%-*-Mode:erlang;coding:utf-8;tab-width:4;c-basic-offset:4;indent-tabs-mode:()-*-
% ex: set ft=erlang fenc=utf-8 sts=4 ts=4 sw=4 et nomod:
%%%
%%%------------------------------------------------------------------------
%%% @doc
%%% ==CloudI Statistics==
%%% Calculate statistics using the online algorithm.
%%%
%%% Philippe Pébay, Timothy B. Terriberry, Hemanth Kolla, Janine Bennett.
%%% Formulas for the Computation of Higher-Order Central Moments.
%%% Technical Report SAND2014-17343J, Sandia National Laboratories, 2014.
%%%
%%% Pébay, Philippe.  Formulas for Robust, One-Pass Parallel Computation of
%%% Covariances and Arbitrary-Order Statistical Moments.
%%% Technical Report SAND2008-6212, Sandia National Laboratories, 2008.
%%%
%%% Welford, B. P.. Note on a method for calculating corrected sums of
%%% squares and products. Technometrics vol. 4, no. 3, pp. 419–420, 1962.
%%% @end
%%%
%%% MIT License
%%%
%%% Copyright (c) 2022 Michael Truog <mjtruog at protonmail dot com>
%%%
%%% Permission is hereby granted, free of charge, to any person obtaining a
%%% copy of this software and associated documentation files (the "Software"),
%%% to deal in the Software without restriction, including without limitation
%%% the rights to use, copy, modify, merge, publish, distribute, sublicense,
%%% and/or sell copies of the Software, and to permit persons to whom the
%%% Software is furnished to do so, subject to the following conditions:
%%%
%%% The above copyright notice and this permission notice shall be included in
%%% all copies or substantial portions of the Software.
%%%
%%% THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
%%% IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
%%% FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
%%% AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
%%% LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
%%% FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
%%% DEALINGS IN THE SOFTWARE.
%%%
%%% @author Michael Truog <mjtruog at protonmail dot com>
%%% @copyright 2022 Michael Truog
%%% @version 2.0.5 {@date} {@time}
%%%------------------------------------------------------------------------

-module(cloudi_statistics).
-author('mjtruog at protonmail dot com').

%% external interface
-export([add/2,
         add_from_list/2,
         add_from_list/3,
         count/1,
         describe_distribution/1,
         describe_kurtosis/1,
         describe_skewness/1,
         kurtosis/1,
         mean/1,
         merge/2,
         new/0,
         normal_from_log_normal/1,
         normal_to_log_normal/1,
         skewness/1,
         standard_deviation/1,
         maximum/1,
         minimum/1,
         variance/1]).

-include("cloudi_core_i_constants.hrl").

-record(statistics,
    {
        n = 0 :: non_neg_integer(),
        mean = 0.0 :: float(),
        m2 = 0.0 :: float(),
        m3 = 0.0 :: float(),
        m4 = 0.0 :: float(),
        minimum = ?DBL_MAX :: float(),
        maximum = -?DBL_MAX :: float(),
        cached_kurtosis = undefined :: float() | undefined,
        cached_skewness = undefined :: float() | undefined,
        cached_stddev = undefined :: float() | undefined,
        cached_variance = undefined :: float() | undefined
    }).

%%%------------------------------------------------------------------------
%%% External interface functions
%%%------------------------------------------------------------------------

-type describe_distribution() ::
    normal | uniform | logistic | exponential |
    gamma_family | log_normal_family | undefined.
-type describe_kurtosis() ::
    leptokurtic | platykurtic | mesokurtic | undefined.
-type describe_skewness() ::
    highly_skewed | moderately_skewed | approximately_symmetric | undefined.
-type state() :: #statistics{}.
-export_type([describe_distribution/0,
              describe_kurtosis/0,
              describe_skewness/0,
              state/0]).

%%-------------------------------------------------------------------------
%% @doc
%% ===Add a sample for computing statistics.===
%% @end
%%-------------------------------------------------------------------------

-spec add(X :: number(),
          State :: state()) ->
    state().

add(X, #statistics{n = NOld,
                   mean = Mean,
                   m2 = M2,
                   m3 = M3,
                   m4 = M4,
                   minimum = Minimum,
                   maximum = Maximum} = State)
    when is_number(X) ->
    XAdd = float(X),
    N = NOld + 1,
    Delta = XAdd - Mean,
    DeltaN = Delta / N,
    DeltaN2 = DeltaN * DeltaN,
    Term1 = Delta * DeltaN * NOld,
    MeanNew = Mean + DeltaN,
    M4New = M4 +
            Term1 * DeltaN2 * (N * N - 3 * N + 3) +
            6 * DeltaN2 * M2 -
            4 * DeltaN * M3,
    M3New = M3 +
            Term1 * DeltaN * (N - 2) -
            3 * DeltaN * M2,
    M2New = M2 + Term1,
    MinimumNew = erlang:min(Minimum, XAdd),
    MaximumNew = erlang:max(Maximum, XAdd),
    State#statistics{n = N,
                     mean = MeanNew,
                     m2 = M2New,
                     m3 = M3New,
                     m4 = M4New,
                     minimum = MinimumNew,
                     maximum = MaximumNew,
                     cached_kurtosis = undefined,
                     cached_skewness = undefined,
                     cached_stddev = undefined,
                     cached_variance = undefined}.

%%-------------------------------------------------------------------------
%% @doc
%% ===Add samples from a list for computing statistics.===
%% @end
%%-------------------------------------------------------------------------

-spec add_from_list(L :: list(number()),
                    State :: state()) ->
    state().

add_from_list([], State) ->
    State;
add_from_list([X | L], State) ->
    add_from_list(L, add(X, State)).

%%-------------------------------------------------------------------------
%% @doc
%% ===Add samples from a function mapped on a list for computing statistics.===
%% @end
%%-------------------------------------------------------------------------

-spec add_from_list(F :: fun(),
                    L :: list(number()),
                    State :: state()) ->
    state().

add_from_list(F, L, State) ->
    if
        is_function(F, 1) ->
            add_from_list_1(L, State, F);
        is_function(F, 2) ->
            add_from_list_2(L, State, F);
        is_function(F, 3) ->
            add_from_list_3(L, State, F)
    end.

%%-------------------------------------------------------------------------
%% @doc
%% ===Count of samples previously added.===
%% @end
%%-------------------------------------------------------------------------

-spec count(State :: state()) ->
    non_neg_integer().

count(#statistics{n = N}) ->
    N.

%%-------------------------------------------------------------------------
%% @doc
%% ===Describe the distribution with a guess about its shape.===
%% Only use if the sample count is large enough to cover the distribution.
%% @end
%%-------------------------------------------------------------------------

-spec describe_distribution(State :: state()) ->
    {describe_distribution(), state()}.

describe_distribution(State) ->
    case kurtosis(State) of
        {undefined, _} = Result ->
            Result;
        {Kurtosis, StateNext} ->
            case skewness(StateNext) of
                {undefined, _} = Result ->
                    Result;
                {Skewness, StateNew} ->
                    SkewnessDelta = if
                        Skewness =< -4 orelse Skewness >= 4 ->
                            math:sqrt(abs(Skewness)) * 0.5;
                        true ->
                            0.25
                    end,
                    KurtosisDelta = if
                        Kurtosis >= 9 ->
                            math:sqrt(Kurtosis) * 0.5;
                        true ->
                            0.175
                    end,
                    describe_distribution_0(Skewness,
                                            Skewness - SkewnessDelta,
                                            Skewness + SkewnessDelta,
                                            Kurtosis,
                                            Kurtosis - KurtosisDelta,
                                            Kurtosis + KurtosisDelta,
                                            StateNew)
            end
    end.

%%-------------------------------------------------------------------------
%% @doc
%% ===Describe the excess kurtosis value.===
%% Only use if the sample count is large enough to cover the distribution.
%% @end
%%-------------------------------------------------------------------------

-spec describe_kurtosis(State :: state()) ->
    {describe_kurtosis(), state()}.

describe_kurtosis(State) ->
    case kurtosis(State) of
        {undefined, _} = Result ->
            Result;
        {Kurtosis, StateNew} ->
            KurtosisX10 = round(Kurtosis * 10),
            if
                KurtosisX10 > 0 ->
                    {leptokurtic, StateNew};
                KurtosisX10 < 0 ->
                    {platykurtic, StateNew};
                KurtosisX10 == 0 ->
                    {mesokurtic, StateNew}
            end
    end.

%%-------------------------------------------------------------------------
%% @doc
%% ===Describe the skewness value.===
%% Only use if the sample count is large enough to cover the distribution.
%%
%% (based on)
%% Bulmer, M. G..
%% Principles of Statistics. Dover Publications, 1979.
%% @end
%%-------------------------------------------------------------------------

-spec describe_skewness(State :: state()) ->
    {describe_skewness(), state()}.

describe_skewness(State) ->
    case skewness(State) of
        {undefined, _} = Result ->
            Result;
        {Skewness, StateNew} ->
            if
                Skewness < -1.0 orelse Skewness > 1.0 ->
                    {highly_skewed, StateNew};
                Skewness =< -0.5 orelse Skewness >= 0.5 ->
                    {moderately_skewed, StateNew};
                Skewness > -0.5 andalso Skewness < 0.5 ->
                    {approximately_symmetric, StateNew}
            end
    end.

%%-------------------------------------------------------------------------
%% @doc
%% ===Excess kurtosis of samples previously added.===
%% The minimum value returned is -2.0 (Fisher’s definition).
%% @end
%%-------------------------------------------------------------------------

-spec kurtosis(State :: state()) ->
    {float() | undefined, state()}.

kurtosis(#statistics{n = N,
                     m2 = M2,
                     m4 = M4,
                     cached_kurtosis = undefined} = State) ->
    Denominator = M2 * M2,
    if
        Denominator /= 0.0 ->
            Kurtosis = N * M4 / Denominator - 3.0,
            {Kurtosis, State#statistics{cached_kurtosis = Kurtosis}};
        true ->
            {undefined, State}
    end;
kurtosis(#statistics{cached_kurtosis = Kurtosis} = State) ->
    {Kurtosis, State}.

%%-------------------------------------------------------------------------
%% @doc
%% ===Maximum of samples previously added.===
%% @end
%%-------------------------------------------------------------------------

-spec maximum(State :: state()) ->
    float().

maximum(#statistics{maximum = Maximum}) ->
    Maximum.

%%-------------------------------------------------------------------------
%% @doc
%% ===Mean of samples previously added.===
%% The value returned is the arithmetic mean.
%% @end
%%-------------------------------------------------------------------------

-spec mean(State :: state()) ->
    float().

mean(#statistics{mean = Mean}) ->
    Mean.

%%-------------------------------------------------------------------------
%% @doc
%% ===Merge statistics state.===
%% @end
%%-------------------------------------------------------------------------

-spec merge(StateA :: state(),
            StateB :: state()) ->
    state().

merge(#statistics{} = StateA,
      #statistics{n = 0}) ->
    StateA;
merge(#statistics{n = 0},
      #statistics{} = StateB) ->
    StateB;
merge(#statistics{n = NA,
                  mean = MeanA,
                  m2 = M2A,
                  m3 = M3A,
                  m4 = M4A,
                  minimum = MinimumA,
                  maximum = MaximumA},
      #statistics{n = NB,
                  mean = MeanB,
                  m2 = M2B,
                  m3 = M3B,
                  m4 = M4B,
                  minimum = MinimumB,
                  maximum = MaximumB}) ->
    N = NA + NB,
    Delta = MeanA - MeanB,
    DeltaN = Delta / N,
    DeltaN2 = DeltaN * DeltaN,
    NA2 = NA * NA,
    NB2 = NB * NB,
    N_2 = NA * NB,
    M4 = M4A + M4B +
         DeltaN2 * DeltaN2 * N_2 * (NB * NB2 + NA * NA2) +
         6.0 * (NB2 * M2A + NA2 * M2B) * DeltaN2 +
         4.0 * (NB * M3A - NA * M3B) * DeltaN,
    M3 = M3A + M3B +
         N_2 * (NB - NA) * Delta * DeltaN2 +
         3.0 * (NB * M2A - NA * M2B) * DeltaN,
    M2 = M2A + M2B +
         N_2 * Delta * DeltaN,
    Mean = MeanB + NA * DeltaN,
    Minimum = erlang:min(MinimumA, MinimumB),
    Maximum = erlang:max(MaximumA, MaximumB),
    #statistics{n = N,
                mean = Mean,
                m2 = M2,
                m3 = M3,
                m4 = M4,
                minimum = Minimum,
                maximum = Maximum}.

%%-------------------------------------------------------------------------
%% @doc
%% ===Minimum of samples previously added.===
%% @end
%%-------------------------------------------------------------------------

-spec minimum(State :: state()) ->
    float().

minimum(#statistics{minimum = Minimum}) ->
    Minimum.

%%-------------------------------------------------------------------------
%% @doc
%% ===Create statistics state.===
%% @end
%%-------------------------------------------------------------------------

-spec new() ->
    state().

new() ->
    #statistics{}.

%%-------------------------------------------------------------------------
%% @doc
%% ===Convert the Log-normal distribution mean and standard deviation to Normal distribution mean and standard deviation.===
%% @end
%%-------------------------------------------------------------------------

-spec normal_from_log_normal(State :: state()) ->
    {{NormalMean :: float(),
      NormalStdDev :: float()} | undefined, state()}.

normal_from_log_normal(State) ->
    case variance(State) of
        {undefined, _} = Result ->
            Result;
        {Variance, #statistics{mean = Mean} = StateNew} ->
            NormalMean = math:exp(Mean + Variance / 2),
            NormalStandardDeviation = math:sqrt((math:exp(Variance) - 1) *
                                                math:exp(2 * Mean + Variance)),
            {{NormalMean, NormalStandardDeviation}, StateNew}
    end.

%%-------------------------------------------------------------------------
%% @doc
%% ===Convert the Normal distribution mean and standard deviation to Log-normal distribution mean and standard deviation.===
%% @end
%%-------------------------------------------------------------------------

-spec normal_to_log_normal(State :: state()) ->
    {{LogNormalMean :: float(),
      LogNormalStdDev :: float()} | undefined, state()}.

normal_to_log_normal(State) ->
    case variance(State) of
        {undefined, _} = Result ->
            Result;
        {Variance, #statistics{mean = Mean} = StateNew} ->
            Mean2 = Mean * Mean,
            Sum = Mean2 + Variance,
            LogNormalMean = math:log(Mean2 / math:sqrt(Sum)),
            LogNormalStandardDeviation = math:sqrt(math:log(Sum / Mean2)),
            {{LogNormalMean, LogNormalStandardDeviation}, StateNew}
    end.

%%-------------------------------------------------------------------------
%% @doc
%% ===Sample skewness of samples previously added.===
%% The value returned is the Fisher-Pearson coefficient of skewness.
%% @end
%%-------------------------------------------------------------------------

-spec skewness(State :: state()) ->
    {float() | undefined, state()}.

skewness(#statistics{n = N,
                     m2 = M2,
                     m3 = M3,
                     cached_skewness = undefined} = State) ->
    Denominator = math:pow(M2, 1.5),
    if
        Denominator /= 0.0 ->
            Skewness = math:sqrt(N) * M3 / Denominator,
            {Skewness, State#statistics{cached_skewness = Skewness}};
        true ->
            {undefined, State}
    end;
skewness(#statistics{cached_skewness = Skewness} = State) ->
    {Skewness, State}.

%%-------------------------------------------------------------------------
%% @doc
%% ===Sample standard deviation of samples previously added.===
%% @end
%%-------------------------------------------------------------------------

-spec standard_deviation(State :: state()) ->
    {float() | undefined, state()}.

standard_deviation(#statistics{cached_stddev = undefined} = State) ->
    case variance(State) of
        {undefined, _} = Result ->
            Result;
        {Variance, StateNew} ->
            StandardDeviation = math:sqrt(Variance),
            {StandardDeviation,
             StateNew#statistics{cached_stddev = StandardDeviation}}
    end;
standard_deviation(#statistics{cached_stddev = StandardDeviation} = State) ->
    {StandardDeviation, State}.

%%-------------------------------------------------------------------------
%% @doc
%% ===Sample variance of samples previously added.===
%% @end
%%-------------------------------------------------------------------------

-spec variance(State :: state()) ->
    {float() | undefined, state()}.

variance(#statistics{n = N,
                     m2 = M2,
                     cached_variance = undefined} = State) ->
    if
        N > 1 ->
            Variance = M2 / (N - 1.0),
            {Variance, State#statistics{cached_variance = Variance}};
        true ->
            {undefined, State}
    end;
variance(#statistics{cached_variance = Variance} = State) ->
    {Variance, State}.

%%%------------------------------------------------------------------------
%%% Private functions
%%%------------------------------------------------------------------------

add_from_list_1([X0 | L], State, F) ->
    add_from_list_1(L, add(F(X0), State), F);
add_from_list_1([], State, _) ->
    State.

add_from_list_2([X0, X1 | L], State, F) ->
    add_from_list_2(L, add(F(X0, X1), State), F);
add_from_list_2(_, State, _) ->
    State.

add_from_list_3([X0, X1, X2 | L], State, F) ->
    add_from_list_3(L, add(F(X0, X1, X2), State), F);
add_from_list_3(_, State, _) ->
    State.

describe_distribution_0(Skewness, SkewnessLow, SkewnessHigh,
                        Kurtosis, KurtosisLow, KurtosisHigh, State)
    when SkewnessLow < 0 andalso SkewnessHigh > 0 ->
    if
        KurtosisLow < 0 andalso KurtosisHigh > 0 ->
            {normal, State};
        KurtosisLow < -1.2 andalso KurtosisHigh > -1.2 ->
            {uniform, State};
        KurtosisLow < 1.2 andalso KurtosisHigh > 1.2 ->
            {logistic, State};
        true ->
            describe_distribution_1(Skewness, SkewnessLow, SkewnessHigh,
                                    Kurtosis, KurtosisLow, KurtosisHigh, State)
    end;
describe_distribution_0(Skewness, SkewnessLow, SkewnessHigh,
                        Kurtosis, KurtosisLow, KurtosisHigh, State)
    when SkewnessLow < 2 andalso SkewnessHigh > 2 ->
    if
        KurtosisLow < 6 andalso KurtosisHigh > 6 ->
            {exponential, State};
        true ->
            describe_distribution_1(Skewness, SkewnessLow, SkewnessHigh,
                                    Kurtosis, KurtosisLow, KurtosisHigh, State)
    end;
describe_distribution_0(Skewness, SkewnessLow, SkewnessHigh,
                        Kurtosis, KurtosisLow, KurtosisHigh, State) ->
    describe_distribution_1(Skewness, SkewnessLow, SkewnessHigh,
                            Kurtosis, KurtosisLow, KurtosisHigh, State).

describe_distribution_1(Skewness, SkewnessLow, SkewnessHigh,
                        Kurtosis, KurtosisLow, KurtosisHigh, State) ->
    TestL0 = [],
    TestL1 = skewness_check(TestL0,
                            gamma_family, fun gamma_kurtosis/1,
                            Skewness, SkewnessLow, SkewnessHigh,
                            Kurtosis, KurtosisLow, KurtosisHigh),
    TestLN = skewness_check(TestL1,
                            log_normal_family, fun log_normal_kurtosis/1,
                            Skewness, SkewnessLow, SkewnessHigh,
                            Kurtosis, KurtosisLow, KurtosisHigh),
    case TestLN of
        [{_, TestType} | _] ->
            {TestType, State};
        [] ->
            {undefined, State}
    end.

skewness_check(TestL, TestType, TestF,
               Skewness, SkewnessLow, SkewnessHigh,
               Kurtosis, KurtosisLow, KurtosisHigh)
    when is_function(TestF, 1) ->
    TestKurtosis = TestF(Skewness),
    Passed = if
        KurtosisLow < TestKurtosis andalso
        KurtosisHigh > TestKurtosis ->
            true;
        true ->
            TestKurtosisLow = TestF(SkewnessLow),
            TestKurtosisHigh = TestF(SkewnessHigh),
            TestKurtosisLow < Kurtosis andalso TestKurtosisHigh > Kurtosis
    end,
    if
        Passed =:= true ->
            lists:keymerge(1, TestL,
                           [{abs(TestKurtosis - Kurtosis), TestType}]);
        Passed =:= false ->
            TestL
    end.

gamma_kurtosis(Skewness) ->
    X = 2 / Skewness,
    6 / (X * X).

log_normal_kurtosis(0.0) ->
    0.0;
log_normal_kurtosis(Skewness) ->
    A = log_normal_skewness_a(Skewness),
    A2 = A * A,
    A2 * A2 + 2 * A2 * A + 3 * A2 - 6.

log_normal_skewness_a(Skewness) ->
    % use Halley's method to determine A based on Log-normal skewness
    % A = e^(NormalStandardDeviation^2)
    % (A + 2)^2 * (A - 1) = Skewness^2
    % B = A - 1
    % (B + 3)^2 * B = Skewness^2
    Skewness2 = Skewness * Skewness,
    log_normal_skewness_b(math:pow(Skewness2, 1 / 3) - 1, Skewness2) + 1.

log_normal_skewness_b(B, Skewness2) ->
    B2 = B * B,
    F0B = B2 * B + 6 * B2 + 9 * B - Skewness2,
    F1B = 3 * B2 + 12 * B + 9,
    F2B = 6 * B + 12,
    BE = B - (2 * F0B * F1B) / ((2 * F1B * F1B) - (F0B * F2B)),
    if
        abs(BE - B) > 0.0001 ->
            log_normal_skewness_b(BE, Skewness2);
        true ->
            BE
    end.

-ifdef(TEST).
-include_lib("eunit/include/eunit.hrl").

-include("cloudi_core_i_test.hrl").

module_test_() ->
    {timeout, ?TEST_TIMEOUT, [
        {"random value tests", ?_assertOk(t_random_value())}
    ]}.

% 1000000 random floats in [0.0 .. 1.0] within a constant list
%
% Created with:
% file:write_file("random_value.txt", io_lib:format("~80p", [lists:foldl(fun(_, L) -> [cloudi_x_quickrand:strong_float() | L] end, [], lists:seq(1, 1000000))])).
-define(RANDOM_VALUE_LIST,
[0.7779093128442551,0.5602545451350247,0.7766852376495581,0.9456642438377318,
 0.5990680481881936,0.047728379072267346,0.6586301304420381,
 0.012254783431982719,0.043885545402663206,0.6954762542392063,
 0.6166421525865262,0.08520025290046163,0.32141882605947714,
 0.9491578130339128,0.7676909159419413,0.5978960396563437,0.6586190704559778,
 0.5420784849708833,0.04963360638083392,0.018027253608207117,
 0.1455830378177625,0.739503834577978,0.41401933401844115,0.469807458374774,
 0.1707761434686763,0.5300297684613087,0.31700184862168646,0.556650914981896,
 0.1593775953149703,0.5234266161384629,0.09497197842604355,0.7577930121361439,
 0.8886013869049368,0.4412648984842056,0.9375472907328632,0.10637635755949204,
 0.14426795030317674,0.9823135065282288,0.6029382152183882,0.2810693108319612,
 0.43998476653608765,0.9720183422937151,0.5149035795910012,0.5311192106119298,
 0.5124394032004924,0.6978845298598773,0.3934114310425142,0.5373092586638265,
 0.059783358190530156,0.07756884285044352,0.9957386314231177,
 0.882570998517984,0.5078797318377497,0.8412208077238221,0.4871749049412779,
 0.9995592174447343,0.1568791828252939,0.8628787822911164,0.3415477516597967,
 0.4032277859695923,0.06411823600977717,0.23901126561601682,
 0.3105824516745782,0.6902515751833761,0.7074114037928012,
 0.025794563251372193,0.20619880508770516,0.5923734953402301,
 0.2840757197100313,0.26427712019886473,0.26426316531354443,
 0.5911079984555643,0.017691758494883025,0.4397289571082156,
 0.7570412979766362,0.3775823133273025,0.6361766793648878,0.6670318781988545,
 0.5314225027835107,0.03192846288429796,0.22651278211799175,0.877892965188197,
 0.8726194463387068,0.9123648316112792,0.2768066672235937,0.595341809201966,
 0.2740653058948471,0.03185435785033153,0.37646445900707837,
 0.5349924328326392,0.6741608248476372,0.03616114080831212,0.6092361546918373,
 0.7022250431043517,0.13743850662345802,0.2135675106927023,0.4356453020996648,
 0.6779143434744073,0.3782823870773727,0.32273846640238,0.7679845132019063,
 0.6855852498357049,0.8944193210050789,0.1773360908568562,0.5969208231996094,
 0.7674196813119952,0.8612327554665106,0.9633454231273411,0.2748163940909649,
 0.7697372046826931,0.9670235279831384,0.5770775718718574,
 0.010142738577649468,0.5387892927515198,0.3643238038124109,
 0.8953932201440729,0.44928333323771896,0.3011190797743677,0.7239069761404414,
 0.6301102140745577,0.4640811311441474,0.05919057243840842,
 0.12207872677295406,0.6287918643796546,0.44509409540335276,
 0.36430493010974174,0.18413443348974612,0.9930757471720493,
 0.46421101446686597,0.7797238971278565,0.8621170751071993,0.2395193350246726,
 0.466691721139358,0.9103540418392103,0.23286919729705724,0.932498907871863,
 0.1960346011056231,0.21061446705496956,0.6886362994062323,
 0.20500350033901027,0.5981142711433677,0.9277983913165752,0.8131556221308534,
 0.11766656714939816,0.1318576602081727,0.14621162829439682,
 0.11527930120149432,0.530637493708844,0.38716292917046957,0.4789674105160763,
 0.24420240777032576,0.7784796329281315,0.9715516764866958,
 0.36349060097611163,0.8969167722152707,0.8138422831128489,0.2062317416619549,
 0.7694109773742664,0.922873550903463,0.3527185032846507,0.4126519700291644,
 0.3083532919579509,0.09744427575949932,0.7362834004398989,0.6064814090655164,
 0.7978911118231204,0.8730007860063538,0.03681357993925993,0.8272019251617722,
 0.386988537875818,0.7833710177810391,0.228882004181023,0.09170599980792749,
 0.9587688999271146,0.8280582846971708,0.7515103872127389,0.8655797153742544,
 0.3168758122063825,0.3351959462353694,0.27475771999921017,
 0.10358393694108203,0.03619725665459217,0.9316269468846785,0.818778651141109,
 0.09078932523747041,0.07064145769330277,0.3307245115440781,
 0.1700914199534277,0.2882416257373821,0.5879867883027484,0.3877852266508376,
 0.3313567527596514,0.7282388787526225,0.46483386023131457,0.8613662592701393,
 0.2583608508026922,0.29859328849679767,0.6118891191619168,0.8405205744719131,
 0.3024082278136223,0.13578173724645604,0.31745303362541866,
 0.3650313767195261,0.8416720728328855,0.31250800425274206,0.8454713522747658,
 0.9971870960785074,0.25032613752909416,0.01519416999290557,
 0.18864290501229042,0.9833320575052352,0.44201561665046263,
 0.11021225282366687,0.5700695456390248,0.5527393783717125,0.7402681555288787,
 0.8264798718475758,0.7310631982720819,0.6574595273931094,0.32978422216309555,
 0.04067368967271934,0.04718418623062781,0.6532939410109981,
 0.27077429999318325,0.7755880344103784,0.9733201496239007,0.9094073971025736,
 0.5877791226783844,0.970753419689494,0.018216987962047848,
 0.19746225714231125,0.6778441348163704,0.22735423530880028,
 0.048856521376948625,0.6887645599366226,0.237554389139647,0.2614498261215603,
 0.37637492853414944,0.7601154821055566,0.4010689383320222,0.6000792127324515,
 0.07594190998505668,0.46532155331194924,0.28515870006832833,
 0.09616741396652484,0.05618488126243082,0.31969058992938604,
 0.8684754781653755,0.14934755659666132,0.2203289361028946,0.662795462833963,
 0.342589194977165,0.467192117471113,0.03826683687495802,0.9978827198430674,
 0.7895293996115098,0.478966587496386,0.302027584374843,0.18978613862938998,
 0.13830887853127305,0.934820935278554,0.5016683161725651,0.8379888001964212,
 0.037245619156541565,0.7813917770216804,0.7437868592626928,
 0.5441648052091017,0.08622491840825952,0.8512574003228086,0.9166289643180581,
 0.483066511567411,0.3777704704490774,0.37346613282605257,0.571082997026451,
 0.9673459368689632,0.06894458316427277,0.23929136461337452,
 0.5598853618293599,0.6998892906732472,0.5114689381231587,
 0.026557901673122664,0.5347524622040307,0.3081547643936239,
 0.015435769276428757,0.8176329476266588,0.16538596615669932,
 0.788800452778979,0.8435499155832945,0.46932389768843963,0.25220914424293295,
 0.8138977530616877,0.9459917604366024,0.5737336151981971,0.831983971007256,
 0.7015703756596665,0.3099207624632956,0.08377915446702155,0.3957115217354138,
 0.7587230277572056,0.8588723683564433,0.4362002843605508,0.560017684769747,
 0.827455861906196,0.18768269011986216,0.39624003502810723,
 0.15748282536828861,0.0058733514631624795,0.5643726321890155,
 0.478476121456901,0.6022303105744987,0.17925856541737883,0.4229564787910388,
 0.9745069024142103,0.1938436082623738,0.16348127968108628,0.9820337649814694,
 0.8392368701085365,0.13844452323779588,0.7633504652823117,
 0.26054070662262574,0.4148660713027561,0.7486668568019589,
 0.17729494601538154,0.17400799764158825,0.18171810678205436,
 0.6156756816479004,0.10004143565940704,0.7407021866921751,0.7730193531847235,
 0.3076018046260778,0.8200732607379663,0.6373358111215879,0.3722483098832885,
 0.8393789657843775,0.8629752467912734,0.305608197284047,0.7643129548849076,
 0.4429090661626871,0.7211462476943138,0.8982840558991267,0.01674945908704273,
 0.6536862480895199,0.37373059951131316,0.37807263062163565,
 0.7317259650685073,0.7809586344944642,0.8571279763677611,0.8220997098747821,
 0.031119652454269287,0.020070990732322325,0.3194471360557759,
 0.05506185965774513,0.1236525172850177,0.9545618299318319,
 0.46690313698988173,0.25698014286210336,0.9610196546355358,
 0.3874568768816562,0.045401977962794104,0.7681485336438904,
 0.9659306672301029,0.37779327985867794,0.8860932743409962,0.5169405239635763,
 0.5909254836060864,0.8102188388249656,0.455416619397623,0.5903458146757283,
 0.5440826185945151,0.5948652674751491,0.6193389925229914,0.6987834260988844,
 0.03523086759491867,0.2445412772481309,0.4327673577720367,0.6375166336141491,
 0.35684496531337306,0.15730861662617712,0.8686860491291636,
 0.46394111871201793,0.3580266410511994,0.5611664887451181,0.7123976197338082,
 0.6712819804935639,0.779695200081259,0.9724466567324861,0.771516349015384,
 0.05374843653745809,0.4166358393102464,0.15749781215480396,
 0.08968217484599061,0.4328470245997075,0.6385186627895689,0.9984964460020409,
 0.7665683295338049,0.21743158978172772,0.5713726885251693,0.9512820205946881,
 0.07754432343812112,0.7573680662544326,0.3680737842911843,0.2891063965091112,
 0.06668799416696714,0.875383796511694,0.2503696906231163,0.28249302817643984,
 0.572422345394765,0.9244277852777384,0.8364729147811466,0.28384227036086385,
 0.25937487664164927,0.4274546275732638,0.3870880299930155,0.4724927062135922,
 0.8850948980252675,0.38454507689734685,0.15730778131590828,
 0.05108747329853314,0.9448166900655603,0.21264307022421192,
 0.31701918632668147,0.6901722479281813,0.7243608042055061,0.558582980112106,
 0.9373698833822872,0.22473689637696803,0.4922819595517075,0.4190292338592455,
 0.20920899749402067,0.7729932674806571,0.9719335802911272,0.3847382206064909,
 0.22928923733834516,0.9486724031452457,0.9615759426353356,0.515688200421343,
 0.7768338209951724,0.22079526939332828,0.13447247876993507,
 0.7643870700681306,0.9047062215422603,0.2856717802185622,
 0.027105250824591698,0.18859277338990188,0.47272527568116207,
 0.23487522969140784,0.9957600443160906,0.7135361153999399,0.5492216574120293,
 0.240680148341319,0.5572667411597647,0.5253535472262646,0.5816220875535566,
 0.37227875379961517,0.7876379623014501,0.031204885802302185,
 0.8050541868865623,0.4291388692055135,0.8637403959307876,0.48549398560934753,
 0.9615345361245718,0.859969889553833,0.26458718630827593,0.30094889043201967,
 0.11483624228971157,0.9521592006178067,0.2054109974109355,0.7709353386724744,
 0.7461814670064821,0.4663212620226148,0.4763671390155414,0.21742226093636852,
 0.13973043369184301,0.6725726651403687,0.07357017396464016,
 0.48275805152640505,0.5813982302061522,0.24797851445893793,
 0.7813846486393201,0.14221661259557816,0.24136417876508076,
 0.2002109891110896,0.49016727593759346,0.2778452064031317,
 0.19457807256429438,0.3491888773256847,0.38218848522605875,
 0.9549940624880234,0.9717048669400216,0.12034476806126815,
 0.08615730427819035,0.24334769726620198,0.7284964523837341,
 0.19142444285842164,0.5350055468049097,0.5837981756185618,0.7295283721280964,
 0.41999124131373444,0.07621042411734835,0.6211509925818592,
 0.6062113755230106,0.26278073265417456,0.19339210840225063,
 0.8489809258897852,0.9471293132300513,0.12853929677260356,0.3344589555665173,
 0.23146122897465715,0.7200281723902621,0.7375942371632741,0.8881175326847696,
 0.42814704582091545,0.3707398811869339,0.216238975336018,0.23837392108737576,
 0.47393540309877413,0.6555120210460692,0.7075280938725397,0.6748226870257378,
 0.9245975912774983,0.28548922145871947,0.0288748213619876,0.6332052719019968,
 0.8388304525901324,0.02464830135485796,0.9901385272658668,0.6323058055882025,
 0.8835343941264885,0.4154005725908726,0.45108084665874526,0.5045984322104448,
 0.5577521329137958,0.7558956568384306,0.02681323197409813,0.5844884839269325,
 0.38345407896348904,0.9518036906831304,0.9528638660782389,0.6197093210104874,
 0.5298184762999387,0.22385550926639397,0.859980328454069,0.907703121074487,
 0.9789926596170858,0.03390910519786161,0.1244026228253049,
 0.11637127882626519,0.2737380521144652,0.951589305044545,0.13185577827320005,
 0.46865819953390575,0.252747307552983,0.5289337251419455,0.9475559655285013,
 0.7518316014759677,0.45356698542079854,0.49856308579369213,
 0.49269923195047904,0.07310747665829342,0.05502614051775545,
 0.5705862016635596,0.4920183617780901,0.4561217792510248,0.5824568089653174,
 0.04397102917125084,0.0026421432870232797,0.311327976175641,
 0.35884765914811745,0.22148416045886543,0.31401876987538746,
 0.45469213422558397,0.22598834781091515,0.508118249075337,0.9899102697753464,
 0.30131203486996216,0.9692388915181548,0.7985875649980483,0.889993212523197,
 0.46269249648336763,0.510781091239705,0.4606350115105783,0.1392328428503289,
 0.2752819087087478,0.7017714699172755,0.055899862557731717,
 0.5408597407869726,0.7021437275900243,0.0019557728665539065,
 0.477164837285999,0.06888059466560115,0.911385740668767,0.14356497830725845,
 0.635458319830524,0.31222185229523247,0.5592212016871735,0.6520566466761588,
 0.3569589039110904,0.32876433369824276,0.9988232670854504,0.5747484775689,
 0.8273606949427831,0.6855601899963771,0.5711296397239065,0.8480884855345735,
 0.06980520353282893,0.16808316428811942,0.5817776115119724,
 0.7791620800825529,0.2569368919053039,0.6295717466189346,0.5649410848139454,
 0.7200672739003795,0.05677267487674498,0.03122895600294684,
 0.024192843479205584,0.6721595281604287,0.3273045869255181,
 0.20256221476372316,0.5991661539628248,0.6321413394043097,
 0.24421305379942393,0.8777046184777069,0.3003007306260469,
 0.13206552424454165,0.5723811757144556,0.9977293386845103,0.4190883502507413,
 0.2640471544404881,0.757488071154678,0.23579037482703746,0.38956841594704184,
 0.10221163167551417,0.2746280755887006,0.27538636867905086,
 0.6434659437589569,0.9334270675749022,0.3309971247065777,0.09316958959320121,
 0.1275738657012514,0.8074401470812508,0.28930078485367317,0.0612754216812782,
 0.6008086726900178,0.6735649467936372,0.9311301483305477,0.08040126672543679,
 0.900251809510872,0.395236828872633,0.6165264694617786,0.6822724051527947,
 0.6068053153028141,0.027479258797889372,0.5601795834767612,
 0.26397367491271495,0.7424173688023118,0.8184497870085488,0.31491412114492,
 0.06471176888735697,0.3712645231043624,0.3976179453558052,
 0.33526703357312493,0.14266908721858396,0.47805977181903014,
 0.041374032701516295,0.7823222279418407,0.16615746030061485,
 0.9671442672129735,0.9762055989648175,0.7751708156402647,0.4197960997219863,
 0.21759014401574572,0.5736695448211884,0.4935593963516385,0.7690688318747332,
 0.3323648982847939,0.4578400420004236,0.47546151030572503,0.867331507860872,
 0.7389901153705589,0.8113869851341071,0.6648912574971944,0.5707057805821518,
 0.9698365927428777,0.11885071743642128,0.3071633782836951,0.6178226149182954,
 0.16240370783779745,0.2636729481056638,0.5463867464111821,0.9028482196730931,
 0.8057807023677902,0.09324970906030805,0.18575866242907912,
 0.6237233024965487,0.44438563195051317,0.8041742907887637,0.6945770270923544,
 0.31895128062164846,0.17068696018932772,0.6914255348796193,
 0.9757158182480141,0.3507480007970615,0.03279833747980998,0.7613525338017969,
 0.16377342485979773,0.4449321018769693,0.845424368501244,0.176315303680169,
 0.15154599198641072,0.6124969879945604,0.010591645030471719,
 0.7346065634774468,0.8137671815364957,0.8234976616134946,0.7796126966800342,
 0.6650581098873795,0.2853388077066169,0.585900699577364,0.43584695484239866,
 0.047370159927473954,0.5334825246345734,0.10563080858045293,
 0.3855341918983144,0.4087767754075724,0.9813967871664679,0.43872317215940104,
 0.5739666205274148,0.7651814728705828,0.41530338630341024,0.5455824933153264,
 0.7189472544196674,0.8570116608349926,0.9864204655347489,0.42640135440868177,
 0.8276445546703158,0.115298086106977,0.38180348574702183,0.3211728063784579,
 0.0507543577254973,0.29271631976622414,0.31990444254930805,
 0.10122444386415808,0.21881078932711684,0.8859742955516599,
 0.5846227919877796,0.84323561334968,0.4954101936076657,0.24890062778993327,
 0.07096726013956434,0.3530775474674195,0.47320367120508666,0.152989559684499,
 0.5966278261104119,0.4797280186744718,0.24317451764652487,0.6335988962505429,
 0.042276667477992524,0.9928409279759757,0.48702732268406945,
 0.006228379031011144,0.9251638929167998,0.2913496452181883,
 0.5507772900712877,0.6460926246895344,0.2950501731727063,0.08435179202632215,
 0.8687462087677577,0.7349412057570573,0.8383976310670018,0.513641158186451,
 0.9923767988652126,0.806787079102118,0.9650943811449292,0.5530476005485969,
 0.04782989685206118,0.9536091685126198,0.20443876999768174,
 0.4021464694689565,0.8218922979477794,0.2909313714791132,0.6638596272233316,
 0.23723912167792094,0.09585323553945713,0.8478158109626398,
 0.5906598493194324,0.3396903594817354,0.5236503181008577,0.3212225810707705,
 0.0638030488234147,0.4243920285206113,0.8944723291240168,0.7796926398242939,
 0.9532769157690344,0.70878530088474,0.7625516167614249,0.9788051582350602,
 0.6115397675415152,0.4943722163272012,0.1656865923100186,0.09720384622037137,
 0.149799272964525,0.45602824906820205,0.486995689542449,0.588392494240631,
 0.4543201593482066,0.2784816096599134,0.5949394826250636,0.5163751497733523,
 0.518359028185517,0.6338591600840829,0.3014132512144967,0.7148981173643411,
 0.07955287880359496,0.9515406384063506,0.10858144981244588,0.680835568860436,
 0.3487594939415706,0.27634315220705297,0.6358103335879448,0.8524471024693426,
 0.3586994194678982,0.8288519777833181,0.7413790112144583,0.7431924805191463,
 0.09264697075906636,0.5260822539631013,0.8297381172114878,0.901110333957049,
 0.3750479470517639,0.6886448128150137,0.6529232260890083,
 0.0018868201811931273,0.1261011918189624,0.23962883727747797,
 0.023405794310624728,0.45452728487372474,0.8947120907862829,
 0.3476190005229616,0.46981413703314223,0.7317206541845931,
 0.19721551394146752,0.6452457832254627,0.6235171061773492,0.8787411263935191,
 0.157958905362235,0.6612476333883462,0.4873984230000591,0.8686487706827396,
 0.8058638895365183,0.5765827292501102,0.939809341305232,0.6626910353121783,
 0.2943996927883994,0.9655300962466723,0.7841144299474956,0.5159808260586856,
 0.9540788565211015,0.13386529623863586,0.07142060502454461,
 0.6055997687629389,0.3020748650356164,0.4781251758955667,0.20393939339954537,
 0.929975452062013,0.21646620814761686,0.21649507986216587,0.6072705059928907,
 0.06297449022942603,0.37320106248097906,0.2316435690630797,
 0.5673907711946367,0.6006828606473985,0.05112725236638804,
 0.020768261028182633,9.94053707468856e-4,0.6244010574270137,
 0.14578491355595646,0.536768414951264,0.757544693488143,0.6342642011902204,
 0.6782348555192259,0.18551663054072753,0.9445592959217064,0.748457685227941,
 0.7323963090818224,0.35737277498476927,0.9196115893328662,0.2498707918897094,
 0.5930490118831917,0.0351982827196583,0.36286281057427705,0.9912110474766476,
 0.0030169604380727577,0.3385955343899869,0.07807787322849091,
 0.44907992014339915,0.8255932084239621,0.20612220726064168,
 0.6781232762543777,0.03671555110965208,0.10962010753611906,
 0.17241126265096962,0.3892841223719813,0.8844725232971709,
 0.003467647739658908,0.20560727647539478,0.6022575898992675,
 0.6715159704569288,0.8477526260161963,0.9154003956830503,0.9375540548759855,
 0.20141602224316746,0.3952240266543666,0.21857856367729767,
 0.6490937831411787,0.9023833871095334,0.9911007413984475,0.8116074578906841,
 0.7394536279519529,0.285468298052313,0.9214788418659381,0.750382800366806,
 0.3428018589569616,0.9281158328057499,0.8373976162981219,0.6986129303734261,
 0.09781663665706164,0.14359678887680793,0.7679713484883304,
 0.5112687892025679,0.606326632441829,0.8536547318295957,0.10110809507619423,
 0.8676141385831574,0.6932152498980273,0.7391687087626958,
 0.055915891648776084,0.1606579845143773,0.6228876965955136,
 0.15607783261260455,0.5779942788242578,0.6808735751632727,
 0.16353645903213154,0.5481649808114368,0.44747313453113646,
 0.10033519267868629,0.10665549718591016,0.8459762170314429,
 0.9865584210127847,0.16581028803392472,0.9668692661455909,0.2649661894579649,
 0.8046901980780264,0.26432244907459335,0.7063300750299039,0.5471410061549109,
 0.18510965965331794,0.8588605122815733,0.27364246803438375,
 0.14779237316170324,0.8501656067598508,0.115185179985141,
 0.013613499451959887,0.8729288906742408,0.8871748719034388,
 0.22037405214283745,0.5634353669994517,0.6845572955452786,
 0.05051706325517702,0.6458274185085086,0.6637348744702057,
 0.05141618788613256,0.4834459159493607,0.15323853380243435,
 0.19513931440766985,0.4532794350787699,0.8278433924435309,
 0.05515578755341355,0.7902725241127535,0.8109409157228799,0.3489173976069907,
 0.34599920471313794,0.7355986831770749,0.686108879431305,0.4779756864625069,
 0.37315886918157526,0.1940643904806727,0.5225724387948694,
 0.05856966159288124,0.6539692961037067,0.6033600147971789,0.8488145308102437,
 0.8049098223368509,0.4493571588647747,0.31544612742641165,
 0.08776372944773025,0.7569813869161343,0.5908365047976087,0.6436045012586948,
 0.103998193598274,0.029923769746480033,0.49093007681808254,
 0.2912092577923101,0.9484281626214606,0.5021108536416353,0.21022301638849272,
 0.26498087397418957,0.5634253776676318,0.22672949383535412,
 0.8640771890205218,0.3960593230248095,0.6885769436925127,0.8036828809739355,
 0.4544167951899575,0.3382680837467278,0.9812182706363769,0.6222797462688938,
 0.3379465902131278,0.798843912660186,0.21107572549914044,0.40148629574259875,
 0.9137096479932836,0.7957149947903427,0.5212181376518503,0.211081791665379,
 0.5291506663003157,0.43036114664625547,0.025619562237734095,
 0.050472751188047615,0.1186935714887064,0.30160746155756646,
 0.6721155367631197,0.17562173028751893,0.22663863685283658,
 0.7165964567455767,0.9987546758798767,0.7430767550350268,0.5727281147416832,
 0.7376686422469576,0.25727966369870425,0.5398505957175307,
 0.04106983753600035,0.8226657138296488,0.3641268647174484,
 0.14621275469148054,0.9635794811634867,0.9634804930246444,0.3681740199314171,
 0.5555852819622458,0.08524304613882316,0.7408972413507147,0.3815729552606968,
 0.051059776438092674,0.17390412109478992,0.9068713906645455,
 0.45694950717418625,0.9493116847627263,0.7995638365867842,0.6103469410626869,
 0.5924977724075032,0.9634105652375011,0.5489961111420579,0.379324605453929,
 0.8208776841923914,0.49052106308839394,0.7419078020903264,0.4412364327488788,
 0.823828648697139,0.806466251044994,0.1326456790246201,0.721379603579134,
 0.3469041473588259,0.010760467562469955,0.5043499044987514,
 7.525345640151393e-4,0.16001077485885706,0.7733301490366833,
 0.4583942067365989,0.9466104984871719,0.2551260426149913,0.688103499729111,
 0.2706557615444608,0.11949540081340548,0.6942990943527723,0.8975468504913916,
 0.9940652271261702,0.6265873962700157,0.28385583216491495,0.7398646963568634,
 0.7507124756314899,0.7749856978968037,0.3856469761134701,0.4936916507604121,
 0.5641295501483776,0.6643824974595858,0.8109985396746092,0.12348576250209964,
 0.9987268533625112,0.08576926171116994,0.02121950680274276,
 0.15484571740770892,0.0307785570146909,0.05807318830436148,0.501918686212289,
 0.6993226441897976,0.004355587644476744,0.5965381781331631,
 0.2895356881020974,0.6710703222536237,0.8331707400254326,0.9607194410829106,
 0.4104956094539369,0.6755239336954146,0.20733427159461215,0.8665655488799919,
 0.6100664442873068,0.41083626478081325,0.3168951492730674,0.5855309206580754,
 0.8995129739008008,0.8945739639072353,0.29341609098752774,0.9997293989675918,
 0.9540508488001961,0.3016175830070318,0.9850533798192905,0.5231233676218567,
 0.7075156459328605,0.7843468658358385,0.9010180516224525,0.12174267093776459,
 0.7461072883509304,0.18147974360922237,0.5454123053779415,
 0.030585542276174715,0.7442866481559038,0.17080272153333287,
 0.07296641618903532,0.4792339575722726,0.841621460737098,0.795464968060157,
 0.7198731961369452,0.5435531321178169,0.7319735173574093,0.2989947016603701,
 0.4550953451502242,0.6008768915023704,0.04200418442640452,0.5843949827610233,
 0.42129218467180296,0.5764853929712875,0.3871724430428479,0.8612642197301007,
 0.37683099444392565,0.5976253632739572,0.7032561982924292,0.71424448558794,
 0.6537348026012515,0.018740942661213178,0.39732711570898516,
 0.07019099635190507,0.014166376672794856,0.4686893637871966,
 0.3532105284331025,0.5391557982127245,0.7494139280974715,0.6219477874303655,
 0.17445174332679392,0.8433112523490784,0.6630826963685711,0.7059319323237369,
 0.6841625386981861,0.4398433028987857,0.5593715719537324,0.46383734983986746,
 0.9859523239494771,0.7604385303652919,0.7090171464033427,0.8879698506472081,
 0.769115494691775,0.6143209861699195,0.42617905822598945,0.4937125429964154,
 0.5675403387867627,0.5374389424660906,0.6081530221637509,0.17387858733188066,
 0.8033314535881187,0.17467128589050618,0.8316168205289641,0.0831828280046153,
 0.9051293935769665,0.3588748490531225,0.9021456393118439,0.9534608121303281,
 0.2860222681531964,0.34367982572381717,0.6097583103246945,0.5742652946387965,
 0.7258950221689326,0.4827641844352524,0.992138963511122,0.017621442199686532,
 0.8878304429999541,0.4873820870346174,0.9915863686553426,0.6359549850517575,
 0.49584117086053725,0.7956727865959854,0.1414946724042454,0.5736958026104827,
 0.9691314416360928,0.7632523035328421,0.48100625938731534,
 0.07332250651907013,0.43044434338441206,0.9671808020418259,
 0.1535472073801133,0.379105676813568,0.04108897392308686,
 2.920664261868433e-4,0.4929902932590259,0.7707097143823393,
 0.7847927730610124,0.18890689834720087,0.453010234612246,0.5510738871331069,
 0.08806089135308326,0.1894888467433028,0.4635098666745239,0.9728537236118414,
 0.4407604747141468,0.7453880774205458,0.3731933188500123,
 0.015652666820231254,0.8918925510985632,0.8783504989692895,
 0.8480117363620363,0.8610620025892506,0.055680500381397824,
 0.06467230565254611,0.65672644376092,0.5385822357769781,0.3983588119788345,
 0.8697710477381027,0.6180070411256926,0.18005094436239266,0.9884419420573,
 0.4247293915346386,0.468558846071236,0.6176737442727543,0.24519455724986394,
 0.3552644336852676,0.25132267377144335,0.8835889893473201,0.2718752786433396,
 0.13191769437468537,0.8754288391017506,0.4437988749776437,
 0.06285055175179699,0.42078006250586664,0.682677213940549,0.7852694492618408,
 0.9387403537307311,0.5859238802285657,0.3038245743043525,0.980372693654247,
 0.6735189112491416,0.9043582665675116,0.6941870400809352,0.10170923323449221,
 0.5343424245981768,0.35508060369373584,0.05545621846021276,
 0.24167530046962438,0.3772412211439723,0.21222439391712933,
 0.3723450190061891,0.7881119182916032,0.17473815605719034,
 0.22696937896087865,0.5943813740892386,0.3159203677067438,
 0.45959136589614036,0.8935675651921827,0.6368549627271737,0.7157115530547021,
 0.008488211998451622,0.8094610158206532,0.393497045060724,
 0.03943849121919685,0.40848740361886793,0.1422842910265134,
 0.33109472022270536,0.3753425046648531,0.5143002741931523,0.8551669306648206,
 0.9444021997933366,0.7279000028615066,0.6960467755282864,0.4210901914069065,
 0.08498066672518123,0.786388816547995,0.02689797081810652,0.3933525260175228,
 0.8344604929287476,0.7677509267012419,8.429991214800037e-4,0.435490508111929,
 0.9863022578397176,0.5063787566925213,0.20726384140607912,0.7564853424281196,
 0.006169254602819141,0.9000021326000632,0.9263973892048618,
 0.22767578710886127,0.7655049041992315,0.03500373644347532,
 0.8641310084793347,0.9935321244768598,0.10776473093784078,
 8.512282964486806e-4,0.8349018696640352,0.8190504691171808,
 0.4566437182079218,0.7715156409924202,0.4356322838595511,0.01279735836747875,
 0.8515713427721363,0.1497781176809826,0.6024455786898839,0.6705845542585263,
 0.6595872869111178,0.6022518638962839,0.23710727247456764,
 0.10328033998209485,0.889608865508509,0.31812009239775074,0.8578365773956439,
 0.6186386873149928,0.4775369026318891,0.35752745143026954,
 0.22829364542673647,0.746749753747096,0.5664214434930293,0.17556087138287935,
 0.17635162516719127,0.09871349589929368,0.7930884921772832,
 0.052981568283753555,0.10583298854002532,0.6014094454027349,0.21379450744007,
 0.9186791146804473,0.41177765951428924,0.1395811015671291,0.4363768662505496,
 0.04427958416046396,0.48303628663510256,0.2901692562305086,
 0.9181088867643079,0.053466593939544005,0.8040249290832288,
 0.8889209312485242,0.5720780589424596,0.8147985672922493,0.3471309679509865,
 0.08527642011148373,0.18007064892356106,0.9277381454937957,
 0.7580620176076848,0.8287657114276108,0.3815265502313264,0.7497418445195475,
 0.5200617703507956,0.39585566625516144,0.5007734364645483,0.26285652997832,
 0.020548341845790907,0.011295360286374678,0.39448182172485813,
 0.3421210350356044,0.22755160725336343,0.11368585358907135,
 0.33597352963012284,0.17809192245197047,0.5117116921496607,
 0.3872085058024578,0.9567724511100578,0.02115973909958946,0.9781026329865204,
 0.14253518972980983,0.8388811972126193,0.04724105055140981,
 0.2849223286536442,0.7316491420532323,0.11838896434804247,0.5775033405622301,
 0.6818241125158087,0.13964075159162048,0.18975770775253864,
 0.7189179746055938,0.8470332595423777,0.6273731411037554,0.5518775493561305,
 0.9029062727947822,0.5960973208893239,0.9493079168335312,0.38190971751266534,
 0.18832215814014075,0.07912106371860428,0.9191732084792708,
 0.21507615326758733,0.9474896190594143,0.9530616249723565,
 0.12574415798202454,0.48215893265417975,0.6911877866835384,
 0.7419278617319175,0.47592043255616445,0.7233190699111015,0.6879998938580374,
 0.10522474487593447,0.5897217403801615,0.3000929055620465,0.8819114367628147,
 0.018792133076280026,0.6813721174384181,0.7257916822104032,
 0.6457028819341669,0.057276455143848826,0.3654768472606331,
 0.0024558119069859607,0.9246643571640947,0.6358179920747714,
 0.08723068537891843,0.9838363913332069,0.40379929795249003,0.531420250385664,
 0.43640010449146693,0.8126552618199121,0.6700561400427962,
 0.16418001615829136,0.4077454702823753,0.0010146086114249941,
 0.03648120507132968,0.6977405506160161,0.6976106257356571,0.9131565867923694,
 0.03815051719577145,0.9987872279082263,0.8152345891051509,0.9330921246679796,
 0.5795637368084495,0.9469578816541832,0.779062993818723,0.34403433281949547,
 0.5168601669293446,0.5246967157370633,0.47876221868599333,0.6827814143926125,
 0.25862211810131663,0.9295818173499949,0.37539259653572343,
 0.8883301159749187,0.1796477362989095,0.513560788369691,0.46334811538953924,
 0.12702103501971662,0.9093530678002574,0.48184287406679316,
 0.7995944883166151,0.41509454880857677,0.09342892445946227,
 0.6518364600678122,0.8998970272170104,0.041574769149342106,
 0.4666666255533225,0.37434583132178756,0.40248591247556964,
 0.9972704694578859,0.4657407136401145,0.019329959898991333,
 0.48964090783007386,0.6224953903367471,0.23751991200068723,0.213635261370822,
 0.694410844704603,0.48224781983347187,0.04446104490154257,0.5301130552596461,
 0.8975015687177504,0.19347464942555292,0.3723663850493615,0.2566548503287439,
 0.4508112907508701,0.459750330040674,0.5422315859884818,0.39053927201002536,
 0.08877495991930717,0.7419324644992674,0.9113615201484268,0.1061356152469749,
 0.7670685414353696,0.41502432600449624,0.5611912572450383,0.6199649462711897,
 0.7504155539281793,0.7690059169676007,0.8143009096367982,0.7054267461237046,
 0.7235650600358827,0.8324962219493797,0.9740530145488449,0.28328034133251945,
 0.9158491053305233,0.18463331394541072,0.2919693998849888,
 0.33283485331486484,0.5815334458819587,0.6974846927809027,0.9661553190432514,
 0.09962887728054104,0.43514973768547394,0.9367695121640605,
 0.48510072640258495,0.3450505866563862,0.11079997083409043,
 0.8022102413811222,0.3204433026589151,0.1372081485915677,0.06102647000231143,
 0.38044382381046693,0.013793586902258848,0.22927363363419828,
 0.3560013626921469,0.03893645742388019,0.5142748303558826,0.7195736221966792,
 0.6686820296331873,0.020052158871865733,0.8486273542273313,
 0.17716597200615092,0.13798287170800505,0.803907959027804,
 0.07001598180298785,0.35241540914221803,0.26390182733724565,
 0.16723722107626104,0.968310601706549,0.9878736576302678,0.5768446968414367,
 0.6242081986875458,0.6812171255545745,0.5790783434267662,0.5033114679557409,
 0.6795513674394913,0.25503013791406803,0.09865251190606317,
 0.9609297711065476,0.5739422677247988,0.7253892560533214,0.7426617255042343,
 0.6528492663175403,0.5255648651579299,0.11410718756268656,0.1528369639018562,
 0.8339596462568182,0.9759693358508277,0.434814160225396,0.3470904716858588,
 0.040941752127884645,0.838959600452702,0.21905296073523184,
 0.036601993330612315,0.1426057221214564,0.7550141049576864,
 0.49162285659453064,0.10631309694873847,0.6144135182301561,
 0.8236012536089881,0.9604750648468062,0.9613611637323352,0.8258541972549446,
 0.978617015329941,0.47056795103977633,0.6113880587112385,0.5808649386315724,
 0.7551943696300535,0.16129494325519678,0.9227332879736766,0.4098286612218389,
 0.1704859366692576,0.5938654791098591,0.8074264386344456,0.9960757215849649,
 0.6188279155704172,0.7259423451760653,0.5219820989242233,0.6751185017482823,
 0.7585341028407585,0.15340024870828517,0.6218438382927364,0.6638285105830418,
 0.7602130058069642,0.3201403375789972,0.9506657816089811,0.9113351454076455,
 0.7292082237912826,0.2571318592532287,0.5050229387174838,0.1776473488229413,
 0.7207360933443279,0.762772949689119,0.4414768097821632,0.9370979253919172,
 0.9309617695292679,0.6206054463667002,0.6026330479267391,0.37479384702341967,
 0.2835552367668728,0.023808306996965345,0.8260843563347893,0.291638135319239,
 0.6769661541965754,0.5325983455429751,0.44185729774084725,
 0.41717618390886924,0.4288838273649417,0.06782742460419511,
 0.5991186410784612,0.3760471238339448,0.3252315676237799,0.5145379997654018,
 0.5832980315304753,0.40154345992290486,0.30783110778897504,
 0.04625606557797546,0.025886885478120902,0.08037880031459421,
 0.5281206780351962,0.6730149507683237,0.6438937520345729,0.4737862545298931,
 0.00914740011854398,0.15821352390014798,0.8187307358436705,
 0.34953504408354497,0.2584217714031021,0.844894655897748,0.38615558477633793,
 0.04526173243374232,0.7379842143827997,0.7792674978556235,0.4476033004715281,
 0.975633867658629,0.9663254730096563,0.21135730928114327,0.8168389833998378,
 0.6686407852434257,0.3479816458590854,0.6835955577328057,0.6039238644393805,
 0.6264566443269334,0.17227455045802798,0.8518791430173275,0.6457826638354178,
 0.8087583145884226,0.07297266020057824,0.7587031805749251,0.8220954216043141,
 0.7070255594967559,0.6067201765865586,0.2766654048669419,0.7169544441863591,
 0.6070936675266119,0.5311600542178823,0.7313716032168975,0.7131282150682523,
 0.03083522256925686,0.9170384115938454,0.254515874416112,0.09036349958634249,
 0.6850793481948678,0.8604524762941329,0.7526854215329832,0.7250398450922778,
 0.7301561146718515,0.06777511931422608,0.31969201489045374,
 0.5512617777260067,0.5243419595253687,0.7921341216688057,0.8131684254193842,
 0.9016874014099894,0.9597466933603229,0.025600107494485824,
 0.7999430038342167,0.5829340405435217,0.6574327271204577,0.06420004871793905,
 0.6057671158000745,0.7957641691175071,0.47715112608003196,0.7877847667312861,
 0.20000863663354274,0.22346527151181672,0.5363989603742038,
 0.6989425374186939,0.9355903690676451,0.18836944596115102,0.8042207270173041,
 0.14368893250446457,0.5208620550557922,0.49936180843975786,
 0.5245832838877759,0.7687522348256486,0.9544112203065394,0.7957870020172794,
 0.5712208212601881,0.34632771656948635,0.7257026952349337,0.9716026853085048,
 0.9558884581017879,0.019825573045095246,0.6041653711958561,
 0.6786904753401971,0.5441764643111152,0.9088875751899845,
 0.024994735865109696,0.2391238920165637,0.33545335326937165,
 0.02782908959413144,0.9948891171206676,0.35676518843265925,
 0.4558161239539862,0.02038230500718119,0.4713585264753858,
 0.32469497098566646,0.5488380087877213,0.34960992864807416,
 0.5098895282992439,0.01644027185159358,0.342860788808916,0.8407278320008614,
 0.4489599953353627,0.9640545345904742,0.23193816846964954,
 0.39464465264188686,0.9973610236504163,0.7528010785966818,
 0.10370272914310519,0.9215582899061239,0.5997561893416254,0.9854952965731901,
 0.7204171764799409,0.99550223514225,0.04925847673895378,0.5736309699877404,
 0.3034438150012502,0.291529166749379,0.4671666566166349,0.6908312084793077,
 0.5608258088518278,0.7108169586066981,0.6050976235314944,
 0.0032151438978981206,0.8935783041813422,0.7737805073781577,
 0.14771664363056702,0.567675838242837,0.5903410582221648,0.20566172960618256,
 0.9228774556385145,0.8487085820298406,0.9127396098349877,0.9362455550263253,
 0.9241826214541493,0.8722961557884295,0.8400821252724612,0.7628788828524932,
 0.5525450408440619,0.39062495068212455,0.3173681182163599,0.7199802609513716,
 0.9209334445016206,0.9763538034129741,0.7992343291121753,0.9000649783061865,
 0.154230443327513,0.019667245587794646,0.4256861031834611,0.510362571891485,
 0.6412702210289545,0.8064978388695978,0.6894830442226446,0.32112157117493534,
 0.06294967459279999,0.9152630406447976,0.6672669373412644,0.3991484460009087,
 0.6231294618901773,0.15926403126089284,0.9103097647027452,0.512309810931601,
 0.4118141417549366,0.8315765336488192,0.2541334149285789,0.9455193984573614,
 0.2685756502171476,0.49320532190510624,0.10560737052579361,
 0.46763087720903473,0.18266859817111325,0.8532541505719812,
 0.14433261730472124,0.27038601002651297,0.7052173199978622,
 0.16713132879049875,0.45512203029724263,0.6516338154431807,
 0.22144550691571063,0.23319337004285612,0.34259058777605855,
 0.06232945282024793,0.20791690181702882,0.9584233750482634,
 0.3544331450496436,0.7209692583692587,0.7448628356050275,0.5097350377339781,
 0.9883829791325452,0.7815492785085819,0.3939151331364814,0.0470935508158149,
 0.09044024989314114,0.7546309589889639,0.08317912579772013,
 0.34539093246776176,0.09939566962710311,0.9629366911739325,
 0.3785630238638231,0.9929873702518154,0.40799107837339377,
 0.14357543875435785,0.6031254543233688,0.11525781770221988,
 0.016877483841810093,0.7774779734131906,0.8565894733955206,0.612515795031164,
 0.9534675301585925,0.724956157594881,0.9997441073181189,0.5853250706087559,
 0.027248583173835716,0.5177846316740661,0.8952827048247685,
 0.5470464176046346,0.3571112192548581,0.18016509884461973,0.9196041032582006,
 0.032798615484347815,0.7069585217283894,0.5532457850683499,
 0.5917656227123884,0.9220907795476233,0.07256908212211666,
 0.11015729304106348,0.9956515508562779,0.845362975690883,0.23906295741316042,
 0.8317205479579415,0.6779112180674098,0.010829223473371341,
 0.4466747014386716,0.5814849759171203,0.21660706532210583,0.6339474333496202,
 0.41130839623677784,0.972903136783346,0.7727901267338961,0.6343515009309364,
 0.804151571930018,0.027777027369294216,0.23089283712403308,
 0.6438518792738966,0.818355289674798,0.38016193701891776,0.2314517831628936,
 0.9366142671799157,0.15312432412565968,0.7105607058691008,0.5575233267277645,
 0.7032567282560928,0.23640397753311737,0.3480324239233161,
 0.016270789971403965,0.046735079850371175,0.7502122545998315,
 0.45457121621892027,0.004941995490791995,0.9916456765483307,
 0.8117464223572796,0.00466046210327109,0.5083564306241314,0.9278518054857294,
 0.1843104588574307,0.3779065971916704,0.3399877080310949,0.1284323187379276,
 0.24485672212981546,0.40127314517602464,0.38074191353721165,
 0.7834343200137324,0.21220480498386285,0.614513217003862,0.9581914231227916,
 0.27228058945889644,0.9454348784209357,0.030756618077907638,
 0.7372691592788352,0.13325078731755757,0.07680431775659757,
 0.07492391264028309,0.20005247745827182,0.9184146561772689,
 0.5457371069944377,0.3153781635019943,0.1948431270985017,0.06119576914717406,
 0.9770124254812357,0.10517292425434821,0.21055081350903226,
 0.2885453508545772,0.5022123668723274,0.7039582565000567,0.8423605264769483,
 0.0074775867470103785,0.007409517975296542,0.9925303800081756,
 0.9151981301123753,0.18523393457873194,0.7783975231851518,0.0881360802122163,
 0.37633784878183285,0.4533030672351195,0.3729634664212822,
 0.24959386187897303,0.6751384920768769,0.6223046002622433,0.9314940707400056,
 0.5217172746786541,0.08204778504844634,0.1679379152811925,0.2032163657424232,
 0.847972083266893,0.27126198907694354,0.977583798592346,0.6777418102656126,
 0.3200567396585524,0.058359538664811694,0.6545044814638421,
 0.5446271025171745,0.6853945890892901,0.13202231934763997,0.7778661759001853,
 0.9529099878656437,0.21115534854353424,0.352980666039894,0.23351926280749757,
 0.21592726913293714,0.9721810126222349,0.07410260386736622,
 0.12118454909987264,0.46625276367134283,0.6732771596119175,
 0.5201417537661861,0.04015592129864898,0.6976841391813021,
 0.29633264501074574,0.13798841301709963,0.38672263989732125,
 0.021907780246142594,0.4266728871460673,0.2841820455818599,0.834540915072657,
 0.9933298748994545,0.5242525645873817,0.3615496290939245,0.960071243805096,
 0.14300249647150898,0.6507148885599088,0.05731833662443242,
 0.9261038265128251,0.7528621565762424,0.8098868902679714,0.4330544043050336,
 0.03666477732422613,0.4196015768079775,0.4198460392891581,0.282089574390087,
 0.9891606870121845,0.42342634085387687,0.6236171273178529,0.7662078171562846,
 0.5837260197997827,0.9046755682653281,0.5592051971448853,0.5620860630828025,
 0.5400623371981111,0.7497457808491296,0.7177408355706955,0.6373313586534993,
 0.3160843147230865,0.22229238753316127,0.6844447047268658,0.3411916397595144,
 0.9332066355505761,0.7436691377680177,0.5515475250034276,0.8678346489683822,
 0.9035003444691823,0.9785671389164025,0.29426760040192734,0.6662861270322016,
 0.6396428818834713,0.03770224155352053,0.4027685160554346,0.7420329654979363,
 0.7757416124234976,0.8104222390825169,0.9761275872029473,0.8627806795675131,
 0.020738257076408084,0.9734014360173391,0.49117400616401385,
 0.9013021055222725,0.8664712323233266,0.29766950165840966,0.7691269943287481,
 0.10822265112058682,0.5472683783296761,0.8820880407442566,0.5499625230850104,
 0.26956137405557534,0.1274950908314787,0.48971996978739807,
 0.2412744936939486,0.39155268709547886,0.470783702550973,0.2197940101082434,
 0.2806319251703192,0.4293473513848066,0.09187900601330445,
 0.09295481385204074,0.18652220689944832,0.3784173895257299,
 0.012854726389769588,0.9255601340493468,0.08095946367216367,
 0.42375542118171394,0.6000829185208062,0.03499116687045656,
 0.03112017734967376,0.42246074773476694,0.2036104103245694,
 0.9215011080198333,0.8635042236575143,0.9359489825331774,0.6427440036120085,
 0.6729501974913803,0.6140816588322904,0.20726902281297033,0.6059160043079947,
 0.382122132144213,0.6744719472462287,0.6242736947590654,0.9769283398775628,
 0.7200227326675788,0.37425968891338246,0.39793229831613486,
 0.9303596227172981,0.3288750069991321,0.8890784259795868,0.7558581328281109,
 0.39981894529791817,0.13603880927329337,0.8594571669047485,
 0.7418044140779914,0.6241368990914464,0.8998087806931162,0.10260506404223979,
 0.816199925671256,0.6429374282710572,0.31635484459958607,0.7335158558281711,
 0.11687913992636145,0.2464556621478754,0.7089934540562827,0.8526496208895837,
 0.6966968520755423,0.5889363781572805,0.9989769915988836,0.44344122893778093,
 0.6556135958441663,0.9649396988417023,0.21239961902310522,0.2034040450548561,
 0.7759850021614135,0.3409216850102518,0.40701951869406894,
 0.41256569882841165,0.08781975608054149,0.5281896379756544,
 0.39547843500938196,0.8165415355517404,0.6885910948730973,
 0.08284713616864614,0.10134828109172744,0.4360432402343235,
 0.16015308506301384,0.3536365581870582,0.04634858514377316,
 0.6943855584413359,0.2884171100214862,0.8112470957288411,0.3863625923314852,
 0.6617000736868592,0.8160709119183399,0.300594011938633,0.3918629392042009,
 0.9814885643534238,0.2964173610142208,0.022485152567660482,
 0.7328511255456817,0.18094741476535214,0.12059079593295208,
 0.3238772525615059,0.40951887961407096,0.1230215643676289,0.6708162949518065,
 0.9533790412839649,0.0568034612728332,0.112955207618893,0.5657688469407329,
 0.8219399045647682,0.5591176310294451,0.4663109788628039,0.23013459789713242,
 0.2552542706152977,0.2969843076396609,0.8535703201745689,
 0.018127324859936955,0.4048703312019789,0.37229478147719386,
 0.6832357760602952,0.6698961235294625,0.5817544890844186,0.44587921832634503,
 0.6547052010862756,8.649733163273998e-4,0.3964571486218196,
 0.8153330084354782,0.09283330129589051,0.48661203083652727,
 0.2986145283558883,0.39589694367512984,0.23709111024752583,
 0.007611556172853473,0.8711986013974005,0.6755908747866137,
 0.20743127889159596,0.10489297273964304,0.5926136430095368,
 0.3487412742410897,0.6378630481377363,0.4737352626358401,0.8073373206019302,
 0.841550783098928,0.9193424652177868,0.9043985443753672,0.4716707732049914,
 0.8538485555248653,0.9670513032479796,0.7979915374845981,0.8300093304579368,
 0.6086645013604423,0.8861726513887661,0.279924105197471,0.9031343513185343,
 0.7356401832231386,0.2982306618131336,0.6257256122953037,0.9179769106125624,
 0.5172159970121337,0.3851900497722178,0.32632532407207737,0.5998463002557546,
 0.17017275084064254,0.5031439271873663,0.2885468359305433,0.2799400754470449,
 0.3403794780010887,0.13171318165819657,0.35055182569972176,
 0.5397460816525753,0.2078313894471826,0.8115662678433647,0.8867376785548451,
 0.08969090724041417,0.19016120019824967,0.9493912246151717,
 0.3304362314189001,0.5070387206142284,0.19617651275802606,
 0.18172416648199885,0.5688181719615892,0.5283917203189294,0.0671484602023451,
 0.4316997026430822,0.17922289460311913,0.007938323354670818,
 0.3860578794692746,0.3765610694970519,0.3450393778765082,0.7482838088519699,
 0.6592291971664127,0.5205792304676322,0.22313655366414276,0.9849415960208562,
 0.3820867624185873,0.14889748421749072,0.43653803705331795,
 0.3508714976223132,0.7590150550819215,0.6737828890473678,0.3457522258247766,
 0.03001569537327231,0.260905190366465,0.9983892913345255,0.6165707162820767,
 0.37179735967958705,0.1960696496509955,0.9802691480811884,0.3326227913657359,
 0.3100822276807912,0.8328284497006708,0.9706630047099408,0.6884930120440305,
 0.438839846534804,0.5111252544344402,0.5347825794528509,0.716981914138431,
 0.20450651094703376,0.08670306464825428,0.13045055934879446,
 0.8589331539318354,0.602881694806983,0.19117975553453836,0.7221824033793449,
 0.17644167410039513,0.3482535978633041,0.20018910593855443,
 0.25481160060402697,0.056458276012027575,0.13439784731329651,
 0.672242343315165,0.011591372834297897,0.9842064123107367,
 0.06877437114826501,0.5477076770498434,0.8194152890147893,0.2153476436240913,
 0.9911111825285313,0.5157663320538464,0.4263670155163275,0.8769606975920408,
 0.2264623813077028,0.6221707114411923,0.41212770715016267,0.5385121728570478,
 0.43843663179953885,0.7529690162163737,0.5818238230899324,0.9664520594856311,
 0.1840362518969444,0.3341082055658744,0.08076432196940608,
 0.08799608847470286,0.23619754000919746,0.6854600778917385,
 0.5601968317308996,0.9858021563935975,0.45293111227870975,0.1856978328582105,
 0.7431139263413069,0.1540961316653532,0.0023431757806433584,
 0.25760815217078703,0.8157901585415694,0.7689934854770747,0.3575997757929811,
 0.41846489284809574,0.23430318063973865,0.15326142519774077,
 0.3394881773456454,0.7490529924661297,0.22667437723338346,
 0.05617721933534425,0.6575596090166258,0.7507071549142659,0.8390161358164112,
 0.29063404054685515,0.9407594483732383,0.29040204126379354,
 0.3550491079973077,0.13589408737908182,0.15026688371465613,
 0.9140468790509653,0.050791345488381934,0.6198465469640928,
 0.28291625601305814,0.552218553299532,0.33852522217095016,0.744271451989498,
 0.03860745725719428,0.01699630309929323,0.5235361033042365,
 0.7795386730846134,0.05504876833501149,0.8158786683186068,0.7398632698203733,
 0.6279967798720958,0.07720395267355507,0.33126441093899384,
 0.5509724258237259,0.7058606309447716,0.0038444254510754394,
 0.17483316980432406,0.6874203348731668,0.05661990823247265,
 0.35648765903251345,0.11211126705351604,0.9630573834792194,
 0.30012159224408175,0.34286591533485933,0.692990687285003,0.9997838194404254,
 0.8379733928506948,0.7648290121436033,0.23129275381517256,
 0.17119287401630556,0.3820584910225293,0.5144642802948789,0.3671231519532714,
 0.6060701292101008,0.032227979377729254,0.9917597458109961,
 0.18924430212689713,0.9323245783766935,0.6174300354619596,0.7849162714178695,
 0.04040895437316028,0.4725586622349357,0.3183892363912565,
 0.12493304554386331,0.017000076425080612,0.22168250543026413,
 0.9439314717788909,0.11037885661553426,0.16639347258758097,
 0.20931121281674459,0.9719749900151389,0.41423596947509633,
 0.30472949097236846,0.4564313624634292,0.002096640390237381,
 0.6065501132640276,0.05533673049533949,0.3257019652442671,
 0.19553593742418363,0.7595998633848644,0.4140862278276439,0.3196382663542009,
 0.8765908508779148,0.2281961598090696,0.11731521958876712,0.9120823819461915,
 0.7359930086749294,0.6329394420222862,0.6928936967185885,0.9208140350130647,
 0.5607180331802119,0.9945327169805386,0.6770967079450825,0.7211571206531375,
 0.7172723658542137,0.1065947464153072,0.8564831623747675,0.35665792985121814,
 0.6111750457698424,0.006941902214812057,0.315962726064422,0.3013245022401362,
 0.583124244449284,0.05036759490452658,0.4048883186949813,0.643164869321184,
 0.4014695439804826,0.48312764597566327,0.9463297462295215,
 0.19917788952241267,0.22667754892548464,0.8742674323470612,
 0.13637689093114924,0.18332472274849532,0.11553125414415932,
 0.4337109147569159,0.8315079930855563,0.11521607143068968,0.7588704341341694,
 0.33769209994301064,0.4667398568673755,0.3607391660115903,0.7767227466011567,
 0.03454779927322771,0.5689430935072196,0.042777242248162706,
 0.5309971168856022,0.34393402020625274,0.19211728797984962,
 0.9339511850760653,0.9018565182737428,0.23706574482241527,0.2826534760808731,
 0.20362915624106992,0.11290268651844937,0.9455731413199782,
 0.7871668162772908,0.8339820315139113,0.7039346193957822,0.4362431071724213,
 0.8591654112090991,0.6660470785002319,0.19182092699373454,
 0.28163077064607844,0.10669963475678634,0.22071175649674024,
 0.11908599284834898,0.7233692882807387,0.4336844649003355,0.6485063082922821,
 0.14265229543204694,0.7649187480864771,0.8419744428919077,0.8268182311023361,
 0.567746776707968,0.8097815713936996,0.4926421694527997,0.3543747111939981,
 0.5779056726452712,0.08949770937772183,0.4530690921884434,0.7761761900023283,
 0.10314318089367402,0.547737165966974,0.47439655612421594,0.6695375448611103,
 0.7913892372663967,0.5011746761514285,0.04680364197797826,
 0.31871066200734555,0.9971604935946415,0.6492525553479217,
 0.45523305527302693,0.5834013128151025,0.9343987484651937,0.8749233895606258,
 0.3150530652307266,0.8237612524325901,0.8596192385953161,0.9253258359371561,
 0.9857158881733712,0.6163946156438852,0.9213809761032031,0.5340040528636455,
 0.03990559575924901,0.8127940801532734,0.09074167519939136,
 0.04710268616896707,0.18584706481635616,0.428927018784544,0.6022097295884511,
 0.10448513317839703,0.8326039511193819,0.27231595948494913,
 0.31293466131795933,0.863952416427829,0.04205622136244935,0.638116575644706,
 0.4601712170371336,0.6899092932685942,0.5434534311862039,0.12896156169988804,
 0.7616379676821524,0.9943084842905732,0.019839459709631857,
 0.10941857410189082,0.8541654915596034,0.05438333332598477,
 0.6912451629569816,0.675960666525421,0.19591026954383806,0.2988304324095089,
 0.7756171399032276,0.1551782493574001,0.3736206596354166,0.728145874504948,
 0.2503690959974286,0.9464735323102239,0.01810595335142018,0.7114050900091735,
 0.7638501496901232,0.7060431841480836,0.3194610519913451,0.5318798668830277,
 0.8367892683949569,0.5339024214910957,0.13921515958609432,
 0.06554040438447628,0.7258282321762235,0.2005069665310163,0.5510736957824083,
 0.8602209452568581,0.8550724519514056,0.91103432554188,0.9062487061092559,
 0.922511225009721,0.37564428679812056,0.6705831180547435,0.31331876063108455,
 0.19891643290275385,0.16665648893989826,0.3267508968466156,
 0.9087786616716319,0.3098826876525942,0.23385283280301394,
 0.23883390008375727,0.9273363067719653,0.3518277785382776,0.8221333383264001,
 0.5332602687506414,0.926044712650368,0.33821836837695785,0.6512506915899844,
 0.27984479362556236,0.11083482660424326,0.7764130748573187,
 0.6449376892072478,0.645348158412356,0.009604301935521087,0.6355992611123343,
 0.9430553286820071,0.18335103872179415,0.9359935911176023,0.7489829385515346,
 0.7142578269808593,0.24293734544577483,0.6552943479485627,0.1885059620524273,
 0.019766673346484742,0.5227274090001754,0.6902145457431105,
 0.0015729832527069387,0.1734147296199735,0.3134295979593146,
 0.7111163108650027,0.35183885012547156,0.39447628736246454,
 0.8027350890379147,0.7366618507438997,0.812811610642552,0.8733388162703993,
 0.564394715465732,0.45338292374186673,0.19106254885084084,
 0.00376833602758464,0.03350538579566553,0.4782706138717764,
 0.05744403832401479,0.5844224529976595,0.08762806464065376,
 0.13222843124284955,0.2930408794199785,0.8578528520154979,0.3727189901444373,
 0.39645025530438727,0.5899177435752744,0.2653472868988538,0.8959930714651121,
 0.220609744440606,6.579861371647233e-5,0.3580883291720782,
 0.20035874941488596,0.6794764737831595,0.0072585436290872885,
 0.44485808225294854,0.249505707656151,0.08986416899818461,0.5780719177419347,
 0.7948811318231629,0.6624538388460292,0.061704098991303136,
 0.8459507313625779,0.049351268075042865,0.7770300717512976,
 0.3712810395269336,0.2650383434187189,0.34847659558597055,
 0.09938077547029567,0.4722595981923082,0.15647877607472194,
 0.8501294178880712,0.7372282033746157,0.7440630254643963,0.3495815582927827,
 0.026097242960336553,0.6937971857847575,0.10623351908530343,
 0.746567255659414,0.6841331239335433,0.21520771711993913,0.7396818946670755,
 0.22610016334456873,0.8105958468993572,0.6997741535849135,
 0.006739652829522491,0.9982463591513746,0.41248154109140245,
 0.7675798592003605,0.323849758078561,0.5500246703133198,0.950426085691789,
 0.8327441132592158,0.7442905989270325,0.8177640975192357,0.11155830857982951,
 0.09589329586385975,0.6821457761301051,0.2047056013936347,0.1468671905440625,
 0.3846117122767634,0.4828642965389086,0.29287585749947953,
 0.23057247733223518,0.030915015885528474,0.9526019117981177,
 0.9721512612735678,0.7270601503744476,0.5779676640435352,0.5120542456387464,
 0.3765731282378888,0.2889923253737239,0.32588899475930244,0.7748117979304039,
 0.27151941135601365,0.40484454540296366,0.9822252901424263,
 0.36667830478353713,0.9745804396811243,0.37794854755732166,
 0.26216732755683014,0.40642044711135916,0.190033464702277,0.3598624153804374,
 0.39893676211222595,0.7107717999494979,0.41236659909887274,0.348567701407548,
 0.31565623085133654,0.4240220757107295,0.8157108849667567,0.8337277091135253,
 0.6688710007993558,0.6975079322469299,0.13662465820504577,0.6311856774992621,
 0.17646473956794695,0.17138259410612344,0.056304962365085975,
 0.12687033652368318,0.36287214046571425,0.5443781651151649,
 0.6556107928203551,0.17384834239971092,0.05581263750160581,
 0.6177872254992721,0.18996844583434447,0.23603538071545838,
 0.027289213933171963,0.9159311822588824,0.30275859722669063,
 0.36111742755017806,0.13995486232432275,0.8569742847634382,
 0.29926908022020726,0.9746586233507953,0.23873468108219542,0.320351067803044,
 0.08469489396617758,0.16789943284129072,0.7656884776161446,
 0.9176185708260663,0.7070049194525115,0.7340785773372156,0.9315986937753564,
 0.6215124867615285,0.8656065317351289,0.15545003023307613,
 0.43780654197253766,0.649645588496129,0.43235677344939893,0.6041540056332135,
 0.4927230603038818,0.9291562993913459,0.33805042230096993,0.4858549109609698,
 0.06333476799649307,0.6628073788541007,0.2361886352536522,0.4334472653087956,
 0.31065303554851387,0.17493766611928174,0.5333895795840491,
 0.9005597093700183,0.021127435932123295,0.3996440582441292,
 0.08915835428202701,0.472336493062127,0.9967300000382893,0.35707639571113636,
 0.5288108545895454,0.4980544340482187,0.9744611551036975,0.17066034400528696,
 0.42050767749639295,0.902838674073392,0.159457183920251,0.376693145603198,
 0.48452993312632253,0.13841821327775283,0.0442646850422741,
 0.7791465125934954,0.12806750238988696,0.14230450310427933,
 0.6303555293344237,0.2864070244594594,0.3851606820974972,0.23543606120046312,
 0.3083133151040426,0.032125781707642465,0.6303067475029104,
 0.15837740670025124,0.4049547561520781,0.039519456873495784,
 0.2287219212938072,0.9785222532968866,0.3788731258220477,0.3513594812636238,
 0.24639876284259055,0.12641431520700053,0.38192693673655154,
 0.2778607974876679,0.5507022436770137,0.5895323192650807,0.9961198989733577,
 0.7347764205258434,0.45927182178736203,0.0805424321072642,0.7153494115563802,
 0.9626507984799801,0.7545491157608397,0.36927275782351865,
 0.10857411933143923,0.8287501360767129,0.22776834659320666,
 0.10658375469457726,0.9872202209213653,0.24520135163925338,
 0.21057892071006334,0.3251619936674873,0.3621866976974082,
 0.24576355110214332,0.9079472004483843,0.7157368996083482,0.601033573063202,
 0.04126595290547319,0.8744593145759589,0.1730956298696551,0.8091476469116184,
 0.6956324677546164,0.9721331539019268,0.9647981509579777,0.1206958211586956,
 0.7290279122380703,0.45302294746624383,0.9380665569787257,0.668127832948704,
 0.2513107287693439,0.28328474421036354,0.08818378978768326,
 0.5873897621834436,0.23855610447427666,0.7956195526866012,0.8643904115255541,
 0.8327352408958782,0.841365293957653,0.0032305772759257945,0.500083478605955,
 0.0861290426255592,0.465165857976125,0.3536197219140127,0.9242248809868677,
 0.7329651411724175,0.16912988692178377,0.7980207207532275,
 0.11532689372458205,0.03290547628546858,0.6838961754618107,
 0.6493461036528166,0.8547313457734846,0.42145868400298725,0.5321036902230364,
 0.8695604519682669,0.6033605337679268,0.13741899493441523,
 0.11670701994019239,0.9840174536031518,0.809553319560356,0.6659749498743528,
 0.477701983912582,0.38917940188571165,0.23162313727094508,0.5438102551710486,
 0.8791069641778703,0.4905815395612968,0.058947026042658,0.32345689491743923,
 0.17195243274860927,0.7927503930074905,0.014450009049632673,
 0.8114058578546114,0.2255791634812666,0.872398744756585,0.9149296896505179,
 0.4125867450273819,0.7075738659727204,0.7305767812902619,0.8556209681728064,
 0.5004180416194912,0.9714343609040395,0.6265967174990801,0.5776720752472971,
 0.0749393750661832,0.9818821548454155,0.694963184105218,0.08608926945172835,
 0.8218705130391653,0.6554445052619016,0.803714428194674,0.7391088123788248,
 0.6197208444981092,0.8862864568692383,0.7821137186461299,0.31628503128858954,
 0.6936096330114387,0.41952765069818965,0.6458311887655711,
 0.14317509113655436,0.9434020180472942,0.8820281174483793,0.9614979161503961,
 0.7481997597952031,0.0934244594637379,0.4530025918790713,0.9538094698007222,
 0.6719760893861366,0.4344646949562555,0.2598573305198729,0.2935733728626424,
 0.18126319316055517,0.9881269952402265,0.30740066035680325,
 0.3990520659288296,0.39819206426758313,0.9655043145598095,0.7102720235716745,
 0.6915414460334975,0.11260014388988915,0.7856483665272255,0.8323900957355678,
 0.17147704067351632,0.6534001995094061,0.7551258517169875,0.5001205401961452,
 0.6244017674087851,0.855261844822344,0.8909467348507862,0.1344071185362935,
 0.7385351573514314,0.30432072076195016,0.058030160409709874,
 0.6548950543511418,0.8300490371961177,0.7364529896032543,0.09396081732963213,
 0.013013441527101377,0.09873333388792427,0.62517470152893,0.655660019471064,
 0.35818829415921605,0.45272274360528253,0.12208439113931835,
 0.5926572261359283,0.4026890766106803,0.39377994010733763,0.4788404787557943,
 0.5795482131438202,0.17361055962986505,0.3839539089674918,0.9339261224706007,
 0.29827943618096975,0.38839269070033455,0.39850269818488593,
 0.49517500228306166,0.5060250406286054,0.1862842950992628,
 0.27361968356232924,0.34394655162596466,0.5436254450108948,
 0.10954896421585414,0.45958313723118904,0.15954911872495847,
 0.4008076907394117,0.871386534554289,0.17166171176694744,0.9309819952773842,
 0.0952207108113815,0.6066605980968869,0.38741716757262523,
 0.46456615171390747,0.2581042976705732,0.9601512057516712,0.6739470047109718,
 0.7746521194251182,0.13717335960103527,0.49611681126632967,
 0.027545494177424912,0.630200673210962,0.7258666669092655,0.6598304910774645,
 0.5920070057359863,0.03838718988067258,0.9474599030913071,0.6734421567189217,
 0.4100808530391552,0.9480242633926209,0.3774763038033767,0.02600152289482416,
 0.5370611462484872,0.5565743422821051,0.6103348002789637,0.1396684196757999,
 0.5083230530551499,0.4513023087557476,0.024478780086875762,
 0.20608553710776356,0.9594885948219382,0.7795943056830067,0.7134007285152594,
 0.7306208274100683,0.9076977816371666,0.7272808334893823,0.719945444444273,
 0.2662325261388453,0.6297319501927104,0.8868083593965636,0.8080561502940435,
 0.9176333268958452,0.29379416492448773,0.5590565820093705,0.21508801429432,
 0.4520459664881531,0.6592484521029471,0.13724860865771005,0.5794606381769729,
 0.1430657687542063,0.08433578594597746,0.04428392235826506,
 0.25024514880825655,0.4665599252755336,0.9086596366825616,
 0.27249049033180905,0.3936230726140688,0.34063777183202004,
 0.5556865280938924,0.06638755975145927,0.7007121534160712,0.7835753319986141,
 0.33780821034548936,0.47486377283722414,0.49227374210432495,
 0.4881272835373004,0.5516191160563726,0.8004951847818264,0.862425705851655,
 0.19934489663519506,0.37837928163487267,0.7098335185575447,
 0.7926293439129569,0.9159536946410965,0.2523916234989292,0.1756090322499918,
 0.9125019644612112,0.7734638620176352,0.15807773628670718,
 0.30011597490158526,0.9686739145300158,0.2885271731714142,
 0.44438861866053536,0.04491344809220443,0.33834999822811906,
 0.7319800929783798,0.24563179390905787,0.9139528249742505,
 0.23062208401781914,0.7385896844249538,0.8863071273363602,0.6725749969174561,
 0.020604732243917567,0.9436240073962465,0.920691518451417,0.8247798178442364,
 0.7363912071528301,0.1927448018503447,0.9355949268610543,0.5370432240458964,
 0.4321260152509666,0.972279669359825,0.8932376482203817,0.0956691388998544,
 0.35291186585585577,0.04352783811482819,0.01992598943474755,
 0.2127109774358793,0.34575430305643373,0.01240338455214296,
 0.07171979042461141,0.3194492976747748,0.21112298497856263,
 0.08946653175210995,0.4248103795669902,0.4651996548538673,0.7591769921735316,
 0.24963736300709005,0.4794451527505359,0.5304872072950404,0.9128542857888364,
 0.6606974705797889,0.7276001828170157,0.19186203065548668,
 0.08524821951227546,0.46145941065284835,0.13940924609946648,
 0.3084131158510699,0.4727700321516466,0.8795420320314821,0.5571599668692973,
 0.7753608805434007,0.19106360072134132,0.951009123019258,0.20747805569300892,
 0.11309297309676725,0.21963370685696304,0.49630045578912985,
 0.7191867765010069,0.3953276275237283,0.7818425798978658,0.8433547512660171,
 0.734309891247528,0.3673469933232555,0.23732752914953392,0.8837352073245048,
 0.20169340569679983,0.517990518421118,0.3581410575972206,0.7597042443711884,
 0.16098022955186186,0.5416615600609579,0.5502041259473432,0.6172084726345216,
 0.9738824277570857,0.011887476526427854,0.667572930314826,0.5922463581885723,
 0.7472984257176069,0.12156365046331119,0.08598485314714732,0.9315877667823,
 0.915849117008555,0.9169646791091779,0.30067592571272883,0.6725509218138138,
 0.9075362316767851,0.16829810852471372,0.4891291723408664,0.828521166264389,
 0.8496205406719571,0.3177272090841329,0.7649463135602709,0.6533043205062575,
 0.4277516883049477,0.004044093262660642,0.8239415098031981,
 0.7029314037531852,0.6035426512452554,0.8010754953607157,0.22151508052561253,
 0.577916681518159,0.2226619221252888,0.037591643134692654,0.8286465927471808,
 0.19511690230075607,0.4988304441226129,0.5530594039319044,0.9003736069935857,
 0.7131992509696587,0.8052461034951823,0.3055289441541156,0.11392498752250613,
 0.7459809707138617,0.34612749915366425,0.1354619911838313,0.2918923285204762,
 0.6163717592549102,0.6004549839150126,0.3144171222002239,0.9549159978829581,
 0.7974620292722999,0.7049098818789631,0.6976055273751214,0.5980703694954201,
 0.9595131440879145,0.23532571650076828,0.3528729385753727,
 0.24465017183996673,0.8679844000820327,0.8104450631949424,0.4723391693918326,
 0.8038736820653755,0.7368213612365191,0.17009434936904977,
 0.16063938752582796,0.14677891619382544,0.6713145152345779,
 0.44953600325000365,0.08261412518496991,0.512650833495431,
 0.49399496624847883,0.14001379952222714,0.7756796055800356,0.861570591247869,
 0.6584850814505928,0.10020521185231823,0.24435627040243535,
 0.23348704649288388,0.37532909656471813,0.9335101347607196,
 0.49289761626677564,0.4512716808648486,0.44956968669751507,
 0.23990606766930533,0.9203723765467876,0.719169841519078,0.4939154538918502,
 0.3552337319248574,0.6779077177545371,0.7317441860566316,0.6054909560466989,
 0.8625576315941831,0.2592799918440778,0.38387436821542875,0.861783960050611,
 0.7379036770456958,0.8141320658891021,0.6119210241250459,0.9975337706449783,
 0.4818439335657819,0.8501826289045339,0.11356086416549094,0.4272015280275505,
 0.7674838525152856,0.007081267296088489,0.5405293290663025,
 0.4955623483217477,0.21040288172267463,0.4628605271716698,0.5674771225868473,
 0.2645477553875756,0.08836456748111776,0.5267069140372425,0.8954387984455574,
 0.7036844818921403,0.5501592864140995,0.7692970952881715,0.40773334284544216,
 0.7369273257507624,0.2035352696326902,0.264017680676077,0.9588461307420321,
 0.7635311169705442,0.26797897420789407,0.18766816471159786,
 0.8278349912949042,0.21027739204890394,0.8872255821192981,
 0.44639905053592377,0.5796743214420275,0.9045168844372503,0.1317987624456034,
 0.007051142344185579,0.08716057025048995,0.6395116821822462,
 0.48689433152347106,0.14392727263578553,0.8263241777886553,
 0.7765499860649703,0.3007736729249467,0.46979957759865487,
 0.27525138967779317,0.04670760157930842,0.33830178240040154,
 0.704567366162089,0.07298700761843646,0.4889730330797408,0.46017814134133883,
 0.9331694897895565,0.6582911466265684,0.06720765581410115,
 0.15566603310348337,0.693037336106916,0.8965511915226406,
 0.024607866179908844,0.19430162363936299,0.4128556555406607,
 0.33079367464454523,0.6676603048886633,0.5899591759153401,0.7025101896993055,
 0.3649001280134079,0.8250316731389056,0.9117510966691946,0.0916051829784994,
 0.49586148805225994,0.9605467583517336,0.9910066937674564,0.4713742865363899,
 0.4928608737740444,0.49788286485545474,0.2708580772964946,
 0.35382580065903624,0.5641274810075524,0.27647023291010653,
 0.17444043805957532,0.17649535294784124,0.3588991950218644,0.403994603372387,
 0.7040956475541108,0.023929579274753254,0.08260331025997447,
 0.5097290400725313,0.14262946902034723,0.40572247724924226,
 0.34512376338789896,0.6997049459084469,0.07081413325301911,
 0.32162427046709996,0.45719257527225554,0.8267074082079459,0.1343313522481,
 0.35891076132069233,0.2835395326325183,0.2862323107816349,0.7708426794015086,
 0.8827471378213844,0.10492627182905567,0.6471253762205134,0.6711299052023781,
 0.6972677810727762,0.4706154137850582,0.31098859844790594,0.0315941405723974,
 0.042665737443925766,0.5960839155225597,0.43821151884825493,
 0.16262836399926883,0.3985921704642552,0.4141992319630464,
 0.18427074698942536,0.21677101048780634,0.013190589580910794,
 0.8599392342209501,0.9346246952665019,0.29891429857402485,0.6435754472964716,
 0.10506272932004446,0.23061912917542982,0.3559922750446781,0.960697289267297,
 0.4073229390454177,0.5649410599424409,0.17669730891712965,0.7917778415802253,
 0.5869545172429939,0.39752890500167404,0.640406279435714,0.12955943572597406,
 0.8673410030155481,0.7669127134887149,0.04404890643874726,0.3355994871234681,
 0.25586823020326055,0.737831770796416,0.71532616909167,0.11517786958930842,
 0.31962749106712673,0.14940850514467174,0.9395652149197793,
 0.45237230247637317,0.6571790422028497,0.08132010035811221,
 0.3572030390702645,0.8108981448612234,0.9915409947993001,0.553104198009595,
 0.8123615454148623,0.015751507375793672,0.5037558850425935,
 0.10392487966908548,0.192241578072193,0.3330686292889955,0.11532828818102914,
 0.219752064615527,0.44204645093738293,0.15082590098505133,0.4842352825219387,
 0.02323926597785453,0.19944734266796293,0.5220608607541647,
 0.3736633863742804,0.456634913599387,0.6387236569000653,0.5764479875258126,
 0.22413759137764044,0.9473769916649726,0.8208550130478933,0.7334752880768836,
 0.08695638508610182,0.26984669298816777,0.5483333287780904,
 0.7449564923912461,0.7891027706598972,0.2840050612636933,0.5621334227352801,
 0.730064708850798,0.6560752706116121,0.3094779945669224,0.4920905111681594,
 0.8072982909751848,0.9644803144269422,0.49635903694428574,0.4454229935216638,
 0.07226021203792932,4.65809414305296e-4,0.19382688223823097,
 0.3192934244573238,0.21892879918651786,0.690748442252657,0.8951249779777257,
 0.766173954675615,0.5620705930760218,0.16260082504831275,0.7778979771665144,
 0.6581258651085495,0.09673076654730683,0.5650794236312957,0.4233912460481193,
 0.9996566104094757,0.18832750785298302,0.8375652472361726,0.8622182318889362,
 0.07280296675778697,0.5406487907629696,0.34915165964329664,
 0.20434584096703678,0.31271053064021326,0.4859022539856669,
 0.8491316518192161,0.9507069017103102,0.7280076556910354,0.09001749608197251,
 0.6222861195718211,0.08294264711632171,0.3883780519310853,0.4727352347293058,
 0.06203002159122495,0.7498543359492503,0.24230725596797076,
 0.3726903273568678,0.7841663922885808,0.561329115547922,0.5425762311451253,
 0.40540033831980404,0.8184868331439333,0.28798925323363045,
 0.3256969762764521,0.4726672990409152,0.6248254130829477,0.5816835241913152,
 0.08298930261476711,0.04916268553589154,0.32281517482295996,
 0.39235476528479296,0.9301324923892342,0.01997744878060681,
 0.7614070035039535,0.5287423909658101,0.226409099173273,0.8508853896427093,
 0.8000295722806822,0.5452235007527826,0.8695431457998298,0.2223138369979648,
 0.4103447998329719,0.2522501116038637,0.7094256587735797,0.7147794926968011,
 0.9131835539134912,0.7479403988462477,0.9139152489866723,0.5832255713996157,
 0.886020909895096,0.5872620167208608,0.9336084623117673,0.7156786512833925,
 0.7058404731253618,0.8427750277605722,0.4211605306285101,0.8728008157073648,
 0.5460248077137221,0.09534573903659749,0.787599888313862,0.8149512960647024,
 0.9234547808914806,0.9984392550599808,0.7978641235953763,0.19136102444462044,
 0.4457040711992033,0.7762716326745837,0.4819267021110972,0.9290589824476815,
 0.5945223717139696,0.5454812056014512,0.5760087250963806,0.5962657676420249,
 0.1565613208227542,0.1377263669302502,0.8636378805592436,0.331139927402336,
 0.1721958571555996,0.32923316789598345,0.10677138649358331,
 0.19083720470132048,0.8224474821436877,0.10276437818457707,
 0.7180898543116023,0.6447863928564129,0.2045092844251193,0.9709588009519793,
 0.3636394514290858,0.5588988782392471,0.9096810472584718,
 0.057962695336645775,0.11990752025983775,0.8280764073700799,
 0.5591211787273692,0.49715363744548935,0.2558725824617357,
 0.39827160081536006,0.1369264102120349,0.5140300000659245,0.6453405927470278,
 0.38424595233070724,0.1383154534243658,0.13777096432966918,
 0.4043604851412308,0.25620918438818696,0.4503764552265094,0.6897583788684557,
 0.5304895042101578,0.4903874199894185,0.8632187487666328,0.9877340268311442,
 0.808769774782938,0.3481374537066093,0.8439746435955117,0.6749275411306823,
 0.5027350677250675,0.4232275240218385,0.954214351642756,0.42856915243858174,
 0.4663493079180706,0.39644299026562835,0.2741988760200855,
 0.07896996379714849,0.38930239222614893,0.8877930737813448,
 0.5250389888681467,0.3283237282212461,0.8718669053216159,0.2883855864530471,
 0.4962891528387734,0.8148545094438313,0.6482289378952463,0.8401863059218374,
 0.9883434598081928,0.06576506319472464,0.9306172488509001,0.6202809689801478,
 0.5468696675066698,0.6333642437982562,0.9105485932412097,0.3150215097786495,
 0.15911349665065244,0.7275285886028248,0.5295433413731352,0.9087967264922466,
 0.5427556134540654,0.4096105967173276,0.31432690652138906,0.414438043976617,
 0.9131705360538411,0.2300982925502315,0.8471025465300086,0.7406309305403763,
 0.8107812698606441,0.6818559922358151,0.10467036285674347,
 0.32717388682342974,0.18046260588274998,0.6354896993356138,
 0.8984834008271426,0.31209418331568517,0.18747824326641416,
 0.6400153815827138,0.5386785013895229,0.19407707164333454,0.2857926146892428,
 0.06157394429075891,0.1508482268536172,0.7432256494958371,0.8577147075101247,
 0.3096967980644484,0.5165673331000108,0.3409223120025483,0.5118259234492644,
 0.7397622610762087,0.14122359320614508,0.08452235251714146,
 0.05718094652212646,0.5407043870411883,0.8413738054789475,0.2825481565812913,
 0.29187113200087955,0.30417897730831533,0.7759390144475341,
 0.6392059217197416,0.15732638310533353,0.9882262757393249,0.4295666405736218,
 0.032867491162021256,0.4640257050002782,0.8464541240512388,
 0.4384953678981558,0.013655020665798934,0.08292385202153651,
 0.2088733745903537,0.3671478504661898,0.9310756492833745,0.9599799231996318,
 0.3630469400307711,0.2158089322409782,0.8633880127827905,0.5132986049368007,
 0.5299446845513609,0.7717421362559748,0.7774164368484292,0.03525328300551556,
 0.6208540280021086,0.5570967920146207,0.4811527763618437,0.08125068969468086,
 0.6558360448877751,0.6452166819237002,0.6406715414727563,0.02189203120230576,
 0.42410414223389503,0.654953679629377,0.22902030262802764,0.2867149800720501,
 0.004510370732105473,0.6012618595364843,0.7609909448448868,
 0.2998893344590827,0.12688309540716136,0.3089575069228625,0.2457267727731527,
 0.4407208135699443,0.21764035476991828,0.43534499217947864,
 0.23797159514976984,0.9507086763374579,0.713999153096236,
 0.024783606648586254,0.28642381816165985,0.6487960758320869,
 0.5600238524345209,0.5363353638566727,0.9490089175270586,0.9549923868627973,
 0.3358485981355791,0.6401494318989254,0.6153417532812278,0.8321982181596665,
 0.08234555705023316,0.06475915951538558,0.5765653137384603,
 0.7048298999398995,0.15716368228698385,0.4260978071031023,0.1844050336524895,
 0.4749652677346847,0.08400280162093776,0.6621005022216986,0.9634722572061373,
 0.11896748687470149,0.5830477878255063,0.7651965054957024,0.7044675302153052,
 0.09390243421008948,0.2105452502976961,0.42046612028163544,
 0.3712702231995598,0.470283423853173,0.66680227785849,0.6982872568010565,
 0.11241122795440772,0.9544098819524448,0.7137900530013788,0.513694826644035,
 0.47127988808595933,0.9269684893416602,0.6248307876560558,0.7105248058390143,
 0.1726934106056882,0.2837166769183648,0.9417596515059702,0.52807573043828,
 0.8971587472310958,0.05885851602072589,0.3375444042474097,0.6379562625875375,
 0.49586941884007074,0.5347426838190887,0.27033115306806876,
 0.9103054596190716,0.5033895733668613,0.19043413635511186,0.5118836743874817,
 0.04506582698109929,0.5988091069614653,0.9756435072223689,0.7209218019992778,
 0.5997268722200779,0.7015632346363928,0.6085319935786688,0.8983075231619221,
 0.35964923861772746,0.7336468086641434,0.5027018429520558,0.3404719943126062,
 0.3415907404014261,0.23225470762223,0.6514246785953114,0.5282240147805538,
 0.03212787575682896,0.893716730583382,0.41104807036329716,0.8458505053711173,
 0.20017068631651092,0.14618990022214984,0.09671647655901927,
 0.8137085103972895,0.3379044491862119,0.6067302699169604,0.8991040453831236,
 0.16803059133398568,0.34417591009545256,0.6167222419963935,
 0.26926478531303066,0.13504601797778304,0.3105507282002692,0.647911878884688,
 0.25348193823888177,0.9502557819735518,0.47126383046617937,
 0.7427156708272664,0.10894493608919875,0.29015923740333793,
 0.5347855040279762,0.8839239910569602,0.14372135655882412,
 0.10682865721933221,0.48536352956964235,0.4900998541216183,
 0.9139764506252628,0.3219616825207551,0.5876979705313227,0.5058171675115595,
 0.004713931294495821,0.4992627328063177,0.6616890020609982,
 0.27728790782231716,0.7750405946143688,0.5976195415622674,0.9099785726290394,
 0.37546925515917473,0.3014748512440657,0.5409847150729058,0.7385814150591454,
 0.9163666180215608,0.5443750594744154,0.2863016933595527,0.990757538756628,
 0.2685471870960924,0.01881980202680833,0.22968618546803776,
 0.9658072272235996,0.23016372208242097,0.7904353965338559,
 0.45622180369632026,0.7952479214348762,0.3943931549589148,0.4516419329316026,
 0.015805014250405036,0.9187983028296325,0.730749950436265,
 0.030271462361925905,0.25691719036333494,0.008435291212675944,
 0.6929154176463986,0.9435423710674338,0.7829210259920207,0.9207196398852087,
 0.41518574880629344,0.7736304237337603,0.48905141226390836,
 0.01938197719416146,0.8059539837489015,0.9468124652723563,0.409300120501326,
 0.9965737546067236,0.3604255178215934,0.695831435603904,0.864039531606452,
 0.03641593432930268,0.00396442820687537,0.6799990090948304,
 0.08583924059256809,0.6117432551988705,0.030573819706928185,
 0.06506870011784271,0.034520139093371816,0.4076500292385825,
 0.4540354168055539,0.389509040319756,0.9952423633225315,0.7646889706431669,
 0.5809670015255772,0.20707259633546815,0.36417362387982677,
 0.7651639807741117,0.2879592899882073,0.6350107068149958,0.7531131776028541,
 0.04228861184065891,0.2871158380772578,0.5530385998486921,0.2781548129850444,
 0.1487771423282066,0.9713423439289658,0.19467716036099525,0.921588493516336,
 0.9008298222234309,0.16156675647144347,0.5384983497290589,
 0.07923475840161098,0.24107546680070757,0.7426697949704342,
 0.4738998162440632,0.46537005152044186,0.33140114897120476,0.756550897516361,
 0.5737946873765943,0.7005649978656426,0.7944303230033893,0.9808317114023615,
 0.6208021847259743,0.30446502312716284,0.583215921430407,0.04098932315430126,
 0.6244313062789696,0.36975529776334515,0.8613981780893936,
 0.07545756468666498,0.6332939286967987,0.6589776144277549,0.8794322414162209,
 0.45224848713616084,0.8195890767057662,0.33701966893791313,
 0.3003314666834894,0.572856349931244,0.9358382873857078,0.8066067198655772,
 0.7574466089473565,0.8446429191655019,0.2540781997624709,0.30686233818955955,
 0.7529932864994096,0.8046276732511848,0.17014196528471937,0.6089041881599805,
 0.8360361770220658,0.16461150235676492,0.2722102236035775,
 0.08839165672014082,0.8323209285003748,0.5553874836383939,
 0.33245183124909805,0.10919719259559335,0.16846336239765225,
 0.9061538939738549,0.1202188082146145,0.1846691785893484,0.06931134090615976,
 0.46027545169755046,0.4790878409342352,0.5271906616082322,
 0.42997177132049136,0.8815754297206773,0.5316218574827788,0.8059800538052833,
 0.7552432949479378,0.30816381746870447,0.5306942742399937,
 0.23538571715861623,0.3790935948483606,0.21440389175731867,
 0.7864136185724525,0.11805102976225534,0.41192340833913743,
 0.6215954891663598,0.10880445858161758,0.34310787522905384,
 0.8859151302153813,0.7587360997630076,0.8070614816186099,0.9379513357979657,
 0.4872394572744366,0.7232395790610724,0.24516849300875965,0.281909000428607,
 0.07704541854201441,0.8591832972011985,0.40193097060898086,
 0.9718469792803051,0.9860430300121031,0.058059489108190476,
 0.2281720753983627,0.13227084045012794,0.7169254700852722,0.6687808383180819,
 0.5460420319956816,0.3082657491571943,0.6736377759496304,0.28783964825476405,
 0.2969286808618865,0.46810173437049285,0.5132072819131521,
 0.06910491215476267,0.41721907583724815,0.7735006163007079,
 0.8027005461658839,0.2693353979165466,0.727560986276669,0.19628121580513935,
 0.530396903542559,0.026196250628327422,0.04180367792379136,
 0.8713152417235708,0.6149331633250339,0.21502286600813925,0.8744590773101796,
 0.8239078352478281,0.7124102978651674,0.6419906685317804,0.38872060433862365,
 0.3916954722522181,0.7771495484851535,0.0874031203459823,0.6320870971928675,
 0.8369365925069053,0.26756433376744815,0.19612348030685234,
 0.13435860726413829,0.3456490555598407,0.702701891384239,0.8247134454901838,
 0.7620719475625303,0.7077691893989005,0.45296389388638403,0.6338969644829745,
 0.30620979234099277,0.3595042458715275,0.21388067114023346,
 0.6732539187807066,0.8379472001119969,0.3257792683132228,0.978860488884369,
 0.0035985565478876547,0.7556684119684128,0.23214892802208686,
 0.5757954175121524,0.19676651424385028,0.7990511727607296,0.7202149754438945,
 0.8771866092805364,0.35990423459291054,0.3475621622354509,0.8486551616906343,
 0.2287322882271451,0.3381943533514481,0.08366468782589365,0.5670373776910713,
 0.2638263101311342,0.35608160210939965,0.8010201304331463,0.6050718323458576,
 0.7381458844227297,0.9675106910702037,0.5150990815082893,0.5127476551445465,
 0.27829156418555356,0.8397000169819482,0.8941444486243434,
 0.25686503776291414,0.5980485823894495,0.6251415817106595,0.477722499983921,
 0.8350791268523413,0.9411939612321514,0.6486208063329146,0.05351783336777949,
 0.40058120109242557,0.6477705661935873,0.6820004001362235,
 5.683643905374236e-4,0.38021447851308865,0.5485623043835677,
 0.6811409830750532,0.8370605766719721,0.40915572333173134,0.6380934962027668,
 0.23555516275018462,0.8299927188623564,0.32675693050899335,
 0.15733763313138338,0.1812677423203949,0.8017683000729706,0.7054814196295273,
 0.526884599097003,0.9111628893848718,0.4899992514545669,0.35482794114877125,
 0.2867735536719611,0.607309531617044,0.2865641857237411,0.5734357135670453,
 0.8930089902739373,0.9256706647952198,0.6475743301246388,0.18252615863857136,
 0.0404504098688635,0.6498646583703179,0.0772633747815914,0.17335803300164443,
 0.8413147462486813,0.9227861214822164,0.5832871975709342,0.48608860996139724,
 0.8428677881988956,0.35375570211120155,0.32003792933093833,
 0.16267278015097075,0.14082861732039598,0.9390344870803894,
 0.42147956575453893,0.5051361345953145,0.9719528252875643,0.8835016105161004,
 0.456964477016795,0.49477627888925724,0.4465903428410961,0.2382898908267389,
 0.7849531128363308,0.5344159641037838,0.19933968238613942,0.1235054207858649,
 0.23119905330613288,0.9229580686247276,0.6914507612543103,
 0.47861501778456805,0.22851720755024707,0.5072890142301996,
 0.27251845024596855,0.8488273384225162,0.10696778181563915,
 0.8454260965855201,0.16854798732372844,0.40473483816376365,
 0.6746023559673188,0.2494938040903426,0.64286207693421,0.6061436023731457,
 0.6199268486292507,0.6713330463106763,0.25519507865389246,0.8760461992307216,
 0.7699661823447744,0.8191779402243946,0.31375052682272064,
 0.38083618857502766,0.9763049472506156,0.21018000667862968,
 0.35298532968975704,0.8494674668642346,0.8066233075340675,0.34671127465859,
 0.9371183242365081,0.5340041460221108,0.9936961543384022,0.8819404258366892,
 0.590254066737193,0.7105384187705447,0.11336901672198885,0.772398585956904,
 0.3163344160426429,0.33199052477081603,0.11398501290722707,
 0.5772220516162556,0.0025046574824966195,0.17006556276141005,
 0.5948426717199761,0.4490811613761255,0.4125258016917538,0.4051689407915461,
 0.826913412444153,0.5839443770452614,0.7219558166932448,0.8155730572151512,
 0.5403695276605508,0.08465597298383176,0.02270374391244845,
 0.06499321295343186,0.3645023222996888,0.9255788883922157,0.988617195928424,
 0.4425598173590535,0.2998266882503967,0.6136427103897049,0.9935688020510848,
 0.37778685405270873,0.46068219826530077,0.22751128271319543,
 0.3734365768136022,0.36278323528805656,0.6266539068734444,
 0.45756953589963656,0.9435612142773535,0.14397811129715454,
 0.1437740066217803,0.7680907219110383,0.555986536156396,0.2478856162858316,
 0.9747447425639123,0.3638671428856225,0.3447868070452227,0.4755971891717775,
 0.2846278197682812,0.7354636676022311,0.8285453778126955,0.7240126167211838,
 0.9173090920168034,0.9232012934820901,0.9686432699128653,0.8631874532366469,
 0.43559229812294853,0.5792568869393486,0.6632953570044379,
 0.44685976571347696,0.5540544710695932,0.5287604816965045,0.2855595340934991,
 0.7110926073169367,0.24158691273842248,0.3094771268713026,0.0833532055226317,
 0.13184454308026206,0.9542900118570156,0.04531047890559936,
 0.9634259405424475,0.49013607809554793,0.3797264226942132,0.9187445740094743,
 0.33462645373883804,0.16253116985965566,0.7242552586761659,
 0.26827878963113283,0.8511138093948043,0.8555193285047645,0.4420058048274974,
 0.36416698020897975,0.3496965389788921,0.6321153519118755,0.8691915380394344,
 0.36408921943444317,0.6969036561231354,0.37177513701609355,
 0.7093000985426068,0.8468802286033214,0.3236676134384612,
 0.007012598396875069,0.10085350211039057,0.2825937486379766,
 0.6051980323964333,0.5521135208768585,0.5139704615439317,0.7733135591775046,
 0.6726442326324583,0.5611344170097906,0.7201620154282233,0.09249109085226837,
 0.1607160603621426,0.425478885729204,0.8968417824039749,0.5986147878411311,
 0.6890210140476851,0.9566666177752237,0.3138305630301407,0.48226349441590777,
 0.4652603932844873,0.6750712717342775,0.3195793427700183,0.7084696536776102,
 0.21086052588839088,0.859152345061275,0.7330072327598746,0.8008387227651437,
 0.22820950258921624,0.9056978619775896,0.5623560610610338,0.8071432077021016,
 0.822448195418025,0.1275701741215276,0.9159166716862533,0.9433910994158018,
 0.8592965381959625,0.24571057432442722,0.8473129280438836,
 0.017754474746118465,0.4084503860998826,0.2728248699568375,
 0.8942909942753557,0.4315967375041819,0.9287090632039253,0.4097975754770673,
 0.7551302798806772,0.6064513305977132,0.6688704468716836,0.658395188471139,
 0.1662266286415358,0.04217874673209587,0.15406290809170298,
 0.04099649650958215,0.6887935566456914,0.37327156581620147,
 0.14268234586016715,0.5680741337623799,0.3299162962675607,0.3643230812396412,
 0.4691109861167886,0.8350046815540251,0.3904620208268831,0.7788547560603857,
 0.5244362280218071,0.6479896765214854,0.4179580229105998,0.9521264962972071,
 0.23428342897446552,0.17571754428847053,0.9891116975917681,
 0.7938918012077344,0.24867717043181647,0.5557888371510172,0.5512957689812366,
 0.9136097921114753,0.5469665626096365,0.6498869894077104,0.7360144012274499,
 0.051165817709597494,0.10983786599186118,0.509009034830039,
 0.7000351611217586,0.1399918284112821,0.6727765515243493,0.45626891931812896,
 0.7866281460649148,0.033328236628694,0.023042874171439576,0.4734452129555746,
 0.2542611158810968,0.9297448050701946,0.8708650070647956,0.9148633900522006,
 0.5870865585035302,0.7505049658799291,0.5694740213564443,0.507773476852106,
 0.7007046223771107,0.32882180571425657,0.213566271563573,0.9745970115262504,
 0.15957209303878583,0.7723639172481087,0.876202104573395,0.3410634022445762,
 0.11244370091618694,0.28073203310662964,0.26746023201203417,
 0.247017391729907,0.695964679454364,0.9069061868064947,0.5019890484485342,
 0.7704400763193996,0.6502445163001402,0.519519084039249,0.5970497408643961,
 0.3791842991479413,0.4374143543826575,0.8788512292959532,0.8625139282871223,
 0.0077533896322002915,0.27040455200650304,0.8278846771391158,
 0.312170984993839,0.740946719432606,0.22482075245220934,0.5459270468415238,
 0.5500027880261512,0.5799123358096878,0.9572246292969069,0.3489634778057654,
 0.4049447570293577,0.42371928072208564,0.7480119488853773,0.7412693910754032,
 0.7372561157123229,0.420323346792006,0.6787948542693631,0.3035782748810052,
 0.014602986773433635,0.7861710590390901,0.6994512335666838,
 0.4346271304356276,0.7314070322467169,0.20103780479186673,0.9981102635967269,
 0.011115451246419106,0.08187884122423206,0.5378700970898126,
 0.811124465008924,0.5029594890766684,0.19182926883652718,0.9056766796644171,
 0.9042342033162432,0.4183727624690219,0.09546960150108819,
 0.23927976368907322,0.6849197457436651,0.16987879449249854,
 0.8205044643581364,0.5028845284561417,0.16362814754513189,0.4398476471615316,
 0.5163413430430291,0.05610688086558546,0.23655239184061527,
 0.8007344589088906,0.7259639350947718,0.3567500902362615,0.08265164173016026,
 0.031197270542085986,0.008228584418430884,0.8358320965766777,
 0.6502944315043832,0.5305540896498917,0.25148809078956647,
 0.12774794490960184,0.48051907630474056,0.9476679449492954,
 0.07246686818354242,0.6497908288972215,0.8804510558762164,0.9468356787022015,
 0.5280355037827438,0.7731417723464084,0.41438886702482225,0.8051251400462793,
 0.6682386200081855,0.6573836884567061,0.6952649514876768,
 0.006559644128201869,0.6474280490610681,0.7271528915776988,
 0.18076502246710335,0.9513820475360112,0.2011331675593151,0.949404251250138,
 0.6996486703252369,0.7006466795732619,0.10280872655380335,0.5357434717815169,
 0.7801402637533456,0.6807999169203507,0.37144258942389796,0.3442435730993485,
 0.44310521002579617,0.026083466353620577,0.24202225544445088,
 0.8471652610527658,0.020587267739690995,0.8328586318144038,
 0.9991346721849986,0.37817778017698456,0.0728880288226661,0.8299371578015621,
 0.6731798859131528,0.13244192512166908,0.8467113631998444,0.9444470102426366,
 0.14075050223848595,0.6400402616120254,0.7739839524090971,
 0.23815663650870178,0.16607085508621178,0.8324671027081344,
 0.8399197742347304,0.5454449369482307,0.18616086686665612,0.7727637975662084,
 0.5831558509676839,0.5755835962397673,0.7617421666333997,0.18869609284686972,
 0.21239550156879072,0.9513864968442322,0.1809970222027284,0.5691474782815771,
 0.28020167209994196,0.9077943464377021,0.45908213802884545,
 0.9594265562599821,0.43020399885151883,0.6736126034638318,0.8740579434882491,
 0.1754400484647114,0.3041274338102219,0.07099322701690336,
 0.26205599012510417,0.08289781567299481,0.12062542132265297,
 0.04168511636225136,0.6305794940050669,0.25599204723544944,
 0.7826115749949317,0.6988787621133887,0.4400298874898494,0.6768235102523611,
 0.8477819522630018,0.10024593780915958,0.8540568386896045,0.604042876513811,
 0.9169725406247536,0.47135276548691907,0.6328958464668178,0.9909893468630656,
 0.3130236811403606,0.15194705109713436,0.8760825798633246,
 0.11693831219696527,0.016491333822278853,0.8703396143628147,
 0.08312480734457639,0.34946856233286816,0.7629178300058904,
 0.40753301343819526,0.8124063319834489,0.993886064975929,0.32363036944000056,
 0.3326127548799015,0.026510675496428493,0.9870072954761663,
 0.3499262335132498,0.151363288669222,0.24690343936091463,0.09374446062357533,
 0.2194635838387805,0.3587495267650822,0.8315617187178267,0.8775114209003015,
 0.7996499807142697,0.7269908482973307,0.7840760699671792,0.6991625518642004,
 0.5006524772778185,0.971623205109087,0.17382452235601376,0.4613966569036383,
 0.32640510611111884,0.7444961207300322,0.4856343409583296,
 0.05896512192590775,0.911876178256367,0.1013378491784458,0.6084421183388007,
 0.9354387755244142,0.40422457975530746,0.523398357652179,0.6684699195673129,
 0.43386607920112563,0.406373365199472,0.6417171282039593,0.6180125133908111,
 0.5378702527758555,0.4115209360838519,0.6448527190564816,0.9162684684270772,
 0.3710955459415699,0.6835183410655147,0.24071562975467486,
 0.01956149588645062,0.1692829741446492,0.7377632126745123,0.8407554545979723,
 0.3723826194766299,0.7603793862635135,0.7542339199609399,0.4313100213442276,
 0.4856675152197488,0.4879805418025497,0.5208369839770209,0.13710558998449207,
 0.9635434290720276,0.29563397582802287,0.5434270103229466,0.7560915320085096,
 0.1488078533819549,0.4176629689689031,0.549314790233106,0.15973801315691705,
 0.7177228166237601,0.8635475474011898,0.6089280344398645,0.2936634946319956,
 0.23752237292093337,0.6987021427262685,0.260153123316735,0.7874543041233536,
 0.04028652998353521,0.9072662489629593,0.11159104449224777,
 0.6095003156041643,0.44124444221410386,0.3142112590896037,0.1349542397125535,
 0.6504837573341098,0.6273087419280601,0.4695383389909853,0.7886404143655114,
 0.6214453045605318,0.6159133354292773,0.9284994482919005,0.911954087758321,
 0.8292831337564803,0.7662178957039447,0.7449515521071395,0.4548037963923969,
 0.4955589294182101,0.41496370538541816,0.7069839294064926,0.2602933220884628,
 0.7356851811187809,0.6464238201461356,0.25629368474284464,
 0.12036492115003838,0.7038626207322376,0.3179090928913436,0.9443113105946851,
 0.5329522575408765,0.7802461551757541,0.8278911647066899,0.9347961519341313,
 0.040671946127226355,0.6801243463854966,0.49911804379840696,
 0.6956090536717912,0.13078361204303746,0.3090045718721284,0.6522929646822407,
 0.7159151186190496,0.08669334489874381,0.9936879250156601,
 0.42701215665993053,0.05983555238489369,0.4194818904337295,
 0.08660057987195269,0.679029040186478,0.6157538703806442,0.44366432981033765,
 0.295956811980035,0.019785080260784293,0.40442882018330417,
 0.6193956806934603,0.533446271378161,0.0025639891076180144,
 0.1840231148732583,0.48086456281311973,0.13531990352001588,
 0.7917950148770694,0.20575185754054537,0.4416265552992499,
 0.49508781433893145,0.08166537397285434,0.5885149128250536,
 0.18859508342821218,0.06068016608147642,0.8794274576875447,
 0.5355482402305257,0.4990060936785884,0.5135557880666464,0.6471003623982017,
 0.9165326402634396,0.01935764521411898,0.5438726197429576,
 0.08138558745154745,0.6307380486095886,0.18519481413410688,
 0.3523315529007708,0.38008282492010326,0.9482342367057847,
 0.17077194840466192,0.008182580832022501,0.8734356141330831,
 0.5207449524214091,0.8323124827539814,0.6998881651293057,0.3692498630884222,
 0.716185754791109,0.4862695388662608,0.9334207651042101,0.2100926333989659,
 0.32091389861646014,0.008008281898103098,0.9146200970055269,
 0.21001363965421227,0.0436318385572072,0.689800626788359,0.6338944538674006,
 0.5075083468409478,0.4160548805752561,0.9202825276243458,0.3522888142386249,
 0.7172939303785953,0.27155457168633546,0.2165364449224989,0.8995129893601267,
 0.22045133223503477,0.5476057280763965,0.0669685137373448,0.518271134878353,
 0.20128273044715483,0.26412872644386753,0.3770100001543635,
 0.1667958672115918,0.05426411992531044,0.2391267343401341,
 0.24669660368309154,0.4850165197513139,0.27894728015575265,
 0.6708318376380207,0.6251564689666858,0.9680600545225817,0.7628514477168408,
 0.1501248891443111,0.36448797359304896,0.4341887154798634,
 0.17787911672510792,0.9696912871881775,0.2278069400343643,
 0.06302451260533604,0.7396225522158065,0.4052703461415649,0.490076157298604,
 0.8285494551343475,0.11846850111033347,0.8935909335960461,0.5312633208710207,
 0.7940652253778984,0.6359260440349019,0.5087421188923263,0.42296450791576345,
 0.6940516985101373,0.127021791461556,0.04345955790799849,0.3263200323496329,
 0.5315836911888867,0.9606479047691848,0.2254667073465605,0.4543890950309861,
 0.5799352669103855,0.8117555587256154,0.6930343466755085,0.9884473789670852,
 0.7624419660155115,0.12443167676428002,0.9902678346960626,0.592733811658496,
 0.9230565295757904,0.6702933073401796,0.4149240366522281,0.6693580311725403,
 0.11125851482322935,0.429660122482901,0.8415392866757155,0.2545548343653262,
 0.5558567198509556,0.2688766718617229,0.6894572002345525,0.9951392812284328,
 0.7900158432055766,0.9984720670947697,0.7829450474003197,0.6461086696536963,
 0.44934177766500427,0.315279812331802,0.12400790805756401,0.2042514494019615,
 0.6755851052965616,0.7924419400786298,0.1529753619096238,0.0470542494318158,
 0.784924444662997,0.9114947378694196,0.9723809859629231,0.7687775677295257,
 0.1340263210475755,0.03908406452518143,0.050912348975700406,
 0.5594746043223743,0.7113906869309328,0.45171532687051286,
 0.39948348740155015,0.06666805376562823,0.08233625291672131,
 0.8279392698960663,0.30374993914839366,0.23759909154993442,
 0.7754382219130117,0.2910888532603091,0.12197816795863192,
 0.42858927559351456,0.41555292858214854,0.6888332052436215,
 0.1586035622184816,0.5152596915631984,0.29717211354368633,
 0.11602373956047929,0.801473120037576,0.6038713087095451,
 0.004811376906558307,0.667908410049542,0.3755611717569334,0.4853172224401907,
 0.7096671588388719,0.9760965340304261,0.4217886778560376,0.27989027001297706,
 0.8463222183443365,0.8908788658278347,0.9236261527491267,0.09965096579228594,
 0.8581883129226361,0.05920808818003054,0.5620725831921726,
 0.18905190692851181,0.954761457738769,0.11194854939655074,0.5215314618662265,
 0.7230261203629887,0.2654715687175304,0.6471008186778628,0.7615182633615147,
 0.13874626854609518,0.6992709336270385,0.9562608557456082,
 0.23356207207443613,0.9178836822221523,0.16397118796855636,
 0.8795948449099054,0.8168992038069574,0.16175313271365754,
 0.43178294024058506,0.777388786480127,0.5724634701557662,0.6375486893212992,
 0.08692544320849271,0.14140126496862104,0.2551155064586291,
 0.47885404275250865,0.865627819528105,0.1244902634045465,0.8175807164721627,
 0.3994796761888073,0.9181797000597062,0.029207368215875107,
 0.7203321954187974,0.42694208769330955,0.33285641430313617,
 0.03046439036414883,0.12778256510191977,0.6676270609971978,0.615730504524764,
 0.5209614395763812,0.5110471276816763,0.537310679516019,0.02500744039608238,
 0.3821937478800371,0.9193312082745777,0.24275581821089653,0.6531888811030574,
 0.781839636752343,0.23615358768246808,0.867963692225722,0.8477526275634982,
 0.7447271906291463,0.9751500283652744,0.10822490793316109,
 0.16492818742932036,0.660580303534209,0.15428541245086724,0.5419752269327364,
 0.023451651545517205,0.07407311578755693,0.9343903776851407,
 0.32545481876502336,0.6588032279110163,0.024503680243847192,
 0.929363762201973,0.0697959193978609,0.43493162801360163,0.546878512667014,
 0.9869070059348721,0.6535782191017052,0.41158044693354934,0.8012101583012314,
 0.6023196590983841,0.07955846021206836,0.007713388038819047,
 2.5654300455291157e-4,0.36105210196642756,0.7157190860015333,
 0.24433140730227543,0.3313974209606799,0.8911349537337245,0.5104702908045996,
 0.9140060639504677,0.9542288531208752,0.20392991686914075,0.9133432484869379,
 0.10781217371359375,0.26158493570707453,0.31856004526668447,
 0.8501123097447029,0.2826594050667256,0.16661237645626992,
 0.039652960541909255,0.4987744406868405,0.700365540127404,0.923030744999183,
 0.5941817363533772,0.2925271967066593,0.04859517926836843,0.9816865892514705,
 0.48233690321579603,0.5467651179969061,0.4974946896337463,0.727582112976097,
 0.5670814403037582,0.06421648795156287,0.022070500063239407,
 0.8208516466507411,0.4153866934221787,0.31498583244214273,
 0.41831792555917124,0.5741359684307911,0.1525951139644508,0.1785346870145228,
 0.911771162468272,0.6940940194196396,0.4597394040878502,0.022275900950877836,
 0.8315589791809258,0.38636359618848015,0.13803371094490013,
 0.7239216078481902,0.9876134556942893,0.849618745339103,0.9816679341382737,
 0.8945927478878118,0.7440691902002328,0.012615116039219165,
 0.9830115428389571,0.12645041216318909,0.7149434452799373,0.1889744029080963,
 0.29030260759701176,0.3219276611221037,0.08362294008761917,
 0.21671218311555962,0.29108887285102236,0.32922729481903557,
 0.30873543033704165,0.5456388228482034,0.6305092378869284,
 0.07802395713959731,0.39954422767854925,0.29403520486999435,
 0.18600331234184553,0.8329381541774364,0.2548068845082764,0.3843075722948964,
 0.4209744065394848,0.9813294221159506,0.8094800171445027,
 0.0011296493224581816,0.5452140452514835,0.39241462711055264,
 0.15652059025422294,0.14788813136638668,0.8891672524830402,
 0.037417453088320674,0.2251234327704682,0.936149765572017,0.7040756410652146,
 0.4789559920077028,0.49835244553220825,0.7019578883195805,0.9009662206542357,
 0.2006033399500352,0.13778260666247844,0.8238558869298174,0.6852924827937336,
 0.05559831979002561,0.371272908835787,0.39210378135537416,0.2885820045554567,
 0.7468856438645267,0.764196640081782,0.41815698969707704,0.9702104933145121,
 0.17100367960208573,0.8512151546964409,0.6587820703712128,
 0.38994558281786285,0.9994159106993611,0.9585158834104772,
 0.039834362657831135,0.10433743130180317,0.9010906578192901,
 0.4819697515352629,0.4867792986806919,0.9148530549121539,0.07597838072287755,
 0.9539562673341025,0.43115083803703325,0.019032989256220656,
 0.25053660215070517,0.15501945773547077,0.7213772819992073,
 0.9261651609132997,0.21623183073613084,0.2368062898317912,0.5313996089498575,
 0.5529995730048459,0.013590988034392182,0.7976108604897696,
 0.23860888426672577,0.48975666742634705,0.600846511688801,
 0.18274116153733522,0.5736081836393981,0.28498671533396025,
 0.9723331718564988,0.8855552971281098,0.8951386438450498,0.6640870550019867,
 0.7810360625965848,0.15970812337297657,0.6845853909018437,0.8119588693835282,
 0.5887711050540095,0.7885111123847877,0.33513165378589616,0.5048431719067039,
 0.2916612412419066,0.1693770771742219,0.1738178745189719,0.5976199180975708,
 0.9557585140007021,0.4180900777588231,0.365692743183905,0.8065706858464778,
 0.22140132404090795,0.9158169540597921,0.3227579662788227,0.5107195252115003,
 0.695481406075323,0.7476112174756686,0.9630451142884341,0.9534530774688409,
 0.4586852488951838,0.7006823820074152,0.17962392688735462,0.3363597800449757,
 0.28045126900489203,0.18468150649750237,0.6160361221566394,
 0.5708183819873794,0.7739745904212245,0.7910584729052076,0.9054100906960201,
 0.17142010431318355,0.6731520402475517,0.7660524832722981,0.3663096064573189,
 0.8525924996134143,0.5051981823812606,0.750186143227214,0.02091725590760951,
 0.19287761560908212,0.6652638898864178,0.5520444941892656,0.3549384628160538,
 0.9787045822361542,0.03339450827125623,0.5674951615507575,
 0.26884829013093237,0.30988889636548245,0.4561609599424583,
 0.5936547766680218,0.9665876861148902,0.1437713527439084,0.4420144064413868,
 0.8478261923536945,0.44825531005511954,0.555339879303898,0.2556211837887462,
 0.5142569070612567,0.5826248984211769,0.23677671171358416,0.6839297928906255,
 0.6792376180790757,0.48867542968394373,0.12956346101044147,
 0.08943788139919095,0.7281826977402198,0.9556864515790581,0.8963284819899887,
 0.8534359080359519,0.503606970954687,0.2761302311723842,0.19388909529178266,
 0.5797640283323359,0.9912526300892289,0.7461328333579553,0.8996995881831229,
 0.5860648562241298,0.4808632002013401,0.6183723735307642,0.7527851093225718,
 0.012293831839801261,0.25748931936010944,0.6383679489167905,
 0.6837971875061764,0.25396092219008937,0.566878411833074,0.18532043706917445,
 0.7108837806086468,0.44173583708476616,0.5865789903493522,0.975941431717753,
 0.5659905510062606,0.07727525199587804,0.30581077943408796,
 0.23892046801352929,0.8166241492328203,0.9716467109217934,0.8781426598687557,
 0.2090806708003392,0.1692782721107835,0.6315679521102158,0.08364580182444581,
 0.4111624141413086,0.3215932145159339,0.7820292145066268,0.6549637987294287,
 0.7203392615918988,0.683320592663031,0.5066460237442718,0.466899959299098,
 0.9230373924171994,0.2643993221841052,0.8927408636564494,0.5732665281525355,
 0.844904796198067,0.8727579922109706,0.5004282722754777,0.18823126655437206,
 0.6935013347299769,0.8096450740418306,0.7217727468260383,0.9332563307299189,
 0.7545407172004703,0.8513113755662386,0.9189140356262362,0.5964189511860066,
 0.2911919067444255,0.5320214517672763,0.905176584447483,0.1355505367881854,
 0.720551356365626,0.7239193837475325,0.5199479762914136,0.5083761525772281,
 0.5455617237844427,0.49879658402546345,0.48238536479109095,
 0.47667583228421395,0.09011587354009465,0.14073049890252365,
 0.09177918715322164,0.43253429786652076,0.49501965650353374,
 0.4391415135928237,0.34353284482299207,0.6732640191808804,0.9026418132691904,
 0.9515054267964139,0.5719290981832231,0.024589206960877408,
 0.22397467849346508,0.5429969275926665,0.745619047424789,0.22008962283454414,
 0.6783790064459386,0.8622199725922399,0.9773823895823108,0.02019571077195559,
 0.4632910789438698,0.9074218351629006,0.7497526912225221,
 0.044249176528055045,0.17798801343446702,0.5149615024816186,
 0.2828000368621463,0.9521619553315025,0.6505909296098165,
 0.060634527939952276,0.01645952895830427,0.8912885724796344,
 0.4183078586346396,0.0022681982002484835,0.14278689764119823,
 0.9243311128423576,0.906697108022911,0.002858945329919149,
 0.07688323784633133,0.558522277714252,0.2980155893074057,0.5652033924885176,
 0.19828179988529016,0.9938739879260491,0.9417912265077197,0.3310508519831439,
 0.9393581663130168,0.33993995855470693,0.2917798922887054,0.282051392813613,
 0.47550220070992455,0.32610956378927114,0.02390737638172702,
 0.8476501594329208,0.17681623468044472,0.07176227890368969,
 0.4918257048561251,0.2751100482430333,0.7412805077778766,0.77456545495177,
 0.6270632659522963,0.45500934335590215,0.08766793978272125,
 0.9760339466471262,0.33977751175088833,0.46531922677133486,
 0.15637904886471177,0.37462732428236867,0.0340918695355622,
 0.6977909285084126,0.9465178233215201,0.7669689024911978,0.09180078410496073,
 0.4935826089693337,0.2273982031959242,0.7893134750529617,0.7508183918697521,
 0.9203330348307054,0.9499539307803542,0.3079050935658386,0.6274924424780642,
 0.8603340286796897,0.37185552270535804,0.42702922005313904,
 0.36586351612964496,0.4788854491214366,0.24963414435914877,
 0.03581245552157897,0.4701234537524006,0.08106890425703228,
 0.7491450837758217,0.14596929537086334,0.568848908449136,0.5379573473757991,
 0.1741128680612971,0.017096447929017877,0.37450472200800466,
 0.07576128443367669,0.5791819426398127,0.04789098139271497,
 0.28417379870660775,0.2832971688852972,0.8805598655837473,0.6041894498767241,
 0.389400693562427,0.4548544464293627,0.6165288791187546,0.023913363820766587,
 0.2668280514325869,0.42659467566310194,0.8207776996363477,0.5735274153038815,
 0.9574975800225167,0.24488167224365176,0.4075276175670377,0.7368399459764118,
 0.5443247205195519,0.32886695988879333,0.5651177092090486,0.7825609527130372,
 0.15909221152871278,0.05543954826654407,0.25487467784232076,
 0.22477478196007006,0.06647230073536325,0.9594028338823971,
 0.4369689021479882,0.18765485598816667,0.41630011369390874,
 0.9187665906182703,0.8472447627485672,0.851355234138174,0.7096089272995707,
 0.6453217013263575,0.44772027275257065,0.3790044461789177,
 0.11539014833964178,0.4696993261633139,0.6078961044709115,
 0.45693977428134736,0.5702929300815102,0.13634214226164287,
 0.9348737931658203,0.08205950778477722,0.6807214031252296,0.5201570798775197,
 0.43055457913658357,0.5075877668249692,0.28148218550455617,
 0.9362336374497058,0.8347893643948043,0.6145575499983081,0.9549613459615347,
 0.08454732706359602,0.6350500957371326,0.3725618599985949,0.5492473429648514,
 0.18489512787280193,0.8644778509609272,0.4149009816262371,
 0.03621746002549975,0.14770166084853076,0.9884090623009836,
 0.3016391067818449,0.7462345468761172,0.38087535387142657,
 0.008030064768528522,0.5351500533560378,0.8752672279427941,
 0.7578798213511024,0.5573042152528003,0.3022472357793142,0.9390928965357918,
 0.16720949853675138,0.10212767393261568,0.5658251453893298,
 0.9045584494944938,0.9978918841279208,0.006871589515392906,
 0.28668175975226484,0.35644154423892316,0.9362187718116067,
 0.9363431334536012,0.19739925274708003,0.4072721359925595,0.4091645564874281,
 0.8875540966695286,0.11520451914001217,0.15889225536307494,
 0.8547857541406068,0.9832619946404234,0.6109512062857957,0.3097280860775973,
 0.28190414410060494,0.7224652574297266,0.5768450228252415,
 0.16483117848988948,0.4128368965664083,0.27082228291655774,
 0.12105065405768223,0.35506484207974076,0.5209770875210117,
 0.9567084473954378,0.06347642580902846,0.4833348622218494,0.8906261690281018,
 0.6944929526497574,0.8769848743642118,0.27891485942577354,0.9808260652797157,
 0.9893691395315799,0.2777818411273929,0.16042562397513127,0.6858403204605492,
 0.35375113446361983,0.08241788915796722,0.8682317076368586,
 0.5790081130970031,0.0651877439174483,0.0985694446969344,
 0.037905222143460815,0.5223622926205352,0.0025910357116301075,
 0.6828069718839422,0.8906406458751246,0.08142927714097692,0.8039860237255196,
 0.7315368366831866,0.9268287710885719,0.25289073733837575,0.6996149527027685,
 0.22750925779652376,0.3490128996539835,0.6127582149100713,
 0.20389336692651694,0.7627628724565002,0.22147482483160752,
 0.1105344902583596,0.9004798131982232,0.5023152434716449,0.5378049785759461,
 0.5305131757423207,0.4780587474934783,0.35521256919516064,0.4385396294357836,
 0.5878794789068087,0.7131045816680641,0.48758197328362074,0.8351381744704653,
 0.2048081607582668,0.9791340475752981,0.9269359759671839,0.17098407928934223,
 0.14125021605132404,0.00583703001549285,0.3369445044940905,
 0.46134421304025064,0.00481464749889271,0.4786366862757795,
 0.1475428258135928,0.13264964592492978,0.23355436463990864,
 0.013845768242927492,0.7334817206665601,0.49828443774380804,
 0.18830542083824509,0.2734360016876699,0.6553047012723934,0.1589187887826501,
 0.2739217152582647,0.5770118858377297,0.14479295509566192,0.79185946796732,
 0.6705865670818797,0.913334245384173,0.6834108529477941,0.7716030463486133,
 0.8296354571092138,0.4509233306970776,0.4331821023665424,0.7599274780849784,
 0.9400885810946497,0.3295258983484043,0.6828483834060214,0.40869318686324063,
 0.7915153580715949,0.42650232174860403,0.5884920824687803,
 0.31161633630014396,0.9529536505387668,0.6375526893974214,0.7597565543759847,
 0.8954782840107752,0.0498929650864155,0.42082502337249283,
 0.006340178891385162,0.7275345555292715,0.24443786963325376,
 0.1959973758050444,0.42127676039682826,0.12500398219124653,
 0.6455156266623491,0.3891134495036237,0.5495517135619665,0.6921698674476208,
 0.6169388964110514,0.9715416177717819,0.6839260799996232,0.6650018850140738,
 0.8665726493905125,0.4434028558057096,0.055139207477740926,
 0.8837700552286507,0.18779661868881992,0.13074184006676137,
 0.5268572514692489,0.5216101242801631,0.6387929585116565,0.3689119711113793,
 0.7616375967088119,0.37294399509370024,0.3594095820377248,0.8325682330179873,
 0.18777126282201473,0.30364202261272843,0.161999907409052,0.9657879462190463,
 0.09076478738303417,0.1794226571243812,0.9974659888002927,
 0.007697819354890001,0.23532314717089597,0.041771219470496335,
 0.3633663251177932,0.39568410137816257,0.3224166795552018,0.6619068843351738,
 0.23555008167920277,0.9884100627130534,0.22768766785653205,
 0.0498180821392592,0.19583039644462719,0.3286251892267016,0.6405446158155963,
 0.9388938670126052,0.13675653247100705,0.8715132961649847,0.9400266181720491,
 0.24995498711465614,0.798845320086263,0.6477106494116606,0.14361920162861297,
 0.7632474674774335,0.9951248866280494,0.8827110161144059,0.649634608384739,
 0.3469277457064185,0.8156079675501555,0.09360128775090715,
 0.06914777327815769,0.8031209656730965,0.7671630623070856,0.6268679694555935,
 0.25627124801806467,0.7236680863011201,0.9458783084764089,0.853833854126963,
 0.40582736829059407,0.4447308605971342,0.8090569090752165,0.9124093382881301,
 0.9424474411749455,0.8613275134490853,0.8931772081564704,0.8536662377060701,
 0.6841333446386709,0.5877131089183375,0.10142049387803276,
 0.06455645353509365,0.13372527515636656,0.8209444157719877,
 0.9491799163540531,0.040353992702765296,0.031946746850430974,
 0.4828017275467892,0.6830185901763242,0.03192598111587264,
 0.33691940230139406,0.08664686150409451,0.21123247344320517,
 0.2117175567895513,0.0041996073158429015,0.9889933658025882,
 0.7341880358629131,0.5077800582256332,0.9762372165325588,0.19828833438681226,
 0.7765976415411298,0.9835883685272326,0.060953369797748236,
 0.41397948595589096,0.707621369571588,0.6415279576805306,0.8216020957040199,
 0.8395947222319738,0.957006629982421,0.9530084941102502,0.7972561888607069,
 0.2327837154376038,0.6442160579220103,0.029682036291711178,
 0.9840176052614856,0.41664160471205014,0.45642926706377895,0.633912062385457,
 0.5211124637511967,0.5051140085136855,0.6979283789282382,0.816889544973809,
 0.16506774747519304,0.8741826199314442,0.5587674704487067,0.7140019674394271,
 0.5101651296008675,0.16774641757859388,0.008009860520870271,
 0.40756555491705226,0.056700160278860934,0.6044337165905712,
 0.1669795220242709,0.4250431536634225,0.23423622335298577,0.7795920725844598,
 0.5359665743588747,0.2915032938695542,0.8922684899137338,0.9238939083092021,
 0.4872292960085033,0.7872695645635347,0.1470944482061911,0.6840466053393265,
 0.09998159429288445,0.06863588088039507,0.5318831176225501,
 0.8467927982134866,0.7574176512121178,0.7552181571268,0.561521808160026,
 0.9062119620968097,0.3366463583055954,0.19167376144532977,0.9436420194289793,
 0.7287790907843886,0.41579006211728287,0.5484679622022658,
 0.19548210956129808,0.8583793253197541,0.7633218228653899,0.5490014513945295,
 0.4936217775080828,0.7382642986738158,0.3517843831134443,0.09453283113090682,
 0.07345409711696294,0.19476005665745533,0.8060449357570956,
 0.2690619249552927,0.3150053447113631,0.05341453893069592,0.646906687272928,
 0.5225867040201049,0.7430392599153877,0.24706861270669167,0.1200460121450021,
 0.036864852779704416,0.5460540001099844,0.2942780464784758,
 0.8413863188021972,0.7083212114632282,0.4589535683516882,0.10048601396192702,
 0.7503449607063092,0.1433389451328677,0.007232710443316526,
 0.3557302700419903,0.7731782670199094,0.8273810542841711,0.15407112625193642,
 0.43357361230201763,0.8601649372071142,0.8979468789856786,0.6443341057610594,
 0.9901302031989323,0.8367633715479672,0.9623708262990892,0.34549443766277854,
 0.32243686071286304,0.6195200241784119,0.29776219316282215,
 0.8034097777645707,0.11981396231529151,0.6901505089643468,0.7427013944843243,
 0.6190381809257711,0.26110947079951996,0.4125838208252407,
 0.47357363595176194,0.4372127775524738,0.39203465272091975,
 0.7506211680509651,0.4470112176590876,0.41617430130792477,0.4448685489889195,
 0.1485135698776563,0.01652437807825191,0.6859246868602771,0.5816344796800045,
 0.6244765594001543,0.9931262745144595,0.26445566042441293,0.5874318182624045,
 0.9344071629950785,0.025730275440645878,0.2541025706922998,
 0.5898356189052812,0.977175673807795,0.3221654570397474,0.059580446268132414,
 0.30804608349988716,0.09564669942900672,0.25437283202159255,
 0.44559099450654327,0.7733814393478726,0.3708326197494144,0.4665932556312865,
 0.12269300704036934,0.6603208892433119,0.08379902717067178,
 0.44617217763993655,0.30780663199241975,0.5377365041096007,
 0.1310754127317737,0.3286031785045479,0.553633436106604,0.2758685682154475,
 0.30169239235275125,0.3230225606274887,0.2024073806593356,
 0.33901598268917843,0.7711732329600329,0.5121519732270642,
 0.13525677644351197,0.7998543786586412,0.3583578375635682,0.6207202485879237,
 0.926107864811466,0.43557142608747723,0.13190628117872427,
 0.39171734101861067,0.9598445952330179,0.3084043363221368,
 0.32317196260062275,0.24447003167371706,0.9276151331007919,
 0.5918228447515274,0.7845754152813456,0.9426346280822087,0.49910572292909994,
 0.5985302385891673,0.47482058701526364,0.5975115836137422,0.2767189371281077,
 0.9919298186005409,0.576571098246117,0.20086142515188032,0.16952041149020747,
 0.7737290670898471,0.6388987629206078,0.13559306605139465,0.4107599808586304,
 0.30748499340447566,0.7667877477726855,0.14813776950812363,
 0.04252339128648097,0.6923452068782461,0.7309186453260764,
 0.22335134384755273,0.4809838096119614,0.673710086874822,0.9939731762300026,
 0.1783654646559505,0.5702991107027832,0.603770036353261,0.4079572423069394,
 0.5889904567959678,0.8631094729542135,0.007836917839364688,
 0.027224353594193174,0.4958799716345518,0.12979209451889262,
 0.36583720637369754,0.3812194252225236,0.13337749519049313,0.551765955619644,
 0.5453273152962576,0.9480335503783597,0.13147898753263598,0.0862941055976516,
 0.41738235034074267,0.5088826218298619,0.5855120505431689,0.411622841603526,
 0.488079534319681,0.6337737468933538,0.5659345768895,0.31641029267684906,
 0.7697717151661831,0.14512944886272972,0.604461445532741,0.5828010874852112,
 0.3627272883216861,0.27503949385089455,0.6699219383001505,0.4480564376365056,
 0.5914107334156822,0.1906894917904336,0.30518983823326495,0.5924068180733845,
 0.06290007706763523,0.9997455819919717,0.2347405428895134,0.8948174963670221,
 0.3098989583787338,0.25504107381655516,0.30265786116329374,
 0.8158172313548514,0.8918522280374729,0.7701165818879315,0.07912012495380838,
 0.07004438818999958,0.3014700229723557,0.9150635065295419,
 0.23440549644524578,0.6513636041772122,0.7851172963006651,
 0.13464947329679444,0.600625963913562,0.1098810267865662,0.6076038913891278,
 0.07348456466559461,0.6027807874592218,0.13956659028139584,
 0.37203298082880043,0.6688852911291651,0.3499047851048457,0.357919463717415,
 0.08082282590226642,0.0889516560461664,0.0068724864303981725,
 0.4826227633245689,0.3966878617208962,0.8716588507084908,0.8441631856560993,
 0.4702773115332176,0.7086691669533235,0.07209283298729174,0.4772708382924308,
 0.20320867726881986,0.3632013860454978,0.7268798015584751,0.8514886640891954,
 0.4303979249623777,0.3081604450546296,0.776046507699325,0.4581107381005928,
 0.5633947519389152,0.47664493637279937,0.4988563245548632,0.8442444849518793,
 0.6501022892239237,0.8560156957319676,0.8452928758991586,0.7322624432339379,
 0.08035476329522051,0.9448870671626417,0.4941933453352365,
 0.16354555611463895,0.2852227272840071,0.6239382878644537,0.8839094475949469,
 0.8248945110937603,0.30298512545496303,0.057671002638585156,
 0.44746450409514194,0.9156969441629366,0.029695449721899192,
 0.06395945263897196,0.08174899230454524,0.5071934435365535,
 0.9663672843758747,0.22526634500206988,0.9899802656916614,
 0.09220017209203046,0.7451482355425733,0.32075531272158353,
 0.6526117182535996,0.06482973522084023,0.43279113088366994,
 0.44179812048253697,0.07714039277285578,0.11537382327882084,
 0.6549696240574091,0.9885210868503201,0.945203029409546,0.20780802881160576,
 0.7198443466900291,0.06131795388271133,0.21712434346575682,
 0.6986441288404958,0.0673813511524598,0.9332025925047333,0.496107883116656,
 0.5044266189069836,0.22510767883562877,0.6696461438653631,0.8360019433502409,
 0.8045866839006665,0.0035844848037195343,0.4563255341954986,
 0.5086980552878546,0.6602643508489644,0.6988827255176782,0.1798208982030136,
 0.6591567184780404,0.9777036005995811,0.39988028174045076,0.5218148089335214,
 0.5114776406145916,0.30856900827259637,0.6022597879022478,0.8702276288274059,
 0.22896484382774773,0.27170489190412705,0.030514415458936184,
 0.29820832377216266,0.7891051349109548,0.7589048416354613,0.9357656019292809,
 0.8955382882661687,0.8430996430365281,0.7974186388702434,0.5964686772376199,
 0.023570851814286864,0.6396111315613536,0.1540624588416355,
 0.8947348326669801,0.7745069768912259,0.24751063426619369,
 0.23672623137036197,0.8869826392547988,0.1466282847351308,0.9762244175995005,
 0.7780307574836997,0.5234509455345547,0.995461691053188,0.05535889803218197,
 0.8868795222108189,0.6671616879613357,2.863398693270236e-4,
 0.5051754382173527,0.4400684994091254,0.5669086376552194,0.3440457054631466,
 0.44229999179975876,0.8622894173771497,0.39440546661448095,
 0.8495752024749903,0.49327076900542766,0.7092219652561387,0.1073549842032262,
 0.42770082423652844,0.0537416025054126,0.4964707109152712,0.877645966157417,
 0.8564574400152419,0.4605255709795598,0.9086871611469589,0.5110064996616542,
 0.1664274181550679,0.8086452519548363,0.9866641856855785,0.3313884744852701,
 0.07546786180472942,0.09050245632902554,0.2193956024498266,
 0.19587000719968184,0.38696471031880264,0.992444456563536,0.750738358715334,
 0.4000713773832848,0.15901542181826356,0.11128039290896774,
 0.9301702820314908,0.8181843570582621,0.12926071487978286,
 0.19432496494738483,0.22440496375327812,0.8672901009425876,
 0.46782957482886145,0.026193131007438253,0.21344314272171205,
 0.7474111487771927,0.370924342021103,0.8068482640442606,0.7218126488912076,
 0.8579081690416401,0.006794367065609208,0.37295216924104835,
 0.6621676752054982,0.4241325806709053,0.9555318233129309,0.8266591536112112,
 0.35369300834100614,0.4475625871783421,0.1514313303029754,0.9152795932700982,
 0.9980858712926849,0.4298240514229993,0.015757353937513208,
 0.7109334667356124,0.8558941864387447,0.05863356848110801,
 0.01473528183606565,0.8800207342823423,0.20053576758965053,
 0.22251774425716775,0.7925171330295958,0.5857784863676667,
 0.09189918945299014,0.4271830310605311,0.8512924799332342,0.2594979432736023,
 0.3396913655657495,0.966814942243634,0.5143706614265717,0.30583844512029146,
 0.5969702009788435,0.29350397302168973,0.8040011020232397,
 0.08945031015978921,0.7146422610277545,0.4030150792481958,
 0.42241381860459415,0.47133122290794716,0.045317585579898956,
 0.14651303080944378,0.2853815955019312,0.7749569085184691,0.8667942508844779,
 0.13947291845819443,0.9184650689082824,0.2690139080910299,
 0.12348793488916021,0.8672257894910457,0.32122389813043295,
 0.7571377441408979,0.9107416560411297,0.30121813102958284,0.4953571025133823,
 0.42783020862728194,0.047252607377762046,0.4042469797238655,
 0.8668184352513608,0.6386673403002499,0.005917803012569611,
 0.6901948692915036,0.4174288883894782,0.044477822077709006,
 0.13437853577596304,0.8398691949540791,0.8364001458051246,
 0.38397822266892856,0.05171415989199135,0.504082035336068,0.7922103199392979,
 0.7125872840806503,0.5934132878588642,0.24520858650622968,
 0.026036666654417395,0.866696380672187,0.022906148414230887,
 0.5439875425615732,0.6550826619067712,0.642104113140884,0.14882419185579632,
 0.502332964222242,0.7039567767644049,0.7973187253452046,0.2843093785119245,
 0.6388358299070522,0.7559692876023366,0.7151073456485911,0.5334284007491753,
 0.446050300148885,0.21540356743879607,0.2879921582337529,0.11512902645904655,
 0.4256610746376924,0.8799032040081465,0.863906797523197,0.9744785004995942,
 0.6359759560934332,0.9618994986511222,0.35823778080808566,0.9278644008871171,
 0.5277737706009383,0.3768121884970207,0.5991599586075883,0.3756718245406252,
 0.11334119824623712,0.2355231789772143,0.27837304043859146,0.997812586139012,
 0.9834631529856696,0.8730828837828284,0.22485765083590392,0.823124183131164,
 0.7885765358382593,0.067153190316109,0.7553179559856406,0.8922413686585692,
 0.8746897827184896,0.42702104515280126,0.9322198884415417,0.542971607925687,
 0.6155894304966427,0.926056071135386,0.7248615262280274,0.5074140580170476,
 0.1999573072994908,0.14986040019428903,0.138245555288804,0.745031493641984,
 0.4099751249278284,0.042804473054934844,0.9822157934883212,
 0.39848745996673574,0.8891038106075079,0.4060930465560161,
 0.32197488115929795,0.6119046636073523,0.015474438728620976,
 0.32504014612941734,0.9216071279903755,0.9143077121133295,0.2402547012645191,
 0.603369863649823,0.7648101636608672,0.940201037910167,0.6466983724990346,
 0.41916620380989744,0.09690639711073523,0.16774732274173554,
 0.6736470137351173,0.812970872973021,0.6150811507305018,0.5831665966336569,
 0.23674494700421334,0.843940776741795,0.4628659262132748,0.25904880243361517,
 0.7821655452797331,0.3594409286039779,0.5131558848421837,0.35448153159639884,
 0.5906765982344587,0.17962422668610023,0.8471978940942979,0.3289798054734083,
 0.63951019480419,0.9387337686373082,0.6357416297726495,0.5582858770412495,
 0.794605864431353,0.4000935309470022,0.5460763680411156,
 0.0035553134484174187,0.17716631525142035,0.7874934822565185,
 0.7142120418460689,0.6312036085061623,0.49293333972162034,
 0.033525188830661024,0.08491301177795518,0.616576594588079,
 0.4951869677347531,0.12716781809169553,0.9385509756908249,0.600483282252885,
 0.6286149480346135,0.14226020340631873,0.5990840568656636,0.9346487788463669,
 0.45795535299664636,0.5473360414914,0.6545305401268464,0.9494843230031511,
 0.8446928647187067,0.49211610883323553,0.3532286607709897,0.6885896086415364,
 0.13527576590561463,0.5963217126142216,0.4168223612126941,0.7637007558785367,
 0.8401498754652498,0.3710164501989007,0.45113916632967965,0.798214903817608,
 0.40464029011793956,0.8479649908981599,0.29579205829344923,
 0.4020470445458243,0.16378911927637807,0.4210093988714779,0.6538671131863094,
 0.6183792456691661,0.3841579199784353,0.5609779907706329,0.9922048322076472,
 0.5415984240732713,0.837667168965503,0.7476058514391296,0.6080886189298809,
 0.32661805683766953,0.28399502380742037,0.720371902120231,0.8773044764516456,
 0.5939421636756987,0.5378219729948299,0.4975408482913185,0.21127912347124067,
 0.6052323535083585,0.20875011358695905,0.6993639034579567,0.9876750667884774,
 0.5120712741229548,0.49808628914886466,0.15019869932813468,
 0.12415991170617102,0.5371535759950957,0.13165311017970716,
 0.8442989511234799,0.05268450152802073,0.5082313283826624,
 0.17971645774657607,0.3062838669645802,0.2538777998285877,0.8478829168504045,
 0.021885041185013554,0.6099551623257842,0.28375506416382656,
 0.8362284147083598,0.6975521245876384,0.8370817946424788,0.8575403201508026,
 0.08478179291821619,0.050613524156227996,0.2379665061519154,
 0.42257543536242514,0.5499509137500471,0.6974717613285116,0.6240988003117849,
 0.9889263190829902,0.9740344180737194,0.588283359852978,0.7389566998446205,
 0.6584823243426869,0.89294153843488,0.16492494916150036,0.1880276228937492,
 0.9516030740501997,0.5071858016705592,0.5994151090289809,0.1925420637577211,
 0.4011192272952109,0.677038339372195,0.1726494612408076,0.5346728257409253,
 0.4091043193255781,0.012088986407965185,0.778996025392637,0.7217214232050523,
 0.25571547231819314,0.15255051960817045,0.8334072228916017,
 0.8202554279725519,0.35434174836943666,0.71537764135758,0.1362719113067158,
 0.8962589102490429,0.9859169279943232,0.8974906571640258,0.9155350924295139,
 0.7116443543234946,0.38451746113075314,0.010060956892744666,
 0.07719617819532487,0.49108524187717084,0.5633893000280094,
 0.47494961018308113,0.6652426166440422,0.38372426777696034,
 0.5234889692272315,0.6180920847085561,0.6024007574053825,0.8430023227160192,
 0.10522098360848287,0.29422491855280186,0.7850911116569159,
 0.19976363631728877,0.5174902421332999,0.425159500964856,0.37919935981960695,
 0.415156127612217,0.06965812584769338,0.3074232099518318,
 0.048787172290547076,0.6014384677537634,0.5280458005759526,
 0.09331256522604015,0.8131151614505336,0.49035584282300293,
 0.43688447163481314,0.1094264106703201,0.786280034057925,0.3336235822758231,
 0.45818416177244004,0.3737125584553803,0.8962202571042198,0.1375913340244408,
 0.022829506959367363,0.1705201487257062,0.9498492619328537,
 0.13281842374869657,0.3454851828944778,0.5754053140353218,
 0.42996523574019685,0.25228997362179995,0.9751293290323142,
 0.5931619996034334,0.3083022118755143,0.2498482493756794,
 0.010605581778624917,0.6632467757797075,0.009380282053932487,
 0.22966145966184448,0.8797984656756188,0.8559192130343855,0.45916291078342,
 0.6478605470308875,0.48782005297770825,0.17787618132595706,
 0.20322044187969535,0.37623790967291504,0.6099535339221303,
 0.10505426427702202,0.5852522252889241,0.7056691979628608,
 0.16542845422623875,0.2149679935778559,0.9447951705823464,0.6383267720809765,
 0.80770573980258,0.08289470354675144,0.623869100618336,0.12808992809137798,
 0.16489623504816509,0.1897749842337999,0.1993343076740991,0.5743086795026822,
 7.754884206671031e-4,0.8506431332683031,0.5836140133601426,
 0.35160826739134343,0.2801608687506343,0.7535089759091654,0.9177870056433411,
 0.8263815807203706,0.4327879791077751,0.6128713631612303,0.2797949494156172,
 0.7392805018933153,0.3918569896174183,0.4976258952435141,0.8363250897269066,
 0.2894823980157848,0.11520171961015957,0.3107952361788304,
 0.0060132759462259155,0.5333066682738723,0.007851426731952826,
 0.7818428016939232,0.6801530965981694,0.46510713028892037,0.9379258187501516,
 0.9618471625912596,0.30271110920415234,0.9682913839807018,0.9077451152866332,
 0.7665322129316187,0.8943251539620998,0.8193126478531735,0.3385434925711862,
 0.19819281877952644,0.2635788209718132,0.20485481415154594,
 0.20506957993759134,0.9440492952851913,0.8704757465252887,0.8821502436212992,
 0.4444507250335863,0.480950271808781,0.16198847718387488,0.7132759562879558,
 0.5120534453434927,0.4515563682684164,0.9158199973574448,0.21407922302354865,
 0.14866205229454488,0.19567778055038887,0.7315839836862921,
 0.04454824125584522,0.978870713922872,0.2587799631921722,0.42489544787978717,
 0.5026404445452154,0.20416899668622934,0.6750071608845604,0.6047606715795074,
 0.8778709706140986,0.24357891863298475,0.06735988204449361,
 0.08444102180118951,0.7568729637140181,0.48890448293567745,
 0.004170091483139293,0.5982646018043448,0.5544337610478443,
 0.6102748382867716,0.4299498296995634,0.036862591144013024,
 0.8230699028452886,0.9050019038772721,0.6173808757844351,0.918079493786787,
 0.35140593520189356,0.7191743363932206,0.29260860355945895,
 0.2332945850487539,0.8992030108575488,0.011181445139113677,
 0.3895453446677374,0.03400896423401367,0.7801507681545639,0.7764322734975195,
 0.3382789950545855,0.4767057257471492,0.283350238728587,0.906969189774131,
 0.4822693546324007,0.9570779095456274,0.35960011359219046,0.2017061636983608,
 0.27743028636994893,0.9809121994986227,0.8174526902507077,0.8264806189143833,
 0.4917459266988048,0.8138988485525289,0.324401967694686,0.3119486791840941,
 0.06037655172280554,0.33345338416537085,0.5819740047328984,
 0.17812378353797564,0.6977642717409356,0.44372341382544844,
 0.41161398378773884,0.7922739756745698,0.7768262808851547,
 0.08562113655497106,0.9665930097715165,0.22777782473864028,
 0.8544667547869628,0.12383228029116855,0.5440092889960881,0.5683225843824261,
 0.9459311948063813,0.3649382067306922,0.1943891720821449,0.06532030240966369,
 0.25756399307161215,0.783197420224908,0.6933239470489359,0.8310134514460396,
 0.19495272871406055,0.728922737718036,0.8849636195479371,0.6507588349840939,
 0.11337715222766886,0.573936595396837,0.9733589842562411,0.592994854507133,
 0.9406958375305983,0.5369093275356817,0.7902382224845484,0.7906984729140207,
 0.8920503700740784,0.8456066338784404,0.29338059876601985,0.9098267180598392,
 0.6330211214812266,0.547134056373753,0.6169019103610704,0.31879698581390814,
 0.48664328486154584,0.49704372897363824,0.8945392947042515,
 0.47042193644560193,0.33671321678500776,0.8709731733205208,
 0.6038761898063578,0.3078212719159279,0.14651276185717044,0.7641239456089498,
 0.3987377970027637,0.07464432078413186,0.9810142731073818,
 0.18865978704161113,0.13598611026183705,0.005117667224866729,
 0.008665562891668421,0.6063827574704326,0.4925412901261089,
 0.3625233674388544,0.3934425036478837,0.6360726868414731,0.4620894363301653,
 0.0684968150718247,0.6903548213089036,0.7379099268342856,0.9582177117571994,
 0.07817407440455948,0.08178067393457245,0.5575621028050751,
 0.2231193975645126,0.5329622051101865,0.7849584147518812,0.33930185009676095,
 0.10545767732054756,0.8968805894464708,0.6934267294700964,0.5873811348594363,
 0.1830719358491406,0.8535308444598924,0.5920206624906943,0.19983218053588925,
 0.12226065044102818,0.3294267612120092,0.5530379284383568,
 0.36665270679270756,0.0801977368419361,0.16213047754072885,
 0.7608890733729433,0.3548038768191686,0.2438377711382732,0.7661232584702811,
 0.33954308132377997,0.05751326468449591,0.8227657333599748,
 0.04209536069979314,0.8336592830213295,0.7486113662373007,0.8577803128198652,
 0.31077283455194704,0.8455723628047874,0.7421968980390891,0.8267613718709593,
 0.2792542200016548,0.4972521770335111,0.4907218328283943,0.31335363678158,
 0.5304897031150501,0.8615749307872401,0.23862992868944732,0.8608124144880946,
 0.7009660619609509,0.6596307344520214,0.03242439028721866,0.688176476022011,
 0.5949685452892953,0.7241199433247575,0.3818352255129114,0.780486835454041,
 0.7180535222668251,0.755277191559135,0.10182946330071052,
 0.016409952973779784,0.25878606664105297,0.5745519255003236,
 0.4618458953092608,0.22312190442103985,0.780385248947404,0.5476753337090939,
 0.9522358123942172,0.18352998659604725,0.8840663128191215,0.5540948821622353,
 0.5440589648061528,0.270637918792933,0.00986752500333532,0.21238906301639127,
 0.1297633058814115,0.04863543187426289,0.3799170441173597,0.79379148031147,
 0.7157400789522236,0.6653485005613787,0.27138977535517217,0.0761806055259091,
 0.9512152620656729,0.11383719731711073,0.578062550644103,0.13914079012213698,
 0.3555105688119321,0.1350488036832651,0.47021245476460616,
 0.40974747909323006,0.3441993031863474,0.8433716408397288,0.7026141500026404,
 0.2198708170113186,0.0060435087330755755,0.2913416418752721,
 0.7043631063420966,0.5907163257195004,0.9844380273758279,0.6088525716978729,
 0.5184794437751166,0.03747122998294483,0.9397664685849749,0.7673612227927643,
 0.5537937890749993,0.201280552070348,0.5607621328981479,0.8239609885155014,
 0.8267789235665804,0.02511836737060158,0.4455299251406699,
 0.06670677680884585,0.21697301582668216,0.6553792521797537,
 0.3588603506386451,0.18567000346282725,0.20355044849313397,
 0.9087471942836569,0.5281008302814759,0.7378529721652137,0.2056585524251694,
 0.3714920316501824,0.9546584947948533,0.5240616790431446,0.6134367377059095,
 0.698930460646313,0.08044955825768418,0.886310874135723,0.6683961235547666,
 0.11057113666689633,0.38608307316388113,0.05912115522076378,
 0.9209119760263135,0.856625608438508,0.3105033772446536,0.47754945070142296,
 0.372745589534365,0.7327179086060092,0.552804892331486,0.4656289820321823,
 0.9617038701193877,0.7320707977446078,0.4924676636088533,0.10777833889790311,
 0.30723073341370255,0.7415834098791924,0.12126989984975567,
 0.25607389570249783,0.588815009243351,0.8922789635010016,0.7724309841107168,
 0.5814784301055281,0.8559922743841312,0.5139467438573564,0.39945930195305535,
 0.24183642825281204,0.22386232517768523,0.6217266281710453,
 0.9801707133462292,0.26917101916667907,0.18305064771703972,0.512715617240708,
 0.04772041197678978,0.7931938016515921,0.4193670777131695,
 0.04227161950055136,0.38391420798405884,0.2781403347219249,
 0.007602030240302149,0.47715459007163197,0.7657616385073072,
 0.4310857398559913,0.031455427472038555,0.7151418436090468,
 0.6161155527446295,0.7641447039785917,0.806529666801035,0.24509438856319943,
 0.12529762708757086,0.9557246492557971,0.5613227184303468,
 0.07427100374056872,0.2145840553924001,0.7073387935502804,0.6938609112414996,
 0.07693360036640562,0.5375542742449961,0.8458502217944357,0.6082979603573974,
 0.8596943816273567,0.9254926339259713,0.38778127959606024,0.6503173197669289,
 0.06785915180342195,0.686273686373587,0.1686268912736224,0.5914529668120698,
 0.39965346145234315,0.16694658366392368,0.8393055288813984,
 0.9364207467675407,0.7905392527052209,0.7564685115825337,0.7922158462661403,
 0.5546114233425943,0.1803826091564964,0.6092138450089887,0.1400738159473408,
 0.27795714379244596,0.11533462016544915,0.07446438681073475,
 0.8298816101688625,0.1134723543926418,0.9990830687467738,0.4759356385469431,
 0.8753673714196977,0.5326271769992054,0.9581017037996863,0.9883261103917691,
 0.9215065540587591,0.8991358926934221,0.6100977761010912,0.40048202605097283,
 0.6413376585040315,0.2544153603318575,0.09839339503783684,0.9953898238018941,
 0.23603609914781631,0.27632360823590973,0.06422265672926875,
 0.601061041441343,0.4024298714352089,0.34677656308197047,0.9639485602916577,
 0.5900163481556614,0.7732934524038168,0.33010010335847095,0.9499502264597969,
 0.9999916926740395,0.38447921631599136,0.1299974700543466,0.1445566865327662,
 0.23426086170154237,0.2668047807494388,0.36817663363945297,
 0.7195352239313956,0.3330652693415521,0.5691393522890243,0.6655211021738304,
 0.5764216802720915,0.736947154415421,0.7574750701611345,0.9973026984371346,
 0.16263320524578695,0.8087757207977386,0.5277913888777742,
 0.0070979845759849836,0.12151994882538808,0.9138823147392742,
 0.6058244863007274,0.4606285739535356,0.644512922280141,0.06162317685797691,
 0.6484719352434742,0.9672687437671091,0.2052633215542261,0.8734981863242705,
 0.6927861072102827,0.502134931253533,0.8594294522514117,0.1652095048502441,
 0.24697869533570982,0.8210824438474776,0.8443537857216318,0.8630433340512671,
 0.729670240889184,0.49364637519744303,0.0628328825783826,0.3855588010115433,
 0.8196650297393008,0.43394609326207967,0.2712368556870183,0.8118138308477139,
 0.5588318497074739,0.8717349359210392,0.07182506683072276,0.8509162789431113,
 0.8671485456498815,0.7005896138051164,0.7249960220216749,0.2845265593271695,
 0.8830286561014261,0.2657473486288434,0.5003788678016995,0.9600655479886546,
 0.7650873963207268,0.9049490715496952,0.6072957147425828,0.3593042410462256,
 0.8101831179489742,0.7152307536270978,0.27070204302284717,
 0.05483551585344337,0.7292308871985422,0.3099231056339634,0.4977098803793669,
 0.8536450042061269,0.8678716819406503,0.7686548200521894,0.08586321629008409,
 0.14200540785310267,0.8104386799124644,0.9733209614383311,0.9290476251399123,
 0.8970599605860322,0.00757917832284849,0.24259308311775674,
 0.6334812392176695,0.889719161603576,0.08780476109620194,0.5226257252189285,
 0.3138811736166224,0.5321808032618285,0.7164805514971595,0.42127679111043304,
 0.509772014418904,0.1954060576917258,0.20497063845022478,0.23009751102642317,
 0.41279964957915904,0.6246871133329605,0.02914714829250975,
 0.7756385468203033,0.42643252436768686,0.4366653811251586,0.8476400995638508,
 0.23988685407343313,0.4201630179070912,0.9556668160597918,0.6572948944825714,
 0.4669162237262664,0.7519647378918731,0.25189014514490626,0.6716637301427859,
 0.28526240945222037,0.6952756325087768,0.9439530787299029,
 0.06305290178434275,0.42689407203952856,0.5384023286183457,
 0.6162068090053345,0.8317648916413666,0.5521891035596662,0.22468199765407104,
 0.31079299167593466,0.49168240838962496,0.25458809339182253,
 0.7398202151131024,0.7151728920343693,0.3668831636053518,0.8707976103579158,
 0.5975140486663179,0.12632478772149658,0.6613729229813278,0.594278166931686,
 0.8715674045982317,0.7578712879443031,0.23845919699830576,0.6206517263491221,
 0.06333138789073534,0.17371272413787198,0.6959522416322295,
 0.49587045106722105,0.7108015716689167,0.5884570375801865,
 0.04379903679958852,0.6655889431002329,0.8134214058010215,0.6083458268796152,
 0.23098698903588866,0.6235472633281155,0.25614754917104166,
 0.5690745293932358,0.07343342244571383,0.9893856563930192,
 0.14505971132665252,0.9600650975027434,0.900592774590862,0.7092189807791036,
 0.3628790938758166,0.342601564894478,0.06956978599269892,0.04872449226791464,
 0.7894452091553397,0.4214608806853126,0.3505732327602389,0.7017742730209933,
 0.5141826572143766,0.121743194500337,0.031216923159983145,
 0.39086035128218066,0.5537281676113435,0.7989205906824952,
 0.31446844328511836,0.9660843555823634,0.3465240767635912,0.7379601364740856,
 0.4741325158750461,0.508991395045793,0.9867458170786269,0.5926837567004292,
 0.8736044616736774,0.08006232083497056,0.2185326039207819,0.6500617679320674,
 0.364840680800224,0.5495992748951608,0.35800287432850386,0.980248953448789,
 0.2028285354253172,0.19173681121441533,0.4336866363538251,0.6005757423704098,
 0.9936537994636818,0.6630758909541669,0.8951962189579666,0.7620513041017124,
 0.09579235237613515,0.3260770081996125,0.9267461959215197,0.8464084693687122,
 0.2567831590971824,0.8266094918475178,0.809227627199107,0.5597835137139553,
 0.6238397963491985,0.3458818649086607,0.19641269637190895,0.538482022923802,
 0.6273692023837522,0.824711320481048,0.1494348784754851,0.9972250865042239,
 0.6888427419648991,0.9376717911359722,0.957264372832727,0.26689208933798414,
 0.4805750409043499,0.5002917766647464,0.5485766444230441,0.4361442679116422,
 0.896020173308965,0.7739903443636584,0.6026427197456979,0.5380627243525145,
 0.8286407146121165,0.97671928387173,0.0980351745711383,0.7678047567935317,
 0.3864434727664362,0.7972656811030461,0.8175694909898625,0.9221443824036035,
 0.5871739395322825,0.5757800608330957,0.5365000064868073,0.5775550529034058,
 0.49150446180254814,0.1156943376838594,0.10825892674501925,
 0.1751869046666803,0.9775016692656324,0.27638203803713846,
 0.44050256323016435,0.4676175656435897,0.46187926072188323,
 0.5049679069919574,0.624477579171251,0.27113561148664755,
 0.024143629329147154,0.2669063460702805,0.8459065835646388,
 0.7945545343227983,0.7541662346727713,0.5477570159394293,0.6824411352279206,
 0.8677541268138568,0.26759611009217266,0.1751228855664515,
 0.36608470756208644,0.23024620363758785,0.09092803897719515,
 0.43281271051840475,0.9455415935773934,0.42004538784312606,0.856829424011166,
 0.3922787489746361,0.3422471395635167,0.44297117556963106,0.5024628667778609,
 0.008550852589371716,0.9445844475698953,0.40714703288061205,
 0.49102557630154875,0.20696968928677717,0.1192661308346562,
 0.04690581973698105,0.6776140673789786,0.5737746664835633,0.9970245126692406,
 0.06014347579254853,0.6308455348486928,0.015000395806200606,
 0.22196554078515485,0.7661459760195684,0.49424370892378267,
 0.9458356887883129,0.1996600316734013,0.5343851098417036,
 0.019771087789436703,0.35298595361511365,0.07831328337607302,
 0.07126539452929848,0.5327761697331923,0.2828145826824443,0.4487156875584859,
 0.3989538890851454,0.9734741160781539,0.6505847501782127,0.6200338424067242,
 0.7973485772310072,0.18323232077442053,0.20108472740970895,
 0.9264218890591475,0.8532522466336672,0.15594128043487399,0.6364101942820317,
 0.2603147796812102,0.8761964822682741,0.12204727600173837,0.4545232039598359,
 0.26439830962985134,0.38073012199431977,0.18755720768735606,
 0.598182985614038,0.2186793274865947,0.7665899413412087,0.33931067008343985,
 0.7023381850506898,0.493806814168387,0.20791517831501571,0.39026518286653067,
 0.6600981693847734,0.8535635959947314,0.6507835566653885,0.359682375547532,
 0.9550807536741609,0.8253813314428686,0.25691037733765665,0.8936925692000166,
 0.8518717929314121,0.33304164050055984,0.8738928905088473,0.379320620076974,
 0.5510061362814371,0.13969911695270587,0.39159929258987203,
 0.5495139683323642,0.2779651811096404,0.20275160550023585,
 0.15676250491220356,0.2695680059158584,0.12165296402417591,0.512049458316046,
 0.8347303354028446,0.07659223322061459,0.8488076105765658,0.715532220289038,
 0.05997627119191418,0.1098294803405343,0.7852069162520436,0.9137097766110849,
 0.2545531395537528,0.383882189369652,0.15960641229459005,0.31721019808571804,
 0.07056858797756793,0.7906805988874688,0.5917638134073253,
 0.08332135433575305,0.505685959906908,0.15026048145046345,0.7155727165588459,
 0.42198685358235544,0.5266947564532601,0.63631427099525,0.6722607634493393,
 0.9393130046877898,0.8628945758089582,0.15450960440637673,0.9644460225078814,
 0.11503248919717646,0.17975176841584595,0.11077552800430923,
 0.23398647321152344,0.8952922293057158,0.39256065704415755,
 0.2973996704332995,0.949771378033942,0.9625347372155246,0.4555366946130879,
 0.27071006986051427,0.09549947547100823,0.9885325187754739,
 0.0640796422161174,0.5254459836645508,0.1500219786649043,0.8785734413105338,
 0.040539175360803203,0.5510206452716311,0.49915721270755053,
 0.5542872712136139,0.14124186550822126,0.92728855941562,0.8788216145925866,
 0.09943482849060103,0.5830117685387202,0.8020242229571865,0.2119081672239892,
 0.9312016595722403,0.890968333565311,0.7951897119911534,0.4370720030324522,
 0.08607958929048598,0.0017987733095757852,0.28291229996568446,
 0.8210925636244515,0.37367692712987266,0.019548650160225334,
 0.6203679208308688,0.8179818681349874,0.5098077267682556,0.9130047256377621,
 0.6263302137007197,0.6903018346562607,0.1981563513390574,0.8919429243132877,
 0.6628593763074909,0.21710346971419892,0.5939115727225648,0.8708030288999014,
 0.7408886440253807,0.9713725436092505,0.2965666338741595,0.6927813042095583,
 0.7224230943077072,0.5044435380274676,0.8100665019999516,0.9729878808169585,
 0.2925087876417115,0.44468342483411827,0.5210798503126075,
 0.05041830176027662,0.5672382511755191,0.49299474443659264,
 0.40140691409901696,0.19193337160334267,0.6906183866012929,
 0.20934310332123418,0.5377434419498057,0.9426744144851079,0.7125813127649213,
 0.05354506348115118,0.894135764596956,0.7826817205351162,0.6416262295741976,
 0.7289379051093337,0.8413316438490717,0.982980895740192,0.16347234305477398,
 0.5544503460373365,0.13556559703922655,0.46016473490857845,
 0.0942696745154441,0.4204129011718667,0.6604603809763748,0.4803461984070999,
 0.014747570076317107,0.13754582379758185,0.22044157349840343,
 0.9373773166070746,0.5385532647502865,0.8559621869239428,0.5264094914167258,
 0.6946769252525721,0.4855869121254559,0.361691919614505,0.4955708007464168,
 0.3184618494605239,0.22562196034239868,0.19753396216830643,
 0.7145117907630946,0.1495691060212735,0.6372382722590944,0.41494450661075877,
 0.4953657614727379,0.8620880249529232,0.7790828972205336,0.246963163064163,
 0.4442666093241949,0.5576203033419357,0.1501240208561998,0.7112170267376655,
 0.032546110774244186,0.2548231412836026,0.8909930117328376,
 0.04737163387812271,0.40858298773655655,0.7193985460334279,
 0.8488695893055921,0.018080155844534596,0.551816054702659,0.1869203241713685,
 0.22413627389285595,0.022130620892083064,0.7024115805437575,
 0.592193692333476,0.17656612087466483,0.18119991371620414,0.2801085006765859,
 0.31860009202607986,0.3403125513955909,0.653111960736307,0.18131488250655026,
 0.16942950925210598,0.9512565633580385,0.26597550822878746,
 0.9569977449566369,0.5811878093140574,0.9418192115572048,0.6609490370176995,
 0.4128341891089029,0.7479193135331433,0.30162573252642866,0.926712242942798,
 0.3213498838730017,0.9945658077696397,0.027166338882034458,
 0.08559357430663384,0.7518089217127714,0.2328335225495184,0.8063927642259232,
 0.16719698112738524,0.6785221155320561,0.06450855918879017,
 0.7592035844350247,0.33632362109086034,0.8086770260444216,0.8745957842069311,
 0.5154816659930339,0.29053841337351094,0.8407017057337457,0.6965797267241488,
 0.06450610476915886,0.788241020151549,0.6804330283970059,0.6311906506685114,
 0.1601365601120186,0.42049322265910694,0.7874144413722601,
 0.27649983652322385,0.9970882484276526,0.7060371747515157,0.5570563132376455,
 0.1295493372910601,0.8515654744532224,0.7349052154032799,0.08227579020503417,
 0.3948820149954847,0.4249187883583172,0.9200659622141393,0.8952506542175552,
 0.9188414600741494,0.293649999473669,0.7265764416588246,0.0404460964395138,
 0.2812345778830734,0.16679356241537602,0.7015875612432305,0.3674996291090491,
 0.18951622704568583,0.834735315690712,0.4261103676186794,0.9799269143747663,
 0.909253315206353,0.7592874146687082,0.12987696861646159,0.4634068462482861,
 0.35025714188451806,0.8232679134514567,0.8274508325297677,0.5039642757431334,
 0.6135650102967715,0.6727016765244979,0.29244154474204853,0.2621545701063258,
 0.5275027409263526,0.38700381691174834,0.7146899451613563,0.5755228323448485,
 0.3487699987570765,0.6069998902775645,0.566360280624341,0.502202562262371,
 0.17084346242851456,0.7792803519531891,0.5913896766678773,0.7453955638601756,
 0.24156343946844383,0.10379790669556088,0.4485210680375643,
 0.9035749024199636,0.9884664085124519,0.3104270960735487,0.49095263717971505,
 0.7930163228966965,0.04759611718809853,0.17257168167833392,
 0.18142429399733417,0.4968630958680249,0.8389044661913743,0.662247694382637,
 0.9665421765731615,0.03732052274623232,0.4436824580782015,
 0.21797298604703097,0.9041247057463027,0.30307794283432776,
 0.0036752856690398517,0.6653004935655792,0.558846412967447,
 0.11499645040611028,0.6440974739080351,0.3462295875275073,0.712670544783409,
 0.6408801739806438,0.9615744615328895,0.7851951164810757,0.2909261020740117,
 0.8818653365902872,0.3416325302596994,0.36411563383011547,0.4248417122061394,
 0.3774096773012581,0.9185427631447558,0.45086890095628407,
 0.013942191185379427,0.6605531081178572,0.4955552598049755,
 0.1983136084336703,0.1348374400435478,0.25232048463954126,0.8933853873043404,
 0.6944314431394208,0.23106863458779447,0.5387610891719781,
 0.06598520076893422,0.13479118157674574,0.6843273487195175,
 0.37780553735783695,0.14651207770194496,0.8652234464756808,
 0.8820188255846418,0.5990892472382762,0.0011139829538403534,
 0.9954859042821961,0.7732813387162805,0.32666974736995813,0.1514718923480528,
 0.015802328097400564,0.6233119816827075,0.7597019425199141,
 0.7410343060238469,0.8780413705164594,0.07725722192327666,0.3529865221603231,
 0.927103746651539,0.8866219147250213,0.3655594699836181,0.005997130225304126,
 0.20622215782656672,0.7095218718358882,0.6257377855524263,0.5600996983955847,
 0.16915757722669633,0.845119306650089,0.2645362196149019,0.09400585910234593,
 0.41961185411074553,0.9206042955791909,0.8588430760751199,0.9968684722688633,
 0.6116802475387383,0.4360736054196648,0.02016090302944462,0.7046339501380634,
 0.6866663949019872,0.4231889318066444,0.8002457336282461,0.7578143995561841,
 0.6955355960702552,0.25790121465083793,0.8916529973407396,0.7750069591687087,
 0.6653608731763089,0.08323757668441767,0.26323122307992874,
 0.02522674274645531,0.8276605904739043,0.18596441644898742,
 0.10699141675457213,0.2022782480343024,0.2549735262991052,0.4304092819756138,
 0.7547962979965055,0.4611331148173822,0.4024555415533345,0.5337062704573088,
 0.12694958882521623,0.4860877010188248,0.5346983683133114,
 0.09825302097506361,0.4094304847180016,0.237237679313253,0.9829869086136301,
 0.19705479427103767,0.7231784893984964,0.2918145911488893,
 0.14240869837504522,0.7833653210911694,0.022931598760945415,
 0.8928448359791819,0.34457366530906897,0.8155544382980785,0.6129462500074768,
 0.7578850371090435,0.611684861355614,0.13685292086864354,0.9550160904118045,
 0.776307525629008,0.6671515244671874,0.1091182959517204,0.5618797668252377,
 0.10736301419822014,0.6971515687537785,0.09225589545655699,
 0.6316848100704975,0.4681382819128742,0.8743102663558239,0.448632137840548,
 0.16459732177976105,0.558053994379135,0.3834646998282112,0.4712994157919691,
 0.21756782790295648,0.23410660509924963,0.2545839955975008,
 0.2692277493106684,0.9230392574769406,0.9759747909421166,0.20415784841569629,
 0.011819687197930029,0.41011146677242516,0.6682835461916087,
 0.2666825425491781,0.7681851437269809,0.45553306707193086,0.1874522128926852,
 0.734572895973377,0.08515917975795972,0.0014258210551946204,
 0.14773678967139636,0.17235843475854173,0.40739180989541157,
 0.5931739109749481,0.0040147142452359175,0.9936240872223461,
 0.2676079692224287,0.16366341758964353,0.32873686595501295,
 0.6689790326596067,0.18041270414145094,0.4892667614907671,
 0.19522541215014844,0.8697101453195213,0.3923996655510975,
 0.23217399463830846,0.892880003890137,0.24648238858040972,
 0.30425388176246626,0.6615492036393504,0.6772021640269155,
 0.27391734395950695,0.6767525538422566,0.09370044484465623,
 0.6092781779197701,0.5317793904327589,0.800095562274286,0.1979687278857104,
 0.20701983967997883,0.5409679258624437,0.6516654899398707,
 0.29279187289583186,0.9263100328881185,0.6784967205242053,0.6809715428721153,
 0.5055440576038185,0.14813908659756148,0.32337355896836906,
 0.34967380420081096,0.15918397794325168,0.037907377482871873,
 0.9130146905811999,0.8496508418958563,0.06216341733591113,
 0.26942219183723104,0.8675329400002348,0.19276934490878783,
 0.5447015938826238,0.354238102161024,0.2624117980121149,0.28080662408470225,
 0.760035659212792,0.2954809253285976,0.6784123389730118,0.5687932355419103,
 0.054595779496389496,0.33285956902020775,0.5969366829576311,
 0.8874320797572244,0.5475040326775115,0.5180441880804117,0.8376544432978555,
 0.4586651756534331,0.5833739021656612,0.788553177890179,0.8913366330146396,
 0.3835609310473338,0.3115334896543601,0.9732227477552664,0.8659153622792657,
 0.7955799476018138,0.30963043274835933,0.15692936662108226,
 0.5275512559253094,0.3668930715741199,0.6168200393346724,0.7363183713960854,
 0.08968679249129052,0.5247107236101799,0.19040111757160616,
 0.6186269449860304,0.29929758249074656,0.4288088620092283,0.6954364665332224,
 0.5936365292368468,0.8592361097778303,0.5458955233094019,0.7611425928157098,
 0.7050750523001017,0.915408141467335,0.07014115837960178,0.6025167541687063,
 0.5648835941743428,0.8946795229034407,0.6422546042116153,0.9774853543999604,
 0.7435863978717137,0.20901866904412403,0.7663528685773022,0.8617384649652597,
 0.8742296784694749,0.628082181679329,0.021183752015249913,0.6360966461642978,
 0.18838856596218556,0.020939739328838525,0.19282900498613365,
 0.4478111743757376,0.4920573712143749,0.8725552999523349,0.06953719122751323,
 0.9555689901206903,0.37714761266320107,0.16132594992650795,
 0.8373042256480934,0.6300688004256031,0.09412297196866948,0.7036858722673628,
 0.5563899605168894,0.1640145388609212,0.4374288711322951,0.9236885419931091,
 0.7804218864508671,0.010509332271210536,0.408907734585217,
 0.42627743153159803,0.8939300931692269,0.3989798722543171,0.3606969797140548,
 0.5659477803686134,0.16369235260588566,0.3495838232409696,
 0.36527682810698425,0.6037337308522931,0.2037207728948922,0.2226500637114155,
 0.8217377868831797,0.21544309724962518,0.3268582833881758,0.354779337896444,
 0.5103494772050607,0.6450267007275924,0.17140720272394683,0.8491932031401139,
 0.8213905925091094,0.539527673627986,0.8223591832988643,0.447554922661451,
 0.8343587026886304,0.8662584875702152,0.769991996621294,0.8661098448617964,
 0.8735851967237701,0.5925904538488296,0.5169341946784864,0.8355800671865686,
 0.655590027993568,0.21218465296935263,0.2226877383285195,0.2949856545994194,
 0.9695124822204801,0.9623738418668114,0.20877406977449076,0.473298152700083,
 0.8481035294074258,0.8223870718327604,0.6386779800109331,0.08376776953526943,
 0.4890007165410243,0.8892690099982338,0.5030685826565556,0.269663019804842,
 0.9460871625510117,0.07605420586433065,0.8942577715408181,0.5235116271365261,
 0.5969413092670355,0.5481045078189963,0.6714972296292862,0.01538153919849461,
 0.742469286469983,0.647183459459084,0.917631354567028,0.9341658462158334,
 0.3049293757685656,0.5047497060927211,0.5173606186842806,0.10081442632462467,
 0.32062823261608353,0.0925601949254854,0.18808693186851044,0.200910238741835,
 0.2107195184851567,0.1584743691466921,0.3096337900409839,0.7173221135846962,
 0.6321526578505698,0.33406192746025254,0.405984172119388,0.913369463237101,
 0.004315372021568065,0.4162420363991082,0.576866310236143,0.8982239042530394,
 0.1487237079348659,0.581050622318397,0.3427171176917231,0.6356441822757282,
 0.462617529722646,0.6967002127287046,0.6993377728156122,0.08087092929004436,
 0.019749053740846678,0.8734661149120632,0.8153913986381166,
 0.2933185249304757,0.5685179217729345,0.9407228710341929,0.153566326896864,
 0.27067400104018013,0.717955397571329,0.22257310292282362,0.8789541799889743,
 0.7770627349705748,0.5647039571555099,0.8866562621721179,
 0.003258749858087162,0.7295647275763801,0.01903531309630957,
 0.2118598249637842,0.34965949889141335,0.0601849079564476,0.2894419778166203,
 0.5346636403786303,0.30271986376934923,0.5168910564917135,0.5575569678548121,
 0.15024336730482168,0.2245779273458708,0.6475914607208704,0.6099864759504603,
 0.7954432236854508,0.9569344038480164,0.6596046982373838,0.3384070042711447,
 0.5057912471086967,0.8920156789073922,0.5901928481154908,0.9414827538111481,
 0.4346912053540649,0.8060869878286449,0.05271514856848336,
 0.009098187524888113,0.8092803283240235,0.12786222012272952,
 0.0574986936232319,0.1260155469109936,0.7890556860055694,0.5951824570551052,
 0.6011407265328116,0.5625195924090778,0.6299340800610811,0.48946177350271136,
 0.8761398492845427,0.6530139291819131,0.4207909506271612,0.7878074475843859,
 0.934595442256632,0.7777565909595334,0.24239020763995,0.02448425396708842,
 0.0012362394616872718,0.6351039789317272,0.6019390142384531,
 0.37282615372979644,0.8238257861243216,0.28055580134835945,
 0.23611153152934372,0.7841897230064921,0.7093866161199522,0.9292574526927663,
 0.39939881813673894,0.22838577242084623,0.12450116512210818,
 0.791356548829838,0.9685468649700867,0.10927202507350053,0.8795274319666425,
 0.11767618103529442,0.569063647981085,0.8992996894132138,0.6575742154843226,
 0.02437753336029136,0.0742277729853772,0.6465719894517855,
 0.18289193920887492,0.7264061155430507,0.9217480664699625,0.8049053092749563,
 0.5261220099001074,0.9391958247093479,0.09704353970167656,0.9058311928348427,
 0.8123531511219713,0.16340076744547205,0.10000179583885249,
 0.5371829121309512,0.27862103082755807,0.9717331728105015,0.5865561794349727,
 0.4356800740107113,0.31574683733875264,0.8430240602620237,0.8715679020168903,
 0.34928323422523244,0.34642250137255237,0.5365797239314237,
 0.36873339971399965,0.6253759629089921,0.2083485435325274,
 0.42873969110496823,0.9705374794861706,0.29888891156321007,
 0.10506264685278455,0.9661731308457339,0.1824312278267607,0.9588580482023459,
 0.21822455141044672,0.7270827845536263,0.6208175258167645,0.9349774062189784,
 0.7315346052615054,0.1154384799122441,0.0648981225084806,0.04855351611684122,
 0.19586346574651559,0.2655061801980978,0.9806642454933118,0.0915884038870779,
 0.21332681710591384,0.5953278985845152,0.3290661864204558,0.7021387952098752,
 0.20858618993747646,0.7062661796205986,0.9075869748252081,
 0.35848091604173815,0.3761397679044707,0.14036718486941147,
 0.31037548148932204,0.8822631071855973,0.850634092775195,0.5977168863280256,
 0.6440131898495722,0.8241302615631054,0.9245216335977008,0.03565238022029571,
 0.42350199692433876,0.9819653127320045,0.9200632374592409,
 0.006251093508519934,0.03481764350478789,0.9184094985384933,
 0.6010114301206779,0.022696513366029625,0.1936279385855728,
 0.0459852253376799,0.3575380853774264,0.4434868161128275,0.7616352276603966,
 0.6039612495561413,0.6129448430890534,0.36936095478104736,0.3972325817223634,
 0.6491513665900938,0.9772632771245333,0.07154940849402036,0.5962034117226085,
 0.09981868101089075,0.7924186184322024,0.9380275022120639,
 0.17008340691191437,0.401504626655526,0.1434723414110488,0.5214838465418327,
 0.4497054700030988,0.29047754176556895,0.9608540731757934,0.6336197758666647,
 0.8751752543042356,0.7943869848291543,0.7139186140063359,0.7168064865313447,
 0.3318593785874131,0.769299245066073,0.6602413173961016,0.12271279481995678,
 0.4422027816436739,0.8692741906132834,0.11961608915022004,0.8571908440565398,
 0.9065348048411769,0.3336322090496536,0.6369103608257775,0.693275911621473,
 0.46041386148257113,0.05844539542604921,0.5360675014712227,
 0.7815052371245245,0.4895260690943011,0.6227422237891167,0.7458931402092271,
 0.3751355474941158,0.8405381599506043,0.330869698268775,0.04785444122748517,
 0.06286973406554686,0.42954483586674896,0.5959841335371723,
 0.22472531663116535,0.4099882505581107,0.33171819044950623,
 0.9291592380715799,0.3759150208817629,0.14624445614242876,0.551168256346829,
 0.7384891967941962,0.41600350610860903,0.20142530749963683,
 0.9324663926140664,0.7761505320264374,0.6756190807474013,0.3312920830904368,
 0.6086355895424869,0.75155738007631,0.42576250534779314,0.8978275324295895,
 0.6745301856863843,0.9584744473268584,0.8610071129044236,0.5798133213806597,
 0.8456080671660681,0.4637399307819994,0.03419074006096967,0.7018118786648148,
 0.1442505948164684,0.3632623407831175,0.8083150720227363,0.8024745141529016,
 0.04705519018824156,0.035694506168167606,0.3432092063094062,
 0.9269218139238476,0.7832541112497894,0.7501536814483569,0.19686135920581516,
 0.8715364682349981,0.6822094070457281,0.23314123592975966,0.9735186921454929,
 0.6426392715498076,0.6508339247647014,0.5058485695079251,0.18020593822859,
 0.8972789042357062,0.5999952557194176,0.5608066593721558,0.8844722294827091,
 0.4264395875458191,0.20024383522500921,0.0727140053780867,
 0.28841031166254705,0.834509987745232,0.5412637386567818,0.28768285060416765,
 0.7131542795295902,0.9539653692543463,0.7187644067601879,0.4354111182189554,
 0.16718518116566228,0.23498630969002188,0.9926083915151842,0.941823955843286,
 0.043119190639560045,0.09977801936560471,0.4738303065628058,
 0.0656831652572839,0.03726492713998397,0.8550282059128615,
 0.05439283552652263,0.7532410153089301,0.061672296840458496,
 0.533228111729247,0.46255302335370974,0.49408027643302355,
 0.23889155867910783,0.15709681641409778,0.7094029023334733,
 0.3440593612898699,0.19271518709222213,0.262373980027524,0.712239116522504,
 0.5149083686002064,0.09198079611546217,0.016667761040493057,
 0.1269480575862626,0.5380134653298418,0.2861704378343439,0.8867589821302797,
 0.19651224590493854,0.23888505324107268,0.10951178282158336,
 0.006086673606780502,0.5625390220247722,0.734617609950473,0.5317267348698768,
 0.8940610370030178,0.21437934905268308,0.10659464574195565,
 0.2800166599377705,0.7643433679384355,0.8792630095053058,0.7253530654478336,
 0.8127205181591932,0.4397895428850266,0.39070439090845077,0.6989081983169247,
 0.047939403122004176,0.652256776239083,0.16747584737780097,
 0.027600504334118425,0.281154447083407,0.7514202158427289,0.36882381654774,
 0.5187593851013403,0.4581204528101034,0.30036175147537725,0.4183406011283888,
 0.024405717884007583,0.5343822234461294,0.6482668623451574,
 0.44967509882516987,0.42442779505531225,0.04247593045034792,
 0.37337229459427046,0.04961333619729347,0.04357912902163985,
 0.18964337770828676,0.9371509265209974,0.3549109060665451,0.6552608792961039,
 0.02477270212290761,0.6519770326601628,0.5513175583011676,0.4856157642850537,
 0.7569564075427713,0.26479168721314295,0.32696455044723083,
 0.7913129471176878,0.7481315484271447,0.3453827105647902,0.5611409399350855,
 0.1263842850907958,0.6164542061369981,0.6417153377506248,0.8362674410507329,
 0.4321937737647238,0.09743355468570547,0.39947100619949805,
 0.6908042465430364,0.6600428749709224,0.569554466450135,0.6678082691795107,
 0.35644127895615074,0.2739968120013748,0.4162145726702373,
 0.10092741425536977,0.34342722490083355,0.7283768102417728,
 0.2681466734663217,0.7990864305045271,0.4267543013541806,0.5975678359895774,
 0.8068240446947005,0.45731075902815355,0.6425015415156675,
 0.46337984321685166,0.6386310581585761,0.027867051155354705,
 0.5962598386336112,0.7583494653722155,0.3390065326371723,0.32626370592839926,
 0.5779844937331413,0.15630512147746223,0.45850120406110795,
 0.7548146793264935,0.2647678627595248,0.6595929879556743,0.8433293533784686,
 0.7015725794164077,0.4404244804919092,0.15977996318850085,
 0.09023234117390821,0.5069307994487797,0.6882367315198575,0.6311928674867061,
 0.9882936862672356,0.19611648564502104,0.0637834685211568,
 0.005504834652723556,0.09094459484437134,0.49293528791701324,
 0.6638234322702536,0.570787937930838,0.336702314675404,0.15027166400021152,
 0.6499400145012669,0.6355067025653064,0.2941268861302204,0.6613759328946716,
 0.6137063621595836,0.20637553382105067,0.6958384559333746,0.7396196827091347,
 0.6257238170332498,0.5238061715242663,0.1237075486199618,0.5405335112831807,
 0.6540576983119861,0.6137973167573968,0.19700815364280122,0.8228251388626351,
 0.07705722484266364,0.6060542042292653,0.1256874089940887,
 0.24019408421212718,0.5423111401814782,0.15131373756503708,
 0.9734778253303085,0.16821580257501123,0.6537803914509638,0.4824570779851086,
 0.32535650755516465,0.9073934826967403,0.46287272407201185,
 0.39517052299041566,0.5294898056840052,0.680531278665697,0.9939752066610478,
 0.34002745448079175,0.041353347594438206,0.7723617892465676,
 0.9444837600471067,0.6883450438587735,0.018594083866795508,
 0.04668275539040456,0.7867737046874546,0.15017005666843442,
 0.08530479614086173,0.6425527598910189,0.42654528728376606,
 0.5777295313802991,0.2787597688428276,0.13076018454907268,0.5066369169833753,
 0.12845263176023136,0.8126742666566178,0.9693626912639476,0.4586593407235685,
 0.03668667467872644,0.19135954994207938,0.8684115950951001,
 0.6459792476387627,0.0874541602015464,0.6414918518900902,0.2030216668542949,
 0.4825676559033655,0.8897978809385378,0.6502605954528199,0.03358876695070245,
 0.4846601928143617,0.6712912711349607,0.42374271920617557,
 0.010520262261783042,0.08603951616535088,0.5832059638669804,
 0.5429768232262376,0.3114362513382083,0.5807345098791962,0.6270001989737146,
 0.01738590367925774,0.8482044885819932,0.37253380298313066,
 0.9589220845625133,0.23746150709312008,0.46068506756052663,
 0.15090625703172889,0.46538491423498485,0.06447765099789926,
 0.5267031228802304,0.5187704587296782,0.3677430269033545,0.700288851506701,
 0.9592603859245776,0.8022879618301311,0.1686610341121315,
 0.016748036129227595,0.852831612955802,0.5472383219837158,0.9965012543137507,
 0.33621120306002206,0.32580364177357435,0.12201029414434217,
 0.4484049778608683,0.9604927315875519,0.47108327879038925,0.4511067015601854,
 0.4806389552963546,0.20888438238046525,0.42031331453670884,
 0.9268986487176041,0.010691692515190976,0.5084569134541558,
 0.4141973883625806,0.4761430112748949,0.43310928686894434,
 0.07255390753753632,0.5251522938711073,0.4872229724002112,
 0.01463308216697845,0.8359951368782662,0.9274814087005008,
 0.16741087855170667,0.19219869859212402,0.17231429471054271,
 0.19037264863817227,0.8611716528372644,0.6985724510811491,0.689546475737508,
 0.35624640568031984,0.33004475490413787,0.7290517799637846,
 0.42669449594965314,0.226269440630962,0.8855551664513024,0.5890750677832941,
 0.2776634369354749,0.18166461516994015,0.25272002105269786,
 0.5869429494727794,0.31113832854766077,0.4278242450424632,0.6996970982508615,
 0.0812671051062781,0.1830147118225982,0.1568208473572259,0.9539790862679183,
 0.2261085753225478,0.34893026211977385,0.31488612888451717,
 0.8129947096640793,0.4802884055361393,0.7862512563826088,0.26489393783879955,
 0.8688334738573241,0.33596396397921147,0.1948711866784999,0.5186360864082159,
 0.09597444559557078,0.8138326938566849,0.14462968927307773,
 0.6599925748610197,0.7559187697195431,0.9785300817336593,0.15656182913957306,
 0.8287729513840734,0.8526969023980417,0.3433754532880483,0.6688763907154032,
 0.8783587231796813,0.7614745980885762,0.8032427940048172,0.4474775809685423,
 0.05056457404462045,0.9591218485389558,0.28567764219187486,
 0.24085122137643789,0.38885808375610265,0.1879962854346532,
 0.25723103807846937,0.25794727453435795,0.4340948514696096,
 0.6164045485840578,0.21639003467302864,0.6030785105473465,0.7622994038073082,
 0.32158025428976034,0.7147893010916696,0.25028613821325785,
 0.21561417429063812,0.747896546072924,0.06306193317283249,0.3064177094976197,
 0.5830938066641318,0.7603643737012733,0.3268365926642489,0.9645773671503635,
 0.6101820698029439,0.9277488951562147,0.279707476572677,0.30903351882835806,
 0.051148762821288773,0.395648732007027,0.7601259125952123,
 0.007528766784583141,0.12716249910919308,0.9755170439106192,
 0.42293241400790527,0.8524489747325003,0.4129687087984374,
 0.20957725926142723,0.21799353132157517,0.8489664286127503,
 0.8913570441634124,0.2641415331028292,0.44033439154330944,
 0.15505697240125194,0.32120617503245563,0.7915899242428706,
 0.7053106771128236,0.9107425075537742,0.019702974317381994,
 0.38914717354135975,0.06987714710460857,0.32784429418635397,
 0.2181398598482467,0.6477503072581482,0.1987010494291075,0.09060514780552631,
 0.8521234213587671,0.3334370071701175,0.4053736483538276,0.09128174637985786,
 0.8549170618181856,0.8123669229122197,0.1841186966193119,0.3736356042213318,
 0.21806586640820746,0.37161181530421894,0.8413667395546948,
 0.08527611516829059,0.5089812790718039,0.7203074401450223,0.8147728886952487,
 0.4386337965500443,0.36738439102696674,0.3054479835517684,0.2621380014598502,
 0.2960536835396176,0.12643661407559903,0.016934628815766573,
 0.9148893212153631,0.8049899806114568,0.06409216498549508,
 0.21835329804873904,0.5552604789696045,0.14621651619684306,
 0.9277345567464557,0.27024196537216694,0.9510060603330126,
 0.22260523551552902,0.6713141419987012,0.8717089385843324,0.8883821616730878,
 0.49537747278846,0.8135417412407034,0.823919148872738,0.9064895936405443,
 0.9975220892476055,0.30855403689171834,0.22492952573675562,
 0.4970804597998765,0.7797676287680846,0.7397625555339098,0.9752449635679777,
 0.11317996791092233,0.516795331579706,0.28087759511924415,0.7443108650061437,
 0.8609069745540054,0.9836319782890454,0.9150732705820027,0.6790746733326464,
 0.7298486533186547,0.2997273571220419,0.12266306687900619,0.2994901850811832,
 0.5904654312655355,0.3154145402945545,0.6857211067145114,0.05595876574711334,
 0.23828000576598252,0.17890314618356062,0.7158399677325008,
 0.5940061219403511,0.9085686153825036,0.6819299910238931,0.36992411405173264,
 0.36361960082374567,0.09816583209272178,0.4675874280898049,
 0.3812489129895942,0.21109758856025174,0.16017470906220443,
 0.11280376589940344,0.6821698135120609,0.2498094194460443,0.5396093358415077,
 0.40336904109020055,0.5503490876421987,0.6508655865640848,0.0681181852304692,
 0.931374480172946,0.0035796066295356033,0.3058385795925136,
 0.24764369820409937,0.7902195007445278,0.2285177106819194,0.732618592635991,
 0.18290411105001314,0.7875430076668322,0.335317772243401,0.09982411185821627,
 0.7495630573214779,0.5722283188892018,0.6533131663115437,0.1830879414893205,
 0.13885610661847025,0.9000673487747911,0.9541634902275187,0.7748474174320882,
 0.2037689996437314,0.7650817989552383,0.8859813833674437,0.37680275490386905,
 0.12822352856450248,0.13351815047354132,0.11630532404997518,
 0.919601010489677,0.3015805617804973,0.6197245418099211,0.5592268919446476,
 0.5351599798922632,0.2554781961221324,0.41976983070164664,0.9057431029765048,
 0.2016081076801186,0.008006659281463935,0.31569079169173875,
 0.10597315017469888,0.9440799208979831,0.7094821907591892,0.8826674737396103,
 0.5894766335565561,0.5498478871877931,0.7413490207704408,0.42554302497850494,
 0.739261399441988,0.2186774271638231,0.9370480666950909,0.06497566779518982,
 0.9335490392447864,0.10040769117767079,0.841252285008988,0.5715993549959033,
 0.7946996206734351,0.5960083817890256,0.18239390009495604,
 0.48415352040477033,0.3980256242412017,0.498233223828046,0.2890666305038311,
 0.8635959656470593,0.6392949863998711,0.9377546110042418,0.5544273821843047,
 0.09668289467725122,0.12817033384259968,0.8813266337473366,
 0.9596245539674799,0.37219594868502526,0.2970006189944151,0.6133769256688012,
 0.4547247145975096,0.6723292075814595,0.776554040390709,0.356445237518242,
 0.8667142051396335,0.9477025443048505,0.8479667783357195,0.08853358777702691,
 0.08006623326134288,0.9217921332018406,0.3670617391009263,
 0.07195467254620258,0.8590771837247225,0.46189656622660935,
 0.33266848764414125,0.00609843560301937,0.0837513513951108,
 0.6535279873340372,0.3904915764766701,0.7718909631579952,0.1362915025025937,
 0.05747043021104192,0.267970944253792,0.5561661812156212,0.20523040800277303,
 0.2450239209624676,0.6008386485746872,0.17196840193691165,
 0.08132872405937264,0.7432729201416817,0.4242747064046847,0.9126769916831429,
 0.372334713711702,0.3169259615572504,0.19168770980005911,0.1467131544431155,
 0.04099198666990744,0.35493850068175414,0.5860151117240886,
 0.4120461116479188,0.6241623380135332,0.5376878329378866,0.5435426123324387,
 0.656478262386722,0.13335034369973453,0.7328405714041516,0.7333956951621731,
 0.7536116929984461,0.6074282447885114,0.9080638926791127,0.36099016168165876,
 0.3323833109916837,0.6070003407921766,0.60614224962536,0.42427232392325165,
 0.5896660960430545,0.6234986895874495,0.6705153986162335,0.09224494084252932,
 0.6628222943356055,0.33936386155573683,0.05482843406133364,0.635790759673235,
 0.0022613423055676263,0.47174826374521595,0.5512184600389972,
 0.6538789399344742,0.9191988621850662,0.8542109544633119,0.7364894927873614,
 0.004823839537526076,0.4380123630419792,0.42262656165887547,
 0.8237089614516157,0.6913063323747622,0.9135281671697383,0.19582072806243755,
 0.809006222405844,0.3588291362563947,0.9070827102203032,0.40734107597669467,
 0.14735170124513375,0.23011024078649034,0.1161832627490531,
 0.49169780741636127,0.9947662811454236,0.9007161864364079,
 0.42358998086278843,0.3073423326586058,0.02115713366352001,0.506347319528583,
 0.34299457663353083,0.1669191264540908,0.5437446852169654,
 0.42991697340769175,0.8447115061761089,0.011124146282082026,
 0.10355314212734268,0.10225300194885278,0.5730735743896642,
 0.7081212540617774,0.11302873609715147,0.3997842495310069,
 0.17494865252381275,0.6417413618591187,0.3348442664009267,
 0.27046704823271883,0.025977854585671478,0.7881491108930584,
 0.44217011457641553,0.5522882158742057,0.30768675906042164,
 0.5180654362602458,0.1303355394112643,0.7941663889236412,0.7924198998048279,
 0.6328572063286722,0.6823486961964009,0.6854370426228211,0.7842600578473151,
 0.8668534392709646,0.8484423538959852,0.38062987708427165,0.4607581262686503,
 0.19806050551532084,0.3999912398384866,0.4179450189065743,0.1613744806187778,
 0.01307182794164441,0.755495864928427,0.8492501577355563,0.5169547106993746,
 2.432619135905867e-6,0.5752870028287289,0.10728029419945484,
 0.9861255946334911,0.37538551312370594,0.8333443624369806,
 0.40355906858140833,0.14975074732903793,0.5074000884537295,
 0.15874189136287242,0.2512394339047973,0.1319113510468155,0.9579858873127656,
 0.9629068740528727,0.7822576146465292,0.2922333555604977,0.4401518749422746,
 0.9256015313852751,0.8544320297900586,0.2857338125092336,0.5855199782637851,
 0.6940874029913566,0.5821852892958632,0.7026247394582751,0.10957735113440903,
 0.9746664986084717,0.4997536939974504,0.3969630243714012,0.18484471192459795,
 0.9866676280450872,0.8266603847906344,0.5286766540137355,0.6069389016469745,
 0.7759980967402464,0.6581463612403973,0.2827212977439506,0.33121004030586965,
 0.6732224754173622,0.8383806682423617,0.956659904816952,0.1784756957909902,
 0.8133505057678072,0.33220525956511504,0.5665076816255217,0.4962365276474763,
 0.6170602427594654,0.0279560057820466,0.2546474541019701,0.03227369953783443,
 0.6739872165544492,0.8800081660809926,0.6404820258618922,0.8970355906981645,
 0.33063851018055146,0.03712903144818336,0.8585014122493336,
 0.7916440304816205,0.12291776916931885,0.22380322499160965,
 0.8330599209603565,0.1989385648176074,0.8178036527555116,0.04435149263226956,
 0.05305412440743751,0.16734478018687826,0.005111354856768502,
 0.029656630442325804,0.058864570023438256,0.6302676202681347,
 0.6367556297145595,0.8511381895640002,0.46297375434704613,
 0.12418745706169276,0.7902092562734595,0.5634751526785434,0.7350666502177766,
 0.13749529060128174,0.14310950934083244,0.844365431019869,0.7520900680410089,
 0.22647620734808827,0.6146765635954167,0.41931861102455903,
 0.3927754366479692,0.7938571223985391,0.7180048495351895,
 0.040582138754648445,0.06669566860605436,0.39789582171410665,
 0.6557823184417615,0.9602818809570343,0.3519251511790652,0.9467951932815649,
 0.08776620737770413,0.14096571691971194,0.3983961599198268,0.694074576561289,
 0.519746538505417,0.7573839981203125,0.5557738977553712,0.9592307863851508,
 0.8429781382724734,0.8666936636543289,0.300601908848001,0.27410004337023575,
 0.8466282848995693,0.075449113165169,0.9890816347803869,0.1571482553460587,
 0.3708500620394193,0.14489837050842092,0.9161642037178691,
 0.05426264799089231,0.37436988318019404,0.166537663450265,0.3720702883945084,
 0.3938988505298573,0.3540188158330455,0.12090051606323693,
 0.43199821427048857,0.07416870474556869,0.8579974959948515,
 0.7509145667933851,0.8654148797801631,0.8756757204121184,0.7143316050662141,
 0.14064981116471553,0.7566164998120949,0.171645160494576,0.6550446930138482,
 0.12596728151860526,9.501931554893872e-4,0.5460299379701541,
 0.43692992411002285,0.09094795540651357,0.7057844516211548,
 0.11250405930814777,0.0026890233695220367,0.5125790747058683,
 0.3500680212877806,0.9706013681854826,0.003046741447259582,
 0.45588439687447035,0.6380098511481117,0.9440209978693375,0.6861389555998898,
 0.9911419667727793,0.5964713414700652,0.15953742627648893,
 0.16278441590362025,0.616594607433826,0.5384391466569326,0.06241732397437838,
 0.6145613819906777,0.31014980913114343,0.7181044784728688,0.6856226236543709,
 0.047536368246162475,0.6679042095423513,0.2812355528104288,
 0.5551469376538719,0.6298762036212853,0.3014843908360887,0.0625552115197916,
 0.7492527414738634,0.81192814828847,0.8625314217562651,3.486075755787521e-4,
 0.737954918931636,0.656497535414064,0.5591190943107399,0.29267571827157324,
 0.6817296800287224,0.7343946435070224,0.5929165809768532,0.12347097697479237,
 0.380798399746724,0.4129329743015858,0.4166590138028501,0.4947418991384739,
 0.9339521784075403,0.865135578938457,0.7004303782228238,0.1773730326855848,
 0.2420074561454133,0.023073910586418278,0.783791946304832,0.8001951286052658,
 0.01241774471913959,0.7773808624005433,0.5467735086041168,0.6328028162127514,
 0.7566204850057947,0.8473960867607346,0.6816139960587592,0.5952370515415066,
 0.11143963158482861,0.32118185663846077,0.4105502727986351,
 0.6081944564876407,0.8455617642056571,0.26452007758554874,0.8530544529090104,
 0.21723888534549807,0.21692931891423461,0.3442695105606882,
 0.6156954908940778,0.8719925236842849,0.6349124163120422,0.8214773411297404,
 0.38768698097231236,0.8240315327394493,0.22415211295124826,
 0.6622717888440124,0.6688289467834452,0.4718362786468838,0.3877349753816788,
 0.43409617799844336,0.5372425946312699,0.11626940931363472,
 0.3302121368039601,0.943743995121846,0.983192537443683,0.5251166606522459,
 0.17893238541536072,0.6291333778224103,0.3835684039797217,
 0.15455790366742816,0.9927925518062894,0.08369095205971988,
 0.9448390913948193,0.7710928187785455,0.1265858496417659,0.1522278599022583,
 0.24818573660465448,0.2398853235166768,0.04969848482162509,
 0.4344934529878438,0.0041620576844102475,0.6892039852845211,
 0.626454449217822,0.8252044995241233,0.9187301699377219,0.257668230435359,
 0.1275755629591021,0.277579745464136,0.48623193244890406,0.6230142202228203,
 0.2916023912242657,0.27518466276458686,0.6931605048473306,0.9403786806118969,
 0.8488071796509375,0.7294148956149139,0.096405024590693,0.7646107657817306,
 0.6344872985850792,0.8064251670999004,0.9537101139055634,0.1979352693535369,
 0.7208289272794982,0.8430060841524435,0.6688996340086678,0.9008171413951086,
 0.6402614526887951,0.9703528072151043,0.9321672272152697,0.15721332606274263,
 0.12053966902103752,0.4862481738283254,0.3505614532593252,
 0.05809730188856399,0.10759809970140144,0.9009919628158511,
 0.08168272925855258,0.46183940554827163,0.821261597933283,0.3314133394044776,
 0.22863640687676234,0.627428480847243,0.4871364715500096,0.7918469315857011,
 0.34342122748982984,0.4105085309658012,0.44724764094973246,0.383194978696795,
 0.3387795064601592,0.1393568488912318,0.20121655349943446,0.7299044527284618,
 0.4002157363841383,0.8466595554863534,0.11229880021846861,0.3640425882386421,
 0.5219835802569411,0.45185650053447446,0.5765848405350896,0.5975629978059653,
 0.36989750394688237,0.5838961937584208,0.7876592327824224,0.2894655696710261,
 0.550492228725366,0.5112083411121958,0.9388280033287505,0.3678583701270225,
 0.9369422418098539,0.14777493886176485,0.4228549020168706,
 0.38800131551078476,0.8453684550125511,0.8078127233591389,0.8789369510118813,
 0.7483460947708686,0.2916585807060238,0.8352701013521635,0.9395022174480446,
 0.13923724373762547,0.8728792861763947,0.029720344674688093,
 0.3959958253239061,0.2910770233829557,0.8661515304579529,0.6875643952412462,
 0.4808326032536263,0.09939544166281622,0.8453629156434366,
 0.35867462119184657,0.2914612703746883,0.26658706229630813,
 0.5134354918081392,0.6934845244085549,0.002380718776516799,
 0.5844415732013324,0.19628055918937315,0.43677195368235844,
 0.36778916461292965,0.12125933018892321,0.10160209222708794,
 0.7877633453266991,0.5043782313901342,0.14645649904646552,0.5522170800268875,
 0.38202437365684894,0.22502578509696658,0.49308307725991796,
 0.8472890602109494,0.9046112444407985,0.8722818757099688,0.5033604404256476,
 0.35565919139511026,0.7035682613200183,0.7675234928185376,0.6257375905138532,
 0.8741020761281254,0.8859082300528831,0.9920455876987851,0.7662826421776973,
 0.4110526492507991,0.2237720961370071,0.21030376286885866,
 0.24126768544759247,0.28084959582498226,0.480112976008241,
 0.36322797855692535,0.3135971781306496,0.9964894453608604,0.5714181790933799,
 0.6666406057322702,0.42242125562250565,0.067903345942534,0.42150814015489857,
 0.25812027428447293,0.6431311196554389,0.8146609069192605,0.140779840264723,
 0.464743591384174,0.09380353419391463,0.4162815691520153,0.9958672589060865,
 0.5103779404515048,0.8205145919774265,0.8848163419671735,0.9195316741833762,
 0.8549253597620701,0.3994955488223413,0.5678682714518914,0.3372688922250132,
 0.4808658748453867,0.9534984234783885,0.2348753635032853,0.03884553452702166,
 0.1801510997089687,0.06896832874060377,0.3994632781540378,0.9147815761757003,
 0.26651201189897944,0.33853710060545794,0.18704272747720685,
 0.6139704000448128,0.8407964694979555,0.21668235333102126,
 0.45123312876308885,0.7691339657293492,0.4908340943500068,
 0.17905455153710992,0.6755539806909707,0.3603202776833965,0.548446338991441,
 0.11184768967959846,0.3367662818447875,0.47344781377323153,
 0.4825843601884595,0.875036230032604,0.8743300773091042,0.679525341860573,
 0.5581649111177607,0.8006859485492345,0.5996243645733815,0.5434199268055908,
 0.20419018514338805,0.045415100361151706,0.4629841643541991,
 0.9062202381179055,0.13900310623447254,0.4433044536231776,0.8235534945076387,
 0.5797137532454785,0.660604495301314,0.7626955274253003,0.0546271501606983,
 0.9814005912861824,0.5131572243479181,0.2881460584485658,0.8792752040295583,
 0.36154792166679084,0.6162428594804931,0.46952429498411463,
 0.8790753033330533,0.22008322841204941,0.063600610826965,0.4615806936164839,
 0.15654689767507235,0.20360453151092772,0.8664399667465085,
 0.7092958300875961,0.20716701221938494,0.6169524879140947,0.7811606425530804,
 0.9610275653151837,0.06858493217363093,0.7890201109440268,
 0.17010289640642695,0.6197553852271305,0.3874719070528534,0.3859605011219769,
 0.21280776185484918,0.07302157575230495,0.8182482519298387,
 0.7676604525000296,0.25464104921838715,0.35901300199735786,
 0.6598305140917994,0.9379448760766835,0.18080247274678873,
 0.13700471313030693,0.0022122748728391484,0.32527458037169954,
 0.966344083869603,0.9923934267927226,0.7562149929464681,0.29568494876022133,
 0.12889737160935744,0.2401697151393427,0.7157497930572174,
 0.050961691194317815,0.9692675671042256,0.7980301519149356,
 0.39790632413695337,0.761399291753199,0.4936603056977449,0.5914979478972537,
 0.48715041241360746,0.3003908361829428,0.6953764529190425,0.9531513690881793,
 0.40018976086240565,0.8263679656699939,0.3513886567679436,
 0.47764504114866546,0.5705938388637147,0.530395005357931,0.1646892904832129,
 0.7058960012660593,0.9201246267784481,0.963493898460741,0.6011571563923787,
 0.7191511625729307,0.6830853930061807,0.3080383799141907,0.2998507638239528,
 0.42257088363893636,0.936898874978653,0.9561076373699404,0.17671292299004315,
 0.8217490381055461,0.027390274035634543,0.5482178747148581,
 0.11033733681143543,0.4661843328129498,0.542001313451784,0.5002611649791071,
 0.9541006901647422,0.5733132259417651,0.7144332717309845,0.08312906725276858,
 0.4834383645878295,0.30057672257147916,0.85571811541641,0.9275131876071374,
 0.2038823756819147,0.7767850396616948,0.3060751516336181,0.21576665902056058,
 0.33344412793991474,0.8131776319571211,0.9153347213091404,
 0.27755204093058417,0.8257151302697988,0.2933015442384759,0.6695529636167337,
 0.5003813520610886,0.10668204235213685,0.22112132518094296,
 0.9879311094261005,0.9973564886726398,0.9488512866677866,0.9103307017624263,
 0.9011362793290781,0.9287451468971002,0.6497095653121495,0.3080114893281748,
 0.1709196650983178,0.7620075582405685,0.37994166398607676,0.7129983615673776,
 0.9539183255517466,0.594198950126219,0.43244736617427937,0.11072957545797224,
 0.7668803074797617,0.9114357301820363,0.49927697269327276,
 0.08461478424255264,0.0837143048136132,0.6150244184313641,0.5653470378266576,
 0.3490104948911743,0.639233924831403,0.354457007506475,0.7659454444564825,
 0.4136317517163717,0.017698615056532696,0.13463931779502525,
 0.4135334259158957,0.8279063698665157,0.5419914685217716,0.3643205349318648,
 0.23907741597478138,0.8132455286790791,0.5264401402827992,
 0.28402040199328005,0.4802179519086529,0.9927221983221941,
 0.09560998146336364,0.30750288677424564,0.41885790556717695,
 0.1427085850719637,0.3942353399065798,0.3346922438255243,0.7482600055417361,
 0.8832399439707019,0.2678475192380877,0.5952299670433466,0.3190382711112475,
 0.22105811454205215,0.6150412674466715,0.57257893048055,0.46357715788681964,
 0.832789629494012,0.010329700734533653,0.8201518865439157,0.621890322393774,
 0.5379773935716217,0.45577043440016174,0.8131805546802678,0.7122863016173148,
 0.4127761927247974,0.7593170453767317,0.6592218203510807,0.2566986616599123,
 0.7732404501402863,0.6415890233170808,0.5168981687425621,0.1799717014322707,
 0.7326773858236963,0.39337508470075155,0.6586815727656946,0.6533406687302413,
 0.4224133666908071,0.5625399083920296,0.012921588657653471,
 0.5464570445639041,0.29724415565969786,0.5627789928798161,0.5944903347249726,
 0.05310417039300186,0.22028924086028046,0.2264313215255097,
 0.5384534532061733,0.27951486691380734,0.7105038635585552,
 0.03364392549774131,0.8503432923077139,0.8459251376115197,0.7640962971444905,
 0.9063941322888197,0.14897362350464727,0.318905560531224,0.20244881947412074,
 0.8432098673578937,0.5172362221559026,0.11981352235169596,0.5925974508889171,
 0.8506252957284183,0.5916706511389415,0.07551393771670545,
 0.40966589511611595,0.2308356407555764,0.7975764768440008,0.6612644882165272,
 0.8378910650165466,0.6933285109001276,0.15106002853989886,0.975578553769544,
 0.10630338620326374,0.6074030820453401,0.5936305347177102,
 0.05269961169989423,0.9821746913149204,0.28218801742291033,
 0.41111024014675623,0.46792652121207523,0.9624973012270749,
 0.3182184235987733,0.6306844795709077,0.7892995101473318,0.8531945138235776,
 0.4361059960841853,0.8171011820243893,0.8592658879510511,0.3626509138995456,
 0.7273236864954612,0.5641828218232492,0.7719955369800684,0.16774739699488184,
 0.9867125202018765,0.9430021665238407,0.30144600459604376,
 0.03645164623167818,0.6575056175461435,0.3455891175974093,0.7664828017274503,
 0.7097379295321793,0.14595773018789737,0.37859828577212873,
 0.6778972498982699,0.17184182324739716,0.5231503594853741,
 0.30290229847646233,0.9803497149253952,0.9291592528290287,0.7645177484017054,
 0.42006192653878116,0.019644348678153345,0.943035920233009,
 0.6223293395587155,0.47719624019747864,0.1010522600999414,0.4786749799015382,
 0.4985827550631675,0.24152230168773137,0.051004670980922984,
 0.702416335342947,0.23296604116277664,0.6723865979898146,0.2644149009449419,
 0.38723829063105764,0.7675039999106068,0.7892537849695306,
 0.08279416391210881,0.7121911478152361,0.048011888590501206,
 0.8196479037782853,0.1593794687378165,0.6946411323471231,
 0.029168753005064052,0.4070199880018893,0.9611731848024175,0.802414975128697,
 0.4776778128112895,0.7144909435873734,0.6273356306170343,0.34483327919663165,
 0.5902003419348515,0.10668360550782574,0.4566787507954535,0.4747688461788302,
 0.8481657822935119,0.02871535965575267,0.1251065462579719,0.9903829863177641,
 0.8407133129099857,0.027613590703814972,0.3355652035446187,
 0.6759279979944434,0.5565885785531506,0.8780840495232995,0.6005554170648059,
 0.7254423385205079,0.051503967166735665,0.6659180992855938,
 0.2038057186875868,0.6419172304071469,0.36231834180979205,0.4569437140753515,
 0.34766020554563637,0.8357452562046567,0.6486358057742955,
 0.28720152405943256,0.27499363437428204,0.488400142113786,0.5153198041853262,
 0.1539839269383526,0.17053681722822955,0.44918381859375855,
 0.46549038157417166,0.41605543270982026,0.2965596233549196,
 0.25071301523111134,0.9831292664502219,0.4009376165339327,
 0.08913554952853597,0.9080895596400658,0.9190690174607331,
 0.38098478007177683,0.19153068806556683,0.5012410737643832,
 0.5737684107490528,0.3996618158351879,0.8751383260996035,0.14693716088708464,
 0.5356112572963699,0.27089032891898757,0.6287603335834518,0.9715753847279434,
 0.5137659383256857,0.7715524094267944,0.4805113279241079,0.6473912098572245,
 0.6934884338665158,0.030303396887997436,0.03640097622378091,
 0.4403465228463823,0.45870308789851055,0.7621563304458189,0.8568287310675452,
 0.7241826935161813,0.8620179720330919,0.47505811424038336,0.6444501006607778,
 0.41173877045708696,0.6605803546329666,0.7674803994652802,0.6242818909946657,
 0.4802756561071997,0.6092510597058423,0.8331531888531318,0.31843636219289395,
 0.1939658086907946,0.39241865638223294,0.660358264331488,0.8275657198734759,
 0.19550231611407576,0.6402723575877742,0.5985614643202902,
 0.19014674378260388,0.6167046333202478,0.928680689692373,0.8058902644000304,
 0.7233477663933673,0.9092560580135312,0.5760225366112663,0.5574574402560223,
 0.9403055732356066,0.788927804659394,0.23024188113466737,0.3084124916039048,
 0.0022885775179959378,0.6837454022782776,0.2629538421129721,
 0.04289143829598818,0.47103842515155203,0.171526863960612,0.5084907498016071,
 0.04981960305533606,0.7367318948083253,0.9484753038413501,0.7378942984511423,
 0.9696969287158983,0.046679567626999874,0.39406988984928304,
 0.287764173673543,0.6666606233500165,0.003683260367886576,0.7100725181033934,
 0.2426534353550499,0.2585196797467252,0.40705099726021876,0.0712610898123387,
 0.05388961124351577,0.46043002978548264,0.47605037050090726,
 0.705127597196261,0.1514275211176237,0.4576475711228354,0.7536694356137784,
 0.7432960486327285,0.26472771373443027,0.18244800294245367,
 0.6218662516837755,0.19065436960480675,0.06212294607521629,
 0.45111696410402924,0.8722099186580573,0.022136090071650072,
 0.217455253007382,0.5744437746659614,0.4680530240828429,0.22176983603112976,
 0.48286834103587406,0.48055746329093485,0.007444840068676517,
 0.1782787131346396,0.6019466789099697,0.5773898331642983,0.3813157685691131,
 0.35039995407895774,0.7767438256438635,0.17948956495555757,0.90512275237482,
 0.55154963810188,0.8272567680971831,0.6290707205114457,0.8733034049869499,
 0.3520413538454824,0.6478315990420106,0.9824813297698428,0.7309267096680183,
 0.5102552410955452,0.7106288085326251,0.718381568594058,0.6599315976478001,
 0.93046326075531,0.622386631935699,0.3000610022622302,0.4817114539435676,
 0.6994013771767669,0.6818993949028425,0.2939640796113654,0.7926159832355102,
 0.8654501930834213,0.10124151284313632,0.8892999235080292,
 0.23306755129641954,0.6497129573491298,0.6494670657524212,0.5276738330249616,
 0.03117978220300155,0.04669887019732277,0.5057814406411428,
 0.16198360429710024,0.8552775942735007,0.9923880782542971,0.7578632102394157,
 0.8656277931611585,0.4767804004183517,0.6425560013071135,0.941887349405629,
 0.8594349904911279,0.477520375358348,0.27068686412113185,0.6126118615296542,
 0.24347606655807852,0.47961270720223603,0.5675960904888027,0.206500954535227,
 0.5455723304890892,0.504362160263029,0.993324690772777,0.02198822286029689,
 0.052231135868914036,0.4791185080977314,0.5442778456079314,
 0.7751962027332255,0.29574379329737577,0.6777073291787931,0.37523991532388,
 0.2277399615904867,0.6656753233519707,0.5139315120502489,0.6019038848855324,
 0.7125800524531921,0.4461645850006434,0.5384826141022244,0.4948640194222259,
 0.7950530527278299,0.03390970590878739,0.21637707855795207,
 0.7932548497985725,0.9822235150914139,0.9581369562211314,0.00887001759379591,
 0.0888953142219786,0.4703696615440588,0.9482880524983721,0.29784683932826994,
 0.9413702416434312,0.6082399655765326,0.5458260491456959,0.5071918213757852,
 0.3641743532870777,0.3225708405736052,0.8014144240853455,
 0.024981993925116797,0.9934303625456163,0.876384715251493,0.5853671442139995,
 0.5451604185462918,0.06544442911714365,0.5255746493745563,0.79404437395662,
 0.8783127941302952,0.36775061006802456,0.6685362261399045,0.6365780887424032,
 0.6291632139578731,0.7646701605841025,0.586882270310271,0.12508098729110273,
 0.5707604475632637,0.03210575578829333,0.4107948954895695,
 0.27930196900457416,0.9499187510077717,0.9590071829448805,
 0.060704635025879305,0.5748137642766374,0.8593810912398049,
 0.24031813319802897,0.26189353305925933,0.3573948993777698,
 0.5019598657457607,0.6218511086196644,0.8267667673441303,0.19997676691314425,
 0.6206979417343855,0.06086142380640447,0.8528109436776967,
 0.13860240054237394,0.6199706289248117,0.5558424821147204,0.9172247208910509,
 0.5063399552881577,0.26007407280593964,0.9956796044701883,0.7988449125765126,
 0.6642394048266445,0.2601111941463089,0.9964061631900829,0.8836943270607559,
 0.6097603739128876,0.6472377134985688,0.2892056816316676,0.7567576852187846,
 0.9126907798918009,0.04278705319809295,0.9532161604797219,0.5242702340503922,
 0.9757158372401227,0.5407914516112192,0.8963735711590954,0.45715301807620945,
 0.26710785755536814,0.7198862417759008,0.835684387900887,0.4252133449318526,
 0.5042980076338593,0.08320557549198582,0.5269680787591632,0.7352653417194324,
 0.07048848262456386,0.4328660119619029,0.3924311446853238,0.9378630343721283,
 0.47083011008853903,0.8769156455357566,0.39144719791344873,
 0.8932216319596539,0.4892272858599912,0.4742222221699902,0.10089917224263134,
 0.6905426974812389,0.28521224548759694,0.9715832505586041,
 0.04200188001881844,0.4893164502981078,0.8148046582865555,0.7951974169388736,
 0.9158491494379691,0.29709678089195535,0.7086718090775755,0.7697574276559336,
 0.8235960338496932,0.6453497487049425,0.1983818408733069,0.47999995006400964,
 0.5960866702573092,0.7375908321375472,0.3217940352280094,0.2027645134055729,
 0.7971524245907976,0.7508168617593837,0.9431866943137491,0.5033420703435099,
 0.8164209647279526,0.1520095101751876,0.46556675536335457,
 0.031800111285682275,0.37601528656810423,0.29955621283390343,
 0.4556360644140234,0.6023023613818279,0.6949669749439186,0.21683937782123242,
 0.8122004616145156,0.7892473435531191,0.3026681741806738,0.1870867926532339,
 0.37234238974124145,0.16412668190195823,0.14976934932515185,
 0.9343479579163761,0.48756216657752083,0.08620940686743173,
 0.5687457797653426,0.429309254573111,0.03841808789097734,0.9215716685415277,
 0.8077606384613919,0.4537152780773429,0.793894837283653,0.007243545761859704,
 0.9274796214681389,0.09513205486321985,0.3941170815467574,0.1186916468325191,
 0.8383942276732528,0.4429476702549371,0.6451363814054866,0.20987763817458416,
 0.44252516720052193,0.9253520749757822,0.904596043134768,0.21760859882820482,
 0.38885542590734323,0.948458797565222,0.21346390061772402,0.8769891483747838,
 0.187645302436667,0.850403921390386,0.3750946317062156,0.9941983753850714,
 0.7592924080194781,0.008564193437656353,0.66589786513475,0.03625104455669759,
 0.8222235825227729,0.3594239569480383,0.3825697421473703,0.22833576646324993,
 0.2547630593506828,0.8196130625930957,0.754534677103584,0.9621939986330135,
 0.5662421561964273,0.7011293768920273,0.09994343029615271,0.8381445542996178,
 0.5342660579328676,0.5897507386086396,0.6555024154864464,0.7328827389210186,
 0.10674869695067168,0.9150227276165428,0.8281124912604056,
 0.004076366896079819,0.9385917103748486,0.8492519891092475,
 0.46370540948410144,0.2939897070738374,0.5028350391744649,
 0.24620034020280712,0.3058909290213575,0.6369880967888144,0.6999273302332666,
 0.023475462166852035,0.4876809674768976,0.3460947882016697,
 0.06112452252130829,0.17937381475630387,0.8357358201373268,
 0.9267834823261125,0.7891072726663172,0.19090532954557948,0.86052830753218,
 0.9513616077239372,0.6422296356626418,0.7939211775372695,0.6823366397551215,
 0.6713345955439302,0.42645006889535586,0.022637752807577294,
 0.7368842256487736,0.06381752626809167,0.23095200000344318,
 0.7600669350562067,0.5351401253280247,0.7238992502698571,0.9961336899359066,
 0.9082524958980206,0.3182137949807383,0.9406266910186424,0.2804307912762568,
 0.6430470433145471,0.3456700146648911,0.3479844341102625,0.37847588038289426,
 0.2101608891350888,0.3417721589792969,0.6467025666311558,0.3525356993124207,
 0.7519476981002329,0.8228171986547979,0.7180477618710945,0.2210829447594509,
 0.780379269144933,0.7953160276597298,0.943434143758667,0.20383653612423658,
 0.7021506319554237,0.7679871810784014,0.6266970780173636,0.16165678824885876,
 0.5078251958779806,0.07716413557938939,0.9515816469207562,0.6442451994375611,
 0.7665599611252459,0.9719198949180164,0.09040215968390386,0.7111294729545807,
 0.3462084980714767,0.99687496561102,0.04561837772995592,0.25265397665002975,
 0.8996257042464677,0.9719824897855786,0.37909765131805084,0.1607825035249565,
 0.28703058770456635,0.39148317960630363,0.46688847815028656,
 0.4069352863436989,0.5578604271489611,0.39950358705458655,
 0.32315263128601057,0.724743179432203,0.15883582257477624,0.1879865125690885,
 0.05404088427488529,0.8316937478416349,0.09493701471966831,
 0.2880711845449154,0.19697758118258524,0.8653649182570582,0.2850105010763777,
 0.5537535732852334,0.1261453085067492,0.04758798544747311,0.5978076661125694,
 0.7078988177310549,0.9608831063291517,0.6440742587474783,0.16777234916713757,
 0.35529726880632095,0.9647947156784155,0.03463187694153935,
 0.9069118568331356,0.014774902699596915,0.5640988122421863,
 0.7557591985543273,0.5722640163190078,0.874595668944872,0.7437005446946962,
 0.6075196398718491,0.07274470351146423,0.6041324235104946,
 0.034426720609829764,0.7846657883545631,0.45270246446823403,
 0.7499862785550283,0.5508081492372725,0.24129200627898162,0.3073879129637386,
 0.5336502644317087,0.7513116041480393,0.8402312521212515,0.9183813212848208,
 0.9369560095948531,0.1933135174628292,0.7284614997058974,0.8390985473068716,
 0.037249825442976725,0.28222792875244385,0.6344656061082022,
 0.3447966239933684,0.4487100121571397,0.5763439330311204,0.3066127041306762,
 0.26771660911420037,0.15403485139220152,0.9849028195616588,
 0.26363947849505653,0.8386943708723391,0.8188391078641272,0.5581863666339824,
 0.05502522295842471,0.9299029809425293,0.5977929585535695,0.5502522292683724,
 0.4261565131472834,0.2461938193074823,0.7284741845505943,0.5325350786041555,
 0.14878694109554524,0.29511373742575997,0.49948326808008603,
 0.8844612590642696,0.23734044903884022,0.02619854722882531,
 0.5058600249874127,0.8766101820441862,0.45590499217454517,
 0.02041936011855705,0.9617050914035952,0.5487891131883583,
 0.45149209107457633,0.557965396861092,0.4245949757859473,0.12382511034512922,
 0.24566022327285275,0.6301818301912371,0.4860428631270054,0.4678707717011934,
 0.4463229425410058,0.4056490452422572,0.2309628998777139,0.965734820594417,
 0.31504185977969623,0.8255938440528476,0.5923621889881484,0.8909154847260198,
 0.5727815992352161,0.3654427843395681,0.31348654004347865,0.5031334173220345,
 0.7305027921846553,0.28253602016301627,0.11313958237881538,
 0.9801129571806418,0.4670678259587199,0.42587507957819903,0.6757298872027216,
 0.8708872087028324,0.13652139196066426,0.8998640928425544,0.3532769517018375,
 0.002662430530914306,0.5665545097882247,0.028844524372152014,
 0.2522235918165733,0.6239590706318224,0.7690603368938469,0.9859301178146594,
 0.1373193346782975,0.9821264056884182,0.36050055724930175,0.6267252145874133,
 0.05444761614072313,0.5204803630989553,0.14304377778200528,
 0.42922239487667135,0.8831353678890725,0.6780180739917686,0.2872511800970369,
 0.45265525260163186,0.4335170844710946,0.7896951107857387,0.6789530106305028,
 0.8492577897682559,0.30501314670032376,0.9800928655448193,0.1664421556891177,
 0.3412419678467512,0.8454179287117226,0.6804473154918044,0.48180040034594973,
 0.402749265008165,0.6243125089437801,0.9316504091278628,0.9173987825436278,
 0.7337983590768795,0.532335480456387,0.9778431148320613,0.6697916884464713,
 0.3617990570588816,0.6398267065431436,0.463475610472881,0.32289843219031533,
 0.011520379933831926,0.9955550944206919,0.8508818034834054,
 0.8746440484894734,0.5256020722794759,0.05133446515536355,0.7310962238035295,
 0.843133066765061,0.8390142623001626,0.24047723120509346,0.33287074490310564,
 0.37546641068424624,0.6575003857837126,0.620917084878929,0.49636342597814564,
 0.8767100397044135,0.9326298932809379,0.6321408782867818,0.11529681646804002,
 0.4330222850856493,0.8088943142607892,0.6551341319721441,0.6303350987349953,
 0.4566851150610447,0.7330507578380688,0.9434770096097359,0.7410750123766755,
 0.5532522408635211,0.18987954458213174,0.411766267825434,0.5294411984174582,
 0.678222166348974,0.9300245960775468,0.7496592212382559,0.17692922087406138,
 0.4004430366926851,0.11501195706094614,0.047526057502933705,
 0.6546785218464691,0.8119727341645043,0.4493931331344163,0.9886983898763713,
 0.8170389015482294,0.8227822776132946,0.14940606495756048,0.3920273252277341,
 0.42714764553670925,0.6889585891089046,0.06877632228096087,
 0.9210608793081081,0.3742890356429276,0.5390049102119239,0.9402932291579371,
 0.5852594131530351,0.38182241482727164,0.4345657556169866,0.8034569233837114,
 0.8158077854334261,0.40780909523549524,0.955262901497433,0.1529692520613669,
 0.7180005746296484,0.18893712450114208,0.27968592162987627,
 0.4891437890774578,0.8046182936264241,0.5763430661597828,0.3699038244156513,
 0.13301966554701705,0.7617899848118399,0.8008612905525998,0.5589841285620479,
 0.07040737227931115,0.9846916951858987,0.8530387069776618,
 0.41406240005703354,0.8948194799182799,0.4645615293907839,
 0.030291679232541724,0.15681053285769986,0.8321696581927136,
 0.6550248799008188,0.5764752538088375,0.014346484247514502,
 0.7409794814477461,0.7924211562936245,0.833802629322013,0.13802271608404093,
 0.29057494639740755,0.3694155094284517,0.47561526014726674,
 0.5653325383119787,0.5717899730724157,0.14385158626654904,0.7687666539678437,
 0.17994427415377046,0.7865236591564948,0.09196791678056448,
 0.031827647429876316,0.1730129903185006,0.5147170859102626,
 0.0718143115439629,0.3300139843298394,0.35908868981378095,0.8867811430251078,
 0.6069828796392556,0.8267522116407865,0.3033676615347697,0.1730038500976785,
 0.607538669331587,0.0603263147085642,0.6491687644507615,0.7951762444942317,
 0.8777729385509494,0.15982085264486945,0.15258739622862594,
 0.9128971012130761,0.07205512631470856,0.7536365727843626,0.6312377870634436,
 0.3224451752257671,0.8873182591141351,0.28421214108471293,0.7430299786915199,
 0.7419621069426209,0.004246113470695478,0.9157410699173781,
 0.7183179833531061,0.3178539522538222,0.17134985258768864,
 0.16031626518794828,0.7272856256060826,0.8024615364296225,0.9143770752004221,
 0.37240531831606294,0.5040683468837477,0.11295023547035188,0.225231339502317,
 0.6630646223839001,0.07845301040991781,0.37136345876503185,
 0.4198411404874719,0.6485723942927366,0.723270485193244,0.10589222001323861,
 0.5727219344182919,0.31652102602746734,0.4788980026248383,0.5301870843295096,
 0.598383777241813,0.06622013112486402,0.3519374376993738,0.5232862969510611,
 0.6544433065520481,0.8595274991794114,0.32034286439219084,
 0.28146654455994236,0.6329339201679473,0.2765190113902678,0.6176516365711744,
 0.7951705675269857,0.17077911515496658,0.2012282381260938,0.3630076551133121,
 0.019366386778273403,0.639456682616539,0.6174213030306059,
 0.07672534398244668,0.36274865552752455,0.8762894664610344,
 0.9253293303719659,0.553842676295246,0.3929587113704228,0.06870932291984222,
 0.9617166104872962,0.3462158659494875,0.7948390683475187,0.10575969142618868,
 0.49232610380625796,0.5416850915431501,0.6665151397487307,0.6090839786974003,
 0.9615005336273676,0.6722158098588293,0.4600125873763711,0.18557499734058813,
 0.3458212555205695,0.18076467543622932,0.002203869493665911,
 0.45399134254628737,0.5429475049328408,0.6946669741641449,0.9400059148370011,
 0.43780270574271174,0.349825067967824,0.5830260775613133,0.561263800614358,
 0.5798172962190501,0.4623847152169184,0.41260163159284546,
 0.10441388770733495,0.49723637526800024,0.973836597031617,0.2753548677309202,
 0.9989204442446423,0.4297246925998712,0.4314444864454371,0.33255823990218814,
 0.7137712397082867,0.6127620571240542,0.3000094680998504,0.43235998564369105,
 0.3195382555471863,0.004230994047897618,0.6889770213131455,
 0.8576573685833366,0.5428219889735392,0.9839000037195218,0.4555394107924393,
 0.7722110440331935,0.28874097063371995,0.21159420607727397,
 0.25559223401309794,0.5089225867329817,0.2808093617128793,
 0.059993434593562545,0.36514877574986304,0.688752440068806,
 0.5270391893581057,0.5593369512195278,0.14133986701680734,0.8968842202709008,
 0.6831122599497192,0.40432804449833915,0.2569651614467009,0.84396765148379,
 0.06099814158040273,0.15426925320090645,0.356771374587722,0.6378266383547351,
 0.7000866762407959,0.7903003313400869,0.6748373831052933,0.25536232769593736,
 0.4811121598055774,0.8357448412985375,0.1401126673457166,0.5229388736425359,
 0.34404742564962976,0.10790552517013519,0.3811392291054915,
 0.7316730813121829,0.2692845560512648,0.32436065415077786,0.3811833903167611,
 0.5239040388041135,0.32846359045758877,0.6438982848453396,
 0.36295173543230685,0.035448908158859926,0.47310630490651406,
 0.05552753375624753,0.01843912084625654,0.6618046423913763,
 0.3066382090507972,0.03807108269405812,0.31229210166522303,
 0.9283517705322846,0.3883177831320378,0.7009142933989908,0.6549755340123671,
 0.3220732505828695,0.6075097216908781,0.01556360191699635,0.3354982139130014,
 0.5532070892309227,0.6875785584328888,0.062450446668968596,
 0.34331656390887566,0.4711202228139102,0.6476532279668071,0.4507619277333119,
 0.20259575588812262,0.24441224398507333,0.5803304982211117,
 0.7158232333843372,0.47201720765700395,0.6090544632480773,
 0.41885409433816534,0.5247710179660875,0.034612461634341174,
 0.9844293956974587,0.6630147718613351,0.40792395332997333,
 0.025306110894306344,0.18447792000103946,0.763119530167055,
 0.21671094505510013,0.7379452088683558,0.6183927938731711,0.3957000909452849,
 0.8440423174550846,0.9870588686368179,0.2410027151017149,0.2563621535335804,
 0.8910196695062957,0.10681511840506197,0.12337045844172212,
 0.23403383657636911,0.8806210334292226,0.7650420092609476,0.4491879472073096,
 0.8788534720276402,0.6877922144700456,0.37541175192658016,0.5472441748076765,
 0.5679123976967702,0.581137621309103,0.38999742142129157,0.45465480202225617,
 0.43843119042552914,0.4174715908034393,0.43421383331891894,
 0.2569412873068855,0.37995499584344805,0.9070903099034022,
 0.10403733001166804,0.7315441034672614,0.14091933032914172,
 0.4286660508516853,0.39929100780392035,0.009231293717010036,
 0.4569048306163149,0.2854890768003764,0.2073521539545815,0.22897491755379673,
 0.4186872457091021,0.09943780844450456,0.10812655933880977,
 0.8036105972161994,0.6686328722650875,0.44763258619213775,0.3167005892963126,
 0.5008628023657895,0.3551407040180664,0.43440907044891264,0.723143307115311,
 0.14822870996492565,0.5890243044058445,0.6337468306276344,0.2571727669884043,
 0.31824760918476747,0.5894984677369992,0.8179673411066433,0.9125179058156835,
 0.49865651551365076,0.9373439620001868,0.11355171427383925,
 0.08832653030646453,0.9166132311811485,0.7796685767014081,0.3375725938859059,
 0.5682861538090496,0.873386061296914,0.8667645491711478,0.23304693009143995,
 0.7490506332723013,0.1393565720350588,0.737336765390061,0.1209732149629631,
 0.06876876711084623,0.48905883359659785,0.03935544995716589,
 0.021863048731306067,0.9591314709688805,0.4117233337344095,
 0.0790366462910741,0.6297259026955201,0.9464892844361572,0.7747955212776061,
 0.754207250866218,0.49223378468201584,0.8027624758252098,0.24605516672862027,
 0.1578403026555557,0.7370973815615667,0.2591090215040118,0.9716990939581506,
 0.5784126674260242,0.6820693521572507,0.6630315433013926,0.4327769597639829,
 0.8015469272570108,0.02488932149393097,0.8032670698677755,
 0.16751364698651572,0.002633798288899758,0.549911766540257,
 0.4036173668167451,0.6840272872661408,0.3850100409984337,0.6645188034764242,
 0.7126917255651417,0.3538140224095032,0.15176520696754292,0.9786390393797048,
 0.9158197949904894,0.4121775651570986,0.7452078967517807,0.36040816670191067,
 0.09519172615272253,0.37397613442212274,0.3458663627767611,
 0.6710232440777668,0.9320964018394539,0.05690630577130684,0.9639051059503193,
 0.8563475044586313,0.7404355112933693,0.4647213811047365,0.6760511030106275,
 0.2387669530663944,0.00861922974438356,0.7766769551785693,0.545581660086796,
 0.3229790639239155,0.21116638301253043,0.8862170438638894,
 0.18634397784527457,0.8481920692470549,0.6345782379662247,0.9989684181948761,
 0.939510080044934,0.13235252890991078,0.21599207183513913,0.8794108744502989,
 0.550889732008131,0.2552210074917891,0.42185184994686453,0.7647746781646511,
 0.8524617034185226,0.4812549399271,0.36786829047859504,0.8886430747408772,
 0.759806554410857,0.5964838203320862,0.8671330441536093,0.8202714863840448,
 0.11657093999421508,0.39442190128011234,0.32753229946587437,
 0.7443515796891879,0.5525662777645162,0.7145841934852225,0.4635100653858938,
 0.5888688400018838,0.32860244912725356,0.13737846886111837,
 0.2918338313427018,0.27438849169118895,0.4505749615860808,0.8458061032084214,
 0.19787939198866278,0.04438341117625366,0.8414844770816085,
 0.4032825656485083,0.926403178887184,0.32939495118057593,0.8621161783855954,
 0.6834583255018493,0.8991525326810823,0.2673383886089058,0.7614177283636944,
 0.9690825452726586,0.4352469345185509,0.5937528009568251,0.5314384914237819,
 0.13249570739351635,0.4859348761141856,0.2974074082995235,0.2885722923090722,
 0.08660026560721856,0.44650584345592403,0.4516050453313857,
 0.7068991817366757,0.013582049677471564,0.09048112217701976,
 0.932085359101508,0.5487180520271647,0.3226497571463155,0.018798622103398555,
 0.041009853222345516,0.3905865063822399,0.9737471072989127,
 0.3476053471568513,0.973477975097076,0.4046055702067033,0.6953128838107232,
 0.47879393773368495,0.0032071993391609155,0.1402257949417034,
 0.5762036583166129,0.7536258222438438,0.556915950547027,0.38288011135344635,
 0.6430067429459465,0.4083361909685097,0.672608345256605,0.7521068754022612,
 0.5801003319310947,0.5196339708880024,0.39214624631135864,0.9395977905614972,
 0.6087868605241363,0.81062306518962,0.8741851311379922,0.7686372005286238,
 0.7353443314504626,0.6666353346139857,0.12421146479073042,
 0.026559124674467416,0.14552863854175602,0.2604886762076203,
 0.12563109711961695,0.7552413696650596,0.7889645597871181,
 0.06453254055321889,0.6199904332298452,0.20042141955988257,
 0.7376843202544187,0.7986648911721465,0.8952118395573374,0.9683479822706498,
 0.45154067569663203,0.008098339974476176,0.8053877270864639,
 0.9798293988085003,0.10708343077156068,0.11821672437331732,
 0.9939345140917825,0.6395592379194852,0.8772117490784965,0.07039583063856236,
 0.28005749922552137,0.9866025508476154,0.271041274045833,0.6876821506849438,
 0.03130646495995326,0.1738846190012977,0.01265718657163728,
 0.5368359443879972,0.7465571302133337,0.41847573434879226,0.7577345565321427,
 0.9335163006489524,0.9074844939329575,0.36693758982665436,
 0.10724079198139036,0.050112800892767484,0.5667854369037447,
 0.9782943492592162,0.8851562073845359,0.7526011594456025,0.9149142739676469,
 0.4191071715435192,0.4517298280942442,0.016497161796697357,
 0.8461776426709295,0.8251420853980497,0.2487272989542666,0.4857917734263416,
 0.2117340027523439,0.853675416674835,0.6322616777240916,0.5120034583100275,
 0.41272937325086045,0.9258156178238148,0.8493537362449134,
 0.15684616960505238,0.49365120515159755,0.8732607262866549,0.14626657204353,
 0.7899973637956085,0.4157159465416652,0.956601318973016,0.7418145439959394,
 0.8724194999457392,0.40789198350798517,0.526506102255775,0.7796751835793442,
 0.6547738029441273,0.47537192673720097,0.48436050991744684,
 0.8011571328700765,0.9068139258375689,0.19023752415112816,0.8004843376520149,
 0.5054050949540873,0.6314449684543186,0.886011497496401,0.8756076092455103,
 0.5637716675413462,0.8817330378042433,0.8304357428217004,0.10656230157191049,
 0.38971725059462503,0.17784903070884683,0.8047844030272207,0.813557832758421,
 0.4229707326879082,0.5909565018148072,0.9258727638179894,0.14999742494171853,
 0.49725163612520995,0.10730732310163249,0.5829158663654841,
 0.34355915342404986,0.985528446065471,0.7890804189999507,0.14820032705010833,
 0.828482305311007,0.2910888823486081,0.9072104397328017,0.5399510744400008,
 0.7069142283791826,0.23954359062371577,0.28903507654089255,0.443942935491485,
 0.7883380840080033,0.35938883494552976,0.10387871181311725,
 0.43087215912348364,0.0757771755888973,0.9170996304560094,
 0.15156308213418523,0.23166675020381522,0.25850218205784015,
 0.5208641003494382,0.8997665053361406,0.2487759297804668,0.6245777043660353,
 0.8676424332173583,0.06693943482426701,0.5754708669980149,
 0.17936124022662991,0.5017887976777804,0.8958270238326397,0.947619942805473,
 0.8169120724371176,0.9923938486860098,0.26401196452778253,
 0.036930075077001745,0.009789278016776093,0.7928892512082234,
 0.2976778713820586,0.01528789075808068,0.13997722347624109,0.729262350506461,
 0.34225921086040645,0.0834825291634751,0.7005580548102927,0.5476980827387818,
 0.0931761934540023,0.4906420182084844,0.9870342001135838,0.8643321771191144,
 0.10319525757071712,0.8072953586325806,0.6275273333489098,0.2613757580720666,
 0.7431454503040391,0.01805107766008407,7.585066543671637e-4,
 0.5480071552475118,0.48381850052801734,0.3222398112817375,0.4871836101373711,
 0.5735449064783668,0.14359360674197286,0.09422019271133153,
 0.6821453688129094,0.5175156237445208,0.7470472803539779,0.8811714837658925,
 0.15480203710720009,0.2178388568251831,0.027880646682791665,
 0.9005607784146579,0.9301182826264228,0.6328343495217409,0.43992836717891526,
 0.2285239653449911,0.18111398468856155,0.4308927533182664,0.3408791039896264,
 0.13362581923142802,0.3339156605252308,0.7428698302101419,0.5175542174055122,
 0.5687868819044699,0.885978588316505,0.6822644336958508,0.9636433987621801,
 0.7057463702089487,0.0034992176836209365,0.39271816032410767,
 0.8294317948706406,0.24268250246731304,0.9500735446680645,0.2455235123170656,
 0.9034105728996208,0.4590353266008145,0.15747145494634363,
 0.10318335258398881,0.47615738126572893,0.29828918012313066,
 0.36603576371203084,0.41959871053819664,0.9321102117736041,
 0.17750871251025324,0.8140426591256708,0.4343868498812756,0.7418981288609364,
 0.2840658302999628,0.2658558066982446,0.5299538261794232,0.8150202882012991,
 0.126581415657247,0.7363025945920556,0.7121952402016788,0.9890734113907984,
 0.8715929929302156,0.761166434483166,0.6813230136830554,0.8899876835147583,
 0.13177736853261868,0.636303177462655,0.16257913621291775,
 0.41796771304032254,0.998050440008271,0.43728776310295303,0.9488150528609453,
 0.5803151485014202,0.07236919496426086,0.06267254640967124,
 0.5680946370457765,0.19512567552503923,0.08578396988306147,
 0.9856019872963157,0.3678988592950614,0.806901150785693,0.08768972309762613,
 0.2207048660412575,0.35205405260240064,0.8146464318920731,0.6435077548703214,
 0.6840596369893807,0.11440163293138172,0.4074356788377891,0.8048534143689893,
 0.9804616986258753,0.05752369194689799,0.053118377239525394,
 0.8480742552923816,0.23177667173839323,0.9207347554282475,0.6474140532302853,
 0.7952947137551916,0.4543622549790127,0.1500003543897631,0.13729222233095717,
 0.31984409864941,0.9450201319461711,0.41215701433126806,0.9543209226153563,
 0.5253550316675838,0.8962057171619496,0.7387165223946155,0.4117612097172453,
 0.8926081784833931,0.36770405078798696,0.2956810963740192,
 0.34463891523833545,0.07829767185953918,0.2113407367064456,
 0.3517480494877495,0.5026741922266303,0.5142019554336119,0.6268310313347718,
 0.3840465449513627,0.923771076275219,0.7196554922241536,0.5957432931120903,
 0.5070121579810469,0.8006185387094616,0.7444776049307417,0.12026196631235242,
 0.12871747289761715,0.7982620251154184,0.5396153442370084,0.808451240373261,
 0.4024656758244344,0.3472032386732754,0.9973010359890623,0.06629931115885634,
 0.6523344160413894,0.3019632329144316,0.1494951989510045,0.4551571807565271,
 0.27046902398257866,0.6391265299951229,0.6455598699446404,0.7860623634591348,
 0.737215625501409,0.34194942976119724,0.8491604228685218,0.11324960586809008,
 0.3360476401404088,0.4659813094103755,0.2576045283829739,0.8646406829993374,
 0.8903454252481304,0.9071512353302803,0.8895923929780688,0.5343880466550475,
 0.15033433785062722,0.7455099509451311,0.8627622169200233,0.5469543369197487,
 0.3257623087928839,0.8976117091238872,0.8197401575982259,0.9154479116490563,
 0.5676027667274062,0.9345275673237238,0.8534468026756182,
 0.015703572807013177,0.6009840923139569,0.7437040708611967,
 0.5722005331693661,0.8050880380552848,0.2349461488397,0.38410998599540913,
 0.40342444507694786,0.39691534915536064,0.18150256461807834,
 0.8569426653374141,0.8029916116412132,0.3359513889669623,0.7386627356187754,
 0.9581269281908583,0.6498454355792769,0.9483865802812489,0.1533898348955468,
 0.08220752151440192,0.18567972615066153,0.7051873633667223,
 0.9811022214817046,0.7673055981133614,0.5481390366109397,0.3985349718370116,
 0.9904667628820599,0.2706785410469069,0.5455002745086807,0.8504165744854449,
 0.8107859362616336,0.008851174439866272,0.8684943586540264,
 0.5341973534277378,0.994389860774057,0.11561782739859994,0.8825248160726349,
 0.11590527456803223,0.9366085567010851,0.8861590270738408,0.8643017646320493,
 0.3499582079156808,0.25096960583829375,0.49505614333044223,
 0.4526062717488435,0.4654741822331414,0.14167590049126033,
 0.36601818410563447,0.7905170142543352,0.7075177836735246,
 0.37655495693428365,0.6861558373471455,0.4907427109923612,
 0.008767700198397788,0.79628079496561,0.3494022790907212,0.22143523076697746,
 0.09676655306234139,0.9868742019765777,0.27133960009116975,
 0.2095660324955707,0.5010927681448073,0.4140431778906708,0.9173820930365921,
 0.5521353027114216,0.23517403653429447,0.811802168872475,0.9885788058732019,
 0.7403522557488403,0.026717161893783148,0.3080664741000272,
 0.5959802208890238,0.7806364925783296,0.22445765229324155,0.8466508529638127,
 0.26772844920268,0.9307938681787563,0.046888912017102324,0.2539052007472916,
 0.7366625518077607,0.06813105223736815,0.9823063745387327,
 0.20385549640516443,0.04141832977073012,0.6533221311126352,
 0.5086389625466214,0.31904107331325104,0.48752140526698096,
 0.041812483230043274,0.3677953360179147,0.3220004170915167,
 0.13413243947433906,0.11960997788542438,0.9498447802228481,
 0.34529034009785675,0.7377133798618084,0.3631082605980469,
 0.31119907109415423,0.7092931889605204,0.18086210683554027,
 0.7807659944560138,0.7766418313956175,0.8618518706616586,0.40068066371509126,
 0.33728979977546436,0.9055609457336358,0.7927593809870856,0.9037197250824661,
 0.8893032889905798,0.8756729303427969,0.15986626492370182,0.4107402256721686,
 0.9620378656470061,0.33008601598679777,0.45937415619225097,
 0.2751359640824551,0.8673562446192996,0.014660338330041545,
 0.021761944990362103,0.09722307356397819,0.20264383136644848,
 0.7476765934586205,0.4131402342709243,0.007703789534183492,
 0.10569308502322461,0.31431218823251805,0.3501273659709535,
 0.11652915737124192,0.11535311965212869,0.05345114416635666,
 0.16121837647791004,0.35553282686424437,0.03773752071447689,
 0.9674968967751252,0.47878570147613,0.11826276260679791,0.022710349215093295,
 0.7003520473686314,0.7024847533863273,0.21107709374817207,0.8596167228832872,
 0.5029365936427335,0.3818266265175153,0.9962728577340226,
 0.030218182544035743,0.08287682271784524,0.2679968001566554,
 0.4500098950742807,0.6905447144351332,0.17130265656297905,
 0.09367392651138673,0.1681872263727615,0.07215696289897011,
 0.4806904892801148,0.3751956064637466,0.8674616037935707,0.7756967499726052,
 0.24204415767536158,0.8255664340310224,0.9043270837633722,0.6298828212627872,
 0.2796399566956177,0.05702581682493313,0.6405211411396504,0.6445754528696539,
 0.777745660059327,0.11308843261740043,0.21437176553131432,
 0.10083992045237733,0.9889874146022593,0.4284064353334113,
 0.46383384810691775,0.5143126185124588,0.23700686091438938,0.775212698830336,
 0.4215139605632586,0.8265977422127412,0.31891170777584377,0.8322330505638691,
 0.8684423119638276,0.30588133648327653,0.04284762195563396,0.974659318716442,
 0.6376681758826858,0.7874555318600879,0.3581350151070326,0.16852190462478678,
 0.7037344710338509,0.598474046234985,0.06815680624402709,0.06815627771272847,
 0.011716899555748506,0.18496814389260585,0.19642039951146117,
 0.9320015755994959,0.05107736612105129,0.33957411217401157,
 0.9835528101893762,0.954219183540792,0.8184098051991504,0.0764745140332268,
 0.9779297046838248,0.9020277289826837,0.21218688284549458,
 0.13344787942148384,0.8362459715272369,0.9947917407883897,0.2718389658024716,
 0.04648875113878115,0.7451939749507385,0.9607592151809972,0.9264955887332335,
 0.6775830483672528,0.06847221233604683,0.35525898486240326,
 0.8405966323080915,0.7086524492008515,0.03282791017624609,0.4525791831739838,
 0.33059782015158745,0.5308901126146819,0.6632658058487922,
 0.17233531795618962,0.21817849966705338,0.5063429838101586,
 0.7488961918971455,0.35835677299934543,0.39339695320721324,
 0.5804714381444371,0.761897034487393,0.6179538478176065,0.38808132078656465,
 0.11491963336214595,0.6135023351652071,0.025911071195830027,
 0.5702367213748265,0.9075337493697534,0.31823103950799436,0.4902553318031936,
 0.2831769024398064,0.7284405779308295,0.9482551827758727,0.5108058221671664,
 0.9934206880650749,0.3491142031010651,0.8390947128096055,0.6125745810300861,
 0.37145639363453997,0.6584737208725863,0.5220968787377569,
 0.21597291477548752,0.17578107763406614,0.21875843533043526,
 0.42056536004563205,0.28402773182708296,0.8866083923106116,
 0.17217545635671916,0.04510473611347965,0.7811758738977016,
 0.9318616581072693,0.37993324523154537,0.2263985544301047,
 0.22852993666302168,0.875331762961752,0.151249952072677,0.20084127980395483,
 0.2397086510240729,0.9517677021153912,0.5463219048009494,0.6479703681327441,
 0.7805230429251335,0.3518852230929924,0.18589590269739587,0.1436207953288805,
 0.03721052069419284,0.16054332904720903,0.44349867225405515,
 0.2559378617257858,0.10126706687580422,0.6823145700910885,
 0.11012120131662173,0.5585146403713536,0.0659432048275157,
 0.28793871352193856,0.16247033823962242,0.5615474108768786,
 0.0983060866463612,0.7429901054472494,0.8300847611017256,0.5027283190640951,
 0.9731231132553277,0.7636021045720536,0.29135228170361427,0.5365706665804308,
 0.4926751011402146,0.8796114263938114,0.1995228728517663,0.32472681293607675,
 0.7556092941324449,0.01872207392721592,0.5219744865393979,
 0.32770595809031233,0.6878459462687596,0.5365315778523865,
 0.08447917936088489,0.3161863944864137,0.8097580233945684,
 0.06231508124015417,0.43190304135322244,0.6380429137184918,
 0.2262898660772551,0.7764225236829355,0.28343432941476776,0.6608672621085584,
 0.10190920051019114,0.843314889008514,0.13229226491371537,0.6268364068957982,
 0.22488488127048267,0.08333332601300847,0.19076517782158797,
 0.510183602928617,0.46472898762131265,0.03882600359400379,0.7646776585897189,
 0.9910497813874619,0.11796503126271951,0.8053775173587083,0.7395165321268117,
 0.3903917691750828,0.11537713422392426,0.14193435988093683,
 0.13283763687229855,0.4754877252320674,0.7573094640391981,0.9570682382991866,
 0.666300854255749,0.013323329370422576,0.6261737693934927,0.6287511628025235,
 0.8495071346395008,0.04959146321698105,0.9536882533965072,0.1658122561840185,
 0.006510128841651808,0.3574249580624873,0.7851966350930198,
 0.2567604469388185,0.6331632089692797,0.9562798643742797,0.4059520146347356,
 0.9720784301104752,0.1208182881115275,0.6305088132063155,0.7242616985575733,
 0.20785764486301384,0.5509347650496978,0.62714884233611,0.41700467897563454,
 0.5083763463588724,0.7710087796031205,0.3357486651431192,0.3672621667196082,
 0.20585728377938362,0.9650088433936739,0.03545598447303089,0.181935748837733,
 0.8195929158473232,0.3537108918130454,0.48850452247516085,
 0.42770882091599616,0.7369266253287728,0.9211304410514973,0.9758459274324441,
 0.4024195503777347,0.7910516029536897,0.9360159545180036,0.37660453337611444,
 0.06099917267077426,0.8195237912882589,0.46414983538274324,0.623985323747532,
 0.4629577648131752,0.011526748518447594,0.8423596425562939,
 0.4482414565025351,0.4266743787934061,0.11245987670773439,
 0.02966700498873487,0.01756171584027677,0.9195706676569387,
 0.2659234241384735,0.9356060923616624,0.613097379705284,0.6485801857522312,
 0.20963313177099352,0.5462669211366004,0.8390264563850889,0.4657319376222496,
 0.5543520889950241,0.6448273807995503,0.8368164352951397,0.4230019198626327,
 0.46446951579687235,0.6717546004219079,0.07859655725546433,
 0.23161506957394684,0.5912610107867916,0.6508954933576694,0.6500205811563482,
 0.35335535246815397,0.40066888471337914,0.3860581135554866,
 0.30422923236739585,0.33177364651383834,0.37059984308711535,
 0.24823559630306158,0.6467337950285404,0.9250765606069684,0.8584088401450088,
 0.21897094463332367,0.6203007520316144,0.8363812962219768,0.1567351583962483,
 0.04368960623615181,0.43872730329398935,0.16579913065031715,
 0.6787520974155495,0.01793550401979027,0.2531039422476101,0.4375765645227486,
 0.23839427701744198,0.9170987343560274,0.6185810021217045,0.5441211432821089,
 0.6571613319404765,0.423682163282904,0.20821683803685043,0.1816327153310694,
 0.183596602307864,0.14805076093880432,0.18790334972393008,0.7334298172662392,
 0.34856896358378175,0.3369783669711861,0.5998206979463852,0.2441777344087045,
 0.39703535406393775,0.4557320153805109,0.7408788760621211,0.7273485982184669,
 0.3075426294361039,0.27957977719535476,0.785624996868856,0.868258177066545,
 0.4299460689679143,0.7536273417326205,0.9562638249136896,0.9711912265417062,
 0.24458792356481862,0.49531679918897076,0.9364959375972586,
 0.5175830442661149,0.6562197669882748,0.1880247120992109,0.9641734850806574,
 0.5672960746287921,0.6386584174874038,0.4310331840993681,0.6594782449254339,
 0.7855581631040395,0.5061851990869326,0.6052229505182812,0.8733078819591658,
 0.42096672974368143,0.479595603695839,0.8777486740023273,
 0.004055778703068946,0.9644298208791602,0.2945514683112236,
 0.6231918816389542,0.4532920646904459,0.05103806583703374,
 0.19546725619348038,0.20762066625453002,0.27625497843632907,
 0.8856518693404429,0.6763572861095382,0.5987113410051542,
 0.019124461942966242,0.08455530826405311,0.04674493634456911,
 0.7077442700921206,0.5640296349039964,0.4237585025347206,0.2952664877680281,
 0.8719770007723774,0.6952661513578097,0.259340010350152,0.40957618720238487,
 0.12549263225279716,0.41726187133187953,0.41165178597604546,
 0.8200658163962391,0.09558349607487693,0.5323899870174871,0.9730139438657477,
 0.8720482314709884,0.6828264258506528,0.2963902367326563,0.43852929894089865,
 0.6235430122080551,0.12677821609041684,0.3821684304608445,0.5080572257039777,
 0.93617042359619,0.7694953661659979,0.35922236227307025,0.96782545532291,
 0.5952481988405838,0.327472265909582,0.8903929029234254,0.8820229727992634,
 0.3951335573785133,0.9019523084571041,0.46400888606729274,
 0.09874243638932678,0.3613318521216602,0.791772910645109,0.6297395202847779,
 0.906941924608042,0.43546728241929156,0.3962992940021063,0.9345442900139671,
 0.04036546334915847,0.3107702674699915,0.1489728209708775,
 0.002592775406880965,0.967323656472523,0.2238173455550434,
 0.21352047473530966,0.6178850481971288,0.7530354329012816,0.9905590463478906,
 0.13135879440937848,0.5510481257926872,0.9014653273025882,0.7390239938801103,
 0.575658605054157,0.4723401429128793,0.5906751316696593,0.8030638623893844,
 0.07382678496764083,0.38883235367949975,0.7004998070163772,
 0.05199555077682383,0.8664645479296945,0.663033831845665,0.6283211205168957,
 0.7639183432886545,0.04017264255908648,0.6913948735030576,0.3021959391567799,
 0.3003621541397178,0.8855523325059106,0.9126464619464295,0.9443526986569689,
 0.5118866646988548,0.8242149937518866,0.8730969089811825,0.500088574001001,
 0.7551053344570045,0.963982386204234,0.2848745187377577,0.9047782275937143,
 0.2013740600719255,0.0322973757260826,0.0975760157714789,0.1029346663167574,
 0.8468087009549718,0.5453158585319559,0.31057008060923197,0.7883223778290692,
 0.1627598109395203,0.5199519076075809,0.23536692519956592,0.3535618911170565,
 0.6483141733392784,0.5005507662523205,0.4426535719510619,0.2064639783402632,
 0.42619713183398056,0.5889542944799255,0.6757365950285733,0.2532371958626717,
 0.7436821624013422,0.03647578495668091,0.9872589355679582,0.9269566610987199,
 0.08154330941243015,0.8341578315772517,0.3629689980405031,0.2843433155977757,
 0.9496359307883142,0.6533787615702603,0.8127356050092363,0.12164379587596597,
 0.47501208991012267,0.48687408619403216,0.9605351636860958,
 0.08630749911426172,0.7310093810999788,0.36690501157859945,
 0.5160498087536264,0.8027072867481652,0.18815488637564748,
 0.009522584092140685,0.2156135374639251,0.22073955554622127,
 0.6358936151036998,0.8840582623421704,0.58846798472066,0.5167158821705221,
 0.8586604903525732,0.48286713474144094,0.9229533619483801,
 0.26300195216038236,0.5850187167994869,0.9276877993889836,
 0.38132010489337886,0.5552838609368348,0.2276919164619884,0.3429069595833477,
 0.21251684252624758,0.7998431289281857,0.5615577829554637,0.5974089112361055,
 0.9642026279417761,0.9063355832338847,0.4162978758275384,0.15921000463662682,
 0.8471620049611408,0.225393185193835,0.4665656944737123,0.7096870551888396,
 0.7619668055740815,0.7176755361346449,0.7874577788103917,0.946676314220017,
 0.8805925904348044,0.5576560923776279,0.3619802124403103,0.23292228339253518,
 0.539457455397625,0.16642683728822072,0.8581513031674167,0.5069744535553761,
 0.19799017164948973,0.902250684009468,0.5887312981767516,0.969497132932015,
 0.6573619149349652,0.004275797670607817,0.49338644499854967,
 0.9337441294393004,0.6684789929709927,0.28637759683346675,0.8171358917500917,
 0.690318035330669,0.16607324280833236,0.7551858680902034,0.759881334335238,
 0.9030355569460332,0.08876537700093567,0.3083091939120045,0.5785175673784066,
 0.8746473228657932,0.31267102586448914,0.25090011960826164,
 0.8227073074276989,0.7288035836979986,0.2040658026405059,0.211794268771474,
 0.25302153896225466,0.5502517518837012,0.5852110647122418,0.880860299700002,
 0.9634097823043003,0.37380053790266876,0.051624161925549306,
 0.2262391533685565,0.681904880218034,0.5545565552281597,0.3360230825609978,
 0.9494143743311442,0.895816583664,0.5699985077027756,0.2879573345649603,
 0.7330962790538275,0.7418228697446019,0.8088302387070843,0.20252470703391168,
 0.023355125223186723,0.1348645770465018,0.5173154871025406,0.946572381441253,
 0.8222858571159114,0.4561301415536887,0.42657072503629523,0.9984074636577095,
 0.4325895403743172,0.6032379019336007,0.880062181697847,0.43137201921270096,
 0.7453011453181972,0.3946290610588735,0.6037393985226894,0.45350417368625084,
 0.5987879851402131,0.229377475713737,0.862652260434091,0.7994163538840158,
 0.6709871701780116,0.6929032528023624,0.8439339574457154,0.5053797259627772,
 0.049163553819093986,0.5996525709193969,0.9968612627112609,
 0.47836591200061274,0.5678766486935177,0.944251784774407,0.30975102856864356,
 0.5581284810034663,0.28248386183846896,0.8361485886615713,
 0.33123053732905305,0.6479662096899,0.5694184535382095,0.7552192751318222,
 0.7228387488152845,0.369617582145665,0.7763652505669173,0.13498416981108308,
 0.43142271211818084,0.06177655341537991,0.25837983752210214,
 0.914740953415018,0.9709453527410699,0.38960979768540493,0.3738118900408759,
 0.8243638909428211,0.1935156382455343,0.89687238103581,0.0319672230685103,
 0.027373103634407148,0.6275140762819932,0.7227961362650726,
 0.3758652599324992,0.3972724381727252,0.8516828958514724,
 0.011611315281342738,0.7268848822078607,0.9268505885118462,
 0.8206734529272258,0.2416135324963311,0.9031698002736074,0.6933912296366688,
 0.2899736223357934,0.7014868228717202,0.023847808624058886,
 0.4573478919342361,0.5194503136326789,0.8484733224911012,0.6567207904843184,
 0.18721549445053198,0.03536558172743032,0.14656790648587747,
 0.11239995008016823,0.4909796455369505,0.10328551546759601,
 0.9743390863554464,0.36427177802076427,0.6911089346409599,0.4412156484166243,
 0.7727763492557411,0.3593825739130928,0.8394788207267426,0.4184876803918067,
 0.8410083030273628,0.7380496502560896,0.5312426901837329,0.20017892179797336,
 0.7726136539934975,0.3933059888232152,0.5536160484087252,0.8048490573800493,
 0.7323160021004536,0.28999440830654755,0.021819068170817135,
 0.6266721702712577,0.065998688908384,0.109969524508292,0.11004597449713982,
 0.4613968872382439,0.46999356519771185,0.24273161769687446,
 0.5128518744203403,0.09039971378360934,0.48377210895015954,
 0.5682494566221689,0.45666685546868946,0.07071544380123229,
 0.9923472185756502,0.7474518193638117,0.5204700113716004,0.15631248811064669,
 0.7026690167894222,0.2299861241966049,0.5374044002514987,
 0.028755391425421717,0.7442571119395807,0.9969417672117444,
 0.6253260952317093,0.5068308757340761,0.6450365039300069,0.6659434272824892,
 0.44140906758053167,0.11668781179047383,0.5937500008721149,
 0.28716505645558166,0.9056582389111806,0.24165502605455824,
 0.18044628144426533,0.5110787112485438,0.027074363637048182,
 0.4345439646309345,0.2955318726883134,0.9383643712454136,0.4050564456472756,
 0.9434018324818123,0.9048662439822369,0.5078006909737979,0.37044260953798036,
 0.5525954258185789,0.604824317859051,0.9890194659965678,0.9805094358034037,
 0.43989934458149427,0.259033301587252,0.778109088168089,0.5537532728941974,
 0.5566527260917512,0.8082807231003694,0.3043540297158984,0.8692094979473626,
 0.9909662580077888,0.17217664336737648,0.13632443173881925,
 0.1563603040066296,0.35172848702831905,0.18267538007178785,
 0.48413283396682005,0.6005592687204072,0.10335345822236086,
 0.03365164201217019,0.5661301411651668,0.1522696385894705,0.5950283454415999,
 0.7333176765475036,0.042624964792929654,0.8663466820216779,
 0.8658612246831747,0.296790628499376,0.16811210903823348,0.7375260063141664,
 0.44689385821138994,0.5748528279468976,0.9192966860898727,0.6466436587535589,
 0.8600314975132937,0.015821733547889516,0.3287124403272037,0.925017933125028,
 0.6594351322214763,0.5029758244720375,0.0836928564991033,0.42549261673120076,
 0.15820492777085615,0.48686597076193183,0.076413385636799,0.6549824002019325,
 0.3895382000447124,0.039823121067251055,0.5803791105248947,
 0.5858185853518951,0.1044594734157468,0.879369084503854,0.4278697995797158,
 0.6722001021859312,0.5944990456837298,0.930860300638342,0.3851693392129186,
 0.665919315416097,0.805702529171028,0.3383669997333745,0.5608795688376365,
 0.42706794464180575,0.43044432418956424,0.12852649612260025,
 0.07600309875472122,0.5141518148503501,0.9075577585930193,0.592584396783617,
 0.6691123263821995,0.7484959959124041,0.8686720107201233,0.9323907144158613,
 0.21034114197379827,0.31706450074264203,0.06179379022930476,
 0.919132107731959,0.7708222771746654,0.8474883228676261,0.20440063401393538,
 0.8654707622574744,0.9685643261992514,0.26195936689666344,
 0.40791082534294765,0.18095898264633192,0.7353993848926561,
 0.6461510262217878,0.4276434510458753,0.9225055059794207,0.08935489769572136,
 0.8513526115000283,0.7416240944226491,0.8040241277244362,0.6655174211515011,
 0.24087055711381544,0.6433000283411467,0.3553195878214592,
 0.46540499854160244,0.25485442680030557,0.3689141001462707,
 0.7600142012077324,0.7304302882587995,0.8093183036793198,0.8905132230398758,
 0.33421062047162875,0.25066928729076343,0.9529343059546865,
 0.8531664050802432,0.754438498675666,0.055790941565513896,
 0.09526933161739337,0.33664305081436385,0.048007858791075386,
 0.25237475491094097,0.2891210241375465,0.5292449005556459,0.7252587004034328,
 0.9868515198500173,0.05818385269094151,0.7449174536298335,0.6333216597109014,
 0.3951019733801613,0.7142159123963571,0.36122460805310974,0.7442290532526119,
 0.20372291078260862,0.7819316850498811,0.30919076394301037,
 0.7743952943437692,0.28503016983867047,0.7641185618586129,0.6813659770986837,
 0.3557906699346398,0.17459039243111008,0.7309863531116205,
 0.24751157523004175,0.38258187487781503,0.4426258656647932,
 0.07440706544015452,0.1610599842769347,0.04246561289626849,
 0.8392526708784579,0.005116939449011482,0.4367971813712309,
 0.9583970742528821,0.2371769991703273,0.1813296900018625,0.22607900497249278,
 0.4675829229241717,0.06859713800507294,0.7586611191956905,0.689878611232573,
 0.6574866648678911,0.03136509548120414,0.4246868103385001,0.260496131064776,
 0.24259089132567202,0.26937715078275604,0.8687577313196024,
 0.8600995060849245,0.49956597005764225,0.744126936925882,0.15847380941099187,
 0.22242759539400392,0.11332771203494696,0.7492530652854278,
 0.43076609125260523,0.37914885011105537,0.3410449563254685,
 0.47993085477058395,0.035383238249234505,0.2895331115323393,
 0.8474120928533612,0.9291790327009903,0.8457464198218855,0.42807741973476066,
 0.6877575943021739,0.025874776860820514,0.1906072397541917,
 0.30917405509374185,0.8681425665084709,0.31097523898049606,
 0.6303126712864608,0.3172132321612232,0.9012088816475673,0.8979757975513668,
 0.1653579250546081,0.6821880614732824,0.6311447225798759,0.7419274334647871,
 0.9918621746285344,0.2639009219191637,0.4585400026287223,0.4232199740633531,
 0.5127964193293068,0.11229817813264675,0.0427908419414601,
 0.46994132628591667,0.33472385210185773,0.6665828378227369,
 0.9386699998458957,0.2830809593335488,0.5805620431267862,0.7251848811471382,
 0.29668311655570223,0.2447756526616649,0.9611577875077931,
 0.08855987563939405,0.4317527043277902,0.5308724580698846,0.9532722229502034,
 0.13022837473674143,0.2737453578923137,0.4108874195918565,
 0.43920375578146176,0.13906449801963594,0.33234209938080184,
 0.460838838518115,0.32999078134063875,0.7550058784065198,0.3467858022872542,
 0.9561901956802966,0.76862722758302,0.9915872482781318,0.949424963465953,
 0.5363886647917098,0.2845479585948374,0.5154411237726879,0.4599639338482312,
 0.5358564733497839,0.9194913357095803,0.6582999667079432,0.9818255838573917,
 0.6247943410473726,0.3090994913216485,0.816899338925038,0.22675067908606228,
 0.9655657841993868,0.9299034887510041,0.4742219872739353,0.8971945486838192,
 0.7634945740491864,0.23689113554380414,0.254270753774702,0.5148585365817191,
 0.3398086743274459,0.6370045964962757,0.6420614921757314,0.20505559309339183,
 0.7908958947021111,0.5219144497093411,0.4818392237021011,0.42254827577367327,
 0.2615191407248515,0.8793865801441093,0.0224931053320111,0.17160024063356127,
 0.8671851504544,0.48114212592440364,0.9551429879991122,0.493857260730887,
 0.5986153211193255,0.9638029248026768,0.893146337217949,0.8143392976061626,
 0.2930945021702841,0.024221664499395,0.42967211917720816,0.6846425191433503,
 0.37513009290811994,0.7312862500051077,0.4767589120958421,
 0.001978758367185418,0.02508718678191557,0.6202391794458926,
 0.4167159302142309,0.02447818941631197,0.35276293725923713,
 0.39517952018029046,0.8790362825426232,0.3530068884451504,0.5873488362153548,
 0.7863765650508675,0.22434626785666734,0.4651041964359469,0.7651231666310712,
 0.25245265982374676,0.7366285501254681,0.4561219021208548,
 0.45650877293768843,0.13446851315261943,0.043097230376691886,
 0.9278219895261712,0.06883619063838509,0.5126560142108203,0.9898335644941461,
 0.21309017011868958,0.4990820247544959,0.8311631595135117,0.654046634826397,
 0.4589517109823765,0.6987027827174179,0.28676221390821566,
 0.15910692306508611,0.7781886814007616,0.9924546519244154,0.2744470538238707,
 0.4149174807646395,0.456298262955536,0.5896622723708588,0.07542523579846228,
 0.6917907953487071,0.4887008228137638,0.4141622899155383,0.7115693069615303,
 0.8469515270632105,0.4418154462034105,0.08985011901772433,0.669178611698145,
 0.23441158569066223,0.5442959945241705,0.5927311464367901,0.7358799834378874,
 0.24560981899841838,0.54931935899791,0.5461251808712018,0.995022795789979,
 0.016175503379529443,0.8138655579904079,0.9203361889514979,
 0.15884549735769038,0.3597913945299247,0.4740181533666369,0.5428252650607245,
 0.8374129233208778,0.5156768700479549,0.7114232517030702,0.6538190886874032,
 0.31402450946157456,0.19553323133091904,0.7013374433882683,
 0.5110003892921672,0.15772039120885362,0.6161752024324204,
 0.24417246963080275,0.3752172523157561,0.48893080986472814,
 0.4549533508109477,0.7802989176001295,0.1642435328883335,0.7075952669190566,
 0.607658525075203,0.9083188107860858,0.5240668364068504,0.39224813489755184,
 0.8159738566135408,0.864998384245072,0.15801456127042268,0.3258011237368591,
 0.09495996020062603,0.09834080309096682,0.5437366300652163,
 0.5006205496667612,0.6728337631080065,0.8737674060143807,0.864021218684235,
 0.17083778721496035,0.18088937615508838,0.6565529282151863,
 0.3336990171039881,0.017339649570613624,0.37389616327749187,
 0.8751862649130739,0.8553845604886842,0.5639288556538402,0.16241318522240045,
 0.6307984774041675,0.946742453419915,0.0037355213949493704,
 0.1580231189792497,0.788182523759988,0.16585756096286597,0.6279257794204861,
 0.11571699316560558,0.819200967772115,0.9880134058819863,0.10019227541134557,
 0.5963273801546191,0.8222174143390152,0.2769109879693896,0.7463876375690601,
 0.5249094028319731,0.6297187708223828,0.040682275962498404,
 0.8485492321505802,0.4135734359597453,0.6511530884841289,0.39506277035079906,
 0.3965425542936928,0.09908615824838929,0.3970215762382294,0.3192215401736579,
 0.5473459873444992,0.7338026358021268,0.5011243054546175,0.9666337676003605,
 0.3831095499110344,0.9220493274224191,0.19891549793967056,0.5615704490044376,
 0.7613543725086112,0.4221563626576358,0.5773349037876625,0.10854743337215089,
 0.942927381534022,0.5845788084134742,0.21447771690699768,0.3604481926662009,
 0.7594698949063469,0.43257891777254565,0.09183883533764681,
 0.08775257766061861,0.015419644858046921,0.15618855275553245,
 0.9466699340637573,0.7447180921235566,0.9743823753009898,0.4800643269004363,
 0.47846286172049735,0.5298498374197163,0.9022410265048126,
 0.21467656381938327,0.48185615372969637,0.2748415084119036,
 0.9544413321197199,0.9501723902604188,0.9157383347744112,0.5532059786996573,
 0.7222033465307931,0.0137671282920373,0.9503532737856955,0.7756272388966203,
 0.4252196948422591,0.8661970748207298,0.585739267989145,0.9401268190588844,
 0.12536507378498962,0.3988062251867448,0.9218801953706544,0.4472323632682669,
 0.8557829017261953,0.5280041148646573,0.7444411238655224,0.10865550815785485,
 0.5086707582717169,0.6952294616426562,0.2226870408668612,0.44026914110350346,
 0.8863884602461072,0.3843247879540279,0.7055967851148838,0.9499021185774792,
 0.025895865840622423,0.4355776826294303,0.6209908723459634,
 0.7503080610821341,0.6082528273465926,0.4035592319742308,0.15098808437996025,
 0.025259512264734663,0.001453685337129751,0.019006405200493015,
 0.0898476734492537,0.4570153413146334,0.041712394070515746,
 0.9291021870485143,0.15957873715728566,0.43655168612089057,
 0.6428040129182593,0.14412090742891304,0.49104474814544263,
 0.03991412985060028,0.16765689940315687,0.7247578558984695,
 0.8359225337477276,0.12721612563219498,0.18660522219094666,
 0.11920047252368693,0.6511289173571903,0.04491693270786379,
 0.7798478438161143,0.42559088128412703,0.33567210611686504,
 0.10272513857875154,0.5615253225943694,0.1301181638205018,0.9295399905676592,
 0.6738542607596028,0.5783110021496469,0.5836207818976986,0.2911686276476372,
 0.4959285267811774,0.2726518238940199,0.8837227789022469,0.5994897861484587,
 0.6246902294740401,0.8482067992762442,0.20369648092486103,0.749574044421897,
 0.5004350814649773,0.4425102816260612,0.47852993227296636,0.8761250705928851,
 0.9870763621953358,0.9436254811985093,0.5972906985350553,0.9052626553392048,
 0.854901693284112,0.03583199120189906,0.4793084628563974,
 0.016914368278982295,0.3192687452514189,0.5922239435240195,
 0.7479031411443955,0.913238908519029,0.6778011460242015,0.2842816331658591,
 0.2954437602988458,0.44737100531222695,0.6148171946361065,
 0.22605722563791708,0.0779760332937679,0.5478393330178957,0.9727375059926124,
 0.29885645064956035,0.7511912191753378,0.7536494674970087,
 0.020696025555346997,0.6005334188276146,0.3644396375769663,
 0.28124495373314273,0.8292256456908486,0.22684950553766003,
 0.8668594582206943,0.06988733666374825,0.7208363602052994,0.1993001326166014,
 0.07666284979418825,0.43802779512911627,0.0341381916580078,
 0.004866701798105666,0.5887241978403468,0.9300852439952617,0.854498429043331,
 0.8087066775109374,0.5903131414001307,0.6752024375288229,0.64170515113252,
 0.3534920787057798,0.7142310237862332,0.11110314853877656,0.6734036479235798,
 0.1978273613572713,0.5589714203368189,0.8720781343945435,0.7438301906441013,
 0.4805731605562429,0.45850567428450173,0.7177832852157223,0.9102140592330519,
 0.3626062127344203,0.3648927310493628,0.7065095488012588,0.341936206943388,
 0.20936557421780733,0.3704410799357617,0.5632130335999185,
 0.17374347135385015,0.9349461353090261,0.9050147291975748,
 0.24244805732254548,0.6431057448544227,0.5612961246199856,
 0.056273509821667345,0.1716733548709748,0.7220590531884697,
 0.4703940552772492,0.9106225100893383,0.029479408089822412,
 0.5461210584879426,0.4306073776209014,0.21928538167915623,
 0.33939523313486464,0.492313643582727,0.044538690867982345,
 0.44347255705915123,0.5488577951727116,0.37195988058501994,
 0.8296217797522519,0.17906974223810745,0.20340141274751367,
 0.4191488941997421,0.3600500080918916,0.9371013455385104,0.35803829462312164,
 0.7775081167230982,0.2120958476186957,0.9180943253960273,
 0.004932881764424524,0.21001159675568382,0.44866392827867374,
 0.44320437081157826,0.15773587828215896,0.007289781749902358,
 0.8298794961678324,1.9025339455902568e-4,0.06640190325148376,
 0.5933389850022034,0.6129334327001069,0.7900181890442977,0.4451163891627766,
 0.25399022816185457,0.007757505976397794,0.8853245724624375,
 0.7975258907416747,0.5236683573504486,0.18160356014548762,0.7529740928067128,
 0.6516030264705837,0.8236710522631856,0.7443205422823896,0.8748455550558174,
 0.5280209532715964,0.7672336084576811,0.752459569522768,0.8665158393463394,
 0.5287375197168336,0.21190960880572152,0.25171120083645726,
 0.8865732569878965,0.3387104643927443,0.2744877124667797,0.12039875091238439,
 0.6804270808736035,0.22664935851129941,0.12295541189455661,
 0.6943602766394719,0.8399381749842695,0.6569176055873943,0.5314718659088009,
 0.02763203234885747,0.39477511728917225,0.030543991629039424,
 0.6508831264326607,0.028675007991075585,0.8844601310113029,
 0.4828055128885599,0.7844753773245614,0.18015930952118075,
 0.08010324511194822,0.4196163942370599,0.9815464012290697,
 0.09236220949212426,0.7604556139428302,0.9406357353804131,0.4810002406407178,
 0.5122137077225503,0.7303327044514066,0.6212100400103385,0.4306615686623668,
 0.07191524320844245,0.1187675690202985,0.865249019853984,0.7803335256640309,
 0.1195499421473647,0.2653479066550237,0.6444005091545966,0.7590126633508761,
 0.19601921170364534,0.9361358307790856,0.6199481589608915,0.571020159326968,
 0.2795101640560451,0.9324693796798262,0.31595386030959416,
 0.09666150615849034,0.39445937707943746,0.5830951275690388,
 0.7452936601656681,0.4539125197116576,0.12675066186719464,0.3035047463074606,
 0.9969489104253894,0.8635433098981298,0.7168544978416224,0.6642141288866401,
 0.09892148071586337,0.08499129751543955,0.6316449474296859,
 0.5363218267622065,0.04561342307786609,0.9531141856083879,0.8216049945344805,
 0.8587518072778259,0.003903909703701247,0.45475587374442705,
 0.6743225234694589,0.15220434117061787,0.8650069682734213,0.8457313517571458,
 0.025211226439547985,0.7883269490326912,0.5915654915884916,
 0.2985341937003798,0.9095424757090187,0.973359404697658,0.5830307178572887,
 0.6993268321306233,0.39755972116204663,0.7024714331665528,0.7484637996471467,
 0.7058959301562686,0.3679861731146774,0.175040961283621,0.25218645762512637,
 0.570652498413943,0.12315770878311671,0.030486712278882555,
 0.48876471296256285,0.04619088102058411,0.9493133550943258,
 0.9964635492872083,0.009770148010388335,0.37895888092936225,
 0.014688787308736795,0.6745693743063016,0.7422851275888025,
 0.49221011521460245,0.656888761644375,0.7638023922512999,0.43940291601565074,
 0.4896769590823906,0.5493784986858075,0.5188120034883633,0.541482631869486,
 0.21968893518892174,0.7782853369399854,0.47509317370635895,
 0.11055273656418585,0.3032837853638314,0.6072602383962425,0.6746687221892161,
 0.3361478549500624,0.8324461606001132,0.8413838173602642,0.7335032740295414,
 0.14488693818738418,0.10007856709165419,0.8106357850520358,0.658543944387717,
 0.9620729232419847,0.42622200579983394,0.33624772095058053,
 0.7617134556641975,0.9562662570909632,0.9883949493047156,0.5176688484151358,
 0.8608009049356632,0.23709780165833716,0.6578193560777721,0.8905878266259447,
 0.6853455782528821,0.9285454837199706,0.7259274611720185,0.8871398390874105,
 0.3021230895415947,0.9486579645258657,0.6368665845013876,0.8073526791683181,
 0.23936029586016772,0.7688972313571755,0.5552038286767729,
 0.02900607168445779,0.10129164151667058,0.39175677285722343,
 0.8841534346703019,0.041779483582604415,0.9700308155981445,
 0.3479760917374537,0.7873028983411889,0.3796911292317303,0.2154057908311362,
 0.5094360557192117,0.10486177946363995,0.007664592751879362,
 0.20895899370968318,0.1681539330522629,0.5517637649230916,
 0.23809783778624372,0.29486217086812727,0.8796630060084799,
 0.21343930963021607,0.3286336320917437,0.8557850955934043,
 0.19507271373551494,0.901362847529644,0.7530664212016228,0.1612020688624458,
 0.9115275406811297,0.7789342041601723,0.8324923739492769,0.824249224273053,
 0.34978517362843264,0.8362459277047998,0.32664111288352926,
 0.11218118877964589,0.9639937592277301,0.4154028668361781,0.5187537282459319,
 0.25106606632342254,0.6881183338605347,0.20402143018962682,
 0.02765236308244312,0.1714676569792516,0.025001910013912942,
 0.8051218942360998,0.361670208089536,0.7392702949386616,0.6534142408561224,
 0.3111200096724144,0.5391065463530283,0.5252233047909645,0.1700306870245678,
 0.9605589037238338,0.049362675736643546,0.665286735413566,0.9020095835454149,
 0.0012109052913441554,0.34508518465136395,0.9612059442340272,
 0.020396001763614557,0.42303270875131516,0.7265003330481079,
 0.6326911839837353,0.6671459633361787,0.04004516092796939,0.5391421664650288,
 0.9690721708834582,0.24980107634211823,0.20300062771337468,
 0.5569800547345082,0.9684554269368089,0.5962346480491104,0.9458090341226888,
 0.11399427782051175,0.3064970485728109,0.505819823196425,0.31909601878931115,
 0.9963860171594495,0.028417675254264085,0.7972389354251324,
 0.9512220609348427,0.9498717170519568,0.4875807425829709,0.31038069971291105,
 0.8112122257103409,0.7710351024230273,0.6701263900195532,0.2910436269419654,
 0.1186132616386133,0.7604606625640107,0.6605063623379447,0.26106189912360345,
 0.6508191079262977,0.6274128669066628,0.8611648595286174,0.33616929792922046,
 0.36676707970883193,0.883156758560988,0.789110082295038,0.19011608075362152,
 0.9849518419537805,0.29164262546967024,0.40211358460301416,
 0.5476062504714546,0.6248173481071191,0.8203258053080631,0.5198423589260808,
 0.7031312555900515,0.8040860326994677,0.9670052578059326,0.4489509466277021,
 0.25319133488028167,0.36074693377411404,0.2658849945453913,
 0.2664576202939214,0.3688435358965788,0.6195758134688111,0.594446090991497,
 0.32090126431064847,0.8423674205954579,0.0032261045941885946,
 0.30271579145509164,0.841124937140769,0.050945605501028446,
 0.7424379471669936,0.5657783952651089,0.13472555142532283,0.6833891078788802,
 0.3425232884432139,0.047621143328821725,0.6609813151225288,
 0.17300570964669737,0.6476880953946126,0.7162494639314959,0.5771262164557409,
 0.08378493012617605,0.3854324830408292,0.28907238744528163,
 0.9801001979138375,0.2017734201056346,0.24092461337900473,
 0.07338530853630154,0.26752018887797524,0.6858588494081326,
 0.5472022840631859,0.5564697864046546,0.9868369059876214,0.6303034278804256,
 0.3157117260326683,0.5514424835605104,0.2814920034210624,0.660914053627557,
 0.8931589873909433,0.2553553016438348,0.3197230941745274,0.6179733378954217,
 0.8009014787748362,0.020633349371272924,0.3934524105621896,
 0.5128548484200345,0.73389724501251,0.7249883208464888,0.43003759684170817,
 0.23201781177939007,0.7165337753280573,0.3563081166671843,
 0.33485826477964986,0.8541726679123255,0.04457099285156385,
 0.1616138538772135,0.7377896232361605,0.40288787308742646,
 0.19095247917240366,0.34553187010094877,0.15931739344769624,
 0.410730057821304,0.8350964910818958,0.9367508516101183,0.9470138021794512,
 0.2572076342710792,0.5523544407885508,0.32218335157100697,0.7575179048097694,
 0.22164593695461654,0.34196711759388543,0.7551520744302651,
 0.4487019033835976,0.9244982829137963,0.03389475002521636,0.4996762102471862,
 0.17923710615101585,0.7590444949647754,0.5294006511597192,
 0.12926244701014733,0.4837792777762441,0.5358932820983551,
 0.25072291144475545,0.8638676894023083,0.9695043196895513,0.5690769071957179,
 0.12343255497088257,0.5473306972688795,0.7581987675155664,0.6953276359630173,
 0.676502606473148,0.5054064298993064,0.35909210455460105,0.8518724985601392,
 0.6324681379696794,0.6639832291158984,0.7812491769636076,0.10298552641066316,
 0.6162346947604862,0.9387485710590532,0.43001750790049065,
 0.01207249796146359,0.2389203676887699,0.8917694206462644,
 0.24673457808186872,0.23451080269638303,0.21095027240277442,
 0.5341540463861081,0.6340084027394438,0.42914281672273535,
 0.10036061074923297,0.32741057574167964,0.38881510471131575,
 0.21393609604369312,0.20190844897765114,0.4711457803766361,
 0.1826905558238281,0.15951830932364242,0.27957629677916407,
 0.14703369849820347,0.04701497532557486,0.8960570235903086,
 0.025562124364670535,0.0463783342074473,0.4840577188551183,
 0.29970149105186095,0.3402162426270694,0.1320414909764136,
 0.39906973577078364,0.09186148353632162,0.12673797850546986,
 0.2660561572594009,0.27716642610926134,0.11900148390204024,
 0.5149284322012824,0.9410880884755237,0.849987340691093,0.8552111770556317,
 0.44333888976732205,0.012251644007388318,0.33235939436481265,
 0.10042522225363681,0.9498797655444285,0.8045187611573569,
 0.02213433253092434,0.1961735851369314,0.516246614302269,0.2963538864341374,
 0.6018658631508601,0.8031873416976153,0.7798656732247315,0.5591728633268424,
 0.4514831502709824,0.6665313768252874,0.10387379820195952,
 0.45732917391687633,0.290031620889423,0.1559492750507252,0.2908158449748448,
 0.3674814194002908,0.4892802419125004,0.2091124672095912,0.6403469033221282,
 0.5195061490588407,0.9266821198080015,0.7521027718449773,0.4943389526017332,
 0.6542969640761386,0.6087889616499642,0.9314953618155299,0.9278739668199317,
 0.5566410177171023,0.40857695509766034,0.762755321057566,
 0.024835737322548734,0.8623040759418678,0.2757640603283691,0.328812357362985,
 0.12667795019449868,0.5543589579293099,0.427412633535143,0.3305725339905826,
 0.9923554025764499,0.5548832558971769,0.5703042392801162,0.733464607417842,
 0.8563346458392396,0.2200253889505427,0.6660038755544586,0.6451235447422788,
 0.8601851559222442,0.37300589356756575,0.47603357639511035,
 0.29700005432814514,0.4534892634161716,0.9581431080370443,0.865746907260288,
 0.9347317672655672,0.9487985647966118,0.5273501288104496,0.657128052738248,
 0.9211850377411526,0.3819729023812519,0.18949902563911092,0.1990662137849729,
 0.3582670991766921,0.15776970208149776,0.4511269037439336,
 0.022732799153187933,0.38208579479745264,0.17668353710964146,
 0.5294308753588469,0.73512244675029,0.7868843540408224,0.5686462914437515,
 0.4316457798550535,0.8956403132121245,0.9381641085137485,0.15307979024772167,
 0.20820548330010658,0.5587868240501874,0.9327941453763186,
 0.08407637358688758,0.5526434715193222,0.7665090124068447,0.9740342937959087,
 0.05378175516923722,0.7472045835206893,0.7210348439773998,0.3386057074650476,
 0.7000675517592966,0.8986290770343832,0.11598174046624932,
 0.18354682159619584,0.5953675108738753,0.9966529854128283,0.6697381479495955,
 0.5929857060791702,0.7711675000211562,0.35321990025673733,
 0.20682609033706412,0.7462893968692754,0.5805061807172892,0.3473574586561301,
 0.003328861666445526,0.4665546220888792,0.18095363709801682,
 0.7211779788905868,0.11340716740960588,0.14605132834603318,
 0.09067302775784092,0.9966738366952951,0.15662645664391917,
 0.20709552615054727,0.6044737077609346,0.9738568011615251,0.5776764646843612,
 0.5069297312044284,0.8440948940554599,0.16657378222697217,
 0.043048004903000536,0.13132540321367714,0.17150171608709486,
 0.7331787259345509,0.5741269939560677,0.5462213293079651,0.2020838313089377,
 0.7752258617186991,0.7042911058336507,0.8809035137498552,0.16339954121271072,
 0.4217722611195436,0.5610626891681151,0.07465572927768227,0.3162729733978539,
 0.2600781271202004,0.7985549405767574,0.27295335030156287,0.8810613645774639,
 0.28701821814971995,0.8919917122422092,0.9865227844992457,0.3318804753828717,
 0.6577312008952966,0.23451272778812082,0.5920492028504772,0.8762111327801932,
 0.20234152762700885,0.8474737911559671,0.8973131147199968,0.9437121236938778,
 0.09207384916286654,0.6663095886742286,0.020402274891053396,
 0.21949945384646619,0.7605686929460709,0.8880741649284587,0.4982145675203664,
 0.8470056966269607,0.9657747802095048,0.7721214717615931,0.13900051035039518,
 0.4465880262059385,0.6603071285303937,0.2987953050955451,0.977253684981555,
 0.5855336391473165,0.9867705643552676,0.0030525704480875993,
 0.955330919138782,0.3189161674064017,0.8507304345318031,0.6287094181452696,
 0.1029210721254823,0.3538766152244457,0.2128560144429853,0.4277175757639301,
 0.32331304572839437,0.6759115059483806,0.47071720284777196,
 0.2940201829183219,0.518026807967453,0.7970674797033226,0.38870811562395313,
 0.9791181659094474,0.34568333034096876,0.6761765585324198,0.3006811526550105,
 0.6514838555756455,0.6018607573649221,0.36825522907923514,0.6125467787288303,
 0.7961083579160269,0.29528768083248425,0.9065636328127232,0.741243418427081,
 0.5330203682398502,0.5236426628952953,0.32776815985092145,0.6571076607928487,
 0.16868456818421484,0.5862519876716675,0.6456730200662643,0.5877728381136437,
 2.4611301296728616e-4,0.21701782842533335,0.6498813258518289,
 0.009190264873851661,0.13761583468978278,0.15136140807319287,
 0.6390775719580463,0.677924929868629,0.8770728528412227,0.33079030586940317,
 0.8794588829879645,0.889870793240528,0.07655387181823892,0.05911289089202798,
 0.7361440753808124,0.07321930078427974,0.08329292547978462,
 0.9068412464433785,0.4067261855378268,0.9364683545364194,0.878674063455881,
 0.09517848200744095,0.9842260569461112,0.24847557454699887,
 0.7955909404800374,0.23297980876590074,0.10466719491778664,0.975001160712734,
 0.5329049021137142,0.5384307128686341,0.6810000986054356,0.6721376174375497,
 0.9487033426999318,0.2321956881428775,0.8710413335189002,0.5822856626604286,
 0.2706929923763385,0.5553335746242426,0.8842815494799998,0.3031465293135389,
 0.8734043695184598,0.12297067294279396,0.7323618560377109,
 0.028806335283110385,0.21356369472207226,0.4909619133960078,
 0.9533358007648919,0.20283732443636204,0.6201044886870053,0.6534575616063936,
 0.39825215296620053,0.5194720610736288,0.4286327229496979,0.5448278464060275,
 0.6391184064418957,0.7786620660837115,0.81574136698573,0.3845888289976115,
 0.16275265666452832,0.9821229962181607,0.22175004379358487,
 0.8281281838396672,0.8223363601159533,0.11981552549002783,0.1251794404951092,
 0.2123407757836201,0.8875190704904384,0.07992828284803022,0.7471484852844992,
 0.7190464079481466,0.9486135055380871,0.32212923387347236,0.981978893686699,
 0.8550159570515211,0.44016437270518427,0.27709659517633667,
 0.5869001602062509,0.08462923625576024,0.008402458945012903,
 0.9107246748204723,0.0919350945560139,0.03068866961805372,0.681158299577769,
 0.5246004271644992,0.9256645142010379,0.2608776440595062,0.8330830418164507,
 0.6537196348140454,0.6825525594439209,0.15601607750672442,0.7094300461433782,
 0.36037906588743196,0.9968209277628456,0.35914426404423805,
 0.09542887184397608,0.6944657107039964,0.6214121208775896,
 0.18690371200969258,0.6820890192993356,0.8513705558845817,0.6671377678280658,
 0.44549108493781375,0.8234852733894162,0.5703715608930372,0.6152035697962988,
 0.6123369171134605,0.9025925521583681,0.43855384355040183,
 0.27310576055799285,0.5765192929409005,0.8093655913782482,0.5652334210805147,
 0.12351316147316627,0.8283331005690737,0.7501407394115853,0.8389418260291188,
 0.29576241132555436,0.5192497084446989,0.8902036242416888,0.7793185964681373,
 0.864665478666445,0.9502834656736251,0.7651998570730278,0.41732646549433583,
 0.17625100706644892,0.1895093263637302,0.14545801134108838,
 0.6225341073790811,0.7708220011407598,0.9340317112458019,0.35718299777274765,
 0.7593901259664957,0.41317175652757454,0.9142194001234136,
 0.17165770386351475,0.5373846188405661,0.4819642125296878,
 0.19330761677358543,0.7075075265088924,0.07948667128877196,
 0.13459074089297174,0.481715979417101,0.8767658024869395,0.5228893753136681,
 0.147402724892026,0.8039416011689313,0.5745021745274865,0.955023390537702,
 0.04463672785773021,0.05026641535829668,0.14853260475133356,
 0.8317386003447466,0.33864150812357174,0.971699664812858,0.18827585399283453,
 0.6664929151247562,0.9813213098599023,0.6979510804252825,0.4008844825092428,
 0.19079481481176852,0.033770387496670606,0.9714844689582338,
 0.27515023464374033,0.024933264223271134,0.32183185968537165,
 0.10601035654876179,0.3414361305820628,0.08658954077354963,
 0.9845266209343069,0.7002987990989381,0.04665152323263222,0.7675508521881027,
 0.7945268965839406,0.33497049589705286,0.09887290261641402,
 0.9752712473956059,0.9462224844347055,0.8135008898211868,0.6025184756579284,
 0.06631442860875092,0.5855555141235321,0.6601065975979951,0.7293028813792298,
 0.8379382026902554,0.5571960953427463,0.4018140305924708,0.9471897916363895,
 0.4148172078597542,0.27298079457968927,0.15613613390101644,
 0.05112080964912791,0.7291553615355294,0.280516921186832,0.6964576901553943,
 0.7869753190204988,0.26308827769095877,0.7991494375564362,0.3175715139208324,
 0.04347950694709368,0.7712867395305165,0.7944213878651966,0.8787521798548222,
 0.387003093317313,0.6723164616340489,0.36753661624200107,0.9959284376876428,
 0.4822820193371752,0.6471478377758018,0.1990705114867276,0.46912205744338975,
 0.09755640147125344,0.8148737484970651,0.15926386459090847,
 0.4589604737229127,0.5763271568812567,0.13766988967250515,0.7365862292738385,
 0.3402958029662565,0.6396867205657046,0.7138995247277017,0.7381337114142478,
 0.529283451279187,0.7145856264410925,0.16597408026382443,0.917593158952625,
 0.42984328834238206,0.1997334820371014,0.874499014681174,0.7537143534903954,
 0.6210690961648184,0.938077983249669,0.496673578535757,0.5792641571537361,
 0.5685883167315822,0.605676211875274,0.2773028463086584,0.5450268481643996,
 0.7864784476069037,0.4726846215888516,0.03758319483591577,0.5317361328606325,
 0.6077895745271236,0.9096278022599369,0.018637765789408012,
 0.7975314033569226,0.6595944344345595,0.14525134811371898,0.5048637776785266,
 0.9159229854290611,0.009171445829259994,0.18156980727129213,
 0.4776664020018859,0.11281075134204788,0.266085924860068,0.5521512051177612,
 0.7555546800381211,0.6038341789949075,0.6353394503172637,0.36375283293753613,
 0.13982495796681726,0.9336074559752767,0.528455422407484,0.5772339268782696,
 0.8692013153860622,0.6550626341646553,0.6395479723025572,0.5528634186931077,
 0.3069576389592157,0.34054791284965336,0.011736448813676637,
 0.22130301694955656,0.7064666388375638,0.8853593094767113,
 0.21812094539799498,0.6955291103196319,0.08078993513696808,
 0.9035929154357659,0.9773998527666131,0.2670677832247643,0.9506684615807922,
 0.606239966834092,0.7340945847650144,0.9088449770555782,0.21040449070194467,
 0.46435627108541744,0.7912761253398709,0.8221384357827347,0.6306030170223116,
 0.09041019339931033,0.989930463309706,0.5941933029058798,0.5463756389453005,
 0.3716838667085334,0.9822437652093277,0.1276303850498346,0.3473143652527637,
 0.11553633395830121,0.4379808542110679,0.01046216768555519,
 0.7884987474707934,0.2875904171823117,0.8912771525381457,0.8129506343511675,
 0.3323672950683392,0.7769661442636064,0.7118363772662966,0.7475187903435224,
 0.6123793367153321,0.2578601074191522,0.3879982673495397,0.37942295530892556,
 0.1289602759816173,0.7085188796800862,0.8735859603539496,0.05151236261174241,
 0.31124610891906546,0.4635598691709062,0.5615429226370632,
 0.36066219772747254,0.3085071641188377,0.3922066306440477,
 0.18032818805700834,0.2836816061177805,0.4915301184006471,0.7599493922751352,
 0.41067434775574285,0.13359391615761462,0.09701778993776011,
 0.17585104617363623,0.5358304742863398,0.5985935164742113,0.4666247050816239,
 0.37429176463089375,0.11363456173348518,0.7788266380652362,
 0.9798906774185169,0.5008382299146419,0.4873463625266341,0.1004049815822955,
 0.9027671557075135,0.6405469461558826,0.7563712022764532,0.6583128828621805,
 0.3974340771498236,0.08684061351243777,0.7177347332819649,0.3294646641496042,
 0.7797568485860843,0.6095783790861186,0.1767206699677698,0.3538073820443488,
 0.30731686764955846,0.6703778788081043,0.05595023212057071,
 0.8811927355364624,0.9307496746043862,0.8645143802033691,0.6968831943481791,
 0.9077523471523993,0.4001279265539901,0.18183250700791698,
 0.45116209339785607,0.859487642861011,0.4897897165135363,0.577136824879868,
 0.9162966310781568,0.8970200239703554,0.10015337072997355,0.9383791455220919,
 0.6112047841183141,0.16891286179220832,0.3681666700409698,0.1543484977281444,
 0.2734625339534731,0.46012608204212047,0.31578849375035856,
 0.7875843907598528,0.828691500514964,0.18622799841357784,0.9948580457199739,
 0.3299149506252621,0.8265834317127236,0.8199064327963846,0.455417673921656,
 0.9958383726605526,0.8317286656489262,0.9388721034393548,0.27059330944551196,
 0.1369351500424285,0.8008749780946496,0.10247197448751977,0.541261324099499,
 0.5775497650238955,0.4484089848550159,0.6288882870107035,0.36652638455145026,
 0.7229089633114679,0.2770115871572203,0.8582496398330249,0.3039477545926864,
 0.11710154371791814,0.45455780881369434,0.19860151794418246,
 0.6875339971386062,0.45734158747551656,0.26612470181986814,0.650150823790505,
 0.289371065051765,0.585675947224543,0.03918486724516246,0.39719136004787214,
 0.023905754736690987,0.12584371191304933,0.43332923961889325,
 0.6843646712446017,0.3409115606304074,0.5165630481455837,0.3391997774676111,
 0.6660290576573932,0.4380455419584598,0.5408050076643597,0.12392658805118106,
 0.10513940475529104,0.3214213997368023,0.6433049104742263,0.8484225828434053,
 0.294863081637053,0.876766343846744,0.17816444470012327,0.9738684961929909,
 0.026201463742307474,0.06469227007628842,0.18714387229680862,
 0.7116149078100821,0.2468919592932839,0.8718525686582369,0.3612369848461525,
 0.7556724602722262,0.0061485407761859845,0.3870645423092126,
 0.11456669151426657,0.9105321826799372,0.047871051366280226,
 0.7897755780779477,0.13239475560650027,0.7652702617477167,0.5908333579324963,
 0.40457661581043713,0.82980890647778,0.3095059631118291,0.7182797848154328,
 0.1348002963839301,0.8793648465492991,0.6779147256128712,0.16992486009648677,
 0.023121856443331756,0.3568731059771014,0.13521180153650425,
 0.3975300961458965,0.9296270406978356,0.3221886920644814,0.3988722145056587,
 0.9122381728519569,0.49168693533854546,0.8538533829512729,0.6113584109983788,
 0.49466127406724036,0.07595589751438869,0.4538365835115393,
 0.10164423441711468,0.27021800507821525,0.26973719703767407,
 0.9340206186482635,0.5393351349610435,0.44347595662513994,0.977925946673283,
 0.4065432038110235,0.6655352641255786,0.621695345714417,0.4788003539346718,
 0.9042679131129179,0.38297347826132144,0.0755232602644218,0.7193499956316068,
 0.33944582668049617,0.1867099754720244,0.6812038904940166,
 0.07717941046028243,0.47828478299512234,0.1744117134432701,0.359448881998259,
 0.49736538949793085,0.5094340396831969,0.12318968206562841,
 0.10861733076868674,0.5511808740511036,0.49042710069281625,
 0.2864247324443083,0.5947521241480567,0.296650227661156,0.2777161869640121,
 0.39954859078389715,0.3201045540277566,0.24729552873201954,
 0.4070287835658387,0.3609869327689872,0.5505148377881014,0.3871141092017817,
 0.8140766801732455,0.8760642595657099,0.777549058139404,0.7290872210761606,
 0.39266198620015746,0.635048056740068,0.13355216432504557,
 0.24205920933208636,0.040895206207179347,0.26645319502631026,
 0.9622556059786009,0.5444966652755675,0.21547773279058924,0.8614838191486013,
 0.34432788055695707,0.5659652637557386,0.762244995313234,0.4035550760800146,
 0.6314426829411915,0.11997014416900331,0.13837130936713093,
 0.19221772093229506,0.5536770057810281,0.8773312739983583,0.7265823386005512,
 0.5015606576522004,0.2947300741550052,0.4798393853389694,0.5345118029558086,
 0.30747317564626786,0.13648270195930923,0.08560258321069991,
 0.835142600194823,0.16926317825535397,0.7786118611541748,0.23559854834985494,
 0.656933005932918,0.04772189200423016,0.8865798841178893,0.2012140213393243,
 0.228855028321425,0.9821023735675563,0.4023992573015899,0.09223805983018418,
 0.039278292044677476,0.8136514626085528,0.9943860546377852,
 0.2838619405672679,0.12506719322686632,0.8069539872008493,0.7926284339485813,
 0.930988542454402,0.8695590433989606,0.29219342340620025,0.5597889317629824,
 0.17261542959592935,0.57491078367658,0.23332438331620797,0.2210062945301562,
 0.2110670359705097,0.7486272205454846,0.343983320141663,0.7032233395819028,
 0.7581905103440557,0.659962908066225,0.5738699003438861,0.1657397487342912,
 0.45598373493773936,0.12494897742832356,0.8034273237582529,
 0.9130311892276213,0.07459895039954845,0.41015741466551336,
 0.3495580999865745,0.48336789906856914,0.09429280829722542,
 0.2831213114962603,0.3740520165012915,0.7086033482818364,
 0.0044486901130738055,0.7491891180175348,0.6441498402457465,
 0.4100970125558264,0.8395718224296554,0.5703640060288618,0.2865644669380022,
 0.09706274234594436,0.613187313480747,0.4142119610091033,0.7395298382985935,
 0.5193187774023914,0.8480822997825483,0.75853259781446,0.9709093685846782,
 0.408128421304777,0.39242138567806306,0.4880631324005956,0.459506101995393,
 0.7113813548776466,0.33961745527435017,0.23867542230638583,
 0.6417449615812231,0.12684058055760838,0.9282346296442839,
 0.18975919623263626,0.9363948940823735,0.7242021234354158,0.791253257879686,
 0.09365369404775159,0.28202543011790415,0.19657722384772258,
 0.6124068222710146,0.5124476779047493,0.9922800352281238,0.7086864921508588,
 0.5728352177396057,0.5701458232592589,0.4308880491461261,
 0.043487585004297835,0.5148761979436406,0.11860142780769356,
 0.7632586557451284,0.7556637182825285,0.8710494216079743,0.30694007216504315,
 0.6800485319792483,0.48826203838958193,0.4114523826324369,
 0.43713600580616396,0.010053209766771376,0.07876374748232706,
 0.07726673825961283,0.38845569434895777,0.8260668944257393,0.857520505115909,
 0.1264074103626447,0.7061864748731883,0.3288596072649298,0.6537574244252553,
 0.8796630635465529,0.4046417170132788,0.2625838485076787,0.36918457452991815,
 0.7714796716497947,0.3159092494281529,0.6803505279302707,0.1758578928657183,
 0.38442151621306175,0.6565997450570633,0.6980532574074967,0.6711341104981953,
 0.22996646685440425,0.9334492842646765,0.4823068785256962,0.2847141698854111,
 0.8798357413007181,0.4708158450993072,0.3761371334612921,0.07777224447977982,
 0.7920680626018158,0.6669906514631897,0.6821935203983216,0.20257846243085864,
 0.5800080113239079,0.3700481837388714,0.679642468092876,0.6737855863575684,
 0.906129070945455,0.15599429756902916,0.8890916844794726,0.09444015489254709,
 0.09000339669713464,0.32357969672338704,0.5607223872163156,
 0.16640471304055948,0.5964017303823107,0.2996505267923225,0.4108071307892014,
 0.3418627693101516,0.5207556048292735,0.09378414137716862,0.9605328255609957,
 0.20373772558769887,0.461342337080502,0.4903085792813596,0.5225372832011993,
 0.2927759962525961,0.6984494429838278,0.05686898117215433,0.0486959904386971,
 0.4296116994216277,0.5616072712553544,0.07403832651800824,
 0.43773569932548184,0.7973786162839218,0.7567665781454544,0.7681222680398497,
 0.8399641938910569,0.08472047551319484,0.20766371239488324,
 0.7273124845477857,0.047701559632361334,0.3566659747046277,
 0.6739344905052223,0.5231250675191366,0.6465339633495059,0.6487009295412336,
 0.15014819150887804,0.3119136966269873,0.32356529673248735,
 0.36408796388506515,0.11113734110231976,0.3890340844521759,
 0.39470913448987366,0.7419349990623758,0.033181763216940885,
 0.8944625993467081,0.09805824913798078,0.06536131820191493,
 0.06204243976581991,0.3838460786111155,0.756307741808861,0.7856205659600778,
 0.010204122011054029,0.490322546908322,0.6923854404568603,0.8966846496234543,
 0.9402872573196194,0.8075999600231507,0.5996233401249864,0.17998217551400997,
 0.5941163465689436,0.10705206817529933,0.5431500136028623,0.6421696557164728,
 0.7963658348673869,0.37634735337197855,0.5557503587783924,0.711787576999953,
 0.7486956187900539,0.5654245366534808,0.15323578260960213,0.6813296021346549,
 0.5758800507514835,0.9548045699209464,0.9029718224576176,0.9529239076621643,
 0.8225831278648147,0.9095500802148532,0.6189574164303113,0.07153673950784645,
 0.002739456874260804,0.8666220263565099,0.7920043094613691,
 0.17946971564929293,0.9859577122976179,0.5426650440824989,0.9467591291451579,
 0.7327210894065427,0.07962312657061044,0.3767334695824601,0.7378609567664342,
 0.22914822020318482,0.1646565706665496,0.454894307219491,0.17193912081572493,
 0.22278424182287349,0.06437487094728689,0.42289943572328326,
 0.7190707394809175,0.15944629332814475,0.03395557236984881,
 0.11067136473311912,0.581692445089366,0.4273659912582991,0.24465559375343326,
 0.7409067006700153,0.8073975191446806,0.642251163072103,0.08959911364694784,
 0.7157735957358379,0.9749715801805188,0.24897256572086768,0.682558456156613,
 0.16593213559137754,0.9877894895491419,0.39281198660616334,0.957232656575728,
 0.20223953638562342,0.8624233604359381,0.3978106016513515,0.9509524012386397,
 0.6323473110358202,0.33529939211839044,0.2009873017309981,0.2865996363164873,
 0.5101478135146337,0.9048633877045005,0.16637297323516098,0.8996395760443714,
 0.5746918557763254,0.03854631394504171,0.7523789703915849,0.6747944327497597,
 0.8015714261506326,0.7477546384556872,0.18474426755557782,0.7646194979053663,
 0.3545888536937183,0.44958475688100874,0.3054410442988885,0.900367736582323,
 0.4752973400216097,0.5151857899574419,0.4798720994073137,0.37885652546352677,
 0.4290342070950026,0.5031446690578961,0.9670664981520256,0.3720000743261518,
 0.40540702737177514,0.47925691130795434,0.40410953063353927,
 0.1053208776502963,0.011134911085386778,0.25700007473572295,
 0.15252514406292494,0.9954584371922056,0.7589665931070885,0.6855602858461842,
 0.4187421310217644,0.9998847263860858,0.86666395171963,0.24846227777661978,
 0.23580655821273155,0.7975356184756066,0.09653863745515312,
 0.5140446396925907,0.7840267860668405,0.40081179899705144,0.6320988416510878,
 0.1065488924971345,0.279038228271154,0.11407229277307318,0.46623384114059285,
 0.9287538841542793,0.301610481750044,0.6159428493240329,0.13143527725634108,
 0.25824815601505857,0.8796340984328761,0.8209980967313684,0.9761665723333374,
 0.45404364950597986,0.6935507018601914,0.9648235520323316,
 0.10989287146050908,0.16099032648421985,0.3421715637463464,
 0.3258722137694411,0.9175720333371237,0.8903731741184916,0.9381995185709028,
 0.6322501741140142,0.8822670331315371,0.27858812831374014,0.3318807722996413,
 0.9945739094529237,0.42290667507058866,0.6072112653824178,0.8047673606349481,
 0.37483942260857084,0.9335662334549282,0.34345335519512543,
 0.38271484753951657,0.7263414987147834,0.3463232619454656,
 0.44883288953062705,0.11194307770678247,0.9665650644231418,
 0.4901765292443503,0.9742036904957403,0.035415649412288586,
 0.06540937887021248,0.09703222038136583,0.44424451888568994,
 0.5482568317674145,0.3438127560352645,0.6489181673141493,0.9080816614955773,
 0.23554280852076948,0.6809325971757787,0.02715801392280992,
 0.5386641948615679,0.36142294504137007,0.8022860989629496,0.7672874086969139,
 0.4690417438124468,0.1912111780017599,0.7128414686453995,0.7928613045097,
 0.8360184420303018,0.7386079305314704,0.22741368660182337,0.5491926262191823,
 0.6462261449366199,0.6061875509424198,0.14136062521856108,
 0.16233783245388378,0.7953823040585518,0.466026314766823,0.10154030275999926,
 0.3970256758213154,0.24080671086976757,0.9467099321536854,0.3956078946914747,
 0.4985037988655503,0.35913994944965455,0.09082722039835223,
 0.9498516129046604,0.9619681369398879,0.6180077994021481,0.20377940367929814,
 0.6830663840628018,0.5793394476475644,0.7696752256903842,0.05904891906912457,
 0.765033594369374,0.5185873859417658,0.5505817659015338,0.6945932902784422,
 0.04870775980826936,0.5936222922493204,0.3330341489255587,0.8366701940417994,
 0.9046943956694059,0.5157801535515716,0.08464813392253101,
 0.12072441784826915,0.5987220605825789,0.951924520973981,0.5690013834338968,
 0.17698347602866138,0.5352006815065932,0.8424402514725734,0.3939885283609187,
 0.817335035232584,0.9819439699748141,0.6376635163980959,0.5813499554189062,
 0.7254868054066185,0.6005135823629535,0.07850398496592914,0.9449157396874989,
 0.041623134485936575,0.2508593396054726,0.9868087061482296,
 0.40243910945841865,0.6379999524561767,0.8937357441555961,0.6487615866764531,
 0.8755352353053738,0.7922118316008357,0.4879207640926798,0.9958983395522869,
 0.30858530821374996,0.7786545502810098,0.6363925939051821,0.6187330746306837,
 0.5435331725950522,0.37951868402605504,0.37659973574888617,
 0.31079356563382043,0.6320668215376524,0.03718082286666058,
 0.7142612579270996,0.23741044341160367,0.5721724472575788,
 0.15408244424923978,0.2373889548766055,0.8044602313593952,0.6153923083857571,
 0.014209091464141288,0.05769245599897377,0.39122799401042396,
 0.5779240875796718,0.8125557862765076,0.5247986588813951,0.20165103390979322,
 0.47134900424474346,0.5584716194633195,0.3974054068321309,0.9864638907179607,
 0.07223674157411708,0.9576709017365385,0.45763209298951835,
 0.41939750312500534,0.21234473430767498,0.4074933753856541,
 0.9194374298989741,0.013231274021408224,0.9255150605244701,
 0.48000677797518776,0.5990903408324423,0.6775949459834655,0.5122988117847269,
 0.9409272057094892,0.07977861521678997,0.6443280931122689,0.5703319916957552,
 0.1986949216303343,0.30915459671601586,0.503129941480477,0.8925017704223989,
 0.5297479450145386,0.7991337359974082,0.3767802403285726,
 0.009153987000118602,0.6966394573231032,0.07859495973241237,
 0.11713791315900346,0.9113741093585411,0.7795873106864448,0.2411488963197389,
 0.06388006802085178,0.0703174596953905,0.1842804807727645,
 0.005758362053379984,0.09672192209943231,0.5748258982104727,
 0.8551540072833732,0.28074532194626,0.535905923074423,0.3402281150144566,
 0.8193490267178655,0.2902040993779189,0.103614497109351,0.7666553376662285,
 0.1744199243375255,0.5395374298402881,0.35501726855618754,
 0.00856628113685376,0.6207868544829437,0.22049226008446665,
 0.36064669524005855,0.9583822124777397,0.3012138819660576,0.8018751405554905,
 0.7520866223966198,0.10462638966476234,0.5472388930780461,0.5649570850651743,
 0.6837117781214983,0.27933659614426465,0.5513851618721839,
 0.29380404921836945,0.018752613571470844,0.7840817435558735,
 0.1041008240728245,0.9307343296008004,0.3519903363736737,0.01051605764669361,
 0.08420846346691424,0.8773340889375966,0.49414713250625664,
 0.3686270091677003,0.34606225738128815,0.6305100131999943,0.7416872019800443,
 0.5246642318666382,0.058426987033657674,0.6901488031527152,
 0.13856172280397483,0.22968079556773802,0.9038476754154717,
 0.15035762701114774,0.39696542397961265,0.3983692740093926,
 0.1317748610379128,0.35438144980357666,0.3204263388368611,
 0.17137720694942316,0.4562510003040433,0.63625675293027,0.02472544775375063,
 0.21180895238072273,0.8940363832189178,0.16811724871222467,
 0.7903843954951055,0.5832678819592436,0.6414936500585968,0.6588698515538364,
 0.8370414179095441,0.9888171508914045,0.6010890361832245,0.09090847448181827,
 0.3998377684460527,0.5525888580680091,0.6874073574767372,0.5934178571627674,
 0.9416244072469051,0.768851823724728,0.531766384160572,0.6150092168828042,
 0.15297888460247544,0.9438612798273398,0.36300624897782396,
 0.011709066383926658,0.6947888429024102,0.3980487023744219,
 0.7970718290256936,0.5364783259777224,0.2563540052865543,0.4654686707572434,
 0.6803259198747137,0.14231962874673565,0.3935203824621345,
 0.17057071051050898,0.5213286112708263,0.8525920248480493,
 0.34702831814005497,0.03530940271194494,0.49940960844280236,
 0.04898444379033162,0.9994510674406196,0.04129809288267783,0.930525058058688,
 0.6612951492592631,0.36949103294303853,0.2580488771601479,0.6733859572470278,
 0.78743778587757,0.8648945843521886,0.42627008772188935,0.5736098558684322,
 0.4961634402102295,0.35238414573203214,0.05222560038415058,0.367416368921611,
 0.24522980327003296,0.5182378543831413,0.8060878692846978,
 0.02037495781059806,0.939932388418081,0.8085898089473676,0.738074943097314,
 0.6948685872497206,0.11533605914179235,0.7007869125188814,0.7887842370764916,
 0.07176923862956275,0.4605129128203648,0.9642374384665008,
 0.18737949351739736,0.15773914605306705,0.1657715582792706,
 0.5830169599397941,0.0992937149834553,0.9742874144070124,0.497178464635718,
 0.7205981773498763,0.40295473605400167,0.07487240849258503,
 0.9274104916903367,0.40642327274532886,0.6828104995997364,
 0.25794935731153346,0.6063302565455065,0.48805801408337457,
 0.9940845270020724,0.8502885091639232,0.43906105954960006,0.7664786708848355,
 0.5067462563959153,0.30472602593569953,0.28525951659290816,
 0.4545664968701628,0.06872783037728836,0.643406615415058,0.5109919857923081,
 0.3655578381412069,0.09378154427095986,0.8732875671487634,0.5060376561050891,
 0.31040675621644986,0.2679324320749683,0.28686548831896663,
 0.15866701544971085,0.008685099113772798,0.5987041796604323,
 0.14068823747459613,0.15520506160980674,0.3281242990916923,
 0.9234460493725458,0.7902131384209338,0.06516347203668815,0.2972222573594854,
 0.602456899171487,0.6705743168342452,0.25084984185069714,0.6724049343076336,
 0.8575703149168645,0.8982361053592165,0.020086473388412074,
 0.8901242561284292,0.9236867926351928,0.2012188653268734,0.3977751197579643,
 0.4973886688950454,0.4788296697877372,0.390511580937807,0.10933750599417469,
 0.8900245653105502,0.617050617234376,0.8610043937827107,0.23449311146196994,
 0.22349472505629286,0.7057141614218492,0.9426148528486569,
 0.17700254736628906,0.16923888078751081,0.5413078970853831,
 0.40047154498824944,0.4178600251202236,0.058730978986829,0.8774171917380515,
 0.9442353066726601,0.6579160017076645,0.8871569497183065,0.3595979027878352,
 0.9099975017808087,0.6752571154748062,0.6301739546694792,0.7514026203327634,
 0.4249376077915158,0.6803575806665686,0.7641722215405585,0.04129983627726108,
 0.7566785168694032,0.8656451703546011,0.990967532969418,0.004859621433218164,
 0.10273518613185639,0.04285534455933582,0.7872550422770832,
 0.7433823028330652,0.007906340763495168,0.04427233206151271,
 0.3013297963225151,0.19915286980955904,0.6053995478951676,0.7943095427381839,
 0.6694711907267611,0.8442370101656594,0.4606997549819426,0.7546470871117359,
 0.559244547311879,0.8866433916016547,0.5293061405728396,0.7931786864535593,
 0.7442083001937093,0.26357630064633053,0.347334427198432,0.45431103854979094,
 0.5072259009963783,0.7112202996668077,0.879763321417752,0.3431699124015567,
 0.13852862640845987,0.9588002985522387,0.8222028928718904,0.5095453455558914,
 0.8031555801944791,0.18288218055623873,0.9708499286807385,0.8793304452481165,
 0.40203622747918977,0.6896937664550479,0.6516441447603788,0.786570165589156,
 0.14865642116753996,0.35919814032831776,0.9052807413297498,
 0.39510697246628235,0.792988402209575,0.8453714454115818,0.9996857183646541,
 0.059152023931819286,0.5263502193107588,0.30987305307963264,
 0.3704343323944772,0.9595680796984994,0.7571931339856163,0.2720150417765319,
 0.18124004767552104,0.2693263647036509,0.6151089082204053,
 0.16829645780686042,0.926527146040606,0.061232702156331587,
 0.4523501959224546,0.6786649925642015,0.09530697133805399,
 0.16621682495564316,0.336296707638465,0.9068881637056686,0.848595764916119,
 0.012974748609183018,0.09993549712965599,0.14522336712632444,
 0.5251089499189213,0.7970925921346382,0.9469144699776367,0.6524849073469408,
 0.9116910557334948,0.1987673180120636,0.007179445777029447,
 0.6809650490034869,0.4572319297600511,0.6867566648795123,0.38281378716356784,
 0.935701630881502,0.8579355647052265,0.09289106715465567,0.2973923789823022,
 0.01418548671321107,0.407782499589164,0.1473200841747021,0.21945728525160624,
 0.2240767062101262,0.3393465436455855,0.513417755837756,0.9760951324433702,
 0.34436771767396945,0.9588805988980794,0.6017808914384718,0.9867198466847698,
 0.9443269642536907,0.6575947296804316,0.8686662170828359,0.50210935085822,
 0.17686990948413261,0.34698219700277644,0.7143992215252977,
 0.5346280503467773,0.5913446591381819,0.6632842666412723,0.5448207521087864,
 0.8216531502837794,0.6494489402909245,0.9654222456027775,0.626872795272414,
 0.47887404876270623,0.21669211020694534,0.8850006851809593,
 0.03659186562972894,0.007945955532817073,0.6117976328939382,
 0.9533395939711112,0.45931449086888543,0.2826303501844536,0.6262470001723593,
 0.8412571926605003,0.5076290124929869,0.7263055561352851,0.2007586276604767,
 0.028941198581295735,0.7375647592027837,0.5927250179142673,
 0.9797470828208297,0.9302278711426591,0.029127741072637492,
 0.9588909874017786,0.6018223013305374,0.727589235939118,0.698292344768489,
 0.9265005742647185,0.3741385687284209,0.23105964890548192,
 0.39372845185316985,0.6751485297602547,0.44632308947454846,
 0.7808392673240133,0.5514728635336459,0.947095336101635,0.8455517490037823,
 0.8666654867352497,0.9657917101253821,0.45784487221595205,0.9278932472434535,
 0.49521803603679415,0.18781493005493854,0.8547508582036935,
 0.48796685863986633,0.6222855026020788,0.8211597030178519,0.6473491087969797,
 0.5097490880793845,0.679260377880459,0.8148504066858848,0.970290673511324,
 0.1291238522552467,0.36828041305381665,0.022252721811568454,
 0.5647681797185384,0.7477642562136053,0.6883443691394814,0.7335364481693508,
 0.5780092883734077,0.6157408892377373,0.8973567327307675,0.06883373724383823,
 0.07405492392792357,0.23800278773984795,0.3621116710336699,
 0.06645790766087833,0.20535139314503748,0.37150420292230146,
 0.7105934261651121,0.40003692059519547,0.4167860136227184,0.9628568157190872,
 0.13075699493962345,0.8541409945445564,0.4214743867403765,0.9343115145971193,
 0.1640141109271691,0.6372260896051992,0.4821708475136868,0.4647992275340962,
 0.43711572949377264,0.8352291010886976,0.4519903701066823,0.6319819919242,
 0.40191917803662136,0.49938713998802997,0.8680844305557959,
 0.6222070138790227,0.3886232737650638,0.8074051881726413,0.8207753800933159,
 0.3192859223985268,0.029537864797849478,0.33496268795628914,
 0.28158383592024805,0.4640287162426677,0.9541964309220816,
 0.35223638977404914,0.9857411896654298,0.8709379158236322,0.5065354992188605,
 0.8160546044515503,0.707820836602538,0.18304300653463745,0.08214708510109359,
 0.5217420834136348,0.9115500752688594,0.02326605327884468,
 0.32185564396487465,0.26851844579217143,0.9700521201426269,
 0.14368220970347867,0.8317822585497584,0.6489039157384324,
 0.33777862518883517,0.8327950666979229,0.4246828528484423,0.6040660702042223,
 0.13533878696604962,0.627155163800432,0.6832038135559705,0.6620062146149289,
 0.4178158794419342,0.033009167857493193,0.8245686465373485,
 0.1316216178528633,0.6285018559759548,0.3895822797980375,0.31986761425247157,
 0.7180780266479263,0.5939824441406515,0.44476809510314774,0.8234209577324411,
 0.9753491472331755,0.6300922177711489,0.08089513806471216,0.9951686404152243,
 0.08952861818326063,0.27753812891328833,0.7834922336622661,
 0.8473423726964575,0.6408950814929918,0.752495040431412,0.7057208254953303,
 0.9088199429592808,0.7669848420476311,0.8098104291705028,0.6822560497483471,
 0.43329599421744336,0.9724211784065045,0.6664343677122392,0.1687917979474095,
 0.47640978474113793,0.8949023581925067,0.3547559269744196,0.6254292833194812,
 0.6330310939405898,0.9311942331068565,0.3187213384079641,0.9797455683174283,
 0.6287991267633147,0.5180037356748775,0.3430152257091291,0.7295882973314258,
 0.0027650228539131128,0.21056986823369128,0.6317343961009814,
 0.5454606866842553,0.7408111478419165,0.9180407051366283,0.13763288406656116,
 0.9432945503635133,0.8856233186651549,0.9121088945920522,0.49757786476721255,
 0.07631729255789643,0.947430852657007,0.3269164683427518,0.4741617121208589,
 0.8526940211197231,0.16786329575356607,0.7613684063816128,0.8117777508272827,
 0.4868113633393314,0.3714040083105741,0.5230137164546149,0.1944758977626082,
 0.6710640347146096,0.22010784466492683,0.7258392434036631,0.933138206390317,
 0.5914437991536886,0.5618547302891683,0.9904951346940707,0.04977060821330748,
 0.5951408531031468,0.6226839812169446,0.7347167934259529,0.6489154762394792,
 0.7443031611823432,0.5580539611132755,0.7931803760853863,0.7546958126320737,
 0.825493359416688,0.1024818523295169,0.4405156766637661,0.37322731168764434,
 0.08270597464085505,0.14328656962462638,0.20928879108771115,
 0.8480007029036509,0.5759230153323547,0.6569470327186451,0.9685300777335449,
 0.10358622552582153,0.7639350001680046,0.17352779417134334,0.910824727169868,
 0.6973377231490739,0.7633499600825773,0.17868364501440048,0.9307886466316612,
 0.3144451576251327,0.6748970393995313,0.0059207183268302765,
 0.48080604388872916,0.33785272271876343,0.6556416064968664,
 0.7657085058053593,0.20393762763318346,0.30258896247952083,
 0.16738334195213245,0.7884671167314614,0.7926499927090044,0.8972325423766958,
 0.8888800692436604,0.2541860312825527,0.06383697943870081,0.7528650019402323,
 0.8005526341256862,0.8317736486714642,0.007349729512744041,
 0.9930214550589651,0.23413977862742907,0.2904533135040729,0.5961435094879864,
 0.9520128987740868,0.8633367288916329,0.03227854657352791,
 0.21325860143990527,0.5993474392690206,0.2896677344883083,0.8713434860577617,
 0.7235070914233781,0.2359374777008003,0.866844579158316,0.49067721920911345,
 0.9606407619588826,0.7946261875460624,0.6557901886227144,0.43769417007246114,
 0.412748110823281,0.2505067051677129,0.9686321607180449,0.9710507998812887,
 0.60628383517836,0.3673092422968727,0.3403933906208386,0.003470589593309703,
 0.20410559915312443,0.7883579750087575,0.2573171616949208,
 0.036743112256046984,0.7812379315742686,0.1301332252888766,
 0.1470086527999973,0.03848354372455254,0.7122047862343528,
 0.039020255425966655,0.5195246840983949,0.6001987648202814,
 0.5087727313521492,0.8363565204032083,0.6135936560777759,0.12018403524950194,
 0.8085697231210868,0.2747949485229342,0.40079835966602073,0.6588527599123691,
 0.08291898392280994,0.6401684452308717,0.044055712661157376,
 0.5528252358440997,0.6964168518102645,0.5193653098112438,0.5985179814214712,
 0.7884997291261838,0.2249150529815369,0.13874003049484374,0.8751889823312156,
 0.6701947194940161,0.1860999750619724,0.3257271818478742,0.9485944401857056,
 0.13672410959471315,0.5662238274170461,0.427795468532571,0.5989308618380321,
 0.04192786503469126,0.08802540165593986,0.6806150755389021,
 0.9458850334740053,0.8475465516334487,0.6991253927168506,
 0.004799620718717312,0.17753099617392332,0.14442857214479443,
 0.1874309601071189,0.8594703985748162,0.8936770150468537,0.10540614305522,
 0.942105109309733,0.3586106513460534,0.9493817609581077,0.22009280189464198,
 0.24237441590144804,0.5476368967919176,0.42482349810954356,0.335267768358246,
 0.6100842519006688,0.2183862343559363,0.3377735669406996,0.27484294087558325,
 0.4940992281832999,0.8257300175162026,0.8856358289225467,0.3671868746830027,
 0.5777272816898628,0.770127857973648,0.20294897490895647,0.14415600648262328,
 0.3536859651282457,0.9598636695651152,0.46537975584956204,0.6865420686219644,
 0.9991219747092825,0.0813216753010142,0.797041521418918,0.20844800104631744,
 0.924211167397059,0.1669054842066181,0.7010238142820043,0.5863379042198934,
 0.6093827058598549,0.12275070970316404,0.9284471860537136,0.742016438304933,
 0.4039295076686953,0.01601449686941281,0.19820671196119788,
 0.9668804492709532,0.0676465115281003,0.6427293793402159,0.8334303171124153,
 0.5438509320631497,0.13202007585136344,0.5835101675718187,
 0.24505568501257857,0.2546534875765737,0.1745630693472051,0.8091746226546718,
 0.254958682103246,0.16645810202197397,0.6236503764392686,0.638750817281512,
 0.8961211638555037,0.9455878618344583,0.9999309552001433,0.6636362487795477,
 0.7208424620337774,0.9496696850643733,0.6704499209868487,0.7088932857527637,
 0.1322605870926984,0.8895676697307038,0.7860092804543476,0.37279976485315514,
 0.9118558563355978,0.7456035239305403,0.1361167262501055,0.24478071562759007,
 0.20921505188806344,0.7088129405417098,0.9313657606861825,0.5439306654923781,
 0.3009816249935534,0.1779722211501723,0.9521259699635471,0.9658455785487727,
 0.4180425465027352,0.9908658431780554,0.3315270612999254,0.6694455058775225,
 0.6352214030737373,0.0795337201191052,0.39061836398121075,0.3861232270946403,
 0.1533575550417393,0.15041077302582095,0.4818438830270355,0.3846782932922864,
 0.39172249632031075,0.04530651750998227,0.692836113756766,0.6207177176133316,
 0.877455165548116,0.209286296108251,0.3022822976317119,0.837751847094951,
 0.34716568601345144,0.7772581368448662,0.6381702158756044,0.5819578390618727,
 0.23534595896556587,0.38646890904725584,0.36494205834691684,
 0.4809136275360616,0.869938724675034,0.011748505311445223,0.5383174735418725,
 0.684637609989991,0.6795496223852245,0.9784310538505813,0.32861684132467894,
 0.6718120617770478,0.15060321239012642,0.8231277822552066,0.8207707518895108,
 0.9289708961812193,0.8385682005630263,0.7576115374447029,0.7040965899063761,
 0.028169539812191613,0.494918151030324,0.9645053945931813,0.2656377782494269,
 0.8654205542917186,0.37922348713866094,0.004730172546308586,
 0.9575119187623278,0.7308416342116196,0.13296427059480065,0.9253753854045599,
 0.8544948261478335,0.5201884046507569,0.6901509682375544,0.725962000556935,
 0.9559114553582198,0.7799866429120172,0.6692620464412521,0.8734930090832878,
 0.8824232731283296,0.680655112991079,0.28958580428948477,0.6878984730143543,
 0.7576830538668725,0.12816297608627647,0.21023059984674497,
 0.8974501532389512,0.6075665082336549,0.8910370647375948,0.8330956638535021,
 0.41943189183813556,0.9307501557621165,0.8441564443630558,0.5169194836450511,
 0.04809267605275647,0.148256563619933,0.27480680911142985,0.8119862434091664,
 0.5375112343086142,0.9602608670836783,0.6437119993676785,0.06691890789238375,
 0.97102276689572,0.901357462932352,0.6150041827488797,0.858479085225003,
 0.7182774407042987,0.8087013263214184,0.530462609390734,0.18656276343846256,
 0.9424514999422651,0.8647249938100019,0.32641980553934347,
 0.31418799388937235,0.6131941115828918,0.8958315062177747,0.7747973598784563,
 0.4169895311504588,0.9729695650393577,0.08684704228397921,
 0.24203433473370461,0.3789919478420265,0.6251301013008103,0.8136304272603067,
 0.6109372805225151,0.06065106471069537,0.017733080821696,0.07579343042568609,
 0.31034596438047735,0.8688763410273893,0.6926391369923293,
 0.06597437364245773,0.7202108122437695,0.84807706585079,0.3967637409986051,
 0.04747147268392127,0.3934441483823723,0.5868591414529073,
 0.10553800779663591,0.025588829921529,0.639421868978652,0.28147750506199,
 0.2765131930526453,0.6037688001160957,0.5015742146354578,0.5677591559207492,
 0.6461121252331459,0.3201727634328082,0.7161842926973615,0.3869341588993387,
 0.3893759396491443,0.04043756928394604,0.981335508345011,0.910619834458012,
 0.39671469161864625,0.20957498207274594,0.6173812323259267,
 0.6893654802829189,0.7669918239441548,0.4018773106271536,
 0.044493887588957226,0.7281978310107475,0.9201612922813821,
 0.3706229269571185,0.8014268155583799,0.5112242764724505,0.6693757162754217,
 0.39451158347533555,0.723502268049441,0.48906536952840773,0.5021339936706836,
 0.8340662964338091,0.6396275904739767,0.10853001249386263,
 0.48959503223648315,0.009556679255686595,0.30091992423374747,
 0.5582835187429067,0.5420086324911378,0.7682054933431098,
 0.040950828383598536,0.47622114407971905,0.8067383972862716,
 0.42871570763669475,0.2311649858391116,0.8059129344935714,0.3091608219432388,
 0.09154897921131522,0.7757335775265124,0.2573977636805984,
 0.06810083592560467,0.5496590949698044,0.17581368241257644,
 0.9914682172795855,0.9323702803025467,0.8172003282661405,0.36253188920307067,
 0.4241785327474529,0.6606077395238865,0.8350679444823078,0.30322870500426113,
 0.2594819787783438,0.3327510166570965,0.07121149985168751,0.2762690210322307,
 0.09228391431849015,0.6780661552601374,0.20676219039495314,
 0.8654508470905696,0.017903733377296116,0.4590653952997934,0.9335917494949,
 0.5783901001204392,0.6499749323955514,0.9454907086271783,0.363737063420014,
 0.9419601526336099,0.19921736843984017,0.9150429759569064,0.8569479830600855,
 0.7523382594687533,0.6449130099814262,0.545734065411826,0.8335549643494159,
 0.2150580984825654,0.3607754913794059,0.9715439750900543,0.3621902498883389,
 0.2982528349125485,0.4341991811657425,0.05258054743931739,0.126846524116059,
 0.6840525727283475,0.28013027395054646,0.5996778882348566,0.6965077048722338,
 0.2879626907492123,0.7865864807984373,0.8600259620325108,0.7801495025079436,
 0.8797037435433479,0.7299830399553056,0.6381737507472618,0.3595198847428662,
 0.5265611858203407,0.5911127489097575,0.5458620751034189,0.6814406808615946,
 0.39287279534429886,0.7441913196649236,0.8032739079050529,
 0.26772502344151095,0.76035705003439,0.5017406422139806,0.48818438299616174,
 0.6756404406077745,0.5835684637399535,0.3554620752667915,0.47910959453166135,
 0.016133164268471667,0.9861320212203227,0.17606694017820346,
 0.9355622008031087,0.9949848642073961,0.5322725643088547,0.6749222030967998,
 0.8191107669094693,0.5473941384009954,0.4516365184823221,0.02721521094457502,
 0.7249092119321149,0.24245250785048444,0.33063712677754786,
 0.6838329684690414,0.8731365903191616,0.8482653820438224,0.2559246241574158,
 0.06283822960409069,0.7778507437724765,0.5470073873798773,0.1475848395459195,
 0.968398927303465,0.28066254318401307,0.2448048581703144,0.5586287094092381,
 0.9232565859510021,0.2622005750474733,0.658058418727779,0.5137199013752263,
 0.8545407409756829,0.21207251901978574,0.5885051119991785,
 0.19843547738317946,0.7066405266883918,0.32726638120615303,
 0.3872366539624603,0.6519155790213894,0.6882112570939324,0.9476258479571461,
 0.5822394440462795,0.45535270434808783,0.20764195813249342,
 0.0364142875256287,0.7302331764022928,0.2701365909664951,0.3267127283331023,
 0.1330336470785648,0.5234274076754626,0.9226189398622715,0.9148620615892109,
 0.32779706174319645,0.39991730800717695,0.6137457865630331,
 0.0376037912560111,0.19484519765185948,0.730069400112106,0.421141908693591,
 0.06979646756037117,0.40187681421973387,0.10364368301839832,
 0.12479575561339862,0.5408907975594973,0.5106240646874911,
 0.17676334239033809,0.964467999646513,0.9100024976566496,0.218738392611613,
 0.5667674561075112,0.32176431026119934,0.7564321269192464,0.4046346708587778,
 0.7672339820516493,0.012964580871150932,0.4488397754274648,0.359346655400803,
 0.6519353981397101,0.36804827129200457,0.06563304593908292,
 0.5369400989561631,0.17089383289576265,0.46441532768578997,
 0.027368516997140935,0.42705108011937,0.6440936877227266,0.03118443880688815,
 0.39222541067693106,0.0511988385486315,0.020153901457646062,
 0.6465209487618716,0.782517065889046,0.011816723852805633,
 0.33467063187840707,0.025855655452938997,0.38387562594587166,
 0.4842757748936688,0.3965379707432072,0.7271274677103013,0.9998680513608277,
 0.009582985047618786,0.144779224233621,0.3758411921727731,0.434195354510282,
 0.028176320690299406,0.500738616713785,0.5705677330067982,
 0.04184210344694439,0.3406805507310312,0.9767524961836856,0.761540725044402,
 0.5674387454879354,0.977395001656172,0.6670221691073904,0.45107267524253414,
 0.08633604108712412,0.2798215833763048,0.9758789557826608,0.7417337719818554,
 0.5829044403925213,0.0348903191559633,0.8657126375582291,0.17266105293815814,
 0.49555474884580175,0.43175846773975113,0.9043665077711293,
 0.28559716425476245,0.07907054788028622,0.8258278144639256,
 0.6361139450883345,0.2577366838471451,0.5129628746096614,0.23995330455590935,
 0.28799305243144413,0.586067563859333,0.5692959942972233,0.23365190835995753,
 0.4205450605742894,0.9285488206469247,0.13221469873244907,0.6581167620856888,
 0.9776950355506902,0.004685250466214708,0.8845558314078907,
 0.21599992323735462,0.9636043479059025,0.456196171465288,0.4936777424486688,
 0.5394184445145296,0.32420808935615,0.1347165495305982,0.9083951926237561,
 0.43892863866740117,0.527155235902365,0.3766572697274384,0.8477385321270713,
 0.006325687306790506,0.7135949210719615,0.1842578463307153,0.946423652593076,
 0.9049376156461919,7.824885007390314e-6,0.7035634337355628,
 0.9224704451086096,0.4867459100682999,0.08830068329155849,0.8052459181293089,
 0.2565562152275489,0.6066850962454953,0.6695913047465304,0.6951478898091061,
 0.7434775028174925,0.32040463931223273,0.08325143234376609,
 0.6640449771511575,0.09697466198033311,0.6484777892749848,0.8323470691860818,
 0.5133917562034681,0.35357129497506257,0.2216440456777341,
 0.24199480431548792,0.994974230737922,0.6697467203745593,0.07684178614914994,
 0.8888066562646182,0.06976397639678211,0.14534630750800004,
 0.3546550687808472,0.9276885230286895,0.296214254781645,0.9454543818055552,
 0.61212899178779,0.44539114388720913,0.882734765396665,0.41621906440676026,
 0.8737049958478867,0.436700938270171,0.5842402934971771,0.5381746388356048,
 0.2826375556972135,0.633838813599812,0.1344399304628684,0.4433243333878857,
 0.3223155492881322,0.5446342808133512,0.89519900322931,0.39894285446962285,
 0.5027638369449907,0.5574561522960851,0.027052807014942726,
 0.9380548735259245,0.7359639736279355,0.8032266663379732,0.8423833893711296,
 0.4138072029773999,0.6713664870808014,0.20919610504751196,
 0.12845332929823727,0.5426297232052857,0.36720535487237416,
 0.051589273417233295,0.5820893035723049,0.05761841374689847,
 0.7644593953260916,0.7933624795323166,0.37756329700505264,
 0.18859341028808374,0.8179780966246515,0.10253166158398841,
 0.10689274413347583,0.8382823898709704,0.3853420569314364,
 0.07897796051236083,0.3854073848628876,0.7762294813744198,0.7756053079012883,
 0.5619986624148097,0.8474667312311253,0.40245893720757553,
 0.18718793125851874,0.6510273149433584,0.4368367155608549,0.7196489704067773,
 0.4123111014485724,0.8227193315650944,0.40939185965430247,0.4512905227816897,
 0.757022924878642,0.5695608128311863,0.4813412694374717,0.014424274521360858,
 0.21695796475521956,0.6564878140180795,0.8141693669999139,
 0.22262854545579036,0.42171332236055104,0.9158483004169273,
 0.25455281679167174,0.3040493461185997,0.5475487873964452,0.4088186159030741,
 0.7476670320096158,0.2974996431687035,0.41513444778762376,0.2605707444022496,
 0.03393265505595666,0.03639572335765551,0.6247047882026348,
 0.2775439907423428,0.8777445255545195,0.6266515553103327,0.6064371458500734,
 0.154116861408102,0.6408297858618363,0.4189220624194465,0.7611974951622948,
 0.8251229295928508,0.12100221416961354,0.19331214202162617,
 0.5885153562097755,0.8237538804039272,0.2955187664741886,0.9376524526967915,
 0.471385331879132,0.3752995041123599,0.9266664026627687,0.9853232793489539,
 0.9950183880363237,0.36773053573036285,0.16751839247176759,
 0.042413506096710685,0.8590585697477939,0.49416112586671834,
 0.1867194512469036,0.07032646617107707,0.9400843512773347,0.5425710384833398,
 0.815059689945637,0.8598711638052408,0.5372535802336769,0.09439439168533847,
 0.9350342902367976,0.4919867041544158,0.4484243241295215,0.424018265337866,
 0.2837186636279877,0.807646214920054,0.5511080944765656,0.2742719489731936,
 0.6860171050132985,0.17233940581656193,0.8337631449640497,0.7719384402390889,
 0.018014840884764793,0.39103408066624035,0.6880970215089726,
 0.4471298507542294,0.694398804829962,0.5370147624419058,0.9333272113868414,
 0.08174269534133638,0.6978787903948906,0.19753076949089232,
 0.4738273570587598,0.8121710152520291,0.17334392015249767,
 0.07217617206471116,0.20030346654193432,0.18464835843149185,
 0.8479351717950999,0.8278648100114938,0.2459721877398291,0.7025974849563643,
 0.29179577301899196,0.5969192553810301,0.4026744961388885,0.7525766394353726,
 0.4586726086398699,0.6107715042705492,0.7064790596006473,0.8669351574493493,
 0.4128844701601936,0.18313439436310175,0.6089552798165916,0.3037194036251497,
 0.16251193386545792,0.45763140720495776,0.02343275100686759,
 0.5945085020095215,0.717850849981488,0.8434421582180724,0.5359135319174829,
 0.07531581746878946,0.31983243468857125,0.8139993994024004,
 0.21138342047617276,0.9738084861827967,0.3253760676878883,
 0.37473353995149405,0.001199138543675149,0.37009802917183177,
 0.8652698853602373,0.12246322799375176,0.09209354851265295,
 0.044818961793087086,0.6303360070969924,0.021288511824488432,
 0.05063478952109579,0.64323395436884,0.5674816775703567,0.0460354653451619,
 0.15225891057559804,0.5799375299412006,0.04706147823113249,
 0.5320624882113941,0.5735928270290115,0.6630787436270332,0.32822412539959245,
 0.1517825666154623,0.2459143902794595,0.1759308135864206,0.5570111438839953,
 0.1362709911121952,0.5489226144990272,0.762775751772368,0.5690946059061383,
 0.37768797173370294,0.36948011267854586,0.561460759419578,0.4059116736965954,
 0.02239419015620625,0.8289641898212637,0.7097953271272786,
 0.04063486520371795,0.428635140174894,0.9918701814184735,0.8020840657770831,
 0.6656985467209132,0.47284452309087655,0.7153137793854323,0.9457248231769914,
 0.8391594207488827,0.8950397733774033,0.19836570168853374,
 0.44655226694442973,0.2997611431460425,0.481687436207889,0.3739166932544308,
 0.8964797203526979,0.23131004791971443,0.9789353686465501,
 0.045924843262111104,0.23549776311382964,0.3050135499478869,
 0.8214630361464598,0.2514143463249623,0.13687644246689346,0.2514136404508841,
 0.054180634011565365,0.024226333498519548,0.11629300318469382,
 0.30642970602706376,0.18569187759004968,0.21400147547141024,
 0.01148916382776477,0.21003352820683907,0.6716757754627077,
 0.2033995572782893,0.22032467123761346,0.8252451354015004,0.2929046953350072,
 0.5392439983903792,0.2237989029918248,0.07207726428078332,0.7707236681899661,
 0.40798953480933386,0.7158082961344157,0.5257058821603218,0.8930141075103831,
 0.24943525109515996,0.5918691616929769,0.6332449798145229,
 0.42715142905403525,0.4154738220489237,0.38088961738420357,
 0.06455482620424169,0.8965023225358864,0.6114324213588187,0.8899524308896513,
 0.7676285092070884,0.3473887059234749,0.17469540088831093,0.6527306424499435,
 0.8626117438756473,0.7802074291029973,0.519929354270386,0.9859429708046618,
 0.8316710430662732,0.23004312390649984,0.5076813399916247,0.7105018022560494,
 0.9280847096345883,0.23100576165912012,0.5160929959090451,
 3.18446892429769e-4,0.7907244047610299,0.08044326825183878,
 0.6169721951160456,0.4731711011089156,0.6346876877019717,0.9910752647398906,
 0.29663173734050974,0.7763806798451951,0.6325491852669072,0.7553053846052434,
 0.8893214271210226,0.25417643633840636,0.7701545094240165,0.7238131618135305,
 0.060907045974521234,0.3391044955542363,0.38313726825960803,
 0.34239701825326985,0.37375319160429976,0.9907243127435109,
 0.6926536260237597,0.47112513787067434,0.6918328664222752,0.7873869703699234,
 0.6176739738112202,0.449319132418571,0.14235603924407692,0.6687941027223022,
 0.7716298760447294,0.18046822117967343,0.9435783131350007,
 0.22531876900544479,0.5009038121927328,0.6764238266135674,0.9072526995138467,
 0.8269465344975846,0.29727741633625804,0.39296411384232166,
 0.5849127286135445,0.5503610595372547,0.17351585154146154,0.6674894573779572,
 0.8412395448637562,0.3177674235681537,0.13032919583569202,
 0.46533222292794363,0.013349402695801427,0.6403894842646446,
 0.3940233665922176,0.30134730946516086,0.11381291868131005,
 0.04336208637835037,0.13160807902048788,0.4762269151619618,
 0.7749193460114343,0.31112382815921014,0.45238561333693883,
 0.7535983866182158,0.9774640662338774,0.8198461367426185,0.8702800411248895,
 0.1374586724533723,0.015319674997742783,0.463353055994857,0.3704482736390202,
 0.09158146379653676,0.15141752076615644,0.08019664606573396,
 0.34396008273616974,0.24023037733789598,0.17104938826712834,
 0.6131170850054838,0.08643753586772351,0.5867274724910356,0.2787218422476241,
 0.9694288146123992,0.6940048164432362,0.5030481958649307,0.3290167600644869,
 0.2437863314559685,0.7594427471530077,0.8002568088975986,0.27042441444071386,
 0.4531349171979987,0.4366300218392615,0.33699916890809745,0.4072930466856959,
 0.5977622440670771,0.1853844308677377,0.9385347881283175,0.32784932180062876,
 0.23999950945305992,0.8993133975310934,0.7795732030909084,0.0893882183531074,
 0.05957360823630076,0.6762976463144884,0.42869598057600566,
 0.8380127933617617,0.4405436638720579,0.005180604254238719,
 0.09036719421314576,0.16720065488673663,0.2668257798716829,
 0.3476101821802309,0.6310018200675983,0.17412263340767764,
 0.30758657215455965,0.45381375222038034,0.40675197485348236,
 0.8950214468279556,0.04388759745788329,0.3884783686845853,
 0.14502050232420016,0.26478690522643455,0.5070386906146582,
 0.5887842281171296,0.9819697756807819,0.7197371592644255,0.7534572526627463,
 0.07853188727149396,0.3070272921294509,0.4858629467263854,
 0.18555430424440122,0.30785287429787644,0.76828551072777,0.6975530812552321,
 0.4748497135175678,0.6522509150915702,0.7721954767201582,0.2654698647131565,
 0.7671492377188258,0.9087010261737044,0.3540672846191444,0.5059066793380765,
 0.6983287741778593,0.7803203005781971,0.8397936507437184,0.08579870001475942,
 0.7773354711424802,0.6765354922268322,0.7909156928135409,0.4581542576261266,
 0.14851210266908577,0.4465976366755018,0.7089187034486296,
 0.33784449545653406,0.5989633201794775,0.9281559079729076,0.9543514383144361,
 0.8315391435949051,0.42892473844350854,0.6440677299660633,0.4030947043900325,
 0.5851836975136133,0.7738530673955466,0.7833130331918459,0.5220356531820612,
 0.841441849972314,0.6189748241344843,0.8875927312688613,0.5238452998373233,
 0.3144667079104483,0.12441401637433291,0.5429267388509591,0.0827321705424644,
 0.7474015164249658,0.8748835254307892,0.2511687094149999,0.12324481712534774,
 0.8502213847214986,0.9257597368652687,0.008303088572522932,
 0.4378471417566795,0.9405751716567429,0.20362055709159854,0.2373858574677059,
 0.8203325548800958,0.4267200182413585,0.9822410851724933,0.19688544479688486,
 0.545341035759244,0.8716858680244263,0.5256816068341521,0.15547481712104794,
 0.1863175893458534,0.26124733853532156,0.8526618838692941,0.4981272764198996,
 0.11048779026402045,0.6838113280213938,0.6225733185431087,
 0.21310345945390818,0.020443729132429644,0.2814021084925201,
 0.41316690870549067,0.2533094960964194,0.3676360068132445,0.3250870181486324,
 0.3517319920239932,0.2898057257108193,0.15506522401645062,0.6349805318945366,
 0.7796251982610171,0.6528012838019228,0.008199296007233903,
 0.41226677899991315,0.9972982148353893,0.39078384495198526,
 0.6759786271443181,0.5897632126778021,0.7401540635026731,0.8649791939934246,
 0.49406898626370055,0.3260895902768802,0.29762133779651545,
 0.4824076951454783,0.7584607242129037,0.23457567841459204,
 0.33840163062108164,0.06465911692292925,0.4554273316128441,
 0.1813434491971524,0.0044541812111206225,0.07807904517973208,
 0.2763506565014151,0.570737876732275,0.27042159604512306,0.5437888508406125,
 0.06098954976695459,0.786965894847977,0.8834864746335208,0.4574177098084734,
 0.39304317058717764,0.38987695532890543,0.11603314821702904,
 0.7031136037771832,0.5510683319118866,0.9191067948673174,0.929888109631876,
 0.7276942836870117,0.6144030274838719,0.07047506212664412,0.8720053145940523,
 0.004899759591827535,0.5576161724177998,0.2773449284570497,
 0.3027456394014346,0.979987168393874,0.19344175783003326,0.4640066629795687,
 0.0369130027224579,0.1628105416543677,0.9754643766570695,0.429029266287308,
 0.11132250844476366,0.05945282810272179,0.5852772758279543,
 0.5381819898813244,0.5498163139290535,0.9673613836064724,0.7856168427201325,
 0.7767074259879202,0.9609088236494224,0.04976759659366026,0.6816878796520522,
 0.8961287626879935,0.5782320361585033,0.18796982999846046,
 0.06588224512042551,0.8375407651180546,0.1788919306319121,0.0785291133721212,
 0.5634353875289192,0.42539847608730175,0.7064219755401963,0.8909739515756129,
 0.09747264008199064,0.29263753795372416,0.35428337787786357,
 0.7205138077124054,0.8375090118044135,0.5424090542766562,0.7810124782102775,
 0.49888147648659864,0.13539753563688894,0.22853050354497195,
 0.7831079676484871,0.6906723741425671,0.922280813555351,0.8384121579392358,
 0.38873549262715934,0.2110242032579881,0.7260709718135351,
 0.40656219521794446,0.5446924678636047,0.3033527156173065,0.6388195766391473,
 0.670612834574714,0.6378832028059085,0.7392839322014856,0.11880698062313733,
 0.70432643340187,0.5913410029004313,0.7966194959417788,0.37673343769825085,
 0.1716028110447324,0.21650055917332767,0.9504098360066398,
 0.11076811821234722,0.8578836012361909,0.794683222110967,0.0609432804841642,
 0.7487484082653255,0.4919973223514025,0.3725602536563265,0.4939113717906465,
 0.2182673330806827,0.41446755818323056,0.7685873273950562,0.1289795529149329,
 0.7295775599663584,0.9743685113453864,0.5368636302689498,0.657556189091886,
 0.9973089748600378,0.8106079204662707,0.6360144148703759,0.19739243805691076,
 0.3011508276274011,0.6050571532347996,0.45604824182762593,
 0.47942701188218795,0.3511981563097045,0.66183256336122,0.7448246156052867,
 0.005484603644233643,0.053453978953208914,0.9614051915299071,
 0.628150525142926,0.4677071837215414,0.5614445339451185,0.9691501946294214,
 0.39589929722167483,0.6394462015686155,0.6964922904275034,0.5074517343708015,
 0.3064003102535684,0.03345809118772436,0.53024744845379,0.9463698293175871,
 0.9942068933935416,0.31564290158282005,0.027328990959685218,
 0.7553760659296382,0.1317802328671196,0.12759977524713584,0.4494205403727126,
 0.23066719265911884,0.4729181577763385,0.6712439350559152,0.6058167609703633,
 0.17892032331646024,0.41891905838728094,0.6579099178688553,
 0.18607096545186064,0.794846374630564,0.5543550307783354,0.10865867028809595,
 0.27521182030819635,0.47916188682820116,0.34973004640618366,
 0.8569825203297203,0.4968694998225356,0.7394947448400185,0.5756671228431548,
 0.8760816343218661,0.31122507343657224,0.0732921010001687,
 0.17279753082928317,0.4296110185924431,0.7704599031465541,0.9277708467961245,
 0.3450063425470733,0.021008668563617605,0.5893462130937078,0.371544959866691,
 0.2360414253310834,0.6334955839464119,0.18668703165324274,
 0.13756809223448496,0.6727750585824095,0.7409649860759396,0.794826523484331,
 0.3163632929305019,0.3486528431682302,0.33366168134308416,0.6105660332442261,
 0.6034305246640705,0.4616752661518594,0.7621522194654138,0.7936586967572589,
 0.3898876329956926,0.9597341430221668,0.028391255409115135,
 0.9666477512185176,0.3066674027019646,0.3575131405692872,0.7983252257657458,
 0.9132260205948539,0.4046524509712984,0.11553156543741028,
 0.19204186569407866,0.03700124834153895,0.63939713326159,0.2521284131390006,
 0.5701831977913091,0.7877955205131109,0.40966791807632297,0.7042012653103735,
 0.9952336867644056,0.6736154400597923,0.30113574609161586,
 0.007672087579852382,0.2891206167023197,0.4620205852003988,
 0.8488135020297028,0.5564795898614731,0.7573088907458401,0.8740916219886842,
 0.8763717963583968,0.003338046854420962,0.4033201546580222,
 0.2567349650433087,0.2853311126630702,0.09581856838497327,0.7886254593027054,
 0.5534645513492891,0.907289381014345,0.9199576812959027,0.990654105860268,
 0.504347483848516,0.7887184044240777,0.6300467515177595,0.0578584582620747,
 0.5511294411942055,0.7388947600829874,0.6062006515092104,0.7915450010844998,
 0.12427738327337845,0.45125753808068236,0.9896665695129486,
 0.4176385999605836,0.10020764161870666,0.5565252685650904,0.8104617327089313,
 0.7750720865751697,0.9742376589679312,0.47724809715103234,0.5874951773275168,
 0.12507081214660842,0.8648684149651497,0.3009683812767475,0.4150293293321884,
 0.4910234998021963,0.7206843738041392,0.4389447673282463,0.07982306020093344,
 0.20081593000683517,0.31854665446784747,0.9266136085968464,0.32046537484794,
 0.32497230094003904,0.6734652441526824,0.803032586680475,0.1420829289156854,
 0.5194227679048622,0.7609235854322838,0.6839521954744671,0.04354377534907572,
 0.8071608029831988,0.7705107659676336,0.7883634133260194,0.7818162948254254,
 0.284128743676604,0.3892374998439193,0.8120345121928473,0.6101375718549625,
 0.7074423857095097,0.2919761796696049,0.3645770818148387,0.03169871487512266,
 0.5889255747318474,0.8883048996444338,0.2531777704109548,0.8290148957203765,
 0.3972528937565991,0.6617103246000896,0.19112453540466434,0.2331728184953663,
 0.47537251082844134,0.32123477636765174,0.925348063841943,0.8778056048524997,
 0.9961470391533866,0.8151758547476703,0.1312236127000549,0.7150971728698906,
 0.5825232401308029,0.9330498430012222,0.13240077827085328,0.7813354539029902,
 0.8952839977208641,0.3116432341816393,0.1283700644418172,0.1355729447324291,
 0.5666269834146406,0.8653784223412953,0.1936463484112314,0.17301480761675758,
 0.6680914093850248,0.15998542088121048,0.8761857408376128,0.6744771198692923,
 0.4973194951538026,0.5771464926328271,0.8695022644420133,0.6653281147179828,
 0.9169031864453511,0.7841539647743968,0.934719018921111,0.04759969236257522,
 0.6271992901264677,0.6989683330968414,0.7898962008482403,0.7436088563310278,
 0.2870349086008802,0.5706874608125677,0.02348340561234341,
 0.028566632762648703,0.744359894359805,0.18137543416514712,0.629367675767202,
 0.44822424923350457,0.30234763416069366,0.1762501076374292,
 0.8542646536232148,0.4154329168417674,0.47524314188893924,0.8875703885896293,
 0.44030667667626067,0.8194725275533467,0.4919453604874776,0.5460573448075063,
 0.029652585768478068,0.6722697545721171,0.42859033268619096,
 0.9686267498804009,0.27243613022568125,0.7874812877889128,
 0.47643808041900504,0.5587444540632308,0.09943157956443938,
 0.13976416523175805,0.29573013841156015,0.9694523111703487,
 0.17491180126349204,0.7589439624111495,0.5593009276047485,
 0.36417180677957495,0.5154638517942566,0.8444543828640093,0.923801445810219,
 0.8341811109014292,0.08510078357534367,0.8042350609366214,0.3083848683199709,
 0.08605007687915733,0.5468784846716298,0.9376875528761289,
 0.00878945611205506,0.13192744592922667,0.4994694737878582,
 0.08973650774639963,0.058421117781058074,0.8133416680750304,
 0.9429841888990501,0.17196034079907419,0.4926989022120779,0.711101123013925,
 0.8495243224606687,0.4857129532039376,0.857779032153844,0.3665575397733787,
 0.345608363776136,0.5042568519288353,0.6656660144345399,0.3493766310800247,
 0.3177410437281992,0.6426420487653647,0.9471885953931749,0.2984174144889048,
 0.05420849619238144,0.26047941336200475,0.3201681354807837,
 0.046165474644056625,0.2978040296921789,0.6600387719094206,
 0.2801132334411235,0.9410970475986675,0.5754298714420686,0.6551687614820113,
 0.9991590619177575,0.4588215787094433,0.1244409521616876,0.02725689289273414,
 0.42820140178808286,0.10870478711831044,0.717662840902525,
 0.22159149714731807,0.6605778359673143,0.71252570417177,0.269616164191817,
 0.917592436615218,0.13676288406850334,0.10887324177676472,0.830576216753269,
 0.04613205867204351,0.3100767322916238,0.331883762652458,0.5907364122529306,
 0.2942856478817335,0.41916864016805344,0.33858570643046215,0.827744998906629,
 0.5505584192719646,0.6421299604059161,0.610846197127636,0.6337426033267072,
 0.08179332563945751,0.4539863615688957,0.37000968303923876,
 0.15146365569767029,0.5320239557020827,0.25158201851297257,
 0.09343093551184767,0.45956159184719814,0.19761993606198502,
 0.13966596304327117,0.3692472484994598,0.5230803672414647,
 0.005800484400579675,0.7980895399206153,0.7669389874483525,
 0.2735879984431583,0.940919437893381,0.5443506428987571,0.1290571318633218,
 0.11508161125057303,0.16058291766736132,0.6689270288996249,
 0.14226999587084377,0.40012426019718983,0.3881202525814741,
 0.7358797044214264,0.14684703130814303,0.035868060228218646,
 0.12787287909251033,0.660766264316671,0.5921326384990312,0.353753362194845,
 0.6930435117650269,0.8003748719258971,0.3417889079345918,0.7645330767092297,
 0.10735124452569078,0.1938756790298073,0.31504373020548915,0.78552113714575,
 0.9614204268559038,0.3603974273492693,0.8818603303277501,
 0.042236625095987135,0.36505461063299793,0.8208876430711264,
 0.1389840202052286,0.5165025444773617,0.9142040272623935,0.142806281452018,
 0.997481063560888,0.5175655053395508,0.44075548015392096,0.4495132040778709,
 0.811096243879056,0.2754606336345511,0.969250553029916,0.517711030232716,
 0.9034249829421004,0.011659134079475275,0.44812459994732223,
 0.5372451612759841,0.7383979891929644,0.24695550641444952,0.5279318926325738,
 0.17876288477978308,0.9516839445261736,0.3513000171837558,0.8756387165814187,
 0.4995916661196306,0.8140476076835963,0.747645208382376,0.13003459068069145,
 0.3160530739901145,0.6952435921854319,0.3887354346095666,0.24346788999325286,
 0.9541605952236676,0.22017702862681554,0.8644600216981178,0.2494178369578537,
 0.3440683889382272,0.7787094311564654,0.9698672611107048,0.42225531460931287,
 0.6818023187394744,0.7043849703217658,0.811835122235032,0.7319324901515121,
 0.3952580866053387,0.6347864473345061,0.37050161576859386,
 0.024258210418522386,0.26018414339468965,0.5458779952392814,
 0.3557832189828133,0.24029074937160255,0.515587141039492,0.3546795603814892,
 0.5813585663422166,0.8016630807418157,0.2691558334560161,0.09211885538355769,
 0.679094702182268,0.27430813034157875,0.8270239220287778,0.40218111995740324,
 0.48994521760736043,0.7849696294018684,0.7837790697023989,0.9880041376094489,
 0.89488954433642,0.09946791955333711,0.5678402455711251,0.7552008943910348,
 0.33072948032975646,0.8208023835114177,0.10684390591623893,
 0.7611798439739786,0.5089779861940384,0.624895515753347,0.6801624064104344,
 0.6344732584513146,0.07623928832776916,0.9322853380787988,0.435424249026627,
 0.8204421584194762,0.5080927466809083,0.6949240206084645,0.30586062336000175,
 0.6860817746899672,0.18279544794739533,0.49912619060610885,
 0.4145677542952685,0.8663203027990825,0.23231996298130242,0.6998333361267961,
 0.48634878338030485,0.9381318546754881,0.2353312718348165,0.6000649362625277,
 0.1281128371119059,0.07550282528013619,0.19722968828771592,
 0.7997440142113577,0.538718045532786,0.6482304438831893,0.5753763599139202,
 0.7203972179623755,0.4777646026261002,0.8903190961724949,0.8137216869378207,
 0.031684459809555254,0.8022374716491821,0.3830162609063418,
 0.04628529645835655,0.5745440698786626,0.34415010592632145,0.647906476312505,
 0.11057059049110474,0.22479466676697335,0.3308737302062441,
 0.30424115459584355,0.8616995142338107,0.7382544076724817,
 0.33533786861932124,0.6818736199577994,0.9982184865114927,0.5117624406459459,
 0.9093303259106373,0.934789480617692,0.5454376565927936,0.5094139548183177,
 0.4594148884985646,0.21252647459509777,0.3436330530226065,0.6232770468455266,
 0.3113662567526516,0.30985732227847684,0.8712913394010816,
 0.18892595216695207,0.737792076348655,0.7129993145314752,0.2905646572319136,
 0.30599191936375025,0.9142578824099096,0.8713851182459894,
 0.14089699448229276,0.40585814363397177,0.9441508285667068,
 0.4523658095353006,0.1138800900527579,0.7060323592819664,0.48846802622038654,
 0.32039771235949865,0.3654981938548624,0.987958960731575,0.3094584839191009,
 0.03809139845976828,0.05946727622993353,0.8807714970586226,
 0.16155982008125058,0.5314144658627636,0.6304995974922354,
 0.06159796614150248,0.858699482459759,0.5268487881655421,0.8760270302746739,
 0.6145458146156413,0.2930889211678336,0.5356927824773059,0.2212766613639422,
 0.8312677105568748,0.34709365843358453,0.7680810228761814,0.282949248274331,
 0.9740922610572472,0.06782191766059975,0.5049905142527096,0.8670989504594607,
 0.05736856311925942,0.8895452124400943,0.4887969177864686,
 0.34630244650402064,0.24100395130654095,0.471175725106456,
 0.35840804416442185,0.7564274852403882,0.30531409105953455,
 0.9840135624682026,0.3212942874520812,0.5712587159550297,0.5556958394932331,
 0.3726325521342777,0.4948311352479312,0.5065332804595833,0.7149848648793262,
 0.3616952523130883,0.9717291044467135,0.5666706426397785,0.5771909180709351,
 0.5363201960260586,0.5513299723541194,0.8793460025279191,0.7909680400014578,
 0.3566633981747257,0.3881350190969849,0.07172423684408635,0.9031509338430755,
 0.8902405225948598,0.3696047976039739,0.812562728499475,0.29204656672257345,
 0.19176508936694203,0.725408576594175,0.40328579436718204,0.7249764792520511,
 0.021501798514399706,0.6448232127741929,0.54594453071019,0.6610064453217669,
 0.3800802723230352,0.5921310855617902,0.38467778127456576,
 0.21278826155881359,0.8929184952226915,0.3934756516652831,0.7400934191629284,
 0.6241311127069511,0.27707628102838733,0.13327030192670686,
 0.37225859872561795,0.43974863986929313,0.6531192402849715,
 0.5258082565306706,0.7212767143336635,0.3605616168447858,0.14838619845957046,
 0.8766848075995538,0.4783799742635565,0.6746171376857415,
 0.058615613376811426,0.9826846997801857,0.5566331446929195,
 0.22945273358107532,0.11294060729621946,0.30264696091178456,
 0.3969115623155173,0.14085803094065974,0.2580537161005667,0.3132454733063369,
 0.7566861468429521,0.5820017027296716,0.4305819813282392,0.397149149302942,
 0.7141997056002511,0.12010918666696535,0.9684268744014232,
 0.012791602857960549,0.03206504965507351,0.6901076018496342,
 0.7867544096317409,0.2054611288165914,0.10091299138538179,0.9756605780844977,
 0.5761867488322466,0.5482487063036645,0.20166176642500822,
 0.016906300953665188,0.6342867237469214,0.5998224850028128,0.213458932048184,
 0.38975452534839317,0.6015861481475215,0.45260702317677626,
 0.27056939048634776,0.4812170563227759,0.0069923593394563,
 0.00395419526712637,0.8396217045223677,0.12165172092997456,
 0.6667972588212318,0.6439781291908281,0.857584930652251,0.3194890590920897,
 0.3124569550870612,0.12142012007041203,0.6492541200863565,
 0.43333842089344876,0.012345786818200644,0.514065774171588,0.573343691668558,
 0.19871268828265598,0.7984116725128947,0.6648041387886532,0.6061616564331062,
 0.9881062568768301,0.4803541143348625,0.31195290695676636,0.0715613152564556,
 0.45077122047531437,0.7713441607913338,0.6710980850774647,
 0.43464490887767115,0.9896752166657135,0.6646858153202196,
 0.24404742840065174,0.6953034796531424,0.45034502997028014,
 0.6439087896214546,0.4299888458209644,0.7333093081234177,0.28551147740720906,
 0.607066187448319,0.49093493805329325,0.539933925181132,0.22845747993101784,
 0.6812921282961021,0.8471652310923872,0.34875913555403015,0.3747727825060846,
 0.8466471824086694,0.6923233838202857,0.6106202369672956,0.5177214795911537,
 0.0031290160395082145,0.9285606789653288,0.7568027703084429,
 0.8302562961757768,0.041705961124381186,0.057750281685872706,
 0.5888541076815752,0.13869247740955526,0.8842289778535386,0.7340237064093281,
 0.18535611909474448,0.5422378133708532,0.08341939435938939,
 0.7240227316358113,0.16408586580515283,0.665475907092475,0.6410509645184582,
 0.7000153165578359,0.6706833488914584,0.7618895074406324,
 0.021259287144128836,0.3121845898873693,0.6188470090236234,
 0.9343460461773276,0.4089129483363637,0.3219217818642187,0.5998697267041799,
 0.47449850782758174,0.4563077657087016,0.5017038392193045,0.8611427390485715,
 0.8568607665389716,0.6245167590000242,0.1317742402429608,0.9404133240326168,
 0.6376592809482756,0.06725688787000017,0.49332468113922734,
 0.6836295788699203,0.4276677965947636,0.391207635442836,0.10508027754633664,
 0.6458018720937181,0.6514214771775753,0.5936772899152601,0.48963304336193325,
 0.4017472132080232,0.30415388729383863,0.4969128106567051,
 0.41119693708484484,0.045975988389130196,0.2684268474988688,
 0.05405972351228472,0.6066635022712284,0.44745351276874057,
 0.8311942813466272,0.2081853392137104,0.7369939615015079,0.17846414126639776,
 0.35625312878355675,0.4628236493549829,0.9976267319582274,
 0.11052161084273338,0.8673906861171118,0.9154869769950416,0.9546321477656992,
 0.33993145650150214,0.21265315800361018,0.6388746584564432,
 0.5902714132826709,0.9688223690742329,0.587555258869065,0.7558808822273181,
 0.3217474425203145,0.886779841948028,0.671127435744865,0.44982078282529003,
 0.42252315519629624,0.8749871334106293,0.8753874250006222,0.6296771022210622,
 0.23441206233098155,0.8133546897099141,0.5138421858442801,0.5030606207991944,
 0.40888777008043264,0.7352172685585738,0.5261345660388491,
 0.44754314390042926,0.6282309929021678,0.18032998148108603,0.654538987260735,
 0.5371172153982448,0.632143521943637,0.743008332413049,0.43230559012818315,
 0.23346598030124788,0.291496164755471,0.27961428252615383,0.5368281778407687,
 0.028539886563257788,0.3581340276198227,0.9518678099098925,
 0.5410208686319748,0.7340257576762552,0.30031419170127605,
 0.31286495799894065,0.7250461760667511,0.4390239955435322,0.4967334777269756,
 0.6801857174864487,0.43017168009570395,0.47881418443167356,
 0.7624441517619478,0.01852767847914949,0.7866223734811821,0.9353512470633143,
 0.0795603874338453,0.5561014582524239,0.9867944700762016,0.36149075603070435,
 0.4286020916947013,0.2671112438229889,0.9076173115272728,0.7532591941446035,
 0.18831118351723997,0.28054098138502903,0.5236999310943273,
 0.7316646598257096,0.7312436331799191,0.024264383421819025,
 0.8612086644804267,0.009469614575468355,0.77404902701504,0.3548512932694009,
 0.5320489464348253,0.9107918115543655,0.5277275128035493,0.997791812826832,
 0.9287624647794671,0.8257344788334806,0.1446556179467221,0.5337220021855993,
 0.6615676312634255,0.2810381886251344,0.9310237326431148,0.9623373345363334,
 0.7301807586242306,0.2491886312907452,0.4555351896740283,0.5190285638400393,
 0.43408601099225375,0.8605933936105782,0.9405746412585003,
 0.11334499711049473,0.7067027562139443,0.44048481823197205,
 0.6396323054759674,0.11958136371708117,0.687044660387571,0.9295478189008345,
 0.16653894800411262,0.7853622443375022,0.2622972051441248,0.3970289525285964,
 0.5479074907529549,0.2837417120060153,0.717837663200565,0.2889058459995021,
 0.18181679119344063,0.2816487506160327,0.5296389528993267,
 0.15389083582836516,0.6728887510603436,0.6012069868275156,0.8697935945674893,
 0.9173382710675879,0.7995411199465857,0.4440935962077466,0.6469934381217312,
 0.04865132657244897,0.4814367079902432,0.7826890696737162,0.8454925081019826,
 0.40001873029903723,0.8016213325153513,0.996197090264627,0.36034429783480126,
 0.9787758726350237,0.12315222658833802,0.3708876426568619,0.8400542263636965,
 0.4272601915007783,0.6105560160284025,0.1103579276139861,0.49605938514370906,
 0.5129788170790496,0.740925316939932,0.1438649370608286,0.6885482622200108,
 0.2663986768304515,0.30405059865833695,0.37769903445706265,
 0.09042737990220284,0.821829571366499,0.9251409110819416,0.9960877779799783,
 0.42569303203333453,0.800399498834227,0.2887277182930006,0.7558452189477561,
 0.056543254591806424,0.5215066704339336,0.5613336950795924,
 0.8129312997473471,0.057724921669467855,0.17962617452915508,
 0.7317898686204526,0.4520628566484767,0.5237055188053096,0.09220538117338883,
 0.9325989043105132,0.5756273974811146,0.17456142277504327,0.1019344839823475,
 0.38867220495804045,0.9628391732127668,0.3625813474492341,0.7229854516673635,
 0.7400366122513861,0.1798319052092111,0.06748869425673554,
 0.21592094848987808,0.17165546017182642,0.03182858178071113,
 0.31316414819098715,0.055909247182799326,0.6891638307469435,
 0.19727574274840787,0.49763948860008755,0.03723948568597957,
 0.8994972614349691,0.9699117828266811,0.9709021006530967,0.21109310130958636,
 0.9158667433122808,0.8908214811564015,0.8670747692018472,0.9209554000663269,
 0.6016350775396818,0.22387482931266145,0.2540556154215087,0.7600428954809818,
 0.49056295232406233,0.49515520611707475,0.24772796810489606,
 0.4294916127044429,0.555727564033949,0.9515344765211766,0.4318968708166696,
 0.6460979411379034,0.814647529238871,0.21787743963417683,0.38119808748536943,
 0.3072093021267387,0.20050350584200405,0.4155835131929657,0.991411135580796,
 0.0598309466221848,0.6126827338535235,0.8443732034237637,0.34190132060518785,
 0.13151976115694908,0.22832147466546904,0.11603966241281805,
 0.5454381275252183,0.29727661075579626,0.8135464436397326,
 0.44324606369305086,0.7110568775717134,0.6260408929386286,
 0.33048314448203864,0.1292439563018084,0.9712192334982399,0.6547025639238514,
 0.24831241891514944,0.49209130080927654,0.16109366089031296,
 0.8755780188815816,0.5736091781651926,0.11846509372515512,0.9294460681193634,
 0.2814390073457641,0.8244397080237856,0.6815863931962445,0.09647376895501503,
 0.17754055306046224,0.3662599695154418,0.09622763965186854,
 0.9102791236535217,0.4031437904024592,0.5774416627772166,0.670396169062768,
 0.09645537709322494,0.4646033110164547,0.21190829915919418,
 0.5939699038650041,0.48896443986089455,0.12231237885337543,
 0.22158107248588543,0.01837956273891228,0.3694598547720265,
 0.49037909364555954,0.0926317636267785,0.6122336973059626,
 0.06723507199789636,0.6362490648978945,0.022156785580029426,
 0.2792462616003666,0.7117689706030855,0.9367384257763026,0.13793418825226544,
 0.4157389576942213,0.7794438469905556,0.6449112968607879,0.280317622403662,
 0.16400952271417013,0.8574263858254587,0.18919442884289384,
 0.5936973367774077,0.5628579379860573,0.5762553493566062,0.9234482449607889,
 0.44497358084832284,0.010410385448129889,0.23886639047646618,
 0.23229928414507095,0.06663099949933637,0.4645403811410489,
 0.08347974602038954,0.8904736252481336,0.740151289386281,0.8550978710722925,
 0.28687944447412805,0.751489639487731,0.6687375035439379,0.6948366972267306,
 0.9892882231183081,0.9639067010936283,0.7626981826879768,0.7689703889971461,
 0.5114549781233626,0.7527155328389215,0.9601342858327566,
 0.007035313812717603,0.3221099769503645,0.6282777376184892,
 0.3269133837818271,0.35748951705822807,0.9501206429724506,
 0.43194683827314106,0.3424075004138557,0.6252813438308914,0.7715684336765538,
 0.5836005553021062,0.2635496245914005,0.15536385807845443,0.9428105190828762,
 0.45002781847491646,0.9988162243510675,0.49056214391038144,
 0.42179795967350564,0.4679810059462328,0.9575873098568572,
 0.42456437929584623,0.5311763980248062,0.9711302997889073,0.3604615835466386,
 0.4755148141044947,0.2837685529310917,0.36824626314737874,0.620852286019048,
 0.46888399887167265,0.9280059431013361,0.7321283764400742,0.9389447039177187,
 0.2017496619213217,0.9374899663005797,0.7786922632025804,0.5518687506941886,
 0.9273599080655441,0.5192127557078728,0.5421922244347135,0.1249489654741679,
 0.7315180176677267,0.1901653666966806,0.23887718671052605,0.8448812534252654,
 0.7758640685499492,0.4307904066410918,0.4528066428922327,0.9599427718901136,
 0.23547682218317612,0.5859337157979456,0.013044154822056453,
 0.35646135695538805,0.6815383095320112,0.6746295949050023,0.8697007199478474,
 0.7179837639561438,0.7990660966512274,0.6534458709170718,0.14992598777237065,
 0.8996355078579404,0.5009292672821842,0.8329072370625874,0.49014985113078435,
 0.5536283598033849,0.14581776349788544,0.18298808654556087,
 0.48553159329126105,0.4545374810270095,0.5790266887889162,
 0.029068150143208182,0.42247676713475124,0.7153882012729497,
 0.8906792432694924,0.40465547245177647,0.057964460035290766,
 0.7125187480912692,0.16676039590616742,0.45584908449727624,
 0.14238728446940818,0.18653351648048488,0.5311139236370683,
 0.9794220895684623,0.45503454008269706,0.8362544814739866,0.194355698361935,
 0.6575631975520375,0.4282536755906573,0.41499818316758097,0.8258349295801609,
 0.3658974884854558,0.5732245331495003,0.3376046368863276,0.586580841083494,
 0.18842514850300118,0.8179607857587088,0.28457431480770945,
 0.6255265089436151,0.7360716435841763,0.5503077799785462,0.8719220837181934,
 0.7370262134489607,0.8267394782586905,0.12842276988790624,
 0.11322020131700294,0.15670537421112685,0.42033489859268736,
 0.7772558900840391,0.17359402412739044,0.2660811020540025,0.5188130793792421,
 0.07376634937085558,0.7937396940086163,0.14596106214175797,
 0.8677992307058549,0.9670930625548977,0.5248416206602413,0.7495155007136715,
 0.8658524336819378,0.7160668500445333,0.13979844523567397,0.8510437934351481,
 0.9788888030494941,0.04900240481343554,0.3655702845944162,0.560296334229367,
 0.03402598145995028,0.9200380145293507,0.18878376563281019,
 0.10319825694924578,0.645460891977377,0.22935945993386153,
 0.04202594235177459,0.4543250999695979,0.598350358058444,0.806006973643876,
 0.7444031387977106,0.39869058296705273,0.6356075653448016,
 0.15669468251099483,0.6389381491200136,0.35173671774555226,
 0.4619248307574514,0.6353396269228817,0.22285371186282865,
 0.17140845058233434,0.9482273537676476,0.4547665011264006,
 0.08454077407754901,0.6466201509718259,0.902264070325603,0.23286019952088655,
 0.16356502040230547,0.8263115453119435,0.0077204485916094745,
 0.7355414352204008,0.31677894561117725,0.2885795752547876,0.7827125511276818,
 0.6150720397309386,0.263183710220085,0.13968525139813603,0.5746052555723249,
 0.29151406242793054,0.5316727626221249,0.009772090106532683,
 0.0308285962000191,0.10392348234536919,0.9627539440384187,0.7082248349252729,
 0.7253266644758919,0.09163003028972971,0.6739887863833934,0.3960152457426287,
 0.15968985985228723,0.48648734982651365,0.9429621417711235,
 0.2243583691874783,0.9271315238851879,0.506146330247073,0.7252920403965615,
 0.4130521999605543,0.05685182476438544,0.27055314900960015,
 0.6552593204588905,0.9560285097383433,0.8176457891256346,0.8080579849115525,
 0.4971461760506807,0.2935959517295337,0.3294828723328711,0.036343180543984,
 0.46818306615198724,0.9747784637658075,0.5954978887200615,
 0.06943712833639071,0.5252576006803168,0.29452890792892183,
 0.9134703373186116,0.4269399973767687,0.21102895798458532,0.4282580739093521,
 0.023001606925976747,0.3778465980785838,0.4767999204311201,0.689085430460922,
 0.27033644497568854,0.9602544744341117,0.2493119281433449,0.2598728491121266,
 0.20856417198414667,0.4842620288250328,0.8913542846140016,0.7592282433716439,
 0.001614567370525677,0.7593830678744762,0.1598132684644583,
 0.6698569559143815,0.40877132140819306,0.031642325863182275,
 0.8429871898170245,0.001067830158132388,0.3247461289078546,0.228408333604041,
 0.773548738208229,0.26132595441158435,0.49163076887430235,
 0.14235199178008995,0.9327757071874774,0.4852358817340806,
 0.42370456711322135,0.7156735062237018,0.42981068766068076,
 0.9159390190163114,0.1711529840467454,0.577413969749264,0.035562511848503675,
 0.06940209285279764,0.3326042838653699,0.10664870946495919,
 0.6494104099188638,0.40858914979140837,0.4421733369779347,
 0.09724610854655635,0.42465457808790474,0.30775068813885964,
 0.6645801241773953,0.8493149151284456,0.16125540310618502,
 0.08334233665063517,0.3733799612764428,0.2295115201313903,
 0.28645372386869306,0.2890179042163732,0.9001267887505586,0.4755538975207123,
 0.9669559869548834,0.3200456950648668,0.5965749464355585,0.07892020936540334,
 0.5596473636543786,0.8182671888871785,0.6104249836929214,0.9751438712120961,
 0.8492005869442655,0.6676285718065682,0.172986724801244,0.2010501405918207,
 0.608961215910967,0.11048526819425741,0.12074339234709686,0.8435484878410319,
 0.20700848023949447,0.472103908172701,0.261550899029655,0.5004403558875729,
 0.5432617027311166,0.9041936990116866,0.8506305342046095,0.7549351172217981,
 0.7218739551545699,0.9310867193117457,0.8550311595086613,0.4130995465939762,
 0.7200266595709475,0.3839588420054091,0.6509360466023587,0.63348873348413,
 0.7709245381049298,0.14216407084206695,0.9413312204503478,0.2120503932007619,
 0.40213770201337984,0.26807988088039636,0.27746057137471003,
 0.7700164608566107,0.49480746565477773,0.8388281382594841,0.8874118550306466,
 0.5102956794383703,0.01652905585657527,0.9870548424574794,0.5801435187945263,
 0.5993888520619367,0.43130138716668776,0.4842699156939847,0.6939500459815723,
 0.7724787232236942,0.5535919996763281,0.23067163912254784,
 0.24187849291431784,0.49883959756418716,0.09482657666267968,
 0.04122036282023678,0.9663961820135194,0.23349086073586034,
 0.2556779808335671,0.8118749417091652,0.28263630162383835,0.9885623224413366,
 0.6011178429838167,0.7347403911910333,0.9470059942572066,0.9993954738836445,
 0.871645978107443,0.17588021224549166,0.7589520622012893,0.5160597820894576,
 0.20238306435691678,0.1479843452936439,0.5399677451604553,0.4678531220917066,
 0.46381513319757295,0.2853057011773006,0.3983236548014425,0.9183570320850434,
 0.49767229027311266,0.9789197305726082,0.6727375293598078,0.7623948507375016,
 0.35864813612736646,0.7306577274651294,0.9952664588879011,0.7873150783058411,
 0.8313976710848185,0.8715446347997265,0.7765078575093908,0.5704458657527837,
 0.7427417159978346,0.722330837085632,0.1368796426063521,0.0203059342953561,
 0.18309303450954362,0.4729637810047955,0.9292853963657183,
 0.11966993129177717,0.7510357119256894,0.11615832951712257,
 0.12695100216367405,0.7304575387357864,0.22897572509882602,
 0.20809956367493887,0.11360218331264282,0.4968024697132699,
 0.49908158706834205,0.2739055648620974,0.2267898056075347,
 0.17174428876608272,0.548695911667422,0.7740045499572984,0.6719483742031622,
 0.2324343547031551,0.5139800317949977,0.14572803527531353,
 0.13677206937624453,0.011097705400535895,0.353705925047447,
 0.3936002877347008,0.7052747774904201,0.3626736893095055,0.5351664549221417,
 0.5669856472413749,0.2986508837406132,0.14761490258980137,0.254175744072908,
 0.6526359962069582,0.4162720212406411,0.6957611529793292,0.10840275868435445,
 0.5887557557317862,0.910931673911879,0.1792349992322062,0.6919649919552393,
 0.06084189060907419,0.19807230159183564,0.8605592324228524,
 0.6765449709545316,0.6936749570932236,0.5029242900119618,0.7452681345573543,
 0.5864311252386649,0.7717710750597805,0.14661739198623758,
 0.34769875140543116,0.3638305412794187,6.846493590700886e-4,
 0.7128791951178796,0.4826576105581477,0.23690989980828825,0.6182722662723702,
 0.6645354033845171,0.5278128733354546,0.18911662040740918,
 0.23323598072893514,0.9927281307273534,0.6464368310861184,0.9304285666914582,
 0.35442135538727404,0.9921127653961921,0.6937457106649525,0.7237069035476935,
 0.5046089558804134,0.31145794562306994,0.4060730613610436,0.6309631874147128,
 0.0771230936195284,0.5952053780743964,0.611324447022544,0.331182872296907,
 0.4657122333078826,0.3286855086382674,0.9979056681246052,0.3895944521970104,
 0.20375296607343107,0.5160530544376397,0.5666723730188741,
 0.21113521361260723,0.17364201536023316,0.26648467046092605,
 0.21352493317073296,0.28552914638742055,0.628650065662109,
 0.012330209048618479,0.9904236183545738,0.7932536974600263,
 0.7114356300440375,0.7872204260979572,0.8044419235919634,0.8089396046804151,
 0.034504793604349215,0.10761062205719618,0.881425740423241,
 0.05915335511893982,0.3337143014967128,0.9124465365875386,0.3753754342430887,
 0.7118280936310064,0.30143291486872714,0.08456116183319129,
 0.9910372134762692,0.8682796335013185,0.2005497729373069,0.09816493255046765,
 0.5242288153630714,0.019529740091334902,0.3060598597522898,
 0.32713538689143606,0.17386949993575995,0.6661643549560453,
 0.48630429412831755,0.7012777153447077,0.663855343106309,0.12508170659205353,
 0.939372222746973,0.39602494792046317,0.1150578641769806,0.4584792866113875,
 0.4213468837959007,0.06793753974132011,0.658667869799588,0.36998667821811504,
 0.9092701377498875,0.13645405744277705,0.3924297279288068,
 0.005042609989545532,0.3034537681208568,0.08986934753412101,
 0.40649961187009365,0.5568917824504911,0.40110557736061503,
 0.12857409749804216,0.5370163998950257,0.44495858816962097,
 0.4443550268046147,0.594630125961784,0.7372857892270704,0.4010428473996872,
 0.7250266786497311,0.39894784560615515,0.7295197409869068,0.7539894878110766,
 0.5165515843194535,0.14976048582208545,0.7858696100608683,
 0.20227550591861565,0.682283705977568,0.3463526799427872,0.03191020460200511,
 0.3621712326857296,0.46963900922335966,0.976471375793542,0.8631871936188498,
 0.8681422365287432,0.6450770497156469,0.07485148764147453,
 0.06137144655664306,0.5204213447015691,0.2920467106788971,0.2277947732693163,
 0.09192191700945473,0.826364998768587,0.40423065268824643,0.917446993240138,
 0.48462760064721944,0.7103959454759389,0.04951327698244479,
 0.012975638485477115,0.9750527305842284,0.601642565104431,0.965005740921701,
 0.1783402867261954,0.1559579384176869,0.9355588878519101,0.4907289665238106,
 0.9185516775053394,0.028719755763859545,0.1882202184848668,
 0.8690990423548389,0.509999678323571,0.042754541723429185,0.46184634871536,
 0.7402132753171414,0.10939375114604222,0.5079795337328818,0.6210232442160553,
 0.8318570946290629,0.9089760399247125,0.8093393131860018,0.615210340285522,
 0.9406393545774037,0.6861330330070317,0.24721434511158613,
 0.24398712043462434,0.7763244173415956,0.7610735495915811,0.571886601886057,
 0.8209286361145999,0.2584304436732693,0.8610892494697965,0.9828142820747227,
 0.463992066118482,0.5909921687703588,0.6072645550682871,0.42188394439352794,
 0.39067196876652865,0.6845891007857989,0.699772180857641,0.6601660229668831,
 0.1784847592660448,0.668287568443067,0.4510603373395502,0.7237210861680992,
 0.5821330782428167,0.3128859117569931,0.8953048256258648,0.6991443847462132,
 0.31073691387752644,0.17144221141214844,0.38718101445422515,
 0.9529341113821708,0.09157184998469359,0.19328808397217523,
 0.6460028293113392,0.7115582913626273,0.3494138906761529,0.6794103311231643,
 0.24356235612200272,0.875817460616376,0.8808846209976675,0.6106341225041565,
 0.41742620083436754,0.524810675418263,0.627655689907871,0.3337443276452714,
 0.5145564986813123,0.47642478979007863,0.09994085192573798,
 0.4720324799887081,0.6112235595239455,0.49446611330875645,0.8266549924230002,
 0.841365841026201,0.3148406388498538,0.7006050735925525,0.26762960961769666,
 0.022573418764006092,2.47083477222132e-4,0.8003174063861258,
 0.3152026019408536,0.749337593441036,0.2571393287912308,0.9755546337929203,
 0.6482265220690159,0.06640280500854667,0.34212865791883307,
 0.3066354024846818,0.20514888713754476,0.4441091703954585,0.1742267503933015,
 0.5631339788273401,0.07128477548048562,0.9402525594483485,
 0.12685470696399126,0.8216529857604975,0.8983404937916617,0.4013035401495242,
 0.8542952088670928,0.98847399594084,0.09960000289900017,0.885869486539722,
 0.11614011563688553,0.6114376189396493,0.049918792777041254,
 0.625568970442012,0.06044525209555951,0.9265290023645045,0.826738456495966,
 0.9761483652989766,0.7199529785584596,0.4345305216704942,0.7692530146857652,
 0.12078584051125663,0.17395887615567474,0.8140564906949738,
 0.4623202423196797,0.38826669517457446,0.9119432581779203,
 0.19291399301757595,0.30257243266717704,0.613395116288273,0.6463405349187376,
 0.8559488801952355,0.23793941809586905,0.46607823388336567,
 0.8787161760390975,0.9410973831356056,0.8974015528958339,0.4318160580913367,
 0.5772978408875417,0.33877929231364423,0.5476803876109856,
 0.22295723868814354,0.9744115811270538,0.43036675381123146,
 0.8406610038297386,0.2335884588538466,0.21190386612205436,0.9833257193806743,
 0.6377429160374855,0.8078242693253814,0.6266457262240719,0.16200634995705665,
 0.7724628413698548,0.3046620479811488,0.04248172420641738,
 0.29650814285440874,0.3267215053716044,0.8667876141734198,0.2360057526878666,
 0.30399288820145765,0.27480371469788034,0.23927678424602505,
 0.30741443354354525,0.24173497585034454,0.5370239598930384,
 0.01831141345107634,0.6836951326261422,0.8661752060881192,
 0.17540466888374606,0.5840030759021186,0.8179437615932201,
 0.04460954809828077,0.3275621699083957,0.4155841348286834,0.7883036991552489,
 0.8934830734384517,0.9957864973201682,0.8094385416489155,0.9449482597435647,
 0.5947533429783713,0.03999458585195348,0.47856503204073475,
 0.12098505820519112,0.7110441458753246,0.8712879766847719,0.5480534608251583,
 0.624243491642733,0.5509749570945403,0.17326779758163469,0.9987992733536104,
 0.1950728323399753,0.7142058284381404,0.9719338672612382,0.12799838551064335,
 0.2131791578415645,0.9544502565624895,0.052324292706569686,
 0.3124991851082384,0.7818157272907145,0.8218652229528461,0.28010587497862305,
 0.42941167388674706,0.3348365931840338,0.7644008377439288,0.462175585673429,
 0.4625741647471512,0.7289169494521228,0.6036904204505753,0.9805267430578003,
 0.7870196456850665,0.40570051309500155,0.27883004580189985,0.579100129417392,
 0.19212657060403437,0.653253549845108,0.43915495141708727,
 0.19440535023334715,0.2582414700851964,0.7497581851763967,0.9631425295494574,
 0.8544768851024744,0.3004123000124941,0.2662612300443612,0.6793979969333351,
 0.9514209255354088,0.6417068519740109,0.3732343329292064,0.9715803821212521,
 0.5766522327313215,0.6087808011200667,0.978403315065644,0.03242021892054903,
 0.028806649995164246,0.7400641085782316,0.39801149930285484,
 0.7518280590945821,0.3932881067422984,0.7052563151345426,0.15972363079387053,
 0.81370062145623,0.1972650655916287,0.47167496150276966,0.3339170945805341,
 0.010087639945258742,0.380200002942203,0.8424588114303663,
 0.30475627567308206,0.3534638714256708,0.09037128276613204,
 0.24604861808141365,0.3441916787643644,0.4812685822976426,0.1010353613789522,
 0.6786833788734183,0.005050778097978048,0.028379334515694743,
 0.42306709161231304,0.8050698196701389,0.82767077951193,0.011142065413356272,
 0.13357811021837596,0.781062322019253,0.3282291801660615,0.6874044339207425,
 0.2677230665341036,0.4070413879395727,0.017520240711578983,
 0.05163704485515641,0.5861695456187427,0.8318805695586691,0.5163196379946364,
 0.04174249898862337,0.22011379243869855,0.8229131849170772,
 0.6390481491727366,0.39487042907526015,0.4680391056247045,0.8212212346891139,
 0.37359569145228144,0.9124013569990076,0.9333116012977988,0.740155410127668,
 0.5737307190190505,0.10117778720401638,0.5045109620721858,0.6897085305779924,
 0.08706882329515686,0.5758343979225747,0.21761314921207608,
 0.18135888234671438,0.36467445262992493,0.3833486040239804,
 0.8452448950549104,0.24946981692367542,0.9245009471669711,0.2949678248974088,
 0.1993679423325,0.18753949685898996,0.14148124311169352,0.5423997498620476,
 0.33783125654396684,0.12946836857142163,0.25688427628022337,
 0.7620574615816424,0.6050215666994053,0.3358438678436454,0.22777435646570077,
 0.1719578452695133,0.36667361880820226,0.9586338118010324,0.6832236305919483,
 0.2640169160170164,0.5683406530965331,0.39965464130186845,0.8601145089377491,
 0.7247420578821453,0.7365736644952868,0.3634972352117979,0.33290441448075747,
 0.17168017283599213,0.07636865843254859,0.6065455112335865,
 0.5130729357002446,0.4910274930474806,0.004825379659273765,
 0.11948915942339156,0.6533242412875586,0.562600834934165,0.4280904180928312,
 0.4277853378172134,0.530085741873758,0.8338628857209062,0.756684843593806,
 0.2422655433749753,0.17092913222221773,0.01159154819133601,
 0.1709585471898175,0.08881595429000322,0.08143142981913587,
 0.06446193513175258,0.7518500997482926,0.5710808068754074,0.6226127254429246,
 0.22265100453702957,0.21664455148394812,0.24240528743794587,
 0.8613097539838827,0.8999555202356868,0.2673349844662325,0.8111702622000602,
 0.15898412572147969,0.9492403779035621,0.6446775745250931,0.2858534505967707,
 0.2607611619999306,0.36704325720318387,0.0759962842131624,
 0.06637884110275638,0.48843558628879225,0.22581928733092826,
 0.6598060139286074,0.9193091295735889,0.2535101129833225,0.829042640976951,
 0.5627896703833877,0.7297247349725255,0.8269967973375254,0.8815456683137403,
 0.4521662471135951,0.19319123837604013,0.737673463392826,0.6112308320774629,
 0.7072506139589751,0.5712124158535165,0.2959234913728428,0.5575174343689068,
 0.2826591233277773,0.7092748629055845,0.6330441686535013,0.6705091815999441,
 0.37443365003004003,0.015850805549350033,0.4142696206405072,
 0.8538536867946025,0.9895696261777086,0.7986301126210199,0.8314275157171815,
 0.2216906489027416,0.9374821065078157,0.05648333405914996,0.8295865166817199,
 0.5714326359257071,0.37024723825609895,0.6803812904542667,0.4019185213400779,
 0.05415247765480635,0.9788485928835279,0.39730617002932245,
 0.5615796957132129,0.035812747694102476,0.6808427905550282,
 0.02079725518079434,0.5565672234864065,0.7820530021754436,0.4415144275841961,
 0.9540689455116551,0.5885905017312091,0.07978953848875847,
 0.050768698713380966,0.953657904582951,0.8405242496498877,0.9578426678215848,
 0.8183855361495265,0.11901573806860144,0.3756796159158645,
 0.36586384520732396,0.9122669037153778,0.5735711999974219,0.72031936475093,
 0.5318375415459096,0.9153314601976952,0.8090940059914619,0.2971368941717921,
 0.24217310661987002,0.4023516048829653,0.2234107517627658,0.935311974250588,
 0.41252180292823726,0.08134098707830228,0.14964043422653017,
 0.8598263897226596,0.027955938289568993,0.1081396280072866,
 0.7755754389705749,0.9482616042198868,0.11602736816037162,0.7290231515527419,
 0.7605797318689231,0.05585401526458755,0.4384789185012513,0.7308620374747848,
 0.1320188537457233,0.4642247891472865,0.09029986200641016,0.9394642611636297,
 0.772280213898052,0.603010886146468,0.9819879544000198,0.5108923197505608,
 0.5993042733017102,0.9548214577957419,0.10901210987584886,0.3328437497368786,
 0.3961590098247182,0.14900647722478988,0.559927735113215,0.27648891709593637,
 0.908405795975406,0.11506790819904822,0.1764523133708913,0.09266393559179875,
 0.038213249600467014,0.4574743766600389,0.6598302885228932,
 0.3976577662753492,0.9573448583966525,0.7486109360868167,0.5617417692130408,
 0.3375381045052256,0.07470065966869044,0.3279353289668999,0.8836892800461149,
 0.6338772175292544,0.5470623020698828,0.5964176038550723,0.16558043558654179,
 0.06336384634695957,0.8291693736436159,0.6554378433102247,
 0.44885204637042275,0.20994914612151172,0.3351901720424353,
 0.20394020662100087,0.724166252630425,0.8864815979069468,0.8560890503780468,
 0.9681808083265723,0.6464903557310616,0.647627453279591,0.8347749904642311,
 0.11240268231281203,0.9429836661153874,0.79200246157858,0.6069553504349307,
 0.3402446759386071,0.6336433091122464,0.11873763426558426,
 0.12557404021267649,6.0918523397957e-5,0.13100799016245246,0.888723751619861,
 0.7535622902270676,0.8081116482112641,0.009889161532074975,
 0.6349044777705352,0.527801247766767,0.21640887304825818,0.808231524311139,
 0.8910094896163576,0.21783992487503712,0.3573535714390691,0.7371487493410849,
 0.986813956546897,0.36902747588566975,0.24371373282456155,0.4436277756238669,
 0.5223228969207364,0.4576402562602616,0.22090414409356618,
 0.042919239412496646,0.4732838994339966,0.4619034461962418,
 0.11269671742130705,0.5123160301279492,0.38863971441917744,
 0.3670238623935803,0.004421662139294846,0.9546133412341861,
 0.2665607178316698,0.9771776020746425,0.5630612703214113,0.4680113826458775,
 0.4147206391159417,0.10157195690720788,0.6491551124768637,
 0.12173398531383117,0.6688813128682501,0.03263749132792837,
 0.8960207405012032,0.9015707042830019,0.4397289020368449,0.17931021870936492,
 0.3799193053846994,0.5700634445360623,0.08543512259339603,0.6720907542018575,
 0.4264821365566164,0.006934154449425134,0.09216482317811048,
 0.3626815442516076,0.6026094011423009,0.20273713636817647,0.654026570031686,
 0.08724354114111976,0.7140816291057674,0.896538828313417,0.5956478943722923,
 0.3964565195991554,0.3864220789709867,0.7145206996038334,0.30410822900309153,
 0.8362122367090817,0.19850813966452385,0.6848856507866492,
 0.23780555275513937,0.6121742996268486,0.2804026906869317,
 0.26629686870060654,0.8304744574546656,0.24217010625873714,0.616297545972502,
 0.2968157539490759,0.42961887101151996,0.181804404556547,0.857724629035075,
 0.24518096941912593,0.7507995576517764,0.1158044902003097,0.5579202167791801,
 0.6145138020881605,0.6620076831811472,0.2291966992327905,0.9712519806757459,
 0.5364284389751874,0.19187013224559868,0.16694123641251069,
 0.1286946052240463,0.7536966159043061,0.027263683887229773,
 0.2265932990004551,0.20259671259558054,0.8555278549357255,0.3889973956304411,
 0.1290546112474238,0.014322176811404796,0.8364066232390245,0.880494786826213,
 0.45383041540981295,0.5752421916578059,0.38794358695841713,
 0.33735644484967575,0.8419735320984596,0.06089398734812357,
 0.03472840263543808,0.3995735097439146,0.553494398909824,0.158694278660222,
 0.48693662035628205,0.24508093763859895,0.20198811783924464,
 0.9193863111097524,0.873097705126723,0.8522298741188198,0.32772663713766526,
 0.2031014032359354,0.8228536602780819,0.8754295129199127,0.11061669207511393,
 0.865999706143608,0.5625759574183989,0.0018078640064730367,
 0.6404772584136074,0.7625022545034816,0.8180782216415179,0.1407472498721507,
 0.5625509533082486,0.9386597066924018,0.9434351082241441,0.03971520161188202,
 0.495186487440868,0.5201252920255788,0.32039126111778027,0.382768556426595,
 0.6974715044713686,0.9739806549639354,0.35522088859112655,0.7537259008824105,
 0.9250799278310486,0.23875529721736888,0.24319191288383357,
 0.13718810628980138,0.07888137150314267,0.3370023349798903,
 0.5036365267713687,0.35170410994970114,0.02583100874156341,
 0.9311806156516059,0.9700201418993615,0.024844290653839862,
 0.6354591778652915,0.41762899790411767,0.8061849703532833,
 0.25948435089702315,0.7573970543508616,0.3340303404282188,0.5823165021083682,
 0.753247676738312,0.5484438465249293,0.23321417620360874,0.08087295771799552,
 0.04837395438754921,0.45149070925129,0.7948261501729357,0.8264637660683414,
 0.30580354309520064,0.09018506140286953,0.028896512113767647,
 0.0569296484357521,0.7682004405556022,0.36568056670707827,0.5760755035830205,
 0.056703741235594074,0.9001145059699999,0.3755045479970188,
 0.5090096325862228,0.8972502815281748,0.6857255513032346,0.5542198625172607,
 0.13532805152430694,0.928127040071472,0.8057868495126561,
 0.008394516565339516,0.575876661736574,0.5673109025199842,0.1441818515747696,
 0.048726932450814275,0.8635949792246365,0.49318496571445014,
 0.252341558360492,0.6623651364341678,0.8514449804187773,0.35980331898069984,
 0.6203340937470581,0.5965802359790066,0.39688963319502435,0.339566819713253,
 0.8467795148313789,0.011897300369641206,0.8818661303123322,
 0.6452297474818719,0.1624782272902885,0.5181100754420724,
 0.002943782575117737,0.5165249622563919,0.3172851375817448,
 0.5495277526354752,0.6206315263615686,0.6956685790313432,0.08105238465312481,
 0.8290338211075085,0.4110782497780061,0.8110063429837954,
 2.349292657264357e-4,0.14744410394718432,0.2855390708347635,
 0.08607121339604185,0.16566425381657401,0.067695893097246,0.3333505908592237,
 0.04224041698564107,0.9792275097432185,0.44077455971537904,
 0.12537021701568118,0.33547588698585207,0.37237934897927916,
 0.2699308339680401,0.07514670493389952,0.8988965783725581,0.4699335527153766,
 0.13527595125370573,0.5205289023393128,0.6621745555578606,
 0.21017535156495426,0.7712619021885203,0.059966847718830985,
 0.4602238126593221,0.714774332284691,0.6212432932925454,0.5215364994090106,
 0.5519737331525907,0.12464353448428422,0.5992636444201263,0.4993696708062003,
 0.0772779833571896,0.25926707391219295,0.4696139842352576,0.4772486013333074,
 0.9284084362269184,0.9366307541658513,0.6078800854500067,0.7162967963452588,
 0.6578883914606956,0.037122362900365524,0.3815984130394984,
 0.2685390165754864,0.5110138303831909,0.5764147137653003,0.558797639712165,
 0.6014608553163421,0.14781925606426094,0.7831649015750913,
 0.16961043803522446,0.3601464151274807,0.3941440644351373,0.4204088328995562,
 0.3080607752303258,0.5929035928468894,0.32626954648036877,0.1891736560942351,
 0.7378145150857905,0.21327367786630946,0.4398581475812453,
 0.09722318570647825,0.5110155207799456,0.22155007028855833,
 0.6178554783658525,0.7672043322586722,0.481233045733801,0.44925979474215383,
 0.5053856253104532,0.5799159269110017,0.9171351559384234,0.21653375864890112,
 0.2407913477531064,0.7763121135882045,0.7799778179666509,0.1846738523521404,
 0.8140357027933757,0.5443397893575812,0.3309698732736901,0.08933967753781635,
 0.11626511646039828,0.19688358159604347,0.516594868909682,0.5022210520766299,
 0.20984766152551182,0.10867641489134494,0.30365202049689866,
 0.056874553744012735,0.7049647464231124,0.29525307777545706,
 0.5803887498860105,0.8733952557101333,0.5444050590259045,0.11780728739782786,
 0.6284421823664295,0.12482890900709132,0.412545937410427,0.21548588326740437,
 0.37701696899488124,0.2838682122204902,0.22438551818900787,
 0.5588952243690449,0.5034160192174048,0.7170215360717732,0.31453508241319916,
 0.58335603473292,0.1540810064502971,0.08099835057467952,0.7980264684845416,
 0.9308978983167917,0.8145589359559126,0.5573400146669676,0.876509001931094,
 0.9716892801417224,0.5748090081819645,0.7864500742891786,0.979429147380063,
 0.994786798363607,0.8502901708732999,0.3162184111648767,0.3509478549948237,
 0.20207400872666403,0.383766303715418,0.034725130836004925,
 0.0217688023948458,0.009441024017826094,0.32459011072202315,
 0.6770091478758669,0.6239973885794342,0.5380675762313063,0.586353465746815,
 0.8998722248871043,0.2100189931612041,0.09188557705087852,0.9993730298250525,
 0.8080406128681396,0.45421664886671465,0.02518966622048846,
 0.08602900072320097,0.7111646290119555,0.9530492413522168,0.7101964551036484,
 0.2409948427258245,0.45366125746726504,0.8536091103090796,0.7272557296266905,
 0.33234414096920695,0.08201971832514443,0.2243785880263679,
 0.9623136279366016,0.002142480321852158,0.5336762925896785,
 0.43885329945417284,0.04735728115013027,0.47411604295351006,
 0.046186659094731186,0.7661547472756974,0.0536938527787818,
 0.5716336865556861,0.4705185133878974,0.9456909188318088,0.4076930981222089,
 0.461640258899622,0.8369373003265711,0.5585534376448709,0.7878927482178204,
 0.2364910145880572,0.008318368244202734,0.9371246751121662,
 0.8021196285852292,0.3861661953047738,0.4777762921599056,
 0.028020634752922824,0.7048713177713369,0.9327386149950677,
 0.3706055992901357,0.7622670186623247,0.28422992822504567,0.8320340235491175,
 0.4356751377858177,0.9057728405122807,0.05531219228881401,0.9773307433724329,
 0.3598605537800381,0.8621553641324063,0.5802151523945716,0.6611240218590417,
 0.8169369410346959,0.8158005799597935,0.8792008970635231,0.5082929149814149,
 0.12228777596217832,0.12483308540574911,0.5169935649739961,
 0.7869473713138039,0.753132935756166,0.6571256456479372,0.8461114796434349,
 0.6439453170249738,0.9237867248673298,0.6864562499442812,0.9362335426978273,
 0.38961798927149216,0.6640520838708233,0.45731944043123507,
 0.8135535782682431,0.933855453392035,0.09442821206298668,0.21264183931420888,
 0.10015062961926402,0.6775742782810982,0.3898770695313143,0.6720487351176114,
 0.28741615945262167,0.007877974148911115,0.7981731183396973,
 0.7375505950468892,0.15371951797475392,0.7581124486883982,
 0.43579178787547834,0.900587291862362,0.9536925247011249,0.10753878290025576,
 0.3467447333432494,0.4008193502038131,0.3467591435131233,0.5206641035784937,
 0.8658087255351599,0.08202983185987289,0.3690055122930159,0.7145250560843702,
 0.47718629253762357,0.8271206339296893,0.4663997321122999,0.3030341492316584,
 0.5779358260347042,0.6974784437573401,0.7024715555111998,0.9683966271487042,
 0.7726716323278644,0.6533693315724557,0.8871318471826072,0.29533824719040014,
 0.39888550259483013,0.20938759910729898,0.12729562367330471,
 0.8355429840759528,0.3369864221967467,0.6668884796225558,0.2708036820357802,
 0.07907646118228973,0.8085076259109624,0.1720647584815722,0.8827715126277299,
 0.4846803238919488,0.7228394569582489,0.08508629498916798,
 0.33305593568799197,0.060937894736387554,0.06545276485671137,
 0.28835312050426076,0.49941109120902893,0.777611279514477,0.7656847254310476,
 0.22563040474792206,0.9567755432315448,0.41161258770319387,
 0.22770392260368144,0.26405902442081797,0.49577821658938903,
 0.7967136287323415,0.4240399285414602,0.871117665993567,0.53195694395407,
 0.6365604178402454,0.7391994588740004,0.5147842145495751,0.7357771254136226,
 0.8731396688109275,0.22695951042669527,0.2373244164912678,0.6396470579999527,
 0.9439911167592205,0.7562862184509747,0.9136261388984137,0.47008295812210754,
 0.5903089277098832,0.6801604699035726,0.7837390336921997,0.9667930842402385,
 0.46308430723949356,0.8353273257110284,0.334969526087448,0.2746108612583411,
 0.525430420674285,0.9937476621289317,0.5491793891565782,0.30580109733102,
 0.9749335835319831,0.8772745795814584,0.752248597991379,0.862097129859121,
 0.20361323651227314,0.8902045260965282,0.8981560431152278,0.5549641124476362,
 0.19984028984611524,0.5395397545519356,0.15020356982260252,
 0.12811455130217353,0.46198134324333384,0.5462019439851927,
 0.5374665356374275,0.7991936331825582,0.7113146369816665,0.232514933408851,
 0.48087229820500776,0.08276887673691258,0.9968481018366364,
 0.20652009101999325,0.14171320458571235,0.9021852480822627,
 0.03643093394872732,0.39073445244158556,0.04005650080871881,
 0.3000259394268676,0.08076510859501385,0.9194235960247823,
 0.21618078624974824,0.7737050894825649,0.22315375990426112,
 0.40665825199184635,0.8201825631320157,0.17617387279855468,0.55367326882251,
 0.3471468215518957,0.7515769515943326,0.9600124696748654,0.7685947014385569,
 0.8199078710277679,0.07580385293862546,0.23671946724336823,
 0.6296355674415388,0.5596215687542252,0.728931585102754,0.8919584886128228,
 0.22899888865516094,0.10767044562429662,0.6413227042040007,
 0.5713391816586967,0.02939028746020378,0.9886261170885136,0.1336117464974962,
 0.9405967458367077,0.8482137629165432,0.08242281090137815,0.3166094028062677,
 0.5301194198740369,0.4210287773610225,0.7419996712522999,0.8773218805500753,
 0.416841420083922,0.7242854178422846,0.930825427341906,0.9683498948199534,
 0.9347654705665078,0.125042888086632,0.09786415186615971,0.8598551211360925,
 0.42676378113265034,0.7485234766400378,0.33761933514969744,
 0.14513286229727584,0.7432232485334113,0.4374133466908682,
 0.011713210176043765,0.8718756132977276,0.20432450592060558,
 0.23562064276477723,0.9425698872540959,0.3820673861317545,
 0.41002144699385534,0.5782016397264326,0.1971497084684064,0.8692676741162586,
 0.774888242098604,0.2828151060866644,0.01395988012701499,0.698785838688289,
 0.36480322597261294,0.8434492470751769,0.2087244298206632,0.3042533634980866,
 6.485911437230651e-4,0.16731943945828087,0.022246366655381045,
 0.5104836518327185,0.014934556326722137,0.02202477466648145,
 0.37902875862050345,0.8536795425213485,0.3837248885838034,
 0.28922546837997454,0.407387721006056,0.2415971719276926,0.7975365117989859,
 0.2937063634198318,0.7548272329357962,0.25954569680796136,
 0.18505236980484674,0.2156029273162725,0.2936657738717824,
 0.018338873825929913,0.5266358558365019,0.41203912294191836,
 0.36079066214778666,0.7572769036877799,0.5816939163356797,0.8503038245075116,
 0.609133603596836,0.92844940635332,0.2790116806932087,0.31690200819615144,
 0.2339129863737257,0.4068357673402814,0.16994245982976197,0.7257658666410101,
 0.8568707260590747,0.34660325685541205,0.1866243898159582,
 0.08716535366178957,0.15936433010034934,0.5052069652986566,
 0.3664524192321771,0.8219146266295786,0.38384837256595594,0.2174872909861858,
 0.7685292551367159,0.9413861814714304,0.504560618181169,0.9876643083919945,
 0.4153727187780163,0.08436982611839328,0.5119848771790764,0.2729094942377893,
 0.4669866216076536,0.5826272402950228,0.2761568568943581,0.5333822904380869,
 0.15509995985596658,0.6975363800516402,0.9436842415463166,
 0.32904898611766653,0.6417915198434905,0.11849601539965615,
 0.24724571465294654,0.10116304590816638,0.7879989594497232,
 0.5089493354646201,0.42025345833673666,0.3581872762033963,0.2943126096673171,
 0.03740814854636265,0.005966917297037355,0.43253946976398283,
 0.5883558155324174,0.7381034669029036,0.6021300474998074,0.9712620899396064,
 0.4691924028103698,0.48741848386689834,0.6313104339570169,
 0.17615226429569386,0.2894926250708689,0.04209580506997168,
 0.1645502928711856,0.8821775657342993,0.8571350574096995,0.6811836196689685,
 0.7608988547127793,0.841117171995601,0.30102720582858533,0.510238562104824,
 0.45751647156601183,0.32972680178521774,0.29873222756401985,
 0.349378973561256,0.06865337136325411,0.8117314054872763,0.54538722470623,
 0.34201607398122547,0.9405370356454393,0.5385363216724921,
 0.45063173936104384,0.9282718086689123,0.4417915249933295,0.7480714537948793,
 0.6135300987835566,0.5626919032009403,0.7006820276465582,0.22114260122224783,
 0.16982011564465116,0.27283977809059856,0.5236139201899948,
 0.7914549392738827,0.8916054263186245,0.2530321459476855,0.42213719886657963,
 0.08929680935605233,0.8775734351970477,0.6628011825652906,0.5276239982237505,
 0.2732242220265596,0.5420055185308899,0.1981769963691571,0.5342107515363171,
 0.4257063547896349,0.6045033322203659,0.927872323316811,0.6475839639110785,
 0.7249154505775907,0.37777999730486933,0.28474134938543083,0.512053821434111,
 0.32134063288757697,0.857007004955527,0.893659512880295,0.6889960276098684,
 0.018688132521797907,0.9933499342737022,0.2516185692106544,
 0.7411670597149682,0.36114506728423235,0.022944039287777107,
 0.802175419381522,0.359439674748242,0.886048645318792,0.7673027728711939,
 0.5562974513304667,0.18016989144628281,0.30150233455324393,0.2629649723006,
 0.9262801604919935,0.5554358687259973,0.7740629392461595,0.7914511667682383,
 0.9799095529992015,0.24679124821169163,0.2774456922725924,
 0.09622162226049535,0.9134250122039664,0.12672388584732197,
 0.44442213693440824,0.9855395583996219,0.2839173616246976,
 0.28064757657496575,0.4871965840000362,0.24355443720818692,
 0.3218653463893214,0.3918284684411426,0.38169856169647787,0.9574368934470089,
 0.4626654740180456,0.28053657712104685,0.8457379202903014,
 0.03339841928760956,0.05766370339175142,0.506686551667659,
 0.13976065090403123,0.05030768654558038,0.26876607399985264,
 0.6133884654699372,0.10551123470231571,0.6051883573606992,
 0.08956076595534712,0.40246105089793405,0.7898041205344464,
 0.3355808197632646,0.10586668657020948,0.7073043200135855,0.9716782178208434,
 0.9334544069807796,0.31020758119458425,0.3795245467963676,0.3828427750271366,
 0.27908481561335796,0.2811218564074959,0.33604366268169295,
 0.23526269434483527,0.40021601379705873,0.10996131962482703,
 0.17134751912784674,0.31402909771762366,0.8386868554867968,
 0.8005093454049781,0.750391274728783,0.2999920640660946,0.42658490525091397,
 0.42397729168871356,0.9682539292581022,0.46407487349627496,
 0.27344296827447345,0.7559007300437619,0.27402755568576365,
 0.46691012400442156,0.2897401032504703,0.5578390496232492,0.1842859182762968,
 0.6894989231694656,0.989427574625029,0.055521158828650274,0.2861108957741635,
 0.17075163641439528,0.5048662146257464,0.3336695066497407,
 0.37828740797577853,0.9775045712922321,0.7606546481050106,0.912664579636762,
 0.4325304444869821,0.5537829206995167,0.6537545726679215,0.18176642884421979,
 0.48311202600741054,0.26726609455371075,0.3248815971552048,
 0.5726158936577554,0.9840991062428255,0.18138786682900232,0.3160495116888944,
 0.864719951907631,0.9326401773769496,0.6747005206513275,0.007590031630372085,
 0.49411319613784965,0.45158849736859297,0.44154968262488403,
 0.6678341293796043,0.039001271006426586,0.9718925115560987,
 0.3454513618766406,0.4496365930770423,0.1398526373252239,0.2753147201468622,
 0.6589314453384323,0.5179692557152065,0.7034638216892778,0.4220419531611638,
 0.6019385314263814,0.2808920078805881,0.2557641788984377,0.72039866561569,
 0.8192822832182387,0.8602312102292596,0.3548277657699296,0.9261280148593163,
 0.9432979987349948,0.8049690711666687,0.2468220254808745,0.8823724364196093,
 0.45510987764794686,0.1463815914216734,0.016997689000062155,
 0.8518290205390422,0.5342290813341545,0.5226226160595261,0.29261754178899424,
 0.8539605214284363,0.4215135988351958,0.19688170946200667,0.9179022519394068,
 0.6442385759365103,0.7431224911422708,0.15839551593864765,0.3584636540675873,
 0.8939304985715294,0.6933163386800267,0.6456744857736826,0.2632383724769899,
 0.6940209722619725,0.653583279229373,0.6899145056317522,0.9745513464539369,
 0.0508325481825872,0.7816412328285366,0.34000464963838906,0.528071727634809,
 0.6037613394438801,0.20470726339640666,0.7712022926290543,
 0.07713856893669069,0.5659380599306897,0.5169075673362719,0.6370005124018994,
 0.8390254079375654,0.8924903017137321,0.8490107057075525,0.9454086452740738,
 0.7494490094395895,0.7558222228650274,0.9628568364626171,0.12357324232314693,
 0.23493436187010275,0.6781211262173501,0.5833186187129532,0.5077684184278488,
 0.8754543788195149,0.3839620066364723,0.042283828585018,0.4996400925069805,
 0.0671463629673481,0.5075280081530807,0.886752365359032,0.715270884345826,
 0.6391889655308985,0.7349702064126397,0.38436357756793404,0.6193484805608022,
 0.9842098278240105,0.4194207696481681,0.6556231139467215,0.28177514364638256,
 0.3552199778954396,0.7597875966039837,0.4105454441811581,0.3189730610005991,
 0.34434677364642363,0.7093917829838707,0.9634793730235757,0.5884930165102955,
 0.12822429620016718,0.46239611974029393,0.19993475397097793,
 0.8572693085268837,0.26791970804716336,0.46273772028283955,
 0.04882694524478515,0.13706756793636155,0.15708777213379332,
 0.9124614909425409,0.905997166231598,0.6794093737061471,0.3845395941979841,
 0.6809772943501067,0.7501091643852359,0.05645269843243916,0.5172403460924416,
 0.11573309254200015,0.7270637096049101,0.8628595663955677,
 0.44828437001291455,0.7012896677735889,0.41590472142340074,
 0.43909915862177906,0.03440196092264125,0.3892999991988202,
 0.17425028209426507,0.07234474858644235,0.22572084400325032,
 0.003537887400013151,0.4008902594581013,0.31546598767681033,
 0.8358450842748827,0.21205056879542972,0.7058825165815334,0.8344709011645327,
 0.8288823913232989,0.4364025601170275,0.6174662678526024,0.05267415445172774,
 0.08596919617243348,0.5252426208363379,0.888296576574353,0.41110385802303384,
 0.8406189857421826,0.6649215390303389,0.7554849977026268,0.2490350724039322,
 0.18960913322892858,0.29663166394370344,0.7380357053869914,
 0.7812501469989325,0.4459918269534877,0.6790163788817164,0.6389133218585725,
 0.9405014003144705,0.43013881019588474,0.4134314368643358,
 0.09480459518312612,0.16866176395726318,0.2524554044826389,
 0.7623516915017068,0.32965336413799196,0.9057451125792036,0.914136106300233,
 0.21722849754742635,0.8996705601382752,0.4809587252829971,0.8910820886427854,
 0.624139540626887,0.21039260219849376,0.4490812960676942,0.55798504656059,
 0.8139591042256423,0.7490967222279351,0.3465483416491473,0.11407671291359645,
 0.4734156979808888,0.6397333901502679,0.6140132867070038,0.5285936483014015,
 0.6484985998732435,0.8881710623012382,0.11908882735712667,
 0.04114954562922146,0.6650219119318652,0.6204500348088077,0.3289395347494858,
 0.27825023369449264,0.1953250805786647,0.11609812896406502,0.422937623410376,
 0.40847507897333346,0.9354423938606506,0.5151505600833375,
 0.06370006038774312,0.30311972876188276,0.935707365973533,
 0.43961961481471046,0.8968512897813735,0.29440572715499924,
 0.6952439748485099,0.383296966836834,0.9694256203843473,0.26547675301458507,
 0.48498297597791884,0.0609548094040836,0.771861318067803,0.23910119205903146,
 0.4038164486075275,0.18645156955119435,0.9356915663449883,
 0.00509997708547949,0.22636575259849678,0.8228969778796087,0.766004545881613,
 0.661856993572266,0.9876956413999107,0.9861128453025901,0.2830699349857282,
 0.6974797029762002,0.27754217470797315,0.39006161441541776,
 0.5999210417888244,0.3469248072226261,0.13352500268700718,0.5415078613172838,
 0.012857530987074206,0.5881623482561894,0.7153978052054878,
 0.8815125042316913,0.36725381753419917,0.3736234518963125,0.7619261761709091,
 0.9959184756462169,0.207284980082597,0.3608151623548439,0.12685403748073976,
 0.28727936812942123,0.19813767522671022,0.7968552536330349,
 0.7864388592844984,0.2538386549843472,0.180042754101139,0.8820017081485,
 0.9750973291746835,0.09663906985932436,0.7717436569776757,0.599881526715884,
 0.7171896397808285,0.8866396830312817,0.3524674527082362,0.48990255778920866,
 0.3673041375709266,0.975615215849841,0.7023835807927046,0.06859617043705424,
 0.4868456191830798,0.7078245422059847,0.4627658670929302,0.30056613814553834,
 0.6497672310634278,0.3680633160479754,0.7204905342979143,0.02378524526247272,
 0.8191196197414625,0.1296326991647868,0.8050285719101872,
 0.0015369599396163336,0.9501822295150052,0.8263380311215175,
 0.6316420037336327,0.5630038072799893,0.7810616844624145,0.15449345307237372,
 0.6995961107934573,0.8778801394308834,0.61046789832845,0.9242953129890994,
 0.2512024731958621,0.6704911526977458,0.49417391159202095,0.3377756416397746,
 0.7623359128269767,0.021589698275582436,0.6827021002331655,
 0.7311669396685692,0.561846359213013,0.8760345654679584,0.7937196755573809,
 0.47626542966539553,0.15713471140652702,0.7385207157099581,
 0.9504741998163146,0.5685918530409373,0.14861548063870889,
 0.04952079983818858,0.3728082572774305,0.9275104064449237,
 0.11302772483376666,0.10198562479738182,0.7672504440173796,
 0.6880909861333805,0.175120841326189,0.5310733079379447,0.6912511470648538,
 0.7992480436167754,0.5882228195528199,0.4628606599970465,0.11501412372550601,
 0.8457600756389904,0.5596793698897935,0.36221894351275974,
 0.14864617060074226,0.8297018945935279,0.9028786995524191,
 0.43559725304238583,0.8006904289250548,0.7370458633495661,0.7718500582272406,
 0.1663440936650794,0.1197930292427446,0.28377170342915836,0.3482747278882504,
 0.3982982877743946,0.09162916586726477,0.14473102798260884,
 0.41109573087925444,0.16240799564080421,0.4874780683911957,
 0.28278231698279366,0.3672721468399598,0.7759581001492877,0.7660911879065541,
 0.7448089207090507,0.7596913074540044,0.298466625711423,0.2935573517634962,
 0.5325651112400462,0.19727023667962373,0.21568682106772152,
 0.14069000446428592,0.054059340518916255,0.9110238029096863,
 0.9655458256258832,0.6290395035820184,0.5549406369213261,0.8589081130779433,
 0.9650128541976073,0.1379177714397961,0.7704327112374763,0.1095349966191499,
 0.6518622080237201,0.13398028907736392,0.6987268068011124,0.9618648688342025,
 0.7354935047893945,0.38239196284300525,0.8833254087758159,0.8654420102834972,
 0.4575797992165843,0.8505155556089278,0.3110566663224278,0.4801138288933783,
 0.7318607189671508,0.9446620699639303,0.3085689130395136,0.7949640524833602,
 0.20288317993678107,0.7468202946923973,0.15326979042286726,0.860252608905522,
 0.386426701248897,0.9819464699300544,0.712554145622888,0.05809424436819988,
 0.7898275268482342,0.17230626380233738,0.4538840600238514,0.9412213254952698,
 0.8671387839248843,0.02270587467547258,0.44840278558553714,
 0.011457220756434627,0.7080310082615004,0.7017873745682526,
 0.8933911583796122,0.6317423312978633,0.5703537852178895,0.6579581315320812,
 0.8421440892501965,0.2804100432835479,0.7833735720044783,0.3997460727741592,
 0.053333421946447856,0.03584127205497456,0.06953262657491177,
 0.66155966004141,0.952980844817596,0.9273029988761902,0.2747799148770078,
 0.8915851674212758,0.6341084760351666,0.49016097305047346,0.8287022209717374,
 0.5709260545210247,0.390409178890628,0.7757430627847544,0.7397619890410815,
 0.5687073519570721,0.7275873271498516,0.11338638259311873,0.442871141187807,
 0.30253399638080347,0.9270630103027051,0.9044674233543392,0.3893150171981321,
 0.10303328421775748,0.8673962730972323,0.466576460347512,0.9884110779582368,
 0.7130703878596689,0.6441038948699174,0.02305607547756583,0.2515525242526887,
 0.9224645778669098,0.026683782137746248,0.6430930073422861,
 0.7126331482426213,0.808430189895093,0.812980519460625,0.06959711512816968,
 0.32165343182010553,0.9001297400386672,0.2511284955665216,0.2518801727533525,
 0.5408743010184636,0.07740447679356599,0.7315099946288904,0.9236116944706002,
 0.17841884042138723,0.11167828934401147,0.17571487273360498,
 0.4789072856418699,0.5408805356711602,0.07499370639500735,0.4840354592805075,
 0.6410611189831128,0.9259287277947739,0.4713969602288549,0.12502504765622924,
 0.19253198945930983,0.7616350487801539,0.3898028512174908,
 0.08288456267720523,0.42690818732514635,0.4050949722553784,
 0.9013363265445379,0.8355959955817043,0.0739319053055093,0.2960581969051085,
 0.3152490816479879,0.6388632935616255,0.8270502855725465,0.4264283100874552,
 0.6031340177458488,0.8072608310642714,0.38722873369260846,0.9049354855881078,
 0.5106752698573804,0.8586284639620287,0.900333665914981,0.6063207772158947,
 0.8603109049022031,0.7502188162152658,0.9811213338008171,0.508442745198599,
 0.32466652181628053,0.03436756933437779,0.30464338187535567,
 0.35729005604735176,0.029560940246338263,0.6108363503513453,
 0.3038595054245774,0.3955443831512746,0.9967263282532539,0.3431039533362815,
 0.4228396894849128,0.9326219216975656,0.7865994544381067,0.7252290726078255,
 0.1901986339341144,0.2840318837174731,0.9674715266534525,0.16359110282997547,
 0.20241532616369384,0.9710888214131586,0.13786701321381556,0.777795193392706,
 0.3046366621201969,0.3575364210293225,0.9994811213111482,0.5492785154466406,
 0.18946394421512924,0.5163776263790437,0.9765188237287113,
 0.36129239770532173,0.5568020725563868,0.6830979638551884,0.7195471860903048,
 0.899218800293383,0.9109265290259166,0.8957501081105852,0.7818354111136615,
 0.24357838022568445,0.3458740231834867,0.6726119249923311,
 0.22736248896060351,0.43288069206286917,0.30990310207007443,
 0.07962472783248065,0.49869349053294665,0.9619040927496474,
 0.2622378944493515,0.6876143889659079,0.6526434886048328,0.2677380368286508,
 0.41935840065023244,0.6090372653936809,0.8993417423007833,0.231507077652215,
 0.2312257539588397,0.4438256055278984,0.8958319281177473,0.5363935982508327,
 0.22061190037226885,0.7633449623405625,0.35953336696834426,
 0.9455354867762943,0.09612402550838717,0.5393748454389462,0.3471389158910073,
 0.8000534800540962,0.7708995607063938,0.9103203149452056,0.4316231657199383,
 0.4458524021120437,0.1462463175018719,0.1834221101915191,0.8267874216197172,
 0.5669041228812667,0.37188752644397727,0.6760981831079829,
 0.28720284854989075,0.6354898294172076,0.9844994582995813,0.9251658758469089,
 0.456397283461759,0.3909401664514818,0.49539602319728193,0.4680885382315655,
 0.6963529217253969,0.3541172964646103,0.5975414533753917,0.01166858303306173,
 0.9506904423216739,0.5556431241778749,0.4293608855722659,0.08682694765384658,
 0.8682411381694497,0.5355892015900836,0.7757928989279629,0.38986995102426103,
 0.9912121711291363,0.07505618766010425,0.8125923470413733,0.6770962050682386,
 0.590105631002031,0.5198354012080153,0.5718113689046794,0.1102607838613453,
 0.12897687189248763,0.9492977306750168,0.7524413437975925,0.2732588020105996,
 0.7734503961460296,0.18696383444474307,0.12122977744030061,
 0.7646686414403634,0.38715094521399174,0.06321746132056161,
 0.16316637599056472,0.5954964666304396,0.46065307370061404,
 0.4921794631729215,0.1166268772807697,0.38710343601875397,
 0.45867661994750786,0.24856872689996423,0.9084314516448321,0.806591555851516,
 0.4131680513359419,0.8045583199123856,0.534503550641779,0.769357292607737,
 0.42775203712286713,0.846438994627598,0.5129114431938757,0.1787720574767243,
 0.6154149787872102,0.4186133048667092,0.5303084540819509,0.6063340884298587,
 0.9882104339075344,0.5357894576315256,0.9731760156566704,0.8293808262919147,
 0.37157287914581016,0.761771759171358,0.7988459646649314,0.8823778981629321,
 0.3779377508930266,0.45153242577767494,0.773018011647564,0.546180532430006,
 0.024055665940134086,0.4616083907143741,0.12003882750779549,
 0.6644319687966769,0.773618108571794,0.7681167879124551,0.2191689330025659,
 0.3487906169007716,0.48591005704978685,0.6297345958667613,0.8799972014530588,
 0.9773101779348798,0.2562151979863745,0.8240841403482018,0.5170039513633767,
 0.1096636352016197,0.5180690938489114,0.8735032907316596,0.957827868735639,
 0.5174301859819018,0.6335348430177309,0.7673268438536407,0.2874351646823675,
 0.37116902549266617,0.5639409661597439,0.4528925548306476,0.9608472542492905,
 0.9520835674192851,0.351896919437486,0.43195474270700873,
 0.028703139154666157,0.3351858277152294,0.2774296511643688,
 0.6162540373902475,0.07422851031927213,0.2124933985875631,
 0.14145344648045743,0.24124686129646689,0.8396700796400207,
 0.7689396679023671,0.8416273784387072,0.38787497416094696,0.1485702182644767,
 0.8893715734036161,0.6452163362441684,0.27996211575056096,0.7361102608391092,
 0.511289682553956,0.283580300772278,0.29365278114280857,0.5249032009837828,
 0.6767822259013704,0.8215674978059172,0.8116447021269337,0.24485531047277465,
 0.549646788351359,0.7677970744517565,0.4344248998252903,0.4183272604018401,
 0.14857366722804055,0.9075755251961797,0.9710365879456838,
 0.41248133240190377,0.7508845104289251,0.6238105102293084,
 0.07792804692138355,0.7119736015823055,0.9596360735550896,
 0.24929466557391777,0.6447852031943377,0.6502534859049851,0.8714098277443312,
 0.3260152212418015,0.24485328930425776,0.7970982900009668,
 0.045290326656936886,0.5319747260967089,0.14551472639352003,
 0.13227268759136812,0.4048158937425286,0.243069853016313,0.5969504137081064,
 0.6737607997131586,0.24314303468602483,0.9478673506349169,0.6586346612363707,
 0.4174102815788352,0.16949952219002695,0.4779974179767824,
 0.33832250758545557,0.9075642398205928,0.041754256104603926,
 0.9896676075751334,0.010370092701529776,0.8218889163645167,
 0.6321317890969294,0.6071212596389258,0.9847012557960595,0.8532990269655096,
 0.18258122750057704,0.6008827549025,0.8906966310663292,0.8934869713256631,
 0.8395518481106874,0.9435527165844425,0.50748389899134,0.8149096615702905,
 0.6759368103137603,0.8855687014199348,0.37631983954259507,0.7016832480247792,
 0.8791544537037669,0.44365033191591896,0.5470987146188145,0.8671016604782357,
 0.4926491569154179,0.9211694157246066,0.35776689316839005,
 0.010179254422986506,0.1390611781881037,0.2683856539748457,
 0.016633323098829367,0.18813012118603767,0.7265940385765437,
 0.08227239844345957,0.6042371355731705,0.8317853040034845,0.8941629270754113,
 0.8144019420626759,0.21967573181925182,0.33265684018198016,
 0.4765604196600236,0.571458976352194,0.6699333312943179,0.9902296115361795,
 0.6931568690203056,0.7170706431030168,0.6230636620982819,0.3181781046463734,
 0.24819620211043192,0.2382532233191248,0.3004472763896243,0.9818043626242948,
 0.6106081869297793,0.5839931139263104,0.3795978174820789,0.4105403097248147,
 0.32246652171813617,0.7124354856232754,0.3739959347931483,
 0.20019175018821445,0.5981912263895911,0.347408222682256,0.6876982935120527,
 0.0828377325816454,0.5615952709443411,0.7705993710121015,0.9135351883869867,
 0.5893000925429498,0.33279922339292023,0.967838133563944,0.18513765787598335,
 0.08277004875071026,0.2247773148748483,0.08728492712713554,
 0.36248327330926233,0.02903072525548933,0.4257703822210811,
 0.8561798841007082,0.5267282817193135,0.13498027785841626,
 0.29366286414540466,0.141872696655579,0.48485524126487356,
 0.29732754014115115,0.1592365939826097,0.033795640667916516,
 0.07880083206789512,0.5387000750243336,0.6843730851062341,0.8161398642673671,
 0.46707200853815556,0.3667703816154345,0.14497358761372792,0.32022502658632,
 0.750978520442577,0.7580700976659648,0.39722928582933026,0.28851409356873503,
 0.2775907704157774,0.555122203155037,0.283931440600934,0.27780189792232646,
 0.2480638019790875,0.3107917225960607,0.10880043691872665,
 0.07471602914013453,0.14956548413486115,0.5787414782182212,
 0.7403660466481757,0.22006247713613447,0.7632140436595223,0.6266467970351303,
 0.5349095182403666,0.1256203803677418,0.79688474080425,0.22024124992247407,
 0.08811940199098556,0.9608536447374746,0.38350604371575503,
 0.5771713245405228,0.14172171394641586,0.584460005791066,0.8545698068378212,
 0.9850769301400243,0.3157242510069498,0.6692187186186369,0.708319765538341,
 0.3231471275375918,0.4759185269904461,0.7743926734291963,0.5028518503897575,
 0.0962361199617413,0.44779120130477623,0.41711622502834134,
 0.4657350215883448,0.6397993182822477,0.5260689985058805,0.9382023082301253,
 0.6443328576145366,0.31743132825530096,0.20440113841760843,
 0.3694275128359372,0.5873675489238713,0.7481413366271014,
 0.016570954812483718,0.09684827515853667,0.8706929985019621,
 0.920720095753604,0.21314866081613448,0.24507162843650887,
 0.46827106149640507,0.5259251853382919,0.47302036554058857,
 0.7383140345808147,0.7194844694991782,0.11256701273395742,
 0.46355131688182705,0.7504547643379262,0.9187183071085109,0.8533910229590081,
 0.10715049880508487,0.04367024516673246,0.08725155668207718,
 0.7087130592534729,0.3836013722975483,0.9633499241168657,
 0.032442682971478454,0.4851704648744709,0.2656941869113286,
 0.7188345892396367,0.4078568665759331,0.05926181116037055,0.5016514796066883,
 0.9643169208395362,0.06435677008358842,0.7485432804332403,0.2953907197641539,
 0.18742775402492085,0.875545417884572,0.5784433606857923,0.5367728638864329,
 0.7646176905568404,0.6049195534363462,0.34140411057858633,0.6231228174869374,
 0.8112489948525353,0.21293319096886298,0.735811818132978,0.7098858944865015,
 0.15133631435580353,0.3754357492010053,0.2605673816323787,
 0.031517824718331644,0.08678961776428906,0.11666950095312989,
 0.6242807091207444,0.36454371679952435,0.03868459936089441,
 0.2575755600256212,0.5854583902557355,0.09595355993630772,0.9939219974276415,
 0.4887985387448962,0.8408868480783859,0.3677871030752815,0.09906689332097063,
 0.06935215638464565,0.7424926181481897,0.2646536643301649,
 0.007133419985386635,0.7523634350826613,0.1412664433400771,
 0.6711728020984226,0.28894917880436166,0.2178101693564639,
 0.39111793396004146,0.7481520254762685,0.009983290715974702,
 0.7942762641987624,0.5827790462775572,0.7363788598393762,0.7262370209241193,
 0.6809659937009488,0.49638594569164485,0.7313257587041989,0.5711481881503658,
 0.695758343056506,0.4941358401796856,0.25667098496596963,0.19814673437740438,
 0.9841021886041528,0.6734790700264407,0.18733795810081577,0.5276905700373897,
 0.25293348461421783,0.23298687389758932,0.9318666816064912,
 0.23000782394939412,0.3371373322060385,0.3590174919944126,
 0.09165682703093503,0.033892802194183336,0.29895093411170304,
 0.6400025753718285,0.12228969755070773,0.3586062432360255,
 0.46046433397187014,0.7174035054438024,0.6462706124063051,0.6801611309973875,
 0.4574952600093286,0.7969997076420171,0.3974104469102181,0.6732797169735308,
 0.569869805586366,0.5482937785928337,0.6860909336585361,0.8617297711765608,
 0.6599155425961892,0.2730467595392494,0.7434055000746584,0.6590861803064385,
 0.9066957394550565,0.9547905973694326,0.27009158238304554,
 0.03155422942902342,0.4799797351049532,0.023504403815093222,
 0.03952298086734696,0.8057727132287404,0.42094809934549793,
 0.2020254406180465,0.8586581593363469,0.7829016035859724,0.8012781322317458,
 0.07351176957248351,0.9580373435780171,0.47247433911254866,
 0.37578562702850393,0.8554612450467279,0.19251482101456774,
 0.22597784161821455,0.9787962197621046,0.545772231805338,0.1253160263825024,
 0.6390564471690537,0.2287622832793318,0.6998719600903534,0.6165697472718952,
 0.5854054076952692,0.3259056324013704,0.6864484059017597,0.3115848368711467,
 0.006107043764865039,0.9941843370770925,0.7080595569393723,
 0.5274480330645731,0.11237325640496698,0.7924140253453159,0.4486966520157688,
 0.6478687081486768,0.25392469604904655,0.6749387210303497,
 0.18630428641787733,0.15886983815294664,0.846150742631658,
 0.15430929766130008,0.9351501486946984,0.8709542281218121,0.1942050919071875,
 0.7267148930686634,0.1960475984836857,0.9944139491656733,0.1946203378054443,
 0.11841340527919597,0.41851161520309177,0.904161670798866,
 0.24555952949144977,0.10415601671732944,0.3730687606899149,
 0.23245748122313836,0.672165425579855,0.7048616740830307,0.5028269404031727,
 0.8115974361699401,0.911312606051972,0.04926149287653703,0.6268207758107013,
 0.1858903776236317,0.7564159371269927,0.12273071152472259,
 0.13222402930231103,0.6169351196645174,0.38952818820255053,
 0.2510103376470558,0.691923192242172,0.6429098137317385,0.7496775408821665,
 0.6355551842451275,0.32425038707968334,0.2508063376097531,0.9948972066333105,
 0.024258172191460314,0.730584587069033,0.9871054719070281,0.9012858623703764,
 0.3391820950194767,0.22320914131033787,0.4732663093110989,
 0.12236137915689049,0.0820908211051602,0.3416066629167086,
 0.49304329937251123,0.035595418234975806,0.5626940515775471,
 0.41335257048310914,0.1893263571470407,0.03543480892116557,
 0.7201423110655918,0.8307774196114117,0.16416963846687016,
 0.02307017722954019,0.7416263549731656,0.9809646876622548,
 0.36604411376115453,0.24287058831841368,0.05518460206026854,
 0.6719498269686679,0.37748078456363265,0.17514762682026574,
 0.23676004818291707,0.40907865741819527,0.5072907835746954,
 0.36000830490047475,0.6494316229917044,0.4375104104753831,0.9060445136193659,
 0.6648311952527909,0.6678547742259876,0.1952000168713217,0.04824733837093331,
 0.1804902869733146,0.6936506321421501,0.6790246639701943,0.6336209760136293,
 0.9471016702151045,0.007072524634916921,0.5157081794684599,
 0.028369114120728733,0.11100284181190978,0.2861431138096101,
 0.30847174464015237,0.13527531484439737,0.30842875340994247,
 0.17176755858958714,0.4746171162142728,0.5644762332805839,0.2072384519470366,
 0.811350144253264,0.5758118368992763,0.27200752258179606,0.35314352142395944,
 0.951854721929864,0.5391248152296644,0.6781897624327665,0.564722966203549,
 0.6131409518795816,0.8029139413002759,0.9046588788134512,0.5319882342685043,
 0.6511646349697805,0.7165496734937626,0.6155863235742014,0.8138236738349961,
 0.26932784089333617,0.11260147587823832,0.288010131184823,0.3496534043920697,
 0.2530086009025644,0.17857313400107744,0.7426768882075298,
 0.37046051743133157,0.4138255924910974,0.8022977876778264,0.7603817957566126,
 0.3828930481800672,0.7041924082620549,0.327858826070351,0.14282456882036176,
 0.6044966447494956,0.823884892482378,0.03443326647838374,0.3596735468009268,
 0.6047072026245537,0.16930857682699096,0.9303123801832417,0.9034044407860996,
 0.64181562698883,0.02091109366419075,0.6501675990493366,0.43354565958187097,
 0.7353538322736989,0.007339507515242261,0.7164978538000984,
 0.8631912049225132,0.44681001081066274,0.6122708241844003,0.6180965305083692,
 0.18086327628614052,0.5411070292224358,0.5313265102880202,0.3922298773606724,
 0.04639669976952876,0.40089736020814704,0.5210400963741332,
 0.9079236700553462,0.1270973390921275,0.996958312432393,0.25856318535925316,
 0.9539628982122099,0.8644133721593775,0.30425725166429995,
 0.24758116807322306,0.5077917532658013,0.2661394101387652,0.5977413856943635,
 0.7391540266185725,0.3385684303503428,0.6448884973614412,0.4002743982506448,
 0.57786144597194,0.9996031367609403,0.9711245827440047,0.7289803199419305,
 0.9981930372750332,0.9407893762336579,0.45004322087212034,0.7423628563198988,
 0.2976354427705695,0.9542303186970515,0.13942781284214878,0.2853784235027501,
 0.8063221046531158,0.9394218842233583,0.6230029647963908,
 0.040947466767563556,0.12545954162305017,0.41799403050227746,
 0.07207822902193484,0.8160350431686743,0.8099457343561383,0.598598059516613,
 0.8064269979128266,0.3015656278182798,0.7306054474055478,0.3217880161061719,
 0.8520015683702294,0.7683737502981912,0.4704466614177548,0.08278706831008242,
 0.2841636854012545,0.8499550924121809,0.8873264061956277,0.6671354613681834,
 0.6531457310151179,0.6569989440507026,0.8245232104667783,0.02026482245026051,
 0.505575118176097,0.1824854376866082,0.20148581268380317,0.3057673561219503,
 0.6764009415489985,0.18919408787840097,0.5746904622654836,
 0.21010134806823821,0.7111900597731698,0.06011360485847095,
 0.9384810147962568,0.5945835312489818,0.13825116846501906,0.5629456063067353,
 0.4327395015163643,0.5192206868817758,0.7928173454946165,0.835103839515323,
 0.0578134461144959,0.8306386026492618,0.13122140308034713,0.6551953695462933,
 0.7502409791881923,0.862810633317014,0.34817819717626375,0.7534818863768075,
 0.6593847150457982,0.3928026035267017,0.7349611532732718,0.19335117553720416,
 0.5144556745343115,0.7655388658407467,0.8528405881391191,
 0.013700259137440773,0.515509658478804,0.22538221800266756,
 0.8318689334716578,0.7802841301961244,0.0353428485961913,0.5612032643039393,
 0.36310581299152833,0.37101699773065744,0.8738208768508055,
 0.3273395809260765,0.014426722394638603,0.527238378814168,0.6918667949433115,
 0.34133841462918646,0.5551416038183543,0.9148215527174228,0.3203014803637758,
 0.3044451253294441,0.443350263776634,0.6512483140756644,0.9828142862129816,
 0.6123921534294864,0.2681866856082785,0.5670831711292089,0.452635553434912,
 0.4891469913232672,0.18037643968095796,0.2510974323664873,0.2634865717992926,
 0.74903159518956,0.2949671126934662,0.07338430307578125,0.06522324095293497,
 0.4784628013906277,0.7502529428267625,0.4897261298921378,0.5544285243908232,
 0.545045112440894,0.171536393074754,0.7360486385257593,0.8859253189376081,
 0.7213024595844513,0.6657241803438886,0.9217937639601268,0.7674914050839067,
 0.6279712427569548,0.5204096097043628,0.5567623678163319,0.7024730932107205,
 0.43780802018605813,0.4162686162778959,0.429188958102397,0.0902665217435612,
 0.9381681105158217,0.3604942728786833,0.12505113056483808,0.7106927379057961,
 0.009795330393192025,0.031052986635703328,0.4270790691561466,
 0.5476596876161116,0.3607693241218687,0.6948752101665444,0.5657349778653444,
 0.6849389046778683,0.7635779261359145,0.044746161747137636,
 0.10226236716039983,0.638554077985928,0.6924264317738303,0.4420051992328815,
 0.8424663961739444,0.9376186581863242,0.7720313809761412,0.8895290253767653,
 0.44647451441049657,0.30878654902026037,0.49934359013943286,
 0.4624245448330846,0.9127677279220592,0.6046003535049763,0.39092078587010604,
 0.993852979012283,0.435206358721432,0.06884984873715472,0.43619270104905516,
 0.7871245033058772,0.2800257505055619,0.1097641185424777,0.9573583262852648,
 0.9289080094963548,0.2736366113607869,0.9781469072304791,0.964670569728161,
 0.3746362989517229,0.4194968685080184,0.9986576633808595,0.4068577912222665,
 0.9664744159320728,0.37065031779319324,0.13480726014688493,
 0.08487321844875118,0.33822407801917176,0.8021944005234278,0.504802396223657,
 0.9907354964664553,0.2736767993333482,0.038922696054945805,
 0.5208684665887741,0.14184403861442596,0.5623151358016304,0.8016161258654282,
 0.7337587333298503,0.19808264283145116,0.11715966154299828,
 0.4541431730901173,0.9519598357183878,0.9198852314955751,0.36536516076368486,
 0.45888413527694594,0.3228647961069959,0.6449844478488675,0.8858117820074104,
 0.15794352138560785,0.25438359546004774,0.26132028318151634,
 0.9387391423298935,0.04815731305900661,0.10091439368421284,
 0.6699223478005526,0.4322273192354815,0.3339593281335328,0.04099393277242047,
 0.48745600780892306,0.5706306512580416,0.5439854096847072,
 0.46171508629149194,0.734606303360663,0.5215578906793281,0.8075273379400276,
 0.27040624030963945,0.9285969598414958,0.16450250121815368,
 0.08799033004028955,0.5929529283331847,0.03430255715108821,0.967409295479945,
 0.29035393184709324,0.1563744314821025,0.36475204614950907,
 0.9943388371180781,0.032569612911440005,0.08798450328019236,
 0.08429345451061054,0.9164140966779974,0.5189986907929917,0.9283173766338944,
 0.14875937453493826,0.283350694302118,0.4623297378675786,0.3684734995599539,
 0.6400390534026164,0.6549200144872435,0.36215146410841936,
 0.19860651011316854,0.9460951134750201,0.6460894771459238,
 0.11381298819114372,0.44172872719401723,0.2609355942083649,
 0.18809708686578708,0.424303928007559,0.6806858750536242,0.14762041561420736,
 0.06608274842764028,0.48740370643515196,0.7502739849651126,
 0.2617034988426368,0.1612984786869075,0.4861917713077768,0.2384473278219248,
 0.17386377303961054,0.09650321426078823,0.940581369855798,0.4200395871935386,
 0.3802671196057741,0.5201950681050551,0.39837240241828986,0.8050007477763533,
 0.6744963685241412,0.6662053119709699,0.7304133072631614,0.04711853594680859,
 0.6794859837338301,0.6276275324198471,0.015760120140903555,
 0.038255482439429556,0.429348527194202,0.8729950766361757,0.7167271874816208,
 0.45453311417630327,0.8028978321772597,0.9323988138642887,0.5616809842330112,
 0.08932943306190144,0.3126712818672187,0.5751922369457361,0.2214200330928675,
 0.8782642287969636,0.11030927184916517,0.6304701380264677,0.9769196418860815,
 0.5640892649860948,0.2925492585103312,0.8551530322990482,0.6812876403017925,
 0.0051606713295279105,0.4436099872761422,0.21021057039385926,
 0.001674854502818679,0.1252694847768373,0.1381819860866237,
 0.32713476908729633,0.7894123584635002,0.6615859477437214,
 0.06363221932324692,0.44075515250340225,0.6691887752065667,
 0.6240114904637866,0.0592364002920609,0.4912662441087312,0.4767570593009899,
 0.7925224569241532,0.19516129166118013,0.7299044942489792,0.4885999809908702,
 0.7602546647750846,0.7347508858858244,0.43875720409858,0.6596472046803032,
 0.7875488276077436,0.11807832448376587,0.689765092696105,
 0.009151724979473475,0.8720378748225383,0.4903520720557638,
 0.21736989437327425,6.455314984810112e-4,0.46671239587502034,
 0.26330083100977697,0.8835801594820704,0.9081069554369517,0.9685173598075916,
 0.9124568070938504,0.2101078143679257,0.27871295813683594,0.6938203110493074,
 0.5719240596545702,0.533560746809708,0.20172612388369748,0.9130832409321581,
 0.6009053942566849,0.7904530579778458,0.7211783708310224,0.10093214056951783,
 0.986329376069014,0.39299733261007763,0.16120458239360114,0.7693717097830118,
 0.40254314027218574,0.3755290502653603,0.9133656960470702,0.5173804560132484,
 0.24943429837740405,0.8668253114572518,0.8566009803214435,0.1398393364597068,
 0.07870692432309456,0.27562962502729815,0.3053656969091878,
 0.9965630281510329,0.6718521416955479,0.5631346885672182,0.20853159511534447,
 0.5944905266219171,0.31043406912987814,0.9896029339932936,0.3587421250743823,
 0.36886501869332733,0.9933406322017084,0.8947475245603183,0.7088339556316889,
 0.7196515947516963,0.3632805485644214,0.1503530728411604,0.43701332628229306,
 0.4631202459940936,0.5796349603612271,0.5792643452500149,0.21043317897575164,
 0.04505208818036044,0.5829834828197997,0.8268194079625478,
 0.16429696547410455,0.3181282636118774,0.37577022242752556,
 0.8882515319435714,0.43825596494511,0.4452757484033758,0.6739613462303082,
 0.9621839168032302,0.9937937909229427,0.23927972844471523,0.7711282405650336,
 0.7498752974824089,0.7955994886294869,0.6837732251127097,0.3190803139351487,
 0.3358153302343858,0.3364503009679515,0.4347708907423261,0.6094887622502496,
 0.555962815220561,0.11926085879189874,0.5451940019108011,0.9639451890197592,
 0.6986769669040882,0.5351428957335052,0.021059074276257417,
 0.9379046689264862,0.6196898390716542,0.14417870028594593,0.4207326222730201,
 0.6018470784643621,0.4710271161202103,0.39271764017526956,0.6521929405267608,
 0.7171605468648017,0.39195933479743805,0.40226473328348167,
 0.8965374590349529,0.0013135967220375688,0.15280286228602402,
 0.13969006644246396,0.541436323224215,0.4321057416227658,0.4259287986801671,
 0.16157377167774345,0.4975577089234615,0.8144628211229501,0.7140806337766338,
 0.22491237264705255,0.10570992999196416,0.9969362085991169,
 0.3960507514569025,0.44499535236557053,0.628449844817991,0.3053402450935828,
 0.41683700753875663,0.014522464533232915,0.9026125311866183,
 0.828211383121487,0.8065043897676677,0.300196977647594,0.8376661166667896,
 0.5301325940001028,0.4820664094042397,0.3735441364019839,0.04861682400757139,
 0.8851293772273041,0.3317498784321913,0.14978177093639766,
 0.20756676540361307,0.34380480086310483,0.9705124954166425,
 0.9762896319223421,0.1540461786424916,0.6367106164740937,0.6130309568231149,
 0.1968424326764523,0.6365101472968846,0.1280090566309804,0.585633841744818,
 0.8335799983515213,0.2752964091807478,0.39748065274688626,0.5628254062143433,
 0.6583187771901832,0.23506374720151202,0.691207489674513,0.8964044752916084,
 0.23534892748391922,0.38660505340700213,0.8051055149131351,
 0.17312000740536748,0.1758160586142845,0.9343141186485455,0.8715068646034266,
 0.9716716050101972,0.2631769338234524,0.800124892355804,0.7422971797598809,
 0.0630139161793346,0.06743616440754685,0.10509194921371112,
 0.047523065998457015,0.684763760683909,0.09316850264965326,
 0.8995150512741341,0.05792651352245637,0.098202858702516,0.948058269138523,
 0.35414228866577235,0.05409580114677459,0.32669710017404696,
 0.16551313759646635,0.3030475842912118,0.07031123193390809,
 0.11622046193446045,0.43560222181414465,0.6867146070447291,
 0.5400993076434861,0.13493984304261164,0.44648933978024674,
 0.0748037564612406,0.40481040963283155,0.09180701476667841,
 0.06451414929755894,0.1065633168731831,0.9547997281424442,
 0.01738921550880912,0.7877747120866958,0.23225226337802818,
 0.8226370402940132,0.9884578947909909,0.5806994458325223,0.4693767661734931,
 0.46851198120118054,0.8518849615649706,0.6571164140850676,
 0.18450618744654057,0.7037555542246798,0.04966332112021177,
 0.018477988086824437,0.8389861224320894,0.9301971059502885,
 0.46065477822571244,0.912794487428892,0.5651059129972263,0.6052437248754442,
 0.9432082037552313,0.9239208825986562,0.6684015164017286,0.8792152393843192,
 0.19463954541565154,0.5542612001636975,0.3800349953612734,
 0.03012556068188732,0.7404868062869941,0.7630510413496981,0.7155410848441159,
 0.3724986040569803,0.721376239235662,0.5692529848355916,0.3928771067695741,
 0.4033363195600391,0.7444436858926727,0.4184826284054124,0.7986749229459918,
 0.726021304639803,0.2511916535944042,0.11467310171360356,0.31887463993206744,
 0.8453029479920258,0.7230487520564718,0.8983416643662495,0.9902881391780842,
 0.3403259568850354,0.866329738476625,0.902517132649501,0.09847559707083142,
 0.5085942036325957,0.7423048759186358,0.5464182804990649,0.49401591117271004,
 0.6337057517389194,0.6627267698350093,0.05757736148950621,0.8895180766423251,
 0.40443012302367853,0.9993278482644051,0.027279385618663166,
 0.20874423594703917,0.6122583917061117,0.059243328159754305,
 0.45490211557934335,0.14737254828262825,0.6819524641757184,
 0.7318815329954499,0.13920133383609679,0.8300243107558356,
 0.48194900334437063,0.9534435827594855,0.15885532638125854,
 0.9743221988990789,0.6036044568854579,0.06642472792082421,0.468685787470725,
 0.1547279947488127,0.6828879283720296,0.020393616751656296,
 0.8424751712510689,0.46240611847674384,0.13751889549320495,
 0.6208074815535567,0.7243114096040935,0.9367578873256377,0.11113782623326174,
 0.4052395717292201,0.5704499314455284,0.052137962175764385,
 0.8806399189802234,0.015216041541646685,0.4593305418821223,
 0.0657080383857338,0.20499657585386688,0.578424054485092,0.9811688262915073,
 0.059110329355806956,0.9022635445302547,0.010949799462771637,
 0.7548557065094745,0.7889974012165267,0.0690745808257125,0.36116312868463907,
 0.349631064590775,0.16378107416394627,0.34014115158128466,0.8956400457779523,
 0.5783547424917669,0.10000129611534181,0.012494057311686624,
 0.8347168194360743,0.5906800110438624,0.8434540437698105,0.8961225978305387,
 0.14497740574601958,0.7727167150322538,0.0355126129004828,
 0.19563585081141244,0.40195029415220307,0.38188526293676994,
 0.34957791590636744,0.4868787644616922,0.8151627190631764,0.1368272675654606,
 0.9080673694940228,0.9801504560411491,0.8128651746605573,0.9225769486278245,
 0.9886797953130518,0.5159067417422829,0.46009508610611083,
 0.09393607222670886,0.9031030766805481,0.47110121468533284,
 0.5663029876730656,0.11932971713135632,0.9180432037138807,0.496858996472872,
 0.7329766988652194,0.5369491174506681,0.7727798830812556,0.5644731439372479,
 0.9683327562237147,0.047493652695022925,0.814675539627714,0.4592023884741374,
 0.41609347158438725,0.6526866423293936,0.27339458769162317,
 0.6784029151455094,0.1462561754460957,0.8272233580742507,
 0.017160236205832402,0.2682601547628657,0.19433088274867494,
 0.9636081747581352,0.3296639909234742,0.0716892319353214,0.19102164938086008,
 0.8828227555363716,0.45215795539411374,0.938223869223248,0.9962769379583452,
 0.9101190372865916,0.18100037160483085,0.5570417219771812,0.8602171369942759,
 0.6842496064194983,0.8753241556936385,0.7999012900297287,0.15211590904039973,
 0.589474275896693,0.2054081881230433,0.19297018994432014,0.6421182222732112,
 0.34873933554666803,0.32022851949139397,0.3590772348858705,
 0.2038440932384772,0.3693219804328087,0.9120840508657356,0.7478900791558664,
 0.3618540959416302,0.39543067423918055,0.8975492125682707,0.9644556388428631,
 0.11225670197570392,0.717897349331908,0.7511633818102691,0.20047265436503958,
 0.27995370672831266,0.5375941792403005,0.9398094409668727,0.3641580765850232,
 0.4933825905132867,0.5101942992800897,0.4749001941380314,0.5197967736718615,
 0.14613334344380124,0.742483028779986,0.1404677005295295,0.40775818382175233,
 0.9477513330739644,0.23291656004129724,0.5702644158962394,0.5855629140571897,
 0.6879629486532708,0.7335901755931844,0.9369635702634156,0.5817059339919082,
 0.5267272769846789,0.8144812806417926,0.30720936681238886,0.8640149797804951,
 0.09043353931695886,0.75550165596349,0.14687445198717008,0.4020425994111825,
 0.9045642304246762,0.13643466852369235,0.43819048884488876,
 0.01568716475422416,0.4345836883788431,0.22849600349567067,
 0.11687925058176984,0.5664996480887401,0.6696260144645575,
 0.09873803327360098,0.1941755961992211,0.3821138127742302,0.5897003686000651,
 0.6199031723763659,0.13271716367460928,0.2935531341250953,0.8785614201746358,
 0.678635279619509,0.6064474499178627,0.3428916990175572,0.286390954095879,
 0.01738984121503251,0.5280904849935693,0.05008148316281702,
 0.9564126633563735,0.8070225875065725,0.6130979710635781,0.37172504313105903,
 0.09947029617217618,0.8638680895058262,0.5495958211623301,
 0.41628309663091956,0.5694936796632082,0.3652817004750488,
 0.05127818435895459,0.6787807611880569,0.44319951281462155,
 0.9630342703518179,0.816333029889593,0.17682202865480867,0.8866615885722632,
 0.19868480161110147,0.6688917646524919,0.3466565276220207,
 0.17988809375246095,0.7810040097177513,0.35458373600657767,
 0.17358860567952428,0.040700065669641905,0.06848718059717718,
 0.251515487196586,0.8128439033684156,0.5284783257825414,0.7274268751393833,
 0.7746592841516197,0.8220274109952225,0.06498632667720106,0.9465669703948849,
 0.4848289968188835,0.41360625875340984,0.1928743084392449,0.2508662825645325,
 0.014553935615547409,0.5809150432578931,0.18775634750778092,
 0.8616266202441187,0.12825567996887666,0.4890808922065698,0.8274546802823071,
 0.2001617019966878,0.4902507934164325,0.5234560331989699,0.04282068717546672,
 0.9030017995983411,0.8246426590463165,0.03165999477933534,0.6042221951817682,
 0.5985184105567947,0.9853217150806431,0.6334907061295973,0.09166801611947306,
 0.15369406572087096,0.42149955410519924,0.8975276413613203,
 0.43670894538438276,0.09586334509294914,0.9793561806623217,
 0.2439355493408163,0.7427552429221086,0.7479563883082971,0.48939758001302536,
 0.20494498993631016,0.323496711021645,0.978511995877745,0.023692656810937573,
 0.5868060197234203,0.13348250410763196,0.6845589139633141,
 0.31367014915924574,0.7736148631307854,0.7798550133378181,0.3008222661086387,
 0.18603476642250172,0.27299873937248065,0.33137392231443763,
 0.4044988790314581,0.31950185543815923,0.6754186296414441,0.9701064515026839,
 0.6038576958622364,0.9268730300880469,0.24465546511071257,0.3869289949789496,
 0.9121341100185526,0.7179141590924378,0.714952765323691,0.7572617512892538,
 0.17456029532134176,0.03691373073647619,0.3355318962310656,
 0.20947124360554614,0.7990904312259876,0.996842488527057,0.8856314929344644,
 0.3324264831074282,0.8157707920691203,0.5409150947798587,0.398035263501197,
 0.8056911230320861,0.3877426158482955,0.951551117144185,0.11248075227680909,
 0.7548150849953891,0.3832429208507101,0.5014290003710355,0.33214021485308887,
 0.16061919577609174,0.04152672744784924,0.5373947495279274,0.824992015028472,
 0.1667681662620173,0.6402978379831752,0.5734516736596682,0.10395792983739616,
 0.06574443763330495,0.7089960925551618,0.21995269645171822,
 0.2716384027689964,0.21690978045135312,0.9737571468015854,0.7051998427242108,
 0.43621291855660616,0.5021598063724486,0.7076115328020599,0.4877856839758724,
 0.8393355602941874,0.5397162805961656,0.4682203842707061,0.7878228468961649,
 0.3850107845831572,0.968879616353101,0.12685638570892155,0.7612840265115968,
 0.23440229192461703,0.3424692460826383,0.4938999525038137,0.6455908214519358,
 0.8149164921370035,0.17128619244820986,0.446092639112338,0.8362631113777295,
 0.013404670202000757,0.13328757595489438,0.1833988457116652,
 0.38282178223434027,0.4838662538573252,0.9926818685868289,
 0.18829329013114016,0.505916869600256,0.2060028577634696,0.35410036671615863,
 0.20720358662661575,0.12793893961861258,0.0562289395445813,
 0.7208585980529685,0.436808531164519,0.4093578348608925,0.3668405547844301,
 0.16980840027382005,0.28821328112815225,0.6502177033303096,
 0.3346373552438433,0.7468423044617188,0.1385020450587351,
 0.009492174276197085,0.3499565511442584,0.5552223785307743,0.88424285794315,
 0.9428772427220907,0.31622314961943865,0.943507081731581,0.45593371618245226,
 0.21932693758408506,0.8515668808432244,0.42017765353209025,
 0.4255202369169818,0.03200377217341077,0.5407070887732013,
 0.18357903319431101,0.014428478761025043,0.9366216628197745,
 0.6532363578926509,0.775051466305728,0.8757072600912347,0.041089274530969666,
 0.5292333385246184,0.8671771470901588,0.6326731257410768,0.8139338425626375,
 0.3138973006181497,0.5528132950228621,0.25290216681814304,
 0.20827061375409117,0.48497363812033,0.09088933874990579,0.4082132546161551,
 0.4574220960383377,0.1340370115131796,0.9940710341131529,
 0.026353847077608794,0.9133902558752731,0.7955028536427511,
 0.13169835191970558,0.5190430843339778,0.6765155148909607,
 0.12219601373747258,0.5654590327538023,0.4523441714049692,0.6072486179939102,
 0.893381556322909,0.2907177349940604,0.553382142883314,0.21161444875987012,
 0.21120266633661777,0.5062968770118217,0.5080353482658426,0.1368496249078126,
 0.8331067272824532,0.5218133570349822,0.9271249528450521,0.7323089097945847,
 0.7809957635033509,0.4901506591260153,0.1629315660559656,0.5251139536174937,
 0.39255110356453127,0.9611120772267979,0.8924159923820252,0.6195128087461282,
 0.41301115465287797,0.3686985221718123,0.09621619234173051,
 0.15688883360744832,0.3984852068725311,0.7912613089195506,0.3112940746662002,
 0.9703604130051035,0.9741888281087292,0.5472338379564109,0.5704095370425508,
 0.6926939326385915,0.9294995803990042,0.5907909774549431,0.25483661362631993,
 0.5725034286293464,0.8441585246744809,0.0890831198033778,0.12758033123872503,
 0.7597694808898154,0.036221979107788016,0.9215617518432312,
 0.7930856268044111,0.1760024972353792,0.6892029400309959,0.2372595971908803,
 0.8177322202140674,0.1432409561239345,0.18144962904267892,
 0.40514898174505054,0.2128684171850601,0.25421216804311264,
 0.7492766867046632,0.4346209334403923,0.12711710177412572,
 0.015850561763740223,0.19077946641303767,0.938817637749814,
 0.1647245588459968,0.23573617587489348,0.433987568154504,0.2354064877426323,
 0.9209619827503344,0.8125010194345064,0.060387175439341245,0.630559544797884,
 0.9041534792554035,0.8840230129901544,0.8204203774460856,0.8380719796399629,
 0.019646835763400317,0.6422299412288545,0.42885168264904694,
 0.7971513314698853,0.2649016699642137,0.1933748136908603,0.0704351651192675,
 0.12283633889852974,0.11404229234126406,0.7960387290411891,
 0.3479069524214087,0.8034632435370123,0.42860493347143025,0.0817001750632047,
 0.9657875201867748,0.11352283715405254,0.23237621661046226,
 0.4942213000978487,0.5230463309279408,0.7769243718392684,0.9161442183280674,
 0.8600739655412792,0.5973877445151835,0.27853095260396477,0.6032538046531569,
 0.15899182621809438,0.6618752842721186,0.6296820840559724,
 0.27337491156506166,0.9835698852208868,0.27553438016522225,
 0.4286112224529598,0.15289115156621125,0.480600947336457,0.42456556351359087,
 0.5670533756661379,0.9988495714704264,0.3883454633644753,0.11374485538195256,
 0.029240947546623097,0.727999272839999,0.16284551274095016,
 0.8679317040336025,0.2506302036368383,0.5640547960582907,0.9244253771159995,
 0.46647430546251556,0.34405534477390864,0.9702614722654788,0.870451670253926,
 0.08758044970766488,0.55550264579523,0.3345188584072518,0.6480168615225591,
 0.0949328591736569,0.19704615999409425,0.3383781966216137,0.7099214363806944,
 0.7798776016269577,0.9209565797743808,0.8481606295360239,0.6774876580912857,
 0.3862812870411342,0.4271444118937392,0.6204768982305977,0.22196469273904973,
 0.5928308994744557,0.48009631977210276,0.17243447527209466,
 0.7748370680400033,0.3521878259472365,0.3175067094900228,0.8297932671708442,
 0.8241793547531306,0.21320951513778208,0.46649779769568744,
 0.3229823510356786,0.33442433870072064,0.18778746081015796,
 0.5885178951160435,0.06408471677811389,0.5836168147127951,
 0.026381149433614803,0.8484112661584745,0.23424133505935807,
 0.6176447646310103,0.8943163081680663,0.3016692190206911,0.2234255539947576,
 0.8974859896068584,0.6504207423994391,0.5872348741241725,0.16472342701717257,
 0.9207348670728511,0.8672125695526477,0.1617335393688638,0.03470014819310707,
 0.5709568482295069,0.4503764170438662,0.019657900004976803,
 0.28030391979930314,0.662506119093824,0.47399301415786765,0.6663714897827795,
 0.9416251447138801,0.5908381071533179,0.786855947365739,0.06562034463026112,
 0.14870999716379862,0.18417749710551012,0.9191492515003257,
 0.12312418838414518,0.8050153101453186,0.575219444414626,0.9922894470415885,
 0.1409290341175582,0.5857194970632658,0.6200591476605529,0.5297436829964315,
 0.15986949268287287,0.5400865807615792,0.28490914563768643,
 0.5728927753603447,0.28048650516377893,0.7890764841177561,0.9365896017653504,
 0.0964909135421399,0.6521090085155612,0.17407882202858505,0.5568770605740974,
 0.8539085184202493,0.008878918993014184,0.7435227706502786,
 0.04104555892091599,0.05605707369952673,0.9681162427066506,0.893243530910611,
 0.005552418377211099,0.3933355229673786,0.052127066762009955,
 0.7840496635302301,0.811856962294678,0.3202589389866922,0.6150531930912845,
 0.9881533868678279,0.6003595648457292,0.8786113005488253,0.15453121660205416,
 0.5952794084166617,0.6114823572432377,0.1206602976024399,0.459100115718341,
 0.18229253844814552,0.8480862660110173,0.13488979985375238,
 0.4657348552378213,0.1392481186519915,0.4056712075470237,0.9162395887195698,
 0.3069899493036511,0.005858017976362984,0.2466287874341928,
 0.2678064947034634,0.05230182146900464,0.5308141676550294,0.9442214876886518,
 0.16680382020551365,0.1318402791240948,0.5999640580709746,0.6236343571999097,
 0.125044635308326,0.1949651122251549,0.410842365432746,0.258986972767845,
 0.5836743216355205,0.8445638920593406,0.710937970876655,0.38438617441198375,
 0.4388087183586673,0.5671835678420293,0.6025792616457806,0.35876337193369645,
 0.9696796314660481,0.7846913396486102,0.9080272200080611,0.7720420720676031,
 0.41483031880281895,0.8408068459963788,0.9153090264780339,
 0.15086396225509047,0.6005518181208792,0.7002950744348029,
 0.07674453485448529,0.15699891858885995,0.4560806227108092,
 0.6381687351983144,0.4017105461506155,0.0763815288164379,0.5875695819353364,
 0.8158635775466192,0.27231761895853224,0.6131217595306487,0.8110946382423371,
 0.024322253536464444,0.3996589819947285,0.7490609222984178,
 0.7254264251619398,0.994646636573979,0.7033838806105529,0.17302449070073678,
 0.9466026182977842,0.9452217632570986,0.13643627234494093,0.8001932072229838,
 0.03838052689751026,0.9865578031117338,0.8366157527840052,0.9216651581463519,
 0.6997183786338979,0.06189380874892336,0.18292861905683178,
 0.4777649108809985,0.4320436364675212,0.1552887408929643,0.7656119089478292,
 0.08547134417298707,0.1051899673747877,0.8155685221929185,0.1017501705552527,
 0.20445483193303315,0.3837707162610605,0.16895509940413544,
 0.6236797162329953,0.48315287200776535,0.5715593539859575,
 0.04472186596101835,0.47185060475710483,0.04952860100760503,
 0.19159133921649052,0.26531824966630346,0.20368491369090602,
 0.12927124719285965,0.11159195249080023,0.4032629485420972,
 0.6708850377689005,0.29126837083932866,0.6548689302653464,0.6289675385409029,
 0.9490157089639434,0.3867945004574683,0.04689866409632404,0.6317269324977743,
 0.9503517898261458,0.978226909060235,0.9658402178588094,0.24227710590128282,
 0.4682883706518345,0.26375990660788373,0.540463409401862,0.7093255664225935,
 0.5635919552971229,0.633250940152284,0.5739286426008077,0.4406355783049989,
 0.14709351757713507,0.5259576442632669,0.6325480473020202,0.0037720140828017,
 0.623856845865971,0.04375822943973884,0.19901700667350053,
 0.07042702736610151,0.3907019818429641,0.4646385625524416,0.5568662948070258,
 0.4304408043760485,0.44496144303515306,0.7117549593026459,
 0.05941516225669685,0.1067370515403494,0.09928266142853415,
 0.1025999281052059,0.2173404657435032,0.03646205575227812,0.2619103129670042,
 0.791686560113017,0.21949600131445524,0.4210553823762888,0.17489172277263243,
 0.42451193243212704,0.9958110926186229,0.402455477677072,0.5823350436642734,
 0.0022966658067753443,0.27012562143650976,0.7417655416217493,
 0.4195760032580892,0.8616720667996821,0.6604542012777336,0.6717720712981647,
 0.9196853331334446,0.9430226079724952,0.9405036418709688,0.34222248444907677,
 0.32058694517419206,0.5073835414086139,0.7880685325836726,
 0.32109997908448973,0.8092461172230854,0.29255310203107754,
 0.7582072555325444,0.9525767621869907,0.11504758894244083,0.695919991358074,
 0.8382640698905675,0.4801244536108311,0.7302692899680019,0.3506043271815156,
 0.17129452772004183,0.2425485660001233,0.11277980425036227,0.405230962810459,
 0.8136127872187188,0.35611060356062285,0.16394224625643283,
 0.9415404548874984,0.937682329415528,0.6432929118363754,0.828846931400264,
 0.6547427808175419,0.2950092532757598,0.2793460076234138,0.6217102939321388,
 0.9342979757247732,0.8075975654229456,0.020257010613625615,
 0.9233470296071838,0.8672075967211516,0.907271137996918,0.32651494455062613,
 0.3377913093280617,0.5867210994769476,0.9844824989821046,0.4371786181991886,
 0.8288608242512412,0.5036600029758612,0.8670045623314412,0.09497347764890463,
 0.7511169692172704,0.5114696220246541,0.638340482858345,0.2840001051759091,
 0.2038488914164207,0.25021132990987627,0.8235111834463128,
 0.45909457238074336,0.45688228577381584,0.41855525668282156,
 0.862106203219798,0.4062741882986006,0.23360394806759566,0.6334785312900237,
 0.3575945742813046,0.9869544613266857,0.1727133908512425,0.5011507072976431,
 0.9180832559001076,0.13484986378312092,0.5684952264351314,0.6624189590875095,
 0.10386893353822213,0.03318580884807287,0.1317737632600262,
 0.5380947676940171,0.6384284542511931,0.07666892887139798,0.9130420379916746,
 0.28243182984147663,0.24623559221882174,0.23023256531625336,
 0.4508802355615177,0.4835855201119553,0.9683029845847823,0.22521326317965362,
 0.8223646351937471,0.23564499432062058,0.05973851416354348,
 0.1211243971481617,0.798138965552005,0.31999472750496605,0.6865214330441465,
 0.9813582394222494,0.23005701192414918,0.442509700808125,
 0.029896057417036048,0.6953218287776625,0.15861316360829214,
 0.5394589495687321,0.05338595381732214,0.8445631982448819,0.4135785117560842,
 0.18750528892779217,0.020239928588712153,0.8427071595924196,
 0.8621971175821319,0.7930554112208554,0.3018798053160878,0.1818407639494698,
 0.7550639047038377,0.2727813879991373,0.6525575623581656,0.11708108241860993,
 0.3216149868949518,0.0014953487333765336,0.5281833411035946,
 0.5569718287471858,0.3642412259148363,0.25051328752576085,
 0.32176491803197393,0.007749719536585542,0.6879527316549985,
 0.6304876214539056,0.3671086421988763,0.7458967760283888,0.815361772512551,
 0.22218612348285782,0.005219976980973984,0.18534952744747224,
 0.48412861343175495,0.49563575455201925,0.17043037798489225,
 0.3736112526203358,0.09603226898347728,0.8402865823308386,0.3059957622761321,
 0.44665217860737383,0.9965905626166512,0.15676827045830088,
 0.14292684629041275,0.8754781452617764,0.2791791297741404,0.8676980206596819,
 0.018776145341399042,0.5293607351156191,0.6080930219522968,
 1.491032213635224e-4,0.1729582077228632,0.9011631033101615,
 0.4364293360052992,0.7517822565536003,0.7652735051580002,0.4399424605487545,
 0.9493033761428681,0.6778404446774619,0.11145470930254076,0.5961502352045815,
 0.4970804050060055,0.46675690696752825,0.3468370142533972,
 0.06184616658000952,0.7485222987950836,0.05917618614577924,0.419255802724516,
 0.8820384534752725,0.36129584971192064,0.14227936807853214,
 0.4391055630402424,0.3028752574841227,0.6617207960024957,0.23529201478523187,
 0.004640262013095975,0.1853714341156415,0.2759875670163803,
 0.9205853235197162,0.6142289687899861,0.3597948745056584,0.11799027164012654,
 0.39722900930313665,0.9960583604368739,0.5346852646070079,
 0.24991543672150873,0.4515932483631,0.1320884992396919,0.010044723025723457,
 0.8608800727600642,0.9906978716531747,0.048367731674193615,
 0.13130066223931924,0.34457906857009557,0.7119864838338648,
 0.0019699148844680714,0.075687911307318,0.08635615124468954,
 0.18809116702096007,0.7380847117893968,0.19162840360676459,0.690123426104232,
 0.9438235413286618,0.9695188580389591,0.7334770115871948,0.662062559138793,
 0.41293530951611146,0.5318670447153684,0.5127729681950157,0.8349614292410703,
 0.7488592431135261,0.11098194410008588,0.510270131692563,0.5831404416164736,
 0.5350731042009752,0.8928386530993013,0.7503780137007848,0.05435760979813686,
 0.07236888763935412,0.9285723103128575,0.01348149325591319,
 0.8567833768331861,0.4737203984645959,0.21016563582600523,0.952921340930616,
 0.8116048013234289,0.1277384221055683,0.6253105116609414,0.9918768569900972,
 0.36209054568544274,0.7822089262631757,0.32125084491674494,
 0.8275917653590421,0.589390527117734,0.5175218005638924,0.5949926326568108,
 0.5993697397435812,0.5047427757932739,0.6566635881349695,0.1619303699796769,
 0.47386838825075617,0.6420620280553034,0.2217578051704988,
 0.17147596825457467,0.9275123922975956,0.7092281884447641,0.5736239383707109,
 0.10865826526474287,0.481075535508127,0.40102589482520923,
 0.12383761812751104,0.13267887056762007,0.14520923505867678,
 0.7986770876040487,0.47034557265425314,0.3855638579838908,
 0.35431047182859177,0.5517431793243425,0.15936219974831578,
 0.10898285350880854,0.5433117816201887,0.578600170744048,0.5458047642086399,
 0.7618848118209709,0.7145641877124563,0.03774842960242186,0.7951532420332811,
 0.9925462730847407,0.34689677083102577,0.342942268213539,0.05653066899169046,
 0.40776610171633776,0.898640735964855,0.4746740070863671,0.9427522800491052,
 0.040877980979777906,0.445458855750575,0.6867809119187407,0.2939236548915769,
 0.18338332867008345,0.7945321712026844,0.8873654744071577,
 0.45267866895906206,0.22717387385351917,0.8699498863287857,
 0.22128541533439783,0.936568443851157,0.9364266610310207,0.8802311021945411,
 0.8203188117923609,0.8752187696014062,0.09595130394770635,0.7406436202577343,
 0.9209381759449936,0.28142013557388856,0.9345061327596236,0.6986085498204241,
 0.6580244352239243,0.6869895617206436,0.40712481629787756,
 0.24102688626486046,0.007993918470351469,0.2673215804420557,
 0.6593666256707942,0.9255918794448849,0.31734853128893314,0.7755568752670506,
 0.04385605793799929,0.6394853839753819,0.5884389143153159,0.210030922843215,
 0.19306104557153336,0.6966267639610487,0.05013598823370258,
 0.2843086892495821,0.5050278020495395,0.8099176539598002,0.7448943666380721,
 0.9141592768608134,0.03056438458041466,0.8126726789563492,0.8481557402687022,
 0.8727947328057093,0.0398391310110332,0.9771075921396071,0.8204691493897147,
 0.6002583846688321,0.6236430085700987,0.43032516831565504,
 0.19362510701133984,0.8871813106465142,0.36258348506225635,
 0.4962857329148854,0.10737684342316645,0.42402422481579216,
 0.41728530471552827,0.9136336115126886,0.5091091287872254,0.9692294785310881,
 0.7722184003322907,0.3370140012210886,0.03075107074527772,0.58159249577783,
 0.8626934121799097,0.5961174187697434,0.9220578001880472,0.7109774165129114,
 0.8887087539225198,0.7882154894937986,0.7989097923850704,0.284808078284569,
 0.5884689025493959,0.304565524234954,0.6185760785895831,0.2818021282405003,
 0.9519013041305214,0.44835641542804405,0.18945617860811648,0.762529569479147,
 0.8843252292527746,0.9347467339749123,0.5725293628096615,0.9363180583944637,
 0.16601024990401347,0.38647282037375386,0.3733544299174232,
 0.4168470327743756,0.316958818972055,0.41946011335194855,0.3773044974000045,
 0.7912622365535005,0.8490265103130454,0.3931308882741609,0.5371423887179935,
 0.32588486571416075,0.7723661169026347,0.6081677149441981,0.5432096894725428,
 0.07542906887031564,0.35940085685510315,0.17779567144587827,
 0.6364961735630035,0.16704042467203772,0.9296189599297219,0.7344131945031499,
 0.16794322178897736,0.08158285923495379,0.020976571911884623,
 0.6346224958318831,0.4886926149509475,0.31059463985750413,0.5318445008559554,
 0.3192433595857388,0.8454987379015555,0.4677676785913427,0.40228362683173113,
 0.7775867727493154,0.7859193670017425,0.047695667570570865,
 0.9955168233106672,0.45967975662880445,0.5592566973267276,
 0.27376972788357357,0.9100639189779609,0.6767497306022673,0.8897688255917757,
 0.1451270421303682,0.7984220493577656,0.6478388604427333,0.09540760402146464,
 0.5009851715330638,0.33952998322593286,0.6754212280767904,
 0.18427077194134867,0.7688304731290376,0.34427153199210025,
 0.3732279301697271,0.12747908786810125,0.7433226348383877,0.5923970438095927,
 0.3968173592938067,0.8679025872763279,0.8274879979705805,0.18190989141129532,
 0.08467636427244096,0.30085154743715825,0.2433431003799501,
 0.2593328783107558,0.2985134850055676,0.3719000134576488,0.46198313882719955,
 0.9988908710682094,0.03735818510120226,0.5620206702212643,0.5122589486891672,
 0.4739558445415306,0.3804431281055104,0.6323991795953599,0.3446953262730771,
 0.06064080835456942,0.13649900196320297,0.7009095433041508,
 0.7908421759955245,0.772256042320381,0.454817989516628,0.12255907111224074,
 0.13062547101900523,0.5554867809417428,0.9915301769364696,0.8781901221657545,
 0.1053658774007118,0.9750481916653653,0.9886218032836712,0.28103625606517635,
 0.9354738565049128,0.8079245894856502,0.20674257155812092,0.2920040241073354,
 0.14395335336519088,0.020220251059371197,0.5580020940113536,
 0.8274559298964731,0.1179639926817101,0.40700569950527765,0.7489012780539471,
 0.13998008603081036,0.5938111562533316,0.5240186249681367,0.9805627293517473,
 0.5727104077969813,0.7138181156747821,0.11216371362456667,0.6658081018887035,
 0.8114049771915031,0.532571600814046,0.14603150068561133,0.3900470591789845,
 0.06632713184948469,0.19342910826271376,0.11979331022190509,
 0.4978402532029611,0.41733693583207443,0.9394410998899674,0.5915255786143541,
 0.4218243121185925,0.40769838550659065,0.15329834970535394,
 0.09169280170544714,0.8215213711140461,0.12097094664525598,
 0.43136963605610457,0.09505701475057227,0.5637335118456566,
 0.9426175678058757,0.3422853022705038,0.36939491050887985,0.7764309994297666,
 0.7193796153202751,0.6487954150070333,0.8918337704204505,0.7413636805066884,
 0.04337530634290854,0.06209487566887084,0.41385741843270807,
 0.4717273966328368,0.22672887394915264,0.3170406503450949,
 0.42372156366401614,0.2687999398443617,0.044179526627187604,
 0.19374661320723274,0.2012383198084906,0.9273701553367475,0.9021003676417978,
 0.6437733256295743,0.780804308173793,0.8580467029717644,0.7560791020025296,
 0.6013571046122945,0.7759472920813919,0.2801133420742993,0.30892050724261166,
 0.197325834623575,0.9682817999067227,0.3453963384262867,0.2516468286287501,
 0.07635892332715655,0.22258789498693754,0.08571539622164281,
 0.8177851575771118,0.5585619668122237,0.2389518517267497,0.6844956761678468,
 0.14693513826569293,0.3611904498889489,0.5277964803001257,
 0.062086623797116935,0.6555847816841465,0.9613952239819533,
 0.5695558665406731,0.06227049107283289,0.0713108993351067,
 0.023663719673750583,0.6245686048652251,0.42879580187589783,
 0.07509390525430093,0.48230370502518827,0.20714773393356078,
 0.11502091694687067,0.6240134527712771,0.7112515431212404,0.7802395837359961,
 0.891430040824161,0.8618546829770762,0.05714071318147618,0.2471118554641969,
 0.1674444079062658,0.4276547645981392,0.7752505242022325,0.471239158464708,
 0.2675246596209282,0.40520932680231525,0.5612400960656263,0.57804187544336,
 0.18843817212236613,0.6251910722657676,0.7560725998706015,
 0.09048839924182028,0.35269451914357997,0.47226717351940317,
 0.33313184379313365,0.980122252256847,0.8489235861480411,0.7287462068553536,
 0.18426584408275526,0.7174258709521271,0.6304436669747184,0.1701479433482821,
 0.4613602172226253,0.3952645114886878,0.533837836524288,0.269259414359148,
 0.09959784355102774,0.32004095749548855,0.8663433915427204,
 0.44545375373867346,0.7172711078630708,0.8070947121843235,0.2921744622031801,
 0.4231465088952616,0.1679990180485974,0.4518332456062245,0.32068950636829885,
 0.6648274778142259,0.6405936252474758,0.6268331269686068,0.588871102755855,
 0.733392403104115,0.1140833466975274,0.1097793800218071,0.36648351437608484,
 0.12979147606589514,0.8884315353265457,0.489755716102946,0.8514559162658921,
 0.9043974640106187,0.6450658345155289,0.029397583385115,0.8803514587608181,
 0.7788578779644093,0.3108339649178149,0.7745003901951519,0.4251418134146967,
 0.5783208563475117,0.25870701389391226,0.14069582265486102,
 0.48832920464926255,0.627927926851634,0.9340314357017822,0.4901159832723454,
 0.6283409569686342,0.21125630765041215,0.8028570112895635,0.7359996142529899,
 0.03051776265036854,0.7078606287290542,0.04403036638390678,
 0.4174956867365983,0.283520297411618,0.08879562029987009,0.16298035743622763,
 0.9432304124285984,0.9107400165858547,0.7192349938635584,0.6626722852913044,
 0.17724324192937013,0.7081166991140355,0.7354940715635856,
 0.27245544441178715,0.1877861324158301,0.544556224682214,0.22831940858065214,
 0.9869543692782659,0.8733088771237973,0.24915373655336537,0.7108498951318624,
 0.5809306040515806,0.3042544961911534,0.49970930696097204,0.6630983927009254,
 0.5688366959112391,0.7274848269893607,0.5096938252463497,0.24613392867420048,
 0.17277468983013378,0.5728570040997726,0.3008864594157491,
 0.44795290765855134,0.01323527213351583,0.6879892281719684,
 0.7295144400945289,0.8278373848798312,0.3902360781257359,0.7268609445044822,
 0.8732110986856129,0.03460192702961551,0.5692807803140344,0.6530701596058862,
 0.6035348107253005,0.5556024235949067,0.02405683000238068,
 0.14282924891350945,0.50083657755205,0.6264543903459232,0.06203989304880231,
 0.8082946480345582,0.46666969166770744,0.6067867102021889,0.944425987031283,
 0.019763087075347907,0.3120210423254729,0.8313957365356178,
 0.8150514831502457,0.041605663897504974,0.9517271740496807,0.908864679403211,
 0.8961521472376728,0.050563666838603916,0.6549992708259518,
 0.6108626113759721,0.14220157636180508,0.6415814998768887,0.9395878800583852,
 0.4060896344677074,0.8752428149231151,0.8981198375039811,0.4644317367294045,
 0.41132793556535496,0.08542170582315645,0.3401486738712246,0.967693398620327,
 0.37404226769226845,0.5135537726938847,0.6073080231096717,0.7583721171742464,
 0.556782568933545,0.8272261743040354,0.34664822197220924,0.6923715132935137,
 0.3031870073226639,0.44414251398094917,0.10836829257530012,
 0.5545408881469941,0.4659100434278409,0.11642836973280168,0.9577436973918747,
 0.8772143426359175,0.18188456025503885,0.10696752693279177,
 0.6814347516063611,0.38832135740213214,0.9027814276879972,
 0.17000026998261353,0.31928032314749233,0.1815186881393468,
 0.4695693331961369,0.3491055807045643,0.7871357126476335,0.7491541182647554,
 0.746724824734055,0.9090966448195357,0.7452310169657211,0.667250685225117,
 0.7794442445739789,0.3251477175463715,0.3339476127015142,0.7640451648630696,
 0.01763559734031661,0.8524644115686756,0.607990341135926,0.33534667355180336,
 0.34903256109474656,0.13133546139736474,0.4101436488816935,
 0.7215224900502484,0.5232518768307388,0.04757507905505498,0.997416923765994,
 0.24489066591062503,0.20287398977303794,0.3324024352151458,
 0.6376021622348237,0.7399386698809005,0.2918875329008267,0.8041378205577114,
 0.714988553902479,0.8887485801257389,0.7504639026352788,0.8433986230700797,
 0.6913017482451013,0.403447384133658,0.6562112095237153,0.6201413098343037,
 0.8730286985642239,0.7755368571401896,0.27817014257870365,0.7420798503473163,
 0.3486648661331764,0.6577112342185971,0.47403737079474284,0.6030650644306569,
 0.2854606730232201,0.14711499616865642,5.519428989598785e-4,
 0.32752970849147034,0.004362210810679001,0.4159178381687565,
 0.8181641079812435,0.7111720279366337,0.010504807756715917,
 0.04358266536827471,0.10817259921002242,0.26298144289557124,
 0.2045331071994675,0.19054205330898688,0.8700385138636493,0.9669196646564213,
 0.9612577991155493,0.10362254695466866,0.2089440812342196,
 0.35502374808214887,0.04655907304624085,0.18980286727284135,
 0.19041914366171298,0.5267262833189325,0.12956510269269061,
 0.37115478667127455,0.3751484592187123,0.10934346349147794,
 0.5765665169531244,0.8360520287720566,0.6634004671507587,0.5507608193446075,
 0.9674858826098814,0.4230271395127999,0.4596837295787145,0.8288686164443376,
 0.5224527249553311,0.1448793922386099,0.8396590181314073,0.24042186090973738,
 0.4130246518116787,0.561218424719033,0.0338722516423936,0.38877449913580997,
 0.6201977407827934,0.7384724620003589,0.09300209763460754,0.9844045322498571,
 0.6854070742360843,0.6309327500459705,0.5180432730507994,0.35341089697908223,
 0.2287598786926054,0.5946454952402671,0.031677421203614164,
 0.7723813407548671,0.8233543219984697,0.1932151491938695,0.22478013411318365,
 0.5720701656253881,0.5943957513054839,0.6439616416176411,0.8160090677329634,
 0.4723479016927926,0.038970150501627665,0.5806668600709537,
 0.6051637460550691,0.04952770315098887,0.46129740742742187,
 0.10129020673535472,0.5872292680619242,0.41054918543448826,
 0.9722726808828418,0.46476649403637027,0.6545703967368368,
 0.12201261413502318,0.6464835513509798,0.3604914312617863,0.8402004412181386,
 0.9268229685822363,0.0836696739238878,0.9801219438922207,0.13263975691492225,
 0.03959086049577476,0.17289563440764666,0.28160713254411174,
 0.20703293688467095,0.48161088417863696,0.613385745775435,0.523393564096232,
 0.0494766679890859,0.6646072638496835,0.8920780219439151,
 0.019671796156570553,0.842719725005637,0.8300890627262989,
 0.44598479606197405,0.579331552843338,0.2505698616080555,0.7078590170091659,
 0.36399175194483735,0.5427600682307681,0.5737926307670064,0.6784696865028927,
 0.4843082215636536,0.5705154573440695,0.8149367750538123,
 0.014401363558417923,0.38573327250305667,0.5935343830460181,
 0.18658870971369035,0.5863465227489155,0.6140434993919984,
 0.24556187572135801,0.423148649324267,0.2888887716981311,0.6734077823006654,
 0.49546141409013367,0.05141539150051999,0.09164777531780977,
 0.6500514045820818,0.47588648458790284,0.29872006309472243,
 0.8655445932342074,0.17555935483220897,0.06634572794274163,
 0.5575277358025656,0.24041511854787423,0.06962643002704849,
 0.6046193803913279,0.5203993954366491,0.5594748858685125,0.7774018360821378,
 0.5813498552124237,0.7986429888281028,0.7138269706657381,0.8894175207467164,
 0.2698282792523419,0.6898353058119979,0.044092249685635965,
 0.9213136884716331,0.5536852701575689,0.6661986387688997,0.11790193493868129,
 0.1436623389664452,0.0069593430246200905,0.5657268651287956,
 0.14804707448610743,0.8560400762690382,0.872071289716504,0.700571580370742,
 0.7507818064496667,0.026397334108649306,0.3932089384218801,
 0.5927321505039361,0.048140445779554786,0.2309520162616171,
 0.8318267346953004,0.6436742082579426,0.04392112613596333,
 0.46720095825542607,0.10715612672378161,0.8998817995621049,
 0.2665350851845807,0.9947509867143396,0.31749588568761355,0.5711340618770852,
 0.2347039672690988,0.9907719855655752,0.4432281739866424,0.8570346809291881,
 0.7830042091653415,0.828087741892217,0.3922685251662338,0.6482463036209507,
 0.5084690501113819,0.23796377242388067,0.7781816714068867,
 0.41263180608916705,0.06200869101033335,0.7145129270348223,
 0.02487240559408066,0.5129802470451527,0.22624398597529383,
 0.7329025635732724,0.6430410618312842,0.8786781328553058,0.29539904910218706,
 0.2105008719866035,0.24053121722620863,0.5797102573139343,0.796158441068534,
 0.06267296606587369,0.0038129403105765514,0.06401726635566407,
 0.6744151760202414,0.022548849529479553,0.17576831492129918,
 0.7390315874885562,0.24719192696954284,0.9365558589271936,0.6405120751371339,
 0.6486481838991117,0.6472603404204383,0.19883243743774337,0.1471373845695062,
 0.7064680327824825,0.3623588745582358,0.003900148764726996,
 0.43232652370579394,0.2979607543497804,0.32789471442146745,
 0.5752508124346113,0.8764675267070751,0.6993028502361999,0.24930781807433222,
 0.30649393716940676,0.03290083838023594,0.25316593941981547,
 0.14332558013348806,0.40127768341902104,0.04945119171259804,
 0.6449131593345262,0.2724657777847963,0.9924879465647022,0.8530657249783287,
 0.47361710434666937,0.39474967550992957,0.4472554532050036,
 0.07012521828434581,0.07158060534213861,0.16054803907649162,
 0.1749289024551729,0.2927436360259096,0.8123519054534862,0.26907800394557846,
 0.4532677058776968,0.9599246557781355,0.4505248442429616,0.6311699966848194,
 0.9197480456357793,0.9171260761564584,0.525034381122678,0.2908868913754612,
 0.13943200816546764,0.3701344353408995,0.2528622674401957,0.2752258275425339,
 0.39620983098311646,0.6056958248633024,0.539425277727335,0.8972624550975323,
 0.9221772986075769,0.6950763097521424,0.24395505951508123,
 0.17343146078732607,0.49405707932606324,0.9105581796689967,
 0.024145325488288072,0.2095242267171965,0.11291028190994779,
 0.7292092622590035,0.6013854551653376,0.9577748739004752,0.7975054729841782,
 0.06506337418367802,0.5196980454906392,0.647300245684243,0.1783373320559637,
 0.36131391606077756,0.11322156552160456,0.7431951476660916,
 0.4459489578407063,0.25799616008589776,0.10347901334840559,
 0.5644276705586304,0.055841523754815436,0.40341535375290016,
 0.5418857760434741,0.37882443140138955,0.21803865236311137,
 0.7037694132191478,0.7620828094960592,0.005691522887249945,
 0.8371196523110493,0.7284896638644737,0.07760987839507127,0.4974882656963381,
 0.7229870322613909,0.3069825356861432,0.7097738537756556,0.9125208930279217,
 0.08488080129529252,0.6843832453947527,0.44146644899950016,
 0.8224206626109448,0.13996829831751756,0.27969669008240494,
 0.3227620062233263,0.11946248708135843,0.6022205899451202,0.6427840732030994,
 0.1672210673664144,0.046292656306565094,0.3085053889535133,
 0.9349125874135423,0.70379328856854,0.0809303608842582,0.18208978783223406,
 0.830637903549125,0.22191679524178975,0.5927385983042586,0.5297162516964217,
 0.3775629909595114,0.7154552235890405,0.9141771957966344,0.7007337476606629,
 0.061964664264602676,0.19620561703388606,0.3176272257520478,
 0.5208756964085068,0.2484184370400746,0.08236147285464768,0.2992309491787277,
 0.7783374246333341,0.34294656687111125,0.38341692431231045,
 0.8565597746234914,0.04758324388450985,0.16778107204720139,
 0.6137100097542049,0.424694597301957,0.09565120290887141,0.6102872663908963,
 0.6995995337183469,0.6685741464543993,0.3076082370744915,0.2314432028635326,
 0.4030749650107358,0.731882836224372,0.932249738656138,0.7037402846988691,
 0.5040138673776855,0.8042018811852165,0.14312899957688074,0.4676937293622705,
 0.5817605650683915,0.5105768920782537,0.008810765639099327,
 0.3424360874361888,0.7716588419704093,0.06585082734227465,0.5703224648495224,
 0.5411230994965702,0.41697579376640503,0.6555236026432967,
 0.11414721519982762,0.6783889129771329,0.8582065564811977,
 0.15452862384400046,0.3829893812257291,0.1652222822727658,0.7567869572951799,
 0.92171861229501,0.9048608519986375,0.4080260424024813,0.8545510585735694,
 0.6962896626043199,0.8185494826782488,0.7427755848198104,0.8906561006230954,
 0.7621996651842253,0.2183437837195521,0.9835238093987083,
 0.015689535935107135,0.43584138592249533,0.7013179607566931,
 0.6961601646071572,0.4254280146070477,0.4941136297391241,0.13824883719926162,
 0.7927381137222512,0.43783899136036275,0.7148196985888127,0.557519989092546,
 0.10013966674264274,0.8756591164550319,0.47045493909096736,0.373449234571154,
 0.14536878737324288,0.9439725736320024,0.671903238391816,0.08136643472859417,
 0.6363083851078613,0.7318263130033232,0.6620145559969969,0.2876556512219224,
 0.3797982867118216,0.6486635749596974,0.5688961279604658,0.5101821478082854,
 0.9975768082244777,0.9176199012761307,0.4039856865078808,0.0324801982598486,
 0.3449813762955459,0.13274084383090778,0.803618969645256,0.6878470554351653,
 0.5544921876198129,0.08349565976430884,0.45535703472597644,
 0.25294723056334445,0.7208342583281897,0.2528253648907334,0.3834618758916225,
 0.7746102062644793,0.4933131230985828,0.4169205256904406,0.6108039633422773,
 0.3933675381634416,0.0375748778442081,0.5017232794098248,0.02495642720765634,
 0.0757777335759161,0.761668551278761,0.32251959997161295,0.8291649040316854,
 0.36975068792558985,0.005403263578827255,0.14778523012477118,
 0.40606202707760597,0.2625587416018865,0.6397702748029036,0.611047924919876,
 0.9564473402937325,0.14407855460902097,0.9330138290044604,0.5773296272497911,
 0.35740949209361805,0.5201070264551886,0.3419156264265366,0.8621749549489587,
 0.8226083455345443,0.6467871353272818,0.04431960087245934,0.3140281320130133,
 0.14109161565620065,0.6811696038367056,0.08440959704445528,
 0.1318685969722857,0.3434924027295755,0.2407166271979323,0.31949987338127406,
 0.6508331882321331,0.6959170887206186,0.5097638442652113,0.33571438934631304,
 0.3354379510747003,0.16535425902011447,0.13929629098544394,
 0.5708176907637105,0.3642154515946273,0.2922654472646514,0.832031384095937,
 0.5839391037516387,0.6619124460685516,0.11193645995738666,0.4288534586988926,
 0.19801187029720757,0.20423205116419163,0.9681934964684226,
 0.11293254083844795,0.3277164704549559,0.06808789299339846,
 0.6209761017671365,0.992100827833031,0.4112867205033034,0.06067166910279076,
 0.839855355730169,0.8827977752772896,0.682745214198934,0.3757338689471603,
 0.5675849246758453,0.8628805968526583,0.11993304498130608,
 0.035386470653122615,0.8488349804182522,0.7793047878143678,
 0.6882465218395938,0.7030884486321676,0.22699758907562606,
 0.03909098998104843,0.20819947320831045,0.35828573609073733,
 0.43627900170283307,0.2393706706292097,0.514393085064702,0.09213495000344452,
 0.7279773873995927,0.2907454834361006,0.13235327184585488,0.8276168040985198,
 0.6392725703798364,0.808106828660056,0.2833174389937504,0.5091717757899967,
 0.2766779366782506,0.021428052011293053,0.8334884377230432,
 0.8023402618787155,0.6836041351210244,0.9061250961564121,0.728846561167182,
 0.028182416246948727,0.5390170954899934,0.039090813458992235,
 0.23067883486084173,0.0197674153943721,0.6720308206562426,0.6796489977376727,
 0.26409746969181347,0.9388999270642295,0.7898037886610008,
 0.12356470401760533,0.7623535798878543,0.45368405479359974,
 0.37994349545501505,0.5875282671895167,0.5642373324560734,
 0.04349547990358493,0.5549345634383861,0.479496828813304,0.5217108788150014,
 0.13160994046815844,0.9241888800816674,0.8942692301284116,
 0.10878435969731903,0.6069525446215553,0.051369017159173413,
 0.8295413162512576,0.0260664695103614,0.9217962554810303,0.5438228869413408,
 0.01090701857943377,0.8218222383364051,0.17939344428208193,
 0.7934994698226292,0.7134850313255188,0.2029559935883246,0.6727990424319109,
 0.022121684149936316,0.7199471800094948,0.1168658361537055,0.909121705041106,
 0.8536810616926718,0.64321689931907,0.8205291832250075,0.6874964688468767,
 0.8828160299253083,0.06406712318788921,0.202628833143499,0.6367235742423157,
 0.9142282829493703,0.8374920920097696,0.16862832074546752,
 0.12260122616928593,0.33443333313903845,0.5914587144309416,
 0.8519211490737831,0.5622997560537617,0.3556282395893575,0.20434502611208494,
 0.7123327503173245,0.3997227992794994,0.8757042817839725,0.9541887343947582,
 0.3862500390966669,0.28238604030282743,0.21355186837120432,
 0.9583314455203173,0.1671709872920678,0.3597227203538905,0.7542165113566234,
 0.7943484964884615,0.5472074601029072,0.6238020137015442,0.22257523001648172,
 0.32711036928416837,0.2639584130181539,0.4835236059322724,0.72669953729345,
 0.7712114779704906,0.4198244666563,0.5411514863310926,0.6622915370386138,
 0.12344420773745468,0.60600009685397,0.343386283855006,0.27875784242584445,
 0.389454025364279,0.09076073028859677,0.050003465481233245,
 0.004060175696992685,0.13480955468092914,0.7782545129287566,
 0.5684429118939393,0.9482546908539832,0.6377123301285039,0.2127423816995121,
 0.4383293546612541,0.3795693621981203,0.35547132927270586,0.6830391325203911,
 0.8639703365669591,0.37929522465825394,0.8134772266823485,0.0760165422960547,
 0.17379512583796697,0.044857627189490046,0.25274593282402447,
 0.23897417051849645,0.9319837349955254,0.6240885384318061,0.7057334984191159,
 0.13840238406543748,0.8033218959150523,0.47225288688003964,
 0.4810644677704099,0.3384707257203209,0.4147921094136009,0.1923781343357679,
 0.9041272972229851,0.4808746614433089,0.1171003393727269,0.1573694255854745,
 0.7678326487975571,0.21513935405208362,0.13480835325624885,0.678952435667847,
 0.31487073413276345,0.4235539694780952,0.8768171412924263,0.841593235537108,
 0.3883850283074577,0.945927711611368,0.8575854993853045,0.15606399774858393,
 0.7209339319293329,0.5204548062148623,0.9225086626326323,0.424087389891067,
 0.6291254108699241,0.0253510379085069,0.6083935505018576,0.5633786297876671,
 0.18274359686109864,0.6207355976841287,0.5981172286989762,0.3416506909773793,
 0.5669866988601853,0.7837738675826733,0.5995478275561337,0.701753415307503,
 0.0010142303106823736,0.16239321190833678,0.4212735559989661,
 0.28740954078226966,0.7338788683587807,0.5204008753189373,
 0.44398247545048675,0.3969004178315986,0.38806436317836634,
 0.8265138025001149,0.3152215921781162,0.9374386306968592,0.9612601993960445,
 0.5794671015774967,0.16078770930271025,0.9258596851851132,
 0.042316967256102234,0.6169982007803394,0.5551906732511782,
 0.32449630723909073,0.8232326177734811,0.8945825325152911,0.3862618388887382,
 0.6146880848263824,0.006549655417004807,0.19754302485599906,
 0.30570858685063895,0.6194725834314838,0.13660191732384874,
 0.8462855110550594,0.7133499698156864,0.26020732416452885,0.1546745041899592,
 0.3231037074946852,0.013451632074877096,0.69745300245435,0.865467200143387,
 0.12637312301030146,0.8666738633909506,0.1407173172294396,0.6980881982237203,
 0.2263044554613285,0.9197338974594181,0.7195764284567843,0.08005523422471705,
 0.6663806267201863,0.1771364557045665,0.46090530799513496,0.530977149671978,
 0.7606446084025951,0.2485091878654434,0.36748827859035105,0.6674960427842254,
 0.6635727275044376,0.6383236666584484,0.28956720934610714,0.4112196850461911,
 0.22713936863165796,0.45928397718044933,0.2947531708923179,0.714682588028256,
 0.5975334344004988,0.6169684802804426,0.084815727800859,0.9006662994361281,
 0.9714934661938962,0.6605895664177972,0.27173707662678426,
 0.03844667755636255,0.7956839145898037,0.6592501823289877,0.9589619857700933,
 0.7091568510016529,0.13192365016183039,0.8887175764486613,0.513026642630768,
 0.11151921228815087,0.9295828229052828,0.06279494811879305,
 0.6169689080517997,0.1408953572834537,0.9371085603571051,0.10682507261034213,
 0.45816387088540333,0.13562610307343326,0.5682675869012599,
 0.9113296849302347,0.7141431867598667,0.37741074868897595,
 0.09763939395050192,0.46606111493983726,0.7792028251996715,
 0.2845561066810308,0.6330192370332393,0.3594045511812505,0.5136276588171966,
 0.07605225839708807,0.2785442017979206,0.3159269895735417,0.6457496190427541,
 0.8166889671099817,0.8065886483371401,0.25434889580271625,
 0.12341863269103437,0.04631748436242966,0.08851619506793285,
 0.9650840888368758,0.9586669845800725,0.23767039850798655,0.5290948094217471,
 0.35019455106583697,0.3335055521277439,0.5702695505582712,0.7381442136147482,
 0.12245682067790087,0.09913816437704404,0.14457155235241004,0.54141595056396,
 0.6011142832861204,0.3510062390467946,0.3023894224822653,0.0125654107282962,
 0.1708900523605471,0.16789007691374203,0.06606237704864149,
 0.40704449814175925,0.525213938419146,0.5661663530133607,0.5156155849047941,
 0.5688705947791398,0.7145122637450426,0.9278431936040876,0.5733441325637985,
 0.05191794510124237,0.6467696410790237,0.40430421842049125,
 0.35128541913853106,0.9929442470746848,0.7236193147899352,0.5824519967741718,
 0.3535272687305242,0.18666017132627588,0.18080527950186942,
 0.29042361295423036,0.3937283095480537,0.43848476519200585,0.871280258348468,
 0.3010046503793974,0.1547782617159693,0.8708870006604378,0.5874187996854502,
 0.8477476196771829,0.13072084745670576,0.5231659586508051,0.5254880181048567,
 0.40638169733475193,0.4319657544895741,0.306145824773569,0.20430041202810667,
 0.5947869827170833,0.3082286527206075,0.45168014832791514,0.8812796896967184,
 0.21639801426779937,0.5737788101026692,0.7942794811494508,0.8128294023602562,
 0.9152432200160254,0.8923442924770593,0.4074886172378366,0.4197018340352565,
 0.90120790602624,0.15959453414152314,0.40979258363306603,0.5046138458101972,
 0.3390270492150984,0.834975759667634,0.3079890984431116,0.25466645351764994,
 0.2839565997810857,0.8514882368191414,0.6999585911838574,0.42074884538150625,
 0.6875083248215017,0.921408440509801,0.5078999186854939,0.3275016811235486,
 0.808229784979777,0.9763934439312945,0.11424931989490472,0.5431931164069248,
 0.3586915302944521,0.826540460352064,0.34314100809034864,0.4986287823207517,
 0.09412347134420762,0.9221326673633803,0.400528629166757,0.9633016701041477,
 0.285786152120185,0.4187006705584667,0.5544295769430656,0.7700152418729199,
 0.7410745961512719,0.44145258280104716,0.5909456974701092,0.5115629583513133,
 0.0721784205690631,0.4301799770586272,0.8544444099349962,0.21789541542288504,
 0.8612866051487781,0.19119072592596054,0.30147079997316417,
 0.48590728356591795,0.6893367956399584,0.7822825341791718,
 0.09436600078887958,0.6189355955432031,0.6688435704453918,
 0.22291784155620076,0.1692962345023612,0.6901894194106859,0.5628139058229099,
 0.41857247775530015,0.8633243556626603,0.24029861535014652,
 0.15669659452516727,0.731524293612872,0.5317812013990748,0.9306345198700121,
 0.5550511097572561,0.8112655872638239,0.2429911081597681,0.7560378704590187,
 0.3853551895673901,0.630688010857174,0.7472483739741372,0.09671581315324829,
 0.4774720591521724,0.49944015301304145,0.04662818888139231,
 0.13765989601729478,0.8956600208125888,0.03610849259982618,0.447984949756955,
 0.16160843885110832,0.6152898616763883,0.5760211656126978,0.5856607529487856,
 0.8598656813175943,0.457816362793276,0.878450669957198,0.8794804245316129,
 0.8061129511973111,0.3155980275133411,0.03008122026998339,
 0.25074903715325714,0.3105759671858871,0.9790662251652115,0.2839958134397166,
 0.07859934754898334,0.07819048035512921,0.4241435778829076,
 0.10937364994868792,0.5701586767848258,0.6791398213979092,0.639557252356262,
 0.8635645210354277,0.3167841004978985,0.5930008030083195,0.1628232161279647,
 0.04898620249155994,0.4908126773284198,0.9054493385659347,0.5119046829677539,
 0.184069965831688,0.9195016563783984,0.5485217272079741,0.6479890695390524,
 0.3912489226470308,0.5793739782950127,0.9596332740100213,0.575032357512762,
 0.9738235667792875,0.23537599399603115,0.4204732914689464,0.2952900119402291,
 0.06638391802351917,0.46728499795293166,0.6311116504871044,
 0.7127926363280481,0.5121210897601287,0.4290000321756138,0.6897499264185953,
 0.24709930099484056,0.04037526735953523,0.7124626141967507,
 0.3080582521827825,0.8725246537335685,0.3927257008400462,0.9612606630469335,
 0.06734580106134502,0.6460115456816146,0.8955296851721011,
 0.24343637178003785,0.02096168153766209,0.46378646520740896,
 0.11403312331839721,0.14501061460274078,0.854729212849411,
 0.16112565354351183,0.2741394622659076,0.5088600505320476,
 0.22597157069831808,0.05503286634052951,0.8949755702390313,
 0.6356491787542841,0.1865147440704945,0.1701060707985591,0.5223862072484187,
 0.6338654622100353,0.013976981367465946,0.22969824187568333,
 0.9644413712086407,0.8155624518994692,0.7340198917073151,0.06646844875203173,
 0.045455839489617955,0.2661364920426128,0.6012317763654126,
 0.9912883274625375,0.9547590229063289,0.5912529972502562,0.49609121259680644,
 0.5883676353810753,0.1421530809896262,0.4067143546524271,0.1878442490772434,
 0.4392325736317284,0.8077012313376912,0.2132126958386963,0.6858088359307033,
 0.07654899558061334,0.2426183230249208,0.25556140295724405,
 0.7614358645443587,0.9637136554335983,0.719521969337718,0.4360375715358191,
 0.8784641859636761,0.3322382734186021,0.9141034782201731,0.9337999627428659,
 0.643946226545004,0.7106120882015211,0.785276607274052,0.86743909368483,
 0.6877759883440974,0.5692151335796529,0.04735360209751249,0.9532138901218261,
 0.6751783980079723,0.5846450521288629,0.9382323496582866,0.31210300619744424,
 0.7580604568559797,0.41545920450138396,0.3277365584392522,0.4861819400404215,
 0.31503749289344485,0.27870990417090025,0.2334041326049333,
 0.4030597967517071,0.6118956576792911,0.5086132805252866,0.4620013116477334,
 0.7496054732167089,0.2069215838802677,0.2690897622923133,0.5495320942171323,
 0.036413931575621006,0.9701854503580578,0.7286820853063914,
 0.1703951697031839,0.8304787245561571,0.6035845716650801,0.45841393300998723,
 0.5173876992233526,0.20015579180437426,0.3006003566698925,0.1955672831930868,
 0.47208677602968296,0.502056746039768,0.907990249048714,0.5219636144723978,
 0.3201251083177351,0.6650855993606445,0.4771870856935707,0.0924348945591058,
 0.5681419980591594,0.8133754964686286,0.6009735398223983,0.7416433731385765,
 0.5589557768545961,0.10126107287459007,0.22068963186072443,
 0.8114155807458737,0.1725202806903644,0.5125904446969307,0.8421237128231357,
 0.09562335576784253,0.5866770507701328,0.6664601238866594,0.5166348373070185,
 0.871018494220822,0.7796986282892766,0.7231407069663536,0.9866433793355533,
 0.16437351684141832,0.17045535590848038,0.4781324510173127,
 0.8257820289521083,0.36404918269731656,0.6277938035650354,0.9580244180282723,
 0.9422056922475972,0.7666495610606834,0.8466784029429559,0.588017855293603,
 0.26448740774810675,0.6079154159921217,0.8637166242543066,
 0.23794227838512227,0.7317292079856834,0.930673087941868,0.282779398150687,
 0.36521065977470757,0.4895452474943979,0.5379222634050416,0.5373387541366278,
 0.667904771994208,0.24270473019257943,0.9632661610315958,0.7219904512013895,
 0.19690257403878075,0.14118308162338988,0.12058837130892852,
 0.462385078641367,0.4666036200068231,0.16251417517918498,0.5437341306049587,
 0.5400654356439052,0.262852600115685,0.713394068105947,0.07132296862761167,
 0.37638023679381605,0.1992315594386177,0.5036034735585684,0.4047771151588886,
 0.12153021851896939,0.979117433005669,0.6444293839940511,0.9784587990104966,
 0.5708106521446799,0.6793153570254914,0.24488487827659045,0.8943252828932491,
 0.6099194875086125,0.9753180976801152,0.07192943404822416,
 0.49543187112848164,0.523841177175363,0.20897126050804105,
 0.24434588437840044,0.22442220974482963,0.058582891059568376,
 0.7733457628335626,0.18328418330214724,0.8146902871683814,
 0.03502966872849311,0.7371020624848272,0.04751574170993067,
 0.6851942145025841,0.30364193833449626,0.7429982618052611,0.6337118848894554,
 0.9370711002325282,0.8299679454782937,0.21003004744372689,0.7901154684949361,
 0.8600024869679412,0.4343983431351812,0.5304607396915755,0.2552210394857414,
 0.5897091884421084,0.3888830159515274,0.6243013693720345,0.5338982572661722,
 0.10775510532500909,0.8323077945924258,0.42782994586361434,
 0.26275255672943165,0.34140642667609344,0.5416971157428501,
 0.4689446840350123,0.15698870281767008,0.42997313197061837,
 0.8183583406036725,0.026065299435213407,0.2010808105273566,
 0.44555639537701597,0.4122508783218507,0.7745325110285862,0.6945866933954614,
 0.9304754726086111,0.7975146902569675,0.34873853997196036,
 0.40595795441459614,0.6689929047824477,0.0519762980750722,0.8599405315959562,
 0.6637968176670497,0.7660638705103628,0.42615557402349535,0.9937016710313935,
 0.48411286242652607,0.4050712046399024,0.47246245595017744,
 0.7570839455298647,0.12169215555851076,0.968195053117713,0.553090283615086,
 0.30538666497318134,0.3589433016995689,0.793705644146018,0.27811084905201944,
 0.029025536691834053,0.5723897272077106,0.2939981477644221,
 0.5974865884709125,0.01866371094589514,0.5122332220473725,0.8964073240779686,
 0.516613501177334,0.7496327867268328,0.39663535090561586,0.28270096952926893,
 0.39726063338484563,0.4256371059941598,0.9841377868812121,0.8954925204185699,
 0.9836662404115819,0.5752379594943504,0.4981846353819219,0.03577135918584318,
 0.30919653005590275,0.74122289502954,0.1326736431866532,0.22201882911450987,
 0.38614497814524995,0.12415996448634858,0.6600921765689148,
 0.3905750063066121,0.38461533509544177,0.9560117169488442,0.2217516382125787,
 0.8165927525746766,0.15323942239238442,0.7043238325854557,
 0.39549554297851386,0.674260856855114,0.26805153405653503,0.2975116669771385,
 0.9684752004657028,0.9043814565759675,0.7249966404553182,0.9745917483682907,
 0.0918089854044759,0.3127898637486798,0.9360810458718994,0.2958597667417461,
 0.15277167191775065,0.47358116549806617,0.6760503116387352,
 0.9597492957778786,0.7816548070513168,0.8944170525779014,0.21345772345828462,
 0.16215071234452927,0.5705624771892863,0.8926022040789813,
 0.034713661257890216,0.9830707125893636,0.5671362870010517,
 0.7823262636698958,0.9184353082082303,0.12496257830706314,
 0.015697973331545656,0.27023604144059277,0.4655830952965947,
 0.08459107445245295,0.00459613950703619,0.07756690213323625,
 0.5556307338719002,0.9958978715427499,0.21397847031909456,0.3915505403227262,
 0.8231349326037388,0.21334598425242401,0.6788426806209984,0.9227816550410999,
 0.22250552809924828,0.17896159634577014,0.4214005853678229,
 0.5578025176339284,0.3283930725472617,0.5699274038627393,0.8097325735596151,
 0.3636632742161039,0.32087136870541266,0.06825656877751607,0.77450948452631,
 0.375697741325146,0.012130271588203678,0.35399785581255216,
 0.6292552757567234,0.03539966309118392,0.25934477284606683,
 0.07194450967291588,0.7111416248414877,0.022004226729234966,
 0.8194257485366572,0.25696196673462157,0.4819716614957805,0.4848763343994512,
 0.12172550435657803,0.22578077192443669,0.7964302322294399,
 0.3322377045614102,0.586279809909337,0.22967185319395333,0.3882539933541084,
 0.09869137436911024,0.7409432392400351,0.6374601556695058,0.8127691122572144,
 0.4772432932905748,0.9786329575457683,0.6091051767491467,0.4807636921086421,
 0.5657297441260674,0.7194283238334548,0.7237083163996566,0.9895968239848344,
 0.7304438088629966,0.14299937694226583,0.2374821684408941,0.9201686768695317,
 0.31999825174602103,0.5752416090265441,0.3961448372705826,0.4708263957971315,
 0.49588304120566895,0.9605223854935164,0.2015081362928608,0.4014290527822757,
 0.3916919432683714,0.3537724159028378,0.17284452011865903,
 0.35556341062941754,0.10363080434721328,0.7992491071230392,
 0.3724923079875031,0.07339216435347873,0.5671891399006228,0.2289820421947023,
 0.6793789915904227,0.36437889067587426,0.3845129413864712,0.9308384874096751,
 0.23707392171861597,0.9935932512284341,0.739136616310868,0.17853306435079497,
 0.6196969293045531,0.8106766849691991,0.9066980723082255,0.3420933996681246,
 0.5947989117575967,0.04446696745048084,0.9578975454495765,0.9847289025245768,
 0.0018637472855587145,0.12232187611084422,0.8771006980311847,
 0.2379930030137598,0.06954490660476842,0.5956091892393354,0.9298574258454392,
 0.7254602018022162,0.7845728853092065,0.6462069283907903,0.9618189349180738,
 0.6645651249906633,0.9815705908984091,0.7645562811828214,0.6072789352216914,
 0.6760479179178679,0.8172708331095605,0.6538911919812286,0.3849278541853822,
 0.008066619751008752,0.3858575770945418,0.061464046971347885,
 0.12068561587706061,0.07110601773806369,0.10948874102819173,
 0.14418280371341652,0.21185852803956762,0.9682054859378938,
 0.6522460824794475,0.6429145625378179,0.4348960176590049,
 0.028571665811723435,0.5928766533728644,0.3395591552762339,
 0.7717149371698289,0.8289435436022322,0.3417851866472207,0.14255764527249748,
 0.18349827927353501,0.5389480869994772,0.9497874846915675,0.3375457103643381,
 0.23072467925957285,0.6219067198454789,0.05457300868086945,
 0.03485849055512813,0.13648815512782952,0.5275775392433014,
 0.8409691527766053,0.947969532162493,0.7862083407272273,0.7004346891690894,
 0.21303198653771394,0.5215218444910991,0.695091587611285,0.9002094439699845,
 0.5629084288027015,0.0027017732095021563,0.9097054758062791,
 0.6626028675876666,0.5179924114344673,0.7984261362765551,0.10602877867547,
 0.3170918420928176,0.11839458241182599,0.7245869446140213,0.7101544322085359,
 0.9962818000216123,0.9736683591879368,0.3229353420586434,0.6812153851352323,
 0.05247027453320807,0.8436073925221361,0.21728150466690965,
 0.8150062446947598,0.9991833694607816,0.9251412398329304,0.6175534174628575,
 0.40059974247771857,0.5375266152628868,0.8731924408508251,0.2689597913468039,
 0.28553828746694676,0.31018674827350556,0.1235931584405191,
 0.9542082935717816,0.48393547773737045,0.8282777334445036,
 0.34626661063615816,0.7410845738389439,0.13785401380251838,
 0.3490796845126197,0.4186780132007305,0.5095346162474922,0.7153041000957613,
 0.6736848815481,0.2932612479484593,0.39766320261631427,0.5785922418177919,
 0.1957763543138864,0.8971352442300824,0.28985362286571437,0.6999925211832523,
 0.09161423667001367,0.5602337095222302,0.9201149267186733,0.4773908946693697,
 0.5028790908898371,0.8787146648100974,0.48413285733044154,0.8638485169127236,
 0.4911409951731769,0.3065941090445624,0.0019038960024192475,
 0.3264169923128123,0.35225854915258237,0.7758896686771756,
 0.09622171197400697,0.8715502336490013,0.6333928821463773,0.7934731975560879,
 0.6633787504711344,0.5706269172005344,0.9299300688474914,0.2955902970017654,
 0.8745138678570866,0.4755719558339413,0.42478167665077493,0.6334013896425894,
 0.7217411450085277,0.796175158136189,0.5078764001285287,0.47137397839634854,
 0.6690478344900538,0.23570466732128037,0.6089989544378409,0.7886349734106717,
 0.4507735784698944,0.37061508490469597,0.09827140473955787,
 0.04844326623047435,0.9317418249600605,0.9004864567727846,
 0.06833031433822712,0.35382562536739304,0.770796347336283,0.5327716815652412,
 0.050604798367032244,0.5255691881742416,0.7377881730896327,0.907798307083966,
 0.7278818472372052,0.45467930837222603,0.4212144615802331,0.7868134367676193,
 0.06583362409106164,0.012585186805164184,0.12447712342562933,
 0.5095544214311819,0.9076967190172354,0.2063982268469412,0.52131473771881,
 0.387021890516306,0.7332100415303661,0.8159790123785267,0.8270154974482838,
 0.3758606845994119,0.2609477851171592,0.4138893720579946,0.12993666172509244,
 0.661863960216448,0.3704073838262212,0.7024788020453364,0.4145086254474253,
 0.7708005064837278,0.948356357559924,0.9982171803439669,0.2624748913114435,
 0.41829892799709034,0.8963662686838562,0.4265788101526701,0.3360383041636329,
 0.9610923152581708,0.9207925471489384,0.93489996543315,0.31079251658341367,
 0.15547609968312026,0.5585149801881659,0.12096462339108371,
 0.7969622222129571,0.16373353914960376,0.9033017308361632,0.7440605136194925,
 0.5642941894419782,0.8326174821631573,0.9007758447547883,0.15206554822160712,
 0.4302981643457946,0.23060449149932905,0.8934026101538113,0.6138711417755975,
 0.8108120497446554,0.1621020519844364,0.2808101323330199,0.4523770725653069,
 0.8669378856123017,0.7179745960249987,0.33829010590658526,0.979535765181871,
 0.9248607915005638,0.8101266595299609,0.8723122636960623,0.17232073793346614,
 0.4378998677236655,0.619784582698692,0.762286808252423,0.3051300174551196,
 0.7677473677385297,0.5422967409588623,0.13569315241380842,0.9656433028339992,
 0.34919667093306817,0.2556787953801476,0.8376063881186917,
 0.19208605017754532,0.8878734463643843,0.7977634563153981,
 0.016440130279682053,0.00324734261934978,0.8644938988978952,
 0.7289754072975589,0.38816796269054465,0.12716821097206776,
 0.18582552506438377,0.39891777418394436,0.0922705174332138,
 0.3120559239711128,0.22422265370607064,0.5224987727826264,0.6464615682913977,
 0.020528178666423202,0.27967949365738953,0.14149261779096223,
 0.8203234059154016,0.7807675083780155,0.1046921846237816,0.8615542372732504,
 0.25941508091120946,0.6209677578665812,0.12451903588956359,
 0.5632566025045727,0.2573460245517645,0.22398227367003443,
 0.14932149648523785,0.5015427248735203,0.6771918502113523,
 0.023161428463427725,0.45957522281003604,0.8348753785743676,
 0.26898810396574424,0.9482154110582516,0.24239587837894128,
 0.5891238352523213,0.9144784491925658,0.6385463328491114,0.926068371392863,
 0.5193943661198228,0.3671040425848582,0.41059756573280615,0.8899668240623156,
 0.8780472111385548,0.9282286453128916,0.8778406218824041,0.25321954111513123,
 0.9091634655833509,0.11003107322876282,0.9842738855892464,
 0.24908802041316747,0.6882850217967672,0.4157947435123448,
 0.19991024865269758,0.731777996783208,0.7756064816440613,0.2638997095740394,
 0.425340230938387,0.8100378839634668,0.9312335512781047,0.0906495644310018,
 0.07956205161543628,0.22093087123891353,0.6393023481548492,
 0.1565361716924819,0.757133415822175,0.7400627179118449,0.2193127606885975,
 0.1922583387674004,0.06696241056271612,0.27081975459927077,
 0.2692799057228692,0.3238237021881629,0.9029098428600529,0.37747970731009206,
 0.4053185047319806,0.8816068096453032,0.15948205949825434,0.6895481221854594,
 0.8715553876886297,0.008645381956392084,0.9553854975359104,
 0.9514672388622563,0.33144146725369494,0.7427734588742287,0.7896351827974991,
 0.5397765970087043,0.3839067638296313,0.4385649970029236,0.34748103029398947,
 0.8896183005528717,0.3769949527994605,0.41651117529330517,0.75293201605854,
 0.12413361468601103,0.8656650113638584,0.7325771264714003,0.3030510360359693,
 0.37109801174143686,0.18100736929867278,0.07464723537913598,
 0.3831071816745163,0.2214135678502014,0.7670018072881755,0.23110079107212456,
 0.4288453556380698,0.9589503324414735,0.943787332407288,0.8016332279124276,
 0.974716190417938,0.8280325735642846,0.44251627469295607,0.3295699891436742,
 0.4260254751615027,0.4977037379796543,0.3747970410051985,0.8649297089703859,
 0.6709367895665829,0.7771765043121954,0.2429629926560406,0.7463346450523073,
 0.1966394593519033,0.7532520547830405,0.10947072111513234,0.2048892589435286,
 0.7417357462135608,0.21843343321462672,0.4120054752329373,0.7656424300413422,
 0.8165704710447049,0.7286250993333068,0.8773107768805705,0.12839890512165275,
 0.5915093588276831,0.7577091228506201,0.7091132718162092,0.129183266790996,
 0.20578763883682472,0.22606600408054223,0.22603583154245777,
 0.21759732678914168,0.749210521750258,0.47578932522751427,
 0.49274724938034753,0.791803330517369,0.6096345617927219,0.4939368410797219,
 0.5431184256392493,0.09949137914154982,0.7247007099085541,
 0.41635318254354703,0.41546662018770397,0.511524305632617,0.4884348432245249,
 0.8476066784776184,0.21293426785730296,0.14241725901224156,
 0.14925608557867842,0.16657374198886388,0.3163150849140015,
 0.5892460656598533,0.1993674751598058,0.9174191611561858,0.9726393951600647,
 0.39536983500657974,0.9615291782982147,0.19401417744816596,
 0.6825794026896577,0.5290819153173025,0.6392656542343762,0.02089564160145818,
 0.7036589922303917,0.4224523101691371,0.6428074188798285,0.7030317927462352,
 0.526522078528799,0.9151766605693208,0.5112249358745782,0.3137457047349298,
 0.018409442287083744,0.6130736397990502,0.5700567374062849,
 0.8867226343134814,0.08031351474143122,0.4912586947749974,0.7900008282108826,
 0.08604020019513958,0.6350053282732352,0.9930633654414665,0.8891162708636231,
 0.12967305893178638,0.7321409443727003,0.16355408010074868,
 0.18255003813841642,0.5702411217011755,0.7939675995075117,0.5708502719347998,
 0.5959642178790441,0.3258212679927919,0.4199429999561688,0.67610002369176,
 0.769621944837806,0.39077615498010365,0.32533069624561617,0.7274684083827377,
 0.3026976484156162,0.7025098006169863,0.013848946431917475,0.369895085084892,
 0.3434620981795067,0.5833208423867775,0.6053536788383858,0.16624399024063752,
 0.7222865307699904,0.5779986482560348,0.33441695032016516,0.9817215692624875,
 0.06663422637988603,0.846928611405078,0.9655790165960788,0.7360310369074187,
 0.11311126663620974,0.8855139778113239,0.6641435512484143,0.9542471106429453,
 0.12768601785854872,0.9645880527664165,0.03213977902351117,
 0.5741949620784216,0.3722086926096443,0.9512118294991446,0.3618276844824989,
 0.3993769393294673,0.16300497398692582,0.9834096511043736,
 0.25579287884135804,0.19444867945490218,0.7312860787605884,
 0.9327482185684862,0.6468864376107157,0.8802532394204504,0.6978556368807115,
 0.9163103941474939,0.18993559618974543,0.6880206671960379,0.878448191328871,
 0.7685294861555086,0.7867254365580534,0.1209776855639556,0.5744370919838824,
 0.4083550012127334,0.683817161591074,0.098301743055625,0.18513342696161839,
 0.6988133537459856,0.6691188624143204,0.25310770551607586,
 0.12673348960405562,0.09416026083140216,0.4302853965258332,
 0.2617069293946661,0.46100186447752334,0.4795070959567148,
 0.13768639824794882,0.5624723108924689,0.8600014902255548,0.8676152594189225,
 0.6503889726347004,0.7410911291772895,0.828663981374852,0.45958974600163083,
 0.17272930713844703,0.20731293552251517,0.7033289499941925,
 0.5671323157748996,0.09731802689096292,0.22520543064909526,0.7233985393332,
 0.3766453019075562,0.5432379585764375,0.9585053116438785,0.917196039490823,
 0.9367927320278675,0.7722557110257398,0.8314094907132817,0.11635481879981247,
 0.3141065300393425,0.6838414459463844,0.9351881912671638,0.6802053117376722,
 0.7868768765754522,0.22419816873922382,0.09788807945760014,
 0.7853986572095509,0.01945086132845575,0.8581934422174843,
 0.49081401855205464,0.4319449037438332,0.13198016705897475,
 0.4761475248282252,0.6598551401218833,0.508796419625228,0.2363080440406472,
 0.5044118543719133,0.6104250791393125,0.8256291868765981,0.6151559474401653,
 0.6150286668764547,0.9557214259141159,0.19316085281958795,0.8839897437859877,
 0.4354558883225397,0.6063842186969527,0.28278559816470494,0.744585666387547,
 0.7376393286099047,0.4165578276673382,0.15534966465984046,0.5362555215060147,
 0.008232861201640906,0.04251633346344719,0.7982711269274004,
 0.8401805987517351,0.8887435418412543,0.005432759179986646,
 0.4033880363400766,0.5943313228886549,0.5513840979192113,0.7780041323007245,
 0.8246296768356596,0.7119511757433942,0.9794310823761527,0.665003911226465,
 0.21153539760985351,0.6854990696414444,0.5606267234714841,0.8782490618451916,
 0.2313152272451393,0.45613670991602784,0.17768097626917856,
 0.1183459811000247,0.38512118935374673,0.016165397301647455,
 0.8966218972745361,0.431376526705872,0.31434724252595203,0.5127241178780206,
 0.11160506182634822,0.5932009728547653,0.4263515158041663,0.789226126795752,
 0.6098420690057924,0.7539878664605364,0.9358744584454043,0.5956389264767644,
 0.04899186080447304,0.7335814661410783,0.02085831169429697,
 0.8655650374976059,0.8046194547137944,0.5974612973957462,0.38939916156676413,
 0.08080726487861334,0.002265420421551645,0.507810592146997,
 0.20756941181020494,0.543045140350906,0.2586867226177624,0.3485515064940309,
 0.2823959869922894,0.8268978752829107,0.32864393555701477,0.0991895886225117,
 0.5626440943292091,0.1385355238530861,0.2909627895323025,0.5925228065741048,
 0.4731845091576805,0.46898033295280384,0.8931559173364816,0.8402487459203891,
 0.725049329862455,0.2028166811657226,0.9507439769001286,0.5473807667980911,
 0.1058742984047003,0.050851742584079984,0.20736027116864042,
 0.044860331635137585,0.8208410230414576,0.12057665756930613,
 0.27479436150976433,0.8782478806548858,0.06723771701448389,
 0.44499599680527757,0.6837094034703812,0.6137114364554656,0.6213648203466134,
 0.3011451273245911,0.7664652889569942,0.7428980057127833,
 0.037741872615495775,0.014596202331983221,0.6052548523897423,
 0.0018867271371501015,0.499657545422892,0.5037000135382826,
 0.5304989593417014,0.8751795881493284,0.8250753735257899,0.8956084813843068,
 0.7319572486212884,0.7226679119597653,0.6175641095040124,0.02085930332342023,
 0.7363405395075922,0.299288751360243,0.8866192381484265,0.8319244537578805,
 0.3908295046537631,0.8987018888883299,0.8521843571544225,0.9473007391710611,
 0.519134230076864,0.07197314754196782,0.654917390901103,0.80032305093159,
 0.2459807272329293,0.738891187901773,0.7989430999629688,0.6680589268580738,
 0.4832785786767986,0.6594608891279803,0.8124207857645838,0.15091965270473806,
 0.08206017228742735,0.5901618194895246,0.6606780253535088,0.2092991222531535,
 0.3631584693975164,0.915809374423524,0.12784240263366142,0.6914397118090755,
 0.6091426777621537,0.46559339194212657,0.9897556240473603,0.8045135197495846,
 0.8021078005420594,0.4160667201278898,0.669946801234534,0.14464796661645873,
 0.8819132526456266,0.2806211210319127,0.23926463010868537,
 0.024897541068081286,0.1954968910132272,0.7120486910487097,
 0.8274942602414223,0.5083273640294438,0.08523133660004789,0.2776218309259867,
 0.22085071958656455,0.21051284960882832,0.21228024020956815,
 0.8718640081847012,0.8544297234976914,0.9283537404061981,0.7644368732733956,
 0.7851751296955484,0.05687836583423378,0.7396192527651329,0.8419043232323021,
 0.5812449876711506,0.7115577624038618,0.18043011321217517,0.7575840681683832,
 0.2360650214374439,0.7207687662929042,0.01047473049589942,0.2270768911718679,
 0.5023759050957389,0.20170024549250032,0.09776243578657684,
 0.22454642483380882,0.9702491380417304,0.7050887659289338,0.4783696768248801,
 0.8337606530816368,0.2699246892339717,0.09024776833578074,0.9449143594395742,
 0.28586399828493936,0.3939469271339323,0.31998844218279765,
 0.6668534028391164,0.5027783166359745,0.4279038107537082,0.7610232767660726,
 0.7885777022223558,0.3461842814310383,0.2716302011056142,0.19022232155808128,
 0.7237596574352899,0.8971532547422132,0.5580429113074666,0.3588061061980721,
 0.34821299712372167,0.6928417808817946,0.577510754760343,0.16952324329785706,
 0.4374715478070651,0.5863101167346358,0.6736816603879197,0.22361670549272306,
 0.6830590995629117,0.3444872742770976,0.6159783603165133,0.9277558287070293,
 0.2979215078712034,0.39112988009795024,0.6382508596954195,0.355298803414443,
 0.18537307963977268,0.5257223205271466,0.32980553419634195,
 0.20222873339147607,0.05329783508364694,0.7895760841290088,
 0.7955297173655982,0.020797354889896646,0.5563502093602986,
 0.44030780361900157,0.1457277730927551,0.40234086340184394,
 0.35641283815131064,0.5334647915814114,0.4700771415563787,
 0.06538650632461507,0.8080956014573722,0.13226402065586162,
 0.5669161181356273,0.07730773893767429,0.798694914713672,0.9075181524406235,
 0.4495189544430246,0.26755874851103123,0.8994845048856972,0.5542074808335598,
 0.41080058322258994,0.6326795138230319,0.15435350573078555,
 0.6950918012084953,0.8150352922497004,0.15087840546790077,
 0.37137490750912194,0.8562893743159401,0.5040749375242426,
 0.26809376688304387,0.7010432095603241,0.9314103107764711,
 0.06528842299328252,0.837455216774678,0.7765107539442831,0.49058970409681457,
 0.5825878910151725,0.8417222100208873,0.9748893345053871,0.42568203756913636,
 0.8857585703419599,0.6104842026357001,0.5229504730675518,0.7766495104200098,
 0.8902364419770404,0.5565787972850141,0.18789872838343336,0.8028814209420813,
 0.9205377511269873,0.7777995989573075,0.7866845626296625,0.4124677495850262,
 0.4573921920482311,0.5351705530477223,0.8109939891223434,0.6027989538537273,
 0.6706916330375526,0.09712046265064633,0.49620255330325913,
 0.23658045971420816,0.7461694049420026,0.6759615090351732,0.6392287821924324,
 0.46002922192355755,0.32433606211254506,0.9958208233463052,
 0.7638756052322104,0.9720694368112965,0.024699119065851316,
 0.07575089490577913,0.04440673032496578,0.06553406279521512,
 0.7820686762275157,0.3908954379088122,0.36661093308271087,
 0.20177946812269865,0.53340959039516,0.6280200654697461,0.03539414707050714,
 0.5382589754192764,0.24326624641850725,0.858987472795341,0.9853191057410059,
 0.0028654980822747644,0.7767149466781842,0.559485561249703,
 0.7787604520856505,0.7271057744477973,0.9528944010006126,0.19659559909387692,
 0.019156775287522687,0.5311740214949545,0.2862847626654895,
 0.6542050523275188,0.9214773471635784,0.6030156539481263,0.39288015068233584,
 0.12874530723642796,0.31396818661309744,0.8997889822654725,
 0.1878478313642986,0.6513866524432111,0.403152301492659,0.322471900950533,
 0.5948016629597035,0.8137168038848877,0.03428848561550568,0.8098388497192238,
 0.3225932443402897,0.2569460060373463,0.49551852022214615,
 0.20051213359476172,0.8824142482988212,0.7835796072221323,0.9005552852699538,
 0.61270393779533,0.2627355569901716,0.9452176666661866,0.27075842982619835,
 0.037591255902248655,0.3410743902465958,0.02674694609022854,
 0.3190584337801451,0.33139409246324025,0.560704143407642,0.20116237279532168,
 0.2129032914003356,0.5193908362955649,0.5425243870756383,0.8661695849090366,
 0.7354180199643896,0.7480972064175214,0.5956916729533941,0.3472619808039893,
 0.41422430341195726,0.6364375063154326,0.30293443557047006,
 0.3700427534081786,0.13788716966469083,0.2978976224314638,0.6136324093764627,
 0.254692286135463,0.2251762155551731,0.6170918740639824,0.5131593011060156,
 0.35307011314606973,0.10863116897980796,0.10939521705957056,
 0.07628560756887237,0.5026625968870099,0.016222412040237666,
 0.052460872609022324,0.15221299967760404,0.3861638338219351,
 0.5054598259077292,0.9641309866072866,0.6305597204775607,0.5852256451515812,
 0.09301096562849365,0.10586343853171687,0.004409112412238181,
 0.12033491715251465,0.35851876701832464,0.4542631496135484,
 0.05476385832257469,0.03674750793140891,0.01794146549925324,
 0.1827608921668057,0.1757110592203539,0.7398575615568258,0.8577195703630185,
 0.87658546315813,0.8193435158822804,0.6176914405548273,0.9772885953881942,
 0.9012440953457368,0.7698142029883978,0.273768529322185,0.3684042145057347,
 0.6158194803421315,0.029215974857266902,0.6520560626219302,
 0.6910188535979701,0.919707861946972,0.3898375050531082,0.7737554317632094,
 0.10237254419071817,0.9618005008439655,0.4998907258964652,0.6532504680412454,
 0.16520635204092293,0.5235217266634229,0.6560149701233323,
 0.019056860408892984,0.04560268685435043,0.8222071137197292,
 0.21914454707276632,0.0646496616173814,0.7319031114651776,
 0.19990089991641047,0.7225106344229046,0.5415976898517687,0.2909685509510441,
 0.6972079921625417,0.1510456588265099,0.6831826229819051,0.6702582182946046,
 0.37269924018559153,0.5935985776834098,0.4516626877527048,0.4831213611517552,
 0.34398859054827424,0.08119040602860461,0.0898832002541502,0.301848231644947,
 0.49087820062318466,0.3113778205403017,0.005699087477656017,
 0.5170078753477843,0.7540243866653552,0.3642070473983863,0.9709052019136353,
 0.8314330330209568,0.305830117134948,0.911695096715274,0.48379813109236547,
 0.13686907635107615,0.25294338385882054,0.7211931469880855,
 0.1523714146450731,0.7865573080601752,0.2400360540058588,0.872856670602294,
 0.047229643027630774,0.2907137458481982,0.18327665610343946,
 0.2697346286982877,0.9710997936596688,0.35202559088775853,
 0.05611928094890495,0.47970927993824275,0.5431533025241406,
 0.7819078191956853,0.33810567913648004,0.2584921200978928,0.039372371644536,
 0.9409733262970031,0.8432309787861382,0.57379896264697,0.22481259425699884,
 0.4587096325513508,0.5960505484671526,0.9598122574238868,0.04750768322655463,
 0.2554932849074921,0.4393087608042009,0.7025686774684554,0.8990579885785979,
 0.43429567175247097,0.699612250140291,0.014229880367177206,
 0.2126684953248138,0.22069057372496426,0.6670203143778927,
 0.44094074539960526,0.9627735370031733,0.7258626745628998,
 0.20002250457161275,0.1699692929913702,0.3836910364919526,0.7597962609955525,
 0.26668038741660627,0.5496815648138718,0.3230677035046332,0.776112968100269,
 0.767129872924474,0.9544349132844813,0.014563758363480561,
 0.49665422066548304,0.15994991238277456,0.47078706036336737,
 0.5581792403229977,0.98213361755414,0.07070921615575843,0.47783045328538,
 0.9021720510031819,0.6702710759964184,0.17494530230644656,
 0.43852381406294527,0.720142373770159,0.7249384110406009,0.6264013301074582,
 0.10251125061028943,0.4682079025383564,0.5865502147848562,0.9679159325945997,
 0.08632024419796835,0.15343034095657881,0.8237325743262597,
 0.29552647812032684,0.14640061846461405,0.8776743348839472,
 0.1874518486252985,0.1053808944317719,0.9480550897890299,0.7815260163282683,
 0.7935372188574461,0.826215462584747,0.3423780441064217,0.6666084261340023,
 0.9076076117985563,0.2461960239619143,0.655357752639999,0.3840589221513624,
 0.740579316317649,0.8595701054096048,0.41450586738075923,0.4651870032182803,
 0.0700526471248476,0.6827660008834415,0.09001373997547302,0.9747842706685821,
 0.6329989155865388,0.6797285157021163,0.30919250844658386,0.5951581595826536,
 0.624072381272293,0.1540154678010297,0.14037731272106413,0.8322411392923259,
 0.7307478780522296,0.20214385812891367,0.7219020291723128,0.7214948499846625,
 0.3583176645938888,0.9478318615030538,0.4561302726113494,0.5430026111978971,
 0.2521334855867772,0.6760942935598218,0.8036901924614848,0.07534098646582854,
 0.5532751792150701,0.48566021580998786,0.9069600564309096,0.6601896305503921,
 0.1712898751850579,0.22086961810268257,0.8454699082362745,0.8551978580942698,
 0.7567745125504698,0.06157376829054162,0.031748781093823175,
 0.17066801966901135,0.4782275048645359,0.24054204011286118,
 0.06671935148584451,0.9333887781912826,0.9713867620346303,0.0939260756190532,
 0.8803569655789178,0.7559822699353995,0.7034758477360341,0.5211648524870333,
 0.8878557208375012,0.4806018302477443,0.5057706552744069,0.9691945633343293,
 0.6771218030592653,0.8697149299004129,0.17674257771823987,0.5813658888263853,
 0.8263122088841772,0.28727781200078295,0.9280440808572287,0.9856143106631884,
 0.0657128762335718,0.9069277866018839,0.07346315948122184,0.7805769070291337,
 0.12167920914231078,0.3139826015392265,0.7599951088010894,0.7273087431397683,
 0.5591573649110276,0.15659977698023497,0.20392745613333574,
 0.9080512382052257,0.10543082952588556,0.502730457986274,0.9314372675121567,
 0.6175539085221046,0.7197627727819788,0.9201268945620612,0.3975813915535824,
 0.8147980746818312,0.2829184086424614,0.5529141531633104,0.04630437702205903,
 0.33133231127449514,0.11970158435054767,0.740183005550047,0.0888100974273286,
 0.9669048717272479,0.9031168404403447,0.6221460931699749,0.9527578809703086,
 0.33254338170138464,0.9340249154969139,0.38772491928867336,
 0.37647533694918045,0.2855269675097625,0.16062098577375372,
 0.023183521957484854,0.952045434817116,0.6298754516354323,0.4497366343381156,
 0.08407703020065738,0.5138866723767171,0.300711094279841,
 0.011237398964722534,0.4854803093065475,0.9603697449156863,
 0.9534776066896024,0.46975990385544697,0.05668122973801648,
 0.48594055638270817,0.530793866466867,0.08841112331696921,0.5057569482886467,
 0.022980248208625764,0.6027203594022995,0.04133117212034798,
 0.36909137366787603,0.8404050546099369,0.21526615189095366,
 0.8123184900678548,0.704995893154075,0.2414911650766578,0.6673206355249838,
 0.1949408027506272,0.21536693770424997,0.4512481197745618,
 0.12060835785176949,0.9150471894088581,0.5843011800644845,0.7660080726810787,
 0.5480701877279057,0.5507287146088309,0.5751371576428,0.6551728949999632,
 0.15443613731221872,0.6817532516346674,0.16753631452459095,
 0.5416135220794543,0.05286117673823021,0.7730054320762659,0.3236865134978588,
 0.010492433094538067,0.5070847215546522,0.6106631361949066,
 0.3324619575865654,0.1285967418965629,0.20472425396158733,0.8565694001626317,
 0.04781591289497211,0.7251372610614694,0.7927085887690414,0.8843007026235084,
 0.22033051681326876,0.556275777716715,0.4206613462147427,0.1224922753511718,
 0.23293759521245494,0.7417021978296202,0.1129935153596705,
 0.12122676444288538,0.5071045409035697,0.6782898162181371,0.7167460443197895,
 0.44527271485276143,0.7317827547259492,0.6901487257618109,
 0.15173930319652795,0.45940388125717413,0.43878150787876025,0.30715863561042,
 0.45979865775321194,0.16558404642476243,0.1098738870765894,
 0.12018002079277379,0.346500605969154,0.3051199582549933,0.9949708646763316,
 0.7000550820778382,0.3044336996596042,0.1843799199543814,0.21601818847178023,
 0.30018010023250297,0.38754927685285556,0.5593294534509484,
 0.32618828474164685,0.7974718616973666,0.4236031163533436,0.9867921836367798,
 0.5551641959769241,0.8894238280244238,0.3596129200111151,0.220793068390029,
 0.17420577511191737,0.1997623764211346,0.28349098123984084,
 0.31324610433997846,0.4182373355513076,0.024355109427169563,
 0.8838390999488916,0.2639167554016719,0.8852060706504097,0.6532737093047034,
 0.9455143538945251,0.2658187454021945,0.20916462212524312,0.7546602157691705,
 0.7347657014085781,0.5061021667961304,0.9669612725694678,
 0.026348680219639653,0.9668502015378562,0.8774254649902758,
 0.3245733931775229,0.9864407850019684,0.23000070859036104,0.0719100832222278,
 0.8919504152810426,0.14025176445858767,0.8643260786282232,0.7458978511230016,
 0.5860579881786846,0.24314392523638806,0.36927852756646207,
 0.7942363015231552,0.6697520195227744,0.4851767577169914,0.9817321197437466,
 0.7076110578689546,0.36538005629571435,0.17488522515273286,
 0.6030784104879484,0.024185561834812908,0.7577494895024692,
 0.19770671280878227,0.2038933407262825,0.5477699228364299,0.5783743344626606,
 0.34162626121325257,0.6966148048610843,0.4343128273332748,
 0.22222513022510515,0.5434468138371892,0.048019569722565714,
 0.7313131070181368,0.9642590228837837,0.5461951772304195,0.41497489775159035,
 0.5353379248678081,0.32935871223421154,0.6219704448953157,0.747007945250118,
 0.9500241223759025,0.8458824343717654,0.009055561694837011,
 0.02380887515442931,0.9912290266698511,0.6062209725448179,
 0.32307939124349483,0.607412558518014,0.935757753941108,0.6023057544035785,
 0.011822381164818707,0.9576217054102032,0.7282885144493767,
 0.35227629640176217,0.17030150068692895,0.12471739169263107,
 0.25584088134971006,0.7486010731677326,0.3831876701969016,0.9403588020101257,
 0.31870953971341354,0.8964150072346101,0.4938011630062279,
 0.15141248704394517,0.02252141340457292,0.7864443217672115,
 0.7256039206971658,0.5292995675661412,0.9756989840502329,0.2888881353888292,
 0.2770836186866088,0.650156929571392,0.956500750257218,0.9841117641395751,
 0.39989003225963704,0.5111737401878499,0.32396849587206256,
 0.0411442287118452,0.40662774618906605,0.20044071914070138,
 0.40356513612072975,0.02907924885869073,0.9118725540230107,
 0.8015827520086709,0.024296967804353198,0.633123607488441,0.8022007226260864,
 0.8491678756498752,0.5720429661065474,0.760986893869106,0.5060506733333126,
 0.42049992747743714,0.04602489817013822,0.9732332352314249,
 0.41063914823310943,0.9873996290031588,0.605205161906095,0.7609227995380454,
 0.5447127677470844,0.1801774934621153,0.818941812240632,0.1914565179500829,
 0.09801846653754687,0.8556068320577388,0.5776186802525805,0.9093775855998869,
 0.11760493672833239,0.5928602688770306,0.21124467985820505,
 0.16668235926696107,0.9774402131525642,0.49611148014577433,
 0.9502487102573746,0.6640703430287669,0.25183468445210444,0.8542216486626041,
 0.7146738482557542,0.7197264834038635,0.018610634378033453,
 0.26652310357873654,0.1491356550530779,0.1898612644968033,0.4091219118852256,
 0.18442573502357273,0.5288860929409414,0.8325668268139785,0.7057152475787613,
 0.6507622320054988,0.7319915315515373,0.6826784081161553,0.9636338306325949,
 0.09939387162043944,0.42418354452471285,0.7663696103428231,
 0.15703502095650068,0.9970552179114354,0.5053290583363097,0.5521149540020215,
 0.6978505903592647,0.9625181351370882,0.9050455923728548,0.5145610442500504,
 0.5041341923833557,0.2156173458386691,0.7140730807704412,0.36040467210585714,
 0.8856251488741615,0.9933123546933237,0.5878364921076649,0.2228064605601775,
 0.7939884687473445,0.22324523588084277,0.8665709762352535,0.9664580158123443,
 0.21629112055910626,0.014727982419577512,0.8243404684499911,
 0.8706639388744712,0.8015027180904694,0.6309469319063233,0.400092078703988,
 0.2942034777642186,0.44815017837653515,0.35546328109949976,
 0.8558178814865788,0.988052587006135,0.7512182780763469,0.40743034425997093,
 0.3820860255055336,0.10395061702881048,0.12564458177120608,
 0.0076883834555550035,0.6146042568774622,0.21349140538593936,
 0.3553997890519256,0.6569803753377763,0.7456448228163265,0.24676012039688366,
 0.2627952545524568,0.8595687398600111,0.5967869473047099,0.6304265470327826,
 0.06345625045171577,0.056408947413454036,0.8140191108263231,
 0.1828605158581228,0.7094377070259806,0.6438467214965662,0.09911264060545122,
 0.30962857465819094,0.7932982637438923,0.26781256585560465,
 0.7041573875947603,0.23270889988571575,0.4689099030802617,0.8182323703157253,
 0.33247804927288804,0.2407749506755601,0.9001204595735118,0.5341472953068576,
 0.20826478755612254,0.6998687417824409,0.1610333750047308,0.8341806185286481,
 0.9045358815819621,0.93955367982667,0.16566760497604105,0.8978070285659271,
 0.6816151309754214,0.014169476735834019,0.8197480081009632,
 0.5658835038983919,0.07963948266424004,0.6072093112282121,0.742770836238219,
 0.07667190476306418,0.29231474173711713,0.40634287500024024,
 0.04021264725024942,0.130482153208891,0.060961740330929826,
 0.44527892895648635,0.6387218939088212,0.42262675001508754,
 0.4534133658350935,0.24524858383333314,0.1909721896279415,0.8377280827479645,
 0.8457268964082367,0.3695248597416493,0.35685582010469863,0.7819141197790377,
 0.49459414647439626,0.5144710661622585,0.6971460434251965,0.328704455372512,
 0.031048686470977915,0.9965749261741628,0.23964674356970572,
 0.16661904399829008,0.9873393878897173,0.28390317291202094,
 0.9743780483913732,0.9884449586838158,0.13936296702649553,
 0.29777410229747237,0.4202080490698745,0.12497405204849421,
 0.3067531385282408,0.30365064079658455,0.18521506805653076,
 0.1421871947575577,0.25222628941884184,0.06459412357919925,
 0.07501742308359904,0.23403225401594563,0.28442861389126295,
 0.43786370928594065,0.7234486520402281,0.22299439443849844,
 0.22027622342361908,0.7268534862502537,0.3447290960368672,0.9430519772138702,
 0.05748467061936158,0.36554733403690265,0.8700980113996425,
 0.4017260445941271,0.10718128081973044,0.37425942910730536,
 0.5011619057862329,0.303552737105399,0.7278004889063068,0.8353161823672107,
 0.128871116528106,0.004574876148928153,0.3274904273119833,0.6878583540823222,
 0.674372477405487,0.4683981321463224,0.46577908338700924,0.5835573655540214,
 0.47905847323269934,0.8255468367596961,0.9279691089119072,0.7028080528368876,
 0.40895446164061544,0.4033308403116035,0.5159774284789089,0.316284117268647,
 0.07453268225968424,0.29111415392187645,0.9650532970347048,
 0.1250124432756663,0.47791446883448563,0.24342971953089798,
 0.7595329933872766,0.9759262586454633,0.7812535858130071,0.35853221912501265,
 0.5523533562646378,0.18439985632457057,0.9658383717847588,0.6358736747897332,
 0.5461598088582708,0.685731645676772,0.07760726006980379,0.6235571179463796,
 0.12383351213064453,0.44048208002334543,0.6543378844831705,
 0.6315853304959358,0.8045266400989365,0.5364048009642095,0.8852106943314324,
 0.8095226895579973,0.2713546853936506,0.3696689508168871,0.1780549249725465,
 0.9960431756829541,0.4724042692628192,0.24830622804084546,0.2577854029820108,
 0.8234298719520972,0.6217679122399129,0.5943731925826361,0.21281986370021622,
 0.6993486692833518,0.6062253984559146,0.14683157629439558,0.8720510357486583,
 0.3678695290647881,0.5343973425539867,0.6701165714573186,0.5444977882366577,
 0.9481180269661255,0.3272812059150677,0.8668379957922774,0.0649035483154704,
 0.48958210187091233,0.8390500910195652,0.0105811071753944,
 0.40037305709571747,0.8351941152977999,0.3038186998513176,
 0.22767299597298463,0.6149962250950376,0.7126841457838977,0.4136433745392535,
 0.1241730752002893,0.9652305775132111,0.9648233737238963,0.3791700605163877,
 0.046692154027783705,0.8478677490271735,0.5287146098123374,
 0.20323068243312403,0.6001182089291864,0.2593514734497564,
 0.38401737562533445,0.8095056573778175,0.05872611153687857,
 0.6021923914806947,0.30301181162115254,0.32556464156327125,
 0.8580477430510546,0.8346652007186367,0.8367023567029902,0.2676193578212548,
 0.6864228738178937,0.45160474897610114,0.6273968226572267,0.7289303761928494,
 0.17814217757305573,0.22982180072079894,0.862662014148444,
 0.48698750249779044,0.5705829537754894,0.5277294392575458,0.4695790571471149,
 0.7344398490475358,0.5271420469481636,0.0581689988652041,
 0.033180400062787885,0.6875343570525494,8.960019989754375e-4,
 0.5054762348463504,0.8836176321916025,0.48121301840754116,
 0.49479095268953643,0.6782944701769383,0.6729995830585364,
 0.058575805788597335,0.621261998152094,0.4890705354414807,
 4.017786255480216e-4,0.09979455692659489,0.07842403064402237,
 0.7929600838723121,0.3964856766183832,0.573951622964379,0.7374101201013616,
 0.5769921240119674,0.7550249985469919,0.8146350583431082,0.6208180428371134,
 0.6855921588599299,0.6798834770037588,0.8360033543788737,
 0.035463435608105276,0.11531970868239905,0.8879278733035766,
 0.11808120333682859,0.8253242215390626,0.3086564046419509,
 0.10791317883380613,0.9935634616290682,0.817627877916811,0.5832112150524027,
 0.8060934367292233,0.69515696628894,0.6096435183206235,0.12386983026027754,
 0.09728248468198808,0.7141264705177715,0.8230689188559166,0.9953220182911038,
 0.5919245675855254,0.39750922146591183,0.043902514814840754,
 0.21817505191289888,0.5372844829538838,0.6061739742749468,0.7572802634621081,
 0.5221326431821334,0.6421536557287412,0.9194291374177093,0.4038895272991804,
 0.5184649577923682,0.6728445582766074,0.6500000158755702,0.07656923234741564,
 0.9038272393026949,0.5554504966109923,0.407691513276322,0.4873455082153393,
 0.1765668204342763,0.1604511397044891,0.052697702769411436,
 0.6745124664665076,0.07129963787139626,0.08674975950799313,
 0.3318692737532085,0.837407553031424,0.15954591175712773,0.4981904831318491,
 0.153514218342387,0.4683659289307517,0.011842383738660867,
 0.41787754510839026,0.9235738321056739,0.8002440894249528,0.5594294070951257,
 0.241082340289276,0.669762848871272,0.3459225379066473,0.8772822801474937,
 0.5349165370785216,0.7073127695056288,0.01577963921424863,
 0.37391979605777037,0.12262541588195375,0.9130581354604366,
 0.40846848294580396,0.9691456862889044,0.8586923421945031,
 0.08377943432878598,0.7063430659710964,0.7297371607250378,
 0.26584437786704307,0.7546770106924221,0.07965154046840162,
 0.0543082961424739,0.3112963702184883,0.1719490006427723,0.8239496117842822,
 0.8491784985703176,0.6299836443819979,0.45935016030797127,
 0.06170192212798897,0.3642842503704222,0.3939973846974326,0.8003242097801385,
 0.0918214400351689,0.0948583387270544,0.26203622046323327,
 0.16603377491595017,0.046977367577027196,0.1779949703458794,
 0.3548425864584265,0.28882608447079505,0.5784350553205481,0.7921224562078281,
 0.388904780152621,0.3642147013950622,0.32681654559100815,0.6752107172394519,
 0.07873432126263935,0.42083033308517204,0.26137697143824523,
 0.2741581322006196,0.5003948419078438,0.3000379208922367,0.555895446632902,
 0.7796293390641151,0.9998173865924994,0.5356859653223428,0.21981748408012058,
 0.9028805869695246,0.7145081133744317,0.06013350475178081,
 0.14928203825094155,0.537575907321806,0.0712745252108351,0.30813687121132427,
 0.209753562427774,0.8247635141432257,0.7815350242100944,0.32143120345976495,
 0.1893759256570201,0.3885952985234612,0.6083325602571925,0.7218766922105577,
 0.21180394779120992,0.20039330225594731,0.34162578047583547,
 0.23833732824665788,0.5451616332333025,0.8650233800120607,0.5117222166516566,
 0.7667398751084289,0.9455816156896095,0.019477095896651697,
 0.20797262529080973,0.05889066757109962,0.44702912941256734,
 0.7526395530890422,0.332283667052118,0.24585025952797446,0.2082110038856113,
 0.8162474614997953,0.3633067674459538,0.30147519838646664,0.6802075333742634,
 0.22763635643842206,0.3565563317178273,0.28624280280270864,
 0.053780951080200756,0.7721589073490033,0.8317015810344046,
 0.1117566144446307,0.6137224502251718,0.34839382435805644,0.4767262904476709,
 0.4745380613105029,0.33433980423585064,0.4719593401574208,
 0.20963074282955685,0.906122161801868,0.052535649116759564,
 0.5151327409449111,0.7403028508352254,0.9573039111519374,0.37215191662750957,
 0.12189710421655553,0.10917299005910264,0.9545621575435387,
 0.09850734695910657,0.016326253355367748,0.6074467824100634,
 0.6397012595859761,0.4329303189825735,0.42520478364387615,
 0.41476165285838795,0.7975897344006455,0.8531599646948635,0.2530009666009777,
 0.24501801990884686,0.0024155488307032735,0.6953631447986159,
 0.7373322597315195,0.30925635262760676,0.027898988592755303,
 0.9667381479452926,0.5335268505310484,0.09049524389035857,0.9270916070532647,
 0.5216905041600706,0.33610835672479455,0.30213600233118876,
 0.9261038118641889,0.9366203595769352,0.22293955726505532,0.4378403617009422,
 0.938907995250102,0.9255755454322094,0.7481897568605809,0.7753985069268033,
 0.050709193274627085,0.8160298583827006,0.9621502077755355,
 0.9883373202477467,0.12285307114760036,0.7558062341888097,0.8759387687958999,
 0.7445716981953154,0.1978791220732139,0.4064627219125212,0.03129718157678696,
 0.46019318217144356,0.9900890090880288,0.20752115804570204,0.950614120500083,
 0.6737423743931006,0.33155714585839413,0.26051371901640963,
 0.10664624819078516,0.36590922360356726,0.7662329676362941,
 0.7000249701693344,0.6609378184376488,0.6764454631934105,0.8359146138396869,
 0.6751603307216242,0.3518644183883465,0.037660187534016565,
 0.8917136148843566,0.9441035092365434,0.046875254950833,0.6270225427403141,
 0.9008356315693746,0.3352184954689511,0.8715533300739502,0.7555704377935909,
 0.2631944197610284,0.8879893280858551,0.8678035740587718,0.1811905446450376,
 0.94274329757693,0.02613632194213078,0.81443809453164,0.9086734585258153,
 0.036319366486648685,0.23333047560166043,0.48928605707492856,
 0.8740184529439369,0.9814057848037949,0.08482438477122667,0.7201780558283889,
 0.25466911195739206,0.134270460732555,0.5452537681661818,0.30954984738010727,
 0.6772594861282278,0.3660414337097373,0.6587451590323568,0.535023848793803,
 0.8470040076722272,0.16034363731216672,0.9742364214831467,0.6959037652937987,
 0.7094339705786132,0.47898447164437485,0.4492852639314182,0.523816209260234,
 0.17779995094643064,0.7942237858289574,0.8595079241532277,0.3038916129234306,
 0.13653397610634688,0.19502575109344922,0.8044405830175577,
 0.3433361413172923,0.6354919762057554,0.6626462282701807,0.1523685200593039,
 0.24083970614103278,0.2113317393147809,0.23053059537725007,
 0.7209583123042439,0.6690331443091587,0.024874086434781617,
 0.32675297478750687,0.5630488330618347,0.6215602975444062,0.9395889772039766,
 0.2171966005826964,0.427142258131312,0.8733867644885788,0.6052175148844164,
 0.342962004523802,0.9389319048311545,0.32269269377206444,0.0959628493256387,
 0.951940491824331,0.6612798349471745,0.9967766064889072,0.6982282623931615,
 0.28212659471622925,0.8355888243469666,0.009664656100466629,
 0.13866052650261773,0.006423935343029985,0.864145565278062,
 0.8954921940991882,0.03231624675347855,0.7863609049116933,0.7905636056477808,
 0.30433715002131867,0.2058355002720541,0.8026448707089661,0.2599180552548156,
 0.6886078819831233,0.3859037594398521,0.7756223244191243,0.14760968791845408,
 0.4576944298174579,0.2047812119845157,0.542393740893385,0.04941470837247974,
 0.6624505126637141,0.8953721839700494,0.8241575088034923,0.15332060688655558,
 0.6765945331714978,0.39335851895163465,0.8815784808952738,0.4573318630824261,
 0.10887606509226988,0.42711122422656245,0.6618647824555095,
 0.014816878254917176,0.6451691764822536,0.23339751467377157,
 0.9051033082691813,0.9356696738485684,0.031965724059441425,
 0.6463823510088947,0.3549417224397178,0.2906260727412022,0.471936883858723,
 0.8550387058104789,0.15412135181324693,0.30492645530172047,
 0.16231796434569157,0.8324121619950828,0.21065286257873728,
 0.7330474360176418,0.4156093158623797,0.029354784145203983,
 0.2672849516509367,0.2080596104447786,0.6639165834588755,0.9863082345689813,
 0.3708455460558514,0.22206345349102197,0.11418714389585416,
 0.22228729973336592,0.24682393320501916,0.20410585505601542,
 0.3341829624031797,0.9860095095188395,0.640371443666662,0.21941295684439954,
 0.8832339739279732,0.3563731238818705,0.7443713864280784,0.4027190687953116,
 0.8365992181432248,0.36493258715215005,0.5688629985450658,0.3767182242816225,
 0.748638647264099,0.8686478841647622,0.2636584564218478,0.73672434818317,
 0.1364951258674556,0.32402284803936354,0.25207196561614675,
 0.6844764785009879,0.9617586397625075,0.6941411758352202,0.1969782615096476,
 0.3841187847294153,0.8749886592076952,0.5521589253489722,0.5970395998711339,
 0.35126824242377086,0.12723113057123903,0.2231541018162394,
 0.8303078969436161,0.019098087778944506,0.38687350358985273,
 0.40368144947913287,0.9190533605206106,0.8485815417455765,0.716440595905282,
 0.4482028632994266,0.4009899424939264,0.18920276308053008,0.70493338398775,
 0.7309798937381548,0.6288808740244164,0.5051437205884622,0.9891814355094208,
 0.9739259549619155,0.9983057349444537,0.5347694326137545,0.5518436795512345,
 0.7293608461256312,0.9372730964921967,0.029340953169437523,
 0.05224589094640564,0.9602444024098242,0.5496629623174613,0.3678913553097891,
 0.7284462121689697,0.8790321825085748,0.03774141190887026,
 0.43147569922658113,0.19458546400416044,0.9522493788415108,
 0.22672067436566057,0.19810996372556522,0.16114036469193604,
 0.5785453048573028,0.49545543528433433,0.671920989692746,0.39859773118429564,
 0.9758360033855487,0.2512323931587812,0.31712414485689,0.933534684933124,
 0.29804501840966,0.9986690342254928,0.04825182842545539,0.6520894425614633,
 0.14191717021751504,0.5421787522233399,0.7513330041923312,0.7484388429256827,
 0.06370623488151284,0.012562977539128828,0.5132901717445725,
 0.1530063427683066,0.3473881719138109,0.03687354473843529,
 0.37519090742259076,0.6915813670865872,0.3208756987882896,0.9841989371569715,
 0.18889063004002815,0.14730620623803925,0.8939854662858356,
 0.36654841783317815,0.2813428425351494,0.9980129735917047,0.7185242938681076,
 0.1181084608206937,0.713221754112156,0.9207547824925593,0.15659328081235735,
 0.9365066734423186,0.24757832710271066,0.08040427947171636,
 0.4359094420912679,0.6675107754380627,0.9931274320263976,0.2989213178407635,
 0.8417521818647808,0.5800272518525309,0.056045410514948935,
 0.5308169355195159,0.8950682517692563,0.4436714408178305,0.04183310815025476,
 0.8779315830080836,0.22815863517520618,0.3982437306816219,0.4403587172367147,
 0.09963265698847668,0.16347604704202612,0.06849490883214626,
 0.9449062974582655,0.17455046362754856,0.5451737276794719,
 0.39000792291055075,0.5277828938497713,0.975263553546994,0.3850340153681028,
 0.707502040731439,0.6952542605508586,0.6614787445010535,0.45324724575184827,
 0.210678728720047,0.1407145238859484,0.8600479023532811,0.06852657216651181,
 0.840115348207187,0.9832731823580947,0.35370156926485197,0.7594811193146541,
 0.5860976295676996,0.1519411303968744,0.23887569446899481,
 0.10794460199698097,0.3743410913963692,0.4700933386089108,0.3601277075967385,
 0.2446294203599666,0.42674546916619194,0.14225853517547893,
 0.06691444028633142,0.6813062406005764,0.5518451695801684,0.9005647501386684,
 0.5573897551597418,0.5456314038622438,0.7705915769074572,0.13645500188334203,
 0.7868094912972535,0.4299972700319271,0.015423176190703658,
 0.5791239574113018,0.2720039181123548,0.21022875493457605,0.8944788627074575,
 0.6699651478226819,0.41167780240730234,0.11946672774662115,
 0.10995081865416434,0.88427295909225,0.6038917367882755,0.769410147565617,
 0.26462622514026946,0.8400709702115632,0.7711216027084168,0.4020324851114465,
 0.5874928177893133,0.7372070583979664,0.0011508123759771394,
 0.30264585584164616,0.6408007540666204,0.8041786076436347,
 0.11575501147324885,0.3628038004167655,0.8962065416189495,0.6665104067788952,
 0.25931398053800614,0.5921494713804666,0.7864043315811224,0.7759038798746818,
 0.160801075144147,0.6822144000301208,0.20526720434307888,0.52495976188145,
 0.9640031113641515,0.3552106314243936,0.48467710479708426,
 0.014198904491986686,0.6118415335516602,0.6251635347156559,0.682472222326645,
 0.4517338093354927,0.019393174193760743,0.6951979539438712,
 0.6940737033206391,0.22248459968984624,0.39832123793763596,
 0.8239303891322426,0.9246854991879715,0.20599574415376865,0.7058098459335161,
 0.6954561921034141,0.4145751262813878,0.33349632184570455,0.904159472395729,
 0.3592609431528361,0.9211364456422207,0.720004316256946,0.5916742369526059,
 0.7848934482463255,0.6468122703752521,0.30382946866255545,
 0.18516038262149614,0.8349208693536874,0.6765656210114857,
 0.17071754675058126,0.5235842868293592,0.9469625196368229,0.6221720748992049,
 0.3699565714921885,0.4337065855033545,0.5609983583280618,0.8409955535741842,
 0.7130542153537088,0.5147691168220152,0.3437491623355041,0.161137355624264,
 0.802869270890821,0.5866059178835024,0.425775142742865,0.9782808857925279,
 0.18310044591153696,0.7646019111958507,0.7311960950119671,0.8921751436635291,
 0.7563693214033275,0.8077878056240846,0.46624568660169086,0.3641859891857344,
 0.36237198237544366,0.2298021065768746,0.1745529561703807,0.5257485989653565,
 0.5464231787887033,0.13163172529555056,0.06214408927220061,
 0.9928709975784473,0.48872112369251086,0.5945401639132025,0.7364750081722174,
 0.9368026457540749,0.5699901566434173,0.7789407197668454,0.05463360674926532,
 0.6028172602405983,0.5040042297820952,0.620954290684811,0.29479765411264813,
 0.22073018754806784,0.9073176849693654,0.4085134915399362,
 0.18139418967134646,0.28561572083922504,0.20888448070226306,
 0.3909850379990947,0.6754785142807862,0.2896099746987394,0.957098244918311,
 0.07978856197874173,0.611871527046978,0.9221812325755201,0.7295911281261035,
 0.5904093352690769,0.5327473729685895,0.4759410701589404,0.9266914857354255,
 0.17928096324420062,0.0739427252996605,0.8837120197496128,0.5406068726644739,
 0.40179941944618447,0.42337442312585316,0.15888938251291496,
 0.9082033655214621,0.21901432521459874,0.5987094078986914,0.5942257949193627,
 0.08590458926305067,0.4281267535357123,0.5484822379815599,0.9171207893107027,
 0.7156502021838247,0.234204555345642,0.3370436438994222,0.9487256307161664,
 0.7288260864286998,0.4877770572327653,0.29837882784513503,0.6574676384377789,
 0.9520283072962318,0.19141194793917726,0.23324516898464487,
 0.0828477534722375,0.7064747605031705,0.03445189721349484,0.1086366240835751,
 0.7148611193737459,0.7215542978978866,0.05883133116533823,0.4886986148495214,
 0.1043449315477446,0.5327686962952454,0.7036936047119278,0.7819503085385046,
 0.5646772581963582,0.27566966133050697,0.4864161352898897,
 0.32754603659343984,0.7136629245735945,0.9720432283800918,0.087310026877504,
 0.44917936644094925,0.7755473287412759,0.12134416496760925,0.98832562552293,
 0.3839439787838248,0.7606017767593394,0.21895533001809386,0.4473473549936734,
 0.09330237680458653,0.6293031622007184,0.32221354907983435,
 0.9236583128099062,0.4204115233371978,0.8148827022517754,0.2678846993708631,
 0.42697758494152105,0.8781077968233068,0.07037644613369143,0.696312177381817,
 0.1014972786263153,0.2681846038054936,0.9806847006813715,0.35254225148452634,
 0.7397135484658405,0.11246858810366345,0.18845731190765658,
 0.14501792967455496,0.5393450635344887,0.5463224268867234,0.4156948516607363,
 0.4823166668210259,0.8858067060909497,0.594168407334919,0.9151118695452413,
 0.6588285499886556,0.7906566950891475,0.8644001820430007,0.5403076855847049,
 0.25307816692775953,0.8516243587014167,0.8220347089096998,
 0.40630687762546225,0.3156568627428057,0.5378536934422836,0.7871517176488062,
 0.4934217047919893,0.5421853361403912,0.25686172088022297,
 0.07320729104380119,0.07815787967252485,0.03032766862217595,
 0.3505661648765861,0.6896929347281223,0.5197670958302645,0.1779396943089271,
 0.6138851653901228,0.9326460202884264,0.25229084088993947,0.6242936024226415,
 0.9153303390234584,0.9075101968311237,0.6447200093164116,0.2305642833507623,
 0.6290923825615545,0.19051318318987842,0.6885995562739239,0.7522757626993567,
 0.8592940033276368,0.3393666930362862,0.42513189093720294,0.5883669858999925,
 0.2846385416231815,0.011470667785824284,0.22099728469568125,
 0.5807789021085323,0.21377143171571322,0.6534625771411242,0.9989329503390454,
 0.9717276127343112,0.5553646037696902,0.17396802120230492,0.4701759835801822,
 0.49935267693811636,0.47536860258451097,0.5295677640643934,
 0.6426758856287904,0.1201458958702718,0.7762014247628631,0.7526192539568954,
 0.6741804727178029,0.7828266021075648,0.7734198085006558,0.3254206412802604,
 0.15473094998061765,0.41254347268058333,0.49224728754159364,
 0.9676067058079754,0.745419815067358,0.11184360360718981,0.6627437918292,
 0.33167847317502097,0.44376998611712326,0.8702008321563017,
 0.5721469408193233,0.30773315904887244,0.9630209165623835,
 0.05597902402670751,0.17800802115066983,0.362776241232585,0.8122429303360003,
 0.8763930980848361,0.5792641958744299,0.5270260571307911,
 0.014487415553887328,0.3770534163924988,0.30679720148804246,
 0.13253294427607565,0.6377639649417708,0.8756673854584374,
 0.26256867781519444,0.8207947696454655,0.011687972855725315,
 0.11555171127712494,0.11522624070843412,0.7991191941882283,
 0.5424229576088321,0.5934464567347657,0.42161519268916015,
 0.23845872428214832,0.5188063118624442,0.8622989633789448,0.8820835730455715,
 0.9306471876012242,0.10029732819969006,0.11726423460917323,
 0.6901058349225062,0.37685086305493654,0.7924359430090463,0.6879284895835737,
 0.6003435512168298,0.3617080214759636,0.833250540632802,0.590778105745442,
 0.5465887104884873,0.7197835850846931,0.6916407187391898,0.781819637175673,
 0.5258847931290964,0.42968936671407376,0.860388695378435,0.9314808683517974,
 0.46878101341462763,0.7844569453116672,0.2862643626340192,
 0.27045165056184906,0.11839849049591844,0.2325550386650207,
 0.9567530207613776,0.5058559689366817,0.01728405493740648,0.7800519564312888,
 0.5139419289645769,0.3134978381012018,0.1225827473043688,0.3333348911757188,
 0.9428987259085931,0.2818976927223653,0.15597880611023962,
 0.12229336531580681,0.246674562089634,0.2969190952834727,0.3621561481358896,
 0.6107837566082522,0.22195107573389306,0.4376966969823449,
 0.35968212116162157,0.6856703078338502,0.45536228391720057,
 0.14985893327117505,0.6160707103621058,0.02072324474429399,
 0.9892694156918501,0.16983032575659862,0.3775821411663872,0.8664246289399306,
 0.9392503047578185,0.5324670863981454,0.9859083221128583,0.9971425110958316,
 0.9246771711491215,0.3376595713199174,0.5431514196355528,0.8801455447780158,
 0.4603069529987427,0.07936137304756252,0.4038400705910725,
 0.11330756737549008,0.13254160967651585,0.6285779296942853,
 0.2162305571786265,0.8685261015747572,0.16886496586690536,0.2136679698091194,
 0.3560557649356667,0.9662152342692154,0.106068885691733,0.0866273688885969,
 0.74995305330236,0.9048898739677159,0.26311875199569223,0.6668013686025721,
 0.40644891063622024,0.34368103131977945,0.10838187116119158,
 0.6492361321224142,0.7776613331571416,0.05746197427021771,0.7273910325632031,
 0.830831544555706,0.4767653899429828,0.004773632405600137,
 0.30046940135573674,0.7752020032568171,0.9538778693873259,0.7727905581457498,
 0.9114876045887107,0.3914685195551002,0.18927078569547184,0.591663277227314,
 0.6088628727987516,0.71652272143686,0.9337410911745876,0.28363826752014887,
 0.4308255977082135,0.8442272814615989,0.8206781829572041,0.6636302353440989,
 0.3973886944835654,0.5391522305586122,0.784960107224964,0.31214113409223376,
 0.1302823660401664,0.2751257590728321,0.8289889313225661,0.8529474716494411,
 0.14439080152677897,0.7381316675880001,0.3728812072229427,0.1648257637450955,
 0.8946832989536705,0.9962766991540207,0.17868534144586212,0.9797779719236237,
 0.9528382329194248,0.050323172408656136,0.31802288736759876,
 0.7037769419894588,0.963462914241456,0.5094719844914578,0.14556827947833528,
 0.7612414491129192,0.12393027687295544,0.8167702102530929,
 0.41246735197118356,0.03856934144153612,0.8834603415381328,
 0.1875011966260195,0.7363880254586195,0.7538618273740832,0.20633233203600532,
 0.3404178978069424,0.5752191394580807,0.8588080566641925,0.407251900230417,
 0.04903860419793937,0.8451892504356889,0.6694424826442604,0.7656831684882994,
 0.1814085665435895,0.2961881827156634,0.9138935460554442,
 0.042771818674918305,0.008920831380327843,9.467998785398501e-4,
 0.0032291241467432563,0.11483760288054246,0.7757618181296593,
 0.17980946855899027,0.9410988117949228,0.9437450211040644,0.8000837784455263,
 0.22242972717357767,0.13562790591293228,0.7449881722798923,
 0.6190688117448088,0.95781775288649,0.8238817174386933,0.47263159785778874,
 0.32905470341347653,0.8374528347733343,0.28753039176632045,
 0.9409187205531702,0.40580877979022545,0.9084477662363462,0.0594687194507284,
 0.9236747077523758,0.03492000671135831,0.7111468488744417,0.6344875816492099,
 0.5743516184571975,0.5346956991236667,0.42971926888181255,0.5835877734117031,
 0.08253138708778496,0.07076339405960119,0.7428873691745448,
 0.8015720680631356,0.3939207075553265,0.3900574483097229,0.7048856324067979,
 0.06602662131971959,0.20194147226109238,0.39531659467924696,
 0.879628668647572,0.8019047085466741,0.5554086783381957,0.8647040511719585,
 0.31725809640080327,0.24231068519965526,0.07433031666728518,
 0.7398108411317427,0.1941293884437022,0.26947742763839466,0.5599949325680428,
 0.3227218036126398,0.9957356207239256,0.17522387022363684,
 0.41322996189969696,0.07418313902226392,0.924893732429052,0.8873833565688564,
 0.8262463973222541,0.8498132452662053,0.9394863782401065,0.23133972039983786,
 0.37924597614993527,0.5878049689273799,0.5070652741664305,0.6415703664329745,
 0.21326686408283324,0.39133812378960964,0.2862188262147922,0.927631830439095,
 0.39187604516001795,0.7882892457128554,0.22781879301140917,
 0.3360857345425504,0.14786045216796562,0.7301877979466403,0.9216995517159076,
 0.8186528021054497,0.8820204384135263,0.23923947727871198,0.4651109102135884,
 0.7345681027765651,0.4030664493010091,0.04477676407466591,0.7990134885864798,
 0.04731202105719412,0.053158161307577556,0.29459375728784376,
 0.040045861767579405,0.7399255542865507,0.23889066905972023,
 0.42338659949485025,0.6727856620302213,0.38008668832439785,0.833738592761667,
 0.004770062684877785,0.5756967409028534,0.21919696485764517,
 0.12073435984783054,0.08544602634056542,0.882431736918288,0.4519266801320002,
 0.7268720248807304,0.22039046756858438,0.7170062141910925,0.4331080715148914,
 0.7919875198594187,0.2847434981104744,0.9492766172214991,0.826253831605033,
 0.5218895740798549,0.7057363268771611,0.9236945280928883,0.36279894067031127,
 0.7135735782065907,0.9603303726352224,0.17593465346538606,0.4298457799101002,
 0.5720531857403981,0.6006125978204456,0.09916974924674238,0.5409964637535009,
 0.381073217471108,0.8455990623927739,0.2703371612044895,0.02345655873802721,
 0.7631413177529345,0.23410518165403849,0.5023293052059381,0.7013942549983265,
 0.00292780591794517,0.9174345358273021,0.3434016121421952,0.3204973976617248,
 0.5520509911531196,0.5078598361898423,0.15331882933218,0.9300107424357732,
 0.5034860335756114,0.8267894121042912,0.12055729292494932,0.5470745037945975,
 0.7684926876200495,0.4848744664574193,0.20704640486161408,
 0.04338106347024051,0.6190370991190206,0.17132070671952537,
 0.40242597677951564,0.40247915849455196,0.5684683667583799,
 0.5273910539205764,0.7941209476393979,0.9266742878248593,0.7307656489024075,
 0.5058033535623785,0.9921346804770012,0.7789748222805521,0.9458988152605541,
 0.5113075624918355,0.9295408800841842,0.21087841293281517,0.9173635534577372,
 0.7615920629401102,0.7134056125531305,0.32486922884265745,0.4304091921505264,
 0.8047745059861009,0.11534951979269414,0.2551412283200739,0.686791838782321,
 0.6738767710965583,0.3101010579414595,0.2690724128501094,0.9328413898494338,
 0.3765882560791276,0.27812142843987,0.39949790798504403,0.21423276658446266,
 0.9412037909134557,0.819182591562607,0.6331936086636082,0.3381704493190215,
 0.13303498978464146,0.3250743821454789,0.5865064523372797,0.5094772031775551,
 0.1480560853995263,0.8793661502155714,0.3620290164208081,0.6282930734788501,
 0.9844909994914114,0.18809182086173082,0.2002828234349564,0.3921780639132596,
 0.8046661813029711,0.2538055415276953,0.642116265427286,0.10539719270492243,
 0.5198485220571758,0.1043429677911486,0.18617691492458766,0.5575482315458632,
 0.10223863791945942,0.9810709514428816,0.235627752545007,0.42212210498735103,
 0.964771654551564,0.7854659082216611,0.6497455046677776,0.1456320596468622,
 0.6210536580114014,0.9804690975221604,0.4262692041919729,0.8577146596030198,
 0.5467837624547652,0.6050717393676254,0.5889524706060598,0.11022333785370841,
 0.5940020999526805,0.9697782311074489,0.18624811243221107,0.3359059373679565,
 0.12912265943928125,0.9498627587095868,0.8873818882283575,0.284138197936855,
 0.738804334832565,0.4835645505379953,0.9093760792908496,0.4719054482247371,
 0.1896253838428099,0.9775814540162612,0.3216846720476635,0.37726448016842207,
 0.44635363126823857,0.8348135916645604,0.38615539606699767,
 0.08133674198369145,0.13262778791959406,0.4744490655139426,
 0.8943182166084839,0.01637636110627605,0.4577810477345178,
 0.42517095595024834,0.4795272600842043,0.6041766606465707,0.7800821920279424,
 0.8633841195405807,0.6835219752363477,0.30645027712847284,0.3188847919245743,
 0.7685676035273131,0.4123486881521331,0.4012020893797754,0.13412326219916215,
 0.3088012643193154,0.8724173491206613,0.41193155260462544,
 0.11819165683758981,0.8972977855074316,0.8452500774351392,0.5260002157082403,
 0.6206089040040867,0.8060207056452763,0.16764990558582393,0.6182074205699299,
 0.22486509589296522,0.33279212043195583,0.8793565904999595,
 0.3847856739582888,0.6237757589399063,0.37874713366920043,0.1940357580689125,
 0.7277681259002322,0.938472010075913,0.9028701661026285,0.8631455627661382,
 0.48000122046829297,0.7687203188086446,0.579410298607692,0.09881368984169325,
 0.6921201672048637,0.7871876495269524,0.9843304934070228,0.33742320641222334,
 0.8167904644160874,0.8006500242613039,0.18070184723056326,0.6359279518886236,
 0.7222941103089016,0.7857304711014123,0.9924634887333572,0.8999066594726967,
 0.8911644800998824,0.03001335003162997,0.6483918002916835,0.8661803172639658,
 0.9306074917781889,0.3147747960574073,0.30362189643358517,0.652628849781567,
 0.6725783071008595,0.02837456464745769,0.8643452270677161,
 0.26739204188205545,0.23769559098356907,0.3358362684450136,
 0.06748287089081939,0.7358389520979912,0.8229080381514844,
 0.04267134564024411,0.4469993807248611,0.9144863627901835,0.8036078271194302,
 0.4773568046519976,0.6933835036274715,0.8386888794588825,0.32013466017511416,
 0.2513897009302184,0.1566517708988171,0.9809297068336753,0.1953628638850219,
 0.727746219918573,0.2377824396972379,0.955860590134141,0.14042697387794967,
 0.07892173919326761,0.26066270822009674,0.11775776593811993,
 0.5226615909817238,0.7934000746814648,0.7445152678409475,0.7479711393733893,
 0.7261679645643933,0.17838962028798666,0.24745399478450725,
 0.13371144934729107,0.7595351210509613,0.25044982042133535,
 0.27043584458972103,0.20831404476469073,0.026255500984992963,
 0.6915916750527992,0.3543319766849429,0.14455509593721416,
 0.13956295108447614,0.31458587215298384,0.17256183750502307,
 0.3131925762037393,0.7365527757030758,0.35425485216510055,0.532604359484357,
 0.9396045116593903,0.1028792923526688,0.6696497189686581,0.46974013321303,
 0.6210566219280655,0.04982258196530687,0.027539622666368224,
 0.17127343798532013,0.10313532960761163,0.16696906260696143,
 0.31606433760359454,0.34997176923997353,0.45488461093057064,
 0.24451276075600803,0.6317660717074766,0.7206943185312162,0.9090849126164721,
 0.6991215388101186,0.14085009924237257,0.8951439826373958,
 0.06222689769423628,0.998447877774043,0.7963924706723287,0.4156438667871464,
 0.21840765254928807,0.14916169344424124,0.41856945109784727,
 0.5856282234132884,0.32085496600180075,0.04821250212630679,
 0.02186653638559699,0.23601624594654158,0.7558624443391778,0.362246118122011,
 0.9899803832566816,0.6037023096934924,0.843987210121223,0.4887205308745389,
 0.14625350177768004,0.1803382935237735,0.4931406827765089,
 0.24392698348323838,0.8337594807751353,0.7652727823516726,0.2091855755734212,
 0.30703461470389737,0.6508890445648243,0.1385413753299768,
 0.07799416197015852,0.877217657474178,0.8067566685687854,0.7799669438256676,
 0.08165552652456787,0.05205730065666825,0.9287267168194551,
 0.5335111000966993,0.7574868348145406,0.08287576874349467,0.6322810403778412,
 0.467251219187083,0.43759579128614423,0.16176384574078562,
 0.06141030313381113,0.24770620379962938,0.38540099282077733,
 0.5277446394249916,0.26107330123601413,0.7695393550946584,
 0.04950447693371762,0.09798209096710508,0.2600776459758396,
 0.29732967993801673,0.8974451981254852,0.8405677185630589,
 0.25465399273618283,0.48320529995396977,0.7518102755102248,
 0.7832003029584431,0.7624270995647447,0.43707881953211203,
 0.09301500046738753,0.1953474514884609,0.07613850849334503,
 0.002236394051752355,0.5879828949919942,0.4868946028932801,
 0.25073549613367196,0.3588058415536374,0.16286562391132708,
 0.8231602974287114,0.6924567182406791,0.4946007182223683,0.5673054026384621,
 0.2536218819868665,0.9841431600561144,0.9170014536565217,0.9957324148495711,
 0.7302296954804572,0.01930386204949963,0.7269900957922585,
 0.024918349581914412,0.7809732823786976,0.8663203679819167,
 0.6380175067313549,0.25377972018397665,0.6390505180014714,
 0.19600274326081146,0.48386435319996024,0.08125702895621123,
 0.3148281576560228,0.968022246336626,0.8119573711043608,0.29499793829402166,
 0.8369688087684167,0.8275153312396951,0.8772409007039476,0.5866188329289451,
 0.44093658787344214,0.893659804185719,0.3943717631385326,0.16340639765253784,
 0.1763424255115651,0.13868629221436413,0.15003275138675665,
 0.31498445761135496,0.9614963643754721,0.6434304187256757,
 0.40892009705943755,0.7096655943187282,0.5536414946182225,
 0.06005970646373904,0.810306922725508,0.42948031176278123,
 0.18818664128225748,0.345849790774199,0.9549607615689079,0.5239036793700099,
 0.7549152091247763,0.5836929544312839,0.82406467243083,0.8357025176189165,
 0.9174741411121218,0.9000041564360352,0.5321088595639475,0.5434775373686662,
 0.040973378572395136,0.19102862435499557,0.4566088837336475,
 0.01865426360251743,0.4768022367848297,0.37890331203193317,
 0.29350640413443985,0.8450943887623756,0.5960332630138166,
 0.48719565754596017,0.716863561269722,0.5482916927362672,0.2806926794726873,
 0.24751990619918074,0.381968922670185,0.9215885971636174,0.8580773063751207,
 0.07232592535355586,0.26265033304366436,0.43424226154200096,
 0.13643486288703055,0.9204553843868903,0.6013847408524085,
 0.0026304749919485726,0.8324231693020563,0.8077099840868737,
 0.3377203154379155,0.5522536803724228,0.4272089893290685,0.6213017217306245,
 0.7477862588081255,0.2773117321239901,0.35061565988117704,0.8736753422940636,
 0.4611308883965347,0.27709147929884126,0.809836872479411,0.4606225458094637,
 0.9467772955888298,0.7062701616257573,0.9123330642133654,
 0.018362423358186497,0.2853608203692395,0.2560675036371973,
 0.2866357857225347,0.6398663514339937,0.23143780798128188,0.7094550474745648,
 0.7536662381941671,0.317514994831151,0.2613559709784472,0.8555067196331522,
 0.14001383304930948,0.014377168741079194,0.030494887943377957,
 0.3511485689378582,0.6353091417573302,0.587166315567513,0.5960403263124487,
 0.1406789091855164,0.035781775830698304,0.2935656659956599,0.884243348766035,
 0.13262210437647892,0.7612476147213081,0.7524756823056378,
 0.014450080525056253,0.09431703985554252,0.5338984969708284,
 0.7635520750568532,0.48167711819968106,0.6073238014366332,0.767346629116103,
 0.22427572103428617,0.2716428389372174,0.8278167717466283,0.7248892294852806,
 0.006881216484939001,0.08596170115420476,0.6885814358078316,
 0.9470228246777974,0.5154883495730554,0.09592757692470133,0.8383705497562998,
 0.9723307051195998,0.8142860054171757,0.6575473430183894,0.5363593792852991,
 0.8796143481134359,0.3157248656595444,0.9254210992170909,0.231306143734193,
 0.15139006138991395,0.4415834039534292,0.5790246757323876,0.4409532933409849,
 0.9857678509764333,0.9910877058083628,0.6354130870457585,0.3474294436775508,
 0.7069441045752984,0.4274393170433469,0.5733520660316421,0.4581589635499195,
 0.6216458663541301,0.4179425494958757,0.36160473471179355,0.1809098783960661,
 0.40935489316438667,0.6487617826512776,0.25766632934239686,
 0.3117872230236497,0.3899127475370535,0.7406034807668493,0.34629303961537017,
 0.7260579313305189,0.19589334695039162,0.6716230908117182,0.9347670082006853,
 0.6855018655273963,0.727369306987059,0.5655368204785431,0.8706119012174124,
 0.15413381456335906,0.5976608467156518,0.6937861343826462,0.6013988473355867,
 0.18415633700311618,0.8641361204736243,0.9381486066648619,0.3412313767270906,
 0.37538174814267955,0.32879937692276895,0.23178005463960993,
 0.1330056442239116,0.9351249926231127,0.9262785967208681,0.5812875376454401,
 0.5807732275955736,0.652854373819141,0.6220085223479369,0.3008778197016557,
 0.25845063454160144,0.3904852059732934,0.5767819587547718,0.858401512006481,
 0.5838111591968143,0.06434809719959744,0.7269853101277259,0.6713886410065227,
 0.24282949162545897,0.5998613514959643,0.036775824325621986,
 0.6050984848118041,0.3599986996567327,0.872865909831746,0.7110672607933796,
 0.9761928937310393,0.7539535595825575,0.764828442906906,0.49010613112832413,
 0.6962745903521063,0.9036831672425738,0.12186089067633965,
 0.037910873554158675,0.39103832307676256,0.47829753951078,0.3895227670535121,
 0.9741022307043053,0.07819287397668517,0.29389666587452545,
 0.7813710617095103,0.7320215521219529,0.9889133839077583,0.7256104675015227,
 0.5467738645288208,0.5575495324263963,0.7934983105680544,0.5554266677693829,
 0.6068402811606296,0.3356814252891065,0.7872845784175806,0.10109186770947176,
 0.5301089925392637,0.7098085264048449,0.028841327797387795,
 0.4664269401025174,0.08066932155379392,0.5799410333385246,
 0.31376595372980565,0.31151903512509926,0.43417791460862076,
 0.6140318574996716,0.23939851390719624,0.6389107595181678,
 0.49784662206890296,0.2643945884046496,0.085235004623609,0.3402739537919496,
 0.5050364247930175,0.07273951078470609,0.9069725958471564,0.5243185553219623,
 0.49250381087471407,0.8625775051960475,0.3040012816454063,0.5353398805934975,
 0.6758945044473103,0.2320222909380506,0.3547228487227151,0.45910125450636907,
 0.32937364392363966,0.06389241313844851,0.7308174475112579,
 0.8196326348961639,0.304706810287215,0.9039425948978553,0.559102045359768,
 0.2855700689719518,0.6518871103480488,0.5812937172801256,0.0845085157607246,
 0.8735914968118953,0.3516038239286109,0.11825558420057936,0.9310079554396631,
 0.35286276761253843,0.663886873174769,0.0489594419304582,0.6541217590410384,
 0.12460550260406733,0.3054703759245536,0.15112773825830006,
 0.23297318237168618,0.5868491393550214,0.371283974606205,0.23869524812291432,
 0.07178253399396872,0.8925440783284636,0.8346021734410206,0.5930632145551961,
 0.8034058098962278,0.7955210177725489,0.4109597863062219,0.37657939712348154,
 0.10803616111958347,0.6372460704557408,0.15147405540059078,
 0.8788393450870293,0.6366477691791625,0.427685198065309,0.8877370578408933,
 0.28907985108597434,0.1916035291840703,0.8616152370187968,0.8043965242344794,
 0.6242324757398582,0.6994030098936709,0.4341910691862222,0.10070519275729783,
 0.2955285581243138,0.5340866577438373,0.7759246258504684,0.5377861199687046,
 0.9278626234189281,0.1698091650442266,0.5501077712407766,0.15726157687537556,
 0.701886575308953,0.6719789100807483,0.003492028736390873,0.7297077128423879,
 0.4776763797431789,0.7370202104579958,0.3135270622819474,0.37730612404943387,
 0.6591060412715827,0.29947136395747076,0.8143345186273502,0.8057722114387328,
 0.32770896290437024,0.34582537542408487,0.4242628169279603,
 0.9628521626521338,0.8666404999464308,0.13078273403471652,
 0.17394569624243517,0.8536012983114084,0.6494592309163884,0.321093746827826,
 0.7766110655385184,0.42006944335111474,0.06958463690954964,
 0.10609996018496637,0.29624978431807614,0.9169647210153619,
 0.0026844331054018955,0.6811860084629776,0.6690172262245497,
 0.21966061233919842,0.5328041638570821,0.5945121565573095,
 0.08075504672886014,0.2789941541735105,0.520940970700271,0.197335092748683,
 0.1295634021872517,0.3022309789790053,0.09621300476916095,
 0.30796125578592703,0.1027130912245724,0.46859322874117204,
 0.7192400793580421,0.49959858449758365,0.17917523183567863,
 0.7794184952054868,0.48373668516565316,0.9156784317987168,
 0.03522586141560513,0.3687475442821244,0.9396864857358295,0.8237898720084664,
 0.25866437693029254,0.040660714241859686,0.9971306816500469,
 0.5647490863564629,0.30402337723836625,0.890484526643561,0.8971670953488833,
 0.9253436156932436,0.14061452285657716,0.876383020112806,0.412056634783867,
 0.4928432162226436,0.05256487619681005,0.8603115839044841,0.5924890913642806,
 0.6701462095896813,0.2716767457061149,0.537098349033094,0.8298560902475135,
 0.9673151081697914,0.3505760962233586,0.7430837925232456,0.8034437841850999,
 0.3119317783412101,0.16632167234195028,0.017881489094606184,
 0.8760714714317497,0.25564673081365574,0.9879027445138208,0.5025913256228316,
 0.975011716525401,0.5578926941851534,0.864453391743663,0.378415589910745,
 0.3564473644539381,0.25382984568899614,0.5000290492291528,
 0.17564545222745842,0.842633536325901,0.33582049865913166,0.645981786491035,
 0.9424280103485474,0.03303150205228644,0.5862509926123491,0.7946830150949165,
 0.2017634493867111,0.12405317388746573,0.6269245889280134,
 0.24299952405312597,0.17732876218596116,0.06272914942261609,
 6.457229626786187e-4,0.6799102059471117,0.5239020389835095,
 0.7545602242830335,0.9096394259902933,0.6404711908633964,0.8644314979611348,
 0.30509685501973793,0.5576429988785363,0.1512151697838151,0.847460035829421,
 0.6571901700154198,0.059500281122046195,0.5019266654353123,
 0.7133898901024412,0.6111575658315,0.0682961695906521,0.8810379718607804,
 0.15313332089676268,0.5099361876280025,0.36832584625425324,
 0.41540532297565935,0.5070941094186845,0.1950366102814053,
 0.19299889580834795,0.832142807071901,0.38333701668715836,
 0.09144215408376044,0.965527943827608,0.21160797269706022,
 0.34894247726206873,0.4542970765196992,0.08755813258696821,
 0.14409182662242692,0.4465033974805196,0.6589352801866686,0.6620815850452432,
 0.02719729405787874,0.8601976318636776,0.800379145869603,0.6571563405830914,
 0.6263843383239943,0.6317700058559901,0.0674878350696777,0.9086687706992884,
 0.6828228533241539,0.9878392256459433,0.12090553838898443,0.9567288621030681,
 0.16067792958573135,0.19567389405004,0.9408099158806824,0.4817273753782838,
 0.9173674315596468,0.4558361686626695,0.2593283091578452,0.8793386053460551,
 0.8032686481270456,0.6343731458775049,0.6440308740851095,0.7952416335400706,
 0.8047595687600849,0.19185483780300505,0.00855060964182186,
 0.10674878147387112,0.0969961329502399,0.5892319276170523,
 0.32434262791494295,0.12622330717724373,0.19531511210083108,
 0.7632245096794747,0.2845896965110859,0.9615360865355288,0.774523655725078,
 0.5520067834823429,0.9049565143489993,0.5091435092519851,0.550219769209178,
 0.6598379948164452,0.22917282908794945,0.6320337292827013,
 0.22249410520774915,0.3166054631886659,0.19146497356958336,
 0.046110905647940514,0.11579103994262552,0.4601646605972949,
 0.6565265111005871,0.829531296598296,0.22021099597210703,0.6367625823989321,
 0.8106612231842852,0.40160470432251505,0.29539180641272067,
 0.1890521792431329,0.4337428232364142,0.8487497753998752,0.9499027339086139,
 0.8712811656116101,0.16203095937083645,0.8409159194067455,0.3643605451831352,
 0.9023483708068222,0.4314767221729682,0.8795751232044046,
 0.016119899042276065,0.40463646184169255,0.552136636211924,
 0.4598575499181714,0.9550191014815554,0.04050764975667365,0.8869883367636758,
 0.6589974157194931,0.7598553498866718,0.26925982165888873,0.9764795194143034,
 0.2510131151791891,0.7306832716151233,0.5917288386759698,0.46728732744324486,
 0.1573551591001624,0.9913108040921896,0.4873222116212508,0.451456761379851,
 0.43647181835598525,0.8371380971872474,0.4561106624060791,0.9821081399099592,
 0.5247406596759757,0.4595645974977468,0.7107197119060015,0.5793512324988488,
 0.11759150053966017,0.23726756612740618,0.49416390263470256,
 0.8615309306800195,0.8634389466545033,0.34015855247718996,
 0.08471031284933184,0.8528291594917305,0.757440977751462,0.9613347609554359,
 0.941103155072889,0.49000295100130764,0.8345899689154478,0.25416393359576495,
 0.40882488574980635,0.2776309910863952,0.1270530999575894,0.9380905592422596,
 0.18762198705777777,0.006738402593676307,0.17014514610510378,
 0.20652466530861247,0.3534383117081652,0.2682826348480901,
 3.489096252993207e-4,0.6454975711203791,0.5150442125062804,
 0.3678652012991156,0.7356381763544491,0.3397589630860274,0.4151908601763493,
 0.13270317356854489,0.6986330112742594,0.5434145829610416,0.140859498380432,
 0.5918865222422623,0.5192754814548824,0.4676104991261252,0.5091314598061399,
 0.7517648546421183,0.37993623772839125,0.9649591450955264,0.8248026933429164,
 0.08126550771910013,0.2716270635108916,0.4797717374304118,
 0.09408980321374816,0.9139969202171163,0.9080105278838887,
 0.17293667305564198,0.46689864048559626,0.9591128473046947,
 0.5553906816454374,0.5921161459228216,0.14599990862770706,0.2278830240688987,
 0.945080769068516,0.8910579816732811,0.3385907486018994,0.550872425788024,
 0.16633339180994,0.8594481864358026,0.29974348789440686,0.6632330112994325,
 0.0882955546619657,0.8801158730031423,0.1654206444133408,0.23645132338358565,
 0.14047459440048482,0.9176298368149067,0.5303644487734035,0.8753805609488343,
 0.3585010997750542,0.6234798643990956,0.32400001595292993,0.7504566300729273,
 0.7232135531194289,0.16091070258142015,0.017542426719041604,
 0.4595516862621055,0.5510827000772873,0.9678545130621908,0.7468984539559057,
 0.029297882315750257,0.9507980759491086,0.9107418938264635,
 0.41118249041478017,0.40791677154562267,0.8004823219373453,
 0.17446015021041028,0.7856752648794547,0.5462627130148648,
 0.44533655482536183,0.19436110963679754,0.6256356576128382,
 0.5719806959155357,0.4005133074159063,0.6033417616863956,0.8651040314567638,
 0.9615454622733128,0.5391622222677492,0.5596496063088247,0.2104448638537465,
 0.8682490865044948,0.5090092422374298,0.8323665604056141,0.7666406192882237,
 0.7529795715279648,0.6945122337619183,0.904325636782398,0.4861180725776536,
 0.0575797821228472,0.4108449194778273,0.6177280722997242,0.5329415641229981,
 0.8407903764966426,0.4313132001664134,0.13140161271424988,
 0.39850974878070666,0.7864680220489184,0.20266117900231218,
 0.9343965409545578,0.3168371103813208,0.5549955197535925,0.6817449014838659,
 0.20172049800221215,0.6304598712071983,0.6387104926339184,0.6870856756444768,
 0.9510067211072232,0.6722923952422655,0.4432791931323722,0.6776047605782956,
 0.9492343584880679,0.24752444948205232,0.6745347239707721,0.9713666058546887,
 0.5849727354610175,0.6465478369192245,0.005585934264807801,0.564387439429373,
 0.11774639042990187,0.8615163929195865,0.15589538967295036,
 0.26142064562748113,0.6396053740731023,0.23593899325973045,
 0.8269221166158073,0.3232482111566487,0.5637690316338484,0.5013255868340571,
 0.05075986466855176,0.563967386030208,0.5132570307916683,0.2836156763903316,
 0.12884403320615512,0.611080344527016,0.9458122346808986,0.8348556387642443,
 0.26282612182546994,0.391681218577669,0.4385317959778092,0.3257855640587509,
 0.5949061044621315,0.8887124115868797,0.2910437357740854,0.38340824290062503,
 0.6383332493841215,0.36521094070209825,0.23657322547576853,
 0.033830286322014125,0.9460839990297578,0.3787801771409761,
 0.4050757770456954,0.5739349095270442,0.9372760099291958,0.8827761205944,
 0.8037464930645878,0.5414261498330958,0.4705985680415604,0.14235683894064832,
 0.13552449918197063,0.8693130277923733,0.16708071932385438,
 0.8779497351148432,0.6550913764299165,0.23050732394820828,0.6126722842306974,
 0.010353371754257878,0.18552100294059948,0.8584741615948247,
 0.3908641011283325,0.8375953747295672,0.9679590163076838,0.14954382169535918,
 0.20881787120281758,0.19326793634095207,0.5113914157770545,
 0.10519149063431088,0.94050349688039,0.9445654012326387,0.6521805495424483,
 0.42968115827819475,0.010149000669775976,0.290970394640784,
 0.3084980956127743,0.9217413495140535,0.1098413307032825,0.34034902935413147,
 0.5176086221314979,0.9609956046034792,0.34619210678824996,0.5552506958819662,
 0.8904732338043032,0.3983394085416525,0.7258623620563774,0.44980442365227236,
 0.38940374959092316,0.41849032379583206,0.95472369200207,0.05906734656854318,
 0.42470635889662844,0.14333875011414,0.8425126395509389,0.5198983351396032,
 0.8593902731940148,0.6053758544382354,0.05767460241038436,
 0.008232619332106172,0.3826511456732933,0.016521984119639455,
 0.5275299505431094,0.6105064116881662,0.9858032534878057,0.39613208996977656,
 0.16107469144764808,0.20930127704017265,0.7440832114896084,
 0.2229190739490189,0.926813431951953,0.0907885327969441,0.5427300262650829,
 0.02278310753006285,0.5291381706717279,0.0520518764301634,
 0.25406186408496056,0.5250845063160978,0.6585489018525829,0.7750166921083486,
 0.4500935273104112,0.6289260885467617,0.8083627367400538,0.9776174337727971,
 0.38984627041504416,0.22055719183369293,0.21877145528269815,
 0.13209792400941311,0.8787845577216553,0.5839539931897758,0.7061645087399375,
 0.7821586670633661,0.617540271871464,0.6181581200118101,0.2043549337410675,
 0.12876001719663144,0.9315997007520916,0.5214219414400241,
 0.25270837903967247,0.5717848060859869,0.8303182595240067,0.9953574953219614,
 0.6056782100501023,0.1267686003353451,0.13336707775157242,0.2759485196685999,
 0.6741964812590561,0.047080385524369994,0.41339778851905407,
 0.7036275540964925,0.7636251889844373,0.24423663944067886,
 0.33748190237830766,0.5039328890924293,0.8246224882690255,0.6986044553845422,
 0.6466856904346718,0.34900133096398533,0.9594435591267226,
 0.35447195773851037,0.3087548349523791,0.7962845380828243,0.3957685256221902,
 0.6015029552392456,0.3951534967861725,0.1492644134531298,0.6368153042270147,
 0.8071684188668685,0.026247665504360418,0.32817724040861795,
 0.1446238388542025,0.30690072614452524,0.27014468284702864,
 0.29967563620511206,0.32768562498308507,0.9064491664544057,
 0.9239226346869951,0.061299579078209265,0.5890935401884478,
 0.5274067958329538,0.7962600401025609,0.48621472435552426,0.6518636731738656,
 0.6460949992063544,0.020208309118533663,0.6426253267305022,
 0.8569177518439758,0.9331842208449951,0.5371231441807484,0.11794539637379786,
 0.8478615101192158,0.7348298806857816,0.711243938341606,0.3656751252096052,
 0.668900210574752,0.24057210200235324,0.19272363279902682,0.7802421778569737,
 0.16221823601248586,0.8519695059818913,0.9139286014054185,
 0.10854171209281216,0.46497091154799974,0.6393355973624131,
 0.011912026276385768,0.14644341608945888,0.6657031548237425,
 0.053535969033699726,0.3498360761824201,0.14676722317132196,0.85551054169217,
 0.3779086253427436,0.3875064470168448,0.03598257024805773,0.7496207181859921,
 0.36903686140226366,0.11117921380576412,0.984906713850777,
 0.39395188433046224,0.21597009955027102,0.09754819752081245,
 0.784413591547926,0.9999587668875489,0.9486368483103746,0.7403832776765437,
 0.01846079262423428,0.3691099427870328,0.8836607019056232,0.2528362429749609,
 0.07065265106969709,0.027984943648886484,0.9512542188203251,
 0.8311374042383608,0.840603259137068,0.4417914248403707,0.24776288429864668,
 0.6817531384644832,0.8632727271242554,0.2983125179183934,0.8555598925285549,
 0.6346835815169073,0.4824630699125345,0.5681470960736068,0.7189167174274326,
 0.8713109084291818,0.11589639286272035,0.37014366602906723,
 0.40595764836487014,0.9178724279359374,0.3173070021317874,
 0.03861310491946901,0.6716664569844756,0.2814092174184192,
 0.049334467575837926,0.9650189401899895,0.3858980647638066,
 0.29989722312371536,0.5100399406397206,0.9563389104706675,0.8464872469100851,
 0.9051819189469019,0.52453711659165,0.48798287496940407,0.4282206268922739,
 0.16893263697978067,0.43172183251755925,0.9923676300507308,0.1505147770534,
 0.5941881898107674,0.7663346660201321,0.732243517485902,0.03584098893419274,
 0.9016205230393363,0.7977439027367225,0.35835069049688584,0.7413753048117377,
 0.04198904806193593,0.4045972909305898,0.03555785696317937,
 0.5586948390508393,0.993128819612359,0.03784546164349434,0.6591380993806772,
 0.5030090722507715,0.6837652274583683,0.23699551478786063,0.6509586607032284,
 0.16773794542712206,0.8926484884468481,0.9286294286290718,
 0.19507473468311531,0.6998162287033114,0.14291382023803056,
 0.8553364262348266,0.6003208019130357,0.9739175017970402,0.4206987776304769,
 0.4698827799286708,0.5884014701933555,0.18077898003123494,0.1212498031546867,
 0.048712839691491205,0.07490887685532088,0.33495999872883186,
 0.5344489519563738,0.6326602861416118,0.5444997751540488,0.6493374896208849,
 0.3079738787985147,0.7870681797947827,0.45954329386113213,0.7541588710470928,
 0.9482545578321795,0.6634493682289871,0.6487973889150584,0.6999232312779367,
 0.19885792492742782,0.04215590522191359,0.1655043255322236,
 0.4666280821957022,0.5236803123455175,0.9955445997181546,0.4247518584992065,
 0.740353169379033,0.041519818417080745,0.5889255111282923,0.7644339925629079,
 0.052338081943365644,0.8592089988550956,0.5747105775540791,
 0.5963714838302092,0.9104403877320204,0.12480978489665773,0.7212527308848967,
 0.1313394987157951,0.6632333361327571,0.5604678918710564,0.0108640271275382,
 0.9830366468132102,0.8795627297103457,0.8802862890034505,0.21280888745681148,
 0.3345203576298189,0.824908173137826,0.7349922806563809,0.7410300726230246,
 0.5060550775680435,0.33655196412887256,0.6515344621764793,
 0.19030712517510795,0.3735351050358824,0.4022076032147718,0.6720140249341878,
 0.7446080667533336,0.6235933720747042,0.43782035037601164,0.3091851385371083,
 0.3416695937780587,0.9773404659942309,0.823465952563155,0.556245543778964,
 0.13736231106132546,0.8453051078551398,0.06949395643422462,
 0.32771483835577264,0.5551241350041136,0.8987376302375676,0.7229105887060299,
 0.8768960396984544,0.4612459233539188,0.43033531372489164,
 0.30169817503738583,0.65224966598109,0.6074831855695778,0.6442599852163876,
 0.011333859241549393,0.3157388846214326,0.868426263281703,0.9617415178228621,
 0.44692045202236175,0.3027064366024428,0.8975364853793307,
 0.19442582140241693,0.7620874142199039,0.14314797906331467,
 0.5711818221763776,0.21332398484606474,0.16236489335678006,
 0.9370005304372667,0.9458721608424827,0.6716288869681862,0.2882727294842502,
 0.2655735130993828,0.3784247190682756,0.460260540783597,0.6693814462374748,
 0.10264713396040026,0.8045561813359162,0.6619681928224789,0.7759702076210445,
 0.27042825322692354,0.17925137062826668,0.4961871352000482,
 0.22399774144891682,0.32058010646304524,0.20061733511579072,
 0.8139087412378371,0.7133776010582031,0.05248982393072643,
 0.30829069068732506,0.060576559668273466,0.6190441704936105,
 0.22995056385940504,0.9337604397652467,0.5393787750432464,0.8334355638980633,
 0.9894508789947003,0.2704418262176417,0.5649173590513583,
 0.002620525328069667,0.8048445652534159,0.6646475886852318,
 0.9070079622055109,0.03969726904808735,0.5258666010667559,0.7605397485397994,
 0.028552761577821717,0.7927982103700526,0.2944595108956798,
 0.6117564845489555,0.7493665570132623,0.2560694711649981,0.613345806500211,
 0.628125311528424,0.564870907729552,0.15206676856196688,0.6308009572295931,
 0.46183449814757094,0.9618008385808259,0.3198172214178503,0.3835220341762964,
 0.9501563307159288,0.8265015534209454,0.5786973643339264,0.8051401368786795,
 0.3870154890967803,0.7826398760756426,0.0071379925823479295,
 0.2907876146124916,0.03064213946289185,0.17787830271608474,
 0.026414841463850647,0.4744969782030506,0.1095820013887484,
 0.7901769708882683,0.7931827829302837,0.9121050899704838,0.9611892777100574,
 0.5309072071943046,0.37951297407280793,0.9989716186587903,0.508492142422553,
 0.3683623765754638,0.34539949515600354,0.5592954387248359,0.5652239687382355,
 0.43803825403591856,0.9520481480195481,0.058326369306769266,
 0.7694202806192206,0.942805979862355,0.7673406260699482,0.344011082374571,
 0.10560587916422215,0.07311898656823956,0.9309000719590514,
 0.35012200926697057,0.4630536754545749,0.38092296745821197,
 0.02693052904861748,0.17284306353047652,0.13908997017273395,
 0.9847221634543328,0.19510627981337103,0.7926909387900994,0.7528631701815114,
 0.6398929346447026,0.4540403796182785,0.9295714027747212,0.40496867459960384,
 0.009429299133278612,0.241389555327312,0.9888989167003286,0.6786864185016002,
 0.19082340563778355,0.8988097994950461,0.9836944970464977,0.4099301567204765,
 0.38851519801602996,0.6294801103180773,0.5500503346093252,0.4301903038074052,
 0.5489271367253858,0.08083101455529496,0.3673611435397103,
 0.25991979880009886,0.03452498801287829,0.5739775387569709,
 0.2761523536023517,0.4450843312817806,0.8751517107787008,
 0.050395436600528254,0.10122375724552235,0.8377499221381026,
 0.7212005551339671,0.6787332668217593,0.5644595066292998,0.16979878790883096,
 0.17828880505311007,0.1819066573254593,0.9180038756005304,
 0.04997580860800088,0.32501643102135824,0.20379823059791058,
 0.4350859386241943,0.3526664789644838,0.15707499624514876,0.7548908611505089,
 0.655651180451073,0.056290729812125906,0.3174973782129563,0.8492422177513309,
 0.9623657443829371,0.5837476020989155,0.9725018198681054,0.17578579950587825,
 0.06170943398863227,0.8401125156339958,0.8798219967870596,0.8486371978702333,
 0.0560348999546717,0.05134514117500022,0.23825626548844614,
 0.37643208265290706,0.9221320914489364,0.8856517956962948,0.6023914894616669,
 0.6615847377618994,0.7631494362827214,0.6202818050839275,0.2923490846981226,
 0.09208540900506013,0.9246246042591801,0.7869944361009863,
 0.21026161812709077,0.44540196298442336,0.9712712924161914,
 0.23354880809222456,0.9348134683197852,0.5642380712917183,0.4151102212719989,
 0.9553007936058897,0.36976597786218857,0.7266395045682292,0.4133360616020134,
 0.07691793709140782,0.26538689795378323,0.36267736482514257,
 0.2414819727411478,0.038983299033017005,0.6363972749627435,
 0.12544706994712962,0.2981569026134545,0.400513107570859,0.7163000664643375,
 0.6436916589060031,0.4916066464755994,0.1564286209085901,0.5776827251715866,
 0.03925201076300555,0.13199130943015047,0.10763865457990163,
 0.08693010468793017,0.48006472480118245,0.20771405103169127,
 0.21972407306940256,0.5848323560254518,0.9650811866161974,0.5446165975337313,
 0.7002353759972066,0.30827794016500476,0.8911573732999437,0.2114578306809527,
 0.2697461064600791,0.21008878596561764,0.9571284635713861,0.61576816909601,
 0.9233678734999068,0.5355078328441345,0.057043799440629206,
 0.9061550544319844,0.7045728274087147,0.5857777401297896,0.6587573684748395,
 0.8123006163639293,0.9776558343311131,0.16139475381662005,0.9720993024463334,
 0.321386197881372,0.3167617865165674,0.3449585136402846,0.2523611422933908,
 0.7449856600594102,0.42709406224753477,0.5171610301868651,0.8022105348949441,
 0.7888234792606363,0.4552393738181414,0.246674204024392,0.3324455517513083,
 0.7073502714685469,0.22317039214368428,0.771500365522004,0.7142445702742644,
 0.3356624879358703,0.894750464215177,0.6666762883200846,0.23844454709094942,
 0.14064181324091918,0.5794017439107091,0.2217285960637263,
 0.11296434496209773,0.6538099046955552,0.004460276089750748,
 0.5870459066317871,0.45779027405144734,0.4362070897489878,
 0.055500900712186274,0.6185169543089882,0.2970220801090879,
 0.1821289577780958,0.864249769893771,0.5799454529314824,0.33992171994190035,
 0.42165945210694133,0.9299525989038591,0.0020610698089601476,
 0.6960523620126097,0.6526422625196944,0.17969805419651164,0.1411332406673541,
 0.599847676305748,0.23353302910973583,0.262195671244146,0.6144698553932448,
 0.6571899234389756,0.9328316804906822,0.13928378157130472,0.6258137386154671,
 0.08611276215650054,0.1232112177574336,0.2448068076892762,0.8813596640867268,
 0.8343161433917152,0.9821024469418439,0.4306060436617153,0.21130962013122379,
 0.18725720394106793,0.33682463655739137,0.4354433273994279,
 0.06934862576577416,0.7412067652060063,0.11109738534612423,
 0.34950109879185987,0.6506415923764445,0.8180086139081374,
 0.08573649476408718,0.773623430553786,0.047713683741322455,
 0.10957543665716807,0.1263578405217337,0.6667874919991651,0.9763243175464699,
 0.6329225921902107,0.49494146775292713,0.0702146514779628,0.8056520233525734,
 0.39772661158928524,0.5377044070200505,0.39133166262861363,
 0.8651164853777067,0.6121853592040295,0.787137183018841,0.19762889758241653,
 0.3400012956711975,0.5535654165790576,0.8245400122046469,0.8203286141931805,
 0.3764288237607679,0.16245907334374554,0.32195227559577566,
 0.9085029471656363,0.8341582394261757,0.5346440709722702,0.12969582638205612,
 0.18351080715248114,0.8927083619714717,0.13692264232412443,
 0.8289293593044696,0.5602791682449038,0.4721287716461896,0.3168076031194149,
 0.10315816390502452,0.47897496734259737,0.9574625921775647,
 0.42205673677721534,0.560661317610997,0.6959759556238204,0.3270054681191631,
 0.5570282732865139,0.17776523896879792,0.9717094683428258,0.9588403013744258,
 0.8105335154415995,0.7702100243915178,0.9497416928706583,0.6130152045788724,
 0.7211352581526306,0.42234664906355013,0.489494109878429,0.5063279687226035,
 0.9415237813922411,0.6126933749675014,0.057108592498266164,
 0.7159777184116258,0.6353487159735203,0.2819970852332263,0.11186240473543252,
 0.9467681164578469,0.30136574595720045,0.6096153237553701,0.798423230937445,
 0.8861420253081217,0.305672310273261,0.2720283492183009,0.8239852295959406,
 0.3868846533270054,0.39935231974866436,0.22062487013337428,
 0.08200994596756384,0.2680924057805345,0.49276492600640265,
 0.3893494246746153,0.4160701937908664,0.5375876465283738,0.6823185852264811,
 0.5405130976937678,0.42108308492035573,0.6425073994268629,
 0.018051137741213807,0.21774435493322508,0.7734393238299296,
 0.7880607981629935,0.911076945201507,0.9468665930091282,0.45117312891942873,
 0.5761949273309354,0.11738327600132692,0.1970090807618481,0.4814522506487251,
 0.09243886155568293,0.4070621716981433,0.7245151781810714,0.7690423391345158,
 0.9336000534385547,0.8491839137303651,0.20950932827449953,0.8587024036678264,
 0.13843289627572597,0.1326969097112427,0.9483138402057066,
 0.37711702426138094,0.7402098110502014,0.7360369429039055,
 0.05473113407631691,0.33622559137209906,0.6974953118563996,
 0.19046884926080077,0.31499632123667365,0.7086756976423074,
 0.37441320382197263,0.19177109496908995,0.4232222928224172,
 0.6985645444099959,0.09695030150259731,0.3986527976946531,0.6984910077548016,
 0.7559574589708025,0.20925398645420212,0.8089371432037802,
 0.36097447129186466,0.7283046944280455,0.2571419075367568,
 0.16833465364838396,0.6355630149404835,0.21608767876920032,
 0.5957025549313327,0.5333038339197732,0.6759081934609366,0.4037332334892807,
 0.5000989127092293,0.4941436386457506,5.184808509735017e-4,
 0.6922837128004782,0.4041944961944678,0.04372295416914307,
 0.28323940243476675,0.6157886131293926,0.32252389380659086,
 0.8874891945669274,0.16773185436100613,0.8933109032571307,0.8380351578768576,
 0.27041640421449886,0.9920386258303038,0.24566760462843196,
 0.28091182351057875,0.7452274505119802,0.25715699188596086,
 0.2502135934647014,0.6314768266072884,0.515150605462455,0.30592912130072725,
 0.4481445384156847,0.7063703570710386,0.818266229616704,0.21343219751258402,
 0.20137523054727702,0.024974342086493406,0.785446054275719,
 0.02037554049718382,0.11984976676053993,0.3696035861958863,
 0.6272466093281422,0.4665719894261755,0.23498863337505338,
 0.21446538996252773,0.5186657477176889,0.016631623439670395,
 0.04095046213410203,0.46220857732950515,0.23087335552885124,
 0.17780802336828105,0.7884432951577828,0.9076816923521484,
 0.35027822287406984,0.6187074008984492,0.51527461529293,0.11622163983543643,
 0.19612642166332328,0.4847309031909848,0.27743998042043305,
 0.5893401941422483,0.5851454500907701,0.5527156158606266,0.5217846506898755,
 0.14723215697494652,0.4937121867379467,0.5991432687084868,0.8093450957805063,
 0.08230550505139034,0.8377461036358114,0.5070310580742134,
 0.014972282668156534,0.50098001805515,0.9684739720175974,0.406407486383221,
 0.49357053783920235,0.6646166575043159,0.2016850299478401,
 0.18279995472762967,0.1821900917746584,0.3342969571108966,0.8478403453689544,
 0.7196853275735899,0.9993115142847052,0.6022781601259098,0.3891056191004504,
 0.16648259782332586,0.17169393403841282,0.6277809548284266,
 0.3606073975987679,0.20494703540790105,0.641778649123336,0.09876414560020108,
 0.7036185798696524,0.9718257794794757,0.16587252256527318,0.7614970556016611,
 0.1437242458715613,0.5227431650547557,0.10502049470626362,
 0.17802265989306376,0.3318131567465612,0.4167861063967915,0.9932322575868311,
 0.802179728446356,0.09825001583658788,0.19687688078900623,0.8888762922369255,
 0.3650143236796616,0.07432109149345534,0.8965604670590934,0.7682691666034736,
 0.662891065550677,0.3060521787651783,0.06271332161536813,0.42307292982620104,
 0.913631229635586,0.15217572004212387,0.9961379101554266,0.308441668949187,
 0.4204220505035139,0.03152354937613355,0.5519691898516048,0.7980527766211387,
 0.05594868896963856,0.4481540029495994,0.35726708612132885,
 0.7227786901723813,0.9494649034495563,0.24121727826297157,
 0.20752041748234873,0.7428713731555168,0.8977301164367234,0.5393639562008695,
 0.5537700861020446,0.8158570360433252,0.3645436619732123,0.08827267654832427,
 0.20601643547104487,0.45939912611813616,0.3267159090749092,
 0.30851411549712593,0.06216810450358834,0.11845332882332815,
 0.11699213467467917,0.7207982301470405,0.7044890994634484,0.8870103784458528,
 0.9964550554882314,0.4750542195103078,0.8325097733095236,0.7553455591970144,
 0.32442509046767354,0.3838925655592429,0.9130624073978219,0.1745757069783359,
 0.9134445403343145,0.8689516691373737,0.9403205253114602,0.8134784170847653,
 0.7611815704360667,0.7190375293148841,0.6136453317599966,0.708025816289905,
 0.6838623971070307,0.11757460076397896,0.2125338654119151,0.5217707603299756,
 0.8225331432748089,0.9955313200843608,0.7980006332750964,0.14545874406025272,
 0.8917574831376299,0.9107797998148022,0.5927953307612223,0.42383041973484614,
 0.87954977423187,0.7482895437978764,0.6930568668559786,0.0728376657205203,
 0.2368234924545839,0.27179333952600493,0.30948966108755316,
 0.7760792040848563,0.9584501657243454,0.19414620278498507,
 0.16545660476536583,0.11386690311133751,0.7479185241464815,0.305680817038548,
 0.9455777190897648,0.2058514310121491,0.9957979948411609,0.08549588425880617,
 0.3717399694421881,0.5484773209132329,0.08010547895563358,0.6807484227032824,
 0.8317310299980836,0.0953972360107056,0.1259003531513453,0.06745086313093973,
 0.6925183059855868,0.0829020820194123,0.15165892739625697,
 0.09526940907462422,0.14702445333425196,0.7863105656914378,
 0.4068954484556212,0.20891414713846956,0.5738209341549365,0.6012360111052145,
 0.1809939651673731,0.47433970197425135,0.7525701615136491,0.5040746231396642,
 0.6503283770400567,0.6787957310660044,0.8518620414234073,0.8852634430164189,
 0.18445823762944302,0.4383033169290218,0.2788889172885324,0.7582186675844487,
 0.8058155501579087,0.7689000076734188,0.23182480661706062,0.8636499529972091,
 0.9759349904559234,0.33429630259149945,0.7239590063612037,0.9069988584673105,
 0.32270757894722335,0.9073119121364073,0.1216455116389048,0.3373492481519865,
 0.01208476548874926,0.6685727559825194,0.298605650379992,0.03369788456757228,
 0.3376842553753605,0.9265698186838822,0.21181154555394555,0.5692777686246617,
 0.7090450164827383,0.2961975285905196,0.33850166135049253,0.9210221480915318,
 0.319266824530161,0.3174819342890064,0.8457901302543195,0.4557161911469546,
 9.831899413380585e-4,0.9280203082101305,0.6240574122339542,
 0.6126291247589457,0.465643138928295,0.9149195115592179,0.3931592040371469,
 0.6136071774589866,0.40768375551627134,0.7435850853300537,0.116115598211329,
 0.31125969806061804,0.6945802715582863,0.8132832878424459,
 0.07503281212011659,0.05573307105131242,0.8234312404234195,
 0.23500689855800672,0.9845710523024676,0.9196449503782139,
 0.029082561381795213,0.11855122416563302,0.34876010156155757,
 0.7518341694165751,0.1399568419610907,0.5750957921522516,0.23946516749996394,
 0.8119155949102046,0.32105830268683877,0.9547814376791349,0.8254606501397256,
 0.746778763772561,0.4208056266450192,0.10778423383168312,0.26328785971576296,
 0.2666651688588575,0.542574892961219,0.33629237883179663,0.8659438913167867,
 0.4073610660011677,0.8236448657491376,0.5153724882248057,0.16962139609131055,
 0.30143699063336504,0.952941074858175,0.9777450945729657,0.6145869528974798,
 0.95614291290832,0.44256601688272257,0.3322163940660259,0.552739226383446,
 0.8399589364273127,0.8915672074571144,0.4002223137820454,0.3643657020351462,
 0.5139588929654154,0.316422267282092,0.825796968988808,0.13205408484713876,
 0.7161052029535648,0.4283576492086638,0.7769149503263624,
 6.250197724988471e-5,0.7001546646208564,0.0037866243423931856,
 0.15661738184176432,0.2518079021759535,0.16957120254228175,
 0.6394865653338118,0.7531526563454233,0.8526130401471909,0.04358322241442403,
 0.009877659052133558,0.5287871747506174,0.7198131096228464,
 0.22773691169788624,0.895409526981544,0.35151896501519175,
 0.05615043897754868,0.33059021413461465,0.6185111725693515,
 0.8416116259467012,0.6010011607876837,0.4534951116140783,0.7673649497017022,
 0.09650092508315655,0.23661643121621456,0.9045770685568882,
 0.5420313639981325,0.05940701603524012,0.6980589258075778,0.4975464254177683,
 0.4285361844475851,0.8953797153352169,0.691993832435068,0.9014763742130586,
 0.6774451830872605,0.4528678550978834,0.1848138783418115,0.7382339486594149,
 0.3887168455303318,0.5346411908942057,0.74941127851766,0.6269929304312125,
 0.020981438157467225,0.41050372558994674,0.050565670429751686,
 0.8672941076771432,0.7004123186778117,0.4836664405129787,
 0.039887766586772955,0.972997403811743,0.30843964271016133,
 0.10375740913230602,0.898136831247732,0.3124739761378943,0.9386640401062767,
 0.7434859107682098,0.4749402938706019,0.5632781582483926,0.9265591451997982,
 0.30901994820630496,0.5873941651162723,0.7696523060216737,0.9899514122106137,
 0.18805601860128385,0.9591915935938986,0.08955114109916895,
 0.009313390505946595,0.4312520246873711,0.23055102830024243,
 0.7557869365710963,0.5754843235763615,0.782849211469825,0.9157808300772871,
 0.23952386456894614,0.7672218385704991,0.5151735590221359,0.8126504430605257,
 0.4657211569245142,0.1868363629676595,0.2901013775325182,0.5490005478400254,
 0.947884037104784,0.29565624963443093,0.25164370656942525,0.6067454710779069,
 0.6584196945931888,0.3678652990256326,0.18891896575961897,0.6114391456634956,
 0.4509284746279051,0.03914557091299875,0.6107718213360823,0.3384349025190666,
 0.3555106633907381,0.2697141689765592,0.4364451581248161,0.387532764332738,
 0.9179065119494338,0.19442118474009973,0.6875636351689861,0.394436644844957,
 0.14928099518677995,0.6285128004011824,0.8337857352707633,0.9486678627686038,
 0.7994075930598112,0.20530241209897993,0.548927298591266,0.5101673583216989,
 0.8511906872443735,0.5892370250297164,0.08148650555501902,0.7869705991739054,
 0.1056600881748857,0.30417196285375536,0.09098143403484293,
 0.7891817004027722,0.058739671430321105,0.4102385770672494,
 0.05589982533910931,0.33405992607205903,0.46704712043903207,
 0.15664184756542465,0.7179239007008039,0.4990691699236618,0.8870313831809005,
 0.09323014840636079,0.7179658370069596,0.13439037080398852,
 0.6941988209370938,0.1879587405242149,0.2196406362670793,0.6702008402829356,
 0.12222342012579657,0.8864920549500348,0.07657646076249414,0.513055442753742,
 0.4691264429630819,0.14638188861340617,0.3049938754193896,0.7040311434255118,
 0.30728346163279907,0.05730537577062922,0.59542534966901,0.5419963442675215,
 0.313502740157962,0.7516803673466993,0.32529661411889343,0.65120069342749,
 0.2588007517079116,0.1344711318978914,0.23321492093471374,0.1483702547624658,
 0.03513331707501954,0.35365006493252593,0.6193556840083695,
 0.09042771671269223,0.2277588061705521,0.8519664049127984,
 0.10935167800330592,0.6243083648870095,0.5460809310694883,
 0.22408245122359383,0.7288198610246448,0.04591395947263999,
 0.8696527606078904,0.014512323991281662,0.9400215881638674,
 0.7429501305943106,0.11391725383982909,0.9123613548319043,0.5895209303254586,
 0.48722665350655725,0.8529677099142247,0.5888855564101164,0.9251434667089548,
 0.5307750839975167,0.3401513822468112,0.6766979405131119,0.3703643980022492,
 0.8436056063832845,0.793773154361079,0.7751079359630489,0.8543889866998448,
 0.8072995206710731,0.4029912459005568,0.6708449681886122,0.9815175328084664,
 0.557221211320453,0.42407925649482103,0.23240667046311358,0.9587912284264722,
 0.8190004546183769,0.5541077482554675,0.06467943828867873,0.4630489364120881,
 0.700792847823329,0.18445566874357544,0.6214279342579468,0.8604493766412205,
 0.6084144221055549,0.05166930566240879,0.10124847096126299,
 0.26564174896280657,0.024850937926326755,0.6904246010847387,
 0.11066090041501575,0.4578593238403783,0.6542020435630671,0.5299909850029901,
 0.179931327277716,0.3880940454762726,0.12984142887386085,0.9288347272569406,
 0.7717276451656087,0.5626503519982872,0.7212412004190205,0.13043012131711507,
 0.2823200686759264,0.12965960595604065,0.6498822475618291,0.576338767756604,
 0.08835227907487853,0.8651265691628747,0.4266236483466225,0.3257489664409098,
 0.6388372886176706,0.5535575735229249,0.8066942035429079,0.620265410316676,
 0.5772151196003533,0.038243627879423925,0.4829305567718938,
 0.22066240052012265,0.2445874152596783,0.5100006373082612,0.5300136784870453,
 0.31202013552472263,0.9425964875577599,0.44737993162925327,
 0.6243229383124685,0.6700995405625613,0.645363432458471,0.16128006541902484,
 0.17099708243231537,0.04782735033078833,0.8308666048529018,
 0.5775055508218483,0.33056398187516267,0.5921248128743181,0.5912453739006667,
 0.5790448890824105,0.8826103915344123,0.052052792914551005,
 0.42636239463673253,0.9435594437207331,0.22249553945880718,
 0.5938485530072368,0.8949526296074922,0.15965651815757498,
 0.31473001804474987,0.5088541933098185,0.167819873401007,0.5153764196843862,
 0.7115968047020041,0.3074586672591192,0.6303287542889253,0.47661749134564424,
 0.8614893526284536,0.3548129791422796,0.001497713978632631,
 0.9968249144970921,0.24088928557965317,0.44613271098072005,0.706608106905782,
 0.838367154451895,0.9476170458992621,0.8037049350142974,0.10496187507989674,
 0.38345692639850737,0.2031773875326599,0.9534571854897607,0.9333447455125334,
 0.06919647535986151,0.43746339525684763,0.7402409326101229,0.342486227835479,
 0.25817326134918317,0.2151675091116232,0.704843355725625,0.03165288436771352,
 0.9252617303189298,0.9494366233136782,0.22296978537329337,0.6096977556506299,
 0.3639649421277016,0.4619485026622706,0.31446262821783566,0.2754747531258306,
 0.08405724665685321,0.0079686965052298,0.794782316945046,0.10781281763861361,
 0.13786937608906547,0.7055962895916572,0.3478972482691428,0.5461921654539521,
 0.2449875160260816,0.6083282832255984,0.9186583595833931,0.8447658937920605,
 0.001966451867851271,0.6768209641204553,0.46421110102042507,
 0.681497478710093,0.16061949635848827,0.04363488201174759,0.8602309261120858,
 0.5226433283882437,0.7065648303118134,0.10183607505417924,
 0.13426120120454854,0.36437806136427875,0.77857714488837,0.49588839989961886,
 0.5399946536381095,0.17656548479425338,0.7744204419656262,
 0.04478772798712105,0.016024566193869272,0.34953107917746806,
 0.4530673968037676,0.5309126535239036,0.23446934529015373,0.4582867366024729,
 0.044224038236473584,0.8429393296923998,0.5708896352223757,
 1.1987489729947676e-4,0.6771046036258201,0.8329073643150006,
 0.907336072736913,0.6701584149873263,0.6493199629506181,0.9717340292002211,
 0.7770722452401356,0.21945967619321027,0.6276250837574636,
 0.0031376979397913507,0.6060165724592246,0.9966123157134273,
 0.10222321072386176,0.38735936465224674,0.40458790584705107,
 0.770137844616442,0.5998885183080712,0.8881515647641195,0.5295138932978921,
 0.16176466284267632,0.1977928674303403,0.47128501221830965,
 0.6339158609752682,0.11884848523848801,0.5991788659369521,0.7963361517734056,
 0.7899094504156542,0.04781497365151299,0.5169692412896495,0.984865066033619,
 0.8921500184382913,0.21694803369636362,0.9080021031409012,
 0.14236393518396295,0.18200928797580096,0.832122005188245,
 0.05380805139708267,0.006609222131078729,0.5492661799246041,
 0.4623522528837748,0.9967982051379491,0.8661764447354396,0.8869563833053592,
 0.8114732678215193,0.8755965912108954,0.47063510538227016,0.5516989302961974,
 0.2620162368358465,0.5408310694823213,0.8338102494530024,0.12234959318605876,
 0.509550946305674,0.3638033812595233,0.31589643934636613,0.42580449166201173,
 0.36037914068756416,0.5903904953529228,0.4898305651145505,0.6346453994654145,
 0.9659142592015814,0.030301650202300023,0.2541603939147956,
 0.7309773260202833,0.22774856179861536,0.2594415523679403,0.4418270961040608,
 0.8987936318253256,0.4274716839077797,0.010239562053564022,
 0.005333624574058615,0.8650140957343384,0.5502468505862945,
 0.4375709912889675,0.004974404514277775,0.40242778656030964,
 0.6929005753356039,0.3412636690708679,0.6352534718718826,0.9859479110487299,
 0.4266477895508599,0.8364949311399096,0.7681584679721148,0.03541227271261249,
 0.1175652264034649,0.8934041580562788,0.010980787378508272,
 0.47641196872207137,0.11382959664904657,0.884166163552552,0.7924231518661663,
 0.20872012395256068,0.8528106781044854,0.025497846246733236,
 0.9342605011968899,0.9337282763861126,0.33497661769021714,0.9058529300360826,
 0.05132828745237317,0.0018558291046191222,0.1835983597410069,
 0.1611630772365047,0.7406763992846978,0.9497356310282289,0.5717109683365457,
 0.03224262565963054,0.6319515043865662,0.19052299590271216,
 0.3423258596775067,0.017823888685557288,0.2012116502486503,0.627603673411006,
 0.07760678550302968,0.01352001766010913,0.0914262937528183,0.240047005931178,
 0.2575049779650911,0.08057437609211426,0.13076840085393293,0.374615764347503,
 0.995869698273685,0.6565629397300267,0.3931824338065897,0.4143853135967426,
 0.7802758040317268,0.5946520467050435,0.4863295244732765,0.20909521714462187,
 0.47731234683060886,0.02433539961071296,0.9833988775526931,
 0.8962362373838536,0.4495440247889618,0.14238342461124365,0.6984444872725685,
 0.8663476728023874,0.21157140560162502,0.9544203040766981,0.0285219231979148,
 0.21704816322838527,0.8527054854568814,0.3224047824878401,0.597065069016578,
 0.39318075525536855,0.9295168965839972,0.08832987420251803,
 0.09758247765544548,0.9698925069887788,0.32604418148047576,
 0.7669886454559086,0.1936511393300293,0.18210065033969403,0.3966701155391842,
 0.6374356215489849,0.20795086798649842,0.6916150473864181,0.1894006244234504,
 0.9951987808944235,0.5198952562785719,0.7009445176851132,0.2701387817091746,
 0.6466923381691013,0.7422039402809466,0.020074906506711954,0.265632364181273,
 0.5061518881209224,0.24052114744913378,0.5947461842190916,
 0.39988014426250784,0.2801814152184312,0.5119779121311867,0.1586246191630465,
 0.3484549011794448,0.9571006619475639,0.6065767965679301,0.5972520577565649,
 0.42015276425590364,0.896412430327093,0.9796215010031148,0.33746532768122073,
 0.8962675542369485,0.2022229029363214,0.7171153220920377,0.6706878567411295,
 0.6609340132425956,0.7799256224946292,0.16871592602705454,
 0.42896872590548396,0.8488762164521199,0.18714564657915767,
 0.44997583888329673,0.12580809512698743,0.36788895710274594,
 0.8914192622210166,0.4566947027715681,0.20634942633952236,0.8830764770733036,
 0.531466287560428,0.6498538867780402,0.9417965534818069,0.3529990829761417,
 0.42523238099944793,0.7153679429223533,0.923509366633734,0.21288661116150565,
 0.76276593861989,0.3429704778506951,0.8539233090642218,0.6182041456168992,
 0.14027583827838896,0.6445939100835504,0.7431696416524993,0.7674703153225753,
 0.5636940538263604,0.0293423997066391,0.9988451037068822,0.681790852953387,
 0.853879535231068,0.011373084259623023,0.9087399590433611,0.2523781458042188,
 0.5186529550925325,0.06522220929106104,0.06943352668171865,0.132471229750867,
 0.7911452850677628,0.10976382746604252,7.934127609807717e-4,
 0.5140595225077016,0.9502259670471169,0.23833953590674484,0.6124838431744386,
 0.48027246357281606,0.6820165279172415,0.8903107550132259,0.7910171594733078,
 0.6549576102794887,0.7668855613977971,0.45078766615602306,0.9754736709409626,
 0.8492491454323162,0.9929139202082822,0.8609153265426932,0.20997274417717549,
 0.980644358132783,0.30603609723795977,0.7869110558560486,0.5970874656369849,
 0.36402075443350324,0.08752253718742375,0.16709342367543845,
 0.09287791816040325,0.9832225917515008,0.8756917891373148,0.5456894737994222,
 0.285671538570465,0.5717867814733842,0.4607662960846246,0.2415304930766019,
 0.698651022393984,0.7364293095298697,0.638769457045339,0.4714936385266114,
 0.3193279007452874,0.4931715514488503,0.61478582596079,0.09189930292555948,
 0.8615315115742608,0.8076444845859553,0.6951092118637056,0.822405066314581,
 0.9963003627398249,0.8379033435611501,0.41490739778042574,0.7630340305422671,
 0.4402363408568096,0.7476796821590558,0.04095609501553832,0.2772741022048042,
 0.21441902594600148,0.06529435771439052,0.5061093902921426,
 0.3792772054222424,0.37561479955357024,0.4059829608517125,0.8189435020989008,
 0.646429360883724,0.2631185849164074,0.4315217511304801,0.05012634333185528,
 0.8538286647190051,0.6004426719412138,0.6963124054025084,0.18926652804498567,
 0.08286834794321074,0.5160991034168232,0.13021930972417706,
 0.8769307831404273,0.48199061165759605,0.9159931474708483,0.3532693232444769,
 0.9686969806009138,0.47282258207833994,0.8140305989178123,0.5823656121669343,
 0.4735748377221892,0.8353508899659274,0.31404723465212525,0.6298562194143748,
 0.19316326594475597,0.5911842777210113,0.4275404831305972,0.2147436433858657,
 0.10476056338486917,0.2805414677368073,0.06803202298861943,
 0.05466524042276777,0.8300285211271314,0.9493380990814909,0.9661567603831658,
 0.4820961322598387,0.7528337905316321,0.9499236070000858,0.0408341772368751,
 0.7327078845354754,0.25098494788293857,0.7539894330735215,
 0.18752837305477843,0.1529081454528567,0.16706868943565112,
 0.40356204549559305,0.46803138465721394,0.10905598624671098,0.74285007709252,
 0.7579296410248874,0.3531866642409409,0.5106076691739972,0.1518061996520661,
 0.3950892623534915,0.3825896108699589,0.7059736714999505,0.9550524152589472,
 0.7468168870939128,0.05714350866086837,0.563439397338286,0.9559094855613354,
 0.6913303469031464,0.4460493973985885,0.04008254969126224,0.3260203973339585,
 0.02150218727985842,0.5071500045395633,0.9048898297373139,0.9757135755313201,
 0.3449010502453981,0.3219071423012678,0.8770453238355318,0.43255568471521444,
 0.2301528682542754,0.04283098546972042,0.9640177123295837,0.3465561283891334,
 0.285610533028137,0.7997321872194206,0.11695351806310661,0.21077416781176095,
 0.1883620166604455,0.6884848735823711,0.7272790712258673,0.09394630261802706,
 0.5304535802268789,0.4622502637319036,0.24650792679690314,
 0.08591858128651764,0.005710547660420895,0.12512307828423386,
 0.6539115168367586,0.2822838509360329,0.7096461853613808,0.31801295621713743,
 0.7685585268358015,0.6404427741139148,0.7015236298241099,0.6492884918703138,
 0.6344035867548798,0.33810077869837263,0.8703924518510082,0.6720822528024829,
 0.7055185247811279,0.19224896885037246,0.4893305898853493,
 0.29272561318667234,0.7252626885244569,0.2788978382224039,
 0.20627178033287585,0.736173251891426,0.9671544176543844,0.20314992562845213,
 0.7487873725953668,0.17905049313912658,0.5070725808682065,0.9250251477193702,
 0.8481259865843429,0.758974544872281,0.09656335288477058,0.7474505422856494,
 0.0894936879177678,0.8650853672192896,0.12760666158777623,0.7369214988448823,
 0.8086410114245611,0.41137518903459924,0.1535457865643084,
 0.005167639292264159,0.7905904047014337,0.42435968189572715,
 0.6283127512013816,0.9773351318290356,0.44907372091851827,0.6919148515143889,
 0.5703597528713149,0.27418788333412636,0.5859521131814939,0.7435661529901133,
 0.20779411283144777,0.1980974188312825,0.8989486263753542,
 0.034167970537919645,0.8404663158599893,0.8568914804948004,
 0.5086307505525222,0.19561751294331697,0.04112948880378353,
 0.14765627555022254,0.6611694829960427,0.15854156691827714,
 0.4137940547234771,0.3590981068314727,0.17374581345636775,0.6722433970157543,
 0.14250739029872905,8.387570447124792e-4,0.8644320173917873,
 0.6541280819564026,0.1742115837237309,0.5945375440971297,0.3247677281660526,
 0.26736011426190887,0.8036774822980821,0.7068288544690916,0.8052727304577861,
 0.9562806669401412,0.7708720595079759,0.41947722672701326,0.9732159680829039,
 0.5580731052856213,0.899901461640193,0.7377701087946027,0.2103466702318627,
 0.42515879262174183,0.08970585723078683,0.8760798624787114,
 0.12529122684930105,0.7311911346202733,0.19439539676543172,0.863007870624199,
 0.878434221484839,0.9707223359880983,0.7586782280590362,0.04246787566677468,
 0.8797724897845667,0.9294543762886179,0.6261787630063498,0.9366110912952079,
 0.21327991649814804,0.8292172587665662,0.4380564701379759,0.9125418227059932,
 0.2651256052571185,0.7621524278945088,0.22747103675080604,
 0.10339795860773493,0.7837234641083075,0.9968204586535719,0.6481270027177899,
 0.825423044329307,0.8090589564463257,0.4177355347469601,0.7756395230071037,
 0.14830193764284372,0.6319524761559543,0.9438907623642689,
 0.38491327651207874,0.46474641336691236,0.07572526018904946,
 0.5466275087290413,0.45028452203601055,0.27133372338134865,
 0.7835418020205145,0.3264504737836269,0.9168223422321916,0.9674747030704994,
 0.597592446497968,0.42330439547232923,0.2549168456864682,0.9356768025693135,
 0.30965858579930716,0.4900075758327378,0.3503454122765237,0.6930116713466761,
 0.744985105093278,0.19321964569899952,0.5631477292779646,0.7346856309073849,
 0.11255117681664006,0.5300302950883853,0.1453445761715998,0.42383893377535,
 0.42965123714886544,0.864665177560932,0.2565619089484753,0.21452007357271963,
 0.16832844059275087,0.9553326012084795,0.2299205800412224,0.7172490329555946,
 0.19800289996618048,0.9050404155531199,0.03893160370953985,
 0.16912846702025097,0.038724640119436105,0.7652344437607067,
 0.8161129104458238,0.1552368708373556,0.8293676866623197,0.7177648109805671,
 0.13353124261204463,0.15077915595268476,0.14155620474276276,
 0.5356681880708472,0.8592377152013178,0.7419757712780858,0.21064682804797086,
 0.17643140809064306,0.41800154985691473,0.8835648283394482,
 0.5163339536171773,0.604566234992701,0.8472922980972556,0.2680102744851731,
 0.04455353158267994,0.21398746090080967,0.9051081519567986,
 0.11818025928235387,0.6817201659592071,0.21806111341186774,
 0.7877090995274583,0.8482843413756105,0.8607482318993754,0.7196984393525087,
 0.7102737917726891,0.262334517856627,0.2836233652544372,0.759128387604227,
 0.29545620147597307,0.5195239427202532,0.43364300345441076,
 0.9480543322218399,0.16067304661521775,0.47885660457601964,
 0.6428949679523849,0.13021952503619705,0.13519506834993922,
 0.9038218799487309,0.343443631519885,0.4347314273910504,0.4279701930800619,
 0.5171670337980393,0.17273524929615758,0.745049283912923,0.737568212669969,
 0.5522532224997142,0.30884184834111816,0.4135491568643016,0.9815701956408563,
 0.13606669837946317,0.11230801421905567,0.5711397392753863,
 0.6847716020823947,0.11892850330224414,0.6446056266471645,0.4794852012683052,
 0.2852832910865484,0.7687200870649403,0.6965917250166136,0.733305071717204,
 0.09982178750058468,0.17956859216662713,0.10396762177555574,
 0.801150167839133,0.885011470202135,0.7697314006176507,0.9156451410785592,
 0.04535185534630837,0.436945896199484,0.733351837432501,0.2012894706314906,
 0.5806891623277518,0.23341018687020998,0.31322906662062766,
 0.8746803060189458,0.3326538347260166,0.2452886652100419,0.40663785388993856,
 0.1316424610261201,0.8552323690362194,0.5771334233905251,0.6818533114305446,
 0.6675129470769542,0.08406297558562525,0.6965039293528643,0.7126458740643833,
 0.6602549317221149,0.7020065695399363,0.9201300703170513,0.4623772014985197,
 0.9065713167585043,0.8965875999341123,0.21018609837548097,0.9275881664984293,
 0.5375525628890371,0.991434396368562,0.5198353313237273,0.05870872383417003,
 0.9113698122615512,0.7679831844881922,0.9141682148719726,0.7080948500166184,
 0.7730926288671269,0.4160726363420292,0.963925963107812,0.3982853173632557,
 0.4231411032757314,0.408590186705193,0.5586905516604592,0.35648190058669715,
 0.8471610703587401,0.19089497105118636,0.2883670849919103,
 0.20020530398209946,0.5232165281005963,0.716927163286228,0.0136995530140116,
 0.8617577361377382,0.3695336902058808,0.7220006366478093,0.5987795576294219,
 0.05001943038989909,0.40094744969716056,0.9915074278074921,
 0.49778556609681157,0.36922218079538294,0.8339375366153822,
 0.2687473355894916,0.23212715624502667,0.6786409932961823,0.7289432152976121,
 0.6784637908119977,0.19853671739477485,0.48719929515830396,
 0.7179585235868933,0.9570964320712196,0.19118203575425963,0.2518000813636905,
 0.10067884994046339,0.4463299676353958,0.7318856326965911,0.6486778266634893,
 0.8367124943660933,0.0434669851196402,0.16851840671482088,
 0.19301156405083975,0.8039337041184841,0.029940671653307827,
 0.28019083345267015,0.09452181927544767,0.328796484576223,
 0.39253164602096835,0.6151260440582847,0.747591090653515,0.5018470727424944,
 0.1939353349173284,0.027531544195133928,0.6858729620374756,
 0.8558097061240716,0.2317735767166068,0.8270090008063931,0.14982557744719893,
 0.005908110050955373,0.9385387545846707,0.10423158738159866,
 0.8415184838253436,0.854623808461352,0.6466468807717319,0.40609171320096005,
 0.5959739720914291,0.4128721301631879,0.8250328344528599,0.9983121276777611,
 0.25809029506829606,0.8193709815230465,0.5887211003150437,0.8413781509940513,
 0.007288779118341582,0.827818966463827,0.3578627146897019,0.7563602626412832,
 0.4486398117541376,0.6611515271938905,0.4137184004633664,0.4154780642635837,
 0.483249816211033,0.5558761278722307,0.41975800011143893,0.12202925147671062,
 0.8196916235412439,0.12046988933333114,0.5914565839337511,
 0.009166146648410423,0.5033823549180884,0.6726993741104044,
 0.03932095109815714,0.592489489150706,0.6616937574050623,0.8393354018428266,
 0.5658819150449566,0.9040181696250297,0.5435113610069177,0.5875528336949947,
 0.8116944358763167,0.8851581206537353,0.6537329852316173,0.34919412524536764,
 0.6825939219651199,0.5218927101553108,0.9349436178692186,
 0.029913566207829922,0.30978376857213064,0.37482103336872075,
 0.8800285334371869,0.7272069930290406,0.6535771745520987,0.1675096810605371,
 0.957559889811663,0.0374931175504295,0.8635427500727226,0.04451398754775304,
 0.5873503940612949,0.15571074214568537,0.246118465394069,0.3169619809803278,
 0.6857841322417149,0.5432521028852727,0.10173378898919794,
 0.07581632543519468,0.7433804396878221,0.8163554871634826,0.8836328923205674,
 0.7354307922809549,0.19128164698764893,0.08565631141066099,
 0.6695137565150738,0.002980326435375935,0.9527960129734195,
 0.5814170705918273,0.5296363585853787,0.5600727069097284,0.6503266124230584,
 0.6821622176169215,0.4562972354877707,0.9868543271080563,0.7219068095027825,
 0.6773028389614311,0.747401116200513,0.09298586396194275,0.6777155590510042,
 0.28304788665802105,0.06388076264835474,0.5357648046854175,
 0.10192341023808582,0.5797211729872167,0.43669140982847743,
 0.7905861276710318,0.4172089629070297,0.1704307573262489,0.8553377916793862,
 0.5409136281832785,0.8770901999851747,0.4572899551498627,0.6893139753395322,
 0.38172386289362115,0.2570421160811117,0.9932825152100809,0.6553388481857414,
 0.748521647070229,0.018509403618826803,0.1974647978825519,
 0.10918125037609827,0.2606200465480585,0.9788040072089027,0.7502175771638318,
 0.7455983127091177,0.7958236096616091,0.6649539435557494,0.34006444563100935,
 0.9253351453992257,0.8639251841928964,0.7943931896343012,0.6637419882422809,
 0.2724706891996901,0.1283388244876671,0.13698750071506394,
 0.02491623588845937,0.4095763588737823,0.07454273185436178,
 0.43629614293740593,0.5702412266251324,0.10641102323822638,
 0.9837021485473886,0.8090191029682793,0.3815348365976796,0.5324653414395473,
 0.9439346238079264,0.37868573154803253,0.42124922507225204,
 0.2508171005872739,0.1622497694853826,0.7161493252984431,0.09338206263505988,
 0.9581387830465724,0.9637733163187959,0.37253467456202294,0.9639120464329405,
 0.019650101905124284,0.13493651374910165,0.8393208845146236,
 0.4869912425775268,0.11618135502053784,0.3166274291644644,
 0.27405495680610237,0.3215200574924715,0.6665966106028112,0.861876773552009,
 0.40601435640699424,0.731261566160067,0.8147856371618433,0.3164485721331075,
 0.5881347431277358,0.6284791563783821,0.14389529415893987,0.3425624869067849,
 0.052721329318029864,0.5749499678205203,0.1838513401214178,
 0.16401342080953552,0.8794304947359287,0.3328735451583684,0.9074202151589641,
 0.8919764372380452,0.25344738455220117,0.1907846701927275,0.8916871063913158,
 0.9570806340147864,0.736639481162285,0.1565531187181738,0.5621170764873692,
 0.7954883753923382,0.3603842765968277,0.5543906274642608,0.4436328925274634,
 0.22317478110886424,0.03191799673202811,0.21381849428944943,
 0.4958168815821703,0.7853123816715191,0.060901853298397035,
 0.6707518805110757,0.889710701107702,0.40145670824163815,0.13055122589132273,
 0.35928390903368934,0.6481690325447874,0.02478360384083378,
 0.7726343808990661,0.3588578896139498,0.332881323656562,0.9964655222843443,
 0.14112469211197165,0.5905016573812005,0.5285105873062376,0.5479623984394874,
 0.32082703877758445,0.8276807744857437,0.04689698577275614,
 0.48144287931308616,0.9425093345541627,0.42840734292129257,
 0.9285547460525257,0.58880587215277,0.17683378898460667,0.2529462018627471,
 0.16633171102906985,0.6439641080723403,0.9319270809075103,0.3545331508153483,
 0.7181559211790226,0.9829226695656702,0.2423591431401676,0.5286092990409126,
 0.6139471561821055,0.7203956372936515,0.6716504662601135,0.39532615475171706,
 0.5793565669385379,0.07972636396858623,0.8041341343381313,0.8271636160590514,
 0.09990713252713224,0.7051766307499909,0.13413412937519498,
 0.3382684270954771,0.2708847790121751,0.11815420122388576,0.7670343282818223,
 0.29799682748473033,0.5961209753857732,0.6109885848912814,
 0.058702961529230935,0.7757309104034327,0.5202356912914089,
 0.6122003079106503,0.2249090368373463,0.0521624397886733,0.5453839044600767,
 0.9486623953419517,0.25481986069144313,0.3607080411556519,0.8034146686487083,
 0.08991990721429521,0.2752513998014442,0.614854322975221,0.5192933139584844,
 0.18620897793095814,0.10476979898797123,0.04744676498331302,
 0.6322097731435109,0.13439392434388897,0.7167285252310023,
 0.17031247459630128,0.313436137102442,0.56706836750306,0.6026241711157053,
 0.9184992096904386,0.7980605835422252,0.7195835751546976,0.5892780032363826,
 0.46567295258153507,0.8167621231562883,0.6743814210550682,0.7342079918499919,
 0.8077370841683478,0.177676851240814,0.2549699505902757,0.12519003969632359,
 0.7885815633052337,0.46919066415587274,0.6696413155084919,0.6218516099541848,
 0.8973382633958051,0.872330654211862,0.17239139284863647,0.6594398448520538,
 0.009337768522024725,0.26983935318179986,0.8751993667986873,
 0.7574057372469574,0.9344164440763802,0.5961503263351545,0.5341890614090179,
 0.4655003358440054,0.13510751099651286,0.4175648151811704,0.2259988597396857,
 0.6421108544970696,0.7259878463335455,0.18023938522898408,0.5368987336733796,
 0.6279559038207185,0.43487514589466125,0.8613868159212387,0.8389021808020791,
 0.3246107573223075,0.3044607656139562,0.2993144468155896,0.619038057671085,
 0.8852986913003087,0.37454322463253975,0.6165878316568706,0.7235913664722514,
 0.8476873337620915,3.0378816278153753e-4,0.40763436680100196,
 0.044083158246440046,0.5064970988562248,0.8435635285590318,0.869739271082243,
 0.9131524238608961,0.9486595441074195,0.9235549923903573,0.7551844239235568,
 0.0038116811542334705,0.8386679782688803,0.7749234252213096,
 0.2620152748157042,0.2080160509301353,0.8238111317671077,0.29272913397205824,
 0.30222013822818805,0.5007473606478925,0.9263346848050773,
 0.43947610634967926,0.7690913262228091,0.7412833089056952,0.7531352471070951,
 0.2438891704096311,0.6520120756459106,0.6918330020488866,0.4912149979706647,
 0.4940050465611385,0.6667896242645904,0.33615798261000296,0.9818519696523266,
 0.8014687291974055,0.2756297600742609,0.18113077114116993,0.4113128957287585,
 0.5256583307826541,0.3806093437719261,0.38447678054326373,
 0.08380966496800879,0.8552358338737025,0.8235976556630045,0.0722607481009252,
 0.20955239695277905,0.9665137208716588,0.003291682368620319,
 0.9581057189620023,0.5016503892153094,0.23845255560123557,0.6699776303900676,
 0.55838412070075,0.4445954552877569,0.988398330475853,0.848997693978653,
 0.4187076073515722,0.19179150621066443,0.2609113056443191,0.5207332070948589,
 0.2882415879184248,0.9744078679766954,0.1529932303970294,0.10895527583112019,
 0.9068477115275405,0.7196606517677926,0.06390983591755495,0.3698875794536708,
 0.960894535927394,0.8757680306100243,0.11637950786504103,0.5046335090745606,
 0.8681160629504313,0.6833765509364068,0.820123454445039,0.684816152068433,
 0.4660349154089216,0.7209235196158116,0.5985896419965309,0.8504244817046643,
 0.9121209324745695,0.8143880672472323,0.8741495742704355,0.3496420093338556,
 0.5591948701680063,0.3777289143179554,0.724174319474493,0.6123526857914073,
 0.23362585214901332,0.6350493437235338,0.9220730650014948,0.3194253508369881,
 0.013599909566716195,0.25954542677606407,0.077512287584389,
 0.1714687004994736,0.09416722856524518,0.5638091302163242,0.4559719123530761,
 0.8416526643666415,0.41038646695525016,0.24275937892461064,
 0.7217563855087886,0.5946028208446107,0.10457876025605428,0.5790159553164402,
 0.066819638131008,0.8278759109865982,0.8119919478697932,0.36326923500000585,
 0.037728993090728635,0.23385759414890384,0.2841230107161147,
 0.8053946789047719,0.5060162132356617,0.8401015506239363,0.29461414224874494,
 0.9438354772474501,0.7035285492002668,0.34159671636138456,0.9501863425887029,
 0.36318021232812414,0.5257296323829548,0.90414157119765,0.5234897077215425,
 0.8375712370806871,0.2540480996199007,0.9836040026858475,0.7362350760897015,
 0.48703423134361823,0.6455689404789506,0.8332940692532834,
 0.02259730897466239,0.7311139215567709,0.8272616086689921,0.9184830697408394,
 0.5084506252851737,0.49038763181895584,0.029082391254892204,
 0.4582154712790997,0.5477352121496741,0.5628519140764425,0.3826343419416296,
 0.6805797832792085,0.5356323471412012,0.7759364910298662,0.9671012478722414,
 0.648195075640008,0.35012380767824935,0.35044157772427664,0.6306747745283775,
 0.46229425547977865,0.0215536659415686,0.4511219735552944,0.2966134529731298,
 0.3106849035185203,0.26714186723774214,0.023138381559004273,
 0.7768761021959683,0.2532818369018437,0.3534474847833854,0.09781223531860905,
 0.7048937109998441,0.10682377510288543,0.8969212113573308,0.800837340090291,
 0.5811402258406051,0.12925978794670534,0.659820348029102,0.6573012794582841,
 0.586951628592647,0.28563134783782496,0.39920073347994944,0.5720495509024833,
 0.22598699120901689,0.2553465993033782,0.23857006939346947,
 0.6425319900787209,0.7849080358400945,0.31603589506414864,0.8617010300810006,
 0.4619355256810136,0.47742666255551447,0.5880829450574121,0.662537356930564,
 0.757135403912031,0.5739717320365145,0.062060294975584296,0.190286110551386,
 0.5391196652925945,0.41013565701463517,0.3187032684514509,0.8154219388333532,
 0.8377307221486894,0.1235085113244958,0.6186467548207607,0.13615362124704333,
 0.5189663535684482,0.8440326678834722,0.9593310954200484,0.578040685259359,
 0.4533350497801767,0.25272567066771345,0.7815016591112139,0.9621838670106867,
 0.9129109348876135,0.7857022055200976,0.9208056413957348,0.2963708596016539,
 0.8521334095919918,0.2833906632669998,0.17207193533074128,0.7786480148140608,
 0.7056181481190894,0.11576940537901426,0.2293583671288062,0.6863986241438673,
 0.8928410752047737,0.21933349967635862,0.3176222184428682,0.2554205235488879,
 0.5025054007089591,0.607014246239337,0.33601993539617936,0.5149322673577238,
 0.11615949653799507,0.3905729070246182,0.26694893419101495,0.631483401572078,
 0.08087100458733809,0.8410694689246938,0.8278547152744617,0.792404947373981,
 0.16981013824311653,0.09470689021954226,0.1893634981513097,
 0.8429989844048843,0.8948656698908962,0.384823718514527,0.2799639880411674,
 0.0709020184288105,0.7082300253387683,0.18120222064970004,0.2246460422500841,
 0.9840652030840791,0.4300187101061038,0.8168908750789052,0.7696751125085364,
 0.5387107189438576,0.2911731022960149,0.09515669872891397,
 0.42714195695803836,0.047050209657572895,0.8863855586289924,
 0.7610476525661336,0.8757985836134616,0.6967632790354814,0.4054194748823955,
 0.32305231974269966,0.9351260300785049,0.06667578392027462,
 0.9894669717067548,0.9232371563260368,0.8734821457484728,0.23826199754705202,
 0.9147040142427716,0.0010989995017210985,0.08134165794229375,
 0.3550770378854643,0.7363159399377434,0.45317569402901015,
 0.11855616970534344,0.3607290615025873,0.14140920037970983,
 0.9374078751699172,0.02186475547284561,0.4763697155677068,0.8758663361245731,
 0.4213367895398159,0.17123709695249267,0.2939092560580985,
 0.34281266339201855,0.5440007570957195,0.09836215539680204,
 0.8374408943294551,0.46115256615261313,0.9910338672970076,0.9294210392166765,
 0.18276843533765752,0.7583542469434703,0.5585675944585893,
 0.29114027881314464,0.2377837824062603,0.6576507326628597,0.7869920979525096,
 0.4257733983579569,0.22007275186209108,0.7487418717706461,0.9126094579573929,
 0.9811139650096208,0.6294917882150816,0.3785364536901482,0.3639444129777697,
 0.01776881477922454,0.6704286831214892,0.4630837276613068,0.8640341154408847,
 0.0339294268516247,0.7531994494945252,0.2447948040364677,
 0.0036241705881935316,0.09803558063242723,0.01872857019359897,
 0.659454170492041,0.9065032614573431,0.26462120589764004,0.350802083072718,
 0.26816029066205904,0.7944041463211692,0.052416994332191824,
 0.29943218071991706,0.5481093142968918,0.8411009817283196,
 0.19056957228656513,0.24694923206641994,0.051750469842513525,
 0.2728585273352625,0.7364097561831121,0.11469854860051187,0.9664644141553816,
 0.2466026973323049,0.24583757398790507,0.9227364381903437,0.9513960654643708,
 0.5793466837145198,0.5347314938409998,0.929830248104506,0.5020025119409518,
 0.07011264629633507,0.12586966957897272,0.40777759501852606,
 0.6296565487408143,0.17520818388651804,0.33489168509468314,
 0.4274951934433521,0.4350856219269137,0.08648908664385113,0.5266780132807524,
 0.7846581094299914,0.3108464071935839,0.12844262772737491,0.7275000556495265,
 0.4737066737767579,0.8268811722390933,0.6693624360488968,0.8083698224995937,
 0.8704534168804203,0.3520182255632226,0.36059122711557867,
 0.11585688328083354,0.32425450277041035,0.7776627237972762,0.664498916540101,
 0.817188218483745,0.45972930096177245,0.251345384270662,0.865488975905681,
 0.2476553128939425,0.22353741224053547,0.6069547565184024,
 0.027785615236830985,0.04949342153361935,0.5093478094858047,
 0.049471203880877446,0.8442399916206833,0.18733857929782782,
 0.04519495023320086,0.7779280188628084,0.8124397895224615,0.1284273356430119,
 0.8840668841171481,0.8000704606321293,0.0885301324720491,0.5356028736366438,
 0.4458194904550301,0.9165070110703494,0.02455676731744527,0.0898841961573924,
 0.6371892774056495,0.4065318132247031,0.5899816515590479,0.3060132569754147,
 0.06010092278642021,0.4666153653291023,0.18948594973198374,
 0.5565060602992086,0.40430720869089454,0.123920089210668,0.4654552067329778,
 0.24489436150001087,0.7796137180924645,0.17794873230696473,
 0.4246562133445352,0.4438330826674157,0.922966065392954,0.4803202656479991,
 0.7265431157373944,0.10974811134873541,0.750527277236566,0.9951576002762368,
 0.9863105804672074,0.9151794720526587,0.298021276271054,0.9276764936457261,
 0.021898360252036397,0.7309797604252054,0.40744276360375664,
 0.5455210748430471,0.3639247441541774,0.6805709154713643,0.9020969726651548,
 0.6733295054274894,0.7330622495751182,0.3724980517922143,0.352434417123728,
 0.18071249172017,0.9558347218138891,0.0492288554133572,0.3662208694873029,
 0.1935175119264697,0.09077172490168828,0.9119235762136633,0.899628127986452,
 0.993547009155791,0.9938227225855465,0.9181818157166602,0.08475909188386643,
 0.5530485746492438,0.3200240823590681,0.005558993149502256,
 0.3621714488460427,0.7502096051796142,0.5781926023373366,0.4461964588090134,
 0.447028298813665,0.7223946286724504,0.9068395202775228,0.19859121247777833,
 0.21614711815286336,0.8351898198174033,0.08914210118979293,
 0.5638369070498557,0.842047133375645,0.16127566701031681,0.5213030872290287,
 0.92611361589732,0.5175938370417511,0.12820077403097818,0.27233495704922006,
 0.26817607092133144,0.474474256715687,0.5288600745957659,0.46127207241287016,
 0.2673951242917749,0.7970682993164412,0.8167105822006419,0.6501227291892555,
 0.7893961769679708,0.29013903290342646,0.8205745199217561,0.6306824268579904,
 0.7205526684455771,0.6004239653808375,0.2934347578173413,0.33786662122488476,
 0.029413153810349146,0.7262121912713166,0.5903897798494312,
 0.3391268358947006,0.18830126879325815,0.5262218997545725,
 0.10173261847291948,0.04998840653600656,0.38601579905397265,
 0.8497532191836423,0.8209479838092398,0.26896551244231814,0.7421556320143118,
 0.8861669926609516,0.7618432561429697,0.8807035739340608,0.7122647158390424,
 0.19379740216170693,0.5691166728265066,0.35956957389569866,
 0.7566850525508246,0.42643355902011804,0.5781200077841067,0.9503538195068861,
 0.5950047938350878,0.8185091234485606,0.19836982394966696,0.2342853725180074,
 0.716658343383469,0.5770355306367608,0.1244895216644859,0.8823534360461612,
 0.37344685208305495,0.011341269278031585,0.7825139395745266,
 0.658018822770048,0.5932007743062172,0.3174984032204462,0.03544199013826965,
 0.03000076833508636,0.056688402637481206,0.17004050106616797,
 0.08976905316795336,0.17793958872413784,0.23081974582035947,
 0.29638386028680885,0.5273278610917552,0.9737056211383939,0.8756432201804996,
 0.04744316188846198,0.09234973021721204,0.3508497873357972,
 0.21828853540449578,0.2457034599936505,0.015931989483112474,
 0.11987827556440256,0.9493652901807623,0.9854109735396049,
 0.07191017402186439,0.6634642388848451,0.5545213695943426,0.411581016109419,
 0.840256722597985,0.7681865448502226,0.9177251009157358,0.7075909885565311,
 0.12442196752626933,0.18993565141477353,0.911712012463315,
 0.08838582539201612,0.9802713224675748,0.16096234815555,0.6951413592225191,
 0.9042816059104183,0.5964016845016622,0.1645163512380018,0.4098673192416308,
 0.7822457671092766,0.7434621544247523,0.953672751934067,0.352045927837345,
 0.47934764697789944,0.9101124528402471,0.2322676017233527,0.4947116044708655,
 0.48476191140974445,0.43025415940051026,0.5105450288986625,
 0.015818927576159547,0.8106318133356664,0.7052055772953834,
 0.19675439669678918,0.8277668143897184,0.9159481366672558,0.4843036977055012,
 0.6230493598653352,0.573309893753483,0.09189651568135349,0.3770614112364905,
 0.13040648579470548,0.8926815415903582,0.6686695361706303,0.6271708405211182,
 0.2596532091727638,0.5660452307049482,0.22188406527791693,0.505611151140314,
 0.4962084302993016,0.16237757945198694,0.722140361420909,0.7224216179624641,
 0.3064412771555598,0.5020119501631962,0.6059757711351494,0.5489878271802595,
 0.1682238857268772,0.8390554309842754,0.12010800187804194,
 0.14160160935961497,0.40943417020884953,0.9659206152897807,
 0.9740508920294958,0.9368317238482748,0.5376634123194902,0.8091017569510762,
 0.5206972090317574,0.7445906311867708,0.6706486730620379,0.7731456985165235,
 0.4735678060741494,0.3846020186011795,0.5910193914176591,0.879372728627065,
 0.6701044072835658,0.6761801316803463,0.597811163546063,0.3186465401406857,
 0.6847630777126001,0.34885193835516615,0.5715879902498434,0.5785552167547712,
 0.9847735998257463,0.9640071523157442,0.5296713149566011,
 0.044225567539348365,0.8199004569208601,0.8817794625929035,
 0.9304355778492923,0.9264903382684276,0.6254838920212075,0.4650286258372882,
 0.5915068321730877,0.7468549121316788,0.5379456983965489,0.6291370218907013,
 0.5359862232989322,0.9211502653096134,0.33621347416476,0.3244789768999434,
 0.573743478351216,0.09898859147297845,0.37843907452764014,0.7451107926575642,
 0.8115529736479503,0.9884745523864532,0.8840572466745968,0.6198856091497624,
 0.4574857245548385,0.5948610833334834,0.593570645235587,0.6158202778535169,
 0.598757038620401,0.5208457724409267,0.25775004579952665,0.45213474961093225,
 0.645085760502481,0.7192651498266185,0.24022953584189277,0.8477148253486253,
 0.41587733463541565,0.4003407081753275,0.7668940354365132,0.268729432104735,
 0.46638500894523227,0.5318604956136816,0.5479608791837904,0.6880461463328124,
 0.09772326075784976,0.09803770179044957,0.2603794527982264,0.703404738989057,
 0.9636662044083226,0.8932619254521573,0.5501771971279397,0.6489089973047582,
 0.217188738465244,0.4044582400778933,0.7398325493588672,0.1178593279912824,
 0.07503514930666932,0.8765101319046659,0.4023723049225424,0.1418363683932945,
 0.16344449902437164,0.1262279307221802,0.3036179885036535,0.9585206091882132,
 0.7423995910119482,0.8742356002883577,0.5717922893822357,0.7553831530886272,
 0.1533170349915257,0.8739787234712235,0.17381124181665208,0.8341646259875505,
 0.1483996575201334,0.0012916438080088355,0.2187841293360896,
 0.6524433150695539,0.07576362092790512,0.049997916598122916,
 0.6592866983659303,0.7725203921418669,0.5172011553606856,0.8642151964822692,
 0.8777687794377516,0.8820066177263343,0.06563830279309901,
 0.46830425126189246,0.20005505108849653,0.024106830234175036,
 0.929349191393298,0.09764272508998562,0.8834559009624529,0.0627089719117121,
 0.7957123484462788,0.9739852393706758,0.9225648370284144,0.4603117075846228,
 0.9375705173632563,0.7384085629471455,0.513965689576204,0.11090093602496687,
 0.9030460467268867,0.9763722055326564,0.5045453965835058,0.22576396576698043,
 0.15824391357411505,0.8388390106005154,0.21093514367734245,
 0.9936092469580443,0.3390617099864772,0.018792968339439553,
 0.30829816157489676,0.21471025527026155,0.42127450176762204,
 0.5149239153389437,0.09902541698943002,0.8716661265011227,0.5659796596635581,
 0.284436963393484,0.08942786251538681,0.535545879650011,0.6284854504686038,
 0.1017332954422937,0.7401725691898241,0.5050561855319194,0.6313264467445105,
 0.26512254452079675,0.15285322103208943,0.1960564354823504,
 0.3163934790994051,0.840432589399472,0.38296593129841805,0.9914914725819419,
 0.6784075011480206,0.7242334114130179,0.08846431159433099,0.579846980100397,
 0.9902317607924793,0.3140821312978095,0.5246717570024918,0.7367360631192155,
 0.38434689007118916,0.935739203676473,0.03914106492781577,0.4641809068690662,
 0.7065117439464994,0.46575393164793855,0.766485941787598,0.40718908343380844,
 0.8546754957607628,0.19381102432213737,0.16890471094447423,
 0.2063741529388553,0.1915238137774934,0.7498807879977027,0.3185734575066056,
 0.2801589278126878,0.9699560231987766,0.2416827529039406,0.6709217260066799,
 0.8965138139560344,0.39370470244313005,0.2919684682155054,
 0.31816275064240995,0.07529105069117015,0.7298381838155398,
 0.39503229071143264,0.9184134447418923,0.5055945792558636,0.8157617494157031,
 0.5809094122157287,0.014076212987960579,0.27717657511109184,
 0.8363671198600467,0.22740327417032813,0.8679165481036706,
 0.13116092745903107,0.7130298109763982,0.4603365570206466,0.9489345476126063,
 0.8206792128000542,0.5518050390090307,0.8454907829408408,0.03843467763410768,
 0.5491341931615084,0.2510850173636062,0.8643916034740091,0.9213350118205623,
 0.06792960645794777,0.8152150358938511,0.8635372571042733,
 0.03496203311785662,0.054097864441583265,0.8620573685581725,
 0.2499821024092551,0.8675725998661601,0.8471361922950558,0.8622305749552078,
 0.9982015165137492,0.8845941108600329,0.38147584063546414,0.4765646129977381,
 0.7049740442532002,0.29943648054422234,0.550350988802206,0.6839471008040681,
 0.5535174379018287,0.09360159225482534,0.5879552437751621,0.7797970099605271,
 0.9847494214814957,0.7917835274670468,0.27473264732247626,
 0.18061181153658978,0.4533276504964372,0.7936411498171801,
 0.20350558196659652,0.06251803695246783,0.7248277439592649,
 0.9932983662210133,0.7296852348740049,0.001203908103137974,
 0.6035277441159245,0.4450330913839482,0.7153125497839139,0.6706477880565935,
 0.9331205950486674,0.5861254647825823,0.037773074487559066,
 0.7508585709000176,0.3720461593235267,0.14298280004300257,0.548873300389205,
 0.9424891620026942,0.7355992026008993,0.20866396333544412,0.9135546979491735,
 0.1152310701314948,0.6063625611909765,0.6303560567164086,0.6940938794909249,
 0.18839391920161774,0.5610753293281917,0.6084143193596957,0.5607181289974333,
 0.056052996099798746,0.6512930636457137,0.90423340038217,0.8325071437373922,
 0.29728131481062126,0.7058981756133575,0.28935516536084394,
 0.7851450501714055,0.8571572732929101,0.2859851548503216,0.36892248997605204,
 0.8559953020620895,0.4531501012539043,0.7934604604778455,0.11609177897368927,
 0.15522173938179074,0.07965506118087307,0.7627971844231837,
 0.6793680484776106,0.6595335311002696,0.7788450117183011,0.8637910232501828,
 0.5595712802201951,0.32358296168510603,0.5748589735089223,0.1498979805986389,
 0.5167614980446064,0.030988511422870957,0.7476346200290198,
 0.9200449225762577,0.12845700448934383,0.9341770305482666,0.6976621910365033,
 0.24202100517432112,0.5212508823708967,0.693046602160162,0.08677180312185351,
 0.17873982048706905,0.8120008634680821,0.764414260121396,0.8802507460426227,
 0.18276460824655738,0.8311285973162238,0.6444940620293464,0.4645969286197673,
 0.31720717115967245,0.7780425420070557,0.576874125113355,0.5650267979990848,
 0.4237385505410104,0.2941378060370917,0.3435895831498388,0.9701186846938468,
 0.18053493794228492,0.3898976201465628,0.38447021975235285,0.342137003734338,
 0.545539012429133,0.8969381377302791,0.639925061182315,0.5215490425940713,
 0.8031346206611821,0.5081395138850621,0.46820513337100744,0.7065887362403976,
 0.4594875749609255,0.835358175734682,0.4249167732792172,0.6800408664182286,
 0.7206081688220769,0.8830463390908316,0.7056834612160584,0.16788541619741715,
 0.007999821498650417,0.37202164858797804,0.5145499042695592,
 0.6625011511396572,0.7897406394976507,0.4692443026022286,0.13963863822671896,
 0.3298521644708141,0.4639023520188418,0.6672082876563862,0.983000631258476,
 0.32418733864586236,0.13128024966685892,0.48068301956494386,
 0.902936629872344,0.059316194460472205,0.07192073626160866,
 0.49148363843757803,0.31210813283432326,0.18857260278838806,
 0.23577918353329308,0.11500332295998761,0.7326982721345731,
 0.11321450707645231,0.9776089570928621,0.5292723190604562,0.7463599963457563,
 0.15055974226588154,0.29916826292314214,0.8238651732868767,
 0.9458379227941113,0.522522859727177,0.9680921238322008,0.9850588319112773,
 0.3729436630648062,0.02346195667298212,0.775697966965757,0.7364782365127409,
 0.7484762960934013,0.3726933219789109,0.7178993918400952,0.5718276874189572,
 0.6052964660893966,0.42696448022815625,0.5183187221673211,0.8571764429182724,
 0.8604130481691257,0.8357416357739452,0.21114014122460412,0.581368029393836,
 0.9496596289507893,0.8223422154481239,0.4740634281606456,0.2897610986306849,
 0.16162299390897172,0.9115744841156459,0.7679273459355365,0.2979234814184306,
 0.22794686241855278,0.510672214343028,0.13022268217471666,
 0.43472765055586693,0.2376825294423034,0.8637853302717597,0.8597358768134965,
 0.12597725460891507,0.7029982530922325,0.13615945188913325,
 0.027240941150050824,0.5036517851609688,0.06766765375513761,
 0.395148785220202,0.3527996583615258,0.6533001493985896,0.10159905205804987,
 0.3872414175926614,0.09079002565507344,0.4612319084227895,
 0.27733128731759016,0.8776452784993475,0.7925923263346991,0.8661222810947836,
 0.2027551416504989,0.017241659505058293,0.09006613761808435,
 0.5552193865769949,0.35633295306380186,0.09766142004359668,
 0.6850327492754057,0.8970601793943704,0.7066538298988337,0.6388237449959084,
 0.09321233446317634,0.49042069560499735,0.6844963704278629,0.658056878316517,
 0.7215377788531984,0.5873637720505865,0.9143817950787868,0.7803995279468295,
 0.19942393633400246,0.46957821320776194,0.9231524028922419,
 0.9626753874048352,0.47560989106797913,0.3428277590513453,0.5616176092661082,
 0.44466958534881085,0.11695631758165093,0.4424009215123892,0.790652307743687,
 0.07708813460423336,0.17172311173292154,0.37336534391919163,
 0.38401712463365767,0.7699068247111135,0.6017956764130409,0.7012043822300237,
 0.06595513650995577,0.476522401714625,0.6625475518850775,0.12338207108741372,
 0.29373923388875156,0.07151795345976708,0.04135958840871512,
 0.18789636793165132,0.42403797715444524,0.17071991585702273,
 0.5868518571525383,0.822918968245868,0.5997650140213224,0.36902497179112703,
 0.04939460657606143,0.8877011242367124,0.20587787717557382,
 0.9186185243669955,0.9901183262305322,0.021450852504224294,
 0.11030114334162833,0.4545555301929539,0.5749020083857117,
 0.18237468229048148,0.11772195070082858,0.3509930547380662,
 0.9069418705884038,0.6974043683776412,0.6862177496165855,0.2343784840170905,
 0.32193641576750687,0.5653169921601359,0.6670656818459235,0.5448856731617987,
 0.27297486321074804,0.4821117062594027,0.9975262211295337,0.792312213468634,
 0.021182725795776514,0.8494190021558553,0.7988815029985374,
 0.10278568358743656,0.85879340170525,0.8322333442140795,0.3871383857413646,
 0.2532562089570799,0.31550212170964964,0.554855976004695,0.261454354547774,
 0.7430253860883377,0.2514709640157551,0.4568366013479257,
 0.009076156161706206,0.7678460175814256,0.37095911058408404,
 0.9859368451724994,0.11289385353607384,0.47488774362222896,
 0.39861334560964046,0.28371220775113426,0.9911920418917587,
 0.8070201724240509,0.5859695357351867,0.6589444200082911,0.8293769830894626,
 0.59894285721653,0.5198727486405167,0.6025916547667189,0.5704775625838376,
 0.906195800571513,0.8518639817873465,0.9993200798473018,0.9812270603292673,
 0.559197194347557,0.16163136459585525,0.7504567106399144,0.8682609985789947,
 0.71225651941595,0.6637633977388805,0.45725450731065165,0.2838418153882255,
 0.11164900231067276,0.07393219373537696,0.10892425073301581,
 0.4745399708298641,0.8528417872997339,0.6926306039699058,0.24138149303757195,
 0.3166487785901291,0.2069922170056654,0.6449664477971302,0.8772458598472838,
 0.7785335840959661,0.43953924104333675,0.5787234222470041,
 0.31977899831915635,0.6800868162007574,0.023193413080612735,
 0.021994050089255257,0.492913968884932,0.6097756281729517,0.1283700810425502,
 0.6402424148446656,0.0638968680256995,0.1871280310225576,0.23843375701677383,
 0.23132111664795219,0.7133853669601393,0.24082691855793947,
 0.7698484069980451,0.9757975592930516,0.13044285918570908,0.4226902505020953,
 0.5756818919718893,0.4466568619857655,0.058085642649575275,
 0.9594719101101481,0.8713549431621183,0.7080210603473952,0.06923184580584385,
 0.7696193978041208,0.12775726558814016,0.4922983977356924,
 0.25092929900631544,0.9424688264637299,0.0994959603754878,0.6108791665286083,
 0.8789039226357366,0.5975817996772782,0.8921060908802221,0.9531578669681168,
 0.47053842467147255,0.8095369069404612,0.6904605775837472,0.8220613897843897,
 0.17478882214651514,0.9129583771314743,0.22654462214641113,
 0.16398902744010124,0.9133709586997225,0.5837115333907031,0.8368063219253642,
 0.058856526569757595,0.3943043894196956,0.08621099835071422,
 0.714266503988821,0.8040618413321063,0.197744957966017,0.8904271018374468,
 0.4167381432874996,0.5520696788986142,0.25377322038219063,0.6327530036785619,
 0.7306868061142939,0.45815789886541813,0.7403210369583878,0.7897183393872476,
 0.1255119462476988,0.1019477233914653,0.6093639821861663,0.8974105969276028,
 0.9369998158001084,0.8923529327473054,0.3430239274212934,0.9118983784599625,
 0.47223454970811796,0.5884493927034542,0.21870041309779342,
 0.20220790583050852,0.3408683280867837,0.9937574063601164,
 0.14324640819146695,0.5276118652178372,0.819849961840564,0.8059060462805818,
 0.2975146126348843,0.007462855674271296,0.32196800844449336,
 0.6260966051248563,0.6565991769196183,0.4799520298083628,0.8967003239148683,
 0.14701593320635054,0.34031351436868884,0.7580394977697852,
 0.3577048989309971,0.48295894128892825,0.962727813015994,0.5770877720616855,
 0.5655056140011482,0.8941240733114437,0.12630984367191223,0.7080856149759381,
 0.9924611209534333,0.5044839299699915,0.0014371327561255587,
 0.6267751836899759,0.5485214806740886,0.5423345609154434,0.7374850983049058,
 0.17170743305974057,0.21614234885657257,0.19118225628028152,
 0.2136744574575168,0.18454369224327094,0.975795657498963,0.6413845762998274,
 0.33925665091665647,0.0016458335446430938,0.7865333471277638,
 0.43867882840914296,0.6880833114174855,0.09921801475837655,
 0.9497986346492221,0.7125366034609126,0.8736164170553796,0.17147680941080834,
 0.6513189091602868,0.6704182138162493,0.7121321173561889,0.9546217653589769,
 0.47514258924955555,0.35660305857424035,0.825477725383739,0.7276082737248657,
 0.47787229833254785,0.020683469480358374,0.9638492210942241,
 0.8996229690138756,0.12053816962209929,0.67448389973634,0.8871661677896201,
 0.33299889182983644,0.17700859185817697,0.2366190052299475,
 0.7180222417581287,0.569739392920438,0.8178364449564746,0.8833731706982543,
 0.9214834542571493,0.7545249673491383,0.07971155236406546,0.9632987332549121,
 0.9629006829261587,0.7888266243302081,0.5891598765137765,0.7736574189948819,
 0.35956660946533836,0.9929047506899388,0.08586346306226789,
 0.5240234786506075,0.31445603882517015,0.42368436094497186,
 0.9292233316246419,0.6259594975193742,0.9664755727421278,0.289506539942486,
 0.20759531204246773,0.8302932287424402,0.19798619457497846,
 0.5914310192809188,0.521668175575143,0.42224093728417356,0.4968782486008698,
 0.4716114350775621,0.25073636992135295,0.7268813651708905,0.5825158068864618,
 0.1341244534887539,0.6804554537619987,0.922093212120514,0.7681137109631175,
 0.014040711546240203,0.8986698029328218,0.25472292687247067,0.16017071542216,
 0.13980204790564355,0.9038243914901938,0.05287997621842366,
 0.34380995731441444,0.5675708152122455,0.607954919665018,0.5993696610577358,
 0.923874810606177,0.0458019557066931,0.05069600453445766,0.45942846951455085,
 0.47630226968133604,0.13532887337440236,0.5424021251620577,
 0.44252498118845696,0.341530766365049,0.7146096098008174,0.9130543740913198,
 0.4579715264660501,0.8596645767886097,0.23002015309726653,0.7863257478801802,
 0.22634619419404534,0.727458953813453,0.49206809190629197,
 0.26066661739186525,0.09618815631284927,0.3422370251614838,
 0.5886011182806623,0.33038245390542054,0.988732140732328,0.40985914377495425,
 0.9561289920878505,0.10138275672620667,0.831497931036578,0.45040778331983644,
 0.3612105535941603,0.7994373193404706,0.39474795283370345,
 0.007099214932306364,0.4851054905764569,0.7079947811822965,
 0.9095141140203534,0.5060547038172989,0.7576089906934874,0.6759655593945312,
 0.4194704113209837,0.5850675267987346,0.14582452661178902,0.1824873243228735,
 0.15245198275934269,0.23325303377374673,0.25657520283099355,
 0.8954397914231986,0.11379542404325449,0.13191540553389203,
 0.10454795065296929,0.8279591067197092,0.6123011491715102,
 0.18886811760448008,0.932834530518988,0.12693560104759294,
 0.17005393169057137,0.9881036320034582,0.8024401498514621,0.783019823892176,
 0.3963934265478197,0.7093394869488481,0.29682086451462364,
 0.45251857524146766,0.7134972678556015,0.33319733584707034,
 0.4650505180704195,0.6078327458073081,0.7631813477034376,0.39425027305898863,
 0.6531884670488648,0.0774049960178419,0.4933654785306213,0.18759543023437453,
 0.4768972633136743,0.5239407903205308,0.4953701927740135,0.38464863958661377,
 0.6368291688264776,0.06376639220778724,0.5011181300060717,0.9593983740714962,
 0.9235069517545365,0.4823392284510526,0.692140815762611,0.4163548479252436,
 0.6971840208800691,0.05207070999410013,0.8115869720241059,0.879620348473794,
 0.5732949538236716,0.3659341412422312,0.255953844064706,0.8663714991552851,
 0.7597008386880898,0.6576393282309603,0.6071391282744921,0.20977906060539553,
 0.6759588419176393,0.30636692586848,0.6192360794129864,0.8093540443052778,
 0.7813749054386465,0.5359216670459692,0.935018411868798,0.04897960107202404,
 0.7670303028426307,0.4392904082293838,0.013349654873870409,
 0.8662830744652883,0.7731723549948085,0.7866814145620528,0.2816921546471697,
 0.07170346608990541,0.5787247848148018,0.8942823840669223,0.4851220327195501,
 0.46045752221671754,0.3706538411988606,0.6696136425232703,0.7379637229571498,
 0.40309990351363867,0.3198123573929481,0.3332841559150451,
 0.29644321267877694,0.4965648853284551,0.29667134669564965,
 0.30773490645885737,0.12903930756576443,0.017811679434437955,
 0.5631884857220035,0.3552511015089408,0.4765463119937725,0.6006204150798448,
 0.017369777011586596,0.4095121410593511,0.05611101897887161,
 0.8029259914839582,0.9877376176549372,0.017106841917182702,
 0.8240524818107923,0.33184453288446036,0.7227884712555646,0.979305492397325,
 0.6637785827075185,0.7048725695385593,0.2799444889381283,0.23647189473177832,
 0.5583437876716105,0.9052382776588634,0.8346658394012211,0.6325370313868187,
 0.9063733730608621,0.9465126015410487,0.9354157087343375,0.8097894734975202,
 0.4030899896347896,0.16600414509264916,0.1757370738182138,0.9065454320696301,
 0.07172665366456765,0.7009760461264292,0.2639245720546106,0.7608171002966685,
 0.42034054817375777,0.06119166215588023,0.8119041014066942,
 0.3481751197573817,0.6844374196003173,0.7030998375956086,0.9179459706592589,
 0.47086834120778553,0.6979625336757572,0.6235623090884964,
 0.08789184660740412,0.5331099725625692,0.4151214170699824,
 0.31950563560085354,0.8241354963030321,0.11966023958493377,
 0.03909928495150283,0.6235317950209577,0.7558653034707415,0.3224812417308006,
 0.07385712554224255,0.3622110404969705,0.7589062841082794,0.7433296130402086,
 0.8599766424849308,0.6990097626841214,0.682785884472122,0.47966099995888234,
 0.6974845555825907,0.9146454838132452,0.4241354532710908,0.6211031304332687,
 0.2742481951333644,0.11049930371625605,0.1370416724151815,0.859871822452777,
 0.8604396953940086,0.2101979816942282,0.8626005269857006,0.557217153165395,
 0.6783342455839997,0.7159581220076746,0.10135857761401257,0.3580013056914332,
 0.9023630285007915,0.24675260431094648,0.03055876011109837,
 0.29649359181326496,0.7010963900421983,0.576660855146897,
 1.8655305795256538e-4,0.7839455810833771,0.12195145513324213,
 0.9912628437315645,0.7004598961706553,0.35830381382714627,0.6499826018196737,
 0.3246574371066462,0.714206681000206,0.24203835410300434,0.4750410738477451,
 0.9449040687524021,0.8909672497525248,0.3969565886522167,0.07330259921239557,
 0.6190752393890147,0.08348875507844511,0.9754880704631269,0.2053851622088888,
 0.9575901597287241,0.1243897444343719,0.3032842408056622,0.2616227431860233,
 0.6653146880634957,0.2725698333215316,0.19179315320282453,0.4070796886452901,
 0.3572941306064764,0.6123679041334592,0.004630818999475483,
 0.6738719326992421,0.54507095765766,0.6558781628146747,0.08392929557508311,
 0.6824285865959039,0.4649178685116604,0.5021372594116699,0.7929443850631924,
 0.9235527749515697,0.5475246041845047,0.4355744826594137,0.7359324416368481,
 0.03536387213252967,0.589370505805683,0.6656398887212385,0.986950283571648,
 0.68714555660137,0.21656318555721743,0.9978423090183548,0.4008240768961391,
 0.941573028789437,0.8820269866701825,0.10180706634399495,0.6629469448573982,
 0.934473562454756,0.3868381067708515,0.460805121379649,0.028233889972441695,
 0.6268933266775509,0.5548571545711343,0.1647158327307967,0.3925230544756576,
 0.1409468848684312,0.6371592360924957,0.4912541396199642,0.9681301230261621,
 0.7500252684475761,0.2292480842944642,0.923471000844698,0.48823067416771015,
 0.9556506007323179,0.23386391038286025,0.0275056392878944,0.561600249340307,
 0.9676363646497941,0.6160706049740704,0.369727070351889,0.7059788959102152,
 0.10952029691342713,0.28447748911269344,0.5008308103954896,
 0.8606035996362784,0.2399817265065476,0.4213390558371062,0.5962205188900292,
 0.8548674074885944,0.5179937435294778,0.9438260523216855,0.20609071651033828,
 0.9968312586714178,0.018860477950112364,0.2179510194589328,
 0.3315127527366407,0.6466131073473147,0.0717146014338017,0.49547141724624033,
 0.4882120558091363,0.9056063365307332,0.8745846097878901,0.8426197739245678,
 0.737391083403041,0.7392909049837099,0.9367266706890904,0.3662945263309182,
 0.16394961090617033,0.12232996498726201,0.9046099972334385,
 0.9486526540144152,0.10676753334238309,0.21590817623282532,
 0.09926157528014024,0.323332075634315,0.1671276718883048,
 0.059352756152270536,0.8692663446346264,0.47428241584184017,
 0.3547037049654015,0.8554439934251538,0.4008995567662941,0.12205285118669262,
 0.9196187034814145,0.9007253975762264,0.11728901461558017,0.958161026760126,
 0.6971331058010244,0.30816595688308157,0.4545765646268589,0.7970976322404288,
 0.38312835706180637,0.9916616423157538,0.849288715116076,0.6769227263667514,
 0.7441637022373061,0.8635187617880308,0.3993097642869735,0.6284552956785651,
 0.6583085479904915,0.6986066832601098,0.4322932957869453,0.6103254238382851,
 0.6588813552278628,0.3657195883445513,0.23965099726606598,
 0.10925497640091619,0.8425289822372347,0.4395703682249994,0.7953625199582406,
 0.4079483771850114,0.9183815183563928,0.8112146819613594,0.7759301011892646,
 0.6922490280844765,0.8575246324930954,0.5467207137166665,0.4221029642466041,
 0.47717817795751416,0.289770350889645,0.7204568680391674,0.3012483244653188,
 0.34407539972255097,0.8489697697259025,0.33684712229590763,
 0.17344425642702332,0.11237089878336914,0.7510817426387286,
 0.6739632758023949,0.24372413086954225,0.7660635017121081,0.8471107789918931,
 0.9516928333360273,0.7911318214955179,0.48636981558444903,0.9414524198500661,
 0.9021264913018644,0.6200263565388761,0.0865878777249991,0.43848108979043543,
 0.10162552204895892,0.4354587712752027,0.502347625957476,0.9638361336607011,
 0.597544230589446,0.5324274262598296,0.046419935621720976,0.8435757160590849,
 0.5942747020845676,0.3590388937944484,0.005505495610819144,
 0.3928570632153301,0.31816715741510027,0.9005022084393283,0.5626373109605972,
 0.7644585212913216,0.5587254157148672,0.3709088058396456,0.32741397228101976,
 6.584333612708049e-4,0.33631929895612733,0.6161179734111772,
 0.16127954456631033,0.4981192950585197,0.6825849896542181,
 0.48873640574954647,0.7465678999353572,0.6486023451350131,0.8708809381032486,
 0.13172072568614057,0.3737516111665621,0.8571612355783998,0.4342549870263698,
 0.7004551478504945,0.9908697896463357,0.3030347338427525,0.08345242400418895,
 0.5495871864359773,0.47939715679469685,0.5704885757697508,0.8951234968303448,
 0.7197852732201246,0.6048083031435462,0.24677689912706902,0.9734177469877646,
 0.8346754071776216,0.7326249038220317,0.7540815351918307,0.7439858821813304,
 0.37531116065825776,0.4955337612498145,0.045974686955150945,
 0.8160818772275981,0.0617067006555444,0.5837167064421642,0.01652553110175159,
 0.25070212410513204,0.6227357764660391,0.20865832322095845,
 0.9611963090201566,0.20293082828450715,0.2917164176267222,
 0.04865461439920871,0.1789902058980698,0.2680155844600174,0.7082699490116611,
 0.8470591645721975,0.4750256521929792,0.2968127702923685,0.2738729215059408,
 0.9064958756872795,0.15328043905479283,0.4839539865630652,0.9461887617885122,
 0.5976396618460926,0.5832620263395534,0.2430828730501715,0.9524714982360395,
 0.24451668045606456,0.2512228293999945,0.0938594144481183,0.9353586601366217,
 0.0705522505661208,0.11390112343972558,0.5594970292527895,0.9340893693739264,
 0.3889369690688589,0.9965407429029995,0.8729018617135115,0.771955827412721,
 0.9959101533806249,0.17755955352065755,0.7960947588825028,0.5562958308311962,
 0.19766391311247988,0.8481636903804751,0.816889115747851,0.8013474845953105,
 0.37876690414705483,0.012052446932622862,0.6863219462877612,
 0.06364867118596929,0.5966116631046452,0.2603652294772356,0.7173589024151364,
 0.8277546646010137,0.7680326478842305,0.9774261445347917,0.407849529555232,
 0.81089594987964,0.906973959554102,0.15130586529052825,0.2981949316052218,
 0.974700035345023,0.07318383551836571,0.3335184998097167,0.10875434778943804,
 0.31443044126745634,0.5981094171809497,0.616356167998695,0.3473053264206928,
 0.4552805208580215,0.43231231109078316,0.47934494089097235,
 0.6588341395868643,0.7088009131207128,0.8880697887604797,0.6746460373386808,
 0.864739383598225,0.7435109268458583,0.1216353913147058,0.6645928721734589,
 0.48697365660631553,0.10407926571604631,0.33157870835760506,
 0.5612767197670202,0.491079581644427,0.9018804174195773,0.5272245105111376,
 0.2980650354355018,0.9131638351535526,0.8627963593592815,0.3324761494805194,
 0.9135890504894753,0.707397588798874,0.797776024172204,0.2846329303435945,
 0.7053533767628697,0.7563935435052548,0.40474526976048286,
 0.14298315041638887,0.8813443708515188,0.3169293762529084,0.7135741531779132,
 0.6544789270107731,0.2982556353369369,0.667084605035396,0.10380420263511658,
 0.11216733717337657,0.1345685019439844,0.08379578591129166,
 0.05972586151556958,0.49039310822461446,0.6413411291636297,0.688387427016638,
 0.9060514623552108,0.4748253305899992,0.20044897931274264,0.4405586316719651,
 0.2277589153074009,0.4370604271041041,0.9262359528770368,
 0.041339230018728634,0.6516021473561974,0.8896048311128838,
 0.1948989780190281,0.9286230773992189,0.49802391855239514,0.3262545466966046,
 0.3668591241085698,0.02700812978006828,0.17395325751473933,
 0.23773883210955504,0.7593603546307749,0.9376657740673308,0.3360777319234324,
 0.9382743189735401,0.40728982915484124,0.5455560784262059,0.4056402335840643,
 0.28646675959688916,0.23852386383372315,0.6658327251389998,
 0.8149649508785306,0.24781090621419077,0.893388409576881,0.9144235533485937,
 0.5816592070216373,0.9319789506916768,0.6828883729112002,
 0.0039858493822466645,0.39396568404349697,0.12383592545486932,
 0.14606176568152718,0.7588314403020728,0.8321732254597914,0.6775771172825242,
 0.24204843972697931,0.922091351438742,0.328404427140216,0.08337361566338497,
 0.48351968324009675,0.3701936559349297,0.4484580471390305,0.1786988359187056,
 0.27243957990502665,0.8798017481932041,0.6760850565509323,0.6052926266498008,
 0.6144061358632422,0.7712396387234529,0.15597913691123977,0.7154701627328106,
 0.2254554216950977,0.1569342049708281,0.74371756777661,0.3519166897781807,
 0.6544585101775018,0.1821864421723456,0.4065137075793267,0.6978000929354738,
 0.37335218894018307,0.6924783495605672,0.05897389174984047,
 0.6980228281564843,0.09861250649765207,0.9331404266318599,0.7037618621340024,
 0.25555102887937375,0.6947748617886098,0.35645030206581696,
 0.4743421737221931,0.4077200210611037,0.9276533481578875,0.6895938846476493,
 0.042444282511795106,0.8923885449893231,0.5347211510679853,
 0.9522574692802376,0.8308940518161365,0.05734145609958008,0.6275381748349894,
 0.3691165572857861,0.3483652819566563,0.5252627192818852,0.06824510925337823,
 0.23913421016672554,0.40221492564000283,0.3922755821770101,
 0.2511393776445582,0.6065344755174039,0.6950364994263435,0.32755989736278723,
 0.8436624338250661,0.2792952933452779,0.3994456221726804,0.8757952673478983,
 0.5282287921335167,0.5623935952911883,0.9689675214089737,0.45786185261222023,
 0.3468224537125284,0.897336843132865,0.7542299490895426,0.036659859697057096,
 0.6492671818534984,0.46073748156108707,0.8284243512044197,0.7246423354580589,
 0.33376087937533394,0.17509414777251642,0.46666548236543504,
 0.5379109540111255,0.09935595545691289,0.1317862160926886,
 0.24949168121981757,0.6857469199504944,0.7063948236411706,0.7912531573002654,
 0.011005044763062966,0.9878904047007655,0.35471939195538693,
 0.5320963697493439,0.48500650504371834,0.4286911795067764,
 0.39429296842559425,0.15316385422330803,0.21871790076519404,
 0.03098971058270472,0.5733952954247313,0.10988748838299545,
 0.7233217979917755,0.10919624617138368,0.5534126124526304,0.709918482441737,
 0.3523349844306848,0.970077150450317,0.35056176437913145,0.6227396708984713,
 0.6152308138384063,0.2998292748810877,0.8807155309545713,0.518829095311958,
 0.4078818383149253,0.8386822683246355,0.2340549825769087,0.8292711766742071,
 0.33907678631395566,0.10138463568062084,0.460912052157991,0.1849229384884894,
 0.08709189132242523,0.010441799594339884,0.8983487250602162,
 0.6947879016034343,0.6136963355518387,0.6823427244899187,0.9489865136509518,
 0.7506570154499925,0.7917979571351862,0.2275346474489055,0.20397497328525072,
 0.39505532559048695,0.4583735444757222,0.375628435536422,
 0.044032784798173075,0.49070445641256455,0.16994024667702812,
 0.30686062118619817,0.7225931861749033,0.751943018060272,0.1500970829121563,
 0.10245104655131954,0.7798153114772911,0.41987732931573496,0.515334221087927,
 0.8867728168472311,0.6680213503116731,0.6329385958649493,0.4374138973352073,
 0.7820680653295079,0.37967187152378357,0.41498988317963514,
 0.16661760140370907,0.073903565972261,0.8847040783246062,0.21471739990632288,
 0.8133135034682695,0.9819140592913204,0.5115212639519358,0.6237473453667158,
 0.5749645739677783,0.431701186008817,0.993027840775522,0.12049038308690119,
 0.5135287496906472,0.04052784135990628,0.769300633504842,0.5589253596305438,
 0.6199655407674738,0.41760400384142626,0.7028467787414617,0.3269010883042561,
 0.3629492506959957,0.7611728842590907,0.5941003825330509,0.9315619242614023,
 0.3667520778990929,0.12872648325363367,0.2567127941313909,0.2049798371110313,
 0.8373058446137234,0.15865313920025803,0.29509014988937154,
 0.9100763580247537,0.35515281238508833,0.9921543699197165,
 0.22027499888032442,0.2846017595331376,0.4302570447832953,0.7418451432081697,
 0.4114567628079657,0.8846668905817607,0.9916743374644552,0.03378228492200697,
 0.962272690414912,0.2645154155565391,0.2092167732713378,0.18504494007515482,
 0.5999870995642868,0.8073917543406272,0.17644065023055888,0.6855990571595025,
 0.7414885983723081,0.7008792645182063,0.9084437603912429,0.31783103884608055,
 0.4964970851176199,0.058382724564308486,0.7751221420379092,
 0.25535964451096815,0.06343713668875783,0.8609058636635807,
 0.9010560891876457,0.31323293266191976,0.7520879856380136,0.7238308174688893,
 0.15068788147039847,0.5449558882859533,0.395663005650453,0.3470430841996236,
 0.7543592037815725,0.7944656862719666,0.8591246564922397,0.48281116969291293,
 0.9508284854553397,0.47983959384054964,0.9299589401196211,
 0.43932034768666417,0.8367758010837583,0.5744496779592271,0.6297807934615537,
 0.6012189242242486,0.755741078918454,0.7498329564695553,0.23654171196586593,
 0.3723733435016131,0.9215785594062605,0.9262069475673395,0.7268957870065714,
 0.8944626284693431,0.5261615059263692,0.4411992646978562,0.26371116971065,
 0.07265029993941352,0.4219717400657129,0.8760038140146564,
 0.15383048911056807,0.9796386451399188,0.8520748846121359,0.3328866559594733,
 0.49152307859081334,0.2217042178214309,0.8373680630450981,0.3341888181955672,
 0.9458126659980742,0.5550534498374231,0.12317741808210803,0.7116780407049982,
 0.8213850182041316,0.5899653244405572,0.31572827236244505,0.8834823171281259,
 0.39669930312515667,0.41075705079963065,0.09004837485593431,
 0.6903832493581413,0.1905913626745649,0.6813388531593326,0.9987371422451435,
 0.3472389099430815,0.3120633984794935,0.29661907009235655,0.6468602392234449,
 0.9242353055110806,0.7307921128442277,0.4270781920615868,0.4604641001546842,
 0.3254584228323524,0.0289461322421235,0.509666929788899,0.9585505086564929,
 0.13079891487644701,0.8464878657018357,0.6852933588807967,0.2923818502536675,
 0.2688680487843167,0.9579441430606143,0.16644745635359348,0.9416183663120649,
 0.419733586039295,0.7144723179978001,0.8723479517149163,0.3810285424635663,
 0.365964120425719,0.8944893114289273,0.09458577313343941,0.10258533338112497,
 0.7592707363219271,0.26752891026636094,0.15450499041642418,
 0.4121492327903542,0.25303288767225773,0.09746026790617757,
 0.28904852647703105,0.36903535500255713,0.08967114656486386,
 0.3781933331277417,0.5646534658564993,0.39024490115260513,0.802643561106105,
 0.55474591340202,0.47012145016286255,0.22804361876376733,0.5005332945904638,
 0.0935124753462343,0.42168414620264627,0.19061066988735353,
 0.7188357519319263,0.04271744811162004,0.7165287442391532,
 0.26240608552560973,0.9079300510657802,0.11970933833353725,
 0.3626579524742344,0.2739914215428726,0.08161826967849573,0.348997879611018,
 0.5587497968419453,0.21481049253181705,0.04237036016959295,
 0.07878558528946988,0.4322232369426734,0.3725548178269763,0.7114412622877373,
 0.8394319035596683,0.7556493269615935,0.10599865771096861,
 0.30114627498360547,0.9167374892501781,0.17965575431631353,
 0.7714056966641207,0.2846938964236507,0.7212704423853604,0.47811653486714945,
 0.8052048246302921,0.5316739331714264,0.6981013693413732,0.5479750117518014,
 0.13914615693221755,0.1315776061669166,0.44042510710782345,
 0.16692185656934178,0.37983785083720933,0.14996455065961312,
 0.14812471671315153,0.3718931889878486,0.31897058449154037,
 0.1433754350516956,0.8813814474758298,0.957635424896365,0.6693799963461009,
 0.006709266261259894,0.8987162694917592,0.7524397954644637,
 0.1877942069580153,0.6232332290922113,0.9042952666657968,0.9578842273206881,
 0.754442849046919,0.11604928991894903,0.32243891047597006,0.7650194403500087,
 0.2274981251292736,0.9369948831741043,0.060975863646341555,
 0.8013544723394981,0.7061532564583043,0.5431089717359191,0.07948080236660449,
 0.3096580962631934,0.27464653832885255,0.7756538777026645,0.9719577864839469,
 0.6643364921165075,0.5812707076690533,0.7355047988544059,0.32139247220964984,
 0.8571028078828462,0.8901045620427084,0.4525684499085739,0.31392650125457533,
 0.5717700622998241,0.9929051270702094,0.4740461849444202,0.07213941744909935,
 0.9121615995056581,0.5546959228443172,0.38661735066291136,
 0.22168150579668167,0.1310584646817472,0.8636798553007513,
 0.004593143847229864,0.5262385270825196,0.403149921365626,
 0.21672959589864582,0.4331405434103819,0.46859320766331225,
 0.2518655981876319,0.11216828980043758,0.5389519898482913,
 0.15060126914752103,0.6605807273862664,0.1381881029755857,
 0.15110731531232247,0.7243001051085254,0.7729453132834901,
 0.48125268443549285,0.782905027043433,0.6599699525211283,0.3296223725966255,
 0.8887087854303682,0.782736072900191,0.8002401925933594,0.7957765391464715,
 0.9893783542468723,0.0708017778324408,0.6171201105049584,0.46963190987925696,
 0.8068115435053249,0.7796263798925316,0.6598331093577399,0.20976038288641474,
 0.8861732909251441,0.40790850180770133,0.5886365250463609,0.8156751883012611,
 0.908320474592841,0.09732740064802359,0.47097294249870236,0.613953627548546,
 0.5054635050598794,0.5427257451732335,0.43365361786427703,0.8673443530492518,
 0.12464674955420274,0.6806783624350564,0.06762523730736902,
 0.9738355553135749,0.34259360468897493,0.4213345439375774,
 0.12030685359824922,0.21240273067666515,0.8353903481276664,
 0.4069627550580389,0.8422560730758041,0.4011442723410641,0.395135920162101,
 0.9671277862473402,0.25331918795165287,0.41837402782254973,
 0.9872151800532831,0.7378473660138548,0.037479424912707504,
 0.07059830158655678,0.9199955516714993,0.07257379670312003,
 0.02515381972092301,0.6854516139146096,0.2973649281600478,0.8231961161710919,
 0.41846672830445475,0.39154375561985333,0.08040634483213094,
 0.12871697630918988,0.3769146713254329,0.26273324408862375,
 0.7780742321404277,0.01726831825656916,0.00509925332588923,
 0.3144886997559495,0.778100433665613,0.2732981417460233,0.6881766123620074,
 0.15085800281180262,0.08608413304288676,0.8555734305117151,
 0.8335073330406628,0.7694701347129381,0.668960534014126,0.48544066663957797,
 0.15087923613572618,0.753881100597799,0.7723560041183541,0.25162527496043463,
 0.3524511147096018,0.12497468001630119,0.565301032045175,0.10913707995371924,
 0.5789001973894466,0.39314521504793676,0.17384508944893318,
 0.9982039672711392,0.817537743438735,0.5544278833223254,0.056382321164414995,
 0.369660077774686,0.6296287675605282,0.8608932508311746,0.6721346541113185,
 0.9272990539476065,0.9973145358348916,0.409303172371868,0.44918594456428795,
 0.269499631973274,0.8692918208456671,0.11913636285635887,0.6942333285269112,
 0.4302719100504391,0.4082578083539795,0.2823520925876035,0.9103443581319642,
 0.8168260485898172,0.44006264684138463,0.7725878881220144,
 0.24828221325269695,0.40212715711018177,0.6231882582866963,
 0.5252830067402106,0.07276547611609441,0.07502350941400315,
 0.9634507368801705,0.2639869015772843,0.0115067976978106,0.14773119626385023,
 0.6278370780943895,0.14331446390306612,0.4097910526967009,0.9873022677728958,
 0.8206127499945864,0.45414033832293954,0.42852409903350386,
 0.8320619337181504,0.392344333387149,0.6468373412011025,0.7182639693857604,
 0.4973485900855301,0.08415467744705318,0.3330053103243241,0.292271438188202,
 0.4217238146038821,0.8745400513579356,0.2233685741529915,0.18084065413800243,
 0.9969042039742664,0.6256545922470822,0.6325775577595064,0.6335745934703484,
 0.06856350296490143,0.5184261136267938,0.3797412948961618,
 0.02251580558025512,0.26566659441113405,0.29037535078400767,
 0.7090774786556434,0.6834702911426788,0.7464421593562377,0.5820986753885022,
 0.5820106867815025,0.013086666690687032,0.6560779056362317,
 0.3455201793649968,0.21381670157413257,0.851310186227344,0.5619065471152631,
 0.9752449529644343,0.6267611566604486,0.4634383888637513,0.8228296827356288,
 0.6165575595512985,0.07106492141282228,0.7729464216058279,0.2233085208891723,
 0.28906120495252174,0.8564802769051043,0.45587231550127105,
 0.6297511255677789,0.7600007111357658,0.7916842993219418,0.08843978102459193,
 0.6819655045164257,0.39810463734049706,0.16084362085515858,
 0.48616608720695986,0.6219792784534712,0.6547178685321137,0.4793658254022628,
 0.8121248911498516,0.24966653076290202,0.7174585028720301,
 0.10394053118623314,0.9427703652738173,0.3650028854748345,0.69671584925901,
 0.20445241755803978,0.4625553606667344,0.6724471431293956,0.4721267534043576,
 0.7389348109341886,0.7520323428518224,0.9825760625023943,0.12642483804968818,
 0.6795673717835402,0.46496347175161523,0.14799982388467703,
 0.32443141846885715,0.2826053713882265,0.8203764888415997,0.7726114092351896,
 0.5597003509046439,0.9740669383381949,0.6241522351109562,0.8605429802547302,
 0.9616959700494639,0.24353391829915783,0.18525832881627924,
 0.6152455752784276,0.7931117369824103,0.794302154816157,0.9460243183706478,
 0.996143897832645,0.31590147661018486,0.05712907074797291,0.9008380879939136,
 0.4413094264027313,0.3013915632262967,0.4513261117189631,0.3666260285858792,
 0.2837956304651037,0.31781889870587143,0.8548743898539107,
 0.28886855901559494,0.8889282381203074,0.10639849285120695,
 0.6017694069041334,0.8400462288171628,0.8844396568985369,0.726290701340322,
 0.6547711712630245,0.9254349138861114,0.25032603558340816,0.7451475203542545,
 0.9328352247373519,0.6880534544929314,0.12487720860962626,
 0.20422747416873943,0.30595829639306393,0.1452025114984249,
 0.24866438082423015,0.4593340782291415,0.7510894872878459,0.9131543361262175,
 0.9378682985585106,0.27652728938724724,0.4213336039009391,0.2978964937284021,
 0.4399894922928038,0.07919293185464382,0.25164887839096306,
 0.44599853123813304,0.4191089482335436,0.4813672082182322,
 0.029719974815900496,0.9947322252305374,0.17120633349501557,
 0.9622733566167548,0.8425254817653991,0.5621963287112259,0.6445040744012006,
 0.6330073054411373,0.3668037004453123,0.8584860893767748,0.29543200609531406,
 0.7743870306263434,0.3808045865399913,0.8094490023823467,0.9615748969088064,
 0.45129802628087523,0.6225538209255146,0.825860905456792,0.9754105590492034,
 0.03676485987624356,0.43046737085531284,0.6394878076725697,0.936520212881109,
 0.501136508496248,0.20499051891685005,0.6515060863100798,0.9055993701518402,
 0.6167571891482507,0.2732249795134558,0.6594535691888619,0.9468695179240183,
 0.3855105521319635,0.6962807920408912,0.6268420448217621,0.1056837125238238,
 0.3551676623077368,0.9580683518651117,0.645116731943205,0.3179068218507308,
 0.1803488973247933,0.9883393333441969,0.5878719457764824,0.5771364980086046,
 0.7616250952679559,0.2899993587845875,0.5785197296397518,0.19431193376805267,
 0.8982873941258878,0.07277172517905361,0.5778109763530923,0.241974253436555,
 0.9338144464082636,0.9586217574426887,0.6363606590679023,0.22689066186568096,
 0.44323995993281584,0.891093007749595,0.3914058664969081,0.2851789634900411,
 0.6898908665887881,0.17175465764819908,0.8267758094535436,
 0.002871922421927686,0.14821281797872687,0.35839090110547955,
 0.6633315426986763,0.2557896093442996,0.6761710921296497,0.05363741042017478,
 0.12089448656256974,0.40529856177106205,0.28267049083223783,
 0.5846093549806575,0.4362861585154375,0.3331005832165902,0.25962545400875214,
 0.6579634961025382,0.3865429284500286,0.9854601020034239,0.8371619747265662,
 0.02434435031820681,0.7412632022748538,0.17882383422969617,
 0.0742085048758574,0.8046140806390094,0.5996512301918892,0.12754356511827736,
 0.12897906037536766,0.3798328074469638,0.6152210175947886,0.0571540561782542,
 0.25269645718369294,0.07944272118154205,0.04077472803841542,
 0.5839829397326739,0.4246269011351903,0.4837188537741258,0.8337935046582416,
 0.8076997358954944,0.7444651788170009,0.8280837254868285,0.21894433634408994,
 0.37202977512524527,0.6406543803651509,0.8899971312431313,0.4876894948043795,
 0.13491029990800174,0.8900587486907185,0.13251438679340788,
 0.7800940094384118,0.5052658121842865,0.7666591335546424,0.570118531764333,
 0.6897281888988099,0.9907337211171553,0.9156100239786313,0.15851407725155087,
 0.8538454697056597,0.2209415313231211,0.9360939638842856,0.17417494200900396,
 0.7342998715798409,0.7309675717009535,0.18855562171914186,
 0.08914629294558629,0.20941396341087648,0.5557451255694752,
 0.17713663295125692,0.7035167632616344,0.20816298424522928,
 0.6015541454146548,0.7427924641563369,0.6249248036748964,0.18386346450269897,
 0.3850557003176265,0.9467503711901736,0.6284609486354817,0.4148605714899428,
 0.5011714433572475,0.3716180752663043,0.03543381062903739,0.4296512583508514,
 0.46509504694873516,0.87883056591663,0.2883093157803571,0.04438179462372638,
 0.5877215324300264,0.07492562835992689,0.9819220853000815,0.6461575049198695,
 0.2585233668677924,0.18276985822427316,0.18299571303367135,
 0.5426644109733543,0.6166133524026979,0.2804179880053269,0.8939129552550918,
 0.941114143200722,0.541816768038562,0.3038554901640653,0.7096963828454323,
 0.08281773836800477,0.9373313656290273,0.9550377755431422,
 0.29859648644507686,0.7174094363287196,0.8947134676508431,0.9763269260837556,
 0.7748458975958584,0.982603786076249,0.708039413567578,0.6832942402228063,
 0.6309366754835949,0.0228375878475362,0.9870323034787198,0.302775225996589,
 0.2234155465435469,0.20133673417490083,0.9832601936407451,0.2310957914991142,
 0.049448385481675405,0.3729729819334784,0.04998526535522374,
 0.11135635406254674,0.5475127765157294,0.8048077383177076,0.1506689667483122,
 0.6203527790847839,0.7543094042571832,0.8746674611424107,0.30896040841837125,
 0.06681388033642854,0.2766962591833453,0.6027431794030064,
 0.22761302321533083,0.5626121803139345,0.6124917446112144,0.9432401448013005,
 0.4230634643979714,0.6600765828750844,0.22586180495904606,0.8684837992572899,
 0.08615451296022747,0.0872779354315778,0.3529057335520802,
 0.30875613284568504,0.11933051659469562,0.8625429704880092,
 0.11623754394030572,0.5908013133289165,0.9498403110204697,0.6088585403090466,
 0.8865874857477651,0.2794444542317791,0.8378887165416969,0.6971088465255681,
 0.4289325986733351,0.5657545933569862,0.5904741467806384,0.6111615404791472,
 0.8793249800597193,0.45196299339284973,0.7354990152146675,0.8672863900670502,
 0.7945719530576376,0.32878538102128285,0.9017595781029671,0.5609176587166957,
 0.3736093191113524,0.3050828888326079,0.9596176184224533,0.0987023290278396,
 0.9133249007080727,0.09483290019084278,0.2697525127132504,
 0.009595815030833288,0.5480307663917464,0.6146433661207465,
 0.3706023023774273,0.6210352750631979,0.9225373543212465,0.12219362963691249,
 0.5605334601002586,0.8741919487106302,0.9282640710880679,0.9946727874332003,
 0.23621496596021319,0.9060178106704635,0.13041852022779177,
 0.7525505074627142,0.4971901188890214,0.6290789899192826,0.9020567157749856,
 0.2159517648211049,0.38809532435655714,0.966864837817578,0.3252181294707328,
 0.9524002703718356,0.5930915778458661,0.8189178851379368,0.25321180871866156,
 0.3577472409898881,0.6848169544875534,0.21103419556918923,0.6272873634107626,
 0.5198850056875752,0.6506051713601125,0.3883050683285416,0.7540132050790187,
 0.32221097990127656,0.9565170590753054,0.061651959910032805,
 0.43597162111703425,0.34376846526077764,0.530467725297582,
 0.11866884059055316,0.4708258477817421,0.5091758599630262,
 0.35116834761824245,0.9502794869297524,0.761065537132551,0.9010602029340649,
 0.03079219373704023,0.5504993907258664,0.42792119492880176,
 0.3833206256766163,0.39301374189157334,0.044833337232139225,
 0.7350157300382721,0.8825662470687459,0.7933934604747316,0.25160363671364894,
 0.8538264047415947,0.8941341298716297,0.6644725272811189,0.7946944926902886,
 0.10111178706452584,0.6359702850836171,0.878480032760141,0.8819646250928556,
 0.30267898118148084,0.25972797362716626,0.9057920801081774,
 0.5242261808811421,0.8175228591333875,0.6061782880159874,0.5871442048515654,
 0.8322331747412531,0.1706280812229728,0.35226464734408536,0.5878065363737507,
 0.4429013917470719,0.7853123127328238,0.7463084711738045,0.8626586236456317,
 0.5933177000538705,0.013296134320142428,0.42197131266444665,
 0.6338750755573622,0.2496329086517698,0.8015407646877261,0.9709497761459462,
 0.6447889161683821,0.49558326965323796,0.3060288590737009,0.7042034583076261,
 0.5142386541096471,0.24653680005793122,0.5567952248344998,
 0.05936582882816244,0.6431738767556373,0.6478795642687237,
 0.04775794577922532,0.9933232153191645,0.6864613609513728,0.3148437267265214,
 0.735882955483077,0.04614499002917882,0.35021450912603624,0.6603531977260398,
 0.846124764006878,0.66805246917774,0.5208410956258924,0.8204467701544244,
 0.019555823902819403,0.5337560558735649,0.7709999700815672,
 3.8488326688868035e-4,0.5125750373521639,0.37110607804918616,
 0.1422930748513831,0.76889695250676,0.6703460276352391,7.142063323285486e-4,
 0.2428085403540905,0.07016495882543183,0.7120342108466551,
 0.08927428784293368,0.4769848960701014,0.20788195036922585,
 0.6267375497891094,0.6847683953095254,0.5438876867102208,0.720964743451357,
 0.965361600137322,0.6357155906993295,0.12300589497579872,0.7286193147791377,
 0.3230099769012057,0.9358349942789007,0.25779742795004146,0.9167898805290126,
 0.7227015120677831,0.6997977947614841,0.3959712591652569,0.3562909767544781,
 0.2185757220418505,0.12405802656103526,0.8592515826160614,0.9709299269562305,
 0.4047163456762929,0.14513420857738168,0.15016133161965906,
 0.01049778742559937,0.628882384052912,0.20978240869597053,0.7475225685553203,
 0.6403187250238725,0.18069451463751784,0.1503787607009469,
 0.07220553493083381,0.15840821215547962,0.35089372343433,0.7276933377049882,
 0.310535099478768,0.7227184049828906,0.6052919272094918,0.5100230253934157,
 0.9639583769334078,0.46614707621622375,0.34905288455935224,
 0.03542908031828296,0.036907617683164196,0.739033612046931,
 0.4307498247000735,0.4747619853649717,0.8932848078596098,0.32381908056148545,
 0.8384928637895501,0.32428330686432505,0.9534283437781462,0.5022203377282218,
 0.5641049989953286,0.6827031484651561,0.24589725280313035,0.8782366281958031,
 0.6204471690427809,0.1105444489624503,0.6407904997025234,0.33735542302357047,
 0.38692999927036564,0.5784690511571964,0.9298177272028358,0.7324073572167082,
 0.5467923703436687,0.10403995893139328,0.9695351079751827,0.8229572338729573,
 0.5562656231619434,0.7654544119739698,0.03061356636046353,
 0.19309424319585466,0.11453517972883798,0.5991318693041435,
 0.2840578494661037,0.8680937933068594,0.001583641721685014,
 0.8300039580771865,0.6545102378784017,0.9255533461957608,0.23744414692100946,
 0.5582325124047741,0.17754466379419376,0.9198612984323277,0.4329128002152538,
 1.9575981346575677e-4,0.7373066234196641,0.3081052682677061,
 0.21609104473439755,0.37149903299328535,0.11799967908913489,
 0.6026432370885204,0.912491457605302,0.42095192475391974,0.5753024923567691,
 0.681274767200853,0.4005746951353606,0.25495147916645844,0.7137056229570456,
 0.7674294442254477,0.928363370068956,0.3179550698635062,0.2828626969285508,
 0.9370587677493075,0.30136488831531927,0.9347103154736114,0.7428735144367354,
 0.80353191542995,0.019006668883370326,0.4937384441158903,0.8137067869085991,
 0.9015061143473809,0.7138900742746264,0.48453726801319574,0.633343066095252,
 0.715330020647252,0.7581363941831872,0.32176703276308816,0.8647733590735227,
 0.28278951847219624,0.4720711519351678,0.008522605005618855,
 0.05414192213070701,0.9191890686819046,0.9574951759289959,0.8512117353817323,
 0.6086793877161009,0.20132338385808712,0.621516578946986,0.2507309622194652,
 0.2975437787750199,0.17073361753096217,0.8328140057638276,
 0.050490430880322346,0.31457398892055644,0.5527270159790446,
 0.2126445463510942,0.14968634586641882,0.35499510802230283,0.64205906777947,
 0.9498759367834093,0.5428147563703272,0.7709465959710374,0.4355799344618718,
 0.8183085443530196,0.16015745038675866,0.5869596135797203,0.8870508713763084,
 0.94650805988916,0.21581729598789667,0.842284615884986,0.928009605393246,
 0.7220529489837556,0.7568393743330797,0.731638991783414,0.24396761278048662,
 0.14088867913202396,0.5642193148753002,0.17908113071947285,
 0.15142020306335757,0.07417591075511254,0.23061794727845075,
 0.9726911226060564,0.6897278788656412,0.4483393303570744,0.8537676429158296,
 0.9535776203481711,0.24883194054904412,0.304053067838597,0.5157347990213771,
 0.6734048950798261,0.2159861307410741,0.8382530358174254,0.5154923340500208,
 0.8121377045904851,0.8330529738080873,0.1405861979754507,0.9108144287595648,
 0.9935411523895471,0.3533705004038932,0.10822202022142358,0.9932829770136633,
 0.2928590440949651,0.7963798818083747,0.2085545606823641,
 0.040409620228305965,0.8012195634589976,0.990639802928763,0.5675412493706894,
 0.28699260583862085,0.3117824997403027,0.32510117509755143,
 0.1693914883910519,0.7698976209677966,0.7265223349328258,0.5928261176420925,
 0.7865374822662412,0.5054511650208655,0.8170493912624723,0.3326253280268293,
 0.09342711167384743,0.6314225230554612,0.6084397642741237,0.7862347913632958,
 0.94092025401592,0.1743146368550521,0.8808801646931214,0.04236702599206765,
 0.22488803213443054,0.6673592049068275,0.33961417280816675,
 0.4815525517378968,0.3323370913308611,0.19811897630705644,0.7447226642968233,
 0.17472457135010422,0.941602763890028,0.03107823762629025,0.9063162297591836,
 0.7279583907493096,0.24724387971859185,0.46710711676847616,
 0.5382378839954234,0.07894904963336602,0.6849737535119751,0.881619107174064,
 0.9390006106293892,0.8504257787042651,0.13552917526355202,0.76804639206357,
 0.5632087864903895,0.8127213671111421,0.011768073250045141,
 0.7453192558109761,0.5504665955414902,0.2436308084855846,0.3985896986198516,
 0.362718954794113,0.5663488681847563,0.07419073238668239,0.7775139467389286,
 0.3309674127607125,0.5906730651368504,0.28212751583851925,0.1786613681959569,
 0.34805359665910585,0.8421948532573406,0.9780123927081298,
 0.17365356611329086,0.19361049324583912,0.4174094000824077,
 0.3759807504010967,0.8233881754530941,0.9439607752620178,0.2652135447450398,
 0.18709558267810122,0.01226955354530801,0.5256195207838319,
 0.8441485343595438,0.5768658962929073,0.2854757431683159,0.770655193289078,
 0.09825262414900515,0.35452365249638595,0.9600970041543485,
 0.5402046278024685,0.9552632373672099,0.42857038817090465,0.5473153564914923,
 0.4177865658302711,0.7849107057706223,0.5723183222023946,0.9578155047263307,
 0.4170990772038897,0.6794503253156893,0.18944270185555334,
 0.31040835168616165,0.47118789761093516,0.034390581320638214,
 0.9441359233883844,0.1639857706011838,0.4718903259315739,0.3654336034461968,
 0.5960052214002748,0.9797721949714352,0.17089792528714198,0.3388405520257084,
 0.0022304176717217183,0.6848824415986599,0.3360419645743107,
 0.7497605558963387,0.10324353605973391,0.08036450425059305,
 0.3004244709163518,0.6008488482699449,0.6530835978798025,0.39184579689923316,
 0.12433673755706509,0.15414715693187941,0.8635628904876806,
 0.042369936817412523,0.8781344384977636,0.05371948582564978,
 0.5099243250246985,0.250980886200577,0.19163151146138135,0.38837477412823607,
 0.7010668676467618,0.1538423018163848,0.6084274042118055,0.8691111387404259,
 0.7177916316889925,0.0999996570661108,0.7177576720204544,0.5665879399234685,
 0.18562308917572112,0.10760899183968353,0.2835897418077463,
 0.25642065271253356,0.2852050771553255,0.6643801285196576,0.9483233399474146,
 0.4315986018677994,0.6967285989800427,0.49770571152033305,
 0.37896756628817485,0.1538315067636844,0.34154129012663537,
 0.6190355750330255,0.4946670130140257,0.5050152835184976,0.9369956181698343,
 0.08873169314476148,0.653314228580543,0.062143030284602996,
 0.5955454059669705,0.13080253765573546,0.14224125706835888,
 0.7260227827171158,0.2032825707542888,0.17006785386528234,0.4907409794037375,
 0.6956521559121924,0.6079431007864943,0.6093579548737388,0.0185737259940727,
 0.7748492115506869,0.4111633352636388,0.8112119375457192,0.48119070167934697,
 0.23281423897894993,0.7537043889110757,0.8276153140386378,0.3966302228428614,
 0.36743390816851706,0.541475405548521,0.9859854016499393,0.980352302506339,
 0.6128769247237896,0.14945052316586416,0.6021765035519892,
 0.42299792121649404,0.47161489708395565,0.7137881373342503,
 0.07357315510564655,0.9015376671580455,0.45698238586661966,
 0.1451780801362571,0.7422187683873324,0.8025472542898929,0.7088042343713702,
 0.1319940764717944,0.40413982123426495,0.6782708887417089,
 0.27132460249240664,0.5923160288371377,0.2425737121505872,0.9122447383857677,
 0.5342727752796641,0.4591634426743195,0.4103518924573465,0.8281507224797,
 0.8584351937458098,0.16494420613665917,0.7425405164638569,0.5764358667770859,
 0.49427264282191663,0.7763410606618728,0.34531993880404155,
 0.37463508396770995,0.15906218892536272,0.9665685591064347,
 0.8338643812532274,0.6361560376096523,0.6227218915443551,0.215622049789024,
 0.9394211843424611,0.28528310853137095,0.5106908614082422,0.8468363456465192,
 0.8314897870196364,0.13777772292814428,0.531171755404046,0.06386654883522147,
 0.42469172552328005,0.9027866992354563,0.7384376251586566,0.5344207131666074,
 0.5410496869892455,0.35216436179090593,0.4162484399391593,
 0.23996162846502822,0.20580608100301379,0.18031764101264358,
 0.8676490922904483,0.4330436441307908,0.14907168719430874,0.2828699946635155,
 0.3913531638398777,0.3202457243082867,0.7014030641741766,0.3052453695597307,
 0.2190782520903075,0.45174781453986756,0.004005755410817491,
 0.27559885009582175,0.5900556916444871,0.8792021239957578,
 0.42599877412770093,0.957713351452143,0.2252180278873659,0.18451627548135696,
 0.5937079219166094,0.01489785514251063,0.5464515276116628,0.746276230466166,
 0.06138260316616806,0.34484928193212905,0.9527667726446565,
 0.9895296021721768,0.21469620060746297,0.5071336351613698,0.5114913232327505,
 0.831627244802763,0.5319813301498509,0.13019473521610447,0.3113871587356467,
 0.40667729103284433,0.4115503998803268,0.8680944999315545,
 0.34683222188242324,0.4206820275476739,0.7069317853469537,0.839236586160107,
 0.6192740829832521,0.27288244294787567,0.3200658480663706,0.441011548446003,
 0.9904690067833681,0.03796648050886142,0.9418351220768331,
 0.11524603210152518,0.9071219306710014,0.9775297789393337,0.6644506325744268,
 0.8136884797360329,0.30479577634428034,0.07631424298633305,
 0.6246164729879865,0.9348912777466056,0.8239787690021468,0.8495474671857541,
 0.02814824453498599,0.6157920649535427,0.1648080018506679,0.9024090423142382,
 0.16799237837564562,0.765227847000285,0.3545260824363684,0.5745513187661448,
 0.44993876820257095,0.43995978849831674,0.6787268451470315,
 0.11452455694096508,0.21800758249216146,0.8607742254632238,
 0.19610380244720116,0.6628995561618082,0.47650600953903055,
 0.9187895885322327,0.660797747658987,0.20804158140268503,0.614185438251157,
 0.12093918495828015,0.11645722271502934,0.7834087718023715,
 0.27136393917543566,0.9458925197732235,0.5181196361991298,0.3911851008858299,
 0.3393622320450842,0.7528901275954053,0.24235436057863502,
 0.09158420801486045,0.7694740073452703,0.9343562205775459,
 0.09615582034131165,0.22937616044616282,0.1381230655622171,
 0.07754509848741853,0.11005372668310154,0.05180804974373332,
 0.5833263521197886,0.9573896839487186,0.9680277332817175,0.5496397960271399,
 0.9578111427185326,0.8312351603293943,0.2473167549901032,0.6456524982803958,
 0.2082994598316854,0.3318918116639008,0.4021888923649929,0.5402206790701205,
 0.9129326209720163,0.5883204858428407,0.6977093973954928,0.03550977226782881,
 0.5514522138703332,0.054835372227911794,0.03113353079612191,
 0.9276879508224468,0.5001254824440264,0.2457838331324247,0.8342724019243974,
 0.171096756095931,0.17895036219557747,0.3922647268715356,0.7320007869645048,
 0.8527433890995924,0.6899947420257169,0.4851796332059891,0.3706173114492137,
 0.8378524176632943,0.6003588578220225,0.30587614294357934,0.8832698688591843,
 0.008994594053791993,0.7319042325564198,0.4480269408002434,
 0.14775793952669058,0.5880039090926595,0.7483372307561886,
 0.32624520322092776,0.6791913499445602,0.21742067003190912,
 0.33571702248289026,0.2732250681775251,0.5559804603264347,
 0.18208268605139977,0.09991395655207302,0.11860325095743074,
 0.2779821286030061,0.08628608438951824,0.34422786632229063,
 0.8049095449602465,0.896698693637434,0.4848649708144762,0.4234560674085742,
 0.6357618215386609,0.7011471447779299,0.7087760686297441,0.00440087332707495,
 0.6673514371823202,0.6308555096550251,0.6514640228003442,0.4912547389673093,
 0.484859064979663,0.49733679475320147,0.8662683697929193,0.4020541568095982,
 0.9837070213910224,0.10354166480356575,0.3234306717713099,
 0.14458087656664098,0.5580944831513747,0.04366030816028699,
 0.5921289119899334,0.3062737744711068,0.5170526972369195,0.02990946472686884,
 0.7376071201489438,0.2977508987947971,0.6208512169506094,0.2304788639673374,
 0.21823187826065948,0.1050437330123224,0.17232185285359658,
 0.20959680859588836,0.7860045142205884,0.039357776549868606,
 0.07227095017529361,0.7307630006796182,0.10196971506435182,0.897748357064323,
 0.01882265987362053,0.5764515700459211,0.7613083797016389,
 0.12586175320763504,0.7980616137236766,0.5326124067283525,0.6503086403903262,
 0.5972748262486681,0.7001333588309872,0.3090207867079853,0.2552080872262581,
 0.5279361617623974,0.017212390396987853,0.6181560064668461,
 0.5167672721158127,0.19695055247803295,0.8190162752156543,0.740998250625551,
 0.5936643725635119,0.6431578847238211,0.746369545987022,0.28313811293979696,
 0.3972224035265284,0.616856813163974,0.6734145520260219,0.6731722437488323,
 0.7018482877054919,0.7759281312490962,0.3143918601352792,0.5355869771132032,
 0.29170096287854785,0.7692028456838568,0.10989763115352058,
 0.3858412646661422,0.6039346684878668,0.8499213504616,0.2268780576616124,
 0.44938751904582175,0.20481240629744946,0.9310920383202684,
 0.02055076010895307,0.15512403200832092,0.9214510157049018,
 0.3348943204946594,0.7064440199170701,0.1059727035309026,0.17790857517456804,
 0.06331108115928386,0.24645511125509012,0.4623939842521181,0.724779501071211,
 0.9940278558981338,0.8551459965999172,0.3132208146143134,0.521660295457868,
 0.5362242458935131,0.7694968567047688,0.23425908695213893,0.8056553813001666,
 0.692241577223769,0.08196962151743636,0.22292346188844747,
 0.26839041762315574,0.61521058530545,0.9780169142599167,0.39889367458249836,
 0.027667363989297145,0.8470311172089023,0.4065629537218164,
 0.21223425253999417,0.9631987068251987,0.545939237782938,0.12472523238622524,
 0.7954369732689592,0.5174750702597493,0.9991937239273768,0.457898363122929,
 0.3942646308858968,0.5049131907908715,0.549297518909577,0.8133133798575303,
 0.22476330700379155,0.019498024976136064,0.49073765867245456,
 0.4564046781119738,0.062180229711826795,0.541979204449385,0.8982530392144907,
 0.07128259347536503,0.9404132980893029,0.19250891611520893,
 0.32479095176133876,0.47357229756425734,0.8329738743665782,
 0.3966341559297101,0.29877115628311357,0.1842329175983639,
 0.25375357861930237,0.9270731391034118,0.0026022357410954156,
 0.07130835418747461,0.9530752798953167,0.2913392740181069,
 0.011626735897798723,0.5949788023036376,0.03773017570505233,
 0.005717583142656135,0.17211297354240374,0.04072459058974065,
 0.463741225962667,0.16102822749215995,0.6872207918498244,0.9754827760631383,
 0.5126371721517063,0.3251584900960127,0.6002508769757716,0.5088551302570352,
 0.4891632151185805,0.023325376152441568,0.6117480281452721,
 0.6367436341536173,0.6331032500511399,0.6205640568386682,0.2286981206120665,
 0.22830062074613444,0.3216004188453382,0.055692445183215034,
 0.2279501687735952,0.3756746735472859,0.6311731425152176,0.7133445643814101,
 0.8305366674489003,0.9779199300975728,0.752150528343651,0.6769833765522294,
 0.5680454883411048,0.9224113900541429,0.43672130525028896,0.160051924214608,
 0.31072797691001897,0.626978459210521,0.6954760816068215,0.666718944932345,
 0.5400908840933807,0.1605892175069492,0.5480551675826011,0.12928541511659908,
 0.8989900452752525,0.45218263893318134,0.17783481416979252,
 0.9968616370774104,0.47199554541295863,0.6530354888119233,
 0.06034147488019426,0.8943483242055916,0.819073087145142,0.12985968333064313,
 0.9870123110844635,0.4951343958403406,0.12404441423968371,0.9379271327551414,
 0.601785937351345,0.7150640576608108,0.01013912585254162,0.5734722603055948,
 0.05037531264446082,0.024663750488990055,0.8572916221820956,
 0.37197829021736384,0.782388046464835,0.5014766074712922,0.8564749098725126,
 0.9569237370844876,0.3709745792381527,0.5583132768963188,0.5776615541928255,
 0.134315848982402,0.04396621793355704,0.12414984520847971,0.4880703199323928,
 0.3341235061897295,0.5356335041861311,0.21044283673068975,0.5042573148060412,
 0.21850429882319788,0.03669282328500334,0.6246477915123438,
 0.5635458612485551,0.4098126163222471,0.7518609624546057,0.9258175669184651,
 0.35824030041551147,0.9377874829279378,0.3228287895526297,0.687012654068726,
 0.08272268905222246,0.31563337819867765,0.43585052286167913,
 0.21948018687752924,0.7348251135452122,0.9623982102480136,
 0.18765282085579327,0.7268288317424909,0.6127485221579854,0.4264485504973936,
 0.06271561329158792,0.7240151673089089,0.2923760636718644,0.0815786185319679,
 0.46263366397285444,0.20241875543787613,0.7214895648729167,
 0.2021579418655528,0.8052894117995216,0.6895618006635181,0.1879053428236792,
 0.24611793128492265,0.7610543760566735,0.2993406349092552,0.6304352202503871,
 0.48020614961817454,0.6397628567248476,0.622593931286673,0.3115715060443206,
 0.5472688499343681,0.6110060289500046,0.7779741637101273,0.6592126192192447,
 0.15219779974085224,0.34543836539112527,0.6937584862138154,
 0.12494297289813083,0.7854714879491683,0.5810158141576699,
 0.47603946600534397,0.5804311017192105,0.5260342009804218,0.5681948645866809,
 0.9068433325785149,0.7486269743602902,0.2974482415994725,
 0.050729975322211684,0.6726340688183784,0.5190080502177269,
 0.8843522850185301,0.4162675412068262,0.9421744513833886,0.03340827999310714,
 0.38658494598385584,0.029264513751635635,0.8723015873660634,
 0.9032868857617826,0.6168445279253051,0.22108622533844935,
 0.46708804709876617,0.19071183496029342,0.9640923876699979,
 0.2661134677779453,0.9874946189356217,0.5139132765280386,0.9038281918913208,
 0.8254296456173383,0.6273038868039872,0.6447931893890827,0.1267099314894221,
 0.7528664123031871,0.03532842011944903,0.7632740427959096,0.2271781996983535,
 0.7832749283972429,0.7569613888411364,0.18608748737050496,
 0.22302885870870615,0.9869007892719542,0.36239609260731476,
 0.5087898807768947,0.8860500022672142,0.99603164404858,0.04631148828369458,
 0.12387976673387158,0.37717434955359286,0.8393330736434232,
 0.7059606375225793,0.7851500231876724,0.9472931293449623,0.8592495832039729,
 0.04856398678836549,0.054493876186013424,0.5261815661850937,
 0.7280352358042158,0.16707430148581348,0.25472578929927947,
 0.3180057703905681,0.8926896784019145,0.15536578708037496,0.7593605171844187,
 0.6250938391367206,0.9561447287364472,0.15684018398297817,0.2706407127280962,
 0.45443981571350955,0.300789943494973,0.5386272135374501,0.9822939359886516,
 0.39687458848648194,0.3264255790234366,0.827966889989852,0.9048531939452772,
 0.27537987814076315,0.6358576621628275,0.025351744211746463,
 0.13749062079155994,0.15426553508117336,0.18769654858159812,
 0.0581623998279831,0.9669310373307921,0.7096459827840137,0.14255967844132988,
 0.1660374668342418,0.45976769343354384,0.7396105418043055,0.9808313539181875,
 0.5921108719727787,0.6674291019650874,0.5163529450197126,0.3614596622459947,
 0.38072920341646843,0.36652104022508025,0.9396025535659307,
 0.7731773517468171,0.25155812893813545,0.370516896411453,0.1096586365002239,
 0.6507580944255663,0.6865625149268912,0.07040511237739377,0.9177532103435669,
 0.702444919983071,0.8929357037436144,0.43635302915709917,0.518011369524952,
 0.13005940737510424,0.2504927342205482,0.5887684512666518,
 0.26613688605062047,0.7057895889447332,0.13897381843184753,
 0.8941006001858657,0.7480231813790477,0.7020147210752791,0.22816051182775043,
 0.03810304859679503,0.12731138189089308,0.31211357422874897,
 4.386300446250768e-4,0.039577746386369794,0.6456930269891042,
 0.3971914860329704,0.945673242346615,0.5876531263884528,0.9798612759951094,
 0.49772929624680673,0.5472938195340112,0.2743464732637231,0.8131920095249056,
 0.8740629713220037,0.0428450002640729,0.666169090972523,0.07746795733023237,
 0.738699618464133,0.15799229787253488,0.9980709742433912,0.3633454569463882,
 0.15516204330004868,0.2908962429329064,0.16950038504954168,
 0.26579743504012354,0.6535033882607291,0.655600613943617,0.9375220481553441,
 0.7740820241144575,0.11800101530671503,0.3070970650747942,0.2349730701895345,
 0.4821115041333016,0.19582521573920486,0.8715493800059567,
 0.29677021537522563,0.3227435147209726,0.8466878202209829,0.8033474374590068,
 0.011976285147533128,0.9052251604373589,0.6978852068307543,
 0.9763598499891382,0.02153857743296228,0.9029245497058103,0.4567517312592563,
 0.047509394930732585,0.058090507043004114,0.13070826077217268,
 0.9602650958960408,0.3606209328064839,0.8604007550506244,0.8535725689014874,
 0.6941285473425648,0.7737768542979852,0.528863107229959,0.9157662733944859,
 0.44216628888352605,0.4286465194569413,0.1203979355460637,0.8276072391134123,
 0.059648744280009214,0.02319069061769441,0.8512277044941584,
 0.15509295050257255,0.26298047791213786,0.3141532194052976,
 0.034361882622405915,0.7849851337160887,0.7262252250717688,
 0.40487480565488077,0.6745804366391237,0.299381685355974,0.8934014801950911,
 0.805099172171445,0.6181916273393733,0.5797953714938887,0.07987223909221364,
 0.29944166737101396,0.13201116147230285,0.2980127865035206,
 0.7241938491877913,0.2931244547137347,0.8939739458781866,0.2216460572756651,
 0.5591692765585391,0.7352884640485602,0.18052849653633496,0.5488930713610226,
 0.5556809298142126,0.9108429047868761,0.8658110306836498,0.4696134675596334,
 0.4418800568642194,0.8285936965654636,0.6587658635034181,0.28502348917640274,
 0.9631902683350254,0.9159366698787487,0.9933232595635918,0.5052270250408722,
 0.5510138039139145,0.8968257943742947,0.5886752843150541,0.8991388488362465,
 0.5620918862060976,0.2477148911827438,0.9029226120492726,0.5749476964934173,
 0.057572731506531305,0.432361175770422,0.8397713133908432,0.1494475231675061,
 0.15685340976812534,0.4943837453371862,0.5910130604443231,0.6642566414948093,
 0.5004383582947228,0.7634603122056555,0.3615252598931664,0.500728114505388,
 0.1712493302318444,0.16427662393129405,0.2941755081513926,0.6330725210133449,
 0.3925547417719152,0.8903036711590383,0.06483868120128211,0.8309872932881198,
 0.6809150474427118,0.23731566904163692,0.2437276840054753,0.5451484848808529,
 0.708900820003613,0.19725309370836974,0.6320356351827866,0.44443981489706297,
 0.9264540898460134,0.6284806579888162,0.8982576664524407,0.25605572776553565,
 0.6422625825760073,0.254066333141838,0.7535342979562619,0.3385158880121475,
 0.9587840303251126,0.7076031518996566,0.246100098104031,0.8131761786982639,
 0.9576616070888362,0.0739010391153282,0.13163447620644875,0.8887352385792474,
 0.3779696566029639,0.47148191296383557,0.371535700453259,0.4670184354363357,
 0.12127292694250225,0.7337173461609372,0.056060755413699925,
 0.8451962887713769,0.2824076850074164,0.035258811327887774,
 0.9623381245113558,0.048104113409543126,0.9231961067155964,0.18409989113366,
 0.6100458465607774,0.9847811463253405,0.8253909385804501,0.39768318667753866,
 0.8086318677550918,0.27677546156465727,0.09969047466555259,
 0.5287417265675758,0.9256152947744934,0.8058515517679034,0.3881379046477905,
 0.23194507970158962,0.345303558032404,0.9855268795567675,0.19560467358621658,
 0.6326692542945623,0.8819707185486803,0.36240678675875015,0.9553931241266506,
 0.10770794933283001,0.15109904502586946,0.11177652705613861,
 0.5174686354111457,0.820362343686546,0.5520052086963699,0.2670775910998826,
 0.8112952135627434,0.7336124482000786,0.10008637804656184,0.9103148227038265,
 0.6986781983238176,0.2673851544186291,0.5810835145917173,0.8634478211828812,
 0.4206430692273896,0.08438142823599892,0.3303184675258126,0.7535542561320552,
 0.033539558314123696,0.4243752334764097,0.28391251206332746,
 0.5454094352716489,0.6174848896976457,0.5083767631936583,0.9594337539962353,
 0.6407606599139429,0.6078886585249882,0.12349477294711175,0.3743400450034544,
 0.633349264694849,0.8813516682778555,0.31769446280396774,0.08684657028216702,
 0.05128979979245629,0.35548549680540265,0.3228909049705664,
 0.16160839800155424,0.5889362736765572,0.7303929899140923,0.8066601506089762,
 0.23958486734225048,0.9331941002908983,0.16651598016573887,
 0.9965518525325963,0.9540708512234355,0.4070266386506003,0.17156031954884376,
 0.044979526765039135,0.7260987445880132,0.8610080155186187,0.618347432971911,
 0.9132527370537982,0.6263323836879517,0.19734389350356119,
 0.22751376839479986,0.3605113786620373,0.048267065366299544,
 0.9937752551254204,0.04842575217397238,0.16403384665010268,
 0.9602909324921644,0.966989885047469,0.11153169110072259,0.8306078619092169,
 0.4071055232780849,0.7265744059301515,0.9016226685263067,0.46405596823941064,
 0.5197513513324062,0.6932725184320805,0.11773134204934743,
 0.36810459818361596,0.5167503565005097,0.6303686697172511,0.6010394584508734,
 0.9366780105746109,0.04906202935690396,0.7569863188092919,
 0.36519271903112815,0.4479855802811148,0.5842934577433609,0.943450348656949,
 0.7785030359069423,0.6965323114690299,0.9255489621336815,0.9559656220725833,
 0.5556969129263907,0.4441487101896009,0.1999119549884555,0.9580946894096961,
 0.40418567736427546,0.21644539304110577,0.30095459195897534,
 0.8200750706554334,0.8358442643899568,0.05972364327210822,0.7847130411428237,
 0.7724094189335112,0.539735605739832,0.0787903415864818,0.071651733082151,
 0.6286136340103667,0.3358917854850657,0.8004660356818496,0.5332201502711238,
 0.5359680561388536,0.2873956929494641,0.4598813392291379,0.6701139012146554,
 0.5465792782175972,0.511176442215413,0.3657861974471035,0.5141550194007313,
 0.7440399678461589,0.9280178184391552,0.37548595808684826,0.8858832957981945,
 0.5667431944352893,0.17183704392427213,0.6263013363032242,
 0.047122738977792955,0.5657735604480488,0.09163190800081533,
 0.2763592553980244,0.6770297573349573,0.9363916720822958,0.31218338161525805,
 0.7353345602981327,0.835998563818977,0.6004016893562519,0.8459735126007235,
 0.9245858026559028,0.3279604078794547,0.07743499231935558,0.8585543222231581,
 0.039241472061182,0.15968795838429428,0.26170813786330005,0.092768490322642,
 0.5449632105208538,0.7297562691568159,0.5614371346684078,0.5538546120718725,
 0.7782147180240957,0.05914128387013695,0.015916212530276197,
 0.798569266850291,0.07568065811629088,0.26241309537310276,0.8744950933094261,
 0.23465271488095707,0.5624151833780211,0.28724373756948907,
 0.8427672340661176,0.8412864100771796,0.5112938414776289,0.2999087728689832,
 0.9828564564167653,0.13363570313797257,0.08146466585066581,
 0.6967622839179414,0.1288915459945641,0.07384842437429961,
 0.06689468694875544,0.9284456339635113,0.7973459441666161,0.6992287472636043,
 0.981728336490861,0.4154965131290953,0.06102475379613592,0.23257607154800142,
 0.6975135773686931,0.04770991906719324,0.07545747994234508,
 0.24297106954809877,0.6389122054373108,0.25088843554259344,
 0.2615041630225926,0.6552705437607783,0.5290287103249789,0.353448958193445,
 0.15943952328001343,0.4045898107760395,0.6374165373241869,0.9153579971555378,
 0.41264689732114046,0.5931495856544066,0.021073976490865287,
 0.8336403513715122,0.07452630678021044,0.8830291573999598,0.3932443183205643,
 0.9719761512592834,0.43781554297881264,0.5501850219603793,0.8341520788768136,
 0.2305650479238207,0.032609433058333526,0.7688294968888416,0.738798589405131,
 0.19833809609337705,0.09186837663889824,0.5065373674045212,
 0.7113419098653222,0.5383996135273477,0.16177319155756176,0.7908132398746556,
 0.3989297434636573,0.6150710894644085,0.08759511892069682,0.6501959081119202,
 0.31085395798102344,0.30179700858136294,0.0770870452219693,
 0.49616330122764596,0.344327530187428,0.29678573761163485,
 0.40267805075949015,0.8985480556603797,0.2396252167291898,0.4560026831863614,
 0.33580422309759006,0.2165022790826181,0.940277483615777,0.8188007394810918,
 0.16606532093278614,0.335510881223659,0.7814000584758954,0.24934493969893534,
 0.9297009471406873,0.8408031476216137,0.4861105323432203,0.7906738673710565,
 0.39919593228999306,0.10031096140756157,0.5099032890627682,
 0.7669089935502279,0.2749644188343192,0.7986518060502277,0.603403831246129,
 0.14990806944312163,0.36683180482236066,0.652493344289869,0.5477247845532921,
 0.6365098539541272,0.6089012957379157,0.30214277078066243,0.6337594454528902,
 0.9554666549032519,0.48992852608618764,0.6114207875422619,0.7519626596365293,
 0.6025385586713422,0.3584819275839457,0.057447195640764614,
 0.6647098718405057,0.9168605278443324,0.45324553395955447,0.7328281140008385,
 0.10314210625708742,0.014541715067512007,0.0175567467201041,
 0.07256593503707454,0.82796461614056,0.21976974207268118,0.4672499827086968,
 0.19223730653745663,0.21805684064521647,0.530815479615552,
 0.11596901259311809,0.9000817147004275,0.08470226123271074,
 0.5370770751581028,0.84979982604438,0.2797439277465852,0.08818814701119537,
 0.3628223722441012,0.7097636117940239,0.3880484010012716,0.2833411835786446,
 0.03894598944348582,0.5268035150903432,0.07725077301842853,
 0.6919326633927614,0.501934928361395,0.17093393612294483,0.6671787787438249,
 0.6785731232252714,0.7017955553256072,0.7973823563241249,0.15795217300527553,
 0.7238184906328119,0.8849816497161342,0.18525312471984823,0.9688513836827355,
 0.25289262494234477,0.5924635354782012,0.43375802612281744,
 0.5838645392158669,0.6062510253451476,0.538225625178333,0.3530341676491189,
 0.5229002989158497,0.052469884252514865,0.05693413471118858,
 0.8531724285130878,0.1244211525346085,0.5265967321478794,0.6269189864236884,
 0.305459724286002,0.7687940942801811,0.13495304869277158,0.9496628857328482,
 0.2700566021572981,0.16171427617075784,0.9208209323728904,0.8975654057924045,
 0.7446958865585862,0.8740028526445457,0.8337399730118227,
 0.016089756466904914,0.6293756319502725,0.13036024279023173,
 0.32322702915915846,0.9731314976659795,0.023346711100680984,
 0.6061020717331504,0.3185749768023577,0.7230388330380156,0.3119764406270725,
 0.39808216890690884,0.2247225743515261,0.8782375348974909,0.9277690327374666,
 0.7705213587437242,0.2185835802266599,0.8872386644055752,0.5561823751794773,
 0.9777867749372046,0.28974756756584674,0.38620852875725964,
 0.5974961573208166,0.08570579928932465,0.7087518802064077,
 0.47164118570432656,0.34245115307825247,0.3708390729953398,
 0.7144711586567556,0.6919870028376216,0.41459042564089366,
 0.03860148095376659,0.6296637913010737,0.4306794778258455,0.9774909350453046,
 0.5833693167762385,0.9992873854209816,0.07588414774446128,
 0.12078367794404321,0.734494455428378,0.8644325462100566,0.5751373189061394,
 0.6362343125604015,0.61246068326034,0.2551701234529883,0.7771510990325706,
 0.6241316514988985,0.09836570940473899,0.2277621231886071,0.5885523216962638,
 0.6549981611828051,0.9009123274698153,0.6921809525713769,
 3.887382942467932e-4,0.5404877239492305,0.8394154345508937,
 0.5423789342581895,0.7949706562011434,0.5837528158861486,0.5356592399486554,
 0.4394681032236536,0.10165281115320934,0.7085772256361836,0.4639218404304173,
 0.7449615044517205,0.3062657453218175,0.8782159056405295,0.1573310093264204,
 0.1335978056608529,0.84081953872368,0.5709320549485104,0.31831526930573617,
 0.11285252013865421,0.9602733291338078,0.09611323784848214,
 0.3768690913380981,0.4275581231344173,0.15964624203286404,0.9807818517184321,
 0.2841869170472897,0.6108844849837289,0.5766313887156477,0.9499797161381888,
 0.26119341299252674,0.29349513969521035,0.010286262795686563,
 0.5849716144170853,0.5801139851125943,0.18027805351492,0.583570767950824,
 0.4319851887484212,0.7167117213705365,0.017286147391107676,
 0.39314023900211403,0.7231465257032645,0.7929379813601536,
 0.16681857461446992,0.8480239554172707,0.518586755472032,0.8362715230190441,
 0.9779630042194298,0.8056539446742811,0.21927534109678726,
 0.16180638440098427,0.1208255970969474,0.8659326618554151,0.6942855796493009,
 0.5702900973824977,0.7167348530178621,0.8227019489644732,0.5731338627254621,
 0.8930150080031694,0.8527708323505525,0.7564011122594765,0.31055420311418513,
 0.8136355286658974,0.8039125719064923,0.32728158219706593,0.9348837830733803,
 0.2895904756608151,0.7093515127550195,0.8712226001229209,0.38585629527912146,
 0.1261371122257291,0.727496291019523,0.5097582703205101,0.052933838717116544,
 0.6788956573618629,0.16456514319943094,0.2541387637206539,0.9109600499008917,
 0.37525316110317664,0.7619328568693557,0.17497151313784443,
 0.2617549354859453,0.47593575891806783,0.746193151065824,0.723075423822537,
 0.2893448048545231,0.8061590385182353,0.8247473523531956,0.3548491736999011,
 0.27776476414538265,0.1980802632563905,0.16890113302616383,
 0.9468414124914913,0.28386755789176965,0.37893405576801376,
 0.7454804710114253,0.42247915001959,0.5437484280239587,0.45605252537897933,
 0.6139426489226836,0.6927834181444875,0.8790271757267742,0.4267854112040844,
 0.296904908526168,0.9495082215007861,0.2804173181856229,0.858558230361397,
 0.3929552635571594,0.2032160721989068,0.9623768329516337,0.8682754584279108,
 0.7458797402582263,0.9321437523505087,0.25562373266377236,0.5781659681569662,
 0.03782437976387787,0.30847120878413914,0.5446770483628842,
 0.10452307972855279,0.996869692440405,0.45151191878254804,0.2057220961959043,
 0.6359816972659281,0.7011891633067361,0.9405903030926608,0.590351510221516,
 0.7066959213555164,0.39242842747722273,0.6741555343119772,0.6842906911612289,
 0.8479379376385228,0.7110833567934802,0.32779391689229065,0.8452657597708881,
 0.6976355979779646,0.5515856871784278,0.6460344178538142,0.4779705922948456,
 0.1299537642748837,0.6581546303942034,0.6654047345493719,0.9490958592400905,
 0.029367638813315167,0.5447502395481505,0.7184516833512818,
 0.7956918755891089,0.38236804302061733,0.3486931492967136,
 0.08995998080721412,0.35028375491394714,0.27938666340019447,
 0.5786893746551695,0.7642715883092942,0.807202438920995,0.043489723103373845,
 0.3711494947059477,0.097697490256206,0.0019195853710811672,
 0.5782387023719457,0.30393807090266645,0.8982946636351347,
 0.07765004058486746,0.9896357581958282,0.5147608362163768,0.2792938524861508,
 0.44384883733134006,0.9704383147351651,0.36768124902387633,
 0.9695448290420967,0.18094610724332783,0.24396778794901763,
 0.4845228734357966,0.7699187744989155,0.5297257537782312,0.4262893227913157,
 0.9192369532104142,0.3053558203123473,0.8646105951400384,0.7897040743641345,
 0.16144087576979038,0.4551559330142956,0.06815878595245517,
 0.9629706619887086,0.5922534107375822,0.4074088381263582,0.7647384347159121,
 0.3733882859158191,0.7172263783798551,0.829686144472508,0.25808315635977896,
 0.5428095661467606,0.09223211447888424,0.05652103621523086,
 0.28494080847763814,0.06003050712557845,0.5732860759736348,0.916191857867438,
 0.7471985623536384,0.9339323361664996,0.09264354682094011,0.8344069863421916,
 0.4289769150497992,0.9567534412342468,0.5139092869217111,0.521750605185477,
 0.3150682595903942,0.6407922579364232,0.1645341473348365,0.08525908067427879,
 0.06444613804117638,0.9070992839534937,0.010689769713835395,
 0.9750064325220367,0.870961133639742,0.48305445156172533,0.13638546676450924,
 0.4797603906813518,0.98930847650404,0.7293603784344665,0.43921194086886484,
 0.3410000594473629,0.030017863581802362,0.189289823371526,0.6269671456237087,
 0.9470781690952962,0.744419942178451,0.12705220220144864,0.9595353400116384,
 0.8018162546269709,0.4668414373670965,0.4369536470207678,0.670770188187656,
 0.2820155588125901,0.658672940971972,0.1678896643874811,0.9954453940317864,
 0.44736404580736944,0.7773574581935297,0.17949895794730486,
 0.47352814400939014,0.29731155405972487,0.49390470794341623,
 0.3607414253920299,0.21351189590717135,0.4164209205431407,
 0.10645853757404966,0.010613374126081987,0.5188562532950234,
 0.2814016507627899,0.5492002185009025,0.7942768505411126,0.4029994327237405,
 0.37737493753927676,0.8303731077837386,0.47005545129200477,
 0.9746024268649268,0.47669060278198905,0.634238872884531,0.9919834587457077,
 0.9524250256405865,0.7773929146909401,0.267779396362235,0.5945096790357903,
 0.9243317366086178,0.8390447325743482,0.8650000817861154,
 0.057569194285328806,0.9871580181534735,0.467042358652201,0.686987546222807,
 0.9246232891413636,0.5309016727189988,0.22084058960688235,
 0.46784477184655904,0.7423178379609569,0.11839389976347492,
 0.03709298894271851,0.037328314673165996,0.049956263141742396,
 0.8780074665742926,0.3252480372142723,0.863439602464902,0.35755910042081507,
 0.6514767973893749,0.8543708479777644,0.45388804925473547,
 0.30651112772028355,0.6163820563447115,0.4777436350396975,0.7641794533014948,
 0.9177363145451849,0.6360300010426638,0.9786861696414099,0.39041865604807513,
 0.9415520673478978,0.8101974928742044,0.04649069571666786,
 0.41219414257406983,0.35053747410501446,0.970141917874092,0.2683187163452564,
 0.8893870675803505,0.3516552659434228,0.9183759010905941,0.49882441918060316,
 0.12105047119794232,0.24187962173244815,0.27614012400395704,
 0.45773070162441687,0.5384392239388103,0.004784640701469556,
 0.9058906438860411,0.963659176137533,0.3932051459182291,0.29122684401429255,
 0.44525131446462307,0.8761103628312011,0.7910643875845412,0.6438901371782189,
 0.9654754558852083,0.9734695327980889,0.4963097363804997,0.9082896038629539,
 0.1935516562482762,0.5986780022933503,0.009773398614476014,
 0.27643810973734484,0.9348554383022628,0.6574508014712171,0.8200317479109895,
 0.32102454617605336,0.8808007976778548,0.9008865778328027,
 0.21858735405485463,0.06069207422999068,0.484250363189474,0.985096179604138,
 0.6808331557440704,0.551905275590831,0.16194434446482409,0.2622459564992483,
 0.11623262102799414,0.31060668737888153,0.9037858634437158,
 0.9936800363567446,0.7301880956688612,0.8441783448888954,0.5384049217446535,
 0.22993507874139452,0.03739303723878773,0.17972471360103326,
 0.6499253886263588,0.8247258122714642,0.6288954676225023,0.6840911042449442,
 0.1466442509390501,0.8878613363844342,0.13855162301599822,
 0.38946831848244245,0.43886587543519795,0.08938502855555919,
 0.966916678673261,0.3908599619835774,0.967848608686118,0.5749867590106964,
 0.759198884207189,0.6451690434270495,0.7487437219112213,0.9577321890112971,
 0.7565504685360251,0.2883893117584412,0.3844215135535122,0.6084341673182119,
 0.7509832620574342,0.783456681267193,0.8534923455205998,0.1789735610051889,
 0.8155634438364306,0.19986641349413914,0.007738828906589723,
 0.8300866882092623,0.9843001619326971,0.8234656276580025,0.9543667021916882,
 0.07778264538327728,0.1546930123739403,0.7393892308469074,0.1267068618512297,
 0.5395397790485922,0.5301792688417233,0.2429545837961108,0.15475895088211578,
 0.07844421441133287,0.6644242418757267,0.5364102721882054,
 0.47101644771592943,0.48854814987327655,0.06165568453687442,
 0.6439908790449467,0.40574704982421805,0.12058670001238225,
 0.03032700295085189,0.6157623556388371,0.2731976419146802,0.2435374503451857,
 0.3208977413341876,0.42051676355111256,0.2496966888780755,0.5120665236998938,
 0.30506740539160104,0.5460578633920377,0.6440842299954341,0.7878007738793963,
 0.7913058879141976,0.8881322393535045,0.541329079461713,0.48962893375746896,
 0.22418037853926331,0.7062498675617547,0.09669354839050559,
 0.15290478650957107,0.9478760524703169,0.132531738744864,0.5809656495010878,
 0.9173421235906707,0.286533500123324,0.9325125778077727,0.6612288796825389,
 0.37688336982511483,0.06209237431489889,0.5197376240586319,
 0.7117691348452001,0.6781615854464319,0.3847569873137906,0.6236509352936762,
 0.32252660214015616,0.08527986082355599,0.5138761436928017,
 0.20877574623149664,0.38109754104896665,0.8778071163784729,
 0.4013720120633466,0.18135407984455465,0.3031954500693621,0.5519837982833881,
 0.9676079803671892,0.9616744238626515,0.6868911576019152,0.4224337855769653,
 0.4776376576945668,0.32554246518644847,0.5595958559601241,0.5009012423862815,
 0.024157805621641715,0.8153285878947901,0.08472391707128457,
 0.1055026546967498,0.06179700388517895,0.6995685311255291,
 0.10405862909857044,0.4759863122377712,0.3654461470408069,0.7366612764045519,
 0.8615653312696671,0.9078482815268889,0.6991111298222816,0.07783253624145647,
 0.13555936575345096,0.6117062457037532,0.960974909677398,0.5894091371087802,
 0.1890021179955289,0.17362689577141066,0.12429652865448215,
 0.8550344469581108,0.27918017641098436,0.35054229621430855,
 0.3059878265482846,0.025200906611230645,0.9954609820138297,
 0.3828702437137985,0.4499345140467427,0.9539870584527254,0.7502380186152663,
 0.49123968996073164,0.6694338088629217,0.9142889792826278,0.4622946191027105,
 0.7890831334736281,0.3061580645930587,0.6202760376166633,0.9897756282497306,
 0.34097491033254834,0.4940459986944683,0.39457523568437036,
 0.9125206915570651,0.9060499527320973,0.33221642600854706,
 0.025890579625119536,0.40175844987716214,0.4365514632193581,
 0.08007166664140375,0.36382025125773776,0.32943691851095513,
 0.6599785359304392,0.9510288416027842,0.8511467429220828,0.5737396592752246,
 0.24583168596277394,0.06287648441035254,0.7439127545780353,
 0.8394999226421654,0.21500377764166956,0.3334847905320272,
 0.27591912668929974,0.6705579155766126,0.08625691427381532,
 0.4324758310223856,0.1145371103653966,0.8117927538139362,0.6504895236193713,
 0.30999436096543975,0.6030891546559541,4.6330764326574236e-4,
 0.7283341240424174,0.22457747799167727,0.20678660701217455,
 0.39709135907541593,0.3263383510666581,0.9612040979490801,
 0.47181287380143244,0.2727165431608818,0.6602452173907296,
 0.33103185214458086,0.8015079785875246,0.984030002349138,0.3901305133741636,
 0.6434140346089245,0.9030256928849049,0.47847195240453866,0.6620036611648434,
 0.9517002295215277,0.5280520017459711,0.8171334638881972,0.5283893313312709,
 0.04345651472473855,0.7512260161808183,0.3051743118175253,0.7494341302406342,
 0.11371817688300301,0.5765496684690218,0.469596325569522,0.3204203275048334,
 0.48667058125742946,0.3950569338694443,0.41629358140617123,
 0.7883400886545064,0.7122134353597971,0.6421349826417964,0.8398167487704445,
 0.39504137446409926,0.49333330465151937,0.17877745639626752,
 0.9761821115961767,0.7937967090641866,0.2964631324845146,0.1628973163578824,
 0.04432290554001206,0.021543250918851786,0.49207664490657843,
 0.3787862904049468,0.8478979338268626,0.45935627477842333,0.884022671815447,
 0.07480505112148816,0.37358361207329394,0.6350010225552716,
 0.8531018628638098,0.2850508720021202,0.20680636721717272,0.9541752850260936,
 0.508549618685985,0.843999379278355,0.3781124423012423,0.9931036775648163,
 0.8083216967802865,0.237983814924433,0.7944932206201755,0.38924800579107677,
 0.7118689473106369,0.48708917658800344,0.5482426233479777,0.9131266223086446,
 0.11940599689196973,0.32189010350055525,0.6983452437241727,
 0.7278339423999614,0.10415840925958364,0.8474045677391584,0.9990818712550382,
 0.7955513629355698,0.18717284906257337,0.7773390208758157,
 0.09075350710453045,0.3811498530075076,0.6879132291132863,
 0.31757378132182834,0.4465137792043946,0.563912478072611,0.12559937586423664,
 0.9690866202652989,0.5475177325993114,0.8115729354133121,0.7483526914056674,
 0.6303296629169255,0.8030876007186125,0.27513771637677886,
 0.06588478270937326,0.4632003965230028,0.4272872437462778,0.9984588087068496,
 0.9407590473776712,0.32140218323706715,0.6172926655213125,0.6653290837518371,
 0.7458239419093553,0.6547305814741747,0.7106029478589669,0.16435335091216918,
 0.6086949051172138,0.7190624730935407,0.9564344221111596,0.9034954184628857,
 0.142873327451691,0.5353997852439156,0.523639956269451,0.226412308497682,
 0.14989518897278864,0.520997449323883,0.782122487150572,0.8758316153931233,
 0.6660664898022016,0.5106746803303117,0.0577710728435582,0.905888084799779,
 0.059386364135808,0.24684457628970036,0.6167775609250103,
 0.045665824398730015,0.8982356526959213,0.8795102632419634,
 0.5814973591315341,0.4249317984834645,0.8218128184273387,0.5699383709362545,
 0.2822740926776339,0.3908505441884521,0.021158081772074788,
 0.4499921034657065,0.8742792919460004,0.09746803603782206,0.5132703141249066,
 0.431135292360438,0.16339793985400364,0.32655288366052015,0.1048504325588836,
 0.7244040795462173,0.13872775314228025,0.39387253765495556,0.375186544038018,
 0.6281832536312906,0.6851563358515865,0.5371585775448872,0.43164959472907305,
 0.251640893951365,0.7221369145686202,0.8475209880780494,0.724789664649118,
 0.08306119844801729,0.528306300790018,0.8576340806812685,
 0.034532408635486234,0.2796446259369142,0.20235347974184315,
 0.29512820776041315,0.8653045835381445,0.9706810071964798,
 0.04888867179134604,0.16612550679950266,0.8738244806075651,
 0.5656080565521204,0.2627982021484475,0.25677883497266885,
 0.022196952651600266,0.5895208768490393,0.3811929752076766,
 0.7378442507158627,0.38604527688154666,0.4676992677638401,
 0.39885893628298386,0.27001218208782996,0.6895348042867349,
 0.016409647257354165,0.23636955654026737,0.0840301517201768,
 0.8566842348933992,0.059464923646576184,0.2105438823553465,
 0.08575586898003151,0.16664538573281962,0.4431504125134683,
 0.8337807817003856,0.5563178761890084,0.7029106412274934,0.03553990475849478,
 0.5148828718384119,0.3826139489157797,0.6515095255676683,0.22247682025543047,
 0.5567360267449407,0.33816173599713495,0.29456621896700186,
 0.045697996731355106,0.5222095193206289,0.5582753305220531,
 0.5329874940478255,0.3810355743544198,0.16069552487195604,0.4446510655594731,
 0.87530440971839,0.49588705937872524,0.2810901871856304,0.3558520397233055,
 0.4428599018174598,0.8391081002898844,0.857880331465452,0.3916945796737634,
 0.22285775261084417,0.4786766287003906,0.9001133947674876,
 0.11635815191979337,0.5880649226770936,0.7284810363854576,
 0.09597275589190213,0.26042144986254034,0.05465145459311849,
 0.10429201294361301,0.41216377490618916,0.48295656862130676,
 0.6172373926656683,0.18373220130528656,0.8255016039615378,0.8723362644219654,
 0.023515927798944625,0.47493951454075023,0.20239364407428106,
 0.5451245115228625,0.09557011719369979,0.7108212981707381,0.4173839159105349,
 0.660232878531802,0.2833327682978184,0.8321359831665716,0.2624672733740172,
 0.6674869858010583,0.5622914014934374,0.08898590626448877,
 0.14125394980476969,0.026201452362596966,0.025226584036514343,
 0.5681767466038676,0.07051367925286611,0.8625773145959563,0.8569700981609889,
 0.6971082999722293,0.7475782604424455,0.2851905039957301,0.9049505770830764,
 0.10907015997292147,0.900222923833007,0.4812055717794441,0.37147457052513877,
 0.9987751801689455,0.8435857596773584,0.9388238486062314,0.13227279861464536,
 0.04731349357819892,0.3003040074554135,0.45537838330993274,
 0.3136323802007833,0.4027930649277398,0.7551831220899324,0.5195136875507321,
 0.6410262653975606,0.5380326935700984,0.20068648980597736,
 0.023307318402471355,0.07572894542536024,0.9990191854574935,
 0.09825819774206368,0.8049882571304044,0.05909227650547433,
 0.5047881342337094,0.46396885553966505,0.04196687467018556,
 0.45409034652947144,0.16042032004452933,0.006754727876793765,
 0.8413848418753139,0.7122389166121447,0.7639610749792044,
 0.026479288811428248,0.4209744004147844,0.4247155208502984,
 0.18097314321980074,0.07348011277136368,0.8909272385881335,
 0.9569663866247697,0.41623961361759354,0.28826042515285377,
 0.017073791882657785,0.3471247771822743,0.720614050940908,0.9286239899225657,
 0.3036740082314695,0.002644145762074279,0.8316301217347165,
 0.5082297486290445,0.5927959223387648,0.04525796654575154,
 0.10713644209594397,0.9838483492383456,0.5057639831657297,0.2952036778718369,
 0.09827664042180018,0.5036660178694068,0.8667726381352547,
 0.38268342719218484,0.960986649838097,0.8073355622918594,0.816746995387911,
 0.34082586055802233,0.18981989067899785,0.9039693354510873,
 0.3378999058841031,0.06153473281790134,0.8575520353872069,0.8854310110884119,
 0.10033146706693552,0.669038597605577,0.3531168091770064,0.8706167561842263,
 0.7288009974158354,0.4490994196556586,0.38796119243065186,
 0.20928370287148046,0.23326959984625373,0.4382087084737929,
 0.4837343534207069,0.7806474849073737,0.5767107412230543,0.5026149039277866,
 0.9953519956592468,0.6301445736939529,0.17672992059367765,
 0.44848259912352284,0.6796150941273704,0.606381202797262,0.5707694626528761,
 0.392511615656862,0.5212278207016962,0.2196592674929806,0.16004352740284034,
 0.50669794300597,0.7801318900572879,0.7449728644982178,0.6409569901879442,
 0.07449739867593219,0.8112372921452693,0.10863377740486146,
 0.6010811260187326,0.7799992956536035,0.5067881712910197,0.7650837937516831,
 0.8807999751870057,0.48470900311388343,0.02682963792553572,
 0.9624811958910323,0.29522643126149184,0.17611535386369426,
 0.014038969933744427,0.6361405774715642,0.9967784915034457,
 0.37695272664577084,0.14522183331572902,0.9331451169087603,
 0.6358757097448473,0.8609561975908411,0.6517716026256815,0.27416975437806235,
 0.529870042372422,0.4383618586064283,0.13064165148724638,0.7839655850498927,
 0.4073552954615415,0.9019427926766391,0.6437138529384937,0.8067002024388156,
 0.21157697888367022,0.43246764796440007,0.36699637779248107,
 0.2068539698779278,0.33930921931724034,0.4950642985477417,
 0.34296619378948856,0.4265372437000233,0.20611861119084152,
 0.7685725764504722,0.05100497138125859,0.5474424513267034,0.3151959878660864,
 0.17021109612102214,0.4164016460035085,0.4647251656330813,
 0.19312583125589367,0.683274435785373,0.5509534057554765,0.3582070425249104,
 0.18767469411459248,0.8803598550929058,0.732233851331743,0.8228841309911252,
 0.7394293659770744,0.7876225481619195,0.27207513792719673,
 0.30429943118049974,0.029649389800114,0.9037456883704899,0.5661567055492548,
 0.15700743989553234,0.5370906379786575,0.47135457024610805,
 0.6991169210550326,0.35634377818101215,0.9582360151236468,0.7438606312703508,
 0.4367416883739046,0.32390541335032974,0.031480588820458344,
 0.6146008203451081,0.34828121088136355,0.13442850224340874,
 0.10559080262823406,0.8402179586481653,0.36149718393348595,
 0.5651788857913043,0.11733742415556214,0.047560365346459355,
 0.7729509822240589,0.1312589777802291,0.6928242650422132,0.8048156396723879,
 0.6658662615848304,0.18664920624189107,0.6163298507466151,0.8936147720187968,
 0.01652521397541684,0.2742071336943802,0.7735997868193191,0.0735126990270144,
 0.27841495520623094,0.6295520707837339,0.6368878368954198,0.9380640660431944,
 0.206760414082081,0.3244911483411923,0.5488045232309697,0.07901243838583016,
 0.37597812458747226,0.5247680075151474,0.04136628813720522,
 0.5247914247240368,0.20001291670356935,0.7206881377973023,
 0.13258136548483157,0.6027978615312399,0.7214338938291561,
 0.32895666664230994,0.710331766827437,0.6150423368476816,0.3268284055113906,
 0.9581399753170168,0.7881597347498428,0.9793553253035657,0.2254174340637758,
 0.850166060329013,0.06373652387958484,0.8701744352620855,0.3315651237485008,
 0.474916267615332,0.94873132303716,0.3046467780674018,0.5022092208360835,
 0.5159827426603977,0.6841617889120318,0.7975750630333822,0.820122422792586,
 0.385634334951269,0.050926553947455044,0.04836324436371864,
 0.11940899203954947,0.0990528997247958,0.4655980646521154,0.2708261155722135,
 0.309033644946311,0.9096120602468534,0.23448478428402553,0.16617261086056012,
 0.09555731656960875,0.4440391832363647,0.660808075497536,0.07260230965268899,
 0.5670077667151088,0.6600302852438009,0.11976955993098637,0.8031238691997884,
 0.47925862934891916,0.376266751524196,0.06959603358688027,
 0.06356388011008562,0.5772066465512277,0.48904708456376145,
 0.3808653090415427,0.21270697495717095,0.41259583964753255,0.385983525246561,
 0.7863282858320066,0.9686990992040829,0.1489263098146093,0.6956590144290506,
 0.23926411806220782,0.045211580933737006,0.35722260037341635,
 0.32352631862424075,0.2914917310789039,0.8513041439079214,
 0.22786358605560486,0.16583514206330385,0.5440862659962703,
 0.42569876070945245,0.6070371141596722,0.2125151745525381,0.9831944439641286,
 0.3724759603177562,0.9839085099123734,0.4829248994065114,0.24202064616732566,
 0.2062569726685276,0.8515621753259298,0.38233553350263705,0.7705101304001741,
 0.4415518638768955,0.8167058032000383,0.7805057381444602,0.20368707705953126,
 0.09240206770826664,0.5593302955975306,0.4362344634720692,0.9099474380726822,
 0.5224141173103523,0.33604939340246787,0.7440432116693809,
 0.32140967408431587,0.3373463991936858,0.727652303735704,0.08043766967050237,
 0.15176482909004863,0.09967026845870641,0.20368341514529797,
 0.5129442280190462,0.2510174191093635,0.7621439132009206,0.47987414097074066,
 0.40357014471717545,0.20841925225077818,0.16424552282854576,
 0.4360639432207818,0.47231596471009085,0.19370306403928472,0.329224694070797,
 0.009596972138733117,0.5482489768031095,0.06768312424207712,
 0.9630023843672911,0.05563244687139379,0.7242036529446787,0.4728831282301754,
 0.17500628386045025,0.8383429259678123,0.8124821535760786,0.8249479979472333,
 0.18283191337370241,0.7697068523535101,0.6109780958023188,0.4855776408545872,
 0.026329322710136993,0.920156633319861,0.4976719957704967,0.2154168443329807,
 0.6497276504612387,0.9487714927383477,0.06625159124183222,0.3069837909591212,
 0.31606665323576366,0.9192606683848927,0.5938461629226314,
 0.36791930808563755,0.8425112776012884,0.9290828496381516,
 0.43856477241666814,0.8633767077819545,0.45594945529533504,
 0.7368837791553371,0.9609348491608232,0.7578714651030912,0.378698535500978,
 0.6259739033702824,0.4032880331248141,0.5193908003250256,0.07912829372385044,
 0.47234624398671887,0.6796837476273002,0.3534043265841451,0.9486080696941098,
 0.8793393239710943,0.28169518329115517,0.29570111043143366,
 0.6046767728656667,0.8164794096002063,0.1418782691946161,0.6472890790602628,
 0.8756121220968217,0.7332262449444179,0.81753409941187,0.06098560497578065,
 0.7275698316421787,0.2801314667148386,0.2125145153135496,0.5541249073133214,
 0.03261690808939455,0.8314828023037324,0.710053836760429,0.4176998009925216,
 0.5472021805876295,0.20622365554317834,0.7071149837997031,0.7254850760035718,
 0.46513480424175657,0.8860851913365599,0.0691469939461159,0.839070187738068,
 0.6170782687333987,0.3546059930584782,0.7818886768471519,0.0747999647642984,
 0.3473914701335441,0.4750994272115905,0.8094373930788105,0.48313780436350384,
 0.026540162425476854,0.20992518179372877,0.32133543368597195,
 0.27600604942860474,0.6978258182989976,0.28650974369221605,
 0.2043839129081052,0.11182679330677991,0.655530257039624,0.6135400157256575,
 0.01619839864736805,0.10716061099310292,0.9855877794811846,
 0.08483738279582032,0.5626270113486479,0.3420749174811917,0.555536452534504,
 0.7410178606877502,0.3957808425584344,0.9225354307700055,0.8186669916002491,
 0.8048830859251735,0.22584826551503445,0.9328668375270079,0.5028110712060679,
 0.6953955013880827,0.20524061650533432,0.1935371238150987,0.993892920094258,
 0.23002352944917148,0.3956501144960515,0.059502320844431966,
 0.29448381937757073,0.837749307006291,0.8827945118549168,0.882861505182463,
 0.304404482855516,0.1349031482473274,0.24961439635000127,0.9677474045059463,
 0.6730370439241266,0.034810697421089776,0.1041928859252842,
 0.9697823669365765,0.1494145403928303,0.6174996391080702,0.5265136500098848,
 0.16794804530252283,0.776292548976704,0.8436721027962303,0.4935706403521669,
 0.19067900772588464,0.9245520771462395,0.9964371169522516,0.3185380453739568,
 0.9451115734060411,0.8470681894804496,0.7464178066725778,0.8886898337012957,
 0.6485219959931403,0.5105924462888882,0.7208210478738619,0.5128742211824011,
 0.9982716447265066,0.6361303195884246,0.6941792177943819,0.14602640195704586,
 0.9316831780466001,0.2815674694271192,0.11287678475104146,
 0.011489339505932561,0.3641595793446134,0.7336407204954137,
 0.5694151684568848,0.46232814734078265,0.11722747612229567,
 0.6345565262674124,0.8648996770742388,0.2868081012896537,0.7686523373960344,
 0.4104854059135882,0.06303301783187687,0.8465315816149626,
 0.49801829405135656,0.2836439699590281,0.26545822357056015,0.477319831163842,
 0.15095749263345393,0.9867277377507514,0.05616203695047883,
 0.5442339839628176,0.9671973742408487,0.4332335005837846,0.1633687292392395,
 0.723410225598299,0.6415294965127624,0.2732080897517244,0.43834750742113404,
 0.5932318186921148,0.32526775684773046,0.605225013609937,0.6540268130640184,
 0.37941461844893465,0.23561701449810424,0.47499881647071585,
 0.5293010565148251,0.09629844987611824,0.46174036337706426,
 0.038237562232998124,0.5495546850367423,0.6014073950335884,
 0.29109387152658284,0.3275286083272917,0.14475819663795364,
 0.7821832753048518,0.8883946421752714,0.4420877603686053,0.4578423469968017,
 0.020510531108345886,0.7098407989456484,0.9248798757667388,
 0.7241058584767364,0.9395502702525804,0.7641931227432428,0.19414235268746816,
 0.5218696724894416,0.7024100054732577,0.511676764506908,0.769655934290516,
 0.3287308915228615,0.663202010713645,0.4999568200844272,0.16348605381391168,
 0.7535314569104368,0.4419567553499567,0.7995971040882514,0.42514966350955674,
 0.72500903098386,0.2749329472915292,0.7638756484459905,0.10703913496103179,
 0.2606250961086469,0.522900225707549,0.5955182073750465,0.688667124562891,
 0.13446537131051506,0.370602897893318,0.16969260759787919,0.5807615269336946,
 0.10597318680033563,0.2297500863460149,0.7630564657024081,
 0.021552568850855325,0.2860045200299216,0.6409398333481184,
 0.08367438385833714,0.0494009525165513,0.3181171148890727,
 0.29937543705574976,0.280406168794037,0.8923855768959793,0.23441016873240306,
 0.7501116956535742,0.8294355048796471,0.1108009313083207,0.4915286353818532,
 0.9886979757929178,0.24814688750944935,0.3384621403531829,0.8723756999737369,
 0.8470046473967214,0.419633499642462,0.8280761595202519,0.4081014020866136,
 0.6366767420736223,0.22322591176459072,0.9897593616930618,0.912656063699072,
 0.6051304162279664,0.6209171926037561,0.7587808877364136,0.20358716816865052,
 0.9395378859372118,0.679567351523661,0.7505621276318664,0.49715947929018234,
 0.8223105029777646,0.7721222115761506,0.33258687451328095,0.5023214831233951,
 0.47992712962007855,0.5178341573129602,0.7423110045177514,0.3078498628996723,
 0.23163106973967618,0.47483856902917043,0.9867338299333492,
 0.4154305803308028,0.8047763913936427,0.9006904695967626,0.7691925280660473,
 0.3971430676705556,0.13935422577797751,0.4632485483090324,
 0.21609431765778508,0.4477848471575179,0.6928056798013232,0.18358546737309,
 0.07169292781931513,0.8978187256251655,0.5608082982509017,
 0.09400909959013093,0.3246460342426992,0.9674194513665881,0.3051494528084271,
 0.4228312202766592,0.3608727813685232,0.29135585028193256,0.7307300005507364,
 0.3531907260569517,0.023398843316017603,0.8171190757288757,
 0.5935867060006798,0.0785087070283792,0.8674467529136147,0.5224003644437719,
 0.5063347783397577,0.011031553421976148,0.1096398508923474,
 0.34836115895417874,0.1589592128012831,0.7626006715957037,0.8012020305875015,
 0.7942693567242506,0.5403956765603803,0.9606460279827748,0.4946847607125664,
 0.1325962755563055,0.034217038291540325,0.4999495145005537,
 0.8070421116873234,0.7940605248650499,0.715346652774535,0.5852974359357541,
 0.6148430939514647,0.14981179596637584,0.5185386433493829,0.6310984881679519,
 0.11281428379789993,0.16771335399278542,0.7346939403637733,
 0.4903029538051352,0.7538825100140077,0.4557668885471916,0.3895006979504909,
 0.1220162506205239,0.9445296336831241,0.9404761444561057,0.1271986898862486,
 0.33360513148332516,0.44876617091608373,0.19553841924082183,
 0.7349236158774619,0.9086004061142092,0.15074373024315035,
 0.07691681644094961,0.9935038667225764,0.3145830481076074,0.707076662580066,
 0.19945610946770698,0.2708127573790088,0.43397524066637483,
 0.3911184918961925,0.36645400479405155,0.6190352667917843,0.5091975312738306,
 0.7638063026381353,0.8889944439801792,0.7313423377476691,
 3.7481594703392407e-4,0.2731791615808553,0.5648966747854794,
 0.2861077258309336,0.33699233146202145,0.9522170035597122,
 0.33827500011393796,0.9331427117440781,0.06782845788708458,
 0.3255988308389062,0.519294289694957,0.7852398123149792,0.7173315595735574,
 0.82418590959544,0.5592240135790307,0.8087958898124207,0.8876371931597237,
 0.09854413096456072,0.18906137242743004,0.285876195561516,0.7471972849967812,
 0.3390461491636062,0.8686038483326363,0.5239206918644246,0.40487351282938544,
 0.3021440664014001,0.8634069447834581,0.09484731727626328,0.2075705970814038,
 0.19873368792302848,0.3984309710852493,0.6171920092486244,
 0.07587701478525244,0.5499736704657755,0.1321006757332145,0.7005840153261932,
 0.7924651603501456,0.7116727194272179,0.7762070528518047,0.753582226566663,
 0.2744849654937166,0.5512917819305599,0.7479040015355932,0.24307845241417592,
 0.4766453757192325,0.02076040688137859,0.9059275461396041,0.9127404865735054,
 0.8718108864041015,0.5604058360193407,0.9905447204211097,0.44022102364960514,
 0.264111527331628,0.0049724628654534975,0.9221024432263988,
 0.8818421423877437,0.1051200546134281,0.21862031456447661,
 0.28788892013741296,0.548553824633546,0.35161462461160686,
 0.22511894543893307,0.6501048928485561,0.7102557685594474,0.1850568795907367,
 0.8765007188831716,0.4668682422034509,0.24639647462993086,
 0.07811417999559356,0.7113373999436846,0.8478961244579545,0.2698516848210277,
 0.6871396433974295,0.7630523803827395,0.9984763878310187,0.3997867494081412,
 0.13066707705842007,0.2643963322951658,0.4376836336204164,
 0.24059367088814243,0.7149752972052345,0.6170571822677009,0.629311776235842,
 0.1567873744646412,0.21824712057149254,0.8943628302090483,0.3115205366681849,
 0.2848319787729877,0.8465254971321103,0.9769383299796781,0.6173133241443791,
 0.694268444916759,0.10660943387135091,0.5111604498328137,0.8390351881318997,
 0.2784621279715348,0.3406398966288263,0.4328691810009757,0.08067331008268042,
 0.46720863759256215,0.9949866402627333,0.29013417419405385,
 0.9789951820128333,0.5176013328721222,0.7529368956610185,0.36446912477421545,
 0.6722594383717408,0.07216905128166062,0.8471278451257768,0.2059856325516617,
 0.5585819169404179,0.8750974637199783,0.9899500057099319,0.7823337678748186,
 0.1712280669582935,0.2831874564419421,0.06880969247277535,
 0.28578590489401356,0.32814827296135407,0.8495865767359493,
 0.9403170835612522,0.21575701648182188,0.9932261611741438,0.4992994923635071,
 0.04621997298170211,0.11161708351500199,0.7875417570723221,
 0.5432536077113173,0.7134320667010396,0.6375950901285635,0.11682837796187595,
 0.897874264086147,0.10790002199441306,0.36642401885769527,0.9721890837520422,
 0.9342323225161278,0.7740099384907845,0.5289153648928062,0.8186928753867984,
 0.11087858613938795,0.3123869270043358,0.4990959802079844,0.5927494928008137,
 0.0778629455721197,0.8364619197699654,0.32381653458091286,0.1399711023961061,
 0.9920123549661841,0.747779085447611,0.22645874186560333,0.7755807044019095,
 0.6713493800882839,0.7849765885205168,0.9734040926047489,0.04127082111321101,
 0.9784420525227782,0.7370947996008627,0.3542911862520096,0.6468234738540471,
 0.14594198291933025,0.1650541873143987,0.11820272205305893,
 0.18033299731761043,0.8732348764111065,0.43534736909934757,
 0.2809813544432451,0.27240227002776696,0.7413480243777001,0.3435351889298165,
 0.5737977250459899,0.6605388600463912,0.3748465744977587,0.408632180814849,
 0.8893237086859058,0.8315515605632192,0.14858893628094538,0.3170796553888332,
 0.8925018691360481,0.19284712319781427,0.48450545320222216,
 0.027595065026960297,0.1665481915503001,0.5085899548418747,
 0.5005723912001919,0.05422308990218305,0.8470817268330637,
 0.13337386619426084,0.07422043264295519,0.16231718027308029,
 0.04208872515503159,0.7127295575524101,0.003987765406972543,
 0.9087602263545494,0.4947176608147881,0.5290812754745842,0.1184619681001825,
 0.9264547271930053,0.5075267795760442,0.018376451841690677,
 0.9717356334644758,0.9711362869759615,0.817927258146307,0.2451731569428749,
 0.20494243579999072,0.5524166244308633,0.5697549585352422,
 0.028264866386954912,0.3338547950261359,0.7515588688633298,
 0.6617310638677942,0.9290093704274333,0.19371325298309916,0.2129317788633427,
 0.3231111947978048,0.17193856721989176,0.46731974704593204,
 0.8335180699926971,0.90103882750863,0.8613444603652475,0.8046748272074042,
 0.6920950041383266,0.7007778562359407,0.024997116789734375,
 0.8131153749186141,0.8436640539451036,0.7595483882354519,0.6571000134375882,
 0.856523720710335,0.3767634275537697,0.1402691346933438,0.11715930690521914,
 0.2722847492711775,0.8068881986956904,0.9487091163245517,
 0.0024500959446590587,0.026439817110661012,0.005934135356007952,
 0.8860738424741298,0.08515071209659786,0.8101562503302574,0.7437892703403515,
 0.6174001473269568,0.3571200058229427,0.6460870162131179,0.5557316752844599,
 0.9574899891562795,0.33425351508688794,0.5414144994393896,
 0.03802909202056182,0.7088611182518684,0.08120066134286497,
 0.34657070722894134,0.05664997571420827,0.12773938032060173,
 0.8364226245800165,0.3533648686188705,0.22996596187980622,
 0.21897460235634303,0.5618699498966975,0.7601333702524313,
 0.19318669489049922,0.04551951174984481,0.43998040802739324,
 0.3788881764322394,0.8157282381281786,0.9732248564502858,0.6250983014908422,
 0.17511185081260872,0.8864487373369224,0.8762649480686845,0.1568797698047023,
 0.5000395367944992,0.2525479773957643,0.7491128279596557,0.6001222915040526,
 0.5958819207281773,0.3353177773755446,0.07747675486881911,0.2403149791079856,
 0.2482655475148262,0.3384543953177791,0.6245591875734059,0.21161976522006465,
 0.1466706128114902,0.8820343158555412,0.21664528649940884,0.5078396512711979,
 0.7460381106325221,0.4944761864745921,0.3441818794206585,0.5144851358400928,
 0.8307096134212711,0.4041859736597305,0.4968989268878292,0.5234583228105509,
 0.2552267658728099,0.1636039815806296,0.3441000088471279,0.25785115520624324,
 0.5743543702988266,0.7874479754863063,0.0989435433127197,0.9434835166208565,
 0.2331604546915962,0.8820971408599353,0.8120337146685533,0.2166142211828097,
 0.9737096414667427,0.06063223665083006,0.5556053429794364,
 1.517267794096666e-4,0.5599465322509509,0.1334664526374637,
 0.9189947040100362,0.9106682323704623,0.6367902255034295,0.9489323332092952,
 0.925156183164141,0.6179189456959789,0.18117105240661957,0.23060233247989448,
 0.15143080149185817,0.0611197100846097,0.5658660940196467,0.4389870355949016,
 0.8302099728502353,0.5292275569456086,0.049085125811869634,
 0.48097972744180517,0.676034660935944,0.6617660217872225,0.7018134406448215,
 0.6205883282970814,0.32949063320059213,0.31948240512155857,
 0.5290567734963699,0.2447488484730398,0.02877182134449552,0.6115226209386956,
 0.8581618320258211,0.27203954582693324,0.6513668604288252,0.3886906111798899,
 0.38828677583665205,0.42626838928843025,0.6098467534106651,
 0.6359004045638403,0.6000058206403351,0.8462966380072713,0.1537515170066166,
 0.7229683993268488,0.8241106038136015,0.42789228501865584,0.836578616728772,
 0.5776991073482596,0.8304907685936652,0.3063166167399626,0.9392928538565674,
 0.8684579945658667,0.9881461995349072,0.08656214171881194,0.5161914905643208,
 0.9663416355431019,0.21857856561100708,0.18266492723066252,
 0.06105369499887647,0.22496423922233333,0.5851916783225605,
 0.5250857404420534,0.4658339413376176,0.33311380339448116,
 0.20630464640427704,0.8327206761706897,0.27055784483795553,
 0.5164674077572815,0.17633303512286347,0.878373525406091,0.11009523746654282,
 0.7480395119014086,0.23843298026059145,0.26058974947856695,
 0.9608435181106572,0.334704787831011,0.3045433763943366,0.5814514489408065,
 0.3047830127787893,0.9999734369850053,0.23323918254846676,
 0.47423367675277517,0.6266246064277626,0.28134420447457864,
 0.6551218807659107,0.5310569921176326,0.728396501727727,0.8267628715141487,
 0.07852751060582208,0.43359997050400945,0.8456502216183122,
 0.8027901300199456,0.9918278091870172,0.055157921797631015,
 0.9477647731492291,0.3272350742481438,0.6169824093450549,0.928535900823189,
 0.8466738740326053,0.3956185546399543,0.4933763838787635,0.9207685040582545,
 0.27212892598049676,0.5077043303226515,0.9469109534935205,
 0.12049656016413945,0.015199256772276293,0.7032379848240453,
 0.4593200809032303,0.9222171477986996,0.12122526660146771,0.8669905146949869,
 0.41802612799272776,0.11304377532677201,0.9781921276847434,
 0.048258805133501226,0.44513640968053825,0.6603207297417918,
 0.10993350526371104,0.8877996708501678,0.2972506201730034,0.691888193784092,
 0.6820507273788758,0.7976193941997556,0.7587405310911715,0.8203391660571477,
 0.19641259987855098,0.5690481234771658,0.14471258781024343,
 0.09008084009959194,0.6076518966900375,0.7558536956842965,0.6891232952130764,
 0.873625462419125,0.6049216103709196,0.07839944391568676,0.935552103130039,
 0.9678674529494389,0.05266153789207495,0.4121387683729747,
 0.24059807106967512,0.4819289694758402,0.20795547392030367,
 0.5138870832055707,0.3688868336620408,0.18259009647871194,0.4266392360481386,
 0.6365391089274044,0.6482895505794062,0.8393478122337644,0.7490186903511213,
 0.8527073130870126,0.2897426061959918,0.2592357129380881,0.9473364219022689,
 0.97380826263484,0.013261922321904485,0.5548555804386313,0.5058197806143762,
 0.7627631412868591,0.001268694189233921,0.7868956358593492,
 0.6674913856630758,0.8969888642531385,0.036260610878680266,
 0.27216070066491493,0.024056908095247853,0.07367888040608772,
 0.21603581565164598,0.0410024590241741,0.1279353745521522,
 0.49607183442423375,0.7380458103518642,0.10001622619327799,
 0.18772385937878056,0.1712602551091521,0.18442672516295333,
 0.9691507936239788,0.8542642765371948,0.46952627708115424,
 0.42738908556286714,0.15226143416965232,0.7286392269208113,
 0.04100128955799964,0.0454125484118274,0.5034085111810795,0.4514046065370316,
 0.7616211578987812,0.035533014281449726,0.1734909621812616,
 0.5297240180496366,0.4775694251223839,0.8778756260146613,0.15315339748629375,
 0.433719628219587,0.8232900247118726,0.7426359155231116,0.9343996306678283,
 0.34438870187615955,0.14801210197270653,0.7280260495407612,
 0.13365174996551532,0.5931063093089206,0.48126235975186726,
 0.9085492385342396,0.8223032879879587,0.7587894734861833,0.6841497227991351,
 0.41278497137676307,0.13086099698745146,0.42555319779666634,
 0.28849668947034357,0.029062776752627584,0.6752145077798409,
 0.6810838014637745,0.740762647227102,0.15467232551915178,0.41819262864878803,
 0.540694731042306,0.12380733546742761,0.7016364248021483,0.36857784611168465,
 0.4174755336882402,0.0782435324906654,0.17374775437183132,
 0.38417971113941907,0.5854917794985265,0.8755966220330926,0.2504908194014378,
 0.5525990810057462,0.7050935483398554,0.7242537210255379,0.12957838597177607,
 0.9578490143959867,0.30745079970993194,0.4933820523637008,0.9896470657031201,
 0.33673364382049886,0.6465599179744035,0.8352669688582953,
 0.08849348503431231,0.739365106008901,0.5328505490860137,0.44255649846452083,
 0.5402228527801151,0.6300558011985902,0.7574533725576211,0.44197612854377166,
 0.07222685602782453,0.31416065397714676,0.9897609399205606,
 0.7676846664192583,0.08097729252007158,0.2115725228802524,0.8171449042514974,
 0.0878967620713571,0.27545331405503015,0.7509310706380031,0.7987930510945439,
 0.25517209127322293,0.13664167338544608,0.9913981231732174,
 0.8544668423858508,0.4933547210906921,0.9157606958574483,0.5933471470030258,
 0.015229564854874145,0.014209228477288782,0.473028513116712,
 0.1584619227632994,0.610986225731043,0.21397956688228903,0.36155975758373493,
 0.6930920846627664,0.862746328369512,0.09948825524820049,0.4046198722133647,
 0.5167460320743733,0.011614421116081353,0.5679013530246938,
 0.1364503377832642,0.138117642709093,0.10545326489647622,0.6842220038990381,
 0.8510446648485485,0.5989750031786081,0.3359699669616836,0.7066878678804315,
 0.40788812620966686,0.20685881527633365,0.6985816160990885,
 0.5601212299233024,1.9864610072195532e-4,0.7897348699734057,0.36722118429435,
 0.9134002319959599,0.7107809024224071,0.25572201967798736,0.4869965129355919,
 0.7390713316578209,0.4718248375038365,0.5416699507777122,0.3596442128512448,
 0.5543996757666975,0.48579378358048,0.630465712390589,0.9959619938097233,
 0.21753689305768253,0.7079781765376943,0.2674248821623748,0.6410851206899271,
 0.41485509708436785,0.609029793195216,0.41965900535073053,0.7450328564569478,
 0.9139027162486835,0.42519816766554086,0.7102863736062647,0.1393666294072593,
 0.2137287824985461,0.5692524620575369,0.7016782191842104,0.1804679291130794,
 0.37855086268086424,0.08511651974985279,0.7827589516531888,
 0.8004927653415076,0.5494806392548901,0.31854284632281216,0.8597058479814089,
 0.6642965166985245,0.42770924417387757,0.8075970883685542,0.7027093132082285,
 0.06237597113221982,0.17332095444629025,0.9555731615001571,0.260563045113014,
 0.36048188858235075,0.20294578269032337,0.5661774746526682,
 0.32907485972700556,0.8312034894312568,0.9609860137624524,0.6494131812344716,
 0.4811951697060203,0.9847017209671989,0.5875252615817499,0.5912625780632146,
 0.13102482345090338,0.9086992458307512,0.4105291608476114,0.6917189952524022,
 0.8273863293195189,0.7450756610053328,0.9579747066778046,0.7829540731675807,
 0.008936187497612269,0.9428910908337705,0.25638341398527054,
 0.9054339103835922,0.7875289375622923,0.9551576862487222,0.5549741015556664,
 0.5055656592166241,0.3477384440182708,0.47839619240398723,
 0.47879939183764586,0.9993927127785502,0.19687568262802102,
 0.8181921689130035,0.5450098912725666,0.9303129107053704,0.6683765575563765,
 0.2620195969568734,0.49071988157745894,0.5068842347598139,0.5432357275752576,
 0.9245056368585772,0.2183783811637624,0.7333741406104609,0.14806592493342907,
 0.6253766650318969,0.3729842980214575,0.597117103855207,0.17388388035611746,
 0.040632628597835674,0.047801763235549966,0.5171222378990306,
 0.441127329330101,0.47435447826550503,0.5414417542180245,0.3786565068309167,
 0.2972877986915976,0.07080969201911258,0.7873833513912019,
 0.49641179242274036,0.5512665704065588,0.4677628892664377,0.4468304443531983,
 0.6162399617508321,0.6799187624613655,0.24635553177170144,0.8720155257744517,
 0.8666739355152279,0.01181926768946917,0.5482454182420023,0.3449649283506153,
 0.6811585862908754,0.8544674288491978,0.9106560080997768,0.1456247177403931,
 0.5914973676153387,0.555675017534658,0.3855910888087417,0.7679194318422403,
 0.8096231861655924,0.748894133009599,0.4466406720015622,0.4115851118784728,
 0.87842935910734,0.44855377467074453,0.17664010263297547,0.0920446090588598,
 0.2559102223481924,0.1361200125101969,0.5880729415474891,0.5572157906192451,
 0.752835565820529,0.9610942102319866,0.11947247715803522,0.9990800774391003,
 0.5267630729327601,0.4364522701337078,0.7120664558546155,0.3187354261849077,
 0.6859549989355954,0.24305732926658286,0.8615465913484337,0.9947735279630008,
 0.4740339974283083,0.8608713784839602,0.3480566824228837,0.582183763240853,
 0.07064847024366716,0.6247197872568795,0.7915626743588903,0.8187004083309176,
 0.6988344629314326,0.2338947086758646,0.551383432667707,0.6684435798821674,
 0.4073050570909087,0.6286989789908674,0.5907720110253135,0.7018223480588086,
 0.27102754155846853,0.02884350300790417,0.9656477944060121,
 0.5467356434000066,0.3822058522167656,0.3727534407794566,0.9950525798606303,
 0.6091000804877573,0.9933733401390932,0.6336733780707151,0.9838997801244711,
 0.765201606118536,0.23324197807051228,0.998040330265822,0.9269973406493773,
 0.08003551229457806,0.11345299385913643,0.48679259472637737,
 0.06744957861658873,0.6196214313432791,0.7024383325753286,0.4882999764486179,
 0.654031942753095,0.904868538164932,0.8745612051104628,0.8519384367969856,
 0.6916202412775649,0.49678327088878116,0.11214267847494974,
 0.2455280848842355,0.8232567830388459,0.09174563235728617,
 0.11582488800314616,0.8548902864781323,0.512756607696801,0.21746847050936724,
 0.483559789394696,0.29106248632857845,0.7482611993713347,0.882473515361543,
 0.11446301390244129,0.29413757640857985,0.3385365821005507,
 0.04460860640955633,0.6385875804788501,0.45244883769423705,
 0.6879159692209023,0.16468974276640436,0.11878929382236436,
 0.9345124709997497,0.4411949250147601,0.030826660151240137,0.71566773647287,
 0.6676570529230027,0.15315363167447693,0.1931102757462978,0.2747338535303345,
 0.6474619654501741,0.014004203535501136,0.7570158731574081,
 0.6898841573468819,0.5630213215397694,0.4974107619947258,0.4858622309761036,
 0.104889841522808,0.1672222187058603,0.38954990785442867,0.48058482036456573,
 0.3539567921512575,0.35419228041417383,0.6064003062325773,
 0.34804609975471223,0.37013389961809917,0.4514860218081741,
 0.49092249178256653,0.2875075966851077,0.8389031042670038,
 0.32917018141173315,0.28280000016592544,0.5347285882608047,
 0.5790125285605329,0.6573298056478009,0.42172185159373254,0.5245277663146173,
 0.03440727780143893,0.6880074918755397,0.928925864113473,0.19929320271059647,
 0.585513529794618,0.4051616327802495,0.301190083506949,0.5384720664049555,
 0.7819177150806906,0.8439567319818363,0.056081994969104376,
 0.43967225415268596,0.6505886779352699,0.6860936159861941,0.558343989471174,
 0.7212642878572096,0.540549441596715,0.5294978759668644,0.1957454809918311,
 0.3012204500988498,0.5738548911420265,0.14297919923311875,0.5494196418356304,
 0.4006537660178012,0.33680738906312124,0.2574139871275031,0.3089645134186224,
 0.1577203705933603,0.611326064716685,0.25377202038945035,0.02073428914622677,
 0.049431058753532486,0.8205526067150718,0.800717940645785,
 0.04329706761788732,0.5122305175183509,0.7992890777117887,
 0.42615288166834964,0.3339786757290911,0.9426268812690688,0.2346405154665846,
 0.5691274822629994,0.7207460587365857,0.31937937898493707,0.8504576271755234,
 0.8939019388335921,0.9613121287607069,0.6925397315216363,0.06885723154055279,
 0.7279902883944237,0.8271438054704023,0.4652622531702656,0.2148603083688062,
 0.8639936265018954,0.33519904204366224,0.2911335995172085,0.9560067994252665,
 0.8103847111373483,0.7675264750532171,0.9953989825564173,0.9114577603898746,
 0.779488121981991,0.05156333540620106,0.5447367543823907,0.7850124407138397,
 0.753696111578956,0.19487053678939426,0.6271045668384178,0.8504175520004414,
 0.24395217968869742,0.015298599610866215,0.30729102267291586,
 0.05394434524342173,0.4209662457323401,0.5802707261835575,0.1694582055292988,
 0.7887858698460242,0.10451643967473001,0.6501987623723323,0.5440428112031367,
 0.43946663990248336,0.1496937142140795,0.9306078563499823,
 0.48826778982168584,0.4246531344835849,0.31892536027828344,
 0.8820955290975414,0.04870248843960889,0.7626603312115801,
 0.17869683181169294,0.26189523089964295,0.15662070535519457,
 0.06539952267497207,0.6939772926694648,0.012348857358766119,
 0.4054874695334726,0.9373367569804246,0.1982463188354402,0.8977966687221097,
 0.4524005907243448,0.9204071993068984,0.24955224883858362,0.758553148624567,
 0.637364157038691,0.363904042691765,0.8688169951428011,0.056388682872215434,
 0.3068801632060012,0.8321907536683271,0.6057841558722796,0.5219092261554448,
 0.5449530629726816,0.3453193802268094,0.8623864749443317,
 0.026463110271617962,0.08004688987821207,0.7874397794418394,
 0.43867342320450264,0.7106576564480098,0.5251776280873293,0.9049128957451171,
 0.5872844209043042,0.9550774181736834,0.48512494987402666,0.935174196734965,
 0.4559280097405779,0.8126544969452584,0.5292873423935571,
 0.005169943455675674,0.394826455307265,0.9243275892818654,
 0.44792893122464317,0.6102632633075584,0.765937367699025,0.0929810125149434,
 0.6040077262715893,0.7366332642060489,0.4542009706198863,0.8086350499205465,
 0.090604852558957,0.3668078252721486,0.0359126435428988,0.19242013689748194,
 0.16057208836811576,0.6045599215482893,0.6886209772240028,0.8899151545384497,
 0.33947156646736176,0.37344844586898873,0.12779663720175405,
 0.1476976969603958,0.269371245541685,0.21131148762340946,0.7332792798899515,
 0.4725130145744124,0.9764313712507656,0.7873725569995982,0.7409682885769766,
 0.8193057397229011,0.2337360478471454,0.28413535061285755,0.7299981914166561,
 0.009855998857968729,0.5783568485956523,0.5998731094395876,0.80208789846396,
 0.39807722721608674,0.16466191458850532,0.8395458302597371,
 0.9967873674996697,0.6544379965849488,0.2255114865633504,0.803328566482627,
 0.45812290289021607,0.9892366852816339,0.46777328845529464,
 0.32294247080651173,0.9491139501022708,0.6673708326934151,0.5765985648835225,
 0.5970245363228135,0.21939515084204364,0.7638548542002445,
 0.26551304743848736,0.483942469311062,0.7178941380507589,0.5566014209086101,
 0.9192701214820461,0.022460988508795077,0.5018893966455736,
 0.8289706602912942,0.7953605922069006,0.07368752541028223,
 0.24138898017954247,0.3374866654077846,0.52347583296947,0.1345720933709169,
 0.04471293555028555,0.03077412666514212,0.09282971303709031,
 0.015365754715127466,0.3382500956795155,0.6237439587692771,
 0.6458423336819291,0.8108580813635221,0.5117601829085584,
 0.025328589060972573,0.6248967163971734,0.6795461580732334,
 0.6699726367027498,0.9464862640099571,0.29824982379218024,0.6821765571286511,
 0.9874013522628061,0.6191657883003339,0.123675287823521,0.4856514189915577,
 0.7543054507904994,0.012598093573441882,0.8676361343764302,
 0.012168649230480999,0.7497091042150834,0.8949174625753713,
 0.06760303808650314,0.23763316357838737,0.8167119812470918,
 0.8716535562516344,0.538571251015121,0.541691486633503,0.050530266358341125,
 0.8640833274674369,0.6246506990639691,0.906096170467331,0.2916053353372534,
 0.5227272195621557,0.020510044006543215,0.34826303462187236,
 0.17691025214108347,0.9911793554960295,0.8094640481992597,0.4235758162498693,
 0.9060620078486408,0.5606032416604061,0.7088455615755644,0.31448597170051695,
 0.797049999478087,0.34229163134338225,0.7177440225204389,0.6803154825493726,
 0.152404862287276,0.2486007302410448,0.05470457163736997,0.17160685635062456,
 0.9628914794805347,0.9941103209354514,0.1332145023214606,0.5702973068421333,
 0.2348611151133826,0.655907257770637,0.7623920016461951,0.881481946471127,
 0.3211156116954743,0.9708161116314143,0.5760053553084598,0.7095904473357167,
 0.5687025773306327,0.5841759678618262,0.9359767093351142,0.24287848499819875,
 0.6464911092021465,0.3113235603613749,0.32627084136924533,
 0.36281647284850027,0.24544820833646508,0.811645526688164,
 0.047071815991956645,0.8577519641083622,0.8575878175538255,
 0.29565748462279673,0.8077445104465805,0.1214695961283987,0.7349885488565161,
 0.8143721803377397,0.05725826496411013,0.07888259262883679,
 0.6620082901534291,0.41993925500565454,0.8926387587822251,0.3767723293132663,
 0.4396668494287409,0.8457750541534313,0.5751022014249503,0.9949220518116153,
 0.5608422652763368,0.7902612169572129,0.9315019668604462,0.06957657700448305,
 0.7795136390109952,0.6917980777915668,0.5972761150653976,0.47405996249110127,
 0.8647093288583563,0.45767448009627654,0.8738683930407958,
 0.014030414802678748,0.718983201691094,0.5147424003676455,
 0.13677232337479805,0.3266171478550036,0.7360672030482248,0.6262039713376162,
 0.3936747387132912,0.5625636149818296,0.8362964909349667,0.9822405983965522,
 0.7876032474880411,0.9819804924010597,0.8332186471892883,0.06270193536053248,
 0.40274596463391554,0.652626314675097,0.9181845091084124,0.7586878743026839,
 0.6719794650553292,0.4125484772699628,0.8948922765697866,0.19232320862072028,
 0.10495053333366222,0.5917502904278767,0.048594328085200034,
 0.949304576943819,0.4773673599948661,0.3797695417838075,0.016762881064069313,
 0.8736464553723245,0.14068144816373807,0.1710682987303599,
 0.36165028754511275,0.13886649022641107,0.3554521881044126,
 0.16976824237249188,0.42237915559744066,0.7755233563984049,0.936865980427803,
 0.4297374519902085,0.3388228386739578,0.13350329507806147,0.7925148303663778,
 0.835093239943737,0.2682170298862767,0.1307259177674709,0.5670091090932533,
 0.2304993637221967,0.9859235038132332,0.19692648512683153,0.3881036457168384,
 0.6954897131199436,0.9386998397788455,0.7874933356461485,0.926615030869149,
 0.9044669372922366,0.8573633891096557,0.6371272974594042,0.06940346264833885,
 0.28179552752021597,0.6507225304004581,0.0797854703465789,
 0.18055921565971678,0.3159047959154301,0.1521217704287403,0.8164840199126224,
 0.28786560417432994,0.2237129781448023,0.21144045945934842,
 0.21775513365240218,0.12485638756570794,0.9905899897866672,
 0.4864283660566592,0.4209593928357829,0.3268804007120626,0.18159933963231145,
 0.2918169353000545,0.9182036849925322,0.40488707208814745,
 0.04103384838147417,0.29985727459430844,0.7876439606556451,
 0.9300436941700411,0.2341630020523645,0.3639910671547796,0.9219995476393893,
 0.3596238313446396,0.6071743862734887,0.7788195143509634,0.5584357465832371,
 0.3585519966734775,0.7092135740069544,0.18647197807773364,0.8732317539566357,
 0.8406036804384253,0.2620729093572103,0.25814968502965074,
 0.22859898532996614,0.13216281951383302,0.4150595953155636,
 0.054051374751785675,0.21100961148344866,0.69132765645687,
 0.023361705402797917,0.2890967152417919,0.06795865137103185,
 0.5829639548090916,0.2857142567087513,0.012283543792080631,
 0.30275809049070423,0.4265539891181539,0.15079336666080145,
 0.45531122480854624,0.6630628047320541,0.3957933854522122,0.8597249331285398,
 0.11096652840365728,0.746335345795193,0.570807012683794,0.9651033785806883,
 0.08583344715861718,0.8000944555246622,0.8661536462629591,
 0.048978349286082246,0.007279992644937283,0.8711453749565806,
 0.5231927301742274,0.9989717484327225,0.9116102569934139,0.7288361553143733,
 0.8985905275933395,0.48022992446047996,0.40760548471187974,
 0.7839282457188742,0.8459943788589306,0.1587070595684651,
 0.050578415399998744,0.4730476544265886,0.09285526222480067,
 0.7874382795588744,0.6333460185577022,0.9161658823221035,0.6347340109140102,
 0.4394220995193471,0.23441362490730133,0.8645496010277137,0.8486555317598151,
 0.7826692171561972,0.6364641953126484,0.794258916407833,0.5048752130357772,
 0.7114564770948895,0.9208300571161236,0.6210357845961725,0.4366926275438594,
 0.7758650887913712,0.8459848796930607,0.1811343871455121,0.9816289811125045,
 0.7646406184705558,0.6182152855976296,0.09927928899960559,0.780002478205477,
 0.6539277571488339,0.2703488805357398,0.537833029189621,0.7440752432655378,
 0.8436695688877106,0.012164933508516218,0.045799601600276185,
 0.0731880312745592,0.8087438848960592,0.33742005924258367,0.8171595293192656,
 0.4804261330522288,0.2862953596530802,0.4334241998055296,0.05502017448644769,
 0.3586764238556567,0.8325481287684304,0.4880912164640696,0.8989626963953852,
 0.17100504191653687,0.34635780402897354,0.4504615614746672,
 0.3605148638643947,0.39670771840956087,0.9292733938859645,0.8668738928545969,
 0.549723727783638,0.7790376576466463,0.6543186886852063,0.496670470223581,
 0.5332124972659765,0.5325362218778923,0.3183303453816426,0.34155737781183937,
 0.031143835850228507,0.7792443746869577,0.5702113443090957,
 0.40206334796957277,0.7065899595357988,0.32377075731463256,
 0.1666839788074127,0.39417780232975297,0.04774614327195492,
 0.9167674978180919,0.4028634642972785,0.20630332557850406,
 0.03824325774194948,0.14927362688588064,0.4558313729949568,
 0.30272353297195687,0.23402471486217025,0.21567068617413399,
 0.1461458877010563,0.14636753810522984,0.24769677787666233,
 0.5717988385677325,0.3816036480031112,0.31508709168215443,0.3048524817317336,
 0.1726067367757661,0.894305743523344,0.26309999401864526,0.9539984071888313,
 0.2141319545492555,0.5305742494307649,0.09765908195268569,0.8594678214218876,
 0.3024436202404561,0.22069206583102519,0.730004054027578,0.29074545157535736,
 0.8071765846823284,0.4436620547984167,0.28086849409447345,0.5349735952538749,
 0.37777687667642845,0.34504408418180543,0.8258404720324886,
 0.4823575645697521,0.32520244965842204,0.4511110994414721,0.6806090382797584,
 0.506540117246744,0.23083026065402257,0.8263496386244373,0.25521438862604706,
 0.720307632198138,0.21811623019647364,0.40038149021986047,
 0.41703055584787696,0.16680205670992299,0.9072118430034928,
 0.6545115323903731,0.8444510951009753,0.9895423061872607,0.2404948557220331,
 0.6580099288503404,0.17330467350229528,0.9444427307016752,0.7275888246873501,
 0.9443436174910672,0.40452818356128395,0.2750584412503938,0.0917497277746484,
 0.5148584854677885,0.8370812446243572,0.96322861572336,0.4507045765809361,
 0.6756996068977835,0.9539485985150596,0.02388119012084089,0.5736172241761622,
 0.2876255978559109,0.6970707974589992,0.25608571867495356,0.0183051691767685,
 0.8318767674669854,0.10208811047037158,0.6100086223847839,0.3827389567582712,
 0.7130865622174652,0.6479582543249895,0.49849708820343785,0.0890148003808019,
 0.3406706568032113,0.09715849883955818,0.6573427516770693,0.7374163698675378,
 0.8226935566385909,0.9655093081608727,0.4846028153051166,0.5579254795057453,
 0.8926785502135433,0.29219540620082907,0.4963819298327271,
 0.14476893352809406,0.10909851385392333,0.06671693733104866,
 0.9790240580057042,0.421391657556044,0.378456029046924,0.8724228671445473,
 0.6907908568474745,0.2170024748642726,0.7893337051323768,0.9810369646524865,
 0.6858574347779013,0.1298480128856777,0.8600090702110127,0.9610271184204144,
 0.23788812836811457,0.10519063095304937,0.3982763203156251,
 0.9746939116277529,0.16932333180059966,0.6903192813413154,
 0.07362338563857829,0.5034459848517142,0.3317665611049455,
 0.015794053781077455,0.02139764861834559,0.6259044605658678,
 0.1984445244980403,0.23853233602584967,0.44847332233127546,
 0.9036927140481364,0.283606202080927,0.5299138079902973,0.1987311147913967,
 0.24504537420544237,0.3688459063562316,0.8712248807308796,0.5224618398506451,
 0.5196531787763403,0.39307581377910106,0.3120940783956859,
 0.011924630097356115,0.4973401967885984,0.9641770582170157,
 0.29290007369577586,0.5094788036839455,0.5866535912095616,0.6557919193608425,
 0.7170230648302989,0.8546685875479294,0.4769440924959365,0.03856407956924346,
 0.668063493871673,0.6589647047479884,0.450831690865377,0.5403843032382427,
 0.32219679700731085,0.8758238395920032,0.45152258037425475,
 0.9531528581934597,0.8957436156988086,0.3418887223731809,0.832240743891389,
 0.6012278228002357,0.10554106404343933,0.8717061224282632,
 0.49301803396639554,0.719084149880598,0.7807172398143094,0.17383654538398963,
 0.7437712616452886,0.32285839121403803,0.31322579537147754,
 0.20165857932533005,0.0977753441960284,0.5532458487700539,
 0.09881070856493046,0.2883003150033424,0.6537509786916886,0.6667467031465254,
 0.3396761512262183,0.5295265116934763,0.46648339814569095,0.6406567332764812,
 0.8077870887523289,0.7729680838162841,0.09928544435775344,
 0.33786511844043676,0.2988511559290046,0.43415025089525994,
 0.4396534026940443,0.24853782518788858,0.9074529612808322,0.5729829771895072,
 0.1582659439387386,0.3281328284867534,0.38066806682468945,0.9237293838100712,
 0.4722922800409899,0.8578643424769433,0.26761662414328413,0.2889978270931568,
 0.34446041531199256,0.43745626303042817,0.09932776810552923,
 0.4206025027955268,0.04186006479968141,0.18040739172904108,
 0.13239279020079364,0.021469200621326512,0.9991793872819614,
 0.9994446511205726,0.9638521979854456,0.7165750901464841,0.9192063639457447,
 0.8976402569694416,0.5014014826668919,0.3527895710354585,0.2579954168904589,
 0.36154993506035804,0.38781028855154487,0.4755969841574986,
 0.41613869276840787,0.9708332935913557,0.8346849444019946,
 0.43116379258726434,0.914939239494434,0.5690829936902977,0.5959449327967505,
 0.12599458036285893,0.3283444087383378,0.8972446049582581,0.9167268162549919,
 0.5413855061186295,0.43462689567526025,0.5928331109497678,0.9520102522017051,
 0.7594322505124859,0.5077308726925068,0.876110187536115,0.7744519353244055,
 0.30925733582697146,0.5039076406075272,0.3635618788337496,0.9134163427337477,
 0.35510668302219406,0.2623588952307553,0.3604448068590993,0.4910050170101681,
 0.5211781530525291,0.41184282787872994,0.2583221565083137,0.1667727100341314,
 0.17314398623773497,0.7010004953448046,0.554623473628368,0.6983364770114582,
 0.5165001346220139,0.4373366331807602,0.27514137723414667,0.318198551589126,
 0.3218017967601987,0.2285675289874951,0.6410669448869019,0.7778246276234422,
 0.6534731256223347,0.70748589099269,0.9828788476621162,0.4917142467758201,
 0.6246524915321248,0.8616619529664027,0.4020652129773442,0.7812991787557096,
 0.2541624559893163,0.8501520186466794,0.2819995340846616,0.7696477443341547,
 0.6416606470774404,0.02829513152600238,0.3602822526255939,0.9548577794171436,
 0.07934459931216853,0.7817364541217122,0.1352516453853171,0.3003148755982511,
 0.6127829962665806,0.5027101237499766,0.8146699494443439,0.3863131883122648,
 0.9685218514844512,0.38884682732331854,0.18478741267232635,
 0.8443997765954716,0.9955002301163207,0.3643869192368072,0.503397814373723,
 0.7797104602921514,0.49286783821639546,0.1411247724102357,0.6826968966636181,
 0.7843177668701632,0.9448296637094193,0.03879783050026597,0.731889267053462,
 0.40162023602068775,0.4854362424054216,0.16279393284035903,
 0.8440715820510651,0.2169393910045868,0.48505118228636424,
 0.28954986617356415,0.901567567426435,0.8655709783892581,0.14459713772264338,
 0.5927020435244456,0.9675503040991087,0.63487383475752,0.06291948537413317,
 0.9720760889200278,0.4434027231352169,0.5576617746364954,0.578505076845031,
 0.06552752607420154,0.038055088221921496,0.8810561010882032,
 0.9680145594507913,0.27527155289350513,0.7127992095134166,0.9225492316969638,
 0.6826366778882653,0.04477166113007358,0.015456551787504602,
 0.5239856186485315,0.1180231763734293,0.0408105050995301,0.6172750619810502,
 0.2619223414807951,0.19542390321143888,0.17698655932674434,
 0.30501889911040336,0.3268676651090867,0.82527682031677,0.1840301328823104,
 0.6172513823147526,0.048673861643569016,0.06337033517356017,
 0.12351534767349548,0.663527740388293,0.2149176971647645,0.4543408975296781,
 0.8764125369466017,0.5403415078535735,0.6098480174235074,0.22751294958153268,
 0.31501331376288255,0.1245285966155758,0.22662970778980673,
 0.2606225771797921,0.6762246663182677,0.2347288725578257,0.8547521457165801,
 0.2921398966663721,0.1652765079879498,0.4426185290058201,0.3983185540657974,
 0.22954059815973382,0.5402897366897474,0.7253822543375436,0.8422433413293026,
 0.19695320487284929,0.422808157276125,0.6520873297737236,0.11422149547311955,
 0.40709565912885104,0.8029053107304034,0.30127498240740336,
 0.6101152280113148,0.5792264545937985,0.35404327445319794,0.9507271004828749,
 0.2200205171923202,0.3835106970314708,0.09065401436724796,
 0.43512677173421344,0.360584853250668,0.10427954219617097,0.7079585021553436,
 0.3218488251850631,0.8300874837864969,0.03289015008831786,
 0.06289788812828301,0.72265605918568,0.13095512729287695,0.2138600662906649,
 0.288902479258194,0.6154349774856678,0.9140356897800679,0.521230481726053,
 0.13426608228376335,0.8318057215093372,0.31493618401092516,
 0.34510192350356883,0.664147977245755,0.32746609886606115,0.8797210837507508,
 0.02431381324011428,0.48372266290940613,0.3733290233681088,
 0.7789281905425998,0.6517531299910848,0.18512439930471136,0.1826622732393297,
 0.5212171559448546,0.15248846470468902,0.5940584171101052,0.3585221988827578,
 0.48660621268898274,0.02410838588433839,0.4686252775254317,
 0.14549390050142885,0.1473736004607924,0.5597806603888198,
 0.21766172577207943,0.5071228853128135,0.4247520026634266,0.2431926034604558,
 0.8397511994194554,0.8498650914857763,0.46533809359338174,0.3294466363366493,
 0.9972652988574432,0.028320702265822506,0.35658076213999235,
 0.5760819504029714,0.896560080585746,0.16810844946451053,0.48741350427352337,
 0.3092540643436197,0.20617612343488445,0.4172415784279132,0.1739585776461905,
 0.40013026944294405,7.997312894472142e-4,0.8627427984816557,
 0.4395485938631778,0.436467740276124,0.5061947938384782,0.3870296445507708,
 0.4049064739687598,0.3790895575422677,0.5258081915542446,0.6222018777616491,
 0.6000348007346538,0.9689435931100208,0.26255762633250757,0.7298975003538595,
 0.2582574683350478,0.5856070021073078,0.8243689706971358,0.6480902789215885,
 0.6272050285909675,0.6842355327893982,0.08241961003635201,0.5437890561079941,
 0.9095984965444158,0.8596668804623346,0.9718518323315203,0.980319730656005,
 0.8705910581452474,0.8765909074080304,0.5507742351263052,0.3040950448400056,
 0.8532377478860551,0.3421009159690316,0.3632917855819595,0.2918839131997206,
 0.8825135265103916,0.3181300335008145,0.051466340522970944,
 0.41817583643141165,0.9670135250527806,0.9469215497810867,0.8723995813407567,
 0.7829502975044239,0.3975008110354644,0.5345567447688832,0.6399638529254069,
 0.8634964832036127,0.6031705846272044,0.08958396058262452,0.1703591961919788,
 0.13192497785304003,0.3531703414643421,0.021826466433427893,
 0.4783289587378018,0.29514837982938424,0.3081939813361587,0.5223684833681378,
 0.27755157947086995,0.3624001709376309,0.892310383021502,0.40722983358073506,
 0.8505893918838632,0.34966429835223833,0.013374678086851999,
 0.022526106490296405,0.2565409291481513,0.9208400194740688,
 0.5567971184656043,0.47218756911851656,0.6940103285032372,0.7233731319208652,
 0.515702033766319,0.38779950864512236,0.2686917877742271,0.6113938105193225,
 0.1634678839950109,0.46741254815026345,0.8795226481244025,
 0.10486063710330729,0.5127505414034029,0.9132381264130836,
 0.18377166992449867,0.9685203209519787,0.6531879501149013,0.7517646479711809,
 0.4066186803100681,0.6592083842590708,0.9954943487200404,0.6755052626538979,
 0.41177764116480964,0.15203513313049333,0.17386698792855337,
 0.9990391440276746,0.4120580927528673,0.6242206235851442,0.28142717406831363,
 0.5177198402436116,0.9898291044028663,0.1380598222805337,0.7301365023711945,
 0.8439289814962657,0.36393701056099415,0.36727606748364805,
 0.6006898941580644,0.26954342511969465,0.05400287341991983,
 0.4154083060288656,0.22563121880183834,0.925754907464417,0.4174895514643766,
 0.04755290390910283,0.9332628360181094,0.45738933840796503,
 0.032852081360132956,0.4716262339491226,0.4358846177076663,
 0.05277424369019734,0.6130030563419885,0.23323434275566313,
 0.5628967675798453,0.4979100110803436,0.5137977139687234,0.9416467948997992,
 0.012434895416745362,0.22399165955898892,0.976649614952334,
 0.6747726579565173,0.1091664075146952,0.2902229821840354,0.4209527592138932,
 0.5916643327714441,0.05870612165619915,0.03256843875775495,
 0.9755069591187675,0.8095254909942847,0.9537402406140774,0.28873400048389264,
 0.05304842156004397,0.3678600759580891,0.016529330521899088,
 0.09780158607476719,0.906423970572347,0.455245140707232,0.2485633774605207,
 0.14061524511563062,0.7998098281589722,0.21188605568242036,
 0.9966857004201635,0.23450152543805447,0.17554064112020662,
 0.6160287286697017,0.05325630197013109,0.37348500402946827,
 0.5752542149684713,0.546959723727383,0.9723501070379493,0.8245695497173333,
 0.782235994082583,0.35525913812897736,0.9597295784631814,0.08901076478180514,
 0.1120810836924292,0.11101683979767352,0.6798479248143967,0.9240152147129811,
 0.5181993912386864,0.7595816958438455,0.17554346311775626,0.878008251566139,
 0.01351312825779405,0.43989960683832185,0.61457299492991,0.8287278684991399,
 0.13606102963762978,0.3324514427099984,0.6860699904274726,0.8800586868330855,
 0.8462659287880197,0.7646704640793625,0.08613665361729705,0.7643085310755012,
 0.2706801037972717,0.4443428824940242,0.5383786867978162,
 0.033957529249209806,0.17947850847251345,0.1382927988440933,
 0.25327844486106754,0.13996924989039472,0.9287075069798422,
 0.3284341732556283,0.06559111077838742,0.5244676235470219,
 0.05452365700690531,0.1540248135633221,0.6792254627330994,0.295699049881492,
 0.5064879192854501,0.9872396593181609,0.09756707044789614,0.6335690937807648,
 0.7644060917282107,0.6288731415838429,0.8751059125887088,0.43306273478912893,
 0.870846410156817,0.7108432186507316,0.5301555297520546,0.517113036998544,
 0.5271232918423955,0.8563463977869591,0.5164723270313019,0.6556309170347818,
 0.5388773973240767,0.5103429801489138,0.9561118564372972,0.8078157741110689,
 0.928762682546909,0.38252895998133396,0.15958875992949384,
 0.07617750232977483,0.27721223492432734,0.6553543841918811,
 0.5288687490446736,0.11469611869373475,0.2763899617792861,0.6793233127513522,
 0.003474160846618557,0.382557087189186,0.8226472628526614,
 0.38218612053127465,0.472900049079957,0.2683985477981493,0.4143282942176477,
 0.12396643536320273,0.874431447899272,0.2584701611715422,0.04246102571674415,
 0.5646665708633748,0.9059950743144303,0.8007705587522479,0.09041209417321039,
 0.3063519056305668,0.381009964287917,0.3774563840461528,0.4931376418527843,
 0.517936552758011,0.920889408828703,0.02733145143797555,0.9294579255414258,
 0.009924988915027022,0.4931249329498798,0.6113496386356978,
 0.6273084367164266,0.8162752138613353,0.869837222261671,0.4747363305281065,
 0.15994315632372458,0.8813068794715956,0.9131059995014699,0.7253403227151245,
 0.4060981175767593,0.27952658257388663,0.47037163628713285,
 0.8377649817773847,0.9229038411695624,0.22942020937774776,
 0.08501298787028133,0.4157102383668698,0.5892000565133876,0.6247329533344232,
 0.5444912201631277,0.9971417724040789,0.146434648511236,0.19112828408154103,
 0.6541612442752702,0.9994011841140039,0.24256703753955877,0.7666577144703725,
 0.5986340884430407,0.35502468904245543,0.4852930779423421,
 0.44659485516637665,0.036687833294631544,0.3154895313406011,
 0.42041674167886556,0.4880229557616427,0.38273694207656284,
 0.9682795240202788,0.8418059655929337,0.32944870489659417,
 0.19154053658257175,0.11688813596018066,0.8002105028644496,
 0.5711994251697708,0.9105297314028782,0.8061425037226381,0.25320279046045646,
 0.6666952808818307,0.7705377995549922,0.5030437103146466,0.9641745803971383,
 0.03909988108669982,0.3973252372033298,0.5316976743452405,
 0.10894548994324516,0.6218236041456328,0.5181690782168576,0.267008993254102,
 0.9183250036049732,0.19357195483810608,0.39767724825343,0.45121313593568446,
 0.4330550965563922,0.7653567061682616,0.4713118321129648,0.18265750903259914,
 0.4860861603576092,0.7525236463670184,0.77997871338653,0.0879483257478505,
 0.8611589850019423,0.13199996037101758,0.157181387508728,0.5612473011406525,
 0.13551387290790518,0.07395126455179135,0.13556740189485716,
 0.2596230834156559,0.45114376030812253,0.5485095950592614,0.8163583901900723,
 0.7668730103825182,0.4998925789057733,0.3619556057935628,0.26772290666271603,
 0.061392972007197844,0.6198787650784111,0.11408290643842944,
 0.41435601796349764,0.10895447858337715,0.5176216727324439,
 0.7528639277991526,0.6152769314135886,0.36859957720588143,
 0.27923641291582435,0.32525001369812556,0.12344085512468161,
 0.018708901158713487,0.5101231323183909,0.8911360351768676,
 0.6301850809350064,0.5905174620241922,0.21648070957086596,
 0.06115855575265039,0.5859192978707471,0.45455128070813844,
 0.20389059528127962,0.7847498209702275,0.9469751743839938,0.7623260724836625,
 0.2560018033219451,0.5275798276185522,0.0770374725774301,0.46051357280364846,
 0.10145969888790785,0.9427295121387879,0.8731017437298394,0.8640031233010013,
 0.8497039758898094,0.32955576735979863,0.685992611734275,0.10592469372255431,
 0.8339907066586377,0.248166342043742,0.18611460404372837,0.41466494137114396,
 0.12120172571580523,0.2669153771316798,0.40813054993247166,
 0.6874077384603035,0.5304547264255182,0.29554030580142465,
 0.26933325270402897,0.4838731695420041,0.45879182645306604,
 0.9331987634223027,0.7884441989963868,0.7548091326074077,0.7241388049353825,
 0.11226342157659497,0.8212582253928524,0.6749341741354837,0.9372754471392549,
 0.6578459452120194,0.8290456508293632,0.6300211937285041,0.08986151262836684,
 0.7096977499921564,0.037856453268782064,0.10777618457471516,
 0.7375434468124795,0.047665968376876644,0.3033229977457982,
 0.28808484309723437,0.6638894637472869,0.6609492139318345,
 0.15044206510333635,0.7551992980461456,0.6773768092116286,0.9885888186107,
 0.2249970599470742,0.0770507442130618,0.4346401111194197,0.09447731379946034,
 0.39944162380780823,0.18429701907498175,0.5577463923222654,
 0.17900793163681694,0.38156757073654013,0.5020439977538994,
 0.8159176601163337,0.3602979430682144,0.9729718002948896,0.6988251699089817,
 0.9869451462575123,0.4958169043706019,0.7371636579282826,0.7541541218751687,
 0.39193622303085873,0.703809479549981,0.780441492587679,0.032874006755062024,
 0.7345704590738991,0.7013769060924687,0.7761554002061398,0.40019847921520135,
 0.03851131636007843,0.581860257476114,0.12854923950130503,0.3419371348925463,
 0.6510466039013114,0.22259282985422635,0.289477589518372,0.9059535578473772,
 0.7243339157953136,0.9193860972576222,0.8461955765072027,0.852465965670777,
 0.49801020316568945,0.0224614335444413,0.5211065823718865,0.3918323966359115,
 0.5468475983560804,0.1397213156944036,0.7158399849278755,0.45777275500244985,
 0.28660375713143205,0.18415987871145745,0.7303952492706266,
 0.2694886029457628,0.0491419156029711,0.30977436345914033,0.8838515928197479,
 0.920963010624341,0.46635650096945447,0.8690156092760835,0.5958674351144269,
 0.8739036717037308,0.0811358652804538,0.6257920361754474,0.10191957834676935,
 0.2980754387612917,0.5480065359654254,0.7650416812025345,0.09239235611345376,
 0.47560650154632045,0.6683254995194295,0.7781971431076679,
 0.019369348688002552,0.6949670250089786,0.8033246673963381,
 0.8312657479598109,0.8650547435383379,0.5608252997037246,0.9895769584226866,
 0.6119890774566411,0.1579760286991161,0.6154013087590208,0.6386966226359504,
 0.6305311182911084,0.9874397958149556,0.27894534179674413,0.5585852446598891,
 0.2651096099972746,0.3779216734311246,0.7838241972683249,0.524820246848393,
 0.345042774276332,0.7527084232927304,0.9136143695079285,0.21869872282991742,
 0.4328269016673518,0.5863416066523002,0.812849644578312,0.2631297428767757,
 0.14176161331278858,0.33021226332838194,0.8821207165354982,
 0.7243732227252422,0.5625837459798334,0.5938672668145214,0.33064241040231246,
 0.7666951586664966,0.8444743521708674,0.07444534557316684,0.6845691148736013,
 0.9744929838044457,0.0017978755331995444,0.49135475213396707,
 0.7792667197770498,0.9679649924618876,0.4099820023540698,0.548126412582688,
 0.7275170308735717,0.3021874557951635,0.8516744198222643,0.20632378115774053,
 0.7225558988310977,0.39826184797553865,0.13121869951443887,
 0.30930701222282864,0.6114387894439987,0.8456854314830184,0.9087323038986074,
 0.5901835007963877,0.35879613995412696,0.6612417202780281,0.6417370477249114,
 0.5491911775603718,0.4089245738018962,0.055127378783809466,
 0.3309657843172512,0.12115238926622418,0.8071130959892803,0.4883060287549912,
 0.563680890014976,0.44785017398327,0.4649018868758915,0.25035005228790985,
 0.8026962798148882,0.1577819301439053,0.7266181682614679,0.7182891549234045,
 0.9318247933894487,0.9216657077173687,0.632026770717989,0.5694747500023093,
 0.17954016062967937,0.5676839539741863,0.5564397303649492,
 0.16521561630426584,0.43349513119994554,0.016353273653322953,
 0.3462459859223701,0.6977188691238516,0.5336638973025247,0.22429478301181183,
 0.850229831261006,0.9809445265519682,0.49918444978927634,0.7576600449425642,
 0.17092679108911224,0.36788591509731683,0.08427555801289488,
 0.7139802886790755,0.9028078021502342,0.5652670219384955,0.6960714127954183,
 0.604748261155085,0.8042538779274401,0.30937261075737443,0.0883891787161315,
 0.9277941375613479,0.27481660064718283,0.06780807133010536,
 0.8248189190363283,0.5818687352873017,0.8708317874451543,0.24883706888442092,
 0.43651440296703514,0.03229879773409694,0.7829841998870357,
 0.7921859165516781,0.849435040643789,0.17599183654839157,0.09209398782484657,
 0.9393729334063725,0.22686367084300219,0.9149737487678097,0.6306677296250708,
 0.5323046769525329,0.6285468698399529,0.7746464565790647,0.5638477991484525,
 0.520835033055896,0.868763427466224,0.8117945214641477,0.2739621869407335,
 0.6633891664490548,0.026109841308702864,0.5905093971178746,
 0.7834000337978094,0.3774701933583604,0.5275224791873813,0.9244047893053372,
 0.34974642168068215,0.7088980937751556,0.6460216628605682,
 0.42640804096140794,0.944502196106803,0.14701478778141086,0.2000419444921132,
 0.22303033529608285,0.9486426692338789,0.11795800398997891,
 0.49452927752246234,0.5499903692992292,0.5813229508630418,0.7450504346247335,
 0.5322177617197283,0.5849799119154432,0.5169821227552273,0.5329469881649271,
 0.7145701118682204,0.574537720711179,0.728499734432417,0.03739517916072965,
 0.9778339348618706,0.03128638909883874,0.42399298920231865,
 0.6280349440067425,0.13392924065779843,0.986746072458241,0.44955189573075405,
 0.19430288016790065,0.08066056919124676,0.473461061751111,0.1594081575685018,
 0.3391851431997547,0.22147700668540138,0.6714582210605692,0.6266442765307155,
 0.8413957881048086,0.6590021569879932,0.7950443287625377,0.13583807634502398,
 0.25662277976527315,0.8437773149089791,0.9454937899813175,0.5333193014120579,
 0.7667553885831929,0.9147229393481708,0.6889230080895824,0.6133701336793901,
 0.02909339554933399,0.775607893402679,0.6753166944995388,0.6554668934101133,
 0.5542909409547218,0.9596972148453544,0.35096171410974975,0.7341707353373198,
 0.09985978734426926,0.9320765441287974,0.8287432292000949,0.2881386237393583,
 0.8434930505571252,0.004251852895803876,0.7208880538008706,
 0.010103280298156547,0.8599735835353107,0.4298166143678974,
 0.8804249292578884,0.41387077270700934,0.6095322468619179,
 0.14811160126357525,0.09132680738714727,0.6212915199812139,
 0.5049284819154772,0.32161897213788915,0.9075234637821243,0.5320852610228649,
 0.8893164076792536,0.8087988247011225,0.15336399890454178,0.66624604201487,
 0.2997407579883793,0.6083241249611004,0.8490181279889916,0.49815575628512,
 0.8549855787035842,0.931464274946597,0.8850177508605203,0.5243354688537951,
 0.04619748682375702,0.9635160764245381,0.8602161428914028,
 0.13491737755484523,0.800590479096885,0.9892945348629176,0.8325984775199997,
 0.3091822432495537,0.09150477186477857,0.8508680903700951,
 0.08635326140013855,0.04028965427058884,0.30485442841417787,
 0.1989170607330929,0.49490935590754115,0.2746433300231933,0.627264976479511,
 0.1959894654306178,0.39828289939904793,0.3936345434897972,0.3110939660112738,
 0.7780443989678325,0.5157989787691297,0.10666321906452647,0.8494179764460893,
 0.17006413638596962,0.0518283531490884,0.559037922750882,0.2544675154718158,
 0.20960272172623395,0.7615378574464928,0.5907073024643574,
 0.11718746689209258,0.29855783159675464,0.9391784969504942,
 0.08699559009159186,0.9318482039324099,0.8107698114036185,
 0.12306372565023005,0.0660055252272258,0.6531671956140829,
 0.060294832049483826,0.12859148492921169,0.6819047135303204,
 0.9199467748843905,0.13630296936856257,0.7372180655283223,0.4949382639937975,
 0.42794901346884473,0.12459658222062708,0.2967916768782052,
 0.40852924963246073,0.9733369861590835,0.8767828863613407,0.1927202971303863,
 0.602801631395511,0.18108812341704073,0.17066349067535502,0.5434020744950113,
 0.7365050429131468,0.05586641022088301,0.10160581586653017,
 0.5714442306465011,0.9498330287590264,0.6014671242911209,0.8208641750006272,
 0.1865549315038626,0.4976313177315297,0.401781549204417,0.7823041325324531,
 0.35848298180734095,0.9218588220422342,0.6149036423403499,0.4954103792980875,
 0.7365063285049205,0.25445977859865676,0.8513488166301709,0.9709867545295136,
 0.26195502694288064,0.011226340141365698,0.7030960282630411,
 0.9722723197405706,0.4431328928777599,0.019108898584154432,
 0.9626422846612498,0.98703192645324,0.30070016504011565,0.5181125919884065,
 0.2970008822009339,0.45599003641938496,0.2592587542826661,0.9416199590681468,
 0.6632270152705323,0.9275092040753029,0.14998999710019267,0.5069097252341893,
 0.6510127193591011,0.3861965484490604,0.16389696211462057,0.9074005049264985,
 0.3018425722246211,0.6171073505221913,0.15229415256584222,0.2859238356110668,
 0.15615851661343294,0.6257211210638185,0.19508060330416355,
 0.27550105917200474,0.741023960154459,0.8643751766521984,0.17319222208895857,
 0.7129800122044199,0.4558750058940453,0.6104443801972647,0.20843420050278594,
 0.16098630120739665,0.7783079695434034,0.43693591286491884,
 0.8850968855673075,0.14668652377467428,0.471036518040995,0.17395702655829826,
 0.7508896185498188,0.6581377426339576,0.8162406508570483,0.9783640587542787,
 0.2032626987956797,0.43367630539053614,0.9953028713559505,0.2639462737403896,
 0.7908898757541797,0.7118560849311171,0.6750402350149302,0.275747078710454,
 0.8956342652983099,0.7105502613947033,0.9489866822446209,0.24979901049563424,
 0.4387673059107525,0.9047654630835734,0.5627521717716069,0.03942336405718827,
 0.6396226576121267,0.47447516738822215,0.431676918263738,0.316279000710732,
 0.696124305396135,0.03136246246448959,0.5764691906819523,0.409109644345936,
 0.21183384030815322,0.5767048849731216,0.0030666197243299775,
 0.5928489758261057,0.7759936434526996,0.4032935138829251,0.26602001861005253,
 0.6198563146193878,0.6418571446670137,0.25723705305316225,0.4353696934099527,
 0.1643442959621232,0.840132359015448,0.5946554903291427,0.8862514905645817,
 0.8637957384916576,0.8003353369148138,0.9720332012869075,0.09351267248056816,
 0.7814390390987451,0.42423027414251735,0.9717066623174986,
 0.05723818006275527,0.6614455867658174,0.5622121584062291,0.4098545924683721,
 0.6969817909466994,0.6734921178025668,0.08635813113629043,0.8434711236555353,
 0.24085217627529154,0.6425681724432398,0.9532295905913694,
 0.13333502262284025,0.25940874279301906,0.22704151537920325,
 0.9143904791596744,0.6312798301974273,0.4695768776539174,0.5005355988418257,
 0.8641130875578393,0.5292761615225866,0.7412398139950583,0.6801137139384025,
 0.38143672566454156,0.8226023736001687,0.7481016874011985,0.7366853752816578,
 0.75775649718624,0.8065816133573886,0.09399882150179373,0.7948809078889856,
 0.5773506935876672,0.48049075594076285,0.7429625078018218,
 0.16871804341143104,0.19460128899262996,0.9942223026767287,
 0.4893163733613515,0.7147931790570426,0.6466660439627171,0.21675929338867594,
 0.8068975529510879,0.17252571825008045,0.06794741536205096,
 0.46125273241773246,0.27462258619391444,0.2585347957446411,
 0.43039794474999626,0.4663466638276994,0.33951654486771254,
 0.7317988055911511,0.9571833761480081,0.8596185167095329,0.02085134314550119,
 0.9624065828908127,0.34497762765746876,0.15772702242994252,
 0.11102759633665449,0.5861813468600208,0.028489186698809843,
 0.7106579089183346,0.13947389038872982,0.31830510811921353,
 0.46202733028176235,0.14879795007404895,0.08721695107110816,
 0.19880260960525298,0.016713107573373298,0.49472689897964517,
 0.6086550105711205,0.3288436393901979,0.9753799776489006,0.34856688656944446,
 0.8995062457939571,0.8726419562425355,0.7620682411894246,0.7242768920868529,
 0.4495345466748275,0.2647994894840666,0.34704128618855934,0.7734105241206921,
 0.23943198639939012,0.42732398708600705,0.7160516388556023,
 0.06373385747519844,0.294219316820183,0.05722221463035382,
 0.20929813089174143,0.5419578195863608,0.38891264571297435,
 0.4227680940712826,0.3966672380946301,0.7589574056401464,0.7643383596163205,
 0.6331523677773542,0.6367109010435413,0.5237434148459711,0.33425896226049456,
 0.9031925657851778,0.4960275438671068,0.08526310389178415,
 0.41149181508794774,0.7232123775031786,0.4675531864290018,0.7265004652303925,
 0.2217134244323138,0.938527092614487,0.7362903825888808,0.12792105400425235,
 0.14966378961136206,0.5069028082500897,0.8685968898733379,
 0.29276421553110044,0.7132610233718681,0.17532774986476318,
 0.9550587520546441,0.07447116440301527,0.4875713894411706,0.9227107010052273,
 0.8379987824903223,0.49158404631245,0.7050804342493318,0.5769646112556047,
 0.04028156058055854,0.14135757867456178,0.6310067131477505,
 0.0643573564394202,0.1313177563832041,0.15245511526837463,0.8218428086745262,
 0.07598401814388445,0.2879356490306759,0.3589667807481063,0.7960799224792662,
 0.323738816333787,0.3965889897209177,0.05615268105371174,0.23165972399653534,
 0.9887017720423023,0.4271063342487742,0.4970686814718597,0.7629280057638749,
 0.5326674584416059,0.1898213285180157,0.7727402188191176,0.32188055411433136,
 0.7712592242601602,0.1722256457054946,0.1824669224114449,0.4573442248491224,
 0.29157956205591606,0.3741046405370295,0.46332952308820663,
 0.9784975738250404,0.7032450198296886,0.9815150168081378,0.31665024520675356,
 0.7325175364442259,0.1387065821254111,0.10549171501504795,0.9755887372062869,
 0.39875794681567267,0.28374495798761123,0.6451886111282945,
 0.5809386193348226,0.6257448080119831,0.3192518069833853,0.6178011396624724,
 0.8745785352139428,0.14280255009025455,0.009091461065903084,
 0.025792092381515697,0.9768997535266524,0.9140112847857849,
 0.21180720714005985,0.09330494862596328,0.6319690412483928,0.736297255144285,
 0.49996647805080696,0.05281499673214918,0.20757707927430846,0.61792139896911,
 0.22174249087111275,0.2783064558565108,0.7040882850536277,0.3369224172461863,
 0.9059542219945691,0.12067072219755914,0.3498300135344967,0.7150022885254733,
 0.13795450180484103,0.9916385879689779,0.11501646944001276,
 0.7850380012997917,0.7706370291649568,0.6660428669516436,0.6099829867614296,
 0.7037488761741071,0.5244618012814286,0.7852427641568379,0.22675674669501344,
 0.1050258316656163,0.7704082056646314,0.27680549945170296,
 0.07254257185831947,0.35512354341256536,0.5711755401540841,
 0.2628066525400137,0.6107971763199084,0.26707507208106185,0.1563866089264504,
 0.29457623927355847,0.8808018078191404,0.42058038007116916,
 0.37118768875166563,0.6619193631838701,0.4937003137340842,0.8861980734949833,
 0.836038373193706,0.360124257983502,0.944071861022336,0.6399168504983165,
 0.2143737735261595,0.1817066608354201,0.8134939345845191,0.4558301390161813,
 0.43617618737066366,0.6985381219278269,0.3619584049485023,0.698595629641867,
 0.7895422559145512,0.7636956018152351,0.9725845593245644,0.9497325213830471,
 0.08202950846490431,0.96761001786535,0.9077422934311413,0.05255991244330416,
 0.35929735787944805,0.15661899553982883,0.017738993104603695,
 0.9211050342255054,0.24876083300478813,0.7037676063846118,0.6898631751793823,
 0.616568385678282,0.8352907211032171,0.852771186012006,0.9345407010808459,
 0.34931226368941015,0.5519626214781309,0.1936159083129695,
 0.47990033132464227,0.5453122551012775,0.9606668697925658,0.7802885425394965,
 0.1735826279921011,0.5606214331302085,0.9879770615642408,0.11311569945806033,
 0.05490053715387966,0.568660590965894,0.7321086480988337,0.3939560502847803,
 0.21535540152216204,0.37593193582454154,0.6822386604682917,0.720461260307891,
 0.7915769077077245,0.33783095968017274,0.7310190974951619,0.6018971448005082,
 0.05573937874216994,0.5062419549338358,0.7938535739669065,0.9376407150757006,
 0.31549123069951346,0.3194100229210548,0.9756415315330023,0.9153481985781491,
 0.002344280005255217,0.5762541339502856,0.34041496866771614,
 0.868530322718505,0.6597834494660807,0.7654454405123362,0.7216445226962931,
 0.8827746149285874,0.04086819285287113,0.7589878996163644,0.6321120628418513,
 0.4452184081836208,0.5196064998972731,0.15774546335721495,
 0.020866173372519836,0.8733060034600558,0.7999025539805069,0.552835477489574,
 0.009470632680069535,0.6719823327105534,0.8684709800474637,
 0.9332439517047009,0.8506084788385615,0.16830315217551362,0.0527467788774304,
 0.1376905621551141,0.4507379857153574,0.4341807619916428,0.8493398779559538,
 0.7749967401556344,0.04068339138231336,0.4752885947382498,0.6150493677492633,
 0.6445739145879937,0.07886981896627632,0.12623463949696812,
 0.27883868793026245,0.9705285747000252,0.4770743686020279,
 0.06424087482987084,0.9061952791107293,0.5071140863029201,0.9009774765542694,
 0.02910354501330681,0.8183960575462857,0.37366019903803693,
 0.8856687850514448,0.17008821178029032,0.6051499810989751,0.5036806492416465,
 0.6322579881756479,0.8087767497961282,0.779562409869621,0.08473852268509563,
 0.3707862354053235,0.40238639887044625,0.9919472613019091,0.8049113105509956,
 0.529351951146594,0.06106541438153823,0.4878239457334903,0.5203024409310684,
 0.5947826816972083,0.6424205001215013,0.36498186957848255,
 0.12289222152727197,0.010887453451064122,0.43245681869335495,
 0.18990815954635154,0.33981168476084356,0.786609728423364,0.8350647631270246,
 0.7650937858584783,0.593609055294334,0.42263436617393424,0.4147213693523507,
 0.9955367274923344,0.9786734000721518,0.8933583152546103,0.4872292584523531,
 0.5678374610952426,0.7907813377565974,0.3133837195005347,0.9001274313116749,
 0.5980703356875783,0.7649403541768142,0.9767367135726421,0.48690723773328426,
 0.7626220740221376,0.12941092975307467,0.8521914620818196,0.3555493598630408,
 0.8108610584145357,0.43839908095915014,0.25048779552307854,
 0.9309191560910732,0.9407585541227742,0.13357793826593645,
 0.36607346155367304,0.3155899084875833,0.005734103976268501,
 0.42005046919251865,0.994694594038256,0.636591706673408,0.6771405802143039,
 0.9938170304519526,0.5352619009260189,0.25654000336370486,0.701440575293019,
 0.2934597500452303,0.25771710935062353,0.37237839391101524,
 0.12370903566797575,0.9649031923690712,0.4277390948975861,0.5590078361425954,
 0.3126171856741222,0.4709431622866259,0.5453674044517834,0.6089311334846765,
 0.8209525915300108,0.3318665711005263,0.5521532247682167,0.6809587063599424,
 0.5579350020817441,0.5203603043704118,0.21780770960455298,0.9051102647020608,
 0.21970847183943087,0.48758877782826404,0.9965610061830568,
 0.11613765191450665,0.5687298817755936,0.4892488552879497,0.5673312178297919,
 0.9079455423296636,0.7753056859984033,0.09483299099356068,
 0.20291197631063684,0.842098045536865,0.5818268351628614,0.508515844178827,
 0.2698031412730527,0.782593199691531,0.25019928255930635,0.8017289277296358,
 0.9190210061747642,0.020000191455752936,0.36800881555720644,
 0.9697690594779933,0.8155003870055256,0.9186383526065908,0.29082875142079945,
 0.02461304165598044,0.23190098387509428,0.09989375305539927,
 0.8676392137995004,0.013574433716134982,0.2963213106591228,
 0.3782585794682817,0.5012868095744946,0.628073567601996,0.04775525236495648,
 0.31668643348903247,0.03272774651283894,0.5261238768057698,
 0.7686715042639183,0.05192609182843144,0.7484180851041756,0.5016619090149583,
 0.3672536489857845,0.11803948520239127,0.7434060155475882,0.2958466500873064,
 0.4595289406884824,0.8119265710713106,0.39721874072979346,0.7780724799142433,
 0.9175736250765749,0.6483805309999594,0.5611880646733501,0.5500932892165558,
 0.40508698585347513,0.09125030072170326,0.3331139683385297,
 0.7664243277916412,0.7244801455942019,0.292821643661929,0.6650886123592262,
 0.803815151720813,0.7453110486531995,0.40252994382151897,0.5960868421859405,
 0.29933226205281216,0.22774485698650504,0.4778864491482967,0.579052966519625,
 0.6063628744089887,0.9485776533949776,0.7610599184663778,0.47695085340867027,
 0.5456370437037997,0.10645510982348705,0.37981360460725055,
 0.7571807623019337,0.28178063960841615,0.04908657751473067,
 0.5444852966317373,0.8677928324513036,0.6214378557942408,0.9973089807119669,
 0.1004581038517407,0.875240010639654,0.035712660286382425,
 0.14857114782167813,0.4048029414147233,0.3289450246276324,
 0.012510721164814331,0.9434333299770477,0.5080781535021521,
 0.9585767749686376,0.4983016074679496,0.08842795924602165,0.8839981701425832,
 0.3114674243044734,0.7785558052208571,0.13023022934120554,0.8433143359647202,
 0.6945403279469672,0.42836839538577554,0.9129998582621469,
 0.29764347273496006,0.43368322069310683,0.5626869994111448,
 0.5562285487339481,0.8202337703637356,0.7035802503375203,0.41381964906455715,
 0.4027613093476925,0.2920604302743297,0.29656359616134065,
 0.49152448945627214,0.023257212394405036,0.9779492262267748,
 0.05950811850792892,0.20433310790393178,0.5533228468023785,
 0.9191833754417058,0.8920216824010089,0.9475918456146852,0.9681143989311947,
 0.2280869038247081,0.04345787287610714,0.525736391182963,0.8367783315847254,
 0.40457752914937983,0.10664102186188695,0.9256641375911119,
 0.7460229074441297,0.04956603135622317,0.7839884406430048,
 0.13737769982999237,0.9701024806207065,0.999883325877458,0.5413207496967337,
 0.6491245143081901,0.4997518294610058,0.5726770911605833,0.9034420635847019,
 0.16202151988942193,0.06583134696142456,0.7901502496131098,
 0.7921373263196781,0.922679328640724,0.22774114375572685,0.9787600020989914,
 0.8077836579067592,0.6335964327387287,0.8776220039899978,0.6973205117524537,
 0.9770225775286251,0.7833367412348612,0.7009809696970535,0.6686150917246946,
 0.9968386107904111,0.18437478100438032,0.39368830625255735,0.887738587009518,
 0.5034587258214206,0.5065568647250848,0.033542800252762794,
 0.39920365780059974,0.6333251468296088,0.05313745701687267,
 0.5181656384710942,0.6844487813019143,0.595735679146478,0.8612245393011686,
 0.9776168719697692,0.2943367412687694,0.3810900522574512,0.9131128452698762,
 0.9445300701661398,0.4179516432042585,0.46608912617513654,0.980395272370062,
 0.9700318596334103,0.252377126106761,0.21310078650566655,0.3641695441118611,
 0.9363366314403938,0.4595435846084527,0.1950094839753458,0.5642438074499179,
 0.21152129619601168,0.898915032119621,0.6459674369822576,0.42953680656120563,
 0.9413171208851467,0.7053388437158866,0.5260126347756533,0.7093399795802483,
 0.6894766758289564,0.4291797262208734,0.3300405131774773,0.3821160878629605,
 0.8245746359139087,0.6344306991490185,0.9517428444359548,
 0.014756206816441075,0.6271590414066587,0.9816830484400298,
 0.49155133122787076,0.42220914303386137,0.8533409328991882,0.280615301348514,
 0.7692882207176976,0.8150432586437134,0.8480890187829772,0.7193405106483385,
 0.2660531377269775,0.7324343967231915,0.486984681393723,0.16760845608990793,
 0.8021862454857106,0.9371275294911664,0.7424546277854026,0.44847864307706276,
 0.4423841288842737,0.8057201317402166,0.21948647276680888,0.7639129473394337,
 0.5407637680057108,0.5441918602021619,0.8358946257116986,0.38694888979907405,
 0.6578135338111876,0.30778699300386303,0.23301995943906817,
 0.6688654217523143,0.7287691679745647,0.6673486052263786,0.7419933033402282,
 0.24484942174921853,0.6646145819763758,0.013803383003220593,
 0.9258156491616875,0.666857096144355,0.2481384012326857,0.9158570562437096,
 0.8034335382932717,0.31639677037988956,0.6674885666459696,0.2045115033629068,
 0.8798016821743412,0.18690112708636875,0.8235640398560625,0.8996078517136676,
 0.7545808670818754,0.6209361906390513,0.5832483014691036,0.5546859353445838,
 0.38196104474311565,0.027389289121958815,0.07897255416622229,
 0.9926543896167677,0.5923201801120479,0.5071977719401827,0.2673829721684847,
 0.43847933659269744,0.4127394120035712,0.7002343292611477,0.8611342379823577,
 0.42577283414192424,0.5061794239721934,0.6193730210034625,0.2927767405910342,
 0.23447632939547913,0.7202102324851494,0.27130410539013083,
 0.5238648014747642,0.31975049856973503,0.14395506107451772,
 0.22332017892114864,0.9112487805543389,0.8798753807394272,
 0.13159518384344238,0.04798876208842606,0.1232396659880195,
 0.31234949857631433,0.2655749623727963,0.8789482292553361,0.5467305209516604,
 0.7845275322562911,0.04044382005965552,0.18128436451150864,
 0.23592462498374345,0.1917570855644708,0.2816520748685203,0.4168524926156716,
 0.8244594396893821,0.9306496623018675,0.35943164059244126,0.5700124852801746,
 0.8444824351362082,0.13062878258589694,0.05015042526553626,
 0.2945565652743103,0.9430673883121163,0.2079014957241544,0.8133428012483511,
 0.13941298451943474,0.9165517736790304,0.3505089782211105,0.8765131629052778,
 0.39084112463996235,0.813764625000357,0.6708842984168054,0.04252242378473803,
 0.05231720516257632,0.5735956670347666,0.525536325264386,0.9220377292466475,
 0.02855977646781227,0.6160217049780206,0.6967173528152287,
 0.17689451439923465,0.24365833447196872,0.6969917306315482,
 0.14317482290263606,0.2238533471460331,0.18521412601508191,
 0.5132837119103375,0.06079060867716457,0.7874323135725457,0.9040821898220952,
 0.5888339613570105,0.10676461697942607,0.030294797215384084,
 0.16750023284932036,0.46470037371441697,0.03640332530402912,
 0.31850926775036925,0.27309861985219386,0.412383930154936,0.0545321712453245,
 0.19356530284412365,0.601052305993184,0.2508994840307157,0.7898179699857176,
 0.3616365248795328,0.03624764236443567,0.7961467899928866,0.362073698542969,
 0.20859296972044583,0.8020856547805923,0.365485579366401,0.6730211084499926,
 0.7326367673814883,0.4446379005967762,0.02822481461217663,0.6649441342135093,
 0.09358511594794139,0.8329689480259361,0.5880954693270964,0.4377103945576084,
 0.16693321992040044,0.6821383078910754,0.7375894301715269,0.6232865880520874,
 0.1763536936256237,0.5899252921783906,0.8322439636598462,0.3351343459013352,
 0.9967761537999503,0.26630065981855355,0.6968681022268632,0.6940532190132672,
 0.8937383995274992,0.8109010343789762,0.1609037207529952,0.285057239283132,
 0.26134879928757904,0.39666302319200053,0.0048921796951652,
 0.9106548005895546,0.12494498988834501,0.9938679668987853,0.7629584878654363,
 0.18454814011385923,0.4939163374557447,0.5306096791010984,
 0.09301220384137221,0.8868583211801582,0.03466299409646467,
 0.15324276347747834,0.2936300654699382,0.853245279670397,0.2822703992026827,
 0.15100208695141926,0.3384351168949299,0.19756500819070955,
 0.8498990269135268,0.16285428056508444,0.9739449118770739,
 0.09558966435710059,0.5624472861708104,0.9478912262438078,0.2864350460397953,
 0.1848061575851312,0.598311875787364,0.9283973079213199,0.6727893865403283,
 0.31036617675355027,0.5438391961888834,0.5105741288316084,0.454939929993606,
 0.26866426187408243,0.28791966435890015,0.03787374646838704,
 0.05553823891476495,0.9648335288719172,0.5097583777692192,0.3437163908125165,
 0.6298370826867717,0.4399526134686126,0.1391573008381941,0.12300407633622079,
 0.3819031307601747,0.4301044083054265,0.08159662194726391,
 0.30415069237961156,0.454608195969432,0.05526491447359727,0.8129715184877812,
 0.4075625883324564,0.5958472400434841,0.6908082694860628,0.7613581360647159,
 0.08216183035460833,0.6714658853728904,0.8664163389157197,0.9367246249389657,
 0.4624469259762177,0.34173932457396616,0.24383247251031537,
 0.10352589538678003,0.3292365471992055,0.7070531966998195,0.7746701233044049,
 0.48531791852937034,0.18650685772543873,0.47658257698981055,
 0.44873143069607346,0.1602519479895237,0.2519578212655115,
 0.047672350923641416,0.2850553625079335,0.7113987258869607,0.919056793524083,
 0.31611077007981214,0.25154774716023465,0.08796230695008045,
 0.18274906506754074,0.350122732576043,0.5133136556286146,0.42736720317568533,
 0.9019956707345878,0.32213156682282906,0.6688961338847119,0.732336721259296,
 0.5730533636912314,0.2799007081371335,0.8344872496622084,0.6640376113313358,
 4.278966780604243e-4,0.9729625982890799,0.07413821928543607,
 0.3296327730124964,0.1067241247503854,0.005479939481587204,
 0.7123581049537908,0.4241860550874522,0.29332039002544485,
 0.10919941480979523,0.9066440591993234,0.1786214139384522,
 0.09476065292056002,0.6771973300257454,0.3923464517008415,
 0.19293865821138156,0.3934775886788763,0.23649017126390903,
 0.4436819083833218,0.21512580122062597,0.3839104859178296,
 0.38925724093975245,0.8124956008722816,0.29895531562655353,0.892822832470266,
 0.676590534037388,0.04814054396719891,0.2496858707797508,0.4791874684366719,
 0.9907677022650548,0.5113420129763606,0.05708490502678243,0.9688076601378429,
 0.9897352791071026,0.9802327693768172,0.11820614042543198,0.416617033174564,
 0.11675138882631819,0.751644905195214,0.28856149055403224,0.8117989227380699,
 0.9097476055315851,0.8349381824753064,0.7536150176777354,0.6555074120137729,
 0.032486199361531076,0.5403045270386475,0.9123660849646935,0.894068025193044,
 0.8806232899668439,0.9918483705778715,0.9794906859763384,0.08058290768795096,
 0.2908513263803736,0.932158242226028,0.6276986006259414,0.7786333239645329,
 0.01591066821966791,0.9642335424441909,0.5640903566094062,0.6783763189103411,
 0.2884377043333798,0.42685629313197593,0.891458442910072,0.21451036731106232,
 0.6646896038176385,0.04675023109396759,0.046107118009122816,
 0.42739935450104494,0.527439819311203,0.7865871257313187,0.7091064891440425,
 0.8245143937406972,0.8751984931085135,0.16580871208194203,0.2670022529692938,
 0.3799966758903085,0.5684347032476554,0.7433976755497996,0.8792549021640379,
 0.43862152256630205,0.8426385919606175,0.9035098705988731,0.648447876266618,
 0.5521790167918214,0.04694682467527844,0.6822477635509473,0.0382935109742808,
 0.6829008700230692,0.03397402148923978,0.05646427412077126,
 0.16393193873676581,0.3117267891465999,0.6320812463462874,0.6947811249638146,
 0.09775516478411705,0.49823811385229666,0.07378289050916242,
 0.6314674819968571,0.6570406862533827,0.023804166128310045,
 0.5109863198778762,0.29535896487594016,0.3146093968850605,0.4069064737437489,
 0.3622074923157733,0.9502814672877635,0.6388099322525443,0.7925187500308231,
 0.809612142634906,0.40517532946417056,0.4551154983143286,0.4208362600888136,
 0.48515387234228247,0.13575072257530874,0.15456168547427263,
 0.12541660341946914,0.4075164213849828,0.8172037375737057,0.8459012309042034,
 0.6414173834983646,0.1260705795319449,0.7505426783110057,0.33359091483733483,
 0.9367732574898489,0.41415501216858486,0.9659349955131703,0.538592086060283,
 0.050427505655401106,0.221358120553103,0.733120792315365,0.16813180027478714,
 0.7063880959909516,0.12217998910434191,0.5248933760384776,0.9408484932003722,
 0.477533720594418,0.7922074026745823,0.5374519167954256,0.24695206171173512,
 0.2780975228655491,0.9058225189665479,0.5117192247595211,0.9149203009850038,
 0.7335725290534997,0.35880324473011427,0.05106268140356829,
 0.4464753731857588,0.1605213166344377,0.6587671112939757,0.4422865441438296,
 0.8136733305375111,0.011937648868690531,0.8237609742172343,
 0.7946321377176537,0.8577198411347103,0.09831936817253217,
 0.39394142920866515,0.663697554961646,0.7659992886601668,0.2700394679967424,
 0.3823801085049108,0.45953339891898726,0.22436177315651307,
 0.8676035757820529,0.7490183032931151,0.7187245648725404,0.17158236031745733,
 0.15112328354311233,0.22566223559029008,0.44610858317465674,
 0.7750342228511756,0.9670913998665193,0.051337624821829575,
 0.47354358611507297,0.4712129061228624,0.6285572252118947,0.1879462184611459,
 0.9087696863339098,0.44474372978090415,0.8960648421940677,
 0.02321843879336316,0.8188963855783296,0.33996318316098395,
 0.5078795738919624,0.8929641199732639,0.7479455301308084,0.9137910693020523,
 0.37395198070968727,0.7080454032748797,0.17856645063179866,
 0.2049596815692889,0.3489263208795712,0.9913062919710481,0.4532368311269749,
 0.7978495707921892,0.7102485465402427,0.15706371393813434,0.9137657749857353,
 0.9139267942886533,0.45664963316790375,0.682718437668457,0.6209627352651054,
 0.9042828803709788,0.6594609026662266,0.2713971287340813,0.9147068651414616,
 0.9361119689301869,0.4592274523369144,0.18404925022246277,
 0.018414387323338555,0.08028438011639882,0.8553789147212579,
 0.5868850759788825,0.19414861985933007,0.518574479818662,0.5144612388858661,
 0.24778160667897908,0.4717226218110788,0.6443807571207862,0.5545573161485321,
 0.11057435652582415,0.7710534461029829,0.7012358185037735,0.5169241214219069,
 0.9131779531588562,0.46174132883886077,0.6529799147142014,0.6407877064866553,
 0.6318191574134004,0.22923419464847838,0.3447384777500091,0.7919565785938628,
 0.799212565082534,0.558060777205192,0.21239332677065903,0.4468568093302966,
 0.013526135716971144,0.22383536560210748,0.3054581990755617,
 0.39677870808824867,0.9754164052209594,0.7505132980991207,
 0.37172053916364833,0.8415096548560188,0.3464005058146661,0.9420828701996666,
 0.347963719439938,0.6512031562115989,0.35479054370543583,0.21845197153585505,
 0.08881204966906209,0.22897287923331433,0.6875396181614244,
 0.8978288671988336,0.3593215776788419,0.9164519984240297,0.31297888092405113,
 0.5493688135140891,0.605920696236137,0.6532783006413548,0.7482113657102932,
 0.93066419165943,0.6765950038831994,0.896168703009209,0.05825675748586845,
 0.2915039328914196,0.8973246717913432,0.5505347384174789,0.7863807510566313,
 0.7624906599692594,0.7956596936135455,0.24814257887750124,0.7623109772055997,
 0.5151918333503157,0.07996238748263063,0.5057065051620314,
 0.04677752936715429,0.2308234682168251,0.7656327856155,0.6259027620033525,
 0.3217450925973181,0.8952883250220711,0.8882122334317792,
 0.0034607872883691826,0.6086324699514861,0.9714010142257513,
 0.7737061411739975,0.21733105907907346,0.5768519710335662,
 0.21009319137477933,0.8341296246987426,0.7432284093949738,0.3283462510055749,
 0.7762926254234677,0.5585348785862646,0.08130676940505943,0.9852494395762286,
 0.9553134032255995,0.7903323946304235,0.2611499169830679,0.16275847126562892,
 0.5057397856643352,0.6974859702445493,0.6977731604809395,0.26347724268097594,
 0.036524888347482465,0.587128909473238,0.759140228200541,0.4378162472711786,
 0.9818476571453877,0.6219968850186667,0.02571861678184273,0.8484332027136406,
 0.7138562000244925,0.5797673383253461,0.029729648364511196,
 0.9450233891164068,0.08993099438962071,0.25736102708696385,
 0.4723818772196008,0.26472992359327585,0.2682660002991203,
 0.24881853630485196,0.69718215319203,0.31464805555493536,0.6698412309566617,
 0.8374218951906386,0.9449549236876194,0.7397007256961284,0.9993448204321562,
 0.7947025726008202,0.4613143713490494,0.3954260452116103,0.5211683361633626,
 0.9060974601081161,0.32256486265978324,0.26848773261951586,
 0.34347188933791206,0.8964183548631356,0.2125559935576793,
 0.28278908228825717,0.31749506885113743,0.7151341800782721,
 0.9415634390318091,0.4167795818670905,0.8930662294879244,0.652954948377721,
 0.7919585762970243,0.645769421700963,0.6726896095450279,0.29616269407953166,
 0.4438817415553964,0.6536527254444179,0.031463243723561574,
 0.06181269982005666,0.07387567929323569,0.980448393448763,0.9676160417700265,
 0.47047897357091184,0.8157060325170986,0.6595711268291554,0.9892229524073003,
 0.6371656540022373,0.4495128507451236,0.24327520347196419,
 0.38325454625516353,0.6557064814489695,0.20896262458643167,
 0.3463466634500161,0.7632174589688399,0.3189899933650555,0.25082224453040725,
 0.8841559657003007,0.024422378599814,0.40912074128907683,0.505191639867214,
 0.9042872126636803,0.12377277359535399,0.2937455246643349,0.5147955202880272,
 0.9109664917100845,0.06228614891925843,0.540509591473927,0.2052452090792103,
 0.5211523187606169,0.2694233952940672,0.8063720697568298,0.0687676200259758,
 0.3502889731629396,0.7250202161957481,0.8324705292827878,0.8457320732116529,
 0.8771566915351883,0.34469803761250595,0.1453059425750104,0.7248958463822104,
 0.8929183620356002,0.924043732535444,0.20579087673258467,0.285046056378485,
 0.6910788819461472,0.24177899065075803,0.4664254456235172,0.39916329654281,
 0.4900940632960117,0.015489106509800732,0.4663181956707845,
 0.5421741338244723,0.1830487809111392,0.10436272079827136,0.2943286001153169,
 0.5407570595793756,0.6014210569233628,0.4531150057715416,0.8873442482533276,
 0.6426091543816903,0.968899946060255,0.18168118865952676,0.10316982164669086,
 0.11107134018376863,0.2281223681641693,0.39408251934303307,
 0.26105589859123124,0.23774016126571096,0.8087737241198921,
 0.5953982746771856,0.5499958878355931,0.3182668638326822,0.5767606609518985,
 0.7275657379793561,0.009861360735917812,0.6316713588108582,
 0.05724556141656123,0.9399154334919119,0.05728570423748913,
 0.26465590704129327,0.3123411143003726,0.8766655037683501,0.2906310204030964,
 0.2520502014320579,0.907467946723849,0.4204826398353907,0.9395327896256558,
 0.5399454615936103,0.847594581968053,0.3063537595125613,0.4825302278888136,
 0.0702372823065005,0.49594370169282,0.1519790937451283,0.7890944101959606,
 0.5085524729329225,0.797511477554447,0.3960243681426703,0.005526098093300402,
 0.48483239633958186,0.613071192118098,0.2745087330671233,0.111911907820479,
 0.4425386674577837,0.9496164325013648,0.12411036575969991,
 0.46232707000242423,0.32697445334157305,0.07156377309969864,
 0.9701279960495598,0.3901245582148434,0.36143230026616846,
 0.23784055593116238,0.6804318858298032,0.30452321485790856,
 0.30965849149535296,0.2937330878607225,0.3043041205489424,
 0.052416416135726784,0.28158011149719364,0.8085679342992292,
 0.38558644932244723,0.4148745893883288,0.406169924580673,0.40900210426991,
 0.9185521976467269,0.3619651909247369,0.4985190040433468,0.5583864347373121,
 0.43748510953724395,0.8261767053414886,0.14786706175234143,
 0.7033309658915534,0.8745594602783279,0.07591883423442791,0.7479300593500886,
 0.8581527060662458,0.03563746565638959,0.2723249784860464,
 0.27558934272312585,0.32249022607437405,0.12938149978364455,
 0.5144174185232329,0.4341313474692684,0.59233335202287,0.147726264687296,
 0.826170384573673,0.3121445212844767,0.6832620763683457,0.5344115426243226,
 0.5870441915305303,0.4860620177202162,0.48297917065193396,
 0.42205241784532177,0.36847893120270603,0.9657695523901119,
 0.6205806716248838,0.8815991130396028,0.3757651815694639,0.6366187408917084,
 0.302405983481397,0.4665753527625923,0.6714347816834605,0.9333050320314696,
 0.8512678616009213,0.1418975486072862,0.906953056767222,0.6424394243191082,
 0.9783529339833196,0.6566944551235191,0.8987351742072431,0.30394287257085595,
 0.05823809528497792,0.4705418512629582,0.21976867692690705,
 0.6706311794736137,0.9532771537617223,0.9171741551050426,0.6484247135008778,
 0.002828485401344505,0.2918563000359261,0.6300596717577286,
 0.6292274720890397,0.2772232653575798,0.26459772447338625,
 0.11917570480802053,0.050835947042154994,0.6244224177251085,
 0.9140358258390268,0.8439302160060095,0.41029940195633907,
 0.025378571374442438,0.7505242359063654,0.14038480340517123,
 0.7875232403541954,0.5141028249237324,0.22826751145510127,0.5224754316718443,
 0.9408704011807432,0.1452439801240787,0.08894536617263615,
 0.05880877650575078,0.1512420629179485,0.0028289222186923535,
 0.41661533076611357,0.11392198770973538,0.6948068260189225,
 0.8083901950641039,0.9417740136270907,0.4525247500135293,0.39462310604057604,
 0.18008377242997242,0.14102126235523482,0.7178375959485902,
 0.0209259512853317,0.42121696297177924,0.6547899743127902,0.5561564372758037,
 0.8459256808300821,0.47761821294449713,0.6894961616182225,0.5272882743894207,
 0.9733132966833957,0.43162993782128356,0.6722278186052711,0.9854622192254766,
 0.5957531037397771,0.3534717917375455,0.8877591467202539,0.6711031010870614,
 0.6131750849537478,0.5331561106895367,0.03305418189580578,0.9676526174521749,
 0.1514539257376739,0.07861919741658585,0.6661850774390461,0.7042955737596213,
 0.396372137848662,0.6265287962582481,0.9939031815173487,0.505836215997426,
 0.4121099969178549,0.26910318920106513,0.0705674529410133,0.404961154775887,
 0.5898875360238404,0.26082757727465133,0.9400819827229914,0.8079453290482219,
 0.4957543668481693,0.6453976718890284,0.618255116729315,0.7254856711344385,
 0.6571449376959015,0.8391095164954924,0.9563530814689812,0.4162752694899148,
 0.27362095172694767,0.38356448421427436,0.5886462750915783,
 0.38766091515340273,0.5442495562148811,0.8068277524231712,
 0.10300269528778705,0.6888331523212182,0.6901845962016533,
 0.12754225241901196,0.4424675686082803,0.30380720819320095,
 0.44524353075706913,0.77288100683314,0.8156813913156361,0.9511629825044775,
 0.4747230877029044,0.023707740920432596,0.847351489850929,
 0.38094549852961845,0.17938402531344722,0.38620849326712714,
 0.1956148967923942,0.850785436345151,0.9542909817128649,0.8219519742176418,
 0.19391113222946132,0.9903789028322377,0.22358532188080849,
 0.5521092158343743,0.5997925623298122,0.32122334917622175,0.5116342367639031,
 0.6747874114136155,0.8716041305060084,0.4021247473773215,0.8545144617945412,
 0.6484287491112883,0.3660510269142738,0.605963900572511,0.8570993974256241,
 0.6718283523214647,0.4279503058452412,0.7782151204448889,
 0.046741478562439154,0.4771226094937753,0.7292902903070738,
 0.3825141661642919,0.23867176305665894,0.5600322563804094,
 0.20816157210527153,0.8416970948358737,0.46243018035363725,
 0.17614086305185506,0.16030182737124732,0.11916274068040678,0.92158348307909,
 0.5523344738934697,0.034057029118042315,0.1131051619724528,
 0.17742625900227404,0.34283126531494224,0.43310836547505327,
 0.1348969222301888,0.09833735735807336,0.44529930084692726,
 0.8128776195826735,0.1769659598423371,0.2906103373692762,
 0.011640634415141293,0.8030919541926695,0.2766893973424598,
 0.43013724798247643,0.2764883997883397,0.012715084952323985,
 0.6292264068084137,0.027412886865707642,0.5498205723783096,
 0.7931060102754036,0.7734109188194506,0.4415599172570458,0.31119614029868214,
 0.679436764818698,0.025934912416515132,0.41698040078370213,
 0.46059233338003036,0.8459323247252013,0.2645066488074118,0.419797199885395,
 0.05532340329619123,0.18668149352413388,0.48566902891242847,
 0.49007412875608547,0.21620790687269276,0.7477186179321594,0.758641316688791,
 0.9460526979226602,0.4236775238560674,0.7895015564635118,0.37772573881759275,
 0.4447709297689002,0.8924408461761767,0.9548861201392821,0.18954816128127938,
 0.6039602029108347,0.8603162640787312,0.34570849468034315,0.6257865235241781,
 0.3524328336585396,0.2745938247942651,0.6807360660054764,0.7044892630369912,
 0.6923821277354151,0.7209569208579623,0.5852398085195702,0.858361758288563,
 0.11742418615160288,0.46416207823379907,0.5467802850115607,
 0.7303743599993278,0.1279304514601095,0.02596621763427398,0.7065487908834786,
 0.1948189399608785,0.9334205986256837,0.23863341359169243,0.0856721848291182,
 0.2840436295839368,0.9439773849412013,0.05460735989672638,0.7828880272173147,
 0.09244720896431613,0.7336147935374875,0.8325571421194915,
 0.17683970819754524,0.7800488495039504,0.4933870108962706,0.634507779226815,
 0.9747087219297792,0.611770562901763,0.7183999324926591,0.40557943988099476,
 0.3498194341981158,0.49589536548991586,0.7494779401937992,0.6702936351785036,
 0.9361430775524788,0.823164948929778,0.6049108377575019,0.33214459835269683,
 0.4368266068258062,0.3004567394131785,0.20641602127338798,0.2351848463388022,
 0.8185668760363709,0.8036388684197291,0.41347542478428956,0.566620889900655,
 0.3238805041350089,0.46257361367390337,0.6988582862687125,
 0.43822437593261065,0.6292876665251065,0.6597481260218819,0.7144828736139808,
 0.8809417389749126,0.9275612106199267,0.7165256273396551,0.8747612734001045,
 0.7492080384470332,0.23487545631103512,0.14513632892898654,
 0.5459423892512596,0.024426631593693626,0.12751221039734595,
 0.872742714557488,0.43385893198897085,0.9580474651972689,0.7231704431477716,
 0.0878504057398517,0.46664656183267517,0.42626896256507885,
 0.7373919069571304,0.9679083498836581,0.12523144043414525,0.7767899790846201,
 0.39768515365760304,0.6012853680124814,0.2904540673419862,
 0.13055503919586486,0.25906571930637545,0.35381963048829757,
 0.5874567021892502,0.4082667233770092,0.9250777999612818,0.952155838515331,
 0.22783640702864616,0.2857572111550126,0.4519475268657226,
 0.32103341558680976,0.7863218606656156,0.5099885161503374,
 0.25887967115776933,0.8875118543009864,0.026008271409839412,
 0.38084175665278863,0.5460459383519269,0.9600193993941165,0.5202958884819749,
 0.29781498000129425,0.9434137470459965,0.325330693749589,0.06736805397231926,
 0.9737889763976594,0.3120604292038953,0.30190357232549814,0.6648379431291649,
 0.8743097426421865,0.29270454666561607,0.15179295387404856,
 0.5235930053442145,0.7687444438903793,0.9792213021176314,0.6285046730524039,
 0.8129630526099495,0.7412624048660166,0.9682924533574536,0.29883939181570096,
 0.3138889497614459,0.40189017878431177,0.6951862659978523,
 0.016157462940027845,0.46166578419364934,0.026935800022611422,
 0.4146270476507383,0.5811189573401808,0.31613645129990786,0.29624430345482,
 0.2874987614088348,8.909575930206026e-4,0.7162454116868864,
 0.49231279708609244,0.678458934982218,0.2983824810844956,0.7227056352052006,
 0.712232467556898,0.21290110183363276,0.2934936795051626,0.9779769343024975,
 0.4934595721528886,0.4739828950897702,0.03991805353727618,
 0.17287022388361306,0.8102297146458071,0.9319713826706872,0.4680746826946861,
 0.7663970057282161,0.31847857029315174,0.5737176006200929,
 0.15442342880233628,0.6439245948983175,0.5912827469144111,0.9099805939173176,
 0.44443006050727596,0.07569040248707426,0.03840287432160894,
 0.17003308057342625,0.8794729238159642,0.8172163501911064,0.8658677597032707,
 0.49827146688976776,0.9131670358625323,0.9864595948718461,0.6891267698082437,
 0.8040305806170716,0.5217162230430118,0.2089768705188726,
 0.029182685937467978,0.17613740545725132,0.009725168050576993,
 0.7352074807172867,0.5115375197616017,0.7013583358931322,0.5373156860540075,
 0.45463540068474095,0.8460812096493266,0.5522722838802664,0.5555163592529774,
 0.5471221950132843,0.8663567353101479,0.16644711498844655,0.463795736916638,
 0.9384526000900631,0.32985398019650924,0.5985757574922376,
 0.31863443521995116,0.0716184145441584,0.6761858301369021,0.6386289951779097,
 0.3771239483323058,0.16548912534848026,0.6915847904923235,
 0.26218013594939016,0.14108400566248547,0.8309987864277985,
 0.48617706969561414,0.45235427437085907,0.8599969731973703,
 0.8852244205072857,0.47220012643346787,0.3598464794143629,0.6612339503748516,
 0.9449581592710031,0.5999868605164979,0.18099535223750085,
 0.26609398494380054,0.20876933535123732,0.7279837695983856,
 0.28046231525379306,0.02935146712789838,0.6173826368564246,0.558824602652358,
 0.18613478350484425,0.3658479524645917,0.02765095514353011,
 0.8315991788198419,0.9953903929942686,0.9782450284684673,0.9569996556716852,
 0.5059919313583296,0.5512469320370411,0.3403118896395594,0.26578705933158153,
 0.659907379903208,0.3278193666095369,0.176465286895511,0.708307925953447,
 0.7170311443068486,0.6358815234052326,0.3135254022267523,
 0.018245708370834834,0.7689777524912346,0.6404836914791369,
 0.5072555910190826,0.06775523292018648,0.4998129246562726,
 0.16810289456282135,0.07552731706057325,0.6214308695813577,
 0.16461428501851638,0.01733010972786475,0.17716810429748497,
 0.19529971724466177,0.2958159381484713,0.2637012196435291,0.3470854512223064,
 0.4881033498306623,0.23880758405795766,0.1989277118501369,0.3390251234342342,
 0.025564997693729552,0.3297217235698099,0.8421227955726354,
 0.7023301694865872,0.4836429853012121,0.5098406083714372,0.7166967364445859,
 0.05926726313790098,0.665427786746365,0.1767909838458528,0.7020366409845483,
 0.5717132380293167,0.9074345018641781,0.7469629357961189,0.08813641030984021,
 0.3902981751711023,0.6393128695793637,0.502708045666722,0.910939792342448,
 0.9592586006942552,0.8003230932628909,0.8126752128469794,0.588485767659823,
 0.8097754955603553,0.3686217283283213,0.34173084930773556,
 0.43128016008375303,0.9893296761978191,0.008850772467922208,
 0.3560218667404518,0.5235498812183352,0.4921088053465603,0.6955618882760131,
 0.15483804804141732,0.3052409677968768,0.7066113823037868,0.8814633292023979,
 0.6029019675124144,0.6679556696583926,0.7397923169106988,0.6793544919586231,
 0.5346239676104472,0.6499985853117768,0.6672633713657572,0.4489272759046944,
 0.3986753730382433,0.7655178861724565,0.3127544589943603,0.43006835749855565,
 0.14555633035257687,0.8004573610099572,0.9348883720747653,
 0.13631871751332103,0.2843030149811192,0.46414470596724744,
 0.5027941839340692,0.8909954211799356,0.7594012731940679,0.27890513998459876,
 0.21492176902597127,0.006497970529870445,0.33447546807962225,
 0.950831311361402,0.33236755441275523,0.047537617114659314,
 0.8044371988990614,0.31176332410564256,0.7159291449812452,
 0.49007852734500545,0.5577837747533707,0.17599858404837188,
 0.9174378318765235,0.6130945808798742,0.9706293417048979,0.09589056147057717,
 0.15452561000256837,0.9087257227551381,0.8924549380100187,0.7983078600005565,
 0.3520725164096542,0.003774390818400919,0.0437421039980056,
 0.8917863010553083,0.9462782121562371,0.41988211059774816,
 0.012571814628147271,0.3267178535990932,0.709951371052513,0.6930978812545939,
 0.4631195760698855,0.20930007080427837,0.35692041415172626,
 0.8060785430101207,0.27636113529136574,0.5732280100291975,
 0.17281672560898942,0.18046977837282396,0.47684657540781006,
 0.4066187849102373,0.7513276151629152,0.34495159981149615,
 0.31112329251629756,0.4405933595109961,0.5173555979151033,0.9141839597947262,
 0.3501534474307578,0.7920286184008744,0.8742664427481519,0.9430433749541728,
 0.1940255977135601,0.4032271364179737,0.575986849687287,0.06691079313944426,
 0.5831981193371585,0.10933025689097142,0.13077198049911576,
 0.7154466469347681,0.8080090453898772,0.24166568069167893,
 0.30562828482605786,0.2093050880100834,0.5503187400381561,0.8102133891919897,
 0.6176742455631089,0.9108135889834551,0.2705094454041146,0.2127312417441064,
 0.008813204171541189,0.8843918677747578,0.5489351452862403,
 0.5365615437305385,0.06473589074763764,0.2303686941944606,0.9244775357590798,
 0.699618275778235,0.35935218797492874,0.9089994973645715,0.03037121176551283,
 0.8837378216669168,0.8708167489008631,0.16792593043400983,0.1006968253012338,
 0.344646057155265,0.7537685025549845,0.919610120701447,0.707014820943707,
 0.3636909072456248,0.33550842994301555,0.21940179997866094,
 0.8308875287554944,0.3164269187506753,0.042835881644396334,
 0.08609720932236697,0.6893344376862249,0.8848931548091273,
 0.42460680778921345,0.6986223004620743,0.044400606795266695,
 0.5169000154566927,0.24153162948464213,0.012010285695830403,
 0.5493628150928664,0.3252496510072276,0.143535385127475,0.3295536040715318,
 0.5952817034409942,0.19147822801283443,0.1201139619295396,
 0.14943591298370285,0.5550218618496698,0.8672615259876194,0.7534293012373608,
 0.13560237272635633,0.35260341012122565,0.9504653925633335,
 0.011521714134846373,0.10190052087326318,0.5090147883084104,
 0.3672936172051121,0.9720829377836973,0.5895829635899538,0.8302311660039937,
 0.8842514872925613,0.7707570835510716,0.19568721574780035,
 0.06781626799292706,0.7453560499165668,0.3331370581795736,0.8683020922729104,
 0.19504835261018683,7.838826510158103e-4,0.6783860314764821,
 0.8969189346101247,0.8021354569250652,0.28378846157375115,0.7654821322496456,
 0.6981911953158202,0.19648823404712557,0.6106297719419791,
 0.39157510655756833,0.7108316571545517,0.8842437807277778,
 0.20550166631068956,0.5517054661373882,0.4613251273157143,0.437599804360821,
 0.4854070545677145,0.035009576731817704,0.7720747970642875,
 0.3254013439035399,0.41885729230954205,0.37226879414991243,
 0.14248489166789624,0.4424561319650707,0.9381455251953192,0.7285623277951302,
 0.24288503009686646,0.8161813659501577,0.6121523342722405,0.552141620595316,
 0.6489032092621816,0.8047388081310436,0.8437286413977325,0.4888219417563663,
 0.9535943295905818,0.8747705210980883,0.7695743060705895,0.35797472284414966,
 0.7768929694511774,0.8667031714243031,0.34200314321718506,0.292505450140884,
 0.02707731155568094,0.5669826299963271,0.7068137050037262,
 0.44452657450933064,0.7822431623720142,0.6465216503075514,0.5160737437946148,
 0.191685310103198,0.840356053642581,0.7866730623890715,0.40724932451921825,
 0.15723887590557006,0.22997774423182327,0.6471222340734379,
 0.3835987690730329,0.5226321015704876,0.603204227943575,0.19739728309714732,
 0.6849948623246878,0.4124095768002396,0.19770388266822736,0.7817716320769351,
 0.07182693434367726,0.4714662258403004,0.689500509316625,0.9513772963662276,
 0.6467608697260766,0.021128014960343777,0.9623357282285171,
 0.012446436600610178,0.04644712707134657,0.9073383127621432,
 0.46549079906185475,0.7583150060403238,0.5672725486848467,0.6305588974794999,
 0.295290277025531,0.1930254788465331,0.8116609148187445,0.03005450293408285,
 0.6676094837643137,0.061084127836185,0.6560969574453266,0.6862754172746481,
 0.7095783667209122,0.5668614481872882,0.6427579972782868,0.1542104976943367,
 0.9734841833704045,0.14637094076538193,0.8712367789250832,0.5371004414796948,
 0.1718281354455663,0.9785099814380027,0.8644787464321217,0.6136236612894398,
 0.4345826651675334,0.3189371100960835,0.9302840838759945,0.9915665736207651,
 0.4316770816983577,0.6907047398670746,0.7002208994260654,0.8709293177995407,
 0.3613053166811492,0.6979628114390009,0.2606393993236983,0.49315569864240705,
 0.48853917852981954,0.8970328698033598,0.55931080009399,0.38293606890614396,
 0.17773941278367678,0.9418873312436863,0.8380647650029783,
 0.28607102423643005,0.14272841184478602,0.6385528888618637,
 0.4452136431397168,0.525338812426254,0.6311907560722487,0.780319369749694,
 0.5210440928484251,0.70157300327669,0.6926223849113213,0.6719488270511531,
 0.01931836782655505,0.8603112625349433,0.9418144783297256,0.5801652688249682,
 0.6447205541547504,0.7517926610084644,0.4185963010479655,0.36172630650417636,
 0.24954678929251095,0.23722321172723582,0.6013495433656272,0.902958401371599,
 0.18636090776567238,0.7768162846534551,0.8447887511975293,0.9230288361722262,
 0.580817329955378,0.31055231991772736,0.31160868631536665,0.9163105827475223,
 0.073130487788589,0.14292594686466986,0.5690437354221656,0.5019079323101643,
 0.3313807679652231,0.5759992146272739,0.4680050506021063,0.9931951690555167,
 0.12383111435071881,0.2784595986175975,0.7995338876240523,
 0.19989703829537486,0.9366873393056929,0.6269824407096833,
 0.15442013168815338,0.13634622779604388,0.4402837958301191,
 0.04489479760600401,0.10316392849921163,0.4175981642583826,
 0.17116633619644828,0.6483297385096546,0.8558299991687975,0.6148891964492033,
 0.12822299865780706,0.3462156635758966,0.2530118885974594,0.2745871875284639,
 0.6583836896911216,0.43784900421894024,0.4896111646107517,0.6607719098974814,
 0.679570458048133,0.47891886380443793,0.35703738260352647,0.6999283322985991,
 0.6669903408257098,0.19271842899535252,0.6954817401367662,
 0.17869619426381989,0.6855560434992835,0.421364130339856,0.8594212823684222,
 0.5639626791592415,0.9577229649154364,0.3325083897290225,0.7282200668282431,
 0.6929730578803229,0.07524930138473174,0.08443681046994977,
 0.7869294562715795,0.2505503663043913,0.0011388688151531268,
 0.2690424961769372,0.07174055402477875,0.20193218263853052,
 0.9934845762692693,0.943372662018164,0.28035276878640036,0.6966598696675906,
 0.19427467364778217,0.9323660963969002,0.7352459703083444,0.6358344739955158,
 0.3164500140166342,0.5304718496296843,0.5337116478617957,0.36735454917432364,
 0.9397311324963414,0.022047911471974047,0.9218755289940318,
 0.9722855987398703,0.2977701478743471,0.02776447644547908,
 0.032209953425886395,0.6027327989998086,0.6394269759090448,
 0.6720560356456162,0.7466368122696933,0.07055466617610939,0.5234664388404762,
 0.14602422853879315,0.317118887369807,0.8845596190512538,0.9371776648070222,
 0.49300941511418206,0.5121898944825232,0.21235995321811663,
 0.7027540922177095,0.9372803989571988,0.03744202551059461,0.9800045887424881,
 0.5651539532193497,0.8290798958852529,0.5269483372000433,0.23855677465370373,
 0.2818562428969501,0.6456569673600333,0.005931903396276028,
 0.5848229713841672,0.9442667731656945,0.3802543964321311,0.4443449555991845,
 0.7667431904557833,0.18756156643542787,0.0981498559604631,0.0939851252819508,
 0.13878885593362478,0.24691131065703564,0.5906235125641277,
 0.16468516559924995,0.8185643128042315,0.9376553726764785,0.5131178562193311,
 0.3625863426268049,0.13486567142455486,0.6519930191367345,0.8477089257856306,
 0.6072374170493994,0.30724774079458117,0.8414909528132944,0.169721699964827,
 0.711478162612536,0.3350808758949846,0.03119464085702539,0.783687538067455,
 0.8576098353156857,0.034026541716244973,0.552653905964763,0.7751801685923172,
 0.21910569736400975,0.3529655562594095,0.17332164358392088,
 0.02284840690112011,0.700032308209661,0.40405832030889477,0.6106978390102926,
 0.006614177403874599,0.11859591228107902,0.4944028533343132,
 0.016993567875283278,0.2984166813211251,0.3917894201929868,
 0.4037564541924753,0.6085166923950279,0.058500018520250996,
 0.10239293225180135,0.2904720704182576,0.39332123052383394,
 0.7375285571225715,0.24977564318806078,0.1696194662023235,0.3316447262919763,
 0.9701625248670502,0.4801212896443543,0.776170969513767,0.39346253378342977,
 0.5252771698043561,0.22186003365215268,0.02902996237911981,
 0.6584815348855704,0.8937711942940223,0.08098260339928565,0.2117435026655322,
 0.8795027770379326,0.4380380039446621,0.34069715798988975,
 0.06996083194286318,0.5711410042226832,0.3439731658034777,0.8027022528962802,
 0.06821515879648332,0.2888751877043938,0.415730835575422,0.718092565346507,
 0.6112532992330229,0.8072496207701473,0.141558999879962,0.20801800564353945,
 0.6707185518370153,0.6737555024192309,0.5534490007046728,0.18132282604112326,
 0.32542163274260205,0.16626357259890856,0.15071687056798366,
 0.4291797487517671,0.6460361838454481,0.045481349035160545,
 0.6030555416019845,0.44741231061867304,0.557780432424208,0.3356408117840557,
 0.8331103809002135,0.5809410416797158,0.4543013999623394,0.7398353863530543,
 0.7473401772150255,0.2910534966170152,0.8480250106860675,0.7310917234261484,
 0.4456126731736366,0.2887800778210776,0.5998224687595893,0.2171797477904358,
 0.5826373173330255,0.08124194934745732,0.27300020746292675,
 0.3529993221995227,0.4291207291208985,0.28409798121396834,0.5971541048957908,
 0.08379869502470016,0.9760868459312239,0.48259056208729467,
 0.9446243563438954,0.23834455345860062,0.6820924707300181,0.5239147506638883,
 0.013011934217010745,0.19687702724204625,0.7068050144581233,
 0.44486716613262756,0.7804317743826289,0.39130374588993644,
 0.6707534954653274,0.18203064263990043,0.5934521289575491,0.8348568924616316,
 0.5420427139290939,0.0012422377225532744,0.2389101460107781,
 0.7846041382923415,0.6564827156729853,0.8183952303871411,0.11966247767158977,
 0.9916277540976594,0.29114509341011163,0.8850913675065069,0.9353605042919474,
 0.4087220269863514,0.950659015818851,0.872364632694555,0.025277829579403055,
 0.906483907214139,0.3032088037607631,0.4601191665190777,0.052320312445825023,
 0.0437771638444443,0.28619148200100886,0.5208939156523832,0.4861525962565477,
 0.8579169779481584,0.060175932522570585,0.17345966879455033,
 0.35105011245209006,0.6971133237053775,0.10844383599952312,
 0.09549230216400506,0.34902863531555417,0.4938357037420612,
 0.38405862901453125,0.6178216266154904,0.2583610183989077,
 0.20055591174044718,0.6347910790090895,0.6568924625432137,0.8134477736390396,
 0.7509974420456598,0.23287176484211014,0.9815338284847239,0.3497656480840303,
 0.544383913357086,0.3422201899486126,0.20737520003376186,0.28785243777584324,
 0.12622705075180085,0.851686915363157,0.17098397781377972,0.7221746765201728,
 0.4315261374350642,0.7307524467780766,0.29368986211508863,
 0.43331106457815216,0.6490051298850488,0.6962460701857114,
 0.28571614377092625,0.09817291129482586,0.7595673194144399,
 0.7312719564307278,0.9170081427685364,0.2998873371245582,0.615372471428476,
 0.27442579951503143,0.8501903714245944,0.8982424529958712,
 0.27512889514954364,0.08300488479557355,0.42199787702992975,
 0.9479511125292703,0.30122024935925296,0.3927281002236295,0.9621004937458918,
 0.26892891093330673,0.8145944231494955,0.6501196886961593,0.6261046559247324,
 0.4414635223766551,0.23853679521663806,0.8865801722734769,0.6796094236409022,
 0.17241470542423187,0.19065442268014154,0.868287451296472,0.8308770879503942,
 0.20955424085346963,0.537421840843591,0.9458364370567952,0.4479994699320947,
 0.4978331311364924,0.7007396588412191,0.851816097914562,0.6265519379825208,
 0.6572591091981831,0.7933877006931114,0.5793740339352128,0.6801130182940036,
 0.5545710965685348,0.45874931744630953,0.8522581065327156,
 0.18071297220503035,0.16603220616036818,0.21463299408105263,
 0.126011558274163,0.2285390672833516,0.006243847186806595,0.5768586599208846,
 0.1879317699438856,0.22640461017878266,0.2177665409355506,0.8360120324112269,
 0.8076960407416949,0.4693217416568566,0.11014013543695289,0.9394861635768292,
 0.7014542692256733,0.8384456028132174,0.6012921465777828,0.254100485609158,
 0.9424894724537087,0.6024249921475338,0.8715658667996375,0.5182736037407463,
 0.3932848223867381,0.7956025359738145,0.6117766386587694,0.22995111032348425,
 0.46006995958597574,0.8621766397029318,0.789502798855577,0.8107140602243005,
 0.915378566530136,0.6160869498314245,0.5077928174052339,0.6176478665543742,
 0.14054792285022055,0.06928388457913448,0.20760318330097438,
 0.8016348479142876,0.34774162227451766,0.40414904581174715,
 0.2745441534688309,0.4224617747681273,0.040119256324848585,
 0.2789500987401051,0.08570768744941648,0.8103288865162395,
 0.20748934123902818,0.46821037914753927,0.2002619143146247,
 0.2894962271604772,0.7406396931829459,0.09035598867146633,
 0.19752209187663472,0.7196008399919989,0.8570648049559106,0.7034548986516478,
 0.843472473530422,0.7362650256616118,0.011747038685899103,
 0.29905604419520115,0.3711682632253618,0.7720892838317208,0.6103336501464202,
 0.03334653368598561,0.2431023821646615,0.2544061708564702,0.9618532575140487,
 0.4697425198551195,0.35495974192377266,0.6489054796404851,
 0.35150499065302254,0.39477553533267185,0.7081899591382776,
 0.9913428026876755,0.37122132607042946,0.6465360657320859,0.4475185291978382,
 0.96412323925893,0.6723868798407193,0.8606081490264421,0.7041866567693759,
 0.2729680092796696,0.3583084726560991,0.3820026116777401,0.33307297824620086,
 0.9209802048581375,0.6574458354931886,0.7419252350763045,0.6162809284356121,
 0.21795202583408257,0.05723890599837689,0.9850782036000877,
 0.5445909971879449,0.24071308252326373,0.37928744569260864,
 0.46363180284090344,0.6170497208850411,0.14433819090856304,
 0.3489518626162429,0.12683738003577283,0.3013734752849828,0.8651408879172039,
 0.3980050315201923,0.609417322014694,0.22422448406674234,0.6226845638572558,
 0.738885505972168,0.7739749833606848,0.5852723089639165,0.47515684780968126,
 0.9239789223927279,0.940026309096556,0.5717563091321496,0.2480540336898427,
 0.8546037706572718,0.7707009330714415,0.6378537470319262,0.9091640206745373,
 0.4414113324389277,0.6562675642531902,0.3063849422391116,0.5173417133779457,
 0.5354093548736699,0.9115232453196987,0.07880044405114683,0.7380481922964289,
 0.9045202768065418,0.5448595854827814,0.6959812280233103,0.43384452201695956,
 0.8556117048070077,0.4673884303838054,0.8159228370310218,
 0.013422295839265574,0.7652841754899451,0.5381389029830782,
 0.6340048840914945,0.9524598258832342,0.9612296700322304,
 0.053434363424920694,0.11556598407658603,0.7263205897774427,
 0.0831877871290051,0.33338992511072274,0.818175133464599,0.9267216679512421,
 0.6685565378800984,0.15177445894827868,0.4211843200203156,0.9337663319553844,
 0.5486945038868944,0.7714596455379307,0.4671765633880002,0.9040338988534797,
 0.3238138799089594,0.27529809968123864,0.7832030075068309,0.8436127837528417,
 0.43129541117726744,0.9788768017753314,0.6004028513519148,
 0.04828669961864518,0.7787995879052464,0.43467153339508235,
 0.7027126105720854,0.5290839589161948,0.3162241806565651,0.6315478542303608,
 0.9553777645288223,0.8737127145116004,0.05867810887163971,0.6070768104410499,
 0.26049535721612205,0.9298567060211176,0.5234729621951207,
 0.029678994595653818,0.11917433229915919,0.593796570043928,
 0.029358923435005235,0.5085543009420431,0.28309710825116585,
 0.04150905704858676,0.09407200506070457,0.961495336620104,
 0.36794390170640046,0.3622656024790153,0.07081955194244693,
 0.9288584801892944,0.0748331556838131,0.9418264337440049,0.8082442583256982,
 0.6089855613859372,0.5612549155832812,0.9519519453772535,0.6775713535434174,
 0.847175183216305,0.031153215395392286,0.8482625698734323,0.4551555030014588,
 0.5730563318099967,0.7450379442816862,0.45427425312503333,
 0.47241266988503783,0.9692963137985796,0.376510681552755,0.9904242943516948,
 0.5495998633623749,0.6866479640055165,0.06923663540351088,0.9914383981316277,
 0.41618693661023787,0.0983366709606941,0.6440608988253816,0.9483892670527936,
 0.3182502675029817,0.022281918247079813,0.5477552997666529,
 0.5216854012506323,0.9986863561189817,0.6584369801894374,0.13575646140755382,
 0.5221719611282585,0.019340783377557957,0.28535505274317263,
 0.5166678753021579,0.20444515815070985,0.46064335844619086,
 0.9772866225049024,0.6294764570829038,0.2832205413302178,0.763151018174605,
 0.9346878848024631,0.549595674608969,0.7308202715395822,0.15322813907141541,
 0.4010594874751928,0.24715485083870126,0.5496902765573924,0.8754059543249664,
 0.9018920328993293,0.5664700207391725,0.018781496694581845,
 0.4657363382577736,0.8686008724087746,0.0020770031342512985,
 0.8665979867232684,0.6630062069319858,0.22458035642580831,0.6060641284729793,
 0.885066975900425,0.4921177929385724,0.47759458575397984,0.9367566052701878,
 0.1240537725595906,0.3361566626753697,0.1403280461537284,0.18834720676844297,
 0.8337946680168178,0.877082586282488,0.6516198400023069,0.8180060964005397,
 0.7517512712919628,0.5877186011760908,0.03691969214188029,0.7904981381634718,
 0.029101903697511977,0.8334165228880381,0.4430666883983604,
 0.11069480661168196,0.2656729767695264,0.40979489190681473,
 0.37949325279524226,0.13009248221606262,0.8428318811640825,
 0.03430976634498584,0.5663470318251597,0.5278131138136605,0.6832971497521686,
 0.7826150540199163,0.1808014264983342,0.25323349867009304,0.766028908470721,
 0.5907365864982125,0.09684034299916189,0.4924776107383,0.08377621151225101,
 0.9875307078333441,0.8545209866350035,0.00421833333838173,0.3767290634399082,
 0.7291572533993455,0.19710465313241088,0.7439915583056446,0.6370602701216296,
 0.13747190763614547,0.5942993313564219,0.24279847918557618,
 0.9749198186143495,0.8972974034638193,0.15977432690144266,0.9760559843104323,
 0.5029807167413148,0.07896625434156623,0.8396896776508949,
 0.49612452753271075,0.11854912597624678,0.9560266551127585,
 0.6539211272496901,0.3597114347434536,0.8961680003944865,0.43303663070662235,
 0.9790883735584984,0.7380609501297509,0.9383042126701182,0.8443157644702072,
 0.08736757576401233,0.4407461636661695,0.19018325143618098,
 0.4959949047842055,0.6728041358786947,0.3001450360130259,0.09419738781632869,
 0.47291647826962036,0.9092536396202199,0.3076316569278835,0.7915644628971562,
 0.3070184944632641,0.9250938928081138,0.8681681312584092,0.30922431114708804,
 0.7227795387388509,0.47535412327567783,0.5867404384841461,0.7741971523027666,
 0.3148633835900526,0.9381585082718443,0.9510434953395968,0.8877465664307753,
 0.28145765061854067,0.011850779238995468,0.9796591904727876,
 0.11378426116546381,8.557388532302035e-4,0.5629662709902172,
 0.3611502856061948,0.8241436250640164,0.7432665000619473,0.3052221323286477,
 0.914622600635365,0.9101343042373153,0.7321516498941816,0.8610172851748523,
 0.7232464061941619,0.3936794405081123,0.3993543827596687,0.9180674420353183,
 0.03464107707680397,0.2959080571180158,0.5053484558789922,0.8999922271572235,
 0.33167176698596323,0.2310905146225487,0.05266845953255217,
 0.7087333370375034,0.4331045890724482,0.20910780436199272,
 0.36479210433897347,0.029571623977201256,0.687921410019966,
 0.9330137817449642,0.8644204277099113,0.9138413415143243,0.31716574660293284,
 0.2743999696867111,0.527535151504329,0.6542773572776979,0.6796712051206882,
 0.24939805056265563,0.028376035787297083,0.09529195940724311,
 0.9687699689748028,0.9188146697790756,0.14044515158252724,0.1886179264450275,
 0.4635698796494361,0.8247262566050944,0.6603544458632187,0.8212170395004782,
 0.4979701658094863,0.3404632652695325,0.9192829617008662,0.0535698272043037,
 0.9836231731439476,0.27159831313458205,0.5050527500204446,0.8105203763806659,
 0.003810185122539167,0.6718226525722519,0.5137292367158398,
 0.2929045841140957,0.8739545286025479,0.4337787165432089,0.6895071120259454,
 0.7822127728323409,0.12232543782174843,0.9000721876533258,
 0.06001843495543557,0.8557461040229147,0.5303142053087015,0.8038393313505818,
 0.4331955664642757,0.4968651338691845,0.4482875532315026,0.7134993699706121,
 0.6012655657380931,0.36504162505253557,0.47314073447476146,
 0.8393285014789087,0.6663602458207023,0.17844755827657433,
 0.18190597923188434,0.5402917424195064,0.3878618312115357,
 0.08519891673519775,0.7042392218075829,0.1285374598956468,0.3981161033942364,
 0.3991424200752616,0.3405603732358087,0.510562767179047,0.6273218082758983,
 0.10112808382034733,0.587116849952237,0.2065520506095435,0.7028167004121626,
 0.2588721873252502,0.04766561464400898,0.014575053378552805,
 0.554914888238697,0.9337079006869335,0.5768041552718441,0.8971712500972149,
 0.2781586374481745,0.9006648924641768,0.2528969655160147,0.8694229089789283,
 0.9688666631340093,0.9731111061029244,0.8783712924362739,0.48372127586699254,
 0.9037593784464448,0.8495792121697557,0.45654464222983826,0.56285213661498,
 0.4414857014243282,0.4135380538884652,0.05501193205355648,0.6577228829336138,
 0.08658323057275608,0.9694297369772366,0.02520441257271311,
 0.3690238669430931,0.6718451560413747,0.6589760616211698,0.9630307472643053,
 0.5796196719818539,0.4854479751655968,0.2921303054689599,0.1544732345795038,
 0.64275690823932,0.48276280966631613,0.1391744916729969,0.6843616328053124,
 0.44823510664833754,0.3238862937476563,0.9291703724974584,
 0.010478885346727584,0.2530265505534891,0.9558740395092804,
 0.7944589681626246,0.14070003141343257,0.30026252802407194,
 0.22595512027944686,0.03370098436710933,0.7023150227806858,
 0.5663144945574932,0.3458779089955907,0.20360008635150917,0.0773431588779061,
 0.6515669357169591,0.6363862764854824,0.012135815593932064,
 0.4622609016608583,0.28373568597113863,0.34234994391320306,
 0.6790915903024346,0.5864699359524005,0.6511205915066708,0.24011241681023965,
 0.1829553653522975,0.2389008747973691,0.2964382933946833,0.37116950348941213,
 0.04043450156850137,0.5419189839983073,0.4444756926477519,0.6030727750588035,
 0.08774917642473457,0.8676991512314322,0.29629244322737047,
 0.09839413900099792,0.8162283143679653,0.07450476074231649,0.846722554913496,
 0.7941900754010609,0.07246217806527888,0.7022971404878996,
 0.22728601841063512,0.7721122498121563,0.866758545841941,0.48476748074159104,
 0.1064769313664452,0.23540204515529906,0.3502212956937609,
 0.02936356826724662,0.91017926825856,0.517342333047243,0.4001924337476277,
 0.8192965920683299,0.12034389998507744,0.1494229195250033,
 0.40145290821182067,0.937058703630604,0.2677203036242337,0.47280743605041453,
 0.35762128012489447,0.6975737070597734,0.5291346538079778,
 0.06605496457525029,0.3437695506656666,0.9982884510898462,0.2518552001079485,
 0.8413974408884355,0.7317903898816529,0.6649034436781012,0.8463342707448982,
 0.8971178064071058,0.7708485770931899,0.23095143101069693,
 0.05809701720252802,0.5067029468657265,0.8704427688306758,0.3866344550849783,
 0.18756445678234335,0.9369313210161438,0.9491424212047983,0.6329789865767288,
 0.3806976631759216,0.700945267748974,0.6097993627733451,0.2803267189509272,
 0.4039923279416108,0.174610022359727,0.4683314836380653,0.7909131760560509,
 0.5441039358094948,0.11305079768068504,0.6542508375977677,0.1173171578669171,
 0.5798743569229308,0.5318784134070973,0.7750491593217148,0.6482316494073446,
 0.19328710122316084,0.9666129613945674,0.28136910861836495,
 0.30275401425588866,0.038632757681359875,0.0439995044543009,
 0.7545950071061975,0.10546432393803351,0.7722979731803725,
 0.27823815683598396,0.17697308558302116,0.6934098073466451,
 0.5282162705630398,0.94208741243545,0.262721907652837,0.09280388492345637,
 0.9059317077752674,0.7880823586631445,0.27571452727907186,0.8948490522767234,
 0.7319572223971018,0.6020198944998634,0.32265561414245425,0.8421145190397707,
 0.5214803917696124,0.9748908845146611,0.3229708461137941,0.241495689738897,
 0.4215250725875198,0.7640279473844696,0.9503185401096951,0.18825867663353335,
 0.06340767742245856,0.9419805115533562,0.4017937310490043,0.7484604714588835,
 0.721043753885223,0.2857485722208656,0.16262747287885426,0.6674459039920412,
 0.1255193460046432,0.30109136145333526,0.5717752333754792,0.7597385522399888,
 0.1839764710102504,0.570769972779573,0.014261566210530763,
 0.40104392747727347,0.34173632972598256,0.6067364122172029,
 0.9306112385741101,0.36933838376055184,0.8492663100031792,0.2577078305909146,
 0.15124491125342332,0.013559414921438084,0.55647257732032,
 0.07318513502233326,0.4582308107939451,0.25310251777747117,
 0.2730741952288004,0.8305314698959712,0.28736338698439956,
 0.34277676984788275,0.9922252155292315,0.3772994272771716,0.6372255536699319,
 0.8564633825365154,0.6945805844628963,0.48567564595094503,0.566568651523993,
 0.3169502023099162,0.0666721278750787,0.9571528581023603,0.7250651151157441,
 0.3609903731101983,0.33043036328717323,0.8115288422402873,0.9514693168896589,
 0.8509447134707222,0.06990767134217368,0.6323812415496123,0.2457134097474063,
 0.4005761773003861,0.8559341682138352,0.9319401170396728,0.09316932586275006,
 0.8550003354512461,0.5123125065520389,0.7295204787339516,0.15792887985028292,
 0.5599542336955735,0.8085992678594834,0.39865510557890504,
 0.22078182792906376,0.8798597798390462,0.26508630044151216,0.621136923871997,
 0.7242647179543028,0.5470899478360763,0.1911882662313622,0.7682540929068717,
 0.16770426592590604,0.7869607850619305,0.26565574516040324,
 0.3399606776265126,0.6145221973236709,0.736151000229338,0.3145088843505446,
 0.8006444300295272,0.46984353260289724,0.04056011271843918,
 0.7704113849088249,0.3669550639039867,0.7777785973062419,0.9666846765086251,
 0.5799676579052199,0.3660543421382726,0.8694706729350583,0.39419633284408473,
 0.29218811622194063,0.8365356283362676,0.46122821951124193,
 0.46647851477974966,0.25541649587353954,0.9585032137702773,
 0.7200548554792755,0.933159543900368,0.9657694024676717,0.17631223541300012,
 0.8309015641875038,0.8125621596517751,0.29293294141908244,0.7327502540671854,
 0.0849122504837666,0.09892944088774913,0.35892135568426053,
 0.16153916970185134,0.03158240904522047,0.27457118272017444,
 0.8390496725504373,0.9657663402360575,0.44292286501287437,0.5006420029852364,
 0.031075079805791272,0.20259500214151238,0.5426440598778034,
 0.43826681691313196,0.7780048536890892,0.5034869736767166,0.3405841249771445,
 0.773373249731332,0.12205116610231914,0.8480197478802898,0.4722259245939615,
 0.00213401719709172,0.2332864899795397,0.5956269421964796,0.6192590143951481,
 0.5420670319911194,0.6059434610969682,0.15754593043357967,0.3294383900478459,
 0.8772714488745699,0.6047464766490397,0.05856485713607307,0.6868842933724305,
 0.48645529983135893,0.8068098332690306,0.966044865384161,0.9300536351350186,
 0.19360814020041572,0.3575734897908769,0.13391373692545028,
 0.47607180640704294,0.4977593085062798,0.5268636270841675,0.1762168907735464,
 0.47578592425310484,0.2538186968614293,0.8228867544695188,0.6898132105756551,
 0.40477102333105797,0.9927938388110936,0.29354579459985486,
 0.9002937684858494,0.22246199128593835,0.3827469449330021,0.7796170324799054,
 0.4328397876817851,0.96714438048408,0.5313492464568529,0.6537234174835871,
 0.012473288713792274,0.9408077915896385,0.4148677457296285,
 0.8079987012016946,0.1411843905307636,0.1716669909245131,0.21436484619172325,
 0.29043009985762136,0.4082032054696726,0.35183807215415863,
 0.9582415125184544,0.6351190492540735,0.39434356588198016,
 0.06789277647059888,0.3511366219853149,0.024957304076266063,
 0.12821195291184162,0.4467171434068682,0.9210724411895633,0.8945902261055605,
 0.8018236569933435,0.248121699618718,0.908990860205486,0.08705963618206425,
 0.36104179974761197,0.7241402693628857,0.129511277366833,0.17558979402322727,
 0.3892380023223696,0.7078559857476672,0.7131171707339747,0.38251151292112195,
 0.5958435895678015,0.8652442287666787,0.7937752935417028,0.3059206916527387,
 0.3017702493045119,0.16181373604659155,0.7589068417421483,0.352641084598063,
 0.236303472198647,0.7493209963595772,0.8191247568165113,0.9629233681380953,
 0.25897745029168684,0.23250649957973868,0.0698004430895367,
 0.24884706999468276,0.3857284657171828,0.18250189022609564,
 0.7844667945170486,0.6848791586777144,0.6905215438131137,0.12187748170101431,
 0.38869086329166536,0.5298508593840361,0.03253050601743668,
 0.9452757337674907,0.7416298600883424,0.41860968106462093,
 0.29051273978106673,0.9748561773671229,0.9204950101732697,0.760022662216227,
 0.29167089742060026,0.7718856184860504,0.2921830275138898,0.7733923514757925,
 0.2725054380430356,0.5032707047044213,0.2160631496005555,0.09913942565779688,
 0.05640798718825568,0.13779845132326685,0.3719318775260517,
 0.7574949891283191,0.45031350211677357,0.9999457169633,0.1896305804227454,
 0.3040318230334126,0.31054569585092795,0.5209671450460469,0.940183024465149,
 0.45658056229856037,0.6889469286349947,0.10184116334999938,
 0.5981920387591301,0.21097274456163184,0.37361443281510776,
 0.07780534197255418,0.6953493199413304,0.8943962195067753,0.6013989260009958,
 0.6647817183091529,0.935704151398933,0.8430013066962139,
 0.0017201474156371965,0.04435783690671202,0.15915590612836283,
 0.4032685485532159,0.4562735019350861,0.039400290663204274,
 0.9074894282089385,0.7944599279900901,0.63266288495484,0.8753428247876693,
 0.25157795768553093,0.34955597778733327,0.5874871629303211,
 0.6320254506076038,0.6590341574379928,0.7651616524318708,0.8954252231438953,
 0.41703761856179566,0.6920205390705565,0.6530128336279208,0.6374919379993831,
 0.8115267427134393,0.96122675950449,0.4050195693460774,0.8425916748261123,
 0.5069813133954967,0.6532398038177131,0.7107645046709367,0.827549108658274,
 0.31842759365389894,0.28720991667318385,0.9344740360017201,
 0.09655194405053147,0.5477962831571807,0.7337313226040484,0.6882756537953005,
 0.6156782876557516,0.403875771357166,0.5825800168438368,0.6426655967410523,
 0.2334496696843984,0.5228717412166357,0.434421921351242,0.1364219689980124,
 0.9385486461377345,0.1622195159391615,0.4764874602967625,0.7278975659665193,
 0.4527638269292543,0.9284748091175827,0.5446781579605902,0.1063595720369801,
 0.7279013562532128,0.9308694703064607,0.5358435356826405,0.04442206927979542,
 0.8662404440164043,0.6908041789132162,0.22080968541016632,0.5123312442778662,
 0.7869135124583378,0.9511064875344644,0.6414749826858596,0.6863188058938336,
 0.5178118880071928,0.9201782810708953,0.316714725019719,0.2762895128047832,
 0.5118771550058075,0.6138445975474147,0.9701870670354662,0.7046622264734186,
 0.2560063363406473,0.7106744650726647,0.5753599065671775,0.7046666933551229,
 0.2412715506143297,0.3234331770921781,0.15251507479361004,0.431771006880905,
 0.006801435081812968,0.919327545023065,0.3085993172324014,0.6759670872412208,
 0.7206571789918781,0.34603843292049885,0.4662828818578363,
 0.48073521355608007,0.09233179875482089,0.24135761178113602,
 0.40513302733765244,0.9427729026499446,0.5751731773353401,0.5325055007703174,
 0.6678122867557257,0.308844726529071,0.9500672264583213,0.9849576135614366,
 0.02062957580146918,0.7306981355132042,0.3190278859942449,
 0.14666817502585494,0.896883976962417,0.16661806724969508,0.3909465156678771,
 0.5721896718862376,0.7181364337067588,0.9420798631127898,0.5125584413623724,
 0.6301633155369408,0.9371743469342122,0.47185206545612013,
 0.22597552910121477,0.4888601520623257,0.8405843264430145,
 0.16760845043798767,0.8678565455456285,0.21248606325691577,
 0.5272079649179029,0.749568363028912,0.524935729268233,0.36508371914746796,
 0.8181027096604166,0.1549184195897111,0.8122330476544352,0.900394417244222,
 0.2177352615485032,0.29272584360333387,0.7712217835390106,0.1432774467363499,
 0.6625629741854676,0.20182065943278604,0.017370742158667074,
 0.8110665182546875,0.07120408292349711,0.4664307296739054,0.8532142662696958,
 0.3810105498284536,0.6498401562316076,0.5800798349727081,0.2513590697188731,
 0.37654879693423604,0.22070526154732617,0.9272466502291155,
 0.8340834857926778,0.5946485500048317,0.7461428543028771,0.8139224082732888,
 0.21030072908194164,0.9007433307474806,0.06749376287063003,
 0.6588506002171327,0.7728135020437283,0.9308879570013012,0.9978830086107771,
 0.5913593635159262,0.4134308846851069,0.8409490941388684,0.3909040569148815,
 0.7237891160115492,0.5616942463005287,0.16844730976865363,0.0612694233686949,
 0.5721834614686437,0.42604422236971407,0.0339119286277364,0.5119767313146628,
 0.7047409349443486,0.06683122007591424,0.4733707979661874,
 0.45817921592917443,0.4252369707990443,0.6825489841959544,
 0.12044218602226764,0.18762158685558272,0.6091968984354833,
 0.22478822444920032,0.4024923462161646,0.7242237050475862,0.4054255164041066,
 0.6568617536124787,0.08207314101910435,0.9546189313421096,
 0.18598669472267382,0.8230153620537513,0.9004430916948556,0.9601258643902351,
 0.9180946490239523,0.7659496066495958,0.5637518697174019,0.32953828397331264,
 0.12642223647958817,0.2337212343837538,0.044279779596061775,
 0.32657515209961063,0.43357285815801716,0.33612621336296045,
 0.0011892334897757317,0.8589533385955298,0.223565896255713,0.683284569861696,
 0.09657958174306103,0.8818985452681964,0.006324838190781112,
 0.8750666398465021,0.3132895539363252,0.3073936021674749,0.35874430713328176,
 0.356574947872899,0.8827521948396262,0.994508675413808,0.682693298293491,
 0.11715454886861465,0.5538819184305862,0.44072056466452114,
 0.7751821230685525,0.7409152303032902,0.5296529004562022,0.4311763326856498,
 0.627745334659083,0.8477956305934771,0.3110256505135611,0.19878584124603071,
 0.19798237115560924,0.07937693921146116,0.012165539272259385,
 0.5148001240210295,0.7038303689507488,0.5910679152710464,0.2331191293555036,
 0.4724919160619777,0.28135629915435756,0.20923780296743044,
 0.21176391376400827,0.07953618709359422,0.18720519124359192,
 0.5848954497078399,0.6322921122781217,0.6378992691867085,0.6572917401930747,
 0.6499286631388618,0.12218585390404346,0.6463316950394221,0.365204955531907,
 0.1541422509576582,0.8986723684362848,0.3995047091793982,0.6472580198482115,
 0.03311008554649553,0.008311814307331855,0.8779336973507286,
 0.4922046310249366,0.45535882842445863,0.46738552895674956,
 0.6856590987829819,0.6709506326330136,0.47577519665668455,0.7284882564725804,
 0.12433023913300978,0.26033274669470075,0.18692403804185753,
 0.9015271790778446,0.7626897288885038,0.6841461870151515,0.8332052777281733,
 0.46411595336388256,0.8509763656333974,0.7783613717170044,0.3208310161982775,
 0.42107756338267077,0.5211838935746733,0.23938621160746432,
 0.3346942678073994,0.17820936737121107,0.1703118990300625,0.8027807799662484,
 0.42712359790221255,0.07843227690568455,0.7349575386128775,
 0.26888027947018434,0.8691999699224384,0.6388781603989423,0.9157620442457503,
 0.2614478673363877,0.020634067035018355,0.9747604920585026,
 0.15015337990448752,0.7266389382118186,0.3536843029293423,0.6398927341412383,
 0.1413623557048046,0.9173699948125752,0.33114358239580877,0.9266864778441666,
 0.19628681580531449,0.4470768595056849,0.9784320151265898,0.8557218058694193,
 0.11022544822828328,0.6247695582692911,0.540992061662739,0.9850924310213398,
 0.31854152325533547,0.009163489348692266,0.9185221664031133,
 0.2426941957078993,0.8524923562599145,0.7916202310110455,0.28204370728152917,
 0.5683594892990382,0.06789365509438727,0.4422184150609003,0.51181871348837,
 0.5345886138324079,0.14977932771360247,0.5420192749685669,
 0.24108309055588717,0.5702968036914099,0.31372149690035034,
 0.47174212365427903,0.32010490523738466,0.3653629471781098,
 0.7640522784585626,0.633816664002437,0.8639329189019508,0.02267166274959509,
 0.4683898077870997,0.9048888233335947,0.524420666638564,0.04811800563009838,
 0.4237528686958467,0.6453237298202698,0.6385369673727177,
 0.003108382332252857,0.47120459437757845,0.18221799836097363,
 0.6271604243939017,0.6199399468533934,0.5746783646347425,0.39793253611096935,
 0.005898812585413471,0.060965389930464964,0.7692708371697032,
 0.7420809788593958,0.9424498541617423,0.4574675611856005,0.4498297941487629,
 0.7459904362383224,0.32276100492828386,0.03146089310332656,
 0.5679121749181293,0.39254560654609805,0.21047871029993892,
 0.9555681767186522,0.9979004263999313,0.35130188476180135,0.9147445575229468,
 0.22314258300854306,0.7108007887204247,0.3136192718141444,0.6236076554473148,
 0.5270528417967562,0.46200175073132543,0.39962480812930656,
 0.6340804295429018,0.6402466526845337,0.9756570855748109,0.849230885584129,
 0.19389251223308934,0.7271576101586426,0.920632658759262,0.592686426003986,
 0.7939989858212917,0.7015770368327144,0.039456968710810614,
 0.035226541947780965,0.7402161164505779,0.0507788097754307,
 0.04191898208883982,0.14871455168968362,0.5041476740445391,
 0.34188064981578103,0.39351227958776513,0.5561374362541666,
 0.1756319854311449,0.35395483526187,0.8163441900639563,0.030622688787298125,
 0.8514386940181695,0.748217211316549,0.7326689851124755,0.06156152197556897,
 0.9044744805063974,0.7002070745403874,0.6739704873202478,0.5577183988193929,
 0.8893850866660891,0.7238957580501051,0.07935464712490459,0.3948165567090286,
 0.3042690105165198,0.651314172260812,0.27865960962038694,
 0.013767185918941882,0.3379053345669938,0.5799351723159144,
 0.7884523087406312,0.5551401991413245,0.3503835661414473,0.3056192704341537,
 0.37653637768562065,0.6277454527687808,0.2077608964157851,0.2579549220377507,
 0.06976636793848434,0.660125011816874,0.49439555279573155,0.5897824549014189,
 0.04569385850785801,0.8045663076564243,0.48738709634782607,
 0.41859192551007607,0.4098570601734616,0.6694199073176288,
 0.006066477333186215,0.41874170805028244,0.31797874197571696,
 0.11328416360904103,0.6221715416515703,0.8775317033622186,
 0.013754729154746181,0.19433057334441017,0.7703993393245561,
 0.20493982649472708,0.1150813313852177,0.9404776344178466,
 0.13335676069772218,0.44717413181734667,0.7649586704561254,
 0.04923930253727804,0.6147647564524464,0.9707087857574318,
 0.11328631318383386,0.8645229199657024,0.556070888854118,0.5051493145779132,
 0.7355741181784337,0.43300113231594084,0.5369891776132917,0.6744527729172612,
 0.07217995430714363,0.05054404408083768,0.49288877413327836,
 0.8354701556150612,0.0186286716984716,0.8888997035804996,0.2085324848793454,
 0.6455054736296254,0.008721034685615803,0.586028252695443,
 0.45957257941752905,0.19593859469168007,0.012489920588687808,
 0.8798606489785546,0.46914344308809275,0.8151210049867873,
 0.25298780809957944,0.39208853965297075,0.8950245399328408,
 0.5793011365088812,0.7135371788977527,0.759594604496531,0.44139968378823236,
 0.7902242889008778,0.7260503333083744,0.036634099562937616,
 0.6118255301493053,0.47240038171679966,0.8750678165661206,
 0.005184568878044082,0.4512436433692062,0.8277256012703287,
 0.24868313494907968,0.17194442566374923,0.907951821922368,0.5125909029498169,
 0.12802930538000257,0.5686427797539497,0.37739906749095564,
 0.02265283289287734,0.9132069296930966,0.6999241468594068,0.5886680899696121,
 0.7416190762040364,0.66761816121494,0.8812053474590676,0.6197031703847105,
 0.5082512513688234,0.3377598672740645,0.31733028475150094,
 0.053779468119472096,0.3135821261352545,0.6609032323044369,
 0.01272826408749983,0.2850417497209726,0.8919113048627925,
 0.35966285731486325,0.27369004913108497,0.9511935593359628,
 0.9956479374144321,0.2564078938608275,0.8646938536460546,0.397259152118111,
 0.25970092938935474,0.6072268067850668,0.29920785665355176,
 0.08219420679871792,0.36765849294814956,0.15794437979463205,
 0.9069280158683104,0.19996172547413504,0.02515817390159647,
 0.4977231909386872,0.8090447316143341,0.21898810432311988,0.7266045936697471,
 0.3358858956612075,0.2994999141154636,0.06723306648391647,
 0.22030169955135437,0.968698042063672,0.8361040184142877,0.316971040119279,
 0.005068456720509595,0.7594733322008846,0.13561160855818044,
 0.1706056785561485,0.9999406969611878,0.4211224674784486,0.8078622812560362,
 0.18201939339631323,0.9991500020545834,0.12440058400224285,0.544945459928895,
 0.17588854645264995,0.2001617663430778,0.6359988920156914,0.5919081386706075,
 0.9251317292447345,0.38125621750471406,0.4095451456306951,
 0.12135786607224519,0.29881395386302534,0.27460786225061096,
 0.8666286549268257,0.8253581915167587,0.17407455777841097,0.6844779392384667,
 0.6685823104224285,0.48465735812544686,0.18502106690755082,
 0.2818875568149275,0.10920037802957527,0.37173215578583163,
 0.5807610998687411,0.637985213242749,0.09638570598079355,0.19154734698717069,
 0.9752314366271064,0.09626907397797224,0.46075378876753803,
 0.25288178033189157,0.8134860883313187,0.29087150818660745,
 0.9736879392866288,0.8700336379506178,0.37265847928365936,0.6395065291409423,
 0.5318615245725069,0.341197463137896,0.28698526623076304,0.13474133831479373,
 0.5634182306196306,0.94116689874941,0.06099992248680253,0.6460617808502461,
 0.2919201392278914,0.8352616775549102,0.02333378943172615,
 0.31621265946668553,0.5572448504136543,0.6247130025638624,0.4319876664651049,
 0.19667426906020846,0.8926123239756302,0.7007947031609156,0.9004178335003375,
 0.21626114328951884,0.7805588329887664,0.7126680513318359,0.6759373562848513,
 0.6723709967464242,0.6888226922829646,0.3844392117715705,0.8598352288105618,
 0.2386185266018107,0.1592025744028699,0.5793356945760639,0.35713401746650875,
 0.42746887426889346,0.3891844910636332,0.029717217134125007,
 0.6198809822297294,0.1641447032225959,0.6458320977561064,0.33071288303174984,
 0.5828848725695508,0.28764124817301484,0.6914518891155124,
 0.22120442989467637,0.23063421229615533,0.23356720144504628,
 0.8721412869374455,0.34560433826703674,0.5174504484470386,
 0.11582486779506995,0.08248300367224792,0.7787266068589632,
 0.6415182264813847,0.23289190644677238,0.5690015870466967,
 0.10794485314338986,0.8690846457272378,0.11362663223000402,
 0.2667831148807215,0.4913275471849686,0.6425379036475494,0.23512880838964045,
 0.26886175382551836,0.6057954571339029,0.6903831890892537,0.187708550820485,
 0.15450488548777974,0.5698536278414171,0.4135858675567319,0.9178678041365849,
 0.6012872409024501,0.4309380564530182,0.6868695351377927,0.6251481987067933,
 0.17715845954061438,0.4867202202942842,0.6339374800415589,0.4735239536440352,
 0.6081415340227052,0.40320404438692337,0.23416654516550972,
 0.4845940479662507,0.6190953932383916,0.5055793041949143,0.4841650576693649,
 0.26819057780725697,0.8109091212644037,0.6333901551156181,0.6930355010045417,
 0.29582719206824104,0.5491414344758466,0.987070182856182,0.8612656369765549,
 0.9156373743521642,0.01126788163900061,0.4684344274713753,0.2900860141339626,
 0.5915291463972403,0.44144676617877243,0.21339070540202476,
 0.26393137297156166,0.4660503706571385,0.921923301481033,0.24537221848383828,
 0.8699054465804668,0.3063307157847399,0.49014710987324583,0.6590566697555093,
 0.5347549045924033,0.9483210580176089,0.059627208521822905,0.891033339327924,
 0.03303878691190354,0.3504651267557559,0.5044461467570247,0.5943353559287754,
 0.3904451818458732,0.28924459715909534,0.5391571923478302,0.810952543674442,
 0.7038143783788121,0.569537947805939,0.8099576448682628,0.024898714963756508,
 0.5745199341608643,0.0946256839653482,0.6918878302236346,0.557358201377888,
 0.018985683026486844,0.3267500929340732,0.44935245155171555,
 0.018976121047659777,0.9131321377311653,0.31152707754821984,
 0.003903915206105668,0.299722747058944,0.2633495572292658,0.2975525739931091,
 0.5347218885743867,0.8222481558476613,0.3143198023135684,0.04093193119499616,
 0.34383179096896954,0.19980951948402026,0.43794827362714817,
 0.34297768465433487,0.6692250492614397,0.6815913527791337,0.5031933336829432,
 0.4008651954589013,0.12661133738310393,0.4268770227350409,
 0.37438205598234997,0.8908118404768135,0.09794332493628344,
 0.5202303296703219,0.9958207529697175,0.11850750038179858,
 0.03160048109192126,0.8009089549944937,0.9826908559061972,0.2336911416929628,
 0.7054459223031185,0.8075606659649801,0.3553601660225636,0.6722914982559375,
 0.35339003570141325,0.6614421693907179,0.9060078325383679,0.4117464763503347,
 0.08092112610377367,0.566604423440305,0.7350704492665812,0.3366536307556005,
 0.08513493409252848,0.06730376251623282,0.3009796891506211,
 0.16071815885726326,0.23639804964016597,0.5748848160205529,
 0.3701273511030688,0.15433843743770292,0.6180082137771846,
 0.04771162483152336,0.5671187508401965,0.8121215302852962,0.8070065639333768,
 0.7867092674774417,0.08833185743840843,0.8269220008894207,
 0.06294014249613344,0.5202064318917825,0.15804771548004926,
 0.9157903693458012,0.43032888869708086,0.8320278244303099,
 0.39273944703483854,0.5375365580177475,0.27147040389567945,
 0.4735831901276679,0.04929732612822779,0.4049176327220503,0.6831023901360673,
 0.2305024253433876,0.759939612879143,0.6701553237893632,0.5389899784181377,
 0.36944452612426526,0.7911506937945121,0.7720459686246043,
 0.0029600955462333323,0.8694334684792442,0.8952521100850949,
 0.769066173599661,0.6404108647864428,0.819800232959869,0.5132873731567187,
 0.3555889035720321,0.6625534609658178,0.12203656409892638,0.5902488025481214,
 0.8578735868428989,0.07014664939951942,0.09447991956322044,
 0.9629291441901952,0.9594462687097911,0.03278331505987142,0.5042174697336321,
 0.2824077016348674,0.22058788467654344,0.6566137733793183,0.7586874214846037,
 0.978485741626301,0.07532249556691839,0.004444851375853043,
 0.21030543367627053,0.2783067778465106,0.7877775147631648,0.3244787931839992,
 0.5020483546557042,0.04006631827601681,0.47246122820793135,
 0.8350325643039683,0.4720533619468348,0.8684490432231184,0.6451881908416605,
 0.47590363004754654,0.3178482542287898,0.2691070107444361,
 0.14382885498844344,0.8438261769024312,0.4450492261412484,
 0.43173905507968335,0.1246973426012491,0.44137584771498495,
 0.9429498352746583,0.48546706297753217,0.9185522763217882,0.709159404794101,
 0.06494630204930796,0.5067006908590698,0.8847724614488448,
 0.10190253199464872,0.21877093279889026,0.5009157794218855,
 0.8902385999095174,0.9981469382826449,0.8705407319345673,0.8734990532302082,
 0.5064027308641266,0.9248698410124807,0.3549218635112018,0.3003661928131611,
 0.4146713936393428,0.42609072265800974,0.540773493349774,0.09392908603988481,
 0.8839324590253947,0.21724903065675705,0.45736739491403167,
 0.5951174921048632,0.8163194858662893,0.8529737088766892,0.6758845436846626,
 0.36534439710081545,0.5381857715555869,0.16089358368643025,
 0.14570085011794287,0.2812545894284939,0.5705179633148648,
 0.21622405773491715,0.3957140499043551,0.7774420256641926,0.6220892863063398,
 0.7463873857026866,0.38768579990614493,0.4433097278074657,0.8753497965666408,
 0.48653950241067956,0.6511832684405051,0.09050735510779484,
 0.09322440129046639,0.5046244317685737,0.6878785278115525,0.4081620297596934,
 0.2386774492602104,0.07563500367783882,0.1864952314834999,0.5841114879771103,
 0.516953282016011,0.7296218947696367,0.23231960878028146,0.4036306195550885,
 0.6476704180283291,0.03653514361000809,0.3580253004452023,0.7345549641264884,
 0.8886592905454876,0.905531707759873,0.8739422185889646,0.7342836776713068,
 0.6068876054460652,0.8482118796017974,0.7660015033188912,0.31440608833711625,
 0.32183517781628224,0.6714137189819467,0.9875233902041595,
 0.38666111383356583,0.18304491034424808,0.8228849056528227,
 0.3550321154918693,0.2250739496006644,0.25193022320793235,0.535404618318473,
 0.28760891576403114,0.5287272009949183,0.20871325974089483,
 0.6722492983867541,0.3731399845258122,0.6496406639450164,0.8920734083186334,
 0.16012867804579645,0.44314502239892484,0.673440834060449,0.2896499067690155,
 0.3392013441293322,0.10514311735893533,0.8212366948925415,
 0.06554788525188016,0.025130003373570187,0.9212710021395109,
 0.1631156364391857,0.19718110676195966,0.7813468693614399,0.5915562350209113,
 0.42394788311775367,0.9067021799571732,0.9069991164564415,
 0.07914645792865915,0.931236235292113,0.5890124101809237,0.8395716650122349,
 0.6502656856252181,0.946919528308917,0.6243987867518883,0.2781276481446021,
 0.6513927323917962,0.3025395700260882,0.5699207210400051,0.7712837622696537,
 0.6049997756293707,0.4582064659958769,0.14680642697660717,0.8724366095994988,
 0.23370949713720213,0.31552983925034883,0.25921600042111437,
 0.9728555033453878,0.6827833536274922,0.9051014801980604,0.08834634645122008,
 0.6826646108239609,0.6622002665093931,0.009195415652322736,
 0.3089123498953281,0.3254078965577285,0.7305086398811333,0.5421386693282917,
 0.8566433554092513,0.9055689188654203,0.9486559707055124,0.3545269276435854,
 0.766682968432186,0.9620392170284494,0.3619991652657706,0.49296431877403546,
 0.7621396695197497,0.9468578813389602,0.8075237197853131,0.8550954717251879,
 0.8428299052030486,0.7530730834240518,0.5505045382343451,0.0475376939961496,
 0.9872776426861594,0.6940985820680948,0.4768524987453985,0.01311787515335916,
 0.7091620265655121,0.28903536643115557,0.31289040563639936,
 0.9704087739500122,0.19301075843581295,0.9533734275997394,
 0.11802496666007933,0.9324641261098138,0.9644294080576907,0.7797925027403158,
 0.3086916939946487,0.30491718877108986,0.9540680722132361,0.6146359582479363,
 0.25874411545273257,0.7418345649352653,0.7397688913355943,0.7923993934685063,
 0.00418125384238166,0.597270079136953,0.9769074131191452,0.3626200321220563,
 0.6924600275781482,0.01942865436002894,0.8659557270696686,0.4167134191757341,
 0.4215850017021434,0.7731416358863945,0.4400789113345741,0.7246522740455995,
 0.8362674704761076,0.8375917816897169,0.9018579534626774,0.18478741157045642,
 0.6979493522055786,0.35791453678264085,0.7403648582908798,0.5759319468658016,
 0.3921482549894584,0.9917782414265928,0.31511449007878645,0.2780985964361239,
 0.002294839638390722,0.32008720344532704,0.67405596154014,0.7455361283110387,
 0.4355528174090131,0.8631876913114575,0.19446041221390709,0.4454855567654945,
 0.9121963343469344,0.27829086522759017,0.5876939851615155,0.4282972200248021,
 0.8210063790840301,0.9452060777482921,0.9112194350821451,0.7949624161505491,
 0.4189594868958061,0.849162209663821,0.9811673701762842,0.6555264700606176,
 0.21244290955763556,0.5043479190378408,0.821740669843374,0.24146048087778216,
 0.3815302340535006,0.9410488162147813,0.8565792676804552,0.17049606952710228,
 0.7062074142112653,0.4883300323937795,0.7861117947472233,0.04370010990550788,
 0.6879714932792389,0.1091316381505616,0.06983947624075137,0.6930675123345181,
 0.9120488664689679,0.41445013111931994,0.6411513933661447,0.2996600140411083,
 0.09222004752917146,0.9132075515810026,0.28993630452224584,
 0.9948574962014841,0.5426123830550408,0.33971117051702926,
 0.09667792964193733,0.09960962402579432,0.4144676862402159,
 0.1475102548174143,0.40543836152248347,0.7754293824324786,
 0.46139227653599524,0.18771156450597515,0.8270032295148162,
 0.9193394448334488,0.05853226079653362,0.6034438096243769,0.4496427911239411,
 0.5139401978221779,0.9074887604244057,0.8196093338849958,0.35163760344318173,
 0.9525527505766542,0.6613562595765676,0.49304881905405495,0.5869722852607783,
 0.1889424619208886,0.4060674551376737,0.480059958390852,0.01823537335003389,
 0.36623995513253493,0.0027941605057553165,0.15547751279631095,
 0.27058245888072996,0.41908561664559996,0.9151046931794129,
 0.05279162349591504,0.10328491852407706,0.6794159469268402,
 0.6323523136843021,0.2687319186738615,0.03314945187125662,
 0.27546602589757574,0.18717293718350858,0.7190297565329509,
 0.21393508842312903,0.07568444203271807,0.5450165783623274,
 0.8947356516367309,0.18198827695619424,0.12548504627178814,
 0.4415474821960269,0.5791700646088405,0.08631956440731892,
 0.29910032871431413,0.1783679162476428,0.8026197830300855,
 0.16850585778445937,0.4237151371583767,0.705915696527564,0.815192167342658,
 0.21222840733952741,0.0594831206065346,0.9848218296046247,0.5872656432537957,
 0.5579167454117947,0.365142440041605,0.3638605278850874,0.9996910073894586,
 0.29086882209889775,0.416701186355217,0.670162791626442,0.6796016043461484,
 0.9906874848994724,0.16533431038007018,0.11377468492765819,0.694982816636251,
 0.9801806399980605,0.3788973492426597,0.5393360189595509,0.6549774347878884,
 0.09158631107847837,0.8655671648314968,0.6651304872300615,0.592911787112635,
 0.8365775734016587,0.6018356978948679,0.15823272192761195,
 0.46010511348508765,0.3305201643440334,0.49892430198943116,
 0.7757409305530903,0.9939678155107046,0.6509493814933837,0.41493947130002107,
 0.4697747026864132,0.4967380074224361,0.5281015859401608,0.8174450753481403,
 0.02275849898258786,0.6581475600199218,0.8130511722452455,0.596718887622439,
 0.4784424076200148,0.8538303074710367,0.8730831341592462,0.19216269181884416,
 0.256099919727507,0.9240300138915508,0.8485675134964394,0.48717455715902436,
 0.3531953893289552,0.3361343294578367,0.785681946434272,0.720388249870903,
 0.07753452969727281,0.1555459328023101,0.05064649478844141,
 0.5268405290492287,0.25529924452930663,0.6365163581212812,0.2952504666688107,
 0.44482244845272456,0.07444626305518764,0.8029437262663708,
 0.8412613526780512,0.657775559346198,3.776357975247002e-4,0.2968985044338901,
 0.33554361996290105,0.9279901161736194,0.36929267361361473,
 0.43543245993919266,0.7559504326154955,0.9864029718731078,0.9611389071933191,
 0.8782473702217993,0.21339278134023554,0.5834070166731048,0.9704521979050067,
 0.8113735128788813,0.8664246816730234,0.3013574022477128,0.9137322806791388,
 0.03832448860911708,0.5632049297398843,0.6182179957973329,
 0.031107906522426565,0.5839211631545522,0.1271115781662524,
 0.17070455999675072,0.7942187248089583,0.3025909840363892,0.4508360208592965,
 0.22484629238764642,0.8606058082919343,0.6795965577452565,0.4176085813792111,
 0.3270956508507502,0.18337221040674578,0.7992067217448013,0.5982210520656338,
 0.20724363754665787,0.6801754321153902,0.4802042815731854,
 0.04138766365505331,0.9328696705836241,0.2651656375393441,0.7441577800212861,
 0.28503815696498647,0.2975637772753894,0.8624741086728089,0.2526247105860411,
 0.6607788534112351,0.2801490109492363,0.22526829894301015,
 0.20399030007288643,0.6842702542229241,0.2106572259592876,0.2804810175246557,
 0.7391739538296298,0.7981661460604825,0.48006736678657236,0.5886468165272574,
 0.43127540605198456,0.36919978339544746,0.5687848602121558,
 0.9683019915596077,0.19598728785291042,0.7900987555535152,0.884480974992495,
 0.8990519278091617,0.5337105584868671,0.5871405635956419,0.9821106994767059,
 0.9895297774932701,0.4410478482407564,0.7854828611778849,0.3106898398379728,
 0.23593447708706805,0.7376008793419934,0.8685411345182158,0.5512630728053258,
 0.975432697107733,0.682986888251109,0.811148361774239,0.4546046470891293,
 0.06895710707750802,0.5971618493575231,0.02658389748146328,
 0.0350591050956689,0.9286931822499467,0.9589391880410862,0.12841567405901777,
 0.6367816666940949,0.47906928367509793,0.29701319821032435,
 0.9566424912901608,0.6082743743061892,0.8368970953506111,0.6078637992928149,
 0.34820740948990503,0.05650535848643168,0.6552641952284395,
 0.5212927644319589,0.523416364256161,0.6635812144259737,0.30095553013322673,
 0.19065078180235717,0.2886858939760836,0.5327813015067148,0.6807545556633098,
 0.6034196737121751,0.506135194679762,0.2336912786327543,0.9138325850214257,
 0.02356490418201851,0.3431713565864183,0.6082074635604034,0.813476043336248,
 0.258358489510114,0.9099125258814691,0.20512471161595625,
 0.005986253362540328,0.34654687280218,0.29335991553795415,0.7170525656983009,
 0.6539403848259991,0.2136741672976874,0.12617139131548372,
 0.46582933955599193,0.46662519273547876,0.6753708314424706,
 0.6280650472971246,0.9225903249684397,0.0332256751360942,0.22303333772512834,
 0.6578317578966282,0.9573737037705953,0.21782067086656653,0.7849100254004029,
 0.4605125363445952,0.348337784081988,0.0560234748502445,0.15280588289353425,
 0.5798519040371065,0.8121450636767497,0.5087629420071956,0.9007609795729855,
 0.5018611498344546,0.13604511032875188,0.2975920206240773,
 0.14419052646585162,0.08753706401319494,0.33243500642318247,
 0.6729581034965731,0.958920826199086,0.8786611906557308,0.7009655193612213,
 0.6498467013688235,0.8034585478184874,0.0498988375424545,0.2943434087900415,
 0.6989437552804733,0.23866270010914103,0.021572801153238896,
 0.9611290325165056,0.07215684019192048,0.09129862128985922,0.128140337490789,
 0.35139081760954305,0.4243285547578829,0.4669609835378913,0.9892509750009796,
 0.5054822217311332,0.9996134130277657,0.8530866761206507,0.7504612777479235,
 0.9690254291236302,0.7423795443858691,0.574599000577719,0.6469144072241516,
 0.6990440132712814,0.3496205107138636,0.9210452884307865,0.08852497328265119,
 0.8697395595297912,0.6787138056649803,0.038870854498544216,
 0.11682095682302696,0.06949656974459661,0.5829573016902677,
 0.13452991860357955,0.7614017212960471,0.7778501557309964,0.5464701833504368,
 0.5746779956568342,0.24776741229666654,0.702442703065088,0.18762686197895684,
 0.22048959501118015,0.6272987278278036,0.9268684007273893,
 0.13588615110250923,0.4328362608032742,0.8337936287088858,0.2584688254920402,
 0.4340239414628071,0.3504200841235442,0.12094982862568993,
 0.05691653094830862,0.45930971047488434,0.8246414870829025,
 0.4464623926080292,0.012127827484298193,0.041141125649692856,
 0.8502852311819549,0.07354925328061779,0.1766326060647072,0.4184265076073722,
 0.4719091411037287,0.03295377617122497,0.5503798428900282,0.9642373255731658,
 0.0028525637432365025,0.8033565604157878,0.6511222965904307,
 0.6658980838188961,0.15174562046302842,0.9497750677608466,0.9324011249179854,
 0.5902587409763744,0.3771564049662094,0.4980319247604248,0.9802867583864272,
 0.46200913452470294,0.8218766699877444,0.6137183824403427,0.1414711303166707,
 0.5129339535607914,0.4934615037997854,0.6096948302416098,0.3331764102917095,
 0.7439476606218483,0.5452179707648529,0.4469828796969749,0.9233100309296406,
 0.8133795093318318,0.23409550237294263,0.5356977120197729,
 0.03755588161765877,0.9080378769714543,0.8233788511988462,0.9618147273077932,
 0.25813885659717306,0.9236097617277712,0.42917456253358144,
 0.1274291682437797,0.45516083461374257,0.48120269714674235,
 0.7477158444862969,0.6202242843542625,0.2357523457778662,0.6154845576461986,
 0.047747050150684034,0.20843507162263963,0.2076093108872784,
 0.8385622702185076,0.8483105106430437,0.6363032775270457,0.8775185429792175,
 0.09981439645653079,0.9764515634767891,0.9853309898080174,0.607338622782239,
 0.055629987120948465,0.992030994826208,0.43663079778996317,
 0.45814543172278455,0.42492757836863315,0.17092218858545882,
 0.4305817792829467,0.729144350624092,0.37405549152052864,0.9271289277726266,
 0.9417672240822282,0.15563743614085412,0.9303644470874057,0.5781988557606745,
 0.5996840640271243,0.8180843535074832,0.9070432366074181,0.9487243593516976,
 0.8613942141123639,0.6663829376469379,0.9540092345442731,0.8634322082741671,
 0.0921891244984332,0.6329270646458016,0.19972244512251192,0.4217199013563748,
 0.7934453948404313,0.04545043308314323,0.3800185877130776,0.2015181764614662,
 0.880844928391544,0.008305618398241177,0.39111800847030376,
 0.5524796772134131,0.6171601018511894,0.11063139481375339,0.7742503691589172,
 0.3067770207460224,0.010825443959748049,0.2902316511472375,
 0.8522999236580449,0.3783124379796652,0.02024422973642992,0.3182901814967697,
 0.6460667467527269,0.846667648712222,0.4438123116623891,0.4316966963956914,
 0.6660163361944165,0.572239542961353,0.7294451276882187,0.5322387968474814,
 0.451394887340837,0.4379864147166209,0.5120365051798704,0.10006358638681456,
 0.8276674278869708,0.9615974692276229,0.8032093786100818,0.08016231905631888,
 0.3585912272248921,0.044613264483582604,0.279733617285441,
 0.03921414651706823,0.5584223079214586,0.16832066513533428,
 0.2047270279325445,0.1597173252417231,0.8693080335362963,0.06722473873180135,
 0.6878908140299197,0.37101659242455376,0.9977363696745669,
 0.23852046244764014,0.4402198915961274,0.47193347104186356,
 0.4427777538832348,0.17443901739511514,0.9486015561833246,0.9887328548109215,
 0.15504334474253723,0.6000484959689822,0.841784387306194,0.4331515051434056,
 0.12093326573260754,0.6260214389330135,0.04360541865517975,
 0.8662580223027744,0.5821793981258007,0.3016241542345943,0.1714356510160432,
 0.3619986430147826,0.7577605568785483,0.05696162320707032,0.9703554212722862,
 0.731817313219509,0.9544647665097288,0.6196892999289693,0.7515588836961182,
 0.6830589710601039,0.30691005330362064,0.28239084025083094,
 0.7717186107519748,0.9788600203606561,0.45539070300460105,
 0.40538132991219467,0.6317692428470668,0.8823762528846284,0.1374378529384339,
 0.053692542760912176,0.2591385414991254,0.27709063748552853,
 0.4596623052969362,0.8886072248457819,0.06671395998572316,
 0.48060856714557654,0.7035212974397081,0.6273954617583668,0.364285366638735,
 0.6161409555966003,0.8330250542565619,0.7624646102846875,0.9347194668091079,
 0.4404034469742486,0.03032136793946283,0.05896480931921566,0.999229293679575,
 0.477051420278518,0.31167956398866015,0.6193981308068282,0.8832130504650527,
 0.37158087712155874,0.8932115404960195,0.006830744730420335,
 0.1596141185497415,0.9496836277751586,0.8082449101177696,0.7212913524194964,
 0.4825219590813906,0.21826572485200535,0.9834895473552832,0.543821000883725,
 0.5461484962074681,0.5918542485641375,0.7686912888358177,0.5360549044659365,
 0.6156941362970996,0.45594830463194647,0.5220367942465814,
 0.03096906581906944,0.858651171975491,0.9472020445078564,0.33077694120358225,
 0.3527744617063482,0.23153301733325815,0.8596821004455374,
 0.011145952882180121,0.209813838707547,0.3473850380074105,0.8080398480241373,
 0.33583646749411,0.15442232158080693,0.7578403374798923,0.5064963168714207,
 0.889026377526609,0.028921977705547386,0.11831856228499638,
 0.5231046175081838,0.3095082625460984,0.7761534652840142,0.2764325039373742,
 0.4472957215427651,0.3376791999394111,0.96505000193797,0.21339573849660476,
 0.2986053652388999,0.6504254774825434,0.13105374184142726,0.4837263691209577,
 0.4805486841105404,0.839032922490031,0.2666883076192518,0.8733981706187396,
 0.12690843051539158,0.43252818188045006,0.06335570754120778,
 0.5140053276251584,0.6606473113613386,0.8826866444946811,0.3785896020059908,
 0.0011590068895536154,0.6000509742351583,0.006068361523146937,
 0.024796581713269927,0.5244580088637232,0.43996403258115824,
 0.8434830249611917,0.22567009322951392,0.3194533915439218,
 0.09875368807675255,0.18137663528670145,0.3783002040571245,
 0.9374032622822281,0.8380661773396845,0.09470191530651972,
 0.10670133790317671,0.4116471509717411,0.49866015205710745,
 0.9544631649233718,0.7221829309872687,0.09739164966486125,0.7722113291289774,
 0.6408828044823426,0.49691759228430965,0.8550327972955379,0.818725708162184,
 0.9140985857798215,0.6585416044229102,0.8202986984634703,0.9016873967594858,
 0.07235797590243054,0.9307746112043824,0.5380467271933586,0.6456547315626371,
 0.28761762979183414,0.8332636178258771,0.5316482263249812,0.8370044750543972,
 0.7716505362593276,0.2498127509580257,0.09432699977045489,0.922942655756671,
 0.08514200506275715,0.33397138705485363,0.7550871157682029,
 0.8271106284696609,0.2087118866346308,0.5781784188243252,0.2644535347585495,
 0.5235804427158652,0.030168374075498328,0.9219706770296606,
 0.23453690751091594,0.323245889960888,0.8680072814813831,0.5913638042087591,
 0.7616862866868556,0.45535491154468744,0.9674867893259451,0.2364891231925691,
 0.8811048558586703,0.5226478544423213,0.26388863097565374,0.9011753596042597,
 0.40929267183784446,0.7155799088894806,0.12336594072284479,
 0.9313299472307408,0.5937419826471001,0.6972028511794032,0.47552477099597745,
 0.9516613521655573,0.4137899512918981,0.1255772260490171,0.5662991729659995,
 0.8335822627486875,0.7683253002962211,0.3404046973697016,0.8248000193757875,
 0.2975367263231433,0.33547232784865444,0.6392715742623409,0.6127390860869263,
 0.45228486375523724,0.4339121803124467,0.10654525041418661,
 0.02752534111236904,0.9226506175501211,0.4679687665048021,0.5599798726013545,
 0.30695717611499085,0.11554828734776978,0.12536664821921395,
 0.31142567991375936,0.5368598527300114,0.7038826312389032,0.5072791823472251,
 0.3169247649783802,0.067979552513734,0.17075252804948482,0.2022057991451518,
 0.1828445411680637,0.19778716845236322,0.051614441323380666,
 0.2665625942891091,0.7206711640978785,0.6571083479337146,0.33134790125200364,
 0.2662384550488953,0.7614804099450698,0.21187022851082238,
 0.24977349893876388,0.9092585792467268,0.01872811899514415,
 0.49467323497246996,0.25250347259495876,0.46025220117210053,
 0.8665673581459064,0.38417302467811354,0.21072482858518682,
 0.031061864687281227,0.20505555309947732,0.4651778715911543,
 0.17428228343375363,0.8210144133778337,0.1687274537503085,0.0995559013617816,
 0.13664916093886448,0.19303589119909947,0.906620308940086,0.9502057591949167,
 0.36044921492275683,0.37331912542004375,0.6349402104099715,0.578698042932076,
 0.19614014970734028,0.658645304232932,0.9809986352997108,0.8679369548080209,
 0.9835074568275513,0.08928391100665556,0.6710651341798878,0.6613373076986365,
 0.7379167654927183,0.1993849965708615,0.8391947225120431,0.6634034137432187,
 0.4630102789297855,0.6140560830951324,0.3968318472344772,0.2249265819431796,
 0.038488060633222076,0.297385675993199,0.578203203558715,0.5990171131809537,
 0.003955690309302207,0.5650509360939008,0.013245227546374028,
 0.25881403028347283,0.6106188459448657,0.2905240243261663,
 0.33201339789422135,0.38116097529411497,0.8163590638154116,
 0.3067644621427418,0.13209802600326903,0.7072801845415296,0.9591423008680862,
 0.9890688498007989,0.4716097063493113,0.05163270671316278,
 0.18705833741608224,0.48319897262043454,0.15303712628541444,
 0.4137450076145114,0.876601381668295,0.12119369849648354,0.7361501786571238,
 0.5126299737953228,0.17321353841193743,0.29134977345139956,
 0.3505927207824292,0.2020089631200478,0.7084807262187248,0.9393286024708676,
 0.45742247408039305,0.219991162175718,0.6583628878187269,0.9530802058566453,
 0.4142382173560064,0.15898299889252043,0.2649509682470187,
 0.11254262333207077,0.638575613066956,0.830526051053981,0.37673558641118987,
 0.6427003816310679,0.13914964822789733,0.3355121372785086,0.7989227281171182,
 0.8553366551987401,0.34130892043371175,0.31183883168484383,
 0.8239950605319386,0.5425926952410189,0.3764690875617883,0.6691144094436154,
 0.38433212327479194,0.23429926028643677,0.0946444769645951,
 0.8199855459739387,0.13162409033409772,0.5742787383696958,0.7946123574439737,
 0.0027863978266909895,0.5317555091088337,0.2858337947373857,
 0.9431967155517387,0.8828857247956262,0.2804291844677367,0.3414499321861918,
 0.10718030861571504,0.2787087468445214,0.08063869818875624,
 0.7191807428192809,0.1903589369561084,0.02098472011670982,
 0.33699210095286214,0.5406199623860463,0.4456917043051739,0.7852145452517251,
 0.5684495750775564,0.3755843877293803,0.10041308965801088,0.7725812236970555,
 0.472636635589204,0.5727134603198566,0.5131569557305836,0.9503124892533283,
 0.9544929392405822,0.6386578002903814,0.05221342295537856,0.9967237694843041,
 0.5316560500364089,0.9851897548252844,0.6090116942902285,0.07506634623193564,
 0.5963428791613807,0.2848699150698134,0.4296045103240047,0.15776798433922812,
 0.73175294504627,0.703871956198394,0.48712318488277107,0.12375613831687249,
 0.9326041675465955,0.7748138707055708,0.6332096093834239,0.7668654476728318,
 0.5653190274969901,0.9226639038930248,0.5583827982275515,0.7397988612855645,
 0.6798964532424416,0.07559269771649413,0.38315917492993945,
 0.9626781225502351,0.7820993250636337,0.191875444034452,0.21634339607695519,
 0.39102096942345177,0.9896948651700085,0.9218029528583059,0.5978605047028928,
 0.8479391166341422,0.3818636905067829,0.3021821640798603,0.8210035990118792,
 0.7104675090311879,0.37447272728150727,0.12096897764193204,
 0.30491787977028495,0.10637116630035304,0.8169639283425216,
 0.12051473561425141,0.6976724460686295,0.6149156885803398,0.971433220760495,
 0.7840932130448729,0.6601605329684823,0.7868778903945051,0.8252176885297052,
 0.4882999056410434,0.33569090434602256,0.4907228763867415,
 0.37715722544896324,0.9182550553083224,0.4440517357743191,0.268276060115453,
 0.7399112466844343,0.8135290378196124,0.5557195359141953,0.72223973344292,
 0.72345362271541,0.7840534638002509,0.859123605694602,0.913912499965874,
 0.4328761883932528,0.7452174311751057,0.027407624249494167,
 0.23583594736703406,0.8928715019492918,0.5067985364639555,0.626516519616882,
 0.7422073844029458,0.5392319113500292,0.3594225854195172,0.8299437522957515,
 0.38990969789047114,0.004458328821497437,0.4280254726805248,
 0.28323306298792494,0.46466033755654224,0.7811670748110413,
 0.8485997488568675,0.947639342137832,0.1213542781046607,0.2992263845455724,
 0.5457017945108281,0.7037458399452676,0.11978300190697722,0.7805871301538421,
 0.017101792392376414,0.6419352557537116,0.8157545385867664,
 0.08489880725119614,0.23969408106082413,0.6496661467048168,0.505875270843089,
 0.9837529315259697,0.9041345670238058,0.08196879765951703,0.6514872987911399,
 0.6481464394043502,0.749327391216194,0.7529274023823862,0.7307783213049196,
 0.9458952330018453,0.6525085599369387,0.4022197995113519,
 0.025139027167282668,0.20167418543895754,0.18327428854579564,
 0.2495075452051092,0.44466495384439575,0.583208564603565,0.5008970221737993,
 0.1561290382988908,0.03442444676119616,0.4900883517274208,
 0.021408539989677844,0.2549477210942058,0.3277529098549682,
 0.4194977864542686,0.1974420509484046,0.14447909599558417,0.1718926319009204,
 0.3507475300335994,0.6323483902015876,0.6215719839852402,0.9698091989191823,
 0.4402824029075607,0.32989907432803434,0.7046937478862776,0.7968106209021819,
 0.4040901478992249,0.5640604819878338,0.9483729061543809,0.5984257854342633,
 0.3144012647342366,0.2978476523603527,0.48542777863206577,0.1799248545132731,
 0.07150200247082228,0.6387098860147141,0.94292069001198,0.6841583926651257,
 0.21415310676456178,0.5183926195676838,0.38187237035888744,
 0.050535719023954706,0.2537355516487184,0.5102153735149867,
 0.24806605902893197,0.4098374418234463,0.09880357584889421,
 0.22076514698197747,0.9100287185634078,0.6503685611330369,0.5679813314081058,
 0.21393177105056205,0.26953416741881164,0.42960121946410657,
 0.7198906572315911,0.6134437218917823,0.6536723002795137,0.4414472575143724,
 0.39472602120738454,0.4565989426366446,0.8936299541599143,0.3563565999341618,
 0.40877894123495273,0.2443282330035852,0.8222005482482299,0.2173434892007231,
 0.04513252390035061,0.736008814760574,0.22198915494391402,0.5334526063005552,
 0.4000562451271483,0.9893440099021813,0.4391046576067067,0.04150554892031477,
 0.31506407878798826,0.1901004379617307,0.6325381990954763,0.8476449359412964,
 0.7110212265866824,0.6675830405866595,0.017216353078632674,
 0.08074949879350479,0.6480909547609356,0.24064374196388139,
 0.021361223312257205,0.6911489984322207,0.46946439196084844,
 0.05154129985075018,0.508782767722562,0.2668050222321572,0.3936759600071702,
 0.1588776346313071,0.11761845542312621,0.28395394334517976,
 0.6854550755102222,0.5593730679102,0.24000937106049303,0.11187553723515642,
 0.16732141761640518,0.6723467613823422,0.1248960459049715,0.5166616952982116,
 0.6784972935373574,0.020033549007734686,0.49597982558923015,
 0.8289783044863462,0.4039511394262554,0.4864090029802285,0.3261624079500691,
 0.592064648062577,0.20822429665984477,0.5080811662387041,0.47450945336668193,
 0.27656232833315464,0.2800887839753924,0.8231066337858249,
 0.17893362988974681,0.10071859153004392,0.1099685215331413,
 0.2526563064333708,0.3031005917633163,0.18320508987278705,0.8659640421525953,
 0.6019588196350071,0.2972806196495138,0.19095262626200205,0.4538931571575062,
 0.6972212432203428,0.5359695877884529,0.857266847211558,0.5426351888661116,
 0.2738006168154017,0.8521182018918697,0.16123621774461605,0.3915615114551815,
 0.5007140068093012,0.4077101843093933,0.8999905790315542,0.0905304757220522,
 0.8061590654577708,0.23402419334505276,0.6212160175805881,0.6323876387883415,
 0.3200058889685399,0.16680214170342844,0.7187385188252611,0.4728386790780099,
 0.43588600717531834,0.39976121884107185,0.07150940635167424,
 0.38479211891180054,0.9705017370253326,0.6874020812243167,0.239756286009429,
 0.9896977820123581,0.5233951448000393,0.757725806863736,0.31741045556081127,
 0.6525222853795073,0.34964349919231796,0.28624041190842875,
 0.32209468784130346,0.17496231484783697,0.06846486785758277,
 0.48310838607163387,0.48434542630348787,0.9334895340702462,
 0.8083228518491525,0.5702041504771591,0.6559296982894322,0.12823115265019547,
 0.2463494731320014,0.045082806385939866,0.7718293054510798,
 0.3475205767145275,0.14877367548692622,0.0061761097867936154,
 0.33450524400927495,0.17978239864029089,0.7984577423708321,
 0.6928892514760809,0.5254936599968416,0.3899024482384277,0.3786342867956869,
 0.31352822770457245,0.7421197560965146,0.9744052903251835,0.2557519302236738,
 0.8568968821306302,0.9709132585280738,0.9203235912507881,0.5571128327222756,
 0.9650268502218243,0.50477173545795,0.019299737857503052,0.8214864433651102,
 0.664215000679398,0.37400201156706203,0.6025900967008431,0.4544406966455653,
 0.8008660215287473,0.6825499458942682,0.03335513022052272,0.3624182359483832,
 0.6407240819647863,0.9101351772809771,0.9927738871608092,0.19468093672038567,
 0.2528790402271146,0.41730895491897824,0.41589088367515126,
 0.12176417428533326,0.7827589198096578,0.9088009464804946,0.7343855944815662,
 0.3084047605716096,0.3073241170065367,0.7791863253228206,0.725576504921371,
 0.7295198851440999,0.10235234492481637,0.7979629146568686,0.4136901175095222,
 0.2804953238978365,0.03943923766797497,0.05232823998699032,
 0.2731455686574279,0.6963303743326106,0.8398107110636507,0.1100922920435895,
 0.7352774558544547,0.47104460513462487,0.953311129241803,0.2367118315285961,
 0.46017413974270716,0.040376845710213494,0.7478299924766352,
 0.9600775073690424,0.39900777003795473,0.9129332553278584,0.8227315465701638,
 0.16317909654518103,0.0674829902882037,0.45640505279537114,
 0.5247946347814859,0.5847113656978464,0.9843411822745857,0.6825876756780836,
 0.4137104798451876,0.2687614498316381,0.6479064872572151,0.6593860241324557,
 0.5743912653879839,0.026212027760657453,0.9546040427607229,
 0.25185017136549626,0.4469530992252866,0.42471485357419514,
 0.7024044023538889,0.12269593040671678,0.5998705388307218,0.2744054798692883,
 0.1822132087088837,0.6882951079153768,0.10099373845032844,
 0.43718708807342677,0.12761322624175808,0.967701357545131,0.6204842751522611,
 0.2924835211276102,0.9028745880078972,0.47860660855471915,0.8125932979180821,
 0.9590046946609174,0.03383733182180604,0.12324333927276032,
 0.8854233449592064,0.07223534718985225,0.26277437497020606,
 0.42557945730331104,0.8098537986204536,0.4306829151010103,0.2931308997090928,
 0.3009814144982188,0.4676229627239762,0.8617710774140883,0.9007595432005413,
 0.3322975836701836,0.17261642771437746,0.1305585411105611,0.5432020120778551,
 0.5863147026459412,0.5058678972372399,0.66262061839405,0.2667113949412828,
 0.2824884004262682,0.9771826488155791,0.14156194002965627,
 0.09556724717766907,0.20579313576447855,0.7839778990280056,0.546677865949666,
 0.6277864327247169,0.7692270341930603,0.6246302858320206,0.5693721300513078,
 0.7287906044154705,0.12786879440143295,0.857393191535306,0.4299076502000465,
 0.8505726658204154,0.12602477136250156,0.4580927447408544,0.9335529035750998,
 0.4276158960463031,0.3000141118588925,0.3891858950092062,0.8659169216985163,
 0.005564477006880897,0.5298170698409198,0.29224969603332007,
 0.5708988910543836,0.24607743438370144,0.23757064252266813,
 0.9801652171235194,0.9820557247793728,0.7896831526409261,0.7275299756030479,
 0.7850748417200297,0.44545761076195145,0.8768556369796777,0.7130791403447716,
 0.27292270459583123,0.6109769034198678,0.3263414885882072,0.4456825308496083,
 0.39650078860475,0.04632988426805362,0.45019000188399827,0.08820364302227479,
 0.12078951386643222,0.7430378114513437,0.10453691033739188,
 0.9236873658311311,0.9530068733667374,0.8039806587844075,0.23498070248617609,
 0.5238256820327132,0.8938180230946431,0.2840820269064609,0.26619891639123083,
 0.8170069627834062,0.6537975363461964,0.1452683440445115,0.14516564980384195,
 0.1242363263043379,0.7588282841900765,0.8911074981302989,0.688478985747153,
 0.6877537293903988,0.7878676529413958,0.7600808510627298,0.9156298370588154,
 0.2396589285340155,0.9227306479695481,0.40584711941392115,
 0.06577578422443231,0.6071291537549866,0.6939834154744499,0.9945466043599339,
 0.5078579674741411,0.9530232364921283,0.40707681287200315,0.7328242869712255,
 0.118422940741028,0.9259494627572294,0.375823137363964,0.01325330456410101,
 0.44655033891123486,0.8087973716275971,0.19534896047797512,
 0.8436643436729768,0.5238404502564465,0.40299202447033433,0.3591105609247115,
 0.10152886371249248,0.3540953180967076,0.5974291502503011,0.7879426603320134,
 0.3548805425368122,0.3193125519753395,0.9885808946382039,0.8636418978883625,
 0.8171325971667734,0.6068999425007288,0.23006870396010448,0.9818704135822955,
 0.6807752135767092,0.05149859040055893,0.858657767141591,0.20362965436966485,
 0.5240812887741746,0.5755261513709757,0.47407254219531947,
 0.024257810358087717,0.7917278832916155,0.43931641729788407,
 0.44716977575226047,0.10150395039895943,0.9902869383915746,
 0.43496398299695493,0.3743683896183454,0.8181663271187379,0.9415699817057089,
 0.6566861906953084,0.8222635975915831,0.5631483141558847,0.464157051157436,
 0.1453907646307646,0.8371612885823965,0.6236302357613482,0.6596580734427545,
 0.535206760656862,0.4192103367367147,0.1382669645489919,0.43200946318222755,
 0.160672060727457,0.6460988923830209,0.9732439347367131,0.7950176990363025,
 0.01610759369678727,0.1787676570970147,0.6987520493460863,0.3075322499716937,
 0.5238618016111655,0.08340284979858836,0.265271662549626,0.5951671086954878,
 0.807561358754522,0.18015278253826572,0.5474178023664705,0.03646180452169534,
 0.398955200442047,0.09131198357358306,0.07770986200816354,0.3403785176364983,
 0.4053820391993874,0.76088479019749,0.40933495822964305,0.509823889134149,
 0.3233449292527919,0.07627467078275352,0.779779461482065,0.6710216128794048,
 0.5898058341183402,0.31308558488576543,0.9691252183169018,0.5825168897836973,
 0.764967376872032,0.1053992259999903,0.5810980692738981,0.11582024139651448,
 0.3135291752637812,0.7787962803802082,0.27388198200366376,
 0.18709710837299032,0.7420909462489022,0.3625051118459318,0.887346490156412,
 0.24155972955072025,0.954095087690381,0.5624445795137565,0.44491187681334954,
 0.9169341942729407,0.022935452339115203,0.27392825087760575,
 0.9123614752515078,0.4251113174027922,0.1939124343115124,0.30558303754157057,
 0.17284548391898524,0.45718199733359466,0.18128594394648068,
 0.14195823308996147,0.054210089132619266,0.6875559005661969,
 0.21648130239819963,0.9331173406763078,0.9126756385055036,0.783815226540936,
 0.8400740820796107,0.3076931319721795,0.5768518379483087,0.3003119435145152,
 0.46996130076399434,0.21482538027545006,0.13851118939926432,
 0.35532948214869975,0.6618612213177889,0.21974599011565965,
 0.7638774681321372,0.15159646187683784,0.26223916475723985,
 0.6144199256994153,0.5948908978629843,0.09387505503951787,0.5701416159908753,
 0.40661796082846313,0.6282967333746139,0.2484572019707465,
 0.16411227841147102,0.5338483071024777,0.9889235392575942,
 0.042506360155540435,0.9427025604760402,0.48473226242135503,
 0.5335797839902906,0.33198558580555193,0.1852787598244966,
 0.46955549726142887,0.23741892509203433,0.16587757814789095,
 0.9648714068589045,0.6525555995166419,0.25223965041538055,
 0.21733463477585124,0.6989765175150567,0.28750246672867374,
 0.12684473791635464,0.31902633556031124,0.8668901700613475,0.676183244340361,
 0.046288176122249336,0.27643601465255097,0.9754548260595207,
 0.6653748697832678,0.06238076839807816,0.6472594857091578,0.5108393718126769,
 0.8754126104877368,0.9192933095277291,0.993590368193928,0.8531852449419358,
 0.7747738010969185,0.18616726634626313,0.19494761715206854,
 0.3646665289594627,0.4920510252956912,0.6460559719125436,0.5841144793607016,
 0.8786797723811928,0.9020347806011951,0.833598884475961,0.6573320075162795,
 0.1717362364596453,0.33748277452656206,0.2673496154173859,0.8048760138284343,
 0.46928992290845384,0.30207728940783174,0.19595311775536084,
 0.3446506869252536,0.7157646850137341,0.07774447949628249,0.1494875729318702,
 0.2086890735967888,0.7876756956757839,0.968454710587356,0.47228670531607253,
 0.40239504774897394,0.6886329771029404,0.5033540098310069,0.4541226739669063,
 0.07740513137276206,0.11591983568925068,0.9034882450512768,
 0.9086220204754236,0.2664485139443361,0.7474476929673622,0.2213279236389909,
 0.10343867057790623,0.6915177525583914,0.961631011216322,0.8365467721458101,
 0.08898880849135105,0.19733589775729787,0.0761155580364794,
 0.5937526192585671,0.4444934548055607,0.5340327158270647,0.3945963957016887,
 0.5451065987114819,0.5532160586430321,0.204639403293143,0.8601265440107503,
 0.2506443269857994,0.513286046307514,0.23368409595653072,
 0.022835931156137912,0.20077145776619,0.5203894332831469,0.37802645291519943,
 0.8362244476572979,0.13368174286272116,0.3776634676705496,0.2060963756121168,
 0.5551850431691862,0.5529147976564639,0.018505571073401228,
 0.007892440154357083,0.12146143964827738,0.1731489186588695,
 0.519426950405242,0.24378996607584102,0.21977305872957587,0.6909737351326791,
 0.9119462424989331,0.1729058085230234,0.20452252787505287,0.4072345367446165,
 0.6235755853225624,0.028115241441008854,0.9106000715865672,
 0.9068156355908107,0.03799645751447356,0.9541395482372254,0.8215553844297381,
 0.9347956443129758,0.6065635364872973,0.2976425035505236,0.8253158664742228,
 0.31899433268844846,0.6772621329823195,0.7890974458527118,0.90702027176449,
 0.8445405475233219,0.8912399375040413,0.316334585259199,0.24241363000396166,
 0.6114476365830588,0.21800334527877163,0.20329840524732912,
 0.0994558035869143,0.6680633103150512,0.15730887186909526,
 0.33204173428061123,0.9302289734161278,0.7559079420642055,
 0.13180329271683489,0.6952583397790985,0.30502459548112093,
 0.3699159302821262,0.017966498410442622,0.04452036312213914,
 0.226672734393204,0.8025763975024636,0.13954934388281492,0.6774089432390503,
 0.3766602983789743,0.8485946583548887,0.1614404643698203,0.473591196348112,
 0.5454025128913974,0.30973980857715777,0.07788324727516582,
 0.6230095737932132,0.9600586955282513,0.49482014259265716,
 0.17074526250904876,0.18715885810755306,0.06704356841763448,
 0.5773928015970817,0.4794398049524673,0.7682064814104806,0.22529353441947586,
 0.9332763802197703,0.42632139219792053,0.06027768693643576,
 0.5894134143638748,0.5992159606768029,0.515414020369815,0.6277551485291301,
 0.19819406192994526,0.4966362366542173,0.13072822622136704,
 0.0724715704233202,0.25999283185492095,0.04024160863774562,
 0.9999718873021661,0.5933559298801686,0.1284265429534075,0.7384499547605465,
 0.9902341321961887,0.8251014900833158,0.013560268287596822,
 0.4813321471688802,0.4475477573960265,0.8778039325300236,0.4155818234557471,
 0.6121542612586156,0.048933933278058506,0.7360768687764601,
 0.3836332993635392,0.31496911507162106,0.2597296936612786,0.944568761526551,
 0.37390716837712945,0.02549646704316022,0.31802754594620686,
 0.25795652935675495,0.6946315473769915,0.4634924068815607,
 0.42187985802617367,0.31521447130738556,0.6014467730162694,
 0.03695714225573221,0.6644679994569865,0.40522651697999745,
 0.6973855745800672,0.9819118060260599,0.06553493085719309,
 0.03159821147451547,0.6616051500278243,0.8302577897000428,0.9921184186826981,
 0.30358425135153666,0.16121584094830754,0.1987757353716787,
 0.8099877857053305,0.9124996684640907,0.39976322658202257,0.662440297734205,
 0.4730129195407772,0.28519019831814785,0.08245464580825301,
 0.4296211954641739,0.9671133127289918,0.7266062172261235,0.32431809826379976,
 0.04047811127202128,0.30320352089585245,0.2996031787001402,
 0.8066854489316008,0.5437507423074992,0.23647536125504331,0.9863466104948484,
 0.14281985450521928,0.4874061486937519,0.8399526302194057,0.317841405152011,
 0.07755768646562133,0.7888318214178468,0.6209097181354224,0.970928945751489,
 0.7913218778216836,0.7420090940424713,0.17610947592522708,
 0.15267473349219507,0.6670335042383823,0.6495295247847028,0.4117299382369318,
 0.20993686630202668,0.3734028389702695,0.43818426565740376,
 0.5366730478933299,0.5320182844380065,0.82866795818839,0.8426610426502099,
 0.26224645008067415,0.5139170464036835,0.6947191779975279,0.9225041210917299,
 0.2424165590977052,0.6253572336661132,0.6477319871231564,0.9875236763240524,
 0.214750880645036,0.017678400305622954,0.7466193100131066,0.8139780854600607,
 0.9925081438010183,0.6389930779072249,0.08367893817378869,
 0.33954448696093864,0.801012361212165,0.6468037467652008,0.8386791728502234,
 0.026912065295650822,0.8176350711570471,0.060500527870101295,
 0.09637233571998094,0.3726885792062812,0.4523011752146312,0.3982654602619359,
 0.27733665956925346,0.4494717081452625,0.059202743320216245,
 0.1057757585733361,0.8271733249185109,0.4799751093796655,0.8384152094518486,
 0.6783261770804636,0.8117336880347558,0.8827347414765476,0.5181452701824213,
 0.48648834169111577,0.057333214739099025,0.6962724011119087,
 0.3288498803507798,0.7239589011789526,0.2213108058401,0.4203565095195897,
 0.6034361710114531,0.8175119854904797,0.6013934695392443,0.7434907168249295,
 0.6244002884346443,0.7773116742119658,0.8363888445842962,0.3903157585383342,
 0.45671090010592663,0.5769502203875945,0.1135574005761637,0.5833981680690191,
 0.9490325499336878,0.8379501477591337,0.5934940129034841,0.9034278022929038,
 0.1538274662049911,0.971348224756154,0.7460205242510216,0.6150191353240516,
 0.21947118827891865,0.16370482845737544,0.44253208870143146,
 0.10369280711088724,0.5541038045222529,0.09195188681569932,
 0.21451414776170386,0.1309380321534881,0.424432090021468,0.6556911911147024,
 0.41952542630828094,0.4136234130503873,0.2563447621914938,0.8042217231480765,
 0.8323474127098994,0.30806591456934684,0.21031612029961622,
 0.9134151656473823,0.15798161373629704,0.7126383324578858,
 0.07575836947743331,0.7654435298625393,0.1890343510885868,0.7637657202041132,
 0.5335384198991745,0.9020482159764457,0.3972953415445353,0.8316291556161888,
 0.61383229936862,0.37574369464914537,0.3450927491809521,0.29327941621696074,
 0.6810834989485144,0.8454357803438362,0.34717674121729714,0.5565285348403243,
 0.41265318024011366,0.6830478199111835,0.20936578187332766,
 0.8701059460960504,0.029048750937145984,0.08029212085564863,
 0.6491513973485148,0.9166362245114367,0.5840207600848982,0.4188309002315008,
 0.15465080718789614,0.9729672462656117,0.4785967043772503,0.1038948357688837,
 0.7902052865476635,0.3989646312346826,0.2500726588855393,0.38331404248847356,
 0.7311754539933271,0.934489768274779,0.5710511345249737,0.6284664115939166,
 0.11498639719996073,0.35283806690156694,0.34208314419798214,
 0.9094890893793686,0.1391209077210641,0.18633673068554224,
 0.17824650146516507,0.2619479805672489,0.16719228386531193,
 0.07270996194254076,0.10139507300070205,0.03247378524287925,
 0.7026997475819087,0.08106985060680505,0.5751181885649875,
 0.012150805377141527,0.6473321184592226,0.39345972858922684,
 0.1347735894592218,0.29181772790866334,0.18396234016931257,
 0.3500139366430802,0.2233630427624076,0.3190895892659147,
 0.015864566881219955,0.2778258264118243,0.5510652529563055,
 0.9157800732673897,0.28752637014913385,0.41730366303412236,
 0.8453792610460155,0.927391975926249,0.34861118982729233,0.3829198207043605,
 0.7462536875755214,0.11115543420277985,0.03271238888775774,
 0.5758963022201001,0.18995245125924076,0.7169489975179152,0.7623420202486912,
 0.8904051403411558,0.025151932855094872,0.4665380434778632,
 0.5350956985743965,0.6328572377202928,0.7843266234976259,0.46691076991482694,
 0.027456829143198136,0.7734098917844874,0.7866735953491831,
 0.8729706727537602,0.8400495863208856,0.9684036165882534,0.8150565484516646,
 0.3224001827627999,0.43540823649656757,0.9136814560455454,0.8525156771979567,
 0.9002012307759225,0.5678714263953708,0.5785297984978024,0.27877002263931583,
 0.06544550984708208,0.20908542394142393,0.4610624442899959,
 0.20383152561814488,0.6280356372422724,0.7453071696257291,
 0.42863444694591746,0.8428749000345801,0.7268088935707505,0.0955113754430672,
 0.40420561511194353,0.17046310009406074,0.9057552579592515,
 0.24429430598732949,0.46148300615554005,0.32195770045223504,
 0.7080222530981467,0.9131176402288848,0.6829447473883168,0.6560507633006858,
 0.43604840821625945,0.8557162228391504,0.6019934191997909,
 0.40623094601436405,0.43166103779534004,0.16627741392429163,
 0.3278483756422944,0.6257543009774045,0.5666058837129638,0.2828915085075375,
 0.17626594288939812,0.713051268238784,0.7584705647126684,0.9983913949501194,
 0.7985660440151564,0.15474257195586893,0.5458802491291268,
 0.026404905480857055,0.44903409208947687,0.07358895406637656,
 0.1289702671275581,0.7924553783225773,0.4897698015502111,0.5855720471069353,
 0.002356875520931667,0.9276054759222492,0.25327965829073074,
 0.3149736072818642,0.06314708372655331,0.6441041461272997,0.3401227595871633,
 0.3346270243031664,0.9024863711849145,0.1646956047176249,0.08627414588061721,
 0.8558085301870827,0.4989550019678163,0.015949892072413707,
 0.6990052389043943,0.015500792185709877,0.7289389261743741,
 0.13202431352619814,0.6809045099768058,0.8828003198280088,0.4163996387796651,
 0.16269482508947997,0.9676842745073673,0.16182127919500222,0.871307319809346,
 0.4876552657256098,0.08574847096462868,0.9171317454953029,0.6208300474847622,
 0.0791942211244081,0.5466886395583231,0.8845350126955748,0.2657038122670977,
 0.5533962951915784,0.3307016439356798,0.05027292585643439,
 0.33398122884186965,0.33643624809421746,0.6406485606854037,
 0.9924916014556942,0.40321188656484086,0.6054232786617925,0.5852702334779997,
 0.6745416007120152,0.8893956388740649,0.858668963019026,0.7075279848434696,
 0.4919673177187518,0.1854033718850956,0.9729157714942752,0.18909444568303035,
 0.43674016397301685,0.4359525840889833,0.37148315719765834,
 0.16289027268350031,0.6710190247858325,0.42190921360438705,0.835262162542749,
 0.6781676200232193,0.91783704993194,0.7849330915121328,0.9988763389385144,
 0.6572556503057839,0.7560419627412646,0.5813304332526175,0.04692954658482007,
 0.8160613307259263,0.02082834844593573,0.8991765472954139,0.7355413246819976,
 0.42825875887464826,0.035651375738232516,0.28266109035623843,
 0.3989601997657515,0.8158448206591736,0.969523145823239,0.6802564170919934,
 0.8389636208664677,0.7712939938984368,0.0023764290410406286,
 0.8262522693549903,0.9120398855359053,0.9379971181631531,0.8230615800194891,
 0.05316057589666001,0.8621390563881601,0.9557978858549475,0.5045602505280363,
 0.017871182946751163,0.2245867339824914,0.6637967611647985,
 0.45375807251946587,0.9319512487563589,0.47964917895879144,
 0.016781826016642243,0.4698575645290306,0.5765978389266846,
 0.11161860146831748,0.5995783241020357,0.5857676958943896,
 0.47129867412573967,0.16106445667519798,0.6365472931209599,
 0.29118418878877617,0.7843180171844641,0.38984524335862347,
 0.17716995123367552,0.7604780389070744,0.7552225878991231,
 0.31292119161143706,0.8694855965731381,0.7991467712318496,0.7126531741471123,
 0.5909174522084731,0.5339132354547949,0.45023510322435123,0.1991228488301744,
 0.8269280248446076,0.07809341177271045,0.08286509056006741,
 0.18924332317677195,0.9905807468321179,0.1158717345985375,0.9472090163364432,
 0.3801818981526266,0.9102645689137605,0.37401836839408964,
 0.14897194694541027,0.37748806050349515,0.21132351615806466,
 0.5913574741255108,0.6240964885705362,0.3061363699993829,0.44458236269714935,
 0.5526072985070829,0.9799064861467569,0.6676803431754446,0.9885448725907161,
 0.42135685639811926,0.9775206374646787,0.3873875556566583,0.7323704414936292,
 0.02594080738538762,0.9602927756077224,0.6485156046249477,0.8507332004338206,
 0.8382718120240154,0.7129540083161873,0.044835344558300516,
 0.6119142649376711,0.1596608294110362,0.024167214215428334,
 0.08214715165665842,0.8485360761665245,0.17164551104060022,
 0.44829101920055203,0.3360750264119343,0.16724868561868866,
 0.10881361092253927,0.25727870349857906,0.07217566583357593,
 0.9952255861998716,0.8004449229658154,0.19910988747697445,0.7995103155447721,
 0.07352006563834912,0.4639763827165879,0.6546992664869018,
 0.28168198619665097,0.837042815408367,0.19201466217897822,
 0.024311267486977095,0.6280001663098895,0.3391215979804003,
 0.5931113307475395,0.6887625744762624,0.4074468204978242,0.23845743679568077,
 0.4952026831314198,0.18695959928843664,0.11046852631424608,0.583555683747431,
 0.5243343234912587,0.6808600284924425,0.6540239275473375,0.7401992957436114,
 0.6544613511310026,0.9663896851240393,0.920295861463565,0.08007725820811318,
 0.9215787696817811,0.698545264249434,0.3691211861576361,0.7480575235431403,
 0.5865382864694283,0.22554821724408425,0.15615035697116897,
 0.17160390569677053,0.5773690333201793,0.44630940317007695,
 0.1834631347711928,0.64529876919544,0.4194324905597846,0.77468984543046,
 0.366037684644564,0.33688033355367697,0.8795481440229859,
 0.015649862905484313,0.003312697980028645,0.39286442970488766,
 0.590216159427085,0.2132597547217029,0.4948890630609526,0.45870974409301557,
 0.880739018040896,0.3152391215295791,0.6269913085685207,
 0.0039034363438789077,0.5307947102804069,0.6618618831336702,
 0.0761292307343362,0.4755803308748896,0.4741062228169699,0.5831569724063688,
 0.7281969883486517,0.5728495701814033,0.4791969566028944,0.3637075928695995,
 0.8473798195469383,0.7495286708136806,0.2639333733569943,0.29544773681347525,
 0.5787667016341764,0.7858799609518768,0.11638997231387482,0.716318977149684,
 0.6551274798455606,0.054787219217110605,0.06132970317254671,
 0.4315413777932805,0.18364393887086083,0.3485448772220341,0.823214238398635,
 0.2561894257145396,0.9431610932308517,0.4534104639345339,0.4946940149303216,
 0.1630526851899443,0.4500957618152577,0.510558725415416,0.8744633359038623,
 0.3747688731128678,0.8125402839589989,0.9199710316058888,0.36094374935242324,
 0.5966924260730594,0.828288944460514,0.979371187903708,0.9803737877272157,
 0.8887863211417367,0.9508368689562093,0.15452802947388733,0.9895987621132952,
 0.17158531201928895,0.8628697361596508,0.34270841604050417,
 0.6047679463740933,0.5431206128124221,0.3817680008581166,0.3982878702780307,
 0.27284483157132255,0.9346697316642589,0.6289600102281107,0.2049730739852954,
 0.6444330021061873,0.827601644400134,0.5602385353524214,0.05644077989008456,
 0.08189757735315895,0.18104464817079935,0.16081601269368395,
 0.2144811435427051,0.07866600394785728,0.17355615872101238,
 0.040491316642086606,0.9007235829235314,0.4330361611608732,
 0.2427025236378636,0.4208500360557398,0.4557666623295913,0.2587989218802397,
 0.3913933200769294,0.13097705141594096,0.2544104154462584,0.7117660875439668,
 0.5180447694168218,0.14374651060128574,0.9505820901979574,0.1662890274621729,
 0.1924840202533531,0.09889960857752456,0.12328272235434767,
 0.7827623021623251,0.868858311617405,0.6763154504057287,0.6108961957609346,
 0.1429270034272777,0.3524523932287731,0.14832547039793398,0.6608774597648213,
 0.655074024424647,0.05606591403218453,0.5126697300875576,0.23991111559230494,
 0.5406663065328501,0.0854425098104179,0.04879904304029814,0.5294476533601367,
 0.9455693478583127,0.6025166448672554,0.5391689404369429,0.9127988830218311,
 0.3084570308981982,0.29728865093289913,0.7663462992486083,0.6131159022808879,
 0.039980053860036224,0.3336023636659565,0.2882006661939154,
 0.14961822195802854,0.7208753828752179,0.3875813701808628,
 0.10706687960423422,0.7168135385834908,0.46205459527687665,
 0.7904843292738362,0.29378334996669986,0.3371589439428485,
 0.37877517640325964,0.3088100974293364,0.059213862476965695,
 0.13333800838460086,0.6383537330711478,0.06537522719933042,
 0.5897528548004332,0.8444796482421896,0.6365144179478296,0.18265291537371153,
 0.8807947621955937,0.32924501969470155,0.891502914588,0.11792571219915648,
 0.7595280197380511,0.4988484393772009,0.9397883598295858,0.10805705382049813,
 0.8485234251378261,0.08216732175072028,0.35390872113748706,
 0.43108739575311517,0.9635418217219994,0.34612722195895496,
 0.27136528945766714,0.7876112531543201,0.988334702141686,0.04112275149640243,
 0.8409713802400371,0.8548945171726824,0.11799686773874707,
 0.26582656109940694,0.46125606235016336,0.19134565087239552,
 0.8788995381885689,0.942301335216935,0.9055163611238541,0.37622305068752293,
 0.8973431437514465,0.934494153543066,0.5236799201052658,0.3938204930382626,
 0.49329639483363075,0.08046759630092937,0.7053412671885767,0.827362937130671,
 0.9532703602151348,0.26243631760394925,0.02555040074077919,
 0.2979506289234811,0.8916383449628277,0.3467059842901363,0.09216082345597731,
 0.47876273573561756,0.42096284726354616,0.6378418884240231,
 0.41702617783493334,0.16993458794684868,0.5897090026637697,
 0.3200177788521976,0.7537952894377083,0.9510198744407347,0.8882508730993095,
 0.5562122888178657,0.8946047223958009,0.04244089521223948,
 0.14754722009198984,0.4651296272471497,0.61461472674624,0.705185996985666,
 0.8063988836059323,0.8931404862723269,0.6563684384991187,0.5250091203231293,
 0.4313640820288096,0.15780040085061453,0.7471541449343374,0.7866004586121058,
 0.006518643586776562,0.69862703120092,0.3305443900328262,0.0682699044260403,
 0.07036614688822451,0.13944798529674873,0.2373065390550987,
 0.9266034496934099,0.9526976120312901,0.8761073680306796,0.48561109941339764,
 0.6700087690826825,0.1573633977744262,0.23009135569035655,0.2085037025733104,
 0.8461714281606388,0.021227487482476604,0.976876242201356,0.6606472250987375,
 0.6022089184105711,0.5289619570301337,0.8731031584094839,0.775433227312388,
 0.31098718001071224,0.3527617926433987,0.9707532638305886,0.2549072039936452,
 0.28296999366870323,0.3791448800079442,0.20291859206388585,
 0.34507432409241834,0.03984596375780114,0.1474109731030795,0.506389337980684,
 0.0034715824188183975,0.01654050147018804,0.9341145897326119,
 0.30110239444140274,0.6330919892203174,0.8923195370322062,0.4691291720078742,
 0.7805955838482148,0.974057532245676,0.40660742067553834,0.35064899924133863,
 0.40821558255584656,0.18535587748440807,0.052059979511437704,
 0.7212023175728168,0.928629780816326,0.7877638243170102,0.6641889184175228,
 0.7508789322650298,0.4941904071824753,0.9747134760029219,0.7711412997706061,
 0.5718314310339281,0.09255592811500568,0.9572092053269284,0.7253189480430796,
 0.47667586512322124,0.7273712097480978,0.8244624457361547,0.3909210988610634,
 0.1216761785811522,0.14893714401215385,0.020011908499858122,
 0.7438804041464113,0.039838174418089456,0.16858032514497623,
 0.013041025285895458,0.6457883618228978,0.4400156276626551,
 0.3263527882265056,0.01015517776139685,0.139300807165455,0.3880440660289953,
 0.005083167533734945,0.8114257409408071,0.5761730833967066,
 0.5004345555537363,0.007184896832665766,0.807947762566581,0.715719022548413,
 0.8241614019173324,0.6554143089725516,0.11652542670669441,0.8895612288866602,
 0.21002337568650453,0.8093483070281225,0.12993268276940972,
 0.9385461065318258,0.4281821439099247,0.04347730405441819,0.5510544059518696,
 0.7750908997490359,0.3005166338135622,0.45802749458829917,0.4853927608605646,
 0.08829771024747712,0.9695974520557757,0.6794823083970651,0.8013787416460227,
 0.6938534739249584,0.55063263462908,0.2809996848991553,0.20126666989519926,
 0.46602136937838845,0.4916407741368213,0.665993852777475,0.11307544750759735,
 0.47493731440129205,0.2809485618674541,0.1842748805907123,
 0.09273740024063926,0.17810393418340253,0.19760126531773037,
 0.02258546056859556,0.7276794199752528,0.4620302201234273,0.9606504630436822,
 0.6024756576290363,0.2828478381808276,0.19057450913705043,
 0.38746946701964446,0.2518509008512322,0.910247182738337,0.40154545113066786,
 0.6955027617324877,0.5321557831935142,0.16469250248875023,
 0.21932136519999434,8.215779856989291e-4,0.8509020163504953,
 0.8199791538033686,0.4016046102409838,0.14242100316412554,0.8245459931296174,
 0.06388700947151538,0.31296121127810705,0.9422913769204087,
 0.8669212625389789,0.6250043115458561,0.985952574448005,0.45405273847194816,
 0.522458762465143,0.7474068086075791,0.3551383211754404,0.9692264961420302,
 0.86638970474956,0.3570185006384258,0.7020692162058201,0.28524058836571353,
 0.18491090739198668,0.5113102378782554,0.4118411636355016,0.181643953599692,
 0.8368308902235065,0.5731584854351636,0.8146440011482108,0.8491423027258221,
 0.5697404606205223,0.5371800067335903,0.8374869166281884,0.41646594634698053,
 0.662175687359961,0.2270612268635136,0.24820055802965169,0.27208801193694465,
 0.5383559943204925,0.8033458494219827,0.6681883175237224,0.7814828224042126,
 0.2520565980784767,0.058630390488245254,0.7463460510892456,
 0.5255330173424932,0.6447252581837033,0.7165585302768963,0.5880952516165913,
 0.264466749591447,0.952235227839068,0.6010725504016654,0.03787671317623453,
 0.5544011889333563,0.028463449711558342,0.5347331328508628,
 0.1389210045962227,0.7628233284596028,0.22878633691043404,0.2710865845760726,
 0.07944462523421547,0.9940220230594184,0.19295340431091224,
 0.34486479640122936,0.8316282049725063,0.7839554819496418,0.7960637615596455,
 0.16348265292090125,0.21026826278275612,0.821513319498873,
 0.12443801688863987,0.4724900013171689,0.5631189574947641,0.791224817220397,
 0.49982444304075524,0.2026733823502881,0.5561744444141205,0.4126640774212148,
 0.9666413169863902,0.4074159097436114,0.08843587155558774,0.6086233585242625,
 0.617160686569207,0.9712621763855078,0.06961234256714677,0.4845981147560565,
 0.7797158527671353,0.5174015651225176,0.7634718997004245,0.942305949336151,
 0.6782375822312905,0.2613337098630115,0.05512971153932844,0.633397387012542,
 0.7234320919244828,0.30955466616680427,0.33174013238119593,
 0.7199616746247648,0.9284202386853602,0.14797606008771957,0.8070194148539569,
 0.4462104488276847,0.07797715490015189,0.3205131644112982,
 0.23203328648960564,0.4750758568201948,0.5164462190211904,0.5920087905608834,
 0.4559315554835951,0.6860211514852371,0.42021577945404587,0.608759898893405,
 0.713752602332305,0.6759329658937551,0.6970250059446763,0.5346141231078062,
 0.8131403477138875,0.56642380725101,0.5837049472341567,0.41211715210195565,
 0.5676518862729153,0.5662799730983764,0.0884903182025355,0.878569719581169,
 0.749546001484178,0.9502878802612745,0.4264648803092669,0.15309577705205069,
 0.13582219216307978,0.8048604295273484,0.9308849796394466,0.8194321296130948,
 0.9231977938096708,0.37220469135508727,0.012531377064661453,
 0.14043075899698299,0.36157417700245564,0.030434745791780182,
 0.3847227034698145,0.30504047934298273,0.8598127089087845,0.3602236394999936,
 0.07802849350594077,0.7438923932595224,0.4747026924333725,0.8159433999140402,
 0.996809640213016,0.5977787661427704,0.3051312361730587,0.6493139840269844,
 0.7251845659876632,0.41790336885166934,0.19840234980878046,
 0.5049447016317061,0.1193047529241521,0.37897200183548674,0.5314000157428221,
 0.7329179234567494,0.4810537196362342,0.24472915258114836,0.961202717581923,
 0.652435623613333,0.3951064238558739,0.9806684274486643,0.3682494535749975,
 0.7758446729586681,0.4146235652034498,0.5218965494333476,0.17253570805872287,
 0.8366193953672679,0.5065002578581376,0.8253546547250852,0.2111218682419913,
 0.22821505446062562,0.751445953431843,0.7637225893248025,0.6955665788549683,
 0.7126032094962418,0.06649643461838362,0.8638487114251628,0.5703988079381873,
 0.1914206489058924,0.43975012264433144,0.02761396750818279,
 0.7873447189231481,0.14611369987810063,0.9320205383895934,0.8476328197121847,
 0.28694249572062847,0.6401726383009765,0.5793372153182772,0.9513206423015247,
 0.23067415249688605,0.3399361120679697,0.3744452426513787,0.2017837423948795,
 0.41056126566108986,0.1141023346647273,0.24150478976656709,
 0.0632089153161034,0.5378326585145013,0.6905303120189563,0.29170090453472663,
 0.16129294120611204,0.7996508546715005,0.22557553762204552,0.538555181470144,
 0.9384796644097236,0.2710342550543914,0.5160582502331864,0.9684737009549252,
 0.04063421781689158,0.2942046908064987,0.5636634090603705,0.6532505924879679,
 0.5627431739573668,0.20143703824280756,0.6592691113119744,0.9389725493413295,
 0.4962427613940986,0.587410555161167,0.2747534342456385,0.1804859918472589,
 0.3913845774321928,0.26423898914273614,0.9712121250751018,0.7517763196932132,
 0.7718690049606551,0.49746122873949317,0.131547003642963,0.9810935960184265,
 0.847194478414465,0.418739430494814,0.39649624236525083,0.025081680323397526,
 0.34703503070688135,0.33589145311937985,0.9849788699894808,
 0.7789722225353306,0.5337695352547817,0.7403185957250373,0.9513358348321942,
 0.6923712602657534,0.8561894077343364,0.5131329851022809,0.22147750588652226,
 0.6756708334272478,0.37402019344817317,0.8419279525831418,0.5655453526922465,
 0.31568944989774217,0.0668199272190314,0.1277710081079212,0.7141526214111698,
 0.8091810974009577,0.08144290794223397,0.5023811893428365,0.8047989464243168,
 0.46271631642074185,0.35835678217585176,0.4913335651223204,
 0.5196813372090199,0.7894505276370897,0.17016819549596807,0.1666292881997239,
 0.2996812283156167,0.04897991774392785,0.8817463137175836,
 0.015262049039628911,0.9189327499173261,0.6579673118655938,
 0.07393046594485086,0.45512993057238693,0.16160011447721678,
 0.3474350318656062,0.532459173598095,0.8763602127851031,0.8536016974925712,
 0.7700490502830133,0.571398121458379,4.240399321320254e-4,0.5569375381904992,
 0.13427344646753325,0.5671806589873306,0.19848570468190674,
 0.7585537464439784,0.32895836629923714,0.5460696926357061,
 0.22215120072599726,0.9311902119589028,0.9348025750623189,0.2984776782715902,
 0.49308165365819867,0.20720854360212837,0.3611862242281765,
 0.5125878685759953,0.07235977186840326,0.04540594291294586,
 0.04600206094507364,0.08085645326683477,0.6317549147715223,
 0.24077597222600655,0.056717262378882305,0.15353466456239662,
 0.4753061027000477,0.46511605676787826,0.654083354278043,0.6488988945708094,
 0.833751638064498,0.19741957659623033,0.5801425088535105,0.4603012601234998,
 0.05838426862182178,0.25574984401014433,0.7906033653472332,
 0.025878487423530605,0.15103401974075603,0.10267717120273145,
 0.4388176389960039,0.9264570265595575,0.37021584353549364,0.6049312496926772,
 0.7518383511977064,0.5428277940146007,0.339301933962876,0.6555525287371401,
 0.16115411190031936,0.3993358212850395,0.3973582312755567,0.4618926730009647,
 0.4895817823869627,0.5531046579633934,0.6488038072202562,0.763332158010823,
 0.6284773983676912,0.6031135159561337,0.4693765764117177,0.4768814686485361,
 0.8870378341067551,0.913276136107186,0.6690271298035866,0.38378230513388145,
 0.336344877374232,0.3379303197184592,0.40371785968018825,0.12256594730024595,
 0.4811041370320638,0.09071618719874464,0.7944978896966853,0.64751798030188,
 0.2195176624637032,0.8549479633081294,0.6271830922078707,0.8770742722921894,
 0.7436117793959997,0.1285619557878549,0.9333174113220413,0.42179432157732566,
 0.30427924626106406,0.2620588761387994,0.40526581209614776,
 0.4508879334990701,0.5430404554303291,0.8963454307606951,0.8246064446054389,
 0.8479804251434511,0.03198043068698564,0.6538852640005031,0.2907400664201527,
 0.7993881992533156,0.19071717405607014,0.39422053213461594,
 0.12647695406592174,0.17913544097978296,0.22803107295588443,
 0.19122720854074804,0.5231131171105778,0.08240056924107686,
 0.5274068035629591,0.030030944590705544,0.6134173268175501,
 0.14435751422904586,0.9537377370588627,0.8452687818869629,0.3715454024338466,
 0.3498801592845061,0.785096707089667,0.9197105236213619,0.37234850396901154,
 0.3071116380254343,0.17744171189045554,0.3496588322466043,0.4254189982768677,
 0.48419789090760434,0.2675536706372693,0.6135442158715971,0.9501911726083593,
 0.24150893486229785,0.275571363876971,0.2008440201605196,0.533150508400642,
 0.0461312850104868,0.46791427159938215,0.4722601447273177,0.5548671559804168,
 0.6337144640109411,0.31970119989828194,0.6229351967472765,
 0.33939596066403843,0.8873122219098822,0.05919804839507292,0.803542612584982,
 0.013358473431720475,0.9754835777523422,0.38531045540031084,
 0.09635570776483227,0.7354273630146719,0.9205713813110976,0.9459756348977683,
 0.2328354586828978,0.2327396398741537,0.667973752281083,0.20621768995862122,
 0.006337901228868903,0.8622183878500811,0.9702023648650775,
 0.7954139806038665,0.49260591394686637,0.6277568462823472,0.4569786124444668,
 0.605689566812967,0.3417473464864007,0.7978286250092431,0.9707237529588314,
 0.658673240262829,0.03374540548953375,0.02054833319589322,0.6707483376037096,
 0.7768085182212642,0.6994988670956355,0.12015084916037133,0.3821855379459833,
 0.7581887989384041,0.05433104777920428,0.8894450306807974,0.7054446914980605,
 0.809698050048163,0.1631280502646173,0.5479825566256423,0.3465745024736431,
 0.667859160615156,0.4210501511638194,0.4421852153725797,0.017687548416975174,
 0.7968721387921525,0.6513990365902436,0.36284059975373273,0.9168490388324482,
 0.19510113148540664,0.8213727916847638,0.9490940398890779,
 0.24191186739041304,0.9323885964597096,0.4079907418266655,0.5069837476879772,
 0.8922861049167589,0.17908847581316378,0.10771484346542926,
 0.9630834698701233,0.33771113651336393,0.5505093893854491,
 0.08697000104615038,0.16053158680058177,0.4737766855561918,
 0.7724757667171893,0.2916591332651066,0.6202418611624201,0.03546792292499312,
 0.5317027105367755,0.2070187674336199,0.07108089660911265,0.4931137639233997,
 0.06426773535401353,0.5464165650075038,0.21068774105340626,
 0.8146673207794162,0.059577143689374834,0.9224341567531126,
 0.4353880742563919,0.6477982933561207,0.12097558552105359,0.3290855916446992,
 0.7967991602854554,0.5050099191127956,0.8571946610993383,0.1259501217381933,
 0.3126266661148982,0.6572914581352616,0.09209281188150564,0.95987512461804,
 0.9363821236754153,0.6079842916716659,0.7733533206731938,0.03725813616815454,
 0.41025466166962343,0.06932478819133303,0.9672820675130867,
 0.6928308348414264,0.5130217803641605,0.09294021529541341,0.5778273870075367,
 0.0906145255942169,0.38955215726700343,0.8849662529734715,0.5541145758044812,
 0.46227148783368854,0.533669181898783,0.01379517941414421,0.9025234378127248,
 0.14245613115677758,0.1659321044761468,0.004009596155530648,
 0.09209370487615054,0.5611978325366332,0.882191891031665,0.3935771588464618,
 0.3103307539696022,0.3985323046355804,0.5401771849831941,0.5358096651218612,
 0.8864784902365803,0.09840628176405342,0.4653582662221831,0.6570476542401044,
 0.34731797192530156,0.1348565960090441,0.8931289710646438,
 0.15301076907793376,0.5332562007459306,0.4134318250890703,0.9886320628099167,
 0.5702087500561019,0.6512961540036318,0.4346517043949919,0.9319367626344421,
 0.38673112656964503,0.5059372161404205,0.9857332558971899,0.6283344691689775,
 0.22194727668303194,0.9755608859267996,0.3056550337423909,
 0.43450020648263055,0.6496621770030332,0.4775051674633223,0.8344774439091799,
 0.41376581221288766,0.39706838028703295,0.5518127571196,0.996799636393637,
 0.3823067025647313,0.0015695603433244498,0.06489573685422667,
 0.9678971869306822,0.931592618414264,0.06937739349433258,0.2640450270366199,
 0.6710703424302724,0.4177874094472068,0.03709263654624584,0.5412605389312274,
 0.49343905865421134,0.19018847779967052,0.029020409600894692,
 0.0745638001670147,0.4255588629200687,0.9521782574297034,0.8548012318695726,
 0.5462327427761119,0.35325437326754605,0.895040753903741,0.19885796830124525,
 0.9482430135919054,0.08320691743267794,0.6924829026683129,0.9323556206441326,
 0.9650173970441036,0.19876106361553636,0.8226703665950681,0.8358431954950042,
 0.26313869652201827,0.2628412463806894,0.5624941361587693,0.8328073431765697,
 0.36324152988537006,0.4829568522147697,0.8985967391318153,0.3900084799873229,
 0.7161320677070804,0.4681038639163567,0.18354314652425752,
 0.32562143984440617,0.8080396396630781,0.16465826615598955,
 0.049767384299128414,0.9230284630844334,0.964338874413826,0.9543394108224424,
 0.8039880137769387,0.0802185238146691,0.1288411365414902,0.3580001456866535,
 0.4951237486448282,0.6699025146874504,0.4709720201927321,0.7852220602644928,
 0.2969788319685982,0.16186082429342097,0.09245168183423125,
 0.9026935868695075,0.6378888633856717,0.2564484304961637,0.06384273992822631,
 0.06490301282845468,0.64530123669207,0.3485427414440325,0.9791922713323544,
 0.32242981444824526,0.28048603146821927,0.9443422371548245,
 0.8227511156092239,0.3236707724176652,0.21092420451621918,0.527430035036131,
 0.8777107879048812,0.8059601451643592,0.4577437191335457,0.679325743481158,
 0.7294170457082387,0.6554048050181467,0.092168037320418,0.9298560408168066,
 0.3496041802757749,0.8315702144075755,0.6955107823188409,0.46549056468163974,
 0.8489129072246232,0.19095269133204928,0.2357910836549063,0.4923180160220605,
 0.20043644597473864,0.9616880210799155,0.36402448427603296,
 0.8497772875427081,0.3017159064986943,0.24888661378293475,0.1604478523671644,
 0.7961079695465433,0.35467050372472997,0.12238679924049367,
 0.17458342288825068,0.16396492583233635,0.4384128051467422,
 0.40459774342764043,0.5120586442838732,0.6931746097157413,
 0.21257980391917908,0.007979810802658749,0.5333516287095609,
 0.06345207920953755,0.5333229850436285,0.769742706547507,0.34833762578956307,
 0.9763132517525956,0.4802067590015785,0.09339689387058148,0.8795973750108582,
 0.9328869311435212,0.8875290830368346,0.7087707940872888,0.9412295722818603,
 0.6023540405206224,0.10762682779063026,0.09275402917964681,
 0.38730265876375336,0.10581350681047952,0.599804330993982,0.7543035155546512,
 0.19626369442407687,0.45728395316048964,0.7565508132378381,
 0.004855295919649061,0.15602493485502444,0.015380419558225333,
 0.2792251428207201,0.8992749893726157,0.0012253048802908229,
 0.30065108519732153,0.3599600527505352,0.7322657945678193,0.8149399505152963,
 0.6143713808425565,0.5071118804073453,0.8714879508160746,
 0.009174481496244558,0.5169181996702472,0.641433112009346,
 0.038321824964473894,0.0221102893316798,0.27941027841661603,
 0.5004440569523066,0.24210663893016815,0.09738640804131138,
 0.6761815811936638,0.5090099345143555,0.7814915000714157,0.6265242580431389,
 0.9125022265368206,0.7729825261424317,0.4695787805492071,0.9994334840087569,
 0.669473300684898,0.130155172467323,0.03799229985288666,0.5219884479800828,
 0.9004983373009783,0.07819381841075668,0.4898182002890328,
 0.07064813690033556,0.6133162845473136,0.23365922383774096,
 0.048677303292881224,0.9500858708123802,0.652720447300145,0.5790693525448414,
 0.3553317376784202,0.563875930219951,0.4898450237337728,0.17029613629782303,
 0.773994154157478,0.7609415776995997,0.48505003970904725,0.32045063620261516,
 0.6075091083753542,0.39094112288088045,0.08101417651942833,
 0.4279787859648061,0.8773686636805792,0.04011789078496908,0.5687216828060543,
 0.2068600239041929,0.8615438493095674,0.9619940427580107,0.06039149676162092,
 0.9441544303820579,0.6541459692393434,0.9947577008457386,0.8136675224677805,
 0.8928911458167013,0.8773288261845203,0.3625013266358732,0.7863314859707978,
 0.8635231380973521,0.8014129759214316,0.5383011891914127,0.5399645879053938,
 0.26826003527171327,0.8256033060962189,0.022592109908535662,
 0.10997342272956467,0.35710331115184324,0.5092079491349757,
 0.26719994539665026,0.5303932662012617,0.8000953203060082,0.4429172667347353,
 0.25974292814535205,0.9001965093257694,0.553247545763611,0.06849969624326424,
 0.3029505718583728,0.3537572483464182,0.20140071442114416,0.1710971503087998,
 0.8694170052213204,0.9853578831813486,0.6301398423929645,0.29078586324040334,
 0.6325420526716877,0.19155817312442824,0.9024130503581945,0.6266085429670519,
 0.6075397021141088,0.058048830750246805,0.28500152988323235,
 0.769223875817855,0.25474217463070936,0.21990408761658164,
 0.10407765821759662,0.79867279687312,0.5898077703600072,0.40449283680293047,
 0.10827761299354888,0.7142502584316126,0.370707023652624,0.11432575837968506,
 0.11144385821209857,0.8306915790959141,0.07015682132250123,
 0.7636392260539393,0.29231418721892444,0.5064150988029534,0.916537937370999,
 0.9820511814571438,0.4506493123950124,0.9608045101317856,0.6035203029245485,
 0.42004528241221106,0.25850475567063014,0.7216039339092633,
 0.8111448372096417,0.4755652822226756,0.5037863101134222,0.16436671979859996,
 0.7432783015376266,0.9259338211000806,0.6743495056762832,0.33028738467361185,
 0.007260002911176566,0.5015192755437754,0.10458916253121997,
 0.7367648231409436,0.4857776944895069,0.1885928621546955,0.11302428492968886,
 0.5870906590147709,0.004270167319683216,0.7804294006076211,
 0.8415852831328642,0.62837084588491,0.6778765970147174,0.24854364311984423,
 0.4696312977845759,0.8953842507859116,0.5510405821347729,0.541122849029851,
 0.9752506367401743,0.8767214746937425,0.1287358533140216,0.6711406219883088,
 0.7521780650119207,0.7409319565858831,0.5841179693124625,0.18343010001959725,
 0.486450969020897,0.866961615365971,0.5677835438044034,0.44579653566646316,
 0.6821182899287406,0.23627363819628533,0.23244146675186783,
 0.7055596588630281,0.6521995218381681,0.63181893409034,0.06790253740100616,
 0.658209871608364,0.5988574278366707,0.5583924457202584,0.4313709346204796,
 0.8692765585729625,0.5641213479612864,0.16196061698187136,0.5826961632376103,
 0.6562103214123506,0.025558255477108993,0.31873053304448584,
 0.925596971878808,0.4387603317651926,0.021597308041953367,0.8429572421827524,
 0.3842366232862777,0.27024681914457815,0.66084734510542,0.05434364297847505,
 0.126956448920432,0.8882148617275282,0.1838095252049219,0.8901606017229976,
 0.5638972105199077,0.5815007993394102,0.7458294636568649,0.4637203830196217,
 0.7388180719728136,0.09310330223530483,0.22961624321658414,
 0.9058072504214226,0.5806774952015481,0.5957830163490216,0.3417464481283081,
 0.4536928539297541,0.579306037678894,0.31289681622371446,0.44126529940608183,
 0.9245665627830669,0.8843355305574928,0.9179166787193492,
 0.056703926714771646,0.06208459383507858,0.9098650973239858,
 0.48560028578766323,0.19740070229765883,0.5338837533833705,
 0.5954620241107252,0.18871590018303064,0.7171769016199422,
 0.07874378361259349,0.9100415004903669,0.8953523947705716,0.6836055095827439,
 0.0522489989797168,0.478660508795976,0.3374968066710886,0.4880427887083406,
 0.12226100400121964,0.8790489411080508,0.8767560904645644,
 0.043298057714066696,0.38491852107376945,0.005429376909733885,
 0.8917098084375578,0.4748831208483333,0.2017482317831082,0.12308233446061978,
 0.09261905164236339,0.1383796907054471,0.5438105226383881,0.7746768993892232,
 0.21654677032302616,0.2976891874864571,0.010950007215025148,
 0.9982283387866471,0.2875742918802874,0.9686453866173098,0.3255723380196458,
 0.9583720626828399,0.8963953757614681,0.3740323570897123,0.6491848663439341,
 0.3644460210859546,0.8997378556838351,0.711029113806439,0.6661315942668945,
 0.9618115056887013,0.16037185152679023,0.5414300294769905,0.7420600374223599,
 0.43035782894871455,0.5578266279095812,0.8283169239099581,0.5849281084808693,
 0.39111399295575033,0.5205925816645456,0.25742148873712123,
 0.7100511920290075,0.804625566636053,0.742807305259877,0.15697662026469494,
 0.4977759255452734,0.22859958828040183,0.5384658026993537,0.326807193198822,
 0.8219922368126844,0.7322698413617221,0.584503539210827,0.05298728007697062,
 0.03232525183697188,0.8119767599798283,0.0316657154597888,
 0.43990417584981145,0.2059110798951167,0.15848860698098088,
 0.015386968968832182,0.27751183360086684,0.987389051758567,
 0.6909887702001639,0.6851013712251082,0.4264205409523367,0.26415051664936995,
 0.3663656633234791,0.5090349784581336,0.615615531240911,0.6438461094841862,
 0.4450019367031508,0.2889781113846245,0.37913624147731206,
 0.05379985743042315,0.2740551469110821,0.7107866879731743,
 0.36147836173475956,0.13187101901590326,0.7453972711028573,
 0.3943487436422384,0.7482504103015506,0.9156189809976195,0.7068471829636352,
 0.04375193826696422,0.2690990129255605,0.09928870226464193,
 0.3588615824176098,0.7064507522967521,0.8305264846290116,0.609269609467257,
 0.8170472753408291,0.5488762890867573,0.740667013026189,0.6995194513079296,
 0.8076228312804382,0.08906897271162484,0.328472891925372,0.6035764055778016,
 0.8639300768961233,0.1271866503771214,0.8344168424213366,0.10253712725025743,
 0.34413302818671676,0.6937853351370433,0.7448786689456892,0.2116462963095087,
 0.6070224720086095,0.5660481385534226,0.05520988374000424,0.7407382374067295,
 0.04854407260787286,0.31792040981501857,0.2710548047853868,
 0.25566522537826575,0.5886924604572933,0.9607702604282384,0.7297881980716517,
 0.6830056511416301,0.11876144774033648,0.34381052452054517,
 0.6986945492825445,0.018690224644044462,0.48191345512652417,
 0.805020148226517,0.7735737559715028,0.7150403718659992,0.5542638311213969,
 0.6822277687055,0.3957742193420649,0.5881463188598305,0.3433206041389668,
 0.8646234270005436,0.3812954420067617,0.4461782598324885,0.9859015248380564,
 0.41039669989407723,0.9286897043039037,0.8748970597816018,
 0.22588432254838675,0.6815777274363712,0.8864065252994077,0.6675702957700399,
 0.1955414006547722,0.8442051856565066,0.4288546355341092,0.9440886406112335,
 0.4778372652448981,0.6694191956279009,0.32018225179851934,0.4668096371962329,
 0.20890273384015035,0.743688368605885,0.6903813462758228,0.7441161909168851,
 0.4823695023525715,0.027086807167730154,0.5834666227558689,
 0.35195566137648504,0.7931225800900426,0.802785123816288,0.6108538093847182,
 0.5245819537777007,0.004056650418816532,0.48237427225322815,
 0.499868984871231,0.6711381460839305,0.10818701926777374,0.6803903786742729,
 0.8435865737949514,0.5444264851749693,0.5805922946368854,0.5267122567610459,
 0.5061288141580715,0.5418970386992503,0.21571285088441028,0.6000198393395156,
 0.04674351233346652,0.9137861988068885,0.48844438306302096,
 0.5571450149801183,0.5210782207123319,0.698754918061443,0.4686906366165592,
 0.7719312441119022,0.5777237892127195,0.18786315698193312,0.5149623452671389,
 0.09733214457748451,0.616873820301006,0.2730805679822117,0.2212997341391889,
 0.5973373611220294,0.7841131207543968,0.6579695726707817,0.9693465233153343,
 0.04130715074334712,0.7769991534605057,0.3856076622038246,0.744177326254764,
 0.20341219059857196,0.6667260510843759,0.3651453930067994,
 0.43604860029599535,0.6242724998980131,0.8624930100106745,
 0.09104508136883904,0.3073492593474886,0.6520780399519384,0.6522914838409757,
 0.7289172248891185,0.14818809214578743,0.43464433213643316,
 0.8712417003546984,0.3085500995755529,0.97653014699184,0.2640986909190356,
 0.6936457166631401,0.2955870597903546,0.7090568611687221,0.09190991380666158,
 0.06284689611055905,0.4743682552185572,0.435705822777649,
 0.045934284595282704,0.891218888428343,0.6946966066399077,0.5422910636655172,
 0.29148136755548704,0.6813009754175333,0.1242092930841191,
 0.05961907414912215,0.7806754476852352,0.16225125813082697,
 0.4865143124543656,0.6033109168198986,0.8232870221631204,0.13058375966121283,
 0.8494259386918106,0.3691542466347848,0.5077064225781238,
 0.005999358247588016,0.17429571793300536,0.3632112567417466,
 0.29651870873904784,0.11872826818168059,0.2109906877283001,
 0.03489110467493517,0.13126563888652665,0.7563097692972596,0.966414848305673,
 0.9994175893142097,0.2937296642549543,0.22453856805780803,
 0.06871198772953813,0.7135342091807881,0.013260021688473622,
 0.3697101184464042,0.8405405941975221,0.08965036113936842,0.3383298090035687,
 0.2742182487367897,0.5652955249607967,0.45591586708864573,0.5304828317966992,
 0.8318625229692131,0.8883207435184751,0.18785176383908897,
 0.39121919499354363,0.9901954658495462,0.681220446646672,0.5599808327564135,
 0.39716301860951886,0.48205810968459994,0.9644174233481262,
 0.6845202750886936,0.5524210691539467,0.08029775134489281,
 0.11935383477662276,0.5917312112821667,0.29919880410082633,
 0.7250953697252683,0.3001236887692478,0.0690077115523492,0.33008605504393684,
 0.2559461790705583,0.5622710404386194,0.24360412503121553,0.8971182842897363,
 0.9412778894422715,0.9864994631401146,0.055550825731987996,
 0.7652215238633707,0.15620888794914356,0.9641845604722705,0.3410866756236153,
 0.213575153478761,0.6883795239552801,0.42381678270251977,0.08153041004744221,
 0.2740024665367873,0.3515397603221545,0.8568098634791359,0.787709863144905,
 0.6774360158194426,0.6901778274669909,0.2569783385797386,0.10990196575198186,
 0.841955104749624,0.1319588513580815,0.11782129777623673,0.2603526421881929,
 0.4878818880071144,0.47776595412243716,0.9434575866970208,
 0.48787413773908384,0.3779279326201651,0.7580612531290155,0.943586162637422,
 0.14859547791993166,0.7882092273315051,0.40135830055432176,
 0.02423031321474789,0.9752929762300685,0.8316086760314868,
 0.25162847880973427,0.27836394803784037,0.2740878344525175,
 0.38636957802778416,0.051416649680078086,0.27343657103274477,
 0.7376898133804575,0.6693224566745452,0.9484538748236705,0.12273506853509386,
 0.09188779371964839,0.6128206594520363,0.1346041895119321,0.2702024938307487,
 0.5135712774377984,0.8148707806674069,0.3747779991035386,0.07600409435498612,
 0.4885260446179124,0.34178616741440393,0.32889684048684376,
 0.01778318942156043,0.936227980859899,0.4443929753136969,0.9922201611352759,
 0.9198556364788483,0.8736128187608233,0.738262499414021,0.7972334624349282,
 0.5102723626638096,0.01910832668514073,0.36519273248713013,
 0.7108145758027681,0.7149022573346847,0.3141424583626319,0.2709267219662066,
 0.1723560021598145,0.12241734439922813,0.808464156478337,0.09418296652181646,
 0.2529435070875613,0.3694444109226317,0.3041655641929767,0.5320000938122676,
 0.6463958404184922,0.4368359654834525,0.7522495412891167,0.7385324515339152,
 0.39442471094707965,0.39195231657611485,0.46593259775110385,
 0.8512913077152093,0.7234248988877289,0.8533892637004521,0.4638653703092822,
 0.8432886612611119,0.44625692215828494,0.18927004084076937,
 0.9454043642576775,0.6795857560509236,0.03392625715838493,
 0.44626401756850587,0.41777050304137464,0.15938794179887728,0.13236309008543,
 0.08365104676714841,0.5564780514922112,0.19676435292181071,
 0.2705440175392475,0.1421398720268392,0.4306776851129904,0.9643129735352254,
 0.3087951920954025,0.5847269828067907,0.7892864117388997,0.7547467038962169,
 0.5078232870251599,0.5142729457671428,0.6967242672100423,0.5045714445854705,
 0.42183860319597366,0.15320971648138082,0.3452549581836426,
 0.38623888699075204,0.17510863272511745,0.45932657712319946,
 0.14670011120105453,0.12896425489142838,0.8307160797144512,
 0.6668424261570197,0.9624375130447731,0.24849903557818076,
 0.07744992513997173,0.23713989354849008,0.9536240681903627,
 0.20258679481238717,0.9985064393231741,0.6471815983654097,0.1718053584292898,
 0.43288025816170983,0.47937547298152217,0.7719394829186939,
 0.5145060205499252,0.22245334184017906,0.6831687580825865,0.4760453756694589,
 0.27971701015611905,0.7490023671203684,0.22049870569393692,
 0.7976227032312391,0.33458721447115736,0.09751390743832211,
 0.20407667742807523,0.0753021164429406,0.8047654009359027,
 0.31446998183226693,0.9602414280778208,0.02859036141354887,
 0.5436571871525812,0.6948812257405577,0.23306939946550298,
 0.20750385906981772,0.9907074857923597,0.33676389611574864,
 0.4728054892913778,0.5080020432984441,0.5267860951355567,0.6328834407003906,
 0.6344911922775505,0.035346194656948327,0.47569306733034844,
 0.9303520250627247,0.49111449201514956,0.7954631943251987,0.5837479620380364,
 0.47147564924587515,0.2282263106345792,0.26953274992084386,
 0.8664395291256032,0.08348249599420998,0.9561050216525633,0.5478444527736218,
 0.040101372643863264,0.6312285048731067,0.725611361173568,0.6603673373144284,
 0.15841792172319402,0.11743567611124472,0.02705237987261533,
 0.9270342203663057,0.4864421539299604,0.5492442987821253,0.7402014509157981,
 0.28354480180012764,0.9136711572888965,0.48878570664888177,
 0.30243106820094035,0.06350955997259711,0.9618031936477767,
 0.01406127451852257,0.9423064616242354,0.6588286259466454,
 0.09766062320420521,0.5339701124094015,0.5496487570119158,0.7223359344222408,
 0.2794498041214061,0.9249680985017548,0.36568936992637424,0.8142545688514671,
 0.9760756702250106,0.7756770411763242,0.4004998157393026,0.38610304269908113,
 0.37621962056635927,0.40478006497677543,0.6421063629505686,
 0.8185795670511243,0.1438638298616599,0.7510441856777198,0.7010335654061883,
 0.43617549800771127,0.15918565608130653,0.1269147155543987,
 0.47424006540026764,0.6081790138986292,0.6555794517541575,
 0.26026476101973295,0.4523680634815499,0.7469948596670712,
 0.15473876435199896,0.9967514555615758,0.406556656570777,0.9768186611243765,
 0.045219750878440435,0.507632109362364,0.12315904987317228,
 0.9373343615623546,0.3800314058421048,0.4544035235683629,0.7472660729113723,
 0.5938728799501736,0.5974154428456007,0.7730479597423631,0.15299508528435923,
 0.692029982406691,0.7129662824042082,0.1760069358596037,0.11775188741170528,
 0.5407158927822034,0.07634558197802843,0.009149666891074548,
 0.801696781337309,0.469569320264246,0.5019195782604141,0.5452652755566012,
 0.08462715515487995,0.8365974702768378,0.8524335233772242,
 0.13001051337135838,0.6198072363042378,0.4988049337803826,0.3976922413705599,
 0.06396793610628859,0.7641887419289605,0.7330117252927446,0.8645672438800355,
 0.44064948707586016,0.05751283874964619,0.7213818575741199,
 0.4895713418980573,0.41886298799303023,0.08638442785967682,0.473455772558524,
 0.7348265769439554,0.5633259400305707,0.1260616925528425,0.20107050790794412,
 0.7328084765925975,0.9231632458700713,0.47109652922229983,
 0.08247234776694978,0.2536868128590133,0.5150351346673798,
 0.009764822848770338,0.5861466248603237,0.5157188544895207,0.550969691181077,
 0.7043455440714158,0.19161729781262837,0.4479780040431064,0.9045349134253526,
 0.8190669857420585,0.5082099975695539,0.10493300239749503,0.8833415183221501,
 0.5451872252368499,0.2602306793026503,0.32711448317976066,0.8254712967163772,
 0.7086501700222322,0.962976083823283,0.8475535770854922,0.98646403858556,
 0.918229572080742,0.8937714309322387,0.3178069986717684,0.8231002870677608,
 0.07142316178188723,0.7350345390667862,0.1775814985619103,0.9501482600636458,
 0.3315877316621775,0.9978739470015063,0.775481987604007,0.9069304639320231,
 0.607249958685459,0.04283885997236214,0.1541058388587324,0.4105008363097349,
 0.28524121309422235,0.8125381206167057,0.028759706644540972,
 0.1453507405274984,0.302907394616592,0.075215581255348,0.3607731027887603,
 0.36791374613626593,0.5871465461487821,0.8546315757320123,0.64591687961434,
 0.3375726425707055,0.270220791689076,0.8658932121609753,0.047141012121265646,
 0.4533632316315489,0.30454658796847645,0.23007046248094176,
 0.5962644600751318,0.04654500491556879,0.3795707568830078,0.7476414982159388,
 0.3374670058511432,0.25587543685627345,0.9509338608566775,0.5169275360628952,
 0.3359120232763253,0.1658623969396541,0.46912314826876267,0.4194151668135796,
 0.2358612547696849,0.9447920361480916,0.970307479021195,0.09833880207888068,
 0.9437025084224462,0.5786067236020895,0.3146395190930913,0.277443571546481,
 0.7904038766425098,0.11848082491591994,0.23746850922405383,
 0.2865659953294263,0.2389866320227031,0.951439950940493,0.6531096155344679,
 0.3738151051232955,0.11673225344959237,0.2215337560862709,0.5802902849635082,
 0.14343577074108715,0.12714670370379333,0.04650280282580144,
 0.06626673477628142,0.9964436753611082,0.029256927321268722,
 0.7094036625026081,0.8650668725348032,0.37547597197369686,0.6779118676043662,
 0.6098604190548208,0.6329531291504472,0.9326853156043068,0.6258042260714337,
 0.2662617765107519,0.06240727288847947,0.07718070003818411,
 0.09446073742785488,0.9811402889806188,0.9056659573935527,
 0.011645947779989618,0.6451725552321713,0.9969417106042665,
 0.3562819903548353,0.2029816173615262,0.6706316029789632,0.8153228415913366,
 0.8960355486077699,0.6880445673666418,0.7432629720711357,0.9316509892396526,
 0.40045515498095563,0.4055625346677991,0.8684019532026883,0.2535902716047067,
 0.6960493057904819,0.006082371928010422,0.08967805222535463,
 0.9538728143557689,0.2332151284729358,0.015958646237634322,
 0.43814824379930595,0.955884022918179,0.4959398604514256,0.4347528842911801,
 0.8752137809730309,0.427291922372667,0.1251291167471975,0.13257932187773913,
 0.22912514647859838,0.6883122230945372,0.865203382973745,0.06244239569091392,
 0.8747786752794383,0.23697586152038763,0.27370858037572765,
 0.9132182801140692,0.8403591437003162,0.48713438071618864,0.9952852101922997,
 0.049670900649592675,0.44758070106509684,0.132168478358896,
 0.035416058416612994,0.8209856898139611,0.4381579667835993,
 0.5642785201163666,0.7027930155672812,0.5274215219330483,0.19433356614241826,
 0.830109278679461,0.33607575660547173,0.4709215737228214,0.8358011353505564,
 0.8155132266527756,0.5933221105566978,0.9536080725406373,0.45132661165127197,
 0.20714803924397485,0.9791162820573781,0.19891820005700434,
 0.04821048700345976,0.08821220994018764,0.7508382841425932,
 0.3714652764968205,0.9641234361661256,0.7445195087877456,0.5198784925195141,
 0.07343272526636768,0.18077519512075457,0.6075295009232764,
 0.10755488415524317,0.9300988199462057,0.7508234252476957,0.9613161045760961,
 0.9194811229658333,0.10543113930549397,0.45807393597469903,
 0.044456811911592675,0.3885106221494111,0.2755389898612032,
 0.2806216011933128,0.6794192774404127,0.05334858677536769,0.5343545260869945,
 0.11461714465808082,0.09009815229936202,0.11525881747328393,
 0.19119280589042398,0.7501085044941366,0.8599283768755985,
 0.12961401151604046,0.9651166628726793,0.7267567672787532,0.460552201258028,
 0.8832485311985279,0.4890994306037628,0.861910660074603,0.8253545039996627,
 0.7249963774245843,0.6579809735530738,0.6186347621567894,0.8659336422918347,
 0.2057969990003341,0.4843663813092931,0.4271927111154121,0.57647717409235,
 0.046947336817456664,0.49674394328889604,0.8383825002228144,
 0.16543828506504743,0.5503414730465233,0.4097219085461027,0.7671147403715145,
 0.5216710922871017,0.8876693379548448,0.9555113938429465,0.8588687629211037,
 0.4746366215713623,0.7388145138131978,0.6487442743991619,0.6402127625180389,
 0.9658643252934798,0.1903630272074407,0.4038833220438386,0.7592187600374539,
 0.6445788306431979,0.5006907981893788,0.8838047868065031,0.3431131706149366,
 0.7070443162181641,0.13649230823097813,0.5541840433633611,
 0.05054001275501596,0.06700303471163715,0.9200159207765168,
 0.7603931876140884,0.24182903584039284,0.02501083335668819,
 0.9649227425757961,0.8584539630601997,0.8860467221194649,0.3919699811168781,
 0.008495434529594337,0.4695531841037589,0.9252900669441455,
 0.11926960529955144,0.6316825410194812,0.052182753444684704,
 0.8641606070090715,0.6847017729709919,0.26359466704159773,0.7951799304963529,
 0.4988567991960652,0.30845004485125893,0.6290493280558545,0.6316070182147138,
 0.5639738309692298,0.9151891338396786,0.03371929887271863,0.697766597283283,
 0.7133930104549266,0.12645183967214657,0.4225805029274897,
 0.42572108247526763,0.30122313296873837,0.252283631752661,0.615872544363663,
 0.17628924651764133,0.24171088824616727,0.5588021627217544,
 0.2872681017823553,0.359253784916003,0.9095528910268282,0.04649797073884332,
 0.4184574561159715,0.3065999574865653,0.8068048468905941,0.44546007769300555,
 0.5939898377258699,0.3410872007760112,0.25310482917676536,
 0.28358193171394697,0.16030514195703038,0.36964232151583976,
 0.26579196978793984,0.9210671411347731,0.18041625362413938,
 0.06782925080169189,0.8602578485241559,0.8495059521589212,0.5902467799195754,
 0.878307177941736,0.9002576436706644,0.801954842850997,0.0581266669371483,
 0.6046220668055243,0.8052078997734236,0.6628552914980006,0.5241591333062555,
 0.5234471770308161,0.044937571224658024,0.13095841239442807,
 0.1606450326275315,0.3007828787629635,0.32887781905196745,0.3911318919539609,
 0.9868479639450329,0.07850688508684822,0.882980725565515,0.871128277484144,
 0.9038062840360594,0.6521282400891613,0.045672451356715715,
 0.15010662906141092,0.2142347214146194,0.36718472475188446,
 0.7792994032119004,0.39991127746396227,0.7123823495142079,0.426913165270402,
 0.2698761098804704,0.08202329516163198,0.7274516108073382,0.8948003095361267,
 0.9072425844994022,0.34565103364402816,0.012209787526932026,
 0.9209771344677188,0.11353808760242601,0.2357597992806365,0.8007921135770523,
 0.9327423588268163,0.6500637624673696,0.03390480069642765,0.6541867667090212,
 0.7169803146060719,0.6576140747225078,0.3695273255116732,0.7350344294049546,
 0.12479956408207504,0.1665439160253981,0.7860645328301635,0.3533070353958917,
 0.8857719520523879,0.09603795792713066,0.7305851572878155,0.895006056943845,
 0.8354340092397461,0.11302560453524702,0.09872998827610002,
 0.7062015736325965,0.39993301330296616,0.4452063676853356,0.8727906616845497,
 0.8599450696160543,0.9789369506053347,0.7921657651388722,0.275197949388878,
 0.11693884495390838,0.8584815564506857,0.6383320123872849,
 0.24523000719105004,0.5942257875892878,0.3632945281939467,0.3863845317483101,
 0.871969027045431,0.022103391170562148,0.33457895509930946,
 0.08794083741628822,0.41258907903975106,0.9899601793622173,
 0.6798321896220286,0.8381357351970684,0.3429208947014867,
 0.032472693322901924,0.9095411029481584,0.8048802310938518,
 0.014248569241117837,0.7899342910601066,0.682812018131232,
 0.30502450548274074,0.43807998517293734,0.5675871730606122,
 0.1856772692293429,0.7391043645149258,0.36282447777818305,0.8388400239973426,
 0.880614130310921,0.8094390863315299,0.9081264722280716,0.7924842575700745,
 0.33458661847585536,0.01749422899406783,0.5537468215608006,
 0.7213191331981106,0.8083866939479036,0.7911658813480714,0.20667426217200013,
 0.4230975231251418,0.8559032323184719,0.1527773881481308,0.15332056653172865,
 0.20453895331052263,0.06918423664422924,0.4354947400028568,
 0.9466004294285352,0.14599777597742503,0.7765459355908666,
 0.058475836957327165,0.9905447756216602,0.13780083800885834,
 0.8978432733000528,0.11072132880641694,0.2063148660409978,0.3513623980810954,
 0.21467413300346638,0.2790537996146626,0.33151277161440096,
 0.5440257638231599,0.15260678042901965,0.20252843501631346,
 0.5650368746898194,0.6097756838491442,0.0035652513349332393,
 0.7575630753186636,0.04061324340229899,0.392003957830211,0.09778195330670802,
 0.055527561894559985,0.37169305519536466,0.48543138305525624,
 0.34946806940409136,0.22714885978157862,0.1431181617641184,
 0.3064162975755882,0.32644024909760583,0.5561276510416986,
 0.41356737446117564,0.05967453503653497,0.9583557677770406,
 0.6570705907999187,0.6565729172294483,0.741332119989379,0.2899871580619868,
 0.24256493591432782,0.9478104717486098,0.736213067518297,0.12777235026206935,
 0.9233593430507241,0.4536480098810659,0.534092786048653,0.6172442855622274,
 0.9326202351702688,0.36121509549287856,0.9858943008883735,
 0.32984892895131446,0.6599313193265748,0.01925253557524087,
 0.13978084674432556,0.41858284487797304,0.34631679139487137,
 0.9618976302886245,0.5544460099837512,0.320036032128417,0.2702440727120361,
 0.5029955213169034,0.8930421276518652,0.40632293147646603,
 2.802940660964204e-4,0.3028488085681703,0.5200559144408958,
 0.48542015915924674,0.3312780413907047,0.2895294228967036,0.3242521638497504,
 0.972788711095606,0.26893995456173003,0.28286813689055224,0.5778128200279744,
 0.2714466156027413,0.40550474325557073,0.661125344680287,0.9476287123818781,
 0.860999694441533,0.23141841814612507,0.34201227691118774,0.7352081444432809,
 0.5762594378334981,0.6106502258212639,0.35995370086007317,0.3927408495359985,
 0.20832849280538912,0.16885117343523415,0.18382313862326394,
 0.9944952883526976,0.8797479037190687,0.8690025910011252,0.18103775252009136,
 0.7815839466805081,0.02718418846786652,0.8489225137490358,0.6467381373515826,
 0.4469093656971309,0.73643341098946,0.486300733589483,0.22485348601851807,
 0.2073185928513327,0.4795150738597136,0.8392452568250344,0.252266669050086,
 0.9770975486422893,0.04833388589554699,0.44463382320553935,
 0.7872911813562011,0.30733620763227665,0.8684065553083232,0.7094690316031743,
 0.8320365674422171,0.5940627728343368,0.2037151756824669,0.4909334019512065,
 0.21809866564572422,0.28985927061270667,0.5794891411262487,
 0.30892066622521985,0.45767114172061507,0.9566194352579319,
 0.4644222537436563,0.8112703652833365,0.12510588976227732,0.3201122183469517,
 0.01933091063524195,0.31016039448877997,0.12136078252360538,
 0.7034451410508575,0.6156403120772835,0.3784622382656072,0.11630750608680829,
 0.25058317335387825,0.26713911582392935,0.04681826657944277,
 0.8977331384715856,0.599075292308946,0.208086431038773,0.09345589128144449,
 0.6365875315950377,0.34301870364072273,0.05541522564863233,
 0.03595319543524844,0.9889620784112333,0.5957234752931495,0.6528252302714634,
 0.8844510166223606,0.615406675855484,0.38273801206685076,0.11571656498512373,
 0.44974487690987763,0.8203512520530325,0.15116348601050322,
 0.9863160981596337,0.25913680415677753,0.1641316329846626,
 0.39573639585087805,0.7458107409031675,0.7522175295930965,
 0.41338898152186765,0.46921356625437016,0.7471982522982411,
 0.9178605698092182,0.6412442823875069,0.47826021668386964,
 0.38958227367715015,0.07248858864874907,0.04290832133849454,
 0.5540785373576017,0.4083566767301181,0.5506649896459066,0.11111095149748529,
 0.6501188469202702,0.3247664407935845,0.9267240317802196,0.9273051936318897,
 0.47582488762767783,0.4361061787892728,0.08527918386782718,
 0.43043990205713023,0.059865192366114095,0.7832320888460991,
 0.036644404532788055,0.6243007898675793,0.6954122069070581,
 0.28956514934295285,0.2654898561780916,0.7719276404402557,0.2642525675677774,
 0.902034424320633,0.9211214151717029,0.19577984226798772,0.3321221817689284,
 0.1236316706151499,0.8503589165673677,0.5558791244785948,0.9173192493973878,
 0.8478231467631824,0.5077006693873181,0.2456936229321376,0.3233226224586181,
 0.24404880780096094,0.7091731498587294,0.36249527630942724,
 0.26956736461817965,0.3046302451264571,0.3347912114097934,0.1616690353276613,
 0.8861165978109905,0.1853122296698292,0.19069659270851302,0.6274506933520073,
 0.29125458171694807,0.37808911982535454,0.8054884421667751,0.754943247446694,
 0.731753337047056,0.925410655141291,0.08429034664031498,0.05948334244399789,
 0.6559245988266924,0.5487232662070373,0.35416790880744053,0.2547067976603429,
 0.015627385647621228,0.5316937499615607,0.9766441758672354,
 0.6947868334710228,0.21613906574173203,0.6482298367528178,0.7522822524786378,
 0.1684857896797063,0.7411472320348568,0.6492952230475685,0.5331933131438178,
 0.3706952548387229,0.49673137284984237,0.20803390670259736,
 0.7573000795252038,0.6018311875491034,0.39387897663790195,0.672307116080348,
 0.8128086825141413,0.8391062385267533,0.8107691233844354,0.8275500278092837,
 0.9067485188195258,0.4332012609716108,0.04201094707367514,
 0.08684060290465967,0.8613424445301202,0.33533866013102687,
 0.30147665462201345,0.7535717062634,0.17809183867769696,0.9028016652301846,
 0.6897665664647921,0.2842660846910131,0.8508488656313207,0.39418999196394533,
 0.5738445948465208,0.6358621688766286,0.7925221836215128,0.23153176068278158,
 0.3375300153608851,0.04532204336023726,0.07884255219383185,
 0.6033569164588481,0.024195626383117674,0.4626287865919456,
 0.017619202208366103,0.40596941405733555,0.3053681627836684,
 0.18378603426980722,0.13442197242979104,0.17160115359502848,
 0.7186724738607698,0.3146964037837763,0.22378033000236486,
 0.29516303456669446,0.09896640212895225,0.6584998031545407,
 0.06450919370884656,0.7264184454631469,0.9460293344848798,0.5262292330151039,
 0.15842184082670996,0.34783886328967817,0.08354533381121854,
 0.3488135823300954,0.2056709649588433,0.07238349437348812,0.7268546043426517,
 0.26206096516034716,0.758702034447263,0.47606203814364634,
 0.12633760588905463,0.7368438533807014,0.3822960353195186,0.795497673264474,
 0.4368467724178473,0.7947116834344444,0.6205001080507924,0.9843474738810759,
 0.9454752062580288,0.6581082817498585,0.6197484075288344,0.5158347980675848,
 0.32675717993511355,0.3320408551695567,0.4883216662345261,0.1398837821547575,
 0.07208760618898713,0.5779197879632928,0.4025807145829541,0.4989478359495959,
 0.32092826945688857,0.3339906026356936,0.15604217689897204,
 0.18031957062440218,0.43983515170446086,0.6656228048516866,
 0.8203082046975517,0.8063169800427127,0.0431414702851981,0.761046907068227,
 0.6958307681675066,0.541167142710056,0.20751597473288974,0.719410298372198,
 0.749067490194349,0.4211178343760652,0.35404857684779667,0.5792063373715275,
 0.6578046706832841,0.14949259574156692,0.37069602009290525,
 0.9626255642657268,0.42443933007961343,0.5859611372493527,0.4005049827823881,
 0.5711644422202922,0.7057075681181105,0.14705681630959,0.6800583392571764,
 0.4664370379545638,0.8472202129160891,0.9722828416934274,0.8219480983431873,
 0.7005498758539963,0.662981698164767,0.7066874230150981,0.4167345299217047,
 0.7673537653488403,0.906037697156319,0.7752121784062047,0.29160004104086823,
 0.8472642284297195,0.2913418490184677,0.35607289648050033,
 0.29229959050219156,0.09916925492354434,0.06588761762759388,
 0.4202318892308934,0.5138818557391062,0.5966769884837583,0.9493136379703069,
 0.5814375343474674,0.7751994623455257,0.4998388329332373,0.10480606117126168,
 0.30749673215370044,0.625458843997241,0.8038793388137921,0.4028223580709298,
 0.25687912609270114,0.6289934845096835,0.02344340488005414,
 0.6955869347345233,0.7623934474117593,0.4630766497064861,0.4059364394144781,
 0.6634381549430831,0.8262387387613115,0.09511342799683087,0.5413013967883221,
 0.45136008507293335,0.7156567431249579,0.8041948442821432,
 0.029215804843360837,0.7780952107770119,0.34716570992246165,
 0.9813408816324984,0.7262972960581898,0.25577369807758044,0.5920198121446744,
 0.029481380534838064,0.24116223678152127,0.469800951945048,
 0.09408298812887306,0.7575783728373612,0.1258301120310823,0.5595125293816329,
 0.3558375136673232,0.42043266753970465,0.7155681501545073,
 0.19246457823841534,0.06024037495249268,0.22041078636701106,
 0.10237652190870605,0.6295030541603829,0.2568906959861811,0.2768663400688266,
 0.8950839003280259,0.026937753392332064,0.6806425436895241,
 0.41567863717413067,0.7563950166898405,0.8370608228253505,0.5919503338980703,
 0.5705800870310049,0.5144677490817899,0.8385365718003605,0.848451726693686,
 0.8032871776016909,0.3346626952930394,0.4878623798219628,
 0.044342954273894675,0.32438775822997057,0.6851125213724925,
 0.46413897365618484,0.015125806727451674,0.9514984806141262,
 0.443391528090349,0.5636553792899872,0.7182422269836821,0.935035019793336,
 0.6246858310716258,0.49304098400603524,0.4833920461982668,0.5868760085955977,
 0.161432652013346,0.4659932884292777,0.6462707729008637,0.5639934434865763,
 0.40590904938838923,0.8026926522374435,0.6867654821904059,0.6266262215088444,
 0.09743350437658183,0.07356328271676882,0.6493923677500126,
 0.5931901222800987,0.379205650376805,0.16001036978208782,0.4154205952006552,
 0.38437267283211696,0.9260948733755012,0.8539594579916683,
 0.07324649814464967,0.5690663915522404,0.45875236477549264,
 0.5710243034113553,0.40747924809270664,0.9792867054506504,0.8770330863576606,
 0.42468447626901296,0.9549875224397756,0.20742759909620612,
 0.6763724785746082,0.7132070077300612,0.6032479460357173,0.6703896390036159,
 0.9780279055125425,0.7570746016426988,0.3208474595278853,0.31953355682737494,
 0.7257474648374194,0.7143051202263199,0.31530559588685503,0.383400584394363,
 0.8397340303801453,0.39146259229854186,0.6275595981638421,0.748081786281375,
 0.6705895122511186,0.09568077641601558,0.355596910650549,0.20495499783272664,
 0.2755762698699997,0.5185094354400123,0.7800162843550904,0.4337992191298795,
 0.7200620800379445,0.41426697775509835,0.1054888457636981,0.6753909850378869,
 0.3763145099234847,0.9260542330746864,0.3069612559674183,0.37203626127733536,
 0.5350310887498702,0.9234104213771972,0.7957225610038471,0.5813768622509078,
 0.8902371943625182,0.3377554583278909,0.5156346082124912,0.5456785558252488,
 0.029119924359182092,0.3508524639368157,0.21487824393120525,
 0.9934201059828298,0.5313713051488895,0.8916673220550427,0.8305186938776922,
 0.35781816343298634,0.2360930935807828,0.02235372802155655,0.73752919350895,
 0.5867595047127319,0.2622991255067755,0.8782167368743298,0.26306861976742557,
 0.7784695618796117,0.9539014288192986,0.715139112834218,0.4128602307376298,
 0.18830991339744407,0.9595629306085252,0.251721960249893,
 0.043966918268286914,0.966065825847747,0.22534694458869875,
 0.09070755922490548,0.1829994958775386,0.1640987948883761,0.8037598301429086,
 0.3339428448939421,0.0908418892827264,0.7092367995057677,0.34387540222796886,
 0.5540116203597071,0.35856154361611214,0.7321779779644433,0.4674464643108516,
 0.8718932321366235,0.4656930560573531,0.2277428635194363,0.15200122672887462,
 0.054178459125273415,0.11898179060040548,0.7029666991725718,
 0.2341304221406202,0.6196417462825726,0.9907036424255841,0.8598079394297016,
 0.9620140117800619,0.8090730667956386,0.19162320867390037,0.2679421790721188,
 0.043335871152565764,0.40291539236012497,0.6142588638111144,
 0.9998967504141382,0.8903491054546521,0.24505631450769993,0.4144619965750217,
 0.9361307809862309,0.49439353849580103,0.039699972739730915,
 0.8249037322955598,0.518292360340366,0.847106358366985,0.23833104494575497,
 0.016493195152240014,0.5640617504273997,0.5296366067159259,
 0.48454682504403257,0.2376058462457885,0.06938441544817553,
 0.15187774284432687,0.631583391247851,0.43800214433838325,
 0.027331045544289112,0.8835869320917167,0.697574227194362,0.6793262261570978,
 0.11878256623642991,0.616536581578693,0.5322698443714065,0.2882522549780395,
 0.5793654796725257,0.09417556193786558,0.709408445322906,0.7712875522122703,
 0.06841618927287463,0.06120064611472942,0.31689619488640064,
 0.5492537686229777,0.4670447766907778,0.42655522834328263,
 0.07567924886901733,0.2990777746060964,0.891247928994673,0.2138488286562069,
 0.5547306415739798,0.09820164533731945,0.25105122368675625,
 0.2137600529658178,0.1043954904160459,0.6134529122640516,0.4807529660186195,
 0.35885447409486815,0.4115549439009206,0.6256739812147499,0.4908196154868705,
 0.7968156337403967,0.45312258915272496,0.2233664192058189,0.8144867003245267,
 0.2604978687557177,0.426780511936924,0.4597443773804315,0.38433757333595775,
 0.8825068663349822,0.4629466674457323,0.6702993702422707,0.6510675677286588,
 0.6401539327749916,0.8168188100081594,0.022979505459310268,0.968465720334476,
 0.34626536195925894,0.37510324929746097,0.5968433911939638,
 0.6681847423785073,0.023384548858213505,0.5417144666285648,
 0.8604691678432401,0.7213102551305707,0.9192606414105723,0.8756914255011543,
 0.8321626434286538,0.8074012675318719,0.19511695387812034,0.9723913706309439,
 0.6951569319904839,0.4831504091368124,0.061398932132876394,
 0.28610635830806763,0.2631707112488153,0.043067664322680166,
 0.46075584198233677,0.689247540025482,0.05361266092824335,0.7700398209554141,
 0.1898368474176496,0.8863176949729045,0.16749297323223356,0.1755770959254208,
 0.7960999434203513,0.6583908688019742,0.2437298943677071,0.5325274554844783,
 0.37656275079139423,0.972942198742801,0.13199059645107092,0.6670758835734915,
 0.5539924171006074,0.9471528179275182,0.8869617888127554,0.8827112069912705,
 0.24384576170700512,0.803444220330451,0.5595367582281515,0.10337557439008349,
 0.6397093302732106,0.5542351784671743,0.7886060973905211,0.6578072182444431,
 0.41140596734853785,0.5400049681070814,0.6300540818085472,0.2608405347056567,
 0.30507634900466085,0.4079637347769457,0.1850360247615518,0.7260658194066497,
 0.7127495220239312,0.5178369635827432,0.2352768340836222,0.7618977476641915,
 0.3781600155188666,0.5399057259195202,0.7227886863890568,0.49481240847929564,
 0.20694547981134825,0.17519018403863362,0.7728825891716996,
 0.9146360578302406,0.3621357592376233,0.4438812453849217,0.9046290129689089,
 0.9905083295901423,0.6512084528652423,0.10336139574588088,0.6667686836007741,
 0.7690547911510878,0.01788670667859127,0.9609664466347032,0.8987384971696761,
 0.4600031625439288,0.5190863483694056,0.9885783008936796,0.4410886709880445,
 0.4872483720508701,0.3572781610072655,0.29022774766671533,0.5280706553983073,
 0.05603902303836317,0.22419021397752847,0.5023147313509824,
 0.8788408189349276,0.049749618985930666,0.3856659475209132,
 0.39076853619739893,0.4302974427701608,0.6452118176423459,0.3749022623792193,
 0.006621790481219358,0.9344162812943081,0.3255651971517636,
 0.08388784942004224,0.4470989654549977,0.6964964715951321,0.6375596722225617,
 0.6449148129364403,0.3817832541472044,0.4968215534099517,0.4627924284542052,
 0.3200326299863797,0.7582670393809694,0.07643061667422524,0.4426155314821806,
 0.6466689784315391,0.09986772273451705,0.2570342220185873,0.7172762184298678,
 0.08961367844982426,0.7336988755266818,0.6991474459302616,0.1082583840348984,
 0.48992519604548,0.9542865836943897,0.4849714581555564,0.8371728222196182,
 0.359848632423754,0.024508357797213498,0.8507827971123288,
 0.27466351055590565,0.1365374286369232,0.7319293942680493,0.1506097830112848,
 0.6410872544471119,0.016589853494293094,0.9882165315205783,
 0.3602256660657496,0.8817106341944653,0.7019239602140105,0.13353785756161152,
 0.7228993508642759,0.9290000563874485,0.3074374577582444,0.08621212630383546,
 0.40130714444670146,0.9426490201259716,0.7269445706993944,0.237254591309642,
 0.48000857946698894,0.8656520786963287,0.2578456911291117,0.4041418021057708,
 0.15486200157381436,0.1599203304037522,0.05618172083415729,
 0.051655706923146694,0.49699188330903943,0.9431711747023015,
 0.1343739110936476,0.04977095032588963,0.3687452687462863,0.3705938755700723,
 0.9533086771519654,0.4127113649906081,0.3475333240113756,0.5556240330659743,
 0.2378078257515689,0.4812286272557821,0.27726798415928267,0.3205050094857792,
 0.3966111352723538,0.2653003253611367,0.281173811403045,0.08730849258718609,
 0.8878311956241207,0.3122877437298719,0.5134625702957974,0.40037703143096415,
 0.20406919584568606,0.3298366149031158,0.8908393822223902,0.571363009358004,
 0.6882013586656387,0.023231795630995866,0.7858866543689186,
 0.5394598475180883,0.8738072253900222,0.01642313970712861,0.5913532838679787,
 0.7080986619636551,0.18968515381141737,0.05152029413654913,
 0.21169092258161648,0.19401262912591344,0.8251710956696269,
 0.3138014263882918,0.28061508132465807,0.834103771854199,
 0.030317872618924713,0.04373728124645371,0.37077085548762323,
 0.01864296325877557,0.25519691813417633,0.7637743284383401,
 0.6190593098354028,0.6931630811768971,0.06919309086388326,0.8952402468269347,
 0.10784285616258127,0.29792487848711047,0.9005518557989647,
 0.4543555787338097,0.321890476741601,0.557710132572139,0.1830178882217236,
 0.8088064794159406,0.4125745288298972,0.26435553551477875,0.2384020771994554,
 0.9044940529356147,0.25246261267341363,0.8955196869451689,
 0.07496794041802479,0.8949039319926583,0.6543825097560232,
 0.28250355089250845,0.7115769034899313,0.7751978535524544,0.5560325966609123,
 0.671566256428626,0.08021090449660118,0.9800497454782783,0.7339767229382508,
 0.8997310347185355,0.6826689294135946,0.8513853522172793,0.8163516403642886,
 0.7681191958432019,0.4656268034790304,0.5237833027607527,0.04058584926088804,
 0.4129158492188383,0.5257316574046219,0.2523988948481377,0.5967662726394714,
 0.3909747621328399,0.2374579986398554,0.8774372270326347,0.872503619208394,
 0.16650549090780065,0.11395437110895956,0.18201930549199374,
 0.7548385845791414,0.8410995941349174,0.8310243925425108,0.41568610912076087,
 0.9733859726205459,0.09899858527564032,0.5302980068166053,0.7131460705237137,
 0.7482709713268789,0.3597028939936411,0.9761511730385136,0.07974056133181284,
 0.2491143166379206,0.4418954258955645,0.12279801260377732,0.8482547902230421,
 0.12772689126101433,0.49681781984702633,0.22873158002260296,
 0.22982447772093362,0.44944719667939936,0.7352495139250481,
 0.3331127778564281,0.7060829997300347,0.1651429584744385,0.28910359525828255,
 0.7163406668484573,0.9441205156101808,0.9379626884003377,0.5959251777457849,
 0.3535084131095859,0.36379519468314103,0.5390494057761978,
 0.012544437055972302,0.9460742949521037,0.2781283719392521,
 0.4868016038176748,0.9399758495619014,0.34445033219831955,0.983880803739505,
 0.27110000971427417,0.5458030365591532,0.9378360925584234,
 0.25433515869133505,0.4142279725231657,0.014792851768883453,
 0.5522747503727523,0.11969998115266933,0.7743758187244426,0.5079012315666516,
 0.36982678509869393,0.10967021674045518,0.05565780556882394,
 0.15046159353683886,0.6794235030973973,0.7090294864989073,
 0.16577460312728654,0.7203880578966761,0.10650790755717388,0.489340978021676,
 0.7745807776803078,0.3638880252799077,0.19241948735940462,
 0.08723407335709144,0.22569263635846704,0.7675861600158838,
 0.3494714014649276,0.6968499026426201,0.48816935530884864,0.8494707142926173,
 0.14882096980955806,0.5999649299528675,0.21230118604025572,
 0.26912358749194576,0.42962390890440205,0.8063938290536857,
 0.7736746054089341,0.46916478132964046,0.9762653280085201,0.3845928621061002,
 0.7811582647173133,0.758465747488639,0.4881842732635546,0.9391560885222364,
 0.9334976508787887,0.3940251035867124,0.6613539730460151,0.46495058155684066,
 0.746071536513624,0.6466220438896662,0.609959720488721,0.30365820586780135,
 0.639737160877026,0.41569816317639974,0.36900255851529395,0.8676197304366163,
 0.7702401787310797,0.12894743907262352,0.43725334632697543,
 0.7894308647622582,0.2803785441318001,0.9794401542459091,0.6635810591859381,
 0.3279487631293917,0.8704518833772295,0.2442948339783334,0.32827846454956033,
 0.783938894046164,0.9951405088253243,0.5538503467745515,0.4505589174701877,
 0.3595255374036411,0.6671887405204308,0.07665794876938126,0.3318978709997046,
 0.9684030937463499,0.2874870570703182,0.49305942769155686,0.7051344747198292,
 0.40917538267598674,0.5397910490041387,0.4965177498671747,
 0.41404558954901527,0.9701274558882385,0.9467094752196458,
 0.36238898523173346,0.32533363266452675,0.20181880564677857,
 0.35763297429431173,0.345922495351368,0.6105329432348275,0.8063891863675597,
 0.3480143333424447,0.6403656119388381,0.8684573309033667,0.23342848761098833,
 0.30860579511789077,0.3319369063984159,0.5220771948073349,0.4533397221338613,
 0.9529942690128038,0.8097200860909737,0.24334720955077827,0.7182890386359552,
 0.8388920789437011,0.9736130098370479,0.052178717832881194,
 0.9486176249020558,0.063238319539479,0.6669772352363373,0.871635174079593,
 0.7871981253642184,0.7409238605736238,0.8999988863844642,0.8108856043242669,
 0.17909389096978645,0.16635284432560038,0.9060405174553187,
 0.8855438819477773,0.8754190781911382,0.9895952544919686,0.3385644898916593,
 0.6721703903979473,0.7992495361537191,0.8811917123952491,
 0.011244559998095593,0.7017941720431089,0.21188048447674168,
 0.7310881236834698,0.2590235611570394,0.5112659768352482,0.4303903923637393,
 0.02039398966963968,0.3657597329417056,0.7178449824944116,
 0.45340918140128905,0.2903554393886304,0.38177599432672815,
 0.7899103721158188,0.5010818669217355,0.8901240419054075,0.25792452820841416,
 0.9611283787681159,0.5728243162269149,0.5099448538047208,0.2264155902230458,
 0.7325672102018004,0.2515898303479239,0.11053884276789505,0.4457956664293061,
 0.7223325741496045,0.574666747255472,0.8950976845717938,0.17565582119620127,
 0.6021271844249119,0.13320154952839058,0.4824623919684453,0.8890484472773881,
 0.3998725776268811,0.6253266996740797,0.30484996164400113,
 0.42509036600971595,0.39037993571579754,0.5157252291973644,
 0.3548103109135958,0.19092786241802195,0.8695470984079346,
 0.07449771915798709,0.6015949152094919,0.48953104356082,0.9040658153830922,
 0.5486589947865873,0.2709537099778677,0.922008547607572,0.6136249170185794,
 0.6898745091095339,0.9370427728339837,0.3826967545001395,0.9721163373842732,
 0.10414769539430502,0.6741651710411402,0.16324081368102028,
 0.5290908722894725,0.7561656591284002,0.02505386068706794,0.6840227673059116,
 0.6523466161096931,0.5959088028929475,0.5758608741485476,0.6465670232932571,
 0.7058719472205058,0.6796580985246271,0.7421096089148621,0.15232347228913157,
 0.12084886111091864,0.16695333775070587,0.14363856971494138,
 0.20334227483001066,0.8335537790322819,0.0010793749839439482,
 0.38793597038138994,0.2759586522043843,0.8601727901903314,0.6396523788867866,
 0.6064855504941231,0.08335431253478265,0.975162022888724,0.7885894273962745,
 0.03875549620438001,0.6890738558445897,0.297938104010091,0.2771430932724309,
 0.28813099125077946,0.3398115286624993,0.17242856412766439,
 0.3257746842457495,0.1579894038068247,0.4520482989186825,0.9502946606214414,
 0.8143299597246473,0.059961091874383365,0.3118046358649722,
 0.39286973266287406,0.7858306117937394,0.9580184101270925,0.4454666503625102,
 0.9259226018444785,0.5619763106097336,0.7343835417645058,0.01169599254890763,
 0.6004633213362613,0.49890031939266877,0.7128320141860504,0.6411546163833232,
 0.6323704370481463,0.8557345677409381,0.7137061702559379,0.6567653221081985,
 0.4578310931366496,0.7782915602003575,0.4799000311428715,0.6112234734819019,
 0.30722126695722296,0.6268331755963749,0.0636381672080466,
 0.44583594480617794,0.8315730088376367,0.8273814895005807,0.7527477821073111,
 0.5709352113752945,0.9308609076246513,0.6999330996043829,0.4835305755161513,
 0.06554317448302849,0.5192985665600116,0.934102549908905,0.5554732656959439,
 0.050010689926636154,0.054489459229463355,0.1530394822485105,
 0.7338025879855847,0.44578459488391986,0.12273532847288404,
 0.3649442463075706,0.6434337359290362,0.6563859020324598,
 0.019911092386201923,0.33141399522115944,0.9955898657099325,
 0.12615516569048857,0.983191033409826,0.9473948752277053,0.9959983796483127,
 0.45223048348310757,0.25460135370781145,0.5002486018934342,
 0.6413626129131654,0.5244671890728445,0.3277221546848663,0.10300271126113425,
 0.5636734225482547,0.13381453292070333,0.6172647920008932,
 0.37870497221761623,0.10278842799755461,0.20477864476745133,
 0.44292006790435656,0.34983654629061245,0.5103554843669711,
 0.9429100683899523,0.7695854324918451,0.6369297323406732,0.8205486200761326,
 0.7547694485425507,0.004363542627499717,0.49630175539317045,
 0.6868902229414525,0.08675536496893643,0.5693155096406347,
 0.29183649692460556,0.6746016585749195,0.02502967293357561,
 0.5403427354168461,0.1528155401325696,0.010230287853700615,
 0.6789517658989465,0.5206070495188764,0.7435304461267174,0.4840758257544361,
 0.38936644962494515,0.1358691542176853,0.8362048725684451,
 0.20669794895457627,0.021689768252717068,0.9408190589298393,
 0.5240775915890319,0.34251880855883066,0.3970151184621873,0.9679749995874355,
 0.04887729225730164,0.9104557426013427,0.7013701865440449,
 0.11817978621258352,0.8608694904904199,0.3162976935599514,
 0.32164948919169045,0.8904603203254535,0.4788590574306161,0.234251471840766,
 0.00534785475233901,0.8338142501768575,0.5820005077037492,0.3089444767254711,
 0.21086933623516624,0.605985894702413,0.4467784988810559,0.3364628216608635,
 0.1526306214249703,0.10877106378087076,0.3676642276301818,0.634551510014658,
 0.1536313557722856,0.8663044258532042,0.46500708817369074,
 0.06874120931613292,0.936279352273275,0.8060961188550672,0.9661047355634494,
 0.3923303035613046,0.6282977637124607,0.6335198590284619,0.880636047929647,
 0.3267301480894107,0.17173784625531152,0.30203338146780034,
 0.7593316186588395,0.6557040392162862,0.8537010396296097,0.03899611524242308,
 0.2821396945028637,0.5644586018774533,0.8849074757857647,0.3662643108572907,
 0.3688822400337314,0.25692632371450663,0.6356872719323211,0.9505085915807786,
 0.761271280231362,0.4047484633084121,0.5340317839616069,0.45843497203606165,
 0.38284770576392435,0.5017937928604941,0.8046718294756464,
 0.07055818697841798,0.2807625658866708,0.4109270283767641,0.3779123702764696,
 0.7443508589672811,0.7122919291502304,0.5168369003522897,0.7951950153004459,
 0.21176116887194474,0.5483489989658208,0.30790861454104723,0.1694341164241,
 0.12732332961792414,0.22020167326103957,0.7288967717925351,
 0.18837460901946035,0.8601992050107853,0.0973172561882415,
 0.14118167740867793,0.0255566853159378,0.34161883379976665,
 0.8723696364419249,0.7895919110351489,0.07229631432821337,0.2705705938891383,
 0.9704097333069215,0.8664654921211391,0.040916115417729215,
 0.8279543014961884,0.8693511779011557,0.8398165482401482,0.4024970400579527,
 0.9200747733361911,0.5492418664503543,0.37674486010713115,0.7519085936170391,
 0.5740332578421614,0.33558001834713,0.29390786024230553,0.5327840597139416,
 0.8099005177265971,0.6454139779982107,0.2844473102910696,0.1058397174179353,
 0.6697428426508658,0.7727084599060048,0.9740542523008603,0.8757175168525398,
 0.24007409263546908,0.16019489877043724,0.4772349062999588,
 0.07867162184528897,0.31741036964634894,0.4785523972013047,
 0.8459646783598361,0.581443100005486,0.853831944213965,0.4524000841262521,
 0.42683689398856006,0.035809983990129246,0.6407651355232525,
 0.6337975631092864,0.5622944505365879,0.9678206553360345,0.5509172742453158,
 0.49978148814710943,0.35247523335306175,0.5462837134747833,0.931516808102444,
 0.4412310950074275,0.19897875541432408,0.5176922077322682,
 0.17863998268464232,0.5939298453568771,0.19655502921606394,
 0.20480971956644978,0.11771930088881877,0.22733442191574738,
 0.6675234473348777,0.07133645534373989,0.839670534627879,0.4287002398315838,
 0.12655295790479903,0.9417008092373362,0.6398587109801414,0.8449579908051758,
 0.38244123860703405,0.9820851491215394,0.033090625099695736,
 0.47280254038904135,0.8990704750179135,0.48653497388479006,
 0.06569124579143992,0.48754436951865,0.5815711816164927,0.3747804978503766,
 0.11210550056880375,0.18348385018555302,0.19592717880662203,
 0.6428548220913095,0.4332851074246101,0.1338013422993124,0.13862305253496676,
 0.45061023601861083,0.5200599010047209,0.4187709919719482,0.6323404123025341,
 0.839080115504946,0.5513161806720166,0.9284517069211806,0.021755609227140926,
 0.22699775913958087,0.5710794863574216,0.43283168763029234,0.798988193151409,
 0.13190026649566666,0.24453877241928523,0.3802593964858244,
 0.7741799865984654,0.8591801523435019,0.16233572133971452,
 0.34767858052415446,0.4019068399091362,0.3681612666044427,0.7833714811807545,
 0.9648877704423137,0.1643324557590694,0.9689837058622932,0.8098957243972472,
 0.13546134909186303,0.6987741674770271,0.3332206017218089,0.9096045238235136,
 0.5673161340949698,0.6250427804452763,0.08499384018325684,0.9374896771567373,
 0.3069943331490015,0.5932292639932838,0.720344482713566,0.1348325243008086,
 0.2564777539798996,0.9205382567013053,0.7211394350767854,
 0.015202481322177941,0.9707927463757494,0.4727101858554297,
 0.05849786316513517,0.6970435715279227,0.9851738605159285,
 0.18958516461203945,0.6040584104103229,0.5039490953058053,0.9482516126775987,
 0.46450180228568916,0.8090211266370715,0.2171769769322638,
 0.06274177855896279,0.41055669736876566,0.9666988030067661,
 0.8868651743761226,0.8858820935075375,0.10851288605761211,0.6911538698438273,
 0.7109128156237444,0.7091682543660345,0.3080064719487565,0.48624590806973733,
 0.7021837829666387,0.5943572306551652,0.2940665354572213,0.7032955594879134,
 0.33483332215265527,0.33872826169641046,0.8940634888658638,
 0.33719504693839375,0.13955750259438893,0.9599111463470269,
 0.056628919674573064,0.714785660467261,0.32300640963708427,
 0.5185600587311938,0.17405275462926073,0.6901784862742014,
 0.25084172691925033,0.1670255704385315,0.7093788984130558,0.5838436880616139,
 0.10449004092729741,0.16404047595418103,0.9531631452348938,
 0.7251098587219346,0.3218075967146422,0.47716591375352435,0.5696099009035845,
 0.33114534494703307,0.2879283983201756,0.8953559643884756,0.4847642373554405,
 0.7607926312278216,0.8575158001649155,0.46004877467582517,0.6397381157625992,
 0.2639917873998523,0.34811919090612553,0.36343236754314234,
 0.7893410715602832,0.07009392262308567,0.5994056925205917,
 0.27502958307905023,0.44788614561359286,0.665593404455552,0.7327667695852275,
 0.31516321785080414,0.14606451274914134,0.03464456596062071,
 0.3252812954079305,0.40497908812475625,0.6742333530627197,
 0.42082082662221887,0.28428259918477883,0.44080627037240017,
 0.5673480956632194,0.9621120573298413,0.6301029986737003,0.23691596788903702,
 0.5957475769745522,0.12349634561739686,0.48328437001953617,
 0.12346350025856268,0.5350784803867975,0.7700209703013348,0.9914226206219386,
 0.9288037353105529,0.9499417374466185,0.3508327619369913,0.7629636075361413,
 0.3111155866237326,0.8040283495601337,0.7227393713995013,0.5671612248098258,
 0.6435513447422444,0.8110160585591696,0.7484712262029919,0.21505053291198228,
 0.6301686061983027,0.007238359373433978,0.4244294249938684,
 0.25319638331243455,0.4876930970817628,0.7307693909179737,0.6293320613227625,
 0.5846594523318943,0.6217996093526275,0.34450731583025884,0.8760620261330901,
 0.18360926922869725,0.9919814332576378,0.7569486806352281,0.8198034914256283,
 0.7957536420209532,0.5445134074043547,0.4906202754049779,
 0.015226015369612766,0.8016843468640585,0.37378365975017835,
 0.4408073708261828,0.25685503392567544,0.03600881257632382,
 0.8537931820615128,0.7433561086021051,0.7717279401137123,0.3120851640254494,
 0.530327999334655,0.27039393633852615,0.096011780925636,0.4951396056314936,
 0.23678301801331836,0.7796548622349944,0.6990973318938106,0.9836594560217534,
 0.9712615024920703,0.3682172859979841,0.29142821573692956,0.242938498958436,
 0.6836112404554923,0.12494145645651566,0.7038863344958375,0.5496242871798324,
 0.7353589935170595,0.3528744149075307,0.9577679496412081,0.2524087312592408,
 0.9480115457616893,0.5166980537959923,0.43209790130557924,
 0.25463860411509287,0.7501094024703667,0.7656951317454543,0.4404306189313899,
 0.9923423188688217,0.4611609283235568,0.4519716807907692,0.10380264268263573,
 0.2148104872878447,0.30582200131998427,0.518500186614771,0.3650998915271527,
 0.798684288384603,0.8783155168259524,0.697519461088119,0.9340025732878464,
 0.8069402727289254,0.6340296082635782,0.4991529811077511,0.5218090298855288,
 0.7575080302808117,0.34162665601708675,0.8768723622658924,0.9592358260874896,
 0.6072140548332545,0.23575075948419455,0.9657353558463382,
 0.24952073818114318,0.2236065202277585,0.3355985497989442,0.9123730903649853,
 0.8144349766089503,0.7543153139744956,0.7017729545231086,
 0.038431577482245904,0.8257670291868345,0.34797503609515323,
 0.9102321809496194,0.4046270790132298,0.033073949719913176,
 0.9818740349121945,0.7249219712954058,0.5452622141819203,0.2591436163403358,
 0.778939836016465,0.525250666449375,0.011600160674844928,0.7430490199838445,
 0.8015016989389459,0.020137718118241277,0.09376591801928003,
 0.730637188571951,0.8727109958947067,0.6757447935065707,0.021377982899149828,
 0.8023801490051975,0.01754689566436518,0.8836346483723355,
 0.36800438497317955,0.04160136318505625,0.017124718403475847,
 0.8474935014935892,0.16584840298857106,0.8878083745698538,
 0.13945114503740297,0.7681447668094191,0.5251371610264625,0.7708629727988747,
 0.2039902171515744,0.8100182788328508,0.7401587132361374,0.4373011866973151,
 0.6283807902144064,0.35149413364160365,0.8766231806731556,0.8926760482390761,
 0.17622910583289342,0.028938903718946474,0.6541433248137197,
 0.9841782359701308,0.12351248298229234,0.2586835147775952,
 0.09958634297340885,0.19574308610426794,0.6067095401351406,
 0.31739679391654596,0.5993509462031876,0.5723041790142204,0.5093119404104928,
 0.9477822315249678,0.21690746854563348,0.8975061614123204,0.5849834998613487,
 0.6490922998652593,0.14154318308081937,0.6078091596143393,0.1269513330105302,
 0.652859267623771,0.7051138346888636,0.5642132786407346,0.21313227910369426,
 0.2230165911344073,0.43382533768660125,0.25866488510894925,
 0.4221836877939882,0.9110664906113722,0.1036484569181848,0.07411292013224946,
 0.42137230422719607,0.06727729171185215,0.5498112567110521,
 0.03977249156154483,0.13251979347297194,0.9017015896191158,
 0.08186706307199632,0.3858719728077189,0.0196224835393064,0.7026386410472705,
 0.9059793737115459,0.8086275558306617,0.5197194745614322,0.47847455210839906,
 0.9091355041925676,0.2727931458204934,0.6433559630873713,0.5830152660785429,
 0.8617607736400232,0.4050412270549636,0.18902428745191702,0.3474989579296307,
 0.6286103473448929,0.7933161871605814,0.04669122739318032,0.1588478463467361,
 0.13277728781383025,0.7601354048266245,0.8786037862158939,0.9018926757611785,
 0.6051292891018497,0.5833783060631829,0.3465406436984616,0.10694611003004839,
 0.8114717131664351,0.9640450130890431,0.9796639971360932,0.598363656768072,
 0.9498912449252435,0.08932261116613638,0.8767889815221215,0.97485081765857,
 0.009234757723494691,0.6604137267137097,0.07430412945339904,
 0.5053149238104871,0.692167065805193,0.5817385888361418,0.5445152372913415,
 0.893336654768897,0.22650027326916755,0.04045340387660212,0.349695443004271,
 0.9371895908273401,0.24266932535508834,0.1998789730940752,0.7281300257941095,
 0.6736506545992846,0.30130579129481705,0.7295315074939347,0.5212208226470103,
 0.15908248939724656,0.6173778501642091,0.7679577893207875,0.3159811106972701,
 0.2646865954776808,0.46811084404286185,0.006743386916728267,
 0.8726511784226356,0.21884295507646023,0.7755744621633508,
 0.14968246917219052,0.04627078543908891,0.9200093995355075,
 0.018631639395511046,0.9241674173096303,0.6063213893444171,
 0.6945491446196091,0.5570810648229498,0.8530288965940704,0.2987956961824908,
 0.6040409817057595,0.960926279414627,0.26664570669752985,0.33760991805469776,
 0.8656902608169963,0.7349880248821508,0.19976274345829326,0.9679402780655312,
 0.35926468161418035,0.8550979958777559,0.03167504404964594,
 0.1224666758371169,0.7682131281283053,0.6389307587861758,0.31188185458596873,
 0.262238210467849,0.8992630916219132,0.3281639881311985,0.18749810984996151,
 0.7886052663752429,0.3423718744749825,0.6065221990112888,0.2417445711281867,
 0.3010633939736954,0.9901083015231713,0.9542004031104742,0.4712646235145892,
 0.9878271073368928,0.48082500441143583,0.199722102199833,0.7022050682260258,
 0.07770566508317922,0.04339080288322461,0.6748585561142603,
 0.6368663955401256,0.2215699100652745,0.7805316068263183,0.6102697652019721,
 0.8365823485589726,0.9974549928921161,0.3216376071993945,0.9122796322793592,
 0.8652550119506063,0.4526716209975987,0.2073143978580323,0.32327258460720676,
 0.948821206755576,0.6119813101602343,0.15251092214207718,0.638649695182458,
 0.11764388623169664,0.5464262493079567,0.6101415339904301,0.5775882783985644,
 0.12542761918901846,0.38036724089304885,0.23357278521290958,
 0.9098468845136111,0.6477054467810123,0.6907621249797059,0.986876194696619,
 0.2760731566880533,0.7081234330497308,0.8188287868783379,0.6883366629142192,
 0.39632296259623545,0.4597421441132905,0.8896978239975799,0.6930903342795482,
 0.6785466561999558,0.7724151358586381,0.17806645209245053,0.7083857869404102,
 0.5009733801294326,0.20530664650730046,0.7442633185881473,0.9947769329331668,
 0.8443418643573124,0.476821386910662,0.08862305427155504,0.6470150797930762,
 0.6909300668045272,0.7281391000517177,0.8601947201141449,0.48380421954717845,
 0.8461163707100355,0.4937613912036889,0.4462590171905768,0.43197368888648624,
 0.025291247141300266,0.5516287468803576,0.8468559400388727,
 0.32644745344826287,0.8184883049395745,0.8469194117351586,0.5520979557984463,
 0.6524442784239872,0.4262297447890435,0.6228546399961433,0.6715817449112164,
 0.12234051721260708,0.2653899189748188,0.25920337467467525,0.597556984302527,
 0.871336148191665,0.3655114030672728,0.7255481401159326,0.38146126019340254,
 0.5275482218599212,0.5012226479734926,0.33133077921731713,0.7581590904764185,
 0.8128078982929094,0.4820117416534776,0.025708605584710553,
 0.7308494936999735,0.020566095888070235,0.07394486129478217,
 0.6484235673935761,0.27152682900777925,0.7597944561694194,0.4782749114717094,
 0.8207123769509215,0.9142172978693878,0.45066283496592896,0.4986471873372661,
 0.7559957266218171,0.22169278305798035,0.7413440887270566,
 0.45132510825465555,0.8511186014263706,0.11758875819175751,
 0.09932894195512953,0.6247489328085422,0.08361716435970756,
 0.6506431905671701,0.8652632431099954,0.085220948694117,0.7244695934243086,
 0.8102639260559001,0.586649949290494,0.34005096690669046,
 0.057299698433095125,0.7251204760693502,0.1809379041035729,
 0.7289515147841619,0.6833380655875337,0.014471371647267861,
 0.09258079776228734,0.8499359606411926,0.8128990298046925,
 0.49693880708214355,0.3821555096050743,0.4994192919018152,0.7805101361915876,
 0.30585676327813405,0.10967810132166145,0.11161386420121522,
 0.4027597949617643,0.8715132439127813,0.29028394596213103,0.8827562202970229,
 0.08644085915043098,0.37842856471470476,0.00975057830450532,
 0.7188302799993838,0.367536710376391,0.7832298906340247,0.9432876649249265,
 0.2965050315160196,0.7803893725463327,0.7858887787080802,0.23636863530649188,
 0.14909957234007698,0.19170845822912197,0.6626362872600541,
 0.29946995648179864,0.9067631085812276,0.18855248846028316,
 0.012635816211589712,0.4219284530896309,0.5848931074161479,0.313831061003343,
 0.9826653299856192,0.014098848317571466,0.8345039784577516,
 0.11394234187494878,0.6576095977349101,0.7939568684321291,0.478698653039088,
 0.27212781272687414,0.5597315951084593,0.2864870330300504,
 0.24111300532157687,0.19563119592698197,0.20258060662612432,
 0.6516797152941701,0.6391356864624501,0.8684915035546305,0.513458193737004,
 0.5227601171759443,0.4534462978630238,0.6091611607902605,0.07589854015911568,
 0.6040351157308436,0.7414785025108176,0.15762195259728895,0.3892516182842056,
 0.4321310066298104,0.06686914019539558,0.10850424238507472,0.455663972235715,
 0.8010441506065864,0.48653808698464074,0.15272067353031116,
 0.5406163736424019,0.7116268083357543,0.17406715432215247,0.6201659325757629,
 0.1964798158932315,0.605390065746187,0.31236546942560317,0.2895159194563641,
 0.7645293141019562,0.012211070370214294,0.6273727795512067,
 0.8246992641757901,0.32337639730398693,0.9472034491573686,0.2239956261406273,
 0.8448021988975242,0.9660241475291456,0.8276332492758262,0.9531314065947235,
 0.6434518899461703,0.6554027115096804,0.9163955615739642,0.07011726614139657,
 0.4095242434790698,0.24219486168843807,0.6104229852639027,
 0.16087607231719947,0.006681973746663794,0.4545737554871684,
 0.9612704666831928,0.3428919586875906,0.9545220928425551,0.6476909113918674,
 0.8419640827844245,0.6003961573782007,0.8411359520455769,0.6941942612926253,
 0.7096384151317804,0.8088250722903789,0.8991943085664501,
 0.023315054602764396,0.8737875910669111,0.9312880513032007,
 0.1749535717564824,0.3413600913442607,0.09483933590478222,0.9632360518592974,
 0.345088198914945,0.22968943542606135,0.5997075722132603,0.49083380330842297,
 0.22991575281362686,0.8710192184851033,0.01676734995625251,
 0.7437893405861358,0.1510700502133776,0.04126428406960014,0.7493838223832041,
 0.486166350950495,0.17615947260913034,0.05469152843156555,0.6821842141103601,
 0.902092516251964,0.017470260707326712,0.37140902605655857,
 0.5153525213169976,0.733828428651081,0.7239970469656687,0.7339180449495379,
 0.1558451420492012,0.3936204679853228,0.34073951208309583,0.7042590130781845,
 0.6549271431999829,0.6249368009388896,0.6223723085413225,0.9781297316912185,
 0.7675380003525231,0.8017322348964978,0.7879523692255024,0.1372806534619011,
 0.4796216466477735,0.6052640473901789,0.6491122620415154,0.09458143046865042,
 0.9447093635990682,0.8185922364142894,0.8833908642910642,0.7365589383181221,
 0.44659100607761126,0.9529007497683666,0.9910021433599601,0.4876744456950397,
 0.2132338521376953,0.3143020395367776,0.5659168881862129,0.9063299069708505,
 0.5951425072470113,0.572023445041601,0.7058244348692507,0.9205557588821172,
 0.4203454337652385,0.02764569082940249,0.6422956728100581,0.7331717202541504,
 0.8933891432740181,0.6200891723142252,0.8496361912785618,0.32945762001215606,
 0.07756534855739483,0.2565236923961929,0.798453260179104,0.7777224146492447,
 0.4944675713214314,0.11157107029337898,0.308888267893742,0.8424993036574269,
 0.6218961533191443,0.03513399075568269,0.21908327266928174,
 0.7714033163830283,0.36368182429706486,0.2529052057711775,0.3965267187105673,
 0.8825970215123188,0.9828185824809968,0.3346353960459647,0.5741900100100227,
 0.04505151322715939,0.1556321449447895,0.1259171588902056,0.3413371547869166,
 0.4965894364641865,0.8816788515178502,0.4516828306589662,0.36892200931142793,
 0.14836559131739335,0.5693390430926152,0.4596150014909228,0.776803847447763,
 0.8558557778005842,0.20789612066216234,0.772325255562373,0.07480175561117897,
 0.9167212135505884,0.6030644776576336,0.4348999425386606,0.3470395010922589,
 0.8937765035573889,0.8182274918892004,0.09301026678485824,0.4055431060576209,
 0.6980303848621175,0.9924988986413734,0.703168480685256,0.23288140604299667,
 0.5501859123043198,0.7359087553187079,0.29569790553130415,
 0.06123025007180072,0.6829868075545835,0.7801180014977662,0.8753429001548588,
 0.20730226336495994,0.03394107152550807,0.24508194484390244,
 0.5838676073699809,0.7479062440009674,0.12343931443106093,0.6051433816884773,
 0.44497364943078044,0.3329987290606701,0.08565901001848442,
 0.34402587862245826,0.7007589792024557,0.534291901314898,0.3709586022885406,
 0.8358959316774073,0.16789398983900805,0.22610732755938534,
 0.8262068432948763,0.5588494505109334,0.544954033477416,0.18258986856420945,
 0.10389546357292778,0.4769650876778403,0.07538779104623217,
 0.9747586840527936,0.7213588400447988,0.7305362582483623,0.5779074419657632,
 0.7911434420546839,0.9050615804700062,0.9244151274760053,0.6273949689105616,
 0.21342169337562977,0.7992515156387017,0.0933607005022995,0.8692309664288832,
 0.17492144787498476,0.6616857508435605,0.7749565124558957,
 0.16249255766343396,0.8914109404387016,0.8973833545820202,0.2449924670658381,
 0.45188508261402194,0.7457771175577846,0.5365540455567657,0.9836237764455172,
 0.5942388570036262,0.22124684241429726,0.8200935377295993,0.3243452824989075,
 0.10563244463762078,0.002551151092683668,0.027071195328883357,
 0.029681629943839938,0.10990851644714583,0.0632739736456952,
 0.1169563681420237,0.9263809977881213,0.10431609372768913,0.9497764333199373,
 0.1675437584035938,0.40708842148312696,0.5035844728607963,0.6647838304553386,
 0.3372401883846363,0.6770189318322014,0.6827487292590627,0.25516244227099194,
 0.18301262331886237,0.7628244871123183,0.6481746396338626,0.8305302941494259,
 0.3620888201296222,0.2700806578795094,0.5071007650490755,0.5533445994007307,
 0.45815578587959493,0.22370756023498584,0.2916584691418216,
 0.8016688863535365,0.5797675492278372,0.595161605890249,0.015611745543423372,
 0.19192143761707003,0.6143855134079845,0.06440026049891945,
 0.25604000417308415,0.4438960072480018,0.129135412662577,0.4719871818761753,
 0.409402879743449,0.5972524736170034,0.42660055847734013,0.6191550033098321,
 0.2506974794052631,0.2186612362527125,0.4861195581907046,0.8080492103588324,
 0.7726252507984787,0.9636822582411346,0.7600170932860786,0.6374836010021719,
 0.4691848268804515,0.3452633398848113,0.14970197263752505,0.9758982449920386,
 0.28044649941362776,0.7626418815746403,0.9130035080145581,0.271252347453368,
 0.5662014154899162,0.9695429208914,0.06207104535807351,0.6221382200362121,
 0.5857497900074753,0.23283103398430538,0.012110382487308557,
 0.8921981268698473,0.8151423459261691,0.4656122121785069,0.5841435735838536,
 0.5700324129858864,0.16747774440981167,0.7851484313726096,
 0.23710397962785634,0.2290502535414981,0.012339087629247736,
 0.2942886526170969,0.5461453905421627,0.32430625180391737,0.5324247767315792,
 0.7660529529903924,0.05638282592564958,0.9055114223542067,0.5173171810170315,
 0.43795610861675394,0.36701030845177396,0.6480692127391564,
 0.7132093457299387,0.17503585955757683,0.9671884791365638,
 0.01786332134633062,0.24636768007179444,0.0030476792958509247,
 0.21861338683404496,0.229525032023834,0.40454409199896124,0.5113887094251153,
 0.4304790950709555,0.745485258578517,0.1966683916255525,0.11874334154272947,
 0.38829868266301193,0.6577306822157555,0.23789224180726853,
 0.3672639992829426,0.5296537504387296,0.5996179637271797,0.12249364471864477,
 0.6605138588977133,0.04317209044446002,0.18666314444449383,
 0.027462918205978903,0.384278143669576,0.11893572555533294,
 0.6839672014754865,0.06334340074766864,0.19925986265861262,
 0.6770012222060406,0.8506764324361176,0.2886255168304641,0.6972720646458344,
 0.19158267981533372,0.6618449123884229,0.6801379871837033,0.6928212837685304,
 0.7446437694448809,0.019444933768645978,0.07309632015652201,
 0.5629679592444478,0.0723820295594918,0.5706723427287269,0.3158440764730367,
 0.19545669217780393,0.13103313145283169,0.6862840870782596,
 0.28921816118506183,0.9803102437044784,0.9103798049686538,
 0.030236430992243712,0.8055868239208281,0.3661941908734826,
 0.1575577445817582,0.01532945623034676,0.5354994481276316,0.6160898880653318,
 0.11571272765631513,0.4455583518261823,0.9938057293235496,0.8214886115836307,
 0.8700252364682874,0.3573322411124541,0.7651097430145832,0.4596135957399673,
 0.1294373574624097,0.11651686663429128,0.002036912296322857,
 0.2071473722232281,0.6575601801035657,0.2762037624188779,0.49342296882331704,
 0.060361082588937465,0.03404660841247398,0.25967735237052536,
 0.0917470384524508,0.48871654901464223,0.6846069636432923,0.811776073308309,
 0.29560556899925505,0.5179887700797503,0.5435312438821117,0.6118875390016312,
 0.8989283631217074,0.1306025831048695,0.55436661404111,0.011632748399195036,
 0.9688927033604995,0.797852885008843,0.19921808592753953,0.6027368433196625,
 0.6141663762488839,0.02388868157464452,0.3254785209208426,0.3442763380447863,
 0.7292511424254218,0.8765018513383783,0.6238902541579765,0.7792892771798661,
 0.6601237388031795,0.8450666975680803,0.06197108530200934,
 0.019539337613866392,0.6727121129255235,0.42237515390648217,
 0.2434472578196717,0.2619345681035753,0.9964643714771381,0.5291821116670873,
 0.1784696881504091,0.37491063939395985,0.6643180033101996,0.5838214461615147,
 0.9684148926026047,0.48530961449716115,0.9011036024062568,0.3777888535062536,
 0.2406225001104848,0.24267550241271352,0.19530821566139545,0.608880706713473,
 0.7327005878113044,0.429643603029477,0.7805839357843002,0.21767131195752298,
 0.32836701189878226,0.7283400029234844,0.029264779648921824,
 0.2045349297187672,0.49539093154805225,0.9396971761620501,0.3058172330932645,
 0.9481666350793985,0.11438525487615336,0.32090278270498906,
 0.2726231358259871,0.37286688979899174,0.6818503752903177,0.1928370671262788,
 0.9737700640139336,0.422892269298615,0.14313980706689589,0.5142197700935528,
 0.3280452933190606,0.12816733944055825,0.11111449968701059,
 0.42720433232663424,0.4318043259533336,0.9495459815685826,0.315443609054182,
 0.3928671663957929,0.882679492751193,0.6732615246497511,0.0265922100698901,
 0.8092723092567434,0.5172812969914812,0.8191479726437918,0.37918583993969,
 0.19640164151798978,0.7596418899896081,0.5497154558099169,0.7879972377999153,
 0.41724777374812616,0.7833529788991958,0.5943915890540076,
 0.004130058824353355,0.1592711613397002,0.681972421711525,0.752985547139582,
 0.21284953480683688,0.1145864626344324,0.7384960994971425,
 0.35550783800758234,0.35988760851467305,0.9729324210657123,
 0.6339143043291996,0.1036495372812255,0.2442700795246232,0.9297038536024306,
 0.6408129796992069,0.2196732481679986,0.4767002805504613,0.3218367977742391,
 0.41275277874415495,0.5173459921884287,0.9552668150481908,0.6637485475490668,
 0.1191422049265447,0.39931638947958925,0.6538945758107183,
 0.40156881657791255,0.6195175608633433,0.7452970186442138,
 0.40666041383653273,0.27910540499210357,0.9511275646055241,
 0.0922024819075763,0.9274729786223864,0.5993399368584599,0.5485812879343357,
 0.3676973064261747,0.04689326207323308,0.5303341807650798,0.9289524179409998,
 0.05825669061042471,0.8289859764065377,0.09895165523715843,
 0.16376582851484478,0.32742060190942623,0.72820637384244,0.4306634776040764,
 0.7892414113524944,0.3850909107616969,0.30102900960920465,0.3327771338564748,
 0.9212345946901014,0.6572447517920725,0.4642961326194861,0.3982412949186187,
 0.1334738887018082,0.5634911114400792,0.5746960229270753,0.6181753392615181,
 0.6264797916389343,0.7829625007454641,0.7333241182440904,0.8986327247523838,
 0.9770292340336338,0.23725855592960454,0.2539896722902196,0.5353454530577577,
 0.2406505896941208,0.355084935254527,0.24200840775854993,0.07230409578874442,
 0.10607717283511808,0.11206710638036566,0.9748636356655492,
 0.5069716254534693,0.3935836002889014,0.33480906478738204,0.8988787150871707,
 0.04579349155351886,0.9289776162586622,0.44849741356989115,
 0.3422860477987826,0.2791745734716894,0.03576618521241559,
 0.38408682778128866,0.2137870178585627,0.4896632903379802,
 0.04499556665407567,0.14592352229420902,0.03186398397279333,
 0.6971097943797484,0.6042402175810311,0.8620730127313739,0.2613635243894642,
 0.1840529729168332,0.005287712315672599,0.16884142827280824,
 0.44093825372180795,0.20966405959548595,0.5080961812141008,
 0.8801277828401828,0.4965065297254162,0.7280530933358329,0.3600413743114852,
 0.1101372041259353,0.17986488565707348,0.13624643574524709,
 0.49942560361363486,0.3986636212315052,0.8305811283930244,0.8368115434784429,
 0.1427247839138689,0.7112012911383725,0.3112887960191062,0.3742934261232135,
 0.8421709571268982,0.6683626241096087,0.2801283207066898,0.7017251223014461,
 0.6083675950381211,0.5455593192074041,0.09355008635722506,0.2271887713911691,
 0.3603207434914387,0.6966014353925514,0.3148319312292892,0.4205627530382684,
 0.7745634177629184,0.8471570296820988,0.9450838298668053,0.22602683522199152,
 0.00958845016554577,0.5238523522232357,0.605095034532164,0.2994007100077004,
 0.09475227734952874,0.31174507513546856,0.5401159676198062,
 0.41194379027705996,0.753857208654039,0.9575104070200955,0.20208525448758707,
 0.6315823522213762,0.11584468149406069,0.8908553344506812,
 0.06540773382849019,0.3764349879646959,0.8573512464090975,0.6143440628502824,
 0.88855606240962,0.030609398479777505,0.0037898509512067635,
 0.12504346792001186,0.20266839686064753,0.9823941885702964,
 0.3221984362570809,0.19914009328371451,0.19994997970142536,
 0.7526670395488987,0.6273050542144458,0.0018195236302629425,
 0.13140609254671487,0.3997350834002721,0.3182988977175242,
 0.08242606977668376,0.43599502337470153,0.6536050578137095,
 0.6268138238263739,0.7813728306739405,0.5201822727798009,0.47545930544140513,
 0.22628264400501819,0.17811191232423174,0.5559843398739626,
 0.6582873565781054,0.7566464665523579,0.82082063801168,0.6210649133688767,
 0.9258636980882572,0.882933511465269,0.14276113557381132,0.19991503965229607,
 0.4150835637673216,0.6940627859131223,0.9700080270818197,0.9028181351078882,
 0.8561692178645466,0.5461276285734668,0.6720480959051218,0.8115023225086656,
 0.5277243839153244,0.6477327888491418,0.6818665509965156,0.6481252182004759,
 0.7870352903139823,0.44782832448102305,0.23338251248623132,
 0.03058279431292721,0.7345888008306203,0.21813578994080085,
 0.19485200022193605,0.37405619761489084,0.8602920248033707,
 0.6440111071766288,0.642839205023109,0.8676548079544882,0.13378413455121319,
 0.9557401741532002,0.6914744988729291,0.7863138578514328,
 0.025231444083862575,0.3253782550322514,0.5568498345200643,
 0.4070450469100272,0.14646582614915227,0.84543685827419,0.9354524369943171,
 0.49522935640259624,0.11801058890943006,0.9432851494120966,
 0.6743642086439604,0.8348346501008644,0.8404536676525015,0.02479887676879422,
 0.7947541001274009,0.18189533018140547,0.13768475994245555,
 0.3595145622389305,0.5107574707099694,0.447150067756954,0.27112889171576293,
 0.21202661868675476,0.17763176566656413,0.49281503368920965,
 0.7553944189286086,0.9722267096480485,0.284081301616933,0.3426444766413764,
 0.22268356662633215,0.9615855150240473,0.90072509243884,0.7469361056140353,
 0.5012207918117166,0.6479603829189458,0.3016235971729755,0.11660092603035377,
 0.658078042115837,0.4926427535371428,0.23352972338310685,0.12578906624174524,
 0.2890530681499699,0.9627629488670999,0.37280050097059425,0.3906846694375514,
 0.4422435875693461,0.9840935875375836,0.6088817274287233,0.9074959495306506,
 0.03930985231951545,0.20820957579204824,0.49566394758465826,
 0.18382186109446041,0.9480911171794781,0.5753207052017937,0.7576571098200952,
 0.21103452932508016,0.14675895769481018,0.9117198036125984,
 0.4569872765888293,0.2595945064585079,0.21064324214040553,0.8593442477658625,
 0.7353777826583895,0.20312707658868134,0.8825162079298208,0.7463943825506405,
 0.018715700618512465,0.3125458039963289,0.7097146302650489,
 0.6974354148577099,0.9416915855418114,0.7266857713879175,0.8653664389502838,
 0.701881936499495,0.24525522003187594,0.3823437299743817,0.40106696309847745,
 0.8241978612234125,0.519899880087435,0.6436657709914534,0.4308934068968424,
 0.9943965875377702,0.9939523331249421,0.6424196944255229,0.36290740568244484,
 0.938682999528679,0.08283596688919936,0.4248980010823439,0.810536857888516,
 0.025496869991372906,0.5399394443522447,0.7476903081873748,
 0.26505201446570315,0.528198545682254,0.7241249582535787,0.9092289418241597,
 0.3398185048780833,0.6476478124977307,0.9632217805728632,0.5626069434178684,
 0.8546205575206607,0.7011926482414285,0.046775081679481856,
 0.7056906454144062,0.1929895404320856,0.6456291093372506,0.9886819773673291,
 0.11519027920440517,0.7882995364818535,0.8958922011834038,0.3860463469809198,
 0.7053492098498266,0.34383374501171904,0.8784032694034157,0.8039182870337878,
 0.5580509147963529,0.7016285440338698,0.30243377289326745,0.5808520611856668,
 0.6704255357980213,0.20106937951788517,0.6831366140559975,0.7372622562405279,
 0.0408467333191983,0.9598850072105766,0.04713251000319274,0.7153854578252011,
 0.049557996188443476,0.1432532555071776,0.6260137899296776,0.522903252852118,
 0.1794827032288805,0.7517077374879345,0.6022722675861325,0.2746864344465213,
 0.06173482801286756,0.7757807413443051,0.8694000490631658,0.753720958938334,
 0.41239059507845366,0.4307518205105736,0.6995470535144189,
 0.019729876814101432,0.5950519242535826,0.4426613768919413,
 0.7742728448675715,0.24303947197015274,0.05142259805067262,
 0.7129458551475065,0.2841063137857762,0.6311995150769562,0.9250134182583889,
 0.9101016776039313,0.5921235921252315,0.5645486146965214,0.5565291182161435,
 0.09578657253279521,0.7165553996135949,0.5610761767472089,0.6039748780539482,
 0.807729324934414,0.6483613950654343,0.06173065739149308,0.8228957516046723,
 0.5000304333617671,0.14729845167026534,0.3378125473875615,
 0.40310047946653105,0.2961821656103397,0.2659545081439919,
 0.31527581262001414,0.5154061240962825,0.13090976655298436,
 0.3078797366653183,0.2608042997037604,0.5588298286045446,0.8749354940860943,
 0.17012756466759438,0.6795263047890661,0.23315862139246668,
 0.5892471778055932,0.5364079870560354,0.5276031505413153,0.6929564576640339,
 0.05258469498866003,0.9449342497133182,0.6274061126107617,0.6525293150050671,
 0.6223968327899655,0.9970221680331042,0.3890141122127042,0.38629230125376124,
 0.6109597368297223,0.6250474891866707,0.5065944759016875,0.7207655295001091,
 0.8321189819137043,0.11908020865197588,0.6171418425188547,
 0.17535247855861846,0.31573696726965494,0.19726346316400845,
 0.6067751709517344,0.5609860960445934,0.7471495140052122,0.74353547760531,
 0.8126745057163626,0.43814673841178886,0.6071383306088102,0.6840035080564184,
 0.833599706185487,0.14040149659146872,0.22929564149494097,0.5544003629181876,
 0.9215406620210407,0.8880663646807271,0.6484344656167715,0.6190005123991192,
 0.0697592837470723,0.5060814203806324,0.5198027959458624,0.4200631724916035,
 0.4313959806614662,0.7290990996088671,0.18456238701547467,
 0.16508295848034704,0.27465233384153287,0.41829992174762787,
 0.7115044796335417,0.6034143624433166,0.9458647000967751,0.2859808790413607,
 0.5942533687181603,0.7396472191792337,0.5422814344945349,0.6539867823187719,
 0.7066011378413374,0.8445249171581625,0.21460513860839137,0.1448047189955024,
 0.38299184347331405,0.580487938972707,0.5503499190862589,0.9749887604339519,
 0.6138771236401074,0.3218310882496592,0.5963153246871309,0.7788796770120064,
 0.23263913460130103,0.36616192481974263,0.4296231770984634,
 0.2555729265712583,0.6546051354503779,0.23749948672052457,
 0.08391089954332809,0.38979867361541776,0.3808288449380599,
 0.23919590440179916,0.8219812829897041,0.8151824234892147,0.6107275422348475,
 0.7823641224886078,0.1474636775785192,0.04161138792385943,
 0.45520052140113065,0.913572878815734,0.9615035620057273,0.6039719802440621,
 0.2334687108777398,0.10205564592294813,0.8458905665418939,0.7157038158793136,
 0.5024473773017945,0.3931536563821384,0.4734683710698301,0.9720420642767534,
 0.2276953054474885,0.5618398837935257,0.33110535331208457,0.8740779801266565,
 0.17089894083177104,0.279779031547419,0.4344235617527531,0.7967136907098863,
 0.5514337966653393,0.09560070629261896,0.6958830258085129,0.7309949580553372,
 0.26800149758782776,0.5740131047210605,0.24088313438406472,
 0.8704713659311853,0.10174473135589734,0.5756527574382149,0.8192925575021287,
 0.2593040171810499,0.854684187948253,0.8238134194220036,0.5865067633718212,
 0.9500078181914839,0.18658390559417004,0.41759369914420286,
 0.09300389350850147,0.043344374740898584,0.6769825528784769,
 0.043143920984297646,0.6136817513609161,0.1566668781006746,
 0.3225022787333329,0.25684024587691,0.20144122645501117,0.8706046758454113,
 0.3833427912860786,0.5838673032068857,0.3771940914320213,0.14041840562832597,
 0.06468659979251723,0.9877420161780249,0.26060863861875105,
 0.4219680623824531,0.7300223734595925,0.7672369586960448,0.49860299888923665,
 0.48747522680875266,0.8782108710810793,0.3404065878035879,0.6780056955991356,
 0.7645768899949138,0.05919980613792142,0.510639591444021,0.6362952610273281,
 0.14682905723485906,0.0494396810644635,0.33426552399560694,
 0.40395132454538374,0.6018499429148308,0.5747771111037056,
 0.31626403058795016,0.9450097292182578,0.41049848847265447,
 0.15266192903942832,0.006838534587209177,0.17887756719962689,
 0.6822840993469634,0.909561572693084,0.9886085310301423,0.7917710204743639,
 0.1429859925800222,0.8682665765664398,0.6512188932094674,0.8418884493880795,
 0.9169743208449392,0.3443590625362587,0.3765079702104458,0.1893409823252069,
 0.9396034101439286,0.029386978007243436,0.724650772536029,0.5953390528672211,
 0.6706867541137012,0.4668383994233515,0.1907106677549515,0.14677168505782334,
 0.11285032563274444,0.532252357382182,0.014253421999087301,0.929160886219519,
 0.9029260029964624,0.11526481398603805,0.2161665564522226,0.9529918672217086,
 0.5754567515613208,0.05888501783947819,0.28210014185299837,
 0.13905264009860596,0.9822660625108519,0.14660385275672472,
 0.7272695512021994,0.7504841726777605,0.24560044189316366,0.6394342581610382,
 0.7693607405521927,0.8091484964536292,0.5356941494761587,0.9173275230353011,
 0.5817950330677862,0.4401280050489582,0.8089550815952526,
 0.002120853388409061,0.13960998891687282,0.7402074459941335,
 0.7282756740775733,0.7417475721032979,0.3119208412598874,0.49946988252464875,
 0.9675860570654423,0.5555086511474029,0.5954525241607735,0.36396772733529514,
 0.5265046602599612,0.05741144892648686,0.3865452555903778,
 0.24830570799684082,0.600362576983665,0.22909837950870238,
 0.33387914447022293,0.6377622280766986,0.9938375799080676,0.7941634137138137,
 0.9401635027526077,0.32140570775398347,0.8693528170255561,
 0.25385971835038423,0.11650753247513135,0.6021530397817094,
 0.00813071814931221,0.5699856506352853,0.47436130725878034,
 0.5544317472950533,0.7088700126087393,0.9422329613499466,0.1576447609345828,
 0.9559621808135746,0.6424278786314953,0.8134323541740598,0.04130971321717791,
 0.22055945623689532,0.6857454155352173,0.45241327934962716,0.79890441927897,
 0.9810801218819272,0.22747598945452419,0.15923781104283974,
 0.8817264586649441,0.8093328772224095,0.6045208332370433,0.722663584499666,
 0.09420015932333414,0.8865796923557026,0.5909057142414362,0.2566939166614862,
 0.4369848291252464,0.3110027169559091,0.4302422664096647,0.3836753961431879,
 0.4810312632587934,0.28105904696006556,0.5096011645407167,
 0.061705289589503276,0.3906469131679725,0.951915631907593,0.630742913902833,
 0.04157263698435543,0.2552420402021084,0.21853941436932844,
 0.3848432913467258,0.8586083038043166,0.8117959654008335,0.0974978784692393,
 0.3875488724398246,0.5364529873838703,0.2782908155064058,0.4091320101853372,
 0.8274702960694973,0.28368052194857074,0.9143433103063399,0.1655290775937479,
 0.24422433009997369,0.1741508460329445,0.8632810688158361,
 0.04289362135657315,0.1763017076935115,0.7938825249971204,0.5189195114979414,
 0.7837241389158123,0.37827919761127893,0.7262879007451255,
 0.24082510951786007,0.6677231192098034,0.35540697444713154,0.644283532053043,
 0.46291821091814334,0.890130202912243,0.9944081317882005,0.3789481257701407,
 0.37198376714630865,0.6854653320423146,0.42013351449096503,
 0.3514339760446876,0.6334408568593421,0.6584280485643506,0.4011205694347495,
 0.48794764382018607,0.36629886439229387,0.05982369177241187,
 0.42635047306798746,0.788550270856629,0.5614202096000431,0.2640939753370616,
 0.18719651483114175,0.8855735684200036,0.278064272327585,0.08587212520560394,
 0.17258577167433953,0.6163347723694547,0.5659109007595493,0.5334311042274755,
 0.45429431608972426,0.3921567334629199,0.6305224959743657,0.157014826970426,
 0.45737431906281467,0.1015343745312105,0.6854215714929824,0.8069345277303677,
 0.8526356175815611,0.056878528461356104,0.9463820682242658,
 0.13823661316302294,0.44678411998674317,0.351211652343704,0.6627083093149075,
 0.467597835578652,0.8664370407608323,0.40958658056267283,0.9494195362174288,
 0.09193561009472728,0.7386111762260975,0.4217138231072779,
 0.13159463279360561,0.14918708141385895,0.8277424136478028,
 0.4108511076696806,0.21959922047192448,0.6227807798614575,0.9977310549529075,
 0.7553205030965615,0.6294161885947751,0.10760155856374587,0.2557413116983379,
 0.5584518493939636,0.5044374471866948,0.31755261676927293,0.831678207613474,
 0.6635249650945804,0.9541246127944065,0.18796549313946276,0.6972315016780042,
 0.9777144394636245,0.37013877392178507,0.1603045216677983,
 0.46141795831657584,0.32100849168004053,0.10018654501050017,
 0.08298607277020231,0.1348753509186439,0.7137655340931589,
 0.11462601667380046,0.9497370239843284,0.5106684300614491,0.4010614128042931,
 0.6888777371401625,0.5911129404822741,0.2792184876319217,0.7143217804882013,
 0.9553496252160409,0.03743259230434748,0.8968699989453233,0.4600971129581193,
 0.38435215545672585,0.9959239176939516,0.7234020538083624,
 0.03943339271847823,0.1850730839452427,0.05981415951663849,
 0.6438305997381999,0.4740107247512012,0.36657673568709426,0.9087522983594872,
 0.04692564460852111,0.8038173372321405,0.9769108383947643,
 0.10174537422935748,0.46212009580536895,0.5790388811551431,
 0.3595554252847145,0.9740931096378547,0.04213463453361321,
 0.38568765855869425,0.20019786494336644,0.376897853280745,
 0.10138144638616975,0.9206979872256343,0.801082358022469,0.4351596600123925,
 0.47709709141462087,0.3671452940509373,0.45783060406216936,
 0.020550456233449532,0.7032800991489054,0.8397682771138941,
 0.3989616275764265,0.1783436945036273,0.22997691103333828,
 0.15744259699648266,0.26986987974819177,0.4957030318743654,
 0.9648374792035951,0.4372284604472685,0.9848238463074019,0.46864216490585764,
 0.9588634254171202,0.06567263080431318,0.8531984578855054,0.5341056402904004,
 0.7983429777640884,0.3300239683068662,0.5604197121983197,0.4343545205119478,
 0.6471314754467892,0.41041835497180557,0.699664583872664,0.5727303130041204,
 0.47163093279697665,0.6526048898933463,0.020045443427122667,
 0.3614950241519129,0.4464279095065643,0.9175398205947547,0.648105276835467,
 0.6091951792952932,0.11373289343718407,0.009741496628614232,
 0.2543442398683956,0.06167587132356145,0.18212725412007202,
 0.13722382997822968,0.7757609970444276,0.5223776833616363,0.5185429502963714,
 0.8072684425735986,0.966670472244415,0.35961764221240955,0.21082855997583805,
 0.03007508058608921,0.23494057370832744,0.5757306677511065,
 0.44640038338269816,0.9992948434955775,0.3344065184516145,
 0.25001788638799327,0.10171336342270498,0.2870688534338697,
 0.7276761013248476,0.9822320501363633,0.9563367905568887,0.29605175200775913,
 0.17985544756705862,0.2674366680630238,0.7029280924115398,0.506065678179483,
 0.7989205753098514,0.20077312201782038,0.21739543621829382,
 0.9933167356918614,0.6011781924069093,0.08335455282559878,0.2636610894425908,
 0.8760502334868159,0.5055998247759418,0.7229451251899144,0.05988954964458604,
 0.12170784058830442,0.7038024815402152,0.8255863990178929,0.3716019890896135,
 0.8015734974279751,0.5419875665106593,0.15574752780881462,0.8448018224608397,
 0.9927323823997707,0.9639988019454218,0.9465385168760801,0.9756305662129197,
 0.395804971258759,0.454038807935919,0.35943003457915923,0.3871662236996014,
 0.1462069342418415,0.3012047993806838,0.7434460374686205,0.837403201227421,
 0.6836525193020155,0.10990999463676276,0.5765720140905275,0.568459710933377,
 0.6739671210368138,0.5309312399929659,0.1330348062360106,0.38303966052434224,
 0.08620970407823114,0.0818663648268163,0.9692522226946397,0.6305074295669273,
 0.6130255355811434,0.0767583152862703,0.9228146770626261,0.6593271342353945,
 0.1025109233256416,0.5726997786921539,0.41053114883574504,
 0.07408633524802055,0.6399383014072227,0.40860426657524096,
 0.05738703111549348,0.04821875788633201,0.7828547073539436,
 0.8339742838737741,0.28650995667674084,0.5500069778423147,
 0.29947153619225053,0.7900049777673157,0.5331634551357801,0.2290220454876034,
 0.4310019374482692,0.7703624067820574,0.2727103234177177,0.39013923343301027,
 0.5038057234747136,0.7468103016668747,0.28578280022439584,0.7780524494630111,
 0.5509550016520282,0.13227815555997224,0.22078343085326568,
 0.5661471445504997,0.8096557992110913,0.25088446585762103,
 0.45603870512475975,0.8169432344668315,0.21388065227320163,
 0.7223053905481909,0.3456712805453661,0.1662400914906874,0.12421773345588416,
 0.8164629935860727,0.5802994650459963,0.12816263090609858,0.2255989767084705,
 0.9040993500228974,0.2433559157886649,0.1399112135970908,0.9676279518809199,
 0.47476384815948836,0.6409134590788246,0.3822492175342872,0.3641670282117425,
 0.3383447002161868,0.8111446379213673,0.46242575036450284,
 0.013457803727414785,0.007485014853811478,0.7380013428386052,
 0.1786335004172207,0.8652880950640676,0.8146420749538781,
 0.014378988208641474,0.2990688680083722,0.611080366845587,0.3782249654200991,
 0.783517650834754,0.18037408662598486,0.36663266474832124,0.9890443721256068,
 0.15844019307560542,0.05068496633600472,0.6101097243424131,
 0.9955937064775593,0.9662160624819957,0.0075665755050940975,
 0.057951855611199865,0.4943559836250705,0.9472339334178788,
 0.9931185115492089,0.22886010401962065,0.7369423625903906,0.9356053564373682,
 0.9357810733822027,0.61583253748892,0.42646194917176294,0.6199476801765524,
 0.8821470804900711,0.035595507745216914,0.5246651985879902,
 0.8841951392084183,0.07459184603152691,0.8234837195500385,0.4916004412171717,
 0.15377423089049047,0.8317039506498713,0.7408713242920997,0.5982454201169954,
 0.485073958761992,0.561358377070707,0.4093951382651172,0.5932721801776356,
 0.30798954487524954,0.032902261156941615,0.6770796676087382,
 0.2509536963424708,0.7773295599136513,0.39089140961869595,0.3458882018127528,
 0.28439836422405074,0.8095160904086649,0.07704510172402246,
 0.13953119751380105,0.5694473920202094,0.4531397124775367,0.624278244324462,
 0.8314045528319671,0.932345680537888,0.69854467573479,0.19937812732131155,
 0.6714274616946226,0.2766209582902207,0.5268407753349249,0.6642137008035609,
 0.4655183703830832,0.32438219921293354,0.11778410226700198,
 0.4826145414055809,0.6450758485922309,0.8785979056883833,0.37870681426644115,
 0.9665506109903559,0.19453585121789618,0.5821266626163618,0.4442120383222562,
 0.09219080774918487,0.4716921641077837,0.9418016818610122,0.5161944152448225,
 0.202056123346104,0.05442296035532457,0.3578628409198972,0.7775394399252863,
 0.7609831569553681,0.8063060459321956,0.1938900362734297,0.19012019080732367,
 0.0631997141371573,0.0807084544978679,0.36748104857281183,0.7940284478393376,
 0.7409637780698525,0.18002461466714603,0.08547833084561796,
 0.48827106114095176,0.4694942439707652,0.10620162525730414,
 0.7169290728656794,0.37900259092903155,0.7397140768834335,0.6765833643027407,
 0.33696543001647516,0.5480031842627308,0.7786286438931724,0.4884410055825875,
 0.07557847041562338,0.5411402427197618,0.040074628394625056,
 0.42261929244279295,0.06310666290738642,0.9408042126018956,
 0.012131756163247465,0.7252823781523945,0.8076733223958097,
 0.022442774788942077,0.3960811781293442,0.7987953826106676,
 0.6938662007515286,0.9788247476723432,0.5903959028816932,0.8945594010459853,
 0.28207737619747675,0.2887349575011423,0.0019798843551170897,
 0.35631223209455753,0.9116590605864567,0.4131938594240616,0.2898013515507849,
 0.6292478802675534,0.2459298457827065,0.47422857243462424,0.1791657598471451,
 0.5712864606789694,0.08365011607031536,0.2840352356103939,0.4070589472702826,
 0.33222006995254305,0.0964222737416679,0.5279971212929516,0.7249394133474679,
 0.009698785298424495,0.8347430005835084,0.010178215385026057,
 0.2911871417585129,0.8811498490347539,0.40123013720618195,0.8659131021404686,
 0.7945556766920373,0.3719074170184439,0.524482603421185,0.5215878185141081,
 0.5211615436992557,0.5242829479662018,0.8228163103097267,0.7319130131891208,
 0.751099845123366,0.8309743799515671,0.3855925824306967,0.5220056752260642,
 0.4687557107893686,0.20195152219485324,0.931454233820141,0.635370046059664,
 0.5322634486797669,0.8609299710041473,0.7914703298960792,0.7500893438459285,
 0.12029802951961921,0.29386405405842897,0.461464467124977,
 0.16958281722497726,0.3423696222252347,0.2503914717051785,0.8453687765930934,
 0.3913387779991516,0.07557663349421029,0.2696677187627279,0.4975372326016876,
 0.2740469763576209,0.6905445500868468,0.9470851669436734,0.0164842648811091,
 0.8189125471544573,0.047945253744851035,0.43013733209782967,
 0.5978665696718034,0.11923959036140397,0.9125468323025215,
 0.45072690036789265,0.7938331074383701,0.8061248034889344,
 0.10018887705357671,0.6581762462695283,0.6760512937582127,
 0.17572380581164604,0.732589958498835,0.9099033314213126,0.60334747318929,
 0.5159644029625546,0.9971955524390892,0.2778201368954355,0.48217073062896676,
 0.32463865400524095,0.35253985609426763,0.8423871025754529,
 0.13471150187440606,0.6829652029033979,0.20529367913766472,
 0.2663301100143488,0.7498084347950208,0.9134706762985506,
 0.039168483961288025,0.5312131304215544,0.6508609759058599,
 2.883489589513566e-4,0.6547293789022947,0.0484944903400909,
 0.13914349003183102,0.2653522002992271,0.5111243241359099,
 0.36282747366266277,0.6036525859540168,0.9006058559391084,0.9683343856257649,
 0.20758615970119265,0.7887931415351097,0.2069599449767856,0.8227832849058512,
 0.28528292613391026,0.2715159862708827,0.19557412215783765,
 0.42830270995488817,0.6547611503019546,0.6884649926640283,0.0965179521414582,
 0.9306603338240932,0.9352053351227523,0.4793774962794407,0.563082700859017,
 0.8312696759170632,0.21026557859019612,0.7479940290242071,0.3796229077958564,
 0.17602366025666127,0.8124201683804201,0.307945242896534,0.23874391283202223,
 0.11453767830575612,0.18501755656524466,0.14783439233584328,
 0.8798217452877062,0.12293387615271589,0.017762153107062417,
 0.4141106350326047,0.4972858087664722,0.5497543176060499,0.36953775473004513,
 0.3654546087467838,0.0868984552975468,0.5099920009738667,0.2909790686371766,
 0.7961947617649725,0.2810962961811324,0.7797239009222334,0.8266744689398735,
 0.1365046238099238,0.08848770110856863,0.07569820216212675,
 0.8627921719196495,0.3268712290128144,0.6677344106039665,0.36419632082556586,
 0.8242047417586255,0.8019656984642577,0.575470786299132,0.06633633049564269,
 0.8893685996609587,0.5752916019245727,0.24482375412074087,0.6083719500690039,
 0.6561330638255095,0.7354461306708893,0.21940220482729234,0.7783251319875814,
 0.6875072039468967,0.3161347416958775,0.46422891088535323,0.5267757456510893,
 0.8278968771345991,0.6719848955219008,0.8706169421829054,0.04310406045046977,
 0.6453116853556493,0.12670061923928322,0.37497599444691043,
 0.8824389431824906,0.9317121663405917,0.9504616841914622,0.8040876721915113,
 0.7690057781144218,0.3907447340062844,0.734279021285186,0.07848469150299442,
 0.5023809590488507,0.38587024614207455,0.12219047425286789,
 0.1537338498137476,0.17134229980945914,0.6876734395362893,0.4761546589934911,
 0.9022166022699681,0.8425407118775524,0.6952404710297494,0.563009914039447,
 0.8431549234205492,0.48837533912705844,0.3168632513068681,
 0.08278224822301206,0.8385495469665608,0.25415380037003976,
 0.3561609707930288,0.976166271540875,0.06563423559744508,0.9424292885615516,
 0.4606359987950964,0.8895400207809412,0.6377692284262895,0.3381444728042926,
 0.21539989890233757,0.033926405822803596,0.2441507245793899,
 0.03796692366211751,0.3338086784968588,0.5358716541955727,0.549237679293376,
 0.7850879633966577,0.5227406679243535,0.04556359657229747,0.7106312407184525,
 0.44733316279226065,0.6347275521098115,0.5992905862254624,
 0.22452150058142673,0.8398131737011604,0.19260020014021229,
 0.5515167558780153,0.11711824023934092,0.8317860950325745,0.4279977236389385,
 0.4391411144748487,0.002246067868379664,0.7992110784907858,
 0.7855897279861053,0.7213914921625713,0.426928829692237,0.11176709826615616,
 0.5944878164075209,0.3104078009235812,0.7785453278534837,0.8776680537510249,
 0.589797653186647,0.6647033823395097,0.8549048886200079,0.9359225292851235,
 0.6358919925393912,0.0044159070707744386,0.5039762601153924,
 0.9750781544628584,0.8345774900084441,0.7010596240457303,0.31238565422670295,
 0.9948104730445201,0.38084712188021375,0.8437491181985276,0.260459117794695,
 0.10591226977949053,0.1571775959370959,0.9968015763493246,0.2328219054875622,
 0.6768689561713616,0.4449639606703184,0.35409367358084076,0.7313141526992656,
 0.4280041328645082,0.6987799066250276,0.16957243438213476,0.2613028746232916,
 0.5799733561687751,0.9492500657891907,0.22047528070054012,
 0.16746500219542726,0.6780060151058144,0.9797653598166669,0.946426773975151,
 0.455221832446861,0.702825791347352,0.3497498301597589,0.829430170580314,
 0.08753434379635683,0.940954250925114,0.4749178037969707,0.5273431400396192,
 0.3587576041070255,0.7248468707009655,0.8462907798844653,0.643552234901096,
 0.5705950135212327,0.7882604764082929,0.1341829188309176,0.1836294452626397,
 0.35833374341317514,0.06328535388345535,0.14453296966337514,
 0.9874495707581713,0.27591139162670775,0.9503014557042144,0.2220337615011475,
 0.9400875257965919,0.8097935247945253,0.4510457967362903,0.2722968385516934,
 0.5025963554003237,0.23199711617931917,0.40010173965678164,
 0.13934411628013443,0.75237237271863,0.36045519306116425,0.6188153742407697,
 0.35749813275055,0.8323183564002306,0.2254220700387891,0.813827711964089,
 0.5256394612202231,0.0570084438759334,0.19553600454207543,0.9791480095464523,
 0.1869883005852907,0.6043599286627612,0.11640547110855903,0.9813577150249164,
 0.2555278776175749,0.4475613886595138,0.652409787298619,0.7121048608226407,
 0.05919043929225232,0.35733496044021773,0.5121798243545911,
 0.3255858306324586,0.3237528326568301,0.8384770174250948,0.7965937403337138,
 0.10983722081090774,0.44804756687482983,0.8689917534665547,0.598738687221048,
 0.5691816820964175,0.94906797358329,0.29164195198683496,0.7704325593673416,
 0.7528828980305687,0.7881462071777334,0.4076659294117294,0.7534153302816016,
 0.8715486307289927,0.33705994118079485,0.012360700749298625,
 0.8481893384394763,0.12825912574092913,0.6993377580267882,
 0.40703992341033635,0.4821640161600028,0.5038775435541029,0.847350218600893,
 0.9441516718497621,0.23633858600998547,0.9786260774546763,0.3370757419795011,
 0.9425665818635691,0.22520003395024912,0.3055867555455989,0.7910525019654988,
 0.09616345356531453,0.39567529606727736,0.546719906507429,
 0.14065047378582884,0.3755178043138535,0.3359033740175611,0.7913130431733789,
 0.22829905899591063,0.4419217737305403,0.4625218253466219,0.8935212719312258,
 0.2595592586261587,0.5908886381703359,0.3420960346796119,0.06463747911275197,
 0.06249136494093077,0.9066357715107367,0.03403085778340609,
 0.20077596405277987,0.13611728432724124,0.9168910107722006,
 0.16243911499224772,0.8910148043289209,0.862779987961171,0.9957396418981105,
 0.6869681380875657,0.4025449217656054,0.9569555467290585,0.13714521023435766,
 0.0988093214652922,0.15069313779561444,0.5791599520921448,
 0.23226523900665264,0.6932780075038419,0.3632859400968095,0.6981900823399284,
 0.789063326209891,0.7296329256023857,0.09827601831109889,0.5691167006440557,
 0.0715682462928332,0.020080125519892245,0.6470690347202058,
 0.6292756334940471,0.33076853315129173,0.22433402894610643,
 0.4596154448458306,0.7582038170962071,0.6863780567646794,0.11758710292563335,
 0.633056558067912,0.1673798112638787,0.03720089625738365,0.34265714514323187,
 0.9435906709934578,0.5807446096837539,0.9470389980300773,0.7847309655539567,
 0.07134066461564825,0.13674317664865399,0.8085902904548208,
 0.044304018568428694,0.9709533779599506,0.478181949750338,0.8186010718011296,
 0.8250384425965299,0.7663029889003371,0.2551752109727009,0.31511634867775595,
 0.45774087390807705,0.10040279006621478,0.8258877268679894,0.896265023473862,
 0.40071372026059016,0.3653855034893848,0.6908374145658399,0.5149965920670272,
 0.456887208931011,0.3219452915712727,0.4014547419707808,0.005357223580244597,
 0.19625145053173754,0.7150937495775006,0.07600738575198673,
 0.2765055097606879,0.9511773452447868,0.46584332077832946,
 0.07585457725126898,0.42302479769376833,0.6925655721099446,
 0.17676567101738805,0.620118827264715,0.9638359944712874,0.3422128097826266,
 0.1507077179684062,0.3288791453478749,0.3817080400736146,0.31390064454086086,
 0.22757533377541184,0.5176037423708935,0.9320489332878025,0.6799301079801251,
 0.031214158642284162,0.2917389786544089,0.6456981286604812,
 0.9783220026982695,0.2505317751318251,0.21916474494125482,0.9959946907849243,
 0.46905097594156764,0.6249051874080166,0.8705719943482058,0.505086976802802,
 0.379593427093201,0.5412360465981184,0.37794777361246,0.8768534569776527,
 0.8881032342662297,0.03370827144700694,0.09157469259426143,
 0.10752305369080684,0.9234794389514375,0.8307138805758507,0.4815443509679086,
 0.13567450281157778,0.19225210664697667,0.9233680972051515,
 0.8986643711563234,0.282450701890151,0.47020618203245645,0.2568372607935975,
 0.6416003707746346,0.45872179640586963,0.7345301262868065,
 0.015613887538860838,0.053221154089449274,0.013491170164445121,
 0.44993113116755856,0.5536036253751241,0.236213534766605,
 0.0029952354585190966,0.9306434173925736,0.6912211798329555,
 0.2884625683521551,0.4701826663651739,0.8433430848428889,0.23707479525790087,
 0.7743082752625423,0.4558939423386017,0.7601341611642631,0.38451198369257156,
 0.8261768288016246,0.8677214941341287,0.22752458440397094,
 0.22970062768134092,0.2990430726969071,0.4445374652243469,
 0.47270656962017066,0.8745749335241467,0.329334105552318,0.1842747618882986,
 0.27204973501799345,0.908335034954151,0.39870211238410436,0.2828613084320988,
 0.13352369651399587,0.22758688463044463,0.018737043275463328,
 0.5978775650611526,0.42566951642852946,0.8264411470686124,
 0.33901729071507913,0.2334823620765919,0.8938764173053827,0.8696153824054795,
 0.9378454274447265,0.37573515189767703,0.6415546670233231,0.9911404406222769,
 0.5020237024809785,0.08141141500380078,0.07081307097696632,
 0.13992186820554742,0.3052794444531375,0.38222976894311833,
 0.5318204758178614,0.5573896361607428,0.8072516883432332,0.15954050333104808,
 0.8293477800601138,0.13347564961522773,0.9218933639210601,0.1648270539301646,
 0.11269880783889474,0.16301695099518232,0.96092594604702,0.3807196700593992,
 0.19225683068337962,0.8350457103925805,0.045799031106658616,
 0.23719472548770915,0.1657356775428963,0.21787143685975108,
 0.8114999512295867,0.8863439031232221,0.7319425095424562,0.8449924711693375,
 0.5129866296069057,0.7237079946988977,0.9643363493964358,0.13340114518516744,
 0.28443700968727703,0.782755726446299,0.30378516855670445,0.6584821498334149,
 0.5920034365560064,0.6716979345179567,0.9211987291631646,0.4457329573104253,
 0.8752081040353259,0.10697060110242229,0.13496134858258224,
 0.6113655746596858,0.979480357822391,0.3051394812943189,0.6714963008857323,
 0.0021011874522479657,0.3927410474756061,0.2407780737007934,
 0.5884872355473496,0.10875369764622322,0.3631175335368373,0.7566278352303467,
 0.5100271176589252,0.46671512833814166,0.8043150599479362,
 0.13456362873631367,0.9226655108706275,0.22603142954015198,
 0.7517922443218692,0.9311001444143129,0.9340545405138636,0.874915413401056,
 0.050229347583836015,0.4673289400363272,0.8281305637787237,
 0.5387881829633889,0.4841758841944792,0.6595058552051936,0.6466255832159331,
 0.17948022184443913,0.2773111881319358,0.7622634319764977,0.5005533674605361,
 0.2321786849519385,0.5615585737508308,0.8584552922213847,0.5738967400789724,
 0.20089330459125476,0.5281853694440488,0.19235589728804403,
 0.5583251391238455,0.24890814158329366,0.42685331153292794,
 0.9949361397795485,0.7401466482534277,0.8244159869725632,0.9001740718855631,
 0.03289974444982302,0.8724634550082112,0.6579852854923532,0.5876849991243059,
 0.48569824301841635,0.7624994941819453,0.504844388825903,0.5891218403580748,
 0.507563878907612,0.9221959779124516,0.20358582571145578,0.31184352257785264,
 0.3657338112676234,0.06013189539578656,0.47155398762193956,
 0.35052976850653816,0.1693521080349264,0.4058766984377049,0.5790684336213281,
 0.9181295865853083,0.6268358056525188,0.9851734964422939,0.5558861740594545,
 0.5689421367973225,0.12418862892692017,0.2518850173629005,
 0.29926903497625756,0.6546192333323457,0.8006124215166809,
 0.39733438965564283,0.6422970418742898,0.6638784526634047,0.5453267427536254,
 0.5172286224688736,0.1611076247728639,0.6581267981117498,0.6094767303327048,
 0.6817510829799814,0.9658180655646628,0.5476017555935643,0.6522028697973987,
 0.03228226379506283,0.11156522933514645,0.1404909961577494,
 0.19979893190471576,0.23063798423933435,0.36563718900710396,
 0.7153137509274152,0.3464641608420713,0.8054396661099263,0.5052975049146284,
 0.9011122362407822,0.9397785945212379,0.3438060043135758,0.2224406767697874,
 0.396221895012682,0.20048703718539784,0.8505624177833944,0.2981417834493112,
 0.022979608107496108,0.46453289403702336,0.6564668659184822,
 0.421687926310334,0.350571821949885,0.8456109828283253,0.1522664479045236,
 0.0812272173183003,0.27972469259268373,0.8387509766858222,0.3498144237680062,
 0.6009788329733553,0.3138069829508736,0.1826129390560498,0.47296440281080254,
 0.6727474365478514,0.12751969538251495,0.5549894715463232,0.5330199810317364,
 0.7378128745073228,0.9794141171068683,0.6100629851684725,0.4415976237957079,
 0.9849549911987927,0.6575953574109569,0.19431105453029118,0.8818152302056732,
 0.4713114737536368,0.027396709776986494,0.5218899395824549,
 0.7658532004437201,0.012470158874633763,0.8176954204560514,
 0.36987451082831835,0.1976332512651734,0.8794542261364878,
 0.37867014479016137,0.15972260488625534,0.373849927476113,0.285597446097539,
 0.812441412764092,0.7382422417833965,0.10187149621193836,0.49037503503530966,
 0.8435169445765398,0.25341284012493037,0.44612253873953767,
 0.07747167186954218,0.7818001389309814,0.6005074346718837,
 0.21766825704240877,0.7379186637001123,0.07076327751939115,
 0.4061809661146737,0.2695799541075369,0.4161218942959307,0.6465085077515841,
 0.7888706603305916,0.482841967336065,0.6660827875706943,0.4168886127944226,
 0.2662614666935591,0.8632234896064777,0.14541246131554053,0.7954572251785323,
 0.04753916964100635,0.2674871098543762,0.6249712013809932,0.6168734546977145,
 0.6026139283559612,0.2402853641503936,0.4256170317228186,0.8179870674312933,
 0.0434795673225411,0.5760048626033113,0.9629171066730252,0.18703901439372683,
 0.8811370247133573,0.12183501634906402,0.35090595425845716,
 0.026155588360329163,0.9706235224993242,0.012486713843782526,
 0.41368441263837175,0.18628000612696494,0.043038035990870305,
 0.3328834177720098,0.5494210628899434,0.3961408094381752,0.38631235294769517,
 0.3249164222993898,0.5327829474530171,0.5998607848286877,0.0707712050568493,
 0.7701624041260106,0.8049845864939249,0.818217953655428,0.1950740816623362,
 0.31293630768322844,0.5865086547975102,0.5587381290181535,
 0.15117507065518077,0.06466113600599188,0.656559809639164,0.8291745721207088,
 0.8033236407825036,0.6315309188917635,0.5350058296422223,0.8396098257184271,
 0.20675285134032495,0.4351391609301146,0.007379203245972077,
 0.3645537720943026,0.19933623581338966,0.28612838225409076,
 0.8656704145053705,0.8682246019925715,0.9123757703869677,0.9128757986455867,
 0.18772616583426216,0.7275414989612847,0.35333775201303497,
 0.3828737849099053,0.5335076252695468,0.5520999374943112,0.49131196389215614,
 0.1424734716531708,0.8156913824611337,0.6530076023087495,0.7391287857636633,
 0.8496866934550272,0.6505040305173891,0.6164980669955521,0.6015004201497575,
 0.1572823231824837,0.35473797984671396,0.9307363722900931,0.9930138644135345,
 0.7135932453192522,0.17128431801641952,0.7898538209232856,
 0.25145621415092134,0.8731888697322344,0.014734614932433443,
 0.6623964273172266,0.062347547663594516,0.4654202396981566,
 0.9125329118096873,0.035400461576723496,0.5598950046354801,
 0.35242440150671084,0.5245911492045734,0.4182709303802147,0.5022400830566033,
 0.26925441528400107,0.993589813465487,0.3903509313051946,0.06398068572545046,
 0.2571727950546353,0.2642979795955037,0.10948374018826879,0.3353365426166097,
 0.9579271659355113,0.31718569294868504,0.9631768767169425,0.9941072262944023,
 0.49103135749573334,0.5563731398691525,0.9148535805708297,0.9495293354231873,
 0.7067342031349679,0.9325156922271518,0.13951920217944236,0.904445211186539,
 0.35914770245691285,0.3126697420424074,0.7837491341212147,
 0.15609048878850218,0.9955060665041212,0.23868912770775808,
 0.25257247539115046,0.7800709475986236,0.8444294926599266,0.980338656403784,
 0.5156206416417217,0.7266046563603652,0.027056911071454692,0.716872035277523,
 0.657739790936622,0.4092914816550429,0.8687490612392643,0.6986632566818147,
 0.9062335949741915,0.6329615802696777,0.43182389283450595,0.9656128957950355,
 0.451513940662723,0.5106792921640034,0.30852309972247405,0.5062539799945794,
 0.24459908358608495,0.002575350361393114,0.8432943961653134,
 0.3445694525566393,0.06193928779025071,0.49994440374093196,
 0.6045593099398444,0.17642761784163463,0.3799149449700334,0.6792131771869474,
 0.11234011262573318,0.397711657773567,0.8454174421901123,0.03536270332324265,
 0.06422340271996552,0.7608407719851821,0.8227876923637755,0.7369926977820023,
 0.2534448378297057,0.07365292278584323,0.7760765015812213,0.7284197093447502,
 0.38937005110845035,0.01109697875207949,0.6862742510508159,0.76855763871097,
 0.2504403519564897,0.5241866220984104,0.211967058515109,0.2006050633844748,
 0.9154297774109886,0.043075998874000265,0.17845551613370492,
 0.0857949247790829,0.7188805694863306,0.1211828077067395,0.13778216205989502,
 0.08283450788037405,0.374992975414149,0.8300056048752008,
 0.017424659660436048,0.8488234438855433,0.12008166638929407,
 0.03888725505728874,0.3394425273758924,0.5793711126647385,0.707774282458464,
 0.544115196073674,0.2216084417293639,0.7360425800115671,0.7357442246074766,
 0.04769333649012153,0.01839938231255034,0.2626405959185397,
 0.9472774453022168,0.5534241204289915,0.09777292933198178,
 0.18955506853148918,0.4396139687974754,0.9174561726667266,0.5331151358115419,
 0.9764249330790279,0.18273952837818186,0.2132021112498167,0.2350482450609085,
 0.5990572897637673,0.7874195284505542,0.645314497974924,0.8377445597226223,
 0.6649508907164837,0.22989680459396666,0.6699942475371059,0.7834931460303577,
 0.3107461715951859,0.18159855436479455,0.46247328118566056,
 0.9563755609418374,0.021409856031887498,0.12008519150791985,
 0.32515030828649216,0.28686044340582717,0.040013803097276424,
 0.6515783379334277,0.846018380811946,0.7443379979160889,0.44820189714635317,
 0.39261967371927975,0.20707528137250641,0.6262609096243663,
 0.9433054198097655,0.8046008357117892,0.6717039273913388,0.4831548729472621,
 0.850206588000438,0.33422285818217723,0.9285990664936352,0.26964743377372136,
 0.12377916899580943,0.43314657830694103,0.5561039410692236,
 0.49184701546565013,0.3327353589189188,0.3352762033451586,0.1382781136049036,
 0.3652076926315041,0.8531627617216332,0.10130961234083236,
 0.05852672898840461,0.7908174030103647,0.9773657281533248,0.5445822793284976,
 0.06632831929918581,0.20591418026738917,0.3721054219045712,
 0.05047364131218601,0.35743450013122124,0.7108366932319116,
 0.8406128092809213,0.442206441271186,0.5937743464813237,0.2428267060140601,
 0.5705704601932006,0.28350936221919154,0.6672940584403029,0.7658989473169451,
 0.11732065603194086,0.6625169464955694,0.6605221640329987,0.6555797957170754,
 0.34361493543878474,0.5911118511644159,0.10779637766787242,
 0.5763896795674069,0.6870320437615539,0.6321487560259742,0.17610726434461477,
 0.07825196814950708,0.584434566889047,0.8386796913423885,0.11465936206381278,
 0.8075293583886062,0.4361858094744223,0.3735233829664427,0.33872417369333985,
 0.19017932910101332,0.26162563958751905,0.6964562837348693,
 0.8393139105843016,0.5769738788763414,0.6140533028105634,0.7437610342466788,
 0.974324620979482,0.39880188958812,0.20401320076372953,0.415206817197174,
 0.3409849236155327,0.8297340168673144,0.6669085348605899,0.8201212805785006,
 0.1996106821558825,0.5866001242132917,0.8952599657196143,0.7899068429314532,
 0.897311468294167,0.07830074829754508,0.22162538198636406,
 0.09298784254393733,0.15171487795285965,0.7205295206888341,0.981283876312008,
 0.7789682441890109,0.6599543806558357,0.8248370771655278,0.9817699157241854,
 0.21740135949232864,0.2123718268845698,0.3964165343791263,
 0.34031329619909934,0.5503601610178629,0.28537525465828306,
 0.2759466706640312,0.7929726359478292,0.5233151197368054,0.19547801570451095,
 0.44369675151022026,0.7294770311282471,0.954892407060274,0.7765602558373745,
 0.201666126166552,0.25314439621185236,0.5114746126724843,0.44907762086552705,
 0.17289557408585976,0.19320824075397236,0.863599545446324,0.7788791080948698,
 0.7247504739068287,0.1047602669164871,0.639606455674848,0.9417040617132606,
 0.5017559069647037,0.4994657817469562,0.8664762218637497,0.2701169293604415,
 0.17194404635293858,0.4088673792123594,0.8034522197089854,
 0.07563093481294036,0.8737152382362662,0.277951231239681,0.2766508753808695,
 0.6123580000213549,0.6799701991602617,0.44369483238218155,0.4096225672561491,
 0.06302843747996567,0.9454068634503112,0.9085074283100557,
 0.23678764020904164,0.3350616442489671,0.32025939231032985,
 0.32618874879572424,0.685530406197987,0.5503092842928801,0.5824655179465963,
 0.8631075345647194,0.5018019301801532,0.103327132386966,0.5369844320213758,
 0.9388651971676555,0.3838926422873826,0.6535380693307599,0.27039004200666705,
 0.8461289928631511,0.796456479544042,0.5822929829893315,0.762478943405083,
 0.5598098195852034,0.9599286399057874,0.32885705614612304,
 0.39779438332077055,0.9994192070871448,0.36336675711060473,
 0.15120791779854015,0.6122243755838658,0.9294132127380939,0.9757645846440347,
 0.1796981790777069,0.290567063688429,0.5496100238239255,0.6256005777414,
 0.9236435642178155,0.01476483960779984,0.41236661412054043,
 0.9959352796463354,0.1691395908708645,0.9157544572166394,0.13468341625842317,
 0.9025924429621386,0.5939987910546956,0.08577528281548896,0.9403518336835917,
 0.20198231158284563,0.049368893002049385,0.8259513092443416,
 0.6799609322000632,0.5354102872164249,0.8975353295432685,0.8854009173254188,
 0.09307751692093869,0.6941508767002805,0.9489553760935161,0.7837328181967017,
 0.8809718263151557,0.2070100297368791,0.27027563303006596,
 0.07058273063756526,0.949963479187725,0.7686117977711383,0.653224996022539,
 0.9498544822436493,0.5590398654371191,0.8483280368102997,0.9502888630282312,
 0.33763879398372054,0.7328389699645685,0.9637492228373351,0.2800393296136109,
 0.30352159242588983,0.5930315443778729,0.9273604381041904,
 0.17157040885869734,0.9448215985474042,0.18200710977312606,
 0.8328959031878989,0.8169216373541527,0.09550904045115993,0.8556240765586502,
 0.8950777344771581,0.7743836466024944,0.17880110266765603,0.8239052129847263,
 0.41671364714926695,0.03425009361011433,0.7281136667180129,
 0.43595317454038063,0.1946107955771561,0.22174165049232508,
 0.6268977736415718,0.9578233787709733,0.805757929928917,0.8493763513342152,
 0.4702236372192966,0.9563546712971197,0.07460274749891771,0.2977379161134489,
 0.7282487771029502,0.8948807790393692,0.28215020138827285,0.8807826876147332,
 0.15112822531630155,0.7795661620669155,0.09732090052134468,
 0.41852024911068064,0.7360465498346696,0.768884102923268,0.7640303971671263,
 0.8194972561283452,0.04773908482891798,0.04393384066254702,
 0.6531836539203396,0.9225639714537635,0.32778506514222683,0.5419559396847837,
 0.885260333292789,0.9261153487933987,0.727399712652039,0.7584265891518515,
 0.1955387533134666,0.22483982070753417,0.5245725096976996,0.9871187908282512,
 0.22289486305416617,0.3440578620541678,0.5345440209169241,
 0.25919880055578626,0.7074919450067398,0.5056545491730028,
 0.10628380707250484,0.8548150848681225,0.2739850631059524,0.6404953445002076,
 0.0916364828100733,0.00895008575979439,0.08280256768662986,0.851915668573946,
 0.8331925419860599,0.9114890884318226,0.24396120508868369,
 0.17494874640759484,0.17906546828334358,0.9931098462365079,
 0.11036878490242608,0.21635078685738396,0.8845514843189932,
 0.8547550638097703,0.9691402044757699,0.8588127204821643,0.8885676094230589,
 0.7938432157221523,0.648701855197497,0.6736773377176075,0.5312180777663431,
 0.19587154381086858,0.8473348276205434,0.7232160624694788,0.5668498458839956,
 0.8717062043052608,0.9576782311721593,0.5182870374204379,0.13207724239872176,
 0.20162280073104288,0.20359558255161214,0.9391071184933749,
 0.2540606394282868,0.21510396618560157,0.5865614180750088,0.5582363712338732,
 0.06495131301886037,0.23859585122999782,0.007174674051890317,
 0.5926334421938685,0.8925355355742001,0.12538408366287335,
 0.052067703880212535,0.10172313263102684,0.7105829202114138,
 0.6898569760560591,0.9446760764124688,0.7954564638746084,0.02571575183765651,
 0.2543781250079675,0.14244303487003052,0.9367291883012792,0.712835709169511,
 0.18616605069691627,0.686789012018518,0.639508121978442,0.3554711468666548,
 0.9273121500389485,0.7838455349919594,0.5358125209380101,
 0.002970898105554509,0.20495694895842276,0.48484334890753955,
 0.599237683829267,0.27474481759156666,0.06445806407948185,0.7052920480752648,
 0.31906606373581237,0.6514764854053222,0.4334924480733897,
 0.11610678796893437,0.30613662595834223,0.7683997026508589,
 0.24800007401085233,0.4345494976296056,0.7543170649412668,0.1582501074218402,
 0.9909813635784405,0.6519939937695758,0.3677055100798977,0.4668379726841012,
 0.03909264773868004,0.615508661025415,0.5432891176479268,0.2856186169704994,
 0.6909081725428406,0.41748894954306137,0.38103096777712975,
 0.02689829325998272,0.6796321192260975,0.22063067232252564,
 0.45979604465808466,0.8468718806460216,0.954842759901922,0.13338679419371058,
 0.3806482911541873,0.8691817059350282,0.30152299936557236,0.5472649657660708,
 0.3354500247793001,0.5369211075382945,0.9084160701884272,0.4184511488066075,
 0.34940467526293384,0.738104423515942,0.20859084824319485,0.2642592599401946,
 0.37587229741413397,0.31676272592910437,0.6213917774063233,
 0.9493973585256322,0.5236979343003675,0.6389474834119072,0.22652708007908007,
 0.3779079491521903,0.41647061601620805,0.11806542619751481,
 0.8960693558833552,0.6411380175749269,0.8730513778519381,0.4198388211844969,
 0.13919097127352864,0.23992154425677614,0.2140324870888085,0.156335983721409,
 0.07635359190988433,0.06355924967347903,0.3845893199921979,
 0.9660456902635939,0.7082468549849282,0.24665432420144717,
 0.44993697264439414,0.6385790517001992,0.16922880379708816,
 0.42406995024082106,0.5331137417018884,0.10717406723039058,
 0.20041818186609484,0.21208105650923237,0.8121564710051038,
 0.15789986489275643,0.2050913475918401,0.4410196310642036,0.5153546910056104,
 0.005414204989804827,0.6671331350011589,0.15911796094457242,
 0.3460284894275273,0.12318064139703577,0.08004594082133709,
 0.26468550137838964,0.8788230203650788,0.9801089920121424,0.7176560826826526,
 0.8885599271466449,0.6228013043472445,0.2605398183203863,0.687647300183706,
 0.4638690623132219,0.04661381533521869,0.8071552674490244,0.5369646281716395,
 0.5648933896759447,0.06886284688957134,0.5197425310619898,
 0.037831145318382386,0.25916988655876494,0.07254470333301166,
 0.7478283796212778,0.8937448999772304,0.9648183960322567,0.8175893144603241,
 0.6830732230358779,0.44808995335093105,0.8413416061837767,0.8900042299886765,
 0.8043680978724713,0.15716268348603757,0.050167102268504626,
 0.35400610656555986,0.379696454465124,0.21747314470516677,
 0.04093995736299816,0.6654936294874354,0.18008860123055737,
 0.23589895229427726,0.42582641089088913,0.8757073019185742,
 0.31501335911432315,0.24844011441227531,0.6267337260238581,
 0.26090354272006977,0.06251610919294315,0.06644945072653663,
 0.4753476572730603,0.5527859859589805,0.05347621982739448,
 0.35646894734336865,0.17446013674619854,0.83676503656511,0.8525347440638354,
 0.9334624131740749,0.03548482533215691,0.9671893238304833,0.425927689498769,
 0.07871407537816077,0.8225341523700516,0.07505481249061641,
 0.8955478323219124,0.7003866878601197,0.38096661062123915,
 0.21311221049988827,0.7795007157104057,0.9077694854709629,0.811673621736352,
 0.2939464713967719,0.9976045901223459,0.6536311882737955,0.07818614691281389,
 0.9762786183738833,0.43094663474466055,0.14395998782853114,
 0.9867896488943364,0.7042517553071329,0.004128072792878079,
 0.0034435338763111334,0.42358424832383124,0.32247308400243957,
 0.7240885794775302,0.07264278589547923,0.6265852336150791,0.7460338566624316,
 0.058118096162075195,0.06698030049475412,0.9022643495426633,
 0.2740053812592672,0.5774451686817185,0.7164437094748507,0.832118751252959,
 0.9747943732079624,0.41407967655273714,0.42129843034546766,
 0.48888091000131506,0.2521512638868948,0.494346821014931,
 0.041012108649486456,0.1394879562024508,0.3367776381608192,
 0.7818686698240264,0.3038789276172704,0.3213067814084175,0.7137066935803515,
 0.6787181012485314,0.6797874339255122,0.8799949993041762,0.03819120427079592,
 0.8249106010355253,0.11071301452716509,0.9495857195305246,0.7522848474908457,
 0.9545392318157973,0.6406391682476107,0.49275118886437563,0.6621294542900396,
 0.31316752626313815,0.7975392353196328,0.8882363143438045,0.7755443977587383,
 0.11476158801890413,0.2790989684445716,0.3226971909202655,0.4378635048259987,
 0.019757037729643434,0.4535908699485134,0.14510639201228248,
 0.8536104538790951,0.2263650335965537,0.49032420039972635,0.4139255030919635,
 0.9997463691736125,0.49305880170678074,0.6611362276402376,0.6605339480604927,
 0.015630240001614726,0.10325991956987202,0.14504633376568177,
 0.019584134177378698,0.6883392811344196,0.6250347361496198,
 0.4714491687839846,0.05515131079198021,0.11134620757181346,
 0.18606996795794006,0.6094135154935864,0.42477494292627394,
 0.5143861679336744,0.6985898360503544,0.9821045464052237,0.03761023781045392,
 0.27906937932319253,0.6641036963963927,0.8874776929457375,
 0.24781686772719502,0.899517705340537,0.37085076091352864,0.7037471706627328,
 0.25623163895681633,0.13357723509615915,0.6431427858380329,
 0.4062755186025693,0.36779868387586256,0.9655053666692455,0.8200770800597929,
 0.7578788142327098,0.7716084249726333,0.6407087862179255,0.7638930872006106,
 0.7952449242211086,0.35946718138792966,0.13507783145007257,
 0.38032402492216766,0.5430334774633494,0.05812080099844141,
 0.5055183150224457,0.4544501708292794,0.3931621984784822,0.7784260214914558,
 0.5572060359142709,0.6000474102901121,0.19629219465823355,0.9534458967084319,
 0.0939819386804368,0.016008973708994456,0.8378844193628411,
 0.11399500970281673,0.17449541321483497,0.8224993915169053,
 0.7821480981006473,0.49695827554823657,0.3101738673707187,
 0.12077422802056481,0.16090279703984256,0.05011494390985338,
 0.1580742149777815,0.4447807479139082,0.07304348596436583,
 0.05964841337709659,0.5823681500807395,0.6370685640881448,
 0.014863706504946284,0.6276567257553232,0.5096273311140794,
 0.020865710932920334,0.8146292955522818,0.43644845894764006,
 0.8155417826473598,0.3665686062106973,0.4234555672589586,0.5404628999747839,
 0.6253246280060535,0.8757298191290336,0.7504439100092822,0.321046921339734,
 0.8782859637193791,0.5322997688627568,0.9806012204700959,0.17999589267210458,
 0.7430662566729388,0.8247958398125067,0.12616756602421597,0.8510790839801464,
 0.44042555400776395,0.19336327741274983,0.9816678764296951,
 0.9664597504843303,0.6309574944830467,0.046828318682972214,0.871688261984641,
 5.801078428744333e-4,0.950240048911194,0.8446424118077914,
 0.06661640354877141,0.4523423221625661,0.4844312731927649,0.7837533282085319,
 0.46206747660604,0.886777769762563,0.6872270681652685,0.8474571812359991,
 0.01952054824037386,0.9016554857317077,0.0974402781658269,0.7810577605667816,
 0.9839044463424995,0.4920896029139633,0.49856351494705586,0.5977057972255118,
 0.9751845953628865,0.5211073091778075,0.2656410695105871,0.19606755963535605,
 0.5970125148261014,0.8083645424987775,0.6009686499887316,0.7217841801841329,
 0.17342059451195213,0.7862249976955263,0.20321513067101182,
 0.9251817521967495,0.3071676532857479,0.7311950883959062,0.5249745408012804,
 0.030963578503674705,0.7623001351631483,0.2110831866511571,
 0.6139900827267493,0.9687108128444302,0.3224201890273203,0.7311869293199588,
 0.023709152300479763,0.8951364971206603,0.7667817716619622,
 0.8999079639838954,0.8649240031595886,0.8276511185207674,
 0.014015566232430943,0.3784905304548807,0.7173285097721402,
 0.5742869233496796,0.6448273331863639,0.2664430294754214,0.9741948458560288,
 0.12239985680853971,0.07508556632114216,0.5527849373395676,
 0.9619737243253472,0.6399216812540404,0.1796750238251027,0.7874160440556796,
 0.8468473524357327,0.2901850909157694,0.7481454700418789,0.2248732985698042,
 0.4613963313564494,0.027103729668171206,0.500633737227755,0.7190938718029787,
 0.09092028207957548,0.006414964477534757,0.29758530270654593,
 0.568656019493949,0.6174304926262404,0.4991952896855336,0.2544606548978262,
 0.9358273916260853,0.5103194053698427,0.032972864186966655,
 0.7819907505124524,0.3998215657578631,0.5516515147094561,0.6774777973874782,
 0.7126347707681048,0.8122207605670145,0.507178610082075,0.5246970552211767,
 0.8957564766440456,0.6788874706265888,0.11900847844160378,0.3535675304627365,
 0.1491479654552239,0.1408107305070948,0.45917732274928014,0.3639129037071913,
 0.38088043348658907,0.18116013082978344,0.987770691888949,
 0.12513230021360622,0.4732005295270312,0.16796732267223813,
 0.21983142010718204,0.7538856936314977,0.5872896727555329,0.5839509028978976,
 0.4402226728186365,0.7606572463972989,0.6403249709546545,0.6955883908855247,
 0.5721125871051027,0.2998608852007727,0.2613988475639806,0.3960221113615473,
 0.5438978214548723,0.5935822024832543,0.1964377892793735,0.4908273154885283,
 0.6657868529767378,0.6770168780876464,0.2330211939597211,
 0.015898695895848247,0.4153175329064893,0.20303323610772261,
 0.5917392613186048,0.23362249811282476,0.29330813452350224,
 0.06413626625445878,0.40266592965072523,0.008468243816906473,
 0.9078658489356575,0.14889784050055255,0.7211699298072741,0.2934328611271162,
 0.6586249838219519,0.4625996203613236,0.17970743980486314,
 0.13260241448283594,0.4158768213520049,0.9873279132852775,0.2912558412193408,
 0.9135753684343761,0.07086223236138656,0.33297686979123653,
 0.46738238663910436,0.695725276659028,0.13911406592089293,
 0.37395945831618527,0.13048338910397328,0.38878538229560144,
 0.7685073878601738,0.02845676032581579,0.3232974065046267,0.8658112462942698,
 0.8493538315806486,0.08854746446830297,0.26794607190622344,
 0.39683322330959825,0.30370525560660266,0.37000597310420225,
 0.9180999211736234,0.2794908994595928,0.12982248646674177,
 0.004339002531516867,0.24624212525757672,0.4874321785257604,
 0.2574379303430342,0.7608112375575318,0.4988159664123134,0.8080326626986124,
 0.5346092027153814,0.07638364206522441,0.8033406792716863,
 0.19135875292849402,0.4147255794466762,0.8203789425004778,0.534720945965073,
 0.17801461607424507,0.19365128219799077,0.9087629715248843,
 0.8331532492237752,0.08165337993972588,0.33800651974167284,
 0.5326884162736432,0.2587271079240736,0.6738501227425684,0.4061616837367439,
 0.6497352720225643,0.8655946447564231,0.17378674406521666,
 0.21759899118679482,0.48185109921894176,0.949807686511755,
 0.11173708231838897,0.9463652145205591,0.49967582667437316,
 0.21244524636411843,0.32878562370242914,0.8956092203697289,
 0.6758074925531791,0.41184601399586196,0.8028234491220057,0.738833671922846,
 0.543675038965758,0.9412883467261036,0.8803613653720346,0.6093562813545021,
 0.6981991386213333,0.693684802672619,0.05921446004197595,0.4499266053581914,
 0.47348714824412896,0.9464520761024658,0.7579651106664383,0.6453241580716506,
 0.7748140627608741,0.9152000736417818,0.561493794486816,0.27636457312426443,
 0.7286820228530505,0.19276194288186832,0.7411526965158473,
 0.27829892729687056,0.12682188003349726,0.27773480317498056,
 0.22471415160672292,0.11980620078984883,0.43971010133990884,
 0.043153817797781935,0.4124594872410927,0.5109801208477683,
 0.34941757842199384,0.5305390579523168,0.6491672848550616,0.6697126750261849,
 0.40662558249685876,0.9001766990087406,0.26486208239649534,
 0.6429767289850598,0.1983345388256642,0.26550853455086254,0.6601334381818311,
 0.6776460372789237,0.4435230463989842,0.97208222653838,0.02151475610913267,
 0.2085746649882455,0.7394575100638296,0.44235508369956333,
 0.031142277733611112,0.464493348030927,0.37523503927048163,
 0.7907222071338706,0.762426211222793,0.7461145729647636,0.5411269225624666,
 0.2857882587602514,0.16278676708781536,0.9205658083900725,0.3794721558859805,
 0.08232750859503424,0.7629319976621386,0.31534471308602574,
 0.9262715659192235,0.665943081865286,0.8586408765614556,0.837403405332668,
 0.19327853949274076,0.6427833580594651,0.012782588213699375,
 0.7361528663755452,0.02876327477423768,0.8245417435202369,
 0.31488105259208465,0.5560640388343548,0.08803275399655708,
 0.9692736733011466,0.2681750480784191,0.7870997633445179,0.44608306013144594,
 0.7382963546979442,0.09357971429160306,0.09388983865675815,
 0.15069699029128825,0.36553600565059374,0.38875373434319205,
 0.6601993747378307,0.7540032971337738,0.10549504625915829,
 0.04451232163266006,0.3610869031242727,0.8676760784655527,0.9739970600459609,
 0.4993052454710064,0.4480945616448042,0.9708943843872551,0.898698340133952,
 0.44933529556312646,0.5989404375458015,0.9970495285939738,0.1859035917472862,
 0.8489095513990064,0.013101137413767616,0.25834516650319483,
 0.6333555414821441,0.7032150191233011,0.4628969815860404,
 0.016292468457775877,0.41832621780543866,0.7592510953424362,
 0.4504423245967959,0.8429593506834375,0.8727017613401823,0.3348727844593128,
 0.36587046144685675,0.3851906077481537,0.027479018517083453,
 0.9650226012461744,0.1275176034406409,0.8476839172974227,0.6468223130721724,
 0.574272152378867,0.9910469649757174,0.7505862791603191,0.05848519444709033,
 0.8311076324196763,0.46387561040925884,0.558272565058079,0.20885522628632214,
 0.6468892553599924,0.45692008768650794,0.24610154805824092,
 0.17798218385694597,0.7276870254389043,0.1733472191586024,0.0321594894057301,
 0.3962910214944626,0.03160702564504303,0.012952281515730468,
 0.9415998503225347,0.7793057425318795,0.31711355854757217,0.1188147224437539,
 0.24143942775215554,0.9203237537289809,0.4148377954261372,0.1823530993815241,
 0.9242545137030235,0.28558411379905213,0.5282193235791521,0.1301442549719135,
 0.3045821294889214,0.2375479693820185,0.46031120323395147,0.6192439833159565,
 0.9339099065171081,0.7453124176731187,0.45446725269745014,0.7640440845761653,
 0.49853862315848363,0.6388012747310665,0.033689191918801153,
 0.8903926391599379,0.22276507355271302,0.04782710706310289,
 0.6748504682301373,0.6006617829741849,0.07150690120936465,0.8025708067415374,
 0.9876713934409235,0.9084569925371623,0.4721404950751469,0.4685992871497424,
 0.20336987512258686,0.4140085898146215,0.1519298339466426,0.7110032487492753,
 0.16243051815205167,0.5710746374434372,0.9366292071138962,0.5616278440540213,
 0.12134394013391792,0.9379191421287955,0.703989483129361,0.02989746661796,
 0.33569901556512993,0.9412271023882945,0.6917525955542755,0.8810400025068921,
 0.2441351956131732,0.13908346762469648,0.3381430111665684,
 0.20892910494743344,0.41453079343867005,0.08795319004731461,
 0.8488223553995154,0.55811626243583,0.7027165412109168,0.6540410284669743,
 0.2653620152667291,0.7060487512716437,0.4761973489863045,0.3241621086901896,
 0.5520290934175475,0.3136806836018077,0.9356641567895049,0.5533928403286814,
 0.6755870481536456,0.9892333291260096,0.062303684255398006,
 0.19654724354105868,0.3458655152645005,0.9095361304880076,0.3571482118926804,
 0.5733319684756671,0.8442724884812778,0.2153113048240689,0.8662728787529991,
 0.6166043813455129,0.19312831660490826,0.25632690292119287,
 0.21156329337658097,0.2609088859595027,0.11255528182311836,
 0.8056453838366179,0.33499031738369445,0.9091679184444079,0.7173235628926112,
 0.0807065482620608,0.04475658821008588,0.06492734986133875,
 0.3855276804065698,0.5643612926968601,0.6588998421182131,
 0.016426757132028458,0.5188997606795743,0.6550361644770966,
 0.7778832812504491,0.7524324163330393,0.4490636469406778,0.23648619851368358,
 0.15449773684312085,0.9039100978698063,0.29339078219223946,
 0.9217114154286729,0.47308064802274463,0.52691890436128,0.968508340602943,
 0.46252647711097994,0.8194263160775992,0.6295461760613851,0.4435720631263069,
 0.2288645758700748,0.22213451263981654,0.8153297445747143,0.9567208244799323,
 0.20531736699628833,0.844868769451812,0.8133385555259423,0.389955835076913,
 0.17896750591087518,0.9791765866095166,0.3900278874475894,
 0.03725888927765264,0.9916780023810131,0.2580164146656979,0.8683874588617999,
 0.8489985412916818,0.7492380333054054,0.08158647118025708,0.5640647304728039,
 0.7380104939544935,0.18656238215007526,0.6464464554080308,0.597185505512131,
 0.21703357933449274,0.358554593771821,0.002024455328027752,
 0.6299668261140774,0.9008198991813198,0.9984881020116894,0.8810610459405767,
 0.3934153138714377,0.35008346503707743,0.634118103517367,0.07888862987762724,
 0.5098418205100017,0.2674232357882903,0.11574629532927216,0.8007619009229001,
 0.7763324756563148,0.5670950247500427,0.261060083181645,0.3311583950827759,
 0.6718235293696967,0.07576478353239269,0.7497113786502125,0.6303631147620883,
 0.5732842740415162,0.9221286005002823,0.16897406192514774,0.8102135634608805,
 0.26371487635610924,0.8309788860785376,0.3947595967560952,
 0.14243995560810363,0.20947548625869805,0.9222532532473153,
 0.7078422053148969,0.49167077440502804,0.015271614175766168,
 0.11097488207599826,0.8286647112533612,0.5775711980803618,0.7899711262392036,
 0.9853486277472213,0.9698996843791631,0.9421475040721792,0.8359671135959362,
 0.2217776980217515,0.3626540141237493,0.15528319696182702,0.6048229758558873,
 0.8456205415010877,0.10865348591569668,0.5868551274285134,0.5540154969666327,
 0.9442845327965956,0.43077660575595655,0.681016301046961,0.07219102758667306,
 0.48911938331737403,0.20224051284727385,0.07759524011420682,
 0.2956747224402244,0.4167851490836514,0.4737440791797216,0.32080195407934364,
 0.5437658742736391,0.4957215677323026,0.8262510214464305,0.5882732991812593,
 0.5841666002299839,0.2782705221786903,0.25140626663243326,
 0.03438888518166794,0.5758588547370693,0.73863772134922,0.9477205196051605,
 0.20996889908898386,0.07115295033986202,0.9293747124009186,
 0.8528173943300392,0.6329082933648109,0.5594849554984741,0.8501821132927393,
 0.6238076192552903,0.6831029044962319,0.3329771163464377,0.7027420492178603,
 0.761124902507463,0.039082575662753305,0.7720334636839056,0.8768737401592537,
 0.5538209192136285,0.08691445808161113,0.16946728302931502,
 0.4380999134908198,0.7165121017529181,0.1204648925715126,0.524899156686684,
 0.27065578967240866,0.9076435093296,0.7383428850526824,0.00984315191211027,
 0.6899814197869703,0.7834540120285362,0.5345715161596799,0.77526210215803,
 0.92153502808864,0.4920065955795445,0.357327799212665,0.860391156450466,
 0.7884570384378898,0.1985356864086858,0.5610657295966456,0.11036361620701973,
 0.1902710819196468,0.8469599135934548,0.085437751318867,0.48645764087384324,
 0.430799787357954,0.26953979070969303,0.6444475117730681,0.16475763458385695,
 0.996960535275191,0.38132440814260626,0.0061564037660961635,
 0.9044722347656873,0.9576451500222639,0.7119184170159383,0.9078013890099217,
 0.30053272337814607,0.25035492637279755,0.5198467672229927,
 0.8741195199484926,0.018117331326624964,0.4042449353788389,
 0.3191484045346913,0.29240172607359693,0.014586204315827689,
 0.7504423560467485,0.41807157064023204,0.4686015283706487,0.3229629762105053,
 0.7649508179658396,0.43956424734025956,0.1188533335533769,0.9923317805116295,
 0.35746519958343803,0.6635236857929093,0.2124449327821637,0.737457527474222,
 0.7899665729288423,0.721404282514762,0.26567992347408553,0.4172170096557143,
 0.2812012128158292,0.5090418488974777,0.8915889978514818,0.7511740647719479,
 0.9630734361160269,0.5939951102135698,0.48273223331624815,
 0.17870338411303277,0.3820280546800684,0.7524064813070296,0.927773309384813,
 0.3102008441781058,0.8492537294224213,0.9541208284619044,0.5967457561847329,
 0.45468326028669037,0.2565400034605725,0.4273791138842751,
 0.20550668711381748,0.7365966370463167,0.9548484195055693,
 0.18479335112317474,0.8684374804799883,0.9347862249898207,0.4075863124446544,
 0.7927327557546698,0.13902504711172514,0.3615744414374954,0.7991525367267478,
 0.2927660638705869,0.08877383264647976,0.6570416681198158,
 0.022229200766807233,0.36534755612650127,0.09679694299735875,
 0.6497353841734479,0.6451304279025182,0.8919126366861647,0.39241923741128437,
 0.49833083078535945,0.4709323894439733,0.224715191562736,0.27383192659295985,
 0.7168173613218899,0.7494926025790609,0.024397561281237268,
 0.4080770082988936,0.21583107772066645,0.21324375612426683,
 0.4397201292839167,0.8522914223680542,0.22315629582229646,0.8772609963142622,
 0.0014861165204571591,0.947841561287581,0.486160853165747,0.9653285240839934,
 0.7473353267386572,0.3672200876636065,0.32234480399565335,0.9152643092102735,
 0.921887428532805,0.8892236986681927,0.9829585242150247,0.7824445123336882,
 0.8916403792195806,0.10272740495547306,0.5209563244441264,0.6923798168333996,
 0.8116382226172697,0.030342420360365163,0.25782334053513734,
 0.24404007772317382,0.021845559671078374,0.8474446515865653,
 0.2686032969231167,0.7676772662454819,0.6611335221421406,0.7610387702878644,
 0.04589116017167305,0.7954838632962025,0.5887277657235195,
 0.24078746490937197,0.24723245583541076,0.9061812770536706,
 0.8957036142476597,0.41974681049976403,0.20235788823262957,
 0.8685269681702258,0.7264647697210247,0.5727150121512478,0.7282784611486858,
 0.5751955558669121,0.43638757202685385,0.22420016355859074,
 0.5040219284596064,0.8298957981784295,0.08175017838222709,0.8262971749077995,
 0.62078088784775,0.6430424419497793,0.17147936138665643,0.7207783769169361,
 0.22012214497420313,0.04314488371446357,0.7740688376675748,
 0.20347749112208013,0.15969083403062823,0.34585012843901697,
 0.6909611958844971,0.9368842645924224,0.846469632838765,0.327372981291591,
 0.8665487016865834,0.089239935871821,0.6075696990102131,0.2200328835078733,
 0.6893038638811007,0.8631500640380124,0.4023942567580251,0.35127532719878984,
 0.48773364076581627,0.45310207488092624,0.10245008961485935,
 0.2370697972768434,0.7708598671349888,0.5347354307719546,0.8863479112223359,
 0.946874667214047,0.3491505768198482,0.5151252508209615,0.4279029120452621,
 0.7361806188139247,0.9490683000259875,0.9981301858742904,0.5948490325664674,
 0.24758290890910728,0.39142972291810174,0.31400359275332246,
 0.10751050305558185,0.1357250421303522,0.1646285301762367,0.6097338863544729,
 0.39984772497087295,0.7683187131951527,0.4032661650670334,0.9661438106726054,
 0.7054305719808318,0.11214393923677679,0.24059423444640815,
 0.4343022836493068,0.5800585328962332,0.7522998661679072,0.22643651947215404,
 0.9012431436356659,0.06332291580630889,0.3030597186085332,0.3327910304931895,
 0.539937068294064,0.03976563391695842,0.482932381449668,0.8285165370429198,
 0.7861987536602927,0.34164030093255204,0.7776846686736414,0.872519164198522,
 0.9413494522165081,0.7840739515668885,0.891903982602018,0.5892336682838384,
 0.31583208954689923,0.6830392878724685,0.9246522038678286,
 0.07452077831969983,0.8331031042204037,0.04843343329020622,
 0.8045373107514363,0.9232858364169096,0.3760467483500598,0.6219866597868696,
 0.3446474496198685,0.16353464687665853,0.26536439630152386,
 0.26799770381965327,0.8040663883942004,0.697094387904198,0.9614095472652084,
 0.8234900463116289,0.9731415353716929,0.9548245422062185,0.3644186616746329,
 0.7387634767212428,0.8287148145797476,0.28142291910366257,
 0.37100771964975143,0.6046007066880204,0.30961269374223777,
 0.4999225053443277,0.03548732637041474,0.026748318335844012,
 0.46904907408629126,0.532314432385657,0.6406708069037194,0.6757583791318852,
 0.44792490698745546,0.07737468175169482,0.7205705128913221,
 0.2659275691517423,0.4595208394810032,0.6182538053512385,0.07826923580015988,
 0.9748087968627445,0.73460710615143,0.08176894080057817,0.13631788848231874,
 0.9614198389585135,0.6376177997730941,0.599132474955674,0.17336841834000782,
 0.25272205724735264,0.7466521204780557,0.6202385450208969,
 0.14967437017050045,0.09851099640822658,0.23167907239674224,
 0.038344903088367466,0.13264472675097316,0.19051335248348744,
 0.34540156036726466,0.9853817144288411,0.8849271062197224,
 0.16257465090774503,0.9135336852541839,0.28595649575212356,
 0.11407832340855117,0.2686674782244822,0.4313971491579298,0.9644417933626902,
 0.19392765186595906,0.6401869182638367,0.49063109248737014,
 0.3130267251857237,0.8759653500937105,0.822362528201018,0.47053837705080515,
 0.01530374247058941,0.9110907774096759,0.7787711585455704,0.9475005383491647,
 0.7881950477564598,0.6667065033533789,0.6433232818073487,0.4303350387815049,
 0.8059116734338114,0.22909107792937178,0.918339721369316,0.5366932469561035,
 0.07091549393371444,0.3473294851686769,0.40760054773617516,
 0.7750588741981631,0.20355624016224394,0.15462299775274146,
 0.6552311227754795,0.7245174566044673,0.6648431547441913,0.13469557997246628,
 0.8101534313944677,0.09380775782925477,0.6944454095391119,0.5059148642425341,
 0.20955243853581174,0.06635127698825394,0.641649828053751,0.3620424007149571,
 0.6366428314912757,0.8996778197176731,0.11380901337895555,0.8743190827571691,
 0.5228623744461524,0.9344918398079151,0.008767938967447986,
 0.5298740336078207,0.5989814080554721,0.13848686751354167,0.5889438962203916,
 0.7762142468484529,0.9172266231270636,0.21879469915055716,
 0.36965274607237475,0.6738394219593795,0.1749289753194747,
 0.12510394831522287,0.1012561250560986,0.38057954915296044,
 0.9653336860886192,0.7677427399889677,0.25754660198696644,0.9618257164460924,
 0.16932753670225154,0.8927660137749951,0.25136230747627886,
 0.44722367618620584,0.23668478220479283,0.17857541855231707,
 0.8724756090207828,0.31728755498893446,0.10236190192773642,
 0.19646051460043412,0.021354568600845836,0.5827812700729584,
 0.4929036511781235,0.01654847953683891,0.4189406750362997,0.3014968480946225,
 0.21746831994042437,0.20200386381181779,0.266089947983414,
 0.02117495802721947,0.10289592025924432,0.47910605341641044,
 0.5847028162473548,0.7139234176292026,0.1632381521931635,0.6374505709622671,
 0.8192298721669734,0.5764465263835001,0.6463890151836507,0.585325132166642,
 0.9230938788781319,0.021993441296282668,0.6469531239792611,
 0.6541026119956184,0.8923769512818888,0.4145028355134076,0.4132045472146355,
 0.18461213947907706,0.31691750155086207,0.8218916824433206,
 0.8522500830392647,0.8719543515437809,0.9648190293650436,0.33668538191384423,
 0.4153804612230909,0.18474872817836885,0.9752185880280236,
 0.10466536251250624,0.2552698614920689,0.29548696924204465,0.640782762630588,
 0.6050244687009211,0.014044428518476426,0.047338517926707735,
 0.8711470230890296,0.057339481914489565,0.09248293827890719,
 0.0092290729196568,0.2505875021617394,0.9166790494305198,0.9248140872768946,
 0.5483425115945433,0.4599186182073347,0.09699636489260854,0.7502656393191695,
 0.30608237698643315,0.3755586149448945,0.9963101360648585,0.576530290696362,
 0.36322201872259985,0.07658748263430581,0.7134342758125192,
 0.5770830012509267,0.4279399817664894,0.34991939019218166,0.7200286197133787,
 0.8607417033149102,0.6486598753293271,0.9958257797655128,0.5051300232403297,
 0.5554458756360764,0.8228238216203322,0.3169577411679205,0.17880541469407651,
 0.6868245501537129,0.5745136482978692,0.6699693603423388,0.10799064939942216,
 0.18622685664742877,0.654286270541633,0.6253513258807857,0.6487890781716331,
 0.5704792314487576,0.3931519753903697,0.26081443070005583,0.9066342402423903,
 0.17551614392240433,0.5033798966835846,0.7169732896510327,
 0.050380931526362804,0.6940256945135536,0.4925958914036511,
 0.7037171356196734,0.3214535031304029,0.7060560247528959,0.9521439345011858,
 0.15473200897114003,0.46111528339187635,0.6552827497677028,
 0.40885782606246923,0.4550151059888936,0.8877468356058483,0.2807404450124451,
 0.26467001331113627,0.596411633018065,0.6311713949813403,0.9741290477964215,
 0.5779286875252525,0.9603179211451536,0.9117747851536863,
 0.028716909188566175,0.8243877117947632,0.9869921708330066,
 0.16368426790713475,0.5332365233462225,0.4612609835794593,0.8505528655784502,
 0.28378578704021173,0.9059904638187825,0.26367662850119467,0.816040351111871,
 0.9591022280406011,0.2930165073942772,0.24605022210849692,0.8557494925950028,
 0.02943607098131007,0.9163408211509685,0.23826989406240673,
 0.5160545803782517,0.23368253978262377,0.5001787826268553,0.4265012710028383,
 0.08743894217445347,0.6439701107450251,0.25716055868678744,
 0.04848309061070688,0.7345266064653153,0.571193131559109,0.3869338587335587,
 0.9186441425612014,0.6230555503069517,0.8942661543003886,0.5185671479713555,
 0.9357434076029107,0.8996117376256931,0.7485949530071295,0.571603494040986,
 0.2642063444516405,0.6231647975492867,0.7078184457386744,0.8618356279154903,
 0.6828773277597355,0.8940522336568604,0.25991277378112554,0.7870056138830348,
 0.629084797354837,0.3971263742390029,0.9993675349330963,0.9419684023775113,
 0.20939703965820244,0.6459813421295072,0.4172930810944404,0.0723217568375123,
 0.21624491353198683,0.8248436535715238,0.031152903083339623,
 0.3601302443461374,0.6038649203591868,0.8800999621954793,
 0.004389371560130084,0.13348560164957513,0.6063064952054038,
 0.17618331432382706,0.4661538854533236,0.9877209043526596,
 0.19345961014430824,0.41716382242851846,0.033360237619698974,
 0.7567389321000022,0.9746051238774291,0.6203545184164019,0.5658838386853329,
 0.9258407920378283,0.11984923179893303,0.3506733131857225,0.7975995268449648,
 0.4838105720258281,0.2354985831809786,0.3630422677463836,0.7999136907919104,
 0.080330143031912,0.9753385231546571,0.818256619788596,0.1832217239412005,
 0.5738771667498168,0.323091837870327,0.9324097221587977,0.7794128189395124,
 0.2449942092437477,0.2875052903076142,0.4174097656075949,0.6811134912378112,
 0.8955252591108567,0.8323748249085285,0.1531029048103335,0.538238068787044,
 0.783071011958331,0.8517306829031621,0.05821365449344351,0.7353754540293986,
 0.6531208376356374,0.6946420442968119,0.12664239774150365,0.4713018622734596,
 0.49404786022630254,0.023898442231296224,0.0213313693160202,
 0.2931879765569718,0.1377598228379745,0.36874172221349377,0.916896311241907,
 0.3265793386983241,0.5183677909963075,0.3715745335526035,0.7507235259635109,
 0.6956231138980318,0.26256355024009925,0.44964403984100887,
 2.8639627754900587e-5,0.15140762793438822,0.870761289175013,
 0.906246634569482,0.9569893772204133,0.4251931749428558,0.8920311106226578,
 0.36970054378489625,0.4922688296789359,0.8125406570005796,0.9822745713837052,
 0.8475519549852252,0.980099928071357,0.8397395831756781,0.8890129056674257,
 0.27133223213949453,0.0766492286348871,0.24272329155426176,
 0.5270547326291246,0.5199622058041815,0.18581406633584352,0.5670182770384665,
 0.9622594878008122,0.9965484471244337,0.9109227506920411,
 0.021033492630004536,0.04430839372764339,0.3243026172497685,
 0.42077554370345116,0.29052922680034965,0.9788676341816245,0.081180806969394,
 0.5971570963900181,0.6023318915773429,0.30459861563108803,
 0.06182358863237791,0.8107533594102927,0.0806063915053814,
 0.45168380480636106,0.6181716288160949,0.478093153700499,0.6045341635550848,
 0.013647968874381466,0.11796401597127149,0.05398410045210167,
 0.4311312161626917,0.5237944588110625,0.7480826930487483,0.9428134391716839,
 0.49556813892564966,0.7309204651250812,0.7567416286275384,0.2695438121578688,
 0.6142086947874331,0.6481149958823859,0.43588025874235714,0.6612985954682433,
 0.3445243802583555,0.3959521073768314,0.38061116763381897,
 0.25837316501537977,0.14812353804718525,0.5611035822495488,
 0.8579727899516622,0.15470112928998536,0.07042320076739539,
 0.8190300787142677,0.6525914784723693,0.9414079781763003,0.9411661238177806,
 0.42756586905271277,0.7905427407206732,0.8989098280705206,
 0.16665345951890587,0.6547275806972702,0.5882284574258114,0.3263250064280321,
 0.7228856288847777,0.40742487330442423,0.21851476219531718,
 0.033718801522646835,0.555188875620936,0.7411090073849304,
 0.19200231534973133,0.36550494048563853,0.5835077512557422,
 0.38157326880286735,0.5673742754836121,0.728172744525026,0.17652736210277598,
 0.6364295883539338,0.3793094966331543,0.7147927550891299,
 0.031757015393914845,0.038199959913265635,0.6354255744508164,
 0.5528300430316663,0.0034664562190859227,0.041715543498215046,
 0.9908063677155317,0.9611794364311279,0.9326973418677303,0.4851449369577602,
 0.4454302461742706,0.6451178394678923,0.7567770894956719,0.5605270802394525,
 0.4387872373533991,0.7407701422793539,0.2754056538753292,0.4153081691399255,
 0.2075036216148466,0.9776322023078088,0.4397914353163558,0.9504099029744427,
 0.013718125202126785,0.7384275210886722,0.304408039877052,0.6503351295771462,
 0.9973929415450383,0.8826806423150946,0.5705624436782617,0.1450753444306988,
 0.22434451857718818,0.4840607615385901,0.42021930337091806,
 0.8344576874075942,0.7809515202829922,0.0642678132669442,0.3490803600957888,
 0.618423529971754,0.9557252411211488,0.7846798706867932,0.2429647678121425,
 0.5542074646127761,0.8072823719363742,0.3502024513860704,0.6948270551152707,
 0.6994810268570913,0.9301863346135383,0.11481533025730306,0.5606141952689331,
 0.3368364873502111,0.5479570102331682,0.09406309040676764,0.890710652617781,
 0.1384772068366662,0.30355947639394076,0.37552369145086517,
 0.8263276728740833,0.742108019777449,0.2712539476374538,0.8935790084168167,
 0.49732712700133574,0.6222539877838372,0.6490826782944069,
 0.14965465922645704,0.6129812083037183,0.6456502232998632,0.9035912747481973,
 0.25322092854830647,0.5205211325657025,0.19723820909523793,
 0.5062031218327978,0.23007836108102286,0.8758076252084294,
 0.16123848435811527,0.6338052863298164,0.39364447999172736,
 0.9686791106554393,0.6856267958651054,0.44772419012437226,0.724631103059541,
 0.29944032219731775,0.6375543363038996,0.04034453865977938,
 0.30997757619244093,0.46310335982995865,0.8560167249912901,
 0.46966178946328696,0.29470618792040704,0.4995604387071384,
 0.06501477348406703,0.457354377058387,0.28596878813078297,0.2924355810115533,
 0.5103209235994552,0.7369042009122412,0.050555352122617525,
 0.3468593327098001,0.660484013636942,0.9646696607807453,0.21675089042193052,
 0.5134599374902385,0.707248096186107,0.23501856099963936,0.43834519838511843,
 0.5850125885725233,0.2949874534070577,0.07141548060611802,
 0.23142808125241066,0.4935709376755323,0.498349947481875,0.12741066831291703,
 0.6773694466414634,0.10752224456833026,0.565841957535644,0.6462106402318203,
 0.4074768959613623,0.4399164852404869,0.5983984140672477,0.8024824545761686,
 0.2836836748269437,0.9724704196663113,0.222127720035857,0.7110258979925328,
 0.6583609354071036,0.06836672536068122,0.22553258678838772,
 0.6205510359554883,0.09035672611100865,0.33100138821081115,
 0.13966012923660653,0.8435333504876912,0.6877290321271005,0.3438395828948261,
 0.11517417146373765,0.11881870664431593,0.9474629168981766,
 0.9125404867405196,0.7021616078246988,0.07909209191052458,0.2720929044805037,
 0.1222963303650878,0.33455076871608613,0.8271918848895499,0.2983389795290361,
 0.6836069809816968,0.18711127362512736,0.9922070656139993,0.6755609069868114,
 0.9085969009378804,0.8031832175577903,0.008797487172148899,
 0.0072452077628817335,0.20030075381015489,0.6420010209156768,
 0.5309667483317105,0.2771709335828234,0.38501824348238955,0.9912216498436474,
 0.47166755009824324,0.4207653723128688,0.34131842272679425,0.58262238706282,
 0.5085915381511606,0.5265215695427052,0.715137228139627,0.752371690330141,
 0.8995668868781311,0.8683502924424842,0.7664410411774176,0.303157659426754,
 0.7190810651988061,0.746686438574319,0.6896878583353917,0.9826888645112599,
 0.0021742920552304534,0.8893188783977894,0.2198993261808596,
 0.06819651216228151,0.5024041126632259,0.48881110116489557,
 0.7645225001022199,0.46412681104096787,0.6852994286682123,
 0.035658970415549884,0.5698568144969491,0.7994793848116557,0.25889481543715,
 0.7728739567461981,0.7750893990496922,0.5284343683272124,0.42870949790452184,
 0.18929085640938093,0.4704013515320177,0.9950951287959553,0.7954206668954943,
 0.6191100035705084,0.0744897089617893,0.7252195076309289,0.8647702749756382,
 0.40680559428267316,0.7893088193289183,0.38455247146779536,
 0.9797206251584641,0.8945817223888614,0.07780031290890532,
 0.41995104973843445,0.15998278311545167,0.8509526512779018,0.262652953573288,
 0.5391957286646114,0.7358069291302421,0.028327504754037003,
 0.5170355403418367,0.35652397856167517,0.12629779656965445,
 0.12626278989092832,0.15185717809544852,0.9751175420071405,
 0.17686382798077982,0.8654399324020654,0.3577156892308466,0.4057453369360261,
 0.17948182527257273,0.5385073870459542,0.9500918163717998,0.7758378519493644,
 0.3788631199281902,0.2617205702711055,0.9826652873826877,0.41765579011388787,
 0.9426703431387805,0.5591182958105585,0.8753007404190343,0.5326404891987031,
 0.3517707942353081,0.9607822632561202,0.21082312165462036,0.7294154859998423,
 0.427132713609893,0.7253984248903669,0.8243261286322832,0.24301288904002982,
 0.32875591887169653,0.7105872734763531,0.9782861921003042,0.692326895683087,
 0.9346802134488166,0.4087924322030323,0.5329991332943195,0.48805094043321684,
 0.184451115055014,0.4764415436844639,0.6154493096963368,0.8226095590225151,
 0.6099548919716867,0.6298064199646832,0.20851678237853255,0.9995013433196819,
 0.38897078682878117,0.7048627683073729,0.23711801979164315,
 0.3874184901038462,0.25837164650106326,0.42931424094691395,
 0.28682443393056134,0.13822542246808167,0.1593917889368096,
 0.9928463834164367,0.539019257213891,0.9377565449359999,0.24702563441605152,
 0.7692676594845358,0.4955530822979807,0.6346756744935833,0.14494169043600014,
 0.8830879941689168,0.6349989454819168,0.38565870592478746,0.2986033476150769,
 0.9310627254201501,0.6791247894845489,0.4399205129664182,0.6986500995453977,
 0.2153753335097982,0.07316414794337711,0.7486443180027273,0.5680660816941826,
 0.8375013433001011,0.6705562556210194,0.9403024834569657,0.5150062813796199,
 0.7723823513032182,0.6573131935828034,0.6350125752325886,0.9540991610551124,
 0.8546531550794748,0.3439739017366584,0.5184448445925526,0.5477260788084815,
 0.8018221078922547,0.21409758555166725,0.16738718131955554,
 0.24221804126173496,0.6356207558593695,0.8603665874133314,
 0.32731227247694195,0.04380780236877313,0.07016478340230636,
 0.988080479807253,0.03838008892782574,0.14300673782096585,
 0.27635077288628074,0.59644260156582,0.9908391723137142,0.9987068704515069,
 0.5072621737389127,0.556684959119293,0.9923628470749467,0.010524060655619727,
 0.5722120085306206,0.3337163632956104,0.6596116805435933,0.47553825055725296,
 0.8162843386929789,0.9573196598689038,0.07046425671933465,0.2917210352667874,
 0.5747851163042023,0.1780910833442485,0.501517823213106,0.4046233896346655,
 0.6330412840178381,0.5813491723431925,0.5653686847236941,0.3710364618889931,
 0.708402961173186,0.9573136217532958,0.31330888365804554,0.8950204860449871,
 0.9196182152495169,0.5150014508362597,0.34527374960952906,0.526367379756743,
 0.18858267848849541,0.6471670911804479,0.6094679361821778,0.6174304170677838,
 0.4736546858383901,0.7087738435614757,0.07873694321918312,
 0.024872095716263032,0.10160456106921445,0.30962916618736214,
 0.7945439890430677,0.9979740984380902,0.9290688296766797,0.9966630722601163,
 0.5530493030848004,0.1592441635576599,0.044753773666484586,
 0.3736903957483445,0.6674056293453448,0.12692068214964858,0.9600796165748613,
 0.14805690711637975,0.8220259655220362,0.454714784701466,0.6479119371878849,
 0.023872615338089487,0.2571002442143385,0.36794855157953965,
 0.0018387862003341748,0.29285314278667063,0.12419913456815324,
 0.9556244053951187,0.33243550181824233,0.7319374383223822,0.1322014522095053,
 0.11686057425480045,0.7673152547530335,0.916691760883052,0.16301496269216487,
 0.82624872202498,0.16078774908507532,0.7296822592026104,0.8238225058071627,
 0.45823831018723804,0.16685150289546058,0.9669196722641467,
 0.23734418891860498,0.8100620871358393,0.8765705784513467,0.2742846720517649,
 0.8543585040958476,0.5655194937604433,0.5449958305001884,0.04436667218836621,
 0.8256260240563573,0.40889743725523064,0.31958265584520884,
 0.7786135517112144,0.7071827884498094,0.39486270455704875,0.8507326311770594,
 0.8160199052626473,0.7874025350604873,0.6109996202171943,0.8860043140546306,
 0.698090746358469,0.5306263315701357,0.5326932257789889,0.9069810373454814,
 0.7524777330102213,0.30128557996897665,0.821509210562416,0.815564134469469,
 0.7495432138461493,0.10364599566410693,0.6897586378815256,0.9739656154224763,
 0.7882876533980225,0.09425041325277672,0.5448737647921164,0.5276913304438493,
 0.8466934660860128,0.29047349530251354,0.5850942551283834,
 0.47541052996961797,0.03139589813776045,0.5290179992967722,
 0.19189253277240137,0.8323814386304393,0.3205425442492116,
 0.15877730308695615,0.5591970107966404,0.21146815824746878,0.153779799586636,
 0.8179715593461817,0.3548329552096686,0.8063655643222947,0.2754690793650394,
 0.43286024177889115,0.3646963209017712,0.16825711475007588,
 0.9092061208605482,0.0120964297033479,0.35700649793975414,
 0.006385888296367326,0.8376225515822098,0.7614749343492873,0.876963487284913,
 0.20248196006433739,0.7939551121468778,0.02106441119018765,
 0.31938705106286014,0.29038078632476294,0.6894280399530394,
 0.16266081222081719,0.3576293114767761,0.08147714201220613,
 0.12307905024087096,0.7460124828440301,0.3460136691184409,
 0.14331378493219726,0.3809968171320782,0.9390549413875203,
 0.15223189545135107,0.5003708172802079,0.5353234973782321,
 0.17279687333196303,0.11350281866297529,0.3811882854559676,
 0.3034867399319273,0.11420520441954429,0.3895609826690801,0.7331360860996746,
 0.23669166036481692,0.5023490354626521,0.3337124691025134,0.3187569342501977,
 0.12310539013368116,0.7329310676739113,0.2589944146658738,
 0.16386794661671067,0.7392656120915045,0.9493936997140473,0.684775633479443,
 0.04702004937871862,0.2689419241161256,0.9432281051805397,0.8140231893583639,
 0.02681220454999278,0.7178587440012986,0.29403553545760774,
 0.8820033170630399,0.12122196853009737,0.16049949887740778,
 0.4709276887153143,0.3255069524059504,0.544683049601488,0.8417667429331664,
 0.6136140314574754,0.38860971587727144,0.984280676280634,0.7964589225008186,
 0.12204910766960775,0.18243232466184922,0.49380004959444446,
 0.3077004349015059,0.54336382646274,0.05313842985092021,0.8086739439218165,
 0.14121348393805178,0.8250989824547457,0.23115579036663447,
 0.8889504928799514,0.5048869798022527,0.08344253707785898,0.7844961005511494,
 0.04370664613762887,0.8785124519839538,0.595008979782534,0.411027725084269,
 0.4779606557829229,0.2014206980019052,0.2136528023991976,0.743795717031185,
 0.5988531243290279,0.251098505070135,0.5040652792504628,0.3246997211037146,
 0.28339785475376866,0.4461079020949542,0.7193715505311117,0.712009103667695,
 0.7456084318855905,0.7571638495107885,0.4439898783316256,0.3995694179696506,
 0.032059739137473064,0.386289982606707,0.2295661990596436,0.3925297823266477,
 0.6132926338614555,0.6806388672643451,0.35431895078630204,0.3906308369715279,
 0.5278218910502659,0.6971206925006177,0.0923326217119721,0.15063540906565753,
 0.033711241530456015,0.9187953835755376,0.004694185887351376,
 0.6489863417813438,0.48297114433494315,0.6610455624467731,0.295989787843272,
 0.2161234845323654,0.14824957574822273,0.9876600831707069,0.6396458657077793,
 0.9444428408666079,0.5362330219876342,0.008908123562229031,
 0.6673118768117741,0.15314515664023942,0.03550693973549701,
 0.7198870794767923,0.43629800416988507,0.4216329749588479,0.8301412357705534,
 0.9385800792669541,0.289918690329973,0.6460704196820009,0.7262581631398444,
 0.10672565558795377,0.2896234335860518,0.32445372567562303,
 0.5696639739503523,0.09028836967519083,0.748819549732357,0.5870232982906456,
 0.5519559610623068,0.6150754575437277,0.7282759210606535,0.6130103754680603,
 0.207618262158483,0.2031740204891339,0.7324235428825168,0.5912015779752888,
 0.2978805708222224,0.788871200811591,0.6900011676591264,0.25816146671764784,
 0.5122424244064984,0.50138323009487,0.02069092876944334,0.272421104479454,
 0.7292546809406811,0.010491702156648408,0.4394754297982747,
 0.5023266518062143,0.7822337501096563,0.9835756511091557,0.853029518685028,
 0.6782816004106879,0.34991684576093196,0.3906995482631128,0.7253631340923921,
 0.452714683880178,0.7749375734801801,0.8850380767869257,0.3434607503480709,
 0.5326839125345868,0.6131621539459973,0.6115289204661205,0.426912377033714,
 0.8179592770331986,0.12659618578640586,0.12049806315512834,
 0.7909437652617615,0.06794960461654853,0.26307104376858126,
 0.1465766318041959,0.3862775946594761,0.692104812828321,0.996346710251661,
 0.5029890236004491,0.3343623858652581,0.8219443792315866,0.8866826786922515,
 0.8086288340435119,0.009135074719105574,0.18143343194427264,
 0.9252753466314343,0.07056926342276115,0.445056386661298,0.7143729840965342,
 0.6758919771762149,0.7056511147608142,0.11129975893446753,
 0.060647868950408745,0.824810096515345,0.2737562965462942,
 0.027131295556786617,0.1095304398846142,0.6057579400979224,0.503536284406247,
 0.001779297383079581,0.4577631161977712,0.09141216912504702,
 0.11638328766364125,0.6076005506541379,0.7809374972842367,
 0.06561536832888315,0.011358924229538658,0.12005445602417497,
 0.0783680771556069,0.4226758446210872,0.2787002161833587,0.9763414558005559,
 0.9285975062937842,0.7026901415065966,0.6823035415543909,0.10582158025590982,
 0.45714309930376795,0.43872724209261527,0.4283283742546412,
 0.6951225972947983,0.6690930685731461,0.6248336694738599,0.18215623514515167,
 0.7574872139922149,0.47528577283639406,0.7060468943637439,
 0.45576659657842744,0.6785952371250362,0.1446194942017961,0.2322734139728685,
 0.48840246874617677,0.22562156189958693,0.5941960532133655,
 0.4182118377159324,0.9712109509204966,0.3522560477944934,0.6979417753784318,
 0.6551622376887353,0.22041850267973906,0.6715051885107599,
 0.14311817504411983,0.6868647875383576,0.8460725343839446,
 0.24184874528908196,0.8012232266423062,0.623848543917432,0.03557756787853916,
 0.11622458904916211,0.5503823347868133,0.2834311395194906,
 0.16970224179101356,0.38421509950846233,0.46093765106014695,
 0.10165682529410769,0.3776269196024601,0.9315800051927525,
 0.20967820624572076,0.4217758389690518,0.8870282261503348,0.5107824125983463,
 0.3102167060036526,0.8143376828385918,0.6976093108479032,0.7230055016376921,
 0.4823525035016559,0.3921578498609103,0.365794278363228,0.45007948878352744,
 0.26953051724911636,0.707124829198087,0.7184620465308033,0.2200826879941975,
 0.055978290650491536,0.26512911524621807,0.7538464919011315,
 0.3337264459707887,0.8930907559402699,0.4216238133726947,0.4379405094409474,
 0.03896193985608987,0.24928115313224364,0.3544345983145333,
 0.7956221857687895,0.32647493447816134,0.13682730557755074,
 0.7480359015979046,0.8587793953257634,0.847658130364885,0.0839337756455214,
 0.1891514434859901,0.818911392607465,0.14672830635021383,0.9368134291974399,
 0.21656707867900504,0.16812248021171816,0.02584717855100116,
 0.25898593549426585,0.14343371546791062,0.1806878813496392,
 0.11250066868340891,0.9100335854615103,0.35154892727668097,
 0.41468272160064945,0.9538766199887403,0.31610032567236246,
 0.2097591508681952,0.1625715797587557,0.7889514307783806,0.11445773395086667,
 0.4685113238824178,0.8983257322863211,0.10310111107724851,0.8998678389891226,
 0.9137865521454912,0.5805590991184151,0.9196343456355885,0.4350759621310535,
 0.8586810541196093,0.9150945936208271,0.8531640766353952,0.20365219063906081,
 0.7540429562595995,0.9802278305685193,0.792231378265009,0.382544123417789,
 0.35083475974856915,0.5677855914366906,0.2818963716146594,0.253068573127863,
 0.6793685413074589,0.152579710772907,0.4616364913295695,0.37430458049146365,
 0.051696579890315064,0.25230002117418104,0.7926707493886941,
 0.7649138928136627,0.016111884049497927,0.6093588598290807,
 0.09399423972476806,0.2895287371250851,0.7055033553821781,0.8063411039951729,
 0.20084828763292806,0.7317218459264861,0.19472181420133827,
 0.026417782264273826,0.3557233221199503,0.6414936582438504,
 0.4716210294340113,0.5657875107573263,0.14647310938849256,0.9519366735425013,
 0.25484355109614976,0.9880587096846847,0.09775796066820408,
 0.24394122713193256,0.8869994055661136,0.25035335794794833,
 0.6151235730138988,0.20159437907356914,0.011618341613513716,
 0.8471862081304984,0.0038988191362867575,0.7794982116179646,
 0.8706229848814367,0.5093782367788444,0.9195827698613759,0.854173801235479,
 0.03542176289469212,0.4829375962152629,0.12966431145834367,0.866679013221519,
 0.8862109481359863,0.37038793255039293,0.3993957697585637,0.1323211110812874,
 0.2716711143086795,0.6187791754714268,0.648104206408015,0.019998060525945327,
 0.28689294653302744,0.20125202089437655,0.810239863700905,0.7252426343968795,
 0.18966189154078206,0.3939983170809286,0.12059342674523066,
 0.05437807232034442,0.15478552166354498,0.028989338460572278,
 0.8005659899910067,0.27191587615371526,0.23073533290813508,
 0.8769464266482496,0.5733464945899279,0.14923142131375533,0.7349983491137242,
 0.9569581176175909,0.38288781490402757,0.07603342586821082,
 0.5764963943722162,0.5950050336155115,0.9856085806744853,0.985416322644002,
 0.5928871077598362,0.6730525398480905,0.6289788102519117,0.13505314084009412,
 1.916879085717138e-4,0.3580966996404904,0.4747600683664318,
 0.5308675590641924,0.9685820172503393,0.11435351289569806,0.7977709055702125,
 0.6011470692517858,0.3773149306224818,0.3640989424131985,0.08615634154359642,
 0.18513480397891802,0.6480334189173403,0.3912685483434156,
 0.41628770227381473,0.43240189009860497,0.29179681551640013,
 0.03624915327383316,0.5041106109625406,0.4095542434923656,0.6805602142572247,
 0.769358569725709,0.27949549930120277,0.42592845279434866,
 0.04502489178984792,0.0826010070931964,0.9918542972273338,
 0.47182346089796157,0.3465912085829883,0.16123504257198618,
 0.9353260448936725,0.6677316827530927,0.5419588689963976,0.5172078589797346,
 0.12382791250245906,0.30531847275223856,0.30423328678847406,
 0.33439032286948234,0.26387276957818706,0.7154174326453695,
 0.12153536346855365,0.09392166810229852,0.8041878536138917,
 0.06737507811200238,0.1671547039936817,0.6845713379024232,
 0.04309720385334381,0.8118133696092834,0.4932588726714029,0.349704513197305,
 0.45725307103900936,0.6515358888493892,0.19705175000517328,
 0.35017558506131263,0.01536109272340902,0.04336725251691853,
 0.945715200638005,0.6027985405681796,0.6615892667597986,0.6468534899945446,
 0.26798664434262875,0.3783465456624544,0.9718665730478165,0.8579755944211358,
 0.9403303875530415,0.49281658022750285,0.532988458681166,
 0.057998313889557784,0.42336014481971185,0.21027460575249524,
 0.41450102313476056,0.7949647279238578,0.9034142502783296,
 0.08566788587795493,0.8604735681415756,0.41181668161393825,
 0.06634861214691856,0.6398975348264618,0.8559396049568044,0.5277888126014594,
 0.7477987211954572,0.9439056270727922,0.9540698148268466,0.7345540822264485,
 0.4013666785226283,0.5504296309687481,0.036436783826997776,
 0.24457407344429427,0.22768801056800847,0.5745221011317103,
 0.15718334376600118,0.13733539207218193,0.7841011923242227,
 0.8502446871864715,0.07086879827020154,0.4536419791539803,0.8713129636663459,
 0.11786452951995796,0.47947793540821126,0.8091492499251198,
 0.2593054233064096,0.5197573978171512,0.7372836123880632,0.5022629081767027,
 0.21494142458265086,0.8858675872986472,0.8738403008098398,0.5402779421047648,
 0.3942824261960579,0.6803474305971889,0.856239297767756,0.5611823851219726,
 0.17931203116688377,0.43229838734206705,0.48933035712536255,
 0.07413815153955383,0.8026916559387938,0.7399171197589172,0.9184686776191052,
 0.7220633562453314,0.7976723995725504,0.4193538095460432,0.46132363562974743,
 0.42559917697915295,0.7756139448927125,0.39175027211021496,
 0.44260559274594025,0.0416307431913755,0.6812825504962553,0.1893642064469847,
 0.2467005703438686,0.7664304724914746,0.6636094442742815,0.6785833971313395,
 0.8770752395690564,0.03816709785163519,0.6217415852118957,
 0.05147384145360445,0.9235978743428719,0.7350367000713363,0.9082478843682978,
 0.9380628647442093,0.9560623440759061,0.36453339320311917,0.9826040010945826,
 0.4447001805568518,0.01595821638586281,0.5121378567414654,
 0.13794338230740943,0.24913605238871273,0.874966913363846,
 0.16446150887035194,0.911293352914178,0.17716839431433695,0.6637747082964981,
 0.1665618331507429,0.8643594168444413,0.9265620968851721,0.29831198048230634,
 0.8389210822932035,0.9910742143630771,0.01019534129339883,
 0.40149997908136714,0.3172553868782054,0.3059350378622565,0.7628750837998817,
 0.8189697994126541,0.4357845050554696,0.820241027018559,0.655680163339276,
 0.3301646252202016,0.26194705930637807,0.41854422187789886,
 0.14970109355670969,0.6973165259511265,0.982056217019505,0.543905682377855,
 0.7453012592873741,0.19214816962769254,0.697939320410713,0.3129072882932793,
 0.4104724401131057,0.5604512871065259,0.28556442268278304,
 9.096288142756981e-4,0.09656280005332651,0.8071992478976351,
 0.5259990629610087,0.2943236803937158,0.32548639611836916,0.8720276461454189,
 0.8163919478320321,0.6494265584752289,0.45900429578672297,0.590106865855764,
 0.8231567180860221,0.28218564645618194,0.23334318621314798,
 0.04109006399552928,0.5023410803978806,0.03848225321363152,
 0.054772693367664216,0.18438122506241794,0.03211364655278803,
 0.3319167602456531,0.8659375848033158,0.34497071566821147,0.7776425866912648,
 0.4894217700701303,0.4508013791397172,0.7812583358861594,0.15149676164310266,
 0.8550389357512924,0.6354491262877001,0.4537022062784909,0.9738092048473622,
 0.3802473625010593,0.663549146429099,0.6963283527014559,0.5116710867087773,
 0.1730859399779494,0.2741568386291636,0.20484927183434243,0.7794458631674133,
 0.34770445876953426,0.8045209340439003,0.3246064042626543,
 0.11597497161041737,0.8235453726709324,0.7141849675914168,
 0.41446646179639945,0.15525992116164145,0.979780840358131,0.3321305800124077,
 0.6599905784005111,0.12584688284356316,0.9637658223069796,0.178674829527792,
 0.13803046986909728,0.7227439292869776,0.5733568329422177,0.8575571074972336,
 0.005856295379895959,0.8849684774882719,0.290446552675635,
 0.30722966050453926,0.13273861956838084,0.6097127392221833,
 0.2435655402061192,0.21624377406750372,0.02149192766031316,
 0.06542996007604396,0.5937909317357853,0.19213292768927814,
 0.9340372163777922,0.7215637628046377,0.11587118576004496,
 0.14073296510315647,0.9642445048136324,0.08999507466215362,
 0.6368799362523027,0.11812384621166827,0.49807089393127024,
 0.30015572849754013,0.36304426841412873,0.015604995549136524,
 0.9455327344252495,0.35211296730856534,0.05704610386560571,
 0.5412915451191052,0.08306276265297563,0.1918010507412965,0.5484095284477479,
 0.03277983888931235,0.4898572341762444,0.9011830695953871,0.7066829526169295,
 0.37391624680143254,0.1611195975222347,0.698394292886095,0.46638677218723923,
 0.6409545754862962,0.35598279800929833,0.642972977738292,0.9457527858432482,
 0.619994372329181,0.5206587844974566,0.34146024887641546,0.9529987405581335,
 0.15707402734933495,0.20566982787137145,0.9859818212115028,
 0.31718058742046307,0.5110773606454895,0.39585301322722466,
 0.7707473511016306,0.6623569484535797,0.3715683959356193,0.37450019271738144,
 0.8936771977806336,0.843481387274451,0.8142624934469926,0.046623060931308635,
 0.5480064318558431,0.7762414025792741,0.06693027208596192,
 0.43501371694831037,0.3425241337074989,0.9944763347270629,0.1431094662491479,
 0.6894334257877865,0.10169178141843815,0.07838374615404431,
 0.9915971529997037,0.18440257591403542,0.5348730305699483,
 0.31404945633861925,0.2642755859275274,0.13026701846447952,
 0.14880192687479032,0.09081488726593001,0.8940642996121191,
 0.4725881008654551,0.678658196352293,0.9797500316466473,0.3873626792089174,
 0.4996772716953216,0.9253009009126286,0.9769202907889342,0.8439893845255259,
 0.6508172485520252,0.4259194604713503,0.21718531738833802,0.492993413455891,
 0.25219787000735805,0.8903807845910172,0.838430028374729,0.12394976272642988,
 0.11013479456753517,0.43626462491953655,0.020190305864979385,
 0.7541048863512866,0.5516135519573703,0.24680313168581725,0.5340283410324024,
 0.7885778401832494,0.5754283099554911,0.9878505880715049,0.1766803475701767,
 0.46354909276392164,0.5582485376680554,0.010407481585403944,
 0.3204839990592757,0.9462909917544542,0.5090491831126723,0.20885017377437043,
 0.13184557079912373,0.8576657991273846,0.028334345012810735,
 0.6891473379941512,0.31177720737225545,0.7833473472918647,0.1336559174411397,
 0.9100733713659263,0.3119524145448622,0.3361891262378849,0.19566208497855764,
 0.029613142952438998,0.05213733280782107,0.32295281967792233,
 0.3069979349962981,0.698855685028866,0.7643668869224057,0.34140165888086305,
 0.6077196222532905,0.921101646244798,0.3806164377933122,0.9165419972188793,
 0.22015632143842412,0.41446509136282395,0.6923868795070853,
 0.9860381104067011,0.47279524695163566,0.2776961661793166,
 0.19592689431197763,0.7063582640880842,0.8327008304965559,0.9044659841502585,
 0.6594841596843535,0.2058379378970685,0.43084388796649187,
 0.40538555442060376,0.22035240432674186,0.1214810420818414,
 0.7186672141362448,0.06324989806995762,0.3000413628876112,0.9397981947155665,
 0.6069947259298413,0.3436676247494387,0.08557734466189248,0.2587402371381774,
 0.2186036470315389,0.009954583405424433,0.9534862406520433,
 0.8487486075453858,0.6535006663136592,0.11551541109872743,0.3997927382831429,
 0.8828582447430688,0.24241842434608873,0.8867388886114577,0.2959787752177859,
 0.020120565855148098,0.9266244470198505,0.17244615636438354,
 0.6359997470942264,0.358380657859685,0.988179459708631,0.825454644785879,
 0.2501673906447328,0.4998683172587888,0.08424732505858934,0.7078361958874506,
 0.6969499290943545,0.6270111823102397,0.31900175813968745,0.6632761694599362,
 0.4838968055955958,0.057309125883934,0.757497289288434,0.7064636831914517,
 0.795715975152874,0.8766281134638513,0.6528138812138242,0.2058771799093061,
 0.8918667689420374,0.5071458633123042,0.4588486591884874,0.2691786631415348,
 0.9210642769948175,0.4307711859995671,0.4640932159627368,0.47707077700074785,
 0.7617676952532503,0.3207220630262926,0.49915629916283266,0.5456135797810451,
 0.35582638825912916,0.847112115243502,0.8754115803956137,0.2725575568866744,
 0.5238529808434406,0.3281181801690849,0.11145298438465634,0.7637455933038925,
 0.3489309172978302,0.8456162485037632,0.3217288269166434,0.49632158779827384,
 0.9565629826323547,0.6460144035249914,0.7167270560674496,0.9716735546721584,
 0.9550467475589471,0.009720901234930412,0.7358176083613518,
 0.017655001742570176,0.8963513242286405,0.04331397342413423,
 0.6880468571860305,0.08389050098045836,0.4357228700417043,0.9812219475000107,
 0.5291765691473671,0.8689316400219668,0.7142145837664472,0.7150161351805868,
 0.28775849168325474,0.7153745950777894,0.48474259670655717,
 0.14050498256765775,0.27365805766321205,0.8441236200997068,
 0.4587025429185023,0.5883094538122433,0.36401437369375866,0.5693529620450006,
 0.28175235761859607,0.27691524506972287,0.42403238838472423,
 0.8370763446741704,0.20046088844860743,0.8929338677011462,0.5163668222766122,
 0.7380778611853474,0.016245050015636142,0.605733186919567,0.9533195173563671,
 0.8532422030512341,0.43195196806125413,0.6872931741863284,
 0.04720422172987648,0.11081258834394991,0.22988228522804055,
 0.34838708051673184,0.8654928647722414,0.8304726787461364,
 0.28813216519409435,0.9793417079837115,0.8804931315841701,0.9957005215897199,
 0.16911070707973797,0.47151190152094224,0.903629678410548,0.6646838999048655,
 0.19614863448935826,0.24613986616776862,0.024991172134312456,
 0.8061867175429753,0.7902500178698647,0.550316389115041,0.4892350697394008,
 0.768750722973951,0.3801795519736795,0.7306266902782604,0.5772428805630823,
 0.10384646798569097,0.5992545633195194,0.6052767563431949,0.8908859010952602,
 0.09636734413799286,0.06804408648760807,0.8407174662698909,
 0.5103192843336958,0.11373319744049093,0.8485392632208746,0.3266140983257052,
 0.49237548130743236,0.6121952366752422,0.16825266287780594,
 0.49401365348165505,0.023236511136258353,0.6974795229351789,
 0.5251556690200044,0.10275784164456458,0.2779816988367382,
 0.46296683872460465,0.07901330374088111,0.8872348006058518,
 0.4870781114703444,0.9378383272015071,0.34005893109661434,
 0.07178179684475461,0.1974585641128629,0.5799692180928975,0.6273782025522218,
 0.28977031479925275,0.2690877566566151,0.024945933684263344,
 0.12966352429525485,0.6897773273100446,0.4518333910723318,
 0.13401228613094818,0.19002146062226133,0.2396593595323927,0.719425310513261,
 0.6938758586286623,0.32476587045454297,0.3361872317137412,0.5388123672863462,
 0.2899332752047882,0.8377108198460821,0.4420863085424991,0.2520830532248488,
 0.8862146405603426,0.5733859237638533,0.11960224452258028,0.3505579880652254,
 0.6782423683025621,0.8279668141769024,0.7774750880614789,0.6746993497187884,
 0.46444941248367233,0.5703972160227289,0.5502289452998166,0.5189781177658611,
 0.28750872999617905,0.20077182089356738,0.023546195660011815,
 0.9284443130685087,0.5944642965181047,0.21017917514590256,0.7532270845213971,
 0.23514473933860858,0.12868699464917033,0.1873823835324563,0.936860895758223,
 0.6761194354185471,0.33930203607397835,0.11213461332394481,
 0.9914857002025014,0.8972331067542416,0.6535832244863519,0.03583846590113138,
 0.17656052348848916,0.5586865997528303,0.5042038384938591,0.8571223797062826,
 0.6372088118333387,0.9805941053846409,0.39209981593205934,
 0.10265301032847907,0.6598248098057423,0.8390533926308865,0.8280778895466344,
 0.4126917423606916,0.07487647839060774,0.036216079469339246,
 0.33620160651825604,0.706965259435982,0.788630236160147,0.7476806208500661,
 0.12777367110489812,0.09859378984783285,0.03418997601919682,
 0.14043449170261613,0.024549324278280382,0.1491837078575987,
 0.8840250486777443,0.10908586885391425,0.3586523752536125,
 0.14970461952042624,0.2839862972088488,0.8522358374715,0.08075829062854334,
 0.39386535736177497,0.8804594721464063,0.17011811987334757,
 0.10500047079144892,0.14580686195376324,0.4000794975922941,0.739331466537712,
 0.975468793388501,0.05845377139461594,0.3867365334466889,0.5512093769134656,
 0.04311835061991354,0.8843529344415195,0.47797421779389726,0.779060463988086,
 0.8804940790717339,0.9585450565665943,0.26929789065964505,0.9088600347569575,
 0.92591199750354,0.18360661925394506,0.8449181280295975,0.859037029956541,
 0.813153571134632,0.7076743962036856,0.926348843027292,0.14412048219397644,
 0.25178100658746005,0.688683749429561,0.34296133615523594,
 0.05203186765769918,0.3412727731530527,0.9817639157571869,
 0.03476192527645572,0.4667319931399313,0.08610248920431851,
 0.43059075176183936,0.6719728504067336,0.7690553222184394,0.9045350883572234,
 0.335859332883124,0.5560647572130187,0.3773191430933581,0.517338409439501,
 0.26582881098161304,0.13160648806875253,0.14703608701740023,
 0.05666677601022474,0.549728766158218,0.5441072660934109,0.01396737942443027,
 0.11406936182806748,0.3836099322309987,0.5529742941763174,0.8502943426084856,
 0.24986177442405122,0.9513059124455235,0.7400374570694483,
 0.027125385424952797,0.9177106481120029,0.07317715903115651,
 0.1675214383593323,0.9683686344903777,0.6457365913456435,0.40435730852898444,
 0.9576402938182941,0.11928540401457277,0.43677366683490226,
 0.05665493322576809,0.08012772724449313,0.9118449784506226,
 0.2047441073836136,0.17298489068247624,0.3943259714549262,0.2621084153699892,
 0.09430068859850382,0.9843540249175069,0.6175580941220875,0.5047435751045517,
 0.14920659570682882,0.5119131768111757,0.15906959078163374,
 0.2881739961606463,0.07567531991842713,0.17759135629375256,
 0.13567911498101604,0.39697781667539056,0.04064592324951122,
 0.7511035282198737,0.5816819192516729,0.07890224299059234,0.6620853464710625,
 0.40494763100242226,0.3656257764105666,0.1097697286153102,0.7026988020100328,
 0.21571804795677596,0.30999532033646715,0.8572259319283343,
 0.9917243338676149,0.6646132326107264,0.43529444056247435,
 0.42966961694777095,0.07359733934258073,0.26726260519650513,
 0.8222051492787621,0.5316558105096468,0.4902503758250494,0.7477508335244701,
 0.8156820889247275,0.2547639705077305,0.017823870969940425,
 0.43620794675403074,0.6893443042445379,0.9702405197970156,0.4925567259229997,
 0.82488590754611,0.5365356268671839,0.6330563153100759,0.8974846522093858,
 0.12378235098903934,0.5011974393606176,0.5197520722866207,0.742821563422865,
 0.37163665944490676,0.5240234591053925,0.9177417343121906,0.7996315974779,
 0.01539120515621395,0.5999009879177473,0.6722524751623781,0.541275267295119,
 0.9816117191925051,0.5771674061857389,0.5605506861932327,0.17763026287807715,
 0.3677871325556197,0.42176267871452333,0.10575034397687111,
 0.8152967406170174,0.302576910619772,0.13132591417360662,0.7618269288952173,
 0.6792080725452911,0.636451190822217,0.9164215253209315,0.33464238485127784,
 0.42984056310333085,0.6351696474829026,0.8573133687088552,0.5076500503401361,
 0.562564435511733,0.7378584699432745,0.2714293774770715,0.9927790565220878,
 0.49485262168195465,0.6275773549926557,0.4749757823976092,
 0.40364759770158076,0.7471764534069882,0.4756049906565989,
 0.28818475725708725,0.03503385159496408,0.09900922017637925,
 0.6887710125963639,0.9219092441293499,0.15855008349980138,0.3696159715608819,
 0.47666857768841375,0.7730638875322432,0.685356859260451,0.5025388285905863,
 0.07829546421626687,0.25171142623569087,0.16618802595722604,
 0.13435182532109546,0.777781351753226,0.6707644267947673,0.08589709154517544,
 0.597929057596796,0.45413037741660456,0.7131008002233417,0.6472482478146626,
 0.38984622815061076,0.7641218339530933,0.8923183215930756,0.6397013877255923,
 0.29714070251750746,0.28149280043422176,0.3163278943998722,
 0.9001224554414055,0.7897770416107791,0.4026969172561241,0.9363659474158265,
 0.08886069479162073,0.7081010999302845,0.8526140497919584,0.8036463442403259,
 0.445809647801931,0.321253506681762,0.3396089969561672,0.7326893014578075,
 0.8109590809676314,0.5849431684099765,0.8562035777536933,0.3002917844773846,
 0.2002718893841149,0.32207065383362743,0.5722283683709705,0.934159960554972,
 0.2729354037177256,0.47972262228618123,0.3127236965022617,0.1162152158723816,
 0.8277327019392079,0.15956379910265062,0.4676602704883135,
 0.04529377672398416,0.03595624566636091,0.8675496999193317,
 0.1290142451356776,0.4256895442874489,0.7334392608018406,0.9817878340049164,
 0.5323163748050236,0.4343460219070595,0.3013038622358162,0.0959786976483239,
 0.22300046888575298,0.7917593754087469,0.12551097995134597,
 0.8557957197589018,0.35735736530353435,0.5199639484822582,0.7202849785098793,
 0.6181334065703676,0.5850445340434036,0.057119924898512586,
 0.5893489986600847,0.9304742796282603,0.36898906068920234,0.789042269123344,
 0.40365834225542574,0.5416276671107951,0.9840818854030643,
 0.33545646618374014,0.9691824796241357,0.5793928368122072,
 0.24992632966344697,0.9979447539541204,0.395907377404773,0.6991558480194322,
 0.31445602673354034,0.6079591157517819,0.6147286018817029,0.629430711475847,
 0.3598824467583501,0.4895620680090983,0.46388505305502004,0.5173892458704936,
 0.7759027520334593,0.7819088999434007,0.8621709581411736,0.6003888976080336,
 0.4198989710049028,0.16886073495784537,0.6839585823172291,
 0.10668293125872563,0.38578947542892783,0.2020715819551312,0.789119585977973,
 0.5512965921899285,0.5555999102371936,0.11733740774772661,0.434025953856108,
 0.8413729680954232,0.6872433123952625,0.6022288422804101,0.7166457913908426,
 0.5642940360196979,0.34427210369032635,0.5576066646208883,
 0.44766123743794806,0.5108446332839343,0.10885252110088661,
 0.1286303811153029,0.3013773900045512,0.8599066078491132,0.3892610668173675,
 0.554385876484614,0.9282096164660848,0.2235016074677998,0.2836115215438082,
 0.3571720909812801,0.29127919618929254,0.6898679668063704,
 0.26998704716141164,0.42545807761232424,0.45429245312103206,
 0.307763571529425,0.47629759147011974,0.864479036236784,0.2933443320334612,
 0.6930179032617572,0.573689455641033,0.8236352909181367,0.5611691325012053,
 0.28771110642904785,0.0949102113860556,0.69697579476937,0.6745174897430666,
 0.4775836468632242,0.35139046452164724,0.4391607518661713,0.6243200716771373,
 0.45257092947400623,0.9316164755835193,0.8192855433323405,0.9202593153539217,
 0.7477813664362521,0.9386478049323574,0.11964740277110386,
 0.36428719611078897,0.599106908561924,0.4317488223978613,0.6505755271456204,
 0.08818789683658779,0.7098137629004618,0.1129494795435928,0.8778682087045454,
 0.20817945458979958,0.7580887260281691,0.7281781524834984,
 0.19836923255025407,0.3304758196444373,0.8555919226902189,0.4475445738679005,
 0.70465350350554,0.14984002071027636,0.7216877579262662,0.38424740849374606,
 0.6757952528549803,0.6951246345566826,0.37208906802494446,
 0.30427874389499954,0.6070480714170904,0.875302962964804,0.8339541381356611,
 0.18359077536629653,0.7391739554365148,0.2522886168552174,0.9434378888445267,
 0.5401574073828792,0.12451664282207986,0.7840191497452857,0.357763859117499,
 0.9749452273904137,0.775146431708943,0.34577197554725436,0.3611752226373711,
 0.008624375252520466,0.6316007314716597,0.7291440709555306,
 0.4476691269663431,0.1533459734525605,0.27857988044600057,
 0.24958239593382625,0.8214556183592504,0.8246459003239813,0.6847146728120322,
 0.2715939966923727,0.3231004242262375,0.5539639178162122,0.4399607129691667,
 0.7657679734840046,0.8080321224910701,0.5102157760011308,0.833708181861068,
 0.5236412393560212,0.23971320760802162,0.5857201642231038,0.2360561511699073,
 0.1879432678353683,0.9534566130493878,0.48369192244885517,
 0.37320070687186624,0.8485285074374008,0.5700812587451228,0.2301120846559369,
 0.3060796040865966,0.7960679620893456,0.058939566917055264,
 0.28953813301287024,0.7315023207822837,0.2894732530625559,0.8421009274124989,
 0.3391167910634799,0.6451060282151165,0.8167739411295482,0.7265270809518638,
 0.004659332017975748,0.8320865744013813,0.39966258069654825,
 0.5227347264221714,0.2015525627172603,0.6568642260412535,0.4650062676974728,
 0.8940605188874847,0.2442948501404436,0.5074117532219343,0.3752473145699363,
 0.06725968818788897,0.06510038969322829,0.9363661433068619,
 0.9392750575556574,0.6691085414490503,0.5212928399851109,0.25366807971847183,
 0.7693534269053005,0.3650998376508615,0.26757232387925434,0.6940669839016603,
 0.4454660518823994,0.0734125004017997,0.8242270374832257,0.5153736679456387,
 0.48309970692014337,0.8896946120410388,0.7731610141873543,
 0.003846046814742432,0.11914811835766448,0.833781014018562,0.918676472700587,
 0.5504819191346674,0.3655941777873457,0.42702172741892097,0.5361243176397587,
 0.2626198459771314,0.00981646782160639,0.49273104468101814,
 0.4798672473455734,0.6494024971858906,0.852620967123848,0.9958958714551109,
 0.03365787890293881,0.6209198948175815,0.553677665167212,0.04319123240010436,
 0.7396335469980678,0.1338655997113356,0.8372096360380052,0.3262373695140761,
 0.49805119486257876,0.2313429576995255,0.3362580403299371,0.7433652356913428,
 0.9208601335944099,0.49080829696951944,0.9964536099759311,
 0.007878543118046921,0.7088565618585128,0.3781400343595239,
 0.14456971608451008,0.4933369282918658,0.350393948604366,0.7235860544183734,
 0.4998160394789187,0.40535529009754234,0.875330828689568,0.6299995971880108,
 0.35944184695289483,0.06390106879885271,0.7461874905168192,
 0.18749896276166422,0.6507748819488554,0.18841030962342975,
 0.4544770537985795,0.2047442607194232,0.4762168283505498,0.36379337762844555,
 0.018927745729431456,0.23237828318273634,0.4254731356473167,
 0.41734612146838757,0.014696575025168301,0.7023675606337837,
 0.2110022180526444,0.5573430560705274,0.6302217947846893,0.6280035103383311,
 0.5858646272081299,0.22817814733998698,0.5557618136408443,
 0.28262481124631245,0.5349584777282314,0.48815779548864546,
 0.2743980028774027,0.837146851161397,0.007954059798806923,0.4488310469375967,
 0.6310398351228017,0.42630492051715485,0.46977641991570585,
 0.07414326909677049,0.3474813996915759,0.03634198233798469,
 0.8429725304727821,0.2049912268500922,0.5066359837196484,0.4338773316336587,
 0.25860979754616165,0.702011009161787,0.3541093064996843,0.2401565184507748,
 0.3932855287171013,0.8825254453510921,0.6523912314530173,0.9309241301154305,
 0.049797983144508784,0.8438746233291626,0.9639588811349663,
 0.3701628157659852,0.19922747592989531,0.8006758567462022,
 0.35437458356821816,0.3253279552770103,0.4951811722472814,0.9152399801282298,
 0.007104605108168238,0.3495498434790827,0.8043160684191162,
 0.4212283932113525,0.9623792302568693,0.3354868958943438,0.5720484979114521,
 0.9158684331543819,0.5375470657389878,0.6815589919229819,0.29286355990795043,
 0.42097057493753787,0.830869771067785,0.5463549205514584,
 0.027415063557078567,0.927662028463439,0.09378772742344799,
 0.48404246210753443,0.10740053098962787,0.37680511734277244,
 0.357135540946392,0.13502752423155806,0.052185101701188175,
 0.7332835306694422,0.457069683520512,0.2660482855801932,0.9349008229185912,
 0.31644283207774637,0.44115934799237777,0.5008566047520295,
 0.22358151338149468,0.8175040181580808,0.9580760859950253,
 0.38814782168067563,0.7571931983134287,0.7821278202658872,0.3138329012794465,
 0.30737298792814405,0.5227267902230285,0.205681686382844,0.35829628257921164,
 0.7605544401025192,0.42433526172521163,0.6839126632643925,0.8949700763074411,
 0.7590662712699273,0.9278969798868469,0.6716204264231909,0.4862945405371,
 0.6306282264742831,0.3413485706668272,0.28973549023374134,0.7144972377510989,
 0.042573639626266035,0.22825052395812417,0.11727650391762556,
 0.894628764638086,0.8643714639681962,0.5016592147340073,0.41754895514505763,
 0.8750267448172319,0.6446157933062628,0.46704553332989607,0.8756248364846304,
 0.33658424403947707,0.8678028496063012,0.21725510775036938,
 0.029262477986533253,0.711120140950049,0.7831658216370961,
 0.17973469238890039,0.5485175189572901,0.3845319059299228,0.8240575246472476,
 0.9482504298725462,0.07377529961873175,0.8437244229949975,
 0.13970878364203376,0.48403674951334485,0.2568541275224241,
 0.0934079060958013,0.51200644817767,0.9459903172353689,0.9801456612972731,
 0.3440852129453108,0.6107114965870157,0.22458595759014977,0.7067267691301733,
 0.7223608436843232,0.1253167035717191,0.010030159955614848,
 0.03909671358308198,0.2938830908024471,0.03933916094482648,
 0.5168020612377539,0.7758578724704581,0.540798450379692,0.8189400674819448,
 0.0204475222267414,0.9676117760417725,0.6838480095505403,0.10502019425740783,
 0.918644272257187,0.316917856008187,0.9168403409279965,0.6928053049736133,
 0.26246741337705504,0.2031825416093035,0.6322073271479091,0.8017219360409846,
 0.5569466637487736,0.15996494533320604,0.24488075178462032,0.66273548855208,
 0.5508969121879471,0.9973288018369075,0.24927549263735416,
 0.21940183642103506,0.06716827859090213,0.9204090005291561,
 0.9538316238224159,0.010033972311848882,0.9833315231968704,0.930330354939938,
 0.43622832800459654,0.8155295901158527,0.767593683208922,0.25650146223382286,
 0.7486171621797734,0.29269343799172,0.02540071747667705,0.6029342921862587,
 0.3738152856315443,0.7716142880948998,0.13132225521002006,
 0.11859102428406365,0.30147628880722066,0.8385563614236866,
 0.26838318689278995,0.19985786709209286,0.5658507620834153,
 0.061390448317190316,0.1209922220927423,0.5444601916959669,
 0.6721625979626249,0.1790625545420944,0.11439139123584019,0.7856322460610583,
 0.6428902009527805,0.9443653528574464,0.6984719019488626,0.7547477925760795,
 0.23385909393525972,0.2174028465031258,0.47302503382237115,
 0.09300175750492756,0.5593206184295543,0.9695441901539265,
 0.033424162453861705,0.6887380331276596,0.483078742621889,
 0.19731112830768294,0.2944335302979273,0.5980981473935064,
 0.37377235529995867,0.6233957084656823,0.3346853309705856,0.706328398771501,
 0.9853141850630394,0.28567818487380103,0.5299239626888065,
 0.07698850539503554,0.3598279624307966,0.9938422042375288,
 0.44184266119735716,0.7310621235912116,0.485063225645157,0.9239668543407829,
 0.5315786381694961,0.3992137284765652,0.36098505387397795,0.9015904746328671,
 0.6719518635942221,0.5010421240846483,0.25328128256146953,0.0387622200800366,
 0.8314902009024724,0.6432259874245433,0.5543708515357937,0.3195523497757409,
 0.5650118604004418,0.0979509317691577,0.6584372382715281,0.27126606417164345,
 0.6584519680916007,0.03389406010488949,0.6291423327434202,
 0.29348729183105293,0.8830160735093444,0.8999195705411455,
 0.06049435226171285,0.5359368330469433,0.5327381417313408,
 0.02520892920325124,0.9681301578143593,0.8769799226073767,0.5339963243001576,
 0.44412014051071735,0.17874045705491703,0.5471647049680121,
 0.7069221251856403,0.8332181914964283,0.3124634394349083,0.20358318332748748,
 0.039073889454529276,0.38783046395913623,0.3424429789279274,
 0.22587269506147922,0.5298684092469401,0.14093470425096044,
 0.21044907195993023,0.7868903134883598,0.7356714193652926,0.3195631020234063,
 0.29499214274118535,0.7213280159638918,0.07122790120273048,
 0.49348408116080944,0.24100875821239254,0.8487620103762928,
 0.7021296467449949,0.032646746765950585,0.9555242811740544,
 0.06619582238350774,0.7956803892908764,0.8428327939014357,0.8724559787597097,
 0.46514817893530946,0.5838532160914026,0.9937949900771743,
 0.009588054155184156,0.3112282687450936,0.49143705466693244,
 0.799258442958123,0.826186554606792,0.6616146746158194,0.6124990214255781,
 0.05332542879095181,0.6309040070763012,0.12850432440205006,
 0.4304281978618153,0.5511745156562153,0.0888195228893296,0.22100212747909598,
 0.3079240139031111,0.046639557872279025,0.07513516878438387,
 0.05402359598166484,0.46739361574480065,0.4417976509904493,
 0.30861355207976515,0.6887090415598082,0.938903426008233,0.49406643835190367,
 0.5051788015190473,0.6014489789803558,0.9652782190036748,0.48392251331995495,
 0.781155310111209,0.7634115389507714,0.0923065922643107,0.8564625904983204,
 0.19268513212252802,0.4943911144864622,0.6074064944903518,0.6029999597330763,
 0.2209085610485927,0.048296507192215565,0.6079343972893537,
 0.4957155676257603,0.8609851960366062,0.16380267915501157,0.7306846378645372,
 0.5823085603426404,0.8122389925249871,0.7016417784908963,0.21316693788360264,
 0.8653482548713732,0.1630858546468229,0.7193703547050182,0.29444581515199475,
 0.6007370796582768,0.5185883469526142,0.47590371532664655,
 0.11825779458809627,0.3039115237781985,0.2526024829320278,0.2866734952375125,
 0.4312303102665759,0.8374036839858215,0.2513879139741506,0.34784791239550883,
 0.9386393617961255,0.11917183186551783,0.8687265477308237,0.5120681363023158,
 0.34694705865105224,0.6112237471414944,0.5010527788232557,
 0.47239907691187255,0.05713650587776875,0.2856989685875071,
 0.10659153741730165,0.9390230371863522,0.09906473422214335,
 0.21993139337061773,0.12506818826074007,0.26475882022315933,
 0.21898775692257877,0.5714322599835445,0.7370754217832656,0.3851498102316875,
 0.32133480860437036,0.4838212900071889,0.6987783800490526,0.875729674786166,
 0.6406051304745767,0.6824796516443723,0.22185658214973114,0.7758161624091124,
 0.5096381351641832,0.43640706421174313,0.7684155697714959,0.746228081328125,
 0.9718972182142905,0.8485184092027461,0.25476697359559053,0.3837466126114505,
 0.5495891093641321,0.174307513717737,0.23694321626488446,0.2106600352231297,
 0.1707895957440847,0.9373310547040208,0.46377840697121897,0.8277144665423675,
 0.22209128459061434,0.320447527533787,0.7964695951216477,0.2666587948509529,
 0.16256890490661846,0.4605929235218986,0.01883043862728251,
 0.3691686634623509,0.2983117416082054,0.7355282646110147,0.355134591018333,
 0.32416495285252367,0.4567945624195683,0.06424264423854786,
 0.5302485478339571,0.4321080888373664,0.9302890754459944,0.04907266911307118,
 0.10516710158465237,0.39548287205567634,0.9730538269408521,
 0.6865709151811594,0.004307933289028898,0.9872929806878299,
 0.5937356602315661,0.28359945602481773,0.28263012670195997,
 0.5738217804747777,0.3344532178365368,0.7263547135225588,0.49389740476799293,
 0.6809275952877777,0.1478301785195001,0.7456029255642512,0.09492391647856313,
 0.8109443034171916,0.23217218902194692,0.28939950339203757,
 0.47608823701263125,0.22649421280648396,0.543501819464682,0.4434861550880247,
 0.2520108070939652,0.364041369215652,0.4190670412229055,0.6731042243144963,
 0.12926099346040665,0.25537000826516554,0.8814441989466538,
 0.5423876283667026,0.14150810941191938,0.6594348136014389,
 0.41419830420195947,0.9063546770012416,0.9235334595698589,0.7208653273200218,
 0.8480221047711813,0.5311052138113085,0.27094866257480854,0.0433153409606466,
 0.5072210836075997,0.2210801876443087,0.3772360453738347,0.83741722882294,
 0.1438643898505899,0.030874177914360867,0.9469882005005765,
 0.29757291407161546,0.680998287199752,0.8687584376188229,0.3798004886279933,
 0.5293225691989655,0.9961953138906734,0.14791403592875008,0.6917699944147857,
 0.9894332176474672,0.9372923995492407,0.12063465830942932,0.742541655256916,
 0.2719228164278912,0.4072415173567713,0.3101617986897459,0.15458977248267014,
 0.9179556670491524,0.9996558035092522,0.07181012694186384,0.721997271778948,
 0.8957719606920037,0.7342594681242832,0.19991473017664385,0.8767192287614942,
 0.19499415037279988,0.6666829383928768,0.32424551097111665,
 0.8558526493581962,0.7616399302525718,0.7245712514740498,0.28375081599845065,
 0.832539839434459,0.8362868968542956,0.00979856392864964,0.245105827361765,
 0.23658027108842872,0.34056067534264445,0.6157849394719426,
 0.43900712313301127,0.29551621931347016,0.6897796802301186,
 0.4721114703180739,0.7150021783424096,0.3691087927751604,0.6001001577124349,
 0.13666382482203254,0.46304733417082844,0.7724096684532732,
 0.39758433043599073,0.7818613159240639,0.7300349291594619,0.2726604500447971,
 0.7622550555644033,0.33887834744611134,0.4030711694238204,0.5226416811547658,
 0.7118939242876835,0.9681640401694809,0.30576018298703,0.9337632327207428,
 0.8092183436395423,0.15123612703608358,0.3456788152967588,
 0.49422191317052144,0.47300655416270454,0.7850133569663004,
 0.44898686942561006,0.7039607633151808,0.7683551659008211,0.5467352917066538,
 0.5787459472761767,0.41203404507115504,0.7450251035475635,0.6804152479613236,
 0.4077128824494307,0.40085124632467173,0.25597764314862437,
 0.023871705912708707,0.11202034358343504,0.6305757988447788,
 0.7726578965564729,0.9829795721966642,0.5806082550711807,0.4593551356034077,
 0.0155787195142858,0.6800181803953226,0.5198955436347887,0.49654472064424826,
 0.13435297659994394,0.6239383589350725,0.6555124898461298,0.6551835317490184,
 0.3486616158428071,0.38341964010436624,0.12720244802631575,
 0.9133406206080098,0.33119518709238815,0.2746330724369034,0.6102418534470242,
 0.6521369857109174,0.06471529263883846,0.07030617501283809,
 0.9137870503807025,0.48332268957324587,0.35220386245579605,
 0.6530906086171199,0.4153522703876228,0.17272109926195833,0.7856228536435528,
 0.32621250380821276,0.4363367308313444,0.2714046387635337,0.2749275420763494,
 0.8601315864584487,0.7332090444998576,0.7839852584306609,0.30433567517535653,
 0.9556534248747898,0.7134823912176215,0.9792610191702796,0.64140212707284,
 0.5878403079373226,0.08418260832540636,0.6000109816020142,
 0.08197068809620545,0.08295842019693889,0.6921284502299879,
 0.12428708307218062,0.8165072554774072,0.8856599892430149,
 0.11035667817590566,0.35101085555490086,0.31338472873277723,
 0.2969156512701705,0.060737120028080294,0.10396255045943226,
 0.07274788949780131,0.8387130843875195,0.5191849907834125,
 0.023254294085341964,0.8467101993687403,0.6328667267531142,
 0.5341537734630035,0.09198453244061466,0.46724640660460315,
 0.17266037067046336,0.5602407890835849,0.13011993377551456,
 0.8588986142521715,0.5745824602691477,0.46082245021692536,0.5721716040291347,
 0.36437012310095696,0.7662180978167494,0.46748056506685043,
 0.8148510925090023,0.18291929868167522,0.35324762132385645,0.935614243134381,
 0.6847363282355377,0.42295936880106555,0.76848704247466,0.20872173699991203,
 0.9370307024915062,0.9805060818855712,0.4649880565247204,0.7099326840508917,
 0.07023634012746549,0.3649607695682765,0.0043763525735595055,
 0.1623510875396933,0.6644887647840757,0.013584170860825195,
 0.11699106346775756,0.0533829209368073,0.9452966663523238,0.5917226548725507,
 0.8908849237211651,0.29099619927066567,0.15957457904511496,
 0.2700406115259527,0.987272190083054,0.20834944303176928,0.10097270240347789,
 0.20357949336183556,0.5026039364059515,0.5193996383465342,0.1460328212819051,
 0.4467909930502374,0.7205729822503123,0.056492821099466184,
 0.03127075547080649,0.2709111264523638,0.8521792785462756,0.5757281674861832,
 0.7945705749365034,0.9450597784746176,0.7177885659982187,0.38187969507228725,
 0.006218777879956727,0.9216778203815529,0.29551432090043794,
 0.4530077852175204,0.28132507021284126,0.037148995566481546,
 0.5347168161701871,0.7885032036798657,0.36138454038118994,0.5013526162638187,
 0.21784218083213192,0.4542271100896973,0.008227951693607904,
 0.5266257688962923,0.47283620470048016,0.1489685451429159,0.3694621957760872,
 0.6003103784689945,0.9670119715573702,0.08771823208496687,0.113556308363287,
 0.2882475520292107,0.37550690021050426,0.6931481487474461,
 0.017383118395319364,0.8257505563582422,0.7935698245705082,
 0.9482586489957952,0.3411298722249785,0.22275494568441223,0.2997816935278297,
 0.3317973709439018,0.8505408247235959,0.22742799790062762,0.8096616100863991,
 0.1686056588892747,0.6071045053623134,0.6611000691720683,0.4193127673425652,
 0.025037280171454013,0.30507746468755836,0.729346506765733,
 0.9383976616198291,0.06498204317241474,0.24931339797044083,
 0.6556282540275685,0.2936357957377196,0.559416994151054,0.4074233592162778,
 0.3422350368759306,0.9836698335448859,0.9834816705998738,0.6674432001254584,
 0.34968780782560405,0.46481180527582444,0.45183030323220164,
 0.3988953552267124,0.732527319324498,0.9626708993787881,0.48799963280971703,
 0.9196951943113362,0.4259383285179308,0.5595622110836493,0.6114426222583733,
 0.7387348492297978,0.7923240538406418,0.8486577321965145,0.03391336595071659,
 0.6161341598549932,0.1877574577455381,0.2827473378701695,0.11278046576185896,
 0.32663180937692593,0.9766415307768029,0.16828313019435648,
 0.05822176433794257,0.9035524988764438,0.6917640208890814,0.60479797373916,
 0.7810426240063296,0.07757638196552363,0.06026824999620217,
 0.4309621579453041,0.038751014596393074,0.18762868896640983,
 0.5256372712694061,0.17636958376585876,0.4582689022618931,0.6135546622230774,
 0.09685614350951466,0.03205537657954394,0.8275582702409979,
 0.46702718344742966,0.13496679541476286,0.9906088709219032,
 0.2960864678101959,0.6294994809367015,0.12727926146453117,0.1231118488881996,
 0.38410121746845594,0.15084865262134628,0.1127309684911525,
 0.8834611426959251,0.622348093189214,0.20829426659307082,0.338649777427276,
 0.8757078070885245,0.00936269405086676,0.7529608090695086,0.8145862200100202,
 0.3851754612434979,0.6175217361060766,0.7620807115075596,0.8884774824237133,
 0.5769156282237096,0.5399976439815226,0.9403290910318484,0.47402240643754645,
 0.1748113753872148,0.1590631149567847,0.877023580926698,0.8515441920310295,
 0.555256459396372,0.24608002584415434,0.901088675330375,0.9566322521053623,
 0.3420831640051495,0.9987506934495535,0.884479358296636,0.4147628564562802,
 0.651393491555118,0.7196952634910918,0.8492441462527311,0.32654039889626985,
 0.8977235722413296,0.12231552710755822,0.16659654167739735,
 0.9613964503095231,0.28086307642118047,0.18549186579678978,
 0.8965834231367944,0.3327794591820529,0.4211926461407245,0.20438346641614746,
 0.6605285189317482,0.8195195495296861,0.38534642187093715,0.701790584650198,
 0.3085235845673626,0.2152826753857333,0.2867260237320163,0.5469875309340493,
 0.5377201339987923,0.8857865773903573,0.33500099561584884,
 0.11940931321570125,0.6631532521600757,0.2941134457193332,
 0.21418014387159978,0.9049150922741194,0.25540413714892973,
 0.5774937598183856,0.6489481370470231,0.05585932708601993,0.851144025008175,
 0.6238783107012189,0.6302089208625707,0.7896321442371486,0.6374011263415253,
 0.484089025736376,0.05606392553650108,0.07383325202912472,
 0.44341879976379683,0.8457133496057994,0.6903895137025026,
 0.02371044754338536,0.5712727927736869,0.2702517800528882,0.9977399616127514,
 0.7077977125018029,0.537923116785827,0.29750407888849584,0.8465806074968661,
 0.08819701469838104,0.09929983250596164,0.26307965316049486,
 0.06509545098191183,0.7203111954315023,0.6560921101717799,0.3807475746159128,
 0.37754126849083824,0.721379713542363,0.8819473525093297,0.6206683835015175,
 0.45120037705195015,0.24729431800501167,0.591400616321537,0.8561251463502658,
 0.1766716083375045,0.5454303116778267,0.1790175071117045,0.21309470569384714,
 0.2849633408789083,0.937857568792365,0.7216035088014485,0.13030286269816393,
 0.0718659624627962,0.04672013309648293,0.4135124242477888,0.0716592446423171,
 0.7441844299999691,0.005963736750998261,0.30148051706022017,
 0.3712580964825707,0.36140551811887167,0.4423910080756811,
 0.08523926183982877,0.8761010427373503,0.7690640478932651,0.8300957679366422,
 0.919273223263506,0.4180286961804418,0.47974671272367386,0.07373082995964308,
 0.42593771739843667,0.5655032603404243,0.3378630466901731,0.7048150907903147,
 0.6594491792311169,0.6743210212869278,0.09864054273346679,0.8994509348186664,
 0.5800825273119273,0.18670570500526962,0.898248502085144,0.3143645515853212,
 0.35574752980289415,0.1098848690352412,0.7323549298105743,0.7679190587494777,
 0.6055741896944722,0.5521933576735849,0.6882755478439047,0.3524653183228319,
 0.5192879731335162,0.13155047541826315,0.9251546094671493,
 0.03533184929654165,0.3075866208749404,0.6051706360027544,0.8682451123405083,
 0.18558280657982162,0.22321864566973615,0.9737910084463507,
 0.7772514672444307,0.9275811461198753,0.5439320590568127,0.7168977649841982,
 0.7100534543233428,0.23511445229072447,0.8609985926933048,0.4413298842595984,
 0.5852738194961503,0.9889578303525897,0.6152800230557945,0.739513611542632,
 0.9229617744942817,0.31492328194676755,0.8368155162449673,
 0.09248070965975808,0.48780741633176217,0.09397509370436696,
 0.8068522006694941,0.35876660987202524,0.05100007622545877,
 0.3198380871552886,0.3818136245060738,0.6483469101419647,0.7314164739729536,
 0.4229221655547005,0.49424588298277927,0.6168761779238665,
 0.22181224607162808,0.7629900832349142,0.21337433611522494,
 0.19280318394779894,0.7272632789544816,0.18554939748546262,
 0.5239231115905834,0.043527937095500846,0.09273331972916254,
 0.1314801618248349,0.6211187420566285,0.8665111516988655,0.49236086911069854,
 0.9439887595605294,0.015266513936249226,0.985518551902727,
 0.22295676250277097,0.9501844798986823,0.9480000120575812,
 0.32351169443975547,0.39616207877998943,0.5572090453107408,
 0.8212113813567034,0.1994207779635837,0.831814981608231,0.024594785055031654,
 0.8289172569944936,0.4508496564618081,0.7357950576453989,0.9052642870993102,
 0.6111073147508176,0.9119755766584257,0.7546163243253525,0.31993768819598034,
 0.1908220823084802,0.7350137504690317,0.6587703132508759,0.23617680796644425,
 0.13917741925826876,0.6670231044961975,0.8052534369400189,0.6412800129027735,
 0.7153331276851185,0.9324403792812119,0.36917475966138713,0.6171244442120969,
 0.07824272143330147,0.5972360306255595,0.11184720317356722,
 0.8607312354916318,0.36862229674105507,0.8522164826942221,0.7418737754960343,
 0.010138840340317756,0.49974559271625896,0.8993025149111423,
 0.6370930829164635,0.6327662346245096,0.9582244990710712,0.09502818301471272,
 0.31443406562030296,0.5201944201357059,0.16314860510182927,
 0.9621677027264816,0.4767398780391441,0.5576296319018988,0.8379031525608606,
 0.6870321107885579,0.5707335883364041,0.6344985497017438,0.6925498055537854,
 0.7045213596949352,0.49841176325230074,0.2514005119997258,0.6249786996526707,
 0.25227996329562485,0.8195446111117494,0.7529126168237879,
 0.23169273610133567,0.9835756336692233,0.713253473006945,0.47189714692298645,
 0.22477199000152404,0.6064317995795698,0.4587895047128333,0.273981074258704,
 0.13532317419843853,0.6790267282025467,0.31438369494104756,
 0.9627718445860171,0.21407870050966027,0.6044490134239102,0.6368749085684565,
 0.34919257434255035,0.27798427983605767,0.9079334550180927,
 0.7895962554902953,0.45960669902859264,0.7766947886761679,0.8123217406984496,
 0.6729837568810854,0.9479938086613093,0.64406050095223,0.23705843070387422,
 0.23557158248368182,0.014613775242867999,0.6141067846872066,
 0.05205186482702018,0.46971730704111836,0.7600039478912691,
 0.19961704531639912,0.7152304825625952,0.5220386559272063,0.1696470622063595,
 0.34754926473466596,0.7789925256646382,0.8404808609155726,0.8851056667475348,
 0.8507057366068522,0.4208311146850654,0.12896190384079964,
 0.09698514497547817,0.7235913511636844,0.890535377962509,0.3793689762274909,
 0.05621107652388546,0.7051019274449256,0.33026341256225156,
 0.30357073989218897,0.10459153135101551,0.19583218364321187,
 0.13129019206004644,0.8560805456306501,0.7955756631822019,0.4979285314595856,
 0.44352083957580424,0.16744151748041547,0.9254313021327752,
 0.9997695335494649,0.18137231811222254,0.676837532248746,0.1569817850373495,
 0.3618072357268364,0.7717685324063088,0.2311476622893348,0.5751505419842633,
 0.6725440654425446,0.4536419616589026,0.7376052523379147,0.6863443556024296,
 0.8556470578728367,0.7499672475106482,0.9310433778819517,
 0.038890987089200024,0.960980571040453,0.5949024115710151,0.9966443247287962,
 0.03827598517444475,0.45924763130261603,0.17093860012587025,
 0.07368313114266178,0.09311753638262987,0.6408129938091592,
 0.14140630361906992,0.3602533732349381,0.5055704041639714,0.4727073814562899,
 0.4596185345293696,0.7390135589577016,0.9463373199354712,0.9611978664218656,
 0.6587157745167884,0.5410630551365657,0.736359853801972,0.8554609108396372,
 0.16624236494837008,0.025305444918759634,0.1536738671451049,
 0.42350217043479677,0.6279597631499039,0.2881387067223581,
 0.057353394773068445,0.05615559943949178,0.8050060537815545,
 0.5922777612669136,0.41849275930301433,0.885516898973543,0.8071880704042074,
 0.3860486195870313,0.3446224726934244,0.8974701308020627,0.14437920424829664,
 0.372037188582659,0.7731368558513914,0.8484589562249762,0.37200918453150134,
 0.9071225859127044,0.33231200574593567,0.8378599506818751,
 0.10426757970686751,0.5765867981595878,0.9667594995284737,
 0.04383809840114372,0.6566532881371974,0.43341108090018776,
 0.005689773930575792,0.5600881073935224,0.9599799895043235,
 0.5587946918160782,0.6673900701540189,0.058365279771690814,0.694561456942754,
 0.5394308114766245,0.23005414216110287,0.1761670222570073,0.3198675782579873,
 0.8638394795757425,0.29163209405021706,0.02704598130162994,
 0.00826224649506857,0.959725247025821,0.47840466705218443,
 0.011746258222244843,0.6957500376302745,0.29319097128188676,
 0.06874887230485771,0.11177052055602732,0.8778226094608228,
 0.5086553386116223,0.6776291928415568,0.6201254139061639,0.8587709059883246,
 0.6130909997573573,0.012512433831896375,0.5766121268707152,
 0.21738765325503395,0.2087510762238186,0.05023132949750131,
 0.9142237981997925,0.05332076304761979,0.443801377376829,0.44867138737612067,
 0.32860889764343015,0.33924067336846964,0.4297043534589039,
 0.20979835233558208,0.7805730955921213,0.18132496222988248,
 0.06708887498936089,0.5748218803321188,0.18742434377419148,
 0.2159646056872544,0.8831775757321707,0.9646543483253662,0.7378477788974375,
 0.8775384922895559,0.27848030869345164,0.7199956028802326,0.5657831636885768,
 0.9376829925845909,0.7921287341183967,0.04278749434674245,0.748791004914002,
 0.6821201385005543,0.9792595728997281,0.14276469277292325,0.8153084823914112,
 0.9839893295123144,0.41406441668396643,0.022596204891723315,
 0.31581333660733935,0.15858927918150578,0.8032471312972627,
 0.6036192071017339,0.06277783270039672,0.3544299238267662,0.722224860880956,
 0.033832962548512535,0.8349655068111095,0.05347689837085401,
 0.20893658444149033,0.18682267622332027,0.33474996875130403,
 0.6361257965713774,0.9907621385959667,0.8348848130327816,
 0.021510074351749342,0.959411439186831,0.8381293146287306,
 0.13966854704607412,0.10359047734830251,0.9128288090848646,
 0.13357602678128433,0.28527431170248163,0.8887281841238157,
 0.3740280597407998,0.9858527992271802,0.5418150147306623,0.48288170894639193,
 0.989258153497042,0.587890889936361,0.037292077492903086,0.533347257650309,
 0.6093337545813835,0.9106016586561406,0.13277773763183331,0.8275114664769246,
 0.9681101894459436,0.17306956934379358,0.03146865372529073,
 0.2506727134329466,0.19632797342792352,0.5498645204130285,0.9712871852634797,
 0.7087832084863598,0.5621896652431242,0.06894127212565915,0.5101132256766294,
 0.5828379158345288,0.8984742330752362,0.09143792576847387,0.6451506827183352,
 0.6561982999183594,0.14035527151812555,0.6047315790770886,0.7967961546286806,
 0.2985602309547961,0.9923162426782951,0.7772277295711423,0.8340606811883055,
 0.5250574126084674,0.923640994627487,0.20956460614535888,0.7645194038298636,
 0.498117404122551,0.17465926833925327,0.777964142803981,0.7385650329119435,
 0.3735815098503893,0.21169187532831002,0.5798563299915794,0.3496526158480028,
 0.2275528114545161,0.4519208035624641,0.8831376074484766,0.7589869818272995,
 0.7922447010297414,0.894837060580504,0.7945466282924494,0.8850973977687043,
 0.5605394798474584,0.07533481422155097,0.4191528737874325,
 0.11044308509008893,0.7199580200509158,0.9687715843755755,0.6273727255451143,
 0.10418048821561554,0.5080301601994521,0.0860691153263613,0.9658282874347401,
 0.7563875639866552,0.47535052213752504,0.8168943256396333,0.6162709490437289,
 0.6860950955603458,0.9557437200465837,0.8903340463929605,0.432176405213863,
 0.6823818677037594,0.9399948156225518,0.9095800874913809,0.7544314403686138,
 0.8508632506671334,0.14098647266197717,0.25292890344542385,
 0.10281187678924508,0.6971758748194745,0.9506991927852145,0.757548564533773,
 0.7024101117483681,0.22267438624889135,0.7210733909444619,0.6545192535654509,
 0.12233396804940777,0.8138740861522707,0.1534808152273086,
 0.08174984287640819,0.7816469312218554,0.9889972869278899,0.2643379643298035,
 0.8003828322227946,0.7002824559364781,0.42041751760658197,0.5339559640099487,
 0.0809101585006734,0.8858129511345902,0.9912555205106472,0.9515764488055257,
 0.5703080013897812,0.22475323735058672,0.8630139006227767,0.4226709763280613,
 0.6046381249296957,0.09488321152691881,0.9712425167924401,0.4715469090137271,
 0.09190524865818495,0.8961949083219796,0.9755688623594788,
 0.025978125524728335,0.08379599474423649,0.3718557783997155,
 0.5966521739858262,0.21082216269689913,0.9718751249745072,0.7381510799742915,
 0.8451343946029318,0.7847443138197356,0.4561592660478532,0.3950767333995696,
 0.6111108788264612,0.19329016973812008,0.2454272060517566,0.2843826440714393,
 0.7207476898735815,0.08407340774580374,0.8848569345442648,
 0.10152866343339306,0.25766716487513586,0.4270011244945253,
 0.1372815915307133,0.5948661955860165,0.017957887092581837,
 0.3461007975369179,0.928985756707377,0.5124287560880613,0.6972073168513524,
 0.3452243012391738,0.19591665797955005,0.25173082753043285,
 0.08917363992133298,0.3484580098152126,0.27846517320168396,
 0.7501900109820467,0.22466077599586531,0.24169946369724582,
 0.4261768165026132,0.43209550165754607,0.6969045530474474,
 0.24808026979968778,0.40014959485119306,0.1868450238601822,
 0.8056028793221343,0.7454146995819529,0.4508416298706215,0.5556833779178403,
 0.05835307022220171,0.6379460528542119,0.484015011206153,
 0.039808284555921225,0.30186986259223936,0.9450594443521573,
 0.8311391335095829,0.8954152708687374,0.40796813754328254,0.4418423726120194,
 0.129814760550091,0.11043066977480687,0.43129854369442266,0.8632314896151542,
 0.9827970618102166,0.630076571808251,0.33448557957378133,0.5162289515900836,
 0.14820204015504157,0.8686661199112898,0.8441973410841407,0.5931877132452981,
 0.3664034793463725,0.4479278317325994,2.700923728278237e-4,
 0.26260737272163415,0.14610138576955545,0.5110284481408017,
 0.4761717081601694,0.375080093591612,0.4739374848515435,0.4913992404878762,
 0.09821856310022792,0.962699589018971,0.7780412288492947,0.26168915840482165,
 0.8734786594591929,0.28070732708550805,0.034587789832057236,
 0.8005669706062748,0.5398506925560709,0.29281454077928504,
 0.09930907767902974,0.3022884562310192,0.6886738540408606,0.2428266856605159,
 0.08410422166506215,0.17245027682303793,0.64151444065776,0.7741187970733405,
 0.9662749309660363,0.7837387024808513,0.7320696099037505,0.39208086646269924,
 0.037159918357598465,0.46845098366553817,0.4955986330889587,
 0.8331940675703223,0.029841933635241014,0.7257066686133683,
 0.5911465780131124,0.821958550249106,0.7401102590079465,0.011680347358987597,
 0.5387438522236689,0.498422486850012,0.24641177487159915,0.26687719119783426,
 0.8216876902706015,0.31660450152063624,0.40859677003630457,
 0.5458631088802275,0.13903140231834854,0.7041130032839262,0.3353183181862316,
 0.797332539380952,0.6395712302999772,0.5387034844965378,0.18775540566430582,
 0.316926148396377,0.5956449619236983,0.6993531745478327,0.8466409293258603,
 0.34197757239030213,0.1264435613011129,0.11039186224123843,
 0.42103194029808955,0.42948564525303534,0.435240223019835,0.6776981556115284,
 0.048789626730509905,0.8546160645240983,0.21369539221054157,
 0.035217912708884636,0.15785878354027916,0.4707817928003587,0.36922055997672,
 0.7325037632721483,0.16333111868622485,0.30308784258047106,
 0.15858243651575232,0.12534351492549478,0.3431527499829288,
 0.5347112937825853,0.28675376770909844,0.18420525493952078,
 0.6858786923140033,0.807586345553742,0.4904799173904236,0.6312373122674317,
 0.9892349502496982,0.9859705541041305,0.5027511005652986,0.0751312105111075,
 0.6098434350915047,0.3276376944695132,0.6222331770125861,0.6585315719047377,
 0.10598333390311698,0.40123021497294753,0.8432747796708563,
 0.3527859566985406,0.8660435350577683,0.4817903814771958,0.5302993916424418,
 0.7243734880716626,0.5180383688442693,0.3971173848008068,0.9063299525302623,
 0.057556134487828814,0.6231234321147788,0.5194260711877333,
 0.40141534843511373,0.2536611194897488,0.012451183916521225,
 0.3415726504093065,0.23334128112990016,0.8804220501729727,0.8135661185181587,
 0.16928026976246968,0.5334082457289935,0.6322267876950808,0.9074655327306826,
 0.8341050190366562,0.059866088394268546,0.02515658835896717,
 0.48787913524245785,0.12888556550837427,0.5806127644722292,
 0.9943914201865219,0.8398232536701904,0.7504218656850317,0.6738991380709694,
 0.48463136317999345,0.6890832387515817,0.703409833034383,0.8893925108514045,
 0.7772859223077353,0.1524106162238006,0.8965302114268042,0.9261312910538896,
 0.5638658791380083,0.5866196583130829,0.5269999296733192,
 0.016289964463857443,0.8777549389754352,0.3725589541653267,
 0.11796696499406312,0.9857935385890435,0.6086924572339966,0.6131555889204215,
 0.3982540193527916,0.008844164927705678,0.1259891001481631,0.992776855389647,
 0.28513672413156044,0.7253267033266112,0.5812939529926152,0.5816335777731866,
 0.9332155963028907,0.8500808746632839,0.5332180709715901,0.7145177924385743,
 0.09829049378092047,0.8774990234254596,0.7002571552592902,0.5569264977883764,
 0.34626586866866527,0.3568403116517519,0.4386531125443739,
 0.015477578251593793,0.7201096824686042,0.3395522780598432,0.447841091945437,
 0.6875498844363358,0.19977183500423734,0.6005374098257367,
 0.30819279293556334,0.5411405714892259,0.9053357036891884,
 0.056500482633400906,0.40708720464340153,0.05021368455068431,
 0.14015954053025648,0.6395490687733306,0.8385906194288332,0.7687826266944047,
 0.12996283459786417,0.007721016304526929,0.6977677912462794,
 0.7325318198720343,0.3256009226236336,0.590661058635237,0.752950709567974,
 0.2813030685263094,0.3828628594542943,0.5629494115880249,0.3457988122323966,
 0.7297848732384292,0.8372286908363739,0.22444242578244233,0.8795197111674443,
 0.31757743022504936,0.9071901714122489,0.1632437174296799,
 0.034828372130586205,0.6421247767726161,0.5451498032214912,
 0.5876775138538305,0.5446352571624496,0.9347353439189371,0.5126312544321228,
 0.383767453185653,0.0797165824050674,0.44282401138539695,0.9811584784543078,
 0.6669985670161023,0.8960845978076525,0.784591523861176,0.9055341457655913,
 0.7214560929688563,0.9332658461747272,0.5049579495340556,0.0225227353467643,
 0.568655657424088,0.9943057284802853,0.8106275678521013,0.2885224973833327,
 0.4524875229978407,0.6360512981535489,0.11169045562033109,
 0.45754701558970623,0.9552584004724468,0.28131934258561364,
 0.39680530845611317,0.3770282873318307,0.851713146226329,0.40005857134248346,
 0.4564403669745334,0.9632052576555737,0.2854190288153472,0.5790861729795354,
 0.7215515304214798,0.8247988839552862,0.64310721267845,0.41952609946917185,
 0.7275413725830107,0.39772306805278057,0.6447616977811014,0.7929269715800602,
 0.39107748129254016,0.874681234885727,0.9960976131835274,0.36046908746842743,
 0.550787496684116,0.737888996447061,0.25800916494669257,0.5131712607361405,
 0.28694305587221824,0.8766950913237155,0.5361394968842448,0.5563251140376302,
 0.1778741352097183,0.40370118228415375,0.05710911038758326,
 0.1415671533246161,0.8811993459833938,0.950989441488884,0.9185834311559188,
 0.11261950647680563,0.19846571829263238,0.5930146133873134,
 0.28791642161291187,0.8086544196062517,0.1584615889515273,
 0.09042826962508588,0.2820662486664868,0.6216633926382493,
 0.06638924146378433,0.4288674181700777,0.38865603603419374,
 0.8740527807681922,0.0037671834356047817,0.8293678466127979,
 0.23671985907409343,0.05309421263966729,0.7058912939145409,
 0.6954927333389803,0.9116771758672594,0.9938816299475063,0.9345803611008097,
 0.0323835414721515,0.8348344388772764,0.8252803817197368,0.3529119155450614,
 0.48985127385581917,0.7432019886467576,0.41836555993317803,
 0.6088156646003912,0.3300763043017725,0.43413728240924765,
 0.39924285579667307,0.16506311727908551,0.1270594301149629,
 0.8467107630544324,0.08379490124721656,0.6858062748585122,0.5163670683539555,
 0.11296485525567834,0.8481203568503497,0.0765656338188504,0.6184906594840138,
 0.3308481427001341,0.7636232105324788,0.12596050783750012,0.7275711165140877,
 0.6798930086018022,0.8349234545742529,0.4651380945239748,0.48157573134791254,
 0.8337781033070654,0.0588991780340028,0.49284227163206795,
 0.49980666498471393,0.1381821991867732,0.4150749132170144,0.2299540171743205,
 0.6764549848387093,0.51458457784775,0.8209395416607874,0.4704215530488398,
 0.6507137776392287,0.843418146869498,0.9529715963933527,0.24999996004865377,
 0.6631729581310487,0.08894534444098146,0.08490601120148256,
 0.43346047045493896,0.8309003678576784,0.38653570414492433,
 0.34771999505650064,0.6805652684323112,0.27212745202297006,
 0.8702171039832993,0.2594023652681431,0.3594928563321874,0.40977164181903325,
 0.82214569591955,0.41954962959308195,0.7109052966017758,0.07796689447483074,
 0.7637680738712316,0.9122333732531832,0.4499651209724477,
 0.008245111683243045,0.5194029185465949,0.23234941088580763,
 0.19784471748139532,0.45730934474096774,0.5007043426498677,
 0.7196056391744482,0.6766428801969007,0.17831198916640167,
 0.05707064160773356,0.0995035542310051,0.31090971215673846,
 0.41163817904653244,0.888856130313683,0.35630356592764545,
 0.09226324396947994,0.9352904103383997,0.8036605898399685,0.6182742884283641,
 0.16979885405444317,0.7772169654303054,0.18916865344620248,
 0.9282740916863332,0.4947344527888913,0.8219188105902397,0.5794083313837297,
 0.4392578211308389,0.9644700530698072,0.8362895931710914,0.4620628134286615,
 0.9011417838197406,0.2592053680030614,0.4027016162989909,0.34824824975982416,
 0.4142665011318718,0.4205495790766943,0.6629914523210768,0.8139553835698127,
 0.16291641455106132,0.25890449722366005,0.9531851067427943,
 0.13985308074425795,0.8318276267772112,0.8029373802658163,0.516689100654867,
 0.5982472305045335,0.35083494872286936,0.4526472282952042,0.1275749382789032,
 0.6397916236922727,0.6519490533214647,0.8551719640553181,0.6993914459669089,
 0.5783291745818919,0.6449620252351984,0.16796164916615297,
 0.09085554610540891,0.09808787300703958,0.23371103782950808,
 0.8609381208921689,0.7813760464106594,0.4128251466685672,0.597882964509219,
 0.7460499385689386,0.6810470295941322,0.9709297636269573,0.3594392875182919,
 0.8786161396516202,0.2913941539277233,0.008879417181478777,
 0.6454329415220635,0.3785616991888301,0.24554154968928787,0.3110784503603502,
 0.3894502195452815,0.6182992854670188,0.6179105430242012,0.880232672025832,
 0.412794714851513,0.7393128091878184,0.05661530397224601,0.685445975493143,
 0.8965049908991422,0.18860307825044353,0.08975714179734862,
 0.3104354321894154,0.2746757133034531,0.7476170502495089,0.8721415615493202,
 0.013942261822369906,0.10304992205924735,0.9486588720307689,
 0.10183891507091958,0.2550100555954373,0.031423097891209006,
 0.9278629379139736,0.291640037148579,0.963622806366713,0.6574199592814334,
 0.8788212062483538,0.13844091031013184,0.18740512415748856,
 0.3617077131487345,0.8547072932707728,0.23919226856087372,
 0.041983900332694835,0.3655609375156169,0.9065575509515891,
 0.7297845178760789,0.22720252814794606,0.8998548762533747,0.7420025980693404,
 0.0016164425782656533,0.8244214597795978,0.5783430579575904,
 0.30175095549932884,0.5522929451460985,0.8425650421460578,
 0.02667205229426739,0.9454822634974862,0.5246171512948494,0.6796556430351501,
 0.5508031952461908,0.3385512121380444,0.10136222027000485,0.7928256160259535,
 0.13251378524900126,0.8774240226153328,0.7638984644487137,0.1523716023878533,
 0.1385653862646492,0.8038948945811768,0.21795230782719588,0.6289343741158215,
 0.39392705547176354,0.032583976532544856,0.20933024716337112,
 0.975060558093428,0.298992627585119,0.0568095240964005,0.43620233467931335,
 0.49824533188386344,0.600328836266484,0.3892274844452076,0.6715778986914074,
 0.22818892411306768,0.8587292436326444,0.5133780986385386,0.7206542730652896,
 0.6115678563944044,0.9718401278876255,0.1062603063690698,0.6507868384936335,
 0.6981619376522336,0.3487699132359271,0.3309188119224814,0.9099995822494541,
 0.2250679902953754,0.19915447178817391,0.8780148607408298,0.7799249051259202,
 0.4823256223880299,0.44765424805620235,0.6400629385927423,0.0679088239388208,
 0.7957413704069897,0.42319894736271313,0.7102908755656959,0.6682152810256908,
 0.3568696066814486,0.9969478765474601,0.3941209243140874,0.9855787360755426,
 0.676670741892878,0.4294981810444929,0.7302861241293017,0.060772155193992816,
 0.5362941494590574,0.7792725524745546,0.5568465698073103,0.02109008385114619,
 0.8750551085948091,0.132560192747888,0.9726573327257059,0.8438203168864666,
 0.11052093103066296,0.7135202798017486,0.6743702090147845,
 0.05230872309780421,0.7665536395489162,0.9522880741375606,0.6387511474973026,
 0.4871082137861724,0.15147958248318427,0.9330596957872931,
 0.20019141249771855,0.40967467116706413,0.4481254889852797,
 0.18855208321199435,0.9915039353286633,0.4071949397595155,0.7884054329120329,
 0.828409573338344,0.1314554543913583,0.03717422494973455,0.5953634254591275,
 0.4657863463491817,0.8887593521092102,0.04378013287964022,
 0.19215490268291135,0.7516106953825274,0.9594111476386538,0.6062228547951855,
 0.14009608894107872,0.4375474261231487,0.4047158399150418,
 0.15588259545309457,0.9035706599396138,0.5628703366158523,0.5944816490743741,
 0.49902602885607394,0.8318199509743617,0.8388665708775007,
 0.39919326149687206,0.05624479501265123,0.9952175133579402,
 0.7250086721524712,0.050299805418268284,0.7388166950442839,
 0.9278657344808614,0.42582748607028165,0.800588368777539,0.5609874371723264,
 0.43966399536882217,0.3230999069395668,0.872182110889022,0.3481907397114655,
 0.8605158252510979,0.30613462327289875,0.5579140691623532,
 0.25403964308845683,0.8634758922915685,0.8568829236153457,0.3835335269301223,
 0.9553140290984303,0.26820774276889814,0.007169617197173017,
 0.6238817856365023,0.037158082842581264,0.3344045657400264,
 0.6047306159609553,0.5845189863757831,0.6857245451958603,0.9776974766910553,
 0.6969283844220423,0.0355861179770649,0.38075410879369087,0.7313930942901302,
 0.3511467570529314,0.7276307035265285,0.3131881717287416,0.6654489572984474,
 0.5627214101721575,0.6724181934662005,0.7371441798851961,0.2754777677097243,
 0.7030916869229847,0.5836064940016629,0.47002685034682756,
 0.14378146522087376,0.5231729859693884,0.33343873816068326,
 0.6307049668958947,0.4757436675776314,0.4341655653606836,0.6015059491274326,
 0.3231623608205647,0.03537507024333675,0.22095983090885862,
 0.7716078048020851,0.5379480987619686,0.6372852302073151,0.6632774368332636,
 0.0748045302717415,0.14607083405193855,0.5580767413237417,0.6463604830815387,
 0.2426936517216024,0.2707596143654013,0.561646884412855,0.3271890053482417,
 0.749126959453577,0.538192919394197,0.5504870491361229,0.8799267759094991,
 0.8409450459484413,0.8375006706748431,0.8381429216057205,0.34722831886034256,
 0.7909091324512056,0.006117562381245056,0.7148530958111835,0.371957470558961,
 0.2514976242342303,0.473786399943808,0.24459989720378938,0.8083814972212882,
 0.026716063341430063,0.45736291887851965,0.1477131956119978,
 0.9533177507683588,0.8849007233961382,0.928951913689197,0.8116155909077258,
 0.12167827783523999,0.8352722763614462,0.07022309049822495,
 0.31222504710610666,0.4263916762507707,0.8706004410359364,0.1456725883711698,
 0.9237675240755092,0.6081776609277773,0.42992924457736437,0.6955716093784634,
 0.9026335954816822,0.02293794875520816,0.4418307948824157,
 0.43293165614599716,0.4101460908823946,0.18083641213450496,
 0.16216734538220712,0.896990467476854,0.22007145050733046,0.5017366476806148,
 0.9659837659830659,0.3162979292677802,0.9409962541140304,0.9703208273061916,
 0.08393746886701081,0.32061586582281165,0.13641528217949828,
 0.886784623842392,0.20503739466510107,0.4638954722656098,0.29657346575168,
 0.9848850589886574,0.12092030801296438,0.16611206858604544,
 0.4549784633527849,0.2337456052921434,0.5550711554144458,0.7412403563994157,
 0.21377450979504176,0.7399106544754273,0.6412111107545087,0.8705675282124093,
 0.6307021919035033,0.5433947804304287,0.6072600517546216,0.45426606673549896,
 0.8577181335114933,0.10629178679490814,0.5821214198054289,
 0.09598866175287712,0.17105929373329787,0.40287041799396006,
 0.8235765172920803,0.3702971115491027,0.0428932241029939,0.7645829960029029,
 0.3491817871374835,0.79808259148841,0.8762004571601965,0.5470576832299459,
 0.6344172814530252,0.7071272934036484,0.9406854308309909,0.8701006195678789,
 0.03817914933779809,0.4851004275251122,0.4287256880361082,
 0.11790012854627885,0.21769673195939288,0.7412867772365199,
 0.6696451088248435,0.7859311462401153,0.6258754157305996,0.6042992792921943,
 0.2885025896742456,0.16403228618685095,0.33629243966808153,
 0.39597666956011623,0.019879885482121495,0.6987499479383171,
 0.7878208216712903,0.2755641788117804,0.11560367476745403,0.3685931441144251,
 0.08382211225706138,0.9769457846901893,0.025787339188506286,
 0.14484324270450954,0.6450576055752384,0.22171906698114452,
 0.5914312559762288,0.902616309905245,0.4460399528168575,0.2352704085081032,
 0.2244733977079053,0.33096219132447724,0.011778143729523238,
 0.6147879394051651,0.2982796461991637,0.644780110894357,0.6632225059634187,
 0.804663638641169,0.8015906056241989,0.10046795169858103,0.1453080323009538,
 0.7137665779605376,0.854366556619111,0.11811897806587968,0.323568475897004,
 0.15369964875102615,0.7334159536936952,0.7929993274555828,0.6767042662620879,
 0.6466654703959291,0.4268231641785841,0.45285457133235885,
 0.16932236169230164,0.7441355920866254,0.018741490965089924,
 0.5775098301167599,0.9818751246311461,0.8525193300007056,0.9387490585030309,
 0.3648728916049516,0.7948426266414549,0.8055584200059346,0.7071316595683295,
 0.6607156214169381,0.8865635859356825,0.8956759832078313,0.5580848389513688,
 0.061838188332938304,0.6114573631657788,0.5253579276119912,
 0.9824631985584323,0.5139633404555158,0.1523140699424398,0.21552933475247826,
 0.01714926083640289,0.04350137628696893,0.6211049349059788,0.83358414205323,
 0.19809422344653993,0.49461354005027325,0.280038017597973,0.8348537794785789,
 0.6015996697423823,0.49990849722875397,0.23676015106398296,0.681087275246811,
 0.17549193450469291,0.14802452572085445,0.08571462693764209,
 0.27585506450590946,0.12011904289114683,0.4515254690608207,
 0.5619323759917356,0.802452074349281,0.9125720113739779,0.29806144557312875,
 0.8204848573893959,0.4462953233570208,0.1270901272084255,0.8238320320047728,
 0.9737729009572453,0.9738111802724368,0.42241383210853467,0.0558953244663839,
 0.8537844278128118,0.4055959873081755,0.07854572320498565,0.6523320222077877,
 0.5019882963811751,0.36716690905492855,0.5273266305379927,
 0.08535010781224817,0.7155638282391694,0.603279107358229,0.28666540164937004,
 0.9430798957834421,0.3940660903623183,0.7402422301983806,0.8839254531807735,
 0.6590557261628486,0.09599588028856154,0.90303711650014,0.22054983041318366,
 0.36707951520982296,0.47686589396712764,0.5291720015860516,
 0.46611065874119384,0.21661311531466887,0.4155908010351781,
 0.42850649076827985,0.0010794607760900021,0.5069869882421254,
 0.9191299097788789,0.3455491339648158,0.8117181998122173,0.9336692791937431,
 0.8428274740303839,0.9429094443740222,0.12412042574863169,
 0.09884339948661935,0.169224812291922,0.4385847231741493,0.4258860018030175,
 0.14538484072356206,0.8990139636461004,0.8166369426615683,0.5418515310461361,
 0.6987191506238205,0.3465184170297688,0.13671410079560886,0.9075984316069087,
 0.8165750202568797,0.9482931367193205,0.7833568579251224,
 0.005455832960699181,0.05072038216945918,0.7063910026855493,
 0.7469326517503599,0.1562170192686928,0.459809649036497,0.4027896734238122,
 0.7489759287150802,0.9140251808680958,0.4167346001589771,0.6937089098242433,
 0.17636484349525838,0.7000352456693371,0.6085322462589522,
 0.46300337807054115,0.88392853274402,0.30617910777460133,0.4005273373830679,
 0.7230559057264468,0.8993677212687305,0.19902259797170851,0.8828592487947261,
 0.4978948924694584,0.05803138148180509,0.6893300354287025,0.9695616945494727,
 0.4506747347693961,0.5938153456231021,0.7247430440585307,0.47910161572147947,
 0.31235576064013826,0.32776193743429427,0.0651731775955533,
 0.24941104376816736,0.48314603903958675,0.40222724997445714,
 0.6414337687559308,0.6947591328799397,0.2185838864133668,0.5527453394240913,
 0.02793380518862343,0.09054749869023737,0.06680760705065025,
 0.2547356358048021,0.3116144987031553,0.6275861027638211,
 0.005451618839842798,0.9130923780780749,0.9433267887753819,
 0.8514608540295531,0.30940838721496955,0.7128902831062263,0.5143728118935458,
 0.2286969732562829,0.15856163351591712,0.24138910549923753,
 0.5540229655687325,0.10559764597311494,0.7326570216382392,0.8955902982151764,
 0.6451994336156595,0.05667416836892336,0.8626580128744514,0.5559491748153721,
 0.8032767017128112,0.7785152295518906,0.23218048778824107,
 0.15621080743504545,0.753171980725267,0.3616583017356554,0.661453313889627,
 0.4671870840059116,0.8828130872597743,0.392406865066268,0.21730447831024602,
 0.062306863593183714,0.3116222806008555,0.7592284137778774,
 0.5233348008296823,0.4958676581304151,0.6374774091397231,0.9520838724773466,
 0.20061987683152394,0.4065637450944417,0.3004889496950698,0.7026322584102417,
 0.10615706771570899,0.7161628781985696,0.679810132211233,0.1376227847152255,
 0.9333616189865468,0.2919211989657525,0.2932336445257373,0.5306371505004633,
 0.14729336656335246,0.5994751502810995,0.06578740462504074,
 0.5526599550387098,0.16292385176486401,0.24003097155570852,
 0.5038224546952317,0.2346553984329055,0.12154410986745046,0.564076527357447,
 0.16450208190640236,0.8198696564631678,0.89648586190648,0.5669752762540475,
 0.700667705220597,0.6318509911172898,0.5425202285371281,0.5871401556198006,
 0.2585587407237495,0.21263340114098706,0.23266481460443866,
 0.5361713678289838,0.6009408262659433,0.6998460223426127,0.6470849204187699,
 0.7479981574912945,0.3563997797585825,0.9903545031620475,0.6675706082062377,
 0.25163601098003263,0.08860516620939607,0.9332789743786726,
 0.21144893674909682,0.5633242447380101,0.9675965431686245,0.1050101379836208,
 0.15967095701929301,0.30153892164699336,0.5915257698329279,
 0.4747283157090155,0.030906366525558893,0.15835131840095573,
 0.7756927706978136,0.31324179925465034,0.3885936095557,0.23186736045779455,
 0.7062845927740142,0.8958660711492891,0.7729361301380434,0.7127505027843037,
 0.9377736870078657,0.20246049204235272,0.682818133923913,0.3364324403657377,
 0.4622234657367307,0.6072116716758981,0.8469874972504902,0.9925328401061948,
 0.9723727733825818,0.8637635753017423,0.27458633808895383,
 0.49908083006069415,0.14681477731168835,0.4504516438984948,
 0.9188876066274085,0.6080042767563486,0.24791306004014868,0.5160551817826714,
 0.1860701191049271,0.3560538701461671,0.0104947188269221,0.13026389061550891,
 0.096812148839173,0.11546047004936488,0.8313010391058155,0.8662145710023709,
 0.2548046173770011,0.26663518648564033,0.8166218517834988,0.5695398855524657,
 0.007354463529685384,0.9050839569945122,0.08832843275135926,
 0.09577630589376362,0.4253969088891042,0.12146563867002058,
 0.22303953681988142,0.5521403435654638,0.6187284810700269,
 0.30787004777196014,0.4452741863436187,0.19184250267668135,
 0.49310890080443315,0.47768413316713965,0.9897086433139257,
 0.13367567674044412,0.2388907842505048,0.367112827681486,0.9907578370999317,
 0.09840403366996586,0.5574364072922667,0.07950919824131575,
 0.015828195400884537,0.2736359707222875,0.4928021163423698,
 0.13249184223536126,0.699399654656572,0.09521652708135997,0.7267696085581334,
 0.504613437305743,0.47803186384329244,0.690719255753243,0.9452141389003456,
 0.6247345924749155,0.29408556105619466,0.10918467172557433,
 0.9700189725300629,0.27108151915707923,0.2164069278983497,0.8052184291145994,
 0.40069778986116,0.4565299260779101,0.6423830526753296,0.21755675412149877,
 0.45667908444515937,0.39096373620879143,0.2633371746150529,
 0.4862558069197176,0.6230555138273568,0.5067410307752449,0.8130398398139939,
 0.3172859306109751,0.44527739963115287,0.3046135694504931,0.2316569698211426,
 0.7589622905548671,0.5012596661535772,0.023612526367353115,
 0.7073479507921472,0.7338358952678388,0.8438004523916952,0.12850523785330226,
 0.7945789035946603,0.4294738750282291,0.684645178329603,0.07800471894245475,
 0.6040629607924364,0.8552446932014548,0.6066460491672133,0.7633193808319315,
 0.09726287971067848,0.466114369761196,0.700546856044582,0.966592038599338,
 0.2816659110410883,0.2902830887402925,0.6050761965311314,0.7364254900365639,
 0.1662106847744922,0.6069409633206966,0.24362615804362542,0.6305812060323996,
 0.30800461344709495,0.6449788113682904,0.3408195376015387,0.3726060004838143,
 0.8115051236181597,0.9799305099738395,0.9227429500837563,0.13400461121200713,
 0.26674907794649905,0.3307222624365671,0.49617880942254744,
 0.8206450083399046,0.7613201691607325,0.46551266282345416,0.3050985799230319,
 0.27769656588807834,0.1795869834009438,0.8293908174744471,
 0.37425887032830985,0.6797293920939,0.7053412381308763,0.8616086418180802,
 0.0771490147920787,0.0206570462826986,0.44502111247099807,0.467251927449451,
 0.8017081763791162,0.5783243233379208,0.863726907421317,0.5913473683733124,
 0.22133754805455685,0.1819610843416274,0.8922724797292804,0.9855668790067669,
 0.28713209626965686,0.8880435640769192,0.10251281869074558,
 0.47753372125181925,0.1319081168589885,0.8917012940288932,0.1968955325217393,
 0.35784454789389175,0.20642599493935543,0.5977885168344739,
 0.5704159800617177,0.5710638897608377,0.5119055125896437,0.5388731744132876,
 0.7405721357718807,0.7668289149756335,0.3751891173401496,0.35904368816442744,
 0.8232732520247265,0.19275843334876885,0.6074829033799313,0.5251029618366014,
 0.07604456669677728,0.3043021062558453,0.9155211625747308,
 0.30670953882138685,0.519486632828473,0.6259310165265063,0.7079056405504246,
 0.5266825797969401,0.27046802319240315,0.986894504999847,0.9153382298029225,
 0.1347544376278359,0.36426597812052686,0.23007532582066115,
 0.7935904892185512,0.4984949048816889,0.7932445522390928,0.1480769284983976,
 0.4552588082993646,0.060375846136996314,0.858020140369399,
 0.26049699100927737,0.5543116099048082,0.44575760996835834,
 0.0015659249749663262,0.6522718482939731,0.7094306391393156,
 0.7183922090069348,0.34871268908448805,0.1831564373858966,0.6310646576429673,
 0.07285983933290874,0.5556693068043571,0.08543110654908925,
 0.8205246457260397,0.08293845483920737,0.808730196674152,0.6006724386064356,
 0.5613457221739594,0.5199446272867757,0.9757678447291884,0.17913959228894294,
 0.9998388255250774,0.9729264791845771,0.14389284186745366,0.5524567056259594,
 0.5774075017236431,0.7530138638152383,0.6516644426776171,0.9537139358297404,
 0.10279446151474758,0.6607000752324372,0.9466667626946459,
 0.19672855716442572,0.04711489376346434,0.28616582259474,0.9950756850513023,
 0.5189509041241908,0.20119842811640298,0.00877554312214568,
 0.6884760291010679,0.42209821279117654,0.139783048726415,0.8302326635663816,
 0.8070434739627804,0.8748087742142523,0.19729506307305877,0.5190684281975277,
 0.328880214443234,0.9591131449462291,0.18462342642153262,0.30882234396191377,
 0.11605740153488919,0.37054584023099857,0.3357537779337292,
 0.7203978724187723,0.8285384422449426,0.3552763510824718,0.08033734002212023,
 0.2709501629223279,0.9754370504835338,0.8946694269691097,0.26618782791918105,
 0.15180980193128413,0.6968385226049315,0.7492937786914813,
 0.05629032506835974,0.23815328719445994,0.38024277997778366,
 0.6341400442595008,0.281197337761832,0.6985459156302425,0.2568555140289135,
 0.7997430084806159,0.6916555007583531,0.905025074970491,0.43548565204486345,
 0.06260022114577146,0.0020338684593301792,0.8014127977377895,
 0.4593304990165288,0.47006826028516824,0.5272352561954773,
 0.31219228119796627,0.40353528171166464,0.8326861654986736,
 0.2315285602277426,0.1905408404880422,0.2743499720959828,0.7510833581553515,
 0.051996095422322974,0.4638652830841199,0.55322125502155,0.4769318586046982,
 0.02884572120407869,0.39306667640132387,0.314958596764227,0.8833204840454619,
 0.8189805311793052,0.9098453673486049,0.19564744165925063,
 0.33822523116196834,0.8111086355570643,0.40170857841618535,
 0.37820260212649237,0.9657914400227382,0.15020103551564878,
 0.7903401651104174,0.3235052102332949,0.2909156787677767,0.9607623983779309,
 0.9672284220230322,0.4491951815367843,0.9309849454350522,0.6077764192491745,
 0.46835734665296735,0.0037536534421757084,0.30746952289133334,
 0.8178188958137891,0.5238736789056797,0.7501288613547006,0.5434114935615548,
 0.841411048538361,0.8422815576763665,0.9406302377729598,0.2999549449656914,
 0.6653342070016663,0.7911242803677075,0.6511623870360639,0.43604632837087365,
 0.14324373142386626,0.5106069272045076,0.46295851392211373,
 0.06581285361269684,0.9453314762124391,0.7680196777324542,
 0.03400288437925192,0.3233329746077602,0.13807496556350762,
 0.08869401727321846,0.28576396287467953,0.8165706289692334,
 0.6300372160106813,0.16595898761362482,0.598046108128032,0.06187472387551729,
 0.6620757482846097,0.9389167789971736,0.5838795818827313,0.47622657409308167,
 0.560739066852903,0.3367389175968334,0.355637924757631,0.5290679627488919,
 0.03208344315860223,0.8823201623471911,0.5001572484224447,0.402740697950025,
 0.5696244677427376,0.2426291395048945,0.48794064904823364,0.519282269525239,
 0.068653018349209,0.697490700133385,0.8579978567064009,0.16329622050830783,
 0.7597514915556456,0.43793029892143587,0.20466690841679602,
 0.6933915923900549,0.267850978682192,0.20517426523033777,0.6059670354152836,
 0.4361890441203764,0.2052998793255263,0.5259846971365314,0.23279399758657393,
 0.5219572778259258,0.4338358565540993,0.19727746462343398,0.15378737830278,
 0.5832131703953619,0.2921499693595577,0.3687534821442793,0.9008482517225932,
 0.3244733209153323,0.2998115235027117,0.35421742918809995,0.8429192228159985,
 0.6603933035945615,0.5330019337045488,0.9054588734476235,0.4664301994359513,
 0.10721238178757542,0.42790186723447,0.8433709076899809,0.8856207594997848,
 0.08872135265799186,0.417448662175781,0.45365864063813954,0.9923126040999825,
 0.8823941326328207,0.08027691496536293,0.3104129353898838,0.5253078789083759,
 0.9998790999880267,0.632493927370631,0.9701186801991005,0.95722116379529,
 0.677768998505478,0.8844734097644393,0.8139957685140568,0.6090005075447696,
 0.09533397831076917,0.06890996916250702,0.09387185915334484,
 0.6398806647832547,0.7342167761789893,0.4278917005688099,0.5298584677409371,
 0.9583434789498944,0.12415303924432419,0.9852728044830957,
 0.30838593980933127,0.5806354998684734,0.6293940940303363,0.5086493231700414,
 0.38537324661054384,0.7068978552622547,0.9875422245317572,0.751634555237468,
 0.04857704959503606,0.4397598577771634,0.8139581367808945,0.4031757189115097,
 0.7478494649339308,0.33240550406110037,0.7167253626166309,0.656740122553337,
 0.8416191860345639,0.7150976413138198,0.3297414645984039,0.30321407481496865,
 0.03129562988490775,0.5001809597644211,0.6436710125782122,0.8941845232641991,
 0.9434713395097668,0.9620208290143203,0.11239821665635141,0.6305224658242603,
 0.08853474602422107,0.5806458685590364,0.395644146672803,0.48800224708582585,
 0.1088341471192329,0.11739816571726669,0.06745777807631737,
 0.42447099041767666,0.7893221417658365,0.08259112486908915,
 0.4827655934239473,0.009066655454396999,0.840585243807476,0.5220139175878931,
 0.9826473628876156,0.3830249719681865,0.1032306960767263,0.5644891019083167,
 0.07549064235888503,0.9796705115104508,0.6338016094770701,0.523523568062412,
 0.8478531589821582,0.48507068514564355,0.2771690476827656,0.3964252982912073,
 0.7116622856237865,0.5744774052055389,0.1707567689228353,0.3059933127621862,
 0.3416426554069547,0.19101330286413287,0.3581329521859826,
 0.36367402430648255,0.3840589775440778,0.4114237490354006,0.7153564958193266,
 0.761658673122624,0.43465199604894644,0.26732675200558353,
 0.08968928253207487,0.29681837608331185,0.8370517861188744,
 0.9976023073549458,0.9985905875525551,0.329529102742402,0.6689323645405598,
 0.1398191571704459,0.4167418579172447,0.35923452508059217,0.4858149031241661,
 0.5875950420327921,0.1608648827136695,0.7108879018773788,0.6741923252249871,
 0.3146836403337242,0.30572322178128886,0.20969909244658624,
 0.036078906689258505,0.06328524668833868,0.035089967232702746,
 0.11695288486832778,0.9744234413218066,0.666730924129285,0.8009910521005135,
 0.6026692435916206,0.6176071529641549,0.41535866791322895,0.5199700087004518,
 0.02284890053509403,0.1125242364154635,0.6612453695638686,
 0.07421799850758148,0.596140869676468,0.07058030483972111,0.7299655030342008,
 0.12322702647690886,0.07249026608267295,0.6143551507699709,
 0.0803705348607181,0.7575475750374211,0.4922781131917857,0.1491417767705443,
 0.9561042315701139,0.008656696520989215,0.22077769860139562,
 0.7761623414122922,0.09819517612254236,0.8357402916354971,0.8684754557324619,
 0.12248606815520446,0.08975334482250663,0.7954352526250407,
 0.3765104009288286,0.6737214785158606,0.9367603838566715,0.12404070121911526,
 0.24156302006945785,0.447833721324312,0.7004402617082832,0.8262799686260422,
 0.3212839255640937,0.2079144644225157,0.8289986703384504,0.12848233202199033,
 0.003627035327787409,0.42158198784604317,0.8689638680795843,
 0.7018702325703302,0.6561711246546883,0.29697212517455573,0.5759430559883958,
 0.7752218774127297,0.6261672832701224,0.7889207213972085,0.7293085525352817,
 0.9930629694362901,0.6153212713817174,0.8102987846111439,0.20705083365059684,
 0.8026959992264818,0.35907209959380004,0.08421546827142201,
 0.8515243464272242,0.6280354195505694,0.16838343579418358,0.4347184972233954,
 0.7596185229822235,0.8568723889969305,0.9230013543166594,0.5508721038804507,
 0.6679562949521963,0.8142539047893379,0.5855070680544467,0.5172935104918447,
 0.05934089860344938,0.049573474292737285,0.7168137358339436,
 0.4591969256669326,0.08343296810313494,0.3717569183597924,0.5934192500324136,
 0.581633117803355,0.6699168669587447,0.29392299890479867,0.05268988501832794,
 0.9825665468382776,0.11079491828292176,0.8077920000836488,0.2896707382792858,
 0.7202342139348537,0.0649424927209793,0.4330001563441017,0.7169075279682972,
 0.2351250769107116,0.6711289176563249,0.40335288291383464,0.5550198582524344,
 0.38662031683586107,0.8118425001884116,0.3054536325137148,0.5042836008394534,
 0.8931042739406756,0.4824757046951226,0.9294665667730294,0.09323512569381853,
 0.3442648027943369,0.245455707636957,0.7275783007809863,0.36549912854204536,
 0.5648055719860609,0.9525644241108365,0.40758088756634714,0.5302484588328787,
 0.2827740799480635,0.22241380291566237,0.6838567559854976,0.6769868521769923,
 0.5763089910229918,0.8522341715338647,0.33222117392711326,0.6145172485345856,
 0.23747372031720382,0.6993842750671457,0.5311041958359942,0.945635398348265,
 0.9695078667197552,0.9921138720488922,0.09029680645314675,0.8963227424608334,
 0.1814233412080175,0.19376619335788314,0.6152471163041083,
 0.28700694596878995,0.6394573648271735,0.41000203397149726,
 0.047631555583178886,0.294785090988416,0.5227782982015919,0.7805609163949981,
 0.4531283131073045,0.973335247944143,0.9485754918450421,0.7619223502861526,
 0.2540308103084402,0.538699243724271,0.6643091735739763,0.16212698187259866,
 0.7961230972715976,0.866178887569932,0.12940997348968508,0.11704441349943173,
 0.2845364704419203,0.3697176409088191,0.6398040509132207,0.5214417040314072,
 0.40203473776924303,0.7366556914138438,0.2742038479679292,0.1530430346886248,
 0.5073156571396504,0.7837977917170793,0.9733516764992715,0.9971142136038328,
 0.20051130958599606,0.6405349300423264,0.5622825038512316,0.7778121417034359,
 0.4275223607339125,0.7774861184357051,0.5056178914024712,0.6400292527998527,
 0.6355199258353472,0.08385373032130872,0.9802319958733983,0.7231078095835992,
 0.12081736772454266,0.7763884885470339,0.4387666496908179,
 0.18747746705904167,0.4140819212487574,0.010256972521622831,
 0.8949407586646653,0.5713274804021886,0.28218909994811814,
 0.03557354165019888,0.6231426522091028,0.4313088234282618,0.5090842267363268,
 0.43967381090605906,0.6776835374766239,0.40889207979498876,
 0.15643607744405497,0.8089027430041391,0.2017045036090177,0.8340306396596462,
 0.7476968826749214,0.7303411260971691,0.9012318844654919,0.5766887954031936,
 0.5709480664019564,0.9689934662622743,0.1677174018366212,0.3326673094726238,
 0.4448917293280322,0.10638357731601766,0.9059513400335688,0.5136473005272856,
 0.9034967700807612,0.2828972268972897,0.14025799774202985,
 0.32504697520174375,0.33858403881757926,0.08413132852267369,
 0.5563727836620022,0.9864822355934347,0.0015068833771119117,
 0.9659986919178721,0.5483519177708774,0.3858297646492592,0.3804944671973034,
 0.6146620386867487,0.4994957043067342,0.6223446581535668,0.542591147882937,
 0.8302307182337334,0.39634575092936986,0.11486442017865395,
 0.05591221777504807,0.370480408332999,0.17582348252346214,0.341498127917893,
 0.9189973105531263,0.5947026129067985,0.7151230987409267,0.8145497901675397,
 0.8756946294632862,0.5245341918475992,0.6369317085000518,0.6452581733215396,
 0.2690237017029963,0.9533475932942316,0.9304015030854461,0.5234515254199699,
 0.7995813583232507,0.953837547607241,0.02116951454503424,0.9473844955710289,
 0.009090567394731797,0.3177613529649237,0.7140205204386086,
 0.7113958308180511,0.27886436109811563,0.4845786307587959,
 0.02315289180035407,0.5606997287152355,0.8818650948093562,0.5428802385221888,
 0.6965121643244313,0.41338531008007773,0.22349552162809827,0.761289228354777,
 0.0818156504207147,0.2574721849306788,0.47578989455101517,0.8033430533989028,
 0.746028452489848,0.9140168662154914,0.8816148463843716,0.34393916154521076,
 0.7993353195822879,0.9702828210486918,0.9290229817264574,0.7031869091726765,
 0.9714103967900961,0.7013836751260945,0.4197946493535152,0.10905722802426732,
 0.4758343485404921,0.2986462024412778,0.576448277772245,0.3355352888750469,
 0.9070678727928979,0.7583801632475737,0.3582342577639447,0.6102862622178201,
 0.6490663632975437,0.8580404786366445,0.08217892457454201,
 0.20020544436075693,0.05509294864456116,0.5111978597391882,
 0.6673000259634669,0.0067249595722777,0.9582052156599858,0.7480745916544014,
 0.9114457428049765,0.1833424652892205,0.7819373119423156,0.1473508563372592,
 0.9009384310170248,0.9501111094779496,0.45744692705819734,0.3757991293628269,
 0.9062254319987342,0.9586206971373122,0.980601079028642,0.48657878411744826,
 0.43210809662628336,0.2572641577817085,0.49799507772823903,
 0.9263926961169446,0.22596127793557919,0.5786304401864559,0.9134303882651695,
 0.9515445109920432,0.5619646082714094,0.7934808971893471,0.133042441883792,
 0.37969244158850934,0.6865858014481473,0.17744579254731507,0.087510301781484,
 0.478228174255804,0.5316883117710193,0.43793652079034184,0.5041461785211299,
 0.27799909611482354,0.23692920614695823,0.21720507192427274,
 0.1571692541111609,0.5213932080219079,0.41389364051991895,0.227665027151433,
 0.16028025891644337,0.7278938110993384,0.5326030746077749,0.9566080344176555,
 0.7683903645254728,0.5167354840911208,0.5410241264398438,0.34089216847472015,
 0.8240945035840103,0.7685256412989745,0.613079425760256,0.3178930397952333,
 0.2971548764194526,0.5414949808696641,0.4237040144159414,0.18020193934942796,
 0.8928234371404181,0.091803688280369,0.11966280628200898,0.6026432336093166,
 0.2455179768164284,0.07744357291988457,0.25417385346139576,
 0.9844795249082272,0.9532031008668871,0.8438725998610918,0.32697380256368436,
 0.5158585414088562,0.8907682817413842,0.6003182099656958,0.5335676482273047,
 0.3369870501068277,0.38088907377192094,0.23759596828606422,
 0.7726100677621525,0.6858129784176322,0.040659362177446035,0.913979494764576,
 0.31501668831951624,0.31786928570065365,0.8186517902060305,
 0.4073195841496632,0.5517144830435747,0.1160161788242825,0.04016363348555552,
 0.9198357376271226,0.43526791699878975,0.4114233557086956,0.6626573684243015,
 0.9003246252346193,0.17663579503722682,0.670154523585838,0.22932044185924283,
 0.832924349183013,0.7613432005231516,0.21547097421842998,0.6998689819814837,
 0.3030615081119672,0.5757303668461169,0.9097062315450567,0.29105872121181864,
 0.7802940567405776,0.7648323745489983,0.03133496433951133,0.3965531614814449,
 0.29238692773915675,0.9501033601653542,0.9759450936176803,
 0.20316562597296495,0.5919190085936094,0.6227029912563811,0.9652825634675513,
 0.026305655260120853,0.4025032497379041,0.9162513624552825,
 0.3053725241140872,0.01767139660463879,0.36878729716725045,
 0.7095737554940256,0.8466013119473295,0.2506658677269554,0.11073165653837813,
 0.44532242404293476,0.35856744526444373,0.7651905660818504,
 0.10003554364640843,0.41315778812242177,0.4528305769961848,
 0.6267310135992024,0.8963607268693077,0.6846824147595976,0.7483233064363487,
 0.7128261569140256,0.6123721363084501,0.5682584740873841,0.20545677789182049,
 0.9374596312338388,0.17719102669858744,0.7192885501487598,0.5280063027132744,
 0.6607227933399707,0.13357638191187404,0.5329375820770156,
 0.18459815138078983,0.003825999906715505,0.8253715574197722,
 0.249672973065294,0.8328474306829106,0.6681018178618428,0.23475782469078355,
 0.4734855677203157,0.7889307150766112,0.5433840421474616,0.8477049934521674,
 0.5462905494312028,0.1605455023894098,0.18136041241822787,
 0.07088060287881881,0.3368450088051611,0.4459603293629668,
 0.17339978480672225,0.7610816261032989,0.8408992854254385,0.6321361129184653,
 0.6892842540296106,0.7774601718884167,0.2853123054352835,0.9701877454161432,
 0.6650369757007271,0.8235499949769479,0.6788134690278518,0.5708317002618172,
 0.5993725758331319,0.3907696945766803,0.7127660429313979,0.6835759228431912,
 0.43069432059154245,0.6990002383605135,0.7595669924781991,
 0.18565750735648867,0.8000860299446989,0.1532704772330068,0.80462336993223,
 0.2676370680706689,0.08200076420483049,0.8050660979371315,0.7233147939829442,
 0.04787110097205616,0.05645621243916055,0.6955539448175577,
 0.22324997623351905,0.5120954054504816,0.4725619774165657,0.6177709155718186,
 0.6002626593436634,0.26949366600126534,0.17440822656731958,
 0.31699627580592993,0.40047167779497883,0.8237289226439706,
 0.8688089461269991,0.3082201666969885,0.29075703559057253,0.4657345189782235,
 0.5147723286312172,0.16248144424238842,0.29421419481164823,0.782409264318482,
 0.2130407610198939,0.08331750814021499,0.18825582297304133,
 0.5851888311149767,0.8803929414036554,0.9947056353979835,0.01553800601139832,
 0.2881935628743306,0.6817363964514883,0.5315690091921562,0.15303954284639643,
 0.32507601673137565,0.4308593177962611,0.5322972278700641,0.5898055360644304,
 0.48924500536240323,0.1422907293924086,0.4693617226169232,0.6618867665081518,
 0.6175259247043599,0.34249101999212916,0.023881528154221288,
 0.4635655853929548,0.9021938277199326,0.30675109070358186,
 0.021369719286540967,0.9503205098862465,0.6683058633921828,
 0.008653409865921002,0.6225337953633399,0.1293205458430282,
 0.09752639020837406,0.05627665246477742,0.5579616223066284,
 0.33260786369923667,0.5797414819352705,0.45333547205340374,
 0.6444351669144299,0.8158807308367398,0.33841057517367124,
 0.06627431340677148,0.7699846622893654,0.40928647293276177,
 0.17783853458009224,0.9403150250762015,0.327508890508716,0.8927503552429413,
 0.7439771993851316,0.3762186057899529,0.9858083583629216,0.549569992993824,
 0.8279886193023338,0.3403222002685108,0.18948118763655275,
 0.44896028434619095,0.002229427689891872,0.3108955504953642,
 0.8973343592284124,0.10791494090215481,0.4246375499229831,0.5544883797740356,
 0.9454240015250638,0.01619185204820306,0.08502738105036323,
 0.1756890957345495,0.9828653062517588,0.21386379932292654,0.4833345768155266,
 0.6687516645454437,0.490824498452923,0.13640279844636072,0.615640510833638,
 0.41007164035454025,0.8773306277876946,0.32694561329128313,
 0.20395791198153468,0.956427098094722,0.5859119463072294,0.7934607946091511,
 0.12448440357012114,0.07126764950100029,0.32076231770602637,
 0.4840250284046589,0.01591955056043115,0.1974060404541199,0.7212339587322654,
 0.978201422954687,0.7690641472894569,0.8285728276773815,0.388320315241238,
 0.8378547057670048,0.9465072656003959,0.6899335600294926,0.5809845498573986,
 0.9684948963979074,0.6196299775089071,0.8717645269726076,0.1545861604776474,
 0.24399170279406013,0.5258697745420001,0.27120264696610963,0.919390032028043,
 0.2877390538203738,0.4179001637286911,0.9809636392906337,0.6928036592631758,
 0.7108280088469194,0.8591121034839099,0.32424833220993476,
 0.16391746154627074,0.24579237780058205,0.8376194296730722,
 0.23360338636706757,0.7994058751016974,0.26304480603758507,
 0.6848920195592102,0.1678656841595011,0.6226359843241418,0.8897639064620108,
 0.7162642539556195,0.22977310829222253,0.5466697413273038,0.3414955573039379,
 0.4804411823857022,0.8201342497743629,0.9820753689432236,0.664147589402267,
 0.5201729411075366,0.5797745061422507,0.8498287484664951,0.9346725497012597,
 0.7383511673574437,0.9450150466939231,0.13936899694857707,0.5177225397197271,
 0.7063492431729915,0.7396552202777987,0.4305464747124602,0.8668427459707408,
 0.5881826506149004,0.6035468587129205,0.7492723421243904,0.11941568394767532,
 0.11166707390317532,0.06799663094426478,0.3692740829522133,
 0.5850535768249311,0.2622686360736406,0.6155403150290287,0.6149537465543486,
 0.6369799516055893,0.9276125490737999,0.3951766071281152,0.7304867722171972,
 0.5857880974600862,0.7980495035691413,0.5929676175680808,0.6341987264022075,
 0.31621228648624566,0.27283962570032716,0.9935031782773979,0.937837594010574,
 0.5853377398288927,0.3936805369403319,0.6217682839056514,
 0.059486710993731595,0.9953665449580963,0.4369191290471456,
 0.15132298420977885,0.5566992255748953,0.08134503244933478,
 0.9495677560913485,0.03485677527379649,0.2483582596863516,0.5949577162268737,
 0.5303190373961013,0.8736617520424846,0.5873181268109736,0.638464094888039,
 0.2027513409180114,0.3627408769921091,0.30328583348381477,0.8091187813980198,
 0.8955922165064601,0.8187935212365441,0.9863358350429468,0.3508234740818089,
 0.3386479694922262,0.9289151723525212,0.7464025574351261,0.9965828576767474,
 0.9092991528229611,0.7687268199254002,0.4931368919718925,0.9351111834259483,
 0.6807040329824267,0.04181819654435781,8.565570795662225e-4,
 0.14362106360340587,0.9248866529292329,0.7807315035825099,0.6119359502370654,
 0.7513588761591918,0.929400299699276,0.1604342568392071,0.013136895840499152,
 0.4047044348063814,0.03458315889575858,0.26033433762156744,
 0.45273309351154933,0.5576639560202535,0.5639249860447837,
 0.02372294890181703,0.8929560328081476,0.10671016288273394,
 0.03336261922224748,0.3758406505454722,0.3211122356659847,
 0.49431289648721355,0.33480453950990097,0.4577175983647247,
 0.6865192134211715,0.40368980570784496,0.7854792791114229,0.7153389305798304,
 0.5812328337624099,0.23941644117712302,0.4670047296426717,0.2295254037426312,
 0.9208558502888353,0.9121999408124088,0.3215151454409414,0.812751922161284,
 0.1491023818474606,0.6973877508596669,0.2512646951465801,0.8387250776210674,
 0.3092777835015619,0.050738455175425856,0.37746717988088274,
 0.7051310396572558,0.05136043215626518,0.03547617419471372,
 0.011526548601129893,0.2957240347726561,0.7218654706983041,
 0.28921401009125647,0.6039791484878823,0.6362971814952026,
 0.29733247975731525,0.571916368870206,0.608804002605765,0.9970354161913009,
 0.6216081260996169,0.5307083594757572,0.5417548908895802,0.34479401585361324,
 0.42783606386799466,0.05965213027443439,0.6229865665551811,
 0.4045875703481584,0.0025389600378256105,0.5587644332354499,
 0.6464054799243507,0.6728589257325921,0.2473945372652533,0.06452306760626236,
 0.3382484573640461,0.5720363172163311,0.4304664137001476,0.8042297829286361,
 0.6554793489643976,0.2347788440442713,0.6450352176991608,0.37274106790339134,
 0.18232527340864324,0.6611170067961443,0.9069076799432839,0.7343027996240757,
 0.2907137230890696,0.233747588676131,0.13904408428817217,0.999399546044915,
 0.30329508048476184,0.3576984040560127,0.16802956780662548,
 0.5291392375790259,0.6956076381444978,0.789747818050526,0.7990691532581453,
 0.017862833468649586,0.7234312263827484,0.41094589525950154,
 0.7745498359618365,0.1058184362104726,0.06330201917407074,
 0.23023216759466247,0.35607086001971133,0.18753158807147474,
 0.18957415480539774,0.6438668757314274,0.7296865593659395,
 0.36552712501731754,0.5456687421798772,0.17688250911361625,
 0.49453740355202813,0.6283749890593234,0.7320173520627287,
 0.05730619629907885,0.5856775134563018,0.07070074130555104,
 0.2503779685502804,0.9105395443303057,0.7048752420688459,0.268048096471597,
 0.9717585061541923,0.25835384381713444,0.10045834108443885,0.939668782052244,
 0.24314394680241136,0.40604143294839445,0.04750766594595113,
 0.15470333071938702,0.19792787793403788,0.3069190319216868,
 0.16463941335056942,0.9303226667818462,0.063514615336552,0.0892970691606717,
 0.47998665611925917,0.8055690860429823,0.34852201677839245,
 0.18404979550337763,0.8808509724623182,0.9327414406403332,0.3045335547039314,
 0.459587612069122,0.13918147718536755,0.7218623538638717,0.8153261453030995,
 0.6135537341780084,0.4988830072576992,0.7745440772967739,0.8725323163997883,
 0.893365589393814,0.607129513537715,0.5717323900498716,0.5154407281052787,
 0.41181798181644125,0.5068602926098842,0.24215898381164158,
 0.2530248552140022,0.44346540077169505,0.6669282003322796,
 0.17369350026451047,0.7382842801535522,0.5407070802862245,0.9252412316577379,
 0.7298838423022069,0.2532943623877282,0.09994170268545632,0.6146434622500367,
 0.16083130051674988,0.622298822624865,0.17581801581655265,0.9023280389007243,
 0.8376542408924068,0.8824177866409968,0.42424069905604334,0.9544679315970139,
 0.09069039473159579,0.6488180021646546,0.630845394924405,0.11278604914886237,
 0.2791159759842513,0.34265877899811914,0.9229195756987323,
 0.19945323628093414,0.6672672380528565,0.22970474921513606,
 0.9293488091665936,0.8585322044281838,0.07375342087716208,
 0.21784715342798477,0.6018242756322961,0.8992312611906926,
 0.004641157847432353,0.7523078859756482,0.0813230800461735,
 0.33120340595824516,0.42086561319750415,0.7944632577134771,
 0.2802574080501863,0.4852076957497238,0.48276286096413223,
 0.11191344463792119,0.4792629880354342,0.1521717092106053,0.5057459723826871,
 0.5721874253284513,0.9811396245575038,0.36862046558564343,0.6585333025663724,
 0.26070109404027564,0.31608905069560234,0.5456637360224655,0.60590360967558,
 0.7919430780750201,0.8167138954715408,0.14171796297313388,0.7852876164598676,
 0.2677377161735719,0.6660399231798508,0.5783131774644191,0.15222709165101334,
 0.1578373961420143,0.28369981496268093,0.9373076172417079,0.459063216991216,
 0.6524756240110244,0.559231822226909,0.14132164222574373,0.7018061602666571,
 0.41966158296485523,0.5945594651189279,0.006516924272049862,
 0.9219816964402652,0.245727690261334,0.03676364428327483,0.2883628023644925,
 0.8549422250726603,0.61233039697144,0.34698885196430185,0.6554440019193063,
 0.9701008476406053,0.47684930162066674,0.8655428500539365,0.8508774596334211,
 0.8284408375055229,0.4294138138499374,0.2085698619233004,0.3870789880113815,
 0.7030757177931055,0.46722661544243804,0.7160198524795544,0.7662333923283298,
 0.9228631497587025,0.034620121251150326,0.6473225300955575,
 0.1013831302285898,0.09642321851833735,0.5424125645907856,
 0.07128370901424042,0.06207544063642223,0.23232751360468262,
 0.6990435905831928,0.189111067965974,0.9682441225581535,0.9411522731647013,
 0.6454157453963562,0.7338977619810363,0.31090104916420724,0.7558243409446398,
 0.03671627224612062,0.7375966864750155,0.37496218535531955,
 0.5260341962226338,0.11195207286956677,0.06919091174460767,
 0.5726687958639969,0.228087020569275,0.20448465124600956,0.33625470929333756,
 0.4647456461172964,0.4498595302245455,0.7776076801905909,0.9211522127296073,
 0.8315649547277176,0.7281084666800323,0.10040711379042622,
 0.10772145503336905,0.06804689521168772,0.16904303192211811,
 0.4361154175604621,0.7976731784727236,0.7078673566551076,0.6568035342177223,
 0.833266513800525,0.23549027484970786,0.3259160188961542,0.46426741649411873,
 0.12755006790264534,0.7822748505874692,0.2227339040228412,0.8735041591474579,
 0.6801570007809026,0.2497520525072895,0.9507267013650711,0.8415056438066113,
 0.42461771125379644,0.4362191321430494,0.9539198881570854,
 0.0018550677535074112,0.8622254960968336,0.5975018323469831,
 0.16358484459528022,0.5806096212238997,0.608909517850917,0.16913595006944315,
 0.1287130344094275,0.8579490298803326,0.8712272408436832,
 0.023419829460200048,0.40913107334906085,0.0571599870738444,
 0.2477193384952967,0.4397164586920269,0.8653914000271677,0.2850405989716359,
 0.9446665995064173,0.1211101549665915,0.12372612082956369,0.694727949875789,
 0.18432943101322963,0.09004348512010263,0.78950311486425,0.5330369971282806,
 0.6564096751439323,0.013977726930203938,0.3186014747668323,
 0.28833334761114815,0.34464608301609334,0.17773501845823114,
 0.11720659666217514,0.12359036302895987,0.13096791686317544,
 0.8350786098662835,0.9622352469962115,0.5040842924478767,0.4923060513657348,
 0.7909413335678511,0.0013239608779067602,0.7266493350322689,
 0.3918291941863091,0.07207372355925157,0.4294382671362663,0.3670667366980932,
 0.24650187150015967,0.09033869317378507,0.48893521433934306,
 0.0350675540712132,0.284630565412582,0.31561120230042494,0.7746770892502921,
 0.8052481482949354,0.2027014980763694,0.4153721304154604,0.5761872264408643,
 0.15482851260347896,0.25917513437038797,0.3378666110733918,
 0.07820994007108606,0.6844273295739801,0.9284161257429577,0.3429225643871143,
 0.6715826619024992,0.8907454438897433,0.7414513492063798,0.6982555789327923,
 0.34599111141312866,0.28792854833262516,0.667429431562608,
 0.17099614880191116,0.5914303061934575,0.49095986112016987,
 0.9298088690242905,0.5655327439128408,0.021201678026303572,
 0.36551293046359457,0.26072054160058467,0.04946755521842372,
 0.9818066028271973,0.5135471669195874,0.902360568900358,0.4860695184378917,
 0.35115404689472507,0.13142932386187356,0.3915789497421789,
 0.8615235554704787,0.42275809498736017,0.0834551708655259,0.466932104418747,
 0.7019936935075198,0.7256235171776446,0.27298104188279104,
 0.34123983460898744,0.9330082244476321,0.2985785889917375,0.604291141229475,
 0.30459388224729855,0.26495001529774087,0.7785572790732972,
 0.5537789372287939,0.568138808521403,0.0022194826146506585,
 0.6288767471833939,0.5933076355972844,0.12387419803601141,0.7679959542361018,
 0.3570282055312337,0.6042400353030846,0.3875639114789756,0.7868761334558143,
 0.8704670775277396,0.5531064447375804,0.8342878342065095,0.6788511912410217,
 0.8607622838815246,0.06333444417463507,0.8154968604597239,0.9494302791216948,
 0.17593020968815154,0.38488157228931263,0.24653190429537497,
 0.28371575786121295,0.6098569757813574,0.37487159668861936,
 0.5481013376760464,0.4379666308941279,0.21097265961362333,0.2453422376485167,
 0.8158015988623432,0.8225080014896258,0.048823078695285016,
 0.40138603862504396,0.7901823038630997,0.4407253932362074,
 0.28489912053491095,0.24227683896683427,0.6239042027851663,
 0.9526845472565855,0.549048384350017,0.5963336389574749,0.7215836438759059,
 0.6643187105396092,0.5575648703648545,0.8455506633547236,0.5084452373174662,
 0.8105134884822615,0.8594455323428939,0.6964205554220919,0.5675740812978974,
 0.4696802762303156,0.2545604150055054,0.2752130538928287,0.6859960068270026,
 0.17508961570032977,0.037313818117668984,0.34924255233217516,
 0.5336254348590586,0.37804298327188557,0.7315352451176803,0.3467936838630671,
 0.9135933092891355,0.6222987363981652,0.6697864744292009,0.0857900498550902,
 0.8014991252970148,0.5489503309794236,0.4783117204263866,0.41901257135890146,
 0.9351596931753463,0.4470114114879451,0.4068996360676296,0.43741437380355963,
 0.27666958443142897,0.8541030426862847,0.599099785468181,0.6499248186217638,
 0.7349957263745988,0.699206021736704,0.5187911319934543,0.302569839740202,
 0.34416732822176077,0.8357204357224769,0.2443083781941543,0.7323131644008322,
 0.48912080892297916,0.7919062678007419,0.40917113784951376,0.429230208437435,
 0.9108993451342605,0.7288790182560486,0.42502639126116026,0.4377475847704707,
 0.45780875122189224,0.7014624743125658,0.6303779666541459,
 0.27100816400477123,0.46568593863904784,0.9280163370462925,
 0.4291171170472967,0.03084029950167899,0.9820762536753389,0.501597328152942,
 0.2869774567957226,0.5509063235220694,0.6627216240304306,0.3629005536579166,
 0.18444236641499756,0.9358249623783786,0.019116962940609494,
 0.6760115379133123,0.15307997901345605,0.9056768469472392,0.63686291568083,
 0.8713870468756965,0.5568276020367144,0.09050695034222334,0.1620167224994633,
 0.043760960401969085,0.9483328621367128,0.3862073561683913,
 0.7876587320598044,0.9668852905725643,0.6985093616641327,0.6797740357639023,
 0.4278218738849201,0.7294950345437519,0.3754398978311968,0.7194183352297819,
 0.3091842637227553,0.470049695706363,0.18387823761124145,0.16717750100477224,
 0.5623914411666341,0.6252776506220163,0.245493720372836,0.5773439123947041,
 0.3354168117933298,0.41598193497700353,0.41542393877913963,
 0.6266218911504069,0.022732280845039643,0.7279194375491671,
 0.6778297972972235,0.8917431229490315,0.029185448308761686,
 0.9050216352841725,0.8482568252581512,0.6352993997592309,0.7886363376331321,
 0.4303577321867198,0.4713236173353319,0.08326764561955846,0.9400054116807177,
 0.32570314847989545,0.8043555931432658,0.9329296656881123,0.2928292452555614,
 0.3695369760518078,0.935726073687413,0.602803982507608,0.6881202018162716,
 0.8131075525518813,0.2643775926971167,0.03598996759204465,
 0.47830231706770787,0.6435178197125292,0.8886752516809228,0.7978436975079667,
 0.7343356094229412,0.17508464002500557,0.9200406655665705,0.6062143405629105,
 0.6940179304162378,0.32669833578608987,0.8858463606680508,
 0.47186840668268126,0.4468828348988152,0.5884193879443137,0.158787082016978,
 0.9654055286304267,0.5401182505338691,0.2724362465266845,0.24805767870261064,
 0.5729736026933129,0.8299288530873218,0.6775351677519144,0.7984453186505013,
 0.39224712430422504,0.49599294531595495,0.45207532123261285,
 0.7316452729682952,0.6167946927518025,0.2035795635364801,0.05759010918669172,
 0.5463926575420334,0.9857430225979553,0.968169008773947,0.6035941032204226,
 0.8295630139928456,0.858666683762189,0.6836567527788776,0.010832132077351009,
 0.20004671028575138,0.8576806371512541,0.8829168314414991,0.5563739687885906,
 0.5219867215864962,0.17965579904263207,0.3012791893512665,
 7.054188314121745e-4,0.23687520967991949,0.40183502146618,0.3914006517729902,
 0.39426076032437707,0.5932783168665141,0.10527370347163534,
 0.7777351899000989,0.22087341851938436,0.8324679827225195,0.7352902866879463,
 0.43008426750153106,0.8509912453142762,0.44114632588123837,
 0.8876033199901298,0.5397613861420497,0.9565225371881971,0.07120427840603605,
 0.4553237954545818,0.06415214762982313,0.2945143918792166,0.790126126996002,
 0.41300657924638073,0.012165471521081161,0.5325068742128972,
 0.01064359937614956,0.4729140339212379,0.4861641587502862,0.5365257557739279,
 0.9371512214181035,0.8460870819832086,0.6597480339760851,0.8681790617185378,
 0.033692988855330896,0.6732818144299989,0.5953652886566774,
 0.8348836979147427,0.011401152409126092,0.339627522121282,0.955534729776504,
 0.79207557806931,0.3421333464979869,0.16504532493198742,0.7167872132663761,
 0.06872538969523645,0.21139402262364948,0.08231604313378937,
 0.27346252039368435,0.6766334606631513,0.864440991184216,0.4925414325753532,
 0.12141263839880745,0.5955531408595496,0.17367705685054036,
 0.3070745310179691,0.6377843312755739,0.7440905516576075,0.5741638837334516,
 0.45906981167417404,0.39340998376822733,0.4828042521604873,0.266160791122968,
 0.6265042203353604,0.48703563729917476,0.4461550376369836,
 0.29720730851289356,0.19159668568603028,0.8076210886334547,
 0.14989460440536562,0.4048596017878576,0.9143311147393876,
 0.055983893564747644,0.9175736439108612,0.2742708243260451,
 0.9247840025948163,0.8488531133457817,0.5853280798129045,0.2732824535493974,
 0.006920620723337656,0.03663057839057027,0.18336141544688378,
 0.3256402526980532,0.8400462178395101,0.4562138046261419,0.9340112479621179,
 0.3609623128038498,0.30492613536239965,0.06838256699806688,
 0.9438234218908713,0.11938340718207607,0.1226842337283719,0.7997569797901005,
 0.40038887898941145,0.43800931459384895,0.8908465172928635,
 0.1330313796829513,0.8726659750338144,0.033731741570728335,
 0.5506006808683009,0.24554908751114335,0.9617281574597121,0.550084429420679,
 0.8037191797509005,0.3227132366299006,0.5842442483654949,0.31063117109581717,
 0.6115126938234696,0.8037867655912461,0.8931993129324114,0.7651601031357018,
 0.9258466272909613,0.33143592357701523,0.7398695911434591,0.4055483272642614,
 0.5114670614484464,0.1149457630610573,0.3831392052315996,0.9579869923240083,
 0.5375655143580939,0.9859660675738611,0.8584273328000459,0.33508654384065273,
 0.12214757741876614,0.5677274096552198,0.7002063160996242,0.6461631706738391,
 0.08928422337101682,0.5177545743520853,0.2187052777166325,0.7514094974562644,
 0.22516958779191099,0.7161965311269484,0.6580577787768611,0.3113838694509532,
 0.732886066957635,0.24201592144396988,0.07384348287797071,0.8215577497450823,
 0.3229292464478196,0.2949866448620213,0.6754197387553056,0.455229323931152,
 0.9815816572330753,0.5314181458873862,0.1886616046887689,0.19318213017249708,
 0.03748225108483805,0.48456500076768394,0.21740407767199788,
 0.011029457859399683,0.09695693816046991,0.2809178475296038,
 0.32142181333401254,0.19293829749725178,0.228855389869557,
 0.47742786914188584,0.4524870304591636,0.8383416701812922,
 0.34934816828913484,0.24307004172547597,0.9353814820336795,
 0.7584491117344067,0.3599933463492758,0.32278600127560786,0.6637634483823956,
 0.8842191181862109,0.5997207538417642,0.9248230046303079,0.4858842165217858,
 0.77031421296977,0.4282690241680799,0.774427963894755,0.943788455819594,
 0.7488964799009723,0.09850500469485945,0.778691013213157,0.7799526707556229,
 0.34630959511016757,0.4886070186784993,0.9932725722464434,0.4686041856681529,
 0.38138349400365246,0.40788846858816574,0.9741871261579708,
 0.9433783958910551,0.5820494575835902,0.9515975173961921,0.9871017731272924,
 0.7081771141354977,0.889449188631319,0.45046410686746285,0.9682686396192195,
 0.056565354814990365,0.25848626093082294,0.36345794579204493,
 0.37829440711948725,0.747288141191524,0.23189647865901752,0.6118307021866716,
 0.08709749716885173,0.5079438896331977,0.3459941395288517,0.8978851954329236,
 0.5085874161957534,0.9384988765581812,0.2558714405274166,0.02673914296570512,
 0.4890656034584373,0.770093455269967,0.5820409995151871,0.8499109089497807,
 0.6551449588553391,0.37212990903447163,0.43857327165670057,
 0.4004308727725039,0.29024155399345264,0.21068545509607617,0.254797096145027,
 0.7052762526784899,0.42005324198992944,0.2478223354447956,0.4736679207618115,
 0.7584721082163082,0.9984801430581915,0.6137004141373716,0.4244135963224539,
 0.3381646248908682,0.2958493925055249,0.7112598376506677,0.9312203764638893,
 0.8113595511570925,0.814930425657479,0.8179521603821275,0.7761478198154661,
 0.6032215900112116,0.8842788109968645,0.1930228952282328,0.08117796428021695,
 0.6607104418107809,0.764323919083639,0.24556288749797428,0.29201037562714294,
 0.9794441723797905,0.23723961125809678,0.812083146246261,0.7299951812445312,
 0.6789677266199287,0.8547111907652032,0.8424210816405626,0.6949179200825465,
 0.7607508365100507,0.7898066792850033,0.4693214788367692,0.709490963082613,
 0.5663981179197322,0.41088267386304045,0.24610820767464292,
 0.32095059806394244,0.9719261659632106,0.9257922224459102,0.6878473388722761,
 0.5019753136730939,0.6397699013771273,0.9651571656944483,
 0.011065706170965806,0.6577213342784513,0.3783925158095317,
 0.14257013392451712,0.7018193782619812,0.6285308957236484,0.9543460618489576,
 0.5478704193577105,0.45221307498689167,0.2034325687874331,0.8179261669631888,
 0.9698989734935786,0.033784222578989165,0.21504408495848082,
 0.7339604328029593,0.9814577774961059,0.1860519194936262,0.12424339539767548,
 0.9444115443341196,0.31079194243745356,0.8536236528737264,0.6500920938251521,
 0.19736690380616606,0.2886255327365257,0.7455743849135483,0.5036959698323441,
 0.5617583397228726,0.06573490068013399,0.7747154532897539,0.6619737838421211,
 0.39798086839593705,0.046623738079083044,0.4840069065163657,
 0.13837478413557147,0.4195367661162618,0.2714285577678196,0.5540998502811737,
 0.6298510196091893,0.47672901517576616,0.9065203485739692,
 0.20774904844150988,0.6234815049385337,0.950940613045186,0.43687588891890716,
 0.7633498910971847,0.23943801143217547,0.671687793047682,0.617243136253094,
 0.8504415220419149,0.9510348347573273,0.8044877714987868,0.3444001942634842,
 0.5584632731614916,0.4385177833161502,0.6638059554756546,0.6798221412583845,
 0.8408062828889895,0.3363274715237131,0.5292098361413643,0.8964568008945271,
 0.08296409667885851,0.24386613314048777,0.3678644141001183,
 0.7155064366797284,0.5454501302085032,0.8040214108192484,0.8991182498104261,
 0.7533703579632284,0.9001155757404052,0.0971514624502563,0.29770471485175964,
 0.37576683402245914,0.48053752711668885,0.3780084859507674,
 0.7136576614161586,0.81509061338763,0.7637046958086958,0.4654041709236927,
 0.7106358649178273,0.17553569197676233,0.1401772723719893,
 0.49863994541301493,0.13794739082945695,0.019131103393320692,
 0.6583684419733826,0.7755082654139993,0.32031359665988424,0.7950269265072489,
 0.2523196605251743,0.37141642123027396,0.8034137128060594,0.67101790787803,
 0.36459963230649484,0.07649873755851067,0.6993134151370954,
 0.1392132305283671,0.9731173908936368,0.14305417191448955,0.7724439508699714,
 0.9307444479319217,0.29848725708771817,0.5035527886136953,0.4185396012556444,
 0.7981168687402509,0.37164157561929956,0.20859704477934227,
 0.5014286218269551,0.6391486108039559,0.15124066355303867,0.1999740257578514,
 0.011005844326066327,0.14702717679565902,0.4465032156867502,
 0.353399460396414,0.48659643196275804,0.9301002380130383,0.9019309637304211,
 0.2842273775993339,0.35642340230882497,0.39930526738941263,
 0.31623539121712396,0.8452788158172867,0.7708969515264623,
 0.26624165288389956,0.1607169861400366,0.41989874734473076,
 0.2726128129422728,0.13401118382563593,0.06179743926738923,
 0.5646820008234427,0.02384214510120719,0.30924179179517675,0.510271778026702,
 0.5790741190952655,0.9340572198520034,0.3507038377107773,0.6997397862128791,
 0.08759153295885569,0.3270388674557665,0.6756234145828407,0.8421980941186002,
 0.8714466814441878,0.7120288415948958,0.3343141318663687,0.36291135772247396,
 0.13676247806212938,0.07690297498641696,0.4232643428835027,
 0.8003358464397669,0.8218296230478791,0.6092952319468227,0.9198579093905713,
 0.10128408732432692,0.7499023828633268,0.4229361537588796,
 0.24820855127675323,0.96946222265856,0.6772955600321109,0.5593170941170452,
 0.4755658351340686,0.31752403120797046,0.5160495207478085,0.0770173726288711,
 0.709594326302047,0.21724354192576656,0.019783501697215322,
 0.7896088421992004,0.3367699166496356,0.5157033153151288,0.3141618579409079,
 0.9955817644846163,0.4437504757273947,0.4063856873606726,0.30816997893346976,
 0.8686991630009298,0.22537170039482446,0.6250340483261374,
 0.07097280391462524,0.6685203151792606,0.29527245189547346,
 0.29495326267457944,0.5689694058976884,0.9732432606367257,0.2006834377668706,
 0.709194339356034,0.736952529078569,0.4496771371575693,0.49847010027664806,
 0.6295260008191627,0.5986124857125844,0.2787507408559403,0.6981769056922688,
 0.8825418321716317,0.21492707989717574,0.8178944140796559,0.6971764183423844,
 0.869797686372881,0.6546247347325959,0.6882058827992144,0.7825778619757066,
 0.22420492222049704,0.7262463644827551,0.17834882115537554,
 0.7411609548339632,0.05050172875217285,0.017230821929646423,
 0.6276770521616044,0.8887592549158465,0.2954536163380943,0.6419175949006282,
 0.5828292116883615,0.8121067836523845,0.2813903184044977,0.8505897419542111,
 0.8954984348980868,0.027482245722235787,0.5977469480951342,0.814729975204566,
 0.010367800886043388,0.13618152575226694,0.4871429127768414,0.87350452054237,
 0.31632755771241083,0.3582070862574125,0.8796885292975795,
 0.32485635658881673,0.5016658351756466,0.3217135291960028,0.9580469255424199,
 0.642745610666823,0.9803544478308468,0.560717279527249,0.769601464860694,
 0.8625940602366409,0.30205023540226505,0.8610969745227083,
 0.020618882573209873,0.20849572362888125,0.11645467307023982,
 0.6452371139057463,0.6186423367823153,0.6381511655191255,0.5744285534922426,
 0.055426521784184835,0.008733674655150292,0.5116766420161478,
 0.2664249497969572,0.26597594339779984,0.10974074932977751,
 0.5530393045462426,0.731149284609541,0.6304674338776203,0.9562058957315291,
 0.5180217042233254,0.3251629981064912,0.1804089051826101,0.9655642990741032,
 0.4484018484081893,0.12861859032487244,0.15569931939669435,
 0.2141547888523534,0.9456088360883435,0.42604340113296457,
 0.34128931961694287,0.4662693002271966,0.7175671105121917,
 0.32822189372239485,0.10911624745484105,0.028245869676402524,
 0.06615925520557187,0.3826326211707466,0.26283948611585195,
 0.8097352732317392,0.3694235480476662,0.17968665835416253,0.5115552151854781,
 0.45709386267768615,0.6750338212268954,0.8710551473269614,0.7862035083070735,
 0.8307313361424903,0.7437820369599675,0.45892851592503503,0.8079279638776755,
 0.11761158130100524,0.4865440885757958,0.3054525142945401,0.5438545006233378,
 0.15936090695901273,0.5113333563656701,0.4580314679372519,0.1207463060153251,
 0.5972554804673229,0.32450519138557377,0.6192751108715727,
 0.33246018165742575,0.7389010675265845,0.13178957998534346,
 0.8040197692043168,0.5840070215042829,0.06496519867305728,0.8670714210510343,
 0.598249393902703,0.9181715838841001,0.6566902085878756,0.13744971358520575,
 0.85315736073171,0.8863318902520382,0.5177814440668235,0.6553097285701264,
 0.47707876381780057,0.9364194167342483,0.4585379864913749,0.5312780973756084,
 0.3920502819014221,0.4020196947242668,0.6915699135904836,0.9524701852432821,
 0.840640132929178,0.05109934726771004,0.8978357479485131,0.28372469629098007,
 0.2758626241411827,0.6136350764080275,0.9071205600624415,0.9082903424301603,
 0.31562109432791763,0.254739773970069,0.9298469400100993,0.8066580334360169,
 0.831264349488117,0.025683899325955495,0.7497179955535175,0.7053376900253626,
 0.8734855851247431,0.8040859360218614,0.2801967241811725,0.4768997665395861,
 0.8580948554588592,0.49901092286539506,0.9638279834545793,0.2796961144352865,
 0.19207281314163627,0.5773103949254402,0.5477538764274482,
 0.23972942083231463,0.93951947801836,0.8949980973971867,0.5540567197239711,
 0.34764740431163976,0.9061468941658106,0.8933850357440294,
 0.18985601518651007,0.8651616113460275,0.42941274697653176,
 0.9455084371679352,0.299272989373758,0.02735799069903233,0.7891448728316967,
 0.771249484258885,0.8140460843004499,0.8224720539296513,0.6367100514436872,
 0.9120398077109264,0.6685177777236282,0.775918562977414,0.5263011085473952,
 0.27817715085475014,0.4246574228090544,0.12945104030914634,
 0.42124322241446677,0.1418228894050314,0.026093181663982556,
 0.47002235911510837,0.8169614331221664,0.019011546078971375,
 0.5886702799274258,0.30074463822973474,0.1815142459931458,0.7011349168836049,
 0.035414062098579446,0.659853570604414,0.7984041483222196,0.8106544909538699,
 0.9783302628187884,0.07546326466097819,0.33216630898630506,0.666377383174432,
 0.3597832616169194,0.5303503731706698,0.11918731883804667,
 0.17021220235178136,0.8739013573508366,0.4635079690231152,0.4140024752728575,
 0.11406557051026389,0.47785407642742284,0.8729667007782146,
 0.8822932844374985,0.60910495397754,0.5081385254311033,0.5311528381517505,
 0.4000908365031166,0.996975019964226,0.2547709646326305,0.7458680792661706,
 0.4595674587140992,0.6674822116530081,0.23405389080111083,0.9163058363341882,
 0.9294788037215268,0.4626274045677291,0.9531363341892459,0.8442204053542292,
 0.5127212925015242,0.8040731225195621,0.06393028966901781,0.5038810391166618,
 0.5502705615758348,0.9005889245518627,0.3434158736891072,0.7150253275045545,
 0.2203074405887797,0.7516041635745884,0.3636481364557669,0.607656838777309,
 0.5131949263551956,0.18035653842724486,0.5794853140924202,0.6277519657721278,
 0.4173205063733384,0.6366815880829136,0.8977375676817344,0.15150756225515505,
 0.8792108953096484,0.8603480335082019,0.6768234404520489,0.4640905850307854,
 0.4493708339130269,0.3236246756145368,0.5948571068107436,0.7891631814480687,
 0.2512177703738244,0.2342602651266752,0.5937705747353222,0.2078464771315649,
 0.7596063219726386,0.5385484765235443,0.7356597246321397,0.46183720523769767,
 0.4665087877249491,0.8392963241634884,0.6358209316445972,0.6038915733960654,
 0.9730307641716923,0.9552991374534713,0.18549659705015709,
 0.04407962412360089,0.0422085931110141,0.8959340102041018,0.6781881033091652,
 0.28783464101025946,0.5390816170512418,0.34745828213951435,
 0.16151596903147936,0.8609452884312275,0.02640862073362471,
 0.6655318321650999,0.23122550928205465,0.2091130695762352,0.7049242439490041,
 0.4811981934006849,0.4147282585483014,0.7539027357168713,0.9536903628997293,
 0.5238487134479296,0.36826168620452004,0.6366544840463646,
 0.08273457700608366,0.6656896329753361,0.956054993047359,0.26120154521271455,
 0.7068025495447038,0.569515056416289,0.3674195800314092,0.524523875692507,
 0.7973143377990355,0.42959865091646166,0.21736504481804164,
 0.3961905021410693,0.6485696586372163,0.5245430508231823,
 0.047974769221223346,0.8917952167030597,0.7299828796865064,
 0.8395145979406518,0.8787924931664838,0.5755891631347249,0.8425794675761665,
 0.6317690204964795,0.11576543932592087,0.5728157703902079,0.0629471321073859,
 0.16646928038209696,0.23977340911361378,0.9956045794344602,
 0.9205793068559285,0.08733942952876517,0.9213006498025598,0.4905972963992983,
 0.51722468898946,0.998530308778892,0.8755363665549151,0.21367901897921948,
 0.5761802868186806,0.9501223446666487,0.4389807410160458,0.14083544090986455,
 0.11770654771920519,0.6269486569488003,0.44106169975366005,
 0.13392190414603378,0.5911350077616486,0.2971691969217445,0.8344041878002741,
 0.5410806776173436,0.8070249599820327,0.5301587304522408,0.9407671440652483,
 0.41623482044347926,0.8536602190707122,0.7152439362816104,
 0.48648367424682104,0.13011562927465303,0.1989498247053706,
 0.5144964649895092,0.7766033292683288,0.14500815408284484,
 0.05165460680545075,0.0036244865885664,0.9630915971490596,0.8002784141019575,
 0.4541265388679121,0.1275384194903364,0.30300480208949854,0.0426667976342775,
 0.4118681286442404,0.5629619714700633,0.13603293469650557,
 0.27235160297376604,0.0714719659412093,0.706363487559569,0.6637532257127009,
 0.4245300684935467,0.004516362732723489,0.17570833850804823,
 0.2932800925708352,0.19196156384314356,0.2576938653186597,0.5137423735675207,
 0.6258798334900081,0.7805771612938388,0.519554048718851,0.6996326342242032,
 0.08225267488900045,0.3176198945705846,0.06723734052550057,0.358356449801185,
 0.14180727608642818,0.7875694876306378,0.9207989166856794,
 0.32724090711694687,0.7940668904295755,0.7910350088672659,0.1427834720943415,
 0.794815378507621,0.4535276255923595,0.653512286372752,0.6835010782491253,
 0.20158708659856828,0.273885530545247,0.6270614729315904,0.05168967730211138,
 0.9425521549174753,0.86930694514918,0.22137334657160224,0.3645680718633645,
 0.5805066508128746,0.8334602679705986,0.5724093681310197,0.5489283173583099,
 0.37678829242784573,0.8942583672646695,0.39033986492236306,
 0.9083676478846532,0.44814878115264545,0.36370142729810007,
 0.7826901018324997,0.277504958212496,0.5285514859317497,0.09725435419279882,
 0.5917933191824059,0.6666455110334008,0.45925351064040043,0.0248444935025931,
 0.3134924966380681,0.11293482860072168,0.5179399945989738,0.9077771637388574,
 0.15611404361864267,0.5618605006571268,0.10029630205689544,
 0.9650061925166014,0.8689282894401237,0.5562017099873614,0.7805717890723491,
 0.8536223735583758,0.7311131890084769,0.8084971560909865,0.2637479203459089,
 0.7897191111489593,0.6219212949605077,0.7651810793018603,0.490656331667069,
 0.7253785209286268,0.16876670225509294,0.679839864529878,0.8690815138627256,
 0.2480256636289293,0.8658110581967241,0.9292632148222532,0.9116538221704364,
 0.7479910450000976,0.19393992043710928,0.8228916895924249,0.53550177757874,
 0.1661020000911967,0.7943606408431113,0.8935769650526387,0.16049738615178077,
 0.8856436265828074,0.3318330552656916,0.332796580996646,0.5579767365409616,
 0.9910246563319307,0.5364723921691085,0.9569535027009994,0.10126263291719695,
 0.8085630762932386,0.002567783898535403,0.41024404441179274,
 0.6950386904557888,0.40210655345052515,0.36282101722645155,
 0.10439230436181968,0.6639881344560887,0.9522661584779517,
 0.002114562549792809,0.4198891197001289,0.11642160698391579,
 0.4855348052819033,0.019068464206160707,0.7655388036229316,
 0.10383609923393955,0.06420473059735254,0.16597451213390924,
 0.22856849976880234,0.44748430537831585,0.47875395885227656,
 0.6112411109598506,0.1092372514767157,0.728385341816888,0.11846990333158258,
 0.8415551505656593,0.35371586524852405,0.24371834394136838,
 0.17825753397635336,0.8072198943899286,0.09299515593938779,0.893099111893589,
 0.46375658279417387,0.26059296115535624,0.6122184807544655,
 0.7418088366705269,0.17419844785130767,0.5765680776146399,
 0.44536185767851466,0.788382252307661,0.8865358000130193,0.17353266203808948,
 0.22423959739856325,0.5159743421972549,0.29221575530074073,
 0.2685646609142104,0.1719049390375117,0.2147174750434523,0.1637962633774711,
 0.7661169455027781,0.572217241154987,0.3948590876389504,0.5837406544887231,
 0.9562599594119124,0.47191246718649493,0.9464212425053348,0.5197447567870612,
 0.74956963685542,0.9691974910544855,0.6134501187287402,0.505802275607727,
 0.526349287926047,0.5323356885018401,0.45105801928619427,0.8191714716495905,
 0.9062514813768301,0.12369708577694849,0.8503432454476315,0.6066693682349076,
 0.6688751941103633,0.8857850538904465,0.6687994517364507,0.9400467693689759,
 0.7443515812451406,0.8019664185154786,0.7061619202025758,0.0639352696681379,
 0.5042804825114019,0.4404037855338011,0.8159205459626271,0.30511896218500234,
 0.17416339111106138,0.018139811616349588,0.16466067805503293,
 0.2604981389141571,0.8480690336727223,0.5125794264080157,0.37686264191057195,
 0.007834625731014455,0.3952684435002588,0.6500256085260209,
 0.5832933620823488,0.6858944187278927,0.33421306673012807,0.9223870079450726,
 0.039018124262701015,0.6660443713826051,0.5599176881249714,
 0.6612035950834528,0.3982870169923446,0.9609201278957112,0.808775383909824,
 0.6394861736984053,0.9172925330577992,0.87433915728802,0.19710401870110839,
 0.22325049176357326,0.8859397682737133,0.03551356497979519,
 0.7930799421824378,0.7580123261716363,0.2758487656984221,0.6639895591218906,
 0.5741653718239498,0.4859052336987518,0.7002906687152719,0.9351011582996543,
 0.3920927564775558,0.8680121210395658,0.5425562941531449,0.53010311004948,
 0.09347637460731106,0.42238551341490993,0.6773743086193739,0.745387326189998,
 0.20672327252761635,0.9311177522209164,0.41453501549249594,
 0.5553331713386127,0.5756695560403974,0.3041078717530151,0.08846393159883748,
 0.7541851076610238,0.17588391001347525,0.5564274671116349,
 0.44270003897756194,0.6287460523626087,0.24543986394878126,
 0.16974162859965947,0.36620893008384614,0.2655738837943372,
 0.8317082956040833,0.42447359402027773,0.9503787953206255,0.666267234640271,
 0.3744731647479984,0.977621039283821,0.12709398407148276,0.40445815980451416,
 0.4075351593594134,0.8138453703124775,0.265233434216424,0.8543547026419485,
 0.6583857008822408,0.33670767277737945,0.852676614792032,0.999990215566775,
 0.8805146194747769,0.20688204670172616,0.36993011961073163,
 0.3537795486643984,0.04651664238755915,0.2259681757706633,0.839750988834868,
 0.9343527912580168,0.24205619515944865,0.15238063004763636,
 0.8652935359260384,0.9657877722093693,0.7993029283915299,0.081006462652772,
 0.6358668464256981,0.22827630343202399,0.172110699727147,0.2875407225179424,
 0.7030435621770199,0.6677016149532019,0.9173173262424975,0.5620478173709073,
 0.9595464774773148,0.5372728221493233,0.9974252009705695,0.8218344183231677,
 0.8692119508209221,0.4644501269106336,0.6164751603053925,0.23248593700343623,
 0.12141223841544468,0.6110473492570246,0.8621013822081944,
 0.34499576969844625,0.30579557492346654,0.02240980791980618,
 0.054943408117314685,0.21449508944416473,0.023653480690846718,
 0.017645467172554663,0.575201278513746,0.9767551300893594,0.1868304366755953,
 7.332224644701624e-4,0.9862116690564557,0.9228474083603823,
 0.33730068363013177,0.6549703090709817,0.14869784947775722,
 0.7686230475734792,0.5946385059855477,0.24366169074162913,0.8578475753294784,
 0.02495887952761011,0.7017340893219632,0.29924010053466443,
 0.9664037305928378,0.10139170196565384,0.15765822450035827,
 0.3205962370093617,0.8887295868412445,0.593600004599738,0.5172844282308364,
 0.4124987905638693,0.7614108942883513,0.5121316460301457,0.17103552007082157,
 0.40008430640505654,0.7029839349562986,0.351231586342842,0.8151987812757079,
 0.27160585622583433,0.9173689053570809,0.24467542940070597,
 0.8100951665712235,0.0723753269016626,0.8040777037198578,0.6607399610407746,
 0.3992671377504181,0.9147540196938018,0.4893696140448649,0.5605828575940833,
 0.8518700017719084,0.6431807076888343,0.4469017690633119,0.7421959199604614,
 0.8308208157788802,0.23862018587557743,0.4642469817638507,0.3603891021956789,
 0.2507146517985127,0.8894287735805874,0.5603809377921878,0.4146905831073554,
 0.19188339119806097,0.536753459352705,0.5809330357613579,0.7297424323178812,
 0.03978051847518216,0.012205867572504525,0.2633383040305858,
 0.24583141910153294,0.7931245822973689,0.07994771883120655,
 0.5261719384386151,0.7128694151586296,0.7519753002058733,0.8233213719666818,
 0.9933343719632182,0.9955109715063803,0.07612320131406436,0.5925386593728373,
 0.13563678228385534,0.8443815150771622,0.5135595343015695,0.685398113566057,
 0.8836265521680208,0.020488141454350184,0.352381151738832,0.7989161557769799,
 0.2910711149779741,0.2914119477146553,0.6603239167065663,0.12847495122972297,
 0.47398571767938935,0.12956852627405357,0.8852882393213131,
 0.6981643603591773,0.7653760459821645,0.9854652707981444,0.9366002936449688,
 0.785532021191746,0.6408920161557871,0.18304027668469636,0.7581221089196579,
 0.9246733012451053,0.6999112918419639,0.4383895755623196,0.6463500276735298,
 0.9205840876493335,0.956577900890669,0.9385292184129824,0.7801035029498661,
 0.5126998693244403,0.4418290440657151,0.5557852959000505,0.6617377745293134,
 0.7236292134385365,0.5097301871838544,0.3700252813537044,0.8348156874666892,
 0.05053944514430897,0.14925940140349825,0.5841445188652049,0.686955754098212,
 0.8791793519088023,0.43076292567441365,0.6585956719321154,
 0.41900343583946686,0.5417501778080545,0.38358541210705066,
 0.006572090327370161,0.5638890970390765,0.879453658140586,0.9795814375452645,
 0.3641105343688633,0.677429274447894,0.8630894360938484,0.19421150425704536,
 0.8782027725933248,0.3860475774111928,0.252731476345321,0.6017562083932175,
 0.18825895194464803,0.564718915317271,0.06382010358223456,
 0.08945837165309767,0.42893714952056317,0.8736069964661106,
 0.4996682834211781,0.737913331434727,0.12476093369654839,0.4186032117699062,
 0.7737923491870312,0.7529637449493627,0.5529669072408735,0.39247248415107105,
 0.6356332718224703,0.1310697379691902,0.24230122308828272,
 0.04221744207138345,0.426669591456709,0.2693920159152584,0.9592919145970893,
 0.025494850209894504,0.3952614323946303,0.06756703911614259,
 0.9165959879370909,0.33487064243375364,0.3379198232840468,0.5356065025085013,
 0.46501579934776216,0.13284318671607542,0.19625564217382252,
 0.19586617926152505,0.6178844449601283,0.6189482331800705,
 0.44973821015322646,0.14922097272657153,0.4576891977605654,
 0.9775859463742306,0.25427635474666965,0.38229324713930224,
 0.4174578351781374,0.32889657531595284,0.8431182439194013,
 0.014490074596520963,0.24619471608943533,0.8499088889326059,
 0.7705214218215292,0.9144433970691991,0.14429296045865692,
 0.38743107747580097,0.8703926042115809,0.9579677106638571,0.5241711337693271,
 0.14272286045318827,0.8109920821632917,0.7645359167827378,0.4591357947997293,
 0.7973316022310759,0.5120189996005632,0.6170293443274604,0.07528657522365356,
 0.6654867484151978,0.7943948452784005,0.603792897351857,0.4416391846320412,
 0.2938572598877036,0.3744939693807612,0.017761580973934432,0.487579666097931,
 0.5280863142131944,0.1698118486495146,0.8646866153249294,0.9832417680017628,
 0.40271429701935,0.7413710094619349,0.9665987578598195,0.6181823439651687,
 0.524346921071129,0.1309900224216174,0.2122822577079313,0.12691447245727838,
 0.5157973064595208,0.8366365544160445,0.5332196138768206,0.3689516281616594,
 0.1863460315661517,0.7865682683590589,0.6822506203238,0.5536551876326141,
 0.12798392809662806,0.1460989742339608,0.5211874503362828,0.3326308668833404,
 0.7000583139752832,0.9037327880574922,0.8289240854072161,0.6599184738957784,
 0.8931369970614234,0.4586460755627112,0.760301638849936,0.2183305439418496,
 0.04649419218169604,0.48617367134455625,0.13873799448322965,
 0.0627844432916379,0.9964145087425819,0.9662709885703815,0.6871877469071972,
 0.12360319343290604,0.7861766650906071,0.8518999971620292,0.7582589883924977,
 0.8563072494548228,0.9860941085957068,0.6494883504838659,0.12169155640773299,
 0.5054205392125466,0.9089046121267247,0.2576073390117959,0.3939597041890439,
 0.9936284821686235,0.5539566765350735,0.21483770524168322,
 0.40055419267070946,0.8352316830706968,0.7330263342912374,0.7798158882220053,
 0.6702996326262602,0.1123665489863338,0.6193902449059201,0.5109983825992208,
 0.2163057140757506,0.4866588106021992,0.16195496534057308,0.8346658464202863,
 0.04531832871090402,0.08220824397444138,0.5843012411679701,
 0.5208705797883009,0.9792109256694344,0.5241370529508989,0.3106396346247412,
 0.3709774599830785,0.19594773031915913,0.11352810934767132,
 0.4152449659564593,0.9272997101355083,0.13072210495347025,0.8498398656845774,
 0.09251584027354132,0.5431966747522772,0.8914947209402367,0.415985550090893,
 0.395968828342405,0.2743167800937486,0.5431052629281289,0.05217837658884372,
 0.2126456821302133,0.08888555896803685,0.4605051243676789,0.6374519228249034,
 0.6901932001422163,0.702311163109077,0.8178844613674962,0.9605829465989855,
 0.8983222387095047,0.351085207993835,0.5354805625027249,0.22422884449582114,
 0.7718445640240246,0.052446059673905254,0.40183404818717205,
 0.2364857860351488,0.5246872272706431,0.01676783466060261,0.140417178074269,
 0.573758242145657,0.026598620688666696,0.7070332674505301,
 0.31072225466997416,0.7345139713363783,0.16429096978024205,
 0.36284877777243496,0.4605391912756497,0.1183181983503625,0.2195504611123067,
 0.9290217736072138,0.5004468792112243,0.310433549711136,0.2078320229053947,
 0.9060515718247899,0.08418457238950461,0.10902070543474385,
 0.7475537946421555,0.9327765240049835,0.4471903466401087,0.9195454179688228,
 0.9689110072281706,0.3719135115947533,0.34255057269474243,0.2809032486354893,
 0.035012621900725294,0.641181069026876,0.3279741024248789,0.2732519134331004,
 0.1981627826263883,0.6625049300689922,0.9318147957042512,0.5663799345534948,
 0.23010925467948218,0.06727233537363397,0.5498891728694195,
 0.10661893891902963,0.6987336365139095,0.8126051798516962,
 0.08119445543711012,0.9388886563980927,0.8921998551004392,0.9862550510966152,
 0.5292346908070779,0.19068524293384082,0.686139077982809,0.3429861926109262,
 0.5281173066387621,0.781949076453919,0.7669242792732184,0.3045860372613898,
 0.3425250777436014,0.8878893396738557,0.33846414426578364,0.9241172394127585,
 0.812549869621625,0.20610231818851188,0.5098450624699693,0.12164760212343906,
 0.6854950431570511,0.02602044602554643,0.05521275619994992,
 0.5314403144237353,0.42192444656213246,0.5705895204814542,
 0.08156713372936775,0.6874865885848859,0.6044816791886287,0.5790988594682103,
 0.16722507230944395,0.4376891736530265,0.9971169863415883,0.8405310917669031,
 0.8407259042142339,0.13658180521769603,0.6951687260284762,
 0.09008391696423435,0.7056308346613679,0.5078013308234473,0.8786343296202125,
 0.38278265712520654,0.03008547530631478,0.9749627655067586,
 0.7069702383321496,0.16613497459136317,0.6062043228930148,0.2526354573946893,
 0.599789904182662,0.06961834721898796,0.23816793087138255,0.8450383368623552,
 0.43936944178680293,0.9372593814830263,0.5762867983686685,0.6142177865487298,
 0.7585223573657816,0.12769064893819415,0.27327494345554515,
 0.41361744290782276,0.9932937033674024,0.04668985814755677,
 0.3963065335625834,0.9631432394467723,0.8275474370049766,0.6951231198284689,
 0.03170671454395968,0.44935072103223295,0.19593111905660454,0.62515947889599,
 0.22259806974403817,0.49651704436170374,0.8498997883573963,
 0.8320932433658617,0.019796430001965515,0.6445795480641211,
 0.21062963090502773,0.2604128097848314,0.4862965575861685,0.7524852126795047,
 0.15968157714698283,0.6017663484738478,0.8665007108176638,
 0.42269094232440196,0.3379011881020768,0.4534865778156332,0.9080677186840341,
 0.9537777885829519,0.4324691805456249,0.5140223344260829,0.9719822116512351,
 0.7621594493117572,0.9184649937291783,0.9075502700072703,0.32669191713746615,
 0.7378785938839888,0.06987441118819504,0.4854032072684774,
 0.14798203135343935,0.052962153101681286,0.6983641351574109,
 0.9428849198212677,0.038374587704667174,0.09693236479848488,
 0.1032006059865489,0.6993951612346706,0.30819352514696696,0.8150922298599172,
 0.6780605222749881,0.8329919506112541,0.8019670121169976,0.6080464170010651,
 0.5980166665685926,0.9096788710801188,0.4495318238810847,0.8220453037708408,
 0.7528560124305664,0.11667435731047038,0.07094765319193497,
 0.7122786920206099,0.1754292098334318,0.15469470359565074,
 0.09446951214250965,0.9020417622197847,0.8595771020181397,0.7928359608190523,
 0.8671678716830566,0.9711725043022955,0.027561277513047266,
 0.10024565283544085,0.7449315403689629,0.5462458476481912,0.2392890268539809,
 0.41412634487618016,0.18088699134741393,0.5614252032900668,0.64733000762943,
 0.008768558140970883,0.2606569399401314,0.47069624084575323,
 0.6800640925454753,0.9280571774339786,0.4667515354043449,0.03485147443163361,
 0.8717901101114965,0.24316030831324675,0.00211243963660368,
 0.2748006920637457,0.675015282334787,0.05287167033971063,0.2508990670498257,
 0.5027202006466824,0.4053740194354377,0.7646485993571771,0.6501667953235196,
 0.9501336239890544,0.40442033341416384,0.1523353726830563,
 0.12201326326634931,0.550440950001683,0.05417487542436206,0.667580820586835,
 0.46722148768929506,0.3207709798973024,0.3365320269078216,0.4486629783262447,
 0.5544796853199435,0.29104595320741056,0.7277741031810233,
 0.015324575356005288,0.6472682105564683,0.09526223430836844,
 0.6900241597687039,0.21622011724717283,0.26194310869534476,
 0.3462223001383763,0.3973809558723873,0.06035940282049057,0.4379369886309268,
 0.42915859492315467,0.9634001325564966,0.7175349915412604,0.9391343482639394,
 0.4224147492638961,0.8666074185297493,0.9866924358063603,0.1347790888354432,
 0.3125350339088605,0.7762574594451691,0.9889551140226662,0.07556339763307263,
 0.49196356169217115,0.39146849530195993,0.17752803876606904,
 0.17091450971144928,0.03636648453141311,0.7249549017912342,
 0.6302949963440546,0.9544153958367333,0.8068541057982787,0.3981053073559686,
 0.9562160933066128,0.2755902696508843,0.16893906304950423,
 0.28529304577363823,0.03827394936203676,0.3339109624720804,
 0.06423947634144278,0.3442763928991637,0.8540060176584641,0.5945435171798564,
 0.2724446432738573,0.7755910694198309,0.6062908217562795,0.7889828200631617,
 0.8273859506910092,0.08174455657938684,0.3379349775378141,0.6671631019830137,
 0.23915666210762165,0.090405958409699,0.2491647812733233,0.4725890529516552,
 0.247702028305973,0.004828199424489532,0.7274959592149243,0.8280845427152936,
 0.6085613789709936,0.1020339916886327,0.9303759825708076,0.15143384617740696,
 0.9519348512543153,0.6464074313745406,0.7891243447289226,0.285389533539503,
 0.7567844075696182,0.9661939347845004,0.012117780460960126,
 0.03458571726497961,0.28669335806516105,0.2221653206615415,
 0.32998054368758223,0.6335817618143883,0.3908678840492752,
 0.12071039017222696,0.803207203171151,0.6442000981539272,0.5203869230828989,
 0.40517591994993496,0.568797604312454,0.9625126993991436,0.5013708648800284,
 0.20948483914171634,0.7849194094859373,0.006972886323942218,
 0.38192085413592913,0.31181157021058925,0.40595636034816907,
 0.5130064769704539,0.9323030578772854,0.865180864135157,0.4419300695715531,
 0.887852026347194,0.3694625874374986,0.9016631020291604,0.7905079057610281,
 0.6723986263197937,0.1462340700267899,0.5430547043951862,0.6551772978088759,
 0.6297678570434345,0.5088287231430387,0.042795541913085655,
 0.6886682602992653,0.8449805304779567,0.8547988659461989,0.6731263005205879,
 0.614924113916623,0.24802509734939227,0.8650339081499354,0.24673906722247896,
 0.10868070739274949,0.25846426991134164,0.5763423932753369,
 0.6283000600679306,0.4924120990679076,0.24840278363167245,
 0.47623638279051983,0.15219951308934565,0.18358256646893845,
 0.22227430559531924,0.6272846749137903,0.29576967637058593,
 0.2187746642266396,0.9864058042746928,0.5862576731256708,0.7499092371204477,
 0.34185533455544614,0.6670972355755675,0.8978849375586617,0.3305953565107991,
 0.17365893644811137,0.9004999482205783,0.6194650641632715,0.3116304973493562,
 0.4761749158148336,0.005040055672804855,0.3168008899297291,0.886414030246647,
 0.17382516340876886,0.8802265246530853,0.37486917593301006,
 0.23404115159476646,0.9817460465080843,0.9472172355395184,0.3996402893574955,
 0.6521068265796223,0.5252537511265166,0.2707268322439643,0.9212484537796394,
 0.006198892945252665,0.05842365140266237,0.6499998493542533,
 0.028008476976242314,0.03909588423008137,0.16130194255558417,
 0.49350678082102084,0.7116147891482535,0.6799760416988627,0.763046722959762,
 0.14688111976296547,0.41364658958585465,0.5394533229499914,
 0.25764968826343493,0.05696838665898185,0.11940681933577246,
 0.42148482421331834,0.22991013745959,0.9102272733888453,0.5545686482851012,
 0.6973976689910366,0.654290052689645,0.21127696939773755,0.27868112207233076,
 0.13307743742178968,0.9702097969055312,0.9140342981585263,
 0.21908235037437707,0.2645261030287219,0.46134364345961765,
 0.32062941712901216,0.18154980572119706,0.5336992162224243,
 0.9223382897954485,0.35146903075325187,0.7178835125135272,
 0.033748202736070154,0.8200002160147154,0.3347968312592934,
 0.21117086077636094,0.9598307811264418,0.9865724908669598,
 0.19591064732934282,0.23565630315037667,0.3583588530513162,
 0.8765322576105313,0.9839898602160181,0.07030980277274057,
 0.20843667464791127,0.7492875715703043,0.8037378375524515,0.6125717315452079,
 0.6743264326129647,0.4052083810457421,0.3481481812033175,0.7229569083869276,
 0.08369318353332433,0.5773178452216247,0.7791276944057761,
 0.13079018370939388,0.0784816573866075,0.6248865323628598,0.6812908838892476,
 0.4153441803237631,0.21890929418678462,0.2728414684996482,0.8472001109331234,
 0.9701476160718134,0.36793620950914696,0.25814538290587663,
 0.5573351861211117,0.40098346606004087,0.05331059436789021,
 0.6501910080794119,0.47252743444039136,0.8392004317602823,
 0.10100916908613566,0.4222346476902581,0.12249401863946519,
 0.2791513117746355,0.9201274761913496,0.9816812621191866,0.8045896299191555,
 0.1406952883798135,0.39846136485121986,0.4285241938493074,0.4931664950061283,
 0.9377890135185735,0.22649318885055925,0.03080354376981742,
 0.1604660076077815,0.2729049549850687,0.10002820424079428,0.7714988780606139,
 0.5814186349586304,0.15308190362177077,0.02969048689191689,
 0.2896945956088316,0.2972143618288392,0.16022339625767779,
 0.24095107221699374,0.41620925953641685,0.2548549785566362,
 0.013272288317306424,0.997116573793682,0.8452347704507839,0.3416831785301411,
 0.5164445995607688,0.6817525448436869,0.2676152918423925,0.11471135750029038,
 0.2090982641460163,0.16018933536697277,0.33641234330195724,0.57874299977478,
 0.659022863874411,0.11005379990926867,0.8726454446787762,0.7393749361018499,
 0.10457661521554762,0.3417901708953406,0.2325940974794658,0.8076584005580565,
 0.2633807542155905,0.08455578038042233,0.3038152502184237,0.502301960990101,
 0.7211142396046777,0.17693170186304819,0.7088008306851457,
 0.49181345165298374,0.8544843435489768,0.24538094197476856,
 0.07258633632776523,0.14382829647159623,0.795342328427989,0.3859219729026544,
 0.6540933342051015,0.38349480968845173,0.3641639614016209,
 0.48050919064041075,0.965489489042379,0.6588762485041341,0.9667207250768999,
 0.7882184896558355,0.21594220745919324,0.260555885926428,0.8952688663302796,
 0.9711868695263984,0.4747089556595614,0.47629889422835314,
 0.38693723718055517,0.6315920834478398,0.9828463661535809,
 0.08718216101950904,0.482947543471242,0.6139495944702701,0.7278240814879623,
 0.5713000981430144,0.15286505839751763,0.583184631406876,0.3485590860646961,
 0.9741396437218537,0.6505752591771037,0.7375493381748562,0.6603428537190185,
 0.42679933752426324,0.8886693129348403,0.3663849313879116,0.6123056020269899,
 0.13550995333849514,0.4569824212814294,0.5849456414900085,0.409234695469483,
 0.824977992700769,0.8752770499687629,0.7370778174371814,0.5257566715588756,
 0.01266182086439327,0.5353638471154419,0.18046559391302341,
 0.5728338934829071,0.18777149362815293,0.04693904497096957,
 0.3205676841074635,0.7357011167365246,0.24596243013790975,0.8191642195176928,
 0.1853585878313706,0.9891557984268229,0.06629827454526305,0.9693276463614282,
 0.755692008150876,0.2939131554474095,0.5346607381153552,0.590335063891207,
 0.832650888404623,0.8956122176256869,0.16895437837458882,0.07202284693194494,
 0.3689124523352594,0.02385935438801523,0.8001372705064447,0.588321705113788,
 0.8104362971107909,0.8193625838977878,0.5572769023185453,0.6363953608951467,
 0.8950802721793585,0.13485762247416067,0.6000409228383693,0.7176203117069802,
 0.03850196798314276,0.6933794769253775,0.6933519645190042,0.2956009088906788,
 0.7267603006916921,0.3253072908614557,0.8576861459408112,0.34482513253763725,
 0.21948978986593692,0.7787031591702139,0.511069186110871,0.29754272263571047,
 0.39468682971216995,0.7494804412590359,0.7080525362332919,0.4702232501969533,
 0.466907854940497,0.7048998365775985,0.25430091638673746,0.4195397233621967,
 0.6361012881809003,0.11065111235597103,0.8514799385217147,0.9355850833488589,
 0.8673289530825866,0.6093352213152066,0.03391434980337482,
 0.12217291857213874,0.961547737752458,0.4063961654799684,0.9101656607989621,
 0.20197306805312254,0.3223504872675056,0.496749359464449,0.6453423960439127,
 0.19817593736136963,0.4009325279543752,0.7930243982002543,
 0.21272826238120657,0.47205620695026795,0.7113095087137419,
 0.2600463212286632,0.8156547590187543,0.1088753284908357,0.5995965885727761,
 0.2830774259185106,0.06215803539571829,0.39712352924914074,
 0.3236628863782087,0.7843390838062088,0.4440174894244475,0.17914976500868984,
 0.5401170406155288,0.6982131146914737,0.220267272891853,0.0985101725360139,
 0.26671969903457027,0.9279359263074124,0.02873562447692035,0.670294480000142,
 0.4568978486809959,0.9566134362561983,0.692590521615449,0.5812895836878927,
 0.05026284643212153,0.034083248722819426,0.6074664103118891,
 0.398090690310717,0.9249960015473763,0.9478275812546855,0.9402296208346663,
 0.7996981356936502,0.396397309828748,0.789077987643661,0.19658612968835332,
 0.31172343354030574,0.10844629962413344,0.2715264693555097,
 0.34630635506295215,0.8426703054133435,0.0926315419458692,
 0.19412952798474126,0.5578510345614696,0.7013632075941479,
 0.27520981445634085,0.9132454205294385,0.5246124201593368,0.7304889101021962,
 0.49726674936357007,0.8179373515860342,0.8636215432130598,0.9976858480555292,
 0.18551211318155914,0.6554434176306415,0.13182316972567043,
 0.18525218209430894,0.05758258985023812,0.7805937326806306,
 0.6158466840492551,0.8287901664592301,0.5536780027250129,0.8765176377230198,
 0.8279760918558003,0.13478091448804974,0.9440500873014273,
 0.03968158956911949,0.09106458261393569,0.2348752021836309,
 0.3008235666631418,0.942904830323704,0.04256218877218165,0.1171206014891959,
 0.8499453715487738,0.12536974414056778,0.9222605410466748,0.2775347158693551,
 0.1912804880108009,0.5401031955334648,0.9836205693506097,0.07509126936056165,
 0.8877043283582563,0.9590677746270886,0.423583307515908,0.5931892999347298,
 0.6771473635723485,0.198443436166457,0.12789221397564332,0.27438591610578766,
 0.7328964728184064,0.0992714640375939,0.5398478264909274,0.6899968765551373,
 0.7694897107008758,0.12637058365100506,0.3717940245750875,0.2688944016867907,
 0.21404667380112719,0.7597219418760891,0.16184330838138017,
 0.12547535997170955,0.42161008291502255,0.012372075791495574,
 0.8641200889757007,0.9000668271992981,0.9218582982168082,0.3569988823810658,
 0.7830855740640403,0.2681896451821282,0.5158484775440356,0.9983140181154787,
 0.5947175595535616,0.4959983576484853,0.3252705295834878,0.804647027676238,
 0.07463048468932687,0.7643546709500267,0.277754682990371,0.7466048622973943,
 0.8116210318424417,0.861969719707968,0.7639678150691022,0.9565877774925975,
 0.3284046016888079,0.007426176326394085,0.344748322660573,0.0647353693845365,
 0.9632441840489013,0.3298333491345157,0.22134958929926907,0.6156024133167151,
 0.9305871083816396,0.9364012238482639,0.7723285011329508,0.08063804906292127,
 0.06402263857566015,0.25407000617738806,0.07943721072880217,
 0.04527038063379052,0.9131499829623427,0.7575827961743404,0.7214649436784193,
 0.3747835187478541,0.34242434405371225,0.7961076760268042,0.5110737850809107,
 0.780676930271735,0.4008205213348731,0.7463440249632376,0.035221473814072035,
 0.23011645668747416,0.28981540229511915,0.9497557834485857,
 0.42596986190047803,0.7433724744182966,0.23677924983137288,
 0.7883125577304503,0.8922053884572592,0.6038365707067495,0.7713236841150718,
 0.6575724245657769,0.33306053501887156,0.84645614957134,0.574964941387486,
 0.05116375425668962,0.4703325656910877,0.006542732913759397,
 0.29839824855602015,0.4211359465475528,0.6691297361407508,0.863406359816288,
 0.28480784744445864,0.45304998443946254,0.8702929540950534,
 0.19574063923824558,0.8200970594164547,0.25598617181469274,
 0.40618754716323413,0.9166956051170836,0.2778467594798133,0.5363910556722804,
 0.46886334251942163,0.4328386140752012,0.6739846361935484,0.2786428075880302,
 0.8579256207207682,0.20327443669349576,0.9155426359585639,
 0.12359213263641133,0.5240667520242455,0.5052384386066721,
 0.034483769160975064,0.6579398099088187,0.07069345909159408,
 0.5607173795490729,0.1666214320171513,0.6274357071218749,0.9970184831998403,
 0.46437616924847325,0.02669209686891616,0.7747009679221062,
 0.8469814492125035,0.4260968109729749,0.7295058103257845,0.8047301957000473,
 0.3451259149882935,0.1254609019570324,0.23681039625800482,0.3658837468634364,
 0.30137741550240316,0.17962393156715217,0.43373804499254565,
 0.4313460089441661,0.24017709729792858,0.30768238520730606,
 0.7098206331725353,0.3330069162531988,0.5846670591040682,0.7667274005839335,
 0.09408865592462479,0.5030300068921466,0.1532608864860232,
 0.04511660149878716,0.5799178039100489,0.8498569671731051,0.8263323692705262,
 0.9065235596254888,0.603503861612967,0.44621285743905126,0.8181958656345644,
 0.4382875833922244,0.723290317295849,0.17308431676192693,0.2001983863570883,
 0.88608799892398,0.7106160026211399,0.4355768362283955,0.6645295036678864,
 0.28289204150483793,0.5938865368480124,0.4084410439663124,
 0.03466431432670325,0.9043463690554936,0.865486818613911,0.26990426588845373,
 0.23598368232664668,0.5747123241424172,0.2121939375875026,
 0.23945761635995888,0.16978428482857288,0.8195287550246898,
 0.36946077814597733,0.7885442982464003,0.6502725265014496,
 0.07676646143335064,0.6320453818494042,0.4596555448570777,
 0.17578576001325696,0.7628876151907196,0.5128603425314667,0.8805837520717635,
 0.2331262086786139,0.4152631126153775,0.5494176003968172,0.933788211656167,
 0.39117034112904914,0.737296070079047,0.49013955436859613,0.9864238360838066,
 0.3614753517940821,0.159138798290975,0.06173529559137991,0.08403235891603344,
 0.389717372073938,0.7233082077683464,0.07075507283996818,0.40533124683947364,
 0.11716301594604905,0.4292475851092129,0.7672747805097159,0.6609327687135177,
 0.5195045224123404,0.3889143884029298,0.22248514900087912,
 0.13450022403222583,0.445648761333014,0.12532696477224226,0.5235119166267505,
 0.4239870088933122,0.29469593136527783,0.18981412444754142,
 0.39912010482809357,0.998247599247428,0.7864079465807731,0.9912707285282251,
 0.41315904736177844,0.9631526742865663,0.29775527205078034,
 0.2564396360946781,0.25033325054472566,0.4222476002091229,
 0.01694190164512943,0.4376038271602769,0.45086512939421564,
 0.8226879206458836,0.26818029793561293,0.17704516473762066,
 0.7786787116361348,0.8718613063130746,0.5922484772613273,0.09159998732906949,
 0.8502834621064277,0.0520240057858572,0.9532974257247167,0.8482720089119464,
 0.8467803385137219,0.7851775817953615,0.23501007780031968,0.7017488639373636,
 0.6439144740379635,0.5597911611167885,0.4930798866681899,0.6687238528791561,
 0.9618144754734297,0.35632304047047947,0.9158876390293139,0.9397176432481212,
 0.24246694831302296,0.0842730085085921,0.16556296267936876,
 0.22266604225385078,0.5212798864885356,0.7081007390069872,
 0.19341461022271256,0.15229721345151392,0.7986979108392598,
 0.1589400780906045,0.9727430418528796,0.9920833512961712,0.6694630763053232,
 0.8070678855472919,0.7509626405981692,0.9579393977436582,0.7928462030886725,
 0.07441638717084742,0.027458464586915587,0.3362365497453925,
 0.3968501804697748,0.46768673171124464,0.639236388723056,0.9712063466764331,
 0.8326984682313617,0.4729073820355635,0.39532385912948675,0.5603835900458874,
 0.10672578781510644,0.5641931307269874,0.07678177663737229,
 0.07530940037945466,0.1704500875249545,0.06675439953319229,
 0.8593104250211604,0.5249811831937043,0.7078267745916582,0.5754892937278866,
 0.13337859192896862,0.016275701053678038,0.5719784525110927,
 0.10847952550246942,0.9815436720694103,0.8099013658280767,0.5894516516114763,
 0.9277018065410431,0.6572621781601435,0.3266341832147379,0.14698209458517675,
 0.9824210960283797,0.6669144925108867,0.19088605903012512,0.8421665124153963,
 0.6869812737953924,0.7238715618141213,0.7806433790020192,0.41074475637211827,
 0.7675120980950378,0.5550798353141535,0.8851758476436302,0.6753749520701375,
 0.31841248778271314,0.32132156624037367,0.9129152688932994,
 0.2585162408744236,0.23374209104775978,0.3672643262014751,0.0948979259689694,
 0.40387703757830373,0.06469137352432908,0.5274474411242209,
 0.7090882053588576,0.0619464266978379,0.7336132217038243,0.15641470631089338,
 0.5500563088259004,0.06042420500276735,0.668705917161006,0.03607896485128048,
 0.7668276406170031,0.1413025921509199,0.14857466407339426,0.5182906868826357,
 0.7029936519331671,0.6159364976278676,0.21386919997376075,0.3983337833710443,
 0.422961948310808,0.4905776531678735,0.4965220822218773,0.3251534035966499,
 0.6146222039039019,0.042053268552039835,0.0381750676117244,
 0.48113407840584665,0.6988217744274005,0.37007136389305884,
 0.46263664027786155,0.9041853611145413,0.575756251018983,0.3759466088905973,
 0.710146808068102,0.96718712680666,0.523015821522539,0.10193614933947792,
 0.948967404778271,0.5373635673957461,0.969204851417257,0.9005288711409595,
 0.8696470391990221,0.7015470250652585,0.4546697202466111,0.5384158412630028,
 0.2886164794490127,0.5341305875116681,0.7988499620629717,0.14366438240840884,
 0.7148520988505488,0.4638996332147377,0.3511228430597999,0.26035029589750025,
 0.05821912383061012,0.1859293572996208,0.44635571652147676,
 0.5231052491431614,0.18229935095025895,0.9834212328450337,0.9534589584760842,
 0.908476844779195,0.3211075437405966,0.9567570983895869,0.3336701896927231,
 0.007260085209720657,0.5091636646047288,0.6962099501975177,
 0.5415805303764433,0.3312052795993521,0.9801965461761194,0.9584110923458421,
 0.5139821017167586,0.7099143242567789,0.6548214878086772,0.38887067319321833,
 0.6297187314661246,0.5268329898785886,0.36237546237303575,
 0.16319780793909788,0.7080688297477212,0.14823457323936406,
 0.026831040038628684,0.28824607806062785,0.7989258704882614,
 0.07895778627165817,0.4464483339365579,0.6546385005391788,
 0.20134758148354415,0.7391619240028273,0.9653674529253434,0.8309506293807314,
 0.9442105704901986,0.5128866073410874,0.820036932325329,0.9516739655115127,
 0.29088722405956,0.33545216787547927,0.35928937993859733,0.9195483161846033,
 0.2959945287017779,0.043864129354386105,0.3004821550784691,
 0.4820200111046504,0.8804791930550575,0.29113813646916864,0.2067912373557561,
 0.241056672577738,0.2457280929771123,0.5856166847138403,0.006452151863629685,
 0.5783142191884034,0.5892462998751735,0.7288266281445844,0.7619237834396576,
 0.9171760051941225,0.7775483517641076,0.02279761872179986,0.6540866910837358,
 0.7409869760881161,0.42632085643720175,0.1317914123805437,0.7193999758035994,
 0.8717895941193653,0.7860996026285912,0.8010894122332407,0.5587672704736579,
 0.056554623510021385,0.8616072409043996,0.21230883504112108,
 0.6442449870290397,0.5324323007150237,0.2817925990586603,0.7248529756695672,
 0.7297880996056815,0.8327513570617031,0.09935598082007502,0.4771402382594695,
 0.03675605953532657,0.5193679124171717,0.4261140149104521,
 0.17385315835128146,0.18198527986849788,0.039777572815120066,
 0.9340913584146389,0.24644035086216176,0.22245891235936543,
 0.8176294408121141,0.6342523700550688,0.00874464016459231,0.7286167962763916,
 0.8936605295955184,0.5933477125183801,0.8317588229387508,0.2779287256593691,
 0.06423100178501351,0.49040479811777815,0.2509700284255416,
 0.3625516207527608,0.511127441319647,0.9405070808410915,0.3886960494424734,
 0.9427522450903719,0.4765466442421906,0.6369554018157145,0.8512094532191624,
 0.15549702241910257,0.08510640712964157,0.7502297293641381,
 0.3117378092566838,0.14661988296369222,0.16860397706504082,
 0.6971792076226802,0.4226879722616086,0.6101962634519347,0.8436869468556044,
 0.26626136119673127,0.6667845010913477,0.10803409720121049,
 0.4394924655493889,0.3523435535674062,0.8026100023043513,0.9607725548619892,
 0.13454884377204845,0.40001651683107453,0.7407534421218233,
 0.8110635287997496,0.11460325635367685,0.10263545565567633,
 0.30939916949681323,0.7347437822512723,0.4457572309296487,0.794761463257546,
 0.17688594615654385,0.10921780080863952,0.8173691085354318,
 0.9234228309400311,0.3578859677005455,0.3193837421725356,0.7036313272372818,
 0.8967747218550312,0.5751494516885132,0.9973889965680189,0.08809194337936865,
 0.18452740587376226,0.7832720694053129,0.5809819686562582,
 0.38340531328304595,0.5198115640715447,0.2519107416020855,0.6731841981211617,
 0.899972313851876,0.7890800463764563,0.16606239862523486,0.40171164404377835,
 0.33593236364535906,0.8415151344203636,0.019626855175063684,
 0.8208489102784278,0.17254499565335346,0.5891869457992519,
 0.29620649844809266,0.6234789885718577,0.9196361716282677,
 0.28621801199081043,0.7219664163328441,0.6096101206883672,0.6211405901795343,
 0.48003334836581113,0.4578650248268522,0.6637965355470655,0.2829793000066533,
 0.6533629668310513,0.4830416395436782,0.4621473188957955,0.580618880353164,
 0.3003497406449881,0.7840054439164672,0.3826694465237498,0.2328734278280623,
 0.7290192255107381,0.7707294797554906,0.4521627279157109,0.294140049769815,
 0.5997773393187849,0.9250018342019131,0.9533073713406546,0.36062568156675745,
 0.9465381599886042,0.520742827357159,0.7742639531198175,0.8186529756532377,
 0.5808484924992938,0.4191621323815933,0.018019727180537015,
 0.10160722059772043,0.8865543675852176,0.3212650748657332,0.9801367578101494,
 0.2766449718850964,0.37030350325387107,0.8890409976870093,0.962743633768833,
 0.6028143373624223,0.02692097867063159,0.8587487739059232,
 0.36330692304991286,0.8123058081112525,0.08622221881033654,
 0.8249665723397747,0.4493083052420229,0.6779962758063398,0.819522521089579,
 0.4434402817283595,0.22932790363507805,0.4142101715244987,0.5125164346584864,
 0.3400407304437836,0.9448392987760126,0.17416138747434695,0.4051981151745646,
 0.9887095457008785,0.9288708295692731,0.27242406902617433,
 0.09514522293943328,0.699364145614472,0.8232079071055445,0.9507655301282187,
 0.7764413257408206,0.604017823653793,0.5883545514845486,0.7440452225045182,
 0.8838678971711066,0.8208260094516094,0.7519589246477683,0.7931495533117688,
 0.6764418142533377,0.9144295080246152,0.6597578052339774,0.521752976269055,
 0.6718738338810072,0.8468867758782147,0.0036514554029030144,
 0.7939866847055008,0.8400064934793663,0.9147640841381703,0.29436499927718496,
 0.16257860746264918,0.5510922864860869,0.7808373029594183,0.5357597518703037,
 0.632033054305462,0.503959238025644,0.16576866926689748,0.5242383895763579,
 0.5043278830379723,0.518071623213349,0.615644456236369,0.46890095052655956,
 0.9519760866321791,0.5324948814944024,0.317631851854672,0.5407042473529193,
 0.8039156044350244,0.31839868080360567,0.06170411264699682,
 0.9423725403390595,0.27453745972042876,0.7162220246641183,0.7880842644397817,
 0.5224615094275173,0.748758451846588,0.21688663954479115,0.5701347639058398,
 0.08640828993010496,0.5525756471259075,0.764733499691645,0.10590533391943013,
 0.05797408884603983,0.09434739652503388,0.32388489470106185,
 0.7423505471055444,0.13942106653842745,0.9565217269469029,0.5378197626571297,
 0.032614820209605044,0.6605606569825035,0.9644656364875267,
 0.5975761908694136,0.7800061843464913,0.19310776428194698,0.3525886828704833,
 0.7235692781908056,0.034855518989434175,0.652286331788491,
 0.48889443273221234,0.38443047003852826,0.9808228607010069,
 0.6438763315686135,0.8611658107528382,0.3377090190169858,0.11489390641994246,
 0.251461018111076,0.9110375663095394,0.4801989728791034,0.2033285439669842,
 0.6412360680510187,0.951228120316033,0.7347512475070146,0.3249644147842863,
 0.6144851551468992,0.4373948741929218,0.5915509999479633,0.43052127379611527,
 0.38615338920596454,0.6966888992160163,0.6410001802822541,0.8309292104122287,
 0.6331386433691203,0.539091197864602,0.42438002581463796,0.3452861348472084,
 0.5180852054355336,0.886516365861866,0.9916539203693698,0.7588057780314448,
 0.21413648193178414,0.7227693027726995,0.47149041856734364,
 0.3773369296449439,0.2624599744319348,0.39559796640678546,0.6784268113660883,
 0.6305015395758374,0.4871102002474236,0.8213931088506974,0.7765917433216716,
 0.39546862964667984,0.783648320819187,0.44377908862752424,
 0.31774917147617376,0.522984673856764,0.6458862671681395,0.4038627742724862,
 0.28675395741804777,0.8714526200013651,0.6693564581958238,
 0.34282395164890045,0.4008253859972736,0.7992830230163492,0.3218003509826529,
 0.3553552627583605,0.09285176558953767,0.6334360522290152,
 0.004563401988842308,0.4255378602565304,0.8186008015364241,
 0.5102554713272126,0.47984449838958276,0.8864342052884068,0.8370757670894612,
 0.4283389442118971,0.16310006524155962,0.1489035537728497,0.3943433906681556,
 0.9189588901095362,0.7158518058494721,0.7357922556710689,0.9279207608608916,
 0.4642695911780593,0.9198598889276327,0.021958709006664634,
 0.6168828647605604,0.6932105018513334,0.4639307280217654,0.6338344241897955,
 0.6929854389499083,0.027708213734151776,0.4729688615112223,
 0.6366196630437961,0.0575049396597741,0.6057309939955337,0.455868494414152,
 0.548181939520377,0.775543628064888,0.5909858113419149,0.14932015486515948,
 0.17193177977118657,0.7139426103465655,0.32749600334437456,
 0.6466931973293512,0.2946535167322424,0.7622033823699246,0.2236412242356559,
 0.9980945659902736,0.8092074700932961,0.5109291606304884,0.5134651022787249,
 0.5478190641102009,0.4718242154630541,0.703036389032214,0.34693817332702304,
 0.8513483174505236,0.21533576509672914,0.5092486765534888,0.9156002575189852,
 0.06479010045094724,0.5613826856035663,0.26881377739548995,
 0.2717159173742797,0.8416510705132585,0.9221717629645991,0.05296848182203706,
 0.45556392676004465,0.9734787086268157,0.13468924215378963,
 0.3951456382572107,0.7170433643463395,0.784298964948669,0.3386439723921647,
 0.22603416171654667,0.22353498254765092,0.07341227783056059,
 0.7709742792580371,0.2876841910630038,0.9964977344455134,0.03132000113908506,
 0.10170093602247565,0.731181817257949,0.5687447810019154,0.4648055617408299,
 0.34646810995860344,0.7440830160852389,0.09010824253003147,
 0.34141248642564304,0.38192135545193484,0.281406034428498,
 0.28945322910380666,0.43554163826344083,0.6503010669584831,
 0.7380403988088776,0.9729822283504125,0.07471188289396835,0.9772946466484427,
 0.7985773954941625,0.9375106483023771,0.9082743810981095,0.17468570762305136,
 0.24294872296496128,0.6496435220243222,0.6014267736892223,
 0.09202074166933927,0.9584145740559956,0.8111582443253741,0.6276698069812591,
 0.11588205191251888,0.442514703077329,0.4051634875412655,0.4328217307886141,
 0.9135622142024277,0.7621098727857644,0.44332356819078966,
 0.44334739957672453,0.8543744013815054,0.8393543351869343,
 0.15730024707812895,0.667360432066397,0.8215565627857458,0.8800088467048854,
 0.19122504356987569,0.13724704198993687,0.17508778879050313,
 0.26885311208577667,0.32831132547123587,0.6919093756825253,
 0.13790772486303682,0.42379149504724456,0.8619897365655996,
 0.6423551836994995,0.28409750613392803,0.21558145149800234,
 0.006761679628383366,0.06582140282928473,0.3554515351204631,
 0.219416856294247,0.8875147795509948,0.23388425665981494,0.25815228747905894,
 0.47995450109236704,0.16740950786633435,0.9187279616617451,
 0.35207256040815715,0.4358073306164375,0.8010016746701595,
 0.41170554447658725,0.9453594299258611,0.6244683802426042,0.2591850917121069,
 0.7381836371237436,0.16843762939631368,0.7652543877190888,0.9318485358370932,
 0.113252433535308,0.4385230237301958,0.7514545911319298,0.8557910337126302,
 0.42405990834771556,0.9202390073669708,0.5710292550849213,
 0.45984939706167594,0.03907632029073649,0.47818360638547375,
 0.1733630549195596,0.9409168560100672,0.10527899719827272,
 0.46233665502388843,0.8747287450013256,0.7197024918437651,0.5791178978035724,
 0.6240395033338997,0.024785804092344543,0.2612988179853293,
 0.2990486500599123,0.4977792186982771,0.03256679248985539,0.4878009827032427,
 0.9725913841789263,0.8851478606915079,0.44365397134813966,
 0.051986615432927485,0.19860019011408803,0.951464955012429,
 0.014551704045383973,0.6876399422308883,0.3358797847327393,
 0.09156508325045687,0.1526101362372294,0.4527017468772574,
 0.18536970147439824,0.124393946936049,0.09408266420042977,0.0589077027898951,
 0.399235945812481,0.4679484416846771,0.6133141689112271,0.23427221076411686,
 0.6973701140905729,0.2221616588310137,0.25732739694774764,0.6689465422697923,
 0.6898363771924372,0.8122435354365666,0.5491899628596408,0.40032934373076545,
 0.9445892108336399,0.24443557816300898,0.504679023523836,0.6619669705744721,
 0.660810425153006,0.3381677190281953,0.2206457729275887,0.7934190340194532,
 0.7672193512987402,0.8894479191065479,0.21642598866565865,
 0.27850205290696595,0.34159610308807276,0.3179049164219102,
 0.17100986512719396,0.5407600580559828,0.07384044550554791,
 0.7221336142451917,0.18727894649747012,0.8609176926665452,0.76050163163488,
 0.02328534216191358,0.8913177619375214,0.9628746784531383,0.6146860886872736,
 0.0710250799549953,0.6342895220726539,0.06005242897587271,
 0.13199700939942116,0.13786312250387955,0.38341420259329095,
 0.02206083647194268,0.9732398337751882,0.7949750781402941,0.7417171932695701,
 0.37737220492508405,0.9223324242629964,0.47282369968703375,
 0.7899405140119269,0.10561869068416674,0.7540698929692715,0.4702170523343048,
 0.1717724901288319,0.6997972436814844,0.7738289442919024,0.8524324765202838,
 0.18804259390102995,0.44928970347311703,0.8294620872824112,
 0.28428548692598044,0.3582176671123366,0.45770493181712757,
 0.6917147928017459,0.6184071735143485,0.3155390469109399,0.1925624529120552,
 0.5097717809909407,0.4295701531901348,0.6993132705529136,0.6494478036329987,
 0.18719249458245335,0.5692401038862109,0.49522128155885936,
 0.3586195399540817,0.19270611469921262,0.2786099841734666,0.6178519261591805,
 0.6560115488003525,0.9052789013901298,0.5293184897239211,0.04013585967541211,
 0.8166029963362531,0.6551369931480223,0.3247076972804456,0.5113310679814081,
 0.8419744298741553,0.42822467697759325,0.49762879412873684,
 0.08820229100021737,0.32751872396031523,0.7223449005207675,
 0.4745411575162616,0.9224466259459347,0.6906900191111907,
 0.025079553974559743,0.2457161916401116,0.07251844185689416,
 0.14637308991913345,0.30809218434126395,0.828724276705322,0.5129123622505427,
 0.17076489636759073,0.4440230825260153,0.2418658119105901,
 0.27434387011415495,0.6845081108577561,0.286488626812728,0.3134583577345801,
 0.45641167550419504,0.8270012435754456,0.5850534945751362,0.8320667026186686,
 0.3799225600074818,0.1430355185526636,0.5082608053515776,0.40256129330915413,
 0.3685460722047438,0.8819181987547986,0.28900499599214724,0.8334319417333708,
 0.08558159649337904,0.5516271506944515,0.9281362406324288,0.7104485602389355,
 0.8071786532369409,0.5276747686176999,0.16058881543833892,
 0.21515953981731695,0.9260045044490423,0.4441344500819623,0.7460523119763475,
 0.5367065067682647,0.27014837868541064,0.3389202962160983,
 0.42860287661819674,0.24952970476566527,0.12156496328721855,
 0.23166501332453804,0.5679207856215406,0.14484614383894523,
 0.7916456384617004,0.6795946997951589,0.01808663215256734,0.6806604019285171,
 0.6923632722606783,0.49205965728880807,0.1467716374515009,0.7917796689431408,
 0.9120424233820718,0.8983243274896223,0.16832970087901877,0.9718366017538814,
 0.99297741123292,0.5100569925412275,0.4535235025756873,0.3643745468774868,
 0.6894941730071352,0.5747108199561279,0.29831173575334,0.13915664314087295,
 0.21006670351210044,0.47418254914433233,0.5246399339556048,
 0.4668002881402059,0.4880445643749072,0.7997745474782889,0.504342538338953,
 0.33531929612292855,0.09902339080434863,0.3947100400367154,
 0.7143768815347367,0.6362645229246056,0.23419894959832788,0.9519439369166997,
 0.33722657977613835,0.5918058921322894,0.8695129244871439,0.792613885714478,
 0.5644051174598742,0.09398834758446395,0.7277364930766068,
 0.29025614600488747,0.5799199362514549,0.5704990065177789,0.8980379767216039,
 0.5069312945579854,0.13631313057483574,0.29981858309971166,0.939080808164436,
 0.7705444925680844,0.11706320702180617,0.54354899872508,0.8539939637389314,
 0.7766312659344017,0.6753448190422616,0.13258282889913264,0.5690938896667503,
 0.009119196020340059,0.8001236965573119,0.4378004101354591,
 0.24974684122358393,0.41017787269470984,0.35303455086711755,
 0.15092721129749131,0.976091033016102,0.8879339866753333,0.7513074450068153,
 0.7252013771879275,0.5673261793249239,0.30061944491627446,
 0.10277643364794331,0.7446556960845673,0.18323553091489786,
 0.9175693509981608,0.6201748039662194,0.5935230485480868,0.27059976599181335,
 0.33222001602617923,0.5557951201095224,0.7602636840035448,0.9870351333975778,
 0.10337179501965121,0.6113572305396484,0.428720412665177,0.3270374087718808,
 0.23340388584365723,0.7087543580091636,0.5137341381089338,
 0.11736833920470735,0.04729897249058834,0.3288985186911222,
 0.8193000642122238,0.8556551253065765,0.6335324886954002,0.4861946334531384,
 0.24028463431775948,0.19645089149194883,0.2706491592175634,
 0.35256659084014996,0.32128285553044655,0.9616304278667552,
 0.12788171334775345,0.07307635729465511,0.3332279845790229,
 0.7908949953227666,0.33733288861693056,0.9732701602590327,
 0.41665347988638024,0.6127294262547099,0.1429643729204937,0.8455238894054946,
 0.2619055231332089,0.8450360304344702,0.6560913424154255,0.7218790359194316,
 0.9339762678566719,0.5043069767922278,0.16102043231774832,0.8405205733660649,
 0.05721186015788027,0.9442617724565523,0.37653262595905723,
 0.9146562923769499,0.4822528459110218,0.42297504610260594,0.9516479857760447,
 0.9693257041244769,0.9820467832650583,0.2788132754711473,0.7022471664157417,
 0.05058697221144626,0.905189123437409,0.7428460677927431,0.3135692687484871,
 0.41601121327709734,0.3783689210058946,0.6534248930022164,0.5806671301127205,
 0.7281681247163387,0.8781631997472141,0.021640498576405154,0.692700690820123,
 0.8613906454382182,0.6867601468813398,0.9037809448354359,0.3138241946440613,
 0.5455222710801185,0.6926408397653062,0.9708143993474486,0.12212073154685765,
 0.8210667986738089,0.3375360009958336,0.7329717629049284,0.2517067249560361,
 0.7506381015049609,0.10748428867218385,0.2972235361779705,
 0.44378215425183676,0.4197906096093661,0.1787295439722667,
 0.37002774053506005,0.944029039923295,0.7011934129463995,0.48784494923600474,
 0.30497267262431693,0.6077225341366449,0.15566458456144616,
 0.6873466293641645,0.1995799816644508,0.30480856269404955,0.4865247072993524,
 0.18184864215110363,0.6945260812719306,0.8863130809704255,
 0.10552708418120071,0.32990870840548236,0.20674920536658736,
 0.12830886023492138,0.034738447494373714,0.8485812743161911,
 0.4605383761602312,0.8328832078811504,0.2405250235752162,0.3585011024539121,
 0.17216929151251115,0.28850573309785255,0.6515675243146447,
 0.7604055362918694,0.5020894122730881,0.05433241366343633,
 0.30031930529504014,0.7609276229969327,0.652821519354316,0.7722714153784446,
 0.3869469951474964,0.358571756582023,0.3626634431411545,0.48486042184751,
 0.322144992563444,0.8695372728516406,0.35080994207255367,0.3325494780325796,
 0.35680683036351524,0.9771885234858338,0.9528170704735139,
 0.22040969841134428,0.7122177388784624,0.1923349080058887,
 0.39537688403782556,0.7595787001857094,0.34858346797385786,
 0.8376237363245372,0.4004905694311486,0.2882403922505491,0.6438339541417061,
 0.3741466052902651,0.727694642623286,0.22780412136035288,0.8468064141040821,
 0.39134242066036073,0.5263829831144841,0.20319374687973923,
 0.3765842524779338,0.5566844482669382,0.03403221680965851,0.4665653554892413,
 0.5941562431114642,0.2272695996477303,0.5045428236836261,0.9312054751492211,
 0.4684263816335865,0.10178480515335209,0.09577756672988014,
 0.8142877924874692,0.8918522972128622,0.9035883894279537,0.9386957368694604,
 0.7271456443738831,0.7675105101202707,0.6294333526193834,0.3116825613441804,
 0.9711743733090079,0.832139660363262,0.39038675126611777,0.1281956900345207,
 0.5311334176495738,0.8589603909997633,0.7604702670754667,0.9650976971447891,
 0.6495288448188664,0.6669586130926898,0.09290480168359505,
 0.06689575003068582,0.451682482514436,0.7224826180710499,0.08428290260025251,
 0.541354830978667,0.8277077782286635,0.027318955005618317,
 0.38715349956110534,0.30776631790512265,0.6248790835875857,
 0.5229107248620646,0.2432160325374496,0.8190255808617897,0.1684365967381708,
 0.5276026584004947,0.5637740459836471,0.13967829909500218,0.3290537560042158,
 0.5775514717234809,0.6125490150646912,0.9247494244621277,0.959016217024426,
 0.38905180321670174,0.22171685924904516,0.11014803082281821,
 0.7808785363299283,0.061630887939113976,0.42112114768037934,
 0.6067378305718694,0.4150440654626545,0.9687404149737151,
 0.028113292538519685,0.1690796700605537,0.8350031925141516,
 5.173567838469006e-4,0.706083129919444,0.4811365710812844,0.9899345391482336,
 0.17615891851629362,0.745629236819852,0.7067892177771545,0.9715069871021386,
 0.46855529386781525,0.40029865586219404,0.6163849790519741,
 0.06841561556414488,0.19810557868013767,0.6884546198113544,
 0.33322287652595206,0.9292530866702833,0.6027385769445529,0.8275535622422353,
 0.5647025987660107,0.7199024598631137,0.5952700542206233,0.04908339692357189,
 0.27046423885112236,0.3617111261685255,0.3445897748218364,
 0.008166883244275391,0.41239453972061935,0.7404681141033844,
 0.7027782467370033,0.25806244429457204,0.5639941672451727,
 0.31860317450535547,0.902297998572754,0.4644600239939789,0.8263540177070898,
 0.5506045793769164,0.3723158056572159,0.6613253747523881,0.8990166668155684,
 0.23191412609105555,0.1485245400120847,0.19106876569191067,
 0.6473941037064793,0.879415490677837,0.7513934219138282,0.5219811042780357,
 0.04119048072917941,0.6974013081709948,0.9372132453933799,0.7899877446397328,
 0.8419149973097183,0.4942450543893774,0.02007791617115029,0.5265143737143992,
 0.7795633706028021,0.5067251423799235,0.9944218836609904,0.4983427695992022,
 0.0186622525569492,0.9697062004757885,0.1317924610873097,0.903994748531526,
 0.4705788416481249,0.7238703881263999,0.5724378426694137,0.04405001808676412,
 0.7343672980790964,0.8050618975302829,0.3803812083297199,0.41658129179504566,
 0.9581253556598586,0.017027091680414053,0.29033645020016585,
 0.555829909756782,0.213220492520092,0.5024209563141673,0.8197564757238519,
 0.9968588308437207,0.24991782498789417,0.19425466878427022,0.50851479783235,
 0.3245034041179482,0.035375840146150694,0.06026524219099727,
 0.3792108517742576,0.9177594152095208,0.8126887000843955,0.9463439938935213,
 0.30771348287905287,0.9038724656755215,0.6236602446183598,
 0.05905964224558158,0.7796972256610943,0.3233894919061159,
 0.24359816152718383,0.8644487968700043,0.15432330132239658,
 0.3838845725136736,0.041394940103059885,0.05870410118707747,
 0.6029073507856828,0.45493030723415484,0.30657569160608944,
 0.8482163371300016,0.744762774374363,0.18757578782042272,0.06726044683611654,
 0.31363842223027416,0.7603352352407308,0.07222322608524401,
 0.1830458347310211,0.31368307233759396,0.30386706409017117,
 0.7493809737025233,0.15173651346840167,0.43391176561524714,0.655795045450766,
 0.13448351038277828,0.36680796614895494,0.3204418046851122,
 0.7089612348514978,0.5604140695921191,0.4540165470736518,0.8677232469659357,
 0.434994033372085,0.36303178083251797,0.15872511802496925,
 0.38785412876542313,0.3119410213951811,0.7781072867184561,
 0.25197423200927194,0.7492647179614793,0.9761307525219358,0.7823353497405184,
 0.11232855050007173,0.9431005399863359,0.5803271622890063,0.8755866416622646,
 0.9020075780634133,0.531654138728737,0.10713274301711151,0.37684818809556064,
 0.7227632709088517,0.3838096020500993,0.21887098894744972,0.7547726478280111,
 0.08272231867471058,0.626664960385613,0.8008479124829038,0.7076905008812376,
 0.05024629813143633,0.052964201272234,0.9178749147018662,0.071092702052731,
 0.33641203619984106,0.059547237492770466,0.2050570564079931,
 0.9508516256571271,0.04441343346436999,0.10124757523819383,0.60813257203504,
 0.7305069033405919,0.26429404553415803,0.5365207799766998,0.6309329621904401,
 0.15084853732344905,0.06533985950736199,0.1270968209195419,
 0.7558818363415182,0.1781535536003338,0.9638381455135457,0.05484448015707266,
 0.3616321695877741,0.03362257991082407,0.11585690498694046,
 0.2579671248682336,0.37862307315089616,0.02007854447747659,
 0.5313577828603551,0.764035525036927,0.06302433384485451,0.14956188831205697,
 0.39796589513952974,0.6723845232994115,0.4186523318579749,0.8856355274476924,
 0.6233858408791729,0.37896055411340035,0.4430814460184763,0.7106159284237383,
 0.15979063761888523,0.23358511758812028,0.7405250003187718,
 0.3594194806892792,0.4154259666504799,0.08207678012480368,0.2237001376371177,
 0.7773451383218768,0.9315210861269427,0.8485833408953468,0.2792805197947458,
 0.9968252763435145,0.290466230886147,0.0064118457393439465,
 0.18988419240288468,0.2574508201716249,0.39501544056855886,
 0.16893528220610932,0.5158242073551688,0.4749331813951383,
 0.022483620711401864,0.570539939489921,0.8349985161034076,0.2881932686450641,
 0.2369747472806677,0.9728322061046575,0.22771527374714384,0.9174024875127508,
 0.2677954653670622,0.45893675729540184,0.7408670985873405,0.268488493587824,
 0.49253388250500063,0.5305621766334674,0.7146537440463843,0.6642280714206966,
 0.5616424605874237,0.9641939008792101,0.6442417453511996,0.5021276692986029,
 0.13115117615129335,0.1421534371690455,0.2493996533367805,
 0.36496196324214336,0.05296899079760087,0.6166394588520031,
 0.3798626980953921,0.734994871422516,0.6907133333484105,0.724252704912156,
 0.27425095586080583,0.4359401804405014,0.2509358679199426,0.678429185957951,
 0.5544764125951996,0.8535275105034603,0.32515224974113277,
 0.22879719243185215,0.5131522213201232,0.1780464785180671,0.7609745907500731,
 0.2380711187859298,0.9767168484998883,0.020486407571825405,
 0.15791405329000308,0.2754443876430468,0.6097453620648287,
 0.47577877608007046,0.8512327430511266,0.7918769723811899,0.071081341744332,
 0.5327053328851276,0.28806345108219755,0.8384804761359506,0.4764667414668181,
 0.8739228009780813,0.2284838125910077,0.8511894884455132,0.7274633177184429,
 0.005124586951124499,0.29383998154699253,0.9158696898968114,
 0.9461730941625358,0.5229631594084587,0.5974378703244843,0.02991413603546489,
 0.1529433701112135,0.6237522948301378,0.46152284046790515,0.819696191690937,
 0.2767197544227019,0.012002018511292456,0.38437561343508786,
 0.4682433568493488,0.5525478673445402,0.9472952071773313,0.846897039002857,
 0.68085225300037,0.6312052036975149,0.6746036552319998,0.2165946279755172,
 0.4517993159737721,0.15997004738712173,0.5853866045535125,
 0.29425975244772473,0.8417283214744391,0.4869122814911713,0.855582447939564,
 0.1755791601982919,0.8203437378272431,0.03925742856208336,0.7870363306804592,
 0.3659534694450758,0.060185430488897906,0.8709451471374398,
 0.21726609878836256,0.9221659615866206,0.6792123093153796,0.6399211908668551,
 0.39909802584815524,0.5767006629768364,0.4578105075240563,0.994872771326868,
 0.9043560493453404,0.9795925131012319,0.9264524322196426,0.6771125496521283,
 0.7360112119660367,0.5522196746563669,0.4973309181099572,0.16494679226163056,
 0.6555841579370394,0.5658747373016842,0.40365424627788293,0.6008934215902104,
 0.7813706948441358,0.2787297364232145,0.8504186488180088,0.38152286745991415,
 0.5322459763033254,0.895873653353676,0.7440221022417254,0.28842135813363645,
 0.2667837760555073,0.942764270878942,0.8309722844013614,0.24684502940427466,
 0.8346008675410869,0.9679375961599819,0.9876135960455128,0.27819497666058,
 0.5602657588039226,0.9889404646978048,0.8609232384573847,0.2839980242933331,
 0.26852381981263984,0.7265052035388587,0.7476309479243558,
 0.42359199288121674,0.49295235993854636,0.97255093760163,0.8948199871302922,
 0.5570353793042152,0.49973926547855896,0.47145969530379994,
 0.7203114370498801,0.8074711078797964,0.3034531012349627,0.8082943153342254,
 0.8207077810089973,0.1904175824780694,0.1666633570983368,0.7748886234720893,
 0.41621762747673674,0.45439126840755784,0.8703416893434104,0.639832911112744,
 0.27200442188221274,0.8710488636338029,0.8336178930223017,0.4461612723528814,
 0.26591385178335414,0.23373609136566553,0.6254888054556093,0.801493807020633,
 0.6439969757319965,0.9007207478354131,0.9922500616924994,0.32956438712333835,
 0.7279910734138128,0.50598057988178,0.7085513656831007,0.8020207720447589,
 0.7373657621955749,0.0511380799351705,0.4192985097753793,0.15860833606092906,
 0.0493721447238098,0.40635784776669714,0.6319684247131805,0.9932905829031932,
 0.005161727639190539,0.8504905672828739,0.6051985864514433,
 0.7450914542504098,0.2003461109205611,0.2768310421427964,0.9320147455162396,
 0.13081134373628145,0.3596767653766968,0.9807507775924923,0.3679496653367753,
 0.7494536005090106,0.2769938917206547,0.07672138111702942,0.2092855810766866,
 0.9993026432549178,0.6995884525396884,0.7239934368760387,0.4114545953849438,
 0.5163330379943236,0.38931866300601836,0.02877898598567341,
 0.12923940285994617,0.24119771279200497,0.9948815146579693,
 0.2678624704599314,0.8181601698046987,0.11666893580992022,
 0.30198065470407354,0.4799091127635582,0.9013625999546133,0.5384874929270371,
 0.4447750656696483,0.17505897914468405,0.8309075475554344,0.6388463134467373,
 0.4135143364017233,0.7511825944992072,0.7170911872288909,0.31265729195916525,
 0.7175472295144697,0.8629436194153868,0.030929747490349158,
 0.9524174855526445,0.2343240911589909,0.1634681810688745,0.8357255482335988,
 0.6250221827574305,0.6801262436270998,0.5024925858655245,0.14316127388986266,
 0.4370645583036,0.8329590014495045,0.9108954508784699,0.6711413393434171,
 0.963854074736983,0.9252383362854958,0.17462001890267798,0.8511823601337698,
 0.5120146983313085,0.7806603476088648,0.36782377584372594,
 0.011890599076751274,0.4480984683527398,0.480721440321242,0.3357643808288183,
 0.5936197054638298,0.39278605210168915,0.9626097127618157,
 0.011628168872500844,0.13367624755873675,0.969329568863566,
 0.9895850674565216,0.13848696983989817,0.0871565947656231,0.6510671300360397,
 0.7019059765194059,0.2382108232758824,0.4958164338604093,0.21738959309978168,
 0.1309948324407807,0.4534769306747872,0.5023376755963003,0.3691863884316132,
 0.04691878134457805,0.6246832610841061,0.8800264341044386,0.528206295032507,
 0.1743658958848434,0.05104438303732628,0.41963404330730847,
 0.02096028061478372,0.01835654202755943,0.44565075407706756,
 0.8285493795869218,0.07501397435304813,0.057589902681214,0.1815636466032159,
 0.05715195002493101,0.736224373247168,0.3309053484717629,0.23921743418743402,
 0.9286811442366604,0.7203283202222688,0.2179110722334987,0.43888332758262605,
 0.6299638212632621,0.8217350230940281,0.10293910174751786,
 0.14813648829909654,0.32671915153942266,0.8599018997545242,
 0.4775138267021354,0.25724586535635063,0.3766789344256256,
 0.23617029371251297,0.2945655915929788,0.8715876635541023,
 0.21878759644556067,0.15028167483008148,0.1230099864155989,
 0.6867603505126826,0.18745274233249398,0.7540848053195784,0.6817676374462516,
 0.04360113054309589,0.719416235029303,0.3263653410866787,0.926062515879619,
 0.9945394547833978,0.7960919402705071,0.2573093823970479,0.24173265175947323,
 0.31796884906294776,0.3549756760187621,0.9576352999032393,0.5994308827559519,
 0.17675246445621728,0.5015554140553435,0.7242200612841896,
 0.33723223784145684,0.2287773092675519,0.6543123084103195,0.3308322158013556,
 0.797797092290296,0.25955839558488447,0.4657206163369024,0.356978759246684,
 0.297774616441812,0.1318944067039658,0.46788822809966335,0.4581439870934111,
 0.4986853198389295,0.6161050476615095,0.054074244177288566,
 0.7111928337987372,0.9038389021860503,0.49210420640585273,0.6426403100165534,
 0.5719239082835244,0.43956856560039514,0.13542681721039196,
 0.7675587651482959,0.7018387436995748,0.631384775061966,0.8218291126445534,
 0.24341711798254284,0.9950593098706015,0.7239227859887979,0.2500698380892572,
 0.5720486895732579,0.8291711998761286,0.5252522637733392,0.7021275791544038,
 0.9153884983231786,0.4897539638146895,0.28025616284822874,0.558143167623067,
 0.6510226106522802,0.7653953759438216,0.5388050080569793,0.42369588549350357,
 0.08085507979624895,0.20985237470988816,0.1367598508798591,
 0.7461992017022603,0.11150662986379578,0.49494889802323416,
 0.9078851629334739,0.23358906418419934,0.6717901232396972,
 0.07794767162491822,0.7096450401573847,0.5144562569486418,0.7224407182666006,
 0.6195311273148384,0.6027404923266163,0.9215794727364274,0.7063602626546547,
 0.9115082929087294,0.09223655485122673,0.568417977079866,0.8561325562385137,
 0.9983152809194469,0.8679696479512743,0.6194166182323321,0.17514067945626555,
 0.37118087988420667,0.7446346121950156,0.5235367736525939,0.7606843458158233,
 0.06292433803999187,0.63635151940768,0.8543755472528097,0.17440454051084053,
 0.015296669582702593,0.38129435640905973,0.8014080055248384,
 0.5254336797754406,0.5910190644563281,0.4035363661946867,0.4303579679842868,
 0.5022269977366899,0.2817535874268797,0.04207020224599067,0.696712155915659,
 0.5496354910760816,0.12756294103696264,0.5667086798542527,0.7759853962284651,
 0.2964361075099162,0.5554175656348954,0.3035969903263629,0.900586714516897,
 0.7777625300583997,0.4224827405964928,0.3719700080643893,0.3237365609184626,
 0.8816115512763851,0.22046177565191527,0.4682015427584677,0.9779624252324065,
 0.3629904978297237,0.2983615214519084,0.6885858957743853,0.7637323742943192,
 0.33730368419170564,0.44006498140434025,0.9005436225492425,
 0.5020753823402251,0.1516191243569227,0.9551116388935605,0.7174367082405514,
 0.05425607401931909,0.9310794027965846,0.6901665756095734,
 0.23251268611776832,0.5109931239435662,0.3660215806146011,0.1339865297560473,
 0.1702575113708259,0.0311813874584026,0.5200218492590283,0.2261093338252781,
 0.5315768233864968,0.8678534090644259,0.04011255439096073,0.567428926848862,
 0.22750385434527987,0.7639927237356016,0.30277000832511314,
 0.30817062807906137,0.06494830711074318,0.6922742085381167,
 0.2630219296119216,0.42490624946225375,0.6311643666434806,0.688014792461381,
 0.6228441641117594,0.21045358721987073,0.8139689537586529,
 0.17053386211473898,0.8589926730225658,0.05812005659305819,
 0.10281407179212987,0.8162266305121836,0.8413559155425363,0.8902730272211745,
 0.7602269922032281,0.059666721047882065,0.8154168695069309,0.611897016745474,
 0.7674185778319299,0.7040994115622031,0.7215209889232689,0.7868491194409488,
 0.9227517727748695,0.15554669735425186,0.775239252143238,
 0.022391505689687063,0.9578706177751131,0.13247154515764636,
 0.8778101574293989,0.7923535626498447,0.08820295811164758,
 0.34037715281539505,0.19771675170731418,0.7691082095756674,
 0.5220451660795552,0.010300775799505502,0.8297813220663504,
 0.9874503130111087,0.4367999415253212,0.5163150971914255,0.41753981013476327,
 0.4645159166021031,0.8672341285031415,0.5076591554922827,0.12396341741047523,
 0.590244467686311,0.908478485340119,0.08930267957164761,0.7874150436912616,
 0.8553432378186946,0.3447837387989584,0.5599035512133617,0.810331304522669,
 0.30173470062567775,0.16459324105595508,0.723717395188785,0.8114259200368608,
 0.3206403429407484,0.20965276787764964,0.884364548213952,0.28902372766164186,
 0.7686848382656665,0.8186003505875162,0.14710711795261766,0.6484979906703857,
 0.5197865661185795,0.5555022453866073,0.8245796018359166,0.18772812198473443,
 0.18269545705571655,0.7071951793391793,0.5996929770134127,
 0.28830425740780485,0.6458911690370355,0.8785640240267987,0.6795175892695106,
 0.8267069767503191,0.024518426075449073,0.5956468259322918,
 0.14688475384619193,0.2761708879714623,0.8894603359237471,
 0.21132429693476718,0.6510044653980018,0.04803697336763024,
 0.7487778077388854,0.5325102415430437,0.3980481916010338,0.8736956923529304,
 0.9006659813588499,0.586810600830882,0.2492023545007107,0.3544506274343533,
 0.8464636651579541,0.3067665901353299,0.20719094067405397,0.2929725257754462,
 0.29719494827068305,0.6178549673090863,0.3565322686735476,
 0.42608647933422894,0.2917626270317225,0.7581815019212479,0.4117621509092396,
 0.6945731080023367,0.9501376323415712,0.42515963975953,0.756967580640655,
 0.7459578880123818,0.5695777272729409,0.04057156469741485,
 0.11821435152654458,0.45412918540567193,0.7778545649475832,
 0.9354230803258384,0.7595533798138707,0.615391852514774,0.5500885097484313,
 0.806824256580039,0.8933274937776822,0.930773311031562,0.5085880446002533,
 0.43122430338741213,0.43327773606091047,0.3477625136916671,
 0.07215595215829151,0.45667446634138165,0.7282914033061931,
 0.4764882276492204,0.5547540208342706,0.5044488558087334,0.9798311550967025,
 0.1516093040089388,0.4533233563952471,0.16865426292640973,0.802556891128618,
 0.9673464534925512,0.08224458641530652,0.9238287175331106,0.1972454493706921,
 0.6494542783984623,0.8559238980881593,0.742617881500829,0.2445580153563388,
 0.749844699870267,0.2535506977736858,0.7759000192855126,0.9685977299815703,
 0.2727220087748412,0.978649500290887,0.5174147332601908,0.5775583307143509,
 0.6007432205143469,0.6576948062783678,0.050064122564909774,
 0.8820628444014484,0.9430180255513289,0.01781971611633104,0.7777130221316948,
 0.2754292383246595,0.5294783749026294,0.13944743392184655,0.9880775967218205,
 0.058430520203532565,0.11428268091256077,0.44098534940673095,
 0.6125593664400082,0.8600035516922235,0.5783900691003413,0.7934324455527788,
 0.5301683837212836,0.31771203003805926,0.12399913997001799,
 0.7481041496496404,0.09582516134574781,0.2742389956282617,0.3116504997780323,
 0.12648602321433122,0.3439482637604675,0.31853957427311663,
 0.5605364021471727,0.7581872373252562,0.9497386630266679,0.6167374454415024,
 0.7150568489625182,0.6295462591979762,0.903096301510231,0.5860903928179826,
 0.9412796416015876,0.030113776858194274,0.7745888461389588,
 0.40334270775993364,0.636206773044575,0.750710565803538,0.3514520386203501,
 0.6803002006491378,0.8869577616207255,0.14485581978172757,
 0.09333381013710251,0.18261534012876413,0.5641738925732143,
 0.8447003400389363,0.8624220189948568,0.9787548883970513,0.8829993445697194,
 0.8480792891283249,0.8371659581363766,0.5301648160611294,0.6453779026320592,
 0.14628147278886816,0.545991406029268,0.4953823961422317,0.3565692802101903,
 0.19512083519136103,0.6492575996736379,0.43782174561048437,
 0.09631874190069512,0.34620994988118137,0.7379813979772917,
 0.21038512356834638,0.17095793988014119,0.8674319680639483,
 0.9142689555985254,0.21489571031294408,0.7328824821510063,0.7767532196842029,
 0.007130430420077105,0.0047581585159488515,0.2842992103310057,
 0.2911188430258832,0.828656971320317,0.7806650304104131,0.7551852322412312,
 0.9723683285166337,0.7057935620128688,0.3539270161541438,0.5481991110283382,
 0.32546349934657626,0.18965597620136976,0.034459855435438724,
 0.08513329115899448,0.26207056988144384,0.6725321746678333,
 0.1667707837505057,0.27863492447371097,0.235484673747006,0.6854534906095874,
 0.7803405452655623,0.6271660365708001,0.5818560009331157,0.523996232534547,
 0.8159888484953316,0.6852578534269685,0.6812223921197951,0.3648568735809048,
 0.27858133887847625,0.030809970757640803,0.6056280530325178,
 0.2237886685828161,0.7340116076286848,0.717433592347825,0.9245785481508123,
 0.7689094247704884,0.20104095186619653,0.8489943173469948,0.9008349774929234,
 0.16118161746742476,0.34611255194161505,0.2538322510409118,
 0.09864380064787248,0.3405350312898481,0.9470527534662122,
 0.24336883571042955,0.35059147040256433,0.9618462590165822,
 0.5564309587357112,0.8802120066141869,0.814686606395331,0.283577465072794,
 0.4250182408021759,0.3373291614326479,0.0863718425933313,0.20526196533378438,
 0.7368506243977038,0.9685035052538635,0.3931943327126737,0.1828007881412722,
 0.5138024676437383,0.8925130031620201,0.24521343409462315,0.9698975403473413,
 0.6974545330571839,0.8703678471518449,0.6130583564797537,0.8236366511729036,
 0.573376927192731,0.3132433659646676,0.06377848205388159,0.1042701737560443,
 0.2147075904072594,0.0752820873537875,0.9194412638148853,0.61354689356247,
 0.008477907825598208,0.4991105918581149,0.2470703682235008,
 0.5525389697808039,0.6703569466260295,0.7519762902029091,0.03479266888940491,
 0.7299543099024111,0.16593664812077358,0.4419384566498541,0.6259265634794483,
 0.7532867111397948,0.1693393580393363,0.3189903710043668,0.7329930671673608,
 0.8043464757836867,0.9382823440443325,0.6219828458321363,0.12180617807736616,
 0.9607468387109632,0.28234674061090936,0.6465748812014178,
 0.20693281351972304,0.19401665848427385,0.2074845830035411,
 0.1788545431852252,0.6496269678386476,0.7762578716755046,0.4598759235239631,
 0.7212518663489158,0.5245390701192676,0.45971125700397786,
 0.17527028081888585,0.6251751201175786,0.870955452250509,0.9983974678406351,
 0.40461791468590325,0.08507750884685528,0.16895941732552133,
 0.9061454932720872,0.9890909162055062,0.5560451822268073,0.6166967487042738,
 0.5895381806397377,0.2830825889769636,0.003440165049929056,
 0.13788316620187002,0.8324965433911133,0.5772513350810579,
 0.41774341562515405,0.38545341455865023,0.04193393671568413,
 0.7091093106986689,0.6243917531673676,0.7413440251911614,0.7403372959679367,
 0.879701858964572,0.4445860278859589,0.41762785137777536,0.35555003273179864,
 0.43370790994409303,0.7176510521989864,0.9967086939378537,0.6991020803824053,
 0.32053993312770357,0.7730930311242046,0.2072857816662973,
 0.19161518862662608,0.7208690430013766,0.06423703737613007,
 0.6872111163762621,0.8842228014712761,0.7375915755912209,0.9763914897668264,
 0.45000042285322206,0.49664879745018986,0.49597716150199866,
 0.6056329065575389,0.6815850786672567,0.8981846548654692,0.29843921906106086,
 0.591185708150026,0.5545782717666216,0.07894407449640917,0.68081339860264,
 0.32399544855826934,0.29957286263292127,0.5856919372215297,
 0.26852481656024485,0.024608457970754505,0.5507842207121836,
 0.8491185818811311,0.3661076659527761,0.8550509748001655,0.5118973139814278,
 0.9747098068269948,0.016310808755074735,0.7115065853319894,
 0.4926947922271754,0.6384187411758288,0.02812736886081124,
 0.13258779920214125,0.2519341419280263,0.6665152655786222,0.8452955359704347,
 0.0829200224749107,0.5009453459070011,0.32278614347987244,0.6964365989558373,
 0.8720802530728853,0.2847290297996661,0.6204852110208638,0.6693710341946248,
 0.9126066378604574,0.7726226010125511,0.243477727769696,0.12342713314839204,
 0.45645342015960866,0.05310072275341915,0.4673095705021232,
 0.0018736158031575911,0.05702455614865032,0.16168114469399986,
 0.9761993748437499,0.9308268459095458,0.006741479135794037,
 0.09774135912671322,0.7110603388726547,0.43505993063735826,
 0.9871060351738988,0.08987983500670904,0.36810811721837755,
 0.7801962671982091,0.29021245461556944,0.275924887451255,0.9530515097186772,
 0.11999942395504903,0.8735611073424857,0.36409508470618823,
 0.28813549287300777,0.35155831637130774,0.8839810262048207,
 0.5618671367926862,0.6240207465890768,0.8990586215036838,0.6871337373902205,
 0.3621190868342722,0.7325350671665688,0.5698102822670785,0.10543151442894405,
 0.6593790764719566,0.3782552806711813,0.7889379332537907,0.6286300765764923,
 0.278173227323919,0.9834842365049266,0.3786800179333708,0.11472693287837732,
 0.4563212928534427,0.842975188937438,0.022937548225988147,0.3922438764529417,
 0.990685737259693,0.19405950410550166,0.5062138852566616,0.20780456814178772,
 0.41559343074746324,0.5017313090104863,0.23141474713983523,
 0.34051757095932234,0.3869691130643703,0.7271083533258014,0.8087265913116304,
 0.15853451964327026,0.805411377108805,0.18902332513423148,0.4535566986386028,
 0.5050847929053764,0.8877932023583309,0.9842478465621082,0.23837776990100112,
 0.17671986118175986,0.029793566166479768,0.9206535127670017,
 0.4649781714584492,0.43311334347310715,0.672831978908607,0.16816462926668785,
 0.17673173461672476,0.5957180972775055,0.7615765966612087,
 0.34012263899615747,0.7575208265826211,0.31164044918933376,
 0.9528585569438541,0.8955189137595498,0.8951102401002804,0.7307093156447477,
 0.3129558541345556,0.569250003837751,0.7409156881449902,0.189392507836726,
 0.706575676379822,0.014883280871086702,0.17197230200954394,
 0.012710239025640058,0.48661352045662687,0.9673065391307927,
 0.9467298675101691,0.16684862935878508,0.5299857294753684,0.9102068908504892,
 0.24154130458062506,0.303208003296758,0.9708453073875783,0.6600871321650086,
 0.7039990727306681,0.6349232535920679,0.7887403317844863,0.5165172321949975,
 0.09878959555435785,0.16832537951747395,0.9331001353346072,
 0.7790960704783941,0.39366695500582993,0.03622310937254025,
 0.9256848924584631,0.9479344767806527,0.3351513547180688,0.9378939894013343,
 0.7121175667349656,0.579621506267964,0.07676865296212443,0.6356857755709172,
 0.6755665053186197,0.5360226455643304,0.33369733110516386,0.6923504928584534,
 0.9008070509807912,0.09615541784463066,0.36438794389805684,
 0.34136410982476306,0.5331332104385292,0.7705150701735036,0.2588302296068542,
 0.07381052955342382,0.20001850431253032,0.8179030898767619,
 0.43163485687895387,0.13386404439974542,0.4624092965336478,
 0.5378243180013993,0.11804976032559644,0.4041107634205916,
 0.18538596783324068,0.9820379190717191,0.3213343991283334,0.5511807540562909,
 0.9173475970284353,0.6583745895948063,0.12467619442596933,
 0.09833559213291576,0.7729899183806906,0.22072955183325604,
 0.7441471090131778,0.10094146414197103,0.07942432361261409,
 0.23853929038807875,0.7162205246698489,0.8853593100853624,0.2639392395258451,
 0.4411643783563266,0.5320935108989894,0.06423215972389307,0.745640416085563,
 0.3694216754842776,0.3361750261333366,0.8454320475978798,0.9002796035701851,
 0.6163941061155861,0.6392114332550379,0.21326228902793853,
 0.11323879448592233,0.033083178297954596,0.5799472579771983,
 0.03406778647475095,0.9483742085186578,0.41996559380213316,
 0.6354563571198555,0.23848378293488415,0.5117003266888569,0.9355464443952693,
 0.6817553464017141,0.5508152013296475,0.23242681768111884,0.5885384257150806,
 0.10191881655566215,0.4407587908762196,0.1794374458994863,0.6132135501364395,
 0.3706320156801022,0.9325622622333516,0.7586599138202876,0.09935127381513154,
 0.6899532214799772,0.9797119347760611,0.5198807023245147,0.07507158232050559,
 0.31336141562593633,0.838295504990718,0.20361540220074514,
 0.24061390542329986,0.7836581024600242,0.3732697390503722,0.2577621202489351,
 0.6332100235062053,0.7504718717874678,0.42815082344342137,0.62598815531587,
 0.07095113481142135,0.9869430291188437,0.7062612109364103,0.8251605946542022,
 0.8474035108199267,0.46975059103498473,0.5742254776746196,
 0.14774166637017605,0.9682102837565602,0.6858320483434467,0.3365925426749754,
 0.24180696755920705,0.7237755862024018,0.49264959353441873,
 0.7240932829109159,0.4870297712728171,0.30080314850046985,0.0424244750272722,
 0.6651946841266879,0.9566622758937615,0.5894202330411411,0.05211277830300087,
 0.9765459855516783,0.0025792613548272225,0.9283716048039924,
 0.44122529797198196,0.03595416575746557,0.9432069079605374,
 0.21702342573285494,0.19555728969671493,0.7042847040493485,
 0.18448485767227352,0.24241288188437615,0.5686094713419252,
 0.23499353907385268,0.7038972937365586,0.8115650015960055,0.8622612058396052,
 0.4060605539846084,0.21571740784100335,0.1631525662390817,0.6039437332475169,
 0.4465504351956159,0.3575074941231561,0.4604988120711665,0.47149291366644097,
 0.15646534501017528,0.5519564588877606,0.16241369405236183,
 0.009307807144854574,0.14929952397704338,0.7937812713434051,
 0.013321221053166132,0.008911347399057301,0.22696413076489486,
 0.7814182639362244,0.9367864387290055,0.49253410294723154,0.9626821133308778,
 0.5213205124698409,0.8121624707687267,0.13570372441717682,
 0.004340108953558408,0.04449959200486708,0.1713928414353767,
 0.8269950286586736,0.45768172397281226,0.30739724758669307,
 0.30444972297881123,0.9087436249965343,0.7595673849795288,0.7797857752805888,
 0.6236103800048585,0.72901162174947,0.7555855893534873,0.9177857299969835,
 0.5073650186467357,0.0040088576283204835,0.39562635178526073,
 0.30243521959505437,0.3456720061634674,0.2653686129287921,0.2409222960395181,
 0.2362511640793089,0.6101928419125787,0.923583145312778,0.8479670106552687,
 0.5938611892636975,0.9693745440288456,0.38247504862251724,0.9957217617854801,
 0.73995054276543,0.4923327806888227,0.7195529934677903,0.03498957293210381,
 0.09133125213709226,0.3216531054061267,0.9828156806376379,0.7068134414284442,
 0.5348112171286974,0.2971661671553256,0.6388830172688018,0.8682727455548382,
 0.7761899233389513,0.19095520214180683,0.5930075253203982,0.9930382733662046,
 0.1770824565337098,0.37557183950420503,0.8077604875218374,
 0.31051796219010497,0.1100906558626783,0.6763088029623099,0.7134985016820686,
 0.2429319534471197,0.4646799886391625,0.9383589942416879,0.6029384596270488,
 0.8504164374126407,0.4596307183548143,0.44442146132516547,
 0.022707187377555282,0.2513507386256738,0.4135504369727988,
 0.2196684514552869,0.14789088350182633,0.6964167711375542,0.3905387610298803,
 0.7580185370488426,0.6545553545866926,0.2724894810479731,0.5672655426150182,
 0.30005063571523294,0.8691934728558578,0.5371467471863434,
 0.10144403691215387,0.14476933937742653,0.49291656414102847,
 0.35889465082201333,0.5347560861539835,0.5521289737381929,
 0.24565097829985383,0.171716928952525,0.49373172579958136,0.2591946637874355,
 0.19242808059969285,0.12724802257337198,0.6206903100134329,
 0.9094967104850331,0.9220524417006722,0.40798582889565194,
 0.08490579699091005,0.11307700970130119,0.24336688580972632,
 0.6764035963374352,0.34509021511399673,0.6881656887506,0.2914592317760354,
 0.3758305197403913,0.41716997312417314,0.6209128280494247,
 0.43310633209635674,0.11542248385019716,0.734639621117461,0.8108291693379437,
 0.22494906089262134,0.7948558497102605,0.573742294409635,0.1374408669835313,
 0.11046010314904409,0.891316773557433,0.006893775419243076,
 0.5242449036755512,0.9386255132134776,0.28487568844748146,
 0.09640605989415663,0.10487339409554863,0.2412442343714174,
 0.6597056317147445,0.9497059917721562,0.9500615160326505,0.03214695623887265,
 0.35256464914273733,0.08570668054360164,0.8660237958804892,
 0.49625522544936274,0.9854369983839745,0.2432060269199977,0.8990663690195718,
 0.2360042885170991,0.31152152745342887,0.5920600511219407,
 0.06226931606013142,0.006135463349719927,0.3331010074489855,
 0.34781827173076496,0.5074519688609411,0.48556398755702346,0.757839984056528,
 0.27389099490505775,0.6492794161528248,0.726929562668092,0.15355049469183768,
 0.053135577093158415,0.7623454002422787,0.24963336405065406,
 0.8623772907467214,0.4862692253364702,0.9451524010979006,0.975069477061323,
 0.6173470563517418,0.8736753421905297,0.2387724260306332,0.9555019757601779,
 0.00294757144834501,0.8144579823860587,0.502619151280903,0.07508931590714585,
 0.6817811998669994,0.04706557674676293,0.5016994561340544,0.5184550045138359,
 0.8489501604186808,0.6254132637355581,0.4636557911420912,0.8864293987823192,
 0.2154762454806316,0.5649763773295576,0.0844396070836344,0.1604850310049054,
 0.4280308883371542,0.4235801916688636,4.495358572076391e-5,
 0.9132070349530114,0.6591795958426425,0.9408842637147415,0.5626123920134457,
 0.3290631824206147,0.3294469987374935,0.1063681169018218,0.6459410124351054,
 0.4584211936515774,0.006986361805744301,0.25295745513536205,
 0.7213050941406517,0.6531079787020999,0.960723489992863,0.6536953622751926,
 0.3943831271570155,0.38498873166191183,0.5354205102326645,
 0.22962126329069876,0.9456781890382091,0.7307415781421231,0.8568486930321906,
 0.7142007991702841,0.7960069015863056,0.14885781990328395,0.4740873733991049,
 0.11545651999975359,0.37633187877532415,0.06730357294756795,
 0.6571963375610322,0.07758111567412407,0.8745091135887564,0.0909064697874491,
 0.2515443696343175,0.8608630392412832,0.8689759808968776,0.929451182859056,
 0.29941391972586917,0.402776211098384,0.7538195877588922,0.6889630661929969,
 0.3949552561321994,0.16036838234193052,0.6275517321142904,0.4383341566357194,
 0.005467962928105363,0.3720674733514542,0.8373995845554115,
 0.9358483607064615,0.7375540136778758,0.3351642089671315,0.35923379493771135,
 0.7320110333064178,0.36331813121511536,0.9160862078713304,
 0.09861344923173454,0.18893696278371208,0.7410076867724027,
 0.6698775869880215,0.705244223068996,0.9429403034977438,0.9806490642296289,
 0.7116243300188825,0.17688012277715215,0.039911911285569035,
 0.5638708586471916,0.7214282071238859,0.8888345189561847,0.6205756528362065,
 0.5295891809093616,0.8303856907380653,0.48115806614653,0.14069109195164164,
 0.32030702750592277,0.11694404370140943,0.08564786174877581,
 0.484854470316868,0.38507917398209734,0.8779557153679709,0.10567374816093622,
 0.5005751341901411,0.24132750158176342,0.2673250023864382,0.3639536197572869,
 0.5314752498192613,0.4019978388595936,0.7508453384968924,0.32124337638374934,
 0.41632136905442485,0.4070150206784031,0.5420130203246314,0.1583096232103015,
 0.012869817293597596,0.4204429895088373,0.10011398965876106,
 0.9422627470219234,0.2579158447651829,0.21910616297400232,0.493224315254874,
 0.7992782742406306,0.3783234217226975,0.9961603125722861,0.6823587764288127,
 0.03827700457993699,0.8750238023426535,0.803087508669631,0.1362758791997255,
 0.7323160773021729,0.5181721545857592,0.8863523877247996,0.5330986816717435,
 0.9947726292384336,0.9191956250467946,0.6008381362338336,0.3875456013099725,
 0.012040316433111764,0.9341899389413704,0.7731069026874254,
 0.2519903783642029,0.3323308146440672,0.9136297799912031,0.4753123542097999,
 0.36823627603108966,0.4697592567852844,0.06256989688259196,
 0.1653279205617505,0.7398212810644668,0.3858074962364366,0.8760366527475466,
 0.9365530915366109,0.5507604677551478,0.6745061215671199,0.1684863416636586,
 0.8459239095046331,0.984196732107823,0.7235759589100751,0.7719095100141722,
 0.11076817452209464,0.5259817255565985,0.4602055500145896,
 0.12455916325952698,0.5893725752703243,0.5266769200804091,0.7155744595602349,
 0.05213737385861683,0.23756832947228002,0.5838177386589661,
 0.25746845728408463,0.6413780164011696,0.3900582848121159,0.6660373087766931,
 0.9676755813093909,0.05228905842727716,0.9052954502169447,
 0.021986181735873078,0.7381326828777556,0.44815487682538724,
 0.8428324219365335,0.8529307115994099,0.9235514847594914,0.5838011835326414,
 0.3496221168577728,0.9733152822978286,0.08786847670049758,0.7856138014553112,
 0.7995927964844399,0.37548464051005204,0.115993596831102,0.1322886262293761,
 0.27324908286658167,0.8363108056776779,0.42225426559770285,
 0.5926823294112563,0.1394473060248832,0.23615050484128564,0.9483676391161316,
 0.47119081461046963,0.11894061302550818,0.396861061196911,
 0.40682041267011637,0.43642383315575217,0.7156924074134762,
 0.0199172771610322,0.1073600056801669,0.1961212158145359,0.32911077102104935,
 0.6333057480257082,0.8862079992717827,0.5687018663483752,0.9195785628284979,
 0.5124022547869098,0.056099318654619146,0.8033150579635135,
 0.09954757767709266,0.6378331436965514,0.9619600112790041,0.8971612862103394,
 0.25125530472908586,0.5493077966601436,0.3555244569294578,0.434332702128969,
 0.4118027823301821,0.03688572929581091,0.45571913923825313,
 0.49795998411247444,0.12541560318003253,0.7385713871583832,
 0.3640725616444773,0.6589426063113016,0.9796123419990324,0.3758054956039969,
 0.7154841373703459,0.7560768367445225,0.7738262507772258,0.5810288147010847,
 0.29331761005535417,0.2985375642646003,0.205226500064287,0.296473076418606,
 0.29085249929497614,0.6440543892541194,0.5314074475292789,0.8089430563526763,
 0.24823897122367122,0.7284041596184405,0.11800646926349012,
 0.33380655153927086,0.4922124290952111,0.8897118864714902,
 0.014508876949370642,0.2578481467061815,0.051227254090441376,
 0.19134251023488125,0.2116088464628627,0.9828744118317625,0.7598333089206418,
 0.22867342513034772,0.8660756036172528,0.41848636153110386,
 0.27708218822602304,0.6402401759052695,0.0815744262991317,
 0.28351234186916674,0.4802784925732513,0.39724544728989253,0.720905937354353,
 0.5421771696192876,0.4676719131815221,0.8213751825866199,0.6242677809571267,
 0.7862653998404243,0.577576394664726,0.535713460674289,0.5049266350120004,
 0.6124764626296605,0.26718567750538436,0.10730614773198455,
 0.7030073480606676,0.13287924058186507,0.8372200482556809,0.6004713407722755,
 0.9440817378309114,0.991715044649402,0.3637945831417324,0.6434141934436107,
 0.8673209630549056,0.9268297734099363,0.784836991661984,0.607680479064178,
 0.7885532518024041,0.009532956257361835,0.9181392583984774,0.948724028119545,
 0.8221569328767442,0.9070768460149972,0.7636956576244229,0.33211770204512436,
 0.9123460666351878,0.6036720208444124,0.7196673615707005,0.7534609828804378,
 0.23767926972343068,0.18528045925376513,0.17571451702086383,
 0.3289190698992589,0.18238074232484303,0.988365838401015,
 0.022633006487847163,0.9207449949451115,0.7263558477992045,0.154966855681677,
 0.8158773105231192,0.7253058104678413,0.38813363730697203,
 0.09683486527574403,0.20608613518157692,0.3617617638354985,
 0.5079940565027107,0.8106457487668656,0.3969072690146934,0.3869472124843264,
 0.4943146846476698,0.7126903796263399,0.44054713797937095,0.915605221037243,
 0.8181768794264402,0.4254592879592005,0.9072864430300734,0.7159392248663167,
 0.016291574670879938,0.2701659089872981,0.1856668410333192,
 0.29195478946143694,0.7146296420659317,0.8784043960143758,
 0.09262459179692117,0.10725613211918716,0.41686519230050734,
 0.08910313760692334,0.11633708989495539,0.07790466389882078,
 0.580617859644185,0.4160625770786527,0.5375117581515453,0.12830083532939607,
 0.8055546781837154,0.22989792113911012,0.5479082946582429,0.7110839881134425,
 0.5323803158929484,0.8798211980421393,0.5763742225255892,0.2319782636010308,
 0.8623996960755946,0.8633544109227683,0.05051260115580036,
 0.06527618927345691,0.7874027467308254,0.1822241078627772,0.5662342965345001,
 0.4761394886864838,0.6278852004415556,0.9744741245445246,0.5052359622215309,
 0.7339122216112256,0.8923434076910695,0.25883453881813845,
 0.19051914943274006,0.09724275506799684,0.14325614849905055,
 0.7529371711814203,0.0392979439642831,0.9877015097226182,0.08263823268291759,
 0.09675264557359886,0.9614965564084953,0.39876037725464053,
 0.5949007541208261,0.7893917046181871,0.6232415279565937,0.7201449230986795,
 0.9653780039640009,0.2661249575584066,0.9917354844032269,0.6530049286491661,
 0.15106761100403598,0.10073993140723592,0.43136887076238106,
 0.7319396212853171,0.013778554321195902,0.8566844835122502,
 0.7341452320450844,0.8970637979334394,0.5731037105328374,0.3622458984691034,
 0.87204736349166,0.6245965692281041,0.658908879651684,0.9965519857891766,
 0.42423463420245466,0.23400380017976918,0.7748720394609134,
 0.8625230795241489,0.45183210350043346,0.07560335152208031,
 0.5652369494492424,0.7911475355717719,0.4030208115524698,0.8567512960127001,
 0.42410172444501704,0.2764169654553752,0.4875324914799418,0.4850903739033451,
 0.8852304536213652,0.4569428222195543,0.22764943106150348,0.2731076430677635,
 0.586289959602102,0.3846569443054513,0.18904807734873919,0.06591162112222237,
 0.3331332994866425,0.4552296441803846,0.536324917495785,0.09071432932021217,
 0.8456524692731237,0.5585457455448196,0.7095698451837897,0.5600130505338283,
 0.8920285760982999,0.876248307110949,0.9703494539364964,0.8052038334251106,
 0.09630276568349638,0.49747071572950186,0.4757616787846053,
 0.37435441347299336,0.7571016896129028,0.7604890487838986,0.6535838508121339,
 0.6174859476922518,0.21905108760486558,0.8478928653747074,0.071273671799995,
 0.8538661872584472,0.1450288904247362,0.4042391430610436,0.08507009201497451,
 0.7603982742963832,0.975484850199441,0.3083355159448904,0.3607574523203664,
 0.7662831843215578,0.6762106394156467,0.7150586221749597,
 0.021503980666289446,0.3720629296440663,0.9734580054374576,0.925373726415312,
 0.10620464574334654,0.8050469991269066,0.6415321601158688,
 0.37315329995837676,0.9309646658256222,0.6782047066266783,0.5871366552566044,
 0.1629458114174115,0.7042118322252285,0.7442287993243504,0.44634252287327114,
 0.9159733148254381,0.6545234042947828,0.15247431411851498,0.2129234735616249,
 0.27728890417858143,0.08325277503628736,0.41607832857595906,
 0.5506840503425107,0.873457026284872,0.04856542486275228,0.8623741171341629,
 0.5462120607449947,0.9521022366258114,0.24619563861421634,0.2613288132278575,
 0.946436443537829,0.5291752581657014,0.7944275146617199,0.05861571841228552,
 0.925419117182169,0.0540444479454687,0.9860048561747582,0.9901301002656354,
 0.07905569895750508,0.8780781343172142,0.02605820283771476,
 0.6514200131301181,0.32926182795283165,0.46233052931577756,
 0.9402810336820888,0.5828318740524848,0.40858621059594846,0.6970574886936116,
 0.15650459550748663,0.7284168346962553,0.46936138005923,0.697858922300073,
 0.5538658244060394,0.1423442454995606,0.5639537218983225,0.12746679258901317,
 0.3759117301256133,0.5806658550980136,0.8618329525840539,0.7056524142579645,
 0.09743173820200901,0.20157793153680625,0.006260530259227348,
 0.4825745527641384,0.1451636752313592,0.4920318919596407,
 0.005487554569823638,0.23585746105562488,0.31611953847783436,
 0.0019436808340883571,0.48063411377577403,0.22369617619173987,
 0.43801067570017294,0.7559181142237519,0.7396255452895898,0.5234480432808334,
 0.2190659545414133,0.06882552596111069,0.35804085958455967,
 0.1667237446087555,0.19790193374319898,0.5146931585256411,0.3853387687605655,
 0.5669373558406122,0.6867631860404982,0.9896889977535017,0.07411423291506836,
 0.9023585292231118,0.3236513495612475,0.3635528481561624,0.48780533910547574,
 0.898051643150832,0.4242657442353457,0.7632986593574074,0.31363677939226986,
 0.14781452680540363,0.14112300745716966,0.13611565167149797,
 0.8241523024841862,0.8719263862774544,0.6586049365169439,0.4846875941497083,
 0.7402365286573549,0.4496913966992695,0.31614406671543027,
 0.16029859471568275,0.4286876474501937,0.73557409000545,0.6733519687816499,
 0.03184668757933218,0.12420561708594446,0.4209059640412096,
 0.8449795153424655,0.9987941681412463,0.9041181817517291,0.21089106052930207,
 0.4672872300922791,0.9979566519725909,0.5924301697134555,0.16934781654543263,
 0.13027949054864896,0.7580678946218158,0.7232258363149926,0.8505072225859321,
 0.8328308426841442,0.756342707211694,0.4779449333612943,0.9598669651999218,
 0.25641468200529793,0.05342800740163478,0.040059676036094394,
 0.9955308942773495,0.374082898791698,0.9392172869856725,0.5294751689130314,
 0.06778100937423126,0.4086453027580774,0.8206875214666237,0.5760198324854758,
 0.7231413909425164,0.9803105531048508,0.9609258848824248,0.4187372740108394,
 0.799559105404805,0.8945591100223456,0.7295276221072755,0.2996837815102201,
 0.38400799279009823,0.9750829256823498,0.5547793360498958,
 0.01594197356431537,0.7564070077995797,0.5803714832761775,0.8644717553639655,
 0.3134761229105416,0.6643872656608358,0.8759562140874505,0.05542177762639933,
 0.7995355757034618,0.31529854009933844,0.7696404980572753,0.7628244710078325,
 0.4558133892711792,0.7858126045613781,0.40073526235902235,0.8390303237495039,
 0.9012278146468485,0.7093981891003814,0.5488054261562921,0.4192554336778982,
 0.6585077815514083,0.21147183297017003,0.8084102352643421,0.82879550989146,
 0.514047859710382,0.5572277991142339,0.4753764514136196,0.006865728410777816,
 0.1808301488802493,0.9728552117398114,0.19014323169944602,0.3028484901551044,
 0.8340858377654895,0.7997293450216466,0.04715971976895661,
 0.10700094752141509,0.839396897715484,0.13696352021382363,
 0.07988192949395401,0.23554615610789686,0.8348295188838222,
 0.6725826526577344,0.3521542267850174,0.24832611674143745,0.5652449767007994,
 0.45246978977171315,0.7692949882005735,0.005536597081989947,
 0.5111777937862996,0.41734452838416947,0.8436873166611035,
 0.16059242826242637,0.9753896278255805,0.3706238026085332,0.9559740396230076,
 0.256108774466705,0.012414736304470209,0.6413158534192582,0.9315747010183102,
 0.4712650715416288,0.8847094805656551,0.3494405717711845,0.8717224043609879,
 0.48014553972782703,0.7800425813440685,0.5992176697921825,0.6411926225461846,
 0.6952396507630626,0.094556589097059,0.02046260942808542,0.5596037852640472,
 0.5700656716636824,0.9481122830534218,0.09038311009508859,0.9721489834664558,
 0.33153092473169066,0.43343017685539964,0.3893330650277572,0.596993122665438,
 0.5455618084746483,0.9498056424377403,0.8998590074937167,0.8801476643122739,
 0.6201471103507343,0.7622300873909945,0.2376361945185932,0.43822200664890687,
 0.31574434447680444,0.14386360686777688,0.35035596945154035,
 0.5066349059162285,0.5083983674524887,0.9839220269328932,0.5890804794034387,
 0.8265009085550102,0.01996304591757936,0.7247523579848958,0.7437726610501333,
 0.07604843412527262,0.971670179898496,0.2452471779620068,0.7182497791106057,
 0.7013914357006413,0.05628887597371224,0.06500740094183977,0.33334183299387,
 0.5922945634318957,0.38311348444027027,0.7629545920649188,
 0.18081833141542814,0.5374294964693677,0.9213678754893027,
 0.31577517155027734,0.48130017594534036,0.3041378151309553,
 0.9644516619582871,0.9188884353514273,0.30844727399942884,0.8391187643577569,
 0.10342937388803475,0.9548550721407525,0.5697486527054343,0.99669882831747,
 0.4107644755773817,0.9557916773654395,0.2936805611783516,0.5973574080146323,
 0.21089332763607793,0.8114635981515183,0.27974801838673546,
 0.15948975733606086,0.5115222674308482,0.19007790807622527,
 0.6764940395506089,0.428803734497599,0.9129124748345748,0.409102578810137,
 0.7359777815659394,0.9472703114648537,0.5395547556067437,0.9654248921107128,
 0.7625848405098418,0.5944632139496682,0.11499779549889533,0.9590151463991066,
 0.29830541712153336,0.6117645398359738,0.6279691153586718,0.6432071592188409,
 0.8661173417289211,0.6757443494377999,0.17126080864656368,0.9005212139520318,
 0.9770231579697988,0.785060034895254,0.020150250750359255,
 0.09554569181520589,0.5372787926875038,0.7225037478689449,0.9343098735523281,
 0.7376041126064533,0.44207330099525677,0.313215595676131,0.4612640512431879,
 0.9489804890932035,0.25041880357001134,0.7086313060906986,0.3485577256547504,
 0.6072350070992256,0.2681671698316338,0.5367805652178262,0.5392514983291026,
 0.22252734720817202,0.3569882467013993,0.24971799633085723,
 0.8678889610866094,0.979743672257869,0.4732562871424626,0.4888065171451843,
 0.5734309980591965,0.6492584353490886,0.34072197276784466,0.1934064217643282,
 0.9034249225088984,0.8725305321856915,0.6461296737306872,0.7705870178157824,
 0.49360114540605216,0.8346660937653886,0.9093536842494101,
 0.47637522606751537,0.3643968471843224,0.2515127095714641,
 0.021165648650951385,0.4257495777907304,0.6019709504843952,0.541349592101373,
 0.14656516712788714,0.20009110336596803,0.1666129550183285,
 0.5145076759695698,0.5114846203152765,0.6967639260303153,0.6155471113582889,
 0.4040997927133375,0.6906182287193026,0.8113768650899214,0.5505572931853183,
 0.841070448381742,0.7628538346763757,0.34386453022287167,0.7774830956069465,
 0.36331548192277996,0.8737701414242374,0.7048910530867354,
 0.040600135051943265,0.9630397671951307,0.05912644593742078,
 0.30878852439032967,0.004477951554695547,0.376665500661864,
 0.6530125847799244,0.34452662793370326,0.34783149821934534,
 0.015528235979290761,0.07573045938872514,0.09798382307816111,
 0.16383826769284815,0.46560203026007707,0.19926847341203768,
 0.5270054090024927,0.19696955473773015,0.2848485389074735,0.6584840323325517,
 0.7051580324642794,0.5774487465029898,0.49395471623729237,0.9399018445773092,
 0.7803288930867092,0.6742838223650612,0.9530279707563144,0.9425037166830826,
 0.20195104949639775,0.5901898811190777,0.045041066190601775,
 0.7415580056933653,0.7812973971936901,0.5268629977302249,0.26100643310950733,
 0.962140205560345,0.8469037266964061,0.18576283006088978,0.746159213253314,
 0.630335462648516,0.7926996042051196,0.17093545361892515,0.8012386620246958,
 0.7937810810987648,0.3463804879244049,0.3713400012534319,0.15249361399018013,
 0.9012361710234903,0.9130698928147004,0.8376147677320933,0.6663114685041267,
 0.5735235727972545,0.20010875884306267,0.42942859968228797,
 0.15339590125114377,0.10080599110358268,0.8469764972524506,
 0.5564618485269349,0.6728468731357461,0.5122138192947268,0.9617172798441337,
 0.5412738401140509,0.32377702493941773,0.6017145865888257,0.9774941681056506,
 0.21424518091514955,0.7245462804018856,0.44327100672726083,
 0.7413399467172326,0.6316013248228456,0.5238040379830463,0.6963036204760767,
 0.8950476871185304,0.8962276686948657,0.7718263518546424,0.7343764220205429,
 0.5688455182329153,0.841755409068898,0.5961965354596538,0.47242414189092863,
 0.6973871653182909,0.766475197329662,0.12628493251237705,0.832430638051844,
 0.04593028236580654,0.9809019228594267,0.956456846494842,0.4656608068616246,
 0.21884415775031407,0.4480820726952264,0.9834681438595698,0.6840665517013577,
 0.9926719992712665,0.8759984987685122,0.3501630030416989,
 0.0019243919975975299,0.13386769065527904,0.05494240538036066,
 0.7189853163621901,0.5623421371836698,0.5743757952424244,0.8393870191579784,
 0.6261747771407169,0.1275893281347389,0.10305435621530334,
 0.09647627843516304,0.5643861352205666,0.4063503460044622,0.6787498029457711,
 0.7612386305574401,0.31464800821276084,0.910671327275816,0.10878062029058,
 0.22449276683044905,0.6187300068414319,0.25264334955895495,
 0.17352085395375683,0.6069307950787968,0.11736354037595231,
 0.8463008649297203,0.09876499987313991,0.6376728125344715,
 0.09176668665611243,0.6122281981169088,0.21142815847932384,
 0.5440919979893455,0.12556100294189176,0.8311392241913808,0.6335501044973587,
 0.8158471812693874,0.4446209587941632,0.1371496115165165,0.14436837359955013,
 0.4742672149399847,0.7318254910880326,0.32006017716703594,0.1410903055846967,
 0.9056771120757798,0.2241831222859033,0.00801116888072384,0.5047564069507428,
 0.40852175796501355,0.18292347200835257,0.28889656125381313,
 0.6682562588592554,0.8362828332370043,0.5508365573273615,0.726472569277632,
 0.6262289623064013,0.4727340792652599,0.4361375081562421,0.8913180840369014,
 0.1954967498832899,0.2499425656885832,0.7377926209781894,0.7069892593718851,
 0.8898723382368893,0.6015824286035527,0.17094580381973223,
 0.39282831987413813,0.13691289861374034,0.5554429589686829,0.404722856421875,
 0.30068733373753664,0.5633508368043315,0.945734865275015,0.4988572787138086,
 0.08291331688313475,0.572504676672251,0.9076559919485575,0.5236358380353507,
 0.5904735350752903,0.3603675478516538,0.8798122466425589,0.817322339142408,
 0.6193374692848591,0.8039443848306167,0.8058333062652664,0.4325999691658836,
 0.12247974662064098,0.5183477278810247,0.22339548518957142,
 0.04516275174488538,0.2544860571130706,0.8778182294556881,0.92492864642601,
 0.8952367770246511,0.004313095956479329,0.9184064196605444,
 0.7831088217345237,0.012917500215143818,0.3706402848650281,
 0.9047877216297525,0.5908923253750438,0.46160542804726645,
 0.02189963010627405,0.380560487524914,0.7502658618775269,0.4945774678808007,
 0.7110879111075351,0.17594747236230046,0.7664140536870891,0.5841332610423577,
 0.6101882560237128,0.5614411776698988,0.8340951646873255,0.9926553624539558,
 0.42679637966921924,0.5941251925098461,0.43828737289342423,
 0.11002004944558952,0.9708078022384514,0.4349690304463527,
 0.45300379132561386,0.23709032025046095,0.3154068857176736,
 0.30445783598683485,0.692629875631336,0.25642331786700145,0.1251603769735492,
 0.5610483358035969,0.22361815859786371,0.7871815370179692,0.255406732379015,
 0.6516543931918708,0.3708163613895439,0.37503544749980344,0.709551119478813,
 0.8128553806908356,0.9027016330134185,0.20099747166008664,0.796932638452754,
 0.09248908254144528,0.33971916750592157,0.12044786841145738,
 0.2515041994539705,0.3715082045216569,0.6158353709235318,0.02139360847176075,
 0.8191495888827036,0.868813629030724,0.2120252609496437,0.7355437405460011,
 0.821710457451567,0.5466497905271973,0.13164417333814005,0.10796912406629977,
 0.03340021753452671,0.20206175170551077,0.9548841095125755,
 0.6637155807393863,0.36794237352618286,0.23319879196972448,
 0.2170438565132602,0.8104737214211245,0.1828831504763032,0.6626007189376814,
 0.8052522034617904,0.1473977944341398,0.8524874997072057,0.2904469098846145,
 0.9984668078497095,0.6723398363702371,0.20872453238726962,0.9921758333598468,
 0.4901025879595656,0.8808299637378858,0.4806392349667241,0.6420858215193099,
 0.17230829673426817,0.42370918848582406,0.0025077832498022756,
 0.6698534430557812,0.9354468510488636,0.9783561465328511,0.7778536405189577,
 0.500391603770531,0.9465285475146232,0.6012550931727971,0.10584383754645499,
 0.09367978048615866,0.6526278007368443,0.30797320124013583,
 0.46610166663384844,0.18539848395229674,0.07407227783133552,
 0.07145940453152189,0.652850640000201,0.4285142899027675,0.3169510448012639,
 0.2991455841210705,0.04631646477176965,0.836118485432455,0.6473620132248222,
 0.01590580379881712,0.7798063459202755,0.7088420774366947,
 0.49944839438468025,0.11005135229791974,0.035306714824460905,
 0.7813977906175363,0.030506534906277416,0.013493069212780484,
 0.7779490250497781,0.0816576485791004,0.6021645233382367,0.6083874159462066,
 0.060008528447713316,0.176176646286741,0.3576531798603536,0.9218373411618663,
 0.28273556348534123,0.8047772365690212,0.6341850170011136,0.3231123368278557,
 0.7546127146544105,0.44079259523217185,0.8443593900029627,0.6733198977247138,
 0.5165336449911508,0.5107373426963232,0.32056491116390406,
 0.40200270965357443,0.012719838859525212,0.7071726264247372,
 0.7333812606436876,0.22411287839762395,0.39396022906957273,
 0.7604325163403648,0.6135085757279115,0.8029167087103218,0.9711188758631623,
 0.6707322715092369,0.5537552452070432,0.8142124890299051,0.3322590628886498,
 0.8905648345285179,0.5081090699705809,0.4443330141907432,0.13057397562819828,
 0.9866508308694047,0.6597945313806777,0.6144764164147035,
 0.058977195494904544,0.07907348735464514,0.7954216253032907,
 0.9765639882728461,0.5735752002530352,0.07523354497284362,
 0.25711688549910705,0.795827250515175,0.20515523289124804,
 0.06975083811704241,0.18804604925642632,0.5091479837283462,
 0.44114295699017525,0.23503622558856085,0.43152951826388186,
 0.8870314044525889,0.13290960244050243,0.14093935230711108,
 0.5758757597161215,0.2722320138944623,0.5886128178748008,0.4279767913069642,
 0.5985783014195069,0.6478422354975362,0.6839716063345593,0.5799066341230146,
 0.8418759582509246,0.3574457273192194,0.9958129380686875,0.46172966584523445,
 0.37655335999648476,0.0025375172539094937,0.3638211291486879,
 0.8066151339820132,0.04431168658057505,0.07887260264713913,0.449040583748903,
 0.35637477622568925,0.6289416640668541,0.0018048506577900136,
 0.8866114191555013,0.4631958512955927,0.5478791904025639,0.17955130538166697,
 0.9396326288987014,0.21374213170649148,0.9302884823765732,0.7706802283205312,
 0.5254442879024531,0.5965727537104303,0.029606365303283644,
 0.9671383067973641,0.4449644354286584,0.7945797852735752,0.6249093508372263,
 0.013562766458636188,0.30781546635441515,0.21064051701656517,
 0.4369828688723658,0.28337793944396505,0.45722948136396635,
 0.0727292755429867,0.5011689378425181,0.5922279967948421,0.9458995268866264,
 0.01942778270264245,0.8218141786575917,0.04488058049753141,
 0.043517133988431156,0.6257754975802822,0.9525476932990444,
 0.7631641120083682,0.6591677663157458,0.7228093035674048,0.20198728274927635,
 0.22072985810829515,0.7843387583979217,0.43046387021445787,
 0.2840269806796828,0.8393758131731789,0.16066577627074252,0.5939951122897476,
 0.8006452178972062,0.7309251130651006,0.1638679881363052,0.4720636479381619,
 0.9455807555083535,0.9939141824090907,0.5608583742216269,0.7015392689558693,
 0.5133500521603177,0.7952781236476126,0.15437283869230645,0.4736860784053858,
 0.7653644734824158,0.16610563369958298,0.02608321050005391,
 0.9470349600565932,0.5581270797395008,0.6864696772620268,0.5352268489436469,
 0.9007972232308962,0.025334471029894834,0.43531462296063994,
 0.1664604264157603,0.5664830049496898,0.32859302891146436,
 0.040509294434464604,0.4272920019154308,0.8174812261699824,
 0.8049099819692882,0.1696039916962686,0.5288426774748515,0.2883998829078952,
 0.3452067929623246,0.5152269643588665,0.11969525275747317,
 0.39052272221033335,0.7370850509542307,0.7644992626824753,0.8144980421389223,
 0.4581455603031682,0.26614702750518426,0.30197378339219694,
 0.39121802912664183,0.18063803460606176,0.29578213130534536,
 0.640393952051694,0.5420627791613626,0.7377604505775167,0.4682538757878888,
 0.921438219780245,0.08450524250347868,0.08712599239167795,0.5928550032946207,
 0.7112079015215944,0.16267558108139712,0.6099181477373746,
 0.40368011704311824,0.8894178637997688,0.46460313138511833,
 0.12195819027038457,0.16782745558313883,0.26755567153949966,
 0.23086259605586046,0.4832435120109708,0.9856930234588247,
 0.45606271507396245,0.8109054066058922,0.1608073761625476,
 0.05441556438555384,0.3611028521453893,0.41411462543853217,
 0.10646980320098554,0.15131714376790262,0.3999689247644165,
 0.30376192426438475,0.1854233589005918,0.0405263245360149,0.914874934240534,
 0.8584149948600226,0.7391479769913206,0.6645527880023133,0.7628546148689322,
 0.9885531902483528,0.2166908674307657,0.2301648981125688,0.5361651835776319,
 0.7389055231634963,0.19847794488214532,0.3490867078788177,0.7198604487106675,
 0.558135425670691,0.49792733937990863,0.3475137032393806,0.507167837667749,
 0.5172757448570561,0.1624064833309592,0.9634243625581134,
 0.046275699651872504,0.8050149725640441,0.5722074695920155,
 0.8345373391566554,0.7553988101003829,0.22833996285476998,
 0.39368768080211636,0.31357970217462383,0.05297144079523519,
 0.6657859951076134,0.75203611484845,0.44421406378736383,0.23627962186769635,
 0.2708779351557381,0.7965584152935775,0.031114200559771543,
 0.9981476646760907,0.7251122302282098,0.9262194430837285,0.5751347156374463,
 0.6118072277987957,0.2811588262693685,0.9010167478003044,0.20188126326148548,
 0.814260087479449,0.47039033357896765,0.4513404997413686,0.00797392002520858,
 0.2306267651897893,0.2496215347448043,0.43461741067625825,0.6294226363534758,
 0.44426914067376055,0.8188747486155761,0.6442512553640571,0.4038363700945504,
 0.6398138619927312,0.9616283963662265,0.7695862313654506,0.9312410351700796,
 0.8479748802716394,0.39798114264587925,0.8985638568961147,0.6199239046806233,
 0.6854932887827005,0.14702782974533768,0.608748170099202,0.6092660872937,
 0.6590526597395697,0.7139647651343449,0.8006973654942615,0.30573007374897065,
 0.7067147368180053,0.4319136559298318,0.5185998776094061,0.46376360614868306,
 0.4093073853217649,0.2095253682762298,0.38919364446235705,0.9354599450540142,
 0.2674261119132382,0.8134207884567612,0.11655058705440513,0.9725972979284224,
 0.879974991573574,0.07214310026903636,0.01798534818888331,
 0.07413015070739215,0.10437938478719966,0.48920168483931603,
 0.23075108210024442,0.36275356806313264,0.6727977177506675,
 0.8124847934452667,0.07093062073942225,0.4773943442625489,
 0.33163977636640585,0.42499817401885776,0.8098797584876152,
 0.7853134286396104,0.3199509187350398,0.4692508829163522,0.8240608516466643,
 0.3490726027609611,0.5317979158047395,0.8337966213887772,0.5691057595263732,
 0.1425981640172196,0.8203694994575921,0.7744396144852318,0.673090385318224,
 0.8475499727444118,0.7932960551201095,0.6335474993012319,0.6039618352170673,
 0.12413938611552311,0.4105038564848298,0.15392120872740356,
 0.6436391005772076,0.18095869725364722,0.4631959470944905,0.8272491078403618,
 0.4659602685459572,0.33440874365578943,0.17670734037664892,
 0.1542239712429423,0.3689294734011501,0.07782739081964074,
 0.13694060548519216,0.8698276451439295,0.8109821963202086,0.9664690046343558,
 0.04166388352502015,0.21944676649706074,0.8002620938636061,
 0.5457550337912415,0.05996894118935725,0.7065528452514108,0.7079141154474777,
 0.17043601949176945,0.7035419334058676,0.9725089878873628,
 0.39176105385546245,0.24504578541087452,0.23806178813581025,
 0.9251480737158743,0.10393368738687425,0.36705243430212564,
 0.28452227191352286,0.9374176710802968,0.48722375170350474,
 0.3285785698899135,0.38544948143417324,0.18787430909670633,0.389961679926644,
 0.643244222738984,0.8863348192762801,0.7578504436962958,0.1734787030426207,
 0.4182078199080319,0.037349681269839285,0.18133741596735953,0.84907508605083,
 0.5815681351116125,0.2572310796534645,0.6711772857423619,0.5129104097409943,
 0.17526727339200054,0.4968534771146882,0.5486596162055711,0.2996917524863406,
 0.4648610325083754,0.45711527961855436,0.8217678204750285,0.7524426951047803,
 0.5906244529967097,0.8257821069103146,0.3353702497487622,
 0.021876256109188086,0.5640667746614559,0.02647655292892559,
 0.3326591157510357,0.038456633285451014,0.4865635370708856,
 0.37954387394188827,0.8997791534043628,0.08806281425109441,
 0.46296237957243813,0.9895097619981127,0.6858877311759949,0.7975524845239186,
 0.2055285596393024,0.25417891220878486,0.5714759849332685,
 0.052270858933074926,0.12146478242560976,0.9816569116359586,
 0.836879697823717,0.7435620097205446,0.9315076143654692,0.43433769942447675,
 0.8246430668528115,0.5683742102125237,0.0715396810356721,0.20194486557456848,
 0.3361926223089954,0.42863836365411134,0.0070744108609152034,
 0.09371142254764864,0.051083231207189383,0.41122334506302594,
 0.23781699317320382,0.7572870897537752,0.8565147139059407,0.9954425504403028,
 0.765236720146498,0.2693473413093811,0.3640529947108684,0.7745018050703945,
 0.43678637050620517,0.9684458300870445,0.977386399239234,0.784772440605838,
 0.8092011127989376,0.07918605943152746,0.8613146626943063,
 0.20074786130194489,0.3991518728914444,0.14453517843762176,
 0.4034602109010954,0.5788551210181803,0.012751374227747414,0.522540149279345,
 0.8257395339468868,0.4907838861332008,0.6124620892653456,0.2618840014124263,
 0.6732818836943817,0.39605856579389187,0.02866723857387976,
 0.8999111147077967,0.16126430122881652,0.6749252067701984,
 0.23738713463250438,0.3412957802382294,0.5943321103333773,0.971684370640908,
 0.09171746247422663,0.2375696482697216,0.10309427128736115,
 0.15646518725354785,0.825327427993087,0.35345956585113003,0.5391388550332662,
 0.44318832497924465,0.5268309126582,0.915536093227066,0.01830659561533221,
 0.8099393519522075,0.6153624869619736,0.7277645523358678,0.6578141907408663,
 0.3502056947157608,0.7774044859469716,0.2199570688699678,0.2406532519034157,
 0.9662503593905079,0.32587020395688615,0.7848059625906334,0.5302617275908624,
 0.0876709142283183,0.17173853922647575,0.16903424808909573,
 0.3517035000441554,0.35179835358260036,0.3638244496311399,0.5058464003054467,
 0.5518868264767708,0.22768880779395162,0.40572636246088,0.2528910111687501,
 0.1399273137413013,0.09522759403700343,0.9224521406230526,0.6896360285950521,
 0.3584591783802673,0.23693567295575435,0.29870153516419007,
 0.24377416585224054,0.4154480417050994,0.6429313258830587,0.9296264427043955,
 0.03545836740318942,0.2922322569850967,0.4032345401674108,0.7175501396846844,
 0.8086269825139323,0.6538331255801421,0.006232430573789083,
 0.3857137607078708,0.32224140978322835,0.4530500682818658,0.9866661488159462,
 0.10047775905737955,0.3327846682562473,0.9714313464211922,
 0.24626026284108904,0.9064252807879651,0.9548429781535405,0.0602677758953708,
 0.5823412887278351,0.9904727197257149,0.4605779543464844,0.25170740606631614,
 0.703143322495788,0.7654810376170416,0.2602698244959608,0.4172461338680784,
 0.9178072185615769,0.41801950841919255,0.7090571687145166,
 0.15029802613034093,0.6312277080540477,0.08696184082636216,
 0.3803902184948662,0.046395808576784936,0.4662284919749887,
 0.6229111073626723,0.14547938898079926,0.33470183075742277,
 0.4814730411075133,0.08819037785565309,0.021649666762208142,
 0.6457660793038029,0.28042254148828283,0.4145157272987049,0.8595946905589757,
 0.5570989729972263,0.4187499176391787,0.9566523880350015,0.10681046166582542,
 0.5266957954274141,0.20818647410135815,0.500623110468525,0.26344880747837607,
 0.2848790947561819,0.6813060732316789,0.7540583143469676,0.9090031357682967,
 0.24200628970785198,0.0561865894483935,0.14646847276747066,
 0.24162702343360842,0.6765423488962802,0.3443385623727744,0.1701069226983548,
 0.26493541949073607,0.39105957017298765,0.4052317388386849,
 0.42797770743062413,0.8440920857805639,0.22777311348997276,
 0.057000822280159125,0.5515994942252956,0.9552778577352736,
 0.43206215012622184,0.6134968017298875,0.7073229255681027,
 0.08092441772755876,0.34579242098956486,0.8254099471007356,
 0.7073079314027354,0.6755237473350078,0.4421525793586951,0.7774027781090043,
 0.07600278033122054,0.5840597424737376,0.46698169147542723,
 0.04865336381788887,0.003278803889807169,0.06506727740884666,
 0.6300985374223825,0.07866549839687487,0.16049436768359615,
 0.5804186004800904,0.8277305923052304,0.2503557070905994,0.19029014591638815,
 0.8221374175394551,0.3383784866496059,0.07839917881101954,0.8658758030822555,
 0.0010655649786355248,0.5827606695347635,0.7082943180490734,
 0.22334199193612603,0.10421758226864908,0.34455298236160503,
 0.041015285826120795,0.09605750382408851,0.6359900448043542,
 0.5358179390615743,0.8087460160450873,0.5363705548972175,0.12938314024390485,
 0.9057425561305665,0.3541178315529804,0.1521625124788486,0.36135887208788686,
 0.28092440843532185,0.9357085478329844,0.9484854888560036,
 0.062431429217959966,0.46170874609520085,0.7825079821591049,
 0.9388908052892512,0.4552869050301438,0.2991836253076098,0.29157797896556337,
 0.6726642701485939,0.4344252659822043,0.4291642106155611,0.43941840144744393,
 0.1838974784057653,0.3172321255645575,0.7038808084786454,0.9762620706811882,
 0.6194515167709178,0.15983564654217597,0.655068334676639,0.9483523367881076,
 0.9580567722699126,0.027939446984478478,0.5709600419520547,
 0.46955363984764886,0.6439957198756192,0.9273256006956395,0.4976150996175752,
 0.4408948249088257,0.7761873877489508,0.2627227104925194,0.10226237478733635,
 0.8847109867866854,0.28876638094990126,0.2967422695840348,
 0.26311816487386586,0.9907375128129939,0.3971934453886331,0.6234659170900201,
 0.4442831721955297,0.269195585372071,0.3792732580759347,0.7166928598093163,
 0.28627013096384823,0.4463204870322436,0.27432522675724547,
 0.8602792587332894,0.02303788345531066,0.6570104572099839,
 0.35602932187067904,0.49075157861372043,0.7108286871873319,
 0.21774538145706113,0.41585500921414154,0.33148727800527955,
 0.008182773671033483,0.5366985666502517,0.01637896126853533,
 0.6092394350634811,0.37324482104492485,0.36411867163413447,
 0.6413829940531361,0.4664053914916839,0.7339776248954527,0.14334778968349238,
 0.5262555868001806,0.8023463568972526,0.9520517282908645,0.03090113828515617,
 0.5766206376311909,0.6656392009023152,0.11319752774031877,0.540224985810479,
 0.32713443002394105,0.34047662021119185,0.45733161590517646,
 0.4766536281198408,0.725122480175908,0.07745148854724071,0.38076109602104846,
 0.18077221201703886,0.6621666567439365,0.985131091456896,0.2916972528170515,
 0.9408966198644536,0.9306896955755989,0.5041215985843948,0.5470034450495528,
 0.7798278542396403,0.13999504854948652,0.17310005700871112,
 0.029634866539327898,0.9764819365762178,0.7107980775663645,0.398250525740383,
 0.5491208830305467,0.39134885616319093,0.3853338925375275,0.7175500673419583,
 0.6937804437364434,0.700553380508059,0.833912964007027,0.29576564647917913,
 0.8032886846743232,0.47249829611986394,0.3414175889180602,
 0.19495757984958872,0.2535554993746343,0.12377296676362959,
 0.5749715579560274,0.24227548597091464,0.9526124260499691,0.6440491647159995,
 0.21305391054220824,0.4793076130292563,0.9639350392097961,
 0.36254631137153104,0.9436644296715601,0.09301667660696034,
 0.2285162410667818,0.5782016703889221,0.25111381166930424,0.385666628920345,
 0.7380835653272186,0.9778196393018995,0.3312399930700345,0.8183443509505658,
 0.06877936162955556,0.8457191982764799,0.5672120465236182,0.4087736465884536,
 0.6539597696516576,0.0638246786399359,0.5644754091304144,0.6428346641898838,
 0.40789720618387404,0.23469787991020197,0.24493151421099169,
 0.4219903548678322,0.17576880403060313,0.980564539398069,0.33716216016816636,
 0.05639006511696709,0.159191644650035,0.733515079426657,0.29133030320596365,
 0.21384769945489568,0.06625498727382995,0.7195891438471393,
 0.3366718520673019,0.1808274023275268,0.938779349923972,0.4258987431719511,
 0.97570801864959,0.803196240557573,0.3485088887676785,0.5618481608273644,
 0.013644218156499854,0.8088794610986464,0.7991530327853581,
 0.9610702210216931,0.966590052633773,0.43690093961623544,0.8297562672970555,
 0.37351069471366616,0.2984680455939801,0.785025515533848,0.15751737141393263,
 0.9003895110873963,0.5478358973815507,0.46134468704261444,0.4629711418304523,
 0.4732630100960413,0.3100492956399201,0.9065856155376619,0.6799252252099506,
 0.9741855249097057,0.2110702411994302,0.8368877535637416,0.34403411631122716,
 0.45431294321558324,0.6714824781232039,0.5268068636922728,0.2829448225281601,
 0.04649566615400924,0.3111393327480866,0.5494441125678765,0.3250128400060668,
 0.0409047230236802,0.19914481484517133,0.8058532330733563,0.6172667841907709,
 0.1770189541643108,0.8721420397099994,0.16015474290502774,0.8321896168584261,
 0.3757273306502461,0.29472381335816644,0.05259400228951183,
 0.8914951639368266,0.5044716332681335,0.09446995706066308,0.2681529722849266,
 0.6081413827908038,0.6056976516983564,0.052738858766576135,
 0.9377787331745614,0.43221708261807046,0.36646638158199674,
 0.9133254547551286,0.7984267507054355,0.4034634692157789,0.23180408706736233,
 0.3048238103692692,0.7070551735597951,0.08178500953963042,
 0.0032962136472715065,0.861142595406257,0.3335695906742855,
 0.10520182486169194,0.7934921625285923,0.4409607688293449,0.717969221411449,
 0.11601840027407173,0.6969841085653299,0.09185548106747254,
 0.2882528888299334,0.8269299321614013,0.0327347647766123,0.35534220697680785,
 0.36850676542343264,0.16848403446393467,0.9949476530955325,
 0.9478223285699879,0.6199558919297182,0.07660180850704446,
 0.46305493459767044,0.4587146618834459,0.5296224074099665,0.3615668716266969,
 0.32785547325066133,0.46973593802805036,0.9769062917547303,
 0.7357041203389011,0.45198617225456517,0.08051155300437052,
 0.9744662676690943,0.1962720193103743,0.16997608539427034,
 0.27001507529092217,0.3911758662398763,0.7761837088234119,0.8789888809670034,
 0.1593308449082944,0.5535791891328282,0.25909324517105214,
 0.20114319861883367,0.5214156664173072,0.7779355160449968,0.6924625549453813,
 0.08950798826071527,0.38686279585855354,0.08978259926039067,
 0.314029687584755,0.9417408850066757,0.4711726711639095,0.5214285139482218,
 0.8453883386604705,0.9866717365457598,0.68085281274529,0.5974327646429831,
 0.6291052214214241,0.8656370630304993,0.7299943947646381,0.5637958665385198,
 0.145003672622377,0.6350896538822047,0.791288681931743,0.06523573731244037,
 0.7694288449909944,0.5111068054544712,0.46232305093533366,0.8761519566671248,
 0.6452203771880679,0.9836376937940984,0.3370792625956335,0.3023944195953403,
 0.2583727213816376,0.4789984502486082,0.5800190465075535,0.13560218347518593,
 0.48775039460194924,0.39622492238682294,0.3827178980126549,
 0.043579748868512014,0.6072683474496596,0.39466704618692816,
 0.7433543592991432,0.36453634584589634,0.4171012322379748,0.4819430504135278,
 0.7116615227067352,0.6149553319667108,0.2593041706121173,0.6374724550717557,
 0.6880620994053533,0.006478729841424058,0.9940774495846858,
 0.2653932147463933,0.4586755724608208,0.4979900063154026,0.7846394760278841,
 0.20531779899930047,0.661235765301803,0.17527964876019975,0.2626567984764645,
 0.189630220610017,0.1649296019813331,0.31482513777207777,0.5091232901122498,
 0.6685442352034723,0.9770885163220848,0.9875889092147984,0.41544275426420985,
 0.5327915025637954,0.17991867822175645,0.32520688639776973,
 0.08714073191494476,0.06681557505988489,0.08307445268846414,
 0.19263142868226324,0.0669616861685004,0.3698753918575405,0.9425115470135321,
 0.9049651806553233,0.28379995706499417,0.3207815019235112,
 0.17914214271500073,0.3462085122893822,0.09414617774852785,
 0.5324751531390923,0.6022947845305366,0.38167506700256193,
 0.21483528522812578,0.48312452546520024,0.8418890394185262,
 0.29344403466460456,0.9252448483869066,0.931046706927443,0.1826606017721497,
 0.14159186702800208,0.33358661865494743,0.8758537750888328,
 0.1216864280570058,0.23975329122953926,0.1279651284825548,0.3666526134253607,
 0.37113115823327614,0.7528002319267487,0.09597803697695473,
 0.8326864116627865,0.18108472354688576,0.26664459141305163,
 0.2747502979328508,0.8344917638934958,0.8802537877833234,0.07533715612722003,
 0.3759165475714491,0.7992567071826239,0.21267561935803503,0.8854520179183409,
 0.7869553307178316,0.5911204161978977,0.4955060371525337,0.07874292572391028,
 0.31086319481489577,0.16853131868886329,0.9390641183762197,
 0.4162468848435388,0.2324354247988648,0.06494962565535356,0.8942861838061404,
 0.8067826678472986,0.29975698357710445,0.42432501610048157,
 0.5161174749423972,0.180339281781142,0.6933032668593361,0.915456779551447,
 0.7136538746397967,0.42309607382577163,0.9393040427689543,0.9174084571340447,
 0.9969229864113943,0.2699250982138889,0.33036114334399624,0.8114971442229701,
 0.004297919863792132,0.468828095190157,0.031140741354178436,
 0.08076960393131705,0.6366836319012292,0.7963773233291925,
 0.30164779115204177,0.3922766458240068,0.7166025931197423,0.9503247728061301,
 0.27038470949543236,0.21461189556435978,0.559245004877964,0.5170153930802309,
 0.3835440467331821,0.6840647324392628,0.5130685050034791,0.06693960508901498,
 0.05682627550124997,0.17743711375688187,0.21320160102328745,
 0.37173359374837656,0.6796653881532606,0.3148715037289622,
 0.30510858042783284,0.32206879142909683,0.7228077372220617,
 0.4177122892848003,0.9125135313609173,0.7707745109146676,0.8531589119879489,
 0.09894784168317305,0.014751220469131465,0.39079868825411435,
 0.2968144211207192,0.1569647342792022,0.8428516932925392,0.72248536745224,
 0.8533438571510865,0.19825270350104496,0.6997570415767934,0.9423707109064952,
 0.8657237338772759,0.47875383787862946,0.5051627003828052,0.9368181649892998,
 0.7324950478664695,0.5493822534000753,0.9183234800661623,0.40323766335460565,
 0.11933253071890637,0.06469592889353815,0.3928647983089959,
 0.5793379842954911,0.8707992683579272,0.40913996040569245,
 0.20514879679579068,0.19680029649681996,0.47815379504027056,
 0.5250567106973544,0.13114710177059952,0.1224941471021761,
 0.21234013078198333,0.7308795136644028,0.15246130850832662,
 0.08254252450804878,0.1709854373328713,0.20088595252988972,
 0.7705931417226418,0.5667900442267214,0.10594716934559045,0.3909347813043986,
 0.8782242832837222,0.7739812175349168,0.8388750850340665,0.7689091786000114,
 0.27073374075283685,0.08703262722438188,0.9473298112193057,
 0.15422528509582212,0.5581584688316193,0.28233067799063327,
 0.6370033509334793,0.5397159964376472,0.7243044319539218,0.03899291128644056,
 0.28593104575112616,0.49766484438144,0.12459018266432353,0.507134360257649,
 0.770345971807142,0.852677656362391,0.057142901322489115,0.66347499419944,
 0.07748938661644622,0.9513088682602561,0.6941237266637064,
 0.07806795649926945,0.7888032736591527,0.30146182494981355,
 0.1784172604924099,0.026320402843639124,0.22916224847997135,
 0.32573004163836594,0.3033344566749768,0.427866406354387,0.9390458769214269,
 0.24542109878732443,0.17172897374670992,0.21930751703568652,
 0.6691770062843674,0.7406358241752687,0.856473466027886,0.49652427417543,
 0.7331503710850393,0.9932118742971325,0.5132018449494101,0.8767506542406882,
 0.8285956080422155,0.31865905684676477,0.4082405759043771,0.5368510917924352,
 0.10264773920646941,0.04324581380615211,0.8040327655790341,
 0.8829788428730833,0.15842909599796118,0.5954173757209522,0.4865166635162895,
 0.7964049255587514,0.13631276617345367,0.4352117718880941,
 0.011645483951865598,0.02956514836896884,0.31191256238657206,
 0.17568086036847752,0.7916320918951114,0.3398183433659254,0.707250056754063,
 0.0936754939288702,0.6832915035013938,0.36336712909177515,0.2480752661654423,
 0.5525937802003537,0.6607326308377337,0.43685234616614055,
 0.30863906507626737,0.5046217717586552,0.9268910358460358,0.3386227297251583,
 0.8671564153298376,0.5272375651418258,0.20799071095868427,0.6221090097568223,
 0.8852930078383531,0.34665471180705354,0.9783158700934617,0.3644937801654269,
 0.8794705485219303,0.3622578208650965,0.18086286462612267,
 0.06824906576193746,0.9263280193610843,0.039036943645874955,
 0.1515238778047594,0.2747199956003553,0.4068779529985672,0.33757978115268317,
 0.8791677085912376,0.04529232799974059,0.7454229790832249,
 0.11783111984758399,0.8330166510620716,0.639018190620396,0.9097526230899268,
 0.9041624326071933,0.8542420618899119,0.4473222872481678,0.17696163401545195,
 0.32325797003119916,0.8476792111566153,0.3572243913088393,
 0.31842522209119106,0.7099789631543999,0.5042224551770473,0.7885981759717646,
 0.3298222806316159,0.41130028659218965,0.6584701494584277,0.9711009576415408,
 0.1265011356424952,0.8988801810328119,0.23634033707920288,
 0.04155126206476922,0.3359459289828266,0.07034078394672527,
 0.09688766136057003,0.20460602511920856,0.41629623364658885,
 0.7829442780488823,0.26876372972099505,0.31074772021713704,
 0.9828248559051793,0.6618522041421393,0.07872187544879983,0.725709246021506,
 0.07759637130877073,0.4908185844751365,0.10925066980307874,
 0.7438807211805158,0.2542372721600029,0.01268790836065914,
 0.36405942980312445,0.22722496287780258,0.5272724831979738,
 0.24475872385845798,0.7568714161655177,0.09626576085739147,
 0.08588922630590345,0.7878188990035345,0.9074817390957748,0.7640734377929489,
 0.679877047395834,0.24084389917018378,0.5282724854156168,0.7647423455476239,
 0.4507335027441103,0.4273670648219965,0.288321103380098,0.7568866776556188,
 0.8920060938123998,0.16481523055106173,0.30057011262599387,0.784430165948027,
 0.04450490697665621,0.7730281160059943,0.3428536929756387,0.4287505934121989,
 0.8953444937762257,0.4828519308106911,0.3567869412876633,0.9696528780561389,
 0.001286618533795969,0.08790866495564387,0.05720553885200941,
 0.33119612139456256,0.18327502944935992,0.8682540696246055,
 0.8602536883403208,0.7183576472957098,0.31503727892066247,
 0.058159751156191875,0.6841018675823685,0.07012401372005206,
 0.05088748980939406,0.22907758184699256,0.4448969905945651,
 0.21835248734006119,0.74417592262098,0.22803929361707587,0.28350525500678725,
 0.6079589822812205,0.11619848424226176,0.1461998083979088,
 0.41364953931358805,0.5766471088483991,0.40165967865517227,0.584202219842274,
 0.575662105061439,0.806878947770379,0.6906595806765518,0.866306040908716,
 0.5003140823587459,0.06430048026122148,0.2843527286368551,0.9207594613960961,
 0.5854200389852439,0.7035871906646546,0.5010626165123787,0.07183602156906554,
 0.9216474716998719,0.02226549415916046,0.004741394377725139,
 0.7489514977034192,0.2113682843035749,0.8580367926538041,0.09924374848906237,
 0.093299153473681,0.8732566329239017,0.8330979006140479,0.34686885236829224,
 0.14617540686684782,0.6584335189274767,0.2056902488497484,
 0.11561057962463006,0.3659240814810628,0.8597519177985125,0.9360977665357983,
 0.6077741315584384,0.7957377179574019,0.8381082063963499,0.27719944082898573,
 0.7896176192833333,0.07991968786193304,0.2981945697941325,0.2805214356628366,
 0.2497071669607246,0.9821817689324798,0.2196007604183403,0.9783970847911778,
 0.30940449569618445,0.32518239624952816,0.5996544818948505,
 0.10023950533816839,0.18050243304172242,0.16229558134409516,
 0.3077131059414807,0.8116010383356043,0.16682483840960804,0.76889723446672,
 0.13257736279398802,0.8326540476152348,0.6362653760917439,
 0.15671368482593573,0.07663884368514795,0.45688347418005504,
 0.014139373324188687,0.4998467142680092,0.7319089069445167,
 0.9662190317558964,0.33372839618841177,0.6501121483231939,0.8700064100100557,
 0.9493485611241539,0.4911702714013383,0.14628702616999023,0.9023733609109941,
 0.6185420235217446,0.08131035272526799,0.11359072399001446,
 0.4864741968843328,0.9461333566746337,0.8058214363346041,
 0.049262058467740344,0.5627033778565919,0.48725000749363734,
 0.9043446009872612,0.6998536877651429,0.6871415585768745,0.2674009090793713,
 0.42953568102817896,0.44174466471071727,0.28614845898344876,
 0.20804650217753573,0.014899976359652678,0.2381956384815972,
 0.1450572361636615,0.7211657502321768,0.45370429648419575,
 0.02852744186912126,0.5809544957604035,0.20682868221509676,
 0.18772705030349746,0.45890988797672116,0.682042286050775,
 0.36845487724203774,0.5880587373051052,0.11517103155487096,
 0.02380072838941072,0.24006570544491035,0.22854456374643395,
 0.1730562526671936,0.5255914046724863,0.5510401266650498,0.1312766231312995,
 0.2602188262592877,0.6733173893726428,0.8041914290951758,0.20950708993578937,
 0.5304360648682956,0.13155158582608273,0.2153114567513088,
 0.22757671733152318,0.6013259865061799,0.4161785964425614,0.5550588729486344,
 0.11232280771212522,0.24361709214739458,0.2477069393698551,
 0.23737886066623115,0.15255101984695885,0.17560486635434147,
 0.08248229522190187,0.23725320054411347,0.9504411987724367,
 0.9262685200988899,0.5464864198573633,0.6277781884083962,0.4166062534844007,
 0.4836814221682302,0.7357351481466796,0.12840208284631915,0.7327172818522524,
 0.6731687523197464,0.2878323475568084,0.7247723810059411,0.03068100391256856,
 0.1167568104573884,0.3001814953108066,0.5386044907911952,0.14560532809349735,
 0.75226045534707,0.7130530267814041,0.0071593291453930785,0.8627277954878659,
 0.0976944338740694,0.6779430585935304,0.8577031982573035,0.6231180667068217,
 0.11034616656440455,0.5385080589758828,0.4750423825572815,
 0.11397379407457686,0.5518560469902887,0.18694740131711007,
 0.8265578777059782,0.6891795601899662,0.6696634799324531,0.6243220027512391,
 0.12364278970252951,0.631036049116689,0.4571949885723544,0.7635044472202406,
 0.5773774183692861,0.5303442924386305,0.7086383531665404,0.29221492808741445,
 0.3615704436009082,0.06556808201633368,0.4142023803260618,0.2977447708573001,
 0.9177637423734599,0.1542040526880747,0.3030599960525493,0.8240584210573889,
 0.987350670039591,0.8391486145823034,0.31993149033683943,0.5571699319061466,
 0.7675457899641535,0.9829641482253388,0.2459138238889932,0.5077646524670204,
 0.4399371546927153,0.31586198209759364,0.6225490525043783,
 0.03520120767772228,0.14455546342931969,0.4682898439494054,
 0.5377977386431604,0.8781419908250333,0.7507431919564411,0.6448075752107871,
 0.3217926735035981,0.8536814607348784,0.6048424861795882,0.15000995329159772,
 0.7647192306222901,0.0589510499524708,0.13436605600976392,
 0.06688156160291858,0.9507359410581506,0.34774775196194074,0.470562033528798,
 0.7895754470695256,0.7473303487732977,0.6119260568736562,0.4964937566204656,
 0.6617105039141377,0.916088710328271,0.1613037631712798,0.07219548523204111,
 0.9394025685638268,0.29409648286706835,0.05851642887409103,
 0.0019491976462442384,0.6197326356768315,0.24305426194065582,
 0.49031096409145924,0.6612439410954392,0.27275896070179506,
 0.050101195643259766,0.7230215558111924,0.4496828343434498,
 0.9872206577501327,0.13765263519422932,0.49697170976889715,
 0.38222773071542393,0.9208783338465986,0.8098736568244677,
 0.22395650348917318,0.8712209998417269,0.5840348118097386,0.6853363675924561,
 0.2681265300254031,0.7130450119466039,0.2492463805562789,0.34176993436200165,
 0.9392303993694279,0.3108722257470664,0.5326511881671437,0.6090218154285223,
 0.9735330713889792,0.15589147359699873,0.6571511120292474,
 0.13408938886871935,0.9748076143079107,0.834046706591685,0.4399765007627032,
 0.7333507345597398,0.3264076180042874,0.8678122180341707,0.9459977389558708,
 0.22681170357939107,0.6760520915878581,0.7812706405292362,0.3955441417748792,
 0.432273918170728,0.28119932418415094,0.09455201621074627,0.1549059769820128,
 0.06695926432476951,0.6236811642665087,0.3482858126980799,0.5525546943844618,
 0.8345170865190559,0.6566674314958454,0.7366841298245308,0.8097664623217233,
 0.024884739216905816,0.3871291714995432,0.3276560850462946,
 0.6485773856753243,0.7834913785931071,0.5133527076248178,0.35815981622603743,
 0.7731754701536906,0.9095330345014524,0.38994463084501685,0.7708421969083146,
 0.18216808134899154,0.1827901090236398,0.40615131950546357,
 0.7508613358229589,0.36907844335149265,0.17857017713454937,
 0.5525269367087475,0.03220222377387205,0.8806301890859966,0.6174207560814228,
 0.06256447237744156,0.44759562307966394,0.09245092645857222,
 0.44218085983693245,0.5006132687437574,0.7938197299749741,0.6832791305738964,
 0.43146719111782805,0.9649196448550312,0.9778822136822736,0.8651894522758593,
 0.7502842456441717,0.2042726581826294,0.28147861147650144,0.5782247452897945,
 0.3148559020754268,0.8810692785920777,0.19910363231050654,0.520445910904475,
 0.7455051867065955,0.7597968924404466,0.9160241854675646,0.9203397905285385,
 0.8450471308714443,0.4819898629844833,0.20488759383264832,0.6487935204163621,
 0.9651384685328844,0.3507580060737666,0.11815127288291338,
 0.21967189843111878,0.5140862658901826,0.8943515379214283,
 0.37547166318110736,0.9527157237241686,0.7622655127777972,
 0.38533822641043347,0.34087752148556416,0.5605957108702366,
 0.9190402222162087,0.02877610234192174,0.2770058434117344,0.9196152471209822,
 0.7605555508497333,0.8520338938147564,0.9786021014163035,0.8385124503817721,
 0.5633243602033766,0.35155186492721624,0.7536398047991425,0.7991268051888022,
 0.44956365548971244,0.3254543181888623,0.39342726727626265,
 0.9642887513392192,0.7099011359702555,0.8921948733502729,
 0.010934304544573115,0.07373171793981514,0.4310836953970303,
 0.6529533592809484,0.5546964066161808,0.8013722537360065,0.38817041817490316,
 0.03943837238670811,0.25944113280821357,0.8649904281867079,
 0.3151980806354089,0.3056563579342707,0.7093263347553896,0.5074617817610064,
 0.9046101802447413,0.5604275904348415,0.28634384040340044,
 0.22424493807551382,0.18069197871113385,0.2809381371181784,
 0.02083938102422922,0.763329182169118,0.8764654771288243,0.39367930233711634,
 0.4719612657026959,0.7576576922854806,0.038413024404841845,
 0.5870321610332883,0.1993325223687038,0.7337659585780917,0.5458756908609262,
 0.059661570379114126,0.08204329307913938,0.13148174333881968,
 0.2965149563461512,0.6421638516730026,0.6810225309562989,0.5369723612379952,
 0.724727357469695,0.18188222697445278,0.732315357082387,0.5306239833590712,
 0.5836163266452938,0.2809683385472689,0.9399029974323008,0.8912559221284964,
 0.08865554770500261,0.9157224869091005,0.6164846293578814,0.298719858449507,
 0.7174032629468854,0.8681791745032149,0.345006721695613,0.1360144553438969,
 0.4807475379535192,0.770351684274614,0.19859017216731023,0.8392423508852115,
 0.9538019332067077,0.6539482236602433,0.8402427971666054,0.7460455009542856,
 0.9041802747119353,0.3853433889693919,0.32963153347986773,
 0.07345931727310828,0.7718424226512269,0.07438318279488776,
 0.4219424942771197,0.619700006972178,0.6559808632552215,0.927609511106518,
 0.09114380265791544,0.11813227794068892,0.5975467343767041,
 0.9852701028195628,0.02965338737418266,0.31365984713792294,
 0.27488455012371993,0.9361924328494156,0.3125965596402307,0.844672481235181,
 0.32970064333789373,0.9310127897458896,0.1802669094278343,0.9527899231506509,
 0.02462508216153947,0.0609673869895091,0.3401314251244889,
 0.14641856654088126,0.4143350271311198,0.015563798226730374,
 0.8932818135042234,0.6858824609338847,0.75418254475488,0.8195019791684603,
 0.9542902167989343,0.8203517994112488,0.8803081612254523,0.6881826724681706,
 0.5420980197848084,0.005159881400308208,0.3140142973329475,
 0.49445352953124233,0.5384983976169775,0.6037758382302366,
 0.32755081638909467,0.019185287810475904,0.4760006017361942,
 0.09861573793983736,0.4017364677933134,0.7538416866501364,0.8862569881093731,
 0.3415852037568198,0.9388232661569698,0.7195761660835328,0.6945164562508127,
 0.10602789818443048,0.5433109872606806,0.3190508531737878,0.976218153524682,
 0.5226724392752069,0.7242516694810659,0.6102184615994957,0.7433387673202775,
 0.9949575495029284,0.38185793845186655,0.6274819733289104,
 0.010940894498562614,0.18512501001712423,0.3648812049684528,
 0.0874722400719864,0.693028370409799,0.7470867749123051,0.3400723189091085,
 0.20843832460858458,0.6704638476438266,0.751089730596535,0.8430804629833106,
 0.23834246376877488,0.4090291092309355,0.08677652832833038,
 0.24010628443057191,0.22352409195025558,0.9237478363865432,
 0.40236404839004947,0.21516557611800657,0.43717045999776194,
 0.5180241967278668,0.4820566758813799,0.45545809810112614,0.9947491527712748,
 0.8066561028354712,0.2033100846718192,0.6653871733026907,0.4305386643735132,
 0.4370193021160943,0.08093975756892613,0.751792617463932,0.8759604664368202,
 0.0340301274878847,0.8007637918104528,0.8694409976558081,0.4272860770038167,
 0.8779875116181587,0.0991257509714677,0.9646202088210176,0.0507301868882184,
 0.8466989570208564,0.38688082599838536,0.2108870248421696,0.1504500882990134,
 0.35463855821269563,0.21620627044974028,0.21857325951753592,
 0.3794737283124501,0.03824177540831364,0.9120767273733902,0.3425473183001345,
 0.2142867616943782,0.21709497238648,0.8198672275142017,0.06675291707232711,
 0.028918487723428865,0.027960243144754204,0.29479060226100096,
 0.8698255498654488,0.8478661348313826,0.2354622305999351,0.24582317072929805,
 0.48194622653969743,0.5849424955294951,0.9509973553054644,0.5754331687235179,
 0.3257898360985235,0.6831454821227925,0.9437136354324389,0.4953676417083742,
 0.17040686947405348,0.4865340462589619,0.40333178912036727,
 0.6909793567164674,0.5645527656260382,0.8246568493893065,0.9205911494711309,
 0.4831981916718029,0.40039190918095346,0.4341430222236833,
 0.10141858719523744,0.5472462496479606,0.3929659277302906,
 0.33645016818194917,0.8174169310274496,0.7222713114567842,0.9334197576582796,
 0.014235270176859105,0.5799582184468547,0.2567160673132928,
 0.22492949678842977,0.993058041700183,0.06190018218028204,
 0.14931535873956558,0.7416383191186726,0.2996251292338351,0.1923998123295133,
 0.5266485608607374,0.1435133444692941,0.21356065680844583,
 0.38500768479113645,0.5828689990820589,0.6846919104729554,
 0.013175216924241395,0.3720041189711579,0.7637387682705801,
 0.7856318707925739,0.3468956956789757,0.6992356186258722,0.7265865477981321,
 0.6678834893730341,0.7092464776304702,0.5720862029640048,0.7446298943772375,
 0.07592899882005155,0.02263584206560898,0.1428930467484637,
 0.4146912927280054,0.5035068169916679,0.8035831237614136,0.3735418376921136,
 0.05698145389611897,0.27487879812135363,0.1268127343226625,
 0.5747315028504992,0.987705805547081,0.164713786254173,0.9229873209115159,
 0.503401401560793,0.26573017596054893,0.8599935746108966,0.03749049958525685,
 0.7955273328772551,0.5040812320946961,0.15771735092743078,
 0.01923916695547445,0.15944704484269567,0.8884085820202423,
 0.9786224096671192,0.45682944339518994,0.04009829990085856,
 0.41958279658155995,0.6173913885237347,0.23123106927934134,
 0.7907477992863594,0.1994493274420409,0.9566513288639976,0.6346000964553213,
 0.021008065147025357,0.9207073323874146,0.3065204565752535,
 0.4180374463179952,0.7421938027112983,0.4430832130150594,0.6963154337936445,
 0.518489580143574,0.13442612925088337,0.2036448599599504,0.03172495955812071,
 0.34280112184150213,0.693257851928554,0.5669374678590219,0.2668510064590812,
 0.16400229042921444,0.8561951110847238,0.35914604221933755,
 0.5927731834932765,0.9234625305335484,0.8643446102125022,0.9268121393096999,
 0.5189901472108264,0.8813212002799772,0.6894390426805643,0.311733383766853,
 0.19359487396792563,0.3297324460453379,0.1904425204852842,0.9813259375722513,
 0.9722687215620354,0.9025839955495238,0.10743312027374641,
 0.12774177466333225,0.07838020934673506,0.5836465255095575,
 0.6131853563254335,0.687206282077348,0.4170751512396734,0.6423821453484989,
 0.5654256089007552,0.6903619714295219,0.6445208474221102,0.11441814267534145,
 0.07261417656608327,0.5986298233730184,0.47991374410282406,
 0.7597237067003162,0.2992059525013113,0.630518336568503,0.716612230569283,
 0.5309314017172148,0.6078388582219665,0.8738131445929117,0.9434200784580277,
 0.076376648957309,0.5517803610583304,0.5439524921999171,0.27988461078353766,
 0.9547259001854806,0.02107584460084555,0.42761574243075706,
 0.5258709932813705,0.43610380064670706,0.4982044960581209,0.3445736300238451,
 0.05459470110916442,0.9867028840673782,0.3684643361478869,0.5971626355147129,
 0.9213270488335636,0.49737488714465594,0.3719289708384781,0.7966544692512784,
 0.23859502176468506,0.995703974851512,0.4205675443927561,0.7803282306085734,
 0.0850212923960183,0.9440191629586441,0.8184213432430637,0.13329411480413123,
 0.7801534023982062,0.22605719933464397,0.936392022621298,0.09313239840807663,
 0.10715318911662819,0.1936607354115485,0.13883146008125258,
 0.16218068424609422,0.4890043665941526,0.25571925031071685,
 0.6026301819605538,0.3141603780239285,0.522793958510178,0.8091896616305227,
 0.571854648364988,0.7351623865670651,0.8630244617193303,0.06620372249583117,
 0.8112791018566263,0.5122153329373202,0.8105477269733684,0.6429910506034611,
 0.7914882345802916,0.2309658444237036,0.8210667775187745,0.8630262109268846,
 0.6443952887709036,0.82476570100531,0.4472547375376881,0.4087085490439436,
 0.2840268734555287,0.8260121371607957,0.24081022367250837,0.8215631533482878,
 0.060879558485997465,0.995481940257219,0.16804986799283828,
 0.7564360905492815,0.04120142156926421,0.6117125485426187,0.464045345368747,
 0.5414585909549698,0.7256249560452038,0.8095566702432074,0.5794713657504598,
 0.6028952053027831,0.5027888835668919,0.9295749903515976,0.8288018555429,
 0.6390882242557866,0.37820841132457894,0.34696154969220516,
 0.38278700046492464,0.029468971407608402,0.48427948523874165,
 0.7447945972832417,0.21546509773854905,0.8090004681819869,0.5255241194826017,
 0.3184785271771575,0.21144880496638896,0.7515217671329117,0.0288849730897901,
 0.9615756253539781,0.3461660909115877,0.5240976148249868,0.7484130124342806,
 0.8834670540301249,0.5053847171770973,0.9751730197792216,0.2042878497464835,
 0.09321452997889201,0.09991008758557796,0.29695096889891304,
 0.2423038563156853,0.6451118210866692,0.6088992161401793,0.9622511402716023,
 0.5812704332306742,0.5417884605747364,0.31281510379574895,
 0.38091199216650373,0.6288353541562444,0.5954149168086549,0.8319240838819683,
 0.3206474574415711,0.8588558356784064,0.15613005009542602,0.3652572777328663,
 0.7381940238033765,0.9090513072795655,0.5233597091544345,0.5830716329846317,
 0.5940763950448468,0.6995084834320949,0.2708396071712501,0.625814943691006,
 0.36513697016626445,0.25969024434181576,0.8845439956674487,
 0.29042038294889283,0.34295690311942983,0.11764322289090356,
 0.4986176162489212,0.0059362444411807935,0.5466087062732472,
 0.9859697979140394,0.2271676595882678,0.09144366984152696,
 0.20689546398388292,0.10186644182732418,0.9379542650195427,
 0.9221207569661273,0.548624611531865,0.9210318367756466,0.04521183888561464,
 0.2981015779095575,0.9705112217057481,0.20268051958009436,
 0.15316369532293594,0.3507631656087805,0.8166372209534194,
 0.05294400464911497,0.9924974223021403,0.7675826467748584,0.9226138846351749,
 0.44597951155253024,0.8133303805518448,0.055843365848520454,
 0.5599659950174938,0.603861921682747,0.013005164862885499,0.7744771085243043,
 0.008842977930999862,0.8919215640910184,0.6818013469958146,
 0.8297299698241418,0.6732414572826784,0.6056901898319754,0.5088414350611625,
 0.8627851519080708,0.5603089928915579,0.9328685084727246,0.7159124447256391,
 0.0994332412755694,0.3350139417808288,0.741284765830281,0.31297762363902426,
 0.45461444185860567,0.7420617476002483,0.00909766799776568,
 0.2821072917188433,0.8331569052466502,0.5301298265277441,0.8768308094525701,
 0.4797414318876445,0.6779302244653389,0.6606002861943007,0.14854158561539443,
 0.7204814085460326,0.45797341565381533,0.0393600364508514,0.2018381532693433,
 0.2691570827535057,0.8561116960270982,0.5051328025280837,0.8580595573241591,
 0.11741664330183976,0.7981841316808296,0.8493143172917614,0.5363549508240189,
 0.2652099507299215,0.7647305687191687,0.5172843000971057,
 0.019290839376181146,0.446243486173723,0.6210970637811553,0.663208140146566,
 0.6517078241205959,0.6065435855393565,0.8126332131778841,0.07691473253715919,
 0.23215529845956395,0.25079893026141353,0.902940811255576,
 0.05415076178775724,0.14524675981337154,0.6882555401577324,0.915807958836277,
 0.4701029316713755,0.23154112263355986,0.9414712616056703,0.3785184901324171,
 0.18930932994655436,0.1401503524878529,0.2419445554590428,
 0.10335585772453915,0.9753303092477823,0.8574946640291571,
 0.24610899719660728,0.32238584146210436,0.0509401761368391,
 0.15554796439705731,0.9678820511633307,0.44344056314940483,
 0.5971029309110931,0.3669919446189467,0.2123820633299871,0.17985801623707587,
 0.17298315760257887,0.8636852867043848,0.5045733164068376,0.9983694572280895,
 0.3224432324532648,0.4093780778800069,0.003465791921876793,
 0.3076226383528492,0.06563341762243147,0.7021011588909779,0.5192954878302745,
 0.706479587657341,0.24489897062051313,0.22652747235980697,0.7119065839652752,
 0.6407318347969567,0.7094112635096617,0.9007383482681869,
 0.035170341352923074,0.43586621310376616,0.2172476084733297,
 0.42833064749085836,0.7486386142902449,0.4262505942090351,0.5243441998673103,
 0.5064500410177808,0.10048872012433718,0.859098259831005,0.13729724267130627,
 0.5882973743210389,0.1813386952585322,0.9992206030014918,0.8131811697201649,
 0.11592867556594422,0.5757370050841469,0.6972497190203684,0.7231826369051099,
 0.15650044563157706,0.7599760018464369,0.3540818246227002,0.9491567767871317,
 0.8195186079071929,0.10549656361552606,0.3160336099553638,0.6575157874504184,
 0.7733115840250764,0.196363992171684,0.21604807558525962,0.6329056683881732,
 0.9591391711775552,0.5027808535374001,0.34338124923448576,0.6201928772934648,
 0.38326445238230467,0.2891995982565453,0.7446794050598785,
 0.10796954063009334,0.11367173948354381,0.3632472593964693,
 0.1978712961103657,0.3441584245430164,0.08650246915741677,0.7022052921718186,
 0.37187113362450974,0.2727662540422656,0.1833175904679909,0.7282650077098957,
 0.07000721439796265,0.8621963418606592,0.053490467694120025,
 0.9098995228684147,0.6274441652421973,0.31428189577978316,0.6335200529326493,
 0.6085144684209346,0.3994498109934327,0.69381412208338,0.9146200305501453,
 0.11095885636046543,0.15713861461117762,0.43768776356254613,
 0.43738758704698466,0.964290247342461,0.017576301923031612,
 0.4923204513535896,0.49987062896713197,0.8996009550076053,
 0.15703309475860383,0.14235363554891145,0.2788063133203935,
 0.02866448298950386,0.4029108086120693,0.7482128304147606,0.6185144594003624,
 0.38425853491781436,0.27817176642378416,0.6029176555879815,
 0.6023049455793721,0.011423496334340144,0.7773071130098836,
 0.2111239875333849,0.007381857219254662,0.10510437911293391,
 0.916586027926361,0.4626064673773016,0.9395844938868188,0.05008194219330098,
 0.20062825690560748,0.5052033271748751,0.4577307353462172,
 0.37138120990983037,0.4119184382955735,0.05484715502950921,
 0.1199471672949074,0.5269411644481685,0.9698004305580578,0.7011694539133607,
 0.5635685027884588,0.09123908886363219,0.49341656104792275,
 0.11000900942508407,0.2744884286267256,0.3063729265871066,0.7705069386842592,
 0.5219079536735751,0.1485130938293685,0.6949886344580571,0.796527767971547,
 0.20750810521853646,0.44838952143376454,0.3179216883615955,
 0.3677896844327605,0.2609611764879012,0.6438721638138462,0.8924712151928912,
 0.624139205047134,0.8772473105623281,0.9619166241412584,0.6279981202702306,
 0.10952451726785328,0.6746202084606576,0.645925410520582,0.6121634353117021,
 0.39540418219703444,0.556030937754908,0.25835668345393104,0.3833378604982728,
 0.9381807011898785,0.660744557712267,0.9850896009262498,0.21752393537145043,
 0.36687486504452016,0.14623879403847195,0.5601133437001744,
 0.6420123782309254,0.8320570105774229,0.22871723893333162,0.8115159988331716,
 0.024611232880417888,0.9260300217675134,0.04279390341128453,
 0.16491649580451206,0.7476595386228813,0.06480981053794,0.09272856442751831,
 0.44423092634736416,0.5658643074469919,0.7002173390135381,0.4898711017123645,
 0.3237721066540554,0.9107131140092197,0.5728837182616939,0.6892016179431579,
 0.7476270008044618,0.9914832018296295,0.0049355398556422125,
 0.4894505458778491,0.4126762671175509,0.12495224504894598,0.536841011786937,
 0.575780094914805,0.1519473507757867,0.3655517389916949,0.4221263008127031,
 0.5329691968326987,0.11585464976530191,2.8253373412157057e-4,
 0.5123334817533345,0.36305087892108245,0.11254422630041538,
 0.1114367579887996,0.5689200467536248,0.040450194308122134,
 0.2755022872057278,0.16857779963600716,0.14703545406370067,
 0.04047155391431434,0.3654667657619787,0.390155697627079,0.7653758791299725,
 0.9499455254969442,0.967867181755258,0.43440825513780834,
 0.028609819247878088,0.061148065463619616,0.5579440614051824,
 0.09821571681289243,0.3564743387794044,0.4717671483153153,0.9042861728422179,
 0.3696049792987426,0.34852202388417286,0.0119753086777582,0.3210740680203751,
 0.47706916571603253,0.035349580078243736,0.6727278377917336,
 0.8695869731291465,0.425700955441231,0.4777771347830019,0.19361323806698083,
 0.3083400106041233,0.7448131304057464,0.44882034563096107,
 0.06699130551206656,0.4650739626441672,0.7842741151952705,
 0.16432501240076913,0.10666427989757565,0.11927389189949933,
 0.6674927943605586,0.14589405974602987,0.12986347022150102,
 0.4813217161906279,0.32695937644386597,0.38935573855617467,0.670879546077474,
 0.8329639249026731,0.8168858693306038,0.20976386593006546,
 0.45107078810652157,0.057133111199481945,0.4999034083995302,
 0.3908286371665315,0.22290638983574107,0.6835006409821262,0.2988400791194632,
 0.37383786244620476,0.11863290466184429,0.5305248471917514,
 0.7731559822947183,0.4099869096467331,0.1922905602104399,0.5153874831811314,
 0.7692902944663742,0.26978889967562614,0.9017758760919304,
 0.48566852745213873,0.24554610743947236,0.20470495534292255,
 0.5562309256777773,0.6117449829049996,0.5918817319785736,0.44530804895842113,
 0.603631224179112,0.028220641232583832,0.4662232879825997,
 0.17642386735485338,0.23424592327156124,0.914757083081522,0.9857203915257141,
 0.7641853480662475,0.6711467033060655,0.22826774328183574,
 0.027359775939953312,0.4741288181150929,0.8729902777010164,
 0.04381089296105678,0.5112954356556916,0.6265741170637649,0.6502961261906182,
 0.623470687516156,0.45510863776253196,0.5308339800342754,0.6165138569721367,
 0.30541466461909994,0.8292384139321008,0.01335832569264983,
 0.2748050599442463,0.9301295441152804,0.15552335524168803,
 0.02272666719342742,0.274256926070102,0.49899860654856776,
 0.05176060024057261,0.2609181515636989,0.19439263396257678,
 0.2896792322939469,0.11679920010429401,0.19241395802133088,
 0.5779244226696015,0.44686502687050766,0.20253328622628108,
 0.7238250316546285,0.335436485671594,0.5885268301118816,0.2787834755749954,
 0.7614858540106865,0.1885066098170488,0.2939756965691541,0.1085950191748083,
 0.9774793369453421,0.39351966407494754,0.009288045193596273,
 0.5345613771798088,0.48440065158418566,0.49181917735659864,
 0.5515393412515238,0.12696916676466807,0.39212756925212033,
 0.5137128547462989,0.9614116249898335,0.644338022366058,0.7405464130781828,
 0.6003047994942539,0.1747982363590852,0.533726205066801,0.6695771506571196,
 0.5659016936859614,0.7368965849307502,0.3430284445766105,0.8997075094642646,
 0.8919081222366901,0.5758685719449129,0.663890487974135,0.9136303272030836,
 0.8082082312613236,0.18641206909743047,0.2500626684557672,0.9363286306072123,
 0.10432309407969076,0.23202341858818554,0.9236419478997439,
 0.5964485222616183,0.45847163193055096,0.1353888891920798,0.8189791055949183,
 0.06074949222473047,0.5687445715129616,0.36592119696381553,
 0.4728416079445197,0.36621490776269683,0.8364681780416211,0.6462672957014918,
 0.6618397583691893,0.9399620918777339,0.7192041386175906,0.94747287179835,
 0.46769432842314196,0.4479024206601916,0.8036872525571844,0.5183858084358014,
 0.5966034856818034,0.19729293936410353,0.17050490471444957,
 0.13600837890916828,0.7968844562060471,0.4692758458185041,
 0.30640347015666136,0.7297370842964085,0.10740051070374323,
 0.5353126285653907,0.8460743279561557,0.32409590588770265,0.687500511431454,
 0.45698542020252253,0.7135455768140184,0.05876945540306111,
 0.036761202247709956,0.8498581948658092,0.9068880345455657,
 0.6258138626656118,0.43727964747846426,0.4541423617026328,
 0.11777370789757646,0.635316389009003,0.939050387222417,0.3585743354356177,
 0.37949254806952537,0.14398983655473485,0.49609535743147426,
 0.39439683170131046,0.7996860713351079,0.8220975415406276,
 0.49147215379806786,0.795090625869179,0.08786061261268374,0.5751315713125551,
 0.7202173544096161,0.19013697638129035,0.287989353099119,0.45051419410025684,
 0.6277498800498934,0.4439238763654543,0.9534101915178209,0.42480408635530165,
 0.00482205382151879,0.9790737339077411,0.3410869752082485,
 0.33580987490842773,0.1782426266524646,0.7722225767380462,0.6861472873135596,
 0.33594699185632637,0.3415583914730318,0.2506159466462071,0.7298009557960871,
 0.6955426364737606,0.5568423867271807,0.4239377831694746,0.21056775408667505,
 0.5176035867563412,0.8319453077222896,0.6717457531732862,0.04285403941798083,
 0.2233486494854856,0.9116572439707498,0.7829151651151438,0.8310166877599522,
 0.65382510440161,0.2356378520667053,0.7221023332991064,0.9458612324644674,
 0.39577018303415157,0.4915475281598649,0.6602301668244356,
 0.03515215704658747,0.7980302000657681,0.8430867675737383,0.8803616107297754,
 0.35960951415389986,0.09048427410591486,0.760706768473127,0.2822824997070714,
 0.798113673861002,0.6267581240833958,0.8727380824058558,0.8829624824062156,
 0.8277315408838567,0.9660193514602284,0.4582463976772201,0.3628529731879846,
 0.725989687457235,0.8681779357843242,0.5540007750693322,0.2503021092326124,
 0.554151073535958,0.5523248942443756,0.619010990901774,0.27783123047025116,
 0.48168090122532714,0.6165488578527748,0.5548158124933323,0.4987595747248387,
 0.8916325721526548,0.5373475323280468,0.9854408288917301,0.7860403484394385,
 0.7375770951269546,0.44035080038875785,0.15548973107938413,0.441158346086736,
 0.5939748937209606,0.9994821085896376,0.8707888251185496,0.6296334204580035,
 0.3998307320611242,0.37192262011842636,0.36126708758250015,
 0.9204910850831756,0.935946371441357,0.04482936678465099,0.774477717209426,
 0.8797857868317749,0.4175791158726513,0.9497839824295544,0.05019589027524962,
 0.7093102065379927,0.14428982944545077,0.30495046701347794,
 0.1299750069205784,0.4091217545910095,0.6443018921098767,0.06963014175408255,
 0.3729788847621027,0.3380151935052864,0.20871940899683905,0.7819272310152013,
 0.007615334255745787,0.5459730010324808,0.4789911906278992,
 0.8623562446743408,0.3144045009462074,0.5631727909306937,0.32850907281291386,
 0.2699785269842466,0.4669498745305679,0.129275775526546,0.9025378109266581,
 0.6916205198285341,0.13564803309653561,0.8171934325769712,0.1983492416206054,
 0.5116413245428577,0.12963668179163723,0.6586209488552325,0.4126969959230694,
 0.5176760483719302,0.04394185799366879,0.46062636563393333,
 0.4701274601271114,0.5024011154767924,0.9343060625063248,0.3171083280197663,
 0.27901817987375144,0.24829177991481466,0.22942118800134326,
 0.09394923917235476,0.1199804244876218,0.8011493427183121,0.8144651010369252,
 0.9881492904462305,0.14895612977432782,0.6410970587873452,0.6193243495860176,
 0.04157294718571369,0.2651789510549789,0.5034527114314926,0.1434978408476073,
 0.0542282639280649,0.21047116630845375,0.71743219377953,0.8070273025778443,
 0.4214771676569731,0.49799961576237417,0.07387141729318991,
 0.4573946997443702,0.7620005649507796,0.4638072900774436,0.7898182409131487,
 0.5448941867227616,0.1880620248235605,0.9345867775468369,0.1853181769088491,
 0.08230936754453022,0.9674052909537008,0.6368548082735948,0.6460335184109763,
 0.12761795764859352,0.4824765350276481,0.19644599958640097,
 0.2091373607522271,0.33321938138337737,0.5103245554550792,
 0.20400064378974492,0.21444319151715585,0.35010092912024393,
 0.3013540169601082,0.9773298228793688,0.5836762381903373,0.22253942554996342,
 0.8051374896578276,0.7992054474933079,0.07381967691545765,0.9096920759220455,
 0.8662483977451458,0.48934442410452383,0.5002825809173747,0.1937947536193082,
 0.5836577704530281,0.7064673275205293,0.5264451303375288,0.7210453338633754,
 0.9036992412732774,0.19341190908793404,0.9392357745315549,
 0.29039098948955167,0.8978518567922404,0.877505642342986,0.4608817108556069,
 0.5930816220619533,0.5601432696113474,0.47089428725686533,0.8686681572640013,
 0.8257585072216825,0.5045768375102434,0.7569479497052705,0.3162085625330562,
 0.11414912195895455,0.41818940988863307,0.23973563464308956,
 0.5886571734009457,0.26025801189989395,0.3395253056485158,0.5885444510076999,
 0.7985420141766222,0.16280691141763382,0.9208366522448022,0.3864425707572944,
 0.14046616997609052,0.9381150013416553,0.6994696519457582,
 0.27345223568189914,0.8908878494125275,0.41371952746195495,
 0.23571949720797625,0.2865057292112234,0.004066736419244599,
 0.02881079450315427,0.16938760822225174,0.19478141988346254,
 0.4761474675862992,0.8041096654754862,0.7539230495846799,0.5866244131368992,
 0.627998295240773,0.31770982056955455,0.549006319705387,0.7916688903900362,
 0.3633437567986386,0.24069485480019648,0.014513084478620653,
 0.4771392986077717,0.8559691733832044,0.9857797073264224,0.7090749109705778,
 0.2493297043062641,0.2072532389057553,0.8659096170187154,
 7.027348743190442e-4,0.8086480350266437,0.0794867561162711,
 0.10161788256126258,0.3713884349117702,0.5598286163658737,0.8526677269325303,
 0.6283751874018928,0.8655411264801519,0.34410965341685296,
 0.14708700134496755,0.7454030327075997,0.04201871226029419,
 0.024321946202553346,0.23409304413451604,0.2913603921166208,
 0.8857981011437935,0.4348902812236193,0.5721377937266001,0.9656030571394331,
 0.48665512598461613,0.2969901476813249,0.534679794285764,0.11074713400430247,
 0.6156603166560227,0.6305318736264002,0.19735587315328984,0.3022947348695628,
 0.6751356839481717,0.8226936812506329,0.539481079049896,0.142880109183121,
 0.23406639744194602,0.09000326166646,0.8710279900387222,0.8462892004548102,
 0.09510460398405396,0.38077843689118196,0.9195985879736779,
 0.5018679543420397,0.30731905769937085,0.5274142702822296,
 0.019904869321678786,0.11186461392905678,0.11731704095517537,
 0.9640860114810841,0.6582840825670991,0.09965606398011873,0.5936657308792176,
 0.5573561411543456,0.3254021994922498,0.17083796139993346,
 0.39526825506101015,0.21765944939324822,0.7938964306726155,
 0.4054897026108575,0.921288419601699,0.41539334342467626,0.6535330097226837,
 0.6273322364685981,0.7037410700862148,0.8879990148759361,0.08200970333785884,
 0.9382209823720128,0.1195494344971626,0.4403908057212029,0.08919601357742768,
 0.8914960027491533,0.19401604598699218,0.6399987163666144,0.8550332844538416,
 0.33910654357201075,0.5874222500929669,0.2592747996540813,0.9570396674076199,
 0.5951169101062143,0.8036323190709526,0.5586294957043205,0.4581091750212223,
 0.1411691869236218,0.8394855017306541,0.31346224786839516,
 0.17977751939455322,0.190373504007749,0.34112887536635483,0.684270352872927,
 0.6757189069900554,0.4960597537809375,0.535898644280992,0.09015604637475949,
 0.042250190875468774,0.8963834924443372,0.57816461170457,0.41441300269411474,
 0.5422346709299602,0.0795856351372296,0.3280833264060202,0.04943844788907337,
 0.5084747059106073,0.7910852207132422,0.011667913039559163,
 0.8001605586472261,0.46510734322651537,0.6467910677596601,0.4469786177076618,
 0.8166380299699966,0.676508587801838,0.6432060610272814,0.07486571724770397,
 0.6851007527934331,0.7573457326578404,0.9733415699535071,0.0373911136645092,
 0.6135291358336037,0.5690515619965503,0.019912992998632806,
 0.3932755844875806,0.5243758419708409,0.3920313283222129,0.8112918960256589,
 0.16785159398733807,0.4228303696048018,0.134092688989765,0.17538552868298607,
 0.6669305504865166,0.4801554082541428,0.7792245300437469,0.9082034328981002,
 0.4712825621445479,0.03073288481519576,0.04533394110472799,
 0.009999072726233349,0.9949212619365779,0.4152995299313361,
 0.4277116543596714,0.2422646362609242,0.1366153275442168,0.39058611703596824,
 0.4625656131886525,0.7572940700682368,0.6370974904298808,0.2128563931163453,
 0.2337842730532652,0.26916353146502237,0.03463819219996511,
 0.8654177370373055,0.5527423019554436,0.6542658845200162,0.8313510454909518,
 0.609344394468563,0.1409956693432669,0.6302882174125854,0.9694961569254965,
 0.4389317934147511,0.6428135158080567,0.6328723262235324,0.9587476960606195,
 0.07512868051982413,0.8352013627970954,0.06391547045032986,
 0.22033656517198175,0.18388237392540296,0.2041581453254646,
 0.29048175419090994,0.15579810548953665,0.7116314003091934,
 0.8435834232520328,0.2065450166068259,0.45873365082944717,0.8393633734622343,
 0.9890123046529821,0.7677380943060491,0.3657117436025069,
 0.011407454868550437,0.3362422282111003,0.034523442383030956,
 0.8570378832224167,0.22138672020070083,0.6061836722478442,0.854116470981766,
 0.8146273942961554,0.6066751188934247,0.7177830381239816,0.96751888174932,
 0.5295572602894598,0.8679924052565473,0.7311895359779351,0.8137968491849714,
 0.3506846656707031,0.7214127133737871,0.9169049057769596,0.5588861774452113,
 0.20468472637559576,0.12967898148864876,0.6126458301660161,
 0.9544740941381273,0.828203468036403,0.8187045961927241,0.15670924372528594,
 0.7555883083471405,0.8082124991524996,0.22312027900643505,
 0.32236097371534345,0.422737783036089,0.1662029094667472,0.24581405911940268,
 0.7351994122854784,0.10902632265785817,0.8452037430762452,0.3596648535740621,
 0.8154386111149824,0.6855994272974625,0.24665392484880522,
 0.30244065528088804,0.20448330599022835,0.5929983264357708,
 0.17056800218341905,0.9674357633228718,0.01264648945310487,
 0.15308734201142793,0.9198029231600962,0.7407259084982608,0.4775399698309195,
 0.41701280442176847,0.9068838939092849,0.8808359661112374,0.6787046077517157,
 0.5359818578687029,0.2857874957779213,0.8591715948766906,0.8390677815117237,
 0.5790649150435752,0.8998165699912113,0.4673508569021705,0.28636050385056855,
 0.8651615225072129,0.34530164222856685,0.43455797930912543,
 0.014951564415351726,0.5844363429930087,0.43853885381164315,
 0.3736729394284023,0.6351551301476479,0.37394244567503343,0.4119163151583217,
 0.11630331365715652,0.48126178994098756,0.021300858922885246,
 0.5147161366434139,0.13266033043675574,0.6692853615096522,0.5322277251933054,
 0.9993183562327367,0.6850727291987914,0.9152537028421294,0.2826910958887474,
 0.7145749582765399,0.9370188946033767,0.221874087631455,0.09787279580954078,
 0.06758411113374274,0.3057041490328779,0.25360683531006556,
 0.6735860961865817,0.6713419211423676,0.8259573737612571,0.4197173437376134,
 0.6329861006694106,0.38005030478490087,0.9422285901539268,
 0.029570723004070376,0.09255120876566747,0.21704159434219905,
 0.1040040718082842,0.691587872332769,0.9429626667249722,0.5684987365683675,
 0.9280476078031981,0.6684270699032226,0.5135574996484219,0.2558506078293944,
 0.4773314759248738,0.9123070182671013,0.5145257235667159,0.738540278181226,
 0.8614777993193787,0.9389647864129972,0.5485930380543041,0.13069697130990177,
 0.8660553265392796,0.8109331871560636,0.6584320399180135,0.6609061223354787,
 0.007773650702837642,0.04622118390533503,0.34388668112494813,
 0.4040183304729167,0.6732945900547865,0.604964825251662,0.058023520870436274,
 0.30029156152822356,0.5770454288668746,0.6517867087830498,0.8523693443444571,
 0.41190460682513375,0.0945334918073687,0.837321787039665,0.8579914896922394,
 0.5834632617070712,0.2773929079485571,0.9783181353622714,0.4083519057019268,
 0.02317815679673685,0.4597970338664328,0.14043399884515306,
 0.4867553346922463,0.5201053244377716,0.5966987960497283,0.41937829630608625,
 0.5249135651927326,0.04939849510770222,0.656215229505402,0.8662886250318541,
 0.24750162982074841,0.7475427132664274,0.10173046909589067,
 0.001255716512262861,0.3909695545405384,0.9952081095893313,0.421802520487196,
 0.8703602007702957,0.4137719902584923,0.6901690341032185,0.7532713738424387,
 0.8040025773085148,0.2432877111175511,0.9494353588344909,0.5316997217986283,
 0.028727947755986194,0.44658137282134913,0.2126501548104036,
 0.38590978786908137,0.5648220258802911,0.3052088355603667,
 0.40015648505594414,0.2700917707067306,0.7565600712682922,0.4160504689266359,
 0.9900039601182512,0.25001689894834056,0.27221112657790425,
 0.05918977130077363,0.1378461592389706,0.9838923950609471,0.9085930499502769,
 0.6885229576339225,0.04663152181630337,0.3299186807471911,0.1290328437795658,
 0.3290449294931409,0.6848810224223317,0.1668680334961795,0.4862972274876489,
 0.7759797201702353,0.6302542959990565,0.7254184383213277,0.4255332320960352,
 0.4937913123488067,0.37143298192574836,0.10404270758670031,
 0.8413494059685461,0.28904666167981974,0.438933130518389,0.7371076651477151,
 0.32732597317497536,0.01827352152253614,0.8079313140899871,
 0.6690365947579419,0.5970869976265232,0.5609755914131719,0.786036303802478,
 0.08579831717705111,0.6973195132730277,0.08861965788837578,
 0.6077317599613417,0.4281052748058759,0.8199279902065407,0.5127787912365797,
 0.479305973084742,0.27500208105658464,0.5742404647697086,0.7195133344477728,
 0.0547674000339079,0.13656744126977582,0.8861831769338805,0.3139793101427889,
 0.24297934731336057,0.34996324854556315,0.6109814629928135,
 0.37928205110945756,0.9984331596521628,0.42233388600755006,0.519918990926193,
 0.8035660966914618,0.3632030465481736,0.5616968644353405,
 0.020784485132892883,0.7041561161278518,0.8511917376605734,
 0.09415088215239165,0.32910469624020544,0.0638294511055788,
 0.34385172670531394,0.818305306007073,0.2999426187341869,0.23649455498448235,
 0.8526643777994737,0.1814888734598551,0.6982427983865359,0.12852094250626678,
 0.8959691209509869,0.5294477075578288,0.017640779822758912,
 0.005550462819710944,0.8345651747308077,0.1067551847464201,
 0.4090295509520502,0.5897701792507135,0.026304456244747643,
 0.9653022072092662,0.7446849894135895,0.6744523342088236,0.47093816932186705,
 0.8994288366944057,0.2603749606000396,0.42112456550905975,0.6303251101130964,
 0.5707687096114261,0.6319104731899984,0.8214924037254175,0.17408642047350442,
 0.18944907374170283,0.14109853071002498,0.29521967806372007,
 0.15813538621167522,0.6860325942367154,0.9324780913118752,0.8276320845973921,
 0.5022793267530316,0.6810819106829159,0.7129961389487895,0.9194263248550545,
 0.14393388175874933,0.6259227201973079,0.4007825943935236,0.9732479987407626,
 0.30122525413502843,0.06039841205971386,0.12872803359499052,
 0.9495684909201345,0.8934557980114326,0.8262748651468208,0.609691615083335,
 0.42193405565915243,0.5601148334001264,0.6489318693283639,0.6548818492629223,
 0.7251295669702882,0.15466041558739785,0.16003993117127113,
 0.12545681065841463,0.5441673236255464,0.3135853740049044,0.5504036316712694,
 0.19218003757445867,0.8342648767058922,0.5304380030191167,0.530133805342731,
 0.26106309751154966,0.8165654745308761,0.8294036645513891,
 0.45027004717891994,0.8386172478083034,0.3443165876237988,
 0.05830569744997505,0.20008266570027933,0.5491017763530249,
 0.7937704557075645,0.9697113788431637,0.5178752062851575,0.8609202171787993,
 0.6104452289694532,0.814520930888383,0.10192399349131165,0.5298734854396151,
 0.9484343916415114,0.2597345157497323,0.7350516974317398,0.5646633767346275,
 0.01357199433403411,0.3613913324972078,0.029057761120438452,
 0.9254372839931542,0.7583163084478044,0.24489326583616,0.7030267247659355,
 0.885064001149636,0.4781201147248825,0.171539497965859,0.2650421823402055,
 0.8947271150689151,0.3731892622303573,0.7675072526379364,0.8238685748128587,
 0.05844960349974704,0.40907040511130455,0.11174340003434924,
 0.9266770415124802,0.916194205715671,0.6023731923994214,0.0971432903571442,
 0.12831378716412545,0.4788091130070906,0.7867554902166373,0.8538262137921842,
 0.4200541875429632,0.7684694997335554,0.6205028521400974,0.7059043392205981,
 0.9889414637837417,0.5043905889436726,0.3493691726793321,0.32423665001870206,
 0.9991166556023373,0.08181913494466753,0.3757892568281006,
 0.23098841184179564,0.16923747664598576,0.6242971594221226,0.93129934978566,
 0.717188829698608,0.30514706419250415,0.41916781539860093,
 0.012605074114454218,0.012906315852570072,0.43140860635583034,
 0.904975227130277,0.8029920714340096,0.8205055382312486,0.5914486282046794,
 0.6751645131530501,0.07966670139305965,0.9516680820746043,0.5900508062118592,
 0.5639765725690946,0.8307413809460352,0.8841415518586063,0.4094545033279049,
 0.5157329969676638,0.6481405025456841,0.6959646411813226,0.40764460849327167,
 0.9429572623584843,0.7003625442561103,0.7870859834687585,0.876423190835181,
 0.6634009096157768,0.9010748174321027,0.17255917785625152,0.6778012717338109,
 0.5181451239403357,0.6651156955151559,0.8018051519703778,0.5549507124377058,
 0.2450844830813632,0.6894568490208505,0.3137009398268018,0.8629694125895839,
 0.7940883159649169,0.17528275277147987,0.6798044039773057,
 0.32089625147808853,0.09288553078679995,0.3471541396583214,
 0.8060988815176487,0.8870650724723751,0.6429481119263469,0.27957872993545874,
 0.4565705353641494,0.7021386261585087,0.21615399334083518,
 0.47707178907462044,0.8000398112718025,0.247392036896397,0.29156651632584063,
 0.15362799156742524,0.633126446325734,0.8577720580273278,0.2152942931399645,
 0.8500412892662905,0.6516516639786104,0.33320059170176375,0.5781825667681684,
 0.4400525239405201,0.5284982371754396,0.464938786842903,0.26559464040970915,
 0.7243908205854808,0.5755776197220531,0.9675547991902211,0.7538156320311995,
 0.4955677531296294,0.6128845071570279,0.7926805076307393,
 0.035883146129949806,0.29632390964055677,0.7011077398224117,
 0.36032428251117843,0.5952103665406551,0.9280627879517517,
 0.26534819837690027,0.9351237855017094,0.8364481675675537,0.7432157866776112,
 0.8542484184082169,0.909998707550262,0.568294139021808,0.9488697542763777,
 0.7556938937666204,0.3623985606759724,0.4139180755112353,0.7609798143950041,
 0.10501041505581976,0.40239547829428224,0.24403776543790412,
 0.0428904114465295,0.7557010507307095,0.2869270381436282,0.723624063701891,
 0.7265066804383634,0.5188670700824447,0.9529296416228209,0.4662608838400859,
 0.23006570390305936,0.7891151252352441,0.7862305720615219,
 0.03301258380935379,0.8915257949069755,0.5532382312311642,0.664057292247105,
 0.4274555771706381,0.25394093884878866,0.13520400607535032,
 0.6053193514850304,0.11337859006771134,0.505491603814416,0.9908900557136733,
 0.8479914604729504,0.5957053706110041,0.6169358752745976,0.8819406386775024,
 0.7687625936965,0.5743998212916445,0.7404743844075481,0.035416357602797044,
 0.7489738841472625,0.6747766374693911,0.684960641710522,0.5243913155789376,
 0.5053462290208052,0.7895885172699695,0.8068695745683887,0.2039163946847542,
 0.9727120298442253,0.2872561780963546,0.3421987101457563,0.645590658365353,
 0.4607689916065715,0.39874787245268384,0.09126029239863531,
 0.7779888239815389,0.28860125495784583,0.7903695571347175,0.7151972792488229,
 0.5559664406932301,0.32209572242328555,0.3050513656939552,0.7183910835571995,
 0.9817191699055102,0.14355255215027318,0.6190649489055894,0.8433885232101085,
 0.293508729114472,0.04760805949932623,0.8707477166516245,0.05332103051072545,
 0.08500749919090722,0.6401945696879299,0.8935283694392611,0.1367313653112847,
 0.06796315060208213,0.6534990019043817,0.4428745338052368,
 0.10627670916444565,0.8181390945519645,0.04660185686386609,
 0.1539536774996706,0.3964197940127394,0.6168906933043692,0.9013064103457727,
 0.9077712557028292,0.5423761822160944,0.04589985170530819,
 0.027180547272255273,0.346805386167919,0.5594134223521529,
 0.46689247422880575,0.5909239890965611,0.9050232892056392,0.3067469394777208,
 0.8742250061589113,0.4236598058127631,0.6151874868478121,
 0.018624800241702544,0.05663949954611991,0.888947887112983,
 0.03339902625683888,0.5549589141524925,0.07064056388758044,
 0.37604221682272365,0.30638414978634854,0.4852474801131562,
 0.6843875074905951,0.7494160924463628,0.697082506409762,0.47646553187541163,
 0.31612082530319174,0.8377871208607033,0.3166264241187613,0.7206112154289891,
 0.1406646513140133,0.9359855262917492,0.4639712656534227,0.38507924300618657,
 0.8818645755835962,0.299352792213495,0.6652632460968062,0.47055297640200355,
 0.4440030634800114,0.9583207680065486,0.20539056775113818,0.4544023774496091,
 0.2654626916140229,0.995684682285889,0.18621492612433888,0.5846007094306911,
 0.17795504810601714,0.6788759760997429,0.020464199032029784,
 0.7118438595515251,0.9850527378750029,0.7310031163992143,0.16723253876548372,
 0.7673947765385253,0.5740507343710205,0.5975398678375238,0.5412717583822169,
 0.6699229890477452,0.1155096650459344,0.2321580471291833,0.6278274966537736,
 0.7976439007388335,0.3653055026389129,0.78103398797888,0.031211068194465508,
 0.40366166101663914,0.20480570298126866,0.6913214841582007,
 0.4765096453169053,0.4370580949613202,0.007973287946347662,
 0.6358126273251667,0.0015099617278382382,0.2530190702677042,
 0.7772854287490346,0.8869601513786585,0.32122784617264866,0.5538166514692686,
 0.7254046852544126,0.9045498286351245,0.5775861550680593,0.857200955306585,
 0.01387277334664494,0.27277683267849073,0.1582378929019792,
 0.6963408591239564,0.15324470571909576,0.7654408929427763,
 0.04897599792078644,0.38191419735531107,0.7770845270567794,
 0.1833042239800673,0.5292011368225595,0.5080606002807118,0.46796362826932814,
 0.9491622516026089,0.035577774883073565,0.3717908567388364,
 0.5252209623616383,0.958995196298466,0.4140196142163941,0.17842197983455743,
 0.18181550012317393,0.7820371482625681,0.5308934763997392,0.6031953638033948,
 0.9704563942737854,0.08440358565105921,0.02688961676110757,
 0.8293150045713418,0.8045820420835752,0.14706324261074166,0.9062248103063081,
 0.6598448430659305,0.5532457013523823,0.33452280843514626,0.5791228620705527,
 0.9252490451002219,0.9285277912536427,0.6027141061928235,0.3090723518992229,
 0.5565004505749694,0.6084237689865837,0.561353762572096,0.3637118359294642,
 0.4301308311734655,0.45379519839201454,0.640693285067829,0.46079352537022145,
 0.3901554205871456,0.014757350029025895,0.1019295678174319,
 0.5912001208960518,0.8976072135550051,0.22700415561505238,0.4409140714604969,
 0.013962709406570628,0.374786847610867,0.7359623916153331,0.8999314503017687,
 0.5128805142029229,0.8042641899345143,0.20331652561866054,0.629712169461902,
 0.3539108453773745,0.5160388860539665,0.48947602630166,0.916755657381747,
 0.06149204332482916,0.8483481891371532,0.7141536386176393,
 0.09248541496948559,0.9423882622171191,0.9783094846697231,
 0.34336107212914135,0.2915701554939767,0.5499348096824129,
 0.005997889230583198,0.8405088618336976,0.2505964550855141,
 0.8513083865823312,0.20394796511449598,0.7063093163480428,
 0.10078088052493217,0.7630561825585584,0.7440398738034318,0.9735200845698994,
 0.5771130586578948,0.6318969795382979,0.6912004101076932,0.37427170923824304,
 0.8860976106138215,0.013590705045335993,0.32917796549215594,
 0.9539675060717486,0.46845024462439055,0.028847530014458056,
 0.9701599319021831,0.9834991835770855,0.097008421348317,0.2551312198066864,
 0.6305833295553912,0.4676805469034747,0.7971808780193467,0.5284298621436264,
 0.8090664212106244,0.18386735698166057,0.7335292972603491,0.6689256222349323,
 0.3956674602565021,0.8931001043268317,0.792701506193286,0.2277062226274239,
 0.6289159885417273,0.6146746655802222,0.16253313094929633,
 0.47480429007236546,0.4239950462998614,0.5805199234636633,0.6775978684269226,
 0.7470100214073544,0.42199665944113607,0.7008201099674273,0.8567280711136203,
 0.5520187803485491,0.3110983926151374,0.7042412781972174,0.5067287634451929,
 0.36846635409916606,0.45554695776927634,0.7886034962791738,
 0.016365279648668296,0.29907045353733896,0.4573943337206935,
 0.25407231643347805,0.8315582247157287,0.6714292011752507,0.3869735250704106,
 0.47509906093638277,0.5800483325580897,0.1443630968952747,
 0.08338799923373164,0.3204757220636407,0.10152695115773336,
 0.7843643385231442,0.08741221472270344,0.6258668889488771,0.3144140306414309,
 0.036743213237948824,0.5000078479126984,0.2307122576517312,
 0.33796508435986305,0.5272990239890653,0.8826497784957482,
 0.14235925119738502,0.43889894470479485,0.695149693729747,0.3011389324393078,
 0.1709438119309573,0.29104712723071935,0.4904752005737235,
 0.09791037540174974,0.8670261538538998,0.46386035077746235,
 0.8989106000210174,0.6993547108262833,0.20117662246713264,0.6518920867394586,
 0.6332779971214397,0.9085303340346818,0.0983572765411369,0.5843421524849448,
 0.298994713868344,0.29290087448142554,0.02470294694156816,
 0.24742543082073987,0.0944200226017251,0.5944082151664037,0.7260832160772839,
 0.9470267185719683,0.005525975205378786,0.40080284961697255,
 0.9420708578466191,0.1891919483651494,0.7623588076581277,0.18069136821468323,
 0.9111065806664838,0.8552025107972459,0.0511092103399986,0.9007462175179058,
 0.7854355188080999,0.12794427800083152,0.11982723566415754,
 0.13182177261180728,0.6081347662395239,0.5888807136238017,0.8419903595473105,
 0.17205111053422606,0.48238133708259745,0.8335175341301545,0.499869796732869,
 0.6462587117382614,0.39624671563283975,0.06495295187327477,
 0.4029055775334358,0.6209835105768226,0.5416165740132795,0.7705826550882195,
 0.8683180936711938,0.6308626541000784,0.6931994385648536,0.8706793216882743,
 0.7856008186206097,0.3050952878470685,0.6625693291563901,0.05064538506311622,
 0.12821514356540875,0.1943486810377698,0.08039783467094241,
 0.05236544491652084,0.5171950815026877,0.05443512960139285,
 0.28661073379525814,0.9258281637623131,0.4086758019295703,0.1315287015534664,
 0.660563684480391,0.8124807174760613,0.30996575893307143,0.4321256296655035,
 0.4503528069137468,0.08019676084401339,0.12582397804966183,
 0.17767523865005275,0.19990109136987544,0.07485519711347732,
 0.4889710916039791,0.82724529976342,0.34936645684148215,0.5079668488659214,
 0.9314951127338943,0.050238562109342966,0.9330307235615445,
 0.7143606430777064,0.3268498527333076,0.64211683437946,0.9004217857379767,
 0.15943276782403826,0.07841371860894408,0.8710617126776616,
 0.8571356236245719,0.595969621194067,0.9959997564238623,0.44994458083096334,
 0.5935569165344332,0.7662748954726547,0.17271192042706973,
 0.42874811036033944,0.5013889931505647,0.5900175145792785,0.8892214731765938,
 0.40416336873946823,0.7578716498199211,0.8397276026428325,0.4197231514506392,
 0.033031590734100336,0.645990835546111,0.7442998265104607,0.2292597831723968,
 0.1970632561888116,0.7420014286496357,0.3159719425187183,0.6266842801398531,
 0.8920454885176201,0.08120813308137775,0.38608460424498403,
 0.9876665461051277,0.4509436415305583,0.4634181487012997,0.04176219606052067,
 0.27595367579641683,0.25042534316959864,0.8103551792278115,
 0.36304950233048383,0.9935927768841613,0.801345586840288,0.461773928759186,
 0.06908993664912055,0.14245450552277172,0.4851328326553237,
 0.20277577872386499,0.0745266372139215,0.48346415221587213,0.526706481260665,
 0.8926772658293556,0.7136536449668172,0.2363822075874129,0.8977388018254165,
 0.43576745467204214,0.008677018078930288,0.5282218981761396,
 0.595828376377402,0.9052824747224354,0.2675457100151595,0.17753100895398932,
 0.6209395187027957,0.14869095904463348,0.3185108239485658,0.9031314070096543,
 0.15869922557427563,0.5488313645772814,0.1393503134026206,
 0.41636014264419596,0.21639775093541114,0.6888000933063033,
 0.2755410472019887,0.42144669430275516,0.010866230292947532,
 0.4665542169500019,0.10340682248983546,0.830060382464683,0.4901704769640217,
 0.6566012929628222,0.6158245196508804,0.7414537207781506,0.15397381532248067,
 0.8005497000106908,0.21679910179460982,0.568380104858433,0.6048690525509092,
 0.31098284387328246,0.9012769098038736,0.6386682049948287,0.8828864613135846,
 0.7645000978512227,0.8518258608971643,0.3434891738099535,0.10108533579679357,
 0.4124613145510795,0.035948620715423596,0.39583044078173113,
 0.6067875478525006,0.11100701934572266,0.9031853510128033,0.6537077374235163,
 0.8030154432359123,0.3243940845758354,0.11564950277358288,0.3769015749630331,
 0.6190425190342062,0.7423689954695606,0.9507906629191734,0.9136426484867584,
 0.915390472660419,0.8052281734966685,0.9732430543572276,0.11568994164367141,
 0.8489674819510726,0.4771236629286062,0.7872109234010495,0.2834796485728843,
 0.18284353193558212,0.9272083678053252,0.5212230426779325,0.0997216936328511,
 0.41790885156669644,0.5593773800871393,0.5845048867329474,
 0.20606349580378502,0.5543979971544342,0.49657987616397914,
 0.3931027918681069,0.14918186639610298,0.788865890098654,0.9227031198309267,
 0.5176651179730017,0.9087372401632091,0.5609462981172045,0.1770967928057563,
 0.3773648425254763,0.4947335692230429,0.5339529828703848,0.7170044997912047,
 0.6212455182036085,0.45851904179105063,0.5577016275778194,0.9366276445829275,
 0.26590733110351106,0.7922588263413365,0.3371513206590493,0.2688205998694627,
 0.46598521880988186,0.007606978794739927,0.22129823228718792,
 0.5359524832587564,0.5669332143282327,0.11787512780940768,
 0.27566690133380134,0.20206784436738523,0.7261884285314115,
 0.04016143917050563,0.6323756185161582,0.7788485832601459,0.7896393972974439,
 0.480616372939495,0.06989176019160559,0.5045054672207866,0.3187481220812911,
 0.36971762885263726,0.49897568650595514,0.8620452668577603,
 0.5338698112749516,0.8066255389941385,0.47513266311979085,
 0.20402638211092694,0.6223452150397846,0.654199307535116,0.5695922479811021,
 0.861125098826154,0.2604298499000478,0.9694187647738538,0.7466405202114391,
 0.7842295959975452,0.9398812480858043,0.25220568841359703,0.9882567410537172,
 0.21122972825692887,0.01267722856482778,0.653520601401642,0.6297501379589472,
 0.6073013255771216,0.01420290606290786,0.24934268934937986,
 0.5959948983458129,0.8480589071811075,0.251669227940889,0.006361595615567928,
 0.014662868682507435,0.18726557542838762,0.46541780814074685,
 0.9555767559683016,0.8453738558170367,0.7720347062272357,0.2927080817692148,
 0.9991924263037879,0.3103953902375475,0.637484258066077,0.877095707142393,
 0.05135363071960097,0.7262091695659748,0.6465969496698882,
 0.022668232410586575,0.870107838204194,0.8540852951104716,0.6106097497856561,
 0.006327166935064299,0.7095088081676152,0.7305823192254225,
 0.6029385370239717,0.9612989228992322,0.5265051286628578,0.8429015366658136,
 0.37624886198414886,0.582042978173507,0.9660620268750155,0.3056869143741442,
 0.9278995396794149,0.3806000491401632,0.9035740746986228,0.4369156584517483,
 0.7273392185449027,0.1266549373725625,0.6194420548634931,0.790592356897078,
 0.7930496772483284,0.6107019030684,0.40305893904046586,0.8316185347917926,
 0.8503363023311866,0.05700680956163062,0.8228410662659068,0.5746404789398967,
 0.8760388672337761,0.817664152208396,0.8302054216753887,0.3341443047964967,
 0.7803334252255764,0.8023039723788912,0.7092845513921052,0.322311496850356,
 0.18144785292742027,0.0454235120778308,0.5034938550051287,0.3882604978177474,
 0.1977978860682711,0.6337453063652233,0.13392943658748613,
 0.28516129623409303,0.6118383068857316,0.8773123055682686,0.9752682994959317,
 0.4264240573849515,0.597728675817263,0.6418468438648334,0.7677013860236538,
 0.2914485573676999,0.8935576761462978,0.07766507445018922,
 0.27387812900389175,0.30859008781617936,0.5572290152477476,
 0.09020477948339478,0.7097395884810995,0.4798281705708546,0.8611881482598108,
 0.8564350097315975,0.20558236882489,0.3984431428624925,0.052016138486133734,
 0.8383411875041338,0.29442436938937555,0.44418398246503255,
 0.30123351423603195,0.1760658330191185,0.520149445900768,0.8940594331709235,
 0.054357569375245496,0.7658729681406221,0.08058641245332632,
 0.33237466020159445,0.20066130362866685,0.17261451209061296,
 0.8198976395136288,0.057923013468147655,0.845047096834055,0.4370859850997013,
 0.8346811499645758,0.5994171956121587,0.34502565920886563,0.7749519805572584,
 0.525139652056149,0.4468285121206632,0.5321382977166211,0.1426644168228992,
 0.25368181088213615,0.763350803435291,0.1996319360051888,0.00261670039603934,
 0.5706721820847124,0.6173705887492061,0.5276577742846746,0.7607613359680547,
 0.9280028948919803,0.5230771006752123,0.15398970527387879,0.5727474718862947,
 0.2730063382033959,0.15082021440808435,0.17428325087605934,
 0.8109878683792757,0.12883116238958114,0.02359263903472475,
 0.9188365603677613,0.6755561137736611,0.40748019275584835,
 0.34667676715679374,0.42410838006920837,0.8518644426841869,
 0.0897695796968121,0.16588879566532955,0.5614303071966713,0.7128776539573166,
 0.4255361775807569,0.41284298645460793,0.9960553123553479,0.6922805308999058,
 0.3591102237367776,0.13616135164103038,0.21437183116627567,
 0.5680485498013109,0.8961153824726694,0.5528454201705569,0.2754280996717887,
 0.40092529756094675,0.5795772169414479,0.9003130997104866,
 0.06493922077682057,0.5353443160591682,0.7038773845408859,
 0.24472789331107747,0.2933532598331572,0.32664499455754115,
 0.5890907343053374,0.03808140908238311,0.7741478792910286,
 0.09271928507870308,0.3680803585777678,0.5592312872879892,
 0.36758711818296264,0.23264023445972148,0.6899774557093448,
 0.3787506463059115,0.6469269494840276,0.8009691820534977,0.18843586595473916,
 0.24674221834268606,0.7898669602639147,0.0893207826911121,
 0.21852233296780255,0.34240714027190533,0.13487215726759005,
 0.986351445318626,0.7567811357734437,0.7520198527441593,0.8307162668133523,
 0.9261267813245332,0.21296617834977039,0.5119157882777529,
 0.41585060181799205,0.8206779241695885,0.9625839028132954,0.8578000174595445,
 0.1162466099383529,0.1252680392043093,0.10566033949609699,0.4851395918294109,
 0.5413027354543327,0.9121206616219345,0.16402789933302464,0.9639157190663336,
 0.5028610273358306,0.6710694894515757,0.14751069307886144,
 0.45727802564712816,0.04845154853492539,0.03821285517672046,
 0.7983213474049261,0.3622622805801774,0.74198637485311,0.8459067554653241,
 0.2031022424628649,0.7331418412050994,0.15763908434643115,0.6446756969041338,
 0.6009443989064392,0.03879095834179247,0.47097811850038385,
 0.17513948721491956,0.3014710267392472,0.7952120498263822,0.9713527157571381,
 0.9831014430197743,0.8500772130463413,0.756586946241418,0.07680534595956956,
 0.04399632656589003,0.5576785372260046,0.8854433927292407,0.3086213738464192,
 0.7058237445930181,0.786938844103615,0.01729661602071375,0.9259085232511746,
 0.2208580200580792,0.07649300333928077,0.08013030111712416,
 0.8916790808234648,0.8082218165321936,0.25926595079862313,
 0.19688159660643823,0.5257696660818854,0.6686366519765693,
 0.35737867707843796,0.10670054198889023,0.37946724423863043,
 0.8422232087093506,0.56700477915056,0.07533811086323505,0.5876593971320163,
 0.28464030217546066,0.07200469286708866,0.8764893682305628,
 0.12153834654396789,0.10091959472931278,0.7353367799558845,
 0.3990341843358448,0.8213255198605779,0.07841497084886317,0.9499603214827166,
 0.4240423701150744,0.728935547421303,0.12349010403827865,0.2827508428085278,
 0.1537435675559613,0.10485507900760282,0.7021218525142005,0.2381750000175612,
 0.6109388973802905,0.8358774915760206,0.9992143179665712,0.8947759485978544,
 0.7144865052545156,0.5296040204966173,0.6714327277284177,0.932992087927751,
 0.3337426916519738,0.4637060185285211,0.7610837693546267,0.16974085746912027,
 0.9185716751671267,0.5171571536087775,0.3565411930210508,
 0.023677717875226545,0.8689728907952887,0.20636455212664773,
 0.024599338995235365,0.26750130973446795,0.3310658958691115,
 0.2172111137532311,0.21306878070056312,0.03651330020423338,
 0.48928949039086866,0.9388888843830165,0.40265844774640414,
 0.2642926414709158,0.6293202553967983,0.436290038468939,0.036524722804683996,
 0.48045981790333336,0.5994932283047334,0.7984238980839077,0.6034068557838401,
 0.20008858528216855,0.7528391320658357,0.8296810235237234,0.2539282707135494,
 0.37711443680117973,0.7132297169217446,0.9643774824694078,0.6961405233136794,
 0.8294871095588067,0.13536030468307958,0.31347596903735,0.8076575409528619,
 0.31235303578989215,0.23933965977476712,0.15182710658880794,
 0.810233137237432,0.5978522415578884,0.9703274733796746,0.49173305831749636,
 0.626390858535978,0.5900237859566179,0.43085088528113025,0.26437869667975133,
 0.8484265011687547,0.6639092270184254,0.2515719246722432,0.29724231484537655,
 0.942714192328814,0.07980652621725526,0.8528644845523532,0.29522399146881606,
 0.8718542217068183,0.12134426486326089,0.6966904076265508,0.7668528987665866,
 0.48103800600911806,0.8807674852772673,0.7664492180697456,0.749965978474144,
 0.4923717350137575,0.5714428500769249,0.09949609312470387,
 0.11653885799213748,0.8042869619023716,0.7008137056186616,0.8377052109117759,
 0.187597939456578,0.13937238407312158,0.18199901551768638,0.3362745866390632,
 0.8136444480297584,0.4845476660690269,0.2218499533572681,0.450305775022637,
 0.28208992632181573,0.25556514082890947,0.9943194456000881,
 0.25410520100583944,0.09859509183753834,0.6552925289550539,
 0.9305962790422154,0.22796112935347268,0.7018461702848998,
 0.09067402989671125,0.22312570196656,0.8084813093464192,0.9442559551617837,
 0.4956386343175555,0.9551466065829545,0.12243594529654656,0.5709375287039271,
 0.47932112914391045,0.014576416626341793,0.4356932314890899,
 0.12806717796704647,0.4836723246363367,0.39197384662439494,
 0.12747141687268804,0.8962182278463914,0.7624569773643364,
 0.40193046889117734,0.18081292209637112,0.5787170924649839,
 0.6310618068037026,0.7777909767073986,0.37477734686117115,0.8899708851996411,
 0.6002637078140879,0.19980398929209497,0.7249948318501502,0.686407524681337,
 0.7286695156480137,0.008747621671428507,0.4186798524137888,
 0.2543511376058706,0.14867229636859014,0.051340798558785594,
 0.9597198125554779,0.8471955375323525,0.5218620273981481,0.1663030687987147,
 0.33395949056059226,0.2633048462735219,0.4240257456069082,0.6053347064631568,
 0.27697334011379215,0.16507364559179538,0.5048130005565847,
 0.9990652898554965,0.7500831974388321,0.07280297451670137,0.5543613336220614,
 0.313473722691781,0.7067338948980911,0.40495719352792503,0.6044844920395597,
 0.29982164746566475,0.3730626468539242,0.679056259938655,0.7535340131172485,
 0.16238307071414138,0.02657778886335116,0.11595848159203603,
 0.4854637761195567,0.6163482489576345,0.018175694688981947,
 0.09532590518217166,0.937569644259604,0.10785322684774945,0.5335601515509644,
 0.04654189054315361,0.9597472686590024,0.06827885403400014,
 0.6015553891318367,0.34576678892177837,0.3456966578292173,0.5150866727314334,
 0.6462299770329296,0.768318604946006,0.9977408152386293,0.47100002415387854,
 0.1863489445304526,0.36075085776693305,0.34757731166252137,
 0.6844968600023098,0.7506992047086859,0.5577987714532612,0.7701511425780393,
 0.9347468067699642,0.04866095919764246,0.20276679809214382,
 0.17233600074306754,0.8640061405170023,0.4701482583476213,0.5893287472819763,
 0.05639213802437937,0.4346330955598703,0.6276834438240341,0.8335258376851924,
 0.4539653458760161,0.7488680176370354,0.8963527504379792,0.05315811836085249,
 0.6440386208071515,0.3862836303492809,0.25914045290046794,0.8667613350619069,
 0.5881428100274114,0.014182359319688254,0.6379230748170786,
 0.5373059181134524,0.9550037906180839,0.4234908115460515,0.9273677281374778,
 0.34705887764296883,0.9474743126196425,0.23432398887710415,
 0.40313985037489597,0.6800274426612303,0.3538373032132678,0.8368642483308995,
 0.7362186531686434,0.6441704964778282,0.7352846907063157,0.4171257964337184,
 0.3018777542371808,0.15561505898858363,0.8005666866120628,0.7345414673393201,
 0.005738051839264902,0.6121445735805501,0.3305441510039382,
 0.38262056957638224,0.7641014306705552,0.4534975545513955,0.5594333253735645,
 0.6122043592683234,0.12950458350172012,0.16587559242720662,
 0.7404916842936232,0.033492934933229135,0.7212052998171077,
 0.9213445505941165,0.9128388905371364,0.4599943202418594,0.9101873892886191,
 0.6079915061080693,0.625210281121305,0.62455356690801,0.37743249702682913,
 0.9183937235732853,0.5852873223143041,0.7883310766332233,0.6136638041345683,
 0.7189305535277146,0.7761019850932337,0.3650374155927214,0.7791761265362467,
 0.7872043893714734,0.7669749866264829,0.6801453162118534,0.44657960805990404,
 0.07791145697702473,0.01548625448589136,0.6763506317615255,
 0.39737341476179544,0.12314450211085937,0.8420004538642103,
 0.5150666261690964,0.363776109573408,0.7491032674947914,0.38557436168484993,
 0.7366627466878971,0.7006736849950094,0.8946509620541864,0.9908082971650902,
 0.33538878646480585,0.46585411534814136,0.3977966108547888,
 0.37479584348779826,0.7343854173586868,0.018888426434234606,
 0.260480950013536,0.879523224172116,0.9217068903050776,0.022175268053362118,
 0.7740925050810684,0.5674159521273142,0.7771675675801479,0.5728661657995603,
 0.3695853375463506,0.19159268262893048,0.39470339020454326,
 0.9339382001247205,0.9532710177165061,0.029703190953209835,
 0.7280310360104307,0.3734120833275457,0.6258509405167225,0.34590260763793035,
 0.044387174719751066,0.7471528034783876,0.867296731951522,0.2105153029412773,
 0.28397595885814275,0.2801500868904442,0.8602782269816586,
 0.38482955100470484,0.040835852025513986,0.8955223568546796,
 0.11439506707959601,0.2215548568743685,0.8300191319133323,0.7280518590942741,
 0.40684071001241706,0.09958354693757954,0.5317091327868048,
 0.1603071149790205,0.12913460442324542,0.7643248759045825,0.7771816945520782,
 0.8795663412345559,0.15192597551912035,0.4145936314262626,0.6529698772271979,
 0.64101667655903,0.09673383597641849,0.27664101560693166,0.7519527433622745,
 0.9860241122735182,0.05332973740184088,0.8415608977506932,0.2654616840325458,
 0.33586991718243664,0.8281648939607404,0.8974664445742437,
 0.40115945411976706,0.5062456532275988,0.1823950772829509,0.9848150507391664,
 0.39815161276034483,0.6592637919911533,0.1734572895153783,0.7157510806992888,
 0.20241259158831437,0.6630245525318392,0.6252863338208589,0.7274872529598074,
 0.27205816080644984,0.48727324381586645,0.19801261088401179,
 0.09333465535010976,0.6402593505218053,0.7905622928340947,0.3662616489724694,
 0.05979795334463711,0.631727945326249,0.24241549522145434,0.7726168673696542,
 0.710446004715672,0.4226689705619615,0.08257858537947549,0.7402055001067478,
 0.7228861524827108,0.4421124439078896,0.8846313932117332,0.9685243907968452,
 0.6136074114098776,0.10598908434172405,9.402236376745421e-4,
 0.03725556023606891,0.5936063023985372,0.9532573295161959,
 0.19415088309945028,0.8190897712854224,0.05304810728976905,
 0.6444922132751569,0.6846112177226658,0.27901203694877363,
 0.17719891023070178,0.6789767670710719,0.9812750495714494,
 0.30092265895229964,0.3143192660791898,0.5769771730009046,
 0.22605510542584184,0.9975394976044587,0.2650161658401209,0.9767430141606969,
 0.289539103393756,0.7325270802990209,0.3504328768708638,0.8066169949246784,
 0.0661047944615637,0.21436157793455068,0.2852658243149061,0.3417950497711495,
 0.5618455017638441,0.6399553156148605,0.09212401799775416,0.6733137744372609,
 0.4813656489722187,0.9302248209910114,0.21148346592355372,0.9904348406424979,
 0.11110838444592774,0.154143857526713,0.27189501073284816,
 0.33289015401763744,0.7156281821307943,0.5283362954939084,0.9142993261744337,
 0.3977429830048438,0.8419704682717646,0.7138116144342251,0.25042360173932054,
 0.1731735536871447,0.43670562792868894,0.5599809327968738,0.6687984130420248,
 0.3187649498450851,0.9116456479258833,0.44413605559585134,0.6894694938101438,
 0.12822482432623838,0.30134407614508296,0.9533200275602687,
 0.3964719009516816,0.5145151676893959,0.6107626965550796,0.6139593914215239,
 0.5562469619501472,0.7807374268005005,0.8331188529392659,0.39544381116312677,
 0.874035355649289,0.4077871917382173,0.8635089498894175,0.3607835650270659,
 0.6827392354701264,0.591023008881033,0.49271441958995854,0.11648728031150679,
 0.2605902640385884,0.4073550165696044,0.7581861724130801,0.382181438557854,
 0.08697536076107837,0.9555733577281617,0.878136587546695,0.7906176149810348,
 0.31983500633859896,0.7549832909539489,0.9593354434839741,0.1550872101280848,
 0.2871126588951014,0.3342605726272061,0.2523199840332624,0.7002928254183154,
 0.1631938653564925,0.6419536220492471,0.8832988218814084,0.6295680327227401,
 0.8007015712825527,0.6384150693650259,0.49064959815987785,
 0.22972706235641505,0.3754175192141558,0.7634012624824195,0.609310209614219,
 0.7526503276110367,0.7385180460364881,0.5449064683745979,0.5911587084416176,
 0.0247607158154326,0.460199958767028,0.3183733696072303,0.8575252146128184,
 0.6862896727710677,0.002653065154485379,0.572764365889202,0.6835239776628598,
 0.7448390457595706,0.4215296819887271,0.35632274384650575,
 0.13304290516953332,0.2371277670451566,0.5234499630495008,0.8758802090436414,
 0.06071108221460353,0.4267674003544807,0.9254395641290157,0.641782086680429,
 0.968409710719928,0.7340478680328006,0.27537796593947494,0.8871863785916639,
 0.8437927481400141,0.044236438186529115,0.21031705934414857,
 0.3090602969184294,0.8346754281309253,0.8859371492647978,0.2370626074536265,
 0.20406584467485944,0.9680267402434523,0.2603428282653468,0.4405868900055232,
 0.8714880966301177,0.6630012759964581,0.9444962863525963,0.5632369968200555,
 0.9030738930003329,0.5949508545132085,0.15929960380710972,0.8680534900745612,
 0.4120323260644372,0.49428714465462786,0.1170158871199275,
 0.42369378403615676,0.8931186554078253,0.7595315712063495,0.7231354832107353,
 0.4922639731022562,0.5872108340410478,0.1724005015891551,0.678464582441319,
 0.06700081083657106,0.7562364874421517,0.9192379785717917,0.1803196410562713,
 0.6638805682935045,0.2925908542992893,0.5094616840881449,0.8562096143981821,
 0.8397823013039915,0.7549978470633005,0.004090732473562508,
 0.48687382959781467,0.2376044371757965,0.9122322051854911,
 0.27752235674647674,0.21756957495226814,0.3134645835243115,
 0.22611417656217903,0.14435831268745403,0.23892629564023593,
 0.9856396366959809,0.11111605107537648,0.5418238226724353,0.2571333985337084,
 0.5944515243476427,0.62709991244838,0.0844893202852317,0.9298019724495251,
 0.736192269103499,0.8303842963515377,0.884398728938146,0.07687380069102778,
 0.08533842723812612,0.8995477697833393,0.9624508807510802,
 0.14483200519285655,0.9484521651281206,0.3929687602128352,
 0.49751004779899355,0.736575158662519,0.37879967143896687,
 0.05935564507317215,0.497367936875039,0.3772261460793207,0.3051151131948878,
 0.48508870983197505,0.818107633017326,0.5996766170990163,0.5641498050969754,
 0.973137009443931,0.9002046923674529,0.05378853869484512,0.5117065047119316,
 0.4387157421213319,0.4473453129109426,0.6400269067123385,0.3472500186758377,
 0.2720060636305923,0.7879211672266725,0.07979188428160133,0.820410963625532,
 0.25500512141634557,0.7668185635430914,0.4579770053644808,
 0.06564265485282961,0.7729460971001016,0.6344101632640499,0.4264261469866558,
 0.7615132329055923,0.29347788931153973,0.8776041894548784,0.5076361804645804,
 0.05430469391525461,0.07940729453743167,0.6711126146212896,
 0.6267783020982783,0.4469417770778914,0.9964732754912058,0.5255739788191727,
 0.32709773338313186,0.17500541787876012,0.13101230559583976,
 0.5939585291739413,0.06817949475242813,0.04268245157130146,
 0.16552205502236028,0.6235232342559294,0.8908455426798608,0.5274235714557955,
 0.9586399462553923,0.8185840646143796,0.1767647927190772,0.07729317663328783,
 0.1664138716566974,0.04520495257364521,0.5718918517667009,0.355604593474507,
 0.7723386286573878,0.6363119997142963,0.9853503066137013,0.805015485450274,
 0.17755125769372027,0.7386550563619674,0.6364499932273991,0.9207583833152003,
 0.42662915760387754,0.345455684715925,0.07624937738830717,
 0.19249806238631317,0.6365192279067142,0.8066261963508766,0.6152832025074784,
 0.9840909149183767,0.3357086598590767,0.5364166937170696,0.8159769209728098,
 0.01838376678672715,0.4646620457792142,0.1330037131722125,
 0.059309568313456906,0.07063261524349418,0.910659955614394,0.911290955489971,
 0.9004353179235657,0.36543299042834554,0.2216046980582992,
 0.06365162797701329,0.8964374492015553,0.7692295481725689,
 0.07366700118645164,0.7718820145887944,0.6056190703343619,0.8147720382136486,
 0.2583862699906042,0.2552834513915194,0.21803674949131335,0.7527241393337087,
 0.9109990284283254,0.5285055061776184,0.45387856023625617,0.5858526027872616,
 0.7560709878173545,0.5143507374537614,0.18605706587943271,0.6799052393986939,
 0.2762312771204164,0.4901363954893775,0.549896218495726,0.8838413114975315,
 0.6897728659906609,0.9146215731073644,0.8557069546422029,0.45302768610755784,
 0.7574791707692857,0.4563172496504453,0.05042303090406297,
 0.11588966835505088,0.7507934021760971,0.8122556610553349,0.3973645571584957,
 0.315572196404155,0.8938653571120997,0.9766134490476697,0.891152701878247,
 0.9204362800933544,0.5139454687057792,0.7959874194869936,0.02643062247748007,
 0.5012986423139456,0.7403497406715434,0.965850491191261,0.47800282630662294,
 0.36558789377185585,0.5088143207537504,0.14603235441837814,
 0.21756351076450975,0.7192783077093715,0.4857999751367481,0.2200669196566607,
 0.6535429347028741,0.09775143584326051,0.7554001805716909,0.8534036138587564,
 0.12138589306460057,0.08626857744428951,0.5926543948797004,0.762117525330328,
 0.8400405750262796,0.6826532488009082,0.6626570794736079,0.8570059530665468,
 0.5957105153408543,0.2039424092235299,0.08933722459011018,0.8087452928280386,
 0.7555947703105739,0.049056472429964915,0.19616563083387906,
 0.014821740492809932,0.746073724115839,0.6651797674909754,
 0.018668519143237194,0.9266211844349224,0.30730939358256726,
 0.7121748714317402,0.6044422001577832,0.2565293233187893,0.899050997561711,
 0.801738094051056,0.25686483125832693,0.8918042311520219,0.27771638520220676,
 0.08467362432108338,0.15683670557331997,0.07417045938293798,
 0.5315783527980004,0.6962733363448184,0.521647056118895,0.39553859169968597,
 0.20285703545808065,0.8914722865977317,0.8689127029478537,0.7770940595845036,
 0.5711749179768337,0.8910503473590803,0.8709238995735269,0.31909650020707825,
 0.0271729694669397,0.33774891826980735,0.3260372371732865,0.9592472157035765,
 0.2309956036801657,0.023949210421997935,0.5010430260027965,
 0.6454325418626528,0.2789085844900988,0.2610708014061932,0.5963266122592863,
 0.17514534952672411,0.9116435064117698,0.9663720619114615,0.8774002095258637,
 0.8375674668212465,0.15986219473307006,0.8568189810931642,0.7682510375926765,
 0.7890694782267189,0.5562415337705817,0.12030516436549388,0.2976249324257729,
 0.5278958339541758,0.6297511398665772,0.3456986461060928,0.1518599122695049,
 0.8712614362842886,0.9485783599373657,0.8834974150990462,0.3350834544693204,
 0.9787469733964431,0.394006525305281,0.8445295119121391,0.9487290119586528,
 0.2344413442320632,0.6063667548335394,0.7954852084382228,0.6092396117349173,
 0.8113449812689698,0.759434366997083,0.8435388360073289,0.34882630668845915,
 0.2512190430264012,0.19861328124126565,0.12998919638589856,
 0.4006058132105783,0.6644210050218573,0.2774258891320748,0.4688503697541666,
 0.5263173170601567,0.930147195192896,0.2546635025033097,0.6035121936971114,
 0.889947464631682,0.07493074915253362,0.28692380384722205,0.7680651750413562,
 0.9847887034899445,0.5360588527838117,0.21409683441979777,0.5431865148885898,
 0.12235035298512575,0.7762402103685153,0.7003523634474406,0.243339572705483,
 0.05947658050075211,0.2307795440838043,0.8834278191931715,
 0.16125523318057777,0.7174203765549969,0.68275148584746,0.9779978240670204,
 0.5475704178807732,0.15250136240191348,0.43342095295633976,
 0.5957141070401146,0.03722904875905808,0.4332315452331337,
 0.46159035722119823,0.9267548429855964,0.9189574596617707,
 0.03252553457195795,0.7275223609224791,0.23768657421341777,
 0.02191233113655755,0.3965227703133106,0.9961694116938584,0.2892871252432041,
 0.26751537226739286,0.28884087846577966,0.5997254686830109,
 0.18528162066214238,0.37454138787498303,0.1523313444436929,
 0.2770047740009225,0.6972984878615871,0.9594062618924107,0.6163828669912343,
 0.3875697397741371,0.22898171554839308,0.010519416150224448,
 0.37721077276410375,0.27644126284811554,0.921867252470612,0.9433647650828632,
 0.5779500338403432,0.3963137561927118,0.4891434932396783,0.04137705750926779,
 0.7618795035854541,0.39951965090388997,0.9244222129906964,0.8775898804489818,
 0.18529751954196927,0.4730809889085408,0.7590369148401331,
 0.31092129125948853,0.7342129609779527,0.3897830139078783,0.8843038832480143,
 0.7252930466893004,0.4833864157365667,0.7167808279213113,0.9723350915150358,
 0.9681547501940589,0.23663300536295628,0.49266841197449684,
 0.4408141185018549,0.25893299811959136,0.2322661825287703,0.6909433564397106,
 0.8196122720191905,0.8485671906994495,0.975387787786535,0.055498843141265075,
 0.2592790947503464,0.963428273494068,0.8832749507809134,0.5715857947812791,
 0.2904541701075155,0.9355418745217718,0.31085228610169235,0.4594447719344813,
 0.43073751065398413,0.028361699799160034,0.13286048299548125,
 0.705063578725877,0.7169883398983772,0.1874198690397676,0.2568543969356929,
 0.7832163727529785,0.8147772136026931,0.05253101628587298,
 0.43278960067612193,0.7230775983996488,0.11625736341179704,0.420904680914995,
 0.3706157812829788,0.12604562146852605,0.6118505393213839,
 0.03517668769125504,0.011667689644669799,0.8644833951154215,
 0.36293032461261454,0.5733783861625381,0.3506249500284414,0.6790060542528208,
 0.7472624917789294,0.7994279811949867,0.3601324865140686,0.8661877192427653,
 0.3460870458895716,0.03711924986974624,0.029137846355974606,
 0.47146606110961187,0.15214063884483575,0.8162000610115677,
 0.3631117403902324,0.6806714988434354,0.9504422962902056,
 0.025279288721049786,0.6676965933302474,0.3072177116475654,
 0.9057369952803163,0.022983327852235447,0.2922732963236844,
 0.3939849814574663,0.4641178050431718,0.9990035897396373,0.7707240016465031,
 0.5097207764448287,0.18305703158500464,0.40364401311091125,
 0.8790359381445628,0.03168766939542811,0.44423117560093395,
 0.7044479371670823,0.7763640447121196,0.5629484094867808,0.8825016587160864,
 0.6678451980743721,0.24333972328294162,0.9013531770071413,
 0.008150899487940344,0.38850536527227986,0.19392676316496849,
 0.18709417025880926,0.5199450834309857,0.9803618520828198,
 0.20411576182987123,0.9728886072446369,0.09713589723598104,
 0.1652355163137993,0.9977085583652056,0.6612266405591494,0.25052621793420893,
 0.16127642967118827,0.8898507419695172,0.022455395426150493,
 0.22886558108718558,0.0561387545479719,0.8489782169753883,
 0.13299583571802143,0.23409436116803795,0.6645713526759031,0.788907930941025,
 0.3936184295324341,0.540304679556945,0.012545219189726953,0.7471895337814329,
 0.931462546899508,0.9022955121158504,0.1937773411195869,0.5429949224237511,
 0.6327770625895087,0.27046038146635243,0.7470097001617345,0.2973537211925511,
 0.3639278242176298,0.02480595979709299,0.3085323535120206,0.8925482786034065,
 0.26388922735694853,0.5542606455465395,0.8174217397555824,0.7978223714354068,
 0.2619474522528883,0.8870288932149524,0.9223577534880819,0.8464660442001632,
 0.29707968050684896,0.04284767195339467,0.6403516050840402,
 0.21445278107578025,0.4906380437393051,0.757117867853134,0.11463808833882372,
 0.8544789638965746,0.5412055039227076,0.09861023491849563,
 0.41293025214253476,0.9105080335451862,0.22352888999942766,
 0.7854058403787814,0.984746306399401,0.623866755650227,0.704910434984355,
 0.7600008234930944,0.07580429659348342,0.018131687766707816,
 0.6699555555688066,0.6181374705054469,0.6671964206763342,0.3753140998918102,
 0.2990086563948938,0.3745948775253901,0.9088295022966204,0.75698902773754,
 0.17616679792062429,0.041045777178519316,0.4104512542616615,
 0.2952555756819746,0.5451940780917292,0.10950396041405186,
 0.03695674307337349,0.5930855881115616,0.51528059452726,0.3886517400501004,
 0.6535844048201533,0.3509127561910719,0.6863509665889711,0.3624120388168015,
 0.3661222571933678,0.8167681210155341,0.8466997646056028,0.7161956417705273,
 0.21103348958453705,0.5223109808475981,0.10632421953448024,
 0.15364299295429884,0.6150241663046815,0.7174941740303739,
 0.09877830153412126,0.6004961517717523,0.5216564005934083,0.2537937491378409,
 0.963214835490203,0.8796722154609293,0.41216826857573174,0.3895685144841081,
 0.7348762743798856,0.590481399343832,0.774424994886857,0.8447464253530089,
 0.8904462832990921,0.4329921211990547,0.9281218025399522,0.9438465903434208,
 0.8692638120569447,0.11196979131643692,0.5940940784141473,
 0.0033410173173274993,0.3295412470478557,0.5691156879722459,
 0.002056328725816625,0.35247211393129296,0.5710361902846944,
 0.018257853194330154,0.8931861348680937,0.2355557780245614,
 0.49936543494574914,0.1893315669886978,0.9372088916719381,0.6283303657900081,
 0.6136690991590694,0.8204332357551603,0.6186085712914087,0.4111939476351866,
 0.7652298674238166,0.6202543952931717,0.1467660019739011,0.4281477623032587,
 0.21811378247651614,0.3954789323330644,0.393183412966964,0.7773247795467092,
 0.2899915955346152,0.7184430579164592,0.3599250312560066,0.08691781430267964,
 0.3366599191948385,0.7212654228611809,0.4472900069587882,0.21330299501599836,
 0.8790275658609642,0.24148745339626232,0.6833360762405879,
 0.10029112440662702,0.33013915412809824,0.9420859062192035,
 0.16631220226162347,0.08063401980899021,0.01323786092325785,
 0.23433883111280374,0.8799760255712039,0.5137002855317143,0.9560246303573415,
 0.1791793470470756,0.5630837819871747,0.5965402343644107,0.6576974333906821,
 0.05836128072193103,0.6984585858667215,0.012060147278090239,
 0.2259284060359278,0.28393569088409143,0.11684094936707012,
 0.4120446438961043,0.5121826661605849,0.9480182942893344,0.7204922639950299,
 0.14910683633560406,0.28362454702074924,0.6155839416291652,
 0.7608473643904625,0.01209283089585167,0.8010580555006722,
 0.16362371795538244,0.6607531800572437,0.8729161162942027,0.544602525846131,
 0.30295819021861103,0.4210356731882908,0.6666511782616691,0.3326804303955916,
 0.34635482258487904,0.11151727904084463,0.6252038330030678,
 0.6502630690447027,0.09248954906659312,0.24637467721755313,
 0.037287711217465436,0.04876336350491273,0.9511720425237329,
 0.10816199251406766,0.9189940705817224,0.8717708227011252,0.378415038962821,
 0.48351997435592065,0.7631351121702667,0.17478511195742363,0.770989039599648,
 0.5362106139065737,0.46674245307482776,0.09002427072498032,
 0.47381064739076406,0.06546759671449087,0.03977916459867625,
 0.7457850554443798,0.3307637390677901,0.9835096811049063,0.8582163653378162,
 0.018574941257268995,0.9737327961230118,0.23219642776935434,
 0.5691787225214661,0.2970580726958312,0.06989532990380776,0.6340577460015488,
 0.5624895750303155,0.7200298263487511,0.5400649955163421,0.8497826271443125,
 0.7311815870098081,0.1622227165426574,0.11084127943527355,
 0.12695792635369663,0.7891606420830781,0.34544608540098365,
 0.2120753964139236,0.5644167141957291,0.774167186065045,0.7254078155978105,
 0.6855318961328078,0.697213049216568,0.7058049963574508,0.7798829356668743,
 0.7805367815719368,0.9219261781050694,0.09148980583911293,0.0693605078017836,
 0.5253517894449952,0.17518240620601966,0.5057283984826608,
 0.24368423225034408,0.30653808150336614,0.09197669120197949,
 0.18324300553685768,0.995231868124593,0.7877707927186757,0.09855675180586165,
 0.44039076501135266,0.2667734540305601,0.20509796770076283,
 0.9948760246482387,0.31081703604852684,0.8271714687813791,0.5905510349863456,
 0.5939390011542981,0.7011652535441409,0.5941069913050174,0.11152529220761287,
 0.46198262009340585,0.2259804711811707,0.8230536022450748,
 0.10687120294077013,0.08992105402252415,0.3402487468862745,
 0.1915581235434527,0.5006045180750163,0.46527199329478497,0.5489253509792963,
 0.4105421699471107,0.5771360402298628,0.8542192751597094,0.2884502330558377,
 0.4647374187626918,0.7286440081941093,0.7470097520104212,0.24050073314573972,
 0.31928081454473034,0.18091671871117287,0.39602236938603363,
 0.9343635184686429,0.029922238518827693,0.260408675167774,
 0.37774952873110323,0.7557111499077296,0.6962023253706354,0.9637553208275483,
 0.37289984014623956,0.9787436686146217,0.8301176318680678,0.6197949906861401,
 0.7128243504635627,0.5432515190965393,0.45826660772268146,0.7707042498709588,
 0.08301850873944439,0.27192327287045803,0.5816540670427854,
 0.3223291309522789,0.3908521941091483,0.6807703277820876,0.7705480063282831,
 0.8774047174766549,0.3894328066956214,0.9585585160032802,0.9761714223515578,
 0.5404123056210575,0.21493830442554673,0.19519067130338164,
 0.16246380597294674,0.2784751098834297,0.4288071640629907,0.8210387142905867,
 0.7099404792393891,0.7592225851540378,0.34231796490121835,
 0.10722333403431117,0.056098572133174374,0.8476719733299629,
 0.026857263830255218,0.711649697162401,0.37551886226798736,
 0.7929866198324853,0.6524204741760854,0.8600662653724678,0.7949907767089822,
 0.16866951383208473,0.713687392942917,0.4531280841456464,0.5863929195626922,
 0.09996040149812768,0.21133522115512016,0.39586277499452727,
 0.40120798191829743,0.7550566077595642,0.8058160395810998,
 0.24978807788925683,0.4699025638737494,0.5125648957615759,0.2926648083008665,
 0.5325016525406028,0.008478489390698174,0.5635588062216069,
 0.5201030192342522,0.7888621197748933,0.5920286169658254,0.691459315201633,
 0.7297070804225407,0.5188714885701413,0.5632148797298577,0.8355517838309979,
 0.9643573274199004,0.3603660262129449,0.7962749957691055,0.626605015493055,
 0.7904290626439382,0.583749613702255,0.8509755333493959,0.7571809565153702,
 0.2983158366171845,0.7793326556889952,0.4262246498574672,0.9481357337384111,
 0.9267488249174978,0.9728446984169621,0.12057838320210046,0.9118999989389711,
 0.5999984610046519,0.9786203526880775,0.5363270509030125,0.45776246721480973,
 0.3694921161701189,0.4866662573091153,0.7412845374491763,0.12609682686363288,
 0.003076426086741524,0.9055415847688403,0.06500085478936812,
 0.1003133913044707,0.19819351964337817,0.9341952078849913,0.5973193096434881,
 0.7075465285691597,0.5329386515759826,0.6619940182898069,0.4639159358432926,
 0.23983631122352655,0.23580541280032075,0.5843185050620613,
 0.3631203495268732,0.6680939052829379,0.994377160564672,0.5409003301990345,
 0.01726369434626973,0.5505330734888081,0.4572498207766871,0.5433790499829522,
 0.23819605533139676,0.07039567969083282,0.8631366671767773,
 0.32517877237885684,0.42296667832356927,0.6473412579591822,0.596786281704381,
 0.6384254627057944,0.22390297524569425,0.539008972605811,0.972313443249004,
 0.10183355136348515,0.26404056475182325,0.28659777573616274,
 0.7296504368647049,0.570146942395496,0.3787002224637229,0.8777276041289255,
 0.03811165619403267,0.13943361746695127,0.6610758440649434,
 0.7771963749526198,0.8273464102956395,0.2591381848620322,0.5409172573464424,
 0.4037020640326935,0.9244208319947519,0.8250594628387032,0.8440093755993542,
 0.05121958842592789,0.9872717398500485,0.5809500309657529,
 0.34913662306411386,0.10985595801795789,0.6275708966939311,
 0.24719572285637847,0.978253480967163,0.4255542375504612,
 0.004066559220480936,0.3143393871971357,0.7759715020066758,0.095979995949947,
 0.6576467938757096,0.7191142195506165,0.14262362705627185,0.6800636807750695,
 0.8366680675421531,0.30559802439634376,0.7021993106648866,0.1208158028288191,
 0.14343316543408058,0.6018085007796338,0.7116659039707446,0.718196249424055,
 0.3284977118330855,0.2162772684863875,0.47103203076378086,0.3122591338911932,
 0.8803759291280687,0.4569250763829219,0.6758586881633128,0.2748864529354619,
 0.9457800886929135,0.04584650859199613,0.05485435359995483,
 0.5397877567054123,0.009376320021012963,0.8894371483656928,
 0.32677567975451316,0.9519985286414893,0.2977904907254735,0.3126096141629128,
 0.06223619186637552,0.44033779559034547,0.24747912333318367,
 0.008108613119039543,0.9428351959542123,0.84937347322852,0.8396576173955315,
 0.2186881911147528,0.7562836812477083,0.9020558364458908,
 0.011714203050606331,0.6887363559075784,0.17801486029246238,
 0.5748041159322806,0.5830228385643078,0.7437313803941433,0.775761396520147,
 0.43900237913704054,0.4129031810735232,0.39437030277574914,
 0.7541233950467358,0.6046225296545101,0.2590131895447967,0.8942541657227262,
 0.6576483792046374,0.6523307969169905,0.6112413305950766,0.8587539876519832,
 0.040090790635235796,0.24266663962213664,0.8319739606612451,
 0.03735861333524271,0.24362613774401054,0.7212805352084083,
 0.4826760982287144,0.7677992430641869,0.17538980372161872,0.7928408626374268,
 0.3645627544226848,0.04722712940626661,0.8458636752675822,0.7891721000534955,
 0.39095427988039333,0.17365434610952113,0.1412489399496485,
 0.4503104533375031,0.7502374276511695,0.6893128954143013,0.6491688414813809,
 0.32787850253390705,0.6949414790719016,0.3725570001985383,0.6311224188028519,
 0.23362602591805082,0.8491110248258525,0.9058762879154464,0.996894075411192,
 0.07381876904062101,0.27354677563340934,0.06709163723777689,
 0.2509614813084424,0.9843820995419879,0.7463686282342346,0.7718174762846753,
 0.7319537558725886,0.3542919473493863,0.011329164478141807,
 0.7748894498588147,0.7525738886869598,0.5372250332236508,0.9431870284635983,
 0.4580726821485741,0.9857846531347565,0.9786108956587727,0.1545208470332493,
 0.6359777395524446,0.3690888620123307,0.6769869435558408,0.6515972433655898,
 0.7067760783243304,0.9190459153180585,0.29065562645880816,0.6799074975544499,
 0.7682028201079796,0.028415624586108357,0.6841285915244107,
 0.5500009650481187,0.3937210323194996,0.21941402109658348,0.8628126175864945,
 0.020234926234492412,0.07881274557953621,0.17415891660840477,
 0.9979680817034716,0.7529074982132008,0.7661053239683976,0.7939997781394537,
 0.8166955365077682,0.8230123972859416,0.8838271618924498,0.07178607233688761,
 0.5867389505841855,0.41225152016675815,0.9519867563608305,0.6174360781617672,
 0.5820741723819624,0.8740247929257544,0.7645215742204609,0.11488442009770783,
 0.619775321366071,0.21143234586281345,0.3571166452658271,0.6846000313023448,
 0.27003489732909614,0.5879532302846666,0.3531921985387476,0.9787982199781835,
 0.40699821148612914,0.40132279469249943,0.5723547984892183,
 0.5074890982429431,0.03200873399468673,0.23525552216850276,
 0.28183621344963383,0.2929763247457977,0.14978359219063753,0.966023123201585,
 0.437361076817406,0.8491290721042009,0.007037706603528648,0.8709864014287433,
 0.11662147321638683,0.5931353938001364,0.04328633644152835,
 0.16747964811310723,0.3214297834711385,0.916202928458005,
 0.026868341220272596,0.059837491289481126,0.9616173434897706,
 0.7697781720774477,0.6116627002010632,0.8137547810272111,0.9507489829384849,
 0.43733986352388543,0.20929473204764093,0.15836073883533242,
 0.8587791559349012,0.6948368143042628,0.9158558698947294,0.333710045128758,
 0.9210760176535009,0.5009628415176479,0.08961791753118231,0.9333672267849885,
 0.6005628843911521,0.3573814159490185,0.43485984711033976,
 0.20591288730480406,0.8864545738740228,0.7800254415449571,0.4260668947941505,
 0.6589350568739277,0.18822753949574345,0.3254609860937411,0.3148144719059972,
 0.4658105073821299,0.3635034637613843,0.42897286882021246,0.5219967818620298,
 0.971564662235793,0.7308820937941058,0.012641908589684281,0.2966717022869535,
 0.9818243122716086,0.26621745709178446,0.03656182800986307,
 0.40389130921427396,0.13960159179655085,0.49202130360794716,
 0.0034923868484169063,0.14748267606495902,0.1442354541246259,
 0.6944600785309029,0.14137012189297915,0.938356252908401,0.16028747232548768,
 0.454451920926197,0.9203901947189507,0.6159283469108702,0.40246940185952984,
 0.9132466803680211,0.6806864894320599,0.29723704141037055,0.5902995496109512,
 0.3587551453879846,0.7221308594965455,0.7511965694447937,0.12127774904825062,
 0.6957853702391732,0.051613078479912144,0.14598471804978563,
 0.6218029047230125,0.3780792862105131,0.2427728867409511,0.5809666155949149,
 0.12992253798643205,0.3520953987589478,0.7185885190156667,0.7265194036993027,
 0.2735920134035693,0.09473220330486298,0.7817539677387397,
 0.16929303144914898,0.27560079579664487,0.3681525612732448,
 0.44174031912494605,0.5468533276335227,0.857720283269639,0.4745318176765847,
 0.851139401474501,0.7317501756345464,0.03586914532791419,0.7391857585317578,
 0.5121585726027273,0.3887205746970195,0.026718274318988877,
 0.8866575154071562,0.05896806371427177,0.4527317443351273,
 0.14235793620816573,0.5296809247774928,0.9346316567449368,0.4159705461926575,
 0.7322575720483236,0.7218857040764699,0.5342975189941697,0.4518169042647119,
 0.7828706098231688,0.7239249970294795,0.9881226824409798,0.5919194491743567,
 0.45031767417160473,0.8822906653797171,0.36273867215755373,
 0.06306896307543652,0.8770162596905358,0.1042566516529484,0.1987271877797152,
 0.7058197606719606,0.4777910829092007,0.4411502915663611,0.9000507545507898,
 0.8050458873868466,0.14535495997374281,0.3603615276975555,0.3183743643240786,
 0.34999038377914515,0.7658897082287895,0.908251821850754,0.8109914247975986,
 0.8263521410434487,0.519011031680556,0.5376378589257684,0.43229247881835553,
 0.7473650051856919,0.7675875360429313,0.03297099675027004,0.6130910711356489,
 0.8098815097447312,0.8829062527879369,0.35434224217291077,
 0.07578281092782457,0.294736503790786,0.1795560550747547,0.899479858312191,
 0.8874558126610853,0.11793196523492833,0.9269371796959985,0.3516518836860779,
 0.06312397736166753,0.6602851994720922,0.030292109280241286,
 0.25022244668504146,0.925301559376068,0.7439150223181903,0.7575304677347661,
 0.9965731755675521,0.602997218231264,0.49987427351371105,0.8437610275826833,
 0.21069504135401873,0.6157371658971582,0.8501269139470641,
 0.10089068215717545,0.419222736791832,0.2941423352307334,0.6543510407446685,
 0.2600620545281863,0.5372760441524478,0.9906704962321745,0.7348980358418047,
 0.6124752467969278,0.08657216889629082,0.5267322564876482,0.3454255762553724,
 0.463588120329963,0.697140251435125,0.36476715830163553,0.7519884707282245,
 0.590641201325313,0.846927627652413,0.47584335485249063,0.5254641460512764,
 0.6853658594691672,0.4353069548736045,0.5725302148801742,0.1523055051269403,
 0.5117842600644658,0.5739169798625942,0.74947380628309,0.26457850505216207,
 0.8383428725789895,0.5754454679552595,0.28080429846893884,0.1987542242969248,
 0.25608839531152183,0.13987230353782787,0.02944581531734547,
 0.4905295642090617,0.4436341258055668,0.32331327730762216,
 0.39254505241179927,0.7021782312765921,0.5677069636474481,0.527181405951872,
 0.8837184018101124,0.2774833337890611,0.6490524892239139,0.5324476474949287,
 0.24775139942299362,0.571107768658454,0.15898914128792663,0.7263389629661036,
 0.35428610642430325,0.785132445896353,0.8779020751889219,0.4649746124326497,
 0.8394421887276802,0.9009580166216331,0.3614670303910633,0.7192590921481102,
 0.4531102673905293,0.7305602280688015,0.37298567691520124,0.9669006745480274,
 0.4065895623856236,0.6292558853301736,0.2190824874052476,
 0.046237855205147005,0.029806027447062844,0.7683664697611334,
 0.4798500494550483,0.4373750828000992,0.6238851944288972,0.14308463590078302,
 0.35522039410416884,0.7747285106160781,0.17907221703831044,
 0.7146075477957909,0.10025997918696139,0.8638472619374111,
 0.03345284657038561,0.9708725798382675,0.8636320270629487,
 0.24868952332083583,0.17315153389315707,0.9610657630402009,
 0.2922338736745693,0.10778991767033785,0.5619642138671422,0.6389290150211275,
 0.5036478614027139,0.9868443552529723,0.13424049971374052,
 0.38794097740915945,0.8605684960117022,0.23986428497670698,
 0.9596937024323452,0.48535281643851014,0.2006425540013358,0.7923427999611518,
 0.7418932389050139,0.7768687923306364,0.9438975058172553,0.9874899309121034,
 0.7965736902326647,0.97089482360407,0.08290865267537151,0.27747465421458406,
 0.5706427702470255,0.8430590162812917,0.9577988875938341,0.8143142701550796,
 0.3700387006795802,0.9834896877767589,0.26580949639354,0.2508038226041207,
 0.12592737308411417,0.41761274555689587,0.6282152546959164,
 0.5613114068143822,0.558143069388608,0.05841683810412779,0.9397725313198355,
 0.5961461762228686,0.3789082911230568,0.5221580507792979,0.715921410682314,
 0.4908981041392527,0.15459319021960027,0.39674520890252674,
 0.1229043429219625,0.5562461322331308,0.4707868202058112,0.5900010763002579,
 0.5530902978515592,0.9100154251598516,0.5751704905242017,0.7914725887898878,
 0.7771225109833723,0.3516083077163733,0.7190665926083891,0.411869740776033,
 0.3574762913005396,0.7404444939473442,0.9289255521925953,0.379314551276868,
 0.8888211176900999,0.19978511512174135,0.5677589871011266,0.8641603425463853,
 0.7229370364566094,0.06404453427846868,0.21294227399435328,
 0.23661635950998938,0.43414251623666933,0.8422428249436772,0.377432336841801,
 0.03686544055074503,0.9477492054057587,0.9793004446806693,0.0627513673694522,
 0.8380206344321817,0.20729898147205716,0.5453029772425572,0.2533747016182286,
 0.8204139123006631,0.4410227344031522,0.3421095796528635,0.1395016147901239,
 0.35286084512345883,0.8084048640157085,0.8831428632417119,0.5061653181167138,
 0.5649104207193548,0.885166877806151,0.9877118001643609,0.1813708846081743,
 0.07697713033454978,0.12511690387559682,0.26610437155845235,
 0.11021763198663381,0.7370195142990569,0.46034858764607733,
 0.3135209892908394,0.29721172174258115,0.10364954011889349,
 0.7962629121446915,0.8072721074010304,0.9232272151116587,0.603247376967394,
 0.22964550617158563,0.5911177428323869,0.4737591497040238,0.9555473889497456,
 0.06754813085317901,0.6357128654902522,0.036447523237992385,
 0.13641541487831788,0.4419815263290634,0.02836520585277902,
 0.21502645077435945,0.24501403009609757,0.2821325383212342,
 0.15921459227977353,0.6135166824056069,0.9076466631077444,0.330458109062414,
 0.09081354018871002,0.39573011510324807,0.0468595193897946,
 0.7243293686658782,0.9300592358378486,0.5933597937773449,0.07462887084728731,
 0.40346923130265566,0.9362620294338979,0.2624127235791376,0.9380578170004852,
 0.12070102804900251,0.5320315151878798,0.7700479853080521,
 0.38290275205520696,0.5343632848307109,0.48409642211699655,
 0.6340304930392706,0.47061978621681955,0.8006562228037064,0.8543823786615283,
 0.5410357798274041,0.9148518812308634,0.7364552629003797,0.1463260311257435,
 0.9342130763657898,0.4949278799868816,0.28347867376345937,0.7920133632017428,
 0.03966218743145877,0.4814969654298632,0.800705273470723,0.5919782031469358,
 0.3056928407221514,0.29639640173621784,0.8481654883607531,0.504563925952279,
 0.5878960014549557,0.13223834471897578,0.1721166444284461,0.3909236993045305,
 0.8441031391180855,0.5868786567132646,0.6289561240650137,0.1659731624139894,
 0.30247478957552487,0.7658803876527674,0.5767980994719358,0.9400983670382571,
 0.46418617336878143,0.4411727164782939,0.5041369555218961,0.9565260790937329,
 0.3203402683946752,0.3593999544617743,0.03780353427092875,
 0.38051268225686674,0.7613386040607806,0.2188241679753795,0.8367414865720398,
 0.8081296700024619,0.1449692358438439,0.8035674653325844,0.21779757012109258,
 0.9425331928425368,0.36864582941903346,0.13515365838723414,
 0.5501379766662403,0.7370125005272947,0.6683310372506409,0.46827268052496984,
 0.9941339174397409,0.5978874223310701,0.3172989453335847,0.2828841646096666,
 0.19128633981912202,0.33426549099310787,0.0450868403273077,
 0.5293721906150991,0.6665076838493148,0.10500825029082284,0.8649148912062897,
 0.5268061385573013,0.9614991743804652,0.31384319982397313,
 0.027459715966391407,0.8681069433767103,0.416518784836716,0.3501611924903413,
 0.472995005365056,0.9921445156834068,0.6122615386216158,0.4082821130918972,
 0.6488441716239975,0.8956228876845538,0.887435415817983,0.2368948959271252,
 0.23196812066709216,0.19371303966477882,0.6982181942533923,
 0.44386314949476424,0.957576040658154,0.27844736925081703,
 0.17914705812762644,0.7066997982795482,0.2931270684896251,
 0.12383496303158992,0.8915073133164059,0.33197945982826305,
 0.010067046547638792,0.1740478245372845,0.9125478123043086,
 0.6406484375015994,0.4712461811854375,0.33444500244299147,0.4031220224135924,
 0.21168475897405603,0.7588690148373908,0.9899707410982783,0.3260272063204084,
 0.4311672263515962,0.2985915825541816,0.19062911302404195,0.910803526820239,
 0.38741497948996473,0.2927917047528483,0.5192914247106242,0.3940633414361441,
 0.3767336640590375,0.002617616614705942,0.6764462243402833,
 0.6760406104491505,0.03985870468306463,0.5324111689779212,0.9237598764846057,
 0.14861032708571664,0.5438880903183638,0.4703660997230489,0.7048791876979106,
 0.04145279265354995,0.5787065695703189,0.15163238246961697,
 0.39669114551609885,0.5755452864835917,0.7990249300617736,0.8076746589703819,
 0.32530592718353446,0.21752564442776579,0.25014078725613853,
 0.1975048107554056,0.1590582208543726,0.23018098989966362,0.6260157300738575,
 0.12524534542225374,0.2540070920298668,0.9826429739385057,0.6562258369576215,
 0.2545321887927102,0.04817174395326451,0.39216871638446693,
 0.6451843324586524,0.21839460179151882,0.6540901911787512,
 0.01231485227279916,0.6602677641154651,0.5843297551954505,
 0.18832261917878657,0.22361766843018382,0.45141459372979,0.9912533234846759,
 0.21774766948422097,0.16928133710716842,0.35695682920587946,
 0.06822218917296219,0.8116466927104491,0.5626929910433186,
 0.34163387572874226,0.27479857570615707,0.8174205393635943,0.491861484204159,
 0.3656100987626908,0.9384063532351113,0.08779397781487275,0.3806542668766516,
 0.2432941355214706,0.22209790051740208,0.38805858203040655,
 0.7115106962573892,0.8613941874447695,0.2495459555211793,0.48981449786403486,
 0.28610087510113325,0.8594978787206592,0.7476453020672593,
 0.13575307470065245,0.3319357684260482,0.43208354400180116,
 0.9957333285595954,0.9500100681962808,0.18232549138517706,
 0.06501531124058435,0.030302280034042006,0.19053952596902635,
 0.10931369055512519,0.6970380587604623,0.11156852803711226,
 0.7826244577224071,0.7682190524506758,0.11621333075476237,0.7786471769053971,
 0.11023676587176157,0.795467674874865,0.6127569539351364,0.7489352423390235,
 0.39612346895227146,0.8493653590294259,0.5761394025943026,0.370049344984542,
 0.7834681043534572,0.06846147495704946,0.8851048864997084,0.8804886681993807,
 0.32064946119969906,0.11124338281728885,0.10825300353545808,
 0.7501977262742965,0.3394808320332756,0.576253633752886,0.9763261155842244,
 0.23242610101019512,0.756140956915359,0.27655063774264865,
 0.12275467292809217,0.42981738678982817,0.612495023358712,0.3660326812926016,
 0.801409655225015,0.31539916352270536,0.5306128147723547,0.8013056319336592,
 0.8292899329580062,0.8658352422000644,0.4390574542426142,0.7877556521343694,
 0.2471419061429604,0.6619067609997901,0.8150784511737277,0.8710953627837013,
 0.11552512934747328,0.570623041756791,0.9250772821731486,0.40803256174509983,
 0.12701635185229265,0.8212950994730966,0.330598877941195,0.5836091192195081,
 0.7886724055028824,0.3806855407338433,0.060520583524242055,
 0.9506234580979986,0.9369042544928367,0.33213796912290827,0.5515542005245289,
 0.8923394694564458,0.4302074253333682,0.23633048727933592,0.7320049801318136,
 0.8694473631542203,0.7280412103516735,0.8535225719348348,0.24445288687060174,
 0.5219803729299466,0.07076243565509943,0.5434260658367724,0.5773049638149085,
 0.6230264147586058,0.5326446318992092,0.6070851440899684,0.4935861745070431,
 0.8968581071048566,0.30748945408249173,0.9399915132783601,0.369231480004074,
 0.30411936035876985,0.7550905294297222,0.14100920868665612,
 0.11359037413027606,0.20993523521348056,0.04449925587242953,
 0.4187650645020604,0.8412200473540714,0.1280443952572461,
 0.051056608613594734,0.17095742590685126,0.6154774097265987,
 0.8803929551155054,0.2137805368111354,0.3888813119039851,0.9708824131542979,
 0.7840759723238655,0.9422492163730753,0.860147426171012,0.48086961997527133,
 0.7068305812975699,0.5516765511593654,0.6115335569342744,0.00603029580256409,
 0.1653232527636841,0.9829912188338811,0.08571372288890478,0.6233385258089147,
 0.7399840942852152,0.12929862212162913,0.7229676217846592,0.7329056291095366,
 0.4426327440144894,0.7936621302262727,0.11961941476033267,0.9547771409812096,
 0.7690739562778425,0.5505709374479822,0.60660140338065,0.9822620098490727,
 0.6591334352290512,0.8288657428946969,0.4918970345407031,0.7645171339761407,
 0.06731674035339075,0.185400730053788,0.22666040520216857,
 0.12670257909600657,0.8853979708601689,0.3222180814322052,
 0.25882570164220187,0.09574164573633592,0.8425177987107316,
 0.2921076376497833,0.5945041001529389,0.17253832605696195,
 0.11423839017164783,0.31275936030962814,0.3039723137173721,
 0.1909900917908065,0.9347897479787194,0.39614531865356806,0.4620263626484794,
 0.6998245396190276,0.12653494229208728,0.15352496270187832,
 0.14982116453215766,0.13416543199820008,0.22915217696203216,
 0.02543029872293412,0.216986765757066,0.5104675709701282,0.6232624523855741,
 0.983860055748333,0.8085127945830662,0.6331670499959101,0.7787882828227188,
 0.3073324100818964,0.5686423041466159,0.40963303213798785,
 0.04269564272110937,0.8665384509985153,0.5110573384771675,
 0.30373224891015993,0.7758391895875305,0.9811217399240216,0.9930351810258607,
 0.6990276757905237,0.4184542187536352,0.6098045629397915,0.40783886766615496,
 0.7400883429847231,0.03830187585989253,0.2698428657535573,0.8132004408997304,
 0.14031985308091022,0.3276063305154434,0.5345428423444741,
 0.14499590423422537,0.8465667458563961,0.7269722952278388,0.730514239579334,
 0.009776193857009008,0.49430875577158884,0.2440443331965534,
 0.5550666873536105,0.909894050432995,0.2835679584824454,0.46794530140433965,
 0.8977451728435643,0.8200387531620842,0.33234844020813636,0.6462605134666392,
 0.7006400251625945,0.5164456790921539,0.8829132285111158,0.9042599588094831,
 0.680916089971776,0.4444450480980031,0.8237834941648189,0.7972925276583733,
 0.028974760158507085,0.7852379626766633,0.9350613608654303,
 0.21475489007401305,0.6857511362322936,0.2642630567250762,0.7790110400865097,
 0.8761351035384238,0.613977618305992,0.8368326081211304,0.8037848316827203,
 0.5428222803594279,0.883196894260431,0.5866050086783128,0.6290641638712384,
 0.5262811692850762,0.04132425137537554,0.5833395311501809,0.3786682768860884,
 0.34269726717744076,0.5084070258563077,0.41900115974444097,
 0.7877874436860064,0.41290686522616915,0.8003454893138844,
 0.41261529486180304,0.3947627902591896,0.12789518536604916,
 0.5566829407570949,0.4011425095013459,0.5804367630822653,0.4540253555302399,
 0.31870231471285715,0.9479669314352444,0.6658180058803284,0.9502330108328401,
 0.7465628662544644,0.8928242656764716,0.659081028426422,0.07535813088022691,
 0.7857225489442131,0.4308927120472291,0.5078608238053147,0.24854363657501066,
 0.5741309483620295,0.27221245685204387,0.9568817998480433,0.7135426933490538,
 0.12656617224290234,0.18362392880877276,0.3222301787985987,0.337098593982437,
 0.6447835578346105,0.5927378895101474,0.48799229070122807,0.6877697228307665,
 0.5577032887281135,0.12370150356434317,0.3578966526212619,0.5069609828318266,
 0.9418340325984215,0.7381286134202262,0.925283972686093,0.14748852383586664,
 0.4914583424966369,0.04265634099609472,0.5324569354791074,
 0.0010318724563014303,0.9344305166020239,0.2564403032544613,
 0.5011881369067485,0.7659749264669082,0.9416878090849153,0.11810802150715904,
 0.4019279670185609,0.6212351607950757,0.09415244607497442,
 0.22154196934684534,0.7426731118622915,0.7239735775409661,0.4826810870660446,
 0.020631488497579453,0.5450549203756396,0.7536463340407046,
 0.014671013550601075,0.5120589690210962,0.14157400092078354,
 0.8330779957263398,0.22294012315794332,0.7783385623598028,
 0.09494393786450328,0.7141763271862657,0.29508464122815947,
 0.10921382346545772,0.5732851891508467,0.06912888011960627,
 0.4320597747629267,7.83682441761635e-4,0.552005589255955,0.7769342009886515,
 0.34225739521709053,0.18187535599089766,0.43605518752946526,
 0.8809172230044116,0.38345418180207647,0.7989889940736085,0.6380000755209384,
 0.46485527039962604,0.44907302089777335,0.08479219812948757,
 0.3662547418558906,0.05661630005978768,0.7597873564044365,
 0.14380227479080465,0.5362949408581699,0.19091096339181324,
 0.42145990539165523,0.6689524824633697,0.8442752665450545,
 0.32628728298119425,0.13895067423534935,0.9048536940002933,
 0.25289875887560753,0.07672626664082893,0.8550753923234992,
 0.6154263292111365,0.20815134095911614,0.9666468031512443,
 0.14172087597658167,0.9260732326340144,0.4578327213525033,
 0.17388342290070924,0.9187798275144354,0.2813098269717702,0.7232523025213611,
 0.792835850168889,0.5659479732378061,0.28639368739439863,0.4340939609059825,
 0.1418185051456754,0.2884600166534331,0.4339819425288042,0.8837354671215961,
 0.5620220652266227,0.10598244388672873,0.46564442536425343,
 0.5770980597999181,0.12143660989963467,0.5240521618262199,
 0.01176465744568822,0.03146306468749582,0.9555103245644415,
 0.8776771779694914,0.5395378871002086,0.5468888236537405,0.9982843081996733,
 0.7333686734960491,0.8706294224199035,0.7699075467428098,0.42823406685515863,
 0.05869560295791876,0.7269599364476468,0.16117050146232303,0.711492351443271,
 0.4950780993362286,0.4134231098953606,0.48863448319090674,
 0.13672278183901854,0.11977270505162985,0.5549493768436786,
 0.3055210863356851,0.8012889821128963,0.45215007073743363,0.7869576952151526,
 0.22361084704828682,0.8741398208595746,0.508713588272033,0.4587995969858073,
 0.27051103181134795,0.24104053272029002,0.11667598798235068,
 0.9706814670594355,0.35680647061109716,0.6083925271521395,
 0.17246080206731296,0.5505349137142452,0.5409238137895006,
 0.22747354575512158,0.9431896075543472,0.45409913698363347,
 0.5512998743807676,0.43834556237101063,0.2608271695772276,0.7076969534040262,
 0.508901301139476,0.892574505331932,0.9543990841343438,0.9263797318503455,
 0.12693529061101994,0.10998598099332946,0.021490111903974873,
 0.5676735278165516,0.33611799816667676,0.7775375910941212,0.7088229020637963,
 0.1550709579851638,0.38756303277153203,0.7973770462359877,0.2162443961184881,
 0.8426790736685601,0.07341675836050476,0.9397734306651974,0.5260285609852344,
 0.18320316173355922,0.10122089863745043,0.9770829527531895,
 0.1119284289908189,0.706280980940662,0.35543886002310987,0.2521588095010816,
 0.8042605216068108,0.6414208201502619,0.06980601411300247,0.733744209287922,
 0.048617699023503436,0.847337964618723,0.3114877222954774,0.8227721547686851,
 0.6064315224677659,0.7085314767872694,0.5034230791006564,0.07118643046603479,
 0.921703865027394,0.570424857035392,0.6276305247900694,0.24157500765743,
 0.800953710070087,0.027522265546143787,0.5190415580305493,
 0.30371655848656576,0.45771443235893805,0.8457290164439136,
 0.43225660988000303,0.8519542329910366,0.8285748156829907,0.2277278416500862,
 0.8913513166669007,0.17718854463228184,0.8093995680559504,0.8433792123418623,
 0.7129013007716762,0.33293426737781684,0.4713601526967133,0.5870545006148539,
 0.31892750375785117,0.27287775825641325,0.7058074502445091,
 0.46400507795757906,0.45355126098326015,0.9920623199066041,
 0.6982888786477899,0.822362830465026,0.7149201357500505,0.9288630953779231,
 0.7185482343556671,0.4071487634990354,0.04704058516449605,0.6661164709604639,
 0.832955140823239,0.0028157217373186993,0.022342156862692475,
 0.11173504343833696,0.3333380406856873,0.9451361855595245,0.1998499104208903,
 0.16456255201770542,0.46632127217327446,0.3087362377719812,
 0.7281380362345048,0.5996593505052482,0.1163791725561587,0.17053409869107972,
 0.2678209951746111,0.7893638107005343,0.306871647157854,0.020786643928472626,
 0.9470071537441948,0.7651821654702369,0.4143819314132353,0.45362355310029223,
 0.8882238576478746,0.3183510418343476,0.8928371797251065,0.5370382630460266,
 0.6148248042650432,0.6624886928838006,0.9471289350610359,0.7877041058067015,
 0.23081405436715452,0.504846118111258,0.8956932177948008,0.4378009688924157,
 0.01570741122002517,0.6592006344958856,0.6654880180264356,0.7059734587499613,
 0.8229320132685441,0.7680021449290896,0.6424708628561745,0.9524583068994296,
 0.7883351065229843,0.2114802665763712,0.787034583903153,0.8936172146186468,
 0.560895872104429,0.8073947491127569,0.7434876056123539,0.3660982669941464,
 0.4898882753492051,0.528122653559567,0.9980844753013264,0.5096004233597631,
 0.6395632483870997,0.9252667374050799,0.25387485153839906,0.6522619569919861,
 0.3723764228223505,0.634089251302788,0.4810210138397487,0.4554833035428306,
 0.25463106404928615,0.6313709023093634,0.8907353108681304,0.8989438028205211,
 0.986311910594961,0.8524473380066191,0.18964229010129674,0.14924210379825353,
 0.6313139763192634,0.3638680557741961,0.1706884247598378,0.41153470412117354,
 0.634883813011976,0.9172410991975186,0.18545526408613122,0.15024583677477843,
 0.9220815914608005,0.8283751892736668,0.5030446057024066,0.487459665376196,
 0.14113077153822384,0.6610067588946068,0.7620089699955083,
 0.038338385556364596,0.9910286887214149,0.3327348264527584,0.319835746466846,
 0.019646374243232012,0.5147918025011188,0.699332444951993,0.9637585227320034,
 0.32487154423466624,0.7497346934956959,0.3763222459507577,
 0.40400909284073827,0.30280226314025505,0.25088980523512194,
 0.19137471832161312,0.3405462750230799,0.506182786167164,0.5736498824053046,
 0.33622494897833866,0.7942436201347474,0.9851549932409699,0.4213624466022452,
 0.3196729021240128,0.6116784286272309,0.6875590303179203,0.2754693465714677,
 0.9732644034008471,0.6264857971227723,0.24986920229619725,0.4553168101416101,
 0.7334626376019785,0.045150363621181744,0.10941673980842481,
 0.2995988448511493,0.648864436373939,0.3050815750971654,0.9555628984532957,
 0.24887882193427358,0.2780722251174841,0.4386211395677987,
 0.32373289084215906,0.12282214938181935,0.42819628640490015,
 0.257515023889605,0.3153980318017069,0.8638807089452528,0.7156661311359843,
 0.3863030434606519,0.5130242626189973,0.8190721534751804,0.6233731175386901,
 0.9670265052230278,0.2859137208109386,0.1999603337687621,
 0.007408152524329692,0.019474726780154272,0.3121232753591051,
 0.9583406562098175,0.3639905448779943,0.2767526285941051,0.5379161865443564,
 0.06401016770596324,0.49468235907317093,0.5954392758592557,
 0.3057458997704905,0.8717821108193898,0.5828865043669399,0.2800073911082098,
 0.7286293198786731,0.09246608289874048,0.8556386435694417,
 0.10979983837524931,0.3075556792394042,0.11853872353609307,
 0.9541498546058377,0.9995395678895962,0.5717408919201324,0.8877621026402839,
 0.4266894817983251,0.0646791113291535,0.15823892650433147,0.7165134639493097,
 0.38009754695414466,0.29860337305439344,0.8062148399013568,
 0.7869973978915488,0.1755871891653158,0.4005448616216952,
 0.026386778795529198,0.28395215319241696,0.7533465859736682,
 0.6344766776604105,0.0689039840638177,0.28307442957600537,0.5620347985448412,
 0.6741851773976952,0.3049219680051467,0.9439440548773164,0.45618868920949496,
 0.8223406126496337,0.6032511476889627,0.19360853118847066,0.6694739180823415,
 0.28877930893809856,0.034385203112826335,0.8834283663283193,
 0.46795611232054635,0.18531440522971931,0.9896106639956946,
 0.8607271657352915,0.8351999791714981,0.9980675905134961,0.6778731181827653,
 0.003883120866999601,0.3623654939684249,0.5855497294296101,
 0.2820987799576088,0.8025526911516304,0.7521560576064801,0.4840257846053876,
 0.43192448778109693,0.06306652525457512,0.49765179022229133,
 0.27087445621182193,0.7878860922587437,0.69126187313668,0.2044038575865348,
 0.3191114201646963,0.24944554630846927,0.8821885126612994,
 0.25243898610645543,0.902166073825183,0.5816602209238837,0.10636593116448623,
 0.8079513595639553,0.7667703587648275,0.37461196108945893,0.4937277360222151,
 0.9028884723260325,0.9602110849191254,0.4983293458007807,0.264205650348476,
 0.7775885462036124,0.04313228381010792,0.06223892695994737,
 0.6056520090693952,0.2896877322289231,0.2452675465105839,0.6378921988243201,
 0.9523918848390467,0.062357582741647044,0.7412435390923041,
 0.7692623303346329,0.7321723880342293,6.92982161245248e-4,
 0.22528949099418105,0.6914688306590165,0.4043072564059821,
 0.35062506142193206,0.1834141283265157,0.8609191124974083,0.5202315698050658,
 0.332547225335019,0.9047783272950706,0.6453734119918726,0.8304672415557749,
 0.5196199421080835,0.09461714844351843,0.042832102886129375,
 0.0013212254921399769,0.2610636672833959,0.8398159559515052,
 0.6309975981741437,0.6497548277867644,0.1732592377208999,0.9060792890897463,
 0.38591998633230506,0.16493458769558667,0.878636050293258,0.5316318015471992,
 0.8604667766517475,0.7000590671819307,0.5415190151585766,0.6010533980469174,
 0.5582905433653503,0.10034995324426366,0.4470141536056311,0.2503505813443786,
 0.5923008449244332,0.5854208816719626,0.022115095549294805,
 0.9091115971907854,0.6910991368253281,0.07814595325204121,0.848882141997824,
 0.7859881714290032,0.17834444880524414,0.023164886816104957,
 0.714839830815593,0.7834717050652643,0.11957430861805707,0.6089492721876472,
 0.4882342897163303,0.5358799126385605,0.6989652925591908,0.7476645251515407,
 0.9527609839007615,0.36857096118359045,0.5130784636833111,0.7950783364321234,
 0.8852339068211447,0.8563789961100693,0.5873719237098218,0.38717509965322794,
 0.3866164005776226,0.0613458335458692,0.1536683808850926,0.15009150699017093,
 0.9218614460246838,0.052861995559128516,0.9130351931819931,
 0.5329851237364897,0.43995889893336315,0.39052144472703354,
 0.7171638306797466,0.39069416507818056,0.9017409316648346,0.2523646961350128,
 0.9079647323862575,0.29223161664036645,0.5180567594474547,0.6481227897977746,
 0.6116467475383838,0.1558714424359008,0.8035005517922805,0.14370858133970155,
 0.26099793808921046,0.7057128302094354,0.6897238230501885,0.964937978271921,
 0.6355386215125933,0.8667071987334707,0.8920270446873959,0.8068890398444138,
 0.8888098772783951,0.4183150295633836,0.6948660139243674,0.04560372119274836,
 0.5061574569567184,0.5892432624431205,0.9466038329125217,0.6910841424340386,
 0.38654067337776266,0.8766092950441218,0.4400261512859456,0.7110068902805992,
 0.2244928107375327,0.8765774330951027,0.500502229176135,0.6181003764511351,
 0.5848382158095009,0.8380108443230075,0.6937001701147241,0.7036886690487089,
 0.7396306688386898,0.669631716325639,0.7548639271132921,0.2109452280311359,
 0.34180379193466603,0.272615863270678,0.8796501171234845,0.811164039612857,
 0.4609482136094961,0.5043140778070766,0.5649290526108037,0.489976149550467,
 0.057600830206402054,0.18085601150137487,0.15514508735975152,
 0.09495085640972356,0.16520958756729243,0.9763954333658412,
 0.2149624761217116,0.46258081369083937,0.7485899551644601,0.930956320703744,
 0.6125871739952118,0.7706766227159473,0.40637837007310174,0.6646804421557841,
 0.6341931630604458,0.057814454154157646,0.04482543188196264,
 0.37917566838033867,0.8628851318447929,0.3188740664876888,0.691675605113847,
 0.3362975143249691,0.8354177632934936,0.43709703239636477,
 0.38573999259126746,0.6741938693404715,0.8560613986986968,
 0.39437332656361057,0.8896456060566987,0.6455821870244869,0.6441211012182052,
 0.6049536149279745,0.665939714103505,0.9372801856973771,0.35234094952145545,
 0.32395854304000327,0.026947107993644415,0.9873540306056896,
 0.6900062062458212,0.267322262131222,0.4928461969183521,0.8455636693565128,
 0.3533562387500738,0.8795931712483298,0.4713049034233536,0.7126935140265888,
 0.3755521919842384,0.8588101639257357,0.14339210587605433,0.9512801258216672,
 0.9865542223818505,0.9059346176134732,0.07502019962294482,0.6504412354425838,
 0.899353674191221,0.44506939930616307,0.28676237254065984,0.9508815685024047,
 0.04111531692938353,0.47632325458895153,0.8299839679678771,
 0.9064046699774507,0.520305042348051,0.3474114740010332,0.8151072959330925,
 0.15844371508381927,0.2385497168703622,0.316518557085669,0.47007312457057526,
 0.4815704743438479,0.5938647599681829,0.9728931826413777,0.34493678037914755,
 0.39061953755275647,0.21814135964338432,0.5071253369921557,
 0.3463716582228723,0.38733772826023904,0.8893688095212031,0.8228842062979831,
 0.9009026937765067,0.8344923776616161,0.4657285897259078,0.26180955341184053,
 0.7438275280570285,0.6722787143245056,0.47853597099177214,0.8416175842321041,
 0.7355214604800281,0.49410911300814564,0.15838989193826647,
 0.8864710554992009,0.4913211754912803,0.4589769868712251,0.6853198324570462,
 0.46591031660897175,0.6769430468939702,0.27444527232444027,
 0.6610822934031456,0.42258906004839225,0.07980961038768208,
 0.2828298210036937,0.36767061438531135,0.4973708624226426,
 0.37086364466193333,0.5883945846214123,0.15227415791131071,
 0.6305023966998236,0.03960659951574941,0.4813340596998613,
 0.16551749998927945,0.5279217361922226,0.5521681163360276,
 0.45952472644788334,0.3210490324058781,0.9847949535327052,0.8487650568325344,
 0.7326389606210792,0.2583816684838165,0.4780576836156436,0.6571701171599065,
 0.9291537504048343,0.41703987893224137,0.06589296500082131,
 0.09866520547155966,0.8855114724123121,0.7461755549765785,0.967996888852407,
 0.14602697625812022,0.6653916429890536,0.1853805858802282,
 0.07685773140101548,0.35344250902483143,0.16800529988895085,
 0.7481588790842292,0.14844516228273086,0.997709773467989,0.9395990700756266,
 0.2779964957312304,0.9261734753991133,0.6102795665808798,0.18565953816376446,
 0.30458586812128874,0.1452006227451631,0.007174010067756864,
 0.02610306824942643,0.22144847373734466,0.5760566575286958,
 0.5248957935701843,0.08771387616375903,0.6271945041194092,0.400619325287072,
 0.349044754996517,0.16101508643638407,0.5039600841614116,0.6694990263235325,
 0.34475833539691025,0.42359012165831067,0.5093325415592435,
 0.8080362120145953,0.4236660770412466,0.13845369517438122,0.6252503383593754,
 0.48980158221107306,0.6612399756388517,0.1525121750230929,
 0.03279179863070625,0.48210457152982844,0.24282439945166256,0.55170200295047,
 0.7352647946835492,0.2996162175618896,0.602648035085387,0.17819659770284135,
 0.4575981607173577,0.08063956984585685,0.553516463686217,0.826188514308019,
 0.15071033781005594,0.7626001008804354,0.8718246877406066,0.5470055440837337,
 0.6647285458552747,0.357619383163733,0.1380813866723315,0.8347992965070467,
 0.6199321460958164,0.029559366521432318,0.3502890801078462,
 0.41790641642232385,0.115074533216174,0.6464010563160246,0.9460950836327247,
 0.009698012966873781,0.9078809651005372,0.8901348634356645,
 0.12483019577568699,0.9930575296076015,0.12320655434590033,
 0.5378518648573419,0.5446389582800637,0.6256060246045808,0.8039670481324973,
 0.1686829683979758,0.24959987321699173,0.7981692457991508,0.9436466288293377,
 0.6658833745601477,0.017746176843880024,0.2582171353721092,
 0.20289403205961654,0.22923837314801732,0.09525927761129249,
 0.990224074253681,0.851814783749593,0.5161328232286346,0.35266419613683464,
 0.5291709015480778,0.8261500481961135,0.6539329162327563,0.569560840672629,
 0.898551118927725,0.4998703250491052,0.6522886193884714,0.5821272791420562,
 0.24312714242176414,0.35626296910210975,0.3097128631232109,
 0.02652018554059965,0.8894297132015768,0.4401012543197439,0.7441497891678633,
 0.169476730943523,0.9314830010053364,0.9165232464953901,0.42797244963242265,
 0.8604698008575797,0.09040460931262773,0.923611853663798,0.9414864318204562,
 0.25559754579290006,0.029612488301821727,0.8321230330172912,
 0.5007790515166136,0.87885591144879,0.7642090042636536,0.9060214474371536,
 0.9839697907844025,0.10408489198086635,0.24330972523844618,
 0.36606969932292244,0.37060078830147913,0.4721692847917446,
 0.5473482522703452,0.17450814426448702,0.8707328051151534,
 0.10924325905762078,0.7659780869993188,0.2026084210570116,
 0.35423002669128434,0.6263554999018398,0.48048057509955155,
 0.7166644461828288,0.1225987801313384,0.8238716829890957,0.5511998518804875,
 0.9621364865822397,0.3778078037941921,0.4445135430903645,0.9309630505803373,
 0.13206799928491897,0.9898933874124685,0.636557348749037,0.8690284916974306,
 0.7552176452196212,0.48744128625046923,0.32918322770110764,
 0.5131803804429165,0.23854250731695414,0.41079303334380357,
 0.8063437780529027,0.6670654096757601,0.1855936265680206,0.5445407630019653,
 0.6819745951275669,0.2696037327838212,0.12976680931786166,0.6459542257270983,
 0.6831854012155063,0.19260813341279448,0.15086338982236802,
 0.28073071344683675,0.5164015424859385,0.7426586060477494,0.7090911981703659,
 0.3950574065626229,0.8686739074425246,0.06898613859201086,
 0.27591896003380434,0.05871244301575407,0.7833766345133333,
 0.9403273748931354,0.3210457653557707,0.5204061124402006,0.4437263022165008,
 0.4958042018907447,0.8300134258493876,0.16578435115879187,0.7059856744401616,
 0.6929405776793794,0.27812943380670996,0.5727239547449393,
 0.06414739094768951,0.7584128046944801,0.7971806602106389,0.8790777564656961,
 0.8511384259543628,0.23473290485680154,0.5977056692566893,0.8652852901724333,
 0.7547813181540889,0.44192147234780166,0.37433005832739996,
 0.005872030141531481,0.726805383465049,0.8746811091495903,0.0531233387466743,
 0.09846836762134759,0.33582018522046686,0.17208031214407193,
 0.650967005343205,0.3690499320374948,0.18775833949630893,0.0864214602808403,
 0.2342040584114845,0.5186661047154931,0.5585906288946781,0.19341478858770123,
 0.5287101209327542,0.02511421875530384,0.3502887185809146,
 0.02410928015110525,0.6925062494109736,0.37228268606019865,
 0.7213563195657761,0.22916572730971907,0.025751062281496928,
 0.4669354532678357,0.31470663685436284,0.7887031823705886,0.5812898897658257,
 0.5745260048712917,0.2946177907477566,0.47915168967206845,0.5248153404951611,
 0.061303115472928216,0.9272300574947242,0.5429004882166724,
 0.7850118129322088,0.4139605726752891,0.7700844675642315,0.9238790634998222,
 0.5103467247360907,0.5952779124740623,0.5997970561181187,0.101732348203179,
 0.852319035167109,0.0648529781010323,0.006423473715443051,0.244585174677184,
 0.9340452741383924,0.17205334313165144,0.6203394821939253,0.5344177888154447,
 0.6492604056034494,0.9424562110783675,0.7611366026455672,0.38494125436023086,
 0.226463089072663,0.344034293106087,0.10528051064317634,0.8907181001393776,
 0.20750346968264832,0.9786686226009621,0.37813882899309137,
 0.3271422061329682,0.9735941318151086,0.06539430692823489,0.8961522568491292,
 0.09534539859760371,0.11053080658810177,0.9832091628876238,
 0.6782098449557253,0.6218771462192044,0.7119184383828521,0.22326382855686766,
 0.5677374060990559,0.8037437019106841,0.9570386163514839,0.4022745503020455,
 0.7496431557217008,0.12960912075685604,0.32668577929486187,
 0.4836159649884837,0.27564434884065525,0.9184785925063523,
 0.27031948515568005,0.8432583410029703,0.29918302422288545,
 0.9569905839875101,0.897841801142325,0.6564779200582906,0.410345942701156,
 0.9220003478331066,0.9090206897576387,0.05212901161861172,0.5104161429262498,
 0.6047360875787612,0.31825997872508716,0.2443178357686865,0.6145249188834507,
 0.6892088894661955,0.7262554244178991,0.9580010665899117,0.6651252233306987,
 0.48974604797150156,0.7752766519285792,0.9434155264877254,0.3567641975253747,
 0.5371721709170194,0.5478691937988401,0.31534279614130967,0.8097809930007372,
 0.25021422081037514,0.44295160110712883,0.38770044369972556,
 0.6757530289326761,0.2607807213369524,0.32173564302936375,0.7704094858566871,
 0.09143313293393529,0.9508902480642742,0.8208129781303042,0.6049220894461074,
 0.3677213923680548,0.12193551520824408,0.2647727163531529,0.6644392067939737,
 0.019423074241438854,0.5421576754445075,0.25348956449126225,
 0.39874374920409883,0.1464377148846483,0.5976333958142791,0.7630729110080452,
 0.427449699423629,0.9797318205759903,0.9721951695975031,0.37796290512485686,
 0.41094561563027476,0.9402568503416324,0.8394181968439941,0.9725793266916796,
 0.47419942165842954,0.9267346674525501,0.46491602407515165,
 0.3295923622000517,0.7408340469365005,0.16985124768635596,0.6089171228340649,
 0.10137185400319393,0.3341845101612425,0.1371956151350483,0.7246104370341346,
 0.647696257766635,0.8777049133910965,0.5378183165887046,0.27514985037337736,
 0.07295591921289979,0.07480602398895175,0.06567696342168017,
 0.48921908263847624,0.7087183134364878,0.5709772411557796,
 0.11783129460009545,0.3276384904337455,0.3474546711512335,
 0.07142290343045155,0.5490989675031205,0.46152089717317124,
 0.6988598424530613,0.04826706168866379,0.45221422219363794,
 0.1160487951177972,0.3592346835972683,0.25380545437347235,0.8805142106003873,
 0.02540850662546934,0.792690527403558,0.19897658928281425,0.430621620601491,
 0.17369546135139047,0.1773175312894676,0.7043907803242359,0.8589558763140943,
 0.08598347736007594,0.2757190725575389,0.9409402345234417,0.3420759877788435,
 0.030955734131493773,0.19150547313111344,0.6119472154401768,
 0.6497970053128617,0.02468962844805156,0.7308803575167351,0.5381021962944338,
 0.7447355898832362,0.5481815506260965,0.6020829932980805,0.16777550538505004,
 0.5596720402949692,0.8610217107310572,0.03387726230412047,
 0.22386079352885402,0.8605179133909574,0.4813563171617674,0.676180550693831,
 0.37679901733777466,0.798744411206259,0.2290554954251256,0.5067791436823442,
 0.6235702545342308,0.163772485621403,0.996270323153064,0.8412528305258378,
 0.27955877830963705,0.21456830638464586,0.20833297315992894,
 0.9195517584035816,0.6653716178162055,0.8007789438025967,0.32084866052572936,
 0.32125029663331994,0.9647730077940265,0.8763850174051228,0.5331652270278475,
 0.6371265297655275,0.5506679298263464,0.13586553570932214,0.8892397014139651,
 0.49221800505243796,0.2718565008090026,0.9083150097309195,0.7548009977490732,
 0.6736787212209663,0.006162316882344587,0.49708939720556067,
 0.714160438372923,6.5010319809655215e-6,0.711039443767735,
 0.03264348059576083,0.5352455471473951,0.4249068261151454,0.457132800953737,
 0.6085601790900621,0.1774893478818691,0.3506110858736742,0.936768974072768,
 0.056933805607016885,0.7307514838780746,0.31394859671240005,
 0.6377855634792543,0.5060677019579447,0.33023892845539726,
 0.49807389565039084,0.3183022392904491,0.6154675371411111,0.75255009423083,
 0.08541419771110048,0.17697640637898837,0.09409592047346571,
 0.471789284428359,0.2906391534285787,0.5624444477904339,0.08711209743752946,
 0.40358987040018846,0.964013604379817,0.31300654033675956,
 0.060883620618037204,0.7400603979507714,0.11219298246375475,
 0.016678885839302438,0.6136426516437303,0.8311151450373147,
 0.7131315691463243,0.3987628374512848,0.1538486242788415,0.43088800253940307,
 0.20367044223668085,0.5203966824456158,0.8486504894091714,0.2839095892495155,
 0.36860413185842344,0.2793198940546723,0.0953004211024463,
 0.08406007371369906,0.49246686040783283,0.7621095847312211,
 0.3249712091793482,0.12263456634886116,0.24441418879579335,
 0.6330171451347212,0.41396760924106013,0.7262372471900187,
 0.009257941914479484,0.05532099621604203,0.17406985131466224,
 0.948345042900912,0.007006340884882056,0.1690119086209092,0.8561652108040202,
 0.8357794302508943,0.42289886960006595,0.26927659007961,0.6286707264718605,
 0.7887533672618402,0.37653064404306613,0.12384791713441523,
 0.5676777411739691,0.7321964339337822,0.33188857150221385,0.1904610792694097,
 0.5075116297457921,0.9599463656862075,0.6124696680252132,0.1172671316304239,
 0.40338569550628167,0.6614064994270418,0.15124190655265635,
 0.4308319916367016,0.7072578127984895,0.3080368192429974,0.8271713685894185,
 0.3584350433965743,0.056053594085146985,0.7732732406464399,
 0.3261090963397696,0.4519557149080945,0.42092079103570823,
 0.13955359627549746,0.8250600725175657,0.24203484969302524,
 0.10582044151001202,0.3764447806338992,0.07878532519634984,
 0.6357730804697691,0.4948647494354069,0.7102041220583465,0.8317442478768677,
 0.9435399456805275,0.2683480297050429,0.23944387122381694,0.6643679575594966,
 0.8424950745041424,0.12930200730038566,0.689759131790954,0.8006086783251979,
 0.3712040617445166,0.04074183242717,0.3590423582594574,0.6251741693957212,
 0.2124260903051549,0.26147818631549324,0.11434633483804935,
 0.002693825263414329,0.42074403569358454,0.32652106788699575,
 0.8601280013599759,0.5592387889879176,0.321130292609915,0.5899264373982368,
 0.7754119007936697,0.7650778971165013,0.5673207309068953,0.18565235531207536,
 0.2967647130778447,0.6532247081701095,0.20696758128305914,
 0.40382430207009856,0.8081686596724185,0.6670039747357823,0.8948019229865711,
 0.14561772334323675,0.6960642023390187,0.7568474974717858,
 0.24408500292552404,0.542419057739043,0.3657849559798214,0.8849185054062254,
 0.8896569677087105,0.05221156281858963,0.022464816679894528,
 0.0916320677480359,0.433099977999135,0.8482480827875937,0.22677994426892278,
 0.43014191371118937,0.5943463548628498,0.8241537411086388,0.7047497618094686,
 0.8890509594824176,0.431173063769397,0.5932321962258787,0.5205611422353321,
 0.9459546675561135,0.8394151027448279,0.9247716044425554,0.5035981930243293,
 0.4465217574070911,0.6959029778972738,0.6141940267082138,0.620571928742843,
 0.9814105454498617,0.31399878293714245,0.6910409607714801,0.9157438513004326,
 0.6090723904496943,0.49848663077372235,0.23483362665111973,
 0.6488224582365347,0.08895458370009846,0.010452958662414247,
 0.7148927647395001,0.19766630639105254,0.8168750040581156,
 0.48327012343755227,0.7033750592764066,0.6008954750889407,0.7782461124807897,
 0.9052017731489357,0.7595936864205988,0.7326429186338241,
 0.005401443172781906,0.2511819245051169,0.568692285335945,
 0.45914648298767535,0.3294248167014544,0.40119406965994175,
 0.6857327142676569,0.8946930865626479,0.12654776030769055,
 0.13011352924066188,0.9402581475118517,0.8068507884100093,0.8843938598772271,
 0.3636072221906851,0.1541977593049202,0.4694724240649334,0.3950009376871435,
 0.9582080036117863,0.05937785791981376,0.5833031540559656,
 0.24632392153502025,0.02235769107813279,0.24442491554843193,
 0.7633350295098584,0.007855254781884868,0.7952795462764664,
 0.8440031659147107,0.5212193790732064,0.050419426507389065,
 0.34124091850268023,0.5934625618657344,0.7205624112637891,0.5074756105175502,
 0.20229239464215565,0.7675774194242628,0.8835810427863415,0.7177242789859823,
 0.11221629255791787,0.5082475718026711,0.6030845883899776,0.8428408015553592,
 0.5297848538513031,0.11100676335840876,0.27759247493772643,
 0.9650942714500886,0.6972359568461413,0.8473064519783247,0.44886692190713096,
 0.3898518245678969,0.21122697175537908,0.4101881761784343,
 0.48997857824980096,0.4872236224145522,0.7681859574913831,
 0.34362179626091083,0.7627482560585006,0.7757285948792509,0.7318663857671026,
 0.6581142579064265,0.2944262090220876,0.01137319938546344,0.8527232285543007,
 0.08673140473985308,0.4041795398236351,0.29532512736723904,
 0.6276983349197918,0.18980187178845664,0.948205778416274,0.12451794068591504,
 0.024432320409475405,0.2881846724769239,0.4848550525414109,
 0.6981370354357014,0.9750222498490322,0.5477005385832618,0.6635155890073563,
 0.39384816677918255,0.36237943429570485,0.9284639252928695,
 0.8145785980924145,0.1801177975413094,0.5623771990440783,0.5816548916848507,
 0.701666259468657,0.3638332723553631,0.4918853943925199,0.7066049287796946,
 0.6601497036362981,0.23489855822256056,0.3561488174051606,0.9542858868616149,
 0.10808168902972515,0.5592744136600216,0.5790048124430179,0.2592338309482348,
 0.10012858081664633,0.8380250600452112,0.4150077426488411,
 0.04243548699576616,0.643444261959098,0.5598343124823851,0.6859865193186323,
 0.8579476609122826,0.08002086432092925,0.4927768950294342,0.9217144900629283,
 0.696994064445582,0.5947158641803708,0.6171580659993591,0.9803613186600048,
 0.05680681202924842,0.8354676476721185,0.5201513418963889,
 0.16107035725002183,0.6607509471322558,0.6018414968840419,0.5647475135168903,
 0.5013707294961275,0.5837092524167121,0.6941492292642422,0.8563311569223834,
 0.6523964858114586,0.1078371622584463,0.19455728949160045,0.2889294942632863,
 0.11441780363425269,0.34556337550540317,0.12695720557408874,
 0.08295137719590984,0.7414058428124676,0.442447795911484,0.26253204205511405,
 0.2183468283912131,0.9470822626051781,0.7196215767821054,
 0.003837913878442345,0.9601396910529083,0.7106656470449398,
 0.7477058499111794,0.28395672952372053,0.6899061391593201,0.5293145163115629,
 0.030753549163168636,0.188147056313123,0.813153386806359,0.45285768458415865,
 0.6994737014713033,0.851044955405878,0.018752535698604578,0.2123675924102787,
 0.14607071455620235,0.5178271008587969,0.7073455014796721,
 0.008251310394715383,0.7636158103477692,0.9839282867284963,
 0.18182120365466037,0.6428598040250395,0.014799288336156735,
 0.22481846635762281,0.5322945098131261,0.010088884218046212,
 0.857183056300911,0.42649653804521237,0.7238801951672197,0.20344718602931466,
 0.8841162542910532,0.8488471292589981,0.8703027007575067,0.2606423253402831,
 0.011966066838042577,0.17067911242071043,0.7911026854623757,
 0.24620243011099785,0.594606820304182,0.29714104999963964,0.1129007566636574,
 0.25867833537099016,0.4265092470445281,0.4372370901929099,
 0.46362447739165136,0.6457246272422267,0.8173849220976984,0.189826439126356,
 0.4396666433083002,0.34278571762854626,0.6187298612840546,
 0.21937586051253688,0.7737894176290692,0.9211003094807589,0.4328353600562186,
 0.40105210361908206,0.8248090273517139,0.8175232730391464,0.5054072078420667,
 0.1421713373955884,0.6842689569695283,0.42590314654925354,0.5458048013473146,
 0.5475192186385578,0.31416968372370446,0.11275661191572317,
 0.6705334996891215,0.5345126075724298,0.34590947177727205,0.4035372356738083,
 0.8547421091886911,0.7541643618578439,0.4905648668612036,0.18606752947382676,
 0.45498937648612603,0.9469942523327023,0.42295043332830107,
 0.19417309239195413,0.20984687821666437,0.5412596516945635,
 0.11116624312853551,0.9986118995937012,0.524843616634597,0.7709959095232093,
 0.7121290894293351,0.39982677940395084,0.24816427807782593,
 0.017194238134308204,0.5525551058665481,0.8526794897472585,0.856904181070267,
 0.23797951199965073,0.9402592045510938,0.3615184176833858,0.9105557601860325,
 0.515257949384975,0.7389716436210325,0.3050284409572247,0.415355051649839,
 0.001909407969674981,0.8192286385433271,0.5506245741328413,
 0.8451683017181854,2.4145956908305788e-4,0.3994448789234948,
 0.816080113880096,0.6448997472791927,0.3666464861639629,0.39555242641589505,
 0.7603007563526244,0.060803720191078936,0.2534902199150014,
 0.2544455383849633,0.654258637379811,0.9387076074393883,0.0680697546318807,
 0.8610320467687448,0.3507160064574303,0.8358476504475904,0.7025202461007216,
 0.794420796919017,0.7657350543373542,0.025959658196594182,
 0.49719815384347066,0.4289646690375569,0.09599474537321018,
 0.3386762501933501,0.8908259143611545,0.8752414815403642,0.8654720182697059,
 0.2898044792914418,0.7677919303746561,0.7429951830339179,0.12124997547597938,
 0.9457917946099117,0.18047652492946553,0.4008308428003575,0.3468635555628192,
 0.5465304799850317,0.24122050205240309,0.6110020964019207,
 0.04445394957854032,0.8315018626195403,0.8103766061288298,
 0.22733445225099835,0.231398081208541,0.8454290888801684,0.7212540728788912,
 0.32202178283350735,0.4930809434082728,0.2950376179617249,0.7192517155661834,
 0.9016767977737026,0.014271577540239555,0.3853829986377779,
 0.12882622693763535,0.0031184580295535724,0.15847742070977067,
 0.8763242757624249,0.2350325183870604,0.8808210128077243,0.7194098983214515,
 0.9999596709032108,0.7978575372212813,0.6948512491645062,0.5203273441444991,
 0.1589409387809747,0.6059986178762883,0.9511841780001661,0.34627704592057296,
 0.16691095781956056,0.3932133775059917,0.3943625300650049,0.8284990902732137,
 0.12528717150133717,0.8132263454730871,0.5775383068404042,0.4416211419856245,
 0.2933291820691286,0.17138861137513373,0.6053315183271217,
 0.06552817847158199,0.7699907003514219,0.20697506748100936,
 0.31533257155062655,0.21559972519619952,0.7196607032630924,
 0.45452031873783105,0.5260934576420176,0.5163664464534161,0.3329036912837383,
 0.07811003314094389,0.3379076617864094,0.7984943750666252,0.9959680737794191,
 0.8198973350270207,0.8769076216984784,0.04571760013407944,
 0.18879439025876443,0.43212106092602853,0.48186022283183383,
 0.06600410264024847,0.43200266583197966,0.2738686439139041,
 0.6964229452326303,0.5497603707480372,0.3341506364483191,0.2067873284012438,
 0.9448865924495319,0.006649251371830256,0.9836269715629257,
 0.8332268998233722,0.7097551142594245,0.38909912946655933,0.9952613923161916,
 0.059040214455718276,0.5968654653855664,0.5785178500712382,
 0.5021090972530847,0.8344688727070732,0.06873111218452388,0.8953583477785535,
 0.17370860443747416,0.37716650236437566,0.1940608966718692,
 0.3752472461638807,0.7388477727284821,0.8728047971236257,0.6987404050219125,
 0.47129853119181975,0.35052388409567725,0.16458991621158947,
 0.9721318708706762,0.6543742134108488,0.004499382677182973,
 0.6505705800282776,0.5935330254528098,0.2282928276007159,0.04852601057545525,
 0.05392213143468194,0.7264142769478116,0.009319327327197202,
 0.1317952657609145,0.5565614180108901,0.7339258366637733,0.8199096121577039,
 0.6052173079230802,0.6944515267856294,0.08314225900379668,
 0.25085235378613235,0.7586467124545115,0.7481941368442878,
 0.42452192463510163,0.7952937537208017,0.9384437319523953,0.8050004264793481,
 0.9052262177608489,0.959105231587654,0.8653754758683295,0.8963342170024428,
 0.9523697102647797,0.539046094604981,0.4432945657678441,0.14117117888346165,
 0.31651628480517535,0.44346440438285006,0.9081250512198255,
 0.19465778625152885,0.5203273539811449,0.6280230047732729,
 0.022429315898765845,0.06278277494314766,0.34515024728161714,
 0.2613825760720274,0.23706047308093325,0.6264783147283582,
 0.24628682737174856,0.9075468521193981,0.9831960862485967,
 0.15031903750325237,0.052054140409528826,0.6741208667510946,
 0.2465772931812107,0.9535498064924642,0.0954992507690604,0.8590956761488491,
 0.13907928255519997,0.18988636252911317,0.7621696714176629,
 0.2721221675264238,0.23560411676822668,0.38147117817306475,
 0.8016305654734155,0.8028954428389053,0.05925154144233313,0.6432206687402309,
 0.3889399439917043,0.15688755249574815,0.9685743137276767,0.873829181016899,
 0.6275461047699749,0.5493317542951435,0.48324114809866403,0.7361498168667937,
 0.04213149319603071,0.5100749705991553,0.15860320530002847,
 0.3141685798078472,0.22453637888639255,0.44120858850270117,
 0.6502885771851745,0.37085182794164206,0.9131374816397819,0.6958153718149647,
 0.6944131602767792,0.8234141041047123,0.44164632317917474,
 0.18902659793928545,0.16682236929497207,0.6033457047953769,
 0.1773245412525155,0.10355352580789823,0.059889578395727616,
 0.4211656638589767,0.5286492228112747,0.13966047356200761,0.8037084771738656,
 0.8396611980297867,0.39795608121604775,0.6409458011588586,
 0.40312265696504423,0.9759261098495039,0.0802155620448104,
 0.07153822175278302,0.1657556057127877,0.45006154114572994,
 0.8126023634925811,0.16253281837689282,0.648939962126132,0.31505857675429905,
 0.6757094994346245,0.9958904244484454,0.4902670381298654,0.07689789985012985,
 0.8529647220337967,0.8442973638660446,0.3767583580547591,0.5089920585457536,
 0.03216658418103491,0.7295098020455574,0.8281463599242816,0.7422323788681122,
 0.7007021141281223,0.5211121235372191,0.6823164676655349,0.45542494067138306,
 0.7600572431466428,0.45174074043268475,0.8004529479909028,0.8807021058408043,
 0.07885330868595597,0.315340522051459,0.031077839108937133,
 0.03829560842135682,0.34888642496595423,0.47416597087701984,
 0.7953479302409726,0.5528761314017121,0.9626296881165932,0.31565194188281187,
 0.45323341329993017,0.18845611990419664,0.11310250956193402,
 0.4090722191883607,0.21928496053624846,0.18927906298348274,
 0.9595811854184589,0.19728959703049365,0.5114163416080538,
 0.32683561243689485,0.9545346799960648,0.5658887514372638,
 0.23667789248990267,0.7750144478959453,0.1673365167867109,0.7019850104349692,
 0.39444982981354015,0.3443936103720995,0.3650117437385936,0.3784004452556913,
 0.8063062881813258,0.7077999535687024,0.5311301224542598,0.7616011774701381,
 0.8103729658730017,0.5341459876812958,0.39754716637112675,0.590025549856705,
 0.8077259575445425,0.219330340932975,0.7116165837986229,0.5805561656089443,
 0.0353783355620193,0.25463116567911226,0.9738673867639872,0.9454261219545691,
 0.8182214220828578,0.798560804492422,0.9863208981445539,0.20137347080854773,
 0.9255081164669812,0.8328447207082736,0.46252608653931837,0.537314443481973,
 0.5996712528687806,0.829783661562411,0.4939908276258589,0.8298927072945947,
 0.3004792592290758,0.6181048594657841,0.7884162932382387,0.09513513963585907,
 0.16221730161510473,0.5313248498361817,0.5257010402289611,
 0.25656148060585304,0.4722213774512046,0.12062363130332887,0.938438265022675,
 0.22265306156864062,0.728682976904935,0.6829280696535951,0.3844371982396467,
 0.4407906704871386,0.4343441827169878,0.5243298949644924,0.8244911320876128,
 0.37065919855552965,0.8226148016548562,0.3637331764477656,
 0.33384253121001506,0.8390390603542701,0.005086141107192965,
 0.5743261780223482,0.6740653122365962,0.026756801226588522,
 0.7088173760034613,0.6170796537465438,0.06834880980447589,
 0.0036776999343521855,0.5564830607368852,0.3910944076675019,
 0.11087937367500478,0.5822646210335205,0.45665661877129426,
 0.39619409167577957,0.6018841436499893,0.8820511764852644,0.7231946428808557,
 0.060841704568948995,0.8012333401186864,0.515842725506884,
 0.35625363425926126,0.8881169999173445,0.1114195753674696,
 0.42232716579500473,0.11297147285360154,0.6456538545729859,0.266608144636556,
 0.7244095499070132,0.16604114730040365,0.002929742781733413,
 0.941863737865496,0.16936392504971265,0.19430688995175838,
 0.018586832238960738,0.4815946198223854,0.10148800887568965,
 0.6135451437208504,0.40271725825716986,0.33087282829031084,0.282750805455251,
 0.900761849994484,0.02466407359009648,0.8212425190058772,0.4337415179502062,
 0.17928933495043398,0.4022894848618521,0.25804950545801975,
 0.8115063581319205,0.37266862531174094,0.6204257604567981,0.8031595701347449,
 0.2099420919463142,0.7611520836056288,0.8146485261789257,0.668853495829173,
 0.7269352873381832,0.8458736227958715,0.48218936906216936,0.7885219576393921,
 0.1191778923122524,0.6494875456955915,0.6223329592087141,0.437032985181531,
 0.3083759356516995,0.6086616381036831,0.03367884732349613,
 0.09614962355784562,0.47705313200428723,0.6477085071989177,
 0.17191602464548061,0.43481225914131816,0.6948738074831129,
 0.1629755334464883,0.7564297842475256,0.15720454711831156,0.5486471609565435,
 0.542464737736326,0.5466540262817676,0.8187239359869194,0.14796642945376892,
 0.046622362357274905,0.9746294761212453,0.8234403162014551,
 0.7849540764711792,0.22989979800427252,0.16745727814907363,
 0.02113027455347405,0.31487231771052937,0.4270112898203323,
 0.6578076852892301,0.19833210002654567,0.5990363931954097,0.5784442330677811,
 0.23515284968126937,0.25007425572475717,0.07880062373840235,
 0.26264821760175383,0.5265505769619474,0.0819487637608366,0.9021510731892781,
 0.5265581806453676,0.16844222278698195,0.07981477055657382,
 0.5188590401975229,0.9326105573868242,0.7097100803332062,0.5910089249981566,
 0.9731196620557365,0.2621731444586278,0.8188335588892806,0.8044127024997646,
 0.6836756201775532,0.5339068904757789,0.008674685898807044,
 0.7023297550028786,0.05698240966777757,0.7956399614680701,
 0.36014337789519846,0.30034033126607174,0.4220850771957405,
 0.46566542456591264,0.9247500556083615,0.9258508425697333,0.6230476823866862,
 0.10605552737412904,0.0641615051337997,0.3741365147517045,0.7368861910366027,
 0.5676468548082549,0.1689958388415046,0.820555120588166,0.2586520748274712,
 0.5755270134136451,0.9153878219660706,0.8214840538348176,0.1621277550363457,
 0.44422934063946384,0.5522779872047815,0.017792188292964806,
 0.7219389613715197,0.7150970570034177,0.22409602073083812,0.2316496372726634,
 0.4705805331310505,0.7476573926720226,0.857956134069792,0.6239262821898622,
 0.1290463662606276,0.6612702604066104,0.03113755180930744,0.9606827401218324,
 0.4152459300121254,0.01146703695762219,0.9655884622672267,0.4953115178556158,
 0.40265506335751844,0.7945545382847388,0.6620753665372489,0.5376367838845934,
 0.9265630814698524,0.8963363933408548,0.9908863572645468,
 0.026159695484131884,0.8384656293685548,0.39529454031656075,
 0.20763969292434237,0.3157355703227922,0.1436959410655062,0.5799364908197037,
 0.2825144855527465,0.5546278374204799,0.029262158182508102,
 0.4970519881866112,0.31813941442334615,0.5552179371469309,
 0.25916596709482365,0.24226484848814067,0.7109771941592089,
 0.23249317191142005,0.8311342764704888,0.11176075383595241,
 0.4777073519879642,0.8836614265989156,0.10282747597828623,
 0.24838895799529503,0.6220343740559757,0.2850976873644625,
 0.006261360848547892,0.8222099378996155,0.18182507190018893,
 0.3316417330718955,0.43893963069356423,0.6812003554659549,
 0.23350687254285463,0.18807458083417716,0.019794528437694914,
 0.4487673378149618,0.7876246624836618,0.1794156275954647,0.9287353674224245,
 0.6713639584170686,0.9271205374901222,0.9444938130540033,0.9604649826083992,
 0.3808179322307691,0.8519390846562319,0.6595522940053007,0.14002536643588592,
 0.3762486121451589,0.9222275854583027,0.642335421550725,0.02065615544239474,
 0.3585034976950644,0.5790695059517452,0.7948283160362999,0.1647001372377569,
 0.9792229340167576,0.5999113078047518,0.5255071991342699,0.9794753246976777,
 0.48339555422674296,0.9612509949915976,0.7684796584833339,0.792601560405916,
 0.9261151909477603,0.6704273242942991,0.19010024487684762,
 0.17485322870828024,0.9230748006587026,0.12207470731818004,
 0.17648162946024604,0.9321362295832609,0.4794826764513068,0.7793758218569306,
 0.8972569490950683,0.843674973001906,0.9171629364807253,0.3994923472296772,
 0.7985973299006368,0.23076375844501873,0.733145432994862,0.3377415686656011,
 0.4018090928293765,0.15451714223862345,0.15804117548000618,
 0.4462204509433746,0.3639488397998948,0.23104565409729905,0.5885867149375809,
 0.12610033790284958,0.08982023041716303,0.529476164056741,0.339630443158242,
 0.15770590131640383,0.9928307069132285,0.08161059571498941,
 0.009124498395909897,0.9141308854745016,0.7378130531618193,
 0.5012063479674955,0.7737321396297695,0.34764522816014987,0.806345810811109,
 0.8817097141622547,0.09193702188803965,0.04602190605343259,
 0.3726936162722634,0.37155834200030236,0.8377319864662568,
 0.22877168757685706,0.9583050367107782,0.8569566756501007,
 0.33997675755348833,0.556610749762423,0.549936090410299,0.9848155191224096,
 0.48745361847877544,0.7550100914600195,0.5983679771611189,
 0.13806377512186263,0.11501966451818657,0.9519947849165886,0.755779925063999,
 0.6428886129077758,0.22858184571385065,0.5736383625234451,0.9953068382359069,
 0.5498934934624221,0.39773550885820397,0.6346165317900229,
 0.34532425431197555,0.687589167037779,0.7225654332660355,0.41398617936960613,
 0.0306067762183988,0.9937543525890548,0.27236940169780977,
 0.25017705059524986,0.40648373724547215,0.6071603226515653,
 0.5494678779099589,0.05970225234665938,0.4389093272267428,0.8315252904056891,
 0.9825089808727739,0.563551519268017,0.7228050680660687,0.29924348571010684,
 0.5815483439539353,0.45710675460465644,0.9328961412671333,0.525649736970404,
 0.7814283246618088,0.6195598044448083,0.9068076019346181,0.9881039735082469,
 0.8008198895523775,0.8424292694429255,0.05882688014089621,0.8959922347290351,
 0.07509286411389782,0.8660069224886716,0.2121880629685915,0.4833276023588723,
 0.678349529569665,0.2877958379197789,0.3801934486439241,0.039051899131369305,
 0.278147797856109,0.15034064703437466,0.4502351501867986,0.8806464358273841,
 0.340431852633727,0.4596573932821808,0.5092610071228572,0.39344810908148287,
 0.06737167395810928,6.088686734682103e-4,0.7780172177507929,
 0.3371763434540854,0.11031071610638243,0.11868436570554264,
 0.2870812834387194,0.23539990748350847,0.434807673258107,0.8468648937121215,
 0.7545347844113395,0.5634582363791742,0.2786873410938654,0.7031709525257025,
 0.8808942358499516,0.9041017029532261,0.8712195274248332,0.43599673258990135,
 0.5962593231320614,0.12080021591349843,0.9981761185967697,
 0.44805971035480596,0.4141660507658911,0.4338250996861788,0.5701036315381247,
 0.6427667025918661,0.1610504924402707,0.283111473070241,0.34227064002659646,
 0.37709267991714857,0.4877444153707392,0.33034480348022943,
 0.7014069547377408,0.036461346212636525,0.22355241507591417,
 0.5137243594732984,0.46758581473595684,0.6778977522598844,0.6788545967328646,
 0.7302353466871724,0.20313427180474886,0.9103146924748612,
 0.08734580284807525,0.30505382003131576,0.7029416657296258,
 0.3831744066384153,0.6624597469253761,0.15167886082883397,
 0.24831327973995654,0.7591450998194448,0.933613197019269,0.46728068993281824,
 0.29927693864305416,0.7192916625656499,0.22230165743064623,
 0.6245791083296391,0.32847139168247164,0.7064729944044867,0.9150743046677451,
 0.637199141222479,0.6884347546873102,0.7999368413044775,0.77506735677307,
 0.30177192387700724,0.05136172074169265,0.636575116718089,0.7809558177761688,
 0.11080598902555039,0.1264966204213508,0.1900387213337399,0.9757925946210702,
 0.6388024800793297,0.6783059688325691,0.833124769275199,0.02958392066599158,
 0.017846911934560006,0.29430334261244384,0.8820798138842628,
 0.25111703267659624,0.07463035018278263,0.7769960152124769,0.676330882624689,
 0.49055017302701565,0.6432435478608519,0.9825926283136833,
 0.35044259144469836,0.34014413679129907,0.49218265702054664,
 0.339735326418933,0.10814180189995048,0.49171895303462987,
 0.16105788562768653,0.04306745918999333,0.05394929090427447,
 0.2499520500176159,0.1353890432392434,0.9042783999133659,0.4452819164982387,
 0.5622853551754154,0.1662745802643949,0.12908024963299491,
 0.03363746183243932,0.9255895507506607,0.9276924857793641,0.6176912344569208,
 0.20922831974771416,0.6031739028497881,0.8477071066429919,0.9934699847837946,
 0.4408037419929487,0.2035697196076044,0.8403113378453974,0.6680863404761225,
 0.5825077648527984,0.6729709659357378,0.31183752057994296,0.5363256273057906,
 0.1987628497881877,0.25404546856449894,0.3677216588526896,
 0.23987310266932227,0.7093377758441701,0.4994888227984229,0.18598840137839,
 0.7270175781244458,0.9903346807794604,0.23306261304329567,
 0.02945558955344263,0.9861938727118115,0.4378222541016813,0.490903606396889,
 0.12408290738658334,0.5409041777666499,0.8453060028710213,0.3323911156619279,
 0.9906734478598284,0.7829061592617973,0.5293133970900361,0.7136165174680602,
 0.09045730888311654,0.28239607446813964,0.4721956510322206,
 0.9013086057142503,0.9884373704717002,0.34366996645336423,0.9597386962555053,
 0.4759482813532717,0.5655070179924427,0.20566784498667978,0.3124328951933253,
 0.059369532169612604,0.7257393547872181,0.8217299148975594,
 0.9030659299113197,0.39831969306603576,0.6620876462674824,
 0.017703234836659587,0.3428424806348642,0.45165705036839165,
 0.17440786389931295,0.8374281003006573,0.5599807723136737,
 0.028984081676433626,0.27141684902888363,0.45157982454873,0.4251593603518101,
 0.0064846062220531575,0.39581496542546624,0.5950241473745952,
 0.9645948295480474,0.2748098786164551,0.36306270395694684,0.9581627309664686,
 0.29934816948991894,0.8850381548571115,0.07365760812185085,
 0.35951124221028385,0.7380194716292632,0.7994872356625637,0.1374682012401568,
 0.6096040499178965,0.12180382756780994,0.07302471851038883,
 0.5291034168193699,0.8075698007453186,0.36563281564516303,0.5345941213458689,
 0.6413795176839933,0.03876987692235234,0.652851043139068,0.8120198917472694,
 0.7030310379847483,0.35271620694526173,0.9197265059563229,0.2370535834150972,
 0.31100003946588384,0.2656718430013253,0.040610987387008524,
 0.06181816897539738,0.49462149115128096,0.9297589961986363,
 0.7135338121776987,0.14248056430981182,0.5965947139651395,
 0.48503614502031933,0.04179293573352405,0.029006149709277618,
 0.7454651108997611,0.994127855931586,0.39807650817395424,0.2975988738096531,
 0.7142134625003034,0.0820588742695838,0.34909228027118033,0.2594182029799674,
 0.33601185217607166,0.18666085746481298,0.5555252668902785,
 0.5813029365898721,0.06570497908160877,0.8889655665584574,
 0.28683111170637765,0.28046087708448597,0.34756931241328304,
 0.7873015869743515,0.5294428084845467,0.6059460588286801,0.12121762688999871,
 0.5982749507749393,0.5950837818563847,0.5744165663300946,0.48053598004575604,
 0.11552516019731374,0.9190422442561124,0.5348420813156818,0.6115251327286051,
 0.44076925179482285,0.6629592398586865,0.9271359640413488,0.9213472384451312,
 0.342135316939855,0.748556445554995,0.38637135945666223,0.19345679158540818,
 0.6404752082992743,0.024404048045687743,0.8840108598367754,
 0.6273691233778407,0.35345027926285966,0.10128603706369754,0.380355178103159,
 0.6297597920053324,0.12866491759813437,0.8585235565070938,0.5628913169357517,
 0.479458578059354,0.22925597749001259,0.07209908195801307,0.9663423306819774,
 0.528919985033267,0.6340469368078262,0.1558293950850903,0.7302935577314091,
 0.12517460561936733,0.7692632258710165,0.949971270960453,0.6895747577021295,
 0.5376555165647706,0.8992201415813713,0.6637992592365162,0.11222259164086446,
 0.3219806192724787,0.37846745781694835,0.9479300643362414,
 0.03179766247949978,0.8212475311733143,0.34741239929819656,
 0.9407382241996298,0.7498234603070225,0.21414526540730194,0.9564160303107653,
 0.7101365535326775,0.9663505670177597,0.002212948750523136,
 0.47052445724063696,0.0012194926810107365,0.3075010388468514,
 0.20449305599830647,0.9950852916974644,0.04285670755677129,
 0.5819993512565275,0.933226869614108,0.5255045858969507,0.7982773737395724,
 0.38387958232937036,0.9707810536000403,0.6557972243749781,
 0.08047650384857796,0.39337791894472807,0.9133632733389252,
 0.3872364397975683,0.10011528308751616,0.8501168668698517,0.2607409415898204,
 0.824129336510896,0.27311382854684696,0.3895411987382217,0.2872596891273992,
 0.05199863775811664,0.008089160150344776,0.6816986734928018,
 0.5795269189025726,0.9456547413128719,0.36683605696116506,0.7693848059262539,
 0.21741634694947765,0.0408859039259204,0.8539588764665712,
 0.14800549052150103,0.9278517695422361,0.10114103931357721,
 0.6659340995542525,0.8832183369200274,0.1320273897990133,0.8340649258342161,
 0.5095122525236927,0.4718399030672614,0.344848927965572,0.07023224471363099,
 0.35569012458626514,0.9765678625643236,0.021399121446561753,
 0.05871574588419359,0.6803017674849403,0.9638161148212081,0.6520730871041863,
 0.2573042273946713,0.30798343479614354,0.08226293156427333,
 0.6509859864768434,0.0014352646903383892,0.08458320183519707,
 0.4977773316073406,0.6413798892664972,0.6822964966553415,0.5413037884515379,
 0.35962811503900205,0.6236673482897295,0.1844942014990213,0.8794853051438609,
 0.22905343189526883,0.18426895618393635,0.2664727911097827,
 0.5592494558082142,0.11135078276144617,0.7448460396066956,
 0.17191083825007158,0.5315108280727333,0.19381593186423585,0.987725488713613,
 0.5924619950435026,0.929073037803054,0.8242308014910146,0.8081848710050322,
 0.5896818894980113,0.85577385837372,0.24440828352611843,0.35832392795065937,
 0.8666244719288763,0.3369925481962902,0.23928170482570787,0.9024417332508797,
 0.13279900318811577,0.9207579579154526,0.9625345703977567,0.3016181445969158,
 0.6660243404664561,0.8409580404173571,0.9947790270931762,0.37784485963917647,
 0.1740400443046244,0.8530716403250115,0.30093235238057536,
 0.18619069545387512,0.7736638481329473,0.588868832984418,0.7192352987426337,
 0.7961411623998222,0.28678796207353596,0.7355019438868904,
 0.07545342965578894,0.8481806716211,0.2229515682589791,0.1426687140788807,
 0.7885935503152596,0.12552877037511012,0.7688740736990407,0.4457198872652659,
 0.38045651768283784,0.8432228561010786,0.9558853854234656,0.8743386922404534,
 0.6313664835063802,0.3506423482512534,0.8257836589134879,0.11870095292873906,
 0.5667700379286217,0.7703304227091825,0.1565081202367128,0.5345010230608067,
 0.22097111025898264,0.9103947360614234,0.7033622359581887,0.5874182588879383,
 0.3311967131680319,0.8585450986591023,0.05195431786131588,0.3932984289265954,
 0.7740694620393083,0.6684417385033676,0.0895163433802807,0.383162440761069,
 0.03606879768655391,0.7483181874216818,0.9884271870377085,0.2103077526472812,
 0.16043102373025064,0.5891732165450062,0.6389961274348617,0.1754685843381576,
 0.35632842752954397,0.294996227176448,0.61621488787466,0.23391879253204606,
 0.37731897400117154,0.029523730268547488,0.9086108813783961,
 0.35576154680557903,0.9728517828683624,0.4557889432830473,0.3196671485182694,
 0.14297210631689794,0.9054616814005568,0.7365225971486563,
 0.33308472904119835,0.4602260701222006,0.6922610274592369,0.7611667452120768,
 0.08473228847853198,0.37974127911236577,0.2570909283661409,
 0.17348338464616564,0.11211616546083425,0.13882846314001995,
 0.1327333423438265,0.22329512639555948,0.7635305799650866,0.9700944909654381,
 0.7261361718344194,0.2321670784481833,0.6399482404129545,0.7736245568235607,
 0.41351841895630725,0.9740544665380186,0.3547749534552197,
 0.04833704992714438,0.9207982507051354,0.7438463364769635,0.7246279740388484,
 0.8125106326740572,0.15577600234082567,0.17328798711076443,
 0.02713442362701124,0.07338350022630935,0.5164601369112477,
 0.9332302173048674,0.823601309996257,0.09679611425171586,0.42062617174327466,
 0.9858058176845915,0.6815861114622686,0.29824960823249136,
 0.49209418954552553,0.131756171126868,0.1708573818575042,0.19681010769441332,
 0.0825397954451097,0.39868374102273474,0.31868365572856827,
 0.9350500133086256,0.05412415536268833,0.06609976078695878,
 0.4681131807359902,0.7591598066130553,0.05928294237742826,0.5157996910532222,
 0.8306648958961848,0.13877956501509126,0.9606848615054774,0.472300196554881,
 0.5816479804907505,0.587145430979525,0.37238974594814933,0.570085640843261,
 0.8842341831762494,0.25228846704587626,0.11982646299821043,
 0.2488234809410943,0.7582090526716544,0.22246144599419282,0.2959157605168542,
 0.16326149173641236,0.4928472925213967,0.43147469738718913,
 0.4757679085631574,0.1512267851568796,0.3888871095870141,0.5229364199292612,
 0.48032935111872754,0.599626911047648,0.333548155046688,0.856611679483295,
 0.752617693550464,0.26159011140408506,0.6681149672332465,0.43850412978954123,
 0.6424923679686834,0.7597391187823765,0.42685203122150517,0.8446578673732849,
 0.09282071955461058,0.5327264561393976,0.6147759885967885,
 0.20644677755087137,0.3637305632450786,0.605615233055951,0.47691288884250915,
 0.572665483774444,0.06590372400658207,0.4038124816953672,0.7121044455241633,
 0.2296534942956766,0.7813639642356486,0.5780411108398453,0.3655665010608323,
 0.6068184655675339,0.9892908828663829,0.07017560134165779,0.8064566322046711,
 0.15563379619249362,0.2913184315015086,0.44876616520194756,
 0.7672061425729167,0.41812014504991457,0.9868278980837812,
 0.08516959355711362,0.8960388205029143,0.4513144038431681,
 0.028551958586055193,0.7177748338650743,0.09163806299284882,
 0.6968732253756853,0.7076628452465291,0.7493714074513875,0.6574072754636247,
 0.6569315422308065,0.5637825518991434,0.10683769770901197,0.8543453883240016,
 0.6153841790326731,0.001997553566256083,0.6752709613929798,
 0.4975070740311631,0.33528887906274074,0.282135757533885,0.5773798482159898,
 0.9763402819640707,0.3677843620433484,0.9106629143410391,0.8476462608365195,
 0.6914860242253964,0.29857254566062486,0.8284507783077749,
 0.40455494014035187,0.6256076448685957,0.7399462891728444,0.5200279447579337,
 0.5661990047077197,0.11102909215429568,0.09378411954329513,
 0.3193835832516898,0.46696675954442735,0.18991332220682122,
 0.6806959244406812,0.12940963627454183,0.6434944418043491,0.6277938962051812,
 0.6891135326758868,0.7638021395912484,0.36908510479071366,0.9995739207596933,
 0.5439062533896836,0.5760558595682177,0.30448819531261606,0.1594601046345565,
 0.8009748427425928,0.8438034750231259,0.9439887480092816,0.3626076019783233,
 0.2575900483446715,0.908587244221552,0.31409361647487344,0.8907802121584524,
 0.7051774261440306,0.08078794883489726,0.9104289386800519,0.7838799450823251,
 0.6214234728333772,0.36056578571678777,0.9101062196582398,0.2545560803871645,
 0.3276778829709849,0.16891950055108895,0.7694806406938919,0.8518083594858746,
 0.5015187447995145,0.5487033258973788,0.04291279057248887,0.6677101884615346,
 0.023281001484919295,0.7153974565567318,0.5418477355702228,
 0.6152518630562422,0.6523462550570294,0.8511699142241848,0.3104101159763938,
 0.5251668163047205,0.5918762376062259,0.4597557323874729,0.8040170867243273,
 0.5822863598940378,0.9584871384211017,0.8061153713607447,0.17464011575928629,
 0.7955993745927041,0.1197020740910113,0.8906844890576111,0.8657495862271161,
 0.9604238686875071,0.10342859043086827,0.4497052985016754,0.8419893005565184,
 0.8455694392139528,0.4036849416728343,0.13173184793289672,0.4709247145597091,
 0.23146271725453615,0.23087285021132997,0.2756180382050011,
 0.9420961641257691,0.25375598739733507,0.004169882156955462,
 0.8463842086570174,0.8755222826030602,0.7414729547192593,0.43107926219785364,
 0.8636897372514044,0.7963175946484442,0.260247190536233,0.26347774601330065,
 0.4350211844053634,0.29330807563021044,0.45515240304868665,
 0.5975316347410627,0.26899099336450694,0.6860044864565423,
 0.059839780609838855,0.21734799322226772,0.39087818247276085,
 0.24814672952095818,0.71109233892368,0.2624235238994814,0.9338222681602296,
 0.9436105157858786,0.19743919922091768,0.706491403249299,0.1865510850362223,
 0.15923848125196471,0.39349374019767114,0.47678195357453523,
 0.02284991418328608,0.6585044013498477,0.026762396211344353,
 0.10469487532125232,0.2828703705975888,0.6223195637464173,0.294459271955824,
 0.6856586960057173,0.6808174194496718,0.08746959380394037,0.4390516144281579,
 0.637738605566481,0.8271684834812197,0.73498447531833,0.8485372737962892,
 0.5996034752788262,0.7498080496986891,0.22048146843908456,0.1478027614022278,
 0.839952405199048,0.42501872985562805,0.6303362315682174,0.4516571809596249,
 0.9342390571488532,0.6392032174463586,0.22901758264529237,
 0.029827664715270208,0.9124258384095004,0.7143321672596459,
 0.8975345779832615,0.2320892349255812,0.5220195095353883,0.7668096404241691,
 0.2588585644690693,0.5169105240492554,0.13955786030421657,0.506470028723457,
 0.12742035409264774,0.17367891327909157,0.8009723875165117,
 0.5070816481140072,0.5852596766788791,0.6072263260075776,0.7378378563202478,
 0.9584335424149627,0.7933576076167734,0.31092416997191374,
 0.07231779574337271,0.08330453519691317,0.6247405257162985,
 0.5427527424118704,0.2526607612098779,0.3583390887852348,0.5293074603075527,
 0.35824128932828425,0.07739350624313657,0.7851159644742952,0.977107610050552,
 0.4893335099625271,0.37196722545120087,0.26636871103854187,
 0.08956331385839722,0.30399406699994447,0.353323763498149,0.7317425980239741,
 0.6260645309646019,0.5171684821013202,0.4564928624493335,0.14692741743906657,
 0.28867102343820206,0.891772441193547,0.487553551566705,0.7125027216151969,
 0.1490049493587795,0.9611356365950361,0.5573771730343123,0.38285762506788745,
 0.4324015041299023,0.4020622428740558,0.3724608167860928,0.1966018471426645,
 0.9526020312555669,0.6432884087878848,0.7014178440145586,0.20149237383295238,
 0.9585703506349025,0.3614758007958244,0.3083011698048722,0.5561884523458543,
 0.5064868616130048,0.5877368258823686,0.6141450920331136,0.98208155343925,
 0.09359790589916384,0.6298253948778455,0.265954665655165,0.34105178261963187,
 0.3570210893480794,0.3664174402670164,0.602329047870312,0.7505380565376542,
 0.5105813763444601,0.9536865488605982,0.4272947271920602,0.28984936917566695,
 0.5147909400257802,0.3146128143087016,0.5826522076256909,0.5747639962396841,
 0.5055818485225105,0.37639150568317936,0.2997236617005785,0.951235210511657,
 0.6476270501680972,0.16310159558489623,0.16054991407310792,
 0.47752937129939765,0.28704632578873657,0.04898825605479207,
 0.5809183469696434,0.16809437074284572,0.8237931144965367,
 0.24591010664691604,0.010893300299610553,0.27265621950798935,
 0.7480787894170771,0.8643264044612433,0.9480071904261728,0.32432336735301215,
 0.9022996429258234,0.9051888887696954,0.2196386945971982,0.6815896142262238,
 0.8013516112625663,0.11996367555033005,0.8229755467187636,
 0.04663657469095395,0.5987566684888759,0.04122993769480521,
 0.04649100819189511,0.521721290602263,0.7740905613963744,0.2759746664447683,
 0.8746891877710057,0.7126663069380055,0.5697022863548797,0.911871140921797,
 0.106817662566422,0.48299186340015654,0.45718991261382,0.14166338069007467,
 0.12475924795980076,0.22255874826002908,0.27324152868247154,
 0.6276254129748197,0.015283737270741726,0.5192657043237954,0.959493747505466,
 0.10086187397821744,0.9761940622086721,0.592357686447018,0.4137056626303368,
 0.40742059168857714,0.7541671218927575,0.13837850822102227,
 0.4040902032295214,0.7003189266027797,0.23006048796347056,
 0.31091991230764326,0.7508704793132207,0.318979641541038,0.7944262050300005,
 0.9844474484826442,0.8113452670667161,0.022840175457235934,
 0.09936130662171472,0.8877793856863231,0.6741903756395996,0.9340289646526436,
 0.8929912870946517,0.9673723133368297,0.23736459510640906,0.8083772695507457,
 0.468177692755685,0.6188750511038543,0.9428867945830258,0.9457177853795573,
 0.274897010824777,0.5821883076165068,0.3319826330454776,0.03467846911474359,
 0.05302693334082742,0.5717575797888009,0.596240314904189,0.8240971917042169,
 0.5678222765018275,0.2677663751882742,0.3977741765250379,0.6248828506365884,
 0.2840075831768043,0.3795517937647901,0.7855473454720387,0.11606566094548876,
 0.6070724487713118,0.2292967837260912,0.7965450567713702,0.9081077276546937,
 0.21869375397741975,0.715348217253174,0.3739626986716862,0.3663144754857327,
 0.43093377277594214,0.2844626054328222,0.5065944966455953,
 0.31347637703741893,0.5438219728302575,0.04688936942716715,
 0.8123344928165324,0.5596992332700285,0.07504850740690838,
 0.39597268587126166,0.5698334939854484,0.010433798145909279,
 0.7115416750040132,0.5732704245986112,0.3808402494165428,0.9803214210942796,
 0.7174510763455773,0.0394347050173014,0.985057014203525,0.5877398349556845,
 0.4298698801498786,0.1982295568609157,0.7923163085487297,0.333698383537511,
 0.7575167709850498,0.2608445576767807,0.9877191467754248,0.5232461051715831,
 0.2502486887769748,0.001952975345562158,0.13531777088863384,
 0.09187007182858142,0.454753754966957,0.5509307530625739,0.24228821907544484,
 0.479399516086335,0.07381615208670478,0.12250048042621653,0.9176783423531344,
 0.09251651996822718,0.16352223364729768,0.43518606928854964,
 0.9404272002351204,0.629009213699078,0.70115177651455,0.13502370224221028,
 0.4925301107620582,0.9716131610169675,0.37243309003005454,
 0.38583281656358837,0.8222182958735248,0.042067198155342544,
 0.7625892197482103,0.8500434362509383,0.19841942414811198,0.3736394071602913,
 0.07283022306970721,0.5190849848142375,0.2159539840080047,0.5503105825477914,
 0.31937700343232556,0.27569692020664094,0.4605035589058334,
 0.7515462166063511,0.4347597037479155,0.6560121369656574,0.5048039749351934,
 0.8300845688959484,0.48419695614422664,0.6163280076873086,
 0.09351945488206603,0.9174118411615726,0.6664280188076509,
 0.038253515781963565,0.1344773827975123,0.18002733333731746,
 0.06517641571646482,0.9143360459428747,0.46905349468950763,
 0.8855031063053213,0.4513354081757386,0.16697736825071285,0.9622169379669476,
 0.20904652636608378,0.5005658137457447,0.5457182366216341,0.9795388248781317,
 0.031604878497649036,0.48212033203476035,0.5065388188160997,
 0.01907661834121832,0.7011845696992854,0.5068664853039082,
 0.08167356683830829,0.09394236418100899,0.57856186315723,0.4700003806990125,
 0.9719930685722338,0.6059676584566253,0.03985445447072988,0.8579470593495039,
 0.3296769472445583,0.6475233262304478,0.22966082067830496,0.7831409474323635,
 0.9892161655572272,0.9692604139050573,0.3084921974803917,0.8584261426201742,
 0.025020447623101738,0.7963210295059556,0.7747530643539545,0.677237103223864,
 0.5183228247817798,0.6531956787394486,0.28908516645798,0.8654738390617776,
 0.9239673963554064,0.4254022680469972,0.3257198789790803,0.9195111100122418,
 0.610742818453665,0.7144843480628978,0.6904022654561102,0.7397042635975203,
 0.8670401543112516,0.1307937020262968,0.4159749288094686,0.26344031720843586,
 0.527787309172381,0.5574427524142292,0.6423794146461758,0.5662890226212742,
 0.17342175813565397,0.6316733293404986,0.11709391078623166,
 0.6031776265391938,0.808359787917427,0.4776178092314689,0.8903918091455914,
 0.496142712407453,0.7394107057459591,0.09269988354195824,0.2000710565728946,
 0.3392007847664511,0.25330032133700686,0.6715740916760603,
 0.19047229490724582,0.6418606000484461,0.8527678459513729,
 0.38400308856373644,0.2969814646733482,0.9209110837332029,
 0.37278022427250124,0.6900711360266256,0.5734555167718909,
 0.20637224884465877,0.18320433440925,0.9932275509921975,0.925062700301878,
 0.5385198024660828,0.06544760381683434,0.9636385432290725,
 0.19116981396963828,0.6792290996248667,0.7824017893648473,0.6188361732819576,
 0.3044422035805666,0.463964969225904,0.6083461335069413,0.956750179075262,
 0.18813205819224088,0.5443102177473055,0.4603589506782356,0.9956422016971265,
 0.2942940057375901,0.8533688580354959,0.8892564868316959,0.5870111812038626,
 0.2564848273064916,0.26970394801859654,0.4065312558413482,0.6389097535384487,
 0.5970727113668297,0.728137791133789,0.7815504399854987,0.5557737913606892,
 0.9613779112943166,0.0602262836549472,0.13452353857066668,
 0.16877365160826174,0.8996909540956094,0.6956623930305453,
 0.19983338986890942,0.3573760300745382,0.8128955370057723,0.7286365741428026,
 0.5336965911711434,0.8021193143352523,0.6056642020606319,0.925049985276365,
 0.037928842049724354,0.20482875862699423,0.8569760047398925,
 0.5106737883305946,0.4129876318716952,0.07340353269263467,0.503564512351555,
 0.8732215807279982,0.7217334618361836,0.06623288645018688,0.9347487599178922,
 0.4151109959558291,0.6479103648896076,0.9917993689933083,0.3955271438101723,
 0.12876427991426498,0.13820308905531586,0.055306263954053914,
 0.7522123736040749,0.2945911101978189,0.6509221359374285,0.7122944059949332,
 0.3880316801071737,0.5544456625366575,0.6476625966619896,0.254522910490109,
 0.42752522111231495,0.09960888620371078,0.633456116173942,
 0.039815411956941604,0.9127618226827691,0.433048571341454,0.4794183751523149,
 0.9092625426447316,0.19113244210123026,0.9319681059731487,0.4661928515120508,
 0.35541785221859157,0.17200035748505538,0.539049694500272,0.2204189365544461,
 0.22183422397976382,0.3338934638861639,0.7959158832121934,0.9657344985797646,
 0.26595626827907937,0.12157934314955632,0.933240246108226,
 0.13316804563882045,0.2513942472938736,0.9991090756674533,
 0.038467443970683024,0.7979922911342466,0.3515744106183303,
 0.12722189568494224,0.6578122500431578,0.20305432825601244,
 0.01622602954750696,0.4621261279843206,0.36335047743853743,
 0.8805460696119654,0.9589896016731788,0.5276688959884706,0.7087541669069705,
 0.3394792025584119,0.6653050520201457,0.39280962477227965,0.2854416100976469,
 0.7680966229808688,0.39317598514045937,0.2375214710325898,0.4761581557555441,
 0.7636055131566675,0.39367173267521494,0.8560488981205541,0.9617137128535006,
 0.048725461102996914,0.12291117924623673,0.8574777858009961,
 0.956991300722087,0.1786508932255012,0.49870425786127537,0.0817117868521482,
 0.9918459356740158,0.7497715976288587,0.4583279091827904,0.3309228035358911,
 0.10485182625595413,0.08548441763942793,0.42122018092671965,
 0.3461023936401353,0.34022366088356937,0.5707385441394676,0.4420788465818801,
 0.2571631425097216,0.4980770652399684,0.8592182436005714,0.9850647894072295,
 0.03341781158952739,0.7859931416056266,0.21301786357845398,
 0.3330386191241099,0.22420771597280964,0.611447367239761,0.9373531076275132,
 0.0095555113938256,0.554286068912066,0.29360113034497615,0.7773966703553387,
 0.5841987825639392,0.9514824842280948,0.6537615517641281,0.15454025314703934,
 0.6421311759466984,0.17819834414782976,0.783585759605083,0.0994463227890805,
 0.7607257243820665,0.5079824541635678,0.06890823026288873,
 0.20998037412392456,0.7688551799868669,0.8286210818501971,0.5112246185480884,
 0.5143718694720989,0.8188267371799633,0.5944105005438735,0.01700600317110379,
 0.0444228769800834,0.6886011770010549,0.6083776075441881,0.5208918564397259,
 0.9581047345427569,0.03216632215737836,0.4989794926529384,0.9517947336425249,
 0.3466393863878311,0.6318898467618909,0.24082327246438673,0.4150181970148267,
 0.8858814327220501,0.17760834031400174,0.9336178307854172,0.9730200533178864,
 0.8624980959151328,0.9069320591667611,0.3556367397771243,0.2704819845987051,
 0.34949215980856063,0.693622792268791,0.6232763831965515,0.6370070804138639,
 0.48431514385203567,0.7948839648572622,0.47253282310199773,
 0.7674302360704055,0.12989751505104485,0.9124674140425219,0.5688801580905551,
 0.19154673015030355,0.09548010217898528,0.19336905190222997,
 0.3111019534330409,0.9273579586846942,0.031028940131445082,
 0.5012845604756486,0.2725683330926464,0.472572898734696,0.7989520643414698,
 0.8219497964765824,0.5014949579793049,0.37539126721417515,0.559746644382265,
 0.10062243790065606,0.03967781303463158,0.3417307589022257,
 0.10534532355832471,0.8988805949319134,0.23831758957019333,
 0.20411023836458397,0.40845872550358153,0.735947025644243,0.7659123198432829,
 0.010616130610313856,0.6808712589609787,0.5590428549059618,
 0.5309961854959819,0.7296404740558402,0.9206230834970155,0.11558609643122952,
 0.7577556245530574,0.9850132696295256,0.6341291178165566,0.12543178283657563,
 0.5789783271440062,0.06478604151288636,0.41128597003851364,0.622897930554061,
 0.6846172403325851,0.7260225936420133,0.28933723540821643,0.5392225035102991,
 0.524963850673942,0.9638699430849961,0.8402066372401608,0.6019515050257163,
 0.6957278106452861,0.4480570636079626,0.2779103633174127,0.9770969431747218,
 0.16708081970462807,0.6088792159439799,0.022375924315331863,
 0.7474725899010084,0.00416150283714023,0.2172771924157355,0.6646125082742139,
 0.49685706328311363,0.04073730112582641,0.37306854494783104,
 0.6195416047564715,0.8108799863041063,0.8872287556304703,0.4677864377167027,
 0.5111114749429556,0.09308198500978515,0.7447857165638433,
 0.31559957547036355,0.5723076494926083,0.23355634345649834,
 0.8869007802364429,0.8103653615138876,0.24350267971360784,0.7887869846085742,
 0.025997334594978883,0.9956554670086698,0.8309749933654849,
 0.3030834178457854,0.6088701764611949,0.12300868060322823,
 0.07601431996024455,0.2053928068594516,0.9515413473548946,0.3565009699869308,
 0.6641719991023085,0.2297562994718063,0.931408620808225,0.034746097714530855,
 0.6967239379182553,0.6364787941119481,0.4430660932422261,0.5407249239450806,
 0.37781338138494514,0.7616729208720886,0.6028125440595711,0.7313052379166473,
 0.11639169838872054,0.9542848281009603,0.48718561240128344,
 0.2783170996178611,0.20988190314908672,0.6983870191550712,0.3017953922358836,
 0.34019704706641785,0.5733734242900163,0.8048999133381437,
 0.29172603917385287,0.9650571660853312,0.5429867964208408,0.6146069566616065,
 0.5218744467727889,0.6488294576948077,0.3074131356751024,0.5596769996379614,
 0.1234269546295309,0.8585920178456536,0.37768791591767414,
 0.33407098851271844,0.18952936572813817,0.9007315057226118,
 0.9864933274566233,0.5417284947606305,0.8024075082507279,0.7018436478814704,
 0.8560686460804314,0.6587184365993168,0.5468385736600104,0.9495684935826416,
 0.23871655010004222,0.6224185497068689,0.8002863394005921,0.5076566122064126,
 0.517242289641799,0.3265259143620757,0.6708250745155906,0.993101223806909,
 0.09401326325131953,0.92433665177788,0.6815298575721227,0.31350376505804844,
 0.479339108535241,0.8613433543640469,0.6268247784181117,0.9929609893757531,
 0.41343459692435913,0.5938172995735104,0.06436102505701946,
 0.30595003317378466,0.9423844183296468,0.288572234872733,
 0.027296302802130912,0.3608465455573021,0.9885296023878619,
 0.28222500084710256,0.7206168942852187,0.7826370471397972,
 0.15103374433831984,0.5116711081184326,0.1875602314739645,
 0.25070460153956875,0.3646400626831574,0.8895152922857179,0.8951653467654185,
 0.9195682154253412,0.052808015874495906,0.8108051224981121,
 0.8280580586924358,0.18850237822097093,0.7820469534538598,
 0.47162337692714695,0.7950386277175584,0.7248917104672586,0.290198659531533,
 0.8870506886900212,0.8575669878700327,0.9853678586235187,0.8842119212782847,
 0.039938629097253986,0.3978555576887297,0.08893893903182959,
 0.5163073288292075,0.4100865140768115,0.4324986552083767,0.09744032278114279,
 0.715034443774866,0.812493705892934,0.17708997812986205,0.12537369403797316,
 0.6875740900794909,0.8386792205966709,0.6808767492135388,0.377576189363546,
 0.7475520537610852,0.07251881234474178,0.10473572045296375,
 0.6931631445402763,0.14960091046623558,0.16250727563203238,
 0.14200679928921223,0.9674135592065662,0.03940361223954747,
 0.5144696493664875,0.29407445782581243,0.9495912882975122,0.8888341421145345,
 0.004367156998102306,0.6180523371817493,0.2829157221145039,
 0.37633038908887617,0.8914130852039543,0.8278755022242853,0.8131475483312686,
 0.3617005833794962,0.7761316673562966,0.4452838878657992,0.6528330144957849,
 0.40778961823944015,0.8185077869557684,0.7169565435016929,0.6618729742480323,
 0.009097489074409992,0.02985653000645161,0.3371807907536545,
 0.8721581878122193,0.09103400585788601,0.6293445215183275,
 0.46958344520314943,0.500382329984209,0.8482426756141425,0.5474252590014683,
 0.8792266957056498,0.7033298874142473,0.21752291642071875,0.5582592632490412,
 0.4650767920336183,0.29603065359689473,0.20993752714936798,
 0.5552072486834215,0.10271586018688827,0.11115560510744582,
 0.11417912722808754,0.9248812431672173,0.8722092559913974,
 0.17424666347121243,0.5305287553505517,0.40096624736637687,
 0.3073999961101699,0.10618327951900586,0.4349899648500136,
 0.45853580099385216,0.5638752941942031,0.9142481788725593,0.94664044440246,
 0.6860882477951512,0.47720911140115774,0.8652097431346186,0.9005934816347342,
 0.4173698791128495,0.16078951152321308,0.861325165257431,0.8952942151992077,
 0.7069355017479323,0.6342704245066009,0.5267186995458195,0.6580337128682227,
 0.17126940302982507,0.6383835874212118,0.3583420122601684,
 0.22502138581678777,0.6326609507767417,0.9435011255832302,0.8217240412906872,
 0.8139753705448391,0.5826903487246264,0.17790426034900653,0.8797535745522501,
 0.3334649938870218,0.14801927690100813,0.2967786741417602,
 0.46607072174671516,0.1604996983081025,0.7956716204980938,0.665336585772443,
 0.8812387188191899,0.6811989141422664,0.11305761775474166,0.4203263196596523,
 0.29008155470053953,0.3259389327949006,0.10969123994028707,
 0.48625283127799657,0.28634336965960794,0.02105604759051738,
 0.4430624423302256,0.7766166379626029,0.8787936663318318,0.16526040814547038,
 0.892083733814456,0.490561609000445,0.5636229074123155,0.0017726977932456256,
 0.5823624120792902,0.3568167873431901,0.7776127624141739,0.08576667360861434,
 0.7675159308310332,0.44353097889840376,0.09314248970720518,
 0.6294266400399787,0.6547819780512605,0.027167916860970642,
 0.4649164325563161,0.02358398043577281,0.1854066408755869,0.7588658233341256,
 0.9587904249331074,0.5107439049134183,0.18698676911572232,
 0.008082806168396961,0.32881196211666064,0.11719520404570805,
 0.7205159684046991,0.13701055767651615,0.9266859344452373,0.7027266623018107,
 0.13333519030382057,0.1903315153460584,0.18977091936582402,
 0.8660987248031495,0.8319735488063073,0.06885365587115766,0.980057372168902,
 0.10782347366331035,0.3232066896718787,0.2825834004029252,0.751193144887949,
 0.31580200202900055,0.14144156422869003,0.5446435113489078,
 0.2627398689913819,0.39035984787658073,0.053674274473157535,
 0.6575005665796476,0.23303991919287115,0.9738307668905194,
 0.32769357644567754,0.2004414243092505,0.20136353912311156,
 0.7601595734319421,0.5330897077150869,0.9440465487255283,0.8193854195597167,
 0.7445019641960224,0.936296836000895,0.12710307254774222,0.6851089697893612,
 0.2990257198369809,0.21270724974475774,0.11655286128761944,
 0.7858009475507405,0.21490405413811775,0.03386153440321904,
 0.36304968225942136,0.3556403983617965,0.39123636224607483,
 0.8892379361948292,0.11645072096896447,0.2181815311290447,0.5497482988838988,
 0.9169654338538941,0.7545262483713538,0.47097949605371825,0.5576027109697927,
 0.6376730008108378,0.5748886956223832,0.1924047755170214,0.8719849709950138,
 0.9140139428668401,0.2944444641803947,0.4110765119202737,0.01754743551464044,
 0.2711088145206404,0.8306871029257132,0.5568329488364524,0.18060610833512325,
 0.4705561596045468,0.7194778478158889,0.801245007110702,0.6336701442270186,
 0.7697938075465172,0.8174888263827326,0.4310001336104857,0.17886072340206283,
 0.20633880604900834,0.9616192557064139,0.5411411599401911,0.9446478202379578,
 0.5565529141885286,0.31611542364838874,0.8552537437118055,0.4719903897553288,
 0.779688401626503,0.8765062128862567,0.9812370626287656,0.7707353914345929,
 0.49266806749533065,0.48265733499186814,0.5264585477549637,
 0.5247492048143383,0.6382004235522953,0.713357421391918,0.26972592593656775,
 0.5008358686553607,0.8993991680741548,0.07768621065997683,0.2585980233523799,
 0.9787565358614742,0.4821493795449482,0.21223175833937347,
 0.37509118562306887,0.6847241645969798,0.5031449334468623,0.8361051019574883,
 0.11297267142591283,0.823525616826764,0.2570937040246378,0.43380184792648324,
 0.6660305310178006,0.1724043618641291,0.7520123666337447,0.5203804579109275,
 0.8972939329479549,0.10028872936002531,0.6706246687676471,0.3860669919060463,
 0.3373981853242567,0.031668243234583615,0.3707642164558528,
 0.4282863744951837,0.5756372522632257,0.9585499138047114,0.04754749253904733,
 0.84779082554111,0.5381883140870143,0.8240673867943705,0.18845452544567276,
 0.7881803662081674,0.7702182544455136,0.8023480061968905,0.9887658199877454,
 0.3229365337860899,0.7870379510753606,0.8173466747548052,0.7648025385898959,
 0.3396867776492565,0.3240481327647864,0.07918998569546842,0.2285242114887781,
 0.736856840435418,0.7427189434169841,0.021522843278138692,0.9483570748051144,
 0.5987252359585099,0.8569758509928999,0.6934850750968681,0.6078110825353117,
 0.9604457761618678,0.906781608275774,0.12713408809785876,0.7415696647821594,
 0.6237617111161705,0.8729674856411547,0.37323262329301843,0.8759337280438588,
 0.7837147635016976,0.031359131957145125,0.564702897025309,0.6473639251508505,
 0.6470512423101965,0.9640709276341117,0.23238163115746402,0.8511387182010528,
 0.09286758616324764,0.19016940644242153,0.331102769836331,0.4624704418248703,
 0.2425506621745679,0.5095894448974437,0.2520519645387225,0.30910607794931977,
 0.2168614350727125,0.1927629958957453,0.31159100795177297,
 0.28172434523918954,0.6316043454920659,0.5585507055306073,0.7847132445367699,
 0.31497136190740216,0.7482082844827638,0.6457489130551765,0.4916449371270736,
 0.8338423910217572,0.9043672494841586,0.15194105207179798,0.8426694489179062,
 0.394510918254281,0.7266976073216598,0.8471012115508588,0.7809504790176841,
 0.11966136513627346,0.5558925289319517,0.7961556995449552,0.1896805407607034,
 0.7171568105012501,0.4014960767244854,0.013536237167728227,
 0.5245153961836603,0.34451759474178345,0.7537333704460892,0.9537161460529225,
 0.21543735288833743,0.0310989648931127,0.5978568465786964,0.5337254793273221,
 0.8611207259444367,0.9157882035718437,0.967304823546296,0.7246877994103762,
 0.44640860277658645,0.8245142139001808,0.9243676967020049,0.6474064211084579,
 0.15034626646601645,0.05488906865124832,0.2388910845104606,
 0.12108079924566328,0.2817202690053191,0.9310123321718952,0.3889905525973105,
 0.7680853975317684,0.5824779582226011,0.0889391009975723,0.5984828190816254,
 0.597883866986804,0.8474700173847819,0.7949383036684025,0.10089838528738215,
 0.8628280583845558,0.44206374472489474,0.7425797617156668,0.84063787648253,
 0.23028235028836108,0.11228622121777732,0.06314207603573496,
 0.9441604282393005,0.3410367789911387,0.719007904241439,0.6829521743003173,
 0.8964692334086448,0.7590099467573878,0.39056540944770834,0.7021340720154007,
 0.23118538219327156,0.9667729681479144,0.35397824013771895,0.571842081467944,
 0.2672337315757082,0.6448915763671399,0.6077279673355772,0.9482200638641084,
 0.001187693468120643,0.0396404967581101,0.8600901429227126,
 0.04987854214335019,0.9274893251855655,0.4178195600951192,0.707404686545667,
 0.857545125603496,0.40809089602543225,0.5592368231821623,0.29277657350330877,
 0.30308951720784594,0.32881909631057404,0.5786045762926468,
 0.8520873337781468,0.7931518209220936,0.8389843942657683,0.7295716679686691,
 0.023002684148533015,0.24376313873190736,0.5603258068779452,
 0.9501075742923845,0.7964101552566721,0.3607337292971112,
 0.007728068128531995,0.35671192175052013,0.7340815179071334,
 0.41331370950115465,0.42322466387489677,0.8147925331380148,
 0.009266602985729344,0.666138953666718,0.3916276900541915,0.6701894729414264,
 0.12576700122169404,0.44641674220356353,0.1150874803029398,
 0.7423390250770437,0.30776259000773676,0.21800846122501294,
 0.7133673472223379,0.08931995768439538,0.23330258154671168,
 0.42561609588401694,0.3068349908235133,0.11391911038489855,
 0.36150834737957094,0.8284012769746575,0.6916217603800716,0.6829125464680117,
 0.7858406965625635,0.10439237273247692,0.0026364725142967727,
 0.36904517181425134,0.6239007562843577,0.05149538266028486,
 0.3257163221165912,0.24733298562836692,0.7108334684053854,0.9098068873121604,
 0.5145386612502427,0.5002356115197426,0.7444978659350439,0.7662479475301514,
 0.2698850509199955,0.508089046398694,0.5457744200362963,0.07671164798418406,
 0.8058385934883884,0.007190823722089679,0.45335785820133756,
 0.8057283588009309,0.016813400277757018,0.4541367513494191,
 0.8109364683936515,0.5031419009497895,0.21499192204241602,0.602154490595004,
 0.13941559222215716,0.6580239281158019,0.6834588512351625,
 0.33898934072131925,0.7769011254841068,0.046211714041680785,
 0.9434064999376375,0.8903531974212566,0.4708366269464118,0.2043592773257238,
 0.8193320926346762,0.9517913292479421,0.6558245410267289,0.6848027512788634,
 0.5827309780102945,0.2394657199892386,0.4938598680112677,0.6042346253870257,
 0.45651683175417057,0.30149653373943175,0.07451710236119258,
 0.43305472207988716,0.011406002798332482,0.31417200694876835,
 0.09856010862803788,0.44250827048901065,0.3285715406187981,0.615755826449152,
 0.8000066560323494,0.3040222787671699,0.5891289888258737,0.1015860083256448,
 0.4669505697926012,0.23936260276979493,0.439881038937181,0.15725478789985037,
 0.4002984655626487,0.4863901751332259,0.4539161520149555,0.5975278779401004,
 0.8677807415514196,0.6480586347324612,0.36015138121374213,0.6070674268293401,
 0.5579500458265054,0.24770833302667872,0.3681014929715829,0.8343710566955509,
 0.12589136838785908,0.5893919982006961,0.9650501086040351,0.6723874111565231,
 0.40051294334113274,0.5224275857186194,0.3681821711285769,0.9352239829626611,
 0.5980069991823228,0.6036952997262136,0.5884225876344527,0.9782062721864608,
 0.10320764820015038,0.8510116702054545,0.46135813885830945,
 0.8931790989518309,0.7991927390191563,0.23790358878231876,0.2014136215103549,
 0.09657862766394065,0.8053126940652886,0.415022137608978,0.7953459456284073,
 0.7662573574146186,0.06848536548591666,0.8494637194506304,0.9420206270904721,
 0.9260781601046193,0.49448616538690804,0.2469267413683648,
 0.41573513477608015,0.9854682079837502,0.8727496434562872,0.3104043963195744,
 0.40124241448350717,0.08755826531736466,0.8265815490895789,
 0.6278320744638124,0.2753592496549836,0.31146271872628106,0.2582241930448742,
 0.9047376294990098,0.9227353542651762,0.9576190188463071,0.6888750079978597,
 0.4139606653726262,0.07425822677510863,0.8528695568114207,0.2147694814292176,
 0.14070168753807621,0.35788630235898655,0.1883841645236095,
 0.4987173097066283,0.5183933710188008,0.1702450748655674,0.43933495357504093,
 0.4902575909308209,0.676900294164372,0.13396653550143578,0.43034940608679195,
 0.9360781824554203,0.9635317414656864,0.8302903350781177,0.6985950027316179,
 0.39678013635937304,0.5037597598034731,0.769068695442005,0.7299873475776599,
 0.5870772326265384,0.4859067378401004,0.5282288279185995,0.32528812841888655,
 0.9987352670043308,0.5133945565455537,0.48918819322819096,0.9997529930005336,
 0.9292196799814788,0.36738695521771547,0.25943985774530764,0.75913865004397,
 0.7862742235567965,0.2567878484955697,0.7839010576153767,0.8557224013949238,
 0.024884529870359606,0.5289019923119238,0.8914486177321408,
 0.23925937005326525,0.8396827156763268,0.8979785808249094,0.3785737115359584,
 0.18595676401614358,0.8983030641777396,0.7864982815975715,0.7740225808918826,
 0.6044162618704466,0.5300488797881981,0.28332524692829275,0.1328873790534516,
 0.7200851871275125,0.126345690668476,0.7703188955977218,0.9117121445704335,
 0.6356952516784081,0.9839892862279357,0.10495354202347396,
 0.33774439304318193,0.6448325614836852,0.8195295061767538,0.4349345484900087,
 0.06699879571284795,0.022800969736986088,0.10296943258044122,
 0.6408277547633506,0.9634659169813986,0.3187277811809214,0.9705532564115249,
 0.1712132306526165,0.40043916207774644,0.5365198428657801,0.2963890786027832,
 0.8164549297567638,0.5789051755040544,0.7693234178798233,0.03204006316639052,
 0.8749321627370421,0.3157773450450645,0.7534153361692648,0.30217592171861807,
 0.15593466952474777,0.11944296643017038,0.5241483253372056,
 0.007019022165521194,0.1282418145235925,0.2859607356762235,
 0.6039236051026556,0.8748291340480402,0.17138635460472162,
 0.13033397453494422,0.41492960096013953,0.6755310580870336,
 0.32260553062255914,0.005739147406515488,0.3880199472020811,
 0.25232105671743976,0.7513545342620286,0.2030552560417619,0.9347797640035993,
 0.8083467889838994,0.9113096144813517,0.6357454172370579,0.9324037432719854,
 0.04918872496801108,0.9139822212310829,0.9697199789709129,0.7703642080928872,
 0.9638479109758329,0.8788469334509693,0.5146794063959023,0.04472419380529047,
 0.4455878753826985,0.2956986473782933,0.8913332884339878,0.10682459536258349,
 0.4264554692087428,0.32799621963971637,0.02873756989904397,
 0.10832054901892574,0.6805320480870521,0.9573053608681448,0.4179548249873476,
 0.3487919266802789,0.5046373154432681,0.5833789773504842,0.7019555505998148,
 0.5569621012843857,0.6908394556257305,0.8226302596379986,0.6576742781199244,
 0.03074040193310479,0.5608703878199526,0.5708294762250768,0.5812805478996876,
 0.1882750521163542,0.25483102830291304,0.859891331434408,0.49268774319633013,
 0.3702857309608317,0.39051667156993053,0.35374796024783794,
 0.2380877192469708,0.5991012821560702,0.4623781014387448,0.3738856474188239,
 0.18505716418834495,0.765182921231567,0.9114248268913735,0.4961965054532138,
 0.8171497411048008,0.6493777351523299,0.04480525301087268,0.9787774025263792,
 0.6127868146184572,0.9867941829613807,0.8574768730001735,0.6596955163549847,
 0.2740329929279005,0.7357935811204347,0.7159101083314826,0.23921131620870772,
 0.10839577938148592,0.8001186443479552,0.6914283135239062,
 0.45525381457669045,0.03692686791697897,0.995519958181208,0.2292889266320567,
 0.9932492257096954,0.3024745348552781,0.9542971142281809,0.26261109821831286,
 0.11258664618720116,0.07369655821659582,0.24876029874870098,
 0.36852786056542797,0.3127689618847974,0.4786423097835317,
 0.12521135500403557,0.5663278659226298,0.1891877918599043,0.6961228893906338,
 0.9360876917365789,0.17917356458224276,0.47165809989722807,
 0.9455174599668525,0.26314336355085244,0.2950875951165317,0.2756632038082273,
 0.27359078988527485,0.6430030334181773,0.5837256711605759,0.0790699104405066,
 0.1639387887514917,0.3290245747288122,0.8533237835606263,0.116582765539881,
 0.3771003872827843,0.8026302812435193,0.18187205486459068,
 0.08360543724250613,0.7490307724076494,0.26252791646336093,
 0.2143367616332581,0.03498216386193953,0.781665706369145,0.13884596056598486,
 0.022369532572357742,0.80160706156844,0.8108845468124543,0.8685544208935915,
 0.544165929686931,0.4977242582737732,0.299381119680173,0.915303907731672,
 0.720533082384495,0.05332520536555718,0.731229996739547,0.08762277999415113,
 0.8149457399228459,0.9441958479419222,0.25594494121220834,
 0.10878375845475907,0.3744309757182751,0.8619978578017384,0.5510538227819917,
 0.4247044769230325,0.5469154716590239,0.07140195232297986,0.8810775307252903,
 0.25800717364744863,0.8237524005139892,0.011886882190756398,
 0.8168716541574768,0.46598277663983667,0.8469481152869726,0.9831433445899133,
 0.291826380628292,0.27005930542723255,0.11518070190140395,0.9442024851691557,
 0.07803647019915183,0.18966222101286734,0.8686872044427881,
 0.8152656083717992,0.09017530513589722,0.5785752428816258,0.6985108142629227,
 0.7387138008932972,0.46093689143052174,0.18066520929886254,
 0.24106624974550617,0.8275247817188655,0.5048364706849755,0.9967150817135787,
 0.18881021276183085,0.2896729660781395,0.30795562329801296,
 0.16364443580450228,0.25441018741451094,0.8978772916411147,0.462664096734131,
 0.04071909147472286,0.139868417531817,0.9356259597699476,0.3978889243433422,
 0.6347475283544766,0.44319053204512027,0.30958856018557235,
 0.05898646692638665,0.6297245726468913,0.28288265295010717,
 0.5800857368008803,0.7330459271873494,0.06138079888207404,
 0.04287558464312202,0.5674830993412544,0.04815388066287396,
 0.10991546754564174,0.22884006217149477,0.8360085206663627,
 0.3742636730379574,0.3935245034300331,0.9064678722061099,
 0.060011112940571376,0.47932805124471156,0.3268087861193184,
 0.21635158184417025,0.9307368834734991,0.8215338435212222,0.6088301401370348,
 0.5799788916175099,0.46236241668480926,0.7905158370751653,0.2554734054872677,
 0.18534583922994574,0.2573666955851137,0.26659084658018484,
 0.5305446361404317,0.5415565186444832,0.8484879087127343,0.38073121811736144,
 0.1788662365078003,0.5139188783237845,0.4247748610392863,0.19673409598252445,
 0.6733569041884021,0.7846406800776893,0.9441835579833749,0.5461941151000341,
 0.3765267235348462,0.22169944326908708,0.45460923013378685,
 0.9200862294930151,0.9869510930576708,0.9383017639454422,0.5636844698513452,
 0.25945737096462596,0.6235458166929906,0.9908541835629953,
 0.28980864111750815,0.3565857381909725,0.557293042570761,0.05971258485671982,
 0.26170453929757276,0.6281119069573792,0.5380749576977034,0.7710424061144144,
 0.20465570632013763,0.16642778492816235,0.9117636173138319,
 0.26488796761638966,0.5367754071117293,0.9843006728527051,0.8226456487145212,
 0.5627918133698827,0.5570105980027791,0.5285155914720241,0.5145955698757763,
 0.7250122699980344,0.843334716871949,0.8060528196677288,0.6941123576576788,
 0.6242922105062745,0.09735220208910011,0.09248524430820304,0.34360144514722,
 0.35101781644176866,0.9813130307285268,0.4612421565680511,0.6015824531847563,
 0.7186478293385902,0.658365130389481,0.9297642587439022,0.8124464445503311,
 0.6305499678120008,0.34999983439980786,0.003110712884229261,
 0.5301833904976809,0.9756073515657372,0.26442705428285473,
 0.18555235386859337,0.7552218821730365,0.24333567612448292,
 0.19128872404143082,0.29141161327737974,0.5743365264797315,
 0.4494755670518782,0.9223730575531125,0.5561289813821634,0.33477626786552395,
 0.25960293600222983,0.4142107438903777,0.7967780051828819,
 2.27877036031221e-4,0.29685283610793156,0.1622271125639001,
 0.5988519532737194,0.3013297526401826,0.8578974777892319,0.3834638009251873,
 0.0451582995550972,0.7955289439825918,0.22285519405939724,
 0.32976215656451924,0.47600316493546657,0.8646594171612713,
 0.8249669001932382,0.2295398071951975,0.00901371489012781,
 0.09772413807824698,0.23162379880875228,0.390921210096813,0.8470325571678754,
 0.8213887419257716,0.2816041624636356,0.6776365971138167,0.665295065787016,
 0.6334151989840078,0.924794852307508,0.789642198160805,0.6573721238536581,
 0.8119100113551035,0.15778558077038551,0.9792574897942172,0.7147536830334287,
 0.9058754485180989,0.9816982750405815,0.9934940738280252,0.9822311612622857,
 0.1996541778705555,0.49390648643121804,0.8808363026011999,
 0.20381993286904576,0.9618276469480223,0.2619048864255865,
 0.17191964433096973,0.7222735220065813,0.2056807230281431,0.691169534854704,
 0.9926813428900582,0.1888372756721075,0.6571096901670148,0.2647030466418612,
 0.09413026681262815,0.5734182232209128,0.14670395746630427,
 0.11882208564172958,0.7602969531018493,0.7247259648560854,
 0.26863351822848536,0.4698047049687657,0.46180788359770963,
 0.4690931067158083,0.03993919822789649,0.03349647691834634,
 0.01695777676761867,0.3496895340832562,0.6948072504121199,0.9764872584784056,
 0.24339025120319469,0.8855201637670344,0.9836505299956282,0.8031429262022162,
 0.292569111217478,0.5108258022485572,0.9774012939931757,0.02520655100198932,
 0.4452226856963286,0.4447745147105069,0.2632384741327408,0.09610636385811211,
 0.1596255059370263,0.04041676898350843,0.8894623417228371,0.6394338087974846,
 0.03442271763292526,0.6585645427939532,0.5688538178187001,0.7416470038436819,
 0.3756301279643167,0.283469752143621,0.10924153770496037,0.265257314256349,
 0.28469957156872305,0.6131186852446321,0.8652664468793478,0.5896493712119767,
 0.5299987117348639,0.26919088769762756,0.8432808847787034,
 0.15296181085167704,0.9412271864692348,0.09480471533305967,
 0.4314480867409566,0.3674578455748053,0.6701863136669539,0.8505137865528979,
 0.30551214322400033,0.2728575418636132,0.3786090566915332,0.5396383295274209,
 0.683355177208324,0.5151618264713902,0.551106080270453,0.8544441127031825,
 0.9523613535007959,0.9364541008737958,0.5028200002946914,0.10476634718666622,
 0.4788107655925381,0.7012402418834337,0.9289598122608732,0.9041972113340024,
 0.18456096082201612,0.5834776322974121,0.2713283442502453,
 0.10594487411264464,0.8086366784221448,0.8345093286247315,
 0.11335196048281349,0.4908591631583681,0.9514855013990666,0.7193062707802,
 0.8273234856243971,0.9972214625392755,0.6426137963880152,0.25814346718192904,
 0.1127581070817274,0.5762159853979902,0.3656214691688413,0.8748781179725604,
 0.631725934776001,0.794321426128812,0.31119765172458735,0.06645490715971736,
 0.6486130903206652,0.22765037848592018,0.5405088913589609,0.8325344831527359,
 0.9256897756435475,0.21501708794546814,0.6757018929281434,
 0.16623155541413082,0.05548241211829208,0.9850210397062478,
 0.8702723016269546,0.6814460381376325,0.8205937513801557,0.9390165953798392,
 0.12400061976033872,0.1613608009888965,0.32935675328180414,
 0.04000031988579955,0.41603631212706715,0.6708689320233373,
 0.22157946582332821,0.5518631695779107,0.5350333487976556,
 0.43763526530190133,0.0709666943510473,0.09682445648585292,
 0.8251116311745479,0.6239033374701034,0.37331424066896324,0.5395149209569071,
 0.02245431373734441,0.7996387898542542,0.18523079091753014,
 0.8875609879629512,0.38355620261511736,0.19991620120854436,
 0.35146627083189363,0.2241057107374066,0.32484828360958706,
 0.7895269136110684,0.5449303176566339,0.46514300329474045,
 0.08356167522854663,0.07862341418030816,0.38738853581909904,
 0.13443702537960012,0.8984136684512829,0.436584392840891,0.9046363802795159,
 0.17005225318303485,0.21453347646918774,0.6838166543063217,
 0.10724307316073634,0.0596319889135124,0.201815827805379,0.6962266551733917,
 0.058516307904086684,0.06258501707783137,0.0923601097331751,
 0.2156973227787966,0.8104646962292346,0.44119629705848495,0.7975790948421743,
 0.40029592595940255,0.8822138058137604,0.32219343107219334,
 0.4193132099218704,0.36489661581296895,0.28870352948631817,
 0.6272988567911467,0.27325227259634743,0.25450042591333444,
 0.24380284692107757,0.28334319691367715,0.510764294217555,0.0666410820210509,
 0.1341202653103425,0.8584167552808069,0.2979294166496278,0.4796982871029356,
 0.17680346103550637,0.22211099319876582,0.5298718118939536,
 0.37010859646736716,0.465139850508551,0.3819516448673781,0.38651427730581034,
 0.11790060203202135,0.4656668721740054,0.2868608508298971,0.277328228582291,
 0.18268735938072744,0.21441132303851773,0.4404614823272056,
 0.4118122583588314,0.3605109164092346,0.7050630078471373,0.5878148474653587,
 0.103295550165032,0.43033328207195753,0.5576470059812115,0.3593493228097976,
 0.2716780047393178,0.21281004435414608,0.7765671032074628,0.5247565386736678,
 0.32621814859560616,0.8890185003861115,0.658408654642771,0.6218079181187001,
 0.029484042222813756,0.05487800438519119,0.8403937310940155,
 0.3143726705795923,0.9588495942198506,0.3356088008937793,0.5042839601548443,
 0.3368285274682552,0.8941281899684468,0.9523946436330046,0.6224702406131112,
 0.34513870101315336,0.6204893294468588,0.6346114175006686,0.654286287170281,
 0.319364121389884,0.12254696551042987,0.8603348697944527,0.07033866431742852,
 0.18242200933706565,0.009535644119650577,0.19898253811358335,
 0.606372557365133,0.20784250602424004,0.7654447529012628,0.5418437004440938,
 0.8990133463319971,0.5447775105146303,0.7562728063098751,0.25196152219775514,
 0.7288572971721438,0.22697553185985975,0.44541427178753334,
 0.9739914088851113,0.8060074987150332,0.21198120858430614,0.7896112688956672,
 0.7029446102184896,0.4993738607778123,0.6660158846585721,0.22914771912447918,
 0.8687414213515516,0.6016250853243532,0.8420671008027522,0.5834536518057662,
 0.9603020890164039,0.36804208633146596,0.05371880868705603,
 0.3996549722495353,0.2998040017407677,0.14634942678582807,
 0.10040651791388067,0.3561631945212409,0.513023139623732,0.6983007159931578,
 0.5831658833173732,0.6491227631132405,0.540389116022493,0.7515082462240172,
 0.9213312906809044,0.3815522291661908,0.6312306012135475,0.5059755267203522,
 0.2640254867650761,0.6667519673142109,0.9126007103016712,0.7075824938395777,
 0.8345485334917234,0.10135862410908802,0.8797795793627438,0.4698844365485956,
 0.5786324860056052,0.3744176183998147,0.928246755224943,0.37153466260320667,
 0.9040216507841777,0.37580171794502326,0.9354325663570299,
 0.09858541503882434,0.0616949109478242,0.9409279102524076,
 0.07443036244388868,0.012946422340764707,0.9655386675852461,
 0.20032506064609024,0.4290818112373338,0.17017951974784373,
 0.2022279507384459,0.690562196620949,0.7179825975706233,0.02342834225479795,
 0.8374282937896013,0.8921811012999552,0.9395046903311733,0.6756695363694462,
 0.15719917984468545,0.4232120990909224,0.4023163312868615,0.5305179969923016,
 0.4829080158003166,0.6542694513251497,0.3346554999111512,0.309941591561373,
 0.5218515069855192,0.8010967003949454,0.5920280157874368,0.2839842765820829,
 0.9899338891270795,0.868959943645139,0.05919682413007321,0.6029558607127503,
 0.6433098098936939,0.3202879666522107,0.745368360855633,0.59449934441716,
 0.04733782045009238,0.8119810578308416,0.9262525173511901,0.7307649224237074,
 0.2950344675462566,0.7674841052373159,0.4570470703155646,0.06894664954795537,
 0.1488516749896055,0.14817258297692837,0.5300575053244545,0.4471299455872674,
 0.5983685998261774,0.5336155511814555,0.7010787592627713,0.18854728775700935,
 0.09220261206514757,0.8973962861445223,0.30893490690303094,
 0.7806807223406866,0.5800391167813551,0.33471728399334477,
 0.003350759496959732,0.08410863849039873,0.1504122727307865,
 0.21900822074833315,0.3871948731506243,0.7263715340325193,0.438952533612603,
 0.7038154637948648,0.3872306761471378,0.7741496931511084,0.8326262152380737,
 0.052302713531961365,0.6348369869875383,0.6594196466435722,
 0.08098607585283502,0.3169853147584307,0.41281517972289683,
 0.2430218683490616,0.12788339009463645,0.5986825921249472,0.2188090742242128,
 0.1281260788479781,0.4108468934670134,0.2963146378640562,0.6374997343153402,
 0.15415571699767672,0.5951176250967908,0.5515348781529389,0.8050030862826807,
 0.732549228953022,0.7177482021379953,0.37800798764692634,0.28595041026838885,
 0.5317466522984828,0.738979846059366,0.5202948050429965,0.42504700734602696,
 0.3101054327234315,0.40068091630201785,0.3811808626430667,0.5869217702383698,
 0.3610563127962534,0.2667780830027444,0.5111340068993045,0.8483155844928121,
 0.7869160857448638,0.9268533902939686,0.37898757988507803,0.4845042320462164,
 0.2974517505011437,0.5746713301367433,0.6681604869710833,0.767055424138394,
 0.6605996279244956,0.6075573670898803,0.3967676500535703,0.2472401966744392,
 0.970772954821673,0.6951064214649664,0.08967206304840569,0.15786633704499675,
 0.8539425375139515,0.8823707051820654,0.8378099067556198,0.14582511761937123,
 0.1627095766632205,0.4467413806549464,0.6273662651698702,0.5144570171071878,
 0.14184481122677217,0.6622984688066973,0.01062242264628488,
 0.3484452976965484,0.5895377269277178,0.11330487968901115,0.7504295845272274,
 0.716253716985353,0.9356854251318171,0.8997974056523603,0.28497120435250456,
 0.24143289289418646,0.5242943448035463,0.5092651192547835,0.6981804683490457,
 0.8609125605897212,0.49636970410723,0.3199698403640546,0.5588699442008576,
 0.4191404697819956,0.33497096506452007,0.8174075320825547,0.7331377388670279,
 0.8081726232191835,0.9960444319026832,0.27246010357648554,
 0.29538583145837194,0.1806752245353892,0.5798823724740256,
 0.09511800272873672,0.28388064040733163,0.9361866826567214,0.257237224495966,
 0.8893377541272137,0.4621533055032879,0.8272540447717096,0.7044721996377753,
 0.33655844889457087,0.8584254394976912,0.8009057810248801,
 0.31602248651911624,0.8348354399682985,0.5787179002294226,0.5646275786649035,
 0.31291160902899395,0.48312205377129,0.883437853960811,0.2003234735562447,
 0.3740981348839495,0.4487132607415151,0.11138109462131707,0.1549224965688908,
 0.5535704459244468,0.9121797528038645,0.6256383332182097,0.9724392943107502,
 0.40355888111034877,0.9259073532977687,0.312017989696486,0.3820546658278561,
 0.37406556078284736,0.8904012591870529,0.37224343789505254,
 0.8478756225293776,0.9121418457845062,0.1818246336821756,0.5187995957421363,
 0.3192990799366746,0.4670662163153416,0.11188143778657411,0.5212749413571319,
 0.8981339180025603,0.5993296690898798,0.8793520230493834,0.4893133218878468,
 0.14938883570055883,0.8277723378984222,0.09322499487554814,
 0.5258429850316456,0.7758463487568642,0.004884679715831397,
 0.3933935124505422,0.8603046154605165,0.0638013597750724,0.33687144936652247,
 0.4349071110505036,0.42152558054754097,0.7160515566010818,0.5907981756078107,
 0.9815941115806867,0.1814795172972513,0.8661958859567974,0.4379858822044095,
 0.5094400461374031,0.5890157627558605,0.05278340342628152,0.9802487955061041,
 0.2263837554509427,0.3802945445381951,0.2263869498526091,0.6560036639935912,
 0.7988262123261363,0.9910861978539938,0.7069780652333382,0.4068352028358543,
 0.8543143822633453,0.82512162253832,0.43326984007728886,0.040132411399681156,
 0.1308885072571997,0.9931825086915212,0.3391881030078655,0.32245170982346494,
 0.7417093371897708,0.1888484955389217,0.9171943375319945,0.768208623400782,
 0.3604771804306529,0.7352501019124686,0.08070409333908546,0.6924723305954225,
 0.37743722525109846,0.32811523787500896,0.11631790931276453,
 0.5262468497793795,0.7180538344991931,0.6744180713273916,0.17434482063259316,
 0.9105667789319319,0.5944635315683372,0.2695581337274524,
 0.056864325978112085,0.43653657409115054,0.10499913664206595,
 0.4774642476567478,0.9927282508234558,0.9321771583317036,0.3692883562402234,
 0.0037808681894481166,0.5857934977607787,0.6984113234678404,
 0.07018165774187246,0.6690627320589423,0.7153753815024527,0.4016885842758622,
 0.8862989746929688,0.06055653320091958,0.40297956844659855,0.266793218735486,
 0.7716159279950748,0.3554335489747028,0.0235611113365517,0.03171809537085102,
 0.36628533354372383,0.5806251229360387,0.38904501398309177,
 0.3320182195832746,0.13007816771341174,0.18973831243076333,
 0.8905867191708156,0.6881894081587496,0.7955800926569326,0.7588602358350475,
 0.1966909587817558,0.03319331232982781,0.25371279876694774,0.167968362983707,
 0.26862971595613305,0.661959511570024,0.24726498734683877,0.4684458549811954,
 0.5712111100319656,0.014781342206586667,0.443027053473533,0.4977114077038759,
 0.8092130065354299,0.8330162076636397,0.602836978029911,0.2335734097017843,
 0.8278222861135528,0.03515576340502302,0.6460998836667341,
 0.05169623071710705,0.06289950011202305,0.17015830371339213,
 0.41048164257446984,0.009052443581715641,0.6359802546329563,
 0.40689294966402323,0.5502341613185716,0.009407631323653853,
 0.7000086144297186,0.15088466534846412,0.4499002994951553,0.9724324368486924,
 0.6711907359493087,0.17233978747703094,0.6782742157679837,0.4078121996744527,
 0.2610960094214505,0.3450600710606557,0.1316204985180427,0.850589150246841,
 0.5315796450656466,0.6553815627243932,0.8073250574072527,
 0.0022447155152924836,0.871243502282907,0.3977079537682282,
 0.8293793186473881,0.5183703944589005,0.17611919534020537,0.4790108795011362,
 0.5868027228188091,0.6990027891963057,0.38215202619178945,0.9738577916518326,
 0.2933076339295976,0.2999213915949289,0.5597195866471053,0.39273190061447105,
 0.4498540640377604,0.030309976494494628,0.14818541718239375,
 0.7417884923783449,0.6277790971748767,0.8813914732345973,0.7943400922161649,
 0.9114858656192261,0.9209920000954098,0.9770549321344147,0.6669814333239699,
 0.6488421460958473,0.2238374192353617,0.7748610229998061,0.9916043497249685,
 0.486221978557292,0.17405955197462353,0.611908519599941,0.5891527229049557,
 0.9968603551429133,0.4452013363009155,0.5444045030232133,0.6550218072138281,
 0.8362489857996932,0.9315869997099184,0.6380522670288313,0.20220395380777234,
 0.8267108342658209,0.5626165410639425,0.6122697981954486,0.4564457507498485,
 0.9504989189165665,0.8608316508161576,0.8895903652973832,0.2820723469141203,
 0.10832616138624496,0.24596745578419021,0.15558801267042055,
 0.47831933819982475,0.17569960596779033,0.8775860513140468,
 0.6611074246586675,0.11270821329081926,0.4769909165959312,0.9205583376153589,
 0.8093195796443456,0.2840870529639058,0.7116747860873263,0.6223016065465404,
 0.7525686484193601,0.5104534809041744,0.8007625720680301,0.4576032286247791,
 0.14728890794232197,0.4441012006274456,0.2958271428444168,
 0.49496332487229533,0.5242298599629327,0.24411163280962267,
 0.1060643383032771,0.18742589484689054,0.22432945324791442,
 0.9140265483753581,0.41969267187061443,0.12878949704872533,
 0.5231944266344551,0.12277426622244092,0.32122627376173163,
 0.7124449630225521,0.45802631979679265,0.3074939022644756,0.5765495746487953,
 0.23786979996525526,0.2347364391909621,0.6348010007729026,
 0.46653116411324147,0.48044732299322535,0.4307224722883163,
 0.27165336113396843,0.5605695111443879,0.08927401649984679,
 0.020591480220876757,0.38064634141439646,0.49352401552828706,
 0.7390948237912669,0.4015276731196462,0.4377272411770945,0.8280541271546974,
 0.004636649314830765,0.44112514611804376,0.3707728434226659,
 0.6513542270882816,0.8349756628595563,0.06365210767522056,
 0.27659126793341304,0.2674012693291641,0.6308818607974506,0.5216464517772036,
 0.5009688389429687,0.06804975175544381,0.6378130168897377,0.4415280292181728,
 0.6315982229738253,0.9772192191746083,0.7149948175591726,0.30730648479982137,
 0.21390049177694692,0.28713089804561964,0.42644683437489583,
 0.6429885099962478,0.03735929347565525,0.13613820216718087,
 0.17690661321243517,0.6810306694069314,0.6765455821459937,0.958345787699753,
 0.6946757664518093,0.5578286630970614,0.4022832175423219,0.5182431419647144,
 0.42624657788359066,0.9033562847345366,0.1298206375284009,0.9471053968259663,
 0.8205794989003864,0.7524697207580844,0.2641622708022079,0.8391494071117691,
 0.8507991735503908,0.7803143754089302,0.8220624497791327,0.7261053453935471,
 0.22174242791526477,0.9547278689197238,0.5544866637331157,0.4760965823231328,
 0.35924192147462775,0.6954536767858263,0.1947055443162995,
 0.16844674511683155,0.3949450197599921,0.4488133138830087,0.9405153162484776,
 0.7169224496290586,0.1591588321587626,0.1465203489236283,0.7437443605051294,
 0.18076446717338102,0.8076422835530966,0.8355502832643377,0.8824583342753259,
 0.6317570727261613,0.5711016477344036,0.9515422490205462,0.4808100116578332,
 0.7375187688629726,0.35345034892843674,0.8053289886088032,0.4863376177552676,
 0.008013384376652777,0.2666081234451977,0.2923962702560474,
 0.7022706235751597,0.718806947661494,0.06698160690651767,
 0.058573329448244116,0.13822702445963486,0.9181118345942926,
 0.48827297777509193,0.481547983638954,0.16917889390314178,0.8078928644688865,
 0.1908895124293587,0.0032571925713539285,0.43747187671561427,
 0.16611608834786684,0.23046822982041826,0.7749291105741783,
 0.8995950405082916,0.41356057205793095,0.10089138902435979,
 0.7238024200553504,0.4450047638832806,0.33748218037988686,0.8333789462618828,
 0.8854727580652612,0.32787576578555033,0.4388329886473963,
 0.03027540927290595,0.011827245397304176,0.6137286499477681,
 0.6202791154488182,0.011157498304191327,0.8243460018001615,
 0.9543095297276574,0.02187452388634159,0.8925749991652976,0.6036708656598421,
 0.08346394264916102,0.6044517825416509,0.15127645813590973,
 0.8452036299732197,0.062203112810084615,0.35062377218737995,
 0.5590048628451151,0.8734606840130078,0.08019349589752522,0.6894825258339777,
 0.7333833276675146,0.2641776059476627,0.34112512443384435,0.5673626418033217,
 0.12699787125657636,0.0803561552700729,0.8809221168667596,0.5237396562311034,
 0.6915032514563697,0.8293026327554944,0.13699722152158533,0.3574785626454352,
 0.12962006981863283,0.8307396481488694,0.3995821309005495,0.3110737192387074,
 0.21514512833600152,0.6248100490534779,0.8653318867990156,
 0.26029036769766345,0.6975860486015659,0.6503365566096024,
 0.19345494741753144,0.3272798825679353,0.34750673304939583,
 0.41941826762529755,0.16979003387672753,0.28706339949514215,
 0.06008659408054773,0.7530493954418438,0.470407424569369,0.6406670611388927,
 0.3807881198512054,0.7543761876583992,0.3711527885918072,0.9120603354788801,
 0.9171578947708806,0.35766213970039407,0.11324498296372765,
 0.9260101166541755,0.03195489074247204,0.22101926700678987,0.441258274927111,
 0.7370990804284742,0.17750598620229308,0.34371542589458415,
 0.3373686196129724,0.4043950455617955,0.5753649745495251,0.5114660480028057,
 0.6409617796800818,0.5921328347230979,0.06317662479524344,
 0.22651872363553538,0.14009903326517636,0.6374037102875617,
 0.8352800073904655,0.6012173536906497,0.6681396183484584,0.7493469094109685,
 0.30317252817874896,0.3468161242259181,0.3730812286839962,0.9019173153138277,
 0.5904141139660783,0.31689972744743067,0.021429509770430677,
 0.5587538883760472,0.8225095037607927,0.8361695630546678,0.828924265154762,
 0.731165934559066,0.14055366847083184,0.5803824783128693,0.3794888872684613,
 0.38284585619267686,0.8889849939303919,0.2500704794953811,
 0.019472491764904398,0.34981205108524294,0.7703042462114893,
 0.39608373146945297,0.43165326025517037,0.7743940988045052,
 0.12306053868631572,0.7175470636365255,0.7840340427697408,0.8490505802286504,
 0.5381093267232298,0.11368111651682333,0.7858105545469953,
 0.09560559168345151,0.7202475294086202,0.48554737726309816,
 0.8242978346609091,0.7425130297448478,0.48034178047408305,
 0.09229036837629634,0.13177961872194244,0.4557245296899335,
 0.9347956229306972,0.6685275071827891,0.6411255205100068,
 0.023632763624163333,0.22082564494327028,0.0754798500910494,
 0.561247894797132,0.5259632844479842,0.3020332288302714,0.2688249439599526,
 0.6042487964903298,0.3570746088738339,0.8524954563186458,0.6823457901058219,
 0.23291186420318155,0.10440114583485438,0.4007374341815011,
 0.5761032154475296,0.5173711350969264,0.3860318670238335,0.4846396252138524,
 0.24346088974940172,0.5987494198673704,0.8759966723782899,0.6387369737000482,
 0.5644076205629776,0.9134806716975356,0.5169126630118026,0.28555596140397177,
 0.15331515171089116,0.07102891157370728,0.9219534985040425,0.862541513440799,
 0.8553585953325957,0.22439499749888048,0.4260931390933774,0.8051857052935418,
 0.29973939398580096,0.7233176229312376,0.035872256745731024,
 0.8269798465016551,0.15780706323281413,0.5789224025122586,0.2705423676739621,
 0.8848842871807117,0.7974598611387884,0.7618264456084874,0.41467731116943674,
 0.20411001697628683,0.4163785838114423,0.8677335890162053,0.7319053895447312,
 0.6554724121410243,0.5676915761611301,0.268840881232668,0.8943366183674407,
 0.35863985558834277,0.27318231687963435,0.6219142978358467,
 0.9152951020241172,0.5791209395794095,0.44935659714901666,
 0.050677075833705065,0.5652151498735946,0.1857070635624415,
 0.5729186176527067,0.8089563595824175,0.5240482605182846,0.46099989265411057,
 0.42101932971579303,0.551558260083585,0.17273643998004662,0.6440432066875519,
 0.3381907052506188,0.10865936230118578,0.6235660282193393,0.423657372790046,
 0.5658220044202958,0.5277132720791725,0.3838479496064108,0.6693862370955963,
 0.8781821172055656,0.9840948656435722,0.39164748494447854,0.2710491655474583,
 0.4215711157104015,0.9535986280613754,0.9455895830006655,0.5939841639059926,
 0.5120048812502436,0.027449881868498727,0.06297375288382456,
 0.2526808283103287,0.8311021913821968,0.2298881021834338,0.45009836666943626,
 0.670505144329082,0.9629609394976896,0.33110669906749945,0.35206225186417417,
 0.46997811770847875,0.27145294342956405,0.8488499127002264,
 0.05948167658772463,0.824528564747274,0.9497067410852403,0.02685257561087573,
 0.4951912676534086,0.2373004309102662,0.7100693125965067,0.10452076618219208,
 0.3813162682760656,0.500973662319388,0.6311204420360388,0.25079853560044263,
 0.5656782197639738,0.7970543686741556,0.5678662431992347,0.6933433095157572,
 0.806027501548186,0.10804326339014259,0.20640994799655266,0.9849987723560257,
 0.9021161816827188,0.5035201761360303,0.8860402621002739,0.8357983612143515,
 0.6076312523425309,0.3667309956467508,0.7630780225955953,0.21812855020851596,
 0.23284652875893042,0.17165116379032452,0.6945092186215587,
 0.9735259886883872,0.8781580120094804,0.29675259994368786,
 0.05062023560921869,0.40533971828334725,0.8774834073075661,
 0.8440006455247758,0.4721652217341141,0.71072101277353,0.572582102969992,
 0.5118520712374502,0.24985872253813213,0.9277330898107596,
 0.11417271686080421,0.9165519057955602,0.32740783235348636,
 0.46490736065683136,0.04158584547318345,0.4810175750960367,
 0.9728142550138091,0.07032960989917236,0.07721395290600275,
 0.2091417273090337,0.8731232138164882,0.19936621882613437,0.6918023732768529,
 0.5147399691176122,0.517006180646754,0.2766799047635443,0.6364102633446418,
 0.7168330849738639,0.7572397385852805,0.6358092206554828,0.43433179510545716,
 0.22949620878900534,0.24861588583193062,0.21804460246944812,
 0.4096189142488206,0.9686813058436741,0.7663859457175144,0.14133047879022664,
 0.8718108016017939,0.738796509302371,0.4096363550670039,0.4985616521408478,
 0.8700691572967024,0.5273733538170422,0.11305568979930303,0.9715748642098656,
 0.9833359387266101,0.7928542958351567,0.07285822117723717,
 0.38026716099209723,0.6648600277010753,0.7012505547880086,
 0.004141863667041834,0.5211722980406122,0.5066193042354027,0.581496352707304,
 0.24976849721703975,0.8122543188040819,0.5138758630584083,0.8435165334475322,
 0.9142849404713125,0.004876076637570281,0.6210788937321617,
 0.46033449388388314,0.6686187872975108,0.6807590951548347,0.3240953168342662,
 0.2331907720469446,0.7475490139096738,0.050007029504283707,0.899283652914383,
 0.25901282583314456,0.7167442424299428,0.23886062032290312,
 0.6621876679976344,0.9148013020694392,0.06545933405069715,0.9803131256106079,
 0.37446553265674676,0.1570530085267876,0.6976876011981427,0.2766364958370733,
 0.4407165543678828,0.32904942601238274,0.5220635399679938,0.816818816549272,
 0.25677787986638245,0.5385687160512062,0.6001423516212333,
 0.12181159163513877,0.4523052604152987,0.09278851074511907,
 0.9235326578574602,0.8811195552800148,0.5551086564060964,0.3882718763064985,
 0.07840318525676215,0.4424923341579954,0.06678447170459367,0.845894700442427,
 0.3294816452900795,0.893850339173796,0.3272639860847568,0.7392467043938561,
 0.03260508223248881,0.1598614371746826,0.21413356825853247,0.762089389334501,
 0.30114601716944045,0.0046025033258091375,0.7981464596246101,
 0.32959639455695,0.10120160308267023,0.3718916022391111,0.013938851722641421,
 0.7483156734411678,0.48336457793631515,0.5413098751409522,0.887290233765437,
 0.76076765444579,0.14897731210961318,0.059657900501184025,0.7987684694275438,
 0.6218679527659493,0.3602540960348424,0.5680661869891335,0.6189765194347441,
 0.31827668362324624,0.8094061926429242,0.04061846898065846,
 0.4486124123616503,0.3606629789856306,0.21198041121126,0.09079594632066468,
 0.8616778966799453,0.1802482341781917,0.9269655480538419,0.4083545881577343,
 0.020836391942714694,0.5623720669331179,0.6890106372366429,
 0.5235043100681386,0.7531308577758032,0.05004671617431733,
 0.023239957479031625,0.4939980390754529,0.5321550095658564,
 0.012564465425412341,0.5812542462900553,0.7759492056003064,
 0.33073831898876127,0.4385775896177627,0.7002597236945844,0.9595815130001061,
 0.026270197709757026,0.8738714685687252,0.8201940660990442,
 0.8730639597754687,0.3104356601241264,0.23126275645678906,0.159763157601438,
 0.5294261721323569,0.16259364502690743,0.5116703119905752,0.7843409719720855,
 0.8802421825592857,0.4009495787822599,0.014384322829669949,
 0.8818713995907511,0.5493432155717677,0.7396723718519806,0.7263148869641464,
 0.6586585942823813,0.6730785921917998,0.9126473719290268,0.49485887237716364,
 0.5822121819146486,0.4633508630466803,0.9341320248554047,0.739191100107071,
 0.015255382963449304,0.3392461258427725,0.53831518191125,0.5020951342929083,
 0.6139038464397607,0.5862902547152333,0.3600939373306351,0.3639930505167611,
 0.7138757994550923,0.4002802519805809,0.9386314799593289,0.7969439386304031,
 0.7357700891607908,0.8005322138770318,0.2229941254645038,0.9801520677586845,
 0.6457611797699757,0.5617926079129878,0.18101703657356583,
 0.17832254165309758,0.008684150319386297,0.8453989277342785,
 0.11642273985087148,0.08711657011604057,0.10457702401117286,
 0.23243367033228868,0.9311944754020532,0.9041830485634594,
 0.45835201800026293,0.8813062300049188,0.028970242974023286,
 0.21560740719783944,0.876359017028099,0.851379552212909,0.21899571197761258,
 0.5178413361385655,0.3688739263344879,0.0977992798555063,
 0.014853536981941584,0.5217781573784835,0.5907728990563235,
 0.4606762879438112,0.4868153088784096,0.31275116360667543,
 0.13335343846624126,0.642911121391296,0.1874488920153201,0.8402625867748473,
 0.1350263666599122,0.7208817046390315,0.8070796814268955,0.5871220542418337,
 0.861081504459868,0.23819381294542796,0.5687137088417722,0.7752173880106026,
 0.4934461568954497,0.054753615865321015,0.58259552889443,0.646087512656063,
 0.9157142316353486,0.3740013873537108,0.20126374850824325,0.3400810587349332,
 0.6627235168610055,0.7203115712021974,0.964278209697063,0.6591192484861985,
 0.05144661779223181,0.59243900649293,0.6171466696650273,0.23099749038964146,
 0.032856260649030244,0.2716297790356448,0.6766069661646974,
 0.37661224782695035,0.11389274400339344,0.5857555085207917,
 0.24376610920229824,0.0908541887675075,0.4546118619092375,0.6641552460945428,
 0.07443099445780321,0.14144587783061535,0.34084193697044907,
 0.06301266825323826,0.09656236465150514,0.3099262392888992,
 0.7744354366519055,0.4121557806089984,0.2772524937791503,0.2077210826844207,
 0.04593423560073229,0.7067035802797612,0.007432883754533215,
 0.6157937124557735,0.1175545337725945,0.7562503816187988,0.7797516570431818,
 0.09661775894522329,0.18141845816597368,0.9596883552411252,
 0.40730884698752035,0.5496282962872204,0.5870759123452324,
 0.16518724174298471,0.37336902115733794,0.9981822102898299,
 0.0633835569447031,0.16610534797023224,0.4089710857871599,0.7127495421493588,
 0.7981692612009006,0.55715448333494,0.5132004669846681,0.08794210976221539,
 0.6638808886737055,0.5146016626773404,0.8699702760664786,0.20641527800387105,
 0.42283973907462,0.061856176323479395,0.7778287639644847,0.1685337067001701,
 0.6740752727377711,0.7368231621258404,0.6582555580523197,0.857406500314547,
 0.5466527445537639,0.6125253695534002,0.4776759841507584,0.2938306134485025,
 0.24546215614658395,0.6032104054587274,0.8264203760562627,0.5755914719695896,
 0.21003296633910962,0.2872045656215457,0.5219719053462768,0.8513400960623014,
 0.812974708839426,0.890794534578725,0.32604496812471173,0.22615227536776117,
 0.8608423177965472,0.9964351063591809,0.46529525971869357,
 0.33680001965384265,0.33366019985014617,0.23206955234156112,
 0.3704092822090478,0.6306874345105012,0.10661756290373814,0.4013036879530082,
 0.30727454249944275,0.13853793145995197,0.9232387860161404,
 0.11024705260269663,0.8421729497578835,0.41389634905066397,
 0.4558919997937001,0.684141137485847,0.5577317088329605,0.1365845114183749,
 0.2531228623173353,0.5703587843636595,0.5405387596713563,0.7333059632997847,
 0.9874421474807507,0.548491890968413,0.7231189726700801,0.36647212588622746,
 0.921667276612627,0.1955070955987961,0.9218637376772969,0.2755872388245679,
 0.640907726224541,0.9157718346150109,0.41940001674887994,0.02456625401396806,
 0.3008981888052624,0.36690613287905227,0.20255805867628396,
 0.06086973432320386,0.25036564783070847,0.5357757108432687,
 0.43124796277866095,0.08726728779413406,0.062438603010992466,
 0.8061385637249494,0.9564041060064579,0.7965689219090089,0.6004966242043576,
 0.6006601573857928,0.4585898268828922,0.6539222464075244,0.37696943269159,
 0.4979575874502129,0.408704404781861,0.6417596897775897,0.7949644417398468,
 0.8665497163956247,0.6464596616892234,0.02597201512136149,0.7298877625869286,
 0.011695885677339368,0.7469304225653601,0.012325149873279706,
 0.6894731685064894,0.32281322726926964,0.5729764590201554,0.8624757455560438,
 0.118925338381909,0.2746696093977571,0.18320256306325622,0.01222011823568292,
 0.7814923450298459,0.4861110947509333,0.5954880164992231,0.7672142647745454,
 0.5893245647510733,0.3793358787845874,0.8762591380891745,0.9089888728381752,
 0.2590844844192357,0.6891358653720454,0.25593525060456257,0.6504036727618369,
 0.7946668276209681,0.23600944420541825,0.827485776768509,
 6.472969542138385e-4,0.0343399756452728,0.33136856742352516,
 0.8187298359846903,0.19770193500992195,0.11693516828300798,
 0.6634277690579794,0.33469428057879036,0.6487140050080411,0.4416709441836669,
 0.10456120704450267,0.8174266105211874,0.058569643577376373,
 0.6629190637307243,0.9445545514408804,0.05390640848085615,0.2849575816703772,
 0.8272646078121785,0.8728028371564406,0.4344001505155446,0.6945433708700179,
 0.49857630073969983,0.12806495633090997,0.0417844213648052,
 0.2708295875673673,0.2499527833511238,0.04341559754167834,0.7473908013017845,
 0.6721683944189834,0.7770614317728832,0.4423358394985418,0.6609243887144776,
 0.7905382771978389,0.7687312124232131,0.009133738377180234,
 0.0491745797647799,0.6111994601436785,0.4121529099034663,0.31969237803931383,
 0.6477187298344331,0.42568576845926187,0.056865384179621614,
 0.4753704099766509,0.42683287876678233,0.9485536414650405,
 0.15251003659200424,0.9038122141713518,0.3603950367791957,0.2974786472770572,
 0.9876797266946218,0.973986811394006,0.20629216125961103,0.17889715545249718,
 0.1750024938422614,0.3969002748321382,0.026292668065132863,0.835845886246182,
 0.683672915173313,0.8547008361982341,0.341243992420315,0.988534255613343,
 0.11332754631230224,0.47249380701170485,0.5477973908833236,
 0.15163269337254293,0.5521977160320961,0.7107187949666407,0.9827942414372459,
 0.8316070171178543,0.4909354818674071,0.0044946403213731445,
 0.5813900962418558,0.9241239610125228,0.1273373717587808,0.29836073260035056,
 0.6161487645905965,0.7626025544447276,0.4430554162890751,0.6079982523332346,
 0.06119880489971519,0.5002989898672365,0.07969228289248054,
 0.13609622077745132,0.4242231788196752,0.14567424750228142,0.22294917795654,
 0.5668478386756546,0.12893114213976853,0.4179568296640189,0.7689309421212193,
 0.29591376212900544,0.0019090523867487663,0.04249693222365003,
 0.09492866877777695,0.45113158810579046,0.11707903386258645,
 0.0031851709841078657,0.6757916231032655,0.820623395117634,
 0.4667311936805435,0.6532180238367079,0.6913536889257736,0.7643244159525507,
 0.0595536835682029,0.29571930427838755,0.42266870664528067,0.321207099722582,
 0.8463770795699072,0.9145424109898865,0.04198165558046485,
 0.18800586873016067,0.738064322466881,0.0966221318322118,0.21126386116539686,
 0.6663794428763324,0.4098943233362824,0.44621839529899465,0.676478469485642,
 0.8470603654325504,0.3799963935334546,0.39039670074441457,
 0.043028684723460775,0.898421986847979,0.07877736052031403,
 0.03824641514459082,0.0675981002674032,0.09793901694580343,0.153010662050436,
 0.18254902009065976,0.37768133961330896,0.9331275378426045,
 0.8635853570236166,0.6288712596346516,0.49951514976520384,0.6469211582452384,
 0.9210379201260002,0.3530942604766374,0.21394472506521567,0.8528258808899514,
 0.7954345675829351,0.7125464100616468,0.8490017575133384,0.9464204635180663,
 0.16145497648745777,0.11067385356563475,0.52530961145656,0.2895139863943358,
 0.02342363729888941,0.9146528496499545,0.9267984866795094,
 0.14668727379602597,0.2277550402045352,0.7418753846047343,0.8591762580700464,
 0.3569258737347488,0.8552739529847943,0.3145210701658232,0.7328623125131497,
 0.17994124978303583,0.013942664097023716,0.24227043865313735,
 0.8202299142880053,0.5987504682063559,0.3332373608563731,0.3489374958248501,
 0.8889307968327497,0.14191363119687983,0.6505151358842287,
 0.07822705041768796,0.35590389479971907,0.9806968405428692,
 0.15427608627261413,0.3887180460059585,0.10562381997820491,
 0.39023217308604563,0.5451872306838433,0.13091260487950485,
 0.42698987030457003,0.27016693982534323,0.5996857000664594,
 0.3207049422356165,0.3064483806045276,0.44773499303276365,0.7689370450429139,
 0.784040129611704,0.2664799999809465,0.373817344979231,0.30669517860805096,
 0.4651813742213453,0.6605562172290499,0.9203883740426778,0.4684699769290597,
 0.5059353180678929,0.7332017020341018,0.6361625202158264,0.11170965891129991,
 0.2197295582071972,0.8941021113502386,0.7717032361398406,0.19435454220004023,
 0.0666956637010031,0.210012449079249,0.2978767667746367,0.7885023575393375,
 0.2625526005471769,0.013853030390629573,0.17879680140405985,
 0.3044527600538609,0.9844832490117321,0.8880430864094273,0.28733860887796303,
 0.5662946851902694,0.9948640453586364,0.4623124268661064,0.80865365130717,
 0.025684654798782236,0.5204513565562234,0.5196415423398119,
 0.3770736037571135,0.6395760490228241,0.5818364644738899,0.19431807771676923,
 0.02432202218469326,0.7092871838468131,0.6460299867107756,0.6768265546057273,
 0.28301136192725806,0.11921567377899711,0.7148369299543276,
 0.059742772114100506,0.8125503568832827,0.4244754435104948,
 0.36132047470031525,0.09373136345369737,0.5747547339974233,
 0.04185436647915419,0.057217825048025084,0.018118642474988067,
 0.05677909676276782,0.51742517738437,0.9543204198434008,0.09166984125390332,
 0.2604115557581167,0.9364046573969574,0.07653140849774931,0.4023836276589998,
 0.8262334940461414,0.8174416957403154,0.6226381102289886,0.8920541829940904,
 0.5014499484894931,0.17523745321716044,0.700628967964898,0.16062833080157857,
 0.8345611650934388,0.6674547568613802,0.5608433857705121,
 0.014650863658454139,0.07243574734974734,0.14418299058506368,
 0.11441928698954529,0.5549356621398456,0.3277628383839951,
 0.07075344249959248,0.25492961816382154,0.3343981852544199,
 0.3526978409379353,0.03447213923649439,0.9365492635377458,0.5440522549216789,
 0.4623184016536206,0.4249121675493083,0.045425302899991804,
 0.2980000305681676,0.6538373344626195,0.034018770343741234,
 0.12284112283116244,0.4171375704721306,0.21261135846389867,
 0.9190845298677401,0.27129814788543394,0.353255216072773,0.5946805651932913,
 0.813947802655973,0.2942525694259379,0.3261602963709527,0.8790896547617391,
 0.6406099233641348,0.36010796885354224,0.9569965467765131,0.8470064249964825,
 3.340378464533611e-5,0.3742734843566228,0.012810688448918639,
 0.984917998652839,0.09829570358521511,0.02631624569957003,
 0.13900693686039312,0.28062228858967975,0.6151216504810901,
 0.4229892348209612,0.34183181341319513,0.7113209012633577,0.6979947027803087,
 0.11274338460995437,0.42984012527059623,0.3160603477059357,
 0.8203458583370675,0.42334105690508106,0.29215531229656744,
 0.7068755823609488,0.6037882832138617,0.25987689989218954,0.4307742418212367,
 0.19612690092069796,0.11845515073996182,0.33417260570885265,
 0.9637945232955488,2.3093095326054414e-4,0.3714003766084809,
 0.2885972215436551,0.6864798301067178,0.08100217827240896,0.7397698117930517,
 0.34978355685066087,0.9651096630143625,0.8416855648265614,
 0.40537909918751225,0.7694780440030926,0.8443874226099392,
 0.36307503802063745,0.9260328903469953,0.5192993157875632,0.5949120977321521,
 0.49145774636934836,0.37665553887762826,0.5905494663279163,
 0.05920838599672684,0.502422964619068,0.4815937672214341,0.4091895595841414,
 0.08589410382656326,0.014673211826839339,0.38177800174440535,
 0.6120689761631807,0.5496302046309433,0.008963928128522203,
 0.9512279509461549,0.6342833812814728,0.16670851396255992,0.3505689685050982,
 0.43879102504317224,0.6509817429724073,0.9949230189834289,0.6486646101474305,
 0.9766089428383234,0.8038617870544918,0.3272110321348578,0.9188617964707196,
 0.08125107791031527,0.09276637783065711,0.7497140987801548,
 0.5468199278541422,0.6380262214640965,0.98831052613635,0.8538171223128564,
 0.5210096509614863,0.6477451770652303,0.17346642258847356,
 0.11972762056777497,0.8883127991004937,0.3503613429534288,0.235711052035811,
 0.12510146873463024,0.32360063056834165,0.7808838177566618,
 0.4537610302015458,0.44671777089835285,0.9163063339147646,
 0.006951355557545114,0.3489469754226514,0.9246724522393217,
 0.4439486444379179,0.09322136397717862,0.8075730147732367,0.8058750081701025,
 0.17183589805746668,0.1901176729568893,2.419904055658284e-4,
 0.3402484528380437,0.5128028894048587,0.5640712103900466,0.34363161202696857,
 0.155229872455797,0.15365942488370665,0.2458545709674993,0.36997169990699297,
 0.4897872865577003,0.37789900565597656,0.8080808379453677,
 0.07684421767967853,0.19631726487449486,0.9561192727527934,
 0.6059649679310164,0.21487454463101163,0.6486969893654569,0.3245066985644208,
 0.9885125101759629,0.10564301751081195,0.16002390683196344,
 0.48200548672969223,0.6562177540443845,0.3415257891954975,
 0.03342734520353441,0.1376049929580455,0.4817942272070834,0.728376540106361,
 0.09071219595135671,0.33755079952552336,0.8330607905343798,
 0.10560194210251561,0.461532604980209,0.4325358951249486,0.903084066436084,
 0.013929202792099526,0.6510189913040931,0.8189642458764038,
 0.5580697408125509,0.3640593726797592,0.2376753402658246,0.30182236553376385,
 0.4726772291485547,0.039943235247040465,0.46916599089357947,
 0.10469846056800458,0.24930949056288554,0.9504136282776146,
 0.6880749084536278,0.7343980063659592,0.6043062973668056,
 0.008115831442632615,0.30976786751543417,0.8478487933561325,
 0.9946313261123967,0.3440881269890501,0.960838893623058,0.2688389480991692,
 0.2755392385502434,0.04041202912014141,0.19453560227291877,
 0.9208003772616892,0.19725658205259433,0.7137314639693902,0.8856189285763508,
 0.5779435987266072,0.09812693131245409,0.7249493951476992,
 0.10903569564892168,0.5163582399368503,0.6561960105800473,0.3184151714068575,
 0.6434221480583262,0.5053684591225701,0.1555770695253622,0.36059058203161864,
 0.7525802403361113,0.48516814151674537,0.7933577230570664,
 0.47711001657231644,0.8506218627934389,0.36157685078492363,
 0.4652167050637901,0.07688071655964723,0.984127891093621,0.48091000359007385,
 0.17869926912449852,0.8741138040309829,0.4374902912010241,0.6398983504443357,
 0.8451262239045104,0.28396173190553475,0.29009041307167316,
 0.8902426341389184,0.11284080223322757,0.9478771812164597,0.8817279747738216,
 0.379331283740022,0.005217052673642586,0.7601848038187261,0.3795066717227532,
 0.23778507619042644,0.03213088038933909,0.8151799586754371,
 0.25449130711748413,0.8774386834030132,0.6415002704954981,
 0.14938916772465316,0.9934312328586247,0.9852134832374331,0.8836262959574607,
 0.48905441508948344,0.4459452022880499,0.3125883497358132,0.6324861888752769,
 0.06893466242992363,0.7318240470279684,0.6099230740654265,0.4767315792472394,
 0.2147985407255103,0.379627113236499,0.23562780934460503,0.5079492201895286,
 0.09457344092071984,0.31110499737657393,0.4683931382378901,
 0.21842780825949693,0.7958446836127296,0.7527484658654694,
 0.00808784440583954,0.24122606451409567,0.40525830744997504,
 0.3160830788081421,0.8114061237869648,0.8140276845024875,0.6085346625780111,
 0.8557324259837648,0.4649978949359266,0.9254560505087782,0.209876890348337,
 0.21697816752209176,0.6866994268794168,0.264611691269686,0.6310528967248897,
 0.5683674393505026,0.7649783749074538,0.7502647495998105,0.11009174976936764,
 0.2755434391400917,0.47027002480221114,0.861596285870553,0.845198643209098,
 0.05071236757870701,0.4617035106928309,0.7442608393858593,0.9360791990687823,
 0.7738968111487617,0.9859613627327951,0.8038810595242455,0.5930215476982483,
 0.1255578309701212,0.05233508436033829,0.7530390436108939,0.5090369156096706,
 0.14002658839540982,0.9684136716385074,0.8670522538093227,0.2410248964224243,
 0.4540859079321635,0.6704404613040414,0.9768003576133848,0.618193801913602,
 0.7803427524848009,0.8715103781813776,0.17434467641959184,0.7747760218906088,
 0.8342674993989472,0.8473311997115961,0.8798099147692511,0.15740284544823102,
 0.9099779055632926,0.10367748621778572,0.5572194725785263,0.5431203846189528,
 0.8570274026394267,0.5139126744382125,0.8212087682531384,0.9686430481434741,
 0.3456909468716466,0.00832605708817391,0.04566811963105277,
 0.9728163876232826,0.9106209011394072,0.8066800947447238,0.5394405680760932,
 0.45292186942204005,0.7861611815602775,0.7295004034263732,0.4613174764506318,
 0.988250555674052,0.03553936531603186,0.9931468637001752,0.167144632169684,
 0.5280439075825966,0.2632271354218393,0.9694496551005142,0.14899875338731738,
 0.12583791487867346,0.05799685809608479,0.258345673356005,0.3715376980385946,
 0.1120590803192355,0.9069363391188336,0.6496739544862099,0.7872910670747488,
 0.36625254042082234,0.7091989253285376,0.12874437882376455,
 0.21410770038586213,0.3418894733935751,0.6540550294972064,0.7364119610974672,
 0.5668058332681004,0.8519160291449682,0.043363887822290836,
 0.04588135541866678,0.5203138685252433,0.5900040843969295,
 0.22097527138074147,0.5888756872449372,0.11194933320039502,
 0.35786770796212675,0.3382918400407948,0.11588395045922717,
 0.19773625508263626,0.5627539644448702,0.17787975027700786,
 0.41465578199287567,0.7007627229328477,0.4669942697107432,
 0.10543082838802764,0.9896146171666469,0.45830852064647687,
 0.09597285212724505,0.846527491366529,0.1508775700708962,0.8247758274467032,
 0.022051880167217264,0.35342174402091986,0.17569169016657826,
 0.6271674718285867,0.1842379841647489,0.9226991527659388,0.05164811231466171,
 0.12718609422643645,0.07064830312963233,0.6358377163518998,
 0.5545416294284149,0.5640279626661555,0.38560197844494226,0.8797349324736137,
 0.11388839565632702,0.2463614259129826,0.8470286659159526,0.8052531254023805,
 0.42900661898023085,0.435574365615927,0.5381100910123369,0.35081792605701134,
 0.4774316346381968,0.9672539000142001,0.34853683387129397,0.6277930970727025,
 0.3514215191711788,0.14429691464215444,0.6410016302081218,0.3779491435852079,
 0.9358862720502442,0.33417330357938957,0.17289577654824329,
 0.8955862028161387,0.18511280511585793,0.03464948910294574,
 0.0026741980772826413,0.6975498895320602,0.39231429116635363,
 0.3609070808074438,0.9039876576444328,0.10017612860858116,
 0.14709312112880257,0.5125673167242766,0.8161297600363187,0.4320755813771513,
 0.11074171649377706,0.9572463899795342,0.6872268661917946,0.8906149961316175,
 0.9664966233163931,0.871426355137643,0.3745263025624286,0.11949842315196424,
 0.8595683569814297,0.46711049848269726,0.6912618310399571,0.5765617398167113,
 0.21719517258129284,0.13792564338253266,0.8078263436494436,0.209135263866704,
 0.1371115912110037,0.5399121497666283,0.6136366044298598,0.01465784552396765,
 0.7799710814958143,0.9068897442476681,0.3954104019697666,0.13740099054071342,
 0.3908709711376277,0.4306375462127453,0.3305898410495993,0.9214486009666968,
 0.07373099446999232,0.911756615641518,0.8147178669258447,0.12681566640551756,
 0.22012759846485097,0.8368676446063753,0.24673054477038814,0.818030025139824,
 0.22947407168427847,0.6678479067480866,0.08154846183797015,
 0.27081506099543506,0.6582218637966804,0.6978651494847226,0.576562869036639,
 0.9706280177117854,0.03857310369253897,0.29246967793913536,
 0.3636110309380097,0.4351794006503531,0.6376771360656085,0.02711042403306152,
 0.5934208892803313,0.37635253904209165,0.8258057989265379,0.9963144632560356,
 0.8010933191647774,0.7122288002862868,0.3018547233651966,0.1605746483242949,
 0.0863925551713095,0.7947929144864153,0.15940593870628939,
 0.03812415788491297,0.1772644054139415,0.9944895292600644,0.986709157399815,
 0.8259754607988775,0.4454762932057047,0.23366425390277124,0.576138923578672,
 0.4311428640292865,0.480514141470627,0.7503258353761446,0.6628125905961763,
 0.8699651259929986,0.39677722912894786,0.9386339991368648,
 0.35529445806931514,0.6156877704382909,0.1266527876679,0.3893018025849976,
 0.56883596256629,0.25548199979615194,0.1455411608209547,0.3805237422291181,
 0.30190670691099364,0.07866640871480524,0.7708577775378771,0.769744461570866,
 0.4993429839976433,0.1308241722621858,0.8750359330482769,0.16248659862831571,
 0.8799319740683986,0.9113975849507365,0.4335447154654066,0.8274646117767562,
 0.3443697259659587,0.28432844132140656,0.9990326834281809,0.7675367600595959,
 0.32580235437582994,0.06273774303157731,0.1911901232889448,
 0.35214160689435925,0.6674345875430563,0.9174639550443165,
 0.45760527832913556,0.41637428386875186,0.29778936123427735,
 0.9751235597806525,0.7784829212552685,0.3135870680585189,0.8931944722599416,
 0.990202322374258,0.6779072352636668,0.7478770641652647,0.15197559619853107,
 0.8774401637301358,0.29365569492550403,0.1381973400349319,0.630597280494719,
 0.803511023193381,0.727572169121254,0.3385015269726429,0.693751762258524,
 0.22896845058368065,0.8361486482356835,0.9767679729479524,
 0.40019284921541354,0.41339932706469595,0.14830287579762824,
 0.2417188742302312,0.08062681869662247,0.24426452063219872,
 0.6406564671609345,0.6220643020132087,0.35030215219492755,0.6430085857655158,
 0.5604685203010644,0.9562197135659845,0.20644146005706987,0.5930318737620572,
 0.6816631418907596,0.9281936122895975,0.8704343878107815,0.06346555155858147,
 0.3387894101408926,0.6015919682620792,0.11253439354254113,
 0.018899144753382835,0.22859595433069657,0.3636306172483601,
 0.5174927185827449,0.2609671051570538,0.6749193865555535,0.5494086695391655,
 0.6156978212442205,0.2269444885556965,0.8868291650063975,0.3155198459959724,
 0.1163830707364768,0.2908024737210141,0.5115903292666097,0.6294389227593052,
 0.7983574432786145,0.6113840763014871,0.6458146157110761,0.9793730503435567,
 0.17033118886321774,0.8394069130650849,0.011772998987435557,
 0.9116584943568673,0.41784511218933984,0.2210420579289426,0.7672367028295535,
 0.3645928167056588,0.22778047938604118,0.5929317171795732,0.9200587428422663,
 0.9630439760046456,0.32187317019887096,0.9945930261213176,0.586880811658936,
 0.11174603201378153,0.7106179029194848,0.3750945105524994,0.6519638512939446,
 0.85843230836159,0.8055705910231153,0.5675185813966417,0.6637073696221703,
 0.5113562642383687,0.7984589964117814,0.22837477138865003,0.5086600838301758,
 0.5026912865935591,0.01844171804939576,0.8308194436942643,
 0.24310837334389845,0.9665009404786421,0.1602382421757076,0.2653137811899986,
 0.8212956707362219,0.900181052075291,0.7803276230458784,0.8482120740273945,
 0.3394050056114212,0.17284905408021045,0.24865197774266423,
 0.15383638815814182,0.533087672968712,0.018471001130806686,
 0.007847873816539064,0.574613146250385,0.9183873033934904,
 0.44740339980880495,0.03147775594720803,0.47148923477798566,
 0.901179904202758,0.1406196322814146,0.0632978052056622,0.3977871265703198,
 0.09293677148297386,0.41690075835249596,0.08188450401962488,
 0.560403405856249,0.09958006879787318,0.7619980600235516,0.4839121584227176,
 0.855241378401439,0.8952807695522826,0.9157081749011063,0.6594414700738304,
 0.6380812253565991,0.08971727597900392,0.4745776053216937,0.5614979044397905,
 0.23399775756909869,0.07719147939034932,0.025794611899498165,
 0.19912167507530665,0.1102009711828531,0.9541612442657798,0.8874287315525003,
 0.04104727764753613,0.08159055921280689,0.3166861299520398,
 0.5739063081841391,0.22630151691687517,0.18682373588005097,
 0.4044029059835751,0.4417978862047708,0.14772238788057557,0.9009032914135064,
 0.11252893332653602,0.6258224605954507,0.5766032413543041,0.8574420199999921,
 0.15867256124832063,0.9769201591619056,0.040132415494048446,
 0.4263255781316658,0.4308380903524287,0.7112600093764943,0.825800641366492,
 0.4573391277372776,0.5717495968557769,0.5225944655443182,
 0.026012464279434044,0.17488980863759596,0.11333656970053962,
 0.6034739155047637,0.4567912881388486,0.527988290644579,0.4596511087986799,
 0.1201515589529808,0.8456930771966712,0.18989093691300507,0.9654783195434162,
 0.6552769820836282,0.3263784763104277,0.362439647019026,0.5717823208538779,
 0.13802060464305366,0.2530945518472302,0.8431530364741736,0.6713314576107555,
 0.945206523095266,0.8720607581659768,0.5253352317541632,0.6305807791229948,
 0.31034391134747497,0.672987415521789,0.47891825531907284,0.7666379937571891,
 0.018727032825010093,0.27706240097799195,0.05464627411614631,
 0.9016985875408676,0.29028790810871186,0.7134671697412958,0.5772568461847143,
 0.013471553417241822,0.8027939698584063,0.4754515916066806,
 0.8553547351981603,0.7016722233157812,0.28486606199291054,0.9975048189418572,
 0.757996078943429,0.740352448520348,0.6825710744639732,0.6412653567889247,
 0.37099960971003043,0.2422094242820274,0.9069620993125097,0.5546226610366478,
 0.6576423225867826,0.822916132267989,0.19309355598768296,0.4161391730726248,
 0.8064714809995541,0.9411749701299208,0.7114792054859492,0.9085663460435279,
 0.8711106769801369,0.5877683960558481,0.7620405312248918,0.8354118925120553,
 0.27533748728703267,0.42348538291458526,0.798572738788567,0.7669948321646051,
 0.6204289035564926,0.08563185192932421,0.665231909184817,0.47241588477496876,
 0.061635304488696474,0.11193400029479739,0.8269463902975338,
 0.8500850968735916,0.8320417069402573,0.7112620337023313,0.1259952817095411,
 0.7651963829692618,0.4822861867630249,0.17631974214325996,
 0.13859296632309037,0.4403954714799144,0.08334674785444995,
 0.6650056229670906,0.1706202221347195,0.18691243865883767,0.2545336080874593,
 0.4254861168851293,0.8474401441591924,0.06415637797728102,0.5046291742333959,
 0.4451790746294192,0.7647639444479284,0.7962492038423064,0.21621968309789186,
 0.5851992938058251,0.713456715882413,0.5855048656530573,0.40695952577823624,
 0.2647307406970115,0.27632892299736256,0.6381303287096742,
 0.07543295897701696,0.16818855403905508,0.5940826177201131,
 0.5764647474812531,0.04722061368384001,0.19331646134069314,
 0.6949523519207201,0.9331074690345215,0.28543303154125266,
 0.025360920934898834,0.4076865961264008,0.5778905946436784,
 0.5237118699860727,0.505057584991166,0.45407139882635494,0.9184028357983363,
 0.31550204460965436,0.45104235889330735,0.06245397805537101,
 0.42239232330036147,0.9679268744247753,0.8099252785392111,0.7309536365691948,
 0.5319185678042828,0.37070936726471415,0.8418735238983233,
 0.19105358208500678,0.781639659855722,0.27699822548483777,0.419658732514533,
 0.5255921470594779,0.2897775102991069,0.6935866597335842,0.3124394964413556,
 0.4622678934432184,0.4261270673283193,0.7071552087627907,0.10937402618539493,
 0.9835527526319509,0.08618526552306782,0.9051552868115547,0.9371459852347368,
 0.49878626809519533,0.9193049258602803,0.6551643236527803,0.520040118707987,
 0.8669942024887362,0.5382415179473397,0.6190809385411766,0.7240922634108519,
 0.7919431498482694,0.3541851866342469,0.49600220640034043,0.6601647335158707,
 0.08787221434181358,0.15818409301708758,0.36648076651878636,
 0.4243305365031286,0.018342808386689113,0.044284456812049844,
 0.7934964957120714,0.9658652404186406,0.7322303878590246,0.2911819705301849,
 0.6033503803075878,0.11097028764735839,0.8569495990699204,0.8326630378114095,
 0.7612389897172676,0.43339713654564826,0.8708374259128148,
 0.26832423196458566,0.3171673914150217,0.1174835564018748,
 0.44199295025191476,0.494588027588719,0.80640149827419,0.8783039278381629,
 0.9513428704566699,0.5532627709751492,0.45980089226506926,
 0.03145502850575255,0.187401394646184,0.6437478658969366,0.2514014285560472,
 0.5046901551258943,0.27909847897265716,0.8286065784222539,0.405377691385055,
 0.8841157630477775,0.8059180821451323,0.3528040353234986,0.501281803706174,
 0.7318125977121417,0.5130075526916876,0.5873996565377745,0.8430674798305319,
 0.2758752825167816,0.6041092501477434,0.5643403987087557,0.8086352987088442,
 0.5514360039807972,0.7052817759840491,0.672811178655376,0.42808299426396235,
 0.9531122324725755,0.9973974518664348,0.11018776215142756,0.4587744576417847,
 0.9265951989545639,0.555632570817876,0.8954548490173383,0.8395451399822462,
 0.12823428936398118,0.1799407513661393,0.11476315825907568,
 0.23345600047061144,0.3267903710872503,0.16361294076778465,
 0.07115460785474426,0.33753676943489286,0.7199210308953947,
 0.13915458356830956,0.7145950333709291,0.2700749262094646,0.9632284798601489,
 0.366440081122832,0.44803490462036,0.8981862169396665,0.4028026572303156,
 0.0029175344662695224,0.08971160020011615,0.13772686111424237,
 0.09534565173018594,0.05614752558039504,0.5756972986556205,0.499385891964059,
 0.14898044591332205,0.28507762311665363,0.6572802075183692,
 0.3658293722127982,0.19791971288015853,0.42667523327179635,
 0.7028657614613407,0.8399290934939574,0.7841701003464205,0.4068637358239786,
 0.3908216775884751,0.13885407369704772,0.6784114029093712,
 0.12822941486013606,0.7900719785057364,0.4316113660319332,0.9828416366838126,
 0.7661722979973906,0.6302513473266423,0.9273534205502668,0.3761710814656536,
 0.2229723607958285,0.9286807051775116,0.1078518531042092,0.2992003885686596,
 0.4244487357460205,0.7117457645392817,0.5291962689641785,0.6627684570722966,
 0.2414326077074771,0.3166208148700561,0.3886023336217991,
 0.050161661372424615,0.7485601678404233,0.23161340662568064,
 0.9673666454187081,0.6989719088605065,0.47628971524158303,0.6672563193502031,
 0.3178107854537998,0.13984024032689424,0.13761926346325803,
 0.3847226268601104,0.8290452933312134,0.14230409854591397,0.5524175034017808,
 0.45365744862716795,0.7319421950491338,0.5285491510902538,0.7005225780230991,
 0.7849588142896567,0.27193102145931536,0.17525437337472494,
 0.11289954513399236,0.41314837194863485,0.9796343829640772,
 0.49320717156616734,0.5562884077207685,0.1746226387876949,0.6340193371894701,
 0.7008057247910465,0.4091409047000252,0.867913991845201,0.3824284104187857,
 0.08074721575311772,0.6672555852633656,0.17327359749645543,
 0.16333321827085234,0.7717688999559604,0.5980832229990748,0.9108356014173035,
 0.12585694527745506,0.28178390464622616,0.9805222936360285,
 0.029003415935422217,0.7752380065591828,0.7366037649186413,0.650705965358915,
 0.7860083758465796,0.27525563214424387,0.33563798164574765,
 0.31245898359348234,0.24854520448229311,0.557369998476719,0.6582713713247668,
 0.9416094761097972,0.6809146202524637,0.8858314820313687,0.07982752683207262,
 0.21971119109020298,0.8787666470625096,0.41909342290298246,
 0.07264765986507549,0.6989186561345051,0.03145205616451896,
 0.30793517839433493,0.5369723542399951,0.06875710579137972,
 0.0025135486506562277,0.2830162668427507,0.8493990246662259,
 0.13610793549823397,0.4476196182851908,0.34010071264645847,
 0.48262667232817424,0.17623530200011484,0.948700712140692,0.9675232717448952,
 0.5324681082478665,0.579607204138778,0.9734339670844822,0.7711637710692315,
 0.2780046683622963,0.6453956995196132,0.4997249416438013,0.56598792624123,
 0.649019782368635,0.20842288297673195,0.7579417320004815,0.04667354217967368,
 0.3538264126453884,0.5950604556457585,6.816184107605316e-4,
 0.3911184327429469,0.7233947759984497,0.994533866859358,0.06811626059261577,
 0.4665670080848018,0.8581603011562818,0.7312831467606679,0.5982164665535649,
 0.7404870050330429,0.5762249498706229,0.43799878159355554,0.4865199657882415,
 0.4720378464620495,0.35241487591180476,0.634941919869494,0.19534234314729826,
 0.497853127018436,0.28780663195787115,0.6135888460391382,0.9042687593255921,
 0.6914388638766132,0.657384556042538,0.09351696239785245,0.9868659504918342,
 0.5987181526647951,0.4212197147545219,0.0232299844838203,0.8088649482865953,
 0.20045394124108518,0.4116716465995217,0.25742912744712176,
 0.17331897452625822,0.49160342304309257,0.8419644262061451,
 0.8221638669273145,0.30929003690616186,0.29432747094822065,
 0.6063081533659894,0.4149326163104974,0.7929750339504216,0.35063880276749826,
 0.39948495644573,0.26153738073234345,0.3393979436453227,0.4695699200187199,
 0.7441596673466434,0.4276046263774782,0.06779120859598553,0.918428890649344,
 0.3532814627993117,0.777117063961115,0.09366677982266003,0.32390565571667307,
 0.8117489876096489,0.8334259593661094,0.18962273536125307,0.459478750850859,
 0.22927425206645813,0.036240339413069145,0.04961479944008251,
 0.2800896824633434,0.3365158926937719,0.06931731661093965,0.6585188354419844,
 0.17393499454060246,0.5591629234642792,0.9700176532459567,0.6219829411238393,
 0.8206770128267098,0.8708680259084478,0.06400399608315399,0.7652961296251427,
 0.04391880002198678,0.27286722049774215,0.5988143476420437,
 0.43455193247354595,0.5794856843148972,0.7681261013545863,
 0.13048184317641176,0.2195019342305614,0.8743522232013883,0.7158158704695624,
 0.6281722033917411,0.49932462074226014,0.2069689846314634,0.5656994137018377,
 0.052570769908804715,0.6758638774662641,0.46780727117397547,
 0.6970945683551637,0.7053130383731054,0.0936096407487148,0.39932507152237406,
 0.40407373278432934,0.5215403397675082,0.2598715669035341,0.4411687901166952,
 0.022962895068833866,0.4697391324596991,0.7256397058381843,
 0.2162926861745481,0.4604751723389604,0.17139975130593343,
 0.35203244939747247,0.6331791849635475,0.265121416505274,0.8700244337864336,
 0.31166042085457635,0.8836552695973506,0.8710070340598721,
 0.29605505377009056,0.1986904751107904,0.17742641879804977,
 0.3003359302817933,0.11262663533397277,0.6939961300880292,0.7208176253151647,
 0.15157036852913863,0.16387838967760326,0.2287652367811922,
 0.9945967470028334,0.9673917415239467,0.5979326103857949,0.8111836703850718,
 0.8802069302156794,0.09337529331550609,0.45727727192548506,
 0.05992637275316415,0.9824752258180485,0.6313209078108382,
 0.45224861446967335,0.21016912708072766,0.19228405928538939,
 0.33548731249765795,0.7672753288732485,0.9836648258970908,
 0.012566686676894978,2.8225598491349313e-4,0.040908233360964275,
 0.5076719302719941,0.6997589842678414,0.8331547994114303,0.5379212285253174,
 0.9746145622679369,0.6256768151646068,0.0761198474607937,0.5977656165801384,
 0.281895303548257,0.16404607315559017,0.5749173997713363,0.22427886403170183,
 0.06631423804846925,0.07738743004360327,0.6246329325718543,
 0.4619419826488169,0.9767402654271803,0.6103769242127957,0.22917114417692508,
 0.0937609681223378,0.4039299528249174,0.32141017625822366,
 0.36321586276136586,0.11998712801670053,0.74828971939125,0.27793485592345646,
 0.02177204669246302,0.23924684849199207,0.5028949207465959,
 0.6162905011422397,0.19903735995487748,0.025182550751455812,
 0.9627288567766444,0.10819956013762666,0.5167733493294285,0.6694456031027355,
 0.5862996356995994,0.8564828869242561,0.20290885003289916,0.9216031210200235,
 0.3270553664986183,0.9467013620853014,0.45260949177089316,
 0.48196944393208796,0.6170063877424645,0.9802794093520435,0.6735804065687029,
 0.736018097043064,0.9295750572617604,0.03333995053589356,
 0.033371265032038244,0.03270051070310698,0.6796948703045146,
 0.028653285664833827,0.1330823185635811,0.49456285295141245,
 0.4189583917337646,0.8745858290164233,0.6015325095290648,0.4488512240545447,
 0.08463956607297218,0.17288308568970479,0.1869103079111062,
 0.10167537234378332,0.2377303719623265,0.6767083324229649,0.856730199578645,
 0.7720064906520124,0.8853104195462941,0.9390047956599487,0.3686910504195844,
 0.4486382043477005,0.6516424441998294,0.18367684183764166,0.6809186505503575,
 0.481572639884255,0.29983608223932934,0.3852810634391802,0.642358184838733,
 0.9761338376394333,0.8402290347388162,0.7304597963570384,0.562212004072615,
 0.26849478059760645,0.11564700583987986,0.5873176213456893,0.520440992380658,
 0.7928080750011762,0.5952229429070761,0.4846942545634746,0.6764751599936993,
 0.19785393982531163,0.3876305794762843,0.1740409726804505,0.7244818771626895,
 0.16984843796600735,0.027861171658021933,0.34919813132894595,
 0.6676741980365791,0.2118031976083391,0.6781581691993337,
 0.026102759791020058,0.32013026022380087,0.22653212496172925,
 0.950675846570112,0.5912316462346227,0.45700832466877017,0.16417086930030833,
 0.6181387905106572,0.1036493045409046,0.09107459104076165,0.1796987802678538,
 0.484466701711986,0.971888310474149,0.9810541546518978,0.40808873359461884,
 0.2881490130714026,0.3233061762801147,0.5962248487360213,0.825883797505131,
 0.7905737646354606,0.29622335984444736,0.09094383272845785,
 0.3416211596046249,0.19508955346332646,0.46695565098646863,
 0.21152223575785223,0.17330377152109422,0.10974798420854315,
 0.2682886185809744,0.08622956146140526,0.17955311142643782,
 0.3309924294321145,0.5323242497024558,0.2328188304026646,0.5393764484518829,
 0.5446827799236005,0.5859574750511359,0.8606026161466703,0.882748637072662,
 0.5264686107093182,0.2171084222489078,0.2023267784752958,0.69793514278844,
 0.7483204134630063,0.9846312437278131,0.5444698959707027,0.7156101848957208,
 0.8991738229407356,0.07325636075864073,0.90324728523585,0.9123908006275864,
 0.90278778947513,0.2839346701357761,0.37747511594315175,0.48157190010526174,
 0.12668523899351736,0.6770897531637173,0.6028146007577808,0.2101024370671497,
 0.12400717402106132,0.8387451147973203,0.3118046559179496,
 0.27584389046503766,0.4050690397970671,0.597987967189528,0.9952770041464846,
 0.8155518148567071,0.7987775626391898,0.5183323547621923,0.6264514495161805,
 0.8487579013014502,0.8141425820397497,0.8755740769599616,0.626422552625465,
 0.7459733277285463,0.9300763782420622,0.7593749748712418,0.7552725845887175,
 0.8274156991704854,0.19802537517096075,0.42940849936641934,
 0.2558106387174388,0.28704890160686447,0.6096233693200187,0.2731833444805889,
 0.7364421539725964,0.7631673204080394,0.7517791418657058,0.06094312442289651,
 0.3821349465906497,0.9207531913194122,0.6801037977669313,0.1613283978280532,
 0.3972743304894971,0.37207960531725426,0.42995777205757035,
 0.9575827128925444,0.4958680569741273,0.9281635829324244,0.7537449915595693,
 0.8536396600776237,0.8163785700738666,0.9345985218139864,0.554959692195925,
 0.6833321671922057,0.7534080664423298,0.9798383061551594,
 0.056195679435650625,0.2598612058175436,0.310903145290768,0.9166472843743633,
 0.41878973621760873,0.13651713105580177,0.444365422922702,
 0.44422614621315226,0.37609865312536384,0.910551839209944,0.7100955363873677,
 0.2634064572948487,0.6040323748614821,0.9556245021711854,0.0959582657651673,
 0.09953514640464145,0.8720133122615388,0.6201624099125223,
 0.40025611497782965,0.05448107313905848,0.015115916785358952,
 0.38592582709139267,0.14350852898458,0.45641761707018624,0.5822489661757487,
 0.9923820600451202,0.5706926724577863,0.88568403735533,0.9783592382439508,
 0.2506908395356561,0.47259814849147075,0.8859229076341695,
 0.33018203841750005,0.5413461093103638,0.40715448932179876,
 0.3931971230267616,0.11893950949459953,0.5671968264028309,0.865125118114249,
 0.09546318450058722,0.6247191574811816,0.5874171546194211,0.9782245858857769,
 0.7364442920243739,0.3374098765244876,0.3316893116712851,0.13761820211158582,
 0.641153404650187,0.9691643910646089,0.20829948854640534,0.9340865521516243,
 0.03198986632676026,0.2357449863703538,0.5220984202322809,0.1690707823074835,
 0.2726296307333498,0.061136708988305255,0.8387656355654695,
 0.7017917457898234,0.030147835221998376,0.7910629350949576,
 0.7762423938010381,0.6377132672889518,0.5173986424794769,0.16005938902239303,
 0.46466491302066093,0.7096178861473813,0.6320691532254049,0.2466811560354707,
 0.779322160083713,0.5008119290823428,0.11626658470563034,0.1898632951456033,
 0.6907124244337077,0.7161866844141964,0.32366640123244717,0.5970236731836375,
 0.48859067325409844,0.26241058097254577,0.4569126321946688,
 0.8046985862001775,0.9427675971458912,0.7192367034041027,0.9498253443783009,
 0.1019700796355495,0.3834372385738487,0.04630263999055095,0.7849468287395861,
 0.5750381735903317,0.6428400978581974,0.4353094263313517,0.6634874233779335,
 0.2701132163601012,0.08084972725901907,0.183771503717017,0.9719877167147486,
 0.9187120634914204,0.7392846276332451,0.6662235560641953,0.35943082907052615,
 0.9621852075160865,0.04780771058208344,0.6153528482787801,0.6198008494390448,
 0.8455851827539832,0.18079791737645334,0.7264486982066463,0.7424125589363477,
 0.6409039663235953,0.6294408118555704,0.3984430608014272,0.7474098537872043,
 0.013876617231995847,0.9077158054740485,0.45629505268407977,
 0.9485620292532618,0.4358329951944587,0.9720895003231801,0.6892124701145697,
 0.006021302081049118,0.5450348333888065,0.3761946115909294,
 0.7496481523292343,0.6290374385587396,0.13367338075540325,0.6012094956979643,
 0.6948569679446791,0.1927549152767304,0.4284054990125046,0.9597009419290615,
 0.37800934025989186,0.049594341095533845,0.9042892310088391,
 0.9844121645577878,0.42659081732654325,0.46716004766445507,
 0.010361770129435954,0.724536841351811,0.4406576549755118,0.8897887684320474,
 0.6264597135366161,0.4263643563657681,0.4035949137373095,0.26527109573149454,
 0.49662849090552774,0.48397134616982007,0.2525565248955408,
 0.4819154451439712,0.46261234591941713,0.2667420720189696,
 0.08277457099693897,0.8430461141270802,0.6039857957733586,0.7473970848053602,
 0.03269925393259454,0.6769944352551791,0.7953779678210682,
 0.17959390852209955,0.014697964194844593,0.9530705406738723,
 0.09176468273715388,0.7969347231483994,0.6411073618554018,0.7042775297620955,
 0.3499374602291755,0.9046844289179345,0.3028137383545103,0.9190204659874193,
 0.6250430531975496,0.6108731493591406,0.2103125630499566,0.9416240478835078,
 0.08979498831453603,0.5699309149739177,0.03412879089231091,
 0.8226287819308051,0.8270837614437777,0.37526456409514164,
 0.27249550144223766,0.2785454890455463,0.5182301769856597,0.3895893144780487,
 0.4740430019489149,0.6691225329913917,0.49231729339075947,0.7135484739165658,
 0.8791523870585936,0.0855726330966815,0.306346739225832,0.9179407640602762,
 0.13428954865022968,0.7393690260151387,0.5170351707002232,0.6427474556605558,
 0.6221330046574418,0.7973923977508579,0.6454318612812787,0.36435208141796593,
 0.9229171572042408,0.37626094944732735,0.5627349706188379,0.3850379128555379,
 0.7168992799706416,0.6535885543743714,0.8402325657978459,0.4396838711506853,
 0.1118588371754573,0.31295475698737085,0.1314997608775471,0.3259340991002011,
 0.4161710407903453,0.21138436668887794,0.6315295576875464,0.3152777183100144,
 0.1500259509677957,0.9317004114527917,0.9408167694155027,0.6127918885315436,
 0.6052835228777338,0.9383744442316215,0.5602309111165296,0.8580316523915283,
 0.5175208227332105,0.31973052267055513,0.09228631337922433,
 0.03253329831502372,0.13969262885515332,0.4389177245631256,
 0.1207007926192517,0.02910512087196193,0.8972256748223095,
 0.35472225913647193,0.22876052814625503,0.16241339177544045,
 0.09643116792099737,0.3819616828592757,0.688303342633377,0.8358396641259871,
 0.1011210658047117,0.8258456287600606,0.03922535180907305,0.6459190546857692,
 0.7693433843505865,0.1909585094258176,0.4396280676909833,0.338840784250776,
 0.713028621338669,0.7018097951587479,0.3859770321401528,0.9773512247995452,
 0.5041931334753271,0.9055434483013259,0.13915686447208198,0.6411314988915366,
 0.1262317593883724,0.9491634344751685,0.04246730807410737,
 0.12265650916957294,0.9005716842934409,0.18815757305022396,
 0.8269091062646112,0.34084947899286366,0.8640587782398912,0.8958150725073392,
 0.5077944976318183,0.7936477742322445,0.05802066194572908,
 0.07625480026389653,0.013516861321863916,0.7641010585248754,
 0.1716797383559857,0.9542217846847609,0.5495632488752761,0.6388760175720116,
 0.37160795578066175,0.9582518999694819,0.7991239314371613,0.5809582966736888,
 0.23264902082097905,0.29046771188655807,0.3783325446493091,
 0.9116925512989679,0.35915555157548174,0.6041494715748463,0.6622090890253658,
 0.6971587870756141,0.989987779555841,0.46469421037156833,0.10964752825842927,
 0.0439721832512493,0.5120597680040705,0.704992882620981,0.45139168989818,
 0.34681361753540885,0.7039524323441084,0.9597630325034439,0.9111858869665798,
 0.21179953817393904,0.03831077581149356,0.39952311883961644,
 0.15259357701027587,1.904905926800593e-4,0.025974595666688027,
 0.6237898380010737,0.9843104956320824,0.9654245021066249,0.9458311454310372,
 0.8039725738847623,0.2033813891066133,0.8091356802183529,0.22276508811388263,
 0.07609391444331937,0.7553725998968405,0.4326303058509412,
 0.24861820723946626,0.6619261024040158,0.9436666144963473,0.8303193801876207,
 0.9962374660531945,0.3442725345351805,0.9341870901229044,0.5052565203730143,
 0.9062703668473956,0.4132792336202341,0.6272759676727199,0.8266435973426317,
 0.39541232492736067,0.608776450753228,0.4057664394495344,0.6920282081023653,
 0.8821453803941257,0.43788122602482593,0.4673955619141089,0.9297919351575152,
 0.041213573633043676,0.9838025740544699,0.05617279053937929,
 0.1143824663794133,0.14742954759926197,0.7743824556043957,
 0.27441852061725625,0.9219400421647547,0.7471529039441034,
 0.36864991496315713,0.5210292229971236,0.9752318680601467,0.4683157390736915,
 0.8975080161644047,0.09068187822648599,0.1928108941859833,
 0.22610740105789584,0.8225609872534498,0.14307497085645837,
 0.15282537297652388,0.9503623060312696,0.4574860480657742,0.9534119703408521,
 0.6571506834803114,0.8996655631241112,0.09759377467834174,0.568918016704435,
 0.7825765561386318,0.2830057673960541,0.713581617529597,0.8833615006293158,
 0.7697067657061414,0.2932139275885014,0.8706854719617908,0.10514973206506195,
 0.19455584238682289,0.5540174067315804,0.5000416993083711,
 0.12052278490335877,0.11007017872106462,0.7187146836560285,
 0.9227126725089543,0.018873031270342566,0.17126422794318596,
 0.1275339391234609,0.19893098225851436,0.6556708163821384,0.6682124540745834,
 0.10654884568454048,0.8255500300999975,0.4889501864684316,
 0.30737945090471785,0.5536841687168172,0.22638394075332724,
 0.7923833852017421,0.4737503246871412,0.02900705218284949,0.4216558639416833,
 0.857052515660663,0.7179169408441345,0.19514624331749375,0.41776030207250936,
 0.9923873808189919,0.18010843065547233,0.4720445201109944,0.6105137802635756,
 0.9161581690668515,0.8583261748281998,0.1073212394272196,0.04579523001084507,
 0.5192703715913588,0.13184414473469885,0.5548412138014969,0.553976087283869,
 0.5884317560101956,0.9204758717494936,0.31107060015699284,
 0.02854559075984231,0.9869570530415003,0.6686932918899622,0.9404969639361558,
 0.7561781000422089,0.3329370516338753,0.5669894163309678,0.43387114230039425,
 0.9768485234201696,0.9879929151883927,0.6484336136514913,0.08328616071312633,
 0.759276148443091,0.47344952247442573,0.6765165202758165,0.17626846599870094,
 0.8867125717515099,0.8120276921507257,0.7138740418996313,0.9760271540165185,
 0.24230056325821414,0.23461693217123458,0.007691334932542704,
 0.19231020730805992,0.8903067176477418,0.5317637609549238,
 0.17493881706699355,0.8928831930844955,0.4710814389681106,
 0.20993629094843336,0.7836329228095851,0.21886217263406893,
 0.2294611082365925,0.16106772601800867,0.01661051627286858,
 0.0016999484079115446,0.20882446123367449,0.5153628246694473,
 0.6011768965571466,0.31118755748195037,0.29395343299424814,
 0.14028135115741325,0.9313851909843649,0.4407431824993252,0.3154726272723294,
 0.9026716011630677,0.7879520503681999,0.7099004810553398,0.5538274100353117,
 0.44241815944363116,0.3054979765564736,0.6935317921448099,0.8855910354237159,
 0.7252341745352159,0.42350302767762504,0.048835717487090924,
 0.7123815101070691,0.8519840373715073,0.344849306951999,0.14081221061694593,
 0.9484803163496452,0.05505098508547557,0.20058066985988066,
 0.8240898652004338,0.32891855593949204,0.9740190724008868,0.7308880057222233,
 0.5186567434369197,0.5640998490657431,0.1861834385347848,0.106718378129802,
 0.18023847110632296,0.4529923450302965,0.4715712760081473,0.7975217618512264,
 0.1684571881581638,0.19944303153588017,0.5332877780748692,
 0.008556811030851263,0.46014869290397475,0.39097930462257424,
 0.14087477353697853,0.6325099632655667,0.8010592874113242,0.2413370173592383,
 0.9408583459936174,0.8042533520799374,0.32076186785750993,0.8380738164143743,
 0.6959203340684437,0.8806214236047987,0.33711780587458073,0.4001995635514317,
 0.24204902285123908,0.7025888516307899,0.20484864492918564,
 0.8091189322918124,0.740802096341743,0.5391026824612123,0.7234340484963588,
 0.28906396459710015,0.619253381443148,0.8956696226426131,0.9425953394768423,
 0.37761958832018183,0.15436937246661242,0.9601165826779318,
 0.057591247151066116,0.4889638088200451,0.6225745574626731,
 0.6713597821813787,0.9723898733100882,0.36734270719651774,0.6146413784149561,
 0.6370606814796621,0.34807143329418067,0.6942051174306961,0.8544390919920144,
 0.9197220474545597,0.4102952120118073,0.9839765596130888,0.7803834963033109,
 0.33950059421815093,0.6722084893590129,0.200142339420051,0.5297793302302525,
 0.42522923870445595,0.06997343035701942,0.9162477697302808,
 0.7308490789180244,0.14570136300454117,0.7069734101613906,
 0.09487389846425065,0.8491118215844934,0.8733534254893481,0.661241001306917,
 0.47042242229655185,0.7789347977007601,0.9612037096194819,0.7046848429915243,
 0.3272841843347951,0.7752555642676627,0.27374272025276536,
 0.48919703876391785,0.7703101200321762,0.313982047328318,0.7163905572084212,
 0.8245532137742435,0.5617176237654509,0.7417827476168009,0.2001544549179305,
 0.7772301655528852,0.5335840404975822,0.8346728338751067,0.25651615483717505,
 0.5779184164808457,0.7574607157353667,0.5505149503887304,0.511155383833617,
 0.6892346233033857,0.4348116630729746,0.381231272877003,0.3931311017257595,
 0.8352895128204121,0.8136887065536001,0.9575620276474479,0.8101225941213519,
 1.285636740919749e-4,0.7633400306900293,0.44526337049065623,
 0.7131307961484772,0.5261421139027365,0.5859227830776241,0.0735616377349978,
 0.17748953690890568,0.6386982668380173,0.5642369318169294,
 0.40263174793703693,0.42388124764172364,0.379175264363521,0.3605569997236012,
 0.6618418708233754,0.42543565543052475,0.490911411159039,0.32437786222795484,
 0.9230632025260792,0.48776017252234716,0.7257054035700082,0.9511054203876628,
 0.8458182366040118,0.3263266445146734,0.1574799921144392,0.900388282018734,
 0.7706418102076024,0.6093641500243535,0.9862845126989225,0.9547991246713493,
 0.29161751168180505,0.807037637294344,0.3999156278154834,0.23259130493982394,
 0.9984300782039074,0.7192810147204615,0.4853952738654903,0.3826126421944346,
 0.5807700277612466,0.6110938405973055,0.04878020648486425,0.5559481544601829,
 0.6523568478521956,0.804883969903647,0.08308565946335444,0.4779742062112623,
 0.40176024669341903,0.5403864611865647,0.46528350766048265,
 0.3876338967909251,0.8566408656777107,0.8952276754279133,0.607251129660289,
 0.9313786155162858,0.3967976708317291,0.2876102714401143,
 0.002348290884075288,0.08631415772658846,0.8107408562554298,
 0.9842386498058752,0.41045172401382934,0.49523808606577824,
 0.8978985061558079,0.8803179569558656,0.12729203593319227,0.8597608095444766,
 0.9534185815739484,0.5733922580871738,0.7996485023381188,0.5955242367769747,
 0.46370513016978243,0.17176884401648707,0.6254811674960964,
 0.2002617345700095,0.43973377964162275,0.572957201130153,0.5520358959800372,
 0.4835322257497311,0.7629963123943795,0.09622682547296157,
 0.07604284508864989,0.4132545102106573,0.08964644071773875,
 0.6830339818362557,0.6254794554173494,0.4735416820933972,0.49093278489637626,
 0.5633245826817761,0.36371252858250414,0.9526768782899826,0.6563657337542871,
 0.7938714573182504,0.08627142645299124,0.7865942662415604,
 0.19914581443573987,0.9516209023532181,0.9500160172433386,0.7910845671934881,
 0.9943676107199055,0.8798824877984424,0.8773112046042497,0.5050628715514058,
 0.4489551621278085,0.9133828668677244,0.28972320374589544,0.2770293910913766,
 0.4726022517777476,0.5779140151979242,0.7974207614758675,0.6995369328285623,
 0.2992324428945491,0.40293781954108543,0.1995508750232481,0.1580869800512814,
 0.8894105907107086,0.7949410059224076,0.31817792540083867,0.4848952689103937,
 0.48739193035599837,0.7891579982647909,0.6773356768783132,
 0.25671103380268145,0.8685023417427217,0.09164229699545001,
 0.8444881535218758,0.44439704853357187,0.006128067849485386,
 0.4581714209137405,0.24765202882778536,0.4537765526521649,0.165417494953908,
 0.4881720247569715,0.28418645913738494,0.9908329066794844,0.809362916004688,
 0.3174209640238479,0.8327306416752407,0.8504459582223237,0.25637072538257966,
 0.44373496708662663,0.8541467128665508,0.7289104510516681,
 0.33292850576969646,0.1460258513110656,0.3871049682704485,
 0.009790287038372258,0.6781910828535125,0.11352772145284007,
 0.5025969297503098,0.03778085938858888,0.6844581637499701,0.6284939794002152,
 0.8558628443836805,0.6450724442664605,0.016729795371399847,
 0.6939553481837639,0.5624786755583385,0.5581397779769794,0.7717096812191276,
 0.5400416300010017,0.664077151517321,0.28066500513483117,
 0.003949385206308609,0.5181413171519725,0.4147355171287769,0.693148190867454,
 0.9364229246142622,0.11977269131413038,0.8415757636345347,0.604975825683709,
 0.7348503568871246,0.8950466138254737,0.22569678580527275,
 0.06144389199853384,0.3870415272470653,0.03626757368781708,
 0.9713188504946857,0.7338791820642586,0.6049267283796064,0.39765345765669957,
 0.5268910767004507,0.8605869992121137,0.40908141874534765,0.3941758902078394,
 0.9047604445958336,0.004091965696965794,0.3083402757542868,
 0.8173325011714326,0.9351978236361945,0.3425455024877576,0.39742901747287085,
 0.5332627087247276,0.5223524149410079,0.6081717546290243,0.5156839407247434,
 0.2817187526234597,0.5138170043035898,0.22567144179091203,
 0.003628083708847263,0.2781506494648347,0.6129993611442186,
 0.9659107581440441,0.731646183236168,0.03602261032983256,0.2290157577940195,
 0.5350968139278282,0.43513372807673345,0.4544463237372546,0.848040782233289,
 0.8345228309483718,0.8380005540203326,0.09893494477687559,0.6073871919513871,
 0.38750803621281904,0.2257873285294515,0.965404015134542,0.5958501206846415,
 0.4264209646861149,0.9284376344688737,0.9014222085444504,0.6425153167554696,
 0.6778873170271631,0.34339920769790555,0.8705842832805757,0.5869782325758206,
 0.03081670768972189,0.5662232626675731,0.8377410280192173,
 0.20303298291403293,0.25925292692859636,0.21506455931449564,
 0.15887858015904655,0.5821134766192287,0.016009632494741743,
 0.7884069527486014,0.22460515542293003,0.5837156111420007,0.5545740187277712,
 0.21390862570707414,0.5868181600615938,0.46492151218124944,
 0.8319395321183426,0.8963855091369853,0.17544248015035158,0.7493630259161471,
 0.20879074224685823,0.19533933863874808,0.29973362315904284,
 0.37057465776821286,0.2944050561146073,0.3957615970138433,
 0.22097273455725097,0.12584332238304086,0.9129811595438814,
 0.30779476443021736,0.06109238505675152,0.4037361980441494,
 0.5864621362219856,0.8044855021097018,0.1444828526096179,0.3141170010843105,
 0.6683455507530017,0.4746527183115169,0.8821223877777399,0.1406533166935081,
 0.3130281861524933,0.47495932369014804,0.1869510241917156,0.6602045444607124,
 0.5842314014876421,0.7320103317457409,0.5512901644884848,0.937257351340416,
 0.45061183703926044,0.8975733863362902,0.323875108103471,0.504679834476787,
 0.8493142443518094,0.47255454786422824,0.0934005554749715,
 0.19043082739017336,0.778887104740885,0.8295518720901468,0.7240746441181306,
 0.03202287349942068,0.7103667820489712,0.8117235650843259,
 0.44197194260758565,0.7145585107705893,0.516550174306868,0.30420104046896035,
 0.5841826581050067,0.3969111413105231,0.6893569485624745,0.4128257181687385,
 0.08902849381693567,0.7097959565000747,0.8110150534274227,0.7504866070043299,
 0.12209625873783403,0.37407427515554037,0.5303661604060607,
 0.9362872906227498,0.37579284012988723,0.09811731975593263,
 0.1609311637222085,0.7922144972416032,0.5098902375111162,0.7718766188962028,
 0.46254889027375623,0.17698047630760538,0.08322180555154857,
 0.7015407059092565,0.023235271187786677,0.014188414975450203,
 0.5381770124542764,0.6580371215508163,0.28003999379870226,0.7331633575033858,
 0.7329262806700949,0.9815265541763716,0.5673927011811262,
 0.0036282179144632165,0.2305104145867376,0.10995780361214458,
 0.31631541401491026,0.4315494369951747,0.348570912083084,0.07935406433086678,
 0.3957806133203272,0.33887380726611793,0.4875496265286926,0.8087515399062793,
 0.8740088626451895,0.5649320002371251,0.04211912695109121,
 0.23498621345722293,0.985332420350751,0.8565061475459728,0.23851853827871494,
 0.5033258162787844,0.5228711269681314,0.7152952406242624,0.9771516317353487,
 0.5499427715906176,0.622817696707032,0.47898135700172617,0.4628880500600723,
 0.696373722631728,0.5380564511694622,0.9684385091877715,0.34620522330959735,
 0.9754223734612929,0.6118261256156571,0.986412806724267,0.10288852263876458,
 0.10961866803082787,0.843429689452282,0.3336452901880286,0.40426934454685703,
 0.2790572554259032,0.6067376478170423,0.31705594938734005,0.6974192103298115,
 0.5002709254540963,0.8474335841632882,0.7970827272497755,0.3251106655477227,
 0.016831199244973893,0.4821456052217592,0.10929918253830473,
 0.3710012228371895,0.27023935895327456,0.6418217053171134,0.874527076728798,
 0.9250607898794334,0.6928846779543645,0.8483255408979664,0.40967216929851713,
 0.37899531906522077,0.2217001136416511,0.6666322597131753,0.9991518649720447,
 0.024898999928706256,0.8139981699220231,0.5115454994697473,
 0.5909158573765033,0.21565869690344086,0.6312456892856909,0.7639109214910891,
 0.7823334872622995,0.1709173578736135,0.40093973084456225,
 0.0013978906400837765,0.38482778972720566,0.2283582414623878,
 0.9690792769258817,0.7501011168459363,0.2515124628731149,0.9651386063734588,
 0.9411201218323921,0.8132542957733011,0.35434992123140274,0.3847399829508541,
 0.6733950500485939,0.17451175710264266,0.5863560719995371,
 0.17321603703938837,0.15199352137253563,0.05661651895111619,
 0.6006658446666614,0.3117827105442831,0.6841411976181125,0.9486664759051898,
 0.5260007633215962,0.14119080260748507,0.5154959706566187,0.2392870951514947,
 0.5651819122307303,0.4012946548005877,0.22349181586989475,0.7587755657113954,
 0.22250786178716409,0.5952036641481042,0.4887236114429334,
 0.031140029416181325,0.9286615826965621,0.3766821317880672,
 0.8884219306863583,0.10429367945356183,0.27984325890706196,
 0.9247760159180688,0.7895224506219403,0.3566845081352248,0.5617011325326073,
 0.07618005892393154,0.8278658084641111,0.3715223746662635,0.8228476730215563,
 0.7479138776195939,0.4448179588059429,0.8267215611796046,0.5690425284184304,
 0.4423751732284724,0.447736543554921,0.7306114069935777,0.40479047267983315,
 0.29659375756549144,0.0928935496278761,0.8321468381553205,
 0.32466605082803635,0.2665023392558149,0.44347992934810554,
 0.6518701956016372,0.732467142073636,0.8013260847408579,0.4096942348610365,
 0.8648004661525719,0.19745119332723926,0.7637845770845797,0.5710801764662639,
 0.7283835354237668,0.5721357717784518,0.16114411501553594,0.0726980893430974,
 0.9339562219006502,0.5278759058321276,0.5300206648104795,0.43268417025397987,
 0.07590024134809059,0.992219701176994,0.2627971581857548,0.9509606534018387,
 0.6907157586876411,0.7905099453193012,0.30016006330008416,
 0.29846703978984457,0.2678800073963927,0.33944748872493236,
 0.6680147553236563,0.8149930666501104,0.6460567916222923,0.37732256812485154,
 0.5311621718466584,0.259723072020167,0.4820614488030447,0.5485323560071232,
 0.4589072255819412,0.12185726846267686,0.2197313124052671,
 0.44835808798687893,0.9574506519752216,0.5744892758328903,0.8782167610632124,
 0.37055670116598094,0.12922094021877484,0.9646610413200426,
 0.7642318412757263,0.5371161081202339,0.3736897101292028,0.3248518413624133,
 0.40467914728385457,0.4269900921423303,0.727398839649408,0.9483265514530869,
 0.9003285457707213,0.24505983018788402,0.614553827643532,0.5697355868580446,
 0.17162341343739163,0.44181868120906576,0.384354532620407,
 0.28449049303077834,0.8389526383027437,0.7594644955820304,0.5817547014264234,
 0.21993784357763324,0.8853559576855085,0.702128599819895,0.6645969243538287,
 0.6650604276178523,0.44771492920308753,0.4589936337308581,0.5689095329776935,
 0.04757178536943352,0.20739404808370854,0.3862506036265936,
 0.19701090832363632,0.22611611232495,0.30217661433626075,0.861876567432653,
 0.2438287156640626,0.9523572695473872,0.10590424634541551,
 0.07208412888905391,0.8338012521318952,0.40649839637288043,
 0.6771030526282552,0.1854735256251675,0.9965402636337113,0.5977037324698831,
 0.9511570896172727,0.11125287735865996,0.7516660717043814,
 0.025822309321802517,0.3112786342298861,0.18027107655739705,
 0.4891310239703185,0.06193326519387932,0.6388972591725834,0.7008228117799524,
 0.7897399207365322,0.12044874724930976,0.5886150774583798,0.9887375025227401,
 0.5901043754113036,0.8203787686148446,0.8686918685710404,0.6739684078913165,
 0.3173068475914944,0.41824684815081825,0.8897382049059899,0.5024447410229963,
 0.8540864864769125,0.13240724778096624,0.44451061045797746,
 0.8322931250962051,0.37201826425679985,0.05104048214985324,
 0.03010065705248066,0.30201271113798434,0.5276811018506001,
 0.08467048953870959,0.2882615941680331,0.4467481467560429,0.8843334811797836,
 0.07068476201086837,0.2307602917228766,0.3311996083073402,
 0.14782937637758853,0.6437629870081403,0.7567324485955095,
 0.46465824514798704,0.8979169840950749,0.911494015227173,0.8710737771908009,
 0.2295887453624007,0.324288492718354,0.9274160706948927,0.2587223128969526,
 0.9164189581955884,0.8269277136336869,0.6344092490875332,0.1369623020981532,
 0.5279254612301916,0.5167665059365443,0.6284224786493713,0.9626741488734866,
 0.664833313761299,0.012293230555683543,0.10162941806969905,
 0.8041991511916787,0.12294299221876548,0.4958646667074119,0.5985061270784824,
 0.7331754999099735,0.8956283133026662,0.870965290696963,0.9450598614778836,
 0.23010185230522018,0.5474463826685967,0.9198233241549436,0.1399733650527174,
 0.9182962060080536,0.8907584099855396,0.8544505563247474,0.283703781967489,
 0.37941268315708143,0.8109203868288094,0.6319218954725271,0.1063663956116937,
 0.4080560968655089,0.10430569160515701,0.04623049632373022,
 0.0836927305508971,0.8741164376170673,0.5685594850448084,0.6449172025795662,
 0.9509545065701096,0.2292138769843941,0.5239944574625459,0.5963407174532968,
 0.7327274490840091,0.9591961091430826,0.05497090698843432,0.4831880621333151,
 0.2412879858914494,0.9833043008171635,0.8775095732200398,0.8955163049672028,
 0.506786169960779,0.28507384680744685,0.7643988460738327,0.6840191464578839,
 0.4570599449917575,0.48432304385600555,0.7760471863452719,0.9804324455415331,
 0.3196401841271932,0.10162258259159529,0.7785211492871497,0.661435839502496,
 0.7237621778304424,0.9907631239427989,0.38922094830969534,
 0.011024167229242599,0.2130891674569385,0.9261563292305616,
 0.6254672107882991,0.806014325820319,0.14027098794867654,0.6989170991391641,
 0.6861914695695064,0.8484428865998382,0.6940826041487806,0.8861432882538033,
 0.1543842691129731,0.8809036298195393,0.7126540016714626,0.7273669040785263,
 0.45055276369928554,0.6326012269728273,0.7767076814532351,
 0.17126412226205545,0.5702533290573368,0.09300416028316105,
 0.40964480730854735,0.8440555856959817,0.5359062372026865,0.925020057628009,
 0.855948209123421,0.2091741284801496,0.24714007493872947,0.3489758944030489,
 0.9016900091090074,0.6511133944664811,0.557788339668435,0.07027264791898113,
 0.5682839051771834,0.6650052716177637,0.7011572839343102,0.04782067078259933,
 0.5763745754649057,0.04849582481210346,0.900346663910671,0.9567039710437831,
 0.4445040209537716,0.5074156104660141,0.1035923248064502,0.7456245700985776,
 0.21655010306442046,0.5179174697343637,0.08617795326289679,
 0.8895300906347253,0.3852774660710895,0.6494784565883807,0.4101713981912567,
 0.4403179291054631,0.5817735813186613,0.6398880549294258,0.5035984926487762,
 0.4412696209578971,0.4579209027110349,0.4113539207858815,0.7971373903768163,
 0.7505239475401526,0.784466186533245,0.8572370593872188,0.9080506377191233,
 0.21126119854157,0.7333209421000682,0.5504656340146888,0.627037163076219,
 0.009536776205573583,0.6477852929560423,0.2992738057963784,
 0.7301907112866859,0.3647399041577024,0.5895989511225385,0.7029515746663462,
 0.2785867965896658,0.7199685868968457,0.6392784098255178,0.4656827287084596,
 0.08329479396440975,0.22366047624635077,0.5720626302110079,0.281242387086721,
 0.16870757922625346,0.4344287612008825,0.6098728971236155,
 0.16838310106726306,0.058519631446939946,0.3399280309628857,
 0.471979065023161,0.30090863609264673,0.4362146900716847,
 0.038617606893982415,0.32008487970523114,0.9831298813474354,
 0.4757778250910676,0.3239165518345296,0.11962068877434295,
 0.08882356741189179,0.019922907526520373,0.12206953055230008,
 0.760876756274584,0.027095150662901912,0.9672125940237264,0.8148069831043699,
 0.7223216795431503,0.4496509041842691,0.834098544279908,0.9825484286852105,
 0.9096734926093021,0.3229852325367446,0.8284747467188636,0.16998097363034237,
 0.15877482925499542,0.9900811562358853,0.29985917141690677,
 0.1811296856261002,0.2110854867550782,0.9135245536362874,0.9690556430931131,
 0.2028566286598119,0.358557835848199,0.24848880387337857,0.10882690812905738,
 0.49877018959633246,0.7025347089743802,0.5708220828066719,0.9681412047811738,
 0.6611695802283447,0.952735422108152,0.18540879940026622,0.4034471175006308,
 0.32063811067438786,0.6695062421896383,0.02684337977674811,
 0.03777868992806088,0.7405290583744347,0.5819006924003586,
 0.25604503604315676,0.9277492194390007,0.8686769786081455,
 0.49794539936097015,0.573024624058434,0.692353937057886,0.416978706649746,
 0.7394781050865582,0.5541955788940915,0.5212736180272246,0.4670362358784348,
 0.837929245709969,0.30269059190111935,0.22321828205596161,0.9259282432534606,
 0.13767644294958348,0.31027096960400813,0.024312889953062244,
 0.3379268175301402,0.9685697997350792,0.22371410643283307,0.4918459820216827,
 0.4533736170402808,0.44499696575323233,0.23278561868790815,
 0.7490535584215681,0.37644451713031,0.6378087460112273,0.004284922222263177,
 0.5157792317072398,0.2637861534786501,0.9270758721240763,0.6320085530775874,
 0.39246587572140135,0.7374006220739704,0.19890113380262808,
 0.8227677105197438,0.28319587064248963,0.8077835910019696,
 0.06193105601122684,0.39694242999603957,0.5666447532772737,
 0.8357704864432819,0.3422942268449025,0.9539240646370513,0.870765035241421,
 0.9558361498392592,0.4716275405536313,0.8407961380524669,0.36994778365885417,
 0.23017717762225964,0.58310932601138,0.731123518234806,0.7687762834392668,
 0.26347743981498717,0.19014085196125996,0.4018320654429237,
 0.6534085292654515,0.6441722722434302,0.261456887195079,0.5004696537984621,
 0.2521367397354235,0.9431160226392273,0.858381401854448,0.49000331836604705,
 0.9202028254919644,0.5951446393153869,0.5289297705574655,0.7263827877719474,
 0.9652556786503002,0.7454843813473039,0.9035930386558526,0.9580523914171896,
 0.5219140421887137,0.8580120069133093,0.18545220725136247,0.7820939864729091,
 0.7124299428351798,0.49385180730002565,0.08389223736540696,
 0.7957706036179678,0.23398806463325716,0.04139411943687543,
 0.8269755611497135,0.34664263319905597,0.40810342596693905,
 0.7370092503886881,0.4476415056428379,0.26811432617918973,0.2861137752562213,
 0.6817804537839697,0.5956082401356633,0.6608358079327734,0.75282525195016,
 0.9551215639801999,0.9388000607492732,0.5789999252180501,0.17620400104114142,
 0.015408593850162688,0.0647687544620188,0.8310346411121821,
 0.28548392146551127,0.3075306851811582,0.12404346301073232,
 0.8238788379983086,0.7470813832964751,0.6249784146576673,
 0.022538207038530822,0.2893265829518171,0.5995052410804922,
 0.6374945732641448,0.7083583238063875,0.5645590067200348,0.6516278066546829,
 0.35281018053678825,0.05599848204223201,0.25083476983625697,
 0.9795196667390583,0.4013895606940907,0.004550791370993901,
 0.42506430367062165,0.14427749813594604,0.4302457759869798,
 0.7054340377293825,0.5312128479958271,0.8636752737890832,
 0.0056554077581304485,0.8193386395818715,0.0600706673577287,
 0.8954792925718916,0.2187421566788127,0.7063434379109368,0.24533350085273953,
 0.021969796308170975,0.6965952776513887,0.08208471157243413,
 0.15992115650297045,0.6787851272490545,0.5765590850864085,0.1261042815295964,
 0.4120789861865637,0.6555772517853291,0.0229978989028794,0.9829712281144515,
 0.9699082378304112,0.9135212826730106,0.5950434375664572,0.34875999149260295,
 0.3717795481224462,0.9728147678267253,0.7399357772273761,
 0.027903957151555248,0.5339060902516545,0.3360660142804024,
 0.18720647782548094,0.02672647813893181,0.5408678867603228,
 0.6759984261472178,0.7153220520362399,0.7275068161119571,0.5225664086809426,
 0.2997565447037551,0.9407169479403652,0.4654541696942577,0.9623646052866982,
 0.30445304127952677,0.9886511117880601,0.2834899991335833,
 0.20173372178512017,0.8041914110447704,0.16436482872131553,
 0.6058608660949074,0.40181497267530986,0.3900571847716151,0.8098447754726156,
 0.28460933837892877,0.36412647555660094,0.5896310596009199,
 0.46866802867404356,0.869390883124877,0.27287801205322404,0.942207888032499,
 0.5758079899192257,0.9574761803523635,0.10919250003578596,
 0.04188272226420897,0.9530770682615507,0.7275527124957974,
 0.21422920325186223,0.18232922839958166,0.01602641020562945,
 0.2983665064487958,0.8118967915971351,0.6227296373944464,0.313431907043781,
 0.14612095255931024,0.5553177405777455,0.7048902220765327,0.9194111275442581,
 0.03739127983000201,0.1798155436197365,0.8709292646688823,0.6169201927279222,
 0.7715897470545636,0.03533656232355653,0.6387143204590726,
 0.08012802131923102,0.42669705407296343,0.6712316620617761,
 0.5687142927643732,0.9407157981737532,0.6870630146855172,0.16349393832097392,
 0.7876951350417086,0.2618639082324301,0.23300137139836463,0.796474645895442,
 0.6038583354871828,0.33590857183028844,0.4512291587409507,
 0.38092646035218924,0.10040354853381417,0.23751228340974107,
 0.006457769074522224,0.5682115445451444,0.7078174880405541,
 0.10729093478980878,0.06786533132720396,0.6404387686484738,0.468852688314101,
 0.3427518702732558,0.47822237028592474,0.29919367881334813,
 0.12217352571554918,0.10386247199725973,0.3380095962951871,
 0.3021996512171258,0.5190516710116639,0.1422669075388271,0.5773463162145186,
 0.09921858100198278,0.5799523834799277,0.255035439709167,0.24825831834127032,
 0.6664920147607389,0.49137910733704815,0.8572104492303195,
 0.28299021399736823,0.9206549234283603,0.06814374645935839,
 0.29495814186200686,0.27834174336803097,0.8615134908733054,
 0.17867714729771256,0.7841202813598048,0.7811011182314687,0.6417818299936701,
 0.9249473196544302,0.2319646620391086,0.8949452741414489,0.15481810472109603,
 0.673715847743178,0.42314237695547774,0.009334040616108541,
 0.4673476224369211,0.8414136333269692,0.4661013960474151,0.1325377311422663,
 0.01182855142829553,0.9505470670801878,0.03840502854216743,
 0.0891366447505404,0.9154673538466752,0.025381527999432163,
 0.5077746809512989,0.6974547030502074,0.648372948781002,0.4674057935362389,
 0.21334335236812052,0.7828083252143492,0.8245766916855454,0.9374309730560731,
 0.6985042258597799,0.36992892545679024,0.669286302297511,0.20823872790461762,
 0.41038587167205354,0.39369472441565356,0.9780310709248501,
 0.2878421025358402,0.5373099241283649,0.3051989087010393,0.917441852938484,
 0.992779656300023,0.8211553072507187,0.7712481986116324,0.8209231193716064,
 0.7772261280634719,0.5685014967263747,0.045763344662235306,
 0.8054064263314409,0.14498407115208434,0.3184990014088862,0.6321560211233946,
 0.0895813756847641,0.9161909840566621,0.5128758302055533,0.3732754930545995,
 0.4929767388728413,0.9750783419721798,0.2209443872118605,0.1458673673286357,
 0.5581839806111378,0.554773300234056,0.7987037984786992,0.40381990262608825,
 0.7736267186025021,0.8962897129984517,0.1065350877575525,0.29746063165933356,
 0.1018397086271885,0.14643605261768322,0.6619173605777309,0.2789408978366198,
 0.44620031567372664,0.952087452367518,0.2354789853164757,0.15903134857027046,
 0.28490894131924704,0.7734108055360595,0.37910656774665574,
 0.19212015732037135,0.8981666102326324,0.4760439639111784,0.3229075245921842,
 0.40890193723648405,0.6765024707872462,0.847869038318889,0.20463732956527725,
 0.8231317130553606,0.04392657474619466,0.7255444064106169,0.1770546856355668,
 0.47953791587882433,0.38620267225312843,0.7544490475989176,
 0.9985842341107475,0.49019060120794467,0.9489975772443586,0.8204823804467192,
 0.8841989494132794,0.0578546478390618,0.7080825122979947,0.15720596947411436,
 0.6986475615629278,0.40970502853115964,0.9837805553358305,
 0.21672555616354194,0.2508566190273842,0.8006567954719936,
 0.16485578001402446,0.9375615455772035,0.262032877671987,0.8107369230844654,
 0.915457148405689,0.9527568554270966,0.58746197457035,0.7010146528916146,
 0.08011126086175646,0.22961697061636488,0.6037574836156763,
 0.012549870220124881,0.7473332066481457,0.6411376044763314,
 0.039942143884053705,0.36803927655442015,0.5649259050314779,
 0.3786875419998852,0.7184660229603768,0.05989820976651172,0.6798411448579829,
 0.6312007921126742,0.06778565818660676,0.6393883881911098,
 0.24558751301301607,0.8599964250727989,0.31032804900615574,
 0.2624548317421104,0.06832065933100917,0.034077161595507555,
 0.7995246230761525,0.830792550997553,0.7645799643863601,0.6535001839883943,
 0.6394581546280629,0.4164834085760204,0.3581746628064576,0.44452115891455335,
 0.6590044549446544,0.9506467294552475,0.44587729198338666,
 0.11525853415606435,0.2062737909606711,0.9124689772442358,
 0.38394869793753994,0.2644533065559428,0.508621032288308,0.5100257319049976,
 0.09245314344558564,0.30576592682864734,0.4339093346285172,
 0.32520866010559535,0.2544009093432462,0.8897457092281971,0.0352966069180739,
 0.28545868500460814,0.24515953912614152,0.5207398291027452,
 0.9040856387185806,0.7415404151801993,0.4890027012638519,0.37788510076382853,
 0.7792331761278917,0.3442752008255613,0.10208303427699783,0.5655558287678932,
 0.6862748324420587,0.9388712644883463,0.6426996796444671,0.999492857816961,
 0.218774175515144,0.05206569596557542,0.09837777062067266,
 0.043226307076612125,0.5171019189661107,0.9264611410965757,
 0.5847024719276577,0.30335101500183326,0.09735903687840719,
 0.35864287795189276,0.029836617909914986,0.25451838874171107,
 0.19999580960193364,0.6108633266223775,0.3405317587021379,0.9043427816509638,
 0.0061390990849417015,0.3160478681711265,0.9522273612391681,
 0.20946758402001986,0.040902410565561875,0.776561105710033,
 0.28287873668794516,0.7422471375799946,0.10703784177972853,
 0.5018505696070714,0.8845037028504038,0.16411848107828553,0.6678948372100234,
 0.23534298387189845,0.6129324173320057,0.8782031823430332,
 0.15599518010955005,0.8420407574632249,0.6498555961617593,
 0.37290913869397213,0.3348311059736663,0.717763576452103,0.2672434481533892,
 0.12825209043896935,0.25461811821392244,0.2589097315493578,
 0.12068120186687681,0.8317616815370562,0.4194233735453513,0.5698670476057845,
 0.2736575302634806,0.25583665731992977,0.5843296555456219,0.5623309393593282,
 0.03896668519037083,0.5564779387587315,0.03263133982118893,
 0.1374732373700257,0.5588122251417973,0.4868263424979219,0.7876963369606566,
 0.9188207978218846,0.3886388362085942,0.20980050262370953,0.0888901595412257,
 0.5416296955194206,0.20514504327109995,0.45744653548018566,
 0.004804563258489658,0.8205200259102143,0.3269526323371038,
 0.5315606170103262,0.26706079886205814,0.4017681056500336,0.1375332857208148,
 0.22272302378422615,0.18366341770541028,0.9956633249397809,
 0.46515527078993246,0.4344298014481507,0.9624545729086917,
 0.30293070639188624,0.1345674743982327,0.9173889463626022,0.681362569162387,
 0.6660559152858371,0.5607340708884808,0.4172132696990769,0.6831526459922079,
 0.44594826200518756,0.07768738077536175,0.9969210060168096,
 0.22165557427198856,0.6839738572311853,0.7071858686502626,0.7515288532164618,
 0.6988172233827453,0.5814628235620493,0.848828862807049,0.39934283562942674,
 0.035972401221837313,0.5609992183878437,0.5120297690102795,
 0.10476272873784509,0.044304666049307606,0.9834601390665931,
 0.8116292019328124,0.6292513068127062,0.6385767516049666,0.5694558467422008,
 0.6240417243907901,0.9603481729204245,0.44862327017873616,0.37737785235084,
 0.35334353151341824,0.6688050522857816,0.7065550839016558,
 0.024151181871167382,0.8715288008217432,0.44921737256348715,
 0.01701784774228232,0.48464427782373143,0.4971629449300927,
 0.13590345775793922,0.8751106834659466,0.30786348049243917,
 0.8452081523809833,0.22657641349978286,0.3460606505839856,
 0.19615206177088607,0.6096308849577264,0.18807017567654938,
 0.11639067714540852,0.3469574286038011,0.07523938498562988,
 0.47068730946791804,0.06267666560619733,0.008773737064572185,
 0.23291050370585553,0.47677516412936094,0.8796742794366984,
 0.12135068552038397,0.185407678845662,0.8575356709258612,0.6330932069993269,
 0.8516341478453299,0.20740337040698542,0.7430339799102732,
 0.30128384664000096,0.9691445073836752,0.39413950789012886,0.240683483463222,
 0.41111443525379454,0.449889289470381,0.07316200952294671,0.7434290314572459,
 0.5087383296749292,0.9082298761115711,0.688470026848145,0.32865608532765345,
 0.46804198049518164,0.6214100865353421,0.6779392723314953,
 0.24534132807696118,0.45938408786139096,0.5046724852420416,
 0.5477596283475268,0.23838947317684167,0.2831807329803445,
 0.03241867729669168,0.8940861924230544,0.04686277672864014,
 0.18731765461287586,0.46686068609491616,0.16056116165754142,
 0.27335905193514887,0.3109616921524567,0.40363597854381694,
 0.44564754244667837,0.8580317523719276,0.5256396260258291,
 0.17166802036267303,0.15613071141904455,0.7555375227884978,
 0.22134930889557614,0.003849113629332712,0.9007308838424011,
 0.6430611874282628,0.8536629921132611,0.2614426651317199,0.00936501207602125,
 0.15801458355570375,0.7974989139029866,0.11300884969417269,
 0.6146353657373449,0.612864115251885,0.3560276719675448,0.9185210020987421,
 0.6292458296738741,0.7653646499990296,0.7471505129527792,0.18992321405997992,
 0.9061748784000805,0.4285241163765908,0.2920877002503386,0.8903314937503001,
 0.5372652864741059,0.24796555339388116,0.5735730414219392,0.6298721402395911,
 0.4891978685349203,0.2515973070391013,0.5391940600009804,0.4747239685801572,
 0.24607675713491461,0.6256572544747613,0.016743111335478567,
 0.7659459591026168,0.5448692026026919,0.04752162003436222,0.46916595258745,
 0.4005088719945643,0.0208397734150777,0.071611497226857,0.9963380590493094,
 0.9502184698045231,0.5959913740429419,0.9053938959906009,0.8592784401438458,
 0.6656714891856865,0.3765236583088035,0.9117081790929928,0.3524920006555139,
 0.9239399720102798,0.673633576060781,0.8844703206933165,0.21366007115007735,
 0.06993354238666916,0.29676138780376315,0.4000364342299044,
 0.9159568403263354,0.04870853829404642,0.4531894569764183,0.9405067910505595,
 0.49739009227583364,0.4762899397584932,0.5143580689496328,0.4069095828368463,
 0.6496972930078553,0.4882382140698529,0.6828057821113096,0.1864805719627074,
 0.363914802805981,0.11338873389670023,0.3038855392270807,0.4046901999592587,
 0.727655663116834,0.6518461758140484,0.3174364775884265,0.5804307836046856,
 0.4875795710791446,0.538463479706159,0.8330841269035382,0.11294720919485901,
 0.5688419294422654,0.5683885622050501,0.6770701610608865,0.7268221447714058,
 0.7801662478231933,0.9896990637342699,0.7103019960440785,0.6357437272233077,
 0.4954139398578715,0.9220627670965845,0.4652846441971572,0.7536813369357623,
 0.5158400853472845,0.0015883409588991482,0.11946597900648115,
 0.9938542717390634,0.04920476482332392,0.0914817099178531,0.6235531561260607,
 0.6939989761423315,0.5192575137383553,0.7260859781741084,0.30919462205015824,
 0.7567304572350555,0.485952324067195,0.5017344265689739,0.24544617519907985,
 0.24174080942635667,0.9340301009433918,0.9361999852031703,0.825601271585275,
 0.05784406688807919,0.26424410924331765,0.5690899269980948,
 0.07276270281469488,0.19650431005647462,0.9933101257464497,
 0.8002871615698323,0.3966619698040812,0.6085432945722463,0.21044791628640225,
 0.9184891263489541,0.6452868014730032,0.00864823979158702,0.3005819406333774,
 0.5529900166664138,0.9986124474784757,0.7325012641777154,0.8373122864768843,
 0.10278290783052957,0.9853889299683223,0.32013690053573496,
 0.030938104273938283,0.6631092282009118,0.3987092813352865,
 0.6349570924786636,0.6381488574199727,0.23998369051291957,0.6673469379347781,
 0.1538282854607077,0.41718599474928,0.143654505540441,0.20939192174676002,
 0.09458659897504451,0.2519407532938648,0.7879584408213413,0.2462910696100633,
 0.03333496673658676,0.5374054857197996,0.7620173172373196,
 0.20024752108860444,0.026283603272807876,0.28413241849428716,
 0.33845232719492857,0.07685501871542921,0.6913426763747987,
 0.7012370574178595,0.9415363993769874,0.8518504660618349,0.14590876465021818,
 0.8063389623080653,0.35929821950654794,0.04056129378250484,
 0.13934436213991108,0.8643288665081494,0.669128749174149,0.7171921052721685,
 0.4798821509008926,0.04860089986273031,0.48826521997664285,
 0.5846748944789381,0.8260465244544575,0.019467490984884206,
 0.3957525650796522,0.751468889118406,0.8600588524350834,0.0647398602886512,
 0.18078279535632802,0.6788667266712265,0.9327610921275714,0.9130119233979395,
 0.42342500003741623,0.2697646846346363,0.688528014691754,0.4038161404307251,
 0.18667981988071514,0.5539074726540187,0.9867108341885117,
 0.13141153479731327,0.6070310616907109,0.7037597295522461,0.7319897164764896,
 0.08388372090622254,0.49951931048360854,0.5827688801301901,
 0.24276652756821726,0.28148306554637614,0.21341164642472044,
 0.7261131076156272,0.1719828935644041,0.6657094957900201,0.6023785382849396,
 0.45142116844672087,0.8209267481352002,0.1390104765306348,0.3531812797578773,
 0.5718241481510703,0.37230918897967136,0.7638985433717222,0.8029446974783739,
 0.70937460067698,0.009029347340237526,0.7271716447158214,0.12725450933617977,
 0.7603379195599613,0.3311738208366377,0.4349282511631086,0.30660347376389296,
 0.5821229341908887,0.65791908799049,0.747689223270152,0.7235184961221995,
 0.768692116667039,0.32431017853815935,0.16202411131957073,0.7559615970728043,
 0.4354365668557064,0.9586513963079513,0.940485846386481,0.7245419707939618,
 0.34819388699123266,0.8971374005838204,0.11946084844302174,
 0.7888610505555841,0.8593864107254003,0.9744290742840583,0.44756132663702,
 0.28540446338259395,0.10315056258348054,0.5381646781089231,
 0.6661688488492844,0.15867113940393318,0.5121063380547133,0.1262960993674096,
 0.3502855232630482,0.9678296389224379,0.778247119715291,0.9826570671962155,
 0.5952917523408555,0.6568508618828711,0.8089166486031371,0.21376527325367367,
 0.666795379526551,0.5568707188268818,0.7996926325781438,0.8903492948147199,
 0.8827619803260224,0.13245137945258667,0.14671204794108117,
 0.4669952471554567,0.2155514298745067,0.043405167954772894,
 0.5472859533022322,0.3688160751775401,0.7575126972777424,0.165400005161762,
 0.10743464081309018,0.893008959500916,0.6718795564155483,0.0951803731693911,
 0.24351097730578508,0.6939915629035558,0.9964513127271305,0.9242655918169098,
 0.7891545773636008,0.8313817092266891,0.502750261500838,0.13814533637828397,
 0.22502841952737584,0.25610040878723894,0.817802722430585,0.9176498623845089,
 0.8221925270470083,0.2973871260666887,0.8484614071331987,0.654254475410375,
 0.0331066835592575,0.9240242133695247,0.3241445283986927,0.21455501704933866,
 0.40665894516593304,0.1965222364902559,0.635758559096533,0.6002763327596382,
 0.10183397746699918,0.49164799328015607,0.47717107742537834,
 0.03689706762417233,0.6801061601203995,0.7262018434569237,0.8876584391571206,
 0.8114446177980013,0.3164615767976303,0.8264856247247713,0.9429565839787636,
 0.04443714570627921,0.7404084041900298,0.13845939420390907,
 0.8976132562656979,0.7427643095990688,0.5484287466699602,0.6166092052005943,
 0.26488412942943074,0.2875644310862584,0.5773865699775972,
 0.41610010505553874,0.8319878481964531,0.036679407821761045,
 0.470742787269394,0.6722189191051772,0.4177263386746536,0.2347090639180115,
 0.3007061676321642,0.7045093808348508,0.35958039129691366,0.5214137295383133,
 0.45653133788662426,0.1489503977296498,0.6329602361491693,
 0.38068148670104496,0.79752808126309,0.7861530256851641,0.04137901818515233,
 0.18006232532138877,0.6107085179544763,0.5465691407420602,0.9776160137566692,
 0.17429439466447427,0.15900178716881264,0.5914528590916424,
 0.9827259538553165,0.5712582642442593,0.7557230787155862,0.47090156491812063,
 0.8019992097063267,0.736336584567821,0.012421135778111658,0.5220152599919958,
 0.39410224473696276,0.11091435547853756,0.3420025642962876,
 0.8417087388698699,0.11670917183626406,0.6138813032984543,0.9140278130576849,
 0.7922015568311653,0.024742155955840395,0.4988329987939133,
 0.3661425865057194,0.6778377919848826,0.9145291275466743,0.7759096923689074,
 0.5240265036403385,0.19338769068302375,0.7470866296738491,0.6829112086135782,
 0.5769301230997408,0.2303037352526176,0.9122960420400958,0.7264871818880172,
 0.9744159047656986,0.7873347764888523,0.24995530460274973,
 0.47467568784702996,0.2442663721695162,0.893979514269585,
 0.027029916089379102,0.3317578090943092,0.497206433465349,
 0.29706585170410316,0.7792775767902178,0.24030323691952282,
 0.5896516997954673,0.028172874323810104,0.18354847191958967,
 0.5951650326238167,0.2657048050268723,0.19830159437874983,0.997056251083037,
 0.019428600927662854,0.19135827821557472,0.9801959785580099,
 0.42749018399043903,0.7904607674839715,0.18082697845475248,
 0.6749379845419491,0.625496028638979,0.9435584682323008,0.8084227894364864,
 0.5770800161064308,0.48967832911582576,0.4434164559174766,0.7057525075409218,
 0.6901919948088946,0.40773211577293544,0.16448177198568137,
 0.7442662910385051,0.23209255101564785,0.019360556016818742,
 0.8550605908900828,0.3641292926737224,0.2191132709792345,0.6019718673251085,
 0.04191151566541729,0.7448836752651615,0.12526085488824323,
 0.2568989178381421,0.9587683034685478,0.006939711856566277,
 0.47452064706502795,0.6248109532126677,0.20160998585706857,
 0.8765185861588719,0.9802796057961894,0.35634240623172453,0.5024271607779872,
 0.6828346098354794,0.281596640130959,0.6631922891805869,0.01948013233864654,
 0.4046333133165383,0.5951118776263482,0.6046571987708422,0.8627846743134823,
 0.1300660680214023,0.036933983035873585,0.6671553562436546,
 0.08313242211843741,0.016011510335362633,0.5770902956126517,
 0.430405098886949,0.2529853937270352,0.58972960457907,0.2943902658665736,
 0.32163654477278525,0.4535073406255029,0.16048546026883614,
 0.4611155928844419,0.33850355398124543,0.06510326127679844,
 0.8635170149154466,0.8930853198225254,0.11060401840323941,0.4073829780015533,
 0.0882374712734354,0.9842557432523497,0.619362246683776,0.7821124478358489,
 0.7202995125739554,0.8073931288641437,0.613414656443434,0.6658080615553156,
 0.8989109549694687,0.020252318313024098,0.33158304443428654,
 0.42265129569827353,0.06711964941617754,0.004097888907738323,
 0.28674336377308784,0.2354160363625527,0.1788836298958576,
 0.13355867854733305,0.8328179122984961,0.5209742094766404,0.4567478577909768,
 0.578323174426698,0.9437850163199926,0.5954194428095656,0.5854275371914344,
 0.49150820959691155,0.7150463192727379,0.16613926101801935,
 0.8161745687672843,0.0586862421578358,0.07872889990154652,0.6276055578232174,
 0.2627720433549149,0.8622397315272807,0.5954462663670648,0.0483345735740347,
 0.48215442377390105,0.8810993157961861,0.9692026233941033,0.9776436401986873,
 0.2666656438745447,0.9860942715955142,0.553883480418521,0.24636213125805517,
 0.10784179713249664,0.1278260526768078,0.9829352105834882,0.3121891416000886,
 0.692168647852177,0.3279235113411215,0.2325587624467006,0.6817143935196889,
 0.46629676967020206,0.47844762458350054,0.08397929580084162,
 0.8490882508024135,0.21222880781623166,0.272153926171235,0.30608208173876894,
 0.3563768582033293,0.5878519937216207,0.7787687911507645,0.6684177334858227,
 0.8509517943689828,0.9393006813241399,0.34422234492413595,
 0.041159989292429366,0.32962153620175083,0.36114129647523907,
 0.032832540084250295,0.9938430902182964,0.7814497933908312,
 0.9500224655619822,0.9813022565666866,0.6165889928339127,
 0.010067766192259908,0.525251080132177,0.6476541627437457,0.4250698679896101,
 0.10084042674727123,0.32376840869580337,0.3699222546468248,
 0.9479806129826106,0.6795460013888035,0.02724282203452144,
 0.29540830936250473,0.31144166553457775,0.1442710936106275,
 0.2108392560038853,0.09108794918753838,0.03564718281969059,
 0.21116977691384153,0.7695363076688215,0.017348913081044248,
 0.15056262083820626,0.3080927701776994,0.23622688205185294,
 0.9596827257119851,0.9641520399846131,0.6968337403229502,0.6786455966148196,
 0.7577331396164605,0.7845908163133137,0.14629063841013512,
 0.02230512207092672,0.693825658748,0.35875587449661106,0.8095366128267386,
 0.5292502422305333,0.7155176661795937,0.04408639005075787,
 0.45708082235199277,0.02851858570013488,0.3416710714976037,
 0.8585622998329971,0.2372486209999669,0.5326866871604045,0.588538109618616,
 0.930811012261266,0.8327932938537398,0.5468886139172539,0.6838600918280149,
 0.19912229444552887,0.9679503310199818,0.9721885096146582,0.7772751440491842,
 0.5304116048290742,0.5419750896796179,0.7494732910077537,0.4326100750487061,
 0.6700451352696417,0.06939992940286921,9.362229117857446e-4,
 0.44465279692121806,0.7742135714431722,0.3623770579385376,0.4034210580388894,
 0.3730721911241943,0.8409604443769224,0.8540570873093484,0.7002916080520243,
 0.08055456444022291,0.7474219256507862,0.09348928982317706,
 0.9786168286195138,0.5881554158900929,0.06831829146299095,0.1570121108488277,
 0.898247461353287,0.06516106078948525,0.713960257772623,0.5103834844703468,
 0.6166425690453248,0.23080952748263472,0.8929893780190756,0.5987302939950178,
 0.14776100702692097,0.293707784061244,0.09355664660114982,
 0.25627019708766685,0.29262389191670923,0.8386621002545782,
 0.6468762217316093,0.9036219139202276,0.04682628384868337,
 0.05368882498021332,0.8005460372897488,0.9988837141333359,0.4403202058018958,
 0.9178515202724202,0.6499758243631252,0.9896565709284073,0.6926804379842719,
 0.3072730453271699,0.11399337387259112,0.46151949128500136,
 0.7759753673827798,0.7361761115625458,0.9029347640008321,0.8462243791376097,
 0.9588460493208156,0.05361415532331304,0.9789327617288757,
 0.44148963569668254,0.28009835352301593,0.27216091489866934,
 0.7287266232311074,0.44439257021579714,0.8286032574254216,0.1484548332682183,
 0.189160000546153,0.5140918603116891,0.36821793050150886,0.8328292679788446,
 0.41462588916141985,0.33702724105515003,0.7655909283082974,
 0.5237238000455837,0.7096928332642894,0.051916307500699976,
 0.49297847737833544,0.30860832938400284,0.5402730703824686,
 0.9153494832890592,0.23859844189456247,0.6478050727900884,0.8156685466886058,
 0.31065989160227636,0.33231217227404586,0.201042978473435,0.8204353768111827,
 0.04255125613412958,0.3236620334635265,0.29047484304238225,
 0.08172618337013826,0.6203941677794566,0.48424536373262006,
 0.25987927634337626,0.44296352232502345,0.19334112441841145,
 0.09247721180745139,0.3238885335223669,0.5399475027131141,0.7127356561271477,
 0.41089778658911547,0.48159602472329677,0.6616985938603486,
 0.8944496519748834,0.7918721892237603,0.19522830955727977,0.4314213482065751,
 0.30644989133417655,0.007906998527230424,0.32638469628272193,
 0.26318983040054855,0.7661147234330644,0.7416520154953625,
 0.16353665021796593,0.022007703522135103,0.8885095680610581,
 0.08702381046221264,0.6869871885014402,0.4905033953488922,0.8841578762991442,
 0.006048235092335097,0.22273028205304146,0.5453586860340544,
 0.9351281523033812,0.3297829327952927,0.6140902195080531,0.08476230991873479,
 0.6130028576946762,0.05133185963205178,0.33815982967437863,
 0.9480731306837139,0.2820374409845179,0.16021993272608492,0.8864075405368821,
 0.8659712625953528,0.3038380976713081,0.841656080484421,0.07115161671053793,
 0.3346817955250303,0.8306321292011052,0.4593222286069992,0.7395046219460184,
 0.01532445908589164,0.11309259377517455,0.505434472804452,0.6474496675266871,
 0.5161980437525089,0.3528344148147552,0.38532020129069,0.7047263986466857,
 0.8760546495452783,0.10586313764287458,0.3225660950413114,0.5241859838253919,
 0.08165093774591603,0.05359483478836258,0.34913550814741245,
 0.43948765749748075,0.24757336095057214,0.1354432760302664,
 0.36062567429672154,0.16508742720140723,0.20426749560934565,
 0.6542384517941338,0.3995440819865531,0.7431247554660486,0.11926538248381946,
 0.6912428741949241,0.8420418097306697,0.14042638924547313,0.5744531684667714,
 0.7234258336420019,0.6555148502225622,0.6919584373902334,0.26146367363083123,
 0.1572988449202175,0.9539030140753992,0.9268448962629753,0.34427061399071934,
 0.19422835009724526,0.17304888205660696,0.5255919078817683,
 0.08654252334042867,0.10112513304779258,0.06586503750863537,
 0.24685721002121674,0.8527654191342467,0.8297838254085286,0.6032525677595607,
 0.4288808864524942,0.8893789818257231,0.0910646358005135,0.7690917020557718,
 0.09764101225200739,0.31622988280411557,0.8839226558088301,0.189100994348124,
 0.16938342130115358,0.7902417225597377,0.47601337304075253,
 0.9745989701580492,0.9948266189274108,0.4798625652519718,0.7853268926465089,
 0.7198739765961732,0.6391837800053287,0.9697498303677774,0.19440395549670175,
 0.9708294294461793,0.2283631640753142,0.9730895761255441,0.4637799658032148,
 0.026820382321632308,0.5801161816262813,0.1960693042538746,
 0.8807845046355244,0.531647315258133,0.14151613124492957,0.26253855891086664,
 0.47838380331203734,0.960035673507689,0.05272388667090189,
 0.040182877598252675,0.7054835610560053,0.19191001768154037,
 0.9729556201452297,0.6548597866854841,0.20248573494746303,0.6127383505782618,
 0.8034089140894657,0.9341947662951562,0.09712210128574994,0.5362620720715795,
 0.02456910017555758,0.2722016558041698,0.41499468140061235,
 0.09591142526742014,0.6464070366718339,0.24733971855963455,0.581926699278665,
 0.2286328428467883,0.16818521999676328,0.7661603073217251,
 0.06554469421502429,0.9324659788179401,0.4523430065260635,
 0.39484286220520326,0.8684432253073924,0.5325771884091666,
 0.25022748006967077,0.3965395503502742,0.1928953828619724,0.9411937900341648,
 0.49480850582399216,0.6885373070620253,0.39629476287223253,
 0.16504695137016623,0.091798808827256,0.1396107140817301,0.6322436828315461,
 0.9263248892490373,0.49025566961131306,0.020343428837958033,
 0.43907011754675185,0.03769108788214437,0.29725161218792273,
 0.6547005714612127,0.7127620847364863,0.930675807455357,0.3810613871163818,
 0.016509620083561716,0.6137740858430308,0.44776799207608087,
 0.1500036502195491,0.8327354629970665,0.03217198344414751,0.2330855235953203,
 0.9322964745190494,0.7969507514097444,0.6013818867338606,0.790513213319687,
 0.7256199613584012,0.415294961941044,0.11933762448126528,0.34827726712255114,
 0.5249760807852436,0.6251481563382761,0.9825839636004436,0.9663425700388595,
 0.4818778216641093,0.8991742190906096,0.6200091897181593,0.11473224763655154,
 0.6339451797200262,0.9430000941072413,0.9175812873678582,0.2633125217648402,
 0.7507477438304364,0.5689476623980764,0.5792613124837711,0.7758227870100801,
 0.5545588107568236,0.3283432560379792,0.4144786020676994,0.3579951516268992,
 0.09272036974164022,0.9687005139947552,0.6823462748764281,
 0.002767903602284294,0.8736231695344394,0.8016179260711515,
 0.16132709954148017,0.41603034684950957,0.6980748736541912,
 0.28063621848581854,0.7827285813316898,0.2794052218853338,
 0.40580128739197485,0.11687729711883255,0.12300076167015483,
 0.1343721510155571,0.6980479635857935,0.9832992598147199,0.20791341638717664,
 0.9084466455358648,0.34545092995511906,0.3760112432081867,
 0.35573842163090985,0.5764388996116782,0.31931839012913477,
 0.35384667009898596,0.32318486944998825,0.41673250672100814,
 0.6636102050173854,0.05923605877810756,0.8293015778021523,0.843415495226103,
 0.05825863314635893,0.94610413976185,0.3801875804933894,0.07564453777382107,
 0.4467781633004245,0.06902344726267384,0.5419995453612518,0.9981910651864874,
 0.42310858604848245,0.6225062424515001,0.3964879834571763,0.6367123077132107,
 0.29449156940220633,0.09112545794482674,0.9480885991074541,
 0.5448771427658367,0.3742182223159517,0.34731303838060723,0.9460487426907882,
 0.7634109906496098,0.45455179851183547,0.8667108267643885,0.8233186453850904,
 0.7542377683128102,0.6788394537811796,0.6553642806696538,0.33845448637924824,
 0.5646777038691783,0.9406595884312764,0.8629030801149651,0.754044091541066,
 0.5110238560864194,0.0893034111091231,0.1166731938317076,0.08137921807599746,
 0.16914084880094493,0.11604686353470295,0.9305556912451758,
 0.6261020227921998,0.4966163286805392,0.6040635353411162,0.3551784987593142,
 0.14421716512496652,0.23845177620983027,0.33268319839934113,
 0.673656439698731,0.8372894049553755,0.2718543860695858,0.017459079436944802,
 0.9294028108614107,0.26929581322633034,0.6158848043283363,0.7822287275101758,
 0.8589222461534618,0.5640706402963788,0.26069122774738973,0.7178996668516869,
 0.07854723936001884,0.5578773901089779,0.3977204056410253,0.790287986159906,
 0.11192867700882658,0.5058779044618864,0.130755840544573,0.6275067127481132,
 0.22687177094918298,0.3275008280179854,0.6228287006122195,0.7448855768814192,
 0.39341081859896654,0.6495387013026659,0.6003397595870138,
 0.47246026670719976,0.27062413205742175,0.378253013654006,0.6581097238442786,
 0.18699364231743165,0.8304084123600906,0.17724503298322913,
 0.6277671529539356,0.973483066756947,0.652717428248296,0.3316430706929884,
 0.9087406547579304,0.6565736292211168,0.41846253259331756,0.8883427099412794,
 0.22836758735982798,0.25738560878666483,0.577239294271533,0.5915209026099623,
 0.6715207529767028,0.1286593336428724,0.23436417955371236,0.8172236969372455,
 0.8002040213527765,0.5777589609888022,0.015887945289102423,
 0.9752849463270767,0.12130506810828279,0.6935681199547556,0.4555969196319801,
 0.3907728020103317,0.9718068358079313,0.6589677261847787,0.2964855531577183,
 0.04579796838865102,0.5344621094140057,0.8495982524188901,0.9485887798988069,
 0.35367060617037105,0.2552267966522226,0.3142954660201771,0.820876602295556,
 0.3476966298568852,0.06761880519991659,0.4258430981768968,0.0942171950475239,
 0.2854413000440118,0.6992812145543912,0.6498113282549662,0.6715235080160159,
 0.3855333181930287,0.1937814163823608,0.9857523745636907,0.9049357135759214,
 0.6902450727316495,0.8887046734745684,0.543900586833439,0.5585208412229355,
 0.4994250135977135,0.7432343663043786,0.0668560547107222,0.8159900201250336,
 0.8156829778130749,0.6767383011792848,0.9850587354208635,0.13521675542317246,
 0.40163698937718073,0.06509404539676933,0.9207121234484618,
 0.9339555893359607,0.4710902428403856,0.9345583402046079,0.8346554799094323,
 0.3519279527116187,0.2782294847607285,0.3623041120196465,0.668492362007375,
 0.6472211100720998,0.1161694626703732,0.592621046818498,0.2234353616510414,
 0.748582329863725,0.8817927692986597,0.5683195922192334,0.8264010158131286,
 0.150133389546382,0.8557784120828462,0.9690306472117288,0.1079532326211804,
 0.7261959036960207,0.9893735565282418,0.15937737935265428,
 0.24617485587229093,0.7808197076454922,0.7817334852637239,
 0.40446224192563984,0.20843926224807596,0.11308604473185702,
 0.20959070570857619,0.5986074795284999,0.6005194007313563,0.6516070341912201,
 0.03228491734219163,0.7337806434027677,0.8210005134161192,
 0.06379603944343937,0.5166463425900405,0.3961897956184415,
 0.08300837630302349,0.8166077582356166,0.6053165861707648,
 0.31303954023261393,0.8905492306333567,0.4638445913277105,0.5414948243726222,
 0.5495929807029697,0.25869169756909094,0.9357817576420864,0.7944607134699329,
 0.01569552398501184,0.698239848201573,0.018816688694822337,
 0.3568026037592549,0.7981579857741298,0.503785648466501,0.7529784442914488,
 0.8653836888290803,0.18091472395703123,0.5279013187877348,
 0.49092335379105523,0.14460004036495522,0.8425180480580516,
 0.23568412049932408,0.6982508940600105,0.9398382034601748,0.48103746284795,
 0.7477800112547514,0.70050982721255,0.19939157975909672,0.903471610612989,
 0.19002358579018308,0.4707823164515297,0.08290427879164641,
 0.9316293677585555,0.9110337263420665,0.4213357851384194,0.8141227225444054,
 0.07231964418746328,0.5856958871022298,0.8795549655120418,
 0.35841273334576873,0.002804794303206015,0.8251050489843863,
 0.9513199138793804,0.4575289373557526,0.7381187549183367,0.7928468134796695,
 0.7046746956509746,0.713180445185306,0.20736984002865666,0.43012066935866544,
 0.5799757469273097,0.6042153741434715,0.16821336822149813,0.8942980601807737,
 0.36909912383124144,0.7835408578049998,0.24360441227571483,
 0.5492741722729048,0.13516981788888716,0.16251315077310047,
 0.6318841111520969,0.043701116537645524,0.5921038067421955,
 0.9924062009463493,0.8859595694609406,0.395596452702761,0.9883942917999898,
 0.5466224427316361,0.18355412400297189,0.37461250642403976,
 0.44899426215935656,0.6079811354587469,0.9993516309355438,0.3508824743452772,
 0.9147745911739936,0.011332854930576008,0.6982567289243518,
 0.4977170496001101,0.710173794293855,0.620516671224143,0.8078454366988064,
 0.5819785778116037,0.7298510311797638,0.6144992899444285,0.1538842670849172,
 0.746415196731486,0.3213323181803466,0.061452250845302925,0.6468894071375813,
 0.19130158849873558,0.51749751680197,0.594569928403658,0.32428389960648707,
 0.5967265629445496,0.6167625786535852,0.27801870127960426,0.2843692492459675,
 0.7417606514750095,0.8230657744611884,0.6766314192154635,0.33192100847792927,
 0.8044138454178356,0.0021815419839595895,0.13054941713729218,
 0.024562989076119646,0.02520042059553773,0.2159414143764148,
 0.3147037955702151,0.06663385590319937,0.41375018292506927,
 0.009678276969721122,0.3890963189995218,0.6776395303400679,
 0.18345270003697167,0.7754972140757722,0.4764603066765344,0.7568793203539238,
 0.135622660114176,0.9682783212925462,0.4504022818785277,0.45528471667190507,
 0.3566402685814396,0.6289804197175345,0.5580610883333718,0.1316147889535828,
 0.17545600499212155,0.06861856504865793,0.9899202607548309,0.86223989538984,
 0.1207845779539366,0.19290457479990764,0.3588120280687702,0.9346065641325502,
 0.526968574804577,0.6266885763435166,0.42969369195576523,0.649420872743048,
 0.6349516653112486,0.2820596958129057,0.5715068479643948,0.41511675375923207,
 0.2229422752186737,0.5760903381612702,0.07547258273055768,
 0.24959141528699014,0.778188153866955,0.4635482988541131,0.22549517627230187,
 0.9400401301655964,0.3881468515633575,0.487984449375226,0.656376034602075,
 0.6000338640223846,0.22499988813663518,0.501561083237637,
 0.020075120254245182,0.8111739573202383,0.21638792739855606,
 0.5243022189509171,0.9293476320596532,0.8516527312835851,0.7162296944499013,
 0.6221844054433636,0.954697114869062,0.39752152810077346,0.6479068811231806,
 0.1988805132420297,0.40767626110304156,0.3497828939407719,
 0.49161920751954946,0.8837832548693452,0.25792980617745,0.21602333596984535,
 0.9424969016351393,0.6313906328421163,0.0847814970870967,0.1809646517628406,
 0.34508741869714166,0.15044861724644554,0.5614406800082458,
 0.41049144584481767,0.42548187182292463,0.3528511471486204,
 0.9936316499690877,0.30965715140966676,0.15645186468866012,
 0.6172110816725117,0.5620741125447665,0.8308844305188129,0.12316536769451614,
 0.018054698197655705,0.7388807406790926,0.9961917959562194,
 0.2325430826806325,0.039640420071573756,0.45604918615761025,
 0.3972882258302959,0.45349897888584667,0.6519838060669169,0.1757227409695694,
 0.0999260011876496,0.42321706380752033,0.27953395339860343,
 0.02407784403112856,0.17441684674377933,0.4798775257010255,
 0.6095646696145286,0.6932844026443702,0.050552714500142204,
 0.12176276061985059,0.07243792079370859,0.7042706017800936,
 0.5458025756499008,0.37918952049054866,0.002516010690039039,
 0.9656948375146789,0.571067612423804,0.422544587593907,0.8781308410900424,
 0.6651771579283521,0.34869800894386527,0.9431472168922601,
 0.12333798522635087,0.12399738728001974,0.7084141452815284,
 0.7359857072960488,0.6429253168418149,0.429199957692735,0.06730645773439858,
 0.29978738681618644,0.535957975215364,0.7640374137492825,0.37414533820239426,
 0.09140778039868469,0.47680358626302854,0.02210865940435014,
 0.7699034046818904,0.13005481926374018,0.47204835600350303,
 0.8423624372143432,0.8681258894836495,0.39905593769847836,0.5563350344583425,
 0.06442059361201014,0.08634167464004938,0.7057794089181484,
 0.8452647014826947,0.8977437149451961,0.6113819281950895,0.27319125198633387,
 0.31402647646064996,0.3376129745210845,0.681674305787273,0.2473322991529261,
 0.2950766797986144,0.2884804296661869,0.3628461240646672,0.3084075233610484,
 0.6723178088747302,0.7447593099037109,0.006391452791261298,
 0.2092146023141871,0.014805898539581808,0.9008440235419483,
 0.6273543234261815,0.1908613190821673,0.9467837412940333,0.7840786248466379,
 0.7829184119062503,0.12220938752654298,0.9592602184748873,0.6764098013838205,
 0.89396473422207,0.6160198277550899,0.32251013567058795,0.37313764645269676,
 0.1357570799929264,0.6034316618181725,0.6537558809450689,0.11816727017742712,
 0.7148117184312163,0.0592465487584839,0.8592752894621477,0.27094455297244424,
 0.31578791402033346,0.9592929557692186,0.7219204648275768,0.9732841417287962,
 0.5063672643254419,0.6279121542970016,0.5351708559534551,0.7571894898949916,
 0.6231324659920343,0.7095218485962237,0.02703707558832702,
 0.45307163053751054,0.19278396024617506,0.11914560302258004,
 0.1239668830311974,0.959529483885399,0.6443514770819231,0.12327034219530553,
 0.13275787712141107,0.35518561184875574,0.15854183249194997,
 0.24433297657800013,0.6863553395781341,0.876462173372023,0.13481410968727292,
 0.06373739062877248,0.44244692946123576,0.9965759246068666,
 0.09957836875508674,0.5782233686359414,0.8659697629143085,0.8098637934467356,
 0.20062316611152153,0.105061713542412,0.8499844285884347,0.03475288176286295,
 0.16103460771590383,0.1249980293736358,0.5144164423118145,0.9881312852779504,
 0.39164380092855633,0.16661100207847368,0.7467849005290932,
 0.3519197270449993,0.9417533222472623,0.24381376312407044,0.6053457223431401,
 0.4585099288878408,0.9924608814313889,0.9552069718717326,0.1420137698289975,
 0.8754482513631694,0.15979234057911762,0.7803887048812598,
 0.40383355762121886,0.35484839748387653,0.8752154179875486,
 0.3668140898042387,0.18868110709838082,0.914527580538143,0.49094849886061054,
 0.08084537919545254,0.12802849040635234,0.9863469581767165,
 0.3350035252346204,0.3347768034092712,0.47661017227633873,0.3606517194325439,
 0.0442477585547596,0.8377049682492547,0.2872102196316454,0.8291012645593966,
 0.2765098780995384,0.9069020777124701,0.09164594362989775,0.5626834109652538,
 0.20622351982464093,0.9221962926407348,0.6905493589453972,
 0.48768808234963323,0.9957484358615231,0.7393706253930735,0.5604989629683781,
 0.205285295446765,0.4350104650196971,0.2387673807539914,0.7154658423210877,
 0.5980619293575373,0.58939971337686,0.5002732687423328,0.3239580087505032,
 0.3584859095081352,0.5100901715508626,0.008051263036012335,
 0.08034728986902251,0.5117913512161745,0.8326432286431347,0.5592696918733897,
 0.08442192878786192,0.6535314597336489,0.9744930237736402,0.5384490215034068,
 0.4206829136339847,0.3293231228594429,0.12861251221201386,0.6830838411048391,
 0.9600783957038426,0.726839748589527,0.49835432756972964,
 8.646465115207391e-4,0.11475713447163982,0.8954680301254172,
 0.10255449085593538,0.3823522226827576,0.40253832152578495,
 0.5450178471791597,0.1417747366890978,0.6485002976859404,0.25707494648550455,
 0.3353009710470871,0.12717486875329564,0.7560431776654941,0.3061907403889118,
 0.23001258922864853,0.9587696879567253,0.8607186599447142,0.5985847894206663,
 0.9445329274997702,0.17211544136018952,0.8847339851682383,0.4834103549514237,
 0.395541765883931,0.6703436535548327,0.5263910425700242,0.904334765439198,
 0.9169851329373988,0.36490553621506105,0.6325247395741576,
 0.02555566380701213,0.623735817902421,0.08953527660577776,0.8334554333548856,
 0.8835078318867944,0.8576257562588985,0.2257075364981469,0.9233801367136214,
 0.8649439314384996,0.8965877648263548,0.6898315441038902,0.05084418616908459,
 0.5934671554443571,0.7626313044102686,0.5155281133442494,0.42236854090745224,
 0.819230243355013,0.0430020227337351,0.8564719034425432,0.15421351306543685,
 0.2575956815709326,0.236266622207719,0.1451130040574733,0.8667650958499515,
 0.14576669827121136,0.12426664462213588,0.2660572701853532,
 0.7411939734890485,0.10389669119707667,0.43795108770418245,
 0.6764246557123744,0.27948299491386996,0.28252514477690216,
 0.8725239445386673,0.9544873674176141,0.26686939519284303,0.9994604509091247,
 0.5190704844250319,0.11839491811252745,0.29896263169532755,
 0.5253098447313337,0.9122022373491967,0.41679890070516956,0.9671067561687271,
 0.6654163811293324,0.3679800005311714,0.7353262913558006,0.08657518671459241,
 0.42051303748309243,0.938863144286238,0.5674055944751101,0.0965754113166345,
 0.7164264510081015,0.8992854657488515,0.2573793991104488,0.9457762046591575,
 0.5703854523365353,0.04479420743055729,0.8933579243331088,0.7926380204103523,
 0.8651059985677976,0.35354707118751183,0.7221011305242724,0.1630565303976872,
 0.6174631094922267,0.33893188625418436,0.7314620573576669,
 0.24711273905205167,0.4500821137276707,0.25984924295976297,
 0.9425235562986821,0.7617310952607531,0.4363414277001719,0.7270055588943274,
 0.4913685280905793,0.26282793384565006,0.7739923568107151,
 0.18140377647063366,0.7740625325838354,0.9980240969257277,
 0.12946755031261803,0.48917020812210943,0.01253679914100525,
 0.4347304698921345,0.8881472378609185,0.7217534288663137,0.7362804331488211,
 0.6068557685149082,0.5788512136430843,0.8771130562299059,0.456838896959581,
 0.41258632183498856,0.7433482843938453,0.715088167518977,0.9570750646869711,
 0.4545045967487008,0.3693567655336232,0.9500951492008484,0.1591500438858794,
 0.9927368904321549,0.8005892204732875,0.7905193952632548,0.6171349770878439,
 0.7955822049136904,0.47782051800048286,0.7326157589804413,0.652715539614379,
 0.436403958563358,0.7158965256256205,0.4554604727293524,0.9075556234862817,
 0.15789985179166066,0.2405296439025738,0.14501890732226774,0.855914249126941,
 0.5609769109899732,0.46144458985554726,0.5806697046102894,0.8144171755622455,
 0.9482697637511778,0.9253464875487251,0.2788949377991272,0.8530225021222054,
 0.9267020170420265,0.8371252521666757,0.7695489874354379,0.07800197967816436,
 0.8138415914479392,0.9572994976305012,0.18342115638620637,0.3564465521786283,
 0.7386526054452147,0.8989675730407882,0.1623269219954463,0.5112360037997624,
 0.9834894571551249,0.9730914117760623,0.48839815275829734,0.296535492650265,
 0.7768228147872661,0.0777563730558879,0.5430865589264822,0.6300817799050713,
 0.7802940756244582,0.617296256020632,0.05708346672643372,0.2704468281736321,
 0.10405727883646809,0.31817731365106383,0.5097119487848226,
 0.7612587680017133,0.36804865088161676,0.7277406685764927,0.4438914128192998,
 0.1783987248293114,0.7999542663308945,0.7699990853092714,0.9419889230818647,
 0.5477467207580402,0.038331217670062756,0.7241754122810514,0.897276554880189,
 0.2774970419768612,0.743688937163046,0.6123386140797148,0.27443738803264806,
 0.7218784300188198,0.8416922745833784,0.60003530511797,0.2298927395443,
 0.028222510813018187,0.5858870975386914,0.2681862322578008,
 0.4327050310725532,0.8202896733953857,0.6436439664492335,0.5482406091620651,
 0.8174315075930833,0.06568951237065368,0.08463898895196886,
 0.19717620279227366,0.010883690537506707,0.2995278660402513,
 0.935295640439577,0.8902656725158231,0.20200304684161907,0.5470580382878589,
 0.5051706353549955,0.7515546779291209,0.01320705219757301,0.7409834906816691,
 0.6812476070660076,0.1875596087195639,0.07021159437948754,0.3734757292719286,
 0.33585615154894377,0.2961955019367294,0.8984729912500657,0.5655100436896383,
 0.2500470211000402,0.7898452756252681,0.7482819087132477,0.5042368812806286,
 0.82471919004908,0.2633028557377768,0.49746069269122506,0.5448419659053444,
 0.2926591917082364,0.3094678279606127,0.0350024943185161,0.9118186770429343,
 0.35005165381519954,0.4437426296238155,0.18447191986487366,
 0.18696545342124826,0.5669700689310198,0.2276331872797317,
 0.20290921411104001,0.8870018046157152,0.7483334703209481,0.8682254904651395,
 0.5589402597536542,0.832314318334125,0.8106742675071491,0.40350579521359853,
 0.2355564695329726,0.7701648706235852,0.9460336857838819,0.3977435371900623,
 0.13215318598284642,0.9945893827700483,0.5152621437893479,0.676325246314192,
 0.015049560239290827,0.05724669961812179,0.17965511200835982,
 0.9538922592115575,0.8282295847424678,0.360047838296753,0.5162981303318265,
 0.022598462807174702,0.0155771845006204,0.8141465711618889,
 0.2594956367288914,0.9513905952011733,0.8527334203912336,0.5212552793695636,
 0.7165019469316335,0.30248182967921966,0.018471291276245805,
 0.6180306733473657,0.5569887640513548,0.21540385356066982,0.9864542317951241,
 0.8068059409760361,0.6010166780388472,0.1620471128471015,0.20037012170254054,
 0.3552051673016777,0.3195046155773478,0.37001708746770656,0.6703358199384395,
 0.7853037901320294,0.2917565850682662,0.5182602682196208,0.43461869571332323,
 0.4151925798710525,0.3722808459666451,0.9431967437385875,0.7873220603661703,
 0.6445409873312345,0.4578397205346837,0.49893668301998584,
 0.29563798929783747,0.8036090041304443,0.401877841843031,0.1724961274534198,
 0.3206200667976652,0.28092316315834964,0.19469834406233222,
 0.9246830406967068,0.7608335290770909,0.2103572992525118,0.8813228335830076,
 0.8199958688646066,0.0487588451124914,0.554849995531216,0.6859298487002962,
 0.6496572825022002,0.3012920696760105,0.6535517335574674,0.17288976172520987,
 0.5789298303551075,0.07096568687103477,0.8304806019223392,0.2090293941779353,
 0.9200058967348714,0.9217791255954286,0.5195900614177665,0.8959672600389866,
 0.5862929344316082,0.9303904317938404,0.9679621671505301,0.9515794472728131,
 0.40666600536025643,0.593383424133019,0.969528870562222,0.5424203699043926,
 0.8770552482740746,0.6579589259259374,0.09001408542729583,
 0.31259524823821006,0.35799833699246986,0.798903217243816,
 0.24340471049011303,0.8406666042121423,0.6752484412251686,0.1966539012278422,
 0.08229058414958568,0.13750744296092832,0.29373921657956414,
 0.9599695142174157,0.8284219098680897,0.1487866649213564,0.8297634930463593,
 0.7198759301831114,0.1691773889556646,0.7128759728656114,0.28647531361504697,
 0.94152166885359,0.3526587397527313,0.6139016376227673,0.8865067694350788,
 0.7951291061504923,0.22190005989654948,0.32697930204703984,
 0.7151961631554429,0.4206840450772742,0.13448346769934028,0.7410403555703857,
 0.4632439561264322,0.6536105717363186,0.39912521862830563,0.5982086634556745,
 0.38440893389141895,0.45992034674146043,0.2786814962492624,
 0.7014240291422625,0.03229155656236993,0.2021510402534602,0.3815451697527691,
 0.2412837054115925,0.555427885321154,0.029400198005671396,
 0.026729143505304864,0.1161311698425872,0.05616106527718889,
 0.7795039018849156,0.04872421521229453,0.025133764069016684,
 0.726530491702334,0.18914054997056884,0.690665130316044,0.544936167456989,
 0.3688335525317583,0.5672227470239523,0.19424532924086857,0.3875300280957722,
 0.8280464162881747,0.05141863632920263,0.46919138241519043,0.637354032549213,
 0.5358269315183902,0.20200005949173783,0.6315057899827439,
 0.04194321974005155,0.9506116475424764,0.7817684021146819,0.3474371805065053,
 0.007918473943828164,0.7764858316235239,0.19093234526410408,
 0.024717954430210187,0.9555487442557036,0.5179243349509339,
 0.9991834456945199,0.15783579990841,0.19064277338409208,0.5275165675344915,
 0.48928989900327313,0.08387920249561376,0.6642041378134024,
 0.3279520253698628,0.39935370033770046,0.19494045446096153,
 0.7317922159268757,0.5802439556042683,0.040877303175506974,
 0.6946570255808802,0.5967434266266092,0.6807909757505726,0.9275600551195069,
 0.17410830908002584,0.12524019169500133,0.8539930033956475,
 0.4728714737329971,0.3196916207250571,0.9723343646386271,0.789179969230026,
 0.9687977152818664,0.7244734659865464,0.15708538793566162,0.4373291573658633,
 0.8433329781514247,0.5881401450964275,0.7773034490022442,0.7361381616976317,
 0.2894574112999688,0.8591302205101232,0.15034396756623525,0.9881355322305153,
 0.9845328120646499,0.17547100586440212,0.6255379955639141,
 0.13192630912224024,0.10483826128696172,0.6246870555470768,
 0.9831161542005877,0.5626679578050532,0.3982622373582627,0.1275980418908228,
 0.5011812002613575,0.6573978423136414,0.9070018196771021,0.41917246964070887,
 0.06514445349711806,0.17146528276488082,0.4472765219861642,
 0.02880626412130538,0.5163835191079944,0.1989739200797892,0.9483566985586892,
 0.6069114287299546,0.174012608865303,0.2729546956573692,0.2495317345962561,
 0.8905344614796552,0.0761144551380124,0.44929060653415676,0.4007006594860878,
 0.08694733086476569,0.7915627502692913,0.6260193757816667,
 0.051986266997128494,0.7810800655740531,0.11382926906316737,
 0.8326219795098742,0.39489515815212584,0.2720987570028136,0.669497981001677,
 0.954453484474028,0.8980010038906653,0.25825597195749905,0.39456611047223356,
 0.9545920384800739,0.10483025390755285,0.260410493410419,0.6130454035638733,
 0.04259143730539916,0.19387892993281364,0.6069779660681743,
 0.2807817474958101,0.951062970227579,0.7828056385783354,0.2156934643938414,
 0.11019346836877142,0.6079002552494314,0.8348837686082423,0.2845999836841646,
 0.09975412518179405,0.7757693209064213,0.9245532615228219,0.3301380585340503,
 0.62817126843662,0.666248268498109,0.1700859010658471,0.26634801202245095,
 0.4895243905484027,0.32473599014324406,0.9703176826314239,0.7907053522219545,
 0.9613088658198748,0.8190591789836774,0.5888833965828077,0.09202909518743352,
 0.6331922148871784,0.14305433629774833,0.6736599514471712,0.2338235611150471,
 0.8616183679142378,0.40820539352852925,0.0669493892992642,
 0.32516482353062715,0.7973034210080209,0.5631679620765775,0.9434155833565598,
 0.40037287791474974,0.2485700847709229,0.6307312879176884,0.740796366631969,
 0.8660050508881253,0.005379410339333757,0.2694017801373273,
 0.28324838119981377,0.8336960736071147,0.8673500278533093,0.6762044628911853,
 0.9856312320301879,0.7876484992683113,0.5215600125851978,0.24474104428090804,
 0.1895336044344188,0.49567953971024525,0.062148594363714116,
 0.1880575073163615,0.010320259629296724,0.21721975063331178,
 0.268853340573847,0.3624856247339604,0.05466203883293819,
 0.006011227352396231,0.4900266162456838,0.007392222694503792,
 0.6229168858117554,0.5076986084001652,0.8052482138763589,0.946634958624068,
 0.5823779119461002,0.9276975409814587,0.22441737658376082,0.9959655996534585,
 0.09945100336421853,0.19213922313098764,0.23464366890242627,
 0.6735562178527311,0.45551573839020165,0.676254896110654,0.4117341295795929,
 0.7281524572387369,0.4112056638818371,0.06465604361303856,
 0.028365267237518976,0.8269788843252363,0.06649975788501383,
 0.28853266783864684,0.7374217950403518,0.5029786136142133,
 0.41630435089243545,0.5666815754636443,0.7377807254690666,
 0.35861310212086484,0.5847791585688316,0.8827022955428343,0.859502518069835,
 0.4031749492984491,0.6617157106838142,0.27844074783663064,
 0.07943344992954271,0.8527839697100525,0.02138073872499957,0.957782431958258,
 0.2853227879941038,0.2698939146013979,0.9093370661416826,0.2966055220596112,
 0.25104921822178194,0.7116846505108448,0.6960713868566524,
 0.13314093762704826,0.45766330173584924,0.01710357649215044,
 0.3133265282847525,0.8394765201323907,0.5877743602936015,0.34594746152210687,
 0.565500413981233,0.20610801226804376,0.03826547014411397,0.9958238791116849,
 0.7761699674669496,0.2894059510090816,0.6505851991775845,0.6341709283097358,
 0.7857508103395942,0.20498051433438813,0.443508336454589,0.18041945418244554,
 0.8997278325219082,0.5843222293010822,0.8226534440171712,0.4986482986935209,
 0.3918162784151614,0.998367796056752,0.33558358378517905,0.8768101969531276,
 0.8097848745380252,0.37178033292043255,0.008517460881941719,
 0.2757472412337355,0.9882752812926263,0.29602304356836484,0.1133860545411075,
 0.6876819472773297,0.9351078036801962,0.44407939771021177,0.7957507681119942,
 0.2518924533072582,0.4473646299707713,0.7599976712846613,0.01569766478044321,
 0.41223601035513313,0.6562417827379758,0.38216149114555775,
 0.8463016408234136,0.9247685349598761,0.6876574016731567,0.7799484905781235,
 0.2662730094171033,0.6253600011932575,0.6909588401348038,0.43101166137685576,
 0.7755836019765258,0.20471071237421723,0.44853099192855084,
 0.7183633414043539,0.5895921818653543,0.7306459483569407,0.7574286293504942,
 0.5503370761119337,0.9848969585521692,0.661377459979134,0.2776580837132888,
 0.9418267513216703,0.14665614773646862,0.9707334100694521,0.3454732673326171,
 0.2812225842940079,0.1909835622957149,0.24751480176460972,0.7830840772802413,
 0.7368559369918035,0.41350978443221076,0.7450022399508245,0.6207343294710593,
 0.9481762286419965,0.10571622136197867,0.518276967799125,0.9731504867931258,
 0.36967439952914094,0.3894605006491977,0.1677994732485094,0.7161137468025917,
 0.4622813523638992,0.7128577242696335,0.9312679454678353,0.10901239852105749,
 0.17943574122619976,0.010909925081242666,0.4148315227724785,
 0.08990845972812178,0.7700792019922141,0.14058266846145506,0.275622316995204,
 0.810096721320883,0.5738755452745639,0.8040000241902665,0.31443251326597177,
 0.0016053632963494158,0.33420627060376185,0.32966793273519435,
 0.39391427103789045,0.9993412344104846,0.9618180071328413,
 0.23853241658399205,0.2926789464578067,0.014725210735683225,
 0.24226241470919585,0.5729374621236908,0.6891553728032269,0.7081875501988765,
 0.416683436702178,0.017281614935265477,0.5830949853391125,0.8020378091718016,
 0.371134813558766,0.22739675741362242,0.5867435544930122,0.8938958970817878,
 0.14044971619585578,0.6852743887860152,0.6227702618813367,0.8335425594367004,
 0.5222531788101749,0.8606557827354626,0.20520936628050956,
 0.018732354519466532,0.18024669514255376,0.24300949392985738,
 0.13051419113229512,0.5183504148283845,0.03674426356991267,
 0.42247704843097655,0.08266993518940047,0.08645254020561899,
 0.5171193358405946,0.3230435482085572,0.40198859721187896,0.5740460010547399,
 0.2059178276002691,0.7578651727590947,0.9175540759760739,0.44251715328486996,
 0.2853796649148057,0.30471451771311986,0.1411135874276871,0.5800309361300584,
 0.8175225472060752,0.1979370773107868,0.9169902996731909,0.33880636466531633,
 0.3703199624209135,0.20611286483230462,0.491003597484318,0.9143908773322643,
 0.45005243227397085,0.9958716277814461,0.3190558056793612,0.1453365500420173,
 0.8380936290103094,0.9986407886386333,0.5784906547309014,0.43164177951535043,
 0.16156107222994243,0.9244136286248572,0.997240534476535,0.40768070020274283,
 0.06749842225943548,0.2883820835379225,0.06326791211544036,
 0.4317948735764716,0.8437480909484698,0.7671829087073795,0.09004497184010318,
 0.7032373506026591,0.6378386885373001,0.3251217835299176,0.6002806456071155,
 0.47550951924188356,0.78508113544055,0.8807292805853484,0.05232109447585076,
 0.4941427100778949,0.9008501489193176,0.6668937009120537,0.2800545828901113,
 0.7487528599030132,0.24953022588175156,0.9451263020190208,
 0.16634508643387935,0.6983747251055706,0.3884816822249004,0.255447454584786,
 0.14873016336201017,0.2080660311686855,0.5543354268332228,0.4086178762075626,
 0.018122120179861922,0.6651439679342189,0.4418434577888438,
 0.6647921202185847,0.28077362884667867,0.7176800336678575,0.8079867612557097,
 0.7288061947455551,0.6264636747475978,0.3061563770054393,0.23342265940323514,
 0.7803149813807899,0.629588824880481,0.06586857796443757,0.3810715266951499,
 0.5542079036237336,0.3287305125902913,0.19480716183367752,0.4684715585706436,
 0.9921194961009043,0.36701459628241595,0.5627812621527268,0.4272111941024108,
 0.6451536266758608,0.6520327955602842,0.2867125621686122,0.9852013617810312,
 0.8432374853380327,0.27919471719638744,0.6920838444001679,
 0.024150142629111748,0.32552921087878706,0.10623196241403976,
 0.10675941302726588,0.9327738164601986,0.3801445345537825,0.4478660284593883,
 0.9108505328452856,0.3267022513296062,0.774450953383201,0.520928442627217,
 0.5159605651166634,0.10796966927324214,0.15986182079384448,
 0.6239029057430857,0.6804477344017489,0.4839634317880731,0.7050142853691437,
 0.5459324451409245,0.04432115273932624,0.632160336165829,0.9569574386030942,
 0.7160102881097686,0.3220883903582662,0.2697805357650612,0.12309796575053711,
 0.06796739112629313,0.660940909869847,0.9626086369845527,0.5149468474682428,
 0.0018682896849132602,0.16540162389737778,0.9815613388905088,
 0.5324787218231394,0.319143809957937,0.4817241086542582,0.8402917743831538,
 0.19895826655065207,0.4595029936153163,0.1086406206506032,
 0.08185173185297767,0.24648180534060682,0.8485730325887308,0.622267900019903,
 0.1549418026351791,0.7565013379068725,0.22098688892588647,0.6596213006015356,
 0.640375149281371,0.9111513227502887,0.7685969700025564,0.21609453764845443,
 0.2511391715529522,0.8651432646973769,0.7180912416301399,0.9182287839351274,
 0.3911869181342855,0.5258739444546988,0.5126159132830216,0.9935907189588892,
 0.7561057307651068,0.1405645707846891,0.30204811577927027,0.34723560252707,
 0.7398908731859122,0.7784371291120074,0.7296940934039113,0.12229342683830247,
 0.4794044134561031,0.21877287949031976,0.4408019794185568,0.9615599915377124,
 0.09020001826384894,0.3012235443022381,0.9075184797812766,
 0.04073391413144667,0.08788720542247619,0.7412521909507449,
 0.6519279299333862,0.26002993402585206,0.16248571543656942,
 0.9265615135190017,0.9810459019798234,0.636552931559106,0.7257825630385789,
 0.6973446646415145,0.7702320220574616,0.8056512192388052,
 0.002683901589821236,0.42295003851711543,0.8782826286964781,
 0.2190656955204856,0.8833574678790999,0.3449239035262258,0.16348001236952392,
 0.6026848333625129,0.8424828649116045,0.11659159034292643,0.7192232307161762,
 0.11416912351018105,0.9351532803313637,0.5157447859684484,0.3606090982085852,
 0.8118508591040701,0.32208038963813423,0.7915115030947323,0.9379965744892844,
 0.3661807176499987,0.8315530131339206,0.06891538361505345,0.7703098169334107,
 0.8169140271336368,0.8909576730614337,0.8551155230803588,0.622898876596648,
 0.2725910346054039,0.5176465761104069,0.42524190741121015,0.8300217688121864,
 0.7424993356958518,0.6983804612826127,0.18037003060436418,
 0.16395100047789812,0.25397229768886875,0.6953078355436484,
 0.03919212374916725,0.3499049319768912,0.9279005514387623,0.7076535862220438,
 0.6120943693732118,0.1573982186142977,0.8435946468026723,0.16577475120161755,
 0.549268080029226,0.834391335906162,0.6969417065582795,0.11400124216208418,
 0.5828467177923322,0.19706090224377482,0.40671250501875333,0.520141467615772,
 0.9290993271059013,0.7925026091789237,0.9194495442811571,0.8027090328943838,
 0.9331370338023527,0.4325822078435617,0.5703243634697696,0.6819720062313215,
 0.9427223435752379,0.5918911080118879,0.10879963154616668,
 0.04946018792187068,0.6494602663497542,0.6548978581401904,
 0.21603354261650276,0.6364435389078191,0.899156470658905,0.8042037579536209,
 0.4479144790549241,0.4824984778632142,0.7587375494449455,0.6470663720806331,
 0.17973300035528295,0.5004160001658021,0.822785180041783,0.5195643961507869,
 0.49542257684193003,0.8873013481738284,0.2806224473569108,0.8633721455455963,
 0.31714630647949016,0.8150524790597135,0.9358785094958229,0.2889949284159402,
 0.8191042033007059,0.39216956185845353,0.26115221478239004,
 0.15436563487531196,0.670759330724294,0.05320097544481406,0.9276448928786577,
 0.7204977797414852,0.8773585608923027,0.27816937822421683,0.7805323981745366,
 0.5844699461532672,0.4764788322370834,0.3455642895127701,0.3284747354825063,
 0.5637341504064253,0.40928036670153123,0.4708415513794242,0.5837413931234008,
 0.9360289223177143,0.9523253903455589,0.23991636256721427,0.4159478480850338,
 0.13697167389404796,0.6539658459031006,0.6498358867416865,0.6039089479041496,
 0.21995728814795168,0.1373560918891924,0.16258122775341743,
 0.6259152477124584,0.3145321569467814,0.7573178415425493,0.9304484399405684,
 0.2296405229763493,0.4777015067215473,0.3165940147553268,0.18705172730122632,
 0.3911251308905561,0.4319494077929993,0.2194017220279859,0.7685819270646957,
 0.08189488345328555,0.2593672808614723,0.7063161598942256,0.4859507775629016,
 0.7533248162565598,0.321744904897141,0.2628686338379199,0.3351221777148412,
 0.28858883124818513,0.489469904321014,0.9034709731293962,0.02574812479841193,
 0.9916468527574136,0.48638469397209316,0.8086675222096408,0.4900388043916535,
 0.1139633196301717,0.46926587522579455,0.8759194259569194,
 0.08889036519195148,0.8412865191931247,0.275741352222138,0.8207751936998647,
 0.9347990002272789,0.8542324891471614,0.3952935950757318,0.6526845663020303,
 0.6871622598722719,0.7168146217750441,0.8053063599832687,0.7751316237808051,
 0.7822159546790547,0.05082202158330651,0.9264990274956281,0.7231760965819772,
 0.4941524612845922,0.1515279400490137,0.9953863525003106,0.23225298034347852,
 0.6709378339767162,0.2062753844006897,0.8530857870171226,0.4608971386468964,
 0.8425529184510478,0.1754527013668068,0.6967791458202297,0.24556446178920532,
 0.7898781546679371,0.12419003915922289,0.0964268105907623,0.705195485625203,
 0.05743621125335219,0.08573077439839183,0.09951918742866961,
 0.7409915723994642,0.02222045989517052,0.16479934952238784,
 0.3515541103642553,0.7938970184860303,0.28918688635556467,0.9776804283539825,
 0.6745072406670275,0.04223573368295053,0.7977979431471182,0.9015048915910706,
 0.8414132194212937,0.6089992014401006,0.8706339850332983,0.9181342976060041,
 0.4606491932404352,0.8787401141722005,0.8099486736325444,0.13215087167400563,
 0.877555829764247,0.5024148392484906,0.6635206861118066,0.9623399754375207,
 0.1783248862196124,0.6723021288602454,0.7769771436553653,0.19770354371292387,
 0.4659832083711446,0.33779705952193395,0.4746912703973387,
 0.03559638610214966,0.445715578190941,0.147508093874805,0.7188032590065884,
 0.22701469542998987,0.25703133338812456,0.900929137087538,0.1116457087637921,
 0.47128077629176646,0.01665427259941621,0.10754704668087645,
 0.680191920670433,0.799481881447358,0.46219191470712806,0.32843465737298516,
 0.4506948132182146,0.14271757688969988,0.7928431068615899,0.8153280126741337,
 0.23267121301391847,0.9720225553389865,0.8887233623649238,
 0.03684492251331706,0.19484286317795751,0.10984112260654444,
 0.9705329993214589,0.07937430485935759,0.6261540275226046,0.8387361032148074,
 0.12413773574769427,0.7341235954282497,0.9641363062039758,0.2877042927505322,
 0.6487399470469241,0.21494628197093746,0.9768598202030077,0.689689228140914,
 0.43444227983465455,0.8290182790260293,0.514082895543245,0.9372668219802079,
 0.29696064196571204,0.531714674112526,0.30227782505707956,
 0.09511612352198173,0.3975304930369131,0.34570373568567414,
 0.8265703187219373,0.8709028286837014,0.5806194281627639,0.3772924208519932,
 0.17642954190397497,0.209276668939681,0.8662215931094444,0.6483410655090045,
 0.37616713732662144,0.916134242548242,0.9638391837465938,0.9970469484979212,
 0.6974784791649066,0.36342883798356984,0.04335968556502323,
 0.9902553514910315,0.7354697471116217,0.8094564158490396,0.7730985149182819,
 0.43113569521882955,0.2143194889842931,0.47669365557864485,
 0.05879163216661398,0.18394317283086226,0.11248682959518974,
 0.8869900611710012,0.13783372297653085,0.5202622442778014,
 0.020110298268941507,0.10929337288645014,0.8939817943255142,
 0.26449114502759774,0.06053048203210287,0.40504149878420626,
 0.9613388708642312,0.9516430753002313,0.1687103599450055,0.1417906627370057,
 0.5031385665093532,0.08874362288228199,0.7886134244109477,
 0.48873112408535313,0.24357265311394805,0.8052412413239558,
 0.5649805181150721,0.2569546015104789,0.06492900373410238,0.784257738509229,
 0.5814414777888471,0.9428273086181409,0.6290144414837588,0.7697038048873378,
 0.8059932725143326,0.7235494403273505,0.04146976682842296,
 0.34560467881380585,0.862944998604798,3.688103532325515e-4,
 0.7613029293040624,0.6350879150549515,0.8432902899372552,0.1785805657250933,
 0.6554594716239712,0.7040117296979346,0.4895017883074161,0.27942340400783505,
 0.424975829428604,0.5963262655876693,0.8003866402726962,0.5116935300121533,
 0.14992183903863765,0.07693729447573772,0.09093074826096792,
 0.3136558832197267,0.31976600397875843,3.6550914537114565e-4,
 0.4599257522781236,0.6861733162791992,0.6576979732681024,0.33679128079992315,
 0.6083558327561616,0.6211095672905235,0.06350497027028368,0.5974096853385866,
 0.3352499030984013,0.6102223564499176,0.4180869922125774,0.8859980848386354,
 0.47792252384036993,0.15321041196097374,0.8032844701103072,
 0.31171749539237603,0.99600830136781,0.9514963187198572,0.24030120275439892,
 0.7817803511814573,0.5844547286502847,0.8632203273417272,0.637620492184613,
 0.7325031564820392,0.6888672934765839,0.7466518369388452,0.1339437130767559,
 0.5258765676890116,0.6311672759482732,0.48914156514503926,0.4653969893194845,
 0.7988886464709583,0.23530928624482517,0.6055564029551298,
 0.18131698725893552,0.5223955821793249,0.8061713698823613,
 0.26721811096921055,0.509433723162887,0.9002025434552667,0.4851706412971758,
 0.6373576476196335,0.552776610866714,0.5001363895271611,0.5833174670180842,
 0.285630925186517,0.30237581845264694,0.1809634009843415,0.6335452586351126,
 0.07963803726026919,0.9615137831048044,0.39364163693298126,
 0.42636221443647093,0.04647869758529899,0.9291132054891018,
 0.5256190653100496,0.7965563048821755,0.41445610989954595,
 0.12145645936920935,0.31527516009522105,0.8006130676316642,
 0.08005200309232852,0.9116851643941121,0.10368391913354036,
 0.4452483713963862,0.6832803734210053,0.8857239237356466,0.25602819605399574,
 0.3912226547293769,0.4889025167846316,0.19791376568950148,0.6699497945906671,
 0.9476091348450603,0.30140830071158675,0.41474459142223763,
 0.3616901035378345,0.30492313421320316,0.684107417456956,0.8027639651319658,
 0.1265225312794227,0.6967898132290693,0.6710318834128188,0.17088061249879605,
 0.8549814809341356,0.8023752397468032,0.6316216072674091,0.5051689002040325,
 0.5973024045306418,0.9363660200860958,0.5010792359370755,0.24485242472026403,
 0.8485084518561539,0.14924959547857053,0.09101084265715154,
 0.2863177894748221,0.7154742401286052,0.49366795966081223,0.9636909609389063,
 0.16899412390293944,0.21528927623380145,0.6896242232562911,
 0.11611222048100023,0.7992336941830008,0.16699882916170006,
 0.9482173794723305,0.8701698508034738,0.44004003775348066,0.5957808836840164,
 0.7698454496646385,0.12131461648184338,0.9605387912988881,0.3311635381726594,
 0.2788983588132339,0.5378142160342295,0.8768438034243008,0.12137106358674998,
 0.24906883122469703,0.5616434118647026,0.9272078273038364,0.4281956686022821,
 0.8732152978760814,0.4706762578841538,0.6490916840004,0.9089491051752705,
 0.6415783128297352,0.002397837319502605,0.8411994948959403,
 0.20634129580654703,0.3070683307592672,0.5334094183924856,
 0.008384020707456585,0.9830188353683929,0.21636490890039162,
 0.4583551394763723,0.17232617952468077,0.8718979254888752,0.5219377617991687,
 0.04476458103953129,0.73573614792935,0.12130932860393673,0.02911357754920796,
 0.06708201274073666,0.8112982062784653,0.7940770261783104,0.321448613589736,
 0.45442869896886584,0.2588418002927122,0.5520502342092506,0.174234376710396,
 0.7597350384804132,0.15775508740201172,0.979462974773532,0.8395182307898663,
 0.40524053887314193,0.7280757794389098,0.7708762415702075,0.4300169737053595,
 0.25968366912576446,0.9408012040685996,0.21322314482816374,
 0.7263189186825513,0.8390339843511426,0.8458596334499132,0.32692783684667504,
 0.06787425100137823,0.5567302366300644,0.5887325457298,0.8111224887663204,
 0.07655590510199128,0.4059825010157151,0.8163886880903344,
 0.47383058613606044,0.2545532768580271,0.18686704201191529,
 0.8706783889727411,0.656623760376418,0.40673230739622346,0.5877015593794329,
 0.5089057619836778,0.778238480740702,0.4858622234425216,0.17812587401468505,
 0.9721483734129834,0.9366845290327378,0.48025690617117955,0.5297810297285065,
 0.6750200139107905,0.49428669720009233,0.4361977619229417,
 0.09624658018217924,0.08735047467765644,0.44090824223073743,
 0.1951628578921164,0.1912841734081191,0.03977278328382927,0.9382770540121383,
 0.8095473452222745,0.49792332688485585,0.012291169763196663,
 0.28116807992015225,0.3567943435347388,0.1432942868783913,0.5826972442278672,
 0.41104869739264316,0.9337344362002935,0.2699721193483058,
 0.10451262859969246,0.666769602897133,0.12431215949073726,0.5497485532722535,
 0.48890487842991137,0.5306584882889378,0.8210178909706031,0.9207160328147118,
 0.31633849346832743,0.6055508935380022,0.13392301372348447,
 0.9061204540759591,0.8673659320563466,0.6183079001106956,0.8977871162837843,
 0.12653006590203908,0.6146398547809131,0.5300471595113421,
 0.08230422036053009,0.3513589533769318,0.9519705695358642,0.3102005460229802,
 0.2894374261242909,0.2680066982974263,0.16145648267214863,0.1683753017386972,
 0.29975880349634376,0.188347246561128,0.8969419122610932,0.703736916065403,
 0.8163954571111898,0.012668407054205355,0.6034752870713247,
 0.33816484498109045,0.8998132744608873,0.10078398512333164,
 0.5478054474379764,0.5995904675462091,0.11965362861266438,
 0.41696024539775733,0.4306423243769052,0.5076250350419502,0.6056136146891431,
 0.4470000027467782,0.7776552657238842,0.6696308564214174,0.2635887884901855,
 0.28110051189819485,0.751768031709743,0.5528349890460132,0.7590011427118941,
 0.03404825949254797,0.4937849839367723,0.5392548285135229,0.3639101748017213,
 0.3528881496700549,0.5165354395659996,0.3562203352046991,0.7371100888197925,
 0.9449254175412888,0.8712872031466322,0.09672098090787185,0.3196541719182603,
 0.7774530064622092,0.7258878655271778,0.3270642128111826,
 0.017588417754277996,0.6517829448412419,0.940107495401187,
 0.44524665239865324,0.030065842399545684,0.6084656303547014,
 0.6955865455910384,0.5939740882355643,0.3423604040458502,0.5299687041505744,
 0.8360968248506877,0.38462991777056454,0.08802838955368097,
 0.3171346274300473,0.9657857292075454,0.7715748684977364,0.6042663121030203,
 0.7606456839146306,0.026218665903498972,0.5011543369057373,
 0.5560661011757198,0.2578503445535213,0.550063772100117,0.7244384414046976,
 0.7415077297109631,0.4688307986970831,0.2828916401167164,0.10620766870399379,
 0.05287206687394408,0.6057475860133753,0.7850879606521902,0.2892319363440632,
 0.43906963699890833,0.4132648115984472,0.6353600515523263,0.3855273639743364,
 0.4126429332996149,0.4994224339197447,0.5819110125979787,0.14864148648163167,
 0.5096259110469529,0.05384319956259609,0.5557048218917393,0.743267603371546,
 0.3645504625144108,0.06595153274552135,0.17414103965939765,
 0.5454078217794389,0.5426307806103045,0.7608706717647816,0.9443638588234143,
 0.4901798219123218,0.8736038723498546,0.11679573275157173,0.8590495353799888,
 0.14690181742137787,0.524411542423315,0.8709267981087709,
 0.025833152632675493,0.45839251748410903,0.7398996986283211,
 0.8960487650609081,0.8724179787867782,0.5159491798321013,0.5199233569334241,
 0.22867187452594928,0.27677205298600305,0.5531926918985896,
 0.34076797140340553,0.16832052548402066,0.9728046453374163,
 0.7781166749614857,0.8492460857015383,0.05021577147808198,
 0.04241623410115947,0.9005985363355588,0.24580935679536786,
 0.06628796583121199,0.4608973479850299,0.1323917933172074,0.5925554851914354,
 0.7848591666377449,0.3326072907367742,0.0017690030085614206,
 0.6731254043018626,0.4024854264426875,0.47653752043098463,
 0.12292946663665127,0.18621752072260456,0.5912026632786145,
 0.6296946107444603,0.9845282592337967,0.9108374160864627,0.3400011345010455,
 0.0641088707128099,0.9360337267118752,0.6347517885351955,0.04797962932926825,
 0.4575689438250069,0.9485659269937925,0.9439837106274954,0.6591221970264552,
 0.7318656448056898,0.5016102703104962,0.6723130248146799,0.7800344402025625,
 0.350812591689339,0.26311003225854424,0.1408816113373712,0.06906730304213826,
 0.5718878358818905,0.3357468159215453,0.7461601689282495,0.8294590450046352,
 0.11406902653440787,0.5202781467407513,0.9694515873432893,0.870742163812649,
 0.03605380367352162,0.4085584998028434,0.31539254248231485,
 0.7447420563107797,0.4618499786148266,0.4710990602296521,0.7242512321263468,
 0.914268202346263,0.13549260757648307,0.26170789075239487,0.7699263095907999,
 0.2885043289009225,0.015423873891671658,0.5498054301988855,0.940042133586114,
 0.0973765943548418,0.3081062725402979,0.7968373344206214,0.8979658218670304,
 0.51423959156194,0.6397285164711506,0.9746071258306512,0.5425772941394723,
 0.7078807062138224,0.3382741030354318,0.4497457762180358,0.07030197294260532,
 0.7822389797640679,0.6479530368491221,0.48825150406728524,0.6697863578395624,
 0.2549008442263314,0.8529322548134624,0.7705737729341487,0.7648811352712445,
 0.0783266640558764,0.8673901599974747,0.40697503286182535,
 0.08222971324094697,0.34758469364349576,0.3260112025703763,
 0.2949549609631188,0.8686518009144926,0.8141812085676395,0.02915992297300063,
 0.48813427621263095,0.36645862576197674,0.41640041064417166,
 0.6423177265597565,0.3113088818748344,0.18342876057034663,
 0.05933707585228187,0.8158309124939747,0.7576076006467347,0.9516268149221156,
 0.5556219507321567,0.1366584984932674,0.23575318685108748,
 0.41428603148714616,0.28863345174770305,0.7276081425230854,0.903638954922829,
 0.9262868030634313,0.7825523713936572,0.3310012246374825,0.8572296929542991,
 0.5021948023012734,0.5185430928251806,0.42876807376717996,0.5179939753624807,
 0.6688620893182601,0.6268439508245468,0.3501102070021621,
 0.012000498267516369,0.4956918853308181,0.4867708824761948,
 0.9312320347447575,0.8816166523474551,0.8164321828746586,0.5254267218803222,
 0.17257463503857495,0.5608606681528097,0.7040874649359825,0.5797463421270259,
 0.7538095783636879,0.13722037286483935,0.1238839674143618,
 0.20620298272495197,0.09374846467634668,0.12288969496915003,
 0.2566900301777135,0.8701217608124607,0.5133186449021242,0.614172900595508,
 0.41546880674372144,0.059695572043306,0.8724550534414671,0.6106086401740964,
 0.4751245722430857,0.5748694619719066,0.30833335077289403,
 0.22108734790710738,0.9381190785342025,0.5725890808865286,0.8602468656237784,
 0.9486260213799796,4.268062493245628e-4,0.7568567780024578,
 0.45921169420290076,0.9745858795229527,0.6559878970711337,0.1273531359837876,
 0.14088009026723713,0.6125910752074154,0.8631424893705093,
 0.18566793715579444,0.8114176907271997,0.7126026965905143,0.8293805402521979,
 0.8654821627604751,0.6312160805383705,0.08824258999547663,0.5245331732686497,
 0.5293003000957243,0.33449086716852805,0.8372537957765431,
 0.46467680216190244,0.760044448064921,0.1245501542551618,0.7025468522763907,
 0.8505931669247189,0.5819454161805602,0.13576126908446395,0.8175789230092187,
 0.3647828222650462,0.2541801283847718,0.9605278402750475,0.2709604720364923,
 0.425631713843572,0.3883048736552095,0.44990981025621246,0.2221506494337553,
 0.8086683313388043,0.5674035601079729,0.5192123393411606,0.3154512110546116,
 0.5120850263261472,0.5810302278327221,0.586431649606164,0.5871055569754715,
 0.17296224353408718,0.6464281294136265,0.4151473892333505,0.6327938170875714,
 0.7174392831801389,0.06828145848463751,0.05936325906094764,
 0.6150742262558058,0.4595749028438838,0.43288298873767017,
 0.16604740243031568,0.7208804205701315,0.974276414491494,0.262360213619241,
 0.0161036795221412,0.5203646343397544,0.9393884645366537,0.9812033834110957,
 0.22099563140290446,0.111583927438068,0.592330531774279,0.4536598463750193,
 0.5590932018307404,0.044212315270574676,0.43366073911181136,
 0.008501669399794687,0.42203840205154697,0.8234672199302987,
 0.35961662251590554,0.21096802508911394,0.05480389288993481,
 0.9058294497229776,0.6413555586228787,0.25517308934413974,
 0.014639538025038479,0.10883571026965988,0.3663608422699488,
 0.7628850278327258,0.36637875760030647,0.3746416304763317,
 0.28194596987714526,0.4039446976963633,0.3334747851110017,0.296479341066497,
 0.8961109695509679,0.5440476703486442,0.5997258190945869,0.20729617987593507,
 0.3021733331952041,0.20337422774454184,0.6112296595808566,0.2615800163319991,
 0.33824350254025637,0.2349941407799061,0.9113883426341681,0.885559717697205,
 0.5868810278059103,0.784114518395991,0.9250462826989748,0.6940419169617913,
 0.7706585006112227,0.8400652536705868,0.5959905382035283,0.6263365213514018,
 0.7976054136960588,0.3225956205741216,0.41850869005401814,0.972887163222647,
 0.5962381708472786,0.6572128793285474,0.1400255527474975,0.9713824611135675,
 0.242212729763848,0.3092642494140494,0.7479407867103227,0.5324398519902598,
 0.19420640526376476,0.33378404806977247,0.8874126892247631,
 0.5256092354774702,0.23175470562469314,0.7394213646345762,0.7306721029634342,
 0.23384285204251842,0.15820586745498233,0.37096894824040993,
 0.4555324752018153,0.7699421214664957,0.904215435861567,0.7918154510497613,
 0.462991371058721,0.9594605532473739,0.14115455580512482,0.4073083652945422,
 0.7859965304117912,0.9266878825810909,0.06243144986263183,0.3924136393812786,
 0.3824686651004494,0.6074001005870475,0.4870004089996002,0.8471335315186411,
 0.5557956833587148,0.9509085063586066,0.020625749990035414,
 0.1402969623323227,0.44194599803489987,0.5129367009483867,0.5950144499983602,
 0.717638582211526,0.505758751742697,0.5324541332911585,0.608532873231899,
 0.6100296626897123,0.43101321105679735,0.7475636009812718,
 0.22663571567701868,0.04395607836682858,0.21519343707344263,
 0.12367867440024694,0.004282972728555379,0.8796922364644103,
 0.17373668481043603,0.00583063099425285,0.3742171859646336,0.768235706176333,
 0.9200272919217264,0.2583931444215368,0.9866692205843143,0.10706290416931996,
 0.807133743149759,0.7821342349874151,0.9358686026911699,0.2816122259126481,
 0.9576400356810922,0.7564733782431886,0.8224947590741802,0.38462982007812896,
 0.17985423095793385,0.07722505912605326,0.27688590979831784,
 0.9774042204523503,0.6813058700968223,0.27942234084309026,0.4004609540155729,
 0.5946957603655715,0.35582898730054113,0.18870479925861405,
 0.3297282205066904,0.7667756964226948,0.7014085210564134,0.963282417418892,
 0.4474574515661722,0.47056734970313363,0.08862075159654947,
 0.8900688926542978,0.8774843709617173,0.1665571040250885,0.2618948813364059,
 0.7157852302086363,0.06894002728033588,0.9723829405705758,0.630383869821451,
 0.7573355508532723,0.12713929359997567,0.2664015756764284,
 0.44115074439028534,0.7198495110502725,0.9694873652630095,
 0.18064357235081385,0.6047859622117789,0.14823387356583784,
 0.06684683233005029,0.12573716258910006,0.6969446396050208,
 0.14500977669322768,0.34245650117247395,0.9198607983807068,
 0.7917813799623115,0.9686533236077858,0.17246780657535765,
 0.019162544159651795,0.365023257736495,0.7090198801440927,0.6642826999593512,
 0.9481848188145869,0.7204394760750202,0.4531139116314645,0.5596286813165282,
 0.4323985380216294,0.3171779932238282,0.9131334922526934,0.49083987704922283,
 0.2100196876680107,0.5243977753413381,0.7529968757919272,0.7694580248829053,
 0.655188743554125,0.5726388471950499,0.4565572124923216,0.5030423290380052,
 0.4992100124933614,0.04272523768229297,0.18676135974097696,
 0.29704314926188646,0.39970963466944387,0.35472890391843104,
 0.296986000222339,0.6858358543008231,0.05381163460934746,0.9976145193849438,
 0.8750155915530455,0.8423956392171981,0.3669001340135203,0.207133972670818,
 0.16026539061016465,0.251465011290534,0.8783580859024982,0.14966497904501175,
 0.9473429097185028,0.6686789538591222,0.518570142432465,0.44631531354477516,
 0.08983489756547014,0.4775706870555213,0.6243164596734965,0.5040877306492197,
 0.03393050042640611,0.25836363420794417,0.8546083608608416,
 0.9955977060148264,0.2623977328858078,0.5606920828276207,
 0.025168554607846838,0.9394388078660022,0.4878778163527743,
 0.6861379013926785,0.7019013142447552,0.17989415229021366,0.9231149218370652,
 0.10729162592285135,0.9467950929542682,0.4847196763666701,0.3042182974518933,
 0.9472693995214758,0.8411024398545074,0.1925942053093106,0.8769034753342447,
 0.39161169817044617,0.9599879991921694,0.34451447653081857,
 0.7738279448260332,0.012331332336625422,0.28553038954804155,
 0.736359128899499,0.09479672650466497,0.5650777942714686,0.360525896872561,
 0.8755278406507744,0.8005128187616442,0.08381531708055268,0.843904585534841,
 0.9397648889356953,0.6253825055544732,0.09196574454065964,
 0.09427647436912623,0.7724189002028866,0.3086673627131262,0.6070258440362839,
 0.33510045072517525,0.6310605752410928,0.7959121112062832,
 0.43647375579180603,0.047809600379252526,0.5358435951019374,
 0.6559280301129272,0.643762833560012,0.6135243892637662,0.8028509219782222,
 0.15749119937696832,0.0316967365470574,0.5567365709972479,0.9734522044379994,
 0.8622243759458429,0.7794641393037893,0.08242321584920831,
 0.39225266883980936,0.9747185368192092,0.2048697556728757,0.7298154417226925,
 0.5133511800647581,0.46678580400769387,0.7737931495941229,0.5893463077159432,
 0.4985602907567106,0.4185707675758399,0.7339354190738149,0.6004659562251994,
 0.5248355476283343,0.5094720856571758,0.5596615102669279,0.7301236580068459,
 0.6135092673246505,0.061626004100932,0.5239770653276895,0.1522686957655679,
 0.29371655260542784,0.8078745199847811,0.7044806395367372,0.5351983614357864,
 0.9420727218881505,0.478154112864704,0.13980436874034707,0.9653020626379828,
 0.2853793454893142,0.4947803590395118,0.8282794659942327,0.05576064314425677,
 0.6372677411433995,0.3308109793832844,0.12913058706729463,
 0.38397746379896125,0.9191227596834854,0.18853754139962298,
 0.7587903144473973,0.5648783720699204,0.6877627330201158,0.6280612894031187,
 0.3808272500791213,0.6864436107586687,0.9077965948213366,0.3791394593875852,
 0.6408191420808317,0.7027605619476609,0.768579943272613,0.8571539315565447,
 0.6514951716853452,0.34950732032012377,0.5009707711890751,
 0.43289392080733924,0.01362259875368188,0.5595430193024459,
 0.7641618550887618,0.7390524604254279,0.9432734268471014,0.46103759942908373,
 0.5849130972651082,0.8211335531394102,0.2005725322578792,0.9475974351401274,
 0.41555303589704495,0.41716142016939073,0.03261499595579398,
 0.5253155803726076,0.42032466202339447,0.6327658228724925,0.5683872040820322,
 0.11585339251604343,0.36674810341003716,0.38527923310910184,
 0.14695486556887938,0.8464201343220227,0.38785463914849816,
 0.20364373647707912,0.09621145761840444,0.3272541466746153,
 0.6154019532111199,0.5654243476656864,0.12401991852661687,0.8184756875981011,
 0.2232126418769964,0.29996660488968785,0.9278249376916289,0.7847248378993293,
 0.319717298581794,0.18356468624160793,0.02969711321434254,0.7620255694327795,
 0.8699224600709822,0.14399176765476407,0.4509166301674725,0.9905875501411223,
 0.45564368508516706,0.2821168387071564,0.25044551598285725,
 0.024280954616115125,0.13718313478640098,0.10673260485653768,
 0.8465282108706624,0.3731863371017774,0.08137746133480905,0.5995023631550734,
 0.13847662917666959,0.062031408660687015,0.8414584076643092,
 0.37706545035637773,0.9279295441611538,0.8893727051885838,
 0.47883124467098903,0.9449191586854933,0.1609141160029698,
 0.19557116490102755,0.07442677273155429,0.8483856775242726,
 0.8543385761527107,0.6124498593918433,0.06382631561102337,0.4835906361529514,
 0.6398802820443067,0.180525988901298,0.21351615122341505,0.8189071378829569,
 0.7915473698868036,0.5403779793563408,0.514937483832396,0.7810316727173451,
 0.6871759752448608,0.6185869103135802,0.5149025839642727,0.7697244951936828,
 0.06141595410165834,0.8162530466114629,0.32921078672982673,
 0.09223718340032405,0.41719597023108734,0.832396632830517,0.8198098405969825,
 0.17923078797218395,0.948367579284187,0.25737232570254653,
 0.37659491161207503,0.9365825548566898,0.10385254412060141,0.798248529925427,
 0.6231072583854474,0.3365492642577177,0.8529882355864794,0.33418173777186555,
 0.5316252451697924,0.6443092821483116,0.0909528249120467,0.5995933595312859,
 0.9705226652929339,0.8174341005225426,0.8808867347043292,0.6042579696585964,
 0.42736704289584126,0.9241833286642924,0.5590089410700262,0.9052320004605082,
 0.2628465103159714,0.9550375280688215,0.1593672775062508,0.8014929368867038,
 0.64584576638895,0.8788714549470954,0.023849214974589028,0.14528140765095476,
 0.35811266391426,0.22441417859768464,0.9726047095632756,0.07942995912807871,
 0.6131665103747205,0.9560464354688973,0.9762346391064527,
 0.012675279014029095,0.9021615957975697,0.9335926950057569,
 0.7315692889123866,0.4304383728405261,0.08235206974793052,0.8442167413974313,
 0.6564120824767876,0.34360868888727214,0.4948944356376357,0.6653491410456934,
 0.16183106965341887,0.508197691406846,0.7007393912154375,0.12874474804226588,
 0.11636153240401192,0.3485238628612637,0.30950684067483314,
 0.6072344631263429,0.596944551859503,0.7917206583318172,0.33558288944537085,
 0.19258372232615184,0.555348119538681,0.8374411924575129,0.6431379991914442,
 0.39800899371777687,0.7313676758736056,0.25028799554451053,
 0.05338248445166205,0.7695121208532587,0.7998369520731193,0.8552776328170656,
 0.19063705575177103,0.0870350765866531,0.6509361126531823,0.5414760586693228,
 0.3825788735972798,0.05654310864610568,0.4461433670679563,
 0.13046577128243197,0.6654977268759327,0.8907458349820088,0.7447287793513322,
 0.4969065603972699,0.4012876542243601,0.6569002607864984,0.8558980269786529,
 0.05049018814985473,0.3050343335828627,0.06947885859506331,
 0.006682890741756653,0.6107648508330815,0.07895143454447195,
 0.37677505822723645,0.2047865572658667,0.12487109148582376,0.951077562877563,
 0.2040254663041965,0.16170294983511302,0.8512988574732739,
 0.01638988860981494,0.8256902673333348,0.5209980670710318,0.7057974157624098,
 0.15396395871807012,0.3962933679514582,0.06282846459597236,
 0.09119984194152819,0.44066865730626303,0.5260721301819701,
 0.43325067161461406,0.9502971020403205,0.08032340020499584,
 0.5083356713002014,0.8924276046750843,0.03186075324401194,0.276816770306208,
 0.562758407576571,0.47154227163578244,0.7750754439696467,0.8813954220603067,
 0.12418845737506645,0.8860092915623153,0.3972998793795427,
 0.23118531701794287,0.9358174751587562,0.4185849625099487,0.0676885734988728,
 0.6746820398008934,0.5725101619249069,0.9975118065053281,0.9248163328157414,
 0.547912315575512,0.8734710506528601,0.18618298749957984,0.4464215474477258,
 0.06837149005469145,0.8203304642717454,0.592131485211289,0.5809139767035849,
 0.35380444515739473,0.33658434851051955,0.7350459668372205,
 0.3472758444135102,0.2909332826622869,0.679926685917515,0.271930852670363,
 0.2416046082326776,0.6023529598684075,0.5408960204867732,0.7226259929428933,
 0.03053075716658693,0.1431941957795101,0.7610949925811012,0.6740283782417454,
 0.6060532457120238,0.9130305940253254,0.537648477662176,0.8984461379512864,
 0.5107015812433939,0.8897315038857729,0.5656515960716472,0.07458787387992771,
 0.5665646519423605,0.646595062305589,0.6959162044512632,0.19140145489262417,
 0.5574395992616838,0.028259992315434967,0.6462317160145933,
 0.6388965606995668,0.5242130485317447,0.43933917873402084,0.6986460149695379,
 0.5068045569032382,0.7749635181176036,0.8647450746277909,0.14530547689077544,
 0.7134605216297812,0.5131290976114568,0.694661018261437,0.7980790884528792,
 0.687368884649729,0.10845795959584381,0.0024077912214967023,
 0.4000702290685243,0.12645518010498846,0.1922423394584598,
 0.10191682665804391,0.4429442980772096,0.6330128591635896,0.6806364645949915,
 0.2853550132326842,0.15837678052513304,0.3735525160740333,0.547903384040423,
 0.4047996232850466,0.00587065388149266,0.029469414151933826,
 0.49768914769597594,0.31375866468586666,0.28068892698863457,
 0.4876449488831417,0.8091141782995614,0.9896230056484377,0.08724053923171271,
 0.6331282928635618,0.7865993215693619,0.31849339859076864,
 0.23037766053181552,0.341316653825842,0.4445855733012364,0.351557926590629,
 0.8593416821214769,0.39134241867884023,0.48875297876427704,
 0.7399876757081413,0.9048452904249076,0.21628911783470117,0.5887079758020625,
 0.5710240863379527,0.20551198407922389,0.5958280181226852,0.8106196256176518,
 0.7756154692185722,0.28993179443321215,0.8525470501186451,0.8519691057072961,
 0.81810192719417,0.8099280131478686,0.5575156973842568,0.7050658187869802,
 0.09467643859384134,0.034920277851853476,0.5597146157467848,
 0.9815907961114851,0.3592880903193074,0.8987133164523959,0.38033288746485316,
 0.07452230097853696,0.5157119343475794,0.6759219253519997,0.8790292287238182,
 0.08373801029427641,0.14300624549754715,0.884302338155984,0.8658677235271203,
 0.2271412277428374,0.0996352720362862,0.47956773021389953,
 0.33134874483498433,0.8312140221474681,0.7898981382735687,
 0.45256257840076586,0.9851989987790856,0.18673234947465867,
 0.8417023353563511,0.2187185095102454,0.7554163024898382,0.07861729885638746,
 0.9791980377992701,0.883749490858309,0.49296893198353275,0.5977755135678884,
 0.6436217147960244,0.3192937071800308,0.3170197871470348,0.4204283714494411,
 0.7688907812318392,0.9854008740934346,0.6034283710418324,0.8890486788445735,
 0.47808859560620764,0.17865389666784082,0.20468630093731954,
 0.10796679155879108,0.682150019712957,0.5569165618017707,0.03507739949478428,
 0.8381141936867391,0.5081644166742569,0.3825748172675274,0.13333921053596265,
 0.15639676652769596,0.1941452395621326,0.029660161098261084,
 0.847445241654503,0.44188000470829636,0.3381401813866004,0.5837077108274533,
 0.2685695082025422,0.6837058554502277,0.9422304461524145,0.2435037160302973,
 0.3986207640866042,0.43447438522058546,0.9566985434537848,
 0.09445937975801888,0.6693739167395544,0.5474328913215523,0.5703503805106508,
 0.7179824533788159,0.5091508087312092,0.5233049816830183,0.5406035663581746,
 0.17727903074948725,0.4362160432231307,0.3034012162984572,0.1552942494799755,
 0.08760512234679252,0.5581856378892762,0.850780826150061,0.8409347279444342,
 0.8855627536769112,0.19029674087074244,0.3249809408383276,0.9410049956036509,
 0.802085218619399,0.5127586741692418,0.953383262383129,0.6173670330675771,
 0.5606631761669014,0.8389094879600227,0.770288700191107,0.6993678020038356,
 0.8096286039390707,0.4249206425587634,0.9147491332382207,0.5040627473038676,
 0.7848623962904073,0.48503662960198923,0.3115252013572588,0.8968828940156418,
 0.5792229279853484,0.17754071033329633,0.9262969496463858,
 0.19920509035867895,0.7822643901397258,0.47315371725809074,
 0.24202379591002665,0.1772526546075145,0.059589159856496154,
 0.37916855615974343,0.9720552015474494,0.0788024949252688,0.9313679858548533,
 0.015098749851576887,0.7693765168849279,0.14518487370945543,
 0.5717928913774288,0.8245965674013488,0.9643296076822806,0.21703804332138144,
 0.6078234678937062,0.17438570738700976,0.27545201352909454,
 0.14996324475007405,0.5958042396834468,0.5364298810660907,
 0.47729532633399907,0.6132688888012713,0.7719341126047256,
 0.10023858764303717,0.9309844695961073,0.6851595052387074,
 0.05416387790186816,0.5719482715638234,0.7371856101423014,0.5764002228283543,
 0.41124500590013147,0.5232769354126884,0.6827451584324467,
 0.21918057020590576,0.1410841218246559,0.17736774499894936,0.521107085143536,
 0.21615472362708554,0.6043776264958017,0.899024888928356,0.37094397259461676,
 0.7992559039939077,0.9110816743036914,0.6224090952334016,0.8488352093220352,
 0.6723762472750536,0.11195268727314811,0.1456293931845325,0.9864014536688948,
 0.6950332309513212,0.4329123825013934,0.7247577099226337,0.09797175634495381,
 0.48389684738875804,0.861812252795809,0.07370266141769055,0.4963708993346668,
 0.2289317136182617,0.02282494336544838,0.6499309875630597,
 0.15979887301190465,0.7236469579000181,0.36524029033760164,
 0.6804702724411169,0.8519155320366477,0.6280249371621435,0.943492593795561,
 0.1977785969193938,0.42224469115042573,0.18772018801581403,
 0.4257866357795682,0.5468203884045898,0.9669054252575288,0.7079485460730619,
 0.42231844442479183,0.411659496578928,0.3626665565717967,0.5748528279238752,
 0.8146621309206835,0.18907159716669664,0.304613643175788,0.9756212978174635,
 0.41161956415353007,0.7662907958735168,0.2532599021235026,0.4248130396057769,
 0.11989670655944107,0.836266170120487,0.8671539259477719,0.7530723882948915,
 0.0765436605552483,0.25909844056133524,0.6652969170404902,0.2811123216925606,
 0.2545936447808619,0.021226752707533825,0.25203002607891345,
 0.5551583456607289,0.9951193324397122,0.49767344604661135,0.668355653461741,
 0.6270050555432347,0.43992317331188546,0.16681411888004616,
 0.3932481519006825,0.7435727756889826,0.5524244617826166,0.14014113074657775,
 0.17849921525208623,0.5321251377153244,0.42249498048885925,
 0.22513712777987183,0.7212321369373463,0.710516899873023,0.8784195737552939,
 0.2316594231839476,0.4968386081685724,0.5840388456611243,0.11741835527487154,
 0.9509029928647513,0.02499944741703508,0.10081161623146995,
 0.5700542579588288,0.03878547422159784,0.2095465638504389,
 0.19503934372442455,0.38934055232485876,0.0066318319351654775,
 0.22166345476359195,0.43239544057311974,0.3242885416088408,
 0.3246211738591679,0.7778148537201768,0.5214351353626706,0.338287786810721,
 0.3414930154863012,0.10087216670687049,0.1984567300280442,0.1355164412894886,
 0.8425374279684584,0.2812592193544503,0.2843282243791345,0.1933528619299436,
 0.44516451382509603,0.9710990659560987,0.21436470149736164,
 0.7164248433120521,0.7715418470957862,0.6022148548842973,0.5685720352843058,
 0.902017250164208,0.2953645196505954,0.9282819100504731,0.33724175736172224,
 0.6569542951538634,0.25238245989641384,0.4890726311867003,
 0.41935834162042296,0.4049486004892926,0.12748200822728095,
 0.38158774335647494,0.3638281053084599,0.9508438558627564,0.9926936575783657,
 0.47319798912898425,0.4346012415278553,0.7150737753025502,0.7845966686379799,
 0.4547570105426416,0.6498231837972687,0.9105585051926423,0.4116602506639693,
 0.14101250441582336,0.2592331200974335,0.25084949675688184,
 0.15387582100809793,0.6322070412406829,0.6230752074648038,0.930162952950276,
 0.5709686363659835,0.9468797157422159,0.09267240233876395,0.6310183188193684,
 0.4822569779959248,0.26211792773645326,0.3724897867105341,0.7820584934222529,
 0.9971735456232802,0.21520900188859804,0.9866772843316599,
 0.06953156055820409,0.45367384859193916,0.6191066617922514,0.873380402130101,
 0.31853009143383826,0.8805499656321675,0.41514501257567205,0.844089899370988,
 0.4668433219985547,0.8042643832889911,0.24959010232480716,0.9083525262363816,
 0.8952381249525063,0.024307247664409837,0.5950132797968573,
 0.3704907322294987,0.7048190350483724,0.6194902053204725,0.5405474310982161,
 0.2966087677474962,0.3663212761602177,0.5741508214925538,0.950328211103768,
 0.999139969560333,0.5654133510432776,0.5298428700466616,0.5953357483995992,
 0.45742119642072054,0.9964363256646086,0.015143542911279062,
 0.5499092722767518,0.2842146394721229,0.2960463039785256,0.9360181863413862,
 0.18539103971651427,0.1287680700966859,0.1370322728803951,0.5983798456380862,
 0.0033032954763138056,0.940056847949351,0.8922023943441861,
 0.35571708498598786,0.8419431135496066,0.12510427916474853,
 0.0334837289885106,0.6157278633195948,0.6147933830301964,
 0.007999683022409498,0.6551863252983455,0.4246604095661207,
 0.4474464220947125,0.8121578864819029,0.6978558451333717,0.3499303137699037,
 0.12373555454803276,0.3333597113401193,0.9461290328744733,
 0.059913517278825434,0.021670788103362093,0.285472251322636,
 0.055401129077470634,0.2122570909525665,0.9029177702698461,
 0.11653030053439217,0.771127645747372,0.4954453057773186,0.9225979010062192,
 0.12168941007667622,0.3779756731321683,0.7687187697841821,0.6113178912853957,
 0.33586842498322844,0.09764088536929927,0.3908113816820281,
 0.12859970817535737,0.3541314340179822,0.20511764005789168,0.852584366556966,
 0.6972025346310708,0.20081891426374288,0.7269740930556129,0.8277834545034498,
 0.04489031836100277,0.4088121783769436,0.6285359120491306,
 0.49495745303691563,0.9263799921351769,0.7812374235513285,0.5337640603042744,
 0.13695294774749456,0.0018925815485362207,0.6997035484413803,
 0.79587163898005,0.48288969602891685,0.692847208295009,0.7027303309466344,
 0.08707440735479288,0.12065331215287389,0.5996355361511366,
 0.1754742851008858,0.6330564639243969,0.6946524975098767,
 0.014439534611933702,0.29630794290369566,0.7010207577663258,
 0.35277047539388606,0.3653273725692989,0.6661262615337586,0.9840810424300146,
 0.8092388534332247,0.13222042753863872,0.6332881901078894,
 0.46847400719940047,0.3994696372065758,0.18148087378374578,
 0.8080687210580557,0.8710813235693656,0.8486262017160118,0.6491574333150145,
 0.07355147480038537,0.34676245443125187,0.6898301983076278,
 0.43618134565941613,0.37807090806781696,0.6178332863784501,
 0.2530912689359701,0.6494723199495089,0.9386866630796101,0.6996766822402781,
 0.4818994326104006,0.834830794156471,0.4876882302078154,0.9825648199255753,
 0.26223333792353776,0.6684376275670335,0.3521453779468914,
 0.26917448772071406,0.24780227432837254,0.28887171819863344,
 0.46245103233622087,0.20302583296695143,0.30444571034484447,
 0.8672250712038763,0.529498375390731,0.21478776597242566,0.53534484704788,
 0.31274300186190707,0.5622081325071473,0.28533088816768437,
 0.4723455889474071,0.8831026541486068,0.365366056695569,0.01427648152439942,
 0.4153509512265464,0.1816993163375662,0.06173313566764249,0.7214627723185909,
 0.7774829781137319,0.14337897649372744,0.6702180588397894,0.7581217803913521,
 0.8478384347361104,0.3509099249727251,0.27347723622553133,
 0.34640715802880906,0.7286158033189082,0.9092626510417825,
 0.08423888591945661,0.27667419670279647,0.2606487226487658,
 0.1809912933443314,0.0346369772053956,0.8643436308659277,0.9345480714956313,
 0.5102751503170838,0.014474054561471661,0.6755338719040701,
 0.03980384413653715,0.0996967288020778,0.9967698607202512,
 0.24916444462857634,0.48176944010292566,0.9055150692362449,
 0.17207825565184398,0.16400711454002714,0.12938782312140285,
 0.4981214797799024,0.9206162791129323,0.7908583598157835,
 0.026259114349869606,0.6278875495791642,0.06866047820489574,
 0.4073257174061713,0.6200093614922249,0.1772938938958497,0.6034575720904918,
 0.6550257525251312,0.011404126793970182,0.1936635227608523,
 0.9352252242981454,0.6439710257079438,0.47581383197328575,0.4123078432781989,
 0.44428585198051207,0.5871846789217743,0.2773026072438646,0.6574485438753251,
 0.9419730359490931,0.7592927252550755,0.5383504975717414,0.345508715638298,
 0.6213370544840511,0.3083434143828714,0.7462972949176949,0.745291636753862,
 0.43564195083477275,0.4950419159432502,0.8947688667051175,
 0.28887469318466663,0.7337556688472235,0.916225860340229,0.7346434230829996,
 0.521744331158422,0.28101635042305395,0.5098628285557825,0.834192079917938,
 0.6933607399258839,0.28239111226728963,0.1578989407639687,0.8573057686430243,
 0.0932493266476121,0.3104326132038644,0.6651840889833766,0.15216865023536408,
 0.38828167050882123,0.24012971872756095,0.5548579824222677,
 0.0531783016392926,0.5909298081253743,0.5062240195953676,0.06233373849791357,
 0.1176870559985892,0.29762272967439596,0.5256158174076846,0.6357934200391423,
 0.26806905255042357,0.30016191386060664,0.477819786188447,
 0.07719007099258968,0.6424096186422674,0.06827864369952252,
 0.27720152264898634,0.4992273622596799,0.3731235603616938,
 0.23972628139340313,0.43462004292408185,0.5185870790339331,
 0.7437261214855684,0.3882196497880297,0.9657194611602082,0.2681659582881917,
 0.8694488472577413,0.24332417144659435,0.03475020240731197,
 0.5432264205551337,0.4763627979668079,0.368157625166162,0.34955901252559696,
 0.5254188369026785,0.292402645437274,0.10931815553182733,0.1337714190293341,
 0.8309098036328276,0.1995452322225465,0.6236423147304004,0.5687906695287939,
 0.4709551422932027,0.7290972336969654,0.5443246013506936,0.4609275336740981,
 0.06248906851911695,0.135619167750061,0.926630074380489,0.21122770436766702,
 0.47398230126995444,0.9589940617522991,0.19679934216642547,
 0.33602733714290767,0.7605554023697373,0.5166956511896974,0.7671051665452638,
 0.9393580948796432,0.14729091677745643,0.6720801746979734,0.9667139824896164,
 0.21063967208442058,0.6307155033018677,0.2626584158194266,0.9891117911799047,
 0.6801072972547959,0.6396800433797178,0.04548610956855714,
 0.05109050314892327,0.5822426304042071,0.42874228237546086,
 0.6664315506442104,0.006263107365191978,0.15287987715015705,
 0.9075826056808717,0.20569430421459445,0.3110618736531544,0.8864549791977926,
 0.9738542081256837,0.2825252061097113,0.7031633264650659,0.9969310905835534,
 0.08194478536138272,0.7101938598938916,0.6734724568084635,0.8586544245245885,
 0.10563286648705805,0.36147455875380685,0.5241539987695221,
 0.5950502696685627,0.15923589607827904,0.9973012210236657,
 0.21231264093184998,0.14843795902013512,0.5365101215037814,
 0.09116818677384864,0.5681085180049906,0.008151490002889061,
 0.8348467326838812,0.7609159989463891,0.05520044106418942,0.970971624414105,
 0.14141715123359921,0.004488632309897378,0.018557290495090473,
 0.5439090291844455,0.879527580590854,0.02633322597740151,0.9304079586294972,
 0.5950889318334329,0.07346848039926202,0.7186255081300817,0.7300711233140155,
 0.11930414594997973,0.692378943582526,0.5018561958532406,0.7278468039180984,
 0.868444639015465,0.5671827068901805,0.06964299735494206,0.8718197272428929,
 0.7882426674198794,0.7927477416461879,0.24952591432765647,0.4593062434391989,
 0.6737706135952526,0.9681154003527329,0.37215612897821315,
 0.20854194759220623,0.09831633607682366,0.3761547020368017,
 0.6085630886847104,0.1993025878187028,0.5380477312105435,0.7540895324550217,
 0.10781658889760815,0.5409754528010265,0.34614734113958345,
 0.41664312366711775,0.7518057018141692,0.7729598810624055,0.278985547239502,
 0.2737532187370769,0.3994735303723076,0.9906212004842363,0.34185824153898914,
 0.9428561812292146,0.8815688402363094,0.12826276103396517,
 0.12331378649614033,0.21559513144771736,0.05396681581967788,
 0.004040829253001421,0.47288219061405967,1.648881983248307e-4,
 0.6490672161929346,0.4696796268034561,0.21747112247738876,0.7625025214855667,
 0.4669680127316981,0.952361167419903,0.05176127683092224,0.7731374916573376,
 0.21247072796205457,0.9599024380677362,0.6292311181702388,
 0.07802264145712579,0.5329024603582759,0.40611488472504764,
 0.15837676799561395,0.9504540773210104,0.2657287233460487,
 0.26516280510009105,0.30939500605958703,0.05375398742197668,
 0.8725901827721805,0.9820325671428127,0.4511588732860694,0.4648092495942775,
 0.38858546253385495,0.17681856708238686,0.4828577015418396,
 0.6091336927542628,0.6408248783544543,0.2786468450566545,0.4885739907937613,
 0.5740865281625447,0.6090038364347221,0.44571799320208694,0.5643820554212053,
 0.05147480444988217,0.4040646791558429,0.11190366733577761,
 0.03213997756442555,0.17195540789534647,0.9945509292133657,
 0.4015402437085762,0.4937562926570619,0.3357988304566295,0.5004866433770973,
 0.38779842563347433,0.6576629313473237,0.9744462267223345,0.9184548911954601,
 0.06829565822813388,0.599372411408355,0.8381993065711155,0.11735332780736807,
 0.789235175855024,0.09161332438347514,0.5738478917734351,0.23951098670983018,
 0.4068808565167098,0.3691577511564933,0.6247292434534714,0.18725485048065382,
 0.33446828448576993,0.9699825065871382,0.6592830221382392,0.2110126701168351,
 0.18542592880692232,0.34458702636622984,0.2246900226447146,
 0.4700969582720499,0.7230217295609114,0.3246249068592062,0.8702143432588428,
 0.1546295447835525,0.30710409837951846,0.42209340882828794,
 0.49383856311850516,0.29660066468965307,0.7840060396997267,
 0.6750412756730144,0.45195768220056065,0.8303028926289674,0.2407052249951659,
 0.6886345737352847,0.8707914994896748,0.8309885391629661,0.1707285660629656,
 0.273016707533588,0.44642011549217375,0.7781273342151508,0.3881279198423462,
 0.7278154399916028,0.8615836191070537,0.6360787139032013,0.8702103596066706,
 0.9407687739389304,0.543208545651784,0.29440226026146366,0.8236402070546616,
 0.7645709445463478,0.611091179985382,0.44773530354791724,0.6490604879281202,
 0.3300143916245817,0.38872464619869473,0.6947316051993726,0.8928621992897013,
 0.5933589806011463,0.05532939516894031,0.7737305363653297,0.2794380074589803,
 0.24623446631254664,0.8430059875293754,0.8904904226186569,0.4456110855114349,
 0.9156235006516031,0.5101425098513669,0.6539269667697425,0.9587468961283475,
 0.3804814071742728,0.9865274070769503,0.9935374512561896,0.7583929396963921,
 0.35247039659314616,0.6876736900038082,0.04572830875134304,
 0.34219915904953135,0.1350746463727509,0.5656948074521159,0.9626986005550834,
 0.8783509759635887,0.03822718566983996,0.13292180861748326,
 0.012404132672808266,0.07302034562411897,0.12499124806640616,
 0.49197610582782014,0.5437202434665889,0.3119094787643458,0.8199757999293434,
 0.7978993682211839,0.7596445353791453,0.07253688738773767,0.6945525860521466,
 0.5150114710971636,0.005669627325432081,0.9094298396918692,
 0.9312523282427637,0.6087139688377063,0.7176130398127795,0.19889141431114332,
 0.8886395398452819,0.9594807245387457,0.7825484112208106,0.5130961606291802,
 0.21081397552149395,0.4384376417286009,0.8491790118001957,0.8775656291797839,
 0.6651289187210829,0.8925767923564575,0.635914520697008,0.11806347208224133,
 0.7186537101577315,0.15037188990005546,0.9813745580185458,0.8644690087875374,
 0.4116188814464967,0.3267207202738436,0.7915515290013003,0.11348337166949563,
 0.5815343369384534,0.9887676208487722,0.42786263578671446,0.5125531701593914,
 0.8254095916503568,0.3289036075498786,0.31921440100198306,0.3126057229338034,
 0.4381429575964334,0.5869956684645018,0.9416411434966707,0.5724852196088629,
 0.30548391776823003,0.20645726087171112,0.12731427929367345,
 0.8488621695635323,0.04598094550616039,0.7794765806656043,
 0.39828911220540575,0.9646562052523058,0.0332146642538913,
 0.12106070100477928,0.3650483362583934,0.7849850768472005,
 0.24218257392060316,0.847144520489243,0.7440227497964851,0.9645497611920751,
 0.03356553288175579,0.943590993846678,0.46520573204057036,
 0.05053706472037578,0.8510312010616687,0.5060566543867323,
 0.17010535279410444,0.2483618876862772,0.818600857239389,0.4726559400609064,
 0.11423342017753157,0.331453363625721,0.7126883109400752,0.4520141463810703,
 0.3219650972742394,0.621481779774801,0.8935624212547042,0.4540995461761098,
 0.8977946639767723,0.04263826044669494,0.45348231299488106,
 0.7752266836712609,0.3546582942059622,0.9650788434116925,0.9374402082556786,
 0.5913645808826482,0.7366044941308635,0.19061703506042882,
 0.36123149399818777,0.6493370375700407,0.10839593707150885,
 0.27469685452816683,0.3241942091382761,0.19907683336593274,
 0.27369267253368335,0.4769977944399346,0.3334166877687922,
 0.47984106185242725,0.705765330900248,0.1415687101607611,0.8596589863363694,
 0.10887329356447384,0.2807581059146904,0.8817138826592006,0.6503712721877772,
 0.7689908882374521,0.4435055630426121,0.6178917281278973,0.03157400073997807,
 0.47556976500706494,0.6326629019814928,0.36541237922814984,
 0.4167158921799077,0.3682164807819782,0.6097566216036918,0.6223770886399389,
 0.30830579224318855,0.23801031641202053,0.06477039135472884,
 0.02378906067648623,0.16959375117643527,0.01366230276007685,
 0.5009471091933985,0.4692160782168324,0.1228260471030056,0.4608025964500614,
 0.9742977298365199,0.17996736375751132,0.974653398946076,0.16211929181288454,
 0.39501218242210856,0.4190268523921322,0.4238095513879184,0.5436208168893181,
 0.009998515386897222,0.37536437378756005,0.08582697780746362,
 0.09874051202330447,0.40300796831409236,0.02703574037887213,
 0.5239527278901934,0.4366801084698474,0.8310470319026423,0.6800277373888464,
 0.0593568668482235,0.6054299280229656,0.6146173402377938,0.8070717173104188,
 0.5616117316062881,0.16354955149923778,0.7514309268252111,0.5598551321203366,
 0.4195080977424719,0.2240226251939449,0.370545598875919,0.6943585045611037,
 0.2203686495857642,0.34852586783684947,0.034065278438852986,
 0.8988287195282738,0.2394457968030701,0.9694195080475361,0.49463066005746914,
 0.781162131689731,0.4584819073840095,0.2434068491732958,0.2907151915318529,
 0.34991359210356265,0.5869934704916762,0.7497945525998146,0.853753205350272,
 0.4476307882687157,0.21166621528885965,0.5237728013909352,0.8393419724911105,
 0.6090218994189144,0.9223025743533229,0.11668609231472638,0.6875805691704703,
 0.0014024607881568674,0.08763989108936865,0.3438676445059501,
 0.5158641984199785,0.2556407665394065,0.2505653905052554,0.08790769494200612,
 0.31984829940642245,0.25578569412416585,0.797207579561966,0.5547456712033858,
 0.27396717845297214,0.42821490592776823,0.5306217907936378,
 0.3174879409944591,0.9774364702655726,0.4131023370927249,
 0.014908055286056765,0.31730692424650864,0.5762885211069723,
 0.5617759568715293,0.7018338393603124,0.5819365026550903,0.7384870359098772,
 0.1434804880575805,0.2537561354663187,0.6317915448232516,0.592658776716396,
 0.5300996639633421,0.8792363814437911,0.3483563233118545,0.6996748429991249,
 0.10696251621189423,0.6413666778025275,0.33524890086963166,
 0.7038243824456203,0.2967207225398262,0.06505811603917466,0.6547337614974288,
 0.06303855639152,0.12581193536814783,0.8941553437436955,0.5337335460236883,
 0.6527191886868807,0.6698837023511395,0.7752879682401597,0.27426642333547946,
 0.6154263409008349,0.31085033219304237,0.3428184004366016,0.9010620888675479,
 0.09639895159581013,0.8635279552606003,0.7749392437027709,
 0.10941197455418594,0.30342038034309127,0.8274081839699838,
 0.6384936296378647,0.9556976241015599,0.020153634507003693,
 0.7441870866434307,0.6386933049078674,0.6357161743331513,0.8878192037095064,
 0.7806086423483248,0.2978884101185134,0.8502029151637372,0.15341046819584936,
 0.17497465683988778,0.6337809497160105,0.4352873636304073,0.7531480867416153,
 0.7736462982972846,0.39980712696698695,0.9983023670317328,0.3345523775338398,
 0.7311482107785383,0.2961464587785181,0.9740917528061186,0.32003684178700786,
 0.5445577658778386,0.06721950670517263,0.9734307575517743,
 0.23849992191103964,0.7368845169783406,0.7396735882354476,0.9305385283475527,
 0.6993122355962397,0.27579381682666404,0.2459035086582657,0.551731188071274,
 0.7857726389076277,0.5619511589949173,0.7055693734007095,0.27740881990341504,
 0.4715681021902869,0.31276265761806676,0.5319477628238644,0.3060521372245065,
 0.47936055059053895,0.6175140695952784,0.37574401186092155,
 0.2797919760865538,0.5501837293273012,0.8667577949976029,
 0.009334260496678848,0.5741060033262545,0.05899416483701769,
 0.5854780337140826,0.9718433151789728,0.4285242644395304,0.7715381055158392,
 0.33635485689106615,0.9636625972275998,0.7994074521503745,0.8920380643211069,
 0.4976839121804463,0.6160628373867741,0.8278962840647615,0.4819169717720547,
 0.271299431533341,0.29675032103637355,0.7476937178620996,0.17804901790961114,
 0.8279300781956518,0.45781735383337796,0.28304876167675197,
 0.6326836916191921,0.6230644731678606,0.49997302235031826,0.523102119984753,
 0.7765468490649664,0.3584977679609511,0.9561287506663072,0.511859791986915,
 0.3657288333325388,0.38393691133861385,0.7372188199762866,0.9514645489481045,
 0.4774075403007071,0.14062288931224243,0.1464740237123383,0.8331866101350417,
 0.46069156356093754,0.3660152212892498,0.38578774666752336,
 0.3383736030312451,0.6445356402248188,0.33713675885425276,
 0.13906600106006517,0.8529955921543368,0.8975879469916875,0.4300620031947223,
 0.5393420890694023,0.6894018292440167,0.07112963723999655,
 0.07355633514840443,0.11112904555104253,0.4779089317491203,0.521214533110904,
 0.753074455125641,0.7000028813126783,0.7516611289822883,0.4285322377902475,
 0.7374037827042779,0.8784198783807905,0.25012606482293065,
 0.24238256549154957,0.8077137496628162,0.3984872877398761,
 0.35551887640656366,0.8663725120195184,0.983215489184699,0.6414084148835684,
 0.1581997606937664,0.7964807393911516,0.06739669470894338,0.7391617820823475,
 0.9744033290433636,0.9085171470174915,0.420738559082677,0.5536921586247259,
 0.608987124773793,0.46758675020128826,0.26266596675767273,
 0.15563575925843043,0.31718769629854915,0.07801689392240474,
 0.8666762847175381,0.5389870601802679,0.300776921498368,0.031092074612512066,
 0.0581036332614987,0.9319512289557499,0.518735794550036,0.8225682508566579,
 0.40057223020185817,0.030031182348194574,0.5280080706377486,
 0.9674192552083279,0.14692394190535174,0.01730435508270356,
 0.05680414410738044,0.6298476494574617,0.2640381216289478,0.4078101513942768,
 0.7914451419324448,0.7805458352683979,0.2296726785218084,0.21524096463161146,
 0.09578723955672264,0.5894396768321921,0.8113267887556826,
 0.03535497757205597,0.264926302020675,0.4056114494896975,0.6533045284136931,
 0.7641905988557576,0.6112983243981502,0.625543826813472,0.9015340865714996,
 0.5135241829699291,0.7084226645480862,0.16852674819019298,0.3330337707692236,
 0.24264779959013816,0.6546147328477182,0.18928640248571504,
 0.0541670801272065,0.058539006902229795,0.17765627210337975,
 0.6902999598014086,0.13344902848060947,0.7032563270984479,0.5850447946678713,
 0.4946422859860783,0.3549662266252098,0.22991041578628735,
 0.19199979650984567,0.03590254660207781,0.5747798785319491,
 0.5554513167281663,0.3170467972777943,0.28677329950719177,0.5021194933796184,
 0.7737111623680456,0.9391103057106893,0.7984137978506597,0.14454323862845808,
 0.9508848776392075,0.2053713891382145,0.5585415874134799,0.5313798018313965,
 0.19742256211296239,0.34578129318142325,0.6937232774854402,
 0.1619673091214331,0.7230994881624981,0.661194254267687,0.3824803031762509,
 0.6911754402809189,0.9668911512826555,0.03978948076175004,
 0.19072905690959607,0.7514772372718226,0.2420541809784783,0.9925303944179097,
 0.8399138046744151,0.7941204508162009,0.5448297995272271,0.46351446831997767,
 0.8161891477759479,0.4742433145065392,0.3898256073724641,0.6537083930293363,
 0.461638851843851,0.2905712015112414,0.014163723002690087,0.6061162448104931,
 0.520629972438953,0.41979202811219407,0.16437221138076352,0.7207522106079695,
 0.9340973552199588,0.6042616086805127,0.3970383548188986,0.7838637435503557,
 0.6776241671779322,0.510149300061296,0.10837154363318835,
 0.037077848230188226,0.9118800447450636,0.5901361133058851,
 0.5853503138647551,0.9534075596929233,0.8791268480785537,0.20675607259185003,
 0.8715398114055709,0.6133933198616267,0.2751750334024532,0.4118098284336529,
 0.721937936057443,0.20506056361846903,0.7275515035809413,0.7576733828137445,
 0.8703884342058991,0.7720253474131847,0.5213835119539714,0.7962936954968668,
 0.6583439647367408,0.7125733208522901,0.004896885799848927,
 0.6999542128667309,0.11140950644223147,0.4254831372169142,0.7351689387428928,
 0.8146456381002379,0.23375483197364844,0.2093852083170583,0.4084008994072097,
 0.26456392353445246,0.08841800486818008,0.7972788259483966,
 0.042200027723539035,0.07018977838742868,0.6202393659234573,
 0.7855708727410424,0.6757822396129289,0.28676203067740047,0.3302326806976711,
 0.34485011726356585,0.26665781586349435,0.8264041726227508,
 0.5604940189290042,0.7071080379747945,0.6396658623867884,0.15457542589375262,
 0.7799755788539177,0.5399072625040798,0.22474096722763848,0.6549536355466516,
 0.8253885238780883,0.6731652103108013,0.0984944899283432,0.3880675356084282,
 0.059292252717076055,0.39347517516761366,0.07786028907582088,
 0.5161036891585361,0.569674992146182,0.37021000104497404,0.33923702362242947,
 0.4452140274504688,0.08998226793173991,0.5778016903304493,0.1359762391177579,
 0.31888764024442984,0.9180581928214453,0.7179769596081153,0.3801098278936642,
 0.16939122818645014,0.8604935142181231,0.3840884140058821,
 0.23740821050940908,0.6306358168102373,0.9773904802446401,
 0.35491506629317193,0.3674311948975174,0.15636013886709554,
 0.33258152200029467,0.7757117899923716,0.043756508378044456,
 0.32443224444668173,0.14147468485127745,0.2419919345111955,0.695818793799149,
 0.459088573601114,0.9139875073086864,0.7543626704816632,0.18903886959590632,
 0.2555646073276858,0.12867553310859192,0.37404371181985696,
 0.28456144051279264,0.56191343160155,0.3439515487920711,0.23296850150433168,
 0.6391352915396873,0.5371088677812784,0.46553607910792083,
 0.20124404441721666,0.0955524942829451,0.34270095719942684,
 0.6020864967646138,0.6527145313477346,0.6195248837312549,0.23554421895460198,
 0.6213747065514774,0.3096046481154554,0.7189737014196739,0.8113387479469161,
 0.5294340186222414,0.7142047571660184,0.2781034613723259,0.42807436961240986,
 0.010105662453133268,0.39869069857533246,0.538858209117397,
 0.6541799729192463,0.34124528800641596,0.922027783357416,0.2760489166314518,
 0.6868690493458968,0.03296418775197152,0.7513257138991861,0.5703155573442127,
 0.13716488606938726,0.005056906773371117,0.19871240560337677,
 0.9348428767060818,0.5440185744668103,0.8340967552250528,0.9071913822547373,
 0.9595887169849522,0.9243841598333712,0.820274802387324,0.7481485561311166,
 0.5900759888519783,0.5786640864432944,0.887648464027418,0.02137389480976537,
 0.5719879775685738,0.19291129838813714,0.6227772800662579,
 0.10459882647639396,0.8115533356424462,0.8073782484097378,0.281521334990946,
 0.2069429533979652,0.008116574453403191,0.5732552130983988,0.213555211511666,
 0.495808728545448,0.20759649323703544,0.2851102585916382,0.6442447218099738,
 0.8623754227367465,0.2192148404627141,0.32624859297707576,0.5171778420712597,
 0.8760831608734522,0.08113438381305071,0.568000686585743,0.7928784912246493,
 0.521841065935021,0.8555703693601483,0.42531270379148134,0.544512463143265,
 0.40351733668102296,0.839176733572545,0.5394940846509096,0.9591361835873811,
 0.0703890734766841,0.10265545176087476,0.47408692160666854,
 0.7353551728599733,0.739654432588053,0.0032270415032233934,
 0.6591734125266882,0.5532651629122871,0.37409735936065347,0.3203270240358389,
 0.9309165912828605,0.1760223579461212,0.1183724769274982,0.3547654040600887,
 0.3952609816801689,0.7576246661071898,0.27250850164589013,
 0.19584257808504835,0.10131664528241968,0.9519043198001507,
 0.11533961816728033,0.9135369183160111,0.10625808608332143,0.423969424476229,
 0.04298997007440042,0.7458896004255846,0.4333056145740368,0.3416517160164373,
 0.8210605350772805,0.19297246815655467,0.08730398687478891,
 0.4428226690222239,0.7298832447962338,0.12232886340169813,0.9453472329368835,
 0.26779522854048987,0.08005883121805091,0.03608359207302725,
 0.6493967412676492,0.4931721195369022,0.5037626996938857,0.1064561401668479,
 0.7949654353287269,0.2987784415294681,0.2152103715612642,0.1809466164194301,
 0.6572217282342266,0.7982980768717776,0.3187649802465803,0.5890985466759182,
 0.6027384863897325,0.3947205235003505,0.9445575682942177,0.14408599778138165,
 0.014628938844428396,0.0010185967865876666,0.9744268594883725,
 0.5965035909856596,0.9913584686364707,0.288446978337635,0.5732756787508789,
 0.14119351446282669,0.31163213365086884,0.7933277509114796,0.648949723236442,
 0.344175606258068,0.7417505169642253,0.40980304571053217,0.6708901418269271,
 0.58248940254238,0.18021872809420025,0.9741025245590015,0.676569016716613,
 0.15159719640413116,0.19340888765212216,0.29654178095416694,
 0.18188855083522426,0.36660289944025415,0.26348274913430403,
 0.5304106236858733,0.24560380769515666,0.6260762647456584,
 0.45505390049312877,0.07369459334897621,0.9532737508118185,
 0.0572833696216174,0.3914808315133098,0.9625232773242172,0.2048519519188171,
 0.22212757984176812,0.7014810705048011,0.6047379233547409,0.5547664795041325,
 0.5601597777689034,0.2137331097149323,0.31108471130730453,
 0.36374434860146154,0.7547307956216368,0.6965734228879911,0.6958291122571614,
 0.9380443369897782,0.0429340054836882,0.9636133436380084,0.6924657075933837,
 0.844248499454999,0.13196893607704885,0.771422853370545,0.17443106553812449,
 0.22432418474336635,0.5240720593597737,0.731961743045632,0.9673188633487031,
 0.46226769022060576,0.7630001520259835,0.0760269467786342,0.5017577655498694,
 0.17118449092838173,0.10032559671785601,0.20475820620955099,
 0.7340269382449162,0.8874387384805081,0.13822727068483887,0.9170919892027922,
 0.22015812663079626,0.9514766006963455,0.6845146577352245,
 0.12124404366821706,0.20494923080408978,0.2855578916419904,
 0.43651965430483963,0.7861838394401774,0.5863527745394687,
 0.22344611506038048,0.6692062135730216,0.6997735598181754,
 0.0016932883800997445,0.16313870818070475,0.7700597529229413,
 0.6198862049276915,0.010011809026245433,0.7025381636409582,
 0.16678016183308797,0.6789894209685876,0.676390434763891,0.12677603040303975,
 0.37496783363132324,0.3035877085733256,0.40522658469794204,
 0.6426302240824564,0.9131868218119014,0.026370991463821714,
 0.9895759079254378,0.20316533136155024,0.2833646979069734,0.6333315227873239,
 0.8240470091453137,0.8946785493709064,0.6197898735222813,0.6477958664551535,
 0.5416296938019073,0.09785140252190294,0.49842149519923706,
 0.08144511367708651,0.6365984574978685,0.3987092001672259,0.3450337341752123,
 0.7350127200632263,0.04315176771637552,0.1702035064382449,0.6907086699812344,
 0.015512677343617831,0.10071655092876841,0.5644477339646333,
 0.4896866574768661,0.4377029707003559,0.6298612621942767,0.9961931152522222,
 0.08581030400511203,0.5382247771210588,0.43827889788880015,
 0.5506343268334392,0.6108612904182833,0.09303536119447187,
 0.33080331514759875,0.1197474533921834,0.7428279886315368,0.5761978542077341,
 0.712445185127418,0.13758572690298898,0.8906884198445469,0.15404281726442903,
 0.5630790594160782,0.052910002619389496,0.042934465466692995,
 0.9359339750098825,0.8190045513384794,0.6013161390887484,0.898183973754396,
 0.17768829701365496,0.672406838118573,0.9385070541517349,0.3886261920154269,
 0.760727507348632,0.8920231313895403,0.1907282058526344,0.9209899254778908,
 0.3461326603360808,0.242604473173183,0.05596561322010929,0.5384124970211922,
 0.6679026441707603,0.10688800224072559,0.8554994043829061,0.6738424970803237,
 0.08996832453969372,0.32430104166420093,0.43196138918704596,
 0.47988905864354403,0.9766707857227995,0.04401710722872698,
 0.9612491738859533,0.997464257068525,0.6032321668416865,0.18668779094552768,
 0.8104144654318144,0.020088096926274335,0.6414893555436203,
 0.4329736944213619,0.19413274068545938,0.5189732866116924,0.888793360445235,
 0.6143968783563012,0.33986333827799364,0.10178590707306645,
 0.22241577579086347,0.9297036905100384,0.2686132958720747,0.6266974615415374,
 0.06979320048454696,0.05338583476628289,0.2256619835322332,
 0.4741311880221011,0.36215700490694414,0.7568695745765381,0.9411240894387597,
 0.4272874122100533,0.19314634958694799,0.5640456167372487,
 0.10344446348710401,0.7624795970486584,0.221175067414923,0.8839058053701938,
 0.8586337890257475,0.29442360478108687,0.7561900404943255,0.5507212624454687,
 0.9079605665736766,0.9745414688515095,0.03656033659342972,0.5923005000497443,
 0.23535837658468028,0.5269304948537407,0.384753427155764,0.7297388266784512,
 0.7927292375854602,0.12856756253533208,0.24666174146936926,
 0.6385790448605864,0.01824527440455126,0.656449897215796,0.5659023809527647,
 0.640278949627625,0.4079574729691271,0.7314509090749296,0.38914706107028385,
 0.23306323646600835,0.3734449529954821,0.4959782636137178,0.7228096819113478,
 0.3184449854670922,0.6656807680065665,0.6014695843057423,
 0.059902932254471075,0.49392933601325073,0.3787039681244113,
 0.8519363115798331,0.5222398050828053,0.9982982836184126,0.8174448024078238,
 0.6901834111548149,0.3896724867331042,0.8957632819943554,
 0.002126641997979406,0.8394699183911342,0.04162079283787745,
 0.7516611058286186,0.020305770861245742,0.5966175192477606,
 0.7919664690577422,0.9811128908325172,0.712143454693737,0.7013295595857446,
 0.7318522220818472,0.2565084110533664,0.2744912724595694,0.5072271658636137,
 0.6071788780142932,0.38830683590615245,0.8503270550947957,
 0.49774368222070453,0.20061390767162313,0.7272630550293945,
 0.8066765529597512,0.5237260558429211,0.7526699457566232,0.8129340929921944,
 0.3248666460039479,0.930619566277821,0.6259891686028298,0.03447396880455067,
 0.29251340446314755,0.6829515899556544,0.7981651563984029,
 0.21091415882239073,0.9180474157114453,0.6487568201808084,
 0.27072605628464186,0.8350270209417662,0.8623679177354516,0.6274540952871012,
 0.40693965181286573,0.4221547558767371,0.02396541340546654,
 0.8815846871873056,0.07954557335343948,0.4207534171153421,0.8105553926624195,
 0.11211007605972312,0.850589951246271,0.015522867872870605,
 0.4672764789874926,0.8718088071237825,0.041347816439368246,
 0.5278270645162643,0.9970661698159299,0.4182499315131559,0.3226557854304578,
 0.22670856199360234,0.004563233045309589,0.44693655338718197,
 0.33275516728023513,0.9796498800446346,0.15898621376065314,
 0.0452287971604286,0.6728250835140573,0.8719609921755628,0.24863455784902955,
 0.12929252030519223,0.9726882903271565,0.12336214302155346,
 0.9846705546521822,0.549027237304807,0.11470480652128656,0.6046787032720093,
 0.06507093900488103,0.8042386374356512,0.06866535749496205,
 0.1348020884654264,0.35327164578926407,0.3346836128048589,
 0.38583132527983965,0.3754548893096624,0.16584962535029346,
 0.8167925076509536,0.2421464927348974,0.63455640092846,0.37916377712814686,
 0.06254483993450566,0.6556014654079966,0.7549009032315992,
 0.46990182571066275,0.7710626568340508,0.4929513791664668,0.7432653188669897,
 0.02779962646594436,0.505874384532013,0.9868957739426543,0.35860885595578795,
 0.016495392407128495,0.7092869022267049,0.2612188634328122,
 0.7603244482482853,0.34854872501350276,0.9768915586376553,0.1645346946234585,
 0.6000543313604292,0.3772723608099493,0.769289965778633,0.03545615460895235,
 0.3491937328903053,0.7209909663948297,0.2360411504658373,0.6442628893835824,
 0.5440701173481769,0.04953839136855642,0.7773901344304005,
 0.07899315931325401,0.7345045082817024,0.7047001039087689,0.2277963004771808,
 0.7972962196957317,0.8638320634301092,0.5120880040805911,0.8607151111928012,
 0.6538997832429894,0.25137610243927344,0.2232648231301354,0.9512070190337969,
 0.7795896310419823,0.20761685991963685,0.2101892626408236,0.7383313964414822,
 0.8648459731774651,0.842108041484379,0.5562190720569539,0.7413840692040208,
 0.5515135359630076,0.721989809321637,0.035423901990603635,0.3527600120754101,
 0.8996383744915692,0.4915084827466245,0.07931759013297557,
 0.04525164549008476,0.14505234264128164,0.24676510658854944,
 0.10295725362465313,0.008900444480901326,0.6702818369764668,
 0.8871337995890561,0.470563574737491,0.5105591744209849,0.957887925080717,
 0.9435808752725902,0.6960313437526222,0.16854051656161506,
 0.32511471127379543,0.05617034888059291,0.46970251676188535,
 0.5507692854801255,0.667389553799165,0.1732695051495835,0.7111913545969599,
 0.5699445971172344,0.36392169783010286,0.6566496523607479,
 0.044209185882314594,0.649366832585715,0.34400629899961344,
 0.4853421568178885,0.9861927937927542,0.8295853071163364,0.8265645085358547,
 0.11711815147025395,0.5249588085410856,0.811410250390883,0.8457851458683522,
 0.5705574087432065,0.6615155341380021,0.21697939074669925,0.6973469119098472,
 0.5759977504642292,0.5741493892072301,0.825453019271368,0.19913300276873902,
 0.7436208050662563,0.10177019180752056,0.9156491142059159,0.3177417913954639,
 0.9890673402245455,0.11077772216605686,0.2662405371655926,0.4030343262721464,
 0.8054038823861793,0.4857329889198265,0.9909923225639675,0.9988509682170797,
 0.49846008177554235,0.05089358833654767,0.15188698507378928,
 0.14937404936762944,0.06549900137003806,0.15945649071690848,
 0.308139195767241,0.7437193973666874,0.7585606531471814,0.06795612577770627,
 0.8956902940459174,0.9953378293684602,0.21639586360529406,
 0.056364317564858046,0.22166146603469417,0.330612930899838,0.64996806039266,
 0.7325557690119259,0.8928264955604842,0.8330581883909598,0.8404267263346076,
 0.6612656000857275,0.6337045769541187,0.3595822275419832,0.4728763128460246,
 0.34263802763607254,0.5390489812997494,0.6339220526761804,0.6341898847288779,
 0.2606450336134284,0.3116929327003958,0.203947071265689,0.811682224537001,
 0.8299209472216762,0.22716652539132887,0.490091332736191,0.14280809086195156,
 0.8025886702231888,0.9293280100743642,0.41739585871395657,0.6885955002036092,
 0.0882225079084854,0.4192348175192686,0.9801474188938131,0.8796382942167965,
 0.3463565047651782,0.6234424957750737,0.015031665174941589,
 0.3695194799131183,0.9318399255195166,0.7725403083857222,0.1574961374342032,
 0.44695355355789324,0.7859141120572047,0.0420335191830401,
 0.35645703099352977,0.5360644335752056,0.01810525285392406,
 0.33936281495182286,0.9418334075735953,0.5678938783490464,
 0.007060145127452566,0.5468146533921707,0.7150369524022631,
 0.7366889690509316,0.033845043270548714,0.40726264135870693,
 0.6454104347898969,0.24733565877749497,0.47531023104769154,
 0.6563242678537251,0.2931593930161045,0.5474246790213064,0.48873274620235985,
 0.7798761404366847,0.6447024424299554,0.5065769357716647,0.5412452968079188,
 0.10088946009690869,0.48370872145380805,0.3049079592544316,
 0.5531444667786154,0.6781977631922003,0.36024063898176995,0.8107023315791029,
 0.5210090123624552,0.8754949360528266,0.769118255423547,0.1830436848808472,
 0.2068504165569538,0.4786641181102488,0.651643424119826,0.5352020354830717,
 0.5024171971699979,0.2626944839726457,0.5041318625532966,0.4232088528092881,
 0.5661907544351772,0.5429617030030846,0.5136796145936935,0.48092634826261316,
 0.3261203236888228,0.403879859388642,0.4557741429732064,0.5252321095082899,
 0.9743438641882354,0.17891796254333336,0.4505724833337277,
 0.20254207128835322,0.10370701990898445,0.46243164409965765,
 0.469899078268337,0.16219539164265084,0.9674565400215536,0.2055874480486809,
 0.6831443350214287,0.3485028595391284,0.5685288189290807,0.7778548866306535,
 0.8019699020291033,0.2333534883470093,0.6298317705334125,0.33943053091511566,
 0.851329152834645,0.22257550686779937,0.615078505374432,0.14191615259081225,
 0.9335582107704419,0.9920333900421571,0.14897764834596572,0.7294391225375134,
 0.9871956537906365,0.04058175821214627,0.4743301841093296,
 0.17020944162642115,0.9102583083566564,0.4182410961574572,0.6063131554465356,
 0.5791435826279672,0.13640613825912185,0.7402088736327114,0.4323460951097887,
 0.22955265886535992,0.9556049817690012,0.8044387374604811,
 0.47407125776570624,0.8126177876017804,0.21401291856057125,
 0.8013742137249902,0.13663852246364783,0.9679058180821126,0.8490227709021456,
 0.12027928313614189,0.9373796789762914,0.9162565643012474,
 0.05573022987478937,0.4917617699129322,0.013831471419082142,
 0.06373653597267637,0.7273411243053721,0.3254314911184434,0.7094155974287493,
 0.8833428307265561,0.9974261762936814,0.5111789385102101,0.18030926893280808,
 0.4955576861882256,0.6801407804710187,0.021861864990326274,
 0.5454652831741422,0.3734494942858485,0.18008925960791777,
 0.24265129432824295,0.5897325532855149,0.3307353602077323,0.8151159029718597,
 0.4994514141332874,0.8632665308163504,0.5186756221510018,0.06134461488226706,
 0.8657654110648448,0.15750468899973513,0.1914245097024455,
 0.18272989729742484,0.5691164024561979,0.6015816791285913,0.8650820029031862,
 0.3739743657920487,0.8456498461158027,0.5206704114381524,0.9510706142509778,
 0.6784568421906587,0.7257031441723969,0.12907940559316067,0.3130480218527407,
 0.8966623953317047,0.8463395563142156,0.5257669002430844,0.5974914321794896,
 0.3322843304095767,0.5127841231154711,0.8599741194317073,0.6306585719296238,
 0.09701438790750239,0.7831266444724216,0.9030054154449385,0.6836500264396262,
 0.902444362133074,0.6739765062016342,0.33506909373154947,0.8282478618952338,
 0.7242140693988542,0.4063691592124681,0.9905081812937786,0.6078397308792232,
 0.46116108876557094,0.4826268508710968,0.19741939824221533,0.537698542800358,
 0.2584591329900733,0.4405393742119136,0.8958677182006339,0.8463160245259521,
 0.5689960219675114,0.6893484848576718,0.9219999960129591,0.48423528545232297,
 0.9974725272113869,0.8923193420529977,0.8207023577962054,0.5301170860677767,
 0.6835631872877201,0.08227398753364357,0.38392097054811825,
 0.4742884444744647,0.6714792135236365,0.7600806410158654,0.5453336425123753,
 0.958868307614251,0.14897950892859446,0.7426734965037668,0.7822825286112558,
 0.25913663616711424,0.08495050051663278,0.7696402997286305,
 0.7278394772975293,0.9640207799343516,0.08836062160301095,0.9220322123591017,
 0.4081254749863723,0.17158531935670285,0.9158146608362476,
 0.25009892499570396,0.23645189486466822,0.009683364617168344,
 0.8825235404670115,0.7837847714977556,0.3665671723540457,0.17547323774125878,
 0.1678791456563138,0.2560465576447738,0.6876554678816857,0.9813277160681985,
 0.758730012412373,0.8279542812914722,0.9938841722998778,0.42086472435381816,
 0.8228848634088324,0.5637048308652817,0.274107210057416,0.7887181082218968,
 0.4564372506992218,0.8103888180140042,0.24232549321352492,0.9650689854835908,
 0.16062897550551802,0.4431911062511098,0.3922169398560318,0.8977139806084603,
 0.5107338899010746,0.7612121995022472,0.7877457925140392,0.8324758663320733,
 0.689342838313556,0.49548029639029645,0.7094378377942354,0.30968242193868123,
 0.4131598291944837,0.8846529631496189,0.8314079416240413,0.6054375345304506,
 0.19496585602774752,0.21766403348071872,0.9395391976239347,
 0.7250732413069928,0.24452911933893484,0.4465751238039657,0.5981407033691407,
 0.19183512881583553,0.18885881682936811,0.9274861588902277,0.267092844226703,
 0.3325278977590974,0.7949113943788968,0.5773079894161031,0.09881249447356133,
 0.1792068561029726,0.755930622943038,0.1751169527184534,0.06828575554889049,
 0.7931336212573042,0.2565586562498884,0.5996400060948314,0.9570915217440927,
 0.930081285466218,0.591758913574591,0.04420827248324288,0.940353706359555,
 0.2581430335541496,0.7138350035869493,0.004207212495895174,
 0.2625458539886648,0.5479516669909241,0.07198020188595744,
 0.30561912484465725,0.9138495812282692,0.5199369286330503,0.5765754842648011,
 0.9564139305960841,0.72645712531736,0.6471249904032834,0.9411062554672694,
 0.2522719771641899,0.7889042788003763,0.7587119836013277,0.5402550099654848,
 0.8625014448916191,0.6037173874559945,0.6004947179620352,0.2044211702306744,
 0.732570566098675,0.9884499376415284,0.9029842905902373,0.2433343768019237,
 0.17495696366848346,0.04563819009135195,0.49671580344238797,
 0.33012762816004504,0.19716930864965876,0.04416040213322203,
 0.24445909162705404,0.4655960465484136,0.7597384254584307,0.531430580515173,
 0.50419541088462,0.48728544349006453,0.7144917925260035,0.8890940038835654,
 0.625906106117528,0.4139597992245574,0.616578762448266,0.5310833283586402,
 0.6323319379083223,0.08940729219713672,0.7062473628098406,0.7857316878580568,
 0.08873522352562646,0.7040473498000073,0.6740376771211603,
 0.033132492008894565,0.15665202667665457,0.8360213790829423,
 0.26299182998442283,0.48031358595979357,0.22275011717568316,
 0.9255351175732232,0.2537717276873297,0.2289038142222415,0.16779884210826224,
 0.4028553066815219,0.6940097038171338,0.5740807853472348,0.30599156651582704,
 0.4887018663599887,0.22007420515917397,0.46187817978242485,
 0.11192510136013067,0.9355167986635677,0.15955984541025547,
 0.4010412261094263,0.4049012099269671,0.8484434955791114,0.5763167776901967,
 0.41749670460474586,0.3990532696313116,0.6257698488531785,0.8863443925819043,
 0.8696999544537095,0.19006349308615167,0.8697052995265955,0.3944559417395902,
 0.4940162241648839,0.5257937455820233,0.08006842049846696,0.6369391301361722,
 0.406759866664101,0.3241592320050357,0.3331955531340167,0.4545598301975645,
 0.6884051206489838,0.5056799998534307,0.23677216462911432,0.9373072905795915,
 0.18280356504299222,0.12026841270037536,0.27097965700675153,
 0.6060421259607159,0.6841000619031662,0.2042853483584074,0.6314630246114217,
 0.7286975883180046,0.6930045686530496,0.10615498208523855,0.9697226652444143,
 0.2253041665976122,0.6497410693720049,0.7288069229632765,0.5447061985234987,
 0.9743283480669377,0.9571348338998268,0.27386673706918385,
 0.16387302145828664,0.7571565205053996,0.09629197125184386,
 0.7829253041366824,0.04945897788442544,0.2378081531114895,
 0.07243188498161168,0.8154531897389743,0.6768196377509502,0.4354619164095359,
 0.6991091418052402,0.8170139757621975,0.9242701739361062,0.21385569546514283,
 0.4471870175946152,0.6756233999844552,0.09787240907085193,0.7630229675732635,
 0.5326319143346183,0.8355566076323206,0.5509374449716161,0.994966269857152,
 0.7224982329056613,0.8336493383124742,0.7604581791641762,0.37174883521602564,
 0.4274644305462645,0.9694706459239621,0.011978153041678707,
 0.7245312669536165,0.912571453051272,0.6381333338540377,0.2414904144293869,
 0.18571907238154606,0.6645969109606218,0.3787445821879436,0.6216441753371704,
 0.6615193936355122,0.6060749444606219,0.7351890242432447,0.19892220861147314,
 0.664661700156375,0.9308266282958478,0.6544990451710326,0.038339646181788756,
 0.573665675460812,0.8308795469254258,0.8349983959520715,0.6950518642161829,
 0.2479943727991063,0.8172201427069531,0.295792394319371,0.6512156307882216,
 0.8720140987950868,0.6360514362833357,0.15382961286204477,0.6796086464056772,
 0.37301819604024944,0.8886495250548129,0.03478962999445556,
 0.3428138991837565,0.1399265539778879,0.8448859414892772,0.6536005649800046,
 0.8447265199815585,0.34040244560385824,0.9819261700805901,0.7429880592926655,
 0.4337299016925068,0.44463027367222696,0.9176938655211706,0.9860869870812554,
 0.9490721488038174,0.1978872537966455,0.6060951765005175,0.9553428339421842,
 0.0764987638629765,0.058573296809473985,0.5771732671252685,
 0.21996197640781046,0.7120092987201836,0.9903760798120738,
 0.21019136464877697,0.4118874982285847,0.3052005523767961,0.5868818311150232,
 0.9004669702609964,0.3403548449975434,0.0198077840929467,0.3135850374370518,
 0.9222987950121152,0.9852731022161153,0.9707847298272553,0.5500729500679208,
 0.8032533076212609,0.27414027438358335,0.26560985110761337,
 0.11582811415180694,0.6253683151344902,0.5273758623670493,0.7682333059662512,
 0.531442679888172,0.8825430264159775,0.909741477094841,0.5037095666532329,
 0.28700963862816076,0.9851093561066109,0.9117047442042928,
 0.032087163577102285,0.5433198137591019,0.6822278401511149,
 0.7444907569446357,0.37083348724631526,0.6397089804631308,0.8511706649459676,
 0.14611963437566589,0.37700474659353733,0.7345224323224602,
 0.050341244169104016,0.08548414580710739,0.47615759093078913,
 0.24850765901469463,0.0015204350703322822,0.3678849588529838,
 0.6501012175539899,0.2776270436067495,0.41281770489745573,
 0.18923212885471086,0.6083609932787022,0.7047479670542258,0.730736756732106,
 0.9280601289297514,0.9025087615126408,0.3628701910731299,0.04424017484248566,
 0.14219096911480555,0.6054246534036791,0.27513483102962666,
 0.10493159123797802,0.8307973257684538,0.021430547741301043,
 0.8565113042160684,0.7306530293040104,0.9834351861358058,0.5889544814682613,
 0.8713186328417402,0.23466140104857702,0.899798596666363,0.6672199132964935,
 0.07049387782540317,0.7326594543030773,0.043492346464792386,
 0.08710631733231622,0.5818773469967498,0.8183238936829204,0.9127756120020502,
 0.4274988777302212,0.590226902893623,0.43754860242761917,0.10755226164081033,
 0.9530041729842313,0.6110271964931866,0.7867313993303227,0.23051158546378925,
 0.7337673813362926,0.3313352094486832,0.07671359280364254,
 0.29679598825235565,0.5662617240474479,0.4272017340176303,
 0.31641337879852693,0.164578665751968,0.1632514538184061,0.4708550938030317,
 0.2980305137270145,0.12991943792244032,0.768010964459711,
 0.014066989488408677,0.03805059108965869,0.7473793127436678,
 0.9887369992697488,0.34980797166470845,0.049470403726085355,
 0.14587494308991011,0.8856199686828137,0.11731179559803795,
 0.4815514872301535,0.6375527472995052,0.8382432954964343,0.7830629189585405,
 0.8266771851266925,0.29018421917357706,0.9551601826088751,0.1164950331686998,
 0.19105286829762236,0.1448928996344101,0.43382311174434196,
 0.11905958264342198,0.3935550703431253,0.9805015627272148,0.931995587530089,
 0.6365952347023526,0.48968115051740435,0.3356500020125759,0.4790817501877761,
 0.24388444643175622,0.5250822829926627,0.8292145003765561,
 0.05940199009836267,0.3035931966384534,0.05441368397411417,
 0.9278543665437599,0.827532063232939,0.04412506421936757,0.5057834373666659,
 0.25223114144761116,0.21717618223600643,0.3800268126369767,
 0.7835108297111323,0.9607224185163297,0.11733340813845705,0.8871844367064116,
 0.7166234602891536,0.01143956285617298,0.29546396781019235,
 0.9078355204675056,0.38036476342610237,0.6620358468577267,0.9338193699834929,
 0.8077691181100781,0.3095895493785953,0.9205530737811591,0.4235367722884291,
 0.9324342632953134,0.25332328078886,0.8398858281042709,0.7672969956925458,
 0.579954258963634,0.4405185800160871,0.6264937905748562,0.09576077348875078,
 0.48049198636823276,0.3312695311967291,0.4173320476731516,0.7736936489360715,
 0.8164053956103121,0.11841656289283742,0.24806153634206352,
 0.5281751069999868,0.7053829740907209,0.058519611323422316,
 0.27146303177193665,0.3318092981219789,0.07156441121717316,
 0.5271366189644516,0.5511319110013552,0.4852521084432271,0.4356371404292758,
 0.8840720606027958,0.6753998521809745,0.8305639359424125,0.5181924280799678,
 0.787502113336043,0.3164048046810931,0.26290139981624994,
 0.056261355479673925,0.8794195936361788,0.9368862023916154,
 0.03928514173194364,0.8070321454304202,0.6518531088959483,0.2377014101821231,
 0.7504319241776176,0.3144779509900546,0.021010494753801345,
 0.7596937992554875,0.6849292418739809,0.4330333081288269,0.6682747710473856,
 0.05007056210467331,0.1073950067162921,0.49807450011366605,0.416432315443523,
 0.07997074365407852,0.15512819658879962,0.466920094124298,
 0.18443203251686446,0.17206200070765898,0.5755598653682932,
 0.0388494256383104,0.9878104619159941,0.35797234388595545,0.6133982406518901,
 0.6742414554828288,0.8002867929143059,0.04500525631641428,
 0.0029088345127590287,0.603683437975132,0.3515755645120745,
 0.1646726086702428,0.2016259477003538,0.08922540164924231,0.5443130755814568,
 0.318709658915434,0.44142966241223713,0.4433014568366409,0.6418274672566633,
 0.7250196240691797,0.9927369832433581,0.5635164689177272,0.2701624069709917,
 0.14815250594091067,0.021310250686724985,0.1663702017280958,
 0.9919587286406591,0.5827569729597383,0.8993858442953854,0.07716196644803053,
 0.8531162016684777,0.7568918213278126,0.8589001747696752,0.5431159959136718,
 0.7240133333293463,0.09702687999984205,0.7679064479393345,0.7983179540963631,
 0.08687495035858461,0.1581821562896435,0.8088400106847466,0.636755771084856,
 0.7342934462689515,0.8119227674971738,0.5245427596962821,0.38200775243894547,
 0.49457583165855834,0.7809294446554623,0.28105200795022733,
 0.9470935886578911,0.8950994247403762,0.6480739672484286,0.20062540367083592,
 0.4792552126447184,0.18427873889555402,0.49128431602619826,
 0.43629910463066246,0.09094460891002965,0.8126496611359045,
 0.6215318749819444,0.2403370019094796,0.6981407169656375,0.7051497413783298,
 0.8845412817673882,0.08779800484863887,0.7447568536265209,0.2622005136745392,
 0.5984190763482824,0.6743842511273572,0.5872625279703736,0.6559100890881181,
 0.005822567955019453,0.5134105017208076,0.2756921849663375,
 0.8399153720890439,0.9954860459689556,0.2945400523200794,0.28048758873701696,
 0.8027143281867208,0.82223297194747,0.154857864681624,0.27604859489863376,
 0.7075656886525414,0.6145925203164562,0.340095671545381,0.3816255592116944,
 0.7749119629787379,0.4263895361151153,0.7531490257668001,0.2599853869168156,
 0.2728866213121559,0.0348580694197449,0.963846055111188,0.31667665570373305,
 0.4531669716714547,0.1595131335688439,0.9103301363527467,0.39441248531874673,
 0.9417233130384964,0.8999940180712726,0.6823203868798866,0.22805487657922052,
 0.8924929728390073,0.09021229823495658,0.29719173964224577,
 0.016042390810263707,0.6489159277174519,0.12900794190956688,
 0.10764040198261937,0.25451785216764644,0.10929528831296509,
 0.299083942591182,0.22052010968722613,0.47324225013999366,
 0.08280288533403113,0.5544427799328674,0.06910726324548033,
 0.3944645114891274,0.7357677982276671,0.9615082963102999,0.5729088964741766,
 0.7979122733611603,0.5462109779828578,0.6719298300793012,0.5020712879029366,
 0.1294472006477717,0.9508519320564519,0.9970435263697319,0.3022863911687055,
 0.5039174941929166,0.7690067480655423,0.8583985454185143,
 0.052166893014786275,0.7775423570943899,0.4665800332299096,0.900597659628826,
 0.9531394127914319,0.9543918229486287,0.1026712921087346,0.7575679007377489,
 0.9882174395775806,0.6827778041696931,0.8711511584915859,0.7890450624839962,
 0.5530617899366626,0.5065659366467306,0.5526399701953649,0.6696633744807025,
 0.881624423239409,0.1924102831015635,0.355018701023079,0.48841297996557276,
 0.4248385656121354,0.7901502594068438,0.971375007084869,0.5395534274270599,
 0.2521172197424879,0.21490195095140352,0.6435539181140817,0.9733780977809005,
 0.18838120039823392,0.444070956484117,0.1044944397038432,0.6996273734357029,
 0.9153188962056656,0.9333679525040884,0.8457068116188496,
 0.007541690596978623,0.04453110932665105,0.4346843387385747,
 0.6961199561368773,0.8410045809658954,0.4781104429066835,0.8207619732974807,
 0.21550472321260516,0.07899093707265725,0.46693175141968,0.1024577996953866,
 0.536848605854945,0.4067769463766362,0.9935379151724497,0.6943987557195124,
 0.9802903389350023,0.45987569087102165,0.3833919680834198,0.8629937402329186,
 0.5668656347541708,0.20508357419319911,0.9454227881336935,
 0.22276226750986106,0.9529959300073764,0.7434498851246104,0.9012955804474027,
 0.44909268492893173,0.7512889338790573,0.8838264574181922,0.9277892515515228,
 0.7852330350244053,0.28083721320892396,0.48503933120683174,
 0.6085953087916881,0.9714465671360072,0.4818500533489647,0.8099637523590298,
 0.1972798841261375,0.18629532915937963,0.22735901542465153,
 0.5671404088173307,0.9711302640242084,0.780230495804215,0.2828315723101311,
 0.8987223531719201,0.2054811391698107,0.35373344549000063,
 0.28342984342048416,0.48828914188370265,0.20668553730922468,
 0.8593662486631098,0.3623627803495014,0.10671104991512115,0.358233087413822,
 0.9486521416840821,0.060112783490986144,0.22152354221465764,
 0.3043800031779934,0.09437356117734319,0.6339726767386861,0.2105790915535647,
 0.5612651243586595,0.9803069794655629,0.06589834061547695,0.8693907769989614,
 0.4859915262782123,0.7902947292339908,0.6092224260198276,0.40959908508210285,
 0.09392165233529803,0.1297028980205004,0.6475370119496875,
 0.07788144537333352,0.4677853637223218,0.9468572302015643,0.6939482953166664,
 0.8473713396883715,0.8173509153895714,0.4145207838040851,0.0359539288322126,
 0.5635172974843703,0.21755155335315668,0.3953106367665874,
 0.12693801147951822,0.9234021227667709,0.7492184862616018,0.5062555664571743,
 0.7335003298205814,0.4006865879677354,0.7764717195045544,0.4211972414109878,
 0.8486853762560869,0.8649482811614874,0.5868365072279151,0.3354741188869118,
 0.8869447209575025,0.33351290479075746,0.07884628305876351,
 0.7554702202042592,0.2347814492868706,0.7900730148394419,0.7998476575714115,
 0.39586745099234966,0.3121237829075806,0.9873963938813645,
 0.04992847061292627,0.705467532176624,0.6598773334349843,0.25083526372185105,
 0.16657828010199516,0.8132702943850117,0.8033417712408594,0.411058997543003,
 0.7679672091371675,0.686882692340698,0.1791792257971273,0.7501964267990093,
 0.2573726506585283,0.584895802557383,0.7018084399230801,0.4803862519343567,
 0.3571408305516941,0.5241923197470002,0.14196021936500336,0.8481334830067082,
 0.7428050077472284,0.8445661776839672,0.649001270254979,0.22139997122459731,
 0.9629493243039335,0.7683432157598888,0.2281078063468115,0.819877019157851,
 0.2935189904239667,0.44531020913197283,0.3305926280724978,
 0.29625184511917246,0.8807591363280416,0.8567229868464494,0.7003431378130333,
 0.2606939037561298,0.4521876302359419,0.6335698030366678,0.1455093051519074,
 0.02547862706003834,0.8311609990184095,0.23893152038897914,
 0.5206997862783651,0.45105510370248636,0.9211143089145106,0.7736962174153031,
 0.6582989762631104,0.3942179126561204,0.6861137856781964,0.5605873525673777,
 0.7379501015691885,0.36652469372948104,0.24547181849989774,
 0.7645237875218014,0.1018817357279167,0.320963524304298,0.23127768112222535,
 0.10219407195827246,0.3968122166268956,0.4914851034743507,0.5802984427086612,
 0.3416407848570664,0.5311796826970405,0.22318010251256504,0.7941352004804133,
 0.039322281179470786,0.5268400981140603,0.5965823825987331,
 0.9135244202968738,0.08283281086181793,0.9320358163421137,0.5022606610737635,
 0.3385462566589559,0.7423218608431346,0.5260697677370234,0.5858193645290909,
 0.5418580520135378,0.06777406690974164,0.4252336951471458,0.5490003090955212,
 0.25359899718369827,0.7702737053757697,0.5751499472463029,
 0.09497087164585272,0.5592336623583803,0.5699836384786034,0.7810505195234654,
 0.3751531300317249,0.8025449522687832,0.7985562794598863,0.41938696051526003,
 0.13132119942670928,0.13066783944007454,0.47924798799027657,
 0.47756098621111565,0.1688415697179051,0.5416627498080966,0.5464200947891996,
 0.2723395628803522,0.6647840573378886,0.2787993601596508,0.4408825510980715,
 0.5693017600053623,0.20197361238534284,0.48011215385554173,
 0.6188601030402323,0.5567880996912687,0.9049261165475869,0.40252296603879834,
 0.014873366241106467,0.1181307109449824,0.843166808970958,0.9935682791776902,
 0.5560475815167203,0.9776540885730296,0.5999204111420824,0.43953500970633674,
 0.9511253924263273,0.17448533916388498,0.04814846043899934,
 0.7401629982520112,0.7909340806759855,0.7126022456525282,0.6013131740219815,
 0.04829035326885933,0.3140572708856507,0.30417583531662673,
 0.7919901381082809,0.6778674221374112,0.3802659441319288,0.6501006430171966,
 0.18840738247332878,0.9582256270968275,0.9205962367851983,0.8423945825519767,
 0.1206882647813674,0.8512053212568987,0.09123258258608857,0.7962618406012609,
 0.259543463306958,0.04353510694581031,0.9154307054251839,0.01458445118322127,
 0.004351018602976042,0.7152237104614616,0.1762206187610056,0.401323826181354,
 0.525388691753399,0.49143658360472176,0.8831997072788923,0.6459393003755489,
 0.06822745979642741,0.2158498024380513,0.18109476545115255,
 0.8279922070282048,0.4337929424921765,0.16065988700998646,0.8523875177862785,
 0.3231052256618624,0.714071714599535,0.4947681987883067,0.18046257714770642,
 0.6912204324699965,0.6503800478163807,0.6659854784242709,0.987581410949073,
 0.17038321711520088,0.37116486328408604,0.31059428189167915,
 0.8183861089583843,0.3033048412462884,0.2036189396577749,0.5631108086742009,
 0.8187698626056155,0.9814324811677632,0.983361563356533,0.984803695400985,
 0.019288943336994824,0.6696803027638136,0.03386715754400438,
 0.11207965235651696,0.5044109795434857,0.6846960795360939,0.9997764056487789,
 0.4570084688383069,0.8846938729750853,0.5287601616339528,0.9590096792698596,
 0.5423471284498287,0.3272456146075938,0.772709690333275,0.3603649920879337,
 0.38212072198372204,0.24802857452803095,0.5527040802276296,
 0.44353319797528035,0.9073434464284187,0.6562854008773134,
 0.17046138007308564,0.2553754039813133,0.8187327535681321,0.8131131204950289,
 0.040716300138989414,0.47224782984423963,0.21344977873133275,
 0.05922711389275159,0.5342036441814735,0.056244661223368464,
 0.3253780246710124,0.46269569030281776,0.9419745006171423,0.6720996615404116,
 0.2911575684043648,0.8223198360024933,0.899606910431973,0.8683849590677893,
 0.6377415011639577,0.6257999983334873,0.6663618058341434,0.7824081551836378,
 0.4478996680872108,0.2826850114531303,0.38055769231748937,0.4127165085114274,
 0.8654236059255619,0.1579787832981746,0.48457519633228774,0.8949008682503324,
 0.1689477368500063,0.8032558347611908,0.44663954679131646,
 0.18397419732813391,0.4570944419773366,0.5722133540094654,0.5296925083838432,
 0.0882531995074618,0.8252424245487283,0.4549119445937231,0.10322215795266143,
 0.5055047509851178,0.6558684214866681,0.5995275998306498,0.5091935100524957,
 0.9639226154916019,0.6029876143368782,0.2210518744204153,0.11129947432744935,
 0.7956108531873307,0.4185804981107435,0.9064095166349168,0.9594361661134826,
 0.5858384431806847,0.0764768576755771,0.6123170146205776,0.4747872337000766,
 0.4072572147068003,0.26590097974637195,0.6716290746473257,
 0.26441420558301487,0.6052815030685826,0.6620553942961185,0.2295527040372476,
 0.7732749083508426,0.004001659380759248,0.44246770653662515,
 0.14773971577180656,0.1727521591596458,0.009127675219333553,
 0.7251912360187999,0.8639374118124782,0.5368369791407834,0.2420549543396855,
 0.2487525864648925,0.645865401348056,0.4448425798301615,0.7513719386100218,
 0.4098715558243703,0.9050164669887282,0.41107597784568495,0.1684067999516271,
 0.6846551049499491,0.6984883812684196,0.8385910251838663,0.7119946478092112,
 6.806621945323332e-4,0.8254203270421335,0.577265315671853,0.3389590927778794,
 0.22504390962484777,0.8009993910091213,0.7671783259306442,0.8802034333504988,
 0.10936932538403854,0.5291278935880194,0.4573505178202213,
 0.30952592252366573,0.5821564051025399,0.6066191873403102,0.4535522024486969,
 0.7237821720480423,0.486159184265722,0.3418898724289645,0.2442805736055086,
 0.03858962947388744,0.03418455297003453,0.36675601678099135,
 0.7139664351588629,0.10585108687977718,0.8573212640847316,0.9910845212151754,
 0.796969371881343,0.03162044489495963,0.18846466580377608,0.6992970450828073,
 0.3336711504593297,0.3367257525673073,0.5120500224519444,0.4943634863610814,
 0.38360498855497505,0.8431572591607619,0.3964160320847986,
 0.24034636285450384,0.9612380393602955,0.11467474131717648,0.349295079935917,
 0.7084703249029713,0.4068732830906575,0.09038177594388441,0.3345905590752396,
 0.41254132910757546,0.6387034043587697,0.39580513542771,0.4234749125500936,
 0.6353612366078206,0.37595783968963703,0.05013068799422249,
 0.8213294867379272,0.5437113557725691,0.21207372288168191,0.8542034801917684,
 0.024318726320276984,0.36083563501513016,0.7288220997549707,
 0.048795409080865104,0.23029435564560952,0.3496441135702848,
 0.3679765629780758,0.9757515065088233,0.6384452220078787,0.8689007927135822,
 0.16615358789298074,0.9410592278474977,0.5690078764564517,0.2329250615036954,
 0.38905807557743155,0.19761328144281898,0.01353138392326636,
 0.6983429786610406,0.7210854958948281,0.2122217484150869,0.8419906025551128,
 0.03790979354121893,0.8980163361424989,0.2191253282813349,0.7171919162810408,
 0.3645381524714332,0.9298036060063477,0.17974491788154368,0.8252666870715959,
 0.5435185984910471,0.5043670741911571,0.35994036063351864,0.563358985835172,
 0.19974371604851848,0.7087555697468189,0.6238027346330887,0.4639847178706177,
 0.6469802291473452,0.35808140146593037,0.17850787133334722,
 0.21354281846119505,0.7932838703539563,0.4915221951794214,
 0.03712857816695414,0.8502797752191922,0.11000304675601025,
 0.7313009791830524,0.2676693891323293,0.8806136941248909,0.4435701554790368,
 0.3077329939979786,0.040697439656590895,0.6373644894391056,
 0.11271418036671565,0.5769403558819762,0.9898267190355431,
 0.16496352977497541,0.7235806992945077,0.8095092532234144,0.3866287651730458,
 0.8909538680123605,0.873691994759033,0.953390570121256,0.19594992540874223,
 0.23319651080974768,0.10714400361710208,0.5209547861434491,
 0.5061656124370605,0.6903316555663302,0.7225965606180725,0.7117695381984354,
 0.6451526537691699,0.5292429295719105,0.8316097687193603,0.8214939007857383,
 0.1251660200900997,0.8313034939339731,0.2076096198330355,0.5372761458055004,
 0.23833789134807393,0.7582786400405603,0.33173428435282626,0.494639805329565,
 0.5902226120830895,0.37179420383843964,0.7371912947240569,0.9374704128467886,
 0.2672808194699512,0.12289320576598328,0.6941101193052147,0.6166899249266494,
 0.72095449365178,0.8745852028795991,0.1857140686733847,0.9877600761627878,
 0.013361177469618069,0.2774885630412097,0.30793350658804997,
 0.6810616580914759,0.482568263984957,0.709463210983363,0.5825096889124507,
 0.12093504501918839,0.2812673379404481,0.6298437875013326,0.547995850263006,
 0.45257669713466386,0.03724476480257477,0.7095839911057635,
 0.3052867861097819,0.7681003206756514,0.8578567155653826,0.4369156312415897,
 0.46019932328030333,0.1330791704434965,0.507936466031674,0.7365463304553613,
 0.9196848141716589,0.5655383699071456,0.865660462694014,0.3087121960556122,
 0.9603328688804492,0.6657534246416565,0.08680280982575383,0.8378557523705471,
 0.7619794215503473,0.25610215928769153,0.07346366023921025,
 0.22250970500619283,0.13286895274881505,0.8986665111344415,
 0.21292023176993236,0.6127975973579904,0.6499994017905101,
 0.27551208179372677,0.3509819053946679,0.3018665920735347,0.091321153147077,
 0.5512881896882275,0.5193272130613614,0.8280844351565517,0.5040659818954422,
 0.1924989119780851,0.5107681260380397,0.10404004557880853,0.82213648390954,
 0.5895461142674863,0.9683463407478236,0.7910069285628418,0.05239481846038041,
 0.7169642154670006,0.20459107436569923,0.22478092717791887,
 0.5949654239305788,0.5519126524048545,0.08434832106329604,0.4094994951213843,
 0.8182206647842775,0.6608270745077087,0.2603930008959612,0.01358952436477856,
 0.9693491640466461,0.7607593730035683,0.9791854480155424,0.24288131571910399,
 0.9751578519241728,0.919839381422354,0.16542749348637198,0.18630507723495493,
 0.8160029435042319,0.1803930477461766,0.3586730073016754,0.05641986433289359,
 0.9046823497316323,0.3831066453419245,0.2777160101737234,0.8133835641203039,
 0.4708309720103574,0.38513046514595395,0.04792999637312911,
 0.6860075927209598,0.9553623107855264,0.680293137300259,0.7957995331171609,
 0.43033936472301537,0.1797555310953889,0.26832260219201653,
 0.37361499420029043,0.6652567400261842,0.552451346501571,0.6156793880048097,
 0.0976170048348356,0.35275628464807507,0.6493850481926696,
 0.18905367272569895,0.03456797170851422,0.8300946857832944,
 0.22341526423760794,0.6400158487914538,0.9783178245055655,0.6924106975113541,
 0.606872055762486,0.6137310355300292,0.17445900567726702,0.23404396024016416,
 0.9720448859419258,0.16430064993524007,0.6059017458470122,0.4668475642411223,
 0.9380624706284229,0.9514040615268836,0.08391025301873611,0.4098849170570681,
 0.244262107739333,0.3308846524790393,0.340678780229432,0.13578208578409456,
 0.18139964031006195,0.9377597214253758,0.8263255554536004,
 0.09401462896882862,0.3048565488995424,0.26160674768778813,
 0.22790600484567192,0.34565745588766483,0.07747818739399692,
 0.12999216312433648,0.9890981498418769,0.921592334599699,0.6743249128732826,
 0.9758182216261555,0.9693126744248967,0.16348741638878528,0.3514476743523075,
 0.3107189767199062,0.8761521967278906,0.07315117272978344,
 0.23601262926489297,0.3503483111976903,0.2085437285151538,0.849053127169766,
 0.15680683744285273,0.551170955694102,0.5963968044891059,0.1078058695223868,
 0.8631366270893015,0.8126964804213005,0.08911312447886921,0.2579229719410372,
 0.7383732410499181,0.8113697551013633,0.3888218471790772,0.5495967656008899,
 0.36219010990457634,0.274449391808415,0.9713058540383591,0.36128849664374774,
 0.4952800454960794,0.6177542877013679,0.11847086653377525,
 0.43475175655319487,0.44054393647144496,0.19151022504368986,
 0.7419582933276362,0.6742078988288708,0.6303432286438106,0.4037843330702048,
 0.7346398308849109,0.8994851854785099,0.9519577849294768,0.6121745807730716,
 0.28852466493609075,0.0729001868467235,0.5689533141653577,0.215481378910806,
 0.43976404084078746,0.9161066137507585,0.24199655696662747,
 0.5752767304497146,0.9978229549474432,0.48749253096507716,0.4190707579044073,
 0.9748491966588052,0.41320353162183665,0.021652846242624757,
 0.45877799648778894,0.5256325296770603,0.031045564922500213,
 0.411646710454196,0.8039023086353683,0.256979352913732,0.9736598704056068,
 0.7750177051534305,0.13582930069234178,0.03150044528133622,
 0.4976473250512816,0.4651012378307856,0.07836370391229974,0.7332346271954178,
 0.7065145065957312,0.10665934209171724,0.5295056305453107,
 0.14400274981483474,0.01683080278469573,0.8381336468755951,
 0.46970179240186116,0.3412179685972363,0.9390295518474625,0.5193892671082787,
 0.6077143430246265,0.8459250176891592,0.7218026266595801,0.3686592073785925,
 0.16969586681184978,0.9268159143615319,0.5129350398042405,0.5986353003598517,
 0.5232470302808169,0.6051770506444273,0.5319135701960986,0.9775300188813386,
 0.23880802931515643,0.49178322328581425,0.24102772326330624,
 0.1651675528112353,0.3693453995606478,0.6564544183912231,0.08678023530099532,
 0.0507965264326915,0.3361450739155285,0.4154861102526686,0.9363993158537047,
 0.0268016040237824,0.816743408680784,0.1555192120279434,0.24579852055544793,
 0.012358188406526027,0.8882941581030914,0.7077888187823002,
 0.1727421249991593,0.5390347839305454,0.7982379940767466,0.7687480550056547,
 0.37039137990305027,0.8689757115281306,0.7524695405116572,
 0.14453515636606407,0.19529999925547525,0.8364041442462141,
 0.9240930436800909,0.04379174978355893,0.9515935623355213,0.6831753934457122,
 0.8685047856136541,0.8786953037600655,0.9199556670108531,0.05596543004189258,
 0.902716027576752,0.8936572294917482,0.43147888873320994,0.3832841884724164,
 0.6315493121314594,0.3269944955147587,0.2779264166242321,0.6220184434808829,
 0.33589080347274747,0.1473793787870148,0.08535268086449876,
 0.42514405410848766,0.19596069987399578,0.8008155455067736,
 0.26711372016643486,0.40329788842490577,0.24961344530282537,
 0.9095652512255459,0.2525051813083592,0.07200040908247585,
 0.041461354704931175,0.6610547419662218,0.9785099965934954,
 0.4114930744774775,0.8083641539525439,0.3713508051890013,0.5352682487877936,
 0.4217040511078123,0.19183080979292366,0.6594787786854222,0.313441627925793,
 0.9944732631195178,0.5976566129121027,0.4744387013587307,0.9535538650373621,
 0.3271419901667716,0.052738029442157774,0.23931150963226877,
 0.06265409598888938,0.42649049382342485,0.2544822318359343,
 0.5891172094029098,0.16111376475287342,0.5036129646327477,
 0.016608465160724006,0.43523724558982757,0.6382173367766175,
 0.6027282877320974,0.8850753422932643,0.25900310867466536,0.9575908495053043,
 0.35159659699992196,0.5747353357510147,0.0078557316829726,0.4126322827461466,
 0.8393101291597395,0.36766894660016314,0.5992427392338938,0.5497097180952322,
 0.06055242304073416,0.1892774067502666,0.9862833857660699,0.8245646891500249,
 0.26192746600820116,0.4252536520292768,0.5587043577714201,0.6042430822613296,
 0.6799404976626825,0.062435840212188065,0.7595477448668762,
 0.7989275190978574,0.06177551867004927,0.857657490984353,0.12157628841587709,
 0.051719425384579076,0.10062164632044279,0.0759681191597591,
 0.8515792709469677,0.7298899901061263,0.35360739515686135,0.7860167483204201,
 0.14375422483039046,0.7456461224606545,0.5369227636434989,0.9698132969623786,
 0.6874619749057023,0.9889435928869583,0.36822867451543406,0.9322984257109084,
 0.07671172469204546,0.303628631744421,0.10438085256386809,0.509389799004958,
 0.6470649233648019,0.9616476159191901,0.3909209934952935,
 0.013174483189260844,0.2938661622492019,0.4630494584983016,
 0.15476151863707788,0.3205715305365662,0.7949335366166327,
 0.28585161368446343,0.6518840179429588,0.7235151064631977,
 0.37787757340850725,0.2789925282102055,0.4415534471866991,0.7540054948834587,
 0.12008522339561156,0.7979002411656526,0.878929671407349,0.9107151516197954,
 0.6451771172725583,0.6839060620643845,0.016244568061137166,
 0.6776359205834225,0.544282332900303,0.27639632172906325,0.44293613989529723,
 0.5776136524672263,0.7052708980452914,0.8591582595965092,0.5487428564229643,
 0.7106406188264489,0.7960056976979958,0.39003797972714727,0.6021062745333239,
 0.7781361810436491,0.355951354650489,0.35000624364946575,0.6486444278949981,
 0.16708305089987074,0.1090118788625497,0.25592258852417726,
 0.7423946532987697,0.9659394242228746,0.22690641619704188,
 0.017845414945554028,0.4683961894665911,0.7204302732503114,
 0.46327265748584967,0.5904496797054914,0.3258898890803442,0.900743383728899,
 0.2759480667179176,0.7266299367498763,0.17232706681975085,
 0.05287427442975767,0.21572934433641788,0.6329043262080137,
 0.5291672402013772,5.6592175545211454e-5,0.570897922439491,
 0.15990606100155258,0.9460885132052117,0.12402022641887023,
 0.7930014235156105,0.3612351194839629,0.2886317342841108,0.8127817135277777,
 0.5792141994640333,0.10205844133431363,0.6699131394905915,0.2833692630824939,
 0.6127915840015649,0.0619110314747755,0.15862702519536487,
 0.44316931606453036,0.612431896805775,0.5166485857876055,0.6328127545420315,
 0.014574843783231928,0.8982636156944485,0.12080269200361382,
 0.8192950468178496,0.17546235238174657,0.7410047761041555,0.5028212362586475,
 0.8856530021703243,0.052235253691161,0.987714697173434,0.4394173148131859,
 0.4334077005747696,0.6690168576582095,0.5378688237652063,0.4720757738076753,
 0.6459395623367172,0.1962630881356564,0.943212294111484,0.6642090436887612,
 0.9353454174671161,0.6470493695656558,0.08982057052620929,0.8128952685199397,
 0.6381085187735509,0.5681245879019025,0.18871877786098434,
 0.19942434925698072,0.38303692844918746,0.25839954571551993,
 0.3490002906479788,0.9128107844779649,0.1658232341378465,0.6863625913221123,
 0.9761897601751957,0.6140910482826414,0.5662845201184508,0.6846050283024163,
 0.8730605508587439,0.779084377772419,0.5473825239562036,0.2700735502431366,
 0.44094879840448775,0.23168245689333533,0.38797234070898845,
 0.472358872743428,0.634373293404056,0.48786778247252216,0.9344191032807336,
 0.7516754940644839,0.17689721944586378,0.20974620026123292,
 0.34806562303681554,0.09174958719610049,0.3176784891457859,
 0.05734422908828862,0.41452716479378693,0.38119446842897353,
 0.5865368644751554,0.653190195718575,0.06667665582922166,0.6526040507977676,
 0.44847535992695975,0.1751456797941935,0.045806084944328695,
 0.21293476581855675,0.9205676242725642,0.44602529672054203,
 0.23919530994959393,0.6176451174674761,0.9839844824898148,0.4422345840288645,
 0.28151816085527137,0.20087197167727444,0.9765277507073383,
 0.6531706937767997,0.8163774189538037,0.6225779836111992,0.8998319691690357,
 0.7757028884987199,0.18680531994985505,0.9053866613096311,0.5041926938481204,
 0.6701041537859999,0.6002135150727131,0.15574344257273487,
 0.032714970076021976,0.21220908152987272,0.741167575765875,
 0.0717737725383627,0.9844829002143859,0.6346660882586734,0.4039933522421535,
 0.8451044510817644,0.5156637323596807,0.9068947378685731,0.9910932081718852,
 0.2977999283383179,0.352012138055268,0.9690697635273113,0.2916206941792454,
 0.6211333009207956,0.815874311857162,0.9426801131430089,0.6923579327929891,
 0.2883819215582626,0.6802087566558336,0.21332339701499003,
 0.03869466411804168,0.5789533773838607,0.9599538788414195,
 0.18042458345669743,0.985519267162895,0.9734660324389204,0.43186135877125054,
 0.5836819607348489,0.9438012611543584,0.9775684718596441,0.09762315849554504,
 0.08290828592764565,0.3196798854505649,0.8694619450120313,0.1635755850605508,
 0.11360430311343461,0.11077470033610626,0.5593764154162433,
 0.6257829547824663,0.13291613760260945,0.8812361560311185,0.3499091110502899,
 0.8576115475749668,0.8635535415584411,0.6453215965095098,0.2151512805577792,
 0.49699722740252505,0.1708753651666518,0.6205936109930439,0.3875295467534393,
 0.8068347909134508,0.8504769712005362,0.4750818769972832,0.3512795758890138,
 0.8155340764178465,0.6876337685854385,0.7661585813311711,0.539815575251463,
 0.31245066805432087,0.29362467732775643,0.4417383227005077,0.77453714582205,
 0.022812867434577333,0.02903230981018967,0.8874702253444523,
 0.24361065852958041,0.7632552930868562,0.44345130831024304,0.969897768514524,
 0.0562553551323165,0.3781261970856915,0.26902856674080233,0.734811656203105,
 0.16043982000568635,0.8808075821205813,0.30241860100567497,
 0.6700565529074787,0.10317640215982782,0.22832278206684375,
 0.5599872517247257,0.9354295949831056,0.5233634988987107,0.04695454746157113,
 0.14547206796255052,0.5245729592108934,0.9297405008917086,0.7655462735816386,
 0.3364668519788555,0.32464728518978214,0.346488122297245,0.38424719603377866,
 0.3065196708419139,0.1179860237225474,0.9489826644803085,0.4958480607276621,
 0.10978630754685359,0.16950430516067894,0.07505804479171885,
 0.14647552203263636,0.5799742636848874,0.8836689383460179,0.8489175388757241,
 0.38728909214286245,0.5130100133302666,0.9372315533435122,
 0.30999249696560216,0.30102086453896937,0.016912227501276966,
 0.38528318891939417,0.7518513182779332,0.027210522133920056,
 0.8187877896759757,0.18614490968954878,0.5404546564927296,0.3612913708414772,
 0.08908681162475784,0.5244210060896879,0.7255971863184112,
 0.04378750292936395,0.6739903431087372,0.9470875416471143,0.170807723486889,
 0.345136563525101,0.7737492644925749,0.05857053550654501,0.5582574246962773,
 0.34508537104600856,0.3402599088327183,0.36412092319914746,
 0.2427525284279406,0.2906345688774866,0.9656076710255738,0.9785844246335943,
 0.5147685806313056,0.3945790619356002,0.7870261361711345,0.3457173941607975,
 0.7196998162133981,0.8982955781450374,0.9397027582627507,0.5564641723919125,
 0.058873709632853766,0.1478393852326776,0.9135506559664559,
 0.21164267253731506,0.8026069328396728,0.8027727996593328,
 0.14742036879495368,0.3089611956897421,0.8167787883570184,0.5265607239654463,
 0.07953645222194328,0.2243103947525097,0.2229269905693808,0.7440121575796189,
 0.12816375471740338,0.4509869144598885,0.7920030225768555,0.6023303358138804,
 0.7203478168874926,0.6341147649952862,0.03593348141678998,
 0.10356956078907165,0.6433377229095486,0.9534247982675217,
 0.047563227301732125,0.9060830290069766,0.8572811713103475,
 0.1136489879073398,0.07272179729387673,0.8857105444876802,0.7844790956318896,
 0.14536160225936434,0.5809281544538024,0.013081169895246547,
 0.924607378792063,0.7736179073052376,0.019706130748597284,0.3193897733110359,
 0.0558113289204516,0.21829451927781007,0.5202518203566611,0.81790408760424,
 0.803371792964222,0.5262402448791547,0.5977648633291451,0.32557049834990703,
 0.2945891682378311,0.542880393437397,0.016498570206679664,0.4972367221870314,
 0.9657252911645585,0.49280426087993756,0.41399583081126745,
 0.2850519261522185,0.934889309365198,0.14423531830020508,0.9723528000592411,
 0.6559641642965489,0.27042984250852053,0.9983125557831655,0.5722003955840678,
 0.09009852118868555,0.2279456342637608,0.20396406730788819,
 0.6593259712818094,0.1603760969944672,0.04806614336279502,0.9493124421458337,
 0.7374870324845819,0.37691764168503517,0.9406742678328244,
 0.048757176831129345,0.8186162982564739,0.5046239851782476,
 0.5154939674663078,0.17732702385049293,0.6214979643671014,0.7124062850019899,
 0.7311274612587055,0.7444065546343136,0.8180238828836376,0.9650276858996982,
 0.8316122005800786,0.7855530135179097,0.8734642342076258,0.5017965427544799,
 0.5441653828489017,0.6224851462953068,0.48968343762159383,
 0.42731476992131623,0.38604864718717424,0.80475291318346,0.4521173250024061,
 0.6172264718392468,0.38192127904666895,0.6013577554587842,
 0.40733133450478565,0.9625651050905549,0.4330900989381101,
 0.12390429695079053,0.8675570207105874,0.5501018505570532,
 0.32640326230341676,0.9669191541379367,0.0037388192826876,
 0.34980961096699925,0.8539477938594197,0.19336365433545322,
 0.10096444726787823,0.6269236762045971,0.7008281376767135,
 0.33562922984856314,0.1804368566776825,0.6445254170101675,0.9868790690624326,
 0.2796954395717739,0.08620986607129111,2.1804706331673973e-4,
 0.14720582264853277,0.8535310666778343,0.9446286260723036,0.4478230837680607,
 0.7594109233048706,0.5429916174795067,0.31018254172636517,
 0.21748627569502776,0.10795119564946853,0.11824472515208972,
 0.8047456171372638,0.656207660332683,0.7614240853866858,0.3107023365421806,
 0.6951437774929944,0.1345955507636587,0.7034250770673202,0.642648431450596,
 0.06409753017660769,0.4271951402488843,0.14054036806549475,
 0.21238810753707438,0.3209314393845073,0.8054548301262956,0.6585057972033412,
 0.2205857447224261,0.717786395641593,0.12658515588888486,0.5456663810243936,
 0.015094032151494852,0.07015673385381838,0.054971914947815725,
 0.19721489689863858,0.5391995214834885,0.6828145249181822,0.3546688117044007,
 0.16615731784616472,0.793924507757579,0.4315810934615457,0.9421607719472602,
 0.8028769133714289,0.7448918841244431,0.272995427875957,0.7754592445830698,
 0.2980445716601656,0.4104414463507329,0.8959080721851405,0.3825100865721216,
 0.41172068397568073,0.7398413711465585,0.9754804128349491,0.2350935750232781,
 0.964660862216451,0.5295642166794936,0.20110930746854627,0.08675995379370605,
 0.5116014116588542,0.9973611819685735,0.3269970847122543,0.6777747606048983,
 0.22887772705659803,0.5140759103889321,0.19728452938769325,
 0.43630556859443426,0.2025608123067093,0.2316076638637976,0.3140099251240811,
 0.8554273100224179,0.6410612468892846,0.4392394547872417,0.6780444903549508,
 0.5487017571260462,0.356725409606618,0.6918244033454007,0.9796109468023311,
 0.7773137752092398,0.11772497070791266,0.6212408702977668,0.2715599355070416,
 0.8219942530281311,0.30225947464887737,0.8450401068119363,0.8314618173343566,
 0.0010439350311347573,0.820234635639538,0.7107006200827368,
 0.040655639798870324,0.8578280531909639,0.07684591507597416,
 0.41425282324722335,0.32477831782720945,0.048454991239031964,
 0.4595740764821551,0.46863962175915974,0.780273899351128,0.10667200207539884,
 0.5394748336407572,0.5536449895914292,0.6837263678304901,0.40580081174282345,
 0.1987528598825603,0.3517940248438328,0.22236828169924205,0.5571265611685201,
 0.034814486864138905,0.41448683134404773,0.5820841300335999,
 0.3275076527233627,0.02649606945543026,0.9143230880332993,0.8263667917020328,
 0.2826108374690983,0.1447636602297644,0.8124744433105497,0.7604607721331541,
 0.3108340622492183,0.9149802898863522,0.5109464673688053,0.6552142174042555,
 0.9795162368931064,0.6598581756306757,0.7979068889393981,0.3525605091486641,
 0.4912976438660912,0.5834200462534098,0.3562346890222057,0.16515669671679656,
 0.2389725006289516,0.7244552018913889,0.24944179653298448,0.6026049369755716,
 0.3042027768386003,0.8180633884562134,0.03856267247558576,0.8316591111012726,
 0.08811025949720319,0.8281888899366686,0.5067209058848942,0.9489728958488753,
 0.01610653240380555,0.32023616384339526,0.2140767191871703,
 0.2191779652903303,0.002506789044855351,0.13494432647875332,
 0.6183731229393502,0.549389583161197,0.7253699120507906,0.94723479940156,
 0.963542823142555,0.024332489535328294,0.5734278826575512,0.5300545890868048,
 0.137272867803385,0.5809840024846653,0.06676850854081195,0.46048487110871783,
 0.37802592960455417,0.7728606091164137,0.5930785613314611,0.6237946185036369,
 0.5729033358559454,0.20630078994073087,0.7767292765797208,0.8238850116265067,
 0.936815259158651,0.15545620129940774,0.7721617359007354,0.6774833646851203,
 0.8122874548154708,0.1116880704315022,0.6427261373065081,0.4687755536270555,
 0.8869678856710903,0.1564164497347469,0.9560923064072997,
 0.030905672660821892,0.9694809006019933,0.9436062085220899,
 0.26076309801791675,0.8549237047823748,0.9577457950087385,
 0.09293677101632924,0.70814812989153,0.7692422648198594,0.8169903831583023,
 0.6372042137644601,0.629647128588525,0.34851823504948587,0.5886302730621991,
 0.289017711942528,0.7550945291799959,0.11685517819594482,0.6455824756511283,
 0.941869262630319,0.5263383286200866,0.7981740811220723,0.8007957466763905,
 0.7976873827979641,0.8819815098596006,0.5727984850111996,0.03373883113188436,
 0.3813154733429659,0.7298877934367448,0.9131892203789865,0.1925905318088751,
 0.7804163675780997,0.5413377963337825,0.14761079629386265,0.3820398611692788,
 0.29156329559413086,0.8413819017115952,0.45516923390282527,
 0.6962774839057897,0.308629260719802,0.815306670824827,0.056427136447365966,
 0.7138015884367997,0.1181679185616965,0.8326875014586611,0.8267806530324384,
 0.7781742405800378,0.9904439865809677,0.232419887868077,0.8758513907599756,
 0.8837028319098136,0.8543574442717212,0.5539869776748627,0.5279788150352438,
 0.04184660822958808,0.8405510768811394,0.26323253751046616,
 0.3018548794413626,0.05330220274285569,0.7788549621250745,0.5822498967505626,
 0.5045149204699602,0.1628362885150204,0.8755787851473231,0.6821791240360644,
 0.7891593747214725,0.6700401116723381,0.9976976630558071,0.3349612584474235,
 0.6811961454051514,0.4435764082635443,0.537360754237901,0.168675705469026,
 0.5484809588220956,0.05045253157000584,0.22282241609873232,
 0.07666679851631386,0.5940823427344529,0.2081486731085822,0.8523105429735248,
 0.9023004665597798,0.320678322632174,0.6337803677249544,0.20436975355948905,
 0.25384733863764997,0.3063076924145306,0.6142819528271776,0.8358347233921546,
 0.16120602908755177,0.9579921758875466,0.7005321007085173,
 0.06290238366961631,0.5868293695218743,0.4207435871113415,
 0.25429156387546836,0.8532641870262127,0.7903661448539075,0.2275604523313307,
 0.05412849339971548,0.4384282689083936,0.8975968690336108,
 0.46727189638573086,0.63951020293031,0.1807612565289778,0.16306276318196122,
 0.10886081603150688,0.26859841398732665,0.571717137662352,0.8289001152139269,
 0.09384400910912494,0.32315584285022514,0.6163441297066082,
 0.15903754926576663,0.11065954363724084,0.9987867272925356,
 0.8872326415854435,0.6556698622677516,0.39054856723271314,
 0.06585951940003387,0.9932527994182425,0.25281773686326714,
 0.6255055525152502,0.5239306620688803,0.4696242893751186,0.7607875814594784,
 0.5221858051965275,0.18059173229598446,0.6653882946417056,
 0.48452491937262077,0.5079790643909917,0.5376987926461982,
 0.03807285554345785,0.9787022299411622,0.9952047943298165,
 0.05653375910618719,0.42944866835816053,0.67947725977318,0.23986664467908048,
 0.36584613968422364,0.8503530056212922,0.1242966463231634,0.1059736416657383,
 0.46360523284776334,0.04219763634953133,0.46785562760993193,
 0.48073186713906246,0.14350850082686484,0.369455242194099,0.5910332164505556,
 0.9952578842950708,0.2820935681006589,0.018661751775829072,0.855636931111603,
 0.06698396035469711,0.26172455542999695,0.6237572994459942,
 0.4780650350008221,0.8786509871074569,0.0294828073590665,0.9663150823878816,
 0.6662467925138441,0.8214559888989013,0.8522269374837784,0.08212698928216067,
 0.08113133783038806,0.5552546225012192,0.2668993027446793,0.8135395155887202,
 0.7651127600492246,0.5966656357965975,0.6133762240391064,0.03762135437644354,
 0.47253751835853675,0.8863232085298317,0.14729563022821546,
 0.44624323046982284,0.5545159381512254,0.6666450129534374,0.4681871663214904,
 0.14875053791471682,0.8745655898931829,0.9360225759367635,
 0.05737260098414321,0.04435266911707536,0.4692440216473158,
 0.23361180225680744,0.5042245371013867,0.9532988249387543,0.7175912830443524,
 0.5014026434349765,0.3180447051559382,0.18138082379130893,
 0.23615439893126544,0.37922197646181144,0.1862039960545776,
 0.37858706480606863,0.21253970564452673,0.3438729578075663,
 0.9878779954288214,0.39195627055534465,0.06719725402194399,
 0.21407370641722878,0.7625196675269115,0.2551216660009915,0.409977645900798,
 0.3005336782167407,0.37592601597085706,0.8443893213371134,0.660708279124827,
 0.29512265000158,0.19112158506079324,0.19863259363377184,0.4302252425375591,
 0.734844952283505,0.3198241302136563,0.2788055122467249,0.16629778704346332,
 0.37218650585061774,0.17174842383559974,0.09831232455271433,
 0.3094959654216596,0.5722673966481233,0.5196389751811189,0.8831052449732911,
 0.23019318582532555,0.0025317820733279373,0.537314323346623,
 0.2955293986970927,0.22126488805216693,0.41468643567318086,
 0.06258310629921238,0.7838439921600091,0.42401079490634774,
 0.8646860465464281,0.8669084924386972,0.0042696056640310065,
 0.07629727541868403,0.07872604834997732,0.737244672222106,0.5929226730149817,
 0.9104967994729645,0.35917261303853554,0.3966364866818214,0.8012725143466243,
 0.3440735723059126,0.05282198506738689,0.4723595279625791,
 0.34150452966024414,0.23830948940665586,0.9728491516212283,
 0.31787976439020915,0.8578971144399601,0.9511630796483945,0.7628475411879005,
 0.4875326088934555,0.5102547998163616,0.09838743790608084,
 0.17907064070455714,0.725188016996502,0.029169039061969215,
 0.21790323484905239,0.8563631760708716,0.35715461682239935,
 0.2923158116900906,0.7184885079772747,0.0811842852522402,0.4738809468085817,
 0.6121679550350362,0.840475523709595,0.007430420260902015,
 0.18476898683511755,0.017405469707028165,0.41805517412895243,
 0.4329091399133055,0.31475791510022255,0.5202384003786386,0.7763416281061668,
 0.630686825785069,0.4253128226360229,0.7569328708172343,0.6693725436563102,
 0.5188816598413676,0.05966528486214118,0.1765790290315319,0.0064059010625378,
 0.9224831876713069,0.4495001408772351,0.9637138950605969,0.8190623438129216,
 0.6866957911031326,0.8381895617205916,0.9533451307566895,0.23756278532849817,
 0.36730940731761463,0.32787783537039394,0.9235175356470393,
 0.6667908309741485,0.3292454405539915,0.9759566347688293,0.5050450207944067,
 0.8869368102985873,0.18494437468775693,0.5964992017090569,0.8727431840537012,
 0.7966702136250661,0.378200765811458,0.8718809463159084,0.0950439196777727,
 0.5061983216946396,0.8031928985436323,0.5589336826572929,0.6712999050840016,
 0.34609341712037145,0.31124187006402415,0.8888152931354935,
 0.4002820751284587,0.8413156837600327,0.6496333454741385,0.9363406671653162,
 0.46349883536005154,0.5161440029501453,0.9713157213069016,0.2846092235606601,
 0.5343002606190214,0.2155066979772784,0.985177299319904,0.04293701385250848,
 0.2592200362181082,0.9265176749509096,0.30097646676872225,0.866215645372198,
 0.799295392801155,0.21375653848990628,0.9944598662120897,0.6880942511733142,
 0.3317101622922787,0.9442720218531703,0.6127089839971278,0.5717706598348661,
 0.5543115583557209,0.8573451355743621,0.15501264246014235,
 0.21476579756466652,0.04016191039017203,0.6854140093843798,
 0.07208645204683206,0.9238086181617998,0.7690208419324048,0.3950415496016805,
 0.3439009273607585,0.6842935813768226,0.7331157566770087,0.2229635658274568,
 0.11859117361387794,0.8101798762777497,0.03535538029388374,
 0.33639962684775426,0.5409012180761685,0.48516505398066445,
 0.07492935940074663,0.6873268045844952,0.20384530604645923,
 0.33106557702240635,0.9775932697455705,0.964420561820525,0.10308054058284977,
 0.89778888615869,0.5799467349172853,0.45464360620565647,0.16853418665835895,
 0.9806772439281827,0.454801093407149,0.6426471143745377,0.3125280564229944,
 0.4021096312573692,0.00813874283577909,0.0747986937745807,0.9341312233026646,
 0.16191654255039567,0.127380759763841,0.6426759733537654,0.5176337207360036,
 0.8277655387295775,0.7912953302511744,0.8335975134702606,0.9065297473166868,
 0.30160694891106776,0.9521106661696928,0.5332275449098572,0.7811915610512843,
 0.9950999150837045,0.49900378783615895,0.21601198983326464,
 0.46721191495330994,0.6526167230295012,0.9681183639534954,0.2587377559859718,
 0.4499166830612349,0.8769578375940152,0.44225062389434733,0.6160192253760787,
 0.7342114759950287,0.19829789729544678,0.02125805240031564,
 0.5450507097121222,0.9582543832560376,0.4268890878370126,0.1550936895617292,
 0.3533479020499879,0.9642169607457906,0.42837150452652084,0.5536008827744415,
 0.38098214823110177,0.5744694189632917,0.13645082212082038,
 0.04237665440401872,0.7590055422452839,0.4245889724440125,0.5587608273978373,
 0.7290236738544043,0.6911114226356149,0.31090016559765954,0.3291727817005279,
 0.8117539830699375,0.8835517868706337,0.03200643508110523,0.8380887209539972,
 0.17940168911128773,0.917645384849166,0.26827908311665294,0.6622393155421868,
 0.1866002818691027,0.23648882211434052,0.05836040786929697,
 0.7431321418663971,0.4302320043662372,0.41914818470318205,0.6936344638822616,
 0.9714879150511598,0.666748603281979,0.35020005462238524,0.5067625324415179,
 0.06382252279041933,0.9385659081905268,0.7021594074902335,0.5765530083058674,
 0.6737970555429098,0.009723557527560156,0.5431870135274166,
 0.7289541025203314,0.9440802362526035,0.4558121036621313,0.0894648312976376,
 0.5253880736857328,0.8944441764294008,0.06917833222842418,
 0.08347347518226389,0.4870178417769484,0.5804554573208807,0.7050846727369746,
 0.9541757280800205,0.7670584534316046,0.614971221467791,0.3916557945496094,
 0.32800959068948077,0.17329235816599853,0.06572880802406245,
 0.2670523988413419,0.15297402407531135,0.49714345730259435,
 0.8515176591249991,0.13138794446608393,0.6781219974844575,0.5494618151586041,
 0.7026675763578736,0.9064224990536411,0.834434994527828,0.11205458298275373,
 0.14255262605960195,0.7033008295011883,0.2950593694983673,0.0593736211922441,
 0.0019317797271788395,0.7747445903631406,0.9433566239769356,
 0.07845069178370967,0.6939192439566884,0.9260900794016592,
 0.29028710849706707,0.43733073722807525,0.5305436192477354,
 0.1379716649722691,0.3168867637593956,0.8135016493919071,0.4634488254527743,
 0.417839005221337,0.612679094297107,0.31431514281101935,0.13668825575693455,
 0.3345841484532288,0.6370505611414967,0.9693138517233395,0.7988994726975811,
 0.2337815070300533,0.6057739881220554,0.8411469458410178,0.45579788884870154,
 0.37835005078466455,0.13492364261500311,0.5607572954400686,0.373452427808347,
 0.8033798513008767,0.6260513754306467,0.4100690296532641,0.36223946564935583,
 0.5711190256887916,0.01886461516255633,0.6082832674419898,
 0.16294861780627845,0.9195830787799786,0.6360894336392714,0.8060814035215732,
 0.7114378787144244,0.17245245250398755,0.7271444949258455,
 0.029902663418897024,0.7351579163098181,0.42594147439860985,
 0.04434754987410716,0.8301666342080369,0.1619026743209414,
 0.34212083522359826,0.3688846074475658,0.7863492601797359,
 0.12497743848674914,0.3076815757894038,0.8247115318020655,0.7007379142666965,
 0.4803571158796829,0.796386762159005,0.06381079574852555,0.3656823235132314,
 0.6355305998086941,0.015843069695752443,0.3531395040673141,
 0.5060434142692105,0.34086592421714923,0.7794000204867879,0.9221439317068033,
 0.6663993294040136,0.6736391437049529,0.06275127213442222,
 0.03774592431347512,0.5944930988239931,0.7935111580595647,
 0.16512239126861084,0.1554424283309075,0.14975611812658374,
 0.9399977760429976,0.7732275508294355,0.014204158259404132,0.39081047363983,
 0.22986303807221797,0.1279025285418285,0.28639393776671307,
 0.9781823392422879,0.9928020367419498,0.6420644008637394,0.45157731901093323,
 0.847241473318857,0.5549577206429519,0.7651966140759862,0.744911098275989,
 0.8984062046614086,0.3005891718459561,0.5742535877584121,0.7914114836818783,
 0.8181804714358513,0.35578600406819727,0.32480600561858297,
 0.24202370501006654,0.11875802074406316,0.6373499356672454,
 0.7236839916834885,0.19762549702313237,0.60230459720769,0.9984321350774673,
 0.6708358403970489,0.6727251022636653,0.8775108167329605,0.8159157441126134,
 0.2352647013402116,0.04185148603729394,0.12919071537676585,
 0.38915710736309295,0.6696962939139826,0.6836258048390016,
 0.09517459973097309,0.14423680673041284,0.46839376389141485,
 0.9202435312420388,0.9723647032095531,0.1339529424438215,0.16768418439530686,
 0.8333441878993771,0.017896878886377476,0.448782808555659,0.8355736116248991,
 0.1924901731863996,0.41866550157575766,0.7460876134032647,0.3457781302137476,
 0.13033112111039225,0.9671467733360757,0.6961413527956327,
 0.31236206520093046,0.5976390897588325,0.28718731149530685,0.199750934921464,
 0.8202699086375183,0.1478338172092446,0.6417442704925118,
 0.013472157324464606,0.49152365765350703,0.547331948638182,
 0.061301594344283394,0.46583290689207035,0.2510407499227637,
 0.47624024480169513,0.4434845591402353,0.2663431532257948,0.4433768917285692,
 0.22656686069111287,0.18498017273178557,0.13363153014430418,
 0.8341043905213248,0.6716674124817675,0.09747378316416089,0.534537617076956,
 0.8649562934850488,0.7167444187313752,0.31919797763967095,0.6645664422543603,
 0.6739564456861633,0.14170430260203604,0.370475715769066,0.6998445652575067,
 0.4418025962726322,0.2544571485150502,0.009863802999589777,
 0.22494795438926318,0.5518920437190653,0.02184126509897588,
 0.7992339352911807,0.6985380952915314,0.9421226739642208,0.20114371591599745,
 0.9586855571809645,0.11069830927320112,0.49052717070388796,
 0.5523595286335123,0.20642226230112015,0.7414723329746906,
 0.31384156006613206,0.39223427606361894,0.7537295864953087,0.387990380346818,
 0.8829960974750712,0.7664510386323817,0.9315293737274053,0.2805038779675999,
 0.4747172669393548,0.8047798413720233,0.058453379261145466,
 0.5197857008701031,0.012698057960440767,0.46538877769230313,
 0.8365663789967441,0.1858025301453473,0.6744406280343348,0.37513364431737417,
 0.6982305927954251,0.17788276377631962,0.06848626606436137,
 0.9633318010652925,0.9727045768220378,0.9862855731588258,0.5217899501410105,
 0.992987102992518,0.2516019957816634,0.8709321431926331,0.012563587282812816,
 0.20403814886237637,0.4191762912858841,0.8143441933331063,0.9140247600343324,
 0.5371514157693466,0.27649582980979515,0.17824186564820987,
 0.9593582051694404,0.908055747742188,0.14689228881383143,0.39598879121985375,
 0.8923638209353768,0.02919336424662211,0.598928371965748,
 0.009727070269395566,0.7468214019533257,0.14051839119446874,
 0.07440868881129736,0.3402258648438402,0.032272644472954304,
 0.7577305928731463,0.3690390618597227,0.7807249667057145,0.23680477666199962,
 0.9291573087561257,0.727679785477658,0.33981831492199144,0.08930399972981928,
 0.7408089184986978,0.024363662240103445,0.8938065285142657,
 0.6477429085184602,0.7137963661083059,0.18104500246161626,0.7104363476219014,
 0.6755965043267614,0.21872077889997965,0.7694636225943637,0.8187832795504959,
 0.37508278119299077,0.7222718062956555,0.7320016306608677,
 0.19435758072664866,0.8148677238253562,0.10267397596420713,
 0.9812429250494384,0.03815468366862873,0.15099323392267894,0.702764126181146,
 0.5803580286954919,0.01322803776642345,0.6047801063929774,0.6011464340585486,
 0.5848711226782362,0.6944660949580043,0.40866893223766154,
 0.34328219873162424,0.4533840697959207,0.6688082262598111,
 0.09645155818575857,0.27918962550663873,0.9232439226564576,
 0.6362773292379038,0.1050944351113483,0.9106444981412333,0.4476942834132448,
 0.18342924601664323,0.6752890289000771,0.6254487727749758,0.9628730635532597,
 0.6993369163515396,0.5312191293013919,0.4171304276878405,0.1297251800125453,
 0.9456147614350896,0.417062501291218,0.8308081077170548,0.0552427288742936,
 0.7490041302559148,0.6153981562518177,0.06227647843947581,0.8507990361628095,
 0.457657894299181,0.7214053622417114,0.5868543792228723,0.6249135383156179,
 0.976980038578732,0.23565304282385702,0.5673833237829391,0.17268356141841723,
 0.36757366447198314,0.7511182499310332,0.801331262875047,0.7494221709443122,
 0.771818438097586,0.1258539123914867,0.20600987713480323,0.05235679916376046,
 0.5827540018401861,0.3493749951857018,0.24978554088065674,0.2321425227791053,
 0.7538106753558419,0.7172910756496972,0.6838082701419456,0.7003865697771167,
 0.1331421004203026,0.2344097382201692,0.9057408787066511,0.17358785746781735,
 0.46140784844079963,0.1832043127504166,0.5664195496937231,0.6573711375186216,
 0.016017064558668248,0.42019737100306087,0.8425434718077687,
 0.2120777667262166,0.3822226220777327,0.6859721259742855,0.365155591004239,
 0.8026109432555412,0.3101930481776325,0.3325353441924591,0.9548810560336956,
 0.3374263953987181,0.15574373816333698,0.358839442957707,0.05787461536678684,
 0.29553643673759644,0.9097005886509395,0.9746262976475594,0.988987373152523,
 0.96225523756079,0.6164747622464648,0.7773490069415052,0.13226417466898865,
 0.24483775372073624,0.49397488891894203,0.4670438604761815,0.912642082269609,
 0.3973217698020577,0.10920047932265764,0.7922484851304238,0.7149978448954043,
 0.06256943474106025,0.7072182149036024,0.6745970879545853,0.7725328351040813,
 0.2584351943813863,0.8330085830375447,0.7045633427293261,0.7330491883270238,
 0.5688702553017896,0.5438000078719261,0.7702790314906103,0.2337656602946201,
 0.3745508614062566,0.410050791861993,0.44081165985204107,0.754042634248649,
 0.02317244682423525,0.8884255148922153,0.8523315361355747,0.0886148890395757,
 0.31068235140398914,0.33805631758008137,0.5074753369918524,
 0.8459222029129345,0.6539099167944333,0.32087198549910023,0.7065621266104343,
 0.6494549338739453,0.4284363258547834,0.46825096672758504,
 0.35968883215758785,0.5797970559501734,0.8654681277388038,
 0.09757136057946358,0.1292522568608031,0.4198245138117014,0.5446911038321516,
 0.5405786116669156,0.14910300200545523,0.8187351902499818,0.9783067123475245,
 0.6960931970536085,0.5948998438998887,0.11757016307443302,
 0.23764194720367382,0.2417894562572357,0.5532395495847066,0.7941721211933769,
 0.1774368947088727,0.8743554103820003,0.9305044327586919,
 0.060107426896005234,0.09385274133052679,0.0037150731605566367,
 0.7375788669969646,0.5489141865221361,0.8522716684367466,0.1432034046367321,
 0.7914750408683415,0.46105584969747393,0.6660278547219635,
 0.19957200868551017,0.008573572841602872,0.4292820326321658,
 0.861220669852927,0.3551426551809026,0.8223199048699544,0.882157036735264,
 0.4300503471907706,0.9655386564711566,0.034801136142875055,
 0.5808492411152709,0.8108401126156339,0.7434262282987478,0.8435310205130785,
 0.33853834308103414,0.07511905727761747,0.8987011538815185,
 0.044837157286850116,0.8230230840709217,0.2092160582436321,
 0.22141493968876447,0.6623657320589506,0.31477072894650715,
 0.8364622301326095,0.6367448665937098,0.47228111010674356,0.9469303811888334,
 0.286729978303792,0.11357748278553037,0.22444965145802265,0.1103014987489982,
 0.1872070542725215,0.01139018142151449,0.010076748868404817,
 0.9543523844784605,0.4710165611513457,0.7274363264766854,0.970531878274001,
 0.7661335333382157,0.08256952359155056,0.6430543653401091,
 0.08643578771629523,0.9061568303103972,0.012238055400827785,
 0.8836083010555423,0.8602035817994779,0.5758318604538776,0.366233169175756,
 0.10514843783018768,0.023424384696945633,0.018543169507277635,
 0.4043570947988909,0.026858020562438134,0.6909950357144113,
 0.7726397671900022,0.9902264333450642,0.36062358047399046,0.8830422960137081,
 0.46313837803353364,0.5090137748841774,0.19870374731575424,
 0.4432703921336476,0.5916944053878856,0.7703767754964042,0.20444156791176316,
 0.23116744526657618,0.037010581822298216,0.7327608223672543,
 0.5519124162102792,0.5028671374456551,0.11076541453589483,
 0.19805932298220175,0.3084912438203957,0.8147293765719952,
 0.42466685420477457,0.24744008663532124,0.9245539491860009,
 0.5598498053947739,0.7448009943338133,0.2971218726191177,0.5732401248409696,
 0.7354534389499301,0.6940787062519832,0.22388441609280174,
 0.40738897840042243,0.2986326640334742,0.9969812285401031,0.9477257612289681,
 0.7995635201299948,0.09418386378759092,0.027796445406890302,
 0.7567374266411648,0.5577078644682584,0.661391127356128,0.5381481643334419,
 0.8435098695192184,0.6758676042010845,0.022548008602437664,
 0.7150264903113881,0.638292862081804,0.5872367747303151,0.6078083563831508,
 0.2061281402223829,0.1661095868722603,0.9147638031692942,0.49766204948558657,
 0.5209713466554604,0.3516197139909807,0.0905186039262863,0.5666762426478532,
 0.48147679400123633,0.2873081095129667,0.216504262891292,0.3366683699127664,
 0.062006998370435995,0.32325006046275906,0.5208032621514713,
 0.5402925046575267,0.7730020485301617,0.16273703657532035,0.8780777866671873,
 0.5445377436897516,0.5628622419587107,0.895360321394571,0.5564785500973294,
 0.19117941631402513,0.4738753477851292,0.03927601017908777,
 0.9678742895895108,0.7626892344644894,0.23574358245672544,0.6043540158303565,
 0.9710779640086559,0.871816536622067,0.36478898694651496,0.31720942988263345,
 0.17683451101041892,0.527216845934433,0.7756560328627838,0.19444058121101682,
 0.6141969319421081,0.7025151837075657,0.7297422810530858,0.9390225003493122,
 0.23701073732228417,0.4157542354972332,0.835461483873455,0.5435117502224055,
 0.25409271043417114,0.006745589760403958,0.9262778529790526,
 0.09208063745640882,0.3498383879841095,0.07331722268181706,
 0.7667033388121737,0.011611276854314734,0.8199001132980526,0.839124609240259,
 0.8445273322432931,0.9090565483627788,0.2963435870735154,0.7605742758128867,
 0.857531312539793,0.44960349253029896,0.6922187900520681,0.8946897320874252,
 6.301194565826451e-4,0.6134748016551116,0.981786389314603,0.5612009242528809,
 0.5722076919831467,0.20718092819411493,0.4361356409555842,0.9549659057401546,
 0.03896395734296543,0.36205098296879257,0.04372986008346402,
 0.6262631722993391,0.2119357232257112,0.18290103510499955,0.2996563280179091,
 0.8262064794620676,0.803239201656979,0.23222638424018371,0.5767152642317017,
 0.6689159059710232,0.9151124037817171,0.8364871359179179,0.4678218568306344,
 0.9181068311165882,0.22855023691309406,0.7091155209016585,0.8792749799559729,
 0.9154655396091583,0.5158691723303181,0.5346105019727576,0.10312103936124783,
 0.9444624676872316,0.5281629473357274,0.014318709993317635,
 0.6230295844305638,0.3210155147971828,0.43782191433370277,0.864173295051267,
 0.46517668543417134,0.9152779333082542,0.9752955905397935,0.8388913751133356,
 0.7654298669167772,0.16619970003234463,0.16108976272601716,
 0.9185716763999511,0.9542802293882902,0.21776305832875642,0.1902737883060266,
 0.16919432837196724,0.2705409099364874,0.8511547875562097,0.5532411840427218,
 0.07381733616419628,0.0017339068833855942,0.47803727880766844,
 0.9025070151758359,0.6278966649158456,0.16445038481207053,0.7410492955860049,
 0.5511914093360434,0.328055878363561,0.5881779919288713,0.07370765100251009,
 0.36614268929547245,0.6311189264561073,0.1533759901877214,0.312775139150433,
 0.7703098689340091,0.461444530657526,0.44325942461353063,0.5219086100249524,
 0.9674893715984767,0.3017345508565966,0.5904759926152902,0.24977864973556874,
 0.9332937017480177,0.7685169215686538,0.5138820552950154,0.17860269184479216,
 0.46078541764014225,0.9010025186429438,0.7234330676201509,0.1744888837012738,
 0.22156490105834303,0.4022014473456441,0.753486151172776,0.31588115431879105,
 0.09773618244604687,0.5080677571503774,0.3608257617377586,0.7102419067254069,
 0.515252314266907,0.44231346474222144,0.26980552564306926,
 0.34393111866763626,0.21955013818539437,0.7932538868031458,
 0.04060772608197938,0.2480198443410846,0.8133113679273137,0.0450674376312753,
 0.9058256179945021,0.3846029019629307,0.10729479895295024,0.5897799844646856,
 0.9295413729229278,0.9089660448453076,0.3782424178571082,0.7119137344887762,
 0.9949894065075194,0.5473598608325677,0.4734300383649518,0.9601230480585273,
 0.40760434804244716,0.5549705921818998,0.969864452994765,0.7320180816821699,
 0.4918719064543011,0.3256237396812176,0.664247628067261,0.38980489987585654,
 0.6669607587416819,0.7621384918158906,0.34195071605172245,0.8772176742545272,
 0.3990528362379583,0.15982767549939225,0.26869130300857913,
 0.9564835998799885,0.8710345994854954,0.8905868775015032,0.7562586284054378,
 0.25413034003553003,0.7622121633764223,0.5584767648177305,
 0.15689472652115533,0.8309257296705462,0.7913848562503132,0.9262847088249598,
 0.654397823449863,0.8477329813778558,0.05751575415280896,0.1682885712712423,
 0.8184513664139267,0.7782293173649532,0.7217436939951588,0.7208241146483008,
 0.7532486616189286,0.8397099153659259,0.9914403165354461,0.3212042710332639,
 0.7196886396463252,0.7194643343436223,0.6669017868248397,0.15844140955590424,
 0.5852463167904497,0.9256674419457331,0.28383978005147126,
 0.22714905669809804,0.5417979769845093,0.9400946986450397,0.5799537546345803,
 0.5830749192420025,0.5529569315383818,0.4528767717772927,0.5980067899725979,
 0.9002188911299245,0.4031893448495494,0.7370818222614473,0.2473615826118012,
 0.9071732806985983,0.4486972706200266,0.8605328953960301,0.25160164017110576,
 0.3739928941987769,0.32210417294039384,0.39798946639211463,
 0.42992857245786875,0.08625238308161753,0.08837644559991553,
 0.4879416133010549,0.7391378777440254,0.08866576068641763,0.7744249209973877,
 0.09800195590314276,0.3978320657478649,0.4432254054612573,0.6100730638026143,
 0.4344702901173383,0.5972126251811083,0.44659190618700995,0.8659585819531997,
 0.5405576397915036,0.11831393556517977,0.7908222436145942,0.9625354910485795,
 0.17729734243525042,0.5123258932369082,0.9889746456054384,
 0.051590129286914976,0.8630866546458816,0.9378489373943737,
 0.03972899680929365,0.31544200903677877,0.3461489156340595,
 0.01932482266433022,0.07917109421220758,0.27122932278342593,
 0.6979101891437649,0.9602759586667701,0.6724216445681027,0.9031214190499213,
 0.6571131539403452,0.6942888548022385,0.26409971738074955,
 0.21243397169129163,0.10159728994773842,0.6772454595642976,
 0.7363210787659762,0.43104543650213056,0.4348471083394603,0.8134232151759271,
 0.10159872495592148,0.8820134926661288,0.39309363238978734,
 0.7834873744544312,0.27379232327988845,0.5968333565885747,0.1505090424012756,
 0.08704642030131293,0.643772060826789,0.09237418836522038,
 0.27978750705124356,0.9002281950108362,0.8044887463991787,0.512592966691644,
 0.3402742765277955,0.38793103293784226,0.1074050386802079,
 0.35597478494374213,0.03655581431616428,0.4299226462257827,
 0.9864241663196225,0.5631884987121938,0.41340880499623134,0.9875902132230052,
 0.9992892294993814,0.865255457593832,0.7237549284699971,0.9855748714203607,
 0.7634712957487434,0.36235002640854563,0.4444498869959784,0.6246874799695054,
 0.41246206271913355,0.14386069983387617,0.5061820887310762,
 0.7140050754001882,0.8584189449160433,0.8941442693809363,0.334274455143032,
 0.8842338811049727,0.2171612465547299,0.8644782872831661,0.15874218893521164,
 0.3789277181216443,0.7744695173950268,0.4838908729716638,0.5478443346905588,
 0.8547377777758879,0.09890405020393622,0.5528666435229096,0.2075760320068839,
 0.20012965162414198,0.8032207448563133,0.7553472149705249,
 0.42727356857031484,0.30161098793758534,0.9891390063101619,
 0.5563725389024737,0.10034542059629692,0.5945156391501818,
 0.19793060209078683,0.9780725818858345,0.7799284802623558,0.883160292488817,
 0.4897582540086487,0.5746081759351334,0.19235612309070504,
 0.45361137657139194,0.04414569719083228,0.26015039678081675,
 0.19280365541229394,0.06680867941659097,0.8796847324703139,
 0.5486366742169382,0.9329801645044344,0.06459939509450885,0.4828805326618496,
 0.532849381435518,0.5864525245316459,0.06809827611165575,0.5725713761842915,
 0.5531233482169629,0.3564374658199546,0.14314456367753536,0.6633734480692357,
 0.9325942126415424,0.2843771935233238,0.50389816213381,0.04492429875686599,
 0.46172237502017743,0.7594517859203888,0.8632446684178862,0.8030441501812233,
 0.02605842844760531,0.8224122964443077,0.04232722061267258,
 0.4103081261296331,0.5616051895685285,0.8881957967226505,0.30360659930878464,
 0.7112634406904353,0.15717001417480425,0.17405586130304584,
 0.25071189498501245,0.7539005508268424,0.10511215144757946,
 0.02082740510892267,0.14732438598772346,0.9419160261072413,
 0.7726025880185727,0.7587879372603027,0.8309544655006421,0.5771466807239188,
 0.7180351827808813,0.05772091858915229,0.4037438973908991,0.4888786342991104,
 0.393876487112336,0.671848339707739,0.802366933452713,0.7546117643097844,
 0.40706683806516675,0.5225803102432568,0.7143345074098626,
 0.19354736984539755,0.17131975045155212,0.5915627810332769,
 0.020980460425833236,0.020178505584086404,0.7328618890772834,
 0.9822149131882779,0.7567351470938113,0.10478822873321647,
 0.27211183742123846,0.8403429088757892,0.07904577166933902,
 0.9685650560651469,0.13227585622224292,0.30312228342365255,
 0.9873156477992567,0.2625380092460792,0.9434661962635639,0.3240140922307355,
 0.22572451403130855,0.9204891128671018,0.3281273862242302,0.2035484376419584,
 0.6934972754629852,0.8209619311108992,0.02965748913906463,
 0.23591848701938523,0.3393687129372167,0.5611583692371832,0.3583820681462737,
 0.25096295620661924,0.34009999978785865,0.7726645789228005,
 0.2530341632008647,0.03493407102277457,0.9788315564291239,
 0.18265950682821497,0.30231260690226314,0.3230078032234742,
 0.4986295799144097,0.8994972842104254,0.33513217990991506,
 0.29951885595596306,0.421197805535127,0.004015029582427054,
 0.34863834628260615,0.9955402105508482,0.8603834498152471,0.7969659279064913,
 0.48986787243951746,0.3835570183394109,0.29814134757338473,
 0.22539401114290802,0.509115837982158,0.7736646331891445,0.5714573551764971,
 0.17893956246323162,0.007769621717369057,0.37053139735443663,
 0.33841970433968005,0.7005525149002514,0.7329640204746268,0.7564637900109862,
 0.9510527064699272,0.15771494126118146,0.5335919451725588,0.5398950656426401,
 0.3168648701026159,0.07169578222106854,0.4872519716974627,
 0.11108482245824358,0.9779540766614561,0.09608289220917177,
 0.40065449083624305,0.6443967860864537,0.7178066829640606,0.9086800235902251,
 0.017054080489245727,0.7945900975784884,0.4506143795348624,
 0.5178748685789565,0.36949503161681363,0.0015779758543064126,
 0.5857109685336503,0.4855535529739251,0.019647896695944356,
 0.9546387181221329,0.06032473859950849,0.7700324819149779,
 0.47961311169186605,0.9513242670260792,0.6684215692912322,
 0.012504980404573374,0.7390267982821096,0.47833669534621337,
 0.10923131523079788,0.6134298530684228,0.382681929033731,0.7600468619858878,
 0.7742625387296999,0.39336577112357074,0.10232546219336791,
 0.09919371411575928,0.4200193301412136,0.2881744359247237,0.8900236523758934,
 0.24851469151090722,0.5331457324746037,0.2679280172248215,0.9718728136831829,
 0.13101829655744057,0.8430582916243133,0.48531495437757655,
 0.26041946534778915,0.5347161761864188,0.35030937453205957,
 0.8073959699447044,0.449385822023913,0.16832555952127803,0.5495870334221723,
 0.5047356737468891,0.5778953013011106,0.19668562755009922,0.8767558314246481,
 0.19886253052061986,0.22323534254656607,0.06512391633110093,
 0.23512713966607957,0.9826339276051866,0.9070757138882263,0.4224020695172509,
 0.30667669691347854,0.2396170983853002,0.4294361634124031,0.8144416650911732,
 0.8938415623182563,0.821524028220172,0.45173770635903276,0.2888080733172086,
 0.8701493009343341,0.3172682761987815,0.8542786880330535,0.13564296324528669,
 0.2222517463507231,0.06842787765349367,0.4012106454294906,0.3259958540907121,
 0.7031948373641392,0.11207247495461747,0.31849995733848446,
 0.0978203373788703,0.7970037150840703,0.03619100946362597,0.5219596956006132,
 0.2658925875315308,0.8548170112269351,0.6960008496541911,0.5691099913495852,
 0.7889350921332198,0.8643794934605783,0.9074632140514078,0.18004058015527358,
 0.7471892301539526,0.1287439239770829,0.04114743117635111,
 0.14293797600796776,0.17015732256116134,0.4694997367300262,
 0.9698098847884409,0.022444704917459535,0.13965119484676558,
 0.342310329809837,0.37070972042837513,0.7873051384497073,0.623305986834392,
 0.3572103579597915,0.234521712616243,0.5047970472489737,0.5691075947061042,
 0.981771928805,0.8880164551789392,0.8741873823933782,0.004279506541730926,
 0.6237485145830108,0.247075859632239,0.9539863577913774,0.7611088219702385,
 0.4681177241578608,0.6216252493222383,0.9025874972664195,0.35938670122470884,
 0.3877276987953534,0.266265985635333,0.36048194392545074,0.31283561950578964,
 0.4769916877458069,0.5896536869324656,0.6180023987372922,0.6373667411793867,
 0.21257671227852826,0.3000287246025114,0.33945459083367435,
 0.01357375652148829,0.6488566641580056,0.10408152059097908,
 0.18972074820552165,0.13194320068931598,0.8084224970398243,0.630473477856532,
 0.5637603134498689,0.5088227578742543,0.354531016400238,0.06490277232494668,
 0.06198538307634638,0.9334748454077769,0.835840595507032,0.1817425882242969,
 0.2921918639818799,0.008841723567488469,0.2679116101892389,0.27039674925736,
 0.9890778035826148,0.6855633182783005,0.32462552350300244,
 0.30041391933523276,0.6767318388895934,0.3315270847641444,
 0.26935038214444906,0.6204984194033664,0.321965779305172,0.5723188887755514,
 0.20790054032488892,0.8777147120309631,0.6999138349806042,0.7232249395677941,
 0.7184403549101092,0.5041540643700694,0.018992133786068233,
 0.4149148939063193,0.7316816943472553,0.4862965562723873,0.5940592161832746,
 0.39692171846646085,0.509851641910303,0.90489558426739,0.16668547985406823,
 0.7452377335757896,0.24067541816954796,0.49774351587048504,
 0.8604986572928341,0.5883836494036191,0.5435585777489826,0.8842321081596135,
 0.03251530953944892,0.23327380810194553,0.13889813131206963,
 0.9663479497555721,0.16890185664138646,0.5970482400024351,0.7200961598496047,
 0.2914406956154838,0.506223894948857,0.4824637685246569,0.9665018649367663,
 0.8916821176090347,0.9661226688947239,0.4590797261207653,
 0.004894328013144378,0.1463367899517929,0.5152185889007116,
 0.6272455183487909,0.4873785002427721,0.2302461215963294,0.22726621074892073,
 0.8822647144758662,0.8622033600079507,0.6639007364050113,
 0.002438857762814828,0.5455518947258149,0.22051245780288398,
 0.03359033403395928,0.6308195545051376,0.28709470007236526,0.928424599155583,
 0.6419997991274015,0.3148394659114444,0.7839082062350279,0.49554359391844993,
 0.9608172103600406,0.3457599921958401,0.6089718368017613,0.5848181970727769,
 0.24387153512619242,0.5372986938974045,0.6811231569470118,0.6901897963786472,
 0.4547669804815737,0.12880638924578736,0.18990716524557916,
 0.1485239452512369,0.9085161776847308,0.9186443193136341,0.5383233371785984,
 0.32195143417321814,0.9503061808559039,0.8902197352162106,
 0.20901827421018004,0.026764369150562017,0.2584319385242001,
 0.008227037763585465,0.23510480517482635,0.43251103424262505,
 0.5510859810952108,0.7950424822478084,0.7718827478952996,0.4955857874784373,
 0.6000902628874263,0.10847750729481476,0.6073809751111295,
 0.03815948611858133,0.8009040375601901,0.1908424057025757,
 0.21796700892967036,0.449908377106946,0.5116621780880823,0.37965119151296245,
 0.600605456736931,0.9953662371879529,0.2609325743832772,0.8478674004486907,
 0.6613177048757664,0.25731603802778313,0.8223652266451069,
 0.0049622479508768524,0.6957351418904792,0.7207823714909761,
 0.14484054390782053,0.7176151592782805,0.09272494823755373,
 0.8028586485300263,0.029469882012933657,0.16400391278823856,
 0.35821734782657355,0.09092854748938639,0.8616927500438221,
 0.6356606604275294,0.42448520524886735,0.6911051102536172,0.5994162390701587,
 0.10335849069856806,0.9649667638959206,0.20018305040603557,
 0.9063370340336606,0.2925454374323524,0.5156948350202772,0.5422039078104578,
 0.2558770776973015,0.8650135755504776,0.27084677878820607,0.8979173485303071,
 0.20533111481540534,0.49665446523491474,0.4720779550765185,
 0.15075421024515623,0.3472273280772449,0.9121011409704176,
 0.06359993973170941,0.023870359530822616,0.15035100744160568,
 0.7251812436922768,0.9094793769632677,0.519996484828778,0.7163113316948128,
 0.8674980469529604,0.880268324955725,0.8125651886782004,0.721247411712784,
 0.9605249417446027,0.03717183220011633,0.7366608251672637,
 0.37620499039089716,0.32540798952236494,0.601576035967556,0.8535765497533732,
 0.638823850499266,0.08391186162270448,0.9894309895706601,0.1534570826238406,
 0.3116667280474237,0.7368144025940082,0.45654703456780465,0.3648544781253579,
 0.640546567254,0.12975057235576515,0.1580500126827069,0.49212638019723254,
 0.8314610902060533,0.42763804199003785,0.8007053421821975,0.6668473051065815,
 0.8940717815058643,0.522730126567642,0.31247887469788393,0.1489519014699474,
 0.87225389401016,0.751454871323506,0.042727665263986636,0.28137347312451344,
 0.3072402342046152,0.3936250463818921,0.7558533472195849,0.0187379890075291,
 0.21069197581937238,0.3572431046166694,0.9500541375012768,0.8619197884246848,
 0.2085695285720387,0.09073062498302209,0.8678684670820164,0.5803935023289609,
 0.07107684763682298,0.5910777709251226,0.057526800329680605,
 0.7893565884692345,0.5840018154628817,0.7621826262659183,0.7455546956961169,
 0.8384310024644795,0.6945371875523063,0.40865657291258084,
 0.05603019888631955,0.3469070446738578,0.6527387821295789,
 0.11819426988653292,0.4965618525965002,0.134278607727356,0.5599741378461468,
 0.8638413420706916,0.7960738532999846,0.36484584289642563,
 0.15080617161766818,0.5781804274323966,0.3712835825478279,
 0.49623612672887285,0.3024791244353977,0.6318855825378501,0.9978558603008382,
 0.5002333779758713,0.30323892136309805,0.22515656694580122,
 0.11582373459861406,0.5430743483001901,0.7171939716156287,0.3499091163995889,
 0.60862161031734,0.026809165773082566,0.5405926794262927,0.3191716955326014,
 0.42055197418585066,0.12242273258391612,0.7747229200031738,
 0.2545959578836583,0.018533112213638403,0.9623482879533081,
 0.1462531977296576,0.6236060843151647,0.18220843907554374,
 0.25657006867558974,0.7361414854528873,0.7519404913260006,0.7171475656741257,
 0.724201271777541,0.4131632274633581,0.3410344628901919,0.023952852539497527,
 0.6835534184055078,0.5292037210139632,0.6271351057978537,0.3917528311978282,
 0.4363115290032946,0.7872744621045326,0.4547714297737897,0.4067741119436602,
 0.3349219949017359,0.8975155516730254,0.7192517759425631,0.5861480641735317,
 0.6316663753146955,0.4224057916177315,0.5516274602951567,0.3215797695130821,
 0.7980600628406064,0.4780917557834634,0.673025315485749,0.41522624440069555,
 0.9887215357192549,0.46255846561384983,0.6164810880313961,0.735198007807919,
 0.7098746991039695,0.4290837996102479,0.9693803647483432,0.0948257236650415,
 0.8035036165553996,0.38920801157479035,0.268047976786579,0.3190232343162245,
 0.4121202250853484,0.5119635813665963,0.7609919113350246,0.43202781188614436,
 0.8272800193540152,0.373851141551803,0.3446189406515765,0.47515585272715744,
 0.42335938020147557,0.13965813202446442,0.9241290216865435,
 0.8852638991790928,0.3007667333827847,0.6905881742671901,0.23244117617899474,
 0.6230220097534931,0.4267112099378233,0.2148001488927711,0.6373550594493509,
 0.011771429839607883,0.6006398782450764,0.45138059121517715,
 0.5393744763686911,0.3315860373993448,0.39052222950342363,
 0.06659793203190045,0.32849976198781017,0.002432835146887191,
 0.7952664271939267,0.27312801541143017,0.8252949186776397,0.935236478778956,
 0.6026079903017326,0.2993217143178075,0.980584189818399,0.027800885564303157,
 0.3824537057764208,0.8155547321147656,0.07107448882039724,0.6729524820664268,
 0.7885697806038192,0.49271357204422317,0.13918244523884837,
 0.47357402448149744,0.01805511429227924,0.3224830452491756,
 0.5908168961452046,0.554657750114043,0.7240483849255366,0.9418139376685012,
 0.22871853133371,0.27911042173031864,0.5541627039591978,0.31223467475657873,
 0.5590671166987465,0.8398841436184815,0.8631802286838244,0.8985189057549184,
 0.838246535655905,0.530816512394885,0.5197571773499958,0.8681551250962621,
 0.9402713553498129,0.4313161249227746,0.13299274856905952,0.5584552708375906,
 0.47185571641333346,0.1878508847926621,0.4856644358665302,0.8080534167700295,
 0.13966841888190284,0.7601433545167002,0.4474546200157501,
 0.18597784110230564,0.10013218232953902,0.07895978673678483,
 0.1754609129817959,0.9277952745997011,0.436093476894709,0.49150574256571355,
 0.14374802299091283,0.936561840551121,0.5717762669254363,0.5391894042628911,
 0.24565724591476268,0.07156872722748397,0.3268631596679684,
 0.24066043615320132,0.051042463940535354,0.010628639728933242,
 0.9493002336778813,0.9490236561238522,0.6659320220042807,0.7771843928297365,
 0.28544437144536616,0.18542088156279657,0.00906867454347371,
 0.7159994748886629,0.8087376554836094,0.8219330368570662,0.43550864917416054,
 0.11427930295565181,0.8602176168951635,0.3370594233566081,0.5802764815515974,
 0.6997284316178511,0.7006866046935556,0.13093108633446493,
 0.40598803361097513,0.49986932759449076,0.5256317694873497,
 0.8139825994851704,0.1086365105387117,0.6194246721803536,0.7241510436905993,
 0.5711488606680951,0.1284251839887186,0.7558859796155564,0.8154408454904933,
 0.27504970867975087,0.9328920559194165,0.10592318211509388,
 0.8252597543887896,0.003534865969541112,0.46019048774963145,
 0.6749531161863587,0.6405541419162031,0.36188982455872476,0.3000642164276299,
 0.0023460140340858526,0.4830176023971403,0.615904877050125,
 0.2937726587428344,0.2803382179120736,0.6949609754050442,0.20916197217293908,
 0.7034743403037093,0.8090303991100687,0.7637200089433577,0.3352709422825435,
 0.6802902631775274,0.5279767562629637,0.14372917352918724,0.2879280594899041,
 0.14361616192954707,0.7861908798398385,0.788315454881993,0.30982092414190765,
 0.6790031878936961,0.020489279673676752,0.3937616320775166,
 0.04157380425749213,0.7334253435146421,0.958333687833552,0.7175707895435369,
 0.38757398214972405,0.10784235767932748,0.6727760676646422,
 0.6009855399813038,0.8875645438394175,0.5720637326788348,0.30647357185453916,
 0.4346203625763906,0.8790414374606799,0.041045446435649335,
 0.42347319155183183,0.4237426179195647,0.3394444527002748,0.788494934137222,
 0.45711795812584655,0.17191755592526936,0.12364395890658775,
 0.9079163725841722,0.4211367238575182,0.7914600754233563,0.8203994652178872,
 0.30653587613338107,0.1971822721481279,0.16421802857898693,
 0.6148780498146468,0.24726405709966826,0.9914110224443086,
 0.25023646254182885,0.9842403504138322,0.8009930500594993,0.863706880701736,
 0.5930106754609027,0.3340057259100445,0.06808867095443805,
 0.07217584106398234,0.5606870962875208,0.4019726701467615,0.3394373428013443,
 0.29724255989329973,0.22556973595280239,0.8899902586723907,
 0.7447724219548861,0.02196588044841641,0.7069083148657037,0.5059369817332315,
 0.9088789664520334,0.8431819526023118,0.7334818669572815,0.5811387393747377,
 0.14935029284907708,0.013642057927585438,0.48932138475825415,
 0.6744565905227579,0.5536018829714815,0.15114722041023254,
 0.21391946664216333,0.5927803514358033,0.3579086997987221,
 0.18600907867715377,0.5353150927260669,0.3191736652106413,0.7464877157571573,
 0.8948243443159869,0.020840272076868005,0.10142096767014497,0.43110051912101,
 0.12793851474457463,0.8907837555144438,0.8584167872738522,0.1569582876755119,
 0.22968001865908572,0.5781846624915962,0.3096186819805621,
 0.11584628087247872,0.3455262730108969,0.9498020938149994,0.8007051703833804,
 0.4006419610774281,0.7972068005284614,0.13451639270161253,
 0.33093648470228476,0.8989616361247842,0.15640123246245352,
 0.6752377665866053,0.6117934728289298,0.6557956932061474,0.34087019263937357,
 0.7630211615860146,0.04611029324472127,0.027535758007703692,
 0.054447822573605476,0.2849080303673053,0.11515529891536691,
 0.957442091114379,0.1767185671232856,0.5762453212618334,0.824229172893529,
 0.7497979209771523,0.49300604636656375,0.7086017413596585,0.1154739115166813,
 0.368062287363923,0.5200559937401428,0.5518253147595326,0.263071425254227,
 0.44628226286685535,0.8866111718169735,0.44591762733644624,
 0.4457931823794111,0.526588197659814,0.5296618831370369,0.9152601123334977,
 0.5256202247145902,0.062451507314176036,0.6418038502653375,0.923083728746622,
 0.05467138678517325,0.5321349526183123,0.29509079834263097,
 0.2718178358121389,0.800313166301941,0.1610318225210533,0.09455534100664964,
 0.045452853552444106,0.06358347171498924,0.09212909555296189,
 0.38832762695192036,0.17496836243606062,0.2554616399475418,
 0.0862250768171724,0.8755844284920978,0.13144620255796258,
 0.040610768577179934,0.8819825068447203,0.3326548600928946,
 0.11277801677252153,0.07889265216596597,0.9169699918997378,
 0.0947413406796207,0.5989844081031312,0.7314150025963883,0.8088842118384375,
 0.15576628224336841,0.8913533155525583,0.5426555354556518,0.4448608101387309,
 0.5144620051714268,0.6070735854326293,0.9310262616551692,0.3056611014831655,
 0.44988218455646434,0.7895291924340939,0.2763211460617456,0.2651516614920616,
 0.170274604648875,0.5745948803252444,0.6915541343900392,0.7538821321303759,
 0.4910640930945155,0.31570187908330527,0.5616213485986478,0.7582933264255941,
 0.5875802949857255,0.5840610704851472,0.23997120401762262,0.2704700258844872,
 0.5350594882427966,0.6885330097009299,0.8027878032400084,0.8234120406043921,
 0.5822628761447057,0.4333873956975688,0.8862452989321397,0.7205279383359615,
 0.9504454969933949,0.5688771322708187,0.5076842866633408,0.36592191143040986,
 0.12426355024509816,0.14157386398744132,0.4123020688300735,
 0.9897331036351001,0.3332118519595313,0.11531617103596237,0.740187378529969,
 0.03687262920537615,0.30625431336414377,0.6971248152330355,
 0.1539853797457612,0.4628045042133281,0.5865431164868893,0.3506332306442532,
 0.09209589087404457,0.42242480123981463,0.16776906654033596,
 0.27614165011836145,0.04994747110193065,0.48139109192464047,
 0.046333804623652,0.8792557380501079,0.3000302475706915,0.22798525167027273,
 0.9497373753134318,0.05518770021791952,0.28066697132572915,
 0.5358797181918501,0.3400985863849715,0.34406991108669416,
 0.45449033024242247,0.44768110282902474,0.9021435040356737,0.728220999354742,
 0.5185502769152491,0.5954256835383999,0.7041285961567416,0.11142049987605529,
 0.7194567172624964,0.748612268877243,0.5020749766194914,0.8269861059083696,
 0.09771896938419744,0.7935502232265264,0.6012950711388914,0.7138270149853735,
 0.2898312831965286,0.12007192164682712,0.7041369470339098,0.5251295987457373,
 0.18179703366435618,0.6277468005196458,0.6849395864181524,0.9925149927129292,
 0.031461415263111636,0.47040057332637764,0.5658027936115371,
 0.5070198482109706,0.12555546652671734,0.9004981519498592,0.2382794772108492,
 0.838609179319593,0.10487516414040621,0.7563321835084059,0.5660684421880118,
 0.5305971160520957,0.942101637606401,0.6203152223462705,0.19201606834763196,
 0.41203867036294184,0.058390845714565764,0.7256660476971886,
 0.5530568204258106,0.9759328582157565,0.08705501968695417,0.9308797489544994,
 0.9180129627802002,0.17745326563829877,0.3707873085201786,0.5184660116730344,
 0.042777683396287514,0.23577847657239248,0.9172958717494325,
 0.32406271931842856,0.7671626114081643,0.27582569327710826,
 0.5854812084640268,0.1007814648169888,0.17177961044499312,0.4491074637378931,
 0.4967427259526035,0.22595934093417824,0.42309964646069953,
 0.060163748493444746,0.6407346590920346,0.6969988111094418,
 0.6951912386869079,0.09563229946675789,0.027555917601943314,
 0.4056880613448349,0.052086959034972735,0.2909934708420929,
 0.9098181698789617,0.08398876843087943,0.9591022806255446,
 0.10710828567447739,0.5988136128844754,0.5736721952395565,
 0.13762075874707003,0.45683219554825627,0.4549743311429889,
 0.11869940196405271,0.6437544358603516,0.5692187272077808,0.9540535815690908,
 0.3948021931300726,0.6607383215190583,0.5347356517889753,0.9677315308121213,
 0.16170203038747044,0.25782579581154275,0.3209556241950984,
 0.9739305380445076,0.04785890964292283,0.38343550020050476,
 0.29916961174443824,0.7847989875642543,0.13881232806404376,
 0.3148454975506555,0.8960208691958093,0.9853429762075976,0.4397359878703305,
 0.2775942679140845,0.9426914278005579,0.8272018195674654,0.47008988494906445,
 0.4698972710300303,0.36971610181437353,0.3616897504535792,0.8614049735707857,
 0.04990723523893603,0.9847519189626431,0.9504723576060772,
 0.17725908781485722,0.9617805807934305,0.6777832468005814,
 0.03275680209635423,0.23345893218187697,0.9254887181511845,
 0.38027755063824276,0.3920322246170269,0.1861102623070443,0.1409012175836818,
 0.15346277248472184,0.20134778988577506,0.4177780402010004,
 0.9842313952064335,0.3685643456282841,0.27089364982219655,
 0.17901888213516226,0.5410501534315983,0.35720275222618214,
 0.31593050842651527,0.13738666171754021,0.023988086614063486,
 0.05019452526666279,0.5271092812719853,0.41893876532209373,
 0.7444939353554196,0.46000408536615545,0.5188357961496338,0.9726538815771029,
 0.3848938651223559,0.4569978103060043,0.9274101916335734,0.23871805892927023,
 0.2869187046975006,0.9618951526999809,0.8904485899795711,0.8362194596657259,
 0.31309449772094833,0.8818929751475837,0.9214451097491526,
 0.08679930936320157,0.7564694631848063,0.06593760369215573,
 0.2121369566774658,0.41752705050498995,0.1841854591358958,0.5265367152036898,
 0.3645036778782914,0.466604483559965,0.16723699721371055,0.756878913675571,
 0.967041355018454,0.13444671294674115,0.05723964022300798,
 0.07081846971510064,0.7246301938907269,0.3991182349654585,0.6285619798842996,
 0.9507481058458936,0.5040176534986792,0.2371729484739702,0.6346011925694741,
 0.9840608910840346,0.7533629964072913,0.8464247496078714,0.8813257579996714,
 0.04443657751728092,0.06446927458960094,0.6138067287605963,
 0.18332561793794644,0.25784059146767724,0.012705481800419971,
 0.2631904234058192,0.618763622956352,0.22720732452613168,0.782656097759475,
 0.4115503864419263,0.12017406369974126,0.38233118009381806,0.773649003135334,
 0.5987816197617617,0.645623915230401,0.8334235503917711,0.8869692198954481,
 0.701960594879519,0.9478686476632274,0.2950609822055539,0.863941832608452,
 0.7275696959293547,0.13075190797809855,0.3812633235017008,
 0.04398179572352301,0.3819078111290706,0.32688656284041584,
 0.8187249409222339,0.9545827076038391,0.3186379432262195,0.0686444140664445,
 0.721407275098338,0.0015550636570235143,0.48147654199163736,
 0.6979771990258195,0.22167476510635975,0.4560272158787584,
 0.011692571517489259,0.9111464994314128,0.10266155322930404,
 0.5743855938122588,0.4783551937161621,0.901763849985996,0.189492893470803,
 0.7142055207030187,0.8348108870028247,0.29395716447509246,
 0.30754716427547035,0.775682342686501,0.46403202258828713,0.1919050274824622,
 0.23944658902147986,0.26534093510805545,0.01597462420912421,
 0.321732898783043,0.7909951329836464,0.8583391103836804,0.6250347243273852,
 0.14975001383667863,0.3559865200859271,0.900102405711344,0.6246520259623434,
 0.020949523810018356,0.15630313857978095,0.850416232853636,
 0.8269030500393113,0.020008585604718965,0.5216804307598936,
 0.5746439792366554,0.11458551874942535,0.6500212248907153,
 0.29367590469063853,0.3155106010779387,0.20357612892596433,0.657776747089072,
 0.014827492255473751,0.8645451392140122,0.68270255026804,0.2674516125294316,
 0.14520293599638845,0.3522832214138496,0.05909745203986494,
 0.9386933206884688,0.5421164924614199,0.5838427206101237,0.10284565482530761,
 0.7157529148251459,0.375544682188228,0.6095995908424224,0.5294463219451336,
 0.695834605129786,0.5095043571135824,0.4956382636478335,0.5300478099746685,
 0.9695398371654366,0.6074197249701682,0.27011467372793196,0.6586648733709234,
 0.21683741699752668,0.6418292775034996,0.04815960459242308,
 0.18910007953913843,0.8304044664584015,0.40378785639521086,
 0.09971151170305914,0.4765612566177606,0.7384928981890944,
 0.06118139200731054,0.08729634524471597,0.7966888835924545,
 0.8638812955435475,0.08255491625444755,0.7663772155494358,0.176686565719253,
 0.6320758648222191,0.7743052129232816,0.25607690682172624,
 0.05285682151395499,0.7541472780270898,0.23257939662920823,0.519461928615857,
 0.9800825941118342,0.4258744814196004,0.9423551229243498,0.7344931283235487,
 0.8060283361619988,0.16712210766495605,0.048607851792569545,
 0.9669091217702405,0.122002091625801,0.696195932299893,0.29718917675263723,
 0.474687818357738,0.453584820277927,0.2032784128375259,0.8754778119974506,
 0.008935009204140365,0.030568258746772714,0.5668247712738216,
 0.39109248969568955,0.7711590047697437,0.34834389673878285,
 0.7659521173742578,0.13084010548844893,0.4947991328906699,
 0.03491551368624468,0.446066987333815,0.9030863078459695,0.38381212263114173,
 0.3070212891756092,0.010123587857784666,0.9180853471395335,
 0.8939263063553927,0.951670327768272,0.08318380841669915,0.6190758649844559,
 0.49442950239919525,0.387290631584023,0.2348485780662467,0.36759116603284214,
 0.1388006387284515,0.40625296799087884,0.2514665118588668,0.3898983267656211,
 0.9382958721161078,0.8518267516789929,0.9125646262203697,
 0.017948724122622717,0.8060572553058725,0.6067994684902067,
 0.5177880182268544,0.9743189456503131,0.6006489098417169,0.6263512191368875,
 0.050341430845849966,0.21860521928509413,0.8223949273259132,
 0.8874744979187587,0.6174216778671873,0.9212914336991483,0.5194128014548782,
 0.24153274180454898,0.6001784158665138,0.10849210684061394,
 0.9623095075844685,0.8468010745790209,0.2077058587159698,0.4716260810907925,
 0.3841966449359048,0.3489540878450348,0.20663013537643216,0.6833992232323437,
 0.7113072633846662,0.6151051274575208,0.5495499101937035,0.5217281074580767,
 0.008809796442550977,0.9425047033736257,0.4104981482142873,
 0.4135833818387843,0.7534221364691867,0.5457477190280325,0.4260145078729781,
 0.17221673206183474,0.24323967098240484,0.6546655327484069,0.719444500172407,
 0.8438885964113977,0.9821859529542182,0.6528794677067774,0.3630835989476848,
 0.12885578097945138,0.06622571530907972,0.4741507390830877,
 0.20956470073772304,0.5607539628520432,0.7109323801533051,0.7995986078452717,
 0.0894185020585404,0.33511825717257526,0.325212236281342,0.1364790839244675,
 0.6727478307111276,0.16840501627555138,0.16336970349629076,
 0.5648000565487958,0.508382554092433,0.24908928816217057,0.5186816403287795,
 0.6861782326613323,0.3821987658039664,0.7036725735431663,0.6286458988895571,
 0.7894962403489195,0.679821563882498,0.5315947984060279,0.6998010999228442,
 0.768663363214065,0.007859288053332136,0.3372689435208365,0.5737338969678097,
 0.21634628169752468,0.6816414164911208,0.06932731232651712,
 0.20230794166151855,0.9317531442711443,0.6125449973608197,0.5175492235075402,
 0.057901182195748024,0.03773274132534399,0.3023561477355202,
 0.6880938637980675,0.12402861697226819,0.574727281941593,0.41179977329432904,
 0.7534981948185205,0.9990004568692759,0.09705701800686728,0.9520594341160267,
 0.6340253610332679,0.38061080432856054,0.7717319493447075,0.4852749180268615,
 0.42238354813460266,0.890169057570974,0.7298410696501295,0.18209116012146886,
 0.17923000346873608,0.9618102492087203,0.5788547129745816,0.7742678206380507,
 0.6178716300500254,0.07588283615842861,0.627592429635243,0.8042981619829681,
 0.8059181834407873,0.6600818727939497,0.6584597598172675,0.59538281890856,
 0.7212279775037903,0.9596681092554481,0.8138087176213323,0.05316781379387281,
 0.938990193541456,0.24276039381586356,0.5738948953469916,0.4203635028716811,
 0.412266519196528,0.7947491874756647,0.2085789693196065,0.4380812588684927,
 0.08861377620553201,0.8654247722811538,0.8048993516533114,0.8780594372521591,
 0.6649475482041688,0.5407873103561307,0.8198771591975005,0.9917166736330141,
 0.7491320928715064,0.010051343210501162,0.9847619252267813,
 0.009333274257239732,0.616144931195344,0.06395573068143823,
 0.2829489521744827,0.08332286270448352,0.9205232810369307,0.8939256706142027,
 0.405555811179931,0.2821753452811676,0.566572669694823,0.16471464074959663,
 0.14399896777293886,0.582510534123914,0.25142049170623737,
 0.16720649558575396,0.3276119971620036,0.6171955438702432,0.8695145474898623,
 0.5951546749711034,0.16565206428297075,0.5124223050973357,0.9182424070030579,
 0.7983117688386544,0.2616223418851079,0.3732640277821636,0.6847283613890378,
 0.38769685130495435,0.5193508776090737,0.408963221147878,0.48186689173200314,
 0.5258712212398988,0.44959975854192424,0.6421503989193414,
 0.19437403676051834,0.5469301552070465,0.09898306762392461,
 0.5626243538952386,0.9445658781455291,0.6120416273086858,0.8372484029104216,
 0.37227390158652396,0.9307256034246116,0.2336763578301465,0.6749753914876159,
 0.49760714945195217,0.5317534849197425,0.6790233716650969,
 0.40273546964428686,0.508916781927095,0.3244200564878704,0.8896274818734669,
 0.6039409852284375,0.8618793692129224,0.025073534285668164,
 0.8030656182252786,0.29786579318588624,0.9023763057645033,0.373551120957458,
 0.6687595371832076,0.8210155303523778,0.47807169943836925,0.6089395625581014,
 0.06351658247051784,0.2270734003412539,0.7259264321476168,
 0.22612150479286708,0.5255545121424909,0.18793494821449863,
 0.9554869988775384,0.6087737896175099,0.5298449584187636,0.30731444488680226,
 0.3476847430818294,0.07917204707824088,0.28554375141843646,
 0.40440835228527683,0.18870751346566017,0.08708114054544147,
 0.4872121871798345,0.5484483696412608,0.9506070378947328,0.7154020874652901,
 0.3465112927033195,0.7255060787051798,0.23912372193716147,0.2018140106339037,
 0.2693367548174177,0.7108741608366783,0.8931997939376941,0.9847605644874374,
 0.7159626749325142,0.5745047935408537,0.8223325663023768,0.19072879761406925,
 0.7136871720425738,0.18197941560353692,0.8738033281691702,0.5695764302371457,
 0.3391768214334788,0.6580854634356769,0.43930179092704635,0.6177891736616243,
 0.7853743578386225,0.8909590440894755,0.07582736008592572,
 0.17779926687007097,0.8092413691612182,0.6999856235377641,0.3854559403806156,
 0.8538168472748867,0.058052905991752946,0.4741070782832957,
 0.06394763065796694,0.45624866092411465,0.968188303989341,0.9781940128402026,
 0.7091608467632909,0.1104522094937963,0.5716244567279845,0.5837195546409552,
 0.5714670045378715,0.0013712547623779825,0.05530315737753522,
 0.3102638746586984,0.055156519513929525,0.16963337162304526,
 0.8140796469377292,0.48721831914861613,0.5996735329441384,0.4272578524080486,
 0.1361312148751922,0.47688023355694753,0.4124496572707499,
 0.050062088724945664,0.0417723138046584,0.5748950920044706,
 0.07356401195272033,0.8288018507102836,0.45851798063076776,0.258045557970202,
 0.23200204832453164,0.09020850150729132,0.5461882915312894,
 0.1620077899400436,0.6128888706970357,0.8345409545726554,0.6998409260113819,
 0.41467152865851775,0.8014955424785987,0.5388209031482691,0.8933278222618061,
 0.866358045506265,0.43190074128358524,0.36539318399306875,
 0.34901358399287075,0.016981737658786078,0.05754400070692112,
 0.7211229292608977,0.07601794158491493,0.8371025572215467,
 0.01693121669156261,0.4534553395535914,0.9607918849175423,
 0.004373185255167611,0.41240353034364796,0.7095655578933355,
 0.012309032850389645,0.7250015375272589,0.03318003120764457,
 0.830347867167189,0.2941780557956175,0.7934883156072573,0.7267132836432526,
 0.9664439048861952,0.3179795926927942,0.4451940195257237,0.7242205265302785,
 0.4628822455630015,0.02810258244592556,0.06855142779482648,
 0.8185366874131863,0.8851170180811617,0.2934452204989433,0.2996884701524203,
 0.7466773396934735,0.811515382355341,0.4746470395016035,0.9371582993525066,
 0.5954741735424361,0.1027433553318543,0.6073376187721828,0.7060535264523249,
 0.7041038621416599,0.834000024271464,0.26211544772819473,0.3073446350607604,
 0.12678252220452202,0.24790297783414106,0.30691697736175194,
 0.40223312805210054,0.5979684602563564,0.40050698447730604,
 0.21730413614151345,0.4008878204505657,0.06693141699431948,
 0.7081052909974698,0.9152986978439019,0.8694454945762035,0.6463306205064087,
 0.008721001707077125,0.045287783520169045,0.37462038663304054,
 0.02617799661798026,0.4258181804054034,0.2231680423319493,0.7776902479871414,
 0.4642125178016213,0.43022538541881206,0.09962615909125716,
 0.6967039806392056,0.44702392578165295,0.632572823257931,0.5109026480846856,
 0.563515563845831,0.6209370241929761,0.5703139859130673,0.7366698702225192,
 0.3342769408508701,0.4444520633006568,0.22643024734747774,
 0.08615136238127163,0.8154348879056534,0.2750292890910909,0.991016534874859,
 0.5489489059112976,0.07678675334639995,0.4485046366484917,0.4046380226932689,
 0.9079531750102132,0.7755715109302647,0.3550946410859823,0.8469073675725841,
 0.2801684583127959,0.6485666540001891,0.898710133520148,0.31186758640894996,
 0.9067038732694901,0.47484738163640006,0.9580146239588394,0.6379081616272783,
 0.961355754337876,0.9374396743882805,0.9232499132871264,0.9918330455241605,
 0.6514769111680629,0.039722379489801285,0.7365426939527404,
 0.7395642612914103,0.8560133804456878,0.3752044981219096,0.13075047716207122,
 0.5103607426575496,0.4734777264472473,0.24507242584437994,
 0.40223353969859754,0.2165095748657504,0.6087354370825101,0.8374798565133253,
 0.8654410510872589,0.7706039794753372,0.7236375903140265,0.7266401500710133,
 0.14774943144447528,0.7705550669143781,0.9306355127391162,0.9914071179879794,
 0.867709539081945,0.08468557892186324,0.8046568939985894,0.07974774652368544,
 0.02811585149474627,0.3286776071507861,0.14972126710224087,
 0.4995008402547123,0.903581149225187,0.0299592166037691,0.3624698293364992,
 0.10340733767488053,0.18835190045825845,0.6196923095931008,
 0.9943809129447106,0.3822478924547852,0.6792198728773888,0.5445949312432756,
 0.4694513965074725,0.39869924327215234,0.8626606586165022,0.8779802779553797,
 0.6056145610467263,0.46594729245826294,0.03251198247175657,
 0.03980225545723781,0.3938458742873009,0.6324539337820532,0.5383337253572628,
 0.46715813527752004,0.40697822320933685,0.10195335812458628,
 0.7993563943847765,0.0040932938340945135,0.19715145559353642,
 0.8160187221033534,0.6972991142736713,0.15007398425195106,0.7655136901536642,
 0.8213233003119512,0.6729686669059735,0.7073603676896786,0.359129812579116,
 0.6559081714460647,0.06180995928617694,0.4885172827077753,0.7614126535936202,
 0.3518076134364726,0.5260693275044348,0.5652217866223669,0.7061212422350182,
 0.13576723317849226,0.39211239772615447,0.14139646741998058,
 0.46219883349310353,0.6183774978522368,0.7387153820484582,0.9128105376804484,
 0.26026210546250106,0.8192251191264326,0.9048771556444967,0.5003957534716497,
 0.3320636456078141,0.6159487640840515,0.8961182977568434,0.674431656592049,
 0.5729053786612871,0.9998109374058959,0.5307981219613128,0.6959197764545558,
 0.3179003006640724,0.7247558299904588,0.07350093651732859,
 0.48965792992860746,0.810677257110744,0.7430149926629727,0.616157016043544,
 0.24640602881914853,0.25240477254512317,0.5005986276495631,
 0.38660007741797164,0.4884782963343569,0.6075671862733456,0.2643165752933513,
 0.009181757003298596,0.7476090321259341,0.8589615469658751,
 0.46339972017738484,0.4935923371489731,0.5434147005502344,0.8763656431264903,
 0.6972188568548899,0.3236618503155454,0.20278920045525295,0.5156419168675163,
 0.424216075356155,0.8990130244003981,0.8804469302885476,0.2748230662080453,
 0.7363495587141752,0.3434407400792947,0.1602338471000595,0.9272782822722989,
 0.14242058447129835,0.8029017039580111,0.0625638076916133,0.9436060431317795,
 0.8301998771775787,0.0011613300422966821,0.050015650204233486,
 0.6277891221872981,0.8427810094782694,0.595893063885434,0.18366908046190034,
 0.7709519356057228,0.4091826164875989,0.6149958752293493,0.14381300761646842,
 0.3982034921220472,0.6083370302872021,0.677479454987338,0.6199128941360059,
 0.33054157434281206,0.846040831039251,0.6003407249683227,0.53897353437973,
 0.9088590501326842,0.4971747190069513,0.33272507072851565,
 0.09560346490937421,0.9511850562794814,0.8835483647805868,0.6710711135848769,
 0.3324827406174464,0.6916782261640706,0.3992649223371306,0.903293226140962,
 0.051101677960145864,0.6730611524661804,0.4152344988814364,
 0.8711193033017401,0.1130027341713209,0.3050806746697875,0.9819789399998764,
 0.6103678347985027,0.08431752315287733,0.815103538405136,
 0.008915507196109118,0.11731192597777118,0.89901469964839,
 0.20217159437504317,0.0827733772657141,0.4488343918241182,0.7188848471443776,
 0.89775361433531,0.5267522058689763,0.5703152121751821,0.035981545315141306,
 0.5392630635379402,0.1690883764614658,0.22648556731553748,
 0.11639148915159914,0.6401820461796174,0.7175008497500139,0.9299061451924138,
 0.9062794078965035,0.9425681881644865,0.26400053903162324,0.5273838055524336,
 0.6555534480927472,0.5550492788650083,0.6722980964809722,0.5610795365054285,
 0.8483246211407914,0.7983205713770465,0.36847461509776813,
 0.24851272368692368,0.5798050665631846,0.06744434901012519,
 0.8929028880743552,0.9314130433840013,0.7709049724342302,0.12964182711487549,
 0.6735820606553538,0.9444356898505817,0.8270344336312886,0.7656901541753922,
 0.7517143049791298,0.4158830186366944,0.08333884257634838,
 0.06041960186483841,0.36496288677130373,0.4908091247507358,
 0.9756953427658684,0.019787692821024883,0.7548909590426502,
 0.3022236906866601,0.9335948322481697,0.09220000615122814,0.6234404961077157,
 0.0027526722551207694,0.33516348877382496,0.2055675158992415,
 0.19122563115006586,0.9480942839572035,0.911874607926836,0.11970063452571889,
 0.9909891474976285,0.194309612750322,0.7520228544572375,0.6269310415041803,
 0.6808490734778934,0.47461763370061016,0.2045332318786489,0.5558714052548528,
 0.3050304966230024,0.1647741738198596,0.16647425516835357,
 0.15437474616558577,0.7571789548654775,0.8554638712826944,
 0.28417664916410723,0.06677268945229253,0.09603435193283982,
 0.39681244396949933,0.4400856540023047,0.7281960524565777,
 0.019823381999590328,0.047817632771119056,0.37603374489477115,
 0.4710420998777821,0.4255845034971407,0.5613706994491723,0.423580935162155,
 0.20682670617557863,0.8133053299782619,0.834647266295824,0.962861526516932,
 0.7416405005793023,0.2702509104812676,0.3708589396879828,0.5856569076009402,
 0.7463780920723354,0.3841817500889414,0.051905652253255385,
 0.8231918320359725,0.33997259883401687,0.6478367881861065,0.5874765124915865,
 0.21123064464301422,0.2502182184369125,0.6766519635339351,0.4758532792531617,
 0.530286454904092,0.8511863347434375,0.22136919909446073,0.0829254426181466,
 0.862152782980163,0.05058366394796199,0.6215521960416764,0.6870245092204272,
 0.6342373046651877,0.49757359509187926,0.40563657089740746,
 0.0034673205626811754,0.035343827459982324,0.8655139834069222,
 0.5841339681800998,4.715646100648119e-4,0.502857648858379,0.4364006400934216,
 0.20567518032710763,0.12999832628815933,0.9362440044912689,
 0.39024213615721226,0.5674768108913121,0.7787263745378504,
 0.48006003335770087,0.06473643829165043,0.3524228424397434,
 0.9903144319491224,0.07683110861116649,0.4722321000032591,0.4368683234612012,
 0.9573033745984643,0.3582278609256757,0.07790958215300003,
 0.36010140007774116,0.6038974774246133,0.12031798354489787,
 0.08570363122538438,0.23708092686289606,0.15214815574442908,
 0.8858058080497349,0.6577964770794943,0.40307029642522696,
 0.034433635421117326,0.035451999855919514,0.7761829582832792,
 0.24726713710046178,0.1612671689075872,0.6223624868382309,
 0.38263154432930757,0.5765422064858218,0.057498931736123526,
 0.46634470051857924,0.31665513553870006,0.8846325710284166,0.420594817198508,
 0.14065936948282454,0.74338992610388,0.7431728342594006,0.9055627542347114,
 0.33312610006029464,0.009133999254378633,0.4115776028890794,
 0.28977929562027316,0.4678483546147514,0.06326483506076186,
 0.13304336771128067,0.7671920534953884,0.06636523793175031,
 0.2498260277685782,0.7432456633364815,0.7122144492606087,0.29713389172792615,
 0.8748812189902511,0.3866702908077244,0.7031462034856389,0.6867477098528729,
 0.21903338528297867,0.3518133014431556,0.933275388664148,0.902982108645752,
 0.9145121598078536,0.8883414565832449,0.22043340255353316,0.7829959079288248,
 0.03227847184809152,0.3787592710476727,0.14646960102828532,
 0.6462389090113053,0.41567767470956674,0.19520369307924923,
 0.7258356669077356,0.495538666774501,0.7084671232740419,0.6955584521832524,
 0.007435630588047859,0.7346597764156215,0.86033435386584,0.24980607732767923,
 0.9651791686766753,0.9432633787766546,0.8511535075334907,0.816580171708357,
 0.07429291144874883,0.2350684406334106,0.695973823859968,0.6108387500821281,
 0.4354967327163264,0.2707927688250378,0.9829346981869708,0.5136135158285894,
 0.6826418337482969,0.3576553566902686,0.9728687527819234,0.7314542475317152,
 0.07712518677487434,0.3567866704540573,0.44842680434125426,
 0.45204636417054833,0.8201923337110731,0.8838274217974919,
 0.08558810127628225,0.3521913402364848,0.8598981384612534,0.5815572265401103,
 0.6254528260272686,0.1312195808781388,0.2571122018297548,0.9817226530142513,
 0.45209309054199276,0.695421543699911,0.7560993020409136,0.6856608347949956,
 0.23687160457437761,0.10679143075844988,0.8493669943547335,
 0.16434503675928203,0.33290809700050983,0.5204612058175796,
 0.07991048758589703,0.18362058869691522,0.4152310617608934,
 0.8552443063201898,0.2836275121700953,0.46601349765606814,0.8947611805949643,
 0.2887633120816854,0.8212785481120524,0.6868101359615099,0.802305578780321,
 0.08216844400398005,0.06383800331149148,0.4177759380186503,
 0.5214401635642845,0.18296234162292402,0.1607994157372502,0.5747253608416864,
 0.7641729467776552,0.6659578594506546,0.931209302191461,0.03720688135100225,
 0.9854230976525424,0.7578932942550256,0.2402611220474995,0.7140361949480125,
 0.646337568329936,0.5956465611099543,0.21604753129780396,0.23399619628599444,
 0.4515512912931082,0.5805082787271865,0.1510616846929579,0.8764726617678475,
 0.9729848418221411,0.908914145446019,0.28755323273882116,0.5552360931453554,
 0.49599046534291047,0.025939849985598395,0.6165591374750019,
 0.023121323191937604,0.6773743560289598,0.49736259712732056,
 0.9501303824726944,0.6509093371803616,0.6337145900066825,0.9199651268125331,
 0.6784329211897356,0.2799145804024772,0.6056492270211253,0.9666957858457602,
 0.7715785952614798,0.882462741308983,0.4104518731479685,0.16404499265596217,
 0.45948296103252884,0.8005975978102944,0.21212182793841006,
 0.6195896498360217,0.9284632799916602,0.5780934028856758,0.3066798183725232,
 0.5034151255984239,0.7730551152308587,0.1531580533223854,0.4213980682549713,
 0.7120636335159974,0.539897575426224,0.5324351191197774,0.6586386169661825,
 0.7426375244358471,0.6704413544292389,0.2149578512702952,0.36895430226914894,
 0.8312942250225851,0.4997522915692193,0.6787099871166449,0.6077115393601072,
 0.3954543237912246,0.4675635262835963,0.27946674505603175,0.6191978623791439,
 0.09929568804935995,0.8983146473221694,0.3671804685519786,
 0.28076117351149243,0.4882465260442189,0.35253126783618305,
 0.27004058917361573,0.2682687263732074,0.6284593876111335,0.4903545745538558,
 0.5968332798646985,0.5532142639083545,0.05117007769951709,0.5469803517406293,
 0.5873654599620912,0.21280287117691332,0.0656276809774885,0.8755682227710715,
 0.9682891836185713,0.9815031903723717,0.7453509543175333,0.1253216616738232,
 0.9447051178757119,0.15757328655522307,0.6504752365332401,0.1694768096619269,
 0.7495684488838853,0.5235271231205152,0.5134593663211319,0.9359252417171372,
 0.13648031796602067,0.06514451498744078,0.41895748099420016,
 0.5202991437677981,0.3091510669797234,0.14464649413288,0.953537131024569,
 0.7524820186856052,0.9572971009800105,0.9057802625575806,0.16737920786543226,
 0.6278147022222307,0.5573036892898234,0.30204446995427947,0.630729078050756,
 0.5153589391460236,0.11007082034256876,0.2440704854544513,0.7859665159004637,
 0.371909812474538,0.6316139418367002,0.798575272328843,0.7569090951247024,
 0.6463703978491664,0.1890232649787188,0.8501428514049427,0.3511547454161361,
 0.2798496330186444,0.9552815239791057,0.9062161946888206,0.05591446526270594,
 0.8540182004037309,0.8247271940883432,0.37013398129904596,0.5034159135259458,
 0.293245697858859,0.5105480845460388,0.33946077342278325,0.875165697562222,
 0.09456579747387106,0.18891285635414434,0.11793392031187167,
 0.07661801276561997,0.788944252697687,0.8488227811859016,0.23604826857314254,
 0.7429251252015883,0.29117067576645406,0.45590645632885796,
 0.6915183075962847,0.9752584293383496,0.6052662714776,0.4016130442853286,
 0.15917746379568198,0.5346236262471139,0.8660440994916877,
 0.13564280750962587,0.5185194333640106,0.8313253410043137,0.7536408513667224,
 0.7725372409667958,0.3035928459424265,0.8716868728448394,0.8281039502473247,
 0.1636769459220968,0.8993704992414897,0.8644048167299918,0.2628524875002862,
 0.7142361425655667,0.04705249951922108,0.8665677674121879,0.6224613662273237,
 0.3588620117498871,0.345565177552552,0.8560880042765968,0.9446648498223259,
 0.7294504727399338,0.49402434828152075,0.533763816519974,0.8359565750693824,
 0.13183669378167995,0.3806097064330696,0.3597079519653993,0.5676968626179403,
 0.4639779222026369,0.5413576035012501,0.4621926493772939,0.11584216274991199,
 0.6815493066374354,0.7625553045733227,0.7010273389766482,0.6649345210782227,
 0.7549644759282248,0.02133417071124344,0.6796928192212582,0.5670862563957866,
 0.828347582874678,0.12360462408528616,0.7494583341452141,0.18154278647319888,
 0.9941044484197297,0.17492597228551876,0.704250033497491,0.974412096178354,
 0.3165399842614983,0.7038361345594795,0.18979671484102922,
 0.33073099068193357,0.2870034666631577,0.993996471380239,0.6295462091118402,
 0.8123847726756823,0.36708237256606124,0.16329249419477831,
 0.13528054327750139,0.7051270332819186,0.7935772614705467,0.6697319952224439,
 0.5927705518783788,0.3347768476967008,0.6435434151028188,0.4856254773465025,
 0.953361003327259,0.05390640707074745,0.44373277804339273,0.715446976301273,
 0.3816317237249284,0.670166611691278,0.7584951016227208,0.9608111464732376,
 0.08902770608585442,0.8449333631646047,0.17331001699824733,
 0.061472776221363934,0.7409309487037412,0.9450011099701883,
 0.38420993985410345,0.901754396102884,0.8469732874324437,0.25044445515820324,
 0.7599018977702314,0.3755880388562922,0.7081213581408539,0.4584870326038205,
 0.9433930009284693,0.7497103117491355,0.49583933536320224,0.6168875991512075,
 0.5891321967772997,0.4938793168345951,0.9744674127100157,0.4859523522110444,
 0.49420167202981413,0.3524872046935038,0.016992221090820214,
 0.9951834117834335,0.14703773405864173,0.8267599196472462,0.6876280001141273,
 0.38287931934502273,0.9580008281458384,0.716641502116191,0.17395902924394524,
 0.37399030229305497,0.580315658713817,0.08654722066615506,0.3762174642787506,
 0.21412890462570577,0.4361338386598549,0.747453779904982,0.7201309766497268,
 0.9746453209436851,0.3530937405025746,0.3891429949328652,0.8363711458247908,
 0.18544086170159646,0.8156384948395755,0.7728007593547875,0.5127803987752014,
 0.14057712009191559,0.9180070774585137,0.396134193205846,0.39135574865483935,
 0.8881246641256038,0.7433807992207843,0.9359876634362898,0.46033367028685246,
 0.23183559820513666,0.8487056785069559,0.05481117882839337,
 0.7170848425517932,0.8485201836662166,0.7180518089383161,0.7447834196546909,
 0.1663492321502179,0.47123092886040563,0.4766210630036847,
 0.03681102457737595,0.6581182190206359,0.9781057506497597,
 0.0012014633320482915,0.5368131358894918,0.46868164982521476,
 0.34944244737581276,0.9859105633619929,0.16886915049063644,
 0.05262552281839028,0.5361187118805766,0.004624410148762803,
 0.12838292128498963,0.11529252720981853,0.45907105403620063,
 0.7986573251024958,0.9448551592553381,0.3219662555234053,0.26340105467858477,
 0.41161711008231217,0.4972961979041899,0.3510138957357043,0.6142117918284129,
 0.17005367198648136,0.7028970905029436,0.09335981881155087,
 0.14819121902243004,0.6814904391936637,0.9248540956580147,
 0.32324615716121463,0.12024491342034349,0.7380083007463276,
 0.2273598810646651,0.6246519401327378,0.6794399602387364,0.6073835643266003,
 0.994474905988606,0.9629902798176934,0.6397530961353513,
 0.0031317373437483864,0.11216950290951955,0.5712458505504672,
 0.06318753412864941,0.13289473367204918,0.8021978880792106,
 0.3442984533442903,0.7484230998383928,0.09015359316668903,0.1622515221936066,
 0.21082535186384244,0.9421418390990406,0.9342105461949506,0.6080274870571774,
 0.3749627050100983,0.21697867181224373,0.22566236028252762,0.610853523467972,
 0.2836794628926883,0.18382419189878596,0.5858294937641518,0.6922873584417856,
 0.5024323611763662,0.24938836912683504,0.41285213246561603,
 0.04407727379093307,0.6313244716452822,0.8613174729446134,
 0.31872973682723926,0.9252951117945665,0.8055593679468376,0.2638421603420332,
 0.8301792924661496,0.7495185618671966,0.10548333882199279,0.7635224286462502,
 0.5773680176435333,0.13082423360940743,0.6543193986384133,
 0.09311832913296159,0.8985312811549762,0.8632390651150846,0.0703642187974568,
 0.567478020358051,0.6532120655627917,0.5167104086833303,0.540924303503713,
 0.44046864541620434,0.4970415583163742,0.13255217240498685,
 0.21511125755348004,0.3775143134968151,0.46664335471014406,
 0.2952740877353014,0.366840839583919,0.9701891359914864,0.5264133758296874,
 0.015576495764250686,0.2086952073330275,0.3594994577213808,
 0.9808106810932471,0.8917156754818115,0.37393713460148503,
 0.24804210047254371,0.32664406931384626,0.5765635517794376,
 0.3128840702565019,0.9184624876315804,0.747207388575907,0.630832541415935,
 0.4336573680633623,0.6777520686900185,0.6196850529250794,0.5565651854452007,
 0.5766294645356633,0.32595685048953993,0.06389200654599103,
 0.9453766331322483,0.5716837098263816,0.2677511141091309,0.7142571626437538,
 0.5395774583505815,0.03110652395987823,0.01343553763085803,
 0.02295363376682147,0.5159186441068138,0.5742158791220925,0.9073893595395572,
 0.4478575800493274,0.43698979940919735,0.6948115484411574,0.6020419002410967,
 0.33652413333047615,0.8029862036507612,0.8025313762911194,
 0.036295935322667394,0.8051875732620402,0.9752041717299649,
 0.7132479258947837,0.5646438467985231,0.3346603015554449,0.4079696555668074,
 0.13482742308764206,0.23348410247335838,0.6249461520200411,
 0.9845072611344134,0.24559886504160655,0.17428471329200157,
 0.07262072596865388,0.3709248524001202,0.16866585594802164,
 0.2841027242681228,0.435754422776499,0.1913097252691821,0.6072356576649798,
 0.7910578792898693,0.9240784212806482,0.9569808811840147,0.5292502734304183,
 0.33009457204703385,0.30461914946276425,0.6768357494242662,
 0.49412951897344715,0.7089781594235355,0.4497288522672038,0.3738231657963915,
 0.3603431759508835,0.789533794621245,0.7443125270664063,0.23715222301420502,
 0.4218327821079535,0.6275369599323506,0.32672766269632214,0.0941954871960703,
 0.10668743452025686,0.9618977743843011,0.4995515111147171,0.5669245211412735,
 0.2377105020813418,0.543389295760242,0.5995520802117702,0.3882423425863395,
 0.0893045535787137,0.19597480442883486,0.7660195743320153,
 0.09420127160880587,0.3439667805408886,0.9679061617032284,0.2117580725326894,
 0.09953558654026218,0.7463614455333977,0.22957988963136966,
 0.2715024189618691,0.7418842414675432,0.5883108883627279,0.7993433411590687,
 0.8799538381170277,0.6326836662081022,0.5107593159124006,0.918734652458483,
 0.9501476185082827,0.281396912503151,0.8792621791823929,0.08354744209493514,
 0.07984002386310196,0.5766401267546057,0.9788064755859751,0.4382548709504145,
 0.5304227549624164,0.724271293033758,0.3366957157162861,0.9935479903852994,
 0.1931038578952463,0.04397534813519055,0.5331705167061573,
 0.47164167080669206,0.6910028765617345,0.460775415182784,0.23654817268810457,
 0.8361695690412803,0.8535103153580899,0.7223927527134822,0.6984330401979689,
 0.693213134756598,0.30145479430448163,0.7720348014945596,0.4360396460172439,
 0.5179802073024545,0.6702721358037014,0.14645728010121162,0.1767469903320804,
 0.5549337283037964,0.09321466548381474,0.6301829970781294,
 0.45659972518933156,0.7530485306996555,0.6076125853485331,
 0.15589417204253697,0.1190825609525471,0.0021121392501614045,
 0.7358431136812079,0.6908726282310675,0.8101286223442302,0.8956265410965584,
 0.17689858848423057,0.4556423387960049,0.46635293237529973,
 0.6577872254667791,0.9274995768835846,0.9987334544623913,0.05013150625666751,
 0.5733263691246115,0.9957020844775307,0.37871236174160583,
 0.23111482497217428,0.1721755370970971,0.5367109153636794,0.4594995463499463,
 0.27222193498966996,0.7217943632607117,0.8730366141918284,0.9538254959669212,
 0.43838462778425036,0.843456155222928,0.29557055843792923,0.6740314070775799,
 0.08268228687632773,0.6588791256802193,0.15582517722399714,
 0.0359033866897025,0.09467564741231904,0.7478556808343655,0.6415083374742129,
 0.16659849022352513,0.025657765954522294,0.6767447381157305,
 0.4724192521267154,0.4552955268875257,0.2216722240424719,0.4054813045714537,
 0.17691381258781413,0.9400883731740907,0.5455871885515295,0.5594862195311244,
 0.5936480119553612,0.0270030336873589,0.7481952513268864,0.35751835955980715,
 0.3720936265579693,0.44033081193336143,0.8653770664489303,0.7072354160095037,
 0.21039441492318356,0.9096352097344375,0.34297486108246245,
 0.8058421307377112,0.46238670562791395,0.8748769094048825,0.6625358504590001,
 0.27342545820686526,0.5157639313793514,0.010867201011415695,
 0.4866168104179218,0.47420146201366686,0.26037092680207863,
 0.3208664509494201,0.6758765928350995,0.7116986397649242,0.5882122114829453,
 0.4881591711498643,0.003795803690333721,0.6959005754708563,
 0.8548939045382394,0.7245346837135392,0.802329510574315,0.27854744349093785,
 0.14653311023868243,0.3270752251794249,0.7971012762609475,0.3359543002360097,
 0.3981801233691066,0.7568843342900428,0.9662155987195796,0.21516600638536731,
 0.2798113908349492,0.17643979848033686,0.3303058065365678,
 0.05366553753403369,0.197496757697592,0.7157119037243127,0.2972633381290992,
 0.9725861003110102,0.6443322793036603,0.346709828828332,0.7244957495393382,
 0.18587730139443892,0.9796560452458176,0.5773196054082897,
 0.26629542090649094,0.9243058805225706,0.3784997415538114,0.8884206190659782,
 0.6316757787863905,0.711913471404264,0.9039468565987205,0.9584475421862927,
 0.41985091494653815,0.28694668302995474,0.7351542613211668,
 0.6389740304337163,0.17969603270003065,0.4856419414022046,0.6092526489192556,
 0.9357176019157748,0.2643249047175108,0.2725765030411089,0.36422791203830707,
 0.5671579603437781,0.11774695248798073,0.7164232731705621,0.5202320479971068,
 0.7294510682813796,0.5952071020726544,0.1478052693220815,0.5160714018321761,
 0.37861114291847997,0.06681982564210065,0.6833020556950588,
 0.15781855135919354,0.4318878254912688,0.6378080432548217,0.8566312020603896,
 0.676849597441667,0.3274119235588494,0.15776595113101022,0.951868078499477,
 0.11008551630841734,0.5441082057433502,0.6103394733767711,0.9946468263549685,
 0.897324433006045,0.7676094036895369,0.8846790535551966,0.6019929627936312,
 0.7801858081761694,0.9384167241772412,0.12260482226185598,0.460182518232345,
 0.009498998541100256,0.03734251241962594,0.4085142285390073,
 0.5416522698357042,0.0900714676119787,0.4400074396011935,0.2706658887463935,
 0.15130816691430715,0.8553737843548922,0.5053113214928991,0.9838353836210251,
 0.09776500674647326,0.39612506320103613,0.891544719580135,0.5746873090077683,
 0.9155237706296762,0.38208730286882053,0.4752879303171629,
 0.28765234533046513,0.5056664905833562,0.6307641227068644,0.5871851488631328,
 0.7114476238252202,0.4404734718664667,0.45547957056701727,0.1263990645541343,
 0.19478114703410065,0.005524512221748434,0.1999192678110796,
 0.6214013884482301,0.4796006449949104,0.8168957920419307,
 0.026750980047481776,0.5436147189184404,0.8765666454640535,
 0.4177241018632373,0.402771050534409,0.19847656432032978,0.599060428766884,
 0.7501300935345123,0.05446418567631006,0.9632385171994324,0.3147827975482279,
 0.3902568650422119,0.5286549962833652,0.5109723497547779,0.1916692730448326,
 0.3061500882519965,0.7200447052594968,0.1605712892910579,0.9759300130160842,
 0.8472667761498321,0.6450094651349132,0.3761643889483688,0.41041376840549615,
 0.8924682960769009,0.775317050286981,0.3718821776411485,0.2848479809455202,
 0.997039668454465,0.24517188132520473,0.5347232697095619,0.7385355012942791,
 0.009847615117674735,0.503773968116984,0.9839452374861661,0.700992181647604,
 0.12836777605675487,0.9927899337938203,0.7098576185449224,0.7845101858230347,
 0.22695558107170766,0.998772541630498,0.21834687302226208,0.8417493833145768,
 0.17934003266810727,0.317219291700071,0.522255397658389,3.803635657286808e-4,
 0.539205072988062,0.3585875359439611,0.963614965801054,0.4652960221559447,
 0.4023258547116735,0.8892282699145341,0.7810390191986668,0.7140130698017769,
 0.8702506637975091,0.6314605117691641,0.20106968177083873,0.8435679661785347,
 0.13742697892351408,0.6312550971100834,0.37462069035510104,
 0.4474775373637243,0.5841804544916337,0.03751297522602093,0.6943840970055994,
 0.7694043153878207,0.6641303207321725,0.6306851451717371,0.10348660020600897,
 0.059347644733924954,0.12216049807075735,0.5148691832884081,
 0.4605847522668022,0.2541288616239614,0.6423005351180594,0.6138567295648827,
 0.08066469742476101,0.11034812481050149,0.9814690488395303,
 0.21619528064549876,0.42878670115281237,0.5972779748509142,
 0.38303071259793475,0.25623468367788704,0.845881677438407,
 0.09618884832479868,0.42130165517337614,0.5225541147176671,
 0.3466313709801645,0.28957903113138495,0.603771579013232,0.9390472493466332,
 0.18370085659422164,0.8033850731304535,0.24017298415437005,0.641341914198454,
 0.7613179484871468,0.6425896368009294,0.11336081086275862,0.2897692476431114,
 0.42496694433390936,0.21760708751247382,0.5451597018580396,
 0.6754750609333499,0.48747026235382374,0.42580135645154793,
 0.7505858207130122,0.2468067642215207,0.8780428765687647,0.7723024806354427,
 0.6317613067888818,0.5689542629701252,0.07859305454748733,0.9079753495373452,
 0.5027400295816069,0.1365736890096071,0.9328656167062398,0.6300203280066106,
 0.24403225729742628,0.6807505366611273,0.06699411465474903,
 0.001197130172435723,0.39085175084733514,0.25506532251821157,
 0.01674626308894045,0.7211528795211959,0.09724127469049693,
 0.3634446267654262,0.1343811320125351,0.6583669625720016,0.24610626501254584,
 0.9991042262391368,0.52096645279995,0.8167805179374406,0.6477186972925216,
 0.1147950871648813,0.3025452530532853,0.021516660990715586,
 0.4431778872717689,0.9650115409581792,0.49730502679397104,0.841489723159154,
 0.03215455443082982,0.5958485526096382,0.28963928446456744,
 0.8627808230262309,0.34383808966723994,0.9420049428437854,0.7377452829030211,
 0.9201872506992527,0.6406666560722541,0.6305212751348164,0.3766858985184932,
 0.924086007015513,0.5745098757146493,0.7106139028378594,0.5869922927896415,
 0.6403120578781316,0.9551835965197845,0.33840881282475854,0.4588278633392323,
 0.11098876405731861,0.704515978970705,0.44433856009413275,
 0.06917346294848226,0.7107883173715818,0.03990638816684822,
 0.4309771603953485,0.15555980644194278,0.5349530889034747,0.9121660527973212,
 0.031751035628639235,0.7900389441720805,0.777515745429635,0.2511218837814234,
 0.30736520480697327,0.2689610099697183,0.2363935269890728,0.5096800820907013,
 0.130333319017242,0.11640652985721822,0.010283668544243008,
 0.7617527966199039,0.6044929364846761,0.5893574783308605,0.8738005191823315,
 0.00792730631712868,0.8736389114322856,0.04826207633368873,
 0.7193454438783374,0.0802038188475892,0.17385295150804203,
 0.16576808946516608,0.1805741844300296,0.433775453888555,0.9535447741117757,
 0.605672051642122,0.359806527868479,0.22102766712628996,0.3437369051357122,
 0.15005580335593627,0.5175434056920942,0.19466515526067085,
 0.4445354126833482,0.037149654110327335,0.5352611729619149,
 0.7936951497468947,0.9782044882398936,0.24132613474333475,
 0.02487450518009504,0.5538504110198238,0.5299961447858729,0.9499798953468731,
 0.028158137101264025,0.06474550572983806,0.9679162113554184,
 0.6146916708078863,0.08700427458005289,0.7431370420407758,0.9130553083704733,
 0.122275548093287,0.33304196735322766,0.05481170957387049,0.1503002556867722,
 0.9426116548960737,0.6786656309316879,0.8279501769496819,0.04407516027754532,
 0.5099222882269597,0.17871261807013517,0.7041891708761634,0.7027082769398475,
 0.15877938986409978,0.4587718310518697,0.8303243681924007,
 0.19937460075444036,0.6747308151747999,0.6235722787760184,0.7562781569645868,
 0.525987488866079,0.6610675101430878,0.8201764419742267,0.6593744178851517,
 0.10302843660343064,0.29147944572381934,0.8483508056277038,
 0.16263011152497198,0.9359982136262026,0.8922610960464076,0.7135625490056556,
 0.1259733440273867,0.5709400199138104,0.17250702869792123,0.6630472083728781,
 0.001882597955018217,0.9606847210104676,0.4834873047531657,
 0.27666726890181215,0.6722579579161228,0.623472359646031,0.48985342342689686,
 0.15994736705657098,0.4522948344646853,0.6722427938851058,0.8173341202228044,
 0.49303403420598857,0.11280980430938459,0.057279487220468495,
 0.6757582725202625,0.39393215762599965,0.5134552045227398,
 0.48323132282427617,0.21066198173153972,0.9748997834451495,
 0.8961892470403305,0.4631514481515123,0.29852439638445827,0.8064394603402357,
 0.477329857003691,0.4264007326830592,0.8694032355077677,0.5100598453050911,
 0.11379499044191632,0.3023431894633992,0.4543311873346212,
 0.38459034458864794,0.3429964807500526,0.8888233982588276,0.1979153256063435,
 0.9405101527035769,0.5117499394413051,0.6729815210755133,0.6011682076482446,
 0.5371384348187386,0.7203890712169915,0.884764774968642,0.916794662497806,
 0.6273182472740622,0.8900174150263966,0.7349850132117455,0.9899487021899787,
 0.12302549030764343,0.5123917174474611,0.9476792826674232,0.8518494198937485,
 0.18633166894846898,0.6857343012804926,0.6732476020793671,0.6976951622895113,
 0.31296823908089433,0.21254719637828945,0.27140668234986765,
 0.06811973924101167,0.9687504355346673,0.40779747742967865,
 0.47213694668770545,0.02146365163464037,0.04906069096865462,
 0.4614105614511952,0.32241858096318665,0.6692984443153377,
 0.02436818081688996,0.3240402582845132,0.04107739125124199,
 0.3346566088217654,0.03877677969844617,0.7285084203201524,0.9527109373430012,
 0.1710629430084546,0.5093192039923548,0.7516788697136639,0.6724523937027306,
 0.9116573774440816,0.9407586130314762,0.15927123623736017,0.3591152567423954,
 0.014005425674884164,0.287427263944266,0.844073315886272,0.6790647873644033,
 0.8536881692212448,0.4371212396738059,0.8643770232991521,0.11430879007577555,
 0.5694073340590383,0.563536991467976,0.7175103420234232,0.9155685141153279,
 0.2665907244025425,0.992758004693992,0.9956652174936718,0.46102440903607445,
 0.2576503336992403,0.9112002907181072,0.8335074909595289,0.14462026716803345,
 0.05847539496135734,0.9279378391894351,0.29560508217083514,
 0.46488363209513806,0.33990509672338465,0.35278671183973864,
 0.4098371782205942,0.7137805204291703,0.7845221489131665,0.6615936702720051,
 0.20616029725347262,0.8831876532485184,0.3375216588348614,0.1615163240064058,
 0.8842720940416264,0.5889582806991489,0.37232711595479184,
 0.06548828112220284,0.866264136453132,0.7745191632290322,0.8701034185000254,
 0.8886147021896531,0.21930333394820178,0.5078276696274041,0.4622502740336629,
 0.6759988826218291,0.44100082579176836,0.08908918989709702,
 0.11388385003226953,0.11970872794923748,0.7226811872665209,
 0.7845013895300378,0.9712333362555146,0.44190059011951777,0.5922534227757709,
 0.08646691765733927,0.5756047513600587,0.4313334696561888,
 0.38304018841738374,0.6875303650952627,0.34524117055384973,
 0.12447690926707089,0.6209211931250762,0.9932648057993365,0.2206273311861945,
 0.5444149776554662,0.2521466420406089,0.5346876932459733,0.08383114701729744,
 0.2127700233925771,0.8418870764634201,0.3957656859135136,0.8588061016751438,
 0.2790463846975809,0.5276038179830482,0.8506722656388345,0.26211850198585973,
 0.1521958376947442,0.2966923513021954,0.972812700225635,0.7715222006388838,
 0.09576988935387543,0.3519326130816155,0.421625361291705,0.7514365131175953,
 0.10902866638128283,0.4584610467084135,0.6649285504214396,
 0.12185551611987833,0.931977679172048,0.7151506273895073,0.24092587840032975,
 0.9099415204055,0.7762430321134554,0.8741626682939518,0.8399318963222357,
 0.5590698143891881,0.631276697738546,0.3354623388397383,0.8141095022960085,
 0.9269964828045408,0.23902119119253085,0.7774956320049194,0.7649693670517997,
 0.39920455713890135,0.15419703131881723,0.29595735350853014,
 0.14067682411164295,0.9120443843807604,0.6044252893747768,0.970211757291436,
 0.5045649439229071,0.562800552749973,0.3772971164120349,0.23288833622059024,
 0.6699625664308589,0.21390730084845866,0.38245377513763223,
 0.8734164828530273,0.6235834154391343,0.8042304959916147,0.7397105923632296,
 0.906018073231704,0.9033023560742424,0.6703062715638219,0.19121035812945153,
 0.9766157481474808,0.0868036171084603,0.9234551004250696,0.3868411438232874,
 0.7971762151192604,0.4489020681197101,0.733943403932946,0.3712327370811448,
 0.6813748777826376,0.5083464477896625,0.8509360998528256,0.5143333902372919,
 0.02861306184061152,0.698723742462055,0.5847150307578233,0.2742804569671232,
 0.5543058143879779,0.6935671124140284,0.23369881209273202,
 0.22549474080213783,0.701018942423515,0.5160990636080038,0.7282180166657455,
 0.11267864708861985,0.3612748904414145,0.7339693963868126,0.7121351477878335,
 0.5322812258807064,0.5046843345818071,0.5071302749626785,0.5960321992609539,
 0.5000407453001652,0.8020971555747873,0.540086549021703,0.7198298819342362,
 0.6760222229447822,0.6612319023224129,0.8782590362542678,0.6070983503161327,
 0.8102039530399425,0.3546780886254427,0.8435369844335247,0.19167225918987574,
 0.140020006487246,0.2774698850606646,0.1896684937119969,0.7671372713689883,
 0.42039058314860467,0.6136817115522303,0.7874894529056318,
 0.47660109066052436,0.9645150123217688,0.3900650098957983,
 0.29724187044100314,0.2291680255286238,0.8958798081592577,0.8091829890486584,
 0.4691792272028774,0.9750546447156417,0.7423874323969712,0.9565014421357233,
 0.42884584221921007,0.6682902669146127,0.2223199728232902,0.7684526979297464,
 0.4941240085457963,0.42155390116501246,0.7110305601119696,0.8113696176176961,
 0.9270887394485311,0.6541087705590944,0.34426882070444065,0.6405635629607807,
 0.6416195714808126,0.26529271709195024,0.39305021247381944,
 0.8505212502789661,0.441513388227448,0.2663796356089042,
 0.0012986713459965937,0.5460175753607155,0.8858469934929368,
 0.7347695361492014,0.28259162860870035,0.5120334866118325,0.8636593987490712,
 0.5585593878860518,0.9701361280280055,0.7540392507890616,0.5145520772350405,
 0.22154224341780326,0.746608780291927,0.3081775330853198,0.8086284581283332,
 0.11669423490760622,0.6432852293459576,0.19128449889880095,
 0.24189342081127174,0.3739935221276903,0.03014234206819022,
 5.724546319442192e-4,0.603985815003168,0.7578230584099318,0.8882432541719327,
 0.7664799361651299,0.5480270976682732,0.6181723183364589,0.8328030341175686,
 0.4295286868037237,0.17017087331879954,0.7232831974694888,
 0.06704036456877349,0.0922576247853274,0.2600148385055918,0.8323062343847628,
 0.8931465534625049,0.4814413513974677,0.5667658684117515,0.5261415860395107,
 0.15038265864391787,0.7300921611262365,0.04949302209166273,
 0.8321343110081315,0.7617472612765552,0.5976761690003057,0.970671030705671,
 0.34985469356853893,0.5672345320229886,0.2587480543122286,0.755695831286504,
 0.6075459807846331,0.7192343001988143,0.5045379066853591,0.8752406890258718,
 0.5426263655558095,0.46854616188756704,0.11888476228771228,
 0.5731699712089116,0.648822094857922,0.9258594080176006,0.7863232725864431,
 0.9023133689467984,0.948692382586655,0.27047930546401244,0.9662707111278825,
 0.425288888637112,0.3914941488181132,0.5715493324487374,0.3018146354296696,
 0.7356921911691985,0.333169866590188,0.4227201886463884,0.1966539353938237,
 0.9579372830175542,0.19472816662397296,0.12672157506937076,0.419259255016753,
 0.1886994773138463,0.17205323912746662,0.9437535048529624,
 0.13619115856602781,0.6020368167408809,0.5221212341403241,
 0.01950300055291221,0.2620662221227419,0.2096547060747067,0.5078936742172808,
 0.07807507887254028,0.4604469008490445,0.20036553827482406,
 0.3191177717811907,0.4708749093473861,0.33909760436383896,0.5689361458062844,
 0.4878799574406997,0.16206145093057844,0.5759747450777515,0.5090717649155587,
 0.833902287820335,0.4793607923007378,0.842185349134756,0.5198433358715263,
 0.25349633347841627,0.6738126680455848,0.45566032975888715,
 0.2625849311974834,0.825996676922088,0.8695462114784013,0.9795977303496116,
 0.665497980461665,0.47003460601995284,0.9712096632202846,0.7700531163035891,
 0.17311793334732917,0.7915212253241747,0.5664852317710953,
 0.0074757694251212925,0.5156688208167594,0.5510130185380995,
 0.033185649602278655,0.6800521712995632,0.5883064550687671,
 0.7636593586467462,0.7035434548501646,0.7152299404660742,0.2313729267996819,
 0.871527980277523,0.7369519360123594,0.4019895092304471,0.9980199195571499,
 0.5285054500151122,0.012577285583997733,0.6449820515710825,
 0.6613683543988195,0.44611359082326885,0.23875391572307003,
 0.7748404085240759,0.5972421027181464,0.15443574668790616,0.6559485751717057,
 0.0731862979728215,0.7749663526989873,0.7372807275771072,0.3514779667866267,
 0.2304862182819113,0.09593691099493196,0.05407820064591906,
 0.7584080694520037,0.07031853314293657,0.09599332429530316,
 0.2948278558561659,0.09416680279579781,0.8354144510350848,0.2065391347128207,
 0.9688065106132124,0.37649522982839656,0.5091396717171722,0.6581662585395669,
 0.8791644271571739,0.4182815335107192,0.9696137396050665,0.7991165370905401,
 0.394100183019955,0.7696721668211859,0.10423101845336957,0.6004367856377778,
 0.993037952372574,0.08133757899587224,0.9376947407126548,0.6120452208943055,
 0.08872674594450503,0.5327418553882844,0.27150052110656153,
 0.9017755519686621,0.3549057646869729,0.6654531626001974,0.4602384804190772,
 0.9046824499233026,0.2955474279952034,0.8286131243027054,0.4457826472645423,
 0.8948386658733191,0.014323022120549589,0.5288756914138109,
 0.7661012503338424,0.7926740557501623,0.0483769819820381,0.5247774761062217,
 0.40771082412941717,0.31597311799423855,0.3096208680399768,
 0.7951251582259224,0.03433897418159593,0.2580561373817525,
 0.14689502581872127,0.921179792897583,0.17631666903631116,0.4334436195478486,
 0.6528897849001714,0.4075743616259707,0.5371316040041032,0.1787242139671701,
 0.534788012443235,0.2790258251112746,0.6070507274687138,0.7836336248458212,
 0.0507946999117308,0.6465132839750393,0.9226607361433277,0.7994025219842626,
 0.9214738977629706,0.7287766111709213,0.25544423477269673,0.8676834876503939,
 0.7148845475868847,0.6894093879943394,0.26037414001716064,
 0.07180095244871376,0.17744316622188905,0.8036521862327846,
 0.5875932418609002,0.5899710755816372,0.7512914062865949,0.4416930878423716,
 0.7257695108396889,0.6794732900032731,0.3962524682937898,0.8816075738652852,
 0.12995116622272684,0.06987604802656655,0.8575093123516312,
 0.6556055550877498,0.24421710013581155,0.6220619738188826,
 0.15441618390067013,0.737999636754197,0.37801576274794624,
 0.10602662054961476,0.8054196404986361,0.9049155028635241,
 0.22057495326630805,0.6139415652493833,0.7385646821518795,0.4901470215473869,
 0.4593953702379615,0.9421721837595456,0.7524510245991156,
 0.054101059869820944,0.8508020486565795,0.7317759057090641,
 0.3109135777438662,0.9518322165698428,0.5417530776507522,0.4701156906998367,
 0.8315410046305107,0.5036884220194632,0.303841699894271,0.8048778526563906,
 0.8552842361856179,0.5760147421878725,0.5251872908947216,0.23653346423396238,
 0.6405831850834248,0.030754876260185915,0.6579114155201075,
 0.5883048174809196,0.5853086657726755,0.2865532115244338,0.9297000327888257,
 0.6627248498654765,0.1964679540820955,0.8360692206035244,0.6256293997313592,
 0.9912693934082393,0.5769477915924572,0.8556975464365713,0.93582557038156,
 0.6526780446829882,0.660068899201199,0.28781543382384633,0.4255763128214103,
 0.35220414257182353,0.8523439275433876,0.18372032576802444,
 0.9332814744036809,0.25631032201259474,0.11489740475767263,
 0.7539989399204389,0.3772598102381034,0.46089246117397564,0.5728861512869948,
 0.4074942812711303,0.04714931832269509,0.36764797467074184,
 0.7724822808778536,0.2762204232970218,0.7540803228767805,0.9455927174089205,
 0.321588696759343,0.7829545453098512,0.9457928047432052,0.71918348816178,
 0.41333324660855775,0.6063384239400819,0.448192969792809,0.2615500581954422,
 0.201001530302099,0.4721292068264432,0.47489021770062456,0.1395646539458827,
 0.557877217788846,0.9044940308644095,0.7206580114765362,0.2871836335675043,
 0.002350963397165229,0.13417608768912226,0.22680591835185104,
 0.8011553786096598,0.06622148330174293,0.06438038771473664,
 0.8544746708931312,0.17462711266786668,0.002159584019331584,
 0.2137025612802148,0.3840933366851146,0.6812281929218676,0.7902062812416626,
 0.5597313196520597,0.7621102347639749,0.7106631800753161,0.5383002805488687,
 0.410333949569023,0.08642059007713221,0.44345279520206526,0.5530866610789106,
 0.19545181670054235,0.26168223146584924,0.5478878769338331,
 0.7835265617000011,0.11442220148286186,0.11849486939065057,
 0.31613942886598867,0.9629713944937418,0.6081177053492965,0.9530960648466895,
 0.005922412404092747,0.4481501967654581,0.39097627630970777,
 0.5171770276738866,0.7023796161556596,0.5490096930674422,0.39753972273381133,
 0.15471273821084341,0.33772415357270213,0.6330072949329318,
 0.26469204088258047,0.9379285204309985,0.4834910829793325,
 0.44178558615252295,0.16517944241360583,0.3357326886451636,
 0.06857899877843243,0.590054271499698,0.23358610829477544,
 0.16081660249439178,0.25514299897256254,0.36555945629326025,
 0.44604806926787965,0.4098419514993056,0.1308007459830286,0.8220500483715736,
 0.8399609444435283,0.49294898807362764,0.0025395909811584128,
 0.377605010535472,0.8137050212612302,0.7649043211774613,0.7525425445088725,
 0.5603965422252757,0.3495673860246544,0.582355827789466,0.9691363013460303,
 0.5682136051872714,0.4565626460908573,0.15376790739929058,0.8747744690946989,
 0.9463260565993394,0.49249326581069,0.9290026761716589,0.10472209343163619,
 0.31709766490971025,0.9303413946526937,0.9344778464221012,
 0.006772013726072057,0.5497746419164977,0.042696559147558544,
 0.4943375924144783,0.0506804323084048,0.43801527242904026,0.5174415635404318,
 0.3500162763360043,0.030914951857154427,0.17899221126630727,
 0.31224422650652495,0.5994616910736309,0.8155719597081358,0.1671844111369607,
 0.3746802687951464,0.913276636842081,0.3562959459785626,0.9088140694190465,
 0.4999944274546463,0.977842523909814,0.0953130454461607,0.8808822690375714,
 0.08969348840981095,0.5072182536291386,0.2285456442381274,0.4547496818429497,
 0.3460507191553712,0.9594833501735426,0.7655551681036239,0.6603712122252775,
 0.6904226440658519,0.905656240841579,0.13978942089152235,0.24263645500555886,
 0.24016122222071723,0.24260459997628692,0.9471337666536467,
 0.5758354828068026,0.6281396909414876,0.730153046283149,0.2546157167686084,
 0.050099077631327926,0.6620444190210619,0.22430181872109556,
 0.3730687615683138,0.9522685980771158,0.48064254361631686,
 0.38025969956662564,0.7280917729489657,0.005849094610643868,
 0.14810166098825805,0.43893444213263866,0.9684575649638145,
 0.8355526838728812,0.9595503258122895,0.611053379217573,0.2829912157590433,
 0.5674690942338828,0.24496808327423436,0.32937766919836475,
 0.9242622365808126,0.3769120535144457,0.8408354849852007,0.9535236369734417,
 0.0018385650972858603,0.3688052219060768,0.9434356325149185,
 0.7128808933070887,0.46246788339485556,0.4026412588967222,0.8527804354684948,
 0.6081080386643637,0.25732940241673496,0.6470194059642282,0.5058566920877968,
 0.5334000241490501,0.6875898820678864,0.9673917876984501,0.8970484308660862,
 0.756964155891512,0.18042020183835605,0.13243392037251245,
 0.0075183350974568075,0.4183608153571615,0.9916812645104521,
 0.7787693218814459,0.45158177310418113,0.8973478047850786,
 0.07811704023481858,0.09237082125375773,0.4191532940464521,
 0.1635266018499949,0.2785126597691061,0.9886633814168987,0.6235639692336592,
 0.9312834370589973,0.7320531256209658,0.9194505219575956,
 0.020704682044821676,0.6904574346169319,0.02536135266607842,
 0.9370392039366441,0.11885926586083517,0.4208601113616035,0.4884087239525715,
 0.642794390956112,0.4989403793384104,0.3769819470363235,0.9139633192723933,
 0.30655132889704506,0.9611987803484731,0.479461569589712,0.3805749947252349,
 0.6841026549443261,0.8138807134481599,0.6789187143469853,0.599946439930675,
 0.1290076307247222,0.17063558721702188,0.7124241890806757,0.3425638734426182,
 0.4950761789049497,0.64500965682501,0.9426589410154277,0.4194894681488148,
 0.9954962528751845,0.26140657149515545,0.6355107862357889,0.5440652340267814,
 0.8554186986278774,0.12460633021412448,0.20957708670451203,
 0.5428295156187066,0.16129654370045698,0.2366142388827841,0.6021205197731905,
 0.3558041837783563,0.7138032682833134,0.7536705448642015,0.4326966824505992,
 0.6038463897134931,0.3728053099224776,0.5127208035502334,0.6521205898564442,
 0.3890078638893355,0.4146473114489362,0.03064365146524095,0.5985045642226126,
 0.6799073901583753,0.4105065214770667,0.8021659366774564,0.6653093012310686,
 0.9102155633732439,0.6746538254673923,0.06133291684592246,0.5258253464862285,
 0.27635119467561664,0.2643972440063713,0.7065502944161239,
 0.24121675676605203,0.19893347159362207,0.9827382477837555,
 0.35075564776194923,0.3475447813419693,0.19146293132708292,
 0.7816603690138935,0.6415793269791987,0.947519121682587,0.11698728642350109,
 0.3451960090933994,0.9775180730166282,0.555268737788503,0.46404178072708213,
 0.5728909559021393,0.44930424754584863,0.7958998906033918,0.4330127408677217,
 0.9580745283711317,0.09922139576583855,0.8773046010980848,0.7039240707684077,
 0.8895415191910008,0.8171561109216333,0.49453856445714917,0.7769626303893878,
 0.3267250938080052,0.5488017567364578,0.659994255261059,0.38744447120844305,
 0.18633333248230322,0.32985943719151956,0.8850515027940096,
 0.6031603448379752,0.6377117720399526,0.35008211819130053,
 0.012844867959377448,0.33140757845947566,0.1301049087520494,
 0.08098292324540068,0.10035139018495187,0.1420109177908485,
 0.1777529174752357,0.4613687727780217,0.18395344482900822,
 0.19403483972866187,0.06826948889262496,0.8802630312002866,
 0.01521449692755783,0.11398871014747558,0.9818456038491096,
 0.7028336211171178,0.9089755959034707,0.45592861941442064,0.6690188022033338,
 0.9029795191452805,0.49247242403961866,0.059501778566518504,
 0.1311425886337666,0.5256890762468264,0.8960153693570989,0.4265178545447269,
 0.148741529854726,0.04194453731557457,0.7694904131025222,0.4688849254043578,
 0.2371672543679062,0.27377428421935557,0.6434591517055164,0.4671663723255258,
 0.7511483716389395,0.12554267179917333,0.9977286945284156,0.4364530390433612,
 0.29679953403291237,0.7197003190980499,0.5783327687516132,0.4545702221191106,
 0.463896664613778,0.46896865119013786,0.6838442293328387,0.6293801271989179,
 0.060366484892142025,0.9668589933806232,0.9473878965611474,
 0.1569336787295007,0.4643296581001919,0.3142412615199712,0.6316535939721143,
 0.784650219975449,0.679689176031909,0.8162582862726846,0.617309623536752,
 0.8999266506345571,0.5261881372182509,0.5250485268713811,0.8122522969991834,
 0.8616599962934526,0.10254712745728223,0.9954189205904037,
 0.09681051380225614,0.8847785539342145,0.14494009061141588,0.42861953541332,
 0.35406282169519254,0.10245727382526426,0.7028241990559155,
 0.38025906882295923,0.8269637608656375,0.4097566162671994,
 0.12193720073273462,0.37141938037706834,0.5903106396196689,
 0.5623451664316479,0.9591163160242389,0.5615616313313081,0.7196590195780288,
 0.5316811949048311,0.5977002784902709,0.28552101759387893,0.7103306134551373,
 0.9197209357221515,0.15474167536415517,0.7164143480214359,0.8131226315134007,
 0.8150513610338506,0.8269279110709918,0.651256809653758,0.48179173279578213,
 0.9606589329507622,0.8268881668538883,0.6474161175412193,0.780745058347524,
 0.8271788927341668,0.7563374878139953,0.2925763348022208,0.06828461310763012,
 0.9583491926749681,0.8517509960461185,0.027124331459597917,
 0.9630340944343607,0.3053709459507543,0.1159721378260985,0.9635930635991371,
 0.6256623717888995,0.41091500139810444,0.43409206768182906,
 0.8602542396519244,0.9135455787648883,0.17965862384142617,
 0.06025103287706102,0.8945402056289664,0.37097366408033006,
 0.9229644236662669,0.026146413942954583,0.9738197235645502,
 0.8404890778584088,0.8654499306910483,0.45039359524163125,0.9776840705306095,
 0.16292340496198476,0.38946420955985794,0.03672648388780331,
 0.6038317053179635,0.37254999027805,0.8501943564347252,0.8348458930473595,
 0.0383822756078015,0.9056582472283462,0.1878500033549696,0.6483819421574548,
 0.9844507475672866,0.5990061392200423,0.15660453704631316,
 0.012556591597276956,0.908051423387328,0.6365960191802146,0.7534869350484921,
 0.8690578885023585,0.274701130845573,0.6061882866456108,0.6858161087365346,
 0.479810467179157,0.4804258628300828,0.028277942162767133,0.8382085695138296,
 0.1649934054425568,0.714939068963782,0.5211934484009474,0.28356740631832955,
 0.03736375565768191,0.11081974812871609,0.8872917579978394,
 0.9876770171035417,0.7048434436041511,0.6512208756062537,0.12403095024994815,
 0.39555110708382535,0.25473571383797333,0.28765621265674013,
 0.5549045411158271,0.6285048983846133,0.5713852435041636,0.9333062282512319,
 0.012607173298797125,0.6484730123721407,0.5713615316003976,
 0.8687278372229441,0.11959527587907981,0.09549191986578787,
 0.2793586281635677,0.9256698439114954,0.22403754385987684,0.7499148899119316,
 0.4931590116253908,0.9732742571537385,0.507763849733996,0.6122501698609225,
 0.5775006003154038,0.10204669318728832,0.9012984179889947,0.3823176172948505,
 0.0021462072447027136,0.8018660410409513,0.7343311441057161,
 0.31447460955402773,0.7104809412928836,0.31835397610381944,
 0.48296190561970376,0.9673321848031362,0.8401612438712367,0.5295840278731421,
 0.12949939102313712,0.7402882075793474,0.2732967027308968,
 0.33142172264218517,0.4875535257426208,0.3662558247929938,
 0.010161361890272613,0.8869379749769603,0.2700742908439343,
 0.12612661139504178,0.4987087767177697,0.5579523131853468,
 0.27972976557776985,0.25486794069545693,0.7377753866993965,0.909740220814175,
 0.39585066064486907,0.32289554114847263,0.3016646615644437,
 0.6214577283793499,0.20755694646939848,0.3992280225005811,0.5871024393163484,
 0.6820319524814966,0.5910258716452529,0.8931665490260509,0.39568832103660234,
 0.40951999539934647,0.563609580180678,0.7901095447613463,
 0.059106825533048024,0.4156900238828498,0.22903878304156555,
 0.8344696857510833,0.6670484130771687,0.5879468919294891,0.2768744690343473,
 0.8111815592434991,0.7572754703368727,0.3847038952144288,0.0900079027906403,
 0.45689110366645536,0.739073795515836,0.10683544034287251,0.3984105562578075,
 0.3240950606575046,0.3299594733696164,0.8533822090768395,0.6328186547157065,
 0.8867931430386974,0.37905860619189746,0.01778148989247641,
 0.3513074370199911,0.25457832834375427,0.2296951242522698,0.5999950575910652,
 0.4421376838078537,0.7157895782864784,0.5317563355333937,0.1594202614188096,
 0.7527205785440272,0.04086369684143354,0.2340814297982189,0.9862981293731833,
 0.8805489942324416,0.8908957310785752,0.05534248669770181,0.2764655853384925,
 0.8985754411926307,0.8010861548063366,0.3911943368408799,0.5113523573105416,
 0.475496261472552,0.44727114706580373,0.9549771413846915,0.16304047300788482,
 0.6181542998622181,0.5093726987414177,0.8111787348130415,0.09630770417210022,
 0.8640695881147674,0.9215837280172607,0.26174563603121204,0.6981801936204899,
 0.040435525736905675,0.8702722014634999,0.47613723721041745,
 0.46414441652355354,0.07594655883662604,0.9190289632587465,
 0.9590616825953515,0.6441776104122584,0.734720044642871,0.3068759226090395,
 0.03322400295535377,0.2907446288346216,0.8051572420650757,
 0.05131338483776193,0.529388800100376,0.5687205833581941,0.28008484458251837,
 0.5132353070534195,0.4925005740439894,0.08962557927108039,0.8768223366871556,
 0.3211603857837131,0.20549857818827044,0.40824866055718434,
 0.7450210438565625,0.7705229673293595,0.4263100005338788,0.5457729448812267,
 0.6235489001602278,0.7035941274718057,0.1289360596640703,0.6522254884238264,
 0.43090037223923305,0.6160491616099298,0.7698088669524273,
 0.018283341637854056,0.9992859720161709,0.996480690782942,
 0.04360011870682934,0.2898942441967862,0.6425894876496715,0.5709637307618832,
 0.31742927993756165,0.39555032858965367,0.5734259359481848,
 0.18741719109036925,0.31821994066004444,0.587659926131684,0.0200707383027533,
 0.9166848534513559,0.7860842665122093,0.40750839898744873,0.6472262586816714,
 0.09747082940932372,0.8443834651990008,0.6736320094263185,
 0.05634293714613159,0.01291376358696028,0.4909534267536887,
 0.19131733165060727,0.6145492817412957,0.7139829987496069,
 0.04979484073566165,0.8713983504437117,0.687211388735203,0.14296795449133803,
 0.03472674294896749,0.6471756151333283,0.08723271575210856,
 0.8777624419778919,0.30685032850488925,0.752042998539545,0.812491595422118,
 0.25398073621369466,0.5269457857019353,0.0881702076210813,
 0.12930654821507537,0.18022712743031732,0.05740579699191606,
 0.8844639191758287,0.44507178793982594,0.26322785799406845,
 0.7590316102174399,0.22931861687160915,0.5746275988573,0.9573621980819563,
 0.8061313290668084,0.6724713670175451,0.7569198096549329,0.5823044223501004,
 0.6939287780363145,0.3036103424140112,0.2676861111048282,0.9060681922314849,
 0.4206642176260056,0.5237553208719959,0.47650729632559496,0.8307524718708342,
 0.19922782265383876,0.46110871165432654,0.16318952618537708,
 0.046883391443995226,0.948673894299639,0.3634244668831802,0.9123549286973568,
 0.18352231823890774,0.8142681280824451,0.6495194726005934,0.7142255885772791,
 0.03206584451324701,0.8321441895029873,0.2291029257934748,0.6539087777819088,
 0.5910717047026618,0.26967875386409834,0.3045803899292535,0.9458995352973937,
 0.5344373287947909,0.46950967207903105,0.22587101910154017,
 0.3351999037984915,0.3578467289575269,0.24373817836784117,0.4343991040703443,
 0.14398270831161242,0.506908686308716,0.7254149814900067,0.09945741156639287,
 0.9443183463687611,0.26240252438843914,0.3870270043825462,0.4068918705704201,
 0.11971953120895917,0.40970511462868364,0.27395966452377485,
 0.27705847306561127,0.43935304664522856,0.720311609941475,
 0.33125793439793605,0.04601979523670696,0.2149707976696954,
 0.7904228709184979,0.14201483958477956,0.8700377078359165,0.8396431580221527,
 0.8952156229234819,0.44068793802418227,0.807991992120313,0.4769473140571916,
 0.7442093595851077,0.2752443417589463,0.6727070700504306,0.18060942669938984,
 0.34661175920398846,0.07140615307755915,0.16216958254668112,
 0.5840461508934273,0.6783687943309222,0.32918680643869314,0.5153424946934598,
 0.4635743905279881,0.639982796611978,0.628384085841951,0.007861828480184574,
 0.2410090429653371,0.27910865267056273,0.7601332301993133,0.9887514033639327,
 0.6571004176095437,0.6802472094850492,0.7949654169365352,0.7868905802981241,
 0.34229868001346253,0.6849037564353201,0.2548258634162758,
 0.08135564244773608,0.8604284206535519,0.2388610435550963,
 0.22701227243498046,0.5530335025545982,0.5661066686436497,
 0.07596232167876493,0.7073053073806761,0.09279666845741963,0.472212617967267,
 0.11387033857882234,0.48684184204184433,0.2826797238444969,
 0.5038194368577997,0.13778426088887819,0.6751820610075595,
 0.12182680266708923,0.46297986326605267,0.15827163423862922,
 0.22008742093256672,0.8682569835742838,0.39123854892473864,
 0.4616254113565723,0.8577102721445173,0.7587401845442142,0.484196079731946,
 0.611063897162714,0.8637442996196468,0.19194552350130678,0.4063938811656014,
 0.9692134452705413,0.6819735808377382,0.4299217815596199,0.369450880909755,
 0.12805066105951224,0.30846042123901996,0.38225541042107114,
 0.982519032309341,0.09591619603885526,0.6330363873147484,0.17188443350367388,
 0.1741411058573311,0.46575426944045306,0.5514372559914538,0.3534052932081224,
 0.2658959033845727,0.2619505085835935,0.18233649856851386,
 0.27456161843391336,0.4683702298427549,0.8186888095301684,
 0.47308430570199356,0.28534012103145623,0.2740021107088012,
 0.2853104346223585,0.36671685577890245,0.4669475154889051,0.2656357576972579,
 0.601263022432673,0.4266297435396387,0.40057227570004894,0.4295943383734061,
 0.8260230488558039,0.43762155591232665,0.7807176679450379,0.796806409721127,
 0.004709893304638668,0.6608860990706598,0.5168986891714578,
 0.27934634662452695,0.3748550560195345,0.010612553101781552,
 0.8647130483518723,0.5733949151807071,0.5893135347702688,0.6868957506875288,
 0.37083461269998796,0.4956952003226204,0.017586119622139962,
 0.7808089883820354,0.8298020414755214,0.9207091964207615,0.05460437947715868,
 0.8102626289044632,0.5599094255131971,0.6458326083781378,0.9045313876500252,
 0.08111565181645286,0.9515696633558264,0.42087671016258865,
 0.013080229327078752,0.27699259836480405,0.5120316368591322,
 0.1549198909204238,0.3703491669704322,0.532286076587845,0.2486704867623608,
 0.8238068603730752,0.6064926810912955,0.19357639126395287,0.5898026475413962,
 0.1852030666180301,0.9893037175363452,0.7719988884091264,0.24891181278147612,
 0.24393562423510096,0.5607191128955619,0.9295712054685515,
 0.24453886489934507,0.8616516462791216,0.054383121472745266,
 0.7002633426920836,0.9478498309242356,0.2866906313960519,0.821244128320146,
 0.8768370976980866,0.03455523945432937,0.7244175395014217,0.6903899221385891,
 0.25848903334070006,0.20584775568814695,0.010940478632001205,
 0.3312001657992317,0.22259723725795288,0.22915775667050853,
 0.7019861911481381,0.8813997171768555,0.09905164144903789,0.695135140696597,
 0.01177012817378531,0.7995215104415812,0.39413400707760926,0.830486998996229,
 0.9160805611289446,0.12493020804620303,0.6932719269647118,
 0.13929943037425307,0.5255088790789956,0.20363661525595478,
 0.32621776932334656,0.3444720546990947,0.5886700502486112,
 0.23108378976603094,0.697012201248931,0.8291884261123412,0.4809982961327529,
 0.7705520749239126,0.8141156877153178,0.06494847885214095,
 0.14416803288108904,0.5376223390418904,0.4963319848614276,
 0.12899474651752063,0.5389991500765215,0.8732618236477512,
 0.35106572787045964,0.09141537193646176,0.010644287603330804,
 0.17728917453320925,0.9380135655774884,0.862005333018291,0.1593913942796511,
 0.11745009240871773,0.38269626371847554,0.6712297525190192,
 0.17550850167434473,0.8114998966330658,0.00918666326052009,
 0.2948577060735533,0.5851561895603722,0.10809526338920983,
 0.16596281215857644,0.07948061063658374,0.8510896888454494,
 0.1323535411565815,0.11348401080590853,0.2951487076434045,0.8250318351356161,
 0.46944905993288155,0.6218159263379712,0.597757113596743,0.20268256074999125,
 0.23454249209973388,0.08809729789172127,0.3325552704409468,
 0.8879756551645172,0.7632169821644671,0.2918395457943388,
 0.0017249268170508314,0.6369327828765222,0.5160439873674276,
 0.5138844087485281,0.22721320401288414,0.08241394812168124,0.913614910949648,
 0.8189286905718277,0.18305910898533329,0.8924255231306927,
 0.44998428389747824,0.4440332968839579,0.2826605407147653,
 0.07279663334189568,0.913899176372368,0.8770774661884735,0.30110118335111613,
 0.2957009957473171,0.619079561619263,0.4554629675365439,0.12496876893839703,
 0.5384360863660119,0.07441439569948605,0.21205546222737692,
 0.34266666257564316,0.029424332422843258,0.8765961570179133,
 0.5745055462071254,0.443287361295681,0.16826231171662465,0.6455832565504003,
 0.01787670008057274,0.49602566868212716,0.9445854164366566,
 0.5182919402657831,0.5446944656956338,0.6543709243463888,0.40976253244362926,
 0.12741845390404527,0.039079093165674816,0.47592150070150885,
 0.6114121094759047,0.3245053869561355,0.7439661790889733,0.11523704023927472,
 0.5677581303406427,0.5001703259986932,0.5058575218643023,0.22997718430823244,
 0.30885762093466795,0.07219123539095329,0.3383061129440691,
 0.3948766152299308,0.4379585175333538,0.34423253416785393,0.0962713004044563,
 0.4812071974005695,0.20691472351324824,0.40284206148991164,0.644391902810259,
 0.9874158059692254,0.0917576446899655,0.7123234224922459,0.5437637113759187,
 0.5481729457189336,0.9542788623382792,0.8283298177734956,0.23220501036126528,
 0.943118248823099,0.6889306788520572,0.5851008746996919,0.09404919227565012,
 0.6736169873396624,0.4200017856895545,0.42118151825775874,0.8764887690404163,
 0.9760733186930101,0.316793750170532,0.09983320944884011,0.7976419979830123,
 0.43857112790361663,0.5618993280496946,0.39575529659983144,
 0.20293534865057838,0.545062741225335,0.27941968692853314,0.899150750550628,
 0.8366669813491364,0.8022599775227995,0.16674569809848583,0.4477592289899581,
 0.9905290038584703,0.7185984421335025,0.4136721488851771,0.7057610146121547,
 0.3720016805598738,0.630442980589839,0.15659143022749178,0.8939178971888687,
 0.6478851601608053,0.9221161755261953,0.11598631870355947,0.8081799386664884,
 0.5834703275555958,0.7261819628282384,0.02459615926312031,0.9550701398700794,
 0.45411134798252284,0.29297944926646546,0.17231891795454057,
 0.7799282927825826,0.1713286863647543,0.44570851353494023,0.6700608152802324,
 0.09752274337837497,0.1855331937279745,0.7345172912788777,0.9270168062774244,
 0.04941731327572951,0.8129856406148323,0.997152055514945,
 0.035652597156732635,0.013323167232811794,0.2070116614029286,
 0.3626408835043823,0.4147564642292536,0.560745523484833,0.3286684763080854,
 0.2663857877179122,0.8397919648232147,0.5508798274308143,0.42476324579688407,
 0.819242662482972,0.030112788857173056,0.8642837077014321,0.3875011743104074,
 0.07152626849367105,0.835968053706485,0.19618133703605545,0.9129634469475647,
 0.21453009785758181,0.668688311598883,0.7489151169552074,0.967589548702101,
 0.043834805296598,0.15076772375043768,0.10331800962833915,0.4769195237962349,
 0.07257551041702504,0.9730266200894191,0.12360194493280396,
 0.8196180371838858,0.8919433328603504,0.813398317209922,0.6309291152099882,
 0.1823062030829652,0.43174421996180434,0.7319712481641208,0.3905721432832121,
 0.6921546389863139,0.8835055730108367,0.5728349706854146,0.5764828882275282,
 0.7342841278206358,0.8215334225021946,0.21872333565109126,0.6928493031664038,
 0.7926446898292715,0.06685431763148997,0.02454492226745464,
 0.6905768571719619,0.5040327433587152,0.45526319925940595,0.6629267288891666,
 0.33043374814202486,0.34623193618768244,0.5290461704039134,
 0.8979002933615875,0.6450537375839016,0.280480799139219,0.945332169675126,
 0.21375822884872364,0.9400436247989171,0.41213894474680757,
 0.9989951350057533,0.5905864532788195,0.021749100438098745,
 0.08790476587793761,0.2498468988217455,0.6449928802075522,0.7490703780818389,
 0.1457165094059658,0.716825694138839,0.794050668602256,0.9513432003142187,
 0.41769237892388056,0.7144595224814292,0.2726240590019282,
 0.16673960494213724,0.8631364421537199,0.5650546323037435,
 0.29104578580899043,0.4091865691060338,0.9735742846629086,0.9313324932048738,
 0.18644637432158417,0.7743798332239317,0.3698437783071449,
 0.19289073053942385,0.7336762457581282,0.584696429323939,0.3592122844619742,
 0.3243567102602002,0.6052854724460152,0.08144380844067867,0.8685204874054849,
 0.814800747805073,0.24427666180340313,0.6385958769026139,0.7271287139980491,
 0.508372665501747,0.5299257884297195,0.41035266079267396,0.8589157664460187,
 0.6814609876790206,0.18403519294731807,0.36099734496099556,
 0.14514281886144031,0.8792391716686198,0.9448964824647593,0.3541841057979821,
 0.028216382860058897,0.6338717802830262,0.9868948327801607,
 0.8109905685805467,0.3149965406617712,0.92712252542896,0.8448045993710622,
 0.7872529769927307,0.8869854868664806,0.633568614629608,0.3153321334586503,
 0.6756851021292105,0.6082035346811537,0.13654571597428578,0.7943377363983433,
 0.34383980165999495,0.5816252067263429,0.34479423554701316,
 0.9026304893403021,0.29687922663292265,0.8007112409975145,0.7405298588205231,
 0.04883446823624604,0.8427241109696937,0.9936867194642476,0.274902771113561,
 0.7874093427593786,0.7413125635449398,0.920652199259695,0.053816232435759725,
 0.5655593301757892,0.0833090183449331,0.6865778516248836,0.8653580482769474,
 0.7972437175797527,0.9401334640111595,0.8113823066727487,0.7349852732177073,
 0.4774173042685578,0.030568666413411827,0.9070671747110433,0.786932053350682,
 0.4350050650664068,0.346375022299914,0.006403323184980936,
 0.36018577016932685,2.2207969082510548e-4,0.31654444586513786,
 0.5952174626347931,0.8994174494661094,0.9720298893162505,0.15029501172301318,
 0.620091510654735,0.9944233249163664,0.6292823771661376,0.6550294431302798,
 0.6054221673021728,0.5455502129786345,0.008286178606463968,
 0.41517228446113574,0.3776082147768698,0.04116031850504864,
 0.9903597437466785,0.28737062510606126,0.4207245325598401,0.6130514058503526,
 0.02209501129189717,0.7864022907736407,0.36737036821249913,
 0.28878165169178005,0.521268017053587,0.07123459114066011,0.9697486609084557,
 0.03941858003461207,0.3658699889397964,0.5657846462788098,
 0.23011958095008522,0.13931346354968732,0.9435491957787618,
 0.6943608375827731,0.7813024055613119,0.26404208214769764,0.9502264850474814,
 0.5227764518796191,0.12307636272029998,0.9136787332118965,0.747509450924614,
 0.4951352205907229,0.31911904322226126,0.9831374003861355,0.6086773836475382,
 0.5852658589445014,0.8866995409101432,0.13343723778139605,
 0.14578462096558908,0.6755504838289395,0.383141854739255,0.6496261494349852,
 0.49599944941936314,0.4746561810606553,0.6283058380167915,0.2504422535117603,
 0.6772122090542962,0.3946896537258515,0.8029062722586808,0.7302671859903179,
 0.7684726204521387,0.6363488044158583,0.5009292525577923,0.5252889604516829,
 0.5589294696977861,0.7546234846567578,0.5404545858531451,0.29637344294787704,
 0.4132865792841103,0.2630565296961529,0.7414890719688159,0.8688481232574533,
 0.5200889694428787,0.11477971792368513,0.6640711279686533,0.5831470413592943,
 0.08554785552395117,0.4324135402385896,0.6109636869588398,
 0.22059170884509027,0.8670595516389364,0.8125990582899589,0.9545876615462059,
 0.7436941255001415,0.3717502992698736,0.5536253556972007,0.7994122404989985,
 0.6048269908543731,0.2559774439142224,0.5511379410280831,0.5782941602076,
 0.8500799805946403,0.8990238856925044,0.49678460891661724,0.6459219985539351,
 0.03922791749469501,0.5022999121759363,0.16628816933235258,
 0.5887008112924195,0.2828512114348638,0.5468739352474604,0.16715364840665414,
 0.5926772367516834,0.4366344427986535,0.3440672782446631,0.9582559551963937,
 0.954649200110615,0.9816987240590397,0.3609526770654723,0.6915595539297651,
 0.7835366212244422,0.6738273055761023,0.32469320804412405,0.4249037625290536,
 0.932288816259791,0.16274577951787872,0.24839329319948267,0.9914754114967765,
 0.1554096916627109,0.4248392952599066,0.5738582430045739,0.784745139307154,
 0.5404954006721455,0.9072120615863831,0.32963425537352353,0.940453514485141,
 0.2731345248161722,0.12453728614866943,0.5045522632064793,0.709205002799178,
 0.9336020846469988,0.5929896318095427,0.11922406301285504,
 0.39616629090815725,0.08990836330070207,0.9776951498650347,
 0.7666059652979126,0.5959733294520405,0.9170326621263153,0.8005449682781841,
 0.4990146103679255,0.8005932206684713,0.3592300575515863,0.43872122809495295,
 0.21186998213466557,0.7387108386547402,0.8382686060911186,0.4433960009463165,
 0.9567662024189173,0.46429522738170537,0.2138400448463892,0.5044372147377348,
 0.8479198927875571,0.5492441594113486,0.06264939353229937,0.4881490443632839,
 0.9205655509935933,0.11884100375205264,0.5433619799759147,
 0.16324558868861738,0.42191049768044675,0.8983341670317658,
 0.7922145777770406,0.8686143295454873,0.9729448574729296,0.25090685144325464,
 0.6954116326784735,0.4165913730405535,0.2854403316608376,0.6506480123751198,
 0.27698010772950477,0.11923257140820431,0.09073921239755967,
 0.8845296013352586,0.5961818179529239,0.12546042358653664,
 0.08395321928856614,0.24095173192177277,0.9234252573053123,
 0.9486671643835738,0.9643489236435924,0.7053180005601749,0.7819563896186296,
 0.6862357174101537,0.6408017662466318,0.8243284201366257,0.5304944358657894,
 0.30876010499568096,0.02849175624722955,0.37508700512920035,
 0.5102938841052231,0.7900263071830723,0.49549807093113696,0.6637701851185319,
 0.0277790643673137,0.8644116888234898,0.05437849680267082,0.8447616319265602,
 0.8028632303619704,0.359310152754848,0.43542458604261514,0.3400017209588696,
 0.8702570175018455,0.11427246082681941,0.4577171626205321,
 0.24435641706786182,0.17437442155770533,0.07650808954802368,
 0.4618060125112148,0.09686922326473213,0.11394129851980772,0.653363603148068,
 0.20616086080325924,0.5064890380465661,0.650905881183417,0.5448126934444628,
 0.3688935017534076,0.3671877841115959,0.6144187595177851,0.6325904673125604,
 0.659112674195631,0.354652022877468,0.4176655559950929,0.8296343540208495,
 0.06957367597290143,0.6503312803585479,0.7743939551792041,
 0.24632680453814648,0.08678649979574427,0.758338391972404,0.983102761604424,
 0.8317821635459715,0.5298272721629633,0.4192036319814205,0.8032115046317918,
 0.9690090352690686,0.04208603658781096,0.09066187494865219,
 0.4526582030939028,0.7521343357651716,0.14971625426890967,0.3860733518681113,
 0.29301969759129776,0.4202904246467962,0.31393881404257096,
 0.5067798508544915,0.8949623270525896,0.3625393056943734,0.6564166900294541,
 0.6029921144739024,0.17550451243755882,0.9560433245160066,
 0.22945738506031788,0.5935128697774076,0.34862286550210486,
 0.08447950766422319,0.0670240732537366,0.6670474435485679,0.7474031041539538,
 0.42300378283975115,0.7093472151340862,0.6601466905815587,0.8568833762776359,
 0.7595832225043638,0.8150095758371904,0.694170223480687,0.43178738983299836,
 0.15896385649758105,0.35962986781874695,0.16511403864579943,
 0.8665566279119925,0.32216645824819046,0.728982735843856,0.8673258869670871,
 0.15387753378315971,0.8484586759670372,0.5033134858402241,0.5218806612358188,
 0.18711561863275272,0.8552853545775325,0.7421383500615611,
 0.44716196211397174,0.3031156913634472,0.7647526898352112,0.5729165629212346,
 0.10329980672560835,0.7335621356898019,0.7216476431176058,
 0.017463221885721913,0.9694283931946311,0.5558834729554571,
 0.46139591378930933,0.20634811403391107,0.9100057698483704,
 0.09034935056826443,0.5796384548964502,0.5571075345732053,0.7544858586334915,
 0.6065341388401567,0.680476217728607,0.44156791339549173,0.3712986157448962,
 0.5362807021354086,0.5029852421084466,0.09406857539287394,0.6583162748096333,
 0.7532914222309414,0.17562633930714922,0.2377607858769485,0.707284216231247,
 0.21644781052018414,0.6194484380307927,0.45412152353797486,
 0.9262082693377679,0.2609741988588543,0.5543238556469381,0.7475266093843524,
 0.012600615172617236,0.39610795451193126,0.003350374953557944,
 0.5163303987073657,0.13422618688722787,0.795747803591752,0.6719492026638538,
 0.9682701268180561,0.9130512304460451,0.9561077691357034,0.1338546343925563,
 0.25867012544639345,0.9724293204395266,0.9726019701732312,
 0.27993454829520503,0.27810004821578693,0.1575583956085611,
 0.6498925322387126,0.9884294891235398,0.378419252670568,0.030894301496687482,
 0.07886821968939584,0.6906790074768551,0.8068601357880735,0.5738318459384468,
 0.9702784342575709,0.38159017140343443,0.8415189972018005,0.4609134858605487,
 0.8263521881337414,0.1987947037031047,0.8238689122310897,0.47994020209743027,
 0.14045949070945385,0.00810553425134497,0.28224167978754144,
 0.00972163435620832,0.7583526492656324,0.2916888977023697,0.7001205210189193,
 0.7643547097516787,0.8870270015929675,0.26304730640057883,0.8295246446556301,
 0.706242234835421,0.11602658872296057,0.5165324569339381,0.8745501771910028,
 0.4585720322793536,0.15030846898961459,0.9599829611275207,
 0.002749743512381664,0.061067288514147,0.158321557951569,0.9806145300365097,
 0.3071213341033868,0.9318670715236288,0.9063953874261852,0.24904933395336892,
 0.9649107746675062,0.5306089661674483,0.17536744134043214,0.7151579461026719,
 0.2942387184734939,0.021017088632728442,0.8624300350942116,
 0.5947284906943566,0.3634118135167318,0.720812182413138,0.3836254468571747,
 0.37385501095782936,0.40446012588426083,0.9223028455293791,
 0.08324324687380369,0.623138374229614,0.510510018304686,0.7438356496709206,
 0.8100306555865823,0.22365474424957676,0.9805138914460331,
 0.43199930515858687,0.7777745598506127,0.30286318718625804,
 0.4791450167304451,0.590449957363549,0.5324453971665368,0.09782314541162684,
 0.2009057490474112,0.9967302323585672,0.9639788949832555,0.8473561461918256,
 0.1815095933382358,0.8052250370678921,0.7312095153569607,0.4054003503901632,
 0.30502027778653984,0.641272625564022,0.9769033018571092,0.8090916187652549,
 0.3716835590229042,0.9252622243164503,0.9224853287120419,0.42062240372588977,
 0.9571875419424614,0.5884818136187812,0.1540295550740649,0.6530310350630342,
 0.5154790359039721,0.4577721522446754,0.6969710921619453,0.35198533440466295,
 0.5948734655570443,0.8236762474427626,0.11883511221590737,
 0.04018373489935256,0.8958818670453004,0.9258340355189365,0.9004344048397318,
 0.3174620983274177,0.5574082066069012,0.34140593987766543,
 0.43161316262502236,0.7661248595213749,0.6731443285524665,
 0.28923668716648065,0.02946551045518353,0.019468685114646722,
 0.2733157461542749,0.3568450312217495,0.5059721103164951,0.6279557290625103,
 0.9948489678393718,0.8732795043920907,0.38475949530987275,0.3064652259841826,
 0.7550673589719986,0.4561767020131504,0.5563720157306322,0.7746704000100088,
 0.42630578228527083,0.627787426969201,0.040543858599210614,
 0.9678744219981447,0.9090734826074982,0.8829211687791707,0.9788432529840193,
 0.5627821769005047,0.6739763891873521,0.9238263219182296,0.18268487675373768,
 0.503920448567696,0.17771308055240953,0.3213640899506456,0.05328741043228391,
 0.5065467544888015,0.6052545846712015,0.9548020704160398,0.5044514879148997,
 0.24203313662898374,0.1752508424949497,0.47639089746601937,
 0.4800326943973726,0.4918135882379784,0.2881020165595223,0.6729486960723756,
 0.6412304894786907,0.8614507246892167,0.8066685156955499,0.10264553752107197,
 0.7922433090291378,0.3155417916428228,0.01778114855874846,0.4472964476235407,
 0.28333828002305106,0.9347260963958328,0.3899752360662435,0.5651446826795011,
 0.5841791407506054,0.6371612086192086,0.32354032223795814,0.6649656183501175,
 0.1750314894112428,0.5687738923294735,0.7667709540794767,0.4668625532910169,
 0.28110635601770584,0.5473059409959291,0.3934840208779512,0.6973272844551366,
 0.3612916177860188,0.17276895328540665,0.6378187072206848,0.3719868000239974,
 0.3748775664943156,0.3164539211080639,0.0596475058436714,0.7207067570907041,
 0.3083206155683387,0.922484167165518,0.8732843842161657,0.5553610883848337,
 0.3895573822604177,0.9814257766471349,0.8613672838635557,0.14167276549909402,
 0.6797450703461129,0.03320822471296292,0.2577778298051101,0.6127006755424983,
 0.04142495195922968,0.21428078322811817,0.6823389219074749,
 0.7286405308478934,0.5501954501694003,0.1011355345381334,0.2659098192927669,
 0.7455168157067719,0.128328132756512,0.8951497934385623,0.09816727268475889,
 0.3350574904982865,0.7952763086782739,0.4034082940889624,0.9252603369329138,
 0.6915855283476493,0.3004741328330083,0.9449851285419182,0.9749896609469005,
 0.9673873059012809,0.8758244620315939,0.21093464944938234,0.7893118698604298,
 0.6264397714337308,0.24151867848181996,0.2127124702437756,0.7847561244253142,
 0.8830278341923175,0.8583879997558344,0.28850561687825205,0.6140196386316178,
 0.10902021977821985,0.057765968440168125,0.3658960683121193,
 0.40614333612740283,0.5609040126595418,0.8006604848653243,0.7631853140900936,
 0.11181661822690037,0.5545696963624129,0.5951131955521445,0.2612613093888062,
 0.27328096958155257,0.4158658697615355,0.08529204798413714,
 0.9014571330622659,0.38792763314840706,0.8729478359296815,
 0.005948385256137501,0.39483100706551677,0.791081592031346,
 0.9525773239521476,0.08082551464291654,0.3241251112982194,0.5335468936522064,
 0.5968223293334233,0.12622434380356595,0.307275277129029,0.6419536657889672,
 0.9307285223263313,0.33617045890092534,0.6531238803239028,0.7419938711590034,
 0.4004398802543301,0.948477496628809,0.2177347964104095,0.8858174029377037,
 0.1798830394266111,0.16966430419492062,0.4448597484425466,0.6197714683341953,
 0.5408346526118056,0.694253657419907,0.6882806298627984,0.695768698428113,
 0.9109791513219617,0.6156518028313873,0.5201850864171345,0.9072527076944659,
 0.6933600628016807,0.43349206237182947,0.6092001150166511,0.8842054062061881,
 0.5406141815991166,0.8036348977631104,0.5645784132203348,0.6659820782350808,
 0.45010105245414456,0.963676886999317,0.7016606825709841,0.7355521985674189,
 0.5103977993276215,0.46315118814310663,0.37986554307642606,
 0.37077073007440453,0.12989201033337783,0.5950746582400291,
 0.8352389862100928,0.3883645748952458,0.7113408976738818,0.48211192828287275,
 0.9733309746594986,0.09027135045838786,0.08369142484770764,
 0.053098117189536165,0.17513808937309983,0.030467882630485166,
 0.7743384264249891,0.4496275037174685,0.108671074760892,0.018571696588910225,
 0.43032828960386604,0.45049675762979546,0.2873827830721011,
 0.4440926851812117,0.5384515683562667,0.857382284526146,0.17075040865641988,
 0.21557880715784794,0.4761291032410271,0.16671189414291865,
 0.2895012495930591,0.3350150377198923,0.9020187573757199,0.7683492332960443,
 0.2622585803663665,0.6569027641349406,0.23577882871043299,
 0.010172089110437654,0.2965896277345128,0.6554418558481688,
 0.23634494085049296,0.5785172134038433,0.750643972457113,0.833533658658121,
 0.28722812961425714,0.3581651225957796,0.2841597548626783,0.6566533889615475,
 0.4517774910926089,0.6046387400497077,0.021267335819111444,
 0.3017243025937487,0.5072897262220526,0.09797519996085247,0.5836596400641334,
 0.11565606987747845,0.6548012475848871,0.005909901349441582,
 0.08735097043402562,0.4867151220904897,0.13272765845817114,
 0.4698228987815035,0.9024011922241232,0.33567844715489603,0.8506850923450529,
 0.06484897943175749,0.8663133081996721,0.48768577503748944,
 0.30522952357375943,0.3038825733487832,0.2837643514983771,
 0.10274118661134901,0.2688056240259611,0.13972916671810698,
 0.10678625169582334,0.15099357771071464,0.06446756396439624,
 0.6116435307829277,0.819021997766549,0.4900565374772947,0.42787659159753433,
 0.0747590645282018,0.7182900836529766,0.009021782098525222,
 0.16470648152017042,0.7348149915028398,0.626494987856065,0.18665614165955857,
 0.1511833157997584,0.8641717814205625,0.46046972600731095,0.5213044634451771,
 0.5466363128992426,0.31826255187029096,0.7693421542150609,0.7779476489340761,
 0.0035988506712424373,0.9869585578672769,0.6931172336908823,
 0.20271674062286937,0.25532632361972096,0.4473271700044884,
 0.40657051909601205,0.8018736380164616,0.8638915524309063,0.8522583571887562,
 0.8612122887703842,0.8044185305173758,0.2494386069974961,0.12530146842944878,
 0.6846858972578614,0.4915513604432381,0.5661220945844881,0.4850296487276148,
 0.7852557520373057,0.38257754672056166,0.060691527086891295,
 0.8256466530192459,0.9031002896387027,0.574960728857287,0.40258833563519314,
 0.7393026014611739,0.14699123778311196,0.8174069451193646,0.4336576099505227,
 0.7553184926932448,0.9588058171716919,0.7966180588945502,0.03820041972507193,
 0.07672191750538404,0.5391908263014126,0.18209427888030472,
 0.6267949273219181,0.2158861382727013,0.9437233434162552,0.6246696147866377,
 0.853652114111624,0.08158625358821481,0.5520953130759282,0.6738859324345737,
 0.9078337024853276,0.06472944001716707,0.7325731506596402,0.6753972792185993,
 0.09513118202420978,0.5240081995687995,0.12422476825398987,
 0.1516441064399877,0.9568872956623986,0.2749212059963948,0.5366406911186266,
 0.24495507021652518,0.28643194168162234,0.2727930704100642,
 0.1615271425201671,0.14756994197894258,0.549540727872923,0.75887378128771,
 0.37660331993643525,0.798391048980194,0.49086479509403547,0.7018759224978286,
 0.36502278326088033,0.9208810238063065,0.6876876445226513,0.4190715387959687,
 0.11978809721444406,0.37069996613428813,0.7220841320446206,
 0.7581934223031748,0.22295143658939398,0.4187396550555922,0.8759154780249265,
 0.7122276223549068,0.770501593222793,0.19260280240492178,0.3753393810893081,
 0.4874730474578207,0.9074490072669045,0.10514759370190452,
 0.33447726405638223,0.2249562139271596,0.7165147386192404,0.2813532607382292,
 0.22919895388986766,0.21567055611945996,0.7473593611325249,
 0.5918504720026607,0.30482679668895507,0.12562464394848216,
 0.050958042074347065,0.9039658231618376,0.9685294876126035,
 0.16727789504914348,0.7545566581543994,0.9356371556571507,0.9106806533124148,
 0.6948210278140291,0.7775885405530499,0.8494225012888822,0.7922824286513361,
 0.19444421577394366,0.0440253593899782,0.9364538977825981,0.6249993364526046,
 0.24514294109808932,0.9695971872508681,0.4291733661644652,0.7441098992296742,
 0.24177511242900185,0.9223000535669216,0.05346758866747037,
 0.23402101600437708,0.6499181903762776,0.5901427979744007,0.7082236211643689,
 0.9028224378778504,0.4782755644195761,0.2000368919917993,0.396555384316302,
 0.7422401949095699,0.8773116482562215,0.6146748471296771,0.6047425300690503,
 0.5614421533744183,0.828362410379133,0.6359016363322271,0.3522421301827391,
 0.8757823966762618,0.3049823908301147,0.0541209212128394,
 0.021114662529463812,0.6865663915893655,0.11570231181808655,
 0.9767629538651325,0.80135044143639,0.6241396527303512,0.10502578156094311,
 0.8982622350274113,0.4653414085231403,0.36247792823353187,0.8255416308841014,
 0.5750768667772758,0.35846035711732793,0.233545278131839,0.949877256345331,
 0.4497284666254068,0.09493265989435817,0.44132345975802445,
 0.6231303554276635,0.10327968370706797,0.36874991346374275,
 0.6422957074927463,0.5952662276644144,0.7460361190131998,0.7935012528840287,
 0.9959511557188143,0.5879827424846026,0.5749040313684135,0.9597973440089895,
 0.24524923459558678,0.12861299050395825,0.7551096212186001,
 0.1478299153452326,0.1917766762105083,0.04294212851069279,
 0.36838276542144943,0.9819657787974093,0.2610910600888162,
 0.37724252781224454,0.7979512087812151,0.0926142801762252,
 0.29782313965984175,0.11755545438171844,0.8359066420539333,
 0.14533027646179075,0.2576032739075609,0.9662705888015231,0.8068391432828712,
 0.718017302157661,0.8265084988682687,0.49570142710557563,0.8857144196189023,
 0.3238776133926191,0.040327704981184764,0.17031883195592412,
 0.6912170103646175,0.8857994625432737,0.12448872738561567,0.999478319996776,
 0.37877544273879704,0.0742497556227516,0.19300056603384041,0.966046128302316,
 0.19392574946827557,0.7020013976488918,0.8988586773235455,0.898578102548552,
 0.5199937863048735,0.23447211748367425,0.4135489416985618,
 0.19389426806155796,0.21320992393768112,0.10830153258125896,
 0.3014004357141571,0.18441194741575007,0.2767368962118958,0.6018112529532371,
 0.8846802781048572,0.8759621307215091,0.550030162544501,0.7366464929397348,
 0.29377955154540425,0.37637105819301986,0.40607452728831905,
 0.05726068479198265,0.6963599908937168,0.70933996501438,0.9850295403741245,
 0.6776162814567419,0.14151043362577587,0.17080579360447068,
 0.38805091998676855,0.3273599153569744,0.8593736360126636,0.4581320944846524,
 0.9551342924611352,0.6339472393031402,0.43788904788543015,
 0.19711484368636745,0.9580307411436597,0.11725867745449836,
 0.04130872548924458,0.563517513849128,0.9337071066407717,0.23651459639826045,
 4.0839762765010335e-4,0.02892902546078302,0.46235123792595845,
 0.9977274333678176,0.44190252611974945,0.24806371997907184,
 0.03893008599321668,0.43068102557208954,0.8077590894647264,
 0.5568267935072705,0.3085410500142184,0.9851979349908276,0.6895343535032017,
 0.5235084539184429,0.07273442933398333,0.6449459803479152,0.9483435754860451,
 0.6897191062080383,0.07248797233093085,0.9832872659078764,0.348323020665503,
 0.8642180443064564,0.634536676256255,0.046461028245295855,0.5279221085279634,
 0.8433301535058271,0.7501783229548913,0.14449369513823695,0.7672000986435531,
 0.3581448729093728,0.7960324795623054,0.21604808465829384,
 0.17291037482442484,0.1203209277197208,0.8613044049588656,0.9790723789429137,
 0.3400682890617097,0.25755798060960966,0.1034190010246222,0.9223928843607676,
 0.5901550758637963,0.9618961181761515,0.16589869797309864,0.829363525187162,
 0.7379502915646682,0.6643721864648588,0.9092489621526535,0.12969968145692112,
 0.9551169220776518,0.6774353491883841,0.9980671303047274,0.38954692721693707,
 0.7027311636059796,0.586086233745349,0.3210223412244215,0.739106076384013,
 0.6266977878373569,0.9881480370606085,0.9934195607578551,0.9989975731162323,
 0.05476298707085703,0.4252482524300658,0.16056782207007203,
 0.10213262147961222,0.23853841464072378,0.21476811498630244,
 0.43845314605061714,0.09494970271455039,0.2792009835006408,
 0.8306117139574095,0.5555224032939716,0.7337838393513766,0.41297888558788787,
 0.7096306158602973,0.026961501854582903,0.6899891317942345,
 0.7268374289369638,0.38711549499475173,0.9072707143512501,0.839454347687788,
 0.1492207731904106,0.23552363204308047,0.26283547199676505,
 0.3509145696328946,0.1913559519183876,0.4693759831613403,0.6083945284007589,
 0.8384913174160954,0.2844229094648748,0.622719460383814,0.37116310799629937,
 0.09533617269370809,0.5798132993833323,0.9213609405399352,0.9498153362764318,
 0.045789311176771985,0.09549712894769646,0.6719022904015265,
 0.19776572918514845,0.6190903289184186,0.2379172234199256,0.6667404182637118,
 0.631675131154102,0.9369410091323759,0.12395788829519128,0.6469898133618464,
 0.29093436614748236,0.7808390651274277,0.3294686649156704,0.7702066569441886,
 0.39893134476355263,0.9519318140718503,0.3012512000307094,0.3102962145187793,
 0.06501543892357564,0.09770609840061173,0.08585092475417078,
 0.7224408800631945,0.9424539473777596,0.5377198665592788,
 0.053371920955240015,0.6509470349550079,0.47295510991742373,
 0.2878865656369858,0.015757805916196466,0.014760194520007208,
 0.5576459000523001,0.546030603909931,0.31388056745795123,0.5694212561197646,
 0.6036216280544867,0.49782935434244013,0.21909230585677086,
 0.9213902288073909,0.6978583993952726,0.22758609719176703,0.2659658217026143,
 0.3653996597560669,0.08457905895945816,0.5227201461714455,
 0.43557287858380034,0.04324192188453846,0.17065508236850824,
 0.6631321588575997,0.5777788402094977,0.025577206878641512,
 0.8421075500899317,0.8718062908511609,0.4689491104739264,0.0764216779020147,
 0.39569935274644463,0.9484891508811367,0.6795527527129874,0.3314120475008392,
 0.45519557910545105,0.4964644227077316,0.6606525517822874,
 0.24153908712649252,0.4215423133579471,0.4950099618306084,0.3054578830983641,
 0.9871626153176779,0.0013075972729569285,0.3707342910608872,
 0.31913637151487695,0.19688840934259155,0.2028082056090843,
 0.5293317770051834,0.239704559485046,0.2183562098398637,0.6884933731061372,
 0.4284617719181725,0.37235804827014174,0.7578431533667895,0.6550097328485281,
 0.25321563008693715,0.8689448669274239,0.46278346414617966,
 0.6794306301433773,0.2627936447260113,0.46104288801221627,0.4332452533998943,
 0.9507239546609589,0.9574021678150607,0.4405981010219244,0.28546272220454894,
 0.7454632424210718,0.5576209600589029,0.6802479609229644,0.2772573026433436,
 0.4234299174636462,0.8565499650069898,0.6808437564827889,0.5795856698810052,
 0.9822551070177186,0.8544704260363066,0.07597273393391757,
 0.07242962368195494,0.1117981138950992,0.40099687995891964,
 0.2720830139352166,0.24406588876631619,0.8031588096877973,
 0.05153849982924097,0.46974977356979886,0.1508945660453117,
 0.8393110815314611,0.08886585795365809,0.180247320176126,0.5551456300659323,
 0.496142721015172,0.5553729916876629,0.9865191525427643,0.0748101740936149,
 0.6221635592117133,0.19497855658414542,0.020890278516125127,0.37169703472953,
 0.9794777759149782,0.12664263682476495,0.9766223074877864,
 0.08528525558420497,0.20587632523831778,0.23597985499525354,
 0.25638878840029333,0.8537856573209655,0.343817763271835,0.665963153872362,
 0.17921666039139827,0.4868563905002107,0.6684481323049389,
 0.44295396128611286,0.49063110685678035,0.2212309045234504,
 0.9793925807659936,0.353161725348495,0.4846320430919425,0.4128954884291195,
 0.9000897931510975,0.547523666032707,0.7294075094174424,0.6796499213133445,
 0.019484444714045535,0.7837783855704651,0.6674832011246926,
 0.28058942873828774,0.07206783324566723,0.22379860428731357,
 0.8733570054681354,0.7606123952843783,0.3282912586472725,0.3309472906633556,
 0.2870581674949071,0.5621554338776313,0.18570781623797183,
 0.058754687724010046,0.7318452309706589,0.4917823768337718,
 0.7445096376133882,0.210461216107672,0.9643888451701789,0.49535562585070514,
 0.5904966342721996,0.18906251696665,0.801046529315475,0.8351377557060158,
 0.13980645934913005,0.9521651873515302,0.8936538277619871,0.309009696143982,
 0.5599841175289307,0.21869954367115474,0.12757767775328277,
 0.9440795097066712,0.6497108000124598,0.41484781735425635,0.9900302437913086,
 0.41525842266542645,0.39800539231244736,0.753823375029406,
 0.42618214769133766,0.9491268883418964,0.8824188237093884,0.5896856324607792,
 0.16245355779228166,0.5915587097273879,0.22633313200948724,0.708686973253823,
 0.7601553014939573,0.1296828770032804,0.7967971739244041,0.45738979354385034,
 0.2976846173569023,0.6510741494723703,0.310020008997541,0.8422002632203864,
 0.6552208183873555,0.6597816122673408,0.2582728048811218,0.3290123908188599,
 0.37204008361894547,0.23356487623035738,0.7775119739572578,
 0.7536708811932654,0.3253076456238303,0.49638036149798737,0.7720763934006423,
 0.6352937660193756,0.3174707198577327,0.6411014264933818,0.5164046621642484,
 0.46697713846915856,0.4142610436182983,0.574648804802077,0.5401558760309773,
 0.143439005712917,0.2337545990338985,0.6758465878503003,0.18157349269326273,
 0.05059685127393876,0.4469395739399079,0.7898893916469515,0.3140018406894427,
 0.30349088307122696,0.2977054565278813,0.6920020947539536,0.9358756162494427,
 0.0251595712654914,0.543403526253077,0.20126813594499127,0.8023233277777368,
 0.5546867936663903,0.8376934168457602,0.5893216045762976,0.9347437030229526,
 0.23655438773175186,0.5277210852411254,0.3402458546949578,0.7923414338320249,
 0.12338660882207653,0.964306972976987,0.6195126502387811,0.36875573916516324,
 0.9517880396186192,0.48616791986223196,0.818260877298822,0.3328218796434853,
 0.02657960184925212,0.2865135275096027,0.4739475911357256,0.856082975134916,
 0.9881230627581593,0.9600744067124275,0.9400843723575814,0.3524519330637653,
 0.08363422629157424,0.0678424803654043,0.8454696716768781,0.7681538531088169,
 0.6211956195427226,0.6430010178328226,0.36095917774462083,0.8695702864286429,
 0.386385292593626,0.2891706029382971,0.5626799204987148,0.46798325811056496,
 0.6995577365706902,0.8091298989274999,0.6836975789772741,0.24387838575108722,
 0.10394505928271536,0.388514773374386,0.9420879416152778,0.5240238203259139,
 0.004145023619006483,0.9014797411371113,0.27410312158991645,
 0.7902301936123406,0.5796935733532246,0.8647228153408821,0.7396164017237777,
 0.42383858638929894,0.6165979329645482,0.3062941867630863,0.6324691165464185,
 0.41633329718427003,0.7486528763025042,0.9965980350962043,0.8832241369344622,
 0.7415209939789512,0.07385021838260608,0.2889166256204466,
 0.06054302015337287,0.27583675790126694,0.5991279600406485,
 0.5100224487596506,0.9293191003567367,0.46048329699805035,0.5689552488880005,
 0.9887389406869206,0.37084755911917466,0.10948219289095917,
 0.9824938274703877,0.7544394142257606,0.44854731094363187,0.7187625337398292,
 0.1496069416049688,0.38715968521600863,0.0632890005962653,0.6415529797876209,
 0.5056170350630899,0.5363941047127302,0.18283105784096876,
 0.26789195476514827,0.9125053559536013,0.3392380412800525,
 0.44145766580123846,0.0927001566712723,0.8589605250672486,0.5038972301532324,
 0.9956610139494436,0.39229377899577467,0.8504364325691084,0.822475373045738,
 0.7140752546991334,0.4721489905115107,0.3345346284191639,0.07288813703888164,
 0.4115274363157625,0.00830523619831125,0.40238092804443315,0.268976262350072,
 0.7980981285912515,0.47786985882005784,0.08228066977141135,
 0.4433570856284422,0.2639038354361385,0.306519439659532,0.0770450273814951,
 0.8447093452982575,0.6562863639809512,0.5039437554148307,0.7353026960599605,
 0.3622039926429055,0.0739255185340475,0.9725577181134099,0.4322997287139271,
 0.9429961092670974,0.47355995210584223,0.9818168888113161,
 0.45768221945492493,0.8118871211162576,0.9349320054614692,
 0.32919660720190447,0.364857363290758,0.08146042303560619,0.6127793338836518,
 0.6983398080051058,0.38990409581049634,0.8486701560393203,0.5922744268710144,
 0.9787356315134309,0.008325958277723533,0.544856475913695,
 0.44230327493677835,0.4350882759957374,0.06324626993314564,
 0.32197505458610387,0.3702598246861145,0.3132776956757687,0.6037855548911968,
 0.055385298393103355,0.6580997746645046,0.2021362370253318,
 0.8346156363135462,0.015445294724697023,0.2658930535727889,
 0.8966474971701631,0.8751682004122688,0.6993472862130281,0.5823924531736187,
 0.9409789887470865,0.659177073483672,0.48745407386581496,0.23534273744335732,
 0.8773862752686393,0.1776237662000132,0.5893360681884893,0.7439951835459646,
 0.5153994001619888,0.8551637796837146,0.5770421080507021,0.21181802080857426,
 0.121495046417188,0.3109490583485407,0.3559244899090198,0.38585597825556217,
 0.8508656908706369,0.6338684069790388,0.9419016289263863,0.8465786738223373,
 0.9559754611052544,0.7805641987649881,0.9686005895724061,0.9924785382322329,
 0.9693371163006598,0.8694654030191097,0.7122691487485314,0.4197073474674772,
 0.7274094924808404,0.42221721644653354,0.7855156556885711,0.9690202827972052,
 0.1780025668501406,0.8782668367594196,0.6096500518658603,0.26771683773727095,
 0.31162295805229834,0.7121070781137,0.6652791053428485,0.6956589928210803,
 0.8795262872249784,0.07034686776140031,0.08413895395660054,
 0.8606061931871508,0.9964645129900807,0.9916091329360212,0.01476049817588665,
 0.5262936576649464,0.6418312627291487,0.34676740630336733,
 0.35131703576563966,0.26990095578461204,0.6414978015851498,
 0.9201290148041723,0.6805889643320014,0.12283292310750071,0.7014962836930057,
 0.6630228342288135,0.3617905259365607,0.6900917851646619,0.5586453176253957,
 0.3769831048389929,0.23125233230309328,0.7458111794080778,
 0.037766363002052716,0.13080551362553128,0.8055483594737544,
 0.8621851779709714,0.7614268386360553,0.3524467649466778,0.9441172156912494,
 0.9989045290471581,0.3412538589125861,0.8948975760480083,0.5102849491525445,
 0.6908350659546427,0.9389585304294492,0.6342374680741601,0.03053495383800031,
 0.036322863358825286,0.8410134147441012,0.17535121817639188,
 0.5053158031820294,3.209092251411638e-4,0.11784613014142109,
 0.9992039950131952,0.7378840622461399,0.6627560513653645,0.1593653172760593,
 0.13751960832850418,0.40360200203223684,0.6250910269937009,
 0.9104791294977032,0.7691376847467741,0.8692100605858192,0.9279370526682997,
 0.31473419836588734,0.35067713119982313,0.4781081910110496,
 0.9038069012931317,0.8329405368900579,0.17154282175592428,
 0.09085177666229927,0.5959467773204526,0.06853276429289767,
 0.7802974514193576,0.3832865906035092,0.9406638104083321,0.9540802758305825,
 0.16118005867201635,0.8170454737046621,0.6296889458363358,
 0.20348126976338687,0.29825935198937137,0.2001777921093031,
 0.14437590204646245,0.6123973572907354,0.6954855189438186,0.5785186129576608,
 0.04122708664858976,0.32343918798232296,0.17398331971481007,
 0.4526514885011811,0.7402337206854167,0.1214669182148359,0.7058031912698028,
 0.8019169334484278,0.7955745495970249,0.9333084037436752,
 0.013323065279890467,0.23024354539827208,0.6849445807210897,
 0.4005193870014869,0.05539851832869569,0.10327475266399821,
 0.6697909647229965,0.24188318160950628,0.3243467643168031,
 0.12282614890206522,0.58438569794861,0.20821879001574062,0.6074518900609048,
 0.3950494349924639,0.31354460505645043,0.36973541637166785,
 0.13690875066617558,0.7752313256223494,0.7366443266111979,0.4895234435476884,
 0.041027284495384286,0.4286130869288821,0.9133331258816394,
 0.6324351819306024,0.9417782268672895,0.6170474970283152,0.16533155874657546,
 0.9002190947202254,0.5575476759801035,0.5794054473614819,0.5370091433392226,
 0.15080374374106853,0.7487954657996061,0.6578997260417171,0.7237142046436086,
 0.18866141051439378,0.07311508604640105,0.07239805878278371,
 0.45857759089695516,0.6278356813574564,0.3953636651112461,0.6738422092908125,
 0.26305976915062323,0.7824709331269668,0.7773650272975065,
 0.08498745883840042,0.7484469006420886,0.644170569422311,0.38451461386731156,
 0.34196670098770565,0.08912277899977117,0.4954509075626068,0.747609310502892,
 0.28672631259446113,0.7388058404154766,0.7117048785553906,
 0.44674675486955173,0.5223174698935142,0.6833795059526868,0.2203458405777331,
 0.6743019416576944,0.20103966358194336,0.9320123102963241,0.8491342388731216,
 0.4919133106931265,0.01025013657870022,0.14040678736217072,
 0.8894546205010937,0.8778678544970114,0.5588639205390584,0.7948384020144464,
 0.4322824991946995,0.6041418488421619,0.5798489073380888,0.8900676278976666,
 0.7265023161127786,0.36158326803047325,0.5257259252051988,0.6047949083204148,
 0.9976796764913627,0.9706380612877336,0.11982606731068424,0.8017312223088462,
 0.4899672956229789,0.203301424158979,0.19291362359706954,0.1682466042316496,
 0.909828778563147,0.8890308257017071,0.1318342133489281,0.6601776932243152,
 0.5894509551310231,0.4987528972882038,0.8950364669953479,0.5488683065752752,
 0.9126642004550416,0.1193067441218405,0.01254237463789265,0.7960305047560073,
 0.011372226455039103,0.9066668746084902,0.3731047881905015,
 0.10625285486150304,0.1841628847401433,0.6303145466300276,0.426078319279245,
 0.6411909139323886,0.6056237546671385,0.3809091983464723,0.5116660833417122,
 0.047259839880736565,0.7272903933713968,0.7569080638819297,
 0.046730642466876415,0.15497113985449962,0.7728898261199769,
 0.10689224966084632,0.3463875345690278,0.7025172417765735,0.7796499379381973,
 0.45385023889682163,0.9962062195874725,0.7781806007538155,0.6211479782777919,
 0.5488197986387332,0.6879102796410121,0.4206108207957103,0.6230401562475308,
 0.13117582560528673,0.0973396821402499,0.6376393686412064,
 0.43126789662383713,0.26741952656859125,0.8779205739315651,
 0.6194018382665443,0.5201289968434547,0.5243671457619609,0.8022978688350942,
 0.7405125078543087,0.0025643482552929475,0.5962024012289308,
 0.6567216097752417,0.8195237933601016,0.8331431633576614,0.13531158785421982,
 0.6378737812897834,0.501880945958349,0.3172548773852488,0.19058458777984755,
 0.7574691323566239,0.4283025513053914,0.951520188791278,0.6664983159867777,
 0.6234199709191454,0.20121095152570356,0.8767841126044053,
 0.18899136046795717,0.25307940971675547,0.6548255088207603,
 0.1315830957073283,0.09284277401659513,0.16876208522446,0.33172232362219767,
 0.8969949832433822,0.40662795634088433,0.28351210416502637,
 0.12261115491531582,0.01911972610410484,0.43578494754856434,
 0.09429201317749591,0.10800332831315695,0.11539226390366786,
 0.37572505375056575,0.4092053587035607,0.8662639232888386,0.2096511330306957,
 0.895744972106979,0.6743728104848364,0.10762019360330788,0.13014095349949772,
 0.09227083717615847,0.10190482961718494,0.1286057033692254,
 0.5739173256892791,0.660581084981964,0.9321277445907963,0.294191955271934,
 0.2633321010280617,0.10297149357470747,0.6939664156672458,0.5755942340402358,
 0.35048067318712184,0.5067020172928326,0.9063821683990486,0.7426451521588212,
 0.5629093181067086,0.7721313667182853,0.8431470891603255,0.08150828258087162,
 0.8033994963274768,0.624930364219192,0.3854474135052487,0.6215782653975629,
 0.6258825113291457,0.953251368332207,0.6812542695291255,0.812428104808178,
 0.45523947392153663,0.33861682621908995,0.6589803773878534,0.261955295554554,
 0.7902186654164102,0.456008334341634,0.44604536753087154,0.9161193053528037,
 0.7227863231321755,0.9025091455814127,0.04178890288675685,
 0.23690287853568692,0.10774078361752859,0.9906028313412149,
 0.5470154581996212,0.041117517873860754,0.68647374566529,0.3368183175704894,
 0.5400999620964284,0.13342589844938424,0.7511084713730828,0.6688884394676797,
 0.9744360935392067,0.8214046682457206,0.966221721844924,0.7529522316985563,
 0.20001917879575815,0.37053765070996547,0.35648197902084144,
 0.9870615666637503,0.27383180152488684,0.2104939125248778,0.0705786877467075,
 0.5649427995678935,0.006082538857478226,0.8576956530779083,
 0.1710643648467528,0.5510663825815948,0.32997461034057407,0.9001533477989149,
 0.1810489988877101,0.3583595122674581,0.35139148762411665,0.545529549358511,
 0.18414813070961178,0.8132994129525533,0.9719017859012264,0.0180959156339342,
 0.8383405143109356,0.5900907617938963,0.5377878479191761,
 0.016197608530518104,0.5852808693476652,0.8829472537652726,
 0.7810347632878208,0.8824889709378108,0.7636258397895438,0.4278274581537468,
 0.610224868617588,0.9792498168294123,0.3007893128198982,0.6524584540450588,
 0.9286321996712467,0.6923696819507462,0.8394777477842034,0.5335309824419322,
 0.4474171860465176,0.720853573710293,0.16896604184356045,0.5409347285194087,
 0.7980177720328414,0.3953468863795242,0.7393335557172941,0.5358694823558575,
 0.6647851124302016,0.18945486304894021,0.32144066967293083,
 0.5397692040849384,0.37557679692250057,0.5226514213152372,0.435527939763112,
 0.2901993870047399,0.4293114778208136,0.5216070830543191,0.7837391136593965,
 0.5723413661972245,0.4080722464687577,0.7230870997421763,0.17321762753549952,
 0.8113999860746556,0.954889678244091,0.9183125174865675,0.5901920619865353,
 0.3470201267988442,0.05539086783609404,0.4272407156203998,
 0.23025761814607792,0.019086749935507408,0.10533304524524134,
 0.24159451693017553,0.798856668624969,0.24149381983565976,0.7035536501104279,
 0.05818793463070959,0.721390942982014,0.9575584625138234,0.5221848447172627,
 0.13495834942027451,0.4671335781029462,0.5581974546529291,
 0.47081217516305895,0.10296768152661584,0.23427036565468462,
 0.49341074935711604,0.19893293893463138,0.4497253967490047,
 0.28442220923448125,0.9417091606619782,0.9313145997432488,0.884388018641984,
 0.9094299368247603,0.8723623808645226,0.229389162842349,0.24277953520231443,
 0.36682085887621885,0.6412053699591403,0.858435846383626,0.8845540444540263,
 0.1792452738566861,0.39202915885466527,0.20060712828553084,
 0.9755859253110735,0.2830888523938546,0.796834850146592,0.8273840363772654,
 0.3407873282658097,0.050768400658810275,0.7893544138595622,
 0.6933782766564278,0.9050396981711278,0.5593295701776125,0.8652834686777561,
 0.5036478090122,0.5427693993752803,0.803357277304987,0.7187302466204565,
 0.9275486114891371,0.854790716073909,0.6175630475453711,0.09852322287635928,
 0.9296093802796296,0.5848888077942253,0.4053411599067608,0.5084634899333256,
 0.10581463154000914,0.1758854942110355,0.2756072755320097,0.6087753287522817,
 0.9963011650795768,0.36068504422526526,0.8405662052963252,0.6621399807356864,
 0.9503237092243219,0.7643920707915068,0.2724661758616782,0.01615271580116806,
 0.49095237088757104,0.16074841069172607,0.6708234635103157,
 0.5975537830298261,0.16632846404713064,0.22837015467975186,
 0.8857853597177309,0.1984859239540101,0.8557615175825439,0.8106998587600583,
 0.21451933576347826,0.4038492362177247,0.00985301072640421,
 0.2229614963895713,0.6526719590431056,0.6813754230595721,0.7814517198478512,
 0.8865686448448622,0.24678066573722934,0.5436605667315099,
 0.33001430568129353,0.43893504874653366,0.36000858143351144,
 0.5423330728372652,0.3152274446859248,0.2986753003548668,0.23630327141385077,
 0.3790445791088405,0.15957949653498316,0.981524750504859,0.9756386123117985,
 0.347571849402165,0.3655377366405739,0.48775422131518575,0.8776330119191976,
 0.5080902059035941,0.6990568794372805,0.8269642828538101,0.2536050681466908,
 0.4388480865031539,0.5980615831292563,0.3575600547943457,0.41545932664973295,
 0.40833868470077217,0.33896693790805865,0.8325358987202172,
 0.8874790829713458,0.43933840787935385,0.5595870974755134,0.4812636854461241,
 0.7965184684206642,0.9736875505267486,0.567325075753344,0.611985514931989,
 0.13471427464132646,0.7459086704686917,0.6223374732786118,
 0.31076193857022116,0.35838028251349296,0.40606960951465265,
 0.4012960166008154,0.47968619630948606,0.3402772706903634,0.1748199673492643,
 0.06409974969023469,0.7479581741190865,0.38872070289233995,
 0.8364182960583487,0.5225626957461175,0.02171984658194459,
 0.19269274541201464,0.875448384834995,0.9025047159333351,0.4852775584534049,
 0.44465756516349997,0.5897199786825502,0.8711675140672188,0.7884865097290875,
 0.8730765370616131,0.9163893208374094,0.3340584545106471,0.22612397013441576,
 0.34977728498412386,0.001411776599107184,0.058762531570483945,
 0.34332561917034443,0.6184349751220743,0.8783658302095663,0.885523176758357,
 0.3619930619692966,0.18967417562011157,0.47273671578578524,
 0.48058256177166714,0.05360984031441851,0.911854127693178,
 0.24296072066274021,0.6085995977614646,0.6696874873737172,0.3454679907936853,
 0.15190078287476871,0.9763274907540314,0.015590460450005006,
 0.3035544329110612,0.05970029920271447,0.22211756781423586,
 0.49329951421598084,0.5115469724905178,0.48710833065660986,
 0.8760451455347134,0.30223634025212387,0.6235481365236685,0.503268909252368,
 0.3246998401725675,0.7327827259958799,0.9177389006480172,0.6062880556145815,
 0.06372905005704588,0.6249472648333779,0.4250566350186431,0.7149196858134393,
 0.20236170984995938,0.7426662302414914,0.6949562863783925,0.606534340978879,
 0.16949141978485205,0.38523120950159506,0.31298438216406343,
 0.6225476563523744,0.46932359920502686,0.07894142111328639,
 0.3758142243195247,0.8685530139222324,0.9036247070666066,0.4942555506805323,
 0.5295937303460347,0.19374557159564798,0.8875368661415467,0.3640620345457859,
 0.9786800195284929,0.8285135328317056,0.9265578964139215,0.23025081072112252,
 0.3008746861233359,0.047595628238151555,0.8999490513253794,
 0.30564347830791705,0.42675174202196475,0.5200430948607766,
 0.29092940630548425,0.48181135294878863,0.5801320208126909,
 0.13605882575555328,0.6017676200485859,0.5454883015131565,0.9087711858146699,
 0.4410891714216215,0.572831433547937,0.6544298524077747,0.18259216126857614,
 0.8428256724465328,0.19692105679859218,0.8725180588313395,0.6715442759286849,
 0.01405935670168057,0.10776011606485769,0.5082943762706347,
 0.6994184810515723,0.5875185730695095,0.41904712045801784,0.3088637416271752,
 0.11913691250794456,0.8404554986306674,0.5574360961813082,0.4508298545007484,
 0.7263909384783149,0.7114242758675,0.7379194079102962,0.8784077391157037,
 0.7822081034539932,0.1240201986602435,0.6255585785999574,0.3450198739869105,
 0.9981233929686278,0.4464646193919273,0.11524069682075444,
 0.03799143341057376,0.4587698088943588,0.08369120443110944,
 0.5965837583446711,0.7103664377325265,0.9374108062128166,0.8538365392922033,
 0.2154581485885554,0.5148612249183057,0.6860912461280869,0.6223636975908726,
 0.5196357219840383,0.8442457435214485,0.2931242241296662,0.9479147430641653,
 0.9593367429418453,0.5609396228645465,0.19246627196769228,0.3736187708102554,
 0.8131452196242803,0.057829433852168544,0.9610040242464204,
 0.07407624615191899,0.05352336753965903,0.6831549733655969,
 0.7864446397964195,0.6800315059157612,0.2313294442924969,0.8076420030462949,
 0.19029357747585418,0.9932000809605772,0.04110876474651304,
 0.21089175869018872,0.2403817888155122,0.9395840006107767,
 0.35962993033446267,0.517391883833151,0.6585016536281447,0.7321551459717376,
 0.18297364376378944,0.3991923511339386,0.38138176398816603,
 0.3136980621293416,0.5964603790255754,0.3929923164450235,0.9926112935541519,
 0.024476315931173942,0.38451289349257856,0.007654056086154992,
 0.9795141423818137,0.8851884768957187,0.9576744716649032,0.5950067093257223,
 0.29840909771493096,0.08616735633520067,0.7964769402856987,
 0.9967258358474325,0.6477144286690675,0.29689601172577706,
 0.03846255460800496,0.33211333496172846,0.8076650589028604,
 0.11680934836873758,0.3471202744317534,0.9349591779634228,
 0.08557099441202842,0.0489191651757207,0.6329353452284775,0.3718287050949429,
 0.15347331699937017,0.311576621754297,0.19622544733414427,0.4137255587845904,
 0.8868365039384698,0.3121124847172977,0.39378949005893105,0.869302574537411,
 0.17437064929171953,0.7377433085081465,0.6738045812038723,0.8648263920206781,
 0.5669543264181254,0.04245210152933465,0.022270991362719594,
 0.6717707682296518,0.7844632706355955,0.10176622557146664,
 0.32693618627262977,0.680160964208409,0.5265127440333424,0.8734957975905355,
 0.22431262889882664,0.2059149637168639,0.8730723773126896,0.2832331108180832,
 0.17915863580706248,0.4492801030732896,0.005979840813383541,
 0.01971059142447351,0.6379585748130961,0.950974455278301,0.738956919479227,
 0.6765518977065006,0.5356524306754941,0.6539277608324154,0.1281458715892856,
 0.2542627191454858,0.9539583581157094,0.3514956570442098,0.8874528152608374,
 0.8274466642503996,0.867374710803717,0.12424904143255855,0.4956612299094124,
 0.05762920307109709,0.14920299009621718,0.7226800099323033,
 0.7760797813336935,0.9734129957731045,0.4289082018722862,0.8780827202279349,
 0.27361518950275354,0.9336117919185036,0.7875452629594841,0.815375155038624,
 0.2607413706716776,0.16361965935629996,0.7271661496814128,0.3918962325672579,
 0.5043679279893262,0.9710319736522977,0.2532034096234126,0.3964714414047653,
 0.15253369489010882,0.023717959618224427,0.3758364334782355,
 0.9591831804803228,0.6572942996354805,0.8212912206466496,0.6246776525603674,
 0.72748556955202,0.3145525180448725,0.8880997345346328,0.5950548928627072,
 0.11100777904986303,0.057035536073843374,0.5822843589499398,
 0.045389548312686157,0.29371250094248,0.22978082541058886,0.9383730991804483,
 0.7131239457850337,0.154571356252434,0.7013497982438972,0.2256154706700152,
 0.8159080499850999,0.3618944576285368,0.9046550228165617,0.37101987306100837,
 0.6614568336237505,0.1554070166583753,0.7348718061844258,0.46353972586731096,
 0.06093841183778159,0.6333107523829843,0.23224537798989198,
 0.7526077656816368,0.837535899223079,0.9259522002415458,0.38757066855787514,
 0.3842151544541421,0.5924861766703682,0.4257358171429305,0.07780800272370592,
 0.7697729831480575,0.9272578491700001,0.020652572695626037,
 0.2228104349145129,0.27841842561063757,0.40010351247507336,0.692636580767809,
 0.058202698908472406,0.49458506261364366,0.6421310606156,0.169425026635171,
 0.5078737236151621,0.7250112552367489,0.5444683281923166,0.9151910184608263,
 0.23629257936182513,0.7757716541193144,0.40845047030336157,
 0.0707361386202382,0.20737398513381144,0.5475278770419229,0.7093935451247387,
 0.5919898886777336,0.47575212328917105,0.07248845211468158,
 0.25604012490825223,0.8658542023202711,0.7588178695685821,0.3172319096945604,
 0.663724047088716,0.17776719094048854,0.31332771456274433,0.3540725163349463,
 0.16279171993521058,0.3628714855103442,0.7690992216672392,0.2834871616363803,
 0.9599889194637014,0.38250304104761357,0.9574680468468539,0.7142196147129484,
 0.39625955836001414,0.04518217592105156,0.07610815397596837,
 0.32258397638035363,0.27348619791859774,0.1476189959758193,
 0.9115570283574915,0.023532597264880772,0.7049726212526846,
 0.6444682947352881,0.31859962963992905,0.4630409059123999,
 0.053629658136934655,0.04973409513667948,0.4813966757218153,
 0.8604274396358423,0.77048915099713,0.33756359788474344,0.5475282573044921,
 0.13107111472041588,0.39742926270364853,0.6116453005850955,
 0.4605929513660061,0.7290703954760254,0.6822632976716131,0.6160787156859561,
 0.9653310789006893,0.1268393737410597,0.435902752860024,0.5615154494025992,
 0.49900096523352366,0.06737645155921701,0.0340046045398914,
 0.44703470457858563,0.7284158876891172,0.008375758222780694,
 0.9213842166647476,0.4825152702137754,0.9921722047674703,0.2544247155969551,
 0.3250144905038286,0.7399309077829711,0.4807705951141875,0.1858171089290953,
 0.8304172242230404,0.5295832148615819,0.30955520967843275,
 0.03998623212855401,0.25321577786601346,0.40268432758180495,
 0.277504482915085,0.37363808304962653,0.9312692281005863,0.16274109976190854,
 0.48118391263575355,0.9090212302965136,0.7245323608265507,0.5204135421386722,
 0.33775393878671367,0.5769203265034789,0.8965586262131132,0.7949707231446396,
 0.8571454671667962,0.7272976504741713,0.2576161594601247,0.46417355267872096,
 0.7607567130791557,0.730884491937131,0.5216394151550479,0.08147471305114207,
 0.07478284940199087,0.034227300126921656,0.7222185186894284,
 0.17032772210114266,0.4596362325003409,0.33555311778356367,
 0.7211193616025278,0.507341174233721,0.42978953100163164,0.8073192004520354,
 0.9435809850881982,0.6146805820232084,0.713425245160347,0.7828876823090989,
 0.27200509728743594,0.7278258932038584,0.044983253377149035,
 0.024913511263812915,0.3851540575622058,0.312663670439701,0.375120268495713,
 0.726176016283097,0.08247535105305714,0.503835521045757,0.4042668363336326,
 0.588184769540287,0.14266025004675054,0.572274918027996,0.6708430211070782,
 0.3721391932874799,0.919980198087021,0.5632508296445524,0.6094324836846889,
 0.44059572297385463,0.17419272289467536,0.5644997639307305,
 0.1830157560654172,0.11047997165921353,0.6738309354866349,0.3284088931136042,
 0.3843123607616755,0.553972098135573,0.3624623793587137,0.9895235920899834,
 0.9689895630111186,0.3285690089321297,0.28532179345679376,
 0.44038914038845356,0.41151216407649316,0.09354219683438325,
 0.9565126627696932,0.9796227354551006,0.15910412990794476,0.540356295724192,
 0.18704711764683157,0.490542705079308,0.6256786021531842,0.8576451955497522,
 0.9358114395593573,0.139771192982576,0.6238173585809759,0.05809433820442078,
 0.18204080069287587,0.0059234102333957095,0.6993882132765389,
 0.4662924991049615,0.44056241186907796,0.807192401791276,0.4467373486061881,
 0.2702822936093424,0.21938378794461888,0.5946842074364506,0.6263013422846826,
 0.7673815488162834,0.8494110360616742,0.907329732670618,0.5321824329270506,
 0.4296889009635406,0.17558573776559738,0.16285522644891626,
 0.44225538758034577,0.5846696257458235,0.040936639339777536,
 0.9142125977495771,0.9000651187262868,0.6822757638713945,0.4800558525831006,
 0.8579700857101767,0.7348713995423352,0.2674453602507485,0.6349561350466832,
 0.5730032812383307,0.5221408766950745,0.41579283915548015,
 0.39995073651430046,0.09865638217257877,0.815218309536625,0.4530374776631618,
 0.25147377151350025,0.12766433024269053,0.6657578129525561,
 0.9913546369085933,0.7151779733330029,0.6281936239913132,0.727884113340621,
 0.9578360745145008,0.8350565493427052,0.2962125712719741,0.7194130755506465,
 0.2871690240118827,0.13691361016536496,0.8533436388149689,
 0.09764133719007184,0.8074798255908572,0.5038801301606928,
 0.19407028486989664,0.1362984847428812,0.0940852947112043,
 0.16130420168423087,0.40315341999588195,0.9116534283967724,
 0.07130415337746754,0.03658129791650988,0.885499997340288,0.7157661933506412,
 0.18107328974519554,0.21389365603133292,0.3169480664371026,
 0.04733347802778021,0.30291458756397316,0.12740301915421193,
 0.19231826126746088,0.16268186622565772,0.7756670787245606,
 0.05224309648008052,0.2956328523998284,0.3420114203141704,
 0.18790022991004418,0.6944425687316267,0.1885763593414349,0.9808393634051027,
 0.2135954335596808,0.5733442092046128,0.2026804926647714,0.5687858115393778,
 0.02149042954906588,0.3772046016234394,0.9316540363502511,0.819484445923,
 0.5147584371058036,0.3146142191587231,0.5472245116476004,0.37086251659442016,
 0.788311114510367,0.6573128864700167,0.6370851780901902,0.4342971991914827,
 0.6999179958276087,0.5459320976761066,0.6437258709960397,0.13389126977448473,
 0.18037575495143543,0.24954326793795567,0.06507640856928931,
 0.2534947185363352,0.3710177383630259,0.4605912230155088,0.4568983953904646,
 0.30774599219101695,0.6184625881011732,0.025867211320592443,
 0.6590866014644887,0.8623394436728786,0.1090684914925224,0.19958484276490185,
 0.942420858981418,0.44447913613234846,0.3836189454977589,0.9464984041883053,
 0.6017444566111385,0.16733826210820657,0.45071891755141114,
 0.4598973174338704,0.9483609143388156,0.12675110939458423,
 0.26694364664627657,0.41490328749910177,0.42571270356419677,
 0.43543367263862665,0.22478749519966668,0.6602099066450867,0.904693251050184,
 0.27137075896498253,0.8779420992257312,0.012198580729907094,
 0.36820766092622204,0.19568034252304733,0.46522331063427147,
 0.06392550131954966,0.04240024699118017,0.35409745332110243,
 0.1631597728552424,0.3428413377244611,0.9712945635173407,0.2488904157378864,
 0.0658152938950658,0.6773738615533997,0.5667037096929313,0.713202160113324,
 0.19276878156108057,0.3763133323193518,0.8027386198756227,0.1470933909008847,
 0.7052331633961687,0.892312537624474,0.6334913013064156,0.31291521698262725,
 0.836515226316224,0.37715758839075086,0.4070129877158809,0.6727996355664863,
 0.2844058893273853,0.745564079172179,0.7566568351031675,0.7611947667700739,
 0.05233542956311277,0.5550221464592904,0.44502909625582954,
 0.7059481622265419,0.6704570745381362,0.30367499124743413,0.7474870993991541,
 0.367395257668071,0.49429256626989093,0.44495465264625933,0.673187505660179,
 0.6631466954016993,0.15693324513338014,0.28748857204657075,
 0.03304385644536234,0.8382983161067019,0.7692412346653951,
 0.17557805049662634,0.22793011987596346,0.35152201476639244,
 0.061005969596324316,0.5828491404335259,0.3636668038338634,
 0.8446319158882215,0.43564768077621874,0.32225812301516876,
 0.007597620436830765,0.6578583089622861,0.78384462729835,0.7675023521461406,
 0.4574408517346552,0.6080912598054911,0.7961542195466165,0.7679140968720837,
 0.1566422911975076,0.0097411542826783,0.14650337270370006,0.6614954289418995,
 0.6895528507821823,0.2678236907471322,0.3876399493265793,0.17377514825205054,
 0.7960804925053563,0.6734420300173388,0.8140580573014886,0.4032677865480129,
 0.7029648780575505,0.11780749746140073,0.15791253600340072,
 0.11419916442522826,0.4559762487926128,0.5239849907367937,0.6222795079109407,
 0.28610883748992144,0.5557975381530867,0.21292759817485296,
 0.5280030762992918,0.9681176211268663,0.3354889692220875,0.5111883628014829,
 0.8969922749836332,0.6584088398243585,0.29333320551419084,
 0.022018240607917017,0.8244615493620102,0.21326114906050486,
 0.8324834594290169,0.3685700133456514,0.27915681126013425,
 0.40128369180957035,0.11910945571154918,0.5991202132671002,
 0.7639264251964472,0.916158554190291,0.6905524250781813,0.9154944885519749,
 0.5069886305458025,0.03252611103469394,0.8095992772533233,
 0.021360306989756217,0.591800277110611,0.619948527337092,0.38126158971468094,
 0.7422447223393386,0.9415706972781657,0.7337216195767332,0.6614254262917352,
 0.3651752654764614,0.9435174234663382,0.6194216582014246,0.743876608729764,
 0.7295883074456444,0.9201309240166968,0.20563406773532888,0.4541082400237615,
 0.5241249108931036,0.4258305882154042,0.931996928909218,0.7964799430317551,
 0.8017801390392933,0.5035284254075776,0.1385145329376004,0.03485428760754339,
 0.18607449879499416,0.5954600718003076,0.17552316748870322,
 0.9705739442572445,0.6383762828796019,0.3660953460438965,0.37530715211290444,
 0.6037128321250149,0.47534539396443587,0.9236573506627516,0.2633717691389664,
 0.4942629809483573,0.457355568067354,0.49898467926144774,0.3050825551493678,
 0.7370389110430473,0.41953248077750704,0.501848249700177,0.8078777469056826,
 0.05569246764683955,0.4733194725339044,0.2866316900384416,0.5685946268237132,
 0.24637202910113076,0.883977947616774,0.9543710558747072,0.8503619827382146,
 0.9093852341204676,0.13331905109417796,0.686748216739252,0.9185832358577912,
 0.7599498406005887,0.839532227950861,0.7663846114992423,0.1367911788781987,
 0.4959439425927966,0.6240504416727516,0.9024500232630658,0.3969880560421988,
 0.38792651169760406,0.5992241017638998,0.8619221953393427,
 0.11217448656782592,0.800212423528408,0.8749954847289516,0.41348126614977765,
 0.7253885389150998,0.942627365935544,0.19398151914839035,0.22745343389634132,
 0.7030769975572743,0.16945780627244622,0.1948853343678445,0.6892358103467212,
 0.16438515845609591,0.43265577502265495,0.8572306825293088,
 0.7595194390587385,0.2600818739655867,0.09324202930551428,0.5424717348704374,
 0.6433862592463172,0.9218093159934658,0.688165715918342,0.10371235734821094,
 0.1518256847698286,0.5811747225206225,0.14637067125611558,
 0.34066779099431654,0.9889044549091435,0.6139893783411241,0.8973434045407259,
 0.711128673508552,0.93310438536844,0.9565453130828312,0.14086692200974504,
 0.44800571220667873,0.6044927745221662,0.16178292622349333,0.663425015101226,
 0.35003473595546886,0.34800097870848423,0.9375460178301067,
 0.6418276184210681,0.4943527132907686,0.41270665616770663,0.1697393480929017,
 0.15106092924360737,0.48036672413721826,0.27029534164016655,
 0.4384733965220414,0.36286550596181655,0.10221724940788968,0.647010119085735,
 0.3435491630993718,0.5894097838420415,0.4425006020401092,0.4049992942389,
 0.951425132439722,0.3776425966465715,0.4926430720233411,0.6450417660426326,
 0.375865124504803,0.9014150076782126,0.2105089074027533,0.43340907275861895,
 0.1393847552272507,0.9962835440069417,0.053052117084327555,
 0.5117121221297399,0.9862750186905542,0.8177443788272278,0.18220489698271247,
 0.5855994365005962,0.7438977550928325,0.919937443178535,0.9317896310616047,
 0.46365585975399115,0.7110309589214804,0.36608455130805817,
 0.21259184281182097,0.7146486192059728,0.29607783741067795,
 0.6541585450704289,0.82592408579811,0.9512904313701017,0.75028995611935,
 0.38529725518447866,0.44910598899645815,0.20612620558320827,
 0.671356931440557,0.6261377816626743,0.4563529669263262,0.37549227517212524,
 0.04339535975275555,0.10563925560533827,0.7520228297038373,
 0.4723235682168051,0.4028635812577208,0.3679155198508921,0.23604843025476285,
 0.646378669266504,0.24029855825647262,0.8453774076870952,0.7538586696207246,
 0.3301574325091845,0.8165023632106159,0.037366287223547845,
 0.4141095861807499,0.8097356668014066,0.734199438630204,0.8139451527894971,
 0.5567005751926865,0.1539924472226969,0.06031677113792566,0.8134928723525501,
 0.8006249564057959,0.44926394168217787,0.5910155206334942,
 0.13002253514263462,0.055096224217483614,0.3331585967150307,
 0.8606280159544683,0.15889042775386586,0.5927778355976818,0.5314227470234861,
 0.7093082910898509,9.101189203903903e-4,0.8107623231562634,
 0.46242100951551235,0.9233027841427426,0.8792717841212173,0.4773006441691132,
 0.6239448872337203,0.9455050654369195,0.7628513384175496,
 0.006934825052972604,0.8992901318106664,0.722447666462879,
 0.06623334220846044,0.3905394097493634,0.21366283185988733,
 0.9655504394950227,0.2353326688794547,0.6231913868693119,0.07043584209758069,
 0.02811620765534195,0.46207071603012284,0.8360998737337151,
 0.6173451737830745,0.6177299638156155,0.336985436698118,0.7206559470801893,
 0.6052520625641049,0.8376215108764945,0.774132651245689,0.5352599958267288,
 0.8237307462346793,0.7668900580670571,0.48373004297026,0.08318079475361795,
 0.1363467410400413,0.650964755900278,0.22186563528169234,0.6145983368490661,
 0.17272885105033753,0.5388136443994673,1.2859762904504635e-4,
 0.05758034860881345,0.00634417019543787,0.8307811747879609,
 0.9726253359152714,0.48306927461789584,0.20081794271661102,
 0.9397833420929246,0.35528929256201225,0.332363099816659,0.5698969155258445,
 0.19172655237141278,0.3840973553290099,0.010695606279507275,
 0.7905564301723645,0.17805912577189764,0.8032864545629048,0.614140720087754,
 0.6216548667031597,0.15938407057879922,0.6702805445859813,
 0.15983011855706342,0.5675344935638239,0.32675140010908243,
 0.16103074336260004,0.5516122829561471,0.7804200492701338,0.6153577853283869,
 0.9028633574556492,0.5816525852591058,0.4787865610787795,
 0.057438442131719225,0.1180850636481181,0.4846041765303746,
 0.7867607780131035,0.9504647917062765,0.4630683908975278,0.24328768118309319,
 0.6323595864653184,0.5294705436818059,0.7235285466976967,0.9198501388575634,
 0.23920635041114136,0.4062903363033581,0.09116526484214271,
 0.8186715953332925,0.3832421379198445,0.330525049631188,0.580922058086809,
 0.061272491747972,0.5062468336457984,0.1141884237615226,0.011787924066068767,
 0.8880154896317671,0.27872174652125137,0.5585640258483965,
 0.33266531667074406,0.7870531511863109,0.7854671802469279,0.3950292852138033,
 0.7697457547351126,0.7790086843154694,0.05821231501948154,0.2879352726333373,
 0.9316000636904284,0.9247942465792197,0.3780992810487298,0.37436293184992675,
 0.28428195104480547,0.4021953191430525,0.7576545692308191,0.3094143800793725,
 0.06478646546553579,0.19455244094469226,0.4283109281474554,
 0.20253849428388082,0.5688866719293565,0.8475182797030488,
 0.37062334247819595,0.6215959893263981,0.2671944846056833,
 0.17564200827307686,0.44177166088649644,0.9220440820729222,
 0.7336020228603413,0.8026961686692605,0.919821324022007,0.26027726960976116,
 0.09519950696184454,0.1222097403670751,0.7701845849245966,0.5657353020471997,
 0.2592571867914102,0.6756670528387875,0.03833616195123524,0.8401254708373818,
 0.5089273057826385,0.3207520616800539,0.6186873904683796,0.5540404397580512,
 0.1316696962468039,0.557380655514669,0.5241203803650507,0.09330862803908002,
 0.644836949273308,0.6366557656178822,0.011114817455013415,
 0.23336065584446253,0.9404892545151696,0.43919627269859984,
 0.5636678179866301,0.317066722915737,0.17831713999256205,0.852852086598756,
 0.9464001457060671,0.9386902624161648,0.4222438944766369,0.2871248838393782,
 0.7747256724859017,0.787744729200618,0.7803195367250648,0.3458929211514915,
 0.5709479319795518,0.318403748823579,0.058399783345540524,
 0.23273877669206433,0.7323029992929637,0.9001881770617935,0.9016265269362325,
 0.08035236867700146,0.3656785547422502,0.38006774252464803,
 0.2115581970152839,0.5154915607688523,0.42966240975759273,
 0.41081916005824326,0.453081441156058,0.5806796561998981,0.6422043807302878,
 0.32887446932814124,0.03366746977448132,0.29288182105808713,
 0.828884757037969,0.5399026575271081,0.5287351720939722,0.10655362774215182,
 0.5276040049302886,0.7585344484325919,0.16338612264292685,0.8992053868211133,
 0.7858861595617296,0.2826761986198869,0.9431126061905148,0.07865570296228586,
 0.8991851777223597,0.9073549079059969,0.7080637109493589,
 0.057098266917502305,0.38168253475721914,0.13051715618804238,
 0.9532173868197799,0.7519387771818132,0.312809030319674,0.6844243337447358,
 0.22307098147298565,0.2206716968101109,0.6544736897091883,0.7310193120041698,
 0.12673132845583157,0.22839438603578022,0.9914634967670782,
 0.3944882359397083,0.49099282991770654,0.5285894837064848,
 0.08687383021607042,0.06161716918173887,0.06047697479261849,
 0.3385281896533924,0.05723328344151479,0.28601544242016985,
 0.7802556257877997,0.3285097673991978,0.6314938636440441,0.927069266566324,
 0.5820068451740986,0.577296557370947,0.0758953671425332,0.721682033471584,
 0.385419706573349,0.5510148504745468,0.5410395076647208,0.26685033018972193,
 0.16087796409057198,0.8805882615866698,0.302069406689111,0.16635270431706062,
 0.515176577899972,0.5956510549103853,0.6987424617506046,0.6459442390801785,
 0.3177099003798407,0.6848443248829665,0.7312061683148492,0.08341019796899218,
 0.7562639862729494,0.3047436315607793,0.19016075952055467,0.6990947469240653,
 0.1789080153719832,0.8298395115046614,0.2768580304795908,0.8831109487059218,
 0.8263660207785114,0.6311517700169922,0.28520710246742065,0.3042352766783334,
 0.05354642154950029,0.9486843143756163,0.35497425908010183,
 0.43212733364903433,0.09295402871183767,0.4715556707887525,
 0.1721641827212026,0.008131206906792032,0.23323379456364146,
 0.467269949150638,0.7027858788195653,0.7551033417974117,0.15705803834370502,
 0.015224396859522926,0.9681455346739539,0.5613622301783351,
 0.6193458167008514,0.5641054211788649,0.7203460536214731,0.8278937244113379,
 0.8816712230160976,0.20064273400562616,0.8025599437771002,
 0.29488099995626615,0.37474522150660194,0.6435543474751788,
 0.1000935667356948,0.7459759666440142,0.43387829055508453,
 0.20667276203835983,0.580242467057883,0.9942584744954035,0.4469750326079541,
 0.6468077454746115,0.19578179284888453,0.2778997928857966,
 0.32890873220642447,0.3298247708254989,0.48319552478256445,
 0.5908518283226957,0.6767077827426029,0.3198616355516869,0.5392985880634623,
 0.6766767548416438,0.5932462166895831,0.2949118076753492,0.7326155073740285,
 0.14221784855571717,0.7992157376604707,0.21309249615840375,
 0.3180533457870386,0.2228537929439416,0.8762884148925352,0.2471595262272105,
 0.32178472833266336,0.3646837878880045,0.6859346253891065,0.9092389641341347,
 0.7445342107407387,0.5680356911282316,0.4665773753914877,0.23509840663896153,
 0.4912238068451589,0.8345847823522285,0.8668074580756941,0.4971852670329957,
 0.051120397300461806,0.6548479351440629,0.7653900713182501,
 0.22259066407952444,0.3048373725289074,0.39722770427913656,
 0.3713646865643817,0.393319398602597,0.7544590279276425,0.3129358469518607,
 0.24021821328788417,0.32231757744882017,0.6875437251208248,
 0.7372665803880454,0.06430853595939301,0.5438494648696073,0.7197439326340358,
 0.8137336679135898,0.7094942531460838,0.18738769950619827,0.5128882441275061,
 0.1847011650667012,0.6073283064020759,0.6830852094453023,0.7362809337855345,
 0.522970764082245,0.8937670931726719,0.9802371495228935,0.30233087572225004,
 0.31906003012586304,0.292854443812519,0.2236286792710136,0.4840560549110212,
 0.24300232784816544,0.714742516274476,0.17647189403744734,0.6362056334796836,
 0.5082851673213339,0.35284559171377006,0.6233959235080119,0.4866260550516468,
 0.9761555356680144,0.30500057181969487,0.08546507026531358,
 0.6839210083271814,0.07749635132577892,0.788795674279099,0.20230090341346219,
 0.9737152511351004,0.8341394077095808,0.5736483656497541,0.6644209160320174,
 0.6392790901433592,0.7494817011838918,0.38139817884793825,
 0.43763237189666804,0.5589911462198675,0.3430991442427336,0.9087944042402489,
 0.3837572302358401,0.574612285889809,0.8343230043511919,0.539807433669305,
 0.28305130959367497,0.6298256977688871,0.1478861822359432,0.5260510983523262,
 0.24959532856031363,0.5248146498436457,0.2520550156892073,
 0.008830124323943478,0.9518939598007323,0.21056661050849312,
 0.5312148050060551,0.2831923317591398,0.6358187831418779,0.6669001632962641,
 0.23258518009035345,0.8023639183793839,0.5432679990447414,0.8610229008493249,
 0.2021345624578459,0.47519352432627326,0.5845609557219356,0.953844905763391,
 0.39317984905953085,0.36950415801340053,0.44550440974888295,
 0.07188665843856323,0.9589237168995899,0.969062689832533,0.35862929927311793,
 0.8868020089442306,0.0380325151423857,0.9563419778820574,0.4983283815859817,
 0.1454811680788538,0.9814076776099967,0.273666694748702,0.5003341097826032,
 0.7544488364914778,0.16546957900198977,0.06559167404212463,
 0.09699622317748591,0.8442144548179742,0.14275503718745908,
 0.8080763810438105,0.15832387840222584,0.5436344605812687,0.924267807868352,
 0.05644242190162252,0.03634394806444963,0.18213252518312129,
 0.9334703330148547,0.5819919128090018,0.6423863553093752,0.2210935729485617,
 0.6808303814416248,0.13063493029923712,0.269581543968925,0.08555164641841118,
 0.6339164446165348,0.8755023861935355,0.5035719868914931,0.8265973839735972,
 0.6980717743661707,0.9169973303474107,0.8323869605660809,0.4168660199870512,
 0.4926748507711016,0.9406923420394152,0.9488271454005601,0.5219350756012467,
 0.20672487156341168,0.6454094809481062,0.4244280477123531,0.7218606354943442,
 0.7646296499627439,0.6198967104605262,0.20433174774128715,
 0.17030629994968172,0.2640358655096169,0.88584793342587,0.4456826677975927,
 0.87748058127358,0.9302112036160007,0.9323355273264627,0.20470141033825195,
 0.39662973364129306,0.5228063725617118,0.8113466832805244,0.6017599533691323,
 0.7949931886764566,0.5047134407048428,0.8725040723078168,0.40737641740376895,
 0.6800841313688339,0.7328542927582609,0.9246682527139216,0.3318807173158653,
 0.4177946269306756,0.9179194783130881,0.498440881126718,0.28933293768008406,
 0.4995035704027998,0.6945220824006866,0.6975981194846262,0.8221783007994212,
 0.5024643485525483,0.8871179680690587,0.24409876868625413,
 0.34852512473827524,0.3970448821915855,0.9520898941572365,0.7424651270806273,
 0.4083124355616293,0.9931722255120419,0.7419563753243577,0.41940899125878506,
 0.4757180996424042,0.9446957183912829,0.6432536332102137,0.17049468298911874,
 0.4316578238925798,0.15103345526618017,0.1675337961064659,0.695840657820761,
 0.9610799911353607,0.29171733353782425,0.3882392806139128,0.2816883050529242,
 0.5702716013418806,0.7957698995014411,0.32297716377415553,0.2622888360580943,
 0.8808105672275095,0.07237027867346768,0.11788355608143886,
 0.7490867618478733,0.754900854271536,0.6650447400518,0.787267985398807,
 0.01907115375932522,0.1254910409877562,0.566971926502304,0.11860420108009784,
 0.09953694071624986,0.3084225559392455,0.18452769844551398,
 0.9899749060459101,0.5227147842995205,0.855022888544112,0.6798053229073538,
 0.08094074857368072,0.4835102506730775,0.21831447205578625,
 0.6313365013873197,0.3954800048168331,0.9191760177341394,0.9447067396422523,
 0.646820217118315,0.4007376770254816,0.6005615984717656,0.42088384759444275,
 0.1694037643522358,0.040171199080682785,0.6507955480737193,
 0.44554605825752736,0.972753284598663,0.038861562447453535,
 0.9873756327507521,0.48420471309201774,0.6703474674642852,
 0.47240120545061937,0.8493417255024378,0.3780955124819112,0.3609809649807353,
 0.7667040961843523,0.5904275108433998,0.8766753961383611,0.38593746827065023,
 0.17430665170661264,0.09958163602235792,0.9659230280996149,
 0.1300229612920636,0.3030672049878059,0.030781670084421897,
 0.03847815650585662,0.12361033283178158,0.32297319981088446,
 0.24395610849354854,0.6992794623702204,0.12894147526888844,
 0.2295529800259387,0.5404476088685012,0.15828218710194297,0.5625296816964851,
 0.5067881327371715,0.7028095283064902,0.2832070003956224,0.8306470810563661,
 0.929142234891182,0.368108057132396,0.8485037648262095,0.2367115827608356,
 0.1330856463795721,0.4951569819861311,0.7794431606966029,0.8290569736509406,
 0.36168091506829925,0.2072554327036824,0.6912836024333827,0.8898769446509242,
 0.8999821805113816,0.9726173458809576,0.6260663879755125,0.22230154121647128,
 0.21113930283333937,0.604889407871453,0.14027758648910604,0.915421874719416,
 0.9890147827820637,0.028718760586737058,0.5786055520220015,
 0.19700538678488067,0.12556197890545873,0.7800987361343307,
 0.6853543054233409,0.9262320176875493,0.7712114687035415,0.2036439885899074,
 0.5201753326626509,0.10433807470366396,0.7473583124238065,0.4799695308633112,
 0.5658996055678877,0.12492604367879945,0.9800027014807474,0.5699949073204954,
 0.3470566223913013,0.8574406010523346,0.07713794475883384,0.5570730351849821,
 0.1140177416924768,0.49607559912038035,0.8201431910202651,0.5210484794198098,
 0.6848598576936024,0.22019465154917328,0.2391715894010603,0.3516955716381247,
 0.5693719123989868,0.9714657471605327,0.14330271850770493,0.718452039886669,
 0.711232386182024,0.8853853110339882,0.8922794685407945,0.016813361055560194,
 0.02839551145883179,0.44323094215336734,0.14279519301189636,
 0.9793646126912121,0.13772622202058749,0.8879770506392952,0.2721865764025262,
 0.45872968457182106,0.7970131998550641,0.4207112268869785,
 0.11856938246740534,0.14080709479748,0.1444402425911122,0.3264605626903898,
 0.2687928978237206,0.5581326040726681,0.08686395092473986,0.7937345661466765,
 0.5121659431091289,0.7964989975612559,0.05592666290073378,0.6512949142495664,
 0.6956611198690713,0.5920924081526949,0.5537288001873231,0.46183984048968296,
 0.4313719546150331,0.5703355823039213,0.9021159531955333,0.6347166905745769,
 0.12077885168941582,0.0826712858986064,0.5440655109935827,0.6805383283329628,
 0.7209094052464502,0.5178320181798152,0.7009803003656381,0.6042499482805173,
 0.8264158807336255,0.41695230615865086,0.4820212689665384,
 0.44740708980802546,0.14659631821138064,0.17870301926506793,
 0.04615117911212163,0.2758913175058224,0.21263222744865662,
 0.4122609777675268,0.35337271017025174,0.8693200087088754,0.6922020878647053,
 0.18003942414730645,0.43948364520373073,0.4119044171358598,
 0.3126373626479162,0.618986727470285,0.4787355161530338,0.8436254372451012,
 0.3283641980964087,0.4044312996367936,0.04572076206608022,
 0.36567705081457036,0.57217244725428,0.21493608211054382,0.45453573049129825,
 0.8012504208690544,0.6181462863128226,0.41613438333793307,
 0.023106382955716986,0.38852562330187257,0.5093887839838122,
 0.8759172230800184,0.6908349384274398,0.48304444853889394,0.6585004093137826,
 0.666977094847574,0.23188084247521124,0.9198859156129261,0.19353375164522102,
 0.3923709712316571,0.20727311494256961,0.33438806490422646,
 0.28653562283305645,0.8144531409577187,0.05896001901028436,
 0.6748531471266175,0.9634993694948317,0.6394511323216356,0.1393191807505162,
 0.21450600749182425,0.5842974579735293,0.12700829583939333,
 0.7029878283887316,0.10804632297718686,0.30375446734565825,0.834711856534158,
 0.33980436731854347,0.7120909318751998,0.9439084525154138,
 0.27557482128368505,0.20618941254406453,0.9951144449665159,
 0.2858271968081807,0.3220811784280122,0.240003582288419,0.7904175077366927,
 0.1672467811687104,0.015254400935355794,0.06947870840848636,
 0.27580100245852246,0.4568036138987912,0.5607834158546889,0.7374150670373433,
 0.8716176915864742,0.42938061118261706,0.5949561643556085,
 0.13479323624795603,0.8021944738826341,0.7097409406950568,0.2853074925682457,
 0.7049825501838846,0.3097231689302615,0.5848703149077252,0.7782765581487224,
 0.9280402106373362,0.5557139859804436,0.6369481747999003,0.01412432800896124,
 0.5174457785970444,0.2569872459865171,0.24053695109424333,
 0.22116465145312714,0.9482625559156243,0.9442082649772342,0.1309937511526852,
 0.03413197620567643,0.7569139312430978,0.6252168200734173,0.4720891992845322,
 0.4962324385478244,0.39741378677748995,0.27037056660311765,
 0.7715339355926519,0.2621648804025394,0.890663089066797,0.45090386293478,
 0.13098853475680017,0.13532378922215815,0.8327421238833685,
 0.19488654858973575,0.09299413120452393,0.6739909155539443,
 0.41030992215591666,0.8829693399026144,0.7126017970323819,0.7145683039837326,
 0.367784504341251,0.638048183381011,0.06453865525166713,0.8865484128944305,
 0.442310304335375,0.6591968119878248,0.932780948262605,0.3978232265410566,
 0.6782185302738869,0.37971413704627166,0.8378719437198624,
 0.18496705146549497,0.5266978547670389,0.2569177148191476,0.4845851757258194,
 0.09248958575275901,0.6852601575082113,0.24226452359691986,
 0.6005875003600304,0.4290799267629062,0.04992202746538266,
 0.13308557372394125,0.7313639247361525,0.8807968514644573,0.5171739669875134,
 0.4155971788113546,0.3920534667374699,0.7364330310368716,0.3547356652710032,
 0.544443200739375,0.01133376216210391,0.1798371079049934,0.3601707805433919,
 0.44539229273874614,0.8986045951754819,0.2475883600337455,
 0.16037621194528395,0.21946687792189457,0.9528386986011584,
 0.6807072688250964,0.7885094117631128,0.127273710640604,0.3671855186607449,
 0.6672505485167051,0.10199725022745199,0.3610029935527743,0.630708575613911,
 0.9886145813878608,0.11090827814969284,0.453115283491598,0.17295609966183334,
 0.5269856587188442,0.5271122711917808,0.3684308000558989,0.4810037381757467,
 0.4928037676898016,0.2639560511134983,0.4474171552006514,0.8052008852536855,
 0.24218912332935638,0.6498980961153084,0.12317611813153984,
 0.4952824260762826,0.5472766106253243,0.32830916973658464,
 0.38432307486034145,0.3266692083201287,0.9812777636465051,
 0.42915314398189075,0.616593964828358,0.11933132430729332,
 0.48992311763064245,0.29394956878227974,0.6909705035363263,
 0.6800255514159601,0.5346583446155294,0.18556222572239056,
 0.13029341800463856,0.594215783234919,0.18910434230643958,0.8874794957227813,
 0.9913569319734289,0.1651079032884416,0.2159527257464171,0.4264687126015019,
 0.04949639626383018,0.20048856553504435,0.381234777754356,0.5709336099514561,
 0.3344786508911517,0.21496494490584972,0.6325940358517026,0.0679883095817625,
 0.3557305946758912,0.26163952857735084,0.3345383994620017,
 0.10870774033832487,0.9864285936389794,0.12052281506647988,
 0.9042783396444606,0.011162254239405534,0.6507396865796857,
 0.7320927007573115,0.4601266734784547,0.1548168698387391,0.7416008915507111,
 0.26548524403558094,0.038443521414843085,0.9205029476246152,
 0.2190178583202783,0.02537182761905543,0.9154751971944562,
 0.03545664391082226,0.0952468699562059,0.8468343741644221,0.5133286812134789,
 0.5100126942723836,0.18950030486067382,0.623077967970336,0.38042396268477496,
 0.20646556360816448,0.44667547378147343,0.9478519435090916,
 0.23929557705310944,0.20781425635594497,0.8594819839868584,
 0.8822274034730906,0.7115075322371142,0.1718219785310222,0.23598836289140224,
 0.8172852812263653,0.6439821800490598,0.8970282441073765,0.6169862703072224,
 0.7591594210767956,0.9407095521336079,0.9595518863344081,
 0.028227997592834897,0.8678898543367988,0.39654424458272186,
 0.7507699856072324,0.7163307871482211,0.22956817375188088,0.7066770242785682,
 0.6593372442870945,0.13558393717866168,0.6664075188166642,0.4941875159150262,
 0.6499775754792938,0.06492182110047251,0.35579027672280916,
 0.8919437454642928,0.9429557782052741,0.1933659191679561,0.506412789952573,
 0.3596436608617539,0.4939775512298794,0.15891976497896365,0.9991227080359484,
 0.5426283637529103,0.12797078674263695,0.22116585267125677,
 0.04503264575151644,0.11217507410313121,0.5017012948222266,
 0.5048734723035831,0.11536931685278529,0.0022532900991887317,
 0.0016779178193105926,0.42143180114420253,0.1572912812925572,
 0.783871760627365,0.11061451783912946,0.6387901035934326,0.71429177883545,
 0.6875687386520934,0.0867251856951684,0.7082649740949815,0.5681207307454521,
 0.35250432582026636,0.43231219709594004,0.2989579844589264,
 0.7364245713464961,0.7460823241610935,0.663256476823419,0.23217975944172142,
 0.6624200694996676,0.33521515671524327,0.7429193224864648,0.3993471390171416,
 0.4283623925846415,0.91201178899548,0.24472754170962163,0.35918375158018934,
 0.1409172968614868,0.6940779672193764,0.8477451565678115,0.24674507860259576,
 0.7505277959753714,0.47382303850959373,0.33573748290664296,
 0.7011063337874021,0.9459550769615535,0.3657133100298108,0.7791474229525052,
 0.19789767318185314,0.3806093281459103,0.3845214859641908,0.1659367994035379,
 0.41783526518145775,0.39520396966441385,0.3910536842638943,
 0.4650493004113785,0.7250151012338535,0.09200254523188345,
 0.31891713770277497,0.58484744943531,0.2608998033585568,0.9195024222341002,
 0.8937379310372827,0.5632609732832373,0.3375957601986469,
 3.485019587865512e-4,0.10085270958105497,0.01739801918097994,
 0.6717965568706417,0.1477469891358042,0.41087534688391847,0.7906718496508531,
 0.927625006079901,0.040137008753916126,0.6554659756632272,0.8999597988741805,
 0.7858434079035927,0.39147759715186814,0.7633283154710225,0.6900754877148437,
 0.04252785376578305,0.5816060095861508,0.2584744854892296,0.6458198272064383,
 0.09600568624572903,0.011130757390669599,0.22367157510454883,
 0.2076577679156879,0.42254812756429583,0.6926295336527792,
 0.20290606963072533,0.22682310070490652,0.7052356008264954,0.143770100402976,
 0.7024290135942586,0.008517700266128858,0.9895728789164183,
 0.9941742535779692,0.43682669757454773,0.4946734779582357,0.9020639347446584,
 0.13656085092439585,0.49280577171085316,0.6438067863837797,
 0.013375579075575295,0.5552972577191377,0.8007932566778485,
 0.2643657602096241,0.8496827119204466,0.5539311706493514,0.7792957899508669,
 0.6462331771982193,0.6213080979792321,0.24772363827008836,0.6669905503042592,
 0.7956369414126434,0.7438182476625261,0.5350858669159948,0.5319298129596746,
 0.7621718331624963,0.8065496815912998,0.8581209430569741,0.6976932572780754,
 0.9360988062679543,0.16423096114974112,0.43867306752283786,
 0.031867726422209475,0.35399761655630546,0.3548223549081052,
 0.6724960115436364,0.5769474611684343,0.725142837509291,0.4989738191949249,
 0.2869997505239523,0.11218848262726444,0.07976935734604684,
 0.6578704063216668,0.8338396123990044,0.45025033999447794,0.5267398379445914,
 0.5740904930846856,0.9622440785501485,0.426301918606547,0.8110135766258743,
 0.6396741106624712,0.6443633189389214,0.9934517623764356,0.8668633069944655,
 0.2335966217868788,0.16670429566663048,0.36312096536095884,
 0.17847658584664572,0.6163109584456438,0.6445223795357877,
 0.19993959181914256,0.34431657048294306,0.5437297825558381,
 0.7167850647952715,0.6901311592985149,0.9058175550498407,0.6353827813759623,
 0.6605751358993388,0.44194214192972414,0.4881127761131747,0.5268400909099341,
 0.2838414060343746,0.17802098631054875,0.5363066448432131,
 0.04478941435178718,0.8346447788821815,0.3168782285583588,0.8763522264486386,
 0.8710535696092286,0.050576613097179646,0.14174436148302272,
 0.7804527734159985,0.5840305772993061,0.6873893385601939,0.38692513863560507,
 0.9478477368876961,0.46062730280308883,0.9086495249444345,0.9712506761479391,
 0.6555702510213305,0.13202357191593184,0.6001675052879984,
 0.28019968390971406,0.7600090836753233,0.2018282218964187,0.810210346470007,
 0.46019022453285885,0.4395590334619195,0.4442902239324168,
 0.47532926662828057,0.9475817016323338,0.47856877676483645,
 0.6687752937688837,0.4451513443362537,0.20144940742291506,0.6017796965395639,
 0.6283440100726853,0.7582646225890853,0.37334588890606313,0.7811753900998873,
 0.1131705521467965,0.1390329491343958,0.1491598502885333,0.03282399470817243,
 0.6158558879354581,0.7582378937651891,0.04164392874297984,0.7599199776216315,
 0.2936254801645892,0.5431845563332233,0.9425824944309498,0.38231524410554274,
 0.8049664627476868,0.3444352769145632,0.4036432523292357,0.9475345431850113,
 0.44166125337840056,0.3279563774863865,0.8537728971328611,0.5911379049398542,
 0.0927211274101356,0.6399528303422064,0.8929685367383305,0.7832870114516541,
 0.37745166001512886,0.10065228275553628,0.259152307226866,0.5686219413796103,
 0.4523221512575999,0.23764898728242256,0.540634828741324,0.9031368005307472,
 0.9594353043479728,0.12265252235783242,0.7907238971416677,
 0.29345324643256476,0.7394976172380892,0.46931956296831756,
 0.4800593423512941,0.5035926726790698,0.9630695721222751,0.2152523831233618,
 0.8399994896130845,0.6850622811662698,0.21741859738554936,
 0.15713049044088812,0.9912812882894616,0.10882688771198001,
 0.05514038129780585,0.537471062456667,0.9889443757101477,0.8513470458855302,
 0.03923505051592291,0.5340491627849951,0.5005209646580108,
 0.23774253250215704,0.3037107493970751,0.8806319022344129,0.3048115374204451,
 0.9701677384244541,0.002743369723754885,0.16547255044703546,
 0.021715451452460566,0.5775771469197135,0.92229168495339,0.8019082162494459,
 0.7184317690057389,0.2693634074140564,0.2507681526257154,0.32568728471751296,
 0.5615300361678641,0.8611277227989753,0.5493248959990386,0.629440533870936,
 0.6101141467615102,0.169251715456887,0.7469491966843307,0.16703736145501702,
 0.24256682690928755,0.8581685496272364,0.22289310099050352,
 0.6936859630779599,0.7501733326625076,0.9972982894656462,0.24852142321094584,
 0.06660333478779823,0.7503027957694206,0.39488726199406177,0.763338435141682,
 0.1762407375437971,0.6381995621217834,0.8480879275021701,0.46359786914256074,
 0.2758677928640785,0.4909469018714461,0.44665309275559995,
 0.024174836652835485,0.6641733105882953,0.7902207456567067,
 0.7747914579745753,0.28230804001505794,0.06948039341609913,
 0.26254615403707215,0.619519963648465,0.6079692373632349,0.26776795774956874,
 0.20475293759886548,0.587964300833081,0.47460400320506313,0.9159666565489339,
 0.5281248143500575,0.4448067971705546,0.6481558142221022,0.08384764097931485,
 0.669710332506624,0.29743028502117097,0.8734555936511431,0.25537194982111866,
 6.191996990745352e-5,0.41967467590650653,0.5633223902998797,
 0.7429902948353462,0.47276011107874205,0.6633763725049168,0.6850460183668028,
 0.4423582056226335,0.662291198746353,0.0036740506919723526,0.136274624404626,
 0.17162245487108863,0.22767951223414018,0.7625092627708593,
 0.21443318735662598,0.24455137918335812,0.6994764385077028,
 0.00677779507505627,0.35644166880039596,0.4537794816477272,
 0.5276893698271876,0.4263384758492026,0.8140285365077217,0.5501187936838209,
 0.979913677107629,0.679921813990058,0.11307159898241004,0.6907598015565889,
 0.37514437842667614,0.4800682082627663,0.46955585844316594,
 0.09653520500011914,0.32350102188605145,0.3354132990446598,
 0.9688402641670767,0.8271733532336599,0.04357208223868503,
 0.0014444194157776646,0.8622383976326767,0.08841366116147054,
 0.5509615342221662,0.6812557491374738,0.6391812017253571,0.19011835992020554,
 0.09068264576170593,0.49981437863189704,0.5280160612960537,
 0.6675183965864528,0.3176166774153881,0.15688189321248147,0.656668774825334,
 0.8362490265222271,0.1421726787222638,0.7506226244824006,0.14716286929300781,
 0.5663778987326197,0.16137368162697974,0.7053917856860191,0.6296646740747377,
 0.6956042362057403,0.9538511711665403,0.265070571904509,0.8342898293229276,
 0.25412404370079933,0.250678457155598,0.08242194833524608,0.8806329930496984,
 0.9336797872131265,0.7945084331114328,0.507590326012872,0.3204204127196866,
 0.05072840395456635,0.484442803267776,0.09509817038879764,0.7075235238730224,
 0.22853329619415264,0.06562892238875428,0.6331089586645867,
 0.7926361833490736,0.16213372398811743,0.7498902303876919,0.3092608240938475,
 0.14683882580147734,0.7288075194621574,0.46583474563305394,
 0.3565505922329374,0.4518145186677722,0.6346571650360692,0.39917072549517696,
 0.4945256995508135,0.01033517820498775,0.5577094968162749,0.7439342078087082,
 0.3339512348179958,0.8277715044428945,0.3242957825192557,0.8263759814626156,
 0.07190384026478491,0.5253594890189779,0.5571277815542033,0.9289110710772714,
 0.2914644401790073,0.9221327862021719,0.2953870662652379,0.9987914534775543,
 0.4938801098490744,0.5041983086254775,0.26569137451154967,
 0.12717390728859734,0.7861186948324175,0.6938541062572502,
 0.39573488093437004,0.5098939609338158,0.4543330255865612,0.7724417450062859,
 0.046814691059756997,0.001051042593114504,0.6730431548189986,
 0.4126170697056287,0.5196499153375015,0.8394232442877142,0.3364473234154248,
 0.6754269722446661,0.9589794151670518,0.3162516329497568,0.3684831628000802,
 0.3211711189823858,0.3656919455993176,0.6019944572383596,0.44508246505157434,
 0.7179983382790638,0.18901043888062607,0.7288778911451357,0.844097355622846,
 0.3854266491144186,0.8707995471591183,0.3387030287255006,0.829946839806616,
 0.2643203711042993,0.6451520939178423,0.907170867905999,0.19669829845748366,
 0.27686123953979747,0.051969451426574054,0.5629907991106303,
 0.08646599263920918,0.23990401417072182,0.5393325433720407,
 0.7877746281359451,0.5807858033373063,0.6202892026806629,0.7666718585737683,
 0.4510439609419037,0.11437799765904999,0.09351604611056752,
 0.4537001015610218,0.8802950658311729,0.1856661937999935,0.595570081563295,
 0.4988476423208097,0.7226905452969858,0.7809221843464682,0.04531487169385606,
 0.048924073920956546,0.06462936756233917,0.19445481266897524,
 0.8707045337492699,0.4423905596371961,0.4859905351489765,0.44673299629908547,
 0.09074343896544501,0.10085288551974947,0.5151026990995936,
 0.9618688916105748,0.45925824287528605,0.315155522277779,
 0.016517467588338386,0.8435072725284201,0.8110704813401143,
 0.46399543919219666,0.8781021031886974,0.8024267480000931,
 0.32242765280301866,0.380365610539936,0.5152837944738998,0.8356161276713795,
 0.5030873121919461,0.004445833288345136,0.8946168018220758,
 0.028431901557558592,0.6744577601740155,0.3889780732049465,
 0.03849647445746718,0.8462552700720263,0.7034597898586029,0.7883649741812413,
 0.5252977300070646,0.21431124906207355,0.6312175807366683,0.5138327131279837,
 0.7003587054659433,0.44087953357496434,0.5269534432191375,0.9805004320850239,
 0.13167141009747796,0.760365071404242,0.94530285736073,0.620815090909906,
 0.6474438163735904,0.32680699059906027,0.17962815875280103,
 0.6975249399341658,0.3629433249000428,0.23430263064390855,
 0.02432336178994199,0.13490202372434812,0.4739827555187687,
 0.42865175721372706,0.054776164486914025,0.2577386325615719,
 0.4867252958319377,0.06401003617324319,0.738649070862555,0.6484944540999057,
 0.5149513217874228,0.6671405186952825,0.8907477341357607,0.6225758234217512,
 0.5309947058002294,0.8672254863664978,0.5583780635064236,0.2753365007510079,
 0.06854731796291191,0.9741031035062961,0.7535358405954905,0.980575810594507,
 0.43728082684020886,0.6602495843046284,0.10492281351174682,
 0.25165187213982565,0.7135019716315262,0.33308645575332285,
 0.4877689161534431,0.05856565305193717,0.024760630963190255,
 0.3811091855033486,0.05531819353178136,0.19495223076608137,
 0.2674642595463087,0.2374502565495782,0.03565911545287115,0.7244579137222844,
 0.09594228505332547,0.5786129618256157,0.34111493658856984,
 0.48502284772829085,0.29826731940834994,0.9835758987730673,
 0.21913054916727193,0.4438821118046036,0.12782137162442997,
 0.8683869585474943,0.35946637993806074,0.13294711750412158,
 0.8215561842511764,0.8379211357813419,0.5298314791897684,0.33882769619912934,
 0.8650670092264298,0.5105126721841445,0.6857003921076372,0.09830520383546792,
 0.7448591315535846,0.9539811077905351,0.5370103256787642,0.9855407114793875,
 0.4335665750935077,0.10047649824923222,0.027444813092171283,
 0.7669217076456344,0.257408742281808,0.7963383249363002,0.9094229599796403,
 0.40766646154499586,0.41421616242227033,0.7016639632842488,
 0.34848861395954267,0.15740700408679376,0.3172712712199003,
 0.8134272557254705,0.27105231366816496,0.08335327607057463,
 0.5173767015195537,0.3048932960637146,0.08525191796390097,0.8409381443925992,
 0.2771307276612439,0.29606637635240496,0.4895764974012936,0.7543172734576101,
 0.7979445745154845,0.4787413204012937,0.8361093136899533,0.1602139606685683,
 0.12706233130726863,0.002000775265524335,0.1617376802361089,
 0.8627999213261812,0.5917798548901212,0.24727834145060978,0.8372160626994872,
 0.5253827120007013,0.012801031502420068,0.42720897642048017,
 0.6770670134316296,0.1712028191006939,0.3550472710847009,0.5862727736645862,
 0.19133245328231008,0.8133363954012637,0.598570091307556,0.758438897396903,
 0.35763746326750523,0.6810634051458256,0.962160587045201,0.8757363772034491,
 0.9644795765603122,0.2798377501199373,0.795120912826309,0.3508817923800983,
 0.5531779998290121,0.2548969888109113,0.08829915022740875,0.9364734752203807,
 0.8608741117324129,0.5121508797213209,0.5411392287624703,0.3138814974944849,
 0.08354243615652213,0.22423499966856208,0.32697632546900346,
 0.3766394178944342,0.4438374062664602,0.8822251984159698,0.20371561491421375,
 0.775190314303389,0.29737586666394855,0.3742872906085204,0.09347691042773609,
 0.5957146519364163,0.29935857840475255,0.05629487308013459,
 0.5300117054559063,0.6898941775072885,0.20642529069948834,0.8067269799383419,
 0.8047544397682296,0.029649706904499418,0.9153848608702068,
 0.3353622342686404,0.8525109344105986,0.28206122287651414,
 0.44487074691141837,0.7096684941478489,0.22422161331077306,
 0.9846497680700724,0.3743240208395383,0.43786159858751283,0.9189490821944118,
 0.3172741827225529,0.011712807067498665,0.5475869617239568,
 0.27069292764628283,0.7121664903032768,0.8542450457363497,
 0.15733550400859841,0.5533875824498627,0.8212861281374492,0.8282233119590579,
 0.48587329562311077,0.22447350233330887,0.030576565780852993,
 0.3381293305177865,0.39071759151597085,0.04087253625302556,
 0.38515557734508377,0.42564405863475596,0.9807614832720757,
 0.2473181543906383,0.19291369872728104,0.8710023881811471,0.3639130101075515,
 0.07393510584620433,0.21993053049207523,0.9122921128090244,
 0.6417033618439415,0.292233601669175,0.29364825238952785,0.3693197142196408,
 0.9071597502076716,0.7459192533338088,0.90134785958666,0.4491693901526248,
 0.8753235238383207,0.22892365148492677,0.19234899970070185,
 0.4275565602427507,0.23260147818787846,0.2196890941854257,0.566548515012738,
 0.06786632184079111,0.33670846794172793,0.440563348694175,
 0.08227195867894632,0.3577845639512065,0.19757603467341345,
 0.7804906894833656,0.3146750778248114,0.6237351313281436,
 0.044409522727159145,0.9033665739646403,0.13707217519174253,
 0.7993660557848271,0.22260430156676714,0.07062372117014226,0.97143760751072,
 0.3257012877010995,0.07248212794165021,0.9496851182338962,0.4405413788046164,
 0.584483739974716,0.13090331242784115,0.35368019292083297,
 0.17894272163228575,0.76486294788068,0.7114553662262921,0.8649157748060379,
 0.98941607368742,0.41837917650185497,0.6995105735588079,0.9654476326881267,
 0.6184889080335672,0.70969016077874,0.684373006533848,0.8442636351191333,
 0.2980914307964023,0.6339362356431414,0.5071992950995934,0.5890483255166199,
 0.8427754532144599,0.4583369748667522,0.8323691787644785,0.8915545903764086,
 0.6132340776584542,0.7584045184138763,0.9813353707606818,0.7650101464389424,
 0.14007538214406512,0.42560703734617533,0.4982179481975535,
 0.1802633387149154,0.5590519757918391,0.0975945361447258,0.16864482310065543,
 0.7271484414703999,0.12546454487572,0.23287399999668812,0.7847642976926511,
 0.16831477828001074,0.21565558338633084,0.8561291892395905,
 0.7161200521511155,0.019779756919971025,0.0825636435936783,
 0.5260617866104375,0.35902796620697763,0.9408851584191719,0.5917613577017867,
 0.6509556282443162,0.03334051449019226,0.5885722208857577,0.7349999574314305,
 0.42408661185565466,0.44827447887052163,0.4870833289037446,
 0.1316788046546472,0.19765053203583693,0.36605487963332384,
 0.45297670805177115,0.675456319150757,0.8692747530999864,0.6490905969171651,
 0.2281063129201728,0.22509043320293276,0.3847735147658856,0.7108539402393229,
 0.761828485263405,0.7608681212876713,0.06990871523407072,0.7307543655019597,
 0.26794417909072166,0.04523316500280283,0.5757671038585693,
 0.9280738599143218,0.025367211393049316,0.7857757944368221,
 0.6482076846453988,0.5882741587822632,0.06288436243748985,0.7851859101709167,
 0.4207538654432267,0.26948111702012056,0.5211697523325106,0.65168131736113,
 0.34976111600844983,0.15300648623559487,0.45452041138681953,
 0.8435393174610861,0.8472698842216445,0.40332641644321643,0.6098719631909408,
 0.7799759439081647,0.2746040769802157,0.016269307169651803,
 0.4127785864329979,0.9519008015345146,0.7132943701288798,0.4211384684288928,
 0.7268567980417514,0.17892828197795818,0.1255234845581933,0.5788934967389592,
 0.32103667203776876,0.9123551701949394,0.6401982943413836,0.9575020645299386,
 0.39848385637930295,0.8887785329520909,0.7029581163513675,0.4122757225529443,
 0.42020351222510643,0.9597765428786014,0.4963154668904376,0.4111714133279897,
 0.1833666666521364,0.11705215133619784,0.5609478823211673,0.8874388669815702,
 0.7134489535290784,0.5164059322333281,0.6296348888895034,0.9271895484835769,
 0.2464226328426402,0.9666937014533645,0.5496601679560817,0.4002580461565767,
 0.6596997600592569,0.3089821358095567,0.5504015374926708,0.6010347023426478,
 0.4160465188658865,0.3362629888542126,0.008734414514811606,
 0.5840995803729048,0.848535769489103,0.8598803276025162,0.5842417204006954,
 0.09910206842583835,0.9888277970495503,0.5429077545402351,0.8127152117056132,
 0.6528699311288293,0.7589362251332035,0.9158722550073617,0.954888223424138,
 0.45411559270482893,0.2838525699744988,0.8859946082928561,0.9731799829457395,
 0.45125547883179584,0.9296914397987989,0.11981377304319052,
 0.7709398275222046,0.911924404201051,0.4697139090792434,0.5792649479767542,
 0.9521549702039478,0.33418142672368123,0.4088823151054858,0.3864987675229836,
 0.7261908095847769,0.8180755268195675,0.16368584651827123,0.8200264088730312,
 0.7014871327346236,0.8918233003155542,0.6926712180564015,
 0.010520479462665211,0.6319182578848709,0.7509090151975928,0.933202527711117,
 0.7043333628897129,0.7803585077787349,0.14493070836930777,0.1913783040946776,
 0.34318274601303533,0.34801304957202384,0.059451871828836045,
 0.4090529201784403,0.6562519896477058,0.42797904250000707,0.806532935777085,
 0.6606593633373372,0.6708272890068807,0.8242104442499162,0.7669928165006353,
 0.7352170407375987,0.13841909880257464,0.37637725166555525,
 0.09191901832146399,0.6517832934110024,0.23634950551871847,
 0.4680794647698646,0.240814177270817,0.2102301112812891,0.27087133170881317,
 0.5513787166223082,0.6311414461744786,0.612233832597005,0.7245565059201005,
 0.31586154449927917,0.18058808968993623,0.35149674384988705,
 0.5414306374741387,0.6255425273780116,0.5584018618001325,0.43285341158010526,
 0.17291970020389402,0.5889918449631826,0.03935978409216301,
 0.48696925594580276,0.11753308928389306,0.5461918773414625,
 0.5687180484978338,0.9713153452192596,0.9025439307913841,0.06424111259420273,
 0.31119443511609646,0.9059029880672086,0.7705396072996759,0.5037335399826243,
 0.720108566192228,0.7935249930379463,0.6245893513991715,0.21916448989882376,
 0.3725610985935892,0.23837466484270942,0.7581243711825352,
 0.41348740395629224,0.9732185994174052,0.46445041563933687,
 0.7175011009871014,0.6093124687585117,0.7975361751150125,0.9945055181811646,
 0.10679450056923323,0.07062278827529156,0.321499460048957,
 0.11834452367936943,0.346083562228893,0.5013693227101411,0.1771420428513445,
 0.32731295902987256,0.7477367623347032,0.6933092454075113,
 0.42184003546106785,0.2980737499685636,0.23519656630539265,
 0.19259645250023327,0.8120692185951696,0.8383667879927308,0.4814195061764306,
 0.3328279747459876,0.3441017953093318,0.3450840327663923,0.35333319721978307,
 0.34110231905668853,0.9851270863977635,0.3773785569607525,0.4559593219559327,
 0.8426867658995436,0.37826934364936227,0.025221847694956168,
 0.029534864221450463,0.3100367397241137,0.9897897240364509,
 0.5473668828831992,0.9365524996653674,0.09237932301968199,
 0.055955135185467464,0.09423525146900091,0.4219731845828053,
 0.8298744764923409,0.5777259628380157,0.8382911669812333,0.24718390373530497,
 0.43611303113592903,0.3758765986996282,0.11820297709060423,
 0.7792490428137535,0.029788809909071623,0.7889217842226908,
 0.010198175592063152,0.6814468898308502,0.1441450286083107,
 0.5516776741144604,0.4792765141842593,0.6766221926316056,0.8893755903015876,
 0.4446735071055571,0.9188169784437894,0.6979336043746865,0.8797049676952188,
 0.4964686490107746,0.7866686029527699,0.5907107439582674,0.950750862443011,
 0.0665884600374409,0.2822520867039533,0.8093655583965078,0.5298854882207701,
 0.5499715894898382,0.41203697988469046,0.0599407193301188,0.8383527437500203,
 0.9129111552828697,0.8948506073395278,0.02092276598302034,0.7825443313893465,
 0.27398283462474105,0.2557993847111272,0.49897470110209974,
 0.5242771044476331,0.44790461307143214,0.026712853357836686,
 0.4754679087174035,0.3287248142754492,0.06879180300380283,0.4875206336010639,
 0.7943474969652546,0.4821539960353537,0.8336087557300069,0.12485846428324876,
 0.38528079759070777,0.40125545361508597,0.9574983936064388,
 0.8650561632389421,0.47772804447907447,0.39261326730608415,0.859687516727511,
 0.24923070584527718,0.4095664836114379,0.8284343986277937,
 0.21762390166003331,0.9092877755908769,0.36718904214198766,
 0.2780573287836884,0.8468953730964353,0.4221203037559883,0.5677677693667522,
 0.9663736382379036,0.020389867520368865,0.5942683573943882,
 0.3857219933514554,0.6493110672216834,0.015255156302437678,
 0.6770929175957452,0.6022961582619111,0.5372036825481842,
 2.905140770634862e-4,0.2535511263481195,0.6936889481406943,
 0.09192497680258183,0.06745276389949328,0.5639050145927026,
 0.1344388553276582,0.8478717289648787,0.3609835463288483,0.6930311795655492,
 0.6896051498359721,0.5271828885870874,0.870813478531551,0.39137081835923615,
 0.9741826733021393,0.2829031824865398,0.7878180796338501,0.7800669508396192,
 0.9681775777294096,0.014572109882948547,0.734648885615746,0.1517320872777438,
 0.7866782297042065,0.9044856689424732,0.28104110049785813,0.6811789102314088,
 0.5864335533724125,0.880833185461521,0.594130205301904,0.885269048382758,
 0.8600118062833397,0.5445934222429962,0.8445346169587302,0.28873643761668977,
 0.27969155388463685,0.8608879358790806,0.9100677568857217,0.5943048124739858,
 0.8319281565564537,0.4803586844377974,0.6325698427208961,0.23512341474271048,
 0.3741783683299639,0.941834510249695,0.6435408110729693,0.6504913083635324,
 0.385223855179638,0.18787342075020674,0.8179341874179191,0.7351616984103294,
 0.6698400539932754,0.4403489981698475,0.574550071828133,0.31247334286773165,
 0.12816021770948005,0.9829115131026841,0.6599333187335779,0.9669251071150269,
 0.3874454028635016,0.16272857941936925,0.0319319706610216,0.3310457505347779,
 0.2517543836836117,0.6972122781469621,0.322257690753219,0.9487455132115608,
 0.993149364578783,0.7018533064922273,0.2915336151994766,0.414033042232946,
 0.17043386938093485,0.464647896805953,0.6726259887391599,0.2440462565665742,
 0.3973363498350171,0.8710192359470782,0.27598047904890777,0.6171470579718791,
 0.4938779862552821,0.5110824687949059,0.53744782231316,0.8638361423959071,
 0.7038485797892285,0.19501609493044114,0.7175617349325685,
 0.48986699180495485,0.3727544251470686,0.3383060346315543,0.6538279555371328,
 0.28275898016066625,0.2559920351418239,0.3155420777640523,0.4598257174737884,
 0.9854037243738495,0.9175585548669336,0.25546156666370334,
 0.40741618196294416,0.10454042431184674,0.39363411386251723,
 0.3932004549609539,0.9714051530483042,0.17293153898750768,
 0.37181821263901527,0.712164629526542,0.9873771699181207,0.4400714332359281,
 0.7598331613369799,0.943092909416918,0.7546790019937502,0.17505755521721833,
 0.2956770153504197,0.2343730619811959,0.2862767176938933,0.22752960735598604,
 0.46982405422514417,0.2835264787027081,0.566457393532738,
 0.012647008467733478,0.5244087065788057,0.18654831090816482,
 0.5773570452906902,0.22185613203327748,0.5237060317320041,0.0699264488561081,
 0.14347279491597598,0.4508208160467243,0.8446907666276884,0.1293342168701327,
 0.04764935432601536,0.6765440719665877,0.335752372065381,0.3007904845271039,
 0.654504456861546,0.3237625863679986,0.07139870438510765,0.8936402443215111,
 0.43006013801020226,0.4159588963555527,0.7552657341780623,0.8625987203877796,
 0.4271575925013654,0.014111840057072556,0.11371976316191634,
 0.1651478320227746,0.3847069313820406,0.8726623101371563,0.15201310448759364,
 0.888093837087582,0.7262881780514235,0.8362311650860349,0.48884235211547644,
 0.13225430537106397,0.7498718241263148,0.777217773508491,0.5215537023215235,
 0.04083945678174705,0.15434657520771822,0.7715050294153006,
 0.2921278763653339,0.4893834292081469,0.9827872751751278,0.6652158300034934,
 0.20864622026726576,0.4869780325226084,0.38267014994249116,
 0.7778315743659829,0.5000546482308872,0.8439915200384209,0.8004276371430671,
 0.1951102242915531,0.008861038378999653,0.03811824072456549,
 0.8567592243559978,0.565881382663108,0.9143435720284083,0.8493446212041807,
 0.6684392353752616,0.5059381278108098,0.3218947322670038,0.34870322919644203,
 0.9846706906285636,0.08178344754092548,0.1236569278900036,
 0.13299175599590407,0.6628266215442696,0.7124370533098241,0.6975299615816435,
 0.7664111987925494,0.08379104432290985,0.6651320716641651,
 0.39158387376166504,0.827944294615534,0.7819565401642838,0.8811264431867455,
 0.4094254498403316,0.12015240231790114,0.8037693855445577,0.1460107879795307,
 0.18848973009621872,0.7968206692799134,0.1678584702797744,0.3200487136001394,
 0.5024500377945418,0.9197428226410557,0.41822793684616344,
 0.05013976413729504,0.48181374963969237,0.033234824988141454,
 0.7309689560396824,0.6099910403987191,0.1861999924482277,0.2781051024704895,
 0.0024600086851964686,0.12365481881222029,0.8844246449238874,
 0.3167494799439521,0.7951430946784271,0.990431339512594,0.2623303829698047,
 0.6592558490647786,0.14083878990927634,0.7672400306551663,
 0.37010938467199705,0.6803616813601578,0.8017184906092415,0.72101006385909,
 0.0866436007614163,0.818659401304827,0.929238195377576,0.8075724721795717,
 0.8125428211305331,0.8091469855858201,0.010522043143199267,
 0.6166473627885671,0.8179079228118692,0.2450201536965726,0.2569645785041448,
 0.26401454800913726,0.6512851011893019,0.7674847318759765,0.8213291552902728,
 0.80731574547603,0.9367370898733399,0.945742398759221,0.8767748510177494,
 0.022839585658282036,0.022047298787918446,0.671867330145136,
 0.6228201364025419,0.5425663058961891,0.3931642349082376,0.5406749463217287,
 0.1908375215973227,0.8836483233728459,0.4499033513031351,0.29214034035099956,
 0.17626702554408347,0.8161051959722995,0.8566976914548392,
 0.38776779318725973,0.38932822024752434,0.3427807073429565,0.412569168855901,
 0.27923569847415775,0.3880753966354431,0.22826863680542642,
 0.7687180200564413,0.9953496048387214,0.38675619598576194,
 0.27406144940746036,0.1978267553659402,0.9778714590806624,
 0.06030068462585192,0.019862564517247217,0.9889500424839212,
 0.3297493367307358,0.9255165464801032,0.2461578166918641,0.39940211283649707,
 0.9760817355973208,0.07920869888872284,0.6392862582244832,0.7792023216501949,
 0.36829396624519684,0.550927943336777,0.8580814493714648,0.9438589260533377,
 0.6584518243133427,0.03341114773424292,0.7418925621761576,0.0969322998757387,
 0.4018371988394396,0.22476769693028276,0.31050338898611174,
 0.7024014079640772,0.09439974634483506,0.40289808664213034,
 0.40758358003859607,0.8225444498125708,0.4383123172379635,0.4578257738017236,
 0.6800250515406204,0.18094544215480213,0.5713305293897953,0.2812896964060534,
 0.7492745868686339,0.07673962302194715,0.9295055076382464,0.6014526007699679,
 0.4729427731960749,0.5793477779919564,0.689695217068787,0.25562517351010094,
 0.2923490783945729,0.6903328987735745,0.3433899276958341,0.6679078713065042,
 0.23288950803935426,0.2939064319095194,0.4778352203173275,
 0.32467543943544586,0.5258891390121109,0.6911079228938521,0.7008079573304387,
 0.30647245945329504,0.9831535717956138,0.25393613082266,0.3537613383940442,
 0.5431871769571618,0.15842740484048323,0.9464641782908103,0.3156676629344345,
 0.11354026368149417,0.3616525011373277,0.6788008891374351,0.508905635750008,
 0.4744965077692652,0.4821074917492444,0.2953917242538051,0.8060791065746156,
 0.27703953609074994,0.5547052356421268,0.5019667053197218,0.7287719444507774,
 0.9901491843777207,0.3703374057067914,0.3321381006298424,0.549508211804496,
 0.21633188895910127,0.9062427403264869,0.9908035215700964,
 0.24761686120796877,0.9569996048162791,0.27885837989438267,
 0.5137568903979707,0.48294388069808436,0.0658188600282551,
 0.41030980455884825,0.4347210153410658,0.004134819401553247,
 0.8620935976298323,0.5058160706769351,0.6584886993074456,0.29311802847411506,
 0.5607038892937627,0.9489669850765137,0.7072878487241268,0.3875343638778268,
 0.5493000832721496,0.10628387570579989,0.9632118244256537,0.6032226536938046,
 0.6183835496904495,0.007537387447825172,0.7999535765073363,
 0.18250965614084091,0.7380786339389145,0.1613341994287537,0.5745466077896499,
 0.16346854426323143,0.7571217114488848,0.9512122283450614,
 0.44738026331571823,0.18607324805675196,0.024493129949774484,
 0.7724554857059032,0.4207821290934294,0.7503091900520056,0.8172310643834952,
 0.29278156322764404,0.9420990649926816,0.7540510037342559,
 0.25299885210535467,0.47294424513517597,0.6708488797828781,
 0.6920204470298845,0.9659117881468248,0.8032603764787419,0.10294851456792886,
 0.5906951899286748,0.885452484613528,0.9030909723624089,0.28556992810567516,
 0.5874950202966085,0.42672758321825643,0.4124858796540548,0.2084330284704441,
 0.342500145580627,0.23023165579990978,0.36941820369452893,0.6256062535765923,
 0.7105285017093125,0.8100983530695609,0.46986701558909416,0.7216444215163993,
 0.827536903921552,0.43831568378468344,0.9854993128669582,0.5390765770121778,
 0.41579804989591906,0.24831275375710882,0.48729375675343756,
 0.29372046926421636,0.3897381398031917,0.7353531952534438,0.4032924973697575,
 0.7417934408408956,0.4978265835158665,0.4617569305989756,0.30384730364675516,
 0.8860594365917032,0.11423690102225526,0.09296991163864776,
 0.20539970762010495,0.14440240791907222,0.2915424398969294,
 0.6700562120160936,0.8742252273121478,0.35603337152320325,
 0.027492604628277073,0.6092573653454184,0.23808767207274317,
 0.24203156283849492,0.3272682802851181,0.948404395997668,0.681513607535814,
 0.5526256515677559,0.3226248939433466,0.9209649801225737,0.5447893973823366,
 0.023999188360456225,0.28624706802859035,0.12681010434369833,
 0.3747755891145699,0.8216482133858825,0.002380837119017487,
 0.7602649726416134,0.4645046085301704,0.6036484330659088,0.977379571160382,
 0.39192252625477664,0.5235119804424685,0.4844290205856998,0.5566948857552988,
 0.9315815077276349,0.438689462834567,0.9189377143850247,0.15427970380656408,
 0.5698743645758935,0.44600734734687075,0.7906430807471855,
 0.38357980433398053,0.6862182700181328,0.5171784535695086,
 0.49264047657490084,0.0925592942519059,0.10026562280456963,
 0.9090549277287858,0.30935934725258685,0.9345670523290819,0.9765768406681059,
 0.21902479275441544,0.1517986737692586,0.19757033202669994,
 0.7746846378036336,0.6044782957413817,0.5587326919439951,0.7138794998460577,
 0.9511979732960315,0.4014315129074665,0.473290418634417,0.29049554734944916,
 0.5584866538368529,0.46842752660973797,0.4644657247615134,0.9419285204059554,
 0.5916491774395813,0.9241700811170905,0.9832347025204203,0.865704569440245,
 0.8680627920689287,0.780260046147982,0.7540483993090858,0.015236922001935271,
 0.3003552244608002,0.016247767759392295,0.868804118825989,0.5325579698390062,
 0.48405879048747635,0.8729117231557324,0.39348498642365415,
 0.33204629813334463,0.2003957777808646,0.398334040527792,0.36659051172199375,
 0.48532882067105376,0.04571287772799537,0.39040764629577485,
 0.3998294299561468,0.2717301628033846,0.9263060096152756,0.8338695626482062,
 0.8510669337817214,0.7205265766411031,0.7636783925630855,0.5176492342596264,
 0.6983326729992505,0.30131648497320396,0.8552063528753526,0.4107829522305695,
 0.6677446309734796,0.8490931413086938,0.509532432342087,0.07835659860197242,
 0.44492179667157117,0.5756504101558578,0.05621912848137234,
 0.05529648557689215,0.49353797148302114,0.4245863946178813,
 0.7534120213743388,0.8998655321892978,0.3792804871009463,0.539027546650472,
 0.11882967954634283,0.6901293762218998,0.3448491389413848,
 0.39519717840438373,0.7905322656029399,0.8328091778217183,0.9678017990419152,
 0.6566126557203843,0.5764752917010394,0.7903606733885663,0.14865024180562203,
 0.4146786192069093,0.2664712383884952,0.7208114337803457,0.5441032553988225,
 0.742535194055021,0.5977993891795529,0.9749330515918033,0.42484251332246115,
 0.18302710591996507,0.7122962480465406,0.3091468359343208,0.8023811600617292,
 0.6685544313829909,0.6719244944444045,0.3653961023258522,0.6303227448933126,
 0.9869123627532249,0.8179844206552765,0.012936227970592684,
 0.45801557153381756,0.8688005647466064,0.14137690504442324,
 0.3123506118348516,0.41387295702539884,0.5030129622028816,
 0.08863806307850275,0.40261782416075653,0.4608825103749963,
 0.07627131315719582,0.878228048357852,0.8740121448801276,0.6684117535582587,
 0.9626320907570193,0.8271608793974419,0.05723114976446542,0.4701336139133163,
 0.5234173728580866,0.8707103989836312,0.45626811938281275,0.8254345476509289,
 0.4734949607364316,0.6830155274191024,0.41289840655907306,0.521912857133391,
 0.3408588710612608,0.6264627297260745,0.7900167168514147,0.4933573698740843,
 0.2973880329132035,0.014978655511687866,0.05999881239746341,
 0.36347154472001486,0.5564289856728761,0.19370383798642887,
 0.1274838716490666,0.459407944685402,0.3106766207128816,0.5672767053546771,
 0.9225909162220325,0.4145207245090732,0.6352751130337276,0.4600752600209773,
 0.4512053621909231,0.6334726138285172,0.7224364741232,0.28165024229832636,
 0.046804582179514,0.3040676249031461,0.9634582846452959,0.8810527986888869,
 0.2081929485158852,0.61876591876736,0.6970926741720033,0.16864683812877956,
 0.6717275337717983,0.0823963386369213,0.659966376799186,0.5022009794073762,
 0.6057915043216785,0.5164747343488038,0.6521214778530856,
 0.051462828572088615,0.09122815254493355,0.49846594475189276,
 0.14535051574269753,0.6746424503602423,0.6186187073705673,
 0.46863808538625806,0.9053275455010086,0.6126540367340552,0.5000520705833605,
 0.6248765705526313,0.08545557809061732,0.019719560180304296,
 0.8071802712764313,0.248439549534881,0.03413260389601247,0.4269321395717981,
 0.46586507154940326,0.015873451360834845,0.9414365731396278,
 0.6944248537441939,0.24153054093232063,0.28670365422037913,
 0.7203637102215902,0.20323189493483673,0.712421289715663,0.7100028613363049,
 0.6957397696921042,0.12247264820944237,0.5881662385134273,0.9341052629302103,
 0.8598119389859251,0.13843737706258985,0.8592565166514348,0.125999223968965,
 0.507926903101039,0.6316846133706108,0.6925315667237992,0.6585390492736289,
 0.9121243291643295,0.36444505373726654,0.3674516613248914,0.2159334468954418,
 0.8993170249788264,0.32035494876948833,0.8749342560485032,0.4613691802778792,
 0.25018392246099463,0.9505027305415448,0.8258214779955817,0.6055094036299404,
 0.9604272797389248,0.6492924595888445,0.856015656674864,0.8200831497432428,
 0.9157461028740572,0.6532547202214302,0.9905861428432406,0.8644379585432689,
 0.5330661770410979,0.800813496210702,0.10525625951680362,0.24429783943515193,
 0.3467046824033039,0.25506996612549726,0.6262573102292777,0.8596809024411127,
 0.4415431508054337,0.022945362632712074,0.6920346635864094,
 0.27669639798622214,0.5036335810134422,0.45238097872295535,
 0.5921495981682757,0.5017925185305336,0.8816577677359435,0.08185970004925647,
 0.5149129204636078,0.3834070643664138,0.19952082380816616,0.1791351733994866,
 0.0722377670683153,0.9507617568441666,0.5956522830528431,0.7239988752276475,
 0.5180937422907215,0.5939214114870538,0.5715381767286181,0.4210993460592293,
 0.39616747410045117,0.15821675050737716,0.38913982245107737,
 0.35081993259438327,0.18271990992202425,0.30836639519892095,
 0.31165208745664974,0.5631425850136363,0.9844482919699884,
 0.36373058093611355,0.5582615851183267,0.17042006645613916,0.382526890665399,
 0.3233094112506353,0.019289222595983202,0.15524895105132575,
 0.49085814799850036,0.060958479806108246,0.9459014166773237,
 0.4284340541099445,0.8173903476837713,0.7660826046551507,0.7661724822767172,
 0.00839665467924089,0.883160165165487,0.8411010398011896,0.29833420097083474,
 0.42501636086752415,0.6173644050210583,0.18411356972945125,
 0.7456854384765591,0.8965578055070732,0.8348713384113754,0.8177634015653851,
 0.08698438859631275,0.22021572878595697,0.6953254028264197,
 0.07626271517618444,0.40589102020938717,0.4141134703961904,0.783609846332943,
 0.16313190517697984,0.16745565675432783,0.588786941445594,0.9824011255053446,
 0.7188731855339464,0.3991893903139242,0.8401336590136588,0.842742017063079,
 0.5084160088033863,0.018116422642232566,0.32587222294312446,
 0.17371203643170863,0.33253225923257745,0.6760047360356732,
 0.5598790648660199,0.062177776454397926,0.28981740990908134,
 0.3868307076617665,0.5458877691003862,0.26757415720633615,0.1312962885475959,
 0.9913759600863881,0.39756523028837254,0.6242869919408757,0.8958860437666548,
 0.04970797316163389,0.8908768966656404,0.7918437807460545,
 0.18293492848866755,0.0926054811458148,0.9441683499049671,0.1404069029159113,
 0.59265925980291,0.7283892811123251,0.6720416819119878,0.441637279943509,
 0.3773755916659107,0.3610499577063486,0.6731316688039748,0.20820285700050178,
 0.48563439471614545,0.8494626936995766,0.14591825559686356,
 0.8929089764211213,0.6922443185367368,0.7404185525673566,0.31159867305261035,
 0.22622399229696322,0.41276873580937246,0.04685937466401113,
 0.7332862151200812,0.46238131487898726,0.7633804217022949,0.6898640891412957,
 0.2470021215208188,0.6833538283679722,0.17027118782885797,0.7498769803335815,
 0.509085818168379,0.5714002070728508,0.2206841200784282,0.04695877965698281,
 0.7719091092451033,0.5006293883763074,0.42248812922472356,0.2425647797461421,
 0.7816152075520684,0.12554527343892619,0.040554037826186384,
 0.9599724841948593,0.3297308082141239,0.47644187373164104,0.9450321472583388,
 0.8997690831930366,0.1273521440065709,0.6229268795326547,0.05147345905179057,
 0.6346455536418679,0.26618526827788624,0.47204460810106186,0.604634440489938,
 0.0846174758924697,0.6132249169432746,0.8774762103610821,0.09009065852403386,
 0.020082400214616047,0.5714777973372563,0.8970753142585649,
 0.9485113334895062,0.33094304559650245,0.013576149439264884,
 0.5985451854880176,0.7803307591574534,0.4055779375599209,0.9500729299240895,
 0.36193033698099797,0.011877688599147729,0.6408790156638671,
 0.3176970562538819,0.4396961688051929,0.9811682521848609,0.5802718722834094,
 0.4771454216791985,0.6412094290077636,0.06213942673161443,
 0.17160698595305635,0.38661719377860404,0.36569616832810725,
 0.4233262370810117,0.39224536365860885,0.5298774073667126,
 0.14634993360973736,0.9739747321615764,0.03623996441167143,
 0.32328957192593644,0.8756705270520985,0.8971761621356804,0.194394739846114,
 0.12052766637629941,0.38344047680499027,0.5088792023429687,
 0.0907310990268484,0.09973869204055408,0.6972554592085194,0.8557074732544885,
 0.1269148662070576,0.8484445009842532,0.3910994888715785,0.7057650615853804,
 0.4617488680110017,0.19684896080549497,0.2252099542567746,
 0.11066768562016815,0.664881176368318,0.902994210591792,0.49625871976947666,
 0.050651861348730454,0.465375491094968,0.8514877712646757,
 0.46274283028139906,0.7078398753610015,0.7923554620621759,0.754049887964211,
 0.40042082215475905,0.7933992765819049,0.9978535715208038,
 0.38314927138080435,0.8690790592666833,0.4211867336787575,
 0.036511670702811805,0.570410743689456,0.27980712313923894,
 0.6512707777275141,0.624691719005186,0.7960489566279091,0.11807644870489509,
 0.32200040102263117,0.10160181704575677,0.3390039355490646,
 0.8175295537675575,0.5603586785181736,0.9804218695903537,0.9264299686632946,
 0.780408699286885,0.7996769431779177,0.49020999454660286,0.3929800395234686,
 0.6879236143845675,0.4995946339189071,0.32795238944392413,
 0.16637798824681294,0.26998132404495134,0.6331107316744693,
 0.46888094795466817,0.3285838426271587,0.2033466183634205,0.5771763221287981,
 0.2817143840523071,0.5645829763319292,0.24512501302556855,
 0.10430280780693768,0.5943434399308436,0.27551823184938995,
 0.04597506681447305,0.7389170677803065,0.4431442474028824,0.34524825513264,
 0.287186751310611,0.10726643144909853,0.5646440072637817,0.32279398551234073,
 0.6326741744211434,0.9502124677705959,0.09286001705316038,0.7369421271415371,
 0.7787504100105911,0.8893335486694626,0.2878479138408033,0.7406921077372511,
 0.8490595939606086,0.4064802907776368,0.29503318226708697,
 0.49575114180642976,0.2540468838054257,0.07931338718352543,
 0.5157863849242935,0.059105194739239586,0.5094688808510658,
 0.5943371311753971,0.5204131605324668,0.5453801761976909,0.46752673642870424,
 0.7632343559549075,0.17503807917373526,0.6701863057639607,0.9918224208719509,
 0.9259346823015763,0.5504556053269415,0.9016046110234939,0.9326039649418569,
 0.01806931559039704,0.268406119106542,0.4573511847599764,
 0.018217336175191057,0.9396134812539998,0.7828069523972944,
 0.06386304205855486,0.8310097446944273,0.48594010641929597,
 0.4157539347777466,0.21175374651690337,0.16654425418927454,
 0.2341351132730134,0.5436096256900321,0.8173990450495724,
 0.053832166371535206,0.9702009469963737,0.6960606039475425,
 0.6355036840158649,0.027569667888533544,0.7848892879596336,
 0.14674378739466265,0.6624826930410873,0.3138324798038771,
 0.24121834508123552,0.8021671747364575,0.013358653242595997,
 0.658713178294857,0.585968050994263,0.4929336825583118,0.32372148394512257,
 0.10461603426255639,0.07401779953516763,0.5006543655411443,
 0.11410334865158656,0.5630916441153194,0.9334225749353794,0.7579872654684061,
 0.2356012757015189,0.2327589540788103,0.7901224711300766,0.4280876416880034,
 0.8298710576597554,0.9221678442134418,0.09584247495473941,0.9319906033041828,
 0.6103642029661444,0.4441169001431109,0.2309015068357334,0.7799589733549502,
 0.9818591473296052,0.8838751386593598,0.5350702306282658,0.15242951850706676,
 0.06581902434607034,0.3495733773528342,0.6217515399276761,0.8404578234013689,
 0.8577855389577864,0.6399479620314227,0.34041527970451246,
 0.24429292964281513,0.6338873656902084,0.9694175315651041,0.7754968679886938,
 0.1599966411417466,0.35440993679361754,0.993431176134922,0.17977004417265463,
 0.5986508441802353,0.6040535881104441,0.6953381310298512,0.3858017463083111,
 0.9190339283921627,0.18310745393882977,0.424813610396894,0.7112206777920098,
 0.46661042105678474,0.9559724931911564,0.6662137483227676,
 0.23875671803734222,0.5439199079524327,0.9052539731093147,
 0.15432866378269405,0.6811330915493392,0.4388284645112893,0.6231385784681032,
 0.8128198866995314,0.7193126113238228,0.40186526241541454,
 0.07730323064026479,0.5244560109380609,0.03078261308127317,
 0.5507591204780758,0.41439004603425633,0.9657385322137223,
 0.09109208017106296,0.5413963184447852,0.5914915124279047,0.7508270513198188,
 0.5859730219840297,0.03524258721301443,0.7531644450756477,0.5637667268090062,
 0.20944439845648655,0.11485192594541904,0.5699667693862084,
 0.14250122575120583,0.19729224469025042,0.011458014616406054,
 0.11249427653810729,0.7463999816964392,0.3435640500056111,
 0.052549197078667786,0.6493409381105373,0.5609288836055972,
 0.07900943869490473,0.015755642924074165,0.8761163960706331,
 0.6991240830549657,0.43820762896139165,0.4395458255889386,0.5705274997841776,
 0.05960670774263743,0.7039229490246215,0.8958268112829937,0.568073508356244,
 0.49973776169227313,0.8752437888097979,0.7926701615389792,0.4174771553677583,
 0.31558515173604973,0.6187135882722878,0.11635098591592619,
 0.2786409367393934,0.5625780490246304,0.5766618741791834,0.7071368530993037,
 0.5259727823449164,0.09698899149557749,0.4996834120426643,0.98267130309278,
 0.5697317239634124,0.2013310495948033,0.8136262699057902,0.42624590268253726,
 0.8446795206381684,0.745320225213494,0.37332872550660723,0.1364272809921483,
 0.7798735651139824,0.29652029374769784,0.8730906566476959,
 0.17357826433953316,0.6360620571009612,0.5318543254317096,0.4219522638924228,
 0.6984016090172898,0.3191762677661589,0.8596994746703835,0.3659973896493047,
 0.12463998298894274,0.3715838466810881,0.19583595165501744,
 0.02083433320966943,0.19599518986992315,0.9384493824658641,
 0.8722296620721143,0.8089651178197911,0.39005216091094363,
 0.29489347907923924,0.18835042549246483,0.6082333318250494,
 0.5421664667768558,0.20698832338913664,0.7332212735212429,0.8190945656240759,
 0.7715258584934889,0.08265705144947022,0.8348573226826034,0.6950342195729363,
 0.004550573370332711,0.6465316313029758,0.2800714118691846,
 0.42785051480073655,0.8077700862690224,0.5726119102157473,
 0.11866479855904122,0.5056944090085053,0.517811830284742,0.14868825803619734,
 0.2655074204555625,0.8039175416880883,0.8505959201785533,0.6048125276336425,
 0.9163936823099973,0.7817870951139744,0.8394581524461969,0.519470947719628,
 0.1480663808857634,0.19977107286129303,0.12557704247978707,
 0.16011606540203405,0.4268935106868751,0.6929451484465684,0.9256458784756927,
 0.9997606092347332,0.5958847978200769,0.6252033785805112,0.901939121804358,
 0.2284073859137118,0.5128323519894199,0.7950211257741145,0.8635459711928797,
 0.8742585494564435,0.8714168152387859,0.6573449281261917,0.24454204145049374,
 0.17393710626536008,0.8848011553498198,0.5346970349759455,0.6689664655935046,
 0.046397740438672286,0.5012304849600677,0.8190539787619541,
 0.9973576791717134,0.7371866816956492,0.17352454025151254,0.7809217620833381,
 0.7598535012421281,0.7757158235794096,0.7928748957979018,0.43551314590869084,
 0.5977305371386001,0.7102181769096356,0.6432298320576715,0.4792824973292762,
 0.6862055715883977,0.6464412719079198,0.13209998646198795,
 0.48041005651624735,0.11063824068723982,0.2562123480878643,
 0.5362966977612694,0.5847487251442197,0.5494475360371422,0.6540780019974715,
 0.7476850465896326,0.5172026511145843,0.46845056288476306,0.9521653405855931,
 0.8199654083819702,0.04160985035138276,0.6740389004889271,0.7056973657206757,
 0.13567665155983377,0.06125484708108497,0.7929061529264787,
 0.5430155039207041,0.2165273857009853,0.6071115735292525,0.33013714268153005,
 0.8565610711913535,0.9334934769385568,0.2085292543285947,0.3587471458052056,
 0.501224391805005,0.6482877849983404,0.575862607736521,0.832447618306089,
 0.7019920068868959,0.9737211791104309,0.704433798551265,0.35623274588710585,
 0.06858829890803841,0.371646927778855,0.7312337797736133,0.6680961808637521,
 0.9558582191360281,0.0953729790454908,0.20741203862394897,0.4859168605255467,
 0.47959719677741486,0.15837097009331236,0.8391428344580198,0.797985532136852,
 0.034750382410429714,0.6748572611235428,0.6581433518736863,
 0.45490699177163074,0.03742952372571129,0.39236222691501477,
 0.18261156635290654,0.4617329366630376,0.8492702634921393,
 0.17494545628731328,0.858955784147676,0.6172707369484978,0.17015075746447184,
 0.06044837184027618,0.3421161377080588,0.6032757504173673,
 0.04133347317510594,0.7394629161163978,0.6650017552130288,
 0.16031668388309195,0.8761519595886877,0.8406817914670038,0.6550023204742914,
 0.5097166123765814,0.7648285617696581,0.09476318223260427,0.7170829606780159,
 0.8607750314845148,0.144766170774838,0.3137630789181717,0.73059746832466,
 0.5267081041792647,0.09218240803829947,0.7005755922515393,
 0.19820792813492927,0.39698538912463455,0.13285492306754998,
 0.6272828193428241,0.6775273619234325,0.03503853629701914,0.602354996199333,
 0.30297827225246754,0.051469328207552456,0.3542446818587963,
 0.4832473610327357,0.436363063256094,0.16021083022004912,0.35719147274072394,
 0.11476639695035351,0.22732996364195046,0.17538544380436516,
 0.7668814997399707,0.8890936273579032,0.09588704417035554,
 0.18998647292179927,0.8432966373703079,0.4764791768020734,0.4322961930198177,
 0.2647611531945353,0.6010268242966738,0.892087918137864,0.6875713448731422,
 0.9113997866611165,0.6877779751672447,0.03444127556716314,0.8563371500548886,
 0.7066780775308849,0.40279705398919796,0.4223707426632568,0.9115411495577007,
 0.49181583573225485,0.14962553788271094,0.12136931034800191,
 0.44645414644829307,0.3440130534194705,0.41813352036960805,
 0.4081508113754243,0.5519107684112667,0.6732297802065886,0.9429859167469012,
 0.1367244195055466,0.3378160416258348,0.050613775976567466,
 0.7598994226154416,0.6878994662842921,0.284407041494807,0.11289361472479864,
 0.31829960820621195,0.549305648638464,0.6673149075188125,
 0.027981098219583722,0.9768678234702465,0.7458795750202923,
 0.14619846221762656,0.37902780664583124,0.9283547461104719,
 0.8853922331160411,0.7746318261051023,0.3170885178730527,0.3287541725060491,
 0.6536325613279231,0.02664596060196711,0.5432844865812807,0.2591398317129884,
 0.9690670113135471,0.47362450558604996,0.8868486033675279,0.5499448303613148,
 0.9487323597028207,0.5460979495299683,0.17581534501678453,
 0.17502911705208235,0.23006644844808943,0.5396175409234679,
 0.41092826554554207,0.2084289073011285,0.7662221966720922,0.5622663001750683,
 0.28500038206659606,0.5573913556757822,0.5527908824105104,0.3133427745522047,
 0.13293494111567294,0.7976158948984595,0.3961635337127608,
 0.10491037626430255,0.36928870742522024,0.9610886204983111,
 0.4234955994471291,0.6455395835479939,0.13594014430088996,0.7816907579365934,
 0.3205591960411144,0.18151614207039013,0.9653227368658884,
 0.45795568497797245,0.26709640197673135,0.13906780691386444,
 0.8257755608045563,0.9987905132298802,0.5233734648109785,0.6454095938442603,
 0.08461274120511864,0.2007715242372906,0.5921798630469406,0.90816802635249,
 0.7102419578892706,0.23054433182140088,0.005411035617411186,
 0.5081881174566691,0.8733963458166231,0.2338976476621255,0.8001937469192092,
 0.5700156508842544,0.47219152867397496,0.4437142905313036,0.3899119158738439,
 0.6528270336007576,0.28936701374839036,0.5608550757268393,0.5735662660400589,
 0.9373036327567008,0.6455597609374939,0.32106647155548784,
 0.28302610726104205,0.8208335842416236,0.7804742563726536,
 0.005323818391881785,0.7130388032542573,0.6404973503216902,
 0.9443690267145823,0.6428154877212157,0.4771639167787638,0.179604429601949,
 0.08773329678947461,0.6790674216689303,0.21004050198119395,
 0.9654855179457508,0.1040512315309241,0.07743264277158324,0.8622136124131254,
 0.912466948155074,0.30805857799736414,0.6545840711216154,0.5487675511297095,
 0.5457101492134245,0.6306921922318303,0.9221760255733871,0.26871680414905674,
 0.124294477042041,0.2488354431292763,0.396981662858511,0.6045408841677727,
 0.07146372527849787,0.4193043982476713,0.7249040671198985,
 0.41253204023034684,0.2677018766628986,0.17390370041715575,
 0.9647771219724782,0.9639194371505039,0.5412621694965991,0.19818053956779758,
 0.7638965986798828,0.7699480268033476,0.5280149479269637,0.08324728799493786,
 0.050044560259995485,0.44534692841557977,0.9164857088652628,
 0.1237862956274367,0.8592832979127539,0.12638583407006576,
 0.12212363858468278,0.43065382105562944,0.0565802360619333,
 0.4887551525346131,0.22357693255390232,0.9536761514593279,
 0.33366949901211396,0.1362167720153743,0.9518502161326738,0.6866943327023363,
 0.5656837072431197,0.7322074749881605,0.04977830913997339,0.510943982896322,
 0.2098635640935328,0.941827804974266,0.23620233016100578,0.8521190869209723,
 0.0627935303502537,0.8901769645238923,0.8578571082664018,0.40857852119279947,
 0.8298049603703929,0.7856691551576348,0.4337115760375796,0.2396445168831386,
 0.4936374189794066,0.9316399929260106,0.5326782673915603,0.3907556336298411,
 0.6460029350569394,0.6615610143570647,0.3215453838840454,0.4336158034787545,
 0.42166644667182673,0.9486925237225603,0.1210259709929975,
 0.43213364874168814,0.07535293931928466,0.17310713907064035,
 0.3959105057928799,0.17983113515395954,0.6468707204875834,
 0.008233252959444326,0.5982313356921517,0.5210522448914858,
 0.8925503857236634,0.6068880232707555,0.3573752096850372,0.3608032169327633,
 0.9582650364719097,0.509504626870887,0.560531086363253,0.22904403616964897,
 0.12097457896095976,0.5278105184230412,0.3084811026335861,0.531063123867478,
 0.5301746729214813,0.09813938588729065,0.9170013114932208,
 0.20057278935750988,0.3693102037455154,0.020233143108178253,
 0.6583894928382059,0.9960737031098479,0.8714823966990404,0.5237208640102055,
 0.40788213376733307,0.18707416472408955,0.248607779545721,0.2959845070559961,
 0.9982376378371,0.5481470572164846,0.36626866389862067,0.2942959066131744,
 0.8777651728332897,0.17168685467946954,0.03719645343529576,
 0.42324054560369573,0.9277954776876612,0.21871490286811046,
 0.4433786949342492,0.06712418450123858,0.6683700660647964,
 0.43353846601802104,0.8465157827547303,0.48946938693858233,
 0.9843164132790618,0.3359918931132517,0.41351051361223357,
 0.24412133913912237,0.37759248053925787,0.13195149130950312,
 0.6349377598751313,0.8903334211182528,0.9302778491591189,0.9021026260949649,
 0.7838989650781274,0.999001165956641,0.04228073034751145,0.8999632667569287,
 0.37728119581538944,0.2498133483729562,0.11019287401145372,
 0.00649083577856957,0.7940427555267736,0.4689953038867015,0.454193280292983,
 0.6085349717254703,0.8125713296608903,0.29790054947400924,0.9495063095621564,
 0.11989721728624247,0.12776469585811157,0.8204602961422186,0.156088105594687,
 0.723468478366859,0.4994695319889487,0.09901530227063948,
 0.010241867916808256,0.40007853500097657,0.2792008736417968,
 0.013496049402511923,0.8340878574802777,0.2228630505059993,0.145286965599552,
 0.9189491325699588,0.7806090648857693,0.6578302989705849,0.10482802906578281,
 0.5616995133447932,0.877577235358204,0.149889056864568,0.9958348338771243,
 0.3044655322350944,0.11876324666492222,0.7156906594370639,0.490617290266794,
 0.4223931190261213,0.8042683708462613,0.17096994264632848,0.6398609917690535,
 0.15262629007233064,0.015628604625800402,0.9641336468342869,
 0.30176761469633284,0.47757445844398416,0.04718264382992077,
 0.4416561669647855,0.8362447105745375,0.7337216622284278,0.02365102179023504,
 0.2738138096454201,0.5509340955777219,0.5237811023149108,0.8965586663008271,
 0.4373360788213415,0.7151492955589546,0.1458620799543664,0.9193436698220525,
 0.18063725543755305,0.5196683320237805,0.7045365485564768,
 0.09860888118110933,0.6790156335785744,0.13877554634595113,
 0.2132136166614681,0.8954787235000856,0.11791775342464705,0.9264126038666802,
 0.07172543957453936,0.7625733024440019,0.8022305909979879,0.6859432876054792,
 0.7096493668361229,0.5793587511817759,0.8528626032231271,0.2788250032017946,
 0.6540943040212276,0.6335417491490307,0.26484360228955584,
 0.18448686021865934,0.6491646495544977,0.5295524069126532,0.2736105471669713,
 0.45204879303756995,0.0369453622262742,0.9123352465048256,
 0.15107470360151976,0.47132763922229237,0.2185142141355051,
 0.5547350091191587,0.4412828383189331,0.8252136584639397,0.5352152366159235,
 0.9093526702727037,0.29848488189077593,0.7673723353532341,0.9584949076114357,
 0.1986064515998498,0.16640314204975348,0.5968704356806628,
 0.49411023546658717,0.1952886548766507,0.9918210370886926,0.6361523007872993,
 0.9125080911323126,0.4220694391980355,0.48693099902557635,
 0.39391665480123716,7.56348548868746e-4,0.3554992170102613,
 0.8823890263702947,0.7974686804151925,0.4907207179729173,
 0.057213982665511365,0.557545466137177,0.02058789065022215,
 0.3437909453329131,0.5970068322720895,0.8452679226896571,0.6285823856487206,
 0.26154435125145004,0.9551621627407603,0.5619210573113655,0.8239822107475353,
 0.3714628461938695,0.0682370528592513,0.5847862828052987,0.08595398997328063,
 0.45217573226794916,0.6789831515190273,0.4393632183299735,
 0.21450896459217428,0.47676231115277556,0.8919060310275628,
 0.0893058288961166,0.4167674252022424,0.6206209141987217,0.5816903248241142,
 0.5741898174377362,0.33917345844020996,0.06120366607497152,
 0.44979845506149896,0.071424250921018,0.7208084707450845,0.06423395896417206,
 0.8574317372163106,0.53853126276097,0.6199406040643861,0.9560746450673181,
 0.1720269853208365,0.689008060664156,0.6767911713235604,0.9279431303342115,
 0.5183640604469137,0.5219929695090583,0.9826645261414783,0.5585592416308586,
 0.1626580326444761,0.8968357518332178,0.25321448001840285,0.6738227975039117,
 0.8243358143285955,0.04366217926965277,0.0453381322180354,
 0.25007761357310787,0.10671998130989568,0.995601975646343,0.7814045981971585,
 0.48854989318307007,0.21934941731803304,0.09938326536300612,
 0.6761410723265058,0.5124877281813729,0.06237870034824455,0.05569192173519,
 0.5741142875911028,0.07447762905400579,0.7140891954867562,0.6524220217274096,
 0.33598671363715615,0.310994720699438,0.3739705325255366,0.0561576995670614,
 0.02056393965887393,0.6638551930951468,0.8257743099805092,0.8525575311927049,
 0.09926078017680462,0.8105973933524878,0.4893520013429401,0.5363061284625232,
 0.39798125406871376,0.8054433280195807,0.3273294075376041,
 0.05118894991592904,0.05717134330827822,0.847119425510317,0.8272588798959521,
 0.8110102534571608,0.9814036927268329,0.5818235523924361,0.8075283809417937,
 0.5984534506550495,0.25611792934000954,0.3669948989847591,0.3366983616550602,
 0.6764791883568908,0.955943079929801,0.7566364308204242,0.08113711214482111,
 0.1963539675371888,0.9314833321654025,0.12001872459097662,0.5841274116473729,
 0.25436788625152307,0.4996244022128241,0.4114322075038733,
 0.36915774246020994,0.7328906002614942,0.5489797426993447,
 0.47083182764449616,0.8962077916296232,0.3126317846579695,0.4671555718540237,
 0.547559259995284,0.9553425802550699,0.6829346357664748,0.5645284386347584,
 0.43479790706739396,0.4408460092298867,0.8402973299215583,
 0.05000735197813211,0.9737264866189591,0.5238296402062846,0.266911218891688,
 0.545784710054734,0.44194836935173776,0.3855443982278107,0.8237531516105392,
 0.7901678653649612,0.7553589243369008,0.9688127111407496,0.32470173268412506,
 0.1211553242323471,0.9715212774889763,0.3178332395832463,0.06328850782592166,
 0.8975284891697498,0.9906622674525848,0.48074049593897017,0.4172868814762487,
 0.8030737604812184,0.6039642225332399,0.058395989032497986,
 0.2259731614969558,0.34722662827976036,0.4978284310627854,0.3974905821059326,
 0.3815132741773015,0.5817007348856947,0.79383197174519,0.25337367346778217,
 0.6334210868327735,0.4726671984119222,0.07299899012691302,0.8761501994336163,
 0.019005941212275834,0.4192499512876605,0.44007497635428505,
 0.23837879211994195,0.2869741994582874,0.30826290894607,0.8098486107643817,
 0.9153329053229844,0.3764512230480337,0.712514103816449,0.5439521565007889,
 0.5331203381958837,0.4625573770442585,0.03519549030350033,
 0.32051561036739007,0.7158350626163276,0.04872229324905675,
 0.9610970703245089,0.5985503435230989,0.3879334891677184,0.8210952890071025,
 0.5188981194494531,0.22001568667768479,0.06065271922309379,
 0.3191087790232294,0.17243607448777099,0.7254840611899442,0.8027059121688216,
 0.540586888650066,0.9434375835646595,0.8368908549040827,0.03894842032781187,
 0.8868832477643072,0.937225103383359,0.655249832430215,0.266315163726874,
 0.5021731473747458,0.8653620773484427,0.8425421106136808,0.09919985345144255,
 0.2500305747752911,0.3238863880181446,0.7055644314172395,0.5847907923288382,
 0.902272838944347,0.3226522275234526,0.13579188317579338,0.7864572239891199,
 0.4758322682298898,0.990853282972845,0.14364995681388415,0.1891403991036692,
 0.3112427574996851,0.7405481519203625,0.8858024423474335,0.4551088621970052,
 0.6613232575815384,0.9346060150849093,0.01991242475686872,
 0.11124617172318096,0.4982821761307483,0.3165169277690768,0.2544751210859125,
 0.5262488019192075,0.7869301525936969,0.15368565917260346,
 0.30774772495598646,0.3321414641000284,0.10119393843888924,
 0.37649850523946315,0.38288257617181565,0.7932936770780851,0.297963396905057,
 0.7490220518937407,0.728638209761027,0.14723902184274273,0.8484093899107088,
 0.851211702342877,0.8127507146469272,0.7715130955048706,0.44703824050915,
 0.309691049121939,0.5216405606250843,0.507548129850503,0.6188212316490944,
 0.38264729067195935,0.8043113702258629,0.7707077722538834,
 0.43654142365957616,0.7718305441431279,0.9490853058643598,
 0.39228306021365433,0.4061484461663889,0.7580734180967521,0.2419552995098626,
 0.22799036433441722,0.8689117649293531,0.5960523624333135,
 0.07850404564543711,0.7979549926707914,0.3252331255888551,0.9384258500408674,
 0.816368181801045,0.9318006818105227,0.2046204566271902,0.23908809402897446,
 0.03691159025206614,0.6635408333527479,0.1441516336154801,0.6855938199097971,
 0.14772856361830833,0.5964821375861784,0.06471667889576238,
 0.6808634513218823,0.17080360304967357,0.03724196814729497,
 0.9278297996480813,0.7593801975498845,0.7102479030053994,0.9296495547213366,
 0.53602839557271,0.26569519541200937,0.19638072114469418,0.10284259952195163,
 0.9763147120586736,0.0815757353283294,0.9541156718197324,0.8164932817564088,
 0.8258529948499594,0.342295989454126,0.8475453580781374,0.9708515949052628,
 0.7066383248179919,0.5852475172978123,0.5312314316544027,0.52381909636355,
 0.2998716965022148,0.8831393316453883,0.2431975664919741,0.5897224152385533,
 0.8077350448778966,0.47914676596679295,0.8938835724015181,0.7826056004264612,
 0.9520226693686181,0.5923621940015166,0.6352406420888674,0.07000324580940032,
 0.4848623824103999,0.5532722259396319,0.6478918265536698,0.01503315397306082,
 0.5871711733020357,0.3388628800227579,0.4037324462938755,0.0610467816898802,
 0.3399111846375782,0.6183209551431011,0.9301057429669072,0.8335774000468282,
 0.9635732576628225,0.7590100019324306,0.6246193253874425,0.6178877563153901,
 0.2467282498734671,0.32247609995130866,0.45015781372130137,
 0.2852279436922781,0.02307185420568081,0.4810239160588361,0.9275591887512635,
 0.37356406294570943,0.6061945057423959,0.22311064483084586,
 0.8759293819941067,0.8314968526531505,0.3842971084637685,0.205298602203682,
 0.1494684233078708,0.06565157299778357,0.2510320397583631,0.7989263254239959,
 0.6975297086369614,0.7686954783880398,0.05535905112428385,
 0.09557315713822823,0.5373382215765706,0.5494649022870242,0.8289709159136772,
 0.37595055035414016,0.724657010592038,0.724488455334564,0.21104656503606534,
 0.7934500442006615,0.48443072570454737,0.6492960845033131,
 0.28309294868780066,0.8607251149775544,0.9891529963667598,0.5020410527923344,
 0.2328582795472759,0.8875094855326412,0.6834760992907717,0.9268763005417298,
 0.6960082792050658,0.14541195446098132,0.5691058764531258,
 0.43567482552096237,0.9541030756601989,0.8447783167260942,0.2143595417237525,
 0.21484470272872946,0.8138330454927264,0.36910365652840094,
 0.1674168266498216,0.24028354332642266,0.12854370776769164,0.815601009688371,
 0.6694776057170182,0.5245792174779202,0.4088589632087055,0.7537867289475821,
 0.4844369353837392,0.275655852565054,0.021521076954795215,
 0.035828555740544554,0.0599351822714258,0.9936016949539007,
 0.23548042592756924,0.9030750491647128,0.34355663239223977,
 0.3611042082841507,0.42689436385772095,0.9256606294936316,0.8024363576241064,
 0.9392742913924509,0.9853641853459035,0.4109087827404575,0.31544931703689894,
 0.9565484579988688,0.044433944384123625,0.7312389741911774,
 0.39484991969536887,0.04922943933955559,0.025310128584402736,
 0.28857058032355154,0.29123812001842486,0.9476106312992446,
 0.6341268714343352,0.42458148355092873,0.20883747203256342,
 0.7106207413939295,0.8891041183897983,0.6491094432685414,0.7500984667324604,
 0.9433449051742752,0.09458597641766842,0.19781337549692746,
 0.10523883013190138,0.18497296347292547,0.37090401391133565,
 0.07904896513720605,0.5370248020597492,0.05881997298333219,
 0.9652677252995072,0.3645399597211443,0.8787301816892409,0.9378546009180667,
 0.7089769346367849,0.01334680969706703,0.15500299917119897,
 0.1988051919100926,0.3637220373598423,0.11591519418985885,0.6998526848672261,
 0.9787390144525107,0.34596450758892694,0.12285892205653426,
 0.021972741899940473,0.556588498797294,0.08711696254715018,0.940783964611275,
 0.8893809985819584,0.12151324351282256,0.6027817556114365,0.8685381539635805,
 0.8595940470941329,0.09958786227829719,0.39945793117061756,
 0.3331428710061356,0.8231598434536435,0.7133262788852668,0.8495550339755819,
 0.6783943413854058,0.871317387392776,0.34812920444346185,0.5936101264836523,
 0.37029061860211276,0.46160134961613963,0.22787990042909612,
 0.46758822767295793,0.745490436094642,0.8775697890547506,0.8525993259848175,
 0.25580430699123025,0.5728303366273245,0.2513841703532961,0.7856858850305072,
 0.3546457259030985,0.011840703465585523,0.15627790713807255,
 0.3952442864456941,0.8616469466067054,0.03777905736851561,
 0.07921450934580165,0.7705642751915077,0.569727380181706,0.07395671456588804,
 0.06784531600709265,0.6021207576051324,0.9757785507237412,0.994472498603507,
 0.8407594478827983,0.26704863610273355,0.30194245368550476,
 0.8002639379192342,0.7296510846349318,0.9184943954313037,0.7425461217813608,
 0.5171057007625965,0.6844475148252628,0.6551974906273778,0.9663941442798714,
 0.7766394704556119,0.22841037526227814,0.32736108989426727,
 0.6850015500606335,0.7763285516080153,0.9164633560278201,0.9462506273216729,
 0.4445409695119167,0.1379825231549603,0.2553924745963564,0.3781195396157382,
 0.8820337204760949,0.38349836646764357,0.7695076369542454,0.4879723533903233,
 0.08748838701569073,0.6194489161465008,0.2745247466457178,0.07235063374036,
 0.6669362142009762,0.8474406641326206,0.08978885436292494,0.7950301417720389,
 0.5389040868817456,0.6350418939066833,0.8759234026490218,0.6189956326459531,
 0.6050943064645598,0.47923132264043133,0.7152941145727935,
 0.43333537592489735,0.9841763247520211,0.07846994317756817,
 0.38563355671743516,0.37455780208877565,0.42693473956850114,
 0.7914189555770325,0.9732219798644713,0.21919685418988577,
 0.009767723355612934,0.8052678847735658,0.7563187279531699,
 0.022856964912840394,0.19004885877706135,0.4114906764330558,
 0.45311987003205567,0.43958882203949123,0.8947049285054716,
 0.19610402590564358,0.1683333867473965,0.6573172214662109,0.5349253452675115,
 0.17207964610134496,0.2907391143851561,0.8359516856547774,
 0.07940244060022172,0.015447634774817032,0.1583282909541197,
 0.562025809807177,0.2493948290891952,0.7361929697008005,0.48802726333127877,
 0.35273848988982015,0.6281951123562833,0.7237095361457365,0.6771787596078415,
 0.14787572903132873,0.8377668037198802,0.6218231423196139,
 0.03924994003335769,0.5784938464567362,0.26603321009026903,
 0.14092891414230735,0.07849861056866003,0.7228510383777431,
 0.3024352583727705,0.04065513618644789,0.9621167148516431,
 0.16054047310907338,0.7514288934520973,0.22201653092085427,
 0.3000426566127784,0.2375405480150804,0.6264940395387074,0.6434273922031888,
 0.06938297373542968,0.7521272531496633,0.46493141199542887,
 0.04066785653517824,0.7341766279955811,0.49058242618385994,
 0.42153938944701763,0.3440689369080343,0.8100877535184,0.9085571354114798,
 0.3254417558439592,0.9144330063587028,0.3687600065769563,0.528834900030115,
 0.9432228695949428,0.8917411760789804,0.956112603928556,0.47950859399350776,
 0.624888381465053,0.7687172538411857,0.9640319214940496,0.9175288326713495,
 0.14259111635009303,0.23991949858224504,0.3929944133484925,
 0.5874857708101567,0.7476128412221354,0.23213190325593291,
 0.017664491606097288,0.2772712859351075,0.30636744483395595,
 0.40531265922613025,7.845042710589167e-4,0.5150587591948973,
 0.9465457370861684,0.7761033796660474,0.6076694566670542,0.0837776366844677,
 0.1984080139002301,0.8156484182380985,0.21665362214899864,0.5465540534289762,
 0.4163402328732586,0.29607845517419007,0.5704469298561601,
 0.29929409984300936,0.8019412546375553,0.5166147761615975,
 0.38777108237764957,0.3280597798667769,0.8562000822885237,
 0.013471944639070754,0.5737974383911235,0.08583959081694525,
 0.6322544965771312,0.9266129831306916,0.3227304396837043,0.4604119503036327,
 0.9040082397998993,0.39368331426126013,0.8859608947791079,0.7595792122808239,
 0.4432556281666007,0.7208946935659076,0.4036763673216859,0.6819785267888608,
 0.7276619035764523,0.24979167286210113,0.9898337498833428,0.536479133518226,
 0.1991814769833007,0.27056932772085396,0.2758632232140932,
 0.21907909498971834,0.508899499735563,0.654241505957111,0.7099352733116366,
 0.7474452564480895,0.1903745640361374,0.6297588641574132,0.8917348564002865,
 0.3162529428131564,0.5537569968461931,0.8724404630027052,0.79352491660743,
 0.7115150797960128,0.6632564731927885,0.9038326017871575,0.9579918035557434,
 0.18360974307416966,0.2715439805430979,0.4450595915813903,0.1257793499500064,
 0.25489417673837445,0.8929300062537341,0.5725748367934563,0.8905267180680068,
 0.8503294435905189,0.1396070297284292,0.24849832949113349,0.4091357090758134,
 0.28956432717816594,0.6158819487066953,0.5606013212518758,
 0.39908270875844665,0.869774102983418,0.8995042726890281,0.6288065704567631,
 0.2468292406210304,0.4177264150264528,0.6308004391133596,
 0.019864585368494225,0.8484334751826226,0.7675190821703547,
 0.3484426585181566,0.09310914475298382,0.10348399054526713,
 0.7446743652111358,0.24043689362332898,0.8654764342457958,
 0.21202034265970604,0.22150353895360264,0.030987389019420997,
 0.0792057445867107,0.4277715185948271,0.7796642424576555,0.34020874108382315,
 0.5938731490224521,0.897115371458221,0.389579590578499,0.8512669566101088,
 0.02427416420981443,0.22721981295969196,0.7022165066568994,
 0.24422776268948676,0.41577612196696523,0.4221186610154791,
 0.8720227770158909,0.3649663111045418,0.721587232390674,0.1095953987280126,
 0.6761019501931045,0.6219364702998077,0.7794693252845746,0.29244891163159825,
 0.9557685233659116,0.6347838596050972,0.6786403875269977,0.46005275638017273,
 0.8373340128962304,0.36182827258622663,0.07858444978839241,
 0.5569692855624782,0.4029656077590229,0.47816508523645374,
 0.15600354506015557,0.08971562797805055,0.1754699656841906,
 0.2626984468241894,0.663938847543717,0.09393145768634892,0.9773847336614004,
 0.7994293049493425,0.8242051373288471,0.4423905607484979,
 0.0031768889672666445,0.9643114034612491,0.8593695864391762,
 0.06884930767943442,0.028640176407454532,0.5833354699862701,
 0.7045769616699094,0.8474617716260776,0.4495968064618232,0.7450452940830866,
 0.29540208434748527,0.5894071599474202,0.9563156373504346,
 0.11129646550732042,0.4558698788147595,0.1567023870017853,
 0.37828124599855084,0.6532633196277173,0.5366812950445671,0.8301242424109506,
 0.5433041927102997,0.8091251064297187,0.9071013714282457,0.7635119031225982,
 0.728590534081658,0.6523065459926085,8.124931766507748e-4,0.791145203892073,
 0.7573463412053303,0.20298321498817562,0.4107039039790096,0.8606657044187629,
 0.8672703056179264,0.5944670933547185,0.8569701633989696,0.7724978330814617,
 0.8226711743458748,0.4967271681589205,0.8009452520295333,0.7177819911541762,
 0.4682597160612383,0.48685137666909994,0.9014020329232144,0.4172037728958363,
 0.26157051461169756,0.17922581658414327,0.4318926112073882,
 0.9832540508972475,0.42981725150000305,0.9789146236223432,
 0.46022001933156587,0.7877312238480515,0.5785531123922124,0.7115954721793236,
 0.7141839306961885,0.8541588829976579,0.4290773042948821,0.1795534253492974,
 0.5251362067576658,0.24500112783435213,0.41134725774295955,
 0.9275769961073237,0.997865339479183,0.655603201123468,0.056916084013366786,
 0.25399607397412627,0.08616453816540803,0.6393103583322042,
 0.6704644893792816,0.8112886682471593,0.5479400983077476,0.5135158940211606,
 0.6957759690836027,0.928671958417681,0.3620726218142205,0.3055670089002098,
 0.028006093840036272,0.03712594756851095,0.40881723008658855,
 0.19200499989528508,0.1353250884674544,0.7348438379881281,0.9343429553054183,
 0.9389419329541939,0.22714187922595042,0.7373547650994118,0.4066819361329349,
 0.8289868205294099,0.450604351740136,0.9356176263737869,0.30437502121481796,
 0.36790543078954474,0.2989742262467875,0.2872343222373066,0.5052523455102127,
 0.48390608019528836,0.06100110557397298,0.17678173359446325,
 0.4836318486682756,0.16263639077604997,0.46805285130931973,0.846225682340765,
 0.5826638405515108,0.09452711103255829,0.21786584371169504,0.569683729472512,
 0.354892724695679,0.019642839328107398,0.8252294285702922,
 0.44082042096927154,0.9232356667649476,0.9844646312758107,
 0.21805920805481926,0.3257810404932395,0.35969061568502714,
 0.7975683111091513,0.7630663508383244,0.11304118101997851,
 0.12384114521472622,0.8210265570906703,0.07702047063951634,
 0.5843796749913208,0.6918405212188917,0.013783221113153998,
 0.3592408691716047,0.7421602891868533,0.27939271792495846,
 0.05937109975626276,0.6764509680767863,0.5758500543698819,0.7603096701839693,
 0.9342835820954586,0.037918499748721635,0.8819259172051949,
 0.10857285270158779,0.024364019297179906,0.1156533181370546,
 0.3250830507657958,0.3774119041085493,0.402557019424965,0.6220801072339235,
 0.29355744496544156,0.37409601046851204,0.6727734536240015,
 0.18434637331000236,0.14095610907920308,0.9643442054764078,
 0.42979426834860845,0.4615718603303378,0.5508268930380179,
 0.39903863832747166,0.06638290100567235,0.933770582127174,
 0.09735196956520331,0.8574086540197217,0.24927782823506706,
 0.8991234824658992,0.2561931861798118,0.3666854540918686,0.22655469613153623,
 0.061754524024421054,0.12366167042429554,0.36718991477097707,
 0.2593771095524774,0.6315832512526044,0.22947302702417416,
 0.07091661209974298,0.05516855119815545,0.10424562931654546,
 0.7789294145473818,0.4957842169973131,0.078884756847588,0.4702369807166631,
 0.504607623923584,0.4473939632450007,0.5937639968149813,0.5319833526751035,
 0.13781455779342755,0.08978404576164001,0.6389155836408432,
 0.7578338348079905,0.39171654234583164,0.7725750718456647,0.6619445439932233,
 0.6998801894590388,0.08333514693358823,0.20115410894093455,
 0.9612830513984707,0.26212058299054486,0.2295211800357646,0.5603180880487244,
 0.04890205327931685,0.043370135946712796,0.6291886276417707,
 0.6990743743446213,0.2356413939134886,0.6823136718872932,0.8301996847395958,
 0.4843801186575942,0.7590346791613405,0.43700002152057305,0.9783228400941008,
 0.3562900706817741,0.446833438070078,0.451711344454909,0.8873484967363784,
 0.6800866686430559,0.5707419902226282,0.6698065599542016,0.3958120287753074,
 0.28845047093108733,0.5456096101452005,0.9197172005533298,
 0.32818687451664064,0.7093012296323533,0.4677569232148894,0.7330521541141298,
 0.7067140367558098,0.6299171073455929,0.22045550863037422,0.7975902623313862,
 0.26820628781778955,0.5687525820535518,0.14631952919316527,
 0.6675286881218094,0.37453290765140335,0.07469406845671633,
 0.7928672127540698,0.8694661485952363,0.2917447888416824,0.8401815627406255,
 0.6385532499963191,0.044143517261979626,0.7088425816630692,
 0.03280092069585572,0.15365395128526793,0.41202973809305987,
 0.9625530320471329,0.9128022148171383,0.895400539316595,0.08563377289463514,
 0.9887607420179222,0.047956403078466714,0.18536881580689313,
 0.6346066328624135,0.4067749324204005,0.23707016522603375,0.7064805161094075,
 0.18184021509728143,0.1902298354240981,0.6126029014196659,
 0.24682591036146206,0.977645854151937,0.5384370498033003,0.14513711704797372,
 0.8046704879520147,0.5251611859971842,0.5378485800789868,0.03934201270615445,
 0.7021334321494127,0.7138054021347817,0.4247620070015813,0.26955066436885067,
 0.7270923788202341,0.9605128204834633,0.6222615337915287,0.9670818920954207,
 0.21660680304469215,0.23829945713138134,0.9263285892309391,
 0.7717614032483482,0.3541851363870434,0.4773456609599299,0.06347781151008347,
 0.004249876957615917,0.2935028548313924,0.5991914429143022,0.178693367748751,
 0.1859053607745299,0.30906194685860566,0.9340848208516151,0.4989726243471566,
 0.4420823842534669,0.10722750985447649,0.813621964799516,0.9783446002779241,
 0.3065877563767907,0.052077659705038304,0.23417313985084154,
 0.11582879322340822,0.875450054178528,0.25295616302236934,0.9750185866821791,
 0.36549861735076516,0.6021351045687791,0.47180108143332666,
 0.1619754411844383,0.7972736235051656,0.261624726598498,0.08696255369054717,
 0.6481439102022616,0.5535593548476437,0.7522267698399446,0.17343571348360298,
 0.8882519884284444,0.07873642197312836,0.010765268522345717,
 0.5820183315080575,0.3866254229581436,0.1389880582902162,0.9195727326558937,
 0.7768641697005807,0.20616031826586956,0.21429097936948427,
 0.29235481842677713,0.017411310550141557,0.5015058337553006,
 0.752881216711137,0.21919838693551752,0.8675086150211052,0.8038082489267011,
 0.1333416136129434,0.4562976518407642,0.02392748374906939,0.701418599008812,
 0.9706827194756475,0.8467247870081283,0.5897850727270986,0.8046319971313747,
 0.9487006465027658,0.09878535503147079,0.7712901721476082,
 0.27053015307621464,0.8358486174108734,0.3735626120730847,
 0.49843733863359596,0.746634934568409,0.1940590257862299,0.8832824101759701,
 0.7940305210345454,0.2504351275907397,0.3920027764605001,0.22956345081847562,
 0.2751013348809719,0.3659478139872747,0.8381182032007979,0.1737546780711462,
 0.4335641198239907,0.30368947105184396,0.1850622175702593,
 0.03865272362058558,0.9415630541806338,0.4704899235474692,
 0.12273300387309616,0.3290320858378075,0.38390756290236083,
 0.8089264518665255,0.38127579439568204,0.6513362904466508,
 0.35033669355762387,0.09422429245249742,0.037381447196118245,
 0.31907711068701194,0.503385364458707,0.04187427399314769,0.9717088053782906,
 0.23700997815031444,0.3022740465950239,0.8720587536033975,0.6510048736049315,
 0.32344291607024,0.504189678679438,0.8156497554856327,0.13436846574824857,
 0.5372518098243414,0.6133035739650395,0.34562122753228675,0.8914532486158795,
 0.1519394677283875,0.947249878449335,0.23299326092796702,0.26835111787911037,
 0.7765094472995607,0.3483745321519529,0.6589212119616189,0.953517859042345,
 0.309586451594697,0.15704542339168392,0.46633545018989464,
 0.10732600411098447,0.7302750265969757,0.33160145658277806,
 0.4857601215249271,0.7851528757963828,0.2090594857542788,0.18126749411830245,
 0.33064813541183347,0.23791678514059866,0.7413949887715306,
 0.5581323064090346,0.68312854579128,0.02187956629765564,0.9602709231218928,
 0.9863434169526597,0.35257697867952154,0.12074040981529566,
 0.2826379193179597,0.07404129951182392,0.12307566232104639,
 0.9672086199850377,0.38633698893977453,0.44715281783262006,
 0.9600393501863497,0.5877735824299999,0.8498376161577708,0.4691408044682783,
 0.3862522280475841,0.833801151404391,0.04302387022564269,0.7238537940355432,
 0.48253295100788074,0.8545363908212436,0.9375987653862536,
 0.34603335874022745,0.007486570374386603,0.22846333582215217,
 0.6901160049047002,0.4703752914323477,0.4355782970437224,0.8399486935933944,
 0.29277424728366275,0.8890758658973502,0.4318613462466143,0.4680494273587885,
 0.11650519609114574,0.9558505567327241,0.435714764881438,0.40215803806809547,
 0.8691907421758708,0.36955338189585174,0.13111968172227484,
 0.2112421815864135,0.8987968643161267,0.8961109266330353,0.6658692779744615,
 0.3032453701155129,0.7279699143855961,0.9628408154230735,0.1025309958419256,
 0.3730125721460199,0.28436727215187163,0.14123178633583344,
 0.09575275090546898,0.997511063488095,0.007066408176650274,
 0.21505692316769376,0.36706471861265877,0.5486357239595163,
 0.5712185280238794,0.35762967206875074,0.7894955140867307,0.9067312935111244,
 0.4557230601848804,0.17104203600538503,0.7990359805798447,0.786998728684172,
 0.9510185553095911,0.14071387517133793,0.6668421194633756,
 0.0035761237803709767,0.43264972982906935,0.25454607607116775,
 0.38977659594176595,0.060380988778573674,0.45861447553374135,
 0.1427279778272832,0.4324056627613524,0.8997349043657139,0.786414591707167,
 0.6736017274389685,0.8539733962282082,0.5546013683097486,0.1988174999911253,
 0.15992376436059186,0.11199038207364542,0.20150547223066828,
 0.5880418370977853,0.5323540366786275,0.6436167302096896,0.34587654594745565,
 0.8419073393726654,0.8483502596450138,0.42505337443693403,
 0.24817541188566428,0.5999379865053556,0.9281997847334653,0.6608229338453325,
 0.5556056337884949,0.33763916478288303,0.4990256401409364,0.2014118601037178,
 0.7180559281986209,0.29732836814327646,0.8701800522948024,
 0.10821583418670011,0.5047237313713856,0.9952356513748383,
 0.46534160866537155,0.9027244270712076,0.9100082273867623,
 0.48293589383766045,0.5780415607269663,0.8914173622554619,
 0.28333696763184313,0.9845037069622973,0.8017279529414832,0.1669205751198226,
 0.7851425145001435,0.9924468561567081,0.7198288004380793,0.9375550146495935,
 0.8656789222479208,0.9088297513693089,0.12545799076804853,
 0.14488539915860854,0.1391381510092089,0.9175176346754631,0.7482595979688647,
 0.6011954920198841,0.9616547692518577,0.287521412051167,
 0.0029458283120068485,0.4563903297934122,0.14972513815177435,
 0.6548091909485146,0.5940795935948019,0.10582243932684443,0.6691651444281469,
 0.4999123381542876,0.42190831550794883,0.7113802195272553,0.3906295429102685,
 0.9230986591723561,0.39372190280027053,0.30327885537543864,
 0.47815977953745026,0.8573496480078832,0.2181916766929083,0.6422329065507494,
 0.22608221698835007,0.3378721297982842,0.2169681617355418,0.3733204634294315,
 0.6659242608881668,0.34156907017982563,0.15306456983799566,
 0.6070431721970265,0.2675328233567916,0.900975312876887,0.36040886872995237,
 0.31059751671827707,0.36291247358012724,0.731251602655579,0.6037922854333118,
 0.7224569731169103,0.9880139956676139,0.5257731543487628,0.5948932060868544,
 0.26781423956526473,0.44485832363713174,0.006578073284362884,
 0.30453602606939634,0.43921239643679755,0.035129257835831096,
 0.30028434915427826,0.6047639098532721,0.6251707840728137,
 0.09596102229616266,0.8501225155780642,0.6281433970450201,0.4064664166508082,
 0.002285869617687619,0.31473829913385587,0.8644497994670791,
 0.757579214528485,0.9545485999311724,0.11432565102543002,
 0.048354067236341614,0.6561680771879144,0.055063539701500464,
 0.11190528991853432,0.4191764503760521,0.571899971994373,0.16459946731469233,
 0.6572990548755365,0.01385210429537398,0.16923022068889892,
 0.6434862933917518,0.8603345628290028,0.06835670134900873,0.5049953628645165,
 0.22416642429346045,0.39049999664355306,0.3540109810808263,
 0.3653272070864563,0.6586566823371426,0.48045087380362916,0.8388902149460579,
 0.6413094438817538,0.7971273423156688,0.1728991030558995,0.8008693217671022,
 0.1187786112151179,0.5872364680919641,0.741551855259167,0.7971093280330201,
 0.9206587440569253,0.86036705167175,0.6995100623019405,0.6414979955749343,
 0.09767483966231238,0.27631173088201355,0.7442065529204183,
 0.42684481439899347,0.513101533168923,0.007693636206199894,
 0.3698045412410579,0.31483607185214957,0.06705348584674631,
 0.3418544012336828,0.048835699201517535,0.8969964522413982,
 0.7276125797220377,0.07193145725099359,0.2809965938846054,
 0.19609838257081647,0.17453788517718716,0.7963057934476646,
 0.2685746585741208,0.08290459422326613,0.7607944143954886,0.2059194127966667,
 0.6534917696359801,0.482867851694045,0.077561585627689,0.7322523607137167,
 0.8218947083975642,0.3954385810321269,0.6783772735762287,0.8051753796433411,
 0.7805089851703344,0.9315505513889173,0.48168252047060645,0.9313007842582528,
 0.3255149086637261,0.23020896844230765,0.9209983940868836,
 0.45628134395774333,0.00782571900111162,0.20234358641181127,
 0.6158684329287893,0.025186273133500192,0.08735651186736215,
 0.5210823316218666,0.24509667133692392,0.35638956565736446,
 0.22187899908074338,0.018656219175549493,0.24807006686823885,
 0.5154892544978339,0.49519381650608474,0.9131232282357447,
 0.0016706231811918215,0.14154918353290857,0.7992038360025935,
 0.2568682495963004,0.9791324532017732,0.9066810319736732,0.48590707247514886,
 0.16989486952004285,0.4285692687779198,0.9567091370604183,
 0.38848653035964753,0.6740502102856765,0.5718252424437872,
 0.024463217592458397,0.5623357940098244,0.6845315763485721,
 0.7038674530741474,0.787200669463624,0.5973907905321568,0.37881638048758537,
 0.44641819908528746,0.6889485082703473,0.42602079964713757,
 0.38591780174164136,0.3376706859316556,0.39632375191337255,
 0.14583415679024847,0.09758276911540276,0.030726767271702315,
 0.7886282272651067,0.7670997192686264,0.5579453530550462,0.9521784048131995,
 0.03371003144206697,0.9924976322617536,0.024509657165668663,
 0.9864914779904115,0.3928683610121989,0.6060601940146134,0.9826810904692098,
 0.5011522280833234,0.38841760073299636,0.4716187226170647,0.3696633961027739,
 0.07544254145563767,0.6564234564867154,0.30833815939260734,
 0.6471151773849801,0.6976560733623525,0.23099809144175443,0.8758691143296076,
 0.9996374701305978,0.19067755814348175,0.7471815889517015,0.9010195142982071,
 0.7392788769513872,0.33797089360613863,0.7163056654690384,
 0.18257578779502714,0.4818611785981525,0.7094593776501783,0.4906079391535778,
 0.09460171333246636,0.25827436858390584,0.861608729352816,0.8831852957495648,
 0.708267220297817,0.03331919655982485,0.9524671183847077,0.26574830229968416,
 0.3190017667341414,0.6821236645360343,0.6380243003033024,0.6302738141803416,
 0.5187622035154347,0.41522854248090435,0.11988896644743074,0.953722577842908,
 0.5574907094476536,0.2259562107032701,0.499246998907918,0.09755442631146127,
 0.23078042054973258,0.6262078332835802,0.7398599067270913,0.1413442349586881,
 0.5088299258137252,0.1571333309622145,0.43829823252435096,0.3851699110233573,
 0.816042599241187,0.30476642889230565,0.7383910047816412,0.49161901609204006,
 0.5500082651407258,0.14952424587633917,0.27440801088063527,0.950862370149602,
 0.8247202193077029,0.054987013573717713,0.048278049328984274,
 0.9345630950407915,0.5028481170113394,0.8723046319083303,0.4309789968033948,
 0.9629493136890201,0.6740645288271425,0.8927051580061272,0.822351177657228,
 0.333118636597763,0.4578546495415844,0.22968902211978492,0.4218005387289413,
 0.8753929356473807,0.8279165778575757,0.273338899907705,0.8680515191446883,
 0.3937963161889494,0.2797421337932733,0.2759066238205835,0.5680559347993462,
 0.29397359000640866,0.155607690862947,0.07837582547662358,0.6517914278354482,
 0.6145068530796051,0.758405813312615,0.7990062626519772,0.6060934184575165,
 0.3545509437203289,0.40467862366065577,0.33208926111843384,
 0.9690379362740442,0.8237778129020572,0.47894885151961897,
 0.01584555437655577,0.7040417354628724,0.5676342938294726,0.3742124545731358,
 0.4795188511718689,0.39085081792255105,0.06831930522755147,
 0.6937787519426882,0.5472209793134624,0.4464804591813598,
 0.012820379764869494,0.7443986122899705,0.2738605758699372,
 0.6118779774047761,0.7056903258911245,0.27889194742444257,
 0.39263162337202895,0.5550802919918499,0.3409203832213572,0.9439688237086696,
 0.8848410052538,0.5025000629242606,0.3038572378292199,0.5228804558331029,
 0.010574001537476452,0.8879443329098193,0.25014373276637003,
 0.7500053181973914,0.6924857665549116,0.7719611153000394,0.6532302285264217,
 0.3244715287629397,0.709653548393047,0.7253769411558821,0.5732170299266929,
 0.28239232489907296,0.13704610216045754,0.10208772141011813,
 0.8020693959558358,0.27272979289056865,0.9163921751312519,
 0.18655198521576943,0.23152330098263374,0.04304727091404137,
 0.5347397487250961,0.4015118787262506,0.9647650111360453,0.35766733164177955,
 0.9919069312609313,0.5816203430406278,0.5187077363784561,0.4742218134679925,
 0.49841357286775567,0.0789417556915114,0.8882215630584879,0.3184978469217006,
 0.8292261138126585,0.21475017746451797,0.15133344030353435,
 0.27439100947763917,0.9352795334326954,0.9792580854567955,0.6164643785041383,
 0.11877181253759395,0.8216513935265601,0.4086213433402702,
 0.42549493231922475,0.47645361119023066,0.6887604697921964,
 0.3057917642087504,0.9505938798801458,0.06500551382666986,0.7136163643253987,
 0.7134846690032035,0.42685360151353335,0.809549451824491,0.5050410028261472,
 0.3997374971576019,0.06238000865234328,0.4542020148532886,0.7804468483720625,
 0.1165448983369397,0.35868312873029573,0.571689324098486,0.5488143215429117,
 0.9813655829738172,0.6829763938709512,0.29010706808184683,
 0.16286091479560083,0.445831314585424,0.8468593318775977,0.309713861271283,
 0.8929137835786868,0.7205224045057655,0.9345893384896198,0.7423979214059924,
 0.24173919845519198,0.7758187558576871,0.6666971210740981,0.2401338424501902,
 0.5315297432931294,0.0927750499150527,0.6373907874820386,0.6214332609648393,
 0.8055785533591355,0.595186522955438,0.7734572769695848,0.4062667015721322,
 0.08406719243689587,0.2572917927494611,0.648373133145144,0.4623260275550285,
 0.5238278464265069,0.4207025153921221,0.15605951929322204,0.8571114631851366,
 0.6962719087587877,0.8616004389809565,0.5464012007927997,0.11737256931070539,
 0.18497787221549034,0.8079706959970595,0.9514166131431622,
 0.32508101824242286,0.23867138328386994,0.3104009139809063,
 0.8591680862373478,0.6452631253887124,0.7557973978098615,0.21175285424676682,
 0.7537315979335429,0.11953172863580286,0.4246658263847324,0.6568340948784595,
 0.46461168702342426,0.3343059020853697,0.11544323956025881,0.732321403920765,
 0.030834645903442648,0.21956456311921335,0.9120243855341851,
 0.9331504850365029,0.04668383264523379,0.35883073199044213,
 0.9792710416874153,0.10222678441856392,0.8028356956419378,0.7299091839808721,
 0.9191150758378667,0.35369351137954985,0.09763144434621818,
 0.8276357698686239,0.10979429112963834,0.7711674530232988,0.8783497882217934,
 0.051471510484508376,0.6152352051427705,0.7841421862434843,
 0.08055059468950565,0.21540980983340685,0.22286806466784537,
 0.07230074551559551,0.024485543494999762,0.5892488250401999,
 0.8173990930422901,0.8156671870250837,0.4959147611774519,
 0.027062445105416777,0.2335195307693273,0.20140589848847124,
 0.7039843192371428,0.5077444861777931,0.5732594836098294,0.3743721367437509,
 0.010540738846250397,0.8160704834915664,0.708883440105764,0.5529009612384681,
 0.06689867694372886,0.7437887944584903,0.5511620805960862,0.4463559790395394,
 0.7086775129139937,0.06488279711689537,0.5302612666729963,0.4009151404016681,
 0.1706719641642912,0.10952027445750379,0.27372046292895424,
 0.2671736632073417,0.9313472564818486,0.2585492156790098,0.19433960701781927,
 0.26965520381252717,0.05494471083567731,0.504413646428982,0.8300559420294008,
 0.8488893811531546,0.44406047975509755,0.14660005240090646,
 0.8486114385525845,0.024366428225990222,0.36566660613208335,
 0.5862341543832953,0.6977974324203727,0.4600725981661794,0.6525731917056944,
 0.14900597034360974,0.7158469362674834,0.3507355698527801,
 0.10925387492931382,0.7617343607597354,0.17851996765472444,
 0.43230799582343316,0.8375689371556158,0.8178871192343794,0.6789548230009191,
 0.9032518890138521,0.7715421327607465,0.3635936121622624,0.9961463955078784,
 0.5855036351906591,0.17728217919976264,0.9785631075121847,0.9123851502073914,
 0.10694440473075573,0.8494988000502312,0.8628207308839085,
 0.04784538951932116,0.630353758267505,0.2610330217456184,0.6744307876994362,
 0.9855518435283939,0.8153612298732346,0.49359104653553054,0.7682889622916521,
 0.8671474639243063,0.5137936129943363,0.49545233792079,0.7774107145474414,
 0.20222790660617596,0.43127078743449065,0.4495844438378652,
 0.25987950481082367,0.11122419325430721,0.06284064248596599,
 0.6294337893168414,0.5724360611747416,0.6944437561512606,0.508959770129175,
 0.715966332735126,0.18159454770193784,0.48891371903784897,0.4961861009067582,
 0.4867643798741129,0.26422316574454197,0.13639290633338075,
 0.9634741625371445,0.3555039864169408,0.8862796130836952,0.1724834646565394,
 0.4265178772721647,0.13966810340790314,0.19261213372506192,
 0.1327581465469062,0.7876583651940707,0.03507221018333495,
 0.20978794071972806,0.8862048145189602,0.4427610673834592,0.2823659347440979,
 0.0196899593647758,0.28541114570058435,0.28650151296955506,0.606270197072803,
 0.568457243787613,0.3405610005616324,0.9666217820331182,0.05975042267941055,
 0.2967686639026853,0.087204155629239,0.9934953635126504,0.6564941733905304,
 0.6659064545361211,0.9346789182268163,0.6206345261006345,0.4279464275203745,
 0.5025275351093487,0.2953172343095958,0.005306575081211462,
 0.06826696852615699,0.9787314683360998,0.03813003097707923,
 0.6112792462631658,0.5863218632320963,0.22918392715658953,0.8903572162067868,
 0.4445120294540853,0.6301529230953714,0.19477874873503342,0.7043265252137182,
 0.6501938960069591,0.6608123270732096,0.0033799253750602665,
 0.645634681988279,0.7059108112940465,0.2535231824356444,0.8226850310256628,
 0.7169013374764068,0.7318653949914065,0.28427232634373933,
 0.010876806147338658,0.8481397430487418,0.3057822080886866,
 0.2147896245697174,0.7547614023329051,0.7460600175715201,0.06629438124541309,
 0.304525573296473,0.294265070274513,0.8551539464081103,0.6861083608894434,
 0.5829377913774811,0.06355404535317877,0.2922818361858064,
 0.02295219989386732,0.9394156931073281,0.1347160286693464,0.181506538708206,
 0.4255669043346565,0.35497065331786726,0.9012069835394872,0.1582739140342454,
 0.38392096311682344,0.04590296212786271,0.2704341402789461,
 0.9732412850022301,0.08526171691066475,0.32767970494649157,
 0.7368503175956805,0.45481733763396737,0.9685264352121539,
 0.23594528833551065,0.5095772576060091,0.5470329692472806,0.5632654306675312,
 0.24241369263170653,0.06810535246063887,0.11901538838736136,
 0.3036687813435295,0.6076491460263274,0.16069732935470393,0.7168882776581725,
 0.6475225409866135,0.3498153283110125,0.6316713032844934,0.1668443464969529,
 0.5852217795360907,0.0033213184664457884,0.6092007527208992,
 0.9060131222043579,0.4385519327452072,0.08354330856465875,0.5246285170444214,
 0.913994285886145,0.6412568563269186,0.2781255468294428,0.7011302285534526,
 0.6939862181051276,0.9363024406421792,0.2766100744872587,0.11540329973994334,
 0.9426895996719068,0.2690187220997866,0.37660352592685886,0.771535174357252,
 0.2948860748747977,0.30167406587006496,0.037941281226172796,
 0.9802637320783287,0.2506882934782996,0.25701488659680527,0.3736320948626143,
 0.8624343314278368,0.7631312974958834,0.6287116222471351,0.7618748261345962,
 0.41701147326965404,0.6514804318575707,0.020284196639468344,
 0.24994131994160973,0.7507957975473013,0.28100601611274056,
 0.4971592846109053,0.8350725225337924,0.9791499204073655,0.5736206887574665,
 0.8408985061474816,0.5247598535539191,0.9821498620042797,0.7255393837937248,
 0.9210676479953723,0.9884546418871639,0.6151775280132026,0.04313460225844279,
 0.36344559622270356,0.06499390396199811,0.49892091427779695,
 0.3398114874158512,0.016691232980315274,0.4374923661267035,0.349659332404304,
 0.43375445742165997,0.6737364698516494,0.9392264807165391,
 0.22580080846370554,0.8298281587627017,0.9025600079235248,0.7987591768138265,
 0.3129163171301228,0.9338092672403874,0.07183955795271213,0.3865637823251752,
 0.6387517617377695,0.4809236739248025,0.3656212789385713,0.16321896153172177,
 0.7780101377687832,0.28769468093459893,0.480465679303856,0.9123460919579982,
 0.42999378973197233,0.3822176077832641,0.04355188694814116,
 0.4452281318859056,0.38629541207104146,0.7469910960923152,0.5229701013344219,
 0.29893301258457927,0.05743707372326323,0.8854883452255916,
 0.7089447641217167,0.45813601200520715,0.395773672912773,0.27463856696212374,
 0.31781446528836177,0.8428787117661382,0.5276467482965546,0.9648248164833109,
 0.18732386428268688,0.14224089240311477,0.7418058020979765,
 0.13232568094894115,0.4348107761536659,0.9963491124178658,
 0.20856533073859096,0.2901324298344621,0.38740494249533086,
 0.13577112138867542,0.49326770032456335,0.5051721710804681,
 0.019316366750363634,0.7873838984688236,0.3345017260306191,
 0.27339510105457177,0.8976387553603311,0.43575401522172474,
 0.01746136601993209,0.17833777336397127,0.42059705276961934,
 0.9322373831130525,0.16397838282385135,0.8708649716411304,
 0.060953259775122004,0.9146239811848075,0.9423155392805307,
 0.0470106441357695,0.12116739303813284,0.029962571244611214,
 0.12388491986478067,0.13338655932185917,0.0461124200615608,
 0.13801948565315958,0.41832408445671254,0.9165099142484383,
 0.09393960260720369,0.36758913574652774,0.34424535151746716,
 0.7197762880299969,0.3198720383964646,0.17839365104283256,
 0.027351601162971417,0.9515093163064365,0.21128714994744568,
 0.6512551898939298,0.2326371450471879,0.279902482222251,0.1411480367945449,
 0.950512817721212,0.35420715256822566,0.5482526010873989,0.7269110144118366,
 0.09518556980071435,0.24272308406629983,0.3458880947210269,
 0.20228262369775662,0.36423816375259943,0.7336262325584841,
 0.7307869836072941,0.13145188436370647,0.0072547581071849,0.6184305187584738,
 0.15037886065149386,0.44088605609218035,0.15654497860321193,
 0.7448841539539074,0.8866502502354595,0.3871078768550844,0.243334817632064,
 0.20026068468329383,0.22214391306582726,0.9933686022528382,
 0.7925489929769997,0.40152557187746174,0.9439425338895356,
 0.24421414739728908,0.3135915375448943,0.20955955992104414,
 0.1996571160841344,0.7812667642420837,0.5977078839348426,0.3746617809692785,
 0.35075193395733495,0.9375714930329787,0.1725665346285895,
 0.12391690582951076,0.43824850761810286,0.9066426983279704,
 0.5640201121459685,0.9689229173576862,0.5595902677867974,0.610150685559207,
 0.30794634710693825,0.7259118480172048,0.7426037952117395,0.6630675403030921,
 0.6789296172016673,0.7953719957683596,0.1417532471573565,0.05454639117795046,
 0.545700585584344,0.60825270228106,0.10516373725625572,0.9246877211363687,
 0.5661280942028833,0.16497369974227605,0.13129367650070267,
 0.9835050241652156,0.18187825910574562,0.43358902022140355,
 0.3675740117149726,0.7127745842371401,0.47496293455730176,0.8487390334670547,
 0.7023293817654682,0.23844492148984775,0.7192893757985129,0.7213439194973456,
 0.5612986428328495,0.37567273201450035,0.2875215528283037,0.2139767891456139,
 0.8112459578907152,0.539129982284353,0.33750766829527645,0.26146173119401817,
 0.3265587572109836,0.8247328235285736,0.08158694998537774,
 0.40627000705759275,0.3645342126458051,0.7349542894832638,0.7658962272059054,
 0.8632818081544129,0.09952856507756203,0.2137926796747276,
 0.05131975309009029,0.7956192281863126,0.860448792258184,0.906950939030112,
 0.6772900248550692,0.9245527408901145,0.6645672525460462,0.8261337239901569,
 0.35332781883854636,0.8615882651453681,0.7299884015779625,0.7417331004741712,
 0.6627895999345126,0.17919382867974454,0.8541639047037359,0.5475276833972815,
 0.1938123673747303,0.7649906376662536,0.31607711373233793,0.9935205744081955,
 0.7447739425556097,0.5403710954611388,0.9131562206625202,0.47084077643833344,
 0.6719188403293169,0.7268556843082359,0.7876003605830747,0.10457584131361142,
 0.37637963785648143,0.025187534817402657,0.7673735602989604,
 0.5021351431297607,0.441460776055855,0.8718552084362756,0.21948663971619375,
 0.19549518152515044,0.9490074834942792,0.39060266285728373,0.946321530033137,
 0.2799476069781428,0.7081191742713059,0.12720264374203594,
 0.28641779182592686,0.6831471190820267,0.3046948814130185,
 0.02977370855356143,0.6848977253825727,0.6803590760587198,0.1260738630684829,
 0.44308753095356646,0.41076715587385315,0.9609271058266552,
 0.4460197074653077,0.6701060451611742,0.6750349533578166,0.7289961639740256,
 0.3612505597685822,0.7062522937944066,0.7800471115317352,0.4990454818582234,
 0.7347552370884197,0.8273656349864155,0.002384958913030255,
 0.4754246382785474,0.22616976661847954,0.4853564299525869,
 0.45226595771093914,0.9801044944463788,0.3961262409680374,
 0.15875425137496812,0.37566016287037673,0.3208650873344481,
 0.8602695552280755,0.2784434413357899,0.6758874840485114,0.15047658983217171,
 0.21821966367914314,0.8696135084487646,0.5038505822539625,0.4371696078302396,
 0.04151575760436066,0.9660446542173995,0.9417174081758397,
 0.04195959571714036,0.9218879697946601,0.11193559131201347,
 0.45281541958970273,0.5242445082051597,0.44294002945326116,
 0.1274559227205626,0.36934479143280685,0.24796444070699963,
 0.9211684314994966,0.07796414832546184,0.8402529672940539,
 0.15285404531986746,0.6188004807971368,0.415818322090343,0.31335281582983987,
 0.8111685784698804,0.9118222420173758,0.27303194587528445,
 0.29294653007888505,0.4115427638449166,0.038919641838965124,
 0.42302196629836986,0.11207121734276182,0.19912450353620015,
 0.9155993692222368,0.14082483811725444,0.7800757688210498,0.9595895749196208,
 0.8939032480220733,0.7524788246191119,0.05041960479942442,0.5920188460375371,
 0.5735705085660139,0.7620250065757522,0.13081925845932374,0.6785012820320618,
 0.5276804659954091,0.6520858534431287,0.36312431862533734,0.488700434460547,
 0.6452554060470375,0.13216221162450337,0.14010696736417871,
 0.6445912923319281,0.9249495790090717,0.8230937761035568,0.7721335511209633,
 0.7521175306899649,0.6763744245741481,0.19865263941458078,0.576142304092754,
 0.6929599486375937,0.2706842914870746,0.8966885591602001,0.16491793934530696,
 0.5817208623298596,0.09560144624710054,0.7624155131237961,0.5329460312120333,
 0.301707423124284,0.8955605567612358,0.45938515460231755,0.24297623862702888,
 0.7776052177134709,0.5406023712273164,0.39312899792032463,0.0867824528288429,
 0.17996087217492973,0.265474826927087,0.5097871145452187,0.48001624454584213,
 0.19998482669188122,0.12246900401639249,0.8934006661877103,
 0.3436406125121939,0.9380088873083475,0.5016552896137702,0.8924583584772239,
 0.0390141056135237,0.6282634888633458,0.8382668581488252,0.5254189761672269,
 0.3691295868193152,0.01703673513892434,0.39522266169575226,
 0.7538920587956971,0.3388568660374319,0.7637329747453182,0.7520694777691788,
 0.8019829153580352,0.028319163569070938,0.8582240114114218,
 0.3557227694616909,0.9132394203477027,0.60828725795172,0.5865521232295541,
 0.7890417121099966,0.23670484601095165,0.8778131899117172,
 0.09285787172037807,0.6951969227384707,0.3203335577566562,
 0.008070112806806695,0.7292957396800435,0.48505528927804864,
 0.4926715736256716,0.014496189268661697,0.9010091816631819,
 0.5018258242519305,0.9318413390984299,0.6958070224300082,0.9514023917497257,
 0.022370615231124935,0.3106029784427382,0.5050422853927323,
 0.05244381024711953,0.08416251171793421,0.030716057692720677,
 0.7676168727034528,0.9721130089211006,0.6634855695934601,0.4757895016956636,
 0.11011350450013624,0.7531374007667335,0.5268828310932553,
 0.25284197826498056,0.2846917936047406,0.1772055571716209,0.7537431450874073,
 0.4246013137360757,0.30173357846048043,0.6563790715900928,
 0.21670333661983665,0.7768697331050706,0.7975040304180562,
 0.26452085414997595,0.7108301738330398,0.3653280743511391,0.5694560576102967,
 0.24177398147225582,0.06777357178222265,0.7503033121787435,
 0.3270834835131504,0.45651641199327075,0.6218443982220772,
 0.28630035140331134,0.3622497289516836,0.6302428004391835,
 0.07896041860744785,0.3901629479098786,0.04114213748709272,
 0.7727415572920723,0.41416740122698703,0.21216378244952017,
 0.8028086449873165,0.12245462639545568,0.4447033940970184,
 0.39044969461976053,0.9136659716411714,0.5274497789056724,
 0.25751702194906656,0.9446890265875335,0.5935122447844672,0.6821079137984406,
 0.15348187923629097,0.25451193148633244,0.559224197298423,
 0.10186101903215117,0.22519939508989162,0.589196323362994,0.6178850834085373,
 0.2569958291647363,0.6134852977622627,0.24248758041043972,0.6588253016415102,
 0.4845446583763179,0.9866804119657836,0.5527174991830471,0.8025484011114472,
 0.520338514958293,0.5339375335862893,0.5259989452863088,0.43983474521608035,
 0.7883227146129602,0.5614384090430125,0.39295228658718995,0.9269138145614019,
 0.9566113801465169,0.7597643535272701,0.38574103521186753,0.7891037817136469,
 0.03801465873015064,0.644368267465249,0.3795015952273406,0.8285399776741927,
 0.9953310542463112,0.6452224251481178,0.33584988587804865,0.8712210078088618,
 0.543531072646759,0.8928601408679239,0.4760965752150108,0.9715681434677037,
 0.8208398924701077,0.18723881739642523,0.6461329829513934,0.377631985118022,
 0.07402235261245549,0.5490245258096801,0.7708893227194358,
 0.06778607326623043,0.43926077011476006,0.6040109203792726,
 0.9202063579752191,0.9834754574168159,0.6869293612716998,0.44441483536767223,
 0.3389308188008884,0.34366361867174244,0.2815833011292902,0.9017556825485376,
 0.3566968229333286,0.05966842606577438,0.5545293666682807,
 0.49814460658168835,0.09494617369764491,0.048250786420439784,
 0.39373742929674427,0.660647340580224,0.06911606601932052,
 0.22461506509341977,0.59316082625065,0.5503397738872232,0.1871742498670691,
 0.3352293681679933,0.6202442979189144,0.25828359039619087,
 0.21789872467389837,0.9684700204204446,0.10497495087712005,
 0.2754493895098199,0.7390651828851941,0.450403075892304,0.8612921680468989,
 0.7045165774548853,0.9755060646716438,0.925987846124443,0.5636126454768275,
 0.6950158977703974,0.5293112450519677,0.19958523143372353,0.6039666456897688,
 0.9040427670307223,0.6044469791971914,0.9023250870142986,0.9069950173194028,
 0.159252721948984,0.5802070056154297,0.8578106661868808,0.7343436848725766,
 0.7908328943298573,0.9657003080836067,0.44121731663855834,0.4826971169061387,
 0.3608804226062161,0.4422675206237088,0.7443011271148066,0.8586063801934404,
 0.8945519167353324,0.5881019272705307,0.7873755237212187,0.20655127928556238,
 0.507314512535173,0.8547235644249038,0.880727334804548,0.5946328790809488,
 0.8552188854318324,0.33847470274781744,0.9894728112996147,0.393333014173768,
 0.6767870697276076,0.14841722828248316,0.3903849779920072,0.7083411991692172,
 0.6671308001412902,0.05670578835727513,0.5927876635833956,0.9935488655196387,
 0.2783593880659565,0.7303231327544483,0.9465944410019529,0.6853663659915825,
 0.9861301582544528,0.7568805325425462,0.5016937338548767,0.6100064692047178,
 0.0011680581138495683,0.6413604822577728,0.4693009278950441,
 0.32130839667484856,0.5230897446663733,0.43763634200230483,
 0.38474562135000223,0.8080222066249614,0.25063574768994046,
 0.5881391289376664,0.4765797901287726,0.17723255659399784,
 0.34084002111672507,0.2820871560311282,0.7028666385401069,
 0.20099550772620478,0.19265074830192297,0.2221832924101087,
 0.015120954986305501,0.39412344524317167,0.4344736458179659,
 0.48285115839254666,0.8582135731770252,0.37708629669305815,
 0.3141968431038631,0.1031293062595271,0.5424234783963471,0.3058303972861086,
 0.006569127396703567,0.9874236837625602,0.6430382697477183,
 0.7953274992862392,0.031112358605887458,0.4992852776915522,
 0.8836926825154902,0.04361757357425733,0.9319062723085435,0.8993952630563732,
 0.7944257672266437,0.9410201564571695,0.8661743787473614,0.8794210442845022,
 0.28447572004600363,0.7029905023812175,0.6795387723559678,0.7220625133767095,
 0.4934245142933399,0.15672925084460887,0.6299136080668207,0.1545446166969976,
 0.727530514890269,0.7911572635503551,0.8120273966795745,0.2639559649669704,
 0.3911260376177884,0.8594222125898576,0.8744183009839623,0.30880800351305493,
 0.7954127646028244,0.6103666953023456,0.6213633638247125,0.1259129808701146,
 0.12146962711655296,0.8451855430312785,0.2829526549421594,0.4117266132116958,
 0.8433965049379435,0.5338803432856061,0.19537313763337083,0.9246777670525117,
 0.10069370718534709,0.5027941585184015,0.4861292182617195,0.5528532653972024,
 0.48108392256849764,0.5526779781969865,0.9243542696340405,0.4031969892623305,
 0.6977931374801277,0.5349863576838458,0.7258904584644059,0.41331661274048614,
 0.21657060579087528,0.6320547202596108,0.40707708738542747,
 0.3121969126945271,0.09694959649561985,0.5008656077037261,
 0.12175874232957573,0.7116352371785514,0.806643155384823,0.08689884399741721,
 0.6182324511489712,0.23247735876569564,0.6347526613609521,0.613324230392852,
 0.13824252190365371,0.13786121618374358,0.37845452080298314,
 0.3848675110409562,0.14123932678562656,0.6880778215487465,0.3134571749364049,
 0.32941255710340045,0.5315270036043972,0.11897548892950138,
 0.18047240733115122,0.3209924756528514,0.893173702968775,0.6444822662675209,
 0.9529432055554269,0.3918376833408893,0.5444988676506711,0.4427008684427105,
 0.30469451958152727,0.00261431450503935,0.0751447732373075,
 0.4286392502963955,0.5300495547302421,0.3892455253842111,0.6491462175395328,
 0.10018764485145892,0.3285818882853191,0.7081945214168143,0.3391589955254889,
 0.7877580953135137,0.11916717760614293,0.9195730883720966,
 0.09482147556135834,0.026855145970250294,0.9260088547817539,
 0.3397585429616775,0.427355476446472,0.11261964915535572,0.24398256490687875,
 0.46942895200416646,0.2100167234943804,0.3305521590995265,0.3372970605261877,
 0.6452624983334112,0.9286318657704494,0.8675610332425476,
 0.003496751440823176,0.43747049828580653,0.7168057793546943,
 0.5824987666527617,0.46287959610466245,0.5703684371092204,
 0.45033198942824737,0.31345114818283026,0.9178999208027728,
 0.34902989011328556,0.3412161602449979,0.06134152319212438,
 0.6994342284369698,0.6696342045561453,0.7674563858304825,0.7479254867982353,
 0.948747187249203,0.15930655825377993,0.26732794413532057,0.9380468875291008,
 0.7941315466160673,0.04794893784528642,0.478883616451661,0.45792159732579796,
 0.2514119033613603,0.7055609371650341,0.21134301183784188,
 0.21638123034170786,0.285221883003032,0.10390958665961036,0.2825367777823682,
 0.7001324193398056,0.4848057351923455,0.7526883185647698,0.9810655430106773,
 0.9467219519031099,0.8297285935387849,0.7767804007885671,0.9850187388537487,
 0.08722489571362668,0.7325250373414901,0.3292852067936798,0.6461505942800135,
 0.4776938270156258,0.8714754428086167,0.17221674667559028,0.6607827950033158,
 0.8767039568368479,0.08669852871009442,0.46306281933327675,
 0.37922350159804175,0.8935459553369696,0.5988116247533087,
 0.09376088667108251,0.2659626594142481,0.6966071898576499,0.8235079210254495,
 0.8071146366475634,0.48252112053534135,0.1422307294171805,
 0.09051072568807961,0.39739265000458335,0.45766578932037116,
 0.9436888970648584,0.14976372079413847,0.1022273982403279,
 0.13201514216854393,0.5210754809449989,0.4410694598920101,0.9293208959754601,
 0.37160576266081424,0.03249131342284817,0.26463147158190004,
 0.3533272359732801,0.6763557811895765,0.8483695334563278,0.3080515308725411,
 0.3694854244255511,0.8042899801112255,0.8138252725141991,0.8368790902103275,
 0.18997088527410533,0.5255668144319466,0.17465224298011672,
 0.42217118793843544,0.07995696013318998,0.09427994077240176,
 0.5623573191255312,0.5637089151748796,0.9895064400351099,0.07020270801519313,
 0.8118706538061721,0.6970747800351973,0.5429884757822286,0.18519421383551815,
 0.8019448201863503,0.13363063873932457,0.616192672556957,0.9268687897605715,
 0.08285566049480331,0.5749940607224527,0.9944561842971008,0.5332977587361906,
 0.3270271125900561,0.3714070528515395,0.31579534349443505,0.9463651249468015,
 0.10043632649890444,0.9413773511369037,0.37406710556382927,
 0.9559409485064823,0.9746470268871127,0.6379953249396549,0.07655587695756827,
 0.18538125773175085,0.08720858174702562,0.8445951784375565,
 0.7606143622719644,0.5226951829415002,0.6083956635437999,0.26940966254369114,
 0.19446916984230456,0.6065951525506682,0.7401123947763433,0.9665460484722773,
 0.2046928213443182,0.22623240438500303,0.022817235127662316,
 0.381946676904357,0.7068948762940604,0.04288331067374229,0.2963271563007349,
 0.5672241428258953,0.38815785314569906,0.1912340786633121,0.6065506496773488,
 0.027589249850518693,0.946827849412649,0.20100420381979722,0.702571292718681,
 0.8188924478780004,0.49610342709318966,0.969023966810899,0.8503450055755294,
 0.583107800997937,0.7762164936069113,0.6264869962425086,0.9553241048435336,
 0.7964797990103496,0.5257751785594571,0.7874951011431935,
 0.019292170371202566,0.2284849757265135,0.2223087603932863,
 0.43434193857316417,0.8870236632054609,0.8006826515968998,0.7638805361735687,
 0.9973180367067642,0.9640218051875471,0.6145463855304237,0.887244855476379,
 0.015828899086693826,0.8406973637668267,0.05884092203927804,
 0.3974265200057283,0.4342886663184502,0.43326774631348464,
 0.11309304173241386,0.11864688668354995,0.572878704184467,
 0.006975767745764916,0.8029376842902002,0.2990147768560284,
 0.23168035534323728,0.9282419127117579,0.06864586149150675,
 0.01932940748864409,0.569491323250055,0.03528567141494687,0.4431045736613217,
 0.4431506379828658,0.16756491844459365,0.19234393330928767,
 0.26766625554194023,0.6697703745178618,0.2945854701302194,
 0.20952269138377622,0.5526018003030972,0.1666510198252823,
 0.032515036536513664,0.45730925261661837,0.7256367774715949,
 0.9779034047348648,0.9776965273651428,0.9247658752023501,0.915772120853664,
 0.6406219433901271,0.030638536110684678,0.45257922349218926,
 0.887804312352916,0.7718502285869753,0.5083100138696337,0.6987384735656286,
 0.3587941339267633,0.6045308708960724,0.25164556598253507,
 0.49148740314413264,0.26353157800141414,0.8336890706036627,
 0.7712174083695384,0.8179556590903765,0.5128004469722698,0.6735269332452567,
 0.2560833911735588,0.9736191946691115,0.5996777598715555,0.46278503548071526,
 0.4678068003590031,0.5893995347607255,0.2851769138594419,0.4019841841810672,
 0.7329158013035374,0.6981209317435216,0.2554116640892695,0.5784819369913796,
 0.544095863660117,0.3787860699990899,0.9016161454308953,0.868794322275936,
 0.2951154519829903,0.7644242189847473,0.2723622061425649,0.4198817636496034,
 0.7145011077943049,0.4280069720241503,0.25431975375293303,0.9168454380849436,
 0.5721689574237077,0.5745955258068781,0.05412740879937483,
 0.20975841907185033,0.4537398204585751,0.8303187495580131,
 0.49678753046741797,0.7997916817532946,0.7725318812423295,0.7390665006610541,
 0.4982463401390651,0.6010585822270894,0.6737163761063836,0.4747204599482111,
 0.46960186546697436,0.8377560492910946,0.6912254635803124,0.8633991271583941,
 0.825161528344918,0.5145336708216276,0.34462134857201443,0.3947899327226133,
 0.0712264988220862,0.2011697001780377,0.2735909361785409,0.19884249379323105,
 0.5515266847667375,0.7193749212847192,0.041625632553075054,
 0.31506597148789983,0.8846767828135991,0.16030718284438372,
 0.8766977785770569,0.6272831888538376,0.38376752047852813,
 0.07275297338778008,0.4010026107850544,0.07723040756211208,
 0.8620936770878019,0.7967230940069707,0.6336224931242309,0.4396976307348677,
 0.5004199509566966,0.307409769206372,0.19442927814414002,0.8384975171296193,
 0.02524955983326349,0.21574722521173306,0.5528367882399846,
 0.1471353058680559,0.8330049904888612,0.5783507913880442,0.6955479923899268,
 0.07746171238412303,0.8508903853358094,0.3573039566453059,0.4190611035331212,
 0.9661078008037517,0.11283965063408197,0.8459591493556339,0.8069714993611047,
 0.10955650285216767,0.3131291988837749,0.7926057962855014,0.5068980282847824,
 0.05770592008292019,0.6107092745518653,0.5273173664886935,0.6340214122604078,
 0.4518893384635686,0.9963780511261697,0.6952366579754324,0.5927191577344909,
 0.19654229355728492,0.4886331064059456,0.8543504323945429,0.7925110015453156,
 0.07856855939748908,0.20761481019746741,0.502843603383618,0.9949237302478156,
 0.9002785959657552,0.4628277997749014,0.5944819479505666,0.49277673047963455,
 0.11139223927015918,0.5671011079355963,0.4658172574704642,0.9150262476770719,
 0.669369798791116,0.46784566516805604,0.23992363245898685,0.5404571511852205,
 0.5701854727679921,0.8857290630822969,0.6850667882696515,0.36287141156966884,
 0.27955744053209364,0.10993822550128485,0.12048784164802229,
 0.39546081026094115,0.8144259332281308,0.8241160125175647,
 0.05099371991891477,0.45493481987893436,0.7765701491490606,
 0.9038384520605165,0.7629530631317742,0.9473183546420533,0.9989346402489833,
 0.6913402779007702,0.9495825815325203,0.4575266233408719,0.5191322125742526,
 0.4276640240439866,0.5940247605532257,0.8921817231834898,0.6427232717170965,
 0.0034003997486720117,0.24331196209111683,0.8061552347111809,
 0.40947139344987404,0.1661724005150006,0.8105195841431765,0.704773690387907,
 0.39308620032486885,0.991437188735302,0.9932399710525015,0.11778268267738201,
 0.03144702125508636,0.5071778235987326,0.6956322669812963,0.6969324780953476,
 0.7152207144598107,0.38625295554532457,0.19268451900913908,
 0.40207097452036145,0.6783704653328564,0.3004735574935258,0.6392701340496573,
 0.2738165010988943,0.41134424819791804,0.45926401379756643,
 0.21743353685808975,0.5239708261918882,0.13053825585300038,
 0.17525836541887396,0.7127435701542322,0.9948249303822464,0.8625727944944629,
 0.9177618660893162,0.4558112490245778,0.062310136423637585,
 0.4328253443577921,0.4574172714133421,0.8648473345855224,0.8222188333117066,
 0.5131649366675484,0.627301235820363,0.7212434012140594,0.8836804375277244,
 0.5767162703103761,0.9792997085534527,0.5195849596055406,0.9439744472530298,
 0.25485081118255004,0.10600401447264673,0.5478911388392487,
 0.9990323454485551,0.9872172450810215,0.8947580896608082,0.9797834445108637,
 0.05250876313488673,0.6306375932829509,0.7312488804167396,0.1264916619124128,
 0.8523337489624159,0.9393545258225678,0.8002330215576706,0.06445054786685867,
 0.193661798570279,0.46272250643643287,0.21738623154565628,0.9539585477502801,
 0.6227875132656355,0.6691674911965749,0.2945900176005062,0.24470670078994516,
 0.659456739919085,0.33640111946481654,0.6617943429023191,0.27732391557523683,
 0.5132522166851367,0.7223448650204866,0.05984715681403341,
 0.11667114995048955,0.5500252980726873,0.2837396898034922,
 0.20387107815192884,0.44137486525083147,0.0344744418538665,
 0.7505977987242981,0.5851362037704184,0.2952463301688091,0.725524836635289,
 0.5707194696625366,0.6938150430469737,0.9253426237636503,0.2293810563575973,
 0.3541267104218404,0.4947284074825974,0.42144322189745753,0.5762996201822482,
 0.07734950895623838,0.7597022717677027,0.7953981395829034,0.3626100374314377,
 0.05089467526194391,0.28640258080322134,0.6492976373607144,
 0.6327783596645778,0.5315319514893069,0.03714536238787669,0.8860568488476891,
 0.22063827570288363,0.2222630041501813,0.7384067118919243,
 0.06875802891372751,0.029971765280540574,0.558419958601128,
 0.8853411996583421,0.14050825592592153,0.5949461585127744,0.4252834565596041,
 0.42767180621456324,0.14546968911381342,0.9991068225495775,
 0.4497101057591003,0.550271643564639,0.573778479387123,0.45739637632428487,
 0.2674738618604058,0.3283806838930605,0.03078613377976047,
 0.13779477829872921,0.21115104472149682,0.15559120778668312,
 0.23461363495478138,0.4620848360162706,0.8365504779048863,0.5518763609845877,
 0.9916387591817641,0.14220559814688427,0.7815660149262594,0.6530507938161169,
 0.8565644277232117,0.01893755216603288,0.8887080363816211,
 0.03445881590081257,0.8106446956068422,0.0484150338415239,0.2636303353999646,
 0.4512056652747459,0.041513606991605556,0.6449249354011775,
 0.4301479208690383,0.5587787762860813,0.11512565183000101,0.617721176927137,
 0.29616591369037615,0.8046613285028341,0.35720071651197394,
 0.7055831093907918,0.2640770899781195,0.5285320002839858,0.7206094997256576,
 0.8393143139496435,0.08940144419989005,0.8939055395887068,0.6303919284771131,
 0.9615611157130556,0.9259851559666078,0.39403032510697455,
 0.12404585375656718,0.03750144665079502,0.29968058739905323,
 0.5433204602450386,0.08629801043093288,0.7037084408660788,0.9219625281396119,
 0.09162352563964171,0.14825333970225174,0.9541899011372221,
 0.8186018607961262,0.1594124847771463,0.20259484603435163,
 0.21469519089514744,0.7906408844236535,0.9168360126430787,0.9516134156440237,
 0.5287512765366945,0.623174308099595,0.2528980343290036,0.3774041095688584,
 0.8879825915780061,0.3093380208714419,0.392560961369003,0.8811815547299302,
 0.1485020697816526,0.9746869256009991,0.40005578067612,0.7548078947473192,
 0.09483727433798372,0.6890239565970974,0.5761360556644047,0.7915262063159154,
 0.42479719699333207,0.4791328891668889,0.891317657046266,0.4960748466790478,
 0.8552874824058978,0.25537704105058834,0.6765662803964595,0.7207961050142182,
 0.5548693036089375,0.4559814155503621,0.16363420373345317,
 0.28324514310923565,0.6256641348983214,0.7873053077421401,0.8889352355324323,
 0.10695420805890798,0.7132871233902845,0.46910322463623677,0.443759359930641,
 0.9984318719801992,0.45711666436480924,0.28711508013590625,
 0.13444763203772447,0.15322068211205042,0.04167295477904065,
 0.19358405571466064,0.7121185283882863,0.9408827425439124,0.9601638628461844,
 0.5586148440915549,0.5478809647017147,0.3518396846451729,0.8111373208664405,
 0.8745884352325252,0.014745997186029314,0.1386907640628856,
 0.6039274670027437,0.1526683578247292,0.9278645282557615,0.49673084954286606,
 0.5592468207153268,0.03957331724057356,0.29589925335573664,
 0.17565086942475006,0.9427750901526154,0.5025607918059578,0.4049654190393477,
 0.7950695349255572,0.948666669350008,0.510251228828744,0.5638047196184625,
 0.89693545462524,0.6321613729386115,0.2776054840561667,0.07629355444101205,
 0.8265254454212935,0.19408449195346356,0.2625269872317617,0.7740564914571065,
 0.2254079729346472,0.8568080867067248,0.024457950152421137,
 0.026095277213697532,0.9897520513064516,0.6258749016454479,
 0.22746400808755585,0.9664373516118,0.6612514866731445,0.6692738545294655,
 0.15637683621544451,0.8063115672340834,0.037497433462385654,
 0.18048495926210484,0.20841659707455207,0.5097983373001408,
 0.5305362124072835,0.21289346421392108,0.4074808657759329,0.8974077226728241,
 0.6157665132264372,0.9244663331863318,0.5971031136273742,0.930824821484087,
 0.8482785847787815,0.8482811985499141,0.2598601331513858,0.9342072831371641,
 0.2257269933995426,0.5198204665726985,0.6138487841728899,0.6781444933167987,
 0.11834994125784115,0.48207126089637975,0.053161888066055596,
 0.30176948269721915,0.6131101124475252,0.08295361753016017,
 0.7161871737862042,0.08073914295151652,0.6372710976015994,0.8481226412108815,
 0.8843512180441042,0.13533963888039335,0.11195987812616814,
 0.8255782300643258,0.9796641081092765,0.05423655978983699,0.4229783951532998,
 0.7124037828789895,0.29676843747913273,0.9048087489042721,0.6363855478112755,
 0.729795223323579,0.33085246986868055,0.37168018098812616,
 0.36444879805782615,0.2861255330236856,0.5411513472927931,0.963041160086395,
 0.7471015605788851,0.6778521391108543,0.49476623915364604,
 0.40787781986470173,0.4997969567075373,0.6594265130201638,
 0.11947364442060737,0.043793404418423165,0.6574393960188407,
 0.4003637109350031,0.6239906326919885,0.7827162638277013,0.9510969707951378,
 0.7944914517448983,0.6725875960351182,0.5036852112361291,0.9701895701618771,
 0.3213113631232809,0.9304049736718709,0.021187515755326958,
 0.3356072744644333,0.8989828488409987,0.34252137646678704,
 0.45816037447076197,0.275347371437884,0.8206057220525992,0.8981224119735808,
 0.2754105405565346,0.3485349745885771,0.11155693659942922,
 0.02518971745598786,0.3653014299752355,0.9172088109734519,0.650464276799661,
 0.5851982847491765,0.8376544933462793,0.40797024244070856,0.7387562833364778,
 0.6985740208935114,0.4928045282340061,0.39585768136732025,
 0.08214549615887412,0.9136728858757514,0.7274384152507143,0.9723081332939413,
 0.9008343282275095,0.1179118962635517,0.9786857328397275,0.1551415879659508,
 0.9909120469025703,0.7410339050533126,0.405195234493968,0.16154513321270814,
 0.3810892259998643,0.3847209299501546,0.8201362503146297,0.6257785938637674,
 0.12138651445039428,0.5163771452671794,0.5361116408616864,0.7955481141322119,
 0.14688577573924255,0.6153347384799932,0.9075110185899394,0.6100014835752026,
 0.8798090268515285,0.08479409269761307,0.9559178572198275,0.7990490717499794,
 0.889159012867195,0.29116617648584375,0.022561142552814162,0.27528357497234,
 0.3286992181561009,0.09540710250445017,0.8862172320942139,0.4721632669055711,
 0.42215089622651936,0.7429121019972826,0.8649643505965687,0.6993836268226937,
 0.026543513455000056,0.4539656518963051,0.13140434507348087,
 0.15067867339898477,0.2255170864604813,0.7469889741489555,0.7527571310161992,
 0.09389673369195917,0.1521884579889321,0.80198007359741,0.017347490527001574,
 0.9130813467303689,0.15748236040776753,0.9779288587707251,0.8057039303447199,
 0.8960843936194973,0.7506372971623413,0.23941709145306644,0.5403408587277577,
 0.26492183096815547,0.4547706169258888,0.9285449139522602,0.6050844055985665,
 0.22430339954804457,0.7008118475014087,0.9260455915922609,
 0.43272460526078815,0.9310268462066424,0.4552148209352118,0.3525028098021754,
 0.6448710230443094,0.00987338479125166,0.5072674546398256,0.8398386487356964,
 0.7004989112209082,0.08527226435177704,0.9681216440722488,
 0.45201613808182295,0.47341825416625793,0.9829754459197751,0.987500108975269,
 0.9346821640328676,0.6798835917066113,0.5988303784298133,0.0882782217610284,
 0.179724981446726,0.8834841857227973,0.227156210401396,0.8590104246602626,
 0.7880437093658833,0.03462109813275838,0.6916379157993225,0.0809593074437005,
 0.02969240223534375,0.5862533149595596,0.7081647565911897,0.6272054284701906,
 0.6084980207140598,0.6758787680413471,0.4283038095092061,0.4821100236113982,
 0.8183809936761643,0.037963383020130026,0.05209670436766611,
 0.7066653870395384,0.34401010896148043,0.002784117956182719,
 0.12006140621557704,0.2918711229668648,0.28621236989321486,
 0.6854165863445361,0.08127670816447208,0.3846079237974993,
 0.16202295326152505,0.22493222323015294,0.18287867812493974,
 0.6332570607806185,0.5836674720705531,0.1856419017170574,0.2542940022383333,
 0.27900694609869425,0.6080305387856182,0.26424291952781187,
 0.9884493143192413,0.28644192507786825,0.2823931647526633,
 0.09213060575782706,0.6783111956593673,0.3452870917600164,0.217938338437213,
 0.6998319039625644,0.7800433596483951,0.7746677068841248,0.2504161028564611,
 0.41594621252662123,0.859609565557034,0.13684762233947168,0.4754358755005096,
 0.035209981171088534,0.009915663646415718,0.12327076199411002,
 0.32092873961432067,0.5628665281392062,0.557508498343352,0.7179496873343588,
 0.30721099935489626,0.20272731851097037,0.9925828141390967,
 0.9870222562411405,0.7478205227620351,0.26252960766190214,0.9000967021150187,
 0.742292096269225,0.9053179884276625,0.09571639487753336,0.2682651750352374,
 0.3837643183195678,0.388492786098137,0.1487624492712455,0.2200774514478584,
 0.8735450293965333,0.35956319503248924,0.618544771002947,0.540540432824752,
 0.05297065210061325,0.4370272910206219,0.8974816705615752,0.9727011554282694,
 0.5213793589519421,0.06370627323177125,0.3156808616582858,0.9088011088186116,
 0.34546403012114535,0.7859998942377993,0.27993780206164587,
 0.7990450452330464,0.3171855659019689,0.004469447289684236,
 0.4922646795347494,0.7287561221192019,0.7355872103460601,0.8823011432109126,
 0.21866094875034026,0.17328036156809268,0.026345566395508335,
 0.22916660493515428,0.6548923002895515,0.6309306637896357,0.5530831973266934,
 0.8390081141846171,0.3682757531409062,0.3668872092241219,0.27850180043726835,
 0.6150852167741774,0.1290444709235553,0.020876851171290012,
 0.8093830621096132,0.7849747105457066,0.7250372078219851,0.5076989011299808,
 0.44940088526167066,0.7401041319294632,0.1868529485799464,0.4530840931370531,
 0.4294527360735133,0.20759357819918844,0.42841810241966083,
 0.5557303478414304,0.21691736821272323,0.23334646763929656,
 0.3371292582277371,0.7672508889051485,0.20186008898266172,
 0.39746933938818185,0.3501200134696255,0.970360726133743,0.07639131685417322,
 0.3110002730980863,0.9982643016753784,0.8720218874823396,0.5979774185601003,
 0.6142372530886463,0.41352238071333003,0.8696475965503331,0.558862822749034,
 0.6275316045135684,0.4499735009759398,0.5098664608735618,0.8507484857832484,
 0.5837841647694485,0.8367992078014569,0.35680877393008503,0.7313065642251922,
 0.43983275522589094,0.04626715714576091,0.19944768877275398,
 0.48622236630894644,0.7878218189565235,0.7191408140341756,0.4512085655130371,
 0.5432967116265655,0.2340805287126937,0.9584585357092297,0.21189379543113163,
 0.7620689596790597,0.2676261151833055,0.5370832046314246,0.45146533827508106,
 0.04725719079639901,0.01281714374027465,0.9132829987843097,
 0.03387012161901215,0.18505256215440036,0.6725493250185569,
 0.6197346616947887,0.3273406919940447,0.3398533021385315,
 0.011770818473744504,0.5354775688931471,0.7246683726381526,
 0.6734335961363369,0.9783851228315581,0.5957229309978215,0.06769634048563822,
 0.40899633900072996,0.7918705696129945,0.25648130191786545,0.73371538285308,
 0.24068667135639388,0.5862478836951069,0.4522341278635226,0.9630316312443059,
 0.9497880819252039,0.37925359283454374,0.5591277093558933,
 0.004841089167055945,0.32379709533520873,0.8421399455027095,
 0.9130855678715094,0.8971609371265385,0.7397758580169075,0.47157617559463605,
 0.9676868608419519,0.5636265366379418,0.04402740374443781,
 0.41720555670698967,0.4026877612282803,0.13028676131888728,
 0.9714729445725592,0.5984944555552576,0.14987809064487112,0.707506257018159,
 0.18457626922757586,0.46400471740842764,0.9421895615390811,
 0.8174195336353222,0.748922096432737,0.16213341567234274,0.512508274401093,
 0.056248952070782954,0.6781174587647976,0.12455198500852649,
 0.6243837539631621,0.2347083041359842,0.6813591432235533,0.34663130366689077,
 0.6510903010462671,0.11130726324460705,0.03251359241480356,
 0.5857476587100753,0.8629592015956976,0.17158583350257472,0.2590199253176523,
 0.2088659957074004,0.8557267598449002,0.30076218165668456,0.9814149852599201,
 0.19525278690829673,0.17747406121458953,0.2990770945425799,
 0.2871682278255765,0.6112190851466536,0.0310904699173441,0.16890521670792147,
 0.2821877309875884,0.06401690174091346,0.6098174980189234,0.7271994236660602,
 0.3757172696290306,0.42196542410089244,0.960792466825114,0.6739201786154527,
 0.24206852088737185,0.8327841493632003,0.5468441629866232,0.7019033914578943,
 0.9052569199720293,0.9521396612170794,0.9274761129916261,0.49770371630048194,
 0.182995181577437,0.41498548834081483,0.058425730217742,0.7286372586085381,
 0.7255392811337226,0.04209021372299382,0.2282924064781987,0.3151725102195829,
 0.9336077103834514,0.5809233456891016,0.5797391682898793,0.37951064878089436,
 0.9748690129005183,0.7655081368012928,0.7578439488025438,0.41463419731386686,
 0.07247041639051499,0.14977099685895856,0.5051385623435146,
 0.28897574513088853,0.9740048301420137,0.7890767970751463,
 0.03040169617979971,0.30593149439150824,0.7607138417242495,
 0.9565530587175524,0.30707467083252316,0.2555196839400985,0.716426375278643,
 0.7329337954564716,0.2655053219490928,0.3956216379818822,0.4495554128050062,
 0.359258607111299,0.42916166454533333,0.8152536658756655,0.07854679726693403,
 0.144760269928068,0.6957231814917587,0.17798859501972242,0.33864928795427274,
 0.9755527819143187,0.5443614470589694,0.6641872943210375,0.0965495771753686,
 0.8020637040212923,0.5527977501792749,0.004029062870864131,
 0.9097852927399319,0.7359908802169246,0.6165301126632939,0.5562988416464971,
 0.3054873563350564,0.39291848089720527,0.8419621803135583,0.8809508668522453,
 0.4861885219829738,0.8981415658310209,0.7038145126194701,0.7083980508011126,
 0.8126293366498619,0.32915099754582355,0.6223572788761969,
 0.11021715360575413,0.9969735992258523,0.9583392229833149,
 0.17218650562403714,0.006795482865498026,0.05201124482803554,
 0.5867587804630247,0.6875886900543814,0.6502827539803754,0.7976383116202629,
 0.9055973565419094,0.11810520778749511,0.5577262793730398,0.543420754819884,
 0.13911205111627356,0.15172385012046852,0.6923830380192297,
 0.5180518320734468,0.2078233814246203,0.45375034146700277,
 0.40469347730335525,0.34475003523427994,0.7555050487730176,
 0.3198626956200904,0.21796185749611607,0.863145768109297,0.7779665346007636,
 0.42306432702530883,0.58148654958174,0.15672892235495872,0.4225789205525252,
 0.4479360665840022,0.6234168503390747,0.9925853374325369,0.06061020866399991,
 0.8030679886643218,0.016916672911499342,0.8251727632630956,
 0.3700572833086211,0.22679723465657198,0.2802283039348271,
 0.33360787988926355,0.2512688795727417,0.9345509535454045,0.2694469369428065,
 0.5970081381611287,0.03527550550759706,0.8026110222447059,0.3250135048744849,
 0.6490020150697449,0.07167502151198679,0.19065790794660242,
 0.9143370588787104,0.9594321201530359,0.1811493682755836,0.7512087050020874,
 0.37764266087660525,0.3568533197426572,0.399806918581579,0.9271992307297848,
 0.4662392397212418,0.6801610073265317,0.4409085758245358,0.6292443062178622,
 0.27629655085331406,0.1531677666814134,0.8361028435074416,0.8444570090841288,
 0.822522267239283,0.5833699203829577,0.4495064144115527,0.7737695264117932,
 0.31024114900634014,0.1990120515024778,0.39115311882232984,
 0.23230472906551825,0.7477718504933258,0.21019645611083893,
 0.24514748196024827,0.673371631795988,0.6586871533764396,0.318935702159968,
 0.8997724125816083,0.8363668030374422,0.3915135436410785,0.4028754259417998,
 0.31880102426136336,0.24504126665345216,0.6556842459877273,
 0.7218669654621738,0.01646410648941543,0.06435872216979732,
 0.39989855716300426,0.0031853726793715786,0.9730374917040463,
 0.8321861308300784,0.39675848741718933,0.6303466870689589,0.5333585884553802,
 0.8492827153425352,0.1287965845242206,0.07288478710481372,
 0.39350853914816486,0.4474514105379571,0.23844045071859443,
 0.5025235313895963,0.22674399892907682,0.9978651169926985,
 0.06155709993759473,0.8687082882043817,0.37655852727821615,
 0.4129757500492208,0.6641424492773249,0.8470477919680447,0.320369986246694,
 0.7665957068496378,0.6924305799756472,0.5197313561280092,0.5025752183416995,
 0.05457826312272385,0.3731408482171569,0.24528519335431664,
 0.6374985616113233,0.45874400589677244,0.7033278421590615,
 0.18066730783188445,0.24057055409157568,0.5625044106144637,
 0.9759177601500995,0.3680919870102036,0.17139779767656527,
 0.43995902721754165,0.40734814219104043,0.0556625926899027,
 0.36619864106625255,0.9494182647103547,0.8197528755898628,
 0.25172414083419115,0.5228233107095355,0.6019264219418853,0.8356375374719418,
 0.4762118432638851,0.6619540529523678,0.97705048620467,0.7520647853421076,
 0.6702256873345025,0.29927524718403364,0.8201944865037419,
 0.40773230810777084,0.9914051807450989,0.8934662955912746,
 0.12020119478860536,0.15384234733486735,0.08175389514925668,
 0.2639012681041646,0.4558492027358584,0.14429163863222405,0.7042242623620192,
 0.35574713947490955,0.2080818205734597,0.7178387308533377,0.5978378466220083,
 0.5588971375111288,0.7419122906227303,0.6498971285115157,0.19934231190206997,
 0.6928689727999385,0.503485924462102,0.4642743099747543,0.06547411798139813,
 0.6212777857905969,0.30630904541135984,0.7384212677669268,0.1037903810949552,
 0.6392122213629104,0.3659724281627146,0.5542551327172219,0.9726524774474073,
 0.6915120530284924,0.5210051017221198,0.17496713837243638,0.7579962951183844,
 0.8638799736482398,0.5966078857080426,0.9007666527794993,0.3280393025352648,
 0.7212256944014762,0.5130568542941302,0.028984625698152966,0.751479710412347,
 0.4293901688861538,0.9129970460649406,0.49244052334169197,0.5443024032266323,
 0.7523336277882879,0.4198589441286378,0.2609981848873937,0.6153464556999323,
 0.4657172357021453,0.6455637190517833,0.009998019651203172,
 0.8921187602327607,0.9237076194133076,0.669178414652355,0.5220479846251245,
 0.8011573197500962,0.2603392081678437,0.8535242142906151,0.8685245528257617,
 0.8080094111038806,0.7634101878397436,0.28858914390081714,
 0.17873352144358334,0.795369515721325,0.5483016265331684,0.6996738645226008,
 0.6419440711733327,0.027831414281348388,0.475243810457868,
 0.06372519635556251,0.9542563542448737,0.9416780238939388,0.6649925505000052,
 0.7354183450465444,0.6654243409422734,0.8299864933595931,0.531439347517904,
 0.5714972615110824,0.9388505940536388,0.25367871796406183,
 0.04502222825362434,0.4671493203845942,0.13381684586544818,
 0.34452163778305267,0.22733738639165768,0.7632493507497108,
 0.6554016895595417,0.3156993882541105,0.7843201108732399,0.5301156905575012,
 0.17946934704641815,0.7679997515395046,0.712591026428433,0.7023854273087102,
 0.8364858602545571,0.30320281195465026,0.9309887712888708,0.8810418552013977,
 0.49400439612740243,0.46971652716669166,0.005061660920153477,
 0.886068924572559,0.40793174390822895,0.21101111704109898,0.5448414065060962,
 0.615552105233184,0.7934247519624978,0.9203436582624135,0.00445620738562591,
 0.6882055701028024,0.9043397241950085,0.7466316415999333,0.47128728237194517,
 0.9629191115721888,0.1803747853104325,0.4227917056882138,0.9943470181103289,
 0.3280667220863772,0.06297848054274746,0.34424545743843116,
 0.8968546340566419,0.024186737106750344,0.8822049763148115,
 0.9906346464776968,0.9741958924821114,0.6886556042874243,0.879717964970446,
 0.056595126655186734,0.43468425631606866,0.7594560371577578,
 0.5197567987639892,0.7131939170808366,0.29546648782732254,0.551627878400216,
 0.2925829847902228,0.10145089223216686,0.285899190499344,0.8444103492868887,
 0.4725321335295254,0.2435243011328082,0.9356705809981328,0.9594190603268133,
 0.7251888142499112,0.7023876563101723,0.5545836409892926,0.6096193740851737,
 0.9088235068074171,0.6490593344666887,0.11244216676857022,0.7488759305672015,
 0.2625424247296254,0.07072181318815995,0.3569480690341542,
 0.011405486571026469,0.6798320984213401,0.6131578240352649,
 0.8231870601921099,0.9682907471623242,0.13506191908596876,0.1332811401930789,
 0.4321306837349015,0.8703222509902621,0.9064239861294725,0.9597700239632327,
 0.2977082812392914,0.12245999577215994,0.8138668501477168,0.7327260794632904,
 0.09601810899673913,0.7397674059595609,0.6361578070402722,0.0898598983158625,
 0.04966795439194083,0.2679892233313811,0.7966829123690697,0.4007528770910874,
 0.06785514092276401,0.9027203473968138,0.1414684725846308,0.4822036506485128,
 0.302546190390357,0.7780927357920273,0.34699640895056605,0.05479662540182417,
 0.5395295883396787,0.5717326927602716,0.7020125188921513,0.4317963896028052,
 0.5184417373121318,0.22934448056759638,0.22025179511418713,
 0.8147496728309868,0.4283572968501088,0.5637703465223648,0.8431220656196489,
 0.3352267566942766,0.34304772961810204,0.782217038891712,0.41015625779584264,
 0.9096514906370974,0.6860644731027514,0.4477806607338959,0.2268086894405016,
 0.6149131865986271,0.25056136766072445,0.8728043529877667,0.5466782603203227,
 0.9596467891435867,0.21672798923457348,0.6820675545275626,0.8040862673379282,
 0.35965208171877683,0.2551720664356619,0.5481870736484443,0.9252083937155905,
 0.8676251290413738,0.969750009278715,0.46513995066376224,0.3319830184828786,
 0.5374940423353777,0.8751981928866562,0.828169902207464,0.7147812873708596,
 0.7337040667944604,0.8539711901925083,0.8041817533075295,0.3309299094863154,
 0.7374636089771349,0.126944362331197,0.8272779911768533,0.9745037430673069,
 0.900030037381358,0.9848423812262216,0.8146071293131589,0.5542123192702743,
 0.5507946360460811,0.3973389100994822,0.3996065149997907,0.4320321661110871,
 0.18808230199528508,0.5580123904095448,0.6421047487733862,0.331125130918812,
 0.16416411266062858,0.2363373480737001,0.9985633736814286,0.9121215831835789,
 0.7581143842990178,0.5271083625014215,0.07978116541245861,
 0.18889978572250643,0.7777127488442716,0.19749025581815283,
 0.18981633172601675,0.6553335264527024,0.8069483966690288,0.3311270825394741,
 0.7721254037658789,0.6101246953500183,0.6807942451369465,0.01551829939991245,
 0.9146314635718199,0.19349436636628103,0.8155832561962284,0.5625661023572409,
 0.21741655536256677,0.770503878228858,0.5053021904265745,0.5543958762276392,
 0.03900267101953925,0.7526674653921555,0.5843537570526994,0.7782305957655613,
 0.11731259789475279,0.8199258121524964,0.286744668851889,0.7400210404559124,
 0.7802627814705746,0.7143690357758996,0.8263754513324653,0.12442234527605833,
 0.37441746111134855,0.8927039551527588,0.46945245216333353,
 0.3365917450332776,0.5333434215737977,0.7379929222237989,0.74877409052823,
 0.9418937916062641,0.8319476023334756,0.6947529062624154,0.21064505974852044,
 0.3800206386318974,0.3710517651442349,0.9501387753566717,0.3533946323465713,
 0.21580915893164432,0.4002824308477767,0.4586512423124326,
 0.025751796004686134,0.9698589305917237,0.25455663727095157,
 0.6599986402839654,0.9087478725312815,0.3567319090520964,0.2205202670184715,
 0.15964036324761566,0.08441814815171778,0.06004916580899422,
 0.6002942045836193,0.15242755101721062,0.05343442059456882,
 0.8168629418445549,0.7727447629850975,0.7124275227077581,0.15889764773638082,
 0.6848451901555692,0.03197953305021106,0.504645462085327,0.5807573684584988,
 0.42788664455018854,0.7967943036523958,0.09943480929001869,
 0.6043305227995374,0.34904444338413765,0.050227539713851854,
 0.46505408319628505,0.9816363090409884,0.5741204992365796,0.9735324796564888,
 0.6707535410205571,0.609232607801948,0.693591874285494,0.6556518322114315,
 0.4242760501418952,0.09177435333260853,0.18197406110224035,
 0.45643306049805965,0.7185971830388319,0.604498700770685,0.24905595480084375,
 0.07977299952837846,0.6210065285877615,0.601601785769383,0.2559228373494292,
 0.8916312929235363,0.42469104308581584,0.9098857915145016,0.9230177861997763,
 0.21831772751992828,0.5556796109872377,0.8136453160262531,0.6487254626346866,
 0.03321075299211018,0.37429604527539206,0.9861432567758806,
 0.4782756945360145,0.9077072848648969,0.34381112944574976,0.4783960896947037,
 0.060519376922664625,0.30019724373435974,0.5920007242657784,
 0.30664336485170185,0.5084004498891251,0.5380708448794239,0.5056987438371462,
 0.2674853807172324,0.12933175222344317,0.6595235222446105,
 0.23628009754153012,0.6460401987868647,0.008489454476113378,
 0.4364468542736112,0.6542405150559011,0.2506761326497525,0.29324533829502064,
 0.8421211634077612,0.12974064671998808,0.18327040425290186,
 0.3539638051628867,0.6220933210414523,0.02049956788163254,0.230042772327128,
 0.6106903713084021,0.8438742122864782,0.8410705935837832,0.3799521765157109,
 0.43467486748946416,0.16104566540075194,0.9445173954594361,
 0.1395111383879556,0.2867080893100863,0.22453975823910544,0.81616550429405,
 0.7159121068725582,0.24202441828548926,0.7468515451320596,
 0.14240539519815265,0.12423759430182313,0.5493791168725324,
 0.14669701436264493,0.8076551648234768,0.40865564785904507,0.773548635761,
 0.38653580880458993,0.2872458405893621,0.9585121874599373,
 0.30949559527036963,0.27532674366376786,0.5114354262208665,
 0.08751364122046312,0.4397765756479811,0.3808827220100135,0.4988617689984931,
 0.7349564202552247,0.172231865683794,0.19356943711093366,0.45446139652657813,
 0.9515137925727624,0.6083548210122104,0.8880163725929764,0.5195743359956706,
 0.1314160513155811,0.7026246147719271,0.9375989787200378,0.20810753650825353,
 0.7201520187312737,0.21780484298157143,0.4710803688806011,0.9028567896910379,
 0.9418221617943537,0.7905976669360972,0.3895256015884567,0.0997869026085979,
 0.6107033882806312,0.9565161707556755,0.20917707912925665,
 0.001217655820231256,0.8351196602491177,0.7001619013573228,
 0.6223946285193449,0.5491445242111673,0.09891794621037608,0.6532612328399172,
 0.7461596858504892,0.1014692114113227,0.8053433082313997,0.22126830927193486,
 0.9038751209330272,0.003998686293898546,0.438338352845126,0.6849626116099448,
 0.014195600902932481,0.700624827298226,0.9157876877384759,0.2308838159275981,
 0.05908709802451684,0.7024284046193536,0.6623591744954133,0.947257521059058,
 0.7296598589042704,0.9330607926453646,0.8936908247016676,0.3735973410720571,
 0.21278512128047133,0.9302165805616045,0.4324377888119597,
 0.43702129761869923,0.5147482917551408,0.7847492860194404,0.7951070208026186,
 0.6711790525915967,0.5208509630951933,0.6684014103555166,0.4979236310282822,
 0.9458528765712924,0.5190729892846121,0.7547106234350504,0.5769369628701583,
 0.992554428134117,0.6814547553956807,0.3487740733477137,0.1365226310997647,
 0.6420025111600399,0.17835783468341515,0.2808356301788153,0.2758056230154985,
 0.9961119664715823,0.9305475974499497,0.14910317455319066,
 0.05774448009138278,0.756890807630483,0.4384552322610371,0.2848100487267168,
 0.022756013183071078,0.7758223043540454,0.42696302459062463,
 0.1141560021628506,0.2675524300325054,0.9349945442827091,0.3370353596170472,
 0.6281172586709775,0.04781035258394972,0.926106867678888,0.8216333617036083,
 0.4722100592450884,0.6710548554975844,0.18051360603854982,0.5826196115125629,
 0.4587662189185536,0.6489156976638565,0.005537040769433843,
 0.1632476131270847,0.8696459774762735,0.592497112502309,0.0826899611283185,
 0.21053172136212472,0.6349737249889948,0.602791324091078,0.25756812980772736,
 0.5467760912294845,0.3874854634167282,0.9316889020010937,0.7737392024951928,
 0.6007968916091929,0.965536994387826,0.5060381690925706,0.49912109438057317,
 0.663513808247709,0.5873665740926264,0.835840358298483,0.9208591489912298,
 0.9428439981315926,0.9657676786097779,0.047642648246411734,
 0.11183184849662464,0.6246484298076689,0.46873256674485775,
 0.5395451413099174,0.6350073699056961,0.4264124437215303,0.7217446479896629,
 0.0777292399756917,0.46730208188603384,0.3583043503660043,0.8501331457927537,
 0.6569989489181112,0.7442456949627879,0.23762308004039812,
 0.05894827805870806,0.31436385685375,0.5653542019291885,0.4822948634472862,
 0.9645682975506041,0.28599838964742264,0.4181638786199995,0.3796029789627594,
 0.07059912159130977,0.48074573709331325,0.6638803411527671,
 0.5090327772868283,0.21726535554985138,0.8602296051450893,0.3682128995528894,
 0.7949640341315333,0.9836271388710861,0.42651957231075643,
 0.05687697027379979,0.5572780732483011,0.5096225141716074,0.7847410401517751,
 0.4937341436685707,0.04281846731041894,0.013522048176942136,
 0.9892043351848859,0.7308749935496947,0.038840482805693854,
 0.9461805290927219,0.4777989513999451,0.18433580926803406,
 0.39528225758155355,0.5260481674625441,0.3037746908934408,0.5738881076464353,
 0.7568878022716248,0.9232432037341879,0.7372470100718871,0.916752706735568,
 0.20061189933619727,0.3135594793038513,0.46462235704699306,
 0.19753686506224255,0.7208232300112306,0.6652814026661052,0.6087781905939711,
 0.06461542368923556,0.8832034682484091,0.7781359255149163,0.4410071318006701,
 0.5560728406202653,0.5440403994507453,0.4499213812166373,0.5375614439913443,
 0.05378729903216739,0.6410220232957893,0.8300760073958969,
 0.39155384383977354,0.3510810216231186,0.9967570997727335,0.6378753159065061,
 0.19587811447405679,0.28731703413889353,0.3299244912199103,
 0.44996297563445287,0.40206994780900573,0.964747271230132,0.7485832186491953,
 0.368226412829023,0.06366971155796253,0.7251030053189833,0.0637848208526045,
 0.18439906349127355,0.32364329694620175,0.23510919643365347,
 0.09908818365968419,0.8104173496242143,0.4805688851367629,0.7448780775730114,
 0.3886607066247165,0.5440339341727402,0.16792071663988273,0.7623068741242592,
 0.09558934254883289,0.7683385219658323,0.3357987954276842,
 0.19764877649475654,0.00532081897722636,0.6295042561411703,
 0.8245692002042463,0.6131837472136411,0.6750452647478008,0.7930350520712758,
 0.20115380635617042,0.6346930736936353,0.6516071984944157,0.6760843494716688,
 0.6118227905802994,0.412359014186524,0.1281104505005971,0.8121624987860918,
 0.42359907457667634,0.49198330863418605,0.530456409058029,
 0.03641829993193946,0.15156256536574642,0.6998451715679149,
 0.9784858328697724,0.9574999272276996,0.32672027247125757,0.5916844444962239,
 0.5511348707256422,0.6908790886348462,0.991845421875469,0.9364680839851726,
 0.1253032700516693,0.7387710824988424,0.7512731272997459,0.2358456498299658,
 0.2830204950473033,0.8404761134542091,0.2557640540177126,0.7893569170192808,
 0.4458520591346714,0.3797454735279461,0.6338067324799311,0.9429030766692177,
 0.9269221813089977,0.6475001494259898,0.40732424061251615,
 0.42861156540721734,0.709287311673174,0.4841025147962126,0.3657059398128315,
 0.5612390593898744,0.8338407456570088,0.41205937325736786,
 0.05028437291912069,0.18748115869476045,0.5964946526229795,
 0.1769481043037633,0.8036124080796875,0.4794804810980321,0.37731096485683613,
 0.8609443608462076,0.8010464368969481,0.2766735002051077,0.8918805099408211,
 0.497100326111498,0.7330414254821307,0.9905314584044999,0.4366819227352374,
 0.6628426759224038,0.6429504359354964,0.2918535973335352,0.30486223951294267,
 0.857877786070154,0.2540603275789851,0.09090559388675323,0.9638323919954278,
 0.532144296088621,0.29951304634998266,0.12245182850661318,0.4799055868530959,
 0.9590617799839002,0.6745402930323294,0.5836139454634888,0.8650379004141278,
 0.8884681545702812,0.1827856179895675,0.9199692134744668,0.5830864099442993,
 0.5746132338723985,0.7386384378168782,0.2457490899044812,0.15066544072267296,
 0.4316121861106341,0.6546395603947922,0.06002009734258207,0.7074660206974006,
 0.6726231667118999,0.0025222464902037167,0.6308229474399881,
 0.38511320804492066,0.6743877373509878,0.09454099001235616,
 0.2421662013055501,0.48509209429986155,0.8725326316307916,0.9694221015838462,
 0.6592413822714772,0.3144529774949555,0.5983991044113973,0.5018270701922862,
 0.19413002486357755,0.30661896634413455,0.3820016938695855,0.500513653841221,
 0.19169061936974796,0.24779907987419847,0.35384882489798397,
 0.4483031078612504,0.8523914237763143,0.08692411497848207,0.1764618880460742,
 0.16183089582403853,0.9051803534223682,0.22890244621072553,
 0.014192074881767658,0.8476998361105647,0.47621696678999326,
 0.12151721860876719,0.5524486423246262,0.6678231049460427,0.5380739770851914,
 0.19239158635078824,0.752951763174686,0.5552960923282877,0.55544903700671,
 0.030955856011209093,0.2853329905682064,0.09116874775022277,
 0.5885673559866872,0.7324726947285924,0.6885340516364976,0.9670165864430014,
 0.5919747830042019,0.4749352590469317,0.8215145033695257,
 0.055628291855171486,0.0025217506455761374,0.0916854418053098,
 0.3645069562259837,0.623381084880871,0.5255465776097129,0.0757809393099661,
 0.1318426448811465,0.49303379393045565,0.6420320988204627,
 0.006065353532857309,0.1820226579397758,0.02379072729518239,
 0.39793327838852555,0.0060775306078014735,0.17230364097973172,
 0.4714540715704846,0.7504854665656235,0.6436410937836368,0.9029302577638488,
 0.6944881077918535,0.7837321356105662,0.8887933817510764,0.6241465317913398,
 0.4812101269103243,0.8607623624233408,0.23202161893592788,0.2329907765391075,
 0.242368972282893,0.8926730389190499,0.38932980361859426,0.5874842115585829,
 0.4424096100744357,0.6388491880874937,0.4731444997829517,0.4764979915589076,
 0.8191878560906022,0.26494427192727565,0.4929921851947804,0.8567584052900269,
 0.23388454634860067,0.9234597977662222,0.5397094414086745,0.242452111888808,
 0.4823273902664609,0.34668197917292876,0.6863056618930358,0.2471086819836884,
 0.256592925254438,0.5531222157137541,0.6687329727184653,0.9750263077382713,
 0.3169452196646029,0.9899889903830582,0.58303997264171,0.7964626118646748,
 0.9645458345498801,0.05781348473261538,0.0062671283510070674,
 0.8902994215330242,0.5853211126596203,0.3664274096328958,0.9812191831897433,
 0.20349984588710945,0.8143044790763987,0.46545181343444975,
 0.003984735371321491,0.5424744363123554,0.7735626118874774,
 0.4381341590495943,0.46557870740619633,0.4943346140800672,0.5750878827383709,
 0.21773450195333732,0.47637074056137163,0.6639842894011035,
 0.18802752195508332,0.6695888204067669,0.07068345034251655,
 0.6336900211094744,0.16852844741546646,0.4545905616137679,0.704692618844384,
 0.050253223586449014,0.7415699649214725,0.7843145095951104,
 0.6134449015198423,0.4975333201099482,0.16392614579921572,0.9323069828920508,
 0.9158068727851799,0.44778235014146994,0.48373012577151464,
 0.12894637409560328,0.0035452538787404952,0.07658749671712517,
 0.6567129010042776,0.0904574142247837,0.351308734835002,0.05897709756435299,
 0.09104517791560551,0.18921829020698544,0.00310874002352457,
 0.21070061437968035,0.8427928125633875,0.328414985024043,0.43792774515838184,
 0.22411384529429446,0.5216469052560104,0.20317426588483212,
 0.6776070482326624,0.328846670802876,0.9954199714975064,0.6153260169323272,
 0.29207454043654857,0.23543695254852004,0.36517370632086565,
 0.44665020194941396,0.7481097335331061,0.04589566797394096,
 0.18071392509264128,0.853216349156814,0.10577287676284064,0.8908491848417347,
 0.033860917944061986,0.4742938291626697,0.6882511935425,0.5164155772452389,
 0.20381808182984484,0.543148602325908,0.8872184974468027,0.5353471679838613,
 0.02179439448990972,0.6322155327798384,0.6103124355332559,0.9752127203166067,
 0.8420052776716267,0.950137335360863,0.31998662861826577,0.3083864003391843,
 0.24397555673610882,0.131918551020326,0.47919630108924516,0.5031492922859925,
 0.7751733957779917,0.03233806181090215,0.33067605010420864,
 0.45473970447623413,0.5426314617199723,0.1746049915430511,0.4086418730254028,
 0.6746097102351281,0.8545050612102605,0.8994006851507482,0.37223410106899524,
 0.33009577647653354,0.4571965476593457,0.8424705391381789,0.5832002752212071,
 0.2916148588526487,0.8988061761719124,0.46628107638600336,0.5260516270053534,
 0.7563034034576347,0.9099588645380442,0.4007725128239028,0.06379336250318468,
 0.41785232888799995,0.8560636168379313,0.5735241826876881,0.3454645932311593,
 0.6703799359822086,0.20108980563309997,0.42079055197580273,
 0.7857745173673429,0.19540295253833084,0.9252849711265971,0.5775157978166642,
 0.7463821127642634,0.9058917139360886,0.6598515876306302,0.732924229661433,
 0.3073456130847617,0.9843629697056656,0.4512180303811364,0.01451351158017955,
 0.9608818688109085,0.6044526389297112,0.378287612192573,0.016711654195522452,
 0.8849458588107008,0.4927889066537079,0.18401700113540287,0.4969426549927878,
 0.1252666615789909,0.6520785754148551,0.8464247585386534,0.1788026531678527,
 0.8732985938007721,0.2752054826456174,0.7236527830509417,0.43017908302122965,
 0.49792104186737707,0.3120362633463858,0.25206106397809813,
 0.37952619165093526,0.8519224557538753,0.5382905814646092,0.2575322386569694,
 0.988405785343309,0.722573287633738,0.5756226670852891,0.3041609600872598,
 0.973601912177176,0.4881564225171815,0.7137072826093078,0.8259695075081293,
 0.8197973555699769,0.9695408727430667,0.8489224201048821,
 0.021441988973223136,0.6350678060227548,0.3795031169652492,
 0.8995640375160225,0.5055240065313237,0.6501699254858384,0.405332799413084,
 0.00776324067583134,0.10257258282528126,0.01496535427818757,
 0.30240802384959775,0.07541100548190904,0.7963450220127325,
 0.7101686571893909,0.4058926541882706,0.18200603271415305,
 0.30615856902364413,0.7329336626378719,0.6061299655493811,0.3474900390738638,
 0.0368824057941739,0.35545798346206603,0.762445065163795,0.4788961640953395,
 0.5952893843508492,0.13051527238230842,0.24767055128014825,
 0.6463931909473736,0.4608170458209644,0.417938163063026,0.10538282956061518,
 0.1535652102863524,0.8369375829022803,0.8390417142131861,0.8696828636629231,
 0.430762413926735,0.026572577083843862,0.17597141677456773,
 0.6075404107370301,0.32287025571836625,0.6543677060622294,0.4902039228713334,
 0.002432683710698713,0.16408576784689022,0.2553737253139535,
 0.8270888341483962,0.6322567218144949,0.9564859659942198,0.5071296654677186,
 0.4592198218809661,0.7379954266536002,0.8088562517146233,0.08544275958140102,
 0.03229006050691885,0.09838984230887748,0.198206576203259,0.7338682347367244,
 0.38187757450659765,0.6388523652854359,0.33207764063049583,
 0.10236590147113067,0.8580532432074123,0.07014403481467102,
 0.5663211466682129,0.5664996619660704,0.744819769191546,0.08333422621980269,
 0.4085156743602856,0.7638890306323939,0.9975570771900268,0.22792789932525892,
 0.1662312410182667,0.48341993765310565,0.2925973188631934,0.3851053514762075,
 0.8121180814069681,0.39187841918416644,0.7581920367576156,0.8623538213921502,
 0.7403885811644549,0.7933423435345689,0.20821113596976593,0.2297122417136026,
 0.44940358125634194,0.42018790645391335,0.7277013920585064,
 0.5234065635632476,0.7008312066122153,0.42083873606587896,0.1650838363625975,
 0.20607568983485391,0.23603370602434204,0.45585397053376453,
 0.3137711788613323,0.28957016608987796,0.9047517079390003,0.3589433148126563,
 0.3151200957492011,0.36513514469506914,0.22096645223345157,
 0.7106126594183512,0.8484470275595627,0.4163809934865519,0.428584775458665,
 0.6889125466205266,0.006016118658460523,0.04107647297820216,
 0.12080719491142511,0.41573453683454376,0.7865890755043435,
 0.9777234716086476,0.9750851279875419,0.8860916495800327,0.9858543825450293,
 0.6589658830393205,0.38313582859696527,0.06500807479771475,
 0.06560417243487382,0.5016257370769355,0.0728784000646836,0.6760843326208356,
 0.9080137173853594,0.8825185203287935,0.2494245382716066,0.47686780163494313,
 0.4549775106353535,0.8647293195713014,0.18107432949751623,0.6302288301189914,
 0.10061566279861012,0.42833322322945155,0.8160770041985438,
 0.2915109056232964,0.08480102309821647,0.7230849819246256,
 0.47105400673409237,0.533997984819909,0.7721773518855307,0.316808177147613,
 0.35087272556189164,0.4142688098113103,0.4616164528663388,0.6091922433716987,
 0.34945859460230766,0.36124227323725433,0.014898122721290763,
 0.35817194710738,0.46848377853269474,0.7080087505093717,0.005546133529147346,
 0.8263878330439568,0.7314300736289309,0.34024856734583364,0.9199633772963484,
 0.6589369402839038,0.6809761022914216,0.16997288995398363,0.7152013753133626,
 0.053678344911989306,0.8405432147251879,0.6904368992831191,
 0.07222439285564575,0.4375591471360005,0.6387571124071557,
 0.18863771514665795,0.1619306882013768,0.8332408930607922,0.6981382986391587,
 0.2402492493858991,0.46107626816765823,0.8705554250507377,0.8824073831782526,
 0.3748910418987911,0.3859426030020048,0.25411812187468075,
 0.03981024781881404,0.8817126969652066,0.7215074801808246,0.5694177960592163,
 0.3355521414090873,0.37541830095790885,0.1334094415054191,0.4450980720855895,
 0.33814620027299547,0.9669573894615057,0.7108820146449534,0.4147055321879253,
 0.11225341996864657,0.5422568601717033,0.020560841148194897,
 0.12165888757850751,0.6723851552209289,0.5081513333164921,0.7923048634816797,
 0.017652087915742576,0.9289071314922639,0.6577502846983145,
 0.8284433698285126,0.8033262063564528,0.20652498474698777,0.4482179690438619,
 0.08876379015489977,0.34165993359995883,0.7491804519025057,
 0.02743991526570544,0.6582008805193453,0.20250994188621008,
 0.031795024204488764,0.1636886937294455,0.14031453876361766,
 0.22156103604315425,0.3462073323752315,0.827289852870684,0.3069767650368965,
 0.9120553489830315,0.40029008078372774,0.1287330735767429,0.6043192361983072,
 0.36533435756853083,0.5931150819627292,0.5672248874815088,0.8184834783998768,
 0.13709782186953878,0.13793098473708576,0.7726183996762854,
 0.5980091627387645,0.08332532933869974,0.7416609852927701,
 0.36049605270092144,0.7388538338195397,0.7595682527903184,0.6054365836678732,
 0.9576879223600671,0.5658684051347952,0.8727491597774065,0.7676500714115017,
 0.31140264062463696,0.8889994862267147,0.4279089348491848,
 0.19776775055018792,0.2486566808639804,0.07544749562768283,0.712103565028679,
 0.8330807378768625,0.1733019940794196,0.4291541562380049,
 0.015030536239331482,0.30886826340374307,0.012923248492146544,
 0.5967303949352343,0.8389321802489763,0.3844927179065276,0.7675824944400583,
 0.7232437525637577,0.7728786599136599,0.14997245291668992,0.9086702867406168,
 0.7697400482925362,0.09814720532555177,0.4979040467306063,0.5942955952679807,
 0.6781508717271518,0.5483129602784833,0.999736719770504,0.2523493295229855,
 0.37281665963648813,0.2978471146864463,0.8808710882375537,0.9343926468089992,
 0.5714066041676079,0.5493025381777565,0.6695989018402289,0.8207136901271745,
 0.3419710356256186,0.6620952523539418,0.17195210573428932,
 0.16622288856589285,0.08108668722466417,0.19292468601469803,
 0.20187253825469775,0.6453224242533807,0.19936724161336328,
 0.6994044469726124,0.16727875067456222,0.08751076660789814,
 0.7750657109988731,0.31514263763085837,0.6842021795119582,
 0.41059130293747237,0.48410669291402686,0.7775023501309678,
 0.2352771012480953,0.6048548378721268,0.6413829025240241,0.2739973053208111,
 0.9929339823492344,0.7280657808404842,0.0669076747015318,0.852215923426671,
 0.39644739312585797,0.41652329482773476,0.5105495018495162,
 0.4016985903471253,0.5442626000354874,0.22905788515824954,0.5273590178466226,
 0.9508718121163465,0.6411884833264441,0.09325888077584865,0.7641292227356037,
 0.9665382176725941,0.5944712464534413,0.5542969213292198,0.2749697956966797,
 0.8057199580223783,0.2564553407581551,0.5028048595955319,0.9376626665070209,
 0.275630664108964,0.34086587086391085,0.9981029839602465,0.6399668614409664,
 0.7394377901753483,0.1377030938065228,0.4499697756318619,0.06183100374209227,
 0.019646339507860433,0.31316611893437163,0.06931522226676656,
 0.10613922669252573,0.8398705885994843,0.0467055156958367,0.8334932808561358,
 0.056250089679485304,0.8792070837069026,0.3291759377176833,
 0.38021941787614266,0.7204372759031652,0.892018163568153,0.36052990519019334,
 0.8959502711124764,0.6366300703375329,0.7764967485996785,0.2761329937842639,
 0.35292007190321584,0.1429069785680005,0.9191718882276171,
 0.14216268814895572,0.565987562785908,0.06481562163204246,
 0.24912896965735787,0.08845721962088615,0.39173216628582386,
 0.6325278564462263,0.11525806741331956,0.3037710020926212,
 5.624027968245127e-4,0.6249619751889847,0.08190275574159289,
 0.4458665312325746,0.7371165675407947,0.34262383813776975,
 0.30059368730870717,0.5035021901997678,0.4491480004081372,0.5672655604886853,
 0.794606918591238,0.7120046676179724,0.5809335423381513,0.04862315237629822,
 0.7466586527296346,0.9588230646674253,0.8898165017192018,0.6480152206178846,
 0.4304662110125459,0.10500388657222792,0.11847237951532852,
 0.6705839221984226,0.67890892077374,0.4225466356383053,0.29588930090695265,
 0.5396702591715743,0.3248275551381261,0.8572363934865348,0.9272231731424017,
 0.8457907501734995,0.49330708647668653,0.8143975365924802,0.3497304313390093,
 0.549562955838609,0.4229607301594992,0.05526421694736117,0.8055925748150241,
 0.7519419502035931,0.6517343054612351,0.7600318113331038,0.6866459138753191,
 0.6207720302264582,0.26805275461422406,0.7159072425048851,0.0931472705893358,
 0.8511522081136677,0.04090505549948198,0.586444265309918,0.6226245222134934,
 0.7562885358831501,0.4341253952164871,0.6383384386629029,0.9157568721221814,
 0.696696184824039,0.619587520123557,0.12712815730852067,0.06294216395325702,
 0.17658482535167974,0.7893889474581106,0.36034473492435914,
 0.17712340255252113,0.6321761149587304,0.43669063760243665,
 0.4453256198466501,0.8061055693488567,0.3232417217264716,0.4699324736261328,
 0.9372343592200092,0.4909761501593469,0.4950438115019776,0.44855697254116567,
 0.17716528091954975,0.17340746541354268,0.9747614990213058,
 0.29077215023688796,0.5648422605549304,0.7995092227630475,0.9340900007423156,
 0.12773594717339387,0.13104931119118757,0.02681151202814458,
 0.0626986271590414,0.7701898177039465,0.4864257355306406,0.4190496551046311,
 0.23104527615087345,0.7096077914486888,0.4178561098677305,0.8735172734506664,
 0.09133183099425468,0.0010637286984490624,0.9886075436635796,
 0.26540605369923365,0.741471431201905,0.8275594120416336,0.46665873230843624,
 0.058080875560765355,0.8766332135470376,0.548451271368338,0.260743548226675,
 0.87502322481365,0.5838842119876919,0.8957807044239263,0.15882129192730032,
 0.3606723978617016,0.2812659127422654,0.6943716673928021,0.4079451957800566,
 0.4943731768064611,0.07513755877538475,0.2200250449757678,0.9218350323843333,
 0.5234928879446045,0.39827579412929515,0.7105178289067524,0.4849643628669993,
 0.8607735801482331,0.7654729685438989,0.5699227464213547,0.11133138465958758,
 0.16554256050410443,0.21890405175917826,0.056335767929124314,
 0.06205666641773544,0.04227875677420523,0.759859260083502,0.6960821743838813,
 0.5819882426773029,0.7008241438036957,0.8583926014762004,0.6675062824936306,
 0.08233057688281031,0.6912221504955826,0.18141466032186626,0.783689549708735,
 0.44975120598760077,0.9968375178341484,0.6281386911585162,0.8836540750135774,
 0.8457458060399108,0.8189549753354749,0.16040698643597895,0.892278777227135,
 0.6617668487486448,0.7449362083209791,0.5622871946299799,0.9414147855169963,
 0.024116483475455808,0.8172375862698775,0.26795713977527846,
 0.2790354646757871,0.19954309022547778,0.8541681161018535,
 0.14217117373600396,0.8082857191134183,0.9088539549767399,0.6366153328287132,
 0.009878075564870858,0.05398685746388332,0.2879599090642676,
 0.6877494675592358,0.2901801332026839,0.09758656015134659,0.6094820709634206,
 0.02493366837172173,0.4870640370177075,0.5618400423680467,0.6598460243028003,
 0.2331752485013453,0.3063908742816933,0.17101675711873332,0.561220352177274,
 0.5986688222661262,0.18129316275131147,0.4580971906509501,
 0.13337073443578218,0.9384771679412038,0.31589535908432786,
 0.9224418161694931,0.4835037296191892,0.2183737590269812,0.3741997568007386,
 0.4110481727614772,0.7803360321802857,0.8698336609984818,0.22952340508468272,
 0.1113440122247088,0.8069216351323787,0.13358809746520506,0.5760813329404855,
 0.8902214021148468,0.3508504239729652,0.6423169627389772,0.14153000952461658,
 0.36243200076032145,0.4861046092743674,0.9060595832422716,0.9617657953718378,
 0.13629473941158554,0.10264649659073322,0.3375712231268668,
 0.5489467544954673,0.7347935555954148,0.24446327022183267,0.7680171516493068,
 0.5083366598260762,0.7289059249713469,0.11014876605371238,0.7849813318684519,
 0.34729096700031636,0.43753377509310487,0.9090582338794452,0.499182914897662,
 0.4633023304985302,0.30068875394965566,0.5170057495706428,0.9449502172535194,
 0.7598767628284889,0.34983229902506807,0.009997718285103763,
 0.3018529500266769,0.6696329644161181,0.02561859884366835,
 0.12178270081304243,0.7150504032372104,0.934112348827901,0.9787874683357768,
 0.41668818624739734,0.6075714957450337,0.9732723977212892,0.6548554877324015,
 0.9584464036581483,0.8007416536354406,0.9585706074039255,0.47960853236273693,
 0.16484211649609715,0.9430898518474802,0.22923134026648373,
 0.25962149378420263,0.27182317439433235,0.8025033232840579,
 0.6090298378942988,0.16578008488109497,0.31374712793508497,
 0.6826327761145675,0.659588718819067,0.3777234862181532,0.030871681100331894,
 0.4143875973502744,0.9629802880529156,0.8143358017567378,0.18646036295943835,
 0.8241112189394558,0.017945529552189865,0.8432756630221251,
 0.36871895251047604,0.5265351095961784,0.3016899671501171,
 0.026322046403923438,0.6421480505756415,0.9955773596555856,0.382468622901072,
 0.4803204391365675,0.6597939647221177,0.4853556819577327,0.08018941479536357,
 0.45622119644567916,0.647304127775017,0.5773531166967913,0.2432111632861188,
 0.3898224008655191,0.9195664418897498,0.37466841449960675,0.732538246715235,
 0.34399149866874357,0.4043222718142464,0.19214996998590295,
 0.0657036348536062,0.9363654546537981,0.29897156639398403,0.630455410770178,
 0.31401939238522736,0.47996384981351414,0.8703738167935582,
 0.9590480504764378,0.1909002715665995,0.0790658882139974,0.6537416128176414,
 0.5284458043650212,0.5445848214647436,0.22275393463381266,
 0.17803067935741845,0.5308401013430293,0.9890484384253402,0.6828431124349978,
 0.05455220177643372,0.9165383600454426,0.09963526174798332,
 0.2674421592645747,0.2907865802999485,0.46234089102415343,0.4306500502069035,
 0.9967793140133405,0.0980991577757353,0.37740311244925506,
 0.12439326569954279,0.07261656145822903,0.21897845882678268,
 0.36517440763249076,0.9513829424761283,0.9618447068908232,0.6379478239836274,
 0.07235304513495944,0.8492753900231997,0.20284886772055855,
 0.8510609625857506,0.4543390683214169,0.9728067594340349,0.4569361937149322,
 0.48825600721481544,0.057408941914074085,0.9227111353943993,
 0.8597781628881708,0.44923402605021634,0.9957843334680662,0.7751331579135718,
 0.23213297460730498,0.47601298554132154,0.8831738913823753,
 0.5411720140008615,0.7143143092730254,0.6829819744937584,0.865844400812616,
 0.9079948791088431,0.8069078277971568,0.15612166281772077,0.9171525040429158,
 0.7439057206494081,0.3814295593579976,0.017517632401660488,
 0.4495326039825779,0.7159424234931294,0.6481367031899906,0.2262726585136008,
 0.874899680762051,0.09639809249518061,0.8295793803839414,0.2504262338921758,
 0.05064434480644886,0.1365894769969046,0.21103577120745964,
 0.3020333601645595,0.8757334705063905,0.681341382563113,0.7087845289315989,
 0.3608925117757743,0.9278504151446211,0.295403583652151,0.6899748280668161,
 0.17217923588182915,0.30563578655380697,0.4167309990096252,
 0.6847313547655346,0.5793421299567645,0.5674993503507063,0.38275003642329153,
 0.4772248694712268,0.38900202886234536,0.9707995365133686,0.6748480359053132,
 0.06306299100998536,0.9931151348139771,0.28599154626380785,
 0.9984347091504006,0.881090245424499,0.0904312932034852,0.27504491361875305,
 0.40535549161085893,0.9335911944924611,0.24229315051151112,
 0.30054492025091795,0.8922923330137145,0.5588948128076293,0.7269690404208999,
 0.9808599428114999,0.7919970117605708,0.2283830252047272,0.5430400101115767,
 0.24973120095364998,0.4725803609171696,0.6264355724447052,0.2563027601518667,
 4.8024078763808653e-4,0.02204796795568542,0.8471721233544512,
 0.583724495445992,0.6318434220500928,0.3322248274118661,0.8671645079129273,
 0.6648331673790081,0.7820054902494475,0.8237050078672039,0.8367559016609879,
 0.1064157301247165,0.9155745269759514,0.5716541202034091,0.4046219241550343,
 0.1708502649125817,0.9564112322282968,0.8971896385659621,0.3601045869694405,
 0.40654998060693315,0.9112299930448508,0.10986142867775084,
 0.9605256750610355,0.5540853025090554,0.31171335892009056,0.9845634598589488,
 0.3989841117123054,0.24848806333833484,0.21973524059512528,
 0.7897023792827474,0.4203901848535859,0.09330376100033366,0.2391843357270701,
 0.06409286299475148,0.4472853786765486,0.4508018582375045,0.859856985769104,
 0.13972997736921988,0.7900975092018383,0.5720669440068457,0.8584210157566392,
 0.5626588539244346,0.7326026927183308,0.23199395041389614,0.3120859804865075,
 0.9251143660088913,0.8876935742909458,0.723649945433719,0.8489526802762795,
 0.7641254794441779,0.17483043262634657,0.7763782848940646,0.1832244127945738,
 0.1712312961602408,0.8975822505779147,0.5006694390004991,0.28520876283773544,
 0.15029707273067283,0.7910624645823132,0.5470667115740969,
 0.04052216848305179,0.2811829743172076,0.8122381393261329,0.5650211386374141,
 0.4839109866208825,0.9282157718837585,0.6739207210140414,0.9856348287399571,
 0.32530450122108345,0.994761596354205,0.1268028163500572,0.5246129097872818,
 0.8230989113592514,0.6845190606293551,0.47007035471187486,0.914009018197226,
 0.7630065572302512,0.44452400072866183,0.22685425780237511,
 0.5551287760531476,0.8737900844710225,0.6884662332211715,0.33358422773191443,
 0.41042583627665274,0.6769103623784071,0.1599385046540991,
 0.27540681332851247,0.06808130539454493,0.09665593327513855,
 0.9260312645602776,0.9155274419926517,0.7562621685315079,0.8430253516324246,
 0.8861457099113594,0.39640362231173965,0.8621335846249896,
 0.12079860116712848,0.24776513973969794,0.7512509754162362,
 0.2880671212283259,0.5224172802294841,0.7726222120283505,0.28832968376066515,
 0.8567159956174323,0.19446131282835677,0.5026800945932998,0.6567709692482939,
 0.19974873116583935,0.8124550071430162,0.7830336873951814,0.7442900133574133,
 0.765022160258184,0.6996283530254171,0.2641907396348857,0.6705634216432459,
 0.11383631543229167,0.319769073976397,0.7273423281149517,0.01152391751639903,
 0.17532752483037983,0.7888517518193021,0.8474941883899169,
 0.30702789183105395,0.8842167746676663,0.14291323397796873,
 0.7662065239002357,0.13757986240164766,0.8291316661560441,0.6005841063111035,
 0.6389999565835481,0.3981412593349529,0.670601867345246,0.6152881348203834,
 0.49023599753365654,0.8596539630898289,0.35690752598978026,
 0.26486636545446485,0.9439265180720212,0.5795175204360101,
 0.010585924871805985,0.1770483487100193,0.7938308926797438,
 0.39077037184044494,0.8551745501962268,0.05196685713936666,
 0.09074676632545386,0.7378824564299511,0.11419765058003206,
 0.27001565664250227,0.31284885692526965,0.9316872943426326,
 0.6914490763047542,0.6511374333705511,0.8364519623350675,0.8015137711886577,
 0.6407574642145222,0.6073958096189678,0.0343293082429611,0.8770455756790531,
 0.6478036331206088,0.7860248874440326,0.6754928885918284,0.06378234442549802,
 0.7519803030956667,0.9538742686720081,0.40277519360506875,0.2482859677035918,
 0.39165955536800634,0.8386681603475842,0.21357406441023186,
 0.9638178336660123,0.21548012773734881,0.13499831669808915,
 0.5765586875531309,0.7534030106238659,0.37715001461563036,
 0.40576779617174574,0.6122881984571915,0.04759493238754564,
 0.009623046258641299,0.833111352401009,0.8847573282803676,0.7093279442106337,
 0.45925529117381925,0.6653706672198784,0.21170725436053572,
 0.11831770766805805,0.4353712078778672,0.23391990141214214,
 0.9266810379119917,0.2548615905786351,0.11153901053934512,0.8605333352028239,
 0.6534629887900776,0.14021364074714238,0.438438510964552,0.20195880044810066,
 0.4637660166306542,0.12281480544409085,0.537083201170839,0.11542191259851431,
 0.31139106221816504,0.179210036029667,0.6168805231161264,0.5905529576335523,
 0.45240931561421627,0.7749539763976953,0.6335865570901034,
 0.031982579742056205,0.23278119240934725,0.9905257385819991,
 0.37784128370069214,0.1388745692628388,0.13323692711268353,
 0.9356230455315229,0.6057971916339565,0.9448374580602258,0.4300841330500872,
 0.7594575604811814,0.9944774748457557,0.4038719834984168,0.7096506580339922,
 0.8362834250390772,0.4726160515055007,0.8910630652448306,0.71560507345012,
 0.4506193675615049,0.616054454677056,0.11358390228839621,0.9286045503038209,
 0.6896423644945627,0.06472875421898683,0.3942759241848467,0.8610693916129231,
 0.021538252819220283,0.5403708999283152,0.1177616914477092,
 0.1049298772718027,0.07922214095222335,0.9533131700561127,0.6465560734951525,
 0.20102445754943488,0.4624630105662225,0.7184212326583637,0.2959256832462013,
 0.6313036731409204,0.12894407302138688,0.7965470091939897,0.8930943562255015,
 0.7825310174671488,0.3931718455326758,0.8798434082487021,0.05095114032560066,
 0.3243818127051973,0.3460723728990429,0.1848329345806885,0.8627454630960895,
 0.895301723073814,0.9556617663188546,0.8959746046359718,0.6611072962714009,
 0.791874980696847,0.8195976296623302,0.9562313464130527,0.830987750180952,
 0.026772377682673376,0.8998910646993816,0.8638872260901459,
 0.25409213773832795,0.5226426398868895,0.10523109771443129,
 0.023582684260897047,0.0017774945815217658,0.9937812422027416,
 0.36971162044014994,0.23589696613744426,0.8353082011867161,
 0.8607329861485732,0.768079915090077,0.9404765013073714,0.7932854189840882,
 0.49108371220801994,0.984603381536538,0.8327867531604449,0.2410001157188829,
 0.8410767833948944,0.4100813601730261,0.0949779921852345,0.3963843250360871,
 0.24033233262206166,0.5439409794918466,0.20237434719462,0.3183204413182379,
 0.7858443884495018,0.38464594950657305,0.9119785985253697,
 0.34308131852685064,0.728248094673742,0.6894979781383661,0.557531440320947,
 0.1514127030516329,0.14376015077742432,0.4473610832310527,
 0.031277478833343086,0.5500658467845245,0.8437766808740311,
 0.8853080125099344,0.45120716871413713,0.007104622397350657,
 0.6663474054171302,0.8599205191185132,0.10592920748398427,
 0.006679661478634857,0.8679925487041474,0.42548280137476013,
 0.5566697205627726,0.48850123644687904,0.34183148008181174,
 0.45444544369049755,0.48193348512362644,0.41007213006226806,
 0.8456035141645022,0.4788224376856761,0.9292238386126926,0.4489262960215037,
 0.19233887587813803,0.6331785330252248,0.5668665249763654,0.6250851987785262,
 0.09954448271531358,0.3903802153366164,0.3677840583124091,0.405627395099311,
 0.027611348698307392,0.32218003059251055,0.5246180261435519,
 0.28037035918185027,0.9718364749546794,0.31052484057710794,
 0.5393275912516453,0.5200608127885107,0.060238440356674805,
 0.35711365202152656,0.20609324273350027,0.3389553383553794,
 0.5737769519379017,0.5380145474062646,0.27487569468803974,0.7245479841262682,
 0.245973770802429,0.7601135983009765,6.112201064726319e-4,0.1757395112739658,
 0.23966543333660195,0.2072619230445737,0.5495429699787584,0.8260746298897824,
 0.48232851086640005,0.36138254764288147,0.08378848781625892,
 0.6551695605740788,0.6936584911809208,0.1979042734963804,0.47003069630218464,
 0.09292259176619411,0.8124155272430414,0.8731800816332487,0.8241485830911507,
 0.014006248907478258,0.16367880200050622,0.8152804558470635,
 0.6961728247407315,0.18291624381680638,0.8219628055245688,0.5161887101091216,
 0.9425812338300472,0.38473789224030464,0.9107201776461106,0.7490330838708721,
 0.7882434366054143,0.5562504536628998,0.8030178196365025,0.7877857942298455,
 0.8151145517501494,0.8233190630591486,0.03836102249312623,0.9687711006031376,
 0.33353232944140065,0.3489526997928477,0.9945985313266882,
 0.26811796134125343,0.8419112292183163,0.8131737205268953,0.563963611888167,
 0.5636588949809977,0.2538453314061705,0.5776041116064953,0.6618288465838426,
 0.40887362328463017,0.8300499612507721,0.6410491005088516,
 0.06461416621680727,0.3485057013388696,0.008040274984369544,
 0.3856017339434923,0.7901244193412558,0.7911565227839448,0.3529873844424387,
 0.49101346371614796,0.3815045948371436,0.48155954606997176,
 0.11147298411566331,0.6982498176534816,0.06215678447700168,
 0.12488120856077745,0.5093409795793753,0.9157837929165396,0.6044381411630005,
 0.1599676748486102,0.10429544505952948,0.7416074992110248,
 0.27629295139642096,0.9170834475759215,0.04916678076218006,
 0.9738687439232843,0.0738788572081509,0.6880910190577842,0.33423577890678635,
 0.6667714164190901,0.6228165459269108,0.2596526378563433,0.5304733787780777,
 0.6648319439429012,0.6158150064801254,0.7993521545691066,0.7797041830590238,
 0.023508915224139115,0.2564799557525953,0.36854104353948924,
 0.7261325308908879,0.9145927563514079,0.6803079422385564,0.25284755141126125,
 0.5803055606992717,0.7933121343675436,0.4307834803778776,0.4946794224154196,
 0.08283970783353922,0.5156946856441805,0.15592290968300115,
 0.7313284258153605,0.005124796232436224,0.8945211392919629,
 0.4778947485614393,0.5231019846638604,0.48084219938059725,0.8302807875372759,
 0.0426438453127731,0.6936773415056788,0.35030771522911797,0.7194387923889622,
 0.49810146749218553,0.5071324846032866,0.7813178978340518,0.7362998282442289,
 0.7843551186355734,0.2409985126852474,0.6198815475058017,0.8592962332919173,
 0.531942706025658,0.9929141896169557,0.8493737885918821,0.5862551897382049,
 0.9944179583327815,0.411202552873823,0.3610535707909073,0.6886691185733507,
 0.836673035361775,0.8725599892213483,0.5201063641313203,0.6026712483845789,
 0.7025011189341583,0.21026561227182838,0.9846741629477191,0.5905818672619132,
 0.6028773468556561,0.3877605665498548,0.8957150890973121,0.5756350287976235,
 0.35852286108619136,0.017115736569257023,0.5707394436076749,
 0.39565219081630143,0.3008047019336294,0.9421281542591898,0.3184133358699057,
 0.5084984922283764,0.1538651562500507,0.41096260883322655,
 0.41165618147255634,0.269015363575419,0.9559710146697349,0.9427735829590116,
 0.6945636408572561,0.3895896957383179,0.2295117029638869,
 0.005174338873430995,0.8311521236969248,0.3133911338906402,0.339610660503171,
 0.2659510453785253,0.11756983054642467,0.4130418638782434,
 0.38930197454011106,0.13907355095297724,0.6414403657801647,
 0.1052918058541985,0.6336308521816675,0.04294663368533114,0.6909128464766157,
 0.7376743615220165,0.6842578879209633,0.6869423259870842,0.19872649028355194,
 0.4878558993175478,0.12707784223078045,0.2779165766125846,
 0.11581088080711066,0.507743992060637,0.8439193067603796,0.05237442114381097,
 0.9124231811218643,0.34150186701159224,0.7513370594895012,0.475513164472643,
 0.183112871512784,0.43833774198821573,0.8951035278348117,0.6376133068369956,
 0.28810397993226,0.1954244149321852,0.7508159783589169,0.6455616044270269,
 0.429801590257502,0.9987183493554906,0.8090553548609991,0.6018481813964113,
 0.8253355569250755,0.564693791061386,0.23033765825797348,0.8680331805449477,
 0.36742958027037154,0.12919846627038145,0.7194497422744929,
 0.4863953989615204,0.978663327790798,0.63062950335629,0.8537660758984614,
 0.22179893144010998,0.0017741041805792168,0.7286766905407637,
 0.919538001645684,0.9518519367576149,0.8328027564353566,0.6521308292592457,
 0.6345359865625957,0.8900999672982233,0.33146198364086776,0.9540321418969644,
 0.42929107715128456,0.023245804627683775,0.44589975996472875,
 0.3013140478474797,0.5131306290024353,0.9912869257349441,0.827797773015067,
 0.2508435690428428,0.2197092823982031,0.42110404443489224,
 0.18646450302993445,0.5748550040039291,0.196633276172114,0.33291704484036433,
 0.0910517134841542,0.8314070375925623,0.22871015428443742,0.9840671034240769,
 0.04564542093729751,0.7785620667651594,0.8119171251349743,0.2930207501492649,
 0.6482269303278944,0.25893951507444446,0.231117408488557,0.27600174387391585,
 0.20988034657580856,0.7298907460996829,0.4100245142303447,
 0.04925997499964008,0.6330227659837028,0.5255744304766166,
 0.012064024681839758,0.9096362203105048,0.3222073463867138,
 0.3691229300785237,0.5880039205766411,0.7961242367390213,0.15308443184123377,
 0.5426719715562288,0.06622798780468642,0.12943199749093814,0.365091333427239,
 0.02756122765707325,0.46837737386243805,0.7441960252339391,
 0.6847016485688054,0.35029264732257737,0.5894750884901749,0.6673358986670792,
 0.43235103866151603,0.8600723454538315,0.1441653042772829,0.5783828885427721,
 0.9357015286163051,0.8944949166581855,0.7050059846318998,0.4376431253436447,
 0.15456147906018647,0.013876975963488025,0.08360853825993364,
 0.09839387416795875,0.9251011829534793,0.1092291824175825,0.786609293275297,
 0.7908157771533113,0.24985523340372506,0.2911937991671012,0.3002836344143758,
 0.12123257289154654,0.04939276374350132,0.46683926339414417,
 0.8854476078762444,0.7005309002684769,0.9967133224374075,0.6238980054787429,
 0.8885684017977354,0.7107928857405039,0.10315590641700934,0.522691475943449,
 0.7747596591354505,0.3837001797689621,0.05568231048995709,
 0.27083172199119965,0.4551193395785492,0.012849324281862695,
 0.23293841450935138,0.45100015270055016,0.07013899143589908,
 0.7479348293515574,0.5783642598782059,0.022528780671173254,
 0.9984275251364396,0.43612035440140884,0.31398213770258177,
 0.20968685203529125,0.5855873641918842,0.6995179433234646,
 0.26616347032947907,0.517077182605779,0.7418908066595027,0.23864366253723823,
 0.20196120892946157,0.2186245764799165,0.9272919279156859,0.9803843944274603,
 0.8404034746584831,0.4378133004021787,0.9297987599467818,0.8464251819688554,
 0.3660516658059523,0.7086062358072545,0.7853742902605916,0.8808897041258672,
 0.5047841611816789,0.7392127499521779,0.907276089628697,0.42473744293377436,
 0.9852373143102566,0.41436671859626506,0.05834930048211284,
 0.6782193026388315,0.8220667790272176,0.3310269339732662,0.47548227899807805,
 0.7994388606709464,0.7564647135360266,0.708439495087463,0.13456829841166218,
 0.6098247593552122,0.7971013036519948,0.5207014858210751,0.5330297884706188,
 0.6838389499333539,0.34656321084942776,0.2063004960146253,0.639940120915183,
 0.07614551187489849,0.7889588069746811,0.7752722540082051,0.8532894517668125,
 0.3270166645201732,0.6737727901856544,0.46415202085620777,0.9032701200447041,
 0.7598556934238785,0.03889239421446722,0.44174834376570593,
 0.4059623294933401,0.8339670885678381,0.935208808576081,0.818978481101004,
 0.9700225488559282,0.9624296191270617,0.04391895280447855,0.3998056199797182,
 0.08323857302883042,0.8654702000767562,0.5563253185553675,
 0.12664255072827835,0.13458236293911074,0.39070318752216737,
 0.679315809073411,0.580726999884937,0.43611914971418764,0.8179708256891853,
 0.9281581485366172,0.7389342335186082,0.5498136963874826,0.35680662724163903,
 0.836633712724894,0.8858707134999253,0.07287195270109825,0.5017240107716102,
 0.3812561696972234,0.17316120193049922,0.9450489042114992,0.640301242973539,
 0.5316274338956536,0.5114136360874635,0.19181130915591793,0.2249804108552469,
 0.5477603529118463,0.11007076024935813,0.43602068661846083,
 0.7559252976900689,0.3434348283910841,0.8551391915804065,0.0602972806800649,
 0.9977323444660964,0.2787013725541019,0.3659547221370435,0.521142473440237,
 0.319939843060707,0.005425577426512063,0.5087083831981163,
 0.032185622348050114,0.7249152951118274,0.5349099839889522,0.462190258129505,
 0.7012099284594896,0.424694008008794,0.6731669255015024,0.5324943057339887,
 0.5179399527653564,0.9458875024604292,0.8781501523818853,0.38569685442232193,
 0.2237230797090206,0.5107321668128383,0.24506731083200195,0.4814505832822712,
 0.8615587611662231,0.14001950953294817,0.9409303459967258,
 0.40768482953913543,0.5979530951603914,0.47898176753035904,
 0.9344168097784936,0.8936475789601728,0.7154781166488858,0.1933180752142496,
 0.4381081554463019,0.9596928063786226,0.8612107489034361,0.40627885748757553,
 0.4383358167794392,0.5092935813010018,0.11561386439632726,0.3659322773631595,
 0.8296677037675069,0.6579244325160669,0.42886204059985933,
 0.12015340100872796,0.4052596896969245,0.6496388718245317,0.5414452142484308,
 0.12315484685094269,0.9707503933336806,0.17873868570701978,0.989567562507359,
 0.9376548684755268,0.6023199525637685,0.32431068505226934,0.8070672381024098,
 0.2497702942511708,0.8041233686090574,0.385638854959241,0.7548068642904668,
 0.5769605120611866,0.8658266790836883,0.48419585760030515,0.7783241435958383,
 0.17434813902196167,0.002305905030437483,0.9852831683409157,
 0.6958153615865769,0.3798661009457117,0.6828222194757617,0.5943679441311858,
 0.4314871423059362,0.4937381135958493,0.6829656832203117,0.06012649507721046,
 0.5296786334829092,0.2586437139962341,0.39296799153954887,
 0.22487517291764636,0.11794166684389995,0.24272206723977163,
 0.32659522844921274,0.5209715940603173,0.9369182381326052,0.9974332473377242,
 0.3035995396782317,0.48497856714943277,0.9350028963522237,0.3510199693481517,
 0.8786109359548369,0.9804634559884838,0.46592653997090827,0.9217526133021053,
 0.785775727904264,0.5442751156763176,0.47353965469732895,0.7894556281037372,
 0.8491074661770622,0.6247037422898774,0.2974262140691055,0.38261814798800353,
 0.07706170875691476,0.39804142488480687,0.3309791920077705,
 0.49091306888988684,0.8745124046264688,0.13396467362811293,
 0.11984164575379319,0.4770023814820421,0.17239249224040543,
 0.21314032703491126,0.470423337981684,0.3229439341386978,0.3111788413292892,
 0.5951718283063437,0.33141650430527025,0.7893179124091004,
 0.36072018612903645,0.3477906673585488,0.5921644931997411,0.6786803860845623,
 0.29914186075549953,0.800895693751757,0.5365413144198495,0.6862667595139292,
 0.7355510191660624,0.08664398968999554,9.706467585852696e-4,
 0.2804281305375308,0.5536114546748153,0.13113677547064873,0.518830893106066,
 0.7550159520567017,0.9130654161876349,0.003186159212580586,
 0.1692745861025543,0.6442098574104574,0.6465818135520214,0.9162557546151568,
 0.043258896559916216,0.9593649239498283,0.6619508572209813,
 0.0278181719037498,0.7114306769429951,0.31426129740422726,0.5960945644443656,
 0.2704916459280038,0.029141279538933862,0.0883480724796255,
 0.9164531116615163,0.8187153849292041,0.953707377129431,0.6082364506031719,
 0.40383948583671647,0.2679446030769106,0.2983357450583819,0.5631934245154938,
 0.7417897506554486,0.16529523168335214,0.2802676917938589,0.896930441318273,
 0.8728339809167796,0.4096070201244604,0.8751585452926681,0.9686683915512857,
 0.582838904948535,0.47420612073001245,0.47719653191785294,
 0.25763756167980234,0.46632561242462356,0.2947825764662425,
 0.24007605277804078,0.8186334394333837,0.7330783203160005,0.862743990813609,
 0.1561378039359741,0.032029174571554186,0.5639576956282536,
 0.8511282948929089,0.5446761165244659,0.7677193673432029,0.1289685460018789,
 0.39708024338273673,0.44197458598292316,0.35708756707311107,
 0.9601254308538835,0.29159239085268274,0.9006080114322215,0.3196906349184203,
 0.8454071088103318,0.7527296402023619,0.7362188479775356,0.5298880577353201,
 0.04488738930876812,0.29395294308370756,0.7755741888216241,
 0.17665000886007676,0.9822566490658508,0.7919849421224485,
 0.29065410582501294,0.37965653775790364,0.2798576049168413,
 0.17514210884337478,0.6586305332818534,0.5546902021571342,0.6298321545439892,
 0.386425140736653,0.9112876821139412,0.14950440313921431,0.8695572286443857,
 0.1895644687014626,0.6368947527847975,0.6563641035742191,0.9535353382506788,
 0.2645545027500509,0.05834542610547766,0.2513445828309042,
 0.032657839114208165,0.20268524329704107,0.8093278651626141,
 0.4719253472722865,0.2600752868737466,0.0083062144905236,0.380388879890987,
 0.7006951228824683,0.1392111293838686,0.3545412928123778,0.4984029904827838,
 0.08841043040119956,0.7735534922354624,0.8406130662772376,0.5558741706629737,
 0.35417549221643274,0.08359707047250331,0.309255028745803,
 0.30972107530262716,0.5017509542631955,0.046922718035907085,
 0.5756733083480484,0.7632831672651278,0.32973790083155763,
 0.36125821627369825,0.8515967970514391,0.47324835395582665,
 0.26458081832894764,0.6355986356549449,0.3853045102840287,0.4857835380610869,
 0.740349790042413,0.7436379707892089,0.21427811569537702,0.6329320762314475,
 0.9792853845847913,0.14959726827368613,0.7142129042409058,0.3302638895663391,
 0.4819960320060108,0.34860237416405704,0.34320175906822437,
 0.07460640583889766,0.1757831647527236,0.8452167908704243,
 0.25072624034656243,0.7091607791201696,0.5701602196240667,
 0.20021226421127436,0.814678148780768,0.3842161559318401,0.6778281947514657,
 0.9990226848806416,0.9963808621762609,0.4074649811385641,0.8933715115322224,
 0.3713433918535749,0.14077855335298728,0.2723075639509418,0.2646109812715757,
 0.23208550832121655,0.6835873991169223,0.17031964857601578,
 0.9630065028792528,0.8406399551375748,0.9005877986069263,0.5376082738509681,
 0.5978908724393056,0.797609062290134,0.42737850886497053,0.12727912862184076,
 0.4199957229372787,0.8908145469298994,0.7227596824959357,0.89351814146738,
 0.8963776165772952,0.9460052225132674,0.5613314178521097,0.6753178080696494,
 0.5748273791905936,0.2694833456236795,0.8306554088060784,0.8578578144007444,
 0.4888588569300638,0.7983276204146005,0.8594740004838386,0.0993211104805749,
 0.7679142330314164,0.768752778579226,0.5679835911952023,0.07911853435855742,
 0.7266939705245018,0.7580511140530591,0.7226050995830847,0.9274565433000063,
 0.2694832771665945,0.8911414391654837,0.7618735343783567,0.0341747984707802,
 0.7369228350166489,0.17216097797007734,0.8831834429317906,0.5171067865263214,
 0.48245077170925255,0.684019037942976,0.16209845252575072,
 0.38544676040283576,0.5613857096876249,0.23214471470682607,0.921146474363016,
 0.30034926394913297,0.32932611161362924,0.49344426571983313,
 0.2673433957200947,0.2824227515090091,0.6437963536017667,0.7871083385063797,
 0.35228057296140236,0.8818058537471212,0.1179173832883369,0.8988941822745764,
 0.24428472018300607,0.7582319610673465,0.5549624193352821,0.4641446218650863,
 0.9299077745477051,0.02456308350127423,0.2699696414903503,
 0.15561570933053193,0.8442473259899592,0.8205436125610024,0.4980363528723756,
 0.7373325294694064,0.22796150771914936,0.8901755543599829,0.1886387439276579,
 0.7366159761507013,0.7302266006383399,0.1089283525897462,0.03507157385364956,
 0.1990370649529255,0.21903069567798616,0.7438744931404241,0.6357275731129488,
 0.5259749156042642,0.6105435295442635,0.929371905547839,0.2637636517157347,
 0.3018460628470053,0.5906531138688917,0.3121779135245413,0.32583758225122483,
 0.9883944868225029,0.8773094591350374,0.22946690233617462,0.3141249938186428,
 0.2868411124718392,0.9047379346066999,0.4473187217471367,0.2711108358961185,
 0.48130215451716374,0.2232470585865448,0.2183945154970529,0.9780589873362848,
 0.06786778689224959,0.5278780074944006,0.6613384932732289,
 0.25469813091819893,0.9234322779021502,0.9017368548974051,0.5584169272199737,
 0.44907091389213805,0.6893695863307662,0.13960675768027164,
 0.3178392736055691,0.5349953362749579,0.3216014875810901,0.1390740758935588,
 0.11895604824800854,0.3920319192156684,0.06456954236481194,
 0.14857093809154798,0.5639406956181976,0.5510774170074625,0.7739341551795659,
 0.4602434002689507,0.04280231827825953,0.7312672930739145,
 0.16273661561967712,0.4691530549040601,0.5468488625445457,
 0.20009340386518026,0.7260700039214565,0.12357455274180484,
 0.21650815337493134,0.5515633364330973,0.3969970621037381,
 0.29956041793974064,0.977127577357732,0.3886895620792752,0.9905767376000492,
 0.615552434730646,0.3811070188232033,0.8185920313010954,0.8020543797941151,
 0.12421188566324615,0.2941764953907061,0.7820876705538641,0.7460552426334223,
 0.6056922271425685,0.6595572093496542,0.07363787705144953,0.4223177999311073,
 0.49206583756228106,0.1944567291985274,0.7236699722078599,
 0.47293815384554927,0.3621347793751233,0.05897156744811882,0.425889994409127,
 0.5044590832012885,0.8161762227747551,0.905097838707942,0.31985281177081026,
 0.4374808155198192,0.71737704654868,0.42644188856071363,0.21590686376301815,
 0.3628000444033914,0.31761007906034067,0.39188771672219225,
 0.7056237253414489,0.7717493733725871,0.4457298821271014,0.11213649677793691,
 0.6603498682504527,0.7481176529599046,0.11948985876116547,0.7624959056728067,
 0.8038666068503146,0.6312988377092772,0.878497703426317,0.3588274957615547,
 0.697947456427164,0.010082255053836198,0.17557454430436858,
 0.04044760997210006,0.6612099733008849,0.9295617088655217,
 0.22768739892694845,0.02540259266888334,0.6425762123024611,
 0.39752721369954835,0.9453467103950761,0.1806743323078588,0.5684396813185497,
 0.10086837708578589,0.42514869202745165,0.9775888050720392,
 0.2379657261404432,0.6984343740920637,0.5782665654732763,0.06799285256415022,
 0.7829815339629211,0.03702644962448998,0.32479986913169545,
 0.6928067167002495,0.07729729173016697,0.699453318128259,0.7674179153779257,
 0.18757156708352918,0.6682577635441113,0.1737816891138112,0.3389264623429564,
 0.26377550242184744,0.3866264547931937,0.37611700795813485,
 0.5321536053119607,0.5969929883961455,0.6088188174744594,0.8196859014951995,
 0.8798581040217535,0.9789637170067738,0.5564540339303531,0.5536865119664798,
 0.5523955921667146,0.3531371822684567,0.24606220205484108,0.778833882450185,
 0.6197483830468933,0.2456686453473772,0.9513028273515336,0.6770065409641104,
 0.10002577835320703,0.13046497174080562,0.33760732592171316,
 0.9548182340973278,0.6565644789574102,0.44300153517421037,0.6971291590998595,
 0.6343043656468348,0.6036134677848747,0.9296841409225606,0.8058532153525546,
 0.5032040753788914,0.5732701876433923,0.13526257081349902,0.2132914324096521,
 0.45869718181918206,0.8806327792782865,0.8908456526486316,0.7621460903890127,
 0.39982637725155945,0.9194271019763415,0.4043429297226556,
 0.37116883358697217,0.4273418516531605,0.10738827007321328,
 0.8119865402936955,0.7496290462339802,0.6301712730348098,0.22133715814450083,
 0.06761737417660696,0.0767974725606263,0.7066884648945033,0.9328161126135671,
 0.8790592591827358,0.8968676433750771,0.8985165927008483,0.6273491296261592,
 0.41544712754556945,0.7999437014889672,0.8662374752062048,0.6426257636779648,
 0.4804585786871135,0.05360214990393275,0.3463405771662469,
 0.045968713521041704,0.8985244097075162,0.23447251575403705,
 0.4740179210223261,0.39198175325907125,0.3282892001981551,
 0.45296223506844246,0.8141905959897859,0.7869064969951028,0.9045894983239716,
 0.37293893498240605,0.2537012250851082,0.652144092279989,0.9872316232359997,
 0.38969609335547073,0.05026019766333267,0.2741393055663067,
 0.8370825913213507,0.7064261569353166,0.4546165274005304,0.2109809655059075,
 0.21886251096431386,0.8885239033526531,0.3586396277661569,0.6908456057074733,
 0.247487484550312,0.19201187976146605,0.563533132297662,0.6838687872072575,
 0.23780476407890783,0.7381156758134065,0.1473112489863947,0.2951424433908445,
 0.8328144025780965,0.7103268753946895,0.7639494787070605,
 0.009510343282011058,0.07494738371069132,0.9337241714140246,
 0.30007497623098867,0.35899344733141425,0.017918990822279324,
 0.839078972739329,0.6495759460887315,0.22170405037789598,0.841963858104422,
 0.23972168588763,0.691442665553787,0.4388834532984037,0.6680329781715116,
 0.9004501909885653,0.6688030774293795,0.5750133240879414,0.7416749258351601,
 0.21179785517233396,0.9182262493780046,0.2271765836893862,
 0.16288942905726056,0.9740167654652585,0.8449071193367987,
 0.45305516873253526,0.040984791845972146,0.8845244447437062,
 0.9263591318046961,0.927596157804119,0.30301419139258734,0.7310728206848208,
 0.09833278231221287,0.09519538225232027,0.7926978684911545,
 0.48533684221272066,0.4701236443998874,0.9350840950270458,
 0.017228103028439357,0.9999331190301685,0.48444223346104565,
 0.6276276428296061,0.7965102015898815,0.09023938366566198,0.9402933258534482,
 0.6645047795016265,0.6362086383763667,0.9213772527944971,0.05245392757978928,
 0.4252564713783128,0.47469965660403257,0.20133492461547708,
 0.4166748429214625,0.5487262112089074,0.06616554938181807,0.1804777915714414,
 0.9712077376529824,0.5130787072041736,0.6449108992042878,0.9927154671228797,
 0.6388115728451287,0.5420662195200782,0.808196789094402,0.9048487888516024,
 0.3734897314393957,0.5393775472906397,0.3054073263356747,0.6574218562390456,
 0.09554821963261284,0.8575246802131838,0.9194829895661372,0.4759355885693307,
 0.9081289511391686,0.15907367225106583,0.9677382086307553,0.4660626297528131,
 0.709932241734655,0.9537376839844284,0.3092565862722425,0.4634265588966203,
 0.3587989862197929,0.2759230953527593,0.7008016854096653,0.4863811160097009,
 0.8108966624675173,0.4766703914969532,0.032698765941477026,
 0.6645745100445992,0.6331918923395847,0.33123525553210065,0.539667213014938,
 0.1070890565136492,0.058077690110220126,0.9722539900851561,
 0.6997420111299626,0.7575742002188237,0.39341669563899495,0.7623804195757278,
 0.8472395694063521,0.2922264484812709,0.15554037855986091,0.4224188983975171,
 0.34066324592163744,0.08758612113240216,0.9129709142853404,
 0.4094716551018802,0.4507507104920562,0.7521381314065063,0.5645364502862792,
 0.5741231100872324,0.35558184483082744,0.8252555413376312,0.7049172257830232,
 0.912408827918545,0.13843080161974686,0.9221433203605224,0.29132362292972114,
 0.22997022226113373,0.9032489479079165,0.7981086627264206,
 0.16352022619976025,0.3050488257426657,0.01723687506054916,
 0.8467848154040835,0.2711627902492174,0.23718402088305213,0.5413998428486444,
 0.34937075504309256,0.5721614042587494,0.6943378474565663,
 0.47771501110135284,0.0080065510017574,0.558971241403866,0.39316500737609694,
 0.8220638095205525,0.6557200976223765,0.7454363588035103,
 0.022049762972310205,0.8216894708267237,0.7987513195789632,
 0.10431369314321648,0.8651801824425269,0.3479167282333494,
 0.06914918352296229,0.6067606104087734,0.8426395145819704,0.8413500302137942,
 0.8531238132197126,0.6071645526057863,0.02141126415363781,0.7388423054016988,
 0.23041586035607187,0.8049798755242232,0.3363696889265466,
 0.17896214019760825,0.8716284032280104,6.252123119734332e-4,
 0.6344732145264266,0.33093822091672176,0.6768616103175726,0.9819845683986893,
 0.3724404050546498,0.4042610261719698,0.444095786206378,0.66939026222202,
 0.40168454576455825,0.6116559962538277,0.421659079420154,0.9941126578196523,
 5.965609545199868e-4,0.6994382672184579,0.7496800765894137,
 0.6354652638579672,0.20493715995386996,0.6940081264909173,0.5238406668872072,
 0.37939258563176415,0.2940659778573038,0.31248834474549836,
 0.36591166094678584,0.48805794007556735,0.04800444542796278,
 0.39015680955832965,0.2510475529804236,0.4626668823088196,0.477813985170117,
 0.014249034138999583,0.05520496824533261,0.27954060644979395,
 0.7598885772829109,0.9427943747696546,0.29749947062839377,
 0.11377951181980284,0.3605478721004324,0.4960969608477861,0.5352140137656611,
 0.024000998260228723,0.9844134527669242,0.6510354921728276,
 0.4640608407705522,0.10864132428526574,0.8494423131398905,
 0.06352614116989341,0.7390551973139935,0.17549585061328188,
 0.2506067743629785,0.016220252077564612,0.12191911656191323,
 0.9823972498627609,0.486349183045547,0.922134294239348,0.5095200248875951,
 0.6488126568771759,0.7729551592157164,0.5098610203593759,0.03861560813961451,
 0.7458543997023896,0.4361409060343233,0.09235511678082808,0.3488045921777152,
 0.6385324096174417,0.9798813042723369,0.5939433957178539,0.13522555212581444,
 0.9361798423159057,0.5610652021614879,0.7443788028880262,0.4390606337125863,
 0.4690439723198103,0.2995073481748354,0.9729413415547971,0.4895438870469351,
 0.2599556128923115,0.11268964268396853,0.45804021428775776,
 0.7027385730569905,0.8798815526147841,0.895654056062144,0.5275175495432906,
 0.7051596008556653,0.7541777865472806,0.9449882820731864,0.7589636183322637,
 0.13647661562762547,0.22856439876621482,0.2681174407281185,
 0.30239500333842584,0.21949859321563037,0.26084033070279344,
 0.7645084156294162,0.13708091316868254,0.05043060174549585,
 0.7647663878159756,0.5209469977034071,0.6337944631537031,0.5054291699813819,
 0.8310408304042224,0.763450376074491,0.19571090828430104,0.8667539764660878,
 0.33572131449454456,0.8848429570983605,0.4048602441588728,
 0.49914544004219275,0.5562556170591233,0.5496039454711912,0.7727999457858615,
 0.8495561519296503,0.4670315597596838,0.42091834199349487,
 0.08117706346726561,0.8055650179029539,0.14944910955494672,
 0.6601180196095893,0.45179118039984534,0.42404023789564416,
 0.46096250988894827,0.16266171683926856,0.14485509284989817,
 0.5095968023586711,8.793792680724666e-4,0.26640857789691075,
 0.15842566546408232,0.753358102463471,0.06862619998065167,0.4019570374738123,
 0.535947997493478,0.8364547168693683,0.7612506743661882,0.52086915954032,
 0.37718921128635097,0.4429184370166289,0.4169785052193644,0.5842234915041926,
 0.9670570560643579,0.35016383740701496,0.13843158575971548,0.626341282878782,
 0.0012209108058665041,0.9452156500501228,0.6926562697397165,
 0.189186879895479,0.7717286282005101,0.11775356367951095,0.12101042940835105,
 0.6180402965436717,0.9405276949324033,0.6344521859347253,0.9260728653661658,
 0.3193076225108965,0.4321994112217259,0.9103282058241507,0.4252986332472666,
 0.32400105326370043,0.9218856466311396,0.9678374185500797,
 0.32520577955796803,0.3472098486826689,0.8371823168327446,0.1310005200048946,
 0.07327154731231011,0.619189049923038,0.8826302398788428,0.7841199583977568,
 0.7606785677764033,0.7654684426739579,0.9581652474069505,0.2939954443204832,
 0.26678305103160593,0.6509076314563041,0.6075995839577967,0.2450323776560901,
 0.1401929766596578,0.9361702745648353,0.09761970190550207,
 0.11473330665557024,0.3981591216744147,0.6233392861493193,0.1784261938858015,
 0.3104580226421618,0.4863419911644654,0.825917148773598,0.4261130984038878,
 0.8123148016766607,0.6251426381984595,0.017039954407427116,
 0.2494967256373073,0.5802918480481317,0.8331078057434659,0.9476856445890602,
 0.2279856126818801,0.07618304760837324,0.4790236153783912,
 0.25840868676750595,0.9248181637178498,0.01896976027460795,
 0.6287749402925137,0.7482811047932475,0.15103825677199112,0.3124981315754177,
 0.1833115505535845,0.4678297527753429,0.5834620272535022,0.48624976829461397,
 0.7105045982847847,0.5573040365357849,0.741685258253956,0.7522301420981586,
 0.13988804731580773,0.751941906853514,0.29086067383544845,0.3654427131947382,
 0.7110612697696665,0.23822681062301176,0.40606213463596463,
 0.7759512384827917,0.8773472880899437,0.3863922082363871,0.3309579167146742,
 0.9638282610835744,0.06511523020632726,0.29554844388003376,
 0.01892326313725945,0.8296645725197482,0.5898102611666753,
 0.29859816647272863,0.0020170413494540185,0.4754434379520306,
 0.3355483034772231,0.5755994485736242,0.05593689602612184,0.8536460856990017,
 0.3090838699020192,0.356178044217607,0.343624595778402,0.905983874119086,
 0.298098919281672,0.7637107677847943,0.031681835633508015,
 0.11202242352315916,0.8568060125354687,0.11841914839317491,
 0.8471963874750185,0.7310906704421715,0.7474655237842923,0.2874825361862644,
 0.9428415516405243,0.18906194502954554,0.8559856339684107,0.6963147812714723,
 0.24263762010937617,0.2472544160017195,0.7215210646475342,0.315702474244534,
 0.296188655619508,0.02201244574285166,0.436793364516598,0.2586018042397813,
 0.07533975900198975,0.17051660419014092,0.9665087388561688,
 0.12182440393395111,0.5031324277254654,0.009437374987389613,
 0.9470860525580862,0.10824064068150152,0.8085957015586578,
 0.011284096268075472,0.5229244499708754,0.49938599178079546,
 0.4160090386271955,0.2145051207059615,0.23941442707176763,0.7812040474448716,
 0.7728484733760308,0.9185160758593712,0.14843021829387015,0.7541844735326223,
 0.942913504188001,0.7584084791078167,0.9343005185582087,0.5778204527364378,
 0.9889195426872341,0.11000398276574652,0.952289997024777,0.8224281778436313,
 0.2790319813189982,0.9383040995609363,0.7458619481364425,0.3138861614898406,
 0.37094162157510346,0.44532991886967344,0.4226406437407516,
 0.7737891904585882,0.17670504273696508,0.7297932255742928,0.9285890480304172,
 0.5666572080518125,0.6837199043060295,0.41978642353319306,0.3567892226648435,
 0.031007068120564307,0.27607997027237585,0.021835784602125985,
 0.2734535255017996,0.30379202797031923,0.06755879081839089,
 0.8654562152179903,0.6485686807433892,0.4540654150216037,0.7957223404861866,
 0.8725154576308031,0.7981252619138244,0.9955199886386403,0.24561765037713268,
 0.7187940288988867,0.19538382331619086,0.32766346236665256,
 0.9387995203236814,0.5811882288530689,0.8051966645698767,0.6935898230342332,
 0.6623336876972918,0.05425300037042269,0.7581478949185109,
 0.21134151911429133,0.6375134027871602,0.1971214904559666,0.588069493588193,
 0.9455381043973521,0.2472085210895879,0.5570468141186882,0.4652377477250803,
 0.9068494028982799,0.6844243384726646,0.13953143508527943,0.5867389942166076,
 0.5894643990698407,0.7451988294118601,0.45784755972888036,
 0.010347707330076505,0.5253399979531876,0.8050081291290107,
 0.8625995969086006,0.5717091635695203,0.3851393177979464,0.8941339133754569,
 0.8382501079820588,0.9516304571765694,0.7830986182153276,0.6459019661933192,
 0.6060394626806634,0.8414736940393385,0.978295935474351,0.6938020904775131,
 0.31333543169975253,0.46680596899512594,0.9879787957380858,
 0.5111178062777509,0.3740355782181135,0.46446483311799247,0.7249275312760681,
 0.5043066202948291,0.7894853365635013,0.9623536104629787,0.8883010941319549,
 0.028301986698201165,0.3071291442263123,0.7347100569006006,
 0.9952955025410476,0.7604662797729813,0.9455194649863966,0.9695241273018914,
 0.4790217052161261,0.802170949175149,0.37665617106076055,0.19189450526974916,
 0.3876635206466915,0.3057438557677139,0.13167542682748246,
 0.03896693465444556,0.7162816701055617,0.0542007660343532,0.6491042619465174,
 0.4020042244693174,0.729658092539696,0.21989028021668477,0.8118497401738196,
 0.011023361113194574,0.4740872866720952,0.697691859273833,0.9452792573640918,
 0.14918974058884016,0.7042001734072453,0.43371671805952894,
 0.6633132251216425,0.9372368198170324,0.49827584527248947,0.7005153388909717,
 0.2583351251557139,0.52972245241596,0.36208441252038237,0.25247757189756903,
 0.27472379516708545,0.2514867308354177,0.20745910364693587,
 0.5626004253895369,0.36120539181671196,0.3117347508318905,
 0.21726284533903317,0.025641154810789546,0.0031037700930713896,
 0.7416411415062669,0.2731062233117606,0.80762400143281,0.4268168245142179,
 0.8412445659261113,0.49851868538477795,0.2964185987654855,0.5017153395842143,
 0.4030764980084177,0.2078510039748287,0.282389027507411,0.9154649879421047,
 0.8460060483692357,0.8764210539608166,0.33280530685221876,0.3247981641470695,
 0.8016718480652784,0.015608793649473318,0.5782378268994298,
 0.4456052099756441,0.6820708161678537,0.042531251097772116,
 0.9740409096452893,0.5379288435786928,0.1113027772868499,0.9367241685025544,
 0.722507513137991,0.4500988140850458,0.2646900982147251,0.8772635116057738,
 0.4078772209120102,0.10026162957387519,0.6688904126731576,0.3219382826967373,
 0.7740551980803101,0.2081504175156743,0.06151236252495351,
 0.39522725828836114,0.7241012512111772,0.6012042889795642,0.3500867839449352,
 0.00965201576650121,0.8763434114623371,0.9023154758757713,
 0.28249283270067904,0.2067572620334832,0.7407391218854353,
 0.36550478360157235,0.4381855995556243,0.7372763060286377,
 0.15999470973281416,0.034899175881652034,0.7978344769148965,
 0.6767426448460194,0.43777341568318595,0.8502887865077716,0.9351822237250215,
 0.41319627740052434,0.1456221540341216,0.6148078166694262,0.5116712261121298,
 0.8145558895311759,0.22969516071421536,0.41348513998409475,
 0.7462344149388038,0.9807534362743441,0.43860241756833795,0.4674200222984234,
 0.9449164181940143,0.8141506820647431,0.36419455931281774,0.7034434175534598,
 0.5267278268647991,0.3426438065998244,0.23234234148842436,
 0.14799051480566716,0.733090241648678,0.7350028720115226,0.9545708916037979,
 0.8676743294387619,0.2261990775650895,0.6791106282246013,0.7452228290845867,
 0.6366731743041196,0.41682179871747693,0.8656222332833844,0.4728889927922191,
 0.4202343748957382,0.9170511897599801,0.048428717437635105,
 0.5636314035494285,0.3288394686916616,0.9299230452928074,0.660465579624876,
 0.9392283798674612,0.6472793996395823,0.9464673177509462,0.7989289940490355,
 0.37234139291093193,0.1708243764815328,0.5684997664750465,0.8123049185885011,
 0.3987421005849098,0.9908705334306469,0.6469117751764086,0.5239370028070697,
 0.5751608493232754,0.13168484418790272,0.7122399989016577,
 0.07909781955726503,0.4229041009202441,0.022404032982668687,
 0.8632572418377236,0.16309996168462382,0.7179735373002643,0.3945511381256219,
 0.46925376065898694,0.8678593389899588,0.6295237705323116,
 0.18472764945727405,0.3429487425360933,0.41036902088888283,
 0.6677175011413883,0.9247525442268438,0.851274742185647,0.24393060551208878,
 0.6161201946145767,0.03062417474267798,0.041978628643799976,
 0.9277081872154163,0.9136029222140756,0.7896201793505164,0.9241497710378218,
 0.24773345808099978,0.6141962474910405,0.1528675791325136,
 0.14367057844562292,0.2218358973482737,0.9306530514956568,
 0.17520171609783808,0.6007752491873546,0.34158613314493147,
 0.34925540237693375,0.8064097456086801,0.5298126375131876,0.9516303031664108,
 0.2450197509496107,0.5192819827295392,0.5300165240100961,0.12944481400426844,
 0.0073502912966421086,0.3385197363427068,0.5644914916701119,
 0.6695755178738784,0.22641934831372956,0.9298196125094634,0.9418446300392882,
 0.7589107742639335,0.4394713609271248,0.6367062900078551,0.5357100362384933,
 0.9649280520334184,0.4126346037656381,0.05822409597050837,0.4446895078123807,
 0.5008836716893144,0.9695820714100546,0.9716344668250839,0.7719190161368109,
 0.2346268253953161,0.46819591591823173,0.16540221759218765,
 0.2953356725597691,0.17986992599755702,0.9029603595542759,0.1763124031254386,
 0.7872428166682673,0.44827494435615545,0.37709233663263053,
 0.7909091715827634,0.4355207891264934,0.9244665465191881,0.9937116251745541,
 0.45121835356769424,0.3414754891014008,0.017783988297983333,
 0.183747776653844,0.59581972187899,0.8881110423564926,0.5934383972353743,
 0.755132298881909,0.5139722028366704,0.9689316476455125,0.1796876812617092,
 0.7793322335438734,0.26490604509558746,0.2549625193302697,
 0.09065348555869868,0.7467342840056181,0.6611985920766171,0.8319765240966673,
 0.7549724531462457,0.3636576193355817,0.5651776433648091,0.6446566762205974,
 0.3085182408590078,0.10531181864577477,0.48573155043047056,
 0.38826393382116553,0.7949317161526714,0.09613749287015327,
 0.6742912437933766,0.9980783753562841,0.4629762050943934,0.38628903119604086,
 0.35905859078068547,0.7459482048581118,0.06659721745097613,
 0.1956335421317842,0.3270369197578249,0.4033250279503754,0.3137402245722435,
 0.4164297921997868,0.9887764150204001,0.028807369403167638,
 0.25121448211636244,0.3478353572627483,0.9771441472562072,0.3028102716106632,
 0.36070986952480766,0.865995201424907,0.6534650054715749,0.6421916956235134,
 0.2145302636991211,0.8224657803520344,0.6679775224363141,0.5104738698569897,
 0.347580167212833,0.8999511994292271,0.8301081590439784,0.13711858939443144,
 0.06404318215405846,0.43725578379596075,0.9254572925609421,
 0.9251371611839226,0.9185426484522937,0.38608385030331493,
 0.07683300954094896,0.918395298299996,0.7012914671020409,0.1113349257634878,
 0.34234384102784354,0.3387451446927764,0.8374575610487089,
 0.29964597102601265,0.7391850469986363,0.05880812182517969,
 0.10653633684226116,0.6808449704230642,0.318160687887037,0.8290473475306634,
 0.30701036919774494,0.35281014956305123,0.7167100251035949,0.291497132202574,
 0.13360906737132983,0.1338697092732879,0.9142004578649179,0.3290990967477134,
 0.3519343581262292,0.08380798049407001,0.5125492013432469,0.5434842137144458,
 0.9066844408424224,0.43161155684507035,0.7649196645684748,0.5630642034842563,
 0.3083482157081111,0.9828534533783949,0.7000039650945652,0.6405087404633937,
 0.23847230714604362,0.5417734282365438,0.1182885725754117,0.6735656235777268,
 0.3186756946427932,0.9138442057903872,0.07311291426246769,0.7019131105275497,
 0.03519064854741716,0.6381643623386902,0.5557888243054685,0.6293446819334931,
 0.12151201011799928,0.2378051646012007,0.15521852980504935,
 0.43865926048963055,0.5903054589171297,0.33464448739623465,
 0.29955203604987757,0.4693886488263772,0.08629760939998743,
 0.8954145774220872,0.4221691440356715,0.4789344293577861,0.2579240794203983,
 0.5287024044828714,0.9242756969860408,0.6800946393791069,0.03850214814230879,
 0.07408289418692382,0.33973848921932204,0.4942079245583354,
 0.5442489888558409,0.7856642125977554,0.7727014444769046,0.8228705815068798,
 0.32270641481119144,0.3158178808746931,0.15103973785011982,
 0.9484853433456414,0.5152411266258381,0.8372703216585902,0.4510519949097347,
 0.6861390999932858,0.6304536714151808,0.7813553651494252,0.43507958874223707,
 0.24789154399510205,0.8561063323151011,0.4567661959065499,0.5159873013519048,
 0.447095211837383,0.23904565211362305,0.7219157179963105,0.43907580780733313,
 0.6433070086299094,0.9600708795546105,0.536435598361161,0.6712112852819433,
 0.6665706860222673,0.968398373670433,0.9922962870289109,0.17047414746240375,
 0.6137358326587535,0.24426661170698882,0.4060928233511363,0.735786071030619,
 0.8358155650933923,0.33962715459968984,0.9699186422336947,
 0.032193588689654495,0.8199874544625126,0.7213642917261847,
 0.16517023687983468,0.16298896673571384,0.8195248404851048,
 0.7427265186213775,0.571938378749983,0.3942142900605796,0.3725946421261508,
 0.2525800089453831,0.9584299764544767,0.4600348084135658,0.3241877976749309,
 0.8766627404918645,0.2017253187806608,0.19810977374054028,
 0.16886415619499484,0.3197695604540295,0.2972538200034308,0.8444931213060645,
 0.07284292120768976,0.2702444080972053,0.5190410904274354,
 0.18826938709816043,0.48448813078219355,0.9279377977543267,
 0.5670786801768432,0.5484921375315719,0.37468214534846733,
 0.21767024605848728,0.42593367200034193,0.1325527478735069,
 0.8472460322665246,0.7445502982392384,0.8434281487148019,0.9107891072815584,
 0.3858501918547945,0.5516722696995418,0.5488366371930424,0.42171619129672366,
 0.676884431561186,0.35090831552302526,0.2916649057380333,0.9297379160401001,
 0.4309339361705936,0.7418404603972839,0.8737154550829599,0.22025079936299863,
 0.9493696582146383,0.0948747464663362,0.351618645710138,0.8818890889998614,
 0.8575253285206189,0.2142704011924812,0.6739160661090325,0.5393474751411471,
 0.4534562409727767,0.6993483635451851,0.3995851567895995,0.5941327974072798,
 0.25875224435060085,0.6968859224040814,0.45136506098857,0.13957156818018257,
 0.7475852228803527,0.8904025370231116,0.5103720035589377,0.6675641521603447,
 0.19879818909545843,0.15276159421836477,0.39196884362387086,
 0.8829699370850912,0.1013932302039443,0.19491431771057288,0.8734944193396866,
 0.9413612702466112,0.8505542309920254,0.5556118068884823,0.37241713892540484,
 0.12842725286062018,0.667534246730794,0.5312742221370583,0.25927966731134633,
 0.06339880368789952,0.3023547825318933,0.46217227591522536,
 0.48663296098178443,0.16139515823358552,0.6307328457903157,
 0.9820921647910126,0.020299982410948103,0.14325074608787003,
 0.8277575856081478,0.8886448966023175,0.7343682479409069,0.2582634684985753,
 0.7343257178210699,0.8335382658589011,0.7906419939793827,0.7535111847779115,
 0.4641467503013509,0.6710936381648196,0.7861885044487943,0.1296960873315237,
 0.7364603098049435,0.9187847857384298,0.18327595335919744,0.1005480755751672,
 0.6791434242561484,0.428321570560234,0.9522683230047727,0.5044850539302611,
 0.1980042558156443,0.21027606916243058,0.0994447611918331,
 0.07081811972744256,0.2585708906160946,0.2940766158941832,0.5599056011784675,
 0.4583147216501262,0.05386440933280934,0.17027572182569772,
 0.024399674941275595,0.6998000271662793,0.23743870468954498,
 0.05140988782441991,0.02435721566075566,0.693251009320121,
 0.09582587926847552,0.8670568948509654,0.9782107232520508,
 0.24163256937939093,0.27375395687525506,0.4735880802132221,
 0.12016932281240567,0.6379492702200159,0.7397174304400008,0.5707413741669447,
 0.6661037371948401,0.993043955485016,0.7563761910850444,0.37023386450816886,
 0.37629523370765394,0.6688819194169957,0.4917104994001922,
 0.49409650944240524,0.9765963500116661,0.9743510904755237,
 0.024641769332827024,0.7438309224059269,0.7730240611601421,
 0.6190714591122473,0.5963822234836939,0.2150423996275086,0.4700671884247998,
 0.23805345340943052,0.769513208779058,0.5355770142950017,0.21288827602501315,
 0.8814668918384654,0.7997128751721904,0.5245151960770094,0.2917806033482029,
 0.7656338028692269,0.21465336839730886,0.04491231721317046,
 0.5872252883002174,0.4201834626384022,0.7603983264191292,0.7469172526992455,
 0.28850448512816795,0.5939907219021929,0.2470672893588035,0.884264521907803,
 0.5850835119839345,0.45248885425009056,0.27512875285882,0.9879832655826825,
 0.829834006102278,0.09795648935434087,0.025227659664178614,
 0.31713551673793583,0.20676820727653344,0.9399669312492864,0.650187445614964,
 0.760639236567424,0.26590171908709126,0.37158152789461507,0.6100293809572289,
 0.4359804777460229,0.4046581272314136,0.3303331091117574,0.32073320811121986,
 0.42729086183440423,0.4650206664968761,0.01302485205648507,
 0.6282293381474443,0.9554795378151998,0.98897689081147,0.9314977020660815,
 0.353080685937035,0.45804600833294085,0.01945852938534931,
 0.09347717812250467,0.16259750443272236,0.6543119892711272,
 0.7048461815092728,0.0015663133685616115,0.6494090462776045,
 0.7651483579330018,0.8428674232729614,0.7757421922381155,0.2834419795467523,
 0.987744516590665,0.11571931875943386,0.3376158772538168,0.9828182639990662,
 0.3177866546517907,0.0944417840123859,0.1243794967369638,0.5497826504740997,
 0.8944979008478781,0.4044324568652907,0.8043380362257669,0.1699419756662306,
 0.5322639475738893,0.35061898208930575,0.5699094585165456,0.3274604682085447,
 0.28540724011920793,0.9180774161899372,0.8026710285410988,0.7567936700281332,
 0.5513279212825546,0.7800081875918835,0.7737906006345581,0.04015515401110914,
 0.9089026343699237,0.026044104294287163,0.03790361146430821,
 0.5966050680612317,0.9758619628258528,0.7014658989105583,0.3924338236138061,
 0.9939904841928383,0.9724810342357288,0.3012535667806202,0.9103175193228105,
 0.7884405000927566,0.7692865587138084,0.9136384770441125,0.7726162624988923,
 0.5865189468715193,0.5673099520082795,0.9447620840857693,0.40237939554744206,
 0.6829306466082237,0.012755610734732525,0.9909957140232181,
 0.6391388463837145,0.324153823066354,0.7710756302868477,0.4024305869089342,
 0.5720791481931667,0.5949660512736323,0.7747548627113738,
 0.051806558398421965,0.15463029215169566,0.7213888602804702,
 0.2700704284595272,0.44925686963661116,0.2906684937483406,0.665062262668234,
 0.7081289209053768,0.5703206544541358,0.2937276519857279,0.277108209456698,
 0.09414104407180313,0.22220303931084706,0.47150350762829296,
 0.8384890572891138,0.6907288542431022,0.6745725453113852,0.43283811304795383,
 0.26609545041212357,0.5360637650576572,0.5398150101317866,
 0.25646955265240323,0.679581292928662,0.9831920423898859,0.04116934897588209,
 0.9954741174438486,0.7887010094739318,0.6418411553188421,0.29619010736380746,
 0.7751058366291713,0.5017481279652608,0.8703293502150088,0.9275618086459989,
 0.09333981725598184,0.0662298039242547,0.7673699552313097,
 0.36011932485026776,0.4231492205964409,0.9177883146800452,
 0.42446108531920457,0.35281793763989866,0.9259069840145052,0.992325325004967,
 0.3298125186110632,0.33372221500045973,0.9055322855210116,
 0.41213908165162516,0.2889177791657038,0.7459960714898994,0.8176401832225964,
 0.25529839017951494,0.661439548661296,0.45690659137712497,0.7989902601476443,
 0.32180777411151573,0.37149991829195617,0.06972506192713113,
 0.6832353663478754,0.0768192661109044,0.9358697635437941,0.09191231224721108,
 0.8397585132521035,0.8079864652408744,0.9329455427069543,0.768653096264743,
 0.8071284824534248,0.9571617904060703,0.36133688523339624,
 0.035312451946488355,0.5715377636367964,0.49791223616495783,
 0.36010472991506315,0.759287793866313,0.8747577757645568,
 0.012465280175426141,0.8513385534318438,0.7907737119345456,
 0.8399949621805262,0.0997039173569515,0.722153668211981,0.41949859402170686,
 0.8459962691804549,0.16261136982759528,0.8014417154893418,0.5679213364169875,
 0.20777461816649723,0.9891189797159018,0.41317532639318577,
 0.5769740672418956,0.14221542896692285,0.12240683378610762,
 0.1380390093958408,0.2424798056554318,0.9974611568695879,0.40945797537862094,
 0.7715900185913674,0.39881293924393124,0.9855198166450546,
 0.49792389781160307,0.6657634496045365,0.8134563845683397,0.6813060631243598,
 0.4783476126934607,0.6635632686213576,0.005569935637218082,
 0.29477022283754106,0.8709079015386746,0.576203833648117,0.7067154185540694,
 0.5288516085884043,0.7397177919331336,0.40015007144725945,0.0987566707043851,
 0.18073963993230735,0.40179212495857497,0.6474518244492905,
 0.8969671758207229,0.4982327110862118,0.1304590392405024,0.866941718393324,
 0.5969248835987833,0.2495224108565709,0.07170285056061987,
 0.15466632134977232,0.4782161054584959,0.6201096075821886,0.9052960037237556,
 0.4384804835567835,0.8960470633170783,0.8662932282123353,0.7788745305279431,
 0.34461328557258586,0.16307409167993714,0.2953424736947544,
 0.4160760130149749,0.4143538223853157,0.1602208507121753,0.17506475610117533,
 0.6169918180849332,0.46241445043227025,0.0526176619593266,0.7658119177796523,
 0.9814345162333413,0.700459901976676,0.9547661341425906,0.8566346695276065,
 0.035766076492783516,0.29194826911273986,0.6631680199118002,
 0.5435075469599197,0.692751889091227,0.11756028109446748,0.6294971601463065,
 0.8003255631522731,0.2811841360294831,0.7589633268474737,0.05608340321772087,
 0.44965838319526197,0.804549120364175,0.11608370230377474,0.29231553755923,
 0.5404051958795577,0.5648014053234587,0.22855000042051443,
 0.21444497221670367,0.2263503712816527,0.48942930569005505,
 0.5961647368423538,0.8319391013508881,0.34755269020949797,
 0.21034678536125018,0.14899020883007152,0.7789398948384214,
 0.44902056899803966,0.5241225506872956,0.5192144980330068,
 0.22132396308513558,0.9804544732285184,0.8327818970993007,
 0.33745096957032383,0.9586933851127885,0.8061833356310987,
 0.23184044837895412,0.3792707896956037,0.704458025820415,0.292298196885595,
 0.024958521770060416,0.668632341008348,0.9367575043771729,0.6721167248873902,
 0.7654237598453528,0.6047430825171606,0.056682358485995565,
 0.9704644403837951,0.5108417261560939,0.1771530181470352,0.2653979485306842,
 0.16432075860174788,0.0825114541326436,0.9254646219060225,0.8420042197347362,
 0.2914457531633913,0.8237660590592003,0.6899362371225698,0.9144718887735619,
 0.8049443869839912,0.8631310192051344,0.2728855707172261,0.8133297249570076,
 0.7028858650059159,0.22832659704255887,0.5758955058835326,
 0.017432949064388303,0.5943866713931274,0.5444015742341944,
 0.3452303633893111,0.6397922459085619,0.9589751977274381,0.17658758498988725,
 0.3005549947577224,0.950177057099029,0.896046150831452,0.36456168343878503,
 0.020738896899645587,0.23460881826048086,0.9043811353217133,
 0.18145270536853841,0.12419252194561414,0.9185004937600778,
 0.46592412099568103,0.14857269056111322,0.8512933120244269,
 0.08819570373478136,0.6423994972962014,0.7818101874938551,0.9921001968638997,
 0.5060513891749258,0.7317436504637923,0.33820488576351015,
 0.20181575466922597,0.7080458485183521,0.31678308882197614,
 0.9798948481312789,0.8862182261167509,0.3672142473776545,0.5909070912436342,
 0.44385607923719783,0.6178945703214738,0.3077823220766359,0.7793434232118059,
 0.016395031734089183,0.5692839079985208,0.7178318627635778,
 0.3779483036711333,0.331864734427181,0.08288147933944645,0.16537432449805445,
 0.3411398463073658,0.45667735972105006,0.5659646983026903,
 0.022189062369953105,0.4060035720173738,0.46821882072275056,
 0.4104025593244396,0.5212802249666929,0.6214890548988176,0.26131455939935266,
 0.1442835945603872,0.6822589879672859,0.7493311166557624,0.9992650899020017,
 0.19528650626356203,0.1692665420200582,0.6614124985368502,0.2846828655514808,
 0.0520114030760267,0.49188707794455866,0.8317615899921041,
 0.35074464717902176,0.603642537221281,0.22515160970851722,0.8562512172716209,
 0.5781174517464995,0.8312811041578831,0.2972943892401405,0.48038714561412976,
 0.3378346483621215,0.9404642223408038,0.6552936566967427,0.5676377212282817,
 0.7212704098344689,0.1903340275063845,0.2243201299114379,
 0.021563271782847893,0.10923425454257096,0.14696552156785048,
 0.7891357735836477,0.9113019538257524,0.47694473312860164,0.722357265966193,
 0.7219560584737497,0.002156729553594694,0.4978788923714008,
 0.6534515973889399,0.13836676778141577,0.8014443995481902,0.4075857134366291,
 0.2229076946294576,0.06026701543924062,0.22956294852999293,
 0.17965318876590874,0.03275227125525959,0.6405436966301484,
 0.9868769626629924,0.06913357342299198,0.44315813333752097,
 0.8528554543605276,0.9706874485640586,0.00172845598697835,0.9539679708537415,
 0.82249748574474,0.7114615003786583,0.9429939021475736,0.7495533294333004,
 0.10640704529438927,0.46733685330223607,0.8865708971739567,
 0.8498918924468453,0.7744193432188438,0.0436496527432757,0.22295861657589433,
 0.47493338629884607,0.6770426409405674,0.5109936011780132,0.689569386839376,
 0.4888346236632799,0.47098875958059006,0.44168035966231,0.18970804393871876,
 0.22378815791236695,0.6671064885900593,0.2894966621557963,0.9491823027013361,
 0.4549425277166731,0.6186080527684628,0.4855240224975237,0.14960244051839922,
 0.18223362806528076,0.4291519675835791,0.11446901671111254,
 0.5058941145574251,0.4712318566668041,0.025141898883761393,
 0.8156214954241151,0.09474904691299635,0.03547069046383,0.5529449595119955,
 0.47211457879463403,0.9481362319650357,0.7409129756030125,0.658743361735122,
 0.4197488779040689,0.6494426016893202,0.4018907724751243,0.932322974531336,
 0.3404612543293981,0.4623998397507786,0.617817178597428,0.1490674270732817,
 0.9265345555589439,0.4044568187527412,0.11945497173451147,
 0.11497668427962782,0.2817379430267072,0.9578035836432125,
 0.36022213951466775,0.6620132293267065,0.37465858064916757,
 0.6834091096778583,0.14001386479276912,0.20862777888425021,
 0.40790930704588124,0.2729787023332708,0.42529469411944876,
 0.9124340625572824,0.6270281243317659,0.9638706343421156,0.25843071059324085,
 0.5040973742301106,0.9136258589374557,0.9226617979061351,0.9373024406711509,
 0.34188738000468855,0.4389436782209456,0.6496025131425288,
 0.12852379932050917,0.4709646057293595,0.4709606713693343,
 0.12649251309402387,0.4013055591835156,0.7400174693604731,0.2779963794841004,
 0.16521585962257446,0.026513330367762045,0.9369272616455483,
 0.05794794036456241,0.5367028503069065,0.022006771223624377,
 0.05789878245235347,0.27591730817214755,0.16100883271380706,
 0.4073199369993056,0.13022171979925012,0.8141214922047506,
 0.056561809242675065,0.7248547456800893,0.876873950570116,0.9646221376531428,
 0.8938533966592864,0.5886627137227288,0.5944490456845416,0.8552751842416,
 0.6503310402527481,0.9322355004773419,0.9475225855187039,0.4964050290110993,
 0.582203981617647,0.7610415956307729,0.4769890169024956,0.15626003688158963,
 0.8495458587448799,0.9643307816250077,0.668064756161329,0.33676240157131054,
 0.024703144444724545,0.6209648510831943,0.5758021741374467,
 0.9588902434494817,0.981968747459691,0.9999249763517177,0.5381172413498642,
 0.193445760216175,0.014873463733733328,0.8606082646029163,0.8901996672561069,
 0.9458450354397043,0.4647197582695419,0.40876891126037096,0.7224594961436697,
 0.24890208851124995,0.9641384251730779,0.20086568100119995,
 0.6452943455132485,0.6159611481700183,0.09624456899721756,0.927572414568705,
 0.07418112312862313,0.5231620344392583,0.36130251634239974,
 0.7606184642296184,0.8383242938958727,0.3858719819408144,0.4108693111979087,
 0.10552479802857917,0.10944046740017366,0.8213709443017787,
 0.10523939123504011,0.7478223583163112,0.49952590026090493,0.571081578611618,
 0.021897835590874437,0.2774421231897928,0.9775516291782612,0.8077838544392,
 0.7492251387057777,0.06484127145971386,0.07771660831567395,
 0.5476911201934361,0.0992766100875544,0.5186687240022685,0.42399548492060324,
 0.450655491707065,0.1118005684792529,0.03433010060969888,0.299435722606734,
 0.42919518334432594,0.48694328728658154,0.6457284027320626,
 0.8914897533803194,0.20101262849689305,0.8947340970427715,0.4194581500407517,
 0.7278653963573419,0.3049188748946611,0.13793148406166666,
 0.32608926823163875,0.7679285307833799,0.5226308252461603,0.8950319869548223,
 0.011867403424437373,0.18113211454764544,0.41202055971924856,
 0.7346065874547268,0.03173329659475521,0.8435898031385562,0.852973768030079,
 0.6750415952050521,0.6631656029411388,0.16012330809648578,
 0.40451806046194494,0.5682292129410683,0.7451609717029565,0.6378932888937017,
 0.1381346088168266,0.29595605513708023,0.9816089014766911,
 0.19245230668914737,0.8911541589690068,0.4540416916043112,
 0.02065067325218406,0.18671425066593483,0.3657403016429138,
 0.8599567455869832,0.43673252075396385,0.11204622144668985,
 0.4298872541242996,0.39739089208869094,0.025222464286194235,
 0.7147122045734027,0.6308057170795512,0.7595064963857066,0.7923090015341139,
 0.8525849671581313,0.7899615150422281,0.043523541499421614,
 0.7205785487068082,0.8570634183283249,0.23371767797219312,0.8406659358422361,
 0.6656358021988797,0.05170686474484343,0.7941959275446222,0.8132935749485279,
 0.7973036420630333,0.47411898037065425,0.45265267439115453,
 0.01588631415881059,0.5691210241279627,0.39633291352411737,
 0.8807326938417513,0.9744361312521651,0.524027723031243,0.100309228566544,
 0.9694372415773006,0.18164064783239253,0.14798032259900762,
 0.18317947136227664,0.36273528991486026,0.6624294913729027,
 0.24519471024461093,0.8794628645533549,0.33514292610588814,0.781295846447147,
 0.7516322424297261,0.32973696079750525,0.42869748656861395,
 0.09212465526806746,0.2621932571113016,0.17710498504026095,
 0.15150930282716069,0.006475032827480409,0.007713337344136262,
 0.7514809135215549,0.8384527573537671,0.13640622000254699,0.8086289384493317,
 0.005800202248136821,0.9542587908828143,0.2428808814255703,
 0.8786310875722982,0.5832084661419394,0.28200310558293096,0.9276107956748961,
 0.7732537052676939,0.43956849345909055,0.8858515448451579,0.7893368330685767,
 0.455246556143425,0.8690117221580625,0.15907757028084168,0.30484876971112196,
 0.4319442963133149,0.8016714692682844,0.6414192866723305,0.44701418668332504,
 0.36710394015202497,0.3879720752269209,0.3033574653839626,0.9448285993475884,
 0.4515577043320349,0.3738947503617499,0.22028924107076797,0.9302727339361478,
 0.44384786191478454,0.7830033880994585,0.5820963751257694,0.8415440347471999,
 0.8908354385412548,0.3857647836960386,0.2169151344740745,0.6072243181596327,
 0.6125351912199992,0.01402817907180065,0.4055116543964692,
 0.47309040430299054,0.22936490776788643,0.2768766637691825,
 0.19495593500864217,0.4409403647768848,0.42492166679330867,
 0.6929261012792624,0.9790051561782546,0.9627537373014123,0.17021396533104494,
 0.3996375193576902,0.7270546254409518,0.356182691651041,0.9105525802906538,
 0.3417178471403798,0.591636730327795,0.5226863782547129,0.39342355854116984,
 0.8796771825952742,0.4384007081230086,0.6461668600561085,0.8278638140685775,
 0.7781469073651626,0.2615567146839445,0.22334518980020468,
 0.17743323290605129,0.8681136029050763,0.9450484470413595,0.6545945287993316,
 0.07295371133857231,0.33217635278032853,0.8184705395858012,
 0.49473654846580517,0.961023233290065,0.6227827632774876,0.6942670678563905,
 0.5359105127109649,0.1240967897132329,0.565403322535145,0.9865766711641787,
 0.4107093861670045,0.06362904867563124,0.025386723356405017,
 0.6528392552578524,0.14143652218439573,0.10607370929728566,
 0.7921331083616083,0.08549253609114882,0.903694315919168,0.9239814344644035,
 0.2808655099881766,0.9146531457466319,0.6766077551680442,0.2733466877886477,
 0.48119436392918824,0.9465231126631004,0.4676093213547633,
 0.22253426991233682,0.5193629504032753,0.017159577547742666,
 0.722377164391558,0.5959823321393041,0.9216129580462981,0.8096780026365753,
 0.8976876064557235,0.8332569884011517,0.4986747871990129,0.6574778469033915,
 0.264270362604516,0.26178265318884875,0.7540012503079835,0.8193932276047939,
 0.8428268409310515,0.8615884268865877,0.6720644295197322,0.9648150979628696,
 0.6917950117633114,0.28527472619049976,0.39969928576251024,
 0.5164938167870692,0.6244125774769611,0.3329307209892579,0.43218750065793377,
 0.8671950052824723,0.9949388040610574,0.5351327817891963,0.43508779000782927,
 0.5468131998107256,0.5809046068565262,0.708245719757149,0.08215641183066691,
 0.2885823036493185,0.09183151742385709,0.01547349673112508,0.521651779937222,
 0.6696794385852735,0.5748737660222567,0.9673297414939657,0.4325710372219175,
 0.3777257980889913,0.9295414783160252,0.7785340189492971,0.9623621169323366,
 0.6437208089497788,0.30381258384339416,0.7015822290221129,0.4663818768595608,
 0.6856456989838698,0.3063237043984688,0.22448140007714734,0.5090705775585677,
 0.666130087610339,0.10966270548471835,0.13752929730895203,
 0.25819757013250866,0.10828756609488299,0.5435370188760795,0.272600983699528,
 0.6656511277042765,0.14806753992012056,0.5539246611339708,
 0.22186434285659873,0.5946003001209857,0.7735653605403267,0.9678542468543528,
 0.4829318764556676,0.24785562085055035,0.8304342036323966,0.9626111351208149,
 0.8889326163106863,0.29882346363258305,0.12128263332716316,
 0.07276224658041897,0.6521179425461968,0.6350681727871883,
 0.19201125814387243,0.11197910974774616,0.9971894014105395,
 0.23152207148870374,0.15414178755683994,0.719660155728334,
 0.22784762405852743,0.43860808781196303,0.4765770395264045,
 0.13877348149389857,0.0437061082130783,0.5527874161049563,0.7326929789155862,
 0.4540164723144974,0.8231718436954844,0.7231725454509698,
 0.060472888611412645,0.3866548404047392,0.8494531162829793,
 0.8239831722309858,0.8924977858148813,0.36202576042821943,
 0.37612844970851866,0.6814090763181937,0.6580622366722763,0.5348345844210419,
 0.1491740184173045,0.5695629752945637,0.4615720713225676,0.6227298669375508,
 0.45524917723634994,0.3932008325000381,0.4610373915353848,0.4461465122029956,
 0.032172255228961055,0.216392096667832,0.8405728197461836,0.5579515563143901,
 0.14347148419498457,0.06821602589644316,0.10985091957058868,
 0.6960515629590842,0.3606889120192519,0.3513375500602146,0.8772629820590222,
 0.2131395740343971,0.557680805419415,0.7504914175247449,0.6421456728137866,
 0.1698476164911099,0.594492729922031,0.9219530782115682,0.1759421137873921,
 0.03607958642308595,0.41543742113877746,0.520067097651383,0.5330175944851973,
 0.2983118889235268,0.005119522843722524,0.3825444304473987,
 0.6091008937684228,0.4807060313786843,0.17593848718764526,
 0.18604268278620162,0.17278817531838098,0.2529466645569287,
 0.9773921538357342,0.30391816031860863,0.5899246440528783,
 0.09964049995157753,0.012323931865579674,0.741847035417226,
 0.7016751448217778,0.9752375139180007,0.9125452701422221,
 0.045233934006182364,0.013273802181394845,0.2786390270329612,
 0.22065815961817692,0.5213426185334652,0.7141244419257534,0.6665358788549492,
 0.9545532269150891,0.515869429420778,0.5782246322518121,0.9523719460400809,
 0.4961197538967891,0.6254222514945429,0.629248190061657,0.799260848459216,
 0.8242218275951105,0.5390374634981656,0.008664489912143303,
 0.32693692344391223,0.04786237451717501,0.04748239237309282,
 0.6765373892593152,0.516053350709945,0.45895596721348386,0.41091498741445,
 0.438352945526406,0.1613943850714853,0.9325278773823609,0.1279318040763875,
 0.24503298367001825,0.10134345991737748,0.1712985640618384,
 0.6654581219623383,0.38883869552617856,0.18009661719401993,
 0.4733156260208866,0.7259691674188598,0.7452081771468709,0.226326905096213,
 0.4138654621586956,0.008977783561246344,0.5478235483846183,
 0.8557881967691736,0.943454571580455,0.9419786327794911,0.34991080001557706,
 0.22570631615550085,0.5766837233683508,0.21581551806572097,
 0.7161093544268884,0.8488112070954201,0.8660015183620315,0.8599834039114665,
 0.00445889649116149,0.6369173519984576,0.04520188000834435,
 0.9571935214830706,0.40006049039365843,0.7245630689245974,0.6556052779053491,
 0.8951248419589043,0.16220193532837202,0.7307861017313853,0.855917507689171,
 0.07743393730523573,0.7353205257332442,0.8809634889164628,0.9735255723791671,
 0.1798956203043659,0.9918362675981697,0.3084756796168573,0.3679610767935415,
 0.024362262005332647,0.7638434655246833,0.2803616237260431,
 0.2887498336850274,0.809891154278114,0.6584097991952813,0.39213730374990874,
 0.7731256459644867,0.632189456887788,0.9243343692640098,0.5933947513956588,
 0.7709868003030803,0.5153132768786347,0.29456233817199695,0.2129201816185693,
 0.3275301843041215,0.05894857394269759,0.988975450004995,0.7337950610484786,
 0.045277431514623934,0.9826950679612506,0.4161467739843723,
 0.6327423630657718,0.13327102571828064,0.44565028993155387,
 0.8470819998837733,0.9211822465688789,0.4055941358447809,0.922204658944605,
 0.4420110806348845,0.0032005932047523,0.8750156662879007,0.24072416372962502,
 0.4696978089654482,0.6557725226129187,0.3641602163530798,0.07534632347762882,
 0.17610959875723853,0.04220230560379712,0.08853286441591202,
 0.06928728965436148,0.942556517024176,0.9890538189264485,0.3839835403069788,
 0.9142997033725697,0.7462783683308681,0.5405865757013056,0.4731431097669144,
 0.8957372957555781,0.8392857139821774,0.7055916472416471,0.2141598645382219,
 0.7053630588543902,0.11064509432012992,0.08412179829807676,
 0.6670497443571293,0.9249847716524127,0.38237882503780685,0.7130111114602857,
 0.21037516194279005,0.031913370962110665,0.38897551681314513,
 0.2633127050848486,0.7056617845633807,0.33954295409506396,0.5169032503544156,
 0.2496942915162167,0.6692513785054048,0.9543002006864981,0.44106195768334877,
 0.03521310650599063,0.3209713590457428,0.7709875641323454,0.6032727307259647,
 0.8922309733201328,0.11198290664117949,0.9283929785341172,0.9508785112145,
 0.7521321175947221,0.7618326139599809,0.18293089946746355,0.2819015121780801,
 0.7530222185859996,0.8307916945335999,0.6362767571295149,0.7034945179041705,
 0.3764673000479536,0.6702266143651615,0.19476091090827874,
 0.05087609040963381,0.8634752471455259,0.27486017862590695,
 0.7057841134673745,0.5042965443436788,0.4279141339722372,0.7512369114719292,
 0.9344772888839465,0.5446349099221539,0.17498189429920674,
 0.24038234480673426,0.6566062371175586,0.5304568230723594,0.5981734782387617,
 0.8013289364296201,0.6371624388824703,0.4602316593657775,0.5412664933052161,
 0.75899116289846,0.47395138825504635,0.6670302133493916,0.5984345311263533,
 0.30557536666369256,0.786480330357483,0.31561667813194183,
 0.39284072524008773,0.384128415817374,0.847950461788712,0.6472569211311222,
 0.03599448042720099,0.6939831906313737,0.8792592901412326,
 0.06722567770789556,0.42107956419165193,0.04420327775270927,
 0.9813743291650676,0.09356279381151988,0.23848281260123627,
 0.4632303758554207,0.19044227110134482,0.5317670111671761,0.2826334891724108,
 0.549444331427182,0.7816124988214518,0.18972834429404994,0.31962427150162953,
 0.09529842222619611,0.057947542822594844,0.240190549106568,
 0.44614838862003026,0.5382210166905943,0.5329477942366722,0.8499442366641039,
 0.6204152829170825,0.6817614357236621,0.4612175424530641,0.7141170783746078,
 0.5310125853912884,0.6537042505961658,0.6773577621686104,
 0.017148887720440542,0.921871258882056,0.5849899393201842,0.9433197996102611,
 0.16783628476345747,0.5645349869188777,0.31474267184280236,
 0.7274719936667423,0.2502513668864498,0.12859245659525165,0.7683551784243553,
 0.5980155122926798,0.20484549424789533,0.8583312150227962,0.5500685970477542,
 0.6460396604895178,0.9754585426642925,0.013804637826081323,
 0.9481766338641052,0.9723314305485005,0.7345593742292374,0.9390698786760373,
 0.5009635820260985,0.2501662450575257,0.8783045794192679,0.7842325755276153,
 0.08683820391274244,0.2825552819486228,0.44832270058795987,
 0.8308644252608844,0.039599572628941426,0.9207076540425527,
 0.4134832669448264,0.9426428846691601,0.878703978604657,0.7600861831032263,
 0.7838487139698056,0.40632465854969624,0.4699839915121625,0.9850122499478337,
 0.22486933641791063,0.8438925056800375,0.1989102224434891,
 0.20439208932707864,0.363833488812057,0.34633210746901033,0.6238317203984719,
 0.843409750318813,0.5843522131665325,0.405887413308753,0.12676715066511512,
 0.42252825325109544,0.8984457815185736,0.8200108256830907,0.3799176532605405,
 0.23118261180196764,0.5665638698070556,0.21154575514675067,
 0.41095457110971656,0.4413970661906076,0.7623732374071924,0.5330137308218439,
 0.1571390450569058,0.3753466719580075,0.6195338708899503,0.5231157924149711,
 0.6537154894158846,0.8129582620484856,0.53524342932786,0.6232513427903176,
 0.1906538153253421,0.3072696617285242,0.01286252426684742,
 0.09642367771447158,0.4367291104370964,0.921088896185591,0.6616394356539446,
 0.479295163300934,0.3295031351866261,0.3102332939162229,0.5269428055662129,
 0.16678322761810327,0.1923727908597681,0.6969365847370822,0.6261662881218848,
 0.116184633225209,0.6237861531737702,0.04326932327971167,0.950313959399619,
 0.4300019807904941,0.4299162960073466,0.4254327178502568,0.9787189361258826,
 0.0672740671566,0.39849645811040846,0.32206170363187114,0.8400330106912405,
 0.2280161786821886,0.8001130819821591,0.6643966439911787,0.5238535471224665,
 0.6236694526768044,0.31689823540662687,0.5658249909230667,0.3989497673671837,
 0.09618953994750534,0.09944219275776933,0.8433608389966973,
 0.6867078533290917,0.3518535297013501,0.4676031112609603,0.6628023070528991,
 0.9625514079651935,0.2858832809551686,0.8103633806949697,0.8678994356228417,
 0.9327111919870882,0.03630173395508496,0.19430191719308643,
 0.2615734009663224,0.2786116375184823,0.5921845687633915,0.2072816708181643,
 0.19045859548955546,0.5395423274548534,0.33944381464908324,
 0.13674513443455927,0.10996924960019538,0.487465476757729,0.7927038969071553,
 0.9763714249305321,0.8051845975031356,0.031284176165806254,
 0.060102185617861625,0.7189323003236633,0.651583546641376,0.489775503496906,
 0.08187538607474609,0.104309043824803,0.48497343203776,0.4996466796562504,
 0.5095057881456782,0.397463706865658,0.16308356172953742,0.44591886951685067,
 0.4182724899096657,0.004591293005223207,0.44026074135015725,
 0.25803306725400965,0.7359343887993466,0.6055088053866343,0.9356612296623114,
 0.9602391916919858,0.6418598825150871,0.6144224638884634,0.3410840531692737,
 0.5241416863159752,0.3032445443816999,0.4545910508290397,0.9046919084778551,
 0.3119214229613062,0.62448864230773,0.24353095274608816,0.19258691214387025,
 0.32544054827134306,0.4630149832636874,0.9204619553368361,0.9637433801273828,
 0.012479951683036727,0.7714560173864063,0.5274835840046687,
 0.21072316422885295,0.16599339346549358,0.7981742735659282,
 0.9731467566438563,0.09905572202410784,0.11277197242854797,0.170401614744091,
 0.04759937506046796,0.6273655009735286,0.9818013603903049,0.6185436024471147,
 0.31432938572886604,0.2959377977690715,0.151998156257534,0.5279473117370695,
 0.9883637326739201,0.8167017341022923,0.5078873130820057,0.6391720288595004,
 0.9458748078715059,0.8603588850312618,0.4722140556774458,0.30886707085553555,
 0.22310641663066444,0.6751773330861657,0.9748616085900168,0.8076019077988451,
 0.9543419824049026,0.8689473898652497,0.9617665679763925,0.9662960214321947,
 0.49116188880325184,0.4164482826493935,0.6428388435562461,
 0.44144265903426205,0.9866959344793073,0.4306442073768222,0.4903498128907874,
 0.8436232470170903,0.0163158263715365,0.9152715646386473,0.5958541943672244,
 0.3048987724661709,0.637199857954283,0.5411779378876096,0.12739640704343214,
 0.8767701222644342,0.6340319495142491,0.15073325508629976,0.3062817392283812,
 0.07447840151071194,0.6016519857202477,0.40985826574582707,0.590781661619572,
 0.010647233541129064,0.1285940095036502,0.9904446847631146,
 0.22672965769311326,0.46583289330035926,0.9805163460646344,
 0.7052619066698038,0.6552311146086894,0.8167119432675,0.6506908512203242,
 0.043557201468883755,0.4988107295947337,0.45880741401535896,
 0.08074720998997353,0.2628132566055632,0.37658804982212635,0.915219298268189,
 0.9587467514587213,0.9318522798759923,0.44351218041214724,0.6337572054516193,
 0.09917523392879646,0.8243082200075973,0.6225606277839109,0.3011435937323892,
 0.8323285766126732,0.5298505377720826,0.5780430629138134,0.8217255452535939,
 0.8683120082544427,0.49855572468810494,0.11313808700180494,
 0.3709902065393348,0.8089000209463275,0.06367915200964125,
 0.20665626095570644,0.9661331105120188,0.8575814131041254,0.4058152231326213,
 0.4962107644253373,0.9934148002575576,0.27945359110761026,0.2266880407228017,
 0.3421493788891238,0.4536272449444908,0.743324665327091,0.22236264802047923,
 0.9388504464570079,0.7433775455053182,0.42850756561170744,0.724076619261181,
 0.10433814638291405,0.5644222532810584,0.5774449269314418,0.2842788617495863,
 0.7921329709716152,0.2981474991200467,0.23212565598758284,0.8456003087509221,
 0.18310276745785214,0.004943348516356361,0.523218387377631,
 0.017492688865559214,0.48859021599676655,0.1522091185430553,
 0.6163604251380651,0.8799539433977885,0.15966990748021215,
 0.04493842423510608,0.2836433915221419,0.5250844629333341,
 0.39172411974506116,0.5964278669900415,0.664023728805953,0.3386169640346227,
 0.7494678457278575,0.1525977258555603,0.16295684437293223,
 0.13572471913611495,0.4097152387874696,0.5521624486724298,
 0.45335399696990997,0.7978566904736794,0.08938606263442972,
 0.7792316071740856,0.039531223061969434,0.2546037166105287,
 0.19327134048613992,0.10653505254024376,0.9672064895241658,
 0.09509593144435524,0.5345699302459572,0.18963813441507882,
 0.5562599630278335,0.4355387918907232,0.9862812055493962,0.09328100191045663,
 0.5512730725044855,0.3074802854610532,0.8795580396842553,0.7210109212652517,
 0.4519566479727414,0.8748145769552034,0.9675348947624972,0.30765631753676626,
 0.4075368609910184,0.7530209406495956,0.4990350965227519,0.24985422916572497,
 0.22224342079889337,0.7086383466151814,0.8139612425105498,
 0.049412345611315134,0.08291629756119301,0.7832130534610665,
 0.5459074543088993,0.3847338295191587,0.6026284564843684,0.5494783247335314,
 0.4035922381160638,0.45265885526961225,0.1291293532267278,
 0.011604844812425297,0.8825710221656854,0.5890936537302828,
 0.8570405488513135,0.6552524966214183,0.9851277123517743,0.19932245678538574,
 0.1437799092611467,0.5674047476975478,0.6219524969299266,0.5768478528385729,
 0.02308314647732823,0.6382053783567708,0.41930120681492933,
 0.8380808616501028,0.13481600731921894,0.27250723620179707,
 0.8708661696600943,0.8266729894350847,0.23067988535213568,0.7972172757285084,
 0.7399878962034312,0.7302690986724498,0.43798740058386654,
 0.24953085489572469,0.9752604638466121,0.999757835679077,0.9906119380225527,
 0.06963709140804764,0.3298935864845458,0.7344630327633708,0.4515423189654727,
 0.23057955029157917,0.2854511728564233,0.8764275592526426,
 0.38977771271594097,0.9304927129725222,0.8746218205722891,
 0.11783952051857582,0.9200338717708951,0.4652181265113694,
 0.04476309854982896,0.5060109159200079,0.5570098793685442,0.6275228887467886,
 0.1409651846619936,0.10287295723589507,0.11603715179940766,
 0.7799394151274107,0.24645041587170857,0.7105773087906561,0.69769781779816,
 0.6350654246280456,0.5565696530992696,0.6206157690233861,0.37891970439322487,
 0.598598519836156,0.969313210435288,0.6805773975985976,0.09922279198258233,
 0.7857177974508459,0.3122566143093628,0.06875783653701795,0.9878501787370035,
 0.19369749863502073,0.16473977153060704,0.31383412064888727,
 0.2787172262124459,0.4878079506287881,0.33063107733356956,0.5326402037527861,
 0.37004331826415304,0.7312588261090445,0.37120034276512437,
 0.6743761123953373,0.3043200256353217,0.35131563661518284,0.3660073314581582,
 0.8031046533296402,0.023014883254684926,0.9028004021031115,0.456160192537021,
 0.5039960707670942,0.9263435460557803,0.1427036036960343,0.9937012770041499,
 0.23154529479326347,0.8919840333802473,0.6120781511090915,
 0.21025659746558722,0.6783742549513396,0.19720419221168328,
 0.4530182516113086,0.2123848659969957,0.6750491228323873,0.19293424267326253,
 0.3038145041241669,0.48112172367325234,0.3048708976842964,0.1718784864234416,
 0.9628600384271503,0.01703240087444835,0.3918535538814617,0.9599066864693758,
 0.4615618003028753,0.7975444540350072,0.847778707051177,0.023965532779877896,
 0.18321720870673308,0.4561697831570477,0.6114731565413118,0.2625611462343953,
 0.6405795488046542,0.9518142661508461,0.4230923142500923,0.6809903086319825,
 0.2375614429891616,0.086471484416214,0.760796217060465,0.5959103666263778,
 0.21476651676942016,0.8635875585735493,0.13495842325483098,
 0.6314855958919675,0.42380164429526335,0.45165248516383383,
 0.14747380659362197,0.18580213495346842,0.3885888185020414,
 0.08495119140401286,0.6665494158696699,0.14892691132715552,
 0.06054560538277465,0.7933419069233774,0.4632364826592683,
 0.06994421094448322,0.7293148857282603,0.6767850231386955,0.935622249216553,
 0.6219294861280756,0.7502190458239557,0.32977822316518657,0.6050554203750674,
 0.8120355454347401,0.26067971830020165,0.5540196146675196,
 0.31619783798364476,0.6772429620559927,0.29667195798210844,
 0.11405362593587531,0.31623972000219014,0.7265870277428041,
 0.2423129971608916,0.7410830856485234,0.15955322482437062,0.2570541431802328,
 0.4296639494194122,0.4805326640760891,0.8765575582047619,0.21650883052012726,
 0.24247086644535176,0.08568576018638452,0.02760553837867863,
 0.1848031821567424,0.6058307796039234,0.8335988282426761,0.3708166432659069,
 0.16257387282952895,0.39207035161955317,0.6436016706731743,
 0.8548271351672376,0.8682045321336356,0.956855194167443,0.6477356477688541,
 0.16113140801628112,0.2909656552072095,0.4390570458434866,0.831153593868908,
 0.2574095744823549,0.060345186253693894,0.6509591335011117,
 0.5943149334894344,0.2247226006631674,0.8678720440170257,0.9985021533393263,
 0.2829029780135709,0.687234830129969,0.5661510666052779,0.6157320701156962,
 0.6469912669861709,0.9003064274100238,0.3701943705180796,0.5961863539812532,
 0.19747650163842978,0.4417985134424637,0.8605946139760877,0.8239778009567926,
 0.07531454889464562,0.8756198674600465,0.06311821521956829,
 0.8322460818994682,0.45422400564510546,0.2764265894188713,0.4801452882476903,
 0.05317293018426761,0.16797194892148126,0.28290950472091836,
 0.020060321485084187,0.7496406035037219,0.6699200867735624,
 0.8686984103388894,0.896119532951702,0.6450419371476418,0.074433005944743,
 0.2075747559329003,0.6892397048975494,0.5144471254960239,0.20879587466825134,
 0.3542583359837398,0.2724291434876981,0.5167729741548113,0.7194254291838512,
 0.7171294573437894,0.8837019678445268,0.7573732404840926,0.4233277028374718,
 0.5618448697073494,0.6082508645057801,0.0524028282497756,0.9083704205188947,
 0.09910269317113263,0.2839022211685497,0.2576695067109256,0.5162213118882419,
 0.36458829973888385,0.4132492050801927,0.8736426493378173,
 0.060692458059230514,0.9659675938008027,0.04137914961150124,
 0.06595284735317675,0.24655451519285998,0.5599670252476221,0.907996814681868,
 0.6078236099547638,0.6550244484318816,0.16273995319828405,0.4795456073733574,
 0.09878881056617217,0.3633623382147856,0.5537054047770502,0.5153045895778536,
 0.16278565492006758,0.5761868939084245,0.909188008520092,0.17550666221371858,
 0.17352130527102672,0.3247523622379155,0.6822298402017017,0.9390471691528683,
 0.1742241377548861,0.7928272217325774,0.9732966891624555,0.17286155297603856,
 0.17398478711627252,0.032369000906001366,0.1385355899549663,
 0.617021501074943,0.44724420858144454,0.6998125869329695,
 0.054430374788867586,0.6672455845705154,0.54890468744883,0.3153879945240502,
 0.09420875685575603,0.38316811473957146,0.7875999317015151,
 0.2562836846446921,0.8375437042140755,0.886681698933911,0.5529715139647269,
 0.21995861877360778,0.22409452203229852,0.9719697212147429,
 0.11018405189680014,0.5093165626446424,0.7476264117069733,
 0.45297172012120335,0.7961328556011698,0.5996857644446927,0.8826791359765465,
 0.6348903826618906,0.5494226960243321,0.5262409946164176,0.8978876056148763,
 0.09217805231959586,0.26178998031517664,0.8095878008911918,
 0.10454077093478975,0.5033713688939431,0.6189536665838661,0.5440629798714635,
 0.6132582174285677,0.7071732355133591,0.8733298000829869,0.09446526376176279,
 0.11538207501609199,0.6313902535006685,0.3119250862172904,
 0.07116001145098916,0.17253464877982705,0.3769186117372837,
 0.08603808391919388,0.08780299153112381,0.3952312512679014,
 0.32794891395115855,0.007230515710531016,0.3984848480493891,
 0.011129439822977427,0.49433148902214397,0.2995894157121971,
 0.29972017241204474,0.1023079583541101,0.6359843078286271,
 0.15431415943031057,0.7845702251996218,0.2925519522853953,0.8171448741413067,
 0.9453220690165705,0.8615207463881492,0.4810287179751743,0.7891376277726088,
 0.13511660069216036,0.8505463131262893,0.00114516292784006,
 0.9901340890199125,0.21824310170803163,0.5762008839486843,0.6610288139800972,
 0.9867763031192975,0.3318542519857707,0.6385497372415552,0.7453067947626186,
 0.053584718335962456,0.16642999536103087,0.39764484186324156,
 0.09013340985420892,0.5022227036388673,0.20159720938964265,
 0.05980539405612828,0.7044384072192904,0.9704156072439994,0.3389615504906961,
 0.010919184430755124,0.036017552349772775,0.7957703806968685,
 0.6775212142593909,0.7457970073284019,0.9363024289666613,0.10923836478849314,
 0.024496029602535607,0.4172652883106889,0.7345533773316762,
 0.6594773208389817,0.34459597234905015,0.2924439094901582,
 0.05524953508651864,0.29498936611747006,0.9478884821903637,
 0.07520323964362463,0.8122113639553716,0.26670622865323257,
 0.6488211080909401,0.8794535417028602,0.6824412171587587,0.2453121497103431,
 0.031943177288017544,0.5923245802166761,0.24349004690266507,
 0.7332041265491962,0.21276394247982944,0.05890960812684065,
 0.6840699474911636,0.3371078282916621,0.5765639929315285,0.09048367794103129,
 0.114858748419291,0.9573054314197257,0.09464263474235113,0.988985820479744,
 0.5223657731103755,0.07764126611218869,0.7342652742985741,
 0.14756526387660696,0.5050504467315372,0.5947443312444971,
 0.10160827169999165,0.30884908014703816,0.8802423007358489,
 0.1260445998110149,0.212024959849757,0.3710294094241565,0.7597113871027434,
 0.45317414424532165,0.525032094003804,0.20322706841080707,0.7874589116099041,
 0.3143077512531113,0.24076946887338035,0.8987906054175293,0.6245323576386871,
 0.4972963980060825,0.277872287092227,0.6611253605213783,0.006232502101121695,
 0.5202068537748847,0.3204248527248138,0.7264671465100001,0.8379771597893385,
 0.9430243253658386,0.40213751348635096,0.8521120209292284,
 0.13917261421484595,0.165672860201996,0.035479575136924346,
 0.38758885978093405,0.8012265499640712,0.3050945543788295,
 0.23715795036528886,0.420264166105873,0.7344089204855138,0.49948083864605575,
 0.9642158019234601,0.07979955321709808,0.8031439952022238,
 0.05301945950122844,0.5141776556743952,0.16352010641435644,
 0.9781679040713672,0.7413677511078317,0.5257734269992574,0.941622386784808,
 0.4701756236363851,0.4790203990612213,0.9472211793651703,0.5834222831306197,
 0.5339159118885726,0.41949490998427874,0.8546024285032751,0.8673084420628074,
 0.6555334728156956,0.32109805043897444,0.9071698326618063,0.9388549157344744,
 0.544079201269093,0.6692971995155821,0.07040864539730984,0.8868073698753506,
 0.11169168161323573,0.13051228506700496,0.9072491982258548,
 0.042644224320814494,0.6236018405960454,0.6147971843692931,0.733510861557199,
 0.1960453688912659,0.9799080332494635,0.7785839701517785,0.26545813363662674,
 0.6108132539783273,0.45658593279552806,0.3872565094516186,0.7156211289508531,
 0.9194796146329319,0.1793661691855173,0.6024965320612935,0.43831140547820113,
 0.5180597586944928,0.11141314154629212,0.17442437503103014,
 0.4702588706454588,0.641005350170128,0.2651564469141148,0.7277377822605472,
 0.7338498823842661,0.8297388713952878,0.6002810232474061,
 0.031020916201758175,0.8600168557257957,0.16064815696677892,
 0.7904334086438969,0.6848912820783284,0.7812855708462563,0.7314949704068325,
 0.19621134840265286,0.8336725269566273,0.45502207064355116,
 0.5212683042074545,0.11729784407599075,0.4202602584447974,0.747952096695925,
 0.33725565279987335,0.9475274060907991,0.29946742875308696,
 0.05154902712247866,0.803643849654883,0.2512250373148931,0.15258224917048413,
 0.7975597408855553,0.31953781832988204,0.42212533237891203,
 0.23098562174568726,0.8849512426040306,0.680093087353479,0.15544310043324894,
 0.6520882028916037,0.5047943569929703,0.6890094060607244,0.2896365127783097,
 0.7478193224580179,0.7941656282348785,0.07857856984790823,0.4610105125709838,
 0.40678730384898243,0.2030965626866954,0.03717229298535807,
 0.9137617532105914,0.41726393624862257,0.7357285241006593,0.3738013604296494,
 0.3124309213340166,0.05260556813938555,0.5915350711888344,
 0.33141332881470364,0.659168088408248,0.1561597762548198,0.9625905922155961,
 0.8290507919930775,0.690441154746352,0.439274334394978,0.104628718298715,
 0.8317783330009071,0.8224745620533949,0.5883734027153956,0.1777647454933069,
 0.5265028435859561,0.31267436870533394,0.5474460874469017,
 0.41243467060711914,0.21246768671075833,0.5195004866629027,
 0.9063064866351181,0.9408215434582438,0.26895272588261887,0.3327854418987469,
 0.39527758215411735,0.9729126420110202,0.40194335503837786,
 0.5196315584804949,0.1832368513614352,0.8645191943055666,0.6713018058687998,
 0.2635962954705797,0.9368244600878973,0.14607163331046247,
 0.41208313493355964,0.884032142085603,0.11547819649662072,
 0.20371327195802247,0.08892240694223208,0.2894963740166455,
 0.6073840723145173,0.7106017010286375,0.14033799373688516,
 0.19878622933083967,0.10542992638286441,0.5830687607958909,0.680306346795241,
 0.5999161839464803,0.35545592227926825,0.01756309089766983,
 0.027995354620254043,0.5941773651150528,0.9917492243079301,
 0.0026584972276384455,0.6432223257448108,0.8553691889963949,
 0.09904347876361752,0.30407341280563793,0.44817294316588074,
 0.800669819414307,0.5067244175343667,0.2899154068450851,0.15255129178182403,
 0.5471139591371112,0.6332966192301088,0.264532512076479,0.626592676888343,
 0.4142368209613455,0.49401507065988237,0.7321926913608232,0.8615812748881458,
 0.6228314135132963,0.30578995794962294,0.014410068764815986,
 0.09390704963370922,0.08409477795933917,0.40822491015248275,
 0.7672875176873774,0.2065692085408808,0.19473308649642807,0.4646392744015627,
 0.4239327732057576,0.8439529515613313,0.47901183615674037,0.2599047002193047,
 0.00603739447374696,0.6826219587338568,0.5793238546860794,0.9792474284773808,
 0.328966606470409,0.4302505801064981,0.7728004375932879,0.5877500229671695,
 0.31139472415363023,0.2549454895405592,0.1842311142697597,
 0.036845990538298845,0.33249926773525196,0.47430220758793096,
 0.6621253066504272,0.6455029810781517,0.47052309061046604,0.527045321031395,
 0.19435722332341276,0.07844019150586368,0.9628547216489558,
 0.7365486876617949,0.6821817287791662,0.7411341224298922,0.8221891561392487,
 0.01778330231924996,0.3610985068532653,3.245916881382538e-4,
 0.30470334365092966,0.8802422866010979,0.6575626768224383,0.4401863732254502,
 0.8768853715788142,0.5156250334471227,0.28567119603232427,
 0.08622421404349678,0.4972581034845073,0.9632349683319792,0.588897775774301,
 0.6396524411663846,0.2442545627858108,0.17851271680588043,0.9287332375254508,
 0.07225340045185447,0.5254036914772998,0.5460807825169994,
 0.40729927210094596,0.7325046190695585,0.10836003324320487,
 0.5023410004780869,0.15998452557655218,0.838717938954552,0.7215396974555002,
 0.9104512821296649,0.5563121446779489,0.5400688496018964,0.13509276601403197,
 0.0986250781361192,0.1354753108272455,0.004546646063308013,
 0.42826580187989494,0.33321765971848827,0.37095791195308325,
 0.6443603442797179,0.7439996620186067,0.44066715621350216,0.7672680188076338,
 0.1477541187976149,0.47757785700749233,0.14304221712662102,
 0.3254200914402615,0.0010497375962786881,0.6224644549309409,
 0.0010182744223602436,0.7515115672770725,0.5396345942261415,
 0.5078593991308906,0.957850673391272,0.0722894081706974,0.8177738369761071,
 0.13024358299661654,0.9540573633926841,0.3836404501311024,
 0.27624604664258334,0.6276090586283782,0.7036492292902343,0.6723039026879951,
 0.5600100401506216,0.6506940860327091,0.9546242568109087,0.21481045401552556,
 0.6574785423245768,0.9442522973042624,0.38106034163647595,
 0.08268959793548158,0.9550004307321062,0.5759140822148885,0.6709816414276635,
 0.6776356191520504,0.8806519011714672,0.7474905412282032,0.38288297519923786,
 0.2696115725641347,0.4109363047057252,0.47841959834123315,0.412410594705913,
 0.7910982465172487,0.717938495727455,0.5483990395283567,0.15250893263947074,
 0.461775248193381,0.9117211223148518,0.8682466450915441,0.42681276494537335,
 0.6087567428434083,0.10489757616339368,0.08791493147110896,
 0.09545267038731065,0.6182609696572727,0.5155673790871408,0.3355944873960631,
 0.8711392113231836,0.3716214410694728,0.9798330194046622,0.29572652483032535,
 0.34633632416083104,0.7174400211184743,0.8816733793371159,
 0.11944254098584939,0.9006356860802142,0.24924041522265394,
 0.12003764940492334,0.06418853097485955,0.38069222977246564,
 0.10168682709757582,0.32902998660851523,0.311340828022294,0.8341852834570843,
 0.5093698452451032,0.474567979370341,0.40277681184003133,0.31103925712269964,
 0.20961220311968742,0.6590301149093113,0.9441830267322662,0.6351055476364481,
 0.8144456845265207,0.6827330025778704,0.7768346527843888,
 0.061224910961802714,0.6112893134306974,0.39304929210643946,
 0.7491356975822864,0.4458871572055091,0.8066379046105135,
 0.035028326685321964,0.8330932843512684,0.6407404488766276,0.315388545826502,
 0.7907584148229736,0.029585374857953983,0.21196988881229362,
 0.3006689255510566,0.34283347967124433,0.6263182897336528,
 0.027155692353346028,0.5517952039217799,0.6940037414058758,
 0.4842387156271488,0.058372072374577755,0.1823178365815995,
 0.3379037788705985,0.4667815688828961,0.7039291906961082,0.7811890049449887,
 0.9934801957749687,0.2537284486037009,0.6096989721952303,0.7980564127026517,
 0.6795015839227899,0.23211768755889028,0.4366606959610885,
 0.30386020101631706,0.8807399413660328,0.7885881774466349,
 0.024830804862947753,0.5623776020287031,0.05401666091220536,
 0.24422947905789727,0.6000871768210506,0.025063556734597592,
 0.5607264465453597,0.4227947967113209,0.9371954218997796,0.8942559482684173,
 0.6914634838738194,0.16845374053829343,0.6833576276593457,0.5877364936805519,
 0.8409605636444738,0.862562340725659,0.7440561162962943,0.7952521348706706,
 0.4648827080348109,0.37182045473392467,0.5380488530136511,0.7689729596797509,
 0.6344810735046312,0.13931387321411182,0.14586987486760794,
 0.9193926903690538,0.5889946498341554,0.3362977476940614,0.22288238732032184,
 0.2709062522275534,0.7445765064153258,0.7099060112209763,0.29877321864591977,
 0.8152011077842515,0.6608933846718554,0.2915320712728591,0.03538167473843701,
 0.5313519661025781,0.0023238906505957946,0.34472482899495127,
 0.14372690181882608,0.35900268589337825,0.35640355413072766,
 0.4703784426727141,0.7036175058399096,0.3054804774289085,0.4916217352215503,
 0.9388704049086762,0.021196003134624997,0.15369693506604232,
 0.1789734241156431,0.23838870566683068,0.6094288339759122,
 0.14952010888010603,0.14452008315053233,0.770658328366242,
 0.021823193067987523,0.189130747232376,0.7043270621602563,0.2670654746496448,
 0.09582295887866399,0.07816370116566473,0.26607577346787403,
 0.1156823918733818,0.9798828794896635,0.9668135610942984,
 0.016467466720722745,0.9170926565502124,0.8471859322440097,
 0.49726859781957766,0.3063696961863026,0.6563141719591157,0.6078170892765081,
 0.9880333938052746,0.6322311185037832,0.8189596619589604,0.6656767890408438,
 0.32915983903117807,0.7354070368001601,0.08240719842537403,
 0.5380055534664875,0.9507952680150005,0.13849134716562828,0.9715460531331449,
 0.6632472058161344,0.7819188853805579,0.09606418455859722,
 0.42231966774006935,0.7911975952115438,0.4005064934618644,0.8126793234539262,
 0.709891486336097,0.8319213864842943,0.7764494738468458,0.5677678436500739,
 0.6132012231176547,0.04352664548151719,0.5346304161014374,0.3616953457945433,
 0.538863649739679,0.22637015069781807,0.004139509163279875,
 0.4936583059758892,0.128259547984942,0.36224377334888125,0.9657068039559674,
 0.6996025930538134,0.7375120570066336,0.013229877141693769,
 0.36796147294712156,0.2332743659978329,0.22627000838254996,
 0.13096716645968431,0.4172685010651659,0.4876822568365692,0.9359464398750702,
 0.957272865891197,0.741017175620688,0.4881413468191528,0.3216928644001925,
 0.7663697875666909,0.5554959562523609,0.6325747649202995,0.27118121587149846,
 0.215488734215613,0.9201585813106271,0.18446947614416853,0.9517147316289404,
 0.2711447936777195,0.9608805240435168,0.7167772405881011,0.45546154954767204,
 0.8071972232933238,0.3457225683723496,0.2913895287493762,0.05876431461409071,
 0.18238268923478063,0.5645763103219003,0.24522761637125934,
 0.7468382991776293,0.029364082068538022,0.3029358271628314,
 0.22592248488167488,0.580896681468334,0.5439016161707995,0.46857768225570573,
 0.40511781019966264,0.22360530179095295,0.2362312953013156,
 0.9948728322616582,0.8884404906914417,0.7614215818189649,0.06484419943243547,
 0.6202091660702562,0.9461460192436303,0.19626521105710537,0.5150447233212796,
 0.663341838546866,0.2910049669130855,0.0674907569279316,0.38191193691963254,
 0.494277202487615,0.20998813378008052,0.7954460361965262,0.3550608468733384,
 0.8913631056340269,0.2999192281859676,0.8151779991545938,0.6121458961023672,
 0.6533542860931923,0.6871688994543634,0.6569279104789811,0.7668785823702828,
 0.4310205301002267,0.7312937392186923,0.9488279050605328,0.9781561435150997,
 0.852190002052632,0.33880287103214923,0.021136140560731542,0.801127120031523,
 0.9703535210001958,0.6523739250651317,0.45364237855200407,0.6590395359291463,
 0.9164757527809936,0.44073179882326907,0.20390608149041511,
 0.5016193289118316,0.5685333905922397,0.685815961138472,0.3567162487813319,
 0.3994659033766472,0.4198190864674972,0.46601519172992767,0.3894789488051661,
 0.0317198881096552,0.12017156776743976,0.3228643631981182,0.7653833800762195,
 0.2715333886690625,0.9892099283858271,0.9516903083572783,0.9752811890978641,
 0.3256436749916021,0.6388042940113199,0.5791759615296631,0.8039340562683788,
 0.8030061233078677,0.8393388795211562,0.7706014630375387,0.2341330749399969,
 0.37467695830776615,0.4254516179941492,0.9670047978040368,0.2253611574368356,
 0.9715406364022474,0.6019627777190666,0.8871152154956544,0.6196751747307667,
 0.5790856195340169,0.11796116161638182,0.44120083947787336,
 0.22699839995390647,0.8043893052953608,0.2729459568945689,0.1521443820696018,
 0.04895803537343624,0.26670498993516145,0.47279483995191063,
 0.44317069357810823,0.7507102087861576,0.7405832152396239,
 0.37785931750201207,0.04076116188534862,0.9155470996406828,
 0.4534871312533204,0.733500759169254,0.011445752177972057,
 0.23049845073665376,0.7641862726466134,0.40561034655744677,
 0.8831754417888817,0.46772763394689176,0.8738311876127959,0.2549730680664294,
 0.06678937615336245,0.6849017611823596,0.25476005645246025,
 0.4011043517969569,0.5617718275084357,0.12655076099439755,0.4637507538459268,
 0.8006227450261136,0.9380420979141455,0.06956199957925602,0.8956590835009529,
 0.6534636943961458,0.34307651147540774,0.2110567359053762,0.7669076333363031,
 0.807044812660289,0.10089244232363925,0.8479536496865842,0.5716393565061763,
 0.47596920007369026,0.5074185081570168,0.19987721923289192,
 0.46405053230086823,0.8765114145346895,0.5503389143843552,0.9166517337792104,
 0.7290730673008636,0.8778166112038787,0.025844638597580127,
 0.8529939888666342,0.5854701540072045,0.8885127143998264,0.2819395047695965,
 0.5797936953319578,0.17520320118651,0.6277770812843605,0.1338920676005526,
 0.47196687044403896,0.2568307389915284,0.9305441990226985,0.8667260263553247,
 0.6176644983040286,0.6398404963410798,0.10772764271642343,
 0.08987215125070491,0.2609507460071375,0.9870780181492228,
 0.43448734272289546,0.28004444838536136,0.5940923355394038,
 0.17262709185970648,0.9467879187663747,0.44300715842151805,
 0.42529944536508046,0.6054262626321001,0.0593036814588872,0.6714605860867572,
 0.9755553626632816,0.184099596030863,0.8227686038730511,0.5905884917676072,
 0.4066166106506598,0.7235719373227854,0.28205433809653246,0.9611555471225968,
 0.7944189147621977,0.11621408906527608,0.7068000015765333,0.2544002695313904,
 0.6116256810473075,0.27344605138426636,0.06837291010424618,
 0.2912441078046065,0.6458591741233906,0.8000347048196123,0.2753083418918709,
 0.9557820383742364,0.719705335797042,0.9195392540430808,0.33700156681797344,
 0.5329691953793368,0.9188312080495488,0.2759618097288473,0.4554511355451698,
 0.5732184241976097,0.7878752787410533,0.5572829375747864,0.19900647983121067,
 0.1259701912321537,0.4329439982311065,0.7391085133153928,0.13508696351430183,
 0.986818109901986,0.6382422799335401,0.9061817173689659,0.9513331626646556,
 0.48080003946236294,0.5406374155638897,0.11454806921219063,0.465480894609955,
 0.03706783700934457,0.6897491995418742,0.38535827529386746,
 0.9910544214669796,0.6943102819122554,0.9370043709455168,
 0.025725988722729687,0.5496476607255725,0.22791359828657254,
 0.5794799385699295,0.12823072789535261,0.08318152863874817,
 0.6369042789570062,0.6485724143853893,0.155193599908944,0.8056154686134855,
 0.21911182099402748,0.6482158060537464,0.8180175644826952,0.267736997924117,
 0.4322509359324813,0.09647188872054391,0.46264816128878894,
 0.14229659611922485,0.07873688458738237,0.46149283655974216,
 0.4928525392799872,0.9565675361824673,0.4115133437804258,0.21521444237754206,
 0.393386045232497,0.5851990164116818,0.20638297057388078,0.38109999391407323,
 0.5206150347221727,0.7309642085423722,0.6713289826565869,0.4785286196996623,
 0.45413675704669043,0.6016688362228679,0.0598709631541785,0.6361480905063319,
 0.30487434999478713,0.9714792271893212,0.13613585935007766,
 0.6804883902243721,0.14508777622129143,0.3150579514329226,0.8601770201900036,
 0.4283896351934656,0.6552472816079404,0.08104686989695387,0.8795602238448677,
 0.46441177378276677,0.5527520124686182,0.04424032850421611,
 0.6500091566127534,0.8560696970707249,0.26048996522358236,0.4214680840002275,
 0.02024457361832599,0.3850611918671425,0.128058467623565,0.3440937117799401,
 0.46294684404690956,0.03950822741200388,0.9522662035055101,
 0.14411113881387372,0.303481556262835,0.34743886271336943,
 0.18262720946547384,0.49122234721107927,0.5797549116902905,
 0.8800235794217968,0.7457163494860956,0.0036663250056461205,
 0.11105615295444304,0.8589933127065346,0.6716390255976648,0.9331422003163545,
 0.9766153117103642,0.5935677409853498,0.4332170861429169,0.5455522257699912,
 0.015023704003567229,0.12007774868344778,0.17899216018175923,
 0.5018618643147634,0.08468056518718337,0.6143577130781605,0.8432337321114028,
 0.5935595436560256,0.10889646644202955,0.5634985157994661,0.6410561979122473,
 0.4290847132318435,0.43821065676590043,0.14439658341085326,
 0.3487918418067443,0.8178813542700024,0.7857256103704301,0.2863160674531734,
 0.9768369784657435,0.08357527476682625,0.7595504773584444,0.6720767845471087,
 0.6578887027932616,0.5726622579575006,0.5590953055544515,0.9907066154191327,
 0.5310033552056692,0.983447737444358,0.2756973020400939,0.7628086860598113,
 0.8396260527392728,0.4658429580656679,0.11625732763886443,
 0.30840222429465713,0.6592245837729185,0.34934459893928627,
 0.9848471539203545,0.31867970417419766,0.0676139270044328,0.9082340683846856,
 0.07767364458609516,0.8117724092642783,0.2917979825586109,
 0.27773785117160266,0.3907607837201914,0.3701938921976936,0.6956070209784163,
 0.2002864051861124,0.7256030231438056,0.27592026436776296,0.9171012366132095,
 0.5994248404357134,0.711235780076923,0.08968565199725875,0.5346962196717897,
 0.6877687304981641,0.7364984316767799,0.11776707963635769,0.4799048175151023,
 0.9370126596107303,0.18021861647315873,0.5250532401615634,0.6101718110740293,
 0.9107083816268781,0.2361302820807486,0.24046480184559782,0.4988046099521779,
 0.7599908326041228,0.6986566376513208,0.41545973672791847,0.9287969569798533,
 0.19356681405642961,0.29241251995591455,0.7801512818576248,
 0.5449731547762524,0.3015926414516592,0.827772479778592,0.17282705487096417,
 0.5044465765044609,0.10124943740980508,0.15361614573174454,
 0.2567538049878071,0.9158264493063565,0.8648376461311847,0.6066258571654403,
 0.7384813437560019,0.34277530858675576,0.6415410249628345,0.5557249702386621,
 0.3854572123717416,0.6686254878059542,0.6060669814337839,0.7578849776005135,
 0.9470015890982495,0.4395894670969429,0.13085545948309463,0.9462293821477832,
 0.2589433109140742,0.1716686620262362,0.4312727706006717,0.1924372822529813,
 0.16520984026900987,0.30603343115331205,0.15094372661667232,
 0.8449688565634961,0.5716526370897609,0.42250068003654007,0.160657134440932,
 0.8257065164150593,0.8851036973954649,0.6292990287254427,0.7463976520655323,
 0.9141750065138566,0.23293385959328328,0.5167411988280249,0.4982520661373675,
 0.6421882030779142,0.42185504579114586,0.5557959559297855,0.8365431686495963,
 0.5066422864623042,0.7324502012008021,0.7853045235934387,0.8132904350486146,
 0.1734943306031762,0.4235695030995418,0.30832085036586254,0.7421498562729886,
 0.29683169246473895,0.5151004730665828,0.9269235571596791,0.7847123068832261,
 0.8540012449537837,0.30609140176161065,0.6408813186256047,0.6300991263763135,
 0.34438901251733034,0.5422232008703333,0.692464898530144,0.5290519806522536,
 0.5328878526764983,0.3872577430869516,0.46194720601623585,0.7942808705338472,
 0.2372085941257276,0.09018293614091621,0.9185546717885585,0.7368190800594208,
 0.6613680987549229,0.8516437679710821,0.5802959742036962,0.17954849392919858,
 0.7492820055633657,0.027918030837786767,0.16425935911863532,
 0.2744465142992598,0.9930855264833026,0.06095809768474514,0.8565723587518447,
 0.7477280077260571,0.8757703644724245,0.1256615218087126,0.9288953814127704,
 0.5397523588112765,0.5841437219465673,0.9589956418810227,0.7919858499111048,
 0.3805737966779599,0.005624863534897884,0.14943658488615286,
 0.9977347843756382,0.17536351516700932,0.49612947050469924,
 0.9735198861765826,0.5879002281608855,0.7879971022465521,0.9093857820216493,
 0.44095428806448045,0.32575684768049507,0.30885255483580065,
 0.7285352487617093,0.3222824357827734,0.9267081113955224,0.5056613817344869,
 0.4216530889525847,0.20416237279642646,0.5994324731927039,0.9531348413713622,
 0.24923061624864296,0.49928452591066474,0.6380193209267255,
 0.31145411045416993,0.11398799269283033,0.5853365922480999,
 0.10831759747160796,0.3195740323123476,0.904247440926684,0.10766509357098775,
 0.22675464251406874,0.3682116121894967,0.14145177312212953,
 0.08312247704602127,0.6855177977534319,0.9178341377255838,
 0.14009715530501665,0.9481301052570311,0.9097555280493158,0.6096635537746495,
 0.5781904211893316,0.5358884630445355,0.6221682957993131,0.8975789977677474,
 0.023943165832430946,0.5181228337633755,0.49958529282463005,
 0.26233162985976355,0.01673228125120363,0.3910619190243785,
 0.6540315838587957,0.523399009170416,0.017839389636140512,
 0.29797996401604565,0.03488740552254588,0.031140821243323957,
 0.2837406619135111,0.6197369136107669,0.28929478217088733,
 0.45135202687898845,0.6466379596916451,0.1875387131272639,0.562746144822631,
 0.4618313115237598,0.68893205808928,0.11554018915230346,0.396792719161035,
 0.9077159196608071,0.013380380357686517,0.5397795669308486,
 0.8851068983787177,0.6316567671279327,0.34297750851467357,0.8429116148302469,
 0.7725308993710723,0.47332837128406346,0.4441765116788362,
 0.39268998949321277,0.1714165797058489,0.267794065079074,0.22203449807082498,
 0.687016022236801,0.38216897930817817,0.005831568345885607,
 0.7734912746683497,0.7357041430595984,0.5766707135805904,0.8554011088269046,
 0.9502813748800587,0.9288637342803331,0.035711103475974726,
 0.23861003509958645,0.6937782370758003,0.9902737155351399,
 0.35679223701403195,0.31599494973473663,0.2504227978736606,
 0.5834866020812767,0.13904170383334447,0.99658810504618,0.8470548324299685,
 0.16537670769817914,0.5588961024419496,0.7821579199509541,0.2257274635626778,
 0.87082942615024,0.3646209331486444,0.2543455648393764,0.928378017862404,
 0.2148257186169371,0.6393232939633069,0.5916832598104482,0.5849549164289674,
 0.03519650804959573,0.46917267258112094,0.860673463757321,
 0.14684822166941525,0.9315213369939405,0.7399941304219048,0.5595107724575434,
 0.7100161389525154,0.23130892548474713,0.3425592572992109,0.6917644892964901,
 0.5211802336624307,0.636646484748706,0.5124427847402219,0.6046057736235041,
 0.6682495747793814,0.29155217391116695,0.9887062078092398,0.3652626802445159,
 0.5679921039849952,0.4702109665458054,0.7783139513236873,0.708782023754876,
 0.46999073681961734,0.9428726846535077,0.3851761039369731,
 0.09296854023487022,0.7856557390955297,0.35124359670933325,
 0.4504590396188829,0.13506368928606305,0.364523418367099,0.28568206121745865,
 0.9465386234466653,0.8418089486297289,0.5945426358028968,0.5029685568766934,
 0.754386709659836,0.72172076714853,0.6557967596956177,0.4840660333687634,
 0.895187110760018,0.9125713637393249,0.3038393980074109,0.24351995566687112,
 0.5655257969889523,0.12397056426446973,0.3190426418852784,0.6580906939407848,
 0.6785062969445714,0.2761812647450844,0.23803976958332906,
 0.03102313389435485,0.27853968030824494,0.03176827041395791,
 0.3809077134986071,0.0893311419719569,0.41308637051333585,0.8423567440875587,
 0.6563822446895616,0.3867225701546345,0.6812011778880769,0.15585444167579954,
 0.08498067584262181,0.7852254385447885,0.8773086670306319,0.7955272839486457,
 0.733702182776716,8.610206117964925e-5,0.3980491017610439,0.7186469171356688,
 0.7940546319147495,0.6259611527270331,0.1695747995471968,0.3750402534984677,
 0.05378286628721174,0.015063801997246773,0.5181787252291702,
 0.33451802497868943,0.18329185188897335,0.46909708446234455,
 0.880299771463375,0.6842221396984101,0.5889862173821717,0.6040684121426975,
 0.438976273697398,0.4226551364189979,0.0620026371085719,0.9038792453324768,
 0.6705993926297965,0.009982590244382639,0.6939830179183207,
 0.5091969681379669,0.8036145739052435,0.6863490813793894,0.5955430370615631,
 0.19132695734956018,0.603238287163673,0.4752477953259444,0.15773142881841062,
 0.27360817025020057,0.6069248742598009,0.6574814817052059,0.6427937575601728,
 0.593208130019269,0.6431659567834458,0.12152735005675008,0.6686884275210372,
 0.16347093069107366,0.42130730072676437,0.9557599713934953,
 0.5336583012975576,0.7339410512541499,0.20967332554119766,0.9634406845958718,
 0.013333384661443848,0.23869442314546407,0.6052265020924215,
 0.6618651624890216,0.5250540343091129,0.10988038254578825,
 0.08197765447281014,0.2473996237635917,0.9828533849813834,
 0.45257566472523747,0.36928147815327006,0.48718671077594755,
 0.9950125644430768,0.15827343123538684,0.8991005903408046,0.7117924369116717,
 0.8779422839291997,0.6289558905861276,0.9985556681227414,0.35002706154694785,
 0.5692213914581085,0.9261380767837362,0.04311430837992647,0.9660142211508179,
 0.3843175211657742,0.5351852816203606,0.10991155697500155,0.9148053202662502,
 0.895692548617916,0.7835611262611001,0.14816928622486125,0.5491676535419163,
 0.8875351926572778,0.680567727715405,0.45314040565986147,0.8233160063268101,
 0.10702067516490466,0.2925087553290847,0.6244485993983241,0.5397228672328793,
 0.3426984579718658,0.9283942994527341,0.7323682818731928,0.19298316412813343,
 0.1251499698877948,0.35361003539946845,0.5093235897151818,
 0.29753106361478543,0.4072367027599334,0.640231183112219,0.7142760494572694,
 0.5814510659768324,0.6976513219280404,0.5300643784535537,0.7176111117309776,
 0.4610421938173459,0.2537160538251009,0.23384890091949084,
 0.17478782166119733,0.4108610392367499,0.3897724162912327,0.6564880365934296,
 0.5143389560741091,0.059917682404347916,0.7996631885630158,0.844471385269548,
 0.02211831837312661,0.8429590264169098,0.5009785405720488,
 0.042134724966788606,0.5506927444845213,0.2073454825958626,
 0.06634814734949279,0.9521336014737124,0.5096135257364325,0.763947435310862,
 0.9638603245462063,0.8926767778338046,0.9287570545327464,0.25725265138888365,
 0.6822150016051126,0.2602307360932449,0.794602159435977,0.9753826004069869,
 0.9249500185004833,0.7174253974070841,0.7403916831598316,0.5227218945215254,
 0.29536773045725806,0.46686313988777073,0.6233240299325042,
 0.8738502433257925,0.5202379767064879,0.2331105286931232,0.29785097663455684,
 0.8687890282236285,0.7063800539064613,0.5676843327227612,0.39460480507640705,
 0.923851194636929,0.9293830377795841,0.9336631324674153,0.0286279757838942,
 0.29415512882440675,0.10536828200577553,0.15221602669417045,
 0.7389959983658452,0.4024206726614563,0.12404005291155684,0.6376800961429752,
 7.996169988486734e-4,0.7763157085868072,0.314120203716896,0.588639121659228,
 0.9891534423952072,0.20037796241516648,0.39966485635487214,
 0.9613369901934979,0.06884815147553458,0.31191714888569,0.21909669200209292,
 0.7898785888396491,0.21107453402894139,0.33132641561434184,
 0.6241200214387843,0.35171559724110035,0.16280616426637118,
 0.8690019377879704,0.6783395197806109,0.7960786109834014,0.8833970479468571,
 0.983859209689551,0.3981071751255718,0.9749145655029995,0.23403435808464668,
 0.5411992103854378,0.9668658096720193,0.7959967850743073,0.6096386719112361,
 0.37121294474855593,0.2866706365685806,0.3657119518262246,
 0.00542194485506553,0.5477135000041089,0.9575279072474671,0.5555240521373883,
 0.6207211902616008,0.6576917543874538,0.7420884316892967,0.35448064991498385,
 0.005237508823536463,0.9461437665884131,0.9945092540153257,
 0.6601186021388565,0.8155582321081846,0.1966879254348547,0.9310832860065326,
 0.9533950021132668,0.21150711084916174,0.4454602617697977,
 0.40232683041087025,0.7572874867388504,0.8847094242247264,
 0.29851910241258417,0.7616866408965557,0.34376374018106626,
 0.8896053955420851,0.1198790796966952,0.3304719142836763,0.30044274932867865,
 0.7596353411223015,0.9505292849009421,0.809106054402487,0.45585550256398233,
 0.5022633758139289,0.20279580255758078,0.8202988446373887,0.284070536007479,
 0.6074559236257192,0.1707203320217925,0.36392693793492203,0.9797929619438217,
 0.6771452122888317,0.22161772091935839,0.5960207610548095,0.3104732921365101,
 0.5016855664099905,0.9834717227874649,0.3018702151373356,0.3603222522359666,
 0.36440205461967934,0.20649192518639337,0.8708431285325053,
 0.46599102582130825,0.14081122439593619,0.1721964288775336,
 0.6252220495770575,0.05268087691874779,0.853892739916281,0.1393205403734833,
 0.505310454183301,0.6308086891803998,0.8556544672635741,0.13457798910894703,
 0.27589211388148605,0.7986449575876987,0.9809732986499681,0.8721537787215718,
 0.9046196738303834,0.1656799265974005,0.4147626741172922,0.09724355557247166,
 0.7798770577415566,0.2971520029749046,0.08749793365204439,0.4182891078323083,
 0.25169616279640317,0.7615457207204275,0.8365894881386223,0.3254761883899482,
 0.22118651564704306,0.8997203165933785,0.8795216218536807,0.6285681331444761,
 0.8558346799102721,0.06853300626621428,0.9351071504062415,0.6054438288096238,
 0.7740288489800734,0.6518260027121604,0.8997505556608099,0.04478602896876449,
 0.11720564083871055,0.9043123383650157,0.21746852495998237,
 0.8021177702786685,0.9003146780225011,0.02387062408506868,0.6261783689687904,
 0.6486990666929293,0.9298011407165068,0.15272028833358253,0.3143727054958223,
 0.1632572697197493,0.8364442427118551,0.8564109140484074,0.7050791558331997,
 0.9359410345599029,0.6425792855787879,0.35153151384201764,
 0.38041970132909475,0.9786060472654992,0.42923912448169244,
 0.9422032414383961,0.6638665185552831,0.3426933685489808,0.17987299502931497,
 0.07774622838384115,0.7559637784048033,0.8249481404996287,0.9308438412155219,
 0.8685319936303897,0.47505574484812607,0.35744304027151985,
 0.2008055866721017,0.00691983889625869,0.007513350798977303,
 0.4512790809087548,0.5517221283754956,0.029654483242046226,
 0.9314646156344735,0.059374929047728675,0.8322748025139488,
 0.31039415139135174,0.5907657426795139,0.5816087586716084,0.4621066927572164,
 0.8716590423250413,0.10344324316364362,0.6286903045011337,0.5720011476398421,
 0.08054941088985224,0.16690436133264708,0.8281323635969319,
 0.25687741675612685,0.8662744462393788,0.013360787601306767,
 0.9834398928297713,0.9532928547152265,0.7660744934799029,0.30565906030563705,
 0.3556422397363306,0.6089012470180417,0.61339233115543,0.2853666948426322,
 0.3360831203159341,0.7328097245015129,0.9308569885272054,0.45509158458302745,
 0.7279088042001769,0.2991937404755298,0.7469854100470197,0.6994868373531195,
 0.1375858581042051,0.7418051506987472,0.2998226177467447,0.6225961002518661,
 0.8952559344094392,0.31441567480277843,0.08218967251927811,
 0.8237840336616926,0.33962263930046666,0.9484175176528512,0.307719183778015,
 0.18650293583889643,0.49274795033114827,0.9041439809983415,
 0.9104332551910842,0.23191628906959905,0.5106633103235644,0.1202189314541896,
 0.751694776640432,0.8167214638277481,0.08168816743234875,0.872483869912365,
 0.7153360198391119,0.05047278500650032,0.07515277840834944,
 0.3178968503888876,0.4494332862345862,0.4893020667459922,0.3832172162444131,
 0.8324479273957646,0.9748559846562412,0.4339861143942051,0.10368202851680908,
 0.25991800348884675,0.18116032159956097,0.7289750716080294,
 0.4423984741572947,0.39739850458644643,0.4167382219321679,
 0.30886368530705277,0.8326752304464607,0.5822712135965157,
 0.14322010806343066,0.7313406238763379,0.7373466439376896,0.4716340058557418,
 0.4599379232136457,0.13006729490867286,0.15313556572456954,
 0.06699271372702575,0.4474563965346021,0.1849250600725636,0.6353603375310161,
 0.22108996760480137,0.5154180117956955,0.549955157751264,0.36064557523719953,
 0.35863018521588175,0.451432058404722,0.10715008543166682,
 0.37805018434396576,0.26814118908728724,0.618983853410796,0.7165927756754052,
 0.46679917152526484,0.39319468933297674,0.8145536699976964,
 0.04190249269195889,0.8178919425745423,0.8886025896055119,0.7952042361129575,
 0.42667617513850986,0.7810616663380673,0.18831477466658297,
 0.17308553564817952,0.750015273146641,0.6030321626861861,0.03511285294736688,
 0.10429102717353644,0.7265402754264813,0.11575195714737685,
 0.8482552069312661,0.9483253750232284,0.8793192121771186,0.7123657034319065,
 0.28823847622107857,0.6862906587769855,0.4729963126798832,
 0.37334720327034465,0.48693676234047123,0.7943754354553203,
 0.16204933842680036,0.7631058771365137,0.06276143045403415,
 0.009250389074542786,0.8158055489980455,0.9454169197603272,
 0.18985151665722744,0.18992906672223397,0.8143993372798957,
 0.6456053593471824,0.0036691160754231955,0.5336510749095423,
 0.1138363336578031,0.6082787147666786,0.8548648688819447,0.6838912733915663,
 0.9113794083261874,0.8926512905897878,0.23679540737592797,0.3137561758729608,
 0.6514994442533759,0.10030145257007639,0.057907126362407246,
 0.9920239998715699,0.6889343827263497,0.9526113200111446,0.6926862899230631,
 0.23653392788958394,0.16717622249232766,0.7712819888088125,0.750311794818981,
 0.9670333012422448,0.5023186623766518,0.06401081723550761,0.6404017364072881,
 0.5282088439935101,0.4667992798561279,0.47069873229182646,
 0.41143220530156654,0.9836812794848594,0.6978979555329781,
 0.42253053318350364,0.7183353729213897,0.3000093568897938,
 0.28177628302948576,0.5393971453791785,0.5593793121322108,0.8260530918702856,
 0.7712624885875206,0.45937880184845403,0.6215312818623521,0.3414458266618876,
 0.7165133755959683,0.5284612333889154,0.5455007148279756,0.8840463510466158,
 0.7234915215586427,0.35164279244798047,0.28836464003441875,
 0.5120376222751314,0.0903342017787313,0.6727402071031796,0.9377335387919833,
 0.8943415047515003,0.17166605426941095,0.6824455918212893,0.7565680994850333,
 0.1315222376473263,0.578550317023249,0.6387556441156708,0.08681844940394301,
 0.1635325714578031,0.5292050355332935,0.8720402470000247,0.7502426554060067,
 0.7282097905113757,0.25594224115419373,0.08058186424462466,
 0.14382046289930506,0.06315360617577048,0.03361778296954232,
 0.15357745231020037,0.32656290841802105,0.5830165533565099,
 0.016668655477079186,0.3086117337177875,0.3962477109493101,
 0.5699145205459509,0.738157498894882,0.492676749368948,0.3782239362084958,
 0.8432066355751775,0.5801866128356525,0.24727095247852082,0.8082039925916733,
 0.9091175186503816,0.8215474396080891,0.597469692575039,0.7593084143270624,
 0.17388389068261256,0.7468748284740157,0.16874933763854327,
 0.08733807738172228,0.45229656638544935,0.5633709547287498,
 0.021129445204222663,0.23355286342224446,0.9150389642808152,
 0.2196602062418045,0.46801990111802594,0.9698228231791198,0.2960560124706334,
 0.680413377197122,0.9636975439778281,0.6958998891284274,0.47012662667006044,
 0.2576382749889101,0.45407591648290824,0.3109467001843109,0.5289616086917757,
 0.08099189748116709,0.37130230882275683,0.5310286653327952,
 0.08025817369982546,0.6770440673905986,0.5805769074578775,0.8303897682020641,
 0.27428035715009735,0.4642535253672435,0.29752976162353717,
 0.0759206783605142,0.029524359158717983,0.04092131529795884,
 0.8699460447457141,0.46327724829939243,0.7547063646841449,
 0.20950894110098572,0.9637651976195226,0.5813391518393175,0.3808475222716098,
 0.44474528258943824,0.339553256617697,0.158515576095399,0.07717860553978328,
 0.7792112129860758,0.6431287761043808,0.5390068207336937,0.07459717493740137,
 0.7918899976403183,0.6696862899843177,0.06494851167882265,0.9829369944734088,
 0.13627931781355562,0.45335902241536585,0.6433128551102997,
 0.13099993763260032,0.5804573745083023,0.6233104457736905,0.6815653638081698,
 0.5893726258488494,0.004639221299569063,0.7067247605255248,
 0.5493089542727354,0.6753411763363883,0.9241103042245609,
 0.032751520354843744,0.6047466466678234,0.884838306009812,
 0.13742931313297002,0.8410263352109417,0.9257227792128425,
 0.09124980520798576,0.4879405966717385,0.7945568846742209,0.8494240960892644,
 0.07831212432293999,0.2548403269071208,0.23481981401104324,
 0.6769890757676382,0.9717852434924842,0.39695620172469936,0.4367627535810764,
 0.14012041485013838,0.4726972874631652,0.2459906786687358,
 0.20226835478626315,0.9600284838358595,0.2545227168066888,0.9337084385017137,
 0.2248073139455682,0.37657540152323155,0.7789856430178989,0.6728960197091928,
 0.4672380756946538,0.13899361569715152,0.8964057529636111,0.4246379158467982,
 0.869349565900782,0.6849072746430469,0.9689128407483927,0.8525482550922308,
 0.17955489822419857,0.08787149844607334,0.10641802500150732,
 0.2968034681786428,0.4677969486590918,0.5468337995591536,0.06812488142968365,
 0.28631287976285114,0.11637859646993831,0.3219844454057419,
 0.08450389111311474,0.12525043186210227,0.6557312865511316,
 0.7943922069039148,0.9997148026803855,0.9035208138743671,0.34210625450909415,
 0.1670594698864103,0.6990746136970946,0.48531515972922146,0.6976554359593388,
 0.80267246183512,0.1289596487884319,0.6412173323283109,0.543363033029207,
 0.7198624474539602,0.2736026734438308,0.48422646096828326,
 0.013592341150756582,0.8168691162517694,0.039201668694762604,
 0.8292356334426086,0.7685673120201421,0.6300043618639756,0.12255310157369692,
 0.3096579670158287,0.4824069176173472,0.5561026409533193,0.2932139394854182,
 0.4773094151095504,0.5621810145862666,0.4678809947933653,0.3580325762576898,
 0.3417203351085254,0.17697217350976346,0.34673938063131793,
 0.8616903329918685,0.3196396629585332,0.4488502176019413,0.3693170744935034,
 0.5164530331167492,0.7884376680515618,0.6327026558025786,0.4890971122654255,
 0.45784389386858126,0.9918781335339081,0.6512630999846236,
 0.08070898028652129,0.024397010713711254,0.7797607397647586,
 0.05293229458481008,0.33685120428701465,0.695628420195489,0.747159387049784,
 0.7781182890062962,0.02833485645115219,0.43824035663354044,
 0.7539372219741122,0.27715347515749367,0.3175669518574721,0.6293443691904196,
 0.38603633682131966,0.33774631222442864,0.8004546120044876,
 0.8637765964300128,0.14033854956636993,0.09322260445625463,
 0.9945554010085442,0.7089397947587728,0.6863026371726293,0.3679677918559371,
 0.6057574214746846,0.2689988402871062,0.1406143317265407,
 0.017285640680944292,0.4763852355267343,0.18633030647919224,
 0.8825410229205403,0.9810754257821029,0.4638728560863662,0.893725255802526,
 0.9206252713585075,0.2738326712564605,0.8551993703516569,0.2550620487842027,
 0.18145135691797298,0.9312358902423781,0.7934321083431756,0.5483255557868012,
 0.4441727083277217,0.384969042909332,0.3854392555966112,0.2775956906922038,
 0.37272918020132084,0.8506082507739591,0.9343067527132837,0.8495605510997776,
 0.5376609789987319,0.16963425019482592,0.4225019140894595,0.6721337908604119,
 0.4040656104759186,0.6372792485369445,0.717690651711655,0.41631729544956486,
 0.9111681546438819,0.12567911647061047,0.03849668563075459,
 0.6754751042138122,0.4118892541692315,0.4020051399857031,0.9290868083164461,
 0.093404366708421,0.15076866503588626,0.987189503677647,0.33964091162203547,
 0.20755099737149052,0.48199211382666973,0.8663431551399083,
 0.8756364170087848,0.1153032999401229,0.192706366774825,0.6253379604371402,
 0.02771865790164718,0.8943463173106402,0.11518291206991682,
 0.45638978672217845,0.4491030515713015,0.06739192477713063,
 0.7405239244511777,0.8287194241809815,0.5389268063195323,0.863763791555339,
 0.31515488388985624,0.5668610978535326,0.05765945765807867,
 0.2518222437452514,0.9012378935772177,0.9951627567080626,0.17995708897523732,
 0.6737835650038395,0.45960056770204416,0.9892694548539634,0.4779862304722461,
 0.30467618733786006,0.5391595485386114,0.6359092046218502,0.670587816879715,
 0.3799742265457824,0.4675279945494877,0.716361580846658,0.010678234110427898,
 0.12073018359383803,0.8653743115993034,0.8906601022514007,0.7191497217581575,
 0.49468752007817585,0.003823722816614583,0.8749650311233392,
 0.04514909510963383,0.5733220995219104,0.6667573048527906,0.5099704287768494,
 0.3162388334793217,0.7246028251242902,0.2673803428566851,0.09833639501247726,
 0.07142283022718443,0.3402971128591149,0.6426065736001537,0.291412579608749,
 0.08518554981103166,0.7807640471863044,0.7817360346782756,
 0.34574899802261294,0.21303123406952884,0.5506981636540672,
 0.5904864837944994,0.7031078131471025,0.20520338809087446,0.4351292124451689,
 0.7547237763964796,0.6315120606024771,0.6430138480934297,0.4763735700223857,
 0.5494347748812491,0.08276881259843227,0.63088692461366,0.8601048934032074,
 0.5591285087161434,0.7464234354614058,0.3633665143353253,0.5064447396684049,
 0.47850095194928144,0.29071403602460766,0.5958138798930599,
 0.38404223963267525,0.08331027353378206,0.32894502397156256,
 0.07133795956590028,0.10448482159858496,0.7417373917057589,
 0.5474308790601009,0.7726698282263779,0.30612768452589867,0.2348958939278566,
 0.31260437588539847,0.4513985377833124,0.08142955838491317,
 0.09065520646650038,0.5307519184827135,0.9345167253659029,0.9533888729805742,
 0.5941240279025299,0.34732855014296626,0.17771289920541788,
 0.2973646784083326,0.397087381925866,0.29267627516231665,0.9125441307054631,
 0.8953765258938787,0.9199226116070603,0.005955432665992744,
 0.6023284255916259,0.8497670899034105,0.3866913955982143,0.1408642752942879,
 0.8125855808059954,0.1562370512661515,0.2845739622864395,0.8563903173283602,
 0.9214895928587847,0.8772369555029228,0.5311663399086658,0.17232015516984556,
 0.31490620908925104,0.9467574539933366,0.596292118614883,0.2793085306247062,
 0.6076433507253785,0.9382520117031583,0.26141859374946164,0.9395036994656066,
 0.9023811085525628,0.9686763149725583,0.8035101726456567,0.7302182533627244,
 0.15803389616014407,0.919455806023982,0.2007116409421983,0.13542361927188595,
 0.9495398348177363,0.9073778299530161,0.21607318404307063,0.3533701182497353,
 0.011122021298386908,0.6162223485697326,0.8674937024983564,
 0.6044320137868383,0.6328242544840453,0.1813950529594126,0.4435920996413443,
 0.10510109906303766,0.5902707819250692,0.3108180685065167,0.6349749542060232,
 0.16351871547406138,0.7165429519775214,0.23113665559810226,
 0.3913685321633872,0.38678205884890504,0.002201645475205183,
 0.04386252583017003,0.42595566728094114,0.5657899387906713,
 0.08273473455511526,0.5286148724209253,0.02965187013729531,
 0.0884190745659782,0.3778536961643977,0.5816116187887986,0.5002421253192695,
 0.7747902701491397,0.538155776321763,0.6824924420702234,0.7468888476559297,
 0.7366689299273436,0.9263327266557406,0.40080624979564383,0.6752279537181668,
 0.10173815784902274,0.6176114123515873,0.8574202049784012,0.8322863754192142,
 0.2826472869482257,0.9860465741077536,0.5131580385559582,0.5513062743009588,
 0.481884799048384,0.7476865292832905,0.6176458227423508,0.5738123039398241,
 0.8849189949897049,0.8148228283470043,0.9685721605644788,0.8832569360363854,
 0.6487299559731391,0.779840114817199,0.7135059344832492,0.6094459274870772,
 0.7211799611827746,0.26887525562840453,0.6534045469505524,0.5841066543302122,
 0.07141603962118515,0.6738944763922119,0.5553895947656041,0.90079429406735,
 0.06632303749561086,0.011932684453999864,0.4024786999347366,
 0.0659559453525147,0.1254994221736554,0.6023839638509699,0.4895942793471334,
 0.06882774022928762,0.459539048468077,0.22934181582892088,
 0.20671568599193935,0.3499815874125277,0.4413684088027031,0.9566480775074316,
 0.6391985432992684,0.06102317753220543,0.677312134609794,0.579619813235927,
 0.568123548590868,0.7389750888144934,0.316079731803369,0.49326270650441273,
 0.2983317899768577,0.8818027504777558,0.8566951218306663,0.7186461314877334,
 0.8679665033768968,0.8670241500585157,0.8805127675212773,0.17703995457004806,
 0.16390448336845231,0.914867387322623,0.4225857803497418,0.653170167984522,
 0.41107710371494,0.2070218659597809,0.08533916260251939,0.15595591450725022,
 0.7824886300577095,0.7594758056892337,0.2674796530426172,0.7424395556554678,
 0.9940993354250461,0.7297733041592505,0.5230042357227341,0.4941805313180353,
 0.2565804999359105,0.050134845816692364,0.7602122623477553,
 0.8250245777147825,0.9598844336925577,0.6349260335690161,0.9484346735335056,
 0.3665653732137335,0.7085849696798934,0.8235255866671117,0.19366122545911513,
 0.6995310798262782,0.5965628563641345,0.46476755926249236,0.5005994581698777,
 0.9973518575817532,0.005468209171128335,0.5838042642979057,
 0.7300268050740818,0.816412301580709,0.05753666974269067,0.2048732853045806,
 0.7925032777834962,0.7613173497867045,0.36087742209941065,
 0.15104381790844146,0.5596009865345736,0.022228520975864696,
 0.3453402680910873,0.4104303574025473,0.4155248612332725,0.7028000396346124,
 0.6778939948582676,0.8076731709266071,0.9392383154232887,0.5974172289902243,
 0.14401016923198706,0.7149334077097464,0.6771892723372543,0.5632909671610022,
 0.2980716311693732,0.04645481731809842,0.1916415629933973,0.9313186836551111,
 0.4952766404927714,0.21789249855507287,0.9970989700378887,0.7313627009553021,
 0.5455474808073336,0.9459726387139081,0.481526893091182,0.8778034060630052,
 0.45211391505545917,0.5368382197149049,0.24650827525851515,
 0.31388602504311824,0.3036186134842439,0.13290203137368084,
 0.23462017897279808,0.6188497748801399,0.24125954254367343,
 0.5014385257032637,0.5440252550338539,0.9701878070911205,0.6944761456133602,
 0.7174886821481995,0.5355757144739525,0.46003414597053405,0.3187762465898797,
 0.5570784663631523,0.2886441631138472,0.5128480876042594,0.5846399432692386,
 0.40055591969476456,0.6276782970103383,0.6702048064936138,0.7506728842952101,
 0.3493668588991432,0.47555080221761004,0.9233271187788958,0.5887070203451931,
 0.959626375415675,0.8747161284333116,0.895932548046032,0.7965899700131743,
 0.2971937029820143,0.3531918059082788,0.29608160577178977,0.7614130898389228,
 0.5449352487154165,0.015001802465871905,0.5851790810587167,
 0.19470375318103927,0.7840152391360945,0.905608173353353,0.8517060621662726,
 0.2219023114996942,0.8436268334829793,0.8667768133888278,0.8870157968854223,
 0.7410557505597181,0.15747233267455163,0.4543152491786835,0.9224727342633764,
 0.2758919473439244,0.3392783015749914,0.992454050466496,0.6680038780632743,
 0.5860495186986816,0.6046939842442923,0.012160232677127958,
 0.2625489720666645,0.5670554487723006,0.8621652099155851,0.6983755736108062,
 0.8738272948054792,0.6500864770604986,0.9987535193909246,0.9321128724900737,
 0.1271212186448366,0.11149173427477843,0.05096168198821893,
 0.2216111984260858,0.6247353998754722,0.4371194932757828,0.30469652751001375,
 0.02572172021031882,0.447887273233586,0.13045347101387317,
 0.12141173976438036,0.007967527569559207,0.20025203558672244,
 0.7093523969062508,0.19768077791067262,0.37875458941748186,
 0.4176016422163972,0.1418495679856947,0.12216783530737407,
 0.23618266499943819,0.39039325344935394,0.4325924886897812,
 0.5394815751792008,0.14608256448136792,0.33257786229271125,
 0.6442071023420858,0.7028462594635769,0.8885850842422388,0.37746513504592827,
 0.202845673849816,0.34305479901781866,0.6233045895125351,0.42425171342773715,
 0.2848434843779385,0.7139467205878763,0.1359081494031147,0.44110934698139626,
 0.36356379108701864,0.1815365919101638,0.9691016244882275,0.6180533379541167,
 0.05565178782717606,0.9317523572498428,0.8705768687464733,0.9731223206571172,
 0.39576554733552594,0.5772353348465451,0.6492689544611114,0.9362953958530347,
 0.8438454651720937,0.045209945927111805,0.7050630502635671,
 0.9453468220548803,0.08195468601563471,0.7860301329542744,0.8617042716509025,
 0.771525511342691,0.0427806132107168,0.5066724931659408,0.86962527149864,
 0.23722690788112077,0.09859438263698073,0.44685779134569914,
 0.7635048268509411,0.6558536203359134,0.9816298180529791,0.24626349686992355,
 0.8894172890001243,0.1287656091172179,0.8389057482852433,0.1974301014229256,
 0.9309362441827087,0.8642309982731675,0.04152239695441162,
 0.21324277968982897,0.17458150405855044,0.11296847696541523,
 0.6493866398056757,0.9082775804652176,0.9642470434747539,0.4213202889633826,
 0.9577620754737207,0.05340836342645905,0.9281137985332729,0.1476090167868105,
 0.893198072117711,0.8035751233589438,0.982160773117966,0.14964021427499408,
 0.5438317903668662,0.9287478557062748,0.5251081992513179,0.3335554644839065,
 0.9286859010199596,0.3976013203351698,0.9261157907335238,0.5670820526733122,
 0.9813484603485974,0.5674567803131271,0.6316416724745798,0.5997338453503287,
 0.27610965525516007,0.2107702015729478,0.12566671165662946,
 0.5774261455819926,0.21627689318697096,0.7062248278892569,0.692260106076951,
 0.8458040545447213,0.9665984654518034,0.4635031472372453,0.9281934406385607,
 0.7006821208443783,0.47048752004766237,0.3335370606854662,0.9811484831957409,
 0.9888409237470444,0.8436832429812282,0.6093629872159747,0.8207274161194564,
 0.2174228488191633,0.22417086708454814,0.23657600036542492,
 0.6753536825019115,0.40490695417773637,0.28151218588613947,
 0.20786294081574985,0.868686400833056,0.7173409631381681,0.7648726514745886,
 0.588671626818519,0.16587312886569128,0.2609456205181162,0.4153167338702204,
 0.32955273736191915,0.8134249561356912,0.5254791798766922,0.6643625891269572,
 0.06538196615320535,0.3297752194499832,0.27610959763688125,
 0.7633727374146447,0.8157082463207007,0.2024237766851379,0.8112428533153153,
 0.4895556850926277,0.7007239292517545,0.4000435242079322,0.7986625666529038,
 0.7848386218917427,0.7514498035911552,0.31312572010352724,0.4303668803230938,
 0.6997716074176423,0.7133244313387215,0.5916759807598949,0.8108432937256783,
 0.5196325862144844,0.7489629549478518,0.22032824285169494,
 0.35891598306901396,0.8002975721107176,0.6702701152095454,0.8980994510919686,
 0.7378112957860545,0.34333443665562435,0.9445842116391898,0.8802636402010434,
 0.13516003270698096,0.2145169220888976,0.5677276109236696,
 0.016507301794065787,0.43597703530873333,0.05821284812425087,
 0.32420778117639726,0.7753952220488989,0.5854436405075988,
 0.40847942310781316,0.7005152593077045,0.1919072050483277,0.7027608126254189,
 0.24298745316972314,0.9081096368859034,0.7548752703585433,0.5126671814953344,
 0.37908705453361047,0.8737486711655209,0.5938232959546228,0.8270782777570258,
 0.27889528054281143,0.9201309127228569,0.34361529636355226,
 0.4823967727448373,0.7989791756516502,0.48103062523136597,
 0.16670645871974932,0.9339446807105448,0.10805842825329826,
 0.7298197940191027,0.3230956909742474,0.44422717934118605,0.734799442375347,
 0.3071441858783537,0.1641974985615221,0.49551961009955325,0.5301442178005031,
 0.13303305983570424,0.0805276835907196,0.5655400562598488,0.733865407396361,
 0.3963504163693713,0.7571445503510635,0.6131531508999297,0.7374528374928153,
 0.021469815600269948,0.908074957305628,0.27092225795811054,0.3306187160992,
 0.297601332617237,0.011499331900086451,0.9892283843926938,
 0.27224418874523304,0.17888186982229348,0.30556770106820674,
 0.5474197751833948,0.2036084282482632,0.31249184224462845,0.6161661131695039,
 0.43122657427289535,0.5231428323292662,0.03606708560374039,
 0.1804598199747084,0.16578252934740778,0.7123346758159687,0.3357382479101555,
 0.763318463688559,0.5157531763110691,0.519199302104853,0.3570292736029562,
 0.1698275395138783,0.958722444769402,0.8183100652983627,0.11745256909552992,
 0.8519783157194573,0.36378869006587056,0.2927543306619743,0.5568020905706366,
 0.5990932086676197,0.7118887432846637,0.08102292049375392,0.3772804059910533,
 0.4042939387176384,0.43455109197264075,0.09073675967835404,
 0.8213855745207344,0.06734587270363912,0.012600303844614968,
 0.9612436248036791,0.14962424965693621,0.20156659506080987,
 0.08033890402021804,0.8362706708393628,0.3262454040451458,
 0.35675816599457166,0.8177646964943055,0.7473685618449968,
 0.18489303221475728,0.04340442397864985,0.8358044226986341,
 0.8829386235053961,0.7393009529218068,0.5311956684299598,0.49565855699576766,
 0.8900178166917303,0.8100500295249161,0.08480692716496818,0.7931697755182137,
 0.9416191785023748,0.5615599722849526,0.15314054627660434,
 0.34714209462632184,0.10388042571369371,0.44308440801602655,
 0.7036916461799477,0.6120491284087856,0.5331632503722777,0.7867883945833073,
 0.24631127757390736,0.4380792271456019,0.08248179767710584,
 0.7398581985150926,0.004265124810743592,0.7443727330980835,
 0.3220807958950406,0.5758786573100704,0.5099948926401582,0.49249266236072065,
 0.7322319165617232,0.7180468325435588,0.6326623146129202,0.6571817739068443,
 0.7153838707801882,0.9569293876815317,0.0484036774132407,0.3061443895272997,
 0.3964678648018737,0.47328713003389966,0.6385981623190629,0.5815900023485276,
 0.5822139628069848,0.06641981355404891,0.19847955495129332,
 0.5623082026510643,0.035818912810474,0.8032821592863246,0.15354512738128867,
 0.2611281357094819,0.8353353964708196,0.8848638758010349,0.23672996975124616,
 0.9791658502743528,0.743797178963011,0.2018471450597683,0.10177981804792957,
 0.057170191574336116,0.5270047820457593,0.7032087252274153,
 0.06273854827184144,0.5987250469340815,0.922490807400772,
 0.015052264266050552,0.4328924638550493,0.4179273091969694,
 0.7982204457647687,0.11215255227012366,0.6435050771905751,0.9400021086454203,
 0.7055227945895696,0.32692118227018496,0.03768401642724184,
 0.13314935134587358,0.6006155812029674,0.12490362420585177,
 0.8842520027166753,0.8179113022491014,0.05296083839118049,0.8455909126778789,
 0.22310166871986925,0.7214984108622812,0.6141994943577279,0.5639434967051549,
 0.815099488654451,0.7510357562552665,0.49332076594571606,0.7968049783024586,
 0.469692511007456,0.5614260507558334,0.011450034860287839,
 0.023973857791608988,0.8907879449878663,0.7536996404171667,
 0.7986498999572093,0.14909476385274423,0.6564424665130625,0.7096522284432157,
 0.6652760493724552,0.1918863059163749,0.5363575366943188,0.6812223895294127,
 0.1152468636294457,0.9905229383931722,0.18240177655598544,0.7758008909372661,
 0.09085132027371212,0.18941647659294236,0.7667385757342714,
 0.6805702520329423,0.29276683106124257,0.4167620875114978,
 0.24628175271115982,0.8801405927993595,0.29692740854962996,
 0.9755261573604446,0.6672642167316315,0.5690290707746622,0.5915438264075837,
 0.5004917238891617,0.6766955489355441,0.7536559734233048,0.9314328037335192,
 0.7234748145787344,0.7885535310817987,0.4566610625212948,0.01777658441260599,
 0.05740168630396714,0.3399169662758774,0.5967625060592165,0.1587403888815836,
 0.23349826683270958,0.7813853480099721,0.5771250873799044,0.963584184540564,
 0.07911831021542182,0.9124922797408741,0.0605786632798172,0.9656583088586068,
 0.9909148713787985,0.3601262337154305,0.9225599149328897,0.5514405852661929,
 0.7075808846034568,0.5172039271462613,0.5637345176622848,0.5801580935886055,
 0.08599105934726159,0.419642531353578,0.8621700016249644,0.15478717099423533,
 0.6901437034256963,0.7679946954936232,0.18119372028018155,0.8949523543291985,
 0.9215721171587479,0.1291755369767532,0.8334497504275791,0.8063966556170202,
 0.47033588018845907,0.08860394533381222,0.7059603535551858,
 0.7442841153301362,0.5581123520106518,0.9216745373298361,0.9606148101078589,
 0.7506924054040345,0.7422337584541221,0.10280567662508244,
 0.025924000326534857,0.5828105405483308,0.5046190998750667,
 0.5681815510331898,0.47205052666670244,0.5412470278677105,0.5622129310660717,
 0.302203728787504,0.7914871362913948,0.3692765473484829,0.011436376964592432,
 0.9342369912948785,0.10966381158718064,0.7516831462928288,0.8634087592052392,
 0.29508963356259366,0.6938199806861227,0.4814047282965851,
 0.04820444602113583,0.14846960464341596,0.08268582130214008,
 0.048010130923549554,0.1723702848306642,0.7968524315472699,
 0.9317586785298934,0.8997455057351519,0.022142784882398736,
 0.4985094093615082,0.0776076992621294,0.0711608258006392,0.03501605486144055,
 0.4724037439160029,0.2942616620157531,0.5077399397219786,
 0.023356195517940215,0.25161703737542007,0.2956270593860951,
 0.1387636138635081,0.385149970270126,0.9051323415266677,0.8635561631334051,
 0.7000842953241085,0.29777728971554973,0.8293461439013528,0.7360006757486703,
 0.2536673606892803,0.4453115366439564,0.31697422350079707,0.1947504546798009,
 0.23397810442972555,0.8275885676003305,0.008768015362627612,
 0.5310556995672275,0.06402992751319736,0.674433853081312,0.19019247759090474,
 0.7808936670344857,0.901274237730733,0.7826362106634932,0.7998725502110949,
 0.7680611239696405,0.7823206856706016,0.6510185632164768,0.5095660026448312,
 0.7133156755302289,0.006120332360666114,0.972420271707087,
 0.05334017847493755,0.04624224846980973,0.916813368517412,0.3424280924357668,
 0.0010923270515171835,0.47359455604809386,0.4427191542864154,
 0.8477884538304488,0.4301631575149568,0.6044608068167806,0.3815999859387763,
 0.07180805802208479,0.41161663635002155,0.09022661600464621,
 0.40068473655042836,0.09028534380015096,0.4062391630367672,
 0.16539368802744603,0.9992494832505974,0.8825188322967235,0.791813681507972,
 0.05451405118844388,0.5465423764123482,0.6825277897509296,0.6775065387059397,
 0.17705686764292883,0.5035301586614451,0.4955882497624614,0.8362083517821464,
 0.4434277813519273,0.18505668452558333,0.3619830992616214,0.8814710570729362,
 0.8476781573482391,0.40013911893369225,0.6649195601988528,0.9034772851153569,
 0.9804261270608639,0.30842553760891844,0.9019555426836419,
 0.12479245015940488,0.5705601160707605,0.09030768890797813,
 0.9952571172587984,0.6419077961373889,0.49377920589366053,0.494059499513091,
 0.5164762454199855,0.342721756450839,0.7538433763170576,0.8043215262730828,
 0.3061846313639486,0.48645847465172265,0.09168852450026599,
 0.3148373591890724,0.8285184439242579,0.8925279695647835,0.4026230986525141,
 0.104616769975453,0.6742092851959305,0.37528239595969115,0.8358171469947773,
 0.7822294868473904,0.36820456856438677,0.941182942564745,0.4198827104276519,
 0.6748570265873516,0.9895600496940152,0.9002419323104728,0.4448141291262828,
 0.041981808895030226,0.4111113794657669,0.36386661660184594,
 0.6033812460458755,0.7362987721483079,0.7759354620683018,0.748902190961352,
 0.7689002138314774,0.2649272886079942,0.5383341587487566,0.12203090333030264,
 0.875607127446916,0.5552857323122267,0.10924660320878898,0.4382110227052254,
 0.6937387342007681,0.4589779648505794,0.6404610064774668,0.12214886696128846,
 0.8734023399846769,0.3307466848590961,0.667855327975815,0.07153167387883907,
 0.669001060192701,0.4116445171864652,0.09099242164736199,0.338121461973493,
 0.26371338039677183,0.24651546915481437,0.6397193113256626,
 0.8902147566053549,0.390807475222576,0.20960930555365698,0.7632779185673961,
 0.18685977301484402,0.23432555667801602,0.7908341009174688,
 0.9664628077213443,0.3284458404067041,0.2811991263744872,0.24194476827431788,
 0.3528833468391528,0.6617862669077152,0.15585405435614752,0.7011880639591171,
 0.6603840344761621,0.13729020796652935,0.31401323617162924,
 0.6907376654803551,0.6091108307149607,0.8126955053610836,0.33492313156338016,
 0.9600363724871759,0.27602691029405757,0.906832676786494,0.8018461646020691,
 0.19551621151709087,0.11982464376248914,0.12321369112777791,
 0.8828924627610664,0.6235658674532464,0.5925579414009647,0.7201736972955082,
 0.7037228932878026,0.523631515561772,0.6977707491620341,0.1727142264690591,
 0.45736025168755534,0.7601896953145125,0.8131945920479305,
 0.03508794173978691,0.7605448677091032,0.1231223694424648,0.711289786026918,
 0.2618244969863971,0.6201401773112756,0.34336524949512004,0.5648260274812998,
 0.03612455501804657,0.6997772650621291,0.30091817208479454,
 0.1309457643760864,0.0013787681239091354,0.3583542169380093,
 0.7424724880237162,0.4333320671156331,0.8933227841949563,0.6006748932272084,
 0.046241337180591646,0.6808542609478178,0.5648988875993273,
 0.5635458643702554,0.21552303875827084,0.12587787427806318,
 0.5996391244902671,0.5956294534254281,0.15080658786490042,0.924521165509203,
 0.1454556471639099,0.9416535735910584,0.9748140300452386,0.3241348899912332,
 0.9131625731146167,0.5849858231352345,0.31058605369462333,0.5572596940935581,
 0.7766573680292668,0.18393502960906405,0.661155858445206,0.9094424281246457,
 0.46152763193097956,0.9567618174251482,0.016099945961874362,
 0.731144717600323,0.7158509531864022,0.8709983146711799,0.009365718515591981,
 0.7370561625390977,0.25856049208765775,0.6093187510462214,0.8507730865546189,
 0.39508167656510984,0.9342837740375898,0.6394603264450825,
 0.06868330550093615,0.3555833115868503,0.3053055671673688,
 0.057243370113429504,0.13986975433191706,0.16772674289087286,
 0.3211107802046347,0.0975021251397612,0.20748120814637316,
 0.23215558200205288,0.24720934894326219,0.08144186341197668,
 0.9581399853564639,0.34618042340743993,0.9495997180342782,0.3606631194264105,
 0.8334138814711893,0.37497682506739594,0.7294865253965561,
 0.07105249802945635,0.6908443491658779,0.6785481866307915,
 0.09827089662981814,0.7670439465607984,0.2356019916541623,0.7397854683726167,
 0.07671309386198721,0.755232183978331,0.11454475120358143,0.4783124830380975,
 0.4691384343996384,0.15994182481770036,0.09202550609551652,
 0.15203712982695028,0.6199413896778524,0.002923492648870152,
 0.8676081354611241,0.16532591212916437,0.5856037674897682,
 0.21094417063335602,0.4695368168139783,0.8198110116270281,
 0.07994134994226831,0.09702907045188847,0.9228136802283085,
 0.6961071183826094,0.525822878589624,0.016779923202818958,0.3316703381072712,
 0.11840256447778574,0.4897386185678858,0.814622958669593,0.8348011233358127,
 0.42775034925039224,0.8112184300722795,0.010326262849950751,
 0.12539667765209106,0.05999157264401622,0.8563966279824559,
 0.6211465220652459,0.6027941735339604,0.8295707793191047,0.39554541369144947,
 0.5298486923552245,0.9453566594970105,0.40715754027567863,
 0.21813637239678907,0.19555630478329955,0.5115467238719584,
 0.8178570338297652,0.7999396055428901,0.9647853778414529,0.5806261879403867,
 0.8561061754282803,0.04935345101326538,0.0970790646851507,
 0.16242602176188203,0.6202350156085035,0.867172398692137,0.9730591180059394,
 0.21247387225780534,0.7685907926139459,0.24930996942635764,
 0.025892936505952036,0.3934466446135947,0.30518329597639915,
 0.08722289853590681,0.7557838153108277,0.9193734074772788,0.7175669542949261,
 0.5987410215580472,0.7761802283801496,0.4077604806976608,0.4687619928585386,
 0.7974151673393497,0.26088735369866967,0.20964030312718163,
 0.8127660595684861,0.16007790936636368,0.18591190834836457,
 0.7162194889717042,0.6076594370315991,0.7375601855357936,0.3792603608988142,
 0.4926890701828296,0.6124873573506115,0.653588304419955,0.6676380566356286,
 0.6234610025901767,0.3395466923624624,0.8224576104547946,0.6228275180402276,
 0.11636512441723057,0.8439838755090713,0.0909180379163319,0.8599312952965021,
 0.7721253129752461,0.8150631707203518,0.9903494421907437,0.2485931224780472,
 0.5797401082653393,0.11004574866527439,0.7363924721408579,0.7229919316255295,
 0.14768856037468037,0.7881379504545305,0.58536362417287,0.5785546790422427,
 0.7922693204509254,0.670920418468948,0.6640947769911372,0.8995519750279602,
 0.9780872790009096,0.27409772414545963,0.555938400479603,0.6249811117842112,
 0.6109236297168539,0.9100726263944398,0.7049239915960427,0.9728183744140123,
 0.9477706853690687,0.8521887348658356,0.17556138839242585,0.1861039077217801,
 0.6066230468478264,0.4946951764495343,0.3644324400338388,0.7068800340977407,
 0.7090348764550226,0.401270002696133,0.6774030341030888,0.6668678663441673,
 0.7241581667271711,0.8652553146628889,0.691983238989081,0.9246146082202662,
 0.022530458678221676,0.4122566079926886,0.131201127296664,0.5079772595884178,
 0.44113504888228927,0.018360648438377702,0.5119185665750684,
 0.39580789945531936,0.34560533949949135,0.0553067227975258,
 0.20464456804734177,0.8764802976058897,0.8082271930921576,0.4329587300223747,
 0.8373849101569734,0.03787624256758826,0.10015050238797729,
 0.8458900146600562,0.789971130880375,0.20851659262856936,0.9830299710639921,
 0.5828948760681189,0.7112687687671129,0.14373012459353185,0.5696226171059114,
 0.5145482399303365,0.8256931073050416,0.04359786811463606,0.5498246705751954,
 0.33218999072445754,0.5030719879954219,0.899550989293704,0.8068795440369642,
 0.36191993053014737,0.7759351059543153,0.8751959637913541,0.7979081075105617,
 0.806725322794805,0.0705226661735826,0.6687473628776087,0.14125053020251488,
 0.61839922481915,0.8849032832007169,0.07556085533902357,0.7803720537710525,
 0.5596009800566262,0.24198961942884767,0.5178522125898289,0.9392439734867449,
 0.8022538430431607,0.9188568663546004,0.9561952506377196,0.987597526544823,
 0.522005308711185,0.3877560997953976,0.3678388211292605,0.6007414265117077,
 0.7577991705142615,0.798766237508255,0.10641110373916196,0.26496565982576314,
 0.4095998292449169,0.7725925814248925,0.9175350770307977,0.9716470174444409,
 0.6093487817063991,0.012539363826716832,0.18470370855904095,
 0.4029096307576503,0.6191374015527398,0.07878600276116332,0.7028790362104302,
 0.7559298972948645,0.3518283389764998,0.14859995473400311,
 0.13031847976975497,0.6958160980163814,0.05551566086326909,
 0.12218258920545766,0.9047328016749381,0.6556689903507374,0.7724722116884127,
 0.14400383155956453,0.4650581455223073,0.35810149348972165,
 0.21016528390065958,0.6305695948377587,0.44604301858459217,
 0.6876576376213995,0.6544871568640129,0.24630413780742344,
 0.015092307606221156,0.23410264391982116,0.5742376015558999,
 0.4696417437706888,0.6397719963346401,0.06343586154888492,
 0.13564460972836523,0.8715351198775049,0.17166501031048376,
 0.2724063899960434,0.031867433852322824,0.9340965633805793,
 0.9078477763849552,0.13789816416647482,0.600751443317154,0.9457867875319127,
 0.24020658779724924,0.46962251876115013,0.8407358749074708,
 0.6624140706002538,0.73904891974499,0.8513135739370044,0.9092152362768341,
 0.42436505588235984,0.9190464341973886,0.7146225298443935,
 0.48195740988534086,0.8689875899735893,0.9178413845983999,
 0.09458791620848706,0.6840165783005069,0.49443068939038537,
 0.7389151400717582,0.9851187974032383,0.9051340953302625,0.3205182229089062,
 0.46438493986336726,0.5573459784195663,0.3351908956770102,0.4107791054496748,
 0.893738778562811,0.14715923146153342,0.32253647866726387,
 0.15275727336457867,0.6815659253951507,0.6534487470639881,0.4869464579004803,
 0.9971125031099567,0.5342357234227441,0.2898628901337873,0.7032592277128776,
 0.7740565984843185,0.24240297536671152,0.9949862060433415,
 0.20225972964909744,0.8721076339942195,0.5309863556868606,
 0.21843507575987553,0.7319334953385562,0.24922579408051648,
 0.015696061151392993,0.8615524737887826,0.946982217470645,0.5851750665754175,
 0.351208728282569,0.41976262434535627,0.8116689953445718,0.19505101888131016,
 0.9749665674209083,0.763351215449974,0.9806886633235802,0.44901665336408947,
 0.3010841235456152,0.23905214981647427,0.3315744063371767,0.6653364822614708,
 0.2566700761832045,0.7703645804516929,0.7628435646327811,0.9944240167057596,
 0.0795682881635369,0.6775012746069324,0.3082938816233356,0.9567679453814714,
 0.41042707625034747,0.3619453234507778,0.6497547458962336,
 0.17847184397686144,0.851303166584245,0.4703787773682405,0.24046392499108815,
 0.8924826187795879,0.1749703369756337,0.35385725920720246,0.5975138847749929,
 0.1727535421256422,0.7615165953210722,0.10713419903628185,0.9083438300209579,
 0.8005550983426599,0.8650619303740932,0.34045066185001194,0.6969685345648705,
 0.7221896020649246,0.8261311886533333,0.8161440013311329,0.7841748434463425,
 0.668701174136804,0.3862856103773663,0.08377794638069991,0.02005553858862419,
 0.582257562085763,0.9675669269897397,0.23544598585388188,0.7991153914283519,
 0.710167717906557,0.8594464930532174,0.39612930011640046,0.5815816552821979,
 0.7857679353922966,0.6238706707404073,0.11827427039445859,0.5796982508618658,
 0.6373856326856424,0.7899638022235755,0.8138250412010727,0.4490090318760849,
 0.2610070863928705,0.14032555172193562,0.9628662089740521,0.5810400421630566,
 0.7762479807217432,0.6776036477994802,0.09865886694638326,0.8307430020117504,
 0.5591988620611219,0.300911806173918,0.4983350229246786,0.8841856927871895,
 0.032324506585284074,0.5295835430246849,0.4548737114900524,
 0.9005327697015619,0.2860448268012963,0.44714503449625964,
 0.058626370390550187,0.4944394365703203,0.7584333153004489,
 0.5612390012623159,0.19588109711761958,0.2672230057614604,0.4966583709616037,
 0.0948637976358867,0.548905404392065,0.30747709711135407,0.6983368098349798,
 0.5689728569934955,0.2532413315665307,0.7972129342982548,0.6326515702398968,
 0.8851108634444327,0.7350308511889111,0.8510467575739743,0.11324211224064518,
 0.7986779233224858,0.7083614990139083,0.9065198917892029,0.7528912303978308,
 0.7956464330396718,0.2613523744207734,0.9204172398301221,0.1121256810113671,
 0.8492910921197329,0.33784028810943467,0.9529381818507554,0.5705172522742862,
 0.7943816552672338,0.8967606592395929,0.016457566707169913,
 0.04907696283326213,0.9268607465914485,0.1353320427816418,0.4153013030725007,
 0.4538896403842527,0.45774602506478956,0.9070391583778136,0.6114589689535952,
 0.031418782581950744,0.9200958196181163,0.6724793860118701,
 0.9508927417415134,0.43746578198960384,0.07548490645135286,
 0.8090854667247437,0.9701278213244382,0.8368986547405012,0.7434540764416798,
 0.8362730027159274,0.7979144307020277,0.7825259684383505,0.7755133845285219,
 0.8441250359775924,0.6110948888364168,0.7625357748787903,0.883025960665781,
 0.6555271859487873,0.9843901736138493,0.010391380450778254,
 0.5442668672121339,0.2579528536008344,0.7027275130393525,0.6362942699860957,
 0.08640469458885836,0.47427021660973123,0.5095592036925173,
 0.12937183076086367,0.6345760048823399,0.6627039713415777,0.3620554818492734,
 0.635803140876013,0.05087196204043942,0.005656584752436689,
 0.9947631960094766,0.5359606682502769,0.021922415396999684,
 0.5455418500234814,0.6675576810574079,0.668269859839232,0.6237574165855992,
 0.41572411773767026,0.09818380249232783,0.34968439683794483,
 0.367756558216749,0.3736875512100256,0.967223723213762,0.7342227578881265,
 0.5573472893859873,0.1424375197345411,0.28186236841981294,0.8964689802818124,
 0.4196036926728557,0.9311282273530351,0.7966488956632993,
 0.041999181248939865,0.8253037737062261,0.6895804166466941,
 0.9292075092342866,0.13185410425277833,0.9884701597893869,
 0.17653192113350136,0.8552138922001551,0.20115415264218417,
 0.12279851975485856,0.1990760680599002,0.5015346817755321,0.8808316644332075,
 0.6375103216874641,0.24807188345331288,0.14018902273669642,
 0.6121510108306296,0.2633507171016908,0.8264112743413992,0.8474972181550746,
 0.32026281248971145,0.10906671797852341,0.7971519827295988,0.474856856863601,
 0.24678827994144037,0.23954397441316377,0.3440587815616465,
 0.6733236260236218,0.04673373478838938,0.6607218555782658,
 0.19732103508112597,0.6792270546143225,0.4990628991180349,
 0.46996111626542303,0.09781144888975735,0.4109046233293534,
 0.9553255886949931,0.22351523595776968,0.1016696555282538,0.7927512499802116,
 0.6802305216757455,0.5467749876786554,0.5966276524646542,0.2333544580386221,
 0.18812907357670383,0.7718383285320429,0.8278584117792297,0.8287101235017463,
 0.0814793793471017,0.7775974285796025,0.8777475857387451,0.31166698876422017,
 0.6342855963167227,0.37317276121732645,0.9369540672228506,0.8573269674586945,
 0.9395780518231857,0.2060842747710907,0.3071936687983804,0.7026311692670677,
 0.8216996592050407,0.9768799377566162,0.39403605408075126,0.7637116787674305,
 0.6243253066873328,0.550471800190911,0.1403382443916359,0.3139432058967051,
 0.05515579430250128,0.377184924071696,0.37817259317950414,
 0.23479860421479037,0.780825541009516,0.22393930769420234,0.9336728080066932,
 0.23002088911295537,0.7985052740953104,0.2747703223151864,0.9445819975993399,
 0.6437644132401146,0.3246977385088534,0.513599286151437,0.925105294354473,
 0.25678036279974226,0.6683817403958259,0.7096232462637935,0.7219590367112549,
 0.8441594830720537,0.20203750561954448,0.17555309510991823,
 0.2301222082492136,0.3582670050255877,0.4592584003507203,0.35550566624816393,
 0.08102613714569995,0.892623212738373,0.6142711013049056,
 0.026660567853977524,0.617449682681191,0.989040360926382,0.45300940671139134,
 0.11637604026836013,0.8189383410577488,0.6738057975570471,0.6651288153364929,
 0.3136539390737746,0.42523474055032573,0.23608998088912414,
 0.8497199475847185,0.5893246971386236,0.8665984319454864,
 0.057889623144643854,0.47178686443734175,0.3434275906456423,
 0.7604129376911012,0.002570738702393882,0.6756136291750696,
 0.7968709839269997,0.6487951752221244,0.26139874957886766,0.5500695548587593,
 0.0422693294768377,0.8631676048052247,0.397559894604822,0.23808953035427882,
 0.6928892175651348,0.4073789636014865,0.5613925623716053,0.4567641086204217,
 0.5706146101235656,0.42150109816304515,0.8119887371137992,0.970383305282305,
 0.2715976103411408,0.700911445345682,0.27494832122033774,
 0.045097661446488635,0.040618656172563106,0.7576274866355212,
 0.4752394521329628,0.7418315700559192,0.6491361708803167,0.7876340992918678,
 0.02431107327261184,0.060892724860517666,0.556136098959912,
 0.2430256296241906,0.8049294298577458,0.8670063912038362,0.195571322002418,
 0.5242744525961098,0.06619606791684951,0.9140810300268605,0.3372388442448865,
 0.8971238270771734,0.3876134117868212,0.6372718423177582,0.9996901482784033,
 0.6556587004717045,0.6989516136940361,0.01490411011863979,0.2340250440280668,
 0.7567935607597132,0.09004591132984296,0.5578687003166822,0.7652894716847385,
 0.726688274174089,0.2123762866842368,0.22358332686085292,0.18737049417235052,
 0.570089802498683,0.3976361128975283,0.7360636361343249,0.9853419124499813,
 0.2606190239951972,0.3838764480846554,0.8958552352018172,0.7594487625997118,
 0.6187996278702692,0.5788503812478786,0.18332663824680906,0.3526182941410335,
 0.6238361983261373,0.07988587475442799,0.5039778473874418,0.9420246905997794,
 0.4175187421845956,0.5663376948929073,0.33482606224987743,
 0.18947753953581914,0.5040887854610417,0.429489844860957,0.6595426259812952,
 0.7867154957181267,0.4816543167391475,0.2369090243834422,0.01607795603690798,
 0.862049042283156,0.7007055955192674,0.24309875618242183,0.8540200542672323,
 0.6786502597919734,0.8744517835410489,0.9691883439800967,0.5090781242896948,
 0.16548747177777257,0.013921074099759956,0.7024267056017969,
 0.2353929269127284,0.21825079426721705,0.0029106874738302757,
 0.4071147443058247,0.7916479753150609,0.8080309841809612,0.7771230122487895,
 0.04058411664303185,0.13207900901094383,0.9424706399104923,
 0.8458516361423221,0.7449652287446266,0.09367922857954525,
 0.17109044570007959,0.5258437133620006,0.20702484705207413,
 0.7400035457811664,0.8521275732099709,0.10484758152763463,0.4173417828634438,
 0.8028056229942153,0.16849182472070978,0.93131763471924,0.24480076930063466,
 0.15311213572538895,0.14993866977187298,0.5132280392182504,
 0.8595873066498074,0.2929899841669755,0.6897529077600046,0.14326564385143903,
 0.6833415449555629,0.5911364736522536,0.6013735818225057,0.6177895701151342,
 0.8048193108184005,0.7916089177347208,0.6671326284699493,0.7155245729942247,
 0.6918315782145791,0.7031123040598731,0.944339821801726,0.34653069371598655,
 0.9956445850803911,0.2794527114951819,0.4289544002146184,0.3873168692918666,
 0.2093318274733441,0.7118860314106819,0.6260433699768845,0.5336038385563435,
 0.1682818069470282,0.6027824417049695,0.9107273397562934,0.02483286357794312,
 0.5378900442085054,0.9575741385730774,0.8186081559287378,0.5779498413345904,
 0.7791068259262537,0.5427836114138924,0.6187291718017383,0.9324578820252039,
 0.21339101640276803,0.04156367467703248,0.35077821220245076,
 0.18713493179079443,0.5695305775493422,0.1291575470645684,0.5258179046677528,
 0.9325163594900853,0.11268375137572406,0.42249239482510714,0.319584743940329,
 0.7426745843733065,0.712530699438829,0.8512468498715553,0.8838809157111029,
 0.23219144421545002,0.19957287406355095,0.46474700517647327,
 0.34566480831790325,0.9054710923716481,0.02811899112854055,
 0.3508722610579934,0.38446903014059286,0.798072295564567,0.36568984592086506,
 0.6742369403682993,0.28813693702036713,0.008910393330929911,
 0.6234273227588166,0.5231164624354142,0.2531365195492197,0.16400905735354376,
 0.8494586418789832,0.8141447269736185,0.3991753918214722,0.8887197086502314,
 0.0078041734218587955,0.4868662608993737,0.2616105801522116,
 0.47174721319627866,0.761936551868749,0.8742818015200787,0.5638030698478332,
 0.20979104465203846,0.16463904760871984,0.30840751474042183,
 0.6058808097272447,0.5939177483102467,0.2605750537078386,0.2931085736861547,
 0.5116757330149386,0.34967242118247654,0.7369292361533647,0.7746782060301803,
 0.26592111998977364,0.28971039859053926,0.7510583132989966,0.361083398347835,
 0.5593960794184498,0.6810423595589953,0.6127133047055628,0.06328333991498314,
 0.028764513802776004,0.15239505957972743,0.04692500593295945,
 0.7521961900710664,0.45512101061987265,0.20775870362823656,0.61335839414481,
 0.4891992621571112,0.711657966790261,0.7564056362908825,0.07976109848406743,
 0.05733182834856143,0.7262147394796827,0.11408048182712116,
 0.9626981270180007,0.22816473622708233,0.7623281246825998,0.3865254721707264,
 0.9241065592882037,0.4413931391972352,0.6367083445643927,0.7279622755583269,
 0.15358522474115222,0.7840440787940322,0.7782951020856174,
 0.17493690464143907,0.7998829855042175,4.2427732131700147e-4,
 0.7107960264080009,0.07194554622099492,0.5581651365511566,0.4090535611988453,
 0.556015904520509,0.3942680871464044,0.8614652581390215,0.7345467813136375,
 0.9540384273794621,0.404243150098761,0.5731306342447753,0.26086322462441824,
 0.32403169771286644,0.08321685464817752,0.027718001549945193,
 0.8361538867605995,0.8907228297759239,0.5013981645366781,0.7647390043827595,
 0.2420723982937112,0.669344593024005,0.3867001570977462,0.8065543277725592,
 0.17583745675610374,0.3441983092153945,0.4233559409602582,0.9308775660075783,
 0.16246319903276885,0.3038326796672224,0.5419279474725005,
 0.35974647152962014,0.6152039247350141,0.22666695060986763,
 0.7386187791994538,0.3359449798299481,0.9253794937700934,0.8342509938501534,
 0.7202578325694962,0.23289991715613945,0.2986422345431303,0.683777487480114,
 0.763458100643201,0.09314491488450449,0.4725298628105573,0.03954721486443069,
 0.25710126941638345,0.12976510307177924,0.21988808459656006,
 0.04689037958396847,0.873770806999181,0.5000628172548354,0.6757504045809221,
 0.5264414408248516,0.1547776392782112,0.012288653266149141,
 0.6031904211469682,0.10183577327213533,0.6880580030121412,0.2764528347151448,
 0.422331725112001,0.17480301518064578,0.7796901207562168,0.23362537012414397,
 0.9371167504212587,0.4283900378094231,0.21179338904706224,0.6572739793295063,
 0.13163251986851987,0.664897449962856,0.3583897234150375,0.1335301135856165,
 0.23894629944003787,0.8546672489269971,0.7269060487773142,0.1947119997041421,
 0.5269861279301141,0.4904795049464393,0.4113598957529805,0.9647138234051682,
 0.2777846591555634,0.6522753960672388,0.21606196855085724,0.899138360673704,
 0.8132182813602524,0.7596125072391662,0.5093609850785209,0.1268937573279405,
 0.6386021088823115,0.993136561412204,0.22664929636259123,0.8135984414754314,
 0.4348196353898063,0.0718827895605968,0.8300195197121086,0.687741554579033,
 0.48332702709152797,0.244357297316097,0.12807454046427924,0.3781979268091119,
 0.8461405188012614,0.6754397513033962,0.3204656886263424,0.8491368022134685,
 0.48668292109971734,0.11364205290431517,0.9695338539665442,
 0.4011555928068905,0.9131251017459903,0.11931387661321857,0.9406334104715409,
 0.23488422645287954,0.07531519032736578,0.8130015470818243,
 0.6830726009915238,0.7694646582734148,0.290188700873766,0.6656155001852229,
 0.2126114396864115,0.37707816079223766,0.9842346800001947,0.7139385018560999,
 0.5086171071350232,0.9279584449390468,0.9879919168541147,0.5351731326949715,
 0.475505486813791,0.6559720722140017,0.6956312665616773,0.9064472518943777,
 0.18181192804020052,0.5649029061398566,0.4203795118202993,0.4988207389115662,
 0.059102877672339016,0.30501802312896786,0.23959198073160282,
 0.7947003262692454,0.22857099448544205,0.3851977685571951,0.5355090819373989,
 0.3954460160912737,0.12991291496763957,0.48238631473756854,
 0.04739134610597884,0.44821966521776957,0.9747179519462394,
 0.7732953545596818,0.7304508138788058,0.8596711723727829,0.24823470204896392,
 0.4076269521186955,0.6261441147347006,0.31854765060990564,0.2020920728079909,
 0.5640361527576723,0.47902734896990384,0.30449774437419697,
 0.9640022005995376,0.12175751687746428,0.24593545911985404,
 0.2673527967550614,0.12658074449469414,0.5331199944657387,0.070776991212958,
 0.6851018466982676,0.6348593519347755,0.12962853488736392,0.6180880975374735,
 0.5961706647454733,0.7711214659534774,0.8297600939629263,0.48487123823895506,
 0.8106420076668662,0.7883408327193937,0.8180462483705947,0.8175695788195769,
 0.11545327050352461,0.9483854758866035,0.8688168471898378,0.8212528951459718,
 0.23692324907058326,0.9618492375820166,0.9036314726638203,
 0.03532767449067409,0.8038376867599347,0.21268634786072038,
 0.11220756579383984,0.9014665504550978,0.5070565747053812,0.9779205794204288,
 0.9361922231339698,0.42066207809888745,0.8099852178959009,0.3948136645239779,
 0.5521227312437742,0.5919971773723425,0.0700756278815845,0.7309567185442065,
 0.7332080985565078,0.3124373504159208,0.7614833341170082,0.6576069477306523,
 0.48608413447859367,0.6292770505558472,0.5498394577617594,0.9809201961230097,
 0.7988471437958425,0.3808435396343619,0.5515186182115905,0.7944914529820609,
 0.6213459893247608,0.42793667665531965,0.12178820816071856,
 0.6728270330150353,0.7658460360811152,0.6994871727894255,0.7744069416614499,
 0.5972804487724077,0.3320992645585549,0.1953309784676449,0.6001277693885998,
 0.6877292637182583,0.17089932382389794,0.13843213535278553,
 0.32432100538676545,0.6993657399110476,0.9374898475626309,0.741752279631,
 0.4440771668715352,0.49701034907111397,0.3651666812232731,0.8021128751751139,
 0.6818344961278477,0.32092852558109164,0.2707188752198344,0.4284587130851337,
 0.86951085309505,0.8475185175037823,0.5464837975239436,0.43244449206735647,
 0.012092212753356368,0.24473053013823953,0.9355106545386053,
 0.11964837829734576,0.709688514235899,0.33856853899869266,0.8153152702095726,
 0.20237699548432175,0.11438699780365436,0.6408423873588402,
 0.5332528202592921,0.7620138416721358,0.7036210955286785,0.7089987914963364,
 0.35728497462739806,0.18569966224775902,0.7498059809729744,
 0.8950246024110976,0.6212939864459048,0.8947292653732236,
 0.034217385213546114,0.43674124838002626,0.9697705814902369,
 0.7084742147412971,0.9363836262907064,0.29417962701449785,
 0.09191456254492525,0.5531950756712741,0.7509500960487252,0.5224626346071828,
 0.4650129274716348,0.12182735212077611,0.359112664595701,0.47804159637920485,
 0.06727060841204358,0.6824939669368858,0.9147788886818353,
 0.18942431748870414,0.20124451227328266,0.5459769318563777,
 0.2122146458893437,0.009616805654199867,0.835605072857557,
 0.03355005205939199,0.6505314070575093,0.2810764985459008,0.5328652424598834,
 0.357634887677315,0.45395863573325823,0.8217937015225489,0.4201436761245093,
 0.5293759375220681,0.6087140290393047,0.41288687965639237,
 0.43803027791963434,0.6864643935528277,0.5815027381781576,0.616670678659594,
 0.7798043060585556,0.7809674940889901,0.8429933361775904,0.26701932025965147,
 0.29778521588041795,0.5723887030249799,0.46399208878816056,
 0.2594980257240781,0.3541871910521799,0.6567572736560011,0.8870046134488353,
 0.41698341742279355,0.07469963753534847,0.6072022606479897,0.527286249837708,
 0.8018837646757466,0.4578399823147282,0.2654639068107211,0.7388768311705094,
 0.4852765289917258,0.08412251001289828,0.8750281146271535,0.7341602787988117,
 0.7358232434628074,0.7762169103740203,0.06601055390151789,
 0.07636219460402094,0.7145809755057411,0.668552523074762,0.4787922552275926,
 0.5087972367977426,0.14823913451361326,0.3451009331050362,0.1307286265818809,
 0.4338938313719143,0.2289493383087826,0.14459824827821444,0.9478185288074561,
 0.010137504675160391,0.18936177029638046,0.7235147109726641,
 0.3186119900119393,0.11987896384771235,0.1666044486291426,0.9505040055169723,
 0.06476756735350142,0.06217202834630553,0.43409486740764924,
 0.379896933485281,0.5241060471765125,0.5626273463476863,0.2877302810419381,
 0.25706902296079215,0.5370575002160002,0.3759257985638468,
 0.15387984123897747,0.22276943721103848,0.001255050766675203,
 0.18742990496182943,0.053782088335257106,0.6646736600823455,
 0.8542088162520511,0.4552546482382043,0.8416365289846773,0.7459539346560778,
 0.1853596331015842,0.9082575166861417,0.37508086896871096,0.1095979189186218,
 0.22836721067388888,0.9344399749739071,0.10946963431890966,
 0.45067808187136316,0.8608977888073516,0.6639141515659229,0.2944203851867875,
 0.7205545255018412,0.4527394648220223,0.4811190945394782,0.9733094064181079,
 0.15452454143121663,0.47084675917302365,0.9789295701454727,
 0.31277248775393596,0.3524064036794925,0.5175174942319212,0.2085438648390634,
 0.7692182513945972,0.16360018091072748,0.07584708191556788,
 0.07315506032295194,0.5392665957365003,0.7493615334814333,0.9755810393021999,
 0.4300667060733391,0.405692132991219,0.09926464939808632,0.1596678445029478,
 0.6443039924421659,0.1945737695784726,0.7196176208371245,0.5894337954124381,
 0.3320503698791062,0.9176532801747928,0.4887925786563855,0.8136683400080035,
 0.4988963754470458,0.12076382969145838,0.09743504840220851,
 0.5012653041033865,0.18150931320767616,0.04432223210026365,
 0.15221480309773516,0.023357041805594836,0.261739458698523,
 0.7752144977425139,0.603924892143382,0.24750849263132313,0.5398144496875298,
 0.46253569896462055,0.6371255050858439,0.8461676652174774,0.8801967966881644,
 0.27951237805268936,0.9833933348064431,0.4894489256658995,0.911199778749185,
 0.67293891190718,0.8093698284158832,0.4106799448032459,0.1483305438482082,
 0.47716526538934345,0.24732691392151163,0.9559663364892709,
 0.24486119077692592,0.6851455303753584,0.6647873955708957,0.6625655869843845,
 0.7936076045074991,0.051290520296709,0.20283182751539441,0.5730121328645041,
 0.6919469921641767,0.47781050697929817,0.2492828621020584,0.6213184289549691,
 0.28141634054688436,0.1636954208995155,0.31040486392950184,0.452446860063755,
 0.963916566287051,0.1549501175118153,0.675784451537737,0.8571951677218074,
 0.17584708393907245,0.9016505506338249,0.31479230161328164,
 0.9056101969785958,0.6121247255032682,0.7742501404595472,0.405942610031232,
 0.6150751177281352,0.8738611394865239,0.14926733065606546,
 0.48642116072227526,0.11775256032648518,0.1883825449341784,
 0.7979345650605703,0.5768308405317237,0.9379780758274371,0.5099871747992443,
 0.2859371943411424,0.8843243505057481,0.2172452034010215,0.7282308931076411,
 0.4543690483602838,0.2128003761201167,0.010386682507405043,
 0.6827442376617386,0.4249795535903719,0.8854306676196525,0.19058503162518992,
 0.9270649812267631,0.29262474036547703,0.8269373693761266,0.266468955969415,
 0.6589304878762426,0.43577957234804443,0.44616993607915123,
 0.4213997884771763,0.9210131867106559,0.21719095653522913,0.4041386567047828,
 0.20459460700721177,0.8974765053775199,0.3506293993765389,0.4590210232112575,
 0.3014168985017571,0.559005374027916,0.7294833953406858,0.754005683937674,
 0.7500333868233173,0.36756381921388703,0.9600941611012122,
 0.17518702436815559,0.2024842899891568,0.32606756899150136,
 0.6713063415248925,0.20177189171993415,0.5639548916803347,
 0.11667615784754937,0.009246411619880268,0.6150043609363162,
 0.7154004037370517,0.2402380867486794,0.7254700748878101,0.0592021020634077,
 0.05892778662191933,0.21507509490749122,0.15785764320113038,
 0.7970878587245567,0.1989322933046458,0.03946165352649289,0.9567070952648449,
 0.624003214711994,0.44453967941097705,0.19732922018835386,0.4952604394285639,
 0.5959383728760759,0.3173217849904503,0.026675202052880653,
 0.8238139986213705,0.4493846532816077,0.368689728843854,0.4173857733846801,
 0.22438530122104994,0.12667827110278163,0.03958250279213704,
 0.47082005657664494,0.6870554739400297,0.5947398647844423,0.3107958108555404,
 0.5980998271305755,0.9781165492875301,0.6174395065746184,0.38935036313079274,
 0.6956914773157616,0.13131589003619437,0.44722833416258045,
 0.6465436388528492,0.3898389329397288,0.10850586858385758,
 0.25598670305852933,0.017233703356145158,0.616511996973367,
 0.5082406326776633,0.8486539881900168,0.4744596991099772,
 0.010178509461827523,0.9012672126136416,0.09793481111453584,
 0.8580316382400995,0.3387021487501418,0.8320419575643404,0.8281593654219435,
 0.23808063556972714,0.14201130085310987,0.7553557010637106,
 0.37032944403730295,0.2655222863703406,0.3727916100430497,0.9832790560879379,
 0.9099325568812244,0.8884907673096853,0.07960162254325864,0.2279945473599626,
 0.6188479373202029,0.01609122673465857,0.13044977556494275,
 0.7128665628202631,0.061835184420131695,0.6569439937868141,
 0.6005777359843828,0.5840105092108017,0.5358430002580336,0.7182955375509305,
 0.30321141205595814,0.6502237961092734,0.4521814644983724,
 0.06549845594637593,0.1219222928012117,0.0612914264693708,
 0.14461818162532636,0.10656330455327834,0.9139112521131635,
 0.5980274021613221,0.3246955056604336,0.9862534503644041,0.5980445418334497,
 0.3599991224513849,0.38145969220811693,0.15797561657267656,
 0.45515518918046294,0.6851277834070858,0.13328591976178727,
 0.8920107980937991,0.5052969306827695,0.9235597891964461,0.39234957754521316,
 0.9906295442400008,0.2393986145053416,0.4744357711448868,0.244435707634737,
 0.1924880472834989,0.30904847901881427,0.6679799454609269,0.3070654315293936,
 0.4783426085804001,0.5517604807061037,0.37025824931492235,0.2754569313713502,
 0.4700855263478618,0.014200034143571627,0.3210062104600253,
 0.6507980948687232,0.79723144146821,0.6350306431420917,0.8677349960341649,
 0.16715977085356448,0.0013986337229136003,0.8562910681065845,
 0.5912806075058546,0.1294224809955521,0.9771432789756482,0.8413102293389134,
 0.4726260406779331,0.1745382634346978,0.15647183523873187,0.5130693737349147,
 0.36276718575096545,0.739069274437575,0.44569400888670496,0.9203437520670419,
 0.6445362338463563,0.07608327211845411,0.3213257261984155,
 0.03835307749490524,0.3038951242535386,0.33379059337143147,
 0.3979318108269416,0.7262016243404518,0.715479481659268,0.5316568750642898,
 0.35923522480955183,0.6321259265037951,0.06350927359346215,
 0.8779547979616836,0.04046751743406374,0.5735161693586821,0.7104202130575413,
 0.5920103799330566,0.3239734240965989,0.6901880711586285,0.887343325205329,
 0.5382341526812862,0.6341767369067223,0.2816433984786775,0.3219286159561012,
 0.3341603312584397,0.19923572946552137,0.882612738969911,0.5413632066203488,
 0.0468965556813179,0.933988449173573,0.3604476051932495,0.06967782525634814,
 0.25801726803788627,0.7187048661032808,0.5986515511621828,
 0.024470658494890274,0.3988404368928137,0.4026095633742861,
 0.29745098011364635,0.8281609832157871,0.46271779849511985,
 0.011758238370705443,0.5797700147185569,0.07474495077201815,
 0.40341819836427595,0.7877897695439638,0.186645306973788,0.5020181135291258,
 0.4560281049084691,0.3827446644365444,0.7671267272564595,0.5726666852794399,
 0.09631327513178156,0.49701875049950395,0.2470781154801277,
 0.5808064055575796,0.24643793137646852,0.76429493615238,0.9038473950651137,
 0.890907683689955,0.5817761460664513,0.37493716524044607,0.942864327729175,
 0.11825116973261873,0.6678297970886846,0.8018342090245874,0.5740640407113851,
 0.9468741238544619,0.11967205780807677,0.12521637163649468,
 0.5246048118628992,0.549003987302267,0.13580242363612605,0.3335863716515316,
 0.09609223120985055,0.7529829890812605,0.5913339749688524,0.5315954395911041,
 0.6207144158297603,0.13507015420191504,0.32739958344214637,
 0.09005540507653498,0.4937737945706031,0.6329667376935565,0.8261539742209808,
 0.993491041877105,0.758632943263989,0.47833240138558974,0.7222378801054629,
 0.6278441316264858,0.5152164048021138,0.20594173150010564,
 0.03955540590494189,0.7632012643217637,0.3442468490759073,0.9055665609119288,
 0.40283749145554737,0.7429568466992145,0.44990809042423385,
 0.031625336561073136,0.6801924718107332,0.5830515989861099,
 0.9183030604437207,0.8493790919256267,0.6474754761546931,0.4935911908687439,
 0.7061845044711887,0.31984819352217764,0.4143606554939291,
 0.12463776784000669,0.4205982322116255,0.4243533020707435,
 0.27718955709025817,0.764449300058,0.6464500621396628,0.3073090438721048,
 0.8198134535927336,0.008714675286311535,0.6266299097121218,
 0.24132869005606583,0.1536194476164343,0.5877953644622097,0.6134990919510175,
 0.703640353304569,0.9242563524293851,0.7811064652799736,0.32280235824365766,
 0.8917264426244824,0.4293840305262073,0.6298330573817038,0.8467909274095209,
 0.2764423394966833,0.7368827198650307,0.3999393104260326,0.9239406447100718,
 0.9355551866878328,0.9302891954282764,0.6011239477600852,0.9749504114821109,
 0.7102854146459473,0.8185285601144161,0.6706776432969851,0.32074628211792167,
 0.9744817045410392,0.4162331927809608,0.758308803182018,0.6227317738490991,
 0.4527110206946484,0.13710985202727277,0.067039209314385,0.21307043619204546,
 0.6458468165481116,0.37903608118031784,0.8502548367525212,0.6975216859347786,
 0.3734974449345023,0.3565685011349665,0.7815932546464939,0.6988276030985562,
 0.45224310824858605,0.07627530697164608,0.34775871498101585,
 0.6306911082911903,0.2516768132759444,0.22260247151820556,0.2586835105397054,
 0.5637057711979453,0.15850419456912457,0.7079016778990384,0.0657852320919321,
 0.535344494138484,0.29492645895800906,0.49211229652864963,0.8764458000370642,
 0.42679447254909897,0.8692145496784559,0.6584682888094416,0.656523735152677,
 0.057995533553131584,0.6362008053735932,0.16212105083561912,
 0.7534006488884334,0.9850260737576962,0.6827206261566509,0.7719866750812964,
 0.97516677613902,0.994431708018408,0.9368890563828478,0.8524334132883092,
 0.26618774988518634,0.4840120138854427,0.26350039290028715,
 0.9264444974987747,0.5100028249472539,0.6694208972353494,0.7656435457585199,
 0.9131074880197132,0.5730072411869703,0.6849645594344095,0.3146228254235901,
 0.07128966332149633,0.6610049243065655,0.34622944250620724,
 0.4118496559766254,0.8002917597763078,0.9339367114649526,0.3963258891378151,
 0.3531254079513886,0.08429306820652016,0.16725482262730573,
 0.24931710522229233,0.5479885651488039,0.40767510748095703,
 0.005269729958485514,0.15916244156788906,0.49839623851962445,
 0.08543871992794139,0.8517856707288121,0.6561039512376552,0.68671972548671,
 0.33149697708840276,0.7439525630090222,0.14485538132576503,
 0.9185731747660559,0.7045272121241478,0.8759093735560409,0.6659595997361573,
 0.29059436397349836,0.4922975754443667,0.21646504977757297,
 0.6158695090581905,0.6952806315567066,0.6754334515811368,0.6379551222770621,
 0.8603409959779813,0.40952880247576895,0.4001145262597158,0.7432084029320447,
 0.429082645494959,0.4082674634105762,0.06739904591032675,0.8968785326751989,
 0.3787912371868656,0.616376656707147,0.5522845982917478,0.2877240244000787,
 0.8705412769969894,0.7379795737286238,0.8783549167730237,0.9131884206642575,
 0.8031726074276413,0.6986879736326806,0.808226686769646,0.8620470468195004,
 0.04854934233885111,0.26276441686940155,0.3597460832456828,
 0.18644358664207505,0.6762355607941211,0.5756854931996501,
 0.45025168894866374,0.1466448459162788,0.6475237494959262,0.5636353058555947,
 0.844556515724059,0.8345919700349463,0.4150776127358876,0.016813499740500504,
 0.8402692358923102,0.31752356961962,0.5944632292877542,0.3210730406776393,
 0.5238725931488799,0.04317041854051207,0.2583004246265015,0.7208442019939075,
 0.9835950290607299,0.1410226027637842,0.3738682912660676,0.8310590148431578,
 0.9905493098720354,0.119100596326983,0.7408953253815334,0.10972452510796138,
 0.6247834583429986,0.3022803276176802,0.5273883823835951,0.09293222306218607,
 0.5660995317796279,0.6289524539737353,0.34693102165744416,0.8609157009593369,
 0.16146291208090802,0.11865840301066355,0.46085297432705985,
 0.45499298755878115,0.9625512068373985,0.8933626009939467,
 0.007272996455114811,0.8199001051377446,0.5243493091094049,
 0.038391624241138955,0.5166734849041361,0.4790732447591278,
 0.22225143061853148,0.39890481413205703,0.931340161495252,
 0.13570892870702178,0.756500836893921,0.7222437813034727,0.9629506906869462,
 0.9121814601654777,0.7853407293399812,0.5212316999614726,
 0.018184457395123976,0.4363613227531191,0.6742189667416075,
 0.8915613078341321,0.555066581674956,0.1320840237826293,0.4959741560452393,
 0.996750285701256,0.22245470458355565,0.7023883565054942,0.7629267139726044,
 0.6477345815010466,0.17560523240185077,0.3142643043458021,
 0.15952134299926268,0.6491135229514295,0.23833401906876284,
 0.33469433008135696,0.6560812683979106,0.21986453990142774,
 0.11554124479182859,0.8681688129399818,0.22001599841435449,
 0.49028156154635405,0.9281142835515908,0.3666096104605745,0.9945786105370117,
 0.6539604691152816,0.6160797899337939,0.2375408691674029,0.5169929407245473,
 0.856037310669121,0.8561228216111296,0.4891112492349825,0.2727670801142632,
 0.19497516050858044,0.6309167405630446,0.5540458057907179,0.8960270941489791,
 0.4895991486578348,0.45956013308849764,0.4329357307044459,0.9428309991079278,
 0.7639648132119431,0.4590730229670915,0.3459857405719897,
 4.2722597874811363e-4,0.7852808115946925,0.9002863709373903,
 0.9180701004114917,0.7791240057762754,0.17266498776937933,0.5086195081199902,
 0.47151150479428006,0.5655247749361934,0.8461119357844283,
 0.38114681896395586,0.7656232295820379,0.03078057713260096,
 0.3489094246791047,0.9097478879397096,0.00803664908920454,0.8684615694231175,
 0.8272676531393596,0.1372153821713865,0.6288822809394707,0.4951769472090486,
 0.06089891817787785,0.4210174165423297,0.6139962187672224,
 0.057749091537888764,0.6896691174526842,0.287305232574799,0.6087671678027495,
 0.4445966195469092,0.4081572009110582,0.1637289020662549,0.588744766416421,
 0.1664095416414394,0.09794119352975073,0.7740774750781206,0.36097674977296,
 0.48449128212107617,0.7701521506961193,0.3720818162872498,0.282390420864542,
 0.5296964625711104,0.8587673743596869,0.3639669298767718,0.4028774043942034,
 0.9615960272647923,0.2317047131866491,0.618366390417,0.4948880064521082,
 0.1468918973793354,0.7024265304863616,0.717256297044731,0.36577552522548396,
 0.92757657184011,0.9525401116960948,0.801471903594841,0.2625826173748066,
 0.3508545230441753,0.033469620464444905,0.700086713919926,
 0.06878650002647169,0.8784302710110465,0.5171859684565037,0.1071146441625983,
 0.9830022005287073,0.4596639819735534,0.5196808453363213,0.9230931441075291,
 0.7046193935474515,0.9871635397696321,0.7636783554102063,0.9485945547114536,
 0.19644361063895976,0.2697947864540846,0.7344446107495242,0.5455138522575299,
 0.12561775337739878,0.05518367440357852,0.5555743155208291,
 0.29732072466807735,0.9111155149637186,0.8494635379236414,0.3127507851903152,
 0.047669153368763095,0.4855495106162626,0.5268061116037783,
 0.2572370054985297,0.33491641894619195,0.1176473309913727,0.7050486738365894,
 0.1897362127807456,0.34085893128262523,0.5510387344255844,0.5777527496758036,
 0.7068637371319568,0.4407268408042815,0.6955187166299728,0.9582676726421336,
 0.9445197008184391,0.5126315743785859,0.8888152348508965,0.4825736850214424,
 0.3926978594576681,0.5827062377806708,0.2845924086094652,0.9913217969648616,
 0.020200342922830572,0.8098282327539866,0.40643133397431375,
 0.6890185808037849,0.2826194046130339,0.4416414872682901,0.9128911249670225,
 0.4300178231067865,0.25187085291829914,0.9333303562441446,
 0.20240183754470564,0.5289477158829942,0.20620275073806393,0.329914832155752,
 0.7619826821144374,0.8857821458862001,0.6376838821635721,0.929705231618924,
 0.6105994026830373,0.05154743489351732,0.44665409539822376,
 0.7798841166231643,0.040137436184094155,0.20832985547284444,
 0.9969403442954491,0.8220126460670172,0.6782473050565001,0.521070884418099,
 0.5472664933590011,0.3983779958452094,0.0865353666499652,0.875968571444908,
 0.33831622816977347,0.19163766317818487,0.2779603153430824,
 0.060292293445470935,0.9555325919730202,0.5607803835392006,
 0.7109513222842674,0.589404104040479,0.18682655747124532,0.470127529252129,
 0.5474240650112314,0.2397300318736194,0.59648362149641,0.6522244849553936,
 0.9175546006599155,0.16228095274375054,0.015841886172053177,
 0.006340358151097347,0.2438837962070317,0.6771981927788021,
 0.48839153591325335,0.3611947742154855,0.11188220135174265,
 0.7955636018732911,0.25576985421936316,0.7651788954255287,0.9681694253564418,
 0.7907081830636172,0.8489339189221492,0.10003310340082483,0.5351801423426482,
 0.8146969467749401,0.27105546992668805,0.8160154534634586,0.9304562423906523,
 0.9189889861495224,0.1646815048280139,0.7898780640354102,0.45939151029588854,
 0.5532148365781057,0.9248338515113594,0.7315431649446392,0.7389048327780022,
 0.6633161992213098,0.9895127506628723,0.0943826086249272,0.9241970409122892,
 0.33016907096451464,0.40537823783231364,0.6814160044112505,
 0.8864754651357453,0.7675931344260148,0.8716957354962064,0.6330169395554675,
 0.8711138715035898,0.8540987238559828,0.4905319093984809,0.47355480636373715,
 0.6952370629135908,0.927283663982348,0.07199387371339971,0.4372894058072816,
 0.010081977494521666,0.7732285703362961,0.2618854633529869,
 0.1857250691198058,0.9005765818574482,0.8397621820412007,0.3095235519825399,
 0.32071321602488734,0.8041786594828976,0.967754980400032,0.2620667851191527,
 0.8123397868878478,0.01388098182620201,0.3196687172303797,0.6194774239725692,
 0.9894746575325324,0.5640443503058553,0.4166838604735792,0.9508666460325901,
 0.46160087188844046,0.6267590700673539,0.3480209554303447,0.91146425141456,
 0.6375392022261398,0.4798273276446342,0.9972391966092322,0.4185892797960268,
 0.7922396121505154,0.28034757041479674,0.25795479670316523,
 0.8380603855531927,0.3625147924029174,0.4034902351955192,
 0.019526943990513246,0.8587213884318164,0.2467369388812205,
 0.5207191445352544,0.6805478958996256,0.7541254851249256,0.18805700491574306,
 0.7852712518451631,0.8103869708089563,0.09696544793728801,0.5641572750890289,
 0.9294048068668157,0.8992158056683189,0.5478127618902804,0.5562876307797178,
 0.34993812984377837,0.5391109013916524,0.8993729808632696,
 0.49146133746206055,0.5136148452752912,0.7619148740515712,0.8901357337287843,
 0.7110506023123295,0.47982644370833427,0.6334233696199166,0.6998952879869779,
 0.9761840434297472,0.32120056042181666,0.8743165984470713,
 0.28197485607649686,0.45893993675055755,0.8051086150391817,
 0.6767588150571116,0.2771760423898305,0.5040856933933539,0.1532120221367248,
 0.2046817326688647,0.4803630844155333,0.5949974946514099,0.07480727736643422,
 0.9932915829072041,0.015349796060245002,0.4905036400385452,
 0.05756939655656168,0.5266513239311357,0.6066569033310261,
 0.21041748572836927,0.4250218131805482,0.4806777949904427,
 0.18049403919080242,0.7013721332330481,0.029139365762305358,
 0.9559968267528954,0.742458357329993,0.7265137400306264,0.4649975128223298,
 0.7611651239656466,0.8498747757527504,0.8148863143915924,0.13101835042267485,
 0.9754813576868426,0.9670726635395938,0.6196926793487949,0.6105363535276618,
 0.375940740576848,0.5141978373133385,0.8864113826720234,0.6293669073504153,
 0.5203834885921368,0.5370598422416107,0.03991467791395187,0.9437997957910969,
 0.4440611395484879,0.206041283107248,0.7649087903503108,0.5836981870111659,
 0.12201635810957845,0.32545530146263135,0.24817080457775897,
 0.5278244037647439,0.561036487437612,0.37760741730207614,
 0.010321853385716784,0.841233961813954,0.514386958642455,
 0.036529672017242776,0.9407627249112807,0.15666426075174888,
 0.7087221552696976,0.5162174561458532,0.10009174897752238,
 0.15439492875295635,0.6415981029621611,0.8258883580892105,
 0.44137318427283834,0.9967761525835791,0.4449762430908957,0.7671029360126361,
 0.5624157846385573,0.8397626076711602,0.5652766975437526,0.6424749496072008,
 0.7442788402367585,0.06129710026908941,0.9655328954724897,0.8786430048261212,
 0.9458603036160472,0.5579515835553109,0.4645427575990423,0.5172483480951424,
 0.00397853529463954,0.497012673320288,0.1936896318276321,0.12451235730973365,
 0.5310624316325601,0.3079751364496176,0.1110484291837519,0.8894842303089748,
 0.3758860916378678,0.044960116997137756,0.8345260620667695,
 0.8939159510485565,0.4068194895867918,0.08499373734932614,
 0.46093639660666064,0.2120048344669757,0.5506755579216008,
 0.07540504578538743,0.19839887664215528,0.1305546481716917,
 0.6935368913652762,0.8336509413606193,0.20992320609328685,0.3767908005016041,
 0.6612879107227965,0.499306067376863,0.23557075501517577,0.7792739034688846,
 0.39043657719876446,0.5763552752959116,0.09297717381217219,
 0.16385299118532748,0.8717227066374974,0.8262485224816494,0.7026245113957906,
 0.14806860263899857,0.2745877939111657,0.6593511639928408,0.5804695222646623,
 0.44089740298192004,0.4453912446365563,0.12731907363670947,
 0.25865363271169073,0.5785007569217085,0.25982740425097406,
 0.7999072216679401,0.749976586221921,0.20459175618828207,0.6252937048204945,
 0.6076087234957099,0.21964652326466105,0.6048290596919599,0.7468274210355218,
 0.8833558328957061,0.1633215116351362,0.9033622023546242,0.7124061708528311,
 0.9695335194502597,0.0929620914221504,0.9439273045572685,0.23241236786483555,
 0.40197443573449565,0.8558352555488771,0.4412648671574587,
 0.46177057442867453,0.2322985696069193,0.32547176087529484,
 0.5046397464300936,0.7190255993361113,0.06428010678867369,
 0.09062135337202282,0.4094555575798483,0.2363622832633564,0.9230916620149746,
 0.2372666297868058,0.3336148911603446,0.03261384495422859,
 0.42331993820755953,0.3177169879298557,0.13369613027729288,
 0.9388505995001071,0.18972735603542135,0.4271899215527192,0.6653956759677756,
 0.31053702766583846,0.2978150371396633,0.993942070857524,
 0.006431164224952046,0.29617355713050453,0.7998471349030088,
 0.764927888199038,0.3754697220857841,0.25239117039145165,0.21828867477802905,
 0.21368630784728515,0.036417567006558293,0.546293635689414,
 0.3071695519285701,0.7531442996463821,0.3350915265808272,0.3803890532847851,
 0.95674860381039,0.23605438444227667,0.7598575534717238,0.9375616261988967,
 0.7113405503984948,0.2140512884735376,0.1845835762380844,0.11425000824909248,
 0.1538083910569651,0.4173639618669427,0.42406179718573245,0.892148343133293,
 0.38689014089008755,0.983867107849537,0.6922839292466927,0.26136640675116785,
 0.12259352443187166,0.9080792698674084,0.5828030194643804,
 0.17884408077465141,0.42642571314147626,0.653067041017079,0.3104704480979321,
 0.6256052030025503,0.07490423329221685,0.1763940351893717,0.4673427991558877,
 0.057906216868510274,0.04419153335556214,0.3800918003508583,
 0.8245615414444382,0.3509902802083673,0.22892870695415002,0.8695497941512399,
 0.5985368677581695,0.1456158356280859,0.6155520907933798,0.9232656399414658,
 0.3785106161705635,0.2882228192264783,0.14067704421013527,
 0.49904804746505427,0.0037673621618683484,0.941627541032673,
 0.7260250567371296,0.8817150090109824,0.14258228971073383,
 0.22693952583850896,0.004558247148014671,0.736700790337678,
 0.41531460562961964,0.9154589687529684,0.9124262051100409,0.5116149720595337,
 0.4679920003937298,0.7604003694699858,0.952392409988409,0.31693572558247507,
 0.4801707808293557,0.18822331673140935,0.9213064771053133,0.4042508022135676,
 0.5324870380504088,0.11316841432203306,0.40175168470630296,
 0.03932563465641781,0.6721954568035863,0.17518127578468512,
 0.7354250833040865,0.6683664084991806,0.6487608410248953,0.674907318138284,
 0.8069499829954653,0.5169856229186961,0.5252956295502025,0.3043359753688706,
 0.9050546684418335,0.412893315867924,0.9942224428624624,0.06851686180371741,
 0.9884050214717126,0.01949620651587991,0.43664716425776506,
 0.2684487196778287,0.28805259468064337,0.9641367582332607,
 0.44779073991462837,0.7919821595659932,0.4036483728987239,0.3443677554509954,
 0.3442183266165222,0.47355846061997675,0.6788060503547952,
 0.008834367104310403,0.8404544976686751,0.08842381631856278,
 0.3839411481458832,0.6808491856133944,0.864555121975674,0.35610840194861526,
 0.6467755490913516,0.9579194097589182,0.5684996302019699,0.7243356236321079,
 0.7208664654154854,0.6806485855007225,0.6534649085402939,0.37642980478201915,
 0.40523727202016935,0.1422746538226034,0.8851694294915347,0.7552286961875289,
 0.45722638974294916,0.510198939922913,0.7378382874363643,0.3022494834247236,
 0.772100070300402,0.13141438047781673,0.8587491201885862,
 0.012011666607929072,0.6593495101083682,0.9050256560200818,
 0.3387263416640671,0.9525724374470953,0.1532527260881631,0.6744666527301252,
 0.9116934547714539,0.14603753239908157,0.0329435519471144,
 0.30740884026186766,0.6088834738762208,0.8083231191707185,0.4552980424721945,
 0.24065042780204637,0.7921670950667937,0.8567551785333443,
 0.21780601418931178,0.9074688228524597,0.8155905437549121,
 0.47568648396159496,0.5539169110262495,0.8405694613132566,0.8601271753170905,
 0.5581739231774215,0.5686182435088976,0.8747872526453451,0.40502012711574054,
 0.30475644844355254,0.9050208641181772,0.21353689592419667,0.63887459817207,
 0.7584870366252998,0.200713425188786,0.115291995385012,0.39369451598150695,
 0.5491527133256568,0.5628189971854186,0.22394956818423506,0.8868196587446456,
 0.5907211130177994,0.1529167340426486,0.4779880957617859,0.23006067180605638,
 0.30188251950829936,0.3793502525871405,0.32788697127070277,
 0.6525842654645637,0.03851536629553376,0.8070851759217188,0.4676392612368838,
 0.9434914706530388,0.45645733604128447,0.6360608531326589,0.6063722164019831,
 0.40013043089564937,0.7348752865929085,0.8471918270522486,0.9022144147782585,
 0.4143598296553219,0.0650410166019666,0.558576680087574,0.3037601661960009,
 0.14879362779190064,0.17934362484512045,0.3828608037659915,
 0.056048654516514085,0.12263924980236518,0.7298816554338587,
 0.9534124495535368,0.07516671659939944,0.08730298091434363,
 0.8353727276102846,0.5489116182709751,0.47814897039350135,
 0.20383431100133842,0.67655775802483,0.9012541354675219,0.7691731566437227,
 0.8482216258090841,0.6727369889784814,0.4363686129034495,
 0.015692842843601063,0.8959216877544226,0.518154017196685,0.4544772270754067,
 0.11804659010258955,0.6344090159123208,0.7842775125441248,
 0.20161750959493696,0.16644072514019181,0.6215742044617455,
 0.7202411513675885,0.17964563177026582,0.949502884335225,
 0.0011722682866334155,0.4746474477993763,0.6138717429134876,
 0.35202555268488633,0.35370354371869794,0.5257620993428588,
 0.5072925693120273,1.6746191131999932e-4,0.9269444150331413,
 0.11449426692656273,0.46099873357996446,0.9157330357048272,
 0.20344815510316738,0.3421279113195913,0.3438797264698773,
 0.48215345336767435,0.4377342941284911,0.3000951382983965,0.5902899724931847,
 0.6887040501396126,0.6489041143214557,0.6558804998261005,0.2410173070918311,
 0.7408846288527879,0.8207988818952094,0.09982847947702866,0.552315513390573,
 0.6773084790941825,0.6710763855915054,0.6245568146815927,0.7218480235182739,
 0.4864644522861171,0.4636337658094154,0.7222567632697948,0.7667057978784604,
 0.42713614838920444,0.6369576673268599,0.8818230366029965,0.308656674692254,
 0.942163816049256,0.457211827790842,0.8643534286851992,0.7208338575852112,
 0.49541330360213176,0.6723453519555107,0.39074469334981243,0.924238380313088,
 0.7105452122672982,0.4994305542858718,0.9887893626922827,0.14248270837092858,
 0.2891925217387905,0.6748647171514428,0.11806359978058478,0.7780477968491897,
 0.35257882168636345,0.6407476720794,0.6749167828746672,0.4794435189401933,
 0.1549604772658486,0.07034685269349594,0.24327517951922073,
 0.40336543751219367,0.9891444265248569,0.14404395699374184,
 0.38116467166633594,0.781356035067501,0.18267676921824716,0.3064127526875684,
 0.9883659609848217,0.9900165821359461,0.8002302499423511,0.8162966163721161,
 0.7127072465963671,0.43343958973945196,0.7279445345690109,0.7865961696004995,
 0.7583869329619554,0.3606829345983722,0.43889617611151555,0.3183658756467761,
 0.4774726143679321,0.5270889461490968,0.9238712985849143,0.8184286719558794,
 0.1378104956935431,0.535744629213753,0.640270327373271,0.16600604427341603,
 0.2187246378726544,0.3241552998852717,0.6339829431621932,0.01795527222212745,
 0.0025462185507554302,0.48528952575769957,0.21404845167613307,
 0.860087188322644,0.041201277584362206,0.9137575793384695,
 0.08042835227018696,0.8323917049535218,0.5186469754478493,0.690964134351401,
 0.3810165677835513,0.5554029743560044,0.8118644489871828,0.1737904250221448,
 0.452614174935442,0.26454034380895186,0.6407461922630155,0.2934837748530351,
 0.03413370433768492,0.10775153470522847,0.5905096336366077,
 0.5669715226695743,0.6322602634615898,0.3004429427235895,0.6242282145116838,
 0.7226057156559296,0.8587408326282717,0.6126467349106757,0.9927490715207935,
 0.8983352872881284,0.17795380913783254,0.8740713052694237,0.3473762717418917,
 0.11912339540009531,0.565118546587521,0.7175747395391422,0.1498348879031136,
 0.5752946241922172,0.6190507374667296,0.6881058951502711,0.6816247234771022,
 0.7511497168796653,0.4382427344651305,0.8289923697145825,0.49948874517989084,
 0.9728045445172879,0.27757168809635946,0.07852286767320016,
 0.5525610249374016,0.37575857196496043,0.41004776772018026,
 0.44319208498279783,0.9422907263199964,0.5703974027257857,
 0.043029909953620105,0.6015354021650494,0.6656143789186095,
 0.5433252627986235,0.27557619855715043,0.978062701846523,
 0.026325845423236793,0.46785439397262996,0.7424636642665333,
 0.5808905218930721,0.9428137783509847,0.809676627377739,0.002611107602957019,
 0.4967966188649138,0.09707330604208653,0.8569968949122695,0.959481808945243,
 0.9490564156190954,0.8582198608265467,0.14389437236463665,
 0.13399333472520758,0.649283905924382,0.057942363645713146,
 0.06033524440901372,0.4700061443419594,0.19589560209831347,
 0.39056873727570285,0.404057763866428,0.0989836428784242,0.8695855592336948,
 0.41267956132803796,0.7521314851018651,0.6555873413182715,0.5872168358702086,
 0.890377977454626,0.47344403738148255,0.8839059403515038,0.9270377127375554,
 0.1652270234854578,0.05188981817058602,0.706996873267946,0.8081023137080778,
 0.25246544984757,0.08963931625927002,0.5653628037265286,0.9297930279586875,
 0.5321949362049472,0.05995366557268378,0.4951625790098694,0.4583401826064235,
 0.22980114836482746,0.4925346460259711,0.3731221432188354,0.1495918602777716,
 0.3247603657845207,0.8206918382663598,0.15908205000551712,0.9451026723903183,
 0.12217746856733525,0.2989166438945412,0.40009632357545954,
 0.5957983478033535,0.43789361613738853,0.17897280125282478,
 0.7381893024411619,0.9083443713354136,0.2869826694842559,0.6247731979329992,
 0.8114084142052921,0.5722561817761754,0.9250106721270103,0.32932831360017834,
 0.15803425320492082,0.21509085708781484,0.4657564214035639,0.717161312894216,
 0.9349203503488678,0.957397361055279,0.9152286563053724,0.7155480656561866,
 0.7308298341249231,0.29923779293316144,0.4518522881249515,
 0.04962836751757804,0.6779861157454624,0.4258074567356286,0.74216655857126,
 0.007979047835963815,0.22984496238440622,0.40417324286782663,
 0.11126603237298172,0.17695901316897544,0.49735050837477346,
 0.0013108142186762928,0.02384600885815258,0.2631590385981615,
 0.427056748984309,0.6021078445311512,0.9621151171270035,0.5600964418997235,
 0.7606982583725884,0.5743302770071453,0.6548218320511862,0.49475600911502926,
 0.45783054902196474,0.8715388960631016,0.1508533635869369,0.1455888677220708,
 0.8873119997413512,0.5337870964824788,0.285537316037888,0.23730503488152477,
 0.7985509127994727,0.18240072688773545,0.9450106510591787,
 0.10250113465601729,0.6002578262264913,0.9337300746510244,0.6624761287580058,
 0.9635056335920544,0.26008823035686957,0.2848894329837359,
 0.33991262087340446,0.651145849168887,0.17061791806152837,
 0.11945321778353357,0.198085241724425,0.4107895932118032,0.14792676849945463,
 0.9890198796365287,0.1114290610886347,0.28336345638866267,0.6834626194179285,
 0.00616480572145528,0.3718682999986006,0.5277214570491915,
 0.04968981121117555,0.49935626858404514,0.732829356334962,
 0.34789650146826157,0.9616511113064283,0.1575714359441004,0.9980365651035541,
 0.5791694015743395,0.03834401904905793,0.6202440315927243,0.9537871304638489,
 0.826309585403724,0.5798373164370065,0.10023080478378377,0.5747525271403771,
 0.12787879297876303,0.8797824711015313,0.9114327255104241,0.8225333993754891,
 0.8438779377964923,0.9379483642612427,0.49906113712933364,
 0.32803303750072443,0.9091095642783279,0.7511997617031262,
 0.06431620629329149,0.7879315545268849,0.01549649616784643,
 0.5628357101491299,0.5588046901650154,0.6034347736277273,
 0.026679088061703604,0.6703456707973254,0.84864547009576,0.4482341796339322,
 0.20991835602151665,0.8186199981919939,0.5938183641569506,0.5834658709559724,
 0.7891830494879696,0.94779836404667,0.8273021201203703,0.940000746585313,
 0.7646861414730295,0.7643791284713141,0.8789468393503045,0.09638021830521504,
 0.2425081930513403,0.29605999369070934,0.36067187233126485,
 0.1368318077140983,0.9245447035683892,0.341491342977731,0.6215484711191018,
 0.3232568208776849,0.6018780088713186,0.07089167411613595,0.3706885258212347,
 0.7307030138635339,0.12687662854457793,0.7180713193776117,0.774542133095044,
 0.5980636615801136,0.9447920737156634,0.20996148088520894,0.8785514053036971,
 0.2619925862315702,0.2299628671549716,0.4872587065459819,0.1878161514062887,
 0.8283305346806089,0.6561351912190331,0.7039419821402171,0.9411624760080552,
 0.08990717598767772,0.10832423856894458,0.9029585944027907,
 0.16968078447412882,0.10780765612544996,0.9360873256086877,
 0.5410901795039139,0.05332722139835144,0.10130559443597986,
 0.5888835268721035,0.44227563042713125,0.26294861114442036,
 0.9747175914032158,0.6316755047846856,0.9584485953799031,0.8128379933075225,
 0.36312647580909363,0.6528197339287338,0.4532493976618541,
 0.03196017663660633,0.07208854052984437,0.12305527224741919,
 0.5423369924795713,0.1324769882153798,0.8391410737667128,0.29481349525507416,
 0.5811319524094206,0.8401450743392069,0.6771615741483842,0.21520723543586806,
 0.1065033116553804,0.9838597146536222,0.9669847139177543,0.2124847651197037,
 0.6005078992604208,0.9511647915501417,0.7110777023747212,0.0192830670915638,
 0.029959881500514207,0.03874018612856911,0.2645885701167541,
 0.7337549787719915,0.5401292639598396,0.8989825997006395,0.0840645812155103,
 0.6323005004524954,0.026298609834087716,0.05244545449110172,
 0.2978017740407103,0.8145201031090799,0.026828881173974084,
 0.7316912434830886,0.36340010809967793,0.9012298673279161,0.9597801785201333,
 0.7460050577667999,0.5541559963422661,0.7927040745926509,0.675066455749946,
 0.49908821932582725,0.6453424092016538,0.27089860647773367,
 0.7129569575520942,0.6168186778550239,0.07837863857197702,0.2593751032446282,
 0.15670039736804386,0.2348106844654534,0.057894852402219144,
 0.8375874679210413,0.9302007895361356,0.2616155131406783,0.5919209336167819,
 0.31062897368374487,0.41955323351129226,0.525808089735211,0.9075618760947741,
 0.04308488629237672,0.3657504146784003,0.8330365283124318,0.6949473835628638,
 0.8654354758076399,0.4021364804157478,0.13536156064131466,0.9806591980062087,
 0.2975124718016924,0.4327908730774609,0.6003660376080663,0.9206953718843809,
 0.03665450771697243,0.35831437284723344,0.4029397121738548,
 0.6084021484546054,0.33698816543445864,0.5836872632022415,0.7381379325850631,
 0.31798278605429287,0.33995056108214383,0.8346135542532671,
 0.14637056266378445,0.09264070551341308,0.5182745399280541,
 0.1768508579237893,0.4298252107371603,0.28106454249338453,0.7789944830033356,
 0.6236574277712007,0.8769260558699162,0.46118134433772173,0.8486854277642427,
 0.6838461805799256,0.08312070811381367,0.7620187089205525,
 0.44143346444067644,0.31302129414781343,0.8036548500208747,
 0.8958659804710773,0.5170978677971866,0.23320617312788317,0.7418860556965969,
 0.04910885338096438,0.7476033798259766,0.6391137832448528,
 0.30181665439971084,0.1886823132774028,0.8800778958619835,0.7119902929234948,
 0.47844726398055903,0.29407722561322125,0.2649162702780615,
 0.9686608586931759,0.1746544539606354,0.6017551872087727,0.09529340330543101,
 0.01465283178954746,0.2792502649217784,0.3597963746632167,
 0.36017363328466945,0.9697517909833406,0.7882487742513712,
 0.46433197444635566,0.3487472978093884,0.45745033648131284,
 0.5745512228680703,0.5477924829154891,0.5446922069117778,0.11689861876714347,
 0.7298108287772452,0.2975538551378697,0.337549991199271,0.526142573096943,
 0.006746292829059586,0.19073670663331577,0.7925176405945616,
 0.2820262378667657,0.8715261128967499,0.4653789149376507,0.5817797527673799,
 0.45441584763722553,0.9527379889850021,0.33254352476763693,
 0.7955324746335605,0.06443373412453901,0.3465662848114731,0.7138150301109568,
 0.29734400752425005,0.7254924440871193,0.9016434809451245,
 0.48867805875358905,0.2254867610513115,0.8231684301044048,0.3829990711756863,
 0.9157464139020219,0.11791671762388622,0.4753294584654474,0.7206890296844982,
 0.10573184120557588,0.9908776772372508,0.7277722186364596,
 0.15646698606875942,0.9416101816156733,0.4242333474785066,0.7876148590844977,
 0.46545441797056253,0.6808442037132462,0.6569154050380308,
 0.18167794116677682,0.5958053640835232,0.5081980958194104,
 0.26671969361814474,0.2979442002911168,0.9871948350234198,
 0.07565864892271934,0.4731761086815458,0.17032541991153038,
 0.2554946103026574,0.022088266605034712,0.44312535571139067,
 0.07528474285401765,0.728859597423963,0.6678643229423331,0.8909744341610665,
 0.2409176975403351,0.2125899609981211,0.12988331810916687,
 0.26012822201214447,0.410342768099884,0.6424355938791619,
 0.061342909607989826,0.6062758288982203,0.295779251339114,0.8649717657221241,
 0.9974071215376054,0.636268276454947,0.9475855499249608,0.30118795182137625,
 0.8209543860131616,0.7332183311055931,0.42525316695325255,0.6218029122497025,
 0.7761529925637097,0.9448539832758728,0.1921346388624885,0.26024803151623566,
 0.8106575230167535,0.8751917403353415,0.3231866736359561,0.6493191403393274,
 0.5995144946815366,0.8775486048240879,0.08028172517710508,0.9957140594654934,
 0.9381513340351944,0.449832399604273,0.1360502745399531,0.6776076472418762,
 0.19597932852516498,0.27477261850750234,0.2836809278604795,
 0.020852638749226715,0.380163567664462,0.7094572669410963,0.7379098370463731,
 0.4764699346751373,0.9240816999709303,0.7514199680955015,0.6877066085854456,
 0.8233634836115684,0.038601787081452676,0.20056832484025588,
 0.05977956015101027,0.56477214608479,0.6435154996534005,0.7277302047163626,
 0.379087750859994,0.3158600629995949,0.996600091297668,0.32858083447501696,
 0.17232792981295553,0.8334126496081102,0.6040811448796956,0.3889648302077642,
 0.5642728254326925,0.8130200073250964,0.7716822140337879,0.9461535940908158,
 0.1603003870006675,0.15129450817065904,0.6878260502182253,0.6515427991899777,
 0.826643497246159,0.7131264391395067,0.3135981014091397,0.8344288754395397,
 0.7065106300431587,0.5830315047626891,0.4791580189909692,0.24449374423125547,
 0.07174572265410906,0.2407453777181343,0.721456678518676,0.9917600477235339,
 0.4910734824135442,0.421880276732742,0.3004318261137845,0.8407673399684245,
 0.22134893831195246,0.1526167926231502,0.717096240278717,0.7091326220353632,
 0.8052903883691472,0.17831777512760494,0.13438304167958637,0.975877881774514,
 0.6506343317240686,0.5948400347713689,0.6158015977466924,0.15173360245540957,
 0.5119101397233019,0.611396283936918,0.21064927385782184,0.1313097327674354,
 0.5120525112647255,0.6659963662920623,0.848843961195719,0.5100623401678136,
 0.373378068969918,0.25676964890141896,0.9421937663808906,0.8534804614719215,
 0.1455121438380106,0.12339467113334657,0.7400779895282698,0.4257641817088319,
 0.18702613756039133,0.47434087504339706,0.21789603334826635,
 0.5552218059801384,0.023609826014792334,0.23983181629296701,
 0.017310433464862163,0.8540146605348582,0.750446332161548,
 0.17931226889815166,0.9660886395856,0.08592761777046465,0.6746507498339758,
 0.46727609611336607,0.9128382277357145,0.9705385196732988,0.2999437117517807,
 0.3457036387547531,0.8595496464029321,0.9025423942582301,0.36847853319291357,
 0.9927885214726125,0.7405184939261091,0.3273958118569794,0.6260097866723097,
 0.139503454034534,0.8789795164598208,0.8366549451322466,0.3247092839679121,
 0.9119531796960821,0.7192210190344545,0.28192228700058986,0.9843030831976112,
 0.1954264472086633,0.5537973270202586,0.7556503032088168,0.25351270097740475,
 0.9922434076791757,0.6457155079154839,0.28721845239921007,0.7427608405554649,
 0.35490074505595814,0.9153813317900427,0.49680603303548565,
 0.8598237859764252,0.8087513972822262,0.47698312069285087,0.5372155824488057,
 0.8942096739489945,0.7309080995099044,0.5244408989346794,0.1334294715460702,
 0.2960746520006289,0.0679201178778468,0.5065538971463632,0.7021655393362165,
 0.8543358442309604,0.9431679168182312,0.7211634313555618,0.6388522358375085,
 0.7836162699978481,0.6427461362875566,0.3782488297004878,
 0.008309416253719615,0.07950473958098991,0.18282755142429596,
 0.7897802817008449,0.6396600835106058,0.8044733675227025,0.09939834514583112,
 0.1882185726539255,0.21708677903744045,0.37892979779323843,
 0.7402645097239204,0.8988728681944521,0.6308971886013546,0.31684146061241725,
 0.5608177558269464,0.4805378226788317,0.9673875513751876,0.1735017535373805,
 0.8513868051632969,0.3175161371702737,0.9525951229046948,0.4351890330779684,
 0.9300545783650132,0.1380133327137325,0.10270713537231102,0.8500874314783305,
 0.8544044013403431,0.26209554039047045,0.00857790649545731,
 0.9695862896220803,0.35888525688395356,0.1264381236724157,0.723094939780599,
 0.9267758848038764,0.7866580461387859,0.9896710364049396,0.4993381845233704,
 0.33322458424994383,0.6697763534866958,0.9799393495614273,0.7466999322334078,
 0.26101761103904186,0.9839097757035992,0.8486613256698113,0.4232726948162261,
 0.25083651027051157,0.3806852340408253,0.44760508734875304,
 0.6972704204145425,0.8588189891609315,0.4863239865048835,0.7220854604216032,
 0.3422063862832254,0.9059823148514261,0.4261310524215387,0.3055288873173152,
 0.24658866591069184,0.9758140821930075,0.6254967017292542,
 0.15565275002976975,0.3783681169116684,0.20735368972304813,
 0.42292183014327134,0.2309305669930305,0.14598068217018867,
 0.27172723597629256,0.10195311662888518,0.8196556904161771,
 0.5627694643606445,0.22650005725941103,0.8426286012341865,0.5960818045801398,
 0.09064437298919448,0.6367641169700768,0.14950313719580122,
 0.6323272141970175,0.7617094718589649,0.46250382896550335,0.6975190773187702,
 0.8738929100548396,0.8397027961435947,0.9400371790087173,0.5627253219950098,
 0.0655125870804194,0.7241277714440801,0.8867614713710608,0.593419668333466,
 0.16765283472075487,0.3859383168016314,0.13034775318062464,
 0.8897028179254354,0.35713677754216755,0.980882645806748,0.11977055395424097,
 0.5781364298873213,0.44846227156903185,0.24150206513057748,
 0.27947152190931535,0.9417642616293599,0.8836361171873856,0.1352690847315029,
 0.5402717884545105,0.7045755484826732,0.28599202093412945,0.9675283401091991,
 0.45233704367596694,0.45107662196083465,0.6630264689024687,
 0.5607845670963334,0.8396717719230632,0.7490105114541586,0.44775751421763044,
 0.3998148316936625,0.23496811931689254,0.19993800007832674,
 0.8416687379669846,0.9160482129559752,0.8293886171673963,0.45689928393685464,
 0.3629603012670285,0.18716441802349282,0.9855818879561787,
 0.42737025824396235,0.9839777719385641,0.7729686951716164,0.6629901683524275,
 0.025018228646560448,0.6525472367824471,0.5125891310566945,
 0.5589467314192288,0.8455504172808336,0.9363144609917612,0.3016941104269826,
 0.7323553466921354,0.05543651096011504,0.44517363303146484,
 0.10046648641218814,0.6103886481960208,0.7356260458847186,0.7097195507326522,
 0.6090047237252245,0.23460038716331377,0.23653345844728024,
 0.6341814648046369,0.7591717681417505,0.8821885186783494,0.6918934519254754,
 0.3193384835521912,0.4436121986813715,0.31831006200907097,0.4805370429250596,
 0.2983369474222335,0.5509492785828931,0.3246753303783899,0.609431425619739,
 0.008528297924727868,0.46790239704054726,0.12542124464096738,
 0.38602477672478264,0.022141558156447694,0.10409682323638536,
 0.35123584238831274,0.8657581727913465,0.21068423413736737,
 0.9475625996423697,0.5827291033704728,0.6388462845201291,0.3517490781555883,
 0.09919774542504345,0.9049408846383203,0.6213689748790071,0.9058691263285084,
 0.15078015612598872,0.33573341083098995,0.4209926400826892,
 0.8849864805247488,0.28751509424524246,0.16073876522112918,
 0.9614144514371483,0.2940975831668625,0.6185261036523398,0.9824639972455548,
 0.7618377744692711,0.5788821236600915,0.5953263025304459,0.9594401876345141,
 0.1927947071776691,0.9384573124934686,0.02530816787433232,
 0.09890629923732697,0.3922547626482159,0.9373796344785543,0.9940893119784255,
 0.48790200049016164,0.35414341930325943,0.6835845110719474,
 0.7430445349198942,0.18857267996057658,0.46024108787554774,
 0.2735540174147524,0.7862255606921431,0.13006865592573524,
 0.18043105441220453,0.08480161573183631,0.7168087412753367,
 0.09983610763061879,0.804725897870249,0.8394788486416361,0.01062845861780115,
 0.526394428040688,0.4161907731139417,0.9392435203527575,0.38805309884580486,
 0.15584400489551342,0.4327790538691627,0.4054954715435761,
 0.44899449424299476,0.38206767816290765,0.4791033208684613,
 0.9276282869204717,0.04731846664220041,0.47229943245324546,
 0.20277876968133723,0.07108489299371346,0.79413758629955,0.9261180731786272,
 0.12954237601470575,0.005649388437684699,0.9684875019470657,
 0.041289554625478786,0.6641193116831368,0.09206949190943492,
 0.034849690795193844,0.20172972797191224,0.8841892858448551,
 0.5437322974512068,0.8446681680686012,0.3780560884621058,0.7077581196487125,
 0.03841453868359079,4.34977681852855e-4,0.18135094100135707,
 0.004180789197897217,0.1283836647717942,0.1576973628382109,
 0.4877804576953232,0.19301920717582854,0.5494011386730795,0.2596766597497512,
 0.825901697117329,0.0756485637629587,0.6808678263224851,0.25555174740973297,
 0.35849181936952346,0.742767487077994,0.7024869260051884,0.12060311018896408,
 0.6855718069818577,0.14964679313646712,0.606697017624887,0.4387096679745114,
 0.03932567024314326,0.44130534782991315,0.16117307255816815,
 0.8265205888218488,0.09570078785573677,0.09281092341364328,0.917457366638914,
 0.8406526001869199,0.7934745091324533,0.2933702853616448,0.09028725172180307,
 0.8759440750797718,0.294803344741276,0.1584641280588296,0.9289020714489004,
 0.2630606817955423,0.49277831219183077,0.31681777277363576,
 0.6347997859887621,0.8383592567302887,0.6608232525604297,0.9371014300087624,
 0.6565771945878217,0.9171567366032776,0.35068870979025146,0.5289817996324638,
 0.9968569309869025,0.040487438783932794,0.6605630458634578,
 0.39072349494317904,0.9123960719969955,0.6521929316454415,0.6690376644414593,
 0.35373880409320213,0.7629886745498016,0.28089002938077856,
 0.09152353281345182,0.8417629146752408,0.8835125580143455,
 0.44897850827436936,0.5029726759307083,0.9747097514700627,
 0.04336191381503185,0.5527490868620828,0.8794931051190394,0.5127032725602035,
 0.005290087364992924,0.364317053956118,0.8782108816781152,0.9027933503720831,
 0.6049020415976637,0.7470960499849324,0.7794296402771178,0.3352717381974526,
 0.1957768747367291,0.8619717262803295,0.3385566006680736,0.27136623466851384,
 0.16846968481257274,0.8571403759003178,0.31727037874654906,
 0.20579603906465016,0.031816700754472516,0.7871406112069508,
 0.05575217957790235,0.8223056495839639,0.2151737884497089,0.9820613065272472,
 0.5691749095732648,0.515542039105478,0.25173891434991125,0.5610752805078011,
 0.2874804643001939,0.2690633697098519,0.29326043045971684,0.7160988698070991,
 0.577375530161666,0.6911421076693172,0.5927830266214896,0.3922733589355709,
 0.2927570880638338,0.8820860346200121,0.7988505784239485,0.9366702122222275,
 0.06997741790993528,0.9115622635584747,0.13595720468654104,
 0.008372078716314957,0.8483778389829184,0.3972488050977282,
 0.48844305257349185,0.26556286824695297,0.37581165997525734,
 0.8308630413313285,0.7167401496365627,0.1369289224217327,0.8936240103223909,
 0.10385474571679143,0.08365584889925226,0.1729435067290459,
 0.21154775238153156,0.18772295270367478,0.552207216919972,0.4373812753389813,
 0.6257636198956836,0.9248051455844054,0.9825048492510257,0.10559675174811456,
 0.18583297787896735,0.2140389989967677,0.7704252686836676,
 0.03898088279710754,0.4182395424503038,0.4420872134399473,0.8183475801843499,
 0.12895869877550947,0.43714743067369743,0.9539969379546814,0.863598107635415,
 0.3264011636763958,0.5136182969000196,0.33498456696530254,0.7748211340506259,
 0.6458582091301937,0.3959705500551205,0.003853337716477667,
 0.7946133495467778,0.9639695509794807,0.5699955638640033,0.46814186204634156,
 0.15844948099636724,0.12558898195939927,0.5442195208543152,
 0.15314877320988274,0.14131855414692984,0.8284875242571097,
 0.008458509876551612,0.10749471448804437,0.057778068863072574,
 0.9187924116513911,0.36324172749275396,0.03707304259165667,
 0.5195699213944028,0.6062218062853706,0.41622310150243735,0.9840459958025183,
 0.6759512371125334,0.8558252666091262,0.33987504553320447,0.9629886674383207,
 0.21771907660804246,0.1872297363346468,0.02769359691858586,
 0.4667046679303888,0.26859623911582553,0.9614606272230134,
 0.23357884891965863,0.8895662919959308,0.057097761053374496,
 0.0526056524094326,0.23910556172592035,0.43123981929512134,
 0.11298846984705913,0.3454677916465836,0.6297949790528181,0.2980440669119304,
 0.17201418043507344,0.3893691285796921,0.892999849490079,0.8512772067230281,
 0.12986293202663968,0.8017504768882596,0.7496148808546018,
 0.23445434142976707,0.587754511960468,0.0741896859611747,0.49149287956735965,
 0.19615753958609927,0.29227221657850244,0.8572425397395627,
 0.7202010736702291,0.20217954006645722,0.6473841715024325,0.930000561447719,
 0.8343078068470674,0.1177519625775666,0.21071961541022066,
 0.18324992857146072,0.9917625917309117,0.9083550711705751,0.3787800244563889,
 0.5973097039021563,0.06990453888804815,0.5033690056104596,0.5600506198282145,
 0.5858838057264565,0.6671488134970434,0.31797334544221567,0.8333976120761687,
 0.7883877029254432,0.2787798426537347,0.6912810461645432,0.8515954541682186,
 0.28725264632828773,0.3388545481785431,0.7244738040066017,
 0.10027876219582854,0.3232982187175363,0.9752897199253089,
 0.054083842417656514,0.4316404318586421,0.3206455816178142,
 0.7174411638657446,0.5122069041099436,0.9859619720224572,0.08678552236201276,
 0.16863521123345326,0.24585972943905776,0.6423258619529777,
 0.09415617076067184,0.5407656623832452,0.802826429138124,0.09497491226893728,
 0.038615936852217314,0.8799346947881566,0.5355168675949142,
 0.5690814422014439,0.4075973599867654,0.7609576769559201,0.43466834621392747,
 0.20923013203545193,0.2787767046219899,0.8187796730444646,0.3089582222852735,
 0.1770866415924961,0.9366992172578208,0.09358782829976664,0.5077549365376512,
 0.021480562699549455,0.7958576925423846,0.46332590081552694,
 0.34554779404464797,0.5820235086128144,0.5295497755233557,
 0.40671087876751166,0.9725309881916857,0.7636444943128279,0.4238005499473516,
 0.9192520427037442,0.5141086353964283,0.06625465175093881,0.8073135078259822,
 0.8459413285235118,0.43419777286007544,0.4565855697651593,
 0.24168891013662785,0.3903917774942002,0.08842830900771959,
 0.6595854182129992,0.2748258729052121,0.725832040176192,0.7738494523705766,
 0.8494537851398879,0.7518336482050828,0.9782176256373718,0.592078799144268,
 0.7195007819635609,0.9030195448064012,0.8290631599986013,0.26048685422768725,
 0.6480096540872711,0.9714915064140269,0.9971545465775389,0.3735312659505786,
 0.2467799159334474,0.42704710659120015,0.4047549383511517,0.9139746186577472,
 0.017062497152514844,0.41032710839162845,0.4193231849980511,
 0.19870684944221795,0.30373554520451473,0.6809073494520976,0.462638442325935,
 0.32879624657632167,0.15373235879408265,0.1660404885583151,
 0.18470558617130473,0.38775200188092707,0.007921361124728854,
 0.24996318187942523,0.3008060359012257,0.2490818558643798,0.7811102832754038,
 0.4176057341996944,0.6421387212594551,0.7857932584169864,0.42392081069388776,
 0.7999469421696103,0.32638771709737857,0.34280760049562486,
 0.3007209910600974,0.9229470477014683,0.9122239455428861,0.7887314382367675,
 0.988630883365483,0.7657877933415037,0.88128438171642,0.15531643475809365,
 0.27717855148423853,0.3850030132473218,0.15125156157135222,
 0.1788311310260733,0.5299757751027755,0.6583664449097323,0.4215495003101385,
 0.6817813990399779,0.4599560168831741,0.7451120915843242,0.5809884437038607,
 0.8579799983318611,0.28189632426360633,0.1540533851628667,0.2729644161790338,
 0.3981190103345382,0.675923553284948,0.405645742163245,0.1625485573265365,
 0.9487138616841351,0.083266741036356,0.8539093735241078,0.458974952436938,
 0.8291451285283727,0.9391006753998715,0.2714425494529543,0.6324100180871987,
 0.6747484076220019,0.5823786937180415,0.3048606444517916,0.5714496815987187,
 0.9098580925283839,0.7925834547717844,0.3850772400031026,0.372524092347282,
 0.44239306152941793,0.5738470857459852,0.6201726002527677,
 5.114032304053273e-4,0.14851994619135844,0.5710405518214431,
 0.8499700201732763,0.2321434367144325,0.44440964892220036,
 0.29853066475116696,0.9202103877100309,0.9176801308488008,0.8650760601990322,
 0.46974187207305884,0.6950718023381199,0.14053086116317537,
 0.4275427856780044,0.7659421968230659,0.9914222411339194,0.21865210808592872,
 0.5350385103329808,0.9096025490955789,0.3859249319981811,0.48824631976168453,
 0.3120735664391008,0.8409195619863793,0.9887137871677443,0.9253156630560573,
 0.04327773912352173,0.7682933758707705,0.09492130136432153,
 0.4052377477490472,0.6174756752046927,0.5688742499029222,0.2948306767541634,
 0.9434911572993963,0.8347095543342358,0.9896359066796964,0.338706047802922,
 0.7942354959253634,0.9912121017418835,0.5440006777205423,0.5730959585478528,
 0.5211171851025279,0.0787170480406093,0.9186728184842807,0.7193773998573546,
 0.3712731587691266,0.2891350628403274,0.7142377540627883,0.19628147620397174,
 0.9630229549565767,0.3583328653651441,0.3835676787281699,0.9380048274278268,
 0.49451337101314075,0.5475924279489957,0.6116724406580938,
 0.17125613981137966,0.8805220106362934,0.3010462602099746,
 0.13069838865749317,0.737119287970761,0.14103832897564927,0.5796623984277663,
 0.5423906184800056,0.818965164714783,0.3874687005465872,0.1975758922975589,
 0.05487706253569857,0.08680141880043912,0.4394575965653523,
 0.7786499424438634,0.32858397430806985,0.6501737675676375,0.5047266810648661,
 0.018324626498871788,0.41316334618520223,0.4310816420274629,
 0.19053711930824402,0.21283642468122455,0.7646551386199738,0.729591237573178,
 0.3709756325463752,0.5836015920651094,0.5826957248587586,0.2643383655304501,
 0.4383781413114475,0.015483539831270043,0.1002439349473302,
 0.21131107463230359,0.23588795883101643,0.8126558177480694,0.173730571393516,
 0.7895249899599398,0.7103850630477967,0.4104324695731494,0.8495506323766673,
 0.8717791965692022,0.7788955338404487,0.43157344078925153,
 0.23691089781136732,0.9679526106614691,0.7888875703847106,0.8602140979844509,
 0.4192266387276483,0.8740968347151675,0.033754739227734865,
 0.9204425898230172,0.5317485330447841,0.32509898725188,0.25454757685888385,
 0.6713216432941227,0.5165446123417716,0.12407191000147799,
 0.11686737322087037,0.6708441275869593,0.6272845229555328,0.3431965916632058,
 0.5556404127586579,0.05325165870260107,0.5309451529751235,0.525599485350986,
 0.9504869593465511,0.23900865809315996,0.34292586780543266,
 0.1861776427777233,0.8749273680132383,0.34084955359006974,0.6551768018239482,
 0.411193843767099,0.39328393791142346,0.42802480481273786,
 0.27523134871286525,0.6383177625716702,0.47044316181976376,
 0.5314372995139575,0.5344437002634853,0.23152912272185167,0.6726510889757408,
 0.19204759757067036,0.8985523882253964,0.8340805735051577,0.286494977688898,
 0.5416801955998409,0.535215652556722,0.12800083302942555,0.19099301662051393,
 0.12588888321836922,0.33347609243468035,0.2170759102714136,
 0.35129870376778105,0.40092946939786456,0.47130308717842795,
 0.18265198465342292,0.8817086564198144,0.6113495298866284,
 0.16823877596808068,0.7752753355948765,0.9949182634950666,
 0.32929594867457057,0.38809427451609313,0.8841744907187521,
 0.9278485863742941,0.4766894414297149,0.9518116608219915,0.6113539676425629,
 0.2079266299346494,0.8835596991912575,0.7738427106391254,0.6575831266098465,
 0.6384375142605094,0.5754295553126866,0.26064115643081676,0.3207704329307176,
 0.8471391261341907,0.300562047053193,0.3414746747969789,0.20730633295089218,
 0.7656058903945111,0.37563654651370204,0.04122538239417384,
 0.8129393922482121,0.13838861801984204,0.8044729511807645,0.475774451376728,
 0.409597526812959,0.8616116886901488,0.5349344757359606,0.9818950704061511,
 0.7574534634842524,0.35095300712555344,0.5891807107531539,
 0.47346992596534976,0.23576154623533196,0.7534771565424975,
 0.25847746416903394,0.9791759303125993,0.6213429743855218,0.7607233564554515,
 0.529525267996317,0.4588171328963221,0.6505029412720799,0.9075661309338624,
 0.9445662684439028,0.7839513779759512,0.3351934439140374,0.8985755709401011,
 0.3793435212159687,0.47020207377625123,0.8206108616014949,
 0.44977383483636546,0.7550228981658363,0.990339551000141,0.21736563088056848,
 0.1399043742451478,0.34501727597944043,0.6037179423508814,
 0.15677658332282862,0.550529620581978,0.8325708909305037,0.02053253883438866,
 0.9160350082716124,0.6675502005406525,0.19739056701033264,0.4500440094698175,
 0.8072546667197226,0.8334878482953101,0.288043330586527,0.899921997326678,
 0.13987745822061426,0.19999808971827548,0.13537258313592926,
 0.764561610399745,0.21130359519557007,0.20861768489376964,0.8397302174376654,
 0.5327973310814209,0.5969628737870072,0.06194815287094013,0.7913405813971818,
 0.5160597116460767,0.3902168458586168,0.9637806387394103,0.685681777281568,
 0.4644865880005623,0.11007768365578152,0.7053863070400401,0.50259688863113,
 0.9542311238516307,0.8706781489267832,0.3113050243917931,0.9994183901304539,
 0.549929220306653,0.30327419762483165,0.6267219534904541,0.13837428598828283,
 0.42579067810386095,0.9526386722715289,0.7088287596674746,0.7923177629826341,
 0.9320550089432882,0.47956937174697356,0.3246803765191806,
 0.13897297394894503,0.5298558556424061,0.911796312889306,0.9104640117075032,
 0.49434696814071855,0.7708046956562524,0.19950752981967534,
 0.2125882820911672,0.17112428384801626,0.4746694471923182,0.5449264717125244,
 0.23808329227002167,0.2537008983973663,0.7706452914184154,0.9019551304200856,
 0.20606631703454903,0.10435895178543075,0.9850484131032586,
 0.9382129951499577,0.9317843903368298,0.33358557766743957,0.5080024818286426,
 0.49997099839089765,0.07743983197010662,0.6528903570258772,
 0.5987378457355552,0.6002103193659072,0.16019717599882555,0.2572856454301299,
 0.9851214436445073,0.2998092829155402,0.3930392691051533,0.4950630128146405,
 0.6662710553346082,0.3840293714433817,0.6037202122538176,0.23181830936880565,
 0.6048375840489258,0.3128478142708603,0.48458384329555637,0.9930717109372034,
 0.7514049096506451,0.6525225477959385,0.059666601016396204,
 0.3947650599955793,0.9352266033498224,0.7678165412305211,0.5683176100217271,
 0.2504805731467806,0.5001595967817606,0.38960471580272493,0.4025674960905882,
 0.020161498012316814,0.9732370861189886,0.04620038077258659,
 0.43835555693687445,0.4761861830662012,0.03135774290906468,
 0.9768241853311147,0.6782333867126354,0.10587451372875467,0.5675462792209407,
 0.1576127462557365,0.12729423648007943,0.3207389658876675,0.8867716385952732,
 0.460677766313653,0.5821905763733155,0.7934093161481777,0.14018198449363706,
 0.5769211341278339,0.40969203752171646,0.1498684035003749,0.3532189693736949,
 0.7598393955677919,0.1090041127807656,0.0011700693490412917,
 0.43552559521624334,0.533765860538446,0.9738930597481212,0.2931686804937573,
 0.31877317567977126,0.4298018056018592,0.385433113924609,0.9581630558932497,
 0.7658100439224313,0.9009238621276022,0.7260923873654208,0.6712469557119322,
 0.9157782147361315,0.12556129462694898,0.6503368074152595,0.7527431581838543,
 0.6680195052331324,0.5773544890333503,0.3362506565307637,0.42704082191378934,
 0.014216281282056209,0.34394147000285613,0.01575806238368105,
 0.03653929903533004,0.4032564785338365,0.3124340230432904,0.0823671568169243,
 0.27275151405905407,0.8803425135831673,0.9720731572102846,0.9991347066982488,
 0.8234810118972226,0.4615878165055626,0.5558477469888574,0.12979958779683864,
 0.8807600916106878,0.6311541677339537,0.08594934299518597,0.3046110092850218,
 0.7858735524917785,0.05540308820410034,0.48690528772749064,
 0.8341307136814347,0.5719815899274344,0.6209003121206631,0.5911107912368455,
 0.820144212127175,0.3644977054349413,0.16366243208080078,0.5984157982569874,
 0.6084182404384137,0.8613245415492824,0.5729388491766542,0.8819444703852277,
 0.6696244191360783,0.13732232458240456,0.5178919246249769,0.4635410612830255,
 0.9652244566504257,0.4922998728257143,0.5831443443325791,0.7428533717625897,
 0.34713425536234677,0.17099380117196228,0.25920747945547584,
 0.05062799360267767,0.819230058699356,0.8033603108403412,0.6546903694429822,
 0.12326637491717185,0.6727870466243586,0.5169510220881015,0.7301216605344597,
 0.6851666053937403,0.702533463445883,0.9784040393326059,0.8232962197045312,
 0.9344466100911929,0.6549269914540349,0.8887601483383616,0.2129006160327982,
 0.24269141161047447,0.764064005829016,0.7306587800490587,
 0.006939900132368004,0.4134825687923086,0.37759594665511165,
 0.5218031592066101,0.7814802220080026,0.7787952365314221,0.18917245888151546,
 0.4780783746493511,0.12327862883188956,0.6511472313854124,0.2551742405195816,
 0.06552857773107301,0.5622208259491086,0.20916670920165958,
 0.6007375215296128,0.305193766763699,0.9522803589970572,0.8333006315888651,
 0.8547056528655878,0.3278065739819066,0.052050903643700686,
 0.6269044417521783,0.38417229784234386,0.8064434665302261,
 0.13643326884875184,0.6327592362765125,0.4689725614181881,0.9625388864336614,
 0.3022074449882478,0.037238630117114035,0.4537019592784328,
 0.7200194981413767,0.694581273970769,0.7183573404472031,0.49447740026297504,
 0.14004040768645065,0.5853301536149881,0.23405012430680783,0.558921403277242,
 0.8927995691496046,0.5619113577728053,0.714554949244921,0.7124522272136993,
 0.034215481490666666,0.4132178099038215,0.7689309757526205,
 0.7782088172921681,0.24017509258404002,0.5573176295528337,
 0.37341377242362506,0.44851421382537493,0.02846142271113461,
 0.8910817457270243,0.660844227955028,0.04788133404385875,0.8690768766659445,
 0.7415137790051648,0.1888181784072568,0.15210508279957768,
 0.30000083287913226,0.6561049624497646,0.6280549380139018,
 0.24810302000952655,0.6654378161990886,0.1579829179679818,0.6600682533722007,
 0.7355157849605601,0.562293474280297,0.9725368712143142,0.888840239303465,
 0.3874870991274201,0.68266992364719,0.24792773443159377,0.7542457466647964,
 0.36003358219883175,0.42834698876688826,0.22016462532032788,
 0.7445415654591389,0.33100480984278424,0.4260833229486992,0.0835668960806446,
 0.6141793395667587,0.022754852589301433,0.11938245230075273,
 0.8050491652450702,0.9871372155258639,0.8208802441757265,0.16171045281384455,
 0.43643296129737674,0.4693608351348746,0.7719368664259034,
 0.40063993875855164,0.21032351351299516,0.49197686326557855,
 0.37783401606137557,0.712245070672308,0.986276871286229,0.7896998722429888,
 0.8251956778768054,0.7950673356104309,0.32103196384527954,0.6046036136423822,
 0.27504387270833575,0.41206253166044393,0.7948017547449115,
 0.6844842552940437,0.9902812958465002,0.07493938983158055,0.7189189223626237,
 0.8354522782370605,0.25910334655975376,0.5425321882583274,
 0.20899668893469814,0.9953578500952065,0.09530525784751764,
 0.8795319184435358,0.8935478861979077,0.6626358522314687,0.07589978715388834,
 0.8617718459197895,0.9148198296330042,0.4368117812056954,0.7360652354807521,
 0.8904899716181053,0.9458573013843345,0.4335141395569466,0.4908471270246152,
 0.7964324682689501,0.6724335587604442,0.7311416692097653,0.4383917093897237,
 0.28678101043479864,0.6437980251506837,0.7796787803175145,
 0.34312076980573347,0.9343724809322543,0.2780987605928261,0.9920303436345698,
 0.37262976037794615,0.8161639767390827,0.2943962371489448,0.6514465175593044,
 0.310999912757832,0.3569404268277657,0.32967218242322116,0.43502150749508905,
 0.3373555299854262,0.5907164285288843,0.5427879504458625,0.11988509655124113,
 0.17325804269693035,0.04592856586384986,0.05218343272248804,
 0.5036538672847418,0.337905572211875,0.6146275407224535,0.30160573480782504,
 0.010517991310975194,0.4029013860658194,0.17138236333797574,
 0.2320082457927296,0.8191448813664761,0.03097455752345224,0.826945457603024,
 0.873241808452145,0.1068050692498067,0.18536556276315175,0.33385718843020207,
 0.4701806164983485,0.9275635946588546,0.5236565888459404,0.28149988758606104,
 0.5875016998047308,0.9865333002752031,0.2297640689023156,0.47577967138939536,
 0.45018480846008735,0.3892844470091099,0.31806312291825933,
 0.21591842362797597,0.8440403335144033,0.200761884932111,0.14759069647917766,
 0.9741807539980298,0.3826478105911769,0.42434042494831703,0.8977823777634978,
 0.24213885109329625,0.5481015234052363,0.14007836676188812,
 0.6450867059243793,0.9101424623288279,0.7997106986554907,0.4458013260711152,
 0.9346460254747911,0.8296242412961856,0.061490892982894896,
 0.6645800519605773,0.09551996818130726,0.6961528678016012,
 0.053159137806471746,0.8024065047702466,0.8786859380920222,
 0.8912716593435667,0.6565838998525061,0.9262617837147347,0.27860276579771304,
 0.7079255975209838,0.6154087786811282,0.1437427820001015,0.6274638320683847,
 0.36513852987426476,0.564157706375138,0.4507417832031053,0.09101638238517484,
 0.12002952302008862,0.2192809027852095,0.04403417894782302,
 0.04906114420777352,0.7986072157754494,0.7297904009239715,0.9327105168434694,
 0.5736096811927983,0.3272251909649818,0.12872415764216516,0.5783058242376168,
 0.3419068310836668,0.8302480222886689,0.9942151444075741,0.5117944078626749,
 0.9451330131839751,0.15497999093482817,0.04965921601078327,0.519141501925737,
 0.6531551803836625,0.34373892046518983,0.4954277394139833,
 0.33161939446483846,0.5022007720964005,0.5390220862854076,0.7875414376879507,
 0.14506576612241817,0.04236847259121679,0.8597736409225695,
 0.7740977429070368,0.4649497997128915,0.589976257333616,0.39184873432047707,
 0.42971420470318067,0.8879698645821607,0.5889850453665438,0.8731391320035038,
 0.8237903238739397,0.07634806853224463,0.6096031415966067,
 0.22743847119189453,0.4589129926785579,0.05754519328364238,0.566372074688144,
 0.44112005896674966,0.24057673906212496,0.8231370752991397,
 0.7532546758339532,0.9645585500567484,0.14227281386468327,
 0.01680551728120583,0.2900239335503514,0.7855580251819125,0.7521890249959016,
 0.27471063823491015,0.8710679328419568,0.0764267708287849,0.6200263298553347,
 0.45902304288868157,0.3921494732420757,0.35729430252938643,
 0.9084373134102963,0.43716399189822097,0.014567023022909908,
 0.18906475958358715,0.14208751834705036,0.7379273949219793,
 0.7622322509062863,0.4291753227062144,0.16580744815545478,0.7885723931471653,
 0.6866681349635486,0.8361018829176569,0.21759971504675435,0.2334857434416282,
 0.25957388654895,0.35689570408056626,0.2603987550032053,0.35219874168505316,
 0.5466460914309174,0.6862147264333273,0.5849029451360237,0.4358984853708988,
 0.9429739083689376,0.3787606565083351,0.9794022842357789,0.08180824179778257,
 0.06870187316850296,0.11495468700779576,0.8063121135248635,
 0.8438105692819978,0.3288724484480934,0.1831908189287721,0.5917619754918212,
 0.9121706887029731,0.21650934978822767,0.3342923635046965,0.3017455430962179,
 0.018827382025410233,0.1617581589385484,0.30435438044194973,
 0.3879896517455811,0.13466575054281993,0.34661991840716255,
 0.8313667116335528,0.5169256161212435,0.40846127258207243,
 0.051916308102526676,0.4409432460202959,0.9885162382373053,
 0.3894179014864436,0.4825830020996408,0.1393282317118758,0.6914310367312405,
 0.21593615964694846,0.5187601558428482,0.02637537846223248,
 0.4949961690168668,0.4295952904061161,6.404381758129407e-4,
 0.9307741867260432,0.2863198092370427,0.5810289192613053,0.57769820379774,
 0.8674557804087305,0.5959313102679005,0.7654052158521442,0.38784254131245455,
 0.37234445127638505,0.8717426126105355,0.2884303513220511,0.2440242756985156,
 0.9147431126327151,0.7830533892127034,0.2248430215324999,0.4733671792323252,
 0.1386188863965514,0.2876235220664569,0.9931291910077417,0.3894993192344095,
 0.9140501770909208,0.025536255047175715,0.3393916752553022,
 0.1359839831928753,0.17551513427798948,0.2655152641026408,0.6295154516879473,
 0.2034780464699517,0.09273921620145464,0.6272775161544014,0.9252075409647762,
 0.27775121860725993,0.957264881189074,0.7191668479569385,0.6713043435772788,
 0.04589692957201108,0.6045060377875132,0.7175209583985698,0.6654801181008198,
 0.6475790086215032,0.21190227843252551,0.5913217046523417,
 0.49926572525664814,0.36491122203117243,0.22577331743180562,
 0.22353680794838815,0.8352021491540739,0.3991200491127953,
 0.46174761201033354,0.7054567850030782,0.8408126494449879,0.9654426500406906,
 0.9778155223835324,0.18485623618245195,0.2582712629537126,
 0.28994482092395213,0.8363358894479251,0.03927663471669718,0.571399203629365,
 0.38369472113421377,0.8665947843632135,0.4640121343153226,0.2252521455130142,
 0.32837942862188807,0.641122097713976,0.8172767026337395,0.2975940775905451,
 0.7119349820235182,0.48631077808843715,0.918106883501167,0.8447896953101514,
 0.31278706471257556,0.7878510549862204,0.6581314196509481,0.6143665654093466,
 0.3038968487045508,0.48534416600946584,0.9044038977017885,0.7688770685524094,
 0.3974856024958833,0.852969758127713,0.2959725696069101,0.4678697884204255,
 0.572021585030613,0.6952814666314756,0.8909370514425126,0.7057842826282076,
 0.5428746780350727,0.7804095610552224,0.21339170463828794,0.8187416356370079,
 0.8987735654301602,0.12937718164465328,0.828558552064742,
 0.025040840541855958,0.9579701444963814,0.9706186645279707,
 0.9215409605946769,0.10650764083418873,0.4029814477406314,
 0.09972333173360892,0.4757506352541747,0.7549329174499809,
 0.09500913398734379,0.7126145611980514,0.25876251252371407,
 0.5568216830711473,0.6496457920078974,0.19803129865561853,0.7200545784828452,
 0.6095714507056091,0.2661747708746619,0.5069663711808974,0.9873353320025884,
 0.7089807429638624,0.6954360988193182,0.6783508666654691,
 0.038532291448754585,0.25614280158016,0.19611765090777633,0.8696164392310493,
 0.46803877438342545,0.629274386812184,0.09757134889302455,0.4669751653395756,
 0.47268622437208485,0.5555681203569296,0.6836848573804879,0.8281924611061174,
 0.42751762039128727,0.9971772106719199,0.7429575377370893,0.78194690649715,
 0.5256952048779784,0.20421850125371743,0.0753635447287001,
 0.12843297089345174,0.7238329090518602,0.037753384664137246,
 0.07414238472703161,0.8948290112971863,0.8112494334634206,0.3902136219295217,
 0.6694664419276759,0.5377125089581152,0.6860891684420861,0.740818075153943,
 0.954533220627675,0.8642828433360464,0.2204368471557706,0.4203973720187236,
 0.19176069037769383,0.8816630908042742,0.6132639197383593,
 0.025019192025410142,0.6135350195623429,0.07926647671588638,
 0.8874217065119436,0.46499990223987075,0.14522854032818322,
 0.6039725412232047,0.3487208657617481,0.8384113022241388,0.9778010318568101,
 0.19418666945915075,0.3994662122071295,0.6081636261646735,0.5371108576386067,
 0.9608251298867704,0.7213031101346615,0.26878926272752723,0.920539174194419,
 0.004439282773327613,0.4673248682595462,0.28024493533208084,
 0.9929910615274135,0.7114958821831602,0.7036745360091674,0.5439350069023249,
 0.26728633011676384,0.548732581492212,0.383827971341004,0.728117515233779,
 0.4216748553333529,0.5100194897386721,0.804058837508728,0.18415110005443214,
 0.5197376623620387,0.1628122983924285,0.4474377737955133,0.5221897641105973,
 0.08410773483529266,0.8719959393761334,0.8872404361988434,0.0790309526045575,
 0.5753998921610424,0.28811173305903914,0.030596316301043758,
 0.013389595531672294,0.02536564459867452,0.654966622061911,
 0.14150961670182338,0.9372652776308037,0.9485318069597536,0.2696054067237871,
 0.7229061927537366,0.9634009078497132,0.2650193540958862,0.7008330407745584,
 0.5851709808435459,0.8578876940566713,0.8170517656200911,0.2489374726818182,
 0.9454493151818739,0.516016801073338,0.28159007262105706,0.06044692019627518,
 0.8081270265869602,0.4061137302580671,0.9176295908735919,0.48296026128586467,
 0.4859105975285586,0.1630367365607286,0.8374092927454356,0.06915353303446414,
 0.6573976462829423,0.7116723207221237,0.2544321052739902,0.5041694362963578,
 0.6513624684771803,0.8156453367436022,0.8983889063112299,0.8343745169857726,
 0.6232140918450951,0.6906409338874288,0.32758728505652346,0.4783593939093982,
 0.3505596102511055,0.9511782827640246,0.21602633707705188,
 0.33484854070783376,0.037738684007109224,0.9498037482533267,
 0.8431740847837655,0.5856117977261777,0.8705559642867702,0.5227417133200887,
 0.43117645707892416,0.777703675692043,0.8607514400259177,0.7712581870006069,
 0.8454567797333811,0.8873828067469732,0.4188994931031834,0.7420710144562432,
 0.5654478958470025,0.7961621312914126,0.2577584308526034,0.7751226730997722,
 0.5694785345938514,0.06295735647664114,0.6316369998533844,
 0.27402759056914094,0.13254165919373617,0.3259834217797194,
 0.8431637467300469,0.27549498432039665,0.8267365256896111,0.6178915716485224,
 0.3072594538269092,0.7408140502492806,0.1216606209865918,0.28681444043714954,
 0.4261349973975168,0.801026145451117,0.2978579369900809,0.9652317650097733,
 0.37335243403460006,0.9860833283472229,0.28540511830594506,
 0.38391443589994034,0.6078439207406018,0.40447747354714847,
 0.3023897115150974,0.2174037946843782,0.7708214085241046,0.24411441667896683,
 0.38853638488969033,0.27924648919630424,0.003061880273266371,
 0.6812081034724442,0.8094552492609564,0.18073592552182827,0.8586651347524549,
 0.7212055289797481,0.1442975222755266,0.5154178811328861,0.661945310535759,
 0.2066094437946251,0.6491053231294881,0.9227554913344391,0.5176375676577163,
 0.7216263551816916,0.02606850853053777,0.1010485651787435,
 0.43860851262993605,0.9116781079192106,0.9648992019978432,0.756150509556956,
 0.6544586497103244,0.21604743865879072,0.051201366702812834,
 0.7152133170300861,0.28673689812294856,0.2468616981570282,0.8512117637422061,
 0.17306738422246504,0.1554354826662132,0.941089918203915,0.8581896930738083,
 0.2890722365616074,0.6488508663849538,0.540395116438852,0.43654525550907286,
 0.6239537833949733,0.665954419387395,0.6669092846456889,0.7700215136805396,
 0.4904163762763505,0.27585630550728446,0.5557763200373276,0.3086511855895566,
 0.6588126297145669,0.16435828817799603,0.030625393946721924,
 0.45975080131988777,0.727348712826778,0.0728016570087856,0.9763227036524936,
 0.6592387481047058,0.23547946575089973,0.9141678017355054,0.6519926034686202,
 0.7211794039155351,0.04201486517626796,0.6200383007994756,0.6921960914613152,
 0.3297247096834438,0.2658523778547569,0.8646384176828092,0.804667167245429,
 0.22675249295590105,0.0018532344123740607,0.16414237363514694,
 0.8547119659495201,0.8775097647318433,0.027050385591681825,
 0.9607461731040673,0.584767394638609,0.4937552342215038,0.08124737943712979,
 0.35977116570931267,0.8815271951651146,0.7117475728679318,0.435570590194574,
 0.11664742496479763,0.9994397072390488,0.7032578355016674,0.4787783287850671,
 0.5031284693898331,0.934835537314278,0.5969965230709854,0.9368297161414783,
 0.8571739441231174,0.6723358299879902,0.6853057306498709,0.4975401331936856,
 0.5480976591949398,0.35067629396128475,0.18093983833234795,
 0.1498191988318246,0.8299558115779666,0.8471577186145527,0.7002011777661149,
 0.5529965407724321,0.5774500956693566,0.14256859778066633,
 0.05134887057157478,0.34836214207576865,0.1201859008193682,
 0.3468510346482132,0.8294814574245708,0.7156473702966364,0.5065934301746235,
 0.8284381510985581,0.48642450812454274,0.7819646125411349,0.3873788883234488,
 0.3420816713684077,0.08808015978876282,0.4844225966379314,
 0.45792514764659886,0.4037391795553651,0.6734676019912991,0.1225408861774252,
 0.9134901803191654,0.4013408817202183,0.9434741151489399,0.16660553194715966,
 0.7688085826040502,0.7721385418752241,0.5090115521299096,
 0.007810547299371029,0.32712442037606804,0.07219887745342202,
 0.2230221565146383,0.7449388413713649,0.6741585256282672,0.3756565182951176,
 0.6084214357617157,0.10529002601358373,0.9087958541347122,0.6406856878340487,
 0.21648629509296802,0.3438226379795962,0.5458970387011977,0.8861358709278006,
 0.4058444482916336,0.9654905375369317,0.5258447861878123,0.4408196088807833,
 0.9551402795217061,0.25177419036614535,0.2010581611869987,0.6495060623145817,
 0.26617025560738905,0.7778362966205588,0.5577491465425566,0.3398346523034803,
 0.7917659472181501,0.7865906904015629,0.41454173113699655,
 0.13633288025589674,0.646287564417803,0.3943749073184013,0.3201793975349523,
 0.38536864258774806,0.4723267682763962,0.3207083029375203,0.4440018820784679,
 0.5674662880590019,0.7004017965034256,0.21058026203744973,
 0.23797806945903333,0.941224310163437,0.15006025735340156,0.9608544979417584,
 0.5061376146004269,0.35179885336737315,0.4043414467346319,
 0.13429575496743418,0.9725979379908182,0.39876776194070207,
 0.9079558624300462,0.388521279044165,0.6002984145880533,0.7494288067118555,
 0.8366676435283311,0.39310382191389937,0.7221680715104322,0.5275316282185768,
 0.2376997073217847,0.6609206671144717,0.5390769129893842,0.3809813186374654,
 0.09932639047014047,0.5662786826924873,0.42768353644286294,
 0.7906431266502709,0.345552544429818,1.708452357697121e-4,0.2493821338242409,
 0.1566502313549114,0.2618733691346552,0.1839325970356801,0.9966734109330386,
 0.7450639523875321,0.22030222456550586,0.7312399391218914,0.0248900157129609,
 0.15063104375541303,0.23030052802216905,0.39886596447218425,
 0.9482952164103688,0.8170236743453345,0.22606745308598142,0.0309129491965332,
 0.24107109816052275,0.45547658042105377,0.3311395359721544,
 0.28634159863618724,0.9011039600950197,0.549485553516631,0.1934096199339157,
 0.26101900724014404,0.66193664243833,0.20711225414461176,0.9898046824332333,
 8.856200610166676e-4,0.5837120405691308,0.5897795487937385,
 0.7911510137959413,0.8725757534130503,0.823891943540727,0.6007450864829912,
 0.9692903002403362,0.613340890439084,0.6420219753069072,0.6538482523512023,
 0.8505868877175433,0.6223875358113806,0.16491140414841843,
 0.03640717741053101,0.35464535195669444,0.2504507521230487,
 0.3931965286017931,0.5117791305274748,0.46422420502927386,0.2496448176117998,
 0.26403817508394345,0.7324563864989752,0.7374435004655115,0.6374040439991486,
 0.9086976222450884,0.547794234307915,0.7137463189365552,0.7532609539966757,
 0.7004515410514262,0.5713386799695621,0.1797163222671435,0.5411028941062288,
 0.5728370079889366,0.7002002718295852,0.4357256175485714,0.01804398091857684,
 0.4095779066883747,0.20219721155630377,0.23699441722176195,
 0.6940022702564821,0.07210910510404678,0.2719010226775179,0.4706623221396752,
 0.8949605397888115,0.4152071460483634,0.27300023561172104,0.6182061893078551,
 0.15558177173573795,0.6681211698517411,0.4932880922246591,
 0.42042893996759034,0.5991054945461031,0.06471319383828644,
 0.48896568211688596,0.31988648801592756,0.32740010949236964,
 0.7165684376351451,0.22197601295238323,0.269081277996013,0.5509777358475707,
 0.5339301648732384,0.3951302734697405,0.37652171025157977,0.7528179552017787,
 0.43129252277481966,0.3381988427340854,0.0357166607127688,
 0.44923460001302773,0.19461099444081187,0.6575014456202963,
 0.18718905285946963,0.8351063452046182,0.9096753641469776,
 0.08236994109796425,0.4228117890644185,0.566098892746034,0.6735781275170334,
 0.31962498610097345,0.2948288098248347,0.10546983455683945,
 0.3050696455039569,0.35191617284334054,0.11774537579232269,
 0.5554068013335054,0.10842420404954722,0.41188489689806773,
 0.48880900661432536,0.48927236693958687,0.8321140754059582,
 0.006060963246512285,0.2926712033561265,0.39009855280057026,
 0.24379550183160514,0.2807479042819745,0.1333485606070811,0.4013895056709801,
 0.4370405497013147,0.623060041313376,0.7757637594130644,0.6172648139823819,
 0.6279659416954637,0.004268126346674284,0.719285266851828,0.5129064269463735,
 0.6313119639150541,0.1436342564481602,0.6488234330915521,0.1344259231135595,
 0.01552548959413591,0.5221636492524735,0.4946979972833613,0.4526346359309805,
 0.5973772240664607,0.9222927096878553,0.818330422925373,0.09641540791674308,
 0.38918054643607436,0.07302747940431231,0.5218279145970905,
 0.8905288118897633,0.8526134737787344,0.7085449014957,0.14161010656938455,
 0.2169070728750585,0.9156918872993682,0.6596447107660822,
 0.010315969452412754,0.5906945622186357,0.30106473952188917,
 0.10912178476607148,0.24446224304821496,0.6282896038549469,
 0.4699175018369012,0.575264001196056,0.0031374492441683666,
 0.6898862541450441,0.6585970773262224,0.6944211680921627,0.28886922521979375,
 0.6940950815153918,0.5339622327808,0.31281549037269907,0.42772100904374377,
 0.07820903940930313,0.02888425308507958,0.3112690349422793,
 0.7473708339679419,0.8216328746769941,0.09090961803835518,0.6245410908538225,
 0.1981026487084674,0.5872060639331631,0.09454066819194351,0.0719242129654517,
 0.048721257806791485,0.9402247383120099,0.7670191021485528,
 0.6543101376668772,0.5411221762141092,0.8230056602969273,0.8493988494937907,
 0.4828228846171544,0.4123892353451901,0.9124309997766757,0.5593593324161031,
 0.9667151342642426,0.057087900998603835,0.9300266041916505,
 0.14281826204117587,0.5843813211731812,0.46478554615699086,
 0.9062528124799206,0.7275907188117099,0.4676356768774125,0.3955770673577683,
 0.31509161547054976,0.85652858532955,0.7688795626399167,
 0.0012083150176059876,0.5879441740936425,0.7113887183955395,
 0.7059529897012682,0.7164511900257597,0.42802011785178495,
 0.20731935782670874,0.7166025794279338,0.30968342433529583,
 0.5042180892692206,0.8477972353425677,0.7257567397105256,0.48633154687507385,
 0.8332976254731733,0.43299613169178774,0.86862279021824,0.1284106791805506,
 0.3064026995871292,5.356727180951815e-4,0.5252290762450623,0.925068365284694,
 0.6851171036724334,0.0019739057790315595,0.8265210181838918,
 0.7568820499368755,0.9471147502979459,0.6957622554598017,0.2355920153841069,
 0.9753683867180101,0.9014468456409259,0.15582831964567367,0.3581661370494067,
 0.9202807662509183,0.8114236986606154,0.8786200640592977,0.9415643022035635,
 0.4742157513851126,0.4100928176187937,0.3058256813489275,0.5538155306523298,
 0.41918833394444954,0.2222995523222555,0.753475213634336,0.9231426335509979,
 0.5752102014067941,0.47331822147525915,0.018834073200817758,
 0.23442694023538702,0.2635506090858133,0.5660005532105864,0.1050525401304242,
 0.4728417885757199,0.08399707799251754,0.9973859869487958,0.7600375750868523,
 0.6799959467957364,0.4920253989272163,0.8227062425520587,0.19599881813886832,
 0.34632643163629984,0.06552374234611014,0.3978147783350594,
 0.8367674075318781,0.08611703670394089,0.7333362031129715,
 0.02219325838389663,0.1285430032801549,0.3567790808046941,0.3585748256633383,
 0.9477845252063231,0.2483603417733392,0.3883665935122641,0.25180775978252967,
 0.8227313956809282,0.8668375635517729,0.14207665330864327,
 0.43586149044987266,0.5572850095733559,0.9035599574156357,0.2665455282998245,
 0.9505084174122963,0.5209644140348071,0.48716912085939057,
 0.31765522155980075,0.3744955644964214,0.9353071123760063,0.8376162206861103,
 0.6190261410128455,0.3011521135239259,0.04039196099993936,
 0.48786320705756103,0.7029226182818404,0.879828631471205,0.15704773130626792,
 0.33124290601170947,0.5811922238482504,0.6012055195836914,0.1378597492564917,
 0.4940046324345152,0.06660818194576656,0.9531437821867402,0.8164942681975793,
 0.2101432760389762,0.5593628465046407,0.5204470567010099,0.9213921943691283,
 0.4944733129795147,0.1899201862396932,0.11300155083520735,
 0.46455572279503266,0.1482673442493443,0.8671327223325135,0.9088377640786092,
 0.5427716540074177,0.13935555886873607,0.28459423635775394,
 0.8395457260309015,0.9761468705341816,0.9165750843483049,
 0.008812442891612204,0.044438564473569864,0.695909371586698,
 0.20432242731505823,0.6287540696600589,0.5510491988092615,0.6005232807309565,
 0.0245407639535683,0.0021976790872482166,0.03884743393234835,
 0.7085259501816515,0.2085579333797145,0.958426721464989,0.841575956880246,
 0.3158042037376735,0.6632595707088145,0.5971501553446829,0.5370784148294233,
 0.08180763609506259,0.9072461374946902,0.12894716858253863,
 0.08591313607255813,0.6766426816606375,0.6551796425987474,0.9639413936406364,
 0.06230997467123356,0.4794006420007221,0.9884761504760965,0.8606093064700764,
 0.534423899390676,0.6799932362947327,0.873342551638014,0.7277759321687456,
 0.09143352628493184,0.9409178577866442,0.4579011462071141,
 0.08897788804554274,0.16106589738130295,0.159454665190914,
 0.49402343401766213,0.2719981218433476,0.43901026577355684,
 0.7660843795654837,0.26271534792904216,0.0031040049301386263,
 0.8050091406717669,0.7387958965621383,0.22263452209950807,
 0.05208554568738322,0.5201709475917577,0.8009701042862503,0.9324828862428189,
 0.9362453772201272,0.8825448351007638,0.876739602929917,0.3542004220936017,
 0.16496729562218604,0.4154327589524076,0.838448736666372,0.10407260306834343,
 0.2506013363155325,0.4160027205729572,0.1963285303938992,0.1623467958571142,
 0.16899619453874526,0.5877065117695762,0.49809794973458565,
 0.9861275810732014,0.44377137201805317,0.2842386112617138,
 0.13149562152521765,0.7716560814508371,0.4764696269206441,0.7124805838586298,
 0.5576655651899772,0.816196632600135,0.5363434221746077,0.7230919689691296,
 0.8404214521602333,0.41373140965649524,0.03952774120498159,
 0.9141280935330405,0.6050370052283836,0.4394591617083965,0.9217336827054593,
 0.39878126279725246,0.5476849262481408,0.4402778660415432,0.8903283320943124,
 0.3209628581464472,0.23431033525205658,0.5440490249632954,
 0.20282793536148136,0.8589497571417654,0.06396071054363228,
 0.7967572535200754,0.008018570285333149,0.18598369161836004,
 0.6029542538905062,0.5471029509166401,0.6255303040757493,0.9757219138753641,
 0.15636795639069,0.4300530004144798,0.0301740740990607,0.40464044530195553,
 0.7941576644266355,0.3999349830751624,0.6356268761386786,0.4308499125355829,
 0.6040741091354012,0.8876418004457655,0.6597023995913437,0.5209966177272692,
 0.8433250021768621,0.1477996729672627,0.10588761558553306,0.634344774814756,
 0.08441767297754188,0.6026203164629007,0.7422341866869971,0.2302545174909829,
 0.05844624259632469,0.16994849863187933,0.7233914805945355,
 0.7343715949686574,0.13923708677814395,0.30405984435379585,
 0.10418671053840511,0.3745428504969195,0.7979523329602793,
 0.15504435219856183,0.3521420997393315,0.6843089981949454,
 0.008521854908448145,0.23035544660730867,0.43824960009978053,
 0.3319931257389268,0.5589676776548459,0.42514224067802275,0.8017998012687033,
 0.7145411062643965,0.233950279954241,0.45220523132065904,0.8189118981252224,
 0.048158621087398545,0.8811024449413883,0.5036129367514607,
 0.6206505157792123,0.9258828194392283,0.39598699379718805,0.2982021867434831,
 0.844208023305483,0.837670890704368,0.6875500298575005,0.7108203700822665,
 0.4919913600857365,0.303538020320233,0.39154791377294607,0.7096470241479538,
 0.5089619374638624,0.4953408559112975,0.10052334235154115,0.5756041464116713,
 0.9374724361054606,0.5910074406633825,0.038469323111555376,
 0.5937213935569198,0.590794310312998,0.6306172971227324,0.20980623875504356,
 0.5886333592899455,0.18739078693022448,0.5488028805269846,0.9164783527036128,
 0.8600424402618152,0.58551393251298,0.7074298367369101,0.7894570996414149,
 0.2769355781336824,0.18335590833847992,0.6543307160332598,0.7171710255335477,
 0.6073917122880192,0.8831936347171903,0.4944463059541675,0.5328698662093547,
 0.9175423936266626,0.5037763830802531,0.8214830293082822,0.11424054248046045,
 0.7580783125478838,0.810875467413343,0.9411700394301672,0.5414776446155285,
 0.7557459224902623,0.24668377282987908,0.6014722274783092,
 0.18000996373420342,0.8073127507902041,0.4371808922282604,
 0.040939817229342146,0.8274722570760541,0.8260345600367768,
 0.32507048776473146,0.2202278298230641,0.8237795659073661,
 0.46229609799976346,0.33800465382276845,0.7946226846537633,
 0.07496806483237961,0.5450909660193229,0.22988343445677462,0.778512280416455,
 0.5730805961313291,0.018851201102919157,0.49621357279099265,
 0.39599051013820896,0.17692119772557857,0.4567401311558056,
 0.4604999557367856,0.9053534847642419,0.35494805177156075,0.6259245684141187,
 0.3603356565649207,0.32701476669543605,0.8497409261585198,
 0.41142320380275244,0.9436404358603486,0.4624185023218953,
 0.26125875664086884,0.38098731898977567,0.7977169442636427,
 0.6261140449882221,0.6355052813513202,0.17354418737684885,0.9385525974812639,
 0.1759199880586756,0.09917310820947745,0.6350829330242699,0.5593455218273132,
 0.20753853693013513,0.09387014711564012,0.33353266245940605,
 0.6406302790120368,0.8912321967381283,0.7435910636647244,0.6375337058161161,
 0.5281466859350913,0.670511531842072,0.3362242106921113,0.7179420209718093,
 0.4912257424817822,0.04933856280295945,0.47261674126369535,
 0.9529861928503374,0.512330972196823,0.04829873716171984,0.8358619537903996,
 0.6447184390733703,0.9391113126215012,0.6762444419150914,0.7175774660953967,
 0.4187507988424345,0.6456028108465263,0.15551629440644077,0.59751678359335,
 0.8872787519478985,0.8186899993111436,0.8373448391150353,0.2392240730501728,
 0.024923212862466793,0.8773540761493227,0.5484549678150955,
 0.8800740847291897,0.6658039489003463,0.4850215569535985,0.21021762426736668,
 0.21665816492406054,0.051270863212686035,0.9371337691726983,
 0.17106938125803017,0.6377978531733891,0.8233040562234781,
 0.48067967151688595,0.4410950235032375,0.25288090805494334,
 0.34093075609090284,0.4444999031781507,0.24862983464239474,
 0.5707853838202187,0.10572839296687964,0.5148031678372954,
 0.38113469110083287,0.6427849522533412,0.00926227147722336,
 0.9337690553642297,0.70855197383307,0.20426155500325194,0.979494325527593,
 0.9321786436620598,0.21630710391583163,0.8270415557421658,0.7050084329121954,
 0.5796591773219382,0.5595963050914425,0.6243395551525479,0.2748714256635768,
 0.9556972585205115,0.03178378149601069,0.3142617729424395,
 0.35346050186879374,0.21499329222312769,0.12912196835511436,
 0.018438444671240783,0.19735674333403175,0.8051346786895199,
 0.37459691038084686,0.5287658235785276,0.7221224531298912,
 0.24466713910265214,0.26086070460697053,0.2960225841289479,
 0.45486110494473486,0.2660500956123317,0.6875429695560181,0.7834854575280362,
 0.24249467308779316,0.44747785238422344,0.24660329222177657,
 0.6009602277136568,0.34985063890590606,0.3184347082241904,0.1667068846440276,
 0.5958718501956538,0.3067453629484381,0.7085425451494343,0.5656384249811114,
 0.6719167865045426,0.3147670304089525,0.2092348546143128,0.6135109873149606,
 0.7644445468748533,0.7511078782738436,0.8577632342593023,0.8369159824263604,
 0.09815962910907905,0.8678756024484678,0.8617700675529109,0.6919516395414715,
 0.1591399744037606,0.634955841330038,0.7836182024293075,0.7670705821951168,
 0.8768958803475945,0.783675701911283,0.7354544037955845,0.7423695762479048,
 0.8105282211037069,0.3142722223232548,0.2571654175694591,0.5495932645967573,
 0.6703927052107809,0.05418829234061073,0.5175239747054489,
 0.37275732018617136,0.9548950719252342,0.044403947659351406,
 0.885711690277049,0.1273496891763749,0.726849937230339,0.12687050140216172,
 0.5522139789076408,0.680996106876739,0.462758717555364,0.19084356535197833,
 0.03107176284553248,0.009316468993073368,0.3284558157850792,
 0.016874026139306664,0.5189814159439804,0.9927639219671806,0.793293670671815,
 0.054268565373210675,0.8411121097883006,0.4995053617612303,
 0.3528283070185191,0.11483356722247029,0.807808858267975,0.22335631734505246,
 0.2758240680744979,0.8755437449991986,0.02638638807298388,
 0.47567631605852445,0.47943999537180204,0.6322460996014029,
 0.6924671105977553,0.30251381757156215,0.40413827051150886,
 0.9659647740608244,0.3684597681759568,0.642801554977792,0.147054794116105,
 0.26433615446902137,0.06554970677608507,0.2461791910331761,
 0.6492765714414653,0.5192273436008564,0.26373700125546573,0.8535446255495889,
 0.42438806390575556,0.5050485378274298,0.8617890005635408,0.3518774219234586,
 0.1456131993732015,0.11724714545625192,0.04708272458708895,
 0.46900833026654276,0.5998916414313109,0.47927567590115805,
 0.35269411678361207,5.52359485065379e-5,0.7722695884567131,
 0.8889996112519765,0.5587269272203005,0.11282914632004759,0.9520520924074823,
 0.2567985291069509,0.41781785806963045,0.5925599868800292,0.6541102565839261,
 0.6947964069796647,0.3851840487318057,0.7393804005348932,0.9142389778221012,
 0.3982376170105021,0.338993779657649,0.8133054201086365,0.5764127104938438,
 0.3922074000929834,0.35036749897735864,0.3290901564763399,0.547356192809715,
 0.6460017132479389,0.8186331958561032,0.5639028628996162,0.76176769858007,
 0.609971396209521,0.05480175446925073,0.9630952223453769,0.5504897210025709,
 0.5579116998654097,0.8257368147234808,0.0453007855229991,0.6705916255145002,
 0.03192783710746805,0.3716964141493222,0.15023514227758983,
 0.38840242860925733,0.4320448552054784,0.09015832319028405,
 0.5877874468788189,0.07851480062042404,0.30427154689703517,
 0.2077071523177748,0.634775532137766,0.8608145001688068,0.7407665960420827,
 0.5894191268198212,0.9193809515287759,0.8759237746218204,0.7329965207073181,
 0.39850536254389834,0.839454118846627,0.9073373868390018,0.8768398590766843,
 0.36545009014596086,0.21506312198726674,0.5512245171312229,
 0.9151466242048829,0.554861084673914,0.9686559582067954,0.025829508571038495,
 0.07707783390561507,0.45068269835260155,0.435344294744787,0.9297076114675003,
 0.17644992498518275,0.44948004094771565,0.8078214945226441,
 0.9146512608694726,0.6537851070672503,0.03908406165475109,0.7089962276260232,
 0.5381087398519945,0.5603475016138204,0.964127381704654,0.7703031964497407,
 0.9327237052185443,0.32176593126241815,0.18071576122248356,0.547959024681105,
 0.38637297751597965,0.9958180102960317,0.6730108725542807,
 0.44869441447148795,0.02187806399304071,0.8607392165348434,
 0.6240712950425176,0.087318679136978,0.47401477696800465,0.9195716930631546,
 0.32996244301560784,0.09324391489065109,0.5438743354595454,
 0.31968980972438277,0.45152230538823246,0.6298296960113928,
 0.6930542426480347,0.48156244634602896,0.08951504584607706,
 0.7439268732475369,0.35359603367161263,0.6523072114200673,
 0.47596497681862193,0.6661913225133843,0.974886492041778,0.3163254830678026,
 0.2580067598842045,0.5301477389251787,0.2893763847659099,0.34981260025803784,
 0.5790800715797663,0.06647668702591036,0.5339677493730204,
 0.20420311597971397,0.827252242519801,0.3832732391168038,0.295085250043881,
 0.29397464879988333,0.7970033013814636,0.25500199950882396,
 0.5361636925070651,0.5182174634739273,0.15636099898020694,0.9311098215715888,
 0.1357296735014667,0.3311573475313443,0.22400264917012702,0.5086242297416003,
 0.9107711978037568,0.2645058247462284,0.3443747866077701,0.5471317346273361,
 0.7093866930402815,0.6071372616821233,0.1415341257760887,0.7049037968741048,
 0.1665521573916754,0.8123873687923804,0.707480445424401,0.6076935878088923,
 0.44225569932472253,0.10064880526937559,0.13357507187956508,
 0.7645359073429834,0.6958475029847183,0.541393858643436,0.9903670823222938,
 0.20898616392952685,0.46570593643066704,0.339916742876583,0.6442271372974753,
 0.656204707963862,0.20231231767464464,0.5189751118422328,0.9118008582079484,
 0.08007643017183541,0.09967591306680568,0.4599364559167901,
 0.23852655726108118,0.32749593971463076,0.6204286151073857,
 0.48741902878896926,0.09383559242332462,0.3130169235118281,
 0.9784178387079036,0.31682499942401143,0.6001175092394313,0.7883781157695331,
 0.3171637637840321,0.9614909610122155,0.571761769963961,0.9572090494913807,
 0.4754591755492257,0.6522699126364164,0.4157328614572946,0.508095566917581,
 0.3211336330466906,0.16459207379029384,0.5099424435785778,0.5337811354357821,
 0.4652298865576846,0.6418611305570264,0.0017452550992256288,
 0.2902394220477842,0.5673840999586706,0.994641042533747,0.48321378056757147,
 0.02432672175179662,0.9190871632087364,0.04785371404221461,
 0.2644236680283698,0.17903552683069246,0.4761856013576805,0.8226730263560277,
 0.1572989063966902,0.10368394634967237,0.03450315779811919,
 0.4965327369326512,0.14355744491048117,0.011685577224199162,
 0.3956908290611185,0.3789034044260241,0.11925071520149322,
 0.013498449051906669,0.20485039991270637,0.8354797810064192,
 0.7536201530969653,0.12969968172568092,0.06707383180831616,
 0.49461120745979636,0.4877544741822295,0.6117733005745863,0.8386153484047092,
 0.8289253466738976,0.8666170022188233,0.7386097781170139,0.894550450781304,
 0.2784492449603664,0.854327986236048,0.15190347573183616,0.6294791639111785,
 0.19403231408186716,0.6735820248569292,0.2632433050013908,0.7857967745861019,
 0.546089677358853,0.5959350384016056,0.3062601736764856,0.8583266131488726,
 0.7718595455731501,0.6251293800908818,0.9098345146216021,0.08936062905284137,
 0.8441417754392072,0.5839859299999941,0.9257252137468575,0.7738066354878989,
 0.2594954333349281,0.8923759771986708,0.4926908016466687,0.9010513683141856,
 0.9059157086508507,0.629600817958984,0.8886426286050142,0.8878317770927633,
 0.3832936210696851,0.7025010604260297,0.5531529152260639,0.8448139931843379,
 0.2775976077504615,0.843612641790688,0.7365845364436592,0.8235676580592318,
 0.6694631712177149,0.22093045295461933,0.5084389487308204,0.5687380355679496,
 0.033796365531014394,0.3919211860886833,0.34272772807304075,
 0.29108745556896753,0.6683729559593141,0.657937998957833,0.5282317726049611,
 0.9313320428438879,0.2928509217646852,0.7400926993216391,0.6817488231799385,
 0.43646338973065857,0.5043836455100743,0.7596982875435347,0.6756214898710388,
 0.7057042899405255,0.2570407668625794,0.9705789542194528,0.8173662495605946,
 0.5053070044444808,0.3720506328309554,0.1103946213635042,0.7109915163754738,
 0.9096961883541195,0.11623784794125602,0.9523896202372796,
 0.15782433788029526,0.9095906501230517,0.33701579038727625,
 0.38719875166186835,0.4593414932748525,0.8870872227707629,0.8391649356736979,
 0.65768203031623,0.45084997522011894,0.09385428044334332,
 0.036844875049682035,0.7074818968295593,0.4858107304424598,
 0.37508975390264787,0.42576688084490866,0.45853295261722504,
 0.037530031608737024,0.4562537512569882,0.3009329590020827,
 0.9883665313074387,0.9467818166920551,0.7844704042300761,0.674101426950923,
 0.48057560394764665,0.10948799230970574,0.6009445015064474,0.864482189738635,
 0.2402728881767925,0.9694984507366943,0.49872681219875836,
 0.09156562721642192,0.6188866755396806,0.1489355729199453,
 0.35474868380288105,0.3178816772337537,0.28513742109317697,
 0.12692352282107955,0.2129389028849712,0.8000774521608438,0.5949935317834405,
 0.36776772092322385,0.4245855341748388,0.07709802966946855,
 0.5764911379062388,0.6355724943744778,0.6138436406287993,0.6796105125100778,
 0.4085867451366145,0.5450179364481402,0.008235213285243459,
 0.6970575124706437,0.04895557649533999,0.3361485148219565,
 0.17416557242569353,0.11511205676437464,0.22007388635067104,
 0.4028197886120205,0.3968628754069474,0.5490539356522662,0.24425946258643394,
 0.7792851589633704,0.26009356919215565,0.5743478719097829,0.7272826810335922,
 0.5017684449987698,0.1863761852343494,0.40078579573571016,0.8896441919432725,
 0.914151314057023,0.5890167800596542,0.3521717454124802,0.8875910918612508,
 0.4426830180157322,0.6277018624374622,0.833317488530253,0.22939735005208406,
 0.42225065760218405,0.8081616540394672,0.32645021070904745,0.983724571902277,
 0.6461532773872343,0.8971307259463109,0.7733680986400847,0.9574233739389747,
 0.8617205184884941,0.4823335540923237,0.6432567689375527,0.9032055724298179,
 0.7377440835396534,0.17406120658025026,0.4407057938657586,0.5841236122507447,
 0.9894005144493015,0.3965409615446708,0.7762353202738771,0.0926935140319658,
 0.32831122059658013,0.42349510389444067,0.17937731799052115,
 0.981741594504378,0.35647990783712136,0.6136346344778375,0.7999331235188322,
 0.12756846041632597,0.5084260542338443,0.07417091632325457,
 0.4806876718711185,0.6166620064270516,0.18904184898751264,
 0.11596600631473897,0.06188764200225094,0.22837468800815253,
 0.7676800502860779,0.5645617144364312,0.7860299359080736,0.7496674317885319,
 0.06979668853382381,0.9529516339405271,0.33712665579105616,
 0.6652761240661631,0.48626924167028174,0.9839376330052984,
 0.02024675182809088,0.5200231326574598,0.7733952903077884,0.4932795641623996,
 0.4156704183450356,0.2519302249997136,0.7336240482889036,0.7129476660955646,
 0.06826519548268872,0.3820966952406899,0.12686720992630263,
 0.22556278705015365,0.3305758699973209,0.24832410982673625,
 0.10656513322163463,0.06838255411686289,0.4934644681364785,
 0.35860484916448365,0.6715477102928824,0.7925390835368413,0.8861812807656211,
 0.4134366692863155,0.6121840223988664,0.6170964447051295,0.6053884063780403,
 0.8263972639662677,0.709026819526114,0.7653297625826091,0.25642603555817567,
 0.08179844583900497,0.28281326794392103,0.8315262012491812,
 0.5125984965225341,0.12335125866695607,0.34558593960604,0.7952923076444768,
 0.3755764335519789,0.9992877165090106,0.6363116128317389,0.5229265381833202,
 0.4230423763381519,0.34346851470968176,0.04139535050861953,
 0.09138893764359646,0.40919313267374724,0.0972015364045794,
 0.03974410027086328,0.8331683465743174,0.9761804844218319,0.7911460262081778,
 0.5800546341727723,0.43587577858481985,0.12140156152017156,
 0.7588653259641542,0.31093933335887225,0.3711956905157304,0.6842484596940305,
 0.48250780894491085,0.6894216683724111,0.24371688442800454,
 0.4140765637161462,0.1586950567177653,0.31400320743894794,
 0.07494053124816802,0.589476362298119,0.2753990580205121,0.1238113539206338,
 0.5399602239139071,0.7488848934892346,0.02827580265452223,
 0.020998874814306068,0.2765002191347363,0.46983129107118693,
 0.7536561923158438,0.16243044743760748,0.5648879062519451,
 0.01785312725140431,0.34954602492579556,0.8282614686540519,
 0.9460611887941679,0.13428329884376156,0.8466685342639104,0.7458132380758472,
 0.9508666391130451,0.8850635981280762,0.7354513864487382,0.7968843542135837,
 0.7337541876910624,0.46972031125752767,0.16189453324404468,
 0.7733097359527752,0.6353053962151421,0.43106985346529936,0.7575218891692692,
 0.6243399597753017,0.08064455417272731,0.4348585744162231,
 0.47928389090798174,0.435329453720964,0.5179178696057674,
 0.012204358886663269,0.6928250177046906,0.7938225068444863,
 0.3245420947772567,0.31762099311509384,0.5189294547247576,
 0.42167619339957085,0.9556805528362005,0.6834460083177364,0.2784049510487102,
 0.6936566941372553,0.7474946293566602,0.7320046219054744,0.25212706590701806,
 0.24996635130149292,0.8884654238157823,0.2631315736553913,0.3500339789190684,
 0.12202775449877346,0.9074920652312455,0.4397805526695703,
 0.11764328887999109,0.20964098778813245,0.5882315314894287,
 0.49057196100916467,0.8098086526128999,0.47390849193032747,
 0.7737431341397436,0.22705518157451798,0.7144851917015095,
 0.21901079863363593,0.29164254270670786,0.5535832572055874,
 0.5183838180165332,0.43411839714481204,0.017677322122000683,
 0.6569255366765593,0.7687992159941369,0.08463955386016353,0.5838467371426369,
 0.9881101662978853,0.4186078838487689,0.05503575409731021,0.9172439166758238,
 0.5629387992317049,0.08600911201876249,0.6097077308234717,0.8841541982390507,
 0.6934733778600665,0.3787074523762747,0.5174755149605973,0.30084151451185415,
 0.009803798619190007,0.9742451726994866,0.12603143989907373,
 0.7443935901739493,0.8158758608051232,0.3391041657209949,0.9007431088266081,
 0.7296589653023494,0.41444994811264413,2.9634760475416577e-4,
 0.27062332189109495,0.5802746519226502,0.5485627762561408,0.9724158622372647,
 0.2517958614859096,0.2887804033168937,0.8760048535840477,0.25977822240828274,
 0.46944805002289736,0.3734180652157256,0.6667874187581437,0.8526479541506244,
 0.38387478297731914,0.045511471405398796,0.48931547821837573,
 0.3037590487113566,0.05329905567482962,0.07676197844616928,
 0.8778719337584694,0.7647351065224697,0.7394006724211714,0.05417179887248669,
 0.6973136959127948,0.7677541602220532,0.6075738520295466,0.5683596862836136,
 0.06251572978526287,0.7681521968339027,0.9218931676292972,0.3937528351869618,
 0.04515440715691721,0.2457524430574981,0.5359204905473173,
 0.11608024550015783,0.9160516914858015,0.17129688128549814,
 0.9391049942970796,0.15624872120152322,0.5254250546810698,0.971628123343498,
 0.8139052335438257,0.592550737705109,0.8041020553321345,0.9099289212696456,
 0.4132995887123966,0.10705484706025314,0.8150575608228608,0.4721146033356416,
 0.9597158802802326,0.49118889379834496,0.20422911773249108,
 0.9487144377495704,0.3973072659135759,0.6466176341484059,0.4982052141980887,
 0.11431637033224673,0.03321314581618273,0.789577029462078,0.8725407454005686,
 0.4163527129394927,0.2525609625270624,0.5772736589802586,0.8934689837703067,
 0.577876035426843,0.783772532918591,0.43365342531912854,0.07891938997578674,
 0.6848861497484495,0.705663979405132,0.224704281060085,0.7683789140273071,
 0.1389435101552393,0.35286842713652167,0.5590049700524852,0.8450443883047507,
 0.9387343426294054,0.4020521170292548,0.5333740493134185,0.5003346116138179,
 0.7177356056373465,0.5237684419772507,0.8537499103275833,0.5060380168785855,
 0.42469833825998116,0.324609454646386,0.6876423250950079,0.8746096248988132,
 0.07686668653587836,0.46037889841009716,0.8692825113490691,
 0.27100850090399287,0.37010268961228043,0.8356745566849401,
 0.18820287343519104,0.8988816006979128,0.8883799082830511,
 0.031033484103695663,0.42005698808628156,0.11027240201219357,
 0.7148438934988819,0.8150910377124075,0.3371160604783997,0.8308740929262555,
 0.6808868898031848,0.20635016022664665,0.5880039246549364,0.9646728461740277,
 0.8734294392702869,0.0709940985251396,0.5307006863143903,0.46204553278354443,
 0.6793058708785484,0.13733492860364716,0.9227263570936847,
 0.45573689901241454,0.6411706436621681,0.5003112580032218,0.9556811360773291,
 0.8609517106126233,0.4139625338609799,0.29231276869759204,0.8820309742926631,
 0.9514793307299013,0.1495503894083814,0.6352288674169594,0.5182543315145204,
 0.43768910558660656,0.941417591986205,0.6145723157179414,0.11714536332105718,
 0.33285140888233,0.005720086763084664,0.9036617187938869,0.4345570421352155,
 0.08195762349336144,0.9761876069079476,0.03402199097682568,
 0.8495067128123255,0.78142055751623,0.4741924905447097,0.7554024945593444,
 0.7017543679540392,0.7331645054416589,0.6888586992731964,0.4913135499848701,
 0.44966837801067716,0.8701881173574108,0.5114770928618599,0.7164685065963393,
 0.23504553500059877,0.24298478724081485,0.3580635648234983,0.530239128025943,
 0.9525404203923655,0.8146732892992852,0.4135551943090645,0.2024277344098495,
 0.6120243202237411,0.9337295294085166,0.24587717266134168,0.886746262324812,
 0.2447336980358642,0.5290767677871386,0.4819218366941128,0.8243806386570199,
 0.25979575459489335,0.04148969332607355,0.23552054688337987,
 0.9246688676539953,0.46938370133417817,0.8541716603283136,0.8174891424851695,
 0.08486680233709665,0.8754722821367883,0.6638197258778394,0.8683558926972712,
 0.003398523168867018,0.885085603821174,0.4732635404423847,0.5010320599846653,
 0.6994200797951552,0.31255351092719763,0.3032249758301184,0.5315943874643483,
 0.9132646805142114,0.9945032050970323,0.8849189016835085,0.08592209990359556,
 0.6438499589872071,0.42125344195568704,0.8537631160737316,
 0.41406672457102656,0.9700370618448559,0.5922554531306745,0.7139716596022371,
 0.61609774013377,0.6211948096499611,0.891690714630852,0.8714810294439665,
 0.3047045909949343,0.16022389679834803,0.514978976999378,0.07967675662280982,
 0.22270430136254138,0.01055860120144847,0.9702995139379486,
 0.7347279432079409,0.8336020794871354,0.2890826159680655,0.3943303544942659,
 0.7609475544703266,0.16926789026909705,0.36761761228459255,0.428314724308268,
 0.5060090671359423,0.25000933746585696,0.11796766294556105,
 0.6675891233256418,0.5521507127524835,0.7317881514570953,0.9733663914545492,
 0.46105270630031525,0.2661379382854052,0.1580164441614923,0.7092225292165308,
 0.4537912941075446,0.5462228026430851,0.6433835342968403,0.6455409385796418,
 0.36312728608290845,0.8288037856133552,0.4493477987013472,0.2623487454936847,
 0.9005749038291125,0.8971567868362864,0.8884452065651889,0.43311030375212844,
 0.832685705196242,0.08066217700204892,0.5510722892690064,0.9009315793808664,
 0.8818708174159265,0.3129682998036787,0.6616571444219173,0.8754710358082181,
 0.43765293340208267,0.6077761550164106,0.6093302826634516,0.827746006676157,
 0.0811185565075101,0.9685256473729476,0.7679264960938832,0.35230706208343,
 0.041391882047851936,0.3809505587831774,0.2830316296123567,
 0.8115869273830725,0.38643135759156233,0.1243688621423138,
 0.10344781458686048,0.24632390825394856,0.7810980898147473,
 0.11535693822878834,0.6905769744492462,0.4896248690576295,0.8499416821208935,
 0.9694670124050379,0.4003477932843692,0.46891268897212013,0.3099991256574951,
 0.9322735887260531,0.8550754370237136,0.23701076488229933,
 0.29204829435103963,0.36525441312173057,0.5159184511148015,
 0.8797967611213039,0.3010125708031103,0.6212819091773345,0.8397365484771413,
 0.9871272685219405,0.7750109903636683,0.0908615189170019,0.05673047284253396,
 0.26748534440535543,0.5903549161964446,0.11567220222726105,
 0.2071182865706035,0.6013430954791434,0.5916036740014313,0.9030461269080842,
 0.9740257138610815,0.5558961978090516,0.8248199187332955,0.5270189643361628,
 0.4626969268945773,0.8185983015649748,0.19350290687822946,0.5210169678062343,
 0.2856516009927832,0.6528789619252351,0.4562327507497156,0.17626571752885678,
 0.42254817615833684,0.4546303922518232,0.9761896926364548,0.3361551074292013,
 0.5084714532409981,0.2750699908402817,0.019031738222108574,
 0.6675511861141451,0.22559661198147374,0.5728738935700296,
 0.03328353280710272,0.14964536865784583,0.6451506134044325,
 0.3964588959714136,0.1937597238632468,1.9190478469255368e-4,
 0.7105714709233228,0.12194003711737822,0.36120543325138066,
 0.40918293739536604,0.4759109100058213,0.6246454967673771,0.6904257295638475,
 0.3723181893263664,0.6089403099436834,0.07050627849035673,0.9298004631376268,
 0.7882698840693668,0.48933862266069295,0.27176848557731514,0.426949234732519,
 0.5088457669026418,0.6380384002249644,0.5428782110142448,0.3053466349704399,
 0.5644110713930992,0.5512835008442948,0.536817223382,0.1632828892401551,
 0.8341200629582353,0.25345990067998914,0.7003622365385342,0.1316845554284526,
 0.875368106739745,0.1979872820008025,0.9441145898597186,0.6767899214637828,
 0.7961047101063189,0.9495985657611441,0.9776789419658616,0.17267115273790556,
 0.2138043025658014,0.41296778501905274,0.18391991302941724,
 0.7982025462485937,0.14749091408868553,0.9092909426108234,0.801122993957767,
 0.18109953370353005,0.21069554627608822,0.166825081467878,
 0.49196650819669296,0.05510898436744249,0.7977589734233602,
 0.9759751404192732,0.07345544361663481,0.771130127607223,0.6897255296853367,
 0.37902158887026105,0.27052654499621787,0.8246711876513306,
 0.6623831521648561,0.27063221862298725,0.2404692052976548,0.5314813618658975,
 0.22369828272885806,0.5507797358924059,0.7030606821433231,
 0.19278266914600328,0.69802338312714,0.354466347579675,0.08626257840092477,
 0.1320183677204113,0.6034577796814202,0.6623401565226402,0.8285004612936873,
 0.47295926154620826,0.5405404629707563,0.7778285683409241,0.6920150887986222,
 0.2963526792660629,0.5262903658708813,0.9564123906413838,0.12800148456595806,
 0.6188660823378561,0.6582396997893494,0.17166847978986166,0.5867342905045886,
 0.5421465702831011,0.23270087415694785,0.8871434473126801,0.9963653211461405,
 0.5681683184679124,0.8469478470363888,0.511007805707234,0.8378908536665061,
 0.10142220727020079,0.4567653529525779,0.5880396515739081,0.5494514867134574,
 0.5938874702486582,0.24594232970568042,0.9038885016100772,
 0.042821741255269496,0.7200326601821385,0.7733164857231969,
 0.25622041612764135,0.7954802634614871,0.8376741610069715,0.509368211744982,
 0.329880168180511,0.5424127242162226,0.9850245030218946,0.5739109009388397,
 0.09363801383975134,0.18836987312082876,0.36534529734979904,
 0.843371878121304,0.0013309636136983327,0.9792655040324847,
 0.4037323728893665,0.19829473927602548,0.20025425014560538,
 0.4229353207746256,0.08108826264153268,0.3235454807345193,0.6225913056148785,
 0.9792376421556017,0.3952853527619863,0.4254320624997572,0.22605643831897637,
 0.1739120907775168,0.8285603325477734,0.308160452565584,0.7500796924805535,
 0.5572078037292805,0.5922071041719419,0.8617505925444974,0.9647925121024182,
 0.403165698065282,0.36623555354575965,0.36784236645847,0.07474495468350839,
 0.7236839149031395,0.17501081672753394,0.32669726834177815,
 0.36058436477223876,0.5377429125500722,0.46535270594425227,
 0.6464949023700854,0.11640696399854622,0.8596905038054643,0.2100121688426163,
 0.48592330406162854,0.8170382266061006,0.8630132759470678,0.9372919915684841,
 0.6237329960537112,0.4200311994376168,0.02628095394369323,0.5970877855532843,
 0.8942025738758341,0.7501819385744518,0.4007400231707213,0.19189154204028802,
 0.8391587856137745,0.39360795225401535,0.027638727188224133,
 0.42731299972988146,0.643399634192202,0.9622115035282408,0.5351456769121786,
 0.7263393111735417,0.8519370022620105,0.6395308707580112,0.7478069548313088,
 0.6056637907214689,0.35027325241852847,0.5077716193024272,0.9363227389630346,
 0.04495994328010999,0.5402046334381077,0.4241919932525101,
 0.07735834537486996,0.045778485296341875,0.4451896496582076,
 0.6515308306762304,0.7207067663363275,0.7875102306313758,0.7807524164599546,
 0.846875813516351,0.6828114059702816,0.7091321966829863,0.8218111711627927,
 0.27062010277845183,0.22651609761833613,0.9106420396291885,
 0.2179394702017049,0.7207704604494425,0.14683438551376726,0.717589645748429,
 0.9421440484281491,0.5388818986163434,0.628138206048389,0.02491378569258662,
 0.3596881840052101,0.227111666669084,0.8054705601472086,0.16287153438515944,
 0.7728648924570897,0.16425649767276096,0.2951286072270175,0.8176475596732736,
 0.510033510309281,0.5755463039760871,0.5271305629628721,0.4155249133839103,
 0.5841968810829811,0.2198386335537692,0.8483633922162802,0.6666151753049215,
 0.012803679280460667,0.013276500101981448,0.38807350409859576,
 0.6132226387374088,0.06995796044272184,0.1975376363818111,0.569228366614984,
 0.15668393048949936,0.7358245813925429,0.8517018902079,0.8987868221881476,
 0.20917921787639704,0.3987966692992836,0.4859350029832976,0.7670900929688333,
 0.9479878628527536,0.7138130309347029,0.1289183630746228,0.6628741986954115,
 0.2886681907311358,0.829698944020129,0.5298835716519686,0.8465948384464093,
 0.6655992780128454,0.613343967176977,0.2900092163886596,0.6803032003365144,
 0.6531004075979323,0.3926174946959228,0.6320984175420861,0.4896819541781279,
 0.2040947225698887,0.9746689189364629,0.9628065698406176,
 0.015029078496016868,0.35140332021429777,0.6451239659322676,
 0.7810244849894258,0.1079520279812839,0.9154695144061772,0.6028012535062925,
 0.07330781591312052,0.8848561799070715,0.32617838397089416,
 0.9097396520912349,0.5570250970207555,0.32452210811487137,
 0.05550037456505175,0.06446069114980779,0.6526659148726335,
 0.43463194778216163,0.904594842647085,0.8698191895353988,0.9946146529401119,
 0.5817246231132713,0.9140278255261552,0.3457716926443237,0.29380060373733197,
 0.252277690838133,0.013425275752122823,0.20702239395149835,
 0.5254025792832029,0.8292428488526644,0.41927846667113,0.30385682018426163,
 0.9167075914403559,0.6982979419776173,0.7649742145645297,0.6557500393430354,
 0.12810928112188613,0.7035126828373198,0.6910284336653093,0.1810643322899652,
 0.4297823522381111,0.09313463774322561,0.3065028380390624,0.4558260673693917,
 0.9163615086325557,0.5802354133707728,0.07106192218170226,0.6080322650974312,
 0.5312691717175985,0.6330197328713735,0.5023723241409503,0.2750953824407022,
 0.08396746932382548,0.6469966639475327,0.9058377640880677,
 0.16003014498159174,0.36873398283047676,0.7691279510211624,
 0.9487468985746385,0.16649972268638935,0.7066571611599134,0.846751284918259,
 0.6138589016708884,0.5109697468356412,0.9014775394876143,0.06133624637659085,
 0.2140627382772443,0.9115189325497834,0.07313782659069168,0.7557219053548854,
 0.775791043189868,0.7297527207526646,0.9497000966950678,0.7007569500981042,
 0.01154121826765464,0.8346565484952092,0.20388583410033534,
 0.3557426668443203,0.7476525414866635,0.06133657887510646,0.5900487446231456,
 0.674875590675516,0.37046942314520726,0.6942897127918338,0.8562624355081796,
 0.7047300088349926,0.4053762504262962,0.30101338326688687,
 0.31996872232234097,0.48049970997009883,0.3587475376759397,
 0.7982932028267474,0.7213453292235499,0.8132842088892837,0.5218697993577405,
 0.625465376744365,0.34588927389641055,0.8615708819506672,0.997675390046661,
 0.7658430592583959,0.6662200205736428,0.38151102924112634,0.9233100859761991,
 0.35716250489599777,0.9471351008283756,0.13595882964500527,
 0.7367879615640205,0.062194897107051395,0.8876332888294293,
 0.17389056135560854,0.27716632070119507,0.45785797283206164,
 0.04417100848839706,0.27887396362770633,0.9670241754835829,
 0.5810557264913941,0.15713079849845757,0.8835852797917086,0.546900531167978,
 0.34949426918343274,0.0710545581308818,0.9802191650237186,
 0.15831358742324064,0.05903333499949437,0.48113823418197077,
 0.2656846186423263,0.48207592280301825,0.19927745950279796,
 0.6821890746029435,0.5300506290198755,0.19387718852967606,
 0.0055438425445686335,0.49050215533602803,0.16733771934802433,
 0.6627000414359234,0.7577072713117601,0.976937026867692,0.09820167786532652,
 0.9390277893014946,0.8907480563588619,0.8100101010756128,0.34222743551046164,
 0.5504547379410342,0.020613199783873437,0.8959556421982589,0.884436190453845,
 0.3195597608130334,0.9615272230287514,0.7507573837662358,0.9364796509591291,
 0.21558207701329768,0.8514430503938434,0.12780676643899647,
 0.8219084996508315,0.9691257663078541,0.6142544548811647,0.6599896130918235,
 0.503515661919387,0.12601126412627062,0.3785147901632374,0.5224297238929891,
 0.7770132576666757,0.3018926743813618,0.8450656680649034,0.9127726411760568,
 0.8538800501103662,0.6710065533361772,0.925483315765792,0.3012643265375886,
 0.38209732979706623,0.8666183569227168,0.08164354236801519,
 0.41886533533067793,0.6932060807667425,0.585592617530982,0.8862995676665056,
 0.5174029846649285,0.723613258347089,0.7983082212740616,0.34749844646615535,
 0.9414164842036838,0.6501905839790927,0.41464744080589755,0.9284636908212862,
 0.7963375404079687,0.060664916127555424,0.4817802519272961,0.833810955308052,
 0.507716485199148,0.5716260621547188,0.5420607074910891,0.28589647155639797,
 0.12950030989659334,0.2902294843225286,0.5069207752270647,0.3910917988284077,
 0.3595358830343669,0.4779182082866922,0.8064957750000682,0.2681901720082813,
 0.5819676167136233,0.16781178654554674,0.2201217423153009,0.3362129803373306,
 0.7467058284632256,0.8155850626900848,0.025560303369083104,
 0.4519073344701662,0.37767103780500455,0.0959706630827456,
 0.19559223555276306,0.004713352910151181,0.34158902163228655,
 0.9275058396722206,0.2741240422550991,0.4147925431628118,0.2300586849208769,
 0.40721512384816605,0.21217722482413737,0.9079094263056644,
 0.7036089442908685,0.5697611058856226,0.701847825915138,0.6871375393442472,
 0.8257691625163024,0.9694960523194953,0.33289897673315905,
 0.44413682615273486,0.4553441890835985,0.06029987816507043,
 0.20505370594044003,0.15190285060433184,0.5532928985606358,
 0.24244166227668085,0.9677595965186703,0.9532113978827592,0.2512964352910434,
 0.08074979724379983,0.9801911872285055,0.8468121255589838,0.6425693017573548,
 0.7511658897910557,0.024048002005156666,0.39560005035821943,
 0.3797280105605517,0.7392669181492212,0.17063839302565986,0.7944315327574609,
 0.7438155528464219,0.37346499562057345,0.5574066633507959,
 0.22185816929898627,0.46750458002949513,0.5348261280169051,
 0.6604304281563053,0.02412617658386973,0.03714809103470218,
 0.2956224019996213,0.36717181521965503,0.1793577735851084,0.6034553695915437,
 0.48804791133731507,0.8213840067169818,0.7320589966959336,0.5397196368789884,
 0.6657574418500366,0.9455588793521589,0.23999612409695426,
 0.08667568578680551,0.8930271158619748,0.9152064068052523,0.2275181067667298,
 0.40094296498502924,0.5355401049277115,0.8368638188124174,0.724246058917342,
 0.7637467017608578,0.7607724257179451,0.35572606615014424,
 0.18998517039309382,0.7692089317343069,0.4222224738120892,0.8629151080927256,
 0.4327903981859298,0.7915573751459335,0.6028502715606504,0.20962590257681468,
 0.6622938536642786,0.41834466019704886,0.6757405943667842,0.8450726332706665,
 0.054985475961969676,0.2270999064883089,0.22129621755372408,
 0.7023770031630868,0.9056737764349271,0.3362277626168825,0.9460088628812976,
 0.2404987689376722,0.5599904406065787,0.2569430471287504,0.49047998381913727,
 0.2924790924961669,0.21270599668873913,0.22414255378444636,
 0.5841098809513815,0.3981500161313555,0.7354146477920306,0.6784906266629772,
 0.6659429235427254,0.47528942287163023,0.1277107230594191,0.4601151140634232,
 0.6861653258995168,0.3361986126610319,0.5623515360651955,
 0.030892054078805464,0.9915909452026772,0.6261669276228881,
 0.42439198769048636,0.959503161008127,0.6125363126235154,0.7890010776008199,
 0.9211387103290625,0.9273975986985571,0.736659454623129,0.9823140996599,
 0.6080107581214549,0.2283735198632164,0.6613252880085745,0.04861483142734302,
 0.649669544474969,0.4063289365144983,0.9455579021432612,0.8219392914281376,
 0.5132530591096586,0.13456071894986732,0.013917357369517713,
 0.008455568202199837,0.035808207755954946,0.22231786256099484,
 0.07061388453157746,0.6108211870297593,0.3680246810541019,0.5216787773418381,
 0.8133402710159539,0.07445022101149479,0.011422627345931025,
 0.7411834021135256,0.19325585543062318,0.6976537927324254,0.976350736266956,
 0.3615710254896124,0.10701607413101388,0.04000189006653365,
 0.26896251987496156,0.9776797194891432,0.5824026425194532,
 0.21919330250894598,0.17949024446375406,0.07498300687268611,
 0.7121108586107511,0.10844902188731897,0.9522150505593284,0.6080463344610554,
 0.12333530604616671,0.7405235094222832,0.9318911639865559,0.4013425162959373,
 0.6989392827354901,0.26884200462526076,0.03408101612526904,
 0.03426493707013789,0.19555653529885508,0.7129227639451152,
 0.8549735703170446,0.7558526159334896,0.06711984022850637,0.5018571117641978,
 0.20183884016595677,0.3941594510251546,0.9255155108582305,0.5191004084738088,
 0.07930791982302532,0.13893814185220743,0.7825054478209469,
 0.8258217514901188,0.12786408824989226,0.3930610329453872,
 0.43055459745999625,0.3402325631098271,0.7850589286368969,0.137974445010556,
 0.39144063000611506,0.6763167560283224,0.021483004954392593,
 0.5183721080150566,0.41068516876570604,0.5905492530317942,0.7009703173537295,
 0.870453711766502,0.622887237763944,0.015274696741814808,0.9295864636050694,
 0.10100031550297084,0.3194162470595904,0.11314640848253432,
 0.3641159116658904,0.6259279844153527,0.20971540306600345,0.6012357850923516,
 0.12122899533640374,0.07467635513253601,0.904159057498643,
 0.16034025151057096,0.0065733916045788154,0.053771265573172355,
 0.1974114727480295,0.5783390217751324,0.15686598465817625,0.1379092935466143,
 0.5635183931641428,0.332332384085436,0.44073305431835863,0.8213879424584233,
 0.2965261964071588,0.19989411255279044,0.7216503900734512,
 0.26705966842097395,0.563664829451891,0.38750169555256375,0.7252100747288485,
 0.6282547814672566,0.7078640339475806,0.36909937478052124,
 3.215298197630778e-4,0.28685416915904594,0.9574089036060986,
 0.009346490173099187,0.3498435326378406,0.5158240566371253,
 0.03460650090874018,0.7763931129265177,0.6859922097175201,
 0.22476509942384315,0.7539227179973498,0.9697113659059876,0.7026341745385729,
 0.571327703165656,0.9011448497223332,0.7736865083774971,0.14804771325069932,
 0.4411222772919179,0.6372414463168713,0.07018382397104783,0.8120700988155004,
 0.7219397541358094,0.8922747544942523,0.2852971420545186,0.873055900153553,
 0.4036813233696277,0.002738936251486246,0.12451264514806559,
 0.20066750801737776,0.41425647407676014,0.42287165093022305,
 0.40881231305944576,0.6589402508866467,0.6704069412932568,0.7005289288694986,
 0.6299909175190236,0.7541523604310372,0.9115216343081636,0.2782684670701453,
 0.5937683756931439,0.7027141963747576,0.42231050474753407,
 0.14822157965093996,0.48414327809770785,0.8167402528661944,
 0.18744195132375163,0.7385195703564104,0.7539258563020006,0.5634821537550386,
 0.9965465369015961,0.08821383082366285,0.2787792384536347,0.3684948559468545,
 0.24394316178896347,0.7512307492104693,0.8719130756655148,0.4598342523169108,
 0.4307606763303078,0.2699658969334441,0.4857602992586114,0.9854721164773825,
 0.8710075559232412,0.790183439669053,0.47079801568683544,0.9489658978900586,
 0.5140468506158102,0.3418852313833559,0.9443433052117369,0.7833398828248321,
 0.3420285705285385,0.8699435874139403,0.9762365141314826,0.7036552052138713,
 0.6627052335463096,0.25226350986281054,0.6231419031983404,0.5660932145677836,
 0.017467790873193945,0.2473424312604866,0.06694367268060386,
 0.07297362401008423,0.9981409419231966,0.8382332036274927,0.6504062935664111,
 0.4012160074454164,0.8017253042848306,0.6743633566026654,0.6652442531916416,
 0.23197657011479156,0.8720080934232952,0.8722081475571705,0.4568821690239606,
 0.2418603986064779,0.5456140106317386,0.436858495697112,0.09674756375923599,
 0.3703069519295842,0.1458836642542758,0.6011046892552387,0.9571853795062067,
 0.4284971728289211,0.046723007417937557,0.575733772366355,0.1821961549120369,
 0.6751771666344896,0.9048740860238405,0.6209432353937477,0.7522624437711118,
 0.8555805364434005,0.47838850375629594,0.3392783011640119,0.9942568227915425,
 0.15778347026674777,0.9507570994435451,0.24385272105983735,
 0.035639655816364746,0.4918447533739866,0.8898894419984628,
 0.4063505425986773,0.4923504741525935,0.07631702335520207,0.5309308628088001,
 0.9105000161771827,0.5183946004780218,0.4630163286873913,0.7258227629879092,
 0.2602739157995685,0.47985089998864483,0.3518542616294116,
 0.05066131859762957,0.3935548931136239,0.43505355495900744,
 0.37576789080669215,0.15863531215274373,0.8986540177979007,
 0.4453504006926191,0.3265776061763396,0.06537082909312486,0.5525117443706744,
 0.7676699794616669,0.7613750351278025,0.6443680467474006,0.9083563916663049,
 0.29974428367548833,0.15819261332484114,0.9574053277287106,
 0.4170598085604764,0.2161592769206715,0.028933603661241425,
 0.46838766670282095,0.2538842141601695,0.2592817483526427,0.9196400714522712,
 0.5574222339777825,0.6883320189842939,0.19819326330082077,0.8919232276395387,
 0.7321972876736348,0.47846928075907336,0.7567510689702538,0.2751441594836931,
 0.35089655951579257,0.6786969876565572,0.27174437339960344,0.637176109374809,
 0.2790453653055489,0.13910845368824254,0.5068716863384528,0.8292437903687973,
 0.8752956054635819,0.7111047609903103,0.3398601432491777,0.7704763366824298,
 0.5102591246942527,0.7760777977819899,0.5539051041037756,0.7106094707068228,
 0.014031568186927967,0.16314023847751513,0.49065898455225077,
 0.8588024953543945,0.8354117049085936,0.31244260805937707,0.3159550359466531,
 0.689620485521856,0.23134198181819177,0.9260575274636963,0.23141948896337972,
 0.08822357194672559,0.47324208298519843,0.2279030242664669,
 0.6868178391499415,0.055594280604632895,0.35655275856169133,
 0.8310832687912155,0.24084833426400587,0.04169073734594486,
 0.6377866226331892,0.0036918940337649087,0.46597981042897196,
 0.02937176023277377,0.5611252570995896,0.32176876827973233,
 0.0859053208379098,0.25311496015988655,0.668095157727532,0.547600110603438,
 0.8363802301172545,0.8572058970795393,0.7256616440043987,0.41867542298106464,
 0.5632824065688541,0.45646377863270426,0.058199783820219064,
 0.03279163892546744,0.7405583986387542,0.9729087188442893,
 0.41331495928748296,0.7413120783922001,0.2963533263980056,0.9371271614850399,
 0.8224556244593254,0.49124510085781947,0.18433434548596606,
 0.2621268810983509,0.9331259037124414,0.5797424965703875,0.2955743047731568,
 0.6816910834150175,0.1171379162774534,0.1600887310536666,0.6454547639090605,
 0.02055121833230067,0.004482421918940327,0.7471421609209321,
 0.46226907924172145,0.4965264113897553,0.7126756743575487,
 0.32916662401955743,0.313818442149304,0.5204787024587757,0.8042389426211627,
 0.2618556369952998,0.2235714859867718,0.8391904212874705,0.7571145688144376,
 0.35550509122379415,0.19376494802050914,0.224962239848723,0.2705142134786853,
 0.13518707973313582,0.01811212735580292,0.4856162239372459,
 0.25620973814091674,0.6307932523142118,0.8556333121795422,0.3093503084068633,
 0.9332922819405038,0.5469925444816254,0.7099412646026615,0.8263832929545685,
 0.9724989131703647,0.6815706606493926,0.6279287322200242,0.3813968234609221,
 0.648769082641804,0.31345575893911437,0.23539187130584216,0.4214616555938189,
 0.08811571683971686,0.249359772659125,0.7373497161415755,0.5103557856363882,
 0.822056211236552,0.27242921882000415,0.5392552849142008,0.5311395346202047,
 0.5633536015178289,0.8763728623296306,0.2631929061434243,0.41500521409560165,
 0.8095065061460579,0.6945501222109828,0.668037038368231,0.6705529500855814,
 0.3728125987691622,0.9046595120377483,0.8579877551229548,0.6622215205461963,
 0.788059283740111,0.8658733989933094,0.376797758416323,0.7238277311850676,
 0.6660737246118461,0.5214552805359048,0.1771215539905927,0.8065449238208079,
 0.12588619134546708,0.3708659077199743,0.027689544920875275,
 0.5616682403723671,0.43804280099090953,0.3366457801576016,0.4173585091966683,
 0.5621647246821309,0.6733866814240286,0.46907172676718223,
 0.0014153701418511133,0.6394750894308007,0.7000109404463148,
 0.008634361008756475,0.32869965268054013,0.20289518445231303,
 0.18683451713217802,0.5447977762573196,0.8575620620890767,0.5776446455973773,
 0.9579863717877984,0.38533784742580746,0.1021771594165416,
 0.016450347210008043,0.5937774839609866,0.11015919585194289,
 0.5768140764722449,0.04954978662317344,0.6035007884028821,0.9481820030927992,
 0.8148428701883712,0.004391679747629551,0.3091949978946771,
 0.25906556030589445,0.31609492173410036,0.5558694908743633,
 0.8424490152097315,0.6561596686065981,0.32917698141735763,0.6803474006283224,
 0.24046762428749324,0.49137928064845826,0.576833357630236,0.905798760016079,
 0.1940423929693641,0.2744290718414364,0.7757226158954561,0.5330325346970375,
 0.5866232289609146,0.35569888675661554,0.8235387531728013,
 0.45133546589023577,0.37391285180945133,0.9353511649401786,
 0.7513471186929392,0.8906033387250987,0.43399981938581866,
 0.13571369610358563,0.8459228013208192,0.5018866383151162,0.6937343861490151,
 0.33800061783264423,0.45208950107017454,0.5612357173499556,
 0.2575108537816796,0.6241779840445758,0.8222319802389045,0.7842071096537458,
 0.6498711872066455,0.6049225042102123,0.1106918237519573,0.15468550400747016,
 0.005641374039437541,0.2709458618875732,0.22255234241980548,
 0.8620875071286136,0.1975669457800543,0.528828327504186,0.941019250826461,
 0.5043320778505429,0.3883956652075603,0.04057766895634962,
 0.11570113765727508,0.6213833157022691,0.8165606012516391,
 0.30571296411663684,0.43018452116311345,0.6981235281036504,
 0.6990635085281935,0.8277166160366434,0.44357477149525526,0.3169268356773164,
 0.16036010508101006,0.46157165095668606,0.6391014961921346,
 0.09908876902585395,0.7389367862258702,0.22165134236100648,
 0.1438671764351681,0.332596097513288,0.6061273722098266,0.721742959838522,
 0.028942386896912753,0.7423300511238166,0.1451383810300565,
 0.7053815872713922,0.18739742087293532,0.19362068190414172,
 0.8189618790063172,0.07864126269014271,0.7478293109121655,
 0.27038422903755865,0.6259933398881301,0.8911682799266455,
 0.14467742944588768,0.9618659387003142,0.8224993169578022,0.3358319862592497,
 0.3531655573045125,0.908352396822497,0.42892403682096003,0.37979923048419584,
 0.7401977696768349,0.6770177707432136,0.783580309385118,0.8487657556390393,
 0.4662276991775569,0.051949554503612494,0.07664344380251686,
 0.9778749911597888,0.3522563963426525,0.6394234361048688,0.7568959650442689,
 0.9596994503552204,0.33429797433438524,0.7539383542237402,
 0.36128020330712163,0.0061763620056856094,0.8738212758497774,
 0.03418526916605591,0.3345143857495976,0.7130318006821252,0.6432654255307618,
 0.9352973153359952,0.6799347637261433,0.8561605718877983,0.4574014676436148,
 0.04586922960005224,0.974061544206498,0.11120357739012676,0.7290758643481244,
 0.7805143740695824,0.7664716683123524,0.973650593944822,0.06608640570093294,
 0.514452648128601,0.042339035093398314,0.8194452361318765,0.5891748591601967,
 0.6481844351614149,0.4297973932497142,0.07600297000977263,
 0.08738859250449504,0.6357287896968037,0.8090794535378883,
 0.00994604827329948,0.8450465336966199,0.46527758318017376,
 0.06042179631744171,0.9067775295213881,0.4994507452282473,
 0.21166397916156332,0.7290922494865607,0.6050845778431776,0.6349849699190625,
 0.7679006623978054,0.8168312834508211,0.6826525514760701,0.4582231207586297,
 0.48616233971689826,0.7987850676554747,0.043756980786525346,
 0.6516157039691634,0.1953651622658399,0.5104303220182789,0.8085249910942263,
 0.06910708582557112,0.43320387704868046,0.2585694176182882,
 0.3157321441218871,0.3081074283170391,0.12882068588523166,0.5053302095555299,
 0.13819250896700408,0.1709117506894272,0.5959437657365394,0.9817819577598818,
 0.009657671239701626,0.6274057368026755,0.838325010955724,0.6686317056880677,
 0.5099822742032871,0.53063867459546,0.8959056077786047,0.7718355100345529,
 0.485409420510418,0.4885469118207938,0.7131509655038147,0.2581981661465581,
 0.2002507366494698,0.5851111629217821,0.3285743332641491,0.589717157001587,
 0.2383824478010207,0.6014802884067705,0.14301735872649768,
 0.06506368001480756,0.21897181698696322,0.14395751248443478,
 0.6807237766865161,0.9636269727536456,0.39500722763873075,
 0.44171181471650145,0.22134837967615129,0.8446885093836448,
 0.7358541817158769,0.7225780978392444,0.7106434923096967,0.9421922907338602,
 0.012973246143912043,0.9520477940366293,0.11177984628060011,
 0.0891949665467675,0.08658709156150046,0.3257823134404635,0.2771360810136737,
 0.20204806058923186,0.0862597593981348,0.44925419250946497,
 0.5663080661996562,0.8033238932966646,0.7936218269823426,0.6705194698380333,
 0.02828335706881946,0.7516269437155698,0.3512629130310031,0.913143187071455,
 0.7391930659060229,0.2705042345473534,0.7327548145907234,0.6065956598349613,
 0.25248917056971965,0.2918871459715765,0.3084288740382799,0.9245016292702213,
 0.36504125546114485,0.029574319509898195,0.5493302084373383,
 0.6040215421399188,0.7958535076496436,0.5268996835296664,0.06379478033824026,
 0.5754612511428056,0.04433999664998989,0.40543994451747045,
 0.7517937429339554,0.5449113641998,0.3274937107207806,0.7700791526553924,
 0.03374124061385242,0.394286342340452,0.5684925766722266,
 0.053528890579362276,0.12925697615572707,0.7172859237449647,
 0.9231137681517796,0.514635963819792,0.4910434176176346,0.21663307857918235,
 0.3121091775416107,0.26277403260508014,0.8996408893191318,
 0.36458956962636635,0.6528652643862384,0.9884029823199209,
 0.27748144774364036,0.18508101529735288,0.1074088615646207,
 0.6403959427225268,3.5081613377618837e-4,0.6746470781416295,
 0.29435740707327696,0.2146408855931271,0.5900643192069317,
 0.03707864105575032,0.06586287156347526,0.9142304868328345,
 0.9360384310068006,0.8083501064487285,0.8751285656943654,0.1268355643500585,
 0.4722297463238535,0.5127322442877696,0.909613761722241,0.9971121158413225,
 0.36622851712585847,0.32193562877757087,0.07106678914928755,
 0.2491840232885424,0.2743088460740627,0.5594003507627803,0.6985785592749438,
 0.09602146526620636,0.71828175954127,0.3848048420139907,0.9156808029672809,
 0.6728886572650702,0.011979404659324788,0.8502643299339988,
 0.4737151427887505,0.6120370488183805,0.39892305263883177,0.5925115941804286,
 0.7191611768546635,0.2770267424862054,0.12207080900482081,0.8104831755854528,
 0.20289473187171636,0.9926243628001779,0.29469967407646935,
 0.06917239978813017,0.49809357287854183,0.6695219148660612,
 0.7478326659517419,0.6056857484145705,0.2871762396878901,
 0.021017035877424095,0.854671653366072,0.3537947452306256,0.4699701054938815,
 0.951097989188186,0.9412652436122129,0.09772210868482523,0.7084345477178756,
 0.04597578665131963,0.48494438830034226,0.5925338589673144,
 0.6651042461024507,0.11567343203185931,0.6371302304820389,0.8738392408743004,
 0.8327838503293106,0.8327217011498166,0.9445578115203054,0.5349925611347147,
 0.5003583129451165,0.07872507738016565,0.4681633216264178,0.7753145004658442,
 0.587743117563973,0.36522119593736824,0.22451050235563896,0.9599194966691882,
 0.8237108903380748,0.1173023729956959,0.3750070039095905,0.609368492863577,
 0.2749135873318981,0.39344364107692187,0.28220622145599006,
 0.8304195162878477,0.3018400101739619,0.8299286000601032,0.7383933368594877,
 0.399708559477272,0.6415490227849681,0.8319160079662743,0.9432283956668964,
 0.508971577307289,0.39861784122871935,0.8724453059409096,0.7264145926079215,
 0.33582325425062576,0.2797044318859184,0.5535815267432843,0.1357134319387916,
 0.4966954059107722,0.6735319912599377,0.6856402824986269,0.0765706617784101,
 0.6634509201103166,0.6050767837546758,0.32537882037477117,0.8268679176681223,
 0.7349628069407383,0.6233663644992532,0.8625398955873511,0.6533662519560538,
 0.08913498354935334,0.318224041742611,0.48613924616600634,0.6031011215901447,
 0.43136676958736897,0.7057469252780244,0.7342537986088546,0.7700464614970969,
 0.07505336744689672,0.1981388822807807,0.40216466265757667,0.858633435216376,
 0.45358493672221933,0.3278645349365831,0.6612620581516901,0.7037821891484078,
 0.37180353835810054,0.5482621096956773,0.6126745517995996,0.8854072986444053,
 0.5921450284638446,0.2245314379614456,0.17574118528569016,0.7669430042174894,
 0.5846737992046674,0.3141625590512276,0.4785330062939632,0.12291084845695077,
 0.7989206064840768,0.5712020584619519,0.8533241474784119,0.10869553617045413,
 0.08483322473358956,0.5616761318163623,0.37071747860221016,
 0.9971594201426561,0.625630144976763,0.9496299965219185,0.3435347479689911,
 0.6746576338465097,0.7826086452807445,0.9936217207343826,0.23549624195891317,
 0.8970250827175811,0.9038687823677264,0.25907934324077764,0.8652185756643824,
 0.4461978659673138,0.7355470289176742,0.18910067317616397,0.7072037624612971,
 0.037007831179578066,0.09148500834676687,0.746597828936208,
 0.3835389708984458,0.2836316472004591,0.7944130188874673,0.7116355718070992,
 0.42033029704673663,0.28155160362835885,0.8236116912768354,
 0.15166633637749338,0.21896022995693265,0.6168847824776379,
 0.1258083919679669,0.9873339376554101,0.1323023300184829,0.34938570698332205,
 0.6897295763239737,0.006304565659320804,0.2249389981946749,
 0.38643081906590904,0.3733786369283696,0.5594254283329088,0.5242557537632007,
 0.9357375459815436,0.9848465303155572,0.19986036920651473,0.7058801333223614,
 0.5092753088222199,0.02235562413743908,0.7676592623778407,0.5815692891308161,
 0.2146976056459734,0.021378179970978906,0.8512443268727543,
 0.19191778569523532,0.7465245532105886,0.5883094050486168,0.6406872472234645,
 0.6399731216881293,0.4553663986965575,0.5934879543954469,0.7866222411945416,
 0.39097276070810083,0.3761832585629675,0.43544621815264195,
 0.12789906279530938,0.25732767171946924,0.8753194446162212,
 0.5866789458295742,0.1304901889632314,0.7754225161512374,0.4517442345369802,
 0.668623198078116,0.2981450089072031,0.5536726990651913,0.3538395579214241,
 0.639536250878049,0.6587805123888176,0.04920993459682099,0.3623454538516907,
 0.22004389229081,0.2897272337368151,0.22582447007432016,0.13296992367092275,
 0.6431908593943497,0.04500767510278503,0.15643915306006628,
 0.6549809607489347,0.5722935586276215,0.05896300083691197,0.3312997419186916,
 0.3481472160904894,0.4536848686838426,0.42510685299028506,
 0.03297581422451046,0.403139303343122,0.4052582470837839,0.0907190478729859,
 0.37493347590026005,0.40808402506606234,0.448551959277118,
 0.12287575030216569,0.9527627159224549,0.48395822441626424,
 0.9441508008592758,0.7362869226055859,0.9682259333030551,0.5123463706051249,
 0.8706722721487083,0.895568234035711,0.5001798877855925,0.8486530031960369,
 0.15418735675399042,0.8921858304462245,0.32795816058531757,
 0.5517746529738938,0.23497664022297804,0.816181338652402,0.6343017911091224,
 0.2668389842252167,0.7300725639678636,0.9654492225250371,0.076702030720882,
 0.7765293354990251,0.6530984848447349,0.23964398342743454,
 0.14635855452872226,0.9695220723021919,0.0844817416184721,0.5690798394328751,
 0.12607510281383383,0.5357935207721363,0.5439080175454272,0.5173651143702724,
 0.8802534918108793,0.9129036499923906,0.8150568259904526,0.9936368046060469,
 0.26821596969380335,0.6723816404326536,0.16010370485692538,
 0.6344158618366972,0.1867391764625449,0.9247365551966986,0.26497699439492095,
 0.6328594776569207,0.256255536490388,0.9406108192799372,0.4644940287635345,
 0.36410309595638213,0.8145260380302188,0.5744174883840094,
 0.21459538314943394,0.6559409576032414,0.4033835358579029,0.6002151301272572,
 0.18572046582360857,0.07600520676667999,0.7352495175009046,
 0.8965061437795765,0.09412256185048229,0.854611870107324,0.34675594419716893,
 0.9367731342627503,0.15845499573404098,0.18641934267163884,
 0.32106721712013864,0.9584942446236809,0.2701305930259098,0.8935010226217468,
 0.7164088995218088,0.20532206060216573,0.3759641394531624,
 0.09100054273399338,0.37824772574595467,0.07421118544616268,
 0.502904117406749,0.4960713527344791,0.5826285384225579,0.36217514636070247,
 0.4922509566398027,0.10833263131174942,0.597854555649917,0.5000386617293063,
 0.8872231761189993,0.24273692726360996,0.15902140131211062,
 0.4058113557298417,0.2788964699150074,0.7663202206539199,0.2134786865288777,
 0.428071159411214,0.8500713098997221,0.7575826600623247,0.0823018533517903,
 0.9956456265332925,0.6876980151360311,0.599311823656024,0.05962341047207942,
 0.470438075526569,0.16357122627019904,0.7117918243767638,0.7260586919191321,
 0.5416236220406452,0.8824481265745237,0.4255674613761865,0.2241999844030289,
 0.8759696705034575,0.211256786283056,0.21768646948339154,0.5169058913673227,
 0.9017788531520116,0.0270998447011499,0.26605898171433917,0.5982532469332139,
 0.004885264452770932,0.3520181586571943,0.9350067497892822,0.529945732882329,
 0.22750812772437357,0.16551281123071493,0.5374713188758242,
 0.7936323549834257,0.7673999614261747,0.5832062679143826,0.76366984007757,
 0.29438587614964307,0.9914421286635575,0.9384114900067858,0.459572263347938,
 0.9285707801929272,0.2071232157955346,0.8555874299026072,0.30913931747707735,
 0.07859002914220592,0.47862129841145074,0.8100377431820326,
 0.6788034281571185,0.7485728803231277,0.21441190017467615,0.7069144166262196,
 0.9708039472302084,0.8606153269674086,0.43240554566075706,
 0.015195168167263184,0.860594987338579,0.44663824105892025,
 0.4175654570487978,0.15582104225769056,0.20338545248710416,
 0.26008464542327303,0.8152551694203636,0.9728411537447467,0.880844713180863,
 0.7621038758803816,0.4232266084973213,0.30919534118469916,0.2199815361738089,
 0.8316102483771673,0.05346463242501398,0.10675011661467815,
 0.7382390886327942,0.005471019557652945,0.05515433648778845,
 0.15810340059665218,0.053163809653482,0.3795122250869273,0.9435103851243112,
 0.41622191097367445,0.5358371279682449,0.6285714518722766,
 0.45758172098596006,0.492999702207061,0.8382860435514041,0.6481797380072073,
 0.9347625046051762,0.9253170356191588,0.10016838607150613,0.3191457752067923,
 0.7203858314307185,0.3113075118644705,0.4750249024572182,0.6915456550352832,
 0.6900459519877198,0.9085858173547746,0.8892385128730772,0.42315812274221376,
 0.28486753274424326,0.8404718125090177,0.2571632170117101,0.6412904335490139,
 0.9424228714680093,0.025800624816172957,0.9864915143546179,
 0.6223397439011153,0.37349281855841865,0.8240672691534175,0.6427055895869672,
 0.744018984211494,0.9033155120734842,0.7858002523917034,0.28057046407157316,
 0.6555897256244434,0.30597097797333517,0.07309195474401231,
 0.7230133379302403,0.2647467267215603,0.7559221338505431,0.17610673381494057,
 0.7139788068176363,0.13540708354881703,0.03464995957796657,
 0.20724372776415068,0.9301248781355036,0.5242830867101336,0.8955930420823333,
 0.358210028282847,0.008329098504012755,0.8327457931277595,0.639091136661141,
 0.9047868788259104,0.25269585326318933,0.577332666127289,0.2453313107072589,
 0.39663713494361486,0.5585060536931546,0.9016469583681523,
 0.031324727562397126,0.026646312190149346,0.8161563268446952,
 0.35444445376256994,0.7937231175856433,0.2428692297499514,0.6059143780153984,
 0.4400493819581076,0.2637004908027839,0.2989211377335722,0.9051456429721412,
 0.9800297956792984,0.9243936962740233,0.7226890272484121,0.43693827200250523,
 0.3556351278892306,0.039495887714686484,0.5478450182622768,
 0.021916408959190226,0.3038350335011073,0.7764055209654602,0.681826831354992,
 0.08427842105956262,0.007813189857500347,0.22505332985603232,
 0.1399506924097863,0.13405371857940995,0.7996764083389573,0.5788008631985962,
 0.9034184715005044,0.7311321320932068,0.7700241673133735,0.5837653116715678,
 0.0995368223420885,0.3967780272071233,0.5365055371555332,0.8409761165406121,
 0.23135215969842804,0.3291218912350984,0.4542769734852564,0.9582600194868786,
 0.8079361896439994,0.2867214881635528,0.3418619489031476,0.12488595303152406,
 0.2906192377909388,0.8715157300427924,0.5166634454245543,0.6935842172298904,
 0.722559634023418,0.8470133331887115,0.959846483729427,0.6056797901091463,
 0.8833265976712837,0.03710728488231019,0.6233188544635709,
 0.15811982623970555,0.8088799002832218,0.714240848233477,0.680776847184806,
 0.33417241309169676,0.2786597981833766,0.29065241498412464,
 0.8899905622721551,0.37489176812356284,0.5808379089382433,0.1773714581404835,
 0.08830439130242984,0.7054743089547489,0.7292256408549411,0.8837617800036028,
 0.6653783917790143,0.8518620222594435,0.33531028593435186,0.8289238721346752,
 0.8674445335580072,0.33032567086289244,0.8948237128159732,0.8626334643754806,
 0.47871792335897845,0.9123975284158239,0.9483224528380562,0.9421025257564283,
 0.2139328374520647,0.16090975561334953,0.5663266449632758,0.9968830904705047,
 0.423377366724658,0.7367690302427368,0.9113255435747166,0.10445005524033268,
 0.1729496198059356,0.9505287910133916,0.10321426021158808,0.4025615181731398,
 0.5994816055180398,0.7961104897694742,0.17407460341462588,0.7265262107786868,
 0.13613577726810888,0.4897617257769027,0.7552982468567042,0.6168566401760932,
 0.12841554705826663,0.8187115737349576,0.31163609652233315,
 0.5838795420916457,0.3630760345236129,0.9915483659407186,0.3258218386649865,
 0.7701014370357699,0.8825227586779194,0.924299601203716,0.15153223720379694,
 0.6637415214213005,0.2774321248876399,0.37676386587290833,0.8175741659471418,
 0.17118184280774873,0.24048197829615148,0.10364447084974437,
 0.5674435189078765,0.5341217177283203,0.44589782293878444,0.6976284637393905,
 0.9932299459714616,0.6880137659379592,0.537228704719583,0.7809360091707543,
 0.9229059231458926,0.026569013856763468,0.34293685795785356,
 0.3900325154861781,0.6548476557015086,0.6477501148169477,0.16140224392144653,
 0.9162849419046989,0.7296524397609156,0.4358162062935944,0.7085108672571226,
 0.8748738793319945,0.05226041016551686,0.03740764586002643,
 0.30269405807578587,0.991535487867441,0.1559126917421504,0.9246511780365776,
 0.23566471688559743,0.17084347053386062,0.007262338418472347,
 0.8835538051250736,0.5584056973549757,0.9330068349185752,0.9575935626989446,
 0.8011655397091056,0.026456391089286435,0.8737415562286401,0.598613093849216,
 0.648042438963653,0.7348956109054962,0.8370511008264989,0.4150413437344046,
 0.3427925976240661,0.6287497690139278,0.9783802623399525,0.34621803955705244,
 0.24903300363456515,0.05108286332824563,0.28193765515536484,
 0.26513949787108226,0.2573370792361912,0.4313641103050435,
 0.32025779790677134,0.5092385190423775,0.088505742257452,0.754425116126111,
 0.544358575429475,0.0605143650493436,0.6418480777804699,0.024508690347322837,
 0.9337955034067884,0.23850260620383912,0.6377657691252175,
 0.24957057886286005,0.3094764255493545,0.23866847558027582,
 0.8065483284558563,0.3581855694297391,0.3874845141301092,0.9771967678541087,
 0.8335055056738752,0.8443390789895443,0.5798502933234002,0.2352445270180703,
 0.11317952701033573,0.12571661674967038,0.1820735643861946,
 0.05752831347525256,0.014304249441124517,0.2352441059843825,
 0.5621187771116779,0.28211421933301273,0.6925771855698918,0.6024789752831597,
 0.45460956877643965,0.20241730713530404,0.11336262917138218,
 0.7711330620775793,0.7036819740331479,0.609662210987073,0.232397627867734,
 0.6694170844179571,0.8991957147486606,0.9513735082565489,
 0.008965066088062468,0.0537237298698805,0.7272185514572702,
 0.9552282648767702,0.11508306784092637,0.9954000692106055,
 0.38947322910451865,0.8395539445848531,0.5073141784356503,
 0.30445171373323654,0.20140152859011384,0.4670619893556005,
 0.3234748059424243,0.5099531354482476,0.713744012916647,0.675062897133235,
 0.00444332892841548,0.23436411855024064,0.7228076715374787,
 0.2888462181841195,0.33210107628674346,0.167367560147334,0.8761771645743921,
 0.6258503740343502,0.018578001193785054,0.3294662081401146,
 0.31370219530982724,0.7310639862353456,0.6247694358103082,0.5385131055236111,
 0.17841415623416879,0.8682066699472502,0.10641944328163133,
 0.5999535740944866,0.03679117898231421,0.10009537030411053,
 0.4778960441600052,0.7769343395927246,0.23013805766923034,0.1474660491065568,
 0.14872340354275526,0.6618917833099758,0.7152521687111615,
 0.11945225434209683,0.198907637171006,0.998277882462232,0.7484851064042113,
 0.18247947101647688,0.40565875947230834,0.7903973101311138,
 0.8101517967869727,0.11483810311752585,0.26398854367206204,
 0.5614894565817649,0.8748851224837355,0.6368768530053417,0.588322735558372,
 0.4907058724060325,0.42393683515352354,0.6322310082513064,
 0.47965334153855954,0.20161135444820122,0.4915239159490695,
 0.5383682695500105,0.67549808776287,0.7000117157891601,0.9987775662841789,
 0.846565998524803,0.8531426609661115,0.8938404248690555,0.49476731108736793,
 0.8513041187799615,0.135335105857885,0.06086353283395762,0.8118528226637319,
 0.8078136267254153,0.5176339455342623,0.6484296153822998,0.5316997342966819,
 0.14907648346125546,0.9251602704596421,0.8486595495918415,0.903251332097359,
 0.9056261015015357,0.4175080977004856,0.9876438589271967,0.5583899829004443,
 0.8678238138807841,0.9811105865422506,0.5686143041080678,0.6256815928928361,
 0.6337863483311241,0.2857026109254477,0.4869040370062563,0.5405905988360252,
 0.4480357720964875,0.18671433646881053,0.07141898413085201,
 0.6567000307131757,0.026893473966412396,0.8074066769571747,
 0.29321487510487243,0.8135329228349162,0.8347348640623313,
 0.42115006480633577,0.9838493904571575,0.3054716293698364,
 0.23763180163030684,0.8352550531729375,0.6375867814744112,
 0.12946085405661112,0.8622077725372993,0.655057549631675,0.690332246911895,
 0.5338477788959711,0.6701233333260851,0.08942424156057749,
 0.08668425795176526,0.581747939327991,0.23792482947779725,
 0.42305209496577734,0.11519328058586498,0.46011981766144505,
 0.5870579531715885,0.3013270197524902,0.6928177551402501,0.9779132503283549,
 0.26081459436428023,0.060352656870308774,0.2631899254293184,
 0.8094986437854458,0.39265915517674854,0.014648774109593976,
 0.3772472079909902,0.08475739603878196,0.357452083989226,0.9749626084197158,
 0.6285758198805608,0.9426944005092474,0.9994162445100981,0.7426609922544609,
 0.3662394847470174,0.7577384735744501,0.5195851067364969,0.5476163842223073,
 0.2849287922131082,0.956527845365418,0.15365632938487406,
 0.007514503356944591,0.08765064473868966,0.9863849683879689,
 0.5386674681006096,0.46211142907623115,0.5941247161264523,
 0.49600295571856257,0.2653788912853917,0.6689731431200122,0.6240543642256776,
 0.6206225899812645,0.7224348082255719,0.11750787678713714,0.2985422091830341,
 0.7979284238878213,0.9882767754406322,0.7880231421454529,0.34175627749835424,
 0.5836842101112734,0.6633655268861922,0.6994256577874034,0.2799232034041296,
 0.33437827106201023,0.5695478517478781,0.5820365498065075,0.9964688498698746,
 0.6687686881877455,0.9039013320210182,0.26233745229320893,
 0.41638781136599134,0.35361318589076773,0.38246594339119344,
 0.7405628523654103,0.7164566429416116,0.07370509804965641,0.0813166839217716,
 0.48867640130377665,0.05040092716039912,0.4723624904840854,
 0.15476688667211036,0.8235605121249503,0.49697505404325715,
 0.5949065211511279,0.8824845114296825,0.9863947953665269,0.695664126459043,
 0.6745022281720248,0.804343113176447,0.4345173914948416,0.43815992247721336,
 0.8100193525979652,0.044382269587898926,0.6696581209127586,
 0.5804354795464391,0.04235826293401135,0.7844214799691337,0.5220903091045171,
 0.3185487179903884,0.25154147181113373,0.45750434498584847,
 0.006286263085506061,0.8487407484404439,0.3253739664913262,
 0.9111589008248054,0.5390843800922199,0.8089871441590397,0.6944180180811123,
 0.8691832780261408,0.4129189719278372,0.779843003654536,0.23596922289247624,
 0.321761550432066,0.9665500500969277,0.5135410554808202,0.691039006441376,
 0.9631737923710135,0.7582783426502449,0.5612501976217331,0.1531683573209166,
 0.5352577845274259,0.4661187196610177,0.5456086011590473,0.5771888639265373,
 0.669291789497924,0.5901425555456067,0.5743542450533758,0.6129796279103327,
 0.13703629199885448,0.9657486672353749,0.2852198631078643,0.3421501489601232,
 0.24696633107976274,0.42619592727601907,0.5730146623346162,
 0.23624067538778137,0.9271966809054587,0.38688155841429983,
 0.7321989189407963,0.8910636341504418,0.014450884251745233,
 0.7516627003956917,0.8801634112075797,0.8275324300370941,0.31520504690023843,
 0.8752213918628401,0.39635189982194563,0.9507734393418629,0.5648445415840339,
 0.18123452843517474,0.3961525847111095,0.6743944653064851,0.1185043509942415,
 0.8661420078824293,0.06470057243017702,0.521072730843081,0.2541864906483027,
 0.7867549628113609,0.9898716849768666,0.6821104778383904,0.6426132184029589,
 0.9138911042078495,0.44852118886263825,0.9640498706995677,
 0.38499661207176616,0.2831433649549825,0.38109250701157726,
 0.28748182230070063,0.36704216249804533,0.46398257517415686,
 0.9459709773604217,0.39694042302675625,0.2633491083793589,
 0.19286440671900063,0.8561921929505278,0.757411811103131,0.24155131475206915,
 0.8514220146167528,0.3365261838765662,0.7964062789490001,0.2631116147557082,
 0.15172737820809223,0.5825855050396279,0.7581562986638822,0.6005733101382777,
 0.8025761491761962,0.8042019557995785,0.059262100896300685,
 0.28005500227165603,0.8506390051951461,0.8315558184876779,0.6499076700541081,
 0.7217311164464977,0.048330650672143816,0.1921630895008375,
 0.6742347758691234,0.5930041435542933,0.4348881852028572,0.4797212513426069,
 0.08588687081804602,0.46828395139776224,0.7844236167886589,
 0.9427593906448235,0.9321210272416613,0.7487699853161329,0.23079895490574887,
 0.2703853231921213,0.3646801558500562,0.28368215895612825,0.51818602285191,
 0.8752590721149303,0.2068519688261342,0.41413908957863865,0.5464772635898919,
 0.3685047524781139,0.18678451645698568,0.7514496868527036,0.4225633102277522,
 0.43521294335284033,0.9911772976278441,0.05843110734346779,
 0.47567270787917804,0.5741847318843288,0.6208372535873697,
 0.15412678539758695,0.8595707859202603,0.39365373280386684,
 0.6632351264260198,0.09158308368932755,0.45556752322897565,
 0.2512187676288964,0.9615243492659757,0.9863422892702843,0.8435490313380298,
 0.7831234286654666,0.8678414888255197,0.5411915382967426,0.6340846102057439,
 0.6641845029237782,0.835720773846454,0.36406735892786024,0.8659075347403679,
 0.07504311800028796,0.7250805897797689,0.022733362679737135,
 0.8199920311776738,0.48257279890364535,0.7000829560837927,0.8309031184243859,
 0.05567914364587756,0.014793824303007908,0.6745741536172715,
 0.7872537699840286,0.8942072939345993,0.9173179427106153,0.13016706618189577,
 0.2910143431512785,0.930039915232789,0.787803415285608,0.9986049302844163,
 0.4864509378274374,0.7015489841592741,0.5631857780483762,0.5678059573393969,
 0.2973608366576196,0.7193481930057107,0.5388471581872072,0.3764722726948324,
 0.8386334413654102,0.287571177699941,0.41633613590429264,0.2124039793481155,
 0.5654788474001492,0.6295829681183867,0.8105799484403482,0.9439183476339625,
 0.38090469641868496,0.5207640698226935,0.5261137713067853,0.5981913745992984,
 0.7432106412904376,0.5678326537547652,0.7729416225908115,0.2823389927489234,
 0.7512109768401017,0.6945618977159577,0.42092053557801434,0.8195402889301079,
 0.2419349644314046,0.0632784867408338,0.6365043025630743,0.8806379324403606,
 0.36255876193999115,0.7190694558532635,0.923561586056456,0.6371455817334892,
 0.8550035726728801,0.25989434283439994,0.2890038487253628,
 0.40395448740861717,0.19856809346390691,0.40537357025153486,
 0.7613341257714791,0.21169097502337253,0.9112682935482029,0.5514447112334604,
 0.2667402723772151,0.834045645493478,0.4117949456734391,0.1867621946669379,
 0.3456729143676881,0.920983075013315,0.724956840892393,0.718391831085892,
 0.015336501948131298,0.9474540092061899,0.8894344670946179,
 0.4441445788494691,0.05433090577906663,0.5218228710874486,
 0.47617776015293556,0.6915409171113264,0.5533408265556601,0.7026039618543439,
 0.09234045860679707,0.5762710120983369,0.415304183259656,0.9421444367051712,
 0.22112787529871136,0.6555291003111456,0.5659416544440874,0.9510721364722341,
 0.6698186043419682,0.7986404092940483,0.08219720192695834,0.1261041064430225,
 0.9817651015499433,0.20348782856392467,0.38317924307285345,
 0.3999074412678969,0.012334015539081866,0.352585258991681,0.9165654222019752,
 0.11306033796947046,0.07910289854510377,0.15111746699634399,
 0.4019600211248586,0.5448389314752146,0.8169649732605259,
 0.047800830633132785,0.5477037750850546,0.2519306305866408,0.776660100196201,
 0.9331308341403317,0.43524485240216504,0.3121327220913449,0.3457890821351023,
 0.5671367067645875,0.22763900480952914,0.9710444053725642,0.8801890185626569,
 0.008167914924396613,0.17346142094678552,0.5861681603435628,
 0.7902327818863607,0.5858974179652873,0.18759540631338678,
 0.04410037463825878,0.47950988089000735,0.014282683027940113,
 0.823390564361175,0.9700572012288345,0.41229858597874924,0.31591424568337445,
 0.6174772507158054,0.018458151057264116,0.8723341556515459,
 0.6679520121553022,0.8044028531599466,0.27491433468825155,0.981749260757034,
 0.1842733876438708,0.5386173098943825,0.4648516005326979,0.3076291352491185,
 0.5174793097133554,0.7201124535521631,0.38397634619009025,
 0.22000491955011026,0.9233971793109071,0.08688963937076144,0.956733943679095,
 0.292808632412361,0.6583210535019735,0.6535294286076708,0.4501716397022597,
 0.4022905821359517,0.8474608188789067,0.9336996573246427,0.6973583751223309,
 0.28559998915614304,0.9279777141029965,0.31164861157787016,0.372707239387284,
 0.8059372689967194,0.4311027802883355,0.6876338234280635,0.06713353368421571,
 0.5541326902068824,0.8309052858159999,0.9897145939604396,0.3356302939537482,
 0.5446352706992303,0.2853333578383107,0.9589380835343819,0.14152856033068084,
 0.4486306747968781,0.4791016356709997,0.2864412272629444,0.6088029597095088,
 0.2906102536692493,0.4001250003406356,0.8459042377152906,0.3574781952509297,
 0.9749020708266439,0.4812838157314767,0.2857149248430991,0.6947453947469975,
 0.3306259693538677,0.08055959691480685,0.515670706906364,0.26480921702186244,
 0.04289372538679337,0.3287413304994965,0.25100877868831284,
 0.048819227537576726,0.20176839199783803,0.6271325216380548,
 0.6328403885709171,0.38711975138451715,0.8067831295517349,0.6379789735232976,
 0.18508174354606122,0.7572685048651259,0.5126345696510434,0.8691673965713178,
 0.040162411013841326,0.9095254192088853,0.6251549681380217,
 0.20132839118037593,0.607344944988718,0.46943969992842216,0.5340746460150626,
 0.07758139730779279,0.837709120731244,0.6536009415392632,
 0.051073241096560196,0.21077595042848207,0.6384271246488326,
 0.7144771597095801,0.9342261855118189,0.8822770992596295,0.4445689177092449,
 0.6138982693731934,0.5679073641676065,0.1498409998963356,0.07132052448538972,
 0.46234055463727375,0.2045706955911859,0.6840636755158604,0.9272347876006755,
 0.6293686425961272,0.483115243588055,0.9545782141238944,0.8839681077419168,
 0.5431877730563605,0.811633777052993,0.6721176506685317,0.8688809435329715,
 0.6572368210190984,0.32782202472971933,0.27377005272870814,
 0.011142031218186244,0.06692328338602282,0.05680290616597672,
 0.175465780065989,0.8790396106570625,0.42672269822731257,0.06302385682470368,
 0.3590315631179135,0.9133306874517981,0.6585255840334983,0.8377407119743735,
 0.2271595982168192,0.2731036007051928,0.653127137577253,0.6651108313777306,
 0.8938437792134871,0.7303582012022998,0.26288045433033835,0.5072445690910967,
 0.22840711109610023,0.7092447547437559,0.23425329155638064,
 0.08144445379040532,0.35166838762303654,0.7370110474813041,
 0.7319927582846845,0.5600503001670954,0.1461090846041576,0.4206508130156934,
 0.5773371776437873,0.22498918025843206,0.2108580847222744,0.9568579911,
 0.3941723937711502,0.23302768814473784,0.9540386092513595,
 0.26189161357532753,0.6126926131648319,0.2277537026427302,0.5317565222364546,
 0.47841332371305356,0.29507166296188336,0.06330178429423494,
 0.012400571953832262,0.031625651537107724,0.04161574536563528,
 0.29207988315172884,0.8195575646821641,0.604941660369391,0.1533079266476387,
 0.23596863847882232,0.960863633689133,0.7479748282056119,0.8363084974707268,
 0.27587284181588934,0.6597875310419097,0.5800699164982502,
 0.13599110693451677,0.23626439707299818,0.8944788366121165,
 0.5296162191739828,0.9094381790201488,0.7885315032452079,0.3132275511327055,
 0.7382428912662845,0.7041772950573957,0.2713088522233884,0.6748554004993234,
 0.9519057493071225,0.8564908870233744,0.7865181219682982,0.9474217146026451,
 0.320663118478802,0.177970327290782,0.891969756405102,0.866349289221561,
 0.2147798253735289,0.5727688077247199,0.07505323686050203,0.9402084940954955,
 0.0057602603616384584,0.415001748912348,0.7211647340605253,
 0.5990909614935123,0.12291532778076886,0.7660457753506567,0.9517004702968361,
 0.8310115586109379,0.312540123098516,0.30433710430891037,0.16110556478625915,
 0.3439451437078349,0.06563076572051829,0.5378183874147426,0.9298847089505223,
 0.4194342254750085,0.7366410027150899,0.870574357534775,0.2230030890182203,
 0.8888654779073144,0.8707721038875851,0.15026893681026599,
 0.47047584786879204,0.0889766950743246,0.7513433068220975,0.7267049872017941,
 0.2786541723451589,0.5960851254813162,0.3768485013610863,0.0439489373859373,
 0.032991953565636445,0.10220156105769729,0.9371147942832339,
 0.3527667381950115,0.18408849057462473,0.5523501382892095,
 0.40941227900360244,0.12811247209252563,0.7315340511025408,
 0.22912988132280976,0.6287856919183994,0.3798739659571151,0.6544294160633576,
 0.8939677746126258,0.14117059854069036,0.5647489674600963,
 0.46635477217342125,0.5767149865520625,0.28492111590430547,
 0.8949478505205408,0.7803479302753666,0.9563825307399458,0.7574479603448362,
 0.7096764927827962,0.7574428025286888,0.02157269312420451,0.3654708847852677,
 0.38783850789568763,0.9083090598561817,0.7677412576530257,
 0.11577531125529228,0.9984120365600967,0.11950974692233962,
 0.19721108792412834,0.015655459947770622,0.11334173615824139,
 0.9417734667902171,0.9618204145351121,0.8139544117672167,0.1711680198905543,
 0.501864703591503,0.4465208148527592,0.026780113194198596,0.3011382181216249,
 0.48626246538984974,0.12093617209949936,0.444521520825914,
 0.16482787977056113,0.16565485923754086,0.46077465253889605,
 0.855912795928864,0.460709777455643,0.036097063742631885,0.6623533764243403,
 0.9398715852464598,0.6743526402869487,0.5163241215254609,0.4907655645261205,
 0.07492356906108966,0.16989399480035527,0.9699703201003174,
 0.37688980099322034,0.2829924868022198,0.16789273230544022,
 0.21990937023056056,0.7179986352386672,0.6199677364832287,0.1844295542476102,
 0.5814604156951078,0.7259980921638896,0.7488460923388326,0.05380297810807377,
 0.5287654745293678,0.3021033400421076,0.9434628127716257,0.14882691967421557,
 0.9960016894078314,0.11159344666325122,0.04363759485679941,
 0.9716877558299278,0.1381381779074019,0.5792541468636326,0.45060207085183435,
 0.3828318712236296,0.3053390728941916,0.9538918761393399,0.839647770372409,
 0.13597097990476914,0.38802731082141895,0.23680491811666704,
 0.8259842032428429,0.261721960233834,0.9532220229973298,0.4946101355415252,
 0.732900176659253,0.6014820304838515,0.07239196324273789,0.7886369795654444,
 0.1764449238306235,0.9465586685284139,0.7976522425083553,0.2869187806954373,
 0.48037927289763027,0.6327774634007537,0.1964057843202226,0.788022788844183,
 0.9814233418545247,0.8270414657642616,0.5377209194309195,0.5886619204655669,
 0.7035267082659488,0.4007848036860777,0.10928560843796054,0.8813610426446359,
 0.9471418878733835,0.3106618612057329,0.7115345575619403,0.1040329151557271,
 0.46903658466015385,0.3229274839915235,0.6807079122493005,0.7841114031706178,
 0.8326163058238213,0.2713234556734432,0.14221400432241016,0.674970023039243,
 0.767966182333426,0.0503068101401537,0.705676451366409,0.571705800726624,
 0.6145597604758926,0.31158672402067844,0.6544218932421785,0.5850963936057104,
 0.2973208589080254,0.6607668275443308,0.8096553833474105,0.5258788999601003,
 0.8613022113908748,0.21613013065982745,0.9710274400789798,
 0.19976144133421037,0.25059837429947507,0.3704920222168294,
 0.4006166406794074,0.9351755564812079,0.7329320150344797,0.26526307694516393,
 0.8914634178519967,0.5341819698579767,0.24641738603914476,0.8396295267535134,
 0.7904768863260274,0.6808019687654766,0.047530562532981824,
 0.42080422698476583,0.8107945568103013,0.9307248134991098,0.7252549165025899,
 0.7603989169094897,0.6932500591866014,0.6098896235340928,0.6206866968362006,
 0.7960419950682732,0.8842163121426269,0.25265929468922066,0.5725972147067256,
 0.2756956380155763,0.1796679217792908,0.13747591749592514,
 0.10028974463966212,0.6052831856564898,0.6099312680755897,
 0.06669653463857783,0.9693819047812309,0.4541297110695367,0.3183717416097793,
 0.9038833456766802,0.2315034751311591,0.03538687333221491,0.7650727876411698,
 0.6381846393079463,0.8022186744512246,0.6053711837839192,0.05304629780646497,
 0.7397762459954993,0.22079156386548549,0.07276825446392443,
 0.2807034444356289,0.6201980722513226,0.6581870714453107,0.6136066526835016,
 0.514785699305154,0.12046003791968896,0.31427390903142094,
 0.34329378316286074,0.44439205446273766,0.1260454155374141,
 0.4925103859452218,0.30424298618511714,0.7066034671026202,0.8350653030616516,
 0.16398390353750636,0.8443937816307225,0.33100801073788144,
 0.6148292038613016,0.9091734802902077,0.5086755713585566,0.9323175541607904,
 0.8140693690391593,0.9205308163293987,0.8573571377219805,0.35140801595742677,
 0.6879060111080538,0.9936807651195637,0.1671069445876847,0.6299175379522274,
 0.3780101556049815,0.48649268694699077,0.8710619537267349,0.619336084421946,
 0.37916588619230196,0.05707874202471863,0.7363453866659005,
 0.025976161937392495,0.9354121338086384,0.8350861505925479,
 0.9402713956103114,0.8761613246957412,0.16151974527309798,
 0.48222617718464555,0.014110995284118633,0.08847462159267361,
 0.24443889971035793,0.29368930386039915,0.9526332150436994,
 0.9151882746934076,0.6947140425428957,0.30344965835667304,
 0.15435687369077145,0.4670025111414209,0.0031820224752195925,
 0.3608303237107281,0.5534084663707699,0.6239849173573356,0.29830532051698044,
 0.4132219167273571,0.8955504085051852,0.09292098458224018,0.744922050779618,
 0.38082293030745507,0.060713108138278016,0.499905071828028,
 0.7651692195310211,0.6932497422692311,0.7543957440731258,0.911290924140237,
 0.21046165874123302,0.5034226117160359,0.36791929746717866,
 0.8783833934564678,0.4615037222230701,0.3805167251122672,0.12975178131791765,
 0.5667031330882919,0.10689920299887401,0.47875066163319924,
 0.5661902029349589,0.7546821729725545,0.11496619968034405,0.729755087593282,
 0.2882569576626849,0.48025732552763567,0.9203999136564796,0.1636636046557054,
 0.011280550204945339,0.7042581469967187,0.13584734109147134,
 0.7331841113086451,0.4135128112363634,0.7091801672788611,0.9849195124532365,
 0.23188655055712026,0.5516354706389451,0.939868417107075,0.1717911733651144,
 0.3765847870200406,0.6404699446280347,0.8418678745797786,0.7351741112253746,
 0.010401174929899448,0.7702042646694541,0.662439502891432,0.887209684605996,
 0.3930568069310726,0.4424663601447646,0.9580196216123072,0.5245714907750048,
 0.32001905910617945,0.7701252425293327,0.8393257945759548,
 0.24859597580007764,0.22851764013636156,0.7126545689203098,
 0.9161535037161986,0.1277282318240316,0.31551455190891176,
 0.36627902206440777,0.2857091777027343,0.8314446856954621,
 0.24281986860847393,0.43417043114007614,0.24602038679055727,
 0.20487526940446377,0.7653943439276971,0.8054712189237813,
 0.42625197037591445,0.9989478931918924,0.7266112007366224,0.9224581252289517,
 0.4900842419105813,0.4200236912749814,0.300242525242128,0.41800985079795827,
 0.18987081655367566,0.7226219855742642,0.9774976979301432,0.5221866214112664,
 0.7841215434793241,0.20171157625842895,0.5842917909486157,0.2883455762903856,
 0.6260626831095829,0.6217208008992298,0.6325187469843666,0.8053146692445949,
 0.44335361372228443,0.9484223840158494,0.9432922792835002,0.5943758819771953,
 0.33803822576429465,0.4997010393194449,0.5398049109906102,
 0.21473648055133232,0.01076277297725925,0.6904826489163394,
 0.7117931452399918,0.07481406547653735,0.04485018835972754,
 0.6489581244868844,0.2721617450257513,0.23597092986589274,0.5057613084108317,
 0.5052628745124759,0.6667267664533226,0.6773138483507849,0.5200621322880582,
 0.6865330012132717,0.8635659954675091,0.9852130946329452,0.06699422867675486,
 0.16669360775671072,0.2783689416321372,0.7622760253112503,0.9802605202100477,
 0.637851187864343,0.001874367060453408,0.9606224418534384,0.3410121307348313,
 0.5771395352717124,0.7113304222977287,0.3700646757303775,0.7688150525995424,
 0.08560955549301907,0.4972731348076166,0.7841693881832942,0.897472748324567,
 0.9125598718632669,0.5339264101774375,0.34529598248865956,0.5538378655065713,
 0.869127325628578,0.6369417534009003,0.55917033130581,0.9777293699075904,
 0.9094537672007036,0.199105063137199,0.775336876555504,0.7447583680040916,
 0.049868559996444906,0.28166476574231136,0.7754249696485316,
 0.47942685711371813,0.007075984247834843,0.002018056060871465,
 0.6867361031069359,0.5637875699509352,0.698504405732806,0.06538727846379777,
 0.03048413814556683,0.34825256631879353,0.1838762313115888,
 0.7798791008099131,0.09030227174484007,0.9793148792465658,0.9447741059863871,
 0.42023403775607626,0.5721265135643221,0.5974957473693282,0.7377095510024967,
 0.3399344586887054,0.6168731819836836,0.34269619136855667,
 0.25804421864683025,0.6262791869099644,0.6603151763929886,
 0.23879609938415325,0.8731087889449911,0.09806194851996919,
 0.6522240714904406,0.29023937056974436,0.08426637870818787,
 0.9640579129933199,0.7666144877283005,0.6855464353252996,0.4544563680450857,
 0.2806270523995895,0.5696389421191602,0.10177464582877582,0.7899316183553327,
 0.5804730085875691,0.46114920290253736,0.3797450945703911,0.7850079246409597,
 0.28707516914184017,0.07082975918956003,0.10737253363469124,
 0.6793469162907525,0.11933848745011255,0.5594504107052685,0.4277717790277449,
 0.9046263185202925,0.9900685911973905,0.9261231826511549,0.5119750508797704,
 0.6813903610049967,0.5750342334275337,0.8582166938922242,
 0.026132595015013416,0.6137678312355578,0.045093518446788994,
 0.4402469136945134,0.9829935812134365,0.4812342674439338,0.9241938533820564,
 0.24053443265371066,0.5853014875578029,0.17293371728511064,
 0.5245824089559713,0.4672582010660349,0.8952068093766439,0.21686495411314277,
 0.19247076269606755,0.05433799064085332,0.9147026213586833,0.615465981492643,
 0.30518684203950996,0.5705303543627859,0.7517588578278421,0.9888728665768479,
 0.5194070409414461,0.24838357636843267,0.8587239992446075,0.6854571347358082,
 0.5094976681820974,0.9430115693695813,0.20783919663215955,0.9387172291218867,
 0.8331237970733196,0.7034193646073562,0.6427520203956671,
 0.006952282030775914,0.14048822816200435,0.3649139025062834,
 0.6613659365832291,0.6876733259419059,0.6533619923325963,0.3258018327393929,
 0.7534533643776988,0.39783937295017713,0.3707519979980429,0.9022608311519332,
 0.4129926789256785,0.7985498937179666,0.6814733556955026,0.7084127581705115,
 0.6037320388915302,0.6285158170076125,0.3305336515131492,0.0954104872131214,
 0.6254985423133586,0.07463832772335022,0.20096580116163987,0.295097348982855,
 0.7430207615534331,0.6557118659861524,0.7118440558092616,0.00897652154104922,
 0.002348734451153356,0.2889010855548434,0.11021514990360748,
 0.9922286439027697,0.7000993985507958,0.552134377037106,0.7076871496179603,
 0.5427390032821013,0.9294274177858693,0.485281381479996,0.6232818577418257,
 0.721471874370153,0.11826099902281195,0.973649394500601,0.5116586165244322,
 0.5080702116078001,0.6467750425612175,0.48169898390949095,0.1184604354105373,
 0.9588904878637775,0.8885196699077849,0.7433458078416543,0.6903753901925668,
 0.5006175178607433,0.5808502320160238,0.23843737965259115,0.6675269326688944,
 0.5993168852720955,0.43167308817065275,0.8559729892745335,0.8953267679183332,
 0.8707623621195272,0.21317977451769576,0.19968196850385156,
 0.7308483978650961,0.3744895071215095,0.9675612696395492,0.858457318702018,
 0.22384928425618056,0.7652902674211703,0.6088713400258807,0.3326879693707674,
 0.03890143657532519,0.836511946059842,0.7647627003880719,0.33689901413271517,
 0.8093476757122818,0.8411657105473476,0.6149674491103867,0.38192073234630075,
 0.13513671904831714,0.3389319407792978,0.6574253372118368,0.7265867994713324,
 0.6245705761784369,0.2987648892901591,0.09202952926646124,0.836021220112621,
 0.3477061555987726,0.2859886897744025,0.3461579809322064,0.7106469904632673,
 0.13847413543737686,0.5100062222157145,0.5269501988625143,0.5054348472971321,
 0.03233327130203856,0.11746190448168026,0.6155681950188312,
 0.7649516889774529,0.21879560781982865,0.483637889777221,0.8577963105129663,
 0.9548165138808276,0.6002446830166774,0.9170666016558026,0.20366787873780345,
 0.39917068199543193,0.1679469585814346,0.2873160772397979,0.5665805412176218,
 0.4615410390348006,0.19694863938136709,0.12169349110776317,
 0.6143334324522999,0.28527718319450834,0.38625269180777255,
 0.7709612241721364,0.7781089009974113,0.3810107431282612,0.4652437511446349,
 0.9312135142259679,0.32435264335114045,0.49435726049533546,
 0.4608212309270392,0.7162609002505477,0.4893638073523343,0.8931472203617694,
 0.5426187355806794,0.772180591926327,0.34021372845690634,0.46686665720714404,
 0.2534146102700724,0.12662064300727194,0.866087406860749,0.7630721409099658,
 0.40868500447408296,0.5280043415124798,0.8651667267042489,0.906953335242328,
 0.7255899888040375,0.7384064819844732,0.10222742839225141,0.8319441496593873,
 0.49667749076351453,0.6972945583075943,0.5821277571007382,0.2461816423411799,
 0.3269368099262313,0.40887161136835315,0.6450877495295347,
 0.06971926776389337,0.6670133143270899,0.2708548692406052,0.6461633151172422,
 0.5932148016756583,0.9287503292991699,0.7500611344471688,0.46900388084090106,
 0.8623630675967743,0.7775974674648699,0.1787549042434028,0.47488261133546883,
 0.6891187078405724,0.6609779182018124,0.869183079043384,0.06749313010447522,
 0.7623798910748868,0.27338757343624576,0.6704427350407519,
 0.07656345081396276,0.4761822329592855,0.7246320965414524,
 0.03808204452766106,0.04080439855581386,0.13105384450028257,
 0.4288811707544452,0.19174744865806093,0.8164826434422007,
 0.44228977266687775,0.2283195886300453,0.3069897765694287,
 0.18130059001387577,0.9964289951692873,0.7599393420375159,0.5231462675978296,
 0.27481045625881206,0.1245540736855999,0.9347992299552355,
 0.05533779464687816,0.40237995615749855,0.9775696862687281,0.312469004580746,
 0.2245419300495819,0.45441891450937777,0.979622343770631,0.5974948281830529,
 0.886412459843615,0.45951997136695344,0.4001198259725206,0.8911636243283639,
 0.04899559564111544,0.32090659191214665,0.40274672818701407,
 0.6608995689121078,0.23480520155752416,0.9664496976762297,0.6732904829197668,
 0.06771815801581049,0.1769615877731312,0.4492075138905046,0.7621114493096179,
 0.05043786178203402,0.31129594358671075,0.9135660033045402,
 0.5255224315902355,0.6837360260583262,0.4958301900581398,0.5343803327340383,
 0.7784818851169136,0.7040572992773303,0.19793071837996934,0.4551307880388814,
 0.02028281802702081,0.6444941568121237,0.605187061073822,0.9784125273210696,
 0.8309432769830256,0.9177169226446643,0.6127469284497639,0.4367195959391498,
 0.6916009537439427,0.35910529054175555,0.48245577649900007,
 0.12423872350617826,0.0434314105791469,0.5842714594398589,
 0.25729504730962593,0.6887511285775558,0.5872299355669396,
 0.43469555378722147,0.7919013175243212,0.2850793489981215,0.5268182150393913,
 0.43379609735679747,0.1354417742426096,0.8630052530300553,
 0.43811177813759294,0.8825787928389958,0.5907032768541849,0.9918598364972487,
 0.8023270783954891,0.33597557006028844,0.9343890879105767,
 0.014830331936113827,0.22802032034296638,0.2527778437157291,
 0.789751239426191,0.47662891911744143,0.4022700413527692,0.6169326931546982,
 0.7692059570335114,0.10555940363816263,0.034028959695612704,
 0.4163552666525703,0.808476779780655,0.8734784030572285,0.25954771508973196,
 0.6476047669282495,0.6632181229079466,0.7318553894299089,0.9355595970076376,
 0.34403565126421043,0.9237825444743697,0.525353342726052,0.4780137801708928,
 0.9856364256809601,0.9151675459540113,0.10413936816060987,
 0.24257005834048895,0.818465493788902,0.31211075495815765,0.8668790060432149,
 0.8835083272703375,0.5080045727166809,0.059480000908268216,
 0.5393344581683437,0.42117708545510424,0.2521516890647345,
 0.49205557525666743,0.22505962669500634,0.023397496128730544,
 0.5751949317551691,0.2761102871037192,0.18061753168418027,
 0.34834439279620233,0.24823902324453095,0.8548414859400147,
 0.9663625432716041,0.5468704336406072,0.1780879446763869,0.6623428880273645,
 0.08178449523939002,0.7332932462438723,0.555878266113818,0.22017985968794473,
 0.8916173019184305,0.5451785054387375,0.8911930254216084,0.9148111398423397,
 0.06862688015990304,0.881593547062126,0.7976472929830092,0.21879817153025716,
 0.5981058836370579,0.4198289885632308,0.5499141243831239,0.0903900498682797,
 0.003076193245092118,0.06972604437542573,0.33783906437703826,
 0.8721169019302599,0.2922797942556036,0.19874486224434873,0.4005891173595191,
 0.5559221716341363,0.4930262323058795,0.46097139763266315,0.3823794454050605,
 0.8180524990695666,0.3682929682773719,0.5423421838461445,0.14673846464256501,
 0.7178628536804119,0.2958060075024125,0.23483921160405496,0.4523450773070393,
 0.9445537322721559,0.27047401613205957,0.11146332323199715,
 0.16768123465444995,0.764545867300012,0.308209919786708,0.021151100731899408,
 0.2646474450476487,0.43557773689620205,0.9132349589521691,0.2684596039440179,
 0.28456333444305504,0.7282182099867239,0.3391512161432342,0.4001448244893375,
 0.6804593452649891,0.6388634830392527,0.739622723719362,0.7012396699445455,
 0.8260595336690528,0.5867862830126781,0.7035301785068852,0.9911448278650602,
 0.9304987058344418,0.15080916968748148,0.5450875326402007,
 0.05383720651057833,0.6269607325650992,0.11288340140584552,
 0.06816109925552061,0.3653079367258075,0.20377591078940271,
 0.04098553438884456,0.09007121245052085,0.8898463892773042,
 0.5631359372800464,0.75286239455257,0.35191454023584556,0.8314026374185635,
 0.5892713078426101,0.9806011169320479,0.9496138792066616,0.4540220966358339,
 0.5174693082472305,0.6158781933752705,0.6820881091981587,0.39924446279689585,
 0.3958586946711513,0.8228695714601559,0.22375384975507417,0.6492474388484865,
 0.8589781565058602,0.6248979465552195,0.6999477011878776,0.8391065460959244,
 0.20359079252950307,0.3148663787051118,0.8971016861600146,0.5401292406940396,
 0.23504194652366006,0.7266285167471666,0.15376468287012224,0.700360688891681,
 0.5368911116438833,0.5195275502813803,0.441510576495718,0.6058357859946041,
 0.2923666262140985,0.8555143082670287,0.46180590443285496,0.8135466807223853,
 0.9572573278702861,0.2846928562529871,0.3304849295531398,0.20301591973499933,
 0.718584819582197,0.6664040782669164,0.25755791338023715,0.7220707007667317,
 0.7749766525363064,0.3377390666334721,0.96146197303172,0.2833956336201343,
 0.46991511912620965,0.7141834747707737,0.11877708718077595,
 0.34637212228732883,0.14949343185863873,0.7001133597978576,
 0.20296437239477172,0.9454031692474463,0.03159301482291832,0.640026833309029,
 0.8909212761765888,0.7835519876074073,0.5361766732093014,0.8353285371595122,
 0.3633830128667941,0.08236386240011562,0.9758959931698811,
 0.34750385695613395,0.13602888340383856,0.5871700442433768,
 0.6296810333778735,0.7996245736156895,0.06836915696053425,0.8221010622603832,
 0.8137892936878351,0.9745087585875157,0.6257231708246389,0.17616224470158237,
 0.765145377585291,0.6911692438949665,0.8905390464733278,0.9424206160314912,
 0.03589884374418151,0.7486075143430304,0.36242122923670306,
 0.16532804628249065,0.9772301505854363,0.46987329078742457,
 0.9849231201241344,0.5638379190430441,0.44838836641524416,
 0.08253723918406597,0.4484906900235429,0.7283394697250366,
 0.12324589488503479,0.3649664458005849,0.9108995096804572,
 0.056939010948964364,0.19888116910044962,0.36135653767274245,
 0.026085729489421228,0.6893146614595325,0.6256489261288029,
 0.23676462165118783,0.18712580417427815,0.057417784660437454,
 0.47920017541234106,0.5402555150984083,0.48469318214021806,
 0.14499646402550248,0.4169119293193644,0.47341256795533926,0.842524872820852,
 0.07255969592761657,0.13206982319473393,0.07841935166017044,
 0.3156362002294546,0.09251494798506499,0.787334109647756,0.12931990013453687,
 0.26611786506815116,0.8843850431078328,0.0648527615664708,0.8211974469590492,
 0.10662872466384554,0.016430636083968997,0.776059660620952,
 0.8159299935317025,0.7017647493232517,0.8767569957307532,0.5835079500243756,
 0.4322666784295037,0.3492694813606113,0.9354268398964554,0.1944812944671962,
 0.6493162535909361,0.5354574605362888,0.9926518629146417,0.9632388411348815,
 0.36441711152828093,0.6037705360362745,0.7131736767028825,
 0.04704556139164939,0.60273634053631,0.5326182437427813,0.5309277290931053,
 0.9528416059346334,0.73011270579002,0.14155969647322386,0.363282568843756,
 0.12765672692661778,0.13764272557285473,0.49701441753861864,
 0.6713300340520051,0.7453251838048635,0.6193970882966198,0.5337444036224473,
 0.6278095386743281,0.30515865635547634,0.4172272985995269,0.2768821269509375,
 0.22796519547391536,0.4015205695402039,0.17211662304288255,
 0.5244642893890967,0.04392821168268768,0.6882113158994068,0.9626498526688843,
 0.7062190297878965,0.6408729916304298,0.44890316527548124,0.6042621202264857,
 0.03135656989934865,0.6583071059276911,0.5543833460419849,
 0.30839419772517607,0.646066758704287,0.6161703765826517,0.15210902506386037,
 0.20020307308251706,0.8698482345981213,0.6578852115853304,0.5547519899622771,
 0.800279368722768,0.24031565945791789,0.039190888616122965,
 0.7571457108337651,0.018189561892229866,0.7996309326162417,
 0.6851780872992833,0.17898825563981235,0.7452720717316805,0.5123306228437778,
 0.7272590154519093,0.5514263371103936,0.5758007890818714,0.20508412865303416,
 0.8306317609945708,0.555874022492015,0.06749658110404944,0.1567858186417075,
 0.570746845036704,0.6939310005824318,0.5814831145888477,0.3146333589752486,
 0.9778280951972417,0.9522808338377367,0.8114113502758808,0.942082692042528,
 0.26856923876021077,0.23404549439269273,0.5599980054556248,
 0.28453513162615895,0.06429374433641288,0.15175594894763444,
 0.25072665987627873,0.8845671557653955,0.8401334262147447,0.8163424125724649,
 0.6775629820465686,0.1955206272309632,0.3639581458731581,0.920116518615259,
 0.047649854165471184,0.41244732316773824,0.858741318893911,
 0.22653264454091604,0.05849824498219003,0.717688354823132,
 0.35163062973623516,0.8972767443488011,0.5343740923049063,0.741428847751503,
 0.21100700517735094,0.2932290886590794,0.055907732713220915,
 0.24361247023488564,0.7567063249270752,0.9653338955470924,0.8120439960979903,
 0.7580367930462997,0.3783737808722454,0.8544893942660959,0.21171007552042254,
 0.568476841293208,0.40268723460599243,0.9796356796895448,0.5767887052276495,
 0.9908311580547833,0.5429317857417395,0.06114624182666373,
 0.36343874296746925,0.43874426757308793,0.27384041490848476,
 0.9077649469941776,0.14214080658890205,0.9604890462498398,0.6341172546965673,
 0.23249023844294403,0.828236194604748,0.3190962907900453,0.5517613460355189,
 0.008232821048405059,0.3685367140586365,0.006539862474870484,
 0.9736614025423861,0.10931950251704348,0.9241511990585706,0.8104500151790944,
 0.837457191566692,0.02770319571463231,0.6710432091857534,0.7131985513506346,
 0.9071521876569364,0.8276595972843901,0.7352426281156882,0.13296566241396368,
 0.32028151831662177,0.2822440682411427,0.6604352003234052,0.7073405598440158,
 0.6971833448076185,0.6823582966819728,0.12452032257470369,0.7551464150732635,
 0.382161495258326,0.03639515630544887,0.18024251784417,0.7311432698934908,
 0.19358009731122705,0.7785642218859362,0.7237741405149232,0.6934028839105167,
 0.1227331606048192,0.7195690504743051,0.4444129486171713,0.02648895650190164,
 0.6534650886103629,0.6827969098471275,0.526531997193853,0.45093092909139876,
 0.18872414408532112,0.23103855272947804,0.4634180427732414,
 0.45592254115682074,0.3497901184669928,0.9383821760707627,0.5853421278128057,
 0.6783457807958196,0.283296413094278,0.4281044131746985,0.26268476843549904,
 0.2526350386727556,0.8498611748760662,0.23904725306661778,0.8531225368421282,
 0.7393962288301945,0.5643845859081249,0.6890109040170155,0.6276332880519646,
 0.1181647014687005,0.09722396083258067,0.8243222985779627,0.4363344270817813,
 0.9448154769223116,0.6951185843977673,0.7042387185693281,
 0.020734896618414878,0.068033017033166,0.29622158040825985,
 0.08444319907219988,0.3552806384233812,0.27306104432517175,
 0.1415164420515196,0.18928820407675262,0.052966275498480875,
 0.11205141466416912,0.5787710668791674,0.1025328527702476,0.9497482610855196,
 0.9364784596479305,0.04188765569515518,0.5798954499231409,0.6430599290937562,
 0.9441770600977628,0.2896251232271746,0.6201145046270611,0.3226538462007895,
 0.321996262382255,0.32270829742996887,0.8470372680322312,0.12419149454164913,
 0.0352244729125637,0.18326664616800914,0.3482291040341623,0.889318708451031,
 0.834791804453826,0.4441388661007645,0.9305632409361633,0.7321398035196706,
 0.7594466510147643,0.4980692478927802,0.6594694019725518,0.4872492660505293,
 0.5879882542194563,0.71680148953341,0.24795941797317267,0.3468538187859226,
 0.7041228388319429,0.7701347100706301,0.6223047671639527,0.7962524972534506,
 0.8420995872316033,0.04179096952304595,0.8946927026734197,0.6987314102333465,
 0.7785077036922083,0.9005766109753383,0.8156849597244505,0.9586895421160703,
 0.9461911563616798,0.6473711309432568,0.7305324977647201,0.21627203154615704,
 0.5477406646065774,0.7574043692181447,0.8530448560860046,0.07438977009911552,
 0.1744779542646422,0.3778799096308314,0.9466628487284211,0.14820530598233228,
 0.9930910946162178,0.7514080036647758,0.7872235641176782,
 0.046016674950781544,0.8808305422853037,0.05839606151616583,
 0.1957030780479041,0.903414809238578,0.242316823256836,0.9616089822860067,
 0.9231950300349473,0.7580897091442266,0.5486568687206836,0.13261355768363892,
 0.6655483178811024,0.7002302873789812,0.15932791446213268,0.5043714865862436,
 0.4048707829923015,0.36861204381351176,0.6731498551026419,0.9625752400761945,
 0.03574169236525171,0.08672285547900949,0.011441980191916334,
 0.15940055700139677,0.012334823130750627,0.6028345575880166,
 0.1837844645735569,0.6338109029301902,0.7051173458441875,0.17648245727134304,
 0.00594694437216059,0.7912578251602348,0.8036319922076661,0.6199540773645217,
 0.671097794705115,0.04666283281073147,0.3938578527493344,0.9078755050501209,
 0.35108274006645523,0.8216029420676165,0.7880464399038718,
 0.40816944130369837,0.7661223634589983,0.11935180896932307,
 0.25232894548401075,0.2032397587867465,0.25634378170122607,
 0.7115374550257993,0.9032095700438895,0.4491812765427784,0.6332986421807156,
 0.708454607908437,0.9537142419817973,0.6072472560649902,0.5910148785293176,
 0.2779039162946195,0.5909895615137757,0.7844146605594734,0.3496030726430299,
 0.5761586002823597,0.26550818397112863,0.8534418097733357,0.9504563829781808,
 0.5248714475678669,0.822103301473099,0.471856132862555,0.00508674456516367,
 0.4912686450709769,0.40445411848389146,0.5101837003247077,0.7089578171948717,
 0.2833672665719199,0.7183572360863882,0.2975190174698117,0.928406499744114,
 0.13991462798885534,0.19591050513657016,0.21140110332698292,
 0.3704779567994638,0.6702563910296104,0.6935919390540216,0.5008201173586292,
 0.7528652636943975,0.7773567726037542,0.7706974953143366,0.35948089080683554,
 0.39046675367292705,0.8405116303043836,0.8369072061646755,0.7147264171271289,
 0.8756720176465703,0.7679972378296995,0.8891136675179792,0.7878466647168483,
 0.8688929915178158,0.5837846703233325,0.9372798393200099,0.8353136574139641,
 0.5413335796068246,0.9489326453870272,0.8190630824954771,
 0.030589821037633746,0.1575949933940678,0.6879326003777505,
 0.5486936211176918,0.46691153748027125,0.1344000234407451,0.2809385172309379,
 0.0925499554440048,0.7998517728465687,0.6249350623318284,0.7177262655071207,
 0.09791167844672,0.4828497136753419,0.5407678198140549,0.48709925451169167,
 0.1685603797613805,0.9633276012052535,0.35287865344766944,0.6213272999366504,
 0.7538601421833303,0.8388263402963471,0.46049761391854005,0.3079032115849478,
 0.499546560399735,0.9944903711928732,0.8500120500073081,0.31963941711355415,
 0.9591060858630952,0.4278878502818745,0.7679488504686619,0.7981920977349926,
 0.5260028253269383,0.7813436324085254,0.814401675227393,0.3806597938893561,
 0.0522233983553807,0.5300116278191945,0.6295381091777853,0.7879021378014388,
 0.7459432318767671,0.04899536409664895,0.5650681814695577,0.9760204093672101,
 0.8162180140180404,0.8536389273605771,0.7572066885969136,0.6256212301531264,
 0.03588190991821549,0.0300427714544953,0.3543461472051258,
 0.02455364514524927,0.44897282011615003,0.26828788511950885,
 0.3655871883406977,0.6230872449771473,0.6008312681956368,0.5564533003909071,
 0.7243996112330849,0.7746025914058606,0.23665010988611257,0.2840024879692683,
 0.49483717986801656,0.48186517943458573,0.9922514583193853,
 0.3818779120111023,0.5615091478617485,0.17073179954547502,0.9056727810440928,
 0.10109596778163032,0.10912518063376919,0.6040675833715875,
 0.2316368065113953,0.32614649229721937,0.1462982658568165,0.7378334666824606,
 0.9468398761178353,0.28486118730074106,0.981905376260855,
 0.044022949624832663,0.4578641910199426,0.5237416835189006,
 0.15397259317308554,0.12419227876319683,0.3804183743528756,
 0.3898639796196224,0.714614347242826,0.4024089443494956,0.19020846422420035,
 0.13676935193005113,0.3858125792242677,0.17680959127929186,0.935508896776669,
 0.47238977721463193,0.48028293113144116,0.3721753990964952,
 0.9379431432657286,0.9270196579227314,0.3668849502247077,0.5444738492966952,
 0.30219573591247584,0.9399115713629265,0.562994584290936,
 0.015632996082864947,0.8666802268439544,0.689193396149519,
 0.028202524904934134,0.3717869305366027,0.5481402849729411,
 0.4641435800254011,0.09856139927113028,0.18155118510115342,
 0.5052635422617958,0.3881529238045024,0.4644710733261673,0.8199075364293763,
 0.485341462883093,0.7032377392248084,0.45311968149910364,0.9040541399314356,
 0.4906338212164866,0.19245656132994882,0.2548649859018035,
 0.29173431831155017,0.813381026338665,0.7911503187690704,0.07454327312664932,
 0.752512374482589,0.6974476558905456,0.1744107889801162,0.23961758204382944,
 0.1285552511354443,0.22329055679789478,0.05550327083435824,
 0.8356446710320071,0.5965646918472418,0.5418590013560233,
 0.050339255124955073,0.03999625899095227,0.8721560855770039,
 0.39315162418391414,0.9752443396094207,0.37131720922946154,
 0.9405592069778573,0.3280559405098369,0.9891333274907015,0.6850217784069665,
 0.8791733008329702,0.9520313880874686,0.032826307426115364,
 0.8529136244895432,0.883777190638195,0.61026641557878,0.15485238301954574,
 0.3404714487990983,0.30269248755214995,0.37084438792146757,
 0.16771700695278258,0.9976637162904529,0.21913153411838182,
 0.1215531775690899,0.5287617721096924,0.4035161727653117,0.6580643451001366,
 0.168830741104264,0.6800616447368033,0.528715777732882,0.33098446783423585,
 0.5219889688384737,0.00490935425769734,0.016512628387971517,
 0.34838681198743093,0.6127772318619186,0.20782294589876804,
 0.5238229825065196,0.7502420349215797,0.19197012036156824,
 0.33424670865158035,0.8457598815017053,0.5397335635133769,0.8729758804026679,
 0.32073088562800134,0.7985541709517668,0.06439648064526604,
 0.6725485685576742,0.5192791046381953,0.17198186603210175,0.6867377458386695,
 0.9645574039437602,0.5763617223664291,0.45913394915423644,0.8963475472889587,
 0.23315291486366052,0.5247569505766178,0.4439433519947391,0.4752766887827543,
 0.8091400745233184,0.8394476551603992,0.527524574224279,0.9612506053373818,
 0.39875641840021747,0.6487428628582035,0.06933155479959097,
 0.3290578701063268,0.8698801020849249,0.1439771920897227,0.4859378834218778,
 0.11432234894867754,0.2956436609515424,0.067897562136779,0.6426253988588043,
 0.8805344929587425,0.03292314503949123,0.1661326067212191,0.5448239910510376,
 0.23436164370160495,0.8546427594810767,0.16367124608841777,
 0.059871077895283675,0.4795029152689759,0.9493969941746438,0.541653834380172,
 0.2495160675614968,0.3650996863345507,0.9889326882770184,0.4754461264810821,
 0.2523809663265776,0.5971916740482014,0.14227603868355987,0.6647102650451248,
 0.48023840382757554,0.21380088659579877,0.7687601618458026,
 0.7500196713890964,0.33972503484997185,0.2814454982295723,0.7514148081605653,
 0.07593787179523037,0.16699534760672052,0.43871684417431067,
 0.40933173675910783,0.7242344377916273,0.9019910848585394,0.6906450049411093,
 0.887345786212195,0.6206033200399638,0.9039717345748065,0.5708363224404459,
 0.6962026387561038,0.32727324321552154,0.5245013361234365,0.4149085278865575,
 0.47821109912026905,0.10285492711645972,0.32487365362005083,
 0.29224288201913873,0.17835633936191708,0.15324299957842225,
 0.17246127895479813,0.15421698891967295,0.41896867940925786,
 0.45897706237344826,0.16169131661233105,0.7634721469473185,
 0.2740754767421444,0.1287741991066943,0.6769020345435428,0.567226124150681,
 0.06445908527852684,0.96462489245605,0.22057280297816484,0.8948910220563281,
 0.09869824963877105,0.8606140096146283,0.5884089423980059,
 0.022051005216809916,0.8783837160479688,0.442277216060151,
 0.14067720159310215,0.09760046162934188,0.9565722260279607,0.337978745828608,
 0.2892493176999634,0.021344501221465473,0.03973815407999082,
 0.033788285411702756,0.535628343297168,0.5171805181030613,0.8352984594112672,
 0.2958456925510752,0.34325272120145534,0.4297804792797826,
 0.33086625631492217,0.5533583638912901,0.6696619759430902,
 0.043061844076569655,0.519324743668905,0.09554945629556044,
 0.19773527916707245,0.5596979669472028,0.6635283447498198,0.2218176205998995,
 0.2698129066267122,0.5491802285817811,0.29673076638550844,
 0.18039993536730303,0.4138159856218343,0.9330308304669994,0.6035399688837284,
 0.876950676388749,0.3617895843176736,0.11585334746492881,0.6264698113066458,
 0.3739079381162247,0.2434292121542888,0.36837115210955573,0.9586763704122254,
 0.9241918773677615,0.671001240270593,0.7848397713905767,0.36206123197703066,
 0.8779267685588431,0.9915737408410633,0.6748991580576689,0.4588271518521898,
 0.9124023567083809,0.7552279771185795,0.46236887035722385,
 0.23931900177038945,0.6288605988812362,0.55608538026447,0.531726222645021,
 0.5109260123770385,0.12656014354127587,0.27418994421190024,
 0.7982639076982099,0.08357204465228618,0.9886695805025217,0.726111298818499,
 0.7874938067848107,0.08348479635934458,0.5261127251842586,
 0.22291467153892952,0.34934098322529916,0.3005121742975858,
 0.9791758635874612,0.7529731943606102,0.49132366235586966,0.5077291480258178,
 0.4310878823008022,0.907828810659432,0.027073005686073404,0.7043665780823212,
 0.9402897091880557,0.041981996269723876,0.39374285480261695,
 0.9520263205527186,0.4688771892560859,0.3582645931350026,0.34698017603123776,
 0.39069237515786603,0.23837109770688547,0.3726489307277966,
 0.01388209400417173,0.9568317820646348,0.4914772879029856,0.8302145720447153,
 0.4664884707646111,0.27488856342865087,0.5248900408116949,
 0.05768648635467677,0.14713855052715774,0.506912278705224,0.5875391800631287,
 0.08515460388219975,0.27626420793007367,0.8805595415584705,
 0.005338068434405985,0.047728706691452794,0.9192748820835583,
 0.6469601959332412,0.568703654592849,0.9374768366019435,0.7766097135897227,
 0.21606614222115916,0.4635878736061114,0.18332149541614784,
 0.7988670591058007,0.6564508821717956,0.13338099202335996,
 0.057118085495451165,0.1857937122521005,0.451516315821785,0.9935958624037307,
 0.06616129234164014,0.13683530629045648,0.2003853622678352,
 0.9029187197145103,0.5637582203638356,0.9809321497553791,0.6211338325868848,
 0.8446188884604873,0.47591097920506154,0.5684726286744861,0.4033261295892744,
 0.07491632812328808,0.36017865862712184,0.5140139938848859,
 0.9421325671953913,0.20559418839339705,0.20801418854568288,
 0.12048313573894232,0.4961413420780685,0.7099304366683837,0.5752132253374493,
 0.8736113826950938,0.7925697368887651,0.19146682277818095,0.5744185217491501,
 0.10023853375545488,0.5517786042446535,0.1313474072651184,0.2642848282347109,
 0.7316903303496747,0.9849296284286353,0.11840590754020552,
 0.31369526949923887,0.6558821909026379,0.4132408612094408,0.9350084846490672,
 0.46575531671532455,0.2552417350973051,0.7264774112706521,0.2940632493370652,
 0.1582189608331508,0.9501926346915548,0.9606992662324604,0.6888585439069739,
 0.04358644820889068,0.635405919549345,0.33919620519221017,0.6551906503734112,
 0.5866657029411213,0.8134760583902717,0.1897596288220249,0.9359034048200131,
 0.895010702310664,0.5060697860882657,0.46045007610736877,0.3507114348570394,
 0.6584044393119981,0.47710006774414304,0.6820695537653519,0.5450941508995351,
 0.8172730520475121,0.06493826570187644,0.7059825956312873,0.9267037418953429,
 0.26506558277843384,0.04682716532094855,0.9628004875929037,
 0.3839731918378483,0.28559042124866285,0.09817757096326496,
 0.3257689223499791,0.9176638847396719,0.9869040931030197,0.9268619868026978,
 0.757922279109545,0.2755462721909502,0.2875932499834879,0.4085032636295701,
 0.49502204505685654,0.9514713321763257,0.9410824209151859,0.9696996762888694,
 0.5440548312321479,0.40469234137532284,0.627690441207931,0.3583525182139585,
 0.44534533310402236,0.8737122564241875,0.7091319726147616,0.1805389582751884,
 0.986569418095618,0.4832443719608941,0.15794677124186152,
 0.0024062361528223164,0.3059535117699991,0.9704944746916305,
 0.09747206578768741,0.6333222765277713,0.6193752914437721,
 0.09356607698864772,0.31290722635604307,0.4175037891675404,
 0.003948126935308327,0.8107738725821604,0.015937985198736015,
 0.07879836399275353,0.04940279466745956,0.9030329660782658,
 0.3239480690821438,0.22786520562382828,0.20818307740242692,
 0.7148444034251975,0.20951501809705342,0.8404154787648068,0.6416173954467567,
 0.5574123453362562,0.9610190590054937,0.4990379196641803,0.16891875857456073,
 0.46146436735706886,0.18139203319930053,0.5397861733339523,
 0.24772950410121353,0.7889388516507957,0.5735468454101843,0.305682999735724,
 0.12844198918432403,0.3778744591019797,0.40726872354858246,
 0.9305504490943379,0.5451538276228404,0.2623833547068699,0.296069367514897,
 0.6039439067055729,0.2671307596338641,0.5036480423116119,0.35024726167979825,
 0.28461724772647445,0.45949155809485165,0.810498312639942,0.7936783007907003,
 0.8154958021689114,0.8432137124157799,0.5640010735925469,0.8517296414343221,
 0.19277838723409912,0.2051279746887963,0.6977190880487996,0.8615849537371916,
 0.9903413064772442,0.44110380893721657,0.13415035258956765,
 0.032312814754999364,0.8644331403455697,0.07454851604031043,
 0.6376188009880844,0.11165215229567327,0.9732789448522603,0.190465819895839,
 0.16436408373427314,0.17425619788081892,0.7932381483495008,
 0.5138517413785086,0.4131474924034897,0.24260242446200075,0.2848194689407515,
 0.8216273451040086,0.9464216364095644,0.9707852193172327,0.9404519531361792,
 0.9871541889262994,0.9434350622941466,0.12143331607744601,0.7930057341348096,
 0.7841335384417568,0.11183528658782327,0.6295521286011856,0.8530049564339909,
 0.7728189050560738,0.9926384713115116,0.7841353175184195,0.19591424533229052,
 0.09406554451805571,0.29213092536131813,0.3489183354808564,0.264966654605054,
 0.5165718723612928,0.9064043894615698,0.9513983148042755,0.06487245275338294,
 0.42713493308543815,0.11452617985590396,0.014239315051230328,
 0.17757264963225627,0.8680306820878995,0.2328995892414112,0.5937905379757581,
 0.25437513937124623,0.9597461961821537,0.5412816369371509,0.9743608528752963,
 0.8310370091586937,0.8285262765654862,0.002545862882116001,
 0.2978653638763511,0.32767225484283724,0.9409991144905921,
 0.34292086457440396,0.45717005655641285,0.2862556436334882,
 0.8292872281568285,0.7616466567928971,0.022055548045602613,
 0.33168274763268557,0.5474881504294044,0.7632287869492532,0.8021072582332747,
 0.46454141454360587,0.31725769458702724,0.3010161605033884,
 0.5795670051430292,0.8872178893545738,0.09469924185137613,0.8103104787465732,
 0.9909727677919226,0.1391689713303773,0.2220661612853908,0.6852196277877995,
 0.39228540018490565,0.9551067349679808,0.6759975971143174,
 0.45299709527185483,0.18761537202958511,0.8585347374672709,
 0.7490860501200493,0.42122605908598654,0.036945506050785726,
 0.39989118554250946,0.9205154162074236,0.18048248574715786,
 0.4056374332045237,0.7621688740062732,0.34601096312163204,
 0.11794549327905612,0.21275957021375225,0.26293062930990807,
 0.7040629192798772,0.4612907736330969,0.9693942790190934,0.3762100765526175,
 0.6662423664717115,0.40335870037192956,0.03747893274033942,
 0.8596695271248651,0.8491354156201422,0.8502603994450655,0.8597529808458547,
 0.7554477687489777,0.07905382020205698,0.9419148119616997,0.3046847155843875,
 0.3600615981726437,0.792830313746478,0.7473467951140647,0.4378416450141054,
 0.9254827667515892,0.6515555861849784,0.40856380124983027,
 0.058601355388670395,0.7840251676307806,0.06561829204742409,
 0.6319286282782346,0.6498141595767138,0.81430747154118,0.7849207827093392,
 0.6492027678624611,0.40727331777019093,0.30268192591411347,
 0.9952601490564387,0.056131826573734944,0.19368949350821052,
 0.07224238930493943,0.36637396768853225,0.8779715661748425,
 0.37684038246222706,0.7824996702265162,0.4874234660615496,0.8907431104799097,
 0.20322490182126585,0.6255060722019363,0.757050412780888,0.8961589859187455,
 0.17410839685191015,0.16912509516422858,0.4153679637009239,
 0.17518974168819446,0.2113757977260342,0.8025004600041812,0.4865016876213357,
 0.946620680552864,0.10953246852353171,0.8308706563285897,0.901638060891726,
 0.3381148371049928,0.6809562581174388,0.11528123209777064,
 0.19748184594524165,0.5309648374700751,0.17790890736658527,
 0.08821615873711619,0.323918733244706,0.22092011703708558,0.190611269327298,
 0.3672396810455657,0.49493448370398885,0.3181457516699947,
 0.13265414351112603,0.04099093680216226,0.6212318133632772,
 0.6478433866719718,0.5332025627563228,0.831079974854443,0.17916228842957882,
 0.13811742454129028,0.02059950539053579,0.7150329552084541,
 0.17267147675344852,0.8563667671304479,0.3873351895044612,0.4913354359384544,
 0.2720816512729426,0.4269289967328842,0.8606418346225827,0.6662851777841328,
 0.5724233269751723,0.6022612651873537,0.6157203617799977,0.23923728417287493,
 0.18224038024682265,0.7404393756201085,0.1262796004215485,0.5343815327499751,
 0.6754060929472488,0.14372931117277654,0.07640601816197345,
 0.2939621490863452,0.9599500848492153,0.03456359305807233,
 0.33387112897225735,0.07730522091626535,0.03799243501010985,
 0.4700239792807004,0.3922052952066676,0.28650510036602994,0.793938725007624,
 0.25816873050217637,0.23724441216735626,0.31597250733375715,
 0.01110377129527107,0.0807030291391706,0.25952168757324423,0.585003958163906,
 0.18756025474745486,0.0480362021284817,0.8819228876698672,0.7890166357948634,
 0.24079029907944471,0.314040292627328,0.4271440303660057,0.6229449746390543,
 0.3991597754115116,0.3101056118569936,0.800282151481783,0.4822008104459595,
 0.9014579757637413,0.31202529098930387,0.7943011537220276,0.7384764816945749,
 0.5060880991233552,0.3591759114069517,0.7280386835694987,0.07513369515276302,
 0.540779518800266,0.9711044915172509,0.8137429563852718,0.710709879873581,
 0.30813114989293844,0.4211785789935526,0.8603029033582329,0.5723581217898916,
 0.641449575601252,0.8416851654449778,0.09331009425815395,0.47754927518781576,
 0.7278566383798362,0.31289599457362127,0.5483049076539508,
 0.009240415304967997,0.7695241638880674,0.5173893951743215,
 0.7487510129775153,0.8856601810127689,0.3940745341646007,0.7412026023786451,
 0.4776763346652353,0.6455115144845754,0.5348528803999039,0.6820177335541803,
 0.58375473017293,0.5532529479505417,0.5467615558215861,0.35764477310868614,
 0.15129942564169052,0.654715580189294,0.5624860678917715,0.5221939531623282,
 0.8691650889742937,0.523338158097483,0.6715303218885053,0.9294554253860378,
 0.15469832273262607,0.3814351749092091,0.68986746140868,0.40180293240736864,
 0.04158210539782825,0.47020454965049285,0.5028810475977071,
 0.07260129855633557,0.9663856869978918,0.8780274238208625,0.6618199975204329,
 0.16098218527971775,0.6763923247419532,0.9913628588377039,0.4976400106031319,
 0.22759246098504105,0.5141841685806141,0.7770366167187497,0.6295123486017848,
 0.8309015386226203,0.8592227956862223,0.5838894573766156,0.08118803295801047,
 0.5740003784319618,0.8576372663277994,0.27546646526493956,0.5798195946206217,
 0.03674404442275769,0.31008790394565267,0.13437596152926634,
 0.0369729043860918,0.012279621160947496,0.4861590400844782,
 0.5382695135302854,0.9743177584193473,0.9402574531545941,0.39111889433270786,
 0.34899244098476023,0.261579013663945,0.7556857304313116,0.8181316779211106,
 0.5038203971290793,0.5931005723736315,0.14238596109824164,0.729212595766595,
 0.9640802153307099,0.2874662483488607,0.5504615621425559,
 0.013255298920538228,0.7611614847686574,0.835794309877953,0.5788363341532163,
 0.21133469377840397,0.208069825365081,0.6515747206415565,0.4223342556577643,
 0.2978627309797981,0.12913698049093592,0.8701923322177846,0.9550180976241202,
 0.10196898734480075,0.86368772607832,0.21974614613936305,0.8294134627452111,
 0.9174577436434849,0.03945991867622478,0.006463073244057149,
 0.03219281203714075,0.28977075836559807,0.4322478301256035,
 0.5322997241831136,0.28660303198339265,0.4090945450971938,
 0.26355772856748116,0.24499171649566365,0.4963229090557747,
 0.08849022423342268,0.8417980077320084,0.1443800241836003,0.5919997898311039,
 0.9583969336792181,0.2445118841493511,0.8774052386466479,0.03469275361545798,
 0.3328451573992157,0.5696288784063372,0.8483689683257564,0.4169762075773721,
 0.0996652165677906,0.477816789567304,0.027046895892577028,0.6680185388613661,
 0.9118066454961431,0.6075254353312334,0.9219793315969823,0.4194229396348663,
 0.7676567498902228,0.18337514063721394,0.8586045946540314,
 0.24141894699932287,0.5509929589753562,0.6824570300032577,
 0.11466743671344548,0.4037050705536117,0.304016049449518,0.14618597979701708,
 0.1444160733837787,0.07703997261707107,0.33640288723796763,
 0.26971955447237317,0.7617460378368659,0.9695017651297823,
 0.14690734943149852,0.2682502456398089,0.07178426983492392,
 0.9516592134365736,0.71472128928052,0.26907727813229665,0.18113614546735146,
 0.7227482135966796,0.051005641599037754,0.07475441369152414,
 0.7967621163906896,0.760099130825471,0.8347865406585617,0.019625161193874696,
 0.3285525791075059,0.4582052986243844,0.3932514300538237,0.8202644124419868,
 0.8809248970418798,0.4597782608130222,0.1772845128276802,0.08246740724318158,
 0.013511485038808835,0.08499725242437683,0.5329831607589388,
 0.8439594329816069,0.7021459543213916,0.931057463003004,0.8683834398123276,
 0.3628596263052094,0.38512092748954485,0.7470953222149161,0.8448310053220871,
 0.7922213946131086,0.9040186373975745,0.7244678781666851,0.078321571132966,
 0.5543432901464711,0.8109775302762662,0.4294677187239757,0.6426038197764943,
 0.420810077667639,0.7842258185933281,0.6967339900868503,0.9043456884172434,
 0.1450686153892391,0.5933701256693009,0.7804785088977617,0.413677070469078,
 0.7452990301053797,0.40119553841898836,0.368914772921332,0.05884453109293908,
 0.9540440895043016,0.28818228627733566,0.890981618020789,0.6877833417374937,
 0.34533353642924447,0.6721479929974692,0.4937350521826458,0.4817826686910819,
 0.7357324665789478,0.8199409484003513,0.8823335571902696,0.6317905334619184,
 0.6950818031133622,0.64552448480409,0.4726368898606138,0.07260844747490314,
 0.7364809779748491,0.37434440295271765,0.16525075259663535,
 0.10315310331233829,0.08796195224404257,0.9538818962976582,
 0.2599679215614208,0.3997626284533955,0.04599438935076783,0.73713443396197,
 0.20098160410974486,0.34909861018026356,0.35888423884226617,
 0.06100509858913272,0.27413966008680224,0.8136283633826532,
 0.7103059532770994,0.42433705194945925,0.5267596640051456,0.1655169762158324,
 0.7096657799574085,0.568768217843167,0.23949213505075717,0.09564430431781368,
 0.6245833946367129,0.728619344132107,0.9946422845321229,0.6844117548799686,
 0.2540716813063538,0.6510874078806664,0.49253407115448333,
 0.07425308764135674,0.7622984266496573,0.5535951945732586,0.1978603078797283,
 0.1359594277699403,0.8174605214167556,0.07163521199526346,0.7691336463704257,
 0.6318935949566439,0.10191669597197084,0.12382612050633723,0.564643777364938,
 0.8438513000950164,0.004181820638652911,0.8148205747615332,
 0.10405632329073611,0.336091933169648,0.2122793053259927,0.4768267238007615,
 0.9297613919450421,0.67337377048004,0.045718590856697494,0.367612909796184,
 0.8631863101305892,0.7590064776685496,0.3999705948625766,0.43882397960757824,
 0.29023508281148735,0.5009476418656552,0.006632441691275082,
 0.8013669446187789,0.06500896934253952,0.6828806242356442,0.4295955266960141,
 0.04275864205654434,0.07742749818124006,0.8132778474162622,
 0.5156456632608354,0.33558768056865984,0.23551949430119146,
 0.33136460567291137,0.6516376004158394,0.16228358847392987,
 0.26796556936433524,0.5503318939208222,0.9875804219780533,0.5203799989157482,
 0.15047710629313293,0.7714536322538283,0.7217070890263739,
 0.27616768064128283,0.5470140970936733,0.5352566226318294,
 0.04853895230265082,0.6863713551136066,0.19165288547428683,
 0.18730560688607778,0.12268785943956073,0.33857513169379294,
 0.5433709478540368,0.7537815117351557,0.5802125522042967,0.7451430598904697,
 0.41787401125056467,0.8349349160083398,0.01553468898236976,
 0.8667614517291335,0.28249318151293945,0.2258849316952839,0.3181827691159975,
 0.7155900390611502,0.5767093280516247,0.6020427874421745,0.4835550914239225,
 0.8729729641372402,0.7627462664838915,0.23394363748674596,
 0.11918432893353059,0.8798376039341997,0.4627241394817645,0.3383554173170975,
 0.24010291957663188,0.86507470420667,0.510105044471192,0.9677227191042149,
 0.0117333819143981,0.9524420677857336,0.09640385800230522,0.7691130280511667,
 0.2918640163337939,0.2850201192515833,0.8945556128151829,0.3475321304996072,
 0.9875501385828466,0.12952876021553672,0.5005331406108614,0.579509090352654,
 0.12525386439336728,0.15662369546664057,0.7080343534746195,
 0.05363517918695804,0.22754939822607245,0.9820697385909264,
 0.9284782632843062,0.8976469175610939,0.3713951095510959,0.5514013924970639,
 0.18775725445747082,0.16988685957026683,0.024169999548748988,
 0.9748665333814074,0.6304965709755811,0.6642326603868417,0.900397354646337,
 0.637166611701441,0.6311317066123513,0.694043080835052,0.9323186975414322,
 0.49155469775622207,0.3148577511538211,0.7984664616089515,
 0.40453689278628946,0.1855461449171566,0.6007082968116156,0.9299889260868767,
 0.10218386672019608,0.2655446848608195,0.6062740231739995,0.6942461165996813,
 0.2899130155458174,0.2758893561196487,0.48893534866536525,
 0.06878632407914353,0.17408396490230538,0.25790445110111015,
 0.7313186721198407,0.8342454255491324,0.13085585436467817,0.7435216919246124,
 0.702220029928066,0.3202308857809387,0.49284193317332103,0.4737951560762649,
 0.7454193176101593,0.7889260320301175,0.17827333309595517,0.4136992690673754,
 0.42280668826989565,0.22177006380714404,0.08203316458624788,
 0.5002253583339106,0.04314679736850069,0.7369671284763679,0.6442280110087986,
 0.5645159749783553,0.6990004150470212,0.41405599536220195,0.7627088627200048,
 0.9651204407261939,0.12676442624273765,0.09931478510404235,
 0.0819376443347023,0.8053973111680902,0.8390050933209422,0.4392641671711838,
 0.5855232733577633,0.7458405585556211,0.17758858416890244,0.9952127721902884,
 0.8229247034301809,0.5010769315439789,0.9324934269392203,0.5724655430790322,
 0.3616518533058126,0.28805265323896767,0.04510072757196759,0.772064642563456,
 0.656964761912858,0.20749595598598236,0.39259687530394527,0.402424220576303,
 0.3496635156537903,0.5712114417567308,0.7278319091366368,0.1452052190574631,
 0.14929868080722264,0.9455434553259333,0.971026184190506,0.8852037740328667,
 0.058584842045832586,0.8890690687221622,0.5005526865857923,
 0.11470460093894153,0.019457584982146447,0.6607281345955672,
 0.7527556641629178,0.4795395703916864,0.31887804565723166,0.660612603774461,
 0.7683389918983028,0.8626073942856193,0.5522325234214576,0.9509596118574383,
 0.8376708520379724,0.45145838805230765,0.055919373421831065,
 0.138378601037903,0.10896046810752846,0.6157112392316645,0.9412814546381352,
 0.9466466414642339,0.8778217068456845,0.20546879053135003,
 0.43734992286977437,0.19392013682458742,0.5821804410873637,
 0.8457887869084827,0.4035690265511933,0.8528621823943427,0.2988353912959767,
 0.4583481117629028,0.545819800566332,0.47191677235254303,0.4751291161134762,
 0.4694362474785926,0.7465660875645158,0.14016961959647922,
 0.32154699372952666,0.2433714836209777,0.5163453217609529,
 0.060925472347641785,0.7014562568871207,0.83031367751715,0.554133725235035,
 0.22302333655188666,0.9022387642847387,0.3671872543089978,0.3975218132410431,
 0.6614840365209482,0.7492202130935699,0.04389559375015539,
 0.34040634887079735,0.8489974088750549,0.5028689141505409,0.7832237982512001,
 0.45134677953216673,0.2921403193366273,0.4376171939914565,0.9871188566784421,
 0.941181137002953,0.3873380036504628,0.780080040123807,0.33889990662721536,
 0.6092858029560302,0.5174452012686019,0.22657133919040062,
 0.06388083532541855,0.1804524442358354,0.060606652556786655,
 0.16019829462751434,0.6806549223875357,0.8881481223229013,0.4249966347307853,
 0.8514281506937453,0.9652232718428099,0.5802493065017874,0.5642032707638902,
 0.9877674840018112,0.7182143699212867,0.45148397012130836,0.8144587524451478,
 0.004130307707401304,0.6592995057978642,0.34828524248393156,
 0.2525753979261113,0.03830387497776422,0.9171026021472178,0.667952667004787,
 0.12483794966829787,0.7426954966013586,0.8520160551469688,0.0673768737117495,
 0.4877999623034729,0.11717840572063487,0.3838626147625085,0.3607221097959571,
 0.06514566749425255,0.3242033234051236,0.5784700374851472,0.5547214994832916,
 0.814793265166012,0.9435007246737099,0.1569863619355163,0.9311978159437513,
 0.28437429913472834,0.18709087033923777,0.76724291590062,0.18708039474913984,
 0.4669053750582298,0.5624163559759032,0.5698085777093175,0.30986375894782603,
 0.928764776412565,0.6588076052076994,0.3028532945219764,0.7197819203320336,
 0.8227170610951491,0.10767592423578898,0.23500550648984342,
 0.10522959151420541,0.7023189050390088,0.702990761214803,0.7952480783476651,
 0.4542203274259845,0.3215485426864636,0.22574273886344487,0.4566639647592109,
 0.006676853911475544,0.3688053007852302,0.10764340902959435,
 0.24382421037332314,0.9034788664462678,0.4274231083794813,0.9279525533831479,
 0.3247215782843874,0.31532039984264437,0.03462025558208348,
 0.2914235985608151,0.3869005437962948,0.7392164832633911,0.9375089580656887,
 0.8440944928461502,0.43214682301351814,0.16477004484654456,
 0.13353135301239138,0.8978226390311177,0.9593421461366027,
 0.23974559063121337,0.870660466817751,0.5917609586821769,0.10545457765719746,
 0.6140742870639154,0.4488368112480289,0.9916903545995776,0.999630452683598,
 0.1911111092785559,0.06385013016858954,0.07227322322768892,
 0.8651339967742141,0.5811554341190986,0.13071127274185657,0.8299196035278951,
 0.1780599548745555,0.2509798799844378,0.04121402463441104,0.9385368961947635,
 0.9019575421462752,0.8731758698920326,0.41694671210025225,0.5268235770814962,
 0.31121575460860806,0.2200928795310242,0.36462969233170606,
 0.5576798982312242,0.6260598335026646,0.41024147725603555,
 0.09187629069549419,0.2207311213983908,0.8823093721254712,
 0.34216573694975205,0.3440709503483246,0.551466785965352,0.22656100612241736,
 0.4086366225081003,0.015457942899687604,0.2884676269708434,
 0.7891313476391304,0.05321285591403635,0.038617423215157265,
 0.8386302702362182,0.5666016317824935,0.44493010725853976,0.905565818292436,
 0.05185195882043292,0.5171472539213681,0.8681671067141093,0.9461878483473485,
 0.29456199104246084,0.4290192964085864,0.12676946486959761,
 0.17844928727335085,0.3901947980022694,0.666466673322585,0.5936983673431033,
 0.47733826809789126,0.36465850381014386,0.8720215685596844,
 0.32317334337304215,0.2454087058786203,0.6712212186490851,0.3436050158120193,
 0.12184302402402647,0.9266116152913598,0.741560132085807,0.13445836899557728,
 0.7909392484192451,0.6606122411705715,0.4990156535398099,0.4512272852170791,
 0.5294857446283954,0.6193136187784587,0.6427636637016225,0.3311957730185262,
 0.9932861728654722,0.6393936322556322,0.8237693802842008,0.8936489348182202,
 0.4192200219805797,0.7984959373677528,0.6803127364972357,0.7731338853170748,
 0.09364082138550345,0.8468588021582995,0.15173842610595323,
 0.7966625916272821,0.15470479440270413,0.9951521374918243,
 0.020980705505435626,0.06895418987885982,0.07868436458031003,
 0.40250012669059454,0.23593428367802927,0.963360625932809,0.3459427817535963,
 0.704589438414769,0.3852966400027633,0.5364983148255319,0.7524968743560533,
 0.9990232499835449,0.859846833819257,0.19347604693768627,0.8081355377417678,
 0.18076988147327266,0.9439076733730537,0.8018432355045095,0.9918536577596984,
 0.8973662649961063,0.512147881749718,0.7474773282998953,0.19491831995983022,
 0.7258278365179537,0.1774969744432211,0.221181806989854,0.46099855025314307,
 0.8151994158294472,0.8701464884515614,0.34660326025975285,0.7317762815253874,
 0.2762797118306449,0.9871578508557415,0.4053942341773401,0.4179775447415336,
 0.6094087453951059,0.27002501443146043,0.17520127420540343,
 0.8843816902816103,0.298310025473698,0.22060829849174535,0.14543569773930187,
 0.4014638737901335,0.3676959607316709,0.861849060295921,0.6708135459446234,
 0.7880389763200694,0.4620007053725036,0.4739134359787056,0.3678756297743595,
 0.6614773859336038,0.03872201559882227,0.7669050131192341,0.7019619998317409,
 0.11372616642545286,0.2126493631584152,0.9249336074743931,0.6488026199977438,
 0.16837769125091517,0.767815228590518,0.9510633363496948,0.17848558969038242,
 0.46057034121999374,0.4389795182328957,0.6570315776225847,0.9047451019239716,
 0.27735800372108677,0.7351049404008647,0.23342205786887027,
 0.18787891121863276,0.4401490176945083,0.138832844817821,0.7898354460260882,
 0.08878383600764939,0.8864283234665761,0.5252968856911324,
 0.04893811314161711,0.9586392118287537,0.0930041070904627,0.1391252480567835,
 0.6793212133132481,0.1764052396067194,0.838243220215907,0.33864495178192455,
 0.28195861970308833,0.792741223044056,0.8386184601743815,0.6727036715147647,
 0.6173038832920116,0.8388595955418592,0.8050915448141875,0.5702807726864454,
 0.09875380101027098,0.7608048565559976,0.39322536689772714,
 0.5292415727297818,0.14116073237986582,0.2069201487457648,
 0.29098482953558424,0.7546998883408881,0.3109385582732658,0.4782048756687678,
 0.43587602219882615,0.026385491445263476,0.9127340616319874,
 0.9792548399489197,0.326820883305351,0.06326959206193028,0.4424203439890777,
 0.7809132060164038,0.17624841761799204,0.40303392883259725,
 0.6985613796652408,0.30083551847846823,0.06905731903069712,
 0.010238030648401342,0.33305146153802334,0.7700066338346845,
 0.741766615830501,0.9063222900976505,0.8967585290252118,0.7623972643165385,
 0.6882341035498506,0.7733927930226464,0.713080222328263,0.8854885573049399,
 0.3836035148300696,0.23421398821724826,0.19242504507901237,
 0.9739543067426745,0.18033644641563318,0.7445527022036287,0.7646061685414791,
 0.5748823971737427,0.3107453320298418,0.6148262826752886,0.5697351683397576,
 0.05128288932750269,0.7980730524998906,0.7185677549783113,
 0.19536861362437397,0.12427954111931394,0.9227279240777468,
 0.2617014574917922,0.012689479377416713,0.27370806411292026,
 0.7410304695454887,0.14807580618168636,0.34489445612318226,
 0.6147614952082028,0.25938325648422156,0.21427341277990697,
 0.30816344822884545,0.11758018814478832,0.012545653279353641,
 0.2414120748004963,0.0798151782824662,0.4717660522155822,0.19216762104431495,
 0.32949109488887574,0.19594844223108832,0.30301470023797694,
 0.11838670210534963,0.84152523616438,0.14450652058361946,0.7658818379224218,
 0.7211872300748118,0.15510688130084838,0.051597194741287544,
 0.4809596166715854,0.4239171731267982,0.7893020103090354,0.7485400742278239,
 0.33510406350605515,0.17786251405713882,0.6204923526946157,
 0.5232173365317638,0.6753607310542769,0.75898176587379,0.09857608110396487,
 0.8659404567328362,0.8796448699378494,0.9628297390264511,0.44651090644374003,
 0.5980820268270315,0.2727813885536119,0.023031437264560473,
 0.5029379538494348,0.4353295298721577,0.8736792308017941,0.7725365739541314,
 0.13028651181407092,0.9438499310167179,0.035568326537665396,
 0.5613616874512679,0.7237147397765191,0.4960270244844688,0.02714367912306115,
 0.6484563262697874,0.1617400988562907,0.4642600064578418,0.5425989095247208,
 0.7787242244835707,0.4582680739592362,0.5190286693193684,0.21952070265801493,
 0.6964391215365273,0.04778219463985511,0.8224210062252317,0.4522911768713107,
 0.023838812699112588,0.631792089621515,0.7449276935338073,0.422288088414416,
 0.24224172899995167,0.17190177526933936,0.7762907321918494,
 0.6095675225190457,0.7819229220217093,0.3441577532488639,0.5380469299284735,
 0.7580985257413229,0.31271976250743216,0.7286988060546163,
 0.35495892021475295,0.44498495494404755,0.9096394856736604,
 0.6315186601913059,0.7034841167832686,0.7484000392801767,0.19140794261728566,
 0.4107551201329779,0.8835589550156125,0.18306870758828653,0.3494136311828827,
 0.24128885687687107,0.5283003515136294,0.3117947921293256,0.6450442539510916,
 0.49129597214996124,0.13864622742640165,0.40491252333382843,
 0.09661154881482237,0.4283466585307206,0.2715471349443356,0.5878952841747379,
 0.03286686455962995,0.8967603347234979,0.38339218083712834,
 0.08808788796332712,0.0389255761189633,0.1600538315043908,0.3036076641212023,
 0.1159406455970039,0.9663446682670309,0.2770241738343726,0.23642678324420452,
 0.7435405043820995,0.8248322801892541,0.4524547192694648,0.6415566211323677,
 0.6357544237703029,0.09197249260002083,0.9652436257255051,
 0.01176633601410526,0.2317116653715795,0.8550162000383342,
 0.25035302895176303,0.2821115627589249,0.8849646607756904,0.6118381736839102,
 0.17871942949219688,0.843170683740911,0.0175711963329499,0.5960282569308296,
 0.5025871662284356,0.23470817680063905,0.22951047563489746,
 0.11853066196552775,0.9127405328747462,0.5320795551984838,0.3425528094247178,
 0.36379145523478285,0.3805206198364789,0.7050283112957734,0.3304654980131978,
 0.8497977268830862,0.30341117406787776,0.17374153120262315,
 0.6237828198841712,0.797454880426293,0.2071360115289843,0.61358125178224,
 0.9887911730978864,0.3394329505629434,0.36975440982785257,0.5914777078164589,
 0.8571636643137481,0.7650800757144935,0.5832329273129458,0.31931565171606313,
 0.09220665608002732,0.8007109202516032,0.9644277781739268,0.6498399080472543,
 0.7877678268400079,0.41112600079527306,0.32522015645145885,
 0.7760330428365051,0.4893864569723645,0.6801780768777353,0.5477463371196665,
 0.6463305687466079,0.6017508566240158,0.8780547234893832,0.2596546572208892,
 0.3415900068378911,0.1659384705171727,0.3485254474424069,0.7095559372912049,
 0.975018892716125,0.7749521388034186,0.3557432353260762,0.17001044499839324,
 0.01941054061027736,0.18455024399932551,0.9555817958531971,
 0.7228141101631691,0.3270487373335944,0.6465907479613473,0.4991312873669299,
 0.1585114202535043,0.7161100165999614,0.0017295223514114388,
 0.6410410367832517,0.6234066494466294,0.2612606400326033,0.43033761677315474,
 0.2869415952109179,0.6517079324904135,0.17749864258697678,0.417926941727917,
 0.8217709321145498,0.46475436086110533,0.30338718509935003,
 0.4868955750140631,0.536107603438226,0.473274352777204,0.4947314145960785,
 0.7881447718817035,0.765202503595016,0.7377864731647596,0.30788258244447764,
 0.9855166072457615,0.4834892465760495,0.9612960085519568,0.13301310666640043,
 0.6311911359846959,0.5247077399651516,0.08958389156142066,
 0.07421211758162871,0.5240300320118744,0.15186697476705302,
 0.34706356993100296,0.34660583060410566,0.6488224657099739,
 0.22022354166919067,0.6309787202711302,0.5534300953562007,0.6164000189949899,
 0.1432068985197925,0.5103053568929671,0.9107506459667214,0.02411305264187158,
 0.5097766961710319,0.3330115308389401,0.4875818306915727,0.978487672469175,
 0.689245200252684,0.10721407068893785,0.43166322858470496,0.4085616207120425,
 0.0065818734053840755,0.9593272184809138,0.382510305729439,
 0.061918820939469654,0.3356295720353709,0.05495555765454785,
 0.6158051527998543,0.49418371943700223,0.6843887867994369,0.6319341280448756,
 0.8031407130706082,0.5774983574662134,0.6647031952476682,0.23163305983129556,
 0.38783146439392424,0.16007510223577792,0.3549745067613137,
 0.18322846876175602,0.4925229097590277,0.345629881444668,0.8608729252419419,
 0.13586973125271418,0.36909282441065205,0.44566155516160144,
 0.4056214031515494,0.6856079541540062,0.4625567889076171,0.902625569061355,
 0.8109951537892014,0.04008682629122284,0.23873283450155192,
 0.6925871665646779,0.9092350092567086,0.084978972335085,0.34191433649835457,
 0.724257006520654,0.28758354178895806,0.14057466528247176,0.9432778103851233,
 0.7429207818158593,0.400042131147071,0.7654736913961653,0.6721321767538723,
 0.7052418330011849,0.313363764721139,0.4442025472815241,0.7102442167462127,
 0.9638733966217299,0.041630849476107956,0.9018111496967954,
 0.06854674596426669,0.8486991636475161,0.29047452387285266,
 0.8841650987680892,0.38236442982414043,0.8010983492435,0.14083099008418376,
 0.2815534038527191,0.21277642750266945,0.5618484635107538,
 0.021003362276677318,0.1335485537601122,0.28473924754105906,
 0.15081095335649342,0.6167310981672992,0.2639911484861044,0.6918465217449653,
 0.9147971677284111,0.33117525600039144,0.4710366516774168,
 0.40469623777499175,0.8150226056804813,0.631466690459409,0.4719121301685202,
 0.04447574789868958,0.22130785995293756,0.009447481531284385,
 0.001144052620901892,0.8893832374909707,0.9376630192231086,
 0.9302577686456428,0.43024112179749285,0.3091434796907062,0.5699747561935158,
 0.7077330710676014,0.30388329134885816,0.6326359113876798,
 0.38745384126418414,0.15176517382128674,0.11159563416666085,
 0.5492510270678426,0.7508192020793677,0.4912437511996901,0.2954333420767522,
 0.34479160370793926,0.01862635165686266,0.5676576546409229,
 0.1095170433479441,0.8039834217022013,0.01898808762306292,0.8811090349890129,
 0.10055998286154666,0.9049655361420139,0.37561421083103375,
 0.5800962218523799,0.660301656890633,0.4575072950310197,0.08287604465134601,
 0.9957710693052679,0.6026907450247052,0.7281986520160376,0.7031372611742709,
 0.7718696143269377,0.4012411168986686,0.6235954873414469,0.20109157880905493,
 0.5894351201355972,0.6226373296085689,0.43205392712935864,0.076300330463836,
 0.27241581222699573,0.9109491090313809,0.760565341027722,0.34501140176928247,
 0.2416192927757379,0.5712062247792953,0.6589484486527901,0.902136442277324,
 0.9561192880913262,0.5265353334593451,0.3725542755850585,0.3404929555980628,
 0.7972002339281604,0.6897224099876181,0.5087296123223706,
 0.057958728826800976,0.040993242557236464,0.6133900539744471,
 0.73940045412646,0.7274185067694766,0.030111201832562107,0.4406658891455518,
 0.2364334399255943,0.9943303683552612,0.5978612117730637,
 0.060223561486306254,0.5544003668846764,0.6672847324982809,
 0.24151841876563918,0.6748602533399195,0.5047344679593063,
 0.02825012367378521,0.8020676440633276,0.44196569283431986,
 0.5329492170606454,0.6757035656333991,0.216964840908331,0.10261896527567771,
 0.6040592791968035,0.8120231460257188,0.6615789356903804,0.7768002411110693,
 0.5154896120115959,0.35403293284489346,0.7656556471262358,0.4638867410929258,
 0.27922725737897003,0.8671798672008478,0.7259182123375072,
 0.22490734245674693,0.8083902178419261,0.4780603493163196,
 0.36799980901463414,0.40591620590198685,0.7864058869263296,
 0.05008476904386561,0.17607114450590078,0.7764022571943682,
 0.8552679819687357,0.35316613426465515,0.472597297635696,0.9310276278518773,
 0.4005339013329364,0.7861172116864871,0.8789730900449985,0.792080087610419,
 0.08818828435666337,0.8908680068566027,0.33501182769415216,
 0.41332737233796324,0.7883000425651691,0.4279161097458034,0.957526087268833,
 0.38943769000272865,0.46946650592457817,0.019522080092481686,
 0.5213694827177257,0.03810960945819153,0.8047638602906871,0.9266869507378097,
 0.7989379308200312,0.05027539618612853,0.6906081134296951,0.9825910364375927,
 0.7578511664103873,0.12476153657475542,0.9733431940637417,0.5743548537194575,
 0.7947892180308906,0.2930407583363547,0.743787181004055,0.1655616404967145,
 0.8901931122423812,0.6832147984451902,0.7110323023309366,0.3017889492516591,
 0.7644000591891871,0.7787459986796823,0.6983005732747076,0.878335034983977,
 0.1760684685407511,0.3231360630872335,0.5809580986211683,0.9174418470523841,
 0.8877484940286486,0.7851163460256899,0.19925107408620335,
 0.06617282379209521,0.6770672352428659,0.5054148039609905,0.7343524061390471,
 0.9981389086914416,0.37742683106598296,0.34719460617641007,
 0.017567506446528047,0.11622700418126664,0.07122177206993863,
 0.7060514120446996,0.5076644164510253,0.252378731780766,0.07839354960623957,
 0.25271502539447344,0.9420778899497101,0.6961039126927594,0.540714914519885,
 0.48513498315730186,0.715338464395877,0.3327829938236281,0.6863346747591844,
 0.177209372929222,0.157034521636769,0.4510302431417993,0.9498189160322518,
 0.11461785501245125,0.7764357328014713,0.8809845794777647,
 0.09135005433848287,0.7197622931572815,0.43602645877321045,
 0.0601029968304122,0.1288165600781893,0.6264421710105118,0.19567590881506813,
 0.2816785264945829,0.7211361571904965,0.2503353863887525,0.9858565722229213,
 0.2554347503992108,0.17473457381630375,0.9672677652245335,0.5873772242111788,
 0.21427983242488824,0.19102387297541867,0.7705862882440542,
 0.05509780312393864,0.7066234425927237,0.5841595852299151,0.770742863706823,
 0.4718658802037756,0.6262430827804196,0.796035724761351,0.23233301585180766,
 0.5379786288292505,0.2021938129892643,0.3300688959852023,0.9964489875193038,
 0.049880451084248656,0.065464364650348,0.15088078216176448,
 0.13943330445514912,0.4783544426931463,0.4804732498745259,0.789274070613494,
 0.4006742939273724,0.3936071845856024,0.4960311444323733,0.15696669015903897,
 0.07290746604160847,0.6107588249007366,0.2006183429421664,0.8190751199125282,
 0.8086301665578931,0.04688331260514744,0.7840024341037681,
 0.09388254159766574,0.42306562789003943,0.8386942305860556,0.388247258007433,
 0.01784390857854201,0.3049200847251242,0.2150456424805952,
 0.19421461995548017,0.7723141838332088,0.3744439046979542,0.8174550808910837,
 0.4512411204697868,0.45347583474678244,0.05378402705548058,0.918445245160031,
 0.6277946443388767,0.4816342753301588,0.12614627164897452,0.5126540676089302,
 0.026187708818802835,0.2115165842156822,0.8597319361265268,
 0.5340442944128982,0.43869197537204574,0.826305333555937,0.267494932449022,
 0.06978552950625294,0.39657469817799984,0.24803871581535364,
 0.6346597997561053,0.5427071772947037,0.9536516361252733,0.8919307927396906,
 0.5987162908751017,0.9790434537284578,0.18253406311676967,0.9077690928809209,
 0.8715714655174812,0.38261212319305593,0.16589750401847037,
 0.052504228915369766,0.09810366762059308,0.04870202334702112,
 0.8792858032344714,0.3859029059212169,0.32401610750075416,
 0.0011908841704927342,0.4456007841290115,0.016150552877380542,
 0.7469918832904604,0.6981117982247362,0.3242737978038214,0.12255036304727718,
 0.3661525116753872,0.7241405757878608,0.6674371878648883,0.05560984752671261,
 0.6635146893527584,0.7834285607099919,0.9043603299762251,0.6151308973967384,
 0.021932714369601536,0.48032291666269833,0.284283432717246,
 0.9780438027400412,0.2147341166362452,0.33606729451353956,0.9636744068659479,
 0.13520388980307252,0.30586917347983655,0.20860329368691688,
 0.6014952531284798,0.8961599518373404,0.24743980252258646,0.9274331491928415,
 0.3020474574446472,0.4437594224671688,0.0688113789218685,0.9168606837349256,
 0.21946004388737939,0.22582609680706023,0.7003226077577438,
 0.18372801439751618,0.6081045684448168,0.261175327317422,0.773757277924137,
 0.6726439474588681,0.8897631150671835,0.17708008741866887,0.6370894205625489,
 0.890235594414523,0.3679183340786134,0.41336610091561254,0.49042911950237744,
 0.10961380739816562,0.730988749074568,0.6092991190925481,0.29042651089146476,
 0.657576223027581,0.019351658046802944,0.9659033105211383,
 0.29467708118002467,0.7188681523894151,0.7770584827146064,0.8524680918285259,
 0.02975376422685516,0.6213873677939409,0.8772309791579639,0.4108998644920818,
 0.3639270295758722,0.0805709973166906,0.5524108710829219,0.4106169153232805,
 0.1525373682826927,0.9502486211428123,0.00883790101243953,0.5163764350458254,
 0.15382882235874829,0.6168387656506631,0.25546975078928436,
 0.6666941760512893,0.44067954548661015,0.7037288698122766,
 0.22417495375299212,0.8011445437569509,0.6795920108355381,0.9337726051591646,
 0.9112327774301647,0.0208207834800207,0.6188208904776898,0.5095091893532576,
 0.39269145978383346,0.6042516417354511,0.4158738171119478,0.8464904092307315,
 0.2671181366484531,0.5960430624489507,0.010610357087988853,
 0.8548521882957901,0.5669005514056584,0.49759585699663755,0.9548877378289891,
 0.5756017595579854,0.9393514304847866,0.5207366953721563,0.8473763135851057,
 0.12319926952281068,0.7489634735785388,0.4961695639633801,
 0.16039316759433875,0.8555847682263942,0.6881779446774654,0.3942486535748131,
 0.4181772861089478,0.905836362148353,0.038783889694129714,
 0.18055486427019118,0.9626487888950889,0.1712964446909434,0.6686947775989581,
 0.4511027341441872,0.8452115769175794,0.3529777727489247,0.8728657564859229,
 0.07970737032067576,0.817891040657724,0.15730807444992934,0.5769938089896045,
 0.1523602896573174,0.9645685536657282,0.3479530408754886,0.5689203088996013,
 0.7694457427651377,0.5065004442052652,0.9535209914320127,0.4842185792597401,
 0.4368390851090035,0.7353106307850549,0.3645900962962201,0.5093059404258535,
 0.7414654447034051,0.9392294865674554,0.42006628766006815,0.8127195697146664,
 0.31257594797645105,0.8320432197738407,0.9552740793021502,0.6054154071477529,
 0.6895313264096831,0.720815604481602,0.8220345387952033,0.36745657871630466,
 0.34360823851416666,0.30501954592505764,0.8013980329533666,
 0.4185597096987227,0.7585655623943574,0.10651222784190706,
 0.45001537023644045,0.5753682874996374,0.5596205739996394,0.3879872464845011,
 0.6046609166511148,0.27796132046827327,0.23458108670947186,0.706116418418452,
 0.2434252857643684,0.3604703650357336,0.21585702644194094,0.649253388686659,
 0.7785799729214947,0.5053148057347662,0.21116947681983156,0.4263032603143643,
 0.9339779002238654,0.8922503847660316,0.7445297394443767,0.7097437195746878,
 0.4155758009769338,0.6781841390572041,0.397596283052677,0.4976020063841946,
 0.5129248245932925,0.9244713379169494,0.2861420597421931,0.2479017034173222,
 0.23038857353598496,0.9911683121895416,0.9952269970281498,0.6976103929481885,
 0.849252512638921,0.4373458728704107,0.7838618912476019,0.9831252896317839,
 0.7880959149099008,0.5924421774704894,0.8891045445404382,0.5797860453978061,
 0.9314610632117681,0.5299635742382287,0.8168799635670578,0.25105371030323387,
 0.7884950042242557,0.19813857857505957,0.4684795916642792,0.2899925367200561,
 0.45481617432340127,0.5945226991084622,0.10269526493661141,
 0.43562876785107874,0.15546474404143729,0.448830201771537,0.5041452460162592,
 0.104759096716318,0.017281085584503808,0.3107203035111741,
 0.21723306597808412,0.40326965729763675,0.5784841592073766,
 0.22952728421103308,0.28794201724344715,0.827506724443422,0.7835623964951678,
 0.22274538941880695,0.3860009373847636,0.7332595899132243,0.6155487164152361,
 0.95164952894728,0.673987104607274,0.1599283752338425,0.7756671776074504,
 0.5653385458943787,0.25057162719529247,0.7898097551158232,0.7912157974845906,
 0.6389561522660236,0.6283836089938796,0.11146637292693073,0.9984425058798857,
 0.21761784301756626,0.41941989577574634,0.9123117773824502,
 0.5267612683828138,0.7790493703078317,0.9186419134955154,0.5359426824334453,
 0.1569675413144963,0.693599051718959,0.7298093571507781,0.6424535858438651,
 0.9556182211054269,0.4463999337379645,0.6542103282134338,0.3574263803594445,
 0.9421344673304811,0.9198870426621583,0.9721544348344642,0.8766594202216512,
 0.6706975942688334,0.9092721144364188,0.6383990160285602,0.7765915430242846,
 0.33958776087739484,0.8015482844491174,0.21558770269631966,
 0.01731666031032708,0.13323167586048978,0.9200187201628444,
 0.9868867769489142,0.42122565232158726,0.0951526374321624,0.5480288032008472,
 0.6142746490002831,0.9077515746449589,0.624555332099507,0.40764464927039523,
 0.44972605569354485,0.46308379396228216,0.7386333192683111,
 0.7729737974213429,0.8566933288337082,0.766429667715394,0.014391214540898756,
 0.4982917846178697,0.5789087873302765,0.7232464347078306,0.985789780092295,
 0.7442852885677765,0.25758756098481816,0.008950187117945485,
 0.023467142958961684,0.746810003156498,0.045596829079134205,
 0.04943708052661311,0.46584048583553395,0.14922955664517357,
 0.3425903485234937,0.293587810009482,0.8070573307988258,0.4376949840010116,
 0.35208713639290456,0.5626295068193241,0.253781161157687,0.2773636441672227,
 0.14128630913493068,0.5533848879619567,0.18508359560206133,
 0.7339278141369868,0.9563049818382197,0.08106072608253345,0.1812642355220897,
 0.8067047736443812,0.90171280989747,0.6299333769035703,0.1594009603822042,
 0.2553990447536174,0.6947086107126174,0.508631971346227,0.3802712002670259,
 0.5963301930567145,0.6593240367976756,0.2633970899430592,0.9288162225263088,
 0.3490747367268434,0.2945300500951259,0.8343353795100187,0.9829608130817076,
 0.6302795737906243,0.22476362778453962,0.010140629511418031,
 0.20222506317447797,0.2136061019725216,0.5093820990205271,0.5914497550696727,
 0.2849161341716757,0.1465140792994134,0.4930296230996388,0.7301937530180765,
 0.5274481922910235,0.15406007389774012,0.8766669833351738,0.8085632472723214,
 0.08906600703949596,0.059005561318193234,0.9183402046846257,
 0.04514937370892125,0.7453541603428878,0.9951033803438849,0.6527598397748291,
 0.5800743401426092,0.7687419327504926,0.6344339235645549,0.876673453890917,
 0.3192850666728866,0.2972854665971383,0.2413224711647548,0.7300821089361632,
 0.4983937016073896,0.902589731311154,0.4170067704292906,0.20517647165217112,
 0.14950922579684123,0.818503628421074,0.23982133517596882,0.6155129495680057,
 0.8775808828586914,0.2387803049495243,0.9623277425821349,0.09287666911367498,
 0.08776810547911207,0.6701460186282387,0.8101791994938525,0.8825595780292793,
 0.004821224149403225,0.80677351495854,0.026635880799637413,
 0.6668776523917797,0.5572620954451956,0.5392202990440572,0.7986159474682584,
 0.5949111408952206,0.6473081145245755,0.5539618893317808,0.6293034189568211,
 0.5675742369971758,0.6914398803368224,0.8187532490366327,0.6449558768999916,
 0.7489583026808825,0.9675806095550512,0.21619070892799386,0.5874953416579778,
 0.9076901147302143,0.37019068754119466,0.6116243015500248,
 0.33560131765534873,0.0466378705412146,0.03816427119643784,
 0.6481016087316249,0.4340304698007572,0.12426225353370357,0.6447790319711962,
 0.26407660716662806,0.5742547264175808,0.6717472309167212,0.9727948046239734,
 0.9482150705627826,0.8363529544566385,0.7077743931867453,0.6884013923348782,
 0.15506096592685947,0.8363704513378812,0.18320958273101873,
 0.04160039588531028,0.67947621715017,0.8368074304494655,0.44884613651922745,
 0.4374496118108061,0.08061802629274528,0.9513044708739129,0.8512643042684597,
 0.43659717772947493,0.25126924861860245,0.45079414127181694,
 0.8863348990913256,0.6397058200517313,0.7825990092710561,0.7217025500949369,
 0.9460415316228723,0.07332949578845604,0.4483351811163281,0.8545510531563447,
 0.16851793412562432,0.9137052926176819,0.8523236002676785,
 0.40344909079666313,0.7442851634681322,0.0722802372105773,0.9328724105668584,
 0.9358629112220058,0.7778517665328106,0.5262392228505288,0.5016328239105882,
 0.179548526469837,0.7857489789080839,0.999696437904074,0.05493741087243498,
 0.8026605306961185,0.6599398593767277,0.8528632263069796,0.25499147498684605,
 0.3910068217597925,0.37368217248656455,0.8463089788263702,0.5598686424190626,
 0.6712277078731609,0.4951323910409964,0.08415627891513455,0.7397197066971994,
 0.2458372508142962,0.06970603792553454,0.3209088146533047,
 0.19925986959340714,0.1253417869171778,0.38335663280972965,
 0.48672465860154224,0.7135763203387816,0.20641053996789327,
 0.8628046339360089,0.17824853592226375,0.358816845922053,
 0.029451649215192277,0.8112731762171652,0.682112459301457,0.9298069170489656,
 0.35807302509441763,0.7767505126978163,0.022883467709348015,
 0.32383293447836226,0.82219473988113,0.4181987007392971,0.15186006808902175,
 0.37548640832617364,0.2826901316730066,0.17889308242062663,
 0.6282104684435766,0.7532028494131959,0.3925928710639456,0.9405415356336376,
 0.8378953806832702,0.892814433369718,0.3446201204976306,0.8684609594023311,
 0.5049037183370354,0.252529940132702,0.4855617678762597,0.7942672281476525,
 0.8956352940491327,0.39320074344229194,0.6866594455269073,0.3761009911376493,
 0.027351498959744758,0.7534578554100904,0.1867068356690248,
 0.6520177034116007,0.1785241741482717,0.5710361588246938,0.5781750505172963,
 0.09694567620761851,0.9190304388773133,0.3786461780876277,0.5914305992461019,
 0.05877572015067478,0.44583589411143,0.392429829709951,0.4521029554089262,
 0.6481680156202267,0.7259647186501013,0.06136155009735966,0.1661712506472629,
 0.09889529044683754,0.27612047397166317,0.6988821280977122,
 0.5531931273189755,0.16990008075116292,0.08977866144013136,
 0.10436510715370184,0.5466542393080512,0.500498505187394,0.9646732179185409,
 0.6164223266760233,0.859066108788338,0.2967351655592523,0.28510331720901694,
 0.12940214775461445,0.45209652143448553,0.9980429548014392,
 0.9774587087191705,0.32794861500414174,0.32733658542111266,
 0.6687818753190345,0.4762485612249602,0.1726928538681659,0.5410763537215073,
 0.9424828868706993,0.12357905350787168,0.16099866000089436,
 0.23839109623980825,0.2512143142753557,0.10989150730933883,0.609031981502764,
 0.8126084217014236,0.2878456667084984,0.298530460352055,0.8226983925297167,
 0.49015973889633047,0.2772726959419216,0.930788963610538,0.3475101450379128,
 0.17738540992474405,0.7660755516529861,0.9589814692696728,0.3228295047247052,
 0.546916179385148,0.07162780300626559,0.7190815917818802,0.9284970617917243,
 0.9630428533262698,0.32396625103691434,0.4775942602700629,0.5424688177408172,
 0.5409174345057041,0.3433976635724244,0.6494742286680203,0.1717988658653572,
 0.3597566960992128,0.4560573483208098,0.8083785209917472,0.31712421226357523,
 0.9246682033814636,0.0852805269047574,0.2809269899274319,0.2280044788269303,
 0.7725329044023452,0.5574908039352605,0.6598484661168169,0.14996630173110093,
 0.9629349544667035,0.7433996595396599,0.6892005156567464,0.08977632953265158,
 0.6948188146615436,0.20376799295480463,0.6250615845993818,
 0.30748108474306135,0.15808816414724192,0.07847861575761994,
 0.9783924263115104,0.7302517925742714,0.7845418476005658,0.28663938979958326,
 0.6156401688185896,0.337838860615008,0.7249546076078839,0.12504411341855892,
 0.871704414957205,0.19442282127302613,0.8432478016500224,0.7920301156690467,
 0.04517110878365227,0.3555742581615212,0.424474231186448,0.9331543715629737,
 0.2912849537803176,0.3564899124572152,0.24135871559834132,0.9900674019059231,
 0.35537350810215884,0.9032857585378707,0.2643584473275905,0.6992446281245245,
 0.9846358298410686,0.926320305671641,0.7088224687815686,0.2089791422009767,
 0.26446335006585664,0.9471973879179036,0.6569310061691102,0.7835066475174319,
 0.8887393553671531,0.5861594370849188,0.7540042859124703,0.170870932063234,
 0.9960061756958026,0.2164984113724563,0.27800850221007634,0.9309545995002123,
 0.40735601338868255,0.44137230141925765,0.036634889574992835,
 0.8587052548508601,0.07747923430615122,0.592294605591951,0.3378356367025611,
 0.2380258585779056,0.07030102764737534,0.5503141491068684,0.5255948411967953,
 0.38917417675338706,0.6197827247649376,0.12044492195327994,
 0.5312463978008363,0.7752210605852685,0.6181554250399591,0.9372854652038676,
 0.05261175936766427,0.221571372923825,0.5582241074814792,0.5165310637015561,
 0.2549498584706722,0.5749263279799498,0.7833841914747,0.5760142728916009,
 0.10166927877631449,0.9161577801696401,0.46385906898831764,
 0.37865946308569254,0.5948821627508499,0.7746558616993143,
 0.13484097159156672,0.5712731207253785,0.49354655607506404,
 0.8830893062535955,0.7778311114784154,0.016965351883983293,
 0.6934308998410207,0.8346217131435356,0.20866480116527353,0.4572611665319197,
 0.3212498904039093,0.01250636696626295,0.5334289584924244,0.1681982032349364,
 0.6877675848525922,0.7375228243753882,0.3569610413312806,0.5871997117298126,
 0.32926912825550436,0.35253129518411996,0.7691253264794677,
 0.08414221323579163,0.24178979471102058,0.007347880478191904,
 0.5613076009050857,0.41649384554614843,0.11136331496643648,
 0.6765611927574569,0.48423880534844854,0.12785406024750123,
 0.29762210110943055,0.3466339973822514,0.8461462492125968,
 0.31290643225388426,0.48263931306962715,0.9242326190449498,
 0.10811421896511908,0.33403147896681173,0.29700470284263636,
 0.23892605679165602,0.2833135794781688,0.34544015370646663,
 0.45836406568030963,0.0034307029081818152,0.5956616006484847,
 0.14657361376412392,0.1706359177438328,0.03649532411644307,
 0.9565929011076225,0.1035655418943413,0.963707776149058,0.6796487055715639,
 0.967671361629324,0.31088020236207325,0.6647258142272714,0.7446803015343674,
 0.05686340969790793,0.10157622863611204,0.711530221135286,0.5747571220071671,
 0.3333001220527424,0.804869316070591,0.9253928741202034,0.34799871948767436,
 0.37207104181931405,0.6749877236448507,0.5568948519957116,0.5208182003878117,
 0.36786619545853294,0.8164513410728713,0.07516674655945443,
 0.45219472546287776,0.5711458593076537,0.4814180491284602,
 0.46916636400148803,0.8359205266570581,0.17336022559565,0.48595424412397603,
 0.30294644716164676,0.17091488808928845,0.5329813851535703,
 0.6723034544111618,0.42292778317113955,0.7045853361293732,
 0.14228500611451855,0.7358139134308885,0.20991051500663704,
 0.5029003871912036,0.2680901369857045,0.17986584988976462,
 0.08727103620910326,0.9138539669050292,0.5692475655260407,0.1886913224540948,
 0.4043368431871779,0.38047580651636315,0.756116565821641,0.21952210966044006,
 0.8962574672640516,0.29805906190274023,0.7094989387784104,0.9406466555637666,
 0.8518981962588394,0.6649350776783866,0.21208625986408636,0.7589810743241663,
 0.42471503978719716,0.7159854972329499,0.2575542634204059,0.7109426579445931,
 0.7129065461539841,0.032694412474113754,0.23182837054933403,
 0.8157498042853212,0.5378991469160477,0.1999645276990547,0.8877844657881977,
 0.8768164625560166,0.05332500469599977,4.170151666027522e-5,
 0.8130460204075887,0.6633516432188176,0.8025122171888763,0.124717483217467,
 0.26420519803614295,0.7506047459324631,0.6801441655939502,0.42066709272112,
 0.5616996434045858,0.014454073227213815,0.6525697284743317,
 0.5689085475530691,0.28612582460895053,0.930958851220808,0.8248442138837122,
 0.31543553305850125,0.8462744928434756,0.6582588131952026,
 0.34570131169578633,0.519981250598007,0.31414035865820733,0.695474402898236,
 0.46391761376527285,0.8331122188656298,0.8926313048371507,0.9857011005037826,
 0.046544252568719746,0.958024517192451,0.09915416867352578,
 0.09069674287398721,0.6350015170280351,0.8912341080231542,0.4796427046514007,
 0.7721528224122152,0.5623549492454892,0.6626851250213714,0.11689126378047698,
 0.5104630201930255,0.38629433296008353,0.6059050286567087,0.6545921422967778,
 0.45995937897813055,0.14623242875311993,0.4635581133080233,
 0.5105120598598605,0.7085272558971015,0.17400641098364267,0.8630544302845597,
 0.05052153958751526,0.7745944575176217,0.4128417991257546,0.5354942114687278,
 0.27106479248849624,0.1880926809437654,0.27145610913475315,
 0.48686566544479415,0.9122664141813819,0.7161175032585222,
 0.13222746393788787,0.7637793674293633,0.8502171816035867,0.4699629629815708,
 0.14885682314720272,0.6084506678332652,0.15080115173097153,
 0.6042140541425911,0.9966732964736603,0.6496393645253595,0.3526199719584532,
 0.43125247925550203,0.7906583643026118,0.37185572015092316,
 0.07146422408677577,0.010093944817527789,0.5007116103632315,
 0.4584875734213275,0.6874170180923227,0.8958052107039645,0.8242063527465611,
 0.6831222628784576,0.6358398693653116,0.42091230349238795,0.3578707215895147,
 0.05153411931649465,0.058353985351296656,0.9840540211590362,
 0.059808584430579126,0.00702626093249481,0.2588236939556753,
 0.8774903721389814,0.6638246507587446,0.7795162723225953,0.9325861923870392,
 0.08260136771459814,0.7854026632588333,0.333449473862504,0.03863753625596544,
 0.47748005512420166,0.9696314783945262,0.7047550929889645,0.746763094654977,
 0.9466779030048045,0.46962604779347517,0.3567389050204204,
 0.003180354923584927,0.7476964182262547,0.8713974924756192,
 0.09671815902465064,0.5535838043550699,0.1741085760413824,
 0.004051260885136521,0.7587071319252062,0.8107176483574725,
 0.3634143671718104,0.06557924359434342,0.6107092175921265,0.7367005153108112,
 0.4604743564749988,0.7282433326073565,0.7714856997479348,0.3457969011284304,
 0.6393289688251991,0.6399423038380163,0.7473646841166807,0.2682495641015039,
 0.9526544378692273,0.7265422274805104,0.8922265650695362,0.6946684163774552,
 0.035772811442288166,0.8859835341553172,0.04885621122509609,
 0.7491680056052342,0.23836192617362695,0.7904227333523508,0.8537750813880022,
 0.5043342104986843,0.6608038159169328,0.2719497415603177,0.13211123102898925,
 0.7982608975820947,0.07867846414909607,0.9003758641608055,
 0.05190765999203806,0.3324195522736404,0.29472625541239117,
 0.43169918152786513,0.11463677242294523,0.21334432858128305,
 0.9086351358713292,0.3986055105885239,0.5094230903587038,0.2153705217428722,
 0.9707371750433997,0.34406502423688,0.044137697471605675,0.5706364628645245,
 0.6333911481855445,0.18232518467005654,0.681223161789858,0.0496156803228589,
 0.14849930030508351,0.49992481396220867,0.642141103727894,0.5575361654560249,
 0.08255118005391615,0.07096389202739284,0.0994566469246676,
 0.7257238353624672,0.49717575042447226,0.7966052236795778,0.6065411443548355,
 0.5280019759703548,0.38188503328519197,0.7959194907436157,0.414241976544215,
 0.435330699947334,0.6284740601057051,0.36043929485891835,0.4914436041155942,
 0.6652347206288395,0.14914926263298045,0.15020677059379373,
 0.5887092211318978,0.8937234018722205,0.5324831110841625,0.670303002133864,
 0.770516986253615,0.5486418683521067,0.3682648604410591,0.3476631666081086,
 0.3641453534004132,0.5370253735624001,0.789909926913254,0.22586829095812944,
 0.1874524426085945,0.1757846444764699,0.792096120713956,0.6740570846359625,
 0.6127309885651469,0.6249762960474323,0.5360200713863458,0.3619326620312533,
 0.04154091807565963,0.08185804709721056,0.20667006965856294,
 0.9851410321674042,0.9289950836514906,0.07205934119294943,0.7570908473454738,
 0.9562458820421933,0.26055543448819707,0.36383376031049475,
 0.5621223639543221,0.412517660713138,0.9312434990824165,0.025931993677601017,
 0.4348877042848116,0.810389661701454,0.2388124443248988,0.5570064820612599,
 0.24672193591975933,0.9335294721034183,0.6493367111474468,0.5459535291946646,
 0.877135262150223,0.38522874496942683,0.7175920339490185,0.08926476726036137,
 0.8485076310988413,0.5386599050755062,0.4766799512966017,0.17959210988888719,
 0.8323157782060003,0.5497561926486881,0.7580004988646798,0.21220224722815018,
 0.48182937465731,0.7301178762620405,0.9051746837851676,0.8422386977511016,
 0.4711039636726666,0.2500802754161763,0.6160686807053315,0.9109555485811822,
 0.33443340472491134,0.32936874140801986,0.6604081190112016,
 0.7322180635931929,0.9800636965944712,0.7046263738091971,0.4442589934865816,
 0.4827601522833789,0.32858142887812836,0.5541482649844793,0.5305077415211618,
 0.6539341998429467,0.1801374869665292,0.627473460633135,0.6768737654639695,
 0.846598474008735,0.5332731218028363,0.24414565931897803,0.1601861013356336,
 0.46738505495391613,0.6959374556752376,0.5979994580010618,0.9585024941737122,
 0.264207147755501,0.8187657049632799,0.28840297259853453,0.9921062273217185,
 0.9791200197203338,0.2655998340752168,0.018178019664141565,
 0.6417289086251772,0.8092232546390314,0.652818692278122,0.13151174515321173,
 0.18335194060302118,0.6658602418919674,0.5285638493338789,0.892422353111642,
 0.5430515656237198,0.2146490163508351,0.3092337674555863,0.5331828730132854,
 0.6563797191741572,0.727050613600315,0.3319417686733813,0.04994418077152474,
 0.2739641498756292,0.26266509040099195,0.3962455177728811,
 0.04295325000994665,0.1817363719867836,0.6423522844701687,0.5830121905598432,
 0.7538896631899705,0.3113509355304802,0.9308481474132144,0.870517561096123,
 0.28266199114273016,0.23297463655330808,0.5506457953357125,
 0.16632580405609876,0.015892404040136365,0.4112195526267337,
 0.20298941300495388,0.011340914922923329,0.7861453837793487,
 0.2547438791127641,0.3166446826223841,0.7902739822954852,0.3614081354046309,
 0.9606313864134226,0.6398896126344686,0.5699510007709847,0.6379002884705253,
 0.03989613169372319,0.8826063063242383,0.9986705108214753,0.8732215554347025,
 0.10521543140845502,0.6459099272088504,0.6729735684056045,0.5908787382058269,
 0.0019334601791592965,0.3257788435875677,0.15398448625317562,
 0.21088694205537106,0.9305775201657154,0.0510646690077553,0.8696938111694342,
 0.7445857882537488,0.9112647417037355,0.4802063270990622,
 0.021728347344576737,0.5823033677559558,0.26475109873034364,
 0.7013107027370843,0.8720674599997779,0.8171610697505969,0.3464535494237103,
 0.42318059523383267,0.5792326030845845,0.8664113133162744,0.1847342724379617,
 0.2174927958047519,0.35803762751482215,0.03165805288389434,
 0.8168019382707212,0.44251695270447444,0.4433863571397024,0.4636572694851103,
 0.9505991847150167,0.7311899874280132,0.17999087012906356,0.7169633489139984,
 0.18647437209953188,0.8620034904676501,0.21194188319009377,
 0.05427653174168445,0.44855675666042794,0.3324371147949178,
 0.7783022211990687,0.7557376645244054,0.1402981377700926,0.13411803851559945,
 0.3578341649315898,0.994035054298586,0.5177698286433268,0.39443746402829394,
 0.8210365385503431,0.8654781442113704,0.6391936622023472,0.08793124984548484,
 0.29043758834888667,0.5729286598571528,0.4628132026966717,0.7742807547137713,
 0.4201031298625866,0.33925898846950797,0.8573172047349559,0.1217510450192274,
 0.14062090187064935,0.9822509070378245,0.3945366383281985,0.6318447847896598,
 0.45993125092427245,0.6323135071596191,0.4180378664215151,
 0.020879134806305433,0.7769638610263869,0.5187959616213385,
 0.34078191689197623,0.027739968064615428,0.4078351525260844,
 0.5339956777456719,0.4257326932603275,0.5601556932245227,0.9855254424048593,
 0.4939557345872495,0.8529899811094525,0.580624034272517,0.3148341646890478,
 0.09497007783676925,0.5215962381188647,0.12051109070639887,
 0.30618109913643565,0.46099408748946813,0.26990256473612684,
 0.4736105977990581,0.7824285086038182,0.05056648485264703,
 0.41980952616037004,0.7322286543021351,0.8030859853184596,
 0.46025155418161323,0.44787103738001044,0.8142352323665678,
 0.7890649758151623,0.0436185431528332,0.8392492935680343,0.421094603152945,
 0.7919793406204345,0.5772509628748536,0.26493468821643174,0.8872263448397228,
 0.07490830319560127,0.44281547547863864,0.610728931329701,0.5135077215966897,
 0.2982276122230133,0.07477352441122043,0.021724372972672157,
 0.13246615915114013,0.023280832155392894,0.34406709197343177,
 0.46292872565981535,0.6977024119662225,0.4153355135382174,0.6662829060082931,
 0.9035401060962079,0.10763724611739856,0.1918787468438916,0.5497153956199955,
 0.12479382862979926,0.17631558475381826,0.12310114611002065,
 0.3020453605486615,0.1367710799645654,0.681221377367378,0.6363809849409366,
 0.762735645203019,0.04136839622101218,0.5185464960701853,0.49211817438602357,
 0.9012047052116489,0.6903613921989379,0.8080933419491758,0.35682732861596755,
 0.5802538785185536,0.5511228581189316,0.29711197217256524,
 0.15951456862568847,0.7267364295341807,0.9334380030930322,
 0.21559689387577818,0.4097255873975103,0.4194513491369122,0.9378559302127523,
 0.47993361266306067,0.3505342786138945,0.25082095930481263,
 0.6790812931171551,0.7353581373811862,0.02526504006889818,0.5075293884840576,
 0.6620654727911007,0.833169691425531,0.6593095253566023,0.1691093030728661,
 0.029817421615413453,0.427086970252316,0.5796938125481842,
 0.06632871325577505,0.8636811969976029,0.31237899933464564,
 0.8780379494378026,0.22045670457078903,0.26531778107171866,
 0.2827353247732376,0.5378497415967646,0.9710649536724333,0.3666871328079412,
 0.8542703800048771,0.36945968027775367,0.2932904705208701,0.8363098523713565,
 0.5187051510884185,0.17023879576399548,0.317345257903514,0.4893842560760896,
 0.10964122667061649,0.9797775658066935,0.35650742757934195,
 0.15191638189872092,0.2495094352473456,0.6302007794992012,0.3388054565590999,
 0.6071808335451655,0.9599897972498205,0.8371130167097802,0.5016549000149664,
 0.4116084322250597,0.0024224049586578644,0.37765933956700903,
 0.4009057626885616,0.7932991424500294,0.6207951902766897,0.05152906034197524,
 0.4235288018586467,0.4768328675672384,0.08632002497871749,
 0.05660437881123326,0.09831252391940037,0.383732982392307,0.7427775735973788,
 0.28570276793868976,0.26505103581321776,0.055104888631344995,
 0.7030194504390352,0.8290259808242125,0.9080685319651761,0.29860320180109456,
 0.43786896801180886,0.6835550254455812,0.6095568990049439,0.5504528106138917,
 0.043835636162650626,0.13225603773139594,0.3510521061861924,
 0.9179957548835374,0.6060052687204461,0.57503111796504,0.5719477669337497,
 0.7061396644077068,0.90126592454495,0.30349325884716694,0.733257231967535,
 0.07611261583842521,0.4236200693691955,0.6894625255871645,0.4601205437873882,
 0.54091758473266,0.3413245174382976,0.35383861358679725,0.9030887275506653,
 0.5411324920759869,0.6509987222485955,0.1494393939418498,0.45539459781194613,
 0.884248789658809,0.6862946754497068,0.05125645017413216,0.3506619245666759,
 0.4429153992170911,0.8062707716216494,0.7326478139238929,0.9020598896567833,
 0.7982449730569162,0.7057556575437995,0.5523513034954715,0.3759942011269828,
 0.49968441477090797,0.5853954700949425,0.5303781817695082,0.6377583226257365,
 0.31879147020025045,0.8059340666736018,0.5969932508652873,0.6062326181459582,
 0.8489768086717473,0.025288335073691548,0.3779152699426922,
 0.013311744125887093,0.09174920450209656,0.37765518117902463,
 0.9353694307449487,0.6780100993146637,0.38031690230863746,
 0.41538378777289164,0.5130056935188337,0.31385396452022474,
 0.9125774248675566,0.9306231698089539,0.019767039243678286,
 0.2870939085427915,0.7117900308541185,0.06848290241223387,0.3184501767472565,
 0.6047430384018793,0.31630688934265994,0.7157475323122673,0.978613165469708,
 0.43369149562672604,0.7820946406687347,0.8642290048983983,0.5073565345482106,
 0.9741220445900007,0.4830132255923101,0.7897624540686348,0.04137634852831573,
 0.08998437862567132,0.1882115124849657,0.07847506155316575,0.602737309233059,
 0.6934215481474497,0.17469191772897175,0.40144176862203507,
 0.9779072196372675,0.499384409650529,0.15665596007198546,0.2049561543721602,
 0.9827950412192362,0.9161029977617287,0.6995762725253983,0.5025113872385182,
 0.29037772683517715,0.37627748495231517,0.5355485721176434,
 0.04844075934309844,0.38764273418485784,0.2593948767672175,
 0.05038341242904454,0.09158311249081119,0.9733021747945202,
 0.7449230187287758,0.81080300765274,0.61463477410534,0.2426694728892228,
 0.658121899355083,0.8179874393739602,0.5239317824249361,0.5696431146822016,
 0.6521965831489663,0.8631473104570316,0.4857123400026596,0.39616620080237985,
 0.8081081569325168,0.2609397943251407,0.41185611121969723,0.7738325624177506,
 0.688689532678692,0.12897429713487718,0.9404399600497422,0.9286584365972597,
 0.5840674742538209,0.9592745697055195,0.9624240357358268,0.37074185947261296,
 0.4966251984011435,0.16351981988816489,0.4347089376365032,
 0.18349859045717165,0.664546026692837,0.3078058563211168,0.3083228667899237,
 0.8845488698177207,0.6241467245006745,0.010791758614968505,
 0.6328482095002063,0.9737955394769314,0.05900266963334444,0.9281874392459708,
 0.7086574140544105,0.9754028601922843,0.6817065210030199,0.8482939487018382,
 0.16000780486791844,0.7395175533751995,0.15986410632833326,
 0.16737620842637302,0.23078429103503106,0.11891105176957117,
 0.08677871814739846,0.15237851782801104,0.41425190536546863,
 0.4210934460610598,0.5292577049706269,0.27859159175163695,0.5178862471776604,
 0.003503404607059335,0.6815283429527755,0.2705950269225833,
 0.2382698928917939,0.43984884580359673,0.9390557231750305,
 0.26164053542494115,0.6016438927453123,0.25508340799346796,
 0.6551408698914354,0.29795905048370463,0.31344191978306446,
 0.6546213042839498,0.7787920109794797,0.7829425858317085,0.5362589329533066,
 0.691970059916198,0.525047058901067,0.05269538259430584,0.13621293554796865,
 0.4293025858744256,0.7881068479831608,0.24177276233375178,0.3245821598099625,
 0.8563542577727513,0.1876036394990609,0.22939767958636115,0.3554655186682679,
 0.18897575255223242,0.4557817613352314,0.8823952022398289,0.851849640129574,
 0.7003751302489757,0.743169533865316,2.4780596379969655e-4,
 0.8478113968716791,0.3840673998630919,0.3795724317372927,0.4309548880891393,
 0.8797735405268561,0.7548970158905014,0.2525299463219173,0.2516706944548561,
 0.04350640346667245,0.9848653268672097,0.7467307296110214,0.3888755898837921,
 0.734976485401472,0.1070074883636093,0.7318272524440173,0.38742875090343,
 0.9990317030568703,0.11261381302450446,0.44750479223173323,
 0.35672017937845935,0.4404866254829667,0.08128518843298349,0.657320047790669,
 0.537595659742801,0.2509010892772874,0.5544341745071596,0.5490231658882263,
 0.8196401855289362,0.9708531206949907,0.20344647427046914,0.7538788164152193,
 0.049230434016244384,0.19247103501224094,0.24717646295651174,
 0.7423970963037506,0.18628558805371542,0.7819309490515349,0.9618826529331991,
 0.6233612795708483,0.13450382495569224,0.1965673820645195,
 0.07756117811321284,0.054986318524578026,0.3909175792925853,
 0.45914730489113575,0.4388544641942057,0.2695137363763327,0.8858610352288709,
 0.7064325744128983,0.7587048045705846,0.7080920919161982,0.5259937046116155,
 0.17501523578543476,0.9339285279069537,0.6459689565879246,0.8662207928149838,
 0.510968300869727,0.34876318364820147,0.7991664813910283,0.26040840769547857,
 0.3792746649002119,0.5506027115154117,0.03298413027068958,
 0.10877911603104373,0.07183182560644663,0.3251555648595269,
 0.04113860530983904,0.05931339394617663,0.3443450398852065,
 0.42594265041525614,0.37584954544542926,0.5281789196259405,0.48531260307421,
 0.509323356022535,0.35405534152188267,0.26299124788088146,
 0.38663148837028727,0.6642853555451875,0.11094829427343644,
 0.8152953929591171,0.8353722284972388,0.13173319346717594,0.6531554714552209,
 0.30168922230204775,0.8940126137510882,0.5807079182171868,
 0.49541257416694806,0.8425173950729804,0.1129926151863021,0.3258275793233286,
 0.7599224266989618,0.2817799639812616,0.4008220147272362,0.06034392663157506,
 0.04869051997522911,0.21110663970964194,0.9079895396170723,
 0.9307548307723676,0.42745858943566517,0.6415216357751574,0.6109709620143958,
 0.9755417902934195,0.0953520720180574,0.6108316529925121,0.8652959859956216,
 0.22379925392038325,0.11827286609813659,0.09218492850467674,
 0.8737663893828802,0.2809290133195317,0.3323017947848641,0.6287486857462384,
 0.8372085041373302,0.18257099748166072,0.7279854077169363,0.3753074171367483,
 0.8724081280222118,0.7454459161345764,0.9401971221363703,0.04522184174795274,
 0.5862606593712887,0.13200685285587554,0.057231084043726455,
 0.20232382942560656,0.5012138893536772,0.5853893098951741,0.7056779975689492,
 0.1735269593570471,0.05196004116486663,0.12303033168580169,
 0.1971008057707504,0.7126377200771834,0.3966245655188433,0.12961911482267285,
 0.8214493680759528,0.8402754226089674,0.9482332119641173,0.23918012299748426,
 0.4551025497727804,0.321974554226535,0.47716515762772904,0.7876423650980005,
 0.6337216386359562,0.5206716004688671,0.8435602384548525,0.41504850134027693,
 0.15925920620442968,0.33194517947462654,0.7027518338977675,
 0.07241921834941323,0.21863184723907603,0.9520811338639077,
 0.07387542343165021,0.7992168251241853,0.4521819930029849,0.5229014277898847,
 0.3617788470547557,0.3185168186734899,0.5586760597271635,0.7343193714889698,
 0.6939637931424879,0.5443777611383231,0.10710204197083728,
 0.08084195797423877,0.5760765495329375,0.109898562010465,0.41769684476266367,
 0.7531649335468884,0.24795404332219273,0.8704130232051778,
 0.17236447890210893,0.8037923479022873,0.022843045782918625,
 0.11142487747910923,0.9483433085620846,0.8273547152443733,0.8672281094157919,
 0.7063380288600211,0.689528654453274,0.26413958370309365,0.34535406630219134,
 0.10668817588106416,0.6758047114650062,0.7702167078251537,0.2772057578985132,
 0.0392999288470427,0.5881777316346625,0.732671207103555,0.4970914673439911,
 0.14924660037306237,0.9234462021832532,0.9132190580265868,0.4847768275270291,
 0.36409577051351383,0.8938853986638644,0.7726236890061264,
 0.26886613536388215,0.7049750091164355,0.2957061769403061,0.4711801753353998,
 0.8073690539685877,0.07859074678651279,0.5164810362980334,0.6475365998609306,
 0.39237438115712897,0.11495988605730756,0.7487502469795033,
 0.7593133836491955,0.6058982588285337,0.4543176752365098,0.88595526415816,
 0.6688104908620479,0.9035848466261207,0.06295272470935165,0.7132755517959671,
 0.7601104233583109,0.011754242140652904,0.8662142278453939,
 0.2670542200704813,0.7617463276398339,0.5385200264301748,0.01999342582345187,
 0.5394695536895312,0.638238081071753,0.3464698493614611,0.15106920421426928,
 0.8643432165907068,0.7517167060854687,0.6679130149175513,0.08393612641597359,
 0.6627518209284272,0.8299892221719914,0.032733501744336646,0.609546236919442,
 0.21800569144910165,0.09692650488392118,0.15030395585597556,
 0.15544383519785931,0.34954552558309415,0.15563693447358806,
 0.3342705454597651,0.5218559506898547,0.36544137672436305,0.9550408702519496,
 0.28549967764528983,0.6332296279437594,0.09157332610274416,
 0.2481758744774537,0.29367384166759636,0.4172578949263739,0.5125899685286818,
 0.8892204066139918,0.948638699176026,0.09800130726293699,0.27180988640333936,
 0.3709013931141837,0.30268035420084216,0.31876819562024306,
 0.6141667862958504,0.6894860426953523,0.0530891704924884,0.3670756518098731,
 0.83003629306093,0.004096939126997823,0.45997245493459216,0.5282144517103579,
 0.7852936300964649,0.27605045173215614,0.046325544862774515,
 0.8273059784028562,0.20436629169754317,0.7298426002543245,0.6330321064882184,
 0.7992769255149905,0.3401994783635781,0.16956980216121453,
 0.060780384962307976,0.8359058474794081,0.519207084183368,0.4477793506000354,
 0.570461978083545,0.2530373890231793,0.09789828095075925,0.9459876352014825,
 0.5706333041966454,0.5902704212600651,0.8171681019830149,0.9368963096824756,
 0.3507007341371938,0.3224446537439506,0.3392862615229921,0.9842677048735199,
 0.7484394073231596,0.5061463034255913,0.09532386280430283,
 0.13625259467609363,0.175988886973792,0.09618584977023825,0.5954370789802124,
 0.16292935747245407,0.45813007082520985,0.883785014251356,0.6719382532202881,
 0.7580552738306404,0.4388435516832615,0.34197912710448863,0.6421215795689292,
 0.8139842254738684,0.1477318751369474,0.4370684212376613,0.6133478719227858,
 0.09228777480744987,0.6695693451777707,0.11575128824392367,
 0.7737516420525232,0.5504712892130592,0.7876917173812356,0.4550164616095933,
 0.632043883130044,0.7802548715674638,0.7931465194564693,0.5243930555680785,
 0.9235355358114716,0.42551470183306705,0.2743544505329609,0.4386641799815938,
 0.4969382770238344,0.2960926264839622,0.06840394210996825,
 0.37350013120486514,0.9682993744033466,0.07948482302020177,
 0.7315233870675956,0.9699290360144347,0.14208988113545795,0.5909486926448807,
 0.019128646506892943,0.2702416540282282,0.02764518292893081,
 0.6269978646876984,0.14104102800566087,0.89730793520441,0.47212515426532087,
 0.6123797025436776,0.8460629629324905,0.6705642034283638,0.9025174602581997,
 0.6276999656115588,0.025547852938701765,0.48199337293735367,
 0.47281104108909944,0.8629063740700643,0.16910620062812765,
 0.8724923615335534,0.9141741217457903,0.7812419073841046,0.3093481184519987,
 0.7388991917946577,0.9171684872544678,0.6795770481487808,0.985951565839945,
 0.17166021207461968,0.4747633790320692,0.20099413173539382,
 0.10045116290690359,0.3471221559308557,0.4787330098711927,0.2512523629626887,
 0.9840940018573269,0.7133203879752076,0.5286133318518625,0.06915540055530345,
 0.5628552150289473,0.746610189393413,0.2558434273180977,0.9564062817001545,
 0.428365704525118,0.3271137401533949,0.5432535548796202,0.9339293411728826,
 0.5421648091553821,0.030635792573900478,0.8705609033986649,
 0.2805998467278763,0.8056069276583373,0.6487795109634055,0.44124319948118385,
 0.41027001919736295,0.7987232484763682,0.3883257036444935,0.2550923720700736,
 0.17937611098803463,0.7315897815844487,0.2666948034858353,0.4572063874252581,
 0.489853621500806,0.32007521698098296,0.002125509009632376,
 0.32095113886981685,0.4345964628500678,0.40323110474354096,
 0.41945944849642214,0.18938573518880697,0.8190081357279043,
 0.8202910904506802,0.5408216222632993,0.07327561535374916,0.9773025609335669,
 0.017063963854801023,0.7989008872397503,0.4776960841109912,
 0.7905979215660692,0.8252138343483995,0.9721154823221136,
 0.058008609087197516,0.47569096073743444,0.28706664006191096,
 0.02255119926891047,0.34196650598418354,0.13484279603227367,
 0.6008344655679503,0.28138134280690164,0.6467471393360061,0.3180149606933508,
 0.36331115059948116,0.2575050450600247,0.6086429661339693,0.7748299879905498,
 0.47050553674318685,0.3886034424262452,0.8711572803171409,0.6371061670991756,
 0.26442236023500576,0.26906722569639974,0.7238734531128362,
 0.5505930766131449,0.20059580647511688,0.3314955531737451,0.7706765968652484,
 0.05377418241238674,0.11167741900618489,0.6442369926970036,
 0.3726656195414644,0.9076445047950841,0.40243939922216254,
 0.10672213200084346,0.5606837645698519,0.3802762918582602,0.5164629426399145,
 0.511994623759039,0.7759988545070909,0.007868963616220914,0.1373980250508463,
 0.5802743651363822,0.2570291579837978,0.9944953419627329,0.793961265141525,
 0.3710647841994119,0.1731928135483034,0.625977286886169,0.48013280077015374,
 0.09301721283606945,0.0866665519775548,0.8310372014964896,0.1117116107548286,
 0.5022875622622346,0.13517982030373288,0.40948406141240423,
 0.8775875081733806,0.09504732333989852,0.2588330807253372,
 0.12184088570696894,0.24718843686794634,0.12085674567577531,
 0.8438133078049387,0.7101667294825919,0.3853507289904313,0.784041373844618,
 0.72042231142088,0.6064797833786483,0.577942695908567,0.601711797274516,
 0.7626003659734274,0.9593733550483292,0.5263270011020298,0.05451172356815803,
 0.6198488920700643,0.5657798434956234,0.19383906661092898,
 0.31264605003451695,0.7933307266465476,0.7717365605429202,0.85422195495778,
 0.9608524184065612,0.9786064779235117,0.11307516879448865,0.0940859036192937,
 0.12222083877828804,0.2577015831761249,0.25006426460156084,
 0.7642743335455535,0.8032147310621242,0.8792652395293522,0.8573592178035719,
 0.7016281622555599,0.8497359621275961,0.5650359287679783,0.7173747856758192,
 0.36052704428140314,0.23757517725361343,0.3319214853103568,
 0.004872284740741084,0.7661455376123782,0.0688589318558428,
 0.6925938559513529,0.3281019805440838,0.4880863111099083,0.01979216029642117,
 0.19157495056796225,0.908955358989313,0.6245519530904536,0.13501116585743833,
 0.26803744605741153,0.7452798303560491,0.9055100426310231,0.700981886972978,
 0.4891646413529148,0.36956375043689504,0.40306315771086754,
 0.6706748500540638,0.1380174209082884,0.36331545344060845,0.7217143427161392,
 0.15364519172414126,0.4366521124585546,0.6817514857240814,
 0.20415381523619647,0.7684773703889651,0.11575979143462711,
 0.38523304685319204,0.8513625307016983,0.839611727819665,0.341806369709281,
 0.15118567780348535,0.8175887571235751,0.22214496879436574,
 0.9540000739574908,0.01829275567014832,0.7752450675828203,
 0.03742016645119617,0.3529022130118131,0.6075370944284753,
 0.06842944770952486,0.8147138414663866,0.006744948845039844,
 0.15852494531624972,0.7162799077284785,0.07045739049132427,
 0.2496687128100018,0.29061973392994445,0.03318689251691498,
 0.6723758075242762,0.866540032335316,0.6051279893016133,0.14685304149785272,
 0.5390524251795971,0.48682888204542263,0.8758659740210933,0.3437138390043598,
 0.13474684010877536,0.9160820459465681,0.37835669841854414,
 0.28113845195762655,0.8939693811750099,0.7724993913929677,
 0.11507762897060081,0.7909253589283454,0.40512231528919196,
 0.13562868189279653,0.5049313704643067,0.8985368130875051,0.5955757401858945,
 0.8736717951861236,0.32882280089483396,0.0813011669184287,0.5721164057511038,
 0.9926096643102871,0.4623757615571391,0.9131297792609392,0.7060274115925173,
 0.4449360990671065,0.624823757682926,0.11436841618480853,0.14753822817816187,
 0.4645131255658044,0.33222679622932094,0.2537222263170956,
 0.12965227572979054,0.3815894617257469,0.1459738976094881,0.7857661519494548,
 0.8448934886908634,0.19099851069648444,0.7458277200272216,
 0.36772070136864776,0.8734945365917903,0.18505262582007231,0.404015599268867,
 0.9057800254818403,0.399598051830534,0.9470247432610784,0.8460510505402035,
 0.22102186588936346,0.4099894240790384,0.4761345459422015,0.0722605875194482,
 0.8182877252720537,0.5473610032710531,0.6034531175322836,0.6982276558139011,
 0.37236635498466075,0.6573810336073216,0.6250523492973236,0.5919157388017848,
 0.9946169447170526,0.9210412156881792,0.3817493535413843,0.09634753077196812,
 0.28515681987544306,0.40372464177320344,0.7338479831825669,
 0.9249406885495904,0.4302781231577427,0.7840005482120582,0.6440956798750495,
 0.7239243217701306,0.30952736195259356,0.5274968351301453,
 0.10276318148993813,0.12993958335504208,0.6525984154449063,
 0.8018504161309787,0.6121483881123122,0.05164738912597944,0.8311744418953642,
 0.22588417272988504,0.48624549227530733,0.7166373050733107,
 0.8354032269388886,0.8531960955816246,0.9516239298597219,0.11408171925374355,
 0.6009805446203473,0.8190721831906466,0.049838898570812695,
 0.8148122083040282,0.8677531943432522,0.9852111780983579,0.37190545946852704,
 0.5319421835382524,0.6157203523459924,0.8419902190062207,0.36821261366402225,
 0.15827360011442393,0.5633251095446571,0.2180369671155502,
 0.23149774316331673,0.5878349636313964,0.2969305589059632,0.8067221494508293,
 0.8514236394309915,0.48161711310206334,0.3004333609664872,0.5409965956170546,
 0.2910645627727865,0.15526449752949711,0.2723655763634706,
 0.35553688286354557,0.2593898371747366,0.9041860904101157,0.2563834149043023,
 0.4142456013269301,0.8058522210996538,0.4525838679741151,0.5896551012083097,
 0.9308202106197832,0.27136168117069814,0.5436230549933428,0.5888469731011065,
 0.29457028654132533,0.6800558474200447,0.799326866859638,0.23445250787053684,
 0.8583289475971535,0.3730576191912576,0.32628968093173605,0.5305080430953741,
 0.31183447517255614,0.6552179497361711,0.6681484465699885,0.7205555670212321,
 0.040372538230447175,0.5003055201510034,0.03759688909036618,
 0.8526261349423422,0.7563059651132005,0.643883532588689,0.5128161769643041,
 0.9876977851822824,0.5337810186416676,0.18974877713980565,0.8798539625605432,
 0.9873207703872203,0.6162603980689388,0.7277576109119622,0.9090323987019967,
 0.31985098304795223,0.73578045245457,0.9807512751446215,0.3873760213230274,
 0.6339317245399576,0.424990356486339,0.7783512822206043,0.4118297747657734,
 0.05126834784187306,0.7347165018529636,0.7573472143234524,0.4799359359867007,
 0.7198381541316856,0.7613497162893857,0.32294409829757165,0.2603011276398799,
 0.4610567627003006,0.4135805134149919,0.17606336466158634,
 2.7443713569907935e-4,0.7498774249357617,0.03954377382248708,
 0.9794029941768203,0.7286613420336301,0.8030252758475112,0.4794021950366346,
 0.2743694385198754,0.00462182754841578,0.9147267673682112,0.8217436295832788,
 0.4116650501656042,0.9181371046682145,0.4729784150726898,0.1660267998869659,
 0.9028889716398206,0.2910776007766568,0.5272604640806289,0.9584411228123059,
 0.014372122725016978,0.9108032322007129,0.5418545467190459,
 0.7728796566686063,0.6445413245669342,0.43367634319217685,0.6269470379712342,
 0.11771894909422831,0.8253410941565545,0.24803452423507077,
 0.9780993112010875,0.9393661630517844,0.37863420904237366,0.7039281822487803,
 0.8960182367773911,0.8368637488655296,0.5330746309178273,0.8618091238226562,
 0.04379310819557558,0.9364360184262366,0.8482883984326656,
 0.08345048391437304,0.7036774743746008,0.8261278585542068,0.9682523524525286,
 0.3392244576382567,0.2765692268439005,0.5857093672635799,0.7988876899064155,
 0.8544207832116116,0.18460272570661507,0.010161281718117587,
 0.6821582514691384,0.6128792339143266,0.16034031906629953,0.7924568530791425,
 0.7149437018499011,0.5077605338603899,0.2747826285930083,0.2917283140585595,
 0.8630194556457125,0.3065615076362458,0.45955139346845875,0.5541867316429612,
 0.29838769883167815,0.4866887493996298,0.18005583220819432,
 0.46645618082029094,0.7723256507720488,0.5657037624576456,0.8677507133705068,
 0.020688839834711148,0.3090027130127093,0.7329663111822289,
 0.2465091878786022,0.25187426325851114,0.5420145482025137,0.2756357660122829,
 0.8327129469493402,0.8871663830153992,0.6317525642405767,0.2767463717254861,
 0.22285439661623285,0.42647822728719076,0.5129718233211569,
 0.08516676743829354,0.4350292612422615,0.06437058685655483,
 0.1666074948115973,0.3873332887642792,0.4687154907563219,0.8325056819674914,
 0.28689345669672983,0.5337181126441564,0.41520845714443977,
 0.2320642565432557,0.9733537978608718,0.9644321474573758,
 0.007674952357209608,0.9344992245303249,0.28502367978668597,
 0.05810805666282326,0.9191252507452364,0.5544565524474262,
 0.28475282383286904,0.28743753166302866,0.9442678818392698,
 0.9822011915789728,0.8844102958278739,0.41019291059501817,0.1018032398205565,
 0.4496718775207803,0.13234991662761642,0.5484610200682328,
 0.36630997376757457,0.1252223486570474,0.17899944690836211,
 0.6398328236619208,0.7032288702152835,0.7972404492044511,0.3816941752151811,
 0.16712083282266776,0.09668110493556303,0.522911097720478,
 0.24258842962275484,0.7290360317837338,0.6492487712299504,0.4764508877075887,
 0.6286766742572694,0.5011579671453836,0.5114572297006864,0.4181953331745708,
 0.6300969553888495,0.6656319709021921,0.6321114423974682,0.04824037196135045,
 0.7437490644761975,0.5127569389363835,0.26277772550604783,0.8413197250115007,
 0.7267376541948444,0.06356989026037052,0.9404155074593091,0.7056184699432458,
 0.5119514458016596,0.05115684778751062,0.5977973819905148,0.5913631722597101,
 0.8595535916951511,0.9097955799816371,0.9739618703221158,0.6797461090567252,
 0.305763928039706,0.7171307567389011,0.5858909493909409,0.9489933331339387,
 0.2388155739714326,0.9528184682797359,0.21035932944901148,0.0667091423648215,
 0.7549157125827158,0.6940650733337594,0.5567070432732364,0.32760532974223233,
 0.6046299772548824,0.7296078718460104,0.7869482957605513,0.7991248050215821,
 0.8425831563784227,0.9395113741944044,0.1950494054011549,0.3634865284797151,
 0.7017765945104395,0.6521036045156002,0.8414863955999297,0.7889053444291586,
 0.8465349450625256,0.2808994154873915,0.9721111725112118,0.050345573381679,
 0.36164627997462995,0.9725734483900454,0.5495493645404157,0.7035972527801749,
 0.7459707267501001,0.8740778762765959,0.9092431206149663,0.334236045256316,
 0.0696869954238798,0.4423000466396244,0.9613013317835033,0.4921709256230571,
 0.36446202495891766,0.8917956760420982,0.0734732275260036,
 0.11354533229278818,0.6553308114670169,0.9300123385544813,0.5483080575453362,
 0.3709673830854402,0.2806352241822081,0.38925713733220824,
 0.004212313503584086,0.33009088433599554,0.5530442514612106,
 0.2826979664443091,0.3322100898800965,0.21759406896960354,
 0.44502804268521057,0.662102733531369,0.6107727307816827,0.5388138663836238,
 0.5805234213874034,0.5993558484612438,0.06597201942811726,
 0.0021346724661667205,0.6546146303833482,0.06534097996118338,
 0.5150480512424758,0.27642952605586546,0.735158927160598,0.9159040607978549,
 0.0038914909999520075,0.3794910079764474,0.13511178440603755,
 0.8066461939669799,0.9560134491507791,0.017063809228370364,0.343978410908344,
 0.7641049229655095,0.05718852000225394,0.3609780405676615,0.1549881289873799,
 0.9906822143627846,0.14893670417868454,0.31453296148669896,
 0.9365147502531372,0.27339618937273746,0.23490299464393705,
 0.2296747576672612,0.1763712229665909,0.6575175093999504,0.37129491191658537,
 0.9136583069743208,0.7330516746997258,0.6080104965725134,
 0.004455281736884031,0.32231239046409,0.5314142059962673,0.5792491878652569,
 0.8629356894059248,0.2695011807094667,0.34950851198282884,0.9434742023599759,
 0.7886711724793283,0.016996296450964232,0.3288798934622271,
 0.1738495605239032,0.46893792572943593,0.6268163751858813,
 0.29475204879471406,0.9309458786147209,0.33605640016031446,
 0.4402101867684062,0.5942564353628197,0.3684687506312453,0.5814148074450908,
 0.2513606046506047,0.8347724370772802,0.8373671181772205,0.5707777484140291,
 0.17906686530615823,0.008892439355701831,0.3936278134111386,
 0.31958602650882695,0.4177745049671202,0.9175220835832409,
 0.24305402836331336,0.3445820518479876,0.9122297511387427,0.833537465363459,
 0.9286593928593092,0.6591306266122458,0.7617327106935071,0.4265227708399012,
 0.9989655160798817,0.6815174487821772,0.37573106882813756,0.7511290596119046,
 0.7437383319216596,0.9119542352326131,0.7157279963641019,0.21626540297746266,
 0.018449852479513473,0.9933922084623953,0.332080324266286,
 0.07853078782945278,0.25256684290306386,0.16993534888241324,
 0.286733339989253,0.5774952894799844,0.7847625593359652,0.10655559746259269,
 0.31748026834687615,0.4303113045712511,0.7400536738388377,
 0.09814362881801608,0.42207107753596396,0.6586255669272546,
 0.6014660616838481,0.3974222558361009,0.8174795943878473,0.57702291534945,
 0.5354924456553353,0.4465283569716946,0.818145528681872,0.8154758158872232,
 0.40942267058426673,0.10381258737604215,0.2927797138731639,
 0.18923923247287777,0.9076630485755237,0.37930497098572613,
 0.39411453426904386,0.775506586277069,0.24143037140053225,0.8661586285237868,
 0.5735158660621998,0.9963722888183874,0.9020006796864355,0.07704513440705185,
 0.9701330847725226,0.2796863110368206,0.4648684236857761,0.6606185593785308,
 0.14483963355389873,0.7767503372999065,0.8926382614106592,0.5471565251934253,
 0.4297574245606719,0.8249530501810115,0.10941993490021795,
 0.44277937920695554,0.9608620087142841,0.22360601687859494,
 0.43168070600661645,0.4839840864342426,0.44195712913828267,
 0.9038324295451324,0.5642015566413116,0.268195479225975,0.05256499410429605,
 0.6248089739733876,0.6519554466888942,0.5340566412497084,0.5069266842427409,
 0.3123451093193843,0.9990940495484272,0.5512321292881694,0.4003009226214931,
 0.3189706295166501,0.37980133932407023,0.3251238708818164,0.4013696466643757,
 0.18501403727082744,0.13447442330014947,0.36129600995682676,
 0.6333496969771373,0.39924958258575816,0.15776046647964392,
 0.6232120054280739,0.7528276702581437,0.35573780858172777,
 0.015819934437470873,0.30342510251846155,0.7747984136504668,
 0.6197879841100484,0.15138471343403748,0.6336522360535496,0.7403209866618241,
 0.893612961003505,0.6934987344752336,0.35857375027224825,0.8074093016176674,
 0.5659291435747394,0.15154617764900746,0.9301331860009475,0.2803562735490043,
 0.8032190506841242,0.9231873209566709,0.535541342064618,0.9360836723429611,
 0.9873206888294104,0.5354698783063979,0.36780205383528386,0.8970237887290045,
 0.8632961863909588,0.9501520306452345,0.9707517047657975,0.40025265923208697,
 0.10231309939743416,0.46365948528651946,0.12920964786975186,
 0.704962251003222,0.4779299025843198,0.12532998706560983,0.13883771961983848,
 0.3697428543950759,0.08226649262307917,0.035275864599308426,
 0.15023577623074102,0.3501306127059084,0.20097194893867754,
 0.7089690796742486,0.4964231668090112,0.8770210263037864,0.45468776059005234,
 0.6817074997497459,0.13467487300924608,0.8577329608415619,0.0757222777017007,
 0.5579560095935555,0.5853259238897136,0.0963239438039315,0.7902230283600641,
 0.3348498509677812,0.5663920031062559,0.8172778134990533,0.09332182400961919,
 0.2764106167314334,0.865211877846065,0.4539352675892915,0.4837039973660524,
 0.08701178380081909,0.5915270685230086,0.059534579004474786,
 0.22547041470694384,0.6911642933835223,0.20066828686853688,0.947779478677301,
 0.9789406107712302,0.40494312761324613,0.9443473663962849,0.719074724425289,
 0.24621436511462758,0.820938376041921,0.68233605630119,0.3657942013915856,
 0.7706619925630886,0.9488176334866251,0.20256846870632272,
 0.16389547571456886,0.5387383117045862,0.3206622842920205,0.8332323941488332,
 0.729097986372537,0.9194847523794218,0.7380096373911812,0.08137159764165436,
 0.5889168926279911,0.706003871407344,0.252654960881711,0.8633145393835803,
 0.12149819750240576,0.5306096690263249,0.3593456900072278,0.9132681636400809,
 0.12317894148122488,0.9131516418170552,0.2214606838230767,0.9566141092978068,
 0.8395891324033367,0.47785078015033866,0.03389154455580945,
 0.15349412236284776,0.2575718121716547,0.35616239346034584,
 0.6660609488444216,0.5168870446176681,0.7483932838819901,0.45767268250897675,
 0.3813633596265926,0.07094879336637627,0.17866618816894486,
 0.8953500482849596,0.3752050872272179,0.09892624700018404,0.8251572998339705,
 0.16075179853083799,0.31026328398057634,0.3008863575936451,0.306137669786989,
 0.6169152743343258,0.14210396745040388,0.7366417906037547,0.9230784808675596,
 0.6119413409376142,0.032711792709565146,0.5226704037811725,0.581740988602386,
 0.5169791234582876,0.08786088388164903,0.8786784479688965,0.9094778824254873,
 0.5550952438765963,0.04190283022908703,0.30346453227042236,
 0.13716650581547118,0.9757071909956777,0.14035602347853315,
 0.8785462156303883,0.2883261180676776,0.47869942623035977,0.6849814405159073,
 0.9433575482621276,0.3500204214903764,0.273522283493451,0.7440317056303928,
 0.7239776244166035,0.21156872016877826,0.9566367004758636,0.1708094339403865,
 0.4797872838274574,0.8755795049535081,0.7251002325285878,0.7036465554510751,
 0.7859257583925217,0.5506061504077555,0.39332672525311896,
 0.48461388039755005,0.4461107566348169,0.06630502964254548,
 0.35000533517887955,0.7651310346476169,0.635391686476588,0.4020075781943425,
 0.7688893874412012,0.45879146472745935,0.2375207447375619,0.2689237696258703,
 0.7960524989402195,0.8729708310601195,0.1069954219672199,
 0.024514891612265277,0.6534500000656337,0.6880718659784869,
 0.6942582286765742,0.056100568556123775,0.33571999035922684,
 0.2895379628854927,0.8089547234822528,0.22713738112114112,0.3663816596948908,
 0.9916832238447529,0.6751675163904347,0.18819495095175787,0.87396577625436,
 0.5551708585288405,0.920179693060905,0.6026208126196357,0.3287108377019238,
 0.2629575290536843,0.3857329695347971,0.37180988140239335,0.867512970790881,
 0.8038371721961055,0.491244701301387,0.6474199080604995,0.6024983345099121,
 0.44390960632513876,0.5680092145215647,0.7176235622145932,0.789758599678873,
 0.5685016053361891,0.9109056497824046,0.979699174046528,0.0659564220661446,
 0.7117086594748553,0.06374458560777208,0.27929234561846383,
 0.4808352188974713,0.9733216852818868,0.9048486751243988,0.08520270922241935,
 0.9182633553421604,0.7103895192674827,0.6599592161494274,0.9404383231337532,
 0.41195911541232455,0.22081467315276526,0.14902856140587306,
 0.8950574038678838,0.09198475949073004,0.07083061223274834,
 0.20715464934484018,0.934170432454767,0.22112180739723286,0.7062259543123043,
 0.06843704042637833,0.8927895804531094,0.40434204864427903,
 0.20196242984915347,0.291861914032847,0.7832974283154854,0.10270249035786094,
 0.2762256284951685,0.5917100201961328,0.574717749305276,0.15280544953826747,
 0.8902403457659591,0.5750241252173829,0.22660553563371888,0.6289313072698371,
 0.384054041148845,0.06627644326949766,0.9044174836416907,0.17174287606176064,
 0.18254251377902986,0.18642448190707694,0.7678135771534448,
 0.2975129065079737,0.24457208852541135,0.8937634537315513,0.5605597786724399,
 0.5803457906925488,0.0014149514108812156,0.1577188638103194,
 0.7041371462266046,0.7957668000195115,0.17665251492713752,0.5508461317366574,
 0.2851124728311013,0.9399870298195802,0.3139842283812263,0.46577065526511674,
 0.7932253568813536,0.5600245788004461,0.8027481592514856,0.24391959655274265,
 0.9537517917396908,0.19565012103205615,0.20996282679360223,
 0.9646684905110093,0.2109038806487865,0.012904181754088806,
 0.7419693772144355,0.31349242964414603,0.6964430630456614,0.9224416097070413,
 0.5282948889740305,0.9132552036450972,0.6229973238145615,0.7453493345475068,
 0.12722467755724975,0.6012413122000981,0.6545514906692466,0.3026007555668523,
 0.6632633975772197,0.5537155997442904,0.5807048909390125,0.27104781683507206,
 0.8005065058856877,0.2284692508416758,0.9083778916570359,0.9614512239208766,
 0.1560069259575113,0.5028507755020221,0.8260759653853871,0.13683351617754314,
 0.7293407407626774,0.7356930278465474,0.17169564754185507,0.5410518964872298,
 0.30266671789498345,0.8225638357659947,0.6240436116565273,0.9863998169382211,
 0.81479190763914,0.3409323481092177,0.8049493899317749,0.39444750664324246,
 0.7706384020071081,0.49676994056867363,0.2518414095373188,0.9993120025284734,
 0.8799688554322058,0.17460253759957578,0.02027887468001488,
 0.9155914666875724,0.7594151865136123,0.9112856865100741,0.47274424977240515,
 0.5672963143450303,0.38747773102923144,0.007073815088258328,
 0.43937370146832444,0.44591011311584317,0.24457845018418323,
 0.9627817316498376,0.5124834336200783,0.5213411367507285,0.5939129259824164,
 0.25809617747882385,0.6195748519409833,0.7422936508169254,0.7479476628403748,
 0.5051112776878779,0.32201552855872184,0.850524843391774,0.6597900591888152,
 0.33934142190516026,0.9745128502414679,0.8434110790835222,0.5767817646811688,
 0.612272025946543,0.6916715802060226,0.47003047362546657,0.5986484403086958,
 0.8556580153376222,0.4620338048417354,0.9500654830699286,0.50499027111742,
 0.9775782200528115,0.3678250399441071,0.17130399754191594,
 0.05971263494408163,0.7667854529268499,0.6407977058401528,
 0.27873771209495535,0.9535032867702742,0.3517881154686735,0.4827233824342527,
 0.47083687092181226,0.7438486464059034,0.2841508539676785,0.7157910955934538,
 0.4828567746659975,0.005196643497952458,0.0795393007566424,
 0.38485689701348935,0.17085357122680778,0.5765757380747598,
 0.4639161710933507,0.4518265123742071,0.7951993278471111,0.7261950958935789,
 0.2670530756980303,0.6036382378535329,0.03414096503058117,0.2900944032036129,
 0.9825784240683666,0.3132750353122927,0.14485568575719465,
 0.13453878139529496,0.23084212565330897,0.7473899688516166,
 0.19135916470787584,0.4557642184858606,0.018843483577609232,
 0.29184964117795453,0.5972410733593244,0.25257794814079115,
 0.17842685570585326,0.645146375673847,0.6014354061705708,0.8062122270610904,
 0.7381658391745332,0.13978895060591257,0.19557693731823822,
 0.2872241684638448,0.8637927828546375,0.9461535544195936,0.9291267940209337,
 0.588231235985265,0.4521581262262323,0.5455212214610873,0.5193549226948867,
 0.7869782259468062,0.07907357329910025,0.5236118742738568,0.6656972783615176,
 0.5738483858603332,0.6594757943795914,0.6317390227191412,0.9754289877859187,
 0.2908431444879539,0.06244569817593493,0.14100577411659299,
 0.6729270519092491,0.3292073985709758,0.4424430641720931,0.8739655291077179,
 0.5512846023084609,0.5927249065872587,0.07107331772146719,0.9184529831345102,
 0.4777993343025445,0.6620873737498654,0.2391224661667536,0.8573048356032286,
 0.013946789182694541,0.660593174570508,0.0895970950628302,0.7424148314293028,
 0.3285776124830744,0.9462631703101447,0.9050586857153385,0.6180372996858299,
 0.1201740058360442,0.15281600907375437,0.4266345460071048,0.5085036090646582,
 0.11406093376051618,0.7450935055358365,0.1321799020247727,0.7546434928174659,
 0.0714845093962484,0.9947728523142964,0.34442545474721786,0.1121590183261747,
 0.1068386049896779,0.08805953036133685,0.3686482419717022,
 0.07520449696762177,0.8227596275870034,0.895492018816279,0.3632392592303074,
 0.5552146595451302,0.43996915257028135,0.6864468940627627,0.7763911481617687,
 0.7939201413531813,0.14634630583715358,0.4530376808735309,0.9741913002646377,
 0.6489116799318272,0.2637587047291421,0.5724650594592513,0.969700287784569,
 0.8777341242435128,0.2877804272404769,0.05185709046953002,0.3562057897240858,
 0.27928701342539464,0.21142642373684262,0.22396598831994385,
 0.930060829483963,0.733267040269375,0.25400380429256775,0.19789909937520067,
 0.34012655873981257,0.6726321218483758,0.9132148053702627,0.544237174637446,
 1.4771651764533544e-4,0.061810589065165455,0.3656866890274547,
 0.018641418254109254,0.10608627867886877,0.5000006957128935,
 0.39436910060567953,0.2120499213569368,0.7688822130788041,
 0.25445066726958165,0.1775537489781227,0.9795081080919152,0.3757685422646758,
 0.07315818249884964,0.25016695556606194,0.30376589539021626,
 0.13248579732555288,0.5743358406503947,0.8024831972887728,0.4067719647341782,
 0.34012129952677805,0.7315800089499059,0.04338520840581872,
 0.13387133958553354,0.15882058959857892,0.2884419039180963,
 0.07541533127276323,0.351711314776853,0.04110545877478933,0.3847348452671848,
 0.25915834906070423,0.9935197907099913,0.4850643862017632,
 0.10441487245320658,0.2183539286645525,0.11680745987240337,
 0.21055060975749618,0.28430530723920433,0.8233930726445298,
 0.6537321739631015,0.9039694657674914,0.6335140161747521,0.9622066111332884,
 0.6626375276339599,0.23524843301129061,0.8648907001196857,0.8672522990148632,
 0.0839533510764845,0.1937398884979079,0.5895039257950914,0.3046482817429995,
 0.0189321284981383,0.22815464764276028,0.910453495168971,0.880477792223519,
 0.07074386195669291,0.8568360552119876,0.8337433941888344,0.5374574095804374,
 0.47344353616885226,0.2996110364643165,0.6272520977289624,
 0.18159181134295177,0.962360060878624,0.3546066046226155,0.27723743507204146,
 0.23498529244971855,0.26591543456926026,0.11145132568618443,
 0.6398010647314004,0.9580389358110565,0.5448779309073096,0.9231180747542953,
 0.04657366769687721,0.1075819826335388,0.6307855846131314,0.8341882222654725,
 0.4256044292420914,0.9810198751935969,0.21230200017012946,0.1273030975514048,
 0.18729499989666187,0.5116152393215747,0.6294537350886362,
 0.029875979515977202,0.32497942384835,0.32208214104160815,0.695893264122854,
 0.4233558494574188,0.18153712180791126,0.49933258243397116,
 0.5465136506238848,0.5122372945543021,0.31632875266713567,0.5430993848097545,
 0.9390468561444454,0.8592833238554669,0.42636692327561354,0.9762063435535616,
 0.32243078621101495,0.5503255935164287,0.23978025215597787,
 0.3508203627546338,0.08929606110082311,0.26177220108541466,
 0.5003861288405343,0.16685325155940023,0.8790676024559962,
 0.21839966981194592,0.2943236035554849,0.6140329545485617,0.1088303322242512,
 0.9228456950129226,0.4646509030580893,0.7637390404307133,0.8203412786686081,
 0.10351618591760336,0.5110648319880403,0.7785547464618446,
 0.31931772964895433,0.693359933394332,0.013363606935785421,
 0.5471674836938244,0.4167630115287131,0.9349225137185987,0.38136712195521993,
 0.15206511596748573,0.18320209823650735,0.47472359199764425,
 0.2194362283584358,0.3096513654132662,0.2596147380728785,0.8169047816970367,
 0.417562215792621,0.09266293359188027,0.3436674055488338,0.7387460624876684,
 0.3104480297106672,0.03154165050937552,0.9412045278295706,0.5805768840319969,
 0.8303427894633402,0.09875015962960176,0.6097596488890863,0.7835469200505036,
 0.23646870512348817,0.8852463533009761,0.7862219841602297,0.5511986835690381,
 0.18151613228685404,0.3071886325020673,0.19938101415897203,
 0.8528584025327794,0.8283887749888635,0.21175322001696784,0.867724390898349,
 0.11839898410843674,0.136623015911711,0.5655711754544854,0.8151096921831157,
 0.651904594285355,0.23094228741716105,0.2259452487851128,0.4885345515689695,
 0.20226185511852723,0.8759480051604,0.9134360237490408,0.7849576598869342,
 0.23600313912749804,0.148026917720135,0.03773825152439081,0.7810427347414968,
 0.11307715274531194,0.15826512830235306,0.7572085576914567,
 0.5793505391731332,0.16323001871174392,0.061013878460353,0.5294285656944495,
 0.4899403486247683,0.06433074717549658,0.6448646644961639,0.7285361911977991,
 0.6604420015568006,0.3960882448980665,0.980909281616482,0.7519622686141548,
 0.3917396060033277,0.23726498313492927,0.8893940887197497,
 0.42523673668460793,0.6045449320379326,0.04995874024457325,
 0.40356441931787845,0.3422340028161638,0.4525548402325694,0.6046069593858293,
 0.17090177829145325,0.7345829137649489,0.819100568244175,0.07708704127182087,
 0.2357160545674123,0.022665439112318997,0.7051627227212687,
 0.4071322127255168,0.38904556258283607,0.7755637785822158,
 0.28816124108184404,0.6559726041399878,0.42393299620611213,
 0.8479951793348525,0.40205355890927286,0.1702917107756473,0.8593479362948,
 0.4498532311115653,0.08537245974123042,0.6022815056702976,0.5143459807590425,
 0.3109767359309282,0.0846329276092187,0.20163570170567802,0.9951251663946752,
 0.7403260311146925,0.2020144575130255,0.41825961065926676,
 0.055190202627895646,0.004490265676792227,0.4716877483282792,
 0.26935045157014514,0.846408305789646,0.6830595029088707,0.5519807916868292,
 0.030569932279496337,0.5198603880843164,0.5264065144282367,
 0.6813669403978804,0.9868278581313457,0.5449739705672317,0.28006048131152195,
 0.23583912371692295,0.5249369768253429,0.5927283210682145,0.5228215913093964,
 0.18030417031319312,0.2914732169325247,0.3684236185077916,0.8084904477406463,
 0.41092925377514544,0.4413768914953078,0.5234479292504478,0.8806805172749946,
 0.21460702232302353,0.6286429507089423,0.7109050166248791,0.8410474414409618,
 0.5303257285057024,0.7013398412535194,0.03356117185491381,0.6665826808134646,
 0.7118219582989931,0.4617499411158732,0.9935225527672008,0.3000385371014759,
 0.5886230348749142,0.5826692523055091,0.833545242034373,0.06530420388310099,
 0.28233943783280746,0.9390168088378806,0.8242445447381833,0.9830926990332725,
 0.6845269090500486,0.7306189134623542,0.23662202836708002,0.4865627805385292,
 0.9173605930244723,0.4635388379452442,0.19550418228390531,0.8422343538639945,
 0.20243872996887946,0.2845156103176698,0.20069465232593087,
 0.6456626061933182,0.5110813380030212,0.3920274312645281,0.48376325591929614,
 0.9133368827403061,0.18695245395053717,0.13021849290094523,
 0.5165235857586447,0.698193308186882,0.8446003187226357,0.3483831270348168,
 0.43643431590771975,0.442558063141388,0.7407176008161005,0.10103220312817351,
 0.6111973620820684,0.08035615825746434,0.22064545057517193,
 0.12501461276395398,0.7516665407885599,0.3520299019024463,0.9811213737672393,
 0.6435062440334011,0.5881678771494652,0.12043879734021601,0.6597422250563368,
 0.39775188473637224,0.019527552228599254,0.9624876687318723,
 0.9665929026648101,0.7338395700561402,0.7027855548442987,0.35794617774635384,
 0.1611164836622736,0.8038303723469616,0.21540381136447584,
 0.011647271211591592,0.12429697978428378,0.8813609272239769,
 0.5727690540141395,0.03957132516436512,0.7250998632615615,
 0.27944762415650104,0.15548615891361373,0.026203040746885176,
 0.3659458483196354,0.8300468717309002,0.1289636740046658,0.1813339244889829,
 0.021355819277091292,0.4520172152568548,0.7425999509288465,
 0.3772392180212025,0.10772256683038273,0.9375886268947795,0.8874388652827161,
 0.3678034599521187,0.006565203615819137,0.08369316896489065,
 0.4950589556498257,0.6304398862421402,0.9333167942534075,0.2827305258163233,
 0.025459658871834878,0.8235519764169962,0.9161977398828006,
 0.8456699548645922,0.9399345521934681,0.09135004554750514,0.5908155869327385,
 0.99140775003203,0.9727013167490708,0.5005724600398889,0.20899106503658738,
 0.513015438779109,0.2017330489322522,0.9284946740759336,0.4892108322303065,
 0.7924047023039757,0.6727951039930168,0.009992174389332309,
 0.9269936110653649,0.6591583874316209,0.44572414478652944,
 0.25435646611770835,0.959520138992236,0.3722257451783947,0.7286266584850647,
 0.2777994793683374,0.36332070987232223,0.2021745156548329,
 0.26158658438111637,0.5818029513980212,0.23952710220423046,
 0.015207284496414064,0.7815043494342764,0.9944446698876421,
 0.35574623155909646,0.1521051008255807,0.8879528015608558,0.8641730465353341,
 0.1214225803419886,0.6441253599334051,0.12022139862044556,
 0.053131366481204756,0.6387324930278965,0.6241857335973305,
 0.31356972683070905,0.9073084224625148,0.4606469454134219,0.558509638648885,
 0.4460262731530428,0.8930865576075426,0.2697089679101161,0.8948623052038741,
 0.8344079969666875,0.12321371084964139,0.3782785998031769,0.3237234768923114,
 0.45939444204851787,0.20175205572428467,0.8287215840222728,
 0.6970663711295518,0.7239715674802875,0.848104017759194,
 0.0042009932510785886,0.9978993687981201,0.38635456763495724,
 0.3412038229172868,0.005641233000046042,0.6673405380997021,
 0.4458679671152568,0.3003887396987559,0.24421300319563688,0.1650999272809267,
 0.5463193519239056,0.5415075850030979,0.08360509317399512,0.9731985602740174,
 0.39682445653531195,0.4538250233559685,0.4483070636732319,0.698301367981777,
 0.6986375439040426,0.07068908289796527,0.5981588323415769,0.6309303090321043,
 0.13243774680460263,0.5719980456253878,0.04059738856097972,
 0.24817623341744832,0.073593201114726,0.9724433147953817,0.2753937188322164,
 0.6258430947462691,0.12713908275165642,0.5932286162117494,0.3770287417908318,
 0.9486317235974049,0.26209872038633075,0.029236320713531105,
 0.0760711612740288,0.30332812572023515,0.49008310486387174,
 0.29782830975002983,0.678203729407753,0.9759598348572835,0.6739440786641695,
 0.39758310958748133,0.3030231389539336,0.819950446356563,0.9039553675853724,
 0.8714647043099648,0.26679652347450744,0.26210514216417435,0.724784406356123,
 0.9687060414117347,0.8463974684649285,0.20839089826314738,0.3979084372015663,
 0.6174511009163814,0.7079861594389044,0.013971178483332736,
 0.40124278700459903,0.24523445577262415,0.33985340215723325,
 0.25553198352613005,0.1702889102107652,0.16445853397527221,
 0.11895881420308252,0.9018326343816638,0.1307507207628562,0.1893573500755361,
 0.12877644154755796,0.12751214014414813,0.07588343076336701,
 0.2412185803239385,0.051621395761835775,0.4716998877711769,
 0.9544223548177619,0.43272682231525117,0.8491009901605455,0.7512140741517183,
 0.6465758910167987,0.47448719070499235,0.9292126819898305,0.6903404956626296,
 0.05491227676878263,0.1426852056966399,0.6981279295478107,0.4001413467196776,
 0.4570908770368266,0.7566732630135394,0.4016945094182576,0.40351732852373423,
 0.7262782570901216,0.8266026897435023,0.3762829493119585,0.7589448352221528,
 0.503569424967306,0.07156299937620747,0.03804676429232734,
 0.36943330670404717,0.03637194235543861,0.15815394573809982,
 0.8136978011422448,0.32387523945668373,0.9190040993645201,0.510504715361628,
 0.2654730131007391,0.6458883552113617,0.5017374981138724,0.12794007503984461,
 0.8894412881031963,0.7097088738244964,0.44828654386156697,0.6975143955008124,
 0.6942180025903324,0.6382437122603221,0.4850064583693602,0.13984611773732225,
 0.07505067756749617,0.24956993247700587,0.7167025805008432,
 0.5102515860585912,0.7686361926001385,0.9161189441236075,0.21637614808847017,
 0.6729871537014315,0.22322987043232578,0.3442605573285479,0.9959358191177926,
 0.876430920653901,0.2628526577184881,0.7532483596857061,0.02983529802126661,
 0.41169098013883154,0.09481095607250234,0.8939578388343519,
 0.16959852919378582,0.4269468166198356,0.09947101348219611,
 0.2685499570382863,0.5973951214196549,0.7459391283242566,0.45933675948113717,
 0.13543981398232574,0.3462548680871873,0.0993444861979581,0.9366393065742561,
 0.9027521368385435,0.3683547630467806,0.9426218870608888,0.9011977203107885,
 0.7983694855755376,0.4990192064868699,0.5982758541782889,0.41844807832737085,
 0.38772566539618747,0.27402722690748693,0.7133244714240805,
 0.3661641635326861,0.936121646458616,0.8829060883344774,0.19291832534914777,
 0.8655988820674122,0.7070994015886979,0.7584591023736688,0.05741613569041126,
 0.9055661132415103,0.010831969508640737,0.3917580591796178,
 0.7602213148647685,0.800262983219569,0.5525292622501381,0.018045378168748294,
 0.19150775183025848,0.26666328704013254,0.8581472635600825,
 0.4473545354211462,0.07245955441667506,0.5203320244692873,
 0.06141940533062484,0.3011934903518185,0.718574921383437,0.05512058606960224,
 0.16202591342503658,0.9552183797348286,0.4941632617795001,0.4016194662351348,
 0.401537872147391,0.1450546865360386,0.9567023053140636,0.2593599578067993,
 0.5973617396280394,0.44155367502528786,0.6566528171321581,
 0.08176945945350245,0.8013824238132786,0.07461524230390382,
 0.007951409052879943,0.6310651239438629,0.7172903310330714,
 0.0815575025591565,0.10194385401674866,0.23173500132826041,
 0.5914450144004155,0.029886992739064366,0.36366879025807175,
 0.07706790501200256,0.3994248815051601,0.2327799018348422,0.9997651770050267,
 0.5750099835239644,0.3935302712973433,0.18998976729295347,0.8046538223778652,
 0.7279315064490985,0.0804893639220402,0.6420352989029066,0.16547208382865897,
 0.3336812099701202,0.9773403454003906,0.523893758178232,0.6473037806978835,
 0.35902360015631396,0.6430548770760468,0.8117176053288636,
 0.24485630259507596,0.3933078761109511,0.07745332150447615,
 0.12289067404275689,0.34595803955620374,0.7037530869817223,
 0.4216734721258032,0.3016027731852474,0.854139740213987,0.705363742206402,
 0.9151833128149474,0.14872393563229713,0.6326324612817623,0.420191360780216,
 0.8776669638247382,0.7751807840998562,0.9886129701038804,0.9712758709333488,
 0.9278731104294999,0.7093899915497608,0.8607572939235827,0.7104569224324696,
 0.36415718198415614,0.9906249691092345,0.8974918347120296,0.5571300232433193,
 0.19967458693301032,0.20764131424165477,0.48090842962425295,
 0.416289820088777,0.7703578850951925,0.36928440827954023,0.6892908946558822,
 0.699861433221512,0.9183129307710002,0.7133458515668278,0.5181353719881393,
 0.3480289413739468,0.5909939210018522,0.4394280937411743,0.7053871498328242,
 0.48026994641762977,0.7824781572876491,0.13802747580151886,
 0.30220893288517126,0.9581678672907915,0.6407674903103725,0.8920618344295717,
 0.8067554749605367,0.23685979875098162,0.5479308228456653,
 0.27614925506644106,0.5989038812870657,0.13260611961153157,
 0.23009314962312732,0.5459620088686126,0.7630468894056299,
 0.031150354806107616,0.5214768142111212,0.8305145327425613,0.714264064812296,
 0.36926012265071284,0.26201298381531934,0.47354124795548214,
 0.8009572095888173,0.5309831958956971,0.747714408362273,0.9722147133977126,
 0.04858889941590261,0.08167656418515357,0.7867018438922491,
 0.016517102713286547,0.8247048799421243,0.17233389969018087,
 0.42625979320224006,0.6669432899575171,0.5294296528453262,
 0.14672905028122307,0.024118347690755382,0.6544691796583211,
 0.49529492536681174,0.2881994001956829,0.7934012081100867,0.6644524750808387,
 0.3026137089169507,0.7902722026941041,0.48293232878122117,0.7457351311033839,
 0.10541178210384994,0.8404615947847197,0.9155666674348775,
 0.04252579882356966,0.2732304425980442,0.9507091883686225,0.6111354829026885,
 0.8903974470175967,0.6628082135341259,0.321965975255606,0.5720862221956774,
 0.9211649061541545,0.6946034781481993,0.48997440896076416,
 0.48307067522644986,0.5623039917490343,0.2469315090368428,0.9184240551360227,
 0.5512480188771811,0.49025337775109135,0.28562997331345696,
 0.9709220736711524,0.7450294648538827,0.552834580419603,0.8116997461654976,
 0.6199389881827995,0.5487318002530499,0.5759470864664579,0.808295492503163,
 0.952937471620134,0.2342410251533994,0.8786555457844674,0.38335241246090745,
 0.49348788551640177,0.02414001577763636,0.782592932342707,0.5915988693684824,
 0.20723050191434,0.9767218464666055,0.8504954955878136,0.7131703031704367,
 0.49253724743071614,0.5871505118193824,0.8813848237197366,0.7504807754463839,
 0.7940678589811142,0.9684904875230428,0.9133090898753163,0.22644904933135357,
 0.5923902670046315,0.24391365938286347,0.08612666041199823,
 0.35404070601091764,0.7264229878104906,0.42985347443228317,
 0.7674752504825737,0.8903723397606362,0.6035422964126248,0.6477686022381683,
 0.021120475460260413,0.6446920095805054,0.376546797560372,0.8600289308538789,
 0.31236871443694747,0.08313050366075136,0.17556372565444722,
 0.9239202955305873,0.2963081015298309,0.9358024538294433,0.13078978919310003,
 0.4295573205844403,0.5594477069884517,0.8156062837595555,0.09917528709064982,
 0.798619658911251,0.20497515759618457,0.7769169191427766,0.40984979791342957,
 0.1896493859831042,0.05852324860360025,0.43806904261432034,
 0.5562955338043973,0.30884272782354305,0.4336087384386078,0.5110018290602176,
 0.3912618840777947,0.7023675634239489,0.03220997721265062,0.7920035190556974,
 0.4345111612421796,0.6261869980583434,0.2977502001447264,0.6552807079631796,
 0.4820263946684167,0.8495313395735433,0.8448842012165353,0.7231097546527878,
 0.30462132965751876,0.18382221904533924,0.5380763306348899,
 0.0021066195248302666,0.763113885291191,0.830852722650178,0.3235921795139416,
 0.014467557125441788,0.3952538939678303,0.8745999532035483,
 0.04194835648106576,0.5269318315094337,0.15114311493181265,
 0.11256688031080375,0.052246562104117045,0.5807950759090744,
 0.622113908303778,0.9526378664373324,0.5738340946770111,4.453580732916018e-4,
 0.6388807885758631,0.36052868354752654,0.8900068697205648,0.5552965349506759,
 0.28973705550034135,0.4466149046674227,0.36828979370371107,
 0.2296045616471296,0.29846639683701526,0.467218499313828,0.8917112078345941,
 0.7230140520072451,0.8503250639651475,0.8524213942803889,0.46158884582360316,
 0.14834020844188045,0.021798514601042096,0.5527056683930943,
 0.1137561720735657,0.9590205783902067,0.7428814589505625,0.3008260028845936,
 0.42723690733667974,0.4594681263747231,0.6348384183527379,
 0.22123888388332347,0.569951347090581,0.2769588061796624,0.7755127527644229,
 0.9724683656201655,0.8481453287650956,0.9001279921391223,0.7445987600831412,
 0.2042068164680838,0.23005225587850486,0.44778166062539515,0.945771101349851,
 0.537874843485995,0.02652716127214738,0.7937971303205741,0.43863857738031775,
 0.8475653368320927,0.5014143723608933,0.2515285183141379,0.4107650925200601,
 0.3230083158328877,0.4095183441470648,0.5562174883834919,0.6836895299530686,
 0.23556931719925955,0.18683760795972415,0.1555488073673238,
 0.5507038235128977,0.09249552328842991,0.36312507874914357,
 0.04262828592457657,0.6380250427137126,0.09009617582679907,0.535062223815898,
 0.9913696520406134,0.609934629148914,0.7035839326097032,0.9187457730604716,
 0.9176477988389573,0.30508593343332646,0.2029151210571447,0.2827980803607719,
 0.05312270439416722,0.5864049412821236,0.17985887792163457,
 0.06461626671632692,0.9258324282035648,0.8267739345640885,
 0.06547149186011358,0.43128302650592476,0.11501696210093226,
 0.01370984279048948,0.662251882119021,0.5560919899541007,0.33316154609237947,
 0.05700710849079893,0.6837288905784578,0.45853190670906996,
 0.9974527352182536,0.9467876372237357,0.3200880414595917,0.42647580993501066,
 0.4355827665328199,0.4585487617662378,0.7546551896293466,
 0.007757038953665085,0.22270995427809426,0.6044562455104259,
 0.5855658811052943,0.9924122004825892,0.7557219736972258,0.9719451687005565,
 0.8049354937325142,0.4303840320395216,0.875960227813101,0.3641657604017732,
 0.026245357626787147,0.6120438961720845,0.030174608772922995,
 0.7929283340938557,0.4200406045677607,0.3791896170532223,0.8559710140065633,
 0.7721367888764232,0.18661030427947878,0.8142537067211215,0.9666949796493138,
 0.23545418690566633,0.13584175708241342,0.023601574490852317,
 0.056527678870070264,0.15488032487877945,0.7980032500988898,
 0.9173745785963024,0.599231968018801,0.8431802229636413,0.7754400530506941,
 0.9522714928170514,0.9763738589925106,0.6170198865550899,0.3548120028542683,
 0.3212556332583334,0.6047995399787423,0.5584378245870818,0.3871139839837664,
 0.20156599392968688,0.0885153484429484,0.5537334688434316,0.7733072461030304,
 0.0967555651513674,0.41088637548679663,0.8898529858025982,0.5236420041746743,
 0.8864545496541548,0.2750846187576743,0.23235021518156784,0.9538058782769777,
 0.5393956289620507,0.22120492024972493,0.2548142780981071,0.3411852823200434,
 0.9029719317312072,0.6414147661662393,0.3607356574084738,0.8490250683763624,
 0.8221588317190907,0.8063989957782596,0.04139461018105339,0.4786075239718445,
 0.20469009493478785,0.582019172594369,0.2928292771127581,0.07733296651351296,
 0.8912929763037222,0.81836229885636,0.6777843477204921,0.675717028848348,
 0.7258119109236242,0.2987298636883131,0.09474493097758097,0.2515815318026784,
 0.9067038540337254,0.5363092981728312,0.05198806671537359,0.6191580371621147,
 0.366288454297538,0.6370247795096391,0.8227999397772994,0.8513277121518188,
 0.3238615416861266,0.9172385626140822,0.5138843887392769,0.8930017035412352,
 0.941778371551853,0.03497387414523134,0.7451259293609866,0.9475246821736918,
 0.7058625111869794,0.48135778901627646,0.4636344708735176,0.8811672156603194,
 0.9833583453613804,0.5790236349188521,0.7231696564104928,0.9814941048352269,
 0.06947113023002938,0.5352045210774757,0.25171940111525426,
 0.9281431119388912,0.2831343591713723,0.1140736009970067,0.24848172728954698,
 0.5620190071473121,0.004785097256426618,0.037279614181787046,
 0.6064974453962249,0.6609259340838165,0.9852545929897466,0.33182842483743913,
 0.2163990364381403,0.6103454740452976,0.635101236167947,0.1390250344616707,
 0.11534710169369755,0.9702708997815884,0.5723775420855967,
 0.10296656382992975,0.4157421599363569,0.6721104715220428,0.2283873885706862,
 0.595235352342054,0.5854052102066137,0.08376723298484501,0.7885336901636444,
 0.4453957596808579,0.041475016847920876,0.03166496191738655,
 0.27567933211066753,0.10697653757209091,0.6804368079122393,
 0.8198799415385634,0.4516465132895845,0.8757628385976637,0.4372691869332084,
 0.018364305704018702,0.31904210482047923,0.8151671546060387,
 0.5975673444983703,0.5195918802363629,0.5847501341744387,0.7758636659257738,
 0.9000499092623074,0.06442361520942663,0.9965823903993681,0.7360442281715615,
 0.1073894954009561,0.17586869783327608,0.8974051739413049,
 0.03723415305223865,0.4095196693833342,0.7280580069982402,
 0.37356194708940116,0.9079624270225882,0.07844184471892424,
 0.40570447372950014,0.4787231879649766,0.2509944676554461,0.6940998116466819,
 0.019680727388545982,0.5625104137329884,0.31605917226606617,
 0.09157716218280432,0.7420359978175157,0.5389589186004475,0.5342559058995644,
 0.4721465612275706,0.03080505648976728,0.12070597804460548,0.531726214237587,
 0.16080869665538544,0.3969278741487561,0.7326580210397776,0.8376591830435659,
 0.7350781366455622,0.6291189198912738,0.3464474614173668,0.04369208617532028,
 0.20644329196626776,0.911950210114324,0.8041684440026218,0.6644449193080219,
 0.27078991268178443,0.25512375758209627,0.46461692366016816,
 0.06978928204401658,0.5515794899012868,0.48623436788779617,
 0.7249606811767454,0.9471811100462282,0.606737521867517,0.2479878416895378,
 0.33661188314225765,0.28314390576353,0.1491068443536986,0.02886242751357415,
 0.5586704478528474,0.3911170865589103,0.3525404024242873,0.06501337105178984,
 0.38510744563112786,0.2567946495931571,0.2699425248570799,
 0.45687489415772886,0.648056138351871,0.9709752294096932,0.8592213092975998,
 0.7603922198447776,0.5580096198044673,0.901813472572387,0.4567269104069991,
 0.07178795714252695,0.9615236219185924,0.8211750413639365,0.2379978009953937,
 0.6468994125620525,0.4545630444198364,0.635014387662477,0.9776908424900581,
 0.574171312127731,0.46021861159855615,0.18277013388445806,0.9014820356722448,
 0.3110619664977,0.8148762973212861,0.9994130965181693,0.9785551428702509,
 0.6076144689447588,0.7310830711208509,0.09660912539643085,
 0.21507352927826262,0.043354069665659245,0.2810972174323,0.8317326718302698,
 0.7906049606797457,0.25287618197284323,0.9490284436590049,
 0.36833275402543475,0.8154946180344455,0.4584715296685755,0.5293654642558249,
 0.9309176857938424,0.49603296113003725,0.5738565025964829,0.7425608232771025,
 0.386684699695404,0.8050878443937751,0.2706575127018436,0.14680077035320283,
 0.4998167837057056,0.6721209798946358,0.09921096144184705,0.562963922729833,
 0.10305117345082104,0.5435354775673256,0.6761366769993722,0.4467000018510858,
 0.5370041371855241,0.8405947324383388,0.25815242868168164,0.908694804574579,
 0.6091278602474779,0.04782253893870547,0.26444667947251654,
 0.08752792337248061,0.40480077458866126,0.6648619481866092,
 0.3685691770575389,0.22674284318275473,0.4204812908829044,0.6828691247271083,
 0.10652158709189141,0.13355091003805064,0.6886933410329551,
 0.7611021682408482,0.46412762225769333,0.7315642222619465,0.9871240306225922,
 0.9853569478808687,0.4694271267225746,0.17216667897556726,0.5511716408817704,
 8.617518251827772e-4,0.7687805715122061,0.8895630254417716,
 0.2881151057875978,0.1138691442112727,0.12794458197524994,0.5973590763303288,
 0.6612182819658279,0.3264420133319984,0.21625387550973096,0.5483044632914948,
 0.9279185082870247,0.03479055328312475,0.3376433545115167,0.4262876997264561,
 0.7564760955841443,0.899508027193285,0.26481687558259004,0.5270853021540784,
 0.07907149950562076,0.255477673189455,0.058983926317514035,
 0.6645192087503744,0.023693029180173686,0.19238464247418263,
 0.18005870418169778,0.9689171059045886,0.23714115552225778,
 0.18685442992963586,0.8908981872821644,0.011626613375249839,
 0.2530589383651104,0.7599641675813942,0.5807908405609036,0.5328617824242636,
 0.7884573973590618,0.5259265716613678,0.4076240503599311,0.9028305625923554,
 0.5213771210062101,0.6720131269095571,0.2658845805097281,0.3435352183623741,
 0.2797741981967832,0.925525411388012,0.6154101742990258,0.020815409711561483,
 0.7284693592592775,0.8624963295292858,0.9407425041271544,0.31267790461271283,
 0.1228416209768185,0.7205558757776084,0.680621803216627,0.8788485585709802,
 0.9884198277825524,0.3275955024052154,0.5591291444948494,
 0.009979854609624805,0.7822838355813371,0.4882105929019118,
 0.8246021510142698,0.053973963178790774,0.28778443764079886,
 0.36476345876332383,0.7142619368180644,0.030025134384905283,
 0.18200666976149826,0.9220057080705748,0.17694679800324242,
 0.47277860803829475,0.3021262556861425,0.40680478265283526,
 0.8523988849735791,0.969844335291552,0.1414729632491034,0.8189595907595665,
 0.6241328090555,0.29816182145723313,0.7805520035519045,0.36355780836204343,
 0.8857336890141119,0.3801183790528748,0.6978849481487732,0.249081685597921,
 0.44880138434393513,0.9493176707014639,0.7790496563207424,0.2115293579279205,
 0.1525719175451451,0.8369524838026592,0.4620092623167096,0.7229321599314474,
 0.8578229803058774,0.20806056131723394,0.5381747497969934,0.5750606569134288,
 0.6570500475171527,0.14520753558120847,0.06836136072593135,0.534831935437322,
 0.7645416888685848,0.3196843318353454,0.6985011416417664,0.06870593501449118,
 0.2603951084870344,0.5299139922707019,0.18422234950867022,
 0.009937924148960664,0.5157861031347408,0.7018328442710863,
 0.8280432358279944,0.907072332331629,0.9356524272395856,0.7576766356758505,
 0.8646690748037956,0.40391390314202436,0.5098891951182934,
 0.04925447088487689,0.38231025920198336,0.7281330563354224,
 0.9880662655823717,0.9278071874431075,0.5324729547540382,0.25393902564425863,
 0.1291026059418695,0.052662203038449706,0.17650881669618446,
 0.9771294401197388,0.965695052272789,0.2318618724934809,0.48279362765300127,
 0.5637732400925455,0.2500571315672583,0.4331151911122699,0.2328832676262037,
 0.6096895991162213,0.717223741388593,0.5408814372264091,0.12170997671121098,
 0.28443202685801194,0.8374644204644712,0.951666120126587,0.3427897425435126,
 0.0141466214185747,0.4371095255619324,0.6925685732192559,
 0.012414936131069254,0.5550698690057024,0.10873783658843805,
 0.8399995049023582,0.44737285360633816,0.20368253999543628,
 0.3274483516920118,0.5032842744391862,0.9632881878412394,0.8952478450425548,
 0.6950020721197762,0.6263902097114089,0.8407694349000762,0.21696611122763665,
 0.16140536117955606,0.9578381826626349,0.24220054618406617,
 0.8639670856759564,0.20006910219929208,0.21352793685065075,
 0.8768162714031273,0.9348571858947798,0.4805478449669236,0.16238122608365013,
 0.2699408634852676,0.029793094468626102,0.7006196947418359,
 0.15951105280261257,0.7227897700683107,0.42918786435829825,
 0.47299733863191795,0.03643068182379672,0.20619412518601987,
 0.8525702188779274,0.5667056564806342,0.13216296912987358,
 0.35339514434428554,0.8115120892429691,0.7320434610711467,
 0.17457214134101084,0.4971913032084224,0.28082849279919553,
 0.08409014321257546,0.8570510410855343,0.5205592188943207,0.6926344953544107,
 0.2105919198157783,0.5261440521817393,0.7679929660201196,0.5307453281773521,
 0.9330372583511976,0.4193102491951245,0.9915929091262686,0.1488844803745043,
 0.4703694826586978,0.8997144767684161,0.11357678185998166,0.4742837539111666,
 0.007030345253956627,0.0012616174112132228,0.24173130466901382,
 0.456311359635831,0.428301789907347,0.15831991480618623,0.7432572898495121,
 0.6370394001181706,0.3038916634097575,0.18212243411144913,
 0.36531372266893924,0.01828319295544334,0.32445900624408053,
 0.6880746928776429,0.4443015779560078,0.9139039875149421,0.7372171635475049,
 0.2555194638568973,0.6068277664030198,0.31798306206282234,0.1458042116279702,
 0.2741654888441011,0.02473138517884743,0.2066966161301078,
 0.03156896696828193,0.9174428786301435,0.6072579053711674,0.467247810040305,
 0.5904190046100601,0.4531550097276632,0.14823439405476213,
 0.23878873879676565,0.2633214768289117,0.05310323223805702,
 0.8804610257644819,0.36638038821888086,0.6600375005413176,0.8563381128000854,
 0.05290675020680202,0.5051382625538757,0.6703001240899756,
 0.25852702328020627,0.8785700953759058,0.4674381111210755,
 0.02615396335539255,0.10131367354635112,0.03425627613641946,
 0.34804078010283446,0.34539089443636584,0.24290798506314015,
 0.7333780127648895,0.7175012016819947,0.7293732714325771,0.5889586457340514,
 0.6334788127089181,0.08286058418367714,0.2499475848906655,0.8430394211273533,
 0.19951840114873076,0.04971385779042814,0.6310364879307953,
 0.8729136161956293,0.07443232368425245,0.6928615472508374,0.8631868745463488,
 0.7117271046401733,0.8028497949474085,0.2665225816169454,0.5751981700801945,
 0.6411554400290727,0.41630124625450804,0.4297952614759645,0.8349135102584874,
 0.8782039666761343,0.416487475759185,0.6036200369361193,0.08727223132581818,
 0.19413035575315907,0.007313410009634924,0.04222666104840411,
 0.34959128665053285,0.079665287480844,0.25212171973593,0.5975038634738422,
 0.044572222727251365,0.8233121588800882,0.1515903230674286,
 0.3294152675545018,0.927399478941384,0.22314729132022904,0.7651189124078176,
 0.3115511539816558,0.005214724466123033,0.02283571330033296,
 0.8803761494113459,0.5256720013230402,0.5641306277043828,0.915275257267459,
 0.6482797424238467,0.8526851963982576,0.22841911310789653,
 0.09437102627851857,0.4058235796147486,0.22326816071691902,
 0.13731128561688577,0.7002596806539798,0.5551235858321989,
 0.13425546879570804,0.5597380537717472,0.9278896843375387,
 0.45341574467806367,0.7553224136350325,0.5575569524153469,
 0.15661960639725603,0.25887211668813126,0.5834063186273741,
 0.04255424093372007,0.9550013544341482,0.24868893621001786,
 0.6333859205789185,0.7596769935897769,0.29322584527922924,0.2222605632026411,
 0.9293210151528977,0.8528424726381721,0.7382728460286657,0.5526615892550285,
 0.20334892600743337,0.9449578902376436,0.9866282788604825,
 0.03413043169704166,0.030853270110531406,0.7459736863083796,
 0.030261297910377905,0.7492631316077109,0.2344685214730159,
 0.2685763543506472,0.3600199748944709,0.9526718050203619,0.06433255944605976,
 0.6722019674960819,0.4696849202423937,0.4618119139865535,0.7065913304505449,
 0.6171328964133425,0.7357608714116374,0.3395886682379565,0.8878936328914233,
 0.7056513019717141,0.6851754199153908,0.5114148380707104,0.6759793509919539,
 0.34189228532653604,0.38167286157470137,0.004147885407106866,
 0.17013379499901482,0.32912506369331196,0.15129822102850976,
 0.7392324468247613,0.06954884427473207,0.917818211296984,0.24107656410615652,
 0.41758411748638935,0.5626793142990593,0.77475703215308,0.4180256264276774,
 0.17630812737473356,0.7310180897230035,0.5322685012555662,
 0.08343370941081185,0.05893904267420702,0.9950005031291249,
 0.11899898880104653,0.7979937164505273,0.5757077978623195,
 0.45345814514569216,0.24132972889950033,0.14118170315949308,
 0.8165727387313212,0.5644736091610819,0.3665812081741845,0.31510576407150315,
 0.3343256361262237,0.058582666266935646,0.8135146512179438,
 0.3346890183646144,0.12491933347348827,0.07494854484409119,
 0.27574815229417227,0.3214994187668512,0.07052663917725932,
 0.7282993123281533,0.9633236744360737,0.6570808355647689,0.06116523188376677,
 0.6843203833863502,0.2543748803954834,0.23167123807337897,
 0.15028149484298248,0.32949627538720594,0.41278361994992685,
 0.7533644057131111,0.8648577962109357,0.8343189329044073,0.7215149705088492,
 0.09001201609992393,0.3240912285797498,0.2567301528028145,0.9004307751794309,
 0.27072460734165393,0.02801151079592601,0.3338648550332839,
 0.08061321517130537,0.312807714526261,0.260373357833225,0.5629679586199292,
 0.03427075474593,0.44779862573403384,0.2282542237851598,0.5086380972253874,
 0.06931205021325704,0.6102628533553603,0.2879165815210749,0.5762157346335336,
 0.23612198755250713,0.10628454335805837,0.6527337506722957,
 0.9852989323196215,0.22102353105118144,0.6318942300157718,0.77964679502325,
 0.523079338055439,0.8341093183786114,0.16224455051121844,0.8931868823805936,
 0.9903246358310915,0.8428886342038405,0.4818448858383786,0.8100932399018302,
 0.18740024711561443,0.31708965121676,0.8410817131768901,0.9224181397067027,
 0.9774537767405524,0.1860715881293784,0.9766640047698899,0.2427742281640204,
 0.5216676844749614,0.0747835875895585,0.7973004314486362,0.3616526986004508,
 0.48394910445690664,0.8758412031812834,0.806151487032879,0.25193262291520047,
 0.17472514756207858,0.967439369732596,0.541735366884387,0.3910570032443538,
 0.9474817394885827,0.5182127486690327,0.4171343051112484,0.9373481668516671,
 0.35939452672227745,0.677295581564168,0.043547329108815425,
 0.2570245072910309,0.47521231790211704,0.6484504125819391,
 0.33493915386538176,0.29406027030823856,0.645668738288085,0.8502878699020623,
 0.46579612425945516,0.38280813209298037,0.31952638261424415,
 0.7415684919085436,0.8628552733446856,0.8159986825802145,0.2052820177615612,
 0.7490199213947591,0.1801133681789906,0.7687564866555439,0.18683617925930573,
 0.9912398152019664,0.8722231696734734,0.11300876376024405,0.8231909674494995,
 0.33929688991201357,0.6037105479119934,0.7407936303758113,0.6013087328317358,
 0.13693925968898957,0.007974648677726792,0.5309721630573884,
 0.9372465237591843,0.4505205371602572,0.04671456527770379,
 0.21654771980651621,0.15160592685614815,0.08507129685918857,
 0.5860231575174234,0.6967473696377452,0.05660544466728612,0.861833229909906,
 0.6036720830596574,0.8280922465059176,0.3845971603242271,0.2500684292179782,
 0.4634111537069422,0.5452231198388332,0.6851559022666422,0.10065764812202227,
 0.7010979925031785,0.15555889594831118,0.427735807282389,0.22865273588100166,
 0.9869602385211415,0.898092977654559,0.30727710717563617,0.40514079774369594,
 0.6317157220357159,0.8823708807216135,0.5900991097074144,0.9529294820187412,
 0.28422829127955884,0.9458845246487203,0.9609365517911239,
 0.16636362983776276,0.9028609339373211,0.6743406159419398,0.1377626590358657,
 0.6750765862261051,0.5510891137678818,0.7208353878175718,0.7965672969044363,
 0.3460792291311484,0.09835516040790393,0.19298696576447083,
 0.8103234913877686,0.009809051709855132,0.9592695731126022,
 0.8226935178934092,0.4954200247756565,0.678997259842626,0.4258596627054366,
 0.8580602910612913,0.7831493131428567,0.31870502134518974,0.5024670940652324,
 0.5333011685583634,0.9645209762502939,0.3616330705393779,0.7824589961672821,
 0.8070924046773625,0.513645525787564,0.5059193282749381,0.21459719513342446,
 0.8556111278645769,0.705828844873164,0.3184299862802116,0.3014876535363221,
 0.4534027543338599,0.639998928480458,0.39785412950515464,0.18236845373566557,
 0.29521636727349154,0.01976746381219896,0.5670066442182532,
 0.7074249344151484,0.4547647913710383,0.7236727432502964,
 0.024712055666144206,0.3413891593887206,0.22306921472234587,
 0.6360961068352093,0.6092801884330304,0.19454982472649363,
 0.24293182080363962,0.834015112521659,0.7048943903494203,0.9085986141399025,
 0.6262132287491903,0.8587355157465031,0.8118687397840302,0.7341467499898793,
 0.8660026745481837,0.5983342285182173,0.8191485455161299,0.7493157691781557,
 0.9236625323111798,0.5878240536540275,0.28023903382668147,
 0.10842532592846255,0.7514598855753033,0.06460915793255662,
 0.24819502088514978,0.7180085109428501,0.8425566710238556,0.5662805977645367,
 0.695678949099391,0.24716779011878154,0.8780981579065087,0.7031090483250046,
 0.6577902676917182,0.8929454997705512,0.5095422546322305,0.509870072950971,
 0.1950057628237606,0.9990291639933966,0.16884600656312065,0.9297887384573296,
 0.6088807191128579,0.712337448475833,0.5342455577357513,0.8090369781211714,
 0.766065478669083,0.1334870506657011,0.37998762716546475,0.8941333782471473,
 0.29511481763660374,0.7307986750144024,0.3946675536736258,
 0.22423642734508076,0.11216162818340047,0.3714583574721241,
 0.41884568435080094,0.309357356644438,0.1755581090994296,0.7512792023785788,
 0.484204195314593,0.3502830413173019,0.11493598563175333,0.350197397158892,
 0.87897941348279,0.0604805178482124,0.7501853601502882,0.4615446470505574,
 0.15397915923215644,0.5861234149242212,0.5086000653364922,0.9207801616035131,
 0.9956761899812595,0.5252803890183071,0.8564594280131295,
 0.060644732848798566,0.3036808996787477,0.7188283732028553,
 0.5823299703551061,0.3044904374902727,0.1162423802712873,0.15782413455447808,
 0.7414999035587667,0.7656642044269129,0.5414153238543655,0.7301640832340981,
 0.49882934274253454,0.035352863133949475,0.8447510714092439,
 0.07382985321637914,0.14030032385671742,0.17080948084236947,
 0.3060078146263827,0.5808258281345192,0.7597140853790704,0.975082940642663,
 0.8695645586059743,0.09539027271015943,0.8697124952683487,0.67279127460554,
 0.9135604418599869,0.9588561103899421,0.24839911506359025,
 0.07368180206602692,0.9301364370473951,0.6382869620012748,0.8202106641709493,
 0.2825516459416223,0.43349084152519446,0.01992895766763625,
 0.09760160554610708,0.38593068945476394,0.5489689433975052,
 0.2564672409053571,0.8698923429881111,0.2194055387533531,
 0.002240427175361104,0.20295427194436033,0.8877530000825445,
 0.6538691468908567,0.014729634581785556,0.7208715088270773,
 0.28995887604705706,0.9064631069576927,0.3662277453471414,
 0.41073251127461197,0.5385141461680183,0.295923676148912,0.10395286993876451,
 0.7947374046508527,0.45074386130045374,0.6542450465049635,0.6342124641425197,
 0.5819413075051506,0.3142462624940985,0.1332804921720584,0.24855067368559314,
 0.4015438945122497,0.322690170740905,0.5024772389079598,0.49497026477085193,
 0.06128406379896223,0.36549090925130157,0.5587593354332171,
 0.021689562473031554,0.3511973820848109,0.9494234501266701,
 0.5253032470739729,0.8929748702622948,0.9133490073069981,0.3193335606466463,
 0.11502470707142953,0.02144948330110752,0.5247115480556461,
 0.9316441567278346,0.2576027130710591,0.5860748086424042,0.7048401050941828,
 0.15017779174736712,0.27594513624557426,0.24678946221507003,
 0.3374637091032946,0.8425087547723296,0.42518110756163563,0.7430293816634554,
 0.25918937767285866,0.647187015469749,0.7117277466025244,0.6653606400843953,
 0.6136536484537323,0.28283818224963286,0.7770786968880501,
 0.04485027339144143,0.044270179599113946,0.23126702899915286,
 0.5911924560623526,0.25966118416865314,0.9742064081805103,
 0.42654899304023075,0.8897751373784201,0.8774432853135027,0.9246615913950322,
 0.2604825543662165,0.1412604233710374,0.4149902282277269,0.3877080605210329,
 0.42263324910309785,0.11944806502044591,0.13442965991394018,
 0.06971161890925825,0.5659334964719289,0.21449897218267355,
 0.3598864620916761,0.9385953546157971,0.05238056730105545,
 0.28815278197461514,0.22738053204234898,0.5857409972357873,0.795383648234388,
 0.40005549463252743,0.74733922373577,0.3071045049492904,0.11725354738545257,
 0.735029595486185,0.9308222875294282,0.5599305536838443,0.7454212437548886,
 0.3836536619609431,0.48733327236700086,0.6128761457290686,0.6137816788859225,
 0.8867729868143694,0.6814660830100602,0.305412622727675,0.8787681738731381,
 0.12479527542149615,0.46164682233476195,0.9704956848833388,
 0.9631587682133669,0.04568747472338397,0.11327791087756112,
 0.7604064925627557,0.4838579212378795,0.7107590074947313,0.7575592467062098,
 0.37825629053538323,0.40227496977915833,0.7068766211762325,
 0.9588594964357037,0.2533315027026113,0.9791606171912237,0.3978967771271983,
 0.47033239291283635,0.6469402804191156,0.7632169774176859,0.4218802249104524,
 0.8444723661885623,0.6168691142251165,0.48045436615288484,
 0.32260046756467076,0.5834193392156017,0.4016244432405798,0.1782507584486419,
 0.60366644812714,0.20668141706719523,0.009955352811167884,0.8187616203992814,
 0.07610733848027784,0.026247969868581467,0.8935747067818569,
 0.43957133006131965,0.3584298197269318,0.853217285069389,0.7973236705202266,
 0.10918268627186944,0.014404900532418274,0.4751633500210294,
 0.4407642352461717,0.4821881186247492,0.8256731540873534,0.03740234134209963,
 0.3676085536224788,0.15634852933580157,0.1709680913753926,0.669385858108873,
 0.9734215130831149,0.20035009978251261,0.5644038041866083,
 0.38109726343259354,0.08630820359213986,0.19564914935063826,
 0.15436095250506954,0.2747047086292793,0.1442310090913741,0.6719201003491384,
 0.7024519528324928,0.20751684024746142,0.4000118941450208,0.9302889706386381,
 0.35381054196911366,0.5111935854417703,0.00973362850138848,
 0.7436945783540799,0.15548983108506342,0.7830332026361929,0.5905226689129763,
 0.8069299681402272,0.3682406380136064,0.6601446183269728,0.7046953558624937,
 0.17810854910731322,0.9492401621152682,0.8271380905374878,0.8831908748278673,
 0.15708137261539978,0.8101532819274369,0.9529387977372543,0.8869306599243597,
 0.7909956088954825,0.6303377705330357,0.7209588691268989,0.7368127063615298,
 0.9135762081303512,0.5611532948399122,0.9962635276625829,0.2811828192229965,
 0.9508774365191521,0.9766833671537019,0.2224089925896814,0.9129582336873526,
 0.49514483932182074,0.8821299267849726,0.839260899904302,0.02430959506096675,
 0.8756057423614488,0.09518298146896864,0.24470131506210535,
 0.2996451936532829,0.18975495674120535,0.4360801174708947,
 0.12170374999764377,0.1920998199695012,0.1831707765146542,0.7729367295687419,
 0.07335343536012051,0.45155017941271436,0.7754047517872162,
 0.20474383096818238,0.8217361470235329,0.5599974894762725,0.5363490195602978,
 0.7175437387658906,0.6456471471932972,0.5860218824571694,0.5548796054509908,
 0.37168358262249346,0.1247212094272443,0.7278984682227044,
 0.15574115996217353,0.07921886739429673,0.94051453941089,0.20732260370961697,
 0.21910460812733334,0.7719033066769199,0.14396282397374627,
 0.8143360869611378,0.6170430072447757,0.10105547557415984,0.5120808963087021,
 0.1412647145687752,0.6993330280260381,0.6859789108195974,0.8342244323605106,
 0.4063885733784318,0.6712345537567512,0.07523795931246258,0.84748004018899,
 0.2601929824394078,0.28974982872872634,0.24451351946388156,
 0.38498871875598917,0.9620638009567958,0.09415942625105977,
 0.2258959359244581,0.7785827803992932,0.3239517518738717,0.4371996683446606,
 0.09063776934353762,0.5418990780433574,0.2344349881494724,0.5985316802229235,
 0.6234380170174829,0.34168504185587534,0.09384592797593505,
 0.8364672719033738,0.3940042386364022,0.040112677684883025,
 0.30596213100194614,0.34110509613952733,0.7921470010907369,
 0.5725150825251427,0.33794448852255277,0.7737797447699857,0.6228058413048371,
 0.41680365729796187,0.3591206188356172,0.14349819781406215,
 0.12866127128315052,0.12436809662731407,0.8623688055441464,
 0.2189213569722116,0.8467926561239972,0.6870339434226198,0.9817696675575563,
 0.1657455972214047,0.4212484855580413,0.02170570559304008,
 0.06481636700099036,0.1529426452103828,0.052162517639032346,
 0.3887939568587173,0.23312740915401697,0.9061202810838201,
 0.015594136758264598,0.4650483586270404,0.14752917442228553,
 0.3864352869678799,0.212855645401965,0.6483234029853515,0.3006504162488197,
 0.0940075506460567,0.6202630904992649,0.4778893621226934,0.5481197718401564,
 0.41200162755908687,0.961710412812691,0.19395937338637437,0.5417473238556996,
 0.3246430219530567,0.43798188177918873,0.989416756379441,0.34223712274389406,
 0.8562026812725108,0.24857162343113914,0.38330471102578634,
 0.5041914608166138,0.2196706652631809,0.32635916036056745,
 0.45171518984759385,0.86183745776018,0.5890883353763288,0.8524158412729024,
 0.5483424312580261,0.38959797413094743,0.24457465503930498,
 0.8495263809079776,0.09568592019524047,0.887323540678142,0.6154234177175683,
 0.05653498264033785,0.5124814862210876,0.969293557224004,0.0663189536727734,
 0.5406884083808952,0.7996831345644944,0.07016540244956271,0.0810416881424545,
 0.289512482893561,0.4300287851573724,0.6029340582465397,0.14075962334995484,
 0.7327847400614385,0.8678809712901395,0.7009287780812585,0.24939988901491594,
 0.8497998558760054,0.5027565857559562,0.23195295874515365,
 0.39256910531429356,0.9807923174846475,0.7186887364354486,
 0.26084168255096574,0.6519401201922972,0.46380691557288445,
 0.8316864429551064,0.8812981768910902,0.7418183498956141,0.6406752284061246,
 0.7320381291305417,0.875206947815432,0.8743051632791259,0.9928491330704181,
 0.7303545412294828,0.9787135208637051,0.6486945931548278,0.13016348068504746,
 0.19539294694195564,0.3298812720553497,0.03091010717346765,
 0.15772613233704103,0.08340217589500709,0.7630839277047059,0.797667281800692,
 0.35189866858238705,0.31694459752322157,0.1339741159336173,0.954084483197637,
 0.962411670990355,0.011717540764868528,0.8484792604281693,0.9402535052857192,
 0.3366920056329451,0.4536043400410049,0.1597812264870735,0.537038728913514,
 0.5835873754226067,0.5317458142533018,0.21010446150064865,0.641006618427968,
 0.9770126139458492,0.3131001786153088,0.9715666168610118,0.3683552336696505,
 0.571835563175223,0.2911313501887902,0.5010754442655111,0.9542130427371269,
 0.1587199983062758,0.10112013073007109,0.07922617696314072,
 0.9916682027785804,0.01188695307349219,0.3756437983950456,0.6974306646434286,
 0.6943913547460843,0.07823167999015579,0.5218496702115613,0.7396120534668685,
 0.5987872920088471,0.8611310397398237,0.7103944201729598,0.28121843834605964,
 0.785754617452259,0.50068829398895,0.29291114485091796,0.43100788229735665,
 0.9978716141237961,0.2379785861956062,0.8539595454024055,0.40340378694805146,
 0.10317026755249126,0.3606969992627944,0.5584816398040671,0.8824281233231647,
 0.8245598276107278,0.737412346690641,0.8124336769481486,0.35069625082544964,
 0.7273728023705254,0.7631436232171883,0.23670965412830558,0.7798135603703386,
 0.6571583641083193,0.13447290618669827,0.20187538406529226,
 0.4042392597288239,0.16381513078080112,0.20166176458596274,
 0.5673132833145398,0.567044645430731,0.7656891290020825,0.39629205644383225,
 0.8465442055375717,0.9362472208470968,0.16997876366885234,
 0.27086403989680086,0.7689398172336024,0.7043087023393114,
 0.08940587461433591,0.7272296232943231,0.23884748673103473,
 0.09916786619930429,0.5938616453316078,0.4356467852848447,0.8756520813329468,
 0.6257150542632532,0.2636352500395548,0.7877877707727101,0.41143317895513376,
 0.14874885409708294,0.11336123608550852,0.8306787765636021,
 0.9894173741824177,0.13272018634957583,0.6080697251293472,0.6544115573254065,
 0.7485501780954598,0.30894533231170274,0.1057653691605992,
 0.010559843962211102,0.4505994426764648,0.035185212711069,0.4411621254937024,
 0.09910633218697917,0.6421611620692939,0.967421749488124,0.19464027163205677,
 0.9189034856766873,0.24711061702881865,0.25197031232320444,
 0.8907465173731136,0.4370984201394156,0.5017784257060277,0.03304414181944759,
 0.20659622097733532,0.7842134081548032,0.5749284310710083,
 0.31447096239789385,0.2952753574418464,0.9856560371176503,
 0.47103185851541873,0.7440402216403234,0.03241848526197755,0.967035585997785,
 0.6300088254512813,0.06582010618166001,0.4601443401790498,
 0.46015867096169427,0.9916673387919945,0.9678419130900077,0.7550898517374937,
 0.2964888798498153,0.2894468551399817,0.9536613890219904,0.6473856360083483,
 0.44986158887113326,0.19533075789991672,0.22325377311810335,
 0.1489785855805199,0.09406823902767536,0.037496705249986384,
 0.17356746205338558,0.9358970742235019,0.058794207453493064,
 0.47013074956153456,0.3745037914825098,0.07691899642983868,0.040377288682658,
 0.6587528898782701,0.8959759487084537,0.059964719085604745,
 0.08757514519121157,0.7781262893402995,0.35116505204827086,
 0.09597977653028422,0.9909916951409056,0.5144852575819208,0.8586408545454928,
 0.5201804197981053,0.4758576935836333,0.4184088819387587,0.7335440138694004,
 0.2977789113077687,0.16589986739522322,0.28103398340135965,
 0.5444655955457965,0.7298759337295281,0.9452919314318123,0.9799985141133134,
 0.0631261641447527,0.3782642093634727,0.7036764322724125,0.6580164631643072,
 0.8851384027161789,0.25307175333673515,0.46897033788562037,
 0.30701004777388796,0.10452511841651857,0.32451846466578893,
 0.8418677707926926,0.6648266246554141,0.14530877381850738,
 0.20998769243685456,0.9518263463909664,0.1183720540258566,0.8515079545635873,
 0.8424124124774155,0.7519009933620319,0.5069632910472756,0.9925765382109932,
 0.7814625063875467,0.4128376841033512,0.3257892702894777,0.2860063955315707,
 0.9371637555357187,0.48440881183612283,0.6815579381662086,0.8572465049734588,
 0.5251517267761289,0.4542998066215008,0.49077534701750214,0.4741752076457738,
 0.4688852743454084,0.5303514470311536,0.9577915467751725,0.490533069945408,
 0.9428289347976277,0.9370866090096794,0.18262764356959027,0.6139021049626395,
 0.5860876802298385,0.9060149430904334,0.9303450459870897,0.2580113463693937,
 0.8207247334882594,0.308655237325273,0.13103643565042666,0.18986081514833664,
 0.6674167154263085,0.9957848500619062,0.3586826385473434,0.9020367770252016,
 0.9803820202385825,0.7352486317248296,0.42983929312140534,0.6722375464300685,
 0.5907891973650737,0.5422396138105783,0.5389945833018981,0.39665400519126615,
 0.15161366802540843,0.20985283170170876,0.33590096116127355,
 0.3049484425411815,0.44160939478059646,0.2861547874111734,0.922121954959944,
 0.3752123550135046,0.6074845706200699,0.47733418016292894,0.6827716770519383,
 0.18198967707207536,0.14987261535945218,0.9390750376564128,
 0.39373922487698354,0.6994013162785193,0.9734741562870705,
 0.48348236909130216,0.7301755162621743,0.29807100414154464,
 0.22240823470651827,0.8794399843992229,0.7964961283503427,0.2785770831399712,
 0.7776586632324407,0.29786765487868894,0.011091851005753717,
 0.5378187671432214,0.17476784189493721,0.2769677003166001,
 0.33682870025018996,0.10181217514879048,0.11991803678317536,
 0.7575040592265236,0.8183307063836592,0.13225821374525148,
 0.12197591214059211,0.6564464685951122,0.20811651434093725,
 0.07213931205441748,0.4490374492278303,0.5554678433084544,
 0.030289482042572424,0.9689033705414529,0.5584089967697774,
 0.8930517759071284,0.053958319432394486,0.8719023727772631,
 0.6130565736617458,0.3079775224054402,0.5132088309591798,0.531703122570477,
 0.8992766680775501,0.2584592686811449,0.9456241193078883,
 0.008297176238423187,0.4079155698792445,0.768661605899453,0.2198712406908564,
 0.7232705136817676,0.9847740464273302,0.8624910864197273,0.40215442531540113,
 0.9994616226155616,0.6727975351578572,0.3246533065125443,0.6851797192223467,
 0.8204856751415937,0.8108486636859373,0.5751466585695809,0.26639372731587163,
 0.19369662110767738,0.5045012705622508,0.2563500891165128,0.4280223976315991,
 0.9059750812368925,0.06629081412470006,0.8106769779463991,0.3272916082525368,
 0.23887175200107358,0.8957015881933599,0.9517935632954526,0.8654221922011021,
 0.18126455939400243,0.03438922218684859,0.931753816361113,
 0.37855702569806504,0.06397392929671397,0.5616840001219606,
 0.8954046828701007,0.43124761664555855,0.22115714883263493,
 0.20036924476118312,0.4790889531160557,0.6599576724947183,0.7079878522942489,
 0.9439335578244609,0.6845494866286168,0.7528342519304609,0.5301698186597164,
 0.2604541319424373,0.3408693534785471,0.6089789115158776,0.8095434639082046,
 0.26102768754284233,0.2651281513298136,0.7206721937885125,0.6354128686694934,
 0.6652540097940435,0.4376573240397069,0.528433828968767,0.6039970395059312,
 0.519712503402926,0.2421480018412655,0.023938918487020233,
 0.10012703747626917,0.8600826515344905,0.002395227293841895,
 0.683225849339182,0.6002541784647836,0.5330765194847267,0.021134090908084602,
 0.6429931840633618,0.718800072902755,0.9616198890201063,0.6806833259486771,
 0.9921851165633729,0.8154512296210966,0.25284090367652134,0.4713293946215187,
 0.11309207405058663,0.3846052905333057,0.8630037126402177,0.868276626382762,
 0.03765732901978336,0.5789760880741643,0.6629493642592024,0.5599109832689996,
 0.38499207937746005,0.9619638879676476,0.009647724238658006,
 0.2553013479784072,0.701046641855211,0.8183541871764638,0.3721648105775819,
 0.3018707682387691,0.6350552128615276,0.5211749697487412,0.5928204188314726,
 0.1198751496095426,0.5953160494470403,0.8268041447792073,0.3639653233751551,
 0.5840850956365943,0.3100651967857646,0.24804250688544704,0.9144531623074807,
 0.37876410949879113,0.46532436953016576,0.40676311255628095,
 0.8257634339343178,0.22811172950142022,0.8094537022832924,
 0.16005647447270044,0.4243536882243366,0.6227824836376173,
 0.048117710548599923,0.6273965214178364,0.5223701457280603,
 0.5134267248313813,0.8563038220845078,0.1859153067757734,0.38675164273750917,
 0.8352127593623446,0.7586238074883539,0.9541513695649956,0.16778108945904024,
 0.2558655825429249,0.09628169014718879,0.9824289864899746,
 0.37295573586227193,0.5237351286747266,0.5914541381522495,0.1371252680212014,
 0.6030648499663925,0.13742156862075705,0.0316799566860656,
 0.015476156608045355,0.9248542041890686,0.14999769905340932,
 0.10344919848309075,0.6798505099559624,0.6414578902988902,0.9744069500746703,
 0.9786060209890168,0.5696556490168392,0.3848277112424551,0.8211070680036886,
 0.4251959797599433,0.3619938085323824,0.3341791921754803,0.07628025011613448,
 0.8930722759374312,0.922720198307134,0.0237343107308845,0.37263147759930904,
 0.9800036742679316,0.3077631669653186,0.2269361130219928,0.981700714596428,
 0.11184123268924329,0.6446860875318846,0.062255921507824996,
 0.8947850894008943,0.30337739652927354,0.34557638832184,0.149850516194993,
 0.8125892007735326,0.09498286102595166,0.5234601995326167,0.1314592904481755,
 0.33089511330570154,0.3735420361413292,0.06869769855764107,
 0.4958868570866245,0.659409655352266,0.11031811861860152,0.5781910016053379,
 0.6504162673191918,0.13575011993542108,0.5367902435604766,0.9983824850641697,
 0.6881035755972035,0.6268599073391725,0.47413422532997085,0.8835622580882442,
 0.5801015454967554,0.04860676594799418,0.9829427887749576,
 0.27879754033351456,0.9905501226823867,0.8043987297945286,
 0.37630375200938815,0.5235257125527802,0.8591060023042791,0.7811243535476706,
 0.9303424146823954,0.6634412998161718,0.6020017300407038,0.22264801163789016,
 0.1998827650578674,0.36110890782349847,0.6241112863985493,
 0.15319928292736829,0.6583662730046588,0.5463570165648806,0.29502883073161,
 0.920718635423479,0.9759848573611827,0.4189848403690316,0.6251162285839782,
 0.8426267880188589,0.48244724951816065,0.029726101835703234,
 0.05105841197429395,0.8387737253144874,0.9426663684610451,0.94229975924963,
 0.8003829083295495,0.7368476802108858,0.03641673443391347,0.672735964652556,
 0.9205698342592851,0.0900241032973309,0.9906581579969943,0.19989815928822696,
 0.7777587404092414,0.3586608257865055,0.6188455128123175,0.3964430191710214,
 0.9598003103512293,0.938491610513091,0.5421296445386778,0.7840227760367091,
 0.9757082805702851,0.18276565823679436,0.3139456821066726,0.6540251740073774,
 0.5337334930894776,0.3756270904446073,0.025230575196140914,
 0.8060081968592053,0.7749575556890487,0.8107552377166669,0.1708590709362927,
 0.5825029332620768,0.8642939805410642,0.8002860619406202,0.8665806690444727,
 0.49827975591403384,0.2989377010523483,0.34141671216044034,
 0.5577198597041642,0.5450695504415987,0.7277913328628758,0.7294750847472147,
 0.8606759355581676,0.4223070423079417,0.44137905080536843,0.5792520940936854,
 0.5225157831205963,0.8187944696433468,0.30459094070824844,0.5718012631007466,
 0.3533025872913952,0.9791869430598871,0.7088128721838385,0.8295224729432606,
 0.2973587896909833,0.2017106095757769,0.030155057560348197,
 0.41723935719994054,0.6269745313311068,0.7234658133479839,
 0.14501351807994567,0.8816472426243335,0.7159238848439933,0.4237670477160287,
 0.30594816464059094,0.028584955037256332,0.9283576110839823,
 0.9763238572961629,0.3066392528843056,0.540816079490557,0.890502842159063,
 0.0914821361807312,0.7548304681068023,0.056434029624702564,
 0.14582700301879947,0.41747365640881307,0.9610484213392277,
 0.6370533651834013,0.6712254708084134,0.011138135691708406,
 0.07097778740403526,0.5877396240795235,0.517999061571825,0.34716192731813567,
 0.6022901678719317,0.3572287266671288,0.44965949973788844,
 0.004500006983238269,0.5315438360040666,0.011275066244931575,
 0.13265013662482406,0.9072118342641639,0.3818501238604115,0.9269826357330477,
 0.9591699906974606,0.8070270744557579,0.8576456777460851,0.20445666388118644,
 0.40893627971044644,0.4174511106235387,0.857142740995494,0.7973002029319387,
 0.5421120450087409,0.10919969094928328,0.01481453450542658,0.8793845441525,
 0.7256255769714717,0.9338523659982237,0.19586227059068095,0.9299393978740045,
 0.46779209117626774,0.38525707161566947,0.5343814518075357,
 0.9650034309573782,0.15821502533132903,0.5113435771409087,0.4591784972312454,
 0.6601695157825358,0.04356834217357264,0.2856944230800148,
 0.24045482278702135,0.8154920321809279,0.17497693715880525,
 0.2900826820060599,0.3434113437170051,0.8280311772316076,0.7517705373987531,
 0.11080934713755741,0.4536228126237136,0.5744453272407428,
 0.13067107537933187,0.7914582234212227,0.6151966631599942,
 0.005078245435690132,0.8885543936999997,0.06244446262483094,
 0.5371637301899566,0.48222441080052325,0.2778969947065725,0.8933393915486406,
 0.1570432809345521,0.33278241667389397,0.5401077106967489,0.8936919689217107,
 0.13511217199131476,0.6512080962371191,0.8748839284153294,0.8398019230853255,
 0.156903117487608,0.367777300461296,0.045868924437594494,0.4654668843275689,
 0.8806212555190762,0.3618484760501973,0.5043451063613568,0.957423941815748,
 0.33666987796182146,0.40586389201951967,0.1576282895464749,
 0.19446620597699227,0.901166170149794,0.044940104050332,0.3876081039416315,
 0.13876425388366997,0.43209810193653675,0.5768210356487644,
 0.36791889685371815,0.8408387799483149,0.2776807535757071,0.790144038175639,
 0.6795865671676015,0.9221595585609205,0.6979097693903851,0.3531504239497455,
 0.37289453692344576,0.9477364225426581,0.07529906642621464,
 0.22505893494089935,0.1570069389365728,0.5352056393618392,0.5374837640011192,
 0.593176834544471,0.7640129317162677,0.2217584167936676,0.2787854970677819,
 0.8881093652618134,0.40881018684074144,0.7392280887566702,0.5485843241445162,
 0.4154948017588672,0.6092900140421814,0.025097973690726083,
 0.5950057129767338,0.17847565329552573,0.1291315199164199,0.6915955685339387,
 0.8553103443444943,0.9343555362876037,0.36666846237487205,
 0.43392833614495163,0.6966378519774318,0.9803810223058687,0.9909664492750772,
 0.1492085299353214,0.690088877963041,0.1985065198153907,0.5489155185349025,
 0.8347187311579018,0.2512094121240189,0.0517228792663561,0.12286019535428261,
 0.5997900614760301,0.33337406842962414,0.6590146672771304,0.8702727857549676,
 0.08179512951960344,0.06118747530571422,0.771502490889059,0.8653651796341545,
 0.5784781725143779,0.5840423872086885,0.473820949122442,0.6131462925495661,
 0.011783308902369005,0.5551219839555177,0.08138598336639802,
 0.10674573339508442,0.0565036606845567,0.7828217574250396,0.8177500356090627,
 0.3115876993462867,0.2858422059876232,0.8340676781909874,0.677079855857255,
 0.41839427466065837,0.5405009739020216,0.46173997559782354,
 0.20607475952324972,0.29954597224730684,0.23774903591456542,
 0.915541367989977,0.3428173607295729,0.6526641822109366,0.3483868298752756,
 0.20668974531880835,0.46731590853553706,0.8250469782398445,
 0.7073416576643851,0.4492632343804577,0.5382496132744106,0.8115425853157487,
 0.7641060951639914,0.2553552617945012,0.5284267699416969,0.0768192430312008,
 0.2854635532785432,0.47313649335190733,0.7900343150139376,
 0.06508194226312702,0.6328742554943506,0.0917219927325128,0.6521823651002608,
 0.9756458240778522,0.6290788393556839,0.07155788734909596,0.8370472311028546,
 0.7749838364725945,0.9806713774457416,0.9282127869024386,0.3311312357494287,
 0.9172522660116209,0.963731682025387,0.05187980183950036,0.23208399630053245,
 0.687936621810281,0.24598907765735734,0.18096676134753187,0.1896101832605851,
 0.7070365756553648,0.8179214304085863,0.4450245233967145,
 0.010292260169378231,0.5367731164713222,0.15853905148983138,
 0.6568008353614698,0.7179603075511205,0.822448749224722,0.7880607925962932,
 0.22533534752263706,0.2216069441684222,0.05757671997082181,
 0.7941749958936132,0.2804428809431835,0.19184631880933156,0.9769908287206394,
 0.4929009169545103,0.42163325957524467,0.9700708829477938,0.3648590843115782,
 0.26946828725001903,0.7058694550567713,0.28460904558520006,
 0.06281072604038229,0.4824971237976178,0.6918607051560092,
 0.01761058370408297,0.4233337933757718,0.6712128036779336,
 0.37620266480196984,0.13169180787019297,0.05118168267719636,
 0.6877233715517364,0.7734706610470168,0.2976807015494316,0.4132657744996784,
 0.5663323518545326,0.6228974310014455,0.9451366731910291,0.8723373300909718,
 0.9043006228351044,0.6550549521664577,0.5380237532296236,0.8270148739679025,
 0.5600571084744701,0.5585270993707647,0.9235449513029377,0.08904583475877548,
 0.6404694405875626,0.6029470984024771,0.6184838356251657,0.396089725482454,
 0.744827019147817,0.7230651232898503,0.42608967703779466,0.12647384037921072,
 0.7971291542284921,0.2475697648766455,0.07603614243063161,0.8632327294011092,
 0.7582758451518002,0.41647747757636444,0.8447979356095111,
 0.09142392235025376,0.20100677919435916,0.028602960752069806,
 0.7120574530739501,0.6592690128091352,0.8262201042761507,0.11910722763576331,
 0.6440193008468128,0.737974281222535,0.34975883785125417,0.22958385825213679,
 0.9546066025571053,0.0025359982843266016,0.8096045868999237,
 0.20119006132085704,0.6601372966833368,0.6328003406892886,0.5026402384157657,
 0.8346371276452629,0.7236922133454726,0.6831174766402947,0.23650061616454843,
 0.9417656374145938,0.444073044358953,0.48236883217816107,0.80664743028941,
 0.06188370891763362,0.10249176863660325,0.9804481054501313,0.393439506978707,
 0.5645611270157865,0.6421615087070611,0.722541553881821,0.7089621696802378,
 0.0024467582376906938,0.016558486242325254,0.2037075701504013,
 0.8031789968980078,0.03971114606403192,0.7445367976399895,
 0.20603388516793097,0.247652745893991,0.06423680878502525,0.707182093229848,
 0.3208860086159452,0.05279223403835287,0.045300987386215685,0.58715859305814,
 0.19592858705869232,0.07326242807516337,0.9458658555551752,
 0.5084474695699022,0.8733047200280859,0.0629400354100208,0.470431364492258,
 0.25422088033760215,0.8803735203114221,0.4943341600483073,
 0.07514202668170189,0.3463074171638334,0.7288706938395585,0.4133370878362599,
 0.029214778520679552,0.8887611700130601,0.6858110218139565,
 0.9012509390468934,0.04973319022833911,0.47756263775576857,
 0.6134737701660354,0.5725038365715157,0.15117435710605276,
 0.29986310728689103,0.039416448847556795,0.09043031236169108,
 0.14254023776668967,0.1802705994079985,0.9474352343812159,0.6880753502597333,
 0.2471213553516819,0.9267964088221514,0.7469975486703643,0.12839193633417956,
 0.7919550030240317,0.9926754703259622,0.35555553343469615,0.2917295714539112,
 0.7440639160146492,0.17797711511445047,0.4610713176733391,0.6483843290969474,
 0.481028214377375,0.39780785085741066,0.0034579975205193403,
 0.45701665778102485,0.9445474659922675,0.16268376530044826,
 0.2346821828133766,0.2254926269945603,0.7551414186674746,0.5317045746542094,
 0.19291199011741988,0.960254941506975,0.09452140735254377,0.9984946112800545,
 0.3543501261233525,0.11353404342171547,0.5184565396447933,0.8222736863437636,
 0.7056008637015239,0.2605117421171175,0.6460434037978724,0.8882356825942802,
 0.8689666675744697,0.6773483565286452,0.6687036188719891,0.9195656604269222,
 0.08063356099031738,0.007601458709594389,0.496609254867759,
 0.06175028227644941,0.012168998960340738,0.19696717851929146,
 0.9446633741305176,0.4414395764985325,0.6924746676886708,0.515611942500059,
 0.19894561911763786,0.7344526806537001,0.9730455497798147,0.3051141599360909,
 0.08827881002520599,0.7776711727766993,0.2066061467930227,0.534493979592921,
 0.898412776157569,0.049720164325498084,0.09980295446973575,
 0.6481730074195504,0.3637086062677968,0.024942663330120296,
 0.8479248492366438,0.0515073418740416,0.17400549184030234,0.3123049486252598,
 0.08757767205210532,0.930444777663954,0.23463757486279058,
 0.07830484291074835,0.50118817399037,0.15256170527827462,0.3401373004979493,
 0.04830298725464932,0.2259926674379399,0.5096977004035548,0.7233179597696733,
 0.5186531031331516,0.11731498579030986,0.30186860675238947,
 0.2977976952608914,0.3489369259110776,0.8413214332833691,0.3188818446582906,
 0.14089334042939106,0.7895979560095094,0.2789422849599118,0.8808160896842177,
 0.503656607505517,0.2724060916550507,0.0028230240763784664,
 0.18813876195738566,0.5449971144163761,0.3205135380651659,0.8563286750138396,
 0.09957966069923996,0.9415025962924727,0.807712742308977,0.0190794417840785,
 0.10798194113453341,0.8868462988703587,0.0033579081815772183,
 0.05790379682098401,0.15788412991683942,0.9487429791341679,
 0.7335586372365649,0.0683498547675121,0.3432237178898311,0.3518744799379996,
 0.4837204614036712,0.0655373132189383,0.7353129835321836,0.9989270276235597,
 0.510959945393898,0.9150394973665367,0.09467262268401089,0.4084611779132752,
 0.12341898942071738,0.3368469146185268,0.9917510796147103,0.5733659789748633,
 0.5448464577166775,0.9111922089589845,0.6844443646462689,0.48468192842513225,
 0.7722834640284756,0.6631611441063032,0.43449245567648553,0.7123948934756682,
 0.05180006599671472,0.02958800785884319,0.6896811785191618,0.329407456244757,
 0.9719096516671674,0.6592198833222316,0.36672968446338416,
 0.16234273206581196,0.26196698475534275,0.3340721117846909,
 0.22650626548233332,0.48186170517503235,0.5240268869284541,0.531478361826035,
 0.009753916921571348,0.26942086664508447,0.9763854917570495,
 0.6884859318220763,0.6061662331029994,0.2341535884089554,0.2590158479626461,
 0.980140345292465,0.19311465340372402,0.023496370259284194,0.790089317981233,
 0.16137898514336135,0.2974667245915774,0.026929195820320162,
 0.8967035902855439,0.8839797087128457,0.19542244717875668,0.3157137448880909,
 0.288470753099337,0.6381112668565512,0.6235788563332246,0.6139436868947672,
 0.31032237296493603,0.2905436504463559,0.5233152978467633,0.40033627516212,
 0.3687866824734053,0.5695278480463047,0.7299629070288061,0.29467903626868663,
 0.778268502313397,0.2997646043416238,0.3836108971362394,0.48029733063076596,
 0.5151202115694986,0.9473997069114944,0.4739035096449262,0.12667547952200475,
 0.48869076879355333,0.14037942141617166,0.7892034798930863,
 0.7340029287898926,0.3714139774438808,0.6898635020920922,0.8847802752590199,
 0.3911564477438808,0.7430973570111149,0.08497670506865695,0.7717484584927966,
 0.6799407866123557,0.24270484884370025,0.6648034021823328,0.9910076033951405,
 0.9390961499003957,0.7549091598299447,0.9438003729671622,0.4730998023963924,
 0.26533996931943715,0.5069573042258774,0.22181202336231443,
 0.27612808183763315,0.7119815692898321,0.301873675313282,0.3806860080936739,
 0.11123973472063342,0.667427208123706,0.6116070040387309,0.2097385572173751,
 0.0858826713051567,0.8044983423500334,0.5862214994761344,0.8776521447261656,
 0.6124573486269538,0.7195035944545152,0.7722249616024833,0.7491757944489166,
 0.49579875609706603,0.2628005098525976,0.36481253382472867,
 0.37094174893063026,0.8301675181650114,0.9203988051947087,0.8701061370595037,
 0.8250537992870652,0.8934262255725394,0.38318375944496075,0.5489052073197473,
 0.1241066118394607,0.1873266116595892,0.4695977043764701,0.4387922361779527,
 0.8021816183601435,0.07359796806245611,0.6456110298197824,0.9767035636077043,
 0.566194593598704,0.09514411366737618,0.9279831639787566,0.16446573797866681,
 0.3766575180644075,0.3504006228047174,0.9310249615232224,0.1960066633434704,
 0.3152323195909803,0.7925627829752973,0.5025397841545622,0.5043447223942674,
 0.13983025594311493,0.09247057297585048,0.7677212019604445,0.765045941880906,
 0.5975139125937377,0.32870352930815494,0.8491094536322354,
 0.10623124865329725,0.6769664146707808,0.702551929863845,0.9264631538028739,
 0.7684000251153479,0.2451399390377299,0.7867708730121247,0.47519191077821166,
 0.3271414742102916,0.23953358319046059,0.4717647430655748,
 0.12522959116046706,0.7317408865677576,0.48721862545095507,
 0.07373973835929848,0.19196536065713554,0.6241223986548524,
 0.16948754007286948,0.5297320866917522,0.33579763399715534,
 0.43664067526965233,0.33565424107711417,0.7446907042722364,
 0.03888138541962871,0.8233891162864471,0.9697105520161808,0.8101039089055058,
 0.6909213041900139,0.34458014591456343,0.3621546837067342,0.5304123366243488,
 0.20631446434797096,0.25273528172187243,0.8587227107476685,
 0.41339792615414817,0.7488425885498848,0.8720559320606445,
 0.02631050990290329,0.9783281238444411,0.4692986160489132,0.6324285368867048,
 0.24312618506355477,0.20194282229273053,0.15042943236217987,
 0.7874714253544184,0.6082757243898148,0.26923400116911045,0.5678557309475607,
 0.6133019854863987,0.25056439369505834,0.3822021381675279,
 0.33667824634569243,0.31978414652925913,0.08068177845399394,
 0.09617267924295336,0.3409929784416512,0.06547075520949996,
 0.7069124396382782,0.3207088634565082,0.8996948642393919,0.5035033046953015,
 0.17849650698556607,0.05733994505891027,0.8099159814093085,
 0.5859031267937831,0.9057308172796984,0.4907490309776431,0.6562571865429878,
 0.2531771601926328,0.25245834425910296,0.7624736907681323,
 0.38271157606604356,0.38182899962342687,0.37450888247714476,
 0.5611065182549745,0.2112980473450199,0.56697451324296,0.21563780482690975,
 0.9234761521849243,0.7113954980457069,0.12242877536110752,0.5551400055940159,
 0.6230316047194953,0.42453332573303293,0.35262743609474223,
 0.9933833491556971,0.7841705016595462,0.9475003880535496,0.8384519492332051,
 0.7782882433254915,0.9585894327803612,0.7995319550987235,0.12071209843308028,
 0.0970424110600332,0.8053726734260911,0.09504362972460856,
 0.33868610412111644,0.385685069217259,0.9772234890390177,0.3899807478035908,
 0.6692591574084132,0.9362684512108734,0.4493725440266745,0.664403067639169,
 0.950289189605672,0.22190295436343166,0.07871202856436466,0.5787577407356757,
 0.5741149899308362,0.8352109538059815,0.43715336588738807,0.2125131429756899,
 0.3750833504556653,0.5337222925300396,0.7052688786149303,0.9356632289574978,
 0.6884817501511246,0.2119801261876939,0.09645375004956003,
 0.14512048051662285,0.39817128297109894,0.09259309683581529,
 0.6496256843430448,0.3446723339414546,0.06783744687547266,
 0.059716254194500884,0.6898174962281438,0.09975815291238588,
 0.025836939694805827,0.8553510738318804,0.009958145976679655,
 0.6895125730830912,0.1135942699183542,0.9403942969414228,0.13056303353207033,
 0.7459064649148847,0.7529382305531979,0.44821028441738164,
 0.23462814967995627,0.3421648399473771,0.469187462857968,0.5941259208276419,
 0.12497690492259872,0.6288534811538498,0.5298360933907336,0.780185126551231,
 0.8652213012289601,0.25803398958619106,0.06839738190418032,
 0.33136452926145366,0.7001660516628649,0.792549800368186,0.06542545604560579,
 0.7799530944317198,0.198454770673258,0.6965783411405467,0.05516804437440792,
 0.026145053834275123,0.8294944679750653,0.6149774116801077,
 0.27814893174099453,0.21660814578872678,0.8664799439236337,
 0.18388423709569324,0.7941918119748747,0.1421812946917682,
 0.044919851514139686,0.6039059577614703,0.5067153523734982,
 0.07814376573063975,0.292300329623708,0.7624783856629638,0.08610261780770978,
 0.9574699455298824,0.2592976552102302,0.6771627426977936,0.8003862866166623,
 0.4125014713953109,0.9453083761834065,0.5364033108111349,0.6337333569670303,
 0.7167366908287879,0.5743543791142208,0.07448423312739005,0.6972636031038543,
 0.9177694696264334,0.894547454163153,0.2570894952843612,0.3344022411422203,
 0.05174335668777563,0.858315358112346,0.6695358752820021,0.8720942896386418,
 0.20689848458256077,0.835364693748581,0.3393137069242188,0.5339739402089644,
 0.9016454339318889,0.6781507888873877,0.7056014741991655,0.8901314910177627,
 0.5048540782715368,0.38746532792384425,0.9420092389930053,0.5296115943035536,
 0.7995865735433381,0.9258616318723938,0.9501700358805324,0.8636736513290352,
 0.8236002602255383,0.9540694860279924,0.06574296816908598,
 0.08080696457587377,0.09408717753106532,0.016470314287536536,
 0.2766399650521769,0.8565174167107762,0.055137070640835395,
 0.09728030470325844,0.10394249014864543,0.6815498501095678,
 0.9927293209286973,0.07297736425510937,0.06230949750310377,0.925610096251534,
 0.1551869273764933,0.5490699022313484,0.4763398703297691,0.6893781110995687,
 0.4921809087159996,0.5529442791875029,0.2599818651291175,0.8004586354963273,
 0.5374128947781679,0.2324410119836613,0.5440116113596973,0.5691732335184145,
 0.02110756484318066,0.8188546175812381,0.3434559589421803,0.0873431563540108,
 0.587173118999416,0.7387234448896223,0.8168418098857316,0.4269645071855893,
 0.5515027457629612,0.40137350493660373,0.31494568837795367,
 0.8771531647597409,0.3879550263500433,0.10995339032876494,
 0.030163947837692273,0.3269787618388008,0.3922606410844146,
 0.005418634446578641,0.14462997053157078,0.4467126280960887,
 0.5153026754953542,0.5824942761429605,0.8161741464249168,0.9353130624168765,
 0.6717403428459611,0.10460080331734278,0.48211117612455645,
 0.20770885864073774,0.03213033771646001,0.08421494625357073,
 8.166334014964027e-4,0.18366028086856356,0.7107184816018767,
 0.7102610893089746,0.9262784838554923,0.37514489275098073,0.3797006188081744,
 0.9126164978539151,0.8017919881730186,0.01898136147961793,
 0.07035507705571342,0.10789556590941796,0.112648667418496,0.9846917198423474,
 0.4623252982404634,0.08844141581987353,0.9125040936368469,
 0.36603642208489173,0.45731175305812755,0.6073741020035318,
 0.24601684283273284,0.15636603666342974,0.7503998313631695,0.52328263262632,
 0.33101661978856267,0.21589094044152968,0.5528787616702325,
 0.4301477389871372,0.37874290235913477,0.8790381209838115,0.5509633632881991,
 0.3751855510268982,0.6086749528018977,0.5861950726391395,0.8627471546842143,
 0.33369759621145034,0.28961047682687957,0.8203183145055188,
 0.1575652994059773,0.16976848646735232,0.7573980849623855,0.8262316075682286,
 0.152757768002156,0.036849840152319735,0.8581629895585979,
 0.49841001098064686,0.5781167403167289,0.17026388942030313,
 0.5192394577052367,0.20872314959173344,0.6171990097206546,0.9611445075564768,
 0.24870327338039588,0.9631848727641903,0.10446075042872904,
 0.6169802511719737,0.11585379245233718,0.5606590434173905,
 0.05089365413641511,0.4465511827969091,0.7853837813784694,0.6906225674444335,
 0.8580936762862638,0.13883096887501134,0.08665027252318314,
 0.8692545330464606,0.043015018120129755,0.4812278358175388,
 0.9286063699225189,0.6848653618040176,0.15778134002538835,
 0.22159194505431756,0.8120221836454662,0.42117580518637365,
 0.6686910702374701,0.49972659827025057,0.011973022401648437,
 0.4709997007616946,0.4809248485548133,0.11647193591706906,0.6325035515290103,
 0.03225949267293182,0.4650408324492975,0.6443007221952443,0.9793126435299015,
 0.728020979356977,0.04251602134942778,0.5004130953316352,0.3573252874721471,
 0.7741210664534504,0.19595398179588286,0.9030159468071347,0.7401099678196882,
 0.31351105211694097,0.08355604862146859,0.33246894491147394,
 0.06699694569552761,0.2940623875435293,0.43980769718833723,
 0.16566682045212588,0.6926496888514437,0.5080677850854497,
 0.06541424062460832,0.9247071712382008,0.7618502586676389,0.4789131945378161,
 0.25859826534128083,0.3546144668787823,0.2099322633081656,0.7912408597454205,
 0.7226162305363759,0.3407462220629861,0.06957285521221723,0.2010934055632997,
 0.9586382572355826,0.3511679075745562,0.4457869512794018,0.24850500216410054,
 0.021502688498308387,0.785357087493073,0.959675041746902,0.09873774126807089,
 0.6107944975498147,0.2648227277984687,0.3598224566560818,0.7481318605400803,
 0.3301337394823274,0.18382120947021152,0.36372066261522107,
 0.9834880982883083,0.48475069130101556,0.14983327985698924,
 0.09767928560171701,0.5092774784166149,0.3105581893137642,0.7799752860703502,
 0.6625190186081159,0.8121774775446691,0.16430621265491296,0.8656692493372952,
 0.5527802898817454,0.016590899451897623,0.5151985357655554,
 0.19842216597876472,0.6555733657519073,0.9332259955192664,0.9648635886078135,
 0.1423249986199573,0.7645701762057496,0.07751975430881242,0.2240888749391825,
 0.22499105186958657,0.3699934985843899,0.668132462469462,0.595006807748417,
 0.4371123478987956,0.05585574781831215,0.9254730568542745,0.8010537388861629,
 0.11401496740600048,0.4701850811007261,0.6389163849445766,0.9276667998429311,
 0.8055768124896144,0.3841796023219781,0.47598399555488957,0.5667070820997075,
 0.7201704492922995,0.23736861097091022,0.42774712671504733,0.625286814649672,
 0.6128891946756202,0.07948634292678991,0.9248817418941886,0.3386995769101383,
 0.4682854938293748,0.17874669179935276,0.298783235994615,0.27532744448199375,
 0.5651673961480157,0.9013944258167301,0.7659242727373895,0.9806569239406925,
 0.8048184538819831,0.5492741013423795,0.9494551408459417,0.5059075628556978,
 0.07266318336893118,0.17189841037450582,0.21018833554918182,
 0.047410402880706526,0.6108204757013411,0.6573258343459762,
 0.9801837360586153,0.23542192605580092,0.0938953921532184,
 0.33976915731526003,0.4941922865473889,0.3373398916380379,
 0.08207111105430553,0.7133916197633985,0.3936887506808322,0.8421303977536053,
 0.7959948089432471,0.9616661097448594,0.35455956933580113,
 0.11670610372943235,0.6950388513469942,0.6005726825317766,0.9633385660395227,
 0.5260874160680695,0.33376335471402785,0.25468519979706783,
 0.20711659875346855,0.0672067797268634,0.4428721324027496,
 0.040077577072322956,0.40992374487407757,0.16202386901222698,
 0.09152096169514912,0.3252686387122691,0.3303661071609909,0.7326725218123382,
 0.7485152401248153,0.174556549170647,0.4702214542532872,0.8546836167666083,
 0.2676672911003969,0.8148191380697111,0.4973891480693089,0.8748726753306436,
 0.5554977851744466,0.9253851051525789,0.11776754232406816,
 0.38723881266776417,0.925747599570778,0.9023099859448966,
 0.027438076748644735,0.2088019192463605,0.9004587899865986,
 0.37814398444242636,0.9586489453511443,0.30962562327661813,
 0.5271818813752028,0.7194768867207337,0.1867140612042274,0.5204330010091811,
 0.9373994007284189,0.9703004798434584,0.29608288601978017,
 0.027886703267568036,0.44323402206261187,0.12551577661185942,
 0.29234774382254924,0.4483884871254562,0.8535095042726231,
 0.44127981605305466,0.17915487630152394,0.5148661778608343,
 0.07427906093865089,0.9701953921899918,0.836358349973962,0.482239010426871,
 0.9971748248221547,0.5750274784345647,0.3358151023208278,0.6404407804274835,
 0.16629954699282845,0.8680766042147217,0.4687326560064857,0.822726142409769,
 0.8714222623821064,0.296793954983542,0.9973184804966817,0.7415073934009359,
 0.9844441664971685,0.3119440253990954,0.675462865840299,0.35593147547650084,
 0.9394919705338609,0.6229191671133323,0.34037347404984175,0.2547763693622913,
 0.0241198274178257,0.5632536318856364,0.3889463534638472,0.9448064404728145,
 0.9067745305461261,0.22489895227397338,0.6158894277184346,0.9470248637454088,
 0.80007033531146,0.5541952589762942,0.7529952070965698,0.5368249167998677,
 0.4507273232313953,0.6988811788682695,0.6462428278514865,0.6395383268992275,
 0.19961216785992753,0.7054187716238428,0.9433131647518839,0.3506532756909734,
 0.9266620005723825,0.7614493130875566,0.3376109233484299,0.08177365191236419,
 0.28521793335669277,0.5342826625959428,0.6185714501105601,0.7013808127975525,
 0.35419158558689223,0.13837187351275892,0.8329551642918662,
 0.6197696598640524,0.48119180744934,0.6696905658050698,0.4159776163687181,
 0.5957077140960423,0.9410356444385015,0.8603080911101875,0.1549925956078103,
 0.40366936018739197,0.2658820017444641,0.54124584653631,0.7916288516099914,
 0.39040060629246176,0.3370820004318793,0.5008510465511761,0.696520596624015,
 0.51454329826148,0.9438945207843339,0.9478269378079456,0.982460872933992,
 0.4275039821985481,0.4858357565778967,0.8251350062872218,0.5767530580655964,
 0.2949231661781381,0.2724593341260747,0.013498025245968437,
 0.7963753482553515,0.7641283225586109,0.8126710597401909,0.182548758443831,
 0.5940349983233821,0.789557474125288,0.24809883563088464,0.11717822794078447,
 0.9266025967980643,0.7146259809678318,0.4395413166338642,0.539698607536961,
 0.5088863245917735,0.35775942700831054,0.761630843882733,0.8500640913657364,
 0.30539404327815045,0.17756051984674182,0.03973860511257077,
 0.7122081067452232,0.8372561331943897,0.5952939583403216,0.1462794191801554,
 0.5630763974035309,0.6676807014050012,0.4007099677856716,0.9882607457729752,
 0.7545875133572413,0.8772515681462238,0.2941145286344258,0.6499881963192393,
 0.509250513443986,0.5695502291996472,0.5478811207417051,0.37791075035482535,
 0.15551250333408784,0.2961833847520755,0.41827874988922065,0.145154246233949,
 0.23980610467657537,0.9028425250143017,0.8664242119435634,0.9312688733077832,
 0.7222084607164215,0.7605236092868664,0.3103976139774032,0.5016264342859196,
 0.6851368901880486,0.954609729103509,0.5672157321819958,0.138699424700635,
 0.8676903422476314,0.29578949131581567,0.02923190945748355,
 0.3829206334910713,0.19238664259953364,0.09951313870546641,
 0.02804229688187343,0.34106278307714855,0.07551543877205036,
 0.687285301966682,0.9164441514490771,0.804036184050102,0.37622935552938863,
 0.8432700122604084,0.25300324200636537,0.06125849530349403,
 0.13143515162690622,0.770828204509937,0.47231601823575264,0.774774858959134,
 0.9978332750300836,0.051625717472315036,0.049326786549076074,
 0.26357384264794836,0.3721182256608455,0.5831747471745349,
 0.08940424910549782,0.6516441814951981,0.3836489704660778,0.6072130450049634,
 0.059924245923328234,0.7386659527771913,0.8870581443196628,
 0.7848170264324943,0.8825605859304794,0.44008767402138915,0.7730930961259378,
 0.9609823898352156,0.8174009322510097,0.06252338549407,0.4812186508536478,
 0.6092099717426551,0.04274331934829756,0.42805201365885137,
 0.5149865608611991,0.08816480552121775,0.9474080776238817,0.8460588836063369,
 0.051879131827079594,0.07545625569053016,0.7739649960905655,
 0.5563337254399668,0.6913015742939194,0.8885829918948857,0.8260421982999376,
 0.5132543064508781,0.9478660276894877,0.07903298872706288,0.7867299918908143,
 0.22010429218711025,0.9401281005271439,0.5076390488943883,0.7933798312144379,
 0.5785666282862963,0.8673011536447234,0.2993774938992787,0.8562312416876205,
 0.05222313282411828,0.7741753949247465,0.32943734722158047,
 0.3553884572912763,0.030591731138241873,0.6751641682072802,
 0.7691196022488748,0.843850131575226,0.33523719705812527,0.7834731738684912,
 0.6758224178839425,0.7387216003151462,0.5744132644694936,0.35051234864076275,
 0.261660444058997,0.4152602182400157,0.9068033556398958,0.4924538374267071,
 0.8357579123257123,0.519215610314078,0.2504056432336993,0.8176701065377169,
 0.7453709212276497,0.10798008036314799,0.503445801874319,0.7338814437924881,
 0.9200488661104819,0.9362280749299563,0.023084798539343643,
 0.6827539775621587,0.5215089028601451,0.5683775180325719,0.08915018782997086,
 0.5344717715853481,0.5753579014462001,9.952299424830224e-4,
 0.43207243426296305,0.23484654051812304,0.12485549925616879,
 0.04598551376439708,0.22303086221569268,0.09677781735473112,
 0.8655752852050154,0.6344691500540555,0.9576856921711668,0.09969997528251318,
 0.30753447377992504,0.8901704578253928,0.9368492337741694,
 0.23052678781344804,0.6374645545765872,0.9014429516519012,0.2081309694671971,
 0.507654557805441,0.7072953408147395,0.7159657676630298,0.01752514185026155,
 0.37472763042853374,0.6070446242528306,0.7243651439189066,0.6322739574910138,
 0.604981718984608,0.007665142663810687,0.5343641546392998,
 0.12287300905340603,0.3339917046317169,0.5430750744073133,0.66202523687676,
 0.21549335837945238,0.4299627546871224,0.03740171381094748,
 0.7164600831102292,0.4923596211617466,0.28743034431328596,0.6041981009959889,
 0.6606983169072739,0.641093091173916,0.06707309663598404,0.17582511447528293,
 0.7192886844708465,0.9921638526817435,0.9399517468477098,0.1055354188966996,
 0.9825359192984159,0.26587050396238554,0.9657573802277416,0.7487650871299211,
 0.7921914671728799,0.44215486207452803,0.8448262619294675,
 0.30673337659802313,0.11316537458763676,0.5335894166561922,
 0.7854850637861925,0.8893275486767412,0.28012460925640315,0.643522182401617,
 0.5055942470563444,0.43972601978980674,0.3223797493908691,0.7717200652740442,
 0.7904436898124669,0.6955338920005371,0.8270906533972554,0.6229765625349161,
 0.6423929982713642,0.6698767361275123,0.2118696490484271,0.8111032524157265,
 0.9125534198268005,0.47409089199956167,0.737322589996306,0.35867249239210763,
 0.7281642875867287,0.7734460842727224,0.3515275997467503,0.7305038066501055,
 0.08744491348130279,0.08069659157504483,0.10927104915559405,
 0.9458288246316096,0.27024401923147134,0.9925458679240057,0.4738639352682992,
 0.17501839590419943,0.1597167763788283,0.4529772141234726,0.650939941449648,
 0.005820086208488506,0.7872952445591199,0.4523236517826653,
 0.46558111835392235,0.8013955591918108,0.4438245271911181,
 0.20132052598280936,0.7507662208102112,0.238242425812631,0.6066567502260667,
 0.7230962391359097,0.2476700133402272,0.6054475638451106,0.8373591001219804,
 0.45230924493096125,0.24195208833482784,0.7408370227421468,
 0.21536942331530196,0.5505072716394204,0.39277404376782776,
 0.2850662561811099,0.845296020982717,0.34744183815821406,0.8461908081496231,
 0.04831952457501698,0.7014442072192538,0.39149735678775555,
 0.8092562721021306,0.8343624083098398,0.13264061030257923,0.6658410911140055,
 0.7783789777078713,0.6414148968095362,0.06356369706419684,
 0.18627850255689982,0.3975691000501267,0.8854657491007265,0.8823253438188582,
 0.9880406165787913,0.7211737457267866,0.35949361838745664,0.7447558914111376,
 0.11217314374742315,0.45106743380271364,0.03798993222592384,
 0.5228737493564701,0.0964819668283946,0.03510733318223591,0.8799013173845591,
 0.12613408367586165,0.31826777530455796,0.820623720509114,0.7237577824799126,
 0.4420393029823064,0.5240535698971374,0.5671220478986726,0.05912681621095894,
 0.5430816233728131,0.5668672062745869,0.5718863564507554,0.0592882307729512,
 0.9878498023800896,0.9804502076235686,0.5927575892730895,0.8498080239892968,
 0.8670616289348358,0.7843962182485444,0.1871500984858171,0.4470981198960319,
 0.0720299599900619,0.06555330463721876,0.40152860544028857,
 0.16165989751688814,0.829382142759969,0.19832367106213022,0.3885979944266983,
 0.7686994744712068,0.4430866987972767,0.8882128949255774,0.6941230156088709,
 0.7850784026828399,0.8139121983727087,0.7484822045534227,0.7642849343587013,
 0.7181683243435025,0.18147671743700788,0.13350868302236207,
 0.9109150608167272,0.644667382688973,0.45687536999653144,0.38910517283820123,
 0.688183456786938,0.7558047755874918,0.20347342808345648,0.9035174916160564,
 0.8968695938494798,0.29330909019500695,0.4476959632757733,
 0.18151969755512254,0.375059294431195,0.6335710898159287,0.8400149947562099,
 0.3732255726297994,0.38598146664091126,0.7055224431389407,0.6414278684375194,
 0.20210642398739231,0.21417044303355415,0.8161866034286278,
 0.3510157696410663,0.3806157273633545,0.8513314888354687,0.728765987064742,
 0.6534391548694265,0.22972884754485368,0.8860964662834837,
 0.08294469241438696,0.7425284319997536,0.739577775450548,0.5773299044650642,
 0.21654050688864201,0.6734454773023849,0.5798638838252411,0.6651091305695569,
 0.8764711269400649,0.8889501590549005,0.536902788923883,0.21883666148512648,
 0.3037100802396987,0.48180327232473885,0.20706495874975583,
 0.21878157344727656,0.6969852564222563,0.9741493340447266,0.8519717991183448,
 0.2248647101874709,0.2031441070557103,0.7225860250370644,0.7593862614374766,
 0.9888534388795548,0.4799921135784093,0.656676179954398,0.1957630333923882,
 0.17184593427481698,0.1982606737184761,0.8703667097689607,0.3100436224964066,
 0.5301424318798725,0.6434867154994616,0.20204678956597688,0.4908681662982277,
 0.9531137614283112,0.8329901547202421,0.30538660992350486,0.4713233009241864,
 0.15894837357318736,0.7999539283941505,0.5023694836362277,0.7968597417054332,
 0.869853320042229,0.29231150102183956,0.9459090261029425,0.2285752734394727,
 0.8139953853875571,0.39483916828598364,0.0992146597028658,
 0.22249581267450969,0.510069111475343,0.020666267440792918,
 0.2748840108065843,0.12374090826424611,0.21688634316504685,
 0.4162708924698464,0.437694992483168,0.9603106236563747,0.5799901766037142,
 0.3260729547212684,0.9251666917390542,0.06753322734395228,
 0.35190416791560697,0.4051930141696438,0.8888556607440316,0.5263247530458363,
 0.18985544932909681,0.2503622395617219,0.9282881380619439,0.9525119591187355,
 0.8153099859496139,0.29390421778715725,0.9012604440110666,0.6767638923203622,
 0.5766003598868813,0.5531387025278728,0.06317652656660755,0.9294057704447606,
 0.6957414686948576,0.9323088669166563,0.9591671241926445,0.5831726499126212,
 0.7119039344838867,0.7015010188873732,0.6163490080898508,0.1301682389595118,
 0.9220388442541635,0.5727358254032014,0.09859620203613828,
 0.06491196218113471,0.2334769528040681,0.3259810741855932,
 0.15986846252243225,0.682708326191544,0.182156451791312,0.38354714645018617,
 0.2828439256259323,0.5085545917858779,0.2777267060355134,0.8692828149652287,
 0.8658955222168186,0.754118330928028,0.47266138219027487,
 0.026051732262610372,0.6150208533453803,0.9185477666012878,
 0.8041413845530683,0.9693491753551448,0.8558419364434844,0.7962248024216927,
 0.20952044813276682,0.6314554208472272,0.02397139973276341,0.667472726821858,
 0.3357973939087482,0.30597974560339825,0.7462339715607318,0.8233960369114729,
 0.5758770343940226,0.0034990430582253484,0.14682248587730717,
 0.22423133177916732,0.5103624077863046,0.9664150848985594,0.5133168524638001,
 0.5111475871018916,0.8924442131058055,0.9038434367934868,0.302097663934105,
 0.22635447145937482,0.43996347975892325,0.1639320594897682,
 0.9828991176076326,0.8946363186444708,0.3834118350925403,0.23725409857013802,
 0.38947009177435143,0.00943474239561537,0.3731279067488227,
 0.3716685568539291,0.21578567183685615,0.8153537124604763,0.7073135254643163,
 0.011507898486665225,0.7076404778152816,0.4602820130847384,
 0.9532014916935723,0.8035502497272078,0.24803535175857405,
 0.15036301824511633,0.201067453338028,0.273985907535382,0.21889395548501478,
 0.127387423037866,0.35860646880803826,0.8918115274221199,0.8763468010751014,
 0.5616774501060594,0.5382469070437569,0.9413466462765774,0.26353426538611446,
 0.4172116875263667,0.43822049922314743,0.9871494904952808,0.7640224749833459,
 0.4610862037801715,0.5913114876719796,0.018149330054314516,
 0.4641890512830179,0.21873238175743082,0.2562458827336177,0.5884230127644599,
 0.8172036263804127,0.029303554896121575,0.45145859670377986,
 0.258741829648343,0.18786577419445527,0.38424588270300764,0.6429211646173365,
 0.6108373013321989,0.27265636009670324,0.6908494485331741,0.446319734348807,
 0.4967770633457269,0.6950270692678988,0.691075914928035,0.1297083614876675,
 0.7332600983110951,0.6346890049172085,0.42310739696056465,
 0.18958403331044382,0.06945391118753408,0.25099125937503364,
 0.9494436180860045,0.8183588720981669,0.460813396284132,0.12949536546816875,
 0.3961498904348877,0.9163072237305425,0.79132442335824,0.13442429237374198,
 0.9658771459820228,0.2244521380169726,0.7241004327992733,0.8469582118338662,
 0.3976580623492194,0.02801435611005132,0.6252812750293288,
 0.16753432156118364,0.11999739942787113,0.33383553428099855,
 0.36708191319924144,0.4798792499192732,0.4462596712618446,
 0.24982531371352168,0.19638612942689027,0.6672674686235253,
 0.29047036733969256,0.8057855289310759,0.9823082633697436,0.5384631630782801,
 0.05472138270562288,0.10587904643874624,0.08352513394083527,
 0.20141913777632214,0.5249363919481111,0.890664891482361,
 0.027395228784435455,0.06677107079400557,0.6297498603875148,
 0.29602814922888987,0.41572935069862815,0.0185506995344632,
 0.5111084764858066,0.7355708620016752,0.7097721214713977,0.3937602345832707,
 0.8845605202674057,0.8582411413815799,0.20855415872561256,0.3154973514995658,
 0.46302512137390606,0.8418778566168033,0.7521037446482668,
 0.31902962013099967,0.7501623378306524,0.22707594465527214,
 0.20379996316857463,0.8531774205853844,0.6557337673817306,0.6286255229563308,
 0.37173618593617486,0.4602595514028449,0.576913887854166,0.04614951217885377,
 0.7911014446698428,0.8729093810332145,0.5205664513310901,0.6592892951386401,
 0.17144948381155878,0.32376597398288476,0.08672179072587516,
 0.14724565963377145,0.6354409276973415,0.968091181924539,0.39977603523905847,
 0.4115836567879593,0.456635223045218,0.5999615211100274,0.10466021184783925,
 0.05433996880431802,0.9131187953580934,0.1788108480401258,
 0.08229166577659452,0.48580162361868195,0.42028163181034617,
 0.6574118654689755,0.37449121933755813,0.060674212807721895,
 0.6613786645989096,0.2962345093339436,0.4663141888907414,0.968432171014736,
 0.9075227557531138,0.5596000492961118,0.8685331339013246,0.1600446730571694,
 0.38975823418720534,0.49241369153135894,0.05383620556677915,
 0.07388825528976428,0.21944159281975306,0.5733299470252138,
 0.5599393518215566,0.9848137591270235,0.7183817082022107,0.2546744548010007,
 0.7283009599201775,0.5081630777621673,0.9602427169071175,0.21879478943676445,
 0.3171881703551035,0.7345990180605885,0.5570054119170491,0.6847924925344913,
 0.5296909970037176,0.49689281086684234,0.2319607088659421,0.7459077372670265,
 0.9154223666111794,0.3633366706667067,0.7034913194607154,0.2503876859985742,
 0.5832491602976917,0.41216656107959837,0.4879090981065983,0.8650044951588922,
 0.6989331267908518,0.6000755270583515,0.9659661746929824,0.6631914368374483,
 0.03268012105139395,0.11225834742583274,0.5994544399789196,
 0.8596133161703111,0.13242252036795343,0.4571793190315273,0.4240384072581357,
 0.9842252025027339,0.6416513542926235,0.8168749585974542,0.7237692973204268,
 0.5942868281590696,0.36532015234789805,0.09785782768306894,
 0.8730113478520083,0.4051228221708223,0.8919366947077513,0.32092392124627356,
 0.5110090426180629,0.3887375917638419,0.03617818463543465,0.5018674266820646,
 0.17069110876816662,0.0444956306603882,0.3570531649372086,0.6118069437780045,
 0.3491530110558121,0.6402447616502953,0.8461706037228998,0.9530955875649785,
 0.30550467502118794,0.0932586691037155,1.9460919614977534e-4,
 0.009807763682625925,0.5788338765627344,0.3107647663966535,
 0.3686480606575674,0.6280703443512159,0.06372152719540347,0.7424820175837685,
 0.5473955973659144,0.2695613302334402,0.9323428556373906,0.21621798649951307,
 0.05701640809148112,0.8934781013276524,0.8424538136470512,0.4140900642457477,
 0.8410881868688378,0.9389022816353041,0.07342960378543562,0.27805395674862,
 0.5553604344754142,0.5676944072872572,0.16225484774335963,0.7985305613663699,
 0.1100780692577844,0.8793328483739101,0.19207492031473317,0.4401592849192758,
 0.8550524363751023,0.9345993466096685,0.44644635610851446,0.5712208895577162,
 0.22631641564010618,0.4539413543827151,0.10060898079691283,
 0.6831677997898412,0.22931449104587642,0.5503934832536976,0.2435658601153915,
 0.23947389466237146,0.18076326183293256,0.6123286374863395,
 0.8393139663917166,0.39101685918651463,0.5467744149212869,
 0.39619479442815386,0.30062518152511597,0.8923227908819314,
 0.9128856652480978,0.3564195239297746,0.6753828348931626,0.6518269880820132,
 0.8309244868958111,0.7433163005403124,0.6557423203501458,0.8001833249841083,
 0.1508970776256272,0.7813582108815189,0.2184564414227388,0.8446790526854361,
 0.2984091108198801,0.46427587502462964,0.4439399856171239,0.9464274018139358,
 0.8585859551131302,0.7108248972550929,0.96661111772385,0.7980535405132562,
 0.9468350213113352,0.2171300048282554,0.6662648887352567,0.37092751673639845,
 0.017993968961820817,0.2793583926707913,0.011457248469865178,
 0.16268481165755466,0.5016250712166039,0.3431691073473828,0.8200426023322966,
 0.12059742105380644,0.6339991723341518,0.313141457109154,
 0.057615571995307424,0.11615866072812364,0.461086240514451,0.785116742260039,
 0.6590472098445977,0.5726685202152709,0.3438442338464688,0.7822565622236759,
 0.9065574597555882,0.7984789064473373,0.1533461164482759,
 0.043633900417977545,0.6926413387081595,0.694182726234689,
 0.06243494932543525,0.7894482504506608,0.39829574440923965,
 0.2718471470650303,0.25091029244313456,0.35006821630852036,
 0.07639857997842303,0.08441005520758849,0.5201883739402864,
 0.8848800035853951,0.8094368049932117,0.08129302964279872,
 0.30493163076151986,0.22751390803700555,0.6801949516913774,
 0.9950592099180198,0.37553251808971433,0.8915990933813926,
 0.40537432901806225,0.018714324380679326,0.4685585990273575,
 0.7953208815178807,0.6245858299366108,0.3787902428368578,0.6208977142160466,
 0.48459296041672917,0.5004465164621952,0.9157985985296204,
 0.48524037041442525,0.30374702316556057,0.06708040188782627,
 0.10186871996306546,0.0981520833962728,0.6118896819271977,0.306230272691981,
 0.7971422535056782,0.939668783706029,0.12134554729715197,0.851176067693493,
 0.5233007182133211,0.47093276675288165,0.7198319510051072,0.7429734963842379,
 0.9243929225660624,0.6809829703970148,0.3473460936873387,0.8525509210670705,
 0.4441115596235302,0.46023908830430815,0.3996471372143372,0.5293820401537903,
 0.032498035321179675,0.23069432593588968,0.39104859929804325,
 0.36913446853813003,0.7701622312238009,0.01851885848433954,
 0.6317685978395837,0.18644698445620467,0.895236027063472,0.3935022213812177,
 0.4690300236216681,0.5813692113596308,0.1990921307472081,0.39761473555718285,
 0.5242047711472552,0.0919234504338945,0.4273903792331155,0.8619011145700041,
 0.2187213318480825,0.8795583443482335,0.544057351347411,0.73299373901408,
 0.9603844519349194,0.2892838047872728,0.7022244599967962,
 0.016460628251057874,0.4938117295698373,0.24270666503086358,
 0.36050746147978785,0.9636142455686454,0.2530010990023458,0.600768627495142,
 0.48912642165798004,0.1880348495500187,0.2968312721392453,0.870490681288839,
 0.09103540919251074,0.05952124348720922,0.49290571631728375,
 0.6516920484856772,0.510635664743448,0.5531026508188496,0.33161622894814236,
 0.864546481208144,0.3568078415611029,0.5696429851130239,0.8111423180258571,
 0.8866596599738173,0.5548935631798587,0.28442598839981836,
 0.25670354137420437,0.0012364212406478226,0.9774218767035048,
 0.3041830547426778,0.4016665554810199,0.9076711216000224,0.3922840003281004,
 0.30552386515240515,0.21065527070516699,0.8852348753117106,
 0.2941162584570527,0.7988306744752778,0.175174651833085,0.9857250236405203,
 0.7660980274978426,0.8055335110115923,0.13174497431376975,0.9924768965088905,
 0.6450830502354228,0.8215688435740909,0.9720818114581853,0.8823226758924502,
 0.7879344824655435,0.22867917493493028,0.5341021395937908,
 0.37112225504444374,0.1513427324696357,0.7336552874092245,
 0.30875208224753703,0.9834919927928889,0.26668472811104893,
 0.6402006593216832,0.43435407743003895,0.8588702745546666,0.8845866501419779,
 0.8016340380447372,0.6523520470684373,0.09224797414255581,0.9738707191467452,
 0.4740104322611809,0.09308324030086701,0.024660196463692663,
 0.4576976402680436,0.6089274473809521,0.33937934936030456,0.5662291387511866,
 0.26961554295209256,0.26240988035418145,0.6431285553884459,
 0.24459847642801047,0.22824118183364772,0.165643865343316,
 0.14410506436964954,0.22611413065329367,0.9624822108558155,
 0.30004313878676836,0.34882786502156693,0.608494482248238,0.7084199312313079,
 0.7111744550635397,0.055678131753141624,0.18244910507701784,
 0.26135424117746786,0.608286098006812,0.14205132655967723,
 0.15971938204408054,0.8422009109443811,0.0936354594396297,0.9669580846028767,
 0.6943439166271217,0.8911122345954912,0.22096305560832397,0.5790030582452805,
 0.5062577099832509,0.48909278184888905,0.6888482443428436,
 0.36067484874578226,0.2826374525158625,0.9915552640356957,0.3356651715504707,
 0.17398237809584338,0.2686777731484974,0.06974569152499144,
 0.0076274576500623414,0.022164887929483124,0.03447600656253447,
 0.6964566815963475,0.5902963870322353,0.7590404108758301,0.09593221399764029,
 0.5475003636810223,0.8510996197812026,0.004283884443590624,
 0.19884346990698887,0.307296955616722,0.9099343548216202,0.5062095299641614,
 0.12191740712657562,0.9752529507079015,0.24930665186201628,
 0.31680791707875455,0.5205212019231263,0.913348852225349,0.5813164881503032,
 0.7759931229180294,0.25652849408364153,0.6496544708776241,0.1763887183061994,
 0.2315483475206629,0.33263371829439314,0.7872119912204976,
 0.005088038076174861,0.14117990371097033,0.3449413145102115,
 0.38920890908072925,0.9054220180526431,0.09326662739164471,
 0.5926521164636563,0.6109198997311381,0.12174200830052218,
 0.15732475778866484,0.3150361317257606,0.6831515455412845,0.5047307216495728,
 0.5475148111459489,0.6403282748018136,0.8918024553658745,0.7901049632030491,
 0.9688563535122787,0.4970195915301745,0.5296118918257714,0.477594902401681,
 0.9684590860737144,0.42452985599970294,0.4234158183619505,0.8187939978299786,
 0.30807344891400357,0.5933770798490262,0.28643410605654573,
 0.6203889403072588,0.9433057559194769,0.8453029413523551,0.16232882114450087,
 0.4083436163186288,0.8895017164342279,0.5973433748732698,0.41970141504967984,
 0.11619997118772551,0.13128614504973035,0.7482936335468546,
 0.0043499400767955665,0.8565039962968145,0.8880184009950126,
 0.16468846618957944,0.4556976720452113,0.924471548393142,0.6244602829755623,
 0.5239441992713033,0.6500399325894118,0.9991480371345487,0.6016444017764332,
 0.2824134716269113,0.5860753016238743,0.3839093955246836,0.9476039879513287,
 0.5449712507881194,0.3173738673434051,0.36524340607085903,0.8963429343457953,
 0.06400436053097935,0.7058094057959481,0.03264469132723602,
 0.2449859538577822,0.9798769984590044,0.4380536327826642,
 0.052176555380085476,0.8124286425087459,0.04386471873072406,
 0.6126181787572686,0.36273878260817893,0.57156501942792,0.007799452509673066,
 0.48676113283439904,0.5635952599297396,0.8592229354749611,0.708313044939116,
 0.42726927670389037,0.16735371439691138,0.3098156300175019,
 0.26044309341706806,0.5689486864209723,0.7745443793714732,
 0.13561475949651014,0.20149388685642544,0.7818504608696014,0.132438896161149,
 0.8881069854717614,0.7473459464761307,0.5996932231516282,0.17361550691537508,
 0.9262433952287242,0.5873675206396756,0.9003578342486899,0.47600379935891424,
 0.07741070130478023,0.9542709934471991,0.8753508228429397,0.8061514058800785,
 0.18565053551236077,0.0021419359489777134,0.31390193043617143,
 0.9680352185502826,0.13087543024132542,0.5356115147101295,0.3007890825293279,
 0.28082967897042954,0.7897407369257706,0.8395394953146714,0.9331672952839123,
 0.8130083820371077,0.14591555447564775,0.8872600065010336,
 0.14165719378928843,0.1754418706114549,0.8138424250516112,
 0.24128035943187542,0.8004590564924663,0.7922838081903615,0.6175756459160291,
 0.46431488907186036,0.1753828977034949,0.9971325337173608,0.8702519800327293,
 0.765428571300914,0.07693939671096983,0.661034531796667,0.14745353891882163,
 0.5146507146874533,0.0906668089797934,0.911139892281896,0.6392246157730147,
 0.361201317825286,0.8133936037572861,0.9041780689533576,0.6877186688954207,
 0.2415036406788681,0.941809226972297,0.8578733262851472,0.18402552247615123,
 0.6018545411169038,0.09341223724460002,0.8539417653194953,0.5460044386380055,
 0.9034535350056764,0.540053341866345,0.4318706585797978,0.4292245219516625,
 0.8312054592865138,0.799947714657825,0.5350998771946284,0.8302605056884974,
 0.04698365413236527,0.2592392199058693,0.16371073089018162,
 0.7329247144329089,0.4580332299800193,0.7954904346104091,0.5769624503001811,
 0.8515141366890521,0.5374276906986231,0.5110359205892272,0.21128001965217424,
 0.6844256690282254,0.29789548091422935,0.5742532843940193,0.8337496774887342,
 0.4219231927487179,0.5486115152293924,0.3456648416437599,
 0.015684251867250865,0.9813227171656771,0.630643514883589,0.8487742254785062,
 0.9997494383766286,0.44316062556337776,0.3259611145721627,0.8381146246846013,
 0.9937697186953064,0.7347932116631857,0.685603558138862,0.6283268781355372,
 0.6852423452600177,0.8640733743592443,0.8671903495035976,0.48877822072663635,
 0.13789092790009305,0.6728915081227026,0.6570761672270888,0.7257683226899043,
 0.04334056137591569,0.7434435688068101,0.3457543942146297,0.5122012561111703,
 0.2620187963786851,0.902020817009366,0.013281883497472502,0.0768057308544352,
 0.465697039230452,0.5851421125280205,0.90896855821759,0.6649296799879609,
 0.24187724549751433,0.1339408996972783,0.07906478558083074,
 0.5294655962613117,0.9833786472301089,0.7187118473391207,0.6079837724446102,
 0.5581377366914904,0.6591256856345988,0.19081998468315553,0.1861153119496679,
 0.3633326387801423,0.7554296783219111,0.12365160460272162,0.9616954088897485,
 0.34453337156559405,0.8789311432355037,0.07212632919259654,
 0.8189532435398171,0.9330012411040007,0.32520968891677615,
 0.28270135979006605,0.5204720663887015,0.1872501517814692,0.9085266793993956,
 0.04716296845358836,0.5010971089683615,0.736690672065167,0.8717932570622701,
 0.8072037778482025,0.5758229074705452,0.9957871699041948,0.2506564965273044,
 0.27780737494525143,0.24231537801558467,0.19783118815667144,
 0.8453668756746465,0.36974621813417974,0.20824056682524483,
 0.2546342355310586,0.6717712984445051,0.054906920195767595,
 0.1394332569087744,0.918076203526652,0.4128790224957205,0.419830782312711,
 0.9358922032322274,0.5720748987463561,0.07931084420360024,
 0.012844612513433717,0.28574976956120257,0.7233622934871494,
 0.5841031200461086,0.33453351908632023,0.18437080115240922,
 0.7854365231260727,0.5654039730994216,0.6993158148143432,0.7624149042305486,
 0.5308316998185538,0.48537418846262015,0.6187498283161323,0.7738519705795794,
 0.47936828099929985,0.38626669602998476,0.35121176503551355,
 0.7645593103076814,0.5558800675379898,0.04643490344155898,
 0.39795701432186836,0.7370328766079852,0.617135979816876,0.9015664887308548,
 0.3301926519654198,0.7176989770993749,0.448935727427913,0.7815875818980068,
 0.40408785764162125,0.8720649085637145,0.4271217476872953,0.3733775017055625,
 0.2747119736656737,0.22067599684848582,0.013859779805769135,
 0.8381721028941412,0.8829991464934,0.12663330625912528,0.7908541432891737,
 0.07089813815061097,0.31022341553396504,0.41344222336634895,
 0.014008821627214396,0.5396250072289308,0.08215206665085384,
 0.9283241558159422,0.36802804394707134,0.22411857983607253,0.569915708792246,
 0.0740854677284346,0.9210253509567922,0.6615551350225635,0.7476477386458331,
 0.9975437928068586,0.9763917550449681,0.04382582260279333,0.6855554309005241,
 0.16033952442720478,0.7599694407681753,0.8330334678299344,0.569187500720861,
 0.9190640214438811,0.9357728080701697,0.4566895710547063,0.3788008821720318,
 0.632608665537161,0.08205549633218578,0.8370569923111298,0.8685518404803789,
 0.2648248193448596,0.2753564333323957,0.1903944068563317,0.35014314914058386,
 0.5163282291313132,0.20899772343905965,0.9105913187503083,0.9579506120094776,
 0.5097007867359807,0.36550238174809146,0.2459425120643417,
 0.37984944078953575,0.925402722669951,0.39761973171035403,
 0.47565808262251585,0.21817411978431211,0.9521410758776818,
 0.3402238860450778,0.7853966001085788,0.29563028069536745,
 0.23729019783754246,0.8833926844446084,0.3277031762034659,0.5500286785893191,
 0.06401145862422142,0.04119456359853779,0.32833161533132293,
 0.11135625042462316,0.15060334844751755,0.8049857997592089,
 0.1350984747207039,0.02638017389813674,0.08779503377058162,
 0.44571331861789176,0.8530093269713552,0.8600028228884035,0.8679328739640518,
 0.8184960408188142,0.07996265659709778,0.6183976451394231,
 0.31746832473298825,0.10754038997720239,0.3395716345522115,
 0.4398450665835508,0.8121713847854819,0.8824123356715591,0.5834897342722527,
 0.06142317665203623,0.8592234699227224,0.3121332031144971,
 0.15000926170733475,0.6145125011867176,0.8063735313538628,0.8380569037019666,
 0.6498280463644303,0.9385551130593185,0.9840940168457881,0.2347524135193132,
 0.7698990359678567,0.3411042966321709,0.9605522308542627,0.17797592035993526,
 0.7775726400889361,0.9207219614326387,0.5542855465988535,0.6520558699855665,
 0.4867042406220822,0.2923765879374678,0.20585397925574234,0.9950332500264312,
 0.06983214609438526,0.6462716338167297,0.35042419296055705,
 0.4619454814660152,0.038411860625159,0.535431055787546,0.8220871762437916,
 0.5252102778143627,0.13326592120872127,0.8334891812983997,0.9253255580754347,
 0.19969159717397522,0.723171549651421,0.2497718552918018,0.7701405705674288,
 0.11086251676331893,0.5778923706595637,0.3573409554885173,0.9225311861171692,
 0.6625937271879244,0.6123543502533514,0.7092738382234618,0.9875330515219901,
 0.38592911560544685,0.7139175483571051,0.6358595003481415,0.5810592431156104,
 0.43368693667004177,0.1436804805158255,0.9799417618229155,
 0.012224495662824908,0.4626356282967611,0.31969046671293155,
 0.5373902668299997,0.05204822786841257,0.18254421544579646,
 0.5497174506629218,0.0829609518185973,0.023376074526423984,0.366703681790681,
 0.9914071190647249,0.18524446497116864,0.6422168355115869,0.3008124217794298,
 0.03320167176284561,0.3815294604871067,0.45366452943237257,
 0.7354969748566412,0.558809994970189,0.8508640261106816,0.1550280441202978,
 0.019795502838409584,0.4164715686130339,0.6211780016850046,
 0.3703840390983514,0.21824209323498278,0.5194985598446927,
 0.46963235260748504,0.2088478180241834,0.2104692282885422,0.5035022989499642,
 0.3753976801260903,0.49580104698865113,0.0837879111471832,0.7102508907608649,
 0.4216653521120488,0.8917091090686179,0.23313005647821727,
 0.21066585942752847,0.9836618813334014,0.37906134734383445,
 0.3229782040421548,0.23873468754158877,0.18563077994254173,
 0.5787807641098517,0.34685576645136484,0.3509116250638501,0.5025270542426281,
 0.507398775343077,0.6025000638015487,0.9843840013686256,0.5515182791396193,
 0.7589400521066001,0.028730897881172557,0.14414331260839874,
 0.9440802105869127,0.5398119299570757,0.4620411990825133,0.5508113454474873,
 0.0676832594035306,0.3932984260150628,0.5548389170865697,0.11583720330661718,
 0.5025399886908047,0.9703548368439884,0.9987104660088333,0.7869441900000256,
 0.8442487985634604,0.14531934773356314,0.9821432368046558,0.915005863715072,
 0.9539739057818167,0.993627415825297,0.686001485058985,0.32982623815332823,
 0.0816392634938421,0.7654874795902001,0.7854396213418249,0.7487234847095312,
 0.7411830328401239,0.9916368163628462,0.60220930315148,0.3166824609749094,
 0.9188164144757247,0.2602251537045035,0.2180241814998637,0.7207395238658638,
 0.23745553517723073,0.6034145499203486,0.9887462986854552,0.8079978138477972,
 0.09201168090494838,0.7681299361305025,0.42219600642820887,
 0.9848183195541543,0.5647105182020835,0.9357031104701455,0.7565244545834428,
 0.7530309792069572,0.044788117724728016,0.8427776348268132,
 0.5356450319786662,0.8400511533359158,0.9647760963601216,0.5934532949539632,
 0.25104007217953994,0.8053562136370568,0.8390735887427546,0.5809425554967134,
 0.9643534950173733,0.13971675492950753,0.7931667078881299,0.7824086829569284,
 0.07482280290992882,0.6860355382009892,0.25762065805678713,
 0.9232483327775171,0.0843501904371683,0.4435542972458506,0.5637527481615736,
 0.9573211841823083,0.19105154396385682,0.5737723242028822,0.6539748114241902,
 0.5295779224324725,0.537109591761497,0.5978209873232087,0.9803556537040075,
 0.20879084070537113,0.7600310602458434,0.7938163822290705,0.6450520374885323,
 0.7306446891178294,0.5531435278408451,0.44891153366519376,0.7332542085706286,
 0.9984989998994509,0.5258139319105773,0.3027291207150823,0.4526302948904034,
 0.13266152882013293,0.3255820375335149,0.16501815297974465,
 0.21533205610960582,0.20847465114746666,0.735291857977086,
 0.18438090525363426,0.6787552843350404,0.6564913947326827,0.3680517505407944,
 0.07279808984869329,0.5264943342471055,0.6867235070352983,0.9676696612687549,
 0.6195554651874235,0.6702305073335062,0.6883080970071709,0.47121322860995674,
 0.7515093227148022,0.723350700381747,0.45559324570331927,0.9024758090367014,
 0.7491164443794037,0.7979357009443134,0.06798990845506037,0.8596856569160205,
 0.555164993162265,0.8938915940985566,0.27945963355378733,0.0531598160995681,
 0.06060827360466392,0.4751528254098295,0.1956480559651007,0.2275511280491208,
 0.7641985113343174,0.0037219461038350277,0.17058690876100868,
 0.17805632394212534,0.44813977424280316,0.5861382478255859,
 0.9075816281088115,0.6325574470046432,0.31160933922155565,
 0.08216658885413453,0.8768566811222736,0.512459856862201,0.06981982351982419,
 0.17647372064633116,0.7892235725610617,0.7680994348178579,0.2873344346558159,
 0.3509173203673579,0.08126066557126144,0.9336802121089058,
 0.16933708280256587,0.9667082098972979,0.5388715441682939,0.4999717160353494,
 0.6206320621912187,0.14755921426452412,0.3050894381606275,0.5733275102333786,
 0.35758929304952414,0.32997252222371154,0.37754459545560637,
 0.9230116980090702,0.5101826476840271,0.8865113420747479,0.9849138973630244,
 0.20501846229351106,0.23077731314755445,0.15671737827411902,
 0.45516557942591007,0.8997108071470556,0.5898206042058947,0.7878754951848793,
 0.12608048187799925,0.1713863822034719,0.1176650284888422,0.5526506162339463,
 0.3604271913035292,0.8592178076271809,0.3898241043032614,0.8994837452845391,
 0.5715529199394724,0.39061177790449286,0.17519522901345708,
 0.3763909563853155,0.8384608963200685,0.6216131210259672,0.5309289334111241,
 0.5824514429171475,0.09192413104726926,0.7018691703579308,
 0.38034088963776524,0.2836262265323338,0.14365545836702898,
 0.9831902927662574,0.0970062498444868,0.020189573342711986,
 0.2489026049098233,0.7119204346244327,0.38421737806389955,0.8121355017879329,
 0.595031605024096,0.6100322367442311,0.6896959740236093,0.7673486728100288,
 0.19152366292911016,0.10604631348188986,0.28355901823004415,
 0.4129230044505946,0.7811974319196788,0.45054399552591373,
 0.24919806114548104,0.12132980689868644,0.0664904095342872,
 0.38688796597197916,0.3555777885678082,0.4701701269223395,0.5858044456526962,
 0.6022591135931209,0.1232854777853637,0.5559634756205244,
 0.062045406368700706,0.4506186055981868,0.4144589466512426,
 0.19376529679778176,0.1657932185450095,0.09147971201151017,
 0.013157830560674988,0.09238866150438141,0.13601084134572072,
 0.1854468487454789,0.24330866991142974,0.9299792457510218,0.5153810544754639,
 0.7689333157588314,0.2356982363348491,0.2164832327691002,0.289636451863296,
 0.3396730301712424,0.24119506624452047,0.4382276205097768,0.7110277801908856,
 0.6166719270964719,0.3279182113735404,0.8511928091125714,0.7040057682758264,
 0.35506856133610276,0.5512547781190933,0.325362054660148,0.7454666574748359,
 0.30638673676615125,0.7825041019470045,0.09285088042834677,0.992064440879138,
 0.07837482923788852,0.700304928348967,0.1284657638286384,0.6429129739889293,
 0.391029169743107,0.890815905996561,0.8470889802127254,0.8995861922281596,
 0.8126721764407764,0.2258650301151709,0.9376513045861384,0.2922493314060203,
 0.29353584595258586,0.09109939854116311,0.4992436416312628,
 0.5470503031562403,0.827026013219089,0.7315202625719393,0.4285725744586779,
 0.330779021875514,0.6365352639968627,0.7551980394083564,0.853233043272817,
 0.4801183658009309,0.7059704711155249,0.3495689617390767,0.5220090029050525,
 0.3507220974351136,0.5889875253731258,0.8380484114963893,0.8351722920076059,
 0.03691750602947863,0.3910417204767541,0.6697374686279687,
 0.17648859778271286,0.12570100091857794,0.6226731742580558,
 0.9443413218345532,0.8580066142428611,0.18554910379870604,0.6134676855561403,
 0.39544957892882926,0.8087447989746719,0.7700592415255022,
 0.03745801086131795,0.6631683463750838,0.6859112527616716,0.8865067040588551,
 0.5132545087724997,0.8869854576518647,0.6247840146419754,
 0.051636970159196216,0.12139308788422676,0.12076717395997938,
 0.9881036544175525,0.7284205328186639,0.40932390469386815,0.9581353713403357,
 0.16463418449553247,0.24153080266074622,0.08052321034751142,
 0.47671544112510855,0.04877565305465048,0.1881841526887028,0.527903412265952,
 0.11489212270368931,0.029217801656815312,0.5163322840145437,
 0.7322132856721478,0.6283454458132014,0.05138428521707472,0.8849407328691674,
 0.8162050639631925,0.5408478758414973,0.08473356243090269,0.8399271688123852,
 0.5081916344695331,0.6690542560908541,0.2575065065193807,0.16741089931897402,
 0.16856990649920156,0.7152028640796875,0.40757933683043845,
 0.8983126845056888,0.3552259223370948,0.3110192315159046,0.2467062129819747,
 0.37543496536520693,0.19631186413048307,0.7320286619036243,
 0.4592453478261932,0.5128136258321242,0.24056746997497325,
 0.06125526477837073,0.6303239410105004,0.4947341969782241,0.708117487057547,
 0.26589884846654643,0.8419261143583798,0.8643286719533059,
 0.04844832590261616,0.9126929246468606,0.9349962122622748,
 0.17459118224017167,0.022997015469152338,0.405025948365035,
 0.43883597133406693,0.9692829126496711,0.437464591917317,0.24389617995737822,
 0.6188083006709827,0.7380917937350324,0.2644027182066767,
 0.004910871382127091,0.8650596762918241,0.257523641668542,
 0.26950378142345266,0.6923681693132624,0.34606507229634886,
 0.45191199693200446,0.8955354410875029,0.13351984215632118,0.309808381054953,
 0.2764223824602129,0.9619209735059778,0.11973439553975085,0.9907197583167755,
 0.5194395859391707,0.8585770392613894,0.8102817338059063,0.4755492205468028,
 0.5618076833445622,0.908316753871313,0.2653437548625751,0.5811944270899616,
 0.861819044173153,0.8145288106277276,0.26965756111401773,0.06847335864713933,
 0.8426750714007888,0.644910100597401,0.7035154224596127,0.00795372963536467,
 0.1076569697529629,0.706543444513589,0.11207070676606246,0.9374406614800136,
 0.3204237443292771,0.03751189531083876,0.24048551740189494,
 0.27436341670358655,0.030966585903718924,0.1608389870428769,
 0.31639229487136544,0.6114122109729476,0.7576604752831667,0.6860134639888683,
 0.11411210733958899,0.9933070269353766,0.3201532721613789,
 0.030846756466484715,0.9734244779537672,0.7446210214091517,0.883857652764697,
 0.5704495687706672,0.5630914825057676,0.6096465137505087,0.42386470564195666,
 0.8333236727709296,0.42134602042365077,0.5316934857609756,
 0.08918672733430633,0.07111833123542155,0.5534799762100128,
 0.5501488292956109,0.7955705250602791,0.9592008398734313,0.866266550015534,
 0.03036811937293893,0.17030774177906793,0.9322437445682439,
 0.14466557412782466,0.36975234723604844,0.23515086487709413,
 0.8626157715630142,0.9932890098167467,0.19490211604173957,0.3216230928015785,
 0.6984441149542453,0.18794829932040225,0.5100511802509078,0.7584444815822099,
 0.27269280026506837,0.8814562561581576,0.856248672881345,0.20281363438621436,
 0.1849507277373642,0.8448251756094193,0.7933359817310095,0.16147548575936777,
 0.008182465348152812,0.7644015378874104,0.33028804832996894,
 0.04627282989797599,0.8187398655024973,0.2775629496373204,0.4886364574991795,
 0.8171773801983568,0.8854342097704853,0.6925501583720187,0.5841944635993628,
 0.8390429097107032,0.2702788074234759,0.2227943327593066,0.3042065799634244,
 0.8322325233329356,0.9644479327484332,0.5821472774554165,0.41097235681168376,
 0.3512101593557163,0.20629120418052127,0.9543971518250163,
 0.15568895857571485,0.8902267835896268,0.07380930769163396,
 0.048217720065110714,0.16503688223993007,0.2103408451215838,
 0.5070682143124221,0.6399942502224552,0.22043103363322403,
 0.48034657645238665,0.950691036567289,0.5273001828247238,0.24863514437994494,
 0.07840901241107778,0.7450752316894638,0.7220668542392529,0.2408534503934413,
 0.12032078727977447,0.5292168419364434,0.5982342000688461,0.8018980596916773,
 0.6963983553225317,0.1470747810119747,0.35903618878591903,0.5016539324495243,
 0.551694318444306,0.7527526340384016,0.3858783441321181,0.8975862268404132,
 0.7091769758044277,0.8528624910193388,0.9132101618355559,0.1099481791428395,
 0.3500776733394484,0.2955141209422606,0.6949281089651274,0.02068749686167992,
 0.5584063594477798,0.642404896027006,0.4285519371133928,0.9262966912884774,
 0.7534804533843678,0.31511243667298616,0.42363356787505857,
 0.4507495366627132,0.5099598222825332,0.807052588184559,0.359951841338502,
 0.6466348849201223,0.8038747506224505,0.8707533603417345,0.3993837006255737,
 0.6337527193271071,0.6402113065370729,0.14857853030547685,0.4531408516327604,
 0.24967570640441816,0.4725277599387109,0.7934165843405292,
 0.48097007966640537,0.5423968100283538,0.7762340570460275,0.0427449428732829,
 0.5424895597575236,0.33917586804239075,0.1631909264647694,0.4062806506180553,
 0.9918992545364606,0.657613288380448,0.0853624730137087,0.48769689532476557,
 0.7623418799944681,0.8473932208282344,0.28542374946768145,0.8736841887017529,
 0.340421575441486,0.003110479988313375,0.3295181622734078,
 0.13791286796284408,0.8207504484889548,0.473194707687306,0.9140705857887355,
 0.9528349491913207,0.2056212287819147,0.14168033222401566,0.5704165943012113,
 0.7096927861064206,0.27917348535292563,0.5241613792081258,0.9800206383229325,
 0.7761079286287836,0.962835054615535,0.064777972007382,0.7849889918006465,
 0.02718487612065157,0.04507962650075914,0.696543720166049,0.8296199197855008,
 0.24776791156866662,0.821967954877032,0.38483543011751487,0.9038404730961693,
 0.5945184702155304,0.31353607679510953,0.7565830061299149,0.6980463914226885,
 0.2797758837182217,0.3749497483519991,0.2303204284748328,0.34260496616793545,
 0.6664423601705051,0.8456865297718215,0.3327190500489019,0.35707356576961025,
 0.759574453990776,0.7488717918866583,0.6158676623731613,0.8999480049226638,
 0.5757332057309017,0.1683824959912147,0.9423807633764837,0.25657781636823185,
 0.525386528362561,0.4069296870237542,0.8829442940872003,0.779343994433004,
 0.18769325529987257,0.3703390192320366,0.12521347465669974,
 0.4110869422533129,0.7024093855417706,0.051379953959537095,
 0.34882197322526554,0.956401998358802,0.8658844306644614,0.7832358115022844,
 0.7050393851843515,0.045630863722847415,0.06344019695284953,
 0.8358324175471387,0.9341651043986781,0.37415408332556777,
 0.11864866723046208,0.11870262758681527,0.8060475893185031,
 0.7936475950431042,0.3154438906877757,0.28771359197528945,
 0.17140104680528145,0.07014804579312783,0.9987579866501094,
 0.8845803445887641,0.3489818593890439,0.2083841843214057,0.8953743646237762,
 0.2444465627698198,0.809794657937138,0.3361968495246699,0.8601654632987326,
 0.5435543274050262,0.7113275850669821,0.8229699782450508,0.8613097086986532,
 0.3718368930433291,0.06975824500622041,0.8193983208600558,0.7057543382092234,
 0.2830627132358291,0.8343368176592986,0.6102155137126638,0.9228595468291111,
 0.9084175204674179,0.01419470109229748,0.43661014901144224,
 0.09020049006344344,0.1443615875747436,0.9882535094756277,
 0.16782917477726866,0.9008719533297695,0.7713571152524126,0.4293093500806763,
 0.32861128721903743,0.6771732655078854,0.03163211892860185,
 0.39355558657018674,0.7981453704916075,0.25323519814925965,
 0.06350308429278162,0.0137047415756677,0.9665182894592307,
 0.29855211098800494,0.5049528240402302,0.9069042143250516,0.7381224515049384,
 0.3961059079778385,0.2768606217712879,0.2779724520693194,0.6136465545357467,
 0.5434596253384479,0.5423605787144552,0.969158108126165,0.0551978454836205,
 0.1665741992384232,0.39412527216726523,0.4127401190902149,0.9743905561771895,
 0.12385499458065174,0.10599866673342406,0.8973781799136659,
 0.2893521872782302,0.33826965018418276,0.765152398164566,0.01908136226107071,
 0.28583387867335464,0.9713343925509558,0.7222124218090401,
 0.29247911595558385,0.13984161610071533,0.3398051762394555,
 0.9682298416547148,0.009092685230585373,0.16883824392545055,
 0.07442087491311555,0.05080114954585424,0.2813958272748558,
 0.16926645965798126,0.5381524472545645,0.10867608036457665,
 0.37457875296623655,0.20815527996002703,0.6317706110632191,
 0.3759625158922807,0.989738260707098,0.4051337267707985,0.49940173762577234,
 0.16313820795155487,0.33977269782274233,0.11054974019400499,
 0.5286216259690387,0.922628166815696,0.18093572230223764,0.7929187795062045,
 0.2612784588452174,0.6224044750791844,0.7008118296412006,0.7191651109406588,
 0.40140446306189237,0.06843388497874936,0.5587012719855017,
 0.09452689723065588,0.756442080560871,0.69675835755356,0.8338229964463677,
 0.6754051363322618,0.33884558166881085,0.5731131215114831,0.3924779654658235,
 0.008263210193811599,0.31714501231845593,0.6881179568983082,
 0.8258735952665327,0.11816728614207406,0.6793022591552861,0.5852216848670275,
 0.5639456814638697,0.15621409502464356,0.7484271316951894,0.7293464613974902,
 0.6416970556314632,0.41047752699939755,0.07713608700134478,
 0.8727992235942901,0.6117033916476263,0.9132309627921776,0.8909274748238037,
 0.1548218830188829,0.9616324265363079,0.11615634863411683,0.9616375881589534,
 0.10561212193292324,0.2660198525996582,0.45943699044412156,
 0.6990365217838913,0.7736919086862906,0.3583393783919593,0.7279033952835083,
 0.18024381708708104,0.3647970693996957,0.7175937325109907,0.3901396718244603,
 0.5344712071179383,0.20640221327505837,0.6530794835400037,0.2863116549026451,
 0.9632947714154316,0.3558325611866493,0.37120137411195,0.363348150504176,
 0.11009508644633814,0.41618205623466253,0.4932897344885998,
 0.31765357858447874,0.34880113656296996,0.9351581080728804,
 0.7345950639136315,0.35847111820532007,0.21946916510591818,
 0.32075618906573733,0.5220164258893637,0.013655187374163269,
 0.40204020497778237,0.7355262825841311,0.7938421770849529,
 0.08701831187522913,0.41492581273209195,0.9721561018953588,
 0.5272211660985971,0.15060577566028288,0.7152913113209148,0.2499464452915826,
 0.9015411841139115,0.7698952564656395,0.5648406662689589,0.15264968960196568,
 0.5335220231433883,0.5274616894416447,0.7507595867128477,0.6594390077186583,
 0.4909218832789549,0.8491995901210011,0.45828802931202794,0.7356349002374591,
 0.4079052713947803,0.682600488712167,0.0012802004612224582,
 0.16666595495614878,0.5737128815442458,0.581874428632398,0.8706411211943681,
 0.9132940332395686,0.5871400822181058,0.2585448565957238,0.40029470371671694,
 0.7845806080618138,0.5338901305324161,0.844183778789193,0.167391394519961,
 0.4639790527977419,0.5576957270166457,0.6453285917516903,0.4452023352035931,
 0.32547693219494644,0.7760355274277768,0.3765782819059228,
 0.024998656394752294,0.7281825667435696,0.40095647261402767,
 0.6549473562139403,0.17397274650833383,0.1466344031864668,0.0866298509966148,
 0.9349896051956313,0.29296497984915193,0.4534461863911351,0.0876430367590535,
 0.7773281435497703,0.8708463636937536,0.13209462141480321,
 0.20182631486346625,0.05064200242050987,0.802288462144012,0.0743627093301813,
 0.5231696435422024,0.3938677297497851,0.7844221126435436,0.9244514105018513,
 0.5336731294398167,0.46966782149754527,0.4111541229196436,0.1611794502113416,
 0.8971524962787251,0.014107870170418368,0.27992979258741935,
 0.8225995435920137,0.5005773238378627,0.4621234383151166,0.9247236198418463,
 0.16696238777295924,0.27654968227078247,0.3066200895382879,
 0.19030737366350947,0.3084742029512689,0.5809647730566607,
 0.23397048223302996,0.4039251570897834,0.25177083039299786,
 0.6045001305146959,0.06567244667392647,0.027969710960428884,
 0.7795001129923657,0.1370564723720341,0.6292199627359164,0.5664228664562565,
 0.5077701818579539,0.4094880327723516,0.5283806943718962,0.6346020922543804,
 0.421163047653903,0.6604709422446676,0.06845966897775224,0.6036160743015122,
 0.8860296691545823,0.04182191859086737,0.5497485585894617,0.8178847237496735,
 0.9462657562016714,0.8925588418279531,0.20136144294821645,
 0.47901430113510324,0.2411896273482984,0.4236766747762981,0.9595916198366969,
 0.5546305485714083,0.8068135857234149,0.5014776044485619,0.2847892774586943,
 0.8192751364049634,0.8711427444043353,0.3429211067149718,0.6070201497721958,
 0.3618742227981231,0.8148621019068732,0.8712866867553725,0.2783150281921448,
 0.8444587511142856,0.8279954077440643,0.5622378277135345,0.9157494821527907,
 0.7496488807956323,0.8982322912795689,0.5267191495014207,0.06416052303176778,
 0.15754352650646386,0.8008600319352722,0.7616774633891884,0.8143594499262519,
 0.20518489247334115,0.3341628596887529,0.7984149301987565,0.2136668528634551,
 0.1358095747348771,0.2676479023793854,0.16156043934695286,0.7432517185474112,
 0.20251385011078904,0.8395461394409307,0.750573618212927,0.7888379880528966,
 0.5091491530005461,0.5483891648331114,0.5134177519737733,0.5105779097801356,
 0.8760490018705447,0.557227871932721,0.4754024866346098,0.5637696685400244,
 0.13989685647190542,0.9836503037827502,0.8313699351599082,0.5206426393174346,
 0.8844856799832309,0.6435740652180563,0.028154907825420183,
 0.3028191570566652,0.2939474553882001,0.48126775103140484,0.1922846400267587,
 0.19851691633181323,0.9585550259452128,0.995457088081652,0.8081302082324461,
 0.43413062922647105,0.017321083379719826,0.4415900937170104,
 0.41227127871255553,0.26478430244765105,0.4485458578694781,
 0.4392444277212251,0.40665827497351625,0.0926930151273414,0.6196879847523689,
 0.04999533208961926,0.9033688357797007,0.9327491707232849,0.6994861093169464,
 0.06936369084870853,0.5896981303118233,0.1140740951214928,0.8382442097340882,
 0.9135532674202036,0.41376038407444005,0.14400347125842816,
 0.9488097743673823,0.5615306678478795,0.3872078896919572,0.19841689066012325,
 0.6018291491326355,0.21005394298292845,0.9885873151614085,0.2863143179573071,
 0.4292951788859809,0.8694756184950148,0.8793432836722779,0.865249428839722,
 0.7799276944310677,0.9147985648516586,0.7013892267961274,0.8430204608548889,
 0.0906402873359885,0.45780981482199823,0.6724882625286795,0.7096113137528656,
 0.12306276687341122,0.6883663902317747,0.6168755417793717,0.6368997208692603,
 0.26948222793920495,0.6333923262188875,0.4897640101428625,0.5904845895655956,
 0.47543005125591553,0.40848808945581616,0.3978077991812109,0.961285527460286,
 0.9785339646216681,0.7937471175365012,0.8148409594292173,0.10888871595692473,
 0.41384472886142665,0.42951071337055735,0.7517875348208851,
 0.46933167684784094,0.1704530387658788,0.14813918115458513,
 0.2289932677100871,0.2867307921448682,0.724649121373694,0.17962145764850512,
 0.5408089931843623,0.016854258187532167,0.6726137172053505,0.572022186068239,
 0.25675527218404226,0.5096648428980763,0.3297751352034697,0.5378026702737047,
 0.3307756401927955,0.13241392487467474,0.7438782411464008,
 0.35345526332588084,0.1399976277648225,0.9152067776110268,0.6543568604039077,
 0.029127071116088032,0.3293293382396091,0.07541559128754238,
 0.042455933781110344,0.22503709955777762,0.40112525799828147,
 0.6671759101426277,0.6402799452566231,0.9008289278585668,0.5589938356315926,
 0.7829406829983082,0.09621173006619532,0.7837627566318466,0.2913554470704679,
 0.06788247684605375,0.8903792188515034,0.561753959722318,0.9289531261890577,
 0.30993452850826575,0.06595434771184272,0.8071827416053311,
 0.05093033812255232,0.6541098399072642,0.3929727885791072,
 0.18548361038835037,0.7453049716715143,0.8971217543220625,
 0.03609140089577856,0.8784180960027695,0.17204873987450298,
 0.3149428537795459,0.4270532317916922,0.9595628028272334,0.6508665988460429,
 0.7632060157283009,0.8845878105763565,0.12128351425696249,0.8535904190090553,
 0.5319569699237023,0.10284679902217764,0.49074543113439384,
 0.5532970118068314,0.007300126083296554,0.8146732176141237,0.468816152799699,
 0.23016061192405202,0.8992500338042284,0.03277001144757785,
 0.7398211675510843,0.1274225525000069,0.7874102374917444,0.7032897779234684,
 0.5811405321559412,0.7053626608830064,0.12077353061690044,
 0.28253897176503817,0.9222449289840383,0.21158675837438645,
 0.19721404087050975,0.5261626421577908,0.7015073578221509,0.7257961965863868,
 0.2440261962130319,0.08560471945846126,0.5527924545495768,
 0.017273411819822115,0.700089393339249,0.6797550146466488,0.7795887103467427,
 0.29862593919391767,0.9258364804792499,0.44807696271333386,
 0.3067876792957954,0.9382759451768566,0.03170297159523505,0.6023023125280995,
 0.9142691562849015,0.162323357737726,0.1741759624160818,0.9486751597630101,
 0.3348686828911922,0.7278245281316563,0.49736861429395596,0.733796954658985,
 0.4528804781797834,0.8776455070897207,0.7781976630233608,0.15664685537697198,
 0.5603479059882392,0.7534184062067748,0.9248730595289179,0.7373591364750883,
 0.03996536759384817,0.29919913395260256,0.8903550542884215,
 0.2200990217415184,0.4069884890127239,0.1104419833517194,0.7163088582602315,
 0.7891555220535438,0.8332499975619134,0.7308950043750535,0.778055917857004,
 0.36805114382812754,0.6000656214738371,0.7553429616854012,0.692502409077503,
 0.05988444979863039,0.15484609486132506,0.24721093984952958,
 0.8298667137302149,0.257578817695972,0.46747672832571496,0.5344543245703035,
 0.8083885833359221,0.17499621529804799,0.49377244841089163,
 0.3026451525916236,0.9867509254892771,0.8283569356275026,0.27983011416118286,
 0.5576235497781417,0.434293341423983,0.33428658685405055,0.8424324579526365,
 0.8987770899579386,0.7463434463488123,0.41563575130122066,0.8887602775048792,
 0.6006723094618371,0.7539049417595087,0.06430840812453409,0.6967472408465569,
 0.8518498030106356,0.11926741420291154,0.9020841999105306,0.7416136039925697,
 0.8163533827612979,0.5638073982976726,0.20210676233552982,
 0.15329603177907347,0.8686803717211513,0.36936930664461765,
 0.5758957418191251,0.760834271018207,0.33658054722695663,0.2856937771435606,
 0.9580622506123158,0.38131713648474097,0.09947476079781592,
 0.004570886392397111,0.7624695568062142,0.7461511231503121,0.717980576666036,
 0.6859477977119254,0.9019164279056611,0.6886038266783335,0.1933953175383395,
 0.8266871796369163,0.6695589007513467,0.6999924095006437,0.9449785775862652,
 0.9935403413137368,0.30776417901202635,0.19357995173525255,
 0.7971020638544553,0.04283671686589796,0.2879761569861937,0.6335434925835465,
 0.33223588946021365,0.7955999552802916,0.7366121837665794,0.6290542905425733,
 0.12731704129201116,0.34805219818538047,0.17381062093042854,
 0.026537761652477232,0.8729219612401524,0.8865652026956786,
 0.7999896851305233,0.1951132006208398,0.43696680789483133,0.9899654576881789,
 0.13282801646014997,0.6445545535778918,0.3805269740856587,0.4643644332723714,
 0.5422962797837124,0.8773998819998003,0.22992225348195006,
 0.34407603642388274,0.41348133780705043,0.5255877939928641,0.965493320508341,
 0.9743493245923156,0.8783550449907614,0.9949876895502864,0.7751003514069652,
 0.10559008822833615,0.11552685701271881,0.7070217895776835,
 0.3575430024355736,0.15842807544500692,0.5048055997693138,0.4350838324442613,
 0.9331521664032876,0.7826041493259763,0.8170600178503372,0.31336126425355304,
 0.4820973239426376,0.09937366260729774,0.03430463995356603,
 0.5303340257128397,0.6060318329782852,0.003304788212735499,
 0.37594756365915794,0.45914978706973564,0.8226180638283646,
 0.2975456126757756,0.39793355572385003,0.04114216399444304,
 0.5331675975568021,0.5943008335717405,0.3473542516154291,0.6129440028001409,
 0.38913807363026,0.8660541875872829,0.922159966442229,0.20134150734907952,
 0.4490441870162959,0.72875700101347,0.3809154671753433,0.10127818336756811,
 0.15171233915051174,0.2012162531742776,0.7553729094219999,0.9971881821938882,
 0.561494189577502,0.6155628120894613,0.3947130383100734,0.3781589225449079,
 0.7030316564615795,0.2814135502673053,0.8438970250452659,0.588389274941524,
 0.5894880663406693,0.9569518730331156,0.7912655835963175,0.7701899810945387,
 0.5482335604278347,0.6732389954712623,0.8388882460133696,0.23184600742394867,
 0.5563705328989411,0.1785857654541373,0.4192212327289623,0.1032127609572897,
 0.6644234262866886,0.8286644345978067,0.36252806447482777,0.7132081659198758,
 0.9318779892494558,0.5698777656848482,0.6954650473659992,0.08059532169343797,
 0.12484550606442335,0.9562414222990682,0.028634044779582535,
 0.4304665515612167,0.7289830149780862,0.7551248303729254,0.1752218987103772,
 0.4703124318178292,0.6608659895450475,0.5434662556639642,0.5521608036049966,
 0.09065939920614474,0.2625126205711873,0.7162695003157742,
 0.17944880395128038,0.5986601284071024,0.25823234547708107,
 0.4439748056142846,0.671609165953024,0.6527343072153869,0.8766329026695495,
 0.20238944219067068,0.7056366775525683,0.9801268439732528,0.9291599378327292,
 0.5338277898163178,0.43487089829119563,0.3548340366446403,
 0.22852229986765837,0.25169976352578427,0.3602922194221184,
 0.7851924063518809,0.7196751370890496,0.9264095217913203,0.43329765851980295,
 0.57734022571275,0.30380602406955337,0.16488246824288688,0.72973477490213,
 0.8401563447040041,0.4769049105627382,0.18732192583944351,
 0.06325323720164999,0.3002191344029489,0.8318860232050168,0.1382272238626615,
 0.5270037542022642,0.5266907669274312,0.3345693502875158,0.9816736329268104,
 0.38627086132480404,0.27973507496825356,0.10938385210225676,
 0.9411139989898674,0.5873951417967852,0.8533027183836217,0.0615349046394833,
 0.3321231039147182,0.8174596246538713,0.0413308052801854,0.12971462108828236,
 0.9423513349886145,0.5934763600217943,0.6312946734017647,0.11431794130613016,
 0.12961548360014674,0.7405738571338596,0.19172653888925206,
 0.0953224557649136,0.40382446267493555,0.49242522451972415,
 0.7758146518784806,0.7905331377566494,0.2916058757115114,
 0.042085540064312976,0.1682658096047117,0.24934084931218892,
 0.946303635861873,0.17334903114453926,0.6452542136010501,0.18497654439533784,
 0.4659213644865957,0.7412625968657806,0.10967332753503567,0.3468074395218661,
 0.024480200738779323,0.9949709417680238,0.45747619858548116,
 0.20941395501626736,0.3315163347390887,0.6595117255470699,
 0.33010561319450005,0.3910524820219994,0.67946870755807,0.32369901387169797,
 0.24394562473749515,0.8773046800810371,0.6406402622527387,0.247552374823792,
 0.18815247041017347,0.9287150694695349,0.37850143781962564,
 0.7383670258223597,0.03548389889088466,0.876910939133432,
 0.008968748037841645,0.2870867805680858,0.5996729520903503,
 0.34943582886543156,0.42313738299816883,0.1092034695388272,0.918747873592377,
 0.7994930209296929,0.18088507054792435,0.45854097163273666,0.559355294479878,
 0.08475936226567937,0.43671720181700335,0.043534025928179565,
 0.3133940366838649,0.09700428485333468,0.1834656274736336,0.7637199752987672,
 0.708085970018857,0.7599874654341554,0.5633761479896904,0.29702207779885503,
 0.8413959002134246,0.7729069567130059,0.2551894401241849,0.5264999628721884,
 0.21652017623002406,0.7505142296112931,0.40977485843400996,0.513051133013108,
 0.7929726816062515,0.7789976914474822,0.6552684379759759,0.8124466838127115,
 0.401548431405906,0.9954371833024385,0.7170686894643994,0.31986427528292405,
 0.008725923787657797,0.8799649597444531,0.6728607445782963,
 0.8885349732261816,0.6032294618122525,0.0733794744988383,0.9549588632317082,
 0.18840889309080977,0.42267080776723487,0.1976561429730822,
 0.3294721542561502,0.750212869155847,0.291562000833389,0.8529766115340326,
 0.7632652166879921,0.5418278563135195,0.26953247970303373,
 0.12834551003782368,0.9965279776901975,0.1460273734980353,0.4638505119430112,
 0.7822482109748633,0.7791732792313616,0.041443417449399944,
 0.13990491989322218,0.8500844900255893,0.13745587030502515,
 0.09257717617701877,0.7325729913163445,0.22336040818088088,
 0.1033746444128465,0.7742054947685334,0.2227190622597628,0.12771862592795646,
 0.8377641539573685,0.9374208532506502,0.6028328331269588,0.171596403456718,
 0.7009839406302073,0.05114284686942405,0.739247009248119,0.7912932463221516,
 0.07740110310044057,0.13232677587763741,0.9628298210529382,
 0.6795212476414964,0.1164556783921521,0.6369778519620002,0.9119680970654543,
 0.5000135803338057,0.6805473678403385,0.6055516875424248,0.18832357323733429,
 0.5825712877368534,0.8971694423054373,0.3813498878759586,0.1410097240628384,
 0.11297326392736595,0.8943403742001528,0.36181845963161374,
 0.6572406035980661,0.6650620127757255,0.48148979436288875,0.5160559667731209,
 0.6792382313283083,0.1806920725640404,0.9172599102873704,0.9189973000541524,
 0.8043449432121459,0.78360034859735,0.2824780364873767,0.9619877752904449,
 0.5024536149262434,0.48368667473442084,0.2878252448999824,
 0.038400876591115574,0.8737677339951142,0.9718792663751499,
 0.9152911875113634,0.11977045681600951,0.3153995508888858,0.5635119801645326,
 0.9363519029964511,0.5691819402668745,0.44843238093326776,0.6509581173875338,
 0.6069386887804197,0.020692072020895513,0.4478349580077643,
 0.7335718214658594,0.9988226171671949,0.7446567998003677,0.5477529561327392,
 0.8164625273525378,0.6668210308559128,0.18099692756737196,
 0.033303944362574445,0.9012460914479635,0.034610204826602264,
 0.0023712352214040333,0.3532644209488097,0.17751627218588417,
 0.08423884757844025,0.40442284857060273,0.13872216911031743,
 0.8272039515161559,0.21314609096717718,0.7540194736781433,0.8687165538713252,
 0.10730618848445084,0.5809004947955573,0.9848729333105339,0.526823142802094,
 0.07715000141670103,0.6983489003526391,0.2898034200128189,0.59093784624497,
 0.9929150549397454,0.8444842597779753,0.6993562496579095,0.1712855483902742,
 0.7398790817016642,0.8857557342414631,0.7704472099420558,0.5824683884320223,
 0.9849240966537869,0.025352436450843063,0.7275464798399081,0.592565164612801,
 0.8561754683112565,0.2600421051423869,0.32288048276962455,0.5067129793130654,
 0.6857357988378954,0.0899689101667056,0.4259635625417585,0.6868963110167615,
 0.4097236586081897,0.4157137100823799,0.6573970153944806,0.16199094370353917,
 0.1507464966331692,0.7181990176648688,0.6659095969960808,0.6313408301271695,
 0.334928511091787,0.6336342088261672,0.19949543238742085,0.7794469730323653,
 0.5214826470045204,0.2038525977381167,0.05064267672192191,
 0.08815594859941067,0.4963330256437707,0.42891697600082435,
 0.22596606740078695,0.4467804992679103,0.8593786593948771,
 0.17663060776094963,0.020512127779707345,0.171035670641697,
 0.7757263255168938,0.45230007561584584,0.8426079309960169,
 0.40683786894445706,0.9979364290846936,0.17969651396512398,
 0.26152564765666964,0.7407917498522707,0.8723509472947003,0.9120142608987251,
 0.17281351673440448,0.6749482652778186,0.7799913379587707,0.9302711691577994,
 0.46166945745337884,0.35534946587458915,0.8214079310150086,
 0.6514244904998198,0.3477215496242446,0.1698261137546866,0.26904353425398764,
 0.7911021303761365,0.4056785758513697,0.12147940856876172,0.3392874570175546,
 0.30044819597487893,0.8015848512435784,0.421771476683655,
 0.026603308639085776,0.7097937014721966,0.19692617161066572,
 0.6231188264607227,0.48650294727006116,0.21019650538974954,
 0.7404737282397346,0.17323367920712374,0.18458629062411436,
 0.35853753164643953,0.7969131630933604,0.8044576634679708,0.6544084450882298,
 0.8520764845202121,0.060392564552971484,0.07174667220267295,
 0.08363305346225769,0.43985176427027695,0.36061242563058915,
 0.27808327956427537,0.2658682477058958,0.5433155846983222,0.3471974847981356,
 0.7144206335753346,0.8317018232569502,0.6429998348294921,0.16611586511208876,
 0.6590551957905526,0.34829120361323507,0.4179465440252983,0.5104353393738598,
 0.9881324704185717,0.2892313531930186,0.25498206995044337,
 0.06156027905268846,0.7856531841584538,0.13825832784272551,
 0.8221117990250533,0.24533104067830103,0.5860492317688206,
 0.08043526808443724,0.23994605610841258,0.5523658525967474,
 0.030638881006541663,0.3795220203322873,0.47453592870472217,
 0.13505227684604715,0.3384478722018738,0.8917585004976912,0.816456527741684,
 0.25323975798606446,0.8574873952237895,0.6953730634480857,0.3635984461962012,
 0.6603496852489074,0.6326060003446814,0.8951613597417384,0.8700867159163541,
 0.7914243389805306,0.7511761563069985,0.2616954261571689,0.08836665700074886,
 0.20195106908438687,0.13113840754403927,0.6721433875031758,
 0.6807602365533322,0.768101667235957,0.8819660509869747,0.8403909615875518,
 0.07331196513192051,0.4390362204071444,0.3084978390591947,0.848567550922981,
 0.40292866125161364,0.7700099234192014,0.47939752085876863,
 0.3990963749874106,0.9548220107274323,0.44760742099952144,0.8425443365847771,
 0.3327725886856576,0.594193565692051,0.8804798853042456,0.5762108138433024,
 0.20137364131578328,0.691268382287534,0.7272032323698191,0.7949141629158857,
 0.060343757010694055,0.03515924886553423,0.7540055237845299,
 0.4616803893168645,0.4562871355576936,0.06829677841028126,0.6842548083749108,
 0.24104379254431996,0.6541198283566042,0.8243604867549513,
 0.14323323550719613,0.14427743011222016,0.38133497159003127,
 0.8331021191577386,0.07255933520830338,0.5803876346646218,0.7190427852840704,
 0.9205993652924356,0.010040436485510229,0.2904088105290563,
 0.8230868681005682,0.8733239418623961,0.6192287482951095,0.8698302084730801,
 0.7583776184149467,0.8314791290985318,0.7472287038555983,0.9287710679393668,
 0.15688281691086836,0.9386008771379962,0.6785411410645527,0.1391077797901592,
 0.16097926428583476,0.009936105120264327,0.4076937546218943,
 0.856400994095237,0.37583898061753085,0.7391853037046313,0.3084167480950396,
 0.5101885422311211,0.6074785367233004,0.7698077448061325,0.4285670253945685,
 0.29761272340911504,0.3081899102205382,0.7231002869634703,
 0.17741660373506762,0.6048842041189735,0.025761831725574336,
 0.9765895085335925,0.28321804375655335,0.08651735024542184,0.943943402618558,
 0.2448398981170269,0.22077565808897892,0.8156381997549383,0.2528802405730227,
 0.8220187179795896,0.16181943205428584,0.3984472914973697,
 0.34693734891495553,0.13784760673686602,0.5184708946752329,
 0.8416285950698301,0.9265334378904195,0.6023216236412651,0.4942503222112056,
 0.4521298309682661,0.6811545116046271,0.2705896516886217,
 0.005094803424680161,0.5097311261362777,0.08240001319165136,
 0.8718257151288809,0.8954139728282806,0.25405637431580863,0.5461085822724175,
 0.6075448941225302,0.5301716159246624,0.8028915389797128,0.8604210803424692,
 0.7440049749596488,0.5690758262640023,0.7025191666891674,0.46226870301398393,
 0.7870082511965056,0.30575600944064474,0.213767501270261,0.52993970337655,
 0.9589029128300535,0.6226552444196651,0.6705375316122869,0.8733862266386371,
 0.3679140250385634,0.27812818843691,0.7846297872909104,0.056574638439103264,
 0.5769450816784282,0.29206266473870257,0.43816293884205715,
 0.13661462979831984,0.3107820063957275,0.7479193545685471,0.6326746809867306,
 0.1886716846383003,0.04478706459237369,0.042453761421672565,
 0.8881676527306205,0.5272025887567408,0.4740389966665206,0.9867729894280306,
 0.42950608283998293,0.7565511200625571,0.08392583597212488,
 0.7401034872272939,0.03467313183064469,0.09235864800937699,
 0.28388778411194204,0.1274527769769107,0.9987522986840114,0.3928558936165861,
 0.9983776355151373,0.17252088299584145,0.08086791928973702,
 0.7421834290953997,0.7408831775473859,0.9081466214769559,0.11374627136459126,
 0.9888206745107221,0.7999972558035879,0.6884281328341715,0.07452136189716219,
 0.27874204383647927,0.9861408767303447,0.5730005578306769,0.2565692362433035,
 0.8067689085329073,0.7538927462108547,0.11878448281640908,
 0.15698419898508187,0.6092970523562655,0.45210840596171187,0.855085202550426,
 0.9714257811028368,0.5812787680893599,0.3953712391965658,0.16433143986820486,
 0.707747831591743,0.7259923711277839,0.02188777741089254,0.5105240231476461,
 0.2563035287211153,0.5580607054784378,0.7591480527553508,0.42789589231964253,
 0.06186030134641263,0.8753734266348747,0.9293941352498784,0.6416351127888427,
 0.6614037951221842,0.8446928642469822,0.33520673495513453,
 0.08415614801030658,0.059506008361243534,0.961136529469273,
 0.9460556558893962,0.9629386208340834,0.7038753191566139,0.12721424933628145,
 0.15029771231166145,0.7761425552638824,0.999879987491382,0.9950795496647219,
 0.23934096138596606,0.9390199391060406,0.5217711138970877,0.5032029639440754,
 0.7440249539028982,0.955568021220134,0.7968092875733613,0.778065029908489,
 0.37139028760219917,0.5707660049885829,0.23721583916457,0.818077162230468,
 0.42525510741804584,0.00383251229926751,0.8095406134732652,
 0.2682424865183557,0.7377321791666129,0.8703709485667331,0.07623571969903553,
 0.5342922665093012,0.6635502737198388,0.10842124831253441,0.5634621282326784,
 0.3053248063910843,0.42901045922258263,0.8913021307983628,
 0.25397636185158323,0.5620243733933827,0.5832073853495295,0.7604731961438433,
 0.38438487549845113,0.773968049227321,0.4649703652818109,0.04261252351414557,
 0.222175695624256,0.551916499188896,0.08639310310323123,0.89988467794504,
 0.6305829971567795,0.0451699637150047,0.4946127127933194,0.6002931132835788,
 0.9530609530339257,0.08867105749762094,0.3550783786126017,0.6834635026403318,
 0.5550641852151743,0.4458372803840067,0.3902329079775173,0.16325637168653395,
 0.30073218562870896,0.08667505503616557,0.9518089770281889,
 0.4580574720847991,0.6906570243712933,0.5735211003965649,0.6896034373204729,
 0.7211602593192541,0.3931602277851868,0.29807308911134633,
 0.49528195231718763,0.9863642206369334,0.9705698240649876,0.8788292452876164,
 0.1578916975648299,0.5616357398470786,0.8985876937736816,0.8818591071073321,
 0.43257921612967287,0.313378856244259,0.6643580223584953,0.7916460129766485,
 0.7777945743022727,0.9367087489158541,0.199073323548845,0.8693073492532672,
 0.7499160644675751,0.12695343415670357,0.8540952083392738,0.9882905906727155,
 0.517138764098969,0.27066373661870047,0.6953736647999239,0.145849056371269,
 0.8099341369457291,0.1225777121015017,0.17382521752164237,0.6486847752571472,
 0.7831515882662127,0.6396928144524214,0.06169889458377953,
 0.40678185751696594,0.10528770630084017,0.6807612605155131,
 0.8795541268632893,0.8886055044052695,0.8272844564769667,0.16557763088297095,
 0.7213433916081232,0.03247671354232684,0.573248464932969,0.8841242719263974,
 0.837016792778329,0.8736408073212654,0.8023001329281851,0.31330027698081053,
 0.28844241177745444,0.8819818935559992,0.2750500704225648,0.6195384786947808,
 0.6115299324843761,0.012910106259934717,0.7160354212554804,
 0.8695282286313678,0.7865142979099753,0.12433163436158468,0.8802424680149875,
 0.9549784455303876,0.7097218826562728,0.41793685124791513,0.7558017294613728,
 0.2255836431212429,0.33751080347000384,0.2784233826254505,
 0.01510819026909227,0.499895362925012,0.45917267897222513,0.1636012799899479,
 0.5598586191038588,0.8862129056665898,0.1755589853456868,0.853231901682991,
 0.691208040645448,0.13013447430317016,0.03298366193244506,0.9235805289943384,
 0.34683433506803674,0.8623260505608481,0.446003312868534,0.7339483256609395,
 0.17055281428039493,0.00769779135876103,0.6304846174894875,
 0.5264925782277531,0.4586489683069028,0.5265128488763307,0.18179270696038086,
 0.12722043050475873,0.5888582773060018,0.5585423156583459,0.4929525284136804,
 0.008596467087554571,0.576533288553264,0.11322619537530187,
 0.5938665002181844,0.8987301579024021,0.5343375424666557,0.5408648388602509,
 0.816225199669535,0.3356922247450821,0.7971994837321132,0.026952280823238084,
 0.9707591978638564,0.8415042482660329,0.0773940582912992,0.2456283228843924,
 0.13315407920326994,0.7347952812017089,0.6547042815410377,
 0.21845949205546822,0.8450868073681413,0.7718246721402154,0.2674170063140696,
 0.12103897853807766,0.6743710357488567,0.9710700080661148,
 0.11646349925086907,0.043685438335154925,0.6879898066743344,
 0.6610963253486943,0.951803813042278,0.0026590251281572996,
 0.6200661704418372,0.3827273546984036,0.14159264397996507,0.9002100789659467,
 0.8449268609367844,0.7998376920277388,0.8963950029740648,0.684687322004089,
 0.5563331372706521,0.99411574645217,0.0529890139392305,0.013018077866181144,
 0.9660442592016814,0.338687815668299,0.7025559963403274,0.23242099658554627,
 0.8635555823846267,0.7486343918358543,0.7951447699356177,0.7472264516067808,
 0.5641703526704344,0.21836440978191984,0.8893474561137117,0.4764561377936475,
 0.04169973021834683,0.7934985678780756,0.35316835255649837,
 0.30309999074355454,0.6927873927028746,0.7188101610544967,
 0.28184972407123354,0.7640731807635254,0.04182424317125488,
 0.20044908190879518,0.016611933973747384,0.7271502071064849,
 0.7430329670171172,0.8826946220734082,0.8885497749408414,0.04004948682139731,
 0.13612828556310008,0.09115004640413282,0.31781985251819334,
 0.6509452873427801,0.5164633164441884,0.3419856904702674,0.892088024188909,
 0.7331380956814966,0.050529909685785857,0.06494934863796886,
 0.28766671284208556,0.5998492194847194,0.4070295434445309,
 0.37200096665932614,0.38467151818183876,0.07541222756551835,
 0.9334200401267303,0.42559425270734996,0.8656521300236416,
 0.12187057121522327,0.8833834657030253,0.3819341547683238,0.5931974700164504,
 0.7185964144861395,0.6790294978289492,0.513658305013819,0.4254048430889401,
 0.18364875641558887,0.44925812976665647,0.7872333869206652,0.691100278258039,
 0.9899349288265584,0.16601904462216288,0.8809556078492158,0.694633426513114,
 0.0049245234274910565,0.25284654455127376,0.08876839317685525,
 0.7562871537618604,0.034794462234601586,0.7639123724479341,
 0.38779170029521115,0.7509203798311433,0.297360340834637,0.8456729604529075,
 0.30999588920466725,0.11496355431048366,0.07476574156545834,
 0.46708873728865397,0.08400093167565337,0.625415033971472,0.5723607852967848,
 0.7602935691504623,0.5704899558537212,0.8787532473035334,0.21790892325632938,
 0.38535421753332155,0.3854608095350105,0.9279455587363388,0.5998069025254292,
 0.8566651438561949,0.7606681144713204,0.10986329515564264,0.3540009583086702,
 0.2582451727792183,0.10193618725192588,0.31375750366966515,
 0.013427221876569417,0.4482600310685282,0.39965300802841275,
 0.5132526344220506,0.7663254218414097,0.8538958582873881,0.8115229860611044,
 0.08367079578237901,0.2888552350006318,0.4544583493368426,0.8548529273415288,
 0.33356340346816227,0.21674911392492469,0.8254709942462979,
 0.29256295558769496,0.6741007384090091,0.23172876058710334,
 0.9502644754864861,0.1786604031960224,0.7835798111589527,0.5204936378042011,
 0.33573818113239184,0.574359595964418,0.5209077234969928,0.6691596322505402,
 0.47383671197553057,0.19482986176880612,0.2608015424451847,
 0.27299918867783013,0.79710080524843,0.19145612099749165,0.450725059858469,
 0.655149802933913,0.13826057741514508,0.7659649617701806,0.48441874473369595,
 0.09166633388373102,0.21735955204293378,0.6499314015422814,
 0.5165661469428016,0.338242786255066,0.595875126627239,0.4064460616315413,
 0.9991173109217215,0.6708874396263381,0.4932432436845553,0.15984704149351736,
 0.05968070067382891,0.013043327033883223,0.6363032963007296,
 0.8972297403552489,0.09236686383497694,0.9721669001848855,0.7245716051724991,
 0.9155831834624382,0.9212117066182899,0.5448593074366971,0.541762828706218,
 0.7178998907555412,0.9275130663522304,0.6031319416730233,0.784906881514849,
 0.8215084873593197,0.16644151534874974,0.31167058531571246,
 0.6581143824236123,0.37709033416208093,0.6762476000357158,0.9579201858517482,
 0.9986393957150828,0.9764561300537663,0.5868462997787461,
 0.043330434887198344,0.07198885571858626,0.8901146353442863,
 0.9018278326006023,0.3611553423175986,0.7554894349774606,0.9477338988632523,
 0.9286682102177114,0.6433728996293329,0.6320195769248717,0.14088053208581186,
 0.07469055479728848,0.8065802343257159,0.7463700214658578,
 0.38787740605593135,0.9067500857068234,0.28505483113021035,
 0.02875577239558924,0.5418117827484622,0.28449457110815934,0.57954823021695,
 0.560908582498978,0.28086379695154584,0.13319309515332733,0.7792761852941454,
 0.26381366077443413,0.5400730031560856,0.5245169638606529,
 0.49958072491209815,0.7603903098609074,0.18273938854128424,
 0.5387153940920903,0.7400522116718931,0.7723599575106986,0.7572802975293925,
 0.8233734849978057,0.5796450290608587,0.31311322159502375,0.4994459035597434,
 0.9273995465601396,0.6779657323907936,0.7967945114451613,0.7292511446504092,
 0.3439322366756268,0.7142832889473671,0.6108940226769587,0.3896334510646041,
 0.6291771182447524,0.403766516397182,0.7868465940708785,0.5829418234820364,
 0.562983375209889,0.8962194626710198,0.2413743292775815,0.5168646890025782,
 0.3553195824239721,0.33489608863489306,0.9325066159055347,0.5791671733345531,
 0.47639006488866786,0.14219380327177045,0.11271147442478435,
 0.42659188947920645,0.8110935930593959,0.41197561647668735,
 0.4661555011096773,0.9762477781836749,0.6321744244632077,0.3491391272858827,
 0.753551611608471,0.8718316331354428,0.37649765346833686,0.46528136626586125,
 0.4119768263951775,0.35432993579829986,0.4476271078580778,0.6520850159503638,
 0.5693139634564955,0.6077685729381919,0.31646607057120235,
 0.12217141427972167,0.3332059050749071,0.18041081476033483,
 0.4236709006575965,0.29401654635526675,0.15764784100835516,
 0.9240981968026746,0.6904833855813876,0.3897709428771068,0.8935725549629061,
 0.4047958660695624,0.9626397083317227,0.41224180740380445,
 0.04371052213684634,0.9815031411343851,0.8157349726518931,
 0.20883561340329349,0.015019177342322831,0.10267837530502744,
 0.5825034519475197,0.4507514049769924,0.2268221525179427,0.5761513563395808,
 0.4482056547631502,0.18182997851328186,0.1262066203519343,
 0.45155909553630325,0.7407956035698859,0.5408316291201717,
 0.09666475902681748,0.6451090191406644,0.3837886022404665,0.6297573137619442,
 0.0392735574589389,0.3012826275388233,0.8286269998555416,0.1456971302649842,
 0.6258742674277227,0.27668602502429973,0.25513943685470464,
 0.25843864436627717,0.5076542479689427,0.8210622785337373,0.8740102712315437,
 0.7172663305332722,0.8711246958610475,0.07224658397812722,0.9186676975445274,
 0.09678476332908359,0.19355696560366253,0.4849623671131241,
 0.6225896827941485,0.16703849506570334,0.15294890082823698,
 0.9453022638651911,0.18291084575898786,0.18532085453591296,
 0.8725864740557424,0.5457623896575036,8.433094772651861e-4,
 0.8518860071956719,0.6642905797665036,0.7391320516718606,0.8007793137503966,
 0.336128738606438,0.8187842908153119,0.45777350817503293,0.2556121560398922,
 0.786240701437495,0.24602869124038906,0.9823663310376802,0.6632650860703188,
 0.11392815859108285,0.07958969192543108,0.6217098898632337,
 0.8831086375964743,0.8056872285683387,0.2583437120305566,0.6907590501008938,
 0.4761865564994746,0.3314037727904111,0.49583687473580573,0.7567614878964526,
 0.23886413368042447,0.056395706651462874,0.09548257569309138,
 0.8002346787052437,0.14680776254630334,0.7440689807367381,
 0.12456232996004946,0.21934143353074675,0.6465690453150974,
 0.2505787077614985,0.015610601565779092,0.3535336839079669,0.772271296337256,
 0.5708362567021003,0.36248026934151223,0.4826275492323412,0.8844924693028008,
 0.3428838662706648,0.49234251995858336,0.468591656226114,0.8989947706510065,
 0.3265348111098273,0.5778356955805884,0.10140135125572236,0.9245776879622923,
 0.536911188737624,0.04668337385958177,0.5031461980197912,0.8381645086542401,
 0.7058087072697831,0.2542891712528558,0.1617355213080608,0.3505786706142201,
 0.08183361642606168,0.05265736473176674,0.014151471386211423,
 0.5142582548902701,0.8045453418287754,0.8315174100395949,0.09249086323805922,
 0.08383234200047718,0.1660647853751347,0.3778572861400544,0.3016621417736234,
 0.45812062819258004,0.6718217187219683,0.11812746308242572,0.87179550442584,
 0.4802677231520279,0.9656647861322788,0.7898460304532234,0.6013785468655838,
 0.1241457063694662,0.25831704391964716,0.7149595830765441,0.7175278741003643,
 0.9546569919077157,0.627683355169177,0.9307863450996776,0.35370455497613806,
 0.02724116593248893,0.3247458130550015,0.23342375097478407,
 0.04713065240909642,0.3403206183806933,0.2542976961616472,
 0.06274596344866945,0.49504395716928906,0.9238780465568966,
 0.19998524955664843,0.20989706603583624,0.43638644539116245,
 0.8164985460565208,0.5167295852029322,0.716969075075036,0.4867884677250004,
 0.8460405310028704,0.45689501805937904,0.5789961825911324,0.3788008389958043,
 0.4253331454699971,0.17267136847397446,0.7308519680585848,
 0.21240030131024068,0.6951497884287525,0.7001644393572155,
 0.0024802382360794573,0.6144803134772252,0.5452401551338792,
 0.010098807293492218,0.37352588788691166,0.17911115908337794,
 0.20968257657619416,0.09460545999552483,0.0649452322368339,
 0.43567620151882125,0.892754879823677,0.9080503373266178,0.654211433571237,
 0.589238678725763,0.28247212940167477,0.17721389777110308,
 0.30638159003508514,0.8728315923861882,0.4642404202781024,0.5535947935030084,
 0.7750284467730336,0.1303396185690232,0.9228447110438778,0.42397169944414514,
 0.3438622886527001,0.9329392284953556,0.95144358534825,0.4112231577661907,
 0.12987982694225864,0.8351243705145008,0.9369626164176875,0.872126749106216,
 0.25847278944378416,0.39977051293672383,0.3506237163870669,
 0.4134064512677583,0.15561929618985781,0.9211283368401468,
 0.28010417592602854,0.898363745018896,0.6200763779555383,0.6619186622518105,
 0.4352430829206533,0.03025566992064843,0.6769144044546659,0.4906767825093611,
 0.7705785336541513,0.6488029566072107,0.3406500573801545,0.846735002511371,
 0.33702890446833267,0.8895556743212183,0.7189344492894807,0.672403299054276,
 0.4345760187913349,0.4615974760735808,0.381472901599056,0.8736915920543021,
 0.38158524350085243,0.5184003653731518,0.6597967578154781,
 0.39787346454644046,0.789775254007497,0.09446228180455196,0.5941021789122759,
 0.25908954400502215,0.8545232141595394,0.2191557152479403,
 0.37806247755286515,0.5071840757548728,0.30340927984523747,
 0.9373629557512376,0.7651568481104843,0.8187424977334551,0.7819965518126902,
 0.22917201719876445,0.8136320006504161,0.5518876412641374,0.3955625898519184,
 0.00795622084864489,0.6136790534213378,0.621503293460849,0.33414194222521787,
 0.05472299952302295,0.13526857925840885,0.20485814909810296,
 0.8940233336645393,0.22527225852569843,0.1186480196924996,0.1297854757121062,
 0.6600904443121977,0.7889560167964955,0.9699199195754152,0.5353408076173584,
 0.9272021008962618,0.9145722578473487,0.11441067073892497,0.769701706097268,
 0.5654008605754757,0.4867771269536497,0.8435189372867867,0.7830136119808118,
 0.626985423109272,0.4620103322302641,0.09357007168416098,0.8380481393147402,
 0.8660043132472588,0.44296877254309786,0.7287008735322661,0.2501816836774984,
 0.2912352760791618,0.5983137486882876,0.552406138515598,0.6838362671257946,
 0.8191866245801278,0.01516590072132118,0.9740763780398255,0.54400841676095,
 0.08632920420032564,0.07710364920348123,0.11724094132797325,
 0.7218479933476936,0.4453222687747693,0.32655988532811,0.6711436920158397,
 0.39976865617134694,0.5110858407822103,0.6176765718654649,
 0.022461786746219126,0.672437115252152,0.10508967093573007,0.894472523219058,
 0.1837343159847188,0.03736730526681464,0.8414176773851545,
 0.047860774436897735,0.476706714509436,0.6260203018570419,
 0.16363730109700414,0.304087774471986,0.18169609487326333,0.7851580806719898,
 0.6390141465145186,0.9453367553817585,0.8573192593162918,0.8321900980957649,
 0.27430712840059857,0.23975037719073866,0.3737203118798461,0.525574769332428,
 0.777979514886526,0.6211340326848114,0.5406838459471789,0.9396294399070886,
 0.17779108823751488,0.3642110166037624,0.63830318917469,0.9158512665300542,
 0.40852222235396807,0.4143607338302405,0.9660021329696343,
 0.06456099690934491,0.9951875053887675,0.5386219236702474,0.5304890044782219,
 0.06114506515831131,0.04942481637398799,0.17579685155632874,
 0.710559501871309,0.5927183582313242,0.7834476627574225,0.1527951015163257,
 0.490690843547288,0.020328441779634443,0.6600615795596502,0.7279123679469494,
 0.8284094339099716,0.6819064193527636,0.060775481856749636,
 0.8199524951310608,0.7479319551159652,0.15790449729461753,0.7620368073724372,
 0.40806026404199025,0.36975270981830877,0.19012446477228473,
 0.30528329624997974,0.6589758291895362,0.6678545936163232,0.3543828495672344,
 0.5904920205376063,0.5515080701250471,0.22417289171232624,0.7710649679303934,
 0.8809791726103398,0.25327140445203566,0.05951354822387267,
 0.8801731539928013,0.35036833949419666,0.57962322623294,0.7974637116283905,
 0.9873772430346016,0.6264448746622053,0.8123724280219145,0.7060939389866882,
 0.9105248315165557,0.659217270705393,0.192270741714186,0.13432341841249884,
 0.031674698412277325,0.0944389111405135,0.25318767909354123,
 0.7280821408444451,0.6400942361134496,0.2055325744901776,
 0.004905650829770902,0.4550167313374195,0.33939147750725895,
 0.7832994026453795,0.2298006344353697,0.2295625712318533,0.7520765174495174,
 0.9133671100076299,0.4415142996791962,0.764040589293599,0.6796841658510216,
 0.7237300103122688,0.9260509271011229,0.6643242499655097,0.04621847658052036,
 0.3847855127554177,0.541517883658534,0.9367349150493213,0.7456364684349492,
 0.5421351966563559,0.5090978356230917,0.8730739197540603,0.7685047163698371,
 0.018028402871320903,0.47887209216898663,0.5030563184646969,
 0.23783864741791128,0.1340252552880441,0.00466218316602296,
 0.24646822764133625,0.46005370575485627,0.04468778495785841,
 0.6648641521211753,0.03486751600459215,0.6221313163047286,0.5854857338670446,
 0.8178468177664868,0.7581707974726765,0.7176247674251537,0.8971910510989827,
 0.1832464525396541,0.30715111348363056,0.6407904500357076,0.7866631795518506,
 0.8944233519607917,0.15807596703340387,0.1856605063078911,0.5331037894366476,
 0.9412398011312951,0.5250381487736613,0.3099854119721951,
 0.030526472739294208,0.6484082966831518,0.1527288822722742,
 0.5547657738352395,0.2999597483061427,0.028327475077900677,
 0.8135326766030994,0.9693499699743294,0.5958410809976493,0.3820572329079659,
 0.7455410136861051,0.1026982444902893,0.2415791960935062,0.6624300424099102,
 0.766872342165964,0.6978978178948383,0.5950654830345139,0.4410152579429676,
 0.7817792657130255,0.9085244249537466,0.6452065495163639,0.34832079245909076,
 0.004393540025244502,0.3568339277372038,0.5517815572181257,
 0.9044387703121105,0.31709731939379193,0.7587585712860361,
 0.13626693699426318,0.1743647266964744,0.9128589764184247,0.9361081781180018,
 0.698650747007689,0.0017832533838141318,0.48396412427059965,
 0.05702607724480424,0.19433914422457776,0.00921574601348607,
 0.2686738256099357,0.17534739380909692,0.08758614788015773,
 0.24379289572241014,0.3633318978641854,0.7446472058390298,0.9062627097016542,
 0.44928824161280556,0.5940667327006585,0.17270223563501352,
 0.2566045034806478,0.19520856875948345,0.5788085131232519,0.907863915667308,
 0.48815100657456834,0.2659956620523489,0.10325230047444889,
 0.9926708574441495,0.7323864476026009,0.8745391775382974,0.5886693919562216,
 0.09185607245379157,0.47465040019608085,0.3019681993509059,
 0.020312282450337338,0.271771506826642,0.628396773502823,0.2526621494239808,
 0.05991532367663821,0.583555207146531,0.3301760082449745,0.20294799041850298,
 0.7739950672468958,0.35456123088734803,0.13359741039443085,0.658718595260206,
 0.25648762592469965,0.9038870741588545,0.2601450472177881,0.8215582072155339,
 0.5144099285176902,0.23805997478806196,0.18742373034545046,
 0.7492404162603719,0.43378635119969144,0.9752902806061546,0.7938679702817144,
 0.6235766640641034,0.23013861518274947,0.1255009338234503,
 0.011665402074532039,0.2913304157307871,0.8820177706088279,
 0.49518433885786484,0.9511990845992503,0.9839372358806375,0.9010762538349137,
 0.6296691866250553,0.6307971941875816,0.07023976268596943,0.4856993693059807,
 0.5888218621073764,0.10942757255610402,0.22367280363775233,
 0.11080626253540027,0.5483761054668866,0.9925117381529693,0.8232202923732335,
 0.666540314074292,0.6961445651408743,0.2631691907757857,0.8910049209721458,
 0.3644480144957585,0.561930534630451,0.8148624205206075,0.3261457631370124,
 0.11896784126857307,0.3451250794771098,0.038929874775209594,
 0.4740680582636936,0.38647686245496726,0.27672808524197656,0.485678495639837,
 0.1517901592019837,0.5893025947951396,0.026945090801083094,
 0.3285866789023937,0.9516112204253152,0.41017803216676996,0.4912427328633491,
 0.4104825009245876,0.038102542676400564,0.05191078766848911,
 0.025690299002740113,0.43741561077262137,0.31696589675698905,
 0.5033008721791338,0.6251208022206942,0.6354132824647919,0.3114833415744631,
 0.8931020419663992,0.5663246650164866,0.9575534046730062,0.679591541201153,
 0.012000550750798644,0.3866621948869422,0.563312802827788,
 0.17956476020697287,0.9249866674152493,0.8751774964292472,0.8665462347713637,
 0.653684441190419,0.38673565987485203,0.5590452136684185,0.9119402070613364,
 0.3936194327440511,0.7857452247206238,0.47821183006972234,
 0.10632734422494661,0.6588417088629309,0.74731871701121,0.9303191433495451,
 0.45895453518171336,0.04884900599570374,0.591445957059547,0.8521461949770767,
 0.6802278244183104,0.8034330183276435,0.6759148424966013,0.42925991471748703,
 0.4717530941437308,0.425428130079045,0.15015301684344495,0.5326812436041509,
 0.4873658242381136,0.704016023800774,0.2834358668663777,0.007566110499907852,
 0.6964771377006177,0.3364238717089245,0.9368772158153197,0.9605008995437674,
 0.4247409766750706,0.9115028718127203,0.9880068625681462,0.22876360684772168,
 0.45070990366619246,0.7125274302900076,0.5155464896747295,0.8829098785450389,
 0.7854610204405776,0.332413314219031,0.1502234826729233,0.04241631695789794,
 0.8646966249524417,0.052773805456609924,0.3218417136420104,
 0.8865237693360405,0.625994508346513,0.7068527532308452,0.4771569121695237,
 0.824629516497,0.6085657571667402,0.8868445473090311,0.539413906419733,
 0.6059393448908607,0.46434725302833924,0.9191829703562017,0.872430794670867,
 0.9118365449717455,0.620778702619144,0.6546528547295687,0.6106641928105081,
 0.6604551755026498,0.9019691993755814,0.38875716604759925,0.6846871732727454,
 0.6781182574340031,0.695689601036012,0.573110115578884,0.9833876542945231,
 0.875491824601051,0.774450748622006,0.9092489531370012,0.331949878357502,
 0.530437696208755,0.34420965456671315,0.24078385547836612,0.5118565380508526,
 0.9410124950651644,0.09384947453105008,0.35726864106827205,
 0.26645921283420115,0.8106285045415924,0.9953811004578598,0.8646981377822446,
 0.22223963793006174,0.8456320922039904,0.9688736943022626,0.8811815714955763,
 0.8306336622493272,0.18011956865651257,0.9745601896029503,
 0.49219170830798076,0.23211512985022997,0.7354005237328672,
 0.7779843480647847,0.07777274851348459,0.8711341939288727,0.9778937014096404,
 0.9619738414208996,0.5110445337154836,0.1756797099171602,0.859431131173127,
 0.3219173242583988,0.6398171621706774,0.14701399487680578,
 0.23541429493609511,0.44320017280974666,0.8425272819241093,
 0.5305419902091871,0.9410923437882602,0.7005460625990452,0.9627810170596114,
 0.45461466877001633,0.8767934392461747,0.02008925739569367,
 0.13072027288072963,0.08536740102163742,0.9082262417997132,
 0.5517271200609825,0.42183592078078813,0.21415984837290925,
 0.7530829944165078,0.9484631689477014,0.7717280952951783,0.4854318086502415,
 0.8516931953662553,0.7026228862665098,0.056210526581244036,
 0.05861416568732336,0.7716891426404744,0.4889916252585622,0.685420535316636,
 0.9901776213568533,0.9864315173685564,0.11671332549009916,0.9920320377532537,
 0.30772861074513924,0.624669336937353,0.2789438233440594,0.1910119547920457,
 0.9238174751583031,0.3028551292390482,0.4253324256611897,0.7010648055863408,
 0.8015422111844204,0.512053499385938,0.9078349385562656,0.8009842465664251,
 0.13207525709282142,0.4582787923612188,0.8230449621445077,0.6765786499365259,
 0.48936009923151913,0.5481854085797956,0.017342107785815353,
 0.6799525365115283,0.35223365302925147,0.3418645425745003,0.3130137462201493,
 0.538434338960671,0.5285437399433198,0.6146317243697434,0.15225506396265454,
 0.5891351725045385,0.07759237058847379,0.9653227902426524,0.9819255015852943,
 0.9112723763660733,0.5968270471357098,0.5903134562009293,0.41780520266939025,
 0.3394870482413501,0.6661365077975161,0.56312209144701,0.2694299561174027,
 0.0013558261766236646,0.38221803961297796,0.6808715166695868,
 0.5891615114677309,0.9015967782540727,0.3637738958935257,0.25342351468273105,
 0.5111718287186213,0.5152628887314793,0.32050365604161457,
 0.26125726990531284,0.7199552792091862,0.7616374828976717,
 0.24159786554073193,0.9737430164063398,0.9797510958743041,
 0.36209951033580434,0.21567663981032503,0.28890391634899903,
 0.20088890030298057,0.4749096928115597,0.6960191105023735,0.857441166080051,
 0.9940494082502216,0.34542718704234876,0.7153842934321747,0.4951845175315076,
 0.16547580089750935,0.31667078692042183,0.4773924940212785,
 0.9904948328668997,0.9019525414282374,0.9541729473287341,0.6372705182882352,
 0.6102971037388257,0.31144533140989905,0.5166846014012135,0.1290856474484663,
 0.577051240273237,0.1717381001977828,0.8588214807295316,0.6778386876884293,
 0.9491107040289714,0.1959005881090744,0.16210177024062422,0.3734781822766309,
 0.19264515216179767,0.08825111871784375,0.9051866236017636,
 0.5895607168348386,0.5041211940591109,0.49540023037128045,0.1209094838661916,
 0.21014454115990078,0.8968844456349062,0.14909003951903899,
 0.8982108186595784,0.6992613721468752,0.37962622150957726,
 0.24834236161343992,0.40591379257119775,0.7736259921102957,
 0.6213164211476535,0.9996098063051853,0.9742215854087023,0.14547386741293833,
 0.36294355804602707,0.9120729951045824,0.4037401180299992,0.8981717019602667,
 0.28258309712983254,0.04586533303449469,0.8409068701277972,
 0.7313772894432234,0.8496194853047841,0.2688768980684092,0.5411379770637267,
 0.5006350435231981,0.4127699323536107,0.11333702871235063,
 0.24900991148110174,0.13102953318849864,0.1731655284472844,
 0.07524015091273006,0.9812017637559253,0.44069906980890583,
 0.6712129640833245,0.08122243502509086,0.014009835253841252,
 0.4049054867670957,0.24308936584379437,0.9525851035559529,
 0.01867580357223786,0.826679970256182,0.03392671475065179,
 0.30067222238247515,0.150100481260391,0.480205080261656,0.21514411224430607,
 0.8534179082560185,0.01894449193373071,0.8325095726170568,0.5198824365348983,
 0.33599938108417216,0.7081331012918707,0.18777418879393626,
 0.17257064266670907,0.5093540395829742,0.5704124117946734,0.9389122847218017,
 0.008675450389972528,0.8738357422423321,0.7177497372171607,
 0.22689818118535798,0.12537391634861428,0.4305683878481479,
 0.7918617680132867,0.16268335197233297,0.07409290449763528,
 0.9145830634903556,0.695658041758945,0.4083075952549916,0.06113324458888503,
 0.07857040534000581,0.16569032326805166,0.2671469163438004,
 0.13040777159525896,0.5859441326807648,0.7565237380240212,
 0.30295070107009703,0.5016954923351369,0.06246867346614682,
 0.17827167806854938,0.874430156456258,0.18577448998930557,0.6545499676838306,
 0.8839543100021638,0.780637338523799,0.5704522621713128,0.06806309547782219,
 0.5791851089014629,0.23045716338529976,0.9958477090377266,0.6534004610598546,
 0.22440863599412797,0.7977593588344707,0.2180147206156794,0.7139588239783717,
 0.9614468931850453,0.7047161237214562,0.667749561288739,0.32617948719693757,
 0.7903069599698191,0.7683168947697586,0.5693372953949511,0.39255623786699056,
 0.23756160517353275,0.3117141011393664,0.40809853729673284,
 0.7210115525423052,0.41341390782962284,0.6523414290256311,
 0.14733891248677877,0.34595370407739634,0.88612888479393,0.7957696842458282,
 0.6896068597320969,0.9286210721995156,0.2293052505661073,0.37756198383340456,
 0.13063704872683113,0.09347782173468577,0.4434055075656995,0.929541564627057,
 0.11774598285015114,0.013091141461338052,0.10806105625546758,
 0.33496509418892284,0.5450690508001615,8.956748604749265e-5,
 0.8304580914171895,0.7452192036096783,0.7577272179087616,0.8747925775685405,
 0.8878589352064193,0.3038060917384684,0.7735438344616554,0.47514575947834337,
 0.37105692919110145,0.15423485469213805,0.11957290625841777,
 0.024996055069702372,0.5472790878459104,0.43414323886923,0.1052165633605997,
 0.5696594903782417,0.6955319405603096,0.7172517921219558,0.6474413825097161,
 0.9879701603530018,0.3216728463316604,0.28094889280447843,
 0.06894282268501273,0.6188342923924793,0.23947286432414538,0.797637376468854,
 0.6073870286878619,0.12464054230775645,0.12703343098851472,
 0.04234780826956641,0.7863797283322213,0.7482823562490163,0.7939830669540381,
 0.8327388875758011,0.5053854665765463,0.9009734838711272,0.28792149780418863,
 0.1942616875843396,0.523781470530108,0.9947004126931298,0.2455464385148386,
 0.21093547422138614,0.614572562451498,0.12309981938846448,
 0.24077538114251407,0.0995373136472808,0.28294655985632555,
 0.7690219292203082,0.8782390124623707,0.513279638683436,0.08859151279263477,
 0.8779649996788867,0.4469405176786516,0.24276276650057826,
 0.41350721038147686,0.12615008173566777,0.6731804249423625,
 0.29022674880758614,0.12054662722727916,0.42014572486407353,
 0.164134067003054,0.33835879627912924,0.26537905069830403,0.525000441450474,
 0.40625376418413195,0.673302099519057,0.4942848984360362,0.9904576438943716,
 0.5195186277812532,0.08737353600508024,0.12453876352521076,
 0.7685675813748878,0.46262106861786867,0.8810853818349207,0.4236273039897912,
 0.9588271997872213,0.7792396205428962,0.3249278061326014,0.14418115744893134,
 0.8893786662315913,0.43547982946812835,0.021958359745128164,
 0.029885002206125222,0.9025641327662374,0.20628580666051466,
 0.10707517528998112,0.6618557414087953,0.9897505923425259,
 0.20448358469523342,0.9220421737979716,0.5695839670144699,0.5378453602331261,
 0.09293084592381817,0.9045319821528386,0.19712052374546185,
 0.2823286029818811,0.9185819050233137,0.8255645940746412,0.9445665181646291,
 0.9110339470239182,0.387658808661277,0.056165104097085305,
 0.05519537378319028,0.976046255431841,0.9020471683459927,0.3076529478041823,
 0.039009264704027324,0.051662657367974596,0.9293702848634368,
 0.4223873655840099,0.12603827143561497,0.9250608902476164,0.9544463400247717,
 0.7440498694647814,0.22244151383224287,0.16195400703525453,
 0.8195074794239275,0.17838829853862315,0.14357798618282214,0.646497296224212,
 0.47389415517590283,0.721901172511939,0.7906720828857188,0.6656807507114264,
 0.2159998321924541,0.8901945682700477,0.8684146091531084,0.9194184281240642,
 0.04276075856172301,0.3485710266736828,0.034507337924139736,
 0.7395453806961382,0.5413366068612319,0.9638340249813551,0.7347146845865274,
 0.01675583087298438,0.87582506937657,0.6959416681087168,0.6304596959135487,
 0.3428446383371543,0.19568312679533673,0.510821815429793,0.46954544366776274,
 0.24107161801956922,0.4188830456205257,0.004799698601981484,
 0.8648764534891225,0.18378186676103025,0.41899186354927553,
 0.5142704543968072,0.21228972920867406,0.9907112269084573,
 0.019909657203157516,0.8945419687033055,0.7627623987328195,
 0.6474075751950067,0.8306988265381129,0.02690099080879038,
 0.18187872592379617,0.3387667758726851,0.20781811889414392,
 0.3843459263446747,0.830206280018565,0.1697316308685315,0.1775567715075338,
 0.7968843993605162,0.16646178587928662,0.7685143170075031,0.6827671372382483,
 0.9789813386582993,0.7733051119400915,0.21131721860876096,0.5305973171797714,
 0.636600205560462,0.5483333131836339,0.8207305844355806,0.4046027641083034,
 0.3440314012078478,0.0846380192898386,0.7689297343398824,0.9033447142183314,
 0.5513084386068517,0.3657408578476181,0.01268878405469398,0.6065352327712017,
 0.9000505670191344,0.10004574349416784,0.30350751567998413,
 0.9900205594410624,0.3372872760494382,0.6482572142612675,0.5293471066231148,
 0.8454114702119137,0.4785870362693121,0.8064222742741175,0.13794721054774495,
 0.7459688949527125,0.017704599990422176,0.21212996889768343,
 0.854581035582499,0.7096235004128356,0.9784989301582501,0.06300199588152566,
 0.18060822835778823,0.8300758575376926,0.6598637510706316,0.9333422992574193,
 0.8597277633483486,0.03587754583208247,0.6291964886886208,0.711491009613169,
 0.7250521746240459,0.5698633925209566,0.8349152716315992,0.37390500541556193,
 0.5847080218789656,0.6682847059165479,0.13302156256106057,
 0.38900750516753113,0.7464108759171866,0.5930506548524255,0.6237728801943608,
 0.5966253624567739,0.0967749485144126,0.39950283444440493,0.2766337974951658,
 0.42997697256198475,0.25189029589395484,0.9393227960410361,
 0.30949460993282674,0.8728736542490453,0.6334594999753016,
 0.11330138637346476,0.029990635756498518,0.055994076506430956,
 0.8412546005978454,0.4756853673709003,0.4865013579578362,0.6884028525347335,
 0.17445250186929318,0.0796807338926484,0.7358179758956379,0.4952806828157047,
 0.2558389428953647,0.5749351694064643,0.023551313581937183,
 0.28156642651976393,0.21191750707600931,0.2383370808716132,
 0.16119046517102653,0.17410355433159463,0.9563323489084866,
 0.9394703773633435,0.812937215927411,0.5948668396801524,0.9703258179716884,
 0.04817221013444373,0.7842838846019518,0.3999712519886609,0.3066208353803732,
 0.8239928787587852,0.3759823550006368,0.8744254666051797,0.2912748653010132,
 0.5418452319080092,0.5623856147548546,0.023679807298332856,
 0.6948459664322985,0.002283739286024433,0.2542568632740082,
 0.2861680103331955,0.9687082269219055,0.5391164678411096,0.7304809782770147,
 0.8952997339904747,0.9861790675762455,0.9507191766626859,0.44784990858467566,
 0.13714810856591075,0.8885087449817105,0.7951399180138636,0.0813701266847594,
 0.640116973262518,0.6088151481633166,0.9006752359177118,0.12217700816786536,
 0.8844751245141974,0.972200972661047,0.3469729249569914,0.6750469229531216,
 0.7955227041980595,0.47703624436368897,0.8598198060900788,0.4229899045324649,
 0.06275616163103159,0.9790259671169089,0.8122614267983391,0.7992206431275727,
 0.3958802687142591,0.45864137483979184,0.994975606673486,0.7923170022817564,
 0.06118844789226241,0.9123684739103127,0.7938379651080251,
 0.25611384954311067,0.6383757072344359,0.5057539859992848,0.7182249599999475,
 0.8994597787097953,0.2776193267706918,0.8232508264028362,0.06874566740031096,
 0.11935666483921192,0.16767280054691935,0.841062319366932,
 0.028566349946332226,0.008019175437586124,0.3612406325934885,
 0.6608973900950511,0.7056897308950154,0.0876780607660187,0.9620996846467643,
 0.4006089792597052,0.5204055573200791,0.6148496671620617,0.5196634107245217,
 0.484299225920951,0.9676509292158321,0.5488118014256741,0.05878459249934964,
 0.40818054316281815,0.41825011555278424,0.06239952392750292,
 0.8430510226203539,0.986846671276381,0.801236961832122,0.7305018152886384,
 0.528643205843674,0.5084381728490184,0.5453870195318989,0.28285798988528055,
 0.4326858185039103,0.4821777817168136,0.19104676236215556,0.8359560229029157,
 0.5547516637852327,0.8551923411988003,0.12331344240458952,0.3322683883829609,
 0.6865788967714482,0.8119072482008025,0.00985379004669018,
 0.04244503466877192,0.5938879604000047,0.19036157809764143,
 0.30029360603122346,0.4601478834255259,0.8206509842922537,
 0.15225762838815604,0.06895544998585079,0.24706592676368766,
 0.793717337921672,0.5577521220641939,0.45647171894064564,0.2524956699538008,
 0.7790516826099555,0.3587762071898538,0.20965289779404395,0.6131930159010827,
 0.0321492746898181,0.04808462846690531,0.43876831178078457,
 0.9396836471557053,0.16256107988844504,0.28695971042294854,
 0.07451412095597842,0.31847612570790196,0.21618896063861237,
 0.19440829494950906,0.616457990943737,0.14923731388273087,
 0.43030197244995616,0.6751947507155694,0.7490111339942085,0.8462984273791618,
 0.8027901351667288,0.3209075166064911,0.7068598818168467,0.1448716429181226,
 0.906855574812835,0.4913410844417584,0.9635053768366936,0.44252048197581584,
 0.05432284059283565,0.8495372768957741,0.6257664314523854,
 0.23674419616814157,0.4421736551354396,0.6284626371745565,0.9686653308396147,
 0.42728308409135285,0.7874107470257514,0.6336520527677909,0.5797227446473393,
 0.8494840749503313,0.3900574126306512,0.7899052616257839,0.3503338109998936,
 0.8858917893087157,0.15459706998113865,0.28706630091412344,
 0.18530971820690478,0.2652671692834788,0.8337206192575587,0.3537044000536874,
 0.9740792157445692,0.4223384828480643,0.49902818235125,0.7584868243636526,
 0.04311297189732988,0.7121564873967398,0.8658544337434501,0.0056874512900984,
 0.057495764589008735,0.16171089720756449,0.08663426408200225,
 0.7691271775319322,0.5154260921544097,0.5501483993187334,0.6357073917362659,
 0.10176462403087749,0.724114465524891,0.7931561510324499,0.06542220786917086,
 0.4482383356816593,0.6192388988268169,0.9841724840489032,0.6292355059335234,
 0.949408657395511,0.3098814705876348,0.8077701069441096,0.2550508511694001,
 0.3652979535826496,0.6460924399672334,0.09707602621200107,0.6321441711890032,
 0.3851868718624799,0.4356002891741171,0.2841191110235153,0.9109697778183822,
 0.03193887693684916,0.05955245152864719,0.794699933773264,0.7080248008419769,
 0.47436418944873826,0.606725842670395,0.9510432134795489,0.1347019397247322,
 0.7912492600340048,0.9537269501144666,0.17351664370287134,0.88032484357253,
 0.34884089589332734,0.6082579747881951,0.4452173567149814,0.2479616713998969,
 0.353629676763058,0.2516211184452779,0.6220639349092484,0.8154116844009413,
 0.2759799997350998,0.7161668211254324,0.8692311419644051,0.32543493009964874,
 0.9851423736168353,0.09624837980974077,0.3189516630423521,0.6821873653538365,
 0.702336288799706,0.21132868476625066,0.566694469191237,0.11578701557281534,
 0.9037040540324364,0.781966900087012,0.22040321639627336,0.4329064993090329,
 0.964507823828589,0.6202450408554517,0.8198132224990362,0.7432728213833757,
 0.3342613218483681,0.15961257287069597,0.11259709773040272,0.721498701250574,
 0.9899008125320212,0.601597302490837,0.48968318660332777,0.6495061946684424,
 0.8781589220381213,0.8602478217316477,0.8305686962361483,0.5158230606194831,
 0.5643851805254104,0.8176475032080748,0.4424988797811178,0.2642337175885743,
 0.7939402713485846,0.929563328132713,0.5656296414164301,0.44483845133475564,
 0.8787522825022264,0.24776566005435996,0.1805050132869076,0.5890106545222935,
 0.3068621292597151,0.4227911519865808,0.31060459009488195,
 0.49106155921643546,0.693393094828736,0.05268735488794496,0.9699932568236067,
 0.056422058280603204,0.450682511311456,0.6144694060798864,0.4333143916328982,
 0.3588705961530735,0.8105173466051456,0.9871526111331367,0.23187350095056558,
 0.4571203270520542,0.8903605389529464,0.9875820040614727,0.6653335074080128,
 0.16595533703103094,0.6078459060593346,0.1821558380423708,0.516684484750619,
 0.4760437210097488,0.35013318546033145,0.2891489184652313,
 0.08531974896871164,0.755897461812018,0.3547509370445113,0.4931307265205377,
 0.1169462044671038,0.2723230709698762,0.36981400614086724,0.2698093370043688,
 0.8776169809653485,0.09389647543278334,0.38836630144148954,
 0.6342577676718018,0.5781516760244191,0.8905712101535568,0.7080960095550707,
 0.09834574370020954,0.5949024864038813,0.2491361745246532,0.5915917497495996,
 0.1725410441242211,0.10996974989317543,0.3519174696138502,0.9533618142069648,
 0.8145783577901179,0.3826573564959346,0.5037332081695388,0.1299497353794642,
 0.05080211516178401,0.5259305093492442,0.7294492213546273,0.7368221632519688,
 0.8210295345265539,0.22614494031872223,0.8907456701164999,
 0.17324648486214633,0.7926990939001788,0.09491180963818113,
 0.7101961530898385,0.6154374517886021,0.003888035326763961,
 0.3506630400332391,0.23506880640479055,0.26823115824363764,
 0.7792119719580275,0.3570360112449479,0.8945499087118529,0.956349751540286,
 0.5516981425040666,0.8392973080075994,0.998019293778929,0.7809832416172274,
 0.5373108055690037,0.837015312160513,0.03981536935462582,0.34628931648611927,
 0.001273859762339069,0.6004403590534273,0.18264525753070748,
 0.5007210522595313,0.4808217032069535,0.003271037105695207,
 0.32442198242343134,0.4687943025389877,0.2712442459845482,
 0.16057912143656328,0.7396407669630346,0.8485581143975248,
 0.12142804274289465,0.3899393898889183,0.8829206387249281,0.778385444035685,
 0.8918098773581749,0.7182832707017782,0.07839030675446634,
 0.20218467862540201,0.04024352345329951,0.12487444712332085,
 0.03499741581542626,0.6053761397798063,0.5097793212272,0.23231195863685938,
 0.6448469552465821,0.434674824734789,0.3463453194638807,0.5708195152526933,
 0.25914643927405456,0.15000689732350825,0.27862745824968627,
 0.2543694742657089,0.3574844585329051,0.788621178347903,0.41298219772922506,
 0.25989992621976354,0.7882650983620854,0.4405422793909617,0.5013060517528076,
 0.33067249613565863,0.6939862685260907,0.9993276855937167,
 0.10701084575228004,0.7393468555620503,0.10078555501448172,
 0.4950976390385703,0.4823879115146059,0.48746426488393,0.06936953329800377,
 0.7483635690707342,0.4808945659919899,0.9125336364187164,0.8885694641358567,
 0.6052103017620019,0.05450969886241841,0.3724807265905856,0.4426752032000747,
 0.023335373690855032,0.9301021939867585,0.7121549707923257,
 0.7746199755671757,0.7443808508020908,0.5422735776367836,0.41755007818495293,
 0.8979777922804816,0.29114190581464394,0.4079560887072956,0.6448544004225917,
 0.24426347400367654,0.6833641280701119,0.6432026984008167,0.5215253437359206,
 0.42062640596026324,0.7131588494033245,0.2843254750392822,0.9743818815332042,
 0.9451893097784775,0.44068640171924556,0.16090219018070773,
 0.14328634749865632,0.8827439743184466,0.648853875095225,0.28258974293057004,
 0.4008226103844702,0.408055680118558,0.8044310409246884,0.1541024509389205,
 0.4256795982560221,0.5653771727622112,0.3926945591233747,0.1728463451456016,
 0.023855207462359806,0.3897431649727894,0.26012240627389605,
 0.2599561617005939,0.11354800481807115,0.64903393116817,0.1789681881955546,
 0.7757297354825194,0.48999102548527695,0.2756068933487522,
 0.20835867091223015,0.8681938715984134,0.7786597443977173,
 0.47207033881351457,0.4349343193774028,0.5628525050393387,0.7018589345121868,
 0.9773609676079532,0.8823171226525782,0.45450827946991035,0.2028976614040775,
 0.9170434366772664,0.9513293564939677,0.694679006688979,0.16720568130873525,
 0.46533634441524674,0.7122714992310504,0.3544874668033081,0.1586012081358017,
 0.3054052711296814,0.2797869737788984,0.7355225142432149,0.33486834656371534,
 0.5767835323059052,0.4951284821419232,0.4216582365014728,0.8259909779705984,
 0.7458886923974224,0.17089922679703073,0.6164614720406216,0.8192318050299308,
 0.6854183898833471,0.025904033350897304,0.4939219049974777,
 0.13806465213237906,0.12282282251301446,0.3982635925194107,
 0.7624013021616944,0.11558080793511194,0.738371910067662,0.6620169901394828,
 0.4526269101964894,0.8265291694454068,0.7775799503353477,0.1485470920535613,
 0.49018349544292417,0.8586398291772399,0.5613035321248955,
 0.10685178106249116,0.6571146327596707,0.10552478457522385,
 0.14029960218015936,0.6838669598640037,0.8536661208419835,
 0.04764961439585913,0.9722648190239827,0.9548701324555184,0.6018326373714807,
 0.4088599758819972,4.786341447132747e-4,0.8550293912230212,
 0.6771234134289568,0.43042121393234634,0.4996824151510443,0.7753991569586526,
 0.9631385670495677,0.9669648462358114,0.800334890527138,0.8008614085858576,
 0.2741980157797852,0.4371820245500302,0.9878325110027673,0.6580599104330065,
 0.14958272391357164,0.09802183229622408,0.4867247393932972,
 0.14109768144579893,0.9528532981737606,0.9603143904474042,
 0.39351171894818104,0.7898178650605723,0.8418423804830162,
 0.27257915285215706,0.4420786405756246,0.07974875614170596,
 0.9409565812761398,0.8550735931409804,0.7709022121439267,0.09317871753057694,
 0.91649092925325,0.7850201483541566,0.8310335248401899,0.9199462035322026,
 0.06125787864402543,0.25698245692729005,0.01808361300572714,
 0.01852290351230257,0.7732355709732874,0.27832734358977207,
 0.03573445043005552,0.640469290045165,0.2668137194468222,0.21304395358435302,
 0.24976000255602737,0.6626944236556995,0.22161282638286461,
 0.1607121867926783,0.5732287032173045,0.36224777997170987,0.9738936493904446,
 0.3473508812676047,0.6143817914830508,0.7184591953757073,0.9563618769110153,
 0.3545773152505811,0.8713434752314687,0.11834932942662613,
 0.048994965912852795,0.6083260692701704,0.8926061566933476,
 0.0016703603863541527,0.7163973142069537,0.46275175215655817,
 0.8037859295381736,0.646279420173573,0.7432109071950649,0.9033492883959975,
 0.45729752815887625,0.7939097416234533,0.3309022129459275,0.7953043356218346,
 0.3433348161990579,0.6355628072998373,0.04235161398955256,
 0.11393087993259254,0.6697503901468002,0.41343769027266253,
 0.056362299486424905,0.1593522947383157,0.6874702576320493,
 0.34902892169267585,0.955722222062575,0.7442054319978647,0.10594489881630997,
 0.7946373651375073,0.6576574166025446,0.0015096617792448308,
 0.4653745115505534,0.8385825365983014,0.743814233557255,0.09037943944994042,
 0.29541509632946994,0.9205044318784181,0.40678619089996215,
 0.7825846581164377,0.02930296920291986,0.559878110886291,0.6726555025356072,
 0.9107127012439494,0.6668491944779614,0.8308263447677097,0.8734345825834264,
 0.5914022200782968,0.3354653120843851,0.0019943688829915374,
 0.49232429352073215,0.9765329571414326,0.9965196788715375,0.7506419861786239,
 0.6718821159658055,0.06028860243162315,0.959587730010036,0.3191550362736735,
 0.5851471104708311,0.24023994355536127,0.005605153749054237,
 0.3150224325421569,0.19839899535385652,0.15079293447418063,
 0.5617253488888378,0.7208618692377137,0.2665749839132303,0.18516497263264298,
 0.14818236117980443,0.7811319629673027,0.6588506882094193,0.7335344472925235,
 0.37473766909241035,0.8999228579996597,0.6365354996096286,0.866147426628185,
 0.21650455111486455,0.8444018276261479,0.8454786308319369,0.5641647695542602,
 0.5762710099475548,0.7539056569833683,0.6698571836814801,0.3927118207529371,
 0.004283341390748818,0.3954326614158912,0.21900649999228838,
 0.018706332441892903,0.7454158078588382,0.03884245940385944,
 0.4430642397506733,0.021081616500227107,0.11549962665579216,
 0.7210705548660555,0.19926314013327817,0.7968236525620698,0.2220980526229499,
 0.3969910789207195,0.3595755138423382,0.10091332821829235,
 0.17133834333502773,0.502347030480505,0.17683643503029844,0.9449791333665306,
 0.41742238939444365,0.6542157352191613,0.4534697516938645,0.3483256520525041,
 0.10088930781748295,0.6580235321561156,0.8649749341129147,0.048057418894627,
 0.4291696206781067,0.9932454978170457,0.4500842983903245,0.48059686311545713,
 0.920147401676511,0.4113171710206043,0.15390644356783223,0.6247836412942022,
 0.9452295445257067,0.24891568406719167,0.9913688807832756,
 0.07371811186111221,0.34929109775765865,0.8738733034786268,
 0.7244191122065684,0.9132519049002393,0.7445649169817533,0.39391811583807435,
 0.7320288207657742,0.9277506291417366,0.9955768442884013,0.8080306271334912,
 0.04888397181519544,0.6459340692626054,0.4295128963137935,
 0.06637486598770603,0.13135754073121086,0.7438762443803001,
 0.9831830859444013,0.5566991355735827,0.9241598820424035,0.1422840222045879,
 0.5802500856295411,0.29237343565810103,0.27942490824482924,
 0.1873160066345314,0.49952609849476926,0.5452658280961487,
 0.13952424157210974,0.17465730265641766,0.672500031518702,
 0.019607366093899903,0.5800024860232157,0.6522216590294784,
 0.1170293123631082,0.03958012374536468,0.6913432540839246,
 0.13130075986742606,0.3263208155614755,0.5670774516638888,0.2423129282980263,
 0.05598090244928633,0.47702940704676167,0.02380271288525304,
 0.9080022604252848,0.3242798313988464,0.621170044213624,0.20339083017731296,
 0.34065540885135115,0.4642965717637294,0.1946516363883466,0.6705565726541486,
 0.272796098676246,0.8118086491146573,0.4998325352760379,0.48678878409605186,
 0.2976039218270642,0.6024585230375926,0.007703138001761145,
 0.9588737035747562,0.3598392750011332,0.6597855028475182,0.5254145088995428,
 0.18231725021374012,0.14580100748074454,0.5818867404422003,
 0.30615485955279387,0.24644911104199485,0.43890103087311094,
 0.23821963235416888,0.39225979620896,0.6847953891197661,0.9517964983360184,
 0.25574012417866887,0.23122496765060185,0.48610538173967277,
 0.5967888320626403,0.5155518359404434,0.7616224411547671,0.4529131293192479,
 0.787071040104738,0.4972180779326991,0.5111678931412222,0.2614177507934745,
 0.5856364557077162,0.7818159132919813,0.29302732235313644,
 0.16032911302986663,0.5067499770441782,0.44026827506254584,
 0.11834386272754904,0.908483359839329,0.7813508698849542,0.24555845040610313,
 0.5416937030426024,0.49728323710934796,0.7566823487296861,0.3046680128720054,
 0.6103731657292714,0.9231779612739964,0.9469994695952825,0.0896084911677345,
 0.9867241625866574,0.038934641933230596,0.28394594571002574,
 0.970050253628609,0.37971247965304,0.6597419932298527,0.6636028769738379,
 0.21572261950857274,0.15583218451195746,0.14612533727480692,
 0.5428134151976348,0.9372125779356859,0.8382694761321644,0.4326777107655029,
 0.897527729899173,0.47039693993857734,0.8648768870086146,0.25178068630063477,
 0.2713272870135589,0.9805498092650916,0.7478390798255068,0.6033912881269589,
 0.13608511941274226,0.5078533944106292,0.28515501042259184,
 0.08005001218564689,0.20713846377599732,0.3133585344764962,
 0.5525024746964624,0.915151043053973,0.9056298279378981,0.9341239268496262,
 0.659566794634097,0.43086721427495156,0.3852697482514016,0.8120613407450402,
 0.5244340622137337,0.6166618539758125,0.04240786031506005,0.9038788043866263,
 0.4875383203784094,0.22866833647521867,0.8287124183496524,0.8382440108697116,
 0.5863336073143973,0.3361750677484351,0.20900035983044973,0.9109859747076289,
 0.501219720596089,0.3185050031012967,0.25071138637804036,0.15157604928608648,
 0.709043266896924,0.8270112505234575,0.6195388391870194,0.4767128549302335,
 0.28463367474726564,0.6912691300485417,0.12977515670090434,
 0.9482740154414733,0.8112570067755236,0.11326462690268857,0.7551108387915997,
 0.18339529665293797,0.5166862388983086,0.5251080233621884,0.5941291898816742,
 0.804818283988025,0.9455818020020583,0.9550469478069846,0.0663497182752596,
 0.9616518475952887,0.7316133172709872,0.36380404489109663,
 0.44779672363290923,0.16513820723373784,0.6257994860234157,
 0.6778147225490965,0.9687229197695786,0.7950934265522753,0.7493920532196616,
 0.4511820904190703,0.06780281683457545,0.5090471062691235,0.9691850145727,
 0.27837091424453964,0.1416541973503812,0.3372817670619481,0.2010491310433533,
 0.7313634328128134,0.21655109452380117,0.5039889357107223,
 0.22954956078010436,0.9459801920354841,0.5467996935949359,
 0.12215169152015426,0.8028345635515215,0.7911397860634729,0.1136624545952607,
 0.13262678023701235,0.8288297164720161,0.43705100114093975,0.301547026164788,
 0.17462412292601381,0.5263716772015905,0.17385229975014482,
 0.9566800321901505,0.7301333494968483,0.5151342207138773,0.447985416706028,
 0.18095601863682742,0.4073601237702785,0.14327284795477313,
 0.8150929485604488,0.8096157526953398,0.041083267310077165,
 0.7226009136310031,0.1312612954480218,0.2054877240149282,0.13106741985145254,
 0.3025595230170347,0.05144466751875565,0.9907329817156438,0.7649854980583484,
 0.3917387618045506,0.47474580759320717,0.23269038986983914,0.95591051100207,
 0.7557252934777323,0.727519651456967,0.8651845374364618,0.5315957698972222,
 0.21057327131113057,0.8474131465054269,0.6824940870239566,0.6413634491844099,
 0.1075295613008056,0.14037619264999157,0.8507508646336416,0.947509391290857,
 0.19276681424186193,0.3850986088053723,0.9629403053324589,0.8879357763652272,
 0.07888742178758212,0.07030329732375373,0.2596250010540858,
 0.6079453751554462,0.5134977190110573,0.32285398520013453,0.9651040273684092,
 0.49593473319817194,0.15288666237808457,0.8604790669585121,
 0.34940808183959315,0.86113219881346,0.5829737914602732,0.9380217287214662,
 0.8262437946599627,0.6292466101713704,0.14569334546228452,
 0.00150961561986529,0.36191692985306656,0.3362542865344421,
 0.2124850334460674,0.9944977736748654,0.36084791659578375,
 0.32702215011653857,0.17813972781564558,0.9066857680740956,
 0.08885781332555909,0.8092935334198889,0.8672807100194222,0.3648503657998543,
 0.5643698885909011,0.4291326218983128,0.807943962199581,0.8634599645366398,
 0.6600644400446275,0.08483819895793931,0.729010625457576,0.3092372674848508,
 0.5610747331288896,0.24433943248226964,0.6854861416375049,0.2225773419948155,
 0.06624166983350088,0.02537506886378138,0.26368136109303064,
 0.6067912206743777,0.1299663514150129,0.006820557118199622,
 0.24440369368390336,0.39335193845494987,0.545405180687728,0.4238885165427716,
 0.023141980985599275,0.8289237849226312,0.2736010345604968,
 0.5559447720509902,0.31104433910303075,0.3173081805867354,
 0.20966415984007802,0.4645573770899698,0.13181991870052523,
 0.07414842917864228,0.8232282340834706,0.9773287887043621,
 0.006780514015409578,0.9601360084638438,0.036967065808025446,
 0.7670274606514339,0.8910882642545056,0.07896792279408926,0.8621707219486225,
 0.30533336661069577,0.6101399044936263,0.5408283543789332,0.933757661358883,
 0.08125830282072788,0.3135092715593527,0.7824246386221283,
 0.10297427973662365,0.9273802913095368,0.3468730986749621,0.7340360414042763,
 0.8179075444458174,0.18861228550912268,0.33717938535158565,
 0.5065151165742946,0.9563459722827515,0.6946298491238214,0.6509434819368647,
 0.9521827240686389,0.5571264878440992,0.5921810773756728,0.7846232872033524,
 0.6959003397506853,0.6229778360508657,0.7309014265614479,0.8867989144237389,
 0.5326118503494963,0.39679208088086315,0.218443805985848,0.5312554453548927,
 0.2548353071529339,0.3658154835139231,0.963990681261328,0.7801816707537806,
 0.9065281352227685,0.35245905061292615,0.3094118031426424,0.1503963747774113,
 0.03546135489949154,0.2545964321299078,0.9892977834918765,0.9745620470093557,
 0.6051399556479525,0.2628049997023294,0.6905009876551502,0.35111204082800096,
 0.7699184568562851,0.5922514214006233,0.905311195338252,0.6061995884257849,
 0.1813245220504005,0.40288589270946007,0.521283237014239,0.3037983471700527,
 0.6626998701315664,0.09623875198600385,0.2898449473369459,0.7648672608984071,
 0.2196193288615459,0.17973193119632158,0.40918080113977773,
 0.25979657426039837,0.6981354359017955,0.621258594084339,0.34145635462102186,
 0.23954183480799351,0.028843256930890182,0.8807971504120554,
 0.7954187826723801,0.020727368985982153,0.2986300772530427,0.677131687844185,
 0.4821783476399022,0.45854866754741896,0.47846052975957376,
 0.7129266508072623,0.47793718139036534,0.587741040407924,0.6972907390206851,
 0.5562708796560661,0.6868264818336802,0.9567108869919604,0.9007727975983133,
 0.8539394068355629,0.34010509364896924,0.7295827197889947,
 0.11731316603430075,0.9636298038824092,0.6574875544505798,
 0.16470470374138946,0.38213383948286195,0.5031847813640768,0.87080706995248,
 0.07067717188705602,0.5675905978690011,0.17952803255736693,
 0.9828963603124645,0.6546259608405693,0.06303604884413472,0.548936952328356,
 0.9580125446269817,0.03506101451039767,0.919884298935182,0.8290763871651008,
 0.47248380716913396,0.20864847091129113,0.6588583475783554,
 0.41859922418586426,0.7429603037417856,0.47184160559231425,
 0.10290903050383993,0.39873812962227795,0.25209086136236936,
 0.8801746509843783,0.8275752867566252,0.6632179055599084,0.24081373651296578,
 0.32698233711935454,0.7630022926479411,0.7609484848460182,
 0.11384056115022079,0.59676559711157,0.8246575011576712,0.187729846172752,
 0.8090046026569178,0.9768928180458132,0.5832397219949691,0.4520827855095906,
 0.6613775712388867,0.6642001962104774,0.1608384908957461,0.381124775004979,
 0.25882976747914876,0.8913335563430178,0.26854434749380296,
 0.49113782672832174,0.9983566931454068,0.6950143145530044,0.9258966545955164,
 0.758470785877802,0.8881561954086585,0.3520069875453372,0.03403278883735672,
 0.3186916821759238,0.10219960093884328,0.8899840282370939,0.8980877738363913,
 0.09844315905658829,0.9052792234734344,0.03977359095638999,
 0.10984730158090161,0.7035330967286224,0.14362045977165727,
 0.9260534045116663,0.985509008461656,0.7175021977283763,0.5532359915121831,
 0.6760757145417875,0.15334609749397687,0.6725106247147234,
 0.07412600708042993,0.9168560095845149,0.9140051284821221,
 0.17587440308484747,0.975658651580773,0.5871970574873188,0.8185327512387861,
 0.10738116815974275,0.8697422297218608,0.1801820882530556,
 0.36770644928574636,0.7570838685307645,0.794364733742374,0.3793021846305453,
 0.4145030909073927,0.7800874834683921,0.4666926735416377,0.960727604791371,
 0.3146610171927784,0.5169999298556673,0.9618371241826872,0.9869325312165734,
 0.14141538529364794,0.4359565000065735,0.053409553464038306,
 0.054901112575263755,0.08874235965615207,0.8211565111771053,
 0.29235255939031035,0.37042748999203123,0.6910792616813082,
 0.11678610906266151,0.9690172250733768,0.665576834770473,0.9171652888930824,
 0.38464228539179657,0.8399688931370485,0.26840230865710457,
 0.9228438874922551,0.31252559445832506,0.1709348217302289,0.9236162561465834,
 0.5048945229736481,0.21067501795961563,0.7435202694352653,
 0.14685834786314766,0.7022318983037552,0.9603076192734109,
 0.21120351473689247,0.05166837678067793,0.7663569041769016,
 0.8257497969246161,0.5379075717013829,0.6413897339794496,0.21695893796499777,
 0.7689557922717557,0.7209267215017064,0.6830816369942264,0.15783988174063923,
 0.7715806623618139,0.54927661222708,0.6745870534164361,0.31863943329314937,
 0.8893406623699585,0.4845634184282278,0.18279273905558657,0.5342532493726074,
 0.18487226692422865,0.2980748022378128,0.2729453423375141,0.6626457139306728,
 0.496552645760436,0.8044006246539623,0.7055397521530689,0.8068072857849105,
 0.10221932309126869,0.39750528951805697,0.2858148867498346,
 0.42575145518527346,0.32194759071412204,0.6213645802973432,0.793284987889787,
 0.4107196056214164,0.03396894868413447,0.2633942385459035,0.9674240949423696,
 0.33514083700182806,0.46746835952098764,0.22009054691713736,
 0.7417845491570086,0.29156393843800554,0.8070084643632291,0.7438034434889957,
 0.36780274593827533,0.7261378575284111,0.5100911709722534,0.3393147954022616,
 0.2458866047015964,0.198087174660165,0.5467780850556432,0.34137078068472004,
 0.8634970190346416,0.04248157384096274,0.6685083244444939,0.9002105372731214,
 0.6841690454298502,0.5984670661128466,0.6291238162086261,0.47456360765146033,
 0.5223721695995165,0.5325976909043477,0.5225329999548838,0.8343999193964652,
 0.3463639847121215,0.8204977446085587,0.9677751903124132,0.04076387078991983,
 0.8975710243684243,0.8075834222902447,0.07296049076442834,
 0.22928354874175383,0.4944353965815108,0.4769324419759511,
 0.19164427381171845,0.6738236571284822,0.3654045702454157,
 0.11800449642596167,0.5060530958134177,0.4378095073955558,0.9751734641367944,
 0.18671825540051545,0.8738252339712776,0.6207582869951687,0.6412264860855268,
 0.5825347554769661,0.9601306981528,0.0079429224362344,0.9707200721481052,
 0.33814263937769573,0.017499484480591243,0.4490317748717124,
 0.731610916317583,0.42828728001387495,0.1742821678863301,0.7283690664122571,
 0.8902117402740506,0.52275158992851,0.9413796672389668,0.8259806900169099,
 0.35036024742211347,0.7349112631761441,0.338933819102351,0.5102578003861316,
 0.28884185354754477,0.2466746901072182,0.5685973606124818,0.9386076989373232,
 0.9328328823573198,0.6677719781203153,0.6883880097037212,0.35636160074674184,
 0.8136290965261108,0.15395402765851207,0.6560389289671108,0.6702910387050905,
 0.510963042166482,0.9727960033199218,0.6840530584549777,0.10649119703194276,
 0.6817139205628265,0.1435007205223725,0.7451015652859002,0.8531622612135745,
 0.12480348468928681,0.23780682047227397,0.7558120317665268,
 0.4154462849627807,0.7884009708252754,0.262036917580592,0.9305443268921868,
 0.6282780264723665,0.23555856795473962,0.47496972707684026,
 0.0014685718721010321,0.17682822280334232,0.06800772968431079,
 0.2140699837072111,0.8120838854167166,0.8104262043321246,0.13035993509157873,
 0.636569526833137,0.1740211847715296,0.6911076339241825,0.10105500948940593,
 0.35532759222401855,0.6379450125503117,0.8268150593675418,
 0.09019050996247657,0.7708126404280436,0.49381940561187765,
 0.8207928910009192,0.8935011023661508,0.6009071353946863,0.8753313835990212,
 0.79032695165565,0.4672191825948532,0.8305619904759304,0.7245487478038592,
 0.4723620452749554,0.9964862320255388,0.23995235964922113,0.3850299564951907,
 0.47539484027761514,0.532413863167452,0.3430506235775702,0.23127496521313828,
 0.7358897516283402,0.2663954974009006,0.28476793322885563,0.5754159783474732,
 0.37129695322810174,0.9549508286368423,0.416828120346313,0.1429496604763313,
 0.22182972754793095,0.5453633199953358,0.14192884136145678,
 0.5396134481906482,0.8531790192104687,0.09610575160275447,0.9923828812563072,
 0.7533788768953266,0.11449896057277276,0.9257232384367332,
 0.33630754471703217,0.8531873753927559,0.9616674919620278,
 0.29150042851404834,0.5936260692922265,0.7349521592534822,
 0.09263232325235549,0.7184661445809716,0.826725102317169,0.16419383180845404,
 0.6804465413725514,0.7215429325947886,0.38493308346067645,
 0.37760514175330795,0.7758919624452139,0.9644335172400953,
 0.31517346095376364,0.34651497333064774,0.02327849704227858,
 0.642319035975685,0.2883446825757686,0.9216510122232228,0.8498832864021645,
 0.4796636369078847,0.07399212775190078,0.49209908906215005,
 0.9861169367081243,0.6306034474658552,0.35364358673330587,
 0.029540076010480365,0.2505844584513117,0.08163425741949148,
 0.9112199842328002,0.93890148113345,0.17404785042030013,0.279982824799236,
 0.23292071240229673,0.24345771720049725,0.255315161837563,0.4668710340327501,
 0.7897774844741579,0.044308451622271594,0.018997663323290892,
 0.5950182483933144,0.15649242922915096,0.31719635353208475,
 0.31578599517099615,0.263705623554911,0.8968374323752671,0.70892382594968,
 0.7008217337261866,0.5954470319754311,0.2919030328169113,0.22227693688929717,
 0.5105953252278411,0.2370467508001063,0.706041216160364,0.4882036057015613,
 0.4028692311265908,0.3615727477600108,0.0632721926696701,0.28061356416809535,
 0.8119374582903435,0.021735859237703936,0.7827486233359506,0.66216085972832,
 0.9016026272692059,0.2489887590864135,0.3229755118442029,0.864163098776371,
 0.008026264735201294,0.9552077283183624,0.5198556057676648,
 0.17858354878528837,0.3330813228405637,0.9366999620936349,0.4686840417665431,
 0.39143799902106025,0.8824997637425234,0.05568997247484064,
 0.4243148061305234,0.6082961081884365,0.9173453274628847,0.8559507996391034,
 0.41233229365056256,0.3682873030198215,0.2939765492388403,
 0.20470748589174959,0.9403786231693236,0.2555980341987527,
 0.17127651145753475,0.7934710285943638,0.44296209558789645,
 0.10719340661306098,0.7087622956673401,0.9550655245925134,0.9073952868453868,
 0.6373304804134534,0.40593524774987,0.8327847027166565,0.8334695591527201,
 0.23793272868943993,0.5659192690940946,0.7287941229205783,
 0.00219714072196342,0.7179447400517865,0.093136546220693,0.21824291702010157,
 0.9147862749345066,0.7332201646256695,0.09140170774594625,
 0.20446101091869184,0.7331090965652575,0.15727143971752156,
 0.7534834794081946,0.38294000706991005,0.11574669891396805,
 0.7332818146278273,0.2882581448753545,0.9072125104386012,0.9474489385611278,
 0.39614844927147874,0.5099498356330233,0.9374067558904353,
 0.41502899177439445,0.4833244140631595,0.35466561256265094,
 0.030957463063564505,0.36136263226769416,0.2687785328473947,
 0.6673318546003824,0.6918404302816471,0.45915769729384426,
 0.11054744195609356,0.26154378384481514,0.09660831963633043,
 0.8825035718523033,0.9716104290024409,0.38221816669872144,0.3998440367450622,
 0.8769593006457921,0.21285481046862909,0.28914666729141314,
 0.14819599054184895,0.34173939464061487,0.9690818500810812,
 0.15321564677305144,0.961174323411198,0.2157677453286312,0.3146469560517011,
 0.7777379292955179,0.39575792417190625,0.8085697514517585,0.468718165739367,
 0.8468919340724393,0.6659146939420912,0.9726959162056703,0.8776856697902659,
 0.11420162841572801,0.9088902237591716,0.8489491476467265,0.5885390764994685,
 0.23055041182850478,0.6704500016418397,0.558774527897195,0.5676688250385945,
 0.21355809526029812,0.8369892608652109,0.6928151588504377,0.262904120251335,
 0.5521008310333017,0.09433891506756509,0.9375916288952684,0.4386793935789022,
 0.17876879091011433,0.5307469179918776,0.36531296625979803,
 0.4394762332598616,0.8151062917682266,0.1246434873512362,0.5223109251135352,
 0.42594582093051403,0.9079517382605636,0.727793759886742,0.6189055728246624,
 0.4228590458919079,0.30200301871373303,0.4577851776155445,0.8619773459262843,
 0.08955036007380213,0.5857045125296196,0.6668141449407287,
 0.14780057690440096,0.98279244920945,0.3689776478351705,0.4484304757273566,
 0.8730977943954094,0.08986680888857845,0.07794312999346531,
 0.8534935201141919,0.8457315441752427,0.33691970036332963,0.1867233956217752,
 0.543472316506795,0.974925434346598,0.25793472217099245,0.3213912348251585,
 0.6489785145808163,0.015235001517063917,0.5136700261072589,
 0.36789827070990355,0.32143900210123455,0.5370323744501293,0.703221683515831,
 0.553625798076338,0.972685805003289,0.10526561400684131,0.2221679653372829,
 0.7531267726544396,0.9259097043926932,0.24124309842752245,0.2986878270281925,
 0.12288587154264463,0.16514596417995242,0.6466998233804213,
 0.7092154013239785,0.16314945678405512,0.5112909066795472,0.5505999162583468,
 0.38923093430118716,0.5957929423957761,0.7397290509652802,
 0.23250669221621978,0.04715389368572365,0.07003196732859829,
 0.9179489990691793,0.5545552547855224,0.5932188369081308,0.4874800068397115,
 0.12059290481942453,0.27057054157602933,0.8097629799336237,
 0.26928066187314104,0.16308560315507437,0.37579014436416747,
 0.0604126792942159,0.9812711881041393,0.6038898216879652,0.48831869479403467,
 0.9393750814130044,0.4809255116618547,0.9489085428705195,0.4486399622126487,
 0.11368272012190772,0.8091806482927263,0.9839417123973015,
 0.27950739904266564,0.21098412026538627,0.46123814477913916,
 0.16202141880467713,0.6326806386159253,0.9174378252584238,0.8033037824143684,
 0.9533221731441496,0.716845416829145,0.033415575160573585,0.9060749921515359,
 0.9717754151500712,0.5018772018838851,0.9025912650034477,0.19184286828637198,
 0.3680693823282032,0.6727240489705012,0.18857490180435643,0.4904643323974379,
 0.6386006464506184,0.5424700899478376,0.001072733720259178,
 0.8077797655380292,0.01824901169252302,0.959502800808723,0.16951394036329626,
 0.4030163131787321,0.5273166704062958,0.6673036396461016,0.3087372028088061,
 0.5711799322037553,0.6465971352993378,0.4052508929015317,0.8720001232658348,
 0.9639655223397556,0.015076281232159539,0.3362739253856962,
 0.7142420885422832,0.15530155229679765,0.44148288453197093,0.109966474817213,
 0.5009300929285837,0.2978508069236767,0.35321799476141846,0.3919864960020464,
 0.3439206858120051,0.03914155468180136,0.44811342156655753,
 0.13432393396361453,0.2874283400724227,0.03879742084800364,
 0.1031692523541835,0.7398512931962454,0.43543610654536824,
 0.32398695072608685,0.6284991893400702,0.7278623542803856,
 0.19536908739261283,0.35076729255226624,0.24418327276284202,
 0.4846584109734885,0.40379371967462163,0.07574531572395005,
 0.1098454132934007,0.10582518697659449,0.07601240064498682,
 0.9801090846804535,0.17624927998037665,0.6087505474828622,0.7459806570392002,
 0.6852299751542574,0.2524067909537199,0.11405418735794248,0.9890752432821814,
 0.383235472827525,0.3756544830742137,0.24163273366165638,0.4518820809357762,
 0.8089503419876881,0.08473184177269444,0.17458419834005012,
 0.44341792491171317,0.718758514083681,0.34898473202763636,0.6425445276378537,
 0.381146049079765,0.5134894157632874,0.8696565663601082,0.3867900836488488,
 0.29052059033367084,0.23759923873639344,0.45351105549737425,
 0.5027882554228462,0.579361116897928,0.23666844455355707,0.4414741065888165,
 0.237078276762695,0.5521828081609632,0.5551416998464875,0.5646975432536685,
 0.7958632426390378,0.40637676357807717,0.1664788153059802,0.0682028938190744,
 0.048095400050516046,0.8424370786612051,0.7009905650226279,
 0.7296790236276944,0.8788598971429803,0.8568483387085113,0.5433760397550671,
 0.7671003903807205,0.8379933933482868,0.3952376786588777,0.9196412711528196,
 0.42730673297820054,0.9529971174785261,0.1262717926997292,
 0.12378365950266845,0.261203042610371,0.9585468607235268,0.7676341074259716,
 0.9017234744875262,0.4044632817144137,0.5411821704763335,
 0.0015812987010288548,0.6316421630441923,0.3440239859236318,
 0.6910643790708665,0.4622537370534503,0.7641606877323789,0.46343728479910484,
 0.2371144917175444,0.31592114411771977,0.4950584248944595,
 0.31239606060734193,0.38782505753888286,0.45975672025239456,
 0.12108413420481468,0.37760258017355663,0.08362420491573053,
 0.20565498848563046,0.13292887011414745,0.8820737205703453,
 0.38766836202274657,0.8342792748310885,0.15062408325420684,
 0.4171873907685528,0.6000379232259383,0.5294400887564654,0.33800640753454403,
 0.8649544779286367,0.9266948442795305,0.4063036391724185,0.7426490186962108,
 0.8825795538993715,0.5400765021150505,0.6281127789171297,0.23571224634392818,
 0.8306082499050388,0.15435263107858732,0.9857524565030278,0.404693204881874,
 0.049155851624026026,0.5615590383056975,0.8375534175138872,
 0.0920849724693521,0.19953263968157875,0.663886087290206,0.03140201444512736,
 0.6243144529101822,0.06496481168046331,0.5922313965591199,0.725003153559975,
 0.5509342507880661,0.2018288265268242,0.8704018309293468,
 0.021697615091863742,0.6347231722011416,0.6988551526508491,
 0.6271261579667098,0.1582694080702346,0.4717880827467117,0.8064111598627342,
 0.32298589304117586,0.7808819463261064,0.7003644663103756,0.9204559459045584,
 0.5089105641792766,0.5935038759924639,0.1363137746961729,0.6465425716156086,
 0.2653942166922634,0.8796650252058076,0.0031823679029936436,
 0.9752710329814397,4.1494954166798603e-4,0.26902117235039547,
 0.1920624013206247,0.7522729943752039,0.7680415763433768,0.08491044476835596,
 0.9189864467938379,0.0726726635154068,0.22595734370023746,0.823342106055672,
 0.5291608278059106,0.8898518033244355,0.8691300932484656,0.17219518565338865,
 0.39195124439548146,0.7747193975208503,0.9748178907972321,0.8087721804011222,
 0.09966726064344078,0.4311144397811726,0.34745227411210056,
 0.5525427527182601,0.6863641104914392,0.4625479519774762,0.4795879081495774,
 0.5736340818396889,0.8987295414483338,0.04062889774324174,0.6060196502723156,
 0.38790204288531094,0.4427672646030182,0.7178989106418522,0.5982334582336094,
 0.755726552953234,0.5289347348532842,0.14237966754903797,0.6484710907421357,
 0.0936904501179392,0.2361691312941092,0.9102527085988589,0.0870527017761421,
 0.9959372496592696,0.4543076525533323,0.7315557306506748,0.3823802892085518,
 0.8171587725536047,0.5937135064796836,0.22148129736649536,
 0.48563665287972646,0.7674638031447666,0.6062800835311879,0.6094686312736733,
 0.5068922931851579,0.019357658544266076,0.8196092113899606,
 0.4666716592326209,0.3325537557511533,0.8794409693271258,0.47418681448731914,
 0.4177521209834866,0.8212641208838668,0.44607677201334606,
 0.36274402163564357,0.42396375172093526,0.5198989183254391,
 0.45487856389956904,0.001058585548854074,0.9501193973767914,
 0.27813391133630805,0.4512171822981238,0.5823166590854212,0.5891612119392223,
 0.1493886554910403,0.7225827103914693,0.13222440665814705,
 0.25740983665200556,0.9750125994459439,0.006789656569856861,
 0.07630818789005744,0.42351417909486955,0.9877494575423267,
 0.1349867361766358,0.707801758663634,0.9913216557107078,0.09666888834316023,
 0.2204277017189108,0.19080063540128722,0.39419134676233725,
 0.2169485029040874,0.7352963395025561,0.21465580786051675,0.6815139299081318,
 0.6225284004824454,0.3618699515812559,0.5678172110769073,0.4724669922880109,
 0.37826889080634485,0.9564405043654727,0.7067825105701349,0.3880258074529379,
 0.052123229864323495,0.14206512032643281,0.05010792305894418,
 0.6930245371596216,0.4071799427320948,0.6582424797891907,0.39698391167276015,
 0.8000431210169925,0.6398645213527495,0.7163457068942307,0.8498252371654044,
 0.5010875392622018,0.06978393743352507,0.12237393495594728,
 0.1063096259613392,0.9218207641331845,0.9058781221019391,0.32222364118217095,
 0.5731902629653349,0.18217768820983649,0.4777951489427673,0.6790528304786113,
 0.9335833030658918,0.21394384869715333,0.8200615387332797,0.3706765164776483,
 0.879677064632728,0.2927213913472577,0.3305813764960289,0.025095988052678098,
 0.7746416689712279,0.29020469423136186,0.7915806252906633,
 0.25205653889270074,0.8070278124237387,0.9012081486867018,0.7185049584363876,
 0.30876665656287205,0.9784663074430747,0.45853069843781946,
 0.6791592332110317,0.8634802960563613,0.8712784041368612,0.29058530519591885,
 0.7409066027557188,0.03434626510435412,0.4221642673944437,
 0.03682144472040716,0.549946010397184,0.13377151970589796,
 0.21606898135982866,0.7486746465152812,0.7342054727651158,0.7894197470452129,
 0.3770148295407957,0.06270291298027386,0.6410425787868732,0.3001438170273004,
 0.6338273223689893,0.41260165129869086,0.9281944328425142,0.2771898624176271,
 0.07761673539766756,0.3289753872072426,0.3552976510275382,0.2778832505765585,
 0.22621274010157455,0.4970930279795881,0.6775222369253974,0.5999311622970143,
 0.3444007473440325,0.5936127218093895,0.8253737220877663,0.5277378335872273,
 0.32365816824867977,0.3772548366748727,0.024508750168988014,
 0.5122482344529803,0.8013850401156918,0.40579003075529474,
 0.08331933471408648,0.3162324580197228,0.04586786844745594,
 0.5851253652035088,0.7049887699524476,0.2958511809603047,0.9488016101413148,
 0.8876389828684499,0.5924443280621537,0.007945167355884464,
 0.8072658857161289,0.040564893312565076,0.7928981789759122,
 0.6431607109639497,0.8374399291274622,0.41872433189600144,0.9281940171101247,
 0.04864055187371885,0.15657916022553886,0.13916142529682873,
 0.7263121618400733,0.026957726749599953,0.42724142293768674,
 0.6420766932964934,0.5067368689392397,0.696395893844893,0.8693008850882522,
 0.7562566200573563,0.7259767200990723,0.9983350050240086,
 0.026178744744718374,0.18501259726405006,0.9571656933287641,
 0.6808178438869621,0.5175519621074199,0.18935016362401336,0.792048889132534,
 0.7526608076348215,0.019646209691197236,0.2097592967554066,
 0.20778993645736998,0.8543131121433217,0.6482244486005382,0.6023855879360095,
 0.5164852633767102,0.5214347751823166,0.8130492235950614,0.6996544275895662,
 0.5392667025566105,0.4110777499643694,0.9868572963760093,0.7765426537599668,
 0.741701537024227,0.9387955606444827,0.698994315663965,0.22863080037983796,
 0.13975716637383118,0.742203962578152,0.1690840541871126,0.10434655519376901,
 0.6114976896838961,0.7131818025663685,0.5999459781212754,0.7681161033799823,
 0.8315461482323303,0.8406629079841503,0.7868399556544262,0.6493195797476181,
 0.4624756738465572,0.6079367304276431,0.4805273373807788,0.497153579942628,
 0.9890408016625544,0.7064756986533955,0.2597646615729259,0.36757558149498126,
 0.8550756094828728,0.9068097474571987,0.049324540663660876,
 0.5919931417655655,0.0558259896746609,0.7394254745835235,0.3813230772152555,
 0.6500825244621551,0.5020907072325356,0.5262880892843497,0.3700254867909709,
 0.6209833522267411,0.8490070442635979,0.8734486797725898,0.7196131891893407,
 0.5027090791597092,0.16846306012981516,0.8607134706375907,0.7824641184153845,
 0.2723366725836669,0.020019585439253196,0.636628273778612,0.978927555700007,
 0.049216511086968784,0.7970293626565043,0.20521604753159772,
 0.3504590227665101,0.6697365084101676,0.07205341173172874,0.7481301665356919,
 0.6285085159514817,0.732250875796403,0.1117313713155701,0.6849567634473421,
 0.5520673544829086,0.48366088171769894,0.4697687792605185,0.4481279469823032,
 0.9451110844923148,0.5332815114695149,0.9061381211347892,0.5062144672471289,
 0.24232974922431239,0.8450589741854073,0.9595342463072981,
 0.06052519693202463,0.4286128151582148,0.979436468910425,0.4223731585093796,
 0.14212944484394874,0.1745099047532811,0.696793980767209,0.3636870055566487,
 0.8953996854475125,0.18686467522825378,0.40058551672156895,
 0.6908419886176885,0.5868553153430586,0.29650970013316846,0.1467527001300618,
 0.9088013058475755,0.3890735388618789,0.36506312206512526,0.6190689628998095,
 0.47158714242771715,0.12666981955391765,0.7168444169819295,0.662431333557874,
 0.23432400373880258,0.21107807104289444,0.37420551305592153,0.15885488328066,
 0.10206216988999517,0.17155879875273927,0.4041344619707955,
 0.2410535770754293,0.06550146540025137,0.7168766099651446,0.9539425187748966,
 0.6845476875921761,0.6364255902272284,0.34796293748033535,0.7918354009152127,
 0.4226776441955048,0.4680842398948146,0.5713854814965919,0.7389811656057339,
 0.4920694296577718,0.08578811162858269,0.08480652210001727,
 0.23646303987845818,0.1325597449398317,0.6310953243570988,
 0.21229518728422303,0.8371251842951213,0.04335131357280064,0.881902572769325,
 0.18751230282895137,0.38413860793855525,0.9135009189956996,
 0.7804339046998894,0.4270843344446891,0.5845349497753298,0.8088751028814418,
 0.2719558751036655,0.8573431387534638,0.7142499060439902,0.2953808287769998,
 0.8508684784803605,0.4928306285012949,0.23676622285442506,0.3578564591887221,
 0.7315061221289397,0.5844087338854643,0.9552263282293931,0.08440032066228709,
 0.08405931125890942,0.10765950504955246,0.7202911982545172,
 0.35367657594959745,0.26835193872988605,0.3817006752002995,
 0.20008071179575826,0.5023656836971566,0.6331763893366261,0.618996544523236,
 0.31209558352875433,0.5873001504900772,0.4732086220892153,0.8473263394290717,
 0.9256246311256984,0.7761793725088865,0.11490214402391585,0.9439632984026413,
 0.8657225863004513,0.4545855247758287,0.5601795759942875,0.47064631830383596,
 0.357112532148381,0.21840573825645238,0.7135820891972264,0.08579122243852177,
 0.6649650510713372,0.033029688100953414,0.33607856646801604,
 0.18781190221353317,0.5801264286945708,0.03776793302833825,
 0.6428071196159065,0.7426058082977316,0.8619434497534715,0.13778589893540005,
 0.3522465942602131,0.6143848375471999,0.2279071616347591,0.4762717956628576,
 0.20009648262661217,0.3735207766253127,0.2500233712554907,0.9913059663500378,
 0.9247326984651152,0.8300190463405477,0.4832882836384387,0.8286645125737787,
 0.6113297399103291,0.30443375411382956,0.7753302173114262,0.9202318298075652,
 0.3960929928607093,0.3955233849254517,0.40368253821003897,0.5790645666687819,
 0.16048491676611798,0.9971023493050678,0.09705811207354353,
 0.7639595648657358,0.9231309865773808,0.7920412939763076,0.7190571821395434,
 0.1172105932429901,0.364675609076043,0.45767456160209297,0.5533383634096348,
 0.03221594584840115,0.6767596105812859,0.5386798529737754,
 0.17813489517334924,0.7316847829390859,0.1209102700196596,0.9755691674378094,
 0.16955063660697967,0.3474361164191193,0.5384569541583778,0.6886098990021965,
 0.9420210758099634,0.7978262496954452,0.8172539219695457,0.757917944042864,
 0.7000548506672216,0.055200329008598836,0.1614067862688533,
 0.2763448384405398,0.30847678242024934,0.4101120349791745,0.6105134117556149,
 0.9746735487455469,0.3590840267974913,0.014181625773667772,
 0.21768018338278095,0.7553220100006373,0.9053509592375176,0.0621049840473612,
 0.9533169329238961,0.9656582410286181,0.5977665366430879,0.23507109280366945,
 0.39545682057099507,0.11826228728127619,0.9025433425104584,
 0.8511001637643104,0.24944420885611307,0.27487742751255306,
 0.31913888046376127,0.06000174499817945,0.02251506406010162,
 0.17304647403245743,0.27478767595433595,0.10980596240633422,
 0.924437332515478,0.3777507330881963,0.7534549944182822,0.936225536742888,
 0.6214479044060552,0.33285970329556025,0.8967627736930079,0.6840341952841266,
 0.6798002096861775,0.16179493704524872,0.19000208625396264,
 0.8150110151185769,0.7906451941942625,0.26709930995272513,
 0.23664358669380858,0.09887693725746582,0.25189896804348944,
 0.013369060487292028,0.2892271028587763,0.28393099347519823,
 0.8551689401254498,0.24028969081702511,0.20448376621322673,
 0.7952597598821647,0.5602314050058838,0.31544072297705783,
 0.25002607135801413,0.6397098429224672,0.9622151587411916,0.303068816846079,
 0.6019020743121571,0.8691095134404636,0.42323632003171063,0.5679163873362809,
 0.19941641257524312,0.8747545284809207,0.16934759806367172,
 0.4580532202594083,0.9547171567276133,0.8766963064035183,0.8368257053147613,
 0.8253675639090627,0.2927145222691032,0.7035528078997196,0.13497200468093673,
 0.16440435752384597,0.6503915295784374,0.03570078134909249,
 0.7627624960247654,0.8458038144614938,0.09601642754153072,0.9598723258796642,
 0.19848168055192494,0.891039091333424,0.010912603785377839,
 0.3141571103243801,0.591199614903802,0.9541186257841211,0.5169515998577691,
 0.14414421488663898,0.32949082312747013,0.11496805628396256,
 0.40250898843563965,0.20321084705828185,0.21030977777204773,
 0.3126719621560171,0.28164722118498287,0.487323054462124,0.3792011217282002,
 0.07249783260420606,0.7793119376765586,0.4328487917845062,
 0.23751160576419916,0.05957789531566726,0.48011895999614074,
 0.5614247073058263,0.33104892540701925,0.3553262886140315,
 0.09826677093221348,0.9749468447140008,0.48655755379401866,
 0.2898273666047627,0.4342364395134021,0.23337083203794817,0.6632726614591771,
 0.31665511929530066,0.41630698367786545,0.9801128122966881,
 0.48254691673298844,0.6524434292432464,0.540356910803047,0.260173385609036,
 0.06956353856392505,0.7528437851748734,0.3735102379632721,0.4668671970368399,
 0.16921215583614124,0.9381171270348088,0.8706776151685209,0.6441222266324049,
 0.5839477844543323,0.6609872308403268,0.4642943169440781,0.5168653087556638,
 0.5256517668256212,0.9177277098571398,0.5078468381874388,0.03905329873466068,
 0.9158509878971633,0.9085761474449121,0.18093641612625666,0.2999953344680536,
 0.5949797934158592,0.7418316989608207,0.705091929732551,0.19342988213193424,
 0.5120207264678441,0.40363206669515295,0.8842514938423195,
 0.33782959267374024,0.7989819112181064,0.8335162672692321,0.7846353214474358,
 0.9037375478166843,0.7774894561252649,0.8784952441181908,0.653216754791832,
 0.29020977541824555,0.4831965750739138,0.8439654744186412,0.0821931083340871,
 0.8941020994402622,0.738296613287236,0.1484108746830365,0.2330630136689814,
 0.12202379476742842,0.8209233584246823,0.7156225190303765,0.8286774135866489,
 0.15765026781615554,0.8419970450592631,0.12729236720565118,
 0.47280281661649304,0.38541462381584934,0.5619132277502177,
 0.8293706115709621,0.459942882310557,0.04993990236416668,0.6421162942190716,
 0.2976907549303319,0.6975903356077169,0.21857240855976856,
 0.49135401109532606,0.43516346629797054,0.5141077544330259,
 0.7955119730304204,0.4357112361201346,0.9954018611951684,0.26129538160076315,
 0.9058488687636248,0.9278380164000815,0.7830809855613571,
 0.007919111475136908,0.8838747467913377,0.37603309654830963,
 0.9381231763309726,0.734894974776769,0.7515081044772357,0.5079677093836574,
 0.29983573597813973,0.3720271391024469,0.6305771894843459,0.8319606695899279,
 0.7407975894133598,0.6401324614660612,0.9533911681895417,0.30327616938590496,
 0.7592919260944464,0.8346792504924679,0.12162496198894102,
 0.0047154822054876755,0.7217097498903724,0.20099544689951898,
 0.7000010937672605,0.4155201643345887,0.9798962682549979,0.7944602152101256,
 0.8054106696782474,0.490720861636637,0.4031913918003964,0.7121785139671986,
 0.32571864719863763,0.24896693374134504,0.5268915054939908,
 0.10242630275825393,0.9264810749410181,0.5099696544426288,0.7383622305151487,
 0.6977284496982218,0.603050413402828,0.8384678673584602,0.933813943343571,
 0.4211800988925275,0.28213340436325174,0.9170298462679523,0.7780858733560903,
 0.7127448741020299,0.07109373964329846,0.6286547634545647,0.9946442045162774,
 0.9425405574611587,0.15509036355325523,0.0780946780724261,0.7937368339403263,
 0.8178181186946837,0.30365734829838886,0.7081491112381454,0.7477910851838481,
 0.055186312794669234,0.31388777289112346,0.6612487671679108,
 0.08176286744239203,0.18388936242560183,0.081970686598419,0.6433952626638304,
 0.1327374599821729,0.37636097691594206,0.942164967193365,0.315127158653986,
 0.44959015811235503,0.3833061755438011,0.01154655211072586,
 0.1271905537415925,0.6253949419443529,0.6811669152273986,0.5929567813444837,
 0.6953323387409677,0.2640088409157494,0.2211180537337638,0.4093280525521594,
 0.7165556046507328,0.45970991864737054,0.47035524265413087,
 0.9113248462828174,0.7781896195809741,0.6222745422337691,0.38777674538688167,
 0.2808557089419055,0.3074802407534055,0.7157853479964733,0.5327136135111717,
 0.5922080453291341,0.9150539131056271,0.21217537909352435,
 0.46273484113374985,0.5595790497437669,0.6639143977942004,
 0.11079156275824276,0.5026122018387049,0.709674580388816,0.8766937483043328,
 0.6199662894396656,0.7941692436221003,0.6370954506042446,0.8818970636255681,
 0.41338811411108156,0.5456175553488675,0.2584506486991167,0.8289402614187785,
 0.7660853681032068,0.30573068391734337,0.20856411914688033,
 0.1455180010149354,0.364537035009206,0.7309170779291312,0.12228747959221953,
 0.30953929495448373,0.6863498946426053,0.16215323716578034,
 0.6479130810374176,0.09155356579808649,0.05173110587336738,
 0.3835434868840083,0.842503377445971,0.3085060807226506,0.8272810393979988,
 0.6348507383583888,0.8503952940229274,0.3170980028324696,0.5571405262324726,
 0.4514689132867935,0.6748643645479377,0.4319199915820354,0.6180209708795135,
 0.13159521559525167,0.24929728166520193,0.4473418632755757,
 0.1227850484070645,0.11002542589934139,0.017600572472052378,
 0.501715870904747,0.011222724039934628,0.5346660798531119,
 0.29787289880175283,0.09597783402275206,0.4904411948066616,
 0.29336383892999374,0.5003445718412632,0.012120776602218553,
 0.013586666039104744,0.47842830793349245,0.027428760252708906,
 0.13503324548431228,0.7443843221347721,0.5412283385060473,0.4243902999627641,
 0.9501309747672751,0.9121251321604794,0.8172771195482573,0.9113677593126613,
 0.4371756886701993,0.2314326929803866,0.42283894987438275,0.1594263618453028,
 0.9522541476085056,0.2853968930815661,0.3618920073367117,0.45184817792968046,
 0.1085181485938912,0.6731071518164893,0.8608965907193187,0.24228844580554376,
 0.21125377554815206,0.6157849590674278,0.3434532469108089,0.2497588043742629,
 0.5329465756543353,0.8945862470816704,0.8895479187101363,0.8189250332338949,
 0.37821285175128305,0.7817907953602694,0.10122739224852972,
 0.9022053734289609,0.5729687491746149,0.838195056357753,0.5316901849430359,
 0.9343775745210352,0.8364196591428971,0.621622205103563,0.17705953391660856,
 0.9554583956299458,0.6803281910299993,0.005104146305332624,
 0.06372404528743325,0.3416970569814296,0.9124196832756611,0.5593242684747197,
 0.43162118358971235,0.15555291590820786,0.41990800505008863,
 0.45026562007396864,0.2314604358824548,0.021546210596308568,
 0.9513849435966885,0.39509387282462527,0.6523656507577152,0.8998077544012345,
 0.8301808657846209,0.2842781413329146,0.49967748395486533,0.6028776006353203,
 0.696509681039056,0.909533432491384,0.31510203809736603,0.6230500583959935,
 0.03762979891442564,0.48892855404994906,0.11056469763062471,
 0.8422128654954988,0.5775507064342392,0.4405604188763419,0.0551752576668576,
 0.3394754267827387,0.9308245331163946,0.40592002895496804,0.5866533307449716,
 0.6031640910862841,0.6361659164387035,0.4046589002675266,0.3059396708430091,
 0.12022982190941278,0.8550116588461135,0.5301904563576225,
 0.12838017982782124,0.6076125132202523,0.34148202746821155,
 0.24231811503427103,0.5779739080716348,0.4334564410427938,
 0.39074200534037073,0.5384482163349755,0.6853538921507272,0.5279732491512998,
 0.38835217127804866,0.8660045892305781,0.08842511016101129,
 0.7279581344182943,0.2390588548868885,0.49641855705492344,0.9888362757694774,
 0.33434615059001027,0.36688423535782666,0.12174827359086227,
 0.4481597866191903,0.25181203828714704,0.741663803443068,0.25735913036528046,
 0.7534025660902819,0.2860258798280225,0.26760527345824914,0.6022605432295586,
 0.48865405631085657,0.7321173819349283,0.2800750198711134,0.7058449204242061,
 0.5350067601090197,0.23040393673932125,0.6764606556464585,0.7371503783757277,
 0.9985138242992695,0.08739981037676092,0.8210128659813876,0.8154376651289871,
 0.45585463985680563,0.8181346311302125,0.8781119659157662,
 0.05187311758210911,0.3252204185776957,0.9755165040574035,0.964856264901748,
 0.3105167953057201,0.24890856292361108,0.10760225388410283,
 0.8079510146760323,0.7127808461734245,0.6758332366933446,0.7936530906603858,
 0.3586558779008666,0.063156013266503,0.49220831088364503,0.41918117505316077,
 0.29841963561860685,0.11014668877260403,0.2257666836863398,
 0.7342744074715857,0.12368918400256768,0.3275701066965482,
 0.004735367861167017,0.22504344584874336,0.39459670298563443,
 0.9110002082903241,0.2136315930027919,0.6613132429436159,0.18906096452750898,
 0.2987282944510862,0.24741534326612358,0.157417537348417,0.976526121889441,
 0.8687306421879339,0.41864476278374063,0.35955503899038666,
 0.7458792432697685,0.42044708282411847,0.600871160551051,0.834024966466909,
 0.08540938304030976,0.033405383995262605,0.057145241452718265,
 0.1743141025598539,0.348938041583552,0.3306239674275965,0.5253187040292471,
 0.3530730005486753,0.1994996898833442,0.7337884844487104,0.06640358229501009,
 0.015396014436793282,0.8402831996215742,0.21206751562723303,
 0.37163235573329056,0.08709662988438693,0.81094102452921,0.8207836255505627,
 0.3016764580529061,0.4051851762709239,0.8126921869098319,0.25571128939672927,
 0.9309727132926346,0.7258966127561753,0.9707872812126437,0.6243152078986366,
 0.4691697988877247,0.9184843245402161,0.9620136492891179,0.7448151644741299,
 4.414174097643331e-4,0.6840148206841882,0.838463324168558,0.2860364903092607,
 0.16888407646840575,0.47078735041341846,0.7984311459012993,
 0.9784366924920299,0.6551658741284543,0.24918897308490107,0.4043532631861946,
 0.7763393742122358,0.8927723533334343,0.6371557491136587,0.9743078212030795,
 0.9559168774920693,0.09198288784445885,0.28894598179489006,
 0.07108436853526978,0.0990637050849551,0.8692272945960141,0.6405009848239674,
 0.22273713862991196,0.44718844585225836,0.32584823042629785,
 0.6222167002618066,0.739353249234399,0.041360439973677776,0.5822678972542841,
 0.6890540995221798,0.47984864077410594,0.2177473889816688,0.3762689733083582,
 0.023112418202620222,0.5725007963695132,0.699887719306668,
 0.21016790351028247,0.29710550337778296,0.14862986216958207,
 0.10497096138390871,0.341140743941443,0.3484231417291259,0.7167439213760971,
 0.3999595084762638,0.32687352729491714,0.052778930613497255,
 0.7162854029792687,0.1134215014765565,0.06675242192927555,
 0.13345871902581352,0.6615574325451181,0.80584921176934,0.8043752565759417,
 0.19784188731783103,0.6729910843005504,0.5226892269993767,0.8873037310756664,
 0.3777895247230143,0.269046460802316,0.45436814545765003,0.12036422772578348,
 0.4716125024872866,0.23791742166992458,0.5759392740093151,
 0.32787820237640874,0.6247955134392952,0.36191573399020194,
 0.5479256769441108,0.05535862867583541,0.3402390125566416,0.867676839343815,
 0.6501814831104961,0.8705227477089463,0.49972645884679334,0.6857213299597709,
 0.36454616790611516,0.07395009429100718,0.11566082742094197,
 0.22870036328780163,0.35795873861052363,0.3427722625152352,
 0.9694238323907036,0.3594960933768532,0.057772705608170716,
 0.08189597834054096,0.49008094355367593,0.9368710940931172,
 0.8100157572243533,0.2555174048687683,0.4264080749183492,0.958201322954832,
 0.5290867138328352,0.38553442820583805,0.6683340247987394,
 0.13607871400895732,0.020566935098124683,0.141964651229482,0.795187697941865,
 0.1124546529712368,0.04211802589289182,0.11695221581656445,0.888866846248121,
 0.10468927785751114,0.7248218857905387,0.522063494919608,0.28355340760667824,
 0.5327473049159807,0.9084851361580426,0.07564897347581134,0.9393244760843641,
 0.1831365136893508,0.3654308369338073,0.10604278176050042,0.8702204499193291,
 0.9745604549163629,0.05545961261009669,0.29249860088678126,
 0.8251225854734998,0.25783433488770124,0.7964092420692558,0.7625980366301086,
 0.9232421493488595,0.956033082427745,0.4327796692571717,0.06899662878419743,
 0.4662684794332703,0.20949074567484827,0.02117836942325635,0.818118173115307,
 0.2154562585900186,0.82003678359547,0.3132013530807036,0.9925513409080848,
 0.37071950985860924,0.34731893050740115,0.5603952887731422,
 0.2895624935412553,0.7866913354877675,0.3451539708923942,0.10674346660468637,
 0.694773455730578,0.11050829941402796,0.16066861784735398,
 0.28362503828289265,0.3582927531227713,0.02279945577593634,
 0.5748282043257777,0.0464030622978312,0.47450646677698727,0.636921322287901,
 0.9958493784964438,0.8419632279455834,0.07599531120926928,0.72299363518455,
 0.6962917850398873,0.9280626990873642,0.16087401613364283,
 0.04484089205154762,0.43916276729898274,0.39496748357148737,
 0.9948985603923657,0.0346675667002978,0.10168937129643996,0.2309701509153469,
 0.967009118993033,0.08395331540681883,0.8758291066073712,0.324511152245192,
 0.886736789472959,0.7876021125117726,0.8584127829931576,0.667538443311212,
 0.8042338792941847,0.8156017255191979,0.5814356929720679,0.7669478238673574,
 0.5909770456409589,0.4268135181346864,0.823774982021814,0.4760090981848053,
 0.9677563920253385,0.3746477868421194,0.14475229155424862,0.8612170333107046,
 0.42524718022397223,0.29011275076071885,0.6782424182559925,
 0.43366447406673436,0.5921801419954554,0.7352865813746462,0.82889750352055,
 0.6854113491237,0.895139157416416,0.6784099892777976,0.452676792224444,
 0.8033137190239393,0.49998060357899554,0.7249431893957713,0.4137350309271488,
 0.33104377329670964,0.06040020816592395,0.3758353471979885,
 0.03006942613977237,0.31627587526620005,0.9612090580536967,
 0.8374404235561673,0.38366574586569857,0.043358869603578154,
 0.4682184866084661,0.7571175858160961,0.311168786621671,0.8284771578941357,
 0.21050748583557977,0.9561165927824522,0.9250783054723509,0.6371945127208686,
 0.5986558628296255,0.4022412644324601,0.020105541025154716,0.976117937645949,
 0.7278417285698476,0.3485305015714242,0.1707411387240223,0.31463919378181693,
 0.946548819199449,0.8218467484135695,0.7700635674941707,0.2827956270985269,
 0.7939046263111776,0.3788298858687045,0.3660685288410408,0.4550107828758577,
 0.29755555674266676,0.4369352720604426,0.18795260620789922,
 0.041769265544007284,0.43110124129306737,0.6362153611720799,
 0.028524483776046994,0.24945941076026867,0.5098948955261796,
 0.3829334513125704,0.46141980928831006,0.9156087807206368,
 0.46707008432727304,0.9048279950529567,0.4939321997572458,
 0.35378142218569386,0.4054696930508451,0.7280414821021365,0.9690710837067475,
 0.42020361402284767,0.8443965664882573,0.4562392940630645,0.5709054014335245,
 0.19424716644397588,0.3817667620869891,0.7838576368660434,
 0.40736308883538286,0.8546949681141638,0.5344600344414723,
 0.17162397953865027,0.1476945657741363,0.8240860497043054,0.2975977810554753,
 0.947071078634654,0.8032042952539393,0.8002079440984549,0.2552960964508043,
 0.16702070751135056,0.7021946641915338,0.65417241889186,0.14074883224297874,
 0.1270191314906285,0.7129887577255815,0.42846447745222815,0.3239266855161622,
 0.8929585729117967,0.47833053463120134,0.4816497873883826,0.7291531461290938,
 0.043972326137822315,0.5752972898742408,0.12725584683085012,
 0.05018075082701845,0.6509797776580573,0.35406475586755703,
 0.03458437702851047,0.6585662701675735,0.8033055392171575,0.1980176144185385,
 0.7852652675229738,0.6513812795596637,0.16999468618123903,
 0.12614326762403394,0.2160470561133283,0.22744089691228286,
 0.02124450801265143,0.8793442730850829,0.6444550251710067,0.8651048263082717,
 0.38239019635538085,0.9967450722001273,0.6537276968788464,0.5509897397388049,
 0.2869443110168537,0.9650508772560801,0.8606804269938507,0.5315780969837809,
 0.9315999238889126,0.9232515770334563,0.6283482745223226,0.8865478460806407,
 0.051216457023798734,0.577479335043745,0.607967113716874,0.05215908987874973,
 0.3350366893195521,0.8395347415177739,0.39862169716352625,0.6026018860897333,
 0.5967348522188766,0.6509125748937853,0.7503861656256796,0.20380180410854365,
 0.7882177648839657,0.6931069456123604,0.8500299926384568,0.640346862100395,
 0.21607252044169312,0.48271344485471246,0.2505959919450461,
 0.017006350400488923,0.19226146668372102,0.5077974103420325,
 0.953965893338246,0.07029296309343114,0.8010735352510926,0.41704497540517027,
 0.2398049609340912,0.6523118844481034,0.23331663215625154,
 0.34685376633379794,0.3834806400161408,0.5116910918335504,0.8264239142568631,
 0.1808888401112383,0.31540412705700727,0.9445927208841957,0.5950030664505577,
 0.8177052904885683,0.39208250726314886,0.5149974895447262,
 0.07656055291512553,0.10820299124852051,0.7494549993538987,
 0.8688605244645812,0.5635585321068843,0.5941441089669824,0.21930481804539492,
 0.9566551260696041,0.17003727210308806,0.27981001443999076,
 0.22892789364574762,0.9772678768474206,0.12343018557029406,
 0.45944771225402414,0.18411892544816577,0.28551018125590955,
 0.20102809552618917,0.39059486570956203,0.1486176267815491,
 0.2841448705789591,0.08309793737746962,0.004949201092922961,
 0.2864130743965778,0.8221485043442504,0.3653205918308774,0.5152157340875838,
 0.6374723168642404,0.19974336584081565,0.4507086757634089,0.5371374747164241,
 0.12376543817076002,0.4094971210907341,0.15628233040330985,
 0.9854404450519261,0.23855557150015938,0.5139706349176129,0.7643324649363947,
 0.011348864504693745,0.8473572196483923,0.9616098670621814,
 0.05593481619125851,0.05732403334493319,0.9985220843124974,
 0.6358167080490377,0.20731518046964725,0.6715695007973141,0.9953222522386204,
 0.6643083158426781,0.0779518934196991,0.7653694871546007,0.7121984661633426,
 0.053605852160759304,0.9889826822002157,0.10255430774638363,
 0.6652086451593472,0.9271259674846565,0.0731807551906608,0.2325960627288428,
 0.817541616389031,0.3778075536938732,0.3980662960459598,0.3963608303158128,
 0.20641043572523432,0.5631482624787633,0.1711451418251031,0.5087374238657805,
 0.22118677845420154,0.021954997016556788,0.7018880849122756,
 0.10083706626332278,0.23896686335118655,0.4805391675836266,
 0.23129898173710595,0.3118573416902225,0.6817137794656324,0.7998461363827141,
 0.7212987159500627,0.866548517231456,0.8977732670446248,0.03684444619553562,
 0.24468899202937378,0.694804554473304,0.8114618844326206,0.8289478512313303,
 0.8080553043000264,0.37541597943717087,0.57811067008784,0.516651530736076,
 0.17660268295087378,0.9860855957083019,0.9114864171775019,0.9131021528358185,
 0.6146038052282573,0.02475760210590927,0.5371370037669398,0.9927789849294486,
 0.20480174279777752,0.13413528672601627,0.060646477593097936,
 0.38265248603565183,0.582929410704486,0.02727194666593924,0.5477193637583319,
 0.4961707294355754,0.2647502668078152,0.6540099484193258,0.6058361771109835,
 0.5068873534365409,0.8217871650548703,0.7065265021109742,0.41332049790074277,
 0.15383224281807484,0.4189175112187773,0.8579860226624776,
 0.13034600484488001,0.40214175987663214,0.19278987356294153,
 0.7980534103468487,0.8557875733465671,0.6510657025464827,0.6185521750536424,
 0.09830438507339734,0.22414933672202875,0.02511940431993609,
 0.9623073216559986,0.47036821405858587,0.7743086465643959,0.4829959491989313,
 0.06317794391957321,0.37386514441069496,0.5525364001248331,
 0.0964557316809187,0.4524302236387703,0.14993859539654886,0.4187145063240437,
 0.8677119940586652,0.9793351007961201,0.7895119996331038,0.05745401938519523,
 0.9811121983792646,0.6280612927614841,0.29611542978635785,0.9428291374304056,
 0.07090637052314352,0.6382168589690054,0.9340655147281887,
 0.27817476141629516,0.849747365220447,0.7555278772831243,0.749218111698661,
 0.1497476233472117,0.7206942738873598,0.6522932120410537,0.7392699781287373,
 0.4900362083031816,0.6136873579305977,0.02284059108179648,0.4953622263501031,
 0.261344063109175,0.12871602527503334,0.1377569087060233,0.5881182913223966,
 0.16062885360049772,0.5418075577940347,0.005726167904675616,
 0.5116213732637973,0.48869080727714675,0.8340428234719264,
 0.30955780269670563,0.12535239028462453,0.6406262424893567,
 0.0498041143064869,0.3270480400369492,0.839477567555355,0.7904059562016744,
 0.5706944873075471,0.6277039883785669,0.44809871367496756,
 0.06776368985536052,0.7838363212781014,0.34137718581131027,
 0.7974820989387393,0.07291197047080056,0.0682822657818144,0.9794456407670337,
 0.17682463592699627,0.36138207624865926,0.4525925116408306,
 0.06980600471519727,0.6788414762419708,0.2102647541668351,
 0.33343413325022275,0.8694852727030713,0.8958697575534025,
 0.10296453697800922,0.6127505147552077,0.227611988151033,0.38818136446334417,
 0.9264846900005465,0.2716810261576753,0.8623233098785733,0.7206299179588271,
 0.8424507065644307,0.8751485929209993,0.11481924313547798,
 0.29340279814455295,0.7856255968933163,0.5109915469267853,0.7064323766699965,
 0.3768722418188952,0.7043860375803369,0.807361012676801,0.045623330094667014,
 0.7477067060322318,0.19813879220967467,0.3174786772643452,0.5463701706207732,
 0.3640405234831572,0.28968637744075676,0.3662565419764514,0.8210167616592068,
 0.12392936394106335,0.7480526613588651,0.22350861297595792,
 0.6654258338311642,0.8953166652548553,0.24054119880307623,
 0.36077345434649366,0.8546354374754564,0.4218785978200492,0.9501438999789337,
 0.5973612425573924,0.13501275864528162,0.4185733594208835,
 0.025282803081900784,0.14336617953365316,0.25349809896130493,
 0.4339296007696033,0.5339255594080673,0.12355389351232637,0.7593920008035768,
 0.836029795183896,0.7083665153508689,0.316339566695806,0.2575831123867408,
 0.9736340896653112,0.0972816088289139,0.9837847884766253,0.3949147536081461,
 0.8107681105859466,0.364900291266004,0.841781993150679,0.5515730311312679,
 0.7841829438590214,0.7162945040447337,0.3874113936648861,0.71598910845753,
 0.7705331106256196,0.156529701138318,0.9803879771502833,0.40424084157449247,
 0.8956327543732324,0.5650497148538417,0.36485207838729816,0.223503731974252,
 0.08406974803695044,0.42671059993483196,0.9926027898017799,
 0.9654499412562347,0.2857254378876347,0.2140443458545862,0.08901545816459544,
 0.6536875925956356,0.025763742800048117,0.9476159207872665,
 0.35775079661211373,0.5250101196955583,0.1774006313728813,0.921820146762806,
 0.0935995110930341,0.5197350479992779,0.7188747426613937,
 0.024773484461257045,0.47837658933500715,0.5480453509997812,
 0.8478362287611867,0.9757954802217577,0.20270850288434683,
 0.30677073844143155,0.9749646105607941,0.9519856597649996,
 0.07186694880739108,0.3462898789598641,0.5576761645656386,0.3163443883867556,
 0.8069585401971769,0.983780122718906,0.05791968828161531,0.5386085805888979,
 0.30941849442429437,0.5023167940009202,0.27533815828953356,
 0.5279669643823968,0.17060288420154102,0.9827898892903503,0.516566244607543,
 0.5799832816704372,0.5169305778662481,0.9002734856982149,0.3902266011888974,
 0.16757893410906421,0.36663240311925704,0.4826245071267917,
 0.6993109935509739,0.25923579560902854,0.04567214489399596,
 0.5825988500477824,0.7732647651940229,0.19716817558407884,0.5017534270648,
 0.6846962949418568,0.7619120806622598,0.7350149167763607,0.9057204832278785,
 0.4515475853853429,0.7744689115913855,0.2621789266065382,0.34536716879458773,
 0.6335227559884763,0.15949829455657205,0.20486026359153675,
 0.5683722756857017,0.25361036394363345,0.002686593853108765,
 0.5082709677661658,0.07460298093982931,0.14114641918201087,
 0.6114318171404255,0.29265263967452193,0.44331522425563064,
 0.4355570542646654,0.8063096095630647,0.6123161675373662,0.8751230367867263,
 0.4736169508709056,0.40803922598775244,0.07288489451235525,
 0.7708966596654339,0.04342322327024717,0.22583839834223596,
 0.010157309540204107,0.9329107672808902,0.6317320188479872,
 0.4931119093720565,0.2187160811611284,0.8747909066862137,0.2838826385656388,
 0.4299749072425929,0.364598633029078,0.08428320242506426,0.2915418217804765,
 0.8726307422360483,0.1824723713205093,0.8196133797524271,0.5619050606821281,
 0.8054171799629077,0.21258759240353808,0.3402008038360378,0.6125553304990956,
 0.5866080953509772,0.4609869125654479,0.115179590723264,0.10128450165141989,
 0.10122220747815791,0.8380945131937242,0.809860766510606,0.6085421744662408,
 0.35903695894393384,0.32760970980009774,0.11393695208939225,
 0.8995780015335635,0.2821899877084624,0.8454135117217428,0.20974964349579006,
 0.4214416687868858,0.7166880184428606,0.004910608657737625,
 0.5385261381053034,0.9655888678154907,0.3936174451911506,0.5337083606635343,
 0.7231232356283975,0.8359189825327074,0.4104056073066955,0.05058780308856514,
 0.2123376966874917,0.7868470817896122,0.8117266270313518,0.9515142724445961,
 0.32875949400141224,0.7175617260742128,0.797717569603552,0.1775553344680868,
 0.4415412528977122,0.7873319889592728,0.7372562952154441,0.2064874206639834,
 0.48966215256588075,0.8089864379019354,0.5252653506683765,
 0.10845223204732524,0.2622788920554029,0.1976635427305502,0.9376875319109491,
 0.9159684048442324,0.37750702648259526,0.8310684122199384,0.4861298174319907,
 0.6116666594316094,0.5518922553666736,0.55245676777345,0.994841268609548,
 0.64529337667836,0.7484965909773302,0.5428865010887143,0.6638599827519801,
 0.7645309728192731,0.4210871708603393,0.497495980002441,0.8578606694751255,
 0.22675806427733713,0.09727340152876873,0.7738560589838798,
 0.5083589913354398,0.15881480358791522,0.2244130088741222,0.7524062331565721,
 0.8674900027155577,0.5986629214369428,0.4727172062178071,0.47866421863266007,
 0.26787078268179954,0.9137743043566751,0.2855296190634442,
 0.15617676546380066,0.9987460234305607,0.08188766503004463,
 0.6626839935014522,0.16247389282615354,0.4132217135941533,0.9319044714588147,
 0.15292810165980786,0.8940893311518271,0.7894116752900558,
 0.14243133163885657,0.7447518507254116,0.8091880502171687,0.489135165597708,
 0.7185959832526757,0.43055628900188214,0.22378808268405803,
 0.6286075766134174,0.5417273997049555,0.2786606857744415,0.07191458236022363,
 0.04796079239191353,0.9084303843203833,0.70840599988394,0.3092499976930282,
 0.8271413325912385,0.18221274097135565,0.10586671973797968,0.632425819201205,
 0.8141390116850696,0.2808699891622294,0.2800940507832178,0.3352339937601183,
 0.9014899159527265,0.05357869517849956,0.49369712971918733,
 0.46926496299253617,0.3440338345688756,0.5620735626897189,0.5543667503364542,
 0.6303321516121899,0.9717502298756904,0.5004651966432425,0.42250891016593417,
 0.18576015051876382,0.0968534791307406,0.4500679756362421,
 0.35273590483360595,0.2857880863399973,0.48658549154189434,
 0.10341340301230684,0.09512127325183728,0.8587233825090969,
 0.2623290191486073,0.07041671603190147,0.1153552622007471,0.568736278509568,
 0.4898198466028344,0.14323102824822942,0.9696671775702108,0.6869434132102945,
 0.7490869181456551,0.7051301132578672,0.8421171223577398,0.30660728914003155,
 0.7388185757570525,0.8943442944560382,0.08579043640837525,0.7294124956656147,
 0.17034584687597443,0.37254837937437824,0.04306268210773201,
 0.13614563727304374,0.8717371436663368,0.6682428855513615,0.2843430129255681,
 0.5830611608211747,0.7630890150101916,0.056981014245778616,
 0.7888302350539967,0.022762479056687823,0.6188133218061901,
 0.5371015899542279,0.81153949006644,0.9138785060889871,0.7060276634287859,
 0.6192572021236243,0.6675689831727453,0.5892071040099177,0.7099614954096881,
 0.05783648239774797,0.4231033870704588,0.10238380716141582,
 0.7923423963055641,0.2891829073371778,0.15409012769596997,0.3101096866299746,
 0.6173728685314077,0.18873216641062796,0.06513932852375659,
 0.8565353174842825,0.53550599648589,0.10412073969902325,0.08584813746102171,
 0.2070645665563069,0.5778851123589489,0.7138169287619599,0.8127144529872221,
 0.4943504841143136,0.4744093580447417,0.08536582530799774,0.7605195483778828,
 0.14306008204856024,0.5732095399360672,0.8813243994579166,0.6210047532653191,
 0.9590401219821186,0.5482170515862564,0.8034204206492408,
 0.0025418481244178226,0.9109137487163861,0.27842489303425644,
 0.500682602721053,0.06360671139526919,0.8823831163002147,0.14779686349805388,
 0.29848125985786533,0.9290645634010278,0.16892904714542212,
 0.5340875121583298,0.7185956455505244,0.9001050134492184,0.8832623048947735,
 0.1831749514159694,0.24038187135225153,0.5104371399202188,
 0.04432060515138658,0.5837656995408478,0.1086142166939793,0.5251613528734155,
 0.25989119532483207,0.9927531488645492,0.8958304479204787,
 0.07849550622737578,0.539874814803034,0.8086475108049852,0.41509906583250156,
 0.07156647130181093,0.06680010327638364,0.3742099052358008,
 0.6925936117434861,0.3107869843192155,0.46627486974267374,
 0.34553904077770314,0.5596963590671925,0.1463098253352183,0.4603406709395257,
 0.2038719788933564,0.9307144866853054,0.5469261333646462,0.6102282444784978,
 0.37635740261408657,0.8360271891029536,0.8465786941070366,0.4659849838446609,
 0.46825443054124183,0.0040896912175253375,0.7044377712759751,
 0.9528316776382709,0.11657476290442714,0.6146504758030259,0.9393554292496674,
 0.5956919959607522,0.4204034362286667,0.5319659286531946,0.2710616430416213,
 0.916272072403453,0.10353226356878487,0.022812971415473893,
 0.12310652553293389,0.1663703883687414,0.6480905702089615,
 0.31429149097013975,0.024767972697505747,0.13715695975622866,
 0.5633346676498696,0.633474326581393,0.4968380766259134,0.711777144372693,
 0.9120963745221143,0.189447965215926,0.4519801105014577,0.6300741936620343,
 0.5276026550873116,0.2503375045374805,0.10217674215613115,0.4740744057041182,
 0.8556387024392575,0.12238629299242976,0.38967206763078366,
 0.6163094117729953,0.5036569673866825,0.15051814617389536,0.5254152981898318,
 0.9565124011657026,0.4042979985215136,0.9771644275687175,0.22514203213210615,
 0.8819688931819009,0.7496787531175885,0.20498806502463762,0.7644633952312025,
 0.7184696551698618,0.3564146142992458,0.9895014166623094,0.9962774769977638,
 0.9083302389429938,0.2801264074121623,0.4917674097055206,0.721460774201788,
 0.6955260861942795,0.6233989279083065,0.31219354683807665,0.4806691233611212,
 0.5258303787694338,0.8911718464420937,0.28929563956515814,0.3171443532725968,
 0.26951926969864715,0.31529717747113273,0.28407878459035485,
 0.7360006498173876,0.4649835730546841,0.2941949511240183,0.8246622838946681,
 0.14711532279865158,0.5978848476210312,0.8518552422732182,0.748273511880916,
 0.0612223802876769,0.6378953602640836,0.6859434309017272,0.558392604475944,
 0.8291768477396779,0.5176145878523467,0.4348672664114587,0.2852726887751855,
 0.8314263858843698,0.05295815319649233,0.07231080600323236,
 0.9283960898159928,0.2002852577940315,0.9313775157694713,0.7923132228272395,
 0.1513740072159082,0.9302154763596145,0.5109929920799946,0.4527608414308891,
 0.7447114053169303,0.16379870770605431,0.261074329015351,0.5555636235967716,
 0.661558478864923,0.9092748851582036,0.5268941814055946,0.9300208909104221,
 0.7192177197705287,0.9231505596844503,0.3612486862709273,0.7665299578093192,
 0.9466925049587674,0.5465573698262214,0.2785541423838961,0.8670768608252644,
 0.2542109985013735,0.4872834888232782,0.29537132704704283,0.3266595934001838,
 0.7721346664758961,0.41139144394221405,0.18941633472204733,
 0.012886349977671574,0.5589643894267873,0.981110979902549,0.6884104699971071,
 0.6249230601266951,0.16322651755473216,0.2774318400339212,0.6886575209952589,
 0.46622368691586336,0.387464537884484,0.6917933452132734,0.9502081273932034,
 0.25132618466880774,0.4371466802776697,0.7799856617398089,0.6532972264203235,
 0.38634909183660804,0.954634728604643,0.8619208461970524,
 0.059670460097322464,0.1605025954948427,0.11355861903350328,
 0.35250464795141645,0.2658295193865071,0.8996164741235883,
 0.08757195258333372,0.6425177334395503,0.6467586562437573,0.1687075187676642,
 0.8624024284442174,0.2705026687079236,0.2965381415254509,0.7951877047385335,
 0.7531974368563468,0.3462788780663434,0.5950164790013855,0.9699550185196917,
 0.08297773705994027,0.40660392784069654,0.2166842635536328,
 0.6753407440230711,0.9398871020201782,0.2336020868191646,0.5359610481712208,
 0.4154566868402917,0.7092883838860415,0.9504906438061119,0.8640830205897388,
 0.356753284254841,0.1600470754316098,0.6438994038337629,0.9299499115891162,
 0.8723621725062948,0.12178004527759823,0.9506065818157305,0.9658889254061572,
 0.06791618106103137,0.31604340221550453,0.30908432214349335,
 0.8525994504025227,0.730779808344387,0.3192654238749817,0.5518340079388419,
 0.8841390876036309,0.6358986739514462,0.13548811941618444,0.9274626784795618,
 0.6913565995079437,0.1208367131347381,0.8070501898062519,
 0.0024393152758309666,0.5679320045611209,0.25463648417857576,
 0.43203743141598483,0.7019065208308999,0.3955436520490955,0.710483760630282,
 0.6678337507602639,0.4383909137515821,0.30618467457876675,0.6127451175165224,
 0.7521920992983552,0.4247465325201921,0.6341528033108524,0.29292865280293934,
 0.7221303179103347,0.7175956929183114,0.16653364470860565,0.6105784694929015,
 0.8861079595678829,0.8997900742070897,0.17343146584675617,0.1324050651642229,
 0.09511518633478278,0.38389616818347505,0.7763424963962248,
 0.9994841884392339,0.21125523615659403,0.813260591812796,0.5554495473915149,
 0.8390864137461849,0.6961921629973178,0.3315220217950883,0.9879801335922561,
 0.5021142921775843,0.7201663173835195,0.8058006969012365,0.8630482526825092,
 0.25447408844337494,0.56215753934351,0.04917981054638998,0.4284655467530952,
 0.7642155624826927,0.9339442741757802,0.31115690892383463,0.8615556503911979,
 0.7078457224967823,0.5060216564198118,0.37424896522548956,0.8972269898526756,
 0.9406201173063213,0.7416982874659958,0.8917485274587412,0.508257887050985,
 0.655079124948327,0.5471646426814714,0.04325383298606622,0.3486004433318206,
 0.09472738442054296,0.5176730299233735,0.8317594870009486,0.4762712419397268,
 0.48022241401336496,0.4184635474523394,0.838476094221416,0.4956339868649622,
 0.05603707854720508,0.39538827490995343,0.6660217839914273,
 0.5465579014715279,0.4227501903524252,0.6048730297503908,0.11750484972805486,
 0.6430877722605951,0.5836789274720885,0.6976534972832574,0.9767723956527021,
 0.2359844332800647,0.3205713851674665,0.6939838175964425,0.7337373600047713,
 0.14262789523559638,0.977237178730773,0.8883072612854437,0.8454863386954261,
 0.33657050400042854,0.904114500602372,0.9966903817772266,0.2805791680467109,
 0.7350438024520155,0.031209290853822735,0.10891359680612245,
 0.526576204351802,0.9849979371700723,0.2847426876049005,0.04782811706383827,
 0.4093802267897513,0.7223245470731631,0.3225845881726189,0.9761174559961651,
 0.5910394500870075,0.4280476315952031,0.16257737823244722,0.0387018752312952,
 0.010401615345095094,0.8260103285057262,0.9237017005206719,
 0.33472220407663034,0.01804887562510049,0.5813048656945554,
 0.26768493857737696,0.20776450534485735,0.6473955479483265,
 0.5991502023796815,0.35103953461112214,0.6932314506161278,0.5623626908542201,
 0.06327592230589008,0.8712058356659993,0.26539261217145915,
 0.5999481897335274,0.6357572505948443,0.4270736968293194,0.37182295610201666,
 0.5069117852466608,0.5457372778045689,0.20995835054158396,
 0.13984771406252228,0.14661865692068488,0.11759195481995466,
 0.7122827867890855,0.49844877504335305,0.854095807100781,0.45956680391335336,
 0.9045425380373158,0.4621280184829615,0.754773682820529,0.11847899207293855,
 0.31579246250975646,0.8148930874419336,0.7558408542648853,0.8361963046553053,
 0.5225348843266047,0.40894107669948665,0.6811877435006658,0.7632642478308429,
 0.7061548884831915,0.9365011544839117,0.020194085189105038,
 0.9355518453701317,0.45736726763711266,0.8842218855975015,0.681110655439296,
 0.012292962326982737,0.9267886311337846,0.42135273874772283,
 0.9934571076767119,0.5195638070907933,0.006943410051268084,
 0.6116447859816075,0.2346752532444606,0.4862236290027764,0.8876975585751493,
 0.24000024658461117,0.07108704973155844,0.23539548586069492,
 0.6958256846117973,0.5172655066979864,0.37027997877050745,
 0.08145055991840278,0.1452378963593356,0.860629699587425,0.8831414195989251,
 0.695037350788842,0.30415376264093275,0.9874893046954357,0.3360564754084271,
 0.4217542110983352,0.20710474292257386,0.1531801851420237,0.7685430729938862,
 0.9987278452437436,0.3902558346237174,0.972150112680834,0.4286137533449478,
 0.12294652744506374,0.07658781853240693,0.9014225881216302,0.344369368711437,
 0.2930387087839832,0.19870485871293675,0.9634428251556585,0.7797159228412537,
 0.43615955230556036,0.08585929421471261,0.3067178455009367,
 0.015554279441460817,0.21498205918478952,0.6164233434746462,
 0.6830677516596282,0.20245139854357486,0.5067449153919131,0.7063716897787934,
 0.7154355401820244,0.005864199960049832,0.06056660743453446,
 0.3373609478398413,0.6283676012791146,0.0775465679240982,0.11056110792683937,
 0.237040380049506,0.7932483425103136,0.6506861851147212,0.1743226296156425,
 0.40795217110132065,0.6092044300234645,0.12152138366393117,
 0.5951200839458077,0.046815400511195815,0.5858962990905201,
 0.5727170443154008,0.25713965878362377,0.8397829973890374,0.5729206971167471,
 0.8321776131620413,0.6444759661921164,0.49411314448972066,
 0.22741859085229377,0.11862697771384856,0.9484464071906642,
 0.6327784499736794,0.578950492224855,0.7156925996103004,0.3087762933701278,
 0.8743574483033837,0.7356106984497689,0.32842936142357526,0.3268606632975979,
 0.19613828610913475,0.7617683893815113,0.5006430804478463,0.8787429335652008,
 0.2444248673929974,0.333840455107168,0.4952306855430858,0.21911560443488487,
 0.9158861906337908,0.15265429092751426,0.9012904549877455,0.8518243333864983,
 0.19180594518637073,0.8127326132638742,0.7984128425850204,
 0.07095376498026251,0.028550804403756125,0.5660994436474618,
 0.33084795599329686,0.3328083575907571,0.023562020201762346,
 0.9108276355766065,0.9498106805236056,0.22061813183890921,
 0.21257259642379356,0.3057975351764858,0.9009358021716227,
 0.25678306907085635,0.5048188667088405,0.014022127057802258,
 0.5701545630060948,0.05910586232915316,0.9767935745150356,
 0.18852864350989706,0.9893190511612401,0.8705735396411616,0.6795446626101149,
 0.4812700304461226,0.8377031833425496,0.6763755314932387,0.7691499648404114,
 0.17595331549419746,0.38127365372229716,0.753104519914748,
 0.010901087583270797,0.3980503708587292,0.22384913506826243,
 0.13248614388317292,0.3147854321399052,0.9568168020967283,
 0.09931415941303467,0.43181319235853677,0.7537079088226153,
 0.01761963246751319,0.6099365225971614,0.4698893448906818,
 0.003614060609282488,0.7858466795683934,0.8682516176823788,
 0.2525400216161138,0.8076622450789819,0.7691927545241666,0.5336991002050965,
 0.33156181658452877,0.9709349548378857,0.2337402955435125,0.5827545503821032,
 0.08504979203237362,0.27255140980214,0.3565274521463838,0.8210735422653009,
 0.6211983239036609,0.8462717983218923,0.8649042751888318,0.5773174656716954,
 0.8281848962083221,0.5264868736811933,0.7826467347601115,0.994910980327255,
 0.9274853070599998,0.8104557706416737,0.3227316348887912,0.5724471706041357,
 0.465214031628712,0.7658320395651534,0.6510437816048132,
 0.0013230619220424567,0.1725549359525963,0.11471950277576926,
 0.5054503667847483,0.7311675887587268,0.3913675836002066,0.9818932755707918,
 0.9547072557443176,0.04929023037926383,0.9291071040695981,0.5198590471971221,
 0.6767554319402972,0.20943574227422745,0.3980982896053257,
 0.47336457418170574,0.4002431382062671,0.8406773565419867,0.7347281987931308,
 0.8945758846255053,0.5274826634640744,0.9073023292408382,
 2.800107610062552e-4,0.9877556550303553,0.4502941219493821,
 0.08826524222221344,0.5008777533143453,0.6839247573199638,0.7614419287639387,
 0.8134159871152757,0.3772346295683773,0.4675398544201872,0.5375349216801085,
 0.9319037918765758,0.016936236911686442,0.9579145647031923,
 0.5992550320091243,0.3295137365845773,0.5670860138643523,0.2419233046845829,
 0.9812279069900602,0.5199720449188472,0.042454813794721935,
 0.4636655094091704,0.44595043515574273,0.7914351645955464,
 0.21505171820636226,0.9003557564096215,0.1508108879637261,0.603803017296681,
 0.22206607681340274,0.7181380875088277,0.8285331875308967,
 0.35283923553286123,0.6665998671643221,0.540840543999927,0.3377722424662688,
 0.10161222331673736,0.10352104202173595,0.19375417604502176,
 0.9915994667299105,0.5833256746354623,0.5756840179073499,0.08045871608498534,
 0.38865702656639045,0.3639297450583445,0.19191049205588662,
 0.8405153013405392,0.06109691798279693,0.6146741309137714,0.2051352549715918,
 0.9518486729038936,0.9125495489823556,0.8385530073012766,0.8292434260804767,
 0.9471906053442367,0.9439406433266555,0.6153599077885107,0.729094781558396,
 0.8950849675268058,0.6796383941737768,0.4514605946593875,0.5754564828336259,
 0.803227743592139,0.4312144272182641,0.6674748586206589,0.01911253698678894,
 0.6431042610763996,0.48488138568638717,0.30207590450758326,
 0.7156298387198899,0.39981592129279553,0.9543111682550582,0.9512601786576493,
 0.2606597948737801,0.3615374720071216,0.5285328908657615,0.8823943922246582,
 0.06568130593339427,0.9616338956662037,0.9925625327546589,0.2712950372919096,
 0.33138362086515405,0.13580733985494997,0.9286942034671756,
 0.6015351367263988,0.21432371828503727,0.4336042195630836,0.7761216347411733,
 0.8743614940013686,0.2143884198207212,0.013983835958923607,
 0.7602616055559867,0.8148059335471901,0.7937426211445167,0.8997796088786267,
 0.1392947572801877,0.4253407807948072,0.8904113795343876,0.972964291456206,
 0.6228778137214155,0.5050933143829257,0.024402428794623288,
 0.2558546004928619,0.5549163931391384,0.2821400462275404,0.5266334588272691,
 0.030745553591809527,0.377235326258348,0.41525752818558026,
 0.1287624429619444,0.8623722690390816,0.7932141505429094,0.7704096052752331,
 0.7955735479615003,0.9046905057200699,0.23044363122008404,
 0.04302623224460744,0.34496941894586186,0.9423181943365093,
 0.1065681707917191,0.22648227436279578,0.04492469639180774,
 0.036866655249390035,0.003780029489313219,0.0027266892169565127,
 0.44207590730745117,0.7131896209110878,0.21592009695574133,
 0.5955208320772939,0.6184076417576542,0.3817520191714553,0.9045937664254105,
 0.7965185596085133,0.15450247768598924,0.33328510575722414,0.894402065562825,
 0.3614632006353292,0.43788607410268876,0.20790156462842746,
 0.8180980859529957,0.0028156115149495786,0.18048648481583707,
 0.2355321949827619,0.12276968148004597,0.06973773955599705,
 0.5121572836168015,0.4999837803705395,0.23244703606888906,
 0.20877066159099034,0.335751290177413,0.29977382147015796,0.8944962869927278,
 0.2724040261961185,0.020670373004545217,0.4955666199735037,
 0.31635862905229917,0.5309138769819843,0.5435560632935366,0.4526664690471047,
 0.6505386920318559,0.7278338630256661,0.47893954718261467,0.6053898938733134,
 0.22874053626622948,0.4874825547897994,0.7825558737556376,
 0.48161448563570575,0.5047386959838301,0.2301206620677806,
 0.20113838422993724,0.6089144319050351,0.8552855961200548,0.3924548662408659,
 0.45453023058652087,0.7067885357347525,0.5504233683366785,0.7424967711354354,
 0.7344266985941612,0.32399676591496196,0.910745324743278,0.8288185488773263,
 0.09616738942551872,0.7007976195760849,0.8397244634601014,
 0.21942647597890486,0.4714342147509889,0.8053771421493449,
 0.29436899177290543,0.9347935171402839,0.8240225882279912,
 0.43755064592922577,0.811725892838475,0.2270810708256591,0.5054005364407674,
 0.9848171915213357,0.8237152501967533,0.6590499229995752,0.45429200374919654,
 0.46884388506873764,0.9669928941179632,0.6705006244364388,0.9867431416610876,
 0.32361792036007575,0.4250738114052438,0.4539047047596575,
 0.03744857808436408,0.23195380854769698,0.5035719360405904,
 0.5764957280691665,0.7046261952982388,0.3368785436268936,0.26637439342660196,
 0.5788137634290691,0.7252911177508131,0.8485960995860513,0.05651860954643906,
 0.4347351402790909,0.7295167855171126,0.30469022973394266,
 0.39096636043997735,0.9548242618156852,0.3385248727429714,
 0.14838467846075398,0.4398215350585517,0.6047155743945082,
 0.29197202432108627,0.3096762275193996,0.8333816654766655,0.7861717241251784,
 0.6939758584883474,0.3184100777756653,0.5239120209410171,0.7631520031051396,
 0.34775294175355886,0.16572047473453777,0.6818234608186575,
 0.3222985816070332,0.23989850350487085,0.11455854309647184,
 0.020648700826640054,0.056545040049623685,0.0381658265319178,
 0.9457234722254068,0.26878821585578205,0.9530482739622326,
 0.04652382224128604,0.986829799107406,0.48914272472815823,
 0.12300793485845574,0.6176908734357508,0.329193145601281,0.20131180323829745,
 0.8183888581756346,0.6756810256180895,0.1567172155611185,0.8527069149003027,
 0.6975838560771432,0.4576324224372562,0.9461165633781232,0.9246797470949532,
 0.16003218669775077,0.968782848247206,0.9630628761699304,0.9934216373251176,
 0.5630170039772069,0.6592293534456241,0.11782299893962667,
 0.12758126046196283,0.38355866716995235,0.33663509350209186,
 0.5653407682876258,0.11683419924554694,0.3622048822495709,0.7525165881652219,
 0.740750536326175,0.9865675607257101,0.3480734343180474,0.971666595284261,
 0.4257380421432193,0.5602092933119516,0.31031544921821796,0.7752354321982692,
 0.6522906559295222,0.5768747080441673,0.27441746953780133,0.8898609472553605,
 0.17555223904734296,0.556773911238001,0.5653744828582139,
 0.020584868772998477,0.9171635830470554,0.943229527694444,0.6715004601489547,
 0.4538831028209531,0.9765284112370953,0.7964521440472647,0.8573783937962649,
 0.5625919499963848,0.48454444938508245,0.40945367262009713,
 0.12978070390289775,0.46426904045233897,0.44097261187923176,
 0.3524016878482005,0.05081007661056969,0.3117247106313499,
 0.02501629746436307,0.21635480286541342,0.04554178034167056,
 0.5805381257072773,0.8341579422240714,0.33133180973928655,0.9373446640434597,
 0.3010247648240373,0.32722157258119844,0.9512240574675425,0.7736291519164691,
 0.029154404232274,0.8808008768946116,0.4910307354728195,0.8234561060868334,
 0.6025959910890459,0.40551878043510114,0.7597115477686279,0.9116434865154701,
 0.4651954981224712,0.5604979809487395,0.5031693790530588,0.6868847855088344,
 0.43586484420925975,0.008427472869902308,0.15810721385723558,
 0.5554274138275789,0.8879314781874312,0.09385633833268581,
 0.11294210226575219,0.8252306050611335,0.653266481303638,0.9137012355162653,
 0.9892725919330815,0.11703253379788003,0.506237281808144,0.18641102227431172,
 0.5902636240493526,0.13581131527057855,0.45755822902508136,
 0.12589648233510786,0.5017529655925277,0.32646620061013165,
 0.7605803086653512,0.9777713360362458,0.9173334252262848,0.28909114246388634,
 0.30573199737190315,0.5155411548928732,0.1092891894875262,0.9080313770199103,
 0.773214979149609,0.051901720002636265,0.35550527817146094,0.297808771744265,
 0.4694657578260887,0.859293567683236,0.9202999438152625,0.986828531799223,
 0.5182170595638277,0.9198921725466743,0.24421287625376997,
 0.18154414218425785,0.7781648081569482,0.30004838261990563,
 0.4736592246471404,0.9456174583650542,0.8659973295334723,0.05130711902958274,
 0.36179395007870363,0.6095664759972954,0.2669593462369493,
 0.13019793882515462,0.6386346252099204,0.611161162233581,0.8378286420748466,
 0.1404361423508952,0.3490656848150725,0.5252291180080143,0.33840639497008884,
 0.2528650732898927,0.801974614558167,0.5342694678559811,0.008432389697418774,
 0.5148986584996016,0.20463717611321108,0.8348886536368887,
 0.48021525252866504,0.2384479224680458,0.23463523585995893,
 0.49364447968855696,0.03044025775997905,0.11799980377756891,
 0.5113956723519787,0.945592362697257,0.2804170921465954,0.7882086998192497,
 0.31132109374408534,0.43296360000529477,0.7857735449444787,
 0.1968509136644081,0.6822496728786616,0.07617998269956916,
 0.33440830218402706,0.4156665820096975,0.22574635839263169,
 0.1811905990596071,0.8719051476004288,0.8459474133607819,0.7866267048888879,
 0.8638384734792764,0.7096201741483236,0.9096560508428928,0.4937507430617343,
 0.7581349742963965,0.4184044202495204,0.8009707537773069,0.3456610697509702,
 0.24618667177815112,0.46380508560341993,0.6340819048717747,
 0.08944626808840739,0.41770110269174265,0.6492846676758312,
 0.16183882561068963,0.8968650284135451,0.364203637180602,0.16013489623533284,
 0.05411581364717877,0.08946757360566071,0.3133648314750468,
 0.33738273099271776,0.18681654446088136,0.23865291190828453,
 0.567353290780578,0.46002708151146576,0.8823426292668128,0.4412625380820444,
 0.663316475766988,0.8425402911185991,0.3734462095217512,0.18256064333406596,
 0.5282024526725181,0.6023934386342803,0.3745483679927082,0.41082634079351044,
 0.41670858708142955,0.7904266081864428,0.3942414221846451,
 0.07470399862185262,0.4720573290942427,0.14246456722270207,
 0.09256107825993709,0.8651510891478277,0.6512630068080598,0.9840804497682684,
 0.7693626881681422,0.2833086332085454,0.3142825810370189,0.9993041610593241,
 0.26511788002503844,0.6903624379827719,0.5833833756292788,0.5448172324768734,
 0.9710857105585639,0.2092582575472025,0.2727339654013121,0.7324447349216993,
 0.8011311571833348,0.12472512875353836,0.4917578682756827,
 0.41543569891844523,0.6008941977241681,0.19732909384937125,
 0.8187279021662639,0.029710836900098125,0.2355314250132826,
 0.07291188950384075,0.3048170316402541,0.15627940765000725,
 0.23188486916463336,0.848439619630088,0.03801651704091469,
 0.15411413311600874,0.38812277679108953,0.8878073826574743,
 0.9416171160895644,0.9140729965868645,0.5823153044446951,0.3269197362444952,
 0.3479719595199695,0.6542296129043961,0.40829892864481554,0.7267591544522854,
 0.17360739674089387,0.038757272915381,0.006345959417869307,
 0.22141983999749926,0.8823214371466733,0.316087515146499,0.6487198307240217,
 0.4500880476482283,0.4557834043017269,0.5403641748270874,0.711145703651152,
 0.07018390672470709,0.3634985102483529,0.5978531888270124,0.9111696012069862,
 0.3460654520192722,0.6894993146126764,0.7982441968357398,0.36555516557766465,
 0.24272530192732444,0.5172921172261374,0.5892953481895102,
 0.23973322251935447,0.8899776765019232,0.8559190722604503,
 0.25413028939538507,0.2453120459220297,0.8811239215431371,
 0.21137217587755353,0.9827550425130532,0.9601129605339221,0.360332782605706,
 0.8805767295728761,0.9174953438325589,0.28489798689903056,0.2650301956238076,
 0.5383644376339234,0.57303918188445,0.10982160040958056,0.30204665890815463,
 0.661861973057476,0.7856223704399066,0.5484829755864137,0.11229639538417646,
 0.6453354069382422,0.34158523176055766,0.32395276347029756,
 0.11285469272097348,0.9672661759508896,0.9724147959967644,0.6974240395199517,
 0.10093781390252354,0.005821502512963761,0.10352879179292995,
 0.229950364077243,0.7077791442020724,0.213392125218541,0.7352635087602949,
 0.23373226125379998,0.7934735701943221,0.5290744576379023,0.9726143787548074,
 0.2767230026562387,0.6005003984674092,0.9007890442283917,0.6284927237885288,
 0.1990606800941701,0.8613449599381208,0.2298910384975834,0.44317585105314217,
 0.5238511708443301,0.7712624812788568,0.9950433847519692,0.7540586279043755,
 0.4649828753583648,0.13393117733793514,0.8576139108804888,0.9741111958624563,
 0.9722064761109838,0.9158989282956571,0.6466206394280823,0.5676863257534039,
 0.6555710821088648,0.2275863192594232,0.8763722543138766,0.9377617136159431,
 0.7220744601732814,0.36927469776973076,0.2217260717652948,0.3504914551101974,
 0.9719178584218466,0.8670469425970503,0.6963625950694657,
 0.020753153265609448,0.2921051910831408,0.23199007068020827,
 0.073782695424362,0.2388339916646749,0.20794899122968968,0.6740346454815227,
 0.9646218097913076,0.9084511696659954,0.37734911731444243,0.6653981835210491,
 0.9710038583774941,0.10651230034643577,0.9514133630913187,0.4408102571429724,
 0.9731780935894189,0.9078868656847983,0.9132392854135648,0.2785496287637743,
 0.687627408840145,0.7053025488633379,0.956309107680558,0.8558088837096457,
 0.7216146368316603,0.35475465568158726,0.366993637630984,0.4749220326330804,
 0.8355821127251427,0.4195446529916408,0.5208891388308127,0.25800335566101207,
 0.8668249193527031,0.0024440540125175714,0.24070841310374524,
 0.8970111786636336,0.4436924935464528,0.05192409129608844,
 0.30514292841349666,0.7460629290479525,0.4771695732923332,
 0.0055007383645406804,0.8013135245805799,0.38074526181660817,
 0.8511812204500862,0.6151691454771794,0.8896765413295827,0.7627805940898906,
 0.004533293977524777,0.6492199384165499,0.03139472882424199,
 0.7454436454451656,0.6467601528308782,0.23083045495675014,0.7396138792153819,
 0.32718612458638785,0.7959831764488577,0.7918301221559623,
 0.25927724434119126,0.9603705544558152,0.6392294646111337,
 0.45807982473601383,0.4339681537549225,0.9709379129502445,0.1757882633701735,
 0.6618439165891149,0.8843365495913255,0.5787546658739243,
 0.027673369190797437,0.7829983400844687,0.4388133559774512,
 0.31152522130421445,0.5246969380224303,0.48106110947514547,
 0.7360278410949986,0.5736550361773473,0.9416383845958839,0.7093269851926557,
 0.5011583971917819,0.1088867287920604,0.7609712562852167,0.7171527047603309,
 0.3286431649907523,0.9508259507314152,0.6359680885039981,0.6315135215222598,
 0.8895832461720367,0.037921257116107276,0.40304323262529607,
 0.7961910006868639,0.7223843008340418,0.9280437363122006,0.31452287497115283,
 0.4141698989900324,0.432878169322351,0.45179825944752117,0.26463384523892375,
 0.5694133495673006,0.2933615949823448,0.9134952208327072,0.05715448499341569,
 0.13576903971014953,0.7074687413729579,0.5692653921556452,0.2553372275704022,
 0.7197029620781833,0.03770469629177853,0.4199964928952933,0.5939912419701608,
 0.9578342009940525,0.4238837611240268,0.9114525430526146,0.10665806049206394,
 0.6159351843635704,0.9059730536117159,0.7929059252730524,0.6411658484062364,
 0.6559471984246946,0.12498597917449572,0.3127837183076081,0.2596040260737952,
 0.10423929229970463,0.226379028221105,0.5758496666787425,0.11631041394766639,
 0.09462122169124132,0.16203384347936112,0.4319836591479851,
 0.2672868862370703,0.02838618811365834,0.8981411249052821,
 0.34725567074838604,0.5114871066021401,0.35299963509938925,
 0.13095351192585514,0.7416918818521244,0.9644608193109097,0.6702489776817921,
 0.2638884624782205,0.26727699620877354,0.07485858431963355,0.629721295842768,
 0.8516543817496142,0.7287445952165639,0.2198426716366052,0.7293433971924673,
 0.04480374957119093,0.38039146008278646,0.1744485713130317,
 0.7002199534115535,0.44957726299853207,0.31481194011167646,
 0.4613601300869712,0.40146944126035855,0.13065745737715617,
 0.8089405671613706,0.1213280487703553,0.4435312018843858,0.7841189109425855,
 0.8384057313151898,0.6353841963762825,0.17581782888788344,0.9005287954454851,
 0.8725503784831583,0.90852486345652,0.7455986864246273,0.35148536821426024,
 0.9333390064792508,0.45852190763073986,0.7292323136009733,
 0.09841403615108446,0.3127598723183842,0.5231930549121045,0.6712693466937901,
 0.49613352340753747,0.024984713142352066,0.732348433270044,
 0.1489667282639231,0.07331264679068206,0.3759328650795015,0.5825038525537163,
 0.27487435037573316,0.8558677229250486,0.6720043745912395,
 0.07020390422859912,0.9701264240731631,0.5202338814019768,
 0.22364558039718818,0.49327702818093266,0.053086957162495785,
 0.7146862347653211,0.2452128168246095,0.8679760648795929,0.7746774767266918,
 0.9782974155749182,0.3461975772822852,0.4858857368350131,0.6510570396665654,
 0.7356246836885386,0.41754529460102974,0.09411347563368233,
 0.43484270426925664,0.4118836190035614,0.7501758613104366,0.2131224132157592,
 0.09114168107891973,0.2009612081518295,0.17438807860234173,
 0.3790993723273318,0.27422774795373894,0.32400626978263924,
 0.5878612102594927,0.8828817162496488,0.43334647021238126,0.9151290587813806,
 0.030642204167897646,0.1209222450200148,0.23200988680396628,
 0.45512606430407243,0.8989787769371586,0.8570706007531885,0.5207346627200032,
 0.8894896387352649,0.9770403089544927,0.49542623721864465,
 0.31722616855982466,1.934491837326524e-4,0.5037701897609916,
 0.8837898306909416,0.46029617782670174,0.9347733148506119,0.8197207212702027,
 0.7848420119793473,0.9800162391433312,0.33579272074969346,
 0.23915125138853544,0.48795550143150523,0.16637212341966034,
 0.33454663884616453,0.5394997322521801,0.2504215588214458,
 0.44758837379301564,0.15788474548621723,0.8892110528068327,
 0.7592883812081551,0.8852675462771501,0.7729790969872521,0.18799410732495647,
 0.21705044841113041,0.7511599083972886,0.13002466444624272,
 0.5958373293978356,0.6333762618248481,0.7369864843109366,0.3013063663973543,
 0.22033005313179832,0.269212922728441,0.357569921009247,0.35932650989313997,
 0.5417389488737171,0.6234515547268098,0.28826234178356147,
 0.16888833464453623,0.6179236529831019,0.9963428159791458,
 0.30966075120798153,0.9844714099232451,0.18615355198394257,
 0.9072005648643165,0.5891780312164122,0.4785528907429416,0.9867975227513532,
 0.09632419742675591,0.1286962483781806,0.34749054334918106,
 0.3184040179143751,0.5389031197562083,0.8437265547651739,0.9330370551648528,
 0.18391748856108392,0.045190854157940996,0.5662871323045485,
 0.6978028319075528,0.5130839378062664,0.2817524693484983,0.8137554830192044,
 0.5570499805708423,0.3291799692469144,0.9067043035413279,0.702032479463427,
 0.5913009047468253,0.3222751738198296,0.8989528114329242,0.05092792766990739,
 0.6682562686996103,0.9396930929592148,0.32713585730479855,0.4288623314854605,
 0.32473763846712933,0.6251231464590489,0.6599020283227691,0.5369591145763171,
 0.47891937088379055,0.5265555021008727,0.4397685771576617,0.8693731669140883,
 0.23742304674148718,0.27997981660340987,0.35103053810830986,
 0.8081521783432255,0.5409131336031539,0.1757578632771385,0.18434726107624022,
 0.00781645193171343,0.10523500790423979,0.9527647247527559,0.629300668477526,
 0.505500419665697,0.6101660752629411,0.9776944434490525,0.6895973189695099,
 0.7085442302294788,0.11568371283998102,0.4696473417218612,
 0.15585810738114036,0.3421861498494638,0.11926991242625895,
 0.4057755533815195,0.3583844285854343,0.9120466582012087,0.651701196719475,
 0.5891444728950155,0.31447163978908443,0.33314888316948055,
 0.40334743952409535,0.6261192463568626,0.3112609438579781,
 0.18772835623366568,0.35228258090185394,0.5152133750716634,
 0.7779965471623674,0.5953067888926741,0.523731135138967,0.17888432257875764,
 0.8125464676989207,0.6146476680716653,0.533237478292418,0.8349240444816646,
 0.8998581515737529,0.29566577859864596,0.9733012426103118,0.8620737680573671,
 0.6975698865649982,0.2100750468985112,0.12971587465611967,0.5285999878256868,
 0.474579245821355,0.5308928348415313,0.2536860415622709,0.020812795450334876,
 0.6203841876659321,0.9211600675032597,0.8261555135749361,0.802471271821586,
 0.03636118608183192,0.44092948847758595,0.8210582678287653,
 0.7608700445424037,0.27464995509376067,0.6952701775126833,0.7563191547219863,
 0.3018464227907073,0.2266468360621675,0.22984128664689563,0.702446059423589,
 0.2902157525486633,0.3435490030021624,0.2609595636158897,0.6567366543017645,
 0.2328720477446985,0.5788914948714401,0.9629581954964725,0.22017405682693836,
 0.8895825993611665,0.29612493236823423,0.8788542531721101,0.3480403678041345,
 0.7854837244669055,0.10120799701772987,0.9061484895008417,0.70766487275117,
 0.016496468693710464,0.4497322722324245,0.6634618650967411,
 0.2806270472415763,0.6235902626519972,0.11662683705377108,
 0.19813241549916827,0.19187196078646696,0.9218563958779759,
 0.29223516005089045,0.13820184571763094,0.31948347789103015,
 0.22597603273050204,0.963095272729925,0.9906778592973631,0.19729295744795272,
 0.4993994453139532,0.11511911452578405,0.023876916762842315,
 0.09608578489671404,0.847325698861917,0.5501598951066611,0.04226726542212034,
 0.7812124884816837,0.30579867518169257,0.8749315403557711,0.5366173553182224,
 0.8490540325397243,0.18874600063889346,0.6258003967442775,
 0.30292654247838424,0.8937414183074324,0.5080196265542586,0.7640129381063382,
 0.9881642055285599,0.6329694832769449,0.41090205804131386,0.8680342515049035,
 0.4706346914895536,0.619555602293657,0.4506014103421828,0.6913449748087763,
 0.3385632314663489,0.1109541805538059,0.6622753712174195,0.7116511149355337,
 0.1935555774025629,0.09007514903040692,0.8836362007462106,0.974347954759689,
 0.6398310416718556,0.9716288590537161,0.6216279436096998,0.1806728596124746,
 0.979988130073359,0.041154593470138945,0.1008915693693716,0.7718675075179738,
 0.22464715885330544,0.9456214918534804,0.22283489936434642,
 0.8425183097879534,0.28175195985828505,0.9977674980496855,0.8925198545425282,
 0.19868392653071365,0.5922552036943384,0.35340066240458345,
 0.012091426202011535,0.104694633507746,0.13157986457705528,
 0.12142414514862654,0.32106976182531377,0.19856797532883097,
 0.3711536029768586,0.058319837289551035,0.5443008925113044,
 0.3862881344904504,0.9470615618447312,0.4147529039119614,0.1764362239568693,
 0.5102327375691642,0.7511645964934992,0.2665899755890718,0.5695086261619955,
 0.9135288490196263,0.3716856254318387,0.9032313198561147,0.8152766335756673,
 0.4816592346609371,0.1284291069541158,0.3988648217941285,0.19486574156904513,
 0.21008863778445952,0.3661884900872199,0.23184232239128544,
 0.2979501507736799,0.40132397087342275,0.953426644527894,0.6808349390261077,
 0.40916517032858823,0.8540440585751358,0.2866389898434507,0.2318875438198199,
 0.9888879642044475,0.058582052181435196,0.03304949169034965,
 0.7962789159406604,0.07477508060684057,0.15942786143299126,
 0.10883263172886282,0.6074971232174757,0.6712786622953976,0.3201241485867451,
 0.22546266260775172,0.19654376590057088,0.7631301166862463,
 0.7572367445860262,0.38426257297495237,0.362435400493922,0.858383425994676,
 0.6332247914914885,0.4220223404114871,0.4953923090455866,0.942289723439388,
 0.725537016360558,0.6072883319625452,0.00530319133583379,0.24421589512880526,
 0.7376900329131308,0.7901046785578362,0.13634947678194387,0.803968989592773,
 0.46812094995047826,0.5405148579616859,0.9841091114310487,0.4312834937013721,
 0.2756744263450396,0.37803303703799596,0.21801038987717314,
 0.44554659282222275,0.6648286359289338,0.214915227587133,0.7770551513406133,
 0.7664026957976604,0.5308397423689477,0.17219688671188593,0.3948816711058062,
 0.7696519673505888,0.5401214072421502,0.9270090208388466,0.3810207246509748,
 0.4759343710804719,0.45552836366466143,0.8625918192590012,0.9362309440640012,
 0.7628119019491026,0.24030024097110747,0.7458897127030535,0.4314898173456422,
 0.801268221135055,0.09673539462063963,0.4530732204988618,0.5311369634550679,
 0.27823612803521613,0.08878120225943309,0.9714367316776071,
 0.8727921422737698,0.7179259253166437,0.5565666794206315,0.41151424494417266,
 0.49637315858575237,0.6137895833178282,0.016347331184234037,
 0.5426138620673376,0.8659891078250327,0.5541921228732564,
 0.027029091542726702,0.6440276412586707,0.8745438888667798,
 0.3883753208304016,0.6911922452602036,0.23986705119346663,0.7629258046165099,
 0.17196523829923993,0.1691995162589195,0.13334183986514359,
 0.2812504991983438,0.7898010811311322,0.41220325292695137,0.7749645596607838,
 0.016383672940292193,0.1400260058726409,0.9307758845991604,
 0.9655054610386536,0.10843390084004645,0.05430314028942951,
 0.8874323509282348,0.5270926016977553,0.4935152079481818,0.9971274063529121,
 0.861955574330407,0.2948378329933852,0.5707902598962459,0.5792716897023145,
 0.43635864401451163,0.2163169145075432,0.03514517440674636,
 0.06799894692561415,0.563115212925598,0.036972181568758034,
 0.49724047615495226,0.4664190406456915,0.6401600764068878,
 0.26300145308503853,0.5444472522897916,0.4818930044287315,0.3827583643369653,
 0.25718844732339663,0.6478584313448963,0.32968387722393666,
 0.49694723254019546,0.6022317852414494,0.5084015607490445,0.3003153503338867,
 0.18886287146090486,0.2587172153340308,0.3302465632361732,
 0.26013966843969005,0.16054926025510807,0.3245568680360099,
 0.8007921506895345,0.08938334259159308,0.9490489673934251,0.1467211690264979,
 0.807481096772765,0.8387719598565646,0.8793788389702033,0.4519682865059489,
 0.12155233403374588,0.908614888385421,0.5959998233936623,0.4930390165057712,
 0.8650464553814845,0.5188524936453757,0.40934331643513855,
 0.01604369880122314,0.48944750277878235,0.9418768903948135,
 0.7843070306269522,0.9330149712639594,0.8897114655874678,0.11826346153007683,
 0.05511148301094815,0.36764144185831515,0.5467893168389163,
 0.01605914412139775,0.7449578222477845,0.9714206990015581,0.777648114197911,
 0.5737740714398215,0.6054903210108078,0.03399152490551871,0.7172867236385109,
 0.2847995457193102,0.5823992123347923,0.5863696782354211,0.4823672056752426,
 0.7928027729710111,0.6943179482535786,0.4362844728111781,0.2548005079592204,
 0.14383032429495402,0.32598158937408905,0.49478138012086803,
 0.9100411499297717,0.46409001795371974,0.9462598957703312,0.556419881417369,
 0.565971532030281,0.9878952863340702,0.3421620599020072,0.08318191842196165,
 0.0936447553264661,0.9392008497951867,0.8582374219984444,0.07688097293219731,
 0.8307471656799221,0.3069461726154402,0.48005781245031076,0.8760487267484528,
 0.862175639892929,0.14042251472115475,0.9971379546441911,0.670852637416297,
 0.82183904012078,0.6273726008995545,0.44655579885590846,0.7151491091381128,
 0.9194900547650726,0.6243940716964863,0.8089579185664127,0.25452402442028454,
 0.8304414872815871,0.9563713624596425,0.30783259583181233,0.6343290725553722,
 0.6459983203459935,0.27314443903431007,0.11620273119185398,
 0.7459535462209943,0.532358982327306,0.37287135684250505,0.9304126351629937,
 0.7227959353187674,0.7319589496100013,0.7752067631634659,0.28823836934965197,
 0.8754745245210874,0.4598644041979111,0.8219257119677629,0.8738783038434865,
 0.24970738182773766,0.8967898259600957,0.25516991979405146,
 0.5622612943861701,0.7559613125253142,0.30090418935070884,
 0.41715812530672014,0.5517823559773197,0.9020163350212975,0.3591752177587233,
 0.3602109684163708,0.08628299913027238,0.7022102680242873,
 0.18970123569510733,0.3077622105808211,0.25467298612727773,
 0.07665212280096856,0.5829090143971772,0.35551014892458077,
 0.10648133861645626,0.9754191882393399,0.6766072177350186,0.496659310281956,
 0.9181634003891352,0.09809389853073835,0.03144448733472094,
 0.6312537710352856,0.47159669508208935,0.6380238380056609,
 0.08693487987446824,0.8714359304916813,0.6867016520773317,
 0.29076366970892265,0.30350150210572524,0.8986022428701886,
 0.7488336214322099,0.7363099529399911,0.8265991755945504,0.0431339520027324,
 0.25356400454102335,0.9557909649729879,0.9717581380554862,0.7083075616199905,
 0.29009740018674135,0.46358548205731764,0.7761227488913672,
 0.18644895483830404,0.030485547700374482,0.6695623814625867,
 0.9540213130003378,0.5944043218329537,0.21179843895048145,0.7104884978888395,
 0.901463746607339,0.27154901327010084,0.16313477846007074,
 0.20925917744505618,0.7408847625440245,0.077459242843533,0.4464372600369587,
 0.20741409562875246,0.66524554407943,0.29455559529865116,0.49519258917695197,
 0.4850857900292803,0.13647016796340072,0.7001346670800574,0.9501168852196626,
 0.1261817753288561,0.5916670207193405,0.7863566150091991,0.6101713587183991,
 0.7409184374526233,0.11247449810998178,0.6414514729416957,0.8018264694618082,
 0.9485090515931026,0.08361466303477383,0.6315962067609405,
 0.06338849577197192,0.44541251891528044,0.8429972296465976,
 0.1156475351922307,0.02675267618862387,0.512450356416784,0.508062687731606,
 0.7428993735925403,0.5634576782263406,0.044326420692290314,
 0.9491734064099027,0.6046384931895814,0.8765425765816744,0.3283263383800257,
 0.5252550326742947,0.5224914036994271,0.01567578943131076,0.2395176311181073,
 0.7807015467104578,0.7420195014097288,0.726995796149382,0.5524267511332461,
 0.06118945721341784,0.31861884905815174,0.27838681955803035,
 0.6494833801104255,0.9977446458090408,0.14922022336087915,0.3563976525468122,
 0.7779585829694874,0.31619965071164624,0.1790979918310881,0.1096267135677268,
 0.4773281961069492,0.5594835591259116,0.9662807846668722,0.7024659764687704,
 0.010995373898235905,0.9482048004882121,0.9433651096622439,
 0.0443086337668952,0.05980477070183665,0.708110167265314,0.13715084261571198,
 0.1429058101567119,0.3908836022986928,0.9813705581995281,0.6056216181557532,
 8.37006419393771e-4,0.9846131573307442,0.11529123947569797,
 0.4669635104764297,0.5638452491591622,0.26887063313069803,
 0.14253961130255344,0.8749615023642111,0.8924986109402363,0.8583534344348647,
 0.848124011932336,0.114198715531177,0.022222674912882878,0.4312016503732856,
 0.7458022834375068,0.05525525340490822,0.3611725331182828,0.0919779881372852,
 0.22032761287283997,0.4441133105074566,0.6810625980060949,
 0.09009004057728809,0.1152489510771858,0.8869322891505232,0.4656380839160573,
 0.0740889908920298,0.02005237772719959,0.8909986932941911,
 0.14002141207575736,0.8497078707230548,0.4098242075882741,0.7141331419007876,
 0.9635477254911273,0.2460263151358445,0.435972124910732,0.15171481660458463,
 0.48142021145066116,0.34075364406414677,0.06466263457610555,
 0.8029199396904486,0.9332380920698997,0.5735907532201431,0.08632685240400895,
 0.9789953279276616,0.9321780283831964,0.09537641277273312,0.9528989657231164,
 0.5699303235943641,0.024319360751354657,0.32170280379432226,
 0.14659279674721504,0.3532000471754809,0.3552496351833331,0.668042278688104,
 0.5786198725311075,0.0523638874686333,0.43047796876168876,0.6401634976640653,
 0.9958571821941918,0.3254212458902609,0.31759876308465684,0.7957792697031639,
 0.4848835871024324,0.12210786153069841,0.8130587758305844,0.4646851667023122,
 0.5726657623138078,0.5602418362940218,0.07327939674169659,0.7527799057121263,
 0.07649340982012043,0.27362866318062673,0.49727066745248905,
 0.3732331921044053,0.43649053605178,0.8723514097306788,0.27000644251763006,
 0.5431503896052058,0.9637895364041129,0.0884461417833472,0.2912987959120197,
 0.9324000137751931,0.39133125632160903,0.7805576173012195,
 0.09337427604635284,0.4505909429227244,0.21471562347356687,
 0.24840577945162967,0.0681327842436138,0.7501067747228626,0.9864741112763531,
 0.2276190510909304,0.9068054247862244,0.41952116576342924,0.755195421151458,
 0.07514619760668662,0.7158193541114104,0.822960439269234,0.12780080261219484,
 0.9263087569376159,0.5401327831598596,0.7260112937258676,0.3648451017270946,
 0.14482658289133732,0.725425662426984,0.7932870655469839,
 0.053403045896201706,0.697655993218097,0.9506413083657833,0.854979764085051,
 0.12351267144944333,0.875948830727311,0.6406160245722098,0.01709653812895129,
 0.6706466099970234,0.8106394231532795,0.19395191101078801,0.885380806667045,
 0.9191819060966548,0.47011201310464834,0.570000187426771,0.9678664685601206,
 0.5612559145371027,0.7754471299010703,0.3007605659810362,0.3950949966081464,
 0.1915911315113168,0.13838956952046622,0.41078814761302174,
 0.12324825992724575,0.13286571460206897,0.6265581546250386,
 0.8240540559554216,0.1888980952528696,0.9800848384636374,0.6307701515272606,
 0.7763310704386606,0.4393434700397958,0.42776467649736527,
 0.025061688033433782,0.7734526613304845,0.4305823177102871,
 0.7759866121089869,0.6780217652310129,0.6618826834193908,0.5724499809606391,
 0.82015440846346,0.0679743408600495,0.5183517425315558,0.08542667627369482,
 0.2299719145792265,0.6693077429390263,0.7432595598814801,0.06044149829691203,
 0.1766843268918592,0.5822459839748179,0.14569802091764328,0.6747116670644574,
 0.4267513908626608,0.02287098328368964,0.9506512778152612,0.9153001814660988,
 0.4885813244650077,0.7853294054995145,0.5250273996519742,0.39401372825026004,
 0.5747135221667595,0.598367991648051,0.8464724521943149,0.966890124204928,
 0.9833146628356918,0.41495471890360125,0.61990118189275,0.6149723399761596,
 0.5682437960129887,0.43100925006543955,0.7535808263699665,0.8810184361032476,
 0.585879593111566,0.8669576281193448,0.5861761339675454,0.2600166344847663,
 0.1678661524808368,0.3791639241591759,0.20341844042719737,0.9876531677663216,
 0.053777613162739546,0.16787825956434077,0.4385664767196672,
 0.28632825502201953,0.9965199729122055,0.4250785367945006,0.1760771271854341,
 0.3370301714057061,0.14277319668003752,0.49173394742517684,
 0.5189133223426693,0.11332517788274343,0.649603974369707,0.9533390234318869,
 0.030826858201661755,0.1452245312325614,0.32024754674775735,
 0.2522962977208717,0.42692721154708013,0.5222093485094814,0.9374353636182009,
 0.3475929068498733,0.7078732081611671,0.6204075668123982,0.8495489677178077,
 0.1693323775742792,0.036478613175644825,0.33105546671406805,
 0.7742120563100621,0.9703442833059375,0.37795409068158736,
 0.04376739773073168,0.5607503026436822,0.2155499594823278,0.0872316350398018,
 0.5744756042170839,0.47152588904341464,0.2598394113507484,0.5946661027510193,
 0.9815856122554493,0.10126380509346866,0.9391793539948672,0.6971107643126977,
 0.14651140143667996,0.9167442012966946,0.07194919079777884,
 0.9949406637371916,0.84581918147324,0.988478668852856,0.7255485261938321,
 0.5552793260660895,0.6562202154979654,0.8143797262709627,0.1868019170343015,
 0.34113548274624084,0.29320364186510117,0.10458474874369594,
 0.9813225005127654,0.9664128963026444,0.0723318475630289,0.33887780101891296,
 0.7962991528070603,0.9458117959407858,0.10742907362441767,0.4948646416325069,
 0.2992753442159235,0.25254567465911093,0.110741487765044,0.5361813101937364,
 0.9301675643010643,0.13103892803278416,0.3253914866709958,0.9732057107385741,
 0.670029903151947,0.4569234438167201,0.19249894802469825,0.11672920505664564,
 0.8234936158188522,0.43620263192390785,0.9037512909903264,0.3208118087408851,
 0.928292413052463,0.6284684671047219,0.9949126126770559,0.017926147830553685,
 0.8561346022941752,0.01892153855077716,0.7516474452268516,0.2485349259233307,
 0.5992218587418129,0.8268217308341874,0.4869210867767594,0.7540657172020552,
 0.16557505420691354,0.5961461213853266,0.33969025483626814,
 0.02267135314954749,0.3846385881882908,0.045287995100560585,
 0.16874380680436096,0.3242937191897942,0.06284824352951013,
 0.36843869445505806,0.12643731431081096,0.7984833429779803,
 0.30714472141210303,0.016413121882295156,0.24640068575902063,
 0.169418549935345,0.3621923866175729,0.3638991781527251,0.9178163782346753,
 0.08853738770396491,0.21354914098879818,0.09198309427131623,
 0.9341621817373187,0.8955852712472268,0.3356363004463744,0.06003361741588065,
 0.8933130079184916,0.07457772835166121,0.5262114087812519,
 0.23622704277568396,0.8555446763255814,0.33184156224175965,
 0.1418509036558665,0.9278879935058809,0.7305044703865132,0.8396524423166133,
 0.5404542130501189,0.6040100224832238,0.8430252028988812,0.2677299661201088,
 0.348864772427273,0.08160377250192385,0.42577320049690914,0.2466774267642945,
 0.029174352770555356,0.7233196571721007,0.8821317692999762,
 0.6523141601220673,0.12992239901822655,0.29654092422984124,
 0.09195070364677427,0.9800937592646268,0.6212050467889279,0.618319670488919,
 0.11280011373950438,0.8901400133830866,0.003079428635462289,
 0.7875822921533598,0.33726446477294036,0.12082314428452923,
 0.43287986156256053,0.42829176883137166,0.38539093448650763,
 0.7068024202018359,0.7746403603198379,0.18816488376351326,0.9644680400387678,
 0.6894409120125079,0.1803127582066204,0.013215912213613401,
 0.18060440942120537,0.3780669895954788,0.700203971726714,0.9865831761188395,
 0.1947374827214282,0.4654714523697232,0.8047846733215029,0.23514909042935572,
 0.6369869073178611,0.8582223733640942,0.6068102926989398,0.768205754095625,
 0.28167272811820654,0.10262329004598814,0.3563748301183962,
 0.6081920303336583,0.3941530060745516,0.5893851644046594,0.7195557608324981,
 0.77853676342742,0.640850966895614,0.46241139325417524,0.5575159203319722,
 0.5481967541501248,0.09186042981981934,0.06197854585322515,0.403043594792032,
 0.7328878483547704,0.7135709398566432,0.2823014947079623,0.01752846008724196,
 0.46071718633780634,0.46815789882302683,0.5520758419345592,
 0.3133388969777353,0.7617174555586866,0.4592180105841902,0.9766209332042641,
 0.008800122422206957,0.03204336700523658,0.5405743473263478,
 0.3617383294987945,0.046826605852894776,0.8553333801432951,
 0.9828855818273885,0.7436550306553436,0.6404223050014269,0.6680729988879115,
 0.18886548452078478,0.1673052213369921,0.26002750421663134,
 0.8839662793874138,0.41708201755228314,0.1195782420656386,
 0.00766038362027599,0.2572073332182252,0.15244283782535806,
 0.9404525656166299,0.6579080853071462,0.5690001507107482,0.7556630981779605,
 0.20854989854460615,0.9683158927171108,0.12136417400791366,
 0.9707189013720817,0.4946170828001074,0.5992538127495081,0.6433826882760559,
 0.594231809840264,0.7347555173380246,0.4254005319386933,0.307333032036728,
 0.8152686350059076,0.33889422563952865,0.09972275655575724,
 0.3249855845722731,0.4402437469819893,0.3739174091318358,0.43546853057984625,
 0.2675424850083161,0.67988390766633,0.500236926355645,0.09063649943335894,
 0.50679504740197,0.48417790862374055,0.272531388501145,0.8748233742380327,
 0.7631713900301378,0.3867703273439046,0.8721225648135258,0.1333510020552897,
 0.48071042105329764,0.3010010968654302,0.07746621407751808,
 0.9861760196029482,0.46221510800630516,0.6778279593670414,0.6603791517071007,
 0.6557784179721522,0.9417501411344389,0.1460069616020555,0.8746962806221829,
 0.9049397208211682,0.8689305673960089,0.008857687019331095,
 0.2277648838509423,0.6506272365074446,0.7715710501771423,0.787304610944163,
 0.3972251777936505,0.3931995512697002,0.08959459278674253,0.8622583654278623,
 0.3218504100308077,0.768657603388022,0.47129680363404325,
 0.044485477804172446,0.7861208576913417,0.05334411968373187,
 0.4325138840027076,0.3218116991064547,0.17587886735365688,0.5236766447749326,
 0.06703792609575121,0.18002086379405768,0.8477279660332114,
 0.5769552755311435,0.6846528090785521,0.9361544734104512,0.14999286967015313,
 0.5304048519597752,0.6256566178318721,0.3646722449284363,0.7644593309871625,
 0.05524876552283442,0.30937459374356446,0.7635350958694571,
 0.6932150909442224,0.976115119752528,0.8093933512235917,0.29964150370280096,
 0.4613448501858807,0.0865721603337749,0.2803748835666995,0.29304289230863034,
 0.07000007514856432,0.6002643398430245,0.3803798225260073,0.9223400779447555,
 0.33595225685550756,0.734726313892878,0.9130819597224692,0.2678588115935151,
 0.6024132055612985,0.7348006094234406,0.9051433593978051,0.08806969809603904,
 0.720596422629385,0.5059431145814556,0.1635398012514696,0.9758334515708891,
 0.47384643696859297,0.4121741286752101,0.29467821888936874,
 0.15273835304014127,0.43201989557404963,0.6209323241745991,
 0.5892430160844339,0.6304976565684188,0.43188882746962076,0.4758666431664764,
 0.21344445461829242,0.13638617720657586,0.8625127145601036,
 0.9095109324891322,0.35721524805034444,0.08340589537201792,
 0.004207085690960066,0.3582595130592603,0.35540247158824223,
 0.2664898382014458,0.9289127199329326,0.7281041733191683,0.1469843746731322,
 0.30031596056965604,0.8788506724854745,0.2917115939198799,0.3228629221193444,
 0.6944584629481182,0.5250610711653672,0.9247826626559851,0.05422598774112175,
 0.8484820452295133,0.15288535821832894,0.23267557098687652,
 0.09738911779057313,0.6187586482519658,0.06118561722372551,
 0.9985501198725293,0.6951174590198234,0.9976263594735613,0.6294916477687572,
 0.21458974483278248,0.510548529475453,4.140726129111538e-4,
 0.8257194814016175,0.9223933379276557,0.08128360635767484,0.8511349022892261,
 0.23704212272917136,0.2181504644439728,0.8563642176310016,0.5258453272394976,
 0.8903994912304765,0.9659674541670126,0.7115151741539334,0.398944135985551,
 0.27070310015951693,0.996079053159203,0.6739200587641969,0.5352221394945418,
 0.9196577171675561,0.5569105971511715,0.44962311447056647,0.2747256770944815,
 0.02075290564759158,0.10255825295355514,0.7869484470263226,
 0.7744174340169473,0.5496095663023298,0.5529459599210643,0.11868102236273503,
 0.7975905970545443,0.3295582473443813,0.6502717467268757,0.2090607529845363,
 0.736636734220027,0.015135292964973623,0.35685663149554536,
 0.5263407548834828,0.18641858998570526,0.4292410179467143,0.5016779719267388,
 0.3262820616690377,0.46811967946034205,0.08410549940107992,
 0.8068808130583879,0.2616301367434193,0.621581057671369,
 0.0031541983220297887,0.764114409681361,0.16785892735420194,
 0.8241618209025231,0.7003116640656971,0.3887969916684997,0.31890425037413517,
 0.5274836563942432,0.2900095997587002,0.24111216511067934,0.8126578109921885,
 0.1887674920403003,0.9398674543155823,0.7691263158500704,0.5629866267502042,
 0.14000896068393232,0.3139013455100329,0.6479775670470027,0.9435211562680926,
 0.7258047681958235,0.17136114887569953,0.26691412683178495,
 0.9708296713460008,0.7340360468561639,0.05643843874206522,
 0.06131204349034136,0.26282527542866696,0.4130816674119562,
 0.9276070900051577,0.962392794959794,0.03240723317858751,0.12006805917531682,
 0.8017274045360541,0.6501101540060895,0.9174681828734926,0.07196141640957399,
 0.8077971729715194,0.5386475352153095,0.4894483636239241,0.14317828284496759,
 0.4020733008620121,0.3033880453010649,0.5305983316908867,0.082742995595113,
 0.9817271096778782,0.06689768850474465,0.36609249901665275,
 0.04460754868265138,0.7265603499734139,0.30342365535879423,
 0.9736986065079586,0.002752200409143035,0.7404220387428715,
 0.6175565572541798,0.7072487485343849,0.2159768219585707,0.9181680469512012,
 0.2530710806159985,0.6457061197794817,0.7517585249930328,0.1176307376901371,
 0.36839848398984554,0.2362314538645357,0.4294302754179986,0.5988361153528483,
 0.23879629303851801,0.6449435812366576,0.7424673399659735,
 0.060327619834755875,0.7760740160062132,0.5686113280309005,
 0.8526708927281668,0.14664629242896166,0.7573881021879282,0.9717772853107448,
 0.4435515219632834,0.1423822887614461,0.25962935357375494,0.4149393026669246,
 0.9991619637857967,0.7691489458637635,0.028273445492398408,
 0.42663315467648244,0.252064838276593,0.07679417474895367,0.914405708094801,
 0.9983988269037474,0.39115432823818996,0.6164979640783846,0.5089959798137706,
 0.42509011158321963,0.4790650156699605,0.448499444259308,0.4428958422628594,
 0.29042658657206377,0.91415291423673,0.3997556853229539,0.42830197559436667,
 0.06939431244206473,0.5403056296088516,0.4041245328244335,
 0.32871579955437347,0.9256568943812513,0.30708071723324826,
 0.9771249092018399,0.1239299107960341,0.048394822622886524,
 0.20101549724090373,0.06801384961345203,0.14868236453737393,
 0.9614241496485914,0.5424353591998461,0.9806701926672761,0.5173881809525044,
 0.5616608566213385,0.440533118117578,0.6821114403873002,0.11593125817772754,
 0.8802742921946758,0.9919713695706391,0.9659382660029889,0.4856862343297448,
 0.6093825963615392,0.418494880000439,0.7533582011835237,0.46568390958223305,
 0.8128693336555947,0.2949990152958568,0.9627495531479608,0.48340678808871373,
 0.5897172498655673,0.9840901638136723,0.5369003807238715,0.9314445016444632,
 0.7639644936298549,0.0375402152776686,0.4265451541223928,0.5358962971597018,
 0.42326875907345374,0.2667723967575686,0.18109774262794287,
 0.41622230528882664,0.6678353570950302,0.6114242247932528,0.5030694501637053,
 0.041986238222557315,0.675336608156731,0.4885042207478496,0.1428794997710774,
 0.6811534876706246,0.5224750979725721,0.7898713329759438,0.9189737932600308,
 0.0966033714139155,0.4900397807221367,0.02974650056575745,0.9647424770224393,
 0.06355782367845697,0.3605624160112637,0.3492020981173043,0.6004241202363111,
 0.8581077515036829,0.6048432109573315,0.3653747917077673,0.2822080806857937,
 0.020950037221042495,0.11385611446353028,0.39040817985728404,
 0.9320714552321805,0.46231563948511245,0.509552633924714,0.34118675407808385,
 0.018350607430754007,0.577365074908026,0.7538858461492257,
 0.015248631868073015,0.547203596067834,0.7115259377974302,0.8378308269654406,
 0.7323793500079012,0.7307791628765988,0.5415962948203152,0.4663376533218492,
 0.7931764005851047,0.018627141020621796,0.49603794712778937,
 0.46060038760731825,0.742891843408192,0.5839376438391838,0.5802791371092921,
 0.6358630745872738,0.21307520959099324,0.020986738382770098,
 0.7292156660059179,0.6563461660544838,0.8523051101007422,0.3214933553655991,
 0.4525392993194993,0.24084030736993878,0.8474933956755767,0.8036395638518191,
 0.7756939528168934,0.22481145683149006,0.5056559985218574,0.8622493385110438,
 0.6861256211439498,0.8442782323616038,0.48521773447552186,0.9634396356593845,
 0.6934063517172577,0.981707765176791,0.9624859794951245,0.834032030758757,
 0.777773315535766,0.533040778683478,0.19166784371470702,0.9713377539389781,
 0.08247072271048628,0.5903189021071534,0.07656364624640777,
 0.3719830506677392,0.2812711961768506,0.3747466548795886,0.19396211626385285,
 0.27109035492509526,0.06752108495224118,0.6670146140179354,
 0.1305381633771847,0.6532387469398055,0.3577842522465644,0.07076738399910854,
 0.34872340493656373,0.3303526710631809,0.07545216665948251,
 0.7768473636880845,0.6277300632220366,0.9134049948843762,0.7113435938609737,
 0.5563953311571522,0.35026557253855883,0.9087961490774389,0.317793499310033,
 0.6173920518715905,0.1307852681317293,0.8092517951706888,0.6536600858912377,
 0.8505451193069293,0.011117471227798492,0.19686928349997423,
 0.2526851191249272,0.3228001705194481,0.4578173069075855,0.6621914176747796,
 0.8878118688863986,0.24052204022785684,0.6209645666265003,
 0.45775706970149155,0.03660786807888561,0.5574041199552563,
 0.39250614422887553,0.5856907339672267,0.9920120389744496,0.9300457793097456,
 0.8765631878379334,0.1831166078095886,0.7477163663470308,0.12833537152548513,
 0.4454974953806734,0.5821401836405845,0.3910619750031249,0.6977793185782921,
 0.8092075938201918,0.12508477302625365,0.46818177390871385,
 0.23407755184442092,0.5763727334518768,0.3985223710934843,0.1081682191419463,
 0.7229702233915256,0.24241888409183654,0.22393375837870066,
 0.7032540834361538,0.1248253222973148,0.9892546458130668,0.1702376167501326,
 0.20802091387047783,0.531865870764964,0.7751900040709989,0.46435509008449793,
 0.46069447849052914,0.5946102550351744,0.9798405496546263,0.6616297635888292,
 0.2928166285702116,0.2512285163702529,0.4061651751214356,0.6252266401246485,
 0.9354937082794633,0.9940971858056369,0.5085778575409906,0.0461742739329154,
 0.6803891604378711,0.8635481223578637,0.8614684876752909,
 0.016781589272599207,0.9173709245518926,0.32123472464840885,
 0.5149644765599838,0.29773078256096386,0.29038643106615836,
 0.33665003651818526,0.15461899141696311,0.7929554972722773,
 0.4373999157310241,0.11251122175120598,0.14922894433580636,
 0.8512418429192254,0.5703187677069778,0.6035490325524604,0.5514395249749268,
 0.040483889458156175,0.6853119085293973,0.8803850679678443,
 0.2545881173826422,0.0984038007882796,0.1291545210099373,0.9857051173944742,
 0.10255430825429168,0.5453035923465546,0.59032872461924,0.5559115058185765,
 0.9668697886435886,0.4693393185480207,0.7496464371446335,0.1398309609541133,
 0.5452731449915428,0.592891603877325,0.22742634129205996,0.21700146708025692,
 0.0012752660490230605,0.42468805495342776,0.2384840452494622,
 0.9503972547632324,0.08148955944062053,0.7267858700814301,0.9774417726258793,
 0.33042754366403093,0.38659760168697077,0.4247842730537954,
 0.2996660427673572,0.3906762199422924,0.44143439811763663,
 0.14057872886659273,0.38021410979070003,0.707590590466373,0.5958677548954505,
 0.6523030207933271,0.9695665306781845,0.8777818747937514,0.239803408982593,
 0.8782077693143598,0.7142782280777916,0.3837854358134132,0.24291695054243545,
 0.23793572022926035,0.23019869649279745,0.14551306268971853,
 0.5276714440680339,0.7611432137964542,0.49138314183486353,0.4271224807441346,
 0.48795196123534323,0.324588011301437,0.3482523200690272,0.9063740383724175,
 0.4676019341011033,0.27777265914785565,0.25730008202141896,0.993444128465889,
 0.2776845049171177,0.22980127679204487,0.7697983047335175,0.5021744820468618,
 0.6154096908972001,0.38570831786391924,0.3791970438586578,
 0.08657332203721835,0.7285449174053499,0.7359449453821555,0.3146005709638785,
 0.9002766508188229,0.7523266052122366,0.592112709874617,0.19076050042465897,
 0.6533759002486907,0.7765921003399961,0.13790197182780128,0.8927914510721416,
 0.24647297334095097,0.9559598208576343,0.10464749683425023,
 0.8330914552770945,0.4152170545683086,0.8028755318188491,0.2327748380405371,
 0.23056384491212467,0.7981276119660586,0.5064495205278605,0.6027818871445498,
 0.6991682423380745,0.5664024617646906,0.09820572155225271,0.6789900525325742,
 0.5570048252157052,0.20848631089377245,0.41102943520963353,
 0.6538459873964326,0.5129072255311776,0.2020679006765168,0.22037829796561692,
 0.0027411333633903245,0.3181501404962652,0.27879452041431696,
 0.1410126620868114,0.36671897959874067,0.7812517132720194,0.6433682378013488,
 0.7804632190612663,0.08174133975985465,0.48097537939527124,
 0.20912900873860774,0.865115702934967,0.4566531166723078,0.10680989469629876,
 0.9168062667225787,0.9375044992730434,0.2631894735652429,0.3530917161682957,
 0.43997993514214284,0.7209959887681738,0.5124605089462607,
 0.47129172331814506,0.01638386443852935,0.6649673126350097,
 0.4486623737216121,0.2438011145002671,0.8873971164628227,0.5563781890735965,
 0.3238620600636928,0.9251783397311176,0.4102961536587758,0.9845686095769383,
 0.22141942328733377,0.08267269712272152,0.3778269939611725,0.948885340250015,
 0.7755177644144685,0.11078184577576089,0.6555082143835047,0.3099370472145513,
 0.9743762341782621,0.6548543515196366,0.05786124964363215,
 0.01755533808995291,0.10053685801244983,0.0980108428591353,
 0.13979790773083467,0.19515529982719348,0.018136329180040822,
 0.5203887797230016,0.5085977280426027,0.39138015655921665,0.8261824670155506,
 0.5432704577788908,0.38032993284338024,0.6503366347012285,0.5148184710318373,
 0.8475154355772346,0.7131481778461738,0.40408838706312944,0.9444616135062094,
 0.4948945598677785,0.8039271757159421,0.7484571438008865,0.10328934158246761,
 0.22075165078405823,0.26539737458456436,0.7567116512934747,
 0.16072928930849473,0.2505776771749987,0.3243620595497293,0.3660496430824347,
 0.010383546784388353,0.32860898652063464,0.7391936231992584,
 0.5712554082234358,0.7268915396240696,0.1640054879456445,0.20669666721776025,
 0.3742104892095732,0.5719819808796034,0.3021799206404361,
 0.007949389048279465,0.581793871595381,0.9964495927971977,0.4548723494156597,
 0.3005247149977439,0.5804699119762737,0.2739878980117729,0.9737320942075098,
 0.1410624799799428,0.820670821179905,0.2728469985285653,0.2277531789056505,
 0.048578409579331794,0.961922006801869,0.9893554890225209,0.2055145180666026,
 0.525904998238833,0.5582302508979515,0.8048640112347407,0.6891251398847631,
 0.06431205901045656,0.17024223128162475,0.7768783719610214,
 0.3668372876966953,0.6187747696847836,0.10508629776749889,0.4398819948767785,
 0.8410617336548717,0.28504186148109234,0.3693022147039212,0.5781103130275507,
 0.14146843043468094,0.4922029904614613,0.9227460596739719,0.806070634453271,
 0.14090420976087548,0.54356151304222,0.7901552436503522,0.543318416131558,
 0.006717150275624806,0.04654169276713549,0.8729615174407422,
 0.7575132245272287,0.9802632736390904,0.0194551847457739,0.27602246433189226,
 0.03631998683022164,0.5658570089002928,0.24642302521898785,
 0.30186171866416744,0.7970710722405345,0.8202501559280522,
 0.19405949181221227,0.9488323367271977,0.39335308606962094,
 0.042365361842180316,0.9496767211377849,0.9605087372526641,
 0.31489000642189824,0.3117713246145317,0.21691861294224846,
 0.11145074246302644,0.08538929648254145,0.07629974067745693,
 0.5492485705731166,0.7476453335763841,0.3376160468911583,0.5402888795513704,
 0.5195108829888548,0.8347627727548573,0.15605831227558575,0.8962710954716028,
 0.8041293274430188,0.0198195299507421,0.2681660799876866,0.8371620515575816,
 0.886664876610211,0.6028148438805451,0.6736239767948003,0.31993528957027584,
 0.28397330913720964,0.7154978434018856,0.6523847304470888,
 0.25201315316313455,0.013655909298606494,0.9096762385846598,
 0.9771715198773049,0.8200452671565867,0.18416425193018193,0.7696465189532192,
 0.4230229141153844,0.8143160231510687,0.3691619463439335,0.8990903159973783,
 0.8959873997482779,0.384672296319998,0.09505744640340474,0.17241311235354517,
 0.8692348199167186,0.20103656390105928,0.8467108072687293,0.7643868975953275,
 0.11653257776405335,0.7804025265711008,0.24358422795219525,
 0.28025664301381004,0.6920698749984773,0.04011803130923297,
 0.4139307774067058,0.4917820243330522,0.5314920345020321,0.24723261456340972,
 0.5581063656693894,0.8528357130923134,0.18543062807513178,
 0.27435360305121137,0.14782817685187555,0.29633403676188497,
 0.15298197828234905,0.8765750154226098,0.41191735924455886,
 0.6079074409678986,0.38662560056112605,0.23501323200380586,
 0.6370893398230346,0.1802539550219776,0.5663973397206193,0.3434264379219437,
 0.3125615045852672,0.11406805821069577,0.18991161947978363,0.262592696731764,
 0.5966109088937109,0.3995172613703871,0.5645589894732431,0.5938891183179247,
 0.5930319995703262,0.5759248993189825,0.655410794063241,0.7882667266747578,
 0.010233690664508632,0.5989120306396937,0.8777070028343451,
 0.5518030995999332,0.8533103676007485,0.09496811911266811,
 0.12705954850608225,0.3333647155313677,0.7914770163558137,0.5482158260984806,
 0.5385804713532396,0.7049504429583481,0.4956659233817141,0.6249824793957156,
 0.5071871109398097,0.07195065822836144,0.6805316350581007,0.2398100297909529,
 0.09260097099452713,0.4291440698369633,0.6624429958828011,0.8571244898217363,
 0.4516598126015293,0.5017786036331806,0.7295660848858645,0.6229481964824375,
 0.4867448543450057,0.5803531888130794,0.5553126257965282,0.282002957624123,
 0.7745825785954895,0.5732009558060097,0.4389109096762448,0.8077166331577772,
 0.7012107505160715,0.7218409711478699,0.6009937303908568,0.30686049379151614,
 0.13454413589818381,0.6264241359344046,0.6925295242472778,0.6786157462010448,
 0.5795508191733473,0.39283264077748015,0.01364476143892035,
 0.8753919147402772,0.5397084815741007,0.010855339530680075,0.310306017084676,
 0.6311374563203621,0.6870736807242213,0.37089281600896207,0.9465334604574512,
 0.35090034844845674,0.8339834723088293,0.8616688442428379,
 0.06227361626852668,0.9431876360180663,0.8751092206126402,0.6406064890507617,
 0.8202608726137715,0.7350338941067257,0.005965175136679268,
 0.0787825829010047,0.4262011245816403,0.9202178030241992,0.15127094442402145,
 0.2114067900582075,0.5682252354442834,0.7774926398765747,0.38606673406355385,
 0.046197860086467646,0.6779484880256801,0.6559501640564085,
 0.8505689086881492,0.963624216855147,0.5880290803593734,0.05820494777050511,
 0.5856249023562329,0.5102654695894765,0.14297645467318087,0.7856801328793158,
 0.16529545102405707,0.5655552874436259,0.49210415954757214,
 0.024033952287296878,0.7037965530245804,0.01039048056942049,
 0.6496958439693881,0.6570381622575704,0.8810941057723941,0.7955118354666734,
 0.6161090084487557,0.3824345759221437,0.40295715481194094,0.2590032208331464,
 0.8326869340495048,0.8876596471539769,0.24630921402889738,0.5285040080487916,
 0.7340187920263241,0.38430053463651215,0.3610227445118521,0.6453063623886526,
 0.05251978752234454,0.6963575718250683,0.41811581364374695,
 0.2897980173240975,0.6378740836658383,0.5439209958068254,0.6399003168253973,
 0.16781816435481822,0.6587522282389476,0.7378486269131007,0.5930294997897136,
 0.8084073105576588,0.1039817037646188,0.15212076025151555,
 0.11957480822725852,0.5691974856158662,0.6637777192445052,0.910059551455846,
 0.07424020084390592,0.9611940282590382,0.013694288251712594,
 0.3881321918569224,0.6648729256870483,0.8668359414535621,0.8896099214293468,
 0.047688506539228515,0.8625419968410059,0.7422509843405858,
 0.5317128880058989,0.3672387559593042,0.7458329338424355,0.1496266187721832,
 0.09969228153507148,0.5103793436833004,0.09626468369336438,
 0.24245856886278905,0.8524676746545412,0.17357953546328886,
 0.7847409948227503,0.6759879174130341,0.6566444280746856,0.6430614470816336,
 0.8619324018859996,0.7911226935715125,0.5517671416708612,0.5712270261769635,
 0.7916265449410744,0.9951960095957375,0.44623988778060664,0.9414218076661297,
 0.31818258153753176,0.43887065295875005,0.08159663860791255,
 0.8731690313658671,0.8760959931544688,0.09461699299066806,
 0.01677325890455794,0.6612075308280183,0.47061964288688973,
 0.2684282076404071,0.6385569233186771,0.628569390650335,0.8779344659374074,
 0.6745108382852669,0.5814570792396535,0.35933473917708425,0.5656837503207409,
 0.3786222202998655,0.8868414407264322,0.5050131738931378,0.13899372014364486,
 0.00447729761717941,0.40625417340595593,0.2533587943809016,
 0.4873111764678585,0.34135783726484004,0.7148955817233036,0.4560895044992781,
 0.03957940334179855,0.17070316992945245,0.33395419925380565,
 0.2953444852412215,0.5972800541044954,0.12608345363938234,
 0.27999334711901513,0.1395264970954313,0.9299244394971063,0.554613555416095,
 0.9871974362691994,0.32094302772448724,0.664362562797889,0.10182041580619594,
 0.694704681760937,0.6206472682352955,0.5644532764054968,0.3980862544658724,
 0.2661039681024986,0.5647323605479825,0.23035127969898062,0.7959003739028115,
 0.3928309871779815,0.0643083549459792,0.5430208527731826,0.5228857971386802,
 0.09245986187039401,0.2683753465004246,0.013629287105389465,
 0.8947209128527749,0.07827160994611981,0.09180082823338187,
 0.5727290444639871,0.18447475316566397,0.22913831837472576,
 0.4917333240824526,0.9511642731599942,0.9120048526580147,0.5032661314939317,
 0.3514768610774147,0.2797878235625968,0.09869053633482372,0.7546451677329911,
 0.054003310705245,0.9044435966543741,0.13882624853969316,0.4778020341229705,
 0.9677718846531129,0.8200769790541277,0.7374194282366581,0.06531860353931729,
 0.6930171223662179,0.7738555428548517,0.6954179051605313,0.6064899104372731,
 0.38459989861974875,0.6013924463053577,0.6596367866973865,0.7747106228674981,
 0.2949181403665945,0.7837459563172848,0.251632551780241,0.6328188324482744,
 0.6650313679631344,0.5070761894432427,0.23773679000033987,
 0.011265669574362258,0.9134676090394647,0.2641137293529465,
 0.5525112403981284,0.8438150141564817,0.10374173253818075,
 0.005424478714260328,0.525122083555665,0.9414916666486091,0.0869176056557156,
 0.27123569954424354,0.12017479367690231,0.040502224299398315,
 0.20756006798500215,0.6550416331773341,0.6254091501106028,0.5485492513649654,
 0.3418384271535059,0.4370829977155386,0.3250082201923379,0.7280278945287193,
 0.9619776365168211,0.926908801188039,0.15155315362963018,0.4639505859570845,
 0.8343986006892531,0.9671777985719221,0.26296262464395503,0.6574624206223705,
 0.005631197603852556,0.038263872653787945,0.5849793934657883,
 0.2450574886598983,0.909101892521689,0.334814405676375,0.5404358577870693,
 0.3868170542324286,0.10593436119144761,0.4598654232309377,0.5049207884058721,
 0.24321799471352112,0.6205700418899437,0.3465846733555198,
 0.26070782113204616,0.5166851728738309,0.1362386085013002,
 0.14582666465788685,0.495735224163653,0.059660240967714606,
 0.31323399047807254,0.3796009858699745,0.13598068345584313,
 0.9176758568110971,0.62837232527508,0.7901932601089883,0.46462482711398434,
 0.07771915773896076,0.031219128014552622,0.61383705779518,0.962535081890745,
 0.14828053441218392,0.3090885654543618,0.16267066408939201,
 0.5987725397031791,0.0635777317933891,0.2914337697092291,0.2500845761848456,
 0.5121371287329287,0.7746945270160046,0.8999366106641954,0.17471000732058228,
 0.7974990187180255,0.2463530133212144,0.830103145972024,0.05583696558033879,
 0.928172061733457,0.6588778125547271,0.7384072000341045,0.21842010568378611,
 0.15173348723468583,0.37392827196003053,0.7725436287391227,0.244492943468,
 0.5761855794982068,0.21404737612380942,0.5198764305335686,0.8189573288124511,
 0.34149035543129835,0.5812233830001116,0.5571216374317982,0.5188224307919254,
 0.9472737899578066,0.886666281729717,0.2979735416267131,0.18807776078988103,
 0.5281911607128742,0.467600005701583,0.8446028145487704,0.6742709295722276,
 0.4145473922322135,0.8409603539601793,0.39898462178798,0.791146803550484,
 0.9828582442067428,0.9552917933079635,0.7305528473028983,0.8192615029368748,
 0.9054698190654991,0.25950466569397157,0.8365532048418555,0.8772007069913981,
 0.2912644857962856,0.7636966182755472,0.6918536614483372,0.7653495862303732,
 0.3959048324368516,0.2360433536623845,0.17763743557495826,
 0.48674690598270254,0.34731063838606957,0.7767553323068432,
 0.01924296583887186,0.6771408688881341,0.30936382740452606,
 0.3825394664785111,0.1300889273537389,0.7366961550518836,0.7039396022642695,
 0.7325389071976233,0.49665749780696566,0.35128328418279653,
 0.9151628896443234,0.7409715780111861,0.7864470058851821,0.7688073369815275,
 0.22819240350286418,0.4453931573959382,0.11700955938931057,
 0.4003059808736723,0.26742885973954944,0.5512043219144073,0.9093211597787474,
 0.3297792389464398,0.31540613404429296,0.3098499616275775,
 0.13815178540326623,0.9112323383836696,0.6290527130767748,0.6559363630888815,
 0.6540944895674258,0.37075636437049053,0.47184523964661784,
 0.5889785746554899,0.08772940976759303,0.9565135507659897,0.7258048420168643,
 0.9171038339002414,0.8446934573966068,0.2140187915236571,0.1792887823965077,
 0.35837349432179366,0.95482706448803,0.02307835706061334,0.17175942863266092,
 0.9121120035871472,0.27180158227617823,0.9834916644062505,0.9302180671060872,
 0.7093169327146673,0.7256401986985285,0.9298323399623839,0.3714209925225249,
 0.9196947751483554,0.39015243196174143,0.81426168327847,0.9755395920610954,
 0.4688611446810337,0.33442841013973046,0.3943456587763602,0.578412328239837,
 0.8328626078367365,0.5553166600341807,0.4028805922266414,0.4876177056970715,
 0.6611692319155383,0.15560188096368144,0.5058682594943432,
 0.13610342115373597,0.37334942206247324,0.018617938267054157,
 0.06399487610429888,0.8677675883624966,0.6993803762907529,
 0.36894029491727587,0.616932863387995,0.35065149345891466,0.8978774625628944,
 0.8443138369261362,0.021216733030574364,0.9405235558484135,0.282521894125093,
 0.2522524279036741,0.6592578220329282,0.26211491300791856,0.8185062458379014,
 0.7840683600002196,0.5134640629100716,0.33446994061841095,
 0.05623800420419123,0.0489807711808945,0.29889806382834205,
 0.3650881000445131,0.7195488708338436,0.1327118581321588,0.4753614345399104,
 0.2796952707377881,0.755072321293317,0.3045869509752208,0.7097606272154853,
 0.06727450321703377,0.649468711056064,0.4976525541645449,0.7642859988512454,
 0.6877809385320676,0.9541122737025418,0.5130420586822697,0.46455819461813364,
 0.8116850475022652,0.01823547998848518,0.6241581382814908,0.9040967282843339,
 0.9587319169060391,0.7744516677278026,0.876106708202967,0.8667940661789526,
 0.23834558393272576,0.8052559738791036,0.7705098803113816,0.5489165358793494,
 0.006212366165184058,0.6304761802750021,0.8596318700681859,
 0.47738925304436686,0.8521682989209478,0.19026475508487994,
 0.38096533405236344,0.7741297046184797,0.35944806043011335,
 0.11267098753570948,0.9083908676154572,0.8370873609456069,0.6501864635175107,
 0.717702867014902,0.012165656963604943,0.2615382105456958,0.5655051853369183,
 0.2955453335296879,0.7297293885499051,0.1629424842700995,0.6900551203120472,
 0.9756876488300694,0.9834911975370703,0.12326172970630489,0.6862429711497691,
 0.06525618100899022,0.9953197062273914,0.08901891550000351,
 0.7423007118576941,0.11950099949486248,0.47395092435454145,
 0.0877295660169205,0.6352518074917577,0.6569093020039286,0.4518707443412219,
 0.3596419658358494,0.7680435137217884,0.36930535962169786,0.7626115435885783,
 0.3160210350055389,0.32618235131404494,0.7723523845250705,0.5447999270318684,
 0.3588903469075838,0.9381765066393134,0.0030456562617612315,
 0.6197207232100121,0.9616082526259031,0.35111810721446557,0.4499618808125304,
 0.40166635896341496,0.7530435815212352,0.3315029822556169,
 0.01572937018310505,0.10001586597011525,0.7691246526550898,
 0.017847198518501917,0.5622083039695746,0.4008417531805498,
 0.31446124941605313,0.12810700197481883,0.6607547146350121,
 0.8471989285159163,0.0028900871433592545,0.2041333356413244,
 0.9268858701183068,0.5447183415063638,0.9587292900343678,0.6623888419363472,
 0.22512859041302113,0.5529964892499761,0.6869438034828931,0.5228868194737802,
 0.37840633744513463,0.8956385004737409,0.05563667839177766,
 0.08272968792527235,7.578009412040343e-4,0.38578875417498903,
 0.03400986451648147,0.7257546600283629,0.05568465873405248,
 0.6741619156830029,0.29278464845481156,0.3514437310096282,
 0.21685854423412043,0.3572296745765935,0.29891730287154783,
 0.5467441730015736,0.9130269345479328,0.05373161289623418,
 0.18450151668265646,0.30009530835269604,0.1414554956762093,
 0.3967491009845058,0.08910877053041377,0.3944430943695013,0.8348317884542505,
 0.7248552350779531,0.97489769746937,0.5929936481755006,0.6717032939943911,
 0.5852289930674941,0.519797842076335,0.3507751248184018,0.23869218874149867,
 0.2803825785425669,0.07985411484314997,0.7111577384383851,0.7722465675066873,
 0.4126785096935578,0.33192605148947085,0.3877767181409142,
 0.42884901484688975,0.41184460109028553,0.8521014798947714,
 0.9602869686630994,0.32133979934175183,0.7519486780692428,
 0.48992013663076384,0.6184381964809603,0.2599615845346831,0.7689237830689547,
 0.07326879230448569,0.9573561372541238,0.25945217331840764,
 0.9082124563976872,0.8985026297406626,0.1597143152359517,0.11546646739777255,
 0.1910478274568188,0.20716271922631535,0.29751332256478713,
 0.7430933657561906,0.14080424356873744,0.09671164884359884,
 0.3106716075840684,0.9428713347681563,0.08332429395187124,0.558570833927234,
 0.05330834097401471,0.9730879093037129,0.05833415134309161,
 0.12107564731100062,0.574034675291875,0.35554979508440443,
 0.08906672996886855,0.15009419249569667,0.5272913212913701,
 0.6280100845402611,0.6327952604372272,0.7797989907192474,0.06332850075296803,
 0.4685786162906713,0.3889680212511002,0.26802426055502093,
 0.17310967607253347,0.08080712413305813,0.08180716791743903,
 0.6721689701061996,0.4217396756674152,0.20166137454676847,0.4896433764203154,
 0.6195468706143621,0.5453392108691423,0.28439659008030893,0.5469345627915893,
 0.7950808107949269,0.04376361534824902,0.8036701086820945,0.4793582864387941,
 0.5268338655978088,0.37465453630286105,0.8230875563009588,0.6374132495999025,
 0.9123634332178829,0.48605559203458926,0.8431841968175035,0.2525218997258455,
 0.8522985336978549,0.9319169356644907,0.8164007363109256,0.7839069565013944,
 0.20992875983805814,0.754023600974187,0.3662047317220134,0.02677052525872725,
 0.7661900996465408,0.9294585610449917,0.5695150937632849,0.8420439219078985,
 0.9152855706100171,0.5590123028896659,0.4835214681973866,0.39270198574957926,
 0.15773081878847073,0.5938031658083913,0.48569583120323756,
 0.6444523455033575,0.7719294054542185,0.5985113618780777,0.1982185225184936,
 0.7153731374472063,0.5262013423717851,0.12776665135872056,0.6019509027577842,
 0.427931085654514,0.22730579826562236,0.9342555554301878,0.6331014395540718,
 0.7907634480005018,0.9764783754350939,0.3880948923063421,0.3922715547400015,
 0.1565211210948093,0.4882466701546737,0.7017137685112662,0.34936825246103675,
 0.3954811615539021,0.059235378180437115,0.21822403370120014,
 0.11210916989818953,0.04328743299958404,0.5567320895787335,
 0.5049139419313645,0.1534699651310869,0.790236886819613,0.3555110505446416,
 0.08534501417614826,0.602899045291976,0.5036779016066854,0.11668019407179042,
 0.07316324698082577,0.34778126572190604,0.21967379902325979,
 0.6804911121264227,0.6903396191287934,0.28874301210799536,0.2206802377601904,
 0.8866276877518661,0.9221979270395515,0.5521238124634988,0.20583925319708973,
 0.12320345066869898,0.029359406990693837,0.7897553144602938,
 0.2669259394915363,0.2457750115194084,0.47875338576431625,
 0.08240669566428005,0.8973669927580857,0.7688055856589824,
 0.36971125682957395,0.1390476255797497,0.45050334581236773,
 0.2117250961651601,0.11783497387362551,0.2835258748932149,
 0.45018911993636423,0.39647379527249416,0.05646405344227723,
 0.4848176392673492,0.07869222418370148,0.9993043852127608,0.4506300901833109,
 0.41175445952737655,0.800706185453924,0.8636412703322408,0.6153814765353294,
 0.12046682691934818,0.28215850948123755,0.8148661093462325,
 0.20407831091656525,0.3566007758730445,0.7214279241098438,0.9059780531373076,
 0.5529239841048066,0.7236105624488979,0.10388352772117948,
 0.44050768804991336,0.5956730587792618,0.217524679147854,0.929029427607552,
 0.17449635795328078,0.198073052475762,0.13132234501743467,0.9325724755836304,
 0.024917403570586028,0.07698724087725028,0.4139554431839262,
 0.33072252381721445,0.5140537243822196,0.05761755433633042,
 0.3520971960830519,0.009568723615641095,0.05861615660310182,
 0.3494486300300528,0.20584575831445873,0.008404201485876506,
 0.5934688036628957,0.5275167675841924,0.566278346193043,0.10161591906105827,
 0.24478439389081996,0.33415623551143137,0.7666565866151142,0.923318492104037,
 0.17656108637145063,0.9264552364148648,0.0027204512702104777,
 0.038197588584350406,0.9781641812358242,0.04506124452459481,
 0.9098410367636027,0.2007834207656184,0.2398677097244003,0.3661509025528158,
 0.20307445808602864,0.3486968540747749,0.9639341974960267,0.1219054035825684,
 0.2954477989361789,0.24772483652908361,0.8677990092575693,0.9935896748175981,
 0.9097408037061718,0.7200612253243127,0.09371873077275972,
 0.42424946055154134,0.5442916406146013,0.5612892825897556,0.32722741416067,
 0.8046791379116026,0.6048782268877066,0.11275539144028834,
 0.020614721751188503,0.2974749940852389,0.24292545725045256,
 0.49522184875526565,0.026461027622791566,0.4987137810485337,
 0.7775997171583918,0.9315710696171535,0.3940304730363856,0.6035215087239242,
 0.2647475681529733,0.9026207253244132,0.8100059317212369,0.7439780764620733,
 0.7474601355395285,0.486371589874253,0.8440619186946658,0.7844183723403074,
 0.16913016037376338,0.21402411400120924,0.125119137709235,0.627119891922252,
 0.17261922987539602,0.6115908840991119,0.07419781157523642,
 0.9416921884279874,0.003498378210299413,0.2267205165584636,
 0.6082213911529905,0.668904723477294,0.8372944302764611,0.7737277473532372,
 0.4836688475000517,0.5627253932150597,0.58135332328475,0.6524870868524061,
 0.798126672121611,0.4033049929827347,0.38364601870800263,0.5378845856654322,
 0.8834747537829681,0.21102786761450942,0.14890085356362692,
 0.19819928730089276,0.5845521546155095,0.8332757291852747,
 0.06239612793511662,0.30431384867998923,0.7575686833148676,
 0.015606939990650881,0.6027094686500934,0.5556437985925324,
 0.7317458867813071,0.06131484634995554,0.635763372986951,0.12716372930609932,
 0.9864834906934976,0.2968989993480612,0.5327596267492655,0.7321907485009234,
 0.6707825406101715,0.3741639711362337,0.23501392098097762,0.8407111597278714,
 0.2594230424451903,0.6148662398332368,0.21922141457870448,0.3776088019798697,
 0.9417818193323163,0.5435137344370686,0.24078807117727674,0.5270141957259792,
 0.7433999917367259,0.2650520088625279,0.9293071731843612,0.6678154792640967,
 0.7993273352583996,0.877424987137467,0.9152416523751025,0.2257330854013565,
 0.6128307239003825,0.4703020327246831,0.9009500984675963,0.6065407555402916,
 0.9360185662678496,0.5749366530229346,0.7600597787837865,
 0.0049822389760486585,0.2290106504563908,0.8026905651958975,
 0.6546319910201087,0.5516156279716677,0.8331128721668924,0.8479218336426705,
 0.5377745382767937,0.3779949024011693,0.46212685850029245,0.8509598542180358,
 0.19635047341425904,0.3275633378949876,0.8786751849917435,
 0.27884406772505954,0.8750909787209803,0.5453783913841121,0.4010648977445378,
 0.4005351336956834,0.0844765056383584,0.29954524122358883,0.3947098608794153,
 0.037930158692860094,0.9014528256745706,0.9679604370164843,
 0.5069402765567553,0.717863462980999,0.11203096274672042,0.10832674734852787,
 0.24946959794804935,0.6357889456343159,0.2253871447828466,
 0.14890315235064044,0.9292149497601301,0.6710053117622129,
 0.16661322061728012,0.9086074952222126,0.8874641956281037,0.9877475436286787,
 0.08527569122043532,0.7888437299749422,0.5892377312233369,0.8976605956945364,
 0.7405440897944989,0.6707549279880572,0.5543557836863279,0.3328871043228405,
 0.6601045449665699,0.38696339620399334,0.22269844981750975,
 0.4137997283644621,0.15909099680799854,0.7681948934405808,0.2992400206795921,
 0.21060983308007086,0.4670011383787749,0.9207673073603514,
 0.20243690274398563,0.9718746666599142,0.7749900998021984,
 0.047708601730959144,0.2855367205820081,0.8917130657253275,
 0.9212388425276813,0.9274948702742707,0.4502198837066158,0.19931639520172628,
 0.8270996638125491,0.9697751094789749,0.7959938855671064,0.39168561262586543,
 0.08208528964865258,0.2384825873369335,0.21802986563988547,
 0.5850974567484236,0.15303838593409957,0.2127489122143893,
 0.024894442656455906,0.9727365576223618,0.6812122663300993,
 0.9412086556313916,0.023116497154253235,0.7407692754976479,
 0.5981650074481989,0.6672730949867534,0.10478529785236312,0.901780204692897,
 0.8868481535756757,0.9174450876247086,0.8367444781400217,0.17032195329524402,
 0.7837853456337049,0.9864622177704421,0.06166202715830993,0.8033016414704401,
 0.9494766307044702,0.8923855072180361,0.7117797719339312,0.3404732351869747,
 0.09985934082136005,0.7453282704359103,0.027195102257529635,
 0.08500657873384887,0.9756847234576802,0.5892778005204579,0.5203725628943557,
 0.2823642206408097,0.8281335696392726,0.46012649334375677,0.8266951824274595,
 0.7681772202679017,0.7283161046560179,0.47538824562843807,0.1656316212231964,
 0.40036494083152885,0.28170547591299333,0.6696056620038715,
 0.5898535266439758,0.028431931989273407,0.057461899794681415,
 0.001478085350831937,0.37619949917013884,0.37091643531987983,
 0.13919663885099576,0.2588862693956968,0.2942443320192998,0.5045705667059031,
 0.4627761404833164,0.5148989576405287,0.5095381177971088,0.28935250956975167,
 0.11437947460356168,0.4881149217156888,0.21129673798765436,
 0.0022097538639607173,0.6324866701460194,0.831311375518711,
 0.35923596261567947,0.4775658635294059,0.4098714808279832,0.9081457659493952,
 0.5265630772308985,0.8732386356378944,0.23668782070913064,
 0.15114997492775994,0.050900114229255444,0.7926273048737721,
 0.3817583547994168,0.29769347829527515,0.00674378540240439,
 0.5497885485926688,0.4166372073963349,0.0437825532552476,0.09530276496154766,
 0.5299366325370176,0.32514162460599494,0.5902542155299233,
 0.49916652227094593,0.42205798471668554,0.10614512249899577,
 0.8806916223005715,0.7782757254636422,0.05649542696014809,
 0.25391737298285333,0.27004209016252445,0.4722495987392078,
 0.8649383474905227,0.10711323765118685,0.5686314726156119,
 0.20219188092017848,0.46022510979975206,0.9766053299716652,
 0.3188771295131252,0.7831253079559343,0.5649900274462921,0.03629705496295765,
 0.33646283208874594,0.48981651620117317,0.8196811317980923,
 0.47266270677449507,0.07274475564706118,0.8431655929877717,
 0.06214598145586936,0.5298318531186429,0.9080119411255038,0.7350895375472064,
 0.31313983421608926,0.12780484595290065,0.9121182146950277,
 0.5008546476443804,0.002365494152610448,0.9940516413479202,
 0.9574537322009775,0.3115539376317652,0.24597022847696925,0.6576675785029669,
 0.24577521572923944,0.6109351436798153,0.21374595969822574,
 0.6886149443032269,0.6206886593937568,0.1049916396691647,0.8715654754841307,
 0.8734143321895672,0.49406104382540295,0.7455620606708981,0.4074573604889514,
 0.8086315701906054,0.14900210742481912,0.5013646815903887,0.9416778059068878,
 0.7663643740679977,0.925619325706784,0.7525489339658809,0.49466033260288667,
 0.9575389446172567,0.7253936569241908,0.45611303645166246,0.9663311010585021,
 0.10959404728023348,0.0927609485718004,0.004214112016475591,
 0.08208741934965647,0.7947770998345542,0.07864930150840355,
 0.38472128924228444,0.20120846389176028,0.20171734307592792,
 0.8806813652243516,0.7794241874016247,0.8771554023355612,0.22508115350459834,
 0.6166262914268591,0.19326801059184617,0.02642614103911678,
 0.43704596889740277,0.057462070372709895,0.6033708199434267,
 0.9208849575283751,0.020728359341627978,0.7218805511817501,
 0.9349112155288355,0.27814671991458995,0.49732155018496427,
 0.4425314422360407,0.8396527665516101,0.6720190804257669,0.9448491889851677,
 0.06429254040220833,0.4066070077341595,0.4640562031650619,0.195786477852825,
 0.07333527508372484,0.8980796063222374,0.7513415208610933,0.9243446019663358,
 0.03174529601565712,0.24111784199218556,0.17254260126016907,
 0.7347246719214312,0.28516129639111176,0.3018520129412533,
 0.19059174225057696,0.6540085700493125,0.09046773108288786,
 0.41930754399238546,0.6262821802155741,0.8436683334524339,0.8291111682079276,
 0.2688310817781754,0.551633055372182,0.3504079403883066,0.5454870992540624,
 0.45381669028718585,0.7605521144719777,0.5335689756448356,0.4752811406965949,
 0.10578714106758602,0.5082376680425951,0.9139040603401651,
 0.06419862637554097,0.20483163086030343,0.2032313376092345,
 0.0637565929964784,0.6328161044854024,0.7590361780640918,0.2731529057780133,
 0.8738514431736237,0.8186040118280501,0.4978882605200662,0.6491025973054149,
 0.8113204570508089,0.5292453944637043,0.6378350984531139,0.5781201989946733,
 0.9966345913485851,0.672973892122246,0.6845626073214732,0.20614249642717952,
 0.6514369214942783,0.001251018924204761,0.5224618898487121,
 0.5056212709935971,0.6867802137241749,0.22099606902955504,
 0.21120273853402705,0.3326510509174003,0.22065674371833743,
 0.029597474575091764,0.03499581118848294,0.17457997211902965,
 0.09503243265754902,0.40120948072951823,0.18833744011040354,
 0.7366931709160319,0.1684404696845888,0.5340922147608529,
 0.022430245405958726,0.6008786776200968,0.4452327707080309,
 0.8642588666831513,0.31381849109124793,0.04651709433341822,
 0.6356924796224942,0.1883070764672321,0.08477691473490645,0.844893428671072,
 0.12385701701094154,0.39998192945765376,0.17378781982124913,
 0.8100829381109533,0.6756358693589909,0.7151675322428227,0.7638797307896489,
 0.7450015892866919,0.6477473642857974,0.4307759677518569,0.8764374311694497,
 0.2631101292443855,0.8380354605198809,0.7700796407250102,0.10912063671732852,
 0.915666474213085,0.41517753031004667,0.46861964938348555,0.6261086678782296,
 0.2860642257017789,0.47738856070497604,0.5919177651709052,0.9458638435084483,
 0.3494562150220869,0.649808114697155,0.13403423976667161,0.7877208079508369,
 0.9428473220214949,0.2670449556458685,0.627982995431439,0.006329423597441597,
 0.35451433435910285,0.1800204415993525,0.9855601132852844,0.8045397026914543,
 0.9345859281668933,0.4306685010380288,0.4316985727755852,0.15768062737670707,
 0.8859243783623029,0.22516483172508128,0.24290856133572847,
 0.9367170364980145,0.8370787576989991,0.7349004446618533,0.1898020793361157,
 0.1263810710605281,0.2548030917800532,0.1363001083709814,0.11259130167400888,
 0.189625972970065,0.9455211017500277,0.4526326651729733,0.4081956322946746,
 0.8890282318548611,0.9123158330331385,0.3650735632287606,0.4814179885625003,
 0.8170260827103446,0.07846791381770513,0.7650529252338863,0.9975909312023993,
 0.8266492767520646,0.07316490208067816,0.4641505058942751,0.4990755695402831,
 0.13174950019932075,0.4728834632963945,0.9124499719303594,
 0.12830283582926805,0.1393632077779723,0.32880926631771157,
 0.12142481158365426,0.3833796687501785,0.2904796926040131,0.9794607478699754,
 0.1050140650075575,0.3061489412145991,0.168838501386532,0.5884835370827941,
 0.5679899003961958,0.5887960421483768,0.05545276724362935,
 0.46766632291220567,0.02801971183116181,0.8441131552237232,
 0.7223720579245873,0.839465223172464,0.39905817281095834,0.4782461287183608,
 0.02304025684652189,0.6176038473955007,0.8869791070275357,0.2363585468209206,
 0.2089569692487474,0.11098735776853608,0.37481762103044114,0.582781988367063,
 0.976190164793892,0.7630424729668157,0.09763138365284629,0.7014072907615052,
 0.1514818953869268,0.1790790643344652,0.91535095238454,0.9658979077176783,
 0.9769517235262379,0.5989630998582517,0.4976608054394417,0.35465705162121786,
 0.634784841172307,0.0626560453669941,0.1533593902117787,0.7196415395858833,
 0.36049579584534286,0.1536785015534733,0.11377552823381176,
 0.33717283396649955,0.06354347227166512,0.6442023363400828,
 0.4372861536024909,0.279649312518193,0.2603763487843965,0.08822710030384551,
 0.9649235964241534,0.3503034711898054,0.8376207643645146,0.7421895610808159,
 0.6320248027562009,0.16606674510541297,0.27387264976754266,
 0.7953877607159922,0.7401359006982463,0.5831129348444232,
 5.643338870641168e-4,0.7675410574990126,0.6921439011956738,
 0.049166822507893504,0.7680450828896209,0.16985283359426018,
 0.08769222810327026,0.6287873513950516,0.0802608874293882,0.0688988120613544,
 0.17335292699576987,0.8276649571356228,0.3188694360059443,
 0.01783827671313487,0.7613304931075191,0.7002174945212859,0.9238578433306673,
 0.14778824619094066,0.025649389673975564,0.6765658889152675,
 0.41151140264661334,0.3750063494691134,0.6682614182819833,0.6121567077137245,
 0.042222921651623224,0.5353984127492253,0.7791510712039533,
 0.7795551316597404,0.20956286825144088,0.05314947788939539,
 0.5202663412397562,0.7852323233914165,0.070683746796665,0.07666021903425024,
 0.850126681653932,0.7103410781100173,0.7441213812352407,0.8410018393028376,
 0.5327410297357209,0.635179324566291,0.2300805728745977,0.7902261086820341,
 0.7804699077264596,0.7337908966592502,0.6790034102276111,0.4577215851870695,
 0.6942083246015672,0.2598714222789249,0.27775289524949043,0.5086169770607212,
 0.9712807493535475,0.3880872929043673,0.08685958456956355,0.8722568206436098,
 0.36671259292088265,0.5075244290397846,0.9208633557662477,
 0.24185269265853748,0.8061282368288025,0.8163343893761618,0.3585074042806484,
 0.5363098039474916,0.6297558550174844,0.4378286925210889,0.3371564289292678,
 0.2616624643048895,0.8948434693549001,0.9912116657705494,0.8655350004951597,
 0.6353503231904258,0.2087624802444059,0.7078513351363804,
 0.038606941855065524,0.7204493300902791,0.7391930982053216,
 0.17659900625074332,0.872442616563467,0.10460729602890428,
 0.20845639379083947,0.437591531483449,0.15509108288624285,
 0.06047035174551285,0.4184803996995008,0.7454759929166617,0.1954998502915528,
 0.10981033442847365,0.04470294244545581,0.8479000547211951,
 0.6201092695788755,0.16122457572674453,0.8578241063619133,
 0.28075423196098215,0.27838252963745613,0.5488231945867507,
 0.20912950588361456,0.8856589922943325,0.8994870926706532,
 0.33249778399710206,0.12389627128266056,0.527305160898166,0.7080971276128529,
 0.42409130644020776,0.8785642196529195,0.3309463991935293,0.5921143190954861,
 0.531218259530923,0.4832554812897083,0.4039120189840357,0.4464213691264627,
 0.8509659008665335,0.885785927544297,0.4743842760882798,0.18977812986928833,
 0.25095786406658116,0.14078129205933776,0.838413918030029,0.7099482484770966,
 0.5408577962203016,0.9444026164752491,0.8348505767302137,0.06802666736240881,
 0.26545775618429446,0.43809769016457556,0.10627698907787009,0.65970648790251,
 0.28304890617573486,0.47751423116039005,0.8207227376132743,
 0.04782384768083914,0.745948193381352,0.6315447237883061,0.08213741918273954,
 0.6469017725493067,0.5430578389159629,0.23907678004111332,0.8897116759652173,
 0.366555810744842,0.713894523896473,0.6489099883459558,0.6381343702762986,
 0.234788797152098,0.906431984603281,0.5172990296167155,0.16820149125106887,
 0.6391103623773499,0.9071655750231816,0.9056782668360251,0.4137469320069205,
 0.25346573483965273,0.41550282627123136,0.540220031662285,0.5713604359625296,
 0.9640063186201738,0.9271663830537694,0.9362151238861322,0.952599095546123,
 0.9776328449317091,0.8431064756261543,0.2520277145027743,0.35832953368958953,
 0.2154114821952493,0.35905090974907494,0.5613278048242532,0.3139975949712619,
 0.10331158946353602,0.5710322891972016,0.8950973909750268,0.4710226392960257,
 0.13584326578231243,0.4909278131991569,0.7981220754603671,0.4506354880255896,
 0.4090865095216205,0.8393496996282578,0.861596406201783,0.9290227292316959,
 0.04590992927501569,0.9832110714020881,0.12017721175615148,
 0.2856339697008531,0.5356472841463102,0.6448070313352104,0.19214350280805015,
 0.48570561139735413,0.7661153848892706,0.5404891822300293,0.6429963016952793,
 0.028123982910880474,0.20241986952064428,0.5002578785948025,
 0.18392556484239286,0.42135963239161844,0.6951320390739302,
 0.36596314906845373,0.6415147318040957,0.2597158859508968,0.7884966342543942,
 0.6428091946870563,0.03933533589513527,0.8460806642253619,0.9328469665156675,
 0.42750393601770775,0.013533996097999101,0.8730935270476847,
 0.2654599305883426,0.46461580312827466,0.127826059364614,0.7470102786718493,
 0.5293518144617102,0.9200831447794809,0.4614956007684895,0.8989163031519306,
 0.08317670639920505,0.7691199953518281,0.4195269646288178,
 0.06188985378513234,0.1004722544386285,0.7649812330779184,
 0.08764368329683447,0.6541821316373224,0.0183157904784782,0.869401216520684,
 0.21234340833527054,0.7094016287808744,0.4722379675327487,0.7608321624666413,
 0.24429250748461295,0.8690114495289912,0.7668531423481468,
 0.18725382413497116,0.09847596038290496,0.11969316925315909,
 0.6313974114449917,0.19102823342003883,0.3836099339729716,0.3555220781698998,
 0.31188545096050624,0.24475154020003087,0.4154972811084753,
 0.25569406883511814,0.6154386392501179,0.7990078399684909,0.3765831649662128,
 0.3999359389297579,0.44081663237178703,0.2412693107689764,0.5640169725838023,
 0.6603502645462586,0.6851643091291904,0.48011528124502056,0.8658229171038461,
 0.06799830236993187,0.7828184021732516,0.531966664743469,0.5258210719182721,
 0.5476907433996939,0.5558711794947631,0.7843749639563313,0.8983800813067632,
 0.8265366411750315,0.03486108863142756,0.2721514489634749,
 0.07754740573149621,0.6790055216319928,0.2248559483816608,0.9811043265248992,
 0.46802725939628975,0.15933505145125515,0.6837243181213362,
 0.6791276639843924,0.673181863158748,0.6447309221156888,0.6654824047541861,
 0.7353052217428955,0.5336105371338777,0.2344057726760872,0.12750551612651462,
 0.514522387614202,0.12842963227307824,0.0714552690188377,0.1919869788310048,
 0.4996528929244757,0.7147259144203789,0.1248511081350866,0.12807317579126076,
 0.12646678468080697,0.17981660095485363,0.43691417491944273,
 0.009997100776150125,0.535997044851218,0.2601076829522132,
 0.08768051450417436,0.5962419082860894,0.10736510043546044,
 0.22128723812325257,0.3106092330194604,0.8628576022869189,0.9637072844594256,
 0.35344056686299885,0.1302655796889527,0.9061649939766749,0.8756147265249314,
 0.7707392799837954,0.3504384061149408,0.6061267645660142,0.9763916359588171,
 0.10090986705580185,0.15144036034558028,0.5494582272047738,
 0.9026553848017932,0.39863677828044675,0.19024378069981995,
 0.28185493014682417,0.13937684486121904,0.8378123797306254,
 0.24101969940716583,0.03067956307710884,0.9317321797048731,
 0.4643689170910449,0.1604749169444709,0.4225893775108588,0.5742890832793024,
 0.8469181901446902,0.5647552949909989,0.9129531030090359,0.09242330646983965,
 0.14172838898190065,0.7994540046529677,0.5369802388108486,0.5206561607352587,
 0.7533810044661376,0.03187984679019629,0.9013002287505232,0.6681435697754289,
 0.48208052946714774,0.5867049775503258,0.584521108025221,0.8782231521441423,
 0.35151874452845555,0.03392395266284509,0.21199426880995376,
 0.8349739978870342,0.33206272568755246,0.41586198162804744,
 0.3367948514879472,0.6699566379847587,0.13546768268522535,
 0.43906762091506935,0.3764726921932876,0.6251667767521135,0.623010317637781,
 0.2746461716377283,0.8989561601001663,0.32788733971769435,0.8369438620216675,
 0.5861085655859305,0.9987312693975544,0.7203848537881945,0.8750812435350546,
 0.7471402134169214,0.4903808769768846,0.09776641473992131,
 0.10377133072416611,0.04943323930194432,0.302975517125883,
 0.11480782437610326,0.25415149042126717,0.6749099319169563,0.681260431006552,
 0.056068700475460975,0.5815616979030167,0.15713536362301272,
 0.003071047428549467,0.028606720721960688,0.12345184134655851,
 0.5903556028557473,0.25817582680209294,0.8032125436433402,0.625674349554899,
 0.607228112362361,0.2515781776438747,0.9608843636337879,0.6695580042458803,
 0.5697439258106892,0.44463023034595384,0.7404896168840914,0.782502515639588,
 0.7353763683479438,0.875648201543981,0.798717638315399,0.6971811988522053,
 0.21899417872670868,0.21528391569383354,0.4727852633601042,
 0.6120818196129385,0.23492543613456363,0.660114803356891,0.09856849751937546,
 0.3466044718535144,0.2986618214327468,0.7523525144669306,0.9481287967961473,
 0.6861523502033459,0.8062466519293533,0.6638260265103528,0.332571377238586,
 0.3040136446899089,0.9935761190488073,0.10375722158125378,
 0.26352261142962996,0.564730337672342,0.7616280740274032,0.8768619801267343,
 0.18916752088207955,0.719952808036254,0.7929557654330477,0.9002648589787585,
 0.9738781341610975,0.6990086189510507,0.6284405162108258,0.23928996500955624,
 0.9146414258259139,0.4440750715555122,0.7432800693356452,0.4268078934536156,
 0.8261386057030558,0.7695361320807598,0.3552241552167782,0.5153522994363947,
 0.6330665419905179,0.10980668162322382,0.5067162521187236,0.2227690483938416,
 0.33272943387018206,0.3846784001011069,0.12983008592158995,0.440693121361437,
 0.3779582998292753,0.3837038073254265,0.9845563700315358,0.8549992361327954,
 0.13742716037342972,0.18766062645972303,0.39191877456260005,0.79484974901649,
 0.43230917782163014,0.07059589879657457,0.6759364864443367,
 0.9929203820748178,0.8624323933241099,0.4486807912773224,0.43948905510318825,
 0.21547092685887315,0.9956927750760288,0.7814931520880365,0.74923056738708,
 0.7566601652542808,0.8181032519940414,0.858519067692821,0.2723552993065491,
 0.26605139351463325,0.742645833562926,0.9543757438895365,0.33763130644131345,
 0.09963926226585418,0.9329048337639806,0.6804872683154587,
 0.044138862711569304,0.9901189251935518,0.1124774779743346,
 0.5034887165357279,0.7286751473162756,0.5780136208545449,0.10796804059014775,
 0.7211195377434716,0.3106701821357415,0.7518456003326608,0.8462397648640116,
 0.8009293158890056,0.9059203503529526,0.797922462054333,0.286880341496913,
 0.4227555807660127,0.3437532511633372,0.049178941137537624,
 0.2692788876901345,0.7356783023769045,0.7429556247909802,0.5209755784264647,
 0.5378982865540157,0.19008674666395764,0.25190296293814474,
 0.6766513827074006,0.983516052668513,0.017848954980476228,
 0.22446367653679555,0.6855451004977314,0.9766983874072281,
 0.35430160058878934,0.5257680491153199,0.9166563755874565,
 0.35680636112619957,0.6824527809450237,0.9428848922794186,0.9464672120547735,
 0.5215889175283539,0.7648608522841972,0.8458310291122563,0.2581645959542106,
 0.8774125952078204,0.9167279692536695,0.5782846882438715,0.6826583501493085,
 0.06684009282840175,0.15861313602836125,0.8535458257729464,
 0.12745984191699322,0.750512124113768,0.22815023253887623,0.9121383444770697,
 0.4510818327932318,0.7491421472452816,0.4545747916952504,0.1299903419176608,
 0.4641621449087747,0.5966672548759004,0.48282075831642635,
 0.21793443078069585,0.6260447588037444,0.8571359620731285,0.6652154958527036,
 0.7130677411124394,0.08993732404667232,0.3949020959121261,
 0.39975216799372193,0.7521564809893918,0.8370326728042419,0.8553993554210562,
 0.7775440940982202,0.6740673829811631,0.048853396070965505,
 0.9904628495318896,0.09649455395861584,0.22207773412930487,
 0.4374943215116097,0.2532959352270565,0.0032489948910747746,
 0.5153367686998489,0.19833828681979448,0.05752673382033202,
 0.8728354821871881,0.13793248626444665,0.9585142882086974,0.548882509624144,
 0.8759421775808988,0.38652761550375403,0.6985819574449944,
 0.0019304816946257297,0.771397396326682,0.1050404771799699,
 0.5259597076210332,0.5081107250493186,0.9845844788389253,0.3025953252658935,
 0.515175316989604,0.7259362503610404,0.6376562050059231,0.10832691422018326,
 0.1264198409609366,0.8167563764715036,0.45634096882015174,
 0.04397557732950064,0.7697349322800039,0.807063238106318,0.8760144942076475,
 0.4904320107082767,0.48532185530345284,0.5675951949603164,0.7055674603009128,
 0.9427239905057038,0.7433818722146117,0.8184994983937095,0.420189451699458,
 0.25924551380554683,0.06992762884124037,0.3643094696091065,
 0.6977712459901194,0.7588316131212661,0.04773078252676688,0.9605296413765352,
 0.8692194844842571,0.12758909135890384,0.1235642589424032,0.8592795335165868,
 0.5704864635878183,0.48015205775967695,0.9286461292265507,0.6094363423174999,
 0.5992047100070427,0.4998142336355743,0.7479050778301989,0.28698274948704505,
 0.0959331416460153,0.7195326902791253,0.438684996918313,0.2732842912924559,
 0.10709112163174495,0.33748020942222523,0.26464652620903195,
 0.5436468675685096,0.7063530213671998,0.22929915004824786,0.9591699055446641,
 0.15571477622305285,0.0978893166895266,0.792051147765277,0.6953118363663363,
 0.708386758422376,0.8003610455024315,0.8428649252949296,0.36133586630560777,
 0.7437781505156924,0.5923857810285339,0.48740557843798593,0.6376665808401496,
 0.8142520342565535,0.10482089929846083,0.2128734828790979,0.4563208582586553,
 0.4234392957889891,0.27632090303425005,0.11030414195468485,
 0.26917186003865046,0.5957555570308207,0.5595776557174149,0.3268188166442376,
 0.2103580599799597,0.3368233121408204,0.09568799730135302,0.6127598587139855,
 0.605258863218182,0.18972984989134734,0.28603874121091977,0.8165984067664349,
 0.02921834234437848,0.8722597428191139,0.2951708835894101,0.897800983590962,
 0.9179949878574434,0.8563629511597182,0.9511595013710201,0.04841441323533269,
 0.31045978004750774,0.7848003516352189,0.5763100025676672,0.6686420113604747,
 0.7946411983246537,0.6150596943604691,0.7986926973480304,0.6723731858102748,
 0.8093892234833474,0.3343542719804785,0.8279721105749696,0.9373523331170187,
 0.612441299861702,0.9157795781646436,0.3071652121647278,0.7961329832340377,
 0.14506897850149936,0.043859939764969824,0.734460990932711,
 0.9171697496443199,0.2682943456607433,0.12101545028025795,
 0.011562527100972941,0.03820990530847479,0.935329093991593,0.752962262167609,
 0.8268203166460003,0.9206915089579865,0.6570245031182776,
 0.056375843752352606,0.8753162526716952,0.8959799524465647,
 0.18510448454416284,0.7812458734162212,0.32357596927256693,
 0.0504183085059875,0.1580803171138393,0.16560102121152753,
 0.16749019947364163,0.8278148851681696,0.18636502799744814,
 0.8146510954082942,0.23914299186261234,0.37124410303527333,
 0.44087212280234434,0.44114183212385194,0.8730505687977806,
 0.18213008302595435,0.33674571677007725,0.5465647673393461,
 0.49890815053562765,0.5605100021825379,0.1439700153016502,0.7969926671742469,
 0.06598668624684811,0.8143253636093722,0.7919470472174173,0.6289861857712902,
 0.1749084095973774,0.9213146734424288,0.9618822240949764,0.9506728653801741,
 0.978697259974824,0.8886409711619387,0.5941791979799961,0.04329303603199508,
 0.5910527777732759,0.6551114889371689,0.17937177260978132,
 0.15709602378083276,0.6553238548744115,0.4930071791972782,0.9296991359458004,
 0.6321476386771754,0.4398184663784742,0.9894404667777273,0.42916868510989314,
 0.3854974462601046,0.0020934124406493115,0.7683663624889562,
 0.6793673844724604,0.2546194071075235,0.20037958536761968,
 0.44912712456006143,0.13703086861558922,0.14037818843271765,
 0.7887412056582331,0.5544453365586351,0.766539400831036,0.7773707466778894,
 0.5836718603483633,0.7196150255994658,0.9743819207118589,0.4183500189898496,
 0.7321756621812405,0.5278518824354452,0.5523892153713246,0.13730302426070973,
 0.24316933216394065,0.5442527654635224,0.8838052495317534,0.6121457193827398,
 0.3254802809137237,0.5150988744006988,0.18110968062030774,0.6861871003595733,
 0.5086803104047863,0.43922566968159227,0.29592803135264223,
 0.8107931726244011,0.6876805546409094,0.26713274962506706,
 0.21778276942567132,0.4959614846681796,0.44520794996214486,
 0.36272380561030926,0.8313488891458787,0.97353939866993,0.4573212871848916,
 0.28368186246428606,0.2229199107121017,0.15915324507976603,
 0.3970874227952187,0.780388341804569,0.27550200004697545,0.6644735693406214,
 0.28352745802388235,0.6375716537908023,0.33629548223622574,
 0.42439480990418166,0.4304399150217538,0.936674039035861,0.1946043849660345,
 0.5622349380904493,0.27354320989107506,0.9364793555984666,0.4464590418843982,
 0.11008160255437294,0.11496599795858198,0.9424650405643193,
 0.9618028581127389,0.8967175180157257,0.8709606606202426,0.740969704625167,
 0.36831214288194947,0.07413434937927255,0.2902312869242265,0.945787931582623,
 0.5705840014790158,0.9222217348193564,0.0782292183733505,0.9585523112948681,
 0.012636094870114079,0.9309484572679635,0.08372267926190236,
 0.6459261465528497,0.5331782129574738,0.8726581132386082,0.10171070166389196,
 0.7112098053619249,0.0840570626100593,0.9816191632793188,0.5403419596765155,
 0.9812639110647706,0.7866215984651334,0.4140933599033966,0.5579351110193626,
 0.6383509720697322,0.49891184524164844,0.8552183657360478,0.8944565609972628,
 0.19803376881327506,0.4852112095275043,0.18258284048389262,
 0.2060207379918958,0.04209760374311067,0.1531885680777605,0.554068973949979,
 0.24964046562154485,0.08497004299297539,0.14237238222954196,
 0.3672239010709353,0.7146305807173712,0.5690418459379109,0.5960024468000176,
 0.6617354775182079,0.24375049218175893,0.06470569524216097,
 0.34674110956625714,0.2614058889272888,0.5951714214492966,0.7708103844857919,
 0.0019528178085178594,0.06640726140395903,0.2859351783985846,
 0.3984207291332882,0.9942349320854831,0.27207948911226576,0.688699324808479,
 0.3687785438128911,0.38950652804960484,0.5242312265179854,0.7992302681950504,
 0.21987240021948917,0.5866493178788258,0.1493683716854901,0.9467890956717031,
 0.820309900488948,0.09280490792828533,0.688248369207106,0.15210926792768953,
 0.48761304558325147,0.20122166234716166,0.4574239457443168,
 0.2129584658462952,0.5703908284509506,0.5039082946715419,0.8748909678815522,
 0.4520936573740789,0.7693994221718005,0.49908090500107793,0.7675931895256828,
 0.10386808037964024,0.9399918846232884,0.9194939118623474,
 0.14977294143236053,0.2465259731150946,0.28178394083009006,
 0.1564110975464309,0.47317153087361774,0.6138554456959571,0.5912565662089853,
 0.04659534247849306,0.3048089455307088,0.7154876184400891,0.9342088404345116,
 0.07180391404863751,0.541023472314767,0.7739174814966474,0.6488862878965115,
 0.8017570531190967,0.6269833637137683,0.730821787825839,0.8415566883748504,
 0.16140480283299685,0.8122477546013083,0.32951244262946067,
 0.7245885830464132,0.6392787426714586,0.6752185624644069,0.9648829252595491,
 0.9178111057231055,0.6615808421177055,0.2607315482671694,0.072652290366088,
 0.37992214143231107,0.6094092908925163,0.2176102749794584,
 0.030791541067465067,0.5500316954565206,0.9618560841663593,
 0.20338578107275085,0.5081789391339103,0.9668569106719833,0.8315426763972446,
 0.035335674710366916,0.49208112918049274,0.16844713233239694,
 0.1551113212120515,0.6145791068162102,0.28537389285855885,0.7589716388255969,
 0.6978177941536555,0.13529665813904035,0.3703742630034974,0.9429953997361807,
 0.5469020724382304,0.3237849864069595,0.016148362438513586,
 0.9939470228189701,0.800708523440969,0.19103245211086872,0.10579514989903471,
 0.9109079834219,0.08040271600440518,0.8434787069989861,0.7191056139028458,
 0.3333349544729577,0.5308516633854584,0.5305577946498184,0.7135657208792029,
 0.6924331183746861,0.591437884241082,0.5799494844482691,0.32612479933898675,
 0.5272466562155582,0.8047866429926578,0.19521058730728702,0.5377096265762317,
 0.8729727108045477,0.5693574095340962,0.3689056082988428,
 0.020186413796888347,0.593123013240235,0.5079001705612312,0.5544471110167503,
 0.68336564495177,0.47136086295706014,0.48035858785006835,0.466331408169115,
 0.5376087503882044,0.13235990847878654,0.3623408804873751,
 0.29423976377209293,0.2344584881920132,0.2172099298485084,0.6913135103322957,
 0.1571296041049428,0.7261419961702361,0.9321584383765203,0.2877485518464695,
 0.3111348043542619,0.3625205055834906,0.617634878333825,0.576199204118563,
 0.3623388341635342,0.24057607162752748,0.43357837974049707,
 0.8731376270921791,0.6975799338210937,0.4037136894950416,0.77372059757567,
 0.6473098817040623,0.676249916983649,0.8894811502304075,0.21053669240666328,
 0.08245946818550853,0.8569870930575291,0.1673500675920807,0.2471656735344291,
 0.7991462921376733,0.8921507080010619,0.420518621926828,0.8607876639709219,
 0.030004743541316548,0.6610723154292952,0.15634504447441422,
 0.5761874875505318,0.6556324230575238,0.6553934587080306,0.8511711942833067,
 0.731970465428838,0.9614013411370155,0.20692423210635602,
 0.052880035012062065,0.0887478659032046,0.6368549487418818,
 0.9466199489869921,0.31938586485372245,0.4709674895270566,0.8126367953653769,
 0.5420280606209527,0.8788982715103854,0.4269982995472741,0.04695428948749647,
 0.6183404152218138,0.9732939246046429,0.08771180589912175,0.7032222576465983,
 0.36059221093690097,0.7647917426131272,0.032150312912002654,
 0.2925814555969415,0.35580330573154406,0.3571002942724131,0.1683429874812148,
 0.38372831324039713,0.12482202143369392,0.1685791962164087,
 0.08438093336719577,0.027955861101910173,0.8013931630168376,
 0.33867114416465993,0.497802729390716,0.05241701483081285,
 0.27961904058907394,0.6356369413108159,0.9173504935008944,0.5507367838156875,
 0.6960947484598946,0.6684686758015241,0.7243698941925943,0.5976328036948383,
 0.49807516986516687,0.1474550295103163,0.3208203877675555,
 0.27850736666467046,0.634151472960669,0.5919155581712645,0.12733104717080335,
 0.4521060639784086,0.23066964867488904,0.44307474553545123,
 0.6031391827358881,0.13368759516870543,0.004222826769689214,
 0.7234971591549387,0.6269332057983847,0.09331683052997264,0.4949437958977627,
 0.14985942560799614,0.3176980693273528,0.08279470252219945,
 0.5297792535380489,0.08545181293808035,0.2947415509103881,0.7761492178462063,
 0.31261502522553886,0.94847637161488,0.41332171550854224,0.9912433811930297,
 0.16099151743529472,0.23193428598260957,0.3350373580935474,
 0.7806503271334256,0.47679008620121943,0.5466676125193234,
 0.00417194615285732,0.6768238587449104,0.7625365991067165,
 0.34198084061253176,0.09444334749579719,0.005366015372132549,
 0.02309271026997639,0.035875397435056744,0.9998488190238897,
 0.7580985157177341,0.4381825985070269,0.32737839285225256,0.6113782247786713,
 0.8539108989130636,0.31747368461146286,0.4436835395997464,0.7052294654224793,
 0.5811362575044271,0.693651200135845,0.21002350783798585,0.24842980033903594,
 0.08608550465151799,0.42138012710231665,0.6459466473598293,
 0.7007038933502852,0.8809759807232419,0.314804235806884,0.1566435372475059,
 0.5629017586024394,0.029174198135053953,0.3596081524525402,0.934587167110853,
 0.7986592245781928,0.7015757192480844,0.575895689325348,0.8339585173120864,
 0.7159229617578747,0.012629420908437217,0.8952960959951836,
 0.06471970090790635,0.2923096605128104,0.40759761621557156,
 0.7041875401216791,0.7841744422194415,0.9775810182854229,0.2336282507661196,
 0.9374338444442661,0.9351845423789191,0.5521900888600834,0.7164991097620317,
 0.9142176515013515,0.7318358607827031,0.5656510115373216,0.10931935073850285,
 0.8693005566600605,0.27599089068768645,0.49324645904341713,0.506964199739755,
 0.9688857055696299,0.5206855281504564,0.5646188824903737,0.9060458033705491,
 0.9473235958788679,0.17388143801072453,0.4194412710020451,0.5199152679120497,
 0.476263395083815,0.15506996891482383,0.640799224053946,0.4156666710632667,
 0.5288185329240918,0.320992798575943,0.8972199675005531,0.6070979087308671,
 0.6519259793923879,0.7937697200945264,0.4564516665243552,0.7649402567030544,
 0.8429963962862531,0.5356970391523947,0.8799764056371948,0.1284533999329105,
 0.4032347792399883,0.8137873459972866,0.3595288245855015,
 0.013495872318836089,0.6611370004985152,0.39500236161813784,
 0.1895386826952309,0.5956345408595729,0.7572829295134302,0.22501383847019552,
 0.4572305565044418,0.2049819269940154,0.7575077845938883,0.6515292783103536,
 0.8680573899237605,0.006064840026513307,0.20326812515654213,
 0.8693542347737323,0.640562889927584,0.4429925889828473,0.32819838166646664,
 0.5469508981175849,0.4980481771669105,0.5221905958633948,0.4098943620513732,
 0.537770918178543,0.6220212787251038,0.6779240916149303,0.09812579196652504,
 0.6048957916998967,0.6398393997086842,0.4301593193878708,0.6231359133595846,
 0.47992078419903594,0.42515258446232707,0.6274009640727279,
 0.2322162207123084,0.22600267437190613,0.842701439780672,0.15474223531519704,
 0.09668029968318936,0.7587648453927754,0.7161115379916295,0.9358057049482561,
 0.6603516280947663,0.08323017820738288,0.7182535731921832,
 0.24274401282751723,0.9014231133848971,0.5589908510431957,0.1609525998783652,
 0.34720710527908716,0.7280713097913449,0.09109494348979641,0.852931189209603,
 0.24204633133762976,0.1781448452802188,0.9042557367776369,0.2154427091097385,
 0.8757665228870025,0.08508189626506235,0.9745887596510528,
 0.39069245512037165,0.9338676392658669,0.2965768130028775,
 0.04926055919037109,0.39452263807406385,0.2505284966349518,
 0.5230362236293333,0.25969299768292897,0.289853816835685,0.5496750750824496,
 0.28028097577903643,0.29964059405603816,0.61899859398153,0.2078814786859472,
 0.4313779371737515,0.9988331619830716,0.22899876260474294,
 0.32453592882990645,0.06896136236764872,0.6778587624594592,
 0.6529764670360733,0.4842232290162073,0.23070263496287313,0.2597227325533934,
 0.9091044026376839,0.5059579021303684,0.4516429444947364,0.2046427869092302,
 0.6853064709180796,0.08581089692130828,0.30299019602174804,
 0.8788251432888716,0.03989447207746655,0.684475118765103,0.9325046603469868,
 0.8822260863290144,0.4304929947939502,0.5389191274613728,0.08150473581629814,
 0.722753489655656,0.8616037326928226,0.4483746391638894,0.0969564055218467,
 0.15590840390776095,0.3517650554709437,0.31440498673528194,
 0.33104329703123303,0.4986059415120767,0.16482210800016672,
 0.9964074789812154,0.24386334180649094,0.6992583146945125,0.553672169377087,
 0.6801394763235487,0.5117893151379577,0.054952450795652585,
 0.9549342400786559,0.543267418929943,0.1256473396714115,0.2797440455564906,
 0.3928360990232167,0.9993726966055038,0.6945831473922385,0.07957083283539934,
 0.5487867959326443,0.4822483446122675,0.32442159923485736,
 0.42667786920855677,0.6113002717470026,0.3654651329581655,0.871349774757819,
 0.7459647250803334,0.03509520187435544,0.16069900717085006,
 0.6135463254572372,0.8302219793289722,0.6650312324942644,0.9881019030863228,
 0.574048489149754,0.06692791010392085,0.5088671813721817,0.15474676952294253,
 0.8018360276972899,0.7219845409117963,0.615878554468902,0.21686842575349752,
 0.7340939024099683,0.9739780055343661,0.12786887099791577,0.6361340507106293,
 0.7481986750684161,0.9438239796745238,0.6966921185780733,0.3074751287198255,
 0.918832185429737,0.46265975883600674,0.8196574739809004,0.7354474036014615,
 0.9337189572777311,0.2899502237717191,0.38841179230433853,0.8731629039853303,
 0.7949992277297554,0.8072319971682117,0.27804445072821926,0.5787298650161188,
 0.38202975602107325,0.6706947298658132,0.3816040035425151,
 0.011928086552047024,0.4201244330246263,0.07422031917690997,
 0.6590284633129893,0.6630227425871072,0.9585445277380271,
 0.013690534844788327,0.13035018476903903,0.3383982720144252,
 0.3333619012705811,0.30218141621211614,0.3533833647702106,0.7472614829683277,
 0.2119414642454992,0.4890305916725879,0.5756057467835006,0.18549218914567112,
 0.5178887696530102,0.2320787151594076,0.30939731271756454,0.5035951396882804,
 0.039189087787467325,0.12011374339092873,0.3214114865167039,
 0.5128980024146714,0.013130464854959056,0.8796183871231914,
 0.02863224832572353,0.14072780195068013,0.4560602891888228,
 0.8809265217311056,0.04162285923996423,0.8177588274569517,0.6018547429678562,
 0.0721377814854417,0.6019531771444141,0.5318913550487477,0.8472735137867264,
 0.6744231974961425,0.46167752223827363,0.04224640290612036,
 0.7724649944988516,0.21894474268301167,0.2292383760059029,
 0.32626198951931906,0.7172316957183105,0.6765550621124874,0.4204352809102415,
 0.5656106749985825,0.21279389173996321,0.8758242531372344,
 0.08400344956557282,0.8807565962071288,0.9808276366335095,
 0.14419187997229732,0.28507258623281184,0.484286390721439,0.9090960108196955,
 0.7404983778194928,0.8717806502521906,0.3912316124121822,0.5742799432768496,
 0.9122909331003467,0.6112500975564017,0.5859474021130646,0.6549016981465753,
 0.7025519572832711,0.8739379205069163,0.5563610230730436,0.8407458195133812,
 0.6265311176966862,0.5239499126365283,0.954900553162148,0.6771062727728865,
 0.05362233169628017,0.06941738384900875,0.7395168584209644,
 0.27525639473815267,0.7090348480610168,0.7810143541274056,
 0.04437531100530312,0.9467872294822959,0.7323856038733648,
 0.43442421562815015,0.8028759333664837,0.34634983696128263,
 0.07719890423110898,0.960825223691289,0.7226700524340456,0.8271705500242741,
 0.23346658454203262,0.413324353382137,0.5085783276479788,0.23996742032195462,
 0.8396829779095633,0.18056027595792157,0.4371365576109487,0.8344029823988923,
 0.431564463081877,0.2075782031182375,0.5503879173643621,0.4034125252607129,
 0.5101701994276234,0.37668066547452417,0.06836805099741405,
 0.25145969619877284,0.6063217859974126,0.8355531666449865,
 0.14624869743840785,0.3921120971759652,0.4240481371713366,
 0.04727698245744083,0.3478199347392128,0.7157598967179699,0.5129225143105385,
 0.7952816587668682,0.049033079668237,0.8339686740305601,0.02445803145020431,
 0.8322392786353628,0.9661677874214868,0.34104027316728514,
 0.49544307604142546,0.006265960867600118,0.05352088814438327,
 0.6623340150017607,0.9492075899812406,0.3934330578417282,0.5488202336071805,
 0.6213980800865627,0.8104389994566127,0.2990443393276687,0.6688470160500359,
 0.017347038911510126,0.6511174574952551,0.867350551922164,
 0.06010752499802252,0.6858615765267754,0.9989369160051669,0.5304204134764018,
 0.021754048405288717,0.5199197915384591,0.8045440445169995,
 0.7861433669268558,0.1571908030699659,0.2722051237792461,0.6812531185908284,
 0.9135455806529678,0.38738968900591275,0.49420850036650643,
 0.4986988026937045,0.542334260464287,0.05136936607381626,0.9960148781363876,
 0.6696168327591795,0.46010844900839776,0.17737538541005338,
 0.12656302106525819,0.40997704118506906,0.4708604502094391,
 0.6235675999665622,0.44455086029968527,0.25248477080457477,
 0.024120708608995733,0.42976519420508164,0.12391045741962625,
 0.4385696606851668,0.6716835923590863,0.9952105308034384,0.30237112564606083,
 0.42105483714786973,0.5450589854309597,0.9758770211618221,0.6345347576348589,
 0.638142858066561,0.09765285033310334,0.46652019423892366,0.1492097888252073,
 0.8032712139015934,0.8118268617200365,0.7873490291288283,
 0.021833415638727094,0.35887923194588334,0.24924282992505364,
 0.65604310097452,0.6402889733352826,0.07830900436571031,0.41169727360935016,
 0.3299843450567884,0.7739477424237182,0.1654942467811631,0.6360528044926989,
 0.8181120994455058,0.3524753968485904,0.2944427776573427,0.9294817578143386,
 0.6123938494246073,0.025936698689107307,0.1870066097200912,
 0.016079101523953976,0.40164833907954667,0.6271112211447252,
 0.3234708898081782,0.6736723665380302,0.5558322662598583,0.4838958486368228,
 0.825907180969795,0.09933230401953308,0.9907498639864999,0.05083486444916541,
 0.4407221246115648,0.8029869817897469,0.24548853263298898,0.9348417051325455,
 0.2424945066096409,0.20933733017291312,0.08537684725826333,
 0.8109875801405023,0.7697412597112312,0.39955770883746033,0.7335304189989716,
 0.47651710083616783,0.22606007967722253,0.7872383454921755,
 0.7684203450633382,0.521521824911644,0.7079260463101775,0.7997971605951537,
 0.25531065583415,0.7597458901431566,0.44569586901791525,0.5049526284690447,
 0.30556285885204193,0.2517839472473107,0.9901306460661171,
 0.37423089194174985,0.05041318354198587,0.9059016739297542,0.810378245794012,
 0.6349855869551277,0.34433143664321975,0.22818214947248694,
 0.11909694364347401,0.3028401299717668,0.03537655852243149,
 0.05162132113800755,0.7820264732518215,0.024477750733830406,0.86156762547342,
 0.04588898623153426,0.47535597957163855,0.8896500629200493,
 0.2972607707730778,0.8530778289618732,0.7319015526240602,0.43604963025242194,
 0.7317435142893878,0.2072579592494106,0.7414160176066513,0.5285934998684949,
 0.033592266692771044,0.025021809019551644,0.059709232812340196,
 0.37215045643829836,0.36188049267296896,0.832760707886331,0.9204814554568431,
 0.19924310484077146,0.2481775986512489,0.2993437956945624,0.3453255023306172,
 0.3473250865109364,0.242510067220991,0.9418814270705564,0.2485116719200765,
 0.125567008833113,0.03517878106854244,0.9892909481973748,0.4201203331501282,
 0.5734157137473306,0.3440590750094842,0.13206538778321542,0.271319509792298,
 0.4186324604799332,0.7357243763046079,0.6364638717404845,0.4165150925038015,
 0.47090130611538406,0.706496376055847,0.28588133095516044,0.3941626138256601,
 0.8036477081992764,0.43740981843476856,0.44688445164937696,
 0.12354033078059035,0.35827065183210116,0.7046242203241425,
 0.6536486093677873,0.8226575047612165,0.30510247536316004,0.1309913967879407,
 0.2547430767887606,0.17263909886321893,0.28212859733147744,
 0.7534469228739398,0.912665257925248,0.973173814839161,0.5352876301506243,
 0.4796105373767784,0.7707088310750891,0.4567549270551583,0.41143343937823185,
 0.3047285141138757,0.8665480593133109,0.11681803381614653,0.6224646362832403,
 0.5240248243399307,0.8516432223651915,0.07180089319158989,0.7664666441740483,
 0.14256325890334065,0.20528827868737165,0.5735496818480064,
 0.06531894637348867,0.056605707776933745,0.18068799089424092,
 0.3529195487452709,0.1342470544597647,0.44506812576970467,
 0.17821283047605296,0.58941573681236,0.8057400870626102,0.2821832700612078,
 0.6745141861339382,0.9844282617703481,0.5881171859388787,0.42474193479919753,
 0.49005195895911713,0.2883080387399891,0.4806962400263495,0.5320339322206807,
 0.5713197896246617,0.9790017510066337,0.010408877084070012,0.504156689409421,
 0.3374227586661034,0.20764256977218187,0.7163716716320768,
 0.29377738361876404,0.8560252992787568,0.992539609233052,0.37452874311516826,
 0.9606953799706842,0.284416118647985,0.3159535898017035,0.9108169371355044,
 0.303338370068752,0.7239268394864143,0.23584459914853162,0.15932176999446113,
 0.5591770254351062,0.6090712499239729,0.34322279209828854,0.8914008918323936,
 0.620229147322642,0.9626161445704797,0.3181432383262218,0.7288363358509528,
 0.5143405167382048,0.42371139962186943,0.3662553637080539,0.9277432762908191,
 0.5923311187649074,0.059487117083452334,0.708208849524454,0.2859388413116678,
 0.621324835768173,0.36949134363994784,0.4614349765069986,0.12267121348112309,
 0.5590430847838294,0.5123760368536213,0.12043582680478504,
 0.10564632683986519,0.03318099064228608,0.8244318820770584,
 0.6430115846595391,0.5342011716534467,0.15432608177662221,
 0.13796428948861972,0.6746007249254794,0.7076856916473033,0.5076717291215612,
 0.2047297829531406,0.292904734479861,0.1482970841945942,0.6152636929017682,
 0.4245088690184686,0.20707526041786317,0.7786100105932424,
 0.49915485711822205,0.03437962880213574,0.645748054089586,0.7111793382614029,
 0.46286323852818234,0.19631644054374098,0.18735970618036446,
 0.2831813431933651,0.2960651064392964,0.06660582257639236,0.5289906858492183,
 0.541284590882001,0.36943962981432443,0.5547092039312445,0.10564115066039192,
 0.6806815205390643,0.9505000499479171,0.9368581031766736,0.7024141435552469,
 0.42436292256154273,0.8286159143107378,0.7929690146187223,
 0.16601123541868323,0.4343490404952369,0.3380488648715122,
 0.40459991148626806,0.9671276621295373,0.7410405005414312,0.6097774070587879,
 0.6533147109487863,0.08649343436597445,0.8343170377951682,
 0.47597126057393546,0.8080631532999862,0.1983479463850042,0.6734235115437662,
 0.7442883279081467,0.09474635688440503,0.9863216365007108,
 0.11092676061912676,0.6354169928883769,0.18200508983557195,
 0.23811344784668864,0.034336943386568786,0.9755728792863501,
 0.7824639771314714,0.48133495338939625,0.448355841562017,0.9635111344160128,
 0.8293908313176446,0.31370025227246723,0.9510637040342498,
 0.07051208410735799,0.6753476261075942,0.04378684879451211,
 0.1964909119421726,0.7546244531546145,0.052042004627363525,0.670822511739779,
 0.4352698592682481,0.16141381954376222,0.39936287360951805,
 0.7683043449886645,0.22050019887931094,0.06328111483768761,
 0.35662527659606136,0.17381023436038212,0.14337968091801956,
 0.39741751189303987,0.16734590642966518,0.19973935463711368,
 0.12097131388339122,0.6154050162416317,0.44889492215173254,
 0.7009923846903272,0.5976357150046147,0.9988230488497222,0.2809781329102107,
 0.27800873451141417,0.7410287918951356,0.1027901258725491,0.7218879652446961,
 0.07215603546447169,0.7848897909084778,0.35655456089865023,
 0.4563909437034471,0.3512776801576428,0.30586776662063064,0.720200041043731,
 0.6756389160994533,0.5573425640993355,0.4606711737657171,0.24412775815867394,
 0.9061289717451392,0.4100068412599551,0.7013832301843055,0.4093910617877746,
 0.7542919530051763,0.5501441655680076,0.4979182363800624,0.32561961515546,
 0.970044331231344,0.7503692576115071,0.01400863520922635,0.9675671899003927,
 0.279849179112298,0.46785726537595007,0.830508630133579,0.1352629448586441,
 0.3732558565749531,0.5932530422320491,0.8580234700986067,0.16810059081675865,
 0.49859171936963054,0.6618928461164715,0.8227973919130642,0.9977457229611185,
 0.7566620434288035,0.4558988352986978,0.013852782344029979,
 0.08446688708814043,0.3663266293790237,0.2018509327097998,0.9878087159473039,
 0.061102603210489104,0.6613421632574141,0.4868495019997928,
 0.04370368283987447,0.9775055650108692,0.3075864136716022,0.3854810926991665,
 0.6208214190118377,0.15285281172850107,0.09958241722412087,
 0.8230595046378821,0.3829164682669143,0.5234084258168061,0.713563780644886,
 0.11792870406656308,0.28626923924277414,0.4337277050436392,
 0.49266933805833024,0.5264978758449772,0.4301525774085192,0.520075450859577,
 0.5940237275327379,0.5048189318233701,0.5246857497166667,0.5307103577898367,
 0.2900619217362831,0.43457293953784426,0.12809638557426928,
 0.7166845404812481,0.8486562133485458,0.6583712865733862,
 0.026623080437749636,0.4302404996857172,0.255847503377182,0.9757413723072512,
 0.2918392984757756,0.510840873708707,0.5264149949706596,0.79950318290435,
 0.4183142493549312,0.24873999105354017,0.022252781538898092,
 0.9710445856247057,0.46666418317907676,0.1408258111753553,0.5030564506785319,
 0.11068900512455704,0.9514373951408134,0.2789158249043777,0.9076286348220026,
 0.907111460993355,0.3007272070360486,0.5444387936306502,0.12834761180957255,
 0.7713476738105225,0.7314057668344973,0.4820275404998208,0.22380254737416383,
 0.6804445083710863,0.898454408380797,0.9478897398460582,0.35592783212166657,
 0.4991206763028384,0.7969013786431933,0.7519071612522096,0.41888050904023644,
 0.45758005631306786,0.06279881681966193,0.40119653166591895,
 0.11085415826916156,0.48966891470525387,0.004421108446610922,
 0.7636742430078487,0.8061910084844139,0.5252366549348731,0.7137403573794736,
 0.3645979408087916,0.22687463851530065,0.8375689947073213,
 0.06692448952851748,0.772254219531093,0.3732106283563267,0.6495650140387574,
 0.33549338675515117,0.5730692936170155,0.008084147841904477,0.15812538969541,
 0.6986917061759335,6.850815174639369e-4,0.7514926818814766,
 0.13063115755289623,0.12688232043214598,0.1363446733581034,
 0.3853623017919877,0.9511459240481258,0.29126735570435835,0.9264688574720321,
 0.4061721162875367,0.33632276633509717,0.2080669082046528,
 0.19577353606343173,0.7274581202185363,0.6624732774961426,0.9104930764345285,
 0.458641846402974,0.21833398497564305,0.5802623908305238,0.46267381982024725,
 0.6496338926187374,0.6732267270015154,0.7078721262789214,0.8741705179705017,
 0.4093173653411152,0.12722547795556738,0.5207409997777338,
 0.009369525984275207,0.05301413480537953,0.051364757175018205,
 0.6177664485217872,0.29034721514117123,0.4613535458569634,
 0.057820159564521134,0.6463550274428296,0.7133861220620693,
 0.6326206854378663,0.24285195735136011,0.2200967067609153,
 0.10459712282109701,0.9651795360851292,0.8878686187955833,
 0.13298364389492878,0.526976178466257,0.8625191950188158,0.26148397438167303,
 0.1033665580360622,0.07302296791121221,0.18666550436774065,
 0.9682998894309827,0.8289862584050163,0.281326813028977,0.2525889327860322,
 0.4307778683013632,0.2802988101651308,0.7328669616780429,0.7983214542946735,
 0.9016927892638359,0.7011218555588014,0.6826285282563077,0.3077759648655073,
 0.4005928485459499,0.47834284019347584,0.5317329930821414,0.3931284605547398,
 0.41041700468776454,0.4564536635679881,0.8770867377023567,0.8239713017770777,
 0.04880095102639437,0.5352488194132812,0.44278054068523387,
 0.5318640352556987,0.5876248204956852,0.19822190718331822,0.8675317017204937,
 0.689035756440235,0.5923421812751138,0.7773529064856212,0.9287872791846152,
 0.4112426986108333,0.32647268490757275,0.7574982444638049,0.9378787026940265,
 0.892787782207613,0.5313471623240515,0.49950186883024317,0.08496536443008385,
 0.29312518432128565,0.4155964397728924,0.671512788141273,0.19276394512501105,
 0.7654077141923665,0.5144905253771178,0.3274799066845505,0.40353681651588946,
 0.1644796745949042,0.9747359232753379,0.9580978287601126,0.3705520043668782,
 0.23309223945850044,0.19903436597456747,0.2638906467225248,
 0.3090666601268822,0.00847885895368028,0.5189840290144674,0.2785190608322775,
 0.4521512625155478,0.3023739524844671,0.6287807273687468,0.5469254485972623,
 0.9149545844552333,0.2341860371686355,0.23973127335762157,0.3468512587557122,
 0.9111722684309225,0.10943709972819393,0.7959443849528832,
 0.49651773519878806,0.4768322052392189,0.33232542995460257,
 0.17855898921897728,0.8980688745843928,0.06316460681619263,
 0.08871459831188999,0.05757793956342283,0.4731495371624064,
 0.36339928315483017,0.8193677330954803,0.8286373661511576,0.7858249697640417,
 0.8547682093869063,0.2250513321836244,0.34997981868331995,
 0.12045064524864268,0.39299059006498016,0.9931020592044306,
 0.051450081663773295,0.07596005553443641,0.8950428861614016,
 0.1371971022769376,0.49084941292212914,0.5500935877551448,
 0.23639860529077072,0.34021781214407276,0.5248260366037367,
 0.8702218931116655,0.727045688065459,0.3001985426674726,0.33459989393464784,
 0.9428529610306422,0.9948722635016416,0.24492185209715445,0.5936913456696354,
 0.18985986982359948,0.005006436087042876,0.7838108877204568,
 0.5259915944603429,0.7072691680889341,0.4289810664622892,0.8269650658202129,
 0.8418436998360295,0.7182269720602669,0.6794091804732638,0.6793512339580399,
 0.39907121375555077,0.12113581276719776,0.16115194334111094,
 0.18838900071007714,0.3453726972746144,0.5440723741851534,0.7932144936981806,
 0.9705883756851587,0.7316481101257887,0.2605798370407427,0.04384224000360093,
 0.0815313193833177,0.22011858718376287,0.9344968475747557,0.7854541978958404,
 0.838952934556295,0.8322995729215715,0.047068421285523354,
 0.37646264634973026,0.18972467066804288,0.20593771855939658,
 0.6690493138137618,0.20983115807004216,0.6236794011147934,
 0.00636403901659055,0.6238139960695781,0.06424000559728149,
 0.19254878883090698,0.09655425747752466,0.4074333413134641,
 0.2657090986131918,0.5539071999315742,0.019092251276316863,
 0.33475507673390004,0.04269844437192427,0.6220409131142688,
 0.009524647559183586,0.1564042286467383,0.8378978972559372,
 0.9704750889227359,0.6646380914032008,0.9164336911812889,0.21750867598753443,
 0.9371264022689919,0.41465690642424313,0.8227467956403135,0.4068410864045773,
 0.3253858945481103,0.9913743518345495,0.0671097307855113,0.9584475106961852,
 0.5866364043471393,0.2068064031422856,0.5269552447024937,0.1287885353066478,
 0.03372583248052996,0.13490777124489683,0.6021872120104608,
 0.8902608213458155,0.6944394951161844,0.7340709782353898,0.3686705169704224,
 0.24542606837042025,0.2381232155773243,0.487319848811369,0.30807399179366146,
 0.959745343319841,0.5047644232828307,0.4976014492173074,0.5769021814773188,
 0.8026970047177885,0.033137035014948024,0.7890128149668444,
 0.9288603179227446,0.030340249749410186,0.45622109909960196,
 0.7922852567087554,0.6092999459734919,0.8466023990349697,0.9569854866445215,
 0.1720448343814669,0.7513104876555083,0.0012087551033256316,
 0.7596828949788532,0.01970473882410595,0.08066558195400853,
 0.6105473839305711,0.41774641778596877,0.67144304169369,0.6947491200642171,
 0.7315138718933467,0.7677590536246864,0.6951992878568299,0.20935967561931756,
 0.8772918438378607,0.28033490887423107,0.6225397850571783,
 0.41030589742601287,0.3680393501256889,0.7889663073242307,0.5618720545227571,
 0.8793565452723799,0.5792599037700751,0.15058718738679022,0.7893665157287572,
 0.944975508296472,0.5577608183597234,0.4487752573423083,0.1828110738121108,
 0.003801623867655035,0.2213379662125059,0.0749915642828144,0.752081821792851,
 0.9520647623535085,0.07947002965166228,0.7120156302618513,
 0.13437080536016877,0.8629210130972707,0.29174440465311535,
 0.4212053473519033,0.19442752861244927,0.8839328342275866,
 0.23182870364917696,0.05576973885789016,0.8384251031153417,
 0.5532966007585791,0.4903014335564253,0.7946089391104311,0.2640832131543658,
 0.44797909709130934,0.44867298289949653,0.783118401988267,
 0.20349229012643433,0.825755433951063,0.4547705758158769,0.5360272205592576,
 0.2622271734124977,0.028652323721105488,0.4067542207610598,
 0.4558969753669152,0.06492883033865948,0.38735413127578877,
 0.5019147491277249,0.40527616080340034,0.5037458481830224,0.5734136129737734,
 0.8245447973242249,0.8727280588769619,0.20160592533060173,0.4806411388862034,
 0.8955006229411613,0.9465331029813102,0.05310324569391356,0.5073389413669209,
 0.17451090960157578,0.4398099816392589,0.5471163569465466,0.6331382439650253,
 0.9175806657526292,0.15770375196514952,0.5864310537248856,0.7243981853156204,
 0.30435532553808387,0.23635235980335412,0.1737305314100862,
 0.8821283500269398,0.3645162205229754,0.5396192755259142,0.232677227074781,
 0.1423042662363172,0.28830497656667053,0.2851273139675313,0.2992633491384271,
 0.8641446379006502,0.43456151576656965,0.6643637410936087,
 0.46008915752293256,0.18445347143587842,0.1469531070222012,
 0.6424714321381008,0.12481354085984886,0.37711862808640795,
 0.37027746389491156,0.788046586817699,0.5222331902396216,0.1529582956267942,
 0.3283773561394989,0.06171676771375678,0.29949051822717265,
 0.9308341610452204,0.6387357210421165,0.9770006652557922,0.5180047794906226,
 0.7823596257526141,0.4179352770527506,0.24673044670197508,0.8885827235816449,
 0.032450777348673854,0.17149866937639113,0.006689679454592978,
 0.654578065020196,0.017713888244891396,0.8937676711283059,0.1372072064609008,
 0.18381456365982796,0.8361192906328239,0.6987122139893692,0.7927134441080619,
 0.4526917175070694,0.8776676847961208,0.3830199704663515,0.8638427375601763,
 0.40656596533479117,0.0749010315809564,0.7116635306467296,0.8240172188170758,
 0.3858670591851451,0.14094123106776002,0.10244611051113073,
 0.23541277034213037,0.6041331277412296,0.9693560297788119,0.9405758921119598,
 0.41073031492543377,0.413721980286347,0.7379456424598786,0.29357033464054205,
 0.01368543928618593,0.5097131489270514,0.4251172447756807,0.124683382862112,
 0.15110309576989644,0.5185028554534663,0.10255621858247155,
 0.6387415608665148,0.5483676354635371,0.7029168814146114,0.6972691234052483,
 0.6913569287988915,0.41996361406374083,0.31313734677279403,
 0.9643685526116027,0.47101706300058155,0.734845377010514,0.5737664028223739,
 0.24638586376973315,0.01666121239738405,0.9264005698005682,
 0.16520417948478427,0.6189559359330711,0.8167033411004553,0.7347303047345303,
 0.7065900096493379,0.4700773442091455,0.3596363455972632,0.3930973819300807,
 0.49579852391318613,0.2258914629154437,0.4103754408239729,
 0.38341277443751365,0.8321303102258472,0.3127338553842274,0.9384873855898274,
 0.9218262131901321,0.31236967331712595,0.27004377979485605,
 0.005342876226082982,0.8927157227833915,0.20842677323352865,
 0.6485367245700042,0.28480293130161594,0.10392711461782678,
 0.17741891590212655,0.48114100027652995,0.582062486031927,
 0.022321615553962948,0.21056692099975027,0.417224172111346,
 0.8490760367215473,0.09298621199397628,0.8581852876491017,0.7135356242646995,
 0.45613354832836506,0.7642526474016665,0.41162074220948264,
 0.6579072027942805,0.3598328598997841,0.6900946749088644,0.5946231417561625,
 0.6295161825636955,0.5662488814815932,0.7363140755883679,0.47852641036403465,
 0.6152889243365817,0.6562566764192417,0.7845389681367604,0.02675506741496525,
 0.45061170154185304,0.30014651487833655,0.012011209946757861,
 0.9260717245579346,0.37206188306632015,0.344643934370227,0.3687797374398488,
 0.6849023715569413,0.29721365461064275,0.16925563771345953,
 0.09351615772364952,0.01924640112058451,0.35289745427434527,
 0.2932617833484126,0.42754834281221743,0.6969802090732137,0.4941560091887627,
 0.5395864848666578,0.052948291522832425,0.7679078816901965,0.528526914103661,
 0.31103810910815866,0.4082998129627434,0.4926995339905371,0.5366757723717412,
 0.8768321374010517,0.47185045943416126,0.2367483429208196,0.9946027372944972,
 0.16235421859531385,0.19468428343921218,0.4648802226452876,
 0.3012519615716731,0.06466812671678845,0.9164462739668322,0.7710009132696719,
 0.9887731161734473,0.9258857340707746,0.1163368977617536,0.8449592029485957,
 0.47536753108930396,0.2740069582194117,0.062198147993932196,
 0.9111154701978798,0.7832537809191642,0.8544169001664287,0.8822252859634745,
 0.07086109406229646,0.09997645654008713,0.26271770021684393,
 0.8354002810135519,0.8669474375125685,0.2150920608674436,0.9335885875608868,
 0.18055831275277645,0.29607335092281717,0.015225643124472277,
 0.982599001369833,0.4830633560017924,0.049567630487210024,0.9478554398429284,
 0.5183835083195554,0.30467285036331704,0.3914715727097511,
 0.20849621545541863,0.2826670900943845,0.6424297039839794,
 0.07952675120143826,0.19488083059170103,0.8025855689204762,
 0.9374365440965544,0.6228928217068975,0.6499063440514533,0.9207972519824958,
 0.15158983722629105,0.31439023537464683,0.1357853564928898,
 0.8487146025764765,0.15008105236141556,0.8763524400356911,0.6576296434751029,
 0.7210755988208501,0.07110010099398267,0.8303002610794902,0.809604022507529,
 0.792772334782384,0.10530132835300987,0.421480457834837,0.7659440122162766,
 0.3116704380156874,0.40987510160652985,0.0015055605080277479,
 0.4617401005953611,0.9834831936444116,0.8486865983505671,0.42180434516350385,
 0.5749543499927188,0.48260205458574035,0.3214055083421291,0.0748136718680803,
 0.7964833614538454,0.39182570623504087,0.6904162604513625,0.5418543006362255,
 0.3739752845277631,0.481594011560338,0.206600726700363,0.09751942998638441,
 0.6065581458502195,0.8040209231412246,0.49481004265389295,0.4018493651720746,
 0.02449139839167691,0.04417812357819528,0.4180245659553078,0.617818003183092,
 0.30566932704112715,0.37153956970048596,0.036886877645569305,
 0.10874014882643157,0.015572061290478723,0.29391843932979156,
 0.2702146050149278,0.6128451180242165,0.4435037450587853,0.0159168844311669,
 0.3811066646390361,0.3757168231134177,0.4918692080799981,0.5714619780009081,
 0.21896435564154426,0.9493453846730923,0.2654671278885352,0.8909722650875249,
 0.5430173962056815,0.3828204755401069,0.159417590371835,0.12394354711435351,
 0.9208977594345344,0.9505706724663805,0.2862846561077036,0.147115365647285,
 0.09251675380823643,0.015214622755102214,0.6661784706323159,
 0.021654210910191707,0.3159776351680684,0.37224088010650347,
 0.7064130178516651,0.5223627274763386,0.5953252753547252,0.9724817282803851,
 0.18313814276528073,0.13290497457533934,0.10363021204340339,
 0.009682244368286463,0.6376777843007729,0.2649037323722975,
 0.6950150113189364,0.32126056504566813,0.4086752537845858,
 0.13887609683381053,0.23046428545280928,0.654255131698853,0.40978577100144,
 0.29378890262158763,0.932361930014067,0.11397475535436896,0.1780597316399417,
 0.4850722991416394,0.4841241510507618,0.36426533646491455,
 0.06722794381240538,0.8955743440218905,0.6721499793999374,0.825812309892752,
 0.7371893573181584,0.13320900773352418,0.223679025488842,0.33715814445568426,
 0.285485423452962,0.5508759443195541,0.9783472737445923,0.29834134395303025,
 0.13896736473694105,0.8913999871230834,0.589389138501863,0.18827798237384774,
 0.17701128842309888,0.4702077604743148,0.7959854018375976,0.3023047952338044,
 0.45818640541361666,0.9716404815249532,0.45985693891421875,
 0.5882520090904827,0.9090549636128892,0.6518713633407532,0.3620633568019923,
 0.687990458102907,0.5389399491204193,0.7289742511070707,0.5076112094132119,
 0.6395668473219528,0.5457627465764741,0.6263373179238845,0.9323743161437889,
 0.7388857088905882,0.9060968148198518,0.4444279584221291,0.3007701736968086,
 0.9892546898778273,0.9466709161733896,0.8502726871687224,0.818428823864898,
 0.4931645538182672,0.49616113135969775,0.39366021113597993,
 0.005515935504778469,0.4728525153486671,0.6379870668994461,
 0.07352303883403699,0.49455877024879336,0.9157050425237536,
 0.29286902542853255,0.06717273101153809,0.1275403058691289,
 0.8140396596606926,0.701386212783574,0.697163201304631,0.06418404017742962,
 0.8473482045395297,0.1331572557718812,0.9220492453122521,0.38324506652931634,
 0.7477909821389416,0.8258546903278855,0.4111374828118103,0.6357853016834364,
 0.6883094524683921,0.8042851334687803,0.8497372388687933,0.7213739024835809,
 0.912383097293888,0.4728194815652248,0.2107442870492543,0.818311485079407,
 0.28066834949431874,0.9546070963641505,0.31867932682026,0.018719538757504095,
 0.020279422193296703,0.8044156174606786,0.4973798786876804,0.891327824701781,
 0.4056592405360253,0.6193169467945442,0.838309257643785,0.4514960975754362,
 0.5619250551474666,0.22462277367880312,0.5930459569010507,0.9085205540389109,
 0.7656491419221304,0.10333183361282694,0.07152239890723622,
 0.8760880629180697,0.5002715536386029,0.8484519163950588,0.7436989490703857,
 0.9279100368995915,0.7842414287783542,0.3776262562455225,0.7068161105676933,
 0.2775390434211312,0.5465244617346112,0.039723724045855335,
 0.44112923773381163,0.18385000847300026,0.47334648796320533,
 0.8903274784184084,0.567264612992481,0.8913948495513987,0.8754854093301224,
 0.665005855423751,0.9872191496847608,0.8991976278721575,0.7279613297684542,
 0.23934388788045413,0.49287703228802915,0.2889650558742456,
 0.12650257813576316,0.4563366857355534,0.708555120133137,0.6226180197521222,
 0.9544038994770427,0.1318836774560509,0.5334417680852609,0.3341569254734399,
 0.6397941778550947,0.7662551652058979,0.4885403295568306,0.3936857126109625,
 0.9499937727405445,0.7022069388896088,0.887352740531434,0.48678786687186626,
 0.41124797523441003,0.12612601779532118,0.8446763023537218,
 0.2019452395811384,0.3781770711902006,0.1306024094502436,0.12386997175321934,
 0.048571836259542533,0.06014787989495407,0.7858612009088041,
 0.3043811886291502,0.7262265783774399,0.0478280458838084,0.969922344941559,
 0.5604839165413228,0.14480938854820302,0.7532035168900345,0.9120760654588299,
 0.5982618760056126,0.9428658175996439,0.1178585274714905,0.196760224916252,
 0.7263601888349343,0.07102374579221626,0.25323799596214414,
 0.2222495668091492,0.026188049019734483,0.025458847771663295,
 0.33010713948306647,0.2895033355275777,0.013181786961416697,
 0.24980523159763568,0.015783577120481618,0.6622667183793515,
 0.8032023673389432,0.7455622654425872,0.8890226099558305,0.8434729409790142,
 0.22650771581072005,0.6307131432723845,0.553978469528564,0.5784568107460348,
 0.9093756721972166,0.4685525744443433,0.7763737800173165,0.37729971897241665,
 0.1909305882899468,0.13516736810266017,0.4597044919162604,0.9873313249915265,
 0.6634871086350527,0.2063800173370166,0.9707682801134985,0.5648044801330037,
 0.8000380035622916,0.49588964197497976,0.41535316704259173,
 0.27527155837906647,0.5292421685957616,0.14624145825242318,
 0.2723889577936994,0.5761532860913836,0.3229512158260791,0.7677750030730337,
 0.9361283700166398,0.6292448042812329,0.4753079672657692,0.6835866017333688,
 0.13527305115050137,0.7093826325831888,0.35486443300942516,
 0.15566944105352065,0.2588268629375914,0.8568350829955859,0.4791527894200146,
 0.5088809163902102,0.2032307420733952,0.16030541669324228,0.4845471221729839,
 0.8297505143793157,0.26963384354454245,0.8135558430200788,0.9342040716675215,
 0.7281156943987519,0.4991107253955449,0.9646093944281428,0.24724369371000843,
 0.9905605503901259,0.7104533824604866,0.19520445026400146,0.7189585312468003,
 0.4669551153241336,0.020805918081231445,0.5853897047822919,
 0.09031634514293096,0.5411972168919874,0.47873628611181607,
 0.07854299512246743,0.1013878313925819,0.33511489923054183,
 0.045978234645496774,0.0707284334295345,0.20375451345758566,
 0.7517945101322776,0.4330866638021945,0.5640059997386875,0.8575141253055477,
 0.6567341261620305,0.03844947960435441,0.6255631324784622,0.4733568866719414,
 0.4879678287256859,0.5702518576453269,0.26287092356860187,0.9700764907673245,
 0.7676563689756906,0.2345488879240667,0.3069305925153082,0.9117114831711345,
 0.5060347526427923,0.007591236206404783,0.4211568748221248,
 0.7768330240624647,0.3843736281334713,0.5886931639586865,0.570262136874377,
 0.5717628160501785,0.9882233224079783,0.32027266944669497,0.9525147123827468,
 0.8066224383189227,0.5523316470591748,0.789516120670872,0.10148882225552847,
 0.6919591650483564,0.3810046896090755,0.4992570097658524,0.23527190622077798,
 0.9468037718751824,0.9304368519524682,0.9247520672808317,0.9300459154770416,
 0.48211615279905984,0.08478034081403651,0.32875764741596847,
 0.03868018081504543,0.40910921657543764,0.6575161242753397,
 0.6478876290204895,0.9875206988541076,0.2656061147262062,0.4774646950601944,
 0.8170651965155952,0.8812489162683268,0.22716089748426438,0.9303348696562385,
 0.5637958169706023,0.028933552187578537,0.6158307963531647,
 0.6912994798691323,0.9980540586187117,0.8539763483838092,0.7963077216518736,
 0.7408104202904614,0.5340571316024713,0.0660811841916149,0.6075003576371487,
 0.8742109641318307,0.6495437767048305,0.19555922628878253,0.630323395664601,
 0.8464367398277309,0.5367580749238104,0.28485172269227843,0.3282057552248224,
 0.5861130631391464,0.3627413991208207,0.40270271027122395,0.861506727303397,
 0.07037478458410462,0.7662827850398903,0.4748973430671567,
 0.20501801398625752,0.5650739378642995,0.8510434398763868,0.5698263359338592,
 0.5018163583800568,0.9729146499379888,0.42260049615985296,0.773404586284921,
 0.14799499470475574,0.9893552632957067,0.20818210679579097,
 0.7747036582555085,0.20087950130902654,0.1774289052126724,0.8465260471944152,
 0.16823973934402014,0.7301724689913299,0.43649150407501747,
 0.24282598814955725,0.8217074699574283,0.3024651209778052,0.5927032124107285,
 0.635169685437441,0.10943997797363159,0.8801772936575254,0.3725065353987175,
 0.6016706546779078,0.15787644649513888,0.5160422109118183,0.7760349907380226,
 0.1711703985636086,0.45440352651697513,0.983782545774521,0.1544522010807906,
 0.6502836745510975,0.6300650098231427,0.18559693745823092,0.7862960622122449,
 0.4434623212558412,0.07051576685056338,0.06952891832097985,
 0.8771476573637342,0.6390403302086656,0.5058852046808816,0.6423983515871846,
 0.521394058111552,0.5949079391300061,0.8874517536780395,0.029968431124400396,
 0.6676440052900067,0.40401844395197506,0.9708590696392086,0.4604103068608324,
 0.264812660851519,0.43026758796864595,0.5666847563251673,0.3705339429090716,
 0.926765600743794,0.23033622304813983,0.04013269742825376,
 0.34415266097696307,0.43043807617550145,0.9655515158917054,
 0.5722427323991395,0.7034390625978145,0.5705832592647363,0.6886051471756393,
 0.9342873440154269,0.29231915116301554,0.211843729694138,0.6205985053882169,
 0.1161550308811744,0.28257690960213766,0.238039367877523,0.3869033284712817,
 0.8303995681838239,0.04124538441648762,0.00857611571343797,0.661867344726329,
 0.5917065983248994,0.7837739228803424,0.9145563047356287,0.24159005997193572,
 0.29248504305948697,0.2887092012936461,0.5649501933686836,0.9660469875923938,
 0.018866053718814557,0.2750766877907467,0.5023733496170739,
 0.16651496822414358,0.18520986588021482,0.12331007858534437,
 0.9910994243530311,0.933410005934554,0.057341902002528644,
 0.22122331745733537,0.8715150334352041,0.27579552775431204,
 0.21832509583710036,0.0994388335190669,0.8526769446100807,0.9946669956705618,
 0.7787548173774033,0.216645870923464,0.7119992732857003,0.10687383326125643,
 0.780630933628569,0.7825550525134234,0.892069368769949,0.7071802020485346,
 0.39021312864854796,0.7538348246724158,0.9973551556231925,0.7936038664440193,
 0.8583638369227831,0.2485453923992601,0.2386815268240675,0.4321323837068469,
 0.301216825794422,0.670053237319788,0.7259152174136067,0.3250858281709412,
 0.9463668893681606,0.6699594902712328,0.8610868463764916,0.3263489627608046,
 0.3074972918529928,0.49799786863768103,0.8643022649690986,0.7706387164416351,
 0.5150361163612273,0.02979321815925029,0.6635670254525938,0.4797835290332968,
 0.8984477729412446,0.28311390360778177,0.5920218700301921,0.8794494003988108,
 0.05545714251706124,0.04952782613106854,0.956702852297613,0.8725142453401945,
 0.8115237275043218,0.6609956025591616,0.07781649362753718,0.678094919417652,
 0.22088433520674378,0.7658429907644739,0.9361338642657289,0.8152511504443949,
 0.5644538435304196,0.9057419003565482,0.592467410235921,0.6521645566313505,
 0.08244232066720791,0.07624644689009441,0.09800048062782996,
 0.5986911115467167,0.7834327118821671,0.06831526165417268,0.6918286178965155,
 0.6010493663297174,0.7957720015712879,0.23144267922144834,0.4742800145394047,
 0.23347009188351686,0.3006909636655982,0.19438024949511856,
 0.3312285255159665,0.02496415525966389,0.8231483794098765,0.7609994528941451,
 0.970728682772832,0.25820430391308835,0.4611624667243682,0.9820516623105552,
 0.24670734710748932,0.40467264999259445,0.6625293194028452,
 0.31916241912199916,0.9059006100844326,0.6246068597096011,
 0.36112523968645316,0.9458793480101594,0.7735907150116871,0.8329971693309799,
 0.15907540927377217,0.07382049038845462,0.7178698767302927,
 0.9790108273054372,0.814236135024762,0.7011227870065174,0.006122990067820799,
 0.22857555585191347,0.9073654266806123,0.9465337635771714,
 0.0022554715329394792,0.18515996954390712,0.4430989701313801,
 0.11816854584494751,0.2643001354303933,0.07493413286480832,
 0.13532992100122465,0.11991728333959073,0.396867583125349,0.4323979932576293,
 0.40003165265824825,0.2538175910168109,0.7189012001747133,0.6657045932065353,
 0.8000893638402413,0.6182148858551639,0.01082700527104774,
 0.32449643867924394,0.9054163212914212,0.7926797613782456,
 0.13205373190929792,0.8094806900023217,0.35302954140252163,
 0.5609971696261392,0.5122708636786831,0.5661691150285761,0.708800005025228,
 0.887897156136044,0.9986608030161253,0.5327203808929397,0.861860646741637,
 0.9420887502670159,0.512255948039537,0.7099717228002321,0.6155118721658932,
 0.27142967628348413,0.6743053772054748,0.013581485122772907,
 0.5259372928551964,0.6789931076699653,0.5394361478066745,0.3483759076374954,
 0.9278284384562461,0.8204238901253901,0.854307199243691,0.4813693894722503,
 0.9939872989229742,0.2502473597215905,0.09304639791977887,
 0.40859748197942714,0.7652034373619401,0.756829731465234,0.8600870664775925,
 0.43265525589369846,0.8368743252022326,0.1803409121915156,0.6048394752512447,
 0.08829169161330075,0.19265623528740683,0.3576039555866466,
 0.5794895932863737,0.34814954565365996,0.5567776740836475,
 0.18161661526260564,0.8913446618063204,0.9187967003051036,0.3032011022291764,
 0.0993100797990154,0.2396759920514936,0.6208621580801111,0.25403134701312413,
 0.5198332632765195,0.15593706476850022,0.2868798720253539,
 0.10931637412395956,0.4305477655049784,0.32745027123531145,
 0.5186594462047323,0.3740279779600687,0.8857616858537423,0.6124942657360326,
 0.22032186846544355,0.26516470176952645,0.9331797874410797,
 0.5595547826197962,0.9804402900670502,0.4339819522904681,0.8496643659312609,
 0.49556596100533024,0.8492728543763127,0.857066159768916,0.3103153288556416,
 0.6453231835133391,0.31703388755275275,0.4846199367855646,
 0.13284709501235636,0.746138974888248,0.9466741491017715,0.7422328205994229,
 0.6280545868659031,0.862790833905815,0.5838535677282467,0.0977475806589464,
 0.103489402391432,0.055671725304719644,0.5922466823609699,
 0.14424842669455418,0.6283148445091344,0.9360536346965369,0.627847410133928,
 0.6066825231398976,0.24855432789691645,0.6724417821503174,0.7068049400508097,
 0.10764286631482989,0.25302524430734064,0.7578827283996086,
 0.5061108857903742,0.06472757757088266,0.4550295883441319,0.273601588901324,
 0.2548711348287851,0.06325406985754334,0.7996699306851535,0.929260771088711,
 0.05176317112493889,0.1340151622196073,0.42070565425569084,
 0.7205337001639269,0.0727623036774192,0.48415698113597794,0.5574308983307892,
 0.6596623786936432,0.33757287200616726,0.3129330998128206,0.8882020816276098,
 0.10310530482297553,0.664760597590477,0.01921234628162416,
 0.10961657940875658,0.13580270399111416,0.5023246942468729,
 0.5457888154249595,0.8412119317726727,0.1322772079432264,0.9630484551487833,
 0.11749205154702824,0.4336670900195313,0.23276899556289676,
 0.4678223306555428,0.11525131156472757,0.14821761472234296,0.940684805754597,
 0.6168832490789503,0.7408422957840581,0.6894132844108507,0.6497245989973107,
 0.5021962821424774,0.19082979200833128,0.6047487831157592,0.4111574563637601,
 0.5434519479811609,0.4702298991334404,0.9883406432059387,0.13202820804727344,
 0.4874366701816376,0.24815491190550487,0.635630644650711,0.29584860934390866,
 0.9472617158968799,0.8477099934269537,0.3860918232674133,0.05164305238603106,
 0.8053017696486151,0.13616390196423112,0.7231366175688406,
 0.45712191952778225,0.5904536721416112,0.992454452607946,0.1185043515384604,
 0.6422787200804083,0.9591168798127305,0.6176854889958823,0.8278444252650827,
 0.45043698114001496,0.4892666720328368,0.5567198728572358,0.5437390320324546,
 0.5445171617749506,0.9064579977782286,0.3089544799386549,0.68397558635467,
 0.039903910970380374,0.45452668947455854,0.7041021030778173,
 0.9922371137938686,0.5705034124107532,0.5609835601406128,0.22996729097850988,
 0.2812575124831648,0.9217109401915892,0.5491619272027355,0.40776739648023985,
 0.7983292780027615,0.9027712218290155,0.2924380874489956,0.1256415158440124,
 0.9923306108723371,0.591320850917551,0.7578218721393266,0.08021172894839146,
 0.46613338523590675,0.6998952123060856,0.5503786211689801,0.7653372148326432,
 0.18295212132030902,0.6092030117332767,0.9113021472611018,
 0.04121145576181118,0.6143667557253069,0.16771894628705886,
 0.015999009179097468,0.7796394812523224,0.8871878559634365,
 0.6428287611275809,0.7357251767969023,0.4836368490757189,0.27461196186769377,
 0.8928371547898909,0.7249772151562703,0.20487116580173503,0.6632722561432043,
 0.23412774273991588,0.9882505663327262,0.45932921092891843,
 0.7001709175361437,0.2927768583023015,0.15481798711726558,0.6706706134725344,
 0.5093171391875817,0.9249291047901943,0.9122666300587481,0.4934754800752722,
 0.04881559004571445,0.5661698771931838,0.594607954958561,0.39372505681601344,
 0.8302851031724038,0.1832223523921741,0.9815262822112489,0.2978878877735759,
 0.7236292878928843,0.8441142122532664,0.18751781758320119,
 0.13783943531622622,0.4593273160432564,0.40491607465276247,
 0.8894549970697951,0.12892019054554704,0.761408851344688,0.2860594574560801,
 0.2825705083636195,0.6108688122072391,0.2826238229740502,0.9071534210326521,
 0.7380988468367967,0.411957143025388,0.9553132871281638,0.7990609834912421,
 0.7784754234562901,0.7086249373743648,0.41566068244814447,
 0.27307226435149934,0.3790330851894217,0.45279029651033054,
 0.6152399858277041,0.30043993465405994,0.43345633562224983,
 0.17299037170682974,0.16358016075965642,0.1165807934493307,
 0.4280271434812728,0.21158687221103667,0.14578648362448732,
 0.7623488830896905,0.9784536325276922,0.5007877230629574,0.8277635797128109,
 0.8900001280587913,0.5840357626601765,0.4806600933325984,0.2148153297109412,
 0.40369051769238007,0.27092457912709955,0.6913371971093006,
 0.5631484332906256,0.16727970183323937,0.6862205098282259,0.3314149571177203,
 0.8930017634039197,0.4794960822621749,0.5694449905064771,0.7053399124660213,
 0.1552413200302013,0.13249249049843514,0.032830353864155115,
 0.5973396864173622,0.9316887324375133,0.6608001108048112,0.08703478621565253,
 0.6415947510391292,0.6664087885475001,0.14408955554250735,0.6783791024611024,
 0.925189169037743,0.9258044413187705,0.5665783781482072,0.06386162519677974,
 0.9527411748678236,0.9276313142870547,0.6017102749631436,0.31086747370047696,
 0.9087553414017944,0.8815037050608758,0.9459017894529969,0.08344155649052398,
 0.44228548736197637,0.806633071487554,0.08613356434678887,0.8993676511200843,
 0.21414344657348094,0.7028177823138418,0.02644777016894395,
 0.8703438342541984,0.5323296359532738,0.046107026638105286,
 0.7241966628031837,0.9614022566672219,0.5496387639558603,0.40411535461862236,
 0.006341797259506121,0.0998639031041908,0.5393806137578218,
 0.36660628719734223,0.5994339335016653,0.09359348851865656,
 0.7074178256097298,0.3466903459034605,0.8198473609163113,0.8412145912184283,
 0.27162468343579527,0.08280410771928726,0.10591741012135014,
 0.8410266024128554,0.330515577095577,0.9390750509112402,0.6253015255188862,
 0.4661953710413037,0.34337570423926644,0.6579358155534502,
 0.15411037491626656,0.34072270909884317,0.8004449656604689,
 0.15576952492591367,0.535206220571267,0.7555516784608743,0.32864600655586995,
 0.7808511529455281,0.9484676578914242,0.384790442038103,0.6861251972847902,
 0.3969849065640534,0.7035385607092892,0.34476506244414284,0.4833878340487616,
 0.7819020163256039,0.7420231945961645,0.5980445048593482,0.2867923878270703,
 0.4792339850338735,0.08410494865848073,0.020013962386607242,
 0.47672776620996915,0.6235423761878531,0.13102506123602564,
 0.32402999936595955,0.2966828582813742,0.144176543199956,0.7590511552420128,
 0.004862304504353432,0.956736584643724,0.2342880983747615,
 0.02784575213909868,0.44910447516217755,0.6773404742758392,
 0.06933917670304635,0.46778616850699994,0.5391061696802215,
 0.18545240800061302,0.28558719026002477,0.9457967778442754,
 0.9552097660848039,0.5399056301704567,0.05006522814975367,
 0.29466499111983735,0.4050886744230644,0.8982781639141708,
 0.32261680959140726,0.6531034611532447,0.9313482278470613,0.6131892812281025,
 0.6798870589659413,0.24939167309840082,0.607507618151132,0.2491676853242033,
 0.48953849677280337,0.10040531209312464,0.15416002589377542,
 0.33350199034924777,0.37445216026828465,0.22919221866449413,
 0.49813846990340727,0.3485861197390895,0.23684648304072764,
 0.10954474360578814,0.5248448060749408,0.0035379023010677635,
 0.31569049259936655,0.15925154969085198,0.8557440361756439,
 0.5807398887076803,0.0023394834037477885,0.0014505755088337624,
 0.9368896623082261,0.11846314773220656,0.053160567429755834,
 0.4165230249019781,0.9796654662434147,0.043638396670306334,0.780362713476144,
 0.5527309724484114,0.08399180872131573,0.41235842912357623,
 0.09890350712461526,0.13497099507986132,0.36107493707608995,
 0.1251219729913785,0.4847352556740925,0.5913427657305461,0.39548188172154697,
 0.7408282176705407,0.2988209178440272,0.014817907498555494,
 0.3270805312612477,0.053513376286423386,0.9671481288730222,0.547721208934366,
 0.6953167207820885,0.6692529294800654,0.94081300441684,0.6603569046287456,
 0.6549262823843931,0.16430305876284557,0.3680872485848813,
 0.07705887526197686,0.14865656985738995,0.9776189527169749,
 0.6741099539246208,0.9947935741967017,0.16729573814156296,0.2329574043708179,
 0.04474914416483855,0.35968324388035466,0.013747427913493482,
 0.5578184311090039,0.9151893592781339,0.8634040415077137,0.4319294748509014,
 0.9288429847983752,0.5329153660939984,0.8172801569348684,0.6584559518509532,
 0.08524119074826253,0.11735259266783349,0.07029650619237793,
 0.17790095769367154,0.7532893237693115,0.982104753227401,0.33659910711065577,
 0.2686003483023832,0.29283550687414917,0.8782676552344493,0.8752960645036898,
 0.3428148340969749,0.7645161489720592,0.4076445322612148,0.9891656667399575,
 0.167524579476104,0.6034549155569072,0.09127617366470131,0.21836531754508204,
 0.18093699391436713,0.35466341864148687,0.7849320494050892,
 0.4977163229538456,0.07559848953507886,0.11032241624530048,
 0.6262262254060917,0.7057773669565016,0.938472868163917,0.9502643731467824,
 0.49061055249722163,0.7328154104173419,0.14299206329109027,
 0.7798618107685028,0.9109317665771604,0.7635556099481734,0.7344344991149091,
 0.45378315772162836,0.7878185958288818,0.8150179214235964,0.7969909181720266,
 0.041416004008820284,0.5389454253241144,0.5889569450097638,
 0.3832305185938911,0.718913325075606,0.5918818929854532,0.8609300528983466,
 0.8830380334134658,0.568619432003086,0.1006571821026202,0.8617329238061094,
 0.19287888547311027,0.1259618113848182,0.8201726941860726,0.5531848718680057,
 0.7660316678414706,0.8605210465508738,0.30886779318559376,0.5495574728313277,
 0.3768850420179165,0.9073719451090652,0.13171956756389303,0.8906738712682409,
 0.7959679396467507,0.9577457693083838,0.02800550778454758,
 0.06411287502889274,0.4532365392007278,0.6733531115099953,0.1625426993883563,
 0.9308982319646422,0.40818701783890154,0.5375520291606886,0.7355405992370102,
 0.6536913489534911,0.9995620398097127,0.9851252945585011,0.5996555200710151,
 0.25243912499471877,0.727911865455768,0.39505774945335126,0.2905621266882221,
 0.03497378821707586,0.40399714968532163,0.6645349838247195,
 0.029160597029267654,0.7237561612859909,0.2605735182071822,
 0.5998358056557512,0.3080877069480097,0.18053115468572378,
 0.43830689348355645,0.5094641837232424,0.6392852957699932,0.6950503920508967,
 0.5179249369487707,0.17386244414922747,0.34072639092090395,
 0.9978595626012062,0.7139940255788109,0.4092816223316894,0.7359221666435308,
 0.45685432587814934,0.3977961991081145,0.3385303931078891,0.4535038537947885,
 0.9403547552169665,0.6309125527163009,0.9631744455042367,0.6916637056767388,
 0.25529306032094123,0.9834910469044758,0.5435520746120549,0.5853966393571799,
 0.7278277112173992,0.8117544044500196,0.45170963857790836,0.8156540541519183,
 0.39557121802502904,0.6716660946699655,0.9115828058728673,
 0.07605732765114848,0.8067025291098255,0.14741923650938304,
 0.13974607035081088,0.8048390311181867,0.5554661939783032,0.3768814964685673,
 0.6118373771619241,0.6687319045297303,0.6803405212182201,0.22624296152794254,
 0.5534700025147411,0.38839794515389814,0.5933433998707045,0.5987553124074035,
 0.6234199677082741,0.478560828391579,0.057268500871783146,0.6940425995277723,
 0.8767687072196323,0.9222402164613629,0.7248911121249425,0.04228463974928898,
 0.635612007159997,0.9663055298655738,0.7547091916197137,0.348282263993692,
 0.014023184068318373,0.3179473632883427,0.3538627130618701,0.618248851906584,
 0.6268886823529332,0.9426880801097902,0.784858934762673,0.5352578727342635,
 0.08081017227508269,0.10616862225588752,0.86442246381435,0.6645770028305086,
 0.8225833007861071,0.6307272454204619,0.5609494250307308,0.9999463756120403,
 0.6774757901456958,0.24871940379529678,0.10216422021082616,
 0.09367202393613072,0.6580534219401962,0.41939180406887544,
 0.020668348076760545,0.5708542129651836,0.9257433168264062,0.911598968768973,
 0.3626794763947935,0.7749162261529569,0.24627149860313624,
 0.08656115321838642,0.9210776945398017,0.7926311453872917,0.4344332198898748,
 0.5211827288790036,0.8645294818585914,0.10202516725627153,
 0.44894678692918966,0.0119338213657455,0.2697022408026267,0.9101559054414003,
 0.10058165031979616,0.8440782357929014,0.9637373997922121,0.3019962808636354,
 0.7141131341922478,0.6824994876240728,0.4101829574334428,0.7580014701390346,
 0.47747348610421936,0.3346006482307129,0.05655765001438051,
 0.3162847448434616,0.240382898070592,0.18398349595445784,0.25549111260393864,
 0.9665963084884099,0.664444053052879,0.7473790462657621,0.6641433258247534,
 0.9992789660425286,0.7549614184269404,0.5751655473814347,0.11416647379944733,
 0.5010643547037728,0.2500666628506293,8.206552147014268e-4,
 0.9168230662631247,0.7304152540517593,0.07983615460907567,
 0.07156345961082766,0.5545379373274792,0.284341855002762,0.5667170345813367,
 0.37212676492300945,0.5168190407138322,0.32808114172292724,
 0.6922742764881598,0.30147259092661083,0.6821789975921397,0.7689670508389836,
 0.9740949588236506,0.8798006051735328,0.2877015828738191,0.41087677679726253,
 0.2607178589951469,0.559493319260957,0.5030699063500762,0.9125219146189937,
 0.050713592125356644,0.15561721001402906,0.023279025896296823,
 0.6020952629253342,0.23581207707468654,0.4324505384177326,0.694376120718193,
 0.8468069893857347,0.01919893146442997,0.08719763835367811,
 0.25648044899119093,0.9675957676583732,0.5062653343112701,
 0.09621470365749418,0.8849124728569534,0.49732056534490776,
 0.32753589576221287,0.11487861660345122,0.696035422780623,
 0.47014372075650057,0.6603964849317387,0.7116020376268228,0.6716077651745709,
 0.04317890927720858,0.8248310188542884,0.9209891465663302,0.953669127810042,
 0.9146364321929972,0.6857555395122056,0.5841110291545717,0.9176659242458173,
 0.23427282237625424,0.2752396226009107,0.5203844749142582,
 0.05059647287792435,0.23866206198111095,0.030980125078800658,
 0.3097686519660008,0.8925007711374188,0.9366581268516107,0.8022125710819423,
 0.002751373490884834,0.19344551645685348,0.015267607200615707,
 0.8877012345751781,0.9444546042405038,0.4616950395924484,0.20138657895222278,
 0.10173793489343719,0.32784949309071787,0.9563986703034236,
 0.5263548220737218,0.04357591550156048,0.13781494614363343,
 0.9905744656959513,0.5438157831714342,0.8648442457655183,0.8293489879946194,
 0.644572649411242,0.8582796847965136,0.6046138927782798,0.361820818299194,
 0.9711773914698647,0.43194406084156567,0.9905446601836098,
 0.23243321295471386,0.8467190026424545,0.7822686770049987,
 0.09385320650937912,0.8050552071992182,0.5264924326183155,0.7327076183826484,
 0.06468955419400846,0.5632707345743246,0.6864177780053221,0.5192968626818599,
 0.35668456112008107,0.8816796268664439,0.6082082290784391,0.2972688044403836,
 0.5388346769382573,0.35904706223611627,0.15597802669763017,
 0.00865819162237802,0.007427254506854886,0.8728350726711006,
 0.14210830634540172,0.37094634377281466,0.13198481554539898,
 0.21557147694539047,0.9549745029421324,0.7978436719415785,0.3707748825316387,
 0.9201791683011797,0.3526888376549554,0.2096013954072533,0.13831167590051352,
 0.8351206221573088,0.9607027128939373,0.35205344237396097,0.7004023642886532,
 0.6906881043337018,0.4269299045027447,0.008749489166721025,
 0.8213475739709007,0.5287435110957924,0.2048445644734188,0.04719084736921697,
 0.8945380457836049,0.809538653488992,0.7490784745379232,0.025543650725263878,
 0.9889617577795811,0.5663130683670426,0.07638093575743932,0.0850326009653023,
 0.40633760303528366,0.025889377856314955,0.37090194432762724,
 0.9091017164380428,0.8024580785995806,0.4491152051387869,0.14444003784566284,
 0.06118320575466718,0.4338634639025094,0.43145879765649564,
 0.03912385648422678,0.751211555675629,0.006201238717002622,
 0.6632886247470302,0.7009707202502061,0.19136065413781211,
 0.37059768761292833,0.4772590014296836,0.46484108146912406,
 0.14645342973955267,0.2774523373016631,0.49106042188792887,
 0.2809652145990118,0.8034372320118298,0.745823601972422,0.641171730949738,
 0.1648072483703612,0.9660956269687122,0.8746644965654061,0.4303992744076627,
 0.9380286895900685,0.03977022583305123,0.29773068391351687,
 0.1650819837346461,0.882658333365908,0.20651334204044014,0.5907241020531274,
 0.8632943080423658,0.29174090741434244,0.4688343697151396,
 0.13248559332323162,0.8451556810621069,0.2812355037254328,0.6958432782536081,
 0.0479643099142647,0.8178874291624689,0.8870954242619252,0.05732911920389727,
 0.29625688922845006,0.8850198602771151,0.577154101031188,0.5478303398157598,
 0.5442790785168001,0.2187486222171744,0.5469784274790026,0.2797590475325946,
 0.1483014376552605,0.5743308318849893,0.5677029545302807,0.27707323048878596,
 0.2543253236902201,0.651658541832481,0.9069490004846177,0.10900907472097066,
 0.16877876241411982,0.5790597967047387,0.8871373922909916,0.4963534156611483,
 0.6986944673201566,0.6568962501802169,0.7607264303679434,0.1896718571220345,
 0.6190209331627011,0.8332412631933733,0.23770236438521497,
 0.35819187065562974,0.6998342183401873,0.8313902760035953,0.5481822899936324,
 0.7612407620834112,0.7850118808235486,0.08891422016386064,0.5135574270155912,
 0.5250326825493787,0.46462203164785787,0.37803246337257657,0.662667176204386,
 0.2349807243539861,0.6024227560804526,0.2784982236954451,0.5958543173713285,
 0.06924755893940349,0.7533146380990565,0.3564775271966877,
 0.34388753805692607,0.4957553131041441,0.27702723686084474,
 0.4367044286670071,0.15883631135783216,0.20117509954036128,
 0.1555422205766941,0.7338346161957994,0.4882354943064774,0.7524097418855615,
 0.24307631802844398,0.3374656267712093,0.10840308691920586,
 0.4215715637707801,0.5971365086211765,0.8681748803575335,0.3590932483199062,
 0.1403211025882648,0.950895708657238,0.41745331353443615,0.09632008192615005,
 0.6503991993192684,0.16273656232150946,0.6749802553031049,0.6279450844002789,
 0.30996760698819104,0.7713225367051062,0.9487933553630692,0.5525512980094036,
 0.9434135859625434,0.5736431765813921,0.06747364383936805,
 0.41301933624577436,0.7654183279461481,0.4371104417661852,0.7039973666420158,
 0.7160248276216337,0.41432458878755296,0.5256398089016421,0.9520764741413773,
 0.059174912086325526,0.014406140946613744,0.4422657049119917,
 0.06602719462307527,0.6781616407163915,0.04640948412765589,
 0.010231489490060142,0.9721290231573632,0.542281943981728,0.5473639465719096,
 0.7851332906527835,0.8749360921049294,0.8264743445919844,0.5265783795544096,
 0.7965627050257749,0.4593628814161852,0.10248648699612661,0.8003508404119636,
 0.635409849453685,0.034131608093407984,0.0013064938029524464,
 0.4410211119281833,0.17897761178637828,0.37541143378410236,
 0.20743550693503943,0.4648054143484551,0.5899248042953605,0.6487650770101273,
 0.28607962581433544,0.5959645524399756,0.4946550713188489,0.3749280247304301,
 0.5177170070389403,0.4530342571218816,0.6462880949317036,0.4284780348142393,
 0.6932474738428057,0.2648515856164535,0.530668138142995,0.14550321747979256,
 0.8298684322967875,0.2564658318481621,0.6403660665015495,0.9782211534135317,
 0.6986687761514656,0.1544380701619502,0.5809417748426908,0.9434727910811286,
 0.9341521227908883,0.3969055351473406,0.9732329658696313,0.8862420199830644,
 0.007980843627720335,0.3643480375161785,0.050038853403844064,
 0.06292688293883764,0.7162692144963079,0.5797621714056953,0.5038435419580072,
 0.977925304590687,0.2554577379586328,0.7757307717260762,0.47172254892237886,
 0.8615244476592976,0.16350934406086537,0.26687512024381055,0.784644815897933,
 0.08437930902819668,0.9586112165271193,0.36475227239410857,0.437834976128597,
 0.09956740027890743,0.9971023258197227,0.46917031224180816,
 0.23939806892326987,0.7946654924421405,0.6997467772423008,0.0283038505579708,
 0.10924029076838038,0.5484904057775797,0.6564869166152832,
 0.13077407242246042,0.13811262260641055,0.44908748358531103,
 0.0518049609490181,0.18481892959232027,0.47473028739349177,
 0.7696841248289942,0.9359489468979998,0.392656795570826,0.5123282029349956,
 0.587965350023902,0.2125994112532481,0.885913254120445,0.21062251119659636,
 0.3767414593945936,0.661165269478566,0.1856884391411291,0.30658672492657024,
 0.7743834799669409,0.5098776208346555,0.5786425300333362,0.06111943410788534,
 0.5546602631080465,0.8306359397735712,0.9831925552354555,0.7602855900651879,
 0.5955917482888734,0.8184107808653404,0.944197280698672,0.38161401533671824,
 0.9753259312910815,0.42284514807414353,0.08528594425162439,
 0.5068309981575334,0.2770606054713628,0.6293447008654783,0.7276583680847306,
 0.13443560808033195,0.6442054504058788,0.9273916246640088,0.1460405224602782,
 0.3823687842633492,0.3433738713179836,0.04040431074984707,
 0.016216564686155066,0.584657748411842,0.6341560464013927,0.2186426485563122,
 0.5436750939156513,0.8759229949804297,0.7907304733861059,0.41888630503182356,
 0.8628732424085133,0.3655352258365586,0.09912941174736023,0.5397451330579153,
 0.818277618772902,0.042534751398674686,0.07073809846325962,
 0.6137254960836269,0.41533757059522225,0.4992737506233448,
 0.10955319397682728,0.9185926041585005,0.24300116241113145,
 0.41520155038828344,0.25437365454245786,0.47792578356127213,
 0.35874976125411306,0.5332380142202948,0.09215995491697682,
 0.21527319476703144,0.04696549990916288,0.2238712027094547,
 0.7729451362645823,0.8753043818641125,0.20295124846409873,0.9266923333271285,
 0.8118164485878943,0.23249902423316582,0.34832678212445756,
 0.8997155197341733,0.36342297141818414,0.6302047294981902,
 0.23896113267579577,0.07570974022654997,0.15947675290006325,
 0.43672259913599343,0.27443769321031053,0.0032400980385981493,
 0.9359336089348647,0.24492713902902086,0.12215043697436712,
 0.7990333708488708,0.34894511096709047,0.25172396360294524,
 0.3253254033052806,0.417849361289119,0.6739657239485604,0.45703363305622335,
 0.9748839328145804,0.1337388928736013,0.7139981353087648,0.9663414137762764,
 0.7633674636772795,0.10347452654350486,0.289045408312936,0.4973053263553473,
 0.2884614408856846,0.316847255235442,0.19438682436815524,0.7198441937251714,
 0.6199798273003871,0.07225530258145507,0.06740741196249578,
 0.6247186336513719,0.2647007572548902,0.40332888138587386,
 0.42649575867768663,0.13228822728727585,0.2716191028713685,
 0.2382709071391632,0.3527354066984274,0.16698206591461573,0.6192745186843751,
 0.3529473309915242,0.12189398017723596,0.6801058418607087,0.9693223241091891,
 0.9721203988685618,0.0034380227634567273,0.3091990326391253,
 0.6864786667906327,0.41806442023822565,0.40046443850301505,0.534606427874677,
 0.5744835358480953,0.7744535250675207,0.8619448815136361,0.6726440090349708,
 0.6083150385517267,0.6061617264050059,0.09593612912197136,0.9902423914053452,
 0.032625097201040454,0.4257785281177253,0.44786678106959865,
 0.1331428489804236,0.23815019362646095,0.7568784528724086,
 0.32110989433774184,0.48615360493036797,0.2470299027586852,
 0.9869042039069048,0.30774812975787036,0.8696439072347794,0.6911077956042418,
 0.6233542754337305,0.647286514362652,0.6267231356211551,0.6228058670498111,
 0.5265899958815239,0.6526334669549001,0.5820633139445858,0.3908926569783375,
 0.7564776658257929,0.06320729083480225,0.24376834243929535,
 0.15246664841213597,0.1338598075062416,0.5112592992395015,0.742965708572587,
 0.7584069619008585,0.9893979163957506,0.7574181261031641,0.17838316476818983,
 0.16957655677816874,0.4030378573766096,0.5233149923576634,
 0.43720459610634066,0.9442944117866654,0.04636213126676747,
 0.4622492513246148,0.9860458018209265,0.6273446183214849,0.2003820111557031,
 0.8242614159123172,0.19612435409544904,0.9522910235359331,
 0.03989701441890914,0.7936890808423959,0.6498085063206501,
 6.938942248420066e-4,0.46750720625909115,0.543028501938534,
 0.5032708933431446,0.47844466037923306,0.960249115636379,0.6979559849757667,
 0.45700086411630725,0.2846939882578644,0.9850666213673773,0.919443729279874,
 0.33364731086302024,0.5730875454781305,0.5710856951345549,
 0.20238842496980802,0.39188107872625244,0.9772838193355314,
 0.9567907374501494,0.6310961327760229,0.653466234424606,0.883052985690444,
 0.41891762504044106,0.770914336786291,0.24384005714918922,
 0.09289433769423971,0.07692802821356914,0.1106629815699175,
 0.7986837743859762,0.48243360489855436,0.8692901485951848,0.4033944633941633,
 0.5311007318869273,0.7828223753898729,0.9717034734113392,0.7831099485447107,
 0.40588306932384455,0.810892248726878,0.27133333681046146,0.622294334622176,
 0.04277813538735653,0.3108819073278223,0.7047967020013431,0.8142984374451789,
 0.025475831209749944,0.98869611171031,0.8525185900150535,0.4856040919835163,
 0.28003622064974465,0.21687669822395006,0.8784154001492924,
 0.26177775040442186,0.9796379337466827,0.9576121442293135,0.9855265996326088,
 0.21212658728100653,0.7517999540983528,0.7581602503618582,0.9323292759802146,
 0.6314629380620008,0.9677495454312818,0.7045950832726322,0.06329861732942288,
 0.2336399781899785,0.15849405019267682,0.11271323805873912,0.996029917771596,
 0.7274336680528758,0.5380610987062185,0.6415636088512504,0.49184907824450186,
 0.478231789197052,0.4485530425276122,0.5724955234171996,0.49077646576517875,
 0.1435751875013329,0.9801633826481845,0.1683061068152254,0.5698181315484181,
 0.634166049821864,0.052488149874916235,0.36472077134135683,
 0.4885662431573323,0.5515722421489555,0.2295724080036865,0.7099740375199651,
 0.19470653666163174,0.5901715149377427,0.4447823793322918,
 0.28462295096949386,0.41991093972189,0.12328590372522241,0.7857090111434613,
 0.8411932102845432,0.849577073128317,0.8318493838066013,0.8199070271234288,
 0.8656893217830804,0.723032005365813,0.722198208512576,0.0035604625812518664,
 0.36124655760921287,0.9587036327601647,0.40006342767319225,0.099867007768073,
 0.4692555120799744,0.7415248316132595,0.30302163298914575,
 0.03116474839930161,0.8779909688684925,0.8074051443721846,0.5413631963964202,
 0.6426364456231466,0.2994161339433945,0.15931218940878333,0.1399837211951136,
 0.3847041059827466,0.6175239820305528,0.7438051992863274,0.906497908199097,
 0.8740710668016198,0.961246805177058,0.6791376414321268,0.8072623741922684,
 0.20089692230989675,0.1200034118030131,0.12773157249234823,
 0.39799434904906617,0.013301250163281653,0.07879118831520437,
 0.4779586766156295,0.946268895162368,0.5696314613929365,0.8566074559627096,
 0.180950930268437,0.9470289594705604,0.23535076862358484,0.4421156799304865,
 0.7185928861849246,0.47852981654804994,0.0732711435181781,0.2133623799685569,
 0.1817811987986757,0.46987144220566635,0.849856235774894,0.9945811185891903,
 0.1745053374438852,0.30289136604216815,0.9376637533592468,
 0.47578458222062625,0.3941378075848544,0.4257160336228728,0.9782215939669887,
 0.2957572959891632,0.5955852208115622,0.03160390832828186,0.5463478689784242,
 0.38805228246649126,0.7686513629859331,0.6631003901210965,0.8594918379173926,
 0.31446182445569604,0.4636527466014908,0.4216551514300748,0.55822861788645,
 0.7786910806030911,0.4613720933625186,0.8667117085410074,0.9492477912335149,
 0.8371240431898841,0.549281427818174,0.06248753678833607,0.5605032738043927,
 0.27322843789312623,0.6787175955893217,0.4881835166110561,0.9957409877665683,
 0.9967162294916437,0.04495219873516909,0.07896391075564813,0.722985366242852,
 0.6429015157031885,0.715398236923873,0.7597681547665002,0.022956035686069298,
 0.2597054132983585,0.2720126644654568,0.4243166832053997,0.25708459357580904,
 0.7249770300321995,0.8442443882276152,0.595781000291811,0.7816735082975579,
 0.28106407471461137,0.5255782428734226,0.130013423908791,0.8022321676874228,
 0.6840299214782191,0.3576339753221066,0.27875088711617735,0.498197265163984,
 0.3847590452465083,0.8501263094958462,0.06954406077755082,0.8339622021897508,
 0.7380588928292152,0.7553945451465622,0.2985450727274328,0.7060113498992119,
 0.2593287641571135,0.07337965879932429,0.5332781373551806,0.6389209367017288,
 0.08555939563965631,0.7034197815641421,0.4466149807978258,0.6545875238808427,
 0.5948787435289699,0.5506218306859482,0.36164813019689024,
 0.22294133998733578,0.9087301950293148,0.7595685683884109,
 0.09740760848455188,0.8899504364231388,0.9082535514272024,
 0.09466897523999973,0.05695840117707718,0.3974073166582369,
 0.5156992162078463,0.24395405755188504,0.6095668135636326,
 0.06603110989818362,0.3770426452257326,0.23324426403133125,
 0.43188405067403013,0.679421267467897,0.5003467748432984,0.34022578093732603,
 0.4960717675457732,0.31399912506551675,0.05578912353290055,
 0.3799772091849366,0.807976972916227,0.8119236615048107,0.9223840037251061,
 0.25217867561889784,0.18119560499634668,0.12073170624426643,
 0.9697229661581643,0.01716487961600166,0.2842799579953953,0.8399931485173769,
 0.2721510159485392,0.6428690509978452,0.28309605350847455,0.9811487733904799,
 0.27330431603232785,0.0239411599379048,0.870866325371473,0.23641550775023668,
 0.6044511851775174,0.7690045347944082,0.23798420914884844,0.4185864806000943,
 0.20183646148539724,0.38034461724526747,0.6575954537196683,
 0.04613479640412865,0.4649915560554013,0.5106627236535086,0.7555466133034981,
 0.4022006651652529,0.7197310460982366,0.3966902619142928,0.1408262932786093,
 0.35262675148951295,0.6962497052655378,0.7389657035785281,
 0.15441340761859323,0.20198944947757935,0.5913232182372126,
 0.11027825394146307,0.722442220758363,0.011727057698441179,0.649249913968011,
 0.5500660122991609,0.9140601181133233,0.205235179670901,0.570931955359387,
 0.33302201115982266,0.6551118041250965,0.47128454028579536,
 0.8338778564844956,0.9065401390897494,0.7013556697041733,0.3566571476405389,
 0.9569753791461223,0.38073345008510784,0.5164375491328678,
 0.26368823176686806,0.9867898188183873,0.34597793426330015,
 0.9960817697391668,0.3016376096953931,0.4911833466857485,0.1566997158950736,
 0.8730906019223387,0.8179240726004097,0.894852388774218,0.041822290110659965,
 0.3988295468038955,0.467667379049308,0.4350737748057988,
 0.0075793298503167295,0.6343477240435461,0.011669021069913632,
 0.4717443116459832,0.0783561845368762,0.9788984911700317,0.2901305898488684,
 0.77489858238294,0.48658617224766565,0.03159913703410333,0.7844445474757417,
 0.24327603722571245,0.7112484524792764,0.37546977127378456,
 0.33040990511492907,0.09103652372040649,0.38302791343982834,
 0.7632379276761362,0.7845547871938571,0.8837639678851809,0.23466556083768353,
 0.02348791659021654,0.6790302991103216,0.11251795397236053,
 0.5900250683169077,0.9044372511871489,0.010964344888608868,
 0.4593077288796452,0.3626637282910513,0.718663198039201,0.12744205351532412,
 0.551692330597337,0.00753401376728402,0.7194399247611029,0.14045144158139877,
 0.942054700584934,0.7270061798701132,0.3584267105896922,0.2600216711706015,
 0.5777702761501811,0.6147945874546276,0.7844212491930928,0.2675053795035601,
 0.7965898910852521,0.7095685462216392,0.7159076598461229,0.17233599315487091,
 0.1319460736585828,0.8085400280283893,0.6791673832815753,0.5428933406986706,
 0.5774763258526656,0.3547929241414496,0.8488945145392689,0.356521307207192,
 0.20380813428055944,0.827623508957604,0.07240727722576568,0.8730210564141178,
 0.5775491724942385,0.2172839247846129,0.7514961242611607,0.8797503169483484,
 0.2551282411521095,0.8883404246569311,0.3268927778438022,0.3752407128608122,
 0.8535156889820532,0.444447422745642,0.8236756267208972,0.7981178741506157,
 0.17895275546388545,0.3775441972308847,0.30570108240355454,
 0.5373794239298882,0.3300971596536074,0.41351459364180554,0.375783261189623,
 0.38333835887721934,0.8156794012064923,0.06572948999715755,
 0.12860923741780217,0.6767449723016469,0.7958311980535836,0.881353567793339,
 0.9696747260050035,0.6051390999334363,0.5162468110279747,0.6399575827109947,
 0.342602895564021,0.5458751263225786,0.762015558960259,0.4182005722118225,
 0.7043911802155153,0.5980458525247191,0.7468748326923934,0.34743330774670467,
 0.3144625916826451,0.1352181134826741,0.5764929351969929,0.9288359697435402,
 0.25524852583180346,0.019495513958437072,0.7676178013648919,
 0.7790085628405679,0.8629724284291371,0.71989785319938,0.6730217745739183,
 0.9224278204888879,0.31412033372325876,0.06438681439564609,
 0.6678889474870501,0.30281309507321885,0.7127125951151941,
 0.15170992726524124,0.7739469160628887,0.7069187871707333,
 0.13953095331166665,0.16028492102368963,0.8230901649379861,
 0.27216860609036075,0.39655666832661474,0.453004364048445,
 0.08018135546293281,0.8533518575480924,0.12722311754923266,0.889905575242283,
 0.6616527585353313,0.30659593396609197,0.7204698041165501,0.1749599056386133,
 0.7230602102444279,0.22732076175780214,0.5908067671685817,0.6851865509314545,
 0.18334899336900468,0.8837266798519459,0.7725829415321249,0.44050588224915,
 0.17852735452931512,0.7058535535997494,0.20960180979667165,
 0.9326289013098794,0.7844163710522755,0.9052909644204625,0.9256723040937703,
 0.8992699587217425,0.036132934508468106,0.48301887559940726,
 0.8655629523834237,0.8761533063886224,0.7319327454823482,0.6567251680034997,
 0.17245588981362914,0.5663314788749416,0.04772478073276554,
 0.5184204409965122,0.6491634817770299,0.4120613168672207,0.9006471528253877,
 0.928744286483767,0.6538053310103422,0.9171726142021674,0.9358896672274704,
 0.3521999453065505,0.2844144370288866,0.8527474875047997,0.6656110558836059,
 0.198869067387758,0.8103969826586157,0.935102903957987,0.7360958902923028,
 0.015853779755684028,0.7442842636082588,0.9659560841179868,
 0.7492205298267858,0.09218972327726327,0.42384882830402437,
 0.1146379375994212,0.4825034615641176,0.1825935791558263,0.16106863583250342,
 0.5587469458439964,0.5718232934381322,0.12994578953560953,
 0.22850500688616615,0.27573004868962303,0.5612863180487478,0.928519366090377,
 0.9980143584028596,0.3787613432563337,0.3543468092239709,0.23499439762157337,
 0.6147178830852702,0.6195847578763342,0.09388706313947082,0.8852735879342835,
 0.265418067104338,0.576200599781112,0.10840282386698141,0.8334135247894601,
 0.08244374116420972,0.8534118102512601,0.3767048771369329,
 0.19652708554620757,0.996241242000992,0.004577269362090464,
 0.12480653880338766,0.9722126700566414,0.035639230335635985,
 0.002903151870461085,0.6992019996543062,0.5867303644605922,
 0.17239602675750665,0.3781433969481335,0.8643208923188049,
 0.19966519153920415,0.10727812171721396,0.7941246042940164,0.782614382364896,
 0.8052552231443837,0.33599269348266725,0.7869864190981513,0.6984074856019259,
 0.4227186739371329,0.6308806827028864,0.6744294986683601,0.1236744099245699,
 0.8627217242450964,0.0706401451041545,0.847718337547505,0.18519942010748147,
 0.789174947380992,0.729832021457137,0.31213858077326384,0.14227742002871246,
 0.6861829224193011,0.7339199495953785,0.0831200657394,0.09612341230671484,
 0.717799590870333,0.2621152561056369,0.31263181903560744,0.874633164882574,
 0.19958060563927338,0.5859427120487459,0.6453142852751834,0.9715871588887608,
 0.8269811731495541,0.3680856323463273,0.7395943566978059,0.20862879274856894,
 0.2695655494170285,0.6813588749539713,0.6750145489100777,
 0.047420644069946505,0.23117424666717046,0.6729384502345306,
 0.1633289994347925,0.9000492509983795,0.8283890721219576,0.43541661323027514,
 0.392494575783088,0.502190251936511,0.5798761905037154,0.3495683694748719,
 0.19877639652563295,0.27469213268702086,0.5016505229761857,
 0.08241857843979616,0.6356595966267571,0.1619068642156578,0.8127623544991799,
 0.5561713960285578,0.4697772583026303,0.5772789737716054,
 0.014300580234529625,0.4531677494113271,0.6204800016333684,
 0.8712446391156828,0.007102895679325871,0.2959491926454567,
 0.5112052842629432,0.1870991223769155,0.44512735622107447,
 0.09526562342180445,0.04700986498374815,0.4427893199865588,
 0.42647080115723746,0.8719124389834257,0.08020455969011064,
 0.4965446348555226,0.18203541113658794,0.33965996881223615,
 0.5258577488298072,0.4954472032610515,0.0127914430621614,0.4760631973890581,
 0.10084955227688464,0.04779712490280663,0.2112434146238703,
 0.7921716166995005,0.2621415645949322,0.8506104717929704,0.5138682721377971,
 0.28142503482407333,0.845909027965577,0.6780488675879408,0.1479530456384206,
 0.9987319268933367,0.5476063021949485,0.521272612787225,0.6497519003939826,
 0.9340183182190743,0.18367920267260285,0.6075533294390988,0.1302584518917026,
 0.807695952916902,0.02414756725513245,0.9453375411770959,0.7630231923329421,
 0.7799897366590375,0.8207654319484157,0.4208831332612082,0.3271588412190547,
 0.45658512846394705,0.9083627561474262,0.6473913361472913,
 0.33276591345292383,0.7663878437255703,0.924974887397936,
 0.061058630441592876,0.9720886154977881,0.6006762459997037,
 0.9233006398947065,0.5643223886571243,0.5487570767246646,0.5604351121985787,
 0.4819509449484094,0.5304504495247913,0.424871019297138,0.8757552713140455,
 0.24045188688393526,0.13587328308756286,0.7130824060604857,
 0.9066015413051032,0.469174535728217,0.44510544062869184,0.1139864214758719,
 0.9809675368415851,0.20443521958264532,0.6833044512461984,
 0.33555067174411235,0.9724067665308141,0.6571757997939749,0.8328916178918603,
 0.6854183540395914,0.03843356233557005,0.8531388534594879,
 0.09568539421940403,0.4046080036113954,0.21014845586213182,
 0.7943280447904785,0.5334109797539197,0.023218946723052136,
 0.3288121292330164,0.4218358492625167,0.8937402664724294,0.18730586395109905,
 0.6101931433777581,0.2810674089173234,0.9329634326137793,0.6940477325356784,
 0.2542363765757707,0.027263225324257845,0.701460090508789,0.6086201292206777,
 0.6079903155179868,0.4885175128788176,0.4007450613766532,0.05078651739835194,
 0.4398174697470074,0.7075205981356445,0.17038407260619326,0.3688137263659247,
 0.8254972432306223,0.6722750139497816,0.6062661535425602,0.15506079137032636,
 0.4467293785008235,0.8994151869611007,0.0400486675819649,0.5780739140984603,
 0.6210164532630443,0.9847920835482505,0.4874111061230384,0.16981139862995842,
 0.4465742769412211,0.7301913147390227,0.7596900655574046,0.3060975788899536,
 0.10445838890649783,0.5747748783952155,0.6608079508729673,0.2950794359246862,
 0.8495177966535205,0.12225409797421716,0.37217003507629476,
 0.3009576345474485,0.9926024235762564,0.2231964084292788,0.252891071345848,
 0.6043809851002354,0.6549662360522878,0.7485969954546903,0.8036996755162774,
 0.18227271501064257,0.36016332116531213,0.19882869609397769,
 0.21486611894999097,0.4973441419589787,0.15940226384831535,0.76303618490671,
 0.44148720987348467,0.42812800838717835,0.3596870248753692,
 0.8963902188452877,0.7511032158307093,0.5430680812853425,0.2665372226760293,
 0.8417879041192092,0.4781169097965495,0.2555784077451059,0.47095199767614104,
 0.037152709937046224,0.03670665314121535,0.0842437478578234,
 0.21913847217810078,0.03921407106887853,0.16124413123716108,
 0.23569845391105348,0.4186332917623995,0.03827508754599629,
 0.01279963977939702,0.9752730730330065,0.7703998147523253,0.5194341894240466,
 0.6548733063864769,0.9239848027712876,0.7699563087242054,0.8904804829862759,
 0.039434463084713656,0.4565752086818742,0.016716958688064443,
 0.06683896158458025,0.5024957610215925,0.23551507096055313,
 0.6277938611061576,0.42292276573947407,0.3461623132699745,0.6615359375264749,
 0.8801519660763014,0.21358299238859024,0.08806340978655391,
 0.9182869756247101,0.1403467110880039,0.544936654917928,
 0.0018143198173283803,0.9201896270017701,0.775949029630635,
 0.5993298963404852,0.22188786066313881,0.9792124068037946,
 0.23158762840472535,0.04312070066832063,0.47682681069758126,
 0.07023460653427782,0.03206840502118857,0.19414167102893043,
 0.8509901490700909,0.7479503808845889,0.8670318044736272,0.38085442206264786,
 0.2759646810740106,0.9184525484416438,0.06716623044190606,
 0.06218201423826375,0.32230689349278174,0.02455794698110314,
 0.024970446035847416,0.04804233529219526,0.6497354630835352,
 0.3141492793728592,0.6904731990085328,0.20595203436138598,
 0.23267321592700496,0.25170545287929125,0.2820172468455865,
 0.7474430417340725,0.9975459332851963,0.9636091230472699,0.05447539739973328,
 0.4648557194810674,0.9627681480701501,0.8417318268903596,0.9298122687144552,
 0.41635933437810846,0.5583180609697146,0.5918113081026342,0.2814139647963786,
 0.7201384909423162,0.8320102765143359,0.021319394074607922,
 0.06668919717912614,0.1906475247043643,0.4496183283251063,0.5523361851806833,
 0.5107393429155517,0.3354981133632736,0.37114179128765246,0.7869045609768329,
 0.2310709707067904,0.8703582442316411,0.9052745670690163,0.2599473241408061,
 0.04452724627145732,0.2666246571113161,0.946476128759475,0.5840123038377378,
 0.12354159525802966,0.783234842738411,0.15677630901728457,0.9164546680493098,
 0.48524128799055877,0.9155450188476868,0.45646120654286637,0.507578132830499,
 0.06311553045558238,0.36410667334900326,0.4800969121742983,
 0.3053750439600891,0.614071374894432,0.5719737769225112,0.8259276190866496,
 0.37785218893008454,0.25360263664539306,0.42031655752257435,
 0.05580643581319933,0.2443586622938988,0.16122488649950284,0.801050687171866,
 0.4143866581091944,0.03473065200023706,0.8154135391434536,0.6335171320319533,
 0.11187128991456474,0.6508210357064821,0.054263888039323915,
 0.5485374617041311,0.7918677390857782,0.6170797405531325,0.29304129212762775,
 0.8326390816141421,0.6609187108573378,0.3124867764494619,0.9519096106151784,
 0.9033118879071691,0.8863880142661094,0.20376324170711724,0.6113506924860077,
 0.1734697796516208,0.4330784592454082,0.4937370798930083,0.7457478534495147,
 0.5894893703758274,0.5275016976573663,0.02030887957120797,
 0.11899236007332037,0.625307901857675,0.7531974682374688,0.1467528222760125,
 0.9700476851097506,0.9456116364655353,0.06620258352158692,0.4698519956580418,
 0.9214049204699194,0.6004465576151803,0.7164191490545234,0.4584559749726346,
 0.673410996886391,0.7213478688379653,0.3992261284919518,0.5563045619637929,
 0.8971601002000126,0.1279214706316959,0.986220374195153,0.6724053084684255,
 0.8346952117405169,0.41929790105395814,0.9691157243285423,0.5975686158080961,
 0.758883068994305,0.7725594251777502,0.7305529374594628,0.10355358550350613,
 0.633658935076478,0.17580287366652647,0.47477769020828775,
 0.47122935046610503,0.46796693461502425,0.3450729487464864,
 0.5644073891872927,0.6946769609622476,0.9619923927265098,0.5507038724108252,
 0.019970730721169283,0.7871404280539626,0.34245660218737084,
 0.07885276488350135,0.2839217733642718,0.836248621131125,0.4531501692497816,
 0.26845657844989657,0.44733043689158003,0.9423750332015299,
 0.2823999189358749,0.5180495830312871,0.8791978978718593,0.5911850265059541,
 0.43282202629687494,0.25895170572940107,0.9880130516730571,
 0.8590839115099439,0.19298453870754084,0.0627404921106467,0.3317477884164015,
 0.5937099878065103,0.6740099085655895,0.026066162503425416,
 0.9798244718612085,0.2754197896256767,0.6587734833618372,0.25573614721666127,
 0.6269635730978058,0.25670095354534106,0.3310288233848735,
 0.18063437478511324,0.5283673348863204,0.6648860855582777,0.8620522574252059,
 0.41025133752949794,0.2878685738454526,0.7716458511938473,0.5727172352406151,
 0.09275482356038978,0.661827726763292,0.4740478349316166,0.2322120445715674,
 0.9079543705311803,0.3575105310786464,0.3794673456870272,0.8565873103211118,
 0.355412915652052,0.2915670769096138,0.008025012324067715,0.5368101315464181,
 0.36502900238724,0.6651171259248632,0.05781612668310887,0.19168147403666302,
 0.3565936930294419,0.8791811555159006,0.8339161568473129,0.03196747352166407,
 0.045905154082320765,0.744456279682527,0.9963220671700437,
 0.39843412856940486,0.1943822858212596,0.9179029961569494,0.7961485998861251,
 0.7187549111759589,0.9882908689495594,0.9317810760876686,0.3748741482349287,
 0.18062546842928473,0.6639588654424884,0.422051834528158,0.09122990438437917,
 0.6422615596104884,0.7223603538025403,0.12087609842694635,0.567544957218325,
 0.22582158105105976,0.6285189771556857,0.7226089256774265,0.9537573986840167,
 0.41620798442049134,0.5126605635373254,0.09551379467967747,0.249187102368159,
 0.35155266265915563,0.49651550823542356,0.8439803419126477,
 0.10779667291257078,0.6364272310022641,0.024151552748101435,
 0.4970592295941405,0.12563784821947288,0.5030141229317806,0.808548828688638,
 0.07151961426155062,0.14317737601231817,0.4163441447878783,
 0.8909160771185219,0.8194886746143784,0.4788639257620998,0.8093358104632238,
 0.9202752114478802,0.5743352843200399,0.15997059055992702,0.7708968533436934,
 0.23445813691060757,0.41555316929030206,0.8850899394020809,
 0.7674608300948715,0.9769205150189967,0.040268505635064566,
 0.26764437887009673,0.984708330529022,0.44115021472804605,
 0.25162355592585717,0.9904331249581406,0.7502001449191061,0.7689640641884207,
 0.5533221477238838,0.4584894524610358,0.17980080106664542,0.6671585816657373,
 0.20919762550594312,0.02700292318333486,0.7850210508651918,
 0.6970190991607118,0.14164861615604607,0.598841354493849,0.6401298202102172,
 0.29861521573495364,0.17143331625818548,0.9726746370821605,
 0.23877422028077733,0.8361211948605751,0.22291167042561166,
 0.8254295553846221,0.37847248385507093,0.847817339743165,0.9209609809952667,
 0.7506877632668802,0.903049822600358,0.7316878735838561,0.20029985255220117,
 0.6930197052357956,0.35244214598270995,0.12754291725376732,
 0.15059070285513776,0.6502084438182817,0.8986516906288621,0.3722969123330233,
 0.4597972392925238,0.1726254507807089,0.006611086737585925,
 0.22533565995139482,0.18607234879626144,0.4234836370648215,
 0.6733528514134203,0.0019341724393688864,0.033883207392371606,
 0.990916968283224,0.9779258215581219,0.5244282796918744,0.01913680194512446,
 0.9529456146610027,0.6788583595554047,0.2656209728814003,0.41841080901700056,
 0.8652135191420562,0.9147269449869047,0.7708328135870018,0.7832101771841734,
 0.2867041121328002,0.033809721913278246,0.39627950515733534,
 0.6805111843309934,0.7130133827752133,0.14344218043559176,0.1168356110597576,
 0.06567154540267439,0.12415941483628867,0.31167013334998905,
 0.8039786210481005,0.696225629517835,0.7323135992769706,0.39594186513176843,
 0.15478940499790028,0.8118074238053725,0.2621547110339688,0.6588095939929758,
 0.18232588632093016,0.1629577546172155,0.5878051487117834,0.8408533678666059,
 0.24163359107160387,0.49421130340498587,0.30954334844963605,
 0.20590089606190032,0.7475567696080346,0.8763772464798619,0.4545660368075708,
 0.5139628191763971,0.7115176620148922,0.5744357162256405,0.6829460696530094,
 0.31987475325694537,0.4245433877788879,0.944827709363824,0.8097147953252831,
 0.48435770067349837,0.34345678020377535,0.08221691582015389,
 0.8213266078855304,0.5862436144647868,0.8664218489669115,0.8400898870858411,
 0.07344926560599663,0.9766812233659637,0.5332578449925055,0.3240201419031059,
 0.08851853327914694,0.449196393756203,0.8490639529578027,0.18176993253563267,
 0.11107492211484227,0.12713055501317383,0.36765429284317996,
 0.5734512904280592,0.18216172307365375,0.11863715551174181,
 0.19486294138388938,0.7714781332652642,0.20218804123385992,
 0.6403371364736682,0.15022556434264955,0.2718531536345453,0.80121502939436,
 0.9292034193475474,0.797925255939415,0.8142611647477082,0.7825768735845648,
 0.961962355195502,0.04803445174490162,0.9477592781620277,0.08439475067581281,
 0.08739468665361716,0.8593984370829738,0.4728815367777699,
 0.32567544391601533,0.5805132757204862,0.5086411814364089,0.4709971183475493,
 0.2735818094370601,0.9173734751438221,0.8803065579005875,0.2622719461030797,
 0.02086941950913368,0.16513645115040443,0.18060254169641066,
 0.8766701943677034,0.7939947645185056,0.12519431533396252,0.5231664821011374,
 0.1420876041386292,0.26465136481883567,0.9841709860376477,0.3317187153234953,
 0.8100632641057541,0.9577026100677148,0.4042122126784594,0.4807724412788431,
 0.4119513982311359,0.1736204609755435,0.25165562182552126,0.5677777154056004,
 0.5969959604067352,0.9163865659479824,0.394673284415201,0.3615403526418097,
 0.564304132462514,0.26611184989226555,0.6804018838734434,0.18249928178655794,
 0.7396560422266044,0.5705190942831198,0.5313917337156816,0.4303424763578706,
 0.7306492771101185,0.20466309096063062,0.4895838980646663,0.8146589834853414,
 0.24445762343287425,0.4929383596775698,0.8493128380921777,0.9582194193814145,
 0.7386141587720865,0.12847098410106694,0.7496998476983098,0.6851713527816133,
 0.7455147943093444,0.9638012130045359,0.11756803695007201,
 0.39959816364814327,0.9094053046628687,0.3555455517002426,0.7742766705414437,
 0.5553712335023925,0.4371532712339363,0.10667179567079521,
 0.24317365904637733,0.241345212816095,0.6448694942599198,0.25025685320998925,
 0.1668226269846832,0.12654486113604935,0.608923264474736,0.5856719975967665,
 0.714039750875288,0.45851429547377387,0.05676687294389604,0.635192733365177,
 0.8015959117078323,0.29187337187447016,0.9026979088320302,0.2737077928033156,
 0.2532547352911798,0.17129569997717664,0.2911566256401318,0.3469852823943397,
 0.18361929544428202,0.39281820319104277,0.46103112864815954,
 0.6414889713544639,0.6687019520674029,0.3630150635730448,0.9774793965168173,
 0.9354075583190468,0.9077031266297819,0.01795717730761215,0.5134763841461794,
 0.5074996547026789,0.42944014913527195,0.6161775960789427,
 0.07968894173159258,0.28339452479664873,0.7693447595279387,
 0.5154070023073325,0.8870196880926163,0.7223826635292416,0.8137927254610365,
 0.3198655278368102,0.5508465260540711,0.9982346150583717,0.5085126837463615,
 0.6102878701688457,0.8828444696985225,0.45877242623734926,0.5834836124778688,
 0.530302847870087,0.353763574842637,0.44554642716785287,0.3783256031652953,
 0.5146850003061247,0.7253923273428711,0.8680836757291523,0.4943208034069523,
 0.8517808965958708,0.5753123255011181,0.7446088114217913,0.5233666801879937,
 0.5113323472356045,0.8626861381709303,0.7131957244540376,0.22479363962879317,
 0.06546801104683153,0.11761075479072192,0.12780135706726825,
 0.8195903902280169,0.9173283774806854,0.45836958669953665,0.1762062475523659,
 0.07657381305555955,0.24991520564115244,0.22077775610826722,
 0.8614139517285335,0.6384299027423869,0.34678228273078093,0.4502373386427544,
 0.2246954974038311,0.2622622710551723,0.6113338706857749,0.3761830336133436,
 0.7976840763309784,0.7740428161400495,0.06639987196809571,0.4541215258423026,
 0.39394847196008553,0.16038773502986625,0.6671477230801608,
 0.39434084359608756,0.04932689323155692,0.47968674100590103,0.77126303133486,
 0.8840922932928588,0.8384341436940683,0.904435999861171,0.3982315193515086,
 0.19477703435974358,0.42476861223715534,0.49786625932213513,
 0.07661712610344662,0.6961643072090888,0.5023801191714781,0.7153157102661174,
 0.4361059672751736,0.7030874842225037,0.6235285574683931,0.16426562105695452,
 0.40918178165603136,0.2812458556158054,0.5420827863462135,0.6518364486249867,
 0.5033981094280136,0.3070651475102615,0.9861495848223816,0.19528302266099085,
 0.2974593574664325,0.5933337987030753,0.17902592281417384,0.9110508620704018,
 0.44238699589747177,0.9669587814384546,0.4120424442257392,0.4044259529894715,
 0.5055930573081134,0.8254943496613324,0.20649646774716635,0.4163719076724136,
 0.0022165176366093498,0.48072125215401296,0.9285653320757454,
 0.8521352859089517,0.260890670726887,0.9061806981630506,0.6983996953983135,
 0.10161591881365573,0.7044617033858654,0.9614810949514006,0.6363481145002402,
 0.8306572432444644,0.34369693917904565,0.41126958752508713,
 0.4118525865774939,0.7714487859891643,0.2681134150434158,0.6268290148767142,
 0.9228640932133593,0.3393020055891929,0.0567765960404244,0.2581844170745101,
 0.12237914819552986,0.40788041070345016,0.5282586750522901,
 0.4223827846596766,0.5547995287983125,0.672574353933896,0.704893834717511,
 0.7940237007265072,0.47744364649292725,0.0260784925712918,0.1431166839550907,
 0.25400852404324525,0.5847445841246941,0.8584479540074399,0.4685814517725102,
 0.5907260349086503,0.35147499323098397,0.13964954494042192,
 0.28641524160349474,0.13064104770628493,0.7886329692794466,
 0.7500081019878988,0.010757736284632058,0.6977193431854171,
 0.36322500738700436,0.8855910097280655,0.10789686489975192,
 0.9328521346972647,0.41698861254233754,0.6223774541705405,0.8981431705755255,
 0.09312591633366918,0.49588596054348655,0.977217494020631,
 0.04478509668128572,0.9444691462705197,0.32382071681869806,
 0.8109047633849367,0.10423440654987326,0.7078967754851311,
 0.30017711070076025,0.10884663188088983,0.8906453964633899,
 0.6651270904929437,0.9449479340907728,0.9281501816151774,0.8384481120492503,
 0.28903689175072667,0.043077638113637184,0.3848512754742798,
 0.6040520164359666,0.4087919862219258,0.7773672519790845,0.5555080521008154,
 0.17704560846786443,0.4100478418088602,0.8135350141924379,0.9806584114260861,
 0.35072064769038924,0.2582230723619121,0.3468692318790578,
 0.13701196392015869,0.868665724776612,0.7574445288432162,0.4540105025998026,
 0.14799400901473603,0.22402402788499798,0.3936018870158108,
 0.5769315278013882,0.7345359324661839,0.8257681643894825,0.6339292196554019,
 0.293675084280988,0.15207573391614992,0.9688309756658874,0.7549224676976904,
 0.4619407113112548,0.24383650748046737,0.0946597934442337,0.913710987286545,
 0.22526097877139595,0.13917559306488125,0.7164878138155018,0.737545354155781,
 0.794583531964734,0.30721962824775784,0.68470256280304,0.6721520876182144,
 0.07951683794235531,0.08852708329907633,0.986338496077431,0.1249443878924078,
 0.28205836499758663,0.12042136117974267,0.31703446145887615,
 0.2077514317107847,0.42398000612387365,0.5450782889480758,0.5292559599808007,
 0.541136232473197,0.37376925133081274,0.43019607768122325,0.508048143239204,
 0.6922838208260591,0.5580225338558146,0.005469946553546934,
 0.8916574644905384,0.9830309008347785,0.16409004372356095,
 0.31987093002905354,0.8798119019533095,0.7814769362819272,
 0.043023114531097306,0.8492385313968933,0.739844206970846,0.3952173291437434,
 0.40634746036371183,0.608759950761273,0.43983689948808113,0.9300548906949716,
 0.4990214267579055,0.7242137183445724,0.6313502389699894,0.40709625429065366,
 0.5504151159363678,0.5417495782037078,0.3214128882301859,0.07151709538003626,
 0.6049325120912525,0.44904391492334905,0.12596884786836315,0.487868587927468,
 0.04022527472952875,0.20323631830229227,0.15844388764472772,
 0.03996856171846774,0.6368591430247693,0.5430225747709019,0.8263768658492989,
 0.5809278961874837,0.5124809611910818,0.7664290323037155,0.15856198323916704,
 0.37581608055256965,0.3495929331318107,0.6952782862761957,
 0.10059453991384748,0.7995388956760531,0.09765155019516814,
 0.10988248296439274,0.5344993117208455,0.8238053561076663,0.5956983618632307,
 0.6579286858347421,0.9956784790056573,0.02071254758159269,0.5045315390500716,
 0.3244515470409991,0.1593987237461867,0.059695831856417914,
 0.6663662830380183,0.44735467285078456,0.5613916094238722,
 0.21523479856967054,0.205160942696879,0.414747512688978,0.7199612800566764,
 0.13620968126660382,0.44399573912944945,0.3695391756134956,
 0.4355761308355758,0.44535742034988457,0.3153533328075351,0.4562271542294758,
 0.4607366787061592,0.8735599489288082,0.712645126061494,0.22312028843696252,
 0.3595671705666659,0.5797970344938078,0.12785702730210013,
 0.08958170009569166,0.04867727665254462,0.7738858327749412,
 0.6760983163627378,0.31529349241634974,0.2066820209497784,
 0.26008245305981437,0.8861777534699722,0.9553869598658554,
 0.41909684351371634,0.7053214527684448,0.09064303030018805,
 0.7969123825712144,0.00463373412372603,0.628395598349565,0.9306896122806828,
 0.8575773464773421,0.12739023662500415,0.30131143364350277,
 0.8031728274472644,0.3313571416145178,0.3226076470125784,0.86917537145467,
 0.004629666859860548,0.5492024668198131,0.2766996598830188,
 0.40623162043154437,0.6479580240569669,0.9997562409768084,0.5650644350896343,
 0.008933160803378803,0.21664162969275158,0.5191977956557987,
 0.31978025168671464,0.25310972003452814,0.39286247179092315,
 0.6257080948110502,0.48625192007297335,0.5690837010636104,
 0.17945507564342278,0.22581367631355587,0.6776844183769312,
 0.6332973853912947,0.18267151835399242,0.6645087110648108,
 0.09474396458933032,0.30306770521711546,0.1047792364083111,
 0.6401336841859085,0.8239384528898861,0.2706416504627751,0.39831120511821805,
 0.059113620177379866,0.2729396426166525,0.3369966708479717,
 0.8223033623731869,0.10727909232761201,0.448057717676527,
 0.053506023780259815,0.09946211277746941,0.5362101506147384,
 0.40519777254551637,0.8833148688878912,0.7430086525742444,0.8102512564876341,
 0.3043161596283581,0.948719029672174,0.333463987526988,0.9378865506802612,
 0.04721559952417964,0.09551362367446337,0.7521623536546472,
 0.07655220914936411,0.36455747104864555,0.12537520467041852,
 0.18844465340681016,0.48939731266665765,0.004229285379159453,
 0.9509318420764621,0.9588495270524758,0.6902437001055249,0.6671833950484498,
 0.928450164930538,0.6549046243783473,0.900257931911059,0.11437117025094545,
 0.715658187374982,0.9873566679738458,0.9169787199030957,0.9700288850045935,
 0.19905721131568865,0.705931371845922,0.10761524647565612,
 0.013016032933436339,0.7795728484619926,0.2831134498486031,
 0.4804075271114143,0.8958103560154543,0.45719654590077485,0.1965404072538487,
 0.245058510884819,0.7842451795326003,0.08321906300123338,0.3677802075090022,
 0.44901542432072616,0.14418215360761133,0.40740059791714145,
 0.21962442548294248,0.07285044962488296,0.23236634315611993,
 0.09923007624543645,0.06251413995291233,0.929220050597743,0.9726286160348668,
 0.8148139828641191,0.6738766001448102,0.16903586423984895,0.9123793798523304,
 0.5894684070285894,0.2716034833652601,0.450610333325547,0.8748591389372073,
 0.5930853016106729,0.30099588096455987,0.3351300330249031,0.5910404039195463,
 0.09925646672361021,0.28906905290431073,0.018564927687187538,
 0.13014264250672447,0.6217113379243181,0.05808911051595034,
 0.2984676160326778,0.38673945001326426,0.8813852817286078,0.3637976987689746,
 0.7959099472553725,0.7337965067912294,0.6184148448516534,0.473566032442557,
 0.944946990793201,0.612217463632728,0.48807932131777865,0.05548477266258933,
 0.7095571589817372,0.7732342673189159,0.705718964391792,0.11618355936756819,
 0.6257846587050863,0.29606448460742774,0.4221698608346349,0.8460416187569305,
 0.14027163154814015,0.3592258099409755,0.6717536754856017,
 0.016529767736061518,0.38442366097398406,0.17423590331533367,
 0.04546946047094447,0.6510960702302966,0.6286419741140995,0.8367074325850561,
 0.19919492258863059,0.7449768245997851,0.8292760788423416,0.4655897297545797,
 0.2403558035357144,0.13056302100625772,0.7546352367445152,0.3814063950271659,
 0.7166374497621736,0.946703338744741,0.0652536076901995,0.07984752614987489,
 0.23401061445998206,0.8457159357768389,0.8694657543870711,0.5844407112485193,
 0.0909576371138221,0.49688444608146254,0.5919139620970552,
 0.08776252402714402,0.14882432568810755,0.7441440030515718,0.916887970429835,
 0.9106016528710255,0.9735095609871451,0.09747622345359519,0.6563544103543029,
 0.8933150603463518,0.5809249968373077,0.8050010789305797,0.9696197169661925,
 0.6487015558201678,0.9590755566732907,0.37215417369083215,0.5279933155856795,
 0.6216250465958767,0.9160906656166635,0.8422460242999809,0.10323765881653901,
 0.29889305248219766,0.6678227692088922,0.4455281838632603,0.8705532494557343,
 0.9034068962819575,0.09368566068936346,0.7951430778082538,0.654742478829352,
 0.8072967887509349,0.48202538525471794,0.9112080833943823,0.7405757041381268,
 0.44340210161810567,0.2830562207722749,0.8135386066133021,0.9505944872506948,
 0.9883124938090264,0.7575760584228762,0.8236515153193635,0.4540062090142023,
 0.8250598263531984,0.9906146103362964,0.11850961811330529,
 0.28198543936033493,0.5189291432651952,0.7481210474512661,0.6810917020112508,
 0.5708675009639597,0.05005709937169023,0.32665341531725556,
 0.2493580322758132,0.21596219498208724,0.8989808258350092,
 0.043616051605211315,0.3457608874073669,0.09340540363848326,
 0.18174650360444522,0.33788599411754094,0.2943928866530151,0.949904915088909,
 0.12881347076571803,0.25047549086494747,0.6760298087140948,
 0.6430969453939093,0.8791996941688563,0.04580726991170414,
 0.33947204060035563,0.3040984468673259,0.14030991861148445,
 0.05881890852623928,0.9183970449519195,0.8846362149618462,
 0.033198719622545725,0.9855558559299672,0.7725994039868079,
 0.08261071800040287,0.3132276913898444,0.587314483468347,0.9743454694361526,
 0.14953159690239082,0.8962935078411045,0.019688988993586065,
 0.705342464189233,0.49680213570881704,0.6399600154509447,0.2370927174657168,
 0.7997505746556341,0.713117429703374,0.6154954128459867,0.02160925296078886,
 0.3380708938446252,0.8003270049134337,0.8251141642441739,0.6054764613355932,
 0.055300358599325694,0.303370035903628,0.46358634982929403,
 0.14512502239445502,0.2526354280832763,0.9730059449316535,0.9819047152928926,
 0.8555750868409994,0.09837473471147595,0.8015626724657421,0.9300109611812163,
 0.47544322626830215,0.8179386024091396,0.9428091045998811,0.056042362096417,
 0.21262284106748963,0.46776348789414635,0.8309593282623211,0.667453115079269,
 0.8607907120617518,0.30563403368544795,0.35607136364669467,
 0.06583839053588847,0.5820722980140292,0.7419549607147332,0.2590229259663033,
 0.5646949770553389,0.7357333304605607,0.5660927515198783,0.9691443658056011,
 0.1457251872242945,0.49711450507070976,0.09157028184042659,
 0.9303644743994528,0.5104872966002696,0.27858422850726694,
 0.40937520019527074,0.30900048257355595,0.590712313032516,0.7132478649954453,
 0.07085003125690525,0.8298017829587845,0.5571843705643178,0.6569371582071739,
 0.4837572714248215,0.7836791988963115,0.47161856527676393,0.9150210382238279,
 0.3924637317897868,0.019810129002204357,0.3705766768620483,
 0.12734396628819322,0.6127747284527956,0.7614064329505256,0.7667333290225642,
 0.9111086594070792,0.680108186376362,0.6309995584202996,0.16174070293547704,
 0.02335470795748451,0.385951169725292,0.42723868726834646,
 0.16362185693340225,0.620039973147371,0.046464890824946425,
 0.9303385356224715,0.30054414884899017,0.09928704586482495,
 0.9335567392693329,0.33740050774563424,0.8800226466815564,0.5870376192396355,
 0.347677418299142,0.33283797703508844,0.9758318670991992,0.9063182447646068,
 0.22265065794999506,0.6188954736337307,0.558060804798311,0.8283830308619422,
 0.3669159666442863,0.9227397599802467,0.4603627238041048,0.46685841888152957,
 0.932334751602578,0.46806345503649416,0.4362803415409704,0.8419792030073499,
 0.5731198792936921,0.8997452425370152,0.9763610801470476,0.2609473352255435,
 0.5452975457750762,0.889745376064241,0.74389176820944,0.3041581886652337,
 0.42966521217371567,0.7843209758778585,0.6975005207570485,
 0.25507718641672794,0.30433071999976136,0.8341958924699876,
 0.8859746996510941,0.9671494576245201,0.4968679735040864,0.8869283594122483,
 0.8845069036540749,0.5018122037362748,0.8552058221913944,0.8019249316223527,
 0.5805633317898871,0.7307473682730259,0.690264620558065,0.7869856948095224,
 0.6426597102393325,0.9299266042597851,0.07779351927986788,0.2608170678836581,
 0.831076840473672,0.8998056834466409,0.7811304506804022,0.4281288234224234,
 0.9305137805175406,0.8096195560275273,0.5665858977408116,0.2930478970022139,
 0.32233867256260584,0.3645630900827863,0.06937212905600298,
 0.3587716767430157,0.5526579843526485,0.8994165358480315,0.6063865992440037,
 0.5399621097385668,0.13380718964883398,0.2121734047273044,
 0.20501221390184987,0.4728632110710521,0.7636008123092425,
 0.42179360095793017,0.14912974649147448,0.36549360544343956,
 0.4022204420531227,0.9250444490771217,0.44485448792866045,0.9692598498994142,
 0.20094326987086608,0.9798339054394467,0.31912090073608745,
 0.40960432419838166,0.563334758133835,0.5204677232152702,0.3731261687497349,
 0.26313728513477674,0.9810659172005248,0.5479620264016153,
 0.44869553489739755,0.5199714266491919,0.14703228988703865,
 0.6987912524124028,0.1009458685803587,0.9299481701297553,0.6606643086725152,
 0.8629337405484502,0.7737166075416095,0.7368384864133087,0.8808271263898315,
 0.840789150334976,0.25876123425698183,0.5130516027031332,0.6458791174064505,
 0.903154271093464,0.9944462652472056,0.19517427805824095,0.18332412446528423,
 0.177984701990632,0.8621562352115889,0.04308621928824996,0.31641803579170136,
 0.09061915287320899,0.7902888922550595,0.09385629556586772,
 0.10154740650520644,0.14542670188613904,0.3796502866573689,
 0.7136737153309665,0.2087840437347942,0.7683645230401458,0.9631517972900205,
 0.32252654318554674,0.816932012715035,0.8062809922674027,0.9965749757509323,
 0.33769921561405647,0.8316402026319364,0.803132106206148,0.8098391073880756,
 0.280470229341491,0.9824228561390451,0.26512522856837717,0.22608746411797753,
 0.5675536080406144,0.44776289896572985,0.9717913167912359,
 0.02934461563123414,0.010247555320321289,0.8719202395890355,
 0.2722022323366964,0.7712564889000465,0.9008812744651831,0.601197825650235,
 0.685966390896076,0.9875830048964241,0.7470763113268284,0.912908080145048,
 0.44480939371187755,0.7042808645385602,0.4589001209303739,0.9383162998884313,
 0.5877102762592202,0.3078174313287785,0.12414139090532128,0.7941855602363979,
 0.29321317583601725,0.6755610321173589,0.44469547221253025,
 0.5469944297101016,0.500045788248508,0.7950126938385703,0.2257756711157225,
 0.3261113014820758,0.04856982826524736,0.2395798638079517,
 0.18382884561843837,0.9397552490002924,0.13065065223148986,
 0.2634632070254139,0.9221904015021747,0.048278885303481234,
 0.13184073452576284,0.74396214090036,0.7447562897347865,0.8839697679336762,
 0.6668791254781231,0.8869282501966639,0.016440157280086187,
 0.09144043698662063,0.20149001780999853,0.38268350193834566,
 0.9454266414872742,0.1942103610119973,0.6719866772436326,0.9590431620534355,
 0.02780265870339127,0.298728193789346,0.7983565705114772,0.2433903613014129,
 0.4956867385102227,0.18439976723156393,0.8420326562765099,0.6030172457082115,
 0.30565893024427215,0.6097343223320503,0.4032857439465808,
 0.12543295408316923,0.9789480932332626,0.25435961380932914,
 0.03015165562130362,0.018848237929952272,0.15821823509300104,
 0.728893062412966,0.2671927230655071,0.21931649016087407,0.6134962243583647,
 0.8376013105348564,0.20966880039942037,0.20617215576806913,
 0.26230529439073824,0.45720936313969407,0.09128587906454133,
 0.9439587697878218,0.5968527378952788,0.42508534228861505,0.8102092908656003,
 0.7676940071084212,0.8146174471581772,0.7034681429104996,0.5892161505050431,
 0.17148613639780774,0.10308959921878269,0.7539846732114254,0.985437400998329,
 0.9003505096821005,0.8776243990244297,0.12082206582892585,0.9597992154629201,
 0.731439418046751,0.1329591090265887,0.05267854039558206,0.9035518322029297,
 0.7119892287361699,0.6890154034798235,0.28031072963898485,0.3241243630853151,
 0.96236549455996,0.8059864059608052,0.8611145051331797,0.5212248366239206,
 0.8244445517209493,0.16510306398534424,0.4244308081983631,0.9818673886312442,
 0.2165010496084956,0.703591145835059,0.5580747120474817,0.6692386101307097,
 0.4682349386815009,0.6248427444530614,0.5069502533703786,0.7089520436757207,
 0.6000949647577503,0.098247222817468,0.8138667631080689,0.5679131915885272,
 0.031532148205442256,0.2849909497645958,0.7667425882824032,
 0.45713486716179874,0.3072563035941551,0.4243749215917899,0.7568750799260696,
 0.08650830969310808,0.5980575809708941,0.3082088525191835,0.5031130823134099,
 0.7047499833467243,0.38363808252782183,0.5920392463816749,0.5484692281354792,
 0.046618202240866435,0.001670428670609625,0.14846888500480682,
 0.7709948165008648,0.8485427642460145,0.4662743964438649,0.6349622569303887,
 0.4248600993225694,0.1983838325939472,0.8246440457244593,0.17093585269761602,
 0.4444408556700066,0.899526900545608,0.7488458741390536,0.5398886202824674,
 0.23003722137772542,0.9613053939291981,0.2402607253262493,0.6837895452224129,
 0.7284874671990893,0.6146323409127277,0.04454681593879384,0.7115905540486415,
 0.8796235924690523,0.32373534628757994,0.9655439008299241,
 0.16769005207606313,0.2738652931373239,0.6789753411616518,
 0.09953514624291904,0.08346632798140441,0.37611173960727484,
 0.5502933088340436,0.3910560260025808,0.4534244743644479,0.26226739906055985,
 0.45381022234982926,0.30450667601814485,0.7018763742407944,
 0.6921752323056927,0.15496274815863664,0.02923943443222088,
 0.23733485249467434,0.6276239804550187,0.22214291930857955,
 0.10334956300679499,0.7819810639365029,0.6610550565375984,0.4537648729876519,
 0.38518291823375184,0.4338819908568571,0.28425110581604074,
 0.07741331617686487,0.43111423388693637,0.9611310328253578,
 0.2615967040118571,0.8909033777825445,0.05348252321000824,0.7096574783598288,
 0.1722762553125322,0.39130339031752803,0.8143884870838867,0.6091875185861244,
 0.39905230889534193,0.31836543259958894,0.4813651366106608,
 0.9889059415043661,0.5446246055519119,0.5480043675636583,0.87192295683148,
 0.5806506231521119,0.9043916265092363,0.9119455283506996,
 0.034137016909113194,0.10120670096388462,0.9554076139059122,
 0.8159209009412054,0.08559319725800496,0.7343069754274548,0.7228646022747709,
 0.8194353148163174,0.5680709207392757,0.3505239031250553,0.0644152383506692,
 0.740518784441663,0.891324236785717,0.9498262361950498,0.4492781410855802,
 0.10112094958390694,0.7708392247218342,0.6011425457586305,0.6126749618765589,
 0.4574124048814303,0.0519379507870269,0.38323506234491156,0.2343996565684462,
 0.09880423302861308,0.7291155036664236,0.13387744045017835,
 0.44703285149879113,0.8172288446960757,0.8220682696919681,0.610225684413134,
 0.9468955702248124,0.9656410021847187,0.6923707454910814,0.3651329831989618,
 0.9398182353945796,0.3805019210294802,0.8453717005640808,0.07010988466006696,
 0.6198168921908558,0.07613313030997026,0.7874273858020641,0.6721004054461756,
 0.6146718185784731,0.8737236414256859,0.4485105267361459,
 0.031369142744662404,0.027924385934018914,0.8620296795958544,
 0.23646620111404903,0.7341679128277828,0.3810353798692515,0.8094516238221651,
 0.7213670095627164,0.6824876756198951,0.24105878861534247,0.2898225952126938,
 0.7305958551857059,0.43307655869996087,0.8774526308777428,
 0.18655858539830839,0.007792827802353353,0.027034813568009297,
 0.991885363464641,0.9621181728449715,0.970461753727808,0.9758039142575314,
 0.16523344713051702,0.3800839644148115,0.17734885711636117,
 0.5814936453655063,0.9908942075856185,0.706408688755001,0.8130365103815336,
 0.8620001513656167,0.36634265697801394,0.8097027777752214,0.7622116116360866,
 0.6035385008438301,0.001090753099045827,0.41984571829670625,
 0.14527975664843873,0.15042943595347325,0.19323358042308625,
 0.29444248350899394,0.6328412928083659,0.5009931307448449,0.2608675211253104,
 0.2184965564178668,0.3938613271529756,0.3190361355481173,0.3033962983881535,
 0.2080136932950234,0.7503339771115315,0.7155138897280985,0.4567986746787648,
 0.3344283106176862,0.264043625914814,0.25496619597182457,0.2604332797691471,
 0.257039117675518,0.7734591063398302,0.47192557134164115,
 0.018185935041753032,0.3503438808830891,0.8430034753049526,
 0.5885941307215085,0.9646284945997432,0.21948876277377338,
 0.025091130611964774,0.03125364917144313,0.8012932223996587,
 0.05602384968749041,0.14168027693505603,0.5608657278665391,
 0.7165213189843445,0.5509463465460305,0.5081803925141918,0.9895374100763136,
 0.24773417097534345,0.44932942571856904,0.9321973024500145,
 0.1510138501215097,0.2872197965979044,0.37825733175249465,
 0.48507998769125216,0.02373144984917186,0.3409454844892369,
 0.6086687032659301,0.2733029630580802,0.1788226528474457,0.9798802288267436,
 0.5861862438249454,0.6836973668380651,0.5221748080844877,0.07084571996460276,
 0.07555322089317162,0.3167969042213621,0.2581823378382878,0.3996881872066521,
 0.8916424899685658,0.17794949305629493,0.6283831841937467,0.5918083784190088,
 0.2875714095552928,0.44286956583285997,0.09647503558697723,
 0.2591594591512433,0.4715044124819817,0.9081661758694685,0.7940739476164765,
 0.6357750959773089,0.9440793637602756,0.753004803211044,0.7382369615238636,
 0.015104332029348688,0.2119083903246307,0.35650472362437546,
 0.8771032347627506,0.7369336896696347,0.9818242909936088,0.8586147817677844,
 0.5146004322280845,0.38235004652564186,0.6185457968047573,0.449614307904597,
 0.7045386985944568,0.758326516281352,0.22994750270106257,0.17282280986449272,
 0.9446892422645009,0.3819361861462063,0.560604520997748,0.7602237384462873,
 0.24253136693984756,0.430002494557547,0.773781842345262,0.5091411812284672,
 0.7546025686402218,0.8069755491746562,0.8536206503406999,0.7665801776507196,
 0.11785339906242709,0.8224312132516469,0.8907578027275048,0.4003051175938559,
 0.08894214813840717,0.17593003554342546,0.5590637304164612,
 0.22703189839740723,0.5465619163668702,0.41547645970550307,
 0.38451767365154355,0.4725665712318625,0.4997125167868697,0.7890838104820183,
 0.16221140485169583,0.059951747483119266,0.07853602809535998,
 0.9881862941985601,0.09496910614612164,0.13231636632652266,
 0.5380960903494884,0.5187503815757765,0.6526304538902323,0.7752441773969752,
 0.7762664498734021,0.35205136103462764,0.9535709471451104,0.7064099683112018,
 0.5775148276813086,0.37271453943316757,0.33110411817494156,
 0.6590889154161693,0.06903160603368685,0.5385903362917401,0.0042562455920373,
 0.9989978338061317,0.9339878023212602,0.3795035808130237,0.25574422424729015,
 0.25707118353559577,0.8924526183398332,0.24053196699390877,
 0.5416434342479161,0.8830093047965144,0.6796437570860481,0.6972696392475861,
 0.5982887161801711,0.9273310443305718,0.795217818574965,0.9200905123601939,
 0.14724314996003995,0.17966195622835135,0.6865000008378148,
 0.0880456772619046,0.3434001519519181,0.0435800137743918,0.6524935039032898,
 0.7594001430039974,0.060960512532844735,0.5680217071862365,
 0.06549306119613285,0.6384533907577603,0.3902409621620435,0.6238956728089085,
 0.14031694972135555,0.05155796498838261,0.43986331593483585,
 0.7303363310319368,0.8556308647001758,0.4317454870452514,0.31167179448300886,
 0.06118262447846068,0.30506196789093554,0.5833630346405922,
 0.8168253544092908,0.32706662201962,0.732059647456117,0.3189536704197654,
 0.47162888742255726,0.49623232373735593,0.3350138822567267,
 0.7097448832428169,0.9717569280656521,0.5069842980998068,0.12679734368627193,
 0.809737037636233,0.30810270857234756,0.530827558596831,0.7961613609159093,
 0.05445902363478672,0.2532936417385364,0.9960097935538124,0.747466941942792,
 0.8368104632482423,0.20585336255103615,0.3185544477670563,0.8394675429494367,
 0.9481638184307254,0.11672831244294002,0.25949318760675966,0.811254427111066,
 0.7586295284257569,0.7021089088910679,0.15042999129546963,0.9765990322251599,
 0.5550387206476135,0.10474631492698305,0.4338975097446047,
 0.37722770559708774,0.03353098847103908,0.6701742451474231,
 0.07685060360804341,0.8384807084811642,0.5509866965647346,0.7246103141438297,
 0.9958741183583996,0.08173760078972037,0.7387059094640288,0.6739740454970565,
 0.5503968259923996,0.775907843821856,0.4108034295119978,0.6982008113047158,
 0.19667954474004046,0.8251059564264059,0.062448534818208,
 0.009545836847589473,0.793827306540114,0.5167598753334932,
 0.07250548828451864,0.7205661591516478,0.7782598720703895,0.5081838059556489,
 0.47093836535721945,0.6188277542431547,0.8603295448097634,0.5051086800069382,
 0.8791450528941842,0.08973958450227926,0.4506184253182597,0.8923604756123735,
 0.258092353863527,0.9765335080643839,0.30181012614785996,0.693821572417748,
 0.2488293887453702,0.6549654424747302,0.0754431481678276,0.45373219765476347,
 0.49227616504402305,0.6990244213627997,0.019271715499575293,
 0.002250819525330905,0.22487432948241626,0.657272115944385,
 0.01205802920548038,0.11740966011396103,0.8302913207690062,0.793646571757116,
 0.19132753073107378,0.8406405184380011,0.17741611064062734,
 0.009460704180934587,0.8503014785105159,0.18945769256368183,
 0.03603149476500833,0.7161259213525332,0.3909738277093271,
 0.06946983938438933,0.7662370515742248,0.8631470595787692,0.423596939559848,
 0.6018778987574965,0.3992397556039663,0.14088061031232402,0.3786706769974679,
 0.6567360670804461,0.8555619443885926,0.9180714321827764,0.2758136142419292,
 0.578108234799191,0.7755346177567931,0.6545467401362951,0.012244576311487787,
 0.6154380613047367,0.34179945864589534,0.29728370347447863,
 0.9583007404920328,0.6737080206438136,0.4075389781199362,0.8108284760755812,
 0.7598121741545866,0.9488651478343609,0.7684585304245891,0.1365656350058635,
 0.6364790714979516,0.09074986843130528,0.8002884578314492,0.8934356074192493,
 0.5509824899301267,0.4200965254245803,0.2951867477982949,0.9646994725234738,
 0.32431441668763217,0.33852603826484406,0.6647720992859445,
 0.5131530330424837,0.8132378771513451,0.1002372374075734,0.9850055719291646,
 0.8037652818940749,0.46871726565256977,0.48875268205360267,
 0.4500503053800504,0.9004521540432955,0.491541047777459,0.31957566257462755,
 0.1896459683048063,0.15085586358627312,0.30520437728164074,
 0.5226305441373891,0.34605139384512185,0.18767978965871368,
 0.9081783285590067,0.4972218244651937,0.06504927996521237,
 0.06495311869954301,0.4694219172007067,0.06145649215826876,
 0.06758534300814334,0.7624818651451741,0.05278060904922144,
 0.3364297937955214,0.8848579089303251,0.7571128224743757,0.20818085790689922,
 0.2212245134431825,0.19511228670228908,0.018175735768584,0.7212353998669981,
 0.5895875240966221,0.24027507395243275,0.8202872693768016,
 0.22926027109653702,0.219207839799619,0.28553780098014636,0.8631143337646451,
 0.6321291871822768,0.5333679381829894,0.2135682133884954,0.01748536717366478,
 0.41043472701501127,0.47672820340075805,0.9645790622691649,
 0.1342833465302805,0.006925617189989408,0.035557932238462486,
 0.7706083232205977,0.9800567694871529,0.6893145823883102,0.6687224185543424,
 0.8914235060678298,0.7252289326151306,0.6921367324963639,0.5741232560601373,
 0.8171522258070641,0.04105339086844939,0.20429215686404978,
 0.7694942081846236,0.6824035895200216,0.2093612813174408,0.40428665856019896,
 0.01159099365029459,0.1440069075678494,0.7645393961746918,0.9227645147944683,
 0.7671670746974895,0.9991234635659626,0.1483989097411872,0.5371667140094305,
 0.7870402744608048,0.8879570158294176,0.9862198233679034,0.6612433170758655,
 0.21134437347296486,0.6692745914619787,0.6839645489659243,0.7569992987273073,
 0.3632165929905453,0.6047215106279031,0.04482321536842737,0.3031688874512316,
 0.6345354638821011,0.8355004019398848,0.5983969818381674,0.5874835547122973,
 0.8557455619347057,0.8631784284918629,0.9342777969718992,0.3175397402413882,
 0.6288614464829639,0.751518143378037,0.20421012210248712,
 0.0011105964412230218,0.28525886621811347,0.8281568001422538,
 0.31256340013829775,0.9067446424918486,0.8003470210219074,0.4380465041182743,
 0.44422251073097774,0.528078455653624,0.6621610332928409,0.12147403961748582,
 0.8745528785056463,0.8800611201919928,0.4236161697580063,0.7693741984994874,
 0.06994977161438998,0.7978719875421593,0.1996090626519399,
 0.022792474022868436,0.4485536156360743,0.7857234321397288,
 0.3920756807783964,0.6694469416575061,0.8296833216573803,0.3545930449999056,
 0.4989216293211399,0.34282083616775627,0.09994324282294731,
 0.14912921706869575,0.799692304374703,0.15120004772873197,0.785694326795557,
 0.3732992296825093,0.158064350131766,0.9304570131217632,0.7778530190881876,
 0.606582446284229,0.8153421143111363,0.4865667801379949,0.37220438153168145,
 0.2503592474306081,0.7922247460131031,0.14169705115156284,
 0.34911031245411284,0.85737545248732,0.8099292008477808,0.3800970287326706,
 0.7527427484104828,0.7148483229862646,0.5196599321718995,0.5756081328896051,
 0.7705753384380909,0.941514167765038,0.4336867024003592,0.09362457527086399,
 0.4389443708487025,0.21661439214585188,0.5592105857667188,
 0.15455558814978154,0.6948011629582904,0.9440164327567102,
 0.24335453466771195,0.6716776966830241,0.30920370586716694,
 0.2910730580354908,0.5708427494208285,0.9210680024521738,0.766958097355751,
 0.4964775175356111,0.4470076659197899,0.7055087101590874,0.1763248236627597,
 0.6068885837074963,0.19829058029495794,0.19034157377054906,
 0.24397870835779756,0.8786992521745826,0.7672635382756442,
 0.042727815411191195,0.9531531897567475,0.4640710680067358,
 0.7039898214273153,0.6027849474887598,0.8240524171859336,0.4440153992111634,
 0.7727278270480961,0.5682557604021179,0.16466727855421004,0.6336613579220789,
 0.6029705242696612,0.7149843366840317,0.9803136329605863,0.3827603017614739,
 0.7076500597807076,0.7726953563776268,0.06755187354587977,0.7655883548493895,
 0.8008579797613773,0.9991618150926757,0.02064335686877239,
 0.19785016883309914,0.05344203714192486,0.0016229911896289417,
 0.6121981767747366,0.13283725551324443,0.237997192576771,0.06507183239207426,
 0.6779745292839685,0.02243852785441003,0.08594777443126456,
 0.053203639195729124,0.7919653281444053,0.3095405435844243,
 0.7602600432356059,0.8220788707481798,0.10535169505529385,0.7396399197237704,
 0.8359406048340002,0.015100610766933475,0.37517631938485363,
 0.6790089898879274,0.22335743809919328,0.5749379498386189,0.5046087214559806,
 0.24220290973850245,0.7870751814268547,0.38713950283053444,
 0.019632787326640866,0.8844722979182892,0.22404475690093417,
 0.9524752377190112,0.9491192134466484,0.5918666021231812,0.18170975982830417,
 0.1480911943404568,0.18968755253631664,0.5881912760812192,0.3163330651008158,
 0.7535506784583784,0.9393195325374689,0.8967253492473977,0.16664470302720014,
 0.7732218598338672,0.4003254359155487,0.4223547388603135,0.7054618995454006,
 0.7505409619168096,0.1251734680516421,0.8325651676504702,0.08159770481796169,
 0.9542909820868719,0.7036912874216635,0.43581150915436073,0.9412362029699052,
 0.8319789422820884,0.3013552793267392,0.2217488748403349,0.8686690153584673,
 0.7909238290826911,0.822599044048851,0.985510536715274,0.23090871277373648,
 0.11160767491103296,0.1724066423714029,0.7776366579601117,0.9686767893772729,
 0.25425155401166044,0.6589903876196465,0.38919272405951066,
 0.7539702812852599,0.342850976782785,0.1268270396591723,0.018341933032557378,
 0.7284814020898951,0.8093926544930129,0.5670983761891206,0.8913864011879012,
 0.1946170327703266,0.9281295146468178,0.6942898196869893,0.18823511988246655,
 0.11012480930357416,0.12627428214285452,0.9175231553130186,
 0.6949577605427396,0.21530730084038452,0.1353743343092565,0.7779719794164657,
 0.6834368016862736,0.9783149216682995,0.5537419655791874,0.6892395151488331,
 0.15765154956472938,0.2903682686618021,0.23621596310511905,
 0.7042364934622126,0.19832423971057844,0.907655595480054,0.02881593439702168,
 0.08432058180828639,0.815065755664128,0.099046776499508,0.4685826190808048,
 0.763111892870003,0.5898569144464479,0.12496764956389528,0.5239253626991073,
 0.5038985483361578,0.6616169794836233,0.43023228580876305,0.3280895223086929,
 0.2564396976471961,0.2030182125952189,0.46620671381971035,0.8093478760964896,
 0.5031207385387064,0.8220539125686231,0.5146174396020007,0.8741244684561696,
 0.15052026703172183,0.5523461546477949,0.9617839910817021,0.5620676144724374,
 0.8606912347156129,0.7089716479757018,0.6277899073723306,0.15349425817541562,
 0.165190004053974,0.5835471555029,0.7855780259269624,0.05830486589400974,
 0.49266950999241843,0.8138207705395208,0.21242192175649666,0.731536601367841,
 0.8013144100804318,0.925539011997476,0.2763279477383008,0.6772981819636437,
 0.997300098817894,0.9239297972041388,0.9133508202578112,0.21340628895449,
 0.4450235538325036,0.5651787890765184,0.40419595839294653,
 0.16670447785328457,0.7649726038171668,0.47547641028638277,
 0.8090772128996353,0.3712958272142296,0.6373562722639916,0.8808812371740311,
 0.45326412116705705,0.6785709673572582,0.8848037102131512,0.3855090826798787,
 0.47629369940977373,0.10665376668339743,0.058598106172817355,
 0.6893299260651157,0.9123671191544903,0.37086103073943244,0.8726206832305996,
 0.7120138572008018,0.6292327252112185,0.19374651728552095,
 0.32208334637466884,0.1099870200346944,0.8262706550667469,
 0.15987335513384826,0.8797437941461228,0.6954351651885753,0.5017309610436366,
 0.5575067877972234,0.31276144918310345,0.8157521404361563,
 0.27048825954510436,0.03166304792734764,0.9498605718550117,
 0.3787896076885048,0.4493283342623835,0.580681707005209,0.3567808602465212,
 0.2820927514650201,0.4135735399855306,0.007444040338743019,
 0.007814606039711625,0.5522382067441213,0.6102739151943654,
 0.8215385074070738,0.5797783590263423,0.5072206482152002,0.48007876282000794,
 0.7511194508690758,0.2584496037570513,0.5974410937476708,0.5246928157505311,
 0.10673842065282435,0.18143015453876432,0.6608227098596953,
 0.7875390450917341,0.31807083340748565,0.5395144676908845,0.728534006538467,
 0.6309144505769992,0.748084310983088,0.7300765961133417,0.727128897780203,
 0.46601453099749246,0.36096259384788587,0.9825694116733862,
 0.8564691111796672,0.68124786253033,0.7589078905215338,0.6599335836722834,
 0.7546948309951571,0.5195558446966527,0.69422186600352,0.09225193276264532,
 0.03248994975832087,0.08242098745856741,0.432865340105253,0.9075960499391651,
 0.8981405473044983,0.2225143919958834,0.2939995622767436,0.11232860168065384,
 0.6266060165671019,0.3175446961038456,0.7994872391915666,0.22036567696587772,
 0.34311443411456033,0.9667695894822979,0.4748951635364548,0.7647513891148592,
 0.2812026234119559,0.45237464354345214,0.6979946546111425,
 0.03963103833281123,0.19681310518072814,0.7838202790203015,
 0.7121863613989503,0.9031138805668935,0.5118753894361218,0.6401652953173307,
 0.5340406063008947,0.25862442348141723,0.7047097838869198,0.8258039920933263,
 0.7829625376798989,0.4536884990113288,0.5899182097764105,0.8179737694630834,
 0.617524631801466,0.05010313444125536,0.4703418844723717,0.8714416675056351,
 0.9128072646114628,0.7825829518461992,0.7003095342682621,0.18862702358308148,
 0.6141236206832259,0.9354325232307027,0.7509197553359794,0.49944917264480504,
 0.6364392998745055,0.3439972540866504,0.35637233079374364,0.5263198657724717,
 0.8130939258008214,0.060693803715694106,0.09547032107706988,
 0.743774627348754,0.09474937295912067,0.7083794796800703,0.7444260912003443,
 0.4462883622786614,0.35388007652332676,0.9005561109689721,
 0.26454106541000055,0.31277315101223147,0.731889733614885,
 0.07010192673645821,0.8379372842348126,0.8729444403652373,0.6675985340349898,
 0.03137953199641241,0.9145274183062724,0.8381486375110693,
 0.20271315497629394,0.36281349491581383,0.9334397357078462,
 0.9388645937309389,0.7486073936322891,0.3254753347065842,0.3048646196933614,
 0.0936845624062872,0.09258616028656175,0.5970931353711828,0.9729139041026534,
 0.3005555395468109,0.8106099180110794,0.5407852594014225,0.19958996573846777,
 0.9384180155570403,0.5841042318791577,0.7916813247118055,0.16923575022704485,
 0.8460619313309148,0.26288982794868265,0.08723280887891804,
 0.9181089418238307,0.750844789988031,0.2221052184497877,0.19404073060374472,
 0.026608286009336513,0.2563330039188867,0.26674424913466455,
 0.6058647189576053,0.032128557476064135,0.9766295273617508,
 0.8123094967162925,0.182864471695393,0.3227754337374992,0.13364090858409727,
 0.15264195779277312,0.964779816086616,0.06156852979065364,
 0.38117598862366187,0.5958231168920088,0.8098118456387672,
 0.20379382860354323,0.7560069232029445,0.25145860977028056,
 0.5567690526902518,0.015064430129273276,0.03464690255167069,
 0.8722779369378518,0.16320581439105464,0.05850697785127712,
 0.6340272856633801,0.9002712305231788,0.839836184891171,0.15241892942734492,
 0.4410688996409854,0.9844561084128188,0.4722108335344788,0.7204971411375165,
 0.3087987118770895,0.4830138729750867,0.25004257986043543,
 0.27972914078386335,0.5826756729844804,0.5534396813551089,0.8079434799470314,
 0.695777803881519,0.6583239168747985,0.23159031428162802,0.36209725428800266,
 0.22385372951633087,0.5419676714140239,0.7959609610146566,0.8729321552523246,
 0.64718371203553,0.9655384286701255,0.049106919127969495,0.07403051184806797,
 0.593621184997854,0.7444449689084195,0.7196149221011954,0.8062607527450282,
 0.3658119586848012,0.521685384628725,0.409472364251941,0.3487976693229802,
 0.8914067865413864,0.8792530441417774,0.9686135325422416,0.45474920105744443,
 0.008344595253915732,0.4780034030306636,0.9472703814277996,
 0.09109806384577146,0.0513507688210364,0.2914155474570317,0.3622438356755082,
 0.2698158164074569,0.3870253104826529,0.1265994236355169,0.22451712786905142,
 0.2730007978864717,0.5907278049382628,0.8885011642645655,0.06400465874043881,
 0.6970324367188617,0.5246355018914631,0.9668743694069891,0.14859026585066404,
 0.8242622896272955,0.9643226582739155,0.4584351335221719,0.7610078685015504,
 0.42209132433007335,0.2440236078377095,0.23317345358715902,0.233967566251284,
 0.8055764297263399,0.15171088300263602,0.435246274037405,0.5252522459125546,
 0.32376976221406095,0.6329932708029073,0.7501046604167861,
 0.029476541318558147,0.9841517767847262,0.39778713770992324,
 0.7086860667541728,0.26715361009441463,0.8326964168926194,
 0.09615874056470353,0.531201460018875,0.8374863956903595,0.79769943517895,
 0.38499093290513675,0.5182468333368142,0.06503088541963353,
 0.7720347316979013,0.3123928001671791,0.004348934495813195,
 0.12895788902444238,0.9404953883477845,0.011702712465433973,
 0.735079422917721,0.5716671104269201,0.8713452866142815,0.6257112299832841,
 0.031589068298051526,0.037323618094067035,0.27884528315017165,
 0.7456144765157516,0.5799911211982529,0.12660204635842986,0.5811556038338256,
 0.25295853488888376,0.17649871182162202,0.9746738655776226,
 0.23371172513704963,0.9855370514233279,0.3007176671543005,0.7079947259824546,
 0.08255510320857062,0.569282299225299,0.5803455269293932,0.7166762553098478,
 0.9960600037045375,0.41919809946578424,0.1816305548312781,
 0.03209094991269279,0.9176781599615326,0.7296380383530678,0.7617482362024829,
 0.09950886784025625,0.16014765253724705,0.6921462002488581,
 0.10938944583167676,0.18783794022555922,0.6873630827127494,
 0.1716686819297919,0.3496046155724072,0.6696917873004171,0.17232156873412485,
 0.1364500677378111,0.22409479571726798,0.04744651228369401,
 0.6944947398553513,0.4080496205818507,0.5464301069333893,0.9073753447265169,
 0.18122583960171645,0.7285414693442817,0.5622891102581196,0.6138045070176271,
 0.47570586926601455,0.6769536328390106,0.4856615572507874,0.3429468810358337,
 0.45787523565290844,0.5025467867858125,0.5394094746812851,
 0.44096930327848216,0.050293514535068584,0.7797985498949359,
 0.08477023960119745,0.8550383397912881,0.7883846145684185,
 0.44968793298159393,0.5476158142232244,0.9685376396509011,0.8383291593671709,
 0.3176062902246918,0.629372371540172,0.45312480732784466,0.6401721463385559,
 0.2794951797411851,0.5053862475892033,0.735634983093946,0.08557497732519881,
 0.5373289509928196,0.4115105163609202,0.8355791583295573,0.6496722813588134,
 0.28466936227850914,0.017479388075878566,0.7400642038272055,
 0.036891599502178574,0.1799684389850842,0.8306755237265291,0.727162954601945,
 0.6795749657054468,0.6510508488647943,0.04756495183999687,
 0.13203148664327746,0.4487855180876458,0.20017142606307803,
 0.7045801170013156,0.8889779354628509,0.9316548154904079,0.43645705257075695,
 0.5794788627735225,0.3121272489384739,0.7523775335274903,0.5687602303948781,
 0.5352595193676497,0.7348564725536687,0.976145600727802,0.29373986361779436,
 0.7888410376929808,0.28028801197132147,0.7280904758135597,0.9486573032728762,
 0.09968380253854925,0.4247472099367039,0.4659062002291626,0.7357451972144493,
 0.5944193559235185,0.8069516046731489,0.015045165673721383,
 0.8075171665893918,0.8873329056547907,0.3300302635219935,0.49584571315187065,
 0.13805121490929018,0.9456454147833608,0.5111692360632618,0.4247130090199087,
 0.02008344283119512,0.9899094003000162,0.22517779686294503,
 0.8213828674344935,0.8864235890063074,0.5102991802349705,0.0870834013483679,
 0.37304797202603857,0.83950584557125,0.7949752179599393,0.10028869122422612,
 0.09595003181792583,0.7300576674014855,0.25495239052903274,
 0.4532423124790782,0.6066361869819867,0.5584818110171098,0.5459917080317866,
 0.38307409310313223,0.08063034298333571,0.34460446052516314,
 0.5775902792956561,0.47701174305203775,0.8920895089778436,
 0.18867909023491558,0.31218797613738547,0.0705904381475243,
 0.8701357577322284,0.11289313993374361,0.6852577312862094,0.6539373506157908,
 0.8068054129213442,0.4500325073847484,0.5250454524966692,0.5505030720602223,
 0.4987187683671027,0.6286209923533798,0.6165501153674494,0.24239833385221266,
 0.9261236558226721,0.8542330638552196,0.28348257801914445,0.8321425092636547,
 0.4861367700339637,0.582586985738221,0.03332813991278216,0.8873384614239873,
 0.10432529166551885,0.522277456667728,0.012617740428352286,0.507158739952692,
 0.09157799588660154,0.8621345535744278,0.8986288310286911,
 2.811481529333415e-4,0.9075846418830418,0.5573908544627908,
 0.2026545338287148,0.8278572781865796,0.17574153303539763,0.5062709801345263,
 0.8350964566644585,0.7090161126779745,0.17285421860785766,0.6790856085131595,
 0.5714534042552606,0.23722324169475006,0.24323890346928911,
 0.20393099105759682,0.49481403333450225,0.32604198835068665,
 0.027072989491204402,0.8820095166318647,0.8786363935442235,
 0.2697972018680004,0.4650977366120306,0.6075385484359929,0.03830622248704152,
 0.27464286404032934,0.4548494161062262,0.2633679205203231,0.5894468439048821,
 0.28255615742087437,0.2766972159069485,0.2278976515843184,
 0.04708861615527993,0.4507994323806326,0.6678527038130073,0.5759939090316761,
 0.556152050410227,0.848442011379906,0.18752572728388772,0.11398293923168878,
 0.8097341305167156,0.9992196563113027,0.8280948080707722,0.8154738391136028,
 0.1810240855327243,0.38647882096455,0.672448757723628,0.7533524340596433,
 0.7511472333023053,0.28057428396982287,0.1264095476877175,0.9950793899293722,
 0.2321558772741772,0.9688702269282291,0.7106380694623224,0.2263230141264898,
 0.5580993296064871,0.9204175635838635,0.6762811949462305,0.7484536509130858,
 0.5563196399988983,0.19260943090577398,0.021997219289234926,
 0.5808372652369156,0.6821498502243171,0.7171187444249771,0.8252575255410864,
 0.15914548976569975,0.10554406603699884,0.8144447668409386,
 0.10423751546901672,0.45113429359402146,0.3078773149301359,
 0.32712594575808984,0.9659114942445484,0.5856126241608436,0.9060880982336452,
 0.830075317846422,0.9187551240171965,0.16228824301414446,0.3306824880777588,
 0.6169474046641688,0.6854833050800198,0.6950516686850462,0.32296702549431366,
 0.9810383247364798,0.6690668157110545,0.8615653837437043,0.6161535686795483,
 0.34293087087386676,0.7208477853407639,0.8154810073194161,
 0.13730487547188375,0.061682293487865913,0.4252034607838915,
 0.5294652311019873,0.6511475654287057,0.8904866254312667,0.2057115046361857,
 0.4373647657416112,0.3261128175293232,0.628538043326798,0.3144561717343257,
 0.966338835129489,0.9484386418497845,0.2656959655146308,0.6910491997685057,
 0.3288930753256176,0.47848604097882896,0.9812240583056663,0.7705865075073475,
 0.7688932125499821,0.3693137763129205,0.6407780059399549,0.2353272784011754,
 0.43589911982497564,0.5597892027466499,0.12090210142115476,
 0.6649942487807318,0.8482911082601408,0.3052800236063685,0.7342156993443277,
 0.009185152516833606,0.46676486951535257,0.6513090964518347,
 0.9532081594990963,0.39170596041921546,0.03132141383931941,
 0.40753421106591325,0.2765757349224851,0.006826331802909591,
 0.6218759442393164,0.5914472957662591,0.1584156351780457,0.4180918181592569,
 0.03187826540871985,0.8619324871328353,0.02982677128495148,
 0.12264765097279062,0.5341803501688278,0.2790759131217454,0.846711720550094,
 0.9513245138647283,0.5328537459696497,0.6979175981685579,0.968233994278785,
 0.39928172922962835,0.07186866387290214,0.43292474875581144,
 0.7797936185621629,0.3280940041053143,0.8940210777214126,0.6606374446672117,
 0.45050415194846216,0.6372102645478237,0.5233009655709181,0.60657173534768,
 0.9480729174371655,0.8507083765291225,0.10615960233252764,0.32341014861538,
 0.1868241360196361,0.5727784337241734,0.16907119919775848,0.2636105073002757,
 0.2821185668230619,0.7717210802161515,0.3822621601204368,0.9292872740913282,
 0.6377481084963617,0.8786423529526941,0.1441792298488992,0.5172469365561574,
 0.5253306506950396,0.40174521863258483,0.2814445987596873,
 0.42594574508595484,0.2819683992522184,0.8045780231140655,0.5494486979819023,
 0.18138936611361645,0.5767042154704433,0.4493257757588044,
 0.16443910318531552,0.7892287055212187,0.944705377617388,0.7658462956620273,
 0.9462098120948188,0.8333562864849738,0.6640935111909492,0.9096136716334762,
 0.2559196109929821,0.031771177406769935,0.28836019073297137,
 0.23278041063571775,0.6329468921182902,0.3548337889692489,0.8246731907042452,
 0.2510690485520908,0.8531827178564612,0.21445866234753463,0.8855263106499784,
 0.04063130636285606,0.2037451029507389,0.20166454231571707,
 0.46513008564020086,0.018029994208144773,0.1616213616329576,
 0.8774012403924341,0.9877267837714016,0.5875354520627211,0.8437924580156762,
 0.561034422303699,0.98915517956317,0.13485292610658017,0.2953054353542567,
 0.012781715059203336,0.15785625248158963,0.638784762616696,
 0.26098515965127334,0.057274517604173236,0.36548707510023326,
 0.2572669556584324,0.26974863759657897,0.8716683386274766,0.891470561956584,
 0.6463828346418552,0.7223836419755366,0.9773532068245291,0.5703414865907054,
 0.3557053734360358,0.4325856002599736,0.6942727379361473,0.2905063109812269,
 0.03572314258472076,0.7384303330187753,0.8304202434131536,0.3897389709584276,
 0.15924498348543092,0.6902219833067048,0.6616313367652379,0.853784492824023,
 0.7506740075856341,0.15175081707238158,0.6492895803928779,
 0.26739379506445926,0.7832424656149498,0.42653230828098454,
 0.9957320851333069,0.45531596843331557,0.033191820184578336,
 0.6175201270623571,0.28076421112401406,0.3965523587897962,0.7533691854578841,
 0.8051224547509229,0.9068965451304847,0.914760580476618,0.13433488457746035,
 0.5836784330979267,0.9276478565674704,0.42323915069081175,
 0.18628392158567542,0.26351898418681974,0.6378725540226362,
 0.040573414567203314,0.21918779143345957,0.11739677979529928,
 0.22607987811866004,0.1978966053518355,0.2354637818497658,
 0.39122696636278587,0.02636139770284507,0.2273182944335168,
 0.5956647692640642,0.1297852749692241,0.6284268592068231,0.0642193425535279,
 0.9920775465529261,0.5729853666977349,0.9114466518690947,0.4790731215672419,
 0.45997157477919504,0.579056134706122,0.49877851543411744,
 0.08080212288088029,0.004858868730987731,0.2666428160893176,
 0.6271673674276894,0.11841148533065482,0.7679691782260918,0.664723326065988,
 0.8385955275888411,0.48428022423967554,0.7832715188952151,
 0.12833745183115308,0.3025518747104494,0.4039619875623104,0.7047514541046641,
 0.9452491572205962,0.34969856251399045,0.09173536976441676,
 0.9351161640344816,0.138647738192631,0.26646489548845476,0.4151122241140871,
 0.9820055882704503,0.6206726965324241,0.2845019827807366,0.3627358590067441,
 0.3939678857460869,0.3585148056499007,0.39763775272470403,
 0.33159492117174794,0.6379446045801233,0.8680664058336762,
 0.12619285217708454,0.03835337520105053,0.961528928714392,0.5393477153954269,
 0.9337906287946756,0.8712399053578298,0.06043225504233818,0.580226392607747,
 0.6936044391554897,0.36169544006980736,0.804841126806808,0.8000103273793591,
 0.4452943726338221,0.5423282424047029,0.514206490203621,0.7478463411689675,
 0.6091030140023432,0.2513847396265707,0.5867324731618395,0.05893844853108254,
 0.38725054041091844,0.421425410894547,0.4809786373140441,0.29409215991070636,
 0.33542006650924716,0.09426091033613915,0.47900992746202775,
 0.5753413823350056,0.9525051716934406,0.4090221380392772,0.39230130777005245,
 0.637598395868647,0.4179345861949533,0.17114541528501703,0.4983217517805836,
 0.7408480854822098,0.9884844262358626,0.15718339935633063,0.6853932505894622,
 0.6679566387244071,0.8344090761188225,0.15325898984874253,0.9744277306341553,
 0.4765058730574967,0.5477728574380427,0.48051048000951624,0.6823896403783674,
 0.11625501867519805,0.8931831191925163,0.47451714126860445,
 0.45082439317604406,0.37868875285830905,0.47343526307146166,
 0.5113750388307243,0.8524191447831967,0.17647653137592556,0.6569923928619869,
 0.12881189306480745,0.516085638823092,0.8660621866771006,0.23328139909327827,
 0.5084229826622865,0.9545272046445016,0.3339093687588084,0.1318169670180126,
 0.7836009204764053,0.46049483238429223,0.11709519134636559,
 0.6078803604044218,0.7845333743249215,0.13452365838107916,0.6790949431615656,
 0.21717077164771337,0.6728956081870393,0.7741438363866826,0.9658698715024421,
 0.015689107372604982,0.285179574645664,0.08108490200068785,
 0.19862265248385103,0.39333664368973964,0.1215464913796831,
 0.49722430068170853,0.8646957945405509,0.4199508266946681,0.9211785655602225,
 0.5289816820060624,0.17482369651803276,0.02904773913415759,
 0.8900031170100857,0.29394504409137656,0.8647933489968533,0.9882117375429486,
 0.3995904095775904,0.4576496979456789,0.8018673745334232,0.6699502502066907,
 0.7190403387809811,0.9088160010111906,0.17919809959621402,0.6328214265854115,
 0.24628105333268813,0.04457187424366993,0.07372776507649326,
 0.6659398693641585,0.7989978356238189,0.2320966928990822,0.4754941776646684,
 0.5278523799142869,0.8557792408248259,0.07282587955512099,
 0.059910655952580494,0.8627713169597678,0.7368297918634211,
 0.9888550548782633,0.9576246734737343,0.2537099798883762,0.14220322813945996,
 0.64844516045732,0.9478815968820189,0.1152506671074992,0.31650981835319225,
 0.2704468674967525,0.15302332016076137,0.07311452179161781,0.562993928276073,
 0.6287950106185337,0.06841102974199087,0.9203192583819052,0.1440347070669079,
 0.5490666577201514,0.9904868083345164,0.05862776842596895,0.265406118932794,
 0.33195570651846584,0.07027810800599077,0.643879121624953,0.3180439159726376,
 0.3982773052943196,0.5951881711564583,0.05374221761665243,
 0.11884411459351263,0.33404765000788283,0.7759581885539041,
 0.0677474642217677,0.6857626970340447,0.9890001878391801,0.42923769983815774,
 0.27587917878098345,0.6139029799806188,0.6140478273795478,0.3946706142062073,
 0.47612974496088734,0.23348359913112093,0.5439233426162012,
 0.8402287665077769,0.3062557788568566,0.7548058115669373,0.6146161899280435,
 0.06731829416547208,0.23025424156161178,0.36638800805922345,
 0.5971902522687466,0.7142885367161061,0.4183120521134531,0.01922104179561812,
 0.627153130776509,0.4443198461723624,0.9430907036995867,0.16877056307626326,
 0.8824383166328912,0.8626566002353371,0.09059150350684841,0.6857379073103012,
 0.4566736461710994,0.4213588470326327,0.8267160704180282,0.7922872994421855,
 0.06598011194936593,0.05698158378262752,0.9050287353798449,
 0.8671798594738457,0.8967360184465134,0.8369841856032999,0.13707526562287953,
 0.9187087060683301,0.8042237464692031,0.762550484503347,0.6764607642880714,
 0.7509711899497337,0.7274966462978523,0.5771145664090128,0.24417172623015726,
 0.4773573645932617,0.634691697672902,0.5685774662481969,0.8458156567425642,
 0.4914264951676852,0.3409644174398557,0.18447327561176763,0.8360984712249984,
 0.8139749763648251,0.5788033994604549,0.5753166320527167,0.40139546969088935,
 0.46324461442276665,0.7292255323031017,0.9611825643271545,
 0.24510070770925807,0.8736739639565924,0.16010830238536178,
 0.19108165779504627,0.6442700931300579,0.5057649572357222,
 0.35804361209587754,0.8500405061184614,0.6630829011357923,0.3030973440779944,
 0.034652277670189524,0.8066507480402305,0.886169788936164,0.6081473116904218,
 0.8896314891130497,0.6380045553871094,0.2977087666272249,0.5941043401279632,
 0.008807397967318664,0.038546008311842406,0.362702513528778,
 0.8519450456656193,0.06571275572076662,0.0837621977360069,0.7359795721524794,
 0.7737649074501781,0.9288318022609805,0.808689081862897,0.9699556300868069,
 0.6962708889114588,0.8973347184252756,0.702674336250249,0.19233031671138923,
 0.5416016318150755,0.059496868711958695,0.9575339097466833,
 0.6419145222814281,0.8908547094591877,0.34964976013027327,0.6754542572286614,
 0.7722832939936075,0.2497107624315439,0.7249431135789125,0.07135268738385381,
 0.4179418254484829,0.6885941804494169,0.8043316315104952,0.2749064976344898,
 0.21029503545159145,0.8948967993071492,0.18062022927947907,
 0.6496817938937808,0.011830059671315452,0.38685849618158363,
 0.012350149034631719,0.8894027337760069,0.15846640238164633,
 0.5839000096725455,0.10885057259043052,0.20622831181735046,
 0.7739472626271734,0.08010322681371318,0.36264737116011836,
 0.5933218678271655,0.7393198265127657,0.01997443633133944,0.6131836536545545,
 0.05655898952604532,0.01443366469531615,0.00793943738005043,
 0.20704478584295039,0.12580066874818419,0.0810483429570078,
 0.3968353060393768,0.4355593176213737,0.2780393343650712,0.04010678889282293,
 0.8509401095304,0.9873872899342258,0.14888833021023407,0.6412980362059202,
 0.0903072969792671,0.19827475021444407,0.1688592451367783,0.4659678499147356,
 0.2915845752433409,0.43170351331864587,0.46599222437973276,
 0.8597222044122621,0.3147618002445376,0.23859250491744588,0.8860645860412979,
 0.18447852958330757,0.8058989662910947,0.9441875957211721,0.5569914598804941,
 0.17585424670525773,0.9167838016159138,0.502960183085896,0.8204668214049028,
 0.9514679865744055,0.09747785583994095,0.3839862778575016,
 0.19865750217014366,0.6835717908423552,0.512158624011936,0.589124527373399,
 0.0867031652847724,0.04593335034206303,0.1851285871485301,0.4385645052929067,
 0.6845290192183288,0.3540726158842107,0.5895672444894369,0.5083830095913818,
 0.13966625030039959,0.20344744926442504,0.30315910064538054,
 0.0481096508408253,0.33051819219296086,0.2519592506134606,0.9761104930665951,
 0.598063991019237,0.40067648062671657,0.3108808053348523,0.9335102200914467,
 0.46337908888205737,0.8939735545340065,0.4047835525975011,0.8207636621126398,
 0.29408073560108705,0.6436441048762715,0.527539045927074,0.7277461213924186,
 0.14259336395121425,0.15694430265053316,0.5207910738762777,0.996451183671927,
 0.0030616104389120657,0.7521274251501693,0.8800146645108685,
 0.784500102692238,0.27454506177034743,0.17414775020655393,0.971073627284643,
 0.8845645997787672,0.10617607787444694,0.3076102445655142,0.3903143574022123,
 0.65807475722773,0.5140298863725377,9.706269792766342e-4,0.9915199292868078,
 0.4279729968095024,0.041921227812588335,0.4337413872144704,
 0.13000567218343295,0.34548473019167203,0.6860375543160643,
 0.0038106793783395654,0.27124110651282796,0.6823383682246625,
 0.8185462665618862,0.7396066640848957,0.5553917093491617,0.23967820138785534,
 0.5567849443894611,0.976458396721792,0.33925862731559153,0.682498300390006,
 0.47082545245269936,0.7718465537787442,0.32414578399366956,
 0.32933392539454154,0.4548670711149405,0.4401529407577327,0.5179236961612517,
 0.17817718268280036,0.9149360753348244,0.03262092117050264,
 0.10312399407401307,0.7052376621727431,0.14332290420995386,
 0.9739116484880261,0.8024580661063189,0.7631419573592473,0.41189857825305054,
 0.6628628211295575,0.5535963361286425,0.11264009514116624,0.3956174188258754,
 0.010619021581886212,0.6410215370953082,0.12704543745249375,
 0.44714578195804255,0.5303289089886017,0.1769150926001889,0.9872683607771449,
 0.12151770645214277,0.7405719728192213,0.37404669570479787,
 0.4065817536768489,0.7020871061143922,0.6741715022308384,0.22869733801802516,
 0.5688602922044457,0.5196194378267929,0.7703666928562232,0.4015399873353439,
 0.7626985525807435,0.9999250763040428,0.9222742818462124,0.1497462286305039,
 0.04952690904023105,0.20444926051821433,0.2979787063720921,
 0.2360787688674425,0.16215118587271926,0.4587954280420041,0.2892625684663138,
 0.7139245108564619,0.5577382636269648,0.954393562996829,0.3351410434832829,
 0.23040807639212268,0.6195402734730725,0.08361327451594625,
 0.9525726018342672,0.28956627203686025,0.3568305740820885,0.9432746372613127,
 0.8643337600671523,0.020616687649975485,0.8023520062379385,
 0.9250985622070654,0.44091561290779313,0.6892528766895781,0.5785944457532081,
 0.16557616802303243,0.9556634466783076,0.23839096334630383,
 0.4823820705605423,0.38347544808115674,0.2058752754372214,0.9991280563562208,
 0.3782595645130775,0.7615014709123329,0.18530410039534306,
 0.04112813825640782,0.5129894096224504,0.09291790544089396,
 0.1741082096920593,0.2389132882868895,0.6179707900983504,0.35109079970805157,
 0.8307951320927369,0.6602547222365258,0.999407807650129,0.7366128920203447,
 0.6520158340386695,0.5202672509057714,0.26967778905248085,
 0.005169639247328828,0.41566864729321285,0.4471936603031036,
 0.7355583126201531,0.08581394468905124,0.6732241582898588,
 0.28631246006050304,0.15058436510811313,0.8057027688001525,
 0.6773880228278246,0.23398517737944657,0.13283582620934486,
 0.5100385545671062,0.08787979195743711,0.907212743813843,0.00499776262265339,
 0.348813137619881,0.012723925793985957,0.24510031116698627,
 0.8258174343181139,0.3404396457193717,0.3169301572407316,0.7362493565359869,
 0.6108161607940493,0.939668304922943,0.188364292895847,0.5443755146248181,
 0.9328847163310754,0.6054991577093107,0.19765599477462648,0.4607784302496817,
 0.2665892286296788,0.010210006629057577,0.5982633599559808,
 0.39202998762959895,0.313637456570654,0.5944627725611643,0.5170829475908888,
 0.7564173845895596,0.06956343453839164,0.17280284281116942,
 0.5624419329676799,0.5978955306003317,0.07249460606334324,
 0.18402858908438535,0.503868513162897,0.46639841361746603,
 0.03994928465694492,0.6014496331321445,0.042920703556963535,
 0.7405084697378765,0.5210029556471607,0.2835618430914332,0.2105931604574831,
 0.5998397079352005,0.7855748294134804,0.554387370075449,0.002846481554660274,
 0.15276768161853704,0.07101034021988706,0.9888282717229417,
 0.33910994858673127,0.4566232491084613,0.1714942631236911,
 0.06270056713173766,0.19572557705224947,0.7644447881169554,
 0.8296580245603176,0.7645542626626353,0.1440926438806689,0.9211254498542296,
 0.43959059210755835,0.2690375046876755,0.859150206174157,0.67930874568002,
 0.24355100747539105,0.32752983585077156,0.7763950402047811,
 0.059759656745723144,0.8367481597263281,0.9358012298992004,
 0.1495074524106358,0.5032366847718157,0.08385258407660223,0.5703449922057531,
 0.8778658460316506,0.514040915616943,0.38528919057710864,0.1874568483811997,
 0.19211999652665324,0.9169779420031215,0.5853319103284318,
 0.35414887081003077,0.014791404349779103,0.9055296917511856,
 0.11412598952641206,0.4010564792844806,0.25229263253310963,
 0.44708120388646133,0.24219606466465615,0.765467442438648,
 0.34444381370554455,0.07282480268529357,0.3348143026133077,
 0.5616703954645164,0.9609969875417236,0.9050784051059348,0.5805170358739465,
 0.11390638817773024,0.8277778057784672,0.007617013359612512,
 0.09591233211504602,0.382170528517925,0.6355427492690869,0.34057659598602485,
 0.5623675038279886,0.8227981423816192,0.1845622423993707,0.7014198757224975,
 0.9662628901623788,0.6647256404191582,0.20337113882598568,0.6525808929676727,
 0.720788380276768,0.379891393280234,0.22103782021030993,0.6431863885266687,
 0.8781887176400367,0.8302087844528866,0.09549548903011829,
 0.026672554083667532,0.4157365702779985,0.8635821921881737,
 0.09557073447498454,0.45406000658429524,0.5033756792145142,
 0.4981306791553822,0.14608509997296926,0.19891149736801406,
 0.7142537929336247,0.3826635239827678,0.4050518122492115,0.5343187815787943,
 0.3758580192282974,0.4266379290139488,0.2753571551408672,0.4247296659799822,
 0.1781275991688689,0.24770776809029005,0.22654376615878546,
 0.39097536554586665,0.06691829967284257,0.18865778849729087,
 0.298699613443721,0.06784552087794526,0.8379135510839887,0.3651777099849749,
 0.13313085217890164,0.11644698760822747,0.5016738388745795,
 0.10119100933708713,0.4046837553992674,0.8076358915900469,0.7449898600865188,
 0.9059833344911576,0.37794412373475283,0.7803916137423501,
 0.056733000742874684,0.4186869016911886,0.2114173570444966,
 0.7872957036648106,0.4817327317145075,0.11570796352065937,0.4979319911811749,
 0.40778123029607616,0.33744476630870646,0.6694395395635571,
 0.044618858725781196,0.9393890365738252,0.10271799212310662,
 0.621120290690478,0.572738836471424,0.8022080433386892,0.4155237356348964,
 0.1191133839065075,0.5160555760753115,0.492420381582535,0.4082173927169177,
 0.6587135604992614,0.24782681894604364,0.27905089989555476,
 0.6697420125027024,0.35851903657050377,0.13506370420095726,
 0.6184186346724985,0.5923815695470067,0.6724095776888969,0.8478469114989643,
 0.29289694978389047,0.627973233579659,0.8882981364759733,0.9717313975598596,
 0.047687033349319674,0.47533976616989526,0.5534042643684839,
 0.5390168516618783,0.013139827189243003,0.5682346221243735,
 0.2937534120626544,0.12891635499742438,0.9651638996923879,0.7978382563340931,
 0.29875168739557423,0.3486524698731529,0.13462148485367942,
 0.7678850547780705,0.9450321928568447,0.0032596879737532003,
 0.9874320387234847,0.13971834266742766,0.06557461465717018,
 0.4309416971473581,0.06841200441058881,0.5294801975784846,0.4354037305756885,
 0.1433024908181667,0.3684075201925311,0.9931204977249473,0.39044971621390134,
 0.9496191908587661,0.11746391126217604,0.7180403941725984,0.4965631843727548,
 0.6715805256154411,0.5084650802582518,0.6914242656243117,0.1104882442075934,
 0.5712058731226878,0.9121162054609366,0.5519963965192727,0.939788709388783,
 0.4341568258464581,0.2236542520714866,0.8230341948561208,0.0585157381279694,
 0.5964347147316473,0.13785132985662096,0.9014729577987124,
 0.25753011827138395,0.36056220545620976,0.24461468044630086,
 0.6580965210889266,0.9074291670726728,0.452692081712053,0.8732976297647368,
 0.5411850981502039,0.5491832258179028,0.2534083095998281,0.7381502450809155,
 0.8144358302210849,0.10414722558731915,0.7162895280188147,0.8745519169169276,
 0.9864553765333899,0.18896986779465264,0.8389031151656109,
 0.32347696892050726,0.7789854333534758,0.2243034851412945,0.331622331522077,
 0.6369317864452261,0.2621789419153583,0.6253198580196261,0.9113776967252237,
 0.16937509032147213,0.3343177545301369,0.3636631065488225,
 0.19309756083029217,0.5942250780515475,0.8004474724843208,0.4952551539900286,
 0.7601193319711481,0.7496760160797981,0.9274774877274208,0.8857873026548608,
 0.06692394882420039,0.7848649432535262,0.2829568428633803,0.8947245493005707,
 0.0953326950600627,0.9476933632591874,0.10247084900480707,0.4210212090414197,
 0.2918780326009822,0.023017671726263456,0.037060493982560616,
 0.18337253599954084,0.5239649360412759,0.9535151390366959,0.7496373967564928,
 0.762300894279778,0.1877648322458202,0.7512073406184284,0.694297872547545,
 0.8136920645939512,0.197494773814616,0.31922865358712116,0.10996218361401089,
 0.22098873871576108,0.3550826547494157,0.31664688017819176,
 0.17852032890736313,0.5909149129073398,0.26999587638329314,
 0.8247970987517478,0.5078991514899075,0.6154380913515872,
 0.006572126880549312,0.9365410980971552,0.6599237449808035,
 0.8687392918606093,0.5949778921061295,0.5245123029787142,0.18146669827655648,
 0.0901559654202655,0.5029166268820799,0.6560583394036733,0.14811450244131663,
 0.5681757288913095,0.3411724923738596,0.6812060687351638,0.3716782923431988,
 0.973454618467024,0.7046912232792581,0.06225044315311867,0.49843222709650536,
 0.9390375748702947,0.7484039600729371,0.3887728298104294,0.7760077317136246,
 0.9078079778294526,0.8022703663602735,0.28449757447307444,0.4484110870398068,
 0.2615680864273001,0.7020236748030075,0.7576891455434227,0.5017072692851089,
 0.7100123959937163,0.01719168709439134,0.8400340453751822,0.8454950127204305,
 0.9154361336123632,0.9807439315336074,0.9298760989134467,0.62016606425189,
 0.9401825974436588,0.875717487623491,0.0780958912467572,0.6719300930677742,
 0.6956035029677908,0.2601288805986447,0.37201715307839245,0.4071327447493194,
 0.30014326919659395,0.9225541215675768,0.12501570123416927,
 0.5505084072546357,0.4524128242013099,0.9449155800731601,0.8784458576822659,
 0.6497691911446447,0.8374514032141748,0.8076741773846345,0.4390690318791871,
 0.2888566601730378,0.6646307525223973,0.37617055549467904,0.9272646401377627,
 0.6013179309260941,0.046186869137690545,0.6396320218702328,
 0.4652553267352151,0.3995391185776499,0.30372476934843806,0.2725599128256079,
 0.30097942161811986,0.30749056684878784,0.0846526325421344,
 0.6251875292171684,0.3882242981964007,0.006886951425448262,
 0.9552182366096452,0.8604358519080672,0.8972990744400339,0.13524082132837023,
 0.1506320349337199,0.9720768988823546,0.856420436445163,0.5654418239164963,
 0.15473249980039527,0.34639712043155513,0.057428559696261594,
 0.419728099630291,0.16261666659717688,0.20105586647040508,0.2944956304065952,
 0.8724428023957479,0.8914433838922275,0.5188980095198584,0.48010073465095615,
 0.6330470913955291,0.49734020159900216,0.3500993408580073,0.2662997240812571,
 0.6782651788747228,0.8751585191245489,0.567910301885099,0.016094556890793088,
 0.9120616753576997,0.6166118406139142,0.06835852153666133,0.6517175575063693,
 0.569660747342782,0.08838685394818724,0.008930340294557082,
 0.9349335649415536,0.6840710977467507,0.9332450628534166,0.6301846556763738,
 0.9280881182141139,0.09353583588891734,0.8729528381219811,
 0.12133234029268736,0.8137171764726507,0.6533461272400274,
 0.022446417117296114,0.27646165408869017,0.7271969407124741,
 0.6512159920894686,0.3347095434771161,0.4117716707249873,0.08851605186110789,
 0.9439984095756577,0.6594709990737693,0.09626348026317932,
 0.17330200902172488,0.8599792545134775,0.23325398794431706,
 0.6666786199831988,0.7573956608437077,0.8870892600390758,0.3784258690787936,
 0.728214724971951,0.6260713678108862,0.06813092453613823,
 0.009493049922483565,0.8657101769246508,0.5392847693471763,
 0.8714774849862861,0.020422411489590342,0.580095909948985,0.9003847772849249,
 0.2161739496923828,0.3503453529248357,0.5331864120311521,0.5476662316395275,
 0.7045192395409008,0.8634725061269739,0.5329573705240271,0.33268796126869626,
 0.3976928927778861,0.3023983662211783,0.7319868289508769,0.3131956427142998,
 0.002293102731748209,0.2261567604906768,0.8805835048999062,
 0.2359638606611053,0.313947054471615,0.0944194663839103,0.8599201570224145,
 0.9473533245760575,0.8255026702204276,0.24045673563000136,0.4733676178751417,
 0.10187336684871084,0.17358113184323531,0.7165378599559541,
 0.08243995739401211,0.26325990589703874,0.1388621701385927,
 0.8188627124525929,0.9020772058274358,0.07182486848763814,
 0.35528429931204997,0.4583751827255058,0.012948821360778795,
 0.1489476398142634,0.4364855472268745,0.619028917578167,0.8495508389411198,
 0.5148292114361844,0.3941065845130757,0.31574034996363765,0.5761689150548325,
 0.8296065320860884,0.7633945367407967,0.7695251604776866,0.3972550464003788,
 0.061114933518658376,0.5683144685825249,0.985497704212413,0.9518758365103254,
 0.7248941779951859,0.538899732807609,0.5200648447985879,0.492497099001173,
 0.1637859885008287,0.8787519172886763,0.9374950399754498,0.12676833465742698,
 0.6957807357409624,0.2607040129783422,0.01533102642856532,0.3962184142247732,
 0.538309755047358,0.7550161669691708,0.6367268856562673,0.01573819397777987,
 0.5472108680763347,0.28482103468167796,0.2620136482580733,0.4623711079742939,
 0.9331836176442834,0.427209270198344,0.176063377777946,0.10701844094124446,
 0.8986382256668307,0.1569845980681095,0.22108591821598367,0.3897974766479795,
 0.420369223753549,0.497241342766059,0.7182415672445742,0.5759284361445309,
 0.7106822803840498,0.05378937966760455,0.4120448214827319,0.6779934489350689,
 0.26357691909478653,0.5235052310529406,0.5495159390185049,
 0.25935364981472075,0.31875063441958906,0.8827352224723312,
 0.5496025321464554,0.9031577017288639,0.12501926207522973,0.6986714420274884,
 0.30999506886936734,0.050925632808355825,0.8466283051382689,
 0.3470046436172083,0.8292344709451657,0.8178026232804093,0.4247058672003019,
 0.24327068449429456,0.7117464647143811,0.9196923870817618,0.8443143445642599,
 0.09267048334828798,0.0723738477306114,0.9172127025225628,0.3298536066126082,
 0.8960690836426478,0.8993570989850326,0.6416914056696933,0.7059890345247057,
 0.5591668447999756,0.8896751703765201,0.04622853488519485,0.869449196947568,
 0.18218531071242083,0.5355299501682551,0.47778215553258396,
 0.13308760956122878,0.5601007883687639,0.180988453782976,
 0.005913956651011465,0.520609434879338,0.03137776499570799,
 0.7902113680412705,0.5003098093112266,0.0921218047435276,0.3274889599177708,
 0.7324900013334501,0.1722851204990984,0.42923416760493316,0.9791317855934335,
 0.5227754200447566,0.5201678026859513,0.1972728601551338,0.7606536979217805,
 0.8596731023518875,0.5422858907790767,0.6728594726242219,0.823578454020493,
 0.6020553217288007,0.07588777646705691,0.44217294414675556,
 0.5902740819933341,0.8321884575065001,0.07836949430531959,0.6281159172264823,
 0.9323260645439159,0.060497109696632356,0.5490685292281778,
 0.6982874135680904,0.6813597629167374,0.3406110621207432,0.2844606687577542,
 0.39220970666346955,0.4401963088305675,0.939734582731012,0.9678090534774569,
 0.5034831590801311,0.4579143097351879,0.9064175919161704,0.6642361221689967,
 0.36651950548384604,0.3494802760898078,0.061345984794910935,
 0.5779133666438436,0.32113839191015037,0.14977224827197178,
 0.7659351603655956,0.5926617463849745,0.21399754437521723,0.8038396423592424,
 0.1847064723799814,0.940669242339051,0.4393475728075823,0.20480967605759437,
 0.15859967789927754,0.6547613262335532,0.8508259496706739,0.6044847079453107,
 0.2173719849932313,0.7202366344542327,0.3640634444671065,0.9777997241155021,
 0.3557511076095903,0.8959018024475702,0.517642240004121,0.519916720481714,
 0.43800412190175475,0.17101666365176027,0.3801854990832597,
 0.24884601202904788,0.22842257559169388,0.1664767340835529,
 0.6252623421064883,0.6715564374599037,0.3240830222589911,0.7826590334188815,
 0.6411530610106894,0.7912112501305032,0.7976815697499421,0.7516778296404247,
 0.4459741973394319,0.5793723711447153,0.8620346882871132,0.18023145371886595,
 0.19767959039386207,0.20917422281737452,0.3231587642662471,
 0.21572315209115478,0.17617102162639342,0.5303048680567516,
 0.20226679957750437,0.9709463726803992,0.2840033756658926,0.5814315984918701,
 0.8402824091850133,0.06545223576715165,0.500850625205092,0.5696869893420948,
 0.43301461994917556,0.2376717263578908,0.8785003056839521,0.5080387313984156,
 0.45202876818806226,0.8053784626704424,0.5703098303531279,
 0.18289090235221128,0.3234479516924712,0.2523658051854838,
 0.14284753548469276,0.2875678446034494,0.1768271370305814,0.4675567183916739,
 0.6271326697752405,0.05817680789584756,0.17457043474366274,
 0.8596021064237217,0.7178580800558427,0.48011961825707594,0.8236499903881428,
 0.5671155328024047,0.442744180910571,0.658343589464716,0.6382454969916718,
 0.1784841322569044,0.78131157051339,0.46595856581217243,0.8074022886284559,
 0.815122979384429,0.4775853289035319,0.7059076615559333,0.2644118831734995,
 0.3363347675901883,0.385502653665422,0.5793386905127815,0.5840392101083165,
 0.6491687936796829,0.5718976555201287,0.053932270770982615,0.931748807223504,
 0.9793501893606723,0.14316669251217407,0.6412063077919538,
 0.01674977214560902,0.7776291860863659,0.6464209626867325,
 0.28091291285921727,0.9050510991975328,0.8723044265630354,0.4223830618744382,
 0.029663471166858302,0.3225056256344644,0.15899776691879208,
 0.09080333845421396,0.4890335611675648,0.18021452875931698,
 0.6290804527638683,0.6926286930852601,0.32858301051437666,0.9430190618164503,
 0.09403395193457553,0.979597591132895,0.6338772605082668,0.6275297749254514,
 0.015059019113348615,0.35041546441571436,0.40384766068620526,
 0.5492469697645028,0.9602798445478163,0.7453593744666639,0.7114225126161822,
 0.4422043328825437,0.9505769467447502,0.7162118950462666,0.4467671302401216,
 0.28786898363926994,0.6588212189436979,0.03536505599382578,
 0.004974593914594649,0.32614804262594965,0.4918736624885648,
 0.8156286322433448,0.6859432902561324,0.3308403701059246,0.9959694367945258,
 0.9953422464859792,0.5988512256158202,0.1482568745476298,0.7399586132494771,
 0.37544475999719895,0.8337202412531776,0.03845552402270658,
 0.7324200355163978,0.681048228976936,0.9920416516368648,0.2865081198305367,
 0.7652350565374341,0.20941000241692298,0.49459811115547003,
 0.0592809197072619,0.046622739334574836,0.5219141780897701,
 0.9860646976213413,0.14331537186673415,0.5183938634343167,
 0.42485262621957676,0.986633555825671,0.4824572331095479,0.24166593846237883,
 0.45251295087979615,0.24842388414400196,0.8213635347881927,
 0.5314821903814545,0.23779373516000624,0.4144110446093232,0.5661345148924491,
 0.1664548347235838,0.08271600995273454,0.7474473707560181,0.6108123502525109,
 0.6846410408621597,0.9063757622058003,0.5292646841337446,0.4544432841573731,
 0.3244094281330603,0.31016469575770256,0.6654366339722759,0.2386005812112234,
 0.5793319882172143,0.6469231879484886,0.021535961080036525,
 0.9282926139215996,0.6158002529253276,0.6709275714261733,0.10466221880828441,
 0.09285896564048457,0.9401684179949289,0.9402474939276078,0.8220391798565907,
 0.37438871332947554,0.1448575421559095,0.5858499158522238,0.2512689364489036,
 0.6786723725497689,0.8252153060975653,0.05152673230718663,
 0.05807062500130056,0.8498082966688942,0.13991062361314632,
 0.3734708432175352,0.10194716617061184,0.7850344937263626,
 0.47931615568387587,0.5209107631017541,0.05026161160588494,
 0.6614970209110779,0.026899032326807792,0.09477148819319547,
 0.15344477340042018,0.7224367008305395,0.920080586962895,0.46284791024891425,
 0.7670090388161962,0.510589357887075,0.135392375727731,0.318659400876099,
 0.09163584866406282,0.3827511580350841,0.839515683398556,0.6722732129740455,
 0.08979089507587579,0.3618969803225909,0.4928946253051377,0.952972179090051,
 0.6990898396687443,0.3191120599154411,0.09288654446803657,
 0.05489130020714361,0.41655024779862837,0.7970850920681641,
 0.3312625385166712,0.6729865446783889,0.9623311015726992,0.4438714489114499,
 0.16107656527234981,0.8252252814597183,0.3309348160482476,0.5037373320816256,
 0.31610712097424487,0.035433437128880696,0.5209756502846021,
 0.3312129692586463,0.0441436239063443,0.1619469271014824,0.7837188660380867,
 0.32927120685865685,0.37981310457040296,0.8890028836310865,
 0.8476475584210953,0.5875191493513603,0.889057173605797,0.49127668524049684,
 0.6722056297390635,0.848092128160139,0.7639389250323578,0.873886936943887,
 0.6367870226036622,0.884107594289886,0.028449163187955984,0.926076510172472,
 0.18711897245403775,0.30819947410963,0.6640168203857152,0.9359390908133073,
 0.6458604008701302,0.5759580750253103,0.4282117030950826,0.3731589990800911,
 0.5784733802834349,0.6232681791725517,0.821342218234704,0.7518552124106742,
 0.9090396393773786,0.9130695631937775,0.8661294429525329,0.6173435425766457,
 0.17017123879448437,0.26689865219242737,0.2504651280587843,
 0.8427590518793934,0.057849584116267616,0.4391447300403769,
 0.2736859502627863,0.6182800099281931,0.4446608575732519,0.2871067273611636,
 0.9128842609772406,0.8423018274975194,0.6808217654406757,0.49547406814869344,
 0.7930990076543968,0.6555643205320658,0.12725672276500222,0.7841307887907584,
 0.053372743411236545,0.08818856614878823,0.295463567874079,
 0.5414395799032862,0.24701440821998188,0.5869902715909378,
 0.35535065308945823,0.3021707374542265,0.4690924114017938,0.6885250563300275,
 0.7865734647421648,0.7243887490875636,0.5276837860067314,0.7516912654697863,
 0.455398701715098,0.2670538627949215,0.7063881874219683,0.7982865996741243,
 0.23806943718835427,0.9349068345605512,0.48045114807123335,
 0.4751214150208224,0.4658083299765441,0.9389958296342009,0.34348967610176706,
 0.6879328754110658,0.7850468148874912,0.5664688323342855,0.09521346361916594,
 0.8750579221466284,0.6320528195721303,0.6883178217320401,0.49371667653397877,
 0.21973263516222363,0.35760070861368165,0.42686735026017253,
 0.7609971975184827,0.1838775557966239,0.011566208843101888,
 0.6103935382146852,0.7736869887621327,0.17720461901411122,0.4294662186750341,
 0.9344333953427452,0.4329795487153141,0.834083662134305,0.9291157110473297,
 0.81753659580735,0.8455656129217088,0.7607280789254368,0.23800388079611312,
 0.5316150806319152,0.5899906922427364,0.08518820386617021,
 0.38737996104642813,0.6607274501292889,0.16095310645708405,
 0.8776691668392914,0.25058366874610893,0.4383213360095495,
 0.17877156452895193,0.6960067246646686,0.3735825896023035,0.4006584542587046,
 0.6042401964360725,0.3557608375390553,0.7225890264490326,0.722748699919475,
 0.4473012211715134,0.39317195302530616,0.6471092341892749,0.7054116808917484,
 0.5122379778008884,0.4961551816031494,0.5399545951136798,0.2656903031315193,
 0.7934392744564823,0.5945428398399335,0.7459224305871103,0.874448604531069,
 0.9360705231249983,0.06193868642103395,1.5747394994114483e-4,
 0.21624236990831047,0.931633661240651,0.14702473305879582,0.4824870241385665,
 0.656394889619318,0.39685802546731486,0.3816298472080829,0.45763958786627323,
 0.45736745719029215,0.6488264008255687,0.39175663940369354,
 0.7844225437113764,0.8467020984091552,0.09950254721852825,0.4017444526117041,
 0.8212905879835998,0.3356485949258041,0.28006944349845075,
 0.49860237136435104,0.050009702671706235,0.4311703854695078,
 0.649818657299522,0.7845239020487911,0.8919297645121408,0.7358803136043935,
 0.9774633480770424,0.06544118977194602,0.6023454781687371,
 0.018157698300997893,0.2395735654145772,0.6962692967258699,
 0.6192084036768966,0.3298454862442579,0.07324743946155265,
 0.32457870842865777,0.07644747643070571,0.3374563452464473,
 0.018297576191158016,0.8507497893484471,0.8938482026275877,
 0.9209168825261257,0.1798488347075674,0.5992730696067252,0.9476615773351471,
 0.8682182475380307,0.6215645806621093,0.3985953593280992,0.14912478344516267,
 0.707060225096555,0.21858200590334897,0.31901045733208133,0.6594183475012219,
 0.09005028173714014,0.7020545808829475,0.682986207929656,0.866884179429164,
 0.4138914549584771,0.10419262566324394,0.9160626401202583,
 0.10917031382821607,0.5582770643332701,0.2794660751887851,0.8365103173625948,
 0.3168777407942517,0.1728785314423663,0.5091597214858775,0.2390188712025979,
 0.34240393714411155,0.8155148535305011,0.6484786382959316,0.6217380202665796,
 0.1714339515976797,0.07498810712792181,0.2567356147881389,
 0.029608273458990753,0.41401748463751376,0.821460694331369,
 0.5131093938961199,0.07478139360796643,0.18558596395936977,0.643771090668733,
 0.6086046312588406,0.6467754387784063,0.9183547367776077,0.05080528408040941,
 0.7106649634106563,0.15326853818292563,0.20249866440887565,
 0.1515239942728479,0.8056228369372009,0.2052674106004344,0.06232641969480179,
 0.5730284023509108,0.9787624873786493,0.1964232808585913,0.2832057321340501,
 0.9698820814066166,0.14224405319380518,0.3911667576904584,0.8000275347633711,
 0.10012426959382659,0.11798528969335431,0.35637957435082657,
 0.8458109927159357,0.9941778480256261,0.5596515512199046,0.9513601174688465,
 0.456633857287574,0.32416051610484886,0.1337333517152801,0.4630294144190681,
 0.5094908026026527,0.1425541052860787,0.7205607622651208,0.04809035311949306,
 0.010324581047836179,0.13659596084603942,0.8192810524484775,
 0.10813165986906381,0.06397846195365764,0.310755711978815,0.9611339252508105,
 0.6703664778059166,0.6840804466148467,0.8348781585016788,0.8368593681328604,
 0.18388618655025202,0.6801717144912541,0.9774560686193262,0.878709287545063,
 0.20846318844338896,0.949822803718,0.45203428963654513,0.9255470847634292,
 0.6035680771347611,0.6172420768355428,0.37821821345722195,0.3883284080562198,
 0.120539842839258,0.8177116067598309,0.9722772484454986,0.15342505768873838,
 0.7952791944922869,0.9257131100993393,0.8443077299860993,0.07110248118726847,
 0.3236592889228914,0.017978381401364985,0.13454424624859085,
 0.8091660191772029,0.6131334785211966,0.5098131951580326,0.9927811801616068,
 0.7827933548392662,0.06885558367292899,0.40483836990029265,
 0.33146855496532557,0.6537416495498973,0.3675836545908092,0.8050554641307679,
 0.8469608638138109,0.9070644998331561,0.6427894258801792,0.5795710226688857,
 0.9085779474254714,0.2739365467096381,0.6501181436391652,0.7926292638539142,
 0.9273579034883944,0.8204611002858933,0.31405673076254437,0.7449922662597603,
 0.5027170555589826,0.30517663941799733,0.4799223002158747,0.887954346856678,
 0.007648799818655405,0.4134554108512808,0.0571550562884976,
 0.9125654243337429,0.8459515854598116,0.6841586740533762,0.8579997544700816,
 0.8572766775740469,0.11046103164827525,0.3880321289896871,0.5540139438191286,
 0.7546288285890577,0.7612878063661166,0.21381236729350528,
 0.34015001074850637,0.046805350407075896,0.9753694998163334,
 0.5346311803339441,0.43671478074099535,0.07748405691251836,
 0.8614902809451939,0.6808765116097384,0.6761616801865848,0.6826495125921487,
 0.5749162881616867,0.9921760309159012,0.5826152082274284,0.05101251975210863,
 0.7375176429563295,0.1679917057607031,0.4985881223316748,0.5762674056898276,
 0.86842403972942,0.7434154732058923,0.44083343950119747,0.8419914471664497,
 0.37883485890212487,0.17970269546211315,0.2123155904189854,
 0.4254378616093467,0.3467689362275621,0.9733971794718236,0.9412069340476287,
 0.10688493215831707,0.29381298686858104,0.18640970367272547,
 0.35056814963150995,0.513144703274252,0.3964475530692112,0.6557056823514227,
 0.8940285758615337,0.31065755574686627,0.5005489458977141,0.7650835791334315,
 0.5999955333213944,0.0661932943568303,0.18826229884437362,
 0.40528944974284875,0.47842067559301993,0.5034749781859845,
 0.18470986978257764,0.5251702047700877,0.7968626749306078,0.7839879143266111,
 0.751625205240875,0.2278911057995131,0.05700218593495743,0.47879822897895097,
 0.19720968455219157,0.587500293867473,0.6100067416040752,0.3460014788675052,
 0.9139427604584858,0.5891863779591835,0.9343636167529313,0.01771894909044891,
 0.9079526638470736,0.04142734252675573,0.461051009477619,0.32952348045558255,
 0.5528537954662396,0.737723277834846,0.6248957578701232,0.1630265485829504,
 0.576024013654172,0.027808479730923596,0.2716265204950923,0.4382959958011856,
 0.05679827048947428,0.337781915064857,0.5035603857905728,0.9643172504296098,
 0.3367772302788593,0.2348242083925104,0.9143515351811929,0.5802096773501983,
 0.35359896120444834,0.7455402030907351,0.8359932479388608,0.7187157729583487,
 0.1048265228506019,0.6800486399602191,0.1683114529836678,0.3099815439883209,
 0.24768268411196703,0.28200845217971837,0.6037084851081007,
 0.45494112332673553,0.9828288656289411,0.2716781340638261,
 0.21831214078997652,0.5879851981027844,0.48369269894987443,0.255485373421248,
 0.32990429429278634,0.18149038093755976,0.668444862305801,
 0.20476424767468515,0.5562184638296581,0.7663293281518301,0.7966299880912915,
 0.7216733217734022,0.5542230720230524,0.22995364812332408,0.7181145443455647,
 0.31697651004897265,0.6767969916700625,0.6460943950279208,0.1130223730386809,
 0.5266184369590142,0.22323668779107608,0.3829506178649521,0.7533672091415051,
 0.7010883411294021,0.9473354754404985,0.20837965296490846,0.5757930242899315,
 0.25215767247767784,0.07750981264836754,0.42533845302630857,
 0.9655202564253895,0.983714741301752,0.3668704558508348,0.7599919926108198,
 0.1892596336554434,0.06648218325401456,0.38830071357307405,
 0.27687999470150415,0.34813051008804696,0.24394703103279614,
 0.7384980969956672,0.9422069903829661,0.1358371008949274,0.39842372688547634,
 0.24645406851387786,0.5191359461564207,0.7902728203964997,0.9366866171296075,
 0.8039970376603657,0.20134778705937373,0.2566806400464571,0.5783824497225281,
 0.9643730721040856,0.3309531098624895,0.6963752177128537,0.07581948712791664,
 0.8773614405478822,0.18616670113674383,0.17650166619359253,
 0.7438239518129003,0.880493954027609,0.6815286938026668,0.69472922032381,
 0.1423734339538404,0.7850487734628687,0.3020085382723089,0.49285019056868407,
 0.6755530864162163,0.20232713775485112,0.5966430208248726,0.3175535454142244,
 0.626297161841896,0.6897403176440122,0.2616168322960776,0.9484355079500469,
 0.012134804065962013,0.6028682031366768,0.8364163893185531,
 0.12510207821352692,0.5142208237792943,0.32833906122986345,0.27549976491052,
 0.774116070768484,0.8879236093038085,0.2765576359036802,0.49099209418973333,
 0.2971655322466198,0.6639470316143801,0.9533930685526828,0.21846377186488497,
 0.012769560362412169,0.5769205425136984,0.9130491200467749,
 0.5825337528375683,0.5264060417007562,0.22938690901021974,0.6607342503739586,
 0.08087280132335728,0.10084313747451612,0.1658099367767738,
 0.04462337026245522,0.5342319128509294,0.6644758922818425,
 0.23008895943559626,0.020234221633424587,0.21536452831896846,
 0.7759746117032342,0.6358254224781322,0.8692362193857989,0.9178316818487106,
 0.6871420258767247,0.37488194133508357,0.9680677867112114,0.8924818045178221,
 0.6241477024326646,0.16798123903486162,0.19584826611958261,
 0.4366712825259277,0.8114509238714147,0.3131782599677415,0.2724057159024601,
 0.516734678323626,0.9401246204608714,0.371605274431308,0.6224689033917307,
 0.0070324350969642735,0.5926577222530425,0.2768390502010384,
 0.1707637877519389,0.9160145549941349,0.28748893163327915,0.3304701145422948,
 0.3836524645644658,0.041604288212086615,0.050742290475074836,
 0.6989692698783555,0.8246906050035409,0.5955208840780283,0.9910684844480687,
 0.9110216970025118,0.5249032051825894,0.016897807255511044,
 0.7330443844574845,0.4848576487361741,0.9617377492854483,0.9781330896724123,
 0.8119831965967826,0.43174107099157655,0.9512702470933716,0.8281307095266559,
 0.3566372049768295,0.637675567954819,0.08807181654850493,0.3951875946989566,
 0.8907424124855505,0.603816744488217,0.3916483643558798,0.5012942389895577,
 0.6422585780857903,0.7780902128903563,0.5822354111844096,0.14145711924897042,
 0.39177693516128,0.8342070012899064,0.9926376599591391,0.7794002508368943,
 0.8650550252606892,0.8342663125370448,0.6616268288680675,0.21848732976138152,
 0.9462570943457653,0.30941772644032783,0.49744750866584275,
 0.9231997122693005,0.7825406134056968,0.96129618590343,0.15287021112476162,
 0.7127169038974503,0.5116720078850753,0.1926754923594728,0.3221662070326572,
 0.5198723872640107,0.45495616762196467,0.4237375349210093,
 0.36260338329629194,0.38990076109548233,0.6211625900640455,
 0.0376818794688748,0.5708227797801468,0.646977200201206,0.6312868782804216,
 0.24842687756155757,0.38859127884020683,0.44857346790474717,
 0.35201627738640473,0.8534469905762857,0.8847446464391246,
 0.020850116211325487,0.9595676915601785,0.3941790183948407,
 0.5519690598128137,0.3327953341379426,0.5222514382755373,0.2795397227688694,
 0.017525868100668363,0.10361970761702843,0.768180389242113,
 0.10388987600009891,0.03026699649602571,0.48058204734681254,
 0.32743116402743644,0.4221842333666441,0.02605251068390113,
 0.5625821938400186,0.2285154301545128,0.38447210538258036,0.3752860097545946,
 0.3401596581462558,0.1742613777590818,0.18252582342036716,0.3305464745639296,
 0.19070916621485412,0.42740368525510564,0.3972425232294554,0.811885547293461,
 0.9631490429949289,0.17785643691755437,0.7867720975188112,
 0.024896531097718944,0.5418192891664965,0.36373988716871375,
 0.17534694836339526,0.10067271705434444,0.7151002341285815,
 0.7178318844668956,0.9345038225917786,0.8719945834312224,0.5378764267891788,
 0.9186146885584379,0.6843653936863523,0.424541347569496,0.8099544521345164,
 0.8442883734794401,0.9837092805338753,0.6809664657726278,0.6007775731695311,
 0.9183002616805774,0.626092810827308,0.4811329639789176,0.13299097168362983,
 0.9868659646579714,0.20086184260756157,0.43320222080877957,
 0.6617983126159905,0.37766372204677867,0.6927695795993607,0.992739452068811,
 0.6420733738208589,0.042293709268039104,0.865652562287197,0.8314329241801234,
 0.9093031644711953,0.7939306320892835,0.08249541100621682,0.0529809205338323,
 0.21891094570636294,0.9382690396898905,0.7701161512250722,
 0.10914210359313137,0.04448419959783878,0.5349099626216096,
 0.1676583463304242,0.26267528912623017,0.7270874136100248,0.1295944283766416,
 0.09656028670905104,0.17483449607985246,0.6356160849763944,
 0.9131690439602846,0.8112090958887002,0.48264740711664056,
 0.07080402901492278,0.5059580638018328,0.10886818490772143,
 0.6642576924000697,0.49692135134598603,0.5049348995129002,
 0.37575586996807153,0.07564010835112045,0.8461828177877674,
 0.8555287730869273,0.11186449462056136,0.5295496139038394,
 0.08756967265292581,0.968247660005819,0.9246741449958118,0.3369104054479436,
 0.6335745295785825,0.7933969803964125,0.017377247165029308,
 0.6342196507246415,0.25705502954192283,0.6946466340825487,
 0.45210586113115103,0.41678426803744817,0.8894491033071399,
 0.17881561412744407,0.9793902882254617,0.3270340051042956,
 0.04915407707365116,0.6484860368622579,0.32293600977245807,
 0.7882265000830883,0.6343471820388286,0.7750618164094419,0.42782738344397897,
 0.8053368862800931,0.767984599017022,0.5583194547150965,0.9644481826513345,
 0.6409005727543726,0.730979914138138,0.5967454100042368,0.8996324573083312,
 0.3125132887795944,0.4587190466235912,0.664757967937006,0.3646548733217444,
 0.9369157918105663,0.695901133278574,0.2635508564591713,0.9466939506489513,
 0.6347756795072701,0.37530739071085406,0.1733399345921558,0.4979590737839651,
 0.11705428357979653,0.6018340848175886,0.20314255739209575,
 0.9623264690031745,0.300852235920237,0.13247080352806118,0.5698527616110634,
 0.8700897444177627,0.6999795198205644,0.6238463761959163,0.8324917786769762,
 0.6579587617957497,0.14452791666862186,0.914914080596228,0.9941769069553822,
 0.77357704031806,0.3099870387013338,0.08288213275111733,0.40417368847259805,
 0.7565212203708884,0.5948504774225158,0.9513541627667228,0.04121699309079818,
 0.7679611488425515,0.4800732224679136,0.10013747401555373,
 0.04120454304281329,0.4153469696528771,0.926818915832361,0.06798724879239726,
 0.10929819044589428,0.48758153182208996,0.9641688983108424,
 0.33087491566921223,0.5442850445549329,0.7526835065415417,
 0.47123267358405074,0.7485857061110742,0.689509859421738,0.33801469166896714,
 0.7009945733617238,0.2142278727018625,0.1647841578308611,0.171788552551571,
 0.9818726804666337,0.8462196666602151,0.3732332179658647,0.23541847073407907,
 0.6635139075457288,0.6870709321152124,0.9102002067542472,0.29566790555906564,
 0.3051730557467701,0.2318681426356558,0.8929628735198127,0.5131764344699786,
 0.8441835053721825,0.02000069758546663,0.7521511023064594,
 0.14729204318502365,0.8547876719604047,0.6796080768561793,0.604964744363399,
 0.4094742152980301,0.07231345528058353,0.35594734220737245,
 0.33188730556098744,0.013231383999888102,0.46287734932960567,
 0.5740512571285897,0.9314541543737619,0.22414615341767608,
 0.17675249615971567,0.3885250310084871,0.9289029312262458,0.1802694655575272,
 0.6813790305968076,0.5543480465826897,0.44516709562116996,0.777067670740548,
 0.09809431411514746,0.8208192882562805,0.6969732991262256,0.8468871326522216,
 0.37291038311308156,0.5006256648024365,0.8241858690832617,0.3139321952819887,
 0.7318602313505831,0.6775819135372754,0.1327392132353683,0.8894000384545775,
 0.05876684741118132,0.5877521936578667,0.8056470020465796,0.7539279854099169,
 0.9395484196308503,0.7666797495376014,0.4682657133281316,0.6921841432480785,
 0.32823866191880724,0.36598783967716686,0.5552619966519897,
 0.8016369095180654,0.14044776373696555,0.2442229965013708,0.684802810157305,
 0.7740010816629281,0.7638617510581982,0.42314875714815736,0.6983718582957886,
 0.8106971187873916,0.567715342762989,0.9369571620798722,0.5043082837245297,
 0.8016014740217603,0.45290792360654064,0.7089797256185548,0.6852005046435977,
 0.43662135662644885,0.9605916391088136,0.4776238929101415,0.4963076236366588,
 0.07900250208777693,0.7135090300989485,0.47946545021813436,
 0.5221519397558412,0.07496246401725593,0.9738282698704036,0.8483420900393194,
 0.041663648283529975,0.7162978743793805,0.4088047106839442,
 0.7199591344896106,0.9746687957127051,0.7279564316005335,0.24999095808757354,
 0.13198835944631693,0.2861115310980966,0.7722349859309249,0.640510894646248,
 0.25856778677037817,0.4069176548694218,0.969366041269209,0.6062165002619562,
 0.6674976083133012,0.1687740992857485,0.8936469144417617,0.7793810758603058,
 0.5879162769507881,0.2963437222054608,0.17435638643163442,
 0.41201138407070326,0.18642162285644293,0.29464696694732784,
 0.5029854428173258,0.12377240569463999,0.5212108385465251,0.637210861401656,
 0.9599211906442985,0.2044818407097302,0.7666448997939279,0.8360088386791135,
 0.2125020858978971,0.9019992976167833,0.33674870246261024,0.2590185699841959,
 0.9079928970927325,0.4222814489702481,0.4857285602402167,0.4493351501719892,
 0.47343468203508543,0.38171404526617325,0.3064545339495287,
 0.44908749475475773,0.8916598601448484,0.5993203036233865,0.8689035408497839,
 0.16817857735966246,0.025965225723751484,0.2370902907314315,
 0.8904753754390934,0.360887394670768,0.07779245483889718,0.23228576867051898,
 0.18276803384440632,0.06638418137697688,0.9043254710230033,
 0.5147182341183079,0.917323055811186,0.31717787852331814,0.7338389129184667,
 0.9925235899754092,0.16637499401198164,0.02995833133169501,
 0.27338620980216544,0.8753490660179858,0.017817711791178636,
 0.5855294569185179,0.06377433813669264,0.8550491861389437,0.7083762018047649,
 0.7415831474262676,0.4554542855773429,0.5365904001983131,0.306922962037141,
 0.46560909918334203,0.2834356087116897,0.7489802864148737,0.3117550149569128,
 0.9573891719415077,0.16255811494320305,0.022997233336930267,
 0.8730009318253523,0.6958289074362084,0.417164258256682,0.09304942680226058,
 0.890091048405955,0.027732084607028584,0.11701104815252639,
 0.5054076804769416,0.8929757637300777,0.11898247325320677,0.6012780465360359,
 0.26129882162573015,0.4983353065430719,0.5822375080583952,0.5369216991340463,
 0.39788335614000736,0.23604646630573656,0.5766462401644755,
 0.8005888540487011,0.305349613240536,0.4393527747114627,0.5393880016399625,
 0.9477081170811691,0.9583707125274349,0.42855779950681194,0.4399398466584622,
 0.024481682225024448,0.578257343487511,0.987254484214821,0.2648664035138718,
 0.3966287441281432,0.8743903150946447,0.5911857335959225,0.7743157669976268,
 0.36720564643329157,0.8475820054503552,0.33372997526673975,
 0.4067904544376233,0.440601891634136,0.8566618857038688,0.2719049878853248,
 0.27124718355657673,0.10484574935635693,0.0213811117733137,
 0.6349936479051276,0.04994410661585491,0.6884975701621477,0.3022531670140559,
 0.19074056489534763,0.05426311250747273,0.7528602811200725,
 0.2873358737497693,0.36710003790558243,0.3785772473958946,
 0.32729571920957445,0.605812739263793,0.3682583859644102,0.6575316996627133,
 0.18786064921475454,0.5012947414303125,0.7760818810188683,0.6948475132809954,
 0.24284460286893006,0.45570721815901016,0.050068256776402253,
 0.7032770473220752,0.48636604005336637,0.9394379731686814,0.8202487283468682,
 0.6710960452926837,0.8355154381190966,0.8424171513729204,0.9487887780896668,
 0.8345922620867855,0.34396112181669014,0.6972326453317143,0.7190296039919567,
 0.1473007043359903,0.1632703702260092,0.8800255550338328,0.7552200756288139,
 0.8590565913915968,0.309819902194303,0.3397108392230296,0.11144083723343523,
 0.40848816340082794,0.44120913769407344,0.3328168128853082,
 0.4374439528807247,0.7538972296635873,0.773831090623281,0.29548634770583526,
 0.20779025140123242,0.9017192769474393,0.3915068767187453,0.6681273166083536,
 0.674873602062414,0.08376604641989371,0.4356117296175538,0.5482202785448758,
 0.10992191360786352,0.7935268334473465,0.3033586417650306,
 0.016863338251549176,0.5114140924574783,0.5984456937267812,0.651893565888777,
 0.8913795683097727,0.5513234616189739,0.34195377873998634,0.5444503412000217,
 0.464561001970644,0.6572647520456327,0.27078089594278654,0.308952229482081,
 0.8791911058860667,0.03880155444741418,0.48507828153666077,
 0.7220323325743753,0.6658166286952831,0.5228617005147357,0.39761226116525006,
 0.20755396952939253,0.7890330441415874,0.31114974945832574,
 0.07562685020907012,0.7692883631123705,0.3122474837590852,
 0.051705191614482615,0.05061552718609641,0.33049039925796764,
 0.5827992420145875,0.5801911711485528,0.2026476199284396,0.4543800084392253,
 0.15418946596134742,0.8761738268592485,0.6858828112665727,
 0.032431246377445655,0.28468945531220424,0.39877907744134444,
 0.36118088649391356,0.34295211481200827,0.05196338097618025,
 0.5919622581948302,0.7899018786557804,0.7513199412353745,0.418792388126645,
 0.13481733489660175,0.025349488747868354,0.8789809205347322,
 0.05415679591709788,0.8105658542235794,0.21663411669014643,
 0.39101769631914884,0.24098466746008385,0.5749223436721517,
 0.4765691836423662,0.08688735195474218,0.5208305443812904,0.2950588313498528,
 0.8639082184198771,0.14481164144536351,0.7949737526240618,0.4969693036484544,
 0.9855377547555447,0.039950839093177204,0.18038697561888928,
 0.49543843673823285,0.19739001738898432,0.7338545492989236,
 0.4906297969385356,0.7323271441594197,0.9453538582912329,0.8973693635677117,
 0.30715689300193494,0.27229026523723154,0.05772434021392481,
 0.49111887662775244,0.19248208642162767,0.6821543173052625,
 0.9888967970900101,0.9467789230657679,0.9844162775367368,0.254337128883233,
 0.2722486169273054,0.31868498774689213,0.43980875234503347,
 0.7177109757732918,0.5407516146736223,0.5219073864006686,0.6008910387655875,
 0.780014497382058,0.40666624862800616,0.48860695818187405,0.8012050756940859,
 0.20166324040863748,0.8818943119173254,0.22522442761311123,
 0.9924303527075484,0.44859366469459083,0.6661988261437225,0.5925972775168343,
 0.22219873413519875,0.9324647996518571,0.8171824360912155,0.6430811797550395,
 0.132128875784984,0.526805596584854,0.919379712953391,0.6997794624567963,
 0.05398141215864305,0.30113021089970327,0.18627219761489133,
 0.30502969390108314,0.10311028217613216,0.8108367746930699,
 0.13077765359509852,0.587394612761717,0.9668095750323858,
 0.020413840809115458,0.009434691861542999,0.16359191500939552,
 0.17686194582498893,0.3723595638925121,0.5760260940935273,
 0.40834260551874724,0.6609946265674497,0.9402986479312565,
 0.15190764662471867,0.9856752698860853,0.7724884038782593,
 0.15291471218443675,0.11015724880059352,0.27948021612567486,
 0.9621984490147455,0.9790857609662498,0.8331754406679429,0.2934694700475372,
 0.7558093968350713,0.5502190197999862,0.5100245153183968,0.3703198122511757,
 0.3626312087848277,0.7219646414027597,0.0876381675159521,0.9423663393546945,
 0.09634120907439703,0.24293493175055136,0.9935712650922482,
 0.8636036008492785,0.7865320928475633,0.7952109289823035,0.8368113771719835,
 0.2770005337300414,0.8980247812715512,0.10849938574834816,0.4573253543853676,
 0.8507650408336,0.8732792955709117,0.08399237449559427,0.0952674929993681,
 0.06963665126020402,0.6636983628144957,0.7760621504997952,0.5979689769077486,
 0.2970242382522318,0.7455678598198408,0.9059026543701105,0.8956715297289757,
 0.08071224697613211,0.4846325019442539,0.3595969397135902,
 0.16065327091426684,0.513824397412955,0.4994380577463994,0.5772659123530435,
 0.30098195205047984,0.4773178306422077,0.6849105134373458,
 0.43649480425148224,0.21493093774516447,0.6255284342085945,
 0.31018486846348736,0.5425224985125231,0.703444574931769,0.6048051818474909,
 0.043745419784798734,0.7527535014520498,0.0024703899032567644,
 0.9389492738574972,0.6196997719832271,0.2767913366928658,0.5932577715422497,
 0.5920083725171745,0.713062937327964,0.8227895431463071,0.6184291237673528,
 0.30183493836810804,0.4869447096839874,0.9819415131415212,
 0.0017119427712489754,0.016533347228722572,0.08520108323262465,
 0.20183331342168886,0.35193586178170544,0.12283367841130355,
 0.6357953680333962,0.8937689519337322,0.4424551801187011,0.279860740287454,
 0.3932556474778939,0.5210574038621403,0.12012919911711561,0.9369606692947189,
 0.25556707232787146,0.48144832185772723,0.011567394228370054,
 0.6598079397929892,0.28160113634320416,0.7864343102696281,0.6802447958351692,
 0.5169471591492336,0.08112686176338813,0.8946961794358151,
 0.47696997387579143,0.7126918993582799,0.8158385714510731,0.9843578954124442,
 0.7434745092383122,0.21381840157620258,0.5856332571445224,0.5593375199802446,
 0.40513997232807364,0.43981117281045556,0.8845224022218166,
 0.1616665889029053,0.47983437899776427,0.6315450145110165,
 0.21651589794524206,0.6684216991161281,0.07446725749514904,
 0.04695213287567801,0.10166148716841816,0.41449731419584834,
 0.7695123386578598,0.815141810330844,0.5119200506551345,0.4965450588283875,
 0.9356534798903979,0.7979083845353717,0.3405679867769429,0.5877008018940812,
 0.05278275718842351,0.7388055033340384,0.47531406468168236,
 0.7919537999855333,0.4736575641784784,0.5279902381251075,0.9687230567246319,
 0.08184504867945908,0.49130867681909096,0.006900921111276492,
 0.07322642879243313,0.41911550012670973,0.9807002121981534,0.725687010926776,
 0.5976014633384066,0.26650044216237934,0.0837416696049923,0.6042224498132385,
 0.41746950013939454,0.1768335199353963,0.07311626655222947,
 0.0491820418553921,0.7066222558559783,0.8237444755172078,0.855773243357295,
 0.036628247655034696,0.7347414680458034,0.8641493302645119,
 0.8856682524983167,0.34392804890881246,0.9561639966410852,
 0.013422240200698932,0.9385220494472215,0.8073199163496567,
 0.9910359396431214,0.7637994915267259,0.489253857758996,0.45535393234857224,
 0.8107800221009209,0.9966288373969727,0.8684046761020837,0.25586478995841855,
 0.3411744829154558,0.43757569210743896,0.01827766019618038,
 0.27340981680038234,0.12832729407466958,0.3693371255187624,
 0.33303127424490075,0.8807327168366375,0.22466924543601785,
 0.3717452720865505,0.8534196796619519,0.12640300611115574,0.5783559538694906,
 0.2300614706437185,0.1900873105519134,0.7557396400407826,0.12091389406375486,
 0.6582484570024876,0.04226928476045877,0.12706840845198486,
 0.1355367101844699,0.18500924227605875,0.4673222963728132,0.6911589098423376,
 0.3618349469380512,0.9138196292309687,0.6267424205248442,0.7128960430073004,
 0.8159177478950224,0.5298798610618851,0.5924242826214289,0.10867076852980684,
 0.09596642090172847,0.3690420226340225,0.8687964888680386,0.9060298368871759,
 0.01839299930355076,0.7407085866212021,0.6787259128450928,0.3556667588465151,
 0.952612100383417,0.1359893447326831,0.6291457778066233,0.708712778675589,
 0.14945461940044857,0.5101003095052441,0.212498722660463,0.6098482812274117,
 0.9618823767463661,0.9558442532627213,0.11852218163069472,0.3966528972630571,
 0.6352356405942453,0.2236071364492146,0.15364793967203594,
 0.04017727217184397,0.9602290012254161,0.3878663393673374,0.9580731932114852,
 0.47667780738974697,0.3843033880349874,0.6288850266914987,
 0.16250978087679635,0.4108344731566399,0.9589253386805888,0.9620349261663129,
 0.09038420707197314,0.7017172619224867,0.7369026881806243,0.6989807122276279,
 0.04002152294849359,0.7051732438347683,0.061473555916638944,
 0.7408092529247806,0.7012623032678698,0.6782382946614184,0.2759578187689635,
 0.34776005037713964,0.04747477557860724,0.5536303055657048,
 0.39641944876751667,0.1991368393906896,0.24991932209255618,
 0.6574740977267575,0.7876602684577045,0.305855384160753,0.993740975734631,
 0.07905917657252803,0.027032557913463817,0.8465453278878758,
 0.9223033346343582,0.9049474337948668,0.10587314541734483,0.742650166465926,
 0.20186565379542254,0.054095926163532204,0.8175990499252723,
 0.733539881768314,0.2511028350510739,0.533953404821654,0.765115748559662,
 0.18707722115365766,0.42304448493188085,0.7668420798824647,
 0.38345270486337757,0.3860904703037652,0.5808624417221132,0.34967238285763,
 0.037142538799643465,0.8413937596303067,0.9713605110875939,
 0.5046599163392023,0.5880820121665633,0.5375280892063139,0.5894295041536742,
 0.9937743768208106,0.6516293574551936,0.10758798223391297,0.715501138077263,
 0.9069038956934157,0.28559086814759904,0.9162047811051494,0.4601843604945338,
 0.8714446445891646,0.08980320185110546,0.535979879028163,0.36998185463076994,
 0.6883119610400451,0.042747161931078526,0.4082454961525598,
 0.3614060791831879,0.7756678492482011,0.60027739397102,0.8567041486573164,
 0.47541971746340894,0.991833052622801,0.2609481199919236,0.9042510420805308,
 0.5826461320564527,0.8736330835501487,0.7986311334875181,0.5570124213713837,
 0.9275507863868453,0.7321925619025872,0.2830137700591372,0.5129288565112593,
 0.4991726375270682,0.06789281835359984,2.553496950368217e-4,
 0.2761474129033523,0.18110672326842947,0.07373967618596378,
 0.07071503720001637,0.49640702876383413,0.8297184915494691,
 0.07175705041690761,0.48253609434846023,0.3209839149673732,
 0.06228763271759408,0.18591480820581308,0.6400877867330722,
 0.11528590645904846,0.7321792567062044,0.20070277123655067,
 0.5258180386328721,0.14894403928527622,0.3387698464920048,
 0.02764219039309268,0.6964119737643368,0.7568788493260787,0.0614737707292724,
 0.3537231167503898,0.37363986975585184,0.4317224635925192,
 0.10512881282286135,0.6329525837034308,0.5865826937189487,0.5017030891178886,
 0.024343730088943305,0.9497921255034462,0.91150262937807,0.32448321117311496,
 0.6927467024682455,0.930096471200688,0.3391236691943643,0.14558538499396034,
 0.9435572218961473,0.7426368976142359,0.15216013377116133,
 0.37392227877425754,0.6253395944537169,0.9155931053290685,
 0.30759997051813026,0.0988359899261646,0.6305740452164675,0.4505903598608838,
 0.9486199703418743,0.7753881478599806,0.6276773614240919,0.4954929642231276,
 0.8761074906215873,0.689236573007099,0.11954160783371237,0.9616116303930962,
 0.5809356022213636,0.18589763817555205,0.4763221551234468,
 0.13112132797122233,0.73938575079429,0.49381441684702043,0.8166270016823801,
 0.1947036585150156,0.11441664670160756,0.40753875901561065,
 0.15992138486061558,0.7550429794200353,0.9056250581908448,0.1269789900145195,
 0.8026320863406203,0.15846701413483433,0.13776628136929403,
 0.2446724870648428,0.0854660273197313,0.8327963281228505,0.7274490286798992,
 0.3035076250790367,0.4323015977793365,0.5135667813458344,0.6149378822492169,
 0.6840854527445502,0.5627950095944483,0.1294237855826873,0.6147406047601868,
 0.8167843414992398,0.48071677819533465,0.862767991495057,
 0.036184657967260225,0.2766561942498942,0.30139766117647593,
 0.5833868149953747,0.30877800267609967,0.2666179824585815,
 0.14221482611891578,0.871234127558447,0.041974434316363496,
 0.4822638871987037,0.5085999648289579,0.4019246908792953,0.8961240817271005,
 0.7970179085241552,0.4692056030754984,0.10811810615284456,0.956004931249877,
 0.36686359233512456,0.9908220857390672,0.6535017085508037,0.6736082461290208,
 0.2612977543517643,0.4466965360870996,0.29317223035894957,
 0.19566328143745904,0.19544607876812825,0.6296087209868566,
 0.49663754198949617,0.9615765733478739,0.8320605145224446,
 0.010328198299533575,0.7192072590986895,0.2270861845114306,
 0.12889199163157494,0.657111567635619,0.05340103842373589,0.323989049902617,
 0.999855841426755,0.1943822635907918,0.36216034220898896,0.8270018975301925,
 0.6556704544714974,0.7986400596664379,0.22318876256426645,
 0.22989984538421537,0.7589858131942503,0.20712775701436248,
 0.49296097218258306,0.7352398083987485,0.21470918617742885,
 0.6334977996887061,0.799639592097841,0.8335476198502175,0.3999425634698034,
 0.8626676163101756,0.7228728209966223,0.5886237867260541,0.7650506994135542,
 0.05965649981108889,0.6380056791214981,0.962324087354231,0.16682650196929794,
 0.007924628484422658,0.13253173023622988,0.5990887565942867,
 0.5623818194368729,0.4241864718698136,0.6769397864177712,0.5899986620436865,
 0.04217228424171593,0.3981599225878426,0.05230499967913205,
 0.12638613924685715,0.970293580569499,0.8101769634147741,0.08231375181174572,
 0.5427101106341224,0.39714473041120457,0.9687320550150662,0.1353493984365055,
 0.2826791617793455,0.5384088302023791,0.873588664183864,0.6079792948616047,
 0.22298324030417105,0.692216909452549,0.9908773105287402,0.24117665320679704,
 0.3475607012990547,0.9259836742398385,0.34788094630676825,0.4768505801070525,
 0.3699981802667156,0.9177494351033046,0.7392997369143776,0.4275771155155407,
 0.5296341388645428,0.9109207950914843,0.6134802843790969,0.26312086561349246,
 0.8803667193418007,0.5260216562375943,0.147226524426032,0.25813246744478313,
 0.5511219104611343,0.2694181831897996,0.6087429383299362,0.6420173884637366,
 0.4367346665293703,0.34847612544473383,0.723783091495683,0.977771642927659,
 0.20259930804193682,0.06001534235542616,0.6639493000116043,
 0.4644595411883783,0.3279973023294822,0.6086816599571521,0.8097610289117979,
 0.367159893764207,0.10335700309473783,0.9968021417475409,0.6593228178887535,
 0.7948564413599263,0.5661821407156502,0.8404548991426118,0.30221307799905706,
 0.9212310512974409,0.13727894841703714,0.6767880204425367,0.7222984438039101,
 0.4465144508140192,0.5007598226691875,0.4329084647025495,0.27952852533584127,
 0.09904220036277045,0.3281798028456474,0.5301795417456132,
 0.35709201453540507,0.8562090030732854,0.17773224364690754,0.856106330433741,
 0.2578579070100373,0.77216253383387,0.510959934440142,0.17008764050169645,
 0.11619241909566447,0.30425173802560346,0.6908034281455484,
 0.3097254732695762,0.7641736571201742,0.941402042472145,0.2547642960924641,
 0.5579066183457723,0.25726838207336267,0.699618066696222,0.960356199219465,
 0.44966676380226145,0.30513718082518937,0.3183442050359476,
 0.5653267619558346,0.7996946810377957,0.859422011697854,0.6943098496882931,
 0.2702139306186139,0.14499963626206236,0.38123570472072776,
 0.15921142559755053,0.04361881280818314,0.7682176594378488,
 0.08198018526974116,0.6831336674561374,0.8264040530522605,0.8050331452215373,
 0.3502822725468816,0.4764654340719827,0.7331831572576755,0.7122750364212189,
 0.008998591940312606,0.650903935581869,0.873176493642237,0.12905461079913183,
 0.13225109860918827,0.8176367819208274,0.5317479629546438,0.3765378498771016,
 0.30627503576080406,0.31151945817822446,0.6954657259674104,
 0.25188420205778383,0.6855798241230695,0.20360067160112838,
 0.3311077677038481,0.7980629953275682,0.5562050452041151,0.02481924862307805,
 0.317647744485575,0.6713946412890341,0.42961095568251184,0.0801864761256863,
 0.7148548902006774,0.9260234485835183,0.6247096429959125,0.4454037555381877,
 0.25924504050681674,0.1447390631113804,0.46751788189723764,
 0.7156657566423634,0.06990284129984436,0.639260985971344,0.7527563245084763,
 0.2764390365576269,0.1857634827354676,0.24702213942576123,0.8003923595533001,
 0.7162265544075981,0.4256431696025288,0.9841379107047512,0.9277618103923022,
 0.7101219207138841,0.24212930300629798,0.9440697832414223,0.7910527435554856,
 0.24783684111735416,0.21669126675551076,0.46485395301647003,
 0.6373239064257741,0.5045202806979362,0.4639546090920229,0.3059024685737547,
 0.20570135244987675,0.17365559663330377,0.33747723134997976,
 0.6833703602324105,0.8358100824687826,0.7337966672115004,0.9067819880579557,
 0.5737815751958593,0.4146996574957005,0.8035700217850308,0.9561900446838156,
 0.6294635376029403,0.30475795190992094,0.32336924914465304,
 0.40221987993819763,0.5809107741583869,0.7935388904702338,0.8430350352562717,
 0.4329325800120096,0.4237034157348799,0.3480885071440515,0.2710651295900719,
 0.19802082127701293,0.6836159986201835,0.0733265069202651,
 0.14701951616574604,0.7953647186268795,0.9115626101631796,0.9321549761083388,
 0.31887465274976035,0.7003486953477678,0.8657826039729359,0.695044899231621,
 0.6589955931947312,0.6683512668787731,0.35832571687361336,
 0.15200939589738227,0.27187338413901696,0.5204846628279794,
 0.7369969610381609,0.5919380925390114,0.7122480079850616,0.8022402724195622,
 0.8914791238192993,0.2406802890094245,0.5908559618148014,0.06549581676113803,
 0.6557956644211975,0.277232361428451,0.009208921248795354,0.9804668623017828,
 0.5577801916474855,0.5486054831794462,0.8879980108916263,
 0.015218338713922641,0.9750186511972555,0.43383331136116743,
 0.5584355306659395,0.32871442260964445,0.155047258627664,0.4722084223262907,
 0.376100969721344,0.8821848279575991,0.6890091097406917,0.19360283599375339,
 0.2831084475772435,0.5718131252541732,0.6768148837155251,0.9918104596034009,
 9.071793310641008e-4,0.3350098493795871,0.14691405028778592,
 0.45816006817936106,0.7912278164480528,0.38498949317999265,
 0.17370044191078748,0.15455092055286457,0.5620142751747321,
 0.9223845017425799,0.2193328353607985,0.5442770639877156,0.2326165073048333,
 0.8467654322195926,0.5900222978217065,0.7338903157286958,0.37136077216047825,
 0.6064964930797198,0.39086375299743625,0.33356462133465414,
 0.030951881655015168,0.22954331324310528,0.015833470915353143,
 0.6275693243837662,0.9690751516299162,0.8366664736198688,0.2559701169130333,
 0.4666876258705813,0.05505956789798094,0.810939381600244,0.05436606352319384,
 0.768210970012667,0.09624539006550303,0.9883326106935083,0.9769470629862622,
 0.760990548749494,0.758317833270202,0.4837039635195306,0.21053520839536521,
 0.0216668093209903,0.36677754204031465,0.20720360549688888,
 0.5377635299013078,0.34241335919054317,0.15187824726064247,
 0.5089624109619139,0.8636899654772328,0.9731999149311666,0.02728196738545763,
 0.42066286482134685,0.8859741603776035,0.8820485805770336,0.9760004979055437,
 0.09944573333735662,0.01160493535014373,0.7551045625895668,0.956359182601411,
 0.19122409196833867,0.9618809316550261,0.35395624964100414,
 0.6307838871391731,0.8063180341411966,0.966115461208382,0.35093192745439816,
 0.9761229637289146,0.7736447202424833,0.17412044437330854,0.9554196069172036,
 0.6006320267271181,0.15754635307800047,0.03696687938561771,
 0.18975334143394673,0.04303551719079313,0.6858613736159299,0.518828715034632,
 0.4176895752832387,0.2679150855803385,0.6338068071022358,0.2512737843130526,
 0.0979616338799365,0.09011482384308156,0.8112274384106074,0.8346288738589809,
 0.946350380470669,0.6811423123199852,0.7192055623784828,0.7782142632854957,
 0.01928970496136495,0.8249074299044922,0.09993471285827948,
 0.9784515028438094,0.4442017871791245,0.26372594014462825,0.9484227816952248,
 0.33694869408845163,0.661565575829361,0.8126956660146479,0.3138785363661325,
 0.7509790372628196,0.26596869199369977,0.6096625732506513,
 0.011782109904981741,0.6335776479864582,0.8281850240203376,
 0.3898997660126191,0.6941271712475727,0.38632300524955254,0.4713802509785261,
 0.5554010681129486,0.1914102373388361,0.21104790592365152,0.2122003576677609,
 0.3962134651983955,0.6647211033659159,0.37843634111592617,0.5876618875364749,
 0.10361372675727232,0.8699989510892302,0.08947374613615189,
 0.8241616164251035,0.2965713189099354,0.33626890840354984,0.716341117090769,
 0.4088159893801072,0.6672504258627001,0.36264289930035876,
 0.07021717111491577,0.7656210931093391,0.18845141174794755,
 0.045517212275522745,0.802201510247395,0.39044928036318727,
 0.9246610006338762,0.2792412798239322,0.3741468113010362,0.5419997970526318,
 0.41976037430005986,0.7522495473124661,0.444465804103535,0.6810539935680433,
 0.4626298300336614,0.688761426862019,0.28066991266332675,0.5432046907453911,
 0.18997857378057748,0.3113867239971334,0.13362356999484037,
 0.22940425206493487,0.554781921265078,0.45639603368877557,0.6635805629144886,
 0.1403337568718449,0.47259397857428753,0.9213167742264633,
 0.42697090009198024,0.21674833853214248,0.8135374236743096,
 0.06203967870190574,0.4450474151685073,0.5374747544925179,0.960099108408848,
 0.5060532719977963,0.13393512037842747,0.9727919831418936,0.786896733965441,
 0.7351781744027402,0.5667405711742675,0.2127432696247621,0.4906609780269494,
 0.49899633694986434,0.334395265645674,0.4700658181141042,0.08738417396971887,
 0.8931416283475796,0.9515813946212657,0.2883206244029921,0.9826003487980184,
 0.897555736334491,0.27450073975208555,0.03180393292727324,0.4190382044740971,
 0.04032626817712359,0.7487074388790658,0.8592400937689568,0.5062376350715966,
 0.3647838854804891,0.4970428804826482,0.4851526874379196,0.5120029184054654,
 0.9737058881916573,0.4170419470015464,0.7370557955590339,0.8559269551066517,
 0.9962327139953825,0.5329521870144108,0.9477055772609165,0.4255547972964472,
 0.18700587524273138,0.019627721115835106,0.38516718817283935,
 0.23826859616745943,0.01918426143361074,0.1704503086022532,
 0.5320154380745366,0.2965268050523199,0.21794250296089368,0.48091605130004,
 0.47233722034109404,0.05116722365781701,0.3413662063214645,
 0.8090301147088547,0.2103893799263219,0.9865442111297261,0.2322518073834412,
 0.4216700650868954,0.0480034689973049,0.4005876056286901,
 0.005573398528770812,0.3857829088136664,0.1803435407594538,
 0.9606102432070664,0.38147526592342484,0.5665803888745199,
 0.23176176335198573,0.43438480400418844,0.9865651504924936,
 0.7158816594604623,0.40233105846097483,0.9814669837747193,0.8031760558165534,
 0.6685389643694114,0.6593059782346862,0.8060556633100898,0.6751059842904374,
 0.37863427893131496,0.4025208243105951,0.860960305201816,0.5413687308006979,
 0.9582076689296498,0.2600092867711101,0.03320915208653297,
 0.36561315828896257,0.2676993159939366,0.07834824067124357,0.462679070812106,
 0.10954373056573119,0.8912402188726516,0.7796026579993979,0.8703246243663576,
 0.5663527820911309,0.24304401682359367,0.5756471677333596,
 0.03408362626145289,0.43914754958030255,0.9378973213284634,0.999300000775424,
 0.8984295649115804,0.6234515116004056,0.16616545418996942,
 0.07297464688847455,0.31669643888199617,0.2811950533533567,
 0.8893820018046062,0.013375760328429465,0.07033165668873098,
 0.7889168017248808,0.7388072496850875,0.20581958209934093,0.4507748196708553,
 0.6961810255782857,0.21208821622331098,0.7182213326935203,
 0.23324827538162263,0.08318085675638032,0.3131817600961938,
 0.9810115282370829,0.10543610965184025,0.7661641016544045,
 0.38381486853425517,0.10042293209200237,0.174948795390088,0.5179352037844972,
 0.21466597476524774,0.6161181184080885,0.09671360457234357,
 0.047183509535742574,0.7441723764465352,0.30983676675913985,
 0.4830591494211691,0.0818995484623497,0.7542551877000446,0.99046171338827,
 0.09424045300265249,0.24519510074725337,0.4071504570705784,
 0.1854809069212593,0.4386509813689259,0.493020016607846,0.5103864678201503,
 0.7913964519276412,0.4150282594589163,0.1424730217991128,0.13497567421834844,
 0.15215346225240567,0.6305893573262087,0.5990392995604612,
 0.47364477486314416,0.8241783559870345,0.7598218490767868,0.7802658080399016,
 0.33110347029966747,0.28377906163255695,0.5732759953015947,
 0.7044703320106563,0.9253758150338538,0.194750580375471,0.47659754658765874,
 0.15283708925807749,0.8398674819296725,0.056223098888614875,
 0.6481589745203749,0.4855573548852098,0.6163988015743278,0.9108705805187028,
 0.9674054012941794,0.7195265046242042,0.46959703710451084,
 0.052123715556614436,0.9237309639295932,0.9546943418888985,
 0.4914358410948718,0.7427075374238109,0.6663600844881084,0.27224735984822224,
 0.018612627137075788,0.3392983556130893,0.8885416280265617,
 0.47581761650673804,0.7666420433999784,0.017011010074737487,
 0.4240813971777928,0.3310426442231531,0.7235209207483967,0.11164441569340289,
 0.45636874634713065,0.043616546512251775,0.4073540208919947,
 0.24949661783742016,0.6651512911445482,0.22929050184889677,
 0.5532978564568946,0.709835578820134,0.8958729533696628,0.03433965153967178,
 0.8947417270029022,0.47969091019315147,0.2876552303739305,0.7944446406659192,
 0.3553315236284271,0.25356990474632,0.6346576972408821,0.6378110294870233,
 0.4241383017445114,0.49302760244752253,0.6437565875189523,
 0.06118903863186742,0.3496760200661606,0.6868061909767983,
 0.35864122318890823,0.8513874853408091,0.003095487145699738,
 0.44377511639161127,0.5455920071134404,0.7299934166691461,
 0.45700343512233277,0.14145407476602045,0.5980247799148433,
 0.22791781288245938,0.08062947516774654,0.6672058204665656,
 0.5242522022407871,0.159291682912319,0.5291919936422238,0.917274804710136,
 0.8135619731101454,0.17686511239974256,0.0030793404635164334,
 0.3037868114813649,0.7671929665294039,0.15908558536651474,0.3466197120713115,
 0.7678956211101328,0.4362540194775212,0.849824683965584,0.9479230905247388,
 0.39061690626494716,0.25459410335584554,0.68610478123225,0.4177116766334472,
 0.029483941164190064,0.6422098371230522,0.7500489024588898,
 0.31859227204768115,0.09038410611791103,0.57400536652481,0.6740042109373878,
 0.25010721002453273,0.6611227846359522,0.8869280731727861,
 0.016586673911648142,0.7310142757367933,0.7313340620500982,0.443213186686909,
 0.27469014214067244,0.3577124096772313,0.6643892673762298,0.8130036476867983,
 0.12050823037243041,0.7705970370225528,0.4618816376260273,0.9287735920379007,
 0.923855225118404,0.5590690009636246,0.6953580167497214,0.7590176286908934,
 0.08263619189129767,0.5122251700566977,0.45059687443442675,
 0.8791844273657184,0.9346008758927231,0.9421354737206818,0.5975170166795799,
 0.8269708927191434,0.5483214281131181,0.22369870994145202,0.5739289773460423,
 0.08359806428978911,0.9375991959541047,0.4908408202147797,0.4754374714873837,
 0.8222183016272675,0.166843043431964,0.8178389925786534,0.7586526171017836,
 0.33312326819805715,0.7586590719454476,0.2716938805732999,
 0.48563678007759614,0.0954047304805975,0.5004177727866448,
 0.16191858021370575,0.4945186825096344,0.44558778091784124,
 0.5085272795256804,0.8115541574329124,0.187701670310236,0.9313046370240856,
 0.07276933279951858,0.9106127545433355,0.05401259363558719,
 0.7782339848392473,0.11020490509198533,0.16439199756260403,
 0.34911389837329077,0.09000966127456589,0.9581670203622008,0.411102445715129,
 0.02396238016976915,0.0062844167256530525,0.3822227063069834,
 0.6483338388140714,0.485151428125712,0.8027198952997454,0.42228612828728695,
 0.5299313636151612,0.17234293384855182,0.1809500826177497,0.627994927353765,
 0.6943641478690502,0.04182133925245057,0.30504610136324717,
 0.6974656660836441,0.10386765034184009,0.1020511122076091,0.5218332889118261,
 0.5783382022735882,0.6107655123530105,0.4093894376705264,0.8368269255711932,
 0.924834107528404,0.4916313577514384,0.48877469195926837,0.587439036020496,
 0.8456780362666468,0.5329113240234586,0.13275091051808374,
 0.03673947461073401,0.5599901794625207,0.9571130957660038,0.9904326740227679,
 0.5355113126825081,0.19327323856131873,0.8568555554427518,0.9414691120337487,
 0.31606910423280277,0.03423082478442674,0.32686531363129234,
 0.18354437456737238,0.16554956989942926,0.8200114351626738,
 0.4240689018856856,0.40496165731548406,0.8266328617618094,0.4741541249784341,
 0.31801515971986727,0.5456697001136858,0.4628920603711365,0.5928186784120684,
 0.9939595250863494,0.7623658587811842,0.3065053425430845,0.8705170910965753,
 0.8562939502368002,0.4334279413030556,0.9839658338411498,0.2892296313205077,
 0.7475528427205156,0.054111388718004694,0.39190865285805754,
 0.844675553677824,0.37915920841574857,0.8293764030762394,0.42735995791087766,
 0.6198245904840135,0.8161245987906778,0.5586993021967891,0.64556676715019,
 0.7896851773944471,0.27208668831977967,0.7211876571710568,0.8443172625914518,
 0.9965593763153672,0.44789610491248144,0.340817016685786,0.6040502839109618,
 0.25261621523580735,0.1834312292946716,0.017985717030107762,
 0.5903428223732401,0.7498344884014178,0.3293023008097711,0.6751923031595085,
 0.8697064387600328,0.03826885792524726,0.3185999022351632,0.7349955714214877,
 0.6250398534005664,0.5288206130582078,0.5109077892243221,0.1712021715929415,
 0.9111522333601633,0.7082480712442649,0.7026187342032234,0.7659634204371466,
 0.9645350910068208,0.9034259162742112,0.40889599589477443,0.6677273633208769,
 0.05785382065580125,0.801725640329981,0.8421277744685329,0.5894603870322,
 0.5194175066477904,0.5725944375402591,0.9370771502011274,0.4342481225064131,
 0.48421764550051183,0.5342314472081104,0.4861825228963429,
 0.32163011545779374,0.010672424036919481,0.11555294711772501,
 0.9225174456450734,0.17717610108916115,0.5436901362026811,0.5713392218809671,
 0.22850147782778274,0.6535886154271479,0.29169002929284626,
 0.6935807340695249,0.002924745773372317,0.4338430524673259,
 0.6566826269332178,0.28401410255320636,0.8518133493201635,0.659742871593139,
 0.8778597948381819,0.8859747105489737,0.04560897597101965,
 0.31360203002594533,0.13916874865707052,0.3100079123883299,
 0.5974197593034335,0.32319371640096584,0.643783792541625,
 0.004796275522594984,0.5229216143237607,0.5084777084487873,
 0.06829717098479549,0.3844449176170083,0.26919883221426943,
 0.5155431063855784,0.1384946317286031,0.7219926369012849,0.6163759808872744,
 0.7801971181514736,0.30033925126244,0.8323816656416272,0.3349986438798923,
 0.7965848939205405,0.18526300464479628,0.7781507240855373,0.8131889369060471,
 0.7148101718135663,0.24028407632605597,0.41457901568748246,
 0.22827219597020598,0.6192050985610226,0.620456560707623,0.8199111258878804,
 0.10397790743437652,0.23709281348614486,0.3530958625486026,
 0.41941843856783967,0.3393170225187462,0.6265951642451963,0.1984190426774658,
 0.18621512142033214,0.04099979639864171,0.263222215614773,
 0.38859984873690245,0.7342310531889034,0.9076096187368922,0.2839851886881589,
 0.5598024270818753,0.3988516679984423,0.8221840264015636,0.7016087120485213,
 0.9457386562050333,0.5732385169326442,0.7164945486832321,0.9637159876901074,
 0.35260680283152535,0.86717298743802,0.9833699943508186,0.7397079068146927,
 0.5321483354213676,0.46285560503059386,0.5149193912822624,
 0.29148582189277683,0.07804299092056988,0.07580225175243205,
 0.5692228736821053,0.8924591761428577,0.9437703122314223,0.5873364421795009,
 0.8997335294894954,0.14381813147922784,0.9024074298151846,0.9873838484443036,
 0.22014360477664463,0.5820882381543186,0.8632809912479964,0.8988428416020295,
 0.1460072485706524,0.18811067386590774,0.6063996891425355,0.9065863574349143,
 0.8899366338792667,0.6967100060544784,0.49772267265992864,0.2604588240691401,
 0.8851272673534007,0.5758374040139029,0.7715022561598336,0.18650841837195775,
 0.26944700017088363,0.9840187562924831,0.9990897154582261,0.9654098258465978,
 0.07833985492770279,0.09087756133208258,0.3996352147423361,
 0.22220721255847165,0.0387639837753776,0.43404241347568095,0.945280865214118,
 0.42148398246185803,0.9446880162933631,0.45585435614840586,
 0.3825265463938876,0.782073159096973,0.44391502472112154,0.4167518842913772,
 0.6963516729258715,0.3111682428443979,0.3578080239529807,0.04699574241916282,
 0.7322969537937325,0.07025882195101119,0.09073219546690614,
 0.5424372401003614,0.35378745891417884,0.136358815074062,0.4876258477354568,
 0.22502673419607233,0.7336456310393088,0.9991859515532436,0.8170267048117906,
 0.8293081889349132,0.5457198631372485,0.2446547365666648,0.22032948509899752,
 0.7164271782125516,0.31027284709753766,0.8128050554286023,0.5561001720096421,
 0.2305489579815,0.4967728490329577,0.5045222987986415,0.1374260765468046,
 0.9975524520643858,0.6740187796973705,0.2115719707586532,0.9312598530433057,
 0.45534341092412545,0.7033133725558689,0.617918404816243,0.26704490499621747,
 0.22511811006313742,0.1797631850048126,0.9846623443214626,
 0.13161143104250173,0.9879554635572306,0.34516655270994023,
 0.9347912387329822,0.8110148046014498,0.14348166566766107,0.9200136202042535,
 0.3547231881652303,0.8130657172497798,0.6554773677890877,0.27327725742684084,
 0.5009863846275516,0.919549275460318,0.9837956710781529,0.932739349486602,
 0.33354486918066395,0.8041691403735511,0.37927300562780264,
 0.7016490155831626,0.848944376268262,0.7446473591683191,0.39842405472634823,
 0.7253252804573674,0.42719295093184717,0.7571124824878207,0.4347464095905673,
 0.5649398781711307,0.237844899934183,0.8001795096742363,0.613949688737588,
 0.1164264866099094,0.5065915543252126,0.48558346214517834,
 0.03512105124115361,0.759396357830641,0.18284624249357662,0.7428559198031389,
 0.5162392488268381,0.22967650790993543,0.9580606381254816,
 0.28685543071987896,0.9383631431469429,0.822048382047477,0.9886571421510613,
 0.5694314186533096,0.5667215328027628,0.45467953968508035,0.6765162203296795,
 0.320512722387415,0.5606548015413243,0.41780224285059586,0.7629823833481276,
 0.7096983218732272,0.3529546860466277,0.2465212170606984,0.10793484240011364,
 0.9583970277033679,0.9616771308740075,0.8031972794316508,0.0639576198367382,
 0.4001908518596211,0.13991121933127593,0.7743301464606841,0.0592276161476365,
 0.7149230700078795,0.02878910704872939,0.44610907071704475,
 0.19403685420598138,0.6473421412690007,0.2677697229062054,0.5573131659217431,
 0.5515274542435377,0.5770452211357814,0.6766556078092177,0.07506740296503145,
 0.6674275137638848,0.2405006262598791,0.8572724186622827,0.9718041407561978,
 0.6706109544959151,0.5914562181407038,0.5838725839749634,0.3051800766970306,
 0.09204638191977077,0.9171444106412856,0.31567722433370193,
 0.6884614006276932,0.23754472443884,0.7585376697092832,0.06250171715041775,
 0.7492347039959002,0.7560620427302567,0.4819182034891387,0.06955874307784726,
 0.4295287613737305,0.47148827260103054,0.4624169255324311,
 0.22814536768533777,0.35874801810201573,0.3813365357676525,0.428706113502655,
 0.2775679907789419,0.38063936964270784,0.48046351143066535,0.945090456152419,
 0.41333924873317085,0.3369884945949092,0.9163852215039552,
 0.17574190218055008,0.4677124115816218,0.9709098646679285,
 0.08672959644526834,0.37012369873116735,0.16906586095338771,
 0.14097844431836515,0.4794141456897767,0.2917134948289677,0.8608975701226858,
 0.20374638169001713,0.5735463005717354,0.536846982546997,0.4430694937809373,
 0.24287295032438916,0.961426751493777,0.6566114399161113,0.7975666798218624,
 0.6400974041722951,0.13410918595092058,0.8751194885068424,0.9525716435139322,
 0.0405536423014472,0.27752061959202723,0.009166767073398363,
 0.02184165333050747,0.43728891122707314,0.9390562809815542,
 0.6293165164381465,0.9911237612771424,0.8567852197931858,0.0656994512642406,
 0.5545450417875174,0.9033915390014535,0.131309627830152,0.27028121478945355,
 0.38522687657647625,0.3485065270797135,0.9978369399388073,
 0.10984589169310055,0.5769620720044676,0.09371888370548975,
 0.1415793678388857,0.8091591809623215,0.19393558497405483,0.5935898413972778,
 0.9200379374860014,0.9373267409741349,0.21315279882762384,
 0.48917724379686633,0.27947471675412217,0.46229855804332065,
 0.2640544379936076,0.3513094380682792,0.47775543714679347,0.9181651945502232,
 0.4735908755548953,0.3785763799097587,0.5133223802852963,0.48194738949809923,
 0.9898950603480487,0.9533018365386764,0.5614164309730936,0.2821143175555786,
 0.19676429517070682,0.2684231733678669,0.08361992011789687,
 0.5686064492574295,0.16682675108064648,0.34562573402601005,
 0.30207073097515746,0.8240322456028528,0.44609771106764484,0.808381972348773,
 0.18925925263749166,0.5062124574810458,0.7326366763823479,0.9729844344028901,
 0.725298864829263,0.35773825015015726,0.5530106803843615,0.733826602478615,
 0.3717132690410312,0.04095203587216778,0.7297033416175628,0.7318071451270217,
 0.5440521291308893,0.917445122489964,0.3604127879547525,0.7237359530970953,
 0.3519289181206583,0.49636910607643614,0.2693243284120279,0.7983922249539642,
 0.10972521776672406,0.03686284695181874,0.2804948130363496,
 0.9370786058251319,0.7517988663910559,0.5776926677404938,0.8245527172051542,
 0.4230986490628521,0.8887849887334446,0.6192029202263271,0.14527515674623326,
 0.3272912990331568,0.04716541297479804,0.07336682611367773,0.996422578968873,
 0.7706507281292152,0.8292062988207334,0.3462810931447855,0.9957743434821462,
 0.7570149213986591,0.40020405051056784,0.6090265397787382,0.5481897882993901,
 0.35013213311173386,0.9165427015722092,0.4909488972608844,0.7461805315332795,
 0.9375239630023371,0.08875494768293757,0.9355135026306411,
 0.49638800224844415,0.4636597745991893,0.5233173654762554,0.8651484322933962,
 0.13003080888012974,0.3635098856336695,0.9855658977174567,0.6298023900615572,
 0.5548989981324504,0.15968146503835434,0.4537729450701168,0.621783748830282,
 0.32671511975452416,0.39088112954914067,0.44283259662489016,
 0.771297575436598,0.16146585298044303,0.12548950282092963,0.4477925218441722,
 0.894800702846669,0.7119178013412304,0.49516812909098207,0.26332516440900144,
 0.22076952227123503,0.6345526966083784,0.3615858504832745,
 0.08130957202384526,0.8838308024313626,0.5225914539578749,0.7968166738216886,
 0.20834600408644233,0.20595085343180142,0.19416342212885918,
 0.44565143178686173,0.18317223969300578,0.043851622724316974,
 0.11893289983729294,0.7236159514338544,0.711013368905823,0.10578241076270245,
 0.6425348403096598,0.7022036350101186,0.5108783257587166,0.8177748706165263,
 0.5122526962284076,0.8107867892671048,0.2112563424565932,0.2679120171916537,
 0.8872211250850698,0.6059850315256687,0.002082335806502522,
 0.8660317115495989,0.4711616174908865,0.8644422989190954,0.3761271566233866,
 0.37270481664796806,0.10370405426769302,0.7546248187654034,
 0.08055430561633492,0.8983923901922177,0.5602829473384856,0.6741873906886086,
 0.4590251562780424,0.7952272733500754,0.6509604361921052,0.7697942925103581,
 0.8423690611761241,0.4922353565887453,0.9916965320582635,0.3910507186944745,
 0.7519407349469667,0.6929127877505687,0.08602043321283204,0.9777407372684387,
 0.8315692016012335,0.17720917990477525,0.6613690697310952,0.5746955231883318,
 0.6272415198083919,0.5564888230317366,0.3090946372608082,0.1897507895801942,
 0.6175691731319338,0.47989083204863325,0.19162458224693446,
 0.012941574696280878,0.18371092785929077,0.3383483058840079,
 0.3726790537410273,0.6206617060813607,0.11101978265610479,
 0.25962735857479824,0.4293204154272049,0.2396594419054554,
 0.14937723524102187,0.9154534186666664,0.26072888165838926,
 0.09384342123739753,0.4433393042175342,0.46043178284905806,
 0.12538037530965784,0.16645228814208812,0.1609603528235959,
 0.9010316016922023,0.7948885239631401,0.047791283659931905,
 0.10310308641015864,0.2918530604231756,0.2677523587951156,
 0.12267883851999428,0.4637641390734045,0.7163849856986456,0.816538291691759,
 0.3707100695449813,0.690992501009637,0.1053513428131394,0.2460972135698678,
 0.8576262242471857,0.5032963995997836,0.928052828437724,0.5502767494000269,
 0.22571996985601173,0.0805196608191886,0.10689587319564486,
 0.8701938730626221,0.0048777099303000115,0.5994426445933623,
 0.8377711936611004,0.4041208705853062,0.6078642563904567,0.2468978472573926,
 0.5627071417690961,0.5443019648910421,0.6308700881273733,0.6577214956897438,
 0.34181702666286784,0.8822398671689119,0.11368788593199863,
 0.29480404984181485,0.1680243689974077,0.1444049380839605,0.7612327271223547,
 0.846343256726423,0.2100280534635096,0.0730848174296479,0.8063157958566773,
 0.26304878809376897,0.48006338298041695,0.7317149689441154,
 0.9395174858082752,0.766478390651019,0.3731155813336695,0.562529132152118,
 0.7675459536071808,0.15711057080023372,0.7898808829576345,
 0.39672996874235567,0.9531598726573421,0.6455358434042586,0.9700496295216982,
 0.06111977741046182,0.6544849634409825,0.04751335340818297,
 0.5198147479901021,0.9470055595788226,0.15574803532583925,0.7210899253696529,
 0.732716989176969,0.46139872639445856,0.35191262456565275,
 0.42567988363514797,0.052304169373068476,0.05714055675580221,
 0.22439981420131716,0.31880478586393246,0.5765189280518291,
 0.7443654960882604,0.21970673443997524,0.5724522951051357,0.4511593375235011,
 0.44581843478282723,0.3583990906729563,0.7323566486481558,
 0.32884319804433204,0.7707710686420934,0.8545372144429432,
 0.18995739580680737,0.6901025928809641,0.2526900949292219,0.4777759588744974,
 0.9049183668673793,0.05539222553826284,0.4802846502673981,0.8798754874927767,
 0.12011085020733925,0.9057943118030151,0.20361511426890444,
 0.4678656028171464,0.7895052510401965,0.9468754808849033,
 0.006106838001197823,0.041254531697711094,0.8212494011361682,
 0.4329388502796243,0.460248314903835,0.11476153160637659,0.9795114075168392,
 0.4541961923033119,0.3329575684398801,0.8884546684200603,0.6703253755450921,
 0.8545786805381584,0.9867894731518245,0.7726115355416497,0.194932961407709,
 0.5406035569514829,0.6377343757310969,0.419216392434257,0.708886024153898,
 0.08674772995337121,0.1632175850961225,0.3054643067872178,
 0.0070088893212811865,0.7360550113923839,0.7629429861066145,
 0.7734322945830496,0.2938973441335948,0.016607399008660817,
 0.9945778622454925,0.4705386107716143,0.5466345812062583,0.6396502058503887,
 0.11830056742468942,0.5351332150150298,0.5817180397577861,
 0.11763711417538891,0.5976169173480024,0.26758699780885975,
 0.48232215748362406,0.9182887136224692,0.27086994383274055,
 0.8516557535601077,0.5223743303660975,0.8559760580699695,0.7720719866509169,
 0.7917898375184406,0.6293871550263952,0.9273412390590606,0.21218477894632282,
 0.5149605884846109,0.3711438558949901,0.6198906721035611,0.2846547504693452,
 0.1802036169397253,0.14856969198250014,0.2906057071721794,0.7936361823461111,
 0.6508162342895688,0.9082105743284526,0.5880651863638864,0.21265835021988166,
 0.8050094684899871,0.9985973353352945,0.02246620961951107,
 0.11188011899887518,0.4079411768008444,0.9450254741776432,0.6142169201562441,
 0.03499899556117669,0.7698380304017824,0.7564942740839079,0.9071850424498171,
 0.26110113414052105,0.7537981380657555,0.1516443195939945,0.6630396842040928,
 0.8236272535626404,0.19344756581218592,0.22887063111816797,
 0.7536133264672358,0.2952552408910706,0.5245178885106735,0.6985253938451863,
 0.5077691079517606,0.17684751164116563,0.6178548311946204,0.989733357294529,
 0.9885721166145061,0.8593305459455153,0.9843170502827311,0.580223967051772,
 0.7284500537774242,0.49469173948008105,0.4919770762599467,
 0.016959529942597085,0.6422960256613176,0.6575497879654946,
 0.6824164807180375,0.5002943686736728,0.3339833441661395,0.9428958261362992,
 0.4699404458105473,0.018621904799828948,0.20956556704734675,
 0.16090507547038269,0.844170257980883,0.845713239780761,0.45077470382560614,
 0.9485094465463265,0.514775491189887,0.4580008804736778,0.05014945708033014,
 0.6763887910555829,0.4066485706433104,0.8160838240879544,0.5753325422374063,
 0.6918106268978944,0.6116730411096093,0.11544380513983721,
 0.16510254737643093,0.686137246362015,0.9963703816688481,0.5019439889687856,
 0.463317315915001,0.8882380170329297,0.30053998766963985,0.4622529749336203,
 0.6672004122296656,0.9179386461639475,0.6550106588191241,0.38500440994923335,
 0.1126217069013079,0.8710081855783492,0.8047258665662924,0.5428010185976793,
 0.8513388327160616,0.6016477599154213,0.9410474824641537,0.3701940586566338,
 0.2812244666801772,0.8160094250282932,0.6140965049568118,0.8625424804458612,
 0.6919350548664579,0.12605180419183515,0.8886620839069167,0.3897002209008993,
 0.6789807970984124,0.10434645784808183,0.17227371321134843,0.686603195048761,
 0.7633942406816123,0.7864128677439189,0.02100764473482286,0.4955542204495741,
 0.624709844514388,0.5991857549371769,0.13638179472354384,0.1475064588980507,
 0.1353148333273988,0.8785929646373489,0.8137393125774203,0.7902438706822277,
 0.9597961262888546,0.3284687858063181,0.4157955935133758,0.8786823220348791,
 0.013149991952929097,0.2114399585179484,0.5873833202480062,
 0.4896020786032633,0.6833674641560773,0.26040139597535017,0.3825300828978241,
 0.9209372066232996,0.8389881607436906,0.5048188582974521,0.4730089688428313,
 0.9498545280666013,0.5845031041811222,0.6831455790073789,0.6743324629492078,
 0.04788522196531275,0.3857088326866892,0.9074190325939347,0.4805572802454007,
 0.8223676730284434,0.5598760087723096,0.42572234813082654,
 0.11054622453785212,0.3580682630117583,0.4924684727381945,0.6876159249130245,
 0.4761985732094608,0.646851543968458,0.67871976541113,0.7979817667534462,
 0.9351559816168946,0.9748162575861541,0.10487612112373823,0.9497651348876591,
 0.5171956791377355,0.48303518189949235,0.6796974876815333,0.6269488507105758,
 0.808688649981726,0.8047572949850705,0.67620299769673,0.10804831134190462,
 0.6238791809948697,0.08570035420939881,0.8799600190318029,0.7584060214537381,
 0.8356076133653503,0.6759644000424223,0.1742141940801777,0.06521408683886476,
 0.8460033835844384,0.3853488572162157,0.5729778058471405,0.8515697208548253,
 0.6966775965789576,0.008248116203949718,0.10255717490251326,
 0.09282259817897953,0.7817943868458478,0.21795779335908805,
 0.48330108882201794,0.18589627353091198,0.43650590999171723,
 0.17654222939377218,0.2481782829901169,0.6568325452462306,0.6394417806646917,
 0.43465378713794967,0.5429569513909287,0.8229973179832584,0.7041658018203345,
 0.48124654649045884,0.19605843647218746,0.38050031945441865,
 0.6122234688464662,0.15601233702391792,0.8327853637372672,0.800337558742891,
 0.5976158328190181,0.2072194092322709,0.961926678785338,0.9582341544338719,
 0.8229726151109221,0.5374209665962357,0.5788080379086576,0.9144568152787599,
 0.12986531205368324,0.06748525668066263,0.409662932764023,0.9842100956806442,
 0.21351343273508827,0.6474936132452191,0.24493758784267106,
 0.9345309053712948,0.3672358961268174,0.5484063290655167,0.540241612814053,
 0.8106604707963059,0.931055479008484,0.6619990511244774,0.999581156905314,
 0.11038822860958997,0.45858973697171446,0.4533324096387439,
 0.3779307812810274,0.8037637261831495,0.9625470900492112,0.9840072352666751,
 0.980851557266083,0.381578207526699,0.7394753906361713,0.3984717181969448,
 0.5268518003449147,0.5993274344535666,0.14387696166653763,0.3175681956913248,
 0.9070391633695786,0.8319118440165889,0.2510624190846106,0.5226306848112363,
 0.1153108718014001,0.4991466906028772,0.9636505869495825,0.4300623623264952,
 0.8860885939208214,0.8873792462362804,0.7709446367794878,0.3123043759435815,
 0.5463699824144735,0.08141476138874204,0.20833564766132517,
 0.8820639643795565,0.9502619570943395,0.6609298069049756,0.8177887350362544,
 0.06988381467233384,0.7474704339504944,0.01151733289883039,
 0.40279209954148265,0.32916094755802794,0.379887020899355,0.5888420736997211,
 0.7193447613090672,0.4662827177764315,0.3641607707150173,0.6154197500426057,
 0.9998202686569737,0.6878072344528582,0.24789042479483336,0.8866476928174779,
 0.6721513710517497,0.10983525332713917,0.7166348878951012,
 0.36384469904627437,0.26319064105437584,0.5602360337637131,
 0.22679351086788246,0.8736292056828763,0.3398501926182401,
 0.27207879005618285,0.6085338338989223,0.007268842763154093,
 0.6856539534686726,0.06298399990603865,0.7669937570069755,0.0522714874312723,
 0.6683370903058952,0.2548421458917083,0.2350172297264439,0.469778615674468,
 0.6116448373129182,0.7441973686213209,0.31890329285357133,
 0.017617333937349655,0.4763259912094763,0.29319173143224764,
 0.8760603505273732,0.40806210338462034,0.15256633757045945,
 0.7203731279590401,0.985593801383048,0.8875512611331773,0.007511189276615204,
 0.27938050160014394,0.7711652879897342,0.2958663898212831,
 0.06099735536258177,0.08119296483995386,0.9465637817368528,
 0.8037669949268474,0.24765193066331948,0.7012771038137896,
 0.16356104808192218,0.5184997939198934,0.3354752864525604,0.9868058406982563,
 0.4509329391994744,0.896198239229987,0.5865899704881036,0.7809776473739516,
 0.16753483834335847,0.5893206254848941,0.5488115546589221,0.8450553942218112,
 0.5638530145691215,0.5722095728066328,0.1031803938369451,0.4038292452393941,
 0.5138381456668056,0.7392800747489779,0.13405789132628831,
 0.17275297498377784,0.006919022255559293,0.5215069927283293,
 0.3270135664650613,0.8305419925002719,0.9636131114286608,0.7002929934369457,
 0.38151051284879467,0.21131705973057657,0.39684476517673495,
 0.5072077748849598,0.7951954473873213,0.4134052705778497,0.6987779744857278,
 0.7174496513028723,0.9289475364024676,0.1768692396835031,0.6412666770352067,
 0.3400582345272183,0.16745894270972994,0.9841336007440664,
 0.006704366610963963,0.538250986369655,0.9998503954889159,0.4453614852453124,
 0.1621951632237587,0.651441026415805,0.8471990300220242,0.1435641412701485,
 0.6922442287087206,0.2034446888125626,0.026963066532157742,
 0.7298174590733705,0.7529117341580673,0.3788578056421763,0.7776835310151489,
 0.9902192044100785,0.2718877633073644,0.7193009449368996,0.8390988645123743,
 0.3978547156430572,0.7900974973854646,8.390053289600852e-4,
 0.9253320337545934,0.24361848709112777,0.2622360362241317,0.6559754563735792,
 0.31944602252322885,0.05265872376384351,0.4378287907556513,0.827729590319929,
 0.7979599301142889,0.05977959106105002,0.4661481620895248,0.1788356877029128,
 0.44385659265966315,0.6301554973580231,0.8308790977587202,0.466362837542869,
 0.7033060801206471,0.9763812768065907,0.35304171826439934,
 0.11447490531842708,0.2770353165036712,0.4497244180691563,0.5738801293775821,
 0.7820511980053848,0.8597884036950199,0.2993553992055997,0.08595660725006726,
 0.5541718430989828,0.06596178958802812,0.7029862188642245,0.9803773472506628,
 0.892092772957703,0.29317080735818424,0.43567046542896404,
 0.05762384636018991,0.30312031658895344,0.3116350848152356,
 0.4451997592266568,0.589607254437925,0.7779833007462585,0.8052963940922613,
 0.15959826172263236,0.2900832418429734,0.2175903276208938,
 0.06377254473410754,0.18918874196077684,0.3187968039930137,
 0.2364457186676414,0.3621321976961268,0.8662050366720355,0.2841791303456217,
 0.17622771900759604,0.24757984758021778,0.8908858881911581,
 0.8624939896982877,0.3909194074354406,0.30194429067002326,0.8036965051913628,
 0.48287527935284613,0.7395450832636932,0.5607732123306751,
 0.04391531800229265,0.08311832071380032,0.6305194092858089,
 0.32048839974071974,0.7923609873812564,0.8685126450265759,0.6721622201560108,
 0.4582058833071422,0.6403007990425328,0.17131379488428122,0.334890248167272,
 0.7912838297224961,0.9427036633755905,0.8676372146226383,0.44543928918227826,
 0.14949648559653783,0.08320445031429513,0.5427253144179157,
 0.06757971485362524,0.713946693785787,0.7891447101977167,0.2449684323811463,
 0.7458837196258238,0.4908356978508349,0.5063800105341756,0.7952189352295667,
 0.35669061908727406,0.01783504271979719,0.909973962863385,0.8745893598322737,
 0.18372404803109943,0.9964780770181464,0.5189189237332825,0.7450352579869192,
 0.5117858642945442,0.38589902616983207,0.6009504647747441,0.4254656689423204,
 0.6117764292958459,0.26828071915406126,0.9537953687298658,0.5301376814753068,
 0.18763061226394773,0.438197683458155,0.5436571745189896,0.6868165355439492,
 0.8439116385922477,0.6271563922089045,0.9772078974057482,0.9915157186082614,
 0.05664056704168108,0.24028940199032,0.41867832774817904,0.6250226323749102,
 0.2339483732599381,0.5420692794758666,0.8374886432668746,0.7991126655761205,
 0.30920140588286404,0.5037868398333601,0.9853455959809149,0.7036866946074097,
 0.9499466328672014,0.3038574289746594,0.19150757116605233,
 0.33448735258608386,0.011130579516021033,0.42477263559068734,
 0.6922836121057038,0.052281114466176226,0.0561064476708758,
 0.3516863866789267,0.9835799543352998,0.6464587756564127,0.1789777814632435,
 0.2912055062494364,0.41704176896809475,0.48020300981184294,
 0.5630979984473766,0.13827030722343914,0.8949567153342314,0.4729074469033273,
 0.7120867636875345,0.248507514014351,0.8683588413730967,0.8971733008325679,
 0.5328945286478708,0.8144779458612071,0.23754040525038078,0.9358011625240629,
 0.5636969673204484,0.9418744508624205,0.12651557847286587,0.3138023367263355,
 0.3676229459626872,0.997963796887423,0.16235169289588602,0.7958518952475208,
 0.7324312817643568,0.7401577818557101,0.7771515178763231,0.2937135082906621,
 0.6564553096013565,0.9049704881287775,0.7517230546691873,0.3590475109292214,
 0.2507560473607606,0.11755872054176142,0.3804393997619935,
 0.07875669833369958,0.23417079535679908,0.9964137433357643,
 0.20786021614610528,0.023851352265116255,0.3644259457461787,
 0.35852175066503855,0.23178459082828107,0.5856237894800601,
 0.7247298892503209,0.6411150427845628,0.8196159602489996,0.7813389825476391,
 0.608816390213335,0.5331939360646183,0.6941819271585711,0.8482893482388092,
 0.11732164908961917,0.931830115920612,0.1338414134624093,0.7362516120051741,
 0.9123557030758528,0.6964813698522511,0.5766144944828231,0.3533100170042376,
 0.8678100234843544,0.21661040246109553,0.581733125482527,0.25200042615300433,
 0.8273175692213857,0.8240184791471097,0.7347303976780962,0.8610360624268225,
 0.3259258155381287,0.055007823935191436,0.4247922026024852,
 0.45214301142788893,0.34112648081214947,5.235114218167247e-4,
 0.8245322389324088,0.8248151176141834,0.7148289190317593,0.19760313633478255,
 0.6806045905144628,0.5940167811097479,0.32094455638296693,0.7756254318503467,
 0.9658127872840128,0.6230318750392054,0.09193690836548607,0.5615293716806627,
 0.08616712205855637,0.6572188221475167,0.905565806707586,0.4062216967894843,
 0.39937984253890313,0.7893233617346503,0.290449445124772,0.08793969685542136,
 0.46020397949719916,0.5992547407788059,0.15676608996502905,
 0.2815420367396477,0.669568118368695,0.2145372074907238,0.6042844686979566,
 0.009880426727210012,0.7550733320426118,0.2844675507314146,
 0.038021933422756615,0.4055884448635487,0.01297881175155502,
 0.638278301124248,0.9240045008414001,0.5593774553843731,0.6499276782481146,
 0.27548051246686944,0.45918236856479866,0.32755668876519184,
 0.6150737585771563,0.7414341713737678,0.5460937236271312,0.19246473991010138,
 0.393156535808169,0.552729702273657,0.905707706384055,0.23974555547462695,
 0.36414325165110295,0.5083120404100129,0.17211000957677036,
 0.21493244702557357,0.753719148497356,0.2569859796998246,0.2758434977611802,
 0.6338670103063626,0.23211759110661434,0.09703454386558752,
 0.5718576275295968,0.7638529444507786,0.358147339543881,0.31701104661732726,
 0.6585735274642742,0.17989324602116585,0.4086772815281873,0.7806414850824839,
 0.7313964150872416,0.9372106759011462,0.64060473562387,0.11513711082984934,
 0.589282570539996,0.8640072337062331,0.8258998017273486,0.35607306440289865,
 0.10308827784074515,0.07455208234068389,0.6569606829590455,
 0.7339800457510893,0.8325166229625088,0.7478586833198785,0.9423725087739891,
 0.683898669620779,0.8341821745092808,0.294695585228964,0.04209039231617551,
 0.2288504771857347,0.49992050778567854,0.8858598800933019,0.683968964949428,
 0.6487038030948408,0.8563060653877477,0.06714134771312208,0.3411477918475698,
 0.5309204205230695,0.43207436308434155,0.07323329152933322,
 0.03748858467524374,0.9923608298885735,0.44198875060106513,
 0.6190906668256471,0.7947686743051967,0.5657598071431292,0.4543000880256344,
 0.8966452096268241,0.3921634327730875,0.8007428720924389,0.8182014860721296,
 0.22133500837887832,0.8406728059080052,1.1983852749442025e-4,
 0.9217689164212864,0.6598974215816688,0.7062066496445445,
 0.008710792520068344,0.9660476547068716,0.001705565670433784,
 0.022403605681003413,0.33971736082285786,0.5504636373795503,
 0.2783516780286607,0.33185564338176643,0.8728010537663492,0.1180361497357767,
 0.9406302145514688,0.6967272118073654,0.6160931376865629,0.05702482656561603,
 0.02717389730434283,0.6052779870520173,0.46742660006076087,
 0.6201502108462661,0.7658326569355156,0.6994714366297933,0.5463715127142518,
 0.15060687599475397,0.1617136034982768,0.42984779427564557,
 0.7555136286328081,0.18528710734603782,0.8307739191167073,0.9317969205340925,
 0.9548698487480012,0.8943671156267675,0.5294628048686703,0.22243538501002302,
 0.1475194083259891,0.5789348780692799,0.6338074656276834,0.09194797348458839,
 0.17555411409033028,0.19984866725731731,0.6022887854105389,
 0.5036264750264453,0.004845495283612333,0.9277623455110542,
 0.1739802274594917,0.4368781470363682,0.9124839270580754,0.23516709704160088,
 0.08048207808030872,0.1403476085078107,0.567378123839206,0.8488622092842636,
 0.3713458832571823,0.5061716124549673,0.13841410116383723,0.5776631789550999,
 0.7411935856789656,0.668892978153017,0.8165686936410614,0.8347781378950218,
 0.8724162482417229,0.6847719278406339,0.407054899404498,0.5370169828609926,
 0.5421138105447698,0.502578054978376,0.7570496093912679,0.9527638571789063,
 0.13148803893846195,0.8479355445644332,0.03132961544040225,
 0.20107607425910112,0.5528724660583706,0.7967165217850156,
 0.29835788347561265,0.7992317159147123,0.22056050336010424,
 0.16168382050327545,0.2669009265742288,0.46758618688706155,
 0.8991614132332205,0.42660515879436567,0.32405322187045094,
 0.7011524251291558,0.28403560964964525,0.5516356930683357,0.411286018438353,
 0.11932380504958107,0.6216357276585052,0.29372679979070215,0.40785011688909,
 0.19268472782357726,0.7248920044587382,0.05211564815087244,
 0.9491598688700098,0.4037220012943118,0.944887265285564,0.018475838764584807,
 0.18791024395389977,0.6052295217096129,0.004478477554433757,
 0.8866655889442768,0.7370417226217785,0.3219407097345246,0.37414739873413083,
 0.863629956511689,0.5054181924753958,0.783953447076277,0.3358023097776709,
 0.42784767175731564,0.9781976646567334,0.629149214287405,0.4602580695574928,
 0.2311890479494736,0.5838735791539748,0.36779885523516886,
 0.09123991615165827,0.16211076623774423,0.05527389431354379,
 0.049472496671317256,0.21396122226098013,0.7882823174817395,
 0.5651300428881075,0.8711794026337384,0.6435570654886782,0.4819896404557862,
 0.4751504993796498,0.5314756122579446,0.8753079212704804,0.6016638816178426,
 0.7324834501437383,0.32612640050159736,0.6381275506577804,0.3704249364891724,
 0.023081578901249844,0.09697390217958368,0.9486389181840559,
 0.9364482180357915,0.763770618252796,0.011732649401839379,0.8835333173033545,
 0.03741272750754121,0.9307953037418994,0.4340441514304102,
 0.15039482674330917,0.05063940715955406,0.35926649104207564,
 0.5232916070793746,0.9533470590948036,0.11491612168398546,0.8446104820429946,
 0.22093888356244362,0.8372311646179619,0.7511693978874655,
 0.09018598039284975,0.7207917324214327,0.7470702615684919,0.3767950365700362,
 0.1087099564653683,0.902553316355643,0.06601663014288861,0.4684614793106914,
 0.3044971294874045,0.19593206457286683,0.3867382156394249,0.2104372286648558,
 0.035079892375784016,0.1280300897961485,0.17628364789405537,
 0.08906127240544415,0.6441217154775084,0.30160542576382,0.9218358744403893,
 0.5420111649354317,0.6969841055157612,0.6826947042188741,0.9698361680009434,
 0.18812252872645274,0.16364898467733802,0.8315447405883348,
 0.4771182615413657,0.4813165549286892,0.9570545446278989,0.889627965528297,
 0.3071397233208475,0.5750345468644492,0.4969465495786929,0.4886380693777539,
 0.8656187474717324,0.5068598318093557,0.873434840404094,0.02024293112964526,
 0.6670390972263138,0.70124365376939,0.017900668183253976,0.12006637232324657,
 0.5403307212859817,0.3847741020272505,0.5569512184060893,0.15052484814223122,
 0.1032638906683373,0.30441524666527153,0.6599774026980727,0.8928623389831819,
 0.7132486020632154,0.5952299326267734,0.08005630684491472,0.6077293789653836,
 0.2712329902053239,0.7641407149076108,0.713290240533365,0.27668257968437693,
 0.1689661022695963,0.41711141829683485,0.7819852672493883,
 0.06024190116816719,0.2473009672844163,0.3056612245405885,0.6814415808304819,
 0.40204978351412857,0.03507485023229784,0.9701245869608617,
 0.3820343493376851,0.4749106049864842,0.8246667031486264,0.646946889670506,
 0.9107664817961855,0.07418608042965924,0.9762741404336808,0.5973410610941112,
 0.5881011135863123,0.2812566320232125,0.15211727553106247,
 0.32939717188371176,0.9202630661034401,0.29165707015726094,
 0.6997232706880656,0.4044855495926186,0.9840967671703663,0.8912698071380198,
 0.5548177647591596,0.6254572847055843,0.7061679108878111,0.6726174481383171,
 0.43803963784257627,0.9105063026530209,0.8421727961085795,0.8173253248014785,
 0.6393312940943782,0.6798656470246945,0.8979769916409389,0.7611524526784077,
 0.35841966267180425,0.9450506797216204,0.37584496198998063,
 0.6401737653491961,0.07739771872217294,0.6680093451067857,0.4308839843648118,
 0.7645658210053359,0.38444713117622087,0.6651900365720147,0.9461690736509174,
 0.5462313253683108,0.3463688831638837,0.6217139540997046,0.7607672809220344,
 0.19424781484516773,0.6807496405085153,0.2886318339929205,0.8873308432569091,
 0.492019600339358,0.6747524750082717,0.667877844354136,0.48115870962316243,
 0.1028844034629044,0.255875983007291,0.5210016718675409,0.6232837675388759,
 0.8922481596865063,0.7292300014810261,0.5195495284788936,0.07463903072630373,
 0.9315369089891128,0.5559378892389651,0.9824387213813155,0.22767755620618424,
 0.6223464796566258,0.8581734744626385,0.9836110517847889,0.6932797728371837,
 0.9753675327711486,0.1049285112995686,0.7713212793227315,0.7795725468046538,
 0.7598505881813318,0.03613579185599625,0.8167100747592478,
 0.31413229196945647,0.5446215114133309,0.2750443217499695,0.3121494681593592,
 0.13341617931581962,0.9370529111220485,0.9331369500359366,0.9826349262898377,
 0.7812072906713613,0.2792805335494952,0.2005400653973708,0.927306782817567,
 0.7194360936769822,0.3984193135241004,0.7029586734771641,0.7849672334044215,
 0.13926644495270457,0.365612010274466,0.8435489915759082,0.8362094972031763,
 0.33501412503869665,0.8361231152778917,0.785407322945606,0.4327311977460704,
 0.9485637966500071,0.7554650028754699,0.05183900235987127,0.9950834101775222,
 0.3841623534163421,0.9136672453769693,0.13599261170763566,0.6151316280188494,
 0.05979353692073275,0.005347001560986997,0.24280652193990737,
 0.4301838536203174,0.8164388919478694,0.8353242432600265,0.9813049813745595,
 0.14550909931846578,0.1520213383132213,0.5210111145673398,0.8854876372764734,
 0.3734026536716194,0.9398139688993925,0.28952623879614536,0.2849589671298661,
 0.6712172535678048,0.368051660336979,0.6253046757581816,0.610270835976704,
 0.3085376669424701,0.8973957490739599,0.14678825647287475,0.8569032726658458,
 0.26113073119201635,0.4351228637424984,0.09112840339551831,
 0.1311610188524217,0.6770657988694491,0.684789851916999,0.3918091534124414,
 0.8086991909254807,0.7093996448259133,0.16196000351257245,0.8730491611925975,
 0.5321218921666454,0.4787991853000969,0.7571385611139376,0.08873113395641752,
 0.4866725018632435,0.9124601243462762,0.2817555099125797,0.2859098154700702,
 0.40245032947726567,0.8401275274582454,0.13899010250142207,
 0.4232537970230177,0.24857577201214442,0.22851751151018274,
 0.11253564928979909,0.5428275383337243,0.9658997550478293,0.7523098867262705,
 0.6000534519407755,0.8573284178261863,0.038752905693879236,
 0.9747768028345977,0.20170621724932636,0.18347507384046768,
 0.9383869400664108,0.5046143278380423,0.2968553073504372,0.25330540272316004,
 0.8453344545575734,0.3354130440325256,0.6188537054276896,0.8877198017028736,
 0.8495734613696466,0.09235943722387896,0.16715092976163592,
 0.38795244019300923,0.968677350743125,0.691555124819789,0.38589080281468524,
 0.5393907595995971,0.46737929469784534,0.1253244580036884,0.5446194988574024,
 0.26924129706270294,0.6484535048384983,0.4905009323465126,0.9809056579340218,
 0.8817156488427297,0.69817528538333,0.5719039151313153,0.6472739564212789,
 0.4908805469187397,0.5350244709851096,0.6094572327472857,0.7839735070940963,
 0.7697644317064193,0.031226143233661352,0.23554872467034915,
 0.0520259402038703,0.9713902762610965,0.7739397133112452,0.3063412946611923,
 0.5506981346550466,0.6782468171881704,0.6038991751317133,0.3932827113897126,
 0.5679357992655516,0.01375668551144782,0.1932843095531941,0.0689050751141248,
 0.3565870929981034,0.049736497660579526,0.2655458341567415,0.719724223208928,
 0.36728063004807554,0.2103697230709819,0.13883204509456848,
 0.6676463865605294,0.4102905855552553,0.04368681260999829,0.7680531880497867,
 0.595475941284853,0.742283375412514,0.6779260495617507,0.7595103612344654,
 0.2652414930668058,0.8180016642684251,0.8897815715369466,0.5074376623790368,
 0.8099742952397686,0.423757869910548,0.7617526245772535,0.44219766586374754,
 0.2079729707636554,0.15315361546194795,0.0641263448803957,0.5454502866585513,
 0.14037495208018957,0.5713608784002604,0.12146628251330505,0.680115170563758,
 0.6951087632439212,0.5820923667728917,0.1167125789959047,0.7966777855955115,
 0.7050312858663091,0.19708132325015193,0.554222057920609,0.7515291380458288,
 0.5089360696491679,0.23020078254066412,0.2300454978711307,0.6919775522857293,
 0.6533439302965554,0.994637548884705,0.7317908427095229,0.7049119073265023,
 0.7266423120455069,0.8706095896094308,0.2423731051702216,0.7150974554760824,
 0.49138962463843183,0.20869206209963154,0.06515396394067019,
 0.9863026554890957,0.3420850683643588,0.9500002969113743,0.49291956460661424,
 0.7887823689786758,0.824257680796545,0.15874636900179173,0.5909901911339323,
 0.6739654315099096,0.9567628690772704,0.44794242551812324,
 0.12341922210812961,0.09384609752752693,0.4077534853474709,
 0.1619042829578322,0.8172701321800375,0.39290969180888013,0.2014645914672042,
 0.07947725493353919,0.4337746916001167,0.700677507776069,0.7640181774786078,
 0.2548232208815069,0.17637158490569294,0.5287875571767336,0.4268945822292225,
 0.4453273264405059,0.42952004491300244,0.6842339117601097,
 0.36542716576057255,0.14493848260583675,0.33807242984962715,
 0.7997464699387218,0.961278247326733,0.06076882607948886,0.12622470114744366,
 0.8235732564734702,0.4157838253404941,0.47348474004406416,0.189760309475985,
 0.4238076260379223,0.0010091381849685055,0.03698075582607141,
 0.7947982505725295,0.7660488164466168,0.25140927658962864,0.7614806666824813,
 0.4412588211125449,0.9902022133189496,0.8582401542494466,0.3700155751806097,
 0.7329953465532113,0.45349527957180347,0.556172044994023,0.5984862878964999,
 0.8192642471285533,0.9415433715595564,0.9935600145049872,0.6995882651945264,
 0.7609953731171268,0.16114492131359515,0.2804489904546924,
 0.15510471752614075,0.25596418822162303,0.6073513940570183,
 0.1814626219404002,0.9131309902072121,0.18263265889191238,0.4920338639822618,
 0.2503455283851995,0.0738158657389818,0.07322070509981993,0.285801752180168,
 0.8125900797181059,0.7850822182260003,0.6229270676211558,0.46919469822128723,
 0.055802253104481614,0.44486818798116234,0.2176100533723343,
 0.921315396708582,0.6127985504414477,0.45571935489278426,0.32605290493642314,
 0.5847134759427984,0.5221708836157432,0.936431423798312,0.4065164154015628,
 0.8933298165335068,0.22688646867051832,0.37578930894472573,
 0.01666024585880277,0.1690329524871964,0.5616266630050392,0.7393507026191135,
 0.7083435355240086,0.13966183431211865,0.5291787365820927,0.9043111384883745,
 0.006251984795874632,0.3845464405720793,0.6678912458476224,
 0.1575431704258522,0.6505270363400456,0.8431686843686739,0.3419294575794898,
 0.9277250987460001,0.15573419675336875,0.8465253921985726,
 0.05234885461895722,0.1880991935587959,0.5751942869589398,0.3066079571239473,
 0.7358344243657307,0.3246812615317215,0.07991403014314824,
 0.05929748458905537,0.9639230917690984,0.8563198079162322,0.3242502887023667,
 0.9200284594845228,0.30132354261023675,0.2076837165589872,0.9989953525781992,
 0.547858243283415,0.002331206853736445,0.7193141435579669,0.3665664290035706,
 0.13153170140903825,0.752674478285146,0.08594321467986699,
 0.37287707330710673,0.8303205683131374,0.7210783885198074,0.2526727613939449,
 0.8248112700502566,0.6028391386424294,0.9008314680523458,0.8344285508004877,
 0.3572645672308371,0.18455196610539504,0.7622865125614766,0.6565215030817136,
 0.6564925781244819,0.4977500908963802,0.5946267088631686,0.0684729164192196,
 0.5273330141337883,0.6443867246932646,0.5011168264144757,0.49417390057551125,
 0.5920846297183673,0.8603048088944445,0.16985623949430673,0.8694082238760045,
 0.09074178295735436,0.5384477478931793,0.3265843894406081,
 0.02981598561233678,0.02429979831974982,0.40305190413915526,
 0.8581642763284683,0.47971066302350707,0.36215155108593344,
 0.5319516362200702,0.9272310415507782,0.9989755772408251,0.20866346635479593,
 0.3671099139766687,0.20284613544276475,0.06877504115976396,
 0.7803208119787411,0.662717808705725,0.257641888587397,0.04875978710693618,
 0.5773802886524595,0.20942655926389175,0.2460030037183848,0.2954029794060319,
 0.345883497209153,0.4520087346155447,0.13251678428465619,0.6931111774267515,
 0.049922323850823624,0.533690853046341,0.24409150343291575,
 0.8786197635440631,0.4086548660164512,0.24210144258135602,0.9508299513821294,
 0.3208014559506246,0.12231779114181307,0.2539130928777794,0.784583007619915,
 0.05357867419906026,0.30018269002613496,0.6189532279317023,
 0.5447928463007908,0.7873525633468661,0.6245204187859655,0.06508948431457551,
 0.12483671503883287,0.5680976367900128,0.4206245427674392,
 0.06501545027042144,0.6706989039896211,0.20926954586010993,
 0.28095439117702237,0.6726907676282047,0.13534699448724785,
 0.44358328906610367,0.18619675088771515,0.09117381442933781,
 0.5646834764583952,0.7759407661157205,0.2232979032854303,
 0.058139529333843476,0.21278353632729408,0.9130007937425679,
 0.15867845284103443,0.6351879689818498,0.631192784636,0.32031407399175327,
 0.3245244705323592,0.3670197538461413,0.46120406186343477,0.7746108153284673,
 0.6424276506049801,0.12505530953151056,0.18867009572529048,
 0.5223410893514402,0.9569802865560378,0.914133077695755,0.7574012073790017,
 0.30250094660201443,0.4694971372293425,0.7230081064169349,0.582625590095427,
 0.9720548595163481,0.5846665154514268,0.3060628627404729,0.5764754012332813,
 0.8615555217520368,0.5553880501684728,0.18140574093778017,0.780041441179563,
 0.7016023088241018,0.5686579995913376,0.6014859726708173,0.38939739295725195,
 0.6588112937648053,0.40535374940770674,0.8228825040294231,0.6202467075560374,
 0.061237016123392274,0.9407910849779093,0.004267238714252941,
 0.035979761315521275,0.7205201523825144,0.9340313040524227,
 0.9399999583004918,0.8502523524192811,0.5102287946649939,0.6498578589492853,
 0.4842306086634238,0.7114282540271527,0.28245391558123356,0.7530698050885848,
 0.5856592258737777,0.0013015990571810798,0.4273176058159508,
 0.6876014518281245,0.6717203222575564,0.018330772398618356,
 0.15290568509184643,0.335458163775745,0.6802691837921999,0.12639105030448095,
 0.8427243959698005,0.3455375447382806,0.28847868006314237,0.7977216977991489,
 0.78939575766668,0.11209614406886448,0.25083893022140624,0.859258843217343,
 0.05412436597015258,0.695371832287025,0.985211834678056,0.2297917309883276,
 0.2789950703983072,0.6167231282561018,0.28299862481831906,0.4660593821995279,
 0.786412530508399,0.3330997431578767,0.39861426319335047,0.47672926503051005,
 0.9478549083329546,0.9373764254186123,0.35194290277054796,0.6000250311370546,
 0.28771139696120995,0.8137951618781097,0.6267584410111157,0.3497440884329549,
 0.20789567948337706,0.6939723014190331,0.648989725874481,0.990247516303419,
 0.3065938891206391,0.45090014808357415,0.20851689983685118,
 0.09752751561395112,0.3745511543457565,0.050828738605430135,
 0.18873726713995942,0.3164999539495619,0.6053128740608269,0.5355340711699687,
 0.6388876747637957,0.96586692341179,0.8176988946186983,0.8435293887037408,
 0.5006010486974151,0.9478170912357412,0.763746803284538,0.738060548068528,
 0.6832313594365805,0.6230091196708082,0.49441946010387094,0.2671037809586496,
 0.16752461925500584,0.8920533225526774,0.8702267709418825,0.674580592230506,
 0.7970634633977581,0.9252954914428559,0.6236785369241733,0.49693521348599723,
 0.10979529416052847,0.2152683027596758,0.27976641935746716,0.771720533155601,
 0.9988045855618092,0.6166427764036436,0.7620862483509644,0.321135459425543,
 0.964202229926171,0.28037559010240987,0.1917189367678941,0.9642744762176537,
 0.22160205466064686,0.26911830283378935,0.3580361744726034,
 0.028070267277287897,0.7680134874996771,0.2747818828973172,0.920573516895015,
 0.03250174542126283,0.12806776401800812,0.17468761612833328,
 0.9257636163274707,0.47832209968270745,0.3666670009305977,0.8666276771911869,
 0.48153109456271836,0.46217091800278387,0.8460563559739172,
 0.1841264559674386,0.9507141795255688,0.2863478723749012,0.8446584060997583,
 0.7944570898663482,0.6370283224564941,0.7300574042233129,0.3630183246979105,
 0.31830237221674196,0.464147117270855,0.16327929228431348,
 0.20223773557173397,0.8487396450288146,0.7167360751525432,0.9119470945554774,
 0.9130390580141375,0.10781102350580263,0.06674945316553227,0.892676319930488,
 0.7871331115297642,0.561884232531718,0.46011818759404355,0.7076874072596074,
 0.8052428360559987,0.29853728946317326,0.6444772403015988,0.6496586717232655,
 0.21513297087036942,0.44361560275599676,0.28223792419016547,
 0.060177417122514565,0.6201848539911583,0.6972288719764627,
 0.2475670310415341,0.042042779357602256,0.8688659075305663,
 0.2847914581901355,0.25355140542806065,0.6238367363935894,0.841355135633935,
 0.5429020101064985,0.883534963225511,0.9710580160893073,0.3926017947556135,
 0.7998108035281192,0.04341817624088762,0.9246542560229443,0.9875291488513168,
 0.7020948193973104,0.0735465604661999,0.009796480361423443,
 0.02734531111474925,0.2975105335073456,0.8250041088719581,0.3973804656642238,
 0.6050145620536059,0.5244369557148132,0.6067067465905628,0.49010607641348525,
 0.06404917397524135,0.044490230386411644,0.7552889034517641,
 0.8845077023056127,0.293557463585821,0.7368522630052988,0.8004033796248091,
 0.9091359777880268,0.16138437367084646,0.8018088613844886,0.2404917928085234,
 0.4757626048381447,0.3703645533052531,0.6135423623143605,0.11292997346492417,
 0.951385767981085,0.18098405485742208,0.8408937585209267,0.25926081680016144,
 0.4561833766171782,0.19782470816040376,0.041717298253101665,
 0.28883475092336264,0.5655784196274694,0.2449481706960649,
 0.36935735522151547,0.34568453937485355,0.36220352270952716,
 0.7965697663307948,0.05522163344123232,0.33285118488297505,
 0.7381149006603293,0.5457024924219059,0.38598152333365865,0.2628176322879171,
 0.5636622812735312,0.8781377862897578,0.3291601195908044,0.41524662131924284,
 0.12351501769711781,0.6677548938181013,0.29107117447220865,
 0.6878268203476057,0.02646834213852589,0.7140639185475276,
 0.07295417423147388,0.8215387680406883,0.424820965863728,0.7146725763984496,
 0.8386989580325828,0.8448900571023984,0.26599130012746763,0.8611626757931751,
 0.6467054652916441,0.897209341167333,0.6250194956372923,0.5427134779378345,
 0.8203029554475896,0.5533398274477613,0.24068149948450956,
 0.27108414852710516,0.04978181196857112,0.827609539137975,0.3061325894102561,
 0.4620139326138475,0.11710781270284665,0.5190393544209582,0.7512429338085136,
 0.7073395427960578,0.6243934083602041,0.12713173508217313,0.9243045919573979,
 0.5872967447375257,0.652707434546173,0.39213307535776254,0.03220253872655898,
 0.17096557560880443,0.7995452182743659,0.8472061040746566,0.5842827494877142,
 0.4753995258886572,0.23582020542843862,0.8600007928149628,0.9999693663298702,
 0.6170348919116693,0.5113471783436428,0.97338187580496,0.3079613443533141,
 0.7995424497188657,0.4772447122298218,0.4469857902344223,0.23099659873394052,
 0.45776660881268516,0.5033446213150516,0.601913659748281,0.2688942248171108,
 0.9500697972609972,0.4122912184552211,0.939159074028878,0.5994514869779407,
 0.627289807706222,0.11747267157373487,0.9563057056631016,0.4322248730080952,
 0.9173758855223129,0.26924414940532126,0.8198417404629178,
 0.39678916129310315,0.010004588966560868,0.3143617369129401,
 0.2601838916791136,0.8548709539702829,0.14484063931054814,0.5940353312956944,
 0.8610188683251311,0.20884804822971037,0.2748572915141785,0.4501934984394902,
 0.905910401394777,8.656719559589909e-4,0.2967162830696135,
 0.03812586542004015,0.968530061049709,0.2972223088959203,0.8090457814190604,
 0.8448319913717833,0.9689240832265941,0.7006603779544204,0.3373694465447069,
 0.3952445427136825,0.26452296573328715,0.12361025731540987,
 0.6754109099587988,0.46392447641945844,0.6886373369746803,
 0.22195437635343995,0.5331424150660623,0.7766749641852867,0.8121859682645002,
 0.9570345823420485,0.6024653028156838,0.6798936589620512,0.26753640056726313,
 0.4633527998673056,0.35144009242082364,0.8541156082711543,
 0.16508882032118755,0.42289388530831906,0.02983263088487953,
 0.2036810386860315,0.39966398612945864,0.11649507881964793,
 0.5854960638040593,0.7186651978643082,0.5838899028923383,0.8956378624092101,
 0.598864224235748,0.9345549687363955,0.2577422467071374,0.6656520636175232,
 0.5773046924114857,0.7808104392917774,0.7782039971685275,0.6754755631291894,
 0.5489889495890963,0.04779582225744328,0.9229713145207854,
 0.14163656848830297,0.008090641672676702,0.03449229026935663,
 0.0780313921705732,0.8826304657103591,0.4927595021090728,0.539500718315266,
 0.9874914348795356,0.9010812860327062,0.9700108915420047,0.8119706214473648,
 0.16776732826604857,0.5691134179209294,0.9685814588893638,
 0.005667079813572395,0.9854928870889688,0.6514413043130874,0.806419550956879,
 0.23083975734662932,0.273274586462536,0.5673929702176195,0.40647600307031173,
 0.6188478735080991,0.8081169318203811,0.5382302747735547,0.28286588741236074,
 0.8951385777127958,0.5770832193094191,0.9238583217676886,0.4662615304713503,
 0.38444057156137235,0.9180268481429507,0.8632368200344306,0.8577000712268832,
 0.12118158564847215,0.5015087225461582,0.948590640216077,0.16218575305352556,
 0.4395234244089432,0.3605667978719421,0.45477463471951685,0.6780436589005915,
 0.7989459811509705,0.982658705697235,0.23548296779380784,0.6429868348480056,
 0.04336990217636183,0.38626986630118887,0.6495719226565076,
 0.7475260903568207,0.7103906516404618,0.15776662901586813,
 0.03154268859466769,0.19363801213753018,0.4416676635933936,
 0.7440660705212798,0.765225428640668,0.7290146848884096,0.36444414273844306,
 0.8361824417516025,0.7658002713988212,0.97642834874099,0.9410253398447708,
 0.8457852775859601,0.001388938159397557,0.1599500061522674,
 0.49262026760914324,0.9811629732125632,0.8261980620646794,0.9119273490364771,
 0.48372681421942554,0.11475491431723184,0.3534316569123658,
 0.14882292272015518,0.5277577393384288,0.36123527308011516,
 0.8967222058517211,0.055226006180298604,0.3103927405621778,
 0.5664429175042673,0.6120316961911213,0.6374207047761273,0.6553785558646503,
 0.15005709973638537,0.14035852445241337,0.6602979998638777,
 0.007119402010908216,0.958188879330074,0.8409930036319611,0.6711325189122932,
 0.5781063422542791,0.5057182432815807,0.5426766839072368,0.5256370697737528,
 0.560307835063626,0.04872317044645025,0.3990884382651271,0.8632891068797286,
 0.5339054439494708,0.5472168614749815,0.9093417053252228,0.3544244848175976,
 0.4486559681070684,0.27613581106473684,0.27161751894317465,
 0.12891060630052564,0.11759819585282072,0.6982234360371226,
 0.3243325823093861,0.5726623167968093,0.26933102619057303,0.2963100283487815,
 0.6449519581131469,0.2154596429305019,0.9550250460243623,0.6733759356192832,
 0.9128225196482829,0.5456354264982108,0.5135274750393635,0.70897953662324,
 0.9632539756460101,0.002201215903387177,0.8400440526840373,
 0.7686471473084951,0.7619011241846975,0.18202671824540761,0.4095369875354422,
 0.7359190616366357,0.6957215957879022,0.9771604958704598,0.8176587481753956,
 0.23801470326244833,0.7429452107166751,0.314679990128497,0.4126508897249779,
 0.3132858726103587,0.7892214898576739,0.7885904102267273,0.2469787557324341,
 0.8533505005787427,0.04754826231012699,0.9951823800431913,
 0.47542658317821407,0.6237812424480064,0.45910381760826124,
 0.6973089122715652,0.7811251440373115,0.636033448956797,0.9034910018588993,
 0.582914793730211,0.35423805991266943,0.8279255968324466,0.04984644128833915,
 0.4812481374257067,0.183208879247999,0.6346875852391227,0.5931865171993349,
 0.4726696850209636,0.09278021439053608,0.8416246708446667,0.7925970964756076,
 0.7689901220040359,0.6847452806601438,0.8946546444829901,0.6225677163310191,
 0.8772961167882851,0.3999902408158128,0.648281917783597,0.23690937583898886,
 0.1910053851153527,0.4256922728476231,0.3034564968759158,0.8470153296063354,
 0.5652918459704631,0.27500460455559483,0.4192141236928749,
 0.21060311233266715,0.10246492707508303,0.6087446739515293,
 0.46135555263535966,0.41603691845930546,0.032437070565593196,
 0.4657496977035155,0.16559390619241998,0.5779112816104116,0.5133696948604124,
 0.6772008920884575,0.5779191815570156,0.8847099901755177,0.2702719367694566,
 0.6351217568839889,0.30975686916391443,0.272732721913021,0.9880285112773585,
 0.6529630722271591,0.014754853858978678,0.9595240828700778,
 0.7400628942608322,0.6852111521518731,0.0070171771215655765,
 0.21712373291010145,0.7182888041683605,0.9953830654922504,
 0.08768137054792446,0.8601221587164659,0.556791650666221,0.81488210699954,
 0.17644697558221156,0.6928613321974714,0.3625069498363914,0.8062491780439129,
 0.8429803382815207,0.4093744999054997,0.5099522075102704,0.01641986649687488,
 0.16941202107840725,0.006198806678851421,0.9286310744445043,
 0.8409419298223677,0.7581334855391808,0.714817865583615,0.08210331633092893,
 0.41155400734390046,0.06078487643166253,0.483979215787099,
 0.44211483045774047,0.6638561814083636,0.0026468475778406164,
 0.8634848805260484,0.37476024293454613,0.6370303351021287,0.5093831416765622,
 0.7215777071981772,0.13924542967018783,0.7926756316971261,0.612907923010258,
 0.2794657705250443,0.010335583129604364,0.4971845951260647,
 0.5915066302145885,0.2891696790763103,0.6236509532052649,0.6349636295109742,
 0.6177914564988036,0.06849125453136484,0.6681779522527223,0.2124399013950944,
 0.5288318591371141,0.10360299840243903,0.5680540738042953,
 0.28652110098833095,0.6471810177545798,0.555257546135911,0.5754862142619455,
 0.9865163703979178,0.4572838341383949,0.6458357769362448,0.40912776687980756,
 0.6092265957019538,0.6644169444475617,0.41928079521616823,
 0.12369729941687468,0.9671880891206234,0.19910310660707076,
 0.5035716700454381,0.2054952502004428,0.341919025963883,0.30934837568773244,
 0.4934450554329871,0.1271306068817788,0.11165253092400462,0.3252249242766452,
 0.7111758306801195,0.3056930665194969,0.5206175178097927,0.11345795282096016,
 0.8367223272585768,0.6597759229740763,0.1225956464342507,0.09873962611854081,
 0.6754485746283152,0.11983096050735109,0.12920429922082188,0.529439454300429,
 0.4517963798095197,0.02020718729094273,0.8205512339620791,0.5807741555251301,
 0.7728617432385702,0.5043603550314445,0.41343089076151907,0.4386865079927028,
 0.34117116518903556,0.36425434208649565,0.45205870087476574,
 0.1857798060853878,0.8454393557590687,0.7622986600916147,0.5282126866633823,
 0.5917556922925321,0.963122137284641,0.30663605219086565,0.7687915877772122,
 0.5538995087921192,0.7874821555634338,0.8254987076391447,0.3105341778387345,
 0.8798368068884542,0.6042511940792192,0.2925167741203678,0.6111012790697005,
 0.39709388446269167,0.36029157685941415,0.21368969935046556,
 0.7283163951112851,0.15839258179815074,0.5143974985727503,0.8925117649224669,
 0.35326711591661064,0.48468163342418535,0.8931717436763998,
 0.6306442676060472,0.2799002923460723,0.8184363793212523,0.9819781068198543,
 0.9067397577022834,0.17831856363630672,0.14864266570937423,
 0.5508797724347347,0.6578690288879475,0.7134290723652283,0.6317619141933124,
 0.4953559435122066,0.7878144592438434,0.23551958133184536,0.656991897750997,
 0.6329634994465932,0.5794023185033464,0.6862777431112631,0.1661514489703796,
 0.9016560640159306,0.7732501314477134,0.988433724061972,0.9612938998428832,
 0.025233445255390397,0.5794700781973793,0.7067732550771193,
 0.020253240252855353,0.33380431766610463,0.10902225598492532,
 0.7801060114724007,0.6786732788862195,0.42008765788121516,0.799140208026108,
 0.41341831081831304,0.6746790430937846,0.208340173402771,0.403394461056122,
 0.4565748915195794,0.5892288865674972,0.47207811758676377,0.8848990001356053,
 0.7880992183669339,0.40910724378092367,0.5112726985753041,0.3822097613949901,
 0.8942061449181145,0.731547359805457,0.2794729780529891,0.766373187611,
 0.5179821034558467,0.3899050932471074,0.07617250359487482,0.7830771428806399,
 0.24816057323201524,0.9586382063688588,0.592590142945218,0.9304201553306141,
 0.41665843208389375,0.8600586328388031,0.16291454951679973,
 0.8556497535253572,0.3034577461588216,0.4808283492529637,0.5952663394554615,
 0.9062106203467607,0.3296315011943627,0.34838524716163055,0.8192786202551698,
 0.8293010443892824,0.9543954706667886,0.3799681822338411,0.3109538651356567,
 0.8796883152811736,0.6208733752418425,0.2805831214281931,0.24669096305798066,
 0.46496849406619967,0.9072296322895659,0.7323832556117852,
 0.10377027196638533,0.2305579381474473,0.8003382050111076,0.748989845874804,
 0.8293290199326713,0.380564073608471,0.7426752713402426,0.6267612773269978,
 0.9072354721278781,0.8095030024269524,0.7031287500558567,0.12730749328470947,
 0.8094102216528383,0.8367321897734201,0.357163529972906,0.6720454456120328,
 0.7294555106475149,0.8970073314091038,0.9741467590329518,0.8351355224278914,
 0.798296916181192,0.4124738067570245,0.44354098435797973,
 0.015974169523019666,0.23554756930283083,0.5404172692649147,
 0.5487926038116192,0.3043554354984489,0.23413256386510217,0.5084771062802889,
 0.32317306203057883,0.9056018269616409,0.25854672558354685,
 0.2626272558586211,0.434360936023653,0.642743411283066,0.6437226530595568,
 0.17088244883437265,0.49047610584225865,0.8658923339525275,
 0.8122036805336622,0.12860544232267845,0.6582730413913558,
 0.14901990458011227,0.32351638805634186,0.9168878591490827,
 0.2614588176266196,0.20844982956357705,0.48156602214751987,
 0.12110490706987498,0.4115327195256764,0.33153425011608095,
 0.7285511297267705,0.3548643463110468,0.8127744079085348,0.546010510340513,
 0.0016207172373714496,0.01973037368573638,0.16896427743264253,
 0.3484503160128495,0.8786975085873545,0.13445721251464193,0.6283480262871664,
 0.09521969077633685,0.4650713580526571,0.1551706950630526,0.4805267794295982,
 0.45798867637344365,0.9939764502673781,0.49531984668681195,
 0.9701197101392569,0.6723674985802718,0.8145073669242823,0.35612980181501985,
 0.4635328211658011,0.07654565163691007,0.4872410365231731,0.0525403879400389,
 0.19522574247048008,0.8120797083482258,0.19748597175796856,
 0.7077727791705198,0.7278722934710266,0.7584602599204598,0.114267233993668,
 0.29052341765978174,0.5225019234110185,0.6355383699784489,
 0.07306574214122885,0.09628911783145422,0.5728584828946354,
 0.7162072469644706,0.33940210309551955,0.7148394370487896,0.8983304597361391,
 0.0020253967000727346,0.8545273011796732,0.4172312873344768,
 0.4445929735063999,0.42650514297446174,0.7643018285805011,0.6740833504984186,
 0.003820101137609444,0.46315950541830464,0.7825173431593347,
 0.4152848665000639,0.3227002937529373,0.5837389411741233,0.4702207366300185,
 0.2366534017438413,0.2752379142732474,0.8606932545696651,0.6478021367295386,
 0.24444235994004737,0.9872355461122059,0.2797606803685262,0.5517957609684759,
 0.3577658273066475,0.5425985943899263,0.51038976709438,0.3527463906180036,
 0.7796681335798825,0.9916627846779875,0.6967207931736253,0.9858267261772983,
 0.18835000876660002,0.8099835917630828,0.44242606684305985,
 0.33070397187429934,0.217147528749149,0.9464781748610415,0.24086414065016992,
 0.5495714617051626,0.6932449948586122,0.93135997933281,0.031269886720862616,
 0.944929274242403,0.9482288913655903,0.22515609684068383,
 0.0047604896442424716,0.11541241782952205,0.03989867651253054,
 0.1941232778480353,0.8398868540804255,0.8560642030850282,0.3581106342700542,
 0.1714044524914724,0.8682473792665548,0.40876931995959787,0.6341921380090427,
 0.7985527559525184,0.3890264521664343,0.14547553044550865,0.2205136342740277,
 0.4521074333506987,0.5225104646949671,0.8504872574942419,
 3.211451975836388e-4,0.28390322874752316,0.18471583905945888,
 0.398307252992656,0.29312260833636705,0.7531353470626257,0.8868883916417477,
 0.8189435888814259,0.8802489845907164,0.36650876999890913,
 0.08430806965688242,0.3853846436324114,0.2919948778411766,0.2703744323622873,
 0.9235810636395153,0.5471980890290221,0.6905016006421232,0.7362308219045604,
 0.5712723564723825,0.8590481416515879,0.09951183596496616,
 0.36526042503577216,0.4614227225070616,0.9403852793648984,0.8734959851770677,
 0.941113737105115,0.3902613908405578,0.6638940680902826,0.534793089841177,
 0.8178019612906322,0.9988704719010715,0.756710729829405,0.10561500459148165,
 0.31534458081120564,0.827960816727553,0.3574995962898704,0.7481762452884962,
 0.47614646752086076,0.588936103601393,0.05865825903928268,0.5186435444093254,
 0.8541535951190317,0.20586865961529133,0.05944316618726786,
 0.6575768683753029,0.13716707933294714,0.8150890543533378,
 0.43101354277472714,0.5953894658490848,0.25243507173968394,0.39605305620778,
 0.6636323270223828,0.43504792300288153,0.2897047495398901,
 0.27445814783959854,0.0016886699563553043,0.06666046961013294,
 0.42610711625166264,0.034304731556979995,0.089628758218536,
 0.2142521211509334,0.7389893058344753,0.6319558574888375,0.8428974725397287,
 0.9591290941199941,0.8312104399369002,0.8502343536682281,0.9961048543619279,
 0.3153698036771754,0.2552935262357323,0.22509605002295874,0.9670581395669623,
 0.6942342359813406,0.9720598595255486,0.746864238646387,0.5700163707757763,
 0.5610581839422758,0.37464450406398875,0.1874673869948026,0.578712993438149,
 0.8778997585010065,0.8166607808059937,0.9832616686472934,0.7649372985392331,
 0.8308009510410871,0.9266005284297851,0.6972635898699163,0.08794301879731392,
 0.16891202895199176,0.6422609138643643,0.4795217865396484,
 0.44822606275501775,0.8016918697424406,0.6557101718455282,0.559108743731374,
 0.7501815026234344,0.8109385625567848,0.6637495012887389,0.749815223494896,
 0.18510789500478253,0.1723573788267979,0.0966639570474982,
 0.40477627342044087,0.27608340079948357,0.16359120041255848,
 0.4738462715938988,0.974344081684503,0.8223411469579736,0.37388132108204597,
 0.48584975706063305,0.5504678663475615,0.6569574243901664,
 0.11687543669936562,0.9471893925691387,0.8794334081308957,0.5619703402009196,
 0.08792719157002393,0.09361126114967089,0.5474969239770083,
 0.8183702726836022,0.9425118978453808,0.09549071416345922,0.5162688917941604,
 0.10014241015114189,0.15916577169900603,0.9896864676477071,
 0.9628495176558705,0.2490710422167366,0.6616081544301688,0.7131027542188711,
 0.11971063296156292,0.6844199180476128,0.9282282793436586,0.8201009742206746,
 0.5455276192340447,0.27714415946102644,0.15239060723782583,
 0.6163062085920326,0.20606108274544033,0.11844259903349941,
 0.9120104553283697,0.9709800392102432,0.4674339790344002,0.9046372653834955,
 0.40186053531788957,0.9615353749034987,0.3917057871611703,0.3774488859265517,
 0.23010447411116752,0.37790013964941316,0.13564651908408665,
 0.6798490756777025,0.9729434697546961,0.5946711965726295,0.1520990139310212,
 0.2280423774409287,0.1431139227313839,0.3001841341262138,0.3878161551349655,
 0.9313950966546023,0.07449556415647485,0.716404831438777,0.6846148762670635,
 0.8533544899995321,0.8045466189385684,0.9861804175024173,0.8512947926359993,
 0.273709189164033,0.005116684785897663,0.46680487683682537,
 0.6583563750612081,0.880997067064713,0.34007830987578413,0.23721599912591695,
 0.15755463799136538,0.9055535225531943,0.6884229085730695,
 0.32184184809185656,0.8618411240275643,0.23101415864078922,
 0.20979359587162116,0.697409989616085,0.16506648523084533,
 0.010201128494160883,0.6474338549018825,0.7258656563406345,
 0.31464633741649095,0.7825215892943523,0.2857567506498422,0.970553685591059,
 0.7094533059135961,0.16391804560505707,0.8266920980720022,
 0.07524990776259499,0.814070999674881,0.7431052969979072,0.355120958305857,
 0.10166960830524185,0.912117366881224,0.7312155908032413,0.4950606500813619,
 0.64484319213308,0.787686486699948,0.12897418449186293,0.27516771799011885,
 0.2872065219214882,0.9302237869991218,0.82380770307295,0.22384430038381542,
 0.885333071078163,0.061913099071629896,0.20610756821310905,
 0.33841081599232736,0.06809847951571502,0.38045790304417604,
 0.9565570470536751,0.42731102286570877,0.33217894459665254,
 0.19764152533214163,0.4105409399867098,0.8413760208923795,
 0.25527019010073504,0.21066165759992383,0.6287772270932839,
 0.9533970881849837,0.02001403273176028,0.3830516661793666,0.5239832680456394,
 0.5216351476907458,0.7612446269208822,0.026043986976057587,
 0.6193840426377081,0.8372708486283299,0.8947876973394795,0.31067490773859885,
 0.5685872176054941,0.18197320215665136,0.8635573523308817,
 0.36005511842090854,0.9667883735139604,0.1905221305018252,
 0.10350669018787806,0.6537035774571539,0.13889595996267845,
 0.6930758440763698,0.5950792904645025,0.19882342699854072,0.507910746145762,
 0.349964668600952,0.32188147598505756,0.4914580902116804,
 0.052302421124121645,0.45379011784211354,0.8500212556714293,
 0.8431847514981896,0.7471203678200427,0.5942706645755601,0.27197889573259326,
 0.07263751997723777,0.585639100752374,0.6725729351845553,0.4148833141027052,
 0.4917495968375889,0.8240064879443816,0.4106425046858937,0.8883406591411469,
 0.5561758456605493,0.3459724127704268,0.43294866790707465,0.8090203283751904,
 0.6082131591318167,0.3251323483118338,0.586684735497016,0.20475353939691443,
 0.7265987185058138,0.8878309252577005,0.9398748522261048,0.8559730956895817,
 0.7235316068384109,0.7447197580795429,0.5300951013050811,0.4703797568330401,
 0.5128230471091488,0.6188439240367434,0.9668376275965472,0.6452717648374149,
 0.7844959799545359,0.43348055815846587,0.16026551363452135,
 0.7983308256980989,0.43174828901164775,0.9761028082763695,0.7986671520928068,
 0.9609514208782357,0.8608574220276421,0.03037898691015073,0.4443445337882648,
 0.47973079489598913,0.11551997075722409,0.9813688062784383,
 0.32763347490529504,0.0569706945470968,0.9237838945805957,0.9435081937216417,
 0.29974166449275863,0.8079178858625103,0.8540655195198711,0.7245297731160725,
 0.6452221431100688,0.3714983417391816,0.41197353618949173,
 0.27344724475616844,0.7340339241263955,0.20831612763817653,
 0.36172383084656456,0.11161724308830556,0.39986857241443174,
 0.7767895069530192,0.6683083901476855,0.64917436496476,0.6714425119267574,
 0.588208351716346,0.8598769494072733,0.8931658330228138,0.26093601674471123,
 0.8134780530797563,0.940611469560884,0.23231125794314833,0.5758441413468064,
 0.7413575082511656,0.6785964447926056,0.26361046191157,0.658794845338352,
 0.6648359591846997,0.2599883685173948,0.7890690627018657,0.5061095932486573,
 0.029093072350965588,0.08927897094275516,0.08028914081991734,
 0.20338954715121316,0.05616570503566376,0.17350556528213068,
 0.9060129441330612,0.7633729919587731,0.17755522728209727,0.8935980527906342,
 0.952720436884104,0.9825800898980511,0.8981698653409698,0.6826332506357539,
 0.6465108492012808,0.8217751044170445,0.4686413129446173,0.8157631231742642,
 0.6677172503529512,0.6438713943223362,0.9662802723044402,0.6596776337180499,
 0.40301815496807525,0.2744001361212016,0.6631580524248012,0.5202019538744016,
 0.7200631278160696,0.4718952380528083,0.3374551353239791,0.533254082348417,
 0.6110865778914046,0.2972263796242408,0.5994307009463045,
 0.0010251270661062462,0.32737528496647106,0.4260393425390333,
 0.2857927017304026,0.9482081960960167,0.22925537249929595,0.9746621384720079,
 0.32314090655327565,0.9651304534975553,0.4072114589685947,
 0.21958594691070688,0.16748091982357793,0.06117136387222322,
 0.22234154986286592,0.3909030275726617,0.28460198541347614,
 0.24504827841495924,0.9549907373300793,0.053889059455238586,
 0.5249591760559029,0.5795355501319921,0.5678259191768601,0.6394299247471377,
 0.33304079244619034,0.8090760789667593,0.13921831546873542,0.552351286605971,
 0.9746575012228993,0.8682033417618258,0.4897898814354562,0.45929930293776,
 0.6232524326360332,0.7894653915629186,0.6939118022853562,0.48479253529867683,
 0.913627166389662,0.13273022017508107,0.1796911128665314,0.29353941957478247,
 0.7472833583941614,0.24434811957284874,0.763596207447529,0.3344254636148606,
 0.05246707865608491,0.9233197921108756,0.707739204322112,0.14955699321647942,
 0.41116348941366077,0.639631726619071,0.2255458122854932,0.9194941574696615,
 0.8386934638534536,0.23449969856480157,0.3361744919429619,
 0.08522720987423005,0.09181603979971487,0.6104798797449877,
 0.40218090574757415,0.7181981267107652,0.2595149925643002,
 0.20224778810503097,0.917037226206168,0.6912336837320814,
 0.013809480023037612,0.3863653592521932,0.12396121146689643,
 0.41073411869737275,0.25370006462134875,0.5942205106807378,
 0.004771799938230958,0.019677632111429255,0.17131721919165743,
 0.20225520050422008,0.2860559976996645,0.7012660109606258,0.8735348818829552,
 0.6562388460502655,0.5401277773560015,0.4726070496500394,0.7639058010721952,
 0.44312630206650994,0.2874175966910121,0.1759751332214563,0.8933775819709533,
 0.9104494859724946,0.3754270881465064,0.006695776301038792,
 0.4659828548428743,0.6006181953901004,0.02981107056203891,0.6270766853217982,
 0.38495552288991464,0.2294417803383194,0.46990362493276827,0.37228008070994,
 0.2996528578011791,0.173887154912985,0.8430134248529981,0.6992270335245254,
 0.0921906098901688,0.2248869373425506,0.6572367997518854,0.884173451807325,
 0.7288881728941758,0.483092596548116,0.4431423523370259,0.7143026286281862,
 0.716415974886795,0.7684895361401961,0.9995417387729975,0.45663595802177614,
 0.9433994925974655,0.14274829253003496,0.7874615940197,0.432021635006108,
 0.5656512988643452,0.4421140224978146,0.8405000472909944,0.6750861991759949,
 0.042693852321135206,0.642477562113378,0.13920432719084164,
 0.25999038847930944,0.9851251694099469,0.1983375731185315,0.5895950071110373,
 0.035079861852930105,0.15560833945513664,0.10661317091571243,
 0.16457421372020553,0.13125193271208035,0.5818709806242399,
 0.3778506073375487,0.14691471582389237,0.9068431413772738,0.1337617517250238,
 0.7046464059474594,0.9201204583567832,0.46699085356413195,
 0.27395095113465207,0.6819124102160267,0.03983797349427676,
 0.5263358249999015,0.255524919925156,0.38102815488358643,0.2359634608321436,
 0.7691838974409464,0.7963751061913543,0.7183489340786928,0.7048364392785965,
 0.24916638729817087,0.7759128602318928,0.9901072473542919,0.5655350680195215,
 0.5481490398030576,0.7311858671137739,0.6918816612267265,0.8756176409951182,
 0.5301626572280111,0.806795651978588,0.7706614298225619,0.9035886505779324,
 0.22721511136621297,0.5791475454270968,0.17556740363447043,
 0.9729784418810451,0.8742533816038851,0.8666880674751634,0.9436443106156032,
 0.357964883853254,0.8663301363019412,0.7091461614828236,0.46080594812818343,
 0.7137491805674927,0.8742092356317563,0.8747114293372732,0.34611487800645724,
 0.542203332399024,0.8567052901911071,0.8620578687571464,0.6134126263216111,
 0.7335502458258932,0.3293968431523824,0.8510582170920256,0.5346420194952277,
 0.03806957482656026,0.05351038854694934,0.8613204967666754,
 0.35975736237295364,0.1820860638480971,0.8406568631615092,0.3791531870855812,
 0.8635622857269663,0.42691407121733604,0.7664662645164302,0.6158345981906501,
 0.5395803253820044,0.16789265818565613,0.71344080405597,0.7218727826845769,
 0.2638437158318164,0.05863465127855927,0.12558131796100935,
 0.9946166528946311,0.28997319758769446,0.7877869513410471,
 0.35584995541808495,0.9049870710918506,0.8558539309201804,
 0.06673342171753005,0.6037473179342895,0.5828430888414785,0.7290372970212937,
 0.5400308457322456,0.22168142564604898,0.9767142091731988,0.2775451476457518,
 0.48656571335478993,0.8218256052425968,0.9041073728061948,0.6535428501512173,
 0.7608353266202756,0.7061195129368513,0.2245727630722445,0.44942676381529256,
 0.6945270595673388,0.3916205355984923,0.3001280555163186,0.5663647757109863,
 0.6678273852964782,0.21353757135318496,0.8320261097621124,0.4756017443174706,
 0.5520056206164985,0.6684524593907251,0.8361909246733581,0.24900484657302613,
 0.6224199953824727,0.9310911300359078,0.3731329582910713,0.8185139390251115,
 0.8895692545976227,0.6015480619572585,0.6399086555027759,0.2254192422740664,
 0.4436278052145707,0.048736714031266226,0.21607992143775545,
 0.8674505524464198,0.20256160767930365,0.47193626052170234,
 0.6518009084108667,0.2662482007050405,0.30214228997013626,
 0.42618862435529603,0.20711543935120214,0.22140354101181903,
 0.1759957445188176,0.8712306447765138,0.9514116652484274,0.529455363088784,
 0.9349362695979604,0.8686915032343606,0.06401643370042354,0.4278623316517771,
 0.6911925565950595,0.21020117880280886,0.271307974448999,0.204436332686468,
 0.06549841984779159,0.1276444240854504,0.5185675739827902,
 0.30373911242243234,0.7234078615330599,0.9138745180251627,0.5203204259984378,
 0.2995495438043315,0.6418939065390254,0.32304579348142737,0.6088005894801316,
 0.4224545056991976,0.24629649350927219,0.6186523988865623,0.4112196102773944,
 0.7945996933054063,0.9771677284948403,0.8960099762832341,0.6968541055194881,
 0.416529529083557,0.44510612309260134,0.05240574641081719,0.4697883428649786,
 0.25718867457095207,0.19609642064809985,0.9844054786384051,
 0.7800774658589039,0.15151835091022192,0.6778399155681923,
 0.11028796484091608,0.07899472981084055,0.494605815239169,0.9959787760738543,
 0.5165850998934897,0.4273956556628099,0.9519457820089753,0.11070888586631533,
 0.879181716725237,0.9148381839016823,0.405558465805973,0.35226136187757096,
 0.19397379090450595,0.8762726805142181,0.5517439709012086,0.9062478242414718,
 9.981777485057464e-4,0.4992207914918537,0.5646116767420026,
 0.5419629028744393,0.1515460922369325,0.10736844048844774,
 0.31451362896612567,0.33859891123726604,0.7097082489096991,
 0.4599466758489327,0.6364122832350939,0.9893385378582389,0.7530639261978833,
 0.2255307065683596,0.5126080814671018,0.5504171617889225,0.5243786006972821,
 0.651992332136438,0.654000110726297,0.3763674995021359,0.8104567378129328,
 0.5297296630255806,0.6745964880339482,0.5623862127885668,0.19094592259501442,
 0.6545709085876195,0.5047403795129167,0.6740429353136642,0.6208695919769207,
 0.8508484465599767,0.11375762658406541,0.4015354000158663,0.8502949533032873,
 0.6809086611585196,0.33727766230483647,0.7213499257444966,0.1416048526482645,
 0.893774785090397,0.5249193072533451,0.27757121718587807,0.5591682607063023,
 0.6361076710094157,0.24983293493038383,0.24888058754655706,
 0.07473488627416647,0.5559172810180012,0.21335235311488643,
 0.7259136070602785,0.8531534978597063,0.06015799151846535,
 0.024936666720611678,0.7200013049349595,0.3175345496700157,
 0.20760783022734186,0.12175472206248106,0.3738023045111064,
 0.07746845483495413,0.18665839011708918,0.3210902709776662,
 0.20086487415703802,0.43021779537720184,0.6476790718248324,
 0.5924562314470426,0.5641811326557346,0.6217877066090142,0.9232157798208895,
 0.5697923360323356,0.9787372787493033,0.3054473994126923,0.3353507159326031,
 0.16568002392962933,0.44350962431076646,0.09480596941563657,
 0.9340193157569244,0.41293485032774546,0.883342277597843,0.1366662170811137,
 0.8324888742668726,0.94153285089971,0.17497923691764206,0.7228039765612247,
 0.9720321820338856,0.24701865481528007,0.6390277444159834,
 0.31697489584614524,0.7329310413128383,0.8942136807912365,
 0.10806435016422988,0.7000051246751638,0.9099203885675563,0.8416636632095765,
 0.5772769785368722,0.3538836433610554,0.8756071628746341,0.6663700985862168,
 0.2390476751514603,0.393350026482702,0.6458080247811671,0.41446378992628974,
 0.39036205916220157,0.49463222688901853,0.5680465918607878,
 0.9566905324235283,0.9436926731220097,0.012648809588556964,
 0.1274230665597227,0.9419928486604817,0.34944555410730016,
 0.14579861720908882,0.954687377283677,0.2842340885766431,0.19554290032174193,
 0.8406213409060063,0.6279346983537766,0.07707755539124905,0.3209250667714827,
 0.6488431593891782,0.34199044984958127,0.7097537313787473,0.9233118047924722,
 0.8578021668646995,0.6171709782773953,0.1988922508796972,0.3023287008773565,
 0.8492048554795898,0.7096468913538841,0.9203267232648054,0.7082760520912352,
 0.723237576614973,0.21545801034034184,0.6007452171532219,0.6904375679945269,
 0.49655984682153187,0.4130784543376518,0.5989091821485254,0.5124100412434341,
 0.3243113610350091,0.8640172951432543,0.5683350153650342,0.16620004177119507,
 0.8115177025568278,0.4199646586234693,0.5986456861173135,0.5712082050268338,
 0.48466492520579585,0.5151286835937027,0.8847131663994362,0.5093252705970419,
 0.09117364897030378,0.631262436865065,0.6015106114595521,0.42362006690922416,
 0.6793176139893989,0.1267005417673781,0.36652330783317866,0.7712748643776662,
 0.10808204581315595,0.6565838473088185,0.6702892560574626,
 0.25185014540068607,0.6836804613277255,0.3604795611305228,
 0.06020464051220786,0.31876617506376326,0.34277217882161637,
 0.48194639197789935,0.08941387477369556,0.4449800266982159,
 0.2573134437008733,0.38201014511385223,0.4286291250805777,0.5284932649140875,
 0.3802506511670325,0.999911635643176,0.006374054952187791,0.8504200912693756,
 0.6483380279822593,0.8065291407639132,0.6545342264836984,0.5545272444938876,
 0.14921798128219987,0.23676801055228114,0.012137780732940917,
 0.48677365884057533,0.9404973460793391,0.6416211579550397,0.8685731253324286,
 0.19260784151497712,0.09498029541487507,0.5066322564599645,
 0.3643839232158046,0.40558990872833856,0.694498943011323,0.9866365305425374,
 0.418925430010755,0.42441593785427567,0.7128615860149046,0.6588439496165057,
 0.2698344947208329,0.6936412063319213,0.3350864348266568,0.8420117367373102,
 0.4871421172665592,0.1609428539119001,0.4744117495978972,0.8396020775783182,
 0.6947074639509921,0.04638987069296607,0.07269465733214997,
 0.038987544420415965,0.21245209160141076,0.5053303376730489,
 0.9634555198546788,0.4297539052719366,0.8373693513864843,0.32499244618251144,
 0.9542892679481324,0.2494960080883758,0.9408741311866704,
 0.056179045009367434,0.9758704613604284,0.03636542805055831,
 0.4221464837194888,0.9938540544972296,0.3703228936418632,0.35347356696481214,
 0.6859279148468349,0.865346112849732,0.8257822373805978,0.990502827532531,
 0.6885827337198018,0.40004740276455597,0.2324529940619251,0.8022441371531195,
 0.052493194318659064,0.3130362957063806,0.5492205977547987,
 0.8571812131012503,0.1865755576068282,0.7058423390511224,0.04702267369919222,
 0.711314663188368,0.42477456348780074,0.3930772001592697,0.6714584907779563,
 0.8951440840629481,0.3526144129009855,0.2652079439931794,0.20712597561444102,
 0.32027741004154264,0.06118188112094625,0.555799986521528,
 0.16407862838132448,0.19209784643701555,0.3886053692957394,
 0.9933124749976915,0.044838534563497956,0.006787213771456191,
 0.5588544465278442,0.5413549154011914,0.49389708720515224,0.7938651794127524,
 0.19228831819429948,0.2918123530766754,0.14072712082238914,0.397954752330415,
 0.8547141726810373,0.2982268476314447,0.6950850306707386,0.6882996436543378,
 0.5680567392609325,0.01481553346813036,0.004980005428399159,
 0.722630470100997,0.31752062517077995,0.255575488317081,0.6035384763994809,
 0.6384433181630174,0.5731337846693805,0.2799652170934148,
 0.005602992614696323,0.5115076569325091,0.1557440948862412,
 0.6186366201969081,0.3036258220971544,0.9675449653272518,0.6449734765637682,
 0.2494995912572524,0.6338212538560548,0.6806374590916827,0.9294199406116853,
 0.18340322786781682,0.7354180024939604,0.9696290850766774,0.1211854349640471,
 0.0528754195118889,0.1689498832824271,0.17867826766844386,0.7071603956427387,
 0.2654313865714041,0.05162654822332502,0.9290638974579065,0.4435340619391408,
 0.7519504693502419,0.7042989234911733,0.09319576682088782,0.7184857774830312,
 0.7140812697669666,0.5289877670213108,0.4304020127200321,0.4419703601252788,
 0.45420829665439677,0.4384353002443687,0.8986492663440468,
 0.24348125313303293,0.8915810583200852,0.6837305945292789,
 0.12646530780153575,0.20913620848425785,0.9875361502591384,
 0.058223692687644046,0.8580134106197987,0.664248796980794,
 0.34698486804507245,0.6114965951427956,0.3964774495312847,
 0.058254855990948595,0.6490486411619449,0.6085013509174192,
 0.6359013068915355,0.8382917187025212,0.13847142020057246,
 0.12072061113546728,0.4056263479404363,0.6824400924129019,0.4091381262145699,
 0.3220604432226758,0.17774085904115722,0.7059433898278452,0.2486014018269591,
 0.9269644332932506,0.4203309503444368,0.268152563996511,0.300986010627203,
 0.7247490588935919,0.5341225740948271,0.13451588540236092,0.0667767776870366,
 0.2875021732989119,0.28451252099390567,0.5740494702824689,0.8922316414683911,
 0.22867079542725322,0.408705642246176,0.04588396906675418,
 0.21938942731267308,0.8915691091335363,0.8429116806078473,
 0.36120267757044333,0.32696591579183976,0.7119927551873172,
 0.8192942498061003,0.8477853347533852,0.3967400442568354,0.6217262847225202,
 0.20609094828426078,0.8966182506876587,0.23310091109183095,
 0.9684800021950798,0.381036084004186,0.055976355248870235,0.311130030117451,
 0.18224300078188171,0.757694073157971,0.9217737134474313,0.21283440374615936,
 0.6602718215728334,0.16147969016443386,0.12998251221457302,
 0.7947753578304226,0.9614430622568827,0.27355524212119686,
 0.15618655120574454,0.6191901313591359,0.5897185370137873,0.1656727843191006,
 0.7368269705157859,0.625730497144299,0.6209500478875829,0.8815157731538796,
 0.5409093825076364,0.16462796177458705,0.5829791941297398,0.1807051886310561,
 0.4937262267041469,0.9723157373064358,0.7882600896740924,0.7682110358943082,
 0.8197443781421584,0.40287953783616015,0.4773610468412617,0.6883992411053756,
 0.14347317936988613,0.42207477698831986,0.5778093756862389,
 0.0671491230346749,0.31312778318194023,0.19238406874932024,
 0.7357437982412133,0.9120417465067646,0.22540991078639827,
 0.27831755871118025,0.6664345636899587,0.7824128579083997,0.7096716771695403,
 0.9733287952696205,0.5390713202748568,0.8516917330913272,0.28808177329460516,
 0.002149502223505295,0.28972264377516754,0.21908974930268543,
 0.4509369414629154,0.05381938697582833,0.7184658658271341,0.8297289943621767,
 0.6782708501058328,0.520428989709023,0.974996027236497,0.35444702386181026,
 0.3102995015096026,0.22707912756595505,0.011737038635243446,
 0.5756509520930956,0.29131996534292337,0.13387915568708675,
 0.6054823649440931,0.7599110794435074,0.6916590492455361,0.10256064649337382,
 0.4025620215414484,0.4872049565299674,0.6253681357351184,0.49597493092797496,
 0.1255091874495362,0.9819443445122317,0.11554039283013706,0.4065049470392895,
 0.4845652797809197,0.11250295189377302,0.9198049855719556,0.7318401924035736,
 0.813758627086271,0.6686693064426108,0.6466463196288134,0.5399667044457803,
 0.8664447991398732,0.8368648671232817,0.5208135002563826,0.973938736009576,
 0.02936208530504103,0.8965442718319104,0.9423735602786305,0.3083130448183742,
 0.36747361558263325,0.9280245101421573,0.10617982703978557,
 0.3751441524749498,0.22137674976797717,0.6932791444722491,0.6886964011971962,
 0.9095678042592414,0.79220463874563,0.5596786787240938,0.11471783368729249,
 0.6047418646188157,0.2846271959028116,0.5201133200026105,0.14516109039789438,
 0.6555675113537568,0.10502591795268001,0.23245703647090676,
 0.02201906682682786,0.20956266685944047,0.005124982732138128,
 0.26941301320324074,0.29875330759371266,0.10911781312267843,
 0.2292676623162757,0.1868610950827344,0.48224080852071705,
 0.22795837262605068,0.6846999822815603,0.10578704438278719,
 0.8069309267975121,0.9320191756733921,0.19185895731535874,
 0.28782359498276044,0.7694448046241282,0.8769096070590896,
 0.09664651119985967,0.7431464179528249,0.30185096319045135,
 0.17855527173581331,0.9551065574320673,0.4201475406228917,
 0.40019352143824316,0.5172934403532353,0.23376496983624728,
 0.5169869102404231,0.10872420638012348,0.849387591347676,0.20278943425279217,
 0.030017288390256303,0.4113056158655859,0.6482370932159137,
 0.6954235756420409,0.16046180326967474,0.6073909786829099,
 0.06869427268418782,0.371650388600568,0.6037215001091012,0.4018713171012628,
 0.9787134993931744,0.14428552873634148,0.7038244065760862,
 0.08060254118983923,0.5574647976310757,0.2235967851812246,
 0.44025866329059193,0.8783558078500823,0.8282853009871793,
 0.30393595511379545,0.21718369608088783,0.37034704968675747,
 0.4418114831601109,0.8615491673410981,0.02968092770621844,0.1791129865466926,
 0.5272849517246073,0.832245898698691,0.2480433787635723,0.6940495462982319,
 0.4806799039551387,0.43547419938888277,0.44132896471113847,
 0.9057866369034434,0.9592993626728081,0.7800277791659387,0.39879561233313354,
 0.2410475206823579,0.09931030284822329,0.6550174367914943,0.7888734904909618,
 0.7103919107721557,0.14040936641240076,0.29781008573339285,
 0.5030851942192989,0.7634497420487739,0.3435219176325722,0.8828714419288511,
 0.4213526525924052,0.7071384234769402,0.8690051603822225,0.2933974469246654,
 0.6407065265520332,0.7735722210850154,0.536137383706669,0.564760940736675,
 0.4440538652457324,0.5999037888792957,0.5623684069257333,0.856885038103561,
 0.9041275427597537,0.6607664236778992,0.47370681874069254,0.3384371515649228,
 0.07745499832003033,0.2604080153752488,0.06075824397157481,
 0.5549955566762097,0.04371593226617809,0.18637743629190096,0.678704465155555,
 0.43624851733998016,0.6714102799332092,0.21724941516249252,
 0.19775307235818607,0.5232771830087829,0.4552411380706344,
 0.14397793103833978,0.878905328233639,0.18971218859663197,0.1667127840842889,
 0.718033461314207,0.45929166319593395,0.960459757692689,0.8845377776817711,
 0.18310617913673588,0.19550815647954978,0.3894125990205236,
 0.3090228636245014,0.21285625308981782,0.5353897534681105,0.9589838365787137,
 0.26422416340016264,0.15163978112342436,0.7274882792923073,
 0.4528333942880627,0.9251512961144366,0.6028304267523219,0.9942054285096998,
 0.6346492873062493,0.5667572867667008,0.02101786830951946,0.8809943936877354,
 0.3921899968596242,0.6217258546732245,0.5040133474865698,0.6979995800689999,
 0.5016889811239297,0.434466559039866,0.596008475263541,0.29927462268090177,
 0.16486828749643123,0.11265763112084104,0.8244464431641003,
 0.5268453556613439,0.21843754498714785,0.11784792994871207,
 0.3742550188574595,0.7349436849050277,0.4802568416937114,0.17691280853987434,
 0.7196907966058519,0.5880836610914475,0.1735304774800681,0.9510357498390786,
 0.5784625814203183,0.8787639812262888,0.9777946840279483,0.3971470756512634,
 0.34818269489359366,0.20096095761540478,0.9861280893047537,
 0.14891145221978563,0.2400679809525108,0.09205910959664398,0.394166588613585,
 0.8984995729186274,0.39650040752612126,0.9213413409752452,0.9900290183812208,
 0.44254269290103265,0.41862994931303776,0.09703316177727839,
 0.16404579281651843,0.753918739671969,0.6142930838877846,0.8252232004049814,
 0.13077810540867474,0.4660857262912548,0.6785988375956941,
 0.43997182916992794,0.6853234350232701,0.6721311255647497,0.7964912688127099,
 0.9795486523316479,0.36010855670530295,0.1183731321046303,0.9640526085944413,
 0.9455241677743105,0.5638388269196255,0.4317902591853696,
 0.017119439688634053,0.6937326887524577,0.669322698669723,0.8955656475516754,
 0.23177602040957512,0.5206593918655916,0.3611620673122774,
 0.11038634625365629,0.15790465066414694,0.18755681887139175,
 0.2704620883860358,0.45677234815964873,0.33430008444557413,
 0.6796921579165206,0.8961038914084679,0.10998331721678578,0.8376912784619356,
 0.09367745682343509,0.5262977298541128,0.3633861143277728,0.3526584057634453,
 0.8635627287930515,0.22241050020114406,0.3870166413760806,
 0.17366897106708568,0.7365391467725318,0.26867298897216485,
 0.03958516140096924,0.5653539701058603,0.8852636224194633,0.909549511459564,
 0.41267475491881556,0.33502633300894613,0.6158831236358411,
 0.12754371928298325,0.3243090448912225,0.16908957512179124,
 0.7166586651857368,0.29536597895483374,0.1477352342718199,0.6607807935127566,
 0.2769000378156026,0.6579639249610337,0.6745007663831841,0.2602556439201499,
 0.5031093291619956,0.5930283593258038,0.5675069007346496,0.42966826464091357,
 0.8224462328823052,0.9658519389976485,0.3331247706528744,0.3327528751551442,
 0.2952130486316418,0.7518412302796479,0.6745246230255376,0.9067014858971671,
 0.6459667100592105,0.3630596966797007,0.822564753952788,0.7805671325162503,
 0.1166745060001041,0.4747791534598693,0.20923738840696082,0.3560733948716339,
 0.2068169132635408,0.4059499899224651,0.6153943407745771,0.08756639912815023,
 0.34544598883599664,0.4873424714545874,0.8738126451622259,
 0.15941745603687096,0.4853183175317177,0.24608487159091352,
 0.9555884237897189,0.20511651974651446,0.5401584072477592,
 0.02821001950513158,0.2459682547158114,0.34723415478047803,
 0.29879359033635977,0.1394602344368555,0.7953528681668143,0.6336300466915894,
 0.022255135479052446,0.17801825198430266,0.11839574517185725,
 0.85306939749893,0.12065724845794767,0.9766530039495768,0.4272126071720952,
 0.6782154536584737,0.4555477415349988,0.08221336983118999,0.6246047414762765,
 0.29912641840248855,0.136172588457941,0.264015523782671,0.13932196294539123,
 0.4571878263727289,0.9745335210161332,0.5838335242772146,0.7639199285316003,
 0.06390310146267897,0.17624274385100813,0.0585020383160898,
 0.9305519794797584,0.5269823324921608,0.4375851248253335,0.8545468673125112,
 0.27440288297910753,0.04863294840815158,0.21481769805014383,
 0.28882632945992914,0.8055397354946112,0.9715680465999014,0.9753639215529041,
 0.14877882943539444,0.8892935120101799,0.2710075559202716,0.8933175001457668,
 0.3865172929082471,0.31120660330357386,0.6114437889962563,0.6138713703976584,
 0.6546791367596522,0.7748861660537442,0.9718144924632426,0.35022844449353097,
 0.17661006492339815,0.638437114551577,0.8340737092727543,0.6625884180985,
 0.1330418351873477,0.44079599843292105,0.22798722436586794,
 0.4981102079848546,0.7376968476746903,0.26872534964231654,
 0.18058972488693936,0.27735390318437403,0.5300494483857514,
 0.029321073440985224,0.5416621897413548,0.9303530937433482,
 0.9062250787325268,0.1687172350444457,0.6369303724452077,0.30641275475305074,
 0.8319980069311913,0.12098793161770849,0.3734753558060462,0.5041679184711539,
 0.7174988930676215,0.3939820397423044,0.7626866053429541,0.11151048844637712,
 0.30898193857707046,0.48359814316078387,0.23026230775311252,
 0.7431051306536686,0.3611083340387784,0.7698532203950457,0.18707791789154737,
 0.4407407615924688,0.379856563338725,0.7919152229565551,0.8911986902155875,
 0.2653156938978434,0.7948676670091926,0.5046142721924639,0.2963522823568058,
 0.9775520045548294,0.31115610015246664,0.5576760375913395,0.7032260278920338,
 0.25204564975805666,0.72215644586817,0.9542197985782567,0.028931906560338283,
 0.030496404636074836,0.25660509299724166,0.4457411063343819,
 0.3677527800849082,0.42804950725434643,0.6376463564418197,
 0.24152888186845578,0.25874586647525866,0.6305457123509446,
 0.16821331156217945,0.7550111605728457,0.09083647559633645,
 0.8424655437662859,0.8567437909272726,0.379706193767275,0.5563526984872458,
 0.2629108817000768,0.5153575367877979,0.09007459914224936,
 0.043144361241877816,0.8099780370724938,0.6312361166974567,
 0.9437490389296982,0.05827148868890464,0.28116102563770706,
 0.6059911002471918,0.6498402927011828,0.26376751033314083,
 0.20164106249042768,0.9997203989497324,0.8997225153394443,0.3952225020516046,
 0.3765910425885137,0.9229374748447453,0.7074706360171905,0.3924979569202768,
 0.18896939480014563,0.7118066168630687,0.1931353439714104,
 0.06025229451801062,0.9987755540785941,0.1879675564999066,
 0.16489399108493408,0.8723923000559834,8.574091552384644e-4,
 0.7304361008647218,0.7476377240315578,0.9386209897379142,0.8757607922415387,
 0.8092069893766188,0.9314638845149469,0.15242320809742782,
 0.19071931597718728,0.39429440306674013,0.7294580931697318,
 0.6407696124921877,0.23132728340969055,0.7454523464751572,
 0.06263748758879462,0.3776849959794749,0.21801327613299581,
 0.1533054087415715,0.02458558136403799,0.950513652118562,0.7960132996706971,
 0.7541248439061926,0.4991405149612016,0.6551679181706076,0.891230943888526,
 0.2149086064250787,0.9105085890296764,0.44311318517631315,
 0.16712456768522566,0.15220927880738788,0.4820372167511765,
 0.4732185554628209,0.6741746392861859,0.14459593486203604,0.272771626404685,
 0.4778761215655667,0.18593839546410706,0.6927532129060875,0.6751434062781094,
 0.37871714973615556,0.5797391127004177,0.7502536198026442,
 0.35680836099184265,0.07109786487047587,0.8188536406319173,
 0.46329807095919207,0.21790941925376006,0.15315438545825344,
 0.07086847557510523,0.16877079422066854,0.7648216562974053,
 0.11464921060900735,0.3557395041582745,0.550549174130752,0.6056374493415844,
 0.6631111827491961,0.949630768289586,0.9120340582700651,0.05544329241897772,
 0.0467315702716683,0.5703936163747618,0.838116392408606,0.8259206852934932,
 0.6581440458609231,0.005349549199475945,0.34222140266320844,
 0.33893988979126133,0.3986421535009742,0.27296893984703197,
 0.27821998541002946,0.989526747517185,0.3005913256328998,0.5247613052602149,
 0.7577994706125669,0.1488270719619924,0.08727252810104558,0.9345589558961601,
 0.98477498302807,0.3405263766513027,0.3973652400170278,0.3962769788845373,
 0.9469559447638816,0.4122405139189741,0.29733350401052616,0.5279026639968462,
 0.1467959715531144,0.47047070438784455,0.1983595144239758,0.5506387039272337,
 0.042302051624066195,0.48605416312069594,0.4480679191944028,
 0.35788297261077584,0.8172914511090111,0.20962881597329575,0.846690046530567,
 0.4662674392088816,0.30909076507816835,0.02467421906078493,
 0.36492225676006207,0.23250237238224714,0.4989156868565219,
 0.9634685419280418,0.8733740022652992,0.38914929941090337,0.4526123663568774,
 0.16918218743013846,0.944841970903159,0.5967042047893173,0.08728345352637379,
 0.21724156888777846,0.47990813441659197,0.9377244649644293,
 0.8221041954765541,0.044612959142517505,0.6883541778314622,
 0.7453904337191721,0.18439584921555296,0.9406050522329219,0.7539560928219851,
 0.3686340427211494,0.6563163667063133,0.4178873359431138,0.6223363783532614,
 0.95463523657186,0.7895453464806564,0.6861303560579116,0.9755027133147746,
 0.18093145154951917,0.6795179938344708,0.09306656749976261,
 0.9055907073952606,0.8482574392850193,0.5000470484969273,0.5502717210885126,
 0.7650444500291761,0.748985709154806,0.11530007204585013,0.31814978734034804,
 0.47052582327485404,0.4400252201050585,0.8779039904020441,
 0.20470711903381122,0.11177838833432008,0.42705427570712406,
 0.771488298497897,0.15331710244654717,0.32622723402384346,0.2867208455342396,
 0.3705929577346465,0.7587892616161126,0.5025230976346059,0.5912766018211646,
 0.9286591368592025,0.991005313929385,0.7303631954969085,0.9573942352384696,
 0.5031140840719595,0.8419506966235769,0.3164128819146268,0.6778962892566681,
 0.5656467806662109,0.2086334541698811,0.602232844328762,0.9269059452849644,
 0.7392227714764076,0.340443815586867,0.7345827835708448,0.40854816964568486,
 0.24059166076111116,0.968511044790448,0.6171727803018133,0.5012804250194155,
 0.7924135473516047,0.9608385682239584,0.6168068298466277,0.29655466581083534,
 0.240799399492567,0.30035834172408404,0.8511979524178732,0.8276300521153394,
 0.3500399728826127,0.9042319879273258,0.3183933479039086,0.6371459895075214,
 0.807456428850925,0.7111691242501775,0.1826658871758441,0.9914121878234545,
 0.6218122155333514,0.7170044109365322,0.25822749556591496,
 0.49837793327179136,0.9546557425479806,0.2570874673628538,
 0.16778971722188918,0.29659767363761524,0.666756640675157,
 0.15060629579456009,0.598109556622708,0.5495818318685158,0.6247769417451133,
 0.753470986830065,0.9678064136597647,0.9289384653866236,0.29890282192534534,
 0.8518270001374102,0.20586807820859943,0.6136758895432057,0.5350265759813428,
 0.8247471098139861,0.0940894555512376,0.01763694113433034,
 0.051653634441346186,0.5986837091547872,0.3208234957363191,
 0.7162672818157504,0.2493550158080643,0.8007638292070179,0.5646043999762145,
 0.26758842131073957,0.9151732600756776,0.10442624202335149,
 0.4287144726185309,0.12388587690656538,0.35834700456639823,
 0.5921876082782433,0.2804342385240757,0.33345312091373636,0.5220162885606577,
 0.7604419921359171,0.32738921498810614,0.5043778780674458,0.4133213480077388,
 0.5292134385846462,0.8331651968264828,0.7005461430130693,0.5259665599306484,
 0.2849077163398718,0.4801038727302931,0.1670582065101901,0.17881863550317,
 0.8777673591579974,0.9649384940132966,0.02339599223139277,
 0.41511415800447304,0.5893270080272848,0.22160993096059445,
 0.6235313701080692,0.1506633961206003,0.854038054089679,0.5587671994301264,
 0.2970352504654914,0.8223537163510269,0.6014192471077554,0.6738026507600796,
 0.5942311618986675,0.7778433113328976,0.21547029763258818,0.9322697106023072,
 0.9284904336776363,0.5032977619141684,0.98551525598077,0.2971753834082622,
 0.009441256202961767,0.03882358645666528,0.6584784466339064,
 0.07105273595232164,0.790738620282102,0.4578536001390372,0.6695539406986895,
 0.8801192023418511,0.676467005512542,0.5841273342200578,0.885252836847477,
 0.18519390636577715,0.3534463293937061,0.4795722326821854,0.2981928837660388,
 0.3335186779744892,0.7976912256063222,0.8583494400795902,0.6451498884013475,
 0.6771719590896245,0.19371430844542514,0.031111768367389714,
 0.07154175659500639,0.26801964458550676,0.23990018591958062,
 0.04414181663600936,0.9728255930536678,0.07035761371062654,
 0.2987163266399335,0.11133724363033581,0.32598851878943,0.9009258452236857,
 0.7048450836173168,0.2733682529522439,0.5533360305129351,0.4822274553042969,
 0.2408283535141411,0.7435555502048301,0.7585987029502794,0.2627906652099128,
 0.9343803486912929,0.6962602095567488,0.38629959832073724,
 0.29422146362994517,0.9096677774420484,0.5734617724140548,0.2901561708801673,
 0.016348905318658402,0.177357189233463,0.4651186499128772,0.3734369322566652,
 0.22996990240510917,0.4179725564858342,0.3729340227683553,
 0.02071419161788124,0.6748260872657013,0.15737016070438836,
 0.8599875683119137,0.08229374692956148,0.5339478424230506,0.4211141005921222,
 0.6102034958364837,0.5154273778374912,0.07738880755245092,
 0.06719756956370371,0.6613944656826909,0.7086908838601013,0.739424440855405,
 0.49908970706768463,0.17106795213216552,0.9990488720907978,
 0.2764908028920645,0.996962048174422,0.9430198596236932,0.7467096798749628,
 0.2669980867218805,0.3501353699112604,0.45244953105397934,
 0.43452409148887794,0.3049600756436036,0.4964357065323922,
 0.10719556770985872,0.3352070318443716,0.2705855221949527,0.7990807308493895,
 0.5594778627486953,0.848529000863503,0.05702977037104384,0.08235672413802031,
 0.19934804819191465,0.3778147384822257,0.8968671889740267,0.4015532889176654,
 0.6045956662982591,0.8798967736879274,0.19076218346493046,
 0.011530750854724547,0.3681224910154818,0.3455546835838892,
 0.7278717140751781,0.282230565442498,0.7011089417892206,0.10119079724347146,
 0.5315545612914863,0.20252977150223894,0.400135912790033,0.9361273495806892,
 0.3012093290573574,0.7223866448008192,0.6962088991437475,0.8528083558976242,
 0.5323621719904869,0.5599824623972727,0.3711321730657232,0.2673694069573782,
 0.029620769263847047,0.9351277672269998,0.572924872269958,
 0.031583642070142304,0.6791323505843653,0.3995426459198248,
 0.22827312019293655,0.39344486269763923,0.46506023639744665,
 0.8720154756975244,0.1831777866788349,0.23212662107587845,0.5143777887332558,
 0.7212923543098,0.9044926886065101,0.45644889038743697,0.1381145574255337,
 0.6063176448091955,0.33073555950331823,0.811465314194397,0.5294582984893696,
 0.2745160050669131,0.5468888181167827,0.16440744820330555,0.8233979729573995,
 0.13170169529621223,0.45760895678466684,0.4356303016729046,
 0.9605027062841494,0.9801299816707882,0.46782051532649327,0.8064135549585312,
 0.187227866369148,0.9003594978598457,0.5464348398984482,0.2487861104385003,
 0.6864439746513028,0.4598310903498384,0.04944200629020756,0.9181915597523518,
 0.6398352464738488,0.5647275462308864,0.6165151433651385,0.35677705210003285,
 0.2908791822897929,0.07841757923303494,0.5305451424348867,
 0.046324858211711506,0.6140871422314167,0.6892679934329241,
 0.2541746928369025,0.7428146875350667,0.12670668769556181,0.306503005592738,
 0.0813029308698392,0.5992389776457164,0.4609296839749595,0.5662912846300266,
 0.34205859238367053,0.8444690671528801,0.7928551842632906,0.7903797050619715,
 0.37003432556683646,0.756632311115371,0.4120138415211879,0.9389772935138099,
 0.4190212125704059,0.6461436112239661,0.41516638066556555,
 0.15386677033393736,0.09573757103050295,0.1367112644392866,
 0.48799266034427924,0.43982793612710314,0.6407872857193307,
 0.1525483134892729,0.6776602458058363,0.42031116406093016,0.7465894452821364,
 0.8596889682719766,0.620477110349221,0.8228754591965106,0.9553505270148436,
 0.8771955764369245,0.17393579072937293,0.5515014166428207,0.7669573631618571,
 0.07578578728093799,0.5012932169259675,0.3028171692950149,
 0.07757924360485391,0.06814778036133662,0.8473629425757404,
 0.2470535890582769,0.9693258442892466,0.573127565259766,0.2884612801992532,
 0.3245013151614694,0.6606258798056884,0.019899721564812678,
 0.30195960102093833,0.37951830638808715,0.08221116973173792,
 0.3809755557412481,0.32023456880659207,0.5146473210545981,0.6417125282049898,
 0.7066167082358428,0.7831192335855471,0.09872988371182423,
 0.47788695038290274,0.7849614195972328,0.6212873780303108,0.5990299303423706,
 0.25545402635668,0.052579689929531104,0.7165847379287622,0.17812092870780238,
 0.5823315555991743,0.39690115945878357,0.1260188081261635,
 0.03958818688772714,0.4311484751152156,0.9706326832686762,
 0.15018040914234398,0.03395123879435058,0.44715424772768564,
 0.8348942119758203,0.36155602813319865,0.3502645313723548,0.5356301989794486,
 0.5361605258937635,0.056803329864418495,0.7835715453530092,
 0.6078930981909217,0.7376959158185958,0.8388803583749948,0.8175360046799083,
 0.18520414646759797,0.6341223499449152,0.4500071862696,0.52606046050001,
 0.42404397275767125,0.33594775462460913,0.7152060782263712,
 0.0035018611217871864,0.8330540117443723,0.2951327322721411,
 0.6562264374483767,0.7161294066106618,0.41872050585190856,0.852369073414063,
 0.977315137314131,0.7258842483419442,0.07335940915747385,0.21815856670299416,
 0.08206281802764226,0.8220854425829915,0.9097588732488235,0.6164245032978802,
 0.5303581259148336,0.25111643242897475,0.20398250740584312,
 0.6538091305399953,0.450919437537289,0.7580298389653249,0.039605417057070214,
 0.3671490241847911,0.3198096289524397,0.508168027278625,0.8363175668521275,
 0.8306215434692414,0.5384823069428751,0.20172758241485678,
 0.08054891408438147,0.20760705069828678,0.9685705482679433,
 0.5674052996788078,0.9721072211386426,0.497464371375255,0.27956523331259564,
 0.44604076068505116,0.21880326855427912,0.9135073711332101,
 0.19819848437338328,0.609036178284224,0.42335998183344414,0.5482866638093357,
 0.6363367638346666,0.4377605852347225,0.9495005520523334,0.8036232795439264,
 0.687875411945713,0.11692432673204167,0.15710701327063847,0.8833579982518767,
 0.4219622529402929,0.961073952374427,0.22994069905441428,0.7912389545726556,
 0.7658760201884101,0.6459804914623994,0.22851114807721085,0.8475790660421321,
 0.30357759114598726,0.2124255565404456,0.35138249201982275,0.615096139365608,
 0.7899082120480941,0.06619009591938751,0.708636009984117,0.3014621545181191,
 0.6948031319659299,0.5967917339348956,0.34781715430936444,0.0613928872734717,
 0.638863049934815,0.4353007201848602,0.3917938178514533,0.6170939562849613,
 0.7757264223177109,0.5532572196016678,0.9536053265044327,0.2967550644704766,
 0.43707984600900607,0.49187361147074715,0.921355867773845,
 0.39787432161505565,0.14815734379156364,0.39021661424634335,
 0.4550685726001341,0.2701032659531626,0.11364173456195903,0.6882175923611241,
 0.8322631298487616,0.8879821967442613,0.20327296994869326,0.3920282190507328,
 0.8441481519086328,0.8303681433547762,0.20031389257038212,
 0.31824437486977564,0.6803775435237664,0.06802335664564996,
 0.8623073924797608,0.9553636687311814,0.9734373382301753,0.22905666309203732,
 0.5583494986388574,0.21202052872572585,0.033224944090557096,
 0.2183551751321492,0.08295223002600005,0.4788219487815669,
 0.25808822593069225,0.20780812375709523,0.30977055101677486,
 0.14238533019395472,0.5246737556401163,0.028706062775566976,
 0.8575628885258765,0.8973348811785217,0.2215500423723673,0.9138865735880753,
 0.8696415939135109,0.4044750866836312,0.24372677820962418,0.6047866197813825,
 0.26605297329575694,0.8607157340829574,0.08956339527541446,0.983068207164411,
 0.9713990482837956,0.42553122845002034,0.23035108635327362,
 0.7325586057215993,0.6720365116272015,0.3126878572466134,0.7425638200692948,
 0.7050667216115203,0.40282204869452365,0.7823568126914973,0.4798072353033156,
 0.14146950783555923,0.3775886402380093,0.7182097145810972,
 0.19424188516354313,0.06803367430437723,0.29550398327988914,
 0.7112666661162429,0.057632513243959016,0.4629246774267133,
 0.9983058174493173,0.48261689062198543,0.38732584152212424,
 0.8957102578766767,0.4372107661792539,0.176275266731991,0.39065825678395827,
 0.30211557311704673,0.3979022667717025,0.45574209907425167,
 0.9516176728782076,0.5339304965997396,0.47435517360708745,0.9362676772866912,
 0.1537933152123252,0.8958642309446665,0.14241784662131285,
 0.03068510240500999,0.4890442140593402,0.4631316503135434,0.6121375365538846,
 0.7438442948478118,0.30852045790740434,0.7701442728938918,
 0.14087227456874007,0.9819719561660902,0.8399341206478034,
 0.12722435629854834,0.5704779863478596,0.4314751877323857,
 0.44392759724551856,0.8615414377752184,0.8215492976728234,
 0.060291762797510806,0.21083879639833636,0.11037157583473622,
 0.5988173429617674,0.6707319842162568,0.14718070975635866,0.6432829589572205,
 0.4931495582370814,0.45913288218161963,0.1146607058393796,0.6004524329161571,
 0.2346771677377274,0.2561634640183298,0.7206176110718825,0.1715343534661945,
 0.10726040061457287,0.09042460618343795,0.7005001656352722,
 0.19532484653931947,0.5300108422464701,0.43953084987284563,
 0.38303623170945045,0.7526346240065559,0.0647948801485172,0.2310734159793032,
 0.01795673392848185,0.015173181603234887,0.8727688987118456,
 0.3552849528702585,0.5236098067799578,0.26573537662405766,0.9771898989260743,
 0.9487108810915319,0.7052328346543827,0.18155272427801417,0.9206888443310722,
 0.3223393164076338,0.5603054625912214,0.08966666177378124,0.8102913090323378,
 0.1876498799063745,0.23395108004061982,0.6576682915243343,0.9647283818290021,
 0.8744467244190672,0.8989489342178496,0.26841605381531475,0.9773566787461065,
 0.5690820463531289,0.6424190647777489,0.8914494170098668,
 0.0023340247462219077,0.01457943470986578,0.7014232942397244,
 0.7137735537644956,0.2003484404011301,0.23093339088251486,0.8761306897661323,
 0.6520425499759821,0.07385629659368709,0.30338968176794234,
 0.5661913326859243,0.452671935174225,0.6648647089563546,0.8100405799486314,
 0.683089991920575,0.9301352202262667,0.10591387532816177,0.4953668831502853,
 0.11441261296599692,0.44282132094496685,0.3212494829528191,
 0.5360806592952696,0.1244483149132315,0.6059719756791914,0.6961538520694363,
 0.949552620691823,0.8204868615554384,0.12089500279871468,0.7777971884268341,
 0.21016735385131446,0.2810721411950271,0.8350887158658467,0.4005564793643618,
 0.35992318603634144,0.5324386646395192,0.5263137792406901,0.6199233643036889,
 0.8924900308709682,0.9424665746723414,0.6607253607314562,0.8551646705313887,
 0.33543940960888663,0.6108518317222454,0.8403513725029946,0.5770533718895822,
 0.3772522455502376,0.33192378664218714,0.9924972118291032,0.4649180699282789,
 0.369857327277786,0.12042596460023092,0.35786843204062757,
 0.06665570186504866,0.2970539850689248,0.08536489303757289,
 0.7471733370535434,0.17225548598359175,0.5840007911970816,0.9560049063382202,
 0.3822655012344759,0.1304264242587777,0.06437819215502205,0.6714824293608651,
 0.7109815273530947,0.8767153817669191,0.4690346726897684,0.28465432977420435,
 0.4904846534184434,0.7345683021706173,0.0328632345943074,0.27338071294263533,
 0.43900332006721776,0.2801233188206712,0.12534013057363613,
 0.36004223434993954,0.8750633748769103,0.27470815408396765,
 0.08606648202275702,0.29933857438131706,0.9287248177571107,
 0.8643777206781925,0.23874305791979666,0.3067546912678806,0.7170377430508206,
 0.43226810159693085,0.18935677118292182,0.09217783306482896,
 0.7058467940735874,0.019070978380092973,0.16303851229732869,
 0.4981636565709311,0.5567236400631909,0.7659966760029866,0.03358413937086702,
 0.7724581207469078,0.30578000856518006,0.9571670184897861,0.6449936681383005,
 0.41054995878454437,0.1223962365024932,0.9805802614567575,
 0.15392512453711604,0.1692717228863193,0.5260640309473044,0.7104774972061974,
 0.7163606944323149,0.8677527162711598,0.41575830389654744,0.9996455887760565,
 0.7420445931609682,0.34945285311207364,0.8792346347120709,
 0.008410312811202969,0.030921623012522992,0.5145951220455253,
 0.789883692127899,0.05777452684885265,0.7404167996576972,0.2976123636514143,
 0.8001572897950606,0.4071154739672245,0.36055114301604585,0.8337254817165048,
 0.6352040865173888,0.19027192117088698,0.7841229231281632,0.4799387569989684,
 0.28970280347141775,0.48420533357341344,0.7299157603925969,
 0.9615376405119229,0.15999764609611733,0.12486649927047277,
 0.19050683804471635,0.9230234796581217,0.3838643031599148,0.9574973438952114,
 0.2919957187696436,0.5131291665685664,0.312619186346607,0.5330343959174919,
 0.10554460796791887,0.3670526381688678,0.2877526454731014,
 0.16292946775967143,0.6579392218160306,0.9954967393124077,0.4084792705054002,
 0.05789113097862675,0.057975052807877736,0.26912330046799193,
 0.5679741272821938,0.3641155462337672,0.9983113313424259,0.41807720586088226,
 0.8290100813550189,0.2912647033889869,0.585542806117486,0.9678595574118876,
 0.43984539730880257,0.9629113540131394,0.14745430444387464,
 0.22707953879727572,0.20767161937205358,0.8574276467705265,
 0.10619530551737488,0.9229042264411043,0.517854854835243,0.9737129675395422,
 0.24426921964199266,0.9992240116956727,0.17947371682436608,0.893667680225512,
 0.024266956999783318,0.6436706312447519,0.02560075117644911,
 0.6329665410935041,0.9386144788327127,0.3977053242817291,0.42377338488869165,
 0.930479096255253,0.7606029113675215,0.3573158447201814,0.3463690505576352,
 0.5575115126704353,0.21406630639041235,0.3481565373644929,0.5504089401300735,
 0.08950389153289506,0.9246975219868016,0.6498758911783903,
 0.40247763444063556,0.9204311105684551,0.015217398549588923,
 0.6457119497251982,0.22620942237662733,0.7159983024596204,
 0.14545077344285373,0.04402283313970734,0.5538873948309193,
 0.8003606710532574,0.7088465928378191,0.4931295288264459,0.8971580355233709,
 0.7813983234962129,0.38552760161101074,0.5331128463304508,0.9325647131404303,
 0.24029305993554395,0.6506361728179043,0.13658988602731437,
 0.013910018308171734,0.7644995662142983,0.10249754485809626,
 0.9000597629116233,0.5135335895382933,0.26292128163062833,0.6441030311959757,
 0.12833985582529195,0.053337186686881144,0.20077972828527413,
 0.28463974850547125,0.6052145747558354,0.831387396556468,0.6900707319257143,
 0.9030392271940495,0.549651916864566,0.9371857883221516,0.5686377586590416,
 0.709790083876842,0.5986313141663235,0.7944217296172587,0.24013480303447554,
 0.3508329660807106,0.2820536402632302,0.5802258999647866,0.4092320604612979,
 0.24734080108797418,0.7261211198487756,0.6186274745979531,0.6692059492991854,
 0.44163911397155975,0.9813434219376381,0.8213311184117629,
 0.37883398829296666,0.6572208923398636,0.9702002389511936,0.8925861086617152,
 0.30765018689505985,0.3167558757686203,0.993196176926396,0.45846772745456277,
 0.4508778994457583,0.7080951933527877,0.049339657796432346,
 0.4283776909242697,0.5996482901026539,0.8888068379948859,0.8126734039129473,
 0.1495587057847093,0.35719925019626686,0.9607775565400526,0.1701932040290156,
 0.04612320484545884,0.5438019794517645,0.6895825450822006,0.9218071200698881,
 0.4537725020718084,0.1779674285233479,0.5039153191534023,0.263153278050008,
 0.6750371765237794,0.9954357120603098,0.7040147662972441,0.2703103869876893,
 0.24149328687466964,0.9122841171272714,0.3556805806423031,
 0.07447763061340629,0.7658593658881369,0.9387959705642682,
 0.14296975464385386,0.7613948833654341,0.8662872005130445,0.626512855457209,
 0.7852875140684643,0.8310342958266623,0.12811046885642707,0.4068746736369725,
 0.5332785749368685,0.3706969828511918,0.45889201309932504,
 0.09294846410661861,0.5801155407291378,0.6131142799921775,0.4292199154547838,
 0.9729917483139526,0.2196733256382012,0.5114563552317961,0.3273493453013353,
 0.9645125834421147,0.9301874583883076,0.13279231410102632,
 0.03546745732363521,0.7156108561759361,0.43356367414606933,
 0.26015747647960885,0.8368868848567541,0.1487416276380461,0.7393455745912842,
 0.991578606755992,0.9063020269825459,0.9084802704959534,0.2895777764329106,
 0.287735919650818,0.6776248542262874,0.19248957337056805,0.7243477267630369,
 0.28577475927450813,0.20510064625392443,0.43103661648440417,
 0.9038530615021299,0.8997243261521689,0.5352851034543319,0.7432369154194383,
 0.24664146358558003,0.14332618284731502,0.041183662607115035,
 0.6877742902485348,0.5685798135345942,0.74613718761251,0.6784480038718373,
 0.2929379496354674,0.6326622555076294,0.935585175594704,0.6975551626430316,
 0.6225777645434065,0.6910670979808285,0.7449457573234303,0.7550380102972165,
 0.40163004040516936,0.22906689446973194,0.6505591889831682,
 0.2206609458274198,0.9959620283457149,0.2699673914744377,0.35550815297165106,
 0.701928769276154,0.7657267572128821,0.8973905356446048,0.6947047559646715,
 0.6026779684724602,0.28125770777877745,0.14084805189547567,
 0.29559512685309586,0.13542501408531638,0.2612190139828968,
 0.5779599147880753,0.2538725232089194,0.21327453862524948,
 0.09130394639479478,0.5761327696158886,0.021170161708758073,
 0.5415705745823468,0.6667495623934658,0.3736020730171016,0.3267550683854321,
 0.6801895204703842,0.8390902971750052,0.3582716222699135,0.9677128853925284,
 0.3991941637158326,0.3447585017132475,0.998513956282505,0.3454888317636665,
 0.364635419506827,0.5635694184878461,0.14600370273317975,0.44038549918365466,
 0.75523566935548,0.4990448980882284,0.822561558578502,0.22212167387993131,
 0.14524906024669348,0.3653841976394345,0.22094714457135023,
 0.9035965203609297,0.2435995202162411,0.3839834005022591,0.3735750520459393,
 0.7603762318498003,0.05266989197914851,0.7051684084648406,0.7625851136807631,
 0.13879134001250615,0.31181842870324916,0.3252503426803488,
 0.2857298606689336,0.470568789232488,0.30078794904365314,0.3237776925299053,
 0.24227177453442006,0.6030768539028052,0.042042930009127844,
 0.8641772572151262,0.002203272778135479,0.6066965940254919,
 0.3023464366283406,0.7756814082380201,0.47991685397967976,0.7320452722077998,
 0.03827447620405966,0.23116206569940956,0.638547148210317,
 0.18107436775472785,0.12915651504691594,0.3369131527109366,
 0.5310099777486615,0.5196203915400814,0.8060691809626314,0.02800736966586026,
 0.9438664882083588,0.5440729849211452,0.7697147025923363,0.23595617395713753,
 0.4885258528356343,0.42655573200464525,0.675383969335239,0.35763143360652583,
 0.30420503200480387,0.9179289299810984,0.2635042262879854,0.7192196691404735,
 0.14536794792735652,0.3411801666831171,0.8112292884810147,0.1883133709171534,
 0.5978868952159607,0.603017229725031,0.7168488475127857,0.3645923565208824,
 0.268715862057578,0.8311299896832275,0.6266823038586629,0.24924184388814197,
 0.4110258572782043,0.24244651837561682,0.16034225336975094,
 0.2680017259385612,0.1492534134912592,0.9492883003333973,0.8091970569750361,
 0.6218650140062345,0.8935370624080351,0.12510888947628307,0.8101472723130451,
 0.7290021402626364,0.47670616833765445,0.14996863558494955,
 0.17032780098116762,0.9934854558419124,0.33625666091625694,
 0.7534542754397969,0.7816662078210892,0.9866603036523276,0.7578155042822435,
 0.6144673021559653,0.07751716287472787,0.7521241695525179,0.7514428123028589,
 0.49462029575532906,0.2616847432100742,0.8161308535206273,0.1587708518058244,
 0.9775038710623313,0.024193537670952225,0.9205606305311186,
 0.7428177469220986,0.8612593917941888,0.8161254382186015,0.9862702589013135,
 0.6228998856701632,0.0897049034156352,0.10181445390487609,
 0.10539670162145842,0.3297742945920684,0.2749444173772093,0.6489104273622339,
 0.5104933676039394,0.9823374779662971,0.7032902158239994,0.4322626428641019,
 0.47898738146220354,0.5266036680191689,0.5058325010453276,
 0.20484887166615384,0.5069544707708347,0.1188813833379272,0.6814562772129599,
 0.08377603599139638,0.08546922096287579,0.5798023701788156,
 0.1290093333921476,0.5481750633951156,0.6125122729242418,0.9966058435023373,
 0.5270031128449973,0.24023206211328418,0.187827214536176,0.6552307501966932,
 0.09219563887892757,0.5917582860118198,0.14613522822838654,
 0.2593891265840764,0.574476397872104,0.27324479305766003,0.8813489862125464,
 0.9746342585212996,0.9325273950936868,0.28457582726477104,0.1286086803286487,
 0.34106421231924944,0.5237061021825192,0.5181083272059239,0.5985106721710062,
 0.7140676300653133,0.28653549499080166,0.6585076217885212,0.8663386919383707,
 0.5909705122013639,0.4363239575780943,0.39306246917982124,0.9974913493326387,
 0.2795901473074872,0.17960880344282792,0.8425185639635082,0.4244174462189887,
 0.5404713869818534,0.9058429592128442,0.15352625228926553,0.3633470872866763,
 0.08035740310512507,0.4757724647659419,0.5252735200862196,0.4480906710885355,
 0.16573021963709023,0.11609957171921781,0.09772853219136879,
 0.9617294087922988,0.9031918347039151,0.6949098619572096,0.4604872885902167,
 0.3451844728341752,0.10438389345211274,0.09955203522816736,
 0.2861342221960408,0.6205061957588922,0.3320402051901963,0.9108955924467015,
 0.0890274424841454,0.040657630924766375,0.7436164089953576,
 0.5329602722635218,0.14757211914284185,0.7633121480314985,0.4701257735820433,
 0.8626675356913089,0.7168616180050886,0.9674847819613168,0.7933817849086484,
 0.8496012704957303,0.26161079821295385,0.9764643759149018,
 0.23829777115730233,0.9778560416867694,0.7083334691226101,0.9441498323675066,
 0.34487716454494277,0.4666600409031706,0.749543398804419,0.23189784579054373,
 0.287510672406187,0.4671483898734232,0.962821094273125,0.0315126632964432,
 0.2767757772224563,0.8656122139527206,0.23783492771706294,0.9969878542154854,
 0.5144255838479432,0.37104839696212566,0.956353297306446,0.8741228226583501,
 0.13926139575119145,0.03274687501560081,0.5087012662039863,0.931179724863769,
 0.829561659417523,0.727760093165576,0.2670438100341117,0.9982074526527945,
 0.47803616911883506,0.6853034091612282,0.24218938031286685,
 0.5701584583282601,0.5633115330486025,0.8816483570962437,0.3930380334205319,
 0.498095372298764,0.31306303883994036,0.5673578581733966,0.3889454022992047,
 0.9539916120922928,0.33584098124165407,0.8230796473517633,0.9594681418791948,
 0.7090944664550963,0.7886428031313792,0.9876396629197634,0.6315293998974921,
 0.685986875201769,0.46044747523583884,0.5061455707379556,0.5218916330369625,
 0.5187331520034468,0.4698796510048593,0.763446819350323,0.11164550945432772,
 0.5683948871655354,0.25234859613464067,0.6135717436558829,
 0.06981104247732839,0.9627732548614415,0.7687801748941289,
 0.40847023557668105,0.289205475575922,0.08896553226303316,
 0.30244628966276166,0.01830926699864832,0.13851296898107013,
 0.7631645632304663,0.8762724436685323,0.8640631789434465,0.9009946388590359,
 0.4264178718513628,0.26829802694686344,0.3414762148834082,0.9359736929688964,
 0.49984893967063626,0.2550185525603078,0.43570221502678774,
 0.2984217895738114,0.5886206600766436,0.5531409663235457,0.78227964974122,
 0.08736594810210241,0.27458459373869426,0.901810627037036,0.7323746731154219,
 0.8952960872303894,0.6841411329904726,0.6096986476878063,
 0.027854863550969777,0.32889522448631126,0.6967925489513045,
 0.1720442900494129,0.050368220102230854,0.316590153656092,0.6152451705241894,
 0.4403609969499118,0.33168526836947354,0.989739411764738,0.5170933036242391,
 0.1668876472735319,0.5821367159212992,0.8399583033672536,0.977970002152252,
 0.4398615332204151,0.7279839974110822,0.9129938189517314,0.552823470773903,
 0.421963061650127,0.253412039835631,0.9196356530038744,0.34822933337697604,
 0.18893940680246268,0.148215615840677,0.8425876023289998,0.39213174238152726,
 0.1766597193210665,0.20518396205125122,0.72234591734031,0.9849028983542917,
 0.9834916510940575,0.9422021456929803,0.683954791601858,0.4814491777038131,
 0.32220107129562403,0.9126666432198418,0.20071029218243108,
 0.7023219089442013,0.11700692596897844,0.9307944231980136,0.5487845351400141,
 0.32320425732998836,0.8288356378455354,0.13534874294799826,
 0.7056922236574614,0.5185883080694844,0.6266151268601816,0.7097392412018572,
 0.2174130862155319,0.12639686840999897,0.11334956607277713,
 0.17592963255864946,0.7355473604531925,0.6058721132124092,
 0.08594442619421316,0.22565485741902214,0.20801945890840756,
 0.34828656223409815,0.5194339027005836,0.29824875721633826,
 0.8479408956736878,0.5370701224620547,0.5979639029575113,0.6810443085409714,
 0.7777165994678688,0.9297374829424588,0.8592783839215771,0.20602799317773846,
 0.9616839249781661,0.7087129081425417,0.2621559098451879,
 0.008460988822687754,0.8191918975262114,0.6929786757787407,0.782315110895839,
 0.7130807785105246,0.3764477965452052,0.21518197492771962,0.808507332026634,
 0.8975869874285071,0.5456602101897182,0.5899449174139828,0.19756015308555752,
 0.1946152968044279,0.5569877842812293,0.5732306655886354,0.04226602395699164,
 0.5816982790948286,0.2855719462728654,0.18770718425117927,
 0.14051908772517419,0.9169243407385881,0.29246388537113244,
 0.25389678011261263,0.07745210200367392,0.015894073980266765,
 0.14222137860383965,0.9482583486537319,0.6296613601753187,0.0700252561258159,
 0.31725205998654116,0.013363962448312172,0.4443267921431071,
 0.44781699032726563,0.20418678464665996,0.1279240867522341,
 0.15806700897155068,0.06049961548270755,0.6878726685923378,
 0.5718642108180637,0.26920334391535294,0.8409724082089104,0.440685879606292,
 0.7593253503709787,0.8155557188898084,0.8730913211084732,0.48540526449185617,
 0.16508325580856154,0.6990640145876407,0.16828422446070113,
 0.30978475792583815,0.006424526521068619,0.8737010079436379,
 0.5252941498146438,0.7058305520365649,0.3042038362720859,0.24055431095296864,
 0.017727594325173568,0.9121341841488231,0.516956419032657,0.9805854999287148,
 0.5839215424880944,0.10019304609349233,0.13717448060690185,
 0.1881961779203326,0.909487461554452,0.42482270891810736,0.03336869287582489,
 0.7822339329375744,0.12314882378160341,0.942007737615973,0.21322047695607593,
 0.17618252309820148,0.45275787682150104,0.1303569449141141,
 0.1898707605181259,0.7013627006109079,0.29565211688403725,0.7776972774782765,
 0.20968188030735502,0.1115068925862559,0.8170001493840015,0.4480071061035742,
 0.030681476797325913,0.9695395299742359,0.12156800446514238,
 0.975029352195069,0.6365486599013135,0.9949219801948062,0.1392268846035748,
 0.14543749404804018,0.22473523343576307,0.6335340386871801,
 0.3768991169088508,0.24962639940240805,0.6019830984293782,
 0.20178962610327633,0.17253835133336948,0.6699075592204219,
 0.5641772584752452,0.5314109534043695,0.6567636778969667,0.4673396864597208,
 0.31325224564987475,0.7560760615399422,0.5197765278445139,0.5357276786156117,
 0.04199935727066595,0.30261446803387504,0.09603427190088676,
 0.8415874672025007,0.5489076063247937,0.5361214048450443,0.7970641508533218,
 0.9495789961666901,0.7648191294475147,0.787925409294615,0.38036330234254123,
 0.9924758250249135,0.8795146285708297,0.32737156624335173,0.9196338669339107,
 0.7188227188254226,0.08330014155845655,0.25285691527301,0.34488191558530756,
 0.8441967749540968,0.789772037166515,0.28419942420795685,0.20618485339963388,
 0.872548956829488,0.8920714312036779,0.35392310088859746,0.7606419297471194,
 0.769445152925372,0.18654180695484923,0.8910073262868297,0.9264686911526743,
 0.6448626717559419,0.2759525339731369,0.467346071570101,0.3328476796296469,
 0.7614888165944933,0.02904219089172544,0.8350291844958136,0.4140491629945957,
 0.48215677869339146,0.7634874248787127,0.3817053138185492,0.5132862118471926,
 0.8982050689897132,0.44182577706676807,0.7357109585110325,
 0.24483900119837354,0.9551577210229445,0.28851546983579934,
 0.5053109908633637,0.8870558368134914,0.16883853992000397,0.4225281755971816,
 0.03287647274404493,0.1831496521837298,0.4212076506625768,0.8115335734797028,
 0.814809839996074,0.8048339927983474,0.7317599256482911,0.084248645320526,
 0.19097762909676674,0.49406580826601,0.9174153150760808,0.7257615190912188,
 0.8271349949442227,0.7458266097699519,0.1096239336809689,0.21343917972487925,
 0.029900659171174238,0.5220373744973323,0.6777419716745726,
 0.09626377170698652,0.5643428089247062,0.21930743598045244,
 0.18531026093460723,0.9909659332287082,0.9205293228703065,
 0.34485956893921643,0.3727719532518349,0.11897451241142709,
 0.9155887481300298,0.1723984021095546,0.2652188461259837,0.8594962440914896,
 0.11220693723263442,0.054650073340782535,0.45883870085812883,
 0.564446720479287,0.9571356152352097,0.06165664509812552,0.2847531589047284,
 0.861702763568512,0.09399081900463557,0.08340074737900272,0.3394193710194905,
 0.4849397355569597,0.07750095905152177,0.5543356468287816,0.3745452680901905,
 0.7474094832062188,0.3845048609511371,0.45829314356859663,
 0.22193377348970966,0.6864267381992853,0.9321886773163481,0.7268575703086054,
 0.3877810739535288,0.8638204464296915,0.9022060703147295,0.8500595683811855,
 0.2534990940622155,0.7556494844014058,0.7741538366463346,0.9465077569038787,
 0.8324321872544023,0.22178022545716303,0.9243304179419868,0.4921111647076337,
 0.01090773826139202,0.8707672156501269,0.08872838766520097,
 0.48393838821796553,0.6896287158763706,0.18590198409354342,
 0.33961591807698355,0.695572985427879,0.3099233628885958,0.33725504048843524,
 0.409221225746884,0.9178839859907736,0.4857492225562642,0.2551591994949314,
 0.4936548237576941,0.12277585501082078,0.8270605641075739,
 0.06866281741988822,0.6041794143872968,0.01993417915040796,
 0.9998375706206866,0.7797840120522934,0.6731768271916451,0.6046142621151089,
 0.026109429120480088,0.07376943692500892,0.8182463025246132,
 0.695078416466906,0.2739597998804708,0.44927307677728134,0.42802048475574594,
 0.32610048860269203,0.024079415033865903,0.10639321476138641,
 0.620234548194167,0.8887299582481367,0.6563999944731931,0.6185224496641619,
 0.9776441812802865,0.7981663483486925,0.5065651819005014,0.6301696620337643,
 0.41022556330133,0.3606963562418197,0.15252465011557315,0.5533924033342794,
 0.4660649151528765,0.09703491645223195,0.12278733110318374,
 0.13884415806997863,0.4225473563377754,0.7246779141729291,0.6567819693031158,
 0.6458727642304637,0.6797083978797062,0.8160188838755241,0.04929097111187719,
 0.13720060220756225,0.8260990341889323,0.20067526842150307,
 0.5970300102204322,0.11402745349187282,0.06680466160245324,
 0.7969375391725473,0.5563921710670483,0.48687457554905755,0.7121076280041146,
 0.40486887605512156,0.9168704699254991,0.8384328075019755,0.6514889991226192,
 0.715532916057286,0.8298832412126648,0.9266444761485003,0.8493799984597027,
 0.48417721551356785,0.7360490501865533,0.5815546901230213,
 0.04250360791192209,0.7171058051206619,0.43058899839535936,
 0.03926280921768455,0.2075687773121283,0.8601682299042916,0.6693058669561398,
 0.18712059346379817,0.34856297193750896,0.2645347270558732,0.705330319118025,
 0.639383829423496,0.6390200242108566,0.29354975836638075,0.9798271833649858,
 0.7160060771714907,0.8177776790780834,0.1668427622804456,0.12409260829384028,
 0.12741680611566164,0.22059321350146932,0.5424596028098918,
 0.9943015301098975,0.2680998483772066,0.35872056482310144,
 0.09866338227549143,0.5764754489769308,0.7434829426465531,
 0.47591218713101047,0.39050918318684413,0.05074871823163751,
 0.4421447397096585,0.6898424831156225,0.424380547409995,0.6519296607254412,
 0.3075340671614306,0.11346109828024065,0.27296577128460153,
 0.8897467713815144,0.4681172259431192,0.030769800316559404,
 0.2739714241954395,0.8738149710609833,0.19127186229476323,0.9901666432440359,
 0.9970545467107732,0.8148349742324388,0.9416132794848694,0.658129930565556,
 0.49883822596705485,0.9002940008085244,0.8530577766493025,0.9377396127925515,
 0.21497043610117406,0.16489227607360157,0.8481651877428494,
 0.28385682285190283,0.7265364759627068,0.01763941885662068,
 0.07034530333364697,0.33496261056211885,0.9158328914114282,
 0.14060004288807848,0.10115056944721101,0.7425910795490053,
 0.07203458654505057,0.610934913814184,0.9664285332562956,0.7613755115513964,
 0.761886818761618,0.38683275900764846,0.3178020437364809,0.6291818699348246,
 0.7587564606676538,0.2512285879771925,0.8566639351230553,0.3969518924957859,
 0.5598320234757602,0.6223596929183125,0.9830143929463854,0.884763359337854,
 0.4571905355024726,0.9388421220886353,0.6976025928538709,0.8890625055409311,
 0.6487720401018722,0.20565705562778558,0.08528751237449272,
 0.35082643417373793,0.3107223450014891,0.9042214923678444,0.1462193895256223,
 0.05527142151905884,0.44690968128667563,0.004051803680265698,
 0.9626300353213323,0.28319966363306626,0.6407173475046145,
 0.41918621824871805,0.1405783449088428,0.30351337463939565,
 0.6687832454535472,0.13735049549216116,0.26344837373637586,
 0.9311347326542306,0.2089988563392241,0.4768498527613695,0.14106664268106195,
 0.43696715506149963,0.10560460976656583,0.4227918908878656,
 0.24280288648041448,0.24207827020256323,0.5566083207180265,
 0.3771513503999748,0.7984246010395214,0.4526422118686858,0.7352120767386252,
 0.8569873244697362,0.7193009975151536,0.32375570415865296,0.8477178359718878,
 0.45038370538712624,0.26025134515936466,0.9151625675490239,
 0.6409775758954008,0.5083922971391471,0.5204661851333409,0.4803228405961316,
 0.20040740371787547,0.8472568533653323,0.14982958716382488,0.48722642438871,
 0.5444355867488125,0.5995935572624081,0.8211568026652052,0.1887205702400433,
 0.7213836940536514,0.43981889365824534,0.5065376551994173,
 0.12383771009576239,0.5091828552812616,0.7961098459486172,0.7764948003450959,
 0.878220049143992,0.5315854897561808,0.9263151507701514,0.27543427712620616,
 0.3390807182528911,0.30691581657073697,0.519106668090868,0.11916807322798262,
 0.8762741315394093,0.8995464110742949,0.805956910226793,0.5653835479678937,
 0.2021024000801338,0.5363722473206471,0.1588518934803681,0.849821003468082,
 0.7903495393378613,0.692653719816998,0.43836901351636415,0.157711013822753,
 0.7879526405439445,0.48084777987084903,0.6965651625094842,0.9517530040587427,
 0.7708698676466075,0.7696394533293015,0.3578357029935103,0.5580289604259706,
 0.7669331944789253,0.8295285806618735,0.46954201987263244,0.9793811167638307,
 0.27252713497465153,0.7638603586176611,0.26208426477954716,
 0.0958508527467371,0.3766340939395231,0.483008957047795,0.29081668447335596,
 0.4296425751481182,0.8888558074554643,0.4482034577277608,0.28687131259536636,
 0.469499914111954,0.7061751803429521,0.10813391842702635,0.7495475690878532,
 0.5180684264533915,0.510581132396944,0.9834541817839829,0.036943820072706024,
 0.5540192567203245,0.6344714812784709,0.8915606139060577,0.9215078918102794,
 0.5231086464134422,0.3874975394670608,0.9556194817157132,0.0552616546069058,
 0.13902202614274173,0.38876349480393024,0.7234032831935657,
 0.5336500982294533,0.30730468221168494,0.3976658896636819,0.711962895047166,
 0.09506431120175052,0.3869970891114932,0.16013177769662257,
 0.9954142310150028,0.9977802850800837,0.004773949840367164,
 0.21284230540117488,0.3815796825637183,0.7777016179415819,0.6602387296392002,
 0.920167719248649,0.6753672505288084,0.30987769745136184,0.4622573895718719,
 0.14125021393515835,0.8944981260495416,0.23213961838644137,
 0.7467315842521278,0.25377478754858984,0.42393541547664415,
 0.9968229225269865,0.5201562220576702,0.5430118508318434,0.7194952380514944,
 0.5139995478832468,0.2830310938235876,0.31289535406084756,0.8090062477410253,
 0.6715912325426373,0.07535520277867724,0.9099164175040966,0.8425736653115903,
 0.9550206027501008,0.5364866354140988,0.32915748587993443,0.3658019748167919,
 0.6392791327382242,0.11598380672044883,0.1978915608654993,0.5584509592695405,
 0.7398643077177126,0.24179101067684672,0.7012474573795968,
 0.047582021740160796,0.2808998531269198,0.7781862798279828,
 0.9903149654758954,0.35758204995069975,0.06796568516672008,0.643673127915382,
 0.33923842511223856,0.5272680966329133,0.8861135395749654,0.8708267293782724,
 0.8713505219228185,0.724048653420973,0.2647953906224729,0.5927825221577686,
 0.829745220104695,0.0673192831352536,0.48762497130225835,0.2625743656046905,
 0.5246346699931506,0.18796784677233802,0.45127128885549384,
 0.25330986931767474,0.9515381524458777,0.7326473159581258,0.7157030503487694,
 0.9390700242989026,0.26239452384353057,0.7920205318216859,0.610114252429634,
 0.5480858046492448,0.6074493250301106,0.8808090011366391,0.5127246313383492,
 0.19400101261646707,0.2567516558761732,0.6640389217512782,0.5158711314137719,
 0.8380521008660609,0.773355584600285,0.0069511722637685125,
 0.7318146560001068,0.5645117874085192,0.5332244432883892,0.4320679490469197,
 0.9954923650649411,0.542998032165385,0.291858762507319,0.43112486137287587,
 0.20504159161657465,0.23624040425723491,0.0920368174441889,
 0.11585697678504137,0.5762011332028257,0.9688245995078978,0.8239498062514012,
 0.6640560992003259,0.6970986200232723,0.37446919203916695,0.9102211641321607,
 0.17086668710990338,0.6275370451081279,0.6093259481807962,0.3820755217770865,
 0.45291428113421894,0.4745071570242029,0.009651522621814568,
 0.9344202445205595,0.65404156613029,0.46124750308675166,0.6178472310178298,
 0.9607689594100615,0.8826236160442719,0.502810122041344,0.08033938257315532,
 0.5229500476536492,0.6871334177674906,0.9008058857690885,0.8715741136464271,
 0.9078137992061615,0.7192281581967229,0.5962582627977717,0.43035434556536845,
 0.9808501503113999,0.06750522810795367,0.38980041068343696,
 0.8409485885599255,0.5174515802050691,0.48909835787282285,
 0.08302478437161354,0.14504865963602798,0.40215035759783035,
 0.8422073151902219,0.8372797754666227,0.9301465432970388,0.9769626339737897,
 0.347067778699817,0.14689800850635515,0.06970884217750362,0.321612110438108,
 0.5394960856272585,0.7749856752201033,0.5536926898275553,0.40768716414862616,
 0.5545008455555839,0.8394981809431377,0.7358066065774849,0.9163240959623424,
 0.9092458483097092,0.7016312947936879,0.2626958457560091,0.02494882660618214,
 0.1463840177118474,0.3031989730218584,0.7258824995978193,0.5251512960359018,
 0.7365060514778229,0.6704625994640351,0.7913230637013613,0.7409986741700756,
 0.6459754088556735,0.483488890226902,0.5292391434780587,0.6101149132470755,
 0.7635581414891136,0.4076107328046532,0.6609355058947002,0.6510806589871839,
 0.44435173346210843,0.7376965453357615,0.8385443406791978,
 0.33863797415955244,0.649850560402757,0.4484329705064315,0.671361046890743,
 0.038277430445183835,0.7653183446412003,0.8328197371755808,
 0.3238695731551188,0.48272292060515676,0.9706512957034812,0.944224050049943,
 0.8328679369851617,0.23097950131471912,0.4463089951135034,0.4739900811409873,
 0.10193339256573586,0.9962548499709385,0.2822953998252553,0.6635107209710657,
 0.2506798425520125,0.29642192458062977,0.024860505083691864,
 0.8597159087001496,0.5281709059552258,0.10737268410647116,
 0.40322516421563404,0.8517351413354003,0.31954541259766256,
 0.6230209128403368,0.24602222841414767,0.7773641085478854,0.4736359813855703,
 0.16003075030961922,0.5869740788405396,0.9866803194424556,0.6467749324628672,
 0.8858819288762334,0.6257252979709755,0.41858250129451324,0.5930441356641183,
 0.42485706147487756,0.8199570881202259,0.006667807949418125,
 0.9011739330712276,0.09211773690845249,0.4953818703857198,
 0.44682710894492006,0.10893762732959811,0.23458547033511046,
 0.690538513719577,0.8533671722245546,0.0015471412284503439,
 0.3387183480115643,0.1405758549120758,0.4947142613473572,0.1868174789158411,
 0.3342104096669135,0.6885415118415452,0.6721482272242963,0.7421302281219776,
 0.58118680501488,0.4040152014219053,0.525972382946452,0.5897403777652915,
 0.12301112121871782,0.3129350678499443,0.9354850644443018,0.9542227151358638,
 0.045099209965241194,0.15947031011625712,0.9054670179849515,
 0.3094198381095542,0.8341657176210874,0.7590070063805814,0.2886456967067892,
 0.9051570074467068,0.24431909308198552,0.9029753683244996,0.7915798264235668,
 0.46147682819979896,0.7896700953219667,0.7517221683821004,0.5161488337346748,
 0.3176476533628343,0.11813937646238049,0.17361928924477588,
 0.7907693571216942,0.6606566356399507,0.5685195484784282,0.21897186899263021,
 0.1654523548209904,0.6205254645631545,0.8887098145227138,
 0.047097829115599454,0.68235202504132,0.8555416551120038,0.36997976356354245,
 0.5818785977934491,0.3565514970224887,0.8845141732216119,0.20115054166055768,
 0.06588965913318323,0.482376790028166,0.02969352052963592,0.3103734170032615,
 0.9986478459955772,0.7162230619278543,0.5313241053720084,0.6372882865184153,
 0.6692212410959563,0.7678294077461307,0.11035465351697393,0.9516527888252906,
 0.7262042704267866,0.312402295336741,0.7043535788673086,0.3009753496496844,
 0.8407715250897347,0.6134938908383799,0.27843441238078614,0.8005523548375537,
 0.9088771581790859,0.8105803282057761,0.8616705501211978,0.9998419388844186,
 0.7796341645313265,0.17452038371653544,0.5151827811755643,
 0.20189140523585303,0.0204526779558879,0.14709684215074237,
 0.5456432358581803,0.36576679132331014,0.935328878089882,0.05924273795559598,
 0.066769622424393,0.6053486818727983,0.24505167587781296,0.25151300475709726,
 0.7222024516039834,0.11908551898237651,0.7103117451471569,0.5849491691740117,
 0.4692713762202403,0.5269301776588081,0.8114499348579999,0.40599659031144375,
 0.2929055242126808,0.3848731395255227,0.512729961040055,0.3628052564715598,
 0.9326616102451939,0.1081636308058107,0.789935040813736,0.32945614017227853,
 0.6180670077975862,0.33023168706766104,0.9623876390265769,0.6943739395918251,
 0.48931507973727173,0.4037782138626802,0.8972085930325933,0.9502435091055367,
 0.00327681182989481,0.09407019225456859,0.49474768896853494,
 0.25275089812627294,0.42477968307703595,0.7968807988881822,
 0.15746256271886538,0.19741291535248728,0.36430649340595467,
 0.47732100913732856,0.63597165496043,0.8768137588375123,0.030932613055545066,
 0.9228388233312581,0.9481277184071718,0.7282065325817412,0.11835088727882104,
 0.5361411141275767,0.6279616774541782,0.43891844658879786,
 0.32304016985020134,0.14089400172862554,0.8475629413406505,
 0.8809973895989286,0.9004845311856783,0.2318087188326401,0.3748950382797813,
 0.48000660949985674,0.9413858144113652,0.6556429609987493,0.6757015311835957,
 0.8467430208030304,0.7781324524331117,0.5262352631047119,0.2991987635440777,
 0.32281757478395723,0.4847052303373346,0.19412197195444691,
 0.6135687021381044,0.6166091792172191,0.3431862045346499,0.4710263797890232,
 0.41712325266026473,0.6041517213997586,0.30375190292218435,
 0.5971736454920137,0.019469902514497095,0.8428980320993728,0.9523676427702,
 0.7518926903172941,0.3110860685020379,0.41624913318848644,0.4829270159289728,
 0.26098464889710826,0.39995698873149566,0.6227220151478452,
 0.032463560042141126,0.8431950054781819,0.6518253357420338,
 0.5157123000823998,0.24959633726198538,0.8356333308993443,0.3808793337803811,
 0.8592758189943756,0.6901209767766789,0.9185461274475648,0.5603857072893024,
 0.13422051930690082,0.4644769818782497,0.8130664560957029,
 0.06231941375273453,0.18995760690564945,0.08854539142465823,
 0.02954133750641086,0.9763970710265186,0.1382382037458123,0.5611567167786362,
 0.06361465923847975,0.22653910254211385,0.44262092543110054,
 0.1602910123979312,0.7608765731638742,0.1082734149516067,0.19814618140016882,
 0.3651411216885989,0.5773781698453243,0.5125007697568223,0.02160811409191865,
 0.9837369414367599,0.29312443973268143,0.5513556101293482,0.2668949354652699,
 0.5211346094634938,0.26892081906247933,0.7477097698672991,
 0.30406136175583076,0.9047319333957682,0.3166696543898896,0.3793587097076492,
 0.002398249904400074,0.3629429286044541,0.875207138942523,0.3370308994918205,
 0.42531271440794494,0.9744749502327952,0.48014930520843235,
 0.7705100818466883,0.5186075461804376,0.8919156000281138,0.10244111013079904,
 0.45466660804577863,0.8258981312035442,0.7784557312781929,0.9996679745786463,
 9.468661877750639e-4,0.5107466639464842,0.31662505537588015,
 0.08655339270730045,0.9094563819545102,0.6582412928697998,0.6110138784763025,
 0.5300779040947631,0.46323332807178186,0.6828964587506424,
 0.41247096550531004,0.2676120531853804,0.7468965455296449,
 0.42620006129665655,0.3468690521890765,0.6207092055404476,
 0.10098686581520233,0.3022327129187521,0.41513595168210904,
 0.20846371019900845,0.9996860759366079,0.18467621121278655,
 0.05448229090447754,0.37978982810980266,0.19903361057269786,
 0.3040519125844985,0.02218694965651402,0.8081418457688446,
 0.09123459958461555,0.7090437904191661,0.0060358437849348245,
 0.5579051150304752,0.38951240329201287,0.36538871282549656,
 0.1252485769588002,0.4414737662418724,0.03510854190112833,0.396363547287941,
 0.9541382388504079,0.7877449403274985,0.48975171323449807,0.8167480800025682,
 0.8840726728228602,0.7661870780705932,0.8985020574367808,0.37150244800767795,
 0.42017894287875057,0.12249013581970492,0.865983040525653,
 0.23629284416657959,0.8927937997536113,0.14049439658658436,
 0.4259549457027906,0.881692058486523,0.5752442514422472,0.065766641525424,
 0.3286862022287389,0.0068667289649905205,0.41805330896543635,
 0.6525848477102298,0.2802166562362739,0.6343869461296088,0.5483968493718571,
 0.6225375115563961,0.5777329857397082,0.13503147391302084,0.6708173283240424,
 0.47711099875812146,0.9485874813855804,0.27432667373377884,
 0.4239978092823946,0.3982228445455449,0.4121923483224922,0.8929424284401883,
 0.7636090290561997,0.7291004350087723,0.10752455558046237,0.7603127172086614,
 0.05706496990875565,0.05121747398698007,0.333424991863872,
 0.37801083414653835,0.49465137592003117,0.6477264521953896,
 0.9368165300203376,0.21096055892338816,0.8706519161803489,0.6871803043054054,
 0.008067401600959645,0.5150962837628147,0.677793685085696,0.7077480974739182,
 0.7543860155626563,0.6878645211033032,0.03083494718737856,
 0.029577131354129316,0.20060852554082387,0.35665089951425355,
 0.5813116317318009,0.2892647159224161,0.3742372467527787,0.12820295626200573,
 0.9302936287867944,0.7098618397715362,0.11520923254174176,0.8847626703252419,
 0.03416322608862887,0.5883299744819305,0.6948322459494238,
 0.048988489514120936,0.6287545574322728,0.23059300421181717,
 0.8843948412713095,0.7896809237850666,0.7898179453039529,0.500460338843758,
 0.656069997852498,0.025201162813811684,0.10906267319081187,
 0.9875078344756668,0.02273534653658571,0.3227375257137455,0.6081656907714859,
 0.353875777194816,0.9037392215402416,0.05697450965156059,0.1670921739774731,
 0.03460393790205962,0.3365822344971464,0.9622672787831165,0.770674728772498,
 0.2710366881242712,0.1991395332114133,0.39866186535227877,0.8793156785716445,
 0.45293397388875334,0.6168578202449153,0.5245770208456535,0.7667064502821316,
 0.8187657793053407,0.965097961015501,0.20389135117569313,
 0.053471581414751834,0.1152455872132575,0.23578659266731739,
 0.04321160857321438,0.3228912993680899,0.8412120610820447,0.6477700448233817,
 0.2510200823083636,0.4377189907977561,0.3347180154427911,0.2216692893613157,
 0.9583923394413635,0.023620589864770225,0.12979864209983605,
 0.5181152487578813,0.9976862405786844,0.5935337162051812,0.25928793347021684,
 0.8124013760183855,0.5331841011566381,0.3167676582002875,0.33769324691621994,
 0.38796822090652006,0.0671794861040802,0.6580162843703926,0.6739874212232443,
 0.8644924341885464,0.9091932065208373,0.14488367217158282,0.7587252788401284,
 0.13495401552562103,0.49437965426858677,0.3527946108768124,0.627084388012613,
 0.23540622081927365,0.07628826922508436,0.20341413222927518,
 0.4907696806975902,0.12588149640568347,0.5218881102471504,
 0.40143260131260705,0.8875897798882841,0.7724647074772,0.8064099626304468,
 0.922122160894431,0.47072528403730785,0.7193393016138822,0.25145726564387805,
 0.6724088734667635,0.0063294303135909,0.90028080517706,0.575024014024806,
 0.5120591159926646,0.17211449363181397,0.8945588699971668,
 0.20567332761326207,0.10952093220135983,0.9382379582835585,
 0.6509027605353054,0.97558317404946,0.725983370149183,0.9405552972389035,
 0.7560915037199184,0.2760576956685894,0.7026492576528913,0.5860417877030096,
 0.3640075824172323,0.6012182073656254,0.27379336386991127,0.8713854085522262,
 0.026556415083565543,0.03347638946187226,0.4011103847777848,
 0.3556125006708759,0.11759411976108936,0.1422271710233015,
 0.0026502945664581556,0.12625944329150496,0.37972626356124684,
 0.7894924841274581,0.4308693724920576,0.2980017367983374,0.17060971781462797,
 0.6469017083797338,0.20678620490272925,0.20869888220197852,
 0.46853377854677003,0.012757002030034492,0.6423683244439087,
 0.6144269873516149,0.46237077066664833,0.8197400511405827,0.5280575605116444,
 0.5421336483728337,0.5372304094908141,0.28087102498404215,
 0.11530673234406152,0.931414047060691,0.6669253611142232,0.9746042961762014,
 0.37518829927502007,0.7097307785621932,0.6531706161006814,0.6905032281675042,
 0.7888945614611225,0.9111490415844787,0.10437767970350442,
 0.08039809350558169,0.1855646396795212,0.6808066153059442,0.845421946932135,
 0.028955276129655205,0.9072104813421605,0.06831911881132002,
 0.7090816818748542,0.20677925920680318,0.12975520162226406,
 0.9770089722329067,0.9412037385358618,3.719158976338566e-4,
 0.015161740461226425,0.5267216551469482,0.571862700364213,0.1954010707051398,
 0.7859529249186624,0.8494198538150942,0.8431744185056093,0.9890846994296,
 0.6883082798843474,0.3330076088310874,0.6201999332891549,0.3582357340322384,
 0.38174002662494944,0.1764947482540784,0.5187558564116613,0.6111913122509739,
 0.05474787312675855,0.13530982064251718,0.318532940870069,0.9843883419874397,
 0.6671347742602908,0.6304947561623506,0.7316850480509708,0.18451605941979055,
 0.5772262082484194,0.5966667286383387,0.3515093906510839,0.36513255707785464,
 0.5837017931531896,0.4358242551722228,0.4346275233651289,0.4004019416321132,
 0.9746669170327202,0.30321140807542646,0.34157494457981563,
 0.35240041114258325,0.009135428496812925,0.6910743207880787,
 0.5377939308387887,0.865535826074488,0.7572108138434606,0.9796251243987608,
 0.22970262508592032,0.5637028585681929,0.6321426470655052,
 0.07012228451043268,0.014286384180160261,0.8633692626640566,
 0.39905299702506014,0.53465458949797,0.3610679446866307,0.6643413097439589,
 0.2810599512954396,0.9886724487662906,0.7624364989010703,0.6223425041960556,
 0.2619344930098948,0.3495996706099007,0.9636866250702687,0.6268673341727409,
 0.30256693781595634,0.7400623170642849,0.3182229148936917,0.9099294063356221,
 0.823061116831035,0.6544464217733694,0.4710494953688138,0.9193462741468568,
 0.19571492681861458,0.04723447654690105,0.5735413582456369,
 0.4305215379647561,0.24965943785549383,0.3548160921799519,0.7991146222007562,
 0.9587230945837119,0.1077155328302064,0.10110644748381958,0.6455377791753439,
 0.4254995964742898,0.045391633505366236,0.9960375050425313,
 0.8539762609418818,0.37708926574707446,0.5690430426715287,
 0.26984522371642916,0.4700579938121495,0.7884872083113669,
 0.46066412616934227,0.44575529103188605,0.2815278077083758,
 0.6684531727360195,0.42767688448789776,0.32060353866884017,
 0.5217710249790891,0.6065884526496149,0.5336820763720053,0.9084113431708343,
 0.6120349243933034,0.1906541623381317,0.9948997033969257,0.8529354013976017,
 0.04089435652357165,0.2304350355635374,0.5433494681224474,
 0.48296871674752506,0.05254581634973465,0.12701321764720963,
 0.6157798100886016,0.12975110164671821,0.21104570891139118,
 0.4833974128382087,0.8746247323002811,0.6567102503882762,0.5175464982677042,
 0.5405594340120684,0.043320639146368634,0.09159252727096656,
 0.7178488586916896,0.06953741604584485,0.35539132649768035,
 0.5421615405712441,0.2873991153572588,0.29125366638281225,0.2986987413912988,
 0.6961864178352614,0.08265187330260082,0.7334734987549587,0.5830571405416018,
 0.7914454951900043,0.9360663450636847,0.037436152516630905,
 0.2887839917263515,0.04880953118405451,0.1650132552221537,
 0.018569622830182708,0.9419054795967277,0.7790673965275985,
 0.5420484842216629,0.02366589041160927,0.3004259507587137,
 0.27025937538745226,0.330922134062843,0.04623572490511729,0.2671237297213851,
 0.0020796718783326895,0.29581052380783857,0.7914632376172732,
 0.5272504111081812,0.8361371100252181,0.155141031694789,0.9621020683122404,
 0.9907424209715818,0.9746008033163645,0.4193073183344689,0.8075598923621262,
 0.9355930240696465,0.1143042174936949,0.7331921546826559,0.33451372332479434,
 0.9270770768172267,0.17478782409098892,0.12533216332605146,0.608294334711896,
 0.1915042483526752,0.6490352959444843,0.6059354293279581,0.04592035758142465,
 0.9134667873318867,0.4944130044228675,0.3865201132800554,0.9656715490761387,
 0.6843779518852408,0.28814925511609557,0.591455977152651,0.9386623710426332,
 0.7573669654391464,0.18006436075902676,0.03368471924074623,
 0.08529402172258604,0.6516650339555967,0.9435733110668277,0.8371332671589069,
 0.08693577719325618,0.07028255500081104,0.11181215148464674,
 0.3254911870993069,0.09104753327797,0.796017762272925,0.6969274883046094,
 0.7788405489011311,0.8422325823234841,0.7727349235613217,0.4759436987726471,
 0.3193902116663494,0.5284052225108073,0.6407909003238776,0.3049233169279021,
 0.2222476630221395,0.6317550076437248,0.766232403740037,0.3975182080922809,
 0.4303805615298433,0.9107291793170319,0.09303968461217693,0.596469535777344,
 0.2774265874798265,0.05388609241013098,0.35051693044077703,
 0.3958638063107859,0.4496856956052758,0.7947653530186668,0.6176401180089038,
 0.4443593662062829,0.8612741230891477,0.8498866920479373,0.4413297279785169,
 0.6763427211200702,0.02934291492089114,0.07632758031159192,
 0.7478823598312138,0.9431180264640125,0.29315337740741554,0.8963556084725929,
 0.7530298075755554,0.44009322650035043,0.6436770764037045,0.3286367588786,
 0.25998639053115036,0.3517909816181375,0.012962254152120267,
 0.9300252789491291,0.7343674003488433,0.6609303898914252,0.7935541222369631,
 0.15490573199897184,0.10991707734035361,0.6022198422423168,
 0.7354745769956459,0.03733989008091443,0.8859476665423096,0.6789966387548968,
 0.4536230466824498,0.15665296502286863,0.7286798465724309,0.4112296613285812,
 0.5731850975229899,0.2090100272503076,0.6924964083385955,0.5812288450829768,
 0.22016426942438894,0.5026907904420993,0.9541419217745911,
 0.43135311642924457,0.7168420639633812,0.9953846949088497,
 0.22854543161633745,0.9949400795780133,0.24521116201635362,
 0.47475974031966006,0.5726673515413516,0.8407646381986312,0.8618268879403698,
 0.49393954545667185,0.17091291302338096,0.2699136971266143,0.303495195167407,
 0.46446198084718215,0.4707306578832692,0.5476142339004842,
 0.03267444357296567,0.31302893500715256,0.6900030489598548,
 0.9477167191220954,0.3886208402219834,0.07213613433679089,0.5002168394251594,
 0.9626850882592503,0.513732123111144,0.7313424278779188,0.8912498574605744,
 0.9905771185996989,0.8723001724750252,0.9542926936779917,
 0.012526164910058646,0.7314816474229676,0.6813463592231267,
 0.2863079204386221,0.6994340164892424,0.8647734917401888,0.11140920716629499,
 0.12865505358132212,0.0851910804032393,0.7204281355650024,0.1573043283957014,
 0.7466552631000828,0.2896374138931528,0.5593090895028028,0.6048361123745472,
 0.19731104709561442,0.612068440382832,0.9380976222074006,0.0658110614508397,
 0.5992242223213687,0.0756667982783078,0.5886555462218417,0.30832478510436856,
 0.8483587690930201,0.5844975832421406,0.8753752544433396,0.5863422155627835,
 0.147475057247763,0.29455909176364947,0.4259020115572747,0.6334401107696799,
 0.2808329865472625,0.2037450421462903,0.9350747144805198,0.20740702937924904,
 0.18242935406863225,0.2371989518671418,0.6647104562040874,0.7233934076550427,
 0.5528428594624308,0.5010435945463781,0.6967148479034322,0.09560132824946521,
 0.30480267890725876,0.842626152080581,0.9321611358732393,0.6814865570923234,
 0.19113939747090591,0.5844191223273236,0.27943986181583325,
 0.4130491968385932,0.6515142141618591,0.8189197096652835,0.2294173241604529,
 0.3540671565036384,0.7198365035859196,0.9307033529696475,0.1589170327867626,
 0.5663896530245875,0.7004559538359838,0.011677094685529577,
 0.3706600690939895,0.6638073878490982,0.1930635872167663,0.6120015279379567,
 0.9897635766549913,0.4898817975041895,0.2014845655517452,0.6838808240871433,
 0.6549999862491862,0.6457279160157074,0.7848663135220988,0.9460882253067285,
 0.12938926224433978,0.8661101093002418,0.9799867278047254,
 0.24493388543295247,0.42760088541209507,0.08875367680204183,
 0.7872020003992057,0.08733551042103938,0.5960070827426057,0.7289733177428018,
 0.23604903558375756,0.9715133370785053,0.16591086093074203,
 0.4234133211129214,0.7302381002021361,0.2719430039330837,0.2627093155662411,
 0.04823595108501555,0.285117183975804,0.24278595036566475,0.6878746971551368,
 0.6629101221480578,0.2843527358346034,0.19238928286632462,
 0.17481005515771786,0.8967800880718544,0.08684712656568006,
 0.9566707700402712,0.2858270160911738,0.6374688605970439,0.02414590062808175,
 0.2101436571005485,0.4383914390084359,0.7892590268933876,0.7907651479022415,
 0.2360459141165726,0.11347393382729032,0.97838089181594,0.06304008743600897,
 0.6571483340392728,0.6601031140023529,0.6300050026052302,0.9453447020540694,
 0.461472835511752,0.5035489260409768,0.5016797888596011,0.7242407633770114,
 0.5019951760346026,0.15272666876119734,0.13845121360192936,
 0.8326748835239762,0.8787570077909754,0.062023656272794714,
 0.7943621987336721,0.6367485204947849,0.7661671954433145,0.6470010992611852,
 0.6347931143627835,0.5557187802019604,0.6684597821876482,0.3733628440463951,
 0.5385486567708034,0.5202180067422064,0.4053608004050515,0.2231255395947812,
 0.7315749933768115,0.11598809501522578,0.5871715344872331,
 0.034606614917801926,0.7243001737389643,0.18462948936033108,
 0.5564788371207832,0.0709163632414439,0.7536190756693872,0.26889181413126284,
 0.1525555489474122,0.7016745167542705,0.05934685520921312,0.3359409342053429,
 0.4512291749369316,0.36751096345364376,0.19166949390003984,
 0.49766661471808415,0.6819586184223474,0.979958706128643,0.594030895729555,
 0.8661347896807537,0.020522670691618505,0.20543938861566624,0.51974823619338,
 0.08321279873022336,0.79833490705626,0.31757934573522506,0.4616181631643539,
 0.0753514007358348,0.1023799195488363,0.18224914657239466,0.576269678076492,
 0.7278980786480787,0.5288784461161647,0.918262862017756,0.019043314492566576,
 0.021495979545719024,0.8608823567431175,0.6176086981834205,
 0.37706246213319594,0.06739098858093384,0.7436720371177641,
 0.21510690834629975,0.5188374915766623,0.2911388383499013,0.9294834874135366,
 0.23225697579060578,0.3799410838903853,0.3808009715238464,0.3156974440117679,
 0.2636386989748274,0.2136931076476184,0.7826672869005584,0.8811389691561009,
 0.4206927984098504,0.32490532174165043,0.47897765452328167,
 0.3060709571425617,0.2569366905014149,0.4808774963391652,0.5756771047150424,
 0.706575231019984,0.9293497659424483,0.7791181924265734,0.5041093839340498,
 0.41286966523127444,0.9677610013968827,0.22058724976799127,
 0.4014407225154274,0.20985454432890305,0.6655003963194877,
 0.44671003333006376,0.3752394323388858,0.15775247355305433,
 0.9881621483835646,0.24200444442925872,0.13349263166617076,0.682755449834979,
 0.21051438625397456,0.5440016250855115,0.11118970668519823,
 0.7027599580694746,0.8578847278822276,0.41477225338685475,
 0.009687578329522606,0.5415240005126432,0.10658422307696458,
 0.6000881033528783,0.08953204939236181,0.743571735902781,0.19918192540782287,
 0.9529448197692162,0.6009916749662004,0.9995916508017585,0.6119112126943985,
 0.18708830797878861,0.5544191686082801,0.215148468322226,0.6620072846604548,
 0.3003892331273974,0.00286658539632334,0.7676577913289498,0.8425945724332974,
 0.48595303547362545,0.206297338628335,0.9519542708935862,0.16755645392321816,
 0.5780654117291765,0.8250502030822757,0.16263458980445378,0.7107651896613765,
 0.37261741741983734,0.4607431053072254,0.23877109875414315,
 0.07007141964639119,0.6782190369348174,0.15549057034868818,
 0.27486455293792345,0.9252859012493638,0.6387739754085417,0.1492147522817039,
 0.8871214324099961,0.5876415319848121,0.17928021384179293,
 0.17303015711492087,0.026395376342225885,0.8070148908033525,
 0.30724696204191215,0.14733565802347326,0.0885104316930273,
 0.17182903414539175,0.7203001158904077,0.6358162004677017,0.5765588244688222,
 0.5500410026292333,0.5364046262345257,0.6960414641957866,0.9471345398349993,
 0.04649364570531689,0.5051896621223243,0.8261171642689457,
 0.41290420613179557,0.8640241126903009,0.47179582014847066,
 0.23654032188343144,0.31983067751552885,0.678311432955215,0.660520381362187,
 0.2478435576194964,0.9529615728706967,0.9214541669798103,0.1502083932557955,
 0.6110444066275337,0.3794552963062008,0.37751832053649437,0.5669714417464434,
 0.6232328025688206,0.3900206416814078,0.14333233072631468,0.8021198755789738,
 0.9157670098851313,0.677791509092944,0.36532489702216775,0.3935373245904549,
 0.20232869506395246,0.7675147421998072,0.05850520631449485,0.450302681436048,
 0.6141256231785681,0.8301140877244456,0.1442601995347289,0.9109563752942963,
 0.010947378852010803,0.10645203670136116,0.5808979897675888,
 0.5950311078480142,0.8484871523351228,0.37960544094682136,0.4404091147213698,
 0.7951003138871948,0.8254894644943864,0.8782578087399551,0.6348976150900493,
 0.5105726298971482,0.6709196514464135,0.8607223546521373,0.5445815879820888,
 0.09637160111511878,0.745664820444625,0.5944826153432559,0.7637377543523999,
 0.3775992176146672,0.925870654273524,0.34448900813286987,0.656722908390714,
 0.3199291529649837,0.5535916914120751,0.42717643006397565,0.2664348112914148,
 0.9175857266305043,0.5269483292261969,0.48045503213212304,
 0.07473123123426495,0.5820581328230124,0.5282383924896098,0.389351924928797,
 0.8698833714066975,0.9411529429691705,0.5991245927129758,0.21027250380284768,
 0.4658098838994702,0.961573282956657,0.5837600734363347,0.03995939482616773,
 0.1104110123842642,0.568989726599088,0.9309315389726885,0.5570656441505482,
 0.9492730654703014,0.7492031678488321,0.730564677863428,0.7180001936944606,
 0.8230025187371023,0.9400790749471879,0.3996839437173483,0.6990322774655059,
 0.8652046767059947,0.9136269606098584,0.04580916333747864,0.4356725367369516,
 0.7103583937272475,0.4851005296752966,0.07786279542036256,0.6415906638886513,
 0.6927453122118072,0.40541978424006286,0.7357932949660109,
 0.11410916429317985,0.7485038130942463,0.7570274126937209,0.3472575721463813,
 0.32895567538908366,0.8588879468016755,0.1432701645881297,
 0.31181126887265587,0.04653771468196821,0.17635903026249333,
 0.7656895531530195,0.7880239922394672,0.5401752176961958,0.7323712106131647,
 0.43820358102147616,0.46140516329838466,0.8769634278725875,
 0.37987334787014737,0.2292075588782343,0.9695959559365036,0.5112771045508087,
 0.9390012336441872,0.8749663659349485,0.6187004159471562,0.9959251966314304,
 0.9569985104034107,0.25831295822974976,0.11177907180763846,
 0.6465120843832708,0.27713899836826994,0.4211288426520926,0.1251679044323315,
 0.36437930120584594,0.17471307218581789,0.8045413776683736,
 0.9294914247996006,0.5884160425507696,0.13529888623630848,0.6049053979133695,
 0.7786644989819211,0.8730535334666142,0.5060723126939907,0.40152110643390715,
 0.15738005297066715,0.9435205467731834,0.254060322419724,0.3396846991422623,
 0.08315843931599287,0.0064378204719856,0.7695657839335329,0.9434376851065631,
 0.3349767419051115,0.310484963854041,0.8184146980162645,0.281994071667668,
 0.28060696871754576,0.822257963238207,0.32471522178149326,0.8366520574991331,
 0.4219486627504211,0.2105381177887593,0.5274666196450314,0.11672737339111827,
 0.9907186540421732,0.5634521888702101,0.1642344863175712,0.03999344188885978,
 0.6441453595993097,0.5793918196740938,0.33042462062060507,0.4837706654824293,
 0.5799385449505292,0.35274381954578393,0.9433613798999213,0.8520381201181341,
 0.3125519012232818,0.15144057030922908,0.43761329472844135,
 0.0711988688713866,0.21817658274502794,0.7995010011399256,0.9682316089644922,
 0.8463007414336948,0.30974967837675416,0.38907354716972475,
 0.9404987231537721,0.9327465467953454,0.4630663452768694,0.8293467367525325,
 0.2776388107792829,0.9138066987275354,0.11973777118271633,0.8287011495651263,
 0.9221809530228915,0.10941623002074741,0.7601857254828289,
 0.22377046182016913,0.6552022527115954,0.6317077196461793,0.2285022085579378,
 0.396992149602161,0.6376018453744596,0.1080395194770456,0.7753808313683901,
 0.5998489728115775,0.7031596110926045,0.9978313421674081,0.9162000494251311,
 0.5465578491597374,0.447900888940412,0.4359585203735836,0.6926298814954988,
 0.10824669888928384,0.8627598822014148,0.07807809102744268,
 0.5277566950467776,0.5452835932733793,0.817660329100151,0.8424897984489428,
 0.8891139858897288,0.36097032857079514,0.9878106329281345,0.7055688477782053,
 0.4888293621637527,0.3869767448424778,0.9254589900069085,
 0.004126660492138945,0.815556322470735,0.18349433275617855,
 0.9103527442922691,0.498920979491987,0.4561859488537613,0.17665386598983435,
 0.01365570904248714,0.9269718007082927,0.22555440476838085,
 0.4443616124509395,0.6900245043831101,0.23537465478489916,0.8444308097496471,
 0.8225431708825937,0.6869604949651328,0.44471832486949214,
 0.09341043172548924,0.5780986968270215,0.14094966073109616,
 0.22998767760740157,0.9454175777135693,0.7682590417262191,0.7421423280657216,
 0.7531316272123707,0.1912089954228975,0.19374390106002737,0.2800415664892354,
 0.6433549632781637,0.9209249312834793,0.31249391510947766,
 0.20377606000934778,0.3744606618083305,0.01610225911030927,
 0.5236262377473206,0.717952005029549,0.9434454259863004,0.22448501415190025,
 0.6955722047881436,0.4187284637766273,0.8510172060288972,0.408275813066976,
 0.7457275048952413,0.8259176392483021,0.011737214975629184,
 0.47243500792701143,0.37631879671922885,0.788675440605972,0.9084946083732551,
 0.004890198461254602,0.895853386290815,0.7019701007225541,0.8084272984069353,
 0.6719858200754792,0.5016176179613305,0.12901885349543307,
 0.027089799523346114,0.9909616012798252,0.14924215045100953,
 0.17937612199340425,0.9224791045299637,0.8216774414965691,0.7888283205715836,
 0.004765552098106807,0.4318681139384243,0.2591363486106347,
 0.7162506262255075,0.21877373586184012,0.5187755084090591,
 0.35726834353428805,0.6939579201886341,0.9177921305569573,0.5902128395487234,
 0.5701737429606963,0.8817320986661629,0.834891326525501,0.11754062552174471,
 0.4011286670727666,0.6032192739049957,0.5019720402532349,0.6949344249714776,
 0.25348711321379735,0.35234963065888125,0.2074253416691122,
 0.7857660374984469,0.8753107819709001,0.4296709741525845,0.374284620404863,
 0.1335325773369942,0.27263799197724514,0.9871685380311129,0.3271612971235188,
 0.8291393405948931,0.8298056053529079,0.625283662291036,0.13466168644312904,
 0.9337103672647167,0.5518042367061753,0.7659937470070443,0.640326161965944,
 0.9538171861190868,0.6389106698406557,0.5841293029250281,0.749264866054254,
 0.4552384830221473,0.15932426168713154,0.06741382977201027,
 0.9922416949177855,0.04648247325928001,0.36288521022216014,
 0.39004482952622765,0.3070242178015187,0.1572415998351081,0.5336655962283275,
 0.7742146418900016,0.9563877617146803,0.01617325906611733,0.9524150501008075,
 0.582701857158883,0.4396835899554037,0.022739114140750583,
 0.44965681119990064,0.31883452916719346,0.9470726012048455,
 0.18834487018311907,0.987587359299437,0.9946026741024397,0.8395839977146206,
 0.12329313631552119,0.720202844655481,0.0190594282643769,0.43462575391638214,
 0.599877691314323,0.10687786792921661,0.6189517099571712,0.09139706209962972,
 0.23477266037013456,0.7391525765228187,0.4709228042300905,0.6144221260968048,
 0.319382223722792,0.19211972909992747,0.32024598764366274,
 0.29497752932513654,0.1884131998525811,0.372345215332735,0.8048131155726537,
 0.37025069030032487,0.7385868477220815,0.34581384159956463,
 0.28873563783676504,0.1150506772738954,0.7589465286344842,0.5201606106616048,
 0.01336823676272847,0.748204072725157,0.20965926423360837,0.7109545832538495,
 0.024025061984411344,0.29879788965431686,0.589645344951048,
 0.3379937656736173,0.9432175894380057,0.6049103708360642,0.4493948565514023,
 0.9115696469035226,0.41991995804442406,0.9590544280375533,
 0.39027398459381035,0.9242854008615237,0.7287200534763867,0.6315295580960381,
 0.08758521462357072,0.23350112476374418,0.5880507061752044,
 0.7141628838754042,0.9571796278497794,0.13422120233261436,0.6278823516826405,
 0.330950421131329,0.8993318796515919,0.31592784276746666,0.41297323848195233,
 0.6368665960730893,0.7184251455882944,0.9460167759964477,0.5534321176483549,
 0.8753998693761972,0.9554995508017666,0.45258320825504006,0.3631002633113597,
 0.9713492269516583,0.37059920353153986,0.49494432560424795,
 0.046665420141423164,0.7324821443599867,0.5320090684624693,
 0.35251231585098197,0.7822917831019389,0.27013531808865654,
 0.3687696171696381,0.667760222011159,0.003103512132592412,0.3874768263949292,
 0.6098641002259746,0.6791180265155925,0.498685796380789,0.2916908919977287,
 0.6415569619361435,0.48679042079391177,0.4170849544116356,
 0.26058024680794944,0.4161616237042385,0.11522096970247886,
 0.8374234901532414,0.18887698062629643,0.5149136975063274,0.3087316150347047,
 0.729874629322175,0.46732090653925495,0.34678469329284234,
 0.32343187982300414,0.9227404546950365,0.2553385804590159,0.6118449892505438,
 0.653280127807412,0.7885866838236888,0.1423090111805504,0.8867984056932477,
 0.3999533697498935,0.4996841284377742,0.9001576357217967,0.45390238089223534,
 0.7638286054411052,0.30305418954315766,0.6831793243421551,0.6068843722667815,
 0.015256834490932292,0.5941913949501793,0.10244576531864102,
 0.4103956447112006,0.5707341013348707,0.6967753843263775,0.8506945138621426,
 0.45437858486971094,0.5670780288492072,0.4979257083092855,0.3960142110425495,
 0.5134000773959307,0.7522755762057187,0.7655271294948286,0.20733119952441104,
 0.9267358125573563,0.2596952305198016,0.2762306515565168,0.9650660206733211,
 0.6173131485809261,0.438804716546358,0.9581380480025177,0.9532171014143597,
 0.07301319337001366,0.6016504615200408,0.42225470585217606,0.825639912317833,
 0.6656665966588777,0.4334077067895724,0.16042829812501747,0.8465977345267994,
 0.11756551574921414,0.8597345508547718,0.6572315444120621,0.8542723071853118,
 0.6390221697764401,0.04787912976911468,0.4387518905435084,0.4581936043247037,
 0.6787955626496907,0.48923527709423786,0.9459091822526555,0.7210880926663373,
 0.6813340329370609,0.21552476403275567,0.20913104067596022,
 0.3196057863010917,0.3421913514921061,0.623756668846015,0.4294407916319847,
 0.8410810243325332,0.2998443362734491,0.39484977061443294,0.2763112506909534,
 0.05090323989624834,0.2810789035189526,0.5448164108886074,
 0.40016541847736276,0.69422559948518,0.20306683720701146,0.7577055095266018,
 0.9996964761481788,0.7544376213830736,0.7497254248003021,0.8533475230514875,
 0.7664098268254043,0.7785249686005509,0.700395249092074,0.40212459507294973,
 0.22548983735090566,0.637743082397138,0.8773887263809916,0.1683536977713983,
 0.6633087641748612,0.5071937526035476,0.9293500736618241,0.7597626198240165,
 0.4154807184487239,0.9978449899204197,0.6604398368292156,0.1212744167738008,
 0.17277163111899874,0.6137943879732004,0.9200477421277334,0.0498640034307809,
 0.7600881084028555,0.12699076266194553,0.5624488445291009,0.5194187925824659,
 0.4506106091103447,0.22137817163980955,0.4544384454091962,0.4778223476481963,
 0.8301282524878012,0.37913079910699854,0.5724802774857282,0.8043288855559824,
 0.9745609680008369,0.43198021968521294,0.7695107990862056,0.3850280974475425,
 0.5748177433841322,0.584592117801552,0.7793263412399363,0.17103285403743507,
 0.5349041251081805,0.8928131952601991,0.607898492052603,0.8100957585126023,
 0.347837895713908,0.16722963982086003,0.27810243545424085,0.9576330048291466,
 0.35687532335893746,0.3051074489871658,0.5453916420950827,
 0.04157337235545311,0.9165770588277807,0.34280958442637044,
 0.4483875613175491,0.6789935672997649,0.13349435539626642,
 0.32045612238406895,0.46051012115204937,0.23856521369037897,
 0.5004292096458318,0.014908292199159456,0.244251686411049,
 0.10908050257084756,0.42876535168201957,0.04569026671801946,
 0.8295227771169829,0.09801635173783452,0.567952293796617,0.14337454255969206,
 0.6037720451471585,0.39159451265849055,0.7738206239635996,0.2587275075079698,
 0.3411633104831947,0.46669836152557764,0.45689240904076234,
 0.7878799821755887,0.621958890392045,0.940636968218085,0.8089543199637875,
 0.44132380716139574,0.903211032364596,0.7112027388337755,0.5678691185993927,
 0.21709360501206154,0.8250669638711283,0.6323430452749974,
 0.16012086513893475,0.9538983079609169,0.7770722228352954,
 0.38491788853597675,0.845699904659244,0.8344846947272904,0.14157130857718847,
 0.864360391338763,0.25998473092009233,0.05912487308026604,0.8468845845723728,
 0.791723075643287,0.9106623829446058,0.5730303537264925,0.29593089180984855,
 0.3907670762421781,0.3139825095833284,0.5179748858373265,0.7448909133353795,
 0.17825822734069507,0.9366696135157798,0.702490662248611,0.3772448681012517,
 0.05697541183264443,0.9115274724892172,0.2308786800770899,0.8432188544697066,
 0.7192936025866833,0.5647552165123826,0.7444408937522726,0.20660208697842153,
 0.4547599691549431,0.4029315622452183,0.21202972126689867,0.8221856809930373,
 0.2763760974442979,0.3058147000236715,0.3588397154504329,0.7806423518314299,
 0.7186771565866162,0.36577554299913884,0.6132426972115611,0.4206476688534049,
 0.83095178815022,0.3776206896509342,0.9575105345352084,0.8574757960561621,
 0.08767714732521059,0.773996496076804,0.12851724978908385,0.9809303382617174,
 0.5613203850393803,0.4716522120281553,0.007950830115935004,
 0.3689618123231283,0.7366295028905404,0.5789967948375517,0.8337152983981801,
 0.12915184795096057,0.21031234803147392,0.7811782237116139,
 0.7786017251952047,0.8748938675950898,0.255549303727337,0.19952959987500862,
 0.33450845654231454,0.8813864087878263,0.5964911636532353,0.5919285937324866,
 0.5321776475858522,0.4689379727757317,0.4017796477384208,0.7465370789544494,
 0.6149341674039402,0.4806540349966455,0.8225894002271233,0.20196221093018007,
 0.3033723755642995,0.348610277477572,0.2018098347259797,0.724426077784497,
 0.7424862156111358,0.38173597581754715,0.7204652591392549,
 0.18767083410137608,0.265873435497773,0.35338356245966485,0.3388827055460467,
 0.28469182378917346,0.008255271252419316,0.6793341260602241,
 0.5700575673294291,0.6514552588079653,0.3460732375623955,0.9707984678790466,
 0.4413412616420075,0.6066913990611869,0.25962244906149723,0.8775505293409032,
 0.8439068987621782,0.8017151218437325,0.17746137975941056,0.6308982890658483,
 0.3666918183375467,0.6551627230740117,0.18371043263374787,
 0.43942547480490757,0.19058340388569794,0.8732334325999533,
 0.9469963795487822,0.8197997855653434,0.5798155085457972,0.06534568889528292,
 0.16918731229452744,0.4844679772496665,0.11783378082617457,
 0.7530807314059348,0.7039603901642214,0.7996272157820787,0.33115000844950493,
 0.7645862826701957,0.3720967894529874,0.7589752234822709,0.9542630816275565,
 0.49289624364254403,0.7163112269502265,0.8190134967114824,0.4205044737915785,
 0.31844991893384667,0.11986560245801348,0.9343779430547833,
 0.11325682799928205,0.40031885229877284,0.9873149068032364,
 0.09573073823829248,0.6109194272536771,0.10666683106742325,
 0.7335201190355712,0.2779368630829103,0.6276952440608644,0.13264324551158457,
 0.6449986736579707,0.11899212323148722,0.09730318706387397,
 0.9152176791327192,0.12675671780344688,0.480606912327758,0.1520630662886865,
 0.5434871861013291,0.7676266513023366,0.5059653464655149,
 0.045448383411704096,0.1878719038015737,0.7857619262798978,
 0.9211626444503803,0.6664509211173298,0.4522849754836883,0.9946409459074411,
 0.9316693123206055,0.13587997900094784,0.9217998799054135,0.6402189904324432,
 0.3078489090098132,0.06044625620680033,0.8087508034178286,
 0.13374765255888066,0.4882163818845442,0.03178970635935907,
 0.6225107665709674,0.5809293049197295,0.02241637739736535,0.3830957547262762,
 0.6281635050259576,0.5131064463130345,0.6961638698191088,0.267941231321973,
 0.13162119938646388,0.25701146433263744,0.3692158312183522,0.970870294319519,
 0.6798825349438896,0.7470864063266729,0.3499193470343681,0.7792433154100279,
 0.705162070168185,0.1670272906296698,0.1639249118611268,0.002829853836947671,
 0.8125193432445791,0.7390942573555737,0.7386657647020128,0.2426039824272992,
 0.24771027521172073,0.5087637387305286,0.3495182353535171,0.8778642186026512,
 0.046187355734237046,0.16030607016222875,0.6235954456529632,
 0.27702254258040404,0.1222610362669373,0.6862864291924459,0.5998828323844954,
 0.11815303287967183,0.43275972674323737,0.18712560943033352,
 0.8590849313739091,0.5658083488340356,0.5731418923222644,0.8802458425017654,
 0.9374812427072892,0.7315777514993453,0.68661720668571,0.09844170741434666,
 0.23764337263248703,0.8801085990918117,0.23173817967819588,
 0.20223642141946163,0.7939585003842711,0.4180851013661193,0.023506591139276,
 0.42692974807760353,0.5596193380437057,0.9253186061389649,0.3542972240931218,
 0.001538077217166367,0.9895001324062689,0.6466912362423183,
 0.9665089518865645,0.8072296469645373,0.2677878620540636,0.8402852632537107,
 0.9296708573630365,0.06263529343200858,0.32706034909520243,
 0.8142286053769008,0.7221663188004486,0.38219179191181607,0.3677045811751315,
 0.1372350750417296,0.2706121067405183,0.9516442188610451,0.6605225885818403,
 0.4208934582267648,0.20461096883141683,0.6427242129191023,0.9919776992348521,
 0.8792849143166582,0.34937014837243996,0.5189647404880703,0.3160694723856804,
 0.8634423381768288,0.8505315548106726,0.38379971109353006,0.4290463061097116,
 0.050327707641664055,0.8794028780535251,0.6139289012184596,
 0.015125472516836158,0.48809019581105584,0.3202432888252429,
 0.9121871241507725,0.3417142826565137,0.9441826072413074,0.59393190465984,
 0.15817160722552037,0.6276124717249486,0.1477122447555187,0.9214964981493353,
 0.4758722298503917,0.3025262115291,0.7772846003475974,0.8856391037366106,
 0.3727797976999706,0.669065626476808,0.3446159596023475,0.8537287969249483,
 0.5732326860375373,0.14249233657845817,0.9407410713308677,
 0.014399525092261722,0.45352451444813513,0.46333492078647043,
 0.7416733530814101,0.6927624409472768,0.8784693187650887,0.601070879407068,
 0.2902570504504767,0.6157107358177,0.8254850505469512,0.8444860166987549,
 0.952495167239649,0.057384652692803906,0.1984923751759382,0.3929383803194195,
 0.5500595553939113,0.48327519429285326,0.7084125873975712,0.97238671274688,
 0.7379705104092494,0.6461242076624154,0.7175072759341505,0.2857179272797048,
 0.3535196495700472,0.9073148220126704,0.4735977999900417,0.27162153224128016,
 0.33374558064930826,0.3782426916068885,0.18647119074350937,
 0.07857794913528882,0.36939855763850127,0.11726789598752851,
 0.8697414833107651,0.6146399870130616,0.29242704650207807,0.479026271092141,
 0.9723733675985149,0.4657939480416199,0.18102352570235936,0.3537796737138984,
 0.7859662458897428,0.319119365171187,0.09093503463272956,0.6336612224280332,
 0.4285860669369703,0.19122031621144242,0.23907688634529423,
 0.5900491190402976,0.8723415460550715,0.05092457354125057,
 0.42790168157916875,0.42893729084679255,0.7254586925876325,0.791947375040928,
 0.9798756424863996,0.525544711996514,0.3286185827813227,0.6943013581950512,
 0.4879878606993904,0.6991454814211895,0.14203148200507698,0.1950253276319529,
 0.1306788390900281,0.7747738026016645,0.30160237780013777,0.325296377471058,
 0.72974668505811,0.1116634674539776,0.16275519319266785,0.717626018200012,
 0.5625389020318151,0.9267662613458533,0.07683484768671822,0.7050349768238285,
 0.10307295227904989,0.6603850171732364,0.991792153905595,0.4228317570053901,
 0.7197436072368822,0.7701878195657025,0.19589376155916804,0.8703364755646095,
 0.5762377913579048,0.7464878814008784,0.04138645482056924,0.7004274793094545,
 0.2180118603657486,0.13300147390615757,0.3624013442619548,0.8733650129934093,
 0.7759040878006979,0.3735000755865244,0.5841535095550388,0.33979955890353564,
 0.34827772341194063,0.8793968116539161,0.2590893819099691,
 0.045896948866686826,0.26373482353454325,0.9334457672692443,
 0.6912050131155707,0.9849807384583077,0.32268798768696005,0.5704808374950792,
 0.8029181894847887,0.16531613990475258,0.378663596442673,0.15704521739855948,
 0.6184805566809299,0.015835004262326646,0.4162662237291531,
 0.22395842298170743,0.9760419393538238,0.02990925439815384,
 0.1334761353989754,0.01998475812870093,0.39197430655114385,
 0.3856667572175184,0.8661288604348564,0.011119282779518791,
 0.3505794065588508,0.24849405593713547,0.3369494093927584,0.5019885220498296,
 0.13748113899758396,0.6384648126646907,0.39195134260894515,
 0.8523801415215261,0.08751308113411427,0.7226725652350046,0.8927958721062702,
 0.7734561167392529,0.601033650189066,0.20709819314342193,0.9534531946168651,
 0.30055371821664567,0.7392843304310243,0.9509769836435056,0.8829917821765112,
 0.25474737183392226,0.603240467656166,0.4810927385861309,0.4307871683533281,
 0.9130360732668146,0.8701809306461341,0.5810305899797386,0.17831049881335304,
 0.7676801308664272,0.6488257357447589,0.23914867201052037,0.5646331976120512,
 0.7027725794486451,0.8423005886181626,0.2717035366249787,0.9830525951237635,
 0.9272711592283397,0.5172349276391942,0.6363183788172981,0.6703732084633897,
 0.357413554981877,0.6372281574883303,0.8737237493482901,0.3922988986266378,
 0.48486984534330335,0.1873223134625953,0.13264986704055404,
 0.2940763381676298,0.08602551858239071,0.888713359374532,0.12227555933182976,
 0.9589192452818722,0.6695174466082854,0.6956633328658344,0.09861757874242272,
 0.5805247307220534,0.2100725434228976,0.5938866566661456,0.538543232361788,
 0.15893982185571887,0.5888053431347986,0.12496384765454926,
 0.7019551288972734,0.7674177085408064,0.9729680903169825,0.9132790820461822,
 0.20877691100676943,0.10423480357443216,0.2362126212986838,
 0.32468871953240086,0.5058013320493756,0.7841935632137833,0.8508626744790413,
 0.7953215932269743,0.6876894489390465,0.16675570573498633,
 0.11406096050359904,0.9882035083363266,0.3394885142922477,0.0723134437482531,
 0.18239450884837727,0.8828179310309113,0.5754769525664787,0.998243821131754,
 0.9519308502565879,0.904044842029434,0.6495293009408634,0.30655530218078564,
 0.32941887826210636,0.7926142547675858,0.32508033731311825,
 0.3989901184130209,0.4833353622331801,0.14024261014420025,0.0663874309896052,
 0.331553721240862,0.39339544312182884,0.9934030872789793,0.7518265103060439,
 0.19677713244420758,0.21204910964474388,0.05802216419658868,
 0.15313605638641292,0.5661634737149418,0.8105645437746564,0.9880797878745177,
 0.3760955252881416,0.05650909778948543,0.27766045909135784,
 0.4131625086457993,0.8211490396865037,0.7197183751078464,0.3167442320427247,
 0.9164824700963834,0.7526900329264821,0.8328837089348268,0.45949719692991675,
 0.8058894437103573,0.9670281793586143,0.4511663812565758,0.8792113207917522,
 0.745741622193544,0.8456930149384869,0.5192917665875488,0.334595738197223,
 0.4361271730016997,0.1945822071989448,0.49934168757298236,
 0.10044053749578996,0.15667678005796404,0.33352562672997843,
 0.6811170746295937,0.4308781513166826,0.5699015529661128,0.4285826162065629,
 0.9015014919268971,0.45443280088832105,0.5049003819183913,0.3180345561920589,
 0.8551636809206166,0.7299426350480464,0.3429871305862875,0.28182645227871217,
 0.10742810274183334,0.14132185120159046,0.8979643026063872,
 0.5047434753653869,0.9248913972207218,0.13118747708063638,
 0.27538341815445655,0.9461529354057261,0.9752695479667096,0.296351366876611,
 0.22598296431868248,0.76867998431874,0.47648373085615725,0.4087409206953688,
 0.09440750006395249,0.39403272011578205,0.23450968142180373,
 0.8226998492451413,0.6968426632901118,0.15455607047792042,0.7435797316047056,
 0.6223841667281812,0.896137390641496,0.3863816054506567,0.7040486439725903,
 0.4139530943421522,0.1988173201391623,0.8658290282862356,0.8775513724681044,
 0.7166446993452513,0.297533179971863,0.779977177635486,0.9570890988252613,
 0.6163079084455554,0.1259519214035777,0.5392247461694919,0.8239558271024183,
 0.6393057438810041,0.3910955596235798,0.3723320151557925,0.9797812596805012,
 0.9040252458780226,0.21241195424075765,0.8380405214228988,
 0.27697723835418675,0.24083184832109128,0.42876043473120984,
 0.9066012602798608,0.9144285060913498,0.5452784184462484,0.4600548052096919,
 0.9605220196486786,0.6790286640637802,0.8999606536208327,0.19582511171706396,
 0.039267776611233374,0.7064682374813551,0.9223123848694212,0.43042290696995,
 0.8893301963239427,0.3788817013824577,0.223472389505733,0.41917169605053217,
 0.23845732192031988,0.03887020012418907,0.7478239066705636,
 0.07683400823406505,0.9400478960955619,0.5774280112530252,0.6800377703161475,
 0.6444824945006141,0.8272642377493595,0.430558692063593,0.3686437458874947,
 0.7591728056176086,0.19534498401829725,0.7212629887746008,0.5432544377979567,
 0.12121129343325887,0.010760413639144706,0.08967989124482478,
 0.8525354220899377,0.7546046131314698,0.5751014196532825,0.2191198406859366,
 0.63980950808604,0.780583726448123,0.056381749320257746,0.0240042965515479,
 0.1644644283966712,0.7329524524603557,0.5701764738100085,0.9713533291547373,
 0.6726622551861485,0.40948632490231596,0.28707331343696385,
 0.2753764740067315,0.560563145707958,0.9964552705655644,0.9953132984775395,
 0.6316677835961095,0.08168160280344294,0.7667686122371856,0.675975733416647,
 0.9738104841585977,0.10018488096749589,0.7713007021722055,0.4807488118171648,
 0.9498852805718745,0.16587777270993243,0.882035351754587,0.7368791438653345,
 0.6574023526369921,0.8727087186955176,0.8537613234956821,0.49093755157414,
 0.7151177008736407,0.20355054769562053,0.5839440972741231,
 0.18683592260972604,0.4275302819506168,0.8683132529668387,0.3114633475002947,
 0.8423987545176315,0.98579215902957,0.8789704180893475,0.6141383927693836,
 0.3841136920768776,0.05365200953647997,0.1685915948541643,0.3878643195247209,
 0.1700407536368247,0.02472274770924021,0.5153561292797373,
 0.20232435739114363,0.47223531199314295,0.7088526171183785,
 0.38977456783971265,0.829877833865232,0.3894497242307148,0.17305804939752956,
 0.27355865166437576,0.2629292991675575,0.20300458786706266,
 0.09306337210751736,0.6020458070227331,0.4221111960633668,0.4771202996836774,
 0.6678879958653507,0.1822708155325662,0.3013490144772727,0.2464971801226541,
 0.0686224825769407,0.5488800413489882,0.39652182711099915,0.579481578188776,
 0.30155838377383093,0.49523250781154904,0.9332331524431896,
 0.7135097925010297,0.4441661969580195,0.7788198983461363,0.949067834610678,
 0.5479471904821687,0.6432149404928539,0.8547128822783395,0.4675905945078007,
 0.7154613898654871,0.8301261069007237,0.44923354084387124,
 0.30169684528063523,0.8278560300695628,0.8414763442227831,0.5476225667755111,
 0.029005208893823853,0.12190447243888247,0.3674448536761161,
 0.7988594755634159,0.5138933923588751,0.5129946605542847,0.7546213993606521,
 0.5531864140858601,0.6675342871311329,0.5779260904059673,0.8370793405929483,
 0.7064972883217853,0.19566471533610297,0.957967022397087,0.6883787990445568,
 0.30315683386571535,0.5767836848335395,0.5770647548281298,0.8274630899155097,
 0.8665846781012267,0.5510041222190137,0.5891252413742845,0.6019048952335224,
 0.3879829106590652,0.8829129210877971,0.1858576290470655,0.6386086600453744,
 0.13999196473579167,0.31177715794871075,0.331828854759493,
 0.29139256279969983,0.9364534777601301,0.06632941101487588,
 0.37889611450520067,0.8078818254362241,0.5394796107968348,
 0.47058265199537297,0.0018532292398705241,0.02051786209902484,
 0.36845824158040064,0.10350200748824101,0.18330128307258464,
 0.7224505229206225,0.4687751732200224,0.319083565845352,0.3840615123384389,
 0.901205582298467,0.7474317368137722,0.35412092503544035,0.9524526383172043,
 0.6969721411727963,0.3891995619481782,0.24134226682187576,
 0.19909381095143686,0.6446722184421835,0.19791038651073667,0.884921040907173,
 0.09591549495217988,0.631695259603575,0.11663310989897069,0.780571495262191,
 0.6217905498584523,0.6558679688641098,0.6648428890350264,0.6384809895418246,
 0.9596150771018687,0.8051828105340663,0.0048408994985715825,
 0.8077611977640508,0.6659071141571954,0.10723177859614663,0.6386990847212742,
 0.3699934787498418,0.08238220837242116,0.7359229610846199,
 0.33570882695922655,0.8384487133060653,0.017214855282657204,
 0.5932573958939394,0.6787957064133913,0.21419957552214608,
 0.17311238770734694,0.05931107231492061,0.6292660702955709,
 0.20135953787817062,0.5005548060259359,0.846318741907626,0.18286854853803114,
 0.06580197439355828,0.2071828387356921,0.06982087377094592,
 0.42169843663773576,0.5762800640940037,0.38960060873006397,0.601240423882776,
 0.8521896491589378,0.8944344628116098,0.4537713487330912,0.8618866250864432,
 0.5476139100810019,0.26287709013523175,0.7927676421895646,0.8333967964427221,
 0.1995408874449892,0.4962527347059865,0.13892685399300853,
 0.08982785330976939,0.5538993474179524,0.2993812044065809,0.2101680559830057,
 0.9075926124689963,0.5150060254771188,0.378364608959774,0.9994777874083215,
 0.997351061649843,0.0816876267828045,0.46489018537785765,0.7030711674304909,
 0.9090415480192688,0.3457520472817963,0.5815369863695824,0.8446556728360838,
 0.8936052555044236,0.0884124237706091,0.06432656270220993,0.1716373794113345,
 0.7588318001574654,0.9088201654314408,0.27541522965354126,
 0.056750596497872396,0.057808913456322886,0.2715832830660112,
 0.40846048387235456,0.48389810533766753,0.59888940149217,0.3099181354632996,
 0.37746168055126594,0.9653289521276368,0.5670937116408578,0.3650154007469759,
 0.760633767031057,0.8842165615710887,0.3744378720516922,0.15542030328670786,
 0.643165198902001,0.0029963294678839736,0.8440332812286965,
 0.7695680498807685,0.743660563907448,0.3264449592119981,0.6186074476986344,
 0.394569690936333,0.8667684379231517,0.5751365281729708,0.22408264111035336,
 0.7264989534294038,0.9722286997906796,0.9657395136191571,0.2529725881150183,
 0.1466569132926141,0.05462679569185125,0.6150828930825663,
 0.21129273654747227,0.9388080802151477,0.4413266751894821,
 0.39222188003044034,0.5817992532277879,0.8364750854026193,0.5873947632558938,
 0.2772151251482201,0.820771074101416,0.570220650990571,0.5943626948729862,
 0.9576510196452472,0.08292667061614378,0.2856755194653915,0.9407035710727074,
 0.18347581108766875,0.6538412394590695,0.5231075151954203,0.9499160847808795,
 0.32346274534803743,0.05055698882376647,0.10546126832952984,
 0.4433757397867173,0.6546593468214821,0.7006315313954389,0.8070803336558249,
 0.1075432624463789,0.2045103424364889,0.6589731303156903,0.9131091772818222,
 0.27845497561627175,0.08451576360436641,0.5342717674705451,
 0.5355225170561179,0.8293933779141331,0.4030395797141657,0.04927272546622519,
 0.7292515131046046,0.716056590233836,0.23176923792858162,0.6783253331341327,
 0.629604077921597,0.131814785582281,0.08203465588193559,0.7220587250771507,
 0.1303244452529314,0.7202211698381479,0.3015706632885514,0.480137651872369,
 0.37221009806507,0.2570836289847589,0.9341020884729638,0.028476755857595237,
 0.6002877603917754,0.49532257983288186,0.3613855565200226,0.7653843534594488,
 0.15325389855692007,0.9809242557686204,0.48360287806239,0.19490783444067772,
 0.7380102112132759,0.04367883931136973,0.6463738096559329,0.9253089326087004,
 0.5197519632680653,0.6210338352532726,0.3094598696671299,0.7135834230542458,
 0.30069447716668407,0.0185343402661704,0.5687517879347351,0.8123535839264371,
 0.4278909218981751,0.008477855927594424,0.37423997831548605,
 0.018945990908017407,0.21831243561188052,0.2628687511982286,
 0.1422035081992975,0.7056407242639497,0.5716058857345274,0.19795586299457835,
 0.9826988017563956,0.05728691423157917,0.5390113925254736,0.5882272827356583,
 0.14674493103378627,0.4405467163908122,0.25454813730226267,
 0.9088527904453573,0.6643582189088065,0.9660420853786932,0.15078332446113574,
 0.9872621768963605,0.8276118003591619,0.8659534908580808,0.6752538896872229,
 0.03426632194757473,0.39641490174351734,0.21435296386473546,
 0.7029877514312501,0.5586660876913007,0.3452547864825236,0.8033989970217633,
 0.07511554205142357,0.9004369818638356,0.6112232993081399,
 0.39496363252984934,0.8046130497202169,0.8732821138286397,
 0.17182037542379192,0.4129675711876232,0.33856464421569643,
 0.6935699790121257,0.5251589193123232,0.855657327149235,0.6039990533150136,
 0.4612893338207035,0.28732573468888645,0.08913803987437163,
 0.8942965547567016,0.8851538673637164,0.9235718416050106,0.9591585501034436,
 0.835931830999822,0.06602155712432767,0.24959527366024226,0.475230633417944,
 0.4471475727489803,0.849869399921474,0.3238493850382822,0.35693832546936843,
 0.7535542992415731,0.28691958261961736,0.6346089757487993,
 0.33960737371219185,0.6839825337005826,0.3808857421698921,0.5828876633856729,
 0.5664635616676765,0.09443012026743502,0.7496466130988589,0.2746565430394904,
 0.49519873281109017,0.44530703938951344,0.2515770603367363,
 0.01189938906525212,0.262473505739365,0.9966797846656712,0.5729396933532234,
 0.4790353634354728,0.08364124142667806,0.4553848728694141,
 0.12259039276431716,0.948578309331269,0.7502626875289098,0.7311367957686751,
 0.6021186881520537,0.7090452569609127,0.8117890320595006,0.2864004637782118,
 0.04760832218623978,0.9257017199570673,0.48903017782183433,
 0.5088930734922579,0.6264301606308806,0.028862530841710088,
 0.16406338856441482,0.4919957597359399,0.7321049848565554,
 0.025381682404710393,0.15573527823057676,0.6135991123109811,
 0.3933263971554949,0.9813980098487334,0.8160846821906307,0.27556121932646815,
 0.42805595815112796,0.22907234617691818,0.4973904378882882,
 0.4186399813795618,0.41326944959797873,0.6352271181693492,0.5387742490337748,
 0.4646389569800111,0.9588990192238729,0.7821771124702805,0.05097928211163871,
 0.2351604750191122,0.22889094666332388,0.18869804982064198,
 0.7062352079239891,0.6791651267124778,0.507926796850764,0.6598930471829899,
 0.29133986305110426,0.7648492649254903,0.4348338656699128,0.061339377518561,
 0.0469057393159138,0.10376267193022204,0.8143306538547378,0.6750116929317894,
 0.6816547863591027,0.2772810982828241,0.8269930152464116,0.6156771003666277,
 0.6238196217681199,0.29757029176443406,0.014978113552288508,
 0.3542881865310302,0.13643993157649037,0.9212658751519215,0.7146725769306473,
 0.3178337058020878,0.6177325557889481,0.8925072731598619,0.12807646448444032,
 0.20090564844893688,0.8514362837687288,0.8913525334094389,0.508439274772201,
 0.9784070511380373,0.24947964950443358,0.02355989165538308,
 0.7411526490670598,0.17959813781659717,0.7807926043985319,
 0.16741430804443658,0.7174118732238047,0.924492897445975,0.2806849511740015,
 0.42385497595547794,0.4370748258476772,0.9855303306419934,0.233064825382039,
 0.4815428202478378,0.8317683273113498,0.3536449343964586,0.45956930385459305,
 0.604808356566395,0.014184640131032267,0.16403108000105815,
 0.9811216410141554,0.9119779777387446,0.7736540992119573,0.9289511380401388,
 0.12781913447677395,0.024877375562212145,0.5718630382258412,
 0.4578149437155309,0.9092236776858463,0.7224866064023039,0.7002686611338533,
 0.2618273299308854,0.5365337435527052,0.8529752506187143,0.7684485313787046,
 0.6979550885355328,0.14747070832722953,0.7193088305403103,0.5087248366799625,
 0.4272248181145689,0.5921508240789395,0.6550112790187713,0.2691301790834507,
 0.9042541552262654,0.9012105914798241,0.9682143098983624,0.11420533313190462,
 0.43984991894896397,0.22622883899328972,0.7180893698136981,
 0.6834493245945317,0.5884980656115787,0.38452733627136315,0.8128994581561604,
 0.695939143223452,0.7576248966198694,0.6980578879126121,0.4771049991533184,
 0.6460509856458023,0.9568556540268531,0.1755140094136507,0.422561505610752,
 0.21180217392907108,0.1542039276844518,0.08814428594501555,
 0.42104285605826175,0.8000464577312727,0.4639655648116996,0.6373047136181411,
 0.12449667554269872,0.6948294869746183,0.12753246252600936,
 0.5195095471502444,0.895821154847234,0.39873286434945154,0.15532482680719872,
 0.33037518812956845,0.42733711706595057,0.10210485952629922,
 0.37419107037466137,0.5690196343810471,0.26257596791575055,
 0.3064621829922849,0.9985605460710629,0.827516597389975,0.7317994938629792,
 0.19161156307072724,0.19507293930306935,0.0012661022931872434,
 0.7000189190265133,0.6365105229322029,0.9408016971763874,0.08481378022746577,
 0.4075710980552867,0.8671756801401295,0.9733694246209089,0.03488145014440647,
 0.4292572327352474,0.9769969684970368,0.9679768368359105,0.2100175773439914,
 0.23345475031181662,0.5338427522588934,0.0375443356817341,0.9592679951341684,
 0.219166330955206,0.09961431995548453,0.19237862643337977,0.6412794643199722,
 0.056397285786153994,0.20214128774634899,0.6339529387109742,
 0.8570781248328684,0.8421972036932547,0.46940938623104833,0.9377591573229328,
 0.8304634507580558,0.351418946274223,0.39747106341225436,0.34893988795124464,
 0.904832308460071,0.31600707080717627,0.5182868756392858,0.647505246360219,
 0.699361683303721,0.9726529560277544,0.5432830623829631,0.37702821790835606,
 0.12149839527518114,0.7058103954829248,0.017731098158278447,
 0.2991306159754432,0.5442620909592465,0.9321432863306786,0.27934628940108197,
 0.11980453448434403,0.6792145624875973,0.036296749457181754,
 0.808669909297308,0.44305245914623625,0.960470800965558,0.7047517513011508,
 0.09226279523301673,0.9089522878718165,0.850439191978504,0.6447560633241852,
 0.2640018699301989,0.8585060749386345,0.13436870642091858,0.8774360444063497,
 0.716272917771501,0.18072893147132396,0.47447858963304923,0.3939414216905346,
 0.30634906474599677,0.6857662056198869,0.9553637907135112,0.8938513523801556,
 0.5017634809152687,0.22854652607191084,0.4066250835926195,0.1713812002213032,
 0.4075012779194124,0.9524470380734945,0.7781887807976499,0.9899971360487929,
 0.8886538080885692,0.8958085375347399,0.6912127583333889,0.8769864210844659,
 0.7699847007695519,0.2412797318811164,0.8804395441271738,0.13742672030667424,
 0.051164032806380155,0.8826157152846239,0.8154370971059498,
 0.42617975806021513,0.5458723947257896,0.5247829864347356,0.3379719914507294,
 0.7455971924945994,0.7665180914515601,0.43056293053778105,0.7682585176958792,
 0.7872742214260123,0.9409441443838192,0.9118533549272246,0.09999670069173816,
 0.8630715549976182,0.8670711592499368,0.18078118071085814,0.063722550471499,
 0.03289455615318859,0.27133456799536004,0.3426083564845521,
 0.3776423515238889,0.781954224685899,0.30158228981213164,0.982418514268639,
 0.07997767699779346,0.6210415543984009,0.7324072397621161,0.2889511113051624,
 0.14190675950106424,0.17118864328957673,0.523459109246175,0.2835773592239176,
 0.36606359387335585,0.08683452856785268,0.02217918937482477,
 0.5077841721942654,0.5831378864837881,0.012074352765089724,
 0.07316494072837054,0.8345297314567779,0.44780798759952456,
 0.9647533121272589,0.749024042006909,0.02108563115822326,0.05735059756059668,
 0.5659296258750953,0.21633456708175114,0.4159022540555696,
 0.19757665652213263,0.15884645028069388,0.0450307943223226,
 0.6950409200706161,0.7215019498179216,0.9599072586606731,0.23033640275713096,
 0.2881464428456889,0.6158771961679017,0.7705139974441173,0.936118603626292,
 0.7315575731254153,0.7077852881096501,0.805369856426787,0.23851578618059455,
 0.885926036422903,0.2773575135810342,0.29385040694451214,0.04323718165468615,
 0.043005966405208707,0.2505533268149118,0.5520322039942132,
 0.3648699451712214,0.43338080974221993,0.4104645025072433,
 0.41206441382033576,0.8115618420800467,0.03430792399952676,
 0.10927357321846853,0.5617147609819023,0.08674621869724497,
 0.6914215824981144,0.8826892175821021,0.5006143190118947,0.6373697002771707,
 0.7643217262588686,0.8826275421142262,0.03392588058470292,0.825762359720875,
 0.019969426271875146,0.8894894744893711,0.2655256622909813,
 0.37381721099028675,0.028339966477752165,0.8193180742973862,
 0.559506847719537,0.4308250149240652,0.7970798782659557,0.7356011844699366,
 0.7191271004652062,0.7682870139263334,0.1927119536801618,0.3847521250822198,
 0.4760677123884768,0.1490181582731337,0.562638138165249,0.09354485120025147,
 0.540785901560748,0.7831834309864851,0.5851393435581536,0.8626612721240082,
 0.6455308233408218,0.8753498816932673,0.022730647546428395,
 0.7619148903605647,0.7196539954048977,0.8243051023866902,0.4490128004678893,
 0.41571684939656517,0.6026161155106166,0.8415178067301874,0.7374347967779549,
 0.292498478962822,0.9555772203398116,0.9060157717016346,0.855354170326562,
 0.3562897689854412,0.006639554367314449,0.659641381654203,0.805134789967765,
 0.49081572645133487,0.36748458289414654,0.5815390970531927,
 0.8952272414845355,0.8001392676218279,0.7581134286508853,0.8020145013964506,
 0.7227670348797814,0.3949789999281511,0.8274493887372816,0.3415299602202233,
 0.45997849613111796,0.9090424910253705,0.5064373188978898,
 0.12773654283928004,0.19698397583868854,0.9123032051712303,
 0.7473708178626414,0.5202730220810718,0.5674017937855262,0.16397040203381597,
 0.8129360063432406,0.15141327841842356,0.1834871150755779,0.9817393866361159,
 0.9136161692901993,0.6333739012171847,0.9302841158061461,0.34605001663276613,
 0.2147318035947663,0.8077916810444947,0.0215772878813707,0.6870780301202651,
 0.831399969706107,0.5736638678144405,0.07969362956928883,0.09450364174134018,
 0.33886693729188,0.5563639964127554,0.8700756481290897,0.17497720519579374,
 0.6891371402860378,0.6200761179284653,0.12273618605874292,
 0.20697296581365443,0.05843164358606212,0.7525547695908871,
 0.9747648728371321,0.5357659689023652,0.9627647630694449,0.3017830000781285,
 0.9801654617572424,0.22033049255352954,0.9076889518274208,0.667260918479687,
 0.7181711771685708,0.4772114395721053,0.35598417820627415,0.9990325588959011,
 0.8754606710842054,0.11326369184089835,0.5891422929014636,
 0.43248150437831234,0.36604296997331387,0.16878339513691476,
 0.13037367160509317,0.999311266197182,0.38504632899360414,0.860840429878107,
 0.44279476148965524,0.4370890284195531,0.8357403507862844,0.6997106479565884,
 0.5379169834342871,0.8711898222241191,0.9609319605630451,0.16987693323650976,
 0.004135073976665904,0.30440741843575414,0.5093490654919017,
 0.5004885686416916,0.9161959210396312,0.20602630199891692,0.9754813263938494,
 0.3578037992558095,0.1762687026379247,0.6032171292156421,0.7292145924292047,
 0.16236898187716842,0.36477709832822114,0.19906734024391537,
 0.8690908761840095,0.1097804424921931,0.6742221228448131,0.08647721434536981,
 0.09351852599538413,0.6165584333837341,0.46466309700913344,0.259271648104646,
 0.4842152540419311,0.5665625269298106,0.0040818765455136186,
 0.8290449402693044,0.9012768272861781,0.41554715046558977,0.2995535302547825,
 0.16872789260750098,0.8327158170992013,0.9410018348837544,0.6144281151569226,
 0.5427766281917948,0.8404902985061881,0.779728513278924,0.2804083954389952,
 0.7250217416819827,0.13605986107684476,0.14323788697978368,
 0.3545881108020452,0.34770527670762275,0.5425540335176164,
 0.36595933303398354,0.42136161421613016,0.3473601052047226,
 0.12227470014209474,0.6718344176737112,0.6018869300390576,0.7349589337996456,
 0.4500471129599334,0.813411451579969,0.42393866107844014,0.327244365763967,
 0.6090017480167348,0.908968437702638,0.956202494085424,0.2954890858539678,
 0.7894140241609426,0.863687313818977,0.44206346489958737,0.45925894611430085,
 0.19867895806940505,0.8745264674313393,0.4738067852982777,0.4017700571432984,
 0.34321305463356244,0.9284120659812979,0.7230671412357665,0.6282515473432633,
 0.47950899974610406,0.6492590195302612,0.5632251018874119,0.8655326051387522,
 0.8367107292610784,0.5500347543189462,0.585766797142227,0.4847791540672245,
 0.06106380698658387,0.4001919149518134,0.4622460376476316,
 0.11471834356698452,0.9166735045944955,0.7234954428142957,0.3109882052443861,
 0.6339652343038527,0.9703769753275213,0.44549059689014214,0.9267117133107816,
 0.8374310544461838,0.5843237285682056,0.3070541831718572,0.23507336473094775,
 0.6315259148545391,0.6694559884371664,0.6636042741069684,0.4074567083471402,
 0.09549170814166996,0.5027378425066681,0.7705939265212524,
 0.08728173875978362,0.2065365834774141,0.803814015844771,0.7079979432962605,
 0.6804554782532883,0.2352645445616225,0.5376364469778983,0.9138134582688998,
 0.16129628931081474,0.16936556034755101,0.8352440052771765,
 0.8199763072145945,0.5002343991174998,0.22811062304049579,0.7612953618463251,
 0.734134316028414,0.9913490907639044,0.27196902513924104,0.8513100826914006,
 0.4451212212717123,0.4773304425716389,0.4070785592124202,0.49896302465208664,
 0.6939744761272566,0.8572651441898381,0.6101193982109555,0.19179820420198535,
 0.4068588242851462,0.9379602943018475,0.163648922715664,0.7672941446812263,
 0.6182557499105621,0.10823535207237633,0.7925606289155573,0.6297729175260908,
 0.6309400553183282,0.9439774768662845,0.27098560109238845,0.5979719622993095,
 0.20019172438364674,0.5424292474853496,0.656913278780829,0.22402431747904328,
 0.8521307641059781,0.7929032078411876,0.3536306466568868,0.8314648286043281,
 0.9213265495685976,0.5374951633451492,0.08059634534755278,0.7241254146456796,
 0.024053553810307382,0.32847158216830585,0.021056408290707762,
 0.5877735113376148,0.7056961075592766,0.8440071552493053,0.24174972458151045,
 0.14825311284359988,0.5192390905553564,0.965599250727134,0.4062350759868194,
 0.9881283178599684,0.3603401465708931,0.1520877661374791,0.40373093188413633,
 0.9993819470761209,0.24776334154203938,0.7190998835836386,0.4308234957339423,
 0.7815315967646324,0.16359173733012478,0.23525588304422285,
 0.45653216779166705,0.8806836454455279,0.6814961135118804,0.8087047203013172,
 0.3497647289965423,0.47831996833296064,0.9744558235258097,0.834720317155447,
 0.8046937276558537,0.671297989556445,0.3023130756051423,0.782452246939568,
 0.9656795339055168,0.5598465161511998,0.5374792862817908,0.16122782136903546,
 0.8962216058096986,0.5133150954793788,0.9516163098770709,0.24331014034374676,
 0.538530796061782,0.8701470973891745,0.2665563029135787,0.9647498953110596,
 0.9562734022674236,0.19606081833199107,0.010596466393806403,
 0.8852037522564877,0.40412928765364764,0.3580968615813003,0.5797758097602189,
 0.10347096189251692,0.8847886764424202,0.9048901492285009,0.6993087439121883,
 0.018086694610267462,0.9440293168703296,0.9974905885959009,
 0.5562635595208968,0.4595486131386297,0.5351245690324867,0.8065675767284082,
 0.044881055295669325,0.9591636314942097,0.5437843351913554,
 0.41018752923309243,0.6846505480511277,0.2318399621830718,
 0.09784969215696082,0.5708582661949896,0.49599053696484585,
 0.2950932303465781,0.39146240160088586,0.2631362548240781,
 0.30004520063267004,0.07461430172129069,0.4099435224984215,0.87594502708322,
 0.44341468410234386,0.7264208115799388,0.40849673911579276,
 0.5475438116380197,0.9961214637021807,0.4929447889479729,0.5013289580756157,
 0.8191244888006569,0.4985425425798431,0.6016829016803011,0.23599674514087732,
 0.7397682347924228,0.36969849567904356,0.5653442205282162,0.9937191331214746,
 0.13427892135763586,0.9025067077714596,0.9144579013701464,0.7953106847477842,
 0.41806304599211497,0.4799462868319181,0.552298298024326,0.20462270135829153,
 0.7415022572763613,0.4205782249051,0.910212181653799,0.8039925381334005,
 0.15555815046260202,0.33201879001326373,0.47050049632474844,
 0.1262979574418055,0.7023497110992303,0.45435589202836113,0.570329147033614,
 0.7999174024582628,0.513946882117558,0.25650357323408457,0.5622069612703882,
 0.5013596965009661,0.4912420398871493,0.1259561765916678,0.3096498322719635,
 0.2534677495497375,0.8235865388203526,0.11183262593575805,0.9403361706629656,
 0.3476255994103772,0.175271059263453,0.7656622487684825,0.8240846902668509,
 0.6978439899467276,0.8363900558972396,0.6068713210564738,0.633979670832977,
 0.055892395094982894,0.43142439127866017,0.8149077463820253,
 0.6103662905148468,0.16796646268059157,0.3262537591996474,0.8679728894990408,
 0.8413708372990676,0.6875294424499074,0.010405240348612343,
 0.8495956611244323,0.6073073993274962,0.4081261931798331,0.6586183005363003,
 0.875259344612443,0.013687719210450866,0.22918537780082826,
 0.21403437286743465,0.41789937999061044,0.8658692466900274,
 0.8496205023405173,0.5314099590278049,0.8925537843693887,0.14561432024009824,
 0.3445047385060934,0.5716029861063262,0.06912640538996329,0.8813659343679093,
 0.04926206409478573,0.5529692290855475,0.8432702486025311,0.9505270986655975,
 0.022120516425779946,0.5763508223898357,0.8410307615072383,
 0.49750501883394005,0.7843012601526735,0.35930519389169746,
 0.02679080610805551,0.1416640044923786,0.9275338381187488,
 0.30970421246497526,0.6651354447899515,0.43742079405263135,
 0.5493404282410767,0.8888958222474702,0.9470339474774225,0.451273982052656,
 0.7381915967510251,0.8110729485619604,0.20037761313280278,0.383580711664812,
 0.5188170851428064,0.9459105356215578,0.3608894171828062,0.20612682631845136,
 0.1931199698221976,0.4194816328504516,0.6228678143609795,0.6170472610501975,
 0.7335892480405021,0.24160402527623526,0.609197734035554,0.13596442816022014,
 0.43187193581059247,0.5329345908098064,0.3253216700443824,0.2877969040946118,
 0.8410987971319723,0.5047708855875325,0.4554724254366579,0.12611122613684123,
 0.07966809194908187,0.05687488348797953,0.6921249742991462,
 0.46681796145983423,0.9380398087660755,0.9285309201058444,0.8375670847518193,
 0.4669745214560943,0.3005565465094092,0.3408909774925709,0.8358952816883963,
 0.5181674446857314,0.8866997214489523,0.33031234042161695,
 0.41456339669214215,0.527325648122544,0.09453538608569612,0.6727671212151308,
 0.07440338178827188,0.412134560232325,0.37013619218454574,
 0.24105238069254498,0.7809258239598785,0.2789716871230157,0.5701840195662395,
 0.7504571676154862,0.1018379833731996,0.8326615989843053,0.24574921185083298,
 0.02261191218268299,0.5651367757132962,0.26698189715741594,
 0.03779439992450595,0.08105180919532085,0.6981172870554105,
 0.7442032030942645,0.8462244623096409,0.5610392707013353,0.18269096883265945,
 0.515718394209011,0.5304742678944461,0.19321429380362565,0.19175432051117403,
 0.7306415996052004,0.7145627439992306,0.6595763145947968,0.5378125475467018,
 0.05727996371145516,0.7617760528524244,0.8406179516489589,0.8828437868198106,
 0.9825948866776741,7.697217633223508e-4,0.20862610872371046,
 0.16214705254088257,0.6801526300147303,0.03714425620192008,
 0.24276413273616293,0.3496696014109507,0.1262472430631434,0.9273233857344612,
 0.26648200939357036,0.7206287500277665,0.3632598398875875,0.4480587742900911,
 0.46844834566163274,0.7923260523655281,0.7364456504982579,0.5186007496133825,
 0.25389814972429137,0.814811592262002,0.47379744521555167,
 0.17063440179103329,0.209505362668982,0.7138844075062829,0.6825161568149841,
 0.5682184990182452,0.0598667432859068,0.8891530850157073,0.07533735895601956,
 0.6106236870354992,0.8136091033603815,0.39111798491433114,
 0.44798727035363806,0.8774337803658506,0.6944959185349263,
 0.11498271025133877,0.31301123223400307,0.3692715299422391,
 0.4854903078624617,0.7792368073750991,0.6991512728256267,0.8473305534651971,
 0.6549127226292463,0.5790156175329364,0.8077127411651974,0.5915732686057018,
 0.9564568727021545,0.5480001154225557,0.22073891833086556,0.5343991519277835,
 0.24449582296925576,0.2642696880459653,0.24522840341975272,
 0.3133057203659285,0.24640275092861696,0.07766922359360784,
 0.16021669994594945,0.4761451238285168,0.054344575246274784,
 0.25629442683769077,0.8502188411190411,0.7482805270371943,0.2826021679710258,
 0.19187726558779328,0.47328251289213175,0.2999346979623785,
 0.2942145968584733,0.9429664598489399,0.20331880787503653,0.7692884392971112,
 0.36715329939753827,0.5379553576805849,0.36979612741449264,
 0.3688719385592406,0.9296315754478154,0.4215434143853529,0.38340595924664633,
 0.4182040615397695,0.8871181460111487,0.10206259344697077,
 0.02031387731516343,0.03587050132187919,0.7927140997193934,
 0.8548047415605029,0.6119921187449087,0.47614553168814555,
 0.18496141740744887,0.6720025285740133,0.7802219635160779,0.7008542229557786,
 0.8505345213739678,0.2891913656448485,0.4115938092478528,0.20152145876260008,
 0.9656765187590068,0.5216678216310283,0.630512221997943,0.3764746329056762,
 0.837286330495637,0.7780413586556858,0.38890857716081917,0.23607236550532396,
 0.08987839477587833,0.2248759979403182,0.5129008359511744,0.3875298999461819,
 0.4408102512327794,0.9563687178938671,0.11790363912577484,0.4504905576066699,
 0.6661184532723574,0.052602843888985396,0.6575750040285543,
 0.4422611625947721,0.0876901503105657,0.5498987289462045,0.5959785352409849,
 0.7520465492482653,0.11612622190280131,0.11645502650984452,
 0.9580755470637143,0.9425297073554675,0.9021700104630751,0.6722923925397579,
 0.16307081081647934,0.6053090933290747,0.0034918432182539094,
 0.4326416210078323,0.6212434989928439,0.6490545740140399,0.7790368438933949,
 0.47411421918616836,0.7948104696939398,0.008650107688020658,
 0.4700021614969906,0.17686057865621996,0.7017556767886316,0.5618662062313478,
 0.5783873937529913,0.36424055260277555,0.5163779384510788,0.3781158920728881,
 0.5415121487242944,0.858280255520523,0.1853603568848654,0.7853036414529627,
 0.1748548043653988,0.5023043313423481,0.5788935338092911,0.4535317088122204,
 0.967898072376382,0.9674463614734198,0.6967828355021028,0.7489475511011082,
 0.45627320678221006,0.30123145251025896,0.32238326736988254,0.296501925986,
 0.9194214329009648,0.7108383840576422,0.44219494641157286,0.5941875945080193,
 0.19255780015890822,0.9078833701882911,0.393019602498097,0.4024045251122874,
 0.2652606958704805,0.18517461167324845,0.08069969708256208,
 0.9752491547975911,0.1500729041694222,0.1137864730890028,0.8245421590128683,
 0.9791618526761253,0.5054820723662199,0.46988260948705507,
 0.32748368749761336,0.1484396290968406,0.6951040593366717,
 0.27833245615769475,0.6820279651104907,0.6833657858352284,0.5027503320382917,
 0.26815671054219226,0.7372639539843049,0.8726649224715235,
 0.07982352554956107,0.741833608494129,0.6208021634170107,0.43860441162834096,
 0.3332321031671259,0.4000158349432682,0.8803454813162709,0.608285450536086,
 0.12698287765584804,0.9466829877079779,0.9044090977087498,0.5399656234941904,
 0.4563985128219068,0.998334690672825,0.8432747741544868,0.10223092638812525,
 0.6314314279070253,0.6400765374477891,0.12244916175101117,0.7945437878706985,
 0.6810833556933793,0.6239372385552254,0.5369894474319915,0.5690921299056111,
 0.5798699999892393,0.951890515868338,0.2843878502451247,0.34104663339755115,
 0.3757804821251407,0.2059337554719184,0.047913003219734795,
 0.5798925896751176,0.8444676661241487,0.3994841614625977,0.9732122904537598,
 0.24546241728173723,0.6827683468050562,0.1530309032736611,0.9734024870729799,
 0.13873856203628732,0.7602501065267966,0.9627086342106196,
 0.01985951100980954,0.2551581766196023,0.6618231537164485,0.7699542154299958,
 0.7882235941044274,0.49137511439159376,0.19662594442590586,
 0.9841000638263491,0.18722138445285008,0.40661205492889807,
 0.48005064968467004,0.23564120526962262,0.07559523462794093,
 0.7827555724590904,0.4850919351266234,0.06296054494227421,
 0.20838895556953296,0.7841642380769468,0.43127524779750237,
 0.12569039504164814,0.11975163227058905,0.46518635650423523,
 0.22001491357403813,0.6414997208711146,0.8476532354195762,0.9519050284349787,
 0.4819860867355841,0.8724906047454466,0.08867580934762265,0.8044948724254388,
 0.3083120718665112,0.4983101021124172,0.4718480151058635,0.6704245872956001,
 0.23586536562566152,0.03434536404994537,0.27317177771918055,
 0.39151875891538923,0.5054017423463129,0.0022856768389405424,
 0.9309776764743489,0.332937570095834,0.8112984906688049,0.7923997352214516,
 0.1725209540256346,0.02322136641146766,0.740944929593897,0.8578146794774161,
 0.5987439417809061,0.711298973821161,0.4491180684078626,0.8965236505606347,
 0.28941449381592577,0.6595645266845483,0.5846214036074785,0.9984747275997221,
 0.11403115037032208,0.8247248111435794,0.7047812897311349,
 0.15452142728310458,0.5413868346935828,0.380620606969158,0.45954607228986055,
 0.899080554824784,0.939289796493915,0.23287129332558254,0.2276667801676534,
 0.1761431689806675,0.928288862913394,0.9182914714063948,0.8080639885837506,
 0.6191082762874045,0.0026583250813362325,0.8461365209454808,
 0.15026665351739732,0.10026471177733365,0.2622483799683347,
 0.7154347571877795,0.23436449445766627,0.1761855412642751,
 0.29912054726295845,0.36869174681807204,0.5308284041999426,
 0.2887558207270049,0.06669084683603776,0.7186189911250113,0.9853550189759469,
 0.5597158641142506,0.2839551170143172,0.4004793067059492,0.05789923668225272,
 0.5984514391388295,0.25802874780812113,0.01203341316338502,
 0.9344408675123559,0.4846699221723856,0.010942527884195696,
 0.3426322638528998,0.7230789397270366,0.6892830445937801,0.6921560244145266,
 0.7900267234008005,0.5296405352813885,0.7988735458669889,0.9332262275643028,
 0.13657106043867384,0.19762061209738657,0.2010212436307517,
 0.8798855233852672,0.15885903397867174,0.36929885057154765,
 0.019934837455264254,0.39964997950872383,0.4438599716304632,
 0.30529897953220086,0.7027307712643833,0.9849569531296839,0.7081818412633554,
 0.823642670938941,0.8493581147283675,0.03337002841461789,0.5188416899395975,
 0.5170943382257644,0.8760239976043439,0.18494868184827318,0.5144764824132432,
 0.40206243515885864,0.6884219659354553,0.33950968015125205,
 0.09432129366150577,0.6030195808224563,0.34859045716208237,
 0.7462651646352907,0.06585138984719274,0.523927549207843,0.2801397450977353,
 0.36460791433684947,0.36248150934243095,0.0775561979032604,
 0.08942757053112538,0.6829974599264562,0.9966246457300107,0.7896152828250269,
 0.2011797271102277,0.3648213061404587,0.3584060007367321,0.9021974884201842,
 0.89025658898161,0.8389397536491654,0.664093705335196,0.6693124901429669,
 0.37731340651028733,0.457281722042178,0.9273168855384398,0.5970397929926211,
 0.7297722586194991,0.3921559331501463,0.15367737486305255,0.8004801543677097,
 0.3270538987152013,0.5568568884641927,0.6815050841790138,0.08289063987959222,
 0.5757812923758131,0.538962438385011,0.08369783182504065,0.12517989289392295,
 0.2321100199695083,0.3505721275139855,0.09061741735836448,
 0.15657902852986694,0.009613468536498804,0.7863917717605866,
 0.9455761710735361,0.40614440613080605,0.7227260950626855,0.846945482655266,
 0.9433148693895795,0.7087648177715745,0.793047483879001,0.6687028415001597,
 0.3813276569055457,0.026824732163057807,0.5538612424656064,
 0.6913766154345355,0.05347904044293139,0.9104457755356719,0.6442492779799469,
 0.46359182870014415,0.11098967482856004,0.3443391232574161,
 0.9469787273659186,0.2620414650476395,0.5394807246705017,0.3768708961827617,
 0.2753555790733453,0.422531830622892,0.15739012835771315,0.6802646047539753,
 0.4332106101301324,0.8341816112194516,0.4846413304499053,0.9621479956201752,
 0.1128388583622163,0.7692569749252065,0.7249655247478066,0.772402485652236,
 0.8588722911631084,0.6701460535508986,0.06990219708437506,
 0.29742472552798893,0.9679015054975629,0.4554602381686178,
 0.17671051312049035,0.626902072996666,0.19548749779367236,0.5862120473806842,
 0.4071115445009442,0.5018536899427201,0.9805929970949827,0.17383850333582418,
 0.7487006508377592,0.8636176467608937,0.06701369224668585,0.5249258919029605,
 0.6434887983994382,0.6366042509939579,0.18765353352031533,0.9580106340509958,
 0.9747337497460657,0.6605320613069441,0.2571184622800592,0.30780737216159204,
 0.4288770638243531,0.17728140408921178,0.47253273075705815,
 0.4334265667180561,0.45959202006230226,0.35661130001024255,
 0.07058896713754192,0.9708949403184812,0.4458572463048419,
 0.49965372175381473,0.4908731302587248,0.8879886335370641,
 0.30647203537833356,0.2971658511236781,0.12535065816654667,
 0.9319367498420086,0.12292280277464274,0.4983612267297467,0.7072884840353936,
 0.13999948436064913,0.8922379163822133,0.5894830490426466,
 0.29877550241811834,0.9567545119749669,0.6387419104798823,0.720016205124468,
 0.8208583353271597,0.7034561210230578,0.420029896881564,0.127685982024493,
 0.41523937214362394,0.4093457260196893,0.7147162207150192,0.9799004714525275,
 0.9065033916456404,0.36599991626180073,0.2629374782602386,0.3434422841037297,
 0.7373068490921596,0.21547406724533702,0.0287386751180565,0.5240451422001737,
 0.2780964410212423,0.02815261253441592,0.41871763218378466,
 0.6151623490897291,0.8985485195580629,0.6443465016848989,0.31401867964097996,
 0.2468616340769313,0.17450191614077304,0.9676993035216577,0.3003568201478509,
 0.9090440929594885,0.17366970600540232,0.6860257970889085,0.7837929935251698,
 0.5606812459370907,0.30187792206465025,0.14002849862356848,0.674116882589212,
 0.7774959120474269,0.9799392065086276,0.20168167988509067,
 0.30152359753283264,0.5487218471372312,0.11808361835615266,
 0.24974830160540318,0.9668560633820132,0.6158067405698391,0.9660261237678835,
 0.8763521065122326,0.2694656438139724,0.5546074946354811,0.5262384529695169,
 0.41845599602777717,0.29012417750312425,0.13519688434140242,
 0.5389642814459098,0.6894204199124403,0.04715246347119251,
 0.38458727156615125,0.18762492457712243,0.23631025977636289,
 0.6902603019707081,0.660835346202121,0.42708705960486015,0.5648045703948412,
 0.9592298642492412,0.938931157791306,0.5113054313201791,0.09894145346776662,
 0.4727615875937047,0.6729502359763972,0.09584691661636835,0.5727884617351717,
 0.542504235348539,0.2657442494880857,0.6139231415198829,0.5481188950254962,
 0.4074126626000877,0.1885648825039673,0.051598551083499755,
 0.5783529137356337,0.5374248487089844,0.2016875272530535,0.2525130305792387,
 0.4635185611460674,0.6849094782315266,0.4258299197382235,0.47331408495752103,
 0.022545183686731862,0.7550498405367245,0.4733061085236948,
 0.7246196927510129,0.4020166164362122,0.784629143700615,0.8728581713203297,
 0.83271694011569,0.9547164015133952,0.8826505777606504,0.15757094397850224,
 0.17329500823130817,0.5910938650129771,0.9865799897846232,0.6830484338453354,
 0.4923769883034862,0.08181762815864213,0.2786596196101959,0.6031656186746391,
 0.6942659908038583,0.9768427421503149,0.5586667950016504,0.22057020323496523,
 0.8979803286370501,0.5829266804871654,0.49608274585403456,0.5879541173891439,
 0.13922830877454895,0.2717040707951931,0.7496080451154437,0.5029853269926782,
 0.8250242615966357,0.7663696574717979,0.42077874249661795,0.8532265791544313,
 0.22427012932084978,0.03688278095113873,0.4574918692507791,
 0.6704850481427808,0.9246671758991608,0.9715344453420655,0.40549215522495907,
 0.008499262646673111,0.5705623330607512,0.5772567610265443,0.373696199445714,
 0.691490032530481,0.2365511692735811,0.7168724981511037,0.5905282431558843,
 0.9161050174353679,0.3590092447518841,0.9511599896769238,0.41750916783780634,
 0.8291627935971708,0.6998891379546449,0.6743709174829201,0.37273292763577537,
 0.11497103313250567,0.047635978890442354,0.6459984548162025,
 0.12706815086666146,0.5232251401773205,0.6901715551174183,0.5664922928070246,
 0.18806350303065533,0.34888971738829944,0.4934441371461429,
 0.08459491389000773,0.2931865364550871,0.11580970020666825,
 0.2626093655739391,0.42934773129897286,0.3938152411789685,0.8093987800734366,
 0.385116144088234,0.0013920937520909282,0.4599858292643826,
 0.1526013288066893,0.026176781685748107,0.6135230234662482,
 0.5683261821135057,0.8201571017056963,0.8238376880493067,0.5158209934222706,
 0.033568538237014556,0.6813547757266839,0.5424891292372809,0.68818040887096,
 0.9433304235132276,0.7936673685969838,0.16519496971690106,0.8841324964901478,
 0.7265804682469326,0.5756601599230078,0.46397600782981985,0.5760259259140297,
 0.8639319636286552,0.7067452198330252,0.3446742410426037,0.4484429595177184,
 0.8271293811545866,0.7760603641399488,0.8156296720727235,0.8644157826989046,
 0.7186483948642427,0.24161703299152948,0.7235825378538582,0.6774375471482874,
 0.757157953444493,0.3873084469987014,0.44684536502700956,0.9835811751057727,
 0.6109052986262918,0.4351694556460337,0.8655445937671298,0.15610388923768048,
 0.20319595895856213,0.18807115075844005,0.10236651547804831,
 0.523148632901354,0.07108047093578818,0.059294475277893555,
 0.2782769514687994,0.036008011292467224,0.8915766610203825,
 0.04724735280920456,0.8998985772112111,0.43187488561603626,
 0.24978188400122958,0.7863788398018627,0.6766914562107161,
 0.33935055509637857,0.40438656697329367,0.8129094617646547,
 0.23296305949124163,0.7465739817743422,0.1959382409202749,0.8688278528539034,
 0.4277323312994441,0.37792173710314236,0.6993450436985956,0.9223033589779369,
 0.534185228843725,0.6592901715847138,0.6944802250027755,0.6393513815713056,
 0.9392795130418046,0.8460589762653876,0.3152878950954048,0.3123095839161565,
 0.6411953504659174,0.6044734565232085,0.30157624875925004,0.967672978864502,
 0.3668773975846501,0.5375769680218061,0.6972325098405662,0.41146792459662496,
 0.8436742842252342,0.07143550075560712,0.7904462769736648,
 0.10116931060723122,0.8503093613621532,0.2579495064995607,0.1789809331832123,
 0.654162500529382,0.6128666061596176,0.7143545846812362,0.6895018712366051,
 0.1288152289819059,0.010209283398600943,0.0870895056303923,
 0.6158018307198377,0.08217252383295659,0.7308965855374588,0.8536007790921295,
 0.0507209561998897,0.774881728788084,0.05164780492117038,0.9866140538648857,
 0.8029878674256945,0.34119472748534285,0.6335643747501122,
 0.43927617986568446,0.18696752669470995,0.04795111874116331,
 0.4970275535208998,0.13577360979862074,0.5839008553757652,0.325367346644996,
 0.06950408520133455,0.48008075278825846,0.09297922893253197,
 0.01257303288499223,0.7270999395998121,0.9564895033897869,0.5455230694168508,
 0.6315980672787719,0.64532557302115,0.22216356444350982,0.5584385330014008,
 0.5480535178204559,0.35679059504375155,0.2990890657480749,
 0.13396845620332642,0.7299009178482547,0.8437223075020329,
 0.32323020629098875,0.1423059854497516,0.19114069656823018,
 0.13648280323011264,0.01427558773018145,0.08106598149162103,
 0.669821142601192,0.0794313868616433,0.09850919466980268,0.09934248224864994,
 0.17561981643754776,0.8841588901379559,0.3260470308830581,
 0.13248463286383938,0.6480693672677404,0.17766592301582596,
 0.03251485133836729,0.17907567138253622,0.049675606650450344,
 0.05234270974290689,0.8212410623530577,0.2530173957863113,
 0.11459645833811394,0.40203389647595267,0.7850922255095235,
 0.5740583297193285,0.9612631663292618,0.28484041480507427,0.8029556092691977,
 0.8026225872955763,0.9485211358514145,0.7141586634712346,0.06900414428612867,
 0.14394826475904887,0.9542337506413733,0.2997893060665595,0.9150770345237043,
 0.5467069583902404,0.10509097848884219,0.33073462293875433,
 0.8033357155035252,0.6222990628693377,0.21008167755373508,0.9793838817642556,
 0.8034959838221781,0.7014131114778638,0.7013092164001394,0.7765312385962907,
 0.7070822573142311,0.7786296551758011,0.7788662946953451,0.840777304434596,
 0.07863162639610877,0.7929046708713878,0.067522646220319,0.5753882251572011,
 0.671164624655184,0.7963723602671783,0.5604563156662085,0.257979033985705,
 0.7725999983431863,0.9516734003623015,0.029621239686109302,
 0.2069435936030396,0.6515055827569854,0.14686813868837523,0.906074291303909,
 0.8840542530637802,0.013529513174704966,0.3408893780938641,0.651106428667316,
 0.7862541888235659,0.28237049791511626,0.5631453457739173,0.5571654518614803,
 0.9684010607003424,0.1715538009460399,0.529557036368546,0.40873655524669583,
 0.5587581218691597,0.6361732151976055,0.18170163380853321,
 0.46278351608240886,0.19159844000676196,0.27248089020891986,
 0.04248820706090406,0.8996743672416633,0.8311197220367044,0.7110179979623861,
 0.10458868823367706,0.9635978925334672,0.675943605286748,0.3910600797405953,
 0.1613362054828983,0.46178565387982395,0.8768064149619704,
 0.017705272294672558,0.4167687884246558,0.9366469013803628,
 0.23420950414921637,0.7893242541459973,0.9591773058591226,0.9736958775459501,
 0.870188617773988,0.9415491142800749,0.5846914682859738,0.2713959448517167,
 0.864442553430367,0.04428689049420198,0.22861791078088745,0.7366902781386121,
 0.8791009052583291,0.7959296242058892,0.25126225089319587,0.3448222370744888,
 0.6411556885841728,0.9351968471718302,0.04860500428989167,
 0.008908642475210726,0.22589696862991782,0.8458341290427869,
 0.8679129989791774,0.7713197633170933,0.8538782086807937,0.9525584756379692,
 0.20229574733321376,0.3498364069608857,0.5076741833341336,0.659065729118457,
 0.19495721272898847,0.6488678907314148,0.7120218064713586,0.5364834902076058,
 0.9402305457843662,0.6176330803526058,0.2967893728552643,0.37152334541733334,
 0.8790118836594993,0.3402442427192168,0.4626546270864239,0.32483458802508003,
 0.4603112403674684,0.05201041942703699,0.7996839296176947,0.6996838914014135,
 0.8730029919457891,0.7909110662986123,0.398798953372445,0.32924076688770143,
 0.15085783756316484,0.5460706209979829,0.4701288774532767,0.9570640351808779,
 0.16606112583753008,0.23458197263308578,0.1480462157815119,
 0.4624855013666611,0.4452250683070873,0.5350612550679574,0.594545887363615,
 0.8525338402292816,0.32792601080810435,0.335872761875712,0.22326079409137112,
 0.7732579084260877,0.8997769093334343,0.6327456849172114,0.674505764415111,
 0.27247435781756657,0.9777247348517747,0.34287549011015794,
 0.13201058460746629,0.3644125239209671,0.3824837118942024,0.9223401446754591,
 0.4254528327725543,0.7666528969289357,0.5976208119451506,0.9962426373455409,
 0.02458581955750405,0.5975273983392456,0.6015906707044281,0.7201945679034332,
 0.17392530462033062,0.9128105744890745,0.09187523172595558,
 0.7863563495892302,0.0739947933862708,0.16897696269386198,
 0.37267228354877324,0.9602057016703335,0.5078536157273184,0.702199442168787,
 0.4562230901467603,0.782751810872263,0.77199783458631,0.12043177484448286,
 0.5481734238328309,0.7254731245914634,0.7969871478947659,0.7328923095586465,
 0.49002390384016736,0.5667992978446377,0.04121865855718998,
 0.9881996960291062,0.45892765041649386,0.17081604045984533,
 0.6298535979553339,0.24884026649072033,0.02536856218389727,
 0.9035243418583832,0.5564543584373167,0.5483796944577138,0.43073238800399927,
 0.2999510910199952,0.5521858782380403,0.03730036647954449,0.8975458993530737,
 0.7503217344854464,0.03605653947967302,0.8560998866207733,0.6420616937536213,
 0.5806855277568362,0.9434501692868923,0.15964596446780688,
 0.22098727395923812,0.7230227785443871,0.15244063153442722,
 0.8141618504920264,0.5978775218482708,0.13813616968846776,0.9424124107200377,
 0.8582835200501235,0.4133083211505203,0.2856705797859941,0.14854360364599384,
 0.1519187475009638,0.6117470997236545,0.11255054525064345,
 0.20828384350063478,0.936692108323769,0.7992490573225807,0.39783725800879977,
 0.15514381081163264,0.08784804529037338,0.6896065028430707,
 0.45017341843027336,0.9899556771707604,0.10650282241482212,
 0.5974043918698659,0.7838496028130248,0.8247357460373208,0.03981353751028571,
 0.3138565914059698,0.7930762334270696,0.9481445059824333,0.4922027614723121,
 0.3803572601819458,0.7283571863264441,0.8957266132132141,0.855463598357549,
 0.010158045506842961,0.31451458988778735,0.09354306181602867,
 0.7136919519091801,0.170467681186801,0.7701449118022008,0.3181536235374586,
 0.5872410170174158,0.8389481210343791,0.14179504290351486,0.901277196163967,
 0.6196959809193807,0.9267490413986559,0.6707865603118656,0.6730349758374735,
 0.3003753726752417,0.030272151205165687,0.03172971895687626,
 0.9424880226482335,0.8457045560554675,0.31507070055382724,0.6764233337801568,
 0.41561758657997583,0.4512733147835515,0.7420047370063181,0.3178742277667904,
 0.6101646077112944,0.8640802655106049,0.23769468675702732,0.7967508141181738,
 0.02360208389736318,0.8532042830339955,0.3597835999500606,0.505475699896205,
 0.5667750440334306,0.29339943919235667,0.8322525529402193,0.225936945939467,
 0.4502917044397946,0.599527234329792,0.012681414630728804,
 0.26464456969225014,0.1447981657631272,0.5655937629126232,0.6913102850798735,
 0.8415781549944475,0.9623613391206923,0.9410486785778298,0.3811765715331251,
 0.00891633729792729,0.767204460038001,0.19688841655109512,0.8332442287637227,
 0.16038368024746497,0.6621750878083021,0.788119846949426,0.4067924790391333,
 0.3616411955654234,0.5645697460510014,0.7721467891807684,0.5212935892966973,
 0.7801306661823203,0.13812634606034258,0.3868016314599745,0.5486671773323089,
 0.46336722447051715,0.801185079234243,0.7515874115741694,0.17042148972331617,
 0.9388216720801718,0.10193514303155882,0.8406139041318064,0.4678422277410723,
 0.03903108578538228,0.2641437782353727,0.04471724027726698,
 0.3159367460243985,0.5402681959314067,0.9662910731164521,0.11049724028366947,
 0.4225438283173103,0.6653623744851022,0.55430099345612,0.923573888202702,
 0.0898101525664633,0.11578213566919171,0.7259942329844101,
 0.03126806240690683,0.3865930316581665,0.4367877756753622,0.622728209263643,
 0.29073101529512535,0.71632847564972,0.24051805465447773,0.8311330510260627,
 0.4708518954132058,0.012953619302490282,0.5957744698425396,
 0.4247982982312217,0.8526701766628265,0.5291782989785644,0.8208807922764618,
 0.43792761805559877,0.3396596480566474,0.3434970964779197,
 0.061182928836314376,0.7427786516032543,0.22646857171320178,
 0.14866612908063337,0.5985353456359579,0.9850634499171969,
 0.09287423323406852,0.9308193680625229,0.45815925685490355,
 0.2726847789232626,0.6087676818122183,0.10592007812523052,0.6215787718455132,
 0.7460101532241957,0.5284187148600967,0.8194841390029339,0.0253096826367486,
 0.48516443859077085,0.19246267583386678,0.023324234750747852,
 0.10596094053373051,0.5683409453623844,0.7116639740784849,
 0.29535026935348896,0.7400009429537153,0.4331270280178774,0.9424719177538732,
 0.3441836259478025,0.10392988455962482,0.24134551528849912,
 0.32809266105174817,0.9710387828695498,0.4393464620985412,0.7863566445169075,
 0.9792415204372127,0.09453490807061926,0.13148955615603286,
 0.03015213880745038,0.37705859711293555,0.5834490097064592,
 0.012518266841128556,0.6749357346564581,0.7970158723303896,
 0.8417649507198288,0.20691926643680747,0.38128067643546326,
 0.5922470733295772,0.9733270683630839,0.12675572013430303,0.5627322755702849,
 0.8851422153651244,0.39782344677582826,0.6047152462204398,0.5732777787347777,
 0.9659369404454805,0.07926713677420172,0.67396161334966,0.8844904345951743,
 0.4365032934405104,0.40990768455375115,0.2908910309647653,0.8081796583803885,
 0.0683064269129613,0.9019400983513185,0.6280187782762043,0.26472748909523525,
 0.6969775139259664,0.5800254493082473,0.16140624241330137,
 0.12244121084757731,0.7253356438561542,0.3441500628850991,
 0.48500508730317515,0.4343888970067935,0.3117810922561045,0.5241017114513028,
 0.12372483044762461,0.12118611009306934,0.302347591517385,0.7351244175640378,
 0.6940646253997187,0.24253877519494016,0.8149984894369339,
 0.39605037894325024,0.23197124462303542,0.3674144304897434,
 0.7350906980251031,0.6628812476904352,0.9758749234588218,0.8876105589977614,
 0.4253736974937169,0.33976229210491593,0.13162150052689858,
 0.7176374693583076,0.04061419268188948,0.428659936262358,0.6662171173062528,
 0.12550594869393694,0.5739382200676231,0.1300356529404011,0.7809561730050664,
 0.7258216249830111,0.15353624867427962,0.265208323486116,0.734678428104052,
 0.21951039583643084,0.6424444615176456,0.7846197652403547,0.7108443346200636,
 0.9626389673017811,0.199181240483582,0.33418958784789776,0.05704209267186344,
 0.4007455612308931,0.26864655009828853,0.007771383263014742,
 0.5580309939619406,0.5012020227992878,0.856235417539457,0.018845265200560846,
 0.6338404954420029,0.7164845510371056,0.49784080177141243,0.254805436572603,
 0.7383270700852947,0.17759638963857083,0.3155625799064161,0.598104724647357,
 0.13729441169142287,0.9587061816200844,0.9270677134305338,0.7991635551305052,
 0.9933706684317942,0.810934158646728,0.7403779384820612,0.7090408643302806,
 0.42336668008696166,0.636131733818001,0.8662350445745475,0.04352797144554699,
 0.9721587490236956,0.7014968112851878,0.5842119672641406,0.5704451245018849,
 0.15522278856784189,0.8371311484800799,0.748961456804519,0.47586974706903573,
 0.5610891720724551,0.5122466515621764,0.8342617355780321,0.32201130093037444,
 0.4101263324646628,0.18337823828644473,0.22652968540187446,
 0.9851540082302809,0.5288738258422453,0.1593936344538317,0.5183827158871201,
 0.3834729551342366,0.11262373337936837,0.6367218729431773,0.3398550122439804,
 0.3254367727826525,0.08891504826872787,0.4755106657830379,0.4857699349548129,
 0.0782843970444439,0.5687443134274467,0.6323267431803458,0.794528312893053,
 0.7716078945784337,0.46615928855324873,0.9662982050960518,0.3319304272842889,
 0.40809635523147114,0.8963252565783322,0.30338847283689174,
 0.3789809677886259,0.9551702765417934,0.9890379127727766,0.24547777214123112,
 0.3702095478945284,0.8852582707430167,0.37770453085299605,0.996868844871493,
 0.6251146845949482,0.25126051412964945,0.6455611612100488,0.8596876619065238,
 0.5307745017235547,0.9156576182019901,0.8349046830583742,0.5593605937711253,
 0.7405031107368736,0.35137776666711995,0.12095830257181972,
 0.23626151381135596,0.9075136766258869,0.6056101539794901,0.5303800795359364,
 0.011810798455664462,0.31896187925769004,0.4713745248890976,
 0.342148005074279,0.31601721198312516,0.8556113580546982,0.8678006182392701,
 0.06664022658653834,0.36136908437860316,0.5866891857986737,
 0.6847458773116251,0.1424148755198722,0.14952732306722116,0.9004203773703161,
 0.4130936282959994,0.6269478433501251,0.9389698955802022,0.7114703807294424,
 0.6731455960210748,0.8912340525972009,0.8772559741919874,0.17149324395784948,
 0.03240714129978006,0.7518758975866242,0.9983889089503046,0.1289807419982989,
 0.811158972023576,0.3786922225121675,0.13516456036292412,0.0460292345201736,
 0.8549894358183554,0.9293880204021098,0.4712906377676628,0.06379742323555104,
 0.7559762366812428,0.0657021537093746,0.950618793858622,0.6079206550661376,
 0.6459888761756905,0.10374774467671655,0.04646692101696126,
 0.6035010588596271,0.06321174429101084,0.7991600655961039,
 0.42506404925972374,0.07853549953594008,0.47330436381532126,
 0.6910710474338223,0.3097385986378409,0.6595343764718054,0.3053284104067082,
 0.9531485324785405,0.08162453345259069,0.7084384128086296,0.9873797213090684,
 0.5924979959506219,0.1978207161910076,0.30361914266743406,0.193884027196639,
 0.39620221493682717,0.5953398047611228,0.05214761043130811,
 0.2007655387389532,0.5370024617196396,0.6137605801631495,
 0.053508581543346256,0.10069603950331452,0.7936572261496456,
 0.30861066979636587,0.7928309396863251,0.23237375739093458,
 0.9479619685843319,0.5541487547929574,0.5374061210773607,0.3442759870068206,
 0.4828414493029368,0.703268956524135,0.9095314077842304,0.412294302781294,
 0.47701364010165126,0.20650736853917462,0.943280394286509,
 0.061679446647427305,0.4424202118784608,0.48664902863655535,
 0.9192515554188578,0.6182405800896179,0.21929111768363974,
 0.27525301512490163,0.6829495501020711,0.7602185142886195,0.6138034791471413,
 0.06499051819245572,0.3390990517426473,0.7043262564967716,
 0.023864611352905296,0.8000795284379564,0.8496245316090106,
 0.3619831649209019,0.9025377218589176,0.8384669803168626,0.0332947332924034,
 0.9229586427390366,0.5210349803399177,0.33732720436408614,0.6905850006233686,
 0.3799807195207321,0.29467987649302274,0.9100543502740197,0.556260866886252,
 0.9434550657961837,0.932324533845897,0.6851556267233095,0.11499396283189633,
 0.3894081817794881,0.2785224843414197,0.7183073533879247,0.2671978582265404,
 0.49387265285367543,0.4854526334025032,0.6383285633221808,
 0.043259328550625975,0.15907199363634716,0.6472059221214053,
 0.7163344756155186,0.5499127040855187,0.2776035863626144,0.874436303548741,
 0.4732076185087769,0.7198581614359487,0.16465615440971826,0.8441679225580152,
 0.36510566325877714,0.6907959615008726,0.04188021975016254,
 0.7471839368416467,0.4000708828547034,0.7901099453778563,0.5599701204981077,
 0.25961644238233506,0.7480590277660383,0.7883756992844897,0.6598583152128632,
 0.7261523312924953,0.4109815379945567,0.21785919329046333,0.5282351686040098,
 0.42111667597576796,0.010344356781880726,0.4951037731122234,
 0.6724843647880017,0.772299519161874,0.6826191988312076,0.7497030766947451,
 0.7130987157941434,0.953121591323183,0.789648619766861,0.3915974617803206,
 0.607086224346876,0.9343199665771452,0.9124160425970723,0.4869345272994843,
 0.8689787557921818,0.5958367434788024,0.22990976934966456,
 0.10031878354992474,0.20430158200733162,0.9816685275005981,
 0.5371462272354668,0.08173422695899446,0.2296195809760725,0.7505054993125679,
 0.3956369990378703,0.06518068596216842,0.9378292847259387,0.3588904287278559,
 0.22825050616258324,0.6314690411391941,0.3975719671512944,
 0.29607592020784046,0.721828899733204,0.7370093787223375,0.22790732913907508,
 0.07501584391533334,0.36222135900977837,0.971471172856648,0.5112690085291812,
 0.3626212452982126,0.5110408618411163,0.41426543932503046,0.7645675540839837,
 0.5216624507271861,0.006215280067565332,0.9020107326079907,
 0.4171651468901555,0.27075264456468984,0.9419680930819918,
 0.35271939194248947,0.6616614499303499,0.03286295774539383,
 0.3918068560451874,0.7703400893173051,0.559362018241703,0.43997452612112,
 0.025943984144198073,0.3971756079959793,0.9630501397298744,
 0.11960212138437398,0.9745849778188772,0.6982514510640981,0.824946461834872,
 0.5005287520305445,0.6175618045021601,0.5806284501521852,0.702472176486437,
 0.45355168259322265,0.1928282959343841,0.14227336870536522,
 0.019354770635853003,0.7113824096880805,0.5158384059304679,
 0.6705756848346652,0.9517968243583859,0.52379583578042,0.4821483912941599,
 0.8088823765203871,0.5329006331839299,0.259787336776261,0.5051006761179394,
 0.27288327432934356,0.5128107278912243,0.41492508869184563,
 0.2752622122600755,0.8370701837775626,0.20194376507327993,
 0.05045410177465515,0.26464606926841194,0.5437060759227373,
 0.024870895311203456,0.3595381210587567,0.021439026976710407,
 0.16187240139068815,0.0048560167332956095,0.6750520426100706,
 0.3880594766260166,0.3641818584811486,0.8907460800453565,0.9614745914424135,
 0.24163890631179674,0.9474744974626155,0.13267914112993096,
 0.2743361480487261,0.6183121174281275,0.06255222073724642,0.6373969816767037,
 0.5876390959997313,0.41064109641313784,0.37735297305744886,
 0.8912287208267455,0.08785611815448457,0.07208111345321688,
 0.3746593970360038,0.8070370338995762,0.9796093798367757,0.9328291718618336,
 0.9798084820708985,0.27004608369691496,0.09801100600607415,
 0.07512042831016164,0.3865213437017463,0.11238368741774674,
 0.3978044281742511,0.52607009141179,0.8648952057429514,0.014347279002837277,
 0.16360435411422047,0.24442614272614926,0.5637963311565075,
 0.8181829438783818,0.30508756963031347,0.29071468169664105,
 0.6138797347865274,0.6132607414173403,0.6957820017501446,0.12400981189266802,
 0.7436491333096451,0.12303531636619147,0.4083692296854273,0.9285329275296555,
 0.7186276366882841,0.9046420163773027,0.6197984287096749,0.8144280398378354,
 0.4126338317933813,0.45728561356409425,0.558200212526966,0.42296806381449736,
 0.3975876815265543,0.7683341073968512,0.05724806892167966,
 0.33264743904450045,0.0582895394189471,0.5161469950249781,
 0.25812226535951244,0.5749102941676738,0.8834564446455334,0.9431501369842833,
 0.5887790815313488,0.9426134339211308,0.213940519886326,0.5090215203832381,
 0.734963051099596,0.9852410191085545,0.20372146399378532,0.5939110992960878,
 0.01348615177150192,0.6703158162813578,0.2078724199102393,
 0.48952174396927606,0.060364739781286425,0.5022117116467749,
 0.6566703878287695,0.7080530722759653,0.00770514683171053,0.0883047268132312,
 0.8621750784506109,0.7548741079926723,0.6005712984351738,0.13666839371220163,
 0.7175794497987794,0.9865855704714291,0.5713870402668778,0.07725262191136917,
 0.4820628051434296,0.704142451304002,0.505167339862044,0.015370629485309872,
 0.4447735311532254,0.5021991244230173,0.09285172890473326,0.1394048707559683,
 0.9546959759138943,0.8635574120805931,0.9319721006631432,0.560504300671078,
 0.7882844196986569,0.6803386422431998,0.7115941296999456,0.6966679831949264,
 0.2761332486471816,0.4220556329889299,0.7350038524238994,0.8095419004469849,
 0.6323473212607359,0.40746131480256287,0.770143396646247,0.5853305903889603,
 0.18563710567103542,0.1586472488584456,0.7059016947824193,0.7980965225145877,
 0.03933862294508661,0.1062783330797098,0.45924963066988267,
 0.9695137065110576,0.3618231071797481,0.8450155736189027,0.6080844882834289,
 0.7178123152097612,0.8749880975522979,0.11841315658435703,
 0.035646688231883106,0.5907882886453774,0.3125982573381977,
 0.5320615700538597,0.6068436000850771,0.6795212703259951,0.5140019712530567,
 0.6460330174177368,0.715529649588201,0.6554190334647512,0.38973260162447143,
 0.9133203404002993,0.8778347806642947,0.24207353553389388,0.2566602366329015,
 0.8462376149603357,0.8357676626602036,0.46003285530239146,0.77650608242336,
 0.1953822504919357,0.580580413961195,0.7279509755259975,0.9007870668104863,
 0.7041508449929798,0.25081742952642516,0.004761202698092104,
 0.7282340628205699,0.34810046966356156,0.520331169773315,0.7521815404823502,
 0.7055887143073127,0.8391901669482703,0.745073445480324,0.3534291334347692,
 0.21040281701609664,0.9487911622776102,0.0600470485706015,0.9905469193595132,
 0.4505831845723066,0.3054841900394133,0.19350992179832005,
 0.37046822514699485,0.954952176437979,0.38767702159842643,0.3614164138766496,
 0.3625388141589,0.1153959230424687,0.400602726167138,0.6614358905725551,
 0.27221952914971914,0.6677452780569205,0.6369483774236228,0.4385656210660033,
 0.6372893063458431,0.10601453873974254,0.46085628350882635,
 0.4560370934345477,0.9316347701955686,0.4743138677395403,0.5192294668043724,
 0.8190680376594944,0.5192563574887704,0.2177087450273646,0.27311876894753306,
 0.006577710388738445,0.659559281560207,0.67941201833199,0.7616250885369694,
 0.8030807558767673,0.7081205707104983,0.43507918122351463,0.5240335110507662,
 0.4369903703895951,0.6260122372117123,0.9864836734658292,0.8314518835099836,
 0.46426099528810094,0.9615600262756407,0.6198633476774922,
 0.40895197385010196,0.4903767337824514,0.7913108686176072,0.1278452897201302,
 0.9511792471564346,0.9966079731274571,0.9530013320977776,0.03737481473780757,
 0.10679830330815765,0.35828274477999444,0.7287622760177774,
 0.34837456762027463,0.1534560459579999,0.001822403289407637,
 0.8476519514286346,0.18437522583073263,0.2700042385824789,
 0.49043930763413923,0.3780876853753463,0.9683307521906666,0.8037494911018936,
 0.5102416311849965,0.1204395081620776,0.6591795872579884,0.9620985068894105,
 0.4786561079158481,0.4776818776268549,0.4367039170613075,0.367478075626985,
 0.8886556832794658,0.638727506308473,0.1580612893433656,0.9538437192562732,
 0.7720172509995233,0.17798135170021745,0.4171433385156573,0.9565215414762909,
 0.6762246095073197,0.5126937051244266,0.9622673374960984,0.789284966721926,
 0.8148330973746193,0.5719446403024683,0.5517364939630957,0.28094288986373317,
 0.12692348875617732,0.6281428980003736,0.6860023003652453,0.2322691235291262,
 0.3913795922209462,0.6112666171783164,0.7391433754733896,0.16121175084215789,
 0.5309049777674433,0.3067874045824909,0.538059677337016,0.8430789445348876,
 0.18173286546797074,0.506144224247623,0.40251523564633496,0.267899939346526,
 0.9719738216091077,0.8245376807007597,0.38049689832806666,0.9552211964656117,
 0.40893772044187604,0.3893044309184486,0.2756998984044807,0.9485930089041035,
 0.9981690530087829,0.7871229813285623,0.686037919355404,0.6091176874608429,
 0.7857897953971675,0.9191261444184656,0.5847124805474154,0.5656432583081556,
 0.4283783325164954,0.204535726490933,0.4814536531497293,0.027270203337505072,
 0.4606490599734617,0.05475298487755098,2.1893947734985542e-4,
 0.16616581826359467,0.640373581052161,0.09526192694221725,0.8624925916031163,
 0.8661000115847056,0.8649619303920629,0.05594839409760777,0.7528786889459255,
 0.33513795029426263,0.37649901828601584,0.18142776213791556,
 0.7323610962205415,0.007387515492170649,0.6737385123639008,
 0.8148432721458789,0.5495729699452666,0.7684171117143038,0.17515098671180318,
 0.7102961510720281,0.7072814965061238,0.9290539178810692,
 0.036880049578250085,0.7721486343659364,0.62320756207465,0.8088724829540594,
 0.9479594850196413,0.6662563476031318,0.47998916802216984,0.54841279803085,
 0.35194641595097254,0.2683947493994897,0.5443674071635478,0.7452297979406275,
 0.5356600329798711,0.5954387107162297,0.33432231064289164,0.9555672119224184,
 0.5794795394867152,0.34608767672682195,0.7838758849535159,0.8333609928397564,
 0.06708520972897758,0.21781990268863083,0.31147116642109507,
 0.015029720783577116,0.8322453701872177,0.1937170662447837,
 0.25811016612208504,0.2657800841365042,0.7860681116772184,0.9304165233829631,
 0.17220965149526113,0.2150927228058881,0.7311250941443526,0.6814232611824627,
 4.772575193069395e-4,0.6447930552835033,0.9716512237954915,
 0.7962946034135973,0.24281913188569582,0.8664290598975861,0.8813051241512087,
 0.8227580852159417,0.39059317887607703,0.1825555457847905,
 0.011446838866138287,0.6756443977137032,0.18015522158965325,
 0.06713560603833901,0.48991890784614167,0.45622202827735836,
 0.6178358587865663,0.8385735485880669,0.9730207959315598,0.5167022290073502,
 0.8490531411499541,0.4356916820554564,0.07018893052780772,0.4409306355928245,
 0.5976237225631167,0.4041277487621132,0.5736173153593631,0.9540255681482442,
 0.4052114244438956,0.23314419950605725,0.17128946782500365,
 0.4010720461741826,0.0874207308039916,0.10797412415649565,
 0.40514497307536124,0.09371903332384979,0.4313635105932516,
 0.8453860248937449,0.6535762302734159,0.5687196369720707,0.49687663748060884,
 0.5687341041343354,0.20855216012966382,0.06390114084955523,
 0.8323695796493555,0.23178951193019726,0.1823779190094893,0.9320108836272393,
 0.8820320512182845,0.5919923287758482,0.5921091465160715,0.5443919601130659,
 0.5254282416781038,0.6024882479694624,0.08502956724291288,0.8477719750143627,
 0.9275194605556435,0.6640799009677378,0.04161308494911775,0.7277573748087568,
 0.7575542356690833,0.8402791623083765,0.46058115642128106,
 0.11079690399309738,0.21904876086939917,0.05898651213633899,
 0.6225544845197075,0.41235868705689516,0.6762104221844868,0.1498392010097147,
 0.998287621302644,0.9622661691050861,0.9438669252011385,0.2504872257981643,
 0.5926305788833544,0.8157124189395846,0.9954341859916968,0.9391433167754711,
 0.5854994907806148,0.9017505689474141,0.11863876997428069,0.500989395337039,
 0.6883312914338585,0.08767725808850613,0.9155555769679287,0.687188990837011,
 0.9905039223897785,0.640872888414865,0.6759298635646857,0.939140929507634,
 0.4339242324339133,0.44286621768236467,0.4122641350832328,0.4966581917378897,
 0.7589716994464445,0.516508863824564,0.6165170058555226,0.39635411864248193,
 0.5086435888817638,0.7276724098951564,0.6115549881416124,0.12153687345331565,
 0.5564881205680855,0.9505719165854105,0.7705007873061148,0.1114467400542497,
 0.7972493027636345,0.32697829434352244,0.20477218051293322,
 0.9945821045776089,0.379872009211929,0.6448351601830128,0.22616311811815648,
 0.29794763652562617,0.834287372045746,0.38497719851953105,0.7455349353701044,
 0.5704077030586576,0.5597810039067348,0.055008588741150444,
 0.32030321399864203,0.3413561826454963,0.7511273131538407,0.8041851712140169,
 0.9565099888892309,0.10852184286446331,0.7441874748833232,0.3515300550563354,
 0.7182690362934792,0.6200973335909631,0.39965756977474765,0.3984846774905476,
 0.4495610186412934,0.47223024557402093,0.29870516637279576,
 0.24859466701770982,0.7261022889350869,0.30931141564849074,
 0.6704833183458023,0.2462391548972268,0.06797484294129963,0.8616917124441554,
 0.3506873680291388,0.3584282288932291,0.5098361795851544,0.14544792870254564,
 0.04647322128697573,0.47996516507023146,0.7667714110091125,
 0.11509770017190846,0.0946862010860473,0.006545539796195876,0.53260264405538,
 0.5820681835050779,0.7764507448202372,0.16066268589254573,
 0.49161746670187734,0.9563401397762208,0.5317772711545864,0.9749583965954156,
 0.6395306288129903,0.7629005984489439,0.03858661277983222,0.3935345937697269,
 0.6024143618592375,0.09193294164461097,0.1852295016126816,0.7055451127558273,
 0.9723650203998795,0.11366475381128593,0.9488269023626903,0.6022149831457976,
 0.6751555511692241,0.20651818504113184,0.28965641249701246,
 0.7483319424680823,0.7309157135569226,0.04224870697608085,
 0.33566654764173676,0.8804426228077455,0.47237989287864,0.7588830812319193,
 0.4290698241078308,0.8997262302426453,0.9269281119357503,0.7461326484277059,
 0.055003280216762485,0.3762293117603558,0.4050112850755284,
 0.027729518178750023,0.4128697989587128,0.37051736081510744,
 0.7157983928109194,0.8135729638309089,0.3422874576663436,0.8995535698518601,
 0.7017712761418411,0.37469431338125647,0.9670873943798475,
 0.09709176701198352,0.4335686040169575,0.7723826562518787,0.8061189549325878,
 8.431494642960891e-4,0.226804915592424,0.47980167718251077,
 0.9030433565596684,0.2666733986276413,0.8772046651418223,0.9987936276728249,
 0.12608020991768498,0.27379019674663907,0.08414225962812583,
 0.6421258067005889,0.9985147326708178,0.5992257414743006,0.5264385091916288,
 0.5775483806132079,0.3425473377527516,0.23722579821621814,
 0.10003511564123668,0.7098780121021495,0.9201634265169282,0.3670138074103615,
 0.41052791814756573,0.6631210213311989,0.47927285875288483,
 0.1358710115386288,0.9985947164747121,0.9074105248643672,0.7625998971824064,
 0.28944395281714663,0.5860793550766566,0.16704692205347993,
 0.5947781270084092,0.7838910258466423,0.030259760890414933,
 0.7732306565132205,0.14515225912791896,0.6834994619049819,0.6203329189571533,
 0.9961304793438883,0.14902320284014237,0.05216817774597693,0.801596570782464,
 0.9037379843837723,0.7507284820415209,0.0674075658858937,0.6807596527163453,
 0.1576628170428067,0.26052147596211983,0.507736314511841,0.4336998651751055,
 0.13759316966850343,0.7539164502399562,0.46409576055377166,
 0.1317370033682791,0.3054979312127829,0.1605350533369262,0.7199558220669874,
 0.9267807917275234,0.18679225962347124,0.9829041241432064,0.9245752879961308,
 0.05176161272894042,0.21681439650780976,0.8335876392169705,
 0.21229936300697094,0.14240212722791923,0.24686899222550562,
 0.48569021091197917,0.29006210031624036,0.5570034414542071,
 0.8032135432336539,0.6814075416554183,0.15002696709613872,0.9240852859156519,
 0.6304645364987368,0.9854479259883409,0.9956814320420985,0.7925807401348831,
 0.9134771926223271,0.5246180442244517,0.2070367310765716,0.36651269481166493,
 0.0016741776026339972,0.18735465065887547,0.03524609884120211,
 0.9081446645113055,0.8380741677309067,0.42307617646846474,0.6489638300105766,
 0.695493726422301,0.13971075442290004,0.20488656022784113,0.9114410383405036,
 0.6569400344844729,0.6046880785464112,0.20955857937659228,0.7515154217674396,
 0.045647225021236415,0.5003602689699782,0.14594967801573522,
 0.8034377041039894,0.5208109228526957,0.14036592440278384,0.7460664024050759,
 0.9673830780191841,0.06856153750211313,0.25937462466599703,
 0.5698038169602522,0.7685586097016041,0.40786299631726675,0.9693161277560243,
 0.668369721483431,0.9915514852848578,0.2903649963411311,0.2311086980280379,
 0.8219170500753007,0.4664415946376621,0.966438419158469,0.4050165226472815,
 0.7907167914805313,0.306606381205748,0.09491195659924268,0.8170836764189325,
 0.7826710808637414,0.08711218976615198,0.7593778853092191,0.9262752190144106,
 0.02776480267914927,0.3219393486125808,0.6134193179819779,0.9252408635031195,
 0.35292937940724356,0.37274117511869387,0.9580505275023186,
 0.7241908476990873,0.4215000472608771,0.17804205993675903,0.8535572310365027,
 0.6920307169022175,0.543521090467779,0.06193685249551051,0.10619221230063713,
 0.08772126849197914,0.23996923599643583,0.5087519712994669,
 0.7365192253209891,0.5255979051514235,0.7696164356279733,0.9780062270147034,
 0.985423484576175,0.46225593566631185,0.9378835492245929,0.04803856729815825,
 0.6202690800635716,0.2832066268881577,0.9702825674232809,0.5353936665018582,
 0.05939253171258441,0.36881579405533416,0.19457196527555576,0.91565551306324,
 0.7409683571255319,0.21575547258122063,0.8663436164740349,
 0.12857287866208134,0.9774127386753751,0.8102953541096549,
 0.09844250271618549,0.35353344826246325,0.6032608483349398,0.584286930809006,
 0.35413684370020737,0.40704033081752744,0.45673963731223144,
 0.2937278847565551,0.5292440716762077,0.3150376550155465,0.7105799868910069,
 0.35538173135833306,0.05111330829191896,0.6714205210377129,
 0.8624490778854087,0.2362878941016281,0.17880883225270638,0.5055376691849667,
 0.3932691626082152,0.866050992573157,0.23922130075010262,0.2664947252210852,
 0.5256017226461919,0.8572580124156446,0.19500137877745216,
 0.07612524543710442,0.28272063773049394,0.22618334742874158,
 0.5262454236792236,0.9772091493502888,0.41368169092898155,
 0.25012949098267356,0.5784700203335875,0.3830241299065714,0.9994876461136288,
 0.18171586682559693,0.4568975337491563,0.11497768898502359,
 0.3210613496618634,0.0838284833072297,0.259770087323445,0.2202441369878343,
 0.5942510999057107,0.5527152101399065,0.405776862852988,0.8675775401471156,
 0.08952241243467551,0.7647057234965756,0.7692857942637026,
 0.48681193069657414,0.8082106252855715,0.3974753511907705,0.3014559700927545,
 0.24401428904528077,0.8132445056708084,0.6211956787242682,
 0.41939399109098885,0.6440045294656056,0.4074586960653217,0.7035486676564492,
 0.034302912081584314,0.5844238240328103,0.46857234987116025,
 0.5640111488377791,0.47928005538565055,0.80221570100568,0.49149364002761853,
 0.7921729059818973,0.40191609555030827,0.9144219803492543,
 0.13024552507518028,0.5301981704937859,0.771332368903909,0.35359085740361273,
 0.19431458307278515,0.5899758716516202,0.04524289199268228,
 0.07251633845659633,0.9542349075609223,0.5086492622711207,0.5737213797077001,
 0.9816335365957437,0.06002070071002186,0.16467710456527107,
 0.5574530128920789,0.3474282312764183,0.17782390254062475,0.8379397146223143,
 0.044095774342573724,0.3203848272466787,0.5219231334713724,
 0.5728873878384738,0.731062674891038,0.6582089476372474,0.7904034461124414,
 0.9007830520057694,0.3131391394850098,0.7861381572547977,0.7673254429472888,
 0.3078994255536087,0.9032755166460558,0.506458742838636,0.2347260725103134,
 0.31345854791939254,0.9712496274885529,0.9544702462621418,
 0.17998195980225307,0.08863922228688592,0.4401310508638995,
 0.9887332549050564,0.11888968800005617,0.3111267451929547,0.6793359853921895,
 0.022464247661352132,0.6755238877156653,0.9509306728330503,
 0.7551879986986151,0.6593853412425433,0.5703690533728533,0.24204315856138092,
 0.4216728606910104,0.11492035266960299,0.5701514444107358,
 0.033474238540820656,0.17490639083711834,0.16816863399370396,
 0.1396535731780545,0.2993725045409753,0.8376288095051108,0.19564069640796355,
 0.7750341896380932,0.4463908597100492,0.249744107910574,0.2457540625281205,
 0.04518991475526035,0.6505129901320538,0.11894463110188935,
 0.09771906431182054,0.9732800153059654,0.8544887903165074,0.9092337619303328,
 0.45835762101921607,0.8237303546765322,0.7888550215286539,0.8558184861707244,
 0.26660958112686417,0.1475066365982981,0.3654581370200124,0.9376760339294578,
 0.4919834731369883,0.672990451291096,0.6226489968822858,0.6755983529557382,
 0.5388754466855634,0.9457818148789677,0.7610024070275162,0.26728782878248847,
 0.9784189013876622,0.4378492692610412,0.6727670666446844,0.902691782645579,
 0.9800950593530313,0.8361739587933039,0.6745204570212335,0.5574387757832254,
 0.032420295903408,0.3268239379262673,0.5582354883328724,0.5669650547463092,
 0.9999106904179094,0.8404558396867159,0.763284186959879,0.8744203230197108,
 0.7511616270407397,0.9335254997937277,0.156400333014032,0.6592997450469277,
 0.591240248188568,0.3882377721415766,0.08128657897649949,0.15280958459033178,
 0.09494478659408678,0.3041220563257119,0.8509308033480042,
 0.24964648258629962,0.1305081439133149,0.25603154162891795,
 0.2033664862820943,0.6584158732304404,0.050873423276234786,
 0.057854984650197006,0.6501177073730373,0.5665489793572724,0.930947781738737,
 0.1374965393488654,0.8425683075348823,0.47304477300533576,
 0.36463911355212375,0.9870142650211976,0.22546910740874126,
 0.4383636540867495,0.20910199861749101,0.4556957510853733,0.6100735082866215,
 0.41922573030503407,0.7387533491640184,0.1855035265261047,0.5723407920260798,
 0.922453248322306,0.00813531935887668,0.48436947301650346,
 0.49763437802302957,0.48007935763209353,0.3168728331213714,
 0.4671726849544062,0.528476715804635,0.4252130669563816,0.23033326158378808,
 0.21527225691376883,0.35421921941040835,0.48550732350668735,
 0.6503552092489832,0.6026595766867301,0.6249550011146142,0.6147943257569239,
 0.4392690342451264,0.7859572621087498,0.6533168003513774,0.21377882523887182,
 0.9448165470600097,0.37835078881038875,0.7887459310951356,0.2508913884561291,
 0.2487236697707278,0.24923426973742746,0.2878368324726842,0.8370474579206303,
 0.501702698903019,0.7077763984870863,0.7600641695141354,0.09625267100777113,
 0.0912916903154487,0.594479847524049,0.2901227432414947,0.9604237021769104,
 0.6977718208579179,0.3214331395187804,0.9737317382102771,0.33986154764550003,
 0.22231693921900475,0.22168378588863802,0.10260002248079125,
 0.774927201205874,0.09447926837307419,0.9216036593876049,0.7712664656683428,
 0.714659424182311,0.9260985991357543,0.6480940857228146,0.6774239869617077,
 0.5817730409342275,0.5831779331473819,0.9327304174181396,0.16386491688453053,
 0.9974433998439834,0.4130023766575246,0.4591680082994458,0.7021331643002975,
 0.7297217871469396,0.7616157926674602,0.7535604014119556,0.8538061352017794,
 0.6515333399780869,0.6399162901683046,0.5694920859542444,0.45110753073240184,
 0.17489299402053937,0.40618336812656475,0.010626063194563828,
 0.854781216847127,0.05321077166939525,0.27270750062620297,0.9495167608270509,
 0.27760656470857326,0.2273496237694489,0.8017528854555485,0.726667549846458,
 0.6399535099007732,0.5839844803045688,0.9963943528469889,0.03797991574503179,
 0.6146079759826432,0.8351817854394862,0.7603359055970929,0.4543169744053204,
 0.23752032631374642,0.42616116823599315,0.7575066437641894,
 0.8429519261820279,0.4368246273043902,0.7340622146678832,0.24289263683389684,
 0.8014004570193215,0.33892432822391794,0.3345926889618238,0.617189040328299,
 0.9363577081666417,0.5599940756181885,0.4980391414449389,0.2850336656151823,
 0.8155807539681881,0.9499419316772261,0.5020263790909756,0.6484545551527557,
 0.6546905013247624,0.9410607327109368,0.4271088584811934,0.4812493462243834,
 0.5386424565542239,0.5495240095979569,0.928685543690004,0.6703450383204026,
 0.7004990899542648,0.13092527923516062,0.5977619392057333,
 0.36909152200845197,0.31327627517672463,0.7806588615549678,
 0.8528187652125612,0.72897779696838,0.45345874839738853,0.9424329401295691,
 0.40149769350289743,0.2406723491991002,0.10218956355769704,
 0.9604733692266556,7.455008628699522e-4,0.7244352381104947,
 0.6192814722647912,0.4809066633321488,0.4423126391644223,0.5380720480731578,
 0.2862676742115149,0.15274569615611355,0.9407340028600724,0.8914489594074085,
 0.6400322405088393,0.47301712015522435,0.8029112040217649,
 0.08773661595767734,0.6181192358551661,0.3788357403064264,0.8868188909675522,
 0.44650679138223903,0.9868462724485,0.7662611725892632,0.84256630028787,
 0.8390363881810226,0.6774538525500783,0.5783676996018872,0.7400229364844143,
 0.04718396393038338,0.6815772384230724,0.43314852627565803,
 0.0805628327402571,0.10799078153684694,0.061809589078129834,
 0.614066921796638,0.42649242846200985,0.7116068697591809,0.694599510488617,
 0.08260056342019151,0.025367999219294446,0.5850638521139303,
 0.8448314947568483,0.4647021057787243,0.6577179580369893,0.4751555408636866,
 0.11413770468244755,0.7560697569552814,0.8551044265206925,0.6196671243162309,
 0.25070064862320396,0.8620675808158401,0.1679188107784535,
 0.0023483341403753855,0.18492010778248158,0.7968740731159261,
 0.4739748339750727,0.5399009295692939,0.7495278439961045,0.5474719802075525,
 0.14001801620320842,0.26891215690280057,0.5101778358118318,
 0.7028790389227941,0.8457805615852931,0.2747016514651073,0.20170002276398125,
 0.3845386366657745,0.37087649475104834,0.06097739453851658,
 0.8168385247069365,0.21477059670564946,0.5800829533609271,0.514706116503711,
 0.698301836637312,0.9178931386909899,0.8309205388293908,0.08260146805638013,
 0.7369030328014357,0.6369566499464179,0.22298058358812634,0.9353062501917682,
 0.7999563208077127,0.8087654769714985,0.27316841430733185,
 0.42890897863572985,0.4024114439363242,0.5048660727033372,0.254823561270321,
 0.7261332869753128,0.8517071567333584,0.09985735094886827,0.4862246277442722,
 0.66969428755797,0.9557994395864781,0.5346752765487008,0.5566959552713769,
 0.9311044643707264,0.25201432383397504,0.6729519216887984,0.3606073715062791,
 0.12285729284193758,0.7997596132503748,0.6142867146428562,0.9138184370451501,
 0.8342905316513799,0.7486670229054807,0.7548984104674558,0.05424719239157694,
 0.9202361101205496,0.3530588448964289,0.0932682060496185,0.8112491719884574,
 0.49535732195422355,0.32713127697517497,0.08792324728083056,
 0.44178677175483194,0.40405457425064006,0.6637874969527432,
 0.5220471419304482,0.611381618253436,0.8058548962743407,0.9047594926794259,
 0.26151757279485843,0.7967019092052617,0.7299347899188579,0.4032682170331515,
 0.6553247658419575,0.40574873860475824,0.003404577528696162,
 0.5564230633873718,0.7547835008720675,0.47370901497174955,0.241973117245611,
 0.936473652655398,0.38887469479218484,0.9967177260863157,0.38321207308509453,
 0.6568562926843688,0.27263689991328777,0.955387943163418,
 5.557994157212587e-5,0.10575557204808306,0.5908691981976053,
 0.5997388136095223,0.4202310487077654,0.23385792650712778,0.8142803460520394,
 0.5851960158560825,0.7317260471879143,0.9651484479291111,0.8880272883511955,
 0.8936485229311182,0.6491241719796154,0.25860950252250925,
 0.38668843881654325,0.03313199074882356,0.32989532559172463,
 0.2722111069447084,0.14617145688764843,0.15707202515666008,
 0.18931970889414085,0.7774532839347899,0.6662192639429383,0.2565800770354908,
 0.7086704781048603,0.5031150119161257,0.4683631007823843,0.7121604416281273,
 0.8054934567136887,0.21682578072673975,0.2638291377100649,0.8886454360932395,
 0.9626112955992966,0.36078981435346036,0.25233393209899857,
 0.5089916810862904,0.4460275295231815,0.8163630153592911,0.44838948193840233,
 0.6327068775129399,0.7665538398877533,0.22600254697737,0.7736786149362888,
 0.4187358158488633,0.03386221751794616,0.9787430885067526,0.9061534294679399,
 0.9298695285262268,0.585816690231365,0.6011351974099963,0.2660657475150706,
 0.6368989868202918,0.2220227113285379,0.5114713397985221,0.07739293056105079,
 0.0076576620004158436,0.8726186932812197,0.6652468172306323,
 0.602239406068713,0.2963780929925388,0.6454992578144368,0.6622396592201759,
 0.91325525023365,0.20676714629944293,0.42781929934821017,0.7821890789613601,
 0.27387240304850147,0.3946778132287828,0.6449470278171173,0.7004724876815868,
 0.6353787321780989,0.7063236765272088,0.3378308480153782,0.9064656655633302,
 0.6255229255446664,0.2730266814831548,0.14030328934113068,0.1154411654305112,
 0.7697905870148152,0.6214434413075772,0.9301982859602108,
 0.005893607065365125,0.8337502236858536,0.039149288967702756,
 0.3374019538426267,0.9196424051613804,0.514425373955901,0.675224921675702,
 0.10152670233305172,0.3484538829767402,0.6897734760262983,0.3666204239001233,
 0.30875124764488127,0.031968849173053315,0.569904953163519,
 0.6281083744047908,0.47473610007642775,0.29514855355344416,
 0.8118628454474612,0.017876303337193256,0.3556855767856436,
 0.38743827449701373,0.9294645267762867,0.6429942779962633,0.6978870678447265,
 0.9485788006912629,0.018294349596299275,0.716445179046398,0.9742830320124521,
 0.9525170984026938,0.8666745972669686,0.3839369614946656,0.9055448031738481,
 0.9880210153707396,0.20726230611988394,0.6174859822946558,0.4314393192062166,
 0.07269658954243285,0.24849041260024007,0.51762506709978,0.5386502671306447,
 0.20613473697067974,0.5905379034032461,0.8443765804363628,0.6092325732845958,
 0.9490770284263954,0.544437660604161,0.8258222359125592,0.5096822071244089,
 0.027238257571039926,0.11812553012611215,0.4511870481537257,
 0.5220140281726391,0.33561084283077725,0.0060178897393864395,
 0.2243583199813023,0.28676599553566595,0.3849035467074625,0.8031595237724056,
 0.4161056286251036,0.39263586091302416,0.16840687941187715,
 0.1512124159887599,0.45895676718135003,0.419701038554653,0.5999002807536644,
 0.14765538485344554,0.4461808432265819,0.012217577172521255,
 0.7866326694902461,0.05608091247676805,0.6582891038579985,0.9638270312668293,
 0.8443036936410676,0.4424687955683896,0.7950965115141094,0.6818821392878869,
 0.10753731661457366,0.006277355443635169,0.9465817408305655,
 0.9523648127746802,0.45218537691123306,0.029600447287555842,
 0.45487682018874376,0.6291918860453746,0.2545060123353501,0.5514440562598574,
 0.021994007972034324,0.20155002865239713,0.3348789565429531,
 0.48391962630719043,0.45142036304012834,0.1789116263539211,
 0.4799992376550295,0.52918853917299,0.7648909828869228,0.46881282519201384,
 0.7074927175030004,0.6946852264891579,0.5254377830232689,0.20904917703308845,
 0.6198202425939207,0.5017464549104991,0.2506121624688339,0.9059117210777796,
 0.8896070822629334,0.012483995311964091,0.7170277589536105,
 0.4899787944342787,0.2811161386007447,0.24152708332464612,0.8839216053986715,
 0.37191597912678087,0.5680699053478048,0.8341566988925777,0.6261859725073444,
 0.0623757136278974,0.35863339246521553,0.7505479192218978,
 0.05140371176815728,0.4745087366665488,0.47022080258584475,
 0.27965849766761086,0.31775288704978144,0.0742433144014123,
 0.9734989860388114,0.9769425657342292,0.8448528264925969,
 0.009084400457865627,0.1932549210029494,0.4851677195343067,
 0.18018085674026418,0.6054255596186027,0.812629505202559,0.805983369960927,
 0.05589256143657173,0.7015704641311356,0.3095194010454744,0.9292983565785022,
 0.6918104007005151,0.8661311957256391,0.11965969430936674,0.5878956738476403,
 0.297909867903819,0.013473781932759055,0.7071912631272503,
 0.23072456688553322,0.21608390847805414,0.4325839417728691,
 0.32048614859809865,0.9328354052019739,0.14458421960176626,
 0.9324284209998533,0.9478832653333056,0.8607892631464482,0.07928038896060852,
 0.6256369271665606,0.7959613761682993,0.040098688726581844,
 0.6294289253250129,0.8551752213890752,0.7463728279790409,0.6698590094109087,
 0.36167989020427493,0.599215701405099,0.9472958902934479,0.6339492192060993,
 0.9650644059371206,0.024582658922158274,0.44868697197005847,
 0.29531198583707285,0.3676101976215683,0.2092100339027676,0.365899703513784,
 0.36487241885874544,0.8486589045581794,0.1402555580279381,
 0.15487097987694376,0.6490333759105207,0.6755881411573659,
 0.22586858099016793,0.27764324350468705,0.864967565079609,0.4255807296421701,
 0.3385472973824347,0.07800167531584967,0.16887234723173228,
 0.2146692657186835,0.32145592178460136,0.9325052464978315,0.8096053203721775,
 0.2325558905215792,0.5960467887878119,0.14444572280375145,
 0.029931233482887065,0.34352742612172915,0.5335690366391213,
 0.03518869322730267,0.7341835357188221,0.24243732478341096,
 0.12250672897649129,0.813931047069063,0.7637478147668232,0.18034828591740548,
 0.5817880902985494,0.4035760354844311,0.601282098020215,0.9267100778089722,
 0.8000538678951643,0.0391220734809955,0.6159444244162257,
 1.5031226298722888e-4,0.7758184426496991,0.570541719215649,
 0.4701274048293679,0.20684959976323325,0.18088702638478948,
 0.2974769059075113,0.08437313472241381,0.18330331869188377,
 0.7545358859629681,0.30800583284191896,0.3138493355801272,0.6624249809513523,
 0.0019505913388255802,0.8347590152041496,0.18482315869048427,
 0.07421021822938478,0.7500832820031368,0.048473017839552046,
 0.6097652732253575,0.7530784713239572,0.34138378985430107,
 0.16444625228411036,0.5854074924094628,0.7393365870276881,0.9310724987298274,
 0.902404866457126,0.49857320550471984,0.41458065532068444,0.7949354049214569,
 0.5490054311843623,0.9710090235478788,0.7735084016808236,0.5107339199512737,
 0.42451597293218235,0.9281776592970667,0.15335228246610544,
 0.6818514192561171,0.9551858559292804,0.574425208325031,0.6540447711348418,
 0.48992367539661774,0.9644097320839099,0.3570816523576289,0.7098124213155814,
 0.5463002568163841,0.04261930774974998,0.7470348156492973,0.1419640098165611,
 0.05569574819808665,0.7895619078088502,7.253224965942939e-4,
 0.023214633762407844,0.6230476647072535,0.3335106240073127,
 0.4582438172834531,0.8303076865143275,0.8543402179230861,0.17555266104154188,
 0.11637675980863016,0.9046885604439124,0.684924997910584,0.9855177477931234,
 0.4008644254968491,0.014901598225478652,0.06013065729309808,
 0.05438173682760428,0.9889219940613181,0.40054144984317486,
 0.7675793716114399,0.24415981050444113,0.28094218721873354,
 0.3029007590548447,0.7491682201151465,0.7076200470241807,0.7716778638530042,
 0.7078106295385407,0.9602979149026748,0.6955190695555306,0.6044861634961378,
 0.20478697536957324,0.7231612154292818,0.2813253643072562,0.1847465958116783,
 0.07866561765768987,0.20021938467720835,0.6661171358184461,0.920921179350892,
 0.846278834105715,0.976846819363883,0.4359789765814045,0.34273114150039685,
 0.49837900406956925,0.18769809614274713,0.4398303128430707,
 0.7776731616225175,0.8496186853556398,0.16277528639846373,
 0.09689811459530973,0.41989406213464464,0.6927432019207017,0.779506610147357,
 0.540746219601125,0.4825899115565222,0.32174516748227144,0.06468762464737932,
 0.18596551882975232,0.37750359305013304,0.7935313218981169,
 0.8767529775015764,0.12823240077140885,0.15341773549324134,
 0.8551684005207846,0.3228184449555688,0.7048055952848656,0.9411714293618679,
 0.6775576435195054,0.2621254762062243,0.5076756837261439,0.1766034047593662,
 0.2039869725140443,0.1389758654609896,0.6326452739099933,0.3308645499338466,
 0.05596611919996475,0.31356832619854125,0.5920551223680145,
 0.019652607463851557,0.32841229162686636,0.038458073019971395,
 0.1951176019244234,0.37199428632377884,0.5468489939131923,0.7322105631419006,
 0.22277194496168262,0.4590157027978615,0.32462680734376537,
 0.21014608526109302,0.17241592614696577,0.11189188319774523,
 0.49495013940590726,0.9717533921361383,0.10360920878885138,
 0.7740600195952476,0.8198939413477037,0.6304319238001225,0.29452974871099036,
 0.8287798963026132,0.8121593836565064,0.6686129969196368,0.9609325946057892,
 0.07012724821960259,0.5833933529006893,0.01444921800373089,
 0.5599251300361673,0.35551453223828544,0.5160710028815356,
 0.04606390452012643,0.1251572545341233,0.255012368487039,0.05570072191631226,
 0.5540412866141916,0.7428065801416738,0.7854458309795106,0.17308873209521014,
 0.22149587305877738,0.4991282388339907,0.2048269832804519,0.825153055422709,
 0.512823256202235,0.06843718043902747,0.3828209886701661,0.6645812324089811,
 0.7831363827471813,0.9985075221792041,0.15724026597946938,0.3063256892483043,
 0.4722323558774192,0.5266615501767509,0.909728831238244,0.8880434844298768,
 0.6602218474519116,0.012714295126027175,0.606393338405327,0.8119119138805557,
 0.06398183059691398,0.23113769484286373,0.7029700895181379,
 0.6023226053170547,0.780060930315605,0.8127440013799115,0.2153346497628431,
 0.2767136855533704,0.6943354028551165,0.6977721376312515,0.8509973344532025,
 0.2795016730533123,0.3664061867728242,0.3646624821951425,0.2327413832700358,
 0.021938150990172378,0.5520572252416155,0.4636710709545646,
 0.09371179597535262,0.5995526180217214,0.7312832026097973,0.5836494906369729,
 0.20958055309970214,0.8483776533186034,0.7824425545601051,0.8453153622906733,
 0.24475951217289482,0.9373466576300502,0.4386755102420822,0.3116900596937138,
 0.9270235232018824,0.37586203537315643,0.840856188257374,0.4094506230868852,
 0.24701834714727278,0.9236826853123726,0.5465737874885594,
 0.29567596616929714,0.9657965060961444,0.3156887794387295,
 0.052633609956159066,0.2971914350924122,0.6198735640475034,
 0.4465526698676541,0.8428211141912725,0.0064955814689773295,
 0.29647887536291806,0.31494351191594594,0.7596828069259808,
 0.36539018710786353,0.8510522467615077,0.8183871752563414,0.7965060796383304,
 0.5050066327912252,0.5909241867292702,0.17699628991256555,0.7878895775316286,
 0.19508656606295394,0.1960116727381075,0.7965603062191673,0.2149497767253914,
 0.8170694258186539,0.2629783215481222,0.9682141684712359,0.6930162416602644,
 0.9731193791643987,0.5552963899979731,0.24712307852586402,0.7914432940736851,
 0.3610046452766632,0.10287065453421163,0.9524757541579812,0.3455751429386801,
 0.868199377992134,0.435540065743634,0.007043852724186839,0.3420582702959182,
 0.7397727728357328,0.7013282034493447,0.1925538230039523,0.9486312497212233,
 0.3786617117447637,0.826757597632181,0.17903085662864182,0.05422801218692841,
 0.4848120821463311,0.4690990908024182,0.1680597365622557,0.6250292029263326,
 0.07202488922020422,0.48461720153818266,0.570848479513251,0.9709450428194032,
 0.25678283884700925,0.903670091151335,0.6417375508790455,0.23062305006955375,
 0.69839603401753,0.2765263549605337,0.9858736031815624,0.7568724976933713,
 0.9568520234350785,0.1868015245440261,0.7849960749270679,0.3563024826450927,
 0.0878523314424281,0.01928462923276153,0.7520955130460167,0.4688478774000854,
 0.3470809850557175,0.74212451865888,0.6656653660407774,0.4174585965978046,
 0.057283081242215594,0.6444692406277682,0.3631691586145064,
 0.16082936446378215,0.7845103346310828,0.06154796580997979,
 0.8249321074952414,0.8670460174438226,0.8893547025286926,0.7481933449145727,
 0.5484581036072037,0.6780957958188916,0.6928085243596679,0.5791190495938913,
 0.28853427781822927,0.5874342882839917,0.1306971942665075,
 0.35095171612691345,0.5590186905107748,0.7578185831408454,0.8458788836346325,
 0.3254617277453735,0.5737900855200619,0.8846228008801356,0.7380801653047836,
 0.5960113703310137,0.7392633244855713,0.20266280246737112,0.1685875020043116,
 0.20205776132466025,0.8456603532828574,0.7273245245471339,0.7079378082316017,
 0.9163933417602983,0.7152599100610347,0.4643627807255898,0.23893657098660848,
 0.8181944633999673,0.37012348952817176,0.48078145121336535,0.567578115009516,
 0.9923416580267825,0.7406954415928545,0.6472439062222339,0.17958035458201815,
 0.8177697080052578,0.3428389323800819,0.43332545013790535,0.7317176292757968,
 0.13126821409956668,0.9934064257417105,0.78370565994804,0.5168398944632941,
 0.9909712487403489,0.8850939140255136,0.42453845246694466,
 0.33082140403608895,0.29905050280771295,0.225901347046072,
 0.35410193954725533,0.2560263793701857,0.27038098317007797,
 0.06248375734067657,0.5437489084299566,0.9343494965641127,0.511535670458122,
 0.28090558369524554,0.8992044685779865,0.9062325715824585,0.9623415223935353,
 0.36289038852908806,0.12359033385521456,0.8284874415547482,
 0.7649569357004196,0.00405779964571229,0.4964386345827815,
 0.18722152686006543,0.4501797132562785,0.25196342111224235,
 0.009914342552755917,0.17784153878420372,0.08887786968893019,
 0.29677768223277223,0.9536875742151877,0.7290286059986398,0.322276121028679,
 0.9901830060353726,0.3743969145399171,0.48014477640459297,0.9513290490240508,
 0.09566787128676113,0.8484595647498622,0.5717135558684661,0.6766059660352186,
 0.27971110583301373,0.2593490621103788,0.34503048169368566,
 0.1637231960964447,0.678634955445651,0.38864934374782034,0.6714393549180466,
 0.3968582174305699,0.06057242491642567,0.24715055193939717,
 0.9327772363276527,0.31177549299577945,0.029557650340217068,
 0.8411800934004371,0.086200449668007,0.289845973641533,0.2822947163486035,
 0.7564872223617093,0.6621064946516462,0.17048765233650087,0.7763107458689552,
 0.4674847281971839,0.07776516231218467,0.032040144090958234,
 0.8633572688284549,0.7275688763318695,0.16963809491498838,
 0.11117587669881146,0.9539152783852528,0.026251869604356104,
 0.5784286589194751,0.32208482382995596,0.9739674186905016,
 0.05137062248591051,0.21331914106177496,0.22431580775905824,
 0.22595921098644223,0.027689333729503107,0.944947879700259,
 0.15214533767390226,0.5744682156758351,0.0430249241167846,0.5812278381984652,
 0.865913115335961,0.40465733224332623,0.09834449871202033,0.3025440005559411,
 0.4749829190047665,0.15221723270568066,0.8369479662449166,0.6675214068461431,
 0.12182252419213124,0.854935800941741,0.38950602432709613,0.9744584883690115,
 0.49952272375547135,0.3885632820594367,0.5266591522204057,
 0.40096071284845447,0.8215110948708972,0.6897420163452339,0.9054584967717888,
 0.7136077307831439,0.23621408961964563,0.8549602875808039,0.8609563966461227,
 0.8246312632918879,0.10450803652038676,0.5770890292643586,
 0.33258181575145185,0.5522497435607717,0.6269308719380768,0.8890125306908538,
 0.5176915062999673,0.8328159998649418,0.4541297630539919,0.41214493609020364,
 0.2872907100091294,0.35838549557446986,0.4702458146184355,0.3351160427366723,
 0.033897458800155045,0.17949060838881015,0.9586461279326456,
 0.011508547044401074,0.819738410647811,0.45940622483763227,
 0.8901958665785599,0.9904488337937097,0.5545273718882356,0.8804131699162436,
 0.27210491816019544,0.013983560286321306,0.01861496517723016,
 0.1311255126735683,0.6085979726449008,0.2146976956506117,0.3767679121933618,
 0.3133185239583003,0.3096627624376145,0.8614827268613598,0.5247459914713714,
 0.7905399991209032,0.4832283799885341,0.08435152537820834,0.7494515792561122,
 0.0695737390205855,0.6203745834183251,0.8518151930571666,0.35916960273311904,
 0.7210187881050855,0.29457832551357566,0.8972426524610609,0.6374845434652969,
 0.03447810448839006,0.723441977309857,0.4258547478162056,0.8430512913314058,
 0.016797602144651202,0.02024990850286368,0.9036692137028876,
 0.23948996492761643,0.44526866585653024,0.5977129942241606,
 0.16274114608943258,0.889157458501775,0.9617774243521631,0.714492644443593,
 0.33065455702222446,0.5245107980954526,0.3074608194786248,
 0.06378466302911412,0.9312581043889193,0.691263266079538,0.6574559127039067,
 0.6622750770644624,0.37600182398501303,0.9092101233709425,0.9729143404516397,
 0.6509278446800358,0.19713394155486852,0.1055045253816026,0.9740549319100642,
 0.10527820603429494,0.5550393583499358,0.8379470950740275,0.4223742766062606,
 0.10533954893110453,0.7743908561356334,0.8069214006584086,
 0.061716171196593006,0.33508271734597916,0.3411649545017553,
 0.17008736335168106,0.4510339456879978,0.3824686708014392,
 0.14778017956131895,0.5051102435054345,0.9078592310124916,
 0.49717543227556826,0.03421082351928473,0.8867115233672641,
 0.2441170665143978,0.6758408457492505,0.15865690786610265,0.1628003485813162,
 0.0011565858418639552,0.8457735090897175,0.9342689105674621,
 0.6018099479717458,0.7305300590433359,0.02356160597492618,0.7728173689794969,
 0.6180616618417688,0.4727063900431008,0.565666401831561,0.33429285909076933,
 0.7693758967199875,0.09310932966963614,0.49584830905279154,
 0.4720892020747738,0.7401168174264738,0.33981561765961066,
 0.43578546007366126,0.5934608765919993,0.4894052132809591,
 0.16782597679756617,0.5270347356162555,0.6487501383579536,0.7423857765085538,
 0.2721330742501872,0.40199566794480324,0.2220343507927376,0.9167978231734083,
 0.6332714327669724,0.3993618423094719,0.01341180827568167,0.868901952421072,
 0.18320022072033493,0.8864914738269537,0.34018652545613004,
 0.7089982568827853,0.6573409098769387,0.8951157622225328,
 0.021563732625993648,0.5898787972532998,0.3928150709360436,
 0.9154272530800178,0.8283619096114416,0.3614167954296197,0.34312488121141027,
 0.9757058807301002,0.4888781593834113,0.6022934283111365,0.9593007427810825,
 0.6316458983401774,0.9135898932331767,0.1894702413199666,0.11504760277654857,
 0.6784048612563852,0.3459621026729225,0.6800264367682983,0.9105526062292011,
 0.7159605023275817,0.2570910177099991,0.8936179759925272,0.0542981034085771,
 0.3276020399303129,0.9808837143439806,0.7593205652291254,0.35025642142311375,
 0.028722879279020508,0.3748081202690755,0.3977371821083334,
 0.8215009641439021,0.08513474291764456,0.663159470440921,0.9324705688837696,
 0.10380036074090504,0.4596057391864944,0.686968647720238,0.33422267335243994,
 0.3107085335501031,0.43569528275154557,0.4208899911198015,0.5584749871363465,
 0.10561224015706894,0.7622974305855796,0.9368675793222365,0.660620602034541,
 0.6186726409728271,0.05250202359961453,0.07930679487697412,
 0.8188022210523523,0.5768874321593701,0.046541020653740994,
 0.47124929804957805,0.8809260219538175,0.3795603568243494,0.9092161449028681,
 0.3084260473061786,0.5591987846505946,0.527611316164237,0.8437637769709515,
 0.4308238076655526,0.9477062191485862,0.6056943861412425,0.12279586344706239,
 0.09639569436719031,0.6132589035847432,0.17705856885060722,
 0.8108535157789154,0.14526099759408806,0.26074491956091983,
 0.8326972429990289,0.5605765354313668,0.1771140805778615,0.9634233204493927,
 0.3905247602959536,0.4377858123855105,0.4327227492938639,
 0.028268194868091512,0.6776965053075124,0.09593291732563358,
 0.6754483476891091,0.003195348312194368,0.03293767619860233,
 0.09031085435503006,0.49802810346635784,0.8568390898617931,
 0.19810131728778702,0.39630504402887046,0.14105290796192438,
 0.9658343321723762,0.8070438988882852,0.5849140163280124,0.08087654830954005,
 0.6502341255350886,0.4050848543828668,0.9637757865499765,0.19760084812539203,
 0.9995860393267323,0.412564041436711,0.8105950838302804,0.12505379885047696,
 0.14780674395240312,0.28868052910516817,0.9960919001307557,
 0.3113319051949337,0.5247911333180361,0.9399968078107552,0.26093052464833466,
 0.5165900114827895,0.09665578653772489,0.21482688065522315,
 0.4822203473358899,0.01611584170462166,0.9745334761900216,
 0.14944853493534316,0.7389433279397658,0.09578380326426783,
 0.4077268686433415,0.06665673980618292,0.16964911809871552,
 0.6058616954576024,0.7030319011620564,0.6920536157682963,0.4405184930967888,
 0.9505897848847937,0.5806009313310438,0.9397494347046838,0.1868725748923149,
 0.03404970011942687,0.6390062543181411,0.5324080189696057,0.8264099175802496,
 0.7503990758027108,0.0962975789248749,0.12270037327289152,0.80626380602542,
 0.9977916847414888,0.29299093943030075,0.18961031884079815,
 0.9771612042188437,0.470591150438595,0.9965580964294284,0.30314414341732965,
 0.9829362812913159,0.13503001097681389,0.514735202734775,0.853603088925432,
 0.22528532445637173,0.034229085843743845,0.6773561245562597,
 0.5488175095765953,0.28153680999015096,0.8401744469787924,0.3765140623316239,
 0.7589269998983165,0.44707338711539724,0.9036063134525588,0.861349714681591,
 0.7713946374618539,0.9601401865597605,0.28901018942621537,
 0.48992335583379065,0.010210777186259734,0.6315489910488464,
 0.9879536105684796,0.39602683603115274,0.2360045142261925,0.8772590704073199,
 0.3617214030016078,0.09577599157933925,0.01800749702829041,
 0.037447198945505744,0.8465165263428792,0.5719904023720539,
 0.020417079996737697,0.3601910244089166,0.9011875686980018,
 0.8722696362475011,0.2480118796865436,0.8486431461209106,0.9764878540333054,
 0.8116151744294925,0.22850268113017302,0.1661036402255186,0.9295770968105584,
 0.7505874809841074,0.2248815328157565,0.9540701120982804,0.08748772926852533,
 0.8447676077223673,0.742547081769673,0.06278913092270977,0.13589372852992065,
 0.9663022294014533,0.7339245733931461,0.29210850557311496,0.9416733683584768,
 0.22043915219648558,0.9512034731098901,0.9873729253220339,
 0.05476661510435432,0.053284832521543035,0.24034466356141382,
 0.9442041446299771,0.07461362610001421,0.9662920186718305,0.9695151057538467,
 0.9202147088374625,0.9939832010322404,0.46250115415753856,0.257704083946257,
 0.10365209662974095,0.7755487559181616,0.996084710637091,0.6797305450428666,
 0.8841967030378326,0.09387677957003715,0.867389071041337,0.7762519371693417,
 0.39808624226394984,0.44917511207252636,0.854181876059358,0.6609049931962547,
 0.982556898342225,0.057746413628655624,0.02438350784364951,
 0.4887363230107096,0.9947667920614942,0.13264405670251,0.5812049120172084,
 0.5508462570921054,0.9940678559236051,0.8469686456407154,0.20675546242604015,
 0.1470192519071417,0.5959106506931859,0.10840228836730226,
 0.23062418013305575,0.5684330409504987,0.26705435874238737,
 0.19264015254956357,0.9629465797565055,0.4831791227418999,0.9314013353685167,
 0.9703596659483845,0.6095541499587598,0.45988847325659465,
 0.20702280486804703,0.004217798002549844,0.9726416105531045,
 0.5336646959522806,0.28215740216141383,0.716506421987907,0.8500944170528643,
 0.24852521559340834,0.22845810742488448,0.1804792785695516,
 0.8399443156911484,0.8847785372135076,0.4500174612055049,0.7937678403018329,
 0.9287080793780437,0.659248714041577,0.616382936571469,0.5723765823230094,
 0.686077133088928,0.1487412578242171,0.44372986504992584,0.49510155029353253,
 0.03485294517593174,0.6822115429737704,0.09960356479022292,
 0.6471567939954846,0.25413492140599214,0.23566548515374863,0.518556350656732,
 0.02287969021649361,0.6687415562151283,0.3889338182410689,0.2893845233080091,
 0.6551784328562354,0.24390937162928472,0.14232412190473753,
 0.7193233100146469,0.36072500888953274,0.8043590992807018,0.3067951584267592,
 0.7417676381600194,0.7005840126887978,0.6670607209249945,0.7489107490193503,
 0.1590693454289992,0.6866233767665425,0.6160415447067457,0.3016617097503912,
 0.38188459948836684,0.7980941596761794,0.6508567892547507,0.4082562735098205,
 0.8126094853535103,0.20492586799010293,0.4553189814491908,0.7983735882860383,
 0.271291479568449,0.21842046213367805,0.5545444115024225,0.18270889567668303,
 0.05634820630791382,0.9239269918344414,0.6189992561676142,0.3217487161204703,
 0.8534794904067601,0.23373636590234104,0.17648439881832223,
 0.6628459358164777,0.13572764220737443,0.6342923880679765,
 0.11777072865102023,0.4985279520783992,0.01762739137438596,
 0.4090422170485226,0.9145876026701264,0.6294420129162781,0.20319131943749025,
 0.33849087503823083,0.4087488360217064,0.09650323128852234,
 0.19243698644892282,0.645748953079604,0.4911117437253951,0.7296210448000291,
 0.8888222156912513,0.65923568835931,0.30747962063280576,0.6473413809275906,
 0.30037107544430475,0.7140618222534387,0.5450763069956855,
 0.45663305338755367,0.897245610769024,0.4988850742184924,0.9845550552454437,
 0.7872354381574165,0.3716823081000509,0.41519407175183576,0.7451883723687103,
 0.5672805475697444,0.10036184824733974,0.21725471793344375,
 0.32788901547314286,0.007898910008097948,0.1768918646241331,
 0.9744540728427509,0.892502722766002,0.8477881530015171,0.33955568658610236,
 0.40101861292086416,0.234091307201028,0.8548742549567755,0.967268300908859,
 0.7082374577726518,0.5197258767890578,0.7922415080162397,0.6755927805818138,
 0.45512030376017654,0.670342886604827,0.573124827406164,0.005311208780390708,
 0.11579499673826377,0.6818454919870702,0.9378040698123248,
 0.40584311674572393,0.006541878145032265,0.11506478301574263,
 0.7188966344302157,0.19460757879752433,0.6210763231267871,0.7112687689598812,
 0.2425541654540212,0.3720271235935284,0.14187988156022735,0.5779259757567243,
 0.16355733543038276,0.9700372605698405,0.32755753594997283,
 0.8681085911769416,0.7879087604988696,0.8323397481741226,0.7809824772056729,
 0.26951982366059635,0.4401540964751508,0.05312125690822167,
 0.003387199543072983,0.9034015907227053,0.8741773045646033,0.537009711514102,
 0.30062220244120197,0.5902300420754217,0.9565517111722494,0.6869021985574058,
 0.8856130069970061,0.029112409699775244,0.5260743738744442,
 0.7716344369598136,0.6591999481362157,0.6387514734680063,0.7389779125299176,
 0.8539479879325681,0.09669820912771232,0.07875442908095032,
 0.9735823833134333,0.9097136485443907,0.5705644548037955,0.3834883912896484,
 0.5927770145560392,0.34221541169183034,0.57425538061547,0.7326347449398707,
 0.3545073433113195,0.5866781536111958,0.287360501204456,0.84150673607616,
 0.9746408900633973,0.8080272145945261,0.6081140244505463,0.6329938191395581,
 0.6236783181166219,0.5262058302412375,0.27135642146525984,
 0.026842621376718334,0.6261809180529108,0.5368330977884898,
 0.7360489208846441,0.27564209125846884,0.4637780500689307,0.4558852280678207,
 0.11580004225899576,0.23825833915252348,0.4966061474530812,
 0.04544932524607659,0.15301397304126407,0.8700042621684192,
 0.35467347725924403,0.7164921138216079,0.10245742891702514,
 0.21869162509546602,0.027524965905995513,0.4676678708423475,
 0.9965466867943276,0.6153245610744159,0.5946770817901911,0.04238313676886807,
 0.05500262593953675,0.7400615068084656,0.6903801013264785,
 0.16117747283507622,0.8370521291915444,0.8098946750034819,
 0.017530053465979356,0.27240239452358017,0.37489705704260634,
 0.5514778073879205,0.8485559225731337,0.02254277415835859,
 0.41031153934526543,0.26873907016907683,0.19093808725618977,
 0.2103814212253441,0.22531051097807675,0.22769251963400416,
 0.8892959492444724,0.4038243893323341,0.36918451053338575,
 0.016777457322904366,0.055556954531468916,0.5172246629408969,
 0.7880261494839912,0.27142333145453523,0.9916868827871561,0.6093319458788824,
 0.5401554593644752,0.4996742420554475,0.3302576235376923,0.35964932558419693,
 0.580013773935141,0.3201186981536983,0.7316992220429549,0.8581344118640091,
 0.546523695485271,0.8349720653946101,0.499513254615217,0.9393575941339304,
 0.21431369998479732,0.4383407330735708,0.5081916499559763,0.9751857144629228,
 0.10189786224656161,0.7275671558837035,0.6425287391555998,
 0.19940946800314507,0.8160025265789935,0.19477153787975154,
 0.43092434005051017,0.17895864302572584,0.12540761547502333,
 0.471328567937908,0.8918091330471342,0.19776577849535748,0.01999092792693291,
 0.3048019105504526,0.0819789726280592,0.2638975691892088,0.48520732984831005,
 0.2418833467288487,0.38962097291778164,0.790164822803124,0.7663877829646539,
 0.090029246281213,0.28112170255657776,0.40795404486970255,0.6539086875107121,
 0.019737662321767546,0.497249271610132,0.8300500382770056,0.5073060654058662,
 0.4815576038243139,0.5719562226353947,0.9279246721790427,0.5086071574152219,
 0.5043085614676642,0.26070264801456555,0.19447902690898633,
 0.2534388996538116,0.38403444178503177,0.12234146178677008,0.306736510854019,
 0.05720689771163423,0.5121148611790248,0.16123458268415336,0.377823680855187,
 0.8884708436514066,0.43355704094243974,0.7791731375076729,0.5899971627984127,
 0.07296490313493631,0.09936145058854995,0.9174036309319973,
 0.5666307197999584,0.675140814820763,0.631252545928627,0.057194555828171345,
 0.21730065945177834,0.493794900587728,0.8253623796290769,0.31147501006789713,
 0.7452869135320721,0.5195173488805784,0.07973318527285822,0.8053691828658752,
 0.09678222586670626,0.5102428002680728,0.7664939177553124,0.9585243947900864,
 0.6869647092558172,0.48838810237319175,0.40917318376874157,
 0.028245517267263764,0.5002487546100365,0.9724101958942815,0.847201004616835,
 0.9547953535371833,0.3875195406999409,0.20804243035508885,0.1512821463145304,
 0.7903153888936512,0.33086615319423407,0.3002241236648884,0.2689050456323989,
 0.5591436026247387,0.15586339941934624,0.8171244674209523,0.4580531735080685,
 0.8306201257894579,0.9053863952564792,0.9162628047509307,0.9950626990478703,
 0.02464803977032537,0.6471876220378535,0.5732418110497984,0.9584047545750285,
 0.046372761092836834,0.15090562440958277,0.3655312778980242,
 0.9628455400182343,0.8639996507552175,0.4090306825384751,0.6133922476522974,
 0.22400344763906554,0.39630961542686793,0.14881969502534653,
 0.8326671407014731,0.1009133731266072,0.13207117376584332,0.933475015802301,
 0.09770843866596468,0.4254883880106648,0.955988556986242,0.8038940918828535,
 0.7398949941271825,0.6499097202226629,0.07602379884380661,0.332967496774245,
 0.1453420739058776,0.5706494388726978,0.2562132770967871,0.14904747457010414,
 0.3496853671760165,0.20920965165934424,0.16534696587213626,
 0.22706565515883304,0.6891094473134872,0.2175107757455077,0.3799575944048844,
 0.2280826465910043,0.9092348809415162,0.5505342828384275,0.18982030589676047,
 0.9289010395472531,0.5650905298467962,0.5936925382406125,0.6251163612834398,
 0.7574015327590867,0.7131341780614173,0.1749020922128065,0.38937591535661187,
 0.8986157162387785,0.2696316356020153,0.9738663420514988,0.4674277933471014,
 0.46849262647544654,0.44789837694148027,0.07795449541185129,
 0.3766497547813509,0.7190535800213012,0.3021355865092116,0.08543333352663363,
 0.3356688950310137,0.14586375115280203,0.9406174847452058,
 0.43223679628192213,0.6021975552663529,0.029776876437488276,
 0.19975518058683495,0.07614420954746237,0.8865948385951773,
 0.6895608928850357,0.935120885798509,0.7190919388159009,0.3247245480908899,
 0.8092535649626311,0.2312260504694017,0.4423657367883691,0.3055685298724131,
 0.44419903617764067,0.15906389393120035,0.8176004285857275,0.787511922635559,
 0.8508893569963607,0.9986544740260297,0.29181771167738124,0.6670001287670321,
 0.5777132102279233,0.15856861188887306,0.05924494638763966,
 0.7260918674625519,0.3951187012827154,0.2877715870620523,0.10831362159246072,
 0.9155612610137255,0.8550261305370332,0.8722157138838612,0.6617374064992986,
 0.058798296776162884,0.3508438806509687,0.940386997991705,0.8190470277045444,
 0.19736755315466747,0.6980243890513443,0.759172896449831,0.8190569486068647,
 0.9216002039858978,0.7427582611702125,0.9554850765043417,0.9583886024470133,
 0.7303635319538703,0.6923808606973446,0.2551952000670219,0.845675558632062,
 0.3456761492058815,0.5810843361395248,0.8645056185428769,0.39450941615835866,
 0.6421182014441573,0.5529649216653055,0.6566833157373403,0.17244175712006082,
 0.5957721630526732,0.7402730028998727,0.977587416819686,0.1422632964775845,
 0.5087656562883384,0.2793585065266392,0.25114437808220524,
 0.46918023881861415,0.9322283959065636,0.16087184581438252,
 0.21025278333888708,0.6290581133576795,0.6132782751281352,0.6556037548388189,
 0.7125360959597821,0.28744058370331305,0.6955333401994702,
 0.08541239122164968,0.773528302025386,0.6379027671969303,0.6365638696466799,
 0.38390794942222817,0.6263985696263604,0.8514393041418403,0.5470557858030832,
 0.9890336812825046,0.6513629317540431,0.344644134803194,0.08258647325181134,
 0.2594941019130572,0.9200042762672216,0.4124790602304147,0.4700557593319943,
 0.14294676940629658,0.23224137140236722,0.1929043444338877,
 0.46324888043990164,0.35491390044990834,0.8287018793396839,
 0.14683134673554565,0.7791713436653354,0.6030612781785669,0.3346585079385974,
 0.27700752999048295,0.010041939943962075,0.7166354328928966,
 0.15572523252520243,0.8130761237294688,0.40435247791356876,
 0.6305006621287536,0.05782423982862506,0.12457315954368109,
 0.07093341615904758,0.5950029432055906,0.9536204960184642,0.7721888096966494,
 0.8814267921314182,0.24600750499838442,0.0543028184954899,0.4672370332130824,
 0.17455335344591416,0.46968979949654066,0.624756252756749,0.5814257300897427,
 0.031531135753165795,0.20612702209928757,0.23719932473438043,
 0.80834172350999,0.7041408223777499,0.15419610308928988,0.05552090993694625,
 0.20723560091882365,0.047600370548698145,0.3087575213583791,
 0.45872676771294396,0.1852015153255182,0.4733015416877794,0.3099465004432548,
 0.410472387867878,0.4928090742139625,0.5989238247745808,0.243363587420788,
 0.9465367005155388,0.23755269568662485,0.8282611013742798,0.6368086364122256,
 0.7186233686403593,0.4830831028334447,0.6457927282692895,0.6063655229464883,
 0.9108666496247747,0.47459611871150886,0.004757817288808086,
 0.23558160970760056,0.9866949621778209,0.03477992699392152,0.535887498405334,
 0.8432726656307203,0.3235143573844058,0.6877536996405739,0.2275813719418226,
 0.543837176674866,0.6704068947685039,0.5688407560075869,0.3617126561919064,
 0.4818027627014797,0.25708598909997327,0.6156450959096617,0.8495761317813706,
 0.8142667720275067,0.9448952289271456,0.9463229256048771,0.31320336069739996,
 0.10237245077557888,0.9383745145409041,0.049726864965509154,
 0.9209586421934421,0.8016771767856895,0.9963383761945483,0.7842595952017029,
 0.622019436301506,0.6647312106964117,0.01414793137982484,0.9757201658010686,
 0.9008195151733552,0.09682439617679994,0.31771662365299935,
 0.06719718943723041,0.9785566778591783,0.3269696685963813,
 0.26432636440383106,0.6422703100609684,0.5651341357848512,0.6448893079680594,
 0.6008505152723479,0.6752498500001457,0.49640100879467675,
 0.22478908232974648,0.30122140227872296,0.12920385433800352,
 0.19682357853215215,0.9834355142050842,0.21773006773294623,0.874307891940627,
 0.869133789654811,0.7521250088367057,0.036944946041697624,0.7306397281017722,
 0.1892989031863247,0.2552865454205615,0.45720301386702666,0.6041944780146852,
 0.8638672095496527,0.3196717794377738,0.5791367258156083,0.6562161577192243,
 0.12028509976544655,0.38385944979431985,0.7866569018940973,
 0.12424644729379708,0.9236606627279081,0.9963593389212861,0.6118287236624959,
 0.6477510933950218,0.518667544896914,0.8512048286644621,0.28924539596045595,
 0.5408995915663724,0.315893585380156,0.013988282292297094,0.6738537594864866,
 0.05559490530580802,0.5779749963521426,0.2999619810863339,0.9330513900260441,
 0.6746005916554584,0.7775510828211266,0.3357218433351308,0.9119525710560921,
 0.37616777489086717,0.1849077483417465,0.3257428365027353,0.9831507343386039,
 0.580455587655853,0.9048974934481879,0.11138530244303768,0.468085570235964,
 0.47336348777675974,0.13743534657249146,0.7982374034089199,0.928029760968553,
 0.05335562128242033,0.2222441093882308,0.8756030443186933,0.1060283154127516,
 0.3900648166568318,0.17962623160623992,0.059362655082607185,
 0.9596779113548027,0.5753474406844152,0.4245795888457088,0.4340197578860515,
 0.27093394261493586,0.537951329846838,0.8805067434449917,0.20276840156832132,
 0.6475761360019291,0.24131543040800973,0.9682022681076944,0.6924890337038283,
 0.4601574613759155,0.6027472033732951,0.1868407138781567,0.2624149024465635,
 0.9317135307611233,0.9881104864242305,0.709693289585653,0.4389624489313585,
 0.48924714565516936,0.44961220808690583,0.9997287261116743,
 0.9878409066668801,0.12929731056973215,0.5695412622134117,
 0.13882457569500695,0.06303246666753637,0.7912793326558448,
 0.8239450267141498,0.20718558224962969,0.6509309457033862,
 0.14275961887038313,0.3104746388329237,0.8960872008498413,0.8406822522303846,
 0.26415996139627407,0.8459296166663882,0.07898728021746682,
 0.021559661482707937,0.44887768396997063,0.5087505877259877,
 0.513329583798232,0.5520680343520921,0.3196677523197665,0.24595184301805573,
 0.6333640810597617,0.395843144902136,0.14700006162167545,0.6457500851462438,
 0.1143542371650359,0.32364624727330893,0.33447827996354595,
 0.8259923219850852,0.5767332932897619,0.9489010701670365,0.3743215118585541,
 0.8266679273816464,0.921487486901715,0.9767958841774276,0.9217120629643015,
 0.99026655873724,0.5200683688011258,0.3428053714646644,0.5671280543974857,
 0.8018152466779275,0.8876258265594366,0.9669183704099932,0.9874763071081543,
 0.22379831364126312,0.7941549040460222,0.6742612746434685,0.7347491407909313,
 0.7524972605241927,0.6371806670940937,0.6541170642981122,0.5650897752064683,
 0.19199517169165037,0.05733323432297355,0.15138157927676343,
 0.42754951466691005,0.5017356631436883,0.3715766897789753,
 0.042675345984736324,0.4886610769950941,0.47115975714232505,
 0.1283793982751732,0.6898012313514571,0.4812929194437304,0.8855449457640993,
 0.7242848737763442,0.5458948137891733,0.7722854094980476,0.49666493154072466,
 0.3669829831484259,0.691431291012415,0.6940559058463771,0.06060693466062539,
 0.7293096255363096,0.09852829364557014,0.8762765761444447,0.2893300097803283,
 0.5727086859784286,0.40572879809847684,0.968545251844043,0.9049661922514929,
 0.6226528226759745,0.759452732274747,0.7626987594961034,0.5694688382062488,
 0.5452587933768162,0.8008879834309542,0.7684773758625774,0.40514949862237637,
 0.23722836982819717,0.2243119090610175,0.7822470740795388,
 0.38761711090841877,0.9042355514159733,0.6697497455576863,0.7037051786665177,
 0.911801980414141,0.6951489819980554,0.29441096417574164,0.8383664659634017,
 0.692221217306285,0.2923089461961976,0.7172608930173326,0.9614942362092448,
 0.7219379088601513,0.5621602928955637,0.5200785525304361,0.4840368873703349,
 0.38391452092398715,0.23375738422365233,0.04709364052419407,
 0.6790457126197137,0.7194608789274114,0.6544786966954014,0.45145655600759727,
 0.8215056804915195,0.8147796034528394,0.5656791230752866,0.4523895587229404,
 0.7107137499669854,0.681815519986468,0.7595332208500409,0.04542343822807238,
 0.7825913812684998,0.6217339503674844,0.04815696421040816,0.6687333154832782,
 0.47629402832434775,0.7928325233158793,0.3926699477983051,0.4224665558383457,
 0.937723986433549,0.21702946748726926,0.6085669581151887,0.6679027534868051,
 0.6071763165301588,0.6612869552571691,0.9659116377746887,0.25599214668614734,
 0.9369646556747486,0.047501863166734104,0.4717316319689534,
 0.26723113684279376,0.7537107893507888,0.05159252253307711,
 0.4599547218421761,0.6052298835363094,0.08866226770586638,0.9980520864180941,
 0.049830722729714894,0.4729011934016596,0.6730410611627777,
 0.46872227053626847,0.27078709968660075,0.9134081466179843,
 0.8194139875849045,0.44998745549249297,0.7421760869031211,
 0.08075196653650851,0.9216538791080383,0.4252549961490877,0.4752244344906633,
 0.7325782790407828,0.08075063278932482,0.2986041840506729,0.5942334233394135,
 0.14556407556533013,0.8074971729608806,0.26403563926583407,0.613618533154747,
 0.45860498094233604,0.2277137166313563,0.6571266183652863,0.3112909540730836,
 0.7606540838330781,0.5468463593429078,0.3194050684114248,0.6603242443085228,
 0.11583578090038171,0.7017626794019473,0.38060210669833183,
 0.6706454159854126,0.6111109759164997,0.7965610307217622,0.847188453903062,
 0.47217496346738197,0.39542471453665706,0.5853541630115632,
 0.5247761170543721,0.7050837139807077,0.8421105026631008,0.8197983199476144,
 0.6911557435135234,0.6704157089524534,0.6824776040400141,0.8852198147411587,
 0.2781459192145852,0.12073354424651017,0.967560093302195,0.26799877384875825,
 0.2917069075141122,0.3477456887588297,0.6708379196558036,0.6644076464465009,
 0.7973346636241786,0.7292252202999709,0.4765192606214971,0.20574054857558155,
 0.6692941679086285,0.1836460482554586,0.3947728111886726,0.24000477937299092,
 0.08510365221143157,0.9151510547686017,0.30537517807010695,
 0.9330039186511445,0.6969891329120931,0.22282707056767814,0.3325539263967704,
 0.3597424115946244,0.8317699579760011,0.11056953316416929,0.544365981225438,
 0.8712633647527918,0.7424108831902826,0.7894580587454365,0.7770286368807883,
 0.7497665816364698,0.37654522558108605,0.9315507221052132,
 0.03476580158952214,0.31228105072485124,0.3667040651158703,
 0.0994761231784348,0.8519883654546422,0.5059311065276889,0.2577850499078197,
 0.4057866757658959,0.08800806888361945,0.7204452684713588,
 0.23866354591966432,0.3417118494035123,0.48221978955052636,
 0.16635410730061062,0.9037715208665564,0.5782271279340141,0.6032173043736927,
 0.9580304202702015,0.348382599271606,0.7299074438573402,0.1812497852385896,
 0.28768996245576006,0.7904604632549609,0.8483818957970625,0.2485314134247124,
 0.6957948840136505,0.6602923417868908,0.9209548038812204,0.34827734543827726,
 0.2763888486275169,0.8288986583172508,0.07223196956934053,0.3634514997136138,
 0.882448389623636,0.6224702660073744,0.9030958973098335,0.7900975498351841,
 0.6130805797356352,0.09677691646319253,0.8985484341074784,0.5579233427605456,
 0.2689308720174627,0.8608814350027776,0.5880684926812438,0.0849372826003455,
 0.33435312822247576,0.7306968727046703,0.15915313706449974,
 0.9564044703428838,0.7263364101982362,0.9142853057799285,0.4589849127244864,
 0.8082828526064374,0.796886554309254,0.3258377100483265,0.1676071756906764,
 0.7518515701011742,0.7863068329382995,0.7494661663224693,0.2915061806657263,
 0.8366999762054212,0.0514212709606543,0.7031173014026123,0.5961837518462064,
 0.41420966306408236,0.7365232853858331,0.2070276412319617,0.9649599605201973,
 0.6590058562996708,0.016666795655251643,0.3499772084366153,
 0.41779570950389444,0.6135362202696136,0.08263149411797233,
 0.25621516877032735,0.26527564034002904,0.5340520839997768,
 0.9984302578483292,0.6369630598436433,0.5220196112431049,0.21732175195424785,
 0.8521137307755924,0.4580004644380652,0.4729530928630632,0.9175588296733979,
 0.9185625947815784,0.8278211779029478,0.8673013715914492,0.43074785444855046,
 0.9215205538055382,0.06619334532513199,0.36596972114120396,0.553590239115889,
 0.8176903137340237,0.3778558119565373,0.6641870908431954,0.2982334681240494,
 0.8698921127029492,0.738727316501049,0.25126135214074086,0.7801600320842295,
 0.4554844194068406,0.22174534987935002,0.8636359661508778,0.5938111022493034,
 0.9933614501162424,0.3416921594622945,0.6943055099887714,0.08272579972185512,
 0.5934286734280709,0.1528158666823568,0.4868143278438186,0.7631924717383275,
 0.5227791400628111,0.32694616515754416,0.26522925718277046,
 0.8915180344809976,0.0469811140237506,0.3449089634260245,0.2403909383056949,
 0.707875102722155,0.019692642729762344,0.36319901806300614,
 0.4261085519338855,0.058102216071287516,0.7529767629082255,
 0.7204159469952777,0.19531510363767157,0.20531013222124772,0.699622674006089,
 0.8302602468667264,0.22345568447199504,0.482047358298472,0.2134308860328853,
 0.02205285446751326,0.027819500708235356,0.09931420444539962,
 0.13254979284307178,0.9594571035355817,0.45113920214322156,
 0.9816353885492755,0.3487919058184976,0.5400321338973842,0.7415590347015528,
 0.3695176546516382,0.7669773417605324,0.7333368823624489,0.5315317267027797,
 0.23595079141656394,0.4486285084569218,0.33055673594935076,0.362197716264395,
 0.8529814655517589,0.4783717809742706,0.7948406969046303,0.8661169154917499,
 0.8427229588228016,0.32712983675716834,0.710471071200404,0.6135816057067577,
 0.19554489104816153,0.9960320447785552,0.5790766365632303,0.8833272550502221,
 0.5160581854912785,0.47691215662947184,0.08441290983148775,
 0.6873421096989065,0.5324990780745081,0.19580364899806657,
 0.13209508583958152,0.7118363973113162,0.9509386700301916,
 0.06561548835643549,0.18630973755639324,0.7581271198801272,
 0.6033708487731508,0.9070009411184221,0.3710295289326604,0.11020604694150982,
 0.24285016341553767,0.7988030818428994,0.23921360661640667,
 0.4750422530262769,0.22157075647113667,0.12439773499704043,
 0.9969172092980768,0.3108064279570605,0.22648880739760935,0.4542467905947595,
 0.5241727465056281,0.15210171371886572,0.3291655810628965,0.605076289870346,
 0.8201437989456746,0.6628716643246766,0.16685255868681337,0.4537048261499994,
 0.03723291066209988,0.7757847362316533,0.5208146765164037,0.9169415414399488,
 0.8275096643982647,0.9864909299424995,0.08686550753118705,0.8651375226610611,
 0.3428236415331323,0.4685644092572321,0.9171005046985888,0.32103719032873435,
 0.6275357399050274,0.22101372358563476,0.01006138896008213,
 0.6525676661924323,0.5067792543879436,0.059933287850206574,
 0.6401205197764412,0.7064951524979498,0.7483257074016731,0.9374584556544674,
 0.9169858742125392,0.35717121749064396,0.5803053622172647,0.1441937087040489,
 0.07024082995115233,0.8417602190816071,0.3918157677659866,0.7168496251651736,
 0.4668004333271284,0.05360116069041354,0.9174092118048157,0.6237457442634555,
 0.2863423120316716,0.9174942014046141,0.8670311541078707,0.5320851977676905,
 0.9749609031878911,0.3701724770089624,0.8014874703139844,0.906634270063642,
 0.9446549029515616,0.5511558697239066,0.9616354152414928,0.622766007426376,
 0.6851734718990927,0.6039573324735221,0.1766547501474014,0.4358251719222622,
 0.4259492576406092,0.8534620618131172,0.606985465553964,0.05634705517580474,
 0.969910184155477,0.3601823699094735,0.5922710968593105,0.7590438286667299,
 0.8810278261329324,0.28771491462119303,0.582855438033114,0.7170641526534894,
 0.4955489756571143,0.3995376890977985,0.8356286753856369,0.4846148112385529,
 0.8436529146896561,0.015041224700462985,0.19249385063385582,
 0.10132604612970497,0.6644462167276783,0.9149369165547729,0.6079340902562368,
 0.29619911463020077,0.1551599539011792,0.03737887826046882,
 0.5265251286382211,0.2532350148090672,0.34885467923843405,0.746283906283087,
 0.712860001485796,0.34862306579091873,0.2503628743912457,0.4648142803486016,
 0.9981151377988176,0.6740893615226925,0.5895375991510403,0.5429878728334412,
 0.5383048214296864,0.2822429813174008,0.5408409526458904,0.7198001764824788,
 0.976481130163993,0.6801428366262928,0.2394549529514871,0.9279744779552725,
 0.5299372726632814,0.83353523603696,0.4399289361824723,0.2058751650900168,
 0.458302854305099,0.7725022159973449,0.5108837880503039,0.14692700771092537,
 0.3832619544944036,0.8267068691434424,0.013750342494747825,
 0.8787720586767765,0.46058182201028375,0.013872984613538697,
 0.5531585326978927,0.6101516201475313,0.19852710021775466,
 0.11272335235218067,0.149509499217777,0.5330066567666057,0.40854573149803086,
 0.4021708421321821,0.49453446545151525,0.6682089870936886,0.8561447659640397,
 0.4121994477518711,0.830687222208557,0.15140065788535295,0.6966337967881452,
 0.9093851079712424,0.9133743712221011,0.3285448894652665,0.45689576805363374,
 0.8319206716812529,0.46602492165765896,0.8670027523421321,0.5538444150488364,
 0.9632867906535019,0.6635394342399643,0.08647011073791988,0.8783569632526083,
 0.3127277396186998,0.7650634599090426,0.562604594992645,0.5032707296030272,
 0.3702245228726394,0.7911430170502549,0.3247452431205726,0.6712099544127692,
 0.5475333309491462,0.6131108650560931,0.325239236197693,0.5159093361528441,
 0.8464197198614948,0.4306219835345767,0.9441282528651975,0.6863798600109186,
 0.34492632177589666,0.2906683081267104,0.8839078765907744,0.728446779442997,
 0.9843750448520916,0.6263243747591614,0.47864896928326073,0.882208681396617,
 0.7675197155466373,0.06789459095638095,0.3810854294382797,0.9385562239940242,
 0.522959503168933,0.1995585753294461,0.08697516244087478,0.8358448105678636,
 0.8921103759287711,0.10451800798311384,0.5186577407454408,0.0679867149017147,
 0.26427608117277923,0.43660370255959835,0.9038061120960363,0.653948046279289,
 0.7793818863858115,0.4671204832328093,0.6675941586164726,0.7011586532529154,
 0.7738532363338319,0.2211794223076673,0.4019600672203838,0.23641449176347862,
 0.10125848505723589,0.3502026604708237,0.8694867921795859,0.5354847287478356,
 0.22893447215874407,0.4268150785782504,0.8124813527334827,0.602867170013728,
 0.0253689104115431,0.7913018204002836,0.9843467194340554,0.7980764575495297,
 0.8562126446782629,0.3731912135234573,0.48566510212355984,0.6031470866359002,
 0.6769015499898964,0.20233732659602277,0.6391320616406332,0.7683575474007311,
 0.956033699646713,0.6963810552119604,0.02775491375446515,0.8186249141698682,
 0.01958610899335511,0.3681367626100479,0.37023478897423456,
 0.8021361663332317,0.7194301836353826,0.35145387673064343,0.1688481242762051,
 0.4696111142690065,0.0969047749151627,0.6955562906712431,0.6346450577922583,
 0.4106739223079783,0.5942540737548205,0.633142636378951,0.9544927531753387,
 0.9046384305165037,0.8865499164640782,0.9529407201020562,0.4038687539130852,
 0.5102714042054491,0.24014147182388745,0.5069037573745102,0.3875406075796842,
 0.18356760344120338,0.3388951873419237,0.680327910821482,0.872780361346282,
 0.17772972606929704,0.1441625923692893,0.8872326801405607,0.6949244585937401,
 0.25420024079681514,0.8399739616134821,0.7639870025493457,0.6924086953947757,
 0.4990506524918784,0.10175369007598767,0.08212547456101904,
 0.9594928242363823,0.032548690272067526,0.828528999990711,
 0.20674811547228722,0.5107001351364416,0.6176775302481324,
 0.20738950216442853,0.21976508269131967,0.9999262940228802,
 0.8330191180008399,0.6118984211689222,0.8192430484737233,0.0428551540956823,
 0.8516782631808014,0.5194791747247507,0.824243298318831,0.1516416751578502,
 0.4326430568608356,0.8339452092536386,0.06500416235205486,
 0.12589284326685846,0.3263779811114892,0.6901981084146812,0.8719314978345786,
 0.03122724749826622,0.5966481194078328,0.5404890250822383,0.8715886924485666,
 0.21172615934280214,0.793428105578914,0.79258110111389,0.5850277099975019,
 0.9151869609713861,0.7666816444564525,0.5434179901850447,0.5774567192096788,
 0.5199918116006437,0.18837322602744566,0.5458840288848861,0.9405610465433665,
 0.06507981672805685,0.5703382671978,0.223704299618817,0.6404431514402456,
 0.9854144094018997,0.3450957992129553,0.8536605944004483,0.10672205669129475,
 0.0333803644825581,0.5570432921344314,0.9623802050553167,0.5689242923840401,
 0.35477185150491464,0.14124304922109987,0.059038581940083046,
 0.35568012215797395,0.38571331026890987,0.35976989311452057,
 0.6631515140586287,0.6224171701473231,0.6303153969800531,0.8979014745793211,
 0.6373030954545914,0.37438686812917676,0.8094590910693517,
 0.08188980081425978,0.19322774104550366,0.7255963112309781,
 0.5561896787134434,0.14609464090521973,0.6604663694661279,0.4348700707752001,
 0.6343513181487862,0.6525878370027349,0.828602496667635,0.390543801550136,
 0.39487719983303227,0.4867727670085126,0.4196159108004044,0.4911414672353891,
 0.37263292884215937,0.008256516851657625,0.47778822322687065,
 0.4739787312042739,0.6140072051034633,0.6458786673312306,0.12306594095002543,
 0.0867489763704008,0.6603838932034509,0.8739056376427677,0.3836585128929393,
 0.8223017299096177,0.834765382378333,0.26156285292725123,0.7359871489263546,
 0.7041856920224215,0.07963503721548504,0.9055670633482387,0.4779893896051589,
 0.9074031723757766,0.09929864812590228,0.6129238647859587,
 0.03917006180286087,0.22713531143835253,0.05669203597982442,
 0.1734720302345859,0.8982349513115151,0.41402993658752874,0.727560036571801,
 0.6707191308875702,0.01810965330740011,0.14946958085135542,
 0.7491179187791059,0.17162418552504588,0.6492143467857,0.1960698717827647,
 0.28313093897569996,0.2838518758122651,0.7772155640021948,0.5699797903553541,
 0.42702497235862635,0.296542510753118,0.6680852371360274,0.8713876701644303,
 0.9173526640693732,0.9130945693824092,0.8509014751004939,0.4279481745121627,
 0.5913093936697285,0.7170764276316672,0.26556840064531784,0.5281562008639497,
 0.579292673255705,0.8310084451609093,0.049385442317228745,
 0.29760045483904296,0.9473663560349229,0.5662163166746697,0.4195563380580605,
 0.7068080577511392,0.40638537895518123,0.4658066428122638,0.5278225961349796,
 0.4897723335657722,0.7674563458389091,0.6754045406700588,0.6866477128669161,
 0.6901803623767652,0.9981058779163672,0.28668611200025096,0.9367549542330421,
 0.6111190239638917,0.1874535046159015,0.4849553532365377,0.08983020366903183,
 0.28380934318613915,0.08642278310086593,0.8231526822833592,
 0.3827822810936047,0.7069852905744122,0.0130049399099601,
 0.019634545030890682,0.653003801104609,0.6069658661107976,0.4586248376056913,
 0.12854344225774206,0.2228114345311767,0.3449280960555434,
 0.49987026016220304,0.12624038572893803,0.6632975525224512,
 0.5005768233683021,0.24735110801021087,0.18961148839913444,0.480983433744263,
 0.7362920159719868,0.6232961749056108,0.6864646211255183,0.3915213205857002,
 0.6554994342743061,0.4068944371736819,0.09166644310034133,0.6501010059090214,
 0.2137236853310024,0.05635509364367952,0.5366090676198979,0.8006763422789029,
 0.27946387096678893,0.1541018463192234,0.20787805626635925,
 0.5421151747757341,0.0778444481415248,0.6879155526702302,0.05807345204165093,
 0.7864599273642344,0.8360524613525742,0.03793492866114001,
 0.05402660373965218,0.8404362698525852,0.9572738846914282,
 0.10944540885151177,0.7010673679323199,0.991158381596408,0.7487223406147163,
 0.7846200605834972,0.35913540838391333,0.9367871691701546,
 0.41821754094136065,0.7533835008055703,0.4048850453339816,0.5112654878975039,
 0.23926252347470756,0.8318080472371452,0.7172952911489988,0.6468237026926942,
 0.7672549207611806,0.8040222015929867,0.4966771978152844,0.27097741204133163,
 0.3899022504604277,0.7249553187110072,0.27329822855363517,0.7743245892470353,
 0.6535920607060003,0.6609086000587643,0.9921864097768014,0.3644461917009717,
 0.29343692342474914,0.35134132934604545,0.8104965513579497,
 0.5354036022495741,0.95755262663019,0.9852037119605302,0.9570220707141152,
 0.9676255258449391,0.8444798378792269,0.9871302771084064,0.2799055633361267,
 0.0997847287923549,0.009277519116827726,0.49786728466622077,
 0.6321216471044151,0.4429914047990481,0.2369672944858331,0.9787736885630774,
 0.48821035944690416,0.9868842273601616,0.10062663275252592,
 0.6276825186316185,0.6983593774344178,0.21174839765806575,0.2070542165376683,
 0.2998439462602638,0.7539968507517051,0.2783030794538687,0.3383988463607931,
 0.29407899531053594,0.16223886429684198,0.22815708117679734,
 0.9870527141421134,0.26272410515150924,0.4801967289914618,0.85171301163888,
 0.5399073288673548,0.3669825622556184,0.6880744132956796,0.0377622788900438,
 0.7843146098953423,0.41285290972107846,0.11458553758554146,
 0.8091173613069209,0.06396746535194586,0.9276230868511145,0.6484684905534981,
 0.6620768429995438,0.22089238303844938,0.11586919305754506,
 0.061120341401968314,0.057092844196030446,0.3474207204540457,
 0.5684104377444547,0.7438498437244394,0.2817484897584316,0.5250677304847453,
 0.992589929288737,0.9087567119118305,0.2582291140262013,0.6708328013284801,
 0.22872945894372176,0.5357916612153094,0.0108656086829787,0.7185085630535458,
 0.8836716600616505,0.9218752865351105,0.07231788757744695,0.6316829420845402,
 0.020854470348595555,0.6888559701951993,0.06779659065605204,
 0.13508151999145934,0.8396149510531938,0.8385290646722051,
 0.19821000093628982,0.7752832978304733,0.7168305401522429,0.6539037595060464,
 0.6408720296579001,0.4157888109673724,0.617003062962629,0.10753135215047671,
 0.21096779092367646,0.4723387717368275,0.390508378355945,0.21652206419117082,
 0.4685056129973426,0.9953714504988822,0.005584809615375663,
 0.8378453715995265,0.7719035423311937,0.20043549529749782,0.4608420128802373,
 0.5941194371425335,0.885134245209118,0.6705751896563583,0.6700867809657689,
 0.9666441729210686,0.9915147372407503,0.7516531874032442,0.5693129192651317,
 0.3327370430471063,0.06423240952134146,0.8555210807323769,0.8898685891129022,
 0.8816049476041681,0.039983386172394164,0.7963995285773061,
 0.27106103178194985,0.793689506395244,0.5914889029080369,0.4130808584440454,
 0.1827075440205571,0.5298269448941343,0.7313784002907692,0.8767383548237397,
 0.4626744326105118,0.8226435878220575,0.3134996314284336,0.4215232037268384,
 0.6043389335430779,0.6255880385590894,0.79799848947396,0.5177288931103597,
 0.680902806205812,0.753247294645116,0.5172478017062987,0.38616372420803435,
 0.8894496827033924,0.4801015838141437,0.19599450405278918,0.5977862375204543,
 0.7781327137614504,0.3835757525994379,0.645836188417698,0.6782737704066415,
 0.605268141647087,0.5228321614140682,0.032549793806976424,0.9900143384485167,
 0.02465491499246475,0.14383927133461283,0.7328829760078143,
 0.009762489505754202,0.8746516414640675,0.2711140294419132,
 0.9340793326712816,0.6697853233340274,0.9828658662569577,
 0.043487482458654636,0.0959154849588354,0.5634606600741643,
 0.6695374370232186,0.42567086336431204,0.29169870306671797,
 0.1702504408656793,0.8558712920603174,0.4588455960643377,0.31105665443918684,
 0.7498713409964433,0.8991273550061797,0.03442438859067687,
 0.07911642609516023,0.28354729633490916,0.5455494622911489,
 0.9401478092033411,0.37874460249906605,0.76486282031566,0.930734216606879,
 0.7031812103165531,0.2587106167007567,0.8834918256862186,0.6080696880131542,
 0.7853778265651222,0.7295371616387204,0.8874174403744155,0.2630954212064116,
 0.12511384883555365,0.13413669684910123,0.6387526211274426,
 0.2448486197906421,0.5420457032312616,0.5759225276081171,0.13355079239085677,
 0.7579859284812089,0.09025505661555544,0.7048307666443074,0.8981390677807111,
 0.21577330608868828,0.6768824810611085,0.28542869044836583,
 0.8498626938241215,0.1341100040942348,0.9062393561222182,0.3474072644969388,
 0.10175058440948415,0.5763733102930767,0.37218194864352583,
 0.4491255561251951,0.4524724754450157,0.10699690533239414,0.9257790086242697,
 0.9742069331333918,0.7626216991896467,0.8769628028245079,0.8045216548843874,
 0.9788363892528464,0.9157911490374857,0.5931044950803881,0.5650408449915604,
 0.9982581423868421,0.9330957656060626,0.3717650783541697,0.6348601906867802,
 0.565840263543677,0.30022877379224566,0.08604220779288518,0.8932707576336898,
 0.0516379737443764,0.6309286984795455,0.27256837486219576,
 0.37031292097722834,0.6517006316531402,0.6839387886627634,0.9387278863640665,
 0.7364411371331271,0.6078261960431578,0.5344142634764716,0.22063851622292252,
 0.10921676822149595,0.15572541527355377,0.257622262900269,
 0.15656855129182434,0.9478665183984003,0.6522971367301335,0.9780992184847392,
 0.38893969489055136,0.5276014725343585,0.4243433506959473,0.3736435473318531,
 0.3441378663835347,0.4866318673897768,0.009841072722884703,
 0.26920158289318075,0.6222185431808738,0.4848742119334254,0.9128686634084194,
 0.6914380414767487,0.7078222365927581,0.5645914062044384,0.37712959442005445,
 0.7966499090958943,0.47438644677111386,0.6904205397643803,0.9693975897246058,
 0.11440615111850716,0.9193370059352438,0.30098451618564004,
 0.7901280774237038,0.3485278217557013,0.741364862131299,0.40177516667081004,
 0.3248660121421578,0.8563339147678837,0.6500376952897965,0.12824612683259007,
 0.7034255713299851,0.1280167100363192,0.22629459290580345,0.6034547618092473,
 0.3349223655302028,0.17106637137930414,0.5488724167344776,
 0.012083943618160786,0.6961312219467763,0.0209894798525595,
 0.8647208854091243,0.43577075550629674,0.3197148826111089,0.6282815916056392,
 0.6866732752940145,0.5328627589594881,0.07014097571445976,
 0.07036915374498842,0.24271448076989954,0.8767125960059626,
 0.7316445549901434,0.4410720313090546,0.9306104959128485,0.44030303753893585,
 0.18911625723644065,0.18809921249342654,0.7707058386061445,
 0.08611377505875295,0.09771417894684642,0.7428502718080814,
 0.18078563397803893,0.4938398043138552,0.5121702749517394,0.8206535064593313,
 0.7861935590284007,0.3809229599315981,0.7310792389154054,0.29397087795418997,
 0.821580239690442,0.027217156126316477,0.761949372710519,0.5579065996070777,
 0.8188203647831499,0.22200574362052217,0.9014338094088022,0.7934254188293649,
 0.8715009922412545,0.9441934863612449,0.8562718258032131,0.4564560228902853,
 0.4261641988482301,0.22186851563807786,0.5277055022698434,
 0.19283789751966307,0.4643457015586888,0.0013532262299567055,
 0.8963589423598203,0.6714740128991746,0.6556869068713888,0.752332835397469,
 0.5853887992578315,0.319218909743075,0.2297628417183123,0.48172871957266516,
 0.8564329400470642,0.42938228397843303,0.2998767235472555,
 0.20291969013003408,0.6032728380558671,0.8839620069133118,0.7270800260971106,
 0.6080918606308064,0.12230432950994774,0.555241778117608,0.3762197166510185,
 0.9421570526564076,0.3083559074868827,0.5264176417204401,0.20865098643355695,
 0.27472061596280095,0.5122548683284424,0.14188741359085522,
 0.3507427881665296,0.5200436470683231,0.8896176875111268,0.674444442779671,
 0.9881619752403598,0.17435482957505666,0.839245025323838,0.11904521934076273,
 0.28437716998217377,0.4624125229331435,0.5316499079353241,
 0.18630961797534207,0.9416125288482912,0.2881319937708269,0.6289423608928125,
 0.8619802273011649,0.2996623414658033,0.8179206431471578,0.3582402371601142,
 0.7359416599626646,0.47979368870451167,0.9954180688910037,0.9046654226510524,
 0.8170769287729068,0.4268503835075558,0.40515161274395095,0.9888066699591452,
 0.9822374835956741,0.28422906546903715,0.3737300525866284,0.5071145312996467,
 0.9259007637741168,0.9357294101959821,0.36021390343817283,0.3808415069278378,
 0.48383585474831103,0.2278306367449663,0.4668019282885145,0.5279375306681404,
 0.40519838214908277,0.788002476582971,0.7991747123436481,0.5132292327594151,
 0.6850623536091065,0.6212600202355345,0.8779039877668025,0.5092925165199422,
 0.5960601100849949,0.11155323362017677,0.6043502525288961,0.740918648133989,
 0.9988701340330396,0.7616472892765717,0.5861996508509216,0.24213379630751408,
 0.18482065720545815,0.03708856115550585,0.7797809188926489,
 0.8466194788392015,0.7702711598061963,0.6060381360650605,0.15102707616982425,
 0.62737213442335,0.805909315046324,0.6695307290287035,0.2110365225570997,
 0.40379518666346437,0.35398821434319205,0.23159143381440805,
 0.9751192507883707,0.6920842319964192,0.28861572476186415,0.7749913865242718,
 0.07731075235828244,0.7161734090091573,0.8896254906125812,0.7706000273009638,
 0.07585966648059916,0.677219148727865,0.11375087669207762,0.260297963065266,
 0.021929565696985787,0.8244512608120311,0.45692045251997704,
 0.9192186370911878,0.7953122251292596,0.38794717747407226,0.9661536641437001,
 0.6168276806804621,0.2305578673572477,0.9398023565857815,0.6625791024725091,
 0.5801299864513627,0.13272186509841266,0.8683183028745606,
 0.15355752661557476,0.1936526929848449,0.8109509422438606,0.7263124976091279,
 0.544446371591639,0.9414626416110973,0.4944645004634718,0.6466037646168787,
 0.29769151968450613,0.2467804554603289,0.16322765160796393,
 0.24875534603188043,0.7513384935107326,0.01155768248880329,
 0.2255273318834703,0.7033208492220995,0.5497305283570975,0.2805844879214271,
 0.8171282630056959,0.8967856044068714,0.35492051233049426,0.920365101274648,
 0.22620067777571307,0.9509758081419926,0.9132357040282564,
 0.022365419348453153,0.423975527767702,0.37697323800507676,
 0.6938745251716124,0.8398769330988004,0.3178164736306138,0.11525329679608409,
 0.7208071158800453,0.42072159676869747,0.1744395692749483,
 0.008097288619186793,0.9761383312963431,0.14786102285480263,
 0.7753291652703382,0.9200644640012915,0.7005904930972789,0.31880098917623956,
 0.45590369189559365,0.8007070267086801,0.20402548077865656,
 0.5996695573618972,0.259300186846794,0.857796711983284,0.4970536017087356,
 0.8388622435496156,0.6580195715824521,0.40800785378974436,
 0.04435509226548762,0.5167325302289114,0.8017134503277085,
 0.22791906356216707,0.7570878582991152,0.9899661963106384,0.5572293536893963,
 0.22072691261837707,0.996560811107562,0.9457422294258899,0.5420431675576982,
 0.6130465968374829,0.6996053292776963,0.1691614788145105,0.4900088871236259,
 0.13057421420162207,0.1284680601719711,0.7025354551750431,0.6193469669662014,
 0.008595530052125211,0.23648187477860017,0.4354520774054449,
 0.624169655485642,0.8353222982827178,0.41405021077562265,0.3906938609816971,
 0.9144988032542299,0.6449377605576075,0.7367553091644331,0.66745747588704,
 0.9931012997578021,0.24753641577490426,0.07628206185837327,
 0.12748592618703658,0.9204472769361952,0.7622112826449616,
 0.058416443073331825,0.7584882696300719,0.8988544366836421,
 0.04232874236803996,0.926160362930935,0.15160689718657194,0.9111724470064453,
 0.560227302263636,0.1650018819388771,0.17667707300427193,0.40492833316064125,
 0.5896984741381195,0.6820522231009607,0.04920555112318081,0.4475123603539286,
 0.8573062999413987,0.563506846291916,0.6183561687331408,0.5464605433091718,
 0.2701403836066396,0.6651903897198832,0.9777223547773178,0.03554942608975775,
 0.6607289792818677,0.4486777438739171,0.4526273041474499,0.23935966998126768,
 0.10649875310402512,0.42682595447312066,0.72181472363734,0.6566138721204359,
 0.6333941609104451,0.4927501408711211,0.7586163367470844,
 0.022124174633001203,0.6690444597324571,0.10936386346983262,
 0.11660149598553571,0.673564875639422,0.09086545772577992,
 0.08475198517636695,0.26544440014246806,0.3184424569481594,
 0.3785439290829635,0.4098353378513897,0.6320026522548484,0.8681131961769947,
 0.5263610241314904,0.075467983303666,0.7651884618665912,0.1733931178015108,
 0.8818668844006631,0.362176731643875,0.7683427118683602,0.14099372974227464,
 0.44972134206525405,0.7281774744676095,0.5582334382661096,0.4860694954602004,
 0.45186017764935194,0.7459652594823398,0.5633349267355746,
 0.04071940638533855,0.6977313069400274,0.48201731091761013,
 0.5486164476484781,0.08319467077748188,0.37908830909682556,
 0.7108538390526361,0.09705956497673607,0.1247810708250976,0.9876262919743288,
 0.5874567128881851,0.6973913004553797,0.9439997142979112,0.3114139085229223,
 0.8444186852849479,0.5462408072997221,0.2400340719743791,0.9330807712611432,
 0.15925272659791057,0.44306576535583464,0.7768270598120316,
 0.2821539666781385,0.7679872168061462,0.4751956245960144,0.9871376724329914,
 0.28765766102725865,0.8145227810542857,0.5206871645854124,
 0.14677335831428118,0.8577697697879066,0.8598724127639645,0.5812293297771041,
 0.9494770438595809,0.880661142169917,0.5855411488926234,0.5587724853223888,
 0.4155967471069887,0.10460352503224046,0.9759472588064542,0.1002777085047819,
 0.33857980916590447,0.294552893882421,0.39185697491029925,
 0.047226091471856635,0.5231715447211754,0.3151957225789639,0.386457053180552,
 0.9032218045723175,0.4361677744613667,0.15686950797527677,0.0178769808894077,
 0.5356893813376683,0.4614033860613447,0.6020403303719433,
 0.020611726917443618,0.4348636518072332,0.6976672007057975,
 0.2137143917410239,0.3984271914372689,0.5027714029846483,0.47198702420894634,
 0.8096905611170313,0.8770928217547044,0.20142032252807496,
 0.36505795538499775,0.38010618941375596,0.46826798541032355,
 0.3844897382043325,0.7377676400015376,0.7321709913954642,
 0.021572493018249594,0.9767826665754707,0.9775400292406463,
 0.0455154067046587,0.33589134930312115,0.40606847831916304,
 0.0811680511193813,0.11778747717647886,0.12122691422712928,
 0.44685109921530175,0.15547672839809545,0.42258507971218384,
 0.3966483890110035,0.09018674624048284,0.1728366597457356,0.9513600529539048,
 0.25635252710926915,0.6507258768676362,0.665177515382309,0.5956919693163559,
 0.4487001258553427,0.5858796333137397,0.8512818157705301,0.7952912073962995,
 0.28182007621984795,2.5342708207720666e-4,0.5176444267593816,
 0.007895559140827624,0.5396293595757365,0.8579612536621518,
 0.12192812726719049,0.3538340737518699,0.10932374475464024,
 0.0684524633146647,0.5066669243762678,0.34518728688867417,
 0.21477661155306538,0.9151009921501344,0.48667718796423176,0.693530122399024,
 0.8519482969323018,0.46793771882672075,0.3231235722604249,0.2321989598305425,
 0.03145272006386324,0.7837258336037212,0.7924636714391886,
 0.12321598349608875,0.6662325685182184,0.19824497914433092,
 0.4308354673760081,0.09078180724233409,0.41655265490693594,
 0.3538921758343675,0.42638592734403213,0.2590574978909951,0.7570011708657846,
 0.8322266059394577,0.46345122854795484,0.5174905254717493,0.2806338677639366,
 0.2805292596210994,0.09462661967700525,0.09011317522561535,
 0.5388192294389889,0.14066786679698384,0.18048803633630994,
 0.22924178479278357,0.4794886738530415,0.016587616955919482,
 0.419557271957005,0.8853308287090849,0.4900547892756686,0.8310219745291186,
 0.8379060479327646,0.15727853510227596,0.831728720213756,0.5753392141420846,
 0.5686061290341496,0.3642095154369763,0.41718965695735866,0.6803774665764024,
 0.9385189699646913,0.5493286875113513,0.21620709607179844,0.1888730670230936,
 0.545803203514099,0.9639863021370002,0.8539319406438656,0.1644208759802267,
 0.9064724430018967,0.8224989355009695,0.8485093235312282,
 0.061481990784694185,0.7167829235255911,0.4361983004718434,0.329448063385528,
 0.9450691625803171,0.7488797519022047,0.7527563996164348,
 0.041058961224540114,0.42977670224685727,0.7204897185440183,
 0.38798028400796225,0.07337124601772615,0.8786473790373068,
 0.5256892425718265,0.399982125237207,0.613863089289094,0.3564992620141916,
 0.051040499506860626,0.9218105085590481,0.5053739998174794,
 0.4194916352113355,0.8726916186459158,0.6056195589197358,0.2059847117171909,
 0.6000651420351276,0.2577561236038498,0.6543623945271809,0.8785473604848139,
 0.942095516221961,0.5052267548630398,0.33979059061789885,0.33518991313586655,
 0.6308888881726006,0.733718851765574,0.28359295306755194,0.743523840610826,
 0.43863927541051617,0.5008467718550456,0.888865756617864,0.3632702134488549,
 0.742770698704465,0.5864776669772355,0.7925030230004424,0.16038258734386746,
 0.605481110385849,0.9375595837981582,0.5884352988028188,0.8672714550286811,
 0.0037444881698053267,0.09470785061446174,0.45442988773630566,
 0.03653492937653524,0.7559270783505188,0.1699054928548075,0.8938188283267811,
 0.6434482815387691,0.4673609464990778,0.8643132343389168,0.3037607196381734,
 0.7944445331107252,0.442396980029929,0.3422648080819962,0.9500236304641866,
 0.5718560566867326,0.5918043793924304,0.8175638642437669,0.3846165186545577,
 0.8438659249014501,0.39163069472704837,0.4733080372385212,
 0.025404119345318055,0.47210154987192565,0.49651130987535974,
 0.5821957391256388,0.715387010956446,0.2509360403418285,0.6309370747898855,
 0.0875568222504618,0.13602733649818577,0.7801006346070143,
 0.39717996327241534,0.4627710767636951,0.22178830029206076,0.098980462527955,
 0.6431573693227853,0.6610355088155208,0.9151604814906128,0.2913575431310249,
 0.005994030733383671,0.4720693017680184,0.6884934791352948,
 0.09497067877104037,0.41930054152564833,0.6495879279328127,
 0.6173779464298366,0.40596410037516695,0.8548229642154659,0.1496512290832459,
 0.7969589309177874,0.10213564269365627,0.691151383640805,0.3100239141424971,
 0.08450424067562434,0.2982226598835327,0.12180931387867788,
 0.8950859931727712,0.8259182954899547,0.13442909418823779,
 0.26058922099503534,0.9858927521843883,0.9923855342271581,0.9463074070124357,
 0.19343357156286323,0.6430524999375824,0.8113415965624677,0.371703138530746,
 0.4053303477787762,0.34960424871398577,0.34015390851145844,
 0.28730165172514255,0.5709806293404187,0.408981335056669,0.6150018409760168,
 0.8678969182631402,0.4750171008316695,0.17824177128018004,0.363899443924684,
 0.9150217482871922,0.3705695206300228,0.523215659096783,0.6186635941412199,
 0.021306086398291235,0.2913005050283347,0.7772641996761085,
 0.5065430185158124,0.0710066478275454,0.8469075317892013,0.04649079763235775,
 0.9510925329236838,0.7010767914316501,0.5045646890702556,0.5523826788594992,
 0.7444994975925067,0.9717559546178708,0.4487535378248535,0.0335205362140788,
 0.20734684205234843,0.9094909450544562,0.20530447114022643,
 0.6741687028876534,0.6557500343824162,0.6909653266984723,0.3312757918379864,
 0.3542812332767432,0.4328502320732609,0.4363995176081693,0.5582013956341662,
 0.6861812126755008,0.08083394044004089,0.06882387318611982,
 0.5087909876807126,0.8965803650070897,0.4161097055779063,0.5134491692503136,
 0.6073369243240907,0.6163065054909665,0.6548755834657177,0.7577385377707007,
 0.6095418037096264,0.9466082924513738,0.5908629676230259,0.9959394860877986,
 0.09160108618486229,0.21595087514827482,0.6270273841696326,0.721577274017541,
 0.7461625015976658,0.4170775767789676,0.10647080747540527,
 0.12242731282778219,0.2567078977998374,0.08717393108347671,
 0.3466253902268748,0.7657006627802283,0.5426896125083036,0.9430721480981005,
 0.24688990368977293,0.42420534499012363,0.09472721433597486,
 0.3365668182562741,0.3746293694651932,0.6052211500754584,0.19596363806419137,
 0.5100678875666783,0.4496757157839575,0.6769101980510299,0.3414322822274918,
 0.12309061809836408,0.4934735306062943,0.35348339621889147,
 0.26427368482633495,0.2404371345229277,0.6554545152188849,0.612117842745018,
 0.3617864368749597,0.477960333174445,0.25381251626800827,0.5152261199129939,
 0.8114627050362629,0.17030089267577142,0.7262395735583912,
 0.43663641447429713,0.9804126620483312,0.6568148085679042,0.8259723425886899,
 0.5069997338235716,0.35420530419860063,0.5397927866898028,0.6610614386827172,
 0.11192067020292007,0.6641917233545109,0.6297269118261681,
 0.40151844917499835,0.9693976044996809,0.0325597892514764,
 0.11477961188300423,0.43785018190130365,0.6461073817290486,
 0.9894585007554303,0.45792597328532625,0.5236510504391118,0.7190073257367222,
 0.7855454524049506,0.4049023843897248,0.5432948620480518,0.8513887583766533,
 0.4699726971679332,0.03713643271171685,0.5255264549824462,0.6112737324177829,
 0.5799132066331092,0.08641041489561185,0.1414577926691919,0.9429592630227708,
 0.5120294875578764,0.548371116668967,0.7983483770219184,0.6794987646005565,
 0.8458991214427644,0.6125698269053704,0.4127678670938518,0.9650807636728745,
 0.3963841428223478,0.8847085476725812,0.25086818328642835,
 0.20779684908581053,0.9625468339668483,0.12484700889550104,
 0.8399027153574697,0.07403121471528595,0.35043131418772755,
 0.5820833055058101,0.6900294989927386,0.663449345967612,0.06593902020192832,
 0.6119874646833041,0.49032058712379634,0.3481841280317498,0.1758103281426574,
 0.739764170788329,0.5773011772975115,0.35375853593336115,0.8870779341159339,
 0.5606241561079962,0.3930162304651805,0.24344466003997434,0.8962615253145043,
 0.697238450912171,0.5811901700426937,0.9648352186098131,0.7695143305974881,
 0.799175323138851,0.7086528050441588,0.8689837866012771,0.5057234593125278,
 0.68995576770981,0.9826521157637986,0.922079888013999,0.12317871841906203,
 0.9900077070216863,0.11980276768366283,0.5822501015780233,0.9184307659263875,
 0.3823248709813143,0.010285701982437057,0.04421775897637559,
 0.2919001361411122,0.3324243373108977,0.8428560996547225,0.2518821685580065,
 0.9715510859098024,0.18599700358880022,0.382987344576315,0.2406941769866382,
 0.08178779113120893,0.49393029849676884,0.22349856600381246,
 0.6082863241021276,0.4479291907902303,0.9933832881534551,0.7062175078759686,
 0.31413111672897065,0.13860995292132738,0.07669216482660346,
 0.7038811400675875,0.17978526028654152,0.5285823045649919,
 0.31655306572356456,0.9888921352581408,0.8462805144565394,
 0.36018710176485536,0.3062939881211446,0.40179847403549995,
 0.8567279727746518,0.5371208668167224,0.1114170187219874,0.3166009392944962,
 0.9627954632985204,0.5988912529486148,0.9445153864090122,0.2923808504689531,
 0.8945344237060485,0.667888808799935,0.6685995747418347,0.17157327004369116,
 0.27324400154934836,0.4023200415130359,0.9290897894474338,0.6112947642960241,
 0.6445130598157686,0.4346189613511108,0.24593712523604205,0.2697529479400732,
 0.09657575329691548,0.5405831835689511,0.7587664278484015,
 0.24030459394464287,0.8311176708088659,0.38462668038721437,
 0.7320740657767267,0.2348929767857214,0.881472389696062,0.03233091951341582,
 0.1254914183554826,0.3580057229587802,0.8556688302534122,0.5457648887259047,
 0.5305049694793587,0.5881367360583935,0.7234798190120617,0.8899785830300171,
 0.95839620689758,0.30061865555412015,0.35775220030810795,0.5358017429860965,
 0.7203356259791787,0.46081761892707984,0.25914834612336,0.4451892648843436,
 0.37076505508889923,0.46818252656966697,0.04693603965878912,
 0.495891265952055,0.015031727057926991,0.1671948108661957,0.8184524823933341,
 0.6729894897057904,0.4526262885273745,0.9562590208502764,0.7938383810419535,
 0.7074222626248577,0.3180024263653636,0.6049330698685338,0.2648089391030668,
 0.6480288167816552,0.6801602475662485,0.21236946424419434,0.0730801011883172,
 0.5013649696728506,0.12081221836685385,0.46049718250735816,
 0.2670557684749709,0.8625265600410281,0.9104186208635366,0.16296557400175649,
 0.7723153238344196,0.9332867021268414,0.21950742411779156,
 0.13431020420234818,0.9900449719201194,0.6429444956574619,0.235786680636903,
 0.5766911108923489,0.48970662384263897,0.8968779809852196,0.9181841010085884,
 0.13572521414862437,0.7903092735152994,0.6509841167197593,0.525111220000324,
 0.3409619051902728,0.03734123367594355,0.237115804433919,0.41700649280819635,
 9.174732405121677e-4,0.7013275660666528,0.6366491121141308,
 0.8449417663428593,0.40384790863052444,0.2944253284819749,0.6879627841059663,
 0.6291683587169518,0.7136901025232757,0.7247390144810122,0.06426102613938112,
 0.39561291050011527,0.7003630729109113,0.5848736386121897,0.9622592456454762,
 0.8015648661642739,0.7747099377584155,0.7352090214014843,0.6161288217746956,
 0.7084229392657899,0.10055508156055148,0.32822471734677616,
 0.4240889175082959,0.866709284547174,0.3268538098529947,0.24627319724625674,
 0.892967559866687,0.0816264905890628,0.0844044815478735,0.5318507338602179,
 0.9107613423253231,0.895812908591379,0.9261875765879841,0.08999677138071949,
 0.14489637574741598,0.879874654850264,0.715326001417032,0.06782887693041784,
 0.2991522372030794,0.5861963609777986,0.4482094961545192,0.9973961757871849,
 0.7474133784604315,0.49987310165913224,0.802589504899308,0.582863687074225,
 0.22459987987743046,0.978558510441871,0.8577900271489615,0.11325567941865144,
 0.4175035846151277,0.3814610766433386,0.27773117168345385,0.5922669667390796,
 0.44281496879588333,0.9024855644950389,0.4937834289606282,0.5374972577019816,
 0.32637877657638603,0.8271874990997101,0.8322203089493988,0.4415819604803902,
 0.3621947863244953,0.42545845533017435,0.7835048115170087,
 0.45352230534782867,0.3574124007765225,0.978446325429033,0.02454823629662517,
 0.9339413031573575,0.009188284792143064,0.3906645325298522,
 0.7246416630720938,0.9392255056375723,0.2624220412767019,0.19584802804013457,
 0.051063737704855616,0.11064473302766253,0.5851697522194113,
 0.13627318305215796,0.27086628625128073,0.6930201318789808,
 0.07987988863086026,0.5893415256567665,0.723292439641607,0.19306339376040504,
 0.8334488206238404,0.4864950317929444,0.7578941103791879,0.5671389943215226,
 0.03197796169719935,0.8797277087201167,0.3206854315060522,
 0.008775971708799735,0.43418063605507584,0.7581676435128635,
 0.7969251395079208,0.6088852387459042,0.6118069747482718,0.6041351547233305,
 0.7145244013908398,0.5720793874393159,0.33194897397319223,0.4418574660051582,
 0.07955045966031804,0.026894998462041242,0.5259943969906569,
 0.7458330559754988,0.9150585679988783,0.3310815295073619,0.3707319405985555,
 0.9573772741894666,0.020469461135009692,0.6030768096789239,
 0.7742300738739805,0.6781064113369921,0.1893933341401577,0.3586614923031769,
 0.8564114895067036,0.3329436492738076,0.10625363607097227,
 0.22309330869420319,0.944010436233713,0.5354371293492579,0.6782098347698127,
 0.16194406308044484,0.2670623338675725,0.9962394746262603,0.6295142173431368,
 0.7825053905571149,0.8354076457484951,0.1644690447039835,0.07324212188917678,
 0.9475008528510588,0.8815293454072646,0.301205575614499,0.9327847165072287,
 0.15904679688749168,0.5266541596266902,0.5900126472997979,
 0.40036437353209786,0.6598855566752823,0.033586105785576637,
 0.38346002307354576,0.8679101295694324,0.13021474668141664,
 0.6428022180287611,0.3919546576051024,0.19681274188946463,0.3609003404006432,
 0.9773688062125673,0.7808343185438941,0.0414664587462199,0.40841165236824606,
 0.35848085248780903,0.5986441613426137,0.4587434362037017,0.841361502885817,
 0.40760570160681353,0.01640901248677984,0.38355961396300287,
 0.5324716378104017,0.9134677361297187,0.6455300145859455,0.642008974343417,
 0.4941880426296401,0.48652494235063437,0.9850940728753121,0.7276982296364815,
 0.7259454480648296,0.23107035530075726,0.8769017559207212,
 0.46708336766948844,0.18070340106822624,0.22898608588855562,
 0.8189498975819018,0.2653497519624286,0.9313590463374889,0.6467824206772995,
 0.42596388778328353,0.2460009979223381,0.3725356753286655,0.4903059717021966,
 0.5921859995671741,0.36386218383829727,0.08844531050147886,
 0.4279971261569634,0.9306199080290644,0.24303753908613013,0.5080614067306252,
 0.5695914664335942,0.027165424153171136,0.6838552723004293,
 0.24728591948523448,0.08601061893577788,0.9787101392436738,
 0.9707853121674821,0.5121205704378371,0.2505012168842241,0.34889440593939436,
 0.1642870180625814,0.5693293142884077,0.1741363646307431,0.6830047868518876,
 0.6737766872558419,0.2575332193564792,0.36398437800984196,
 0.20547402694801353,0.25795473961669513,0.37125274801072283,
 0.7608582030755112,0.5275386117238121,0.6556715693156765,0.38040501508000646,
 0.4908908616274835,0.2479465882398424,0.8541910913289151,0.07915918090358443,
 0.4578322075561603,0.38727864880322116,0.6558813312946008,0.7421936188332253,
 0.25143743016807774,0.8714308287293053,0.4935414178272932,0.9804280862486195,
 0.06744799823319714,0.1766223416271131,0.15944509903543735,
 0.9489314663002196,0.053509697183831695,0.9404983792902639,
 0.9741602233101153,0.6530975445238028,0.22955048164850644,0.5968483591438905,
 0.966536195179503,0.8026160793317664,0.3971292548759344,0.6534415946205324,
 0.632131626698555,0.06011062276011059,0.6652155832046396,0.1039801843709266,
 0.2219909343558938,0.362597958731926,0.03729978232076048,0.04321254283812004,
 0.8950987059549697,0.9006368988021763,0.5790416928262345,0.1612704931260569,
 0.7553383440875912,0.740558220184596,0.7357315958208721,0.9710664491023551,
 0.06741829453579395,0.3450439507996821,0.06284390060159517,
 0.2123887452744775,0.4199589930739488,0.36176315336546805,
 0.13271189976237907,0.647397524614646,0.3908231350795106,0.2280528837230209,
 0.17436282840863515,0.07567167750582604,0.35226225663914323,
 0.904951515067488,0.7903064859073576,0.604215596721323,0.43103022523427725,
 0.7393320936027898,0.7016116209436953,0.02359565807807218,
 0.16872399369951363,0.7331222518225117,0.39835892319091337,
 0.5449911427859586,0.4336822165807094,0.6201106197427321,0.3917608277118029,
 0.849692143226167,0.3609257953319872,0.4080895469889094,6.275656597681811e-4,
 0.816513331906277,0.6728249492248194,0.8716052803723188,0.851642147581051,
 0.5230072226763962,0.06341483574022888,0.4572194712396698,
 0.35509436441012465,0.6761336005735585,0.9105313227425285,0.76806714123215,
 0.48697479680078015,0.05854607008074708,0.7198254944818882,
 0.8271383336751134,0.26587543114025414,0.5482466813410234,0.9812312126840728,
 0.594621200356873,0.31582142260224544,0.3284198092523454,0.5930354414006599,
 0.7704310706990058,0.9867953362385012,0.024331099111175924,
 0.9777137225005258,0.6867465904993872,0.7586501142210077,0.19141125164403971,
 0.7490354128555786,0.9394248181995499,0.707209394989765,0.8284147492919389,
 0.20757241519170677,0.9843614450299228,0.8429521595217555,
 0.06469346388338149,0.8790388755969979,0.3063496415325543,0.9025033506549914,
 0.1371894029344295,0.6451903958237158,0.136350739724168,0.6736514700837274,
 0.5421539491933821,0.1932968945506892,0.17834158821884927,0.5145907577829725,
 0.018871597290171405,0.7512219595247422,0.9338152161809781,
 0.9657532768650039,0.37500475550583656,0.7777424564433318,0.8722642369191693,
 0.22253864769247955,0.331176584943369,0.6388733278082049,0.29607305478450974,
 0.04481756921171487,0.12898679074147312,0.392160798750004,
 0.046230962170384804,0.7131226743439136,0.45203254554419525,
 0.3152715141089947,0.7084383430545281,0.6906401195540495,0.20244575690235533,
 0.2083599816476871,0.9508177266241242,0.08419931456785879,0.984832497505167,
 0.9064807958038045,0.8169420113742099,0.29285425411011834,0.6617794921571477,
 0.8081506338560771,0.2231247065756644,0.5218728194080474,0.35633115472238863,
 0.8576054131336007,0.8816219170841643,0.5438932293257389,0.42949359525382813,
 0.6730051328203834,0.9974276947801916,0.8014429364559078,0.4358163432168143,
 0.2842532540579402,0.9580019155946289,0.03105879561481739,0.5163912050538195,
 0.12744855036292124,0.22497692223687238,0.2635509050260564,
 0.5453003952424412,0.9417273071698682,0.7113592737074694,0.3482807602369449,
 0.3176435799903128,0.9109674881251463,0.05499944150200997,0.7676629002093558,
 0.9341691135533154,0.13479442986804635,0.005778390325643423,
 0.2967295302146671,0.0625362882933922,0.7395236886022707,0.3504555074223924,
 0.6677916938061195,0.8850677375750862,0.11272979101905478,0.475306166836313,
 0.09421737337621616,0.2772158807788463,0.4428933515490001,
 0.45213038957286444,0.6926182471773812,0.49037230599370407,
 0.06929447976435521,0.13582944677179143,0.9487757510319926,
 0.7711680174587789,0.49675248383488835,0.6284359827699326,
 0.39927576053823566,0.19666888201055144,0.7796043151553275,
 0.9714903261703449,0.8782117131707301,0.5893981074289197,0.9664607670375159,
 0.4348270778683272,0.9491431002467775,0.9342667511572327,0.7983965168653676,
 0.6755446877047181,0.943932487773958,0.09281821808099433,0.597369059733808,
 0.4303033996898421,0.8852515934527557,0.003855383447892824,
 0.34670775605390036,0.5973428513494093,0.7601764539174953,0.8189406389473857,
 0.5608124661725826,0.764474929422337,0.7953977876843406,0.2871966487616767,
 4.992785266788591e-5,0.530893731703006,0.6476637254965519,
 0.11096163933727843,0.09299030690411558,0.8198316280323164,
 0.19794514487803672,0.3769075682921442,0.3589667204090099,
 0.19897004072630686,0.497501391130759,0.5477662106193294,0.33946635432003225,
 0.04752928578362137,0.6117306600472247,0.7777490236752229,
 0.35886976063620535,0.5233082683810215,0.8326923714851943,0.8708359530101579,
 0.26929423460092805,0.6582020341568635,0.1599271733177705,0.7562796938840528,
 0.9067770305547453,0.6402132088473027,0.4922212171363559,0.04203298113148635,
 0.5550268603442431,0.23508560598864525,0.27718063687172945,0.464552012215558,
 0.4361489294666018,0.8952747569858838,0.4357846590120379,0.3954815450717353,
 0.2804965468875613,0.711201991602777,0.5000943388499032,0.8848569388253295,
 0.9463799059954079,0.12316325455545374,0.9448305802847978,0.998092096629105,
 0.8682747848242474,0.5794105926059903,0.3803376850936915,0.185646746544049,
 0.0016965557834041078,0.2090715010311366,0.42346274574556897,
 0.0780034422942224,0.21279808418893054,0.14206675044806727,
 0.7729995772103176,0.663478496575027,0.05360212896007355,0.5112585588036416,
 0.48543959187173646,0.9647148057491115,0.445199913841993,0.8165680581938741,
 0.36788522595574336,0.08298598095186216,0.7352250704817522,0.524333851225106,
 0.10924895238934085,0.687045590313704,0.7651257562212012,0.8959207500898029,
 0.27612667214827136,0.08548687071382943,0.762345883920135,0.654197449612549,
 0.5551182481741574,0.5299332100290864,0.12285653245943151,0.7794391765392865,
 0.6525851744265193,0.9776486068178706,0.9729483594518208,0.7013064528731391,
 0.13995091544099392,0.771436220998984,0.13083224630116153,0.943061114610452,
 0.24748391646518375,0.25445234340792433,0.556988581311633,
 0.07574754464034239,0.22399993725543998,0.7622596956411045,
 0.8687213873363212,0.7885944248422646,0.2748700562451072,0.3570174232602258,
 0.5191029950581898,0.6410150155074137,0.025671749173919123,
 0.5127367714419798,0.07259307664609527,0.32154767187519095,
 0.27817872335088945,0.8625411551438765,0.9918645850056246,0.8253110726162726,
 0.5621647351022884,0.4349489365066057,0.06006540860068166,0.8387146304850541,
 0.982087848488127,0.5552843998974659,0.1593520147888211,0.24118321610501448,
 0.9958813569764022,0.9066548514758885,0.9487038720611493,0.6683348249096757,
 0.4364842151364493,0.504979984002394,0.37190844137513546,0.7605258903536392,
 0.6943874248921987,0.8457361263755053,0.5379048688230664,0.43835082961414285,
 0.3788491640808752,0.4496625265222519,0.3651282035108119,0.46711345032207485,
 0.21191420801216665,0.21795212489223048,0.9609576318563856,
 0.3614352980782263,0.8786242458415817,0.6002797403904623,0.5151883244744437,
 0.6193066495902312,0.2442487916015157,0.7042557658509369,0.8732502722936528,
 0.12866480655195778,0.335388551627281,0.2598416438841379,0.7948553281891609,
 0.3401188765462252,0.9348661718961596,0.04398911044147047,0.5681515350586009,
 0.13336635854756007,0.8748522132354551,0.05944859050854223,
 0.7476516822728233,0.604469444700742,0.6514696821070849,0.9779055182528287,
 0.6398824719744535,0.9007831387143597,0.6671206725935279,0.34786471673908204,
 0.5710821379726289,0.8348609514517783,0.5897706698139262,0.21233784572772685,
 0.5162817963632037,0.7247606614545912,0.6990938371168235,0.3327977847142267,
 0.6183995970502613,0.8553516968059118,0.48914814382710214,0.7367230551383525,
 0.22128109125151718,0.35017819466633027,0.5735982931920315,
 0.8959377233717885,0.593583126285048,0.14558580979811575,0.7036737837585061,
 0.8948926024105478,0.6823951788751845,0.5428454819091812,0.2504731101351564,
 0.5456845819550973,0.3997145605067003,0.9312428468754507,0.5309133461782606,
 0.7575242113901408,0.8024142392584723,0.7276498516067588,0.8711591478815617,
 0.33131806349665327,0.10622328193377051,0.6209799947688223,0.538815323142648,
 0.14953387660033624,0.9654462866306692,0.5200136794823673,
 0.21754584571893087,0.23404328334201396,0.6576927681638602,
 0.33396908841685957,0.8647714321335535,0.6789236895786492,0.7432225660681449,
 0.7840211281520814,0.8708029997960417,0.8979261522734374,0.34943388040333423,
 0.16940953027930572,0.744661009650518,0.3590792829408801,0.21561257605381312,
 0.4446031484488784,0.44694683328960383,0.17009168235221395,
 0.07908415116319611,0.5492713252818032,0.6111865152103205,
 0.040643020592401546,0.5250493215723573,0.9655109268960093,
 0.7502508299204004,0.5171043300524417,0.7851679407197748,0.9773958958933271,
 0.20536372856908447,0.6964529485088171,0.30114733811568173,
 0.5110384851435583,0.7042963177105953,0.06895114029364524,0.9842258308148684,
 0.9919227748621945,0.2471097859687258,0.1823114381243509,0.45685710558332415,
 0.09127504371961415,0.831553495924248,0.3532953703268542,0.18261683331498202,
 0.7514480391162826,0.1578700337985236,0.28376069228451384,0.81487668526878,
 0.32166832621943964,0.4912264757121766,0.8113170284990849,0.0215000454523403,
 0.8098678464846647,0.6167196109454006,0.9873689762467348,0.14352866582844626,
 0.06681348325633263,0.8120711993568034,0.1924594743860485,0.5748111191771672,
 0.36706462435691267,0.9879836607997041,0.9841869780338258,
 0.16432967644630625,0.15754645361336028,0.5732328047467204,
 0.11436404598077299,0.10899834071640568,0.8575230640564807,
 0.22914099256504739,0.5825866882507336,0.021051395109592308,
 0.7560381900246563,0.39511705732625824,0.34818581137355564,
 0.8492016354128721,0.8079060773190709,0.9831364611218808,0.2971334272228613,
 0.7140505187914519,0.32958953458377327,0.17448826158705677,
 0.27388953776616354,0.048781794197794826,0.9950700060529392,
 0.9793337128558947,0.37024547440990696,0.44113546457925723,
 0.8804766522362446,0.7644518315749966,0.3083632804103009,0.17192210036560485,
 0.24622558599882083,0.9684033623295986,0.7227277854544059,
 0.03249073321276463,0.7803168117570299,0.8740808521185343,0.7407765290879041,
 0.32744106628513325,0.9943459613564951,0.37900315542029417,
 0.35086438848704327,0.24138985259790702,0.44364068119199407,
 0.7808932513308238,0.9382464356351374,0.14869958143813977,
 0.02370817279366444,0.10267070464296735,0.6111979610102172,
 0.3367726867866061,0.10223307621413014,0.7478880847025134,
 0.03427419823183664,0.9436365692999501,0.8166198541194809,0.8706732017635919,
 0.7187125007052392,0.44649124349828306,0.8734244655792854,0.7629350493729559,
 0.9240471406440137,0.7199725296346955,0.6922585866738964,0.3202776804668632,
 0.44790790146959425,0.8424070242514254,0.0125434487484255,0.6541734226621198,
 0.8580039568925448,0.07433898044873433,0.4515631771911083,0.6606832499290245,
 0.39480412618859895,0.4457766101382237,0.4176382400987866,
 0.06593517148631178,0.868368320684587,0.27760731458286625,
 0.46879842491224755,0.17709284964625016,0.8717707392784418,
 0.9778988167418148,0.8027140161487889,0.8796110468740579,0.8678231538063856,
 0.05704155866493821,0.4792514970429379,0.608868208078332,0.6346406414406462,
 0.1778957084619659,0.24976200837782414,0.472401667482911,0.9620522665415245,
 0.7460030671869693,0.6227428033481801,0.4258840789131926,0.5005411785889665,
 0.4055604268017431,0.9045046098934779,0.8471694509442472,0.3389792284951375,
 0.9405277486824435,0.42050727544414246,0.08825933413339404,
 0.9927988770971888,0.6534168177274357,0.7891678934052897,0.09959182379014508,
 0.19587077445698864,0.6061998682121904,0.14895367596641462,
 0.8968235761307128,0.7164673643926996,0.304001366649134,0.9974084191557216,
 0.9916765658069676,0.031041580241339406,0.5696571529341045,
 0.19132236661877688,0.38430559197433767,0.6879430444151362,
 0.11878120157060468,0.9952568340927487,0.41997841215466614,
 0.4935443513865564,0.05028965190110579,0.9266365683750549,0.9837480672528734,
 0.8099272921642695,0.25670079552384506,0.9885002924492755,0.5123494282415006,
 0.09633754289421514,0.002938501869099497,0.39114563931775137,
 0.4442833726690266,0.9974665449703229,0.16739701547097108,0.7527023273062999,
 0.9012636704614887,0.7620902545567566,0.46324032914314994,
 0.15644807392077864,0.4469770655044941,0.3621119125911112,0.7675898128563254,
 0.6708490688065158,0.03552469279677728,0.9130487391024775,
 0.12458114020845956,0.7891521397751807,0.9904252070831595,
 0.21710509925612154,0.17572078195482288,0.6948197038450898,
 0.47229136092158386,0.39155504640767513,0.4104082230700459,
 0.6390150020228715,0.10743025030862918,0.3067920498876777,0.8042590072258222,
 0.13017637789734293,0.8020887314685953,0.9215974503977697,0.46146399070679,
 0.6421866516567865,0.18663455222458558,0.581793259696062,0.6495983511977108,
 0.6974184336410808,0.12456974073962068,0.5224967548508728,0.9781825623463011,
 0.7847264897350525,0.6704718154770866,0.8898838028017996,0.3802325642402614,
 0.4735927050436143,0.8339333407169466,0.7520881538336578,0.7051857446216611,
 0.39583151607794853,0.9084702395354468,0.19416600730747802,
 0.3741764488175129,0.7531773103981337,0.8803882996909594,0.4517374338692648,
 0.3137861512120901,0.032154011500126534,0.48239603034576206,
 0.8943883556793834,0.8501439651766362,0.27090382330805685,0.9010451794342724,
 0.8615771047772555,0.590229446043942,0.593533721443541,0.47972372872638236,
 0.7220673860894283,0.36037394302930337,0.30969904862526043,
 0.35371300249725646,0.03859388764687488,0.4765681147028301,
 0.2151450454118421,0.33240308596884394,0.03392789318279421,
 0.8089011037902301,0.9247243358052997,0.8852213166845015,
 0.004607011437893327,0.43922041295626224,0.05656964850098112,
 0.8208945521163407,0.6608137001687122,0.40295664458779257,0.6398461922738268,
 0.4400863680054977,0.6609266877171882,0.40854821335211977,0.8376061353999509,
 0.2346705794114644,0.7265103588240202,0.07876972773607116,0.9882355863845618,
 0.2416837863015847,0.02713261119488286,0.9995751628809554,0.9676548431203551,
 0.8506524811086151,0.4462018752383735,0.8816003469746279,0.04465815731306633,
 0.09156545062826738,0.7958479076485551,0.07157115532460923,0.495957715508627,
 0.4013898930414206,0.28788711022512137,0.5731681844591887,0.9423065664897794,
 0.22716815879430718,0.6081759483127064,0.09175516586755517,
 0.3432815372401349,0.4218495625600548,0.6119670176046232,0.8433689667479498,
 0.20939856413338132,0.36602258561351086,0.1611181360582925,
 0.023361237189407524,0.08666948513949169,0.12955962714348535,
 0.28890907078767125,0.3006423077868987,0.22206730506705485,
 0.9325866414646656,0.5282813842445175,0.7234729780757295,0.3623987543668945,
 0.9305953007861374,0.4402769420968282,0.21328683863865716,0.4930301805138465,
 0.941301166147013,0.05276958995225556,0.5994678107596757,0.7397647489285246,
 0.6945999345130535,0.19423281274553217,0.015845644101489254,
 0.05654047063661527,0.40396896891225476,0.1900612893119391,
 0.6279711595364899,0.08009553737925823,0.24851703112700063,
 0.43965192617837734,0.3552410669846985,0.3366697323059281,
 0.25974464249075735,0.6150930385667134,0.3410171473808782,0.7782056204547491,
 0.4950221091244167,0.4120790148895158,0.565806585914333,0.7131752778287593,
 0.7732149072585346,0.5822170522318619,0.9324318716050013,0.06542511526193939,
 0.3398539598101117,0.07430042231093026,0.7620047400311803,0.6783677831695548,
 0.783961992765293,0.9734669087863685,0.34651606074217967,0.7817537585486047,
 0.36922545132136797,0.8145577962369378,0.17265833740992942,
 0.8608422182075843,0.7768245332776107,0.49109117399802815,0.7601242809524874,
 0.46457845154643285,0.8701326355712845,0.9341640578408308,0.8457119897149357,
 0.7074533471447126,0.16013093118749178,0.4131638238527928,0.42404990648577,
 0.9495345917617375,0.2198136778397457,0.15577397422535633,
 0.13917543149722467,0.22828509362619176,0.23217643630940443,
 0.7870149728770317,0.8534869900172624,0.9517533448704397,0.6490083598255323,
 0.009546482015931335,0.6193424378564315,0.9890049473750665,
 0.32313445125861817,0.5387204300622815,0.8122123942428044,0.9112253303180374,
 0.9617731728550227,0.05172550839767631,0.1867512075816541,
 0.21333726677432652,0.39055706074465224,0.8314820979962081,
 0.4487770725218333,0.8425623902502962,0.9738361267574708,0.5614762714324704,
 0.8606188897025486,0.429943122612372,0.7268271863296379,0.4502926043346188,
 0.7456453488680522,0.19615199678110673,0.41689424626577865,
 0.10126532429247492,0.15965340410543893,0.48278433939097554,
 0.7351975863170009,0.25348975739188384,0.8969900038265661,
 0.17590214506095714,0.9643602648524752,0.9699280569669534,0.5689542518288487,
 0.046459197772957816,0.9730834032038319,0.1795427356403977,
 0.7471513808785073,0.40135960255641745,0.6378638229672327,0.9714975881671755,
 0.6299143856942221,0.5536042191478044,0.670122538009671,0.05723868683665345,
 0.9603699372963308,0.0660253761296048,0.22506883921466025,0.7581514896485321,
 0.07524240560834394,0.09811403414662023,0.7300910684915365,
 0.8438221357714468,0.01456977400635684,0.3461703755882538,
 0.27230082864544203,0.7454662758737572,0.2227230344916895,
 0.24812820177730643,0.5157450922683471,0.8563227903204023,0.8083985625886432,
 0.8997466614422456,0.4509744211013278,0.49918745958118715,0.633936236544862,
 0.633484789169936,0.4748056064314825,0.500598957960165,0.8226957662073371,
 0.6025560074828261,0.6282263585997476,0.4813678041909538,0.1638848535513142,
 0.26601009745851123,0.726604800479293,0.1569306544601562,0.2773977660738369,
 0.3038672125170655,0.81888645625433,0.3983850145671666,0.6873522427115191,
 0.9727400202729622,0.9491672680748329,0.23101934111673905,
 0.21184883856397474,0.19285177153108046,0.10414641133118707,
 0.9395441969578997,0.4683219563199861,0.10186927369422705,0.4447729134880192,
 0.7661160763822403,0.9892986172376076,0.8325096980916543,0.6066196711384303,
 0.8063988270448957,0.1269797947573963,0.03940491506339183,0.8236082251929182,
 0.06856288798418131,0.6202036947051963,0.658013973480183,0.980407502632599,
 0.21382994681618428,0.6047778220335127,0.9681220518990201,0.9960267543546807,
 0.9111742639630457,0.6780112352496571,0.8621381875674722,0.21516427473942423,
 0.4816063934610961,0.052388401819723596,0.882149727122559,
 0.27255597571399337,0.143736478133208,0.012007000358103537,
 0.16620898204929624,0.3629570100519144,0.5254417381545613,0.648216364743643,
 0.7172740467706309,0.02946444536385673,0.2732225840448912,
 0.35024297963656315,0.8519670243411758,0.5773063320998674,0.5084833508263485,
 0.7265374261932384,0.5610089296641035,0.508389131307952,0.040349716977670314,
 0.9731563479398508,0.0982836907340302,0.7893202817326794,0.5883070708689735,
 1.1796726118051826e-4,0.17709819344312505,0.005384316449074,
 0.10705184351897057,0.02514813683816952,0.15547046238155504,
 0.6199753059166434,0.7365300557448711,0.47489931142345687,0.4736050009796221,
 0.7015683863111604,0.4454075523910165,0.7196349667574291,0.3818548433051212,
 0.6621609692053024,0.5448389898714463,0.03558510172862284,0.5384947704197819,
 0.3470119022573175,0.6091468714489664,0.9582284078496002,0.34193257330911553,
 0.6671759273429958,0.14033833229461956,0.0634416436000077,
 0.11043385513641524,0.508188287241435,0.3846431170450765,0.3251174948950083,
 0.24843299249717965,0.37021162415512954,0.8594297522935185,
 0.014165420091426295,0.6551603821998865,0.1089063492392972,
 0.2935471595826462,0.47419805610697496,0.21470370863224475,
 0.9334558362346016,0.7202958873789828,0.6398574643595701,0.9723290986467821,
 0.7858215789455735,0.4186820744440778,0.20155818130956593,0.3919412487269677,
 0.7463298797911874,0.07369456866457402,0.6455013855309277,0.9198472964451505,
 0.9782230295527212,0.9600210138864075,0.013155093929441586,
 0.3066157650364931,0.4548288793646722,0.8063091355398673,0.4701348747441676,
 0.8268639065507131,0.5414418106968503,0.3655603757419117,0.6621604073284802,
 0.04586651758332594,0.9278262009066902,0.60245788964079,0.15394203133677864,
 0.4358465102959581,0.24266281076266416,0.3606895705540434,0.9035325825036984,
 0.893219572349251,0.6564991179953708,0.526185986482471,0.1964847194157937,
 0.9682996781322506,0.769365881182362,0.4419730366912775,0.5029422757171692,
 0.23619437905006468,0.05853879491762115,0.6009418754150705,
 0.07888655734114103,0.17038553773702036,0.023642513439127932,
 0.4945111594740135,0.4557030332518719,0.4009903112153218,0.7055596244428265,
 0.9755973080458334,0.05705321616893566,0.4490145434844459,0.5014448225412538,
 0.20043767179230298,0.8480885358142648,0.7364571516301337,
 0.009930605021017724,0.9475626760624724,0.3001233513670728,
 0.10411892681930501,0.4288933715992068,0.987297832581389,0.8690849904866571,
 0.7786743313215175,0.14224316268747228,0.056506165418737875,
 0.23468032313150244,0.3381095166343131,0.9871504128512109,0.7722562445767496,
 0.48149236135205054,0.5827266235883964,0.010950344808955847,
 0.3762631314521927,0.5764751360610098,0.7646209823499209,0.9068871960761274,
 0.17233388733383093,0.3508308468397139,0.039052754555969016,
 0.745936542566767,0.2959661621681493,0.35960081510194575,0.62734329924231,
 0.31896316990907747,0.5560049338756555,0.1704531632813473,0.8162441572032951,
 0.6394647360241635,0.6468898469952575,0.05536065029273507,
 0.018145350385354475,0.2976059319793528,0.6818699069266976,
 0.39259499542538845,0.7933184964808531,0.31769481338267447,
 0.0179646523671676,0.6086329786373349,0.15822605556062586,0.5152083529604959,
 0.9260674778563758,0.08869233952894973,0.6723889066213526,0.2448103411984005,
 0.08559688544234223,0.14473511924661653,0.5352696826223902,
 0.8828697749259284,0.6636507047717946,0.1841095869894669,0.6705899932096828,
 0.196859818875744,0.194612521654556,0.9387041568627246,0.7739821002339948,
 0.6328283912168516,0.6081236789899763,0.48317686262194237,0.5496435368956843,
 0.5847306751843074,0.7256740008540241,0.010475996093329054,
 0.8582189915523689,0.28068525328877636,0.5028201902383744,0.7437903706208869,
 0.8350077865633762,0.24018469408112386,0.926801159131094,0.3622047052684444,
 0.33909899017103406,0.928739539398175,0.036502275547053875,
 0.8312227075598304,0.2173653951629454,0.07781688753033089,
 0.016132151946991358,0.19148408608587353,0.46105056071451767,
 0.6028128957852017,0.9360594052798138,0.32957982482924286,0.4306115505449275,
 0.7624009038727773,0.37992621806996796,0.6791713404969135,0.3797749861007914,
 0.22040868303096706,0.08223678591192585,0.3743880899481684,
 0.6874960890547295,0.6736704531306432,0.010953499393180022,
 0.9142295096292973,0.72440385882917,0.3906133370670347,0.5987957939287963,
 0.45165545756187553,0.9365990867475056,0.16339036547232788,
 0.5925375129061338,0.9392504845864295,0.38181917647960617,0.6073569617970925,
 0.7639039963930419,0.4507322357507475,0.7615805567210746,0.7129117230761688,
 0.03534078146325603,0.9759317059670882,0.18419984140167345,
 0.05855454061600207,0.6629374090589666,0.9570071239430743,
 0.015539845055468127,0.570770132005724,0.21266477567583586,
 0.43859082815447137,0.698926472226044,0.7147193550160025,0.7223191757666345,
 0.0724407814433331,0.34823237650202243,0.23345618944693003,
 0.2962377163346792,0.9532671818864265,0.9851960861761393,0.9441069066369774,
 0.7733970345010094,0.7418310654526968,0.9442346770311422,0.2348159063592581,
 0.12025171355337083,0.08999551434364927,0.9087690161450467,
 0.9524959785820404,0.18750496122884663,0.15139869200291722,0.775386487354864,
 0.699025336926026,0.7498952658020116,0.33944220652789403,0.5028318923670422,
 0.9200458029610552,0.6679176574450485,0.608745684831831,0.1298275201783986,
 0.9151622632469741,0.3540004729473961,0.21139123879654553,0.8792385124593872,
 0.9909994883840751,0.7298775242595549,0.8882447602197647,0.7053929475573657,
 0.7725147513072881,0.3680601684453173,0.6215675775484705,0.9649689466222825,
 0.5994492026879151,0.4686565319552479,0.6357239516864108,0.5668279316372496,
 0.14773111628606272,0.03599261854421143,0.14313419151732965,
 0.07790551292618086,0.3353967246072356,0.09464158720778615,
 0.35934048475603086,0.19751086915252636,0.6308182009461875,
 0.46163363131028545,0.3349451068409831,0.328283396076012,0.742076908951807,
 0.9680364193573348,0.4706257859875923,0.2610204222680417,0.3149427361524211,
 0.6323870630252896,0.45497348440382956,0.2478076104694421,0.2956398122907602,
 0.5972617611465437,0.5856584595569325,0.5252360673407843,0.27996847170313877,
 0.3614840747046829,0.4715374481077702,0.13574657592475614,0.2308146224408406,
 0.8894066808036557,0.13622807948202798,0.8425880595835601,0.7500107935858348,
 0.0015591165437240928,0.9152089096779662,0.32254738702366337,
 0.16412207808117718,0.260658028318826,0.7107978782715655,0.08517822697388477,
 0.3919811046074022,0.5614334617017828,0.256500008650432,0.7398476421223377,
 0.1577030345640672,0.23358406284014843,0.9807035948394334,0.8390078318957971,
 0.16300917740553778,0.6104849815559656,0.5858343785776496,0.727648298838164,
 0.8626707636744124,0.518096622174057,0.3804307783327072,0.6367260283674822,
 0.16081494782833927,0.3521394771075085,0.056560856354462175,
 0.8503054230395288,0.5141890872330515,0.826878063313017,0.30908413073972774,
 0.5792418017059927,0.49511223550109296,0.5445711537584182,0.8906546856382102,
 0.8295356323893398,0.28851615176696055,0.8011754790037374,0.3344973146889626,
 0.19004866588216618,0.4993091303639817,0.7633572817137344,
 0.35390332658318047,0.7591664378148821,0.41247016128702496,
 0.0030020223916182864,0.8137968587209427,0.6032811274724438,
 0.5921594040716008,0.06205143465659324,0.6395695976617141,
 0.016067672740234795,0.48203310646446695,0.5465690333480483,
 0.3565607183452215,0.40211519104839577,0.8622149231977766,0.6210580543763219,
 0.9106062242129828,0.01246260420998302,0.33317926385791896,
 0.3981163889880903,0.953714978470202,0.008900380812111863,
 0.47807582668668946,0.3709371510790074,0.15993944370865898,
 0.4479432280441845,0.959014120129663,0.6879589729639113,0.3666019634814006,
 0.18499055798548913,0.8144302036034824,0.6304800551525195,0.396134646975696,
 0.9222724964578404,0.38018309834084174,0.4523016735940675,
 0.14034836116561789,0.09380267114054908,0.6127916391647172,
 0.6196758702814921,0.7569152323287647,0.9392602039857598,0.44973800855297696,
 0.5879436663298472,0.30206829203264096,0.846219348636425,
 0.017140682167282928,0.9011136908354491,0.45572668854835385,
 0.9538234232783975,0.7513388567538976,0.0863114843602939,0.06082146673472266,
 0.4098289744661835,0.39682103896574183,0.10170824567072956,0.640945474913749,
 0.20145932822030277,0.28659847121349336,0.43083415404107483,
 0.5765673226447032,0.4426101193239287,0.08878820077408567,0.4371225423027506,
 0.44330158966058,0.1353368270416615,0.4357843948265644,0.6651089032765761,
 0.3240002175007859,0.7926569082523779,0.1697768647422504,0.41022225815653723,
 0.2010460942909209,0.08225814517452901,0.23749833015269373,
 0.5445108732741117,0.21791997888885062,0.9359584760347764,0.9004299325719713,
 0.16982180797734736,0.9423078479181286,0.2496518178460484,
 0.32277546461790696,0.6909458602595291,0.570176508007284,0.3555747537024597,
 0.31214684267053694,0.27019306263705,0.4353187928696147,0.0865855792696657,
 0.03545761725790919,0.8941958489039563,0.026782962505012553,
 0.06947433986999596,0.29414107471101525,0.5447559878342718,
 0.04511352094601995,0.15689642358498757,0.24322984324367836,
 0.3068861147101122,0.9889006966006885,0.5080966015756464,0.8028670747557052,
 0.8101502172097951,0.7797146550839128,0.8947887328332185,0.22422131135682866,
 0.2213297089873011,0.5243892997510178,0.052472463591575313,
 0.1415486952442243,0.23661146904050723,0.12686363758352937,
 0.4672136979153433,0.6484750062901484,0.40818394739658903,0.2194424177625821,
 0.5284821569006845,0.436367167533277,0.29367653130048466,0.37121433387993663,
 0.7464387992678284,0.04541237538028908,0.5434942817243107,0.4013867283044886,
 0.01374783619644071,0.9082065804121939,0.3936398924552925,0.9803579257952849,
 0.955061270176575,0.9739718670875812,0.2953514522886528,0.4832706721680713,
 0.017477298050422307,0.22545343193946077,0.19281964557178188,
 0.3054972546119519,0.6837013323709445,0.06465214442854172,
 0.25892057551302794,0.23305407286339797,0.8716434874371926,
 0.7525867379932923,0.770552100581959,0.18554267391957613,0.7253266301281472,
 0.6333868609110017,0.9129364014020869,0.09923570759503808,
 0.059863734418352244,0.8280152323365566,0.394744068271359,0.5299954992708182,
 0.3576475763753453,0.005947398735375642,0.6662698501492951,
 0.6190976334365614,0.29284288427623906,0.8320785554579248,0.8472187558050344,
 0.10319983677044708,0.381001146390358,0.6758476608850122,0.24582563313992056,
 0.2577381984858167,0.9495343661486706,0.9472157055885502,0.9704807155573966,
 0.7594803509577347,0.6523295545694131,0.7670081291758347,0.6335740428825964,
 0.5642731478928765,0.6436277438443199,0.6668457944976667,0.9863742286927726,
 0.614808731108677,0.5451024506497569,0.07941116264974113,0.9225377102669468,
 0.01593650781191236,0.8183354561327512,0.03801426139110797,
 0.8908611913181502,0.38965880985186174,0.1633824777662921,0.9021414551986067,
 0.08885522989232775,0.7568145996415266,0.4879023478315889,
 0.11407898803498395,0.3232066304516852,0.7247429131892277,0.4077256262934812,
 0.3627769424492089,0.8634756348278904,0.6496240140943229,0.23314684227916593,
 0.06731345029633451,0.19651830745590126,0.49649225012471776,
 0.8043390970635999,0.17532098284532993,0.11227775914197624,
 0.8524657005169995,0.5687029095313261,0.2452228582954561,0.7574256186351398,
 0.4732974827148244,0.5371071963727193,0.33681190737753597,
 0.35365196325722414,0.5810088089914459,0.39164982476079035,
 0.6466272133561748,0.5489501321154874,0.16502504323888312,0.81112886331225,
 0.26192433864422593,0.9101206403660118,0.4405117917850754,0.9345516939287879,
 0.5566798370170402,0.7300011467801946,0.16410528426136395,
 0.42799533707992365,0.9664349553783048,0.1370776882607463,0.4348416748484486,
 0.693941292203528,0.6181274022245871,0.6931813016491669,0.8677629477086618,
 0.1698906107938125,0.3371181140658701,0.6404706058003364,0.8230376191726275,
 0.48600935651425714,0.29923534952359776,0.9316469435302712,
 0.3966442662645352,0.2759248433043868,0.5899157663698785,0.5128021399971937,
 0.3790076011889296,0.29016515026520306,0.5298079497088175,0.628220115793289,
 0.6978527807166139,0.12905937754892904,0.8043451274213849,0.7390980014568047,
 0.9300665919491136,0.6093919820505693,0.8699331268296656,0.5922035537548892,
 0.5089978585528131,0.3133699214211565,0.7984988832263131,0.7105523490140557,
 0.3723693744563996,0.711146009033224,0.3217144588409556,0.7687817739891132,
 0.5872119223048794,0.7375188334043744,0.9178805461303312,0.21695308030309435,
 0.9893165515189093,0.07637058203289682,0.6814359853824186,
 0.11442725867243297,0.09105741784960375,0.638454657725629,0.538210700343463,
 0.18884182084974688,0.6358918327819992,0.30837805314955036,
 0.02401868818230224,0.14890697129033614,0.02559732011330318,
 0.18656931329531068,0.5000590284068539,0.23664483054436047,
 0.1206203228787659,0.6701827319722748,0.7447935699120732,0.02124955718490651,
 0.07187694664801136,0.5782508492806355,0.022472677454512535,
 0.04137927292481658,0.6721643476310443,0.844747555743845,0.3766745859765249,
 0.7965572635210859,0.8168976124697495,0.47616134377175334,0.5404471965940673,
 0.0014323365631931928,0.6812234160926515,0.10178008129136273,
 0.18705526400937922,0.7191200270708095,0.3623947830965838,0.7485703374428816,
 0.4751781890939094,0.11892087319034306,0.3080429914546571,0.4668378435387388,
 0.42191641564050364,0.8581516071376649,0.20570414341822785,
 0.018686181072110064,0.6084146930984443,0.028540959254487075,
 0.8706638825797072,0.13969504489029694,0.20580084149209243,
 0.24771315045150177,0.8045586472930308,0.34520606854506986,
 0.5618780009195712,0.8519777512520716,0.6704581990725972,0.8992598391281711,
 0.9613819239718003,0.4614383919175735,0.4605313555322844,0.984411248063564,
 0.7330364420022315,0.7739540195935953,0.41924282833641624,0.8949951843301597,
 0.9918817295267099,0.7318450918237764,0.4409369135702763,0.7910094289867644,
 0.38333158891302166,0.05949748321860404,0.3938830517675539,
 0.8062357412146873,0.0704382234516614,0.6146609453612679,0.06341390415054515,
 0.7326270824039703,0.05384957388953593,0.9082841825280997,
 0.18917744218324717,0.9722625915959338,0.9615966586611726,0.2691112011540161,
 0.6509591319795147,0.73840714759236,0.9199007698383591,0.14323902480398332,
 0.7594307126908543,0.3281708203987693,0.19456149165345638,0.1188670523651626,
 0.34194661854573083,0.1706943275494499,0.11733254177910835,
 0.5015458893616374,0.18528500132115222,0.9949211437857401,
 0.03492771225524893,0.5487863519665347,0.5646198160576611,
 0.030147247129836408,0.4526752114658845,0.9485394107475904,
 0.2694062909778223,0.5767345195345059,0.728833666316561,0.8694373271940583,
 0.23024761115705894,0.20159186457516898,0.04223211951263395,
 0.28186164081942167,0.1968227536559094,0.6717273200425272,
 0.06983240898061505,0.6737728726790356,0.3710035345988122,0.5359239732189528,
 0.6559466460637668,0.34761546818450384,0.04751246586518909,0.434804791193674,
 0.7311419650752456,0.17616091311952498,0.6812605418573702,
 0.16563498279694766,0.1645495494831769,0.48311899691053306,
 0.07945552799737132,0.7592188893393418,0.6945228269838458,0.9901581879103462,
 0.8828714645198773,0.7496554204176927,0.36624995871055877,0.6804764650820817,
 0.6734650821775958,0.9018280728459285,0.3140245642787215,0.7788556755249637,
 0.35962003290254674,0.5308396025090157,0.22257954201266028,
 0.015023705273646049,0.0711505761065967,0.6169612575964756,
 0.2930649227445735,0.758287883118125,0.6321636356194983,0.8655991668598956,
 0.7660062420420359,0.526870819200849,0.3335270018437556,0.10646083305945575,
 0.24747792860926765,0.3470059534469071,0.17557045401812,0.21935820415112517,
 0.7370032226978225,0.3101531579542859,0.17374045919379555,0.7144743132038538,
 0.47394018078376,0.6244929655935304,0.007881982704453172,
 0.017584766916938888,0.7314011323655478,0.11921071429330077,
 0.025099564267247465,0.8153232369937139,0.4835771970912024,
 0.07996440049339659,0.12550112366784183,0.10585175897421406,
 0.230714953973423,0.6313865190509893,0.18074181809484158,0.6106697463648689,
 0.8272591484845387,0.07986075002654924,0.5609568729302428,0.838512940142792,
 0.4624186164492252,0.3727460931413269,0.41063065829463696,0.7320866925864123,
 0.3075708172723657,0.46144060737891557,0.4384469367895931,0.5535683132973352,
 0.186303863507701,0.5571147608340812,0.2854253967142113,0.4791797562509007,
 0.4660556719764092,0.7725475254572051,0.9580493601947532,0.7963704266577278,
 0.6120805912300623,0.6734258874050583,0.6521597236667678,0.5395528752335104,
 0.343254229034085,0.23226766615255567,0.5880344084998831,0.20724620207913358,
 0.9461125263891738,0.7133245925876128,0.18019460645385754,0.3696832711665158,
 0.6638834880803884,0.19813227760977126,0.919150111004674,0.29632042457398533,
 0.3565663321106236,0.9373867705378696,0.0689985091643871,0.771172400228712,
 0.8465348515546504,0.3762978712297541,0.05569967311500312,
 0.018731674781819918,0.9181610317347166,0.6992483566644121,
 0.11820827914777476,0.3864242857508967,0.05365189996665076,
 0.5375991909122835,0.4581215177799549,0.48237148285994746,0.3079786805619946,
 0.4146117665423682,0.02798268098165302,0.6937193967726261,0.5240396455975453,
 0.360969396628229,0.7287901288871187,0.7833390926815357,0.9208543593285363,
 0.2992826422068924,0.7563346459631636,0.9317557576953108,0.5089653941916912,
 0.0033188238892140065,0.8213269436171281,0.25044242757737334,
 0.9614632060115149,0.935955370403919,0.49778220675313745,0.2779357551163114,
 0.13996132647466897,0.8586023579629601,0.9656201198008573,
 0.44839070200444264,0.7421633924873174,0.4192021595536791,0.6148143100816018,
 0.0489588859457577,0.09901543549157887,0.6649018732152563,
 0.38781540287375416,0.5168826754506399,0.4357091804000748,0.2116527023249939,
 0.8686619379177611,0.22788491413620848,0.9522417914971116,0.6077759900391683,
 0.27126396878757963,0.2131450283698222,0.7638234474197053,0.1068204664802812,
 0.6837255877101585,0.9728227938396083,0.008434602967528737,
 0.12273758999177742,0.9512606689562397,0.1702229147539408,0.3937463665502875,
 0.4951794267860672,0.4041087934424834,0.8915070159486496,0.6293667699990402,
 0.31204432241822533,0.31772849846391815,0.41584875291895096,
 0.23951799462196655,0.49608916019493154,0.9228440095936996,
 0.07990084354054727,0.30701058077786636,0.46416517725788353,
 0.6752186695425142,0.3510844629737814,0.31227131721974444,
 0.43365962363599386,0.7862035411879352,0.43506844721311144,
 0.7009494644242584,0.3507510079951227,0.8145647787676633,0.9997322744041653,
 0.5964526049628522,0.08562638428254499,0.8659639026029906,0.7542003572954915,
 0.5811633290307429,0.6317819586382667,0.23107347112281107,
 0.34474303973281606,0.17737213343395175,0.5028853089616209,0.580249789005847,
 0.9473348970176333,0.5146493202510019,0.9014556495805415,0.6948981276270355,
 0.8199842813519831,0.4236906105293432,0.9320716869915282,0.5371983984514871,
 0.7954205980605074,0.23567389328433574,0.10828459765428688,
 0.1689857232256905,0.6912753422950191,0.6203973029493726,0.7050112038494023,
 0.7496982732915045,0.6213835411541101,0.039817635311598876,
 0.6064671037867533,0.4148045725199552,0.33879941998571506,0.5172642521106517,
 0.5948791790797119,0.8919992505083841,0.9289904445542566,0.45282781753577195,
 0.11822644171740793,0.22086563478039212,0.9825310405537506,
 0.9363092794666394,0.6268758703817909,0.24602319399991834,
 0.13029983727856653,0.9196140355777803,0.3055861759832059,
 0.05122867845492485,0.09146272783163001,0.03190757522196219,
 0.3766349628091903,0.7971522442812364,0.9356206294040207,0.7737224036661078,
 0.7461348693475651,0.9571106149813323,0.34514455179409376,
 0.009334253654381897,0.7002336026240498,0.12587257927643825,
 0.28583023092793713,0.7278160976452139,0.13420126899982665,
 0.8387659075138149,0.4044726053348404,0.7216115490775781,0.14656266796449557,
 0.7625484929408579,0.35560197100592206,0.2507423677862751,
 0.16092854996548434,0.20136948262033305,0.606371110345775,0.5455373521189438,
 0.30966104479572676,0.5920903505843373,0.09317922254788724,
 0.17709422825140841,0.44220770475089033,0.7995357161732961,
 0.9919307347981742,0.006599536024089581,0.37577413795511516,
 0.5158293699946793,0.12429829479082266,0.29311519263311936,
 0.8593425901029558,0.05804162293927517,0.7618284941213299,0.9145330610101878,
 0.6071225274091989,0.6202632458547989,0.2308896316181095,0.2911259378718146,
 0.39188083215146785,0.7398421781109847,0.0032698008273303047,
 0.958537461190548,0.35187024036275805,0.1880022473654518,0.358489061003346,
 0.4066694970052749,0.4772435499272345,0.003708613378558989,
 0.39372791077281644,0.1857567762029283,0.3329850099850402,0.8355053478771395,
 0.6819518035068336,0.6203205140850706,0.6239668100050534,0.8217504712303333,
 0.7791446574095952,0.43762572528410615,0.40087864561123776,
 0.9700133800842401,0.6158106589814875,0.3658683234405494,0.6156157083491308,
 0.05008544821295691,0.5868865185646925,0.3722175047597659,0.6081334120623341,
 0.856189617292568,0.9216696548131564,0.5947216185323492,0.48635349720422094,
 0.4374500168864529,0.20419019839738706,0.026869991153045003,
 0.679410878432252,0.10252498469343208,0.8942303530123479,0.05439117209948752,
 0.8777737983334433,0.3564651222830445,0.18729402700678488,0.7641429536376984,
 0.22069252079768753,0.8321907351052771,0.00753727944353999,
 0.3659068057123841,0.6823968425091267,0.05348489410544344,
 0.48867346016126645,0.0962499997574282,0.3831346509736031,0.8371918134693027,
 0.07245171554244945,0.4673919384268902,0.2346411741389156,
 0.16400648904175963,0.7000718872705309,0.39197303491898916,
 0.5338400088364363,0.45112036268560574,0.6894506677560323,
 0.08400273757388166,0.9506719303562543,0.48355886739351706,
 0.8761752558974226,0.49841118173115817,0.025153050337272553,
 0.14900950770762422,0.011489436662287855,0.5531612676015578,
 0.26294938493593467,0.04917039405717838,0.6516591158148674,
 0.6933180997181105,0.8098140121232862,0.7337068182197214,0.5367442538392483,
 0.9579326822847133,0.6547012957214007,0.42949853711664276,0.6711909812831282,
 0.3664430515644608,0.7281686291398405,0.8577416760960674,0.5407152942866117,
 0.5692816468547232,0.938416806232516,0.9099523594477232,0.6313001029967064,
 0.784706270146258,0.560358338079895,0.5101368606606157,0.05943337505511215,
 0.8351731292714226,0.7500126700836018,0.3424527311929898,0.9605221248933626,
 0.5732664537636027,0.2762297956605927,0.41881439470436577,0.8233610363028991,
 0.8649961689274568,0.29047401935883554,0.44424226041493964,
 0.052764057512739604,0.456237198958343,0.4071764753741288,0.8889874583061973,
 0.6770830922155378,0.7788740192576061,0.7263113454990409,
 0.008035054760523686,0.15634100711468357,0.9498633729028673,
 0.2247552225733277,0.7793050224748405,0.31045948374925114,0.4111094433878829,
 0.7764120327016227,0.8795793853009524,0.17049187858501857,0.5108525078448946,
 0.18333809900673503,0.9383751589650395,0.983956292758159,0.838455249905026,
 0.7712398158533538,0.7867723621008884,0.644443028568584,0.2315824062090892,
 0.044556937295964216,0.5679302020019337,0.7846263610220598,
 0.7871044027471388,0.7020349426505369,0.9392674828585683,0.6957710396176792,
 0.7168992152167375,0.4370738329673125,0.7344338065803792,0.9185903927953366,
 0.760710847408464,0.246267756265664,0.040664143747006554,0.2706240523624954,
 0.2563467209834568,0.8396102305928022,0.4910639514155609,0.8403336434682577,
 0.7858966549979002,0.5333768770689942,0.9275177892990384,0.7444727539437446,
 0.17233263928506815,0.198289474202075,0.6290226729151566,0.26684933695726476,
 0.8988013914566808,0.08476780365967551,0.9769371732794688,0.8896442714674985,
 0.14642604339891785,0.050882600012973844,0.9217301821433149,
 0.7873877669874693,0.8296420912130061,0.016939013519160606,
 0.6589887776061804,0.8244335106883974,0.8605562395876274,0.8766449307751145,
 0.09456053069334769,0.7582692336734856,0.2829639321619041,0.5285549677163509,
 0.8507737293564821,0.15294223068594903,0.7663679126310506,0.7881394391083466,
 0.05696738178349592,0.07445391102961896,0.5088358821175761,
 0.41194465410586345,0.7458049620818596,0.7647151363390521,0.6251841833745906,
 0.24059298213908364,0.19809918392691606,0.17107296199651445,
 0.174889149009389,0.8099588810192286,0.39588234471106165,0.505087558024004,
 0.391851028135098,0.269242947011726,0.07228136106577565,0.05047587872220349,
 0.5099184429094501,0.9656139996099151,0.6080737392939176,0.2131491572980524,
 0.929347979207664,0.013233785076869231,0.41333755845920206,
 0.33771810320648976,0.524569433624204,0.5274841405678792,0.5576181160631637,
 0.9413655817180984,0.8018370322245089,0.13733707118477712,0.942823870455197,
 0.4140917411270014,0.7809424415530881,0.7966941577807788,0.502107837937875,
 0.8274924821016919,0.28748692802770026,0.5996888794597443,0.174613504320818,
 0.7567899650308713,0.5169147333648269,0.8481832540215059,0.5812586785334031,
 0.709594557425658,0.14014725572147402,0.5039845434122207,0.11719593303184139,
 0.13419318634687427,0.16299161410707663,0.02788575982989483,
 0.0033669588614021473,0.6930703424076047,0.13301128953482733,
 0.9850988906448443,0.3202117719555212,0.7474125028151305,0.9167314968926813,
 0.8246973626095966,0.04070849167132862,0.8758314872871217,
 0.20559569566671188,0.8767252920729234,0.1506081848506522,
 0.45001633869218804,0.04535731897860906,0.8865226656672884,
 0.01903134909706594,0.6952524730868453,0.6624656081976499,0.8649507480389711,
 0.11050988854635868,0.7420347814929049,0.3344346035584975,
 0.052643903135764725,0.4490820726599094,0.6934310923816612,
 0.15388283864985242,0.39324931021849685,0.8076739868934224,
 0.10514322269137544,0.01762846957722175,0.2493076105594021,
 0.6773981479849902,0.8691918853185793,0.8235419978208147,0.6473560101476102,
 0.6624348327755137,0.5676614077262117,0.009286644509733266,
 0.7131201220407405,0.4362180682542124,0.12361841583146527,
 0.22716071417309514,0.9762611723418495,0.2654749023723808,0.9892389050268431,
 0.20829590516231244,0.5560840384629849,0.11188219757997153,
 0.07475721852745743,0.7312902470399002,0.23486159945338247,
 0.4697671913552881,0.1197537107773593,0.25090337047976774,0.2694914454890923,
 0.2324212631474778,0.5893146150476496,0.8668642553426189,0.8461027807500796,
 0.6363842990857307,0.08424887863479824,0.9344518453282221,0.816337863871465,
 0.3905146377420249,0.6571840627024231,0.21614058773613332,
 0.007369232098427991,0.44738434822644046,0.6195090215359034,
 0.47856000668355825,0.8300254076068797,0.39462516817543436,0.426469841057715,
 0.30779782629088226,0.12519179018396098,0.8775040871780552,
 0.9976790367213952,0.1181488396617637,0.22774268570396383,
 0.11464390411435399,0.35145916379278885,0.06749342789667334,
 0.4837855286717474,0.12361055809340904,0.5728720885580784,0.9867759034342637,
 0.8158455160531283,0.10844337773716639,0.33441936981758813,
 0.5475104888400714,0.47074752145663046,0.989361071735788,0.03246195190754564,
 0.40064288628932687,0.3091898623666527,0.3662738816758566,0.7480268430625807,
 0.27802020190481125,0.21935137146676897,0.5260544207543364,
 0.49829490447458546,0.10956374195324425,0.255362990070899,
 0.30023767940397506,0.9922869527466985,0.6782562807256128,0.3204946516876377,
 0.4629394696121566,0.36469759544686065,0.4140568768917111,0.3324774444363743,
 0.521995882931004,0.2830181172198216,0.546586398677603,0.9973693811293791,
 0.023858090798730025,0.8400467242432111,0.5819873247616106,
 0.7226037999831555,0.3092386649809278,0.2963877131031951,0.466063250348108,
 0.04135766798785989,0.36607296002530054,0.8449690097960783,
 0.5047108136779594,0.530608835482734,0.13563297285742493,0.6355552962795584,
 0.4445575393479406,0.7100248476735306,0.5452368867930243,0.7833650341874145,
 0.1812059322187668,0.2211139355877042,0.4480130390855447,0.525672621319025,
 0.3785254159113143,0.19872558861052436,0.11072865170072943,
 0.4423674129796885,0.6718377533008428,0.5917057464670742,0.38277382206386656,
 0.29790061994701467,0.2833419035494177,0.7185239703932456,0.99437778493351,
 0.9769313188604147,0.8053576852295936,0.9588285344865749,0.7251262214076015,
 0.312459183545289,0.5510842636055007,0.4833507567421633,0.10195321505487653,
 0.42955755314757627,0.15777626685216073,0.658183790141296,0.3839356797339222,
 0.1645103490147375,0.6899668636527825,0.2429892318910617,0.20159512365697507,
 0.4526654256886794,0.21083599528797758,0.8840328305089316,
 0.05796530637570707,0.5777780168816352,0.8750003747879255,0.7628876546559453,
 0.6286323357279409,0.6501355596667083,0.26068506073485265,0.4826673611542531,
 0.23631924192246156,0.07941574413528574,0.6417666034577745,
 0.4120527940092853,0.43171832857668213,0.23487643353580678,
 0.5757521231196842,0.71551851716047,0.8658179825020553,0.654472449670964,
 0.3766973460601317,0.9995417161813885,0.24353672647840408,
 0.09607773958542576,0.5623714513490965,0.6298240917348558,0.7639679747552749,
 0.47199263178558903,0.232901946881378,0.5300172305980909,0.29994717888168554,
 0.1109772986232368,0.34127102449328817,0.470464682409963,0.7646426088822179,
 0.7502337992079251,0.17418990188585437,0.48193247901725056,
 0.6884015789739567,0.6638558692106595,0.7971097875291319,0.2180586510408148,
 0.9838374369372965,0.19883649571698148,0.9642009467770586,0.58244256995586,
 0.26589066503420966,0.32097480399264133,0.7599611156420418,
 0.9809625718584406,0.7450915587488444,0.04947058143602345,
 0.23289289152612258,0.20237048115910128,0.04681624756362679,
 0.6388010612237024,0.2605134214919491,0.5384972344975719,0.9794154395170994,
 0.2799149062007974,0.08813566155294017,0.21493316586342448,0.411292121058042,
 0.8953035808148524,0.15295368433594136,0.54660467199011,0.8938016003766583,
 0.26322632124286893,0.4653731729468009,0.8386330316429295,
 0.05630199406128755,0.7635664755427773,0.8145638869772831,0.7435619665924975,
 0.20418689054253814,0.22749212978025646,0.4479432879684826,
 0.8667088237133027,0.8137540920648192,0.7080396493145267,0.31981728581918034,
 0.15236933930026964,0.4715369896257968,0.16007024642916645,
 0.6740045421516916,0.6258986157665615,0.9310978241909577,0.7585849387439322,
 0.28051005314300936,0.5979591283323825,0.8810527962541911,
 0.11478209040241882,0.06405560244880593,0.7415493916581221,
 0.3132694890939278,0.12344476981762853,0.15698199053980755,
 0.11320496748476316,0.3627945688654375,0.4606462173668724,
 0.24590495859858585,0.5652245809985122,0.4689262765955696,0.8815001883635897,
 0.6528966472206499,0.4260482254282262,0.0429613358596298,0.9902235329743405,
 0.020323556713127577,0.885519679968388,0.340195971537518,0.04932967690095991,
 0.2152070163346479,0.7967116035725842,0.5331357433684399,0.8801647428194537,
 0.7615546344921377,0.5944885171810428,0.6134256728312613,0.4155497990759025,
 0.04401581896034468,0.36496897414359974,0.39860875172998245,
 0.3109769687818881,0.7317541338224557,0.1674965396602195,0.6685767176234017,
 0.30071358131321246,0.29575015361379886,0.994574872358188,0.9226925831413966,
 0.8826659740387756,0.9988780410386647,0.8026252607285258,0.6426134945041988,
 0.23991407747918259,0.958850273148105,0.03854996544821676,
 0.14854524614884057,0.7957524172626883,0.3693730102565347,0.4164506747987824,
 0.27427465264929374,0.5125878686747474,0.8307345358897463,0.5596380625609336,
 0.07668251390178693,0.6942879300836784,0.6235141462623195,0.659675211875192,
 0.468775907602463,0.13473941253094956,0.37676303837867076,
 0.13838656912793124,0.6459283883469483,0.11976853933904918,
 0.05407611947455471,0.630473118275735,0.9693227528052624,0.6070792617531081,
 0.9486890643059736,0.7837532881370615,0.08203247861997653,
 0.012644925860354705,0.3018080172088097,0.2243640757608072,
 0.40212082935172555,0.5271798059617017,0.603853616762762,0.828363299113734,
 0.3359851318794661,0.4323075683694203,0.059604208228559385,
 0.24885457277347633,0.9477778666831452,0.7718370315685119,0.3899638286268645,
 0.43147119865528094,0.6400033766927675,0.9339223050265554,
 0.25910981785092035,0.9980287004820659,0.01767019789717017,
 0.09032542835289503,0.4696170152834731,0.1754441333591088,
 0.034914412250146354,0.14881847411048754,0.25037272364722796,
 0.842834484589138,0.3373283585861103,0.8889669327198589,0.43874592437500015,
 0.9536303868428729,0.059688848423737895,0.5112384695147939,
 0.4769082248311246,0.6607431736225877,0.709871434013118,0.9538902474372986,
 0.8411480345068637,0.6436976419684213,0.6951985610689175,0.07271762983119956,
 0.17313614252136345,0.8231073631922827,0.06794613588623166,
 0.3893481362036123,0.24401334837054944,0.8909686223159975,0.1307084898615588,
 0.6747559259621715,0.25920477347565596,0.417085289125273,0.7613262035649823,
 0.5819876146550195,0.3823471963328603,0.887338567149819,0.07895347665186747,
 0.4826216551154384,0.8845986708773032,0.7262339069497713,0.4412842561124506,
 0.9662458173925076,0.15183134425245937,0.6445721835921873,0.7619639950926734,
 0.535358174368529,0.19832459391746904,0.3656090236188798,0.7851623425348323,
 0.620519059402118,0.4307261330297655,0.2046037505827325,
 0.0035221556643384444,0.6202341706077347,0.6555376942426063,
 0.856804504015673,0.581857479717783,0.6891507907248335,0.94095816285021,
 0.6219872679321,0.5642264597458824,0.37223101718109997,0.9637176522545186,
 0.4085092328777974,0.793683657429968,0.677172726550161,0.2723295782613382,
 0.5615436122731111,0.10751869607210651,0.08825835571667551,
 0.11138043934927577,0.45306217851917185,0.32720455616159905,
 0.08958539182508762,0.03599411969226485,0.16315596458354853,
 0.8419509242283526,0.19971968710089438,0.9317900393989077,0.2982222442540736,
 0.04479660905664984,0.6193067612801659,0.19281871030797648,
 0.5084829133085178,0.8652548986092179,0.38115148235232943,0.9639973044455435,
 0.512002450309327,0.018389861007199748,0.2061155619935756,0.9512438023773859,
 0.9874872834144799,0.6841599258349467,0.43536870871123223,
 0.11261841717973564,0.7583642563485284,0.33567976986340897,
 0.2620386145356649,0.2947199791294811,0.5187880142191854,0.42080358620797365,
 0.9060871579490454,0.5163434135748332,0.44503802131023473,0.7436861098315564,
 0.730294731419583,0.6625193176600005,0.5851365320074211,0.09009042895939334,
 0.3245200365501525,0.6896393218676538,0.9993884917658601,0.6996275280226087,
 0.07003022158666627,0.791639942116109,0.029704093514479113,
 0.18147692375904556,0.660115495450326,0.5279191422146595,0.16889603361389416,
 0.9291079722184039,0.7279675342104371,0.08522613686602565,0.8801861073795105,
 0.2861856165040815,0.1397301947581242,0.37768437537832356,
 0.13894191483950713,0.23354915308241997,0.6690772859007188,
 0.1173521890567083,0.2765008716721086,0.7244778279987047,0.6379401180892067,
 0.8724451282520136,0.31317414467576776,0.18382483738185307,
 0.9201422094814412,0.20551566023494794,0.37574070166232965,
 0.02919463754602203,0.8616535110459357,0.33184272277946614,
 0.4008386852659488,0.8942257761745293,0.9470183925407125,0.39866646590607424,
 0.1506183556681988,0.7219629337396722,0.7317583963393028,0.659200566360875,
 0.2817752085678683,0.1287171099583304,0.03377949233805955,0.7764140756176888,
 0.06689977826038918,0.530027710037253,0.62753549015346,0.9692576154700334,
 0.8249893906849612,0.26637991980282927,0.3427763871420052,0.2051308819597406,
 0.1357609907594055,0.7242573463345506,0.8391606600171583,0.6328946937364563,
 0.16853907488352204,0.11601608425344256,0.35015743072467376,
 0.8269150783371716,0.41875989670475555,0.3713106664272834,0.4169979608108799,
 0.8631784113180954,0.49324867945032025,0.1745904453618251,0.2178017792908039,
 0.6010499976348416,0.3978884875769867,0.7094223244264629,0.7484934350635424,
 0.12542385370502396,0.1607832921857152,0.8023171285929408,
 0.38686333966595343,0.19021519307860757,0.630248275776793,0.5825740048179006,
 0.01392869504195382,0.14623356296343992,0.42036507992853656,
 0.31817389906867233,0.36086063610065566,0.7327479898407574,
 0.2976576035018228,0.7589968841880046,0.7793137486597704,0.6841832734575685,
 0.6573166219745895,0.28664694628159404,0.020003792281202526,
 0.9416980397251472,0.9896053330079005,0.12586224839496596,0.6995800864076391,
 0.5434531200489856,0.8899772110124698,0.7722255609434244,0.28605057992242244,
 0.15857206432824922,0.9241365615647563,0.729132714536029,0.9055658722166697,
 0.08415910645692615,0.48560869127631767,0.9042483954843108,
 0.6153347445731678,0.8925497386008848,0.2291441179521504,0.07101949188015089,
 0.3533932674999214,0.3394056119422949,0.11806935645220318,0.3072791665493747,
 0.5882504832114006,0.9432641017772059,0.9072169106688659,0.5259893868410841,
 0.8754536455066576,0.41534863584140447,0.8200861743510566,0.7089105083073218,
 0.9832694980346047,0.20965260487260406,0.8115868629947097,0.9354630987851696,
 0.765713952197366,0.042256991443301306,0.8146934455022159,0.8542617864339419,
 0.8784830924297383,0.6760504057868743,0.2580190589929443,0.42860829356241714,
 0.9774315278791784,0.9516917071259369,0.8005185205622758,0.9843002555512677,
 0.2699748208820082,0.48004147546184106,0.17635002553887613,0.58151431202055,
 0.056546260483469224,0.722449113611575,0.36303120316274673,0.822716750367288,
 0.6869344211807468,0.8359428870057262,0.6615995203115833,0.7771774835502785,
 0.11638067534591112,0.9036395561812381,0.8319479867199873,0.6645912177990181,
 0.25329296543024116,0.7000613315059526,0.8408360365190567,0.8420709243890129,
 0.30334336510017323,0.5339443053254063,0.560078442151755,0.508203067512751,
 0.940607050776335,0.7845415756486881,0.7448730149790136,0.4472759706625071,
 0.33152738728979936,0.6715982570862024,0.22748964929891158,
 0.3503238695400994,0.23484267141097992,0.8278687710592705,0.5108950672491398,
 0.024382590279988126,0.8252079201514739,0.9204649514144696,
 0.3420896149791075,0.34720738314334954,0.1145019701211546,0.3017921501179913,
 0.06026620647618641,0.7848853874763366,0.756567513608037,0.2920575172725759,
 0.20282121216444793,0.5818268744504532,0.5789590140049956,0.2445784317320292,
 0.20997416303444438,0.14258496302952972,0.5673247374798067,
 0.2597450025466238,0.5855753933745387,0.9427642533856267,0.48590390195280186,
 0.14745879415935992,0.7904346336888446,0.9323651413844878,0.7062103609285719,
 0.7606690716051787,0.3443609583895595,0.08159654635658786,0.5285140518612946,
 0.8624478951888185,0.5276400957398802,0.5048768209595673,0.3500736460589685,
 0.27776558342442026,0.5376981316944004,0.6521418674464101,
 0.028261721109712146,0.4063018100083088,0.07313767427457829,
 0.5134443612148369,0.26524590330307285,0.12515793239892958,
 0.7995688496145362,0.31631615364597854,0.10457274949947393,
 0.7137962226245846,0.3538802992052722,0.36089862025727204,
 0.17901850367711647,0.17491972577864823,0.5289503064038439,
 0.4945532451639726,0.6949102318044464,0.6767078600648982,0.6387849310957905,
 0.2383819941184112,0.05956137855031418,0.6943140843946988,
 0.40702736156091013,0.6395162754243471,0.2631867767122623,0.34157598675029,
 0.3014457414498444,0.16841261934854068,0.9768377008527777,0.3014560235141881,
 0.9694817220001255,0.5760461851949149,0.48391613044658166,
 0.39387337672010003,0.6788344510441626,0.6778499565166446,0.5271220722265678,
 0.2770596654937514,0.8980992776489346,0.7652550253623547,0.7812555566144014,
 0.958558329432789,0.5066392375882045,0.32568206742938877,0.6741647647237561,
 0.025962431875238057,0.6684099501602451,0.28502136277272994,
 0.5536236839462454,0.4377987347038135,0.6760437917597196,0.6461754527421325,
 0.8142195148751789,0.033173877897170856,0.7826145984066505,
 0.7949516766177246,0.9450585955197214,0.3974220815274775,0.45583475393043327,
 0.008964935357943404,0.7561961699712882,0.3662530676447582,
 0.6087377893202103,0.6346967132156025,0.9848358861674698,0.06766030073175755,
 0.8398473125987613,0.7817161093364363,0.9437160083350266,0.8636746585447785,
 0.8607109712209946,0.2918167799685887,0.3789256242352662,0.2587744071133745,
 0.8821015887968017,0.49400832904227066,0.17813408641080186,
 0.9395805367746948,0.3170732358178122,0.1621031679092988,0.6205044742174047,
 0.4119771735953527,0.7503265511971867,0.9046802158638125,0.04617892728128725,
 0.646320045713846,0.94973480658742,0.5157142749554877,0.2372037383004162,
 0.704234035010949,0.862117901803815,0.32679286003746677,0.9367625133740428,
 0.48319528517129695,0.5110497568774823,0.3105957564620274,0.9643400376188125,
 0.4732094818076098,0.7608939619032378,0.14113577107690856,0.5498292495313074,
 0.2084059730613489,0.7424949584204747,0.0885406784005126,0.04722599102119773,
 0.6008382324577287,0.845624305758067,0.41796288746142574,0.06967244033716402,
 0.852404264707018,0.5548374944435008,0.6302271971425103,0.3041847906617269,
 0.2538799772979201,0.6524891357565592,0.7726727653667556,0.22007433409460142,
 0.4938295595326413,0.9337114968007295,0.9484776205125748,0.21610279817654265,
 0.568079916471621,0.7354534840044785,0.8343190802580489,0.45895524319215153,
 0.1813339759594541,0.6904872173846381,0.3010687477767744,0.2860454834214434,
 0.3859423688903374,0.9932884036948239,0.23436798948015192,
 0.21179880943859897,0.10017931542083036,0.29579198066923984,
 0.42589321250328216,0.4908401819433824,0.6608383434357071,
 0.40980014792506947,0.6954475157760726,0.22757060322162004,
 0.3555407587683199,0.7377117861221237,0.8233101326152796,0.9046012608608558,
 0.4248093466035132,0.7356671916363307,0.8520686210048227,0.5910533247808379,
 0.44914971531835446,0.10411268426856624,0.5774944411680174,
 0.7632588352915616,0.599825690747046,0.06460240524027172,0.1981557295272499,
 0.6550821658251457,0.4457498430441019,0.5356539852034135,0.8799789018204306,
 0.4476920803704879,0.5458231909429723,0.1772409813307798,0.7665267373258904,
 0.2547021743046852,0.9997012670149715,0.1843774844310948,0.5551000837204954,
 0.5112614843638844,0.4773243372333209,0.0713717647181299,0.7531136806865999,
 0.714936158046037,0.19255477457661963,0.5518281637153237,0.8346248934888486,
 0.8818852621056602,0.4886991522723071,0.725756167032575,0.5040051759359799,
 0.37012116031634856,0.16428995772407207,0.1896246509530568,
 0.17337910518248634,0.12412352635564106,0.5447242017990647,
 0.10370103909300687,0.43743352101656496,0.6505088836846261,
 0.8953817261084516,0.4938908598867128,0.9415438022647543,0.17215158421724708,
 0.2830351107234582,0.659603184772134,0.9958587196915398,0.24053684044033463,
 0.7324112652028087,0.04261501093661835,0.7581052296910116,0.6041882916487457,
 0.753756672270936,0.8089106325763117,0.5926667387757593,0.9952428768757345,
 0.7892388987025604,0.016583639055498867,0.2189747362014154,0.916107345031553,
 0.5845227033441731,0.27665481237551903,0.2670553853982166,
 0.09052631069447803,0.009110843675330105,0.9874773103488051,
 0.16821487388574674,0.6982640188440474,0.8361759954585669,0.1696493365407843,
 0.9245981051776286,0.0566668505525586,0.9874843988262846,0.5465037812444192,
 0.6812272317237583,0.005908795006722811,0.4069797111165173,
 0.009612076171318096,0.8553249985737837,0.20878212121453654,
 0.5599023549077673,0.4138422944784417,0.8996653971031782,0.5641485088055394,
 0.6397563583426833,0.863292205919822,0.4063545709442815,0.7358869882757412,
 0.04152721901864842,0.35078139960321086,0.47912368005390893,
 0.6652402004115359,0.8612429973013199,0.48205639439615056,0.9932079875722274,
 0.30456396375964356,0.1800262599214254,0.5438049821012815,0.8445050701923745,
 0.16590010744565986,0.67434774225547,0.781823645464846,0.7420478892499065,
 0.37831538660193365,0.2290799052400988,0.12687812856926683,
 0.3719778600234984,0.7985529445256466,0.8959833793487451,0.4804199173762933,
 0.07115924065957646,0.9500086687264684,0.43683823231104757,
 0.4673109189111978,0.39996736760744045,0.7211413375321213,
 0.27979572480802284,0.7995638877611867,0.13882585913192336,
 0.4606199555584294,0.23620953041115011,0.871530795050275,0.5798130307154679,
 0.14891698418043975,0.6402129179504272,0.5096527284209874,0.7653720111430523,
 0.42766466440739404,0.8592862664147436,0.36365238108588427,
 0.38810595898894273,0.9156247280351129,0.13211329348813416,
 0.7631649169348281,0.9500794874748186,0.012770145756420326,
 0.5104639590689207,0.10533908709443529,0.7988126328095666,0.3527982476502153,
 0.08252142222217163,0.9328113055527419,0.09231257485857536,
 0.4703081217890215,0.011682149537930009,0.22045148449319318,
 0.9540829263689165,0.7061095425684223,0.9320577235236885,0.6964994417418048,
 0.6199004386831163,0.19963009722307312,0.6301052358566025,0.2514699501311035,
 0.6581330147555555,0.5810338339486449,0.4958457786041076,0.4094336398360814,
 0.214170392754214,0.008210863284277892,0.1580509841533343,0.3401016056993793,
 0.15838590804371633,0.5759344097500412,0.8129567929950999,0.8654832273207647,
 0.5394137662352213,0.8328746369288096,0.3317945598774149,0.7602840355013609,
 0.12167845495820939,0.5103079179732785,0.9414184855169866,
 0.36048349198093954,0.13071267193023794,0.8967092286224548,0.689666007559553,
 0.17487531529196498,0.17058401395334477,0.6154556770059916,
 0.9275468173853606,0.21642101407296854,0.339107313939526,0.39937136313440624,
 0.5021815107390156,0.8361133237322654,0.40600396290102714,0.2131358467982285,
 0.32288060080519054,0.5471216106941439,0.532605996809179,0.7961627324036975,
 0.4202673897819297,0.24524694481618936,0.669113486859195,0.23176101229169843,
 0.10363301517891388,0.8181284483058566,0.09412875819241318,
 0.9667797609290874,0.8036857903953339,0.5123484433714045,0.6063563869514008,
 0.012631328233401518,0.6514226090552011,0.8993562051432756,
 0.6236072109108749,0.2404364078104475,0.7394875626155257,0.5649250593652437,
 0.34892137271322,0.5429307714885122,0.14077984912903063,0.6020042161494384,
 0.2433417147932615,0.5592659865954025,0.5804246271198686,0.4923147628320321,
 0.8792824077928766,0.7271870699836609,0.6668244917967981,0.13925052237991775,
 0.976551171435821,0.08166858059114335,0.31092698204297453,
 0.030977989460902022,0.18996521452201853,0.3052592164610247,
 0.8019444433460442,0.13731707358800171,0.02165291302726391,
 0.9153903071715115,0.3850899791896447,0.3008307176435868,0.5941460787492407,
 0.4560288991567665,0.4200125842825405,0.2199433383126752,0.0182947977635739,
 0.8949869595370197,0.21107294208185157,0.9328582338920437,0.5438448208490054,
 0.6147970164012759,0.9974836529263008,0.8762924405024559,0.23404806547555213,
 0.6787471630407342,0.08401372501549198,0.12042683631871087,
 0.2567407072561376,0.6971729671149859,0.44384289954499034,0.4760882482718719,
 0.6486997285144347,0.09976576899792311,0.9474671584582175,0.6283974296743803,
 0.5613558558945292,0.9706570334244868,0.4971523504723173,0.1974140428545147,
 0.8201077100835624,0.873499185609562,0.4107021940726977,0.8754277680449071,
 0.9284238716257728,0.14151271101044383,0.7382497070767313,0.6959323564687384,
 0.435017616913563,0.05051992223541035,0.26706939846826394,0.4880187581792146,
 0.5916111540825756,0.3474524592325039,0.38115833609577765,0.8391775013414544,
 0.5788724879987215,0.9992108177693007,0.19604069888308895,
 0.29898436842391374,0.037279437467059306,0.9170915075601839,
 0.9225386468783188,0.364782410018477,0.18848074460659459,0.7760182046152002,
 0.34644832484755894,0.7146115419034649,0.7013178771066297,0.5640649335053383,
 0.3002117674203445,0.6844202117626629,0.22307087528480152,0.9056311677577017,
 0.625603689285989,0.8962923182892192,0.31490526725198886,0.02432974633574092,
 0.17513434935539218,0.8324781800439492,0.003480200200732808,
 0.20190598023259265,0.33421786139501886,0.19660893175658234,
 0.6062646317074694,0.1366344646138279,0.5620426850748252,0.17746164483803628,
 0.2194438175234441,0.9122630937572653,0.737604339693776,0.2713009558224655,
 0.6796828090628375,0.33527307604605483,0.3307087121267628,0.7949426449506725,
 0.6542967850051521,0.41853398904055505,0.17949657000656138,
 0.22309899735129524,0.02669966632567189,0.6136912644705523,
 0.24305784561459332,0.47885541934657017,0.707423570682982,0.5251803535371997,
 0.05448048388222826,0.3748250907405485,0.2730413077849789,0.8689714951617384,
 0.25266737284564145,0.9544887066395933,0.8093871429215106,
 0.08093341802744825,0.3857464274143416,0.5395507563625448,0.8563472794356772,
 0.9110765807806426,0.7538035190711028,0.8424877420892394,
 0.009450184603198908,0.09709999905710942,0.012120704770462898,
 0.4049409615229488,0.15461068149704893,0.5617993114173033,0.5842091099847787,
 0.8395124750202839,0.12519966627001233,0.5050463572138407,0.9834435581234822,
 0.8440310460977561,0.4470049376861104,0.07486956507523757,0.321876551024911,
 0.32664376145191243,0.38476721929455715,0.7041181293485415,0.271191927418556,
 0.9487947830339123,0.06843637667429803,0.9261897099528145,
 0.44579259120500303,0.6510086648970856,0.8040329503176101,0.3930063684201984,
 0.5058861008199957,0.5342316310093802,0.7046662861766073,0.8114865914166474,
 0.6412581676613149,0.7721053765720881,0.7504319188810475,0.7946060954421361,
 0.869412353445599,0.9210302351008569,0.9930856890558469,0.3226178449461644,
 0.3687665059253704,0.8513295188933607,0.44723831002760017,
 0.028018386554808594,0.8252428211825709,0.9166047327561372,
 0.3156157371883991,0.2103137225557966,0.6042920630553882,0.6950046698747004,
 0.23661898571948592,0.06890370345232,0.13298449824868075,0.3416408316345242,
 0.7700614329885217,0.09377630064077225,0.735985607911543,0.6376640695565051,
 0.22237258146846428,0.4017709936915187,0.9011861283297451,
 0.40180930618699506,0.8962218349366411,0.3293098841446145,0.4429145273332151,
 0.5812625405826061,0.5104541811077551,0.30288062309183805,0.5339636378525486,
 0.25477737541607914,0.9441410489013262,0.5886159199549252,0.5162841476588836,
 0.8920928401545486,0.5675086919566331,0.5495300240029886,0.5811258423266613,
 0.37932781080870015,0.06014349050860868,0.27476537658685074,
 0.23288675431786077,0.08611505584751489,0.37871378062639416,
 0.7918908456120687,0.1647602448325487,0.6326881227780192,0.2696725189599025,
 0.24968600277176667,0.8554764949757959,0.7119746486765353,
 0.15038519598212297,0.5075226195625705,0.030434650853003564,
 0.8501258387210313,0.3567183787980217,0.4067593219890644,0.6139945399801413,
 0.7161915586492824,0.3243115208926588,0.33428883938448184,0.9616727468712871,
 0.03195023867829305,0.13994729726487765,0.624231585846692,0.7767953267678809,
 0.7105643266513798,0.08598755405093417,0.9447955720617267,0.3109202572534817,
 0.9208552678038944,0.22367290880412716,0.7233779713055379,0.7316675387646353,
 0.2946615214533781,0.5042688287114495,0.9641378688362297,0.5588531724929251,
 0.4070879116116499,0.1696122430241771,0.5560850254400014,0.1761633830256807,
 0.029894019852382048,0.2914167051741223,0.6488036584723536,
 0.5235463757761873,0.4556636062513637,0.1813567644883377,0.588341245048368,
 0.5559365982366032,0.5650799246170515,0.006222426709034945,
 0.7849460358914582,0.7269717971251495,0.0427120851344408,0.8100985382842203,
 0.8079526873391805,0.35425274971229437,0.746287605204553,0.8378237565921507,
 0.44330039787293085,0.9170243924289573,0.6003995161723747,0.8910742487258342,
 0.9722453736565255,0.7733074356397754,0.9336673296303247,0.40667200752048005,
 0.6176108996470618,0.738787254225288,0.5174060253568923,0.5826524852575936,
 0.9198203088414437,0.45930104946426376,0.8597163158172817,
 0.19020693246409603,0.8207530852853226,0.43093260701181757,
 0.4734174594805546,0.9319100633350091,0.15973650236891812,0.9822263920010442,
 0.40081356467894136,0.4412903104489293,0.5475748893956912,0.881500264120154,
 0.4981862385061965,0.7832969215491279,0.38128411984742383,0.5168003513672853,
 0.1299093391155185,0.34844625653193717,0.7134615199637442,0.3466531099501522,
 0.364398530372794,0.2717515800943424,0.779415177333507,0.05634723550621523,
 0.5886855860521528,0.26231913765638637,0.7695570070877498,0.8857441938378469,
 0.02622994518164845,0.7725284625269782,0.3083697207127257,0.7215079289845201,
 0.009575361133186022,0.7599180824576892,0.10427210354028427,
 0.9275556270710771,0.9506024633810971,0.9882267903397454,0.8092712944923803,
 0.41720997995793163,0.6920703435377573,0.8445095166721508,0.145930066361879,
 0.8962898298895796,0.22183768303708862,0.4155238170334895,0.3750685780008989,
 0.673624964723234,0.009504543181328096,0.6164115841934383,0.2769878267073489,
 0.8524420544697042,0.024080209191164492,0.35696274923261173,
 0.6517916612169793,0.3979568799632235,0.4788439000437418,0.2053388218105805,
 0.6451062412020939,0.564402105011589,0.4002889480779359,0.3085112252792732,
 0.9614844233002349,0.06520837503811683,0.696112618285355,0.587432690118345,
 0.12034990033620319,0.4420652380713478,0.32140946535191606,
 0.07186673898382068,0.2514531606174766,0.8105173959346135,
 0.40164580145774775,0.930372690172163,0.6089176246931843,0.6403690915759792,
 0.6357464890191684,0.27694899750780244,0.5859882930658661,
 0.05653375745992539,0.6931860388326009,0.5361529560645122,
 0.14208478038499817,0.506486442031741,0.6780298116558314,0.845712998761999,
 0.4757823569545079,0.40681944396880054,0.8189174889763337,0.6882802871675513,
 0.8330796679677659,0.17419155472894654,0.7005357572054934,0.6111923243977588,
 0.427319475736304,0.576210639289701,0.16069457081091099,0.11535138783695797,
 0.2867611886236765,0.3459542315232029,0.7958041223843437,0.758463150778793,
 0.23288675935864822,0.4556123160753919,0.2674976031807863,0.9621398486897643,
 0.49847609751467836,0.6014967204847634,0.5500378364513664,0.9548285486565076,
 0.4030528136899779,0.5443176956082851,0.9925651375811001,0.8889008804245735,
 0.43622652460904454,0.32424067042447713,0.3819090067550399,
 0.006890001179728489,0.1115966769072374,0.23432458952595114,
 0.8447437969946325,0.6745502535883587,0.46523538020369093,0.2526033860971014,
 0.9614140860546607,0.8778281780081267,0.09843513141660387,0.7804191473767563,
 0.4563608942811054,0.976243509242671,0.6940226878074411,0.4420765877026064,
 0.5445404030551089,0.045013488434555216,0.812464429126256,0.2059873552375997,
 0.8008360069689748,0.11364681754942985,0.026427003216675682,
 0.5232099326567913,0.8422458458888565,0.8256602525531164,0.122147576517882,
 0.055613137002034074,0.4532144153996235,0.2616927883776684,
 0.3123408237086598,0.40595304887687367,0.2608272193121075,0.5738790371751743,
 0.7083977786358724,0.9411765197759119,0.759495368950526,0.15286219091796127,
 0.9064465111541222,0.6422470198818798,0.6817518572286076,0.37600196752034576,
 0.9071221126362977,0.778550992720478,0.9238444595011138,0.3595849287436893,
 0.3948032835653348,0.04772341046666029,0.9052693729279979,0.9646156454487037,
 0.8759194120502298,0.06661501426737293,0.0998530056415744,0.7729370511717744,
 0.613115568840043,0.9445601184911454,0.2162190585179895,0.3254520999755357,
 0.9111924131749782,0.029651230394615258,0.08374959374707769,
 0.7738812985329675,0.5134419940927851,0.6706738941345228,0.6774107687915868,
 0.4218039003539761,0.8912147599027093,0.15228094369879464,0.9901665289597599,
 0.5773655908233755,0.12679263067832758,0.7997984206682562,0.0686681905070311,
 0.32296761073571656,0.005743644343737753,0.6807923851697185,
 0.4468742097370798,0.330857629584998,0.28600269993897265,
 0.055573911522830866,0.5053893793353421,0.8116962394955214,0.985443040520451,
 0.6959187067009007,0.739824245065553,0.7163842396029874,0.10085326234291503,
 0.7311723237527418,0.3506729591086978,0.07061802157682362,0.9518188352374094,
 0.21312855356440585,0.08153147861174048,0.4131928168491795,
 0.3854263967783944,0.8051818207218863,0.419314357893127,0.09501583980115758,
 0.029439077534101576,0.033949726246652934,0.43616909333835785,
 0.7118627322670305,0.9189198403163222,0.8061783443810641,0.8651639976172929,
 0.6712815298779454,0.9009285291836115,0.11886899424380182,0.1235086612784928,
 0.6042383172503278,0.6013307449273642,0.7324551839713822,0.22066008935638415,
 0.22330819025956405,0.02762338606111736,0.833649635722654,0.9350220438554513,
 0.22492559286014058,0.6510852088527898,0.2576996741227936,
 0.08713501123737621,0.6703362662701626,0.37871825284354255,
 0.8647191239671413,0.4553947984001836,0.1995279390209007,0.5974024253839919,
 0.4913997267717649,0.5643119144955935,0.639483316189614,0.44794863528186757,
 0.24157738919879168,0.7726007096712514,0.365660774402195,0.535635636754989,
 0.6722476225119339,0.6732135300235391,0.14830201853272307,
 0.43821847452477913,0.35710141073452517,0.11946325773074729,
 0.306433554951266,0.6215670767844835,0.14033836735154825,0.3130772484438712,
 0.2725334561294471,0.4916938095411467,0.8000371608622394,0.7870906046882776,
 0.607005933805014,0.8228973667812934,0.5341698394531884,0.8242043716688886,
 0.13554230917953936,0.19639293544033898,0.6385880948167718,
 0.5818664326071037,0.13093048158314735,0.5359608216396128,0.4134325358722517,
 0.10658933035215834,0.8406648470108081,0.9140333819429779,
 0.31932262696916514,0.02282050516456613,0.9983697807881374,
 0.3783446210404664,0.9132799651370227,0.9312709715679233,0.5441274381189163,
 0.9642655446270623,0.9935839996971869,0.5789073830720303,0.8356335612033109,
 0.14148907931859223,0.427980530219896,0.8781121730604267,0.24598370698441596,
 0.8767516887980147,0.02145722020178653,0.8589104723550537,0.8656769317680351,
 0.8091457848950306,0.5646911032162858,0.299998593940617,0.408393523858976,
 0.9847636585679922,0.06226309535980801,0.33097614646181495,
 0.6723218168314334,0.5555596344786266,4.812433518297521e-5,
 0.6692181584860923,0.6130539053042944,0.28064334910326183,0.7157444893102436,
 0.5756159362078089,0.7561777171273761,0.5557598388990559,
 0.032792639840391824,0.2519582932607367,0.944866246078707,0.5973265159508855,
 0.5979268609154162,0.4533336274604338,0.32784071477586174,0.5017074917498616,
 0.9747777714250693,0.157238772452426,0.656158687286704,0.7489929329131093,
 0.0443788804529226,0.16457859673487973,0.6586223967525109,
 0.35448160734568135,0.2316379295605563,0.5447579153751446,
 0.20548077984833968,0.7204730174462518,0.6782514928294322,0.7828963772873037,
 0.43650821825279695,0.8948739467084132,0.2853008084737255,0.7906589891884875,
 0.3960039906082856,0.7964629327345366,0.08861681084396877,0.1808086043131324,
 0.6189561262858136,0.2515557521375782,0.6130176686728158,0.6160857400232246,
 0.24289251684088942,0.1729023104412829,0.6480297512835397,0.9545824329180213,
 0.8201497634889833,0.34646622783126635,0.5595952993194815,
 0.41622590866809295,0.0011238022611486098,0.686476008047145,
 0.8347281093128313,0.9764063653996106,0.6086720781483671,0.3093535221208812,
 0.5475119700057545,0.04742802180129391,0.8443979380581723,0.5585275224225301,
 0.9052248382153678,0.8875250625387481,0.8393735265238875,0.9266231847238113,
 0.9174868445264752,0.177317092144071,0.8097588344783678,0.4108260267487117,
 0.13476943607739333,0.9029524844740605,0.3589709649244548,
 0.49395664285018603,0.5115720781406519,0.6722911305035346,0.9721534642670699,
 0.26070172294665306,0.6371671586167289,0.07588382966411755,
 0.9755782224552283,0.9411676011100891,0.7968249950294543,0.9856206156213079,
 0.9284180709952219,0.07781042909764657,0.10866917901585338,
 0.8899354498106574,0.17882208313313097,0.462411860194604,0.19584309795590515,
 0.639816624589734,0.5520017416239598,0.8530041857760535,0.016990851782940686,
 0.6815133781781048,0.44451681426076306,0.07664271170068826,
 0.4422104364379663,0.08838806515485165,0.7278762585887247,0.7507556805131508,
 0.9337894862994924,0.8331397553755276,0.6527663122543721,0.6548008822901932,
 0.15704563950113792,0.5977755480186342,0.08673042221254312,
 0.6736960307622217,0.31162066790345044,0.5440211211296986,
 0.03686159508827902,0.15676866987365035,0.43059957977575447,
 0.5108172078862857,0.7221246145895573,0.8290905809906925,0.8737292177073739,
 0.2462480682492696,0.4385385016464137,0.045250785969917495,
 0.8032185741231201,0.45228454363419734,0.25857421372231437,
 0.5342020574568336,0.9636608521625553,0.6235973528390404,0.25320849201439455,
 0.03696516862641264,0.06606098721221687,0.06748331287683629,
 0.054477100273803036,0.29918572476708194,0.7057069825671575,
 0.2201905815601899,0.7232280448746566,0.512182304180813,0.010001138246173324,
 0.31069141826875957,0.9111313149929496,0.598756631740708,0.9275224016283002,
 0.5674874367953323,0.9095948684684841,0.32640099944145884,
 0.15602924262420126,0.791142328078481,0.5242830025512056,0.3757315529262064,
 0.4489522353224291,0.19507263300169175,0.48639252402539046,
 0.8904989435361006,0.9704858131180802,0.735112054418126,0.5533589460041041,
 0.7367177335778827,0.9811551879721886,0.8337772068906703,0.7528216865665089,
 0.2999516185016845,0.5723312765616989,0.8948867741274056,0.7841769495591047,
 0.47434944502263376,0.004745854303709773,0.7614508528201124,
 0.15551831242533887,0.7182597370177412,0.032532427835139766,
 0.9053269689510416,0.7362991699257221,0.8903597017656987,0.16455353160313912,
 0.041383050569632984,0.2955106029430662,0.013545365443161761,
 0.41815063960682575,0.32787546242540344,0.614501908687352,0.790184465988407,
 0.6362817995126043,0.25933164069140646,0.5329680831974846,
 0.28699113058782544,0.38938834925159027,0.7482495785165599,
 0.23311766785525978,0.7727400543004382,0.8399453507875794,
 0.31468955277253996,0.16566757427107615,0.9705125233780796,0.865401099401051,
 0.4785072195587119,0.3487070749648796,0.41654048730382287,
 0.011812937012257607,0.9754969109337275,0.017289453936726096,
 0.5900188817699413,0.045164901785794465,0.07471533134928121,
 0.6145461938827073,0.9396106402619857,0.0032633544191035257,
 0.4363035839631071,0.5965379116830407,0.07467411290945158,
 0.38689033174979415,0.33247094751350437,0.7172482462051704,0.674823349309874,
 0.2375899097445331,0.21328420273112392,0.2856758306088527,0.5636356601195107,
 0.47537516724036555,0.21296885404211685,0.4522376521597653,
 0.8094795544672705,0.40002701872413593,0.6712662675512789,0.7192020023981086,
 0.7964706733636621,0.49886056855444527,0.97534737063909,0.6924536633031606,
 0.9967224335918766,0.027603560180872955,0.162311530534187,0.7985036917332625,
 0.7343103812382037,0.6122681513027415,0.8511545899518997,0.5722167197356223,
 0.3450350026599265,0.014338035143727001,0.3676208752654061,
 0.9519499245840073,0.3500831946058003,0.006079262083624637,
 0.9754850617671319,0.24374543387195502,0.5617869280153882,0.4476645509497178,
 0.004190699340093196,0.6683011067501311,0.12447803100355104,
 0.8748507262122049,0.19622031526684913,0.7994841929776533,
 0.16073862957327756,0.7259000456108152,0.4950660954596606,0.9956986667542588,
 0.9401675905310782,0.6502311642042922,0.2236594970017095,0.23967738685808515,
 0.04777040543542044,0.17350388354826063,0.9663295078445667,
 0.8906140402078555,0.6955422868771327,0.8884416585879245,0.827704387370166,
 0.6270778604740426,0.02679641221691509,0.6150683989550774,0.7615738830526488,
 0.3639577785502559,0.9622909162010937,0.2574737846242843,0.6645707220821099,
 0.9091337132296444,0.10622507868944686,0.4799026150901786,0.5261746032119181,
 0.9856531738310588,0.16057255717208951,0.1637147910041301,0.5373776285424305,
 0.8566426715509856,0.13995521175968662,0.14168038388384896,
 0.5716054421408572,0.5128268909348349,0.6998880581285009,0.5516979687546427,
 0.4976806147049633,0.4498046431996682,0.951904880979664,0.6423744304056807,
 0.08289526616893639,0.8313334102512542,0.11700456622532396,
 0.5268392632607588,0.5238444083970919,0.8692333350482163,0.5453622435802155,
 0.27033075455009437,0.627434791874836,0.4064230684511415,0.834300916171817,
 0.25359747001320865,0.8664577943984446,0.8522979906712239,0.8319405168673515,
 0.7779397970117752,0.23255301555140173,0.522715270337222,0.7962838409495417,
 0.7056734734618222,0.35160080983671216,0.18511692003909197,
 0.3894091804736811,0.9429711822367199,0.755038294399722,0.2083413091998848,
 0.9789718000087281,0.87018636821566,0.16816272739369165,0.7087208416983105,
 0.6583101716349528,0.143637513992953,0.011615139341426106,0.8932972187548673,
 0.14526707239460446,0.5468044051630833,0.7387477690015632,0.7780969894116894,
 0.48023696854100484,0.41951514273245505,0.7959069229264794,
 0.17036090981095253,0.1592817124542406,0.6776986534693725,0.6600857593071087,
 0.7819136078533864,0.9225789601019168,0.2725287886942508,0.7234814136863784,
 0.821384768620705,0.6477151572711871,0.7931157313026039,0.5452031489832377,
 0.019022489882584592,0.47162637350174863,0.6710047127573909,
 0.8916389075282238,0.34583747482120686,0.9497970182789126,
 0.42199516486844624,0.4324842473247954,0.6386009529412624,
 0.21460006338498983,0.8797811160774249,0.4657754840539541,
 0.41797016857083025,0.8492497527830811,0.6076974635268669,
 0.22150860480981893,0.6694692905830298,0.1745360486099129,
 0.09294398007889582,0.9832076906571101,0.7593193721847986,0.448064535223749,
 0.7519543220167139,0.005230345452852947,0.29094971871841135,
 0.09926886340041408,0.24138290612040658,0.8155473979690816,
 0.2079939524157337,0.9050210528016798,0.05199825374098155,0.8625536908729274,
 0.564358933707707,0.6707090822385487,0.8479512320309165,0.12126339100607852,
 0.057587239403380175,0.7846983019744606,0.9971668604634993,
 0.16631124601542258,0.7879873481669714,0.6323065555030032,0.6098457630916774,
 0.612910298767696,0.5553609013978068,0.14809629372458155,0.7286028481777896,
 0.5866642078971054,0.5290526997082223,0.7055769968470208,0.8959301527199772,
 0.26316604553625356,0.2844046433259334,0.5005225104656358,0.566203452325816,
 0.8193047834581989,0.003540657328036767,0.38007050167787515,
 0.17941796531302012,0.5002251510686866,0.9896299052174595,
 0.21369005156086085,0.8959605602730918,0.8392226691095401,0.8326921393803485,
 0.08062095474342701,0.9496168414984502,0.853549312784642,0.8108626722475496,
 0.4290247246263338,0.0831623053330739,0.8253249751694645,0.2786059984301341,
 0.004475012771108067,0.2850179178426586,0.7968860663117056,
 0.23659366552849237,0.8237405585631883,0.9869367635856893,0.8923098864685475,
 0.36934097705121804,0.44743386737841007,0.12902369460678365,
 0.6367537756074306,0.2615970999178263,5.521006691636687e-4,
 0.9373270832406135,0.6877186400243719,0.043711561531691534,
 0.24499828850487815,0.3804327183943661,0.6868477016721847,0.6419411104522085,
 0.9159787476777248,0.2241582883489669,0.3186476627730144,0.8443159559564279,
 0.34503233457291926,0.4900108511734227,0.012004112205955764,
 0.7376223634226087,0.1787761773987655,0.6561802460948047,0.27676877239044195,
 0.8347807279171567,0.09338373259122534,0.8405593906847458,
 0.25332412215087907,0.9459462870998372,0.2544155073426544,
 0.06107950727633027,0.15245384424739616,0.7406204540631401,
 0.8109335515538352,0.741825528554769,0.4796185620790837,0.05371876914147178,
 0.09707849140272917,0.9208458741094515,0.8113492226734591,0.8898886985619141,
 0.2420479549038329,0.557745145982043,0.6518192393356489,0.9578233885447368,
 0.6035569202939813,0.624644374241778,0.45791729769901046,0.9937744393643928,
 0.33488665357625647,0.8674120825622189,0.7827427898858959,
 0.26198506131907207,0.74466764900606,0.44430610367790413,
 0.026492395534744473,0.18850976019715993,0.20687106395114585,
 0.5163677521045463,0.709073578082596,0.8042831971021919,0.21360151188998677,
 0.7888736964835121,0.47792837085710227,0.8954514882727269,0.5600384525222505,
 0.1491428180429556,0.8293236026249726,0.6862176116455321,0.10300826482999026,
 0.5434938380065338,0.6772383381590932,0.3938703844696263,0.7629290484479062,
 0.5583623979690368,0.785108322032848,0.04058185195213104,0.7261453442867016,
 0.5465077681995177,0.7627208460214141,0.47746476204334143,
 0.15331685226299951,0.9223141298957394,0.6431489215609953,0.4637940607666421,
 0.7754612861218705,0.47760607512296716,0.12331641359437306,
 0.5558898271947875,0.1814984977108347,0.9386636910362266,0.67190412434291,
 0.9266055067651002,0.9605382696728674,0.20786989924425303,
 0.25268696901570264,0.9291068783443218,0.4912951936582378,
 0.10313171514725528,0.47780366894904447,0.2070299279235921,
 0.4616035276660406,0.710596233652392,0.6364042904362486,0.17723173559050942,
 0.08465477292054846,0.8537079986549337,0.8963691818375024,0.2765995414664667,
 0.7604782971540187,0.7465250331234894,0.10149139803204765,
 0.09732386312594465,0.8478146672098009,0.47807712806077696,
 0.1766368649153106,0.2048239957338508,0.8608388455414583,0.11615860586186944,
 0.5323885813446315,0.730141511336585,0.7669416427504314,0.502901601714574,
 0.634483723493687,0.31295338978225595,0.4952292387380066,0.09878648224587883,
 0.889229990531691,0.8588011746697836,0.30981475426284544,0.35810961388940343,
 0.2838113111342483,0.5389756706739077,0.7420543638476407,0.01602615869197699,
 0.21985985357380133,0.7780367228176235,0.19933032023115194,
 0.4957771643087563,0.1424797814817852,0.5682434540910246,0.6677779180145704,
 0.7400516385673181,0.39934322433912495,0.9315995250285276,0.8510484057504626,
 0.7356171350548673,0.5448179802068032,0.6473995800828629,0.9309461233697367,
 0.46826183473644756,0.05314693353709121,0.7550483742637059,
 0.7170343703949842,0.1660543732657832,0.579495607030857,0.230121268876879,
 0.20414642163928953,0.5150353022457314,0.6224486954640496,
 0.48654343730866545,0.8906432050029305,0.7368620897308578,
 0.25212127359671377,0.27280978460310856,0.5561794457275367,
 0.24429871435749184,0.9450403416774867,0.051630439735751876,
 0.8824832266510553,0.9992790961026292,0.6439067902910315,0.20585004594472844,
 0.6864052481738271,0.05501918395053529,0.7021834209141197,0.6617700062444852,
 0.432892044765506,0.39908088935183283,0.1682522186427058,0.07728282316734114,
 0.2684175713620014,0.5810490568580979,0.0040692701992992575,
 0.12612286717892962,0.7502281527143699,0.6906864480543318,
 0.27889035706183873,0.5036115322274966,0.4901811631973819,0.8331621830950405,
 0.20120725909672,0.8017270059316249,0.3490088737180199,0.40558304055953387,
 0.4508631409187659,0.4686624912424884,0.8249835252562745,0.5848759361409546,
 0.7089954399497421,0.35381900715060344,0.033384920385375105,
 0.22864416527916076,0.7817437980004641,0.190649992506562,0.5966833858685012,
 0.877315833976006,0.5389977256766504,0.31825992065081665,0.26676735025565745,
 0.2024570937341439,0.867643635454864,0.32212579418855125,0.889921040942027,
 0.542405704331628,0.6456733060846043,0.10394539478274312,0.9428712761513006,
 0.5351120969522415,0.4955023971316318,0.7647505475728055,0.8404496081917285,
 0.2967110203413237,0.03458191679366618,0.48996077500230706,
 0.16394441482533506,0.17256789110526716,0.9099145240394545,
 0.7257455925269839,0.4599390873516932,0.5858571941897861,0.3590333377262521,
 0.16244017582470482,0.48489615360480787,0.3206338054275847,
 0.6591527364301714,0.02205590399497892,0.5963720454400249,
 0.15620442637096088,0.6392358303665597,0.8541199845275534,
 0.09375181596759075,0.2769168230728395,0.0560956811858645,
 0.22949856684342174,0.9331885868468909,0.9695153753305896,
 0.30872133551673864,0.8106935565812841,0.13498790457450682,
 0.8150359203472488,0.29018578011361573,0.23455479147725966,
 0.5345696346655534,0.28179970492265727,0.36440029081729075,
 0.47699484527492664,0.5640326153163532,0.3186024461399848,0.4167216009825243,
 0.3203277613283513,0.379595262811409,0.5612275719853758,0.7098710062969535,
 0.21489205735950256,0.8914153729793529,0.1942955525522736,0.4463891914817062,
 0.132878129154976,0.28514262866380746,0.49728973020477796,
 0.11192076915634197,0.34412511032017234,0.4207340742865926,
 0.5335897240177341,0.8255382966968037,0.7700172723475136,0.9326924983001285,
 0.7451911531093257,0.9992458272766117,0.5697418870367444,0.7432814974603602,
 0.7367172456237767,0.3482210618143181,0.23873004611128257,
 0.32052928696876437,0.14590883689136602,1.8847453952364202e-4,
 0.2495730127267205,0.3671715855757909,0.5886240365184576,0.7490020551872141,
 0.7793455972567501,0.2095846607061972,0.14391777184628352,
 0.15912921455461204,0.0368958330009459,0.813185780401895,0.7967629885499254,
 0.3596169153013886,0.3871441411043788,0.296727331976511,0.07828158717114331,
 0.3332212728031049,0.49447077011010965,0.48749967394498483,
 0.7662010160147349,0.027444253533073337,0.9822681164583238,
 0.6857299928460457,0.12477310633921534,0.32398164994259626,
 0.10302062493546571,0.5617701072212774,0.46544192298368336,
 0.5895242569013175,0.6589667314376453,0.9206682071171339,0.40894113459237713,
 0.8958945522371788,0.6347931202993071,0.8107321003615628,0.39255566459999247,
 0.4865958768386581,0.532683616368925,0.0895757403221128,0.9114093417092158,
 0.6239526099505406,0.2644207826066912,0.42121603802717533,
 0.008592970412283796,0.3446869166153189,0.02595133001761052,
 0.2406796076744553,0.18719403679141744,0.9075643169785238,
 0.11633121092271792,0.2193714555323928,0.9982859461420365,0.91753710155337,
 0.4203255261500186,0.3789457344955923,0.6878499086158547,0.2814412143533671,
 0.905656455057463,0.901670439909207,0.8852014275546983,0.056333746495372816,
 0.1788026816794992,0.6746863124717061,0.6275294223317882,0.8149528711740189,
 0.067428697966333,0.5921055377963088,0.22659519857874422,0.8790827155985226,
 0.13790782446056205,0.1065537227604002,0.17132796938441164,
 0.03672942024902115,0.09779418323244016,0.6191887992555924,
 0.8728669780735957,0.15001903752931023,0.30656133827348475,0.980687846227053,
 0.7430676653910919,0.9907065803284107,0.25289344581449547,0.9737950692739924,
 0.27394310025029456,0.8527672471874508,0.8648626705281389,
 0.10077634561534865,0.11560150018182858,0.5297110805820173,
 0.6461128744822578,0.1407508019550604,0.9027564347002007,0.32859424437722207,
 0.13912255490765324,0.29615879426053504,0.5295240458543444,
 0.7273162349321298,0.8395244029267375,0.9234132762901306,0.8932082581380834,
 0.5836779006821888,0.0016400783652497086,0.4359196476841549,
 0.16942807101467716,0.6583606834726742,0.6856265775428426,
 0.11498972232994054,0.12899748346668216,0.9806890789911187,
 0.5709977686095908,0.5839515266980775,0.4132242620461516,0.9722289401488832,
 0.3153719869504712,0.24632382335092395,0.5697454844833131,0.7246837412947322,
 0.8967692688584438,0.8149801590136445,0.17074761766524327,0.9212684914964165,
 0.0779408619318731,0.3741708460660217,0.5939152665693533,0.5856986505961747,
 0.20002787759068563,0.30897266947842117,0.3329497368215685,
 0.8511432011519686,0.3888578527744385,0.7378473503242005,0.01150382850802123,
 0.08732239860514157,0.39413495150699707,0.9477498488734032,
 0.8685880305558488,0.08495783203440099,0.7850387136015602,0.5328963393003847,
 0.30972922491634025,0.5763743512267833,0.7979162948109251,
 0.22493552675048134,0.47419994509302466,0.02942219070536889,
 0.5937661047966097,0.7318953751798225,0.5229784863206742,0.48692769769664734,
 0.26458001671564757,0.01533928865598233,0.3117807583134147,
 0.5074770205529802,0.015195699399584317,0.9175742972910628,
 0.1632019183163792,0.709490366258352,0.8643377687841389,0.2124872836800713,
 0.871561829092258,0.9506047030292909,0.03736273858435046,0.341509500725353,
 0.9007503333599117,0.7687348877947682,0.0923929220495272,0.8249261975317377,
 0.9893266585450674,0.2110449985268108,0.20673785713729975,0.7007832353942044,
 0.16149201333302365,0.4735202523473804,0.008055694553718684,
 0.30238298341310565,0.08918992738971865,0.6183888796139498,
 0.3534603076646552,0.7478626828706045,0.17528960103319324,0.4465555392645295,
 0.0029809886162143195,0.9245148829116332,0.2695896392282683,
 0.3032217074854272,0.5562829002170104,0.3715082414826607,0.44806062852278816,
 0.20717992792623807,0.3069183922539742,0.8555783789288745,0.5565951287468565,
 0.659784181714899,0.6495164144171008,0.2163150017874903,0.10484204637830064,
 0.06129468312745168,0.9975686717239142,0.44221309943367926,
 0.47134991975212825,0.6769599698515033,0.9511517610284977,0.3825128834456273,
 0.8407603719506881,0.4360502873041793,0.6670115412448919,0.9746768684324099,
 0.5222173099898044,0.13390779027052224,0.0423665474335948,0.9861766510970918,
 0.9240215977059357,0.977075495511073,0.4649007247999266,0.07560310573610807,
 0.44874102860822507,0.6082351978024383,0.18423221374880938,
 0.05453894330734399,0.12042148017080545,0.46321643863121853,
 0.12008350381314858,0.4914610616188613,0.4687491240031393,0.8873809687486742,
 0.03366117378478817,0.2736873281913994,0.7332378720313081,0.3570947324581334,
 0.8139774156006946,0.3939921087203765,0.03986386582895518,0.8230289818377485,
 0.16293979693812666,0.7662106025629676,0.30032720023482307,
 0.9156040734138573,0.5293522800744083,0.8997187125363036,0.6767487778292143,
 0.7635741485115984,0.4915276162610397,0.3510615910087458,0.7140489250789649,
 0.12391316241750716,0.9681424573181713,0.9952888960603617,
 0.13486425075085717,0.753957809641572,0.9769237409561388,0.06992754513953081,
 0.2825106718792638,0.8285471597326042,0.3493640483002737,0.597313805327837,
 0.26526457566964523,0.13481258139740382,0.6796754770535309,
 0.9284151504388988,0.6704013329151123,0.7750063949943077,
 2.3405224632855326e-4,0.30284810049056166,0.4646975353542636,
 0.013772377883270459,0.5334975467232137,0.858995577539678,
 0.46999494157099964,0.10653259995379916,0.9545909625337166,
 0.5082360693250599,0.5863793061258028,0.42762451509565613,0.2632051741375441,
 0.03791863998419087,0.39488397368701844,0.4907604605157072,
 0.06985642752921473,0.22575774122453318,0.27503482154844383,
 0.5141400319310789,0.638210480702922,0.8192507021763294,0.8189296297371101,
 0.7430495124867535,0.8140663597946898,0.512714059304732,0.2973283800090041,
 0.13043481991240968,0.42450174794423867,0.8426901719445364,
 0.11201433387218451,0.6430801155405304,0.9215111112933475,0.6638018415550938,
 0.6847666042267195,0.8803916983887273,0.8691455440976682,0.28003932437157075,
 0.5490008976408766,0.04157743589316021,0.7941194783265129,
 0.38722467016010453,0.8379000922488521,0.7865218692523032,
 0.24691233754107433,0.9384572721599694,0.3260889815414275,0.68823906444926,
 0.8664655503797176,0.8104834773944781,0.7505815185959313,0.6571922880170711,
 0.7380678026509414,0.9929801625682501,0.8181622030756192,0.09404253867648482,
 0.7216364888244913,0.15807424147886728,0.6443650687016187,0.5083627141259891,
 0.5119392255451285,0.24657954866754628,0.37551290002982396,
 0.2815596456681476,0.9205543905936294,0.995141793576792,0.9734769308120236,
 0.3042618358511945,0.19729417472520328,0.6817801318647951,0.7124758285154018,
 0.8777447283829976,0.7506484817971,0.10650892596486605,0.1143233193839569,
 0.7270813986190173,0.8078865535551628,0.9199497724183144,0.7594819478865407,
 0.6743220384213614,0.46436066811300336,0.3838383078291673,
 0.06076672515509607,0.7446919273286127,0.2213393121846432,0.8214818856339376,
 0.7864943214851334,0.4685900547385605,0.785057233843973,0.3917199097105959,
 0.8560986494732963,0.637866525284802,0.5398038124513997,0.701452900817595,
 0.7816240853968832,0.6354832112667766,0.04001325070733275,0.4163815246695336,
 0.035147827630092054,0.25488456920128866,0.11082218534395605,
 0.6495972778110535,0.6884789919755538,0.9604444344400919,0.38004228146291774,
 0.04595748990568016,0.6405878268933585,0.5939455889836447,
 0.17424775483458943,0.4825941872087901,0.7493403442708841,0.5116735611372925,
 0.845397545215363,0.02693418999686359,0.5529823759887114,0.6192278868603222,
 0.7158133631048261,0.6918251394709755,0.19686365573831888,
 0.23535234560708296,0.08930916889741258,0.6875843555864917,
 0.05105420384360693,0.8663386323711814,0.5066058502359136,0.7194663514859237,
 0.4964870828179937,0.5604725938506212,0.30942795021654534,0.7938410139648612,
 0.9904106654490957,0.5699118977616219,0.35555874130070897,
 0.04681923879946792,0.4090049377158873,0.30068514366933596,
 0.7525599861465386,0.7887994123322188,0.4319600385885395,0.25610669647875195,
 0.4585884415176438,0.5451432299483159,0.6934433808610945,0.2263724113815524,
 0.3907479411589311,0.720217902308031,0.5109749542585358,0.3396977238583978,
 0.25062669746160726,0.7656786562669646,0.5021671755456676,0.8076832311016807,
 0.6096986531711678,0.13138458295559763,0.03117432619057492,
 0.7883455601948285,0.2311181658544098,0.5320084608694199,0.6552659373793089,
 0.9336154095243925,0.6970719699124981,0.9995298806548063,0.6669140548984981,
 0.5894922187679457,0.4263987663198099,0.6834102270548973,0.7957395873866174,
 0.008252716686383943,0.2409831275336386,0.6751339138904231,0.354040647610511,
 0.5102459131013781,0.989427547428263,0.6567554474350302,0.31791710422377584,
 0.9613232150738886,0.27400596500133334,0.5087178496174874,0.9643753152063422,
 0.3249507710361823,0.19603850907597675,0.8679771428730483,0.6285259739044093,
 0.1823807694777435,0.9813706269656083,0.5587926125022447,0.6839386793254492,
 0.4833365834365393,0.4758729323611569,0.7158816642201914,
 0.049306130632309375,0.7186079072336878,0.13734307077687868,
 0.00465317471530502,0.8359201789911244,0.3816297472041318,0.2866996617975318,
 0.9373649051698331,0.6720886399499095,0.4027346343785454,0.5629364287117714,
 0.8962948862535968,0.14757720716296263,0.8895838383992878,0.8556809479488549,
 0.8851475835232747,0.7809134579809268,0.09429906027266832,0.7279583777234057,
 0.5705285016057349,0.7298292899137103,0.48223278837221173,
 0.25540201858972145,0.2656376936741258,0.39071528696774016,
 0.9306512918281558,0.08562843545511234,0.48375934140697874,0.799400001318568,
 0.9334520720114377,0.005925192722812889,0.3080991197678413,
 0.5392537259023538,0.29970461649063174,0.6834564151110666,0.8377227087452582,
 0.4034370164108826,0.04083739092489169,0.19901629670278942,
 0.06663506579373313,0.6142202229203301,0.30358262625258436,
 0.32254406883581777,0.2811534148438861,0.4872596978846536,0.5281940608435607,
 0.5514208672611973,0.21410258825878137,0.8970947186292355,0.3838306181434482,
 0.4722278899372597,0.12796631299336492,0.9751198988753043,0.3432843706513228,
 0.3258320083868783,0.7215540813316885,0.7754794599535904,0.24679900179045433,
 0.20031119242636097,0.5108203793515442,0.6900485052590231,0.3858655632079169,
 0.47868643995746485,0.970505653044199,0.6045117478778115,0.11344754305828586,
 0.47218607273331614,0.1781025598164978,0.7457523416439157,0.4324281424087335,
 0.25399499174912754,0.8100956167869767,0.5308094196176913,0.6702987694004932,
 0.8927400470645889,0.5238178481853893,0.0979536759784082,0.8366709252943724,
 0.8853005954599984,0.7454449472565481,0.8029320692583604,
 0.025229458346549105,0.9098876763942255,0.7080985902613319,
 0.33913131472927216,0.1873869824236568,0.346784382849863,0.2311990846747053,
 0.25877793004124294,0.6751420377927033,0.5612057355909935,
 0.24136945043923452,0.19157611349986214,0.057917113145949606,
 0.6386227017920119,0.9584156026000925,0.5430562514133702,0.3314702625093796,
 0.4324107454085454,0.09624282310772625,0.2700429159387534,0.8347489742080764,
 0.5050498026576153,0.07435712640476289,0.6889528299050108,0.8945633541961486,
 0.0110928248954959,0.9723220816461833,0.9148276547587776,0.2764412653873488,
 0.606977002745586,0.03397361393276388,0.2095025587279945,0.5522045058966628,
 0.8806954866697917,0.6647525652379676,0.4426058022730591,0.5100071286411537,
 0.4435743742450782,0.7595605156349591,0.9538365970763566,0.8166434389135202,
 0.9979897175982193,0.6542731059087263,0.4379378933905259,
 3.2857245766704235e-4,0.45830319778698003,0.6979897542741744,
 0.9822696446724968,0.1943537828719909,0.2564456842529468,0.758703035514351,
 0.56963512068873,0.3187703102678031,0.3447981211253237,0.5101472103438902,
 0.6762285461355001,0.37214902375251013,0.07581941678366988,
 0.4292776995006118,0.6929024166446041,0.7684225348391829,0.8176390303325721,
 0.36374948834413856,0.8562001862448246,0.4186053164952811,
 0.16576980594849866,0.6014529177248776,0.4972266064955563,0.5828777966512224,
 0.5532374684894695,0.1167450953310758,0.9511413072712064,0.8745859046330327,
 0.9300883603911072,0.212113564324653,0.03548316595494294,0.5837009498609493,
 0.3466013897959299,0.5459894944496962,0.05639285994026977,
 0.39513475405547704,0.7214557087426577,0.3712027782047388,0.5752297960763131,
 0.7076206134524887,0.7941214637625019,0.21587427417929994,0.9419598125043085,
 0.5925153204283901,0.1784071301542981,0.06822571247972098,
 0.32790233558466786,0.21401780313066388,0.5059777983869941,
 0.8745659238895723,0.36101758997296285,0.6043935318008417,0.7650331867475276,
 0.2995537103543471,0.4924521691633955,0.9692044644332941,0.5563606220480154,
 0.8905922723599878,0.4165651533097616,0.9501350736260049,0.8032603102927757,
 0.024752586561860412,0.5842928399641272,0.47494753532262934,
 0.9232753112517916,0.8760977514402775,0.19217039036844086,
 0.01895948566151706,0.7669441968797296,0.6915550806717966,0.6740081399498352,
 0.5794285104851711,0.99157173477786,0.8722116063798805,0.18494857963805766,
 0.48861063023526474,0.2980415110906628,0.21492940922577286,
 0.24451185981531598,0.049577933294020715,0.8818227294921296,
 0.48705227199940315,0.6257328929077205,0.1828608483027241,0.6812011061897196,
 0.975148481341869,0.010210839860688714,0.27677180290913805,
 0.8530161858774183,0.5008693705801583,0.9867398437092088,0.3703393154821595,
 0.6909344471217029,0.13419438020384566,0.6139752834727928,
 0.027407861253156973,0.6614155015463175,0.506972287738417,0.7595275494296392,
 0.8999682294301268,0.4865413035604276,0.01649829105037226,0.492008171517455,
 0.17034432111565867,0.07958445396219382,0.3756718740554069,
 0.8641656957103099,0.3319260794957959,0.7119870367228828,0.4313659015110892,
 0.9148262788707361,0.132635886424616,0.5396839886585728,0.6929716410384877,
 0.6982590929297191,0.5333763908673391,0.5267376112984241,0.7276782547805827,
 0.07104210657128973,0.985434102935475,0.38964631498796176,
 0.40153457277656823,0.2915844966063129,0.5375250287607547,0.6938362919451323,
 0.5190079709386369,0.6039306091113358,0.01041445470432234,0.3298511726704253,
 0.3169081004927736,0.504098766119485,0.3124164022207375,0.6102182331148364,
 0.9909951945131042,0.4618348129836406,0.11667183774181811,
 0.28202836127220043,0.07077308608197685,0.2217199106439811,
 0.04111545162068797,0.6448805401316237,0.8159701040948505,0.8048234792251696,
 0.5617646570715785,0.4169890385168449,0.055722073028834807,
 0.9437867005118363,0.41284107893044386,0.9314797635461596,0.2032710815757267,
 0.4037017760788456,0.3919088430662153,0.2378013645667716,0.8089340070928772,
 0.586053520410214,0.547391078891949,0.43538189456557175,0.3446686363145357,
 0.6023615822848866,0.2212326057681162,0.4459102830377757,0.4985422443748625,
 0.010045844657378122,0.36850957686122254,0.7342943106303625,
 0.6172660607700743,0.14891286272348825,0.7544620246697709,0.1561390265909679,
 0.33557999139284966,0.9701633994260579,0.02051757690308309,
 0.11833732528536933,0.014669976456888945,0.5876454112791424,
 0.5004505303448822,0.9447018547476698,0.8656348717996641,0.9903016896097422,
 0.9730376071408481,0.4465253817021916,0.8272581388079344,0.5052433301966803,
 0.07996420018792916,0.1530916246606543,0.9169617392104973,
 0.09530821356858787,0.3818842197267892,0.21886927205963447,
 0.17572244165591477,0.5555536759681345,0.49703277528375356,
 0.16706234950594567,0.34385653571644537,0.9124538960998972,
 0.36198982209079134,0.7855336705139963,0.8597628522573774,0.3939560767131811,
 0.6635751309222102,0.3525374604545102,0.9246285911501076,0.481348408450598,
 0.22660974023726876,0.5581625086461947,0.5372388293493348,0.5809024791766587,
 0.3740069386158321,0.3549623272533855,0.15724705521151117,
 0.011578771963955625,0.5234838507024504,0.7731250039158064,
 0.2852409925467756,0.5055630387107374,0.944592723946079,0.7637926778338819,
 0.8321230429921003,0.8511987930221645,0.01908447293084914,
 0.05635545202358139,0.42046440330625856,0.1847878994253639,
 0.6409881299750514,0.44801470831358925,0.6568096990889434,
 0.035860316329613506,0.28652081651095873,0.4375207888970183,
 0.7901898741158156,0.6289580233679649,0.19972306648323035,
 0.33251619513072694,0.08428612909748812,0.7489318166134231,
 0.24314636976252924,0.3718264255007322,0.4916235988703863,0.567093354769257,
 0.011668050214972414,0.6118048793354515,0.6903684960483083,
 0.1752827713871532,0.4580851839373712,0.23578223045997437,
 0.28040181910911055,0.26941659308899135,0.5636560464756174,
 0.013950446409349948,0.023752755588007934,0.6241252223487902,
 0.8489479949289925,0.9335966968311474,0.956482482321627,0.6293430483576543,
 0.4113742172482736,0.20606604969433873,0.035214135001042646,
 0.21252421140717281,0.5434450555713075,0.9285773118580604,0.6055570458966801,
 0.4253748043382949,0.0943740515718654,0.9652699294001444,0.758544959094095,
 0.014116406180872954,0.8536449056124712,0.3364248710558708,
 0.06860279696248128,0.6651773896894885,0.38168900108848636,
 0.4756313858595542,0.04525182971697417,0.15861878623912629,
 0.1359469071587206,0.01254374335348063,0.3925944590329278,
 0.16937312801732773,0.890822726742568,0.1834050124594797,0.7839369737926055,
 0.9059716933343461,0.6964168214458378,0.3519854914895417,0.9621199502652309,
 0.9754569679734623,0.7081615653808405,0.878842724946462,0.33409185211000536,
 0.23594625131229596,0.9290334021603014,0.7972576816480043,0.9831867071526166,
 0.26373971183842426,0.12678343973636919,0.2831187972714104,
 0.20985103737184396,0.05613092133179021,0.163291773658523,0.495118324496135,
 0.2163140332438932,0.6555104385293516,0.5314192550151032,0.06314459044296983,
 0.8456481399742354,0.1287692552935915,0.6320343539517709,0.40021729562857955,
 0.86740930727186,0.23624913779682533,0.45612160550420844,0.24114869088059065,
 0.670206526380186,0.9975089001628606,0.14078408004828813,0.6507663079217572,
 0.7196286646569838,0.11822977076055097,0.9270467560990698,0.6034009460410608,
 0.7696451528867917,0.33653196521884954,0.4871361546733526,0.8288985161007392,
 0.2601936218439367,0.39114755276602287,0.28442858319157416,
 0.2892481222507969,0.9809062694764873,0.03239171696806875,0.4202614042317534,
 0.31358781927766755,0.7497878557151718,0.6840200751125536,0.6690602801360561,
 0.1492496462860785,0.8208168853994843,0.6094520466558517,0.7069893730366562,
 0.07546888148945696,0.5889035940853089,0.9254348996445697,0.6131728480442703,
 0.17881953270022344,0.7659621511339771,0.998650521136928,0.570431909391179,
 0.8508966324046752,0.6962651767444785,0.578959176515608,0.6129298845385273,
 0.5211746264139343,0.42299569117291935,0.5317832351736312,0.6652411224723612,
 0.5118855570949811,0.35551939477520333,0.4618760014600056,0.5580655888749315,
 0.33707549326228203,0.4912191100779957,0.819502702875666,0.42165386888898226,
 0.9325380748894548,0.6494353926386183,0.5277958680696329,0.4043410071195337,
 0.015342917741693052,0.30737196092963603,0.43653878082638986,
 0.9840766625027302,0.01736005728782719,0.4543938846330777,
 0.32381061758263086,0.95173211075854,0.12838598668409307,0.7702813529465256,
 0.6156041631492148,0.7596939459440952,0.09815265186851141,0.4239034205421436,
 0.24586379385075197,0.9921089794385476,0.40944245289252457,
 0.07992447660695556,0.23236309929689358,0.7229511614451521,0.606903019544655,
 0.5438533780837432,0.6230451576679447,0.04188051795002912,
 0.11957573679985378,0.6091946614205855,0.7840992303221468,0.5170838306441938,
 0.5858976428538546,0.898599074632382,0.32898195588499757,0.9199713341819166,
 0.69816024486299,0.07725733873998897,0.10919021559352138,0.39393953787493563,
 0.2168834330985111,0.5067210937318241,0.8242528533570246,0.5691512495201304,
 0.8634373909562985,0.4476813307477334,0.8445416938200813,0.1308730548634195,
 0.44113723865501475,0.18901094673199303,0.9383920709851988,0.925690994469259,
 0.5504717497010123,0.4903631961175642,0.7108240566912312,0.6716412803104637,
 0.6423805080641694,0.9325027190756111,0.7835067232501884,0.7927891489722261,
 0.7206577581654622,0.32758429426761926,0.7058634341943809,
 0.04732521089412556,0.19748677854235375,0.9010780251429533,
 0.35571569466395203,0.6386567013596237,0.05780109955960988,
 0.2269012846518551,0.22900063286852912,0.6123115555248952,0.8268628143754511,
 0.40596902181450034,0.46706203976711336,0.7997669318641232,
 0.2238614195428713,0.37907323146391436,0.5447160045587959,
 0.30003152944002376,0.48335493099942695,0.4244164616245829,
 0.26080367800241233,0.3763133479650477,0.1010960715984015,0.4945976077895762,
 0.6563661289130094,0.5628802529982806,0.9158238558776607,0.3388074886920196,
 0.8698795628261676,0.7430933422933118,0.5197471099101805,0.3259188503325646,
 0.36253909744452817,0.5029707055636403,0.1883732410616662,0.9246564789538179,
 0.4762195983857178,0.5469424698426788,0.48483663908864294,0.5314258529134769,
 0.6118042920612742,0.290637156105354,0.6594351010128678,0.011486023213514662,
 0.07423105996353618,0.15678341827907594,0.15801794528506374,
 0.8367687296376297,0.774857324473206,0.17532284154313948,0.200743527349577,
 0.3465804791682914,0.14161663188596185,0.931099382566684,0.562220895636787,
 0.9941231941338616,0.9243791547953467,0.08696861683751644,0.1075592962169184,
 0.05107094298083281,0.9437157603383415,0.48701457923928215,
 0.6436445305520957,0.8799373299593621,0.27843739365272047,
 0.24407506589719186,0.57369840178303,0.7982747883152448,0.3867082827040902,
 0.6970659455812973,0.26414264301685986,0.5211573137284302,
 0.23893788074721678,0.3730246257420744,0.15675630067689816,
 0.006859601087082723,0.5049927045082163,0.8804394307976579,
 0.42342957908660384,0.7693820255639129,0.6295388143843655,0.4660938518789859,
 0.934479274986204,0.9913517779651781,0.6931810889844099,0.13212880877082356,
 0.43491567447879464,0.4839476366182376,0.4776730041030173,0.7416889374214864,
 0.5700267785219817,0.3120545701021221,0.4924841807007544,0.6762437896306714,
 0.3710707058318745,0.8970349557289273,0.4173697538707969,0.775557318653025,
 0.6030727505681651,0.08529025297219917,0.5909781729895519,0.370673046295407,
 0.6926100130709303,0.00605248955561688,0.14726720126963433,
 0.14205955568051276,0.6878692732686762,0.05208033176955962,
 0.5811464944030751,0.8101359580923865,0.23904943769511433,0.37312748746589,
 0.36439739486194533,0.8712713862384329,0.3358545770955639,0.3438939346438773,
 0.1299106266052399,0.899257269411023,0.1753853938396044,0.4627240730079235,
 0.30672693087310354,0.4875239929182482,0.4225113373785233,0.6419684549926624,
 0.3021620103127156,0.9547307855692918,0.4621625861679608,0.7988346996550756,
 0.6180799279700844,0.8354517533295813,0.5727044959395639,
 0.011220629540911076,0.06366367052780131,0.14767530511878457,
 0.09723657227580262,0.8584366018481941,0.17782067497055132,0.618012005065534,
 0.13998746665188622,0.11627421610282462,0.1327963232956637,
 0.9133150777656384,0.23916538867499504,0.6017726682381376,0.3661765770027312,
 0.04461439259456723,0.7569783571621108,0.22236962098254476,
 0.09979436433270106,0.9225497024868857,0.5390594372687684,
 0.39366176006810616,0.01019397678981504,0.9470955053309554,
 0.7907803305935102,0.8033039254366257,0.40557349299901924,
 0.16186088600185133,0.31920887404584775,0.13267301469093362,
 0.7361950452835639,0.41060052051436213,0.47556484189391335,
 0.4979017656192578,0.3338911036589097,0.4883200127772961,0.1592276250697332,
 0.5074315015795063,0.5133363879623124,0.33866245767132597,
 0.09778261078277017,0.5409950467051969,0.7165208858778347,
 0.42684119994601044,0.6534422233623391,0.6642944144441816,0.9760520007201855,
 0.12618813132263418,0.05789318165292334,0.3391895574947169,
 0.45782818490176735,0.9049098857530664,0.8196688556017262,0.7753046816446496,
 0.18390703087268356,0.6092636980942819,0.1890379215420258,0.1627049698887142,
 0.7429662947821486,0.4132670706044832,0.06122161829720052,0.6267824792400264,
 0.4571614154438858,0.3451624087913233,0.27836271626420395,
 0.12570139758319832,0.3389064572335574,0.9130756032332351,0.5629057145195153,
 0.2338329515889468,0.5642430143524729,0.7649215658420533,0.32196942596013756,
 0.4414169257952393,0.7808767668321377,0.34805321074432927,
 0.18584681663754232,0.6757211616056924,0.49966751001460974,
 0.1762617992308938,0.6486305300235409,0.16198159226818432,0.516756410408164,
 0.9649463408461401,0.4884711589063033,0.28552266832711737,
 0.27464982166345386,0.2689138135633229,0.9086148541815123,0.5288000078408509,
 0.12960942769221728,0.7849874879671719,0.661223174764388,0.5462988536696024,
 0.0877025218056735,0.5865502542829039,0.37369380820637477,0.9006070209841402,
 0.8578809836812329,0.39857974055429135,0.3503528275283685,0.5598492704653961,
 0.06634647804987093,0.04827114817613409,0.9302661177664566,
 0.8424778598643204,0.7866835057080516,0.6243187058923213,0.9084570064053564,
 0.2748229154035602,0.18634025926427633,0.7383955116712836,
 0.22743107488049685,0.4039109209153481,0.4743724672765952,
 0.39538203850898845,0.266119226680284,0.4277313592917642,0.5722156327160054,
 0.763889409604269,0.5322445707012993,0.9023759201985623,0.638435813901337,
 0.6451553826469034,0.396035255729278,0.7432759497147926,0.32933433285600044,
 0.063366257481407,0.2599529314652095,0.73356511035288,0.581322811463509,
 0.696863022439418,0.04122855245534485,0.13487463431021562,0.7494082804343717,
 0.5831649194592344,0.5457457685480699,0.5698645170427408,0.2323884482963836,
 0.13357180273991698,0.5808572180768451,0.9204408688778605,
 0.01095735915602658,0.5986585445691368,0.9172789331981217,
 0.0077621852103464395,0.3176847230301215,0.5527612163437851,
 0.6631058069390839,0.002259957123734835,0.3414509376400692,
 0.9481757013834492,0.3487900652288295,0.1583586600013942,0.49388433997748526,
 0.5494065746488472,0.35570651958047683,0.16175209102679888,
 0.8558064378997491,0.5396472992374837,0.14538781953866275,0.6062455149289643,
 0.40315816907546564,0.8803913563250745,0.6825170330317979,0.3400423863349844,
 0.8509316778402642,0.7892235512357503,0.39412691464246374,0.790962556021994,
 0.9025940874958516,0.14154966347766829,0.38624463706461143,
 0.4242813211372154,0.5021042846553347,0.035167400354424116,
 0.2938999663321731,0.8316889322507115,0.885767641885891,0.16591163506547402,
 0.09658302824629872,0.3036647080471566,0.5500727521202574,0.6527740062896372,
 0.9741652462429397,0.19264415153124415,0.6488820209150297,0.667674599044028,
 0.1045423093152884,0.23938229774315067,0.9865596235959848,0.8324431190851834,
 0.894260782348788,0.32920975124137175,0.22973603646883678,
 0.02128477503323345,0.013259531821659731,0.32332761912322017,
 0.48022507467963926,0.36212161623095673,0.13181504193106397,
 0.6953239677691212,0.3618423026809865,0.1798196830047737,0.6039218963990092,
 0.2829231461650762,0.03035800837018532,0.5621929066282528,0.4035976455942042,
 0.037673208820751425,0.7620249827684265,0.08153857277934029,
 0.6954316332590471,0.8690992485535133,0.6959297157999123,0.5258558866356513,
 0.250858915535197,0.04973447541596521,0.6494707160242705,0.20310113146215714,
 0.4086473231294808,0.15383903170747837,0.7203741462311103,0.2965646580013821,
 0.7009510644527763,0.025923687654770156,0.2880211021433694,
 0.2922202784646776,0.053477905170049134,0.7087007809046849,
 0.2612443045987428,0.21835233319997893,0.02993876554870012,
 0.8519676088470731,0.06960844493405371,0.6235537642097656,0.608817860332493,
 0.8114387142967396,0.7991687727254019,0.7055741236174425,0.3073487196438296,
 0.5078646531541618,0.9569928321278333,0.8241717192792489,0.5674892107825269,
 0.7611792193405165,0.7939146802250161,0.4445246102932887,0.15456279913681326,
 0.6770497316115194,0.9008507536875037,0.9621473378056891,0.5209618298907419,
 0.3543081659782489,0.7218790549039487,0.9845920971023886,0.47169251741774576,
 0.21760188156958504,0.6033485885593988,0.045717460342731164,
 0.014931898544876088,0.9679546737509492,0.18716061749317847,
 0.6164179913495555,0.8909319150158803,0.9172731072818767,0.3188490543397927,
 0.6707597987600169,0.3034762013114065,0.5619113559337339,0.36784705346414237,
 0.5384383233838491,0.9666516925968561,0.4226117321143783,0.8853287716854232,
 0.7202480297101146,0.5296664811114636,0.5731219950824219,0.8933921456971269,
 0.1526416919538538,0.8667172751019152,0.3093080097020896,0.2562792559791779,
 0.8038006549818427,0.09970371662830557,0.7816982823585872,0.6935544264894893,
 0.5435360289295202,0.07495009904662087,0.89379973834633,0.08642143029228666,
 0.3445563383829856,0.5557377843378403,0.26158183677490277,
 0.10978510416241016,0.8556714215209182,0.4942913193743099,0.2640454442119332,
 0.868550756013231,0.7184332174809309,0.20164439133398504,0.9732416317843026,
 0.7936567865024415,0.15079863615936595,0.15453880913831097,
 0.14849298574583092,0.7921220694987408,0.7684132131066962,
 0.35131310528557724,0.43286072606991843,0.21080943571235478,
 0.2440956941482696,0.18065127202536135,0.45042374549128783,
 0.19458674106015061,0.02952614448443025,0.5182861315321365,
 0.27566046746768347,0.3790041980362947,0.22096251832259528,
 0.013173999440505124,0.4997819832604685,0.6334401831151673,
 0.19271150902256118,0.6455593032049997,0.30863717901179044,
 0.19671699498681983,0.9890777899203594,0.4241968463694886,0.1618090245605246,
 0.855791916815822,0.16775748848228744,0.33953238417357556,0.4485457656033074,
 0.42462520238408563,0.2147122361250131,0.2513093543622189,0.8295981237128448,
 0.9682979673011057,0.8105954185119515,0.19126522548961444,0.7446934700930845,
 0.37321407654650884,0.034910698447346955,0.2696652852901903,
 0.2691787490643206,0.707959316201263,0.6418508110404888,0.6135978539134336,
 0.4788130868681567,0.7232964288293242,0.9240265015564804,0.825210436655786,
 0.7704139593195795,0.779145499521391,0.582491370520037,0.402035084576321,
 0.8921795501981427,0.5284598275176253,0.6017354167198022,0.22525511028824685,
 0.08704517162813219,0.7106852871872245,0.1356434183098465,0.9115206728349058,
 0.6452114005098508,0.9816264680752674,0.1275656696556443,0.43089864378683984,
 0.20996583383379142,0.25032189961098017,0.24270735899981688,
 0.9386438864857729,0.03134664664334175,0.10110678876309476,
 0.13558058328492117,0.8466329799318927,0.8348659471509056,0.9029783866114781,
 0.06288778331012479,0.2694448085364125,0.1060749505097297,0.8269437720230189,
 0.8027030851150285,0.46326799883825087,0.6317831610999759,0.3325954439884391,
 0.9896582697657451,0.5446299897993123,0.7795412480248626,0.6206574091207232,
 0.0894443330169169,0.4933557572254682,0.27726830949533676,
 0.23435223474464673,0.1817418649997773,0.5582151725112271,0.2553253463762609,
 0.4234414569821634,0.1310632839641256,0.8899228195260472,0.1954025215172166,
 0.029245988327537775,0.5091897282348422,0.5296461529371234,
 0.1538756204490651,0.02906070284721307,0.6009352650835793,0.5892217103468426,
 0.8353310282828791,0.30664243621281606,0.7738365093955814,0.9228890192934451,
 0.6981394722706452,0.29042311302847446,0.31096539361306086,
 0.5003300756412087,0.031743024566838085,0.5830579926337969,
 0.5162118196009736,0.07133527831500186,0.028529110548008774,
 0.2177079564085539,0.4154261432339905,0.09049743139098365,0.7454891328854868,
 0.6846152411079265,0.6171281281543897,0.8121564721419454,0.6468471009033732,
 0.28256877613563514,0.8823564370158755,0.3529101364964198,0.9402107877351367,
 0.7146715808484575,0.6596799316115025,0.28410819319401903,
 0.15901689614336822,0.6389734690589125,0.05905902193081991,
 0.7635683683690617,0.5406002028230653,0.15845764849543176,0.520638222755228,
 0.2723564949169208,0.7329578004812045,0.8922645697854347,0.41337901388748477,
 0.5360181656784218,0.5521515260118912,0.48376331290033325,0.8653333676123737,
 0.6302047721853896,0.9072513425079423,0.2980210721897114,0.4520832906250736,
 0.6297100969851765,0.9573625994064031,0.2139759250520652,0.4741793266628044,
 0.2192466084060255,0.6388813068077464,0.1732016350049289,0.5139669137327,
 0.0556429363426012,0.8851560542922092,0.3854544497328487,0.39932250255713764,
 0.5175030020140182,0.45656545365710144,0.34687959207697394,
 0.0012550193423453981,0.7338685373224709,0.45432247887549027,
 0.010632313636968815,0.4739093988146371,0.9505805805263073,
 0.9236713615218735,0.21049123755387744,0.6879155769218632,0.5093980934980162,
 0.9167841891818328,0.23903249998930132,0.9469881198040389,
 0.45242505676481104,0.5384259331121064,0.794291212058392,0.5312183966594917,
 0.42666491162595555,0.8382407108346529,0.37289050957087333,
 0.36807503287442267,0.5154380449134838,0.5162435224577964,
 0.45069997892956326,0.0428845769280467,0.09490350533340997,0.930090237747827,
 0.7616856742708723,0.04466998445893955,0.48265428567432,0.7725773450722184,
 0.6391764449051306,0.7084248204372029,0.7968494617325528,0.48877534666267886,
 0.5636832193322678,0.7284587850552908,0.5726965473327397,0.06370604532231994,
 0.6579832954353515,0.827128833052124,0.6006629635011631,0.7854425213417535,
 0.7876199704338108,0.9616533596816483,0.1299326019368311,0.565847953710766,
 0.5566430182713173,0.9020024297153413,0.7821750647291978,0.6415853238308717,
 0.2083970489134075,0.22862679734598057,0.9140801722924686,
 0.15525841660028905,0.5470608707895961,0.5814875659250242,
 0.23104388273577237,0.5069656010060976,0.19704100953341674,
 0.19051744780517255,0.692664464010886,0.808344781724501,0.14211304376346512,
 0.585146934851661,0.6520676990548557,0.44675504936177257,0.7244357318119676,
 0.07729502788817022,0.6618485959082812,0.12216741102168582,
 0.15060055756943191,0.005362194379089691,0.2403844584851601,
 0.7358566507321926,0.5674726525456241,0.1772388961705651,0.7678556374411536,
 0.7069538477656798,0.7413678915156546,0.8048251101768072,0.6150656287002637,
 0.3864959642612199,0.9286936326123097,0.1300425889140132,0.03534788461034244,
 0.2513180317252627,0.6819947168326492,0.14112724101307172,
 0.25391156765206924,0.6346357841196943,0.35747665172332765,
 0.8548788152409273,0.8093635025479998,0.22885378377170906,0.4055496351178989,
 0.6433156378099457,0.5626957003677212,0.08211262365182959,0.6572434161233861,
 0.9581728007549585,0.4838046313750328,0.6424944600162483,0.7156448827138766,
 0.4878506613217094,0.18540590437395166,0.36454093571934865,
 0.5388657141656394,0.9790376348640919,0.5670018618058603,0.7134192409453091,
 0.2357113855805395,0.04624463851106908,0.7888991284094478,
 0.11670265400939872,0.2870961783609859,0.9381233668343988,0.7206217890380461,
 0.8431523811144404,0.4147011611459197,0.7837460755551316,0.777918755180774,
 0.5257433031017822,0.04168672614551461,0.9147808854176379,0.2580968221308001,
 0.06480872667442483,0.5383255579713155,0.5891788385036324,
 0.45774486033882167,0.7100032405250346,0.06649272360467129,
 0.041917254707381724,0.03876296709165905,0.9799259302424919,
 0.2448659199885771,0.5585801839421022,0.9078172233674096,0.4069878204371282,
 0.41056480294082953,0.5656565285499798,0.4164146191845888,0.8387356652114709,
 0.36842900707521253,0.8726442275493909,0.30172481367755677,
 0.8775739694889617,0.41669216897295847,0.040953172233033186,
 0.5690457067502848,0.15392575397440633,0.2716382505057823,0.8318896343716335,
 0.9448501701936571,0.06923955161257134,0.8114436085565384,0.5901425201097721,
 0.9689906247171473,0.6411821718995775,0.4320612451494916,0.6228835059245246,
 0.26836934792928746,0.6977097820298283,0.1855054882267999,
 0.43064240836152257,0.08277351394429866,0.6753401737794159,
 0.8349880282875477,0.5782032985186243,0.35128387781921333,
 0.04551433376816372,0.9315286870215355,0.4127101201482106,0.568869651578504,
 0.7940809856939869,0.7893590745200765,0.9516488876127057,0.4625767010430287,
 0.593399643946262,0.8140227948838237,0.07093949879188521,0.46160809038652584,
 0.04402154080088183,0.5428679379712564,0.2515511714902844,0.8724091130236042,
 0.3589147921210557,0.8668555126191635,0.538823692996688,0.8652011821603706,
 0.6637232695270103,0.5973483456111532,0.4027209417805968,0.28092872974119865,
 0.6083865409968054,0.9957042425564211,0.8893094838677057,0.19442284021524958,
 0.5992894021981414,0.7874391983457908,0.27655184636122754,0.5077524471980489,
 0.15811930053962098,0.5560183047537994,0.9102800502327295,0.0861439318365651,
 0.5701031794342631,0.5024336781012588,0.0668645935180282,0.5401634207035173,
 0.889893872002594,0.4934486915603379,0.6346626901261958,0.8006458460917186,
 0.8756238132706962,0.4518362017532186,0.578963454603225,0.2111956130239706,
 0.9176212260047673,0.7892826668585077,0.047662724849972804,
 0.22790046986421764,0.08127189113750877,0.938064710560819,0.5538640754267591,
 0.9927773425961228,0.5014346920826354,0.9508417739452619,0.9816155030539506,
 0.8395983102128467,0.701244678950475,0.4817195685554997,0.8605719441260669,
 0.8101561931800813,0.5092165058949589,0.32374310393747585,0.6292975354517305,
 0.7355080248515998,0.5607286282014031,0.1762096128568359,0.9254964182215155,
 0.3318911665643114,0.4527211565995165,0.9003269108848478,0.9960108089801608,
 0.9674971642682895,0.5652502264680946,0.29351437284327475,0.6227047110904368,
 0.11952911007950517,0.965244576349883,0.9527490573154126,0.45686150606581555,
 0.8921935862534407,0.5260977222268641,0.5698539633621844,0.7409836792588437,
 0.5962305036128275,0.9982115440559776,0.46444876481297137,0.5368673833038111,
 0.9925610404066922,0.16259121133369348,0.9821850280329842,0.9965582765679264,
 0.4666648310845569,0.5522940860773439,0.9038267402486152,0.6877225318371015,
 0.41816636640836613,0.7399566928729644,0.28159987234344863,
 0.018891916282888332,0.2504494880605078,0.6981068876315935,0.523930199936696,
 0.8730907679025782,0.23940962166401636,0.26603341868440167,
 0.47694778303534635,0.10930923824285066,0.8251335404438481,
 0.8058285407236073,0.02586679828908589,0.3072979538253938,
 0.045300480163247614,0.6775448092582793,0.0469924301306921,
 0.7883165296621218,0.1938724898378683,0.2528891972353955,0.5250322379180338,
 0.9938233777359531,0.6195208173041498,0.7757675538666141,0.4028828479904015,
 0.8762876913554998,0.4401189508926495,0.4618160168029104,0.30249151677202724,
 0.6299853184835389,0.37704586028074394,0.9382561214777152,0.1150974045265003,
 0.6989385916050381,0.28233880297203284,0.1267909764886861,
 0.026902959175876595,0.21312257635163734,0.9331940998203234,
 0.8161466632808984,0.47406690397550366,0.4542608722889401,0.8479193339899357,
 0.4800098615717533,0.06835738199453023,0.9355965941859374,0.6608907826671081,
 0.9050553331781742,0.8566865951589935,0.31347856956578535,0.7132539217271988,
 0.43663592131635176,0.7660371082387122,0.2894115518700717,
 0.13270657741592629,0.3824784283934014,0.08082947531113782,
 0.46733218512437846,0.09844285856487678,0.2704728024561124,
 0.9010285526290562,0.6382386206318784,0.23199853561546802,0.7490669767427188,
 0.457147544127704,0.10986341942436151,0.8348038795740291,0.6291534570493941,
 0.5674782867443838,0.6748249677311848,0.513823841949248,0.28042403701452856,
 0.7786267478969336,0.5636987667022154,0.4816676276647197,0.5796563479259875,
 0.5270307175641831,0.8041322547586691,0.9242732769324957,0.02004351251910408,
 0.17757213062276367,0.49885098703770414,0.7678056355501877,
 0.13052627068103662,0.8186614991767178,0.5452011487250666,0.8267994840388747,
 0.12175172685635527,0.3335486639893118,0.4481864261040319,0.500294923898344,
 0.5446787616996486,0.41419134812794633,0.5494901003740467,0.3175404676052924,
 0.1212253528214966,0.2566888902077191,0.14601457351517577,
 0.30427245601238695,0.3166697936794308,0.32681369610543254,
 0.9378450437728769,0.3997824677728259,0.08643470570336886,
 0.18340300234249396,0.7015627360789909,0.9332852235489058,0.5038812803800209,
 0.7652423786840131,0.21585061768525904,0.03842099271118682,
 0.6820061068606862,0.5180426257553465,0.7192825636629485,0.397015447581502,
 0.47622692860431604,0.027829143276269952,0.14991226514673395,
 0.05615948107822755,0.5229534381644444,0.6406090045433073,0.9106392893151678,
 0.5526752962613085,0.1039070901023903,0.4540858362201957,0.08602641453398274,
 0.3425434353324447,0.9382359034746162,0.7795929850083855,0.933258931595434,
 0.5159751841045794,0.42152301370595735,0.9951968899630448,0.8414308268549486,
 0.4952342233945044,0.15488475454220618,0.6307475535559609,
 0.12345136644910226,0.19644772480026773,0.17384653647130244,
 0.0973222643152788,0.8386283439876441,0.5692430645840496,0.636530776264438,
 0.23751682146335773,0.34857290287393106,0.6990111030462522,
 0.8141176821923262,0.4614752578846062,0.630083165544311,0.6534366474491926,
 0.33912674582572777,0.5686749958825527,0.7012240763807079,0.9416402423929346,
 0.6863351787953466,0.2422766861359673,0.7468298766217608,0.5773074468512153,
 0.014567286627826248,0.17602616389473558,0.1505975621708857,
 0.7180840314693983,0.9291148030709161,0.7437412802189113,0.43087906822907096,
 0.161952287352628,0.7833626748100161,0.4693553573301821,0.6944350135193195,
 0.43612922470544624,0.1621466823204223,0.6353864892806023,0.7130620286615489,
 0.5032381274070893,0.16502614150904582,0.8869173822903819,0.0965763930491268,
 0.12140224229947583,0.8294909327214802,0.3963661471281493,0.0702645348964237,
 0.6345529161942995,0.16095862889364387,0.6365461172375382,
 0.17681158405813602,0.06457799578800583,0.29070194576830544,
 0.32735072835739554,0.8282428050906505,0.5638615141520066,0.6898607439329075,
 0.029580033418308282,0.5339330926757874,0.36646018872616415,
 0.8212854858151505,0.6982197370391128,0.7814442227468007,0.22869597351630688,
 0.7813471873431151,0.7039720696733126,0.174291024309582,0.6964322040654253,
 0.7540620097768591,0.37133302823600123,0.3718027688352574,
 0.06743871033461946,0.07121651140181928,0.4049404739454716,
 0.4114374484711638,0.09610730654573574,0.2211385824901343,0.8537870858060022,
 0.9913169328773123,0.9982648012444644,0.7221765351111885,0.17876670062511146,
 0.22388003240895737,0.32047202276063,0.9700229894015944,0.1009309702015776,
 0.3197359648052962,0.2990992069139886,0.9391823014598734,0.172604615699539,
 0.06888029590716649,0.6627709219720127,0.3110728939538906,0.3875542337471485,
 0.6268173187662557,0.5140026143805926,0.8321164037299219,0.6112252351183175,
 0.941954421307401,0.739749009186642,0.3175573506989092,0.4520992537097728,
 0.6972243978286663,0.19164140560028076,0.35887410808705966,0.500527932642219,
 0.46918524991551747,0.06235565992829217,0.31865670420180914,
 0.26828842665032226,0.24157837998902532,0.3821241050510994,
 0.7114575041154434,0.3425012683031853,0.1483378741767739,0.6597069503484102,
 0.26834799991609926,0.15149311664251297,0.06034914029639016,
 0.9800920034349268,0.40708634779641284,0.8375059108178344,0.5117499782479018,
 0.8979309448681067,0.10456701381887068,0.6848899929229763,
 0.09086652822694052,0.907383095985841,0.024969792949922187,
 0.3358969900357953,0.6496295819639009,0.1808880820243215,0.796414509642154,
 0.41901991832162333,0.678256105728252,0.14910649729314251,
 0.18671544581625588,0.6218434146215727,0.41065407658356223,
 0.3685530543143313,0.138117892332575,0.8416911059556023,0.9644806969371073,
 0.6718443721612826,0.2505068766777261,0.8501158032627898,0.8932879172773087,
 0.07529306858060303,0.11654747197229409,0.6195703240873124,
 0.2550591449134675,0.23412220526782745,0.006685371106302185,
 0.2666402713919841,0.7606147556594491,0.6231709061640596,0.9993271165116376,
 0.7530718377899634,0.3563175739129164,0.05693343348664426,0.5620943538434969,
 0.6122839403369521,0.13363759431647404,0.5087382797320765,0.3099571993598035,
 0.8709662487023868,0.4141111089015743,0.6691893956220598,0.03187650252740992,
 0.6661440420063014,0.5536074130264383,0.8759769127268231,0.6829657292877552,
 0.9578216214905623,0.5369702689648785,0.9570289717778615,0.06767790178509858,
 0.7150686189686646,0.09734969625185785,0.5401953818342062,0.5381874101496565,
 0.6461771070482235,0.2485374143603024,0.28867820383552156,0.7090391518491326,
 0.49610010996124254,0.5492990632700668,0.7970467046162207,0.5406132334499185,
 0.9140569005312288,0.047496530091708844,0.6261073638593616,
 0.1387645157989229,0.101863229895371,0.8951292279035481,0.8370538562066314,
 0.2912336105966157,0.6921553134817697,0.7228384187449172,0.881667776866274,
 0.2618740191407244,0.9073002421595285,0.516350971665922,0.7116684727449318,
 0.4882694696377271,0.021548592671141398,0.2362634570876223,
 0.5052357214284758,0.7693010453642398,0.5644848890688952,0.28230686357311086,
 0.3953814489106253,0.9898019125502339,0.9635014866377991,0.2244243422811013,
 0.09936911907988866,0.8504044729888753,0.5321334068978573,
 0.39013050257709847,0.692557066772333,0.11121870381944987,
 0.02721240871607622,0.498351696561841,0.4934466718954339,0.4607133560129303,
 0.9537840422583805,0.29631789621827065,0.5065056975841864,
 0.004063508793143988,0.8118209215695747,0.27095827901303404,
 0.11640621722691369,0.7113662161335156,0.2685194750065898,0.9313855582122587,
 0.24617685767750885,0.5976538270582581,0.6465865786169982,0.8043959062482825,
 0.8378263505683528,0.6584660473682494,0.8298362601598758,0.7608717603937845,
 0.5738292603677695,0.57855154985242,0.8314805273084059,0.8735391649641399,
 0.08530718341886745,0.3144715403160414,0.17738158592177444,
 0.6500300660046472,0.9613380487147246,0.507747191190128,0.7728471558693903,
 0.288196311374752,0.7381404096001893,0.6069293460666052,0.350780412248736,
 0.9996359647537902,0.5587609059398791,0.45474756197830135,0.2017888884272614,
 0.5247998901888138,0.6176497628155234,0.3472259308376898,0.9493774090878642,
 0.09466475518815742,0.10051291668294249,0.9089858540508665,
 0.4807183597760416,0.31484502262032343,0.7472877286499054,0.5045731266769473,
 0.9713090991657097,0.7535792655505672,0.5378168969921828,0.29667140977175965,
 0.17461536110330766,0.737925656849433,0.8393070358214374,0.0917016184794156,
 0.6454507205050924,0.6671678433049264,0.11597722158266588,
 0.19727580979116066,0.6615447331086195,0.1333324641320056,0.6504101322885724,
 0.27847163655878016,0.3734852160305637,0.7598889359390898,0.7477221631683796,
 0.3350377100072145,0.524956048663185,0.9116181209950389,0.26476365245126765,
 0.3530507418969622,0.9804617300255256,0.4439745798917686,0.09471700548461293,
 0.21431510264383946,0.2199853574516082,0.8297700786105578,0.621345471673308,
 0.170041641084138,0.3922748106233742,0.8734018659562096,0.9162278787593642,
 0.2274030118371264,0.18907343092923834,0.20107480705721303,
 0.9511206567646806,0.8216580875808652,0.39494497610061663,
 0.16694196849354725,0.7099480473834722,0.053878362309600636,
 0.3996066342269604,0.8349384890309846,0.6083069269558536,0.08255421630183846,
 0.4586095162188819,0.225381368237772,0.21365014520977743,0.17783308406603504,
 0.14559761924398462,0.5744260825667996,0.6848636272667525,0.9671104638938162,
 0.9001348435026101,0.7493844331016958,0.48611121762069365,0.4059464339724578,
 0.5135278028596332,0.05022128887440569,0.0877639394788875,0.6806602198403376,
 0.0024139506638410912,0.8854991843439816,0.6363716934567365,
 0.7494493025416302,0.9180285320455238,0.919294904849116,0.7609585870944136,
 0.9033188690922084,0.9785966899460908,0.91421398378793,0.31170872957228446,
 0.25559395307802724,0.6355606236520681,0.7783300824691495,0.856385815757422,
 0.5678642884339682,0.32527538935065425,0.5812360958710732,0.7138416701897925,
 0.24363362618260498,0.22633692051991083,0.6234512270840652,
 0.4008208521527965,0.004920941724533168,0.11780928531425583,
 0.3952852688933466,0.29981200093100446,0.37347445254694056,
 0.6003423084435491,0.20186049248238114,0.4291087792073004,0.378735197412539,
 0.5958936522239754,0.30822069949774755,0.9570251623482974,0.5028238421597302,
 0.29169276538215816,0.10155439145782785,0.9973865291991574,
 0.2571626239427348,0.1599001314186692,0.7648700730253919,0.8640178921849307,
 0.11069605782747649,0.12936949051788893,0.27543653491034004,
 0.08168088740127244,0.7396345739065067,0.9651361496587485,0.5898494436205471,
 0.8810514020861442,0.17974421494767556,0.020344283625894244,
 0.13878436331696076,0.25899186201156055,0.8166148365903851,
 0.2348474076924033,0.783690740144062,0.6389766622657413,0.20461239137386378,
 0.3799313501070023,0.8051178225844059,0.3463583885653502,0.8717730581147434,
 0.2303055536795564,0.18904769431537338,0.219665192808301,0.7188792150660818,
 0.2927958014241687,0.6855558937139141,0.7985923341216059,0.7755743866882261,
 0.008511896239499439,0.4592766089316458,0.715415132021834,0.3074960837356123,
 0.303115120439604,0.934656237491809,0.1235274484201464,0.5828728794192757,
 0.48815005303024595,0.7973802695303068,0.004354455488524311,
 0.8679228350401053,0.2671091872699365,0.42172627517168293,
 0.19753977633876518,0.7617885019449888,0.31999572867435344,
 0.7149629017346064,0.2577953329351689,0.5833014344630953,0.6755283788048804,
 0.5789381808482775,0.4565668254781998,0.16732154709726976,0.2205083286105265,
 0.4228357070284344,0.701970289506208,0.4044728303275974,0.8439358271153341,
 0.0418972417574881,0.0066053620323348605,0.5628900018933704,
 0.8063416014930268,0.11470678596844763,0.9428743249191966,0.4067432773410645,
 0.19141489435204284,0.14172744281132477,0.14020953446273943,
 0.6076868479049219,0.43777545513829175,0.6960679667737234,0.3574702425876385,
 0.5295931300025855,0.7089503856008066,0.03384844712717339,0.2425771349803041,
 0.407962466087629,0.10102940774843727,0.05571534774020148,0.9791354813842575,
 0.9061739010008559,0.2555717450322287,0.7161484008019814,0.3206031172495838,
 0.42201503628215753,0.3393069941369088,0.3202875063524723,
 0.12809082152676532,0.7242866321192263,0.3011464206676735,0.6444294132246368,
 0.07776707463769361,0.23446825153418094,0.7058995784196053,
 0.21898002377688675,0.8541801477238492,0.36531263630354793,
 0.9373723421604987,0.5155517297199507,0.641100440858101,0.43056602484504236,
 0.35467142097693216,0.3003575907900031,0.22756513647739196,
 0.8130553242969153,0.5679397122066552,0.24880913798738147,0.2983806075623293,
 0.976928884538256,0.6498385363890196,0.21136615946717552,0.8379909490348247,
 0.8702479882948302,0.0703187740216703,0.33601994877439456,0.8486339065413416,
 0.10182427751239997,0.05657763092427037,0.5849731294732051,
 0.5301571018667437,0.13365746637596299,0.771439082602591,0.2030832446700228,
 0.4869416491688112,0.9991620447362504,0.6241979759257751,0.9583518953806489,
 0.24573734291076355,0.4289270183678662,0.5647953393921361,
 0.11012827108870749,0.8809099211989876,0.8752471354443708,
 0.003404179531617535,0.9292374343237909,0.814116275887401,
 0.41542911193134757,0.9416288072356724,0.5294261195490925,
 0.23543972349736098,0.917661109674307,0.55172698904497,0.5168603434501066,
 0.7933030025694279,0.5857288578822158,0.27370497104842695,
 0.05008536765119054,0.8029397608630043,0.5352567556179564,0.2777353706567627,
 0.552693482701425,0.011332191121206758,0.35630276849347553,
 0.8651739985204782,0.2836173876444529,0.38905684394099793,0.3826002984878756,
 0.6605442656523022,0.6758968379263106,0.1846791498190763,0.4030130672530705,
 0.41238618709706343,0.9084910459353532,0.7877637763292847,0.6065332793851138,
 0.19534675492031717,0.8298316475279787,0.6409458669212095,
 0.031440405758028644,0.5437190735967794,0.024386774374851927,
 0.37498265289062294,0.03875509850274661,0.9806636110256401,
 0.46452315099107355,0.2530881081419609,0.12454661288924485,
 0.7407379250384407,0.010185335838695297,0.1952031915748249,
 0.6426031306764118,0.2912718537953489,0.8850313859360663,0.9386497404455083,
 0.3069399699980081,0.7752839076259619,0.7174516691261278,0.7214800372315731,
 0.3945536622652489,0.7469345691630904,0.09950007964077978,0.8279491136127403,
 0.8575770974011772,0.49795533870538244,0.8704756159371915,0.4705956584689003,
 0.27889911136499423,0.16736291991189922,0.5734441703525162,
 0.8124927825386892,0.6455479633485747,0.5015752339501255,0.19396387262448767,
 0.5406322213068061,0.20590927555738037,0.7761439062201033,0.7939161133407336,
 0.9625105736018741,0.5807059538133373,0.22781770410925073,0.7118464994045794,
 0.15749505352410786,0.4569440997122083,0.4860147138588716,0.9412918702860363,
 0.6781667576742931,0.8512134693598732,0.16888003364681492,0.975516641871036,
 0.11448159271463654,0.15020294777726073,0.11919483962293409,
 0.8642527269637884,0.21977258968621838,0.5323942122623906,0.8524898952369289,
 0.1340159458788982,0.008415780472698087,0.2729642948995086,
 0.5083322993353202,0.5399375391402716,0.48817206581508954,0.672308204576256,
 0.03421046214018986,0.6411944220821133,0.5907382501492622,0.3363897343320502,
 0.8809996767320196,0.21184236127127054,0.47288549425535464,
 0.4194218688268031,0.6884532812351599,0.6664852346409045,0.24728313216015274,
 0.8363172787546689,0.4672058123072199,0.4640521602501899,0.7055321350931889,
 0.0012142870157393348,0.8676168853063632,0.00593863183567056,
 0.5506756594704546,0.6333901556351511,0.28554425666319927,
 0.03368442589921228,0.976678190788247,0.47594236599507944,0.9349564458835107,
 0.4969242270437969,0.5116227610710036,0.03690894065953709,0.7732776275803043,
 0.11232157671190146,0.7286716139993973,0.6052090113350197,0.7395043758223592,
 0.19289726444081834,0.7065159915631009,0.8191592209013798,0.9713044768953367,
 0.6453165155100845,0.20444504022928867,0.03230882193163642,
 0.13498465208027388,0.9242280574596492,0.07193734450726874,
 0.34848040493030896,0.2293548401624813,0.8778260857406776,
 0.11841062806308933,0.2850150021467929,0.39710569235727666,
 0.7589427167135128,0.8670709661271153,0.12754123603142575,0.8170069770970059,
 0.27067657051024374,0.06394479461775171,0.4005274605256858,
 0.5598226202446636,0.9437002343097219,0.3098023539704703,0.7146609390486165,
 0.7264659758476898,0.18965910851714207,0.39835051148893985,
 0.4993637040560156,0.011001161456781983,0.6088839755300769,
 0.48495948496718344,0.7132890687983937,0.9834030896267889,0.0403371402457553,
 0.7786652946940387,0.10791571478792761,0.9862420278498123,0.6581312677855301,
 0.9515900037904034,0.3483747037696987,0.6939980603464803,
 0.022929670902579025,0.92376113318203,0.7949155587127332,0.7224728083022384,
 0.4748328547644207,0.9029559592376877,0.1549140032253915,0.8349828238618014,
 0.6674963785353432,0.36465987628555885,0.07517797982702257,
 0.30841004924020776,0.30683058167734556,0.5120832079531444,
 0.9887492531868599,0.24869317156730875,0.4755003100773365,0.9796701789806291,
 0.23223727301559283,0.9864732396117343,0.1664295561383613,
 0.18058558008071834,0.47673983867826275,0.8177306175491452,
 0.8588621731835958,0.5182933154597255,0.6825447283641503,0.202199118194073,
 0.5069198248750897,0.09475655702144559,0.48464932071495714,
 0.20374003517092865,0.9682220114520035,0.20702713816920737,
 0.5260385220619426,5.759952908179766e-5,0.6518062193363711,
 0.7772380798764289,0.47818440418863495,0.7254381344914643,
 0.38377769555198127,0.8136363812090788,0.6631040305154038,
 0.24492049347399636,0.1487410051378053,0.2479879327514143,0.3215355981880017,
 0.4103866407526322,0.9518056686107251,0.8764097952862966,0.11260852916811903,
 0.006005893817385699,0.1694851950219547,0.010441281110771361,
 0.8328714985295991,0.7930720294665298,0.0058704392171315956,
 0.6785986465802285,0.6506521779818322,0.9732351615965437,0.30305057031814575,
 0.5729744261914642,0.803474383695579,0.9687044922851651,0.5411686801302237,
 0.82495815190978,0.46321800809219404,0.979128350843294,0.31284813332645267,
 0.7504116091009966,0.9818186093450266,0.6152380684702876,0.4555304142117732,
 0.5950204021216358,0.930410382033897,0.5588416486573466,0.7271100679996734,
 0.38006016977661217,0.8856286167090929,0.34287450846919787,
 0.5655223910482765,0.4364718502434025,3.90874548633402e-4,0.2391820345277551,
 0.19013563950286783,0.13459901824442366,0.6555826030747556,
 0.16904603594712342,0.8054422127791391,0.49959792658586943,
 0.4273057211943063,0.08724489695501203,0.7014658157844906,
 0.05518664383864702,0.6929850870522749,0.7929415369448259,0.1146426492923317,
 0.4777621932017895,0.713682889606367,0.12617323100983024,0.23072520752759962,
 0.17444980395804732,0.054446736979004706,0.9778991179713016,
 0.7507441486145411,0.3139400840798625,0.03292449371229744,0.6288415107537688,
 0.6327285028194624,0.28246688758652805,0.7476712860478592,0.8815000891952303,
 0.5666190057771027,0.34852274058938515,0.3200050183438814,
 0.22046147318058307,0.5134765958819201,0.7362258550550836,0.5785932660133686,
 0.8971956127302496,0.7218926956012237,0.14561720123594668,0.5407740920691461,
 0.4328517074379925,0.20248229228304881,0.8996684259422507,0.3639790977446,
 0.746663752512879,0.9587737133460283,0.4915458483429096,0.5854302166234456,
 0.7087006346814092,0.5058967544109232,0.7158644062274145,0.6790443290699868,
 0.5892236903512567,0.3236185582566502,0.8962548719920667,0.7440242108393447,
 0.21866697218120112,0.6609186357517383,0.11528253978141478,0.707475766886508,
 0.08924531462307306,0.16106706077871347,0.6113759202808612,
 0.4019041263435298,0.5207999649863552,0.4980780281749221,0.9604707851260312,
 0.5016610167279405,0.4496586820658629,0.5432688052367143,0.5141019705110739,
 0.9514923007107905,0.5296625001517008,0.661254123624326,0.25790391371853094,
 0.7922242751525843,0.02568172805428992,0.0509308619539347,
 0.22258540646584934,0.5503010580984926,0.6000313549357913,
 0.03897891007950971,0.6105888566658827,0.3582153374426248,0.9658969823562422,
 0.7629201158323734,0.2578906552721476,0.7700108918726299,0.5644431258607769,
 0.35146636144623244,0.8681720902265333,0.32677826278957545,
 0.023803180357416576,0.7594114947848012,0.1981360032396886,
 0.42563424573726616,0.10183473160094647,0.7773911957795475,
 0.020426010713383902,0.8561613138321355,0.840459097003543,0.9506244310849347,
 0.6177767143463225,0.39054478943771753,0.97521980192398,0.7202141560253715,
 0.9001203639338495,0.30721522135287516,0.7402472981710307,0.568922955382785,
 0.062038657670377084,0.31070888377266304,0.21591650150763886,
 0.99894249265674,0.22761597406907907,0.5162171620778775,0.5563959187540259,
 0.14585419196134186,0.9390959998039756,0.12233121767614008,
 0.8639240566000342,0.31608427305477993,0.29452232919241206,
 0.8512297561451997,0.9640537442874516,0.9424257098034063,0.6903075798048416,
 0.5690122946058946,0.44081961235113587,0.3853753373731039,0.6783635854289813,
 0.8578938564858647,0.6798835784704669,0.9779153184909419,0.4710924345381141,
 0.9035755460386903,0.5806128914605151,0.5139120649254378,0.5890462458236565,
 0.9736565484063884,0.884157622379146,0.16172771497875782,0.848509708770718,
 0.6041938953633323,0.39446266355856574,0.7892743047857999,0.4334800301341647,
 0.4051199222629961,0.7752223293784315,0.7318565086653357,0.09897262388909889,
 0.32888392904823827,0.2828390578864719,0.3528519692628277,
 0.14815396238096568,0.6267013297024685,0.5574363802097729,0.9741027963258082,
 0.4534357714200099,0.5658155751389888,0.9209043614400796,0.9831918146406934,
 0.42428463607303524,0.7521671012686094,0.046997587446558575,
 0.5078628629220977,0.024196299493762674,0.374787413506974,
 0.09898338657042782,0.34156747107744334,0.681502501110816,
 0.13226504422561947,0.36943782453894847,0.10681149770265008,
 0.24103966244775765,0.6489960854067987,0.556963486161295,0.23837744101874236,
 0.8147704507385211,0.1583163736542752,0.6683300749169088,0.543965219678117,
 0.6184541429616094,0.7392007538909838,0.6591902445402567,0.7817675430639078,
 0.43944479276100035,0.9670481863851317,0.6619218923035924,0.7114864609041853,
 0.4258044346887311,0.8032377340976058,0.6248886679560768,
 0.008514517990799386,0.8878753427506397,0.558333998281036,
 0.23817749578697733,0.9026199542551557,0.2506269585592664,
 0.20596879141210722,0.20977271453735968,0.5735390473000155,
 0.8505087724458795,0.6177934249134487,0.3390064402278853,0.36293158489280963,
 0.08613270347590052,0.5788398337904228,0.18836061412564398,
 0.3244446993679142,0.9865307183284512,0.3977000122613352,0.6095699178981594,
 0.02099366139997605,0.10456099073606073,0.6481128110563982,0.353371737040967,
 0.5315672097914852,0.8049231722805353,0.12162337345334917,0.7046478801675398,
 0.9261328740731752,0.0035106467224576,0.6951955128756534,0.3072938242843939,
 0.153153498156529,0.7174414952661715,0.11818516195115814,0.5838410041804554,
 0.813206917423955,0.9771981328463505,0.888729698740674,0.9101103457912033,
 0.7023254125035042,0.5639997433616233,0.6280467111941769,0.8419614205159612,
 0.2656791761911744,0.7740889622773433,0.7114998734365978,0.576552713423913,
 0.38423381659573697,0.049139747925573074,0.6795895398467654,
 0.5752433297326556,0.3182453700016563,0.21393074575083337,0.8087531600334446,
 0.543258238249118,0.26642758909315223,0.313739928891555,0.33146593299067884,
 0.20953293467427148,0.5372443041408419,0.838813095418977,0.21037391845158115,
 0.7305978593696255,0.4103991265532808,0.6646159216270422,0.7296858337045007,
 0.23185898686161188,0.20835171075809644,0.5903814686784469,
 0.4165748452689798,0.017556877498795864,0.6553542260858506,
 0.16037233339967716,0.1845940217278359,0.3862000198373007,
 0.14670195144243692,0.7946836838633854,0.40093908820200197,
 0.01342356560669844,0.5740433032211143,0.4133008260350808,
 0.49593130780957406,0.5041709608219042,0.5068019834804642,0.5340909317378196,
 0.40079970761881234,0.34346097647733775,0.4126503812011738,
 0.8038307687001287,0.17685447136662602,0.23959106462460655,
 0.21412173424347425,0.4901695949375249,0.7828309707273923,0.5358004187719391,
 0.16912188331304745,0.41611443614428867,0.4152050594930059,
 0.5251735417679253,0.7005668791262977,0.14932120294275353,0.8163776203094483,
 0.7206148437647969,0.06647192584115547,0.9210195426476493,0.9083401739659888,
 0.35157185123458523,0.8843903541568505,0.15631169098676123,
 0.3612783786292383,0.04106167258528037,0.873618521195797,0.5701157645256062,
 0.3446806601522939,0.164549665128764,0.02415937738931495,0.45448879206110626,
 0.9135066094936056,0.9371678040243497,0.2036044198536907,0.41497247131682846,
 0.7621099375579482,0.866724030760188,0.6132501896532466,0.3653607850174604,
 0.9622499532704724,0.47915444458323464,0.42416927801044635,0.349432568196107,
 0.9401215210166571,0.9939577366640119,0.4294595979226521,0.8825216642973924,
 0.456857310907987,0.7381910765219485,0.03595721541731556,0.37328707422327345,
 0.9005993220801563,0.34589005071546053,0.09261638013955042,0.843779896262079,
 0.6916787719916377,0.5174262818513677,0.2780540160764581,0.24252090418522265,
 0.4281804699864521,0.7054918111246431,0.49785296756328024,0.8124911471653606,
 0.15960604392036237,0.6136841829458559,0.0797485676184303,0.9651112055619285,
 0.2149701875736475,0.33240551574893784,0.4372236437062934,0.5859119987896965,
 0.7775261164145345,0.9866440620656811,0.021202744038300803,
 0.7798497804488908,0.28744890474572216,0.287943904068685,0.7778297002364802,
 0.2530502601375787,0.9400306491360433,0.9214128634525167,0.04379243469774752,
 0.6845662200382718,0.4026209762767916,0.35451976215642167,
 0.039436628084869474,0.8495559387919687,0.6469908584975376,
 0.7240827958494381,0.6714634077227234,0.6473453628616027,0.22135484960677,
 0.11512265329702387,0.7266869002956483,0.2881169175073801,0.9804787232598541,
 0.1754297364970614,0.014524509172958888,0.385245877425583,
 0.14482675499327136,0.8942787762770912,0.49450369446121134,
 0.8899298646686703,0.7972501674742023,0.5464669658226077,0.6439761259451259,
 0.42781906212368526,0.4686801834873504,0.5583804253055655,0.8510369053594888,
 0.6998595806214338,0.8309038131395032,0.6336955880962836,0.07523490171498792,
 0.8728087744993418,0.3483084643635028,0.9761455094851327,0.6161754320350815,
 0.3854056624369926,0.6123390759677036,0.586438766292147,0.6362253404856874,
 0.6629787850995724,0.9304784033198759,0.07582433920512488,0.5501140017843342,
 0.10412259749038721,0.9231866933261468,0.94412253037304,0.9919784337301906,
 0.23091619750545211,0.30462927924208905,0.8967611303087986,
 0.1506392963265274,0.7429849011234306,0.2733378478752263,0.552724193573576,
 0.009312813978311407,0.2154658463534742,0.22807879891269756,
 0.6825027511938461,0.758053156356423,0.7617395555346476,0.49580326511035355,
 0.4259618962232484,0.7170719331172091,0.19308982649970796,0.9224605822900619,
 0.3581519188416946,0.7550997140914625,0.13510364249552298,0.2477644970145999,
 0.3089933983972707,0.6809845470370156,0.9832231821890832,0.28524537189208654,
 0.7584893945738461,0.4646848115293466,0.9353175863487029,0.22745912397821855,
 0.24188397745535406,0.9584950271054877,0.5853283887153639,
 0.10405777464737798,0.08825707915182635,0.6053333388774051,
 0.9561012947494112,0.7597991281525597,0.35762294508486303,0.9138131127844729,
 0.19681252789916714,0.5942325305083308,0.9749143551984542,0.864361695136942,
 0.18975749299853284,0.23032738362223726,0.11852866103810211,
 0.6985679296452525,0.7779398920299844,0.11802744474337301,
 0.19832810971355874,0.36270758903497,0.6647629537506902,0.33338107595674005,
 0.5140688880707854,0.5660352831573722,0.3736407231145855,0.20854256700218932,
 0.7907926603571169,0.3370270034730142,0.6021026346016982,0.07787142596063634,
 0.5139607827449088,0.6083501481770185,0.6682920398926155,0.3471239342989507,
 0.7938520467234979,0.206754383443291,0.5761074650249001,0.9182250977675668,
 0.9205318078575303,0.49781652026332357,0.8822009047718405,0.9658768777637693,
 0.5844313723773984,0.05323583472722293,0.8175060798133356,0.3708357502997669,
 0.4093517449987495,0.8720599186063617,0.7635932531487754,0.9667584830681621,
 0.37370263799647985,0.792339124556602,0.7319412108224225,0.10944031084005834,
 0.3226027874184799,0.02173899777761701,0.5169044561658328,0.9123056087417034,
 0.4020583990779639,0.3280218296061701,0.2851204369278215,0.8546401088624298,
 0.35757012266497445,0.854302730573124,0.2675499719379887,0.7594605547744724,
 0.9695660257100408,0.7929291440521138,0.6093460611043184,0.05992738963948596,
 0.49452786376020663,0.8671171143347476,0.8848060914688903,0.5404303907375847,
 0.09149147449681794,0.7976396624316223,0.7323442789207282,0.5752694445438681,
 0.3872437176278841,0.9495611047698239,0.4625448216407557,0.6042699056269232,
 0.14512644660040253,0.28841987242389866,0.7521235044325738,
 0.11218504494815984,0.750227135936795,0.7281651312903862,0.1912052055399801,
 0.016498219469413655,0.8908826398506177,0.8311960232998697,
 0.8409768248662114,0.7663255765852982,0.7264438459630927,0.7553761415157636,
 0.6549393062054043,0.7183704080283545,0.2836319210306305,0.20390429637235163,
 0.5110689772305691,0.22251829735793038,0.782888928627651,0.798636493845912,
 0.5761240507049071,0.5071066094222326,0.10877301819983709,
 0.40823775258076533,0.27540937833027634,0.9286575277643765,
 0.3153105849758987,0.6105593735558484,0.1673423900410128,0.7558585039638177,
 0.7569472876628559,0.32404130531510833,0.5359328141515839,0.879845112263475,
 0.10218694277138285,0.11751941477879946,0.008501036151506769,
 0.4752654987046153,0.18823816278593364,0.6630737716530343,
 0.32919239827944635,0.16873300567418859,0.03995362572973571,
 0.6907955370396844,0.4561644598727015,0.8880339239426445,0.8419868493389258,
 0.6237379047163003,0.7855768547394298,0.4527491135694446,0.03474199286905311,
 0.23037699850147242,0.816568495905952,0.9235491094064674,0.8727030986846807,
 0.4911625317666253,0.6419769998236379,0.22260001765391402,0.3604506700397403,
 0.703469942503797,0.680212779732171,0.8466639820989854,0.7955373866494027,
 0.05219588928865404,0.38574459359157287,0.6369169343602276,0.740386702826501,
 0.7180205970004617,0.9265975932112742,0.3528387906066308,0.9231518905099243,
 0.45152504983765374,0.22261047327979344,0.9994687963735276,
 0.12526340253459578,0.26097010097462214,0.569906337372936,0.6976157489445718,
 0.9735761204176638,0.12138643095669821,0.8871991679818574,
 0.31443414582414486,0.5504247088799767,0.9322188794419854,0.8325506700376158,
 0.6344297612902347,0.823705395765176,0.30298768375891616,0.7616146694469719,
 0.37854965272912977,0.6260664840335066,0.7256746930270113,0.7340137001349866,
 0.7872122989577359,0.7135558259356228,0.8053719057607839,0.49437892031200437,
 0.794253522389419,0.1980552393449666,0.9399199383018125,0.055346909039297176,
 0.2825734193990074,0.09468191380707647,0.6250053708142579,0.6270355741992137,
 0.8150823939841564,0.9388100614708188,0.929082567824414,0.7635014807039175,
 0.40994679421969504,0.2907407534016858,0.557225331305624,0.7059653721174987,
 0.4717915195704768,0.9662754507270214,0.4150059438938144,0.684590258184369,
 0.4237914736807187,0.6940189960120209,0.1474540083159328,0.4540621372832572,
 0.47060965735471916,0.5266785135234965,0.2815005553529354,0.8354773673285529,
 0.3634896495062039,0.594650432526757,0.2996038902891538,0.9820652333419866,
 0.18920280273387313,0.36671369311797664,0.30211367413407964,
 0.8419752578182474,0.9066654048660733,0.6091912047539855,0.9908078104830365,
 0.9203820947026562,0.5196542963147845,0.5056004152925063,0.49631347131752046,
 0.046466410943606706,0.4036887504493518,0.06795059498810674,
 0.9425335923992447,0.7134110908370168,0.470257952935222,0.9595172601040733,
 0.8730551303434518,0.23679573337757753,0.3954035791675601,0.3106816249502091,
 0.08041656582516321,0.6333769934606361,0.9267328589288868,0.5350382940707679,
 0.6469933502386614,0.21076573321924796,0.18784528113363064,
 0.7323330100034834,0.07526862729178008,0.8064502200442647,0.5260186713080006,
 0.48318687153215634,0.4671431779845573,0.9597231495457967,0.8303300563044566,
 0.14554608014102455,0.9989110021904952,0.6748002951567091,
 0.26750440201636794,0.2725210183385022,0.2709995033957373,
 0.23390626822470617,0.49575202331226254,0.9773651349762112,
 0.030822758204497647,0.9828372040171628,0.7173990808021454,
 0.3921171370637544,0.0716546010908764,0.9179427708942647,
 0.038069153018606916,0.11420937935594522,0.4834053862358817,
 0.3749928832535053,0.4306667843249683,0.12150742086098087,0.7417520298214587,
 0.4414996579318625,0.7227334324833897,0.08646565966814257,0.8773479320505125,
 0.736529362545977,0.3928581111002112,0.8651789493884192,0.5851474912743453,
 0.16025231417492491,0.03380495220456492,0.4502365757150213,
 0.5685691356052406,0.9865184016487095,0.3903006446977336,0.43928510015038247,
 0.9950107191837766,0.20208234458631447,0.4706777637100531,
 0.049498618893651924,0.2984125938579587,0.046588205043798236,
 0.8753605765949865,0.3924622163419721,0.46977294513170154,
 0.08261493505260487,0.2891426371125868,0.1364430375352771,0.4183142693743419,
 0.8154118716041141,0.494614865759779,0.5360364135754597,0.5326894375083264,
 0.816749415106209,0.6114348622104507,0.957108762255491,0.131280016153409,
 0.3735012098789183,0.9073341150145803,0.9970807797198176,0.10003483917937117,
 0.8239388262098511,0.06272169203194278,0.3209248138696962,0.4158373622370801,
 0.3471807958237825,0.39412505488153105,0.8281592965103671,
 0.0049958064800782465,0.3000735059708278,0.003397333591985019,
 0.8439044686576677,0.7812556980885479,0.47130210487864466,0.7708050861293363,
 0.015292816440849233,0.11068090941948794,0.291003129727222,
 0.8580977909396497,0.1942242102091033,0.09014870299874755,
 0.025948005741548763,0.459872606312527,0.12894594129926162,
 0.7130284066344347,0.982342441933963,0.21622237012347711,0.7332805349596817,
 0.16466847110435456,0.6934854271076157,0.8994294595124536,0.45978314890908,
 0.34298811241324034,0.8350146001192614,0.5799303367473665,0.4571975049843524,
 0.691504929630595,0.7165833812334456,0.8035759674628572,0.9421262393626326,
 0.804950655237026,0.8425259104519741,0.23977531585633072,0.6006711491214046,
 0.6665448151511513,0.507088975366216,0.7523520018477821,0.6163398141559752,
 0.4710342014637142,0.39612728546672826,0.5664093728022741,0.9083980061245864,
 0.11087636618789276,0.6220330802060338,0.13094866019872897,
 0.1683282253096029,0.22443380640674193,0.7305571044763601,0.4455765807539692,
 0.6035663884689765,0.30882554282887953,0.4835506069388221,
 0.42654790817668153,0.09125715644809129,0.9042198257632423,
 0.5437030540404444,0.6521875193100815,0.7684228264778535,0.8533361955014382,
 0.7741072461332712,0.09502731115465968,0.005609921123542638,
 0.5813920370461725,0.6386338999316686,0.007449300830698058,
 0.5105282358786393,0.3982886641743869,0.1703360381035237,0.1194887663862354,
 0.48551937130400247,0.3704420976350582,0.5553713730738673,
 0.02385299881998959,0.13765450668063683,0.5573260247017346,
 0.10250057817619973,0.6236641180306234,0.06422527167864711,
 0.7106770325683733,0.6419177392605666,0.8804670184018241,0.10488076507406008,
 0.636196347733824,0.516495726212486,0.6291631869353188,0.03555700845805576,
 0.483981005004272,0.5845391346811569,0.5260694006408179,0.8319524176329889,
 0.32487779990322896,0.5546260953922555,0.12993669666062269,
 0.2904996558375522,0.7485596942120044,0.024384645640769365,
 0.7942956513000512,0.4294795033559742,0.3721905287174201,0.37437423570773143,
 0.9000998141560914,0.8783408907885765,0.40054372937287264,0.8635418422109099,
 0.8826464262402177,0.9120605610026526,0.35445563894965115,
 0.37948143746304486,0.9291011443913895,0.5559746712617546,0.8500092638798108,
 0.7264053242217103,0.18114575965932067,0.6212022997238297,0.6742377856297258,
 0.09705135676878063,0.0578838704742618,0.4323930370634711,
 0.21750788345319494,0.38287475953593697,0.5193110995578051,
 0.8534574676001002,0.33551459416829865,0.19214651056077636,
 0.028388422503802024,0.24767643096925163,0.4306623299057917,
 0.7418029089768778,0.12723777918843715,0.07470803259700909,
 0.7314046151663993,0.39397573073113745,0.42760110291853604,
 0.9341778344885813,0.35615247213932544,0.4985872693527694,0.6399493834561939,
 0.9907575814623305,0.6044885653507084,0.6713791871145824,0.35272520036021926,
 0.237395827568325,0.9658567479752308,0.9795618458308321,0.5029543171966169,
 0.5231068449921933,0.49168579146711655,0.5183006178688466,
 0.17449086785633894,0.42774183256755516,0.42830509223901025,
 0.6230338584772486,0.8422536231010471,0.5662256935283713,0.15362403361782362,
 0.3721824903673754,0.37289281493537596,0.5538906024857552,0.8746847437895948,
 0.9094887162545505,0.7578367569696751,0.16305623862665664,0.4247463800856306,
 0.6685581989302835,0.8361914810183424,0.07273901290234719,0.6134174109697711,
 0.103786446384868,0.9315108572267246,0.005938552132415453,0.8911563098083024,
 0.1548490495305762,0.9504222121781256,0.5442040033536865,0.5740968414577735,
 0.805321283269659,0.6753116604202095,0.20879966043605702,0.7709864856453802,
 0.7310867508106976,0.2507281724135776,0.22456354215592267,
 0.28870251258684243,0.10208976249365664,0.11762331215889688,
 0.5391883817427285,0.6636459376863372,0.19443801182006826,
 0.34443935964746863,0.14705254754774955,0.1728927999455836,
 0.8092702609573762,0.1458381536073936,0.5488954580611118,0.4004186056624741,
 0.6551668586068757,0.7516077673927325,0.739645495053453,0.12803420245589403,
 0.2931985142143154,0.27614956370942934,0.8352363792277716,
 0.15284720715563938,0.035565674196953556,0.3974467979609704,
 0.3805428782735315,0.0911780656275809,0.03871609569918277,0.9883173226234926,
 0.9290139087297179,0.9468902068755709,0.34547462308235055,0.5441498663137426,
 0.592706169804455,0.14273217521116832,0.141009059686386,0.7848044541843953,
 0.17279301152648396,0.3533894189616007,0.37716409429557907,
 0.7143545905485877,0.949574039367711,0.24210115727597592,0.9537536074018265,
 0.933421395009764,0.29747848692040224,0.4786877396180492,0.2801251176764926,
 0.7290063164242505,0.30262758745952667,0.7716322363960675,0.8277928863433442,
 0.031808300334453365,0.7868512558474221,0.8161088178568432,
 0.4839424866267733,0.505258719568106,0.017923646947138328,0.9122624943912726,
 0.9264978449482771,0.9924149108494804,0.5992351388716057,0.7240074643272479,
 0.5638477001958255,0.8736513482764018,0.6668161944298561,0.6936668321904066,
 0.8263486596560904,0.29481083714108525,0.2813750117012873,
 0.32335082770239987,0.014392712308098465,0.04757382672856736,
 0.05348683562469947,0.9190204020793674,0.9606288554065162,0.3516832981562954,
 0.49972541628121103,0.42937928496274136,0.6107851251562737,
 0.40480937739423495,0.9686078283577102,0.9419485843211967,
 0.025470053913702273,0.47637710338691275,0.8672442749682889,
 0.9644962102739442,0.307054716977631,0.2254153000837572,0.1494967946584128,
 0.5390078379047039,0.9265022263920726,0.21189188580953855,0.1979854211164166,
 0.03732125801403985,0.8236489591064535,0.5666083540440057,
 0.12212597958776183,0.7348369750840353,0.6737390115779825,
 0.01658613471784398,0.6147735056410435,0.728659220924599,0.06637711776927513,
 0.3052124607935862,0.44319572499363447,0.8992486034316434,0.4116916108833074,
 0.7646589154183201,0.9359379477311058,0.5058033360218763,0.5031495986106969,
 0.4158165592753704,0.1398613741968855,0.8678666670551072,0.6259439560393452,
 0.8051455862874599,0.9036614411511676,0.23287630242516777,0.3643713505021664,
 0.08033067637440061,0.9084063054426921,0.6578705801623194,0.7102023208338406,
 0.6241992019729663,0.4254461353255835,0.08487277631460577,0.2912920251115415,
 0.894393285288214,0.7385180211381915,0.9788434919542608,0.4355778328830595,
 0.9482209671491226,0.3129137055648662,0.28157025977443517,0.3323382323566525,
 0.40251496564156275,0.7416821022116292,0.5487239212652874,0.2805724058616863,
 0.5986917159258408,0.39265653740814677,0.8164999093515998,0.7214259350309625,
 0.9944946309756221,0.5208535199494746,0.7409414811634751,0.4012614208501156,
 0.5435827673502163,0.5176538177523976,0.16404273693102933,0.2023172409554933,
 0.8492781377102491,0.33813769824491413,0.8333811121489894,
 0.12740279797453247,0.25624326884099313,0.7735920899797775,
 0.2812896730854918,0.6719476676915948,0.8710855624225643,0.7037046286322024,
 0.5553659892802121,0.10607514018903486,0.9124026696972658,0.4208959418251341,
 0.11555215766939841,0.5570148341043675,0.031095530739654542,
 0.9101198063179416,0.67839802063923,0.43527654575031915,0.7573778267766733,
 0.4805203138507038,0.9533778886338362,0.609295846585487,0.24135219837708755,
 0.6324949421138564,0.03771876307651412,0.9014189515488484,0.658453662738046,
 0.35398872153655525,0.3594950691290968,0.10563769693024971,
 0.20677305839981985,0.40059292832757687,0.5834601837236476,
 0.47434319691526383,0.636082204441731,0.21639342473171175,
 0.17176047022990404,0.5124257430524949,0.6604324723586047,
 0.42881051752959487,0.5109632461105442,0.001278827892924994,
 0.5418910661071055,0.8920893672567743,0.7929992717264418,0.8302901465059171,
 0.5381202744984199,0.25932467145569216,0.8937398681315125,0.4703123481225283,
 0.30711451642719056,0.7786182774661667,0.3979929752125082,0.8846661315457482,
 0.11706955385518325,0.8217813585300358,0.3127459815276108,0.6686641993756136,
 0.7151224983381285,0.7749145516566026,0.7846636333121249,0.8682286486795979,
 0.6723640044390262,0.6399902754895447,0.14962872122999138,0.628306069193252,
 0.8829303328867504,0.8634101534115529,0.655513419368423,0.6533926945532456,
 0.3919755887536871,0.4867481617388113,0.3102193654001998,0.4703543895225758,
 0.6533057030975188,0.5580562837493759,0.13989423849438742,
 0.15278767879143706,0.7825128846928474,0.3395675948402753,
 0.24133976257035927,0.6165980695536435,0.14620872335558566,
 0.5820499724767003,0.8781846492719798,0.2685847117666722,0.24769193529414324,
 0.523140575788609,0.5562151450906166,0.02126343980451162,0.37147867167542137,
 0.9506003048298668,0.962796852505668,0.9740487842529211,0.2905237508901698,
 0.18991537022407723,0.18369385950611783,0.18138583415129939,
 0.2899224979189261,0.016594583075560165,0.9307479042195563,
 0.2791793972010175,0.38017454227235636,0.2524003610693578,0.1476595589284292,
 0.35595307225222095,0.04250710989721418,0.5977525369898897,
 0.007738421433082676,0.9366567215380724,0.5597426830416793,
 0.3991773150140254,0.8009635347342712,0.31800132668166337,
 0.42979999442453154,0.009731664922731365,0.13157406362590318,
 0.21110459819106275,0.42528728456156195,0.19527765042443834,
 0.5707538427022353,0.431017018098366,0.801635298262064,0.8718575599098418,
 0.9358130352763245,0.09646388677042195,0.29984372640324275,
 0.012515163223257408,0.5731961284401355,0.04298847545762885,
 0.9710493698944096,0.2820578479740472,0.8829597035837313,0.5813570005483224,
 0.13714560184919333,0.8175051069446354,0.13534300830023516,
 0.32366953974941237,0.44414704439538577,0.4057348971474902,
 0.006511243964348323,0.40233765893194806,0.27595064010431847,
 0.7326822806669132,0.29667044161938383,0.4029355802374601,0.6893968417765651,
 0.6709336988067538,0.10265235698774966,0.015136175415691788,
 0.30165931646737487,0.9057809741097997,0.1626892014962501,
 0.25993935264703116,0.4208267768795042,0.08361308024384306,
 0.2591343545496191,0.6522495489130683,0.14866006847974256,0.6630463497973177,
 0.3795676414554561,0.0020156164031427393,0.5234591502176992,
 0.20487574021624555,0.25207123512063745,0.6196931952465793,
 0.7556422408415925,0.010055779266877174,0.014708925929071692,
 0.06194362426177291,0.3749745258110898,0.3052950289036549,0.6107311315159422,
 0.3887047674137357,0.04661611331835758,0.38007176043797963,
 0.41548321993328896,0.5704093542332965,0.3505383967259922,
 0.024063629252416008,0.12016863789223053,0.15253317469624927,
 0.3887168461455428,0.7352465147497201,0.04328735165869635,0.442196542590624,
 0.6014791322514202,0.8250110478171309,0.7346452627151008,0.05129335069845886,
 0.7902699664512938,0.09533621100039891,0.9098993429626742,
 0.022076607910117785,0.3077413965810296,0.46667026405395395,
 0.6196089775954786,0.397977716263507,0.8681548546689415,0.9428665882015927,
 0.85866124519276,0.6006162625507138,0.5110378494689842,0.5706532843129561,
 0.7676168229187899,0.16451390809954314,0.3104057248208951,0.8732086197031083,
 0.5557065807986763,0.6272502323746403,0.5118583892887844,0.8568283046239343,
 0.3298639782589745,0.9664314453578011,0.5373957052305877,0.9701746244677882,
 0.4684063088167173,0.03447056583452124,0.46240256188582485,
 0.04070485966602766,0.33472373398723176,0.17261000820089623,
 0.24936137867868013,0.8653911562838354,0.09998273276058778,
 0.6395222358266572,0.3956161491884901,0.6922936005894655,0.6103968865852867,
 0.38721275700556157,0.39222842192649954,0.9732257650319635,
 0.5714330724131723,0.7866556080920285,0.5847411152189962,0.5771206188991563,
 0.9113386144904905,0.8100619768172453,0.02284061581639596,0.5061956663796484,
 0.27304438403955966,0.567429972487843,0.3380792355003528,0.5057803768522094,
 0.4199902054906812,0.13430836650449351,0.5469679762254239,0.9517075949718413,
 0.06971276342725197,0.8243468026977505,0.47226137468348073,
 0.6075222179858178,0.4199542185106703,0.34872936203385785,0.6587200732067998,
 0.13759780738705107,0.9613366755476546,0.5975663520534862,0.7453724161018765,
 0.11888303170341685,0.6837043274497429,0.09665943744096728,
 0.15126123462985241,0.700872036082718,0.665797008571551,0.7304833941516989,
 0.7386749903717296,0.3701180473750735,0.5791158909461466,0.2346418318167084,
 0.655170011785065,0.8136687324082948,0.7649901917132352,0.7978231176020527,
 0.8918753202359434,0.9465362836585852,0.014306061422861283,0.714231693811076,
 0.04555942749892794,0.6224131930261542,0.24044785623391196,
 0.4873282038138145,0.8860899498575927,0.19342127331319103,0.5633528114717823,
 0.5593150885615528,0.4077222996220753,0.8381107517662442,
 0.033381443145992784,0.27312078834339537,0.6302663700764424,
 0.4988748646933572,0.6277527282518631,0.5480420096727716,0.913958703070894,
 0.5847335054451857,0.9871751250813142,0.2891038727390437,0.5281599132432061,
 0.6092378383430278,0.7245828292857825,0.856930441604255,0.7301549399703708,
 0.36166944917982424,0.8547238146164219,0.8078640871764685,0.7756440618819901,
 0.3899332384763695,0.29302535525791473,0.8737441414828101,0.9238611247410005,
 0.6475947881584645,0.7881892390330086,0.17614547632752542,
 0.19165953800897328,0.8121828829424038,0.5403352965129377,0.2433657241768774,
 0.44324899019214203,0.7780116519343097,0.7497693279326361,
 0.19524755793791382,0.034903204394924225,0.16164423635556024,
 0.8849176598471843,0.28910738755758236,0.46404605848397074,
 0.4733543808185391,0.38284193442234216,0.7195417756947725,0.5395132475098144,
 0.3125658241853644,0.9089531982395344,0.4778414200726333,0.7526345534000312,
 0.3880064715716329,0.8903940864853378,0.49844675613598266,0.805399690175979,
 0.18723122317232932,0.6286192717531824,0.2868895563302851,
 0.31819588973677926,0.9053551835708062,0.8548410028490997,0.6266718121188077,
 0.053373985590354334,0.3904823543578204,0.6302711656766625,
 0.38995625892716224,0.5900992211261423,0.7337084597903942,
 0.43196455336139294,0.7269930673950075,0.8227992927833935,0.7105720600316568,
 0.45182279422895677,0.21156782088049608,0.910912896944219,
 0.48800031351238005,0.7655813088270977,0.03673377961195101,
 0.7470102521684115,0.40647109547002513,0.03676071678402515,
 0.019418395371770147,0.12133063366036934,0.7310206187965562,
 0.08091311284774172,0.5790549597752777,0.46162987770946873,
 0.4445345992501242,0.6046282810857393,0.04993681851669507,0.5949825123881425,
 0.26714303845261667,0.9509239855129621,0.9702198483857083,0.3837757027181553,
 0.13625658493603265,0.5903617375906985,0.10152210674154671,
 0.22288489658504018,0.8192778689486743,0.5220618673535057,0.6730748262483144,
 0.6645411723266255,0.7760627612360957,0.09567294335809351,
 0.45894851639923673,0.09670959732736861,0.3216555078713115,
 0.5732474000867227,0.1923833943025256,0.8938434975406634,0.07148450480028423,
 0.5068158078621723,0.6011220880116259,0.5201256645222367,0.5646557875800898,
 0.16779445337333976,0.755107065317408,0.682352775016719,0.6480154204763263,
 0.2967854560073264,0.06721363176820627,0.3396034852286638,0.2039983966127512,
 0.17884804713646796,0.30858466244132876,0.3166388164698687,
 0.46531787027939464,0.5823280581200624,0.7874026832032095,
 0.45295324225840117,0.33773658957849084,0.7089350145974708,
 0.39313024830325427,0.786036679952674,0.6739610263852522,0.23639557569587089,
 0.5236753524450314,0.21795873701781088,0.9289010108899023,
 0.08531539989954029,0.10535121592196728,0.010032835531912587,
 0.668346576445894,0.8719468658435108,0.701575327265235,0.11839547854825383,
 0.42164334005490123,0.36519404374665576,0.682521139666425,0.3379540823694446,
 0.8623207700307287,0.5804620239159266,0.6089350029656153,0.22109003263363192,
 0.18209565905259217,0.7359788023511297,0.4260968758446001,0.9482657368673948,
 0.8523617000614648,0.03417102330914157,0.8616199956314734,0.8999368933456829,
 0.20514544208588836,0.3545782498767621,0.07384648294662166,
 0.3740025215398052,0.40975758318526134,0.23107162415362537,
 0.7013439662403417,0.8968657791584946,0.9521294342102654,0.0663329227304511,
 0.44678891332619974,0.272668742550208,0.5690187128331339,0.4365890026571154,
 0.41806100050964945,0.4774268670953743,0.598782034302472,0.0753165703141584,
 0.07978889228658059,0.4702344721921624,0.3748941759302298,0.2937919906423019,
 0.9073019885292276,0.46005971765913933,0.5865218222893951,0.8926745038842477,
 0.7281331429157789,0.15061920062805234,0.7870716393379322,
 0.35646629920136774,0.8439625985732961,0.1960049930523663,0.7366020319166967,
 0.889004667820623,0.6349844812719219,0.5250405290320522,0.8909475689348662,
 0.988943113147641,0.9899610450604636,0.08513483576006986,0.8921553882413163,
 0.6594130062476411,0.43769856373453786,0.04539131177530098,
 0.9472851901550894,0.18186864856309914,0.3354131847535248,
 0.15063141498326948,0.5130418965160704,0.670386995031011,0.10304907595451274,
 0.47960227475677086,0.8499075932882085,0.6778649268335689,0.4273585403290866,
 0.8962565122954562,0.5873200947193112,0.8066160467926506,0.06549798677459029,
 0.6527143587541399,0.03390477897006705,0.8812734786872253,0.3119012420988485,
 0.7971913756589598,0.653485286581368,0.5691925151031485,0.2758605845322225,
 0.972183695671899,0.8103771246945921,0.060350290319024946,0.9209868167754566,
 0.7310744643771524,0.5383297755333486,0.43603928548794946,
 0.18754020462614285,0.4085600207017118,0.13557942810969736,
 0.2037330183087529,0.7552374658343722,0.9468957078483883,0.05176616298291814,
 0.833155819023196,0.38679338066705826,0.06404182707149753,
 0.13646510453203253,0.4237838166972602,0.007941180453369445,
 0.18419708975988436,0.31755393701065515,0.21702617910187683,
 0.5196384146611444,0.42375753465665067,0.5436737110557185,
 0.37739030410732577,0.7159501470677121,0.8865651200857815,0.6133914085081992,
 0.7444468030064402,0.6456895150663845,0.8685145555492031,0.810455368434148,
 0.7141577568256421,0.1542179798679577,0.8082022748475025,0.36707169325964795,
 0.8637254720197203,0.20140054559521392,0.9600968859472852,
 0.17672162948156867,0.2050415933201184,0.2883087818467087,
 0.26119116719834357,0.10919023058816979,0.3953872918460205,
 0.3653374358493877,0.2993833474000843,0.9282702580200811,0.18290729143051632,
 0.523657841205334,0.19795264254938194,0.6811542882616433,0.46649570645567684,
 0.29458069800288067,0.28217402300037964,0.842816788063566,0.6354536337202777,
 0.573517965811435,0.4869994269782949,0.24793677554024152,0.5365933788398796,
 0.6701712391000948,0.4362965143480231,0.6751641789588197,0.794821311331775,
 0.8508558987903051,0.020811509012980678,0.5008218747238767,0.618005458381712,
 0.970448511697221,0.6311021370804188,0.5871896318424616,0.592195919564193,
 0.1572320130994338,0.5800746820332894,0.7364559236687381,
 0.011953585169817704,0.5695470736504153,0.33038373853282355,
 0.9412969430283166,0.6821573944066572,0.08438518197830258,0.5806017882263642,
 0.46401545572899683,0.9303232772369556,0.45097522695199244,0.546021369364292,
 0.7081779885594311,0.3746045729356907,0.904277658879672,0.5287664832204967,
 0.6069582772055471,0.35055961954121295,0.9327300043751398,0.6626407510435561,
 0.9881150763158258,0.6843115512509865,0.22098362871869248,
 0.16994072991403586,0.4217742321037673,0.16501309734099034,0.891522217426906,
 0.969979529741208,0.34038373521321696,0.6745227952420477,0.6636209624210737,
 0.2659059024077516,0.11414248099994384,0.1737312262294728,0.6682899329356898,
 0.7611228934939456,0.9269194889286893,0.6570864646466278,0.21114758245490428,
 0.8924678379134602,0.646865938327115,0.8782666249541543,0.48313334833916366,
 0.9400002601055718,0.18776420158455998,0.8451762836663093,0.7512483692404024,
 0.1040915987024068,0.10913102627261384,0.3033915968088621,
 0.30733680289923915,0.9540761465625671,0.8663236349572607,0.1297657701503755,
 0.5880657758818676,0.5635652472715028,0.09113512416458258,
 0.33180960526723324,0.7180807916047974,0.5590182194084077,0.2941102447749273,
 0.745385505666612,0.6228770936741521,0.7466204468718389,0.6402510257635263,
 0.3579369089341994,0.35805065392420965,0.0500955012041262,0.7859415057832598,
 0.7935171658654205,0.7145151943138278,0.4347928973282714,0.7648755325006404,
 0.12213225798561744,0.8727838910038773,0.7733548896953851,0.7360581994923793,
 0.03656517449476471,0.4338743129317063,0.9599592294148046,
 0.04068188129267725,0.7968424698618685,0.7456651890064171,0.6691785976814845,
 0.4221020508869511,0.439948659455763,0.36046721144990623,0.9311545885766049,
 0.27230188406211264,0.2994223267819367,0.8125676854970094,0.7695795885000785,
 0.6093508287308195,0.3196267371492152,0.5577703561370683,0.43856013208358857,
 0.011604067405815433,0.7049169047504243,0.24312284819538088,
 0.11702331910177621,0.10361070100389524,0.5666383943261277,
 0.6302510112455039,0.4487614181304924,0.7505164124516898,0.8231856524104354,
 0.8962673610233445,0.3666481167342077,0.015983501110395926,
 0.3927969899684185,0.01133423775781639,0.32884767655932046,
 0.19097840859380177,0.287263469529324,0.49014022731702267,0.2093854118205999,
 0.28388403012672425,0.7022783832913999,0.18285805083865514,0.761183182863985,
 0.3830269938958607,0.29613293678686525,0.9774732901162495,0.4752162434785586,
 0.44955349400943456,0.12543936042289516,0.5700244824119775,
 0.27811618722497267,0.21939681056878935,0.7487241519004474,
 0.4415854885341487,0.9233939532202325,0.7587272307215279,0.6119410358524373,
 0.2704645407162485,0.008340402466795016,0.30658545716192853,
 0.5721642597765466,0.2742723286485522,0.05779358285002789,0.8849858406075412,
 0.029808035621101125,0.64132718929131,0.421060091855192,0.38767478065781646,
 0.7258949424501729,0.7775409072805379,0.27277117014456953,0.7361588387182443,
 0.5063037401496499,0.38075338681089765,0.6197396246007355,0.773860699874455,
 0.3185547754779311,0.29196656227397966,0.3834199567481973,0.6540725923666679,
 0.38193197588816374,0.9702742531907049,0.562555418973706,0.8715059366911934,
 0.4634868776842431,0.5520619456378898,0.01760734772174144,0.9011781869006525,
 0.8468277493646368,0.13124641533346593,0.6158615201918027,0.2259072095845963,
 0.4243376489043902,0.948305165066484,0.5325173178863296,0.8170234769961735,
 0.45118850121800025,0.4563327595233857,0.33632518778920684,
 0.08020775191039775,0.17289645656056507,0.1680792214834027,
 0.12124326569706889,0.5830772797798103,0.363622204570354,0.6766699411126602,
 0.9073611886328341,0.13025482191176918,0.9932863417046824,0.4748556488012635,
 0.0559717828571169,0.17859936666876786,0.3644367192385116,
 0.07958138930312475,0.18533045261784153,0.04287388645497625,
 0.8204044027628766,0.871702255348552,0.19155244238551272,0.10576680678136963,
 0.9605410980592763,0.6290748819726671,0.4764331898846741,0.5143256332817171,
 0.5133955362040008,0.9794671585483212,0.8953268103354691,0.7588061620032622,
 0.7190410338137662,0.09274979959411422,0.9939271262744386,0.2754938485916584,
 0.22252444051524356,0.6067228622527091,0.6581239721979218,0.6960210221991179,
 0.6825367125635694,0.4546684719124019,0.5652992657544925,0.7238329472555157,
 0.31064058215650525,0.8763820348744775,0.19825002333220842,
 0.14452089024009362,0.012200131551388105,0.04892284573211114,
 0.908581200939369,0.9094229580311186,0.9645469759293871,0.5324230692396649,
 0.9208791440251419,0.32004555220153064,0.24353909322264733,
 0.17120598852617874,0.7661470463227195,0.8590670789113491,0.8346978004484781,
 0.3853305419536178,0.3601300448414442,0.8674506975348749,0.5431016010734464,
 0.512018539865635,0.00543455667038395,0.9208233093507795,0.2486996982153984,
 0.2598797401683567,0.5689180179269219,0.260260971862704,0.9016139474347746,
 0.12747363715361626,0.5088409020964634,0.9267449261197411,0.4484227661881015,
 0.934338474920346,0.9637730692003541,0.26094421835044723,0.3308121734403979,
 0.29872122237523546,0.8144198537750207,0.6652150185385584,
 0.49888600094705804,0.46788928874179503,0.3907914594865435,
 0.4558794959995609,0.26409323199724766,0.7452043481086955,0.899798782703356,
 0.8291583155688337,0.9622729397391776,0.6883193161909895,0.622568944505659,
 0.22014347277856183,0.9970898700375355,0.9713912511566646,0.552249112394717,
 0.4579229311790476,0.6170708831059851,0.6568322892385893,0.9396958734338622,
 0.4500564131348286,0.4334829068562619,0.7236717182229085,0.27803300110602325,
 0.03686086553456913,0.0722504593662282,0.8004737448470493,0.8491713848427718,
 0.0018204713294397434,0.588957826616054,0.5794710108559236,
 0.3481347998979025,0.32700484392364504,0.9854449014211246,0.9137905895255779,
 0.4671555659347606,0.34829964011617187,0.25582688618105687,
 0.8013133789764499,0.5932917788109484,0.9882429695064532,0.2971872998321693,
 0.3012997015699339,0.6260713505680487,0.8516057868053696,0.11492834658833484,
 0.4242866645120297,0.09183054875188812,0.12846438793749926,
 0.7657088385765658,0.2977702331186526,0.10826151567561582,0.8628625414133188,
 0.9819998609974177,0.29863155670515096,0.8016240016928894,
 0.24693038571491044,0.7483653947979301,0.4791798307784333,0.6776992920987935,
 0.09390051621240514,0.4782816024132104,0.2171800028348122,0.8649559876553254,
 0.5020696336273018,0.34840841067156525,0.14145450443231833,
 0.2574772343309705,0.8311525035106594,0.009683733862506205,
 0.06441293413355198,0.3568595396148516,0.7023635745075996,0.2838540335673988,
 0.2004057658796472,0.29824215834758494,0.609749457482152,0.10812408445988697,
 0.035045738696994455,0.39185317856526336,0.28955102142354794,
 0.6863357895026667,0.6614425398268845,0.2699950843881911,0.4951468370484179,
 0.3113577786297349,0.27482081715662376,0.3744263728362216,0.8555210619614407,
 0.29470504968315636,0.26542262076053547,0.3784863165188661,0.258466771611386,
 0.4165858230923428,0.4501477194985156,0.5948295369491647,0.4785996679781276,
 0.8390052407358242,0.9486964699509448,0.5557158029943625,0.07730946015031026,
 0.17897124413596321,0.04253457502625291,0.3599559505934301,
 0.03304680612966304,0.27395706980241286,0.5075782856016096,
 0.25405186719956974,0.3509868783618416,0.7526002855877026,0.8944470118883919,
 0.7410786264464212,0.8509177846919376,0.06552289528271282,0.2928839960541174,
 0.216511335336657,0.28214575460975966,0.5490994476081268,0.1112775092145486,
 0.5318677927139163,0.4012701869756271,0.8389458152705513,0.9215928201639916,
 0.02462862290642498,0.9153375024158417,0.2821588342134962,
 0.48874736469611835,0.7019270614252533,0.6738053337214838,
 0.20402310740904417,0.3100998216613823,0.9116335395417455,
 0.11861945015759734,0.5000923457131314,0.8916033602715893,0.8166958096115803,
 0.0522208350647253,0.18429990493062498,0.9845937807152473,0.8391478121589261,
 0.45870679015274063,0.26904099777017965,0.6617200939760978,
 0.08271373369002077,0.9618830924044282,0.9556571263414477,
 0.16067471811585543,0.12270454726036717,0.9269979120878508,
 0.20328825321737887,0.8309124466297704,0.7832619761659683,
 0.053388558990977164,0.9253734214036888,0.645366933085349,0.9467240314626163,
 0.07233214533137522,0.8584732555955175,0.8272537515855174,0.9475497090036529,
 0.1998766510167862,0.964354070398191,0.17506270678448344,0.12795432301419873,
 0.3100116410851643,0.8826305963296724,0.28796460158838666,0.7443544121973114,
 0.7371053905801891,0.3731727383761414,0.998122173877936,0.7114292637454163,
 0.5995757811223663,0.765863901966794,0.8449597845525597,0.8508593131240008,
 0.2584314382489771,0.6941484317932312,0.703376426068589,0.350332478589505,
 0.35034067450412165,0.6289673985756912,0.5818435763142301,0.7035912962854529,
 0.49122288253116375,0.7173034146126763,0.49426537375940816,
 0.7722788178528969,0.21522122534318922,0.5981263913631786,
 0.21846003194355312,0.7725330102190473,0.7564252169759599,0.5890070854577506,
 0.5061985811218717,0.9054031334806883,0.25541598080118366,0.4938867125136336,
 0.26929133912711034,0.5546096052031538,0.4471314409999343,0.7818334180522265,
 0.24095165399382912,0.16783145879530292,0.42584091573448124,
 0.8963163188888073,0.36218212700922203,0.6489827299841523,
 0.44247238160478786,0.7241739987329198,0.4311834250817843,
 0.46840838106766647,0.4167101992125677,0.8984548194115827,0.776771023791251,
 0.3962258051872313,0.1944741432101963,0.8235713868346889,0.35608143036240336,
 0.6431701951261101,0.8517893489202691,0.8182313687010134,0.04788961775233802,
 0.9692729127791807,0.2755708285972803,0.015144475405843294,
 0.9069686933246399,0.07420039498161068,0.6127460813672321,
 0.24514515057798314,0.37206737970198756,0.08277129527218285,
 0.8565978579407297,0.14582084726562528,0.5966311005817939,
 0.09603405585962943,0.39433082643912964,0.739835624535123,0.2768208302176991,
 0.6645587904287631,0.7055987473109887,0.22513066833959083,0.8437862998675024,
 0.580741029765748,0.06526458573420313,0.6524013770286018,0.33753291683361886,
 0.008383533300384438,0.8427749235462368,0.5562951941782771,
 0.9617087539649322,0.8924227123654253,0.8926610726716246,0.5323555122457123,
 0.46259231792110067,0.7554156949836993,0.32633831133530633,
 0.2852522144433134,0.10519554607519332,0.8332027470199259,0.5993755388580013,
 0.2652164777380148,0.8711342307299047,0.5325868789039673,0.19993828728716323,
 0.8908231345211006,0.8494110784881693,0.542692023516273,0.021506937050684627,
 0.2596326193489956,0.9539810063246195,0.8126491548352784,0.743574146985358,
 0.6006445450954512,0.8220128766558601,0.99952260244055,0.6464301032453783,
 0.2742939632289627,0.5802404443979671,0.09775950162038982,0.1595184871720582,
 0.8298832482770941,0.3712024363187951,0.539501946429418,0.63156639303289,
 0.33230656453324303,0.9820030298506466,0.42968537129112705,
 0.7566371810866679,0.9332923486543596,0.27118917162921397,0.1874040629578615,
 0.20621088575279656,0.003422982093842619,0.4574972293441877,
 0.4733986064588548,0.019033673752680613,0.7086978069400423,
 0.9756626033918921,0.14277149315912396,0.12797060235555135,
 0.18121297218758237,0.018499990720219905,0.6546794505765404,
 0.14820831774410592,0.41880770755065333,0.6688752236586504,
 0.3714055266648194,0.03922075912779488,0.20622467467326944,
 0.9892909166543217,0.9872070259472715,0.44881641903008296,0.9854835279809826,
 0.015166294465201235,0.9851102308804834,0.7706270584424372,
 0.6862113745941114,0.41038925091303147,0.6984211179585085,
 0.33814183139270326,0.3602026803798468,0.48377878442942357,
 0.5505552351853344,0.305658695312233,0.26005668072320254,0.04523012914940505,
 0.5231658275223557,0.45172036985655983,0.44766962787283815,
 0.6888845357638925,0.9339505875246817,0.24418744863550634,0.7088622735489879,
 0.44739181404554285,0.9550094459683064,0.6592639560331438,0.9648149279538922,
 0.4433425902623698,0.4712841192028846,0.6718713255338876,0.2738903523683992,
 0.17923188577504923,0.6388864343504492,0.1370352785700908,
 0.12507102479083299,0.5691944132188208,0.34605817603988864,
 0.23728216586468642,0.20011502294440775,0.33398692978026023,
 0.7684245103542351,0.6548435637050006,0.5441536844808281,0.24204782499594568,
 0.09542273982104499,0.37683793395822285,0.7965277444393511,
 0.5414639202140923,0.03855755594741683,0.12584221504177917,
 0.5650870149771984,0.5163631346323193,0.6642936059143065,0.3720630007574941,
 0.8969390261910837,0.36958498205245427,0.8625442925834469,
 0.09803806948453819,0.36596977875861525,0.5191720646136891,
 0.7806884180896208,0.3240471600337972,0.11267949262690957,
 0.36671880863156636,0.5302470176159958,0.5770963385190939,0.6843672958819399,
 0.48891974309375563,0.19062857311953663,0.2600066192125704,
 0.9856931418527216,0.15358728419957912,0.9813076131068902,0.9637067003298599,
 0.7273255962119423,0.5124071778169246,0.33370663933885014,0.5633217586863776,
 0.13245920931186572,0.9022233748905151,0.6107631281974162,0.5582757635506039,
 0.6253078029048645,0.7098459490134158,0.14872843446814654,0.7996916088084578,
 0.25201466541725415,0.02087622435244163,0.939965224008555,
 0.15386032104206349,0.7058073118796171,0.5327085128471817,0.9655537974453713,
 0.511692825970927,0.31063873352775884,0.06850825851169406,0.3607679961572815,
 0.25455232917950443,0.6198347809280708,0.04344175034208253,
 0.5142005589968985,0.9872040671217338,0.053789350150480364,
 0.49883942327652775,0.37262304777139854,0.6961004928470786,
 0.25083087003555815,0.6984433695493389,0.2154281717275679,0.7800982895152622,
 0.5486815036392142,0.9981825683459971,0.2608162300974197,0.21561917058510516,
 0.3447296666994666,0.9873727869486415,0.17214865086032094,
 0.40035234384523355,0.6959348419205289,0.3766447435007826,0.4757328320967592,
 0.04012189547189693,0.17120456304821496,0.0272537333422459,
 0.22903820407230147,0.12297762591204842,0.5817614310360695,
 0.08406104232185829,0.8266514637864373,0.7428050624421806,0.2731914265447155,
 0.19728148080341246,0.170506284247261,0.5813628215645387,0.1773400148229467,
 0.5831026373678269,0.37470289110548427,0.8181219877093913,0.9137531796136366,
 0.9887936410948087,0.46772269816553746,0.42860941865206614,
 0.023082194659263755,0.586701158830332,0.020023740070268747,
 0.8887170329265284,0.13389512927946368,0.1476007085505059,0.9638169514364552,
 0.6927021495017934,0.622681237503749,0.03702936837680404,0.9251011930314536,
 0.1525100398999667,0.9140998218382396,0.4477901841534049,0.7416278526551511,
 0.6797976629237164,0.13755503164635463,0.20488780639212723,
 0.0390830718826356,0.7599757063142462,0.056497706068050935,
 0.3230172327169675,0.1951448329571397,0.6890263943992504,0.7908639390306493,
 0.5767966634075415,0.5475947731023152,0.10675427217002453,
 0.46409258991451496,0.6775146390024522,0.0531019865488509,0.6059521129064759,
 0.5205608218672321,0.5221603333331601,0.5156883313627679,0.10269100140509702,
 0.06769582877363967,0.9762981095182155,0.8413372945047687,0.3512528328374097,
 0.535173559420887,0.8661304376321441,0.1075457635763577,0.189122424124196,
 0.7672846385023621,0.5514076484412752,0.4326592307662749,0.9118877491960925,
 0.5713933974451942,0.15602048752610964,0.9739334040590655,0.5072770445457586,
 0.5872376274293482,0.15215655679137374,0.6660772454007186,0.2356519582745823,
 0.7494517948339877,0.5499352946494713,0.9656951643129638,0.31140643076424046,
 0.8148646308152357,0.1938020271764157,0.3905063410231211,0.23650349477353516,
 0.22377070021642798,0.40916112829925966,0.06146581001267837,
 0.969235982696968,0.9909440203357467,0.8012051132610246,0.5796019980671946,
 0.24415046970599963,0.048971925719689,0.8057172514006891,0.3125485744223302,
 0.17508858323785614,0.910501202249535,0.4339023926117722,0.1906334047303977,
 0.7722403334729163,0.47885417541369313,0.8567263886731056,0.4713770397628265,
 0.3171854559363986,0.8397691324067156,0.6206159862648584,0.23416600574699797,
 0.7832402552773394,0.4930352983153051,0.05318293260079654,
 0.23537740254947315,0.710979628654318,0.8861209061779944,0.1261552014624554,
 0.6891820664460626,0.4061545771327273,0.32269996452293936,
 0.38819824596171626,0.8335962187440534,0.77898177974425,0.038125448077003354,
 0.02930043173048913,0.5353077760539567,0.4382719868425381,
 0.005004453564238287,0.7483714920562247,0.17272483914461978,
 0.1660608652512513,0.722471087428746,0.9534239379666339,0.9544237543302344,
 0.6219124699224615,0.271782039196651,0.15208999203153195,0.8083561647452873,
 0.8806851921154802,0.46535554592068273,0.38006722081749933,
 0.11511106827732243,0.6795213819443647,0.4740147124463545,0.937706400926948,
 0.8001051014143319,0.5145727813067481,0.5817979586325116,0.25559737229861124,
 0.6877741435339482,0.14140696922948215,0.36467671111226807,
 0.5782553123298942,0.3505217634989076,0.1098766557134474,0.8550497492923254,
 0.1101433963912073,0.233361300602571,0.6978264620171707,0.37198412807891124,
 0.5448359050303064,0.5035802441552703,0.4559967578653491,0.9985189168678911,
 0.954494038393816,0.6812424026440771,0.2748609400353277,0.7968906514880144,
 0.12008413536457652,0.8937988481783544,0.25792432851537694,0.749425723232053,
 0.172446577155987,0.690065573876448,0.10671929990481788,0.10046810880521373,
 0.005042893250788771,0.18141174539364302,0.3008089590306019,
 0.11178144942802404,0.6558969365713533,0.20943773126030174,
 0.0019709280345454605,0.6864942303376651,0.678523298781158,
 0.5947612861785295,0.2989341227365532,0.46999215501434055,0.7826651588165714,
 0.8223611064720453,0.45230571187560675,0.284333882083873,0.38367769046018774,
 0.6409340070647275,0.9329656079515585,0.4739032979860103,0.35880502010444937,
 0.646643138033284,0.9964078966138344,0.36513183008326844,0.7225550996859388,
 0.18120549153409482,0.3108033637933222,0.9954444126560579,0.972206949835957,
 0.18436366874992627,0.11168399085403058,0.3839098160744655,
 0.3961827582648243,0.4774235462465046,0.8741314676148737,0.7860397019903139,
 0.40024438294310605,0.8047328846274998,0.5097106453549803,0.7071680079673219,
 0.0052837605503698315,0.34890927191522714,0.2669894450313919,
 0.24675888889154007,0.5310125226519267,0.7442040532133388,0.7096725756489735,
 0.9407758524238532,0.49009958723015234,0.6727921889982105,
 0.36712796792366553,0.4272663222075821,0.6908097026754924,
 0.20007241656065833,0.21558051720384708,0.9741193666339206,
 0.6557271579397363,0.9800444016110708,0.5084514688859978,0.7850171658437178,
 0.8107801588428185,0.22139913961902202,0.25500783058001963,
 0.8724156631280817,0.020619646945687964,0.39441351337265085,
 0.013157811558400678,0.2811909103705562,0.7154750799379259,
 0.03493912717666281,0.009117315979557117,0.9716026686416029,
 0.141714731110818,0.5879611639174077,0.9990883323009512,0.9187355012992503,
 0.09681905275916025,0.49151018512189637,0.5338204977891152,
 0.09279797400189371,0.312892818874982,0.3445043836751023,0.9203142005393254,
 0.9271769886445258,0.848816612506665,0.24461503881365942,0.6026684845045968,
 0.28523223127320363,0.0017085127203826378,0.8836778014382144,
 0.4047247514556712,0.6237045904716219,0.8304193582543322,0.8050701393839892,
 0.10620431868430391,0.9792136188284493,0.7371894044186598,
 0.038729955144368056,0.9329188412331578,0.7389888153586498,
 0.8847482053664864,0.850014399788919,0.32714489910342615,0.7669515945927057,
 0.5785450223005868,0.3814908788998942,0.052995964929745565,
 0.49972209410760127,0.3485534968821955,0.05055344121644534,
 0.34191169181741554,0.04083911328037115,0.8822715827963631,
 0.07617961385516869,0.20626326219480384,0.6490316989519588,
 0.4551792314320018,0.07774688410147756,0.35462850216066266,
 0.23107854538585393,0.8645426525145422,0.5966600319246365,0.8974436446719141,
 0.3855384946386572,0.30031205026941354,0.7000472194181817,
 0.30187420085310457,0.0358046381569892,0.4906513429715167,0.862537588444837,
 0.005058186108865548,0.23301847730497416,0.8089004263847664,
 0.7642397170313002,0.01733277087107521,0.23646156503929117,
 0.9616029163456362,0.39074309886665504,0.06787341932561464,
 0.4885650070831309,0.784610631634837,0.07240453594212692,0.14815793770736096,
 0.5346953243333317,0.3618362298429614,0.14701343435001257,
 0.05686557250409863,0.9573217211934403,0.7967518981518692,
 0.25000537237327813,0.16557787145257175,0.8313903925824268,
 0.3352808131500087,0.3546310470880806,0.38171115495598607,
 0.010572276939277314,0.21803629223135768,0.07766347608312274,
 0.30246937120204065,0.1568815166014712,0.7805356871218838,0.427721194519164,
 0.5970344377931784,0.14572960059238715,0.7408839683759303,
 0.13728941476144407,0.40548402741655265,0.3913050323049153,
 0.48110359294796623,0.6971741635374581,0.17619627048358144,0.70290328498164,
 0.4908774542161686,0.9680618142090578,0.27022325286966076,
 0.22111478391673622,0.90620079291005,0.6524280691951818,0.39636290083426895,
 0.7799029388116738,0.9943051302392982,0.6367251363563717,0.6360860390450568,
 0.26078104862231943,0.6620012194351043,0.3872954585629377,0.2294548769156064,
 0.21334627366506054,0.5584019415418348,0.9167384714046968,0.8704846426430966,
 0.8334463231114411,0.3623008163976501,0.34329972036296597,
 0.0075771969063286315,0.08796138941544696,0.8810457773172969,
 0.02593294457171691,0.5256761266668262,0.5143684696868577,0.7321036589700818,
 0.8314552175441877,0.1802103939136771,0.22737385699764445,
 0.03480528583805287,0.4683363802159307,0.2966108535835301,0.7797985538147977,
 0.36393503565518115,0.18818893354942123,0.902881532759796,0.5250734183789842,
 0.02815756725523688,0.2208376826940549,0.7400322175671521,0.9540091466031749,
 0.45626784142937793,0.11128155532396034,0.8473693787522097,
 0.7512442291959006,0.21397683202762663,0.7989597667020485,0.5420681681365338,
 0.24914259855777265,0.9171902443925406,0.36117761350565447,
 0.6432358282087405,0.4704766214178442,0.567222308227098,0.9280980343853703,
 0.5814128515314397,0.3911410384643167,0.9178635169790021,0.9547813457914908,
 0.8538523509542995,0.9107886221660108,0.6169554296207642,0.15632656629653285,
 0.6510837045126026,0.5110865069179297,0.1877442627752941,0.31460418115834665,
 0.3899223311147356,0.3139625662443061,0.6269657502274464,0.3804398875047236,
 0.11094702803501655,0.06675173104409382,0.18187288592890194,
 0.25111944973404376,0.404010727360705,0.6470336585860018,0.2397924261423079,
 0.08099824278326107,0.3458860839226148,0.7942852752457631,0.5077018200591783,
 0.32734597554651357,0.48899987041452364,0.816063869414128,0.8544986786804066,
 0.9022821547033781,0.574545831767853,0.6302971442912461,0.24398114139988447,
 0.4299447660714234,0.46584212453132257,0.3186020701161889,
 0.15689245268459828,0.02690128551979587,0.6997137872878927,
 0.5690106716783547,0.41803336213476794,0.3757045495205691,
 0.04786100480740241,0.744379333550578,0.44997644545834914,0.505044180917646,
 0.4469140587275645,0.4932248016912021,0.8715364215557599,0.26986448256079776,
 0.03830630091309728,0.1153103415592831,0.4799060094516244,0.7975962206667682,
 0.7754685240952619,0.38814840186772614,0.39474603538495967,
 0.10673713508336713,0.07192731073251823,0.09159296979731513,
 0.8482941872112674,0.5608250301688795,0.4511556937264892,0.8259377488564051,
 0.3623973539339699,0.8245338287394998,0.2722669497543908,0.6831982434659208,
 0.5668682802776489,0.6878842272171879,0.6951350951952951,0.6329829778341587,
 0.9632374033910756,0.5542608915311286,0.18993105221665652,0.6044879822895584,
 0.9564280482842957,0.6486507095410263,0.29525479350808714,
 0.17101115028469693,0.030567764194761926,0.745142218837948,
 0.36619440643643963,0.3843549616550329,0.9866378392172771,0.7575646083324603,
 0.2705444995339914,0.3392383440816392,0.5641509752095673,0.6219633883483385,
 0.7996644849690723,0.9656847193126646,0.45413633763158856,
 0.06630624543007135,0.5184868543385915,0.5494256057363972,0.7952124197570315,
 0.7534946524859941,0.985296552773445,0.30486198097160744,0.1638781706449931,
 0.40905130330946093,0.2526382719180438,0.5130785432344809,
 0.44797621846688596,0.10960106289519056,0.18415800530739068,
 0.9464433470006487,0.43120154228265895,0.22334016755912733,
 0.14569029702552605,0.6853657566596383,0.8182746511785945,
 0.020227802209191514,0.11386881783320868,0.09216955765007462,
 0.726663835040428,0.8514449094539831,0.6076393962681976,0.959391842437089,
 0.8737816937089755,0.3834683129044193,0.9408804625002721,0.8819502746938027,
 0.47752189092380515,0.14361090654605102,0.13453705143228756,
 0.5748852331571741,0.03056126586895891,0.7372514771132851,
 0.00789047626196715,0.5086255311960812,0.055121760320271496,
 0.39325204378445433,0.953315344972129,0.5675965460065826,0.10282094148975329,
 0.007860955645060175,0.8231205961441983,0.7638571813309929,
 0.48325251741631736,0.6029749790095089,0.2873441811622932,0.4873179044928251,
 0.8517029418949555,0.16830883979575917,0.018013626602580657,
 0.9689640245167226,0.8389848177357739,0.7417448626544677,0.28299467313585946,
 0.002004606211132187,0.949399661828188,0.3044218575531278,0.3498716131009365,
 0.16586528482051377,0.8984828940254136,0.2921266843687864,
 0.09987867564799624,0.29654386254359766,0.549021661951603,
 0.0033298621615420743,0.035469453228005765,0.1440655442672245,
 0.7346836147322194,0.3124768635556723,0.6686688775297013,0.9706739307776209,
 0.9202703288197622,0.6985106172678331,0.7679443081674291,0.4972704708880915,
 0.25837059446940147,0.8073490984740217,0.1392052941278682,0.6543728976982868,
 0.7647258763803182,0.5221998654740854,0.2545649519935531,0.9804538684245024,
 0.2564002512918575,0.12572519958602246,0.8133981532508419,
 0.09674062296294272,0.7343632519331341,0.6547356809854183,0.7242967530933058,
 0.9451017463067573,0.6184480995816526,0.30184074331426236,0.4560434842159984,
 0.30012621466339084,0.5580289456206518,0.6897432084760519,0.2768864991980775,
 0.0242846568734153,0.03360703322814773,0.0050149867095730105,
 0.3296005477808732,0.5385509709121818,0.2640909783846026,0.7980571175707255,
 0.6673728523068978,0.3129354304709464,0.48287791550547654,0.5966979810975956,
 0.24080526116022738,0.9759175439683627,0.6922637416479371,0.5114642115750387,
 0.33145045128815265,0.18679333726122616,0.905432000961002,0.1873018919273166,
 0.002996546341094475,0.6739207632522838,0.3492220096493468,
 0.9264474185052469,0.9027368650910548,0.7148992462636505,0.9136642185569,
 0.2525442178056857,0.5260617287268765,0.31320993233937966,
 0.20043677399203053,0.8311586976585413,0.12122746800455808,
 0.36669410013216897,0.6213508130444234,0.08223297942822516,
 0.11516699081559978,0.05979307685116886,0.5423793866163827,
 0.9240050187785298,0.38941636053698137,0.2078266630501947,0.9803016870134981,
 0.8503940687458564,0.18753956717436404,0.9789327355650235,0.2155658163898616,
 0.026684568669087083,0.08616482514517054,0.957310963922953,
 0.07282014176115015,0.18669817497312513,0.17287606208339812,
 0.8015849537308017,0.37852919024903553,0.06778050901232002,
 0.09622482501268714,0.7260266438117136,0.376413249757411,0.4069181165074476,
 0.5592831135766169,0.9244475183002143,0.14639398597892894,0.575670373485956,
 0.7068182926049486,0.8610196833916453,0.06920518703059098,0.6016372357993094,
 0.7167845748753715,0.057124612314242595,0.822720093973979,
 0.34959246286748635,0.36300627542394015,0.24569239187899816,
 0.32651135298583567,0.1641838319440655,0.670998491658491,0.11300292215658436,
 0.27059414482613053,0.7382567674443602,0.9745156934863389,0.969250127334143,
 0.3864948351682336,0.8358694806163125,0.8712152709808091,0.7630083382441334,
 0.42167038792137146,0.6047824807086595,0.7920894978829766,0.7432776317157807,
 0.4376015292701063,0.2491948004550424,0.6076950393476802,0.5193928508897637,
 0.049703787691894985,0.9301834205443177,0.8220020763453069,0.125318599003991,
 0.26958459281243397,0.5364914094596837,0.6021373214636411,0.5528373084428688,
 0.1749360309603195,0.7996618778364204,0.5384677844492781,0.372934497041152,
 0.28282915804932707,0.6931987138638509,0.06590845940033396,
 0.9691815377239746,0.21840516816317013,0.3386483877016956,
 0.22212406365264015,0.15026776286877686,0.6435468360609632,
 0.9137047498863313,0.05308364968746515,0.202139331991242,0.8692744445916983,
 0.5159306218839254,0.4706240668383793,0.35897625710252834,
 0.30121468977379795,0.11502606787586678,0.8393206096002104,
 0.4844702590444063,0.7325902406158092,0.3663286901632604,0.11176305102316564,
 0.8598300366676428,0.43202428040454643,0.40393773066779615,0.488980564820924,
 0.9487601356139161,0.5950058519304088,0.3856283838336625,0.5513205102520684,
 0.513583142378696,0.06975759669996906,0.1875233491011553,0.6759881789343405,
 0.1357529613406403,0.6638824487396819,0.18717838595656722,
 0.37547459946830486,0.7010517110426322,0.5401991430198352,0.7616191049994091,
 0.14695583220523145,0.4887936481174687,0.5087325524624036,0.8651365842317355,
 0.05263629343340681,0.7229751434866571,0.24505023209567645,
 0.20330636618331188,0.16385374171980305,0.1244632188762187,
 0.48806642948124734,0.1310928011578949,0.6583228874855566,0.7847796010553405,
 0.5785117508328466,0.09353574940989395,0.8200123415011259,0.7893632831250816,
 0.30141302624049326,0.2167760918670032,0.13335994548361985,
 0.5213956108932281,0.6719213818738351,0.017727103309999048,
 0.23181775488191703,0.21310316901352921,0.8734872761150909,0.81997097761752,
 0.7197573749654239,0.6162339268984695,0.06562898068005374,0.9403852315884963,
 0.717340108010261,0.29254227630111573,0.10408615036086066,0.8091452392561792,
 0.017704920726573592,0.21585598478842605,0.35871970393232067,
 0.6635453032760807,0.17147775468158877,0.4455417261915281,
 0.40240010691438843,0.8703878900315382,0.773968635096256,0.9626381325904378,
 0.3743067359880762,0.9022031542061287,0.8389391838437719,0.586032768315184,
 0.512331071886948,0.2689710968220399,0.058209571441041996,0.7494761889287226,
 0.755939067971098,0.9674440986648353,0.6882499153891197,0.08229060888234441,
 0.8379697562512759,0.3431473893168281,0.242234382274515,0.3329112849494711,
 0.1141583789924211,0.29255782158767507,0.27004840304310473,
 0.16581298134475664,0.6455990894614999,0.6494515759511297,
 0.16914556328291563,0.42320821131218633,0.5354879169715953,
 0.9394037670265668,0.5137065150445476,0.6786343831596514,0.4411411910850854,
 0.5529236147098293,0.4805206329519045,0.2681350111285542,0.5183922824231684,
 0.028994872031147034,0.7778468793305388,0.22652331715195873,
 0.37387394035449106,0.46544990775066675,0.27042119269534637,
 0.20806266925537809,0.3392820853547216,0.5918212662418479,0.3083279661228596,
 0.31232052445755654,0.8494061411108517,0.4752202892210353,0.5104575557721519,
 0.26472898046908644,0.6185929759240603,0.8600153631006169,0.769840688494133,
 0.6100239366387477,0.3511703066373285,0.8041250888214477,
 0.030225326807266528,0.4115719173241512,0.1267116894908339,
 0.1475455452171991,0.11753520957987984,0.5483407551844088,0.988722748441814,
 0.7485439805106413,0.8144334023109353,0.793131544103428,0.3095257841084639,
 0.44363429365890183,0.030815231527638987,0.3665863486839913,
 0.24692756443507835,0.9421945977365284,0.9463120782735307,0.5718503646618611,
 0.6435514824374962,0.11543071356181622,0.7030431785794575,0.2850043370785835,
 0.9023831424118792,0.021217902482483897,0.6102378081254825,
 0.9461718631485541,0.5275374307904768,0.8240693121286969,0.9638795609441873,
 0.03155443869650654,0.7147241464096906,0.715232003909629,0.43159189136009213,
 0.012293722967103626,0.09930562390182174,0.9216668309539499,
 0.04062838291694526,0.697373726572555,0.8065914822756788,0.9583395625790638,
 0.09289924807605188,0.06289263751491969,0.08547391180439157,
 0.8470970651082049,0.5385904887847434,0.7156843641895949,0.2403979053432702,
 0.24776518632360722,0.4858846415260204,0.3337681906179155,0.7394130439861378,
 0.4487355544153717,0.9839643213876835,0.6294305482687665,0.7816341769295582,
 0.43040561784552733,0.21326252684834757,0.5856270259697993,
 0.31202273462261476,0.10963303031387306,0.011791386045841068,
 0.9914284598039742,0.7592623416097142,0.901467898275175,0.5309986115200297,
 0.9211238888912101,0.11417302290099707,0.029463054683880996,
 0.647292666026018,0.3413176153569949,0.45547339463998193,0.5716164331170803,
 0.8916839844338543,0.25729988486860866,0.2888394549532065,
 0.35306819620072505,0.16191330177095464,0.019580908413396547,
 0.9889347093044123,0.4099458251547302,0.9000192161412159,0.9467803700395008,
 0.5239677599207497,0.542449981299189,0.4694214329699328,0.32579802849086703,
 0.6054493575410644,0.7049877606088627,0.9089656079731793,0.06393072878727124,
 0.8442131446115646,0.9265868026246573,0.37663187203855963,0.7961111253990898,
 0.3002347054443859,0.20799593208717937,0.46732793481522983,
 0.8891112413986577,0.060419776368092704,0.874782274528225,0.4404293443921543,
 0.6654583228736636,0.07407598198743903,0.4531932066289308,0.9271547981287995,
 0.19888078283677824,0.5665235257685737,0.37275413786007505,
 0.2001050700846545,0.42123097647023855,0.11264118213458052,
 0.9913181480275174,0.26257409404500964,0.36861497662167686,
 0.28529208917330706,0.730440945722563,0.9806373498282394,0.08846872400646211,
 0.4365284739679387,0.7922632872978416,0.5140661512236048,0.38831432605922134,
 0.31906562059935895,0.20664050622077346,0.24691318971168463,
 0.7309018250859517,0.40941940601357196,0.7986056416086976,0.1433982603031847,
 0.30365038643928177,0.870005129218727,0.7687739592075878,0.6517299995374075,
 0.8008035391634813,0.6635895337938722,0.0703843778485328,0.10699188925927028,
 0.9461874126724765,0.9128639444801062,0.9709309274270825,0.27095373429261893,
 0.9823371518117668,0.10402431435667292,0.8387234820595697,0.3356414111535072,
 0.6507291120832597,0.30105409471895517,0.6475185135629526,0.9235093347668729,
 0.6096602459268696,0.8969787142340707,0.6365300212925872,0.13928048208766175,
 0.6285430974273217,0.31501311347424,0.7764953462649163,0.019308132236027342,
 0.1855832850532374,0.933999649800965,0.3989686875922054,0.9177064976530453,
 0.9135564723248067,0.1227589556698353,0.27704064390671124,
 0.0010270342463248872,0.48830320575284236,0.8944110508555873,
 0.7439896714850204,0.7123375317289855,0.09977611911789974,0.4207890155104973,
 0.7004711736806055,0.6276197944310496,0.5877668209143282,0.6142413903203775,
 0.045755791918380195,0.2206153964837938,0.4207365732452011,
 0.8748164992517251,0.9692334768424479,0.583277546790249,0.08654452083404884,
 0.42612363613134796,0.3570897284457676,0.6590401014129144,0.5382529784266861,
 0.6402605674372507,0.5926540428390306,0.0462269475862398,0.4671098491897199,
 0.7068461594826112,0.10415530789488747,0.6079647635644708,0.6642729773107495,
 0.6479586469647061,0.8727511796874401,0.53587539830447,0.1302046900833368,
 0.6224225194216205,0.5516226278542453,0.5083239394311235,0.5272037462934702,
 0.0013519474273875343,0.8607867554290705,0.8396748778837908,
 0.07043474652597237,0.8378632730201476,0.8813631203990783,0.7567238814239226,
 0.2890252013958453,0.6027417149517154,0.006337008383793918,
 0.23547315318699036,0.8080567982458656,0.15114210092410274,
 0.11402573542801309,0.12897411962476235,0.35934605389518515,
 0.902591641818149,0.030829823309543425,0.15208015692309562,0.803544468303963,
 0.5224424203119361,0.6033529004575028,0.6760682981865228,0.12480643530448621,
 0.6455663682800346,0.7566525773510562,0.7209645305627975,0.4557170474070641,
 0.6025685293823185,0.5756990949606674,0.7624493025250196,0.07349281832001797,
 0.3372636026354072,0.8649119774740153,0.7757350409366245,0.6649899352952992,
 0.2061808715307274,0.28537328507348525,0.7261534497539905,
 0.22582807141281336,0.15745151106996924,0.9036508230317618,
 0.6527799021243899,0.03661183401054047,0.6274569478547485,0.9074825342075908,
 0.6013080184636124,0.08629968127868537,0.9367513986333518,0.5178776480103308,
 0.9462408624680197,0.461416712221548,0.3279329971768018,0.8791253113245024,
 0.994686624560496,0.6295105863551259,0.7236860583518646,0.12381639318579984,
 0.9448128782389787,0.24605120782398693,0.98605057448689,0.3750185497229701,
 0.8217771234266072,0.017106126228831453,0.809660879032606,
 0.47631062690114057,0.8327034066935776,0.9713422006029986,
 0.16297396253514873,0.03695737384899578,0.9735169292249575,
 0.7052751583048421,0.34220297778242736,0.6444477220216216,
 0.11362325103174287,0.9453997945584832,0.9567269433117033,0.4812584547408205,
 0.5808828087604259,0.6482700354491991,0.37616582161087575,0.7994742187611702,
 0.6976708492588859,0.42453245774523407,0.1915271488584459,
 0.43834703544082876,0.5038412645652293,0.5027068668683516,0.4944653455693222,
 0.525745659118579,0.3930912866581888,0.6769790869123467,0.9452780079830585,
 0.411333631252473,0.032023384732586035,0.24474730334963202,
 0.9240078289016785,0.4654276106441809,0.4242026597448363,0.9429504823849203,
 0.6759258326635413,0.24339524707202564,0.4597877677408063,0.4305562886488534,
 0.6674037256198821,0.0724701918880053,0.7008243892630237,0.7221364244531585,
 0.739641269685152,0.7105507641762281,0.6872955812728505,0.039902131106962946,
 0.8821880184062506,0.6214649381005677,0.40125202504961854,0.6019565759909093,
 0.7403698674180377,0.6601210408429409,0.0343530352864001,0.290701563927532,
 0.3204829501229697,0.09940942818361309,0.8336341772873426,
 0.27728565619619305,0.8022257740411459,0.7273336335165377,0.8276188001195597,
 0.5115637425365851,0.9406390529117591,0.38790768520944074,
 0.26580919236510514,0.43130053483725206,0.009381660028074945,
 0.6286188381288655,0.666547632302871,0.5441102914067178,0.30977638734060786,
 0.813099821382288,0.49747411770952155,0.4173140611933538,0.5290752526901679,
 0.5366980518647412,0.6666153642224241,0.5285536671255479,0.7943555202474074,
 0.8036703695129256,0.05452177343747666,0.21522217943352395,
 0.7138356632677341,0.14483942079296397,0.8127792087875388,0.7820457395827524,
 0.6798348597027356,0.3004189308046912,0.3715262279993966,0.1621472298591209,
 0.9193742071713749,0.8849480316604404,0.8337498198661969,0.40766567551773025,
 0.7003240105640135,0.5363211566605094,0.41377733901566704,
 0.23774628998838987,0.31427401057226445,0.47902509900307455,
 0.8049513014329654,0.8490593103168054,0.761481663072906,0.930644018592209,
 0.44360966027527093,0.012785045327736055,0.6298305797889122,
 0.4287969121546962,0.17266910151611803,0.4210032465383857,0.6720891116028467,
 0.9976761709961267,0.5808100491509118,0.14138570905852865,0.5196310138983974,
 0.36415819688477946,0.9320434992060663,0.9000718486291212,
 0.22245468256539347,0.42466002166389105,0.8375546500548466,
 0.6469411791105154,0.21955902243938008,0.8788174487124096,0.2701963863035697,
 0.09539121224931968,0.24016345771660097,0.9026018392435998,
 0.8378544971046425,0.1474399820427289,0.7393313976330647,0.8915246302203145,
 0.7140848001936901,0.5394937223849976,0.15677445027432413,0.7046237279834298,
 0.7683453195454415,0.8106409818769414,0.734319045550725,0.10613001345991369,
 0.37181489435008597,0.7676217122834882,0.6515590580820555,0.4385278664162726,
 0.1726423072251685,0.7578661431473933,0.3959679968404184,0.9026316113489105,
 0.47788879374523463,0.16054489542951755,0.10293877547642438,
 0.7252148498946464,0.5924200492303624,0.7008744103289616,0.6232933269511325,
 0.05594530552741339,0.41101859704712207,0.5185776140369645,0.853356793133365,
 0.39644190065916407,0.5989776354870748,0.8709399261742473,0.7345980631283111,
 0.8473878286058095,0.7669073567494313,0.6846839798273087,0.47791582168240154,
 0.7811059877468176,0.021504220787307293,0.5307964301151089,
 0.1415024184731638,0.3169592938102993,0.24558103071285553,0.7934493323450574,
 0.16397470950345183,0.3149573284312446,0.29442573801022087,
 0.2051906400605662,0.8258509537546996,0.8918506166739617,0.9860429563598383,
 0.29925934217447947,0.2714222677137721,0.3637301827755294,
 0.18892444947257003,0.9037662865823396,0.8717521382625992,0.6855041215752367,
 0.3373242617397555,0.8191561883893796,0.2660932704768355,0.13304245996103614,
 0.9215724755848937,0.3349430015218814,0.819190889147599,0.8135182227060279,
 0.1526156955144602,0.7666814222082348,0.12894390071304151,0.5178333738610125,
 0.6659739848398213,0.5201575872727132,0.7246955986575345,0.22887790117717377,
 0.057826466854625336,0.06787668197124763,0.002163569797177911,
 0.30358384672513505,0.30092945885396316,0.6539337402325602,
 0.20874148541281645,0.9716658068649576,0.5571337590333245,
 0.35020644042949656,0.44227864300014985,0.031800913918758056,
 0.5784772269002562,0.35633442887815103,0.02647265330361137,
 0.7805346391316693,0.8750198024371898,0.9950539468588797,0.9631413343262162,
 0.24041622902760762,0.456895315004625,0.21316408663205577,0.4050990579398862,
 0.6655049975141375,0.16828141628778737,0.6101612741400891,0.2772905668781056,
 0.07043520454663887,0.7558031633278575,0.7115589720462027,0.6673430376982861,
 0.6065636584052967,0.6643235813614563,0.6610096645756505,0.2881482957027398,
 0.09422488290334341,0.7677453585868759,0.4044286046367348,0.2885586503473493,
 0.27610924710020035,0.4283444464387731,0.6198696526852566,
 0.07753578060996702,0.9350459335925225,0.03495792360449379,
 0.13014170031511463,0.5461208944393187,0.01843516737726314,
 0.7538497807112388,0.9424943428968862,0.2190058503544463,0.3954785783198024,
 0.14540741752588648,0.31179499464260285,0.8804498489542203,
 0.5548668814397116,0.33032709352827494,0.7520947747245085,0.740525520173919,
 0.5541220856329618,0.975209215387705,0.988723238413996,0.20157164028492836,
 0.004828822080195638,0.7864655772989431,0.6162332397055328,
 0.9204564489762306,0.042767327420931256,0.8271189551895706,
 0.3619297509941113,0.7869405297303187,0.5712792288766333,0.8881878290341132,
 0.060938489841091514,0.2528516115687862,0.27503041001983963,
 0.9123167032885515,0.5340654233426235,0.8193558220089305,0.9259679334044236,
 0.6568097051792944,0.891649706328364,0.029201411166575375,0.9904293266714523,
 0.5423087574236791,0.2564557365066007,0.6567343470299613,0.7046253370845575,
 0.3484285513224452,0.2711187677267948,0.6232814957442921,0.7911674460771628,
 0.1935113752913643,0.27432327669439527,0.714839661416512,0.42327168611476496,
 0.5660394570972898,0.4481406709945128,0.6276379583574669,0.727797279377618,
 0.2810271952080037,0.7119377278844597,0.08782092522274576,0.1559471198957969,
 0.6453103854196257,0.08410191623532137,0.12437594693195264,
 0.6501993175939991,0.7176983177031683,0.5767027556205951,0.6176472081937429,
 0.9132953330039038,0.49607280172165924,0.14346255390958795,
 0.5163742500304854,0.8836419912140872,0.915531691281487,0.7766062800667478,
 0.6806782813836625,0.4912905100266455,0.6546766116967483,0.2299134288614405,
 0.9069571919549233,0.774299315774423,0.7131473707027123,0.6485986577983726,
 0.7406897958787316,0.16903284996442847,0.06111797559858201,
 0.7187072273862024,0.4684822469530142,0.4944284833978292,0.27420492883013337,
 0.8177369647446837,0.31775323622204477,0.25724842740779874,
 0.5200001515341772,0.09472005500220304,0.22167490160392833,
 0.8507064217242385,0.2101161166617549,0.7517464028830337,0.6369725249080739,
 0.752310306967676,0.5325885439273915,0.25929785433499497,0.882859106254079,
 0.042857239123590185,0.7440516871870061,0.8397916690251036,
 0.9445698057351708,0.923854460643918,0.5142679991246408,0.5791294089832882,
 0.33701460313115184,0.4616348931405888,0.3775230436688458,0.5499943834126173,
 0.7594997060258367,0.3665827773041106,0.5117464307768643,0.46914800191787887,
 0.23022209187345288,0.7463182466548376,0.6914019486457552,
 0.29170702686570893,0.7636455975190358,0.7397489918974949,0.7424188774263036,
 0.6207392332721412,0.10703360340138701,0.906533453965387,0.980421210470907,
 0.014465837647475777,0.049190521760241857,0.47902859960586763,
 0.285846311491912,0.5969140109912697,0.9039001185941617,0.802511438295043,
 0.750125288205788,0.32609233324456266,0.3940412651528208,0.2455935013726167,
 0.5232089697612972,0.15118595252530453,0.2344070141036939,0.9145654136464624,
 0.16235917617648288,0.5049193562314068,0.12508749436896716,
 0.21712711003016716,0.3179246852930878,0.6151217529730937,
 0.07855150709168413,0.17910304730497684,0.9611176215002536,
 0.6926715476056692,0.5850410996800408,0.622850055912419,0.7106210921369093,
 0.09830953750867166,0.9163506721610278,0.07538327343783224,
 0.27391323507934506,0.5729890877555087,0.31496589647473616,
 0.07132084071306943,0.8709810647415509,0.7751985942389534,0.9452179411989339,
 0.500500705945503,0.36604400003170157,0.7431358421744791,0.7877582850933277,
 0.54784410802662,0.4042894506006325,0.2508113799471904,0.4512737444602064,
 0.4807269442701819,0.7321790606822262,0.5357738183597366,0.16343299987433213,
 0.631559971476904,0.2935987978336184,0.5864515825068113,0.9823809553516437,
 0.4241712580273429,0.702648036588251,0.19582604122452285,0.7698546896277442,
 0.04777296734707115,0.6103494347332701,0.49337921499352466,0.356277568463509,
 0.5469586122802549,0.02525757073967727,0.19884059702611878,
 0.08784824038913797,0.7510293162978925,0.06836368445255114,
 0.10843775383886378,0.697914067065153,0.45650420805657943,0.3695784300575983,
 0.847580353666352,0.7261567680698671,0.17478342516212697,0.5546464598365292,
 0.8198716133779717,0.3212831634752443,0.06766336567155173,0.8191954069339263,
 0.015299686088723119,0.4671485177334733,0.27339549375431293,
 0.5091281454055191,0.3516817530824242,0.7840994106542631,0.10144877417767217,
 0.722556237183203,0.4855333878796019,0.42088670209349643,0.6530958075948021,
 0.7528368651504455,0.8499213665343925,0.49843549110829743,0.9086157715738354,
 0.7874696517634817,0.502930937301981,0.7567971816172779,0.43872022378279507,
 0.15163283852502973,0.41875144997660807,0.8421546736512904,
 0.29990953734937076,0.35723720247617996,0.1880531605463508,
 0.0664680641873927,0.8932278117567037,0.9820774392574346,
 0.036583952607578296,0.969716201053858,0.6052200788227603,0.6702630162906598,
 0.8082468737993271,0.2706557088123649,0.2984860623750297,0.6415138581856611,
 0.6547516381244883,0.16756464007418026,0.8553524588439194,0.2793891146357509,
 0.2663703866337579,0.47083582189059947,0.7870828850815584,0.804352331157962,
 0.8851041994811097,0.8749285519974976,0.8927833104701983,0.9776339453664136,
 0.12010840782035204,0.20937347663591654,0.6157726848052734,
 0.9349819807075621,0.5882254906929288,0.36228130161505523,0.7940986905147095,
 0.6022690915231929,0.6843367303047024,0.13701935518842,0.930405815532692,
 0.6563136177171833,0.10780429884297238,0.9150121186255091,
 0.21656385124101718,0.8821699820419506,0.094632901511563,0.39959708449262943,
 0.6434820535692706,0.6580707108038455,0.18190767528991536,0.6470127307430563,
 0.021712161629642734,0.3809874632915302,0.03147843757686719,
 0.8902320544035599,0.6479050429864844,0.5927258529342857,0.2448827973872978,
 0.8113378357686665,0.6812073276210501,0.8661462560843693,0.9463714676754247,
 0.7106679282324015,0.6994698009891465,0.38416762539703875,0.5592236287089863,
 0.6819010346833304,0.7792803550124443,0.16868088744579124,0.400092250601642,
 0.03701207378295546,0.5338714000945248,0.8523511190685547,0.6362371487234076,
 0.22646686142860695,0.1368758533809752,0.037387115886088895,
 0.08930163373718802,0.2966371905862375,0.4035752038923278,0.5842789649736003,
 0.5379142613050493,0.5298266385741252,0.59621751190197,0.4285473562866502,
 0.7216241319913699,0.8053806118735768,0.6158687511373363,0.4401983761625946,
 0.09174931028531852,0.39781309303821966,0.7758714275734153,
 0.4334144125075434,0.2895770045004863,0.3513833193661954,
 0.039976686277276574,0.8932486716581493,0.29150898844759066,
 0.3534541479807367,0.9999975927034268,0.8385634298687211,0.29305288080191105,
 0.4441171879046504,0.7972146885862582,0.012225222075895048,
 0.49376814590066076,0.12318969310832562,0.40015923450214685,
 0.5977199999174508,0.9622985267925579,0.5446468878646579,0.6070686374457974,
 0.33577099213805983,0.123567069090327,0.017603057173295578,
 0.18351184117352115,0.31519243712574885,0.18459331425099001,
 0.6005756809091318,0.7141239536845693,0.7045939868606725,0.38367571204770146,
 0.24556788624006454,0.32072430716511857,0.4967262277220311,
 0.5792029384766466,0.162691317056754,0.37283956067526514,0.666677798974249,
 0.39176294442764714,0.6908997203565616,0.35751687765258877,
 0.3147021314327576,0.7587140383611933,0.9509444347939506,0.7625039860992524,
 0.3598400529313537,0.21043286441219233,0.1982159273054167,
 0.23261737105757196,0.768515061631876,0.7911682827722296,0.17716574547772557,
 0.558444755449856,0.42029947056384764,0.26010506493365104,
 0.42912908878785705,0.052552017230487524,0.555215065014474,0.917673209353923,
 0.2210753317465003,0.753860036476407,0.9502419957034518,0.6505412849556632,
 0.8376861806053291,0.6461042752405427,0.8743267204566803,0.659359717731892,
 0.8017584971316403,0.39896521536384943,0.35806954533610136,
 0.5356488106988166,0.09285603273834109,0.9821465141490326,0.684406342088686,
 0.07402646720181738,0.6327987656090444,0.7483762712932921,
 0.20142979530207217,0.8550185283246213,0.8926930690614981,0.7013729457422303,
 0.6583147103583283,0.8491299546164365,0.6151509612794693,0.23317312973608206,
 0.7771866152955517,0.8005249298766843,0.2440192264263985,0.7830308742851522,
 0.7246114793065884,0.1621705167529751,0.5744988954550841,0.5292936775286621,
 0.7126582866804286,0.9568471577988954,0.45928164726447174,0.3487726045521714,
 0.8079902608707737,0.6520883978397563,0.1435671271260991,0.5750944825879041,
 0.9069339353392791,0.008263130874416214,0.035592158796430295,
 0.5237931254864578,0.001923529182186634,0.47070866987528326,
 0.6218408856724238,0.39807184002071017,0.8946691523201697,
 0.49054657922290124,0.25151705977229777,0.42607017370807776,
 0.6687827674221345,0.8234541534800472,0.2760530107163438,0.8937711832803205,
 0.6549371020229646,0.7033876982535052,0.7824703805323943,0.3657642057672019,
 0.5983253290188456,0.6445117153556242,0.22491634586125453,0.7949097576657009,
 0.4475504155980675,0.40262893743526174,0.9873497332519215,
 0.10521929802033192,0.37315511114248734,0.46287934354368143,
 0.40341880405889396,0.8584498185859002,0.8434274042077532,0.619970891445042,
 0.6899712007487927,0.19815704190183425,0.04933357803105176,
 0.8954125113301721,0.44483184781116347,0.45956717612600684,0.836472007430626,
 0.4392053635478099,0.8971503653469315,0.013486489740305418,0.80130323721325,
 0.432581324406759,0.9747659707537804,0.6383955232715178,0.5225850131420477,
 0.5861058574613165,0.14482333308671214,0.33868873694837986,
 0.8833430200752098,0.2797034308132904,0.8861391465840428,0.8602202913385077,
 0.14943523634534217,0.2684218435752934,0.24126507221629834,
 0.5098294164506438,0.21697497634120455,0.7360914432053167,0.3647567454261653,
 0.05649480083262781,0.8516785097002362,0.7884961986590421,0.4223639618217916,
 0.05941973801230094,0.53937876918809,0.7558345857635781,0.7146426405317676,
 0.16627114559064782,0.06901377624888894,0.8550473798606388,
 0.2712840632384379,0.8785711364328357,0.2714804284553103,0.9734570957416265,
 0.22251100972167548,0.5476615939903575,0.6284820764841977,0.5297852433304018,
 0.0414110831894291,0.964146105925587,0.6256404333260632,0.19681716984733588,
 0.7451179770221172,0.38484267278393824,0.896803850205161,0.9243075952239463,
 0.06475298470560298,0.14664183191596503,0.08166585215454691,
 0.6721905223958871,0.05159725532318549,0.6895772752828606,0.6135498502914556,
 0.22197846172041125,0.5645375082798882,0.9121807295901521,
 0.07618979370427992,0.07055973711699737,0.9671658068985405,
 0.3096922988672385,0.059588631076818444,0.9749923438967244,
 0.49347516284477455,0.04504985019796526,0.11234934740199476,
 0.872502222522262,0.7499960394975129,0.08528642011798326,0.16082951152611202,
 0.10735324589879236,0.5629664819722143,0.8851164121492336,0.3839102288098091,
 0.21890062763939522,0.716053435718185,0.4798963308511984,0.9858991254822354,
 0.520060486227768,0.7239313523914453,0.9019610550790106,0.555331156752099,
 0.3084379149329709,0.718175020386158,0.6084217975489676,0.7005663463659942,
 0.792436772882614,0.9854140531804626,0.8745551534923727,0.4262291468114894,
 0.7581368598132884,0.5093196113878408,0.37589618593108354,0.7635529711614066,
 0.6149401429512352,0.08138742378000552,0.13246454787647388,
 0.11361824303883183,0.16112586747376223,0.3978411287845013,
 0.19349914785577793,0.9430930631222902,0.5297050238048631,
 0.08225583207731324,0.4500001322190944,0.8799426295490792,0.425039465601469,
 0.9805340979035522,0.3310976779885998,0.3176211032234868,0.43450559470003725,
 0.9831118812779848,0.8009120683747453,0.9184760547934044,0.9922429650071702,
 0.6469237462470843,0.6332979305397118,0.7797760398487148,0.5495940395995613,
 0.9423764061881404,0.6894302067353992,0.6560092497094767,0.0660982757014199,
 0.40091209021951146,0.46382350930364535,0.16756351390985535,
 6.707968375716522e-4,0.042019617359736916,0.1552162163459364,
 0.5533243440367543,0.24143608661016502,0.025887761496619865,
 0.8723274948161076,0.14516697793537248,0.21811707088964172,0.787707083420585,
 0.4199314243503507,0.05833572571052992,0.7184825930091844,
 0.06751184672196453,0.28479765162028625,0.934340571885488,0.7373356010518277,
 0.6508107131501496,0.6670197772154045,0.9768959304718867,0.969709898356699,
 0.2249786088532214,0.43627696542832306,0.8585853978433351,0.691125423590707,
 0.6300882996511932,0.5216050032476679,0.2617887871866327,0.30792750453237794,
 0.9045587558343746,0.08022914594216413,0.2681054068811328,
 0.27104561905727864,0.9256699861155299,0.4260833011641061,0.4410418816434002,
 0.9731786576830057,0.38623575985616676,0.283299429389668,0.8555366622657998,
 0.6215519646701553,0.505730061062913,0.596486505193852,0.12636778007967586,
 0.2571715057153885,0.31366518396629883,0.9061460653996363,0.2009616253965617,
 0.34717651461801713,0.8736461273249498,0.515392688113091,0.5310927566751507,
 0.1269039233227912,0.02748309193674181,0.1247818033004342,
 0.052431779601222916,0.2806307218740858,0.10119020883388519,
 0.43630172082652086,0.791902978369383,0.22098434030748448,
 0.43100493263444795,0.5040177402328824,0.9916693490283096,
 0.09250769302888207,0.872239599109478,0.3125425694611369,0.958508209490536,
 0.566634052052789,0.21334013636228233,0.5555637838188545,0.7801177698223957,
 0.7068750869591164,0.7286599740942715,0.386350293996957,0.7487133467781637,
 0.8004921625867024,0.793873185933611,0.41033487335099406,0.7114375085789445,
 0.3067627435072702,0.13337319999850306,0.5525694425535768,0.2709514431025414,
 0.7132280438696811,0.660730468534211,0.2613520640583086,0.26781070447094424,
 0.8759638184304931,0.00854412256411241,0.7311739467273144,
 0.19506380079925012,0.4477944856845488,0.3634588778779825,0.6323783981759743,
 0.5283002237941578,0.8193114775676263,0.2892235023762163,0.57208556136103,
 0.48911470307504956,0.2557276021618794,0.5741037622113816,
 0.18911961040718117,0.05189873015465418,0.8469452367101402,
 0.20177192099339802,0.24387058229900693,0.6366033791716001,
 0.3476670495253179,0.5099161121088415,0.7670129075566089,0.836475450057487,
 0.8542333424243882,0.41219418601493496,0.5072422893127131,0.6240881474295562,
 0.14775699003359521,0.546084633247025,0.16979658360340844,
 0.23836943217187445,0.8663252840594584,0.40543739456619865,
 0.05223155704688842,0.8203634263620144,0.24470010122385843,
 0.5619977003846692,0.59530225618451,0.9552299808110828,0.12700667100377505,
 0.9423094624112772,0.8253435698465946,0.7409558387425939,0.118289982839341,
 0.43775036087294183,0.8529928077732511,0.37241666397751116,
 0.6403958205753671,0.8996140985102568,0.054886311612387106,
 0.14127133376396372,0.4069136933276185,0.6077529412095137,
 0.07517486903711568,0.189052653415279,0.5260012251650882,0.5841530527680966,
 0.7060168230766666,0.22591431543620955,0.1519763952386164,0.8150570931394147,
 0.5238617117177414,0.9199711003674619,0.5447717637578818,0.9109614518487477,
 0.9792458572553734,0.31518781605183777,0.5269113849191283,
 0.23563746630317195,0.24389164351125203,0.6762514352744681,
 0.8233502601091384,0.9203179980707252,0.13855821288525738,0.8846546197522733,
 0.6183507155254736,0.29688750034386224,0.39571255133166394,
 0.08188422183645128,0.3050411202271256,0.6106642079212782,0.9619559586563174,
 0.6830555561835083,0.1384819809238358,0.03224798418489827,0.3942409991343282,
 0.31170234522190254,0.9157390065439349,0.25141999331061216,
 0.0716918317039763,0.7576198783822622,0.9366033071951162,0.6202991970700932,
 0.9470792356757555,0.34650097088968923,0.22040474631405704,
 0.6693692449547505,0.9372497895021727,0.9341324157377778,0.6041302758263846,
 0.09811203257332568,0.09628659517446092,0.776541587611585,0.6268684490695418,
 0.7770892807448884,0.5099475329446084,0.5464821888583149,0.7795027164932679,
 0.40191003224084343,0.5303999922548818,0.09138779354443582,
 0.9499745350679522,0.6583634557409759,0.3624338185100362,0.7682264053269312,
 0.6988962903392439,0.6620109597545695,0.2197460178571552,0.7015225805823938,
 0.5106543675409192,0.2485803005673506,0.9165411428374093,0.7265410367247909,
 0.11833327904606683,0.3207145383284724,0.07523406411246603,
 0.8449670247633425,0.3015340303158497,0.6902446980063829,0.294640324530523,
 0.16703836838031705,0.8499372268130498,0.06382399769400182,
 0.3306268000779907,0.649003351718717,0.24219605981336312,0.04683891247474137,
 0.7578994097420405,0.8639015519252845,0.13462600205739494,
 0.07669563916334043,0.7343560344126171,0.981592704440606,0.33383624524730393,
 0.3022357668101129,0.058968138535422265,0.6490634030980197,
 0.46818754903507054,0.6293124497928236,0.023011221426245676,
 0.2790550481632442,0.40489277376013055,0.9277628469280131,0.4112095330478902,
 0.9118787103652953,0.10257370634823293,0.9444008457740662,0.3327954443871981,
 0.41030849812636383,0.42545701475971043,0.3832457695345656,
 0.1164086399734845,0.47769908090136526,0.6503714084740997,0.9167356187647719,
 0.9931151393750465,0.21926711439913926,0.8512881324425178,
 0.30599469281874037,0.7191313234552787,0.474979158875929,0.3747333643852039,
 0.968165737171985,0.2864202321384022,0.624029203493075,0.6812053567561813,
 0.1689514349236575,0.5632814311974614,0.1758147067512036,0.7570269817893974,
 0.4752900899394724,0.699054709317273,0.16801268698632477,0.3150509005316712,
 0.21090872353518075,0.9740702676876632,0.1636950627582615,0.6227483628392623,
 0.11934897426658164,0.9071496018879434,0.7900737670381572,0.8350411229706013,
 0.3852926747613872,0.7534524487792036,0.8879393081255982,0.8514327115364122,
 0.9441760078126665,0.2933963989782816,0.39153670469688784,0.1666432380043945,
 0.3898161543407306,0.89795676183879,0.09108492470241047,0.9531548929795809,
 0.9624832251473883,0.8980999881586427,0.4665335050063697,0.978620394715683,
 0.26329238612505934,0.936574236360555,0.9584913314975263,0.5725214301378565,
 0.5155029523438137,0.06899331605684011,0.7687502385417327,
 0.11491388544966519,0.9874272291811885,0.20556465111305688,
 0.6797153654063682,0.19622403208877692,0.6099913900791176,0.2802900489741157,
 0.2186402044470268,0.10049347789939944,0.07749167381465072,
 0.35481913294095657,0.9002689888896983,0.8268941821575437,0.4940286976457472,
 0.12738715224419994,0.2435097941297979,0.7748783830115136,0.5466458291393816,
 0.7446125271164149,0.8743853279420964,0.020688976701303874,
 0.08876080465510117,0.6337714770740954,0.7215829543266823,0.7774010721024038,
 0.12306614660987591,0.8939601101402098,0.8751636914961275,0.1861050064872416,
 0.12321238523669031,0.861659508704167,0.8233602903648762,0.00500378940588353,
 0.43274663958805004,0.45923179980665974,0.3696172273544993,0.99307894466587,
 0.3419161725942659,0.08949001173125148,0.11141224271161276,
 0.6029317858219777,0.034784975997659506,0.6181234642773825,
 0.7706567775320876,0.28914930004253436,0.7582444379676249,0.2526290545622746,
 0.1795801511886136,0.40070139851226905,0.36525829845479096,
 0.06784081389120622,0.791742834943878,0.030603939997002172,
 0.19705690463921777,0.3402325780182137,0.45168232535093034,
 0.8357278044751515,0.1998300047122632,0.03428840362354946,0.3711111707400483,
 0.07010178981972437,0.3562299630844853,0.1295637849015029,
 0.19824284377051848,0.7588571759837379,0.6148018235487503,
 0.39435720077977365,0.3320097867952849,0.9151923001260946,0.9931756861131799,
 0.471078460425134,0.9820497042768187,0.7670108193519422,0.6868788866723233,
 0.030372457140188436,0.6279404743561363,0.6120555651846213,0.707431656014932,
 0.8667630214850127,0.0928141508680862,0.6026361558622487,0.10030926364328174,
 0.8727501439359712,0.7846690757284157,0.4766146413187078,0.48837090810938133,
 0.6885170722256582,0.4564083169525157,0.4649822328651214,0.6924136591447373,
 0.3412367694102886,0.5055876111858036,0.49276844525293295,0.6839789266175997,
 0.010614171618068213,0.9912574039294125,0.3576463405271998,
 0.5319365685717693,0.022823943494107546,0.8714712087919816,
 0.6399618130211355,0.7201410879104103,0.9208783590607011,0.23619888320794313,
 0.4091939442235172,0.8734123521294636,0.32715682871495977,0.7530481306756711,
 0.19076207910931853,0.34443127049326394,0.2808035823580153,
 0.20075558070240518,0.11192817378971653,0.27839328787274675,
 0.8922131844759936,0.3845848071645347,0.33586902558311693,0.7558025495184905,
 0.992593415935748,0.5208336809315651,0.658332255433031,0.21891764062804542,
 0.6423588107415675,0.05018282937449148,0.9675931229501138,0.6730123543175837,
 0.11369793125613625,0.8891382078373564,0.5910505295948087,0.6536083447025895,
 0.04689730261694658,0.5672130394779995,0.3095779362669996,0.6251302962030956,
 0.8870350343937073,0.9411232575845803,0.16169068281714316,0.7874717811039826,
 0.8360316354061496,0.4679973207992284,0.1845161602695773,0.4253749352850902,
 0.38846642587293945,0.43105363647588035,0.4381235071709353,
 0.2112652370610676,0.05014255717202787,0.4004956518345768,
 0.03406034078253062,0.9268604959863157,0.14488383617504685,
 0.9228393591770288,0.9065583803392816,0.8420270196526625,0.9573579602640738,
 0.5258342449519525,0.7954480884075937,0.43722283017210883,
 0.31589224684886297,0.7022797944857686,0.9291021907804541,0.2324572781376837,
 0.39180220691015033,0.033530056996730284,0.9984800074444529,
 0.49718704573986805,0.703636102036655,0.7979461939740328,0.8301009543059605,
 0.2040389482412054,0.32673466464981094,0.9362190370172097,0.382237713073164,
 0.24293347170823332,0.1263411865852586,0.9836944298115516,0.9625439117010469,
 0.6701020855370083,0.02461872737456705,0.9288861257390839,0.7015261150380687,
 0.10145236810445912,0.4630495501915568,0.6647691391804367,0.7809035048746578,
 0.5332422342615086,0.2881256094509246,0.33650938245011386,0.9924151649080409,
 0.4817603109009654,0.8691655463655961,0.3480545720279309,0.44518890821216717,
 0.4245327545317845,0.30737914113524833,0.03583636488024,0.602056876603684,
 0.39237284591571286,0.76776952254155,0.8807945397670613,0.7335118700899407,
 0.13928857710565923,0.8359341371200311,0.07218720520360533,
 0.08425199063976896,0.6573147971504026,0.8011266244182308,0.5401621397757513,
 0.6516658496349317,0.773848106289563,0.9667928912730996,0.0638270733916193,
 0.16234315241119968,0.143409429189851,0.6495063338341714,0.4102089509139941,
 0.5417262795171571,0.08480800538604127,0.7465017919016546,0.4671380401854205,
 0.07753940019692307,0.7895826116069121,0.758917736825727,0.4495596026948969,
 0.21034717675857573,0.6607159444511782,0.9380217454191708,0.6363273930030781,
 0.35137216739897503,0.8190381088952329,0.34541742593549873,
 0.7023791726326822,0.837046058652979,0.9845918371840995,0.6177417630709598,
 0.8790960970162792,0.4165587715496988,0.9914421804941164,0.8699244013652057,
 0.7096846565541902,0.7410193477683549,0.5807809241307481,0.2600058363736847,
 0.6171892594352909,0.34816002534641144,0.5158120441843641,
 0.39193631912488347,0.18753264598836739,0.7481520095594034,
 0.16294357775959778,0.7427358130884922,0.47568973505731504,
 0.5713590400628915,0.12164904302545831,0.1894860631897689,0.7992362541106331,
 0.3574318877233382,0.717693416885772,0.516618321941214,0.5194173466045292,
 0.4029974202907264,0.5293862605201667,0.684833632876182,0.8422816779389113,
 0.8939836257697877,0.4163623642904344,0.5542632673033507,0.03276138765020287,
 0.09809898827658137,0.8867924656765094,0.6789928499214721,
 0.36314762546866786,0.4783421780898792,0.6882313605191746,
 0.37938269909787015,0.051040698746979696,0.25248381662526087,
 0.2979563048955226,0.3255801465937076,0.06324489729652549,0.8113231368487759,
 0.1198296220001418,0.8888640696730691,0.3904691494544483,0.34844150746132385,
 0.008828728918695794,0.11683266570970718,0.8388374217697165,
 0.6232791618401832,0.7247440615620339,0.4121948334495469,0.23807447932808168,
 0.5770247027029938,0.6691559131609859,0.7231110735057629,0.44499595074252407,
 0.29591217678902004,0.26836775107321365,0.5702188924551549,0.949664544665336,
 0.1645101223912413,0.9128929835429839,0.7611623443652685,0.2724683054535273,
 0.776990859677655,0.16667122468023843,0.3180226422649498,0.9462083280955255,
 0.03487953399866239,0.2333318541987197,0.5404212239624288,
 0.12749832659681493,0.4986644324009648,0.41151295851425984,
 0.8673194106620501,0.20700288866609207,0.27683264864718593,
 0.6466370158497553,0.4534179040241709,0.8252345148660469,0.15657096641821355,
 0.5360359078176218,0.9297587895349176,0.45979655002430875,0.909652793919458,
 0.9925328308687243,0.47955148054543983,0.8852786310873629,0.6495514866507881,
 0.39093046421807054,0.8029710693106185,0.9834650766559438,
 0.49541021966072063,0.8275367173607034,0.2406536651629988,0.8627047803655112,
 0.6315430399371138,0.6844633415756076,0.6522535102271608,0.8726435743863158,
 0.16461102664090654,0.4219161399589262,0.4669129236055084,0.6710582078089671,
 0.8023028108115637,0.8324607868942261,0.6173048660371443,0.697979174188887,
 0.5178588824919714,0.22371158718129702,0.04281383773121106,
 0.16410059776605546,0.8950531508419569,0.15762019129323968,
 0.5706299498836088,0.9446646374997621,0.7433235002104932,0.2073095153403992,
 0.17325912182959358,0.28156503741056915,0.31250168430921366,
 0.5313167186760822,0.27266996149367684,0.8437115491809077,0.9514919354384184,
 0.6892439144324352,0.47144760441189115,0.9271154914771159,0.5508623980887739,
 0.5939770711431547,0.1957196669912391,0.17407136539284962,0.9881438494181929,
 0.43318020897519227,0.9615187156333892,0.1091544530377686,0.2394471228214895,
 0.24412935779900302,0.22116560804199437,0.49828331253795155,
 0.3800246511376588,0.9992342527600223,0.8272791249814061,0.14849645821469515,
 0.5232833418899184,0.9274169717410373,0.30551507716031123,
 0.22240783983136747,0.8559441659941232,0.08082749152341528,
 0.6165270111382835,0.3355745372954475,0.6309654625006305,0.15072891918123799,
 0.6422220943799218,0.025451125217705428,0.8493149794596208,
 0.26747480815572866,0.12946894203745418,0.8269557638899121,
 0.8326776416150254,0.4984575186709761,0.4963991814958133,0.7039608069493395,
 0.9194882752674762,0.20011501597066306,0.4916376520156791,0.9237531733137667,
 0.4172268683353998,0.9396429858594221,0.8193964030335432,0.662359348187575,
 0.8172341302033048,0.3886235177928685,0.9628640860402369,0.9189545245514602,
 0.5162971158878681,0.49469716967789534,0.8057354455548327,0.7214856182114773,
 0.719928839689677,0.9637609331378033,0.2547651170112679,0.891193453359003,
 0.2065257468854348,0.0832505414803495,0.26842198351886026,0.8965797608186767,
 0.8147223123376478,0.19802003460233097,0.36865118511836703,
 0.5741823737684558,0.8329406073869196,0.22168319232481914,0.3715127750318561,
 0.643212419821748,0.10472669606335117,0.4498724028347738,0.7455215255722468,
 0.3258423311164528,0.7332909028891361,0.1543493616994256,0.8695544633110532,
 0.9640473694293757,0.6961059737822644,0.9072903905107398,0.4545475360363561,
 0.918523573143776,0.6402269321810194,0.7246742314258441,0.3406408923776063,
 0.3011119247144619,0.5764385774788178,0.7141215362154758,0.8502896476872794,
 0.6759281249333476,0.6601604624836858,0.6982754194587736,0.3100635882142637,
 0.8005125726129825,0.15444146067593845,0.20536391051634417,
 0.0025500901920835206,0.2102850340895518,0.6593724598757515,
 0.597722693408097,0.5034235491245306,0.432230680569747,0.08540892535787115,
 0.25989637535902865,0.5155177216517954,0.16084473496464313,
 0.07038816001249071,0.1716009128837993,0.4065881650547978,0.8518472842300903,
 0.5107439355713347,0.5193088733480584,0.32833440759553245,0.6431300714392131,
 0.4387293303368198,0.14191355392331495,0.8244089874710863,0.4022122954710238,
 0.685175179106849,0.6316455216729112,0.5326685109744597,0.3173780367211131,
 0.2998693262637917,0.5571843607897282,0.12246027000524651,0.1976173034900609,
 0.022558283656599065,0.7721131675869869,0.4897915558763958,
 0.5580486361918368,0.7018054745627716,0.16746054806134603,0.4118847554704449,
 0.9076103236377269,0.16159279077975763,0.4199625432211753,
 0.05448604982217109,0.6329373473068406,0.39202648472951207,
 0.2726123441028474,0.36113426131276694,0.4274095690868226,0.8606857617226137,
 0.5396148105572179,0.4194896741036638,0.5585359851782342,
 0.011282466315508888,0.4247645817599822,0.2787306446825657,
 0.7667934520079954,0.4526431914959388,0.14794383876964712,0.6131209796710857,
 0.1510997225722488,0.4340547264397353,0.6999224613774777,0.7371373345985684,
 0.5277784842581719,0.6086730950750294,0.5719976430428231,0.3828539569690528,
 0.34036225624995486,0.43120018166130003,0.9635499859131432,
 0.5866719198015045,0.26468752617481894,0.15644462602157316,
 0.1306398204243634,0.652653393589713,0.11943676552228877,0.8022370907130317,
 0.29041341991682856,0.47998517690815756,0.11716797152322467,
 0.3835907344085132,0.3774819207957546,0.10445940309868917,0.6329869741710946,
 0.6986338444932543,0.5202335335445897,0.1993539533327271,0.7889985159843919,
 0.5668375547756624,0.9503719533271967,0.7793871649982974,0.28190341824061727,
 0.0899585685580403,0.881547946660689,0.9179594287964349,0.7786082133166103,
 0.506173965863407,0.8784108927736627,0.47415355092072453,0.6154525486406172,
 0.5057958754867045,0.1378599735228767,0.4013066673299853,0.29454556042629765,
 0.4156944719077682,0.6722886308636187,0.014947519474527882,
 0.8119232709674942,0.10871947095881451,0.9625571043042885,0.9187043718505274,
 0.9181371248416255,0.6531064012864692,0.9149785990671456,0.762619280169671,
 0.35583731116966066,0.18164762614970253,0.2404446838916544,
 0.6146560022723235,0.7959767826099282,0.8235420743211049,0.6077845319358783,
 0.5689389330577849,0.6789014679945553,0.8061039584950452,0.6050146626153483,
 0.17565218632847257,0.8858365182098027,0.06493352100449579,
 0.47674327808054173,0.030703450943314725,0.5198465361743505,
 0.8983365315874031,0.6212952931085274,0.8773744908010412,0.4881434097958933,
 0.4991575289998753,0.16424610466951983,0.17790724929945034,
 0.8140661558217588,0.9648115671503407,0.7078421880298854,0.5024871902584451,
 0.9432963413756036,0.914436524611811,0.30999331161396193,0.8583153414914563,
 0.676188411463476,0.4440856245719791,0.039231314404091,0.5528617411595195,
 0.5564883390993115,0.48616819846336534,0.8967386690059782,
 0.18105292306083776,0.9384981030107014,0.23734585100067074,
 4.765586657781151e-4,0.8964277823604541,0.39940204056221573,
 0.6465637222351703,0.6506204943109939,0.19259302507215292,0.8584444067206884,
 0.9888834095250586,0.179528439270226,0.2638745212940997,0.2921973827885128,
 0.20012750870539198,0.044934424566842934,0.5853426901052909,
 0.4368688262522672,0.49416436093135363,0.5541698159322498,0.6436482253059841,
 0.8961022767018562,0.31978615614533357,0.5198258053167938,0.9749295710719033,
 0.6020070565009814,0.8965507702309515,0.8226344054110438,0.7508353704381754,
 0.30844321178015033,0.22739951556722826,0.36764029622396066,
 0.21264720571040818,0.19981968531406136,0.4262205591900262,
 0.07231755656673067,0.8799622198489747,0.22872580445551438,0.810430895585454,
 0.46096405460218004,0.11712148881027273,0.8362654979177168,
 0.3099012091313166,0.23518068789050484,0.21284174468728656,
 0.17273683527514216,0.4720263301485794,0.10093677537105417,
 0.07532426325538166,0.013984137199083868,0.054750316900901486,
 0.42506269230081006,0.039368914405997324,0.6054259862343571,
 0.6549921485043562,0.877112639064141,0.6273740739606241,0.9599760832589413,
 0.2928968906417875,0.3682439940447807,0.6000396832856417,0.08081737044948578,
 0.14482802104381565,0.8673946381895269,0.8990488451241943,
 0.03560430159331385,0.4541612513424683,0.5972164004536489,
 0.13471617362805566,0.6716118412718615,0.970010502437238,0.6191238475495082,
 0.24887387691127805,0.02871968905447908,0.8603430762938837,
 0.6795539795896479,0.5269459133618077,0.4771070801097502,0.21619908839551583,
 0.007859686826225065,0.012972559959418817,0.06089423948922035,
 0.7389976120109504,0.5201976413188512,0.0396866319397815,0.8041446605059429,
 0.12326660994635319,0.5020975744302194,0.8471783227688582,0.7737193043059739,
 0.7216644839400282,0.3268598321027253,0.04207706611487494,
 0.34580493348477037,0.30933957519565336,0.41350806933785356,
 0.9924917239183879,0.4502254391250551,0.3302774498537502,0.6759677669345809,
 0.30921379488810485,0.35015573487167206,0.616229806088624,
 0.04880234107787518,0.8610250738283903,0.7392938194305185,0.6450536140588355,
 0.026717850145429045,0.5883049785984051,0.741119100686675,
 0.32990471881422767,0.5315574606742833,0.03101646253625734,
 0.7089240324003184,0.2527820311131692,0.6657749278904357,0.6666050472136509,
 0.19025262974499513,0.99404746760836,0.5184007276582596,0.7740883513393058,
 0.49654208723792537,0.21836071935961265,0.6621271151512615,
 0.7956301452758973,0.3928232965606515,0.08117830067028842,0.4365969498807003,
 0.9706823963688873,0.5582280736730288,0.8491933143623986,0.31615909982978885,
 0.28572066424558196,0.5267267276550469,0.05350377525281513,0.714625884628676,
 0.07203949488905215,0.9835909556739615,0.33556600944537196,
 0.8031453015694252,0.48592202573111576,0.7845744512901982,0.2179558625124829,
 0.5713455568742397,0.49994608138877716,0.006651448922097236,
 0.1265547515970864,0.3856392973571888,0.8762515619632106,0.7189123688437192,
 0.954726830265492,0.03891572882474015,0.925926184997939,0.703558195122531,
 0.15134626952429675,0.38637420960738655,0.8046297323230898,0.691776944507075,
 0.03121757870390951,0.7506351106561201,0.020809010197785804,
 0.8757744583375594,0.7386347895537674,0.9887433464485132,0.6485757966844051,
 0.05666934879844843,0.2646527470248763,0.5116413796635313,
 0.34478724773664227,0.9625845829407438,0.1258459636795699,
 0.36074450829523885,0.2843312941674788,0.6722976934189164,0.1432503790834717,
 0.012655484439948572,0.27863033498573775,0.8283337853215981,
 0.2759324640015256,0.7323177454587538,0.5358688128085201,0.345425511163334,
 0.3317333038702439,0.9400448226766212,0.3011620182246709,0.5078460302328028,
 0.7807397161399964,0.7616021131367934,0.3882640586152438,0.2703588767712144,
 0.021511374803844574,0.3266861078529928,0.5293661985370722,
 0.02755745200135562,0.4706803835019674,0.6079643309904669,0.9220640783418482,
 0.35240269597692775,0.3823358533185176,0.23991494276146264,
 0.30273854261233246,0.02465558046186811,0.4338373087969203,
 0.5263296537840264,0.5914570945511501,0.9050080577047316,0.400411107968635,
 0.5203538040541537,0.5193554429248902,0.772270651165237,0.22281078220484463,
 0.17331946052974634,0.7616137000693427,0.16911600810076888,
 0.8714293735867986,0.3388012960461312,0.2037028698404708,0.823321944775831,
 0.4977796106681721,0.6989415179786537,0.08995522332136385,0.4456371267161292,
 0.18881855517322588,0.8956343671344469,0.018568651451743468,
 0.19846151604385343,0.6731065970382563,0.7455057027682733,0.9062853937777472,
 0.3297861090100641,0.1758393061869844,0.4452682197424528,0.7203279085693018,
 0.4215208858349041,0.36496420051743894,0.8995887863254456,
 0.14646307621447663,0.8156670966997805,0.31824230454640945,
 0.9802860598581374,0.5633117173492977,0.18545571890678658,
 0.014435572542240038,0.45546415862603373,0.21716558222100202,
 0.5831624553672797,0.17984393188413839,0.5473192602036593,0.8807536625123527,
 0.5968317807785738,0.45095500912292974,0.23530791609309176,
 0.5334155297373816,0.43505562252854457,0.6120193534826899,0.0705230844005883,
 0.8410212968525372,0.33165534699551646,0.005370026080401047,
 0.999711852045605,0.003307697468978077,0.8226625896059576,0.2613151153358383,
 0.8608193395499452,0.3439588095026028,0.36850053662660487,
 0.20241394814120084,0.3624478025756247,0.584647646787415,0.21897296899650975,
 0.6038252749867677,0.4074755846295891,0.8793204861828984,0.866779797530964,
 0.7265116149374905,0.8415596415714388,0.24197694616162124,0.5788080214575271,
 0.7085733147304762,0.01468057657163957,0.5674009187703345,0.9427682843582511,
 0.5747470497991648,0.10006194487482145,0.8610097218258184,
 0.18546475037533283,0.9987628914038454,0.9478005395569394,0.4117065876048538,
 0.8788992881943478,0.11048965642177588,0.36528510701357464,
 0.5294082356195893,0.027475836514818996,0.5177688950733245,
 0.48716512814544544,0.7269654988941472,0.3872214912693325,
 0.39226919632135904,0.8101772066878918,0.18372072186605104,0.701634188056285,
 0.5056781098820615,0.9493286362072916,0.030607433338375567,
 0.5846610379489053,0.27816316193764223,0.08546683456937332,
 0.5542211277470004,0.7725423679127551,0.09370631857060363,0.8374959842412744,
 0.9558369301245753,0.190056087325417,0.6363203962731421,0.05000004884713216,
 0.1814880202343192,0.6554840209484719,0.7361707111509898,0.38357106718850886,
 0.6324307726445445,0.9314266672939199,0.7302249499707268,0.7533957926377843,
 0.7000625678792981,0.9044908658321699,0.6370347962402731,0.26157705234403583,
 0.15191911625296295,0.27053922898126015,0.5646562067907777,
 0.8802619542931435,0.41402529314494463,0.869193517791458,
 0.017526924283621903,0.40123806744431956,0.6491901575440312,
 0.10156863705648511,0.2664010761390837,0.1259597295936976,0.5114886936854002,
 0.68769527108373,0.9192212081255887,0.9817137953613132,0.971910744703295,
 0.2665219355371292,0.2971887651918458,0.7586130284146242,0.8214989925703006,
 0.6512081918357326,0.6060322899237605,0.17626468879291268,0.6083349968910546,
 0.7591235403959723,0.678255555689295,0.4649227643730821,0.6436512496056266,
 0.21125385461331092,0.12910261088801345,2.7557900906083344e-4,
 0.6860311500938537,0.9899151831849878,0.46368031983068425,0.6405513771913022,
 0.7302329776177029,0.20653678499655337,0.18224902816926503,
 0.8023601334037934,0.25966518966086616,0.23070773153408985,
 0.04443483150545724,0.8081951780085648,0.5529433222139404,
 0.06809458898348997,0.145421598645457,0.7827440901342091,0.6537288466250311,
 0.7901280502243389,0.42542557315185825,0.06613233389347128,0.863661251208745,
 0.2848605549549368,0.28077187882278276,0.8796531606785818,0.3902382581871392,
 0.6156776721071094,0.48911251105604636,0.44571126978477216,
 0.9971179012088675,0.14739472001796727,0.04005653013704935,
 0.8182157599951523,0.9613787272698261,0.1779101287965057,0.7310993137138709,
 0.9579390430033906,0.6800062983872205,0.9176882885417895,0.5372378896397514,
 0.8351581738581151,0.0334996528940299,0.09294373176669335,
 0.17877238559906194,0.47296416860376655,0.5969738881233101,
 0.9315908241397998,0.14364352636324884,0.608162804154532,0.10789083578411707,
 0.5952164754628512,0.06746762478236201,0.8857270066059222,0.6305744297296293,
 0.9915824515188411,0.40584208315183345,0.17796632763083453,
 0.7103965391465576,0.45552652974122176,0.8589781437638115,0.4010354314153751,
 0.20245605480919682,0.28123577831975943,0.5317772598072097,
 0.8098477632798737,0.26190259839932384,0.15400169637380845,
 0.3476353184226567,0.7612400993826376,0.6125506036008483,
 0.0019964274135244597,0.6337709369951149,0.9863390815498454,
 0.9894225865353563,0.7452189986080463,0.9659206880544484,0.2524380604255032,
 0.2567958448375687,0.8309877875515717,0.3298239785270731,0.8158968907827632,
 0.5945844661238598,0.13533410936067047,0.31869072309880975,
 0.5196555311234609,0.40893823178566313,0.0724119131344173,0.9764742205220373,
 0.03294995968206671,0.25717454327560696,0.08057772621920667,
 0.15033797578866592,0.8193122426559951,0.8298206221517428,
 0.029229560024251033,0.34985797637608007,0.8753757026482736,
 0.8494650279845369,0.14376883536837637,0.7533110642051619,0.8434104830410004,
 0.9276653902073775,0.14116659129038012,0.9749304324050342,0.1278914976776271,
 0.7081133193596072,0.09644020320346103,0.12160377988041926,
 0.5141768675922912,0.29210916672085885,0.20290509074163998,
 0.2927873498872603,0.6481482669160585,0.9724287482061389,0.09146332640488342,
 0.2293475695538868,0.6317148537166031,0.578989887906443,0.1501783353436098,
 0.06923536335830416,0.306746857481885,0.39240003799757084,
 0.33944376120388386,0.2045752088979783,0.5672983372763372,
 0.15659508743136086,0.10057553453281132,0.2744147623508686,
 0.5725486558093794,0.593076218754982,0.7774876001750601,0.6581821725159842,
 0.45482399024448816,0.0492042133427526,0.41902621442113797,0.439865737976898,
 0.7629232885297013,0.13412796941601368,0.7679277969805511,0.1979092847455437,
 0.14997148525267368,0.4735594921129389,0.6163321289584742,0.5706171297092651,
 0.08056419862775921,0.972350648480555,0.7643397036162449,0.7629220014927989,
 0.3037238617429471,0.5306715836412735,0.98281279652238,0.14129796130742034,
 0.3678215480174418,0.46343730998454014,0.8744830309645139,0.8500749928333198,
 0.2847545961927218,0.048486111230626516,0.9611500250159238,
 0.06885147382181955,0.5787883765913264,0.6751756708646959,
 0.051487090509208966,0.35983677931891356,0.08728444429049242,
 0.22883604961010762,0.553679962508033,0.5822079535220919,0.7873888277229275,
 0.5380883767449068,0.9425191589098648,0.7230315566852588,0.4360948731131221,
 0.6696345529966464,0.4433838940261414,0.6328768582743384,0.8042549743784269,
 0.9644659639615003,0.18338810041231424,0.9103063230964477,0.5417437548734616,
 0.2750314017889224,0.8720912378269265,0.5127600110145617,0.9800715814316336,
 0.16877329708186584,0.5327860100529765,0.988202960967371,0.8468899638983212,
 0.11156164158186777,0.5667226977083442,0.7575282232552383,0.4324700414511773,
 0.7799740605642018,0.9335867915406107,0.1369100293277964,0.47697278953597055,
 0.04793871004220318,0.3389739188170885,0.6203192278005476,0.7014230524630466,
 0.8352585332384552,0.617151830698586,0.8684956869380085,0.7596402217069165,
 0.9358313459124817,0.6022513090735767,0.5198800422286759,0.16880432789072541,
 0.7324551317964464,0.23265052414096798,0.09836791679198076,
 0.7299608949258701,0.012354915864791427,0.02775459326526264,
 0.22161670097851005,0.10762749021098905,0.5575720052957224,
 0.6759444670138348,0.4877082051657572,0.8038100843996372,0.3728746380719339,
 0.8848523767957097,0.2197208245811788,0.43465102438827685,
 0.34791538069085404,0.8440308320195894,0.5812299321987326,0.5941531585212605,
 0.21787246995830567,0.30461824742939103,0.2646703207329626,
 0.04326070002486604,0.11284331825893901,0.4737447160457049,
 0.7995547008411753,0.41668496454273396,0.519629046630677,0.9245120681736895,
 0.19647405982319255,0.02787096190891236,0.734093550476595,0.2830675718870447,
 0.07771155709886646,0.4027863707926478,0.18511362091867334,0.555727511948669,
 0.8805742993542216,0.2696799491390919,0.9345353416156795,
 0.062252071275953336,0.4055443920772349,0.831970216271142,
 0.16587268791377796,0.45960810410683084,0.08049850358198019,
 0.01830572715577139,0.10829717333624767,0.1948853330677479,0.841766268362112,
 0.595616455502471,0.6673009847712902,0.24638995427559696,0.1298424423303267,
 0.8109794565537811,0.4703413251764016,0.22347987475571385,0.7477364088583545,
 0.723633944583485,0.42907014048498004,0.3815681016222483,0.0982819328702833,
 0.2632734690686055,0.7361689841327658,0.5663177701098052,0.8655088770289586,
 0.479606148461152,0.2327240986581608,0.8226343131509997,0.9937389486584591,
 0.8981223168991664,0.009284444658243407,0.6414888866425202,
 0.3545809481271053,0.9998937655293911,0.7943946060512048,0.4144936435674599,
 0.37498701110898447,0.05060720723119805,0.3795967212169227,
 0.5014640593008098,0.8016553657371532,0.5096536484954863,0.13699648973010958,
 0.33481114933320744,0.6271015677951635,0.25050941403254423,
 0.14341696593458442,0.13061140447914943,0.44032660463166995,
 0.9775023211567534,0.44675966903343867,0.273278273062298,0.750380449653253,
 0.9020576518153252,0.9311135705128449,0.6161801333831448,0.5601971646446691,
 0.7570880551340724,0.10560384262478495,0.24598257040599247,
 0.8976774154756667,0.47170517249810706,0.8623726250022338,
 0.26776423947128136,0.5123371399455362,0.5721701383105807,0.961951944726437,
 0.8212847974576334,0.9471706143232096,0.04278705633285951,0.3506066190275713,
 0.708291486250223,0.8953893202383326,0.8319007968123576,0.9355842416763609,
 0.9134001604554226,0.6275021014076232,0.30856021148677315,0.927712415327695,
 0.21404428189391966,0.050294935998345225,0.07579166761541445,
 0.15575734961892307,0.09156251476246913,0.8011300495434279,
 0.12384223189114163,0.177467143429766,0.44712005431251667,0.7842352032394674,
 0.6150159081918022,0.24402163966345103,0.8456014349991633,0.1535838894639966,
 0.7721635654978066,0.2050973968936569,0.8600332649279647,0.08841327679828803,
 0.5882121749256368,0.4465344966373992,0.47676430185115637,0.7849900313450949,
 0.9828200193579026,0.6077130276138135,0.33270084641903186,0.5884660871743218,
 0.435301164083405,0.03986198665435292,0.819010988814447,0.18985147035604366,
 0.050686327642760376,0.13697376518090043,0.6097808252886028,
 0.987268585024586,0.6977488356776649,0.9968939494524255,0.37759263889836503,
 8.411416324471954e-5,0.3158503140813398,0.7379684189730512,
 0.4668071606302684,0.8452781399854887,0.16541900345359806,
 0.43005671420588343,0.4018845678491093,0.7374309090805378,
 0.10269909401087396,0.7992378755883767,0.28779554155142995,
 0.43146912599807163,0.9781188096206092,0.8123095266090709,0.5158787954457605,
 0.6258514946603608,0.863323062633307,0.686173000263605,0.6198598238911734,
 0.057488483731125695,0.8148244435175762,0.791693370693216,0.6979247895214363,
 0.36716999720071863,0.14574966189309835,0.7031190514803298,
 0.18110825199821212,0.7664300222889636,0.8203451419169284,0.91544100074126,
 0.5518695339593476,0.10206790673943467,0.1290409413273389,0.60555903636745,
 0.619168166763399,0.7626872660245209,0.8236147123679202,0.3527454132931429,
 0.587624713902372,0.17762549862306964,0.5802617852798919,0.8192763424704214,
 0.3165896339449302,0.5052541268025421,0.24717542339005816,0.8941696950455247,
 0.14087934128997626,0.33686733893732623,0.38259177529369925,
 0.6107358465637303,0.6088000646355854,0.9303891890614049,0.3062428213030072,
 0.04748480007297784,0.06452476440111765,0.14481260287340092,
 0.7369068162087522,0.7276892069144837,0.4954996316170208,0.5187525867574304,
 0.48749769545871324,0.8150215986414945,0.7910967920873877,0.874828479667235,
 0.6921562481823149,0.39031041087150664,0.9284296928877671,
 0.47097333907858685,0.611931444490736,0.9029739090854932,0.14549323665671465,
 0.8078391486494562,0.9361508569841499,0.6763302672878377,0.907064863696208,
 0.06327623376381697,0.9244780845194276,0.8503330937800849,
 0.11767749998275245,0.8892949725277877,0.6992115928330153,0.3499111390352053,
 0.8206280122316275,0.6246731627781077,0.9130389906747803,0.8539437237668147,
 0.6779435024362902,0.023599894457479453,0.44496548747567966,
 0.9114371133751119,0.08796055271418113,0.0882122610970223,
 0.09281506321207422,0.2669502118763031,0.7941437605313568,0.7905290940625923,
 0.8303884607840541,0.1432127098111965,0.42361600230947916,
 0.06436802102993266,0.8908094712177885,0.08492609588001798,
 0.2068638088386262,0.35382810967088707,0.7083083203101439,0.5261648045886208,
 0.5780211004548695,0.591411083401347,0.37577690689429377,0.6526711943684491,
 0.4859020661014464,0.32994807865636533,0.35079583506747425,
 0.7553929156743641,0.25008127782587986,0.8707226579311222,0.5595684775705967,
 0.4378213372355658,0.7681847793100207,0.9748386433785143,0.17101924560887094,
 0.14743168422481645,0.8961979478291915,0.049999977484242675,
 0.07500608527833752,0.7315080601035224,0.710025410899727,0.49403624926921486,
 0.4533600026723722,0.3517650327955425,0.35059159753349867,
 0.12588966543232194,0.7999679228694909,0.05921422784514374,
 0.4348631606807194,0.1245236613766082,0.0857068069447724,0.8107813193888408,
 0.015584684856115638,0.327181283210638,0.6803842539561513,0.1167986897264548,
 0.4339674196341762,0.008785321312548056,0.15090287528610435,
 0.4701922091698453,0.5865547534635848,0.13744752365836388,0.6258246537415002,
 0.8952524710220193,0.1189691909510413,0.4132695582914858,
 0.011932460035213532,0.566723986788809,0.7699367347816994,0.2857560846854409,
 0.4261771843675607,0.0337295229195107,0.16132290689401974,0.9779316164513239,
 0.2510162248032921,0.8069965931900611,0.5494002588026297,0.16739006024169412,
 0.6704358047241061,0.5206956692715209,0.0012760058218538317,
 0.8154412932154296,0.15835880215006937,0.21430044738464527,
 0.6837683545973668,0.10178146508940411,0.2221479626825832,0.6061839619743323,
 0.4982453454520459,0.15568553256633477,0.45001988163244266,0.386709364088589,
 0.15266771861033301,0.7666063618752433,0.44308691250570875,
 0.12824788790896624,0.32542176554926805,0.6562203035459,0.10939645845790458,
 0.9135698240945354,0.8666496429426758,0.07711349158660563,0.7198969941018306,
 0.8291427907744224,0.07662853949911763,0.5014150168390755,
 0.24007185496681216,0.30806061620636915,0.24396721263679777,
 0.48061737870762833,0.6757461914500297,0.7554834956252661,0.8035633999079499,
 0.8641476451789842,0.44546037671112804,0.4959753025013224,
 0.07226116211461608,0.2979824472093602,0.4685022474334243,0.7430317196228629,
 0.9594268430349725,0.18302894700469385,0.09215949097550413,
 0.45097596996362344,0.8984988113838168,4.2127562258176265e-4,
 0.7352506154433005,0.33478636619497637,0.5904039178825962,0.9016599350667546,
 0.4006087674280968,0.13926620443755366,0.7515576881473854,0.5909431553175868,
 0.9725309329905751,0.842998940059308,0.4659065958797778,0.7070697004997878,
 0.22362679139847685,0.7943385030980163,0.2624880860523401,0.9667010096795988,
 0.9784685329352072,0.005235802082560803,0.6256576021168603,
 0.45871857023739226,0.4236783721195534,0.23230028560490035,
 0.48739878773847267,0.643625518117496,0.4865254563367525,0.3158487810596644,
 0.5180185505176471,0.8257587941755568,0.07849455064814748,0.5464898656997794,
 0.16629028160427373,0.9444956105077312,0.40777643095709815,
 0.20224479052981303,0.8182377071305938,0.42146473773289805,
 0.7584393776959203,0.7280621969027852,0.5396620132218444,0.5145107354650129,
 0.45818079804735157,0.44366646431867374,0.2787980299301037,
 0.14259135187073535,0.1625991653303296,0.5914216118169062,0.3128913300640104,
 0.9952230291474783,0.2866993660653546,0.28745568117016673,
 0.21636754562400795,0.371157868933747,0.8496596747794731,0.8925674120049085,
 0.7433955996334385,0.953471145110175,0.1753264007341928,0.255056417308988,
 0.3650646202685697,0.9136377631654351,0.12000954120743013,0.5291661376209027,
 0.4537289191818308,0.8630654996405971,0.10787157263581904,0.2329379553878448,
 0.13376218625311342,0.06054398657826765,0.8211356998749039,
 0.06609056065638774,0.6283681073410892,0.36539524739068885,
 0.8373575170124733,0.305662238719679,0.37402673985728563,0.9651993025175654,
 0.9237846924678014,0.08898366640481103,0.3237435598258481,0.7493096826279881,
 0.5407253230892823,0.7742548444381077,0.9932658473790559,0.9174064069461979,
 0.645505348113525,0.3395469915515391,0.12538482906182202,0.32438592940122346,
 0.7723814960302703,0.16307923562226811,0.8999406878589232,0.1739227582377083,
 0.3328704017127241,0.8319292873146336,0.532419493587153,0.9465942630724181,
 0.9783601056760586,0.977204026956085,0.25852121101158,0.7917592920777795,
 0.5788799461844942,0.9893519186370248,0.9991135597171033,0.37506709674400174,
 0.36881756435926727,0.5331430988702172,0.34984266406651854,
 0.4103281051542058,0.7640820257785578,0.14363081800712907,0.8197232758348757,
 0.008699148408169322,0.559366916883789,0.702066486530023,0.8676846008483139,
 0.8834772394347319,0.9844302529804126,0.7448227831742699,0.20545792040902178,
 0.14243744773945777,0.4959200501249259,0.571693945847344,0.4973318146040622,
 0.30797625225739145,0.5332346275863777,0.9832721200018762,
 0.26309063065921245,0.23553676249707933,0.2327439463100871,
 0.26899623969349906,0.10328284879057092,0.4374489014353099,0.153379433121129,
 0.9649861442711469,0.9796842182376331,0.34939372350798714,
 0.21990456097623823,0.37733339675773936,0.8083950544775463,
 0.16566123931764898,0.811970493279363,0.3854945110602359,0.5114785096055278,
 0.3592016369034188,0.2754722616872466,0.17324008252571177,0.2478851688159769,
 0.47929154058529844,0.25091378853964763,0.4664290034933154,
 0.6240411749832524,0.9517723880571414,0.4666578407177445,0.9510914110769823,
 0.28938342310220544,0.3143079649903604,0.8531048533439985,0.5531640759927815,
 0.19490687441282473,0.28376158439665977,0.9465401654601235,
 0.9457992043005456,0.49163513071845755,0.037560737532162713,
 0.6805723235374298,0.6409856623890835,0.14279696921400264,0.8345553007330533,
 0.01596836598597351,0.8259292723013713,0.9961527228457655,
 0.19412381924700584,0.6332367656840228,0.09276653335695617,
 0.21409116880109869,0.9508853421280096,0.47539125900769363,
 0.5776833744397473,0.39721841710762684,0.7061717710460509,0.3205269757540281,
 0.781744911043692,0.6692686964798606,0.6043521238787019,0.9225916989010033,
 0.41269192422754475,0.5845770002772404,0.2022196645696549,0.4840454927405369,
 0.8210395397563449,0.36200881809329954,0.46202227028173815,
 0.6292431504567201,0.040496413080232996,0.32690800562430855,
 0.5661264496446137,0.1406076754819524,0.6482963519209323,0.4167302906525636,
 0.6554378602514582,0.2507466183357282,0.65864141358444,0.46525866328319265,
 0.3943626817706114,0.34298111491228833,0.8266988008657122,0.2409017414539223,
 0.4707689030299418,0.6254052121008822,0.7959071572360124,0.7975953875012657,
 0.47196431292857843,0.15498149432391128,0.6469664800925561,
 0.7842676362476909,0.8722564987051721,0.21609008931689677,0.5013680588539143,
 0.38092020621500744,0.38967238208958344,0.4721884585399174,
 0.9433494181353413,0.3266415765430406,0.030872495526601762,
 0.15080978237127218,0.7112089291522002,0.9836745753157832,
 0.38477818574701483,0.30398997856827636,0.0022714030528647733,
 0.1664722696241172,0.7178700864100074,0.2639470372453204,0.2032483463331578,
 0.7936414598649559,0.44384010194289447,0.8412775546684206,0.8545704363692805,
 0.6279939475884209,0.1916325531975589,0.16579160360417833,
 0.18445129619257583,0.4471869924165358,0.389394561275096,0.3208890889298347,
 0.08664548610787626,0.8470063558574839,0.13012938111563765,
 0.24324197247364177,0.9307856169783907,0.3662071739688785,0.1780868962451917,
 0.6048065049793361,0.9057852599079542,0.2514902393728594,0.43252019180936474,
 0.5992712453192474,0.6017546077481537,0.19642302092778596,0.4733721973395879,
 0.8574648613014014,0.487981858845686,0.8511638925843418,0.2695677260422208,
 0.5594787047150557,0.6029871426583464,0.33469810022113833,0.7175738885756946,
 0.14686279446494022,0.304678236808854,0.13308444397453278,
 0.07801930837620275,0.013612611133045127,0.7766952224018441,
 0.9441238684933372,0.20054880962115873,0.9459530316061106,
 0.006483191310402647,0.8272277393791754,0.4667435717264312,
 0.23699667926002044,0.5038113588094002,0.5782485693420589,0.5732452563554089,
 0.5449337373004113,0.35181445401451916,0.7369828755998751,
 0.47618603985823227,0.6187876210108625,0.18713244201992107,
 0.5054040099714843,0.058373706722091856,0.4087375807434426,
 0.6454496890180063,0.07651274934823948,0.22861268008735436,
 0.6317859342764037,0.029472131313829197,0.41826203681693863,
 0.11846539036466142,0.33211069292756823,0.43891707617481013,
 0.9791311975922464,0.9949432910372833,0.8106161390291478,0.7184579558617158,
 0.02145300826876939,0.28144827729035216,0.24145067554896726,
 0.7972718962265096,0.3437867354791291,0.84790860689069,0.5987759206560774,
 0.011346608263349323,0.5821032146114525,0.5614269139623355,
 0.7130003107321072,0.7133090076939144,0.7176853712017192,0.17547250345417598,
 0.8078161796076714,0.08973688783679068,0.731029397020411,0.4383039711452499,
 0.1105523922982361,0.1194311060884421,0.5205180734950711,0.3546901368263612,
 0.21373823561967653,0.03112735000389677,0.03057038934406664,
 0.6461144972549283,0.845083096802577,0.9512603460980955,0.4590789170170321,
 0.30094546029566993,0.19910472441563454,0.2943161302090699,
 0.5407993217855952,0.5922924703860999,0.9111734237790727,0.13029290127135573,
 0.6711193513548331,0.21867095985181206,0.37076512977352094,
 0.8124567945253588,0.5256647129985366,0.013611735408195269,0.627949012353857,
 0.18357793022300728,0.9610574130184555,0.11667999482704172,
 0.9358247326221094,0.5875517324518923,0.01787093924929306,0.6758933985326945,
 0.6007627177248838,0.5811639969407134,0.3214968603511529,0.2877838319212451,
 0.9684065390372633,0.5833593005291656,0.6475572988130582,0.9768830980907309,
 0.6038846894006578,0.41096866827816103,0.31860907239645253,
 0.5405175173933379,0.204385380591674,0.27970093715430844,0.10310928740092506,
 0.3899343014032246,0.14194603206685807,0.653584987017362,0.8005849110208618,
 0.6784232484431137,0.3075935662211102,0.5484030919223576,0.8638978192956704,
 0.11410325884314676,0.5875840757826521,0.4319088332899712,0.797342258780507,
 0.909146634689671,0.9894926985888586,0.835128936532577,0.2635442699412859,
 0.14676957573427063,0.38275660510010767,0.5496039507935794,
 0.5205120503153807,0.670952873620858,0.2855085178973902,0.4888441544442568,
 0.4991534272046959,0.9392727561771715,0.6131730929910768,0.4484258249565364,
 0.380784968964768,0.26221854171121417,0.8981410355305841,0.1466863881732703,
 0.1789918123876143,0.3187980686703067,0.66370716730996,0.8072194742124944,
 0.7207154271160005,0.3785702204760645,0.01050363723161829,
 0.18057558961598574,0.7833139610119799,0.8490512437215225,
 0.024397240220660366,0.8058573746456072,0.316628629127231,0.5997248507176862,
 0.03292405376452545,0.5168913242752559,0.38886624231406397,
 0.3018031415854081,0.2661416166059376,0.7621215421004218,0.7386932822933242,
 0.3256156316330726,0.9279325183840564,0.26685824358203747,0.7069512294202855,
 0.6179813683274679,0.3875347450457346,0.8070882911614535,0.22836615679588568,
 0.9627300092548522,0.639451160247524,0.0213278669163518,0.8033616779477574,
 0.15101743395516198,0.678354532826691,0.8545869084778651,0.5202491511419537,
 0.06866231436046921,0.6351337506207839,0.4015023428659902,0.6417656072022451,
 0.468897048920596,0.9460320619426136,0.16731860089184258,0.2621367296535271,
 0.5303125164589562,0.8313280967005098,0.5785393853776126,0.508699428049157,
 0.02419784353195309,0.3220717205850866,0.5160916546859473,0.4955746558789006,
 0.2918675790259917,0.023123988844309218,0.6307411788946119,
 0.1654480563160574,0.2812458418967967,0.7898452840837277,0.44485242979727246,
 0.2373576289586412,0.3649985021285619,0.8798187715993846,0.6946745132544369,
 0.3832174856393056,0.6727153582115597,0.36992732356235836,0.9932880691399286,
 0.0384060155953424,0.28497159356469626,0.20840605292709558,
 0.9771055221888982,0.14591424212424486,0.4441131372635341,0.513527405528502,
 0.8468553153107742,0.8050854824726242,0.9322026303666986,0.24618171249929077,
 0.5508010783120514,0.37586723572720904,0.6095312226052736,0.504423915910025,
 0.18567635894031664,0.9132813843188601,0.5466577128466988,
 0.27349451403544733,0.8219267023401395,0.8851192656613966,0.7581048175447826,
 0.31780903907685687,0.8293929309379662,0.9797141831993257,0.9657814604577546,
 0.6643133457374627,0.9877393726163616,0.4950659849969292,0.1957100348424301,
 0.8166463007207876,0.4778339100879193,0.13766296478842233,0.4868415977503696,
 0.7143835041400342,0.4813171811013336,0.15071731288193102,0.2350733096847164,
 0.889481650692395,0.15915210752821007,0.7393728665018189,0.345710191019746,
 0.4569746905567783,0.5834674439646418,0.8162118021287096,0.31068693759558197,
 0.5814386961576699,0.2311843243155015,0.4603089537812627,0.8975537997177346,
 0.3964536520067151,0.3607879908832513,0.7053151174561588,0.8566271454269384,
 0.1306972823918735,0.46243259088840827,0.7604393033691116,0.6459382290580875,
 0.5947456707303604,0.5597454262042072,0.3721748788965701,0.5672655034958634,
 0.8409758692066386,0.6219095634165343,0.7603507782139655,0.9720311979825927,
 0.7477653838633883,0.21857806742623964,0.4362297122451956,0.7791120238488667,
 0.5763590018050101,0.11428588636174031,0.2252712005606673,0.9542896595734788,
 0.6333708284561584,0.9988301072483783,0.828200766046629,0.8257603571408006,
 0.8267833108829973,0.753260418575932,0.29832351729759976,0.4114676221082251,
 0.4970544136380113,0.050701515641945694,0.5080852793120687,
 0.8854131332282301,0.5988980792985256,0.9741023907138188,0.7336761958951156,
 0.8386014283512317,0.7747730519264671,0.2942456936601793,0.8882678532199213,
 0.8604784850569175,0.7262292517974678,0.407460436443317,0.5472976455324295,
 0.2731310451086446,0.20570944611107178,0.5280297440647026,0.5802044636921616,
 0.7650744375592786,0.12256562164927975,0.634186712844029,0.10756599670724576,
 0.7507857191296592,0.46834775309862064,0.13692868631254218,
 0.0016667045805567327,0.9840519477793451,0.5871703358812801,
 0.8440872290381212,0.7784165294663816,0.6960370165945932,0.6260496579308333,
 0.9386479457840808,0.7942297539659963,0.07874827716811916,
 0.07579994101424858,0.48008311744144827,0.8187678590246953,
 0.5806302215257398,0.36826471750260426,0.8230583290370213,
 0.31180930043536315,0.38005529899497126,0.9675434973728898,
 0.2630474721293383,0.5377924292872671,0.19597822279297272,
 0.15861030174208623,0.5123156102517262,0.666626037908091,0.9536917342690422,
 0.8720585856354421,0.9301445594345116,0.4988967718051437,0.6503987141540875,
 0.7378561085918767,0.3042264375859244,0.28039549866678426,0.6949198063627691,
 0.349132162875863,0.5785168256723066,0.5689983268524598,0.6420438938404285,
 0.6373546130563519,0.6952526763700752,0.7665436272937268,0.41615137691331205,
 0.3668361252552954,0.7166059572852399,0.7734652409562621,0.5527703308440095,
 0.6275048686028828,0.3824086324060628,0.38362689081694046,0.7557353756606654,
 0.873473186690964,0.9054680253371429,0.17753104580594437,0.5744859368763627,
 0.07146360956350406,0.0044717804868315,0.9317046170283664,
 0.30378753318782126,0.36824244919249716,0.49299226490865844,
 0.7226337575041668,0.3874533856442909,0.46302853706791103,
 0.07248719625443001,0.5090333690120558,0.6751423926455724,0.852159027128505,
 0.9550322395626405,0.13413106121989393,0.7388597555262305,0.562588243631922,
 0.38049309714426505,0.5815573470529202,0.6527211572225344,0.8428224936566994,
 0.28986270619359344,0.5515006896595271,0.1694028978355393,0.4852140472686135,
 0.10051262537353933,0.726802459738353,0.9255521847203618,0.14421818247064244,
 0.8364050682978729,0.44221669299117616,0.6516854474090245,0.6540234125404408,
 0.7407390069582708,0.879291792185937,0.565226372020878,0.2963246112409932,
 0.09150718869627639,0.6120646284999708,0.27999074215797204,
 0.22749260243386815,0.7936752837428199,0.44695735233456557,
 0.46535726356770646,0.7460824311109242,0.7452173625578791,
 0.11251027030471805,0.988193917468148,0.8874858214487705,0.09131613142372397,
 0.6672320289096358,0.17118523815565734,0.9624368114591744,
 0.29084012763713407,0.8653511506284922,0.7020400595370118,0.9304904655729106,
 0.8007325555632457,0.39947527217928813,0.21430784422753202,
 0.9164217232733265,0.609469694426415,0.6039590249233459,0.321314181013035,
 0.5078126738450869,0.4166159262793725,0.5790115222256823,0.7736484452184453,
 0.6603463790355117,0.752420089572271,0.6835150752616168,0.34758197647503797,
 0.6058191704867025,0.982908006032438,0.49514286912646277,0.20133582174646159,
 0.16435279936535885,0.37265904656549564,0.6652098768805206,
 0.013874772281753489,0.02091189813855876,0.6598663515357214,
 0.03473441841431746,0.6519293614933983,0.6834427368438942,0.0537125934576026,
 0.964407236412248,0.8088218292881928,0.3899811260147399,0.8427028840694533,
 0.8337274364092939,0.3013626295789382,0.686697434511581,0.40714112450935824,
 0.14864979249028643,0.1315665745015686,0.1842369522306655,0.7101650896933548,
 0.794045666607907,0.13412000616773045,0.034271821234319044,
 0.4720128202191295,0.9710076380089012,0.39612416262501104,
 0.38513801411604776,0.5124908098311841,0.09512296142487109,
 0.7948898829287065,0.06933710863373221,0.2838326767175844,0.822061264050432,
 0.30617550490750156,0.5850049074103457,0.6957736296377919,0.8765903657926859,
 0.8645010355818715,0.13730192695149623,0.19029661667526054,
 0.7652975300920606,0.9823958756729435,0.6958996667760518,0.272379987923811,
 0.3457674159798363,0.3381921839774015,0.06142959122520275,0.3165986038707074,
 0.06391600236154582,0.6326611611232849,0.5634356790180576,
 0.17198260982944868,0.3043309925971822,0.7942241410993668,0.4315092570140241,
 0.6236671312745594,0.27452588320802607,0.3920575238142948,0.8910030460651651,
 0.15459063548572927,0.19866209825051073,0.300557224496809,0.9766301352310074,
 0.07228536050304557,0.5361154693818281,0.3503160045736806,
 0.43927192459480313,0.14080650676630468,0.7002266501957098,
 0.6821610340056616,0.7665321086859294,0.23850418645078297,0.7709350159415653,
 0.28187713452128127,0.4039723417200308,0.14609301807716968,
 0.08291921964093618,0.7809727350922192,0.44522651712364536,
 0.15118073553648603,0.10191036774620754,0.2788406812646319,
 0.3022785110792968,0.250557778234082,0.6322380403259326,0.31395106885919233,
 0.2543068796356106,0.842341744645535,0.5972310258401031,0.21907834483342903,
 0.2880714900910599,0.46633061885978166,0.613721946578057,
 0.028630695127470962,0.04058291658004409,0.36722213328162834,
 0.9299355786441599,0.5165109621133406,0.7770457401262636,0.1649434686365474,
 0.09455901949110346,0.8952648464260445,0.3767383426143657,0.1914097174552606,
 0.1085902117650086,0.9424636029990175,0.6885441605639849,0.9649909527689795,
 0.06010374270927432,0.8863893969203283,0.5315969797951586,
 0.049200760538255506,0.31276732996907763,0.6731466708531163,
 0.2564830285608507,0.8157166279234387,0.3413851160287835,0.07340845812177599,
 0.7145871826384843,0.24281741067083096,0.3301028617139755,0.7124693167199138,
 0.19175378669329302,0.5320656780142554,0.980728942494425,
 0.033417642749652976,0.05793297592252811,0.7866091090321783,
 0.16330010588019217,0.6237898789171278,0.5522064908909199,0.5169986001275984,
 0.3572828209302744,0.5111323516257681,0.535700194754571,0.5984023135047131,
 0.49065593761961646,0.11001510766103029,0.28581489169730334,
 0.9520853244066011,0.32439627109527414,0.6015507579988415,0.3336810538874676,
 0.09109176074975955,0.9089291628248596,0.8090511518998162,
 0.15340233716707852,0.458888383248403,0.03399255752124419,0.1747903268129315,
 0.4393124627425027,0.30369942322744603,0.5097649428235905,0.7555177759179166,
 0.047189131314841704,0.06769659295023733,0.6512111920949678,
 0.6459958218004556,0.0379673353013843,0.22252900993528602,
 0.46405596959093387,0.9526902578544714,0.843341691565416,0.7126477680320545,
 0.5329037607465473,0.8175091490014651,0.22976344892029432,0.9245689739061531,
 0.7402708291872695,0.6591818357953058,0.9188137024385055,
 0.020876698298246787,0.3830590247659956,0.47033595985265253,
 0.5915471366558728,0.3682956132648948,0.3973011294510572,0.7570119182511093,
 0.8688112572777054,0.3086340090952545,0.6751588055063109,
 0.0015785414239907203,0.6287843932365098,0.952775224360955,
 0.2833210510252855,0.5704602963927845,0.5571081544761969,0.14686530085496186,
 0.04416459497638636,0.939847589298747,0.08267937247565937,0.6773183481946403,
 0.7009169918303227,0.031137124341484634,0.9177347158046549,
 0.7957999141108594,0.6215021426978224,0.36212926131652,0.4655757198967878,
 0.0725522005012208,0.7172807034307164,0.9208610354942246,0.19871989746214624,
 0.2133730008213941,0.17525385717087694,0.49601320729374343,
 0.34248243228810316,0.1398164948442262,0.1329804164425471,0.8324481063244786,
 0.4753757929153276,0.938031623202127,0.4839509875157033,0.703679077750512,
 0.5194011283527558,0.2086077770726692,0.604797249329349,0.6046952816567717,
 0.9079481699867163,0.3506896068158323,0.9734921428246759,0.9417322701484946,
 0.5768129585478546,0.5194927743348774,0.07538136389663563,0.8122794834330245,
 0.10607292296695514,0.3591703086716582,0.2795591819376232,0.3748870638624514,
 0.9965732198435898,0.12018840596458791,0.17950036200564867,
 0.03937517719902561,0.09152821454214699,0.5472772913059218,
 0.8714603963692453,0.9352301073793564,0.09539625103257299,
 0.010312848574675959,0.4611591807183213,0.7705458046645633,
 0.3361054924363942,0.8458006987325944,0.22993958850428864,
 0.05593699653876705,0.8454256377908153,0.334816176714838,0.8724223405636494,
 0.6830861905775842,0.8470577467467152,2.6815029514348776e-4,
 0.08194513168959894,0.955307707763146,0.4766920766167303,0.9287856656116976,
 0.03951430548811108,0.7145756008179395,0.8251074655727381,0.3001979382217925,
 0.748795168382328,0.5931151673195009,0.8808191439116881,0.4313531762634776,
 0.36180408778916084,0.47178549207827725,0.37447163962003704,
 0.6962113480267755,0.3792725740503705,0.3330223247100582,0.5942206175496065,
 0.06233553336069009,0.858711091702402,0.7037564372817294,0.0168246455569806,
 0.945172086040305,0.2909756154413333,0.3630926244451198,0.6029434896603016,
 0.2944590098707164,0.2611196685615277,0.804645722762221,0.8502040684061756,
 0.5490955154914429,0.8853746926041962,0.8935611381057347,0.4303506595758799,
 0.5499856824044076,0.25824594444957694,0.5341983718276196,0.629369079375566,
 0.7858168089404434,0.20849542793339992,0.2621255822895233,0.878041700342523,
 0.011941125498781235,0.8150291350362177,0.5872609342239294,
 0.7828093274655409,0.5857404642520345,0.1611047617270699,0.2145317115274633,
 0.8387794233773682,0.8972194548519687,0.7524605353203372,0.33265026118150964,
 0.5646887217865064,0.8929549818643144,0.45414727586904513,
 0.02644546986287566,0.025167374196599424,0.20932995748706107,
 0.145744903134844,0.6926014383907887,0.3573643209325168,0.677633621909855,
 0.9046517680542159,0.16208160143740702,0.21520722250694135,
 0.9357351595351113,0.9848978658023749,0.8180807362302739,0.6533603724837189,
 0.9041435882062527,0.9936129420901715,0.9475888130398289,0.8023603396234007,
 0.9811895557596207,0.7062830916199351,0.5877642030863247,0.3555578663539505,
 0.6636273958728947,0.7292078754633159,0.27151615981966737,0.8336640483802938,
 0.5176944127903751,0.34313704654511634,0.3185759168029394,0.3865247001379437,
 0.2959392977470383,0.9185192018963091,0.04518392500897139,
 0.16989102232999687,0.3747445243870441,0.3496180645055166,
 0.16655192510935868,0.3503729388933855,0.0223734943768652,0.4635417797142566,
 0.4040081214945269,0.9492514020895407,0.6836064789391069,0.3769127961054053,
 0.9244782825043425,0.7988643623099313,0.9863316719431516,0.3349501063057425,
 0.32336348862442765,0.6876119991796098,0.7912674925496146,0.5253558291144147,
 0.06915292118556648,0.4877628181872695,0.07241540464215657,
 0.41908496202533907,0.7956283503728386,0.8136504111349491,
 0.17457897906288722,0.5048154669197532,0.14345008505229895,
 0.5378690167720548,0.6372136720025576,0.579132011033653,0.14509283329476474,
 0.15148454535640288,0.7739921283100384,0.5023334499498558,
 0.43817169151580104,0.4797054046962095,0.721198672867531,0.2878352408690088,
 0.1932154415896996,0.1035027654589471,0.08097608179659976,0.1710643391975618,
 0.009859357574471983,0.9839009566070898,0.7750514339801177,
 0.7447133661811259,0.9494980512667748,0.6884043944423438,0.09518486030121143,
 0.6032668658708937,0.795704715290817,0.13202912971890068,0.7573144637938057,
 0.3049006103769508,0.4813111449035393,0.5992831113714973,0.7338869342705249,
 0.7732095110850817,0.07123684651740347,0.07607663745247517,
 0.054639358356514456,0.17739960097522967,0.35546809660133394,
 0.9797400182613092,0.0097542396151723,0.8854851679508305,0.15281113144975667,
 0.1570824325195158,0.0067759448068680506,0.9404369257363537,
 0.33409448431101174,0.5650039603021669,0.6394641051867028,0.1518294323516507,
 0.49830697905043275,0.016897384370675228,0.4239049664875544,
 0.9183644156719866,0.5625672106761153,0.4903722405844573,0.7827013395657564,
 0.1658946623981239,0.9448708489087908,0.7741643226227586,0.9849033781681127,
 0.6085121640711002,0.703050386631592,0.32205728280029733,0.5502419397077516,
 0.18700507060447513,0.7930886238196297,0.7462565367350127,0.8194427855427313,
 0.79867905874825,0.32867044934172784,0.617995905240549,0.35392541108850184,
 0.10697224264172145,0.4655959000247517,0.2564608833972818,0.5024904387422903,
 0.47918475755176515,0.7734544577499033,0.6595438414268501,
 0.0043772291639641026,0.18190443021502722,0.04362821726489108,
 0.3242039699533591,0.15531341160271794,0.2528877869027454,0.5798462499513299,
 0.8203532647354007,0.5854907170474343,0.5831910838623634,
 0.037117379726658184,0.3049489189858767,0.29564590903425625,
 0.28232429145705396,0.2217868176379153,0.8416910181948747,0.2149855928560327,
 0.8611834963043375,0.9039061429052163,0.7284509971624891,0.20863911405162172,
 0.44160697884780054,0.6217260484208513,0.5159599114775899,
 0.04021857880900226,0.22055569868283187,0.011048125857369695,
 0.007321387468967044,0.9955635068185011,0.8853103143884644,
 0.6340114166131873,0.7780880833425997,0.8300858424229411,0.9776492393621684,
 0.7634922783700288,0.818103141697988,0.5203138787777378,0.47368476516345703,
 0.34150922389586236,0.21099656657977806,0.6500017934278491,
 0.09086439011891556,0.5659777982907418,0.6110616842954247,
 0.41611344896822045,0.7901104526311211,0.2860441373882425,0.790803973125921,
 0.7551746632521514,0.5665238792830234,0.8000908783093142,0.8458663959419891,
 0.41951910297662964,0.5977745003752114,0.591631518589925,0.7257396281495032,
 0.050310613430270656,0.6766299167517047,0.5014030157533668,
 0.6817898575998869,0.2824889685042187,0.5762831234813977,0.5201600199910084,
 0.1364196579187149,0.4208305565986842,0.8094899106528436,0.4783121394528478,
 0.7667684489753288,0.3836478619224395,0.9004386298370055,0.948290996620244,
 0.0032516897852151727,0.8182804613331571,0.6435452771967966,
 0.48104334941877214,0.4233327110796328,0.30850771710850744,
 0.2343311791266096,0.9813330156114944,0.7546312357985842,0.7501647765353547,
 0.08186959919889181,0.6079949032055019,0.25758940887340775,
 0.5796279280183697,0.24444297350112643,0.03031739956921431,
 0.1411568773524191,0.17217660464298712,0.7799110782625456,0.7655599040468072,
 0.8736910624279302,0.96453406669104,0.6692971402366625,0.7956310764539687,
 0.7777458669749645,0.8722349865319355,0.07399349203353045,
 0.22623417669132861,0.4365087114567191,0.09233845161235299,
 0.004922234136146897,0.965993512444517,0.5657468039911473,0.8139862923212914,
 0.505370427744834,0.17666314069489242,0.5766225883558301,0.15428576360465318,
 0.3326859244424859,0.7588570708440213,0.9792987564687705,0.16740351947299825,
 0.3810764031988081,0.5155392764131936,0.43028592097780927,0.2898254854465844,
 0.5949249649285256,0.42140819161629006,0.47421060249416824,
 0.8347620135211333,0.9617493483426531,0.05580379343428199,0.1708876711944206,
 0.37328069249427653,0.4969333211962451,0.8810651174021391,0.415474103306146,
 0.47454515839662426,0.5564552588857499,0.00293316446808789,
 0.8813301484008144,0.6309624747274182,0.7873496777788467,0.6362742811199578,
 0.7517705812877015,0.9824812594879544,0.3768257151530048,0.1283700422821138,
 0.40655443125972623,0.021417841264826487,0.5676109787916348,
 0.4256741943672998,0.3598861450191613,0.3140864196625216,0.767259689049947,
 0.6278208978763521,0.15770570422176955,0.645908650442853,0.7410598004321861,
 0.3408762050050186,0.4402606239732525,0.7446840402714947,0.7673174866001161,
 0.007541509740762509,0.02404195550367938,0.34586403704868707,
 0.4476787902360603,0.8903447704791801,0.9001920786427073,0.5549700438509932,
 0.8660386829779657,0.7029507687967641,0.7579990617916478,0.7407391047472835,
 0.9211846787125197,0.6693778510728353,0.08198555623452675,0.7074291901597066,
 0.5377217394799115,0.4381104484163988,0.7315158970222372,0.3624608831803391,
 0.20800919142637864,0.2219992646276049,0.8715894134684533,0.648963192221919,
 0.8129273453704088,0.34835111285506304,0.295302485158477,0.05927161559468752,
 0.38018504505836215,0.5715759244105578,0.40573073171155305,0.712828652567071,
 0.6233071011924167,0.769070466190422,0.08613273153965195,0.7653016962425514,
 0.04577208698976065,0.08604281102979472,0.6659750896606952,
 0.17071990333452036,0.5873261000164217,0.036278742121544916,
 0.7040520336396647,0.2502409640747948,0.22139014306489446,0.6962446569085007,
 0.6177036356199963,0.9674421190499675,0.08520665184842857,0.9328605211782248,
 0.13247297998241592,0.26672116857243455,0.7313976183244805,
 0.4617879749526279,0.37851621542726865,0.1752002174746431,
 0.06752734643860192,0.5786084934105928,0.3678235277077806,0.9714441505497257,
 0.8545970233819035,0.0783217217475407,0.30510160176389245,
 0.45880054222364497,0.9675576077841062,0.06864107030203326,
 0.6194659609474473,0.01897597350396618,0.5086876300081099,
 0.33876814942672095,0.8702527410989247,0.7851598257976621,0.8227439213855813,
 0.36284588784659355,0.09013924480757474,0.830670146994336,0.2936406454687758,
 0.7932860522391378,0.20033220737944746,0.6924121415719959,0.948705825650228,
 0.7033113166714129,0.5665701941907898,0.08512253923946755,
 0.41971297036210253,0.7108214341550964,0.5881439557371467,0.8923383100636039,
 0.740858450197812,0.1256139917905027,0.5218858417456624,0.8034839946610723,
 0.6381088644094695,0.16077737968022432,0.8981449076598161,
 0.28472453158958055,0.7540458008176939,0.4942264202065799,0.9048745931914436,
 0.2590800491412435,0.17886181192805417,0.8672933162774583,0.9224215153093234,
 0.09905020164022837,0.6899216269996613,0.7434947948909775,0.578788118922662,
 0.43744488742814347,0.3393960711110774,0.8640428907421499,
 0.44239186948611564,0.15147739897542434,0.641143673918973,0.796683023417406,
 0.49750917079893875,0.7699782984342313,0.5141155852248595,
 0.40599357972864336,0.7245261312618706,0.3035155490121215,0.7561649375189092,
 0.09375788136666674,0.9825987027309886,0.9519940954095625,0.3756622092796499,
 0.0630741145599154,0.29207294514951165,0.4013505837082578,0.9288736765113553,
 0.5088159786225884,0.381006482970212,0.14434604518990246,0.8009504322365596,
 0.7602007459810978,0.612695460756289,0.5736420224309502,0.35068672623844754,
 0.4123468015568974,0.7058428095157381,0.3950543344198282,0.2370601080128023,
 0.5417779036922173,0.10311531706827204,0.5116201906978837,0.302274134520399,
 0.7137142961567481,0.6981791283219279,0.19194370612308354,0.7573952317431809,
 0.5410401338262172,0.9965791019557374,0.6340943932907858,0.9299230239084393,
 0.320413278443916,0.645810984950265,0.17353288866540273,0.8350099223628368,
 0.6192396138303262,0.8154650458830792,0.8944257621769783,
 0.023521377280161904,0.7866403225379396,0.9829122151530713,
 0.6151341175511186,0.05796030787383444,0.35085938313107545,
 0.7766719881260613,0.06900030369519483,0.3569695193199388,0.8220170430749991,
 0.6797824920360804,0.16377813618360926,0.6738627667476444,0.3781389203456863,
 0.32877851851664996,0.6474723892220493,0.3548274672699423,
 0.038030500764157815,0.3061176897833616,0.28323763478852393,
 0.09079198625014795,0.34393057560609597,0.9466227970013555,0.692840101334953,
 0.5840649943445666,0.9821406202095939,0.403300019923144,0.15450007959129353,
 0.1410231426555244,0.09281347811102703,0.3834950157529592,0.5963347294352273,
 0.1832961681516866,0.2667463803800002,0.33223376346462696,0.8132473254786524,
 0.9837282635180601,0.9544985195921131,0.8527419847138569,0.6161076241841947,
 0.41961512507665943,0.8211181308018024,0.16184582064664887,
 0.03755122371019559,0.5829330594680249,0.8814164726928738,
 0.036559337336932174,0.5223726695378383,0.04891558288298703,
 0.9416888393729481,0.33016118193879784,0.8660781882475105,0.6788895462665044,
 0.49687304619856476,0.2520772011564538,0.11665751584011885,
 0.6985790365660468,0.5583920359576591,0.443911991376696,0.13487494098909203,
 0.0499867316629421,0.7876745461493818,0.29745543666223884,0.9452300950206539,
 0.2251542184897828,0.9441709718474682,0.3060938803843657,0.19319607777950842,
 0.5030119385512644,0.3010506362596108,0.5384490417639034,0.5257508588408543,
 0.8423393866438696,0.6946378685801133,0.3455710341696956,0.3786672019253414,
 0.6065393996893332,0.1623149540098907,0.5720449121226925,0.07242996377289235,
 0.09208077800107672,0.2531254784499761,0.21409359879933676,
 0.6799257281646093,0.6923341075453383,0.40020178168971954,0.8215867736350496,
 0.09399343374262614,0.5381731546500821,0.22687524918876778,0.099674253607976,
 0.6547896971072122,0.41128858559641024,0.7329304933845383,
 0.14521070135531744,0.39224445687712206,0.798953346608778,0.886764207009383,
 0.8795019246602107,0.3035208981616534,0.4166825582107633,0.921813768303425,
 0.09431343217244426,0.27649475129424783,0.1436710745944626,
 0.8941575351202907,0.7767324896603187,0.5662806400161848,0.3294469700804059,
 0.9838400822223872,0.07704797893116577,0.7734735712737649,0.3376465963452624,
 0.8134341953079965,0.2001212742685735,0.6009011142300694,0.18393465523985908,
 0.5439423536871877,0.10131948860026885,0.18952306887349568,0.196028609149979,
 0.7739515843344008,0.0956035964808547,0.6274618731172001,0.3892036223652614,
 0.897177925344127,0.7707981939439467,0.48907381066459565,0.2814306760519114,
 0.19396935415518202,0.5724734117351781,0.3861465869366518,
 0.39526498933863397,0.5615157568105642,0.372721328070295,0.9810850853101198,
 0.9147919994283897,0.34752314937104734,0.7992063394962429,0.7606932338446966,
 0.9561107457945311,0.6187951681510443,0.539491649990713,0.07775229162855068,
 0.24767658476506016,0.08987337459752054,0.24205454967817053,
 0.7818754527229388,0.7428705913740421,0.10005937772897,0.3707966407230716,
 0.09922322267563999,0.5880334665709354,0.12519603187115502,
 0.6671856805951787,0.9185941869683378,0.4837389492657217,0.5267753918048693,
 0.4197137957667212,0.6468599436559145,0.23397947987308843,0.730063648298274,
 0.19138001164339424,0.8777702373136208,0.19916115593480155,
 0.8439537291533018,0.2728705484624858,0.7936134034935798,0.9015808187192937,
 0.5379146035913953,0.007731055412687815,0.6500958489977773,
 0.7961579692561173,0.5183490122901501,0.12660021590099602,0.6591020328243168,
 0.8489922942288406,0.9180056481989302,0.13977184344129612,
 0.27519419651230004,0.7560126160412122,0.16400114252331754,
 0.4021672636894066,0.4101706015158423,0.49960765432031995,
 0.34799568685473825,0.4167401195047351,0.018334446152192485,
 0.6768928867397135,0.43049833793631576,0.14352588353665252,
 0.4974572397344703,0.138989289069945,0.5815670703758549,0.12328837773357426,
 0.6139292606742391,0.37502198018161603,0.9016319708893588,0.7687669360519943,
 0.3591596746843704,0.6234807740036616,0.5227154761126734,0.3990305658842558,
 0.7165541822418128,0.6732564219982347,0.9591510893823355,0.7228775691696273,
 0.13317800895125642,0.04360639188887672,0.014044438192891917,
 0.751910996119621,0.689307586921393,0.366999009351595,0.44925233654571783,
 0.5664851726506308,4.5176362619836485e-4,0.719638879383524,
 0.6382319303783174,0.46564851703011,0.32987762668138754,0.9927722371777087,
 0.7654259637804642,0.05964222745472014,0.04967960787294401,0.78123264429096,
 0.010784221854521436,0.9263918191937361,0.28906843800732385,
 0.8945025139379834,0.006425847777277505,0.769567870698697,0.4473262494374626,
 0.07800596588708975,0.6447146098193406,0.0012032818932085076,
 0.025352760435730648,0.9115593421692119,0.06963139346058389,
 0.15483932958470947,0.24579082119152595,0.3891873828004905,
 0.6291595747306857,0.24120356567064472,0.5912173373149683,0.5269477531195204,
 0.2934547461196102,0.8205419552489205,0.7083636222201493,0.8691175348707301,
 0.3389658775682788,0.6090496018683843,0.5775017939767279,0.4841849156877265,
 0.01933879889166934,0.35865050377854335,0.23966283717315218,
 0.8461690693244138,0.6722103676818094,0.4945228902696389,0.9121446715716882,
 0.2908395693505884,0.08096875605364706,0.7455747225314753,0.7488413688645793,
 0.7277052841879028,0.2654741694495728,0.3603443303548861,0.16220298199615368,
 0.01399480084905913,0.08587144201621144,0.15273283089012868,
 0.2786493383422062,0.7809987675267226,0.7558192414986539,0.7998438309309024,
 0.6037971195092983,0.416036615089321,0.9928371382782661,0.361765139174684,
 0.962301120052702,0.15590812130877096,0.42329780541199935,
 0.19768577824567057,0.2310852070750834,0.9165051519109219,0.2551878729795416,
 0.21019838614295738,0.3379538872446495,0.4186085115879964,
 0.08520212413602768,0.3601001589164916,0.5801910295310443,
 0.48887382950367575,0.8440022736520246,0.1801022515321017,0.9266134006631531,
 0.737704377084139,0.7629349029973241,0.2670654187586796,0.05689623977842262,
 0.0969925710209456,0.7257858414479053,0.7401622656761162,0.2388842968305459,
 0.11918912826582317,0.7827687380905449,0.8868663917554753,
 0.39385592972037664,0.6404756445995401,0.49615159632800276,
 0.07645951955588504,0.35778290375230104,0.39779838039984616,
 0.34147165338269425,0.7107693572163362,0.2712258144043622,0.144298154013618,
 0.7704686790728476,0.2499905804245799,0.08033791736294948,
 0.02474263514182018,0.772533877878153,0.8591800388416599,0.6377844069815451,
 0.5220700044482178,0.8916964845968647,0.3042510912181402,0.31249464060958665,
 0.7806055961421705,0.9089302315978907,0.5980301077162306,0.32503217748307234,
 0.479336603507985,0.12827395216693915,0.5724750020605285,0.55630652219927,
 0.2783881901823805,0.013698432484566259,0.8158708970245623,
 0.7785399010889966,0.8899402025935417,0.3550714339289439,
 0.001767094739620667,0.343017685859359,0.3184871897850746,0.266629219585249,
 0.038319598071183014,0.6696428712454877,0.17255678771056449,
 0.254342953043206,0.8285478767477672,0.9576583683187376,0.7322695877994735,
 0.5763842344818885,0.32885783199678764,0.42366272817088,0.10104173669039163,
 0.11328045654358587,0.23323671419669556,0.3268446728675729,
 0.5983168508824024,0.16963431294448528,0.4401589479727752,0.695162697846039,
 0.07936603571274503,0.09237676353372115,0.09310893055665936,
 0.8410935765154846,0.7201456987925876,0.9084739519141146,0.9925920883899931,
 0.011078152834820276,0.0524956549291542,0.28681785568351637,
 0.9231883059802106,0.8747436392517364,0.5475358059785758,0.5664720420585797,
 0.7563618580138689,0.8991670204526888,0.385256255678856,0.2637193327996574,
 0.3274649041490659,0.7393115997340829,0.5443831348998358,0.9114048151329476,
 0.5960731288015353,0.36371542775167953,0.3212034025710433,0.8036719557800981,
 0.6781492099587701,0.19685527421808569,0.4138800758855362,
 0.14315318039243374,0.8764797700212144,0.951813746738366,0.8028082931024236,
 0.4815101863304264,0.5635753251974702,0.7597668996229682,0.7076143050721557,
 0.5817432216512805,0.6305879199093225,0.20768284437543238,0.7972219368876593,
 0.7552245495378187,0.20531861920577743,0.24599665777973967,
 0.9683523205908517,0.054991548715358274,0.5538577688799048,
 0.42274077428814216,0.19869180405021902,0.886175271941879,0.3673385184119048,
 0.502207505686359,0.4477674493021342,0.8332031646353182,0.04199623008508757,
 0.21919369981533998,0.7640715158986854,0.19694047426686867,
 0.9029868375318111,0.9583265855602203,0.38508480856235805,
 0.10369763071633398,0.9786041432468354,0.024282045257273066,
 0.2122252961208394,0.9435096138671385,0.6469660125247945,0.9450032403479229,
 0.07979496168394185,0.21932682596183928,0.160352834947746,0.9836797830484352,
 0.42588952357058707,0.4831407825281613,0.7748659284550606,0.8134257825202265,
 0.8229256949565306,0.21801542984719113,0.06230273867587188,
 0.6371709203651911,0.7983183293009675,0.16762801084005052,0.1480362057752863,
 0.7309087880114593,0.9510355094382452,0.5972692724194851,0.7256980625827026,
 0.26785842095990575,0.4042640873874125,0.2775228096901874,0.7984877668379908,
 0.8801018119719864,0.5111233849907527,0.6644794825149891,0.3502820294502542,
 0.8605160999849972,0.46564659867518077,0.4337623237518996,0.7336235287680503,
 0.10402031055956329,0.5753891457639242,0.22462415144017234,
 0.24219542581801734,0.4689788117426261,0.23921866479886011,
 0.05654452402424126,0.11865702855987059,0.5964677051672539,
 0.062295234261677046,0.2668383924615896,0.4501448418809547,0.72836231272001,
 0.8906539976047401,0.19582080860166096,0.8093302841743503,0.7847089310503921,
 0.46824561611311444,0.7893573286731527,0.20626592811493238,
 0.28797381233073505,0.29237472428735056,0.17316539334140224,
 0.6265331912092897,0.07265533549134451,0.270328553826538,0.8500819113374253,
 0.8370669009725751,0.04570317213181352,0.885434103709621,0.5507179725709971,
 0.33833979043128026,0.1585801856347806,0.16302994846968022,
 0.05117614706457485,0.0522825967857351,0.2618048886992803,0.7746931220599279,
 0.7341727721635418,0.7574523382278879,0.8637150808755161,0.26934629155805534,
 0.5170981358970863,0.08690352844816973,0.4760359894952343,
 0.03298347537059354,0.9725958687744236,0.5181595819365502,
 0.42191797326195335,0.02138201176065757,0.9214748254429586,
 0.7651616839787629,0.5627962990604345,0.5666763243389028,
 0.027509695894649355,0.36578919032692636,0.5985172414584077,
 0.03760510130320871,0.7817670578541788,0.4303885665019139,
 0.34834512492779024,0.17466184789936923,0.09479532643253241,
 0.2011654706231043,0.36582704369143526,0.29700773394467916,
 0.7816425119756736,0.6063468958456859,0.952382701545917,0.4382193279383144,
 0.06811001274687178,0.4505931411364248,0.10672660682379576,
 0.7182828008382213,0.5431324490470578,0.11038053351395849,0.8286365093339056,
 0.31363264521900447,0.8938764869515401,0.38453630921055526,
 0.4275804102498987,0.4453556541488519,0.9789515160918855,0.7599567547299653,
 0.5424795612702421,0.48087059964047296,0.024267568787735727,
 0.4363804000977102,0.27726787322709145,0.07636477642713446,
 0.6183967685443471,0.9024549310503154,0.4204127979616975,0.06537441304029501,
 0.9932346967480694,0.689247219359818,0.09738614494120668,0.2617660302144724,
 0.04280501446045337,0.2118442567235711,0.07626037312213141,
 0.8874854161640429,0.6322860085308837,0.6573554707075212,0.5415180687845,
 0.6985934972131734,0.821270678644543,0.30104861878516154,0.9295792230235955,
 0.2459166089025665,0.327321386476082,0.316861986246315,0.25025225647469473,
 0.7246852981908455,0.9496244570298814,0.49597708296317267,0.861824547693591,
 0.14174063522961722,0.07650433927159295,0.3921056841487982,
 0.08037713351852305,0.503796332965471,0.25831740738465647,0.9413078568007781,
 0.8053847743230478,0.5941091759761464,0.8774078583717129,0.46330105756646145,
 0.49623462410294017,0.32137395843070693,0.373498511363468,0.5330778447090505,
 0.7294226845981319,0.5025783108033456,0.5778558889826434,0.8112038211976675,
 0.8800416312188174,0.2741796319733283,0.26575324616759644,0.7897192682570298,
 0.05506052727094901,0.2494181453947768,0.7235647275224187,0.3492375032222721,
 0.5566867115463916,0.26692695075488326,0.8493153600924488,0.4756888428667416,
 0.9973211038359278,0.6646148069483498,0.07623676003880298,0.96967922955895,
 0.3910984991488753,0.4524702954222718,0.4918564279366794,0.35859197259685627,
 0.15139200367788486,0.20299476867377764,0.6415513767743748,
 0.5923891347617326,0.9375557871681053,0.037833231182419214,
 0.5422284900465677,0.2874653224065531,0.7087510186966444,0.22299435146561286,
 0.35569869282518174,0.8470077461295552,0.9951789092945977,0.8663074315284921,
 0.43663129408888146,0.8868761546945176,0.43866273385114485,
 0.45204661692954606,0.7370939468246062,0.17762655650156578,0.817337745705911,
 0.5041674366428074,0.5235120413727063,0.16809486438975896,0.3230940423642491,
 0.6203970911119543,0.31692472859002696,0.7607640623015752,0.4454588134080526,
 0.5419000084688856,0.2910961903310231,0.2705936269091611,0.2454682543265425,
 0.6267189275655194,0.09418405978680555,0.28133876007667835,
 0.5417327751209782,0.9703417021810776,0.1581000878434331,0.8616705171298874,
 0.31530135886283195,0.3460776837609407,0.6794044086364904,0.6767999502445079,
 0.47688557652161556,0.9630758259933916,0.1733078600555945,
 0.05515172873614582,0.47896915361920955,0.6602760540683968,
 0.2768531314079783,0.6482158974255806,0.12022905766265446,0.400330142968077,
 0.7839793716207266,0.14440496299774708,0.5278323745787864,0.619786255382963,
 0.502054544369546,0.8516078072531441,0.18622888489173361,0.04815957922381908,
 0.6529632730828775,0.9938559535811384,0.023940769611556445,
 0.2871325315262917,0.5097639721878992,0.041431802274348506,
 0.4556327418588506,0.11902633629299142,0.8525078871705788,0.3513089446410299,
 0.6366555049936651,0.19409259277901214,0.14365772872301918,
 0.33283005256119613,0.18698509613875947,0.6060366048934782,
 0.001112463378593942,0.23817457349465543,0.7954198511777566,
 0.08355531494319601,0.36582225604727925,0.3481337428400889,
 0.16928971912785362,0.005278296547517103,0.24549342429901733,
 0.8655925750425044,0.6560907439787946,0.3668973436507925,0.23737144496082085,
 0.655036526288396,0.05312244859444193,0.6356361717295396,0.9614423969374276,
 0.5058704522952374,0.7366643736470453,0.01428163871576904,
 0.38905808342219794,0.7795603069416137,0.9989169509720109,0.8422424160647871,
 0.9032587960424799,0.4509502754294845,0.21038512012044985,0.7016337707944149,
 0.9007322363982369,0.9761149382922744,0.7608162422230863,0.05264024578732285,
 0.9591847051397924,0.12711505341810492,0.07218869092243185,
 0.8298683285980305,0.9025372938338672,0.09314176168714561,
 0.25796798814192046,0.24440638830880335,0.9390920147814152,
 0.7046551142231716,0.04316904033164082,0.9855391202241663,
 0.07893086775306679,0.60547528466014,0.6348169151743399,0.33011099488355256,
 0.3381726174924805,0.2712015381159718,0.2901421349878833,0.809585122760951,
 0.9795284403862404,0.20266468212702504,0.7558064214548678,0.969733977711792,
 0.11941919270153645,0.7769249935796758,0.8483654821814757,0.5020815968950668,
 0.043525998050761294,0.0936983084429025,0.3713876068415589,
 0.4090224991577739,0.7097225518398496,0.37495009325440765,0.2423932136665019,
 0.057057780881619835,0.9008789769825166,0.24213660722745678,
 0.18628798096249066,0.10152234737494825,0.48092918567566956,
 0.576766092918339,0.81201111755782,0.44683010347101004,0.6945645639604046,
 0.8676695774842275,0.1495796843982261,0.2782244585613808,0.387097938887062,
 0.705807541379171,0.4256651817957592,0.2736404966914775,0.49880307036166316,
 0.8486322322394781,0.6093471307672628,0.5041252103856523,0.9902130663416687,
 0.6627337489851303,0.6838617940277709,0.6060417834485261,0.6357684254642052,
 0.23503421324105767,0.5642815221554326,0.5872081944644703,0.8198749785657462,
 0.48968874685322306,0.06393106118984082,0.41327961454634776,
 0.43984828424756606,0.24707714368930023,0.09730436003917808,
 0.8125396639191302,0.8065406746918773,0.22881543913012115,
 0.09608338124162652,0.7318110026299329,0.7812436245721238,
 0.16342514938676567,0.9395490589521807,0.7931963766332658,
 0.17853214300110376,0.7056008920237696,0.7763286500434601,
 0.11998889393127765,0.9141863008356402,0.047988914620262224,
 0.7502759184728393,0.2818960112549832,0.30976789378887515,0.5774241918239911,
 0.04893142176276577,0.4262970700470544,0.07981327176410213,
 0.8719095720683834,0.9877040235602641,0.8105732324122474,0.8100196955726388,
 0.2630018843158719,0.3302849261390216,0.8091844196466883,0.590472001262077,
 0.04520634346516439,0.7132918705728619,0.6718740379609646,0.6910649230686217,
 0.45232133071170266,0.7812163301395779,0.5410494364616592,0.8298023752328442,
 0.6057705523814295,0.27496751908926254,0.1274820188371545,0.996118158522054,
 0.2782089655558052,0.34470679117078185,0.34115869110704033,
 0.24829436297046292,0.49583577993524197,0.15868122158682507,
 0.8625617089401054,0.9543520906868198,0.6711460221026905,0.57076819542039,
 0.28226966759352,0.8526919390136,0.9178454445102258,0.14790183641998367,
 0.9636399981884136,0.5769278615857037,0.1970324693649722,0.4524656704175858,
 0.8095174716444233,0.04363853704036169,0.6358213409976095,
 0.046367622650202556,0.2757675835708724,0.013241381676121877,
 0.42152202813804085,0.595241675762621,0.4838725873525359,0.3607776187651266,
 0.5160629057069643,0.7736584136339075,0.3884414295242955,0.5020237918520388,
 0.4645018941818444,0.3597446326033389,0.5157428996995961,0.7528726092128182,
 0.25252491715535796,0.4832694427564409,0.5282934078841207,0.7301278989921178,
 0.166971611628418,0.11188215589416173,0.004580633916102084,
 0.5013149820943993,0.3490635627348908,0.47646861317454525,
 0.19891820262920346,0.053199099019131,0.1637133352765634,0.8954983271770126,
 0.38794770870380835,0.7854253630843206,0.9600552919628086,0.8278627527205386,
 0.1756074864259548,0.9241806750275405,0.4377963732788376,0.6859709954223949,
 0.5370657770554338,0.16527766994235693,2.371097369567865e-4,
 0.020469044272903303,0.08265458284911653,0.7653222315735164,
 0.22621372016303298,0.15874903655687844,0.9738527398414987,
 0.7871832965954922,0.021534550931443297,0.4288815423284734,
 0.002363374126029827,0.10581035625272905,0.7423973447024081,
 0.8283935031932287,0.7385148658812531,0.8180426427103854,0.9512893987026582,
 0.061219210797399115,0.6552496204729241,0.5463218832860036,
 0.35999033555085447,0.6834669380246556,0.1179527150296662,0.896546107718259,
 0.4383679324926315,0.8576621934566527,0.3613008556338121,0.43768103303830985,
 0.20988596202400311,0.7954613128662281,0.07802829025042957,
 0.055607750611936235,0.967011377845062,0.19734590159076393,
 0.6718057107462694,0.12521199011150652,0.848059718851171,0.2467554407284691,
 0.12220800337539262,0.44912172950104445,0.5790181163134037,
 0.3724905140542417,0.7162565984832485,0.8017902749031863,0.05505512211446284,
 0.8778239424064498,0.5088163739477383,0.3005853755396749,0.3711298866366989,
 0.28878017498174136,0.7269871840804351,0.01709019262595801,
 0.4635575822991511,0.501004422263052,0.6117424427333853,0.7373030530853688,
 0.017147028071393122,0.9064034649830943,0.3477574670780039,0.163389675549106,
 0.6384289744605026,0.022878526862176285,0.16305355598606575,
 0.29942755333534954,0.5893038189243782,0.8018489304220906,
 0.16494137887680038,0.7649902848511059,0.694758018307147,0.5907191789782023,
 0.40363278748445597,0.3161250970869617,0.4932302517401461,0.2546419945639714,
 0.44480036414289825,0.5140278884830727,0.6592137871261,0.9934595387007307,
 0.8590258478667456,0.6257668461433092,0.29377419077192635,
 0.36508774686322076,0.5025204053639295,0.050600693539947006,
 0.7873457180423191,0.20302439637330283,0.035956479402385555,
 0.39433711404044425,0.3680008409903197,0.6530248485649809,0.5203369516208697,
 0.926377278084759,0.6447666268479098,0.8273610716289557,0.7219629139653556,
 0.32811938170198884,0.20298121415951087,0.30326545987733733,
 0.633569780852262,0.4394507616401313,0.3818738128803558,0.07675197404397593,
 0.5574841581119684,0.47608069973703726,0.934119778726679,0.6637663233810758,
 0.35895434007166516,0.28124740590130404,0.8421879740608269,
 0.1166596107927087,0.5677076782487003,0.9521299711827382,0.75502502064349,
 0.14613789688461254,0.49195695101590897,0.9013440226998227,
 0.8551128370393338,0.07495941213270507,0.7193912071575076,0.9370844221208273,
 0.6281497078311306,0.3536022173244997,0.9723486285715888,0.46374808211423646,
 0.4170833084971213,0.4444642804608919,0.8325950824125119,0.0229847226054396,
 0.42627829307779586,0.7940343675437002,0.17929828482159782,
 0.3269185481943373,0.33683396414807487,0.1555244753705195,0.3068708779080527,
 0.9439186851777193,0.16863365307594935,0.13646400210619525,
 0.6832659351082941,0.456460769911763,0.8995387442293407,0.15908235916646685,
 0.11792511401542616,0.19876686925249754,0.8712891034475545,
 0.9246801114253804,0.529606094624222,0.7165101740152056,0.8048353058067212,
 0.8384855169497788,0.15220287850617364,0.17650081259624584,
 0.4386026084593265,0.7010208908838952,0.0921730845261568,0.14814580554442391,
 0.26548946355296577,0.1036639884223477,0.831636757226264,0.12962330267900446,
 0.0249484339005962,0.42822314431418484,0.3993730386634391,
 0.44824625185948563,0.3977338753381011,0.607218528325256,0.2474866671945588,
 0.5661632986977645,0.9149812883775419,0.45924326886651223,0.9318860751475376,
 0.606204144874034,0.5139071235263959,0.9052513865007735,0.3338235415697628,
 0.4942181862137275,0.5668190853978916,0.38909945187798667,0.9330777686136507,
 0.8868376219620888,0.43646147321453665,0.23202105969893405,
 0.17297849663770826,0.29310636500882425,0.05733473398278699,
 0.4564505491684909,0.436608029275685,0.44304969566493013,0.13648366666294443,
 0.981944012281715,0.012835582121076272,0.03462276973294354,
 0.7187893024294489,0.9362795377468089,0.5719387055622702,0.9466909159609078,
 0.8665294964550341,0.5056861370264171,0.8474861769032742,0.862884303366102,
 0.0014743956933808144,0.32210778897548054,0.9665603168515746,
 0.698101265413002,0.11559020196875436,0.4224381335572689,0.7905276662593359,
 0.0569451926007154,0.5281937068619941,0.31158880167902414,
 0.03691088151781685,0.009938485222432902,0.47588616329275557,
 0.2088537202261862,0.5608436454892944,0.33085198297884366,0.6436663137806797,
 0.8059830736936957,0.2439562924146632,0.2082002488067627,0.9940382505203349,
 0.19449293035941306,0.04313410577160803,0.4683068516793616,
 0.42710989812712674,0.25102108721716154,0.5075902180561095,
 0.39569626156794435,0.1199456116396469,0.28699041348187715,
 0.008960144465133935,0.303588163016405,0.48083452166604357,
 0.7286740115227307,0.3494115573829444,0.8715986021864004,0.1526170052478153,
 0.8702089821639754,0.7141632181459979,0.5834734011981105,0.1729067409257905,
 0.7710914688223739,0.70538929066145,0.0563954909281037,0.5711516298495696,
 0.5275996848696337,0.5354108609244129,0.9013507415758955,0.7135485960292134,
 0.44973305410669484,0.43190375457238983,0.20250176852437296,
 0.2058329395165479,0.42826909401737734,0.287055507043539,0.8989993876518396,
 0.4873536074095782,0.6507367071342638,0.9836245780058719,0.21906371091222265,
 0.8265021084274619,0.4926900538739776,0.31206466091114693,0.5379176891368448,
 0.42755542277898884,0.7332012612777558,0.3249168220830363,0.6064949747707318,
 0.589824007829882,0.73696414349152,0.6886742275759358,0.3888813002637391,
 0.017017722548098968,0.8021765438528132,0.16564011681737767,
 0.8983990832087949,0.7251854837189541,0.11018868076631494,
 0.15906511039662463,0.9954322216293704,0.9767404195004822,
 0.11989135821244079,0.2958343619135394,0.7656073250892811,
 0.35602948833558523,0.2948723718058702,0.6866926436071344,0.434343900140193,
 0.7890034876147156,0.19358445737671737,0.8189605075247121,0.5842367603743892,
 0.2266256242211111,0.3646671611557585,0.15132408305007994,0.1563419454267052,
 0.36854426315690225,0.46983602774355215,0.6236383197430079,
 0.3010203594626707,0.8209475841250965,0.9382379938557766,0.9020561755866048,
 0.3769678314021967,0.23505420036459324,0.2312428686412763,0.4969198262593947,
 0.658924578641204,0.9552137586498191,0.9138832699605378,0.2106846053510949,
 0.03664268387556435,0.23811889942193853,0.8061560258096253,
 0.38777956539805336,0.08578368317618379,0.49744620541702944,
 0.27800904646452573,0.47312741671654335,0.08329234397549534,
 0.8002538903911293,0.9909121007685451,0.45964544701947174,0.5033285886447724,
 0.3171371278901878,0.4119920212373439,0.5122529264504588,0.23451019704908826,
 0.7010766467962892,0.20593788384163236,0.5279622289447744,0.6589050201616317,
 0.7455584516097329,0.3206155022815751,0.656935630875924,0.14620314342604912,
 0.5847213779430057,0.8277045020397117,0.2644809750609821,
 0.012849367028021974,0.6855620146784768,0.6196561739760494,
 0.2708880436986426,0.9746036496020204,0.025481099396416074,
 0.33998510056353504,0.9606284122901944,0.7328849850685569,
 0.17294751768289385,0.41753115508797223,0.5544668692231772,
 0.21818058777157257,0.45839848106061476,0.11575414744923407,
 0.6304885431190083,0.7389045989980193,0.499137694033855,0.9451665500210631,
 0.06331264751749943,0.7036195134858132,0.9112918044640963,0.8174325022615904,
 0.8508416666909292,0.6873805005528816,0.34246638089686654,0.8138347102912533,
 0.28138291909091384,0.9675595258427295,0.44737290664285234,
 0.1836154295661292,0.5264389936942636,0.2285296319575727,0.5811444294500278,
 0.7528204690850789,0.5530901831543212,0.7685362054445178,0.9239537240869772,
 0.7961580765498465,0.4464980981262827,0.19257969847048184,0.9963198593967856,
 0.56708663743327,0.9293462264448755,0.8793953831788873,0.7175625979261544,
 0.4729266428371327,0.9519848580141074,0.5964797758241194,
 0.043812742098062296,0.5301189732165859,0.4593025542951462,0.183089399590733,
 0.2860747114406451,0.5972279642222812,0.940022122480302,0.41293310773131553,
 0.963831598211506,0.30810098141012743,0.32823624575048993,
 0.19522357992063388,0.1684468797223504,0.9913519014116792,0.3269686617300991,
 0.17900942801916808,0.20947371872828713,0.5425371594609112,
 0.05698569850208701,0.3358186843607939,0.055617563914130086,
 0.4279446691558688,0.548076970648116,0.9238213152375342,0.6288452986037769,
 0.8771594500620042,0.8017749768625977,0.00588240200582868,0.6311692199602529,
 0.051272694010214614,0.5932587526517897,0.24474963790452087,
 0.33808170889392775,0.4507422723060647,0.11712896869499545,0.457098930627199,
 0.36782662676132893,0.7203568133079176,0.4001295642631575,0.794177764261685,
 0.9520826136600284,0.1597369068214951,0.06963710939172707,0.3328271215058155,
 0.45654858595927317,0.007324777594307941,0.11909069369211545,
 0.050678012297205366,0.34454060663561015,0.12975484908723955,
 0.2836112706646202,0.8242611821784775,0.3458527876056041,0.12323258474849297,
 0.6018049559489149,0.7291844516077314,0.9525259417305822,0.5136850523431209,
 0.6195543093590943,0.9751709701427552,0.25325479616652846,
 0.35525114364694477,0.49269278893423807,0.306111115783943,
 0.37674088815460016,0.7918733853203808,0.6676338960078075,0.2598926555310268,
 0.030381147913268958,0.68753929121403,0.7031867936588305,0.9721323104955446,
 0.564987701285557,0.3750780967184718,0.9599237286854974,0.8958128749907058,
 0.5373550006343653,0.12391159600588575,0.7583570577697506,0.5030919594492813,
 0.9731363381287467,0.08110351077346722,0.2680519244313353,0.9893907772171907,
 0.9759543286802707,0.2900394401491354,0.912024127402697,0.7124025867055749,
 0.4702477320320205,0.40122006684145906,0.18750431746665552,
 0.5146445465386652,0.029686635706160036,0.6687441993255491,
 0.7090967786325196,0.5657239039568666,0.18114251995890007,0.9480896513122962,
 0.05521119855281864,0.4386076697655301,0.8384554207463465,
 0.11347346009785708,0.9047844827243976,0.3205486997611675,0.7847121822945015,
 0.06179275401643902,0.4208684820889864,0.015237633950897722,
 0.1743386127278117,0.22691972367134172,0.053278522071474876,
 0.8079804306975041,0.21657993185781343,0.8148225367163712,0.0930565363305631,
 0.2722527709182795,0.17491260300428702,0.40162486604826186,
 0.6936468328482481,0.03310191394135864,0.6945835595857069,
 0.03820935070769482,0.0011789250277611796,0.7560233756565872,
 0.6777477455522057,0.38356999565069105,0.39755817327283216,0.635747509621634,
 0.20125004280141434,0.5213396461379625,0.45445567631519757,
 0.4327734718957591,0.5338629741854828,0.591179004646255,0.6303973277308039,
 0.3618350303901715,0.26477238469916753,0.8629722938811115,0.7821156424998311,
 0.66159107790416,0.09859452468434915,0.03179888625207228,0.5480100424708789,
 0.6645017501413355,0.033051094232617206,0.2416264392361238,
 0.3752948422744322,0.13440979891797078,0.6518428105607212,
 0.22523186582970534,0.6736711953509282,0.527951691910127,0.982178460639247,
 0.2339652509428498,0.9961623452734841,0.7330956684396469,0.5843453995433272,
 0.872755124471264,0.6063652825842784,0.6909560880622627,0.9932114515876244,
 0.9564382709545479,0.8294328223347412,0.8627091386075149,0.8047458880318733,
 0.4170900965463664,0.48178362764259086,0.5226338703538397,0.9472453147300398,
 0.19803558128451315,0.45861048745161914,0.36472786573424454,
 0.3429846365381747,0.9097498353067198,0.5576593824413847,0.6266563356017223,
 0.8080601144604458,0.5709874231075639,0.8388001941909059,0.9132390400632758,
 0.38432295314092146,0.31059400528118664,0.08405831492743099,
 0.6424723982245071,0.9514812911592653,0.5761658568387088,0.18396294364644206,
 0.2216326889085637,0.16045328514056412,0.05772148230821861,
 0.028571141880174733,0.7690534760589423,0.1999913408968761,
 0.13953531711618739,0.6492455978236498,0.745904978998948,0.6417632373438776,
 0.4133011957961922,0.3400736984781302,0.10451957640040133,
 0.22102955996186635,0.6532463729002563,0.1503477244323721,
 0.10098507075061913,0.3180613593711672,0.4580335421430569,
 0.09612672275644762,0.8272115085153935,0.49809873607737454,
 0.08185761169964556,0.5762303181567039,0.6450331935046467,
 0.018454968501844804,0.4444341706885947,0.49859988089878504,
 0.3899273357770586,0.4684579041614223,0.5777488783447431,0.7056450678545225,
 0.546274436760908,0.8786402320259677,0.564641270185922,0.6656296776585188,
 0.43262080488131305,0.8866542527796916,0.6904958942448327,
 0.057464314413621764,0.3288393526553529,0.711188681102842,0.9309993832716941,
 0.8699932516987431,0.10607788564518639,0.34429258710528066,
 0.16799270474564632,0.29570762985322985,0.4159444948129586,
 0.7194905929674632,0.6573361183996724,0.11084906412345874,
 0.22326693628641436,0.13329705422026994,0.7895516022477,0.3892731281203352,
 0.7767923375156657,0.5139828196093752,0.708322809021427,0.7492043645223192,
 0.5182480760496964,0.08617620845249097,0.4455382060378946,0.9478222521773376,
 0.0996804455964011,0.08717216690372698,0.7862399027506664,
 0.43571166870213307,0.45931922833599603,0.4999634932587603,
 0.26849900716777697,0.11914627385035448,0.2767071015433772,
 0.7347227392897039,0.5957736217055167,0.9468923665532722,0.5294924388887968,
 0.22260779445864887,0.17358110182753062,0.9435457861376627,
 0.6255775421862434,0.6527793145261326,0.7840844800590717,0.0880141830585659,
 0.4756851047284598,0.7656487885612092,0.8267581746582991,0.29689470996444856,
 0.43048805286538816,0.2513375010460277,0.4874068769000959,0.4904413331566795,
 0.3541641616110889,0.5117594027082898,0.6471197648620153,0.6833294248574854,
 0.6955597761015325,0.0973209801976972,0.0637606379479263,0.753433355781595,
 0.4557536809596836,0.5838854517434421,0.889533286570785,0.3280025098470525,
 0.8933953832378018,0.3173065747653999,0.4084442116350544,0.8874178621872671,
 0.2677108388616368,0.3853237461686715,0.24023664917220233,0.6282242086328426,
 0.9265511452619981,0.5355044897883013,0.7939622348198924,0.09305506246262496,
 0.30024739236789366,0.9384015813491946,0.8806908281172086,0.3100608233179174,
 0.76681598070052,0.5084869600585734,0.603558563627903,0.6638877657986283,
 0.1341917422546295,0.6911517111137322,0.059292541474811555,
 0.29633498411080095,0.9849843583052862,0.5529258284967602,0.9133020723781041,
 0.37905007447715755,0.8108940413691121,0.785509730159431,0.8896189459671564,
 0.024072228890775205,0.5184756042549361,0.8670081519565978,
 0.2983067890081553,0.6413715422341595,0.4092564546480164,0.7249972411552699,
 0.26854942143486404,0.7862787707334041,0.11553167650740015,0.692140874420038,
 0.21914103539540597,0.49868986342975186,0.8369921014267856,
 0.9537143607778897,0.25661142734151654,0.6148835819506925,
 0.10399987461934801,0.9255607569255613,0.16312945298987847,
 0.21886076262480592,0.7737343069161907,0.40158905927619226,
 0.25316404607610155,0.9127907487133006,0.8128070631926485,
 0.20870045202479093,0.4834664135638448,0.2335608486901588,0.5179405156549932,
 0.8269488491083771,0.8518058563927339,0.5292025400602703,0.598355763448528,
 0.7729597202116858,0.5819998656901658,0.14641025578500688,
 0.04574952433768331,0.5531272765829554,0.0418600305997463,
 0.16375047923266395,0.5061154117881358,0.021979913451937394,
 0.8188988917395972,0.7394608374786434,0.8214023551988355,0.5017762654712258,
 0.25112092116639206,0.3786559236029873,0.7258390091338882,
 0.43263333110684576,0.5015504375951303,0.5431616019094361,0.7460893965019528,
 6.490626837581948e-4,0.24961894507559412,0.09229246081830167,
 0.1067028670561655,0.5014402209413913,0.7280583825759015,0.9235381971279137,
 0.25480385852363363,0.05082474044188656,0.23687016466982969,
 0.7283968987289882,0.032555124831590065,0.8857933031826174,
 0.8530519318345885,0.07965740025863766,0.4120330683534912,
 0.054008500972801965,0.9558788089409426,0.2014410377029775,
 0.8787753749750564,0.5774532542272769,0.22167496248515983,0.917935891507578,
 0.2932679820281302,0.6700564581960549,0.14755034901309316,
 0.15002272362374047,0.9139055791092998,0.975100585075539,0.49040714730535495,
 0.7400740181411362,0.8688386350290529,0.014661096568674092,
 0.9648258519375814,0.9125313936853164,0.9124675265571012,0.4705010110560026,
 0.9237713362694135,0.3011758948764647,0.8158299134068074,0.1517593596970711,
 0.07118021865630564,0.8258701700447654,0.9755327541761704,0.4868417043069897,
 0.5281725724466815,0.4664225691521986,0.6349875178377504,0.44468674175387435,
 0.48027790919563773,0.24313745854198976,0.0013931124594412925,
 0.9442230862985068,0.40246410318991466,0.16585712438120837,
 0.5484254142882382,0.4130449200638385,0.05318259182122176,0.559334255570842,
 0.6896462304794388,0.06405409166961773,0.4296907175201743,
 0.36416364282774427,0.021323707264204628,0.16100442089068046,
 0.8219330687557521,0.6555321997278142,0.3446636051159845,0.7835404898454066,
 0.0761534759052036,0.06639853461368284,0.1689029103627796,0.2418450076522698,
 0.42501009787707333,0.5747408540109027,0.0581384936967555,
 0.37053547607970216,0.6406187557868034,0.1773776754573133,0.9860968691840912,
 0.16603485952546126,0.7689974695613881,0.9509796203459003,
 0.14179978695266993,0.5893803653096967,0.19016487063364063,
 0.6586187377709464,0.40274588482784124,0.3410506050599966,
 0.24245577055392786,0.7822172029099475,0.2962801974509863,0.4979967409824616,
 0.6651446432368254,0.5497313619230366,0.3755966066143137,0.6788181690856572,
 0.12030422861963008,0.9673847347591046,0.7165722080139927,0.3995854931359092,
 0.01268036512218862,0.7795710401625737,0.38711299831260804,
 0.47715765693467005,0.9582698380110967,0.8250475738677989,0.4221463545101375,
 0.7162566324073055,0.8291454721735966,0.07535744475528294,
 0.17828800461304772,0.020366637533599463,0.9952084125926617,
 0.049189558585523985,0.44189600304811716,0.15501153912318844,
 0.5427375668736486,0.29892857140207507,0.85398300439216,0.6028567183711208,
 0.19184127895125158,0.054198566884704036,0.010994027207162582,
 0.9452355172400237,0.12936518160091892,0.7694804246128327,0.2980170009376757,
 0.551564769283084,0.5597575448176347,0.04639354151462649,0.6807408754753883,
 0.27735651885858525,0.013621532313230156,0.24631113446074193,
 0.6881888777027945,0.4455601075703226,0.20458885658833237,
 0.21910423190829043,0.9344191586384051,0.1858255942652608,0.9439078609837909,
 0.60728114097186,0.7293467394120449,0.8974350579926063,0.3701446622965171,
 0.5933758753278785,0.8298606954878585,0.5539589724452672,0.27401661793214627,
 0.20962092289036538,0.9322337863762477,0.5169162897582531,0.5205731408167791,
 0.7445423533269167,0.5969574419371341,0.05841611746583264,
 0.41380903810066794,0.017730107930434613,0.6769637130231894,
 0.5670978249320096,0.648724347244528,0.8310143742636007,0.3467296874150162,
 0.774509071855355,0.37697011799061075,0.06315523963768055,0.5944790101044192,
 0.10279025384856755,0.9095298885444859,0.8371097712743923,0.6864908751264635,
 0.9875326249102129,0.55971079934986,0.6261744912032873,0.695848044699219,
 0.4071312115352733,0.8208939134616411,0.35687435340760776,
 0.28432880527417703,0.2635793545527232,0.21567042675405745,
 0.5044095257205905,0.9862266990644637,0.8367599330726934,0.4503526607500892,
 0.5541595417283162,0.34298797982363893,0.33620706181562954,
 0.7182420339590717,0.9087090076530869,0.14847687386320774,0.8325601686300255,
 0.42033336365888074,0.1217144340270302,0.22713002977924868,
 0.2689161894807758,0.10524210810680523,0.9239121350200673,
 0.14768867623794235,0.19636602908716982,0.9754544881790383,
 0.4549365447139804,0.9638020074126121,0.45559607104562394,0.9533403963336219,
 0.3605837534477404,0.048998323504444996,0.6995521611412877,
 0.6194723515463707,0.5160271855949843,0.3790449066947762,0.8130571311937754,
 0.7909941049419862,0.7942526008206252,0.5501971877486548,0.7839411149946531,
 0.2574660765828142,0.6897367305747417,0.613298754887446,0.10083392920184475,
 0.42595582053994774,0.8282792175571998,0.5246396148615748,
 0.23949462149167833,0.6082920282490965,0.3615062859955942,0.4994670253954492,
 0.8429572676445816,0.07888190831681519,0.6662934687794235,0.1191053344104488,
 0.7961208036652342,0.150561967559093,0.4691302275820485,0.29624697279461965,
 0.6715577436754003,0.7155259427035946,0.6586798864334386,0.662975210385981,
 0.27642763998851116,0.4616684407213866,0.7579921268620472,0.2880942918926208,
 0.24996615011872603,0.5391022805120547,0.07833080766814216,
 0.13774924524356835,0.6186589913037052,0.940976486861858,0.7823667419889161,
 0.7172483511088297,0.36511938214128203,0.9467712255117722,0.8764123143859623,
 0.40130701136660485,0.9924884601046869,0.3597782400697306,0.728459537700032,
 0.48129887382781866,0.7462675941398271,0.8681208524022664,0.9062014999006588,
 0.08298029852308342,0.17648928029443567,0.3593647706279989,
 0.26240307276547836,0.5290206328777799,0.5359207083137143,0.7790570149065535,
 0.49819832198130465,0.79010593217938,0.4587659283378178,0.05318903272684139,
 0.048109141586906046,0.4154272123931215,0.008326914532664875,
 0.15351418856785026,0.7190222826054402,0.087173962180085,0.3939144112484235,
 0.7182441707052225,0.6645761131956348,0.7099950689453146,0.7489898990191204,
 0.6505441741280218,0.12071128353811877,0.2686782485568848,0.2268114268462167,
 0.9257446803006911,0.9182920338155601,0.02734556582006109,
 0.031275455454750745,0.18137308657567275,0.3673099783060264,
 0.47167480556088714,0.9428964445202042,0.7246648000928383,0.9615897780990883,
 0.42042912199200055,0.23599333472482076,0.10772244995430147,
 0.7364276504204963,0.1663663084682928,0.9531869455048569,
 0.0031401668396370264,0.659845915842296,0.3446182987203872,
 0.40065536379170086,0.27818396002264767,0.9096290096527643,
 0.12483522522947599,0.7640158431957471,0.47206924396353434,
 0.7705234851715526,0.8570033384096365,0.782793821054418,0.5525485364371198,
 0.9124090904757518,0.6286248402939016,0.9171966468973326,0.6056095478673554,
 0.5362260087917246,0.6651368399007135,0.5553695215579448,
 0.008958253413656303,0.467174988488964,0.06885866982216005,
 0.8706307995707133,0.9680811039889509,0.610640200128005,0.6184777042451394,
 0.14130544099599107,0.11631594122942424,0.5476296520208856,
 0.6033402019860569,0.17504220836328865,0.7491780000945345,
 0.13424231514183316,0.26173354580795394,0.5978124015553546,
 0.3014826972945851,0.2878319217454539,0.9486079926605644,0.48286796261805165,
 0.6374965253711093,0.8981473636823675,0.42493060190762755,
 0.10796431226176018,0.34667479274174884,0.3784097926647061,
 0.25827248221012844,0.9137567105910281,0.4634564283123893,
 0.03260987520530134,0.7604462748953347,0.15415416632239753,
 0.1150927929701836,0.43479546669268354,0.8286019602154219,0.4783875497450517,
 0.5131794435305654,0.19639048218479693,0.47397260449795675,
 0.5640127143673279,0.43388406567535753,0.8773654440390484,
 0.30027471011424167,0.7253362838653191,0.7701732718859655,
 0.23512418953881287,0.9355629920171258,0.09595190012221932,
 0.49228518189789794,0.36325683149045174,0.927401187476893,0.6383955400780444,
 0.3256533576309355,0.05268964797205722,0.6197339069129911,
 0.0025880322371902853,0.6913832187749576,0.9436628084969954,
 0.3050293979643215,0.4138749243800466,0.9099805227833981,0.7374108651929098,
 0.03498130691133705,0.861810895975549,0.025567643852514532,
 0.5455259129993633,0.856020572518204,0.6838632884966965,0.3355190583430421,
 0.6435458482850488,0.6871615736418605,0.8179103919272508,0.49637577654565745,
 0.46172570251919065,0.8674676350526958,0.6044609806693629,
 0.021844509949029867,0.8557476599023766,0.8153397148190956,
 0.9465195161023928,0.38829572457781536,0.6671988682139017,0.2688567943282537,
 0.990795012157253,0.8831518018757696,0.777875350828716,0.5541098131561685,
 0.20406967808358945,0.26806006134527605,0.5937270943377464,
 0.21661756599889392,0.36155166893485935,0.8949299044536824,
 0.24665183205728858,0.9688630887310805,0.7773266695319997,0.5899034167179799,
 0.018632478468904146,0.7626382767201068,0.30982880405925073,
 0.5139131858030429,0.7302834200520406,0.20469463234808238,0.7287876957652123,
 0.631636860093229,0.3786801905245101,0.9177764355779239,0.4965833850924295,
 0.022743209008199483,0.5675644637983778,0.690599111323873,0.6139048284923838,
 0.4420256504600031,0.7761696917062992,0.8916960713771008,0.2968112093663977,
 0.6400314838475448,0.034829205174709,0.5039982501770693,0.12945661370509554,
 0.5074709903471213,0.4224720357273499,0.6346572605420331,0.7109423940618816,
 0.22648162012106043,0.9934834310217908,0.7226053449452743,0.83773177004204,
 0.32895410021682536,0.2772030191584207,0.009997082749893527,
 0.781722903553841,0.3108212425826109,0.6128587766678895,
 0.0024551693558946264,0.31210707338394494,0.22622191526775093,
 0.05763590783515116,0.1866606833559279,0.3275400466738363,0.4515905838366112,
 0.8956220190435821,0.8936031024940341,0.24801226303184676,0.9688392207307994,
 0.2921847706535381,0.6075711208394051,0.5572601860815628,0.8426345365141219,
 0.40396339350607446,0.6352877775643142,0.3369497953111471,0.6205704940377589,
 0.4900282941358246,0.7231102859770417,0.7401694922932605,0.689214054053494,
 0.5853904093615239,0.2112602239407968,0.2555075568399223,0.9938262292184394,
 0.6303276998138366,0.7091210943494992,0.06110021766033158,0.7132060789984137,
 0.6310096437668339,0.22051729616405535,0.6003391427767752,
 0.04901145038753818,0.4139378916043671,0.4270142180571681,0.938228626289639,
 0.7350724995572191,0.35550394466338275,0.5542491380621603,0.2350852538356054,
 0.5276609672128958,0.26275879880320163,0.1464956650962248,0.8327654983689488,
 0.7223696902135969,0.16145939863105607,0.4466366783207468,
 0.27310193221541335,0.6824517673679803,0.620836469440321,0.16145034007955728,
 0.1483980831545615,0.6321067634741782,0.31113473643377054,0.3662502153876045,
 0.5075110753068671,0.08266820649164752,0.983306942198986,0.18952491225190182,
 0.4559841425223057,0.7823633033345708,0.6522834629725455,0.08972305837409145,
 0.8813659123492023,0.0576102064126488,0.7322668836622136,0.163976444151273,
 0.9800408368377129,0.2804820039431657,0.4875970367071335,0.07167618855908964,
 0.38705805799035786,0.8948548741663708,0.44333141731746073,
 0.5441958051955068,0.8824945339718522,0.017260656634066285,
 0.5626991481663686,0.6665263504270609,0.6873287031577135,0.8488306890156766,
 0.7284230636029015,0.6721935262801452,0.43810898168681434,
 0.10805890905016502,0.9850708305571197,0.22404653020173027,
 0.7349454691169052,0.24931143313854753,0.7416211546330462,0.5810781211048778,
 0.5192630420296526,0.9716414938131176,0.9315593759848295,0.560949684989186,
 0.6803155869801801,0.21553531086539657,0.1603547285786603,0.6604413656751502,
 0.015303803918477632,0.36610555969127667,0.5379000524099077,
 0.21464240307627935,0.1831441890696972,0.9544314290891849,0.9913369062314364,
 0.10839313151400387,0.3176247095358442,0.9514180968558189,0.3826839804434491,
 0.5695823331141247,0.8583723625700667,0.5285129036746943,0.6070722742699075,
 0.7078110433659728,0.5014631843514393,0.8525413816775572,0.9126187978095275,
 0.29946962755453044,0.3109306136770723,0.6748255595376781,0.3874334373523791,
 0.24981904673608124,0.619715665058438,0.6766709595382265,0.3377090418865082,
 0.33597520489603805,0.4569412827871616,0.9630184497615182,
 0.11261047469054386,0.06282942998995689,0.11920788881975741,
 0.34817790738769916,0.9865360252615887,0.4631361438530577,0.8554491037391845,
 0.6683665238350355,0.872595503414021,0.6785558236265025,0.981265549273442,
 0.5270944638579683,0.5954801686051849,0.06561800646097615,0.5482385046578688,
 0.2161139172488612,0.8332245305544925,0.971057061863485,
 0.0036980312422942863,0.3953755833037299,0.0031105766448203953,
 0.7616313129369782,0.19783153812432797,0.03605248835632868,
 0.8352842914026625,0.2588365459505848,0.1433444383782756,0.9297713861130815,
 0.5422388776903593,0.7659048542582187,0.3409647332774146,0.22747814198925254,
 0.06029471904556405,0.6753321842050384,0.36520639264257015,
 0.9223441684604073,0.8676677844507742,0.7419122828773766,0.14319073803914784,
 0.38863894922802944,0.4612269749564728,0.4392323162768872,
 0.19497744183497712,0.26407916747592386,0.057840688125749495,
 0.5468510302937345,0.8110470329555924,0.775245700905112,0.5838957636169487,
 0.15810793255113997,0.284448569704138,0.016484182619561194,
 0.7928630950340919,0.4960716725984239,0.32884920190045963,
 0.24218214166423369,0.054120393117416454,0.7989579495333111,
 0.9819929735603897,0.4505871328799026,0.6627911822125037,0.4110906121625617,
 0.6476530349900923,0.06475881055124943,0.9490499422856383,0.638989433770166,
 0.8349017658025085,0.7608006408094689,0.7107577700298041,0.6375213457172617,
 0.40399034785805654,0.9993378575284485,0.4910802032688316,
 0.49045296697312135,0.38035481631162504,0.5211943366963236,
 0.20387828107874917,0.36674393991005716,0.8794458841030567,
 0.6496901838784694,0.11118638855950469,0.2971418609451919,0.6298101273437785,
 0.9405986405803213,0.634252409860573,0.9108272991109346,0.4303782198358971,
 0.10855152099202736,0.6905989343435943,0.8930329726109331,
 0.47151908941941467,0.7053610578217676,0.7797145198999372,0.6156975829003353,
 0.5947169501703314,0.4500712920460168,0.19669085981842116,0.2273917675167536,
 0.5278279158382868,0.1411853981978345,0.8539809059303641,0.05556895869133982,
 0.6661736925631154,0.8015161512749492,0.37904048130059387,0.2182001987491694,
 0.5370352599099003,0.8214252448129196,0.08922385247539721,
 0.10139467165566729,0.8190238599461412,0.1482639526619941,0.860445286305173,
 0.2113345044305598,0.7597703130202025,0.6029741419874443,0.7479165633663879,
 0.3068745907370428,0.6447239811607074,0.31633627809924636,0.7928172942693282,
 0.3283380786736988,0.4678780806226639,0.6331846632272929,0.12325725225669737,
 0.11738725857022869,0.9838905846553578,0.46257048242166865,
 0.6275040762043744,0.2608993044584528,0.3657349557934867,0.9264298113455036,
 0.39153764228824506,0.22205863970135997,0.3335259114212731,
 0.9802036231581998,0.12840034791418897,0.825326912386122,0.6474498836577218,
 0.7693687224490718,0.8107057568957511,0.5561884373692011,0.5935609031693173,
 0.5373833579078773,0.5606185552861487,0.6312857081621192,0.07819083252390235,
 0.1165177262280368,0.6686714962841985,0.5997644018295525,0.5258077980928568,
 0.36250575172771404,0.7127371475329625,0.3386306332835327,
 0.42412441355990504,0.35595164388445744,0.527493699745353,0.9432378099170651,
 0.7622470882792959,0.882368959780752,0.36716444284913585,
 0.024186472679853188,0.07708016005007501,0.21721493075327503,
 0.9184282876979903,0.620358944243115,0.3119892135896368,0.9461587168768423,
 0.3480910464287993,0.4377658139501993,0.9654081495638533,0.583631202053104,
 0.0034536393173468793,0.6422773293116751,0.2056566625499926,
 0.9095988423480734,0.5428721047743261,0.8197648087403452,0.9761561526688108,
 0.8674329483332709,0.09109378341175844,0.20351033533296514,
 0.5889380209823398,0.1817370436507436,0.6948815086378577,0.935751061620549,
 0.3594892809655633,0.6640046188001197,0.08498009400573159,0.9981706619276748,
 0.5190213444670586,0.09152502092332082,0.4240187109838631,0.739934165041193,
 0.47312661366363495,0.9432297919631893,0.8059520699165321,0.9414122751598293,
 0.34067397403797983,0.5358369080863994,0.44640673481537196,
 0.2715217068833422,0.46607449381875954,0.4810081572540369,
 0.05613846674949885,0.506419126023966,0.1773953026168421,0.8958481899816916,
 0.7413348733184482,0.9103807189238426,0.8191832832342679,0.903001645627783,
 0.5131923553592224,0.4409380759859195,0.038964642141820494,
 0.5344192033599423,0.30058616058413046,0.2462165122747998,
 0.030377388137975858,0.3807012596774495,0.892085678689642,
 0.49003152459475285,0.9926353739159387,0.8775196760788054,
 0.22823036636899452,0.5355647631072337,0.5254977392120637,0.876266816924401,
 0.8915001789565912,0.7828405135760744,0.7027673792213875,0.3960480430088036,
 0.29939806641900657,0.4734296006332329,0.46032109937412347,
 0.8950752912770165,0.792448904913584,0.6675032102535943,0.05834184733140402,
 0.1979486418699239,0.6900448044735848,0.22239215087731135,
 0.35414875640652044,0.6544241914123629,0.4228284547736394,0.7344239117297168,
 0.004767207149606656,0.8009054675313928,0.18841636391296535,
 0.09250604789576244,0.9130132849978213,0.8139128794222426,
 0.44405629799609947,0.9859217371166337,0.9707588877618616,0.1903454520857778,
 0.224701855172041,0.9566819167865828,0.3916039421649009,0.0737754710578108,
 0.8288633591916332,0.4622928309125467,0.42858531226572094,0.8565353369976747,
 0.7554189433791926,0.6923291204053564,0.1280270697015209,0.5278446729014321,
 0.015701344653799576,0.07954025464206482,0.17221282378081115,
 0.01953817900625865,0.235484010357076,0.5668640372686062,0.8219404228806263,
 0.3734361465443139,0.7648547700077705,0.026624518444001155,
 0.10690771880270322,0.398581271265277,0.1914728927040451,0.2481631407479763,
 0.46367799330594184,0.6516874532420498,0.09855644343047065,
 0.7802873786789485,0.7162525849596547,0.74586221891111,0.9871897532939026,
 0.2669752969755941,0.9498201097004999,0.9107920866356385,0.3978205011968503,
 0.6853060652767823,0.4210179566906228,0.3168122253782141,0.8814173926018697,
 0.6859740427279489,0.6317141493839182,0.7974538744270667,0.9945915237727659,
 0.40097274767751745,0.7088363868960679,0.8929299907839221,0.2988874624764737,
 0.9097457617307297,0.780394563188999,9.197906593075844e-4,
 0.02831058975913303,0.12505559963088153,0.8419859265840265,
 0.6016600767679624,0.42327119884680675,0.7243775080615313,0.8729965535641346,
 0.6384616450602804,0.35440023067338133,0.324172457504306,0.2515950209039379,
 0.17465582305634642,0.24236084657363455,0.07726386539877084,
 0.46954805651619247,0.021177634197099948,0.42996862136473235,
 0.28198144024464267,0.3678466313256451,0.7805004517971275,0.5083289491394657,
 0.7620676196507311,0.04537396620999057,0.5548423669919902,0.8271900670598308,
 0.6129795785080396,0.15762779639074398,0.09138631771799455,
 0.31591157267268544,0.7276994503645898,0.3544793261879674,0.8324221511756387,
 0.893213100116728,0.5723292736189026,0.7968308278666565,0.8008812687227659,
 0.8992801389147438,0.6332420647629555,0.6193180365707674,0.4804929354432138,
 0.023578248558332238,0.8651463121478605,0.6733670384028259,
 0.3285730998545535,0.30006535540949153,0.4046998820820963,0.861660195364093,
 0.05483286627946782,0.6032360237883098,0.21789621097716472,
 0.9374230864592918,0.9020788671963033,0.9757550340733322,0.9666670321577605,
 0.5236503529214152,0.2720974140757819,0.8626848657203714,0.6205200727888359,
 0.10936718846028404,0.06863226048732385,0.6766299214856091,
 0.4559886571441226,0.24041420263606206,0.5979551702148714,0.5002896730744909,
 0.20349726425186065,0.9027405884342394,0.47445148840099727,
 0.9098724832996989,0.9920928230607677,0.9237404579253989,0.7778901595489558,
 0.08998038423882959,0.7497525375304656,0.8377722563297311,0.8226667113777267,
 0.9846628598540472,0.2328917925912174,0.932798346430071,0.22154412285882585,
 0.4273788121611296,0.9259346852448019,0.854359260703039,0.19616223517378062,
 0.625842988372838,0.382351276391775,0.3801513179551985,0.2274439254571492,
 0.5858262565938472,0.7411800319833609,0.6493100578518738,0.08796574372975097,
 0.8590098617624822,0.6423844100791977,0.8858887100649505,0.3438649928574147,
 0.04642909741667167,0.1985618657721071,0.39196479580478794,
 0.10025655481112783,0.5411538176167173,0.35661451887754914,
 0.6339848189979393,0.40228410154968375,0.4550473418769453,0.5039470451917858,
 0.7559858191965262,0.14167684275907833,0.9426375861028111,0.1634793303984291,
 0.869920924247156,0.9464008784144586,0.12507626882653977,0.4444294198570349,
 0.045355596234253115,0.30357171924106074,0.3653610543911622,
 0.2229307284630292,0.3727127178481964,0.29732345936068616,0.9484359684227799,
 0.4686755965656445,0.6217449339099084,0.2527363484614721,0.8938233079244402,
 0.9320302505382829,0.06924884739772119,0.9947978583423872,
 0.18962627834215828,0.010016899519303624,0.649329023051007,
 0.6742407067713913,0.8114352932366565,0.25377833633984315,0.9700742329297697,
 0.6855385719160149,0.7702515406540553,0.9554514553846144,0.06614587709449093,
 0.025280708243418104,0.5067609457437264,0.8553770479244001,0.531427751631275,
 0.13270463296629664,0.9036361848770927,0.45713947576156344,
 0.6886580379018984,0.6413005006936987,0.401492424952938,0.05106081747253177,
 0.7918905238526932,0.3466496262192651,0.2477787476319968,0.7918063994592199,
 0.6282117423989575,0.6563158245347478,0.34844681050976256,0.3612005095901647,
 0.9748015061007629,0.3349101804128707,0.13196046976509435,0.3028071866477674,
 0.6544938098938193,0.8986640722582225,0.4313728022184755,0.37016663907548975,
 0.4304859313528736,0.556651097158765,0.9193874823932534,0.33540919284630843,
 0.8937043010921933,0.3016670506465975,0.5117096270452096,0.2341767693834006,
 0.2085576892321701,0.3859704506173035,0.6097936559581992,0.942364720368528,
 0.7995794639018665,0.32895712282461576,0.4326171219909405,0.8260885003738838,
 0.38071206601621876,0.10473841880677215,0.35660478092808234,
 0.9435700141900489,0.6034643967634566,0.2332850675640359,0.11864920399113577,
 0.7307241489155182,0.280437868010886,0.20436023672077264,0.6492532612056297,
 0.03389042319173796,0.9821817419620124,0.08154642625105657,
 0.6421213822349776,0.8287191324924109,0.14114781783660202,0.6887554029224189,
 0.9999535141792677,0.6862445359764607,0.2360966948710993,0.40509504047034106,
 0.4968166501438118,0.8517662265017337,0.6961299557581874,0.4899890796711649,
 0.16315866451733674,0.9803836880197138,0.36807447705694885,
 0.5784808671403704,0.30883772051513647,0.04143167135018355,
 0.4816028736016009,0.253848742759886,0.2853731571723689,0.8998355283773254,
 0.44549049201025703,0.17586098406681827,0.19133619617128994,
 0.9823209905095572,0.7905410881891877,0.49908016364260244,
 0.03820432069935076,0.05178655744257177,0.5623415514723887,
 0.05712052478977114,0.7126188572241566,0.8742544993087084,0.8086467335642264,
 0.5278323268123531,0.8936499368868346,0.8546909288502189,0.9580951565213451,
 0.5354278421480206,0.04679485621465007,0.9226013250148598,0.876575177151599,
 0.0784808883297966,0.8576818196236131,0.9227164766655214,0.1358335204271709,
 0.205319227265516,0.8980003705061446,0.11808821771604427,0.7384729353641349,
 0.5213502252753411,0.3903817172286198,0.34104793274307077,0.7217512392621098,
 0.9164022244002435,0.7092292443268244,0.5156499066653939,0.9050384462906726,
 0.2157561240544822,0.2294479904567066,0.7272487426818599,0.5196329803922538,
 0.030431534327365894,0.42960148964476974,0.35828113091793656,
 0.5668540297061229,0.8846499484999358,0.7876072924864633,0.17721246430207394,
 0.6212311482733253,0.7457535658456943,0.004631646248533183,0.319790911272943,
 0.007102116645712142,0.9293882884609782,0.519544670715654,0.0429772422275041,
 0.5419453852335513,0.2337384816745922,0.10102369121446253,0.8831561549374721,
 0.943558389554214,0.34314320548732413,0.47808237939804044,
 0.15457479732884127,0.7324795817768903,0.425542874565205,0.9569062321491855,
 0.010217165635398207,0.2916260989560312,0.7974750404474975,
 0.04282172469281764,0.5032299523227223,0.9676085052069038,0.9640527261404755,
 0.4498486624183339,0.8974930902443188,0.39116204155102696,0.5408898205053987,
 0.08971666445907345,0.38260738791749394,0.5695600357100138,
 0.2674652333139408,0.40717060289625406,0.18734309972115837,0.823184967113652,
 0.37240099115479053,0.545458003233726,0.7087694373537156,0.5760818148862765,
 0.38071759305112396,0.17451676557328755,0.890000724100301,
 0.46610111624073924,0.8883876300956,0.8119985154186,0.05515931850560363,
 0.658436195756148,0.415257063681706,0.91618743104439,0.9708267274775357,
 0.4072056262258631,0.33164517483250955,0.57604843084446,0.2587259759424101,
 0.6454206660216198,0.7883941911859811,0.8816806308756289,0.8176689485408906,
 0.8092804039747072,0.43954236824194726,0.7004435976429472,0.7766600272729819,
 0.6100685267450235,0.5394191484463716,0.7551589131650465,0.6106421321528819,
 0.4906749878867971,0.5421654288339649,0.025224162889986856,
 0.8149322466230653,0.44337094042619607,0.19808106441843154,
 0.40325717262970073,0.8291874724523154,0.349753991121032,
 0.024154839370693404,0.3050846903515968,0.7609129807776108,
 0.5428518202180568,0.08185747792767428,0.3092955665045718,0.6479017646707277,
 0.062295326991965605,0.3439244354884462,0.2624245063580428,
 0.8645587150253556,0.33851620546120764,0.8885248581378872,
 0.17037218084715744,0.4858183038316829,0.19687607101202964,
 0.8816933396659202,0.946885670811092,0.8299288431295181,0.49788806773136696,
 0.9522615617713935,0.2972631608807903,0.7889896060413272,0.6028113051510484,
 0.07864839754328734,0.1487708783832511,0.26640372561069914,
 0.39059775540621866,0.4675095441683036,0.28613876632781865,
 0.9719259937168343,0.3548769379310428,0.4749043974495837,0.9611810519047908,
 0.9654465496931589,0.7100857302112908,0.4176924157652132,0.4728826870132228,
 0.3402036803426247,0.6999359649395863,0.30160006002116435,0.5882860164922333,
 0.1601480604729728,0.44665487889893607,0.6389456252869455,
 0.27224832862993165,0.6216504410120136,0.0814955494697227,0.616839352842123,
 0.6261376124855552,0.34611410164098,0.8602611698800734,0.6381175655318977,
 7.654670038538258e-4,0.6665438995130594,0.08066840497744099,
 0.815666492421792,0.8028653113112001,0.3837243809549732,
 0.0030198585851398674,0.9856462251526551,0.93345436496247,
 0.33617774203309325,0.5797203303242233,0.24200511743080255,
 0.5710287755498515,0.23014064988216643,0.1258269662411794,0.6216486330743187,
 0.8244558358887265,0.5244015282773972,0.26576250937025014,0.6104448794276582,
 0.9056626081670288,0.8964889810189652,0.5598324964484379,
 0.026388473938504187,0.46253045897356393,0.715607560352162,
 0.28563638874836983,0.3007812193700792,0.9661117256502038,
 0.45219483088963186,0.3350358881957254,0.9870520185300989,0.4077805035767639,
 0.20435579725852315,0.39879623471459125,0.3251754566908198,
 0.6567059993410662,0.3941735585624463,0.07081940062006342,0.7017934169641006,
 0.3330736378028193,0.29733364873302526,0.509484420779313,0.8750797685330801,
 0.04939501912819122,0.8866104672886608,0.9709450682334673,
 0.015865190724808853,0.6939524862534461,0.7902111199925747,
 0.45741782641643003,0.3790874224953562,0.09259342482103017,
 0.4007326320835678,0.4912806438521471,0.3790251005596317,0.06418254737046036,
 0.9889335619187516,0.7416792228510188,0.9373263431325993,0.4295801685879217,
 0.965113660924127,0.03018616816622921,0.16151841915291787,0.2867225793609057,
 0.7063344100686274,0.7509021591659881,0.8821369335061774,0.16800781075839488,
 0.8754380603133067,0.3035587784000685,0.22732816784846333,
 0.05639929387111553,0.3273035905338786,0.6437451544521087,0.6556774237052664,
 0.40217946279863825,0.7148807287591447,0.8395166119962577,0.2640509503555686,
 0.5877060034346031,0.15444662672819154,0.9522920326596682,0.9914069633176084,
 0.9198853350651957,0.6414393857073378,0.057837527714888703,
 0.5447698786696221,0.88201219862377,0.7743085357343712,0.3049564951550029,
 0.6704388645161021,0.8709163911362411,0.048392902606794186,
 0.8675673419631721,0.6432112426370012,0.16137161099299202,0.8274861951986962,
 0.9688699123793344,0.9320192329139655,0.41572704652461867,0.2557763187991796,
 0.0023581278000804895,0.09531925247568784,0.17119977836720746,
 0.3621903482175446,0.09565204950177653,0.5665104402532259,
 0.26102180732369473,0.5152143929582431,0.20401581843431593,
 0.5932714592314636,0.115123479540709,0.36091169156115066,0.12862364963779094,
 0.6334988557976604,0.09465430591200696,0.29346581880064115,
 0.18157086317013893,0.2850183368799919,0.3038231008379756,0.2044277317843406,
 0.2899632279924942,0.38886022778493246,0.50768012351149,0.09078534048293685,
 0.7544941377181048,0.0789327543419458,0.9191582946114041,0.9741211606693008,
 0.05847236773904618,0.7663818477464527,0.6449190761634955,
 0.11815537680001698,0.7984655346261139,0.06925251009514188,
 0.3471485517970665,0.8220387574016551,0.7198401985884039,0.8518152337319601,
 0.5779761294476332,0.5099909102942599,0.6789693844089781,0.6189985112456003,
 0.05339684536367928,0.01585381410386355,0.638987933863619,0.9304464264598653,
 0.6354738862480788,0.25155241703351694,0.9955833294782994,
 0.18684097165710056,0.8760349103105781,0.14871859474446214,
 0.11013731515776559,0.3236711349551823,0.3619395940003878,
 0.33758943923987794,0.09563676091941098,0.4107907184603776,
 0.3763190693376057,0.880672292796338,0.5490601169785848,0.8764022462942958,
 0.36281560943492996,0.5319240036193136,0.6289291894225008,0.3161829722409668,
 0.8346607048191044,0.45761543880769706,0.9237671553667189,0.9752769827132648,
 0.6656780154781154,0.7927387264116209,0.5432168418551241,0.7738772757516773,
 0.9740209121844008,0.3910799944689548,0.9994979800756599,0.2933666156247936,
 0.2774128700927412,0.39408556403580763,0.16023705519436593,
 0.6631493090039762,0.7553635369842382,0.03376517737709328,0.8438513224210611,
 0.11619550699177061,0.9653090338581465,0.518792393732285,0.7890779516982244,
 0.885048962356646,0.26331393241104395,0.7326933994543491,0.522160833626901,
 0.02288860627967504,0.44739828971839013,0.33223723798231763,
 0.6717335061070808,0.7680728497240777,0.023873708154903972,0.706017154834494,
 0.3810500806196775,0.7139924719139761,0.32474989797747744,
 0.41882501896546565,0.6083890645159996,0.6476751566075467,0.8100796667281427,
 0.10491512177727724,0.8836958953209689,0.6882104533634306,
 0.028335842835410685,0.43621443319439657,0.8674771930452247,
 0.21323399813869104,0.6623143250816532,0.6115405559520651,0.9121589977948555,
 0.07608251928457299,0.049791097894257064,0.10399497297883953,
 0.5003404402377403,0.04371747179241425,0.5443944772792986,0.7573524857551422,
 0.06693798583976651,0.3394847930669671,0.5457675940227397,0.7043928054500441,
 0.46306498865024215,0.13466725985471195,0.02973201254243696,
 0.4567767450153539,0.6998956691898817,0.7759793703173266,0.135061130305823,
 0.7927591283750101,0.2623483934634291,0.030928845030373187,
 0.7167517015122391,0.8601912162944563,0.3007676293566941,0.6738715326414026,
 0.1851344434393929,0.10689636000313951,0.6824299238695474,
 0.06127632921061554,0.8170500307734394,0.9743132611308225,0.8482890594790485,
 0.7071519866288288,0.8748009975846704,0.4148584837436494,0.9257743261844118,
 0.32871183155449446,0.1501620755628964,0.9386133946942287,0.7350758781506004,
 0.29208053476619256,0.9197064238051897,0.7931434263509728,0.2228460450456915,
 0.8626970428859296,0.1133210523927799,0.8913586944806209,0.4602718094118212,
 0.8708269615841205,0.9199317732417742,0.46923989454718795,0.6706074531662923,
 0.8846959945319595,0.5661208891746029,0.34462144973139774,0.6869875666002923,
 0.6600262889595369,0.3416847073535354,0.8296210445246173,0.5071606718839262,
 0.9484939757787085,0.3422085260367478,0.6663348483553004,0.4547515314641618,
 0.2672566735008106,0.7273568062486769,0.9653129296395175,0.28493991255924855,
 0.9472152261011484,0.3877198843939972,0.9180821638005819,0.9266108620959477,
 0.7608238673711745,0.4335857792635809,0.6785757508275514,0.6308310610656962,
 0.7962328895841146,0.26401911511600573,0.47725928321170175,
 0.6503415040718298,0.8280485434448441,0.6707131854230262,0.6163392676294044,
 0.7252640729903462,0.6873054679903693,0.8160850570376598,0.22364215478672544,
 0.6705319650281716,0.7181099267283777,0.32354031522110294,0.4427454504952164,
 0.6087001930498815,0.5383657220723349,0.6764553330964714,0.8146450223784085,
 0.8542809980466824,0.2527919702233954,0.21025266472225912,
 0.25132758598741056,0.21531903954083353,0.04798377541392396,
 0.7311908961963786,0.7240841950970832,0.4575922583057944,0.08995743191129146,
 0.2494469163612001,0.4158124172440073,0.11139417815291575,0.9945741010249322,
 0.5855046955172651,0.525381887207287,0.10127565610865041,0.28017380906891964,
 0.39487469490303284,0.1367334185639103,0.19910248388431118,
 0.5594716092549222,0.16680624132446398,0.8835179327218848,0.7325103767378328,
 0.19153898267585345,0.23354201765709814,0.2630192410337219,
 0.9103758247860788,0.5672899864845645,0.7465484563593542,0.6729746659326664,
 0.7500009905286515,0.17246185888931798,0.8967337794514425,0.9664042428795174,
 0.9004254838995549,0.3723537079190692,0.06996374295132679,0.8265132739968076,
 0.02195978141557131,0.1099526285599195,0.3327425744718787,0.4279443730122957,
 0.13203331074349933,0.2395632086317332,0.3014208610148498,0.4087752314450067,
 0.5002347849285814,0.2585471381648641,0.5416775225942926,0.7392240318475384,
 0.5695723185468766,0.7219564228939882,0.476586096471289,0.7038168215451528,
 0.5762525993217118,0.38779781115529166,0.8524371025350242,0.356584549646072,
 0.43777483948808504,0.03037913920358204,0.6017052406079244,
 0.24753656562977333,0.0857108475841265,0.10210432312127304,
 0.23124051491144926,0.30449829310768584,0.5666262530462761,
 0.08247444184232999,0.363296520230838,0.7861475210389213,0.5004523475400536,
 0.4383177695939976,0.7303482352217291,0.9416960448167332,0.4280354004149448,
 0.10434163853215583,0.697522254057507,0.9496789502092221,0.8468220014634437,
 0.23616061406249567,0.8772845091363692,0.6045461780323836,0.5755789381406068,
 0.26411973371852304,0.16885290352703197,0.6808741192581644,
 0.7065470345503923,0.2892418780440984,0.972569373030873,0.8528986690238208,
 0.06751749639353488,0.8834849434826402,0.7567812614359066,0.3911699772941645,
 0.11522615050290497,0.42227581008165893,0.6561233769561589,0.349757128469282,
 0.07979042350709653,0.6277672919032079,0.43650111460925445,
 0.1092744515868257,0.07417312187328517,0.2276155645211605,0.3376615432541056,
 0.37722107466349897,0.953382850114167,0.9556268012921,0.25182760874059595,
 0.8100372429833802,0.904472401428065,0.9030446264676674,0.5743415860574521,
 0.5863149589218861,0.42791416441801733,0.6658413082560717,
 0.40589306795557833,0.8609722939053996,0.5341286101848839,0.8207620416469225,
 0.10785379396827,0.2033673876874127,0.6853475510504446,0.9937781759563801,
 0.6911812867532556,0.5023347180674289,0.43404579228690165,
 0.42558886719005107,0.14199732072342985,0.8764576793864353,
 0.18576748151650446,0.042269956487885674,0.15621498007674695,
 0.3486492161375192,0.34995038974188053,0.49419886950717185,
 0.10061081606702582,0.3354882327886458,0.8868303429389846,
 0.07270867592296026,0.9033255260415108,0.06680423253253043,
 0.4860422535987482,0.5037257057775435,0.3050021880425352,0.5390969320191001,
 0.7111714435734572,0.10350988997912647,0.4974390000814507,0.5073887955225219,
 0.6956812432426576,0.7749009796483836,0.1835828552911415,0.6126007325850387,
 0.5391629942188036,0.920116738664334,0.06992381973603234,0.490656160874793,
 0.05717366691867154,0.47825304777685895,0.7940704895912128,
 0.7511543044427109,0.02540650669755473,0.25324966408898897,
 0.3195756148767087,0.086958956505644,0.8360133369347907,0.19358590443082224,
 0.23230765224855854,0.5459052791378022,0.3403179532464581,
 0.44001114833748045,0.19660053603928984,0.2618497174558291,
 0.9781335059906194,0.2948859525343196,0.7953369946476974,0.3734702386034028,
 0.23112567838821296,0.08398681970804944,0.43140632890099495,
 0.23280858337782917,0.999213942747312,0.7927779787358673,0.6685534106167698,
 0.3227847632745594,0.7531569225332214,0.2292259305210218,0.03240260439872267,
 0.70063114943063,0.4504832714756567,0.42640533270932046,0.23278349892381622,
 0.08848378949484637,0.1288188754102092,0.6087054897790578,
 0.38640750458215745,0.7641317481386689,0.38521001028238366,
 0.43807866946295404,0.6165467076565502,0.8965704608229489,0.5347260873321498,
 0.2382060918026051,0.1569749962788608,0.6176259024089731,0.3587460156364626,
 0.47935270865611845,0.9790424035708407,0.6845453873122462,0.560629135485689,
 0.5127439678076516,0.3466799466074961,0.5639760942225933,0.6369051162506233,
 0.37687629147829804,0.31105962446098967,0.7253717427647842,
 0.4958283705920784,0.788527889263339,0.21500726905571177,0.1652430289336545,
 0.09075061814620355,0.13981317977110175,0.7038606902656674,
 0.7338910353371617,0.7408637561028445,0.0849993751013155,0.593196501373415,
 0.947522558332086,0.6348544274860368,0.9415862489275537,0.12785656637614884,
 0.7567003130235506,0.6037821925750742,0.2572722165583208,0.7957245928907828,
 0.39589804233696635,0.5963695146804551,0.15238299212815487,
 0.9580839501755404,0.42233213304134465,0.9932270320645712,
 0.06558322576794773,0.5590321508172539,0.2966838058776111,0.7522470777743151,
 0.36231069664168414,0.992908384501312,0.4038090681772346,0.8137229216957702,
 0.0969624168154598,0.6232034041594595,0.23301782414616212,0.2402833398306442,
 0.6887369130150154,0.375397384568664,0.4906549843021495,0.36124335131204455,
 0.05608511893298107,0.040754172472995065,0.6099001317470756,
 0.25933540829142876,0.968259575352335,0.714753029105926,0.033950091650196135,
 0.8476659122231389,0.6730918973331518,0.8477606294847848,0.7134945200713335,
 0.5046300253661281,0.07430091674413719,0.09299949056315204,
 0.009939909058832241,0.3174070523432537,0.2635814355793117,
 0.5470287852839458,0.7117079912931964,0.3952166803526338,0.2473713841923857,
 0.4983710463210418,0.7964975068104698,0.3568941690700147,0.8797646971236404,
 0.2628782124911666,0.7936179498254876,0.8308256415122535,0.7897337160759974,
 0.4176045407326383,0.5181904437511938,0.9406507774465355,0.5835136608768031,
 0.5115103810133695,0.7562153604744661,0.6156669130471013,0.21198614641437852,
 0.5819525940945327,0.09529014661222945,0.9139209527689988,0.5314807392432179,
 0.7871016716843461,0.029827031943404414,0.08151156838736673,
 0.35874203568994867,0.21972459314722725,0.47965462603939424,
 0.11948711973763815,0.5909564662276523,0.051838795460113385,
 0.1895515529730185,0.3979390475318756,0.43752467175656906,0.6672695523980068,
 0.41300213863759105,0.07615362412149462,0.7466787304401679,0.82021430061517,
 0.8858004122944928,0.8299413046711004,0.575883999213579,0.4070568208707014,
 8.45164506846241e-4,0.35293940047476813,0.5204202688177328,
 0.6189406948611221,0.767487884685894,0.02007911519972949,0.5551570544037089,
 0.12513302959166883,0.7276023906200352,0.8040172731404773,0.6006645781623066,
 0.0620935317044502,0.91081487915635,0.33792255531141824,0.9759565850605669,
 0.49407491288232597,0.4660845952573569,0.6465911735316977,
 0.018936293877171217,0.09395518455366259,0.05034637512944229,
 0.48335750770999164,0.25675313954288503,0.8607549623076765,
 0.48071725036871726,0.6605331966745162,0.7715087825074153,
 0.42140692712889805,0.19167075514553178,0.8817290828189447,
 0.4538503764307561,0.286344851603493,0.684949128433451,0.589017100221905,
 0.8550219048884662,0.25662971032466875,0.43403036055086,0.2269435565069451,
 0.7468412350918604,0.3689274119397511,0.28491355084422343,0.2756793959664213,
 0.5209829410156644,0.07039828561958716,0.03789366832814578,
 0.7860450429937489,0.47251251404620576,0.6195719318979863,0.5013498156454006,
 0.8807593143355638,0.4745381214705475,0.8413397810401043,0.25840099333674393,
 0.8646829252687663,0.7039654796841188,0.9834028539975045,0.2571029489168726,
 0.5762547925296915,0.5252499373289647,0.16832573478857116,0.6943655149035793,
 0.734392986770384,0.49299735811207246,0.5424499435442047,0.6688610857459321,
 0.44973321723847437,0.9158441574268833,0.7528546032861736,
 0.26834837954150526,0.9527713331106069,0.34980600956305996,
 0.17566479154093306,0.38513332715228765,0.7701692639133073,
 0.5732613086317577,0.2995759305444976,0.8901431876819836,0.06930386558954693,
 0.9841534752313671,0.5686951594030536,0.772850391994201,0.019234200405069846,
 0.5793710935341343,0.07856613932096934,0.616466071189889,0.804842339664357,
 0.33337615989872604,0.738071682335173,0.36057800746117075,
 0.019574111058099697,0.20554341781586127,0.3734333106402439,
 0.5913931657521978,0.05102843405077784,0.9291802033090226,
 0.46793154547711524,0.8936108669375332,0.4546834031753445,0.9908300618363556,
 0.546210726617666,0.23713763758134832,0.5604872538572743,0.19603685985717056,
 0.45652799304943703,0.31106868626486606,0.27163162581055167,
 0.15338930792095795,0.631021536374357,0.9610710529770905,0.5076160700441721,
 0.09863236015922894,0.780530307720849,0.6810396133027389,0.6649188291455045,
 0.33903576605057795,0.044068763138499945,0.05364944214031231,
 0.08268288367775567,0.9736270900742134,0.19047898886212322,0.468172706906429,
 0.2735344674269855,0.053321324155861105,0.29287888085396907,
 0.3548052692962612,0.3523088160514489,0.053904882396227594,
 0.39261319511063775,0.6459036630353072,0.3400702334415828,0.9006304842363558,
 0.29543082491043215,0.18957988346390897,0.6771200499639721,
 0.18078160822263123,0.786405927638904,0.8250803630059261,0.5436049457520462,
 0.8820803734127958,0.20617991573221195,0.5355834976221802,0.8383109466722776,
 0.19810335814055158,0.06748028452547372,0.6519613759346412,
 0.8848098310899299,0.5694562089992321,0.09063655731065079,
 0.05248521471028811,0.34832699155771896,0.7730219238579179,
 0.25630357479992105,0.423305869258235,0.34732136245196776,0.5781390358669823,
 0.8087012334695738,0.5453399043864059,0.3439140956712077,0.6866917363080105,
 0.9623391859417684,0.19704533550472625,0.27087427109709916,
 0.9723034185076314,0.2942994098931456,0.3925246926049254,0.17317905142966517,
 0.7998688515335457,0.44779050717033964,0.083931967325649,0.0885070051637713,
 0.4577791668882486,0.7939254870270797,0.6064583633213342,0.0699996739833082,
 0.9364900891148489,0.9851432360371193,0.8782868194212223,0.5621365593240887,
 0.21741924615775843,0.9119620749339055,0.3973923850835569,0.2802085678784071,
 0.3038294631727948,0.08235521580304961,0.6133604629476488,
 0.16166135593381303,0.28771084133659974,0.9894811959302706,
 0.5367265247700939,0.36167203813268045,0.34888128086871606,
 0.34749299571661374,0.09707095466647375,0.04300158783842878,
 0.47306853981942154,0.8079706688786513,0.6572674400950662,
 0.27621528631195413,0.11411562687442645,0.31966562235623175,
 0.9730287975821251,0.10480043143532769,0.02911914375643132,
 0.44687816868206365,0.6889706853192331,0.03232958576498601,
 0.6448051188582453,0.521942168609821,0.9774786696251704,0.5072753728338898,
 0.3037231115074186,0.1816392729081111,0.18375137242249606,
 0.09385878810880066,0.5512477629610525,0.09739192335929192,
 0.7747982953031903,0.3089275192996528,0.5745295471579835,0.8451643899025084,
 0.31747199839560947,0.4309881493136902,0.26981235315417373,
 0.11115492094381074,0.41702748702525927,0.795209653332002,0.6125468631823318,
 0.5848192886382474,0.6704854759996934,0.7757009651844821,0.46186523805042234,
 0.0698125052274815,0.6262219331204032,0.8119644305936926,0.7802014638157341,
 0.11361133361314746,0.7750625483128367,0.06157382499073405,
 0.1928562281356231,0.09969108354130951,0.7079110848750669,0.5715645509511105,
 0.2462388171550568,0.7592343598601893,0.5606793351156699,0.9809185312201099,
 0.9880063819919732,0.11187141796914457,0.3982321025907667,0.1752120937665691,
 0.6148879610991804,0.998089617763352,0.5890210176264933,0.30880329153917485,
 0.22442304205478436,0.6216353805791607,0.18816228131879165,
 0.3894512852507461,0.4540323897215809,0.7775982434079612,0.9594324545303643,
 0.43357945517986785,0.8801729609640192,0.6357946606956775,0.5588802001433734,
 0.442584356164649,0.09434498274209913,0.786288475532743,0.5931856868701098,
 0.16691977264090596,0.9024835610073292,0.34997744363471217,
 0.3802076733338262,0.5252795189338212,0.31158073966943056,0.3193474704134184,
 0.7184602271459285,0.8287263834991357,0.48567799687522895,0.9087332704271346,
 0.0705485116240877,0.4539099263931444,0.5900719997181196,0.6289279068311547,
 0.6746936632827149,0.5105918621707416,0.09439264237027367,0.9699347710820452,
 0.30748781162398975,0.18126350289796656,0.7069861754846688,
 0.9950016898260841,0.6072710914302726,0.12656568913381083,0.9091802547544674,
 0.15188714435744655,0.16361946040274677,0.8493246469367504,0.871327947509875,
 0.26322888563386526,0.22319032831046415,0.6981442783024021,
 0.8732821106313476,0.9639556911156129,0.3477156831244468,0.7722405587926405,
 0.020342482635901882,0.8657573280153084,0.7151800741015559,0.756100070360508,
 0.5018690140074155,0.09632919774671078,0.3844450862967693,
 0.033366543395237036,0.21734446339906222,0.08358129474298726,
 0.43392953928182765,0.207943613642408,0.45242153916111394,
 0.09237164196754422,0.8812825550656692,0.7688012214254737,0.9815885025801387,
 0.02512816323202849,0.9581132290922661,0.0703360673317982,0.5513334441093365,
 0.8338281925612812,0.7535788503947677,0.5741110548656176,0.3793450018517631,
 0.8036234573155064,0.6731493032351814,0.9335883836518755,0.11239437349460735,
 0.8248833662467834,0.7848961398420642,0.625048721818406,0.8932545139196485,
 0.3675373976384404,0.37761152235465756,0.40652258583145595,
 0.7634483526519715,0.22909947657818808,0.4642409910418588,0.4971685678003912,
 0.21214605461102742,0.8921897210881219,0.18011911378413414,
 0.9838819520857985,0.32160677822177075,0.7032014971921794,0.6129366725247961,
 0.0690232105007822,0.42040301399712454,0.3987908936526313,0.5747256104510879,
 0.4459725401164505,0.32962796806928496,0.6795173692326893,0.7372313705666083,
 0.952333645903826,0.3517909833081607,0.30418668566164575,0.10334815375256134,
 0.9664269716587686,0.5190599869408818,0.22082450214108906,
 0.09396220950943612,0.890131992362527,0.5465682440612267,0.33278480528568133,
 0.21761932392284422,0.6699649432477864,0.3640487916693339,0.4172359018691215,
 0.9488694406767778,0.4364293680974538,0.6338943297904415,0.29749725980398556,
 0.5289954581940989,0.18804797229989112,0.20248903845758504,
 0.06507684049379037,0.6468067192365046,0.8579128495966644,0.4573697335188841,
 0.6308099090155351,0.6826033039835129,0.5197104757536823,0.9205018873230377,
 0.8502707150396194,0.4288368466009136,0.2968565335300586,0.22545116703682078,
 0.551015964111857,0.9304399930793377,0.4637901506609938,0.12449806157056864,
 0.822339786710779,0.840874348471952,0.16721845630702858,0.3357384234435379,
 0.9700310775492265,0.12485937369043554,0.28852058157610794,0.906446388689894,
 0.5502655546772051,0.5643907483233297,0.8455901291007576,0.677672107962248,
 0.6734711274387566,0.807525494735075,0.995573639867259,0.3089518398282992,
 0.7850501506926133,0.3638266368628885,0.20074660664145105,
 0.11649467347149589,0.496580839529424,0.8969394970125506,0.4457958810903597,
 0.10156748677327387,0.5468249407972032,0.05049756361344471,
 0.2736996267852708,0.3954698523557978,0.7334755181378012,0.7018071691217797,
 0.8321305117632275,0.009405607599781485,0.36929510026084733,
 0.6264175737711728,0.4397013690344521,0.15488947189401392,0.1489235808466115,
 0.5365922751369516,0.1344089051317503,0.7446092772677448,0.8495596693517692,
 0.35121899975105864,0.6655918195290452,0.374363737766477,0.529274118975525,
 0.0720408100864286,0.5325390848372258,0.8192140838182517,0.4800605020734596,
 0.5689707091923087,0.6259692918314869,0.3460472030638049,0.16201974411365772,
 0.34545488156590576,0.2949252602112352,0.5381094051901002,
 0.45903746207912244,0.9693901576822109,0.02161694464492292,0.367976079281801,
 0.45701853161236883,0.11119361763296387,0.7637955898780698,
 0.24942556079316214,0.019017441946846736,0.025317508311242132,
 0.696356970565125,0.723560918356219,0.9291789403894584,0.9607835426045005,
 0.44152725962557127,0.3246207526308712,0.8264282534308675,
 0.35177543262328725,0.8251542832489183,0.8609319018741439,0.8285143144248095,
 0.985301236680285,0.5343427439000006,0.5011590620757677,0.8663020542229093,
 0.8537702523112626,0.9682644662782584,0.23412929207441469,0.7048939913615525,
 0.6374678640348089,0.5747065633531836,0.6941697558427412,0.9797856329553193,
 0.28687460132179854,0.6540944902928026,0.09509202654897653,
 0.8884676557904869,0.5289660792480256,0.6037878432820931,
 0.027859216845855572,0.8172838554761881,0.40694437418598095,
 0.25318189297222615,0.2708911719510113,0.3252414056636699,0.3495507375207314,
 0.948875455292935,0.8811094869917327,0.33136063619680056,0.4074672854929471,
 0.24943979579071274,0.7070678893387286,0.3337173803029049,0.5100201582820383,
 0.4537356499959381,0.02933381946401281,0.36715107084258314,
 0.01654188525684608,0.3008730146980034,0.770026813952372,0.07220791576944241,
 0.7944010278477619,0.3408440923388373,0.5392154100984389,0.42002545426615256,
 0.5495651533323472,0.5632255232608796,0.05749242043971525,0.3743227658578633,
 0.8934029022714638,0.12912688022295826,0.416676467532502,0.5189888352555737,
 0.08370811981557968,0.09842133491784744,0.9160154620379224,
 0.35038320236100073,0.9467406002752302,0.9715910989410177,0.2030209017285034,
 0.7301344253482207,0.9163110938626481,0.21913862559103747,
 0.13860643609432954,0.017976831374312052,0.332204305469194,0.938619832803114,
 0.10413916528445177,0.003230297127124304,0.8683841439524465,
 0.1677384083030844,0.12460042794868043,0.7663700865245264,
 0.14750920488272057,0.5812658198869509,0.6935791945549858,0.5725327791480345,
 0.7787673523750472,0.4350451005430598,0.6927807780124859,0.04364203855623239,
 0.2457008363470622,0.8335840816394615,0.675806922214352,0.8720793983146009,
 0.3208892193131725,0.21958047899612543,0.049553032300393984,
 0.014655673854488516,0.7200677707473713,0.3145539141837077,
 0.2815911687512366,0.44723482520134883,0.8874778433469568,0.8485489768229044,
 0.6318134792383029,0.1767424185196137,0.8291084606683707,0.5751695197602923,
 0.6534791067850235,0.8828215552420189,0.5415132783883962,0.14052814855334506,
 0.017769874591391788,0.7380605552988979,0.10158413249562237,
 0.5784327085570333,0.2505595746456718,0.6504544966842046,0.11593905008733096,
 0.900076650366091,0.1874376507806469,0.07157764014155665,0.2743337221929174,
 0.010853676624805297,0.8986290804834367,0.2478476536604598,
 0.4889969840293067,0.4315622633108802,0.5143751248878798,0.6323866708197903,
 0.9922060793781773,0.00534981663443912,0.08847432543114109,
 0.4330257389159111,0.8648964629008372,0.7817694768249258,0.23860991037896084,
 0.9216513543681112,0.3872782363763375,0.8494387083593151,0.6843589405964833,
 0.4366297301610492,0.2714589964337747,0.5666148478631277,0.21184198283780342,
 0.4440663285841989,0.31657887689924424,0.1555340400723384,0.8463805624933337,
 0.8622887890746253,0.58516939933449,0.938019996438518,0.9028943933515265,
 0.8808632753310561,0.028701376543213364,0.9898602641886776,
 0.16398534390890107,0.02725934942091668,0.5338732441053375,
 0.5823346344224987,0.2614672575767919,0.4964606678498502,0.6763526333913744,
 0.3439903462550442,0.1205144713143802,0.7987542513547816,0.8254321098151436,
 0.25939426666789045,0.753260842758738,0.6991530082075674,0.8326469184122177,
 0.14578325472962184,0.8894563532817706,0.1512862551804215,
 0.28385919363949497,0.3316096873456226,0.9297462211262769,
 0.25607397437251656,0.91474510513995,0.38517150021117263,0.11419758479512321,
 0.5589902733265273,0.8346545906660947,0.6431834342168969,0.6061559428558553,
 0.6534667181190751,0.6603494704128571,0.884289517618195,0.9435579716850422,
 0.5810796570282726,0.8157816822831868,0.5383639472694646,0.5206983997017307,
 0.08277146161457904,0.4427249189500465,0.8721764325939946,0.8356801003227512,
 0.6819349300118861,0.5636161064455661,0.20093855311725972,0.9249174363934515,
 0.49325751988370303,0.9102266871836225,0.9588110395709467,0.8024088167441904,
 0.8637289570551973,0.006505367917968963,0.15172919465121892,
 0.09716941309463134,0.03950859762812431,0.389079940051391,0.1408568802381095,
 0.17009012797914058,0.049667541836727835,0.7921008062593863,
 0.7023925013587642,0.859810379593363,0.2878543122217967,0.7714713720574797,
 0.5094059529908356,0.21263220938212857,0.26824727448000374,0.873910445522738,
 0.09664742410456584,0.055759308491905246,0.35958910115520204,
 0.9153306721114614,0.44542731387384427,0.1788863315319763,
 0.34930839174031203,0.7305855881887654,0.03761130107217148,
 0.8332519883020512,0.3626034147292868,0.253898141086064,0.5085282116343712,
 0.06199217259580592,0.7173877090047469,0.9935112807887332,
 0.04179060339635288,0.9918348603403931,0.6821584319831419,0.92048916173395,
 0.32204096033397633,0.5934536770145897,0.11791219913821305,
 0.38039005998261977,0.11724740625988261,0.0185160544077676,
 0.7493664648078401,0.5080197729072452,0.5699456334310974,0.4645790825291647,
 0.6517273812132979,0.7756797371218486,0.7904363838811637,0.3127692942322844,
 0.35430604393631016,0.0778227845091013,0.703648175679709,0.49764711883179236,
 0.6137371235351269,0.4232006046481308,0.888898280104636,0.217976857845383,
 0.42122489858447654,0.35046291372741323,0.43787780800967,0.2895992665850321,
 0.6019606700077222,0.07637601225593194,0.01870916697581848,
 0.8434522143903947,0.48045283424412333,0.6982665678995172,0.7649961604779913,
 0.34983868356199777,0.13029156985814627,0.5385923801415693,
 0.3898271575534309,0.42998965294314007,0.554692006175478,0.6584327267227102,
 0.9394590058115481,0.6004801130080812,0.8829582513841674,0.3280679703157434,
 0.20477054327754463,0.06359511382873351,0.6100488577795444,
 0.7903093030858874,0.2938056703232923,0.8398044530885457,0.8893168052802815,
 0.8543344156006539,0.8452271624700687,0.7718448790522918,0.9420982134966283,
 0.38708303122171184,0.6903734716956365,0.7708801518566395,0.8897165332033582,
 0.8991532901518107,0.9347700720032238,0.8816747622788533,0.2588356916579255,
 0.41524458032295297,0.44018384089200635,0.09276145486382603,
 0.39561344719496094,0.7624836978630425,0.26761882637874923,
 0.9613425976721719,0.1929272727841298,0.13514945970829095,0.8960186253089629,
 0.3627575867939268,0.48306483058756144,0.8547111636979903,0.6122952215539994,
 0.025975953433575727,0.8236697995910685,0.15933068366883807,
 0.43895721173292446,0.8522349846969294,0.8248989449518817,0.5186617978057971,
 0.08200590204484781,0.8384328241924128,0.5883066165369586,0.4730436611877328,
 0.17688766585346594,0.8269658842723661,0.6721077656269983,0.0726456170711689,
 0.09138378993465224,0.8221578642424552,0.24651202428882757,
 0.6180625247879485,0.4257817301436986,0.9997430764534733,0.26714692272199936,
 0.6755869169074089,0.5047548567874358,0.3738888260515907,0.14338780264722784,
 0.6323366274633431,0.8145943418398027,0.2903088041187858,0.5892213409634935,
 0.767913039200389,0.032066925260435086,0.5233776130775358,
 0.09694876396107432,0.9637328591742673,0.0379354194178354,0.3297994164818573,
 0.4066765715677759,0.573440943686116,0.5449095737385434,0.12382792432982681,
 0.5167526007240634,0.05398559006694137,0.1721251615363144,0.5048798717595653,
 0.09655710478873658,0.14087867613981253,0.3805292732936919,
 0.4909597884723905,0.8625438088899993,0.06479356831314975,0.9907214212214858,
 0.9771021396368831,0.7140177764257141,0.711164040351154,0.8158557189583165,
 0.31063400046629197,0.32783394196892324,0.7760356394207578,
 0.07127693268713942,0.16780941562521046,0.9588171226633343,0.435657729564461,
 0.9585125298798072,0.6388069641328844,0.6448099592192285,0.49367978605908425,
 0.7873885799680892,0.6644841750186554,0.7165936081589759,0.9853981553421406,
 0.849302461646258,0.24287018346900746,0.03362781821985927,0.7033236254175376,
 0.9184036566728313,0.8952149925095086,0.543946103618194,0.5610039377033065,
 0.41671474326691704,0.28871613038502186,0.4732006149807573,
 0.8810056049320183,0.7797327827560581,0.9903435790432193,0.4235307683901046,
 0.639389797655333,0.8352103013799006,0.5539062125075912,0.45717553648973575,
 0.8729232952162214,0.8768869348496535,0.989127418015716,0.3799289857711582,
 0.29007054023954293,0.90516370683922,0.5645210174185689,0.1016860364648573,
 0.8805014135815761,0.07477571093343727,0.37693057945612385,
 0.30088503465940863,0.38947179837770574,0.24733055218058264,
 0.8305729314049011,0.3759966109177244,0.8351346360332706,0.9732565681543999,
 0.5303929226517611,0.6098098827148496,0.1224911562867742,0.1543562407361807,
 0.2876436072246762,0.9050752630551315,0.9864579222707797,0.5752268608223786,
 0.2887327912530141,0.20279685939578407,0.015127746691058208,
 0.5516543595226953,0.10859957164947698,0.613392978234791,0.23061523396745054,
 0.35693757756467237,0.23097999845981698,0.45223428238522334,
 0.162151661969672,0.8002069878444852,0.9723150756798006,0.2649451396223165,
 0.8101480405804836,0.9594434676333597,0.24594692586716982,
 0.38141637104617676,0.8895589124894572,0.2686113989993393,0.4354728700141003,
 0.9656413679873406,0.7730016792209116,0.16332982641307747,0.8289708916989915,
 0.7664698240535032,0.3653855709334941,0.005427608910615778,
 0.3442002510841474,0.7186197021235576,0.553509966869879,0.964869885474607,
 0.4434042347120375,0.8995863281065734,0.08969389621780355,
 0.09384228634447833,0.9497288855091851,0.5868710267245556,0.6867634596002281,
 0.096335452006591,0.09845227705977344,0.5173726007086167,0.35910721198011986,
 0.33947710399755304,0.9699036014351679,0.2145894721538596,0.7242312907662076,
 0.7781697916389215,0.1452344535239417,0.7628458403975988,0.7500407823099834,
 0.78963740408223,0.4027624215461483,0.1653811027088682,0.44307878343082563,
 0.9826442203947325,0.5410855256915821,0.7961091221340012,0.6596533766037216,
 0.5534721234139903,0.6825815144540326,0.9300414882038143,0.13755201604500722,
 0.23503065066038387,0.03694453602310188,0.43297526673921294,
 0.8576492106018588,0.2068387337735128,0.46436111509413613,0.4258292458747437,
 0.5763986903797043,0.8028737248296656,0.7850446696837947,0.9230399604564312,
 0.36417195224980015,0.5378362508586949,0.8046136622076605,0.7130019632128363,
 0.6674604891214262,0.4439219839379366,0.48283622925998837,0.3617203218780136,
 0.27812249293586233,0.681811055715846,0.8309991664589832,0.1431532558888965,
 0.10525351216071566,0.9558727901096189,0.49940220401482516,0.191923177853636,
 0.6020595812767231,0.7202384564231318,0.25707186997763576,0.1201956859573089,
 0.008844347729602409,0.031790567447821494,0.8647014859089316,
 0.3863279873639708,0.6471190599696064,0.6781756125493239,0.36161708116723357,
 0.7465368407330691,0.6304823281956375,0.4805552446533823,0.8405301777793066,
 0.31439557593561873,0.5257752078183847,0.952347859887677,0.08137714696300824,
 0.08351568712962942,0.03058173669785258,0.6933086882691863,
 0.8795908887232956,0.7541446035844015,0.23964096221686293,0.3371240365797641,
 0.5949960448720104,0.2219557403857415,0.6786966892277919,0.7124444763958184,
 0.6475670064683374,0.1396343164965952,0.5059029292482018,0.38093875478629624,
 0.674705075552925,0.8542819383155821,0.12190501455594882,0.2153821788544945,
 0.01342548674370403,0.8018505319022455,0.20543643194298844,
 0.8960288237129906,0.8129348192104296,0.24487315189419956,0.1363674745273059,
 0.3971474542133949,0.787690355750985,0.07725212909462509,
 0.0031283346049713057,0.1940473489626573,0.7500631976027634,
 0.30703713493926554,0.7980401787730435,0.19603590368877177,
 0.6370543230023726,0.09966202105410349,0.9185756760226036,0.8147830617428325,
 0.11760164196043199,0.575234573700809,0.5002247920448272,0.10016546336592835,
 0.3780968988367265,0.43308708968553056,0.8831854627265521,
 0.46075592239176655,0.1140596499668558,0.6534381348262818,
 0.05779119094377594,0.37841753220405905,0.7100997762234987,
 0.9085699875393232,0.15286949396574145,0.5532403261369221,0.5586870501200392,
 0.9734723898330093,0.12766863581503662,0.840126157465611,0.656337635494409,
 0.17953574442111342,0.4059361167177651,0.821528671472183,
 0.051412255341893975,0.543171209915031,0.6098389272952885,
 0.42009879516700255,0.04780350345704809,0.12341122611128819,
 0.47093054722438543,0.063425746127027,0.22205498872483498,0.2167724829965062,
 0.537736359272804,0.36282145550637845,0.3554130640893013,0.6853943777027085,
 0.9816438380544716,0.2737862348066158,0.28872953429942494,
 0.46228567713820556,0.5049672950272304,0.8742952404650426,0.273014449179261,
 0.2862875834011941,0.29646787521604023,0.8920010557968686,0.5248870935285899,
 0.19258436254143274,0.6984652325086012,0.011892263126945912,
 0.2903160952744729,0.9178634292100575,0.43916917262808164,0.4352885576010249,
 0.29071002332754914,0.49763625677984746,0.845515670600184,0.9269941300476311,
 0.7578212488453628,0.6940186512764428,0.46486446391218483,
 0.13831768072453665,0.7143256763401612,0.33686464528238347,
 0.32189120114258163,0.19302509882566732,0.19775048122549344,
 0.6101763447699086,0.14965527459052863,0.017320564842827846,
 0.26010433356437335,0.5232227007297109,0.031059223847682893,
 0.4288511910110805,0.2232510850923961,0.7778519097326039,0.08064796121643025,
 0.4223895228399511,0.5095971652102153,0.31429999631480443,0.886087324028939,
 0.6877346750642627,0.4952437704970496,0.21676621220333359,
 0.23592648497185464,0.8295890595814573,0.9607195906921155,0.6875908019606011,
 0.9662821813758207,0.24004896916227114,0.45564660612878705,
 0.7220833565254351,0.8949201894464914,0.6501848757046673,0.6116695552945713,
 0.029684531168313488,0.7003734926439462,0.923233020796443,0.1956293904116534,
 0.3623981831613664,0.3062432165301575,0.2616762276469954,0.4309697241304017,
 0.4453043153837387,0.9128776349613507,0.546186461500237,0.948184425401758,
 0.22124413708188173,0.8564543174291138,0.9257496719244568,0.708223844634687,
 0.953957946948765,0.11977878890276805,0.9923093985469346,0.894265546936935,
 0.816568904449502,0.19013772602933188,0.4271900336891541,0.3353970078717209,
 0.9928489602343994,0.8947528948443055,0.05068688656723774,0.8788439645366746,
 0.4898379031300303,0.35360842049932295,0.7719105163856568,0.840790357462654,
 0.8061672765178816,0.41684757285927465,0.5744399874496526,0.8804112055733158,
 0.8577067087835867,0.149846019253003,0.9675907954393075,0.13364032237733847,
 0.13379515506724915,0.34508919459783927,0.5349306295803538,
 0.5321877712188839,0.7287996835732932,0.8938692976190582,
 0.004656900689990251,0.48258101875096604,0.6537855453385958,
 0.1238943679796799,0.7535678213368466,0.16252932169652712,0.1811357450868375,
 0.334852687672359,0.34550175734975896,0.5789904791782216,0.2829049502534313,
 0.27306690975212444,0.1102164415201019,0.6704045900266286,0.6712836353764696,
 0.8708609857928361,0.8173530320556516,0.4974730251838074,0.7315136050022673,
 0.5550973603500682,0.9449898487235312,0.597525040471158,0.08865257658963543,
 0.18950579073024987,0.35653749763904163,0.009709458727219844,
 0.0023811776483476654,0.9816696315178955,0.10914875144071168,
 0.20588908355689628,0.34040673507421204,0.7588395179098466,
 0.6282218700734664,0.5434523754108358,0.9938326837652732,0.5582696067234448,
 0.3579209002088021,0.6988456500881709,0.537660014280864,0.4114764884239933,
 0.19294792305351272,0.705600987335415,0.15465295811500568,0.1506230201383698,
 0.4522056434305749,0.2765917093253254,0.26007012841269084,0.7951736117707655,
 0.21637524883792858,0.8091159838709824,0.20047742923182843,
 0.7729564134909013,0.778804036753759,0.3714497387033606,0.22819755483826987,
 0.8407238167673297,0.21487355933320973,0.023888055396264263,
 0.2946805296913665,0.7357426297281553,0.6733500343187263,0.3981671469037066,
 0.29786840813460547,0.5370607274309096,0.3568777485404042,0.7650459262361161,
 0.7392874468290427,0.6987119369108229,0.059111666902854854,
 0.32707590094343375,0.5345229864210587,0.5852292492692364,0.8445104557490184,
 0.2070384666216427,0.20896018686366702,0.8192237075386652,
 0.15791484239365328,0.11863965573883806,0.4848846710195307,
 0.8348875762569512,0.8950824974335295,0.7216982679269978,0.5814802244130389,
 0.7530337255627482,0.3670640012974258,0.07318366792107522,0.5235541663327246,
 0.07890676779079264,0.09279359023512235,0.9841321890070573,
 0.23797122726693598,0.7979183870015414,0.07001242338315172,
 0.35069592041798714,0.4553338574606849,0.7304925289318296,0.355748707816811,
 0.8378549195402876,0.7804815770685751,0.2448646109272441,0.9146062079522193,
 0.6536794267282626,0.09342898203065597,0.3679963052072729,0.6660379161530523,
 0.14619998646174814,0.8812475749076056,0.19481250475149514,
 0.9627418388437801,0.05148439921134651,0.8352645126109063,0.294172453372336,
 0.28828913014351554,0.002284498359805154,0.5504013987080455,
 0.6700622638058417,0.017085269508743117,0.8644091270460191,
 0.8882142227061011,0.9282440509698852,0.1943778152391119,0.14373756138647842,
 0.9240931453689291,0.3447884350425865,0.47117756356543794,
 0.34285378181388293,0.5547417866645151,0.1552789045827866,0.9223290969860005,
 0.802174463690603,0.5615820689101512,0.8530142759164138,0.10043238788670894,
 0.4190044408539719,0.9407511877479106,0.0337717738038098,0.8181663332680996,
 0.76635865125142,0.3945196816072205,0.2941717790914088,0.13099434687943956,
 0.6638380632192089,0.9744708293820864,0.14087801932707378,
 0.025400759660674455,0.6898460347743988,0.3650026351797355,
 0.22191758331365163,0.3558751436398737,0.6753933701124133,0.953085697908695,
 0.16954748236733197,0.22920977856256208,0.43246266531355626,
 0.15262337999099118,0.763439513888964,0.48188138585880536,
 0.10106970225382861,0.546996690005522,0.14659640324702605,0.9677366384825317,
 0.9550926867338317,0.47354000567038645,0.44989769299376603,
 0.5609282829100596,0.5381110113958912,0.5726730880294872,0.4503786152199133,
 0.7116508835488118,0.8511745779706538,0.7456782604769934,0.9501606684420815,
 0.6160266403022175,0.6789209284276418,0.3560140007387027,0.37776074230501155,
 0.413765757703639,0.27498131736133746,0.5428876437534439,0.3095081202095179,
 0.6306296590084843,0.9761185163884524,0.37202555218959243,0.124482990890181,
 0.5542648509724182,0.34410680653450443,0.7212650983242018,0.2530672656562759,
 0.8096957482179536,0.6241846651490195,0.4578876141991678,0.8437038776513821,
 0.22066082031322587,0.9942613336940808,0.6614520082344031,0.8275508795172325,
 0.2636606096336117,0.4855899826853066,0.8436836182714027,0.04155502711990333,
 0.20993296096879654,0.3448505165430735,0.8033966228239129,0.5425291082918894,
 0.44335045017659,0.6104501395357829,0.608558637068645,0.8715143965502883,
 0.3039494060187461,0.10910391785862128,0.18581025186497468,
 0.4083447262437542,0.753662777013292,0.19477899407339982,0.6639288356443573,
 0.864630420152869,0.787615363031337,0.26992353507095557,0.24325243188076007,
 0.005776190191757902,0.27835802547829847,0.9902732222891341,
 0.6868283138969566,0.14412801230084926,0.3847611412455981,0.7991736350151557,
 0.3838672917408966,0.4012465935218771,0.5235556043828931,0.9070812982582644,
 0.11776251191013964,0.7610377169328697,0.6446521626519578,0.2180045494366537,
 0.9543011313773652,0.006568926127936914,0.4772715613930183,
 0.9643635433164869,0.10006408630205665,0.7609782241927,0.4894163734461219,
 0.8572642612951514,0.740970349600083,0.5860686374346818,0.8958785184942786,
 0.21860033964319248,0.22762409432895048,0.4435627032044077,
 0.23691012862545346,0.9020703931276056,0.9702006971226423,0.8801805711089011,
 0.0863906338173348,0.035572312708763265,0.8831762162744162,
 0.5423404154880506,0.971257490770518,0.11551816039113028,0.5077858875374262,
 0.10200083621584055,0.6336798041283135,0.9846601108272663,0.9574637994088829,
 0.008490337022571404,0.5502368675765819,0.38380336603368836,
 0.07455895962885262,0.9646590512612923,0.5391540970418067,0.3371372087315825,
 0.7389345587499859,0.4498211840427764,0.8552249376180803,0.60688212585592,
 0.9782238352578664,0.4912126536174336,0.031691999562377005,
 0.5010800077114025,0.3718952940625052,0.6554527812811197,0.5516357765866942,
 0.2888750715060393,0.2626540090498578,0.7688093584892773,0.5985439788752303,
 0.5727039111855685,0.21007669492398795,0.8465976888133748,0.7743149191187869,
 0.6136431367646955,0.41453463544166036,0.9177037034009495,0.9001183592462066,
 0.29417202763137973,0.04878898694708467,0.4300723230123197,
 0.7507207187148183,0.8136072685377039,0.06805338753063395,
 0.03504859435972052,0.951725176483491,0.046062549510544226,
 0.8375499149552318,0.07182748720157905,0.32895302829272255,0.54168425749108,
 0.7036998682868852,0.1854671473410745,0.37007136671798246,0.4983705797619221,
 0.17233810773996694,0.004159122895885603,0.5753819916183768,
 0.5815871471596997,0.2057852156087936,0.21175722883004355,0.7295165674914282,
 0.7548302497154848,0.48185980355209135,0.14675475668869686,
 0.2728580311772669,0.012241417971847524,0.2690354131761542,
 0.8566052814658504,0.5461205213514422,0.8180572614314539,0.4017081534587674,
 0.3904643019534615,0.9098593611546522,0.9983751601540748,
 0.010086383533977283,0.6635125356381621,0.33354982096835695,
 0.7732444117108438,0.7759420513496746,0.7253511475013957,0.8577403737158478,
 0.20139318504863146,0.5713526395282762,0.914674740649504,0.1362725794633105,
 0.18779814054907207,0.9490962319238012,0.2665385893297886,0.1991575945896974,
 0.14993130207797267,0.21318707681692806,0.6096968893384275,
 0.1977289715912104,0.7486092406811352,0.9396612420306788,0.08221568007607993,
 0.12009343599756128,0.8253683058368163,0.7720092331562645,
 0.33306548428692606,0.10849125236207613,0.7992148795918156,
 0.2054570883249539,0.2547002195014124,0.42373657412329446,0.5501188000573072,
 0.7975430525409047,0.38347417435471054,0.45622055253877913,
 0.6393812732389736,0.2621149633604801,0.9286725639842115,0.4914064960614566,
 0.086214672299052,0.17835379880984514,0.08647513681220642,0.0631622074545316,
 0.6452099165682633,0.22890387938516343,0.43512457974213115,
 0.5248656542131837,0.9893322576671811,0.9341584203515638,0.459654609536264,
 0.7318082398254807,0.2054496962822444,0.46280845724292097,
 0.08385174195212308,0.8427585483962927,0.030042579211027598,
 0.717907958758339,0.38072806965000017,0.06730606833417552,0.1495087042486255,
 0.9751393244193397,0.7472963719695578,0.40075151481138804,0.3546608617418522,
 0.8672299869869704,0.3340021337207798,0.6035757537951928,0.6544418783472857,
 0.32810704293442006,0.5408339457002027,0.5553711156013533,0.6796046463791084,
 0.9109729890001615,0.20126181502030382,0.2929322800019427,0.664524255092572,
 0.3107791454491321,0.06122396344651493,0.5399738082781508,0.7567371073077968,
 0.02684723919175358,0.006019134009375149,0.8554107450800035,
 0.10373286675833848,0.3861544670195811,0.3324549696699163,0.6732554431789115,
 0.3998954216307661,0.9139067417133386,0.3295540787773801,0.08198629996565487,
 0.9593881003813894,0.5491699349053705,0.8742976068845413,0.10318547629589703,
 0.4608239501854373,0.4141678949692722,0.28164422245954523,
 0.46679423105293827,0.3910395526844065,0.8476785372933916,0.1274839727012942,
 0.4181653140986116,0.19100822814111673,0.28847859162136935,0.678922636297877,
 0.39914908375118174,0.6447307259079235,0.9291613030288262,0.5195831781351526,
 0.6758345319349174,0.8283818566320855,0.4186516115783391,
 0.019245218770020056,0.0163036395004883,0.45479369889221444,
 0.3296858106367657,0.8339549897986904,0.43732418701247344,
 0.37589489281646327,0.5133503602920496,0.32464845601290426,
 0.8677634397719426,0.7532648502656519,0.1955663012879011,0.341308162438393,
 0.3197313233028486,0.8709842750018647,0.9011450149717891,0.9827550773515407,
 0.5082106939370209,0.3334825711525079,0.5834900833525449,0.19771170921206305,
 0.8987677295033006,0.4030709258418882,0.1313009398958741,0.829740815039303,
 0.7711532855800083,0.9893772205006562,0.57552325752673,0.9267082861291507,
 0.9981918107292782,0.7903764935252663,0.3166011012406158,0.0636279383908781,
 0.2848637192439367,0.8158402429452348,0.9851330383862197,
 0.028461761050740386,0.499332522467953,0.5637151414781421,
 0.24597369743396247,0.6652367765446509,0.31602689721658794,
 0.8282855086120839,0.00563158738784153,0.5013181341104165,0.37226767554215,
 0.43370208103870944,0.19293840227374126,0.7900287685026738,
 0.7082541158031195,0.6883885536338696,0.9914642747382295,0.2756401927229456,
 0.32533776889368127,0.942986385298453,0.9482755622220167,0.0462712090723173,
 0.19571835727475206,0.2781770308117214,0.9787272975197971,0.7372609276806236,
 0.5875019356739903,0.2464578370117818,0.653690591831925,0.05211945684523833,
 0.3142540984815374,0.4037028846761449,0.784287367247796,0.5013669531073318,
 0.5294780264070613,0.6000127494321205,0.5436825455490695,0.3950833694779824,
 0.3152336833496532,0.2550334405106567,0.16589591478512522,
 0.10578788486501733,0.1661823280439254,0.8260282363866588,0.6744347800752538,
 0.3288851522715517,0.6917666006979215,0.7226885171735722,0.8460536117743692,
 0.1550490600196356,0.9528730138353609,0.8373494008674236,0.6995571743248783,
 0.09272802232059463,0.9099978482362183,0.07336653810630012,
 0.9340577621113897,0.008481432990140036,0.22533314384443492,
 0.27015897188384774,0.37203648071027007,0.6018446204405078,
 0.7661437326107912,0.46439559553723164,0.10355053915778134,
 0.44958433450096935,0.6220950703671082,0.7299276919376682,
 0.07667629544035537,0.729452861239408,0.6236794278003305,
 0.002519021422444312,0.9093396024668063,0.09350951316788447,
 0.8407785934252375,0.967752493158255,0.008644064397252849,
 0.24978748851445642,0.6416617870290046,0.45059641728716504,
 0.8606454971517781,0.4058930385263618,0.09591379205870676,0.2407645529387542,
 0.6563053790197585,0.08750483878098647,0.011967196641284317,
 0.10690665330467619,0.25041625920260724,0.792800793707958,0.5620462139922026,
 0.6353034234971395,0.9014412622096014,0.6470303020988873,0.7037463410730775,
 0.9168993113223685,0.17090632479464118,0.9763452107135242,0.907055842887462,
 0.19688797341439113,0.3022598498794089,0.5827332772690005,
 0.19284317416509733,0.9231306176588583,0.6275595308565144,0.7360979706137061,
 0.9720808690187169,0.3862184409161793,0.8062101359271308,0.3243987242878814,
 0.6975166442329943,0.3334532590136655,0.03727458250421323,0.188787809596192,
 0.42259287901749343,0.7420341990091844,0.7585965977944448,0.5278297978969066,
 0.05485210764587911,0.59441978452699,0.7510332580287119,0.05154721737544998,
 0.9567231473142715,0.11470462842952567,0.12041100285262896,
 0.9158272333262181,0.4566063771873723,0.21454407611555382,0.6589480984415688,
 0.4070527258091794,0.5419609813389512,0.2641701829990044,0.6665913894519327,
 0.5841824366596777,0.42620021058879,0.7874523077295004,0.43524625260071104,
 0.8878500726164803,0.3239498898340717,0.5017510726346268,0.7518937568631568,
 0.03365755181474761,0.49659976692429975,0.15208690365469024,
 0.27182749862024136,0.9172999416051917,0.9311908449151329,0.8158500162215856,
 0.9044771104025893,0.25480443684541354,0.25149452590501054,
 0.3180099247309145,0.8098628562253434,0.02935583850621959,0.5696594066594279,
 0.5475919398665847,0.5327470790755336,0.965763846564952,0.15411382217412273,
 0.33421078713097985,0.6906798334615383,0.3070790596871389,0.8487646816181987,
 0.5348082266574015,0.460051609011517,0.33015183219651656,0.42085618173139727,
 0.231808291802584,0.023617663058442373,0.2561945101098959,0.3396259917331157,
 0.6785545190792668,0.7144661010209729,0.3449701195916465,0.18485820437490152,
 0.4004334375588563,0.07760838760305322,0.47695827535796353,
 0.9253463354274294,0.16369279427879435,0.4635434989327216,0.5625499283524444,
 0.9127946794455307,0.1465331378424406,0.8096386420262963,0.9244303459838482,
 0.19663474923461688,0.3668055852303329,0.8727438158417413,
 0.21926295215896208,0.38039573803770543,0.25889660021650307,
 0.8840829439388321,0.7830165759801266,0.36652405038848745,0.8885648119151756,
 0.05286660306098323,0.585126315286313,0.5954503751166448,0.6900836200407523,
 0.3971019669400625,0.5833313970604016,0.9984785989888069,0.18838252565448244,
 0.9939036711953113,0.019752995438426568,0.4664717319185694,
 0.6127315895414405,0.028912663144880058,0.45710103378218403,
 0.32699382038510294,0.5285132021684146,0.5057196724214025,
 0.27869602320749376,0.6269899497432194,0.8350120821982203,0.9322104257273615,
 0.4164934740244455,0.050061176892705506,0.802076123384925,0.3507938462437218,
 0.6167118301406973,0.9289958833484542,0.7488155055097727,0.6393582786095952,
 0.8549729024618462,0.14351588362947343,0.372142184290992,0.4269247036948465,
 0.5484503953505492,0.3885200620798642,0.0781430021764633,0.4404792070380905,
 0.5526146117677382,0.42754304876182525,0.53742525997417,0.4647753655152309,
 0.08536884832685265,0.5745231560238646,0.1278551828772485,0.6971269651035742,
 0.767137325596725,0.18777845477113275,0.2837028351259938,0.3972074796920322,
 0.2241910887815899,0.4836115231123992,0.9909137363934304,0.8169211842641894,
 0.14634350524969786,0.1185496655940389,0.3706516767896584,0.277676838203199,
 0.3968697162866064,0.15985642912563336,0.1318490278968878,
 0.34793665495609993,0.2171037015000935,0.09648356817598402,
 0.45280132654498706,0.8322998995002843,0.6646762893795505,0.2774667795610113,
 0.1999275089202993,0.7571377185537121,0.07074140536427598,0.1385260489814074,
 0.41756420497741664,0.2392189261767076,0.24634639243077205,
 0.6329755448418308,0.19034794745522388,0.15279806826519404,
 0.4408391580651222,0.5123333540535341,0.3516262714049817,0.6187602979946419,
 0.4892434828393958,0.9677576819527993,0.5709536486786805,0.6309482037439101,
 0.3904483858766118,0.64491959858494,0.9252311222499106,0.34284390127031894,
 0.2775135854393964,0.3778268149687888,0.08020367767042047,0.3438699748198797,
 0.13214898129928454,0.6936560373362813,0.6487375747015974,0.8457910811383156,
 0.4390652155910012,0.11386556102282674,0.3758447081294417,0.202944065777787,
 0.7822358876603261,0.6986450795929448,0.23060707506281253,
 0.0016022504208902388,0.04493363300080366,0.5207657840615594,
 0.1491380521587754,0.4286149420748302,0.5840544191903727,0.24537206856328864,
 0.9787181956446351,0.45999747885899556,0.029524110813591742,
 0.8847703741684414,0.8388476674969432,0.7475047019842301,0.565801879928611,
 0.09357527251728193,0.0794404596007594,0.6528545557317628,0.3975458916847181,
 0.19741178706973606,0.9142599532708349,0.19096804107274745,
 0.10287315283153176,0.7942276122403624,0.1303225107053273,0.731571417127963,
 0.6714057661972773,0.9211431824536146,0.4657074080707452,0.20555351768373764,
 0.5426173411105516,0.5278243279472615,0.01999904129101715,
 0.41911553456020456,0.07552094417853461,0.8943842789570606,
 0.1724361851586249,0.2586516811545453,0.5824644550881647,0.8184806410488908,
 0.9760744334950199,0.8209592474244942,0.7920966192162301,0.7951269619512272,
 0.27530863173524756,0.8414561987649339,0.8626379354061263,
 0.49413941051075005,0.990221101053335,0.48490442525078525,
 0.05160716220547912,0.400016663741385,0.11812210286390235,0.982072593312285,
 0.750770985479854,0.04813451607887398,0.6850397677662428,0.5012608226932536,
 0.2891069730601835,0.7538658986422466,0.13650561504033365,
 0.045806371999502105,0.07532628443663003,0.5188788139835253,
 0.15608769260358557,0.45590297011370584,0.584022503001253,0.9851538648052719,
 0.4041110940302858,0.5005291288530752,0.7466382520506587,0.9550796053622952,
 0.604276263938861,0.9657648458313849,0.3511866363385413,0.21074910721847384,
 0.6420382754373672,0.9278939875495331,0.06271236559031768,0.722580808519966,
 0.14759928566265323,0.05712413751958201,0.5652832649748863,
 0.07136612821995902,0.38507627876401973,0.25789956217767007,
 0.19286129474864588,0.8022805625229332,0.3601019648519608,0.8762922864848405,
 0.12781618146520268,0.2886114710920188,0.7226453331441708,0.4690785185093398,
 0.5850490521091856,0.5779329305877344,0.14919069907205074,
 0.03271553628252477,0.36546798519751245,0.32982794413060756,
 0.1148708131430386,0.8608273037311169,0.4685068803378226,0.12919451216370614,
 0.9860724621220975,0.8915928690348748,0.060152914392899604,
 0.5180526464788969,0.9175505968810798,0.08699028256542896,0.2687639071086668,
 0.9840258869089108,0.7259861176417476,0.8610235519440931,0.2903414323569856,
 0.8702817376885025,0.22575823898588676,0.10530280675677284,
 0.7811403321294565,0.03185034257156938,0.44696055367692233,
 0.07575194981018707,0.030800237046160972,0.08116622064090517,
 0.44506557270048963,0.7772710963503152,0.6119695226926408,
 0.12479390257959411,0.4410553406211334,0.2056488627351173,
 0.08744769762566384,0.7124030991473309,0.5975722498718158,0.7735166354788351,
 0.7024609497031448,0.7551499721993776,0.07855803004797002,0.9146748132710272,
 0.03678018137225869,0.3403568067758217,0.34908468747540644,
 0.8261753406803681,0.07792192530187092,0.7939287334780821,0.9065850622833683,
 0.8716367129751021,0.16302504634476755,0.7761355330415816,0.3884462785496846,
 0.5720569573682971,0.07140080291052753,0.16554873897348088,
 0.9039232004197513,0.38285841305947743,0.38602471480797274,
 0.9768315161919611,0.767854475119495,0.4556874725593151,0.5250004227057867,
 0.8434379618892398,0.19656453674060193,0.059492403717550935,
 0.15816606426966728,0.006599350817924732,0.7306437838440052,
 0.647215073570943,0.6738285359443532,0.40171228297529904,0.9485311359610388,
 0.6541982551705314,0.1099602876503365,0.6359418004243589,0.446024072484388,
 0.1436418499920702,0.2723369533486949,0.8400318325872471,0.28124682183802086,
 0.015329289656675682,0.382044971759288,0.548535333451067,0.892258917051074,
 0.9202414416426336,0.14525311855955414,0.432163895592673,0.47322352498759523,
 0.6857258867668645,0.708061772396726,0.23508158414765667,
 0.038830689110707195,0.22275133583568907,0.5748113214969277,
 0.8743275890634102,0.2285177906316328,0.47342885325205497,0.4802552551438982,
 0.8428264883238171,0.39548116697547553,0.44977724300186206,
 0.10810595797390887,0.1609488431135374,0.6046970149609836,
 0.47223491293860587,0.6506610099012845,0.8296085301392216,
 0.13147396062143435,0.2753852766970749,0.9525757177338708,0.429452297824521,
 0.2671309036730467,0.7492614326033915,0.44186734258052307,0.7121401509318424,
 0.7214444088589801,0.7206018259126994,0.5451177939340419,0.5019506112011318,
 0.7669139096885814,0.5195716221287504,0.6902488451777564,0.7407113976118223,
 0.5654178614063111,0.8919749332205748,0.5834842504254154,0.26834102073976684,
 0.847088695116108,0.11583845970458495,0.671284307692381,0.5104489561568222,
 0.5458289925674674,0.04482856672015523,0.1824131949259249,
 0.37148930414770054,0.45335293285560485,0.8493736495317582,
 0.3700551178590261,0.0037134070305322453,0.5326356766613544,
 0.5333881943870347,0.6465169871709655,0.8587975724759146,0.9994174283789369,
 0.5679384118777286,0.41496044731058246,0.982446893294606,0.2947908661600982,
 0.046434585779800974,0.5268545229736802,0.6131991756575478,
 0.7516263009475648,0.659832346232451,0.13402531699716425,0.4336250129657959,
 0.940152004328517,0.7910478058344433,0.867268159323262,0.8605605642317149,
 0.23283387004732714,0.7288390075362567,0.9446630119811245,
 0.12873938196699342,0.4603153461987206,0.40655381678657754,
 0.19433743136607895,0.16734190378997815,0.908463661053153,
 0.08313332700774956,0.6543160855698964,0.8375198861885927,
 0.19655203951473887,0.8869531218314659,0.3807366625088675,0.319565737567174,
 0.27892601627676317,0.27573260845919045,0.09116146100550537,
 0.12743061998989014,0.1475380995233928,0.5582257689164433,0.5101847892146005,
 0.9897674277767152,0.8119632880155796,0.14313738881492166,0.9132089598477788,
 0.18200278098828448,0.44478140906275676,0.712844231508921,0.5521963529633717,
 0.8278952802719759,0.7849351761224124,0.05949682739468454,0.6288472041782651,
 0.8193760745206249,0.7413955893354099,0.30980492564982365,
 0.039510015769007145,0.7964040565759617,0.21523146589355324,
 0.7678542375080265,0.3612880076960463,0.616640616464734,0.04099444113750428,
 0.692097454939246,0.5941928937667995,0.3891243422343599,0.9097653653756179,
 0.8695438336720448,0.8721196163014889,0.5076219562549402,0.23559717647823386,
 0.49966694818448165,0.9510452478066743,0.18296670045397356,
 0.8326683412493467,0.47813747602908485,0.673429065709072,0.11217229974298049,
 0.7823073533425037,0.00782115050084975,0.733906259599839,0.8781037905467306,
 0.1918713608232273,0.621074815878667,0.5532116022132754,0.8374307024002414,
 0.9187609294343551,0.31833328395549265,0.800625419017419,
 0.007149592363803969,0.9312569419683696,0.06953124295355029,
 0.34524742425834365,0.7452888844675657,0.3715741967228594,0.7512313936541207,
 0.028352087682181448,0.35246192704562684,0.6838405831388289,
 0.6592747306812625,0.1897583553511668,0.3335521281900372,
 0.011478015114285034,0.4497909070501428,0.7090354418791689,
 0.9924148353432622,0.629389218193986,0.172615229402168,0.5122187396706087,
 0.6690078336082367,0.49844309924156416,0.02724714551033114,
 0.10709789626928923,0.2659614469221523,0.9153703655927737,
 0.017243882418443657,0.5498156370348545,0.4220696159086107,
 0.5538045767170918,0.5286611873359874,0.8909983680223511,0.9996153318053075,
 0.21729002626679927,0.31580327749949877,0.9835631341097544,
 0.8631781132980771,0.06725027614823009,0.7462026777566791,0.5247671966414641,
 0.5042622591395807,0.2625993601599248,0.3949490952731157,0.7963293988351987,
 0.82579243817555,0.751612848018963,0.3475903376706353,0.46710375041275753,
 0.19925307305857953,0.053982707261009955,0.79840375780902,
 0.18908955035099873,0.680363598495875,0.22138465332127522,0.6089363544088532,
 0.7975895331159322,0.2581990433906912,0.5561780782156444,0.5063140595711899,
 0.8224901099520437,0.49252712776410534,0.39955085570480375,
 0.9875213056244321,0.46102549679263716,0.7648901410772742,0.8429903267756659,
 0.5666276946128967,0.5320091478865324,0.7143445134269761,0.375762529566931,
 0.25398385763930253,0.8461274181997915,0.9993151408434257,0.863834898096752,
 0.31491311261021016,0.5364469105010189,0.030747910801223077,
 0.09107342814326858,0.05849859035310523,0.6449700222153493,
 0.8270528248979243,0.9632745247331196,0.6608670039525407,0.1631196190977856,
 0.39581646885352784,0.9462427120869746,0.23051238364013837,
 0.07162384241632613,0.9896477478629864,0.30957950487429675,
 0.5705169196947055,0.8360348011778489,0.7477682639462708,0.5452675305733091,
 0.7473577771381211,0.378458753148031,0.31955020820031044,0.11743752453655165,
 0.6349703204402031,0.4650393282944685,0.994203538289876,0.7780499612545113,
 0.14026821461764194,0.138907886390085,0.06012776718439872,0.9794596635381807,
 0.9246515557841143,0.27933660993755227,0.13716290893701033,
 0.8696919451508399,0.34584793268252745,0.6742008305881508,0.6551431863866344,
 0.6055937910901988,0.10748171815345597,0.7902931372248579,
 0.23104574721720572,0.7507311885378147,0.22425774052236236,
 0.0588935463664958,0.2867451501716741,0.3149961963641241,
 0.005504976749012069,0.17338826322300926,0.33147107434503,
 0.42967978947721674,0.8458420499731961,0.9616483656366824,0.7407039243489434,
 0.10658381820404805,0.21963296725027215,0.14456791297354,0.01994644545380364,
 0.7951865635191705,0.14898876756016033,0.4576000503296981,0.971902293355997,
 0.7628046635901614,0.26472140123911525,0.0335345908017981,
 0.39375710826453636,0.23488539884996784,0.4962307168760184,
 0.0077029474947190035,0.2681412198369374,0.616824307906007,
 0.6580445718088822,0.7457669171554127,0.5328341947418499,0.14960322654624225,
 0.7986430257893492,0.6846078451730496,0.28810166323096253,
 0.23561869520614154,0.6660575329645717,0.36369337793870193,
 0.5436493145199319,0.16825350885026102,0.2138493119654209,0.9296916104640666,
 0.8982588539496343,0.6884423024324777,0.5942815714502865,0.8498897515705033,
 0.7713914307505463,0.49570464045646023,0.6491078941756644,0.9418674411234452,
 0.12676484541118227,0.13019059208320705,0.1649829614194519,
 0.27731267816942684,0.6579596480216765,0.16297694565611298,
 0.7764662396358133,0.7591522178917189,0.027916544359862794,
 0.8442595964706128,0.880443778828532,0.8298894375327734,0.38299981612383405,
 0.09453078962053652,0.543425169167038,0.10986253132372825,0.2610217855963083,
 0.10134578634534352,0.45357424336870344,0.8133905807219357,
 0.35175884792670553,0.5075541673855783,0.9522247492026249,
 0.14101493993962466,0.7101845107167198,0.5008775393002525,0.789824810854699,
 0.11173131149657878,0.8435516532617463,0.73535230530399,0.7157681381327382,
 0.622437579971591,0.6107500885137579,0.8259505532161013,0.3327881294032544,
 0.6334110008030437,0.39696621757065687,0.8058448053913266,0.5171981393820497,
 0.4493036653017538,0.12411097869343446,0.3751064508442048,0.2755262648209845,
 0.31227769629587,0.8621726356765295,0.501388574813407,0.49872531323911806,
 0.04033771463610669,0.9419948818406033,0.25748868999795704,
 0.011278162331028407,0.7742591537785559,0.41280296420788776,
 0.2744876320874936,0.9137749375278342,0.5430646410999586,0.19761343595859227,
 0.09422318995118184,0.4160988260919337,0.13390412035413302,
 0.9924314321478758,0.49635203207669853,0.5872164598597658,0.329331140273987,
 0.4079287069342813,0.0020467099638502573,0.8902471172265761,
 0.14606683960789169,0.4825694364098003,0.3797761506154994,0.53677971372401,
 0.9671535391935993,0.29652656013267786,0.5818479533216069,0.6146548430013928,
 0.6523335862124804,0.8697403681080017,0.30330079702437374,0.9149569274786643,
 0.9212993419085858,0.9116559824385034,0.9875811125505226,0.4759712443819092,
 0.024630211597308493,0.9966298862218014,0.8352968623523283,
 0.7285708242460587,0.8888573737821278,0.7597055127537147,0.2501089916704674,
 0.9339944286531618,0.3041341017453931,0.7463751195786342,0.638745468979189,
 0.7928729901070479,0.9073289251303771,0.6497227551931153,0.7810794174277081,
 0.047530433703418695,0.8856397606815879,0.5465939369487266,
 0.038218994317620214,0.0788539353780211,0.19279518725305456,
 0.7380605083944064,0.05288586787755467,0.5270130832944679,0.209846685814276,
 0.5373950482896866,0.6999699662247333,0.028585182730709202,
 0.9484124135209782,0.21941372205537513,0.17232890448907945,
 0.3689533306657906,0.32447588316945297,0.6821125360170711,0.8866231810809374,
 0.2293701612277722,0.9946772122548363,0.5241168761899682,0.5686782669376144,
 0.7879045581222889,0.8260286138059045,0.382859338026039,0.6576230995263782,
 0.03764461359734339,0.14280580763903683,0.4617023138773143,
 0.9044482660758603,0.559949902420527,0.0071551742087589965,
 0.8573499412908547,0.045280365982778514,0.3103477784023332,
 0.7057132903231172,0.3618948047466223,0.6854133386989216,0.20895174495129065,
 0.6524664062295236,0.037191599306542034,0.19248558741651378,
 0.564695172522062,0.05766407512278471,0.6240429694457185,0.5479582110120477,
 0.5686867419678945,0.40015471317459983,0.5747935724278384,
 0.32714172547505826,0.8556119095697519,0.22381115538580476,
 0.20005466834883545,0.20277203272028443,0.33450390207063563,
 0.16812986875680658,0.6041873397263219,0.9054659575421953,0.6149634182465952,
 0.5057575539424052,0.718591565646433,0.02999122193106707,0.3430402757621869,
 0.38567122664739983,0.11778597812360159,0.3192040600112004,
 0.8313506637708095,0.6826856820627605,0.438948916942725,0.35981753923943993,
 0.4451124306707521,0.5037512377994088,0.8479557157435381,0.7886704588009035,
 0.7117761491128183,0.45824914383232573,0.5863915704301063,0.5508547461844714,
 0.7302236035164799,0.5365913771767784,0.7325168642839921,0.8979516644225638,
 0.9579299314629832,0.5295034551001737,0.4641293992048813,0.9058818884058457,
 0.9428593905802667,0.40252391306589186,0.24459288418423264,
 0.012205916202851053,0.5554777551026991,0.32075109727040385,
 0.08215510841393603,0.14335619446771253,0.03565287239295156,
 0.8619514095503877,0.68529066172264,0.5270021565271932,0.19540838296915886,
 0.3638285442228696,0.3936274718072501,0.16036231965006964,0.2761894963325202,
 0.9033107921734289,0.48329064528896926,0.6791658220017577,
 0.22586656496474333,0.7266698824367388,0.4109192832687898,0.0759446101586514,
 0.5158902183352739,0.010084047555599773,0.8854020638538691,
 0.2750251303492641,0.3700201634189727,0.5181547318857741,0.6244468322011077,
 0.31591760944421543,0.396839379851752,0.18238876944651894,0.6704188812960474,
 0.00519908284071946,0.011330858490393503,0.14815323114940016,
 0.026728457583258924,0.3336089806040644,0.8667784067514455,
 0.2020623754206261,0.981053348834902,0.5496253193872488,0.9088010578473615,
 0.5836178520765518,0.15988169259891316,0.6453567470796631,0.7180917578037584,
 0.02643878266651356,0.638758506805079,0.6255909539542538,0.7217607200076661,
 0.5609290308082125,0.8844857950832682,0.29386346289411003,0.1072917916081445,
 0.41156889096209015,0.4076928690157534,0.10640085549486411,
 0.11748655398081165,0.06706925767477667,0.749433988667254,0.4512671514030273,
 0.45938432126663886,0.6481292367395118,0.02030805010562997,
 0.4789675556492061,0.9415546429391088,0.5545504573612483,0.19121176342053714,
 0.4588855255377866,0.661099344930362,0.9280821096824118,0.9326757777793456,
 0.1466354544702183,0.6514847734395504,0.9998095713402922,0.7817475093825769,
 0.44484127665830187,0.7521770046177951,0.7346510184435772,
 0.27400646245318816,0.6296985384001972,0.6629998966266302,0.0673497519199403,
 0.8860616036226713,0.22543017203278093,0.9701559910672691,
 0.17461953976095557,0.4080763190444151,0.9870905864827116,
 0.40425889217183375,0.30491801220778303,0.27916525088912914,
 0.011347782047837152,0.31420630508797387,0.5244268537801353,
 0.26483654302618465,0.0917544842393242,0.817914250497603,0.4315703322579677,
 0.8063986131825999,0.5413117301252469,0.844222754060932,0.11696966094272143,
 0.3570863254162552,0.43876432323102066,0.29171190242295253,
 0.4331619218851984,0.24709619711152897,0.954394595795568,0.2683884727525382,
 0.5094234088577957,0.6146311607348866,0.03921294877237902,
 0.20908493781266146,0.9532417570259288,0.74366084988653,0.9328286666675313,
 0.21697038837446236,0.5779497345270752,0.8029423017377737,0.5217470722644756,
 0.5724578040684616,0.4649945622914785,0.35916251366724117,
 0.30465203041641553,0.16656454136608623,0.4269691029331022,0.879941982981889,
 0.04748634816438568,0.19790542435056568,0.8530356196494486,
 0.3503128960922357,0.200716446070351,0.318159309086708,0.9898807478599785,
 0.268298092779932,0.799734356849212,0.6283557629416272,0.25435941973301057,
 0.6021798532230941,0.8177035557923753,0.672264857236262,0.7598192883131761,
 0.9392940240802278,0.29295571020499467,0.6703324755316626,
 0.08866609068321551,0.22702445540545246,0.7220378763427868,
 0.8348027188549184,0.2747695930611588,0.7778895868850175,0.6668976499773497,
 0.4482611995620127,0.26757311600414935,0.6130722341959999,
 0.22143450274870125,0.3411400918896913,0.6732159265007038,
 0.12347742286811436,0.5432436057830333,0.6818879636353917,
 0.02538805559439794,0.07263093597160886,0.7635261401283547,
 0.01106957628653138,0.742768422287131,0.5755488407110577,0.3198746224683824,
 0.5469211289289049,0.5711261719694926,0.8164273475853804,0.5755378230095053,
 0.9921902983207327,0.21332705906244942,0.5495124962535277,0.6597425594610238,
 0.04128772238996925,0.15191831747996953,0.39849458030648943,
 0.8323257021449626,0.2631695841694742,0.2070813830744681,0.7318951665507637,
 0.3980743672930316,0.07100697369061582,0.9516578505102035,0.4037729435487617,
 0.4199524818892928,0.6796301148710898,0.2019366968339128,0.7358962019797821,
 0.3813535131644782,0.41345176311285303,0.2954236777053001,
 0.09075405157978222,0.6149439120368523,0.4881955657904282,
 0.40688013573209125,0.4666794069154574,0.7534015100551441,0.7502079742789303,
 0.2621101980736129,0.01565606562041011,0.9717971479073384,
 0.03646412703317259,0.4572211940570384,0.40177914360342615,
 0.20342353480011177,0.17197470840589757,0.8465260817302446,
 0.6246805798298269,0.17576118828602816,0.1123414178926172,0.7978194731906542,
 0.3763158871779393,0.6440536284102281,0.8149290400164305,0.4675198880119661,
 0.53594191313276,0.6408843255763292,0.7479989348848831,0.776229348415488,
 0.22000108998750945,0.18599332289991266,0.2700934759931879,
 0.09339830086750467,0.8661511347313715,0.8660842826224759,0.2664000441358497,
 0.766069607989661,0.3099678958989156,0.22152709507031243,0.7948246388283955,
 0.4582171671342541,0.7023489611769815,0.8180216468890465,0.4227085770257376,
 0.6572631942848832,0.5586289480824432,0.9069774117542712,0.618897782679048,
 0.09153392705193619,0.29084090935815643,0.5213835597493611,
 0.6603193332282423,0.45088381470043537,0.574676624701482,0.3549514520399416,
 0.9315107763818428,0.4425837934333078,0.18215278429228043,
 0.30050220352270984,0.6081602302417061,0.0600674019654549,0.5712251633302757,
 0.7592762094692916,0.24673912771401307,0.012660517121862225,
 0.4211308023855238,0.4686343801117244,0.6121692790022719,
 0.008002269024135922,0.4253377292769728,0.9847295890265539,
 0.38191207958614726,0.25625695607523047,0.20468870776464065,
 0.9352198585431005,0.20211914598764347,0.6790557998162212,
 0.030746933079856098,0.06774385061808585,0.6485906198124015,
 0.190056518431015,0.3840515643702338,0.9948394797499897,0.6561625910859274,
 0.6615691517318478,0.8017345307290424,0.28078183777073584,0.8516456133033969,
 0.23344874894239842,0.16485179341420253,0.4988543233684092,
 0.8663950545871983,0.25429521325195303,0.5438877979647282,0.7673017178212072,
 0.6128589201715953,0.6485357651988983,0.4595002776089979,0.6013862735587777,
 0.5565177172330782,0.9697636585692905,0.043121059046781984,
 0.8060085022462363,0.7027386128298008,0.703275463775945,0.7242427490463477,
 0.5027119279270468,0.9586714250669727,0.9977941849808515,0.2578213810865976,
 0.7954620910912352,0.9853077753570005,0.5110841764636129,0.5421683392978448,
 0.5077363237622758,0.8860962047724842,0.19528857149744638,
 0.39140073765447825,0.6430180546930921,0.45189644192762113,
 0.4213535133143087,0.09317668235768217,0.6998791140343679,
 0.10567393339785591,0.7196839082329183,0.6560302636989126,0.3070368420020311,
 0.729598322355593,0.15142793382502706,0.49101552472216803,
 0.49116431863200727,0.10573256973058454,0.08806203020842007,
 0.6026705403231838,0.43134642900748343,0.22400424454641576,
 0.6432521570086932,0.5445468540448535,0.5738926224001346,0.15993146674063785,
 0.21433003038394638,0.822464056695208,0.08888261456884883,0.4140414117720628,
 0.5639884620849532,0.6658733707842905,0.12175974110239274,
 0.12299068880020603,0.7608213727403049,0.7358388419118442,
 0.16585286924274079,0.1287442913775516,0.5090333057190609,
 0.30120036515658966,0.9075750010989526,0.17481674807034497,
 0.7525677355037194,0.4571779171223964,0.28494990181288826,
 0.42049413724375606,0.38986409549031553,0.703222876615923,
 0.14887516648602883,0.7539736681525925,0.1387261820276826,0.7540185068592098,
 0.952789220490589,0.7747990581795096,0.20308001553670219,0.7737118005464961,
 0.8962202233324559,0.09088029851026613,0.5380126520623207,
 0.30331625077791147,0.6616117681845969,0.48389457612497233,
 0.25134365270480374,0.7121670314783666,0.9533632396107351,0.8799670820415553,
 0.8801392892985127,0.455542830812559,0.11624303428920213,0.49889162513511953,
 0.9440741454967102,0.20137215441537282,0.9790284184732191,0.2167458572698635,
 0.4522086723643194,0.2178725344552529,0.5938283087112093,0.2656373426650095,
 0.708508901002682,0.5716532509947548,0.6011627424988119,0.7009627194739877,
 0.5675846876992476,0.817731812765164,0.608387295126689,0.2014018333659673,
 0.4597277511661454,0.7731347493221511,0.1608545463366713,0.4563628420201439,
 0.13784217625507011,0.9630859523445331,0.6168488418583634,0.3820563020777151,
 0.7942657369950988,0.9003095542918658,0.38967943464035315,0.6505314902508692,
 0.8436172522488046,0.7646281559840432,0.35565448973522407,0.8620878577037904,
 0.30326332714953075,0.6775133544467271,0.39158408727002136,
 0.42639805997638724,0.7024115504689646,0.043539837495930156,
 0.48266820490007667,0.6491260392531524,0.8170706926977844,0.3799567978297286,
 0.41872413202455616,0.9722935331734137,0.3121698115830316,0.7990278493004731,
 0.41003945383831586,0.05953890211396751,0.8966316168039463,
 0.0257512464923646,0.7336871574547013,0.15323297311432815,
 0.026743562799532716,0.530432026660895,0.7563278309408668,
 0.46296653051799264,0.3024340414505201,0.2822630596148651,0.9277851716175235,
 0.037250225913191826,0.2559212462482173,0.3937478789930464,
 0.2842789911889637,0.3088095749309272,0.1484503236711211,0.525826515984685,
 0.863796921590333,0.7327556752954205,0.6548011373298065,0.49642814120309586,
 0.382488156544486,0.15907841880306084,0.7574590203545296,0.3427899539809607,
 0.06771460480576896,0.4263919104516236,0.9333066125802939,0.9232115935994539,
 0.9977484119362898,0.8634788220188909,0.1939322618562237,0.23880263988193562,
 0.007834535654715369,0.02606244787919476,0.6502903113094556,
 0.44306877874809636,0.8012664841931956,0.26871857895929496,
 0.4241313628080877,0.6587760735465076,0.7466646886527495,0.7813272631069104,
 0.44612371565758335,0.09634077265382401,0.7276951224396895,
 0.06330648226728275,0.13737943824595278,0.5087700917725642,
 0.3562267089011363,0.9518365859814264,0.16422118117018114,
 0.44862358243607503,0.2340508218601014,0.47502305564728686,
 0.5905582210653249,0.9831016123284582,0.6530952813078988,0.925771550568236,
 0.942397926279704,0.7497351501685875,0.4545704787238616,0.6649488645612573,
 0.4243017444931477,0.6207159126365579,0.8168310760106738,0.3057234803306831,
 0.7270001144936749,0.40199346725964524,0.8119247302595028,
 0.38307904986036956,0.4871430857214315,0.4310833023774099,0.8431090570277198,
 0.9207101424868888,0.2703232770736069,0.4541104882433322,0.33094672938161696,
 0.8971381602080242,0.08246251212678724,0.5715852578041184,0.5168495452984767,
 0.8646978249783599,0.37271096183325436,0.49920052011343063,
 0.46188575107627683,0.8004436220053801,0.6727439558513598,
 0.05006280732861634,0.04390485806099631,0.5581291352656108,
 0.3553553685951151,0.13166264631975533,0.5390028668586024,0.1682894439331949,
 0.9148385352553869,0.5762572827868292,0.8077518903178115,0.6989481258361605,
 0.4907337259658634,0.4454968052988927,0.06334020655092576,0.622148342839995,
 0.2070664122785486,0.8102107618924461,0.11696936712220729,0.9224143452168704,
 0.036035757210486685,0.08751250714444347,0.15915141196082638,
 0.5846428948643014,0.20146236642915072,0.6542401651188534,0.7528032074476968,
 0.5860020414719554,0.4236397982928498,0.05530068348208694,0.1319566440823553,
 0.09427329271072771,0.02290773767469223,0.9730373255521535,
 0.17822662675908185,0.6076206939899255,0.19455816870700204,0.884930676421441,
 0.7920380805085534,0.5894722742253871,0.9064642867862333,0.9452575012467271,
 0.34617343451950255,0.8167736066188096,0.10902331739409765,
 0.36528272179964316,0.9851074511951572,0.723787030869376,0.6943271593322923,
 0.7298601112726559,0.7027853695523929,0.3447452360079083,0.18976187636741493,
 0.24870099579634963,0.2777332423219203,0.26429762233799303,
 0.5037082787983435,0.15847446400439014,0.902473520365762,0.6555976717416939,
 0.46241926492169083,0.9968624153340576,0.9239227553346288,0.7343502617426975,
 0.03780654378016435,0.4942437908694898,0.8698863617975987,0.5372329176379526,
 0.6125920649397298,0.6071418596208451,0.7708220385603695,0.2791363618202596,
 0.3501161019200968,0.4254603216272632,0.8778340073612796,0.06961566502505934,
 0.11944308833066208,0.4225475785166133,0.432836440934102,0.7175581897900687,
 0.8346125705909448,0.813465798996405,0.7559789148267754,0.7129055385992437,
 0.4805250277222276,0.7495293744419268,0.2093822197287859,0.1325401420901251,
 0.4772228710327089,0.011868827260798964,0.8770407179031736,
 0.6469577979111096,0.5698863334416157,0.6321494563876048,0.03244128203361052,
 0.8451774476617796,0.34818263650579184,0.23440537372888115,
 0.0780159294612891,0.26637858827105965,0.09546028516172966,
 0.6922887166191447,0.9078021801006846,0.5458919681641051,0.23830678919655546,
 0.5702718899459726,0.9077171435864381,0.004370223413757701,
 0.13999633737282724,0.6836105956349912,0.9291112990446657,0.764806350047902,
 0.04142926538140901,0.7022564202793481,0.7633295506175813,0.5677661202537245,
 0.5493526199162172,0.7827258591586175,0.7392019544548144,0.3516717991047451,
 0.7317882811983124,0.3927916167528138,0.30262222992616183,0.8696040926824213,
 0.22758490714507085,0.1260640109452742,0.2532833197784826,
 0.013658788944599043,0.3013538587415793,0.43254755503785336,
 0.547835564899761,0.39779038277699963,0.36376685002926734,
 0.19477674881547435,0.7346893917135736,0.7856095803045258,0.3987129733498569,
 0.13919313890652063,0.5010851104719112,0.2963929570848932,
 0.35810852641791924,0.813505496135568,0.8820210928948042,0.39530687978309476,
 0.7842348782105024,0.5532121511290147,0.9911071457930215,0.7726420881641836,
 0.9274256349641753,0.10370407309774177,0.4075904238956517,0.6657833132371904,
 0.7492339823962614,0.19585256184370403,0.3388689444138637,0.6135284796603708,
 0.7422777685167873,0.2003539830395763,0.028970808777283308,
 0.3910469319960528,0.8293166212186013,0.74272297890864,0.10498921925335447,
 0.9482119870887669,0.5916635924400648,0.5349230435516101,0.7403283986788537,
 0.6440407881633076,0.20407712886066476,0.9369468140574366,0.4985598838854475,
 0.2875828611200747,0.030657128960194013,0.13392916328617388,
 0.15096422654971753,0.9752491353096738,0.5985279934609026,
 0.22223834971938705,0.6661718785551921,0.8807258666866559,0.9911093982756954,
 0.24105515935113686,0.3940524672746092,0.3036556394422987,0.372244292309199,
 0.4518719179527696,0.6608113543988857,0.8527201401559783,0.05412235065711013,
 0.04830252252442302,0.4392227287499342,0.8046402108382497,0.4144110507512896,
 0.6158243059555525,0.9845675357504946,0.07739694477543346,0.5767381155634025,
 0.44767774436228225,0.507623968641068,0.895083625859337,0.982060405436513,
 0.785587087270476,0.8453100359469089,0.40470916238104837,0.678939105961609,
 0.2704574561300753,0.273703982479901,0.9788371671366467,0.661288354547843,
 0.8938577520354497,0.5587256078951177,0.8530575367491928,0.8877861401925276,
 0.5340867727312235,0.1177240509841806,0.9788799720529577,0.5786373655419268,
 0.1479184298004479,0.7781794604183417,0.21309810886020397,0.3748832588524811,
 0.5117641985397021,0.21899056698143604,0.3869133930216295,0.7236999351589195,
 0.6336380100120923,0.46534693934928284,0.24259825821250647,
 0.7664244968238765,0.5576548154466949,0.01949354597333164,0.5225766817611981,
 0.07574155729018084,0.684153402619088,0.7387304513683142,0.9325215169327631,
 0.05994154446196953,0.036564772336381,0.7805722798767183,0.7326523361284839,
 0.05970736011955868,0.5202739675827779,0.741340692571686,0.3472179957517003,
 0.25961359531978223,0.5926970980680285,0.013957123579041997,
 0.2814997527790012,0.6072610674454666,0.6809166047194525,0.8622896960598521,
 0.42768730796006493,0.30732117229146294,0.11797953161663288,
 0.5063270111991774,0.31481159339692855,0.08797614264841214,
 0.4368830603945978,0.5908785957845822,0.372083097373757,0.4811868811224903,
 0.9039240233367192,0.054960462208023086,0.6233619319619413,
 0.9479806607137273,0.9924488482747417,0.6549543766271073,0.3906021019202178,
 0.3158676504133707,0.9254262208326961,0.6715297693143407,0.6390588380569607,
 0.2856604738562315,0.32415655246037245,0.27175839534212354,
 0.4359797099735747,0.621121787047234,0.45480762877524605,0.7911957762555903,
 0.9263719835311215,0.6182734078273884,0.273686203083626,0.5516371102405236,
 0.42660014400062596,0.27628113123805476,0.5070180140600452,
 0.5987024082305585,0.5284516214505814,0.7098123445802609,0.8999113922594039,
 0.6602521417027559,0.3738407829285362,0.4787199287826538,0.7251196262158495,
 0.07658810318908305,0.15974698542193266,0.16122706271406861,
 0.9727762737781322,0.289376161493587,0.49716398063613654,
 0.007360249508480132,0.1788723808285454,0.5974441718483635,
 0.20176840514252425,0.9729897845114515,0.6987406231062342,0.9667350426627249,
 0.08843452159386156,0.5463650056461995,0.6326312045719289,0.970762573401724,
 0.5443146487548095,0.2454493715136855,0.5741195384862161,0.2501581140500134,
 0.14347795816529918,0.13394221627640646,0.6283226446356976,
 0.6643759247216083,0.08651270262407973,0.4509546617414585,0.8305998275803373,
 0.2887464379471165,0.6775781692002927,0.6274829651163887,
 0.055593798406071704,0.604137322836358,0.2604499567295857,
 0.12336069863196464,0.3615388394184711,0.7244999520922437,0.5103875205326723,
 0.5645366688339535,0.13160671824601988,0.8530046431008412,
 0.48396128291997254,0.0016218175290967363,0.9674460203087009,
 0.9429540067171258,0.07598774953169363,0.5392690878305881,0.8699161305203633,
 0.9595040836970387,0.38542793701360556,0.03886574781080032,
 0.5461992944722674,0.14026572072612842,0.6547655147245112,0.6118312801717373,
 0.8862212857897701,0.016912353045293926,0.4134098110544444,
 0.7222803572980347,0.8666641623883697,0.5122390922894173,0.18255610001661438,
 0.9897127900875415,0.08029653491128552,0.6479313356782166,0.5905443242244771,
 0.7846285882813463,0.8120121013430225,0.39987770259184063,0.6902795813028773,
 0.06384673736267132,0.904572306763722,0.32384829129546855,0.9942959134936776,
 0.10529059547816855,0.67678117686962,0.8273751403621515,0.01744912663530873,
 0.06185439675738569,0.4475094549384623,0.7729471558077332,
 0.45653604830847094,0.48575229289234667,0.47233459184605564,
 0.9859413038608577,0.4342149403784328,0.32377152426880895,
 0.14537841853442435,0.6966859579778428,0.3140103376680379,0.8330144863867947,
 0.17159539394161505,0.3462914593286629,0.10086601062113565,
 0.2649467863747539,0.7876641885246093,0.21987716567007687,0.2500917023146555,
 0.1793773611185181,0.9244373936435988,0.5659528134694669,0.22422381100451994,
 0.6743405746414843,0.7439046708638363,0.2697667552550792,0.8237865098313254,
 0.7372264481013143,0.32897139455997226,0.07567473303864092,
 0.9001835391782306,0.4621806172763625,0.16348273436314487,0.7583381297815217,
 0.06390554838337137,0.9151735573753145,0.45516529079773127,
 0.8755498819646791,0.22916649667300082,0.17604873321485337,
 0.8620662920372479,0.934686366598883,0.8564171184997511,0.12450244583213965,
 0.9495375997109096,0.9751588942164313,0.4226535125406977,
 0.004270831872279168,0.705205419886907,0.6861207788529877,0.435256382311181,
 0.21108004623494914,0.4830772771062525,0.6403702937568485,
 0.20504894708205357,0.3402649093116523,0.12018504462597768,
 0.4107552492708194,0.6332672183160947,0.9248978732527369,0.6879482354591921,
 0.8802185254621484,0.3268052621232612,0.8596899954612929,0.774203066557119,
 0.7110699968948371,0.4388069899938214,0.15501053514427643,0.3753423306636974,
 0.02316223743621104,0.18304197245700404,0.7034715653803499,
 0.6639855967495603,0.5414159985968071,0.9051503677227449,0.4263380157683415,
 0.8532089439251128,0.6700608669256799,0.5142027481175975,0.933052732219872,
 0.4654319351742656,0.8298607071000864,0.4213570831873972,0.9098416992579426,
 0.24871831553482515,0.19096738183749862,0.352191388219952,
 0.48745203624750044,0.26134961428643055,0.4574194393364813,
 0.25969111072263584,0.11317085871971233,0.05822890574164907,
 0.3586123458069458,0.22617220189643794,0.8244996604824582,0.6292980413316008,
 0.37058337279076214,0.9379081197690702,0.14805694430663197,
 0.08209011488220297,0.6774535314765702,0.3552482631693036,0.9400986658838664,
 0.23559627809197548,0.49050071600094525,0.5818629617894637,
 0.5164847961692364,0.2938580366020874,0.22205455130092366,0.6157582358905818,
 0.3266405505564365,0.8881203748855873,0.7309254801675217,0.3808611875967498,
 0.1893110796712303,0.49052243866928313,0.6888788397334771,
 0.014774401634873091,0.924099404299571,0.21615204662374043,
 0.20211927734754043,0.6354618017753909,0.536392894249166,0.8974120713175258,
 0.41200326800419784,0.10727771705602351,0.7840330062317075,
 0.15808436667609616,0.48025694215806314,0.8009024517621958,
 0.5202884117507479,0.34121486337844986,0.7830415698520097,0.4515213515975133,
 0.3289648518580388,0.19296620526208852,0.2357310173141105,0.6006294961070965,
 0.08018496765455341,0.14481916793431937,0.24594175926481598,
 0.41946361877829474,0.4869727823269706,0.8793079347132124,
 0.32557491868573285,0.46782033111264476,0.8911746302307545,
 0.41467412360762357,0.30362839076738,0.4744272622820376,0.3121129789171043,
 0.9627351429631046,0.08696242134294829,0.9609449323802765,0.5043431778704485,
 0.36689995645396745,0.3830414947692663,0.4998193606595147,
 0.35982909569171917,0.20402130523734208,0.007967539969568449,
 0.3722190326961792,0.4677589515855234,0.29352908493607266,0.6644889404433487,
 0.7034785940015565,0.6820480363175379,0.4418949564237349,0.20955087478377066,
 0.31919794192370243,0.7093964072909789,0.23930728091164344,
 0.6828247951924655,0.6897626011096544,0.9647408072935632,0.08788916605269759,
 0.3716583277418577,0.675527617522515,0.4521350601518018,0.5913081412412211,
 0.8473381203498845,0.5121800264657945,0.5989023333923553,0.40007369824189665,
 0.06787791750736205,0.7811851873358654,0.9200368107700175,0.6944507078126386,
 0.8597384105073054,0.48702085493886504,0.4528508357957638,
 0.34344679839107994,0.09999552485733165,0.4340220645059386,
 0.9972832540174048,0.809317750918758,0.36832361290846116,0.6060860787908758,
 0.8265332961795129,0.79058950634627,0.8174874135839355,0.661110673288032,
 0.9045547082744481,0.24014073560030136,0.18406068954420673,
 0.5481285848340144,0.8818940737104594,0.9699696562948241,0.6537011645434803,
 0.4641620254431278,0.6416409142662962,0.515059310090522,0.5716590862009936,
 0.6790861512194928,0.5839023247911561,0.4338505942664125,0.6450344981796524,
 0.3055957138343979,0.14234870346123984,0.8602604646401992,
 0.30059943345538664,0.8682931153561062,0.8153258938892394,0.3205989497528504,
 0.8510634071651927,0.7584179704348806,0.7490844330712054,0.6989296407893631,
 0.39020750888899314,0.5763267440604334,0.5426659928341238,0.5217880069457904,
 0.62608671307963,0.2789619000025252,0.9413598790818185,0.9235415909963086,
 0.1685634096465929,0.28179454008129046,0.12196300985620634,
 0.07128729659134769,0.479819577344201,0.6882858399157716,0.4350663119336693,
 0.7328872077149735,0.757845853589085,0.8691450651099255,0.12513472994282793,
 0.2677532658517009,0.46027033248914473,0.24622603348578342,
 0.1472782744172012,0.29774783400455784,0.39521045377785446,
 0.8437774959707788,0.30606750169555996,0.9995037238957046,0.8373882158914956,
 0.7371761003527004,0.8151551133671022,0.5398996746100098,0.6684934917177003,
 0.40659841732445445,0.09291789857283705,0.9884747622664202,
 0.7413793869256615,0.3246468610886264,0.3466743714204944,0.7068400178955797,
 0.09781444636895653,0.08793560244738297,0.7735265065413519,
 0.031167650688001536,0.06646416487311613,0.05026598135797644,
 0.052472705229607564,0.6648223292347871,0.0693039286604974,
 0.7311958208507998,0.5154404769836456,0.12057660433760997,0.8801944603650294,
 0.5595635022016524,0.751334397257871,0.004555631633262558,0.3934568980048011,
 0.12818567048671214,0.26563626749932656,0.6035896309508676,
 0.15789686884377152,0.22195935973502245,0.30159534011575784,
 0.8778939337887206,0.5320857266692531,0.43978119200460974,
 0.07503754010120844,0.7800164782990074,0.4429579713426478,0.8337368257083719,
 0.3450320564594106,0.9394439508264628,0.6916031133498349,0.7035118934349459,
 0.3456927296255483,0.4021938128773448,0.8025914689831604,0.858996356576019,
 0.8259051281972565,0.3308669932621885,0.37444827322495977,0.6052342566335961,
 0.6485453411811443,0.3898630843025933,0.41384395281515707,0.6649696752245371,
 0.12651864186131923,0.286073000837745,0.4770292132764936,0.6331429379892433,
 0.853184772920392,0.06972251410911146,0.9109812407205247,0.6513058391022528,
 0.2585916693009177,0.7845403154222722,0.9885362146220563,0.5874260765571636,
 0.7768025542507283,0.19647180549829302,0.981119183275102,0.9281263102418387,
 0.06667488349212003,0.5257878237007295,0.08323149885942549,
 0.9674654716391897,0.5796309270785488,0.16385988805411866,0.6770807358308827,
 0.7733924089588263,0.12141061452864044,0.5964414495517577,0.4591449457778106,
 0.7552021252585445,0.9074653044414824,0.8821777872807938,0.21684390833577205,
 0.6025714394938252,0.1876782971772596,0.5168183657853667,0.32193260726982653,
 0.9139009357496998,0.9669333859233998,0.3366094437014351,0.7837368995324773,
 0.8936378035091067,0.15724784387622692,0.4620004158566027,0.6169788204971469,
 0.3295362236938564,0.49712451435678995,0.5448153802270681,0.5726826078246976,
 0.8971294356744949,0.7976387849848411,0.23708911693629398,0.9005645025148391,
 0.0687930482598148,0.37640046132642213,0.010869737310342242,
 0.6019527370743449,0.793720180052294,0.7015760787786385,0.7638212681391577,
 0.9624004312929128,0.850229178388174,0.2594225313048072,0.76220991079669,
 0.6065496591530384,0.05893364757114372,0.1373520857818824,0.5747264574526292,
 0.9233852297798246,0.6249089404840361,0.7850937714978437,0.07918346229530449,
 0.20460287482401696,0.5026772200928892,0.017950278423254362,
 0.8942242130077004,0.8069107501678784,0.22371838480570116,0.4540537588996798,
 0.05150865776809854,0.7918262984628538,0.36100187735706113,
 0.7933238220344592,0.6500376805910836,0.49341808129189313,
 0.12686659344298734,0.8458771411913584,0.78627478777005,0.9425528009367486,
 0.05522551700301692,0.3125942677424194,0.3328125933878048,0.0441792215753285,
 0.09920758968253329,0.3664794039332885,0.19478437521849568,
 0.3854010485960656,0.43945002902925523,0.5894116083801861,
 0.34609721473071386,0.15047397746317603,0.17298551127799167,
 0.9350119249520334,0.6274624159243264,0.8425514002089218,0.45033477749310447,
 0.11118668296688372,0.7828851763786969,0.28848049510800744,
 0.8145846651776549,0.0050219180917796535,0.6714211590305746,
 0.35317078083198494,0.00802427878113765,0.365869812401157,0.7059081222615347,
 0.1361121051875589,0.18552393419767232,0.6010601063539144,
 0.34089572517572475,4.1893260757275286e-4,0.5238673734816748,
 0.6853984042305244,0.8094686083562301,0.758487459164842,0.20292794776228384,
 0.900583576110356,0.13661786956280564,0.5013322665733694,0.18527186634971626,
 0.7201232394043291,0.9793262225671786,0.1709882985146043,0.8284452470270245,
 0.1322260353779896,0.5955748934931807,0.9109215119423041,0.811001421525938,
 0.19070788446716935,0.9047580918824277,0.1000695308954519,0.383346778551917,
 0.7332274795629302,0.4929638810999254,0.5807897515109506,0.4928639819984234,
 0.9084837216029616,0.16488269145338386,0.9053162972400858,0.5813047931070789,
 0.7789026421440547,0.7354812706643904,0.20333615551952688,
 0.33753540341015487,0.6330129692518289,0.3101455641817621,0.726924179086431,
 0.7574212545687834,0.13221304496479347,0.3898661408157448,0.9732457578104188,
 0.36819137095974375,0.6798060651576519,0.8792152806253177,
 0.40086211180837383,0.821975824921993,0.3968506628968326,0.7892527621701702,
 0.8281292874504416,0.8087792320896486,0.6395509806655453,0.18518653114792216,
 0.12337322023252006,0.00156664785474947,0.0024358584332270228,
 0.5582161758747163,0.6467617193913798,0.017260126088876837,
 0.9884251426437519,0.8080445677191632,0.4386290847565638,0.9363489008341294,
 0.31462076649766557,0.8868207035765123,0.3589137244966524,
 0.03886843526061279,0.8639637005786948,0.6084172623043357,0.869990791297152,
 0.04040050901498449,0.11204792442401235,0.3374781741272096,
 0.32164513465151834,0.05919679560083124,0.17228283099985364,
 0.9718650542901186,0.9481572082939115,0.7241520863452114,0.30717765438897227,
 0.9453936898455282,0.4227804862990525,0.7713974738412588,0.1265965355465396,
 0.01598954056779689,0.7800643672421307,0.7796038102359123,0.5264782247574933,
 0.6287712663128597,0.8381130065054375,0.5142975452648924,
 0.022997506049695637,0.4679464340726984,0.5275488069120576,
 0.6135647137178508,0.19304115088188067,0.4910485980083764,0.6998092881386577,
 0.6355827554093628,0.9384126019306033,0.5659665777982318,0.6940304757268517,
 0.6670914061540462,0.6831669090980713,0.69029911878588,0.3262215179413658,
 0.3574227673996738,0.1207308192582992,0.9028817304850894,0.18633039644418536,
 0.13837150893473615,0.3404145587295392,0.7062968649503948,0.702895771440733,
 0.7225501229552503,0.47129409237429565,0.020951132358101532,
 0.793420986873259,0.07047067144232211,0.9560308038924997,0.44980878093153254,
 0.7585585907555841,0.5334207970223943,0.6066697886586595,0.46157278707161353,
 0.9895092673406248,0.5812813792072579,0.7331712407742017,0.6557988096284636,
 0.6775130045952971,0.44919370527038305,0.33954650951115983,
 0.4897789027393876,0.737532411101188,0.1323177324542596,0.20806663744517506,
 0.9132742050091756,0.6119815428231422,0.8236060360856868,0.19355970566941139,
 0.26211054326294925,0.8258295616028244,0.5666420666895452,
 0.45530801590673664,0.9939783806260225,0.07010869500752315,
 0.8475785862027672,0.09223983963996552,0.44280494694582884,
 0.7889813433482735,0.40699202622504915,0.5381577267618799,0.6170132303461467,
 0.7869768363171233,0.7094128601584124,0.5732645766638944,0.9760236999877497,
 0.4768324824501601,0.31051467579158176,0.563313059552237,0.9983642022487679,
 0.010775951295413644,0.34938825915684835,0.8350752555914981,
 0.2611878395802396,0.8783407066331669,0.3861828218556729,0.36615904925608656,
 0.016493332557275586,0.1677386917861574,0.20934123257677928,
 0.6752404583134325,0.9403628458296547,0.056605994546408356,
 0.5055925950282248,0.7631906555412433,0.13888792144462747,0.9690984407499901,
 0.9356449819717538,0.9681981840570817,0.4541406970497893,0.04578670508629745,
 0.6633630371147226,0.16472975764368103,0.7539291678712231,
 0.09066924239855356,0.6438303022475818,0.805042011274603,0.533362019213168,
 0.8345999974102516,0.28385818820641295,0.8682886510199209,0.9377159642096008,
 0.047956158908276425,0.16389295640880208,0.8912145984164276,
 0.6519554689943587,0.33608492350399355,0.4972985122565208,
 0.21392367208787944,0.5171814928255634,0.9089242579785022,0.5505830623997917,
 0.5435071759813825,0.10020941177584075,0.5125288110053801,0.7118552528004187,
 0.40592657646151875,0.9436169430424965,0.3516421785213313,0.8593677767287505,
 0.9498699073527789,0.04432005378127024,0.3004734411484792,0.8547542154617207,
 0.9278247345669418,0.7016401569523365,0.9797818592293175,0.7396061957310218,
 0.06083011219228929,0.956721325991677,0.49404907386019326,0.6147225076129222,
 0.7014893753091709,0.17729603441271058,0.20657124854077302,
 0.2985262102435978,0.2613678602526369,0.7096685021549916,0.41741467267926935,
 0.00968243818776282,0.571101957056166,0.1923902750812706,0.03308878941032434,
 0.49712786611796833,0.7433001593486159,0.15202324472869544,
 0.2401194130927088,0.10395049442831028,0.9176471064155834,
 0.08846176572061082,0.2062384175127001,0.6608986428599758,0.6682611548422907,
 0.15774169461858134,0.9459539884328584,0.9610519631577378,0.750172708544527,
 0.9103354342452143,0.46383784455031674,0.30323104586088645,
 0.05105757136153166,0.8859687473168157,0.46418933474760626,
 0.19559651620425644,0.716735949455133,0.13945862003807974,
 0.06794799608436108,0.24810115980714254,0.7782850653412317,
 0.7395500851504666,0.5521443675646028,0.17018617534703395,0.8975204891474498,
 0.3247179784230064,0.28671188682195503,0.7097450287734793,0.9010556184743355,
 0.7254444572565687,0.28247865190105037,0.5716222632765753,0.5142557541979563,
 0.540103702817905,0.2662518061153246,0.09406198599311433,0.3290974880995796,
 0.43714057894924463,0.347572090659095,0.31528552168911583,0.8265215580737285,
 0.29765491605928096,0.6998567513825594,0.12482867963891287,
 0.6265328603895981,0.31344673563575065,0.9013098750158484,
 0.09950336670576898,0.6853056382558843,0.6329899638662405,0.5737086677032638,
 0.48810251422647133,0.7269402802274489,0.6217054976392329,0.9550212747585327,
 0.3600862810754807,0.2310852685307898,0.9035719555887999,0.9370193599061939,
 0.3746955180045699,0.7824395983058928,0.9931524202167632,0.8800378047554271,
 0.17642204122538274,0.48973781527729787,0.34307153139408175,
 0.5321291381149265,0.4484274867335094,0.45303200058227977,0.9405202233105596,
 0.580668878104029,0.6154791899905189,0.8478109589581777,0.6746527654331724,
 0.32902548667812703,0.0884831570981387,0.5695618575611536,
 0.48991537623676906,0.5567758339271631,0.34110498617731033,
 0.9519371686747359,0.2341203556245166,0.858229291137918,0.3274189684710831,
 0.6604226145546659,0.4213353410787596,0.70223843194246,0.9563729654793623,
 0.07382734421104986,0.6161372551482118,0.42077645290538235,
 0.48810086881845005,0.30099182594499263,0.877660499294802,0.8873475672880674,
 0.5845394959077953,0.6445906034213577,0.8092344161146533,0.41562769534128463,
 0.4473040710807836,0.5548559857082291,0.3595615862972036,0.6287402594711856,
 0.3069030894050506,0.07987902620470055,0.3765821387142283,0.2786699442423506,
 0.8045501982388866,0.3453492331800766,0.5673613658693931,0.6093048318013269,
 0.5702864154998313,0.2476283247136516,0.2866539648219929,0.9541027940294955,
 0.5016123917940237,0.8229827438327553,0.7715173393704994,0.4887899639001281,
 0.5463178868985193,0.33071307396877847,0.9494766332850649,
 0.38251763639466174,0.5602630508794796,0.1592642380041559,0.7999646681863076,
 0.41355011814688025,0.7464972177701203,0.6467105301246218,
 0.13631364951768832,0.24202050703709688,0.05261165004643509,
 0.6901758977991002,0.9254958394141358,0.30610866442690043,0.203091623636371,
 0.5710648161045653,0.6676207969742013,0.47787415406660827,0.1796676392312957,
 0.42843082416391964,0.35178274434908585,0.5739655701099026,
 0.3628773990970894,0.9629674210105043,0.28892337825726366,0.5821025655071216,
 0.07875364765054715,0.3721678081809452,0.21239378505673923,
 0.6313236690056141,0.8516819610008658,0.5435328217087423,0.13498934219682523,
 0.4254590643132561,0.961796349585461,0.7868814753242664,0.8621494805869039,
 0.6996153780955289,0.26253068172422855,0.7637627720738176,0.3404962250301773,
 0.5268127066236112,0.7870300840758058,0.8745479828257863,0.842117831894992,
 0.41594957294652357,0.04988927136361321,0.7019997293870248,
 0.2237177916538161,0.4974657532000669,0.8593393195345238,0.7556233713917476,
 0.32682445517186853,0.12547217356594476,0.3950630645802634,0.244850552672908,
 0.7738488030605226,0.9619897891138289,0.8191551431576701,0.0220811733991354,
 0.0641847049365516,0.31679742398792976,0.20912400645880547,
 0.4750265195591393,0.27800296932837976,0.5249715775568046,0.6087123567925795,
 0.5433809875237922,0.8743276187721258,0.6387337286983453,0.8772158507329444,
 0.04638663919571384,0.33956174915536463,0.9701932076750212,
 0.9010176444601797,0.48891517504130466,0.6998423713251852,
 0.11141187133359054,0.7647103527013122,0.5837440116914496,0.7008773021337571,
 0.9128767272942209,0.43070088674727247,0.8409815845047499,0.5064319739992655,
 0.8304304157817343,0.21461304274451387,0.5789579069945457,
 0.46505475978104527,0.10594046403584212,0.8159014358809854,
 0.5373529046732312,0.9881787955586735,0.02550837985755805,0.9626116122397161,
 0.9976435399267078,0.04280579028301723,0.20614455824261513,0.572418591760176,
 0.9526333807481571,0.14545728772341748,0.07470789429994484,
 0.7477460383971998,0.5008243844416775,0.9591763244030758,0.10492207950059262,
 0.8514127188611792,0.8389725484866531,0.05624728443351945,
 0.051035106569846556,0.8225241898141855,0.4640698932401478,
 0.8401604139736236,0.9381818377252283,0.9921457046238698,0.838117985751372,
 0.9125541738877357,0.38117409668315516,0.5868698187165249,0.837536025736607,
 0.380571508507925,0.33665115181935834,0.9067524259168505,0.29287332243074027,
 0.037203726882759636,0.14688816828608897,0.7769139297036606,
 0.3230859676091332,0.6376762491546791,0.4946353613190806,0.9436595008163168,
 0.273445150810513,0.5730670194786361,0.472742622742755,0.6443833259853378,
 0.44156241448454336,0.9861914048261,0.26674464906262607,0.8905100074989152,
 0.524063909297289,0.01193532462285507,0.4762474101019414,0.2351323159957439,
 0.6489759037565258,0.8053505900912354,0.16201980354007328,0.9276046507446912,
 0.43465564327151995,0.4158733002317805,0.7747648179744427,
 0.21301695746821048,0.05497038182087832,0.1272256464860625,
 0.5803602571091088,0.1643160088985206,0.7542994599016564,0.4861266590918375,
 0.9172406008151885,0.32199335990641365,0.6274548126692,0.11434426858522351,
 0.06696296439381177,0.9782931782024947,0.7826888806372426,0.9793022134015635,
 0.4661461650236063,0.4505808451522192,0.037670808253529975,
 0.08375263022847512,0.4457556226407038,0.6460437601264885,0.6091298181302286,
 0.2954745693804618,0.9587073395784036,0.5613018016138664,0.32035673849237134,
 0.06034756099343341,0.16580521236893453,0.03692020920063277,
 0.5676853199640266,0.16719616984274543,0.20842997262102114,
 0.8342603684688736,0.2665228236349534,0.6306210759454856,0.5301772025875551,
 0.39059826426134425,0.9027839370037376,0.5151139233372849,
 0.09998408568761663,0.4518444893761727,0.03881655274404738,
 0.8750007871164818,0.5230280663549021,0.012523243173639176,
 0.48584914917145117,0.03038738774867289,0.3757810832059235,
 0.5173493941283004,0.5762688547639836,0.4262236458030524,0.7409268230043919,
 0.5367683768162267,0.12501666448135362,0.23092925741261416,
 0.011696147851591454,0.08588980859958384,0.714362928745281,
 0.9047493102244857,0.9923818470226131,0.684313449269094,0.6261878320245012,
 0.4111065492939623,0.3017579280552176,0.37168302123569563,0.8876311689418703,
 0.714630218621824,0.16127568519192226,0.8268360495914472,0.2896279121773634,
 0.8500681434107398,0.31822110377151236,0.1478038200253038,0.5000191703426609,
 0.51490783691351,0.12585378694718574,0.8727977338977416,0.696503747714951,
 0.16502738565343456,0.06759962427852328,0.26518310380627685,
 0.41493550781823774,0.6180489994401706,0.44255355547362785,
 0.04517665754576905,0.6204385767883254,0.5001034660468308,0.4113942754264214,
 0.6157477312469399,0.39840082749709227,0.5784178721053918,
 0.09498447818111244,0.6687972388895539,0.41889826372959915,
 0.5332831873313044,0.8147723885112388,0.46328351536064005,0.3132339147182809,
 0.9119489553170228,0.4359966272680471,0.7465464869724902,0.2188123084742407,
 0.9599857410392596,0.08130193652908191,0.06165106599733161,
 0.053146195179970546,0.07007944477551764,0.30233154506786797,
 0.5902380453663039,0.4240714283830461,0.5043050900966848,0.04617196044544403,
 0.5074333934419412,0.6583757072843991,0.6881861010749835,0.5107414613194045,
 0.9537364367675851,0.5092695421114534,0.8197373634468205,0.8915138978549739,
 0.3410853737630909,0.6426892109347527,0.48458754031148954,0.5065109937398471,
 0.719843964739501,0.7717732452863335,0.27079732817542024,0.6135565547740531,
 0.9975695404628677,0.5410478799545586,0.512722499133217,0.07798168845781439,
 0.9432360218191834,0.0329210290113483,0.20482072538292317,0.6671298987539879,
 0.7754053357798041,0.4447079908781909,0.3621413235233112,0.21473878919210132,
 0.8423593632985333,0.45449110536532156,0.1497304227312568,0.7003941407747709,
 0.9692313913938947,0.8009405760567604,0.9269999280927604,0.5075373106940847,
 0.194991735757565,0.8571160591281348,0.15974798498567688,0.2756955024495126,
 0.2220061215661735,0.7028655234206,0.7423507483942324,0.7532223322688373,
 0.6698782919294571,0.13637918130213866,0.9435533483633775,0.5883026883015711,
 0.30870777334435506,0.06634745460155078,0.9162386321668144,
 0.5444221383782453,0.017152129617907907,0.3965184270269515,
 0.010066495910951212,0.8820672661071081,0.4623609454250581,
 0.30207551874308336,0.671878634909871,0.9129149545307045,0.27388933343060407,
 0.17063087494336726,0.1748336784559883,0.5540492330173501,0.9346285346451452,
 0.8503774847003879,0.6912361167613784,0.003076111269559756,
 0.40715413798096567,0.11271952084241754,0.6130234171647357,0.668140539279813,
 0.7365566777293192,0.5158559365941187,0.28221187128732506,0.3198638039007098,
 0.5208070508504121,0.7547622628643743,0.9520154608733594,0.7406489745292038,
 0.8711910226414523,0.40373906261272197,0.5215848631162364,
 0.06884040535142777,0.4829325971478401,0.793337114479129,0.5924282384844171,
 0.048501469282641874,0.10727227617687185,0.3873416387903599,
 0.5572851693161589,0.33419287555113875,0.7553927160459414,0.6131364803443684,
 0.39875518854007486,0.6452216784106766,0.965429699181619,0.6405782965891776,
 0.927411443928183,0.13959701342001152,0.28983795173267834,0.6722604881113858,
 0.7374120680600859,0.33737613660773336,0.9850671734767316,0.5738340183887142,
 0.8248715864513043,0.2053757336484241,0.41716348402915315,0.906011924505129,
 0.735167925391766,0.7196023207904415,0.1935899440342942,0.1450748535706602,
 0.4396939132239094,0.6356173396203321,0.34058218889746716,0.4461264102216489,
 0.8201463084960289,0.15910149145850572,0.08174940078001292,
 0.6580941590222492,0.3824916302226501,0.9293015618318603,0.6247330340623504,
 0.4536068320506803,0.4418448582824537,0.8426078584941623,0.10809055243450127,
 0.07030235121253514,0.36396755262221236,0.8268325480991724,
 0.8228856307519499,0.9552215641219911,0.18766325685228313,0.2851089233467027,
 0.9601662600662865,0.07736685121676423,0.3072245327283123,0.1800645284218989,
 0.9121215200622218,0.10602283771372023,0.4389797285671183,0.4562951050314824,
 0.15555005464491667,0.8904733707129315,0.9829540706304125,0.4659100989152336,
 0.7749838142986794,0.9711325591589862,0.04540706216774404,0.6913623322293458,
 0.30656006485224707,0.15076508612299766,0.12683872637428528,
 0.05101864920621468,0.24792091364420532,0.8021042126567478,
 0.3937804803821986,0.7906567916238642,0.8219960457181185,0.7449469814579704,
 0.7716353369951079,0.6044150855643586,0.05956245216920453,0.5872176620717404,
 0.08182133498785271,0.23623719469239857,0.25811906201577084,
 0.2289024671163934,0.7554092659697795,0.2387867378660975,0.7429054075298873,
 0.949036044568427,0.2573864925665069,0.7200808888253315,0.776616930669536,
 0.14242533858203388,0.14877370422563485,0.05717684597717232,
 0.26700901049771275,0.20817889697931813,0.21016591620539504,
 0.449382204506672,0.23057677541107935,0.6010733402937355,0.554344588337896,
 0.09056184206692575,0.48500940574027485,0.021885474281046036,
 0.37881284082838773,0.38043840534554907,0.4867240449463206,0.857836825949986,
 0.4963444560850274,0.3943677613569505,0.1928551574315449,0.9916069114842648,
 0.9843721968542033,0.8320553397499346,0.30431616172352904,
 0.028112238376648335,0.7365785834908102,0.7674035835608238,
 0.38246644273427255,0.5454814716104102,0.10237137326932466,
 0.7648753732988541,0.7212466389790055,0.13667602217773234,
 0.37135259104364027,0.17057135332217077,0.1802544744622645,
 0.4939593512292225,0.9856095853411625,0.2907241156045036,0.21875419658970596,
 0.7294000290837883,0.6761956439849592,0.10570532035281155,0.6232873061428855,
 0.19871645453736875,0.5210994048341296,0.05779641426785487,
 0.41633742981846167,0.3396870227388734,0.3702337158871105,0.9411323851132234,
 0.07316593491842982,0.43548639670166756,0.05827313354223229,
 0.4698924276852767,0.9269022903967727,0.386510993283958,0.017797646837432457,
 0.7733770382018211,0.9008618582123817,0.24519043640476879,0.7779939219602433,
 0.6671662673953197,0.21577851305570472,0.894985127705155,0.02745169283307891,
 0.7218464125099012,0.5424815049420346,0.8672891588579821,0.989617091993502,
 0.7478693902971422,0.9046583140333662,0.48953789734918274,0.6615507499401092,
 0.8298028507351031,0.028850516180823793,0.23005652955740818,
 0.4426161742431902,0.6959047838278628,0.009759677539662337,
 0.6824071504570424,0.8393029851146021,0.8101027054342055,0.8483320769282876,
 0.788825446051146,0.6933586519285957,0.7672824277524881,0.6997661863234903,
 0.41041458729407365,0.0994427612060852,0.10358137111157995,
 0.3584468956896556,0.6197883728479896,0.9252596586107004,0.38540643804705677,
 0.9791844927779845,0.7703817044111175,0.8071803245512904,0.7036146722837432,
 0.729810527331069,0.8253356912905812,0.31810021972057423,0.44735032271561515,
 0.0383624169725959,0.09358133205117036,0.8572544091941189,0.2458662250753647,
 0.9096013454136005,0.3999113140985391,0.25686440681939826,0.5913566640603571,
 0.4499342233380852,0.9949067876065882,0.6034149024337168,0.8871415622846409,
 0.6234809242752372,0.2795458240768287,0.6892520432370891,0.32990303304369717,
 0.5707485540907821,0.8579926305176452,0.061168453988854066,
 0.15347968001420031,0.17515634177722628,0.8533911615003826,
 0.17525004106925968,0.18322498365662565,0.054138813232100125,
 0.8693946010881614,0.6100036661304222,0.3440609843220226,0.3831893898957146,
 0.41080567929851375,0.23150699977033995,0.42660866490980565,
 0.7465594962306097,0.36213743093325346,0.6951908188182814,0.588193672328207,
 0.3884230418478706,0.925110084802766,0.9295614931797332,0.8858010688376445,
 0.8956937962076779,0.12939919888651907,0.32545214634506014,
 0.35819953668714133,0.8034175503994465,0.572657396841208,0.3802394066361866,
 0.08375118773286283,0.5336183987752975,0.684620264538701,0.7637714248230858,
 0.24184602713689807,0.5917681962918386,0.6927293879744956,0.5532644780825933,
 0.8339998144748613,0.553986668786514,0.7496040609667738,0.6409469782384914,
 0.4861485408492846,0.2402111467416045,0.26548730246845853,0.3745613144294152,
 0.23791760591207045,0.8862768093035744,0.5854435988902988,0.6276236328357631,
 0.42862439781161765,0.07066554650460544,0.6140330772343601,
 0.6740257815010793,0.4184431273498169,0.12741656235761611,0.0927599522506315,
 0.8169055738364986,0.8477415071747517,0.42543618891185364,0.4089087389751299,
 0.04334909031874701,0.2635760191464578,0.5634757849246832,0.5209987773006487,
 0.44901965858899573,0.5785689606635175,0.3334935067386967,0.734678096405078,
 0.07533768757901238,0.43492056823860104,0.4150426208035376,
 0.7784709863540255,0.8609565496569394,0.29541087294150625,0.5834563477790505,
 0.5072465010172275,0.7804360866528778,0.5620804153131679,0.6786467093975388,
 0.8485230445884984,0.003018805624550258,0.019635811102975964,
 0.626315156365892,0.3625869870710715,0.2836990599667817,0.2979121603093702,
 0.6528361823232766,0.6894391158156724,0.13796322778035597,0.745346765756681,
 0.9816708125401381,0.07726423746779354,0.7364680901670864,
 0.10495311279373731,0.5061233457820905,0.5412007310508368,0.7171478858232048,
 0.07156687353201363,0.4569806623196804,0.84415150770187,0.1821674212341392,
 0.5090739287644727,0.09109365099227074,0.9295346449805866,0.8987974278702953,
 0.39362417899507185,0.6898849821734702,0.7737358423854339,0.9402940999623876,
 0.9535650097756874,0.8076176307116347,0.5031130350591251,0.0441202090159315,
 0.3442735065713397,0.1990051253702001,0.5531352140496076,0.97400201367173,
 0.5632636914699554,0.4530782071524647,0.005161927336882033,
 0.02057124321653514,0.8603082620872743,0.06355047232955813,
 0.5283936451698643,0.1186888350583385,0.6146670490986111,
 0.050976224643668466,0.31139122818216136,0.24570502642264813,
 0.38944285936570944,0.14580118805277253,0.0030179990142221724,
 0.06920848854939032,0.21172613054106526,0.0034281692917152995,
 0.7365285241979123,0.4850835159714336,0.7901627106111229,0.6070853229180729,
 0.759051743807254,0.5258279505980757,0.2112065788964893,0.8607039456737929,
 0.80607074275914,0.30257411092234343,0.8612959753983488,0.5955673535400351,
 0.3814212730109596,0.2589091485730999,0.36854940321429874,0.8739119841479289,
 0.06822068939619697,0.7122690713512534,0.46752710849231116,
 0.10478007655098898,0.3104369246884162,0.3469500826398024,
 0.06270388796725235,0.44393136597338534,0.795363988675938,
 0.057731323887425745,0.011599722651138622,0.1310797373249425,
 0.25266676131095767,0.050432524082057784,0.05063339917294418,
 0.8133391820919174,0.9408550253825159,0.3840156403317493,0.3108544096190877,
 0.23045192341964316,0.25937352012533,0.6921801591554557,0.7605043746811754,
 0.7036680264705505,0.3097456497018869,0.06396376036272322,
 0.16782839997446053,0.9106416512486878,0.9261544615213809,0.7591825040445979,
 0.8265802878366306,0.6721014904583961,0.29176701718230913,
 0.03375873990244804,0.2678072620792783,0.743918805423339,0.9796852706333216,
 0.314799263869292,0.8067320018982972,0.586490918393196,0.58069798538638,
 0.09150478782356686,0.5089995742870478,0.08219466030718714,
 0.6596618330184865,0.5334388299381709,0.5042636790711318,0.5712281593409427,
 0.4638546345113075,0.9526567521111597,0.957093703086,0.9164002628479799,
 0.5021230563452772,0.12467221982597798,0.4767235754062509,0.7636978649301602,
 0.10911878245716822,0.6176118229415103,0.9320180699700685,0.6438240068592376,
 0.6208773575987199,0.07403083524995646,0.5549969387972896,
 0.32044391275737816,0.5448982847395081,0.283706722831944,0.8384236579210628,
 0.7568724790860366,0.9823146449338674,0.9440508309678584,0.9993014583384384,
 0.9209827837485149,0.21540362128248802,0.6612049093267106,0.5203909164283916,
 0.9864839592469834,0.03234514237832187,0.8414369456063227,0.2776031056875753,
 0.756341159338746,0.744031707803306,0.5007170931402978,0.8945263344664013,
 0.5166260233263203,0.6900546255116672,0.020356920796018696,
 0.17009419738546938,0.1662554567197403,0.8129870836709261,0.8779399221378492,
 0.5859254318101262,0.2578627165888572,0.4866742331806363,0.35973528716160397,
 0.4433678478705728,0.1302285436912658,0.640699411160413,0.5928774471723212,
 0.913323116757688,0.46497856887709876,0.842592019808458,0.4760924810189442,
 0.8988135979804899,0.1971375121200567,0.2980310795435759,0.782171268102632,
 0.8301478228025002,0.8520776395687879,0.37506888682227024,0.866791367111827,
 0.5509115888061009,0.966186121768636,0.12303018497350893,0.9368707401547536,
 0.2751246315801531,0.07979350753797632,0.018431216625915847,
 0.9683438647829978,0.52047201433651,0.8740697987079905,0.51857299875905,
 0.571808189306841,0.3691860820085008,0.6435843581249097,0.09362871714230192,
 0.44263357108101176,0.5012628175058139,0.8623831269478872,0.5235174315092642,
 0.5226423813218204,0.4361369024247942,0.9201496911689988,0.5402124249773127,
 0.29455032694047467,0.82307039433185,0.6379873075895724,0.1328861346949235,
 0.4736960375149539,0.8695403762892291,0.23679128823333562,0.678884694333408,
 0.8888955569783645,0.6807629127031458,0.7040415156809455,0.33346164201074946,
 0.9701492712046751,0.09454454061765583,0.1466765099451809,0.8254558468866158,
 0.6906293552353132,0.9151074542564399,0.0039288200524211225,
 0.29753641628210126,0.3319722825821724,0.4602414788168764,0.9051315990177863,
 0.866522060364126,0.37154607852065236,0.7499816580069041,0.6328879923450185,
 0.7935395564938151,0.03845713817666241,0.04381062089957832,0.693445849421722,
 0.30491520848335796,0.25322963524506514,0.09569241278205143,
 0.1247004407302954,0.8157427245258099,0.41388373933945644,0.9730107888342305,
 0.7228346223184328,0.19487416743060648,0.9103021757266736,0.5538839995338096,
 0.9134506964831521,0.7865068604270486,0.6436879478835024,0.1962936759798507,
 0.9961308855094595,0.09003238714547857,0.18979382889936303,
 0.25804277060684777,0.06845821891866932,0.24084314431116183,
 0.08715196710296536,0.8853653837123168,0.8359794898611655,0.4846122463291451,
 0.9947391501716818,0.477176028829156,0.8461773544592347,0.8944107953220224,
 0.8448856517086094,0.18749423238227947,0.6328508465140869,
 0.036688805084957754,0.2310670614257645,0.5652359646730689,
 0.5166746241938132,0.3437467244181368,0.3815334895690091,0.8952061149906162,
 0.9538626613523772,0.7101180496796844,0.2593908093428926,
 0.027938952775729775,0.17774824226899866,0.11181615032728487,
 0.38736526018332207,0.09922659081950347,0.3192272325881059,
 0.22820733792438563,0.12717747148077774,0.07476909152912281,
 0.1099571969352997,0.4595220733293912,0.9361529663812868,0.8508990842885343,
 0.23395664248443304,0.2844011938601141,0.149458232451942,0.43207960313432825,
 0.6330998437563111,0.9076281616791086,0.9558537479911506,0.03607953609630832,
 0.8671172551567878,0.47608635427152346,0.6847829553594759,0.6685027116917295,
 0.23399622319964775,0.9782710227266855,0.9345848066304591,0.4085262063649986,
 0.07433503343085879,0.2752426873018289,0.587334492085473,0.1195809945262779,
 0.7954905161926468,0.6213249233856243,0.3219395435778083,0.16464048713525714,
 0.006193914805760925,0.1448657280091098,0.9016468508172497,
 0.21771935410688126,0.2617959767229977,0.46882417766569606,
 0.6471381820508663,0.7270016967633433,0.3695230005539971,0.8641007911284808,
 0.34689576096107744,0.2542883201497299,0.236041175990512,0.9739195103186459,
 0.5615453367887191,0.9012619817661477,0.4910400538749611,0.9953855110944407,
 0.8447399396821754,0.3933544570923547,0.4287489673083015,0.19710971960647905,
 0.6540719522694681,0.7772013385196155,0.7107373976529531,0.2629015288998091,
 0.6249901833816679,0.39588293743046155,0.3632390706318118,0.8328502413558535,
 0.9964172834021507,0.22815632868707736,0.7600441644988691,0.8055386879827384,
 0.03512069672445506,0.33267851775926216,0.909409360370853,0.5475515619771962,
 0.7009622714354332,0.8568783014150471,0.03374857822030852,0.876788018915875,
 0.561084240647855,0.8335368457478796,0.5760830963865234,0.5500936424473778,
 0.3054449792983722,0.567582833961028,0.7072105288372232,0.1580129785487473,
 0.7725405146214879,0.1427635012211309,0.13870782943403936,0.9343017222957881,
 0.3292346102033189,0.6511967696117523,0.7877595635325192,0.31242367433061025,
 0.981119250703524,0.8107912012795967,0.34007049963228686,0.8473118723549623,
 0.5309179689713545,0.0395860742482359,0.8746893774303738,0.10392874132990837,
 0.5188201303901372,0.00896906404443154,0.40581007930998936,
 0.2864063915985755,0.883820956107451,0.26147194204536706,0.4703092643061164,
 0.8473578428964659,0.655891633319347,0.765940818540133,0.15534421563500822,
 0.8916263452300814,0.9003877613944611,0.09088819092456624,
 0.13011564466253955,0.392356687893628,0.8165289058361546,
 0.018180326445689232,0.8704338670343186,0.8538298600749447,
 0.35436270927930946,0.4645913346034014,0.20082882473372443,
 0.6430934118589271,0.5978617688917465,0.4983247768160207,0.13959878590531294,
 0.7592723442864889,0.9725988648545002,0.04633421633725321,0.7890180605859375,
 0.07142223968126571,0.2002798058366434,0.05919499497995728,
 0.3999959555349416,0.10293745652402464,0.3170110475111042,0.8909162246398449,
 0.29670264755424913,0.4680713856920975,0.4107881798193008,0.5913824303919479,
 0.523501748889428,0.2256752735568951,0.46565849415796967,0.05118357184524547,
 0.5466800129727085,0.5201849506713043,0.050944186753233756,
 0.6938494312350783,0.6669916670014319,0.31241888718753064,0.5000364969667106,
 0.8862357902182184,0.7948287285233467,0.17614967520364455,
 0.43079428008354725,0.6309013933608785,0.7410020483527078,0.8907774843578399,
 0.6697312553038742,0.11985507607326173,0.0868908125156651,
 0.21202629556208796,0.9409804416480592,0.5218417490572095,0.4006117129397839,
 0.31945784771636887,0.9001061877732075,0.4142586827679525,0.581184678557045,
 0.12127828341956426,0.9837557654213918,0.6441528833616477,0.8383620837940674,
 0.03286929353384016,0.721150831297568,0.029974352176775465,
 0.2938954362714231,0.9061153391510235,0.12554426642541283,0.4678119563860037,
 0.3662993984351589,0.6885848137372304,0.779530712428638,0.10994600562398238,
 0.48667263583334197,0.4987934594871981,0.4078571740449701,
 0.32156177614425285,0.36600304130149286,0.29327327258768376,
 0.40337520786850367,0.7136796079305019,0.10018001914344732,
 0.05409574388480731,0.9484767504058945,0.29154842289490024,
 0.3090690301998975,0.4379263431216869,0.6550349577567949,0.6544257139994961,
 0.3150869147260644,0.005881267544529334,0.6228078481580411,
 0.5545936503001465,0.19818469608106515,0.6100525130124722,0.6907393838724286,
 0.3322593720509843,0.9952823797155355,0.37241743432348695,
 0.20022660656362745,0.17028110689577147,0.5814090265579084,
 0.5402099865415418,0.8628414275450469,0.7332248157933101,0.7788075424938842,
 0.8711192617210848,0.4293532057332001,0.1702851955011232,0.9437776781520074,
 0.39014375367936893,0.37303125903669876,0.11641238649244989,
 0.27929798833322717,0.4075983882618339,0.8691617566080982,0.1955035559191497,
 0.9990393210236511,0.38922506150623004,0.6393361868957106,0.8116118834931759,
 0.5636057737784094,0.6440302014709679,0.9092843862899829,0.6979880703256635,
 0.3206962823111742,0.27061926388032864,0.3256105843551568,0.7417552133873766,
 0.01718070573282182,0.7334331683153841,0.7599464741237814,
 0.14979513585016624,0.6207889840489467,0.16461794222517012,
 0.8948781040309155,0.9156954314595698,0.5841091532437171,
 0.021804257599987498,0.7836702712983633,0.709884661811983,0.4839800145993267,
 0.5469916083612044,0.3897528249849146,0.22253059336707015,0.7052316638947274,
 0.2159236318301352,0.46828994032339577,0.10269902996826574,
 0.2111793289563011,0.7309688324472237,0.6042203888632725,0.39912636561271175,
 0.4638888670528095,0.23375174990539938,0.4435232701402194,0.2572498073994567,
 0.2991548492785817,0.8934648988744958,0.8189960866253616,0.10012967922430704,
 0.3543290548281287,0.9632412415085482,0.5767028724370705,0.9615289119106313,
 0.9769548639955433,0.8660871451784817,0.8381223457031787,0.3046647712406644,
 0.1855992359907387,0.8673464160705221,0.815538952842371,0.1572101707004152,
 0.9044063901010684,0.8177911016157088,0.9596194329573874,0.6474299490432682,
 0.031213550205178242,0.5654197266724368,0.39494598949819826,
 0.7048943494575589,0.626843452977555,0.6482786884992199,0.6600396989200998,
 0.013427539459553217,0.6384241200967887,0.6121532749038359,
 0.4505216538560449,0.5927902826331561,0.5015129223093188,0.4929833703682682,
 0.8432128381409628,0.7503398718231301,0.620510409053955,0.015838793169917986,
 0.479832412400106,0.13635893637722762,0.3750976337137566,0.8015876745984774,
 0.4687899193405033,0.013099064783450798,0.5785827436165369,0.510195079465879,
 0.043576126351841604,0.9243878822472079,0.4915867853573317,
 0.7073431569017257,0.9685481029051017,0.7396471310744603,0.20712639074707195,
 0.10813209980685101,0.35124507055831444,0.42005972635486133,
 0.896246729004305,0.7017020658184863,0.7710461292794543,0.17097560181663163,
 0.6925447940456189,0.6187380787103415,0.9403167461255497,0.9888618041483135,
 0.9266938511746108,0.3164358429217762,0.1280420074145665,0.5391248854437137,
 0.6544173544658978,0.32606465562519704,0.8724467908935827,0.7291114816138021,
 0.7523148081807949,0.05170520950056734,0.9076808722365852,0.5615924280305226,
 0.1559989152592428,0.03587667476880685,0.8125816377678138,0.3597293443554992,
 0.5126932099503275,0.39946226056340006,0.04768522832600697,
 0.008241387141847967,0.5632636395128502,0.5763586327923748,
 0.03261764555329982,0.6935334251361198,0.4752455482550494,0.837888379725628,
 0.10295913504845233,0.8306580820429039,0.7937985845138534,0.5749490784232278,
 0.14011508016706997,0.7424157085745139,0.14931124744828017,
 0.34840828591268436,0.3132648680092106,0.4939794067692791,0.6491843079336768,
 0.9253677319845692,0.8823067976519653,0.18466821108295994,0.5195562006257787,
 0.772526509293813,0.9673078786232648,0.12735003783100685,0.4909021794430021,
 0.7820354752371012,0.13368605187387606,0.1905048422440826,0.3333962067603381,
 0.766393630029726,0.15387724817338344,0.3591121453187538,
 0.018613700946638168,0.8971729511215946,0.6578470943286362,
 0.040510697833085585,0.843894658200196,0.9142456437476569,
 0.015621163551543527,0.050353610922354686,0.9833824648405033,
 0.09540206747099489,0.23062426839503136,0.037150014189805236,
 0.7011316405226359,0.6088771416929349,0.8100080750595638,0.25518778661022634,
 0.4070421779369837,0.5576634024062104,0.11541540944485829,0.5588207404984902,
 0.05573209473223606,0.3671255780416377,0.02394437001273997,
 0.17303756247122948,0.830626397050098,0.7847276241440332,0.4777496978991459,
 0.8533473396419484,0.5737241962888414,0.4486599848294349,0.41340800246153,
 0.6691898955895165,0.7169115695036679,0.20271668099450035,0.5191799489485617,
 0.06558396284081569,0.5003686639490593,0.9525392185717512,0.7347612272327966,
 0.46923279091666814,0.8014886632796122,0.12210764365642301,0.443090459371703,
 0.8172065297643601,0.45403405784502104,0.4677841197387753,0.4655177586984781,
 0.7970183570724659,0.1968689128337091,0.3547383054399934,0.9115380856331509,
 0.8353946727429498,0.7868216082816075,0.5939125579425165,0.31571699116934326,
 0.5582894042231821,0.47878661287837887,0.0801368062779485,0.4927174441421097,
 0.6158162274971993,0.6679732573959789,0.20799400262911494,
 0.35845861551453473,0.8640434453323229,0.9539559364078429,
 0.31682110746060543,0.4819781374664114,0.34847684003733526,0.992344538461949,
 0.27090969246774665,0.030860771365934903,0.7667648493953214,
 0.28350436490447317,0.270121081934538,0.16463528405533256,0.7838753595995894,
 0.8000511951416163,0.45327475031978914,0.8857954177960011,0.4281273951134318,
 0.9351979206092668,0.6364983842389033,0.6588728177297628,0.7088792879354416,
 0.24883755632917803,0.5219779110552185,0.36452609287194815,
 0.4530486532622635,0.3751075413559162,0.12146934193156089,0.5177918586620335,
 0.004093623016020342,0.4227930976031852,0.6045274024167213,
 0.6774093242550793,0.6923400951902375,0.23236337758742975,0.704125268945662,
 0.9854872452346177,0.13189692612515147,0.1454408429706846,
 0.07767789597069386,0.24299829927883032,0.31011316451648085,
 0.479547668912055,0.8694432866297027,0.858190489647988,0.5904789552139109,
 0.12593073848061487,0.7183097319679387,0.26144585861439584,
 0.5634747038876248,0.4540142815784223,0.3162605034537477,0.8028902465562591,
 0.8612083216897739,0.8005010336561551,0.42445096049459674,
 0.009349214871719358,0.40241899276927284,0.421321859753354,
 7.636136727703358e-4,0.4203155623468736,0.6781949130197542,
 0.9550859764707085,0.8182050568252657,0.6415246466349194,
 0.042349087991195744,0.3545615184877243,0.7359986343231142,0.739560018259609,
 0.8264393266573032,0.5795177210743795,0.2787574309241665,0.9550160808609053,
 0.41208065804523575,0.3908373462522794,0.10766735650408488,
 0.023230840463644542,0.8915332428713152,0.9602652858814467,
 0.4773746989548283,0.7140449077244292,0.3831232151186783,0.9604542373677942,
 0.13645468144203954,0.40186744929802054,0.7762698334521212,
 0.2297162019421981,0.800026416844744,0.890358173742146,0.40873578225933205,
 0.32303932443739847,0.9641170770791971,0.8662019202763858,0.3358833403461705,
 0.032591346315030356,0.3069455059197155,0.16928739215439237,
 0.585324921156671,0.7194781367769433,0.901717568097284,0.03508017543391384,
 0.7104886889648135,0.3681266482059268,0.7410229192247899,0.9039558041007698,
 0.7723185758875807,0.5347020047119636,0.10023574941646152,0.8601943530090931,
 0.9103394085706615,0.29771304962571077,0.8418606063701731,
 0.38438112065908214,0.17835399496668658,0.9366214319805732,
 0.2950572677798735,0.2332459194538976,0.7590658074218184,0.5207866450644996,
 0.4407532664062792,0.8256501173306084,0.9991706054268492,
 0.008388822079180014,0.5555337059499734,0.7828451296425524,
 0.18186465745321534,0.9830329939571033,0.7938076412604145,
 0.03520292202262609,0.3738836011886817,0.3102680737058744,0.6367829648596033,
 0.8891545956169343,0.6017683167698299,0.2252659503933787,0.3236459395372612,
 0.3014374628500085,0.3065420586899492,0.026827686741531354,
 0.5425104525717447,0.8841926534700126,0.8518280020219083,0.9880401145122192,
 0.6159872621298155,0.8794111225913978,0.44723818832548545,
 0.0013941813843385553,0.6140922803321118,0.3031146483539996,
 0.5232763190664866,0.08343847948829475,0.039477039865340235,
 0.45923092174623037,0.6465762973172533,0.0734776460144797,0.9934681097374921,
 0.9860447624109893,0.3845600563513335,0.08513296928433578,0.6870099312395243,
 0.6248153450425088,0.493155561060051,0.6324790723906615,0.5674221152532515,
 0.3005060919770912,0.8134257717221307,0.14800155538943294,0.5105958763789373,
 0.5924463248906656,0.19789254259311073,0.8351414752863203,0.83949201647845,
 0.30858463413655446,0.3077907696022556,0.47887870301917645,
 0.8514015963726898,0.3556668902955046,0.3808088573593903,0.7646180672024663,
 0.5354386015351585,0.4078479265142443,0.01696863542358218,0.5043917351332468,
 0.05566587002203793,0.05065765051092219,0.7401624702555083,
 0.3235246404309169,0.9115101766091303,0.5939750648302325,0.14418862276568079,
 0.2828610776607978,0.4549646976018552,0.10084623311789653,0.7130916311603833,
 0.34530623696187823,0.7896517875641627,0.3289354214828879,0.31582398778832,
 0.4574724993789262,0.12902685002702263,0.4351180852633838,0.6269388647471873,
 0.1448180291188781,0.924700117723746,0.1704018050087951,0.31333268365000344,
 0.9228067987948666,0.6036291727615448,0.33333497696338343,
 0.14062621302811706,0.58057445887162,0.8248189940346861,0.37703427797942224,
 0.10994793200738684,0.8998281868580772,0.9045585228460835,
 0.17276621871041842,0.44145412389679806,0.29191928723869554,
 0.08620380247074899,0.23301383126914543,0.38716997433684763,
 0.05023076329383824,0.013189078854109249,0.7805526058262178,
 0.3155932896042766,0.92455721685825,0.20458663101147223,0.30665082350120343,
 0.8474706056075307,0.5853521613220627,0.115169407833147,0.8166683839824733,
 0.5578948564636397,0.333215154353245,0.9680707821655417,0.056434028042360196,
 0.1386604842755328,0.5816586904266673,0.6762930404808815,0.3740974613951016,
 0.3815120365301111,0.23987771486939535,0.06588356994141886,0.968920754703085,
 0.6843560083117334,0.5616069755563291,0.4538921742615739,0.6056015779470594,
 0.3034077291720376,0.7043395649601725,0.23195949264115012,0.7619432892798919,
 0.7043327969672694,0.7016719050449187,0.9512011101487812,0.03950544584138782,
 0.41743345077412286,0.18331543902671488,0.019308841640847674,
 0.9384386024505387,0.676355271097748,0.4541124027035375,0.5617181956466003,
 0.3853741819875286,0.0577874411779179,0.9671792530143315,0.5344011113961609,
 0.03855175502738484,0.6132550608247043,0.6741913447436984,0.5153012990648266,
 0.16754236542811984,0.606451085753448,0.839435148456453,0.9988761566334156,
 0.8289623412544159,0.9569637328659818,0.1295121831855618,0.8219763663525393,
 0.5885839516850524,0.36485747170292615,0.7534603131218774,0.9745690600664142,
 0.04790820829278197,0.9711794233880761,0.05666476827432387,
 0.2535094776259593,0.3661579001416456,0.16915781573070343,0.9523498397706777,
 0.17349965382427324,0.8719788753272211,0.7733635843113029,
 0.05976191907514883,0.6576451501179763,0.09990990679970146,
 0.6858119534537837,0.14215286551248518,0.6710454198935478,0.7119748696492684,
 0.9886065672230528,0.884643718520129,0.9523741519971837,0.9235533290075886,
 0.19460040123238231,0.7461772936231295,0.47675718621236984,
 0.8998496906161574,0.6566897274864938,0.43873177488197446,0.3557690957950904,
 0.07928234709918158,0.39235322643939774,0.15316641860435987,
 0.7539147713114303,0.00983074641579551,0.963004894958412,0.4682774277539825,
 0.4318150810126118,0.6458501368621797,0.7267430520194182,0.06977374248374923,
 0.28890295839504043,0.8662749977846529,0.899558897595386,0.20599266779048264,
 0.4833633320638735,0.30520991510052775,0.0048531255399312245,
 0.04008053378374066,0.8466822379490779,0.5542704256824389,
 0.18662820805954006,0.8823365456262068,0.3656423322291523,
 0.22436341254810754,0.6094812713355647,0.8540693461750073,
 0.28958852786166467,0.12581022433090727,0.9738277679042467,
 0.2002419589855745,0.8317378666444575,0.8067916054547647,0.3271535118688157,
 0.7178940367730544,0.2935027122970516,0.8063035179173368,0.686934723887135,
 0.09163895297515989,0.6081957417834498,0.9431160767965459,0.972505527592819,
 0.302090766554755,0.8524038817007276,0.7547246012683005,0.031009909931464685,
 0.003180789105277526,0.8686478981714025,0.5453341032077387,
 0.6600050872118325,0.3859745873730289,0.5862135442808137,0.05481436685323415,
 0.8756701597182682,0.10671911646570209,0.25914819491175134,
 0.1626195888970191,0.11877270910950244,0.6630682818791341,0.6584489072253832,
 0.04885684632744636,0.4420889530453639,0.773496734097594,0.5453704269791722,
 0.4873120639860592,0.2955925743142481,0.012366688521576408,
 0.2922321046862971,0.36829682648296214,0.42275748029998617,
 0.22076887751781726,0.8138855428402888,0.9748205700006125,0.7754160880387767,
 0.9567594196342155,0.1412151076280379,0.36699619586100096,0.3343443872362404,
 0.7233820150274372,0.35175249523350693,0.5911771783789332,0.8483809936593341,
 0.43948026676666485,0.07544354834019185,0.7345113755558912,
 0.18247253605364389,0.022030077474773146,0.8391649257323767,
 0.1647777259492399,0.5488886147234554,0.037994648582443946,
 0.6382269619008238,0.5346205920856512,0.6146925678012157,0.7915977833817559,
 0.4266088171743295,0.46503782034033847,0.7596730359235029,0.846768758465751,
 0.6330974956792444,0.529157766289145,0.45095829708024016,0.18591236300069713,
 0.4346467592459802,0.42053336323830615,0.05226287405311181,
 0.5784283325679623,0.7991545895952684,0.019040273232544425,
 0.2685800734209597,0.7640213141687828,0.6616690713860266,0.5556896743417299,
 0.9815260758335855,0.7619414842799355,0.6436878789004487,0.658682674945591,
 0.9438421749329632,0.4591990083191163,0.5118057102079544,0.9437668552844632,
 0.06296455217634522,0.5115602543760202,0.21446010021518003,
 0.6064895496058524,0.5161151138877692,0.9092076405945458,0.31507386809538274,
 0.08158114423513452,0.8097967013600541,0.06100370141605349,
 0.16214149425528523,0.5950509673932756,0.5434252012733083,0.9226821550887928,
 0.31249964235452965,0.1972085190630326,0.7929084662791777,0.2681363744488855,
 0.6071898672640796,0.8863710859926797,0.934751154160357,0.22614282718198797,
 0.9934830579225847,0.5891019607129867,0.27283667778240006,0.4907730109361188,
 0.09692898461714461,0.14287332557734989,0.1847200980272915,
 0.2887997318180048,0.858510670101931,0.5084014627437843,0.468151504522786,
 0.7158703203299941,0.31339062198250434,0.39952193455425633,
 0.8040511778051243,0.5515269262626542,0.8835677936526879,0.13465917197373978,
 0.2906224195942959,0.1601314022460586,0.32244809816446596,0.3111755470929912,
 0.528829469643051,0.7828929366757014,0.8959127161111106,0.16642051787492307,
 0.5657408375049662,0.294470093990404,0.07995216356535306,0.17456636721344898,
 0.7227187648876985,0.20190981610003245,0.005394166601663253,
 0.35181251499161903,0.02972883046652086,0.09020664825748081,
 0.20905660213651833,0.08239643173829725,0.26199703333071,0.664323092829709,
 0.080700994645316,0.509281521909487,0.2323800822134514,0.03812704569918435,
 0.17319601977097443,0.3861543920799908,0.6758970320503357,0.5165996633979477,
 0.5225788885144488,0.1681639919047646,0.86848909318568,0.7338267436871289,
 0.05796259196202158,0.7101012162671096,0.020967816508560944,
 0.7745614864841751,0.9133738009694687,0.7299967260381484,0.7488946222998347,
 0.2722357054999751,0.6228998851784153,0.5186723140566746,0.20771890500293544,
 0.27766421283129294,0.38963217750447565,0.731769453998383,
 0.28245122406879564,0.033138251042994504,0.15771654456186246,
 0.3525889071432965,0.7919953574614664,0.3018620008543409,0.6213010087744618,
 0.49857608802263476,0.7381394917403132,0.5391914029656398,
 0.16480323717557988,0.9682565597378912,0.5626169291148834,0.2239995302794836,
 0.23540694275142982,0.8372406463960652,0.733357386534157,0.5994042460087987,
 0.3902676776021261,0.609566449048934,0.04753061303291517,0.46541031335337346,
 0.3661690629736658,0.20530883987337512,0.0701951977697799,0.8897648994791578,
 0.7472941829496248,0.6276940646245008,0.7726572964709199,0.2668913138470905,
 0.3646164303400109,0.1380284619084018,0.1609941544716167,0.3253647633238369,
 0.4164803926959686,0.8773091964438651,0.6111667800017868,0.09226428006881271,
 0.1634923353586505,0.04648903635070434,0.4242100174205031,
 0.10534564965385473,0.35094308865889967,0.254976598427046,0.8901388742040356,
 0.6573175653438995,0.6351456070994546,0.8394358479290405,0.22258719067225996,
 0.5920952047197718,0.9625215845247369,0.8598634768317674,0.16429630983590549,
 0.5142753678665167,0.5542535032913932,0.7602140169759717,0.7796811946855483,
 0.9325852333537596,0.16036249000596126,0.5577973019861059,0.9758989764248204,
 0.27957293068036115,0.5400091232838178,0.4072879223321648,0.2348143459908889,
 0.24356521579143586,0.4259239309780368,0.9148947652054493,0.9645094182436119,
 0.8490111904300182,0.3421456471159898,0.0015003505612047885,
 0.03707093485418933,0.3477200141403629,0.44281090937966805,
 0.9848462508339098,0.5873615502770682,0.525938826920149,0.8818475707470274,
 0.39102837889036834,0.41272361880753594,0.4730632845604227,
 0.29532289080696084,0.9980355924913433,0.8573191243265191,
 0.21208735899478093,0.1422475849210144,0.904220483671813,0.9466250778682711,
 0.21682527926249218,0.11713622135820623,0.6358048977157355,
 0.5496027743210611,0.52656020339717,0.15726903952505633,0.7869805574036188,
 0.26130950380251206,0.26616565998455266,0.5962457080013436,0.990078083409966,
 0.5483331795068257,0.39867612362244653,0.5512562502606648,0.9212965794249011,
 0.8949745270969662,0.9670501170705842,0.5307537696690515,0.5382519193127744,
 0.5162291323288694,0.6887714304475419,0.9853511122302419,0.554831078460709,
 0.6695831434329738,0.8462601173568031,0.9958201091838815,0.16897846149151252,
 0.8604115903602523,0.15553968785163275,0.6932797915606931,
 0.04195370119395614,0.10275044766178731,0.09809868682922507,
 0.6227241750142755,0.47407385082015197,0.5942444100732995,0.5116195178972084,
 0.7669371689746831,0.7364321705202026,0.17835742269036703,0.883859692406277,
 0.5042542678025517,0.6142669840887095,0.15739293174228708,0.8894342849993705,
 0.23477889634838311,0.7462308779442751,0.7557514724829872,
 0.45684880138356854,0.13728970503084093,0.10814344824917344,
 0.5663465511938091,0.20989136534309039,0.2690167360105744,
 0.36269130927873017,0.7361525512077924,0.055350801328931665,
 0.1308536299714813,0.35974564326817393,0.5700418466369013,0.9016941933593187,
 0.6619194865527493,0.3945738552134661,0.3583594009877644,0.9666673560191383,
 0.7405581408998314,0.9588488193059589,0.11722982892273048,
 0.25330551953072133,0.9203636029965536,0.6577102891390855,0.961887571244391,
 0.054800478898060834,0.9929502279403651,0.2345526078667881,
 0.37362698882450573,0.5001193987931857,0.9565810120891896,0.9188787450102303,
 0.41450673145021755,0.8949178191297894,0.5836799656988355,0.4800231744760586,
 0.12578751545149913,0.9393841643227431,0.3249844554173511,0.5427354087610939,
 0.3195343246923561,0.14849494991347578,0.6530296977263277,0.5251753577081633,
 0.2261224372279862,0.37742241271389254,0.6981414541774882,0.5716356190540727,
 0.8297799747819065,0.7511547942557895,0.6452061325008618,0.4460229461892027,
 0.07745728370149463,0.034066090642482116,0.7791751046293031,
 0.677381555212545,0.5078838417964242,0.0759138980133034,0.8509818526663746,
 0.05331250694219769,0.5240174960266593,0.09629793597193259,
 0.6255574908756164,0.33389226564590124,0.24475643645682543,
 0.40788263236390154,0.07214377648930947,0.009390319179934936,
 0.5125085766294496,0.277453380931595,0.8032296606868173,0.8078423688007663,
 0.006491124970031348,0.4966722484624926,0.868014194265401,0.9180441009805085,
 0.5807501989878656,0.9748947486936158,0.8995994348139513,0.5905302861553511,
 0.42574613819444473,0.5770551133399147,0.19630838988176946,
 0.6879360563247474,0.5486847510639453,0.04272280720820543,0.5774512553578942,
 0.12536602304688638,0.9386614138074747,0.5070118252015843,0.0902842473063813,
 0.20997736736064,0.026334940878982582,0.5964307145988554,0.3016474542936999,
 0.2516681612813396,0.6690347795988226,0.04420891316766684,0.6247972882199282,
 0.3162069083127529,0.2564231011285881,0.4822800173085191,0.8863323873271789,
 0.3620120698595186,0.21376292743109926,0.5457434477231221,
 0.028057756076593332,0.650922269226837,0.44924007258450827,
 0.7506115605765084,0.267069626750432,0.17897093690421628,0.9156594866161559,
 0.6406961198139359,0.6037003353821966,0.8268167851764029,0.4401947274852991,
 0.46117715505629775,0.5170706958084151,0.1342460345885692,
 0.23228251881372075,0.9422767430514842,0.869414894870793,0.32843516488435465,
 0.5361374865688513,0.1858782074980727,0.9062678103785642,
 0.006940253189018586,0.31163384818034623,0.8079631037090542,
 0.6331641999481769,0.8354403956779282,0.32917322689446327,0.5993802440556405,
 0.575904829961513,0.5618445511947048,0.4268951155434192,0.39674054184101426,
 0.1272414464929762,0.006225130025111825,0.0542723544333964,
 0.06357905695397392,0.6793095994229382,0.774097464933255,0.5650447853391404,
 0.43812114950055914,0.900515814190884,0.783660409885107,0.9216696522107365,
 0.016045829094085695,0.6095233867127259,0.3430286552936197,
 0.9923630319675509,0.14332006134947284,0.27469327540885646,
 0.45857982738004577,0.6832589821214381,0.14339992095053256,
 0.5899508137796557,0.05477910411773357,0.977734941982608,0.5528514508192633,
 0.9181031475748597,0.5545076707151279,0.9665523783587158,0.46174696209315536,
 0.9969840261121591,0.2685771754605729,0.3028497927046411,0.5637018637519109,
 0.7198217243562967,0.7735924663393029,0.5652609264181861,0.2180331627194837,
 0.5921013755140921,0.7807677700511593,0.297975582877988,0.5050678862150353,
 0.8063903479055632,0.38596820275545995,0.329550744070394,0.3460325717662096,
 0.06292444892242266,0.6293224572117366,0.7856551063843107,0.334874870553723,
 0.3473663629219267,0.7900204537185728,0.9445817705188566,0.9350017536432647,
 0.866316445331073,0.908663962478,0.395617762492833,0.6175668581010393,
 0.6731119554517108,0.6466083412681517,0.28363094071749406,0.6814764446426123,
 0.7033687431537122,0.9625096389165119,0.8867318688240193,0.4788933779502893,
 0.9535759683588831,0.1395501544853862,0.5925347816964472,0.32574050327303006,
 0.8624860241524638,0.5907161910090013,0.2494481556574879,0.30117655151850453,
 0.40289702392086335,0.5232449909618533,0.05906316211700546,
 0.2062644390882301,0.2991854619927584,0.45753192059304526,0.8067357096363165,
 0.8876807016074316,0.8692053760452739,0.9117798777692787,0.7463848887870042,
 0.17592076166851034,0.7391167079920001,0.508255598338007,0.6812007850491117,
 0.5016371391326112,0.9113076901411924,0.07141288613260655,0.7968129628467538,
 0.024336822719667528,0.8676635788261435,0.7473703131832582,
 0.8915350198490232,0.17973825287986356,0.24277458927920492,
 0.6813082278083895,0.9709451627507015,0.1001003492131276,0.18127369990476017,
 0.05711179961492474,0.7552112870934147,0.5140150502072758,
 0.41689001311510565,0.00949407090647525,0.015828522278321544,
 0.20479611512404683,0.8513373186979163,0.13207469626467738,
 0.7713795500750799,0.9245931903137552,0.9153532182066494,0.9802231303592708,
 0.0749711184864954,0.42349263542306803,0.6509276661232684,0.876169287762097,
 0.019830188464176612,0.1756539131987498,0.7077111356840166,
 0.09950802981179485,0.05074292068194741,0.7832502214562718,
 0.24517208849044458,0.37046677342148027,0.48113514563837456,
 0.7162550705506926,0.5745709375577225,0.14058084699515683,0.9423071621762944,
 0.94861258282073,0.27565450003792547,0.44921993568194607,0.8932135954079918,
 0.792400506846527,0.8656178204439555,0.642253874372953,0.5767463678490625,
 0.2863931851603455,0.7925627604461125,0.5772433350433439,0.26949465671166173,
 0.7680063684776565,0.415516378591396,0.5988925635393393,0.8300206042874463,
 0.3354085328529991,0.36046626130981274,0.6687654861820826,0.9142623557848097,
 0.6180688159162767,0.2660937939092778,0.2977511194243456,0.18511984801361026,
 0.7774303438614387,0.5349264177297695,0.28200722174694726,0.5537098143085556,
 0.9286950608226625,0.5466639197724742,0.31151876634823394,0.9514595521184933,
 0.44534532097801305,0.29069753154373457,0.5229955400447268,
 0.7192739357412733,0.8072515119798748,0.5922002154782029,0.6979362976492351,
 0.40844641040665886,0.48093405818954915,0.33354410191173267,
 0.4233595371511021,0.383465907919335,0.7400184324970779,0.598620905541433,
 0.171487417899979,0.6058170431804064,0.32328377909156203,0.9705262554510385,
 0.6989131185469588,0.9963206388525502,0.5252402376471453,0.08742126266048156,
 0.8697526963328782,0.7922335184550449,0.6557341104279194,0.0760050193393601,
 0.8058311349252407,0.12111486127023996,0.2574368596038753,0.4672898605579888,
 0.8609716368517634,0.1958645262322113,0.6191113944583673,0.5359562385435319,
 0.674502270406745,0.363890055295903,0.7610925881740743,0.1589731004344429,
 0.30752647489418394,0.496731214293311,0.37575886819935544,
 0.20732905218157005,0.6483750552158059,0.46854632180209665,0.541687237318693,
 0.7376972789793655,0.3767798066260054,0.7209423668974596,0.48740100657821384,
 0.9314334148151606,0.6818202212026122,0.13210046714216572,0.6569507160349303,
 0.563064220927829,0.44982102314256467,0.7041767113104219,0.02977333299876883,
 0.9019462656832363,0.6262695527440607,0.9935130958244838,0.7788648157507414,
 0.4044507731067355,0.6073907146983397,0.2395416954582601,0.5588925532417628,
 0.1900125392256482,0.5484107471199675,0.2079753598981856,
 0.034964106472806566,0.4610603610138925,0.875827199359316,
 0.44875686532359793,0.5805282671551285,0.9639814647627951,
 0.44595336019951515,0.8526049718285292,0.5382621814772182,0.9217229682980501,
 0.7548237902417511,0.03715792328419565,0.35494893502071534,
 0.31772768650291416,0.5242338276414316,0.9543082517733302,0.9937616138070032,
 0.8602661730923689,0.7631810587907136,0.7784061033381181,0.9358222561086899,
 0.8031452343483974,0.11353271378024421,0.45455547069487334,
 0.6614316682864674,0.0868155385665974,0.3066474865086284,0.2622507254390839,
 0.9893538047235205,0.5451598696895281,0.026676633067572175,
 0.3810857901587622,0.30870915774204344,0.5809646475779354,
 0.024916756499480885,0.2614792614839353,0.36324571198489364,
 0.5344906306624146,0.465520948085138,0.7454532417595503,0.9734885608336771,
 0.1247164405748783,0.34181867434345314,0.14120982903045676,
 0.49948317726987657,0.42046082896965253,0.1623188346725668,
 0.7379104131407914,0.6397834086511255,0.2976624792078306,0.05488465310032031,
 0.45039311875036614,0.5014157129904644,0.03995936240625941,0.774091640431877,
 0.8605702405684497,0.6181199414837862,0.9416508902111183,0.9523380416633328,
 0.24622102807875457,0.3729065074409955,0.3866650710817765,0.557437437637307,
 0.06915080939061713,0.8996685828897977,0.440772207101889,0.2852336363106347,
 0.8092710729169935,0.520985657874755,0.35333701048520105,0.37495106255800836,
 0.17327311345015695,0.8571241013698578,0.7022533513860937,
 0.25537460556174285,0.49441741144411755,0.5574426334840672,
 0.08257845161096766,0.9004141303881421,0.34332481872604625,
 0.8446692891201883,0.920139662912749,0.6321711274685321,0.4494584624204402,
 0.4022916745253148,0.7769658913863388,0.01615605547055099,
 0.47756730856965324,0.7314151503842378,0.5122457666844457,0.5632323313698606,
 0.4288798533924707,0.7526567994793848,0.22562122645163463,0.9801203544030295,
 0.4145890531716774,0.04663097175172004,0.6344739106597277,0.6529536058993941,
 0.1471412185256138,0.62695981860402,0.8656830092027127,0.006179601047164307,
 0.12301312907726902,0.06786953250508043,0.5857378201522057,
 0.04844786253130173,0.29336890685092787,0.7349021895051114,
 0.12019460870544507,0.3872651514675153,0.30735193299896624,0.188303298543954,
 0.07796352348342017,0.552515432552403,0.4912287300159467,0.5402583882835525,
 0.06764687761075328,0.7147994618159742,0.772491254211331,0.7881903326483819,
 0.5586692723872309,0.12179506403389184,0.6129100531686285,
 0.40141567600296113,0.023447056525785492,0.6327961908318883,
 0.5156267738328434,0.2788824284025154,0.9349024477845398,0.3824775211100694,
 0.16243687303454224,0.06313092483473337,0.8208066200214058,
 0.2530145403725481,0.008602525620226786,0.12656732421752737,
 0.19603231899506957,0.060244948880911364,0.664482411572435,
 0.5325153356224545,0.9039479982473664,0.14205261208522213,0.6873596722107154,
 0.521318088833078,0.71167900271231,0.7291493502277294,0.026427944387928393,
 0.7514380200802852,0.0587431057268204,0.25346358233364963,
 0.02639688801923512,0.8763802259643864,0.47502433063536686,
 0.7831464156364597,0.5959425444367815,0.3760327337002426,0.04993023211148406,
 0.39294859530805304,0.22856276787809904,0.1330435473860021,
 0.8030235695176063,0.789368648813551,0.42383613027623523,0.8718386558362525,
 0.5693110588326123,0.9119762138401444,0.06536371814474673,0.7392250745949239,
 0.7580040190107876,0.3737843316203724,0.9394969485595818,0.39885398182039666,
 0.3547340788552312,0.3585461284478524,0.14740897341942716,0.5220872914701615,
 0.7941833750922113,0.220986728545698,0.35064353518458513,0.39054452564341235,
 0.1295324503290457,0.6245448338236154,0.10066959333115988,0.6729174906876508,
 0.759281395895579,0.6805578085521713,0.29647545406960407,0.314008053708545,
 0.8832438746768425,0.4014745472556138,0.5253015231388752,0.12431077055092143,
 0.8711886037027644,0.2006668524488654,0.05258335304582373,0.7979653981950136,
 0.0884660861967731,0.47239429358475793,0.18821512376892624,
 0.05147811595537055,0.4514629553367091,0.49194836081429405,
 0.009050874369195161,0.8047594496275409,0.2988778764175758,
 0.9483581298991645,0.6546060514203017,0.6227415823938853,0.5833611183665771,
 0.6159083811499649,0.9490019585251692,0.8733169009119471,0.6882946770551586,
 0.11267548217431733,0.3508197013647171,0.8566905941854431,0.6906402193447344,
 0.033368468373866866,0.5913309220972035,0.2279936866823128,
 0.4214017587604596,0.8933097033992332,0.19137669009204628,0.5067187668268148,
 0.0027027071149022808,0.6295251758493263,0.6820361884700537,
 0.609156055618134,0.5031502010559327,0.6109660189110963,0.03190368470117977,
 0.495128213427807,0.27401428477857703,0.02757177494665275,0.2871511015263011,
 0.030354416547994867,0.5485632457460097,0.09408891589804158,
 0.43516293805232176,0.8033275800098608,0.4413528067848611,
 0.09696024505120926,0.32774718516238344,0.22244863260135528,
 0.4874335011222348,0.055086591866317036,0.2131760146798145,
 0.5961081173458338,0.39986580512486103,0.38756045471767797,0.992147033572431,
 0.6220177931963028,0.40042174583812595,0.041835180621047297,
 0.5621219834765474,0.14466348442350263,0.03842364187101277,
 0.01834529444006372,0.9797697547468308,0.3566299850279786,0.4777867782598185,
 0.703412314050479,0.5494257027431022,0.09221267100789832,0.9892665994761993,
 0.02861320532418099,0.5524609719867719,0.8477948041807838,
 0.09155271642094576,0.8624806990974222,0.34095609180879893,
 0.6044324367215188,0.9350101025295802,0.34762462871538224,0.1937597964197103,
 0.8238055014909809,0.5822857620079895,0.4612491222345334,0.7453503474006069,
 0.4392292512104343,0.5342520956849384,0.8333773341569695,0.7131665811265521,
 0.5157426783273313,0.5156685631825814,0.9457761040868692,0.596792859266568,
 0.20189488633535346,0.2458500354259635,0.9573516736342793,0.8941481905299137,
 0.5404575485108302,0.5601517018102908,0.17597720441050158,0.5135609816392377,
 0.4484392679064303,0.9176898862652113,0.6062347474021869,0.7182349650105696,
 0.20431462211193363,0.17278376220666536,0.05880266693859004,
 0.6452062036231696,0.0335229444212064,0.8389736878468387,0.04153042960475428,
 0.8715269276776889,0.7307299168592775,0.6588294240894886,0.40065469275679877,
 0.85216080335296,0.874161928924281,0.49558042048064765,4.916567902283342e-4,
 0.958616826675233,0.6683251695035867,0.42352122933997993,0.851759684763037,
 0.5124793961392938,0.9827242510171732,0.9152863252625646,0.22763996392688157,
 0.0866280579534967,0.434539988316329,0.03274591268237015,0.8520917374257421,
 0.0758105516495996,0.6549428547798699,0.5044827102807873,0.3625986232719247,
 0.32029481783153413,0.8910337570827839,0.6313866643647353,
 0.45866945932603453,0.4680660666254012,0.6731726654650005,
 0.40887789609816816,0.9978934800051799,0.3903843684581425,0.640464706266332,
 0.3873405197628492,0.4029345087603944,0.3788972316090501,0.1342930803458141,
 0.9056900132503812,0.5437891701075703,0.030405915212293966,
 0.7501567631349259,0.8985649979143603,0.6226258578225147,0.09390910030918709,
 0.09532705703364708,0.8017026145624904,0.8733893094528501,
 0.45164126884985356,0.08791054686053701,0.21417617009991263,
 0.21224386826147634,0.5840176630317205,0.34561651964867846,
 0.45342757390125854,0.5881428865761256,0.6246788928439995,0.7559048154807305,
 0.8240263880365047,0.3787319276000033,0.252865170982273,0.43872037522966856,
 0.5246107858872295,0.9084742843977018,0.2957138286516142,0.28463628953771236,
 0.642545671161171,0.3422180966219136,0.020242377310501936,0.6302891522847739,
 0.7334333243916131,0.6101848726320783,0.7464981782970977,0.6065442524857197,
 0.4769502908716372,0.7516615511230775,0.7402474285388958,0.05170128123581663,
 0.1884133672858227,0.16037259226497003,0.3287814138516989,0.9808977519978136,
 0.10591310393366637,0.6424892526756933,0.6457255846963973,0.9799953591522376,
 0.6312857909185224,0.9163616198500926,0.36516769497610346,0.8077230792963288,
 0.4923570533077022,0.6043160726587333,0.9756420460238426,0.20391460585496723,
 0.23090948668283318,0.2327475839330425,0.9569199514408634,0.8018392710814485,
 0.27335763426840765,0.2860035001974449,0.7066458038448584,
 0.04916661952696544,0.2294822775675055,0.5395714233078788,0.5199041869914033,
 0.6109648483416504,0.45457315914584995,0.14776555372627953,
 0.3091586155379904,0.08637982278630707,0.9940963712885577,
 0.31569967016732026,0.4114302376054533,0.9448087173589068,0.8127087161414804,
 0.7256438134028887,0.9710915869134708,0.6935377967575931,0.6136739883792511,
 0.8868198435869642,0.7467648281929056,0.9418098524099365,0.41740884077414864,
 0.577150633872287,0.46239341692851155,0.3087209628383977,0.7044946579344679,
 0.800761984670862,0.5314119790683419,0.7692512041661422,0.9714565405554582,
 0.5566341942058772,0.23688496207436793,0.9494969515331446,0.7293569200389219,
 0.8196908759940338,0.5737673932373505,0.5229733007348811,0.7939958036757804,
 0.4201451525164295,0.644750310035076,0.5546672659193425,0.8485739783424234,
 0.6370594783167339,0.4290797169635433,0.6335510243450668,0.3192558934262605,
 0.7422330449806567,0.32746880078562735,0.10566662177300479,
 0.9555390868116285,0.527218616834695,0.7234242677848217,0.550554960000154,
 0.09360370250507033,0.4664321789552863,0.1320334083896365,0.6199807675757224,
 0.92462908226967,0.5079514530681671,0.5882955789559929,0.8837005216140802,
 0.13790447082787838,0.15017090025383206,0.789533312752537,0.3957791523887342,
 0.8202849153212015,0.2596064468659991,0.40234370995081103,
 0.39623860103643616,0.0053093681120153224,0.3842553027200246,
 0.1487875887475978,0.6536364397669853,0.2707445310904022,0.05749246811490005,
 0.03612832904377006,0.5262908137948324,0.4744466233431718,
 0.49875888196161355,0.7269961615938544,0.11431432246251627,
 0.15857444153804923,0.4107354762339521,0.44387386793800576,
 0.7181336038944227,0.7504641322352853,0.4885586868901356,0.7978753084811971,
 0.04091915558519921,0.227142364586027,0.7235419416217733,0.44176275530780695,
 0.7555553335845504,0.4932719346147706,0.38574070006846073,0.9377652564276463,
 0.49251968446083794,0.7487883442637577,0.6985604458162462,0.5339259740126863,
 0.10097550210866468,0.8780492926566785,0.8599512123045582,0.8060259825880927,
 0.8851638317272639,0.11315498081968378,0.07318872526251707,
 0.8022561511322841,0.9543795651983329,0.8721782705552619,0.8028131847397689,
 0.8022442576380561,0.12432570390678932,0.4232943393676546,
 0.07905651997509033,0.7907204989540603,0.4354112607548025,
 0.44589020635380505,0.8442184348777818,0.4925829869734665,0.569728131675199,
 0.03282533972279844,0.12718301519147102,0.6873084913637655,
 0.10618492293347537,0.01631536599647232,0.352659908076035,
 0.20571144947890496,0.5655174709494522,0.697555821039978,0.9190032879486009,
 0.7505102753929599,0.23118848010731297,0.6046447337506244,
 0.18106193138926152,0.6337179895965652,0.723055095896396,0.7591491966071391,
 0.39502513788245375,0.0960977037413604,0.016816255889460563,
 0.32777368459604095,0.5908794558073398,0.004839200758829154,
 0.1382136977312025,0.9267236154972635,0.8666006420376667,0.04542111528962833,
 0.4494606076538944,0.33383084492901494,0.9451576521352488,
 0.027781870755104188,0.873656224074716,0.20935645761370303,
 0.4546591738202218,0.36249388543211214,0.5640500888106813,
 0.03123554429213815,0.1707989756463547,0.4753600507680926,0.4528528880176904,
 0.010413564045070545,0.8205982953142111,0.45664503312966,0.7875479043371025,
 0.1024974480469445,0.2306646210743145,0.8255952333738245,0.9682861537535521,
 0.25144622127387806,0.7281012091197707,0.2585612786629863,0.9604231812406346,
 0.8654661691002984,0.14811197267005027,0.8780829276310897,
 0.43669222420552756,0.6509842212490087,0.9454507509047535,0.8448214422318113,
 0.17746601358478453,0.4656328049830012,0.884936198856793,0.07181076831775579,
 0.4263386882705388,0.7784857321973903,0.5982521070057277,0.4733569854631966,
 0.21309357421659303,0.4029779174210304,0.08970161221451467,
 0.3519050852957396,0.19599969561206665,0.8040565978607199,
 0.19599081232062365,0.3958230399358328,0.3602883103340476,
 0.46220563688899685,0.584644375488916,0.35533817563140113,0.3478086641278392,
 0.1971178681008563,0.11443900443379607,0.6690717217539975,0.7292709414727881,
 0.7013740513274015,0.9404779894409725,0.48824688922757864,0.89290645523504,
 0.21184334872384725,0.24294679611620762,0.5445882448137634,
 0.13457793284354036,0.9932616165346644,0.5333639671444591,
 0.12350463294997804,0.6012843228496954,0.614113933180481,0.1935006645457349,
 0.07261781905107056,0.06525577689528539,0.9304783097588758,
 0.6650648377832323,0.42965794261554,0.04205585994730643,0.8341316440394243,
 0.28895701519277583,0.739326632065575,0.5377917001227754,0.1896490473409872,
 0.8791110094870157,0.17207104824695896,0.7239251278625699,0.1339260153373114,
 0.43598141252154843,0.6466801480196847,0.02306727440454004,0.587866863385302,
 0.3562747980243226,0.053990995432448674,0.6064082816306028,
 0.9697263152053016,0.48961561921831265,0.4737033563927435,
 0.36445143577803507,0.11406228668590368,0.3521197814289727,
 0.45050709181913207,0.527972749345709,0.7964358964524336,0.4961657639247016,
 0.9617437948934736,0.557582947154866,0.5032257742253069,0.060278890723548306,
 0.8926179804143153,0.6023859625622117,0.5813867120058905,0.05818425593021548,
 0.2042308260886595,0.06551327127751116,0.20991220272176092,0.674207443923835,
 0.7462987664829525,0.11898936082458966,0.23266174452745814,
 0.05813042300255178,0.7131704169261411,0.2830164314761262,0.6835112265613675,
 0.6974040200214804,0.7092795368787238,0.02950042888871185,
 0.27977182918019106,0.45548639391485923,0.539107867928591,
 0.10651459553747944,0.5139458468004383,0.41243668357005603,
 0.43209689738083734,0.18497929731887996,0.6665751619450878,
 0.3504326626717048,0.4535990862327316,0.6586238196516621,0.40643054229489306,
 0.2694356862499915,0.9453526247923502,0.5526045456707944,0.6834149191124994,
 0.336871382025132,0.7867352371534552,0.23774481483752807,0.9151420040091246,
 0.7601544032161143,0.79098745766192,0.2926530382228619,0.8035865656417727,
 0.3412019234217699,0.009127639391845355,0.30126330620641584,
 0.18738242606213185,0.10944714242733244,0.4605976897538876,
 0.16567234417601628,0.7636935926885375,0.8224632859085009,0.1844352743643547,
 0.8046848590520643,0.5448259165740392,0.9833354055023941,0.1336033563605139,
 0.4968082668585404,0.40194184123632615,0.6849531661786721,0.1025659687150412,
 0.646244534625539,0.47826707037223104,0.6661159967071466,0.9691473128432833,
 0.8640280923783767,0.8836777802270704,0.2257334516795173,0.5817042654883021,
 0.7944311453406631,0.9543088636395245,0.9828919128573866,0.6935141894264983,
 0.5085955475006596,0.5676011822700884,0.8053680355650992,0.49465116556478117,
 0.4778422840074633,0.5598039689662693,0.021476844254253336,
 0.8001622136434081,0.5775271163377714,0.11930438058814274,
 0.11367455180016839,0.6146621476908014,0.30495749775382863,
 0.17488089018842223,0.7996256314483182,0.6246915496923106,0.2740284036374695,
 0.06097998894903345,0.8232483136429518,0.9313845039281662,0.6331512170515982,
 0.6352017188524619,0.9949541224296379,0.9602922523962067,0.6790925073192504,
 0.3158642967014432,0.6147298218443705,0.5749245969229221,0.6435184850087992,
 0.9173230248908796,0.29362961063307325,0.1768211858493468,0.4010539542065713,
 0.5769017518595393,0.295692275047329,0.2759607143805576,0.67826848504255,
 0.3060514691144456,0.7706157006593638,0.25037058974820303,0.9011634455871427,
 2.742393952790412e-4,0.7590769378124697,0.9619227527275859,
 0.4075793149173007,0.4007968840447983,0.11608281525058073,0.2721967454499066,
 0.8492093654595254,0.25478049938417435,0.677409711216973,0.03639426013035918,
 0.0691088057121364,0.4530998930111898,0.38974370493783395,0.6628276466751245,
 0.9147387511329248,0.8747090786487542,0.7544140276067209,0.6422664574278557,
 0.8284042455529463,0.06797546746147776,0.4393728538818873,0.8793661996910865,
 0.7714229546489844,0.943808180751481,0.5320471239447625,0.645471642477928,
 0.24311850772095556,0.3698881084817356,0.750796973404094,0.5795135284923733,
 0.2554385862428281,0.44807945668231475,0.6064444194217402,0.7703691970901851,
 0.4639619925814763,0.10168490042358957,0.619797072663112,0.668123699040278,
 0.5033138400599537,0.8444970677470833,0.33808375042330874,0.4003168126641182,
 0.0672179416009735,0.47196428317553296,0.19418873239260137,0.332904847965385,
 0.24938598309640458,0.3473386589850057,0.3703926225341736,0.2677249713870442,
 0.5401153322584787,0.5186725958330115,0.025377475917997883,
 0.3455075655378358,0.3334997225474665,0.7210680929301756,0.2885717735978236,
 0.5379373614301002,0.9058184779681437,0.2148033852463519,0.1142802533533045,
 0.19040512611053928,0.09642126628578707,0.4912284829802749,
 0.9422025054345986,0.7074376279649494,0.747687000108383,0.2311689966642798,
 0.5518359078891775,0.3535535829852203,0.052257813261762154,
 0.6569468372902134,0.7115596553972352,0.10601006791127654,0.7808610668940915,
 0.46011817609164996,0.30764136163827027,0.4635746090890409,
 0.17044004716507688,0.34135638602482554,0.9181474466808645,
 0.8392411648438839,0.16514712470881832,0.23381460405090482,
 0.8205626118697124,0.5860341172175263,0.6947280526085546,0.9654642483613908,
 0.7835752991539167,0.8328177452872592,0.6718383228938601,0.6203596432174082,
 0.511069490757867,0.14669122488852704,0.8359372357468132,0.4632713047998913,
 0.5916050564404632,0.7968019260611976,0.8761506178564716,0.8982708283175205,
 0.199823489456598,0.5816048027227656,0.7444379832615858,0.2825894124715531,
 0.9217834612521081,0.18991172591573324,0.9659184494344588,0.8846489191644284,
 0.8225708852720617,0.08570397913101047,0.42534467887201965,0.440109018860279,
 0.6415029019353963,0.41614506303158916,0.31095615923658493,0.342747769791206,
 0.14048902108202022,0.7774636876739476,0.07266275105072983,
 0.49100529991254593,0.15083699907951764,0.8001369648278158,
 0.5884691926909354,0.8220845320639132,0.3990363883101725,0.24487487067174396,
 0.04907434225775853,0.52973419490808,0.5213910443342039,0.010624010332111022,
 0.2516330070761956,0.09405125405780645,0.5253304213167567,0.5036706907711105,
 0.6491590546723988,0.5948338715481074,0.0034440031408856653,
 0.11210011069840276,0.7695063855519984,0.08003866528999415,
 0.3178001600691378,0.38489584307589064,0.31999341745673315,
 0.6000194207611358,0.920103239984867,0.10418981340824851,0.8537650226976871,
 0.4208571745107691,0.6856691713112282,0.07005403977987756,0.9930973845314891,
 0.08894977836655171,0.4053841165376463,0.11225692147927646,
 0.3592942532848734,0.9159675397669782,0.36112551207565047,0.6926840608290418,
 0.9980575296022871,0.2872251598344159,0.8715328329404788,0.6999838917188274,
 0.6815972562785775,0.5395838512061407,0.20647835976500517,
 0.14740374923832933,0.2506858399575461,0.07221037589245527,
 0.6493453925067181,0.8154466899702953,0.8276766768610787,0.07868288730900797,
 0.5123364269046219,0.9850215503354791,0.5776559087406602,0.27849460740777443,
 0.032247632765640355,0.7022888761150401,0.19851919285455777,
 0.2829003042230014,0.4356895418340587,0.13881137070028848,0.5241386632705872,
 0.9635383344640368,0.43462720710203795,0.4078802601658623,0.2749946666087729,
 0.036589537136706984,0.5164843700803631,0.3040005703028016,
 0.8890544143754399,0.31411988467508367,0.24149728120986225,
 0.23722040387837695,0.42058517165065656,0.7226108910471294,
 0.4194346108986945,0.03500074333103054,0.47366658214981605,
 0.6229271024076405,0.07851882330580917,0.8648318186415808,
 0.12569104473943493,0.7741022368324793,0.6738754935342629,0.5311501505924033,
 0.15411937816714183,0.41755167634907975,0.5713041271622382,
 0.06643424412956866,0.6349789395063776,0.9081394090335744,0.3354968358442685,
 0.7840709956245884,0.4772739379491652,0.223227108038215,0.7100143495117393,
 0.4003038422463213,0.9850836764854347,0.9098754574747008,0.7174995659729846,
 0.20455474775719495,0.3741392158006468,0.18292630156502565,
 0.18676658730945528,0.9315329518046,0.9624226366594952,0.32753227797405715,
 0.08184006949582157,0.688734285286657,0.9359302686332365,0.7882814377437615,
 0.6557302196997865,0.635612128998598,0.8922022346642875,0.8227617516626388,
 0.5570298685065572,0.6428694708619341,0.336687378878463,0.016764616341514937,
 0.24889566997021395,0.3149972755848093,0.11207307546885015,
 0.8552150425965769,0.9291346583919526,0.04557450461902679,0.7440475917863323,
 0.4096675835334328,0.2311773373593049,0.6716136965435144,0.43389345100008603,
 0.4613451303276799,0.23082508058890272,0.8562690276009743,0.7130066758424063,
 0.530537511702918,0.39701359399714053,0.5838718077336875,0.6460590618135476,
 0.9848506118684826,0.1270079222496865,0.9385809450753804,0.7804211591828072,
 0.10500672657146881,0.652035886126492,0.695932485123646,0.6186501201942589,
 0.39972156947471804,0.569761393818424,0.4954732923794357,0.5614955596889881,
 0.9512581137351213,0.7597923941251964,0.009524437789315443,
 0.3056309388416979,0.4531306281576968,0.8642651181706735,
 0.007918420394449455,0.1625169720041295,0.08732051823197795,
 0.08991759210126093,0.33977640992981895,0.06317180753518714,
 0.657586146266181,0.8410934736902829,0.9335683862856254,0.09866948465779257,
 0.5494461720785366,0.9713110541206507,0.9748552515461697,0.9659497092313912,
 0.04712504659944117,0.03036102708971511,0.24114420993381858,
 0.4127888101029903,0.548981169686749,0.5398888001109399,0.42905702617495556,
 0.7512683743580472,0.6648061530523335,0.25210330405972436,
 0.08273119620381508,0.9518218658518351,0.17462057390298347,
 0.008723249843647585,0.7991275920250085,0.9455662557500655,
 0.36609811469611087,0.9403872345609149,0.27205551769610004,
 0.4714626639025594,0.9932729541943423,0.8286638638696183,0.6276798216303433,
 0.7642909813630137,0.8896728616879611,0.36202052357309256,0.5653273894266733,
 0.5294225291176562,0.42632026629453545,0.08502773682685993,
 0.2289124518555431,0.535449489126314,0.9078223369098272,0.5983029675016233,
 0.26589059747648824,0.4360896981967881,0.47587919992503347,
 0.2278936665789988,0.5739582682626685,0.5582372038440138,0.9194136084103604,
 0.4993727620658378,0.4409358307948741,0.003834384576122951,
 0.13586828489475067,0.43923905624205406,0.8825265557354315,
 0.5968097299169022,0.94680108588852,0.08475763484433618,0.724405445007626,
 0.023465162494762604,0.18916744426720378,0.8512387722829532,
 0.4569475346883535,0.30013224707767094,0.061835931341546635,
 0.08820724416848447,0.2190228248100129,0.40095175904733327,
 0.030473596567619277,0.6623793731442907,0.3313834398498562,
 0.8707585327435797,0.24410915081380824,0.8620664141715072,0.3316990774355214,
 0.6092882965797896,0.3935551604032961,0.7823266882440825,0.20532036395760933,
 0.549111539033328,0.7174340358631377,0.926474470885787,0.7151669555786568,
 0.34253389511085497,0.9335528166623622,0.9171774585491334,
 0.08767822189852037,0.20016959359603925,0.24977685734023714,
 0.5259832467493532,0.7550145492719836,0.3953149733123732,0.5645016674390956,
 0.7413501391331058,0.4274484065231401,0.7964641491854174,0.6285428392561264,
 0.6814602297894685,0.10466703499091867,0.8871033819276652,0.7505825438630613,
 0.13710293033157994,0.5005755009352029,0.08607520041987404,
 0.13345742118453296,0.28655875165697753,0.821319538168555,0.9712194953529806,
 0.8980266485282008,0.13923193305257897,0.4077326930638818,0.33088786857909,
 0.9706374103782095,0.3949646663141446,0.20174789211252198,
 0.33601544478393386,0.4180842305589607,0.3882477782077328,0.6456208830120493,
 0.5853928155849282,0.8148516739658491,0.2252789275787308,0.7746539756938938,
 0.2040163677175053,0.20869399485724505,0.870550700304896,0.7113252857131649,
 0.6616095837324404,0.601744830072117,0.44938181924307696,0.6031691607909463,
 0.30336042040952305,0.6388397764136656,0.6337540814785781,0.2752750072855087,
 0.5083239936557042,0.3711020743965133,0.6993064690343505,0.10170218782145446,
 0.1754416671488529,0.1832292556668761,0.4997382365949934,0.08381577354293557,
 0.6217935068455609,0.5299216048437232,0.16903695167677613,0.6052054526446496,
 0.8718331627937933,0.44180746545690086,0.2934796334088706,0.8718301732202736,
 0.4893110151683485,0.8622618127139163,0.4145437975955111,0.0902622475787489,
 0.5384259193523379,0.6107843881338211,0.1551059605557059,0.8498861122329976,
 0.7755157817569464,0.9435871999786822,0.33477325712442263,
 0.21804099634706364,0.6267729141648996,0.6579940828025335,0.8922749883388729,
 0.6484540104685419,0.8305606251613917,0.4558388352987103,0.9048694427148487,
 0.3710302513466679,0.8657066985291073,0.5359661822947166,0.18203954766011776,
 0.15778232304308726,0.3398154439198885,0.9141305906359375,0.3278562514447444,
 0.5460742568166348,0.359276165777746,0.45581777488686626,0.5957637033850555,
 0.1381373202504511,0.7020462229785682,0.3346233633080026,0.4507695295661377,
 0.23152862032391341,0.9090307592588931,0.6903811855469087,0.6121852627568398,
 0.4672671215823464,0.8781850453316989,0.4394764199263169,
 0.051660100904002415,0.04526374922120602,0.30178020399061645,
 0.35601300016434667,0.38402849127394945,0.2861132036068621,
 0.9210104006834375,0.2488220676419457,0.21191393758714028,
 0.16076441239058126,0.37182120386100004,0.18529987766380418,
 0.5796556373841716,0.7502827105921767,0.9495809501981471,0.05651946366961458,
 0.9186515026804516,0.37543004171633354,0.5604643819674294,
 0.017123801875176037,0.041828582447221097,0.3833353730128939,
 0.16875930838872633,0.24545766883694287,0.3364405873284241,
 0.32940350012676023,0.2334976105134431,0.9835491830666415,0.2157982838113197,
 0.3854466009507448,0.2683314978262705,0.4220355826339167,0.26973666435211796,
 0.8773944018275631,0.9127393806733953,0.1932409049058912,0.5428544823029612,
 0.9608327435709322,0.9540316081965236,0.8239847227368519,0.660319013938472,
 0.9137882446102182,0.7278498939351298,0.9174726118951061,0.8731208715222826,
 0.574377103238412,0.8462073137595778,0.3494325359667112,0.8510027427441159,
 0.1868990861738533,0.6146875654446512,0.20730748271480637,0.2926560317839978,
 0.3975623638992849,0.6370334947233178,0.8477711124513404,0.5490677562657229,
 0.11861773589370905,0.056248333161594055,0.906013992223214,
 0.5456400094562298,0.08060958726773015,0.6395235277879315,
 0.42640366526102624,0.018510611641098462,0.6301204300023456,
 0.2085719000868459,0.8279811277564972,0.04463830078304942,
 0.09308874183493965,0.5971448863539093,0.6131736561456212,0.7641932302473635,
 0.3651851136248402,0.46043694656664313,0.692228823593338,0.6014334985198019,
 0.885815857161096,0.7956093612716166,0.6771673518631723,0.8248195241611791,
 0.9063318891924161,0.4013567097344575,0.7525751188241457,0.11900779531785721,
 0.6818889155594073,0.2784896394468397,0.9767675086966721,0.6918532179342181,
 0.047684755833571124,0.11563828712963475,0.013770325478080325,
 0.9370512306525711,0.5375968591182994,0.9918330022066734,0.8134330625261722,
 0.9249981707656295,0.3782018969759794,0.9788021560438374,0.3982121911075316,
 0.5994320365237927,0.10832519091038828,0.8157773051611035,
 0.048359438827025536,0.22810154823731998,0.36582487755524407,
 0.7209404665193629,0.418616435637127,0.854704701310229,0.673373552581315,
 0.3994550072190264,0.9287419239862661,0.8673524685983915,0.13923765631308782,
 0.8042570181182717,0.34226198349541315,0.8950291488835304,
 0.31637387552373675,0.6481973963234227,0.23288449488377783,
 0.22675430829671606,0.28270130840306584,0.4950491675147788,0.899025402318002,
 0.9404443210891882,0.49849504811961676,0.6569669551071793,
 0.02505158354984438,0.3194528830803526,0.16263439085802733,
 0.6351326977829487,0.08160559599818418,0.79588482357682,0.6748340733363241,
 0.559365854877376,0.5109209453983505,0.5258254968966098,0.8640356110786774,
 0.3667316012538714,0.45209486981927416,0.15283741747872814,
 0.8768192787103395,0.220410370395003,0.7536867575178559,0.7705713280578732,
 0.9347662003705814,0.7914841231050387,0.89921864805924,0.7893974899579551,
 0.19505217784959006,0.5512731938920364,0.827356385335453,0.4466270328818098,
 0.6533339809931646,0.010355823550692866,0.4634437435323747,
 0.6985459569291042,0.218077099289415,0.13466536266027174,0.657324727041666,
 0.07820071934508088,0.6264325502871694,0.2286394846356684,0.955787118350508,
 0.9542681718985229,0.11601195014914423,0.034791557766347236,
 0.9381467738537539,0.4481005981115097,0.8266538526770785,0.7974647272152247,
 0.6962864782533816,0.46741740964168677,0.8994958497819125,0.6541031199130675,
 0.3092419077758055,0.022412317260831593,0.3440379182247364,
 0.3928390086930632,0.22694236941471768,0.30860385353705877,
 0.7107217132171668,0.7038940930331989,0.4773458139501654,0.90262513758228,
 0.736750613504371,0.8986042897633215,0.34380870916328066,0.9147690933985495,
 0.9134813463821465,0.8056339154147774,0.5942949801961745,0.28798932568408364,
 0.21320474533354705,0.9368602054006482,0.07948506088596596,
 0.6929277856374515,0.6245584692194815,0.2769970156586118,0.8632977798370084,
 0.8116943044547865,0.1777813485687475,0.5728801820190864,0.24332014219153497,
 0.6008108886541497,0.28540205277589714,0.7425413078484107,0.5966205444585654,
 0.18592248701018532,0.06443505837513819,0.10052627857685137,
 0.1566524997491625,0.33595503729070264,0.6775428224630942,0.1322848337227247,
 0.9387416790553158,0.2557044263987597,0.5903763689773879,0.7312095693445715,
 0.6778603619911224,0.15982346017989824,0.9575585471261102,0.1309318827443069,
 0.8623267248854685,0.31258812499235744,0.16444916554754685,
 0.20778079570912167,0.09766105672806302,0.27563470459494466,
 0.16561451907023916,0.7143202620679824,0.08837478926702125,
 0.12563616885348083,0.35473341833028527,0.06767788033311917,
 0.4995458550520311,0.7059289172525243,0.8277573446342029,0.8245129703714478,
 0.6311445945548211,0.25898700047190004,0.7241530085229831,0.8833002754324488,
 0.5050546828500911,0.22606300383541944,0.21500782979630917,
 0.5669860270276652,0.9675613552858884,0.146398932942451,0.5979239182979706,
 0.14578529992240774,0.8017787138898118,0.7757641611585295,0.7163222911983923,
 0.5192071839070281,0.7577876699090516,0.48070803574180254,0.8670495028654489,
 0.11522743528076884,0.6876590373325538,0.14138223639252823,
 0.7776484876058468,0.5571760819307486,0.10159689157156848,0.5622246191951139,
 0.8199105978802865,0.25075283982695495,8.050808034801848e-4,
 0.46298868263739945,0.9904330742374631,0.6188893430312558,
 0.43515773009388237,0.5634757190989242,0.4270730345482906,
 0.45996351152110015,0.2871669887766214,0.28500981113440305,
 0.6190952345190083,0.8205319847892915,0.6541728667750648,0.6756498496378214,
 0.4703177470575598,0.9879645236455231,0.8774127287691813,0.3446511337423668,
 0.49475610922803037,0.3756363629910766,0.8778419159939699,
 0.10774004021093364,0.3158491400629433,0.6949768802464941,0.3266062533933246,
 0.8145718393443537,0.33867185030769575,0.21638026283222178,
 0.08680673421456298,0.3485724016859254,0.847141864874236,0.35248408101602513,
 0.5582053744425187,0.8943446537795443,0.3274746083589536,0.33631938639887726,
 0.11914944025850083,0.10852136178564209,0.5773831004800706,
 0.8825651425172254,0.8928712953995901,0.2065597461443015,0.857715925478762,
 0.11683207584363364,0.976221079807048,0.11138418570257269,0.67055423304061,
 0.7925795421057449,0.9723012773670994,0.6302687354813925,0.9729295407297287,
 0.5234730780576686,0.05601936845155009,0.317458762458515,0.8661080429455102,
 0.628708907307598,0.41981762909698384,0.30699158329841647,0.5070737173655696,
 0.6924516158071305,0.5293185285779967,0.11000482883824958,0.6461871681506025,
 0.5296922519384558,0.615752436352213,0.7992391988767417,0.5049795190499572,
 0.4459334796548544,0.9449609255545197,0.5269354011703035,0.6220236597163199,
 0.5278224845584788,0.41647024106108377,0.3016258624147464,0.8252742573682416,
 0.033183591837551374,0.20831330410943294,0.8053018088916495,
 0.27881028772611727,0.5564532316787036,0.8618439088336516,0.72308674068596,
 0.03787068014737471,0.6890572859107794,0.712596426713759,0.9352788040251293,
 0.35875928533713664,0.20097064245206053,0.4444550325729971,
 0.16451372493578675,0.9647127885163025,0.03767798811197176,
 0.4180948311757965,0.8159056894327681,0.9990388564176712,0.10798051648343854,
 0.3585434778163633,0.9399870704416438,0.11248494841064882,0.7606007796219038,
 0.23963578859268586,0.3402924800209324,0.8076071677761346,0.1686587284625859,
 0.6135525658042955,0.9868782583120511,0.711470311823886,0.8573750945060529,
 0.6019068989703101,0.3106208708199766,0.11571754366037368,0.6469778466696448,
 0.9682989815119175,0.758944749913389,0.06009323934360977,0.5081341364069002,
 0.663083827506342,0.0020711538395474793,0.03400159963547711,
 0.23740229936078383,0.7967419897910034,0.6062387686961931,0.2280964310395992,
 0.22343270519967318,0.6636952192297594,0.11211412100412843,
 0.4202362052324131,0.8804469654230159,0.018809826516046124,
 0.3469399284701712,0.047697380375179455,0.11148544878778022,
 0.31084707295044256,0.6851626933884845,0.21560990313175332,
 0.1599877981051867,0.9564314900762511,0.8625746773219903,0.7401876581050525,
 0.6710916219234051,0.21195839896845137,0.5817037497670653,0.973425593502229,
 0.4227824136723234,0.5580009916834175,0.07439425895341223,0.7381794355582326,
 0.3322626875469009,0.7204190492418263,0.12973841883303872,0.7528607829494721,
 0.44636855019032073,0.04495187321499372,0.7451914982967678,
 0.4694339803722014,0.877832184422316,0.33234030932632563,0.3187353105803209,
 0.8199829723036198,0.4064148103892059,0.8751127269870252,0.9463113495245151,
 0.2633869054469419,0.6181831173897464,0.9985170476833972,0.3593129456430554,
 0.008107733845794773,0.08946282360837976,0.4426895770114635,
 0.13186294815328092,0.6829266917210858,0.39833385115648956,
 0.1644102445805249,0.559810826728844,0.8701175441310723,0.5696268656989354,
 0.7871868481168952,0.5778641996802036,0.159141409345541,0.2854090127569672,
 0.9352652312171996,0.6036208354434188,0.5990651624919346,0.45183223850236875,
 0.038273701262775184,0.7450252040742161,0.1137260494510991,
 0.9310657315919674,0.7735247059885156,0.5567858344488466,0.1446765750767004,
 0.28355786556995843,0.08978406941892036,0.027005093867713548,
 0.01553114867419636,0.9383545331388798,0.0752030101925295,0.6082800909005194,
 0.7425604147368428,0.7801628952626917,0.27944073484052867,0.6955846700306003,
 0.25312757651966755,0.5838693364737143,0.8100240466388764,
 0.36845066980497887,0.07616695282093533,0.7598859224170533,
 0.36219892636122974,0.965894790779333,0.05034070081874242,
 0.04640552920419161,0.542569756850177,0.26077577118706097,0.5587115835032487,
 0.04930809490620014,0.4793600567005033,0.02753778464402934,
 0.056850355805091235,0.07819415775195993,0.9669172433486745,
 0.7875088846882835,0.0854840296809194,0.04927898665167907,
 0.26615300138645326,0.06622723292313881,0.024914036102717252,
 0.29998644210790504,0.9826104819478076,0.44934193485020635,
 0.6330811812816243,0.7532332473478716,0.3289106090856502,0.320021865977167,
 0.5179198083473775,0.5202947732317637,0.3182128760749129,0.37739178083147884,
 0.4065645031842391,0.10289989262652921,0.5845660824539936,0.9692976702805394,
 0.587292641858969,0.9733728397302198,0.7711544983157506,0.4005983827023033,
 0.034106250774359026,0.6615747792575084,0.7521101549335948,
 0.3502658618437362,0.8556159970369261,0.7753308218042009,0.9997765432324345,
 0.15120696499998632,0.15537137935059142,0.1287369483849814,
 0.8090619286746199,0.4462109077474933,0.7688250456099692,0.6315005718445987,
 0.6170351686505235,0.2724897139032251,0.33616195984958064,0.1196507754941164,
 0.002790826296988347,0.7025334672945213,0.6649045179008308,
 0.3612984099469788,0.3081707291818332,0.5340565903232832,0.32044071663404905,
 0.18217625843531537,0.5603422598610293,0.30876324117645837,
 0.35849662383467396,0.4091458518213781,0.761113005790207,0.3887400394584477,
 0.43847493787217495,0.2021977284239298,0.5545647860274638,
 0.013842075760653327,0.16033570567463173,0.2751372311076621,
 0.0528476168454286,0.655591279815827,0.8523906880150045,0.9553111003566472,
 0.42670560085126386,0.19884510910909692,0.12306890318189778,
 0.7498201514774024,0.3639902195800051,0.4368306108404896,0.7130962035817069,
 0.8094493813152611,0.5605346081629167,0.22856841016215734,
 0.16062666013662263,0.5478057302013091,0.28496823333779064,
 0.4922862237966559,0.457877881471905,0.8833685635811792,0.3689464246147991,
 0.5640749713787697,0.488358675595572,0.6638812574389448,0.046351609566053886,
 0.4787932159252799,0.21890627283834818,0.06036232673126318,
 0.6644866338541597,0.5997410723875124,0.147224690747226,0.1354688151732797,
 0.497708123150691,0.24928657581804625,0.7202683461201712,0.1117624393215132,
 0.41566738013755,0.22896774420727584,0.38873017171035296,0.6192539651707785,
 0.9187822114717052,0.5289529888620803,0.0055011940705487206,
 0.4738355527676259,0.9874331723603301,0.7332169178448457,0.1836659224393704,
 0.1998923809647296,0.5309950827798952,0.9712982856119122,0.97595549078602,
 0.12930153542404677,0.8748992446380753,0.14254153022278582,
 0.3231507060048724,0.9314338602200524,0.7805215145227301,0.2219948706634417,
 0.6097066552664516,0.2249947887138195,0.8259695782414986,0.10866336257830478,
 0.9800772568982005,0.501167987515149,0.9128524081236585,0.055842218922390496,
 0.7485903070962475,0.5897206858951204,0.4599675166555437,0.07665257557401628,
 0.7883200508199644,0.9192510908074117,0.9058590274392969,0.7208884359203211,
 0.9370609417040827,0.9288466979483051,0.6992476303444963,0.18605406629369692,
 0.6671277992125163,0.57754396487216,0.28872169265521797,0.543866849487762,
 0.05825117044876571,0.9475550574295131,0.48418546871845525,
 0.9139834633671433,0.6347284876693927,0.15847994706486157,
 0.17027649773103815,0.14181058656108703,0.8485079823545516,
 0.22689411375734725,0.7507368020471171,0.11457692008576525,
 0.7315492327710332,0.23347881340896626,0.45039815087574553,
 0.7757767590329588,0.6336570657560064,0.1991364296017627,0.08347982372713636,
 0.5721234152976273,0.6005488322451806,0.3213436609847411,0.9991316150454544,
 0.9993046361752391,0.8007606152364308,0.9234554264607017,0.323542219813418,
 0.3018158738150818,0.06325009957959571,0.28503634456366644,
 0.8146976409268215,0.4833227026514113,0.18847066901483522,0.6710502066061278,
 0.2699559317483643,0.5408673562182723,0.0019342891287418285,
 0.8213954610791113,0.6633681664004917,0.023226769948524795,
 0.9812530218215098,0.1553760641103724,0.16786424220476004,0.6571447196620963,
 0.11204364838146952,0.1526781653024053,0.9124292944827677,
 0.48070275619563374,0.54886290629943,0.4719033324451971,0.1805586227010122,
 0.034570523769662964,0.20011334880185572,0.8798348222024219,0.5160021120979,
 0.4762148435747712,0.9000638140326704,0.4070534840546013,0.19091829243126035,
 0.7099753876804715,0.5183580375657195,0.394654956477974,0.6574666984686652,
 0.734756271153545,0.30036419646972057,0.3704478187909883,0.8205322485202123,
 0.16528494285880801,0.11084697355361162,0.4099250464163896,
 0.9591316180905638,0.4773311356164762,0.2352040078287322,
 0.020165253040622666,0.1796233173932198,0.3009556880178884,
 0.7201051463968695,0.3533888800647592,0.5969710928380124,0.7431171120667983,
 0.6261077881824193,0.6180006907736951,0.013644611279610586,
 0.02971402926341893,0.05150166303588466,0.23029646652809066,
 0.46664561849384045,0.8898245620029985,0.14903972962029122,
 0.6914225594439154,0.6150588701123056,0.7064780749524441,0.5113222977655921,
 0.08212860695170743,0.2662644810481868,0.8165466046583174,0.1410703474460555,
 0.7978467447455306,0.5337578526078403,0.4375832907906726,0.6838740022290513,
 0.5683567620475207,0.11958908550115599,0.5972314442519134,0.9022731234238289,
 0.8910971343899244,0.04428424044165724,0.6591984475053437,0.2183977139979234,
 0.2739851758046532,0.6302118960960345,0.4117164430609179,0.30423818184931295,
 0.7651959291756498,0.07054813859001541,0.5845460460487111,
 0.023630189075859453,0.017158620792662727,0.18592426298534215,
 0.9843812546967563,0.1936486587059455,0.11867869019011223,0.2013129265222552,
 0.5569077918209554,0.3020147807986625,0.4236602930289056,0.40043622931827993,
 0.9075082359197145,0.9916244657180434,0.502442428513668,0.3776019351910759,
 0.4497372689623218,0.8057502738912697,0.3961864094116081,0.7773725740346475,
 0.6122792277968252,0.9857027499712245,0.7303741234492421,0.22920434459611616,
 0.5028872755493525,0.7318087782923876,0.939664178771027,0.8964753526006817,
 0.7266910136693595,0.682954166583904,0.5886348053571402,0.35966813126634534,
 0.823447683454799,0.7405399761013159,0.008950610910367529,
 0.16219676601174315,0.9425683679107044,0.740084860602901,0.14716937520068518,
 0.6544036414705988,0.06974984505351489,0.26287261142674456,
 0.7774933552599095,0.3977533548097928,0.4808472605511125,0.42668468672472426,
 0.11987960583298884,0.32683309797541504,0.5432141429550733,
 0.20107526311479407,0.47596563316469764,0.18132321176729138,
 0.8683022237088807,0.055965426758812886,0.1558476648231496,
 0.21464338054099397,0.543987863061262,0.4936391516361329,0.549846428052909,
 0.03620978140822517,0.37449150037443835,0.9694465932847139,
 0.03752733638209893,0.514501706221119,0.46416312621363975,
 0.39744955920630665,0.29820774289077734,0.33608847063677527,
 0.7819137945577381,0.9700503266516806,0.6087876086879476,0.8036977702670778,
 0.33759640897574517,0.7011607647739895,0.9189795192267479,0.5538723283989888,
 0.3843785610488859,0.3457027322717253,0.364343427274457,0.6163433912492712,
 0.049536589662304675,0.83005752767588,0.9152803031645158,0.8336378198651823,
 0.07998511523814955,0.7422001147382835,0.15923934468607293,
 0.7927125357167224,0.10241952889869566,0.4242943630756646,0.2998926180102949,
 0.22858256139097166,0.09348288334535071,0.37712593943135053,
 0.7157459916255754,0.35348916921753526,0.7579049963684417,0.3812663833311427,
 0.5263068597091695,0.576429095593377,0.14345919370038052,0.6527977088546278,
 0.5925458133167004,0.6579899572526897,0.9583182818281873,0.1803048810081579,
 0.3528452826624543,0.6977625126150621,0.9981922901753211,0.2914140647455298,
 0.7884401047759861,0.3080329767145984,0.6920137483446627,0.4666109156373186,
 0.31748451933787336,0.5741986283845265,0.4326821479952485,0.8051114434876755,
 0.2557563470990688,0.6244945479007873,0.9233803646288475,0.7776433534815593,
 0.41281491863288844,0.6143910824715271,0.5849948821156437,
 0.27829469467219503,0.9997153773184982,0.4640126530188604,
 0.038312243602704354,0.9207733403873184,0.5711522257745132,
 0.9996978136728683,0.3474348586275401,0.6717269851705828,0.8519590912498503,
 0.6300829929863623,0.049942184615350604,0.6203785532326952,
 0.30751915388399065,0.8511132963241459,0.6123228858949661,0.2187736782093782,
 0.3314913298061951,0.16677991177868545,0.484909586397573,0.7842166320910496,
 0.08486527238555108,0.30383973676681153,0.09205843575865491,
 0.686010662548689,0.2782211624268808,0.1559223543170447,0.18857214597008387,
 0.38788745783887457,0.27173806836431513,0.14660906548482677,
 0.015425276529800525,0.28914911394965703,0.06511028982501033,
 0.5640648530227473,0.006374640219983463,0.42326196842714636,
 0.6113750720835931,0.6500179196004551,0.38125993580799955,
 0.42012199550915497,0.09277943557477908,0.03289555870211247,
 0.987600810308836,0.892823446514651,0.6035500091160648,0.3177330305283548,
 0.23116745269366434,0.8178935784272574,0.5076415355534072,0.6892469499150172,
 0.9950296318979611,0.8297323703533299,0.702879821064035,0.49024570895905817,
 0.09662663245957981,0.3813742094270134,0.2520273572774038,0.5242455707875137,
 0.4279429280608802,0.8317071632006716,0.29023357446385123,
 0.22033575733414834,0.19573563161203722,0.5887836704393852,
 0.5497089388047504,0.8386184758201722,0.8841240251385296,0.67011102174483,
 0.3144507553104722,0.7606547998715394,0.11888127452796393,
 0.10569529366519859,0.27989024289053976,0.2935655469098558,
 0.13575996359643294,0.3802609360036724,0.07221722875169867,
 0.7143476221926494,0.7836594179513188,0.5825562009270304,0.23085598736806656,
 0.24445800309577592,0.31882178116099413,0.9049884769266066,
 0.15014608405928065,0.8300082533547026,0.866177117836909,0.4582536143717434,
 0.4654702623947047,0.6094507675360419,0.04870995029132852,
 0.39509656893702794,0.6757363945021644,0.25037496690772143,
 0.6415685972913178,0.04720334881501953,0.2673091835996936,0.6172566695332302,
 0.3387675459899744,0.5639976004459525,0.7495861637760339,0.30914842499508344,
 0.5837786067682097,0.7193749274912559,0.5844860745074044,0.7647747373835426,
 0.4850497433394757,0.2689854069897002,0.15539174560385483,
 0.008831186193774987,0.7535445393634658,0.18727021010288492,
 0.5988931158266619,0.015826298442969122,0.5375385525967813,
 0.5356482187587267,0.3048755505348555,0.48970238940911903,
 0.027479812801201597,0.4265737249221948,0.34181851340600744,
 0.44483606716377233,0.2543797540653693,0.6353531159546775,0.7537214105841431,
 0.7616010356579179,0.7267593528964206,0.06164569544863574,0.56647383116961,
 0.09838896832131394,0.556710439791724,0.7942957731650896,0.901001440036722,
 0.06986566520877513,0.3774453405625916,0.3823714169271071,0.8041765718539762,
 0.9593901556789387,0.24955995817055465,0.11635706344580887,
 0.7296966678053283,0.16689843601433385,0.12893783655724544,
 0.14600229171976586,0.11047083328925356,0.04925257020397744,
 0.484122753965185,0.12017914197051682,0.874750854521667,0.14851309173766258,
 0.6292855771140783,0.3132116746881426,0.7552969818103428,0.816134218952755,
 0.38058779024850664,0.034446063195904264,0.4568963012523023,
 0.9775536859843749,0.1646479983243907,0.7516608766606068,0.08321769573686122,
 0.2801825088755856,0.26293255306555297,0.3829098993921535,
 0.35671334360882523,0.002752156285352969,0.6644284510424714,
 0.9565343365241467,0.4554718089593216,0.9611270202187342,0.7476074869048437,
 0.34610555793094455,0.36450268223278026,0.8127685851862672,
 0.28800806636903165,0.945450217309913,0.12552271838001317,
 0.39911277674222245,0.7037983528464774,0.5168652747943476,0.8080333686748842,
 0.3100223359671146,0.18920663926872905,0.923079580786864,0.2119656695515012,
 0.5941883327821007,0.7187748222588788,0.038053964778510085,
 0.006061621603583145,0.5377209980372865,0.26518352282411894,
 0.7553142027711155,0.14034659303630992,0.5405826562714182,0.8427798914802608,
 0.34198968909922867,0.013440654833928978,0.3444703242537538,
 0.11905692209802443,0.7148482879193033,0.27540064334345316,
 0.4765911960593081,0.5143867495137073,0.0142599934742621,0.18182354809309798,
 0.9106094409720483,0.975490934863248,0.3224973436219125,0.7923353206336369,
 0.4932929813069734,0.9417647439299521,0.7369588574718552,0.4197071444659871,
 0.8902555112877937,0.0876184142448112,0.590140090526835,0.3940937226491338,
 0.6306654248189303,0.04380146044319322,0.6759856254764274,0.6330707854125631,
 0.7201487861475406,0.4460883007314692,0.9062366954108794,0.3950640847039879,
 0.3042537554511442,0.161789230853712,0.5523472078898359,0.285650158382726,
 0.08451548457370572,0.9559799354601644,0.8583335827727339,0.2903378260511813,
 0.7176149217556884,0.0443995760960646,0.1259636639772076,0.8763074331476338,
 0.9733321080778234,0.22199125294241762,0.5288905869140429,0.7073537886207443,
 0.1626696145396762,0.4990851431397777,0.14898055713044267,
 0.31818531124198546,0.37138005150872355,0.19679883604799053,
 0.9517545609298314,0.016299026248139947,0.19935242010358278,
 0.0972312291587637,0.7207667817824894,0.4524618226076782,0.12482518512186924,
 0.5993593891853881,0.6389564064984792,0.34033942860325783,
 0.045266114037528005,0.018327024346732168,0.00669526620630867,
 0.5109640958287215,0.9662139267555417,0.8576882320741157,0.8051318804214065,
 0.9861546709691486,0.5536637422769921,0.11531565443719471,
 0.12346917927757772,0.717220292755225,0.01901108334556767,0.6015744975323168,
 0.5850781985542463,0.24085186483399745,0.007614607201883783,
 0.6843507220700413,0.6125451443381846,0.26093998776215055,0.5278799771824371,
 0.7112917945785657,0.9274974790948198,0.7309450717322346,0.9854182308448012,
 0.6531879946432351,0.057503345009360274,0.7252573199103434,
 0.8105490844066895,0.2153305574749228,0.8104613010273928,0.5501448186120234,
 0.5405871230148611,0.008056322556058926,0.6909938882833883,0.735471505144395,
 0.4720403363696344,0.6896396320121257,0.6239084436795699,0.7433685732466635,
 0.8105222076899379,0.6677243163954939,0.5027296138867933,0.5983860282621327,
 0.19874399410340926,0.7602651557123263,0.5716395758285054,0.884998869342114,
 0.7845031531835909,0.3157685609308887,0.5325553068534872,0.3721128079180783,
 0.21309376309961592,0.7183285120992458,0.12441534437199164,
 0.9377301392865329,0.524730049499533,0.5149898086363769,0.9525782547513072,
 0.9454567391644233,0.014499416155623113,0.6984991056354747,
 0.9372535128036238,0.29974674300381343,0.38452284356499744,
 0.8340861834886446,0.4183937462151459,0.7669140091104125,0.11320111690224022,
 0.4267233429841264,0.9922873884050897,0.04731751486434199,0.8714541722602653,
 0.9857199183812544,0.8837952212028896,0.011142911078836915,
 0.9501871511625403,0.5777428256938657,0.827837716363205,0.7514583943389652,
 0.5134727346636383,0.7764905075811488,0.3873337139525296,0.23613078625517137,
 0.41551883066861606,0.782707564767153,0.626579710900049,0.8877805049416734,
 0.6516827447566931,0.09586877927158,0.37938227275542347,0.399352248985372,
 0.8182846474854995,0.611805053181664,0.9931789716621409,0.5072115481473792,
 0.577857835275415,0.5911704848706265,0.309925694492056,0.7895606555506074,
 0.6765064531537903,0.6372223343640401,0.9835405579311314,0.8172666267170722,
 0.5500421000944118,0.408381842194677,0.03066506448760986,0.7136144108220416,
 0.11578538514329062,0.172977374671753,0.4747086229457863,0.11082987076371553,
 0.40345644704624284,0.14484954982644715,0.17411318001727416,
 0.19467755405878406,0.30650340306856394,0.8814401653463262,
 0.7811038287394042,0.7140706852182496,0.7042660757281018,0.5807744662548703,
 0.6330982398474247,0.8731711900377497,0.6536498388359153,0.23447089070657456,
 0.3444731245582555,0.10094080948655637,0.36640098907541974,
 0.19548818465726192,0.5592882577875876,0.479161982901521,0.36414287534423606,
 0.14320106555712264,0.3511759789216524,0.5969390419258057,0.1018495541744393,
 0.7741072944083952,0.758640049377897,0.18752612950112058,0.869231893502704,
 0.5665178654267309,0.18896239324747355,0.6216552130749163,
 0.44863797522827276,0.9230312893428507,0.8346782620899282,0.2960624689402034,
 0.04421235487612363,0.22686583924826997,0.6109472964535844,
 0.03198510695437062,0.9075973518722041,0.36730560062907935,
 0.7077542703345527,0.2943941852744647,0.6367551072333367,0.2653146637386803,
 0.9626148411260508,0.8923976829453636,4.159788228985661e-6,
 0.15048541402845628,0.7806998042618674,0.712609259685856,0.6300228996254761,
 0.26563013482991193,0.6420725898276286,0.34529661203624784,
 0.16757100782534895,0.8083604499692628,0.23837880665783018,
 0.5388678882647635,0.9555120525153641,0.8390856529530214,0.34142651893721654,
 0.4738799350831576,0.12435219945928899,0.6814816938071971,0.5331434078538134,
 0.09197592843357649,0.3304074429417888,0.8945753915698654,0.4258300942600213,
 0.7907581963342477,0.14191224094986366,0.6227483054770708,0.3131580505990491,
 0.5858243409828003,0.026702843008093113,0.5114816171063015,0.552425108526386,
 0.7884534073388717,0.4358012449267328,0.9477083011363355,0.8285279017508375,
 0.06921130962907762,0.7470661144198858,0.9195150968902672,0.567333046719312,
 0.4056707393782597,0.19715739776410168,0.18846861823052752,
 0.3033755346296617,0.5081795641280846,0.4446372850228868,0.01537274395046595,
 0.4936991139448693,0.9228013278904958,0.9778997029980031,0.5347251458681034,
 0.16733733192733335,0.4705997398580254,0.8756578877072106,0.6440712872192912,
 0.6963318259247862,0.10195151594986707,0.5650724032077853,
 0.21745644271677056,0.5333393171597843,0.6828834990537389,0.5988197749602964,
 0.5479067970581009,0.37030021991654327,0.38595627482875416,
 0.6853510570900427,0.6582481449282221,0.742269757323829,0.41526436789646537,
 0.5808779266723971,0.49573312255544977,0.7336389747415465,0.6004434862406208,
 0.39398886379987796,0.009373107513964807,0.6770127366212121,
 0.2736256503751615,0.6636048271319249,0.42320536557387356,0.306492636067521,
 0.6226723341980485,0.9270542300007636,0.32210117152178475,0.5526435633363495,
 0.5933122978996651,0.8618720623366956,0.2632276489678532,0.31123926435942806,
 0.5802812034826903,0.5129960271891159,0.05330801411925712,0.6080690757750823,
 0.9549812312748167,0.8279900206580629,0.6435348865156987,0.17668780535707296,
 0.5555319975615194,0.1592629582555356,0.9745572082886795,0.10644282031619179,
 0.6442653500712298,0.22543221131080238,0.0970284554414208,
 0.08746492215474533,0.6586284062944522,0.4115007995145703,0.9639414951365342,
 0.5562576205714505,0.7928514785589883,0.30128654191841997,0.2391216421558734,
 0.7467241047684186,0.5853798058047137,0.735307217292092,0.37244782565492274,
 0.8364714338960841,0.1728990587547511,0.08079817584128512,
 0.18395866949091444,0.15813123584166056,0.7151495835333664,
 0.19843154151267695,0.5851710425660818,0.20421346362586557,
 0.5368220276487905,0.4446589198442509,0.8175503827465641,0.28483739805786035,
 0.8479802177833178,0.37453977670394123,0.7878808225165206,0.6434986333873468,
 0.5373028286238698,0.18767116484582014,0.23586104392353258,
 0.48988345338093175,0.9736049860780652,0.7712200193663548,0.5484725389692158,
 0.3647860701490929,0.6025826329810632,0.8241339813880495,0.21531582070907018,
 0.20429304081030975,0.6765171135231873,0.6948102777193671,0.4160704602612274,
 0.17278674614090306,0.4994974114392976,0.5268170401855151,0.3989488546158775,
 0.825221830988651,0.22090403619222165,0.47316323952194295,0.8420437138983754,
 0.22416421157799993,0.946598982631569,0.9794310738218418,0.8994738646426211,
 0.10397710613111466,0.0463508806879519,0.3010669170737609,0.6911934758338922,
 0.718811069059252,0.25322804362375095,0.9454327887724231,0.2043259638501156,
 0.7913776930080729,0.18514339830773663,0.49956721541952065,
 0.8062446424709941,0.9246688783270873,0.6018251093890589,0.6936087533633479,
 0.05725015713729864,0.6135789137025878,0.08914485465616417,
 0.1525095014247111,0.6499914561089467,0.4549660582419005,0.12100133774876964,
 0.4484699857013946,0.055278366240944954,0.10944951853219131,
 0.5582886923982313,0.510022648529372,0.7733898395865421,0.80924330567737,
 0.7357930615282502,0.05749120591584056,0.11053122034454976,
 0.6885214417749288,0.2334793619953459,0.8500837989753762,0.5889926581400383,
 0.23761914390367933,0.12351653562894993,0.7532816380576178,
 0.4226298598361661,0.04782126309091961,0.13848493329065203,
 0.8243134041925416,0.5022355625650062,0.052324127922916186,
 0.7259260773467151,0.8339603191201638,0.7092689660625969,0.23123979652645354,
 0.5286349630270282,0.20670563598256608,0.296417318634413,0.5185064065854956,
 0.5728389886374825,0.7275292939372804,0.86340585997402,0.32028623950690993,
 0.9022190683369132,0.938092453016525,0.8921565929135775,0.4974750466118857,
 0.5547133337712574,0.8855158355654099,0.8709833131609299,0.1165959681752533,
 0.9734994052879529,0.8460936995568445,0.41752926715578187,
 0.05844289622870236,0.980854830053535,0.5269711132136919,0.11518324770249144,
 0.781404870797885,0.2804851790875629,0.19473022823247,0.37693052506153046,
 0.15737114683184195,0.6822157067516431,0.32381052559941537,0.601912402868792,
 0.31620920276285536,0.41686264639154946,0.1931421135805983,
 0.6422758086935554,0.9284401766096952,0.13248475413091465,0.5058186227742899,
 0.608900703815882,0.6208817609923593,0.4837298034745575,0.6314011003702024,
 0.4879535114105248,0.8238600139626754,0.21953241665726309,0.8749217066901706,
 0.8862621133982035,0.38832364477025705,0.7472593191266865,0.802432940521255,
 0.6076583273320844,0.17943670384401622,0.3334086331271259,0.6060677906046362,
 0.23754718773985517,0.9508339350683721,0.06823214839086345,
 0.2940754972120715,0.47587833865989004,0.20998181417646555,0.34631845353571,
 0.9074385842046601,0.9881835424426461,0.8256068050541934,0.41267585167570764,
 0.42223788984443933,0.5575319209107069,0.5877254510699236,0.7148876139985646,
 0.3789449991920838,0.6692458786055373,0.5019767067839619,0.054824084773618,
 0.4997156368259581,0.25421677978635215,0.3736074533797775,
 0.19927193149816314,0.9499056494392764,0.07800330419044643,
 0.17375862758036342,0.3467934232648283,0.47194529755160086,
 0.8256701681318237,0.05194291402933138,0.101037325615708,0.9659310675146329,
 0.15005272057355856,0.9129473138007096,0.3434501444681949,
 0.33130792310533475,0.8643491299495991,0.14325924680695168,
 0.01063632952464666,0.22133895845676832,0.2853937118289166,
 0.9629894195570126,0.9311289992753978,0.9491525463252153,0.7940492796530807,
 0.7767358683709386,0.3815459937944229,0.008201328086539506,
 0.8905304477033174,0.43110145270075473,0.5546382194116398,
 0.056396735566215384,0.15502095641848412,0.8289625227297696,
 0.22820272158572186,0.34796090031167703,0.605152116988025,
 0.03169487541698379,0.1796054222738761,0.7099600889562895,0.8291724288838477,
 0.277570147567067,0.7076965812253844,0.29375900460298343,0.2529215808070251,
 0.02505943510456632,0.7636879625240698,0.5551146338346709,
 0.01300794551249218,0.3567730767305649,0.5907752887957899,0.9059053748443721,
 0.2358865386413993,0.808311537303425,0.39396828137493245,0.47907713829046294,
 0.9550856386076988,0.9187273263430183,0.008315214525360992,
 0.18068643235379434,0.09898984993874493,0.7539838467045505,
 0.8694494840432252,0.7924877340028081,0.26962249264726734,0.2454298704193345,
 0.1492907357605463,0.21020724659352796,0.710590829970433,0.8702787595966861,
 0.5762324258905639,0.7892240621427242,0.9822186862384179,0.8268412047759339,
 0.005840918556162866,0.939178202875315,0.25692820183473475,
 0.48295115079442363,0.11681691722763299,0.34143340768063857,
 0.6957601230896596,0.32523049641594093,0.8698613362162008,
 0.43220018530060533,0.19080356594839887,0.1347984497354321,
 0.3505077558871805,0.4929254766108142,0.5358436054159922,0.08212551169238058,
 0.22441792363290602,0.20548130870095604,0.36829250535080893,
 0.037162292201895486,0.12244015240208583,0.6912120916373253,
 0.05070252348796589,0.8229728990704828,0.2686520386038207,0.6118327034682414,
 0.6217761255762477,0.015551806989508754,0.6304323948694007,
 0.025544924964104987,0.255335665484439,0.5143050245505953,0.6617459601982149,
 0.4805931594278967,0.8337846929295684,0.6781247315912606,0.07986390742887717,
 0.32465937638857323,0.5120687015359462,0.8990714161571712,
 0.005222595877756864,0.09620957719319523,0.6489066384522285,
 0.2298634520003464,0.8245601087411738,0.2826850527967161,0.14754193672585691,
 0.20681598256400358,0.28035218948362806,0.3001354611572066,
 0.2158077304642757,0.06455797505005711,0.6242687157527768,0.6478565957850381,
 0.5426666778380305,0.11092978367361239,0.909734861934972,0.7484027052161072,
 0.6368677996539185,0.6975849433006037,0.5655066746421559,0.39778478977987874,
 0.8331562937398316,0.1474864411866551,0.643365898329369,0.3081133350058848,
 0.30260786845050525,0.6671215560292778,0.8169364151781083,0.3720394244035953,
 0.7719552660270571,0.3438799488298463,0.6059787604081682,0.9232836830211304,
 0.564760378156131,0.35099340206174223,0.9269637053777192,0.1869339243281033,
 0.8063239920208066,0.31868155215453187,0.636821346004112,0.7806118797516113,
 0.24055355528827882,0.04250568896430751,0.24437273856181796,
 0.0782875980615032,0.514553968459736,0.1385226317735293,0.3356406253465831,
 0.86644589997585,0.0040993323051053965,0.38889426458413157,
 0.18543675551754524,0.6884495609159336,0.0326099177175051,0.3935904761560638,
 0.4747645456390245,0.8162754296667074,0.6449407449792358,0.8824530228631234,
 0.891541486006544,0.14532306260853278,0.7140864945738318,0.7397397513023479,
 0.28319168756147584,0.22851067526105806,0.7146592474640082,
 0.35347180881201523,0.8473574657577287,0.5202379619321497,0.8325268234186953,
 0.611341605797098,0.46239740470811086,0.6332249411739976,0.8133974751248945,
 0.8808446822334501,0.006344485993045601,0.5620185307388512,
 0.9720721793782469,0.9881891516023339,0.07870904900192344,0.2996048390725402,
 0.06499892455864253,0.3538675880993688,0.02148540381353292,
 0.21094329093870445,0.9429619059147246,0.671951857317024,0.7711348687587534,
 0.5914456975762435,0.6573129142833807,0.4124056774765046,0.1821949084193527,
 0.6892683937307593,0.709107963130697,0.014525569726928311,
 0.39806508448160893,0.5687901752901408,0.4054678161080423,0.7911296832626534,
 0.6833162609317093,0.8365395631475642,0.0694560382439885,0.7459955869899945,
 0.5280655157015838,0.383638451088651,0.6327814488167759,0.6532380016681437,
 0.6549745973072265,0.5059432472280209,0.4763338053942274,0.6974427536276455,
 0.7632827023884827,0.9385732687373286,0.8466054920915923,0.3264772454342697,
 0.14111247286734574,0.87906207831139,0.0792129408913167,0.4004481244991731,
 0.9959001583203951,0.4231182697282182,0.40091039491344815,
 0.18832241540102312,0.29528606888675213,0.7526659758738932,
 0.8159184813248767,0.09212901446932897,0.054676619133830706,
 0.7410227252927459,0.8339534911907189,0.15774475476696803,0.7369448645853491,
 0.6978015743287188,0.5872273336075889,0.7183467900158387,0.10989419447683924,
 0.1898798687741119,0.040238226687751455,0.3108047475816398,
 0.16554987447826852,0.43162815978663804,0.9372796511942848,0.298436536465292,
 0.738366214632335,0.5658484411166871,0.7488085640077928,0.13591189333783493,
 0.6271314481899657,0.43482958862715704,0.2681371265915171,0.5491590521952388,
 0.617904676935709,0.5039505880066937,0.8625648700440601,0.45153725167159986,
 0.43919919551653785,0.871679486078691,0.9711622481007542,0.5612075000501551,
 0.48618939680479556,0.456111850370711,0.8522156319089526,0.5204954906502998,
 0.09127360942077045,0.6038726169242653,0.7633116473092956,0.8652778614826501,
 0.7537823148499981,0.8866364827515829,0.006772335909386751,
 0.1367542919044501,0.2130882775619083,0.1519644574579837,0.08897239925282052,
 0.4746110026495104,0.5530114267334906,0.8224078079541126,
 0.003958474492012587,0.5967844185726133,0.9996814864943186,
 0.4938635662735058,0.4235438641584215,0.9963111463783414,0.18968609232739309,
 0.5420175240332555,0.7850105096538678,0.29049043792163975,0.9474449798208864,
 0.1755017374524307,0.9778325128493704,0.5931900243461311,0.4888636558844357,
 0.2695309253643613,0.8998993539300075,0.632476302645552,0.41523705396986643,
 0.7752769486501115,0.8984085439626883,0.43727382850266683,
 0.21419099107601836,0.2662247357711426,0.5180289285328404,
 0.34697905384869465,0.5405398103390973,0.8491774573816576,0.6120015318128079,
 0.9657227378711475,0.7930002179932424,0.827986060235614,0.04534961675730009,
 0.43176230723926523,0.8765615626108215,0.9853987280027124,0.2141534443812696,
 0.9321778053400723,0.9842218026890566,0.8725992666713921,0.8863476203371575,
 0.15780852126836997,0.7989535359850114,0.4010979562517375,0.0963139321767229,
 0.3868680423434865,0.9035459602220688,0.9902790860072532,0.8136164179850139,
 0.562046424668661,0.6378304439238105,0.5926698746357728,0.38613864280931853,
 0.2920524386981831,0.6115138935134548,0.9022128742174808,0.8375959391782014,
 0.12242233041392847,0.33958736330159855,0.683757137939325,0.8389417593079812,
 0.17332729396379354,0.8513436191702041,0.9168689733290584,0.3483348913965393,
 0.47145827198167944,0.2563339979856135,0.3395274609967961,0.3153588148719825,
 0.86124350779111,0.5080636773903753,0.029413352314865193,0.04699449866753169,
 0.2807484553503323,0.1727434944420163,0.0015921303435313394,
 0.6851841972783312,0.14277654943731077,0.9647333114655408,
 2.3471959564447253e-4,0.9630495692517074,0.7799698818668432,
 0.05425714774115087,0.8092717272314941,0.32681600424843593,
 0.3874263760229564,0.29696885690530195,0.7505580669407752,
 0.01871638533337827,0.133540938670196,0.8084018504099222,0.9154351089372275,
 0.742288831351637,0.29387709612094515,0.42895739717848613,0.6194860147645058,
 0.8539365710456629,0.32567334322771846,0.9768559732556473,0.5661312674789852,
 0.3920342577613908,0.48569083216422304,0.7283449927858725,0.7431866780669358,
 0.1993577151602649,0.46270827623923105,0.5862449320717164,
 0.17077085358368893,0.22198577741165104,0.43671061930386945,
 0.5443180765303878,0.72854797883396,0.05224079922387803,0.26143412664453136,
 0.6143692078425365,0.642739170754232,0.37099017576358184,0.8298623734855308,
 0.0678304413031483,0.18098690180497978,0.7376458681352694,0.5065213608801985,
 0.6107081681140953,0.32166545634176114,0.45230338269827663,
 0.2682050006914063,0.527707971912137,0.352469585022094,0.552452381245815,
 0.8348922874139002,0.2602204995886984,0.6382670926380986,0.29334485317755943,
 0.735119581827011,0.6660501033559296,0.8558311070096759,0.5418449584827993,
 0.47537095389155937,0.04843155581993552,0.3489521791519419,
 0.01918298378898997,0.07885859646384774,0.7911872032734624,
 0.010969269550551886,0.3480803028778843,0.1540041382406332,
 0.2016561369637746,0.8909741651707201,0.21368959984219738,
 0.08267473583857765,0.7929412334043405,0.9151983602238115,0.3503476795037537,
 0.5210574342783919,0.9926807123622262,0.07773758895883887,0.5129088449266849,
 0.8996643469753483,0.5085085594142106,0.21129188064533289,
 0.07073869481112782,0.17977398609674644,0.2838011785551312,
 0.3209094985311237,0.38781977930034717,0.28775173903119533,
 0.17092804230280256,0.8787350699763614,0.003306858329594098,
 0.8666390643326449,0.4512976813477577,0.7141136756918555,0.8964243136102565,
 0.4847609041555563,0.15490573840857524,0.40428468561454856,
 0.8484588643040154,0.1954096804719362,0.2930329253842183,0.4379282537204928,
 0.4990094430036972,0.23523819301951265,0.8311507621657408,0.2760448860929764,
 0.2402936870409751,0.13779757015836624,0.8344926435682767,0.55193950312911,
 0.9175482587511372,0.3963113171614956,0.7618674666865934,0.1696561566732313,
 0.40694233903363153,0.38997900565231824,0.22343423072895308,
 0.1963798880189629,0.5366750010451968,0.23731962766320713,0.4717552001222274,
 0.010898129516532906,0.2892728540309518,0.655577610035231,
 0.21101099727633899,0.6975393670336608,0.9053178552087049,0.5842703403197087,
 0.12905055456360348,0.04178457265989932,0.013353993120966856,
 0.9150327830621121,0.17172544447250204,0.30152409946564873,
 0.33275202520842495,0.07780754305632331,0.35408479667891646,
 0.33338707282541946,0.6612825320542117,0.3858972760871652,0.9960820090260294,
 0.58615989030814,0.3644073133106196,0.023835843403751666,0.2895422220528032,
 0.09735019015083723,0.16708907211440682,0.251746716743111,0.6746187562909217,
 0.10172317727545443,0.6899695962145403,0.664338482263666,0.8333373614959937,
 0.19698650643692162,0.3631505444985834,0.0670177041805664,
 0.31781155158949204,0.9109439973372381,0.8168432217951865,
 0.19644703827197518,0.6759245528863578,0.33931818835285665,
 0.5929567508895673,0.9968215938765955,0.574039421003001,0.9901760285565215,
 0.15843132193357923,0.44564435290399096,0.1962022638538672,0.897248581038225,
 0.20615733923341073,0.06925205874042095,0.5685511707987283,
 0.6672563113784648,0.128499881554744,0.8499544362774761,0.6964306990234007,
 0.6375995923772134,0.6695920753621314,0.638708346066791,0.9501073506498074,
 0.5712342513240961,0.2520733933897933,0.01232740044063041,
 0.02533449086539874,0.6438475396428739,0.07660426198703751,
 0.29815911576800347,0.002667973323615125,0.49381825156976966,
 0.11857575927956254,0.8525287176261382,0.07121776301397653,
 0.03981508320688376,0.8868383654495495,0.8365471702731496,0.1801796137146079,
 0.15852811026435976,0.3440655436303929,0.7321289082078094,
 0.47073807105535026,0.3532836909482824,0.41837927997196833,
 0.5481127015106143,0.47716650148577155,0.8941158197590263,
 0.19282830958155306,0.44090694710259204,0.6205104915746352,
 0.6938709123773884,0.6252768587641127,0.24318974024348938,
 0.09271771403320317,0.5881593849509112,0.6652668585962497,0.475384232607134,
 0.8938547856698142,0.5611679352543706,0.3756317067349403,0.7509174211597093,
 0.01684069641456787,0.1646744793974172,0.8644914267174749,
 0.31088835181063557,0.8124645231291941,0.7918133382941767,
 0.025776990590712368,0.9423274322176005,0.47741457545877874,
 0.6917863519234498,0.5715994860388937,0.10802619541115388,
 0.45205880918078534,0.43327229552899293,0.29003869447296027,
 0.07042165609364914,0.5974068989049839,0.5675192662398018,0.9750345924859202,
 0.9034449761454367,0.7432439327408621,0.28745108478721726,
 0.48581994209512147,0.23696528836200104,0.244113397093492,0.3030873337277017,
 0.6684889248604582,0.6597750605826994,0.8094433210289834,0.522658547441009,
 0.028694751083239267,0.4311490560647264,0.49023723149290055,
 0.5440958433643048,0.8713488468419093,0.08736226391262425,
 0.03612832124851706,0.5130462561142225,0.63013137096912,0.008804135864591278,
 0.03616954248723481,0.8034902584212537,0.47651260001303275,
 0.5578224838000624,0.7300615106031958,0.0029452314385078893,
 0.7496469141453038,0.8448551743077407,0.019048718109472063,
 0.9423445183694765,0.43852082588803765,0.9737420099162897,0.2214172682654384,
 0.5217262443718391,0.8878614201704427,0.6688895146953157,0.8367706264896189,
 0.6421059882635456,0.7969792570520131,0.8663335133434239,0.8244093682912971,
 0.06395340365532765,0.4534118498515318,0.29469547656676376,
 0.2813552886638462,0.03162608014813739,0.39675995616812343,
 0.8322313487594614,0.9229691775642295,0.33401970488151034,0.7393815841558798,
 0.6256732033514336,0.9457738837984848,0.980984289408465,0.8846346302206862,
 0.7360256942963358,0.6749061155484699,0.4585788078776123,0.45307341911032595,
 0.05991806578332981,0.7701825955782111,0.7048732522493683,0.5088122517228475,
 0.4244041201201768,0.9327516959488951,0.5676280986211699,0.603238234271359,
 0.8993780408014392,0.9710502846346872,0.3619461447674076,0.08704266942772854,
 0.8849532523884261,0.9029762795036503,0.13945225467149824,0.2784387039022014,
 0.16337904872945852,0.5399212804091774,0.865960323456148,0.9430241492807544,
 0.8299388874319418,0.8999346072128862,0.010607020013270851,
 0.7626490917942393,0.5090766164090015,0.3509039851985125,0.3126296621445961,
 0.7511238328760762,0.6287808971107797,0.8706884898607664,0.5498722446363034,
 0.8490548054598257,0.07960896564630804,0.009326049536474534,
 0.5058639005363941,0.08315442476982349,0.9064500353723727,0.2736717869522459,
 0.064075955729058,0.2951815304937171,0.7616563685066409,0.8842131178073169,
 0.9723362703089906,0.5975833299081245,0.3310558483023742,0.8628165067468926,
 0.8312351169015513,0.45299385419461835,0.13643807020136167,
 0.00921329378027802,0.3569059095523397,0.04684139317002001,
 0.1525129315303475,0.898343574835564,0.07758362047871448,0.32830127583966595,
 0.8445406448859669,0.10188363952473456,0.20948114282338925,0.6383022244838,
 0.29444449470366085,0.03597981746101042,0.34977567472177973,
 0.445929205806331,0.8604246663696259,0.20767275293765675,0.12871412793039005,
 0.48557492408373804,0.18347038817968686,0.9915927671240069,
 0.7867284311949535,0.99607412042556,0.814379515959776,0.16478235167663946,
 0.391242233086566,0.9416653873636901,0.6974143667239552,0.435971178849643,
 0.2937084021406442,0.4692161199864294,0.8444026502960904,0.6220687585324496,
 0.33482584675081295,0.8396586781723642,0.2138144614824007,0.1451136408335919,
 0.3276209063566162,0.5411334009437784,0.31119966073766414,0.880533679686403,
 0.3299660222220111,0.10624230876769147,0.6257204214888373,
 0.057722036988237635,0.4563226903338442,0.7498283157979138,
 0.7254312617069703,0.10995219515656907,0.368070424227235,0.5280459844832326,
 0.32964206405725194,0.10890001741407018,0.18924335731814224,
 0.4615462643832269,0.8054970525311513,0.3130913312931035,0.5309886213938465,
 0.013955476377686349,0.5059399001638007,0.34719386880544545,
 0.6332794935932325,0.5536959579748044,0.7899648770987938,0.5974779710837586,
 0.06441104345317472,0.9292719689440485,0.5676840020201327,0.9579694295689477,
 0.6007987711388301,0.39237878705343787,0.5447635286292548,0.7135660968516041,
 0.011477011111252433,0.9875474854332593,0.2820582282461288,
 0.5938860081120397,0.13532190720146586,0.425313150346296,0.39691653712596553,
 0.32072951427717267,0.13426618590591322,0.6464239006723886,
 0.4366597295884579,0.2682775330465479,0.510491292368687,0.7030796540002222,
 0.5398681157027625,0.1909020087459009,0.09116546038341822,0.5213667840782136,
 0.8433079873589703,0.738780585541516,0.2585881649345705,0.32580565490680247,
 0.28416503549253613,0.7418929925163332,0.24634664100219061,
 0.10408727633242953,0.35971619592540394,0.6125728555003335,
 0.2826925941178965,0.27853980730429917,0.5803030325405981,0.2828595062873367,
 0.6645350437935196,0.23261385260402023,0.5363600333584092,0.5198195241828979,
 0.31917804532105465,0.6323525068303507,0.7259637595548276,
 0.39358282430941294,0.3460694558820707,0.6070752697598117,
 0.46651312047060123,0.8913172673501176,0.08933351324237504,
 0.7459655398986329,0.9447153125483373,0.007115149408351673,
 0.3461767528486449,0.1445354554861088,0.6051005249642378,0.7617021340359988,
 0.26358352765372894,0.761226243532557,0.1940619901822338,0.5167381329404006,
 0.4893178254723418,0.4808125668395904,0.6982555898946596,0.5573248113063722,
 0.04816110187693823,0.2382892269305894,0.33233926225181,0.5778220497032243,
 0.2979773393640158,0.99266567098358,0.5052954253202798,0.824516247131022,
 0.2002230494756222,0.03657959425468682,0.9328414331222276,0.2780090058483934,
 0.2880498596970359,0.5525471990107352,0.4217548013590593,0.20008236606461172,
 0.9571070884975136,0.7780052502362104,0.8889036457387668,0.6477176890502436,
 0.21282496554096464,0.13490971518986627,0.9800240325369054,
 0.19147693353392137,0.8698902650458054,0.6793873430888907,
 0.20532665687762242,0.396849116606553,0.02979768131859717,0.7482051077907738,
 0.3919025138714065,0.056420439251883,0.01365404351315791,0.10853209912498152,
 0.8830757280129835,0.033350589436492384,0.7749752931680727,
 0.4479704937537009,0.7966211112292877,0.7566755885152784,0.39860025667767873,
 0.4990501397143844,0.006784676806553813,0.0170363698634719,
 0.7396521879423877,0.3382300382629171,0.1737921070873284,0.12759915796346555,
 0.8545295210099455,0.9821290232091682,0.6302163257993044,0.18373183782000924,
 0.5774794181351659,0.5453745928174494,0.009441771482747274,0.719071901907718,
 0.6391019759530036,0.31917334405755193,0.24262600771700005,
 0.23854673478111832,0.5229915477160585,0.968709676720319,0.26248420283131146,
 0.8806267586245211,0.9580577512714109,0.9739684591503233,0.9053421791271846,
 0.6291549957442293,0.8718079862668023,0.427349637070945,0.20003598006143009,
 0.05390739691756408,0.8886909723237065,0.14592234943415838,
 0.10218458398117047,0.3600757949768463,0.2316006468672226,0.3485604886422765,
 0.35712808448699274,0.575892249105822,0.8065366723673845,
 2.795923339581341e-4,0.7783989494042363,0.3956417621688456,
 0.5524074772492755,0.00587450404436185,0.5907983220408021,0.9060209994709951,
 0.3852450181218875,0.30198390757256566,0.18336869524778832,
 0.4175574375644592,0.44974757184524683,0.7345430190647175,0.7029040197656146,
 0.6615378990451486,0.30122845625332906,0.5646891242591819,
 0.44053016766734876,0.3497827233271501,0.03096397887705793,
 0.9479295519658844,0.6955263627229222,0.534045889221717,0.9604242279490247,
 0.6128441864671234,0.5271829406487162,0.19085958068347442,
 0.26475127291224587,0.8786265932321956,0.20434860465656468,0.503231370386724,
 0.34983548032833145,0.3594742046448932,0.7937322603495941,
 0.14232206247441803,0.18826076361843003,0.10579053083682155,
 0.960384741202777,0.1746750111507238,0.4153486596038233,0.09820237360819228,
 0.4162995844054387,0.4537882723445812,0.35660016411019657,0.2390396761125373,
 0.8540342445584105,0.4941721158403354,0.6298204901472565,0.6301939496191749,
 0.9410485718121563,0.8902129993284641,0.8199574013517592,0.5812531175409692,
 0.4147995398289862,0.779819652729701,0.8260690591583568,0.16047545738130842,
 0.964440140750065,0.40715878509626857,0.9260404494021302,0.4764443472804276,
 0.6137537572064735,0.5357798878713655,0.7449040840701866,0.3490058603008518,
 0.9548144935830201,0.881729890472514,0.15467092396419635,0.15189725776135654,
 0.8234713621977566,0.11776736586563152,0.4219869661787684,0.4162309670223828,
 0.39372804203860734,0.1416109523712792,0.14804863171058846,
 0.8621124043674288,0.37584740539405204,0.9291744089054281,0.6522410872518778,
 0.6841801928419353,0.630043300163268,0.1800006755928779,0.023563672591190854,
 0.0722881901953375,0.5280832365814502,0.7431120596791438,0.783572164117173,
 0.36992103147750055,0.8412549939792151,0.4875330534530812,0.9837478792927621,
 0.2517994388417393,0.8602021375853242,0.04827679402621188,0.3814363543067043,
 0.05052657946917938,0.3182737799684371,0.04865095554356158,
 0.4177633138795517,0.041631598165185135,0.7364463230250348,
 0.8982459541229186,0.01242406210571212,0.2964815082612601,0.2635190073631595,
 0.23746330435532592,0.09099657437521946,0.9166043281413383,
 0.3122800905407912,0.8988313147942772,0.45020308740456794,
 0.14157368950255644,0.10072625436630067,0.6497112751515447,
 0.41440202263976467,0.8902700476547916,0.7782320512363738,0.2265980504374927,
 0.987608375337773,0.5086291990632558,0.4883043776920901,0.8510479087829683,
 0.16023188271431887,0.4998576568157147,0.2183952582998181,0.4932746314343227,
 0.26513186564799907,0.8630084140325542,0.37148303382524606,
 0.9629758414721578,0.3753465108696623,0.8095934296252457,0.4787867279729172,
 0.13715909227065493,0.12880054815649777,0.9236480400957585,
 0.6313338988284144,0.7919407621023269,0.5456956173576115,0.0358467072916534,
 0.3099629887156543,0.008387249601915059,0.5140006143878377,
 0.5355937901152572,0.029408689450845182,0.9124164366672166,
 0.35133359854672364,0.12135398802668829,0.3929158876678134,
 0.3572774831187129,0.8053365268709621,0.4434914501784363,0.2604921684137512,
 0.38926978563535264,0.35993422339344827,0.6667759367210161,
 0.8011519444131344,0.9511923751904904,0.5046786203974036,0.27469662190851085,
 0.5505198838623063,0.19209045146233206,0.16087454702608195,
 0.6036298696901083,0.5230322714069819,0.9086457705044056,0.2957296026767502,
 0.5912516253814081,0.8212193020179993,0.03075380548705897,0.7801693042481249,
 0.7667222518945543,0.4831460220273517,0.8388266208688165,0.10789874490139473,
 0.44335835672027,0.09499597374757085,0.3537240735917412,0.8434977854477811,
 0.2595984633950972,0.9051537474814169,0.46352381145578125,0.8972176486427953,
 0.48849781422043137,0.12071477109091688,0.7401318110933996,
 0.2449813644936819,0.0743273204897632,0.5618760729003158,0.06859161657672697,
 0.12820977101955955,0.2612935136227117,0.6758865394696505,0.4790265618288052,
 0.5599935667948707,0.2495601385896441,0.8201511899692687,0.09178807271733636,
 0.3455991266997477,0.39607962656299844,0.8595560839251555,
 0.19706688646926018,0.832671781496536,0.9294027925940427,0.8760339337881052,
 0.8495109262740426,0.752906109120092,0.7320713494315774,0.3219628075521075,
 0.04314310552515033,0.8723036550572675,0.8598422289304326,0.1888779306439865,
 0.18736133049759185,0.7224437659106613,0.07903461262545053,
 0.8627760051247007,0.30634402378970316,0.13015034665220293,
 0.7217759636232334,0.3023372241431237,0.05139147270723954,
 0.08192241266222411,0.5309242568369841,0.057908848124988976,
 0.32906902351115563,0.19086209454933645,0.14639335703143375,
 0.5883815624344304,0.3113990985868016,0.1607012763386555,0.41871357798940756,
 0.5573716735437666,0.0025703801063259313,0.22911877450042462,
 0.3273877144969902,0.06858519846858357,0.8627008866297853,
 0.09215854967204196,0.823063451539318,0.033482197687309756,
 0.8995987373086254,0.9559311780571681,0.13122637714147378,0.9783843573785783,
 0.3721165526971987,0.6997937533665973,0.867966715732177,0.4134001760967114,
 0.09108730963472811,0.17682088778696314,0.37980933326508226,
 0.5113803095064173,0.9540995298200641,0.5906285833579721,
 0.007014560415822824,0.06310038470930324,0.3775729394512819,
 0.5370768060170111,0.7708679774850957,0.8447542547798143,0.6904229803871443,
 0.09538346144060983,0.5434310990731364,0.528934517028003,0.29803020855941675,
 0.28545042039777846,0.08035891083427626,0.951806327461943,0.7421575599923356,
 0.38451953799267413,0.7051855693463672,0.6955699928870012,
 0.34063760896232886,0.15561274556534443,0.7108949154692343,
 0.8864577510630457,0.40077096973427895,0.017888334141191398,
 0.368940467777493,0.8840264749901667,0.8106863684538876,0.9266031712213162,
 0.9856807749502481,0.8863939534609467,0.2051417016878243,0.15594429942735288,
 0.6897978198232273,0.09994359375660744,0.8501967198209099,0.6827014658605779,
 0.16451554147101544,0.5599109692616884,0.6779236779519832,0.2605044652387547,
 0.25301396396956266,0.47892722117712894,0.518961184359795,0.1749847239465594,
 0.7511522498258487,0.14019444909389722,0.5264550673125311,0.5765166643264662,
 0.25006794948759714,0.9926518318915756,0.6749051883803945,
 0.23058210370865828,0.7069223533535899,0.6251938405503537,0.5879546116539618,
 0.2861224920169396,0.16787498125192657,0.6968500254807501,0.5921116123700018,
 0.9579487043057044,0.49742280327343646,0.5197171768111615,0.9862186270590579,
 0.619578028168959,0.6923599474595021,0.7234547721957224,0.06014101992346754,
 0.9858720872933209,0.5920214902331248,0.11192437822811796,0.9590723457222009,
 0.2538635269871037,0.1463751546353419,0.20541559288299194,0.7118468073761514,
 0.71786823797408,0.14524601566242057,0.21777757621788096,0.793628561372207,
 0.4130761292200237,0.41061495423858385,0.11615250892900097,
 0.16001209783356607,0.9728457605804134,0.16972115733952176,
 0.2793256530837379,0.22248934596243308,0.19452754909664705,0.532622791287103,
 0.9252215378790837,0.8609942183224236,0.840154254998957,0.7772922063669756,
 0.9212528963967378,0.3980789907997304,0.49495508283993306,0.7619222669962376,
 0.844274233443883,0.28221478013292167,0.8707787856300218,0.1277693954704302,
 0.47796826974223017,0.3191539938688003,0.4309339324488237,0.7260353258296423,
 0.29864901086802387,0.9350738646096002,0.2722484610821212,0.5450323977029916,
 0.653959241796527,0.3736580094349712,0.6864749901820135,0.5575040134292157,
 0.666112435716163,0.16146888912067037,0.8808219446643418,0.2897796877702534,
 0.41855649245288296,0.10412588325916727,0.6121740222032612,
 0.8881586084590003,0.2535161655344119,0.6536553991993383,0.5867060388567117,
 0.035089156617561845,0.8871529459595973,0.015910110358726137,
 0.7762956383135234,0.6172066076689969,0.511590884288999,0.9286775775746896,
 0.35078536258696025,0.47488208681995825,0.49062285313523324,
 0.5443516940222782,0.5090039920302059,0.3190292513552059,0.25548173683001907,
 0.2416206855908144,0.5334005208885884,0.8820854673637899,0.09090079615141611,
 0.9954771023005672,0.06088712754139869,0.6285520033222438,0.2004362356954139,
 0.7062659051186073,0.7624260545465341,0.26380380238997425,0.7155107423887065,
 0.48950693740673146,0.4392093257844458,0.19069285055269225,
 0.6010529183149864,0.26994693090057553,0.3112578676903094,0.9601163577217977,
 0.06697412682849957,0.06289553275612081,0.09311960534730435,
 0.8519875941871597,0.2858930497780747,0.8416531033193321,0.5439877384065446,
 0.1554246642035123,0.09632333217913303,0.1410899882711909,0.6291135414848715,
 0.9906615946736591,0.2389122368766291,0.04878887723445946,0.752189542477784,
 0.9458971450924444,0.97341061128347,0.4144418475041838,0.9288198651051754,
 0.7133416800415205,0.21043619389027868,0.8758706892925248,0.8958398134573233,
 0.25562008776999046,0.18496740128590394,0.17774304896240722,
 0.3140478373471366,0.9172299442022724,0.12381484084876027,0.2933764280210015,
 0.9723608987280871,0.06536357460110698,0.9780047543199865,0.4031251411235406,
 0.6604626899898152,0.107775783087662,0.6767150382713792,0.07367637860681964,
 0.8290625977680298,0.1974939295220366,0.700174240204511,0.20681339326086823,
 0.47525527059846984,0.0594100096028326,0.08712420169366175,
 0.9377999312694597,0.9421067436593096,0.5394260588468901,0.7262708800780242,
 0.7072976919044788,0.8587650855647483,0.6003414539209329,0.5306062688877174,
 0.5769844809023283,0.5764943839127227,0.8442372749453518,0.3461057053705878,
 0.3855309659001662,0.6502877056156623,0.5947904973247881,0.11483543063789459,
 0.5305215983300196,0.986738391022421,0.0712683505268279,0.42377413612839143,
 0.5561461801418818,0.1430249628426442,0.6642398095480024,0.1786623975637298,
 0.3232950978471322,0.04651203699195017,0.24396612619040903,
 0.06313300854653525,0.37947476789279067,0.7848254619521945,
 0.3596474996803818,0.46191129664587993,0.4342102836649665,0.4193697828076345,
 0.369778462683649,0.09305043753992748,0.4371716463284606,0.481047275954259,
 0.761400987564742,0.9878886711624534,0.1783002538186288,0.8428146574567663,
 0.11779155633168037,6.1593467711063e-4,0.1814159711222153,0.5755368194851449,
 0.8511823653456007,0.7056057126588063,0.045901231609262516,
 0.1219654206688564,0.907434594981961,0.911580688463956,0.03931471960901356,
 0.3031145159686225,0.036345037642941724,0.4342133206368939,
 0.8932011267340714,0.27793230103941524,0.2325766961999557,0.2006004444548194,
 0.7807716617774382,0.6311464038407207,0.6884254106398024,0.2671848073064257,
 0.7489248157727987,0.5836733069723508,0.44392338970836676,0.3018074197757159,
 0.8287929857960986,0.2053452228084356,0.5637877368534816,0.11194190361169742,
 0.34969592537725747,0.8573452084655474,0.14570116699486968,
 0.6516099861961376,0.06953409527279397,0.48845033258323733,
 0.9596996003181866,0.06522787596567314,0.6951699513265353,
 0.11031474531274799,0.12403366590629805,0.5645172407391982,
 0.10938355370449004,0.5684806056869051,0.8257368069152097,
 0.18357962378016657,0.6590230361129695,0.2620935557221229,
 0.40897580524409505,0.39427297516156856,0.72834631028197,0.5478540194923752,
 0.02748541979212782,0.7334840300471999,0.9878679990123741,0.5476559906853923,
 0.0034806220429528523,0.42646622580484583,0.17523206718770112,
 0.07891547981183178,0.4478419799811968,0.28913983375832586,
 0.5299842386914154,0.830314451853294,0.25136920848472033,0.3555298682864222,
 0.4309605810678476,0.2753481724645893,0.13354561337692694,
 0.11326415191734951,0.378142645238405,0.8218719613984025,0.5639928067640616,
 0.22780767374266175,0.4867034366690185,0.12679721130931854,
 0.9894277738969417,0.13410274141470835,0.4119689350627561,0.5880804413050958,
 0.9929772472502463,0.16803710758051582,0.5875989744187583,0.8598190482123538,
 0.14820599814823232,0.22118823510665986,0.40551705363698154,
 0.17552523584879254,0.8492464027335304,0.07618669113140997,
 0.6676966690837312,0.3824049587085565,0.3251926612156053,0.3628411754414911,
 0.262483937397782,0.056544331680114635,0.12377032519513431,0.402938161927632,
 0.9244262055248473,0.05744984258085006,0.5653519777847875,
 0.32825923055347195,0.4197602294036754,0.3898268182876137,0.4323149625950078,
 0.03797337098573017,0.477540388976444,0.5168948196158459,0.9093929806637439,
 0.8168683733883639,0.4967972241273385,0.2955529647377497,0.821444829546423,
 0.13325447519314415,0.9573224014396942,0.30565935080338125,
 0.3151616705980006,0.806844964587924,0.7509764600311525,0.5064418006914385,
 0.5555421093892661,0.04597228998062297,0.4923818553185263,0.5863246946108766,
 0.4874262832575338,0.8959877527689162,0.4780868444816786,0.5192417720062462,
 0.6910632977524859,0.09931280982199397,0.11085404920057351,0.290859727564757,
 0.7572641322146513,0.34535525029217495,0.012412160455807841,
 0.0625041630057761,0.08820870222577615,0.8501730505839747,0.5076653064461729,
 0.27390466977574923,0.9062810679450936,0.4520207038932511,0.7132067485712407,
 0.4229189368117724,0.6464370732951663,0.3749036769963908,0.5474458267131684,
 0.03667680597539025,0.7293986936773181,0.9596899400108061,0.9279888717354458,
 0.08739051159201872,0.6190083323478133,0.2522660942670991,0.7041362939886886,
 0.8887110864044304,0.5098868217864216,0.944690483421849,0.644871836373553,
 0.8358840114932111,0.04970165754431399,0.3253122341297945,
 0.21492155408903613,0.3657206253838934,0.5057749657722894,
 0.13294343343377846,0.6109623916179667,0.2306030508307596,
 0.07408593533162289,0.8390089767770407,0.5273957257291169,0.268940791172958,
 0.15285627293159476,0.4826950821641989,0.02759889496012613,
 0.3130450850935612,0.047573299036910766,0.27162740374905936,
 0.9181638324571072,0.1911384919137582,0.22319127123555405,0.3519959984340283,
 0.11984374229568562,0.35536481555514576,0.53981715311669,
 0.019324343734241523,0.8505508444260693,0.5275142442270023,
 0.3758762152874603,0.2739592814992893,0.26717689630205954,0.746065166769132,
 0.76454042609153,0.989531042049269,0.931534290122606,0.992161762542617,
 0.2951225042862965,0.3937589375822603,0.09716315033539191,0.6960701713721843,
 0.07676659263106589,0.155217462444466,0.0911202872263771,0.7937430825486044,
 0.9041179321499662,0.6362843277149952,0.4101740958191529,0.5274431465370814,
 0.2663009456748503,0.9890751592954099,0.3828821352134957,0.6326434604115124,
 0.04383555003937689,0.9857812224231631,0.8651927131891026,0.9847762975679037,
 0.6661817602187978,0.04636077431793362,0.13240477114520244,
 0.46759272598810664,0.19915709502108203,0.2721681468755117,
 0.16960515508223584,0.42105719087669435,0.39215837133996423,
 0.27945673648481484,0.04552780635022713,0.1612841918006127,
 0.40705242921881024,0.38486006708139375,0.29249548911364176,
 0.4586910577394704,0.9732485973715427,0.5239869629985638,0.8905108513662195,
 0.17996575885231536,0.3077164581055618,0.3743703973840017,0.4964173539238247,
 0.934852044214203,0.8936257099830938,0.07955802684005331,0.533204508563429,
 0.7388612312784661,0.28069481026212917,0.4777825309303739,
 0.21097512458553902,0.27657475317607716,0.4720901202342309,
 0.1322554158333047,0.493086354125404,0.6069961677851022,0.7939285484170184,
 0.6309464407556281,0.06105223546497929,0.20661646702086112,
 0.40191369758094053,0.20672513361576916,0.7502241529757016,
 0.5457233938304592,0.33240091960698603,0.01641948371305657,
 0.2202248708507053,0.04429132394765822,0.4369220236317074,0.7841293228960676,
 0.572416317340795,0.2157457828342474,0.641265369209971,0.028009259859848146,
 0.9208443076367117,0.6827025183473453,0.4265769919547694,0.9702805765841358,
 0.9023947958817119,0.003910839716535497,0.9441323461051937,
 0.12358354426079388,0.43176353582979077,0.375341380164275,
 0.49102379835336396,0.2625853284305891,0.6552763463886455,0.0947698891216312,
 0.7692270915660191,0.12253751528260004,0.011676405333371997,
 0.06051375578697393,0.06981639641729076,0.4232759953931575,0.719822689371578,
 0.5582977016206773,0.2595094057579499,0.8532717599233733,0.28846542333964353,
 0.39128468666781924,0.3984528988626923,0.8475941124310574,0.8533777750008715,
 0.8323940259821457,0.49443705732422627,0.3466045456257365,0.3347695433601672,
 0.40493647243555186,0.261876410568445,0.3482469458369597,0.9568802985243288,
 0.11762324491238774,0.5882631353516453,0.22330693159109039,
 0.8899072031056796,0.4348256365324644,0.8775221536413897,0.9005239864850764,
 0.6356795011937721,0.6546913925245336,0.8039347199971656,0.8234995567137258,
 0.6794064096744943,0.7217548262760202,0.7262401506735618,0.9235008672965445,
 0.6880169670838203,0.9021560888018033,0.8930341094133692,0.797873955548614,
 0.39075771279566396,0.8343412434144967,0.0788965844149997,
 0.04443307290455634,0.7629312879068662,0.8614358638058257,0.4048240964805302,
 0.4032078986192549,0.3334123014025433,0.7513433068630577,0.6194132220002283,
 0.6289101138157717,0.46111397460931747,0.182048296550554,0.19006821490845338,
 0.5885102326576396,0.5243171621809859,0.192200095410332,0.4062748076113958,
 0.967233131371216,0.6629558389278589,0.025734176244464724,0.5765573896902715,
 0.011761755616615521,0.36125695860274165,0.910071563479738,
 0.34007424521498897,0.3180902087821753,0.0910368258989741,0.8821241326790209,
 0.317254199249989,0.3218139969821412,0.42061367055574317,0.43650988847114347,
 0.20671306781799614,0.6350309847180886,0.17109305628790616,
 0.042992009797373054,0.47167353225586606,0.6697529627423042,
 0.5211506117498381,0.18802709971751175,0.38748528847346286,
 0.25504730534815223,0.2446259969811977,0.6630924239041867,
 0.00824621227547262,0.018266649898708742,0.3008170073770773,
 0.07545899575129145,0.39115106436552893,0.38305129173445407,
 0.3561579456326689,0.7320042131082724,0.033096067658218487,
 0.6662725140812394,0.2512856098275945,0.25076154680797225,0.9376270302484802,
 0.03757352438124206,0.08788327726877809,0.5772511117488373,
 0.08278020096000316,0.5896917365790655,0.8094406345901065,0.3585068413977285,
 0.9150567352871618,0.17122956160048997,0.16366855915271816,
 0.15796479030269517,0.03495085194763459,0.3312522371311627,
 0.6664884992340884,0.13145130285648665,0.7217790184719233,0.6171545023197968,
 0.00819989411818034,0.8019876349972762,0.6591944614488952,0.4626573257799975,
 0.8603045682748293,0.38645020660229923,0.15628459694545438,
 0.9754290308043863,0.2066946235991396,0.6266770017904073,0.4004076100858158,
 0.1779109390568726,0.9034601003842678,0.6363069878803208,0.2833890414981399,
 0.0987422362121978,0.04583112698948566,0.8931089513978349,
 4.401168168338687e-4,0.43662901981007485,0.8103148292086211,
 0.6570730565351461,0.9449013251315908,0.31646963547717355,0.8424336558203053,
 0.05693639535648409,0.9452233861616004,0.2365782875833763,0.9721342117393683,
 0.033316183247975806,0.5111767882409446,0.23267498833850164,
 0.18113161177273374,0.17367082875742668,0.10686972208411172,
 0.23656217390967815,0.7427331410559299,0.7469204904343404,
 0.43462755060533975,0.432172818906277,0.6976474209534976,0.7786114099229621,
 0.40016726978719286,0.8549232300082551,0.9594442749819202,0.4236184059973558,
 0.26994595675966915,0.9017917723473832,0.19492657568382932,
 0.4556266512035487,0.5457820888480264,0.2957859533333327,0.47129819755646707,
 0.6652400253465761,0.9356163445207011,0.4466776974935983,0.41523341847153794,
 0.6711838935590517,0.6189444061521067,0.6881564975952695,0.7675712825479534,
 0.6359509677004715,0.4963761630944785,0.8150663861879698,0.7929472959359816,
 0.5394163404374687,0.3180506225546694,0.5034588670414489,0.8124270999528569,
 0.9480902362925052,0.9089420856106457,0.4536148309685101,0.3464023808124229,
 0.1967697625721423,0.9285626170825826,0.042458624832612646,
 0.7618950567061112,0.6241913628245584,0.7958019605370064,0.8084288161695513,
 0.4511274205207909,0.1102277713894152,0.8731904437341581,0.5866185170650077,
 0.7108541864904594,0.2095783711407816,0.22536352497419687,
 0.36995729815266076,0.9369818772283423,0.5723200548643762,
 0.02193523326018021,0.4060185761151678,0.439210311879492,0.8559997513907797,
 0.006471128375764579,0.4620710548050354,0.7262059782964468,
 0.2820299802654914,0.7713568175928024,0.5630333111552822,0.7687987949058213,
 0.7066519426642424,0.41491866364839547,0.46145369109147194,
 0.17866481593901462,0.8630852970233428,0.787275702068464,0.9930319698310152,
 0.9293166821299027,0.2654403959200792,0.9671923527212001,0.6449717797527698,
 0.39381543766067795,0.513712086134722,0.9098682147350265,0.46153804332436943,
 0.9514248582519087,0.9803574280844496,0.08276230487430458,0.7410191037774759,
 0.9969882474423182,0.709152426748256,0.7092854923759809,0.08729564404823731,
 0.8731586733521567,0.041438497775801664,0.8812535572192516,
 0.9894526136821222,0.47270350915966264,0.4984774043313308,0.5615811922712215,
 0.7864631175112073,0.12903135350366235,0.7953385372882709,0.416168895362377,
 0.6823756065025767,0.5513351454612967,0.856954005165647,0.49612883647446115,
 0.12206090575955952,0.5192234988499299,0.3800371474824199,0.810246121812269,
 0.7990434722860675,0.4702402440381894,0.41421993024964976,0.2608348091428865,
 0.6021204780961172,0.07329200031977201,0.22519554405873154,
 0.6774233279168254,0.8080196627105108,0.6942045171615447,0.03513870748707526,
 0.9115296582027245,0.9050404073201478,0.9306114844859886,0.9935342671585207,
 0.7806246499075856,0.967931687926795,0.7711990844474049,0.5803710421079415,
 0.2128489130707254,0.4206547892829795,0.8870886913729172,0.1200164376932259,
 0.3702612926732055,0.7342952196382345,0.29370226421364753,0.9262499315924317,
 0.8815385050400105,0.9288465992750253,0.8767225786884438,0.8664687809823426,
 0.4107184675686948,0.8244564126640576,0.5438261959911683,
 0.029236389643823912,0.45875261681847634,0.6577955759449797,
 0.30165367650993413,0.8224688130863507,0.08916597926534287,
 0.5070450772969919,0.47137894716299034,0.4172074867786898,0.5535564110013936,
 0.8191393486652502,0.4601186536547057,0.6963926680398613,0.36136130422606816,
 0.6307789879117397,0.553455716120064,0.6408760157465992,0.5923924332337798,
 0.5248503929045095,0.03579289467956148,0.8659738768521648,
 0.18914152419160435,0.05091766210722293,0.3255535023138406,
 0.43462024009603273,0.6070858244696757,0.44481212452473184,
 0.35686856936720823,0.8662389570922515,0.5961661744575512,
 0.39407043835068967,0.9345731457784481,0.9751225925178479,0.5697129611552514,
 0.9954729260214438,0.5299410464842471,0.5316039375716982,0.07651620113395308,
 0.2706120253893135,0.03190834978530099,0.059088827696965995,
 0.2997271936210356,0.3251259540374889,0.5073248952568059,0.4562171594612394,
 0.030550832572707964,0.42324227152381777,0.9721017685190761,
 0.4072705157874238,0.6163706844942356,0.6163498188800668,0.26185376751030864,
 0.8437508629473509,0.21444299343573803,0.8395048679616944,0.819838161807833,
 0.260270666064353,0.5744461135672918,0.9347679856239246,0.4224348725072481,
 0.44629721475575,0.45675545910675286,0.9312897794410182,0.3771955186728231,
 0.6670513650951302,0.40808906409746903,0.7432560207943262,
 0.13778846101886322,0.9890228921108583,0.8396416745967069,
 0.005476105827068034,0.9369559277686013,0.7819749503372118,
 0.9749488945144876,0.12188255105204471,0.5828047424503173,0.5103351197930912,
 0.5662111723445855,0.9183350120987979,0.8646440687959931,0.5389624227284459,
 0.9187180570771766,0.639456006816657,0.9836233841082063,0.1362457021708232,
 0.30121384645810234,0.6786946360327166,0.8545542125650648,0.504583818634873,
 0.45747449101767346,0.2897403825595153,0.7654847453370198,0.8227304246082668,
 0.3797111006329076,0.24178759703533337,0.01862573234194964,
 0.7952613530839732,0.09606312455515853,0.22499676775966115,
 0.3230164055774136,0.04887187839264884,0.6652559650183328,
 0.03790037538573743,0.9608959202771714,0.099157905029295,0.9554625150192685,
 0.7465961055902306,0.8168542259026891,0.975528482436346,0.208324611664774,
 0.43611632817292134,0.5390198022823515,0.06326611439419216,
 0.6687968733917182,0.43654074518762453,0.20784971789176387,
 0.41511319024036786,0.6475426429373667,0.7888253967336075,0.7193165666988196,
 0.9499058343379265,0.6572689780802345,0.8814066202834505,
 0.025463688735384737,0.3967156988221041,0.34486101147006454,
 0.5842371768502684,0.48331826389497734,0.043173833517588744,
 0.7930516777655563,0.3169467689767165,0.3317916440021096,0.9368647682171235,
 0.1930921390091399,0.8212046731902763,0.24096627175115803,0.6184022518841192,
 0.13516499486939937,0.35118365750366265,0.38257626110949805,
 0.9500261459063589,0.9315636449804383,0.9559660060926695,0.8044000948829552,
 0.28478707830555394,0.19083297219633566,0.9706513098437227,
 0.6236261127742845,0.24040244774474784,0.8696388758672348,0.7623400663322358,
 0.42688300680873537,0.8464410592857846,0.869353678547096,0.8000177402422356,
 0.3001783326704731,0.14303439947774166,0.7644973389786305,
 0.21434430901275836,0.6219386712692341,0.0471651944904673,0.9430432897008609,
 0.33482505452520583,0.4153130721776477,0.40918198000584693,
 0.7204042403585138,0.6209558989394873,0.16465368432744576,0.4913338194552782,
 0.48268238509033645,0.8497853047867735,0.6745043889820344,0.2639916837885642,
 0.5682877340440304,0.4873259648170616,0.8103896400657388,0.8818763925724041,
 0.4720338604122748,0.454282369653092,0.5392073695344489,0.46329634306497247,
 0.8555933620097289,0.13500947409593855,0.506328451302416,0.47181247403087345,
 0.7465627129583183,0.1749167660275438,0.8575306647525262,0.07183866278791928,
 0.702545803103709,0.055010282988244064,0.2540718155363556,
 0.24424180170906973,0.4247087622426313,0.09706746120777188,
 0.7082667961091171,0.8144577923442481,0.14570658305329165,0.6167840387893868,
 0.49752327947825437,0.8825552895158569,0.4728972352635957,0.6092797027228195,
 0.9185998658000624,0.8121277515222155,0.7175366726492478,0.8138105179681199,
 0.20290689143533036,0.5426055773163306,0.13000494401158935,
 0.6671682075626832,0.9336997541570218,0.8979567234979283,0.12568328057795086,
 0.27133349468078527,0.17115202279601505,0.1938037872776659,
 0.8273368621959015,0.646991195334292,0.3039644609181539,0.0895391511473026,
 0.630184639784063,0.8268688859759635,0.17251843696725744,0.11113714676643849,
 0.7707245303680363,0.9069904012384961,0.2877634664547807,0.24608040668211228,
 0.0797761349421916,0.16256047110473082,0.8401672170495968,0.878071688338288,
 0.8671676119876136,0.6584365812866994,0.24646141909229335,0.5993908775321084,
 0.22695831308424863,0.013994630138294384,0.8706179186966422,
 0.37208779319492646,0.6776328139784457,0.6492653496777268,0.4264589138119198,
 0.33965328137081474,0.5159976371399866,0.5792454825833137,
 0.11172341441578837,0.21262967456381598,0.24119934871509896,
 0.619445471991415,0.7178273810193359,0.8378835737452147,0.11018662838893345,
 0.4980197341983733,0.42957707630391595,0.7860090041496479,0.7913131520347496,
 0.766530710942435,0.4363119626153209,0.865687011691862,0.05407938058876394,
 0.1626761292469494,0.6803048266709122,0.8286145269116313,0.09544342221232116,
 0.11005725485520634,0.5156283605148602,0.8843296751874453,0.8484617740754888,
 0.020708261906200454,0.40909482731143487,0.7168101444310877,
 0.02267585074787648,0.6405197197138774,0.6575852598361412,
 0.37407152709310054,0.8373915364160315,0.6790478849326709,0.715002919413166,
 0.7146433324158511,0.6123554434317009,0.849797094648827,0.7587521453490411,
 0.23759620178844787,0.8508818962530315,0.9623547799090009,0.3004401221957913,
 0.30907343672912413,0.2057653375262133,0.6354840070714706,
 0.36750488440609363,0.15200116576844902,0.7946190010695304,
 0.3226189229494053,0.23388998420423834,0.21484799371437635,
 0.16205605778508647,0.8149592239968643,0.8326131420636198,
 0.07844179661106787,0.2912993823241852,0.6403561010110287,
 0.24091951990521132,0.3541474046074009,0.6579707669703235,0.5360292660440626,
 0.9164993531894338,0.9029405221541432,0.2979490880991831,0.9888891822297047,
 0.21342704325937367,0.2042563477635433,0.21876353571427432,
 0.3995318004129996,0.749383877249874,0.3518683579313554,0.7242561275069904,
 0.3907124322067186,0.7659330724337944,0.9291683485247823,0.171909068825004,
 0.2775322061512222,0.5265374238061258,0.1900607987371299,0.6502063575220088,
 0.7959243804420246,0.5989669803876375,0.6675998373900266,0.6939817092104262,
 0.17425372757612823,0.9957366125417381,0.2838658121951809,0.7237536150701435,
 0.3785243515428943,0.2370303789235424,0.5617213721114878,0.5296972857298153,
 0.22532294039158274,0.34499828907863583,0.5657165399169527,
 0.1564876414080042,0.3740614056469662,0.21635041211268624,0.4595442198607702,
 0.8581099322856923,0.5298082018256058,0.2551635716453412,0.26666400814602287,
 0.897877898486722,0.6075832308257157,0.17265657653143152,0.26962968900883966,
 0.8910411018108194,0.40922435578794203,0.8643317940571671,
 0.49196218834116534,0.4248349887730123,0.536165404002863,0.40962310941837854,
 0.24525837989014931,0.7513353397178173,0.15601221188188819,
 0.5998488787523425,0.8035096577340688,0.20419798914586662,
 0.027449911777269076,0.6478100219110546,0.9989853929021231,
 0.26028309187912946,0.29989145602568246,0.3716517150726679,
 0.16420729882999519,0.7246406307757995,0.3206692337509618,
 0.07466552907504576,0.07695080310029268,0.7623076398462805,
 0.7422632903887826,0.5029405334930492,0.6501385494281761,0.33847319320694214,
 0.6380311188769434,0.6245060485547391,0.04002371383382286,0.932634309199202,
 0.20563737419695327,0.16950699403224823,0.329770517884867,0.9731406637627062,
 0.09607788269473727,0.8040645968905702,0.6366925798232682,0.7983154389288705,
 0.5313344961708447,0.013371920601126797,0.2933066056611954,
 0.8087988779483402,0.6956362530555519,0.6321916520429343,0.0819603816851292,
 0.45029602074455854,0.43023696214576357,0.04144283312382391,
 0.7186609664179713,0.6636434975248211,0.21936926112647115,0.8756352941820013,
 0.5450634570877609,0.21386822546328854,0.07227314117296635,
 0.9875761150289231,0.8209631041060798,0.7915552495527607,0.36867064437167796,
 0.6596834863406276,0.06438584653664137,0.8422712512551003,0.5702507935975422,
 0.5748936319994562,0.6432126249557698,0.6143474335325578,0.6889892616044427,
 0.49681827215244856,0.68254179411133,0.16505012841624922,0.21667331482479613,
 0.38600952652096987,0.6499206597213203,0.3670216531580034,0.599130646893579,
 0.030605835449475594,0.3816228504167348,0.6673944200676095,
 0.8375693642543566,0.42641903431000283,0.6761389517545884,0.7405346730084326,
 0.9443713424860919,0.9737487981608073,0.37410084939011634,0.6158719382801213,
 0.5226449484508255,0.28834609175685944,0.4960993640082001,0.6369192345317694,
 0.921509968102409,0.0459228034328254,0.4749823710403075,0.015017872679274347,
 0.3263749985965503,0.6062558719179965,0.35590558915648174,0.7228408673087005,
 0.8218120646735019,0.3372625091674204,0.7902783613319204,0.7474788918974624,
 0.3842316517004468,0.07129052072145692,0.24146376806289083,
 0.13322054977706932,0.19957071460757536,0.23739309277291998,
 0.4717523938166496,0.04606859926610196,0.33171936195503515,
 0.05905203626753852,0.42749857259185553,0.6856335651256386,
 0.0869751468947132,0.3044413822493943,0.9681212213171173,0.9424742785600769,
 0.905765819911727,0.6145231306629785,0.1766756180395428,0.10935464143009688,
 0.47300558665280856,0.7593323304445875,0.8910465544804903,0.2245879583828717,
 0.673671956472685,0.573345518960538,0.888389540361799,0.06373600252495293,
 0.9501238888705027,0.10909578913194729,0.9396542394619914,0.8373411357994167,
 0.6156965824354711,0.13081945232042347,0.04601219919659416,
 0.42355878023721605,0.0294574675277397,0.9474793477758335,
 0.44007048861793074,0.6418887557548253,0.3199851688288644,
 0.10550062016843709,0.8419153260586042,0.564883083844991,0.22578827493004272,
 0.9511516538002127,0.9268858415933235,0.7776692075073416,0.2346541116909573,
 0.9945337588390194,0.7232992345473084,0.3083817869519814,0.8636649622175249,
 0.41416530395808604,0.8304651547009113,0.09373872631447411,
 0.12148471346930667,0.35621512488525264,0.7039168422659037,0.523501819281001,
 0.2235924009576684,0.0021335698497997324,0.3121946157035833,
 0.8550809452159858,0.99987226242195,0.8545065857701718,0.3500738216841076,
 0.8822520858721967,0.04291949628160052,0.9055518750236947,0.3909399665757234,
 0.3687915857834977,0.9502644616885323,0.7590916645478835,0.9433212458484136,
 0.7642901409611346,0.6177669613766138,0.49550805684376376,
 0.43484774782563473,0.9845176948988175,0.4802574197324567,
 0.008462576904119112,0.0838518828673781,0.031150562039970575,
 0.2746236782688041,0.3466709905330777,0.5732647993620563,0.7708944436826651,
 0.5710449827323877,0.2521869836334151,0.2568624920299306,0.18914586390526822,
 0.8735814270175398,0.3273563089727902,0.6073607642811445,0.6892566067434397,
 0.3307406111201645,0.5049326400320244,0.33774692036163034,
 0.011197085255471806,0.914738291797902,0.09344006211893219,
 0.5944837539354484,0.3253032352760915,0.42676672002896054,
 0.016416321165110603,0.4445697464715027,0.3542654422723195,
 0.9832335358288672,0.9025855580095838,0.22381430913973177,0.6459648143234096,
 0.6854856769426853,0.9456620007035004,0.6104816310890548,0.09519314949035262,
 0.7688840694931762,0.8718509256167167,0.9340171780360528,0.6293212221776756,
 0.8687142090079262,0.14582048709518058,0.3978405060496095,0.9933234329583654,
 0.1896913438459651,0.08280652788857235,0.8059581020754177,0.8757892831509573,
 0.02358604091167482,0.35902231843183763,0.47313642771903097,
 0.08676345510700623,0.9334711675375686,0.7099518080238227,
 0.17301654938236477,0.6984978473817297,0.058938385044359665,
 0.6001755797879024,0.4025415434764239,0.5082087180450087,0.9577228523935839,
 0.9153576500356396,0.17762970625189523,0.1369481390173677,0.7506827240760273,
 0.840457833629588,0.07946168664669373,0.46536508626503825,
 0.42062994777534846,0.6675145187476734,0.22168737366952596,0.580741109735576,
 0.9582298228437551,0.6169748864721276,0.44003300138309975,
 0.37112798829822735,0.09402992907798047,0.7919277300724024,
 0.8919290917563154,0.14861045327866607,0.7489604973859637,0.997444827657927,
 0.7275959264800431,0.9823061330591878,0.7253937287086675,0.1998375895342488,
 0.7958152649039748,0.5842935416306934,0.03136428980605699,0.5823713733314045,
 0.6770855658365437,0.5130408304068942,0.24030308761976626,0.5318323076473391,
 0.848884828940494,0.5622959554671685,0.961960391375759,0.022601371015399052,
 0.8609799322929197,0.4366164982608811,0.8014045997715612,0.5019320145053516,
 0.2577474933581082,0.17641647320238696,0.27516919087489744,
 0.6756526138742357,0.6030523084157948,0.8151652378323154,0.701972238991274,
 0.4749885636071868,0.38061940462106414,0.1565887215063625,0.6149530307771685,
 0.5206889088601953,0.700835209681899,0.20888695043919525,
 0.030470760634597416,0.0446942282481666,0.8379081172854489,
 0.1658895186099858,0.7701462623517821,0.3210367720433408,0.4855113372673371,
 0.0793002125405452,0.16476230130567993,0.9847881142407402,0.7200109209142326,
 0.057712944929940946,0.10322169475004395,0.7974420338111029,
 0.6847561738642732,0.9677982805826338,0.7785185094065864,0.6694837760827929,
 0.9344987876356053,0.06286032836071132,0.3283791709229845,
 0.05624769722501244,0.7420798156130095,0.33680523946085517,0.253548952034816,
 0.9041024828367161,0.8037852614859915,0.10163645255015907,0.6922472971658077,
 0.18771811424924867,0.9318520591050854,0.5922202578817451,
 0.03795094403701749,0.8036502285933447,0.5419441502282788,0.5697319676189738,
 0.3922703297878135,0.23416909653051743,0.6622204349289978,0.9024553173696342,
 0.3673473396745014,0.8087286996937442,0.0032026077803944952,
 0.31049943748298026,0.7101056776698685,0.653211643582202,0.7015262860605589,
 0.8093994951600176,0.019593300814426784,0.1176904397558668,
 0.36073342093696603,0.33407291369218817,0.061732492182891674,
 0.5709012332084492,0.47396022529287696,0.8691919156994798,0.8118529382802971,
 0.724822827456016,0.39655210989543577,0.6414689852360164,0.17663885264913215,
 0.9559017512396424,0.4238925034609018,0.9460960455681456,0.49337154098572145,
 0.23971186148209245,0.6328753957981773,0.49406476956322687,
 0.07057221408416536,0.04506160003600379,0.3847944736945147,
 0.6063512334440301,0.4067991609401236,0.6670750365395839,0.655113412515294,
 0.5943965007096188,0.5084868205446272,0.3737386378752482,0.09713397420309422,
 0.9001585667986146,0.7346141588630585,0.2546545835956904,0.05626253256570468,
 0.7843923896941293,0.34271951459205674,0.18228493856421957,
 0.19967481012401656,0.2772837695820636,0.23970463521614016,
 0.9606905191617034,0.09422544228124596,0.8099392716388393,0.8876264006249837,
 0.22204600997456247,0.2862269547376467,0.7168978841819932,0.5320208917627736,
 0.008608605529331514,0.823938607782402,0.14584930839401944,0.584651197429016,
 0.9100223030707385,0.4746779273400523,0.19997601177452606,
 0.21528843520468344,0.6426567057486313,0.5478868402429101,0.6075054568345037,
 0.3921270892840085,0.7768432832272923,0.6526414032945733,0.5871816281552121,
 0.25462800166091004,0.650181956727901,0.5395564650984987,0.5019277099403938,
 0.4976389568831925,0.5589194501755518,0.10104964859892773,0.3305305662830298,
 0.14776548295176317,0.331351557928402,0.49853754733885247,0.5297324029632613,
 0.5482530950297996,0.4167423480158964,0.3383243055863284,0.1335932983439595,
 0.6479171157636794,0.5546253645835229,0.785269467995644,0.5739638317126763,
 0.2455784857942067,0.7313178044977802,0.507269365594821,0.7992650278286011,
 0.7628467690742816,0.7674296786130943,0.4677345782758944,0.29830470728566083,
 0.8347551337876125,0.08637906326884326,0.5294791450733273,0.3216948554249782,
 0.3017978398502227,0.5851613648910851,0.9255704939675424,0.5741872202477882,
 0.23179668967156586,0.5122964492538594,0.37347563834197095,
 0.9795866269222159,0.911996359467148,0.4919121090823596,0.44451731667519667,
 0.5248343597668705,0.5089835634956054,0.4077902013199679,0.8853008065173185,
 0.5645647340120521,0.4338459490788873,0.20311149552872676,0.8169623635994775,
 0.5707898732820134,0.4350829457350658,0.02400151336013412,0.3160648993443975,
 0.10909694111342849,0.8286322964092392,0.7754361854650823,
 0.34856934418742724,0.3289075639224367,0.1282421162612558,0.3193776995808175,
 0.5785134537299729,0.01295158102549554,0.23721351759317666,
 0.04572629748829482,0.21364205510071488,0.5807343051999859,
 0.3994542806167275,0.8993186000212485,0.796890151747752,0.34028172327622686,
 0.49274585780470903,0.13031812539930931,0.7043776314816813,
 0.6701065525579815,0.610249620319394,0.12551274231301146,0.5502787370659846,
 0.6468082278650837,0.918368524222398,0.26343619579954825,0.9834278621814335,
 0.05167429552829972,0.20347948954592754,0.5559244670760247,0.829237692898945,
 0.9500089743402506,0.14798608908345756,0.9092530673208559,0.1897198349206518,
 0.7517891153877666,0.5533419045998728,0.5475918913055576,0.5642976424409186,
 0.5881438084757598,0.5299445787145342,0.7043442758666414,0.8849402141948699,
 0.8673234433741328,0.04761041882301931,0.6843859941464544,0.7914501723483862,
 0.6098818493412858,0.3434000623661726,0.36813706931408174,0.2726601011395603,
 0.541220277392878,5.80447224247882e-4,0.7882843440006864,0.24882505127332266,
 0.6841669050821033,0.8163452729609841,0.31606904869980956,0.7079985190985654,
 0.8171707410312913,0.798948402697348,0.11184875803791894,0.4650551528716682,
 0.42376698756420106,0.7191446489485551,0.33721002596048666,
 0.09990365257200073,0.8801515321438617,0.8829987439453547,0.3432106086502029,
 0.34894354909252623,0.3195585186118447,0.5693417915158266,0.1751996282709678,
 0.6037316115782405,0.2875468914737822,0.9669922879696905,0.371022839319466,
 0.19299830293035614,0.2643496980550075,0.25145150664402316,
 0.43943931941664083,0.2099298655267111,0.2103158575479408,
 0.40350180678474334,0.5726977209596739,0.12637568156462808,
 0.8329675403655781,0.26354919714745095,0.45499735257309937,
 0.4933277550705094,0.5351129009448398,0.4771877434807358,0.06841905167698703,
 0.6812084265906142,0.5999794622729703,0.9565306039617855,0.7268458186592863,
 0.9852579357267337,0.5557479544919202,0.34274596819412884,0.9548083079145515,
 0.4188983304580095,0.14485683881131273,0.07152753054538019,
 0.3797481590686801,0.7279670850244403,0.9412883480924378,0.5767486130488501,
 0.45461738263306095,0.6742450921429796,0.5940186077949062,0.5463510297456826,
 0.6167560924956502,0.20845537230963884,0.41137528149089064,
 0.8640056619285952,0.21038140481828482,0.4132451435221485,0.6819749858202423,
 0.23650253131048804,0.1564540040092608,0.006070636167934684,
 0.8978278444907991,0.9995846927309484,0.46281978604543383,0.9524035550369273,
 0.6724996305099215,0.42851113495507354,0.845386458019393,0.43128521201770487,
 0.36882029945465444,0.48014574938367727,0.19170723949981028,
 0.2010187237394898,0.8358360762943591,0.5980445134096977,0.284560261909935,
 0.7821248936089356,0.466636592840883,0.12085996464081761,0.8696084392471244,
 0.13668630793182468,0.3668493049869743,0.09941240178862876,
 0.42321840728192317,0.4168515921876247,0.312778501113829,0.23379303269370633,
 0.474272496464332,0.9620535685040025,0.27027931367758606,0.9946850072326139,
 0.7650451083494871,0.745269856470602,0.7893389018711334,0.6967197919016653,
 0.5192205885086906,0.5047795376149167,0.5255909366686748,0.4757852798241088,
 0.853570711910205,0.42646005109630813,0.5243094584595372,
 0.033319740760483274,0.8602316819444856,0.3172518939831751,
 0.9898299459617541,0.5987353053010834,0.18363003884430262,0.3520715967045456,
 0.5650819331931917,0.5122510179983855,0.9435934095261804,0.6309067078049981,
 0.4005903137858423,0.12098515391200348,0.006143352998413176,
 0.7844451028089738,0.3120974289739662,0.5381361369353448,0.06823818216846589,
 0.2284665081011774,0.8369670901865561,0.1314646325704688,0.8069589057662688,
 0.1140821530457431,0.6504586546478307,0.43409430154110984,
 0.010762337777924902,0.5070425108053782,0.796454668154511,0.7724393032523151,
 0.439908085790217,0.018393670809009666,0.5404376312325295,0.9295848955353653,
 0.4617996967385677,0.9027992737782814,0.11299283675175309,0.4024781106420622,
 0.9732234908597961,0.7125370541650081,0.12395047708833085,0.9889586741259675,
 0.9370501913857088,0.6801367359156373,0.19701337297283772,0.9165308776912579,
 0.317275261850842,0.4430584111758512,0.5193523211112617,0.17943906492862227,
 0.38627905051949263,0.5476439561993437,0.9104763724686369,0.4984013350282177,
 0.905029869225454,0.5506042861146636,0.9271002101799913,0.19200640222059062,
 0.7197480054478407,0.5364044302402364,0.5581968844944869,0.9484373823863493,
 0.5062450333859376,0.7075876645193352,0.4156558820955114,0.5831604674715837,
 0.6003366934377169,0.7903903426145367,0.19691691133212186,0.8303459675524473,
 0.20201900723833177,0.9042355390489862,0.5705725323611757,
 0.057417332382312725,0.20643554455797453,0.45689369358682175,
 0.48516068356470765,0.4343962623016304,0.5475773306428122,0.3639449204017301,
 0.8702539146842156,0.6505395930843161,0.7704370275347016,0.04934817367114497,
 0.5812148668831557,0.8647952434878713,0.6022502485748713,0.62229096968956,
 0.5498928430410911,0.40410106216497055,0.6501286553949356,
 0.054655480627632924,0.8873012607668344,0.30635987953694443,
 0.7836336885641918,0.6512496475785459,0.23543348629931005,0.594222989880951,
 0.7095152850691174,0.0740623808519274,0.21260524696283345,0.7376081329021448,
 0.9220529571553986,0.8422780779598185,0.7165107349763521,0.10432789138021659,
 0.3552131838726502,0.10492684986411827,0.7062529477357322,
 0.026442377501400882,0.08829203686491527,0.5130675569617307,
 0.7262818011688286,0.5343070609502693,0.5022362048659857,0.6103016635193826,
 0.5723686832671838,0.9077948730566944,0.7083989023531323,0.90384943801981,
 0.3352933650555724,0.9248678931485427,0.7188700299953832,0.07169787029701813,
 0.27010025385980285,0.3064079695849472,0.7238660268578614,0.7736227362053116,
 0.4030019942954829,0.3584798516860651,0.7539531771819429,0.06946921548189999,
 0.29915106733131624,0.22836181205565764,0.2579684183440675,
 0.2649867222105241,0.11605768075723821,0.1839358669855502,0.6340776533778361,
 0.906343267355867,0.01434646607469292,0.22535664316927917,
 0.34977157628797784,0.1619854214969555,0.026666385080239885,
 0.49975239430013885,0.295575761634994,0.12914396332949862,
 0.31494649105437345,0.8854867495902603,0.3232508147395533,0.2841000201509494,
 0.6759368121215463,0.08921845091498237,0.11687837103320342,
 0.6338071929239221,0.10200988278172185,0.27210297975003805,
 0.9272617330834582,0.7966994108812977,0.9118305268036984,
 0.023805221112838337,0.6972841418529766,0.6976239329151452,
 0.5020943364990459,0.005426732094751641,0.13486034461270846,
 0.9551945020145727,0.07191705409144711,0.3438907214272954,0.927520799430099,
 0.19097768783263547,0.07192353632170845,0.05484502906180733,
 0.060642277664117894,0.1152652936682027,0.9528671971846142,
 0.7608986228043437,0.13820812246079883,0.866041160843467,0.7124695872856778,
 0.6519653791666961,0.21710309567311503,0.12602914352050143,
 0.5604515992866719,0.07561903803430703,0.8379734704015175,
 0.37061627772116224,0.30854612039303886,0.9754104691624917,
 0.05321434299334371,0.9186819446410819,0.156136487330222,0.5302862172342699,
 0.1209947583982981,0.9188639552657788,0.6633456002851453,0.5563269369987734,
 0.007496953506445636,0.5421709152093636,0.2050772869562798,
 0.6056545956331145,0.25787635428830225,0.9434674919705908,0.3837606275148484,
 0.48582337527945385,0.17401822470692385,0.8476492201894622,
 0.7195336515581762,0.5068958806136152,0.7285832363179456,0.8720951217929723,
 0.5021102526339035,0.14395040038268592,0.4691804878675393,
 0.39859468857036795,0.4865557383378052,0.3325637571344945,0.8727338164744485,
 0.12186491413985834,0.4536811224112617,0.7064621890160063,0.6208448719402782,
 0.7539400287572519,0.8547106995068727,0.10179542273446962,0.8686110656046381,
 0.215050925598271,0.4171044187643064,0.39863494628449214,0.7222371858490486,
 0.7311172737098223,0.3992478749181996,0.8649927636455995,0.4575683422118392,
 0.991064465495434,0.6335452992040617,0.5457892931825625,0.6029186366637102,
 0.7771413973335137,0.5754623716582565,0.9401669403708349,0.29135377578967514,
 0.7414209760419651,0.36708542835738367,0.6439145792863774,0.7680686347552149,
 0.07412019403335546,0.6010012992277189,0.33033632122752965,
 0.07765326411002993,0.18869294957943283,0.30570748148654336,
 0.8111639221340955,0.5174004418460605,0.8029954790089017,0.9042390410732283,
 0.8124871493954821,0.8720167203525269,0.9995004093451356,0.7942777511251472,
 0.8208604882480148,0.26955559066319457,0.7571280309419648,0.9810297922247929,
 0.5275321905658117,0.10414808297537725,0.5285117230135221,0.8861203529382544,
 0.09795711821754272,0.1479025787926801,0.9945093796194658,0.3443848304876832,
 0.29101453212779294,0.3639663420761746,0.5632813690642686,0.2837953246328928,
 0.9144307249380138,0.7516277319733556,0.8507688524811784,0.41396917063779337,
 0.5181001231006758,0.9789564210364687,0.8456893842186656,0.432417039640709,
 0.6429633128129711,0.40074062598148663,0.5579951855594796,0.9379661967771737,
 0.936334014557156,0.7173237819712849,0.5159893476545966,0.1720424669597036,
 0.23616237779872362,0.10049266929478717,0.04903904202407117,
 0.3808574908135842,0.36091920026620294,0.6891496120544427,
 0.03305498454434519,0.07442140759921245,0.19295409454099866,
 0.674303999768047,0.8111750731949564,0.7223132427627696,0.8019272627214653,
 0.5518377138237698,0.7892898117849969,0.9543011354258835,0.863184813906027,
 0.8751427432268826,0.2487623152998395,0.8522836267617371,0.5291574633413505,
 0.8362327458834838,0.21512483320965248,0.9608707696090072,
 0.17381415887103946,0.46384926889756084,0.6089934682064664,
 0.40230938541340155,0.2586226915534353,0.29177685672412945,
 0.8389710110147993,0.768290489789737,0.39739217229655854,0.1864230495152095,
 0.19003683002879013,0.1515937792087264,0.9007448364950785,0.9029134164720066,
 0.9105269807291516,0.5717943900278841,0.6304623314109391,0.3479977063930665,
 0.4902465869553505,0.31558735777266933,0.4981050280051148,
 0.38602402091726673,0.6529918780774138,0.6709496282275526,0.9594450819530366,
 0.8375713003100463,0.3258660499559971,0.7003885127985958,
 0.020663363329053164,0.2851515129486647,0.18832415837466288,
 0.8938836938992212,0.12909248822822983,0.4656168967584948,
 0.34966928303247036,0.2947172169599108,0.8674032122986216,
 0.11006709489695965,0.35498358942426944,0.05384435629376616,
 0.6894554982006214,0.912236242998495,0.2685956584316693,0.3957261625038925,
 0.6912252102584797,0.02091496241390689,0.08679384806071289,
 0.08763866703664902,0.09136716423934343,0.09420693529451785,
 0.29300536764448537,0.0852552403261091,0.8314145780166815,0.6297570346837467,
 0.04550005740244978,0.3952137851510763,0.16769423662902916,
 0.8069948784734468,0.14758422200204646,0.05204070365700175,0.414061868508523,
 0.6070625264566729,0.5623081826372113,0.3648984653231335,0.12264983985812161,
 0.15222006619658823,0.8008098917104312,0.027839164808442507,
 0.5712727877694356,0.8210575932582439,0.02385018471238065,0.3551518435939278,
 0.6362223316348781,0.6577171857502598,0.6825285684853987,0.49888441626570235,
 0.03671957500226908,0.7458613442403133,0.25502314609450194,0.236239563071295,
 0.23994190948092087,0.24443881184698657,0.7337637232955514,
 0.5908942339044457,0.8775302933031194,0.2946569868884865,0.22874163603762387,
 0.37735249640912627,0.7016273077078763,0.95945723056469,0.6454458951681347,
 0.6278123248961989,0.8540769888523179,0.41334968876077216,0.5831008286927141,
 0.4900735962414281,0.7558769319992686,0.866543656886071,0.30067220080167,
 0.9853769300981152,0.33201672211215605,0.7648637533710521,
 0.15859453487339747,0.6194637386613575,0.4297105659009346,
 0.0029642341288235174,0.7538517078852591,0.7868738284649838,
 0.7830623127909215,0.15176264018233976,0.681278713691999,0.2191045382254998,
 0.7016303177617355,0.8843983336763585,0.9743677730135235,0.29154100546904427,
 0.7252214710514199,0.1724152300605043,0.0045041870288510255,
 0.7045272926643712,0.04011440985905457,0.29617948963018326,
 0.38610383698861195,0.5020533778133127,0.23893869442482807,
 0.9958497117876463,0.3181274038373413,0.5056430467216064,0.11963627727576709,
 0.6115982049525877,0.5557190210392422,0.2742519178657724,0.7964307258242588,
 0.3689195281602854,0.4401061059502479,0.950187803036788,0.23509302828128664,
 0.70840763470908,0.03159921896805262,0.5604717487757963,0.05558378201835423,
 0.488556774254442,0.782760713552958,0.2482651971536558,0.20488467607114147,
 0.7433137897497237,0.2007817306578321,0.1763475578352286,0.9987641153893754,
 0.06229257834335178,0.550646175021213,0.3121824673450261,0.46986446053313935,
 0.08337406108554568,0.40548673283000725,0.8348091714350355,
 0.4534579743370194,0.4471408927602861,0.4549254882554825,0.771851284981286,
 0.12343281071530932,0.7564408716844319,0.8574242577358318,0.5152253174063856,
 0.7453144194076377,0.4583497974839904,0.6083362100698272,0.1694120233720522,
 0.14081113501813503,0.9044269472035604,0.7071819081154609,0.511211673665709,
 0.9494942373457136,0.2845387209997836,0.9614188572555195,0.2639271808107183,
 0.6779666939211962,0.09395049250946919,0.745991725106871,0.03121966312063329,
 0.49835508702507625,0.14986344504614157,0.08011797438455925,
 0.4686447260698088,0.7743119506350605,0.8173579824850823,0.5607939194285805,
 0.4135892460795543,0.10645904222905389,0.08131456858931418,
 0.8898727165808183,0.11545165061305585,0.7416568454970678,
 0.0059829640215995905,0.6492433030959648,0.5133405883246784,
 0.003450939015642418,0.5570932669097619,0.6832931039213351,
 0.5062556148222673,0.565713845848663,0.2845146130644275,0.3291108244716785,
 0.010962505338918183,0.46401279996164846,0.6124324841677307,
 0.4319816680473485,0.4688932225740662,0.8950526547320518,0.7055216646927635,
 0.7511556977464148,0.1567879104190968,0.8506587682242621,0.40922261711738295,
 0.26061774217513756,0.2589279630739454,0.8276611473430913,0.6887098946261981,
 0.5251085909629669,0.7842942673721177,0.8805970690055704,0.6422135684892559,
 0.25915690651197554,0.8893165757616618,0.8918627127863339,
 0.005270406037085396,0.9911519724268824,0.4935299410246572,
 0.4366819561909838,0.03360783428234315,0.13496138271199332,
 0.8962672490584859,0.5919355130523279,0.45485063109224977,0.9099924990814726,
 0.633241479652883,0.005728980693318286,0.011699165654438404,
 0.4746874295444532,0.3634857825021508,0.22765717566774746,
 0.28022064297233373,0.2595143397603945,0.6818483170936437,0.3455439972326234,
 0.801192559689083,0.1553128382103568,0.22587719416370144,0.6731759765092549,
 0.6033449247003266,0.20525569694443035,0.7084397845021939,
 0.17358186620449045,0.33204602615150525,0.6334881709047395,
 0.8672127966039244,0.2844917969822667,0.03610270175765573,0.9893096654259126,
 0.3346454926839296,0.94787627247804,0.5551018377022827,0.16861992566355433,
 0.026656809338037046,0.1482687516778055,0.9294159396620782,
 0.19484487481122514,0.288066287449251,0.25648706334505567,0.6012214658379523,
 0.7885506048691427,0.1459862699365514,0.6583847047837142,0.31362548114172206,
 0.26742632658616083,0.6074481184462575,0.41124300787419743,
 0.8496742143404067,0.3844151493691229,0.947193095269692,0.2845117475268363,
 0.08107658566789577,0.19247067597035405,0.39147693982836074,
 0.5323398902529419,0.19619893869311567,0.08229293698511941,
 0.7306422348436464,0.7435953902505209,0.23716386364293285,
 0.016677228102511377,0.19868777868620524,0.16543938308444361,
 0.21095446737543655,0.8865305590120951,0.04631581994676015,
 0.8180439744691048,0.23818641742960323,0.5753319177635132,0.815582007460127,
 0.7715913351444514,0.8476609566802978,0.8337144045981726,0.23555576725210847,
 0.10436277093544155,0.6002720588714541,0.0937499806312827,
 0.30213442783092825,0.7984899515086707,0.04771790517585939,
 0.9298133021555401,0.28066296359971843,0.44801675079014214,
 0.5334638405164794,0.12359439193873223,0.3575961804480461,0.6890416362344194,
 0.39260168507466275,0.4976183442246426,0.5679167329899937,0.8835935083632098,
 0.09171778072414172,0.33823406150109325,0.19418736254981117,
 0.5487326681629752,0.6079316597148293,0.7985067152071053,0.4889501484554186,
 0.8865654929701372,0.7082663041132096,0.6318513091449497,0.1617702308579212,
 0.7333532120071549,0.7476776701967927,0.2787134597635139,0.8289005695734779,
 0.30609581980365663,0.05862891397685921,0.7417993966240914,
 0.0348820251869052,0.01825093072088646,0.5443707310933841,0.294349019736973,
 0.195548810480635,0.19572736417632353,0.46495617860178806,0.7778785393152401,
 0.12932623085499217,0.23857839709759998,0.6272070070254121,
 0.26749401503670156,0.44442487454030666,0.11876003472229968,
 0.7891087743624895,0.049453660697984625,0.5280824913728255,
 0.025857913365632812,0.08250229608916737,0.9190182224108452,
 0.4017023256736658,0.23843373238162113,0.42100652160300056,
 0.3073123072750159,0.9457048249385639,0.9659515837335507,0.08625448762914578,
 0.7163096778125175,0.8477372972416348,0.10994095069623833,
 0.48526663426330274,0.6335473824658744,0.11657002305193553,
 0.15300321890358348,0.5373847440263105,0.2572987710601571,0.6262287242287043,
 0.29784636222506,0.3280438368848607,0.8454766834168512,0.8168998631668754,
 0.13042852525549997,0.4520934060590369,0.3163782345416001,0.1519847009078834,
 0.5560559310595348,0.11782482926058357,0.18201785607837473,
 0.23774800050485656,0.8881227064917642,0.8034006994394634,0.3858254025899668,
 0.5000304031585143,0.233093972811933,0.7885529942241674,0.9168983526942324,
 0.569716163899651,0.7815589432768926,0.8223295106752243,0.42611252973695546,
 0.008582224806573224,0.9424148899060071,0.5984142388827836,
 0.4902812889136644,0.2824606245471648,0.45858817098642857,0.6164243090472747,
 0.871813007988928,0.8147040708489105,0.42374837292119305,0.30992522281736745,
 0.8140184572809462,0.3743836016844352,0.9959069190116221,0.25565174317332995,
 0.4300336507551682,0.6783825862001915,0.8543730444006309,0.16156712846057086,
 0.42305127599106396,0.5632468647016674,0.9182337429659858,
 0.20533555482287857,0.32971371991568,0.8061016647669658,0.3860005683756417,
 0.8987269165938391,0.5970704531515294,0.6586029612992974,0.9136658520168484,
 0.13327327144955547,0.05718387119300472,0.8865471778155641,
 0.10145805335683489,0.7743199788314349,0.6286878062764846,0.9340707267216133,
 0.6633669103891736,0.561523633681702,0.799531482302291,0.5365565259685143,
 0.5536556128071642,0.47200343360460173,0.6661106242343745,0.8971717976815181,
 0.8078820369720228,0.23768280378204,0.7281631180153086,0.5510393210056233,
 0.24801263686030317,0.37304322878091234,0.0067849163735528295,
 0.0818227718299539,0.7596820498028914,0.8263574017353292,0.9968744013676275,
 0.342221326930237,0.6570740587008432,0.3925784712749686,0.278703058523857,
 0.2895814129384213,0.6148620770258715,0.42675026410362393,
 0.05145139875394045,0.9908103047028893,0.013380811816173122,
 0.9048740464277063,0.04338307783617279,0.5185783642386476,0.3931436605795856,
 0.3704645224206352,0.12186004054172817,0.9471442579920458,0.5596603075725521,
 0.9287521571984667,0.1341471947341557,0.5294652237589859,0.7887897297746949,
 0.9462294441246705,0.9831897107219614,0.5300336334117501,0.038181743863963,
 0.5688360846266172,0.6581411106714349,0.3168682232921034,0.9811833317450488,
 0.5675243430470163,0.22141058605166108,3.3002081745969747e-4,
 0.46175032543021755,0.5978335625933612,0.750630497436416,0.6816986009877424,
 0.7757715367223688,0.9120940235606361,0.18566168039820607,0.6139260437038986,
 0.42172593648828016,0.06371130454167662,0.9134603549559345,
 0.9923878742559438,0.3231127643135978,0.0662711108514138,0.609534324024481,
 0.30603965541584865,0.051859218845546784,0.7631977743409332,
 0.8942530649056512,0.8283171969364888,0.974864969429001,0.1927578040701361,
 0.03811111917417931,0.2444615159666521,0.05349118706786826,
 0.9584473006173956,0.3496425668825367,0.4272289227957169,0.9215617706786922,
 0.37570093814390004,0.5339323128659887,0.7809622509604669,
 0.31718473383947354,0.9798738533290202,0.6188492743322661,0.8362484189337596,
 0.7258578890607424,0.5934834538471078,0.46403984801672926,0.6621438459921469,
 0.5140214748724476,0.8581831982516588,0.6811625111427515,0.7155277190395717,
 0.6060198835136523,0.9204126065938579,0.018449428845882743,
 0.2004604484838367,0.15108541785539775,0.20620362661216984,
 0.27698708527194016,0.23440710327370928,0.5535513991860439,
 0.41603802706323767,0.7913315283291158,0.46840656175611595,
 0.025338639364456017,0.2692110441535449,0.2374192658708465,
 0.49700062120576205,0.08511364758568318,0.6461336842913749,
 0.12994793264502935,0.6945273234745588,0.1842349643677127,
 0.38053645926689406,0.3343483229418974,0.20216360293388036,
 0.3477638526478505,0.5299399936825606,0.16835648778735568,0.6167436403508689,
 0.3791548567234917,0.6530931553280045,0.6612273685230363,0.7106331307684278,
 0.9341159654556183,0.6318196674020226,0.6430691256432933,0.4636896392240658,
 0.32195863619618703,0.6955659813321227,0.5627862876405845,0.9640691397300396,
 0.023840373442249208,0.8938395333976195,0.05549848290733694,
 0.8705899504498006,0.8813033505831847,0.5388163366773638,0.673369202525934,
 0.32661283002885366,0.9217128237131336,0.3610664286445372,0.6148324346736935,
 0.1911751077322147,0.8379152956468507,0.832589087252634,0.19212692943753407,
 0.08217199601474956,0.3872249336498681,0.6172695676401936,0.963336509541927,
 0.0511792134115413,0.33196548345005694,0.5984064034724296,0.6814220597482941,
 0.3973234116704917,0.7920944110710465,0.11607935420517046,0.9832594210095063,
 0.45389815916473075,0.8945431674498379,0.96781846715511,0.6777013975409887,
 0.17653114273687598,0.46742489787944375,0.31916786778588024,
 0.8052127697615659,0.4741323867289491,0.18617976496864663,0.1571927745794246,
 0.03742384702509616,0.2907338089218492,0.5804212768213887,0.2061227655600859,
 0.5607743881466117,0.013591090125192551,0.8377380916729559,
 0.6125023963310918,0.43029939440504517,0.9439206953271548,0.2623350680164568,
 0.32554528012302286,0.11045116967235036,0.3309355953982688,0.338189987209916,
 0.9182075849285312,0.6050562243394977,0.2686018989257617,0.5522962046872114,
 0.6752906828951283,0.7192048282633816,0.7482325415353971,0.3434280439645212,
 0.025512425052860332,0.7971678096508774,0.544683822397938,0.4786261980300993,
 0.2839991784002396,0.3526309554443823,0.5384585112799607,
 0.021507568458823623,0.47583873909066277,0.687610005106953,
 0.2961752700573713,0.531433491020892,0.5061249698951203,0.24959736787876596,
 0.7643842601944699,0.4977249891422435,0.7936993465313906,0.7963247887704544,
 0.3514849516148745,0.021243426871458415,0.5048023977725361,
 0.29169763876968036,0.793165963579301,0.4844322049646492,0.8265730806372716,
 0.30758707413052855,0.611689033434806,0.17241584821643863,0.4558734286849959,
 0.8395925516031755,0.7334709574751167,0.5929168610716157,0.5407804208970209,
 0.2708937643908186,0.16967939388137399,0.9777191869723418,0.2475023166465885,
 0.6497469630437446,0.17659704992526093,0.5252697493395644,0.447512700984131,
 0.6973449407925909,0.5044419299654135,0.32540221377932743,
 0.011268416704124373,0.5804344699757532,0.18512633610922324,
 0.5708501753177277,0.37217318555371637,0.5659543780538474,0.5749379915622229,
 0.8647487573815424,0.537704945489749,0.49597084566376104,0.4639518045548212,
 0.8655984724971764,0.8740252428371356,0.6172204632490892,0.9668183117318507,
 0.8677239981930576,0.12156591519204052,0.7678779647114042,
 0.07811585411033173,0.042028277066474495,0.3501316454976079,
 0.9578739490176724,0.14947520923330204,0.24325732479075335,
 0.3916088157925246,0.5538932796504294,0.8403257837420389,
 0.019451359799210866,0.18717419052119644,0.3959116096782789,
 0.13352097872475144,0.1994909870490139,0.9731783467562786,
 0.29214099199523946,0.6159616532745074,0.6993461597327673,
 0.16749431054353192,0.9541270068725164,0.6728769601278075,0.7716391726308514,
 0.57590181850427,0.7806867420421097,0.2204980691394013,0.5514506734114563,
 0.7186651726424841,0.046965914449357715,0.308232961998989,
 0.45066104168483234,0.9683060132284865,0.24591348137600633,0.606116540469904,
 0.17431438855218162,0.14148660541804237,0.21824555528187983,
 0.20835787156109098,0.34211999515837455,0.36580241228258636,
 0.9582742352416082,0.25829730465941103,0.38962713231934887,
 0.04408033641589404,0.45768641900026474,0.6627253393022534,
 0.056750806921637476,0.499154367609593,0.6973823885878963,
 0.03309130406627803,0.017403913333361887,0.07792062212804551,
 0.7550702308136444,0.3874856716102818,0.6272652845548419,0.9849247534480615,
 0.9529891687063312,0.780389634004627,0.5498845697954945,0.9393808818831645,
 8.914335332075307e-4,0.17226266585123606,0.796755423331863,
 0.8865030924096908,0.8211767071445928,0.7143552426988135,0.7669857235035824,
 0.6993755513055392,0.5451571329251972,0.36014327895332665,0.88264774070157,
 0.6710064412855141,0.6011042056378897,0.8591151216183037,0.49690308895312996,
 0.8967199260030448,0.9991840458926903,0.5709483428642892,0.5167562670165382,
 0.9627386417063364,0.03284256832865273,0.7010012538392093,
 0.36741203472777695,0.2039975315225908,0.004191599298257853,
 0.4528992406116029,0.5005727359127559,0.5167268205270997,0.9740892776429718,
 0.5969253811365198,0.7373485745503902,0.49260741014955556,0.6767982978929945,
 0.8140282454969359,0.8124158669496228,0.8567259654725484,0.24543328092219696,
 0.3993063934530119,0.06700486151174667,0.1751506834883637,0.5139212722141048,
 0.6213561002884237,0.7179209785939836,0.6715765302044261,0.8353935456531859,
 0.11829013883154449,0.2407078926105194,0.09055034820190577,
 0.6056439518522891,0.08617501645243952,0.48832083693703754,
 0.6665432731637069,0.9176816812428509,0.33399795023883017,0.875841262141144,
 0.795849171117259,0.3157572711460296,0.5184675356350789,0.5218663363975826,
 0.8005785601625042,0.4525117843519888,0.2888780997112761,0.6230074791661235,
 0.789423893509783,0.5176093548188875,0.7006701637314622,0.10504326083206394,
 0.36392557783498136,0.3466209929387959,0.30414958292771554,
 0.0776472001052183,0.3834285882651052,0.02801475902969064,0.9325456594737778,
 0.9234207006898338,0.35120356860153157,0.6362691047703446,0.5627027128201162,
 0.7339026502044314,0.14016891078163396,0.964759958306463,0.6538233631800222,
 0.05872457577103074,0.6580349015169235,0.6352725323014055,0.748234528743154,
 0.11102480230211398,0.5816805069687181,0.6990626067963724,0.7977932214760789,
 0.4906012564311478,0.31547613055642,0.40173782665441204,0.22655091759666113,
 0.7216897383348071,0.807622505483949,0.8326988340069492,0.527477922173026,
 0.6013748271303063,0.10244432709495321,0.9851164136131111,0.9636221719968768,
 0.8083409752503832,0.359311185070574,0.36549181605464887,0.9401256041771162,
 0.0032607394067007966,0.24377523226627418,0.6872942753844478,
 0.35387425922695903,0.6899377230287717,0.3075889670180145,
 0.10680234667522637,0.3806839212750879,0.5563768291095587,0.5240863136839258,
 0.11240890316865926,0.3176982320619689,0.06387888574692357,
 0.7562027809871755,0.6991523761428876,0.10975785206060862,0.1291214586847239,
 0.7765508255691751,0.49100165004351737,0.6140176299670238,0.1770994261372607,
 0.43431105754359145,0.9197847563906946,0.04588347148788352,
 0.18032841536578104,0.213617156490143,0.13429286845647737,0.1806692482733101,
 0.5393043674441765,0.31917682060931507,0.8098980242819601,
 0.07348821360442159,0.24533742763477628,0.8582000195304977,
 0.0011769577111229168,0.20558335421349083,0.9766174619423946,
 0.7226959213224825,0.6958290172091222,0.7477907837535586,0.03660847243569054,
 0.8635968551478722,0.744413227257689,0.4229292193999713,0.4926761376862192,
 0.19653316112507158,0.32888835295394425,0.2107732063947968,
 0.44391544847222664,0.6995821238058865,0.7057637417107143,0.2889469882211967,
 0.7847861406849626,0.15205717859564505,0.1545408790389583,0.3176588424491983,
 0.3143670388164379,0.35145779063285976,0.5486293412606028,
 0.34017021061508323,0.8811547425094509,0.6131517438942422,0.8846806965045303,
 0.8951520029220216,0.6538209020886925,0.6368295650488955,0.984522861011631,
 0.27008719461058595,0.5499520028585377,0.07241308742372543,
 0.13745178856192408,0.467132074582321,0.8348120295546237,0.5836659128411629,
 0.3749093532750638,0.343147011156431,0.6858833501893818,0.6605111262873845,
 0.6102867862493035,0.06757848851090309,0.02308112730747902,
 0.04822407581345267,0.6891580584360143,0.8653579666724973,0.7946016017814548,
 0.5976678246930714,0.305175142108827,0.4106060428664099,0.836622383489203,
 0.5090636068984243,0.5409751051693982,0.7157474465584044,0.572876184655059,
 0.023842024688190966,0.43875063814418724,0.3265527873436702,
 0.732951245443646,0.9523740588163306,0.9982374624713836,0.6777376225712133,
 0.01611226917293096,0.22763055053736014,0.6112770892080857,
 0.5048058380959421,0.9846116408312744,0.6016694701913704,0.26820601323671656,
 0.1308244503680127,0.36377214681787073,0.038559021792392456,
 0.5581298602131229,0.33493885540197765,0.41123689957849263,0.95590487466399,
 0.5382843937085106,0.18829516360558052,0.9320326969426984,0.8093922945413314,
 0.12470846170702476,0.18251388248596567,0.9751698280008647,
 0.8223453970326918,0.36663086311932946,0.42296439147040577,
 0.8551524590190891,0.38906071939572107,0.7100633872518697,0.329535435098642,
 0.6556288974608083,0.18190450912368727,0.8248331885942779,0.3632886588468438,
 0.40847796643541534,0.061896612307778764,0.06789722317184876,
 0.7071914084308532,0.49573393155154366,0.2413555660441702,0.5877827816001652,
 0.423397430303474,0.5341389127282586,0.7234462916524759,0.3401552476961036,
 0.4679523286224536,0.3807339986282464,0.5739346130442212,0.8226130575769027,
 0.5536342119686304,0.5506198581730362,0.03993661137061366,
 0.013201942589018745,0.49195748056775424,0.020197160813308024,
 0.14105074213150437,0.835661952379134,0.12100093987919736,
 0.25873476472064605,0.9386467422281997,0.9045025750555211,0.5844290992917599,
 0.9989471850766505,0.9452785537425311,0.8317569849911336,0.2317769309182427,
 0.19193466699942263,0.5952799793009538,0.2132816204564355,0.6646153735859694,
 0.3053499597445012,0.8198047990202003,0.38037815575419176,0.1895575420499136,
 0.5524249079663947,0.32841666845189776,0.11597472133350395,
 0.3414351384732135,0.3472145301358228,0.6948801998907084,0.8498511911176495,
 0.32486615688178955,0.11030398229613958,0.48878939812224975,
 0.4114434329846456,0.8421371220566212,0.49643218594192784,0.8310289897482686,
 0.3230868303135088,0.2724520755335079,0.5203436228565763,0.8802252319745132,
 0.0261489775429955,0.5982463123102102,0.7431941942379482,0.578123693207095,
 0.6565653198373528,0.3456712039256531,0.014316968083822434,
 0.8075823151212956,0.3832633984549245,0.8830691676276553,0.8231378122706762,
 0.25493277761222954,0.017477271444876208,0.22727382019583175,
 0.29506164700294557,0.9591572199714046,0.12168030682203368,
 0.4509714621263565,0.25368717187051937,0.051550038683959465,
 0.7623197059514077,0.5726724223919173,0.9940954610956785,0.7047405899813227,
 0.9618586281786399,0.9587127564833053,0.2749974209740713,0.1630056723226203,
 0.5735643100549652,0.2697701140449996,0.5168328362155421,0.7168519620390337,
 0.2564724956604517,0.5626457361094194,0.20380662334675892,
 0.14637523275505115,0.8655281332980383,0.3303942588694443,0.8181189721454817,
 0.22712618811203933,0.8294991037866498,0.4696184689021543,
 0.13108191373006428,0.8865380709429967,0.8055268060805474,0.3254788049856434,
 0.17129961588143794,0.12226031146277594,0.10431386944729948,
 0.24820306677083248,0.2597803196336589,0.15822312533770688,
 0.7530621582268817,0.8545279988179649,0.06840813679122015,0.9584222948193094,
 0.8980633232242794,0.7293945644301962,0.5198802469001959,0.33354192272385874,
 0.7027823780718568,0.5234384978933174,0.2736087682911835,0.571633956623707,
 0.848225067887319,0.05890015972490015,0.6911439356887912,0.31640897957872605,
 0.8213044182566077,0.09474458690354803,0.4353834161416885,0.8024246335371434,
 0.2895844957789978,0.4610249064026959,0.8869332239952168,0.7512829829826693,
 0.8150424586763321,0.45442389460528176,0.758647737170174,0.9157434421018084,
 0.48625999792801133,0.5479677151134315,0.9143169610701553,0.2615339948521549,
 0.6248025887569119,0.12760302460879647,0.833160888034037,0.8413262134526649,
 0.6804914434437961,0.7268323657425099,0.12698268286509962,
 0.15215099237150387,0.5216490554437513,0.5355769011126779,0.9550161457512357,
 0.8235061217708938,0.701051716163465,0.8731543802412763,0.3883364819501818,
 0.880584256848927,0.303082777905638,0.24961002095005935,0.8460637071133329,
 0.505953718041713,0.20515674612697377,0.4665842266786909,0.6309483248237743,
 0.5228210615963463,0.07044216694841143,0.8093182848968818,0.619487513300907,
 0.49167538522509013,0.21612517500607242,0.9256638955468472,
 0.7707806620931372,0.02556039354890549,0.5455376352711969,
 0.17012440774890802,0.4158431400951783,0.837776851703744,0.6847807751049508,
 0.5473226002974633,0.7719549146619783,0.5424710500481859,0.5597880952929326,
 0.37141104338812525,0.8866770923668529,0.6452841353613662,0.1671210478299502,
 0.9740014544950666,0.5512314325616414,0.5909363875205036,0.8834789822509664,
 0.6428107494219341,0.024334514366948334,0.4440184843574497,
 0.5585295019490952,0.6956088870709085,0.5948270955606557,0.8427159577327274,
 0.1414509146709888,0.09436085349768497,0.12932525916101112,
 0.6473491767295317,0.591556223468259,0.09018557369194813,
 0.008033913928956693,0.22081917915697147,0.4971018717766408,
 0.22063437347418768,0.9507247727968473,0.13628723624478734,
 0.9065896990507273,0.8772557901767574,0.10107610326608607,
 0.15102924983389154,0.7930451653787188,0.6144558885666828,0.4533094706751354,
 0.2434318207138988,0.10967264241700958,0.12898281908004638,0.923403062711965,
 0.6932066737359713,0.0640218997043811,0.8904916827619053,0.03894344733307742,
 0.09208213264982956,0.16258833316282106,0.3420398671716083,
 0.5419488407086711,0.8769218407587737,0.5730882329824775,0.1654771092000079,
 0.16938437515386628,0.7436211531153634,0.0896513878070836,0.8294932846215548,
 0.7891780555952941,0.6257683031106135,0.2716834387206014,0.29399913467637395,
 0.675369805888802,0.02018292454004811,0.9217152212555697,0.8248857330342955,
 0.3211458257114007,0.6141024846690856,0.28347895149090707,
 0.17975219642606566,0.1467874045870463,0.8955930630262715,0.6724338601238937,
 0.7464415884885517,0.02672300111763215,0.6459043886651906,0.2639676589248615,
 0.8412087948455723,0.9415170099662918,0.5422394354809742,0.32622947554561954,
 0.5168937941187085,0.7422722997772758,0.8687321019601014,0.07434875427901833,
 0.5585280769738534,0.7101168122398581,0.6648543588450687,0.3647466478375243,
 0.5906916044577643,0.5406146088996089,0.38003741316869677,0.9234100025966688,
 0.6527815194117867,0.9580091700185946,0.3203345698543201,0.06868145423494887,
 0.7305147381881033,0.814832325391812,0.9055055557009014,0.14286047728045748,
 0.1510599770772425,0.18416632701679836,0.995339295318681,
 0.023421187373416585,0.157177159762796,0.3036558332391449,0.8334337175239704,
 0.9872871458393726,0.017064427482833433,0.777273719961509,0.9189379592674719,
 0.2773553821671648,0.434622238285165,0.30987202681304915,0.39652236054171974,
 0.4728608127878612,0.022407940629388445,0.21199165917815443,
 0.5995305214838424,0.5790057233237871,0.611127126152623,0.4135645147871454,
 0.21017413919456795,0.3468403637638704,0.10397733993282221,
 0.21208930153626016,0.23700301555913772,0.29915130064108064,
 0.6540420811759888,0.14998230531318701,0.5258282094865694,0.7297552248111487,
 0.9428554423073504,0.8028724439783972,0.4317586318147697,0.8542390255047607,
 0.7276178155031827,0.35650078863520807,0.38873416458545584,
 0.09911918765762462,0.45524206054123517,0.9273784064916197,
 0.1772266324187085,0.904238626520057,0.7386063303190893,0.9198784155327039,
 0.6712704320065461,0.8380905518783219,0.7677884812284844,0.9459916374815113,
 0.9302201614414446,0.28598807765482637,0.9349115777596776,0.6188248266398255,
 0.8990854347660074,0.5798514336355414,0.7267605742025001,0.7648742974094371,
 0.604393411088047,0.17569815185296977,0.5774641371078658,0.84809695526174,
 0.06722521358851419,0.9692689029876661,0.6566427428575106,0.431901043917554,
 0.9913127278518853,0.09564989786886735,0.3048000368331263,
 0.18551779897981147,0.11499457275334224,0.7344266560315389,
 0.5793154027471067,0.7245613047164059,0.46123536648179664,0.533442064930166,
 0.3006105538731815,0.36726010613981186,0.07264578411768297,
 0.5613675331049816,0.4682421475820372,0.6899690808488316,0.7835949043944992,
 0.42630841826212496,0.8825809013586708,0.7439734703082491,0.4401948836476012,
 0.182342069357187,0.3915084254085406,0.5075961642909143,0.6703132666370671,
 0.21316967399695952,0.14877276217629454,0.5823299007516656,
 0.8983368673005245,0.9402420031668307,0.366864787452566,0.3857493654177625,
 0.9291401869039122,0.34416988997838427,0.129003793746565,0.6557421376672621,
 0.6897528001368777,0.8138320804028016,0.8162661108986946,0.606617349840085,
 0.8221534734447985,0.2798426761066789,0.6863032385650207,0.06077478280206783,
 0.7714478534865441,0.9103935276931115,0.9929003977709678,0.2077881590223768,
 0.9880956848333462,0.6910394740036729,0.4594428613079721,0.02526967513953915,
 0.9878391798478862,0.7885497055138937,0.7090197550287112,0.11811724323121853,
 0.4033781556055337,0.429849593861947,0.7861120972013309,0.4194766712493374,
 0.6829295189215066,0.6474541376711637,0.81985306905173,0.028304249351424948,
 0.05334712406248232,0.7767380272980047,0.7312048326066433,
 0.03867430246538217,0.5917177001442478,0.9489494379069943,0.8790649664042832,
 0.34589593493764614,0.4313523409886264,0.880966886662723,0.08012921923904404,
 0.8915288916828258,0.43610672251973137,0.28371455062315376,
 0.09786710813776811,0.553013885431201,0.7270503129509696,0.3242315206527271,
 0.5635065818095728,0.17141011992168087,0.08538796614305144,0.790805846699838,
 0.028756524459987998,0.3529201176071026,0.32173487187057914,
 0.7030436866072393,0.8704009225676165,0.36576303020233125,0.8022427841954205,
 0.009630216927661417,0.883452423545792,0.7141072176528068,
 0.10838236299353066,0.18897441320866282,0.9033799647443775,
 0.3996214725906766,0.2315699358996529,0.9514530826058416,
 0.043002727515903616,0.3622567699177798,0.7766005834937385,
 0.9955630940365473,0.8244336061962466,0.7443477966640286,0.21427509701242298,
 0.1838142913078855,0.3539760489035334,0.6335722957289149,0.5272830156445036,
 0.9375636303038087,0.4061942745619479,0.47861236084212433,0.2219508090869713,
 0.25728039882497444,0.2946572369161323,0.7888408859398824,0.7979915615036391,
 0.7480948072438889,0.3167582498837791,0.6194456193252751,0.3129254244134677,
 0.9718547422185346,0.19974195307981846,0.6891295300499196,
 0.07365240590551225,0.8974140043222932,0.9791690206187599,0.7600036488186589,
 0.14070243527876358,0.010656381982688368,0.783920418885178,
 0.7629203593864207,0.26671323901865374,0.6804070530072556,0.8137766978345191,
 0.2346502429135947,0.7991317826088598,0.1705526293187124,0.9411072536305317,
 0.8073228959287768,0.9230626803423805,0.09070149389826987,0.5510782426493329,
 0.7832380245750765,0.6881595012041369,0.735066007738056,0.17113771785076337,
 0.2519658475632116,0.743570939972502,0.640193315365219,0.8877926184378904,
 0.470054444899368,0.30392115472136705,0.5826738582706726,0.7203039831334856,
 0.0028678214387314327,0.46597435226459394,0.1030930574036033,
 0.054979601589427274,0.8159805187340519,0.382664728783659,
 0.005945118776402691,0.5067436921809363,0.6945530643531423,0.542609233792151,
 0.6459874151415875,0.5993524801869202,0.23880259091799527,0.8660409341966919,
 0.6616972391333557,0.482309346877266,0.8065639676513716,0.6470632410066013,
 0.8991143891717329,0.27605360076646135,0.5539632417548751,0.3833224747681465,
 0.9862148444343782,0.9528742118052796,0.04526021734980379,0.8975049362371275,
 0.15972014050847105,0.042775041335111896,0.14815514838789123,
 0.599772053373539,0.10234241260913945,0.6827803436235785,0.757144441991745,
 0.33212560710079186,0.8271133975707633,0.3103259115199001,
 0.02626905251273315,0.8385916631641578,0.2526775169443284,0.5098841070831266,
 0.15533816738141237,0.41769658368157836,0.5400181840903839,
 0.24426932628471532,0.8446093928166412,0.5377620050507954,
 0.14351674406299975,0.5920048949673983,0.020384837128086586,
 0.21034317000065317,0.2532030883847852,0.6614370062569234,
 0.07977381567403352,0.6562040582012334,0.7838494644589209,0.6880113002934783,
 0.47664592819608165,0.07147082015474349,0.4689463285015131,
 0.17924571854081595,0.861877309868098,0.6842130983673871,0.01496291918473891,
 0.27581707256445753,0.6841032300583996,0.24849906983780756,
 0.7041867972649893,0.7217619800414246,0.5030817100764091,0.6305696909198428,
 0.7783690813807055,0.47876836898406494,0.24234184725376806,
 0.13405639471016517,0.21677180888017422,0.18596653911961358,
 0.10060683846041263,0.3521140056290196,0.6106255585096743,0.5757375073122625,
 0.7561257936671407,0.6194043010974296,0.3353064302066653,0.05907598390155944,
 0.4265704259534856,0.03628159630470762,0.9209841635810905,0.6317175754322001,
 0.3631103184962403,0.05745489976603513,0.9636128668447211,0.5111819779711884,
 0.38511967146275605,0.7135723408034251,0.556237375372561,0.05486852824626609,
 0.64704479468444,0.8918670905392366,0.9855778825924378,0.18708802125648538,
 0.22060757137019815,0.6161737392208401,0.4453672077043932,0.8592886801849839,
 0.5141463480300047,0.881310380770727,0.8065987541188534,0.9400979493368922,
 0.12312230812336766,0.2836476085244207,0.7609882220193965,0.5944564819181661,
 0.6681208590364766,0.658624461637616,0.19425706666488218,0.08253629200895352,
 0.34650584459869715,0.1153851066914875,0.7432185324162714,0.158021529190991,
 0.011088285547576904,0.9296814538439663,0.46431599568160187,
 0.9937135877372043,0.02555563231915603,0.4960616116496295,0.2192778729061884,
 0.6013067741089886,0.4874383633678858,0.3116370978774031,0.5448413005569368,
 0.8359578664099225,0.8521658206083127,0.7424685405128865,0.679986659667714,
 0.10685790471613954,0.33259856745434424,0.30315457708775817,
 0.572706272348756,0.04528551208227749,0.7141983116610493,0.6746553503379825,
 0.2622296459690602,0.5040438749802394,0.6773538888993118,0.37706178394853374,
 0.5208148838566817,0.8347018611024071,0.2919426456408569,0.23637634050055867,
 0.794681297265926,0.6086846119721439,0.1880888402147317,0.9265710774787288,
 0.437824424747816,0.263383403948276,0.18291206241767055,0.5538542439668227,
 0.1480913568630059,0.21738283720058182,0.9975266913253833,0.9183230444716926,
 0.17766232063435894,0.18578792637773722,0.8024230686697235,
 0.26710249477708636,0.5866123017017051,0.10169728887102991,
 0.6116236607315585,0.5492327178947335,0.5488123944755754,0.22635707061987143,
 0.4630402569689861,0.6736317813172884,0.7440988208408461,0.27365549251138044,
 0.05801183139253108,0.2900022298218806,0.996598596909452,0.415854784759109,
 0.3697191296164103,0.24499943952497882,0.9146003930258713,0.8927104426586527,
 0.47197412062302135,0.2535547446821548,0.1332097141141917,
 0.23503523388078407,0.12462220082459363,0.13229815900961928,
 0.9332607010207398,0.18045357735236944,0.5312239116216388,
 0.35571016906976827,0.2209415788367226,0.7889863763489263,
 0.46040826624052267,0.8392972288575442,0.3249422154290099,0.7494780872866428,
 0.26150946726262747,0.9499185607515926,0.45340611432459377,
 0.7019990760142427,0.3409041881186847,0.9268458836128337,0.23031304746025094,
 0.44765294399321465,0.7858574920378997,0.9538382137169534,0.7391850114076884,
 0.7470981695905731,0.06732739912826358,0.5038077717905146,0.6548195840640999,
 0.49487642648998076,0.03034214445336192,0.6781040240568537,
 0.8146626831623894,0.6037645333043391,0.3293619908571992,0.7585960189027294,
 0.44735506617131227,0.3240518890786216,0.19706862698935645,
 0.3275553724082624,0.6745502435825934,0.042046389324523537,0.714271164706769,
 0.5630723662415037,0.18075418625332207,0.5281654003980172,
 0.41449414125866135,0.8280196275166105,0.77882710699076,0.22018449889424074,
 0.5142459824423977,0.11266470975045628,0.10254009031488143,
 0.11176596772608927,0.03388461946358223,0.4882464786316506,
 0.6666549147506132,0.4580243976498588,0.30509133188832727,0.44052168621122,
 0.8303960164728726,0.9002844318566425,0.24305173374533906,0.3583953360960257,
 0.6817650568514231,0.0835392836964366,0.9206134727744507,0.3456004662413339,
 0.12964094180604258,0.8111793721714888,0.7202142392202455,0.8953941611826058,
 0.08630848818098658,0.08811072293926014,0.1479137715583161,
 0.3650647072998032,0.4326586496846824,0.7200829683650627,0.7837839981276629,
 0.7895033184658192,0.12330803548467661,0.3103461089377416,
 0.0014167962668151235,0.05023044840071422,0.04781801304466371,
 0.5298380910267415,0.39301832801059355,0.2730288065983445,
 0.45438145089521964,0.9473215813447856,0.6690025488923047,0.8467854378895954,
 0.222319560301517,0.8617803626892786,0.9497194808691923,0.4494934983178861,
 0.7064715798178932,0.07013845971791877,0.9630584991896021,0.082671843743374,
 0.8397949077805762,0.11167392790199537,0.8642661663913918,
 0.28099427611577377,0.8278824889633195,0.9993821776911594,0.7245260639257898,
 0.868272209209617,0.42929275169257386,0.46933416123689053,0.3473736837142155,
 0.5214837177676107,0.021698054222596297,0.6889775754147517,
 0.8856764079315081,0.6453850698592491,0.7512848750185682,0.2984850283810404,
 0.6039272659266646,0.3531550425259212,0.7468111372997648,0.7671086607649106,
 0.43478745818718434,0.9510011143499728,0.8059264159506366,0.2405342743025416,
 0.8212457715333604,0.9619556468623014,0.9846795782387905,0.048655421599339,
 0.7870224514998916,0.6868238314040276,0.476028227415068,0.012234338774926301,
 0.2721374346238231,0.43155348928238235,0.621619902246701,0.9964558624616476,
 0.36797953243218906,0.7349673264142926,0.5759186988422927,
 0.006420574070305385,0.8541823643587697,0.13580368142720955,
 0.09860433157413173,0.7680144907941889,0.6508897685875441,
 0.17528299787781132,0.8925203115839816,0.5091839207729767,0.5527478227766198,
 0.44481571881887116,0.7471394024089327,0.4854335215054052,0.8210639133661669,
 0.46130267876478104,0.8073074484215546,0.08181996039084394,
 0.7340132360747037,0.7742150845303444,0.5023001812280139,0.15957819720449196,
 0.13031727997324372,0.9217409314242591,0.4744846831066818,0.9545372685286251,
 0.11571025614522712,0.4321777293341058,0.6337241481948238,0.9822648547229075,
 0.24400720724116076,0.5859931780298703,0.2068412303753958,
 0.13759567566631958,0.36379881340406606,0.2824329789399014,
 0.4733950518468224,0.35364570053992894,0.025419870630091435,
 0.8196635753837503,0.17231855821153264,0.24077937493728463,
 0.019073945763606548,0.55102419008846,0.4618771307246461,0.170978715315805,
 0.46956012649555245,0.6494829345790664,0.41401097972122025,
 0.08271221458479794,0.6911835671911413,0.7548939179051921,0.6589358045780064,
 0.5000995573407169,0.03760044847993449,0.19746251247448265,
 0.9575078180861951,0.8139816124779273,0.14405824661945432,0.8127601844934886,
 0.909200689619753,0.8185616538130774,0.6814647005439013,0.8629641092230026,
 0.14102116297113998,0.4409148065680929,0.8169049932780224,0.596668705041648,
 0.7608504272128975,0.6161540555668942,0.9101042506490954,0.5428778239004171,
 0.9644744695887844,0.6092069175055844,0.37179793546890616,0.7933276980239586,
 0.9797583508895178,0.19046515253277396,0.05903854664621644,
 0.3996805199334399,0.17606138134460525,0.4731323636892282,0.6012397568782814,
 0.8706039616356263,0.7258145983181777,0.3530153769559702,0.94474545266805,
 0.43094111779515365,0.08587104935665346,0.120266817308882,0.8610239161975279,
 0.9853404363470811,0.014837200101939563,0.7580635673348218,
 0.18999158902393032,0.05763169994469408,0.5927277850056978,
 0.6591640611750976,0.5154055486752366,0.9940558508308212,0.9139258972089054,
 0.035734424925398356,0.13745883002894976,0.9129359590137016,
 0.7376014353376164,0.30173449243150696,0.7947021789501026,0.9494798881309026,
 0.4911087452424938,0.12022021895605395,0.45693187963752413,
 0.2414601766165404,0.86202956954658,0.3445140658386382,0.4385714210589571,
 0.6113625595152506,0.49584310530980247,0.8507161720529189,
 0.20631019828225516,0.10806011787334768,0.02341634386811131,
 0.7296850309120506,0.9255257785856669,0.8147945233026833,0.8095033843185279,
 0.8847535955251398,0.6720514196220825,0.09045471306831565,0.8542734285606268,
 0.7025966616118737,0.15506158712943674,0.06644570701561403,
 0.2965522190252282,0.6237577152446716,0.5446615545468533,0.53386973233306,
 0.7143112571170028,0.6820772943052774,0.5861680055181498,0.3833207277396188,
 0.4313729862380302,0.8438267536235294,0.030301247128718622,
 0.5748724425126664,0.26031190803784177,0.4081436182142575,0.7865279852826447,
 0.9011450107311805,0.22586602272399314,0.5342061516701649,0.638567267858726,
 0.7255360286490258,0.10256051432464919,0.2786236762343942,0.8625259108954012,
 0.5424193116722879,0.2512134062976784,0.8030859462718644,0.3088304514134228,
 0.501511698911392,0.21308213019439148,0.9099808902475067,0.8253762630231077,
 0.27177540517794585,0.7339598679693917,0.8798270927723746,0.7646606725752434,
 0.8504124556549584,0.3522837797492532,0.8882609730303436,0.8686135471710582,
 0.6223116994773003,0.938889392937422,0.6836097383665312,0.4980075855987861,
 0.9091122514699657,0.34487616586716674,0.29846902170079637,
 0.11621646773061811,0.7084177601253874,0.002941989416724744,
 0.9327076359165526,0.07918621208873411,0.7960176938617449,0.5812896948185443,
 0.001202856329899915,0.8496632247259237,0.12742936140718886,
 0.18908460971658614,0.9148928666920767,0.22295901104270388,
 0.44766369317080845,0.9466087421336498,0.1559351220904015,0.879273450000276,
 0.24129704628229753,0.03859916572025601,0.10751767338353957,
 0.9777160893866914,0.5236850703591959,0.22476361671078382,0.9319614111315834,
 0.12190257227271162,0.4461772035818976,0.41518439792833783,
 0.12084646645398933,0.3129733244715692,0.3997969021281409,
 0.057386358642356194,0.07101864198595742,0.194171563250358,
 0.18810116385733755,0.16928032845399243,0.7287732714017211,
 0.7870535928628472,0.5474121445349067,0.24516078976227096,
 0.08062694058553854,0.5718629882718138,0.08336987012714236,
 0.42712276801037097,0.4096784395953783,0.17209224045818716,
 0.11216287685188342,0.1358845361545331,0.8614250979812325,0.5086618562837988,
 0.3180856603017862,0.416416874194657,0.667964246066562,0.12304567160705604,
 0.6324217673523651,0.18662073849259975,7.140729286276848e-4,
 0.8171191062345319,0.40933192273040675,0.6604071464582679,0.7655643406358641,
 0.8294119409118439,0.5405944704983164,0.7847985954954898,0.32951987494920676,
 0.22622577304116065,0.40188249419253097,0.7650489520207208,
 0.9986720320968945,0.08910860067373516,0.7164995603864802,0.3200189546018225,
 0.4628960366852364,0.853484836301657,0.6985100521411155,0.3888220496312109,
 0.8267590824486447,0.22791940485770634,0.8704418540968285,0.5748032566458965,
 0.6176605833371397,0.5461582038372279,0.36555581453962016,0.8660777560121742,
 0.3036076481210267,0.6785085915288379,0.060710892154519636,
 0.4121957804514216,0.7792934378624601,0.13615836322516517,0.3957374521623179,
 0.1251228186291201,0.036235084069681345,0.7593189024198045,
 0.8545283790817315,0.27849969310027245,0.8493230386988179,0.79346398397613,
 0.6040992588703302,0.6320342614254489,0.7056509992250993,0.30578896793059396,
 0.45215702276423964,0.11082632251252833,0.2625292087408583,
 0.7465941560594556,0.941012903716444,0.9829327980385658,0.9801784919798576,
 0.5580749440772756,0.4856321335662295,0.11847752462523298,0.8601372813442245,
 0.6020833187689157,0.5229414410172075,0.3165443615781749,0.06366328131996457,
 0.4459719858261295,0.7579460070961778,0.11453240194001513,
 0.22980444213385076,0.7798264715691156,0.8755900320342572,0.531989213424989,
 0.0820781884419246,0.10910016537998968,0.343772193664171,0.9826178434564358,
 0.12221554528752121,0.14034189715733025,0.1034494798556268,
 0.8070625464518936,0.5296155099631051,0.697429661283954,0.3949289010815872,
 0.7139581932570986,0.2546136847278444,0.35917659722258966,0.853286948718902,
 0.40461301915467696,0.6900781907536377,0.5245412154946703,0.0431739057731364,
 0.3752314608505607,0.8071171735219173,0.024968479723804782,
 0.7794665056606394,0.9780696549205871,0.6423732110653892,0.7038184272613901,
 0.8386140973871641,0.840434056105321,0.14449205517540398,0.34225114376784604,
 0.6858986971194708,0.6324543208168376,0.043238040095783026,0.889242486644768,
 0.20016068554206268,0.6209771742920637,0.41348460700535183,
 0.6870426098789251,0.09330841664184364,0.309395289225581,0.4677280356198057,
 0.5583808039130368,0.09372739027042754,0.0851468813534132,0.868305164539683,
 0.917224212042065,0.6358097682782624,0.3720078697050192,0.09094404714115911,
 0.7292867739151572,0.6839327823153958,0.053680619339895985,
 0.3101199281274165,0.13528677148647983,0.8665690001300838,0.6039733410914804,
 0.7976387555636415,0.36711107507816654,0.8406672510789347,
 0.048773013987931546,0.0011800588443464655,0.49401049722790036,
 0.6102477840966624,0.8155473943247431,0.5429519576108176,0.8634762079357916,
 0.4077048491678017,0.13486144536644495,0.4676698015598483,
 0.006710814918391317,0.6643303608483274,0.2286698944150356,
 0.02659528942502043,0.811220702884453,0.2746429998026524,0.8847923310067553,
 0.06339334840934818,0.2615523039742872,0.06877346863028211,
 0.4917988340079321,0.8239562007109634,0.11787098207409774,0.8928603179613592,
 0.8397913866192834,0.7854683078109065,0.27027982759897784,0.9652959563462608,
 0.7014842100369384,0.048088597145664824,0.760310631756506,0.8048389033880912,
 0.20939020640699157,0.9692296128237249,0.8400200750377204,0.6625101592008079,
 0.5178815097804118,0.7217210760199941,0.156623458945087,0.9069829268365914,
 0.763544053152157,0.9778385835194043,0.4565170520643824,0.01809221207149947,
 0.7041328889337285,0.9336407586028462,0.3361693950665835,0.09649871703866508,
 0.7491051736895754,0.9693122484921844,0.9057534775024889,0.6124442848524637,
 0.3257899811823288,0.09165838460616604,0.21682783471972533,
 0.4169878313786858,0.9727257273291571,0.035907716546465296,
 0.06452108378351396,0.15651915647419812,0.3592688190768839,0.364110222840207,
 0.4190708563974561,0.45077372418756967,0.8228703390259787,0.6328298715345327,
 0.3537948202860992,0.34215995143232114,0.20674299281352904,
 0.48422125100984725,0.9425382692982623,0.5907685904715444,
 0.026796508768671568,0.4385225550109877,0.9161702390931828,
 0.45490293892764766,0.797720882141089,0.7325844190959728,0.35062367069635425,
 0.7054035059657656,0.28713955675015257,0.4500596161223207,
 0.014115495204527484,0.027355119057342336,0.2608275848902075,
 0.1161762165056146,0.8506117384309508,0.6694146610194408,0.5257446983969838,
 0.6766606412516462,0.8045245247682891,0.7160224929646594,0.8388489643611475,
 0.251321644842038,0.4643268454985229,0.985378302424092,0.605840820171733,
 0.11714995849972765,0.5819447893329932,0.2623591987721977,0.3324123944385331,
 0.10412408691375585,0.3927554873653638,0.7024807158080546,
 0.06375918190260832,0.10918233352268902,0.901603851354332,0.2764814158278237,
 0.15915713779338658,0.01687602841881264,0.4453554923742209,
 0.4093694942505198,0.9052360370535053,0.9369324188465709,0.578425348281655,
 0.4863836820934787,0.44675930722060775,0.13005903766664606,
 0.7567516606555945,0.7397778000648091,0.8723197438017507,0.8003256424680791,
 0.09422860090493501,0.42207669007757853,0.7569276636477029,
 0.1662297461310911,0.43848226956196057,0.13822889907921287,
 0.6876481253521242,0.09501730883294657,0.19363219934134046,
 0.08951938762242817,0.32681062615767265,0.4127556504933041,
 0.4057411247730439,0.8768332233539253,0.3621900568117635,0.6866437347765616,
 0.8683727093917184,0.9124564426886832,0.6913532078736779,0.15659875394575606,
 0.887133401357815,0.212590355540473,0.22479385006435504,0.7507555134490015,
 0.8867815595988936,0.7680784735022437,0.18953545065470878,0.5315854518994633,
 0.2322999692545582,0.5438227130925588,0.7465147275378899,0.9003856611951366,
 0.7632371932594609,0.11000494548695305,0.6697964123985813,0.9916907806445118,
 0.12912662257690943,0.007341812741963993,0.6995593082529007,
 0.6319208561893661,0.7809347966336614,0.7321455895298461,0.9118189827149096,
 0.48246206911587264,0.13093720180548374,0.9554488365846681,
 0.2801855162345067,0.4006484401393222,0.7736306686017309,0.3581949628971316,
 0.059874320072132,0.4834632261327465,0.7395309842433975,0.4659782592416293,
 0.1627426823331487,0.9689034920012063,0.4346627434037281,0.3877774477809244,
 0.6063960844139763,0.27327011733632955,0.6913168615672679,0.3463633425673317,
 0.35471249985935516,0.758545589657411,0.7982772080629809,0.2021964770519954,
 0.18868276798981187,0.4933252608566846,0.46550680520298704,
 0.6194380030779301,0.1738227408658748,0.12042276365578664,0.5714692838679193,
 0.30116915608225103,0.3717540839289112,0.1975572181847437,
 0.07239074715095128,0.4922191603132253,0.5418155806357884,
 0.18912330817009315,0.5472581188027692,0.8571566324733922,0.1084734094280182,
 0.04667873038764547,0.4195024251560183,0.3532767008090303,
 0.06351895563547238,0.7046893970458742,0.8307668180579247,0.4174803622579817,
 0.4547951476322951,0.7933057565745828,0.402282437343096,0.6306492056884734,
 0.05379674104276155,0.2304845721251808,0.2823863607449194,0.6460732451911094,
 0.44541026800866623,0.4399618772205883,0.44273190101325977,
 0.9017412550240067,0.5684305430794986,0.1640640742305799,0.14369782310019952,
 0.5161409166827897,0.5310654959348631,0.9502933229691982,0.153320292942403,
 0.5339762358004096,0.16670418415746746,0.5089062497124864,
 0.22445761110498164,0.3696690904296358,0.23523424450830965,
 0.3940408656262394,0.798254413580622,0.3083813802706543,0.968416209147112,
 0.719416751147093,0.06979285779479294,0.5610251652767462,0.8267929958180971,
 0.945174912477542,0.04014179155307951,0.9853890652404822,0.983727204369119,
 0.07288884626160963,0.08494931320203547,0.09973600622795853,
 0.45464654023403517,0.17004220979036366,0.2429718829979567,
 0.7882717171815766,0.17256173582911427,0.31583656197003773,
 0.5439431112669888,0.9760972159483766,0.32032146855346133,
 0.32126165982296284,0.04144424113412082,0.3062191211531683,
 0.9595845947829542,0.4239806774556878,0.6361441936869142,0.8952634286611448,
 0.06575196544333872,0.5585147321725101,0.10993964361243314,0.481651336005396,
 0.9024924713791427,0.1444459120413475,0.9431609634880753,0.8318925102998679,
 0.03406745021444313,0.1661710443436324,0.4204176282039258,0.8727124771786601,
 0.44515145123325595,0.9561797420392324,0.6484142816005996,0.8952535968483848,
 0.4201693169642319,0.02439456769617976,0.2955808783289233,0.5690743046182,
 0.2158125756340361,0.024043098386084205,0.6016884512309301,
 0.3126966411525033,0.4609219036626794,0.563722026561981,0.18684671840515932,
 0.8416184341762374,0.9173270137035041,0.16278676259125302,
 0.12412937058332074,0.8288739459135864,0.6664860933454542,0.2503006404343441,
 0.6239336627224732,0.5822763797096764,0.4925204263121439,0.6261774179569508,
 0.8035286694457426,0.6604393545092991,0.3456068871366931,0.8512127149496662,
 0.7300024171552153,0.18881896497588013,0.8868422777468293,
 0.17736704269250225,0.2856701982036782,0.9098484918581535,0.7733485310845812,
 0.5821060369463631,0.2632710922373941,0.8875383236255493,0.8085338448229317,
 0.7398462441386656,0.20683737417354275,0.1317577746544688,0.3736374383253751,
 0.1672594936077969,0.5971770642440323,0.9996389093862942,0.7681056663238821,
 0.4234662574114959,0.8118491463091143,0.27557298406121333,0.6876463888213431,
 0.15336521857511387,0.2712911010455813,0.8022993377954897,0.3246084983885379,
 0.4913409926179594,0.37817428536649067,0.9514713924100252,0.9585001614356493,
 0.4088811434253964,0.18262963485867378,0.3211137570079975,
 0.25607530814615365,0.12740889499757602,0.8411330054688437,
 0.7424962869869571,0.6365865904999489,0.021290718705284095,
 0.37803734896515784,0.308423292861309,0.29689331496452265,
 0.08188216750647215,0.6457875404122132,0.3971068686619813,0.9641648898969349,
 0.8090373624990187,0.44141095092343996,0.7278126646537412,
 0.14484116570411765,0.18937290083742397,0.9490394392500544,
 0.004387138751275832,0.5572749213441166,0.20120269103147537,
 0.15412347403898363,0.705626553190421,0.8770969126237088,0.6533972215963785,
 0.7961149731377913,0.5194858167804755,0.6825758780866316,0.5038723767514038,
 0.3347506637376728,0.24895782160432844,0.9665094703973552,0.6365321416934455,
 0.14506303079451077,0.2847216265613194,0.7288038193514947,0.628576798966064,
 0.35144480184303195,0.8073196197331461,0.654752533384045,0.0798235843732652,
 0.4533302616435627,0.9619176609595087,0.7887727282968148,0.44951850171864227,
 0.14632073662426426,0.5316374480749111,0.8804887475179902,
 0.38517605627217777,0.9267224690298961,0.5501775161333253,0.9906230379079727,
 0.9457217381005714,0.814289449002688,0.15608734717439376,0.48234999325861483,
 0.0641539654882306,0.10952112228875477,0.925032650651348,0.3564624177382685,
 0.8967003442421801,0.9613611710540783,0.032762021503893846,
 0.9813297039908196,0.572863923514083,0.19880597607243822,0.6102656170411045,
 0.5111957067450099,0.5017621351363858,0.5609675480800986,0.7551903430757434,
 0.5006701750904459,0.1834183608517621,0.519101499167772,0.47430250102947735,
 0.454462962914603,0.376620979745379,0.4021534555588959,0.10242841289229554,
 0.8349727098655734,0.8141837887702604,0.5875062741193685,0.7898735051214515,
 0.21833422743879105,0.7158549011209693,0.8420935490325985,0.795055474854278,
 0.8418257414328167,0.22947909391941035,0.2463492299527148,0.5759099909933144,
 0.0052358185127956425,0.21857813296963902,0.9599461005011707,
 0.8422382303018496,0.92384180679243,0.04745084710877623,0.2225183952973262,
 0.4539280864568287,0.9967363328478471,0.9797694662311873,0.6967033551529513,
 0.8571700470874208,0.67109805434169,0.8985280811542294,0.27349300100599805,
 0.3233869383648147,0.15891455400591137,0.8904654148054754,0.1998768452643882,
 0.6613133721146772,0.06019136947500814,6.7298546096628e-4,0.505235547041349,
 0.953087947864519,0.27351636441095195,0.12145090632007627,0.9188952798418878,
 0.730226529706983,0.8368512059313916,0.010755843080068939,
 0.46029688373279576,0.23626378874646492,0.4934662391602982,
 0.7754505138888746,0.7542689915950106,0.7110621997188722,0.5436227491340595,
 0.3061929250782012,0.6307470653210996,0.793096735109196,0.46776716647507555,
 0.7929716909081301,0.5581837945797212,0.3080249799042172,0.5428561007628134,
 0.17379248895168253,0.28197066363546597,0.9129563381542694,0.722710679429433,
 0.8475762222227398,0.8704198025826622,0.28470807711551793,0.6547454254733129,
 0.03582975427627799,0.9247186432122266,0.06869634215877385,
 0.9408943535184163,0.9932719455481174,0.4216084997948184,0.06784416882038014,
 0.4684095713704368,0.024784115434759335,0.6703008281140507,
 0.18321971685723504,0.019624201149225473,0.24607174169644175,
 0.20090783973907134,0.8785804868483058,0.15550910913257743,
 0.3466718462531755,0.9293870421332215,0.1904451218878006,0.7459483338549203,
 0.9376871256513449,0.5819240535422198,0.8103293979805596,0.3714405270131428,
 0.8885826187473886,0.10766250542205802,0.09420370039596071,
 0.03612521049230699,0.2036903293569815,0.046899662375935236,
 0.8270371894521055,0.9539426441950283,0.7714517475813237,0.10799994787680867,
 0.24036315389015928,0.3698170089897259,0.6810790139226617,
 0.10713040875217428,0.23923233702371338,0.4288061840105042,
 0.6721932134544454,0.8526198098599639,0.6775296207295349,0.9429159921682216,
 0.23940437142140325,0.8030961811081768,0.9124596546535919,0.3147848980757981,
 0.5790375885209937,0.06586386031261171,0.29133266952506065,
 0.45289335458194735,0.41484650673195,0.10116478102382975,0.6462306928948112,
 0.46691924740749036,0.17304214638157955,0.14611247826939766,
 0.10155383007281904,0.8045537638469794,0.5684125440273707,0.9597143012243277,
 0.6541130178907932,0.4955771014423065,0.04542668895111901,0.5679477375756369,
 0.692321810381046,0.8074062551114709,0.10206030459628346,0.7443675568720397,
 0.04919854424855952,0.6507657488604576,0.6053046258091118,
 0.07938456662919768,0.6183396333836677,0.8549375856117853,
 0.07675364436098531,0.23320361666348866,0.37216618486674535,
 0.42818345159390026,0.5050128374143517,0.9304284546148947,
 0.13240315047745288,0.30776963891621245,0.9960245544033215,
 0.47005958255337743,0.3442890885414128,0.7600070378975892,0.9472545168623058,
 0.9782574085133016,0.3283807145333708,0.4799250049872933,0.45494125957938514,
 0.4936325896943353,0.28225966176146144,0.7741334562569875,0.9411060818989141,
 0.7823082258733561,0.1982855342174913,0.7865181383823411,0.14970213575187885,
 0.22336575856350993,0.6003772598732031,0.13781846250230734,
 0.16665139382890182,0.7991865617968447,0.9284893994069585,
 0.17523808019608622,0.3396437219602827,0.2893550356756408,
 0.03185251921056864,0.45590380196022806,0.4009346628066697,
 0.3948614981454427,0.9231918187649505,0.6326275516277565,0.7514698888513347,
 0.6506134077288365,0.8279048793190036,0.16288121766241215,0.8283088823777986,
 0.644879811461725,0.8139787507309337,0.33302649993720634,0.7973929968044857,
 0.268842445473808,0.900059589938183,0.5953150501480785,0.31908692027798446,
 0.5962045584679317,0.6894716739044172,0.10494425390694317,0.9879132529170219,
 0.5387304628475623,0.6845821086789732,0.9394563329613479,
 0.026167932768007995,0.07375917267854637,0.1707139616109229,
 0.9303493607962998,0.5874337715198574,0.8426072047101776,0.47364999881083325,
 0.7679148452838273,0.12647461601404753,0.8523898763020497,0.7700447366215557,
 0.47526199169539085,0.8780316145791945,0.6947534166179883,0.7122191597116939,
 0.11217972957175626,0.20191506860643538,0.9362679287994113,
 0.2029881490168849,0.04872660943977347,0.28793422526541745,
 0.31845468553362366,0.4396451688314672,0.6726839533987609,0.6357854145093212,
 0.9143545784879612,0.510378849028603,0.3442669722330156,0.2060456823556408,
 0.8868906274513556,0.8852556681164092,0.04741828695028316,0.5800331222120052,
 0.2618518440172739,0.06839822942234142,0.10485774594578567,
 0.7936504067448386,0.7117199857586538,0.00678836494447399,
 0.37642828214665125,0.13863226285634433,0.20560243866354344,
 0.3980560729942908,0.5515057927015147,0.7467687292401415,0.8717692566269986,
 0.42693421826621925,0.590039044118877,0.41338167440828233,
 0.10319831698600335,0.5118578050518673,0.7076890046611652,0.4151073790309612,
 0.6302062886295192,0.5794645412720967,0.14477220795187018,
 0.22694577722682296,0.4268829962054179,0.7561034267780353,0.1451397942074989,
 0.9601030764336992,0.5099009515119243,0.05228024141895593,0.4927064507802762,
 0.344797570550759,0.09326450065217184,0.7017186226768134,0.9445854635277844,
 0.5714107247988974,0.08536604847324125,0.9839763078294995,0.7844319707511013,
 0.28243324161647265,0.811539464720338,0.275140219028282,0.6256858355843817,
 0.10489696678298543,0.008421628080805599,0.0030306180605041977,
 0.6103526795927569,0.11952328088509756,0.2515674676535171,
 0.41269502243720013,0.6900812365202832,0.27456292011954386,
 0.15430270913776634,0.10352709223370016,0.8369669566590987,
 0.9453635090731975,0.5298606158393898,0.36659309881942737,
 0.025471527195019328,0.878957629824185,0.8125630359738271,0.8747764352806818,
 0.9013490953352841,0.6470236980549493,0.5870563794348908,0.6309076305060022,
 0.289469939013245,0.29972354260656886,0.6405719618074113,0.5108249440608226,
 0.6948899341390788,0.9719007549846942,0.1225729315926557,0.1427537026958423,
 0.3274633122170567,0.6847678223617798,0.7584376091438829,0.4871835307221285,
 0.657028656547576,0.12181175290574331,0.13117477821355017,0.8998392833949419,
 0.14589187409993654,0.01247959371613161,0.4230491296002935,
 0.9279303968800459,0.8843875023713137,0.17348928153003607,0.5545674835841874,
 0.7636408423553207,0.16399173377451404,0.37495583516451414,
 0.9729722251152969,0.2373136127106643,0.3342574498727512,0.6128889418468666,
 0.23978833338985428,0.47075136919453797,0.9461002873904297,
 0.1968804610379873,0.5694724751847411,0.7358139818112975,0.7978330895507105,
 0.09669406121367519,0.5820781188321906,0.433546606105919,0.44789028496559113,
 0.5681167123240716,0.969165978157282,0.7826145401782049,0.7826697572687544,
 0.6538154094806716,0.6387034580416013,0.07950803003152929,0.6041765399127901,
 0.8491737117112045,0.973741623008105,0.2497538219127905,0.4504096854813776,
 0.7480252350606946,0.4751027631376964,0.9227798109600741,0.8142146874967068,
 0.07788270835088751,0.2236896501296769,0.7769661809510104,0.248229972700308,
 0.48210619651886655,0.28275838318099344,0.5203275992767895,
 0.9924556545046052,0.15193583060642024,0.9065204531148,0.795774809499339,
 0.763651223577922,0.9383390637138013,0.23025354681365406,0.9658466965871868,
 0.9444568487013617,0.4246655531819382,0.9959416932688472,0.4600647116780783,
 0.07192266641882261,0.28009407644793083,0.7151283627668491,
 0.8303574638546036,0.025187386839316428,0.1022869490956606,0.407573405299391,
 0.7241380184654289,0.8208272304009963,0.8038801340302829,0.9675057568350194,
 0.14912881342096906,0.7919867218272681,0.9718772327221821,
 0.07713480233856806,0.7736513818479536,0.01778912617527073,
 0.5820718425374705,0.7321307466550904,0.8861109740430508,0.1404323566709279,
 0.6986807504774725,0.13030817347711365,0.1615488505623225,0.7660999929295274,
 0.3493439982070702,0.6144913929653341,0.9071421049326032,0.41891439134855035,
 0.980262845385352,0.7331603291737679,0.45219224829756033,0.7043811824883546,
 0.4313602575435498,0.9317494678797449,0.9309494753302128,0.08879984385807849,
 0.6219564001998924,0.8950584574846968,0.16901713100254068,0.7160879128018238,
 0.5655324173287031,0.5758271932230439,0.8232791508191701,0.9204746785451869,
 0.16173397565354564,0.5972377370393628,0.3036062892753848,0.4854905623433219,
 0.2829641401879447,0.8788208449719898,0.863546123583312,0.7297417845215705,
 0.5883054580950361,0.5191532619716742,0.7912691903989731,0.2938283948148229,
 0.509376062126077,0.4443335665519881,0.19756151487869966,0.12428795950754234,
 0.973732581882063,0.7574352366003803,0.6328680128288635,0.713651277150818,
 0.5523676757000983,0.9945377452651926,0.586754300332922,0.2975124180361345,
 0.6154606334323526,0.4386238482627024,0.10254988905717233,0.8141425262676113,
 0.9817072856728182,0.4677132586840884,0.7655928456219874,0.19378396355119754,
 0.43453748841982476,0.2786763921236577,0.8199486940324242,0.5817748363956671,
 0.6760203537314805,0.7802687129693189,0.7791351204900236,0.43598310320333633,
 0.6666843567360256,0.04645802847068181,0.7875389144900528,0.5709565947582971,
 0.0986596101382099,0.6055030239360454,0.41187953084725293,0.4431969014501004,
 0.9016377070587899,0.9575543413829865,0.2959115660701751,0.7765219083328806,
 0.553492731416435,0.6161259808273976,0.3142770355492037,0.2554067996404933,
 0.5826428226178709,0.5855799154366481,0.7199356257383732,
 0.034698613173501536,0.41073207583535964,0.22588098351718844,
 0.6438716130079446,0.4212382228533479,0.7860587498041741,0.10740553873351899,
 0.9717718783259269,0.2463806858961498,0.6782492297093976,0.18324880485557238,
 0.03393159322042516,0.9098819812576235,0.9881180703927132,0.7103847330627759,
 0.8874873858898789,0.9370260556538228,0.9683707508398222,0.34123036610632773,
 0.07976795567413708,0.8830009027179809,0.824350273079273,0.8266822287781661,
 0.9747153975100704,0.6638292698309856,0.9958919777582808,0.6170973786019367,
 0.2077977997780669,0.24777048383605405,0.0877914355210846,0.6461192084149165,
 0.09803109657615738,0.0058910314071052294,0.5817243392891869,
 0.4246308515342724,0.5668567684837518,0.9881385285478164,0.6271583052966396,
 0.9040052427058078,0.8532028629966463,0.5303096370067266,0.6005422643010561,
 0.9304465769611862,0.47656573482338715,0.6269384168355876,
 0.06397253899268562,0.47898422039296185,0.4027086010539639,
 0.7470026312054215,0.3788750291328762,0.010870369993389084,
 0.07751583956510488,0.3656075959149595,0.9815742645674437,
 0.03422869943192597,0.31618300411605493,0.720203343803441,
 0.43765845915643153,0.8941307437161318,0.5812781798619926,0.6533849370684774,
 0.27035060985632764,0.30227714052929644,0.3210435043698896,
 0.3850691359638434,0.626182894330208,0.3487640066969582,0.2113594329404852,
 0.7093701498862331,0.912266708572508,0.9376624097169355,0.5576099668126892,
 0.2686818696423269,0.352595949043497,0.5279247762704301,0.20893521478951638,
 0.34968630973792025,0.49598154031756136,0.2013210396577516,
 0.05616178457663212,0.8904950915130566,0.772746188956062,0.7569019773433324,
 0.17914995261521938,0.35423853398200267,0.8590685396966479,
 0.3858321786542035,0.2673282733762503,0.042581834681404196,
 0.6698423427520952,0.9278230952471075,0.8545558652332508,0.603959299063621,
 0.2765560666013819,0.624636144491747,0.6669356865853348,0.6536885308944834,
 0.39539184984995435,0.897734823895129,0.7132116802989427,0.19708213273313646,
 0.43740100165576135,0.7393434483558406,0.6991940944160698,
 0.05122558075912942,0.4153391399593829,0.02912376648632986,
 0.009128924403131733,0.4397564817838697,0.39138638762567823,
 0.0028344135654010394,0.9803265386769257,0.40680177908563286,
 0.9430438706087919,0.30564608317427533,0.745273574036905,0.9343939556675721,
 0.05837359971200906,0.1342000443314716,0.2703310986173174,0.6316900518331746,
 0.47944459261644157,0.30060544551930735,0.9740828038785633,
 0.8682656702341028,0.2875612655298453,0.08800803009860503,0.6355818777220402,
 0.32590350523441647,0.10575626934760263,0.20486900630203242,
 0.2681506958144183,0.07633775907341944,0.36573148854910875,0.332099933741158,
 0.37354008083774304,0.8483968586785702,0.7631149093032228,
 0.03793464851034256,0.7704482895707057,0.9605164590378065,0.9489688794336332,
 0.17019326461253814,0.8587071430624065,0.07359415109943246,
 0.19873406383943992,0.6746601060709018,0.705614569658241,0.5405654087027328,
 0.5525837272994365,0.8965686787695515,0.25818202350342623,
 0.33745156709287427,0.40291439606933077,0.9041460278689322,
 0.37302339613031077,0.0081981265742338,0.4749068552757454,
 0.02539907075975234,0.962438365116991,0.8117112235363778,0.2518534511534676,
 0.5279021577880266,0.337262475016077,0.08799208304708972,0.3506899480336694,
 0.6783663784388357,0.14687792147069556,0.37219237620674284,
 0.9778495148042111,0.4378115537708056,0.03380816657624974,
 0.33286649768192966,0.23732004606846313,0.1266419165916638,
 0.35364417860412134,0.3966105429049842,0.19020021188737557,
 0.31029379493349685,0.8541699209191399,0.42478640900165976,
 0.17535707808993117,0.4387675892595273,0.20802756041833947,
 0.5575225854790832,0.8667915223638247,0.6194394560496566,0.45591072806905175,
 0.32621733745066095,0.3385174701018865,0.13563546177599894,
 0.5155121483939681,0.8415388673670418,0.5963481676194355,
 0.032586206969220544,0.039438674210288016,0.7947784741503024,
 0.7575984242216774,0.8714200185541124,0.962559765063249,0.962669367643787,
 0.9379403128210954,0.339308520320477,0.5094995399742775,0.23426754405844463,
 0.9550297927477224,0.5229318946782491,0.6894880038279969,0.591908733667693,
 0.38707244737820057,0.8347942420795447,0.6304420562795618,0.623934161627185,
 0.9530179129748689,0.27227672422018534,0.3554036145121745,0.9760226195426108,
 0.4019060046887514,0.19399051487260854,0.0813049007512221,
 0.08103215812479081,0.05767866342524508,0.4620240945987093,
 0.5147008835520184,0.9804130463132629,0.08422022433774279,
 0.32868169938969805,0.9158476874113238,0.7163614567309793,
 0.047810751812745744,0.48972829269786433,0.24389503153581193,
 0.9418684231687869,0.7406594711384376,0.014936140926562258,
 0.49781734936767297,0.1750970450462822,0.7219553135665749,
 0.23229052483206214,0.17532744335080197,0.46259911957287225,
 0.16291222664843263,0.19274883534990872,0.519082505834035,
 0.09458165967664134,0.09802622078773848,0.754877786598703,0.4776759810813662,
 0.4912367349171688,0.15905676039543104,0.468446481685021,0.7113206901570507,
 0.5042766960316297,0.7239372877284027,0.811848826732943,0.36346502904730915,
 0.22836266016298346,0.7891918202830098,0.207172190527675,0.06804573131118707,
 0.25719913803554917,0.9778335226021478,0.07944939157346487,
 0.3129500651549344,0.3050862326622108,0.275646940294485,0.7309759079034827,
 0.1971012022110351,0.9677597815577914,0.08362328826896803,
 0.46956294873057636,0.6068176556836204,0.23591296837351516,
 0.5902597564255774,0.9250479501142926,0.8230242840792944,0.4959133243963,
 0.18677438065181007,0.7978662397278112,0.4306845940782573,
 0.22931117312514837,0.7336483778016312,0.8854395171085688,
 0.09086771180287478,0.1294520888234416,0.756127124006063,0.11300614467296632,
 0.276926766047506,0.48946765045656493,0.5646333089083618,0.08948535072333241,
 0.043988641192525435,0.40367418288013013,0.7978918628050202,
 0.9102461713246008,0.32697399733318955,0.06249849223181336,
 0.9624921532298403,0.2692934786844332,0.8270671521036412,0.06217824569350383,
 0.7758485065396807,0.6375334490714449,0.5593103775217576,0.7603918738564299,
 0.46080153330781304,0.003999676391447382,0.8045849179733161,
 0.41261859130553524,0.2709812599241904,0.15850800965506107,
 0.24267715794886546,0.4278649470464765,0.3045228024139146,0.3899560822718737,
 0.6803511167388248,0.35911577667941696,0.6238573735516647,0.4049898871278479,
 0.145722840402722,0.11999345546894447,0.46477071286600835,0.543001067023916,
 0.017684478698787198,0.003516888924205097,0.21054760066085176,
 0.24172798388161854,0.8733516719400054,0.8557190886522459,0.2621395484509519,
 0.2381142517460645,0.4845821788389493,0.7505115373439905,0.36501149433312186,
 0.6760126140521904,0.39428414688072233,0.47949572749427394,
 0.17118691281010168,0.5048531415637103,0.3665662716873822,0.4877627749368785,
 0.6194293259395115,0.48698139468443324,0.2825868918805753,0.4972219061769305,
 0.6681545676863303,0.45172694824485504,0.1714873767461511,0.8118102486314657,
 0.8108217997709133,0.7079198270392726,0.8666853587953304,0.9433419836345713,
 0.3589099719780653,0.6657129593609596,0.6941914769654934,0.3616528521207164,
 0.3881061287271611,0.9503943609575167,0.0905949117454804,0.4537980892560035,
 0.2946170261162383,0.03591467270915716,0.8301615500974298,0.422248352533268,
 0.10086994018001738,0.009906321374319016,0.695587516831797,
 0.22632444957302567,0.03359824362528818,0.7443456760526329,
 0.31920217317788346,0.717026105057817,0.43694923599597124,
 0.29820753611257766,0.32273328410708535,0.8709719634632672,
 0.16579317418581774,0.6019794976803255,0.16654106512902023,
 0.4867443059345289,0.31436116838183503,0.22064197644761452,
 0.6890156109046202,0.4503946515433881,0.03706178096488533,0.3345514950208496,
 0.47118464875579424,0.5608379021855973,0.9108219232843646,0.9028413423093817,
 0.663103334305053,0.6304009173408393,0.018642303745961875,0.2201535267579694,
 0.3535588817999761,0.9265358509163154,0.45807533159285807,
 0.05500564210043668,0.06227194766857236,0.13942164743218188,
 0.3329154263105908,0.9650071344233582,0.6109147969076627,0.14183992486071884,
 0.07496522845058728,0.5157568261879744,0.34610038901379314,
 0.38816787320125123,0.3006165950305688,0.7552883209435495,0.6529752414900312,
 0.44017912270393444,0.980189690518531,0.2730673265180201,0.30124270288244737,
 0.9084077019608094,0.8578717284175389,0.9878323891230518,0.6626676267998485,
 0.9602183685575889,0.5376985384071268,0.37237849875360163,0.2794980062392104,
 0.3785057807369592,0.4868521013204614,0.8672794388520145,0.6782375858193451,
 0.6822465815250935,0.955719188352479,0.9282250124214678,0.30211481763796744,
 0.07525933682126118,0.5793721236315128,0.9463593104469781,0.511229746112758,
 0.45251054287224646,0.6574712719345145,0.27517741801393036,
 0.25764146676683863,0.9307386151689896,0.3397890475509603,0.7066874762385935,
 0.720427427918433,0.969016959832829,0.10695323815406499,0.44497497999031066,
 0.4588807944554818,0.05675434743747587,0.19527761571714886,
 0.5980890563916175,0.541423975236291,0.45672989536954345,0.793796076371267,
 0.8228534135606992,0.06523113216512177,0.16563942137588206,
 0.9939782714388427,0.22928925827220537,0.6239683061521032,
 0.16403249006582632,0.06670514439980912,0.7573326970716455,
 0.9296818292963724,0.5484701150594369,0.674455385650661,0.6544785547093455,
 0.4761424024894664,0.5730735868094156,0.11422749225925999,
 0.012294210223400492,0.15112960965857636,0.5440648968474767,
 0.3365125398451012,0.2391761960806401,0.14995350010970976,
 0.18361413035340124,0.6337146126136355,0.6704428213623909,0.7174434280335239,
 0.25486287767399074,0.09821512197188598,0.30763977966185363,
 0.03724339752414363,0.034655270441682484,0.5393029972932543,
 0.9728708682020578,0.35073742688541,0.5966871882333806,0.30198322060884475,
 0.9990627047540942,0.08222083680737458,0.7791657589347244,
 0.25802812836766675,0.11237825465020468,0.504259327252163,
 0.19601445670649043,0.20402159466734282,0.8084593964766814,
 0.2157826422839355,0.15229312063075318,0.10766387572906955,0.821119171908676,
 0.2535548124483946,0.9066482456155935,0.11981186724396908,0.8378041282621397,
 0.15606332922511668,0.931167396184233,0.5067415849202054,0.2576049365205829,
 0.8320583951655691,0.40195318646248834,0.17123468978619072,
 0.8804332374465179,0.8596792893766894,0.32946487153198556,0.7379131051578164,
 0.13663993050711964,0.9207510061676218,0.4380088049871591,0.380577918945211,
 0.23077685635896394,0.927158432758528,0.20055715654568418,0.677584331180381,
 0.8788173707248753,0.6271820978001431,0.39248865938194577,0.3629049141267,
 0.806252139592936,0.05293188765385848,0.9948366826405483,0.0802767429497585,
 0.0561948381136379,0.029930515613890596,0.6371172040123537,
 0.16084743843079397,0.5571255441899052,0.063393856687602,0.29286115439975324,
 0.5077374785603774,0.4930856000316446,0.17779240421919373,
 0.017241094084457065,0.9360655342604971,0.7360224962503578,
 0.18119837828147645,1.5784733272838913e-4,0.9025586061451456,
 0.34311687880574326,0.19227436979035417,0.8372658429065483,
 0.8639230247909452,0.7800814438392604,0.8743382322459974,0.23351420228074538,
 0.45147969147116285,0.3515232021745187,0.2714530225598988,
 0.18519733065804034,0.7111445079191487,0.1727896657750453,0.8102947429830587,
 0.4440804713829746,0.14628856695952708,0.5613642492141262,
 0.24294246489015137,0.4894010732014279,0.9630614556281117,0.943865171484328,
 0.7782560196860169,0.8971935233820646,0.7820253175332699,0.10299351066750229,
 0.5636884204012774,0.02680400141926087,0.06822385042085133,
 0.7203898528035994,0.09002830106710358,0.9985133694254519,
 0.34302258199411695,0.9259550415521862,0.5450133740754523,
 0.37880168010207105,0.38323841278614657,0.7513374133003119,
 0.9816472923895226,0.38615487332529297,0.6443272445775889,0.5852292147695196,
 0.7348647949554701,0.40611969628684363,0.25283094045840004,
 0.5634835233295851,0.5687578522603544,0.1937486179191069,0.12158999048255048,
 0.2527982986073335,0.2589309330156315,0.38148236127806856,0.9822613321581676,
 0.9542927817335188,0.7805619500201925,0.1284699312574219,0.6042577888728473,
 0.19963487082308928,0.16705966940092953,0.3813473251961562,
 0.8814955774991299,0.5127329546918923,0.4730004025637967,0.23096493237663251,
 0.855174649920087,0.9116633973696945,0.376154140848332,0.4287954328777478,
 0.604634946148499,0.8500228046434434,0.6891167617238038,0.24250847479027604,
 0.9864422079210861,0.2693083705033771,0.7342870556992809,0.5920860732551338,
 0.0257897920590604,0.8630818664430467,0.35612317443879216,0.7963282031241946,
 0.9840794816654369,0.8391886472710406,0.5125606899535909,0.8707906711589376,
 0.4946091979399897,0.6453311542979823,0.9338849972111878,0.7882604380823824,
 0.6807716388715696,0.4045062629138614,0.3521350980672465,0.9940466103203537,
 0.7964017412003348,0.5093653227755189,0.14845824434455013,0.7374615335489653,
 0.09905485451284834,0.7324144559368336,0.33012891601219985,
 0.6565383712434256,0.5639359827751602,0.4691368220837161,0.7809797379686908,
 0.8612406871210809,0.7846808928224615,0.6715791266311917,0.6526602556356103,
 0.8167044256274562,0.7325074407295601,0.34241928376500363,0.8331148332742077,
 0.5431922014688193,0.26808313245874726,0.0663859898872301,0.4055692826508609,
 0.186376271086373,0.40531641103793214,0.7174145156334276,0.4601690352947456,
 0.1598001726573045,0.05039092814687263,0.8259541264461763,0.6249086659368642,
 0.10127545312363584,0.4751971020927508,0.021482720738316807,
 0.9359819610337504,0.43662017591599334,0.5616474285762781,0.7020556407970006,
 0.855767051341844,0.7264734617793989,0.23958048055430659,0.7148555292843366,
 0.35313357533417966,0.05347326259685403,0.4199169033355956,
 0.8174237231968917,0.44699277681388905,0.39412654712540485,
 0.08491334883301205,0.8323017258732945,0.2335256612097415,0.7048580454906963,
 0.7592511465902484,0.33799398372001965,0.8474496815601676,0.2506797558054982,
 0.7109948594023919,0.04551205724506979,0.3646722739598395,0.7517392333213626,
 0.3810130581700424,0.07317381919605381,0.3198444592703298,
 0.18262844389124422,0.9815296053631783,0.8722177988606191,0.6892846925090211,
 0.4043044761764516,0.923881535963075,0.2246229800724283,0.8154503147192871,
 0.17677503178601228,0.7234352975822471,0.12551436923115344,0.854004828284119,
 0.9165790353475237,0.02726829192232716,0.04923458782726853,
 0.4968295906757684,0.8784663618283872,0.7665850772310695,0.18254465013773824,
 0.5113159001956924,0.43962630530455504,0.9297635415469935,0.5990034912455272,
 0.7181061399051305,0.8529565349936454,0.581477291057632,0.2728649215912292,
 0.0740165780019133,0.5047992135173697,0.15469867783236435,0.9012203729170266,
 0.3278683020436627,0.7990745503353656,0.42460926854139136,0.2728410555306422,
 0.7315993010111844,0.5249826766220007,0.2711513518669205,0.1594089536786747,
 0.5528759532469193,0.25885248814263295,0.5032602483144809,0.6837253985108105,
 0.42025533840369556,0.13941384210897612,0.6672449548695278,
 0.19269386061432914,0.5604768514486367,0.10914975125572035,
 0.9594588213604789,0.5641217545604753,0.5445880356179181,0.31711687689451595,
 0.9527419899552352,0.18833347886025964,0.3456146519853551,0.4870162369354105,
 0.4177356909853952,0.2821099318770429,0.020769224678461873,
 0.1190415798127753,0.48317631918666604,0.08598647634797385,
 0.33511477121868793,0.4217030850245659,0.12898386312800825,
 0.6041651437880574,0.2503954206089032,0.7554492547828373,0.04742183413190615,
 0.5634427090422774,0.3329884749659494,0.8073690024821779,0.7985263581520252,
 0.6000946202320571,0.3542033430269875,0.2948073462155547,0.4923419370287827,
 0.3935370960370308,0.7859154181210068,0.5333524104807325,0.04457703503398447,
 0.09507891019372028,0.23841473037969008,0.8259416623372677,
 0.9564574074928078,0.23356282758731317,0.21445984951229524,0.93669465472343,
 0.6548791061790498,0.15300639305552444,0.7966364222495572,
 0.47407820209728513,0.5967921152160133,0.34926977890212096,
 0.21299074898733872,0.35513243394580374,0.6787680349810897,
 0.8099925835608345,0.44629736977613854,0.19372748176697518,
 0.9141706683563698,0.8661227044493797,0.21880556858649836,
 0.12292640860993154,0.2256809407171927,0.2283461674395315,0.4535748120798867,
 0.9182118933954211,0.6444559359838773,0.19456700735932098,
 0.06973566890018323,0.34657728306939883,0.7755410820187437,
 0.8246840630018691,0.5709271082996409,0.3670915174280386,0.6471642085577174,
 0.9673367252045647,0.16620305849291805,0.5059632418501316,0.8814153984793033,
 0.8209293538767752,0.13420685337003835,0.4162937590639282,0.2288257542588641,
 0.7378241853897202,0.4001722138144266,0.6715251982682822,0.460788290334313,
 0.2988250624522857,0.39385856788778406,0.7941491857081655,0.8727115169056542,
 0.03242964317975294,0.79958359897688,0.713185150711422,0.2800169701328278,
 0.9370568583745672,0.22951316989895398,0.7390548226966669,
 0.19145064807476464,0.17742720571238546,0.6307449362347403,0.53584863909747,
 0.13254132993141476,0.04282764534366579,0.3506145083355412,
 0.5946056197080581,0.8022677011032423,0.9547008660129113,0.03755239276275191,
 0.7468842695782283,0.7858522916601686,0.41308064961556357,0.5125113950693307,
 0.7819558835767745,0.7021647969496368,0.27116914198586817,0.90890429447028,
 0.8542571887462846,0.2639002207895855,0.07102031183416746,0.8397062305233172,
 0.6637489979362413,0.9024576009757199,0.3521223116708986,0.4063231899921118,
 0.0098736255960864,0.035713153244565365,0.9992679894420876,
 0.2680587409328372,0.7903454609300329,0.049240580840226644,
 0.3680826030829163,0.06668542578002612,0.8039785179460764,0.9163669015027472,
 0.12303382483628122,0.3785331091020703,0.39898320791936015,
 0.8395535512157977,0.44450080830951044,0.008971380022158582,
 0.23364538878331564,0.23151779878554046,0.30219958576420125,
 0.05723050337429503,0.5609567651649278,0.44118039142247467,
 0.9042680136693981,0.7738356523191179,0.48207331710477086,
 0.34643490437407554,0.7584472247465583,0.9602203091588131,0.8136416443925776,
 0.4219355609647891,0.15945902258908629,0.5765508277631411,0.8723836747899902,
 0.22656415670564622,0.6696774027245815,0.2815718554510569,0.7201104513830794,
 0.5483942437899719,0.2410411653400405,0.9329837062015556,0.683469535371382,
 0.9901274780362164,0.5021869930520461,0.9015933744400214,0.16400726138476496,
 0.8500500905950757,0.35633605961357684,0.9004416694355153,0.7817012913575166,
 0.49148576587198123,0.33861276983072464,0.39259744827012977,
 0.780763956018597,0.0739517103595474,0.146679004720062,0.12579644571074344,
 0.1676651903360592,0.34239119180344324,0.7561915490347638,0.1371580262682499,
 0.9802126692441426,0.3063346926766518,0.9492847645435053,0.4228804756968453,
 0.41711576382230253,0.4151224209001181,0.992997915203906,0.8912331379243454,
 0.7365124221729993,0.2906788960286826,0.341699285390783,0.023563519227207896,
 0.25009718839028217,0.15030971463299847,0.04194277258114221,
 0.2922702319662581,0.6404430978107364,0.47571465977578464,0.9282479584102098,
 0.48956744173076916,0.8752210819314615,0.3648292672958555,0.8163202073987166,
 0.926595849411282,0.004886321737754273,0.3549298481381956,
 0.006772838961182304,0.3547309564348273,0.7384468838755347,
 0.30207104145652386,0.8433500894975882,0.7755664162069686,0.6295920595236205,
 0.32438006424085397,0.9662424901904026,0.2960074088524687,
 0.18081553045855814,0.35521476595384494,0.723431352669585,0.2845441435093151,
 0.02178792275952357,0.9712976800730379,0.29183011910606804,
 0.24014191150971798,0.18441974032688924,0.11911717534609068,
 0.9951340627068067,0.18288964480357095,0.21279424197018837,
 0.4249607885596194,0.2885634247613784,0.3060232913961113,0.19890070889392364,
 0.08233525775571016,0.47052359936637744,0.1559539919881472,
 0.1085869352410227,0.6894567816659447,0.250745153997573,0.05701572804934063,
 0.3007063727314252,0.2223166433136171,0.031300373367527,0.06664684199662763,
 0.8333209317906664,0.9517936825994132,0.858160185735147,0.4995597216868256,
 0.3965446301351272,0.887408022253677,0.1646832659123827,0.6928340517170573,
 0.12988632499523434,0.4133748405598163,0.9710605353575331,0.7995759225848814,
 0.31015504493476886,0.4277916604860904,0.8276807062903039,
 0.48449240618776046,0.9805153236861757,0.9401365308421171,0.6445534707092253,
 0.28345961530515906,0.054940220539281936,0.650039485411083,
 0.9689625777242132,0.38047722275210394,0.9046260444128247,0.4434820714355061,
 0.5202824081276317,0.9263431132351231,0.9452825571746567,0.7274583267006429,
 0.008752052543951505,0.9542625930378528,0.012606779394269352,
 0.9674236488914398,0.1548106846161279,0.08980629442647547,
 0.013421071419569275,0.6678525050619977,0.1278167724407162,
 0.9819278424196332,0.1064211013792784,0.6642232388086654,0.36347709802378725,
 0.23933757380075826,0.47814791548469515,0.8681044513389283,
 0.8577351331543683,0.08632063293233916,0.0980276302123052,0.9061838336530323,
 0.4453153503883803,0.15483613073678015,0.3863128331032133,0.9162777680211118,
 0.31473696608584634,0.971929981010794,0.42367263642995523,0.1537779774695538,
 0.1177385784820767,0.36461961547147637,0.14311570059621415,
 0.41288869270100426,0.422441080392542,0.904146956539393,0.021760280982106917,
 0.27685703044815446,0.6469822603593387,0.9921122887267,0.7708488900593538,
 0.7392214979893447,0.46023970186128393,0.2770749488137888,
 0.17807537466942636,0.11486850885966671,0.7121142922017587,
 0.4346700503614731,0.4340627557331407,0.5006331946381429,0.5665305936305536,
 0.24343550528166635,0.832309716504223,0.8952250903403186,0.6380639101972221,
 0.44362661532100744,0.14124562354966896,0.3839953556814606,
 0.4177556001381988,0.040146448770240095,0.406917880817145,
 0.05800987852956474,0.8575266004521377,0.613238986393995,0.8847492422272755,
 0.5459242983654438,0.23495740624115158,0.3976818904665814,0.2066721618711772,
 0.8021677079001471,0.6131777227922542,0.3773868966188062,0.7473722831970505,
 0.6969069813800773,0.5624276002060523,0.5161911441834482,0.7175946976789189,
 0.933824256502938,0.6933087717231441,0.015031121282882975,0.3535654393406159,
 0.9248020463726211,0.21828818752389745,0.16761299556981835,
 0.5397704549482655,0.2222729967121484,0.8889187071907073,0.4164465227381642,
 0.41004271984190677,0.2959577502259081,0.00981360325677716,
 0.8994240317135891,0.3408808384878962,0.1679876344405522,0.8225296010119321,
 0.0603460659363686,0.4237725668100434,0.8312054559023095,0.5534045598525797,
 0.5620267149647115,0.15874593059961528,0.27687744655920854,
 0.5332346152478011,0.1853109534677111,0.6748176495142396,0.03986755350796645,
 0.5697537372051862,0.6284907330376982,0.20460191202597566,0.2649838530296428,
 0.2321875547855723,0.3410532372320779,0.42081761656618444,0.7571040446087709,
 0.3547012816357119,0.9123142768096596,0.14224304431405577,
 0.05731688045666672,0.29509215412530143,0.4734781842571548,
 0.021205260721993358,0.5535216453976016,0.5509631037317058,
 0.7082896027674783,0.260019272756999,0.94966094472648,0.42673477446606656,
 0.016244391929045143,0.2740388087919451,0.023957518179234083,
 0.21096202417460408,0.18476603588010287,0.11607944257431124,
 0.3121557081888915,0.5179922379255557,0.656883442503396,0.33689186088877443,
 0.13863430292463574,0.2657002350425751,0.24011723165630117,
 0.9987897633556339,0.36697104544842607,0.40800505990333036,
 0.8214575188447357,0.7061477746355743,0.8581057512317226,0.9975179986424558,
 0.16093585939528143,0.5850305805704813,0.12889461978692607,
 0.3392258349253977,0.558725000077681,0.493785312658662,0.497489138157879,
 0.14585803304635314,0.7993023445891136,0.6481014874934707,0.2607201771712816,
 0.03678663879593991,0.4162386871600514,0.5725621164567316,0.440798909097019,
 0.504825226422101,0.9746297038241919,0.6869091811830366,0.2727478551822683,
 0.546999334321466,0.21619685146675371,0.5695645068223507,0.7928483845516787,
 0.04668472894430975,0.8479340593444791,0.7928441810206065,0.7618685393603679,
 0.6707747629065548,0.13524475354716503,0.1432457651309691,0.544843838348649,
 0.28902283521335637,0.3374641841666901,0.16020493335088914,
 0.4536232364497982,0.1568666122339708,0.7863911548948268,0.06075065376303279,
 0.5928052801050017,0.06046691442520924,0.37243034346668935,
 0.27374919824563093,0.7208362995018143,0.6178367469346863,
 0.24798287143421405,0.6890487064446075,0.40737682130135233,
 0.6798885357773929,0.7872710754574186,0.35066126224723093,0.8215543341137523,
 0.5130799136319408,0.04457388882203994,0.7342559124959472,0.2181310183798978,
 0.5932536974896362,0.4270095316701954,0.127084518630033,0.5144013615508263,
 0.47177869314229515,0.06188932913270517,0.745944792232289,0.6699986401020166,
 0.11242008871859976,0.7730523078172131,0.946185835743804,0.9424902280144899,
 0.6992017720912826,0.2741154607862033,0.24390177432481375,
 0.32664042105544366,0.05283562072293857,0.5169553812498918,
 0.6826047467191713,0.2994972696338857,0.3034289238938622,0.7498192720827036,
 0.19725499065132412,0.3779920649290661,0.1275731273267211,
 0.01723543960680296,0.44552396361493163,0.5019668798086474,
 0.060542350260205824,0.01600242353422965,0.9238894345634958,
 0.4631198975928875,0.9143274795698552,0.4358441446377084,0.9665489040668384,
 0.5521633652268694,0.10659128398637929,0.27128380828792054,0.946980382967994,
 0.9553909958689345,0.3995482465445459,0.65765563573575,0.6648235198143264,
 0.8068737512966667,0.6624292593383984,0.20948223004678668,
 0.25492483420435663,0.13573943485021056,0.5058807354210277,
 0.5142702378548336,0.21994825750404312,0.017057207410212993,
 0.25468517196530893,0.7228880083671941,0.6698927124869772,
 0.05741730579349402,0.2407240720884365,0.7404815498337082,
 0.08482146002223612,0.272490910834143,0.04798802914256939,
 0.014838811527327023,0.04821946059053761,0.34199051285222537,
 0.22433342672788192,0.4982202738010527,0.11819399601793423,0.928554490619234,
 0.9578882344358671,0.5261134920101449,0.924703328627197,0.46523627068459616,
 0.45393324818610903,0.2977566888045019,0.5884210123441524,
 0.36432248571256276,0.9381350094731018,0.6227538476858906,
 0.49438715300785574,0.9199133841097806,0.9557120727101988,
 0.004684263765983299,0.030649375219953368,0.512387616245318,
 0.47372570447041595,0.8280396740615772,0.21808441707500104,
 0.5521947324014013,0.9549096822823947,0.06603147010246702,0.7788925415719417,
 0.4321754313942213,0.47201940538344433,0.6482112589632991,0.6923179584210237,
 0.3710737752670572,0.19461430056146767,0.06631794369769739,
 0.6092911800699645,0.721345349795974,0.982817149260557,0.42516954689506914,
 0.3963911313752928,0.9969136309638896,0.949761281039813,0.9138091823787096,
 0.9327700054426008,0.7064499935527099,0.29729181376418923,0.2443509703553549,
 0.30356634285697903,0.04770662397745895,0.06999988986208816,
 0.6017857616936493,0.7985418094930641,0.31070860409210765,0.6728466903120984,
 0.3914570669042715,0.5463588698880829,0.6309511467461556,0.561006553268297,
 0.9885963028087116,0.08800573958564861,0.9412888859119063,0.6384665506786908,
 0.9879012113049377,0.46378050734127774,0.7630511263883198,0.6157716058294579,
 0.047172627942763,0.8691848857724823,0.5234947417833855,0.6756590982085466,
 0.8453816857228038,0.8716848067213623,0.9692060360816821,0.5905097776835292,
 0.30025142195014465,0.3443577259644469,0.861506319509567,0.4428708816520214,
 0.15819364103202394,0.17312858344187454,0.13578507574829757,
 0.33547051385236737,0.8293516274227404,0.23189262671726962,
 0.4556803744909659,0.6749367403195777,0.9602510960101512,0.7827735061067798,
 0.2509855919413574,0.21950008331138426,0.0345933504798267,0.1495571022560227,
 0.7936915248207204,0.8260049155714412,0.505211645229246,0.08836038417854908,
 0.24062323598639124,0.37415383193488194,0.2593179542862246,
 0.1612459384707221,0.6916184969392228,0.07642568503486558,0.7960385357333772,
 0.5794627370936785,0.66363202695047,0.33774866107177104,0.5499153099303844,
 0.017597502312552527,0.6756314717310271,0.7075806728314779,
 0.6280975931025867,0.7146018192293003,0.8744000185395019,0.9745990673499871,
 0.3349801067086168,0.740233101297882,0.11079308866835991,0.5789906845177121,
 0.4292937069909821,0.013128171397882693,0.8528792091280505,0.713310860903683,
 0.701571117898725,0.31197645616419456,0.7379580009627211,0.9396233470527474,
 0.6191436679135348,0.1882759226844829,0.9633012159933224,0.31262374779385416,
 0.8745512393017633,0.23617624803393056,0.3146825578938349,0.8384235044876073,
 0.7317862904309891,0.7798171051008903,0.5936739803064456,0.6998973097166944,
 0.6873849382553157,0.45067734227647993,0.7874121274752527,
 0.05340203953335976,0.925040606547771,0.6688105726836902,0.14393046906131512,
 0.06613782644480826,0.812761779081797,0.23659736397619535,0.9830587590016433,
 0.934744312721369,0.4568769176892684,0.3080251896477012,0.4813617983922447,
 0.9537486201571345,0.16382969206925746,0.8936912051639421,
 0.43061331950335635,0.7506440128800744,0.27657686872392073,
 0.8702964324972882,0.5861047229524722,0.9861358742440037,0.11104422744791176,
 0.35714157903994903,0.4776632685147646,0.5494855888257352,
 0.22188450169296947,0.49465252819391514,0.6295208523909602,
 0.5947557331335959,0.6270669581424555,0.12677070861396977,
 0.47844861495713664,0.30716315547307416,0.9842488619934557,
 0.2439891656469222,0.17231133784197616,0.9451373364091226,0.6946289635616051,
 0.40320723419137294,0.14015551711158425,0.07601170518127576,
 0.4295330897074162,0.660028499953683,0.6721596521841972,0.4473237231083089,
 0.3605968361728632,0.09103966104350214,0.5458597668972692,0.044633553868069,
 0.13193518687280037,0.45230467663375185,0.9927892722498969,
 0.9964150120487686,0.0911811208898855,0.8698456963950622,0.9530644181063174,
 0.9561040365362126,0.7354205807594795,0.847909948810512,0.7724808031908845,
 0.4895053932266711,0.14522242195960144,0.4988785282113909,
 0.06035709557670954,0.6445566003092884,0.3130313647332196,
 0.21271033492101232,0.159060762661429,0.7549244271231234,0.8817963189438766,
 0.19986913019624153,0.8376071592237548,0.22974202933647936,
 0.9683829327569572,0.039378356922519675,0.7480477284357538,
 0.9311383810360165,0.26312014117067273,0.30289644816930683,
 0.0316299492833767,0.20970558466939504,0.15051446083967224,
 0.39110032236059056,0.7585295407900082,0.19298300555617642,
 0.9677612760722947,0.8369844066694196,0.7559320493101777,0.40779399101136526,
 0.7924787934055554,0.24844626218242838,0.12161309965183364,
 0.6762335141920649,0.9855439921358818,0.3646132955306841,0.2643620957365991,
 0.8466616288136655,0.8684067573389223,0.7668070928745009,0.3311249173040267,
 0.9277811423767086,0.2789383169871623,0.3333031503199033,0.1479892480986884,
 0.7982048279296706,0.785031755992931,0.8770396230151687,0.37691102695010426,
 0.6936575625604703,0.6631798131674954,0.7085784455361401,0.5323955405491742,
 0.7226936184321406,0.8059701468685716,0.0037048400100201384,
 0.1235607691740278,0.2079976898110396,0.7099763920701825,0.7221902896853766,
 0.5560758488285089,0.5445342701520035,0.9071561200480058,0.9958325222998203,
 0.04720969009781317,0.7265183788183847,0.34115893338405545,
 0.4452037727691338,0.16187413390429772,0.34111564974978714,0.942736034377592,
 0.29938787031953296,0.02916940799689116,0.5769188443544583,
 9.902118651071712e-4,0.6260401340767908,0.7752306073886177,
 0.9987999140464873,0.13569749849762158,0.41838853604640114,
 0.7440387353410897,0.5800793721823821,0.6046048976827038,0.78494416733704,
 0.9176657267842049,0.5273043716663438,0.32433387609572084,
 0.06741996942571082,0.1461071038479026,0.08598152249537849,
 0.44645260085075833,0.35578861902393977,0.9816295873946665,
 0.15216239086170358,0.15408175642327726,0.6750283459281928,
 0.5227470152589409,0.7195153617303964,0.13437072357717517,
 0.25707001282402064,0.3293680410227602,0.9251514523059147,0.5657224182274846,
 0.12413013480076984,0.2157592631497911,0.9908994732778633,0.3638771201283887,
 0.3892960274591829,0.6240725086971965,0.8067883216576953,0.9242884435909033,
 0.13717597529679204,0.6462789550860549,0.6683847419193397,0.6403766164543293,
 0.5876402154811897,0.9815490105015207,0.6986520799472392,0.48083216850718313,
 0.28567136330834475,0.38579782082372993,0.33101756066307997,
 0.217633422112683,0.1389470805625599,0.7748383632167498,0.5071732461956993,
 0.6576313296351995,0.461602211931204,0.9357240488253665,0.9200183361102331,
 0.07152251312173474,0.17482173902363796,0.746701006181141,0.35226380858784,
 0.2824836825239905,0.17486061987420687,0.3934859957965683,0.4228124235046391,
 0.04115195164499097,0.33793692505878536,0.18262774946338922,
 0.424777241987066,0.6738152139171691,0.28270787551118604,0.7767784088049129,
 0.9019589535265414,0.3149844199674946,0.6174081652336916,0.5934017463653902,
 0.2776340471757578,0.4854116152462502,0.7912956689643337,0.9932886320160738,
 0.3304550312049551,8.738496502707083e-4,0.36779028791725354,
 0.0933059317045335,0.06166127294262114,0.8260605813494144,
 0.14208448309628874,0.3324835331167161,0.7132187800546951,0.9634991666886732,
 0.7861231322812527,0.6743391832161433,0.18989736367132803,0.9681309872530568,
 0.28110289736865623,0.025051399338495317,0.04966462240341363,
 0.8351594951639998,0.5943740019314218,0.7579762057564338,0.7253582464784264,
 0.010683439184086985,0.7212605468473016,0.8591041707301191,0.537904960132147,
 0.5030310984940715,0.19300113092942384,0.4322557349384689,0.9516939606672572,
 0.8581545502524053,0.8290712900549321,0.13682100974950773,0.8198445340672235,
 0.8459450930862441,0.477269744923175,0.2747039834887012,0.44276892642946974,
 0.374719046176866,0.9504492091180079,0.7442497402243816,0.26560099817133964,
 0.5774911931963357,0.07005986631592798,0.31280693089281153,
 0.3348129017289815,0.5398657655188517,0.15819612624732904,
 0.09483656770192828,0.49621708811611276,0.19970784481791448,
 0.05162257094166012,0.17072697553020044,0.5361448070907013,
 0.45621019995060397,0.636648933570439,0.5165731441475755,0.5208910761319928,
 0.9447592250520652,0.594465931457372,0.3575862891155748,0.48456450860961686,
 0.7991290542005585,0.2552743826316751,0.9578508352219465,0.11415004396541517,
 0.31177201225346807,0.5131624480285512,0.5964619512550499,0.791734710711333,
 0.237837982341497,0.1612052676891087,0.3533057672491007,0.32195751251480653,
 0.6711017562715471,0.025983291721381474,0.508118209164287,0.7875359484016065,
 0.33473845257079304,0.9764402011547041,0.8997204522490141,0.7662649805162843,
 0.06811763504291968,0.28908288859911013,0.9043619613788414,0.612327423584367,
 0.8634148758041883,0.1297787298459705,0.6433457176920481,0.20688463872597973,
 0.16878180010137522,0.6091922687825553,0.9914814885905798,0.7126938767296985,
 0.5357113194863758,0.8858734209835978,0.8280106696919872,0.6756632453015091,
 0.9890047936714218,0.8421256858291631,0.7172585731519484,0.48882964988710975,
 0.9144642130206088,0.7591256345505111,0.8640177360281365,0.5095676912384626,
 0.021283920472920492,0.06216687006545052,0.34846923909591476,
 0.8049248253346022,0.9725014266605442,0.27232937870706886,
 0.06168575893789097,0.6196972638372347,0.6852165263496796,
 0.04483271578928538,0.3655010802969102,0.44708024088612486,
 0.8369131413143562,0.4193962210531551,0.947141069370908,0.8665902648404555,
 0.3742270327584719,0.25612067271913974,0.39269374567868454,
 0.8343763927065977,0.40548282079223485,0.6993561132419015,0.5164667130146784,
 0.7921214770835846,0.30863516624980647,0.7730277310003069,0.5978463895214542,
 0.7179580886010303,0.0388324013793272,0.9524064211959294,0.6555875365076721,
 0.918981270533685,0.6952087348564873,0.81471623504513,0.9088960604730255,
 0.24968051638308808,0.2091984460365761,0.8134025798203928,0.9112017870061048,
 0.4868025699565963,0.8476806688746873,0.13616286122739318,
 0.15571310720603526,0.6523249268705062,0.1669735621216345,
 0.02887793559362639,0.6454108723476046,0.9193653005985876,0.5248976285141882,
 0.6616322930346257,0.0035515102014485045,0.1349775966904354,
 0.8300238390340862,0.4049755536808135,0.750335005440854,0.1481964698430137,
 0.3529423535971361,0.22193499992582788,0.08579648637985515,
 0.7298765498526257,0.6868359312155952,0.9713498258707537,0.2567815069340339,
 0.3108110689258834,0.371312488575786,0.7407104518801861,0.12055653910172259,
 0.05085098667329746,0.42620774886472323,0.7522922256569567,
 0.6712497091587406,0.7612013551065747,0.44755263268217904,
 0.04532705903820333,0.8590305857189858,0.6694678015416817,0.6893562751987343,
 0.4632601597626831,0.8368427466943212,0.08348745447480177,
 0.007961419241698908,0.74832190235393,0.7783955540931103,0.535282733591817,
 0.6879023356971479,0.8280547709751851,0.8576775964014357,0.4750447326413032,
 0.10138874668327191,0.30772520414516513,0.6764669590971212,
 0.2556720856737196,0.6715600014281758,0.6578136640177139,0.8101088057196438,
 0.8052964907680734,0.6690242959947282,0.5298700705995304,
 0.007377925424853382,0.21397464657036724,0.9564315857139194,
 0.6451462444775508,0.7003757058852977,0.20701207270589328,
 0.12525916788988334,0.40369413051919656,0.3152057309718951,
 0.1105479878048442,0.11707449348126164,0.500092361802919,0.44900515142817243,
 0.016113937858928673,0.14373897500512534,0.44446925757186084,
 0.4802863304260856,0.25674303299196544,0.26623140359951436,
 0.9854860340037611,0.7326612802063374,0.6629615378033511,0.7800125475375312,
 0.540436782331743,0.5153359803285469,0.05790427652262542,0.05152887018621344,
 0.27353404192192843,0.7477139075625647,0.1062268935727696,0.3171224264518866,
 0.9682380998856395,0.18547448527575539,0.8012904807995102,
 0.49408440745047233,0.6304327576190581,0.8202528311986088,
 0.08062024675500556,0.09834630627408536,0.3031485680978916,
 0.03952706715333332,0.9014381165865754,0.6585363575490704,0.8854145425475194,
 0.21184570862180496,0.3725314566529063,0.40708930314571934,
 0.9279937812116491,0.828541660021289,0.9961665786213992,0.3230933522769851,
 0.89305313378713,0.22853739336462287,0.047115472209305564,
 0.24891479343329548,0.056625826720973604,0.8758826843096085,
 0.35985740923323617,0.5267645070936646,0.39491664320499575,
 0.35900670823664627,0.3048483936855091,0.5732643885137375,0.4982182235631848,
 0.062043264288104694,0.08125074489856299,0.3342586656765264,
 0.623811878215293,0.211586374277547,0.4155772456302048,0.07616284563053666,
 0.8108444509997166,0.6207556618589118,0.21740004770050447,0.3086948330450855,
 0.6752580291557605,0.4347929159603007,0.9251800448040671,0.3412820185423757,
 0.9731398955408359,0.946715265642257,0.7020297412009653,0.31375450726496645,
 0.036085825404169825,0.36516310412594566,0.6770789565293053,
 0.1376270690426653,0.5417754789229734,0.9114071573822331,0.11697635531903561,
 0.7258248933200386,0.5960012282453243,0.7700448920595118,0.8790551096539463,
 0.38545899613564794,0.10074972202273103,0.4372242109011151,
 0.7292652733362136,0.9870805894705638,0.9513570231802179,0.11401988394003149,
 0.1599652720602669,0.8572963563511053,0.21774427705738608,
 0.04288389880564525,0.5040964596383974,0.5754422440580318,
 0.22426641083977827,0.2820678358007973,0.34119148828820123,
 0.08832933256800779,0.9399020832570352,0.3907052461694819,0.8348017436385556,
 0.27540282795723725,0.4979708720522349,0.40511823667151436,
 0.5370698684831731,0.6385383822339782,0.9692700774316865,0.7983298815767913,
 0.6307423361633938,0.13435186757432815,0.4020395233112327,0.9024741593730756,
 0.07587706186118082,0.8355509609209866,0.12281859232040637,0.251027640570752,
 0.08260255298261343,0.9553679994783802,0.7114521928551046,0.9683720334134456,
 0.6357978019536847,0.41408411747515317,0.3341935607250728,0.3100623564548014,
 0.9560394376007639,0.5629053321884416,0.08075144134472811,0.9754674764107402,
 0.7229811023737828,0.15967201836521072,0.855075169122986,0.724178203122703,
 0.9065253904270358,0.06140359799553241,0.9182450023263606,
 0.43742636046894035,0.5004819227613342,0.17629090599902342,
 0.3150552428736968,0.9630020310445222,0.8599720105661243,0.16935382787448916,
 0.20615529226676998,0.6203659000790253,0.8529301145630631,
 0.011880823911827543,0.3004590934000374,0.013803770689518213,
 0.6709763932570363,0.23972538019757483,0.6347252860767248,0.2787526747007236,
 0.13625402199903824,0.24112175983152861,0.2484110262104925,
 0.5533681855488908,0.30611234119238473,0.9048250470833424,
 0.21985683631904795,0.5552149462773808,0.22222379005479498,
 0.43649967608780726,0.1140421610677409,0.06233391426998469,
 0.6814090853014929,0.4872905036613727,0.6280222860003939,0.09198047616655225,
 0.7674108654065952,0.6331006158717162,0.6369560079388341,0.62123915525539,
 0.5856913203919522,0.18378997107114114,0.03903015615036065,
 0.6939717586602872,0.4704994114101675,0.34830408128211243,
 0.041135360777854624,0.41147409802311297,0.08148444907132013,
 0.34191758594428723,0.568101865969918,0.10956264850752928,0.8471098993916658,
 0.9232806104333733,0.293756160668874,0.4829135043195669,0.6689955346569056,
 0.2949214500664188,0.6080384699552437,0.8455050204346948,0.6504662819975269,
 0.8203871247116838,0.8367974341183158,0.5620721428027308,0.44360824788554776,
 0.09353945944573594,0.5667619044081454,0.5468392466565621,
 0.43247075002021884,0.826698899128724,0.7158321763568574,0.2576225748962597,
 0.9233643888565872,0.7471923155828106,0.3512612998550194,0.8448628612902113,
 0.8249809789999328,0.48295925102548787,0.4788678118018753,
 0.11460644331822367,0.24593837715013334,0.2153115352000058,
 0.17276165841628233,0.47099650391817105,0.3877061767674562,
 0.00369742304805043,0.6044743805497937,0.6228637203649637,
 0.061074696446216215,0.12518653876256158,0.32211358741768503,
 0.4021324044686818,0.6951609244861251,0.18785320340996836,0.7342427927018486,
 0.5443617214876018,0.020846978928588333,0.11952103247430557,
 0.3633615922493626,0.8385021159344697,0.03239622591491065,
 0.40243037491325007,0.2190242837857963,0.04055532758513969,
 0.19852700600849715,0.05046433449936982,0.957481417157409,
 0.018043768988708297,0.44920880868287283,0.9175075628063109,
 0.010558545188649182,0.32231804799485675,0.437413022131136,
 0.7667713345394997,0.07889178255547913,0.21038827164377627,
 0.0730039883595156,0.4735630373223144,0.7492340229774458,0.2356600026636113,
 0.08740901185394512,0.07800481873798393,0.7008591790250848,
 0.9270047902532025,0.8365335938694053,0.6288398329302869,0.12288300508987904,
 0.9097859179471137,0.27683998035851687,0.06221160611840837,
 0.5636732432103888,0.1492433095776411,0.16852890218137517,
 0.32703384793238666,0.7473264000410437,0.10074181183928832,0.902364642694747,
 0.5882221343078483,0.7509972748681712,0.15361148668354718,
 0.23967785990825496,0.026860666060285587,0.9668902724304471,
 0.3057539661651395,0.34016633179696687,0.4654642909043586,
 0.18997709118416872,0.36007895655771116,0.6091412206170654,
 0.9223926998770314,0.5027129332365993,0.35866076785892986,0.9248823759226678,
 0.5311801576337863,0.26165745997062306,0.6350037980291112,
 0.26095908537364343,0.5126989289128904,0.6764162983476224,0.9418864222414197,
 0.7423013126156836,0.2649180589869756,0.6134371447804057,0.7463169038395223,
 0.6901631653062148,0.2225086302796666,0.7534540155397944,0.7380108899521,
 0.8035552057327304,0.7355877951176572,0.2804810954978424,0.3797334873575683,
 0.13953714069107392,0.40442610748997254,0.9989344751241502,
 0.3376668257130724,0.5702433977419104,0.19732064582897235,
 0.15221101068761178,0.38912959135430203,0.21782371668536726,
 0.7918325408007114,0.49861128224118734,0.16255428147950968,
 0.48782427704609166,0.14692496162049307,0.45081860647142535,
 0.6496586678213218,0.9361630772212326,0.34142219646822036,0.8828387136707949,
 0.7105853811615058,0.14619293127096866,0.6444287331826303,0.7164754155035149,
 0.24261732895086774,0.9236208074429928,0.5966389743905904,0.1005145376721267,
 0.03155805466927564,0.25883203097096685,0.865123087223126,
 0.06626303707216863,0.02792282147543479,0.5728490516377756,0.534003751544537,
 0.36200132880386693,0.3826730070722515,0.6277281115265676,
 0.40038221122777007,0.33257974911920807,0.9902898580376167,0.752838302037598,
 0.2580023305593546,0.8120198684940972,0.38614017368278597,0.8662241162996737,
 0.8475122445320364,0.5988461254806816,0.8931157476504265,0.15409606764618322,
 0.9712381137328668,0.23978398521848587,0.2182524559095721,
 0.06747783361445059,0.3824590054418452,0.39459290271769587,
 0.5674025884596327,0.41991057835644896,0.7053387735990427,
 0.15379950228359573,0.04761636902010491,0.21731648445599394,
 0.6720010908194187,0.09789667617887621,0.9398460815915237,0.7517696823829815,
 0.24279592630568947,0.9061654277178232,0.40071802394793743,
 0.3198004568235452,0.33672208054120845,0.024718559163751874,
 0.13370618086261044,0.6425772591868608,0.8731636650384359,0.7139279968183854,
 0.3303705111455254,0.3626512934635302,0.8895389079619305,0.16656904457668442,
 0.30380553736436144,0.17615491372701286,0.5281931937200786,
 0.5136745417695996,0.5107201215656487,0.6353578924558305,0.18034477267679994,
 0.8543810450770644,0.14467720756568303,0.09176495908423776,
 0.5831033947694595,0.04885673582100725,0.9676332699876734,0.6597386879796564,
 0.12050847126055309,0.2774007250014884,0.907606784374208,0.848325369178702,
 0.26737241554433866,0.49906283824879816,0.6180550748089042,
 0.5279462257284084,0.3345436224445153,0.6186604095222371,0.3035542809118963,
 0.6310185519985285,0.915911958522855,0.645887456886674,0.17457190421720936,
 0.19691421036056633,0.6480064778085535,0.12386690547627588,
 0.36973304464522905,0.813599211290893,0.1375109748377491,0.26809059269649504,
 0.9400471348262368,0.4549582252811434,0.8031426321871522,0.1858711994039246,
 0.9687109565467875,0.7338191220128832,0.4646331926026468,0.7566864785536583,
 0.40320630154617343,0.6227357098550551,0.8747086394824828,
 0.28363920217223193,0.8757020311914697,0.3633738964621198,
 0.09855484651699276,0.9668791351416818,0.5276165009242603,
 0.051996452367986556,0.7749336990043196,0.2160261086489551,
 0.5433534598128222,0.640178147996103,0.6536783708554887,0.778725462336524,
 0.6542984397773536,0.37276182333341346,0.15918700440229372,
 0.8272770580279737,0.3422006914046274,0.20435382540392222,0.9924567385252626,
 0.7187577769802569,0.6082186318061581,0.4833079820201093,0.8616723297066816,
 0.28765113457486047,0.6707506135729776,0.5436616338405625,0.7623164189636443,
 0.30103164819120276,0.6903304314588689,0.6893848333004294,0.3419583138200242,
 0.5660273797789529,0.7071352699565516,0.8523894019594912,0.9405062666224752,
 0.8610703138987332,0.6720845716453689,0.2434574667816728,0.8296534617424821,
 0.4953865016533434,0.1988797155554135,0.3390543512507077,0.9768883987462466,
 0.0343555872993353,0.319181550670129,0.045481254491466006,0.3477457890741146,
 0.0954467089095774,0.0025999306275019674,0.9782968753862566,
 0.23412662960626052,0.31184796869874265,0.6475071103330281,
 0.9865801989205194,0.5437775520896886,0.2898549372697581,0.3919502020935396,
 0.09355175491672452,0.046640688886750215,0.4088413894351848,
 0.8196706067873619,0.3334670528711724,0.710110548434365,0.3157033884126639,
 0.6318766052260307,0.7973649915652034,0.5440346653192824,0.7932190837225559,
 0.29336649554078786,0.53952095665421,0.7020105345402182,0.12912073606209218,
 0.2693579384343028,0.12554155873216244,0.7740327172784958,0.4011236636450831,
 0.3865714855563458,0.2399205865748678,0.3407180378714334,0.6388759874481698,
 0.31571604339519066,0.9039635452268069,0.04350035569297661,
 0.6842957016251144,0.20527008469575192,0.8945799885101477,
 0.45249908240836456,0.16116501541523354,0.9123259109543694,
 0.17153791484788128,0.7372245445194344,0.5072926822998458,0.3429073285629487,
 0.37203252398141107,0.6402099121183422,0.9400500928627593,
 0.27133021103711663,0.6986488482515593,0.8051184975303036,0.9506998139688886,
 0.3509108556959786,0.48454890466613143,0.1451102443849014,0.7431122817690435,
 0.2844330447369915,0.18591329387117572,0.7652254649157711,
 0.14235486037690137,0.3432057286829612,0.6236433395070109,0.5069047330031913,
 0.3371363188246129,0.7144200797216301,0.020656895217444404,
 0.8140268465644055,0.09979302409715629,0.6567093172500548,0.1481650543293076,
 0.2175878575081791,0.8012964755465207,0.44419460406689226,0.5832321119150766,
 0.7929025750573959,0.9421458621880289,0.4628598248704012,0.5687152805439921,
 0.9034186491039141,0.23518853658543637,0.022307550573556933,
 0.6884612870372416,0.14020365825138614,0.23204348303080757,
 0.25285332910270664,0.39366551349717804,0.5033546399568386,
 0.23667923204354413,0.6772064345974905,0.499911718997733,0.3300162204109939,
 0.013018408144722438,0.8965346138134386,0.08838693258426933,
 0.1900618505233872,0.1780569222532169,0.8113146971425332,0.7144373480627625,
 0.042438419775226666,0.15090352398091933,0.12987304203611383,
 0.7210618530310597,0.2604221970186479,0.4217007119595888,0.49504929952940424,
 0.19073111893576378,0.9863207927159031,0.7097027636711463,0.9984561204969458,
 0.8229304377535932,0.45912214702412113,0.16892499872347022,
 0.7275713247586861,0.05491776799622061,0.16805167204721372,
 0.6212452938044124,0.4319659204813454,0.40219594408850035,0.7333108741246197,
 0.671122101468231,0.24002748197580304,0.9861204050695305,0.8241901977646114,
 0.1212544439582457,0.03318230511253373,0.7298346934865604,0.8771421173326027,
 0.6438653980584833,0.7980247237709411,0.5994294152437616,0.11807106483818408,
 0.9794872430559043,0.6361612456304553,0.006246592478677648,
 0.6504670798710619,0.4863250021705957,0.6883067283212907,0.9640024134329206,
 0.3402308758839283,0.689582764199943,0.5129995428205333,0.6125552248303237,
 0.023101694838914777,0.8090087495409619,0.17871031793354264,
 0.45085789866864223,0.7516848861821042,0.7664135197118969,0.9460983954928021,
 0.24727725415607216,0.9128484622914711,0.0148396159838311,
 0.18337004685700775,0.5478436578893888,0.4470945144001274,0.4733971231667532,
 0.6271033950534465,0.2766496517795779,0.13073502116987668,
 0.39755857515766957,0.8999912290170576,0.008364135466786715,
 0.09607911114727563,0.6348135767855794,0.06187235731535623,
 0.9672010674567666,0.8815881838217844,0.8721694507880723,0.4751644285362331,
 0.6989574364402232,0.7334535932222288,0.39454049192498386,0.6644853992853026,
 0.42874216450656166,0.4622640986290548,0.7700927729429895,
 0.31337492582148707,0.18255362587576895,0.08156291234808977,
 0.5803463461615787,0.4279017695900317,0.17866907640254237,0.9548665541906689,
 0.9099114572365792,0.3738100768278214,0.039540986704796444,
 0.9946599728088475,0.7817671756515512,0.7298568187828461,0.8522904313452742,
 0.6319433830431131,0.5210437083779771,0.5276806769665715,0.5924419304997216,
 0.730975502693144,0.9138262743349802,0.1277645761738534,0.03241961742788291,
 0.14374653005659543,0.2574718203339025,0.3385861637611245,
 0.31165679563738846,0.10274197090336223,0.340788339403195,
 0.09438888318891125,0.8320520354363383,0.2249006055261905,0.1291246509411067,
 0.46610765105312746,0.269990891302058,0.03873845617319849,0.541519036607422,
 0.09516023768372761,0.3378108122414497,0.9476649787408679,
 0.39839093883054977,0.7166850009454255,0.9896726157546397,0.1537011310793639,
 0.851707782516216,0.880758127402164,0.9086186956336549,0.5833502494818107,
 0.6456548101358489,0.7513146122322026,0.8065972385301622,0.38947547536450666,
 0.2561969436288143,0.3798339034429363,0.3788445963744995,0.07091496350666038,
 0.997697253458523,0.44340346812933085,0.606285449123801,0.6362370177461146,
 0.5886450246696525,0.6863144802636133,0.8622401460294128,0.49033622735094995,
 0.39940145108093705,0.12729177601010522,0.5917097870005236,
 0.2300065918779738,0.30479503746881154,0.7175299139155012,
 0.15710937445507278,0.07621481522735807,0.19146221244923445,0.96375915516809,
 0.36804914016025647,0.5967632803235228,0.0843843493700468,
 5.899832427144203e-4,0.8487474482218896,0.9803711872328822,
 0.30869966417919514,0.30379454415330587,0.7868087790141627,
 0.3072079260901638,0.7671606050820665,0.7066474913175902,0.9464050101480636,
 0.15967242342074361,0.5702008543505515,0.7022517792859869,
 0.0040046274463405584,0.6509926813218703,0.7812126156069541,
 0.9029459496179761,0.7813689573354661,0.8878042690485394,0.5805024699454274,
 0.5158807410545148,0.03227799757773486,0.8448202344421479,
 0.05853189699257877,0.5694512599781933,0.3671501298272327,0.8341321542145864,
 0.11061331895927362,0.9159123693729357,0.41738912889360946,
 0.7497402786762812,0.6003376804562062,0.44685990752645754,
 0.23996085061687067,0.7920323605799742,0.04889214125706054,
 0.7973043967669043,0.21353584508569434,0.33802894881512,0.34928840362293334,
 0.989918943598832,0.9980594170758317,0.14182202377043984,0.8050980598640537,
 0.21335706947760769,0.8736776693323111,0.5500511389929652,0.1385495783736591,
 0.9482356478113989,0.6413793784744456,0.4264653665015773,0.15600559530127822,
 0.17432397043328896,0.3880942720169499,0.9796414825014553,0.9115994003213994,
 0.48283119117862006,0.7363029190406407,0.6654128389409405,0.4514586861501536,
 0.3844861920296023,0.8017385483548497,0.9077457390295499,0.18488045023885913,
 0.4075213168393599,0.0852842417900026,0.3095772556895693,0.8196132054848517,
 0.1476181248584435,0.4224436876835892,0.4131775901885225,0.9057983728072355,
 0.12900596862098856,0.2488984791380715,0.1259928034320309,
 0.07606473204212627,0.10958792613704404,0.04855866046889379,
 0.15630849569086125,0.9241654862879095,0.5043822942045366,0.6662200108277345,
 0.024490323084206955,0.05648637105276699,0.18054051345229705,
 0.5615505852314847,0.8477717895769415,0.0278742902807928,0.7214065450762434,
 0.45194237542059,0.4116522218029467,0.3460602477594271,0.36057853625401814,
 0.19859869294907995,0.4538218556190555,0.43099507756990507,
 0.21238219421432847,0.8293130892536258,0.3924598951811419,0.7953646666863989,
 0.12008970734647273,0.2074181082668507,0.3668881846559571,
 0.18053168005698994,0.9866820939382285,0.8292561801087853,0.9625349722465575,
 0.4024118970368592,0.8813219344471692,0.5570919184487153,0.03171128480073193,
 0.00989241355638748,0.6218857104561655,0.9883342819830523,0.9666768647467857,
 0.07058926426891843,0.3311694277184154,0.26955880909995256,
 0.061436796241865155,0.5680063327465865,0.5085324501469555,
 0.6320656088408343,0.23264553762298756,0.6430464172215883,0.5457608338437794,
 0.2342646719859418,0.10326532367106622,0.9435480941650014,0.6254688516122237,
 0.7271607212881865,0.7876443157564175,0.7818395821352958,0.21320738751555923,
 0.6834382416160275,0.9739271691028408,0.4624498188214998,0.26660719451863535,
 0.7295861725043804,0.4613033252828551,0.9070284205112452,0.9564663438780818,
 0.04984926984915994,0.47573742193700874,0.6451451689578598,
 0.06237138979521184,0.9332173109477847,0.7671329493351252,0.560719852081431,
 0.4763876846001788,0.9364183765023762,0.13719084572425944,
 0.19864748359256124,0.5181285233337535,0.7829897951944976,
 0.19453198826379325,0.47932431574095047,0.2745491670560194,
 0.041837091364648304,0.6466960607657907,0.7934203255075621,
 0.5718525486148283,0.7251658907201413,0.5114218182598496,0.11499144785194648,
 0.272922873506412,0.4341890490307042,0.03809877503063852,0.256873824137677,
 0.36557656178421427,0.6819845435127329,0.5303096375085685,0.5590505885067379,
 0.5113625085678176,0.5727357798596602,0.23882312617029178,0.0444557141206694,
 0.4508852383870734,0.6872514777970194,0.2953064768193736,0.4473450634800502,
 0.6169794210028853,0.2314447488614214,0.8873907141807263,0.4646786335389176,
 0.9471743541394612,0.8918607727142914,0.06007494165986138,0.6289244736375552,
 0.0893600086505485,0.6923728411484098,0.5717718881310878,0.440602999921758,
 0.2324004524885077,0.32315040108254434,0.20577877646731568,
 0.09937653206902575,0.4995233100714036,0.35633537777752,0.5942920918897505,
 0.2590389732897752,0.5232703210971413,0.4974519254678387,0.9931518474150653,
 0.49927842960233393,0.25745876145307867,0.49079361193717286,
 0.38755043690578617,0.5688511316119088,0.2623947775144564,0.4384861870189436,
 0.041105813347856124,0.47739013079191106,0.9061461914536841,
 0.7431104889322179,0.4978269716607391,0.5922096406295628,
 0.005609985779838511,0.06572716052619809,0.33590057079168156,
 0.08135813270173198,0.6908727746334563,0.8825327972073242,0.8042070579563053,
 0.032639555322368397,0.6208346336590935,0.040018136575259766,
 0.2176265647474911,0.8058577570688279,0.6591805657524925,0.5525716118004066,
 0.779190817306016,0.9949400902055382,0.09351093082566198,0.8382589245970383,
 0.5120580970745824,0.23179279527539098,0.6546572481739829,
 0.023073758622245033,0.11953624545530195,0.6396888457843614,
 0.9474962080442835,0.22100347152834687,0.0376968441189619,
 0.10259183503648661,0.21226007077143827,0.6831608918631759,
 0.9461838430921312,0.022572690104935456,0.9952441769513904,
 0.11169351437809172,0.8574159735878905,0.4630713055328417,
 0.45679243651594525,0.3628602348834481,0.8146657201953167,
 0.13924818291437357,0.05638420250829068,0.7536855117765833,
 0.7500590140399986,0.0428083050823167,0.739639646089015,0.35869036977331326,
 0.9320346243389361,0.846661416131589,0.916450790400403,0.9737926515565696,
 0.5528813143693017,0.5837902817042535,0.10731142183908682,
 0.29862913718324235,0.9959542203173031,0.9375520472647847,0.4212242534384446,
 0.9447629687491659,0.42376345220115186,0.5114949569775843,0.495179173851348,
 0.34127910741600487,0.6928102362379828,0.15414205818349835,
 0.0056536469364308894,0.6425959549886127,0.41717789921404114,
 0.19700484015933528,0.5068107389989365,0.8184947243219826,
 0.21468243028665634,0.46283335565153805,0.9975018291081695,0.502643996721537,
 0.9614571917268107,0.03724248027445465,0.6800086271481373,0.9448513052520708,
 0.756345189558415,0.319042668569687,0.11453610551740068,0.6004349818964567,
 0.49218436532725995,0.5523958544552928,0.08959243948840112,
 0.3211463498777416,0.2828975392511476,0.2691887645583956,0.6419551173596907,
 0.913145030653329,0.9783177143102136,0.18834635035902514,
 0.042300589098511976,0.9100982374693135,0.1355385185997019,
 0.9008092958617938,0.11298572670716234,0.9023814208461491,0.6293989672003667,
 0.3339906140419103,0.22137467393682086,0.8101520261920473,0.8116689878055873,
 0.29588525673402444,0.5947873692394772,0.03555534131046745,
 0.45831970129760324,0.17253442565253818,0.467339270408757,
 0.04495936108342846,0.11062137713989006,0.08469390994793624,
 0.48234647306882084,0.6497131019324152,0.013171901538040487,
 0.2565117324117516,0.4585102589257152,0.38624332145813733,
 0.47130441037131676,0.184654167919312,0.3458547573590758,0.8366631711859194,
 0.895927586042933,0.006321354744130425,0.4656403718737455,0.7012202039534423,
 0.5908566504515254,0.7179106084387127,0.979453523603976,0.8758976187846884,
 0.13964963088417937,0.7954152711055247,0.15126355735954755,
 0.8165459168458754,0.3296964405939451,0.3772734243294731,0.13084988097552464,
 0.06324593338365891,0.07431821349534196,0.10238921524550337,
 0.015389563867663325,0.3719866123417773,0.571913088342117,0.4855641135066969,
 0.9210860634117456,0.6327826450926393,0.9242098972244038,0.22291098645027707,
 0.48111399260181587,0.5669268740288601,0.9451381301853977,0.1126141750518519,
 0.5685092743300031,0.6239220406602887,0.26068748324343183,0.6942225123923734,
 0.36155223170073103,0.4801698878723586,0.04162745765164644,
 0.5915035052899742,0.5263168240868363,0.38862209505859846,0.5315402946683488,
 0.24071553147593194,0.7249540622263202,0.24335324373925016,
 0.5580801387567934,0.45522314227656613,0.6706432501578846,0.9916450838288005,
 0.6849367708553626,0.1100128668155389,0.901283508392292,0.6328366902291691,
 0.24942967134728644,0.5811894356672923,0.9341208526421039,0.4365807139257043,
 0.6881029437538212,0.7734950257726537,0.8467950301512704,0.43235242985656297,
 0.6852057545048433,0.8831338029462085,0.05916571304325213,0.3038864292977279,
 0.04283252933349413,0.1498707704722445,0.7677933312137692,
 0.17582128736367852,0.45737230210514057,0.5115488797478595,
 0.025477578897459163,0.8024948608910867,0.601545279514223,0.5136684673534397,
 0.007860936473773283,0.16056381096828476,0.9410549182303442,
 0.34150385813160855,0.4627629980033865,0.8723408092357919,
 0.031127588688195207,0.5430198712557733,0.6382172636179207,
 0.2738504048066517,0.0928421499239912,0.42516630687255874,
 0.04704269298448116,0.6460308057454817,0.46008262713208037,
 0.5244665169808257,0.5145481744591117,0.9441532977670167,0.8429528965697792,
 0.3598721574909258,0.6899106137199941,0.42989687223189077,0.4946373461099307,
 0.9965508736951509,0.34252530711692675,0.8588780279512194,
 0.052581703017726356,0.2196949872086259,0.8711195750050201,
 0.7911028370642862,0.08269638593599371,0.44957008959838707,
 0.2943749173861666,0.32014281290854163,0.3900995826381608,0.442332517936302,
 0.5913527113343879,0.8980219738060101,0.22420085449537952,0.7918892321653012,
 0.1228898891053114,0.5796518798467226,0.9826806110752055,0.4597490829145807,
 0.42600891782116745,0.8547356923153926,0.8168028521266387,0.911408611884733,
 0.18960309589767532,0.6874363694641108,0.6261072980820276,
 0.10208380175154841,0.998365442645148,0.7308692451323957,0.2097756906113346,
 0.6000215712196335,0.02741955308066435,0.41597962242696496,
 0.5798362610404619,0.4203520989863212,0.2500766257782585,0.06898380542943083,
 0.005248768750327426,0.28160992127107254,0.3650022741404373,
 0.7468930476265256,0.44677758462447126,0.5756288130204069,
 0.27781764866715597,0.9507791170120856,0.8616308726517995,0.8102329263887158,
 0.7229594627795003,0.1576816116706501,0.4380586465418901,0.6721294534464927,
 0.013986252377943598,0.47109544518544155,0.38565256076288146,
 0.05684951434948016,0.5983757529615146,0.5788708935142819,0.8066331946035845,
 0.09340917781652613,0.46215647831257123,0.8792355863950989,0.729196372243013,
 0.7804330796984938,0.9254877821644668,0.7020647363544285,0.15342317770419678,
 0.9811269316618787,0.5780570848238589,0.911771950740692,0.9752938489759337,
 0.4530778890373185,0.06062943031964907,0.33375373245351814,
 0.2326516936780867,0.9380517254791423,0.998893113382931,0.686499052331205,
 0.6318233613146854,0.890773151109581,0.6413870122346562,0.3786604878813843,
 0.6471745259105933,0.2467347228246103,0.05436335967453587,
 0.05396123588008672,0.09142347576485632,0.6701145030669224,
 0.22451649694307552,0.9409254157215367,0.22157627255189394,
 0.8742836842739358,0.778948128894421,0.3862737805892247,0.6122414174704932,
 0.45675701994836493,0.9897945623874693,0.8794946188679691,0.6070077738677934,
 0.10713628699956412,0.4906663045682348,0.07043287071058935,
 0.8440260803602058,0.459889182938648,0.07055332658046509,0.42261327089613443,
 0.779074430959211,0.3404631347794942,0.9672285096790085,0.9614085977491141,
 0.6646062373181081,0.9108724627615162,0.6055323857511908,0.5264987330094363,
 0.07160017631416293,0.46735471467577694,0.39114263455892906,
 0.7928995522834749,0.250833338540039,0.9234287591290166,0.2017931358915872,
 0.4437987862095847,0.1286039954098822,0.33301547455030756,
 0.10297538563568809,0.9031789435929548,0.6864512955182962,0.5794729249354124,
 0.6407942042584994,0.48775242366632376,0.15044645252849775,0.508361938074489,
 0.6364439258876149,0.4597916780922314,0.381775113764126,0.4316785219746586,
 0.3454939155464344,0.20587041944772266,0.9130652284312456,
 0.38488324741189905,0.7211896354832401,0.004442071805185965,
 0.8759337739953003,0.8249681595814806,0.04076928419029058,0.6821655885525966,
 0.8145683279125536,0.23596013046186304,0.031143496572329177,
 0.23380404136014354,0.7691376902438037,0.9788842713270824,
 0.48837977975837876,0.19575061682418105,0.37463570686888203,
 0.5778896180778825,0.3235759012488558,0.2829998700573746,0.25874935990409476,
 0.9288552204966577,0.6414492079583123,0.7417539289209841,0.4792265038513751,
 0.9490128064562182,0.4609121308163504,0.19168863440056427,
 0.050591818676809175,0.536792155666633,0.5164062670464731,
 0.01807428340419115,0.2535616802632741,0.18377643033558966,
 0.7960835382995078,0.4807252029072764,0.9120798294503601,0.9265533697746106,
 0.8877483894079392,0.6723140768129815,0.24073225142424315,
 0.10882557972156748,0.42758129923173605,0.6960940048539609,
 0.8845850170942786,0.502753088790616,0.7202076441974349,0.20425483922412802,
 0.9170631575025442,0.8708881304903852,0.71448462011516,0.9906912644242756,
 0.023455062809791705,0.775771038460891,0.13442619070062867,
 0.9960796989399237,0.37608206617338125,0.6203515277913327,
 0.009383008673327309,0.890302188126784,0.5066367337741043,0.9234548707754906,
 0.3254852492339586,0.3079704178943866,0.6093661060632161,0.14812695169372692,
 0.6038283344379766,0.13755326931890277,0.001751991712742318,
 0.9977866742243285,0.42204855645507267,0.5634350323131564,0.5066735070289118,
 0.6641971607455165,0.05297369145912911,0.029806398254710764,
 0.6003782688859981,0.10605060794194654,0.06251654177117283,
 0.8635273775277169,0.914358999032349,0.03405475629013366,0.5931106660341476,
 0.4518428970195557,0.19035097002416113,0.7699153682339789,
 0.40823261867319927,0.2434677576278692,0.7971609635051757,0.610501589820313,
 0.7236929853431274,0.7634243044636903,0.9531112279884923,0.8435200388305437,
 0.9813514127150772,0.2056219330158613,0.9993372287448201,0.3162072863509747,
 0.9866256410443237,0.5786649365292037,0.04874551835654328,0.9064449131568696,
 0.5752889343011584,0.3964292837565463,0.8175550610538266,0.31949024145976,
 0.5028841649056359,0.8415130166559414,0.29558096559903124,
 0.41062707571122226,0.9016514525621446,0.5171924746938794,0.5238568207726487,
 0.7250982123623582,0.6380694015597151,0.16959770004252694,0.9387220622885748,
 0.7097842469754824,0.3118962037398165,0.28212325118124815,0.7428005700947135,
 0.948374004671295,0.49600323092733223,0.9351035715123911,0.26445146634049366,
 0.40877594454904775,0.6062860250171561,0.338418874873173,0.6137301055475346,
 0.23912119344538363,0.7000459800284227,0.5312156769131199,0.3415447642207352,
 0.8262320960171408,0.4561660203167178,0.35174111612401904,0.1280298345859271,
 0.03039479395840461,0.3647425712868507,0.6258319843651198,0.6713070708486102,
 0.10871967516476566,0.8286533141139357,0.18958018784774755,
 0.29288221624228383,0.2110786809766756,0.4515774144297082,
 0.41122594289542036,0.699492883579423,0.5945060089752974,0.3075354818622458,
 0.5407010791470648,0.36644037844154786,0.14216854906114007,
 0.9872501867332716,0.0667030923510703,0.8099648639518315,0.4487630300892943,
 0.21581620929792156,0.9564431999734698,0.6700917280947378,
 0.13800067191117538,0.13298314132656874,0.3940002201523396,
 0.7742302008306989,0.27940983738435643,0.3333641395074033,0.5344818182074184,
 0.7113551470444537,0.6865360094143494,0.17080511306723223,
 0.45075922157584436,0.33739874618496835,0.11161292415319368,
 0.888357899728857,0.822577711505859,0.3384722612779868,0.29192222321452344,
 0.45456600898154775,0.5414617748162224,0.5386783855000855,
 0.24117748811099893,0.25274553286040813,0.5234362998602538,
 0.3159232125839906,0.0861267100786427,0.6340750434879566,0.16404154311467656,
 0.37922849464347763,0.9951061175845423,0.04240432180183129,
 0.22777793661477486,0.4276799195939992,0.31499128202814275,
 0.5424381413993725,0.2202523055308263,0.9815907200968238,0.7208455897129871,
 0.46357436329183666,0.06090875085056091,0.18127851104767667,
 0.7215554111093752,0.9692843959354566,0.831205260222076,0.6875413990706991,
 0.9557893178287409,0.47768918464846155,0.045824147823409156,
 0.39393592948467027,0.3810143425007999,0.46059928002997175,
 0.7981063234583803,0.24593886716195634,0.2442843071644143,0.4128726644324784,
 0.8295774075723032,0.905179611392713,0.5621760868588468,0.517936709189796,
 0.585225408151524,0.18358301991065562,0.1410752337543506,0.12674542554645207,
 0.09431824779300146,0.055826908927421814,0.41013823982117426,
 0.19443884921609345,0.37516946540468,0.38086529474011477,0.03474879888567095,
 0.24413403823537705,0.7794884976452239,0.575955098284721,0.18137856885768933,
 0.5735134393021664,0.756033191757426,0.43881266130302155,0.9642435665991733,
 0.5883907720961552,0.08709308164656926,0.867390987967159,0.2799690481039133,
 0.3860444107302764,0.08587427520661617,0.6026944717989241,0.2606826401104725,
 0.06842550386810997,0.43827868036633255,0.2508960384857306,
 0.9957014937847184,0.1442106044196113,0.2572064345753471,0.18553791100666828,
 0.25919350585369993,0.2651290840730939,0.9614181303053468,0.4304609647356936,
 0.6558403520639478,0.02247407671695323,0.7405099563700555,0.4284295275815555,
 0.08802232704919699,0.07369101200454853,0.9836045822000021,
 0.7926183702063894,0.26990520523482386,0.6061400909122355,0.3911701472097554,
 0.2735563485729894,0.7863502881754584,0.4952864334325494,0.604350029681967,
 0.11503695334235231,0.12739519474576666,0.1698260565080888,
 0.5901357184523339,0.4881043859421277,0.8577218382986285,0.8501306020709212,
 0.6916966008919698,0.19758244679683112,0.44211069702284145,
 0.32575771339846504,0.13765787289875864,0.21382768060528234,
 0.581369747565814,0.9003457428946109,0.396995099283603,0.7404438888224616,
 0.21107593426344873,0.5239546215163904,0.5549087740106681,0.9279762259559741,
 0.869328988699837,0.5950616103704947,0.8253881765639612,0.815682107702857,
 0.3061979999328187,0.4792753565758935,0.19952522904303627,0.4523005047508948,
 0.42546175133277253,0.9017538004322991,0.3393752633110123,0.7046729730081689,
 0.6725397332191214,0.6325711424199204,0.7261645400837848,0.6093452338955639,
 0.8586175902552338,0.7379434119881524,0.27705957216695565,0.1762518416252571,
 0.428543911580086,0.2697819288203146,0.5544733402630035,0.021695927175366725,
 0.9677564594978612,0.9846546051538214,0.7896958936288276,0.8145372866108473,
 0.7245351590076105,0.9889815767734368,0.050255996523458135,
 0.7989747281568561,0.40233423473814345,0.5250469419228221,
 0.20129537910445794,0.41939323266577666,0.1435665330915259,
 0.6652926954317325,0.9849434495001992,0.5858564871898503,0.201991595653491,
 0.5930659576812544,0.17980904789012087,0.29740578236160364,
 0.055230213332688405,0.19358343464928485,0.7170516907265718,
 0.16151333151620373,0.8830366392111635,0.5204651207064697,0.8536542907220397,
 0.39449503734633684,0.43318030274382613,0.23456171743492205,
 0.9371678402370035,0.4214196786567921,0.015779231149148698,
 0.5641280155461097,0.9133133881623984,0.30022397927911726,0.2584184235405469,
 0.480874731285849,0.053469678916692076,0.4427718145220422,0.919324623124579,
 0.5746488437308541,0.03762460251606958,0.09575483076529578,0.806376443425512,
 0.8565617374720694,0.5118149597241368,0.897561694528134,0.5209023586835588,
 0.13321375838787697,0.281726421904001,0.23652563134330684,
 0.02041173360409898,0.7661400229404901,0.8813855316557644,
 0.22250778335036636,0.7099624564441213,0.5944107302091881,0.7314568529993837,
 0.8030840102763217,0.8627969599472831,0.48403249955712435,0.9228380426749551,
 0.2047606207922238,0.6291206995443064,0.4750150964864306,0.9750478401485864,
 0.7764564956508139,0.23777711919315125,0.1639940768879219,0.1478517007644613,
 0.26552561451065515,0.08019210868431781,0.2484082608751823,
 0.8310441232002688,0.6299536577407066,0.5212100456239576,0.667789107299823,
 0.9713497245331882,0.006154281222972902,0.715210416446689,0.8412524229585057,
 0.36996602568581427,0.1579206701598561,0.25451629490076844,
 0.8787114113549824,0.7709193749690284,0.4381329131918069,0.8855631007880921,
 0.6892005690148053,0.3509315481259381,0.9475745148711826,0.8740249513477927,
 0.49702948383611345,0.38498083420383433,0.00918581933419571,
 0.07154256043050533,0.6594300674669579,0.4975110487786867,0.9247746515616421,
 0.22795566955173663,0.4182796644043859,0.3406775803016854,
 0.09988163297290364,0.9665843337851306,0.21464239550165987,
 0.8284190353179285,0.980829774516601,0.2259868682491344,0.8882298309154298,
 0.8884212900991606,0.9045821004121518,0.6683256967817792,0.625689409343702,
 0.22356904836739377,0.5550744564757671,0.5539718714032155,0.5529347979440252,
 0.08750574164271141,0.4886953366243877,0.6084659888101718,
 0.08872598700998768,0.9415551755487099,0.8250426191930488,
 0.16856795742958985,0.3870652384735642,0.4438130492695527,
 0.20314219043716808,0.8315570231669498,0.23296908596064303,
 0.5833140586896792,0.3688126354465694,0.33224855293715905,
 0.22786285992880362,0.8333842882853022,0.46106800341098864,
 0.14289602031026705,0.016561626328404078,0.22260522455426068,
 0.803150277112585,0.12452118355115938,0.6596646620980833,0.46494933253968374,
 0.14675284754592477,0.2610515277776616,0.08187189465592948,
 0.4907779865063393,0.1717533482126824,0.621804583198383,0.14308920569126637,
 0.015330287402170684,0.14120856635812873,0.030462084748085427,
 0.5526510569788741,0.07527473709170429,0.229240253296086,0.09715347715809453,
 0.9267873300732903,0.5300899589255396,0.9521726560835936,0.7631725587499544,
 0.643828686604117,0.5473129898499155,0.17970362745735846,0.6221668838858593,
 0.3782016586273017,0.6519914551809676,0.3837049520739554,0.342140559119965,
 0.06291369206061237,0.22168186035423165,0.020090877269249297,
 0.59342148924527,0.31949965514885137,0.7825257899383274,0.7455561102099207,
 0.6632610109661798,0.08104728988143362,0.8416937063335012,0.3577366212331916,
 0.5999752259943989,0.4295754256977551,0.722448804852754,0.8477647017802691,
 0.8847417442601915,0.10822599087834339,0.4109791977630626,
 0.21499282636100092,0.7157134056909101,0.2520949149264907,
 0.22292051414762815,0.06804816830727378,0.08517015951620366,
 0.9715824213100102,0.8834360122216088,0.5169633950378045,0.8885921247837101,
 0.8344614933971136,0.37876238710483345,0.9624317679585301,0.787779543849214,
 0.05447724008650445,0.24835691419137018,0.7063814984859824,
 0.9424002801337326,0.7081586918398907,0.303134654441908,0.37835877850718747,
 0.6672243403236924,0.48516012540229825,0.7386731030324158,0.9203720514507203,
 0.7547149710520231,0.15170420045614674,0.8046509374657368,0.7850548691514364,
 0.017064200862368706,0.16411623841996814,0.5478946056363232,
 0.4596468107057471,0.5431112707033751,0.282511653957525,0.4928597516190871,
 0.6949676941489522,0.06990194783234605,0.308258005474909,0.28454159098560783,
 0.9571601654626942,0.11203033847113297,0.8858317547256446,0.543979813973597,
 0.4689564850355459,0.7702132473541359,0.5750884217687757,0.4602251021264947,
 0.3884574590739379,0.6984626973935705,0.8972416777174844,0.17646810782025002,
 0.24804828341572827,0.20482340946003486,0.17534133149965814,
 0.14806197328906512,0.6157963814920513,0.45072169827353525,
 0.6024326837041968,0.7920973112629797,0.9277279771497051,0.4231770541242327,
 0.3623056326930114,0.6546084037031518,0.09008351749118193,
 0.015250580865218355,0.5055433504772646,0.04540030585209154,
 0.6115493294923305,0.36874965259397785,0.3243651470892046,0.2459000624527209,
 0.3499489625186054,0.6245632482819338,0.7495208840350202,0.48345625661916536,
 0.08370990879123796,0.2794966948874955,0.8695914318462707,
 0.19616603972622682,0.8164202735454658,0.6804829099182841,0.7719019917020101,
 0.5988827613101837,0.9458883997789956,0.29326576384056313,
 0.00541294438876061,0.2688437397752571,0.6895716419300801,
 0.02515241591536599,0.13689617126617082,0.19059948949446626,
 0.8519684288759102,0.942090599780154,0.5794805148895296,0.5745985493197021,
 0.09697853010876889,0.5671542037787668,0.629082725684616,0.26639842536141434,
 0.6032400946897009,0.847093516628869,0.6146133271214491,0.7883224454606371,
 0.10289209476113093,0.04473938836681368,0.14363988695596352,
 0.253385695763484,0.5876361870008886,0.6390013210266455,0.752920572941651,
 0.9106926410708839,0.25998140266664427,0.32459294092428015,
 0.040544821334855374,0.4219830227650697,0.6998778575840479,
 0.046358709688784194,0.515210863684496,0.25313439591195497,
 0.6316269294731468,0.6936272546877921,0.6291597827752748,0.5363730498517477,
 0.12175881070762573,0.5902496236604509,0.8288279621781585,
 0.04078515364088586,0.3284121998353533,0.37586033624994153,
 0.3461927871624134,0.6215773570277975,0.9934331671159445,0.20000690740132576,
 0.642446738378083,0.5999484536847823,0.2812663003129938,0.39173046856617944,
 0.2923399102540517,0.3858585821863948,0.8678192889416193,0.8280746714556302,
 0.7601883369581304,0.47036542924969116,0.009430662694244751,
 0.5443377839667156,0.19147142316223387,0.8854279447874646,0.2500408916344018,
 0.7449794803971709,0.28486730541169336,0.1358928075733925,0.6684849525694834,
 0.5456049921955063,0.830004086262523,0.32602871243369835,0.21587746616002312,
 0.660144372545686,0.5778113439799564,0.3051750149927951,0.012158421286845522,
 0.5629937415841454,0.835095076834363,0.4585768367761909,0.320605193773575,
 0.9266943322165607,0.5134438896047953,0.3955537434414942,0.5939836721093912,
 0.2192628522327259,0.5330999392123342,0.9723363610453215,0.09871283826019883,
 0.30370850363807245,0.454291324144263,0.3265122293542172,0.05561806463041885,
 0.009500244424277993,0.7394279768673611,0.14842390305154063,
 0.13908698022376387,0.5215990170701941,0.8197813448292323,
 0.07814269238179605,0.6490836565907264,0.8692538487251619,
 0.18758074440301065,0.08108893520229854,0.9925056064504543,0.715838638045335,
 0.16027773321116923,0.535190904196085,0.8608425284388775,0.5934305760633383,
 0.9787895177910013,0.5298661198872079,0.8959795626627584,0.7906903504618049,
 0.8903087911160436,0.7492733649796831,0.31464370718632917,
 0.03946554648830303,0.6294193972699909,0.16812086572283846,
 0.21821531555185403,0.39062136991181995,0.29882326652819513,
 0.009048909901158142,0.0902264442854156,0.9958142478095539,0.875734858222334,
 0.5168294869704434,0.027892796148846766,0.2500049604288388,
 0.3183729130741645,0.7094212652393126,7.815050756232456e-5,
 0.34941859040193723,0.748556899403253,0.661890214634005,0.43285827236974095,
 0.04095871607843116,0.9246242153544019,0.4823138085217219,0.5605775830323281,
 0.38346950910648103,0.6264646300977474,0.8585518900326112,0.6604604364551712,
 0.6646883345271524,0.07039829484309224,0.024484350098598617,
 0.4015015968593909,0.9371268610687855,0.09738392909682592,
 0.04306953587544715,0.7705452681552087,0.5305073228416056,0.8443349904958358,
 0.08152113570993114,0.5740813279728165,0.6506005520686337,
 0.22918428396253265,0.6513331570781614,0.5843189267547898,
 0.08914491932354218,0.6097410739020206,0.14626349454148824,
 0.17222090605343743,0.5710027998449642,0.4715130350868524,0.7174713783832349,
 0.9761460863881593,0.5576156692115073,0.5653933325167348,0.7608014884166527,
 0.6891319908660822,0.6266782254871953,0.8492753114625113,0.150384194355623,
 0.8127768300389367,0.8481277082716533,0.5286789183695926,0.20860463386126493,
 0.8654895789000778,0.02031992551116124,0.4447755467978549,0.9499416866107953,
 0.5560572207482359,0.8476655989242387,0.6517255064984946,0.25291388342616794,
 0.19510219795656225,0.7198905160992609,0.9843403474036769,
 0.35348472205480674,0.21474153109652572,0.27817946850941266,
 0.19694051796015966,0.9922411195646934,0.8104069332494451,0.7051943085482869,
 0.7163418633486925,0.7101191056189783,0.8791465902846446,0.6692452250872098,
 0.7446900663168696,0.8154407151228604,0.5960049196994402,0.5870345189295417,
 0.1838368199447491,0.9866518550147273,0.4506077767934107,0.5353638562307093,
 0.24607551241939896,0.5293206699650895,0.6209314320960555,0.8407842079321258,
 0.6059828066302777,0.6332548498875723,0.5976160491618426,0.11366909522058344,
 0.30876354395276573,0.9037088235781552,0.44987018172877113,
 0.09116599735350506,0.7146504269887781,0.4612259967059954,
 0.36965240096413887,0.8324553410447962,0.7632427561683831,0.5401850364961881,
 0.6885163555827736,0.870132310166501,0.8573144550480704,0.03681981161572445,
 0.44314934849902865,0.16753703967993772,0.6237240249654409,
 0.7697510615573969,0.013551094231555982,0.047848753581188674,
 0.6086257324362632,0.4658394080845505,0.4123022489917856,0.22941112357736082,
 0.13479594030458664,0.5438234050939663,0.6388455201116465,0.869298197264799,
 0.90783240416967,0.7531209761192361,0.8114888190909282,0.2478609357442897,
 0.4115962940188336,0.31919013835302446,0.4342598101938152,
 0.07460130516293373,0.4868753316289435,0.9006389662593393,0.6864009033897602,
 0.9590827740499838,0.04383959980290009,0.7368269512726816,
 0.20456701077955053,0.556941515031302,0.7214423588690935,0.9970696750837106,
 0.4497886463123927,0.4949051849023792,0.7532164725870696,0.4420181307233849,
 0.8303620539015633,0.8944088882580118,0.23683840390815836,
 0.13936052940055732,0.08033382110477172,0.1185336909967627,
 0.35045851448289067,0.9236183219417127,0.8573910321523759,0.8254156774959766,
 0.4941177197856508,0.9738013047755321,0.9679209126772927,0.5231960780882696,
 0.5986470100486254,0.7806347481302801,0.6739341815313492,0.8599424819688039,
 0.10789381709201296,0.6964184496554272,0.970261425732806,0.9775053994360886,
 0.3608394051257209,0.5258594274793611,0.5149059110221686,0.4510856412271441,
 0.14474900373005506,0.3353605231746516,0.3313480845405822,0.2142152417357135,
 0.903163977334672,0.01305562470469901,0.41732115173968376,
 0.17911272444613036,0.3629749594847532,0.8062115649127856,0.5927974035406945,
 0.37108099845463816,0.2910306098770008,0.11817026565250732,
 0.7678459028846614,0.5264869292360072,0.5332803108399469,0.6239928018080442,
 0.4782102322094306,0.5780792777884507,0.3141951188875547,0.6030740040581131,
 0.2417452510558168,0.1314014356884987,0.48072995503300986,
 0.051545530992741395,0.8804512396944613,0.3129243507180809,
 0.10744459974266318,0.16017699324941959,0.9875916802522077,
 0.4857781156328399,0.6642080161076631,0.15645900049583783,0.5485721862968623,
 0.8217905193602801,0.7734906407784168,0.3748066093062885,0.8859204915893748,
 0.17923973424670403,0.9366055411909439,0.5627649159598374,0.3430937759353436,
 0.06775441584269082,0.259761651561223,0.3126791582640909,0.04589542682858305,
 0.49300025486068655,0.45589956069868387,0.16590533035745914,
 0.8235699736834845,0.6373262718506759,0.43895509128285104,0.2839378141828356,
 0.6932808390017718,0.20077755158164579,0.6778576202418439,
 0.48855804651168977,0.48372876677774157,0.641054343318411,0.2073485888663249,
 0.925350359261189,0.4948468167501727,0.3232999759424232,0.5377022360389567,
 0.0032474487644386096,0.05561767528049488,0.16059649707787726,
 0.5685538203602444,0.3755767092437067,0.42425394855970544,0.9944305013355516,
 0.7889682425319804,0.8768198201326654,0.11437275605510777,0.4747047664071059,
 0.8229708355164082,0.6057381529735769,0.7923243912864506,0.7997294325008412,
 0.632276997590201,0.590541530659025,0.7942684904903177,0.48399135686634953,
 0.2825088926663356,0.154539088956118,0.6301686334113076,0.3551909488878111,
 0.8879076538880708,0.6794063493855471,0.5424687107117846,0.7915763658935824,
 0.6616921253229231,0.13481622625649792,0.16306501596016598,
 0.08721093923377399,0.42680211646280997,0.8512628599765004,
 0.052399356109777195,0.13199792040488256,0.17932799167790137,
 0.019450590979211757,0.951449971258839,0.926339281372254,0.05057141148391564,
 0.2366379418661484,0.9537217030836509,0.08602385302700555,0.4274017660493551,
 0.5714817556692193,0.7673831747775306,0.8121085138467555,0.7654670218609436,
 0.7517089570440808,0.4837551814838258,0.9209466651368426,0.42548720862790324,
 0.4270999802874118,0.7296358674749457,0.5829759161862182,0.43519235805409173,
 0.8014925519818924,0.0012151785924278125,0.9262096647547742,
 0.20641626605957097,0.4920861238635885,0.27580170139405513,
 0.20023123289959943,0.10489589732020366,0.4403200967398274,
 0.6938630890497633,0.06757743977250497,0.822070952535376,0.8952449565546133,
 0.6038807342311061,0.24011848943224812,0.7184976010052161,
 0.36183016165395265,0.5579739357399099,0.6462829507961559,
 0.43206484387947763,0.08219954746092051,0.35472213166245425,
 0.8991743367372674,0.0030913087495960667,0.7447506066441199,
 0.7225601970530436,0.467858238963338,0.15292878129753018,0.674533360693798,
 0.889677323975796,0.18881992845475803,0.6030171084551127,0.22031196685013965,
 0.5774731590192916,0.4109372231275792,0.17548700490783176,0.405218323670292,
 0.4968913339137426,0.9026377036756336,0.5874080720156977,0.35141588152142056,
 0.7629049737670025,0.9489292758066133,0.08595676761459992,
 0.13799950842808972,0.30779525972349,0.39285432743633775,0.7766369330237464,
 0.7857954649178283,0.7522705185598924,0.11556720289877753,0.9174505146934037,
 0.5233743667880998,0.9750582503090449,0.397240889693631,0.545989266502663,
 0.4732225256242538,0.8858084189611083,0.5484614701414449,0.3260891862639249,
 0.5390420980981544,0.06069236049845783,0.8324678863593363,0.7613761921352866,
 0.7082924389709565,0.0036789389428915076,0.1461842474693723,
 0.9870747637699534,0.3741039742217669,0.36314675026104903,0.8097004001947337,
 0.5897705537664581,0.8353122402861922,0.8120431706171632,0.69943379648356,
 0.5661262839710454,0.7202066371269308,0.22790785347189735,
 0.40494082217170424,0.6664520868890741,0.8483245605579707,0.996932112199651,
 0.45086415254667633,0.26598790147479356,0.22384072757289208,
 0.549548156741745,0.3004264533325981,0.6671337732358007,0.7976116384627999,
 0.011653790417489596,0.29985429846862977,0.623546316966722,
 0.7641252883450474,0.3394236004670793,0.41885748500520437,
 0.22036127724489418,0.5350597436463131,0.23821010005191956,
 0.11537634514379469,0.6552875065668577,0.09519370366906488,
 0.33340488177164795,0.8905799719914919,0.22411368596854575,
 0.09198572342722189,0.9565282214365314,0.6389564753042436,0.7942019815845657,
 0.30542900477790436,0.6492605973378662,0.5401896528651352,0.3310142704714053,
 0.39232617148125104,0.7862288485685165,0.8680618362930761,0.5278085039912692,
 0.2052630551689466,0.10863737107923477,0.35005367916908503,
 0.4480615287380122,0.8691053372489896,0.6777670530700145,0.18915622102545704,
 0.13998367392632916,0.17285670890726124,0.38750221416176145,
 0.7122372309151239,0.9457813691967851,0.07128130548855971,0.6813133223691469,
 0.0300994374703476,0.4770809361264804,0.9670973405400463,
 0.006457920395859684,0.8897999975043184,0.9614656410862777,
 0.3614948412780541,0.32210140303033175,0.5438061484135057,0.7010154423270092,
 0.7467683944880175,0.025233561429815143,0.4157326379801396,
 0.6403921935786892,0.6639502287037907,0.03890938473580474,0.9314760917147971,
 0.467277097047346,0.46050083263734054,0.13636205161717563,
 0.40858042767598346,0.7755649418038583,0.951201356589361,0.38883003667927807,
 0.931216331355858,0.7892131701672348,0.07525433959106065,0.27509684228286313,
 0.8235243313996885,0.13360286495746176,0.4601889039121728,
 0.16824615538415588,0.008433425846603448,0.2889612550218673,
 0.6670210667133021,0.3298213368376721,0.6417189966745356,0.3401319167144342,
 0.06043299739418739,0.41427671527917465,0.3119986579674022,
 0.5833341615079733,0.5573726524125406,0.9197866574334543,0.781968030957033,
 0.6095321485213462,0.664833534018556,0.6774064009193328,0.011567287851907082,
 0.5676957145294581,0.6503319699034912,0.5526157821198695,0.2286156039107914,
 0.08801495759875422,0.17181865497337634,0.512161541042836,0.5917127076006641,
 0.8925236419203988,0.06880669428239872,0.559191761182252,0.38903391305013346,
 0.41090780443560326,0.9497720698069877,0.3833517538878296,0.6104615202546052,
 0.8241874074930462,0.49299555642854864,0.0010301638927784351,
 0.5999289868520054,0.6396753741084731,0.47863872930709095,0.8100848517724604,
 0.6372148504863249,0.27305147007827557,0.04759838206718725,
 0.4289607123014805,0.749784804955153,0.8812898285128281,0.277912934805676,
 0.6109212168564441,0.9834128659586034,0.9157542824309047,0.1852811350255661,
 0.1802631136858337,0.6138165199129693,0.780089104512315,0.8514126127543695,
 0.01928229729050157,0.5793966879611377,0.5894328632114325,0.6639510272389357,
 0.2733076417389009,0.35419390967444275,0.8163939920014762,0.5342638751866037,
 0.9908196303202201,0.8265317370409481,0.7226287573386204,0.8610595152423377,
 0.13044206640382627,0.42212541049829755,0.20344902842801582,
 0.2425540738997145,0.3098504143340691,0.5718391660012736,0.7002817322545326,
 0.15844606692223828,0.6771709615425667,0.02215894390347617,
 0.41822061115120335,0.5548942236331899,0.5927695307376515,
 0.009063485727617726,0.3512314232551006,0.3824488549738684,
 0.6978225760783312,0.7434210940494136,0.7921408447896773,0.6427894580239869,
 0.8241045055917648,0.1824083307404749,0.4676547648433469,0.708387487053433,
 0.266401294685673,0.2927331467158176,0.22422955833266955,0.19464492405997313,
 0.5172646447470948,0.7466518126516078,0.6420019816661269,0.7542625435602169,
 0.9011375193511438,0.3806396758227635,0.9347763703240195,0.10952497727935095,
 0.20834061165536633,0.39696767538096245,0.27370896230330155,
 0.5238358295145014,0.8646777441020046,0.2699634699143564,0.9706261366507467,
 0.42862584971048356,0.4984662604491683,0.9742731615537692,0.5875546346681293,
 0.9262941427863391,0.778681706975623,0.8421255319714054,0.41968520895144334,
 0.7777888979248049,0.8385027029604369,0.4345440587041449,0.7871200543867674,
 0.22443951731201794,0.5368027733725252,0.5261991328458064,0.4015272849527036,
 0.833398386756196,0.978087834226414,0.8857690112866162,0.6928157334287642,
 0.12238354799964846,0.8164056634574588,0.9662433980399403,0.9891532492224858,
 0.5322823222126032,0.6630729432002399,0.02873208548578665,0.0106292935919714,
 0.3875514065737906,0.4893312196611518,0.2142341759478994,0.7026968369672452,
 0.2674007077596965,0.04329235568280465,0.49864024682535935,
 0.7734031756061243,0.373916583768951,0.022288264490553544,0.5795007958324848,
 0.035435992024873686,0.4258079814095157,0.4092702340489929,
 0.4420021638743754,0.568764937430173,0.9681687781854937,0.8260666316369714,
 0.602167001302077,0.6209148582214102,0.0254816821004078,0.3741065954859434,
 0.8754589892106358,0.8381993781863722,0.9970877675643086,
 0.024141654403852075,0.29801132966041055,0.631242115743172,0.630684695567417,
 0.659843383756681,0.4489619712490903,0.4963947719336661,0.6010542777352262,
 0.8170443140293033,0.4574222679971247,0.8016763083241959,0.7223160147326183,
 0.5780143207679912,7.993494763677278e-4,0.43460038858700545,
 0.7428787625656906,0.04674867640176705,0.9661789062333892,
 0.13660232653317828,0.6710293679485546,0.7563773746763519,0.266099867555904,
 0.29729573868508397,0.875139104141613,0.1134033536193031,0.03378708263072239,
 0.3047052437277863,0.7133607296330146,0.4377193813853951,0.5399431020258869,
 0.3322635297242662,0.7636997172774314,0.3953116925496498,0.46297053975755376,
 0.8522175379315898,0.29418090235405436,0.21891169603771665,
 0.8902922833180853,0.35505721656628875,0.10179538759634388,
 0.7449853704206214,0.22900988535326183,0.2485307253514314,0.2180167252469547,
 0.5101529915631746,0.29575798577410517,0.2680060912939253,0.7837633705902673,
 0.9543207493972262,0.7738949776817021,0.961409123272459,0.005540281752984266,
 0.7105067080314177,0.5395566208986925,0.9796385259171826,0.25826863803333044,
 0.47173068760557335,0.3386765875739257,0.28196788920093463,
 0.9418676308594908,0.3178846119112644,0.6149235376907957,0.6381367470430893,
 0.6649965253552557,0.4214930961589426,0.30212473219985114,
 0.25448653475998917,0.631269858537615,0.9118247455676669,0.9721381554986052,
 0.3920848982980498,0.28174926431386593,0.1891270589673386,
 0.09267791131187264,0.9544470699715678,0.8768498046220072,0.131394677484297,
 0.009189774511502269,0.540800790717939,0.06299434177358099,
 0.28618229979116283,0.7947579002680859,0.6964513776288798,0.6696911962898988,
 0.6749918033746428,0.3034007762299533,0.6856359862816513,0.7958977518270115,
 0.8789738323982274,0.9715328637459263,0.7947785650323128,0.8434221103690273,
 0.07562916650711116,0.10968680074012449,0.8570520651176498,
 0.7516937456809826,0.5779238651982985,0.582548851526605,0.17411061044209097,
 0.7167542334795263,0.7423882747571037,0.5451222032979792,0.4386307654791398,
 0.8523184230961812,0.7270406316209452,0.25125248867435845,
 0.10989478908627837,0.952928844249477,0.7860369785228777,0.06177404348439397,
 0.9976963893767009,0.23197684368304639,0.1829734608366026,
 0.33028759463582524,0.1677196313249626,0.16194706427062644,
 0.46975412897831825,0.6601611102982662,0.21765436687315798,
 0.2190720579401384,0.736196148543757,0.8154247545935344,0.4746105748571561,
 0.32899505138916996,0.2803651676348182,0.0946420561085346,
 0.06462066074002781,0.2492643459600299,0.7156362581295022,
 0.18766149604793536,0.25481001637131484,0.877352755368813,0.5295262259697446,
 0.9566362173555484,0.6019268166319317,0.8928780443864002,0.08055934547266197,
 0.6011268713436224,0.10288268409834056,0.8404410775537848,
 0.048691246847521574,0.2826162959682881,0.3123883476345243,
 0.8461084961613027,0.10920104906499706,0.8665100086288547,0.6905092887186345,
 0.4792484012956615,0.2548019172743722,0.7473973375307404,0.6087507817787364,
 0.5781188168291226,0.5544378687248619,0.10960997953931284,
 0.18534432132832468,0.17095407960956288,0.2654494999835426,
 0.09405430904089285,0.1801560737348059,0.19114758795466724,
 0.5807980761622468,0.7931493875826107,0.43472124648300325,0.5144167894189322,
 0.45222338121992933,0.7124919882356119,0.9947306399146635,0.7704729632211765,
 0.4973388509943373,0.6706217663968863,0.5634508123233483,0.23812564496084232,
 0.008655862681024917,0.7837449259883233,0.6339236179794872,
 0.1063763560430182,0.12988295846163356,0.8454379667160812,0.7487726655016207,
 0.22857018851175703,0.28197869693163036,0.6207927374934149,
 0.22661442219399786,0.4408480456977695,0.8494393064338897,0.2675748036959603,
 0.7821693972703984,0.1844736903594304,0.47563482841517,0.452204976626553,
 0.1975628025455336,0.5044118737671115,0.5299446855479252,0.1402757047683475,
 0.25849701449911555,0.49743315300323254,0.4534557170779999,
 0.5874735659497542,0.5494839814434237,0.4393397168229266,0.08920722348065735,
 0.9179365090009965,0.2393338081134322,0.7954703953845345,0.2588499579069924,
 0.2726827462069028,0.48093982758204123,0.2387121928937126,
 0.04875390758189846,0.8781617611485832,0.989887839933663,0.3863666025122171,
 0.5918300632468296,0.8543797735464905,0.33898713877213094,0.9789135493007509,
 0.3286923602031231,0.2805890438975226,0.8858306613218179,0.8301684500965268,
 0.7110349158765107,0.004508540288864493,0.2342100080753602,
 0.9982417225321336,0.28106302305878117,0.059066630406721665,
 0.832326264315428,0.9083700613688384,0.5297629066268467,0.17393029939707783,
 0.44046479507527014,0.7378457979350356,0.32520790592476956,
 0.38972259984699464,0.7440250453808641,0.9377705184764964,0.6661887908501813,
 0.6963645360738895,0.5550116157919226,0.38854072440635823,0.8573268084592475,
 0.02572835785662675,0.6436907203364179,0.3654239056814129,
 0.16979978179307098,0.5401593825671774,0.4551557489024376,0.4834983801640329,
 0.4252477407123296,0.7062082976397365,0.07043873914225385,0.6551735432327677,
 0.6502679471425367,0.5206182727510023,0.14177981430129483,
 0.12664385538980183,0.8362973968776384,0.09053870004311393,
 0.6845020715211279,0.34565331646810393,0.23295441829960906,
 0.12046088711568481,0.8475108851527966,0.18818154269038934,
 0.052304628175064205,0.3648528590520761,0.9056038230018678,
 0.04171719250085881,0.13085474652155904,0.6838539034946406,0.79468775365023,
 0.5503704143766329,0.8747313616455175,0.575842886759427,0.5867121340337641,
 0.8362889809406352,0.665626905734604,0.1202704034765183,0.7495600423871832,
 0.3625521293066729,0.3738855510597201,0.4120570271515478,0.5730677934371053,
 0.5784504153059442,0.6356470196351508,0.9869420262460822,0.18170908583686907,
 0.5410984718455071,0.28446013514797497,0.7436389848817148,
 0.36301073754963964,0.5512016015429678,0.4572063395261017,
 0.11502059875345683,0.8341709944533108,0.39329837554845004,
 0.08415415229773227,0.42315757155260636,0.9954024382199548,
 0.5830311317258244,0.7957427625054961,0.768019359326107,0.5224588915719792,
 0.9098032456899029,0.1533452495298797,0.24284054055508075,0.5496832589348541,
 0.23802118448745868,0.7823532257788551,0.04335713467044866,
 0.9164986391984601,0.7219677944904075,0.7359281170207361,0.32194313357942217,
 0.782787963810337,0.24848091353004575,0.7651073513910509,0.40181174562634714,
 0.5887015893207338,0.26060697748216055,0.6403014775356127,
 0.18678429450107303,0.3631560684980748,0.9578951165148488,0.8834987335957317,
 0.6983041861287382,0.6304823828935455,0.9802297500912183,0.6153849647060641,
 0.27576197239554645,0.5713039445814081,0.029746300021052496,
 0.3353170312510225,0.233535066002727,0.4724731754462329,0.5496928532995119,
 0.6103516778730882,0.9798873340408079,0.26914970724131304,0.5384860208127191,
 0.8626437175333371,0.7105493098750035,0.3684699249276656,0.21506602650226425,
 0.5576136374556658,0.8410812004995956,0.8164674203012924,0.4362788794241994,
 0.26223522236940844,0.6747002512588874,0.1802102369392209,0.5317846538559892,
 0.5342679447103236,0.9887202888350716,0.05726763929408951,0.5498606874697587,
 0.27174753762999293,0.8568145893769279,0.6930486803344722,0.9772645189197945,
 0.20718150969909854,0.38493303937530443,0.24588513920679422,
 0.8935503402512958,0.4896348516545648,0.7830287127055486,0.5414749529684582,
 0.6609561430621883,0.2745682229535553,0.6942727383719975,0.33458017193114287,
 0.08212841438251195,0.6304666955608665,0.9985317539713051,0.6843422417488629,
 0.1086849563928185,0.1185784305827553,0.93101363491211,0.7092840089948783,
 0.4314076110611401,0.36024728284017915,0.7606766283536007,0.5839110176777463,
 0.005426598284780981,0.11025044695832797,0.32438663679013746,
 0.44876493182497534,0.8429885396504889,0.6189330639877612,
 0.045417353374671454,0.9153998867793,0.817377892953783,0.4578105397491896,
 0.5792566532846093,0.4968210414818448,0.00326065227771144,0.7067069827670804,
 0.49424201684181424,0.4416191035437068,0.038183150816652356,
 0.5538941504197525,0.7316595129703819,0.8096441371421114,0.05278066940582793,
 0.6556048089522073,0.7534368910533434,0.1664077231066764,0.16307065801437481,
 0.4021100780522038,0.6009996776025869,0.34802571876876487,0.2578345317563281,
 0.6569640572591073,0.8686713507622491,0.7904501115620318,
 0.040085680411860225,0.082043719820037,0.12841483586768287,
 0.28694862375263674,0.17100101241491783,0.23289096762192008,
 0.6564834221411207,0.3310126255604787,0.1117183427221794,0.0604167885295952,
 0.2717384404469284,0.8533552298143903,0.7730387500937171,0.7490291567123188,
 0.9917998958586822,0.3699957874405988,0.34011088668119427,
 0.10000583746962388,0.33013697667445163,0.9665689525439924,
 0.7830983578476269,0.021512081880672662,0.6153881141916854,
 0.35664168554195497,0.6094304842182667,0.6862034518379537,0.5126900698236386,
 0.027673666315427004,0.430959145604942,0.7326848324004607,
 0.37735996762299706,0.07003070209424878,0.4468639840033841,
 0.09993127330542717,0.707166715843616,0.630449020170456,0.8041959616548138,
 0.49146752306623465,0.5908671011026312,0.17090349509997693,
 0.7103017577585617,0.5701634649390421,0.4394803271661123,0.1946438879335466,
 0.5967575983780815,0.15872074885379606,0.6825330975623458,0.8226604072835246,
 0.8360106422703496,0.8279136595563308,0.5156742840649565,0.5847399498928033,
 0.06634478693574408,0.5550748225059551,0.16340683986892457,
 0.3805324122780246,0.3527190925227799,0.38881787433132997,0.9952385831199386,
 0.37031741704511334,0.5527616814741291,0.10370489910506775,0.389995195180023,
 0.10159289418400774,0.957657063676107,0.5174699309421096,0.7022988542503186,
 0.6870724980481102,0.9495210766928759,0.5884060827422503,0.47220329439008624,
 0.06827625841771512,0.11125241880372794,0.8557077311629981,
 0.030376573286065245,0.5459152483795261,0.9965950672723166,
 0.18803202665823404,0.07935577986820674,0.17202047660363196,
 0.6866416552402895,0.2611299576596071,0.3383724438861193,0.09845535364731961,
 0.12176072269362259,0.5027200506033,0.5240406267383569,0.6453470051238482,
 0.4248920592049701,0.029835455593035798,0.8320915857571349,
 0.44796164151371665,0.9938912492301645,0.10042931840275104,
 0.9263813673811786,0.47591645645140146,0.2373767933684553,0.5669503783743245,
 0.7189972627793826,0.9634167323005697,0.04577549858871832,0.8817305119623676,
 0.9038212121572804,0.5816942350184273,0.07549119170893581,0.682587397513653,
 0.8504376733808781,0.31851987859132247,0.23625209275144465,0.677606563730167,
 0.11856687470035665,0.2727061771319794,0.8453579891217786,0.8125959303540718,
 0.5755599090040348,0.8578121238851745,0.29207209057100714,0.4215733566225134,
 0.7382460990444556,0.9526478993726052,0.03226324075356879,0.4195658774206643,
 0.003964707890383501,0.71536191695272,0.44485516319970764,0.8065331839886803,
 0.6060384639641472,0.09678217265606937,0.5842335997869356,
 0.06115653458635972,0.7653248626904912,0.12220688800245927,
 0.45196272865738585,0.06866799515946831,0.14067911373198483,
 0.8632252955786979,0.20079988659685022,0.8195438030210334,0.935787822407048,
 0.9817812146609449,0.40025949870554844,0.7458489930533977,
 0.05811683983082783,0.49855791772135283,0.20965733848817503,
 0.4130979777814341,0.46498358324211864,0.63694474790453,0.44520305158078066,
 0.7251551900291222,0.5508370536263016,0.6413621915318801,0.42511657597298935,
 0.9018744940018671,0.17538426536539842,0.11012205717778778,
 0.7972850510500882,0.4182972153767205,0.8177605195656745,
 0.022769304149235614,0.7720354974130614,0.6059860777817877,
 0.46436751545451693,0.5605194720374113,0.649772102353907,0.29364683650166645,
 0.6166371641914817,0.9467821632583895,0.46568699131359836,
 0.13160787077638436,0.11171207520459814,0.6210862472989178,
 0.6435993885319421,0.6354223535308018,0.5863185610803476,0.24508762022873776,
 0.6316903999417192,0.798853606158864,0.7471239507486953,0.6592845002606641,
 0.3877745772631538,0.3130207504697896,0.6083697961601751,0.42418666175646313,
 0.32013895213002386,0.6898849496007463,0.45694062111565725,
 0.6611400392914578,0.2725018452739254,0.6585459748665968,0.9858284144463629,
 0.16821359941352776,0.002770589210219665,0.8425938276001593,
 0.2637717785714301,0.3037753126663776,0.49478003107598434,0.7844778292511748,
 0.12813010274928538,0.5915585920248259,0.2021378063427961,
 0.12692143720081284,0.8882432257002945,0.25768341105343084,0.680248072696436,
 0.36786994555295205,0.7435242099028243,0.870766813267399,0.9657780051836714,
 0.5349695421690832,0.6881494458960288,0.2754139142638179,0.05072328896364653,
 0.9520357165397052,0.432886246374776,0.93974572833616,0.020619859446938227,
 0.6430759831512599,0.27280089152511944,0.5911022440540364,
 0.013099106749268619,0.07606945211069038,0.4208660424080848,
 0.3382515564009837,0.5137305197887717,0.8972574058254481,0.5215072031579076,
 0.9223574376804146,0.7143997384221664,0.2607660629754118,0.12750900168665036,
 0.8954536465701753,0.21830953648892537,0.08924994273651621,
 0.8971773839336364,0.8609487682956167,0.2846888625087516,0.28567933522118816,
 0.28252911847760975,0.39060373370503787,0.0027848022874931155,
 0.9389559574474033,0.3209003834135421,0.9893512178182348,0.22004290829982542,
 0.651543235033308,0.6366204402702047,0.574944178356605,0.7474594742029337,
 0.393012155676413,0.36609961076744035,0.20748237693387117,
 0.40509049668864705,0.22515755258920922,0.849352135335417,0.7850461786857881,
 0.6778293264874982,0.46144313272904247,0.24587889022458753,
 0.8560037379175568,0.15117531235893777,0.24407427497421963,
 0.19651463058038232,0.5501035467246774,0.4206672097950074,
 0.22568617290606163,0.7136998827729106,0.2755559337833958,0.6449250770692345,
 0.07509310321967888,0.4779334382902357,0.5025626449453751,0.5733475774568969,
 0.15437213726002708,0.7863176915248636,0.26787085638885766,
 0.7106494624164706,0.21515108697598784,0.959998589035074,0.9307298351176212,
 0.20080618083070356,0.1629074480759617,0.5597383936637148,0.6646479812555276,
 0.2747686061651039,0.5415518045426043,0.7577374679653784,0.7895853036593495,
 0.24277639020896136,0.4160686879286797,0.475589269342128,0.8073113454947117,
 0.39828819060681175,0.9838242571243345,0.6093521822063793,
 0.24278342103838957,0.23764991365374333,0.8972898488765974,
 0.7705654355113366,0.9592494642649525,0.7575785325681278,0.5905422468799321,
 0.4017994337681008,0.8882133739747542,0.3183795181890211,0.20865856590648102,
 0.39634765469446553,0.2821258192932693,0.5694406330496843,0.9449523128668625,
 0.05522690933967811,0.564004602136726,0.23620646394640188,
 0.48246012597058674,0.5312681662016698,0.659270521641854,0.19657331655507393,
 0.3437267492250775,0.7510637116034823,0.8893973350329818,0.11808755593770282,
 0.41054518349380587,0.49501749810890605,0.15661172961837755,
 0.4255511114304643,0.26117357167852084,0.11414815921269306,
 0.11840630467469493,0.5415512417179624,0.4160378420334059,
 0.13422170837192127,0.9235070298735316,0.8991132011435637,
 0.27272182263333444,0.4283156832296908,0.07096238184880943,
 0.46464425903958995,0.8302645195787417,0.5686870794383513,0.8884674518662546,
 0.421853723178098,0.41582024568963905,0.8453904729272224,0.7227464901432629,
 0.8920188429901261,0.46171684174696903,0.020120263854079612,
 0.9551949328388992,0.8919596384795796,0.1109161752113842,0.3793477916021756,
 0.40857881668167284,0.21696999869417943,0.4898900876373281,
 0.7866339455501201,0.8363072830901153,0.46095892138100447,0.503927640522443,
 0.05364849414387607,0.26438212825175,0.07804950938702815,0.8806199130437654,
 0.2829099535215974,0.07149452666254974,0.44971044220754863,
 0.5785381728384487,0.8568423326890416,0.9712420019137772,0.02802055595791353,
 0.8151856849484265,0.1444281993998039,0.49805992756132456,0.4866281263978727,
 0.29732248526484606,0.3642473286958978,0.8936339986250678,0.5337837639869059,
 0.8797988548110137,0.8651126857232843,0.7721143631044773,0.5610559679084489,
 0.6817364470911543,0.8941768733113643,0.07625116394540599,
 0.35949164060718686,0.791577844121969,0.7987712179939479,0.46150800661085556,
 0.03201111739784701,0.7434837849557885,0.2693291824579789,0.5531591422395995,
 0.8862497133045337,0.15424720038272144,0.3551765210392096,
 0.47465962107633775,0.6519737857803831,0.3141286660509274,0.6751290072126935,
 0.9846297092017621,0.23068294091708308,0.3551416143756646,
 0.49941671190093484,0.13614666503108863,0.35721424052332773,
 0.5759693934395514,0.43826923565355247,0.794488475464225,
 0.019218733013017686,0.8078417667940353,0.4706453140778186,
 0.46195212638925554,0.195501754841217,0.9216187464303446,
 0.019111901314990964,0.13933852782355338,0.32163454505891,0.1406824779791066,
 0.36986987445902797,0.9018217378705377,0.061217628671447866,
 0.6843236992911966,0.05388754657924899,0.5534460401876902,0.3028740853243467,
 0.11186957581799539,0.13389619587082435,0.7983059848524942,
 0.9317020199749739,0.44338425827025574,0.2873047295873433,
 0.19364536540677102,0.43222942142634857,0.961484687201383,
 0.21244779958545879,0.7857661280541405,0.758378452685967,0.2007432870221706,
 0.2417730062019594,0.16681561189210847,0.36853317854593026,
 0.7087989507012503,0.45196319149461406,0.1796586909274569,
 0.17753382104636395,0.710587352254442,0.913355238326367,0.48554214209545044,
 0.550992453679277,0.07282703392915346,0.9342346056385807,0.3733807409718286,
 0.01605876040367027,0.3374076740835127,0.3281552936941722,0.9887053705438735,
 0.7192102337075098,0.14323143884422007,0.25033632512297277,
 0.05720071972656249,0.715062742607845,0.7066557709688033,0.6694679260482844,
 0.30549445153807764,0.3676391859977942,0.36205090564949993,
 0.036804902915879234,0.6045153710788196,0.09800465479264897,
 0.3613963034984007,0.5025142888187861,0.5549422181844836,0.11007777019588427,
 0.38238819687174985,0.9659212707228172,0.7193885968322752,
 0.46276500797321507,0.4507736294826351,0.327261612463442,0.03150029067798221,
 0.9349839347974216,0.7370525168518844,0.6307059027711664,0.44389488376327224,
 0.8928316184850782,0.41993498384300654,0.7241128792678982,0.562748641806039,
 0.8326788795478492,0.20276367757849978,0.37748643127645753,
 0.2797280972824515,0.25139128507038366,0.03722964744104773,
 0.44574668916351057,0.6058843075437272,0.5032199739252244,
 0.31630826359718067,0.19191808719862025,0.26566239443951467,
 0.30693172039858574,0.5576808734946289,0.36304458365872816,
 0.8798223812054259,0.5513334203502799,0.6778260815685869,
 0.001553267719117235,0.8655442357333916,0.4711174098637667,
 0.9146598947585849,0.11486471424022071,0.9247769715298784,
 0.10372382030315841,0.8607498728032297,0.8294952044538547,0.8629110822720655,
 0.17961777881618135,0.30366365688627184,0.90840356998771,0.959275004351447,
 0.4029952217495484,0.9680897812052912,0.1981484144754525,0.27863392904540485,
 0.4779334745955618,0.4760449708064842,0.7810816608162866,0.2661449808121962,
 0.12037144385438558,0.01673900954524732,0.4972254499655314,
 0.9614731641786403,0.20552832673798316,0.42610534595274685,
 0.3906295224113667,0.3891852561571377,0.910411069366647,0.02574003759074872,
 0.3017369170166688,0.591054339927011,0.2589053581778541,0.16632175675869576,
 0.5732875849484409,0.3100001321260385,0.0740793789407832,0.7964331393069127,
 0.6185031461489219,0.951315219289253,0.625270456000919,0.022961569724034625,
 0.006113607237167473,0.7519353866286645,0.7412383695227092,
 0.48998053789316043,0.6651440469979342,0.5710648078358547,0.6916412702056881,
 0.5779026491798295,0.9876736800628712,0.10197243336903639,0.7417353180040156,
 0.024035184492789297,0.3449661014621078,0.9938705082122838,
 0.9776551514376497,0.08763140013871229,0.5019056416930401,0.7555430628558604,
 0.2978644409105021,0.9205224249012813,0.9761479098760167,0.932731478945127,
 0.08568836692699133,0.46561211177950135,0.24820919027271615,
 0.23076540714440352,0.6051711801677848,0.5920013770862856,0.7150130880298682,
 0.7333624145511249,0.4641448096117201,0.8445715943212293,0.399710337101739,
 0.7366237262534061,0.4282848532502561,0.3262823974718715,0.06549715392841893,
 0.31881077301676075,0.05006523342986291,0.7966123357734768,
 0.7189646618280997,0.3962331123189642,0.35696844616688406,0.9207646885357587,
 0.6251253350513901,0.8190226294572263,0.29680245714642317,0.7457422685910428,
 0.3228141175030257,0.5464807808805441,0.8575485098326641,0.9253955401005981,
 0.6079376366212793,0.03150402972372346,0.31280850067647337,
 0.42345323989538364,0.913862467150569,0.012047811802407882,
 0.09257963905737843,0.014921954684434913,0.3544344584490823,
 0.017303678890324914,0.5607915366324929,0.38857662961893125,
 0.36131589209511183,0.947391591214981,0.9748436049379263,0.17631586602165972,
 0.2631130864516368,0.6349910843333969,0.29188479527853806,
 0.40217467080886116,0.96429504624655,0.05273193383768915,0.2999838466835133,
 0.5610388189579981,0.5545168459834939,0.5540077092859511,0.14184200810861924,
 0.637754640797536,0.5048334098715616,0.5638714537790078,0.6032018924034988,
 0.948525181502695,0.40950794175051364,0.4372105022548378,0.04569288737028587,
 0.23056900284981197,0.800249170549142,0.637700932564583,0.20589074659134443,
 0.5745726356223622,0.34646653988439624,0.2785995756541172,
 0.42176509810759477,0.4486826567402974,0.828074635022059,0.11413885460144624,
 0.6458791088742563,0.8467647605087607,0.44027923553526127,
 0.34513328317517533,0.7290195220279592,0.27907164962350484,
 0.11411544587262756,0.23933116340715133,0.39868643603817555,
 0.40668931901440586,0.23482283835518492,0.47005142645407216,
 0.8820094538665751,0.3804473229755294,0.03513420984442339,0.8281178937338218,
 0.3311545761214779,0.5546604394803061,0.14767021101858557,
 0.23731646182988608,0.8738916081687502,0.46057263275595817,
 0.09994917553439431,0.43482996060616574,0.7946747036059224,
 0.28205037117007437,0.6611828940431699,0.8049239819775209,0.8163185262237568,
 0.3057006024373643,0.42701916854324706,0.43237315387965447,
 0.7092741968445989,0.4803310596256538,0.11500582121960545,0.8764350950070549,
 0.15470247851789265,0.2286270981213383,0.49446883600049696,
 0.3755517589677334,0.554635686804097,0.46410399474859,0.9164461629774926,
 0.7260271335039732,0.6204979571770659,0.5191421911331728,0.5307358971254422,
 0.6071214377621701,0.7338257372391228,0.8742490165346157,0.26072812824845804,
 0.2513584577808018,0.8764862110397733,0.6313979596424101,0.8400432573864061,
 0.6362543560626501,0.890638825025517,0.9684224342824804,0.7198667870073417,
 0.8520724600510345,0.6077655768708731,0.38248452062593385,
 0.38032215894546273,0.16898108533920064,0.14103488534669573,
 0.6748495018162916,0.5558357466559783,0.7088853058379049,0.5848524760031141,
 0.6416303968323985,0.9415174144187136,0.842826758837397,0.8260777408701241,
 0.08594079792660803,0.0871490496045092,0.39342472731355094,
 0.7179111464034358,0.7536383704803995,0.6158353347360662,0.797829673416983,
 0.27327806659013065,0.49489796708963163,0.7194401205355394,
 0.6606361058620399,0.6811331927411765,0.01986890818929965,0.6886773668252274,
 0.024554289848221944,0.2877854666416635,0.38441896578843604,
 0.03585238133021762,0.11446522361003975,0.42725151790516114,
 0.9010161494752584,0.8753224327707019,0.7284059127636859,0.05297236877133937,
 0.525399651320582,0.6791021283539617,0.9108377834620511,0.27148711782933077,
 0.6330720006314751,0.5994483290613274,0.26733583353667034,0.3377944119673816,
 0.12772272889905423,0.3606076617142401,0.9995792566239015,0.853021718139332,
 0.8216219980746534,0.4856686133139938,0.9699091930603583,0.6470162901674873,
 0.8028229088191235,0.5765742733723861,0.20052238554516555,
 0.21978285292147404,0.4473041762638784,0.1345934102083176,0.8859798135729553,
 0.9134136385566959,0.46322549388412004,0.5667675630584749,0.7266872286191728,
 0.8139758666216322,0.470842060427737,0.6733494302934954,0.9712828822566165,
 0.9157782358598139,0.08480710046220585,0.989191781815659,0.5280883906443983,
 0.2933572474724656,0.310906094931206,0.13161093683553093,0.7316231236446856,
 0.5523197774962976,0.5955005372673322,0.47358996663389974,0.5467905815058764,
 0.5775123498904438,0.6379992717405779,0.4248903810978045,0.7927868657942329,
 0.2618050535800778,0.9317180299330035,0.8550398603258147,0.1883534505885328,
 0.7407549486549418,0.5992371989333584,0.6840993365487276,0.9422172681347051,
 0.5999010972924461,0.5332455147182124,0.4522288725653508,0.13134594693168056,
 0.6802920540264447,0.9189041899834925,0.7143501004926521,0.40135824672875553,
 0.07512044919637728,0.06424883807222714,0.2665154202961978,
 0.9977796677667106,0.8625798921003122,0.4989166298234371,0.9032352429176215,
 0.8596740801183053,0.34799862334420995,0.044088531335228165,
 0.5730826568173023,0.8139199492827986,0.22808216807998793,
 0.03334899674274039,0.7994868767806054,0.05396762905224117,
 0.5797127326008458,0.5002082018273181,0.5070268335239101,0.12526683157310914,
 0.8810003918792947,0.372714651653542,0.1282122402942445,0.20659387635897852,
 0.17944710400394104,0.26276403929178194,0.6182863350304525,
 0.7634460046949203,0.28273523893078434,0.6433309053571772,
 0.44875255033746997,0.7646745601293732,0.2799189411657116,
 0.025916914427367543,0.044185318505940874,0.7337660846165375,
 0.48777297549124954,0.0918214218014789,0.9214590916601717,
 0.31400977557706555,0.9010832967794786,0.5861227996262258,
 0.41451021866393967,0.9804452675114412,0.6577724987461432,0.566301015480584,
 0.5254738680312924,0.18784366935751629,0.09409298978692737,
 0.7672303431796277,0.848251372932897,0.13065740830515282,0.8370939848860718,
 0.10674261771573401,0.44500064367011916,0.3757731904726521,
 0.3763933161866163,0.15807712652024186,0.19138290332144348,
 0.6894930343086092,0.7322042436408251,0.6811036796115788,
 0.031204378741733763,0.1270395548361496,0.25339302499723226,
 0.5287820095311655,0.615079717494741,0.029641600132985224,0.5106540017146298,
 0.3626319808384372,0.649629571850143,0.5273405576739915,0.4304664232175124,
 0.10566645426681487,0.7711244908490089,0.3801285138865509,0.1094725956235123,
 0.8543377126975874,0.15936470705927308,0.2184977547601542,
 0.08809900078307842,0.9294717239767735,0.5131165432662559,0.8901536113542242,
 0.12692035614180963,0.998061170074292,0.2806189188381587,0.9052213761654713,
 0.30677294332824845,0.42162695052993704,0.11275675285912701,
 0.5057289263928291,0.41949460217732226,0.724307226039815,0.21757302154919878,
 0.40896439545738106,0.6298141975624361,0.32866244921138776,
 0.5834129611414063,0.32699823982487397,0.1263393072086315,
 0.31327763539435693,0.6172759017404448,0.9466304797763319,0.7652205626032031,
 0.3732822512796923,0.9945068381782883,0.7083904547038147,0.9622847291009492,
 0.4496827250919635,0.019745071298266748,0.031078247786637037,
 0.520063979910871,0.9718056106389833,0.12018316091065162,0.5553399637896654,
 0.6457637878921637,0.02661609848239488,0.9083434040687103,0.9990376015089898,
 0.37590099619524286,0.39111567325907515,0.9378990671197116,
 0.8152948095055291,0.04125567556181087,0.18915431271882077,
 0.8241842577316044,0.14339525418365928,0.4524400630156856,0.9629870456980092,
 0.05885479052658227,0.08192751957739264,0.6150938584553226,
 0.6105705950055781,0.874255716280779,0.8370810328162492,0.7641604418392546,
 0.3677017748069027,0.403344232488154,0.9723300463984688,0.2492141645894802,
 0.726858626283353,0.19849488368089885,0.928584444699825,0.8900319704150339,
 0.5442999460667648,0.4562190483764863,0.5604898550174635,0.5597682708893483,
 0.04975348270933022,0.4978065437342586,0.17404159922097517,
 0.9083913291536085,0.7289899791214828,0.04976339007695241,
 0.22936748553744746,0.3701438929766043,0.9643063922389337,
 0.20691480360797931,0.9448017166930743,0.506893409928876,0.557437649643893,
 0.8287729491589098,0.8693157426760759,0.40364925621990955,0.4305706454062125,
 0.4507779837279474,0.3284920886124736,0.5809911417771507,
 0.006169271361402551,0.7180963763521007,0.9011863617418244,
 0.7552461540720886,0.07207591568347327,0.41450811923088204,
 0.6280114356173199,0.9058909867388532,0.5380241907756598,0.1807968087257532,
 0.784608338722181,0.12184723023322208,0.005754782749333165,
 0.22039312241332798,0.7394930662759988,0.5539127764865769,
 0.24830201090902526,0.521562763240932,0.3599767369471113,0.6424116332930131,
 0.6257337058286591,0.12853027442116494,0.4595314984169383,0.6676585547505278,
 0.4773505300780885,0.6928119227214735,0.6259020839655481,0.8210702850058847,
 0.7863896161701561,0.3019741894530924,0.9345585560552805,0.6989911740370256,
 0.45513649264551614,0.8695460240442687,0.4150565900869694,
 0.16874672014630254,0.12348771705377404,0.7463297168633479,
 0.8075190427005305,0.48184391223674017,0.7045229582114979,0.3459099545997728,
 0.684529785240167,0.3360448617993085,0.8432417871108829,0.4583161333600171,
 0.7912710699972861,0.9392825620190193,0.6105853824785313,0.5941547045089194,
 0.09174031724658516,0.16072893549710077,0.8488225992473691,
 0.35331833228571163,0.6138283229447522,0.9950155379351285,0.3455541328454623,
 0.23881565994188692,0.8224762549020405,0.3301828515161488,0.8494697120580399,
 0.7179379687608389,0.9175015703478083,0.4362915643609798,0.7020960424840207,
 0.022320726817989578,0.03520682571954725,0.5349090999788146,
 0.8655359495895462,0.7504446041985142,0.6898903455160789,0.39166680739220283,
 0.051792721982015455,0.600829691822625,0.6487047774466403,0.9642298183184821,
 0.5644973048161475,0.4940048711902113,0.6080315537393483,0.632033125006636,
 0.2863124836826523,0.7912491794358426,0.5872459781400037,0.9244986087053126,
 0.6546195330034729,0.3096984956371144,0.18183830880800678,
 0.22087393095866792,0.12967928670233064,0.40850023311392847,
 0.44567339949121787,0.7944944271877323,0.9801283122947538,0.670622249710174,
 0.9973756036367005,0.40417148832320404,0.20442006592669404,
 0.8341745126717561,0.7839659082733925,0.6487895244717512,0.6800884695199299,
 0.09944794384078381,0.20802164953108004,0.5922609064296014,
 0.9423251709162569,0.527290470721797,0.6312122911924535,0.23784440768483006,
 0.28648439509999346,0.6001636634238674,0.7468032597072338,0.2885198764205553,
 0.9126509790916393,0.6688892564383768,0.11324397235089112,
 0.18746489252660237,0.27143551217343675,0.3615678214825221,
 0.42075587681643767,0.5167723405191844,0.02796908383883523,
 0.06179380015141378,0.9486394494835643,0.8969753957266158,
 0.015561258961919067,0.30931595446096805,0.07719475420993649,
 0.6724933814748862,0.35098077741182887,0.5131844686424676,0.6186262195123264,
 0.9961597520169403,0.8026124094459066,0.5718722180305463,0.6757478568861258,
 0.10397440113916956,0.6476772627051501,0.40719620867913375,
 0.22676210001327535,0.6516496851928673,0.2416352347585734,0.6177660734989954,
 0.4015083570917619,0.9276907095893732,0.9888215118095702,0.2396292284902729,
 0.3876801614785065,0.47536395811359133,0.8905871791721532,
 0.34488426858247756,0.5723602785902409,0.5997692844822222,0.9277747562305854,
 0.6806841648093391,0.013077584657075425,0.7506221424637993,
 0.41580822025046327,0.7233145689044937,0.9493128729753734,0.3606421909542168,
 0.3443304924634165,0.893321546273016,0.5289396309782193,0.4172987116489656,
 0.05494842817364032,0.22279174626744047,0.03105445226972081,
 0.6254240812285161,0.9960714419174569,0.18127155041712784,0.1521038504389619,
 0.004480926458354961,0.42445563086567517,0.9378902102238745,
 0.6533228450637293,0.48829884487784114,0.2684585725698153,0.7965202810700409,
 0.6527454773490196,0.8581254074747067,0.7763504791352903,0.10191925629327625,
 0.38955147131111423,0.587720583981676,0.15064702895972637,0.5963141854865988,
 0.6304949647636939,0.2086838096714061,0.9910605830699281,
 0.036162081398653845,0.6091185565451469,0.4401959363320448,0.151448335224314,
 0.8027149303111446,0.33908192653255476,0.4784830083992907,
 0.05196504650158895,0.7422820297957788,0.5851719196345708,0.6930701224333677,
 0.05223765068094777,0.5177426619247136,0.3172567008048407,0.2620669907250801,
 0.3318288219905775,0.11611390431447177,0.6363352565399192,
 0.20130660010065693,0.09793106939445417,0.17624792442342707,
 0.38518559187931123,0.4286018183062107,0.8683903821029328,0.8709816238023834,
 0.5699560228375371,0.29885412821516333,0.40395213320227275,
 0.6575113039914522,0.49353945149511824,0.33740696562868533,
 0.07513209606885707,0.7738893469206837,0.9649117605600867,
 0.19381278186038853,0.1571912719574976,0.2257373374887115,
 0.020634453352327276,0.8597986488735331,0.8494583570787083,
 0.3544218099599018,0.3079267359463773,0.39609867759938844,
 0.09128771256118962,0.21766998975890217,0.2231830252119691,
 0.028207430533071398,0.5665164412537046,0.44238741659360603,
 0.5902304077437558,0.4349132663595209,0.48886926499458716,0.7521328501479363,
 0.8969168308825451,0.6238730932929518,0.890240817274224,0.7709613710008821,
 0.8179297990219571,0.09700767691597001,0.21247994090671052,
 0.5988543768495084,0.3740579641203343,0.19547120127618334,0.3101621286135261,
 0.6531470891513553,0.8448894455223797,0.9060969152964468,0.2332022019092509,
 0.1052720596125356,0.4079381410947539,0.19231612585043834,0.5351553837916524,
 0.7825591258826932,0.5957771572546806,0.3726598460727776,0.43892256861445134,
 0.9970047890101715,0.7966977150420156,0.8381644071093395,0.99052844794748,
 0.361590037982801,0.06527859144440096,0.32079184406829575,
 0.026656928602249996,0.25934694046528173,0.04690368871566364,
 0.20801167951480237,0.08379841105654318,0.28899452818411064,
 0.7259694243329712,0.0667614008715055,0.0012459394980446348,
 0.31417742377261115,0.3973291163317103,0.822210959444303,0.30176947536778154,
 0.2317380089471156,0.625417771409933,0.966226083224008,0.5441356164166085,
 0.3934042134786515,0.6793383822014112,0.6380993254562431,0.9410349111904618,
 0.1462373684165733,0.4343536072622318,0.9704930058720622,0.46609722867513215,
 0.2878384494479602,0.9488703202006489,0.9991204527999294,0.7215026860347679,
 0.31717221743616575,0.16046705130918204,0.737836685506571,0.6859693950575404,
 0.48164650700066947,0.0872504006173539,0.4420253135292974,
 0.16861383810282715,0.0701043865393306,0.4715544323811136,
 0.08720771602418165,0.22434655452318064,0.3843715154199091,
 0.4877770334708025,0.46045081449341596,0.3559808947718115,0.5757813135543525,
 0.5067387347934226,0.5143887480769688,0.7261231087724671,0.35024737982649934,
 0.15990935209160084,0.05116106744814686,0.14605587293067623,
 0.8301183937575843,0.25359700307491106,0.6807469282283195,0.7211896842752937,
 0.9563678918035516,0.8746406056585048,0.5784726931016617,0.3230559007523618,
 0.8409922258431091,0.6571717740722991,0.9434264346840171,0.8940591032746478,
 0.7166928959296044,0.9885740707092487,0.5527261697206041,0.2999759219143051,
 0.9455479728825538,0.45903039064171436,0.33620448821750826,
 0.6684049423812896,0.43897497322738754,0.6000377614714176,0.4134106485566089,
 0.15393750742817647,0.731297108684364,0.04345277381548751,
 0.009017574617089075,0.7234119950322164,0.6111724730549669,
 0.20707294135730692,0.35801014313290325,0.6638619899523125,
 0.8054708222409565,0.794151232640412,0.9754215967723043,0.5120734058578746,
 0.5569451250062777,0.2625169377045796,0.6334842244421409,0.9273990320596556,
 0.17880277391271848,0.7435550358180499,0.6249103238926652,
 0.47962686288730105,0.20295007483164196,0.85522046883899,0.4180417266072437,
 0.4714453438376265,0.8846204650631133,0.7638453140741347,
 0.0015492784712692886,0.8426216935918183,0.3259320183220653,
 0.7945576650406425,0.8920733861905155,0.12411089238926332,0.6636298874261053,
 0.7166366126805561,0.3261642105185556,0.658613664231782,0.34944666044678874,
 0.759525974975897,0.3731037261609661,0.9433331106362188,0.4859584052083822,
 0.4124018687634964,0.07433950130983757,0.3389516949615956,
 0.37910451118868527,0.47778454740293475,0.25950611416622393,
 0.5129392746777709,0.8808049516998309,0.030270804195374534,0.983183866931417,
 0.46333377191333136,0.7444473461670215,0.12036034547246888,
 0.2926276419942364,0.8990684522853671,0.6524951275316991,0.19250730406770666,
 0.27810976984231717,0.3068453823486619,0.3236077394058757,0.6673336424184656,
 0.564893478805865,0.9297353113474804,0.24885493096889777,0.04245033595251735,
 0.23656817518541517,0.8839806818864121,0.8501953827435543,0.62055985680645,
 0.10896224215356909,0.5401775283509396,0.3173213850948122,0.4836991789143734,
 0.8910159596291872,0.17337428905220154,0.6365508805036447,0.6591718242579236,
 0.16805745752200374,0.25487092897190167,0.5305666096066652,
 0.38366700275188514,0.9869334909596881,0.5727748544433073,
 0.17945839283845366,0.5164571876826558,0.47065348912287985,
 0.6121987716228474,0.015265350421079016,0.24202944832162943,
 0.41001178687162454,0.34201039934409017,0.9373507999183635,
 0.6737577350887375,0.3599824065891698,0.7618482253854825,0.9497894998219917,
 0.6516224151389821,0.9486319581748233,0.8955033731707873,0.49420350776719635,
 0.5797426769823489,0.04765367155654865,0.002578772923673034,
 0.22698367609592796,0.8922679217549153,0.9979086358006461,
 0.16007727386617043,0.958558015920261,0.5479246797284812,0.3630158725044277,
 0.07167133683851645,0.008520151625031613,0.39081827146736825,
 0.828803196061161,0.02638022859955702,0.9671336312044821,0.8015586101945151,
 0.04116250689518686,0.8076779196536314,0.35484897169556484,
 0.6685754814522713,0.9295225872303982,0.15301207331347033,0.976655722111757,
 0.09649235909134612,0.9810386831198866,0.21362186276689166,
 0.04756701626014892,0.3627210328220546,0.7729753031392474,0.6107710708068766,
 0.14641639117378136,0.6464200320609377,0.4994702998080167,
 0.46284339952724085,0.22454466855626198,0.4425731349018106,
 0.8766148302572584,0.3847681038974561,0.9501878806290538,0.9641076843677758,
 0.9943441370032827,0.7920235861530557,0.004643072912289603,
 0.7843063266647726,0.635932677137659,0.5478634452175556,0.010752275764889463,
 0.18335537270401625,0.5835677477075073,0.9546495988026265,0.7033479966373901,
 0.15216015916233094,0.36944700028428645,0.8117777911148647,
 0.20834186766923035,0.32280812996665975,0.3888948894603992,
 0.8279545638335586,0.948729626782157,0.6463127777193164,0.4543049788424357,
 0.5523007472361308,0.3731161789208556,0.7485581438395308,0.39812821972681023,
 0.031452605311757864,0.6236010893327528,0.13994921299125218,
 0.7740732372415409,0.6176532188747592,0.4215562608861152,0.7628728946697472,
 0.7707179309852064,0.35107350184742814,0.6800782130209356,
 0.11132006627156876,0.8066902124658486,0.8699117864213619,0.9794080639839619,
 0.6686059124274355,0.7463177482115132,0.3625149915517917,0.9004534447533306,
 0.898100402356797,0.884787926939863,0.112083718645696,0.21560794110641557,
 0.48975028227456674,0.7248713837525544,0.8229843397456731,0.0810100726461156,
 0.9494106059641482,0.6964588137352498,0.6287995738360047,0.7985215681400368,
 0.7592826174018672,0.10698491623155193,0.7704268848490962,0.372688672826623,
 0.6299498279934189,0.6760590354893938,0.43758091495649676,0.9400174495876611,
 0.7537327504575105,0.18994759621303936,0.45629945602304045,
 0.9825221883752804,0.860658300232746,0.8381795449464425,0.8104567031874962,
 0.6615375491829033,0.24442884252433816,0.735738644697007,0.14276628147616277,
 0.7591887545644331,3.5345468004044367e-4,0.31401148337721985,
 0.570686399840187,0.29130389010241386,0.5496017299898124,0.7190617654129637,
 0.3253823398562844,0.7673540413204804,0.33801083311712055,0.963251729686903,
 0.09886382794660142,0.22702346732615108,0.040383339686761,0.6174619484092911,
 0.24293172849370825,0.3902750551768964,0.2691139809538601,
 0.00811406042294327,0.4430732015500921,0.4825890980626779,0.6701414533316106,
 0.6951030897028283,0.036406447860354296,0.6514491715940453,
 0.6586672911938775,0.578659291460862,0.1197313932608356,0.6268493482738088,
 0.9962525447387248,0.3485784297458071,0.34351337586760555,0.3639388023183189,
 0.9153812476729605,0.45461109011143885,0.7313716477718525,0.4122616917643326,
 0.6317969479847974,0.6624918125961804,0.9826850844563741,0.6939296047100212,
 0.7358085616347101,0.7318645206681755,0.2755955685801156,0.7625546619587463,
 0.6873726731954011,0.2040995954095095,0.657752808803966,0.059169284245140896,
 0.6814430692750281,0.3023873170608522,0.8463206469963341,
 0.021508583928303793,0.6196007909608581,0.08744502822730227,
 0.15754052650064843,0.8175685439052817,0.08616209114813356,
 0.06787232678868016,0.7327805041633484,0.1331455823014458,0.8307680476597096,
 0.5106626068786523,0.03064018020805659,0.4033920755685817,0.0787437395524565,
 0.6944210628997486,0.954449391945098,0.724232425004987,0.5158426593703787,
 0.6712462968352692,0.06561445443971403,0.3997023450545143,
 0.26739354939054494,0.8041871380358245,0.24545891297835742,0.382276880611806,
 0.39125424401868125,0.5282348401173389,0.6010398352352693,
 0.025073633450506172,0.8876137004542546,0.25162855455157684,
 0.7506922990241468,0.8182111098458364,0.253819196869026,0.3090726102579895,
 0.7910486239647123,0.42490832319524674,0.05472297558931272,
 0.05377036513127276,0.22989470355794828,0.9068659403711313,
 0.2700553944557058,0.25026629723825955,0.35544893761838814,0.512888046799316,
 0.1387249082411136,0.46269290233640126,0.09852838563807864,
 0.36799328302706513,0.2909461527608489,0.10316678267880564,
 0.6278213425937508,0.04441872494466681,0.8243005098201104,0.5483970521596362,
 0.576898307231655,0.5325074379915958,0.9162572602545738,0.5952597745809138,
 0.44230409757877887,0.3361800161507531,0.04717615660909824,
 0.41960231167615847,0.9794020195675968,0.8556232653112589,0.943803559015123,
 0.2355789308824231,0.29249880087393054,0.33505143184256525,
 0.1485094473322761,0.6769618990779762,0.3008925887075884,0.39465410979754423,
 0.20754634495547997,0.2793832203459041,0.4388234067559592,0.2044149410759616,
 0.41672963668970975,0.5267511611854659,0.9072489872473607,
 0.39829789139457816,0.8942239734030999,0.580234377825644,0.36368256903190654,
 0.24153196985035386,0.5741036498374077,0.3366954988683001,0.7287754084815952,
 0.2789281108498287,0.22752893382026462,0.708561067025286,0.6722665463132972,
 0.8867197927712709,0.4353827133635845,0.8813533054384867,
 0.047260318764399734,0.790116448765207,0.4312313474458187,0.8772107694782139,
 0.15726779024223125,0.1607907544992303,0.2184017661823905,
 0.28383849500120584,0.5895409146666127,0.8001120010166134,
 0.037413076295668146,0.5883693859850704,0.29895737579941106,
 0.5402544700206873,0.5890089806246079,0.6826609325997179,
 0.009569865028476454,0.49948607615635243,0.9781868717186311,
 0.7757159795744665,0.9608348213049809,0.9246774657487664,0.42363394656606657,
 0.5558803650510789,0.48290064618038375,0.5779229353122831,0.7356505109784107,
 0.5542590608790048,0.3550631928236836,0.6147225030861785,0.40720075456695215,
 0.2061870194053924,0.22291875521626647,0.1825451335284619,0.335187003168156,
 0.10772444520146995,0.5929248340200517,0.028146223509047386,
 0.4080215214920795,0.511430953079172,0.8468142225688996,0.530195205048629,
 0.9233283023684813,0.06879937925490154,0.1444844658943324,
 0.22225738698037723,0.1832740366129728,0.6776188206233478,0.038374464721738,
 0.07087119069071657,0.835261727834956,0.3455720252546959,0.5000795730608151,
 0.0030259933369746195,0.450194919313781,0.3616849648303556,
 0.4469898781049273,0.12614127592924773,0.764368426963794,0.9429560681310613,
 0.11303209502607403,0.11737617362155306,0.6294379061217951,
 0.5114339332731593,0.9243637149163516,0.2039683130299197,0.11932042555247313,
 0.36347280748878075,0.6584537041871695,0.31644109093901884,
 0.49340836132305366,0.8590296893656413,0.7090655881131139,0.9219469429021822,
 0.7466916454634382,0.017575008770389533,0.9344977734076272,
 0.013337859787785455,0.6068307941884352,0.677431341006516,0.7272999442781943,
 0.03537438828750672,0.1680233044640692,0.09341866401845733,
 0.9639939076671206,0.8585886086224863,0.8194965096770076,0.7306565442889588,
 0.7744537648395955,0.444770991490135,0.8625436968030336,0.24730589434714745,
 0.4160411229505324,0.552972292744026,0.04965756728331572,0.575856671911132,
 0.5246156857293222,0.9051522193686453,0.4902655219608031,0.7116759701805014,
 0.2865740901633502,0.8286371997862941,0.8658295385759565,0.9832844639285803,
 0.3149302340480369,0.9777696042622172,0.8717635748281594,0.12332911682699954,
 0.2572423436077841,0.4210011472298907,0.4267137537387081,0.9259506097303619,
 0.9441009951906889,0.7764198241788032,0.2257307940411929,0.15637256648159015,
 0.042194052187417785,0.8095876702984193,0.27364367104959086,
 0.20706118472917345,0.7683643186858627,0.5131646254209538,0.883644363878514,
 0.11852671877147825,0.8120677292337698,0.5349469395780703,0.2956471041577138,
 0.31271933545191777,0.5638576810245334,0.2750276413561553,0.5425633468510399,
 0.43932964089175286,0.018484982457814936,0.6242328231542207,
 0.29004265034035215,0.09370626884451472,0.47720830883407206,
 0.3073390464284296,0.45630281818327023,0.4004786603327892,0.9614683582854715,
 0.43749480184539546,0.7455826353455602,0.960566043158228,0.3990909009716136,
 0.4224545212057589,0.772383422953597,0.08126273582798915,0.09459604690532386,
 0.7361521565289879,0.873261098643644,0.6377258923770934,0.9246124730133928,
 0.09220898919685105,0.763112757654122,0.6362336236611394,0.5650617080500726,
 0.6894294112602077,0.7966463965498368,0.22578828208240853,0.475646714772751,
 0.21356074339254172,0.33422092006957105,0.4966650591620493,
 0.4334827775764779,0.7887150001739786,0.04608480598937237,0.5644519988374576,
 0.5665616788819218,0.8272103911124085,0.2859718527688865,0.4850587976746138,
 0.41320127238861193,0.8403594661947048,0.23413024686456574,
 0.10976134403442028,0.620558958794954,0.10693311884997625,0.1842696831055296,
 0.9960745591777843,0.3417376466868661,0.7474377089289134,0.32901446274612733,
 0.6433324990846394,0.5478019031509154,0.9264884004449458,0.22431861673271059,
 0.9757126616784024,0.023371546942132504,0.4174385813585221,
 0.31237715642527775,0.5419320648002453,0.2859306807738726,
 0.14462426345232238,0.10911859386899359,0.9960574015748744,
 0.22940162871237602,0.1385655244216457,0.8533184008623994,
 0.44843105637112335,0.80797173431897,0.9808349726591757,0.911004971546843,
 0.4264456017237017,0.38004867399225506,0.9811396560827433,0.8175528687576693,
 0.5970246429213284,0.09695265869289849,0.9966952462011492,0.8525594072198424,
 0.25510982224746226,0.9270591295019625,0.6861021423351394,
 0.05480591984627314,0.9220794771878129,0.0149183346745686,0.8991607724203953,
 0.7443354387199724,0.024158198893004412,0.9340453486150848,
 0.25895074811391094,0.5701592808045451,0.511967410030952,0.8727734622435797,
 0.7008968955841383,0.5869944258039997,0.8479434331170566,0.3564688358440832,
 0.2899597894080217,0.98680259493582,0.8320717911881012,0.4254754261282223,
 0.5073244911857621,0.18488438696337905,0.022877261991446685,
 0.4701674932770441,0.1958130267371243,0.2084949745878376,0.4549364517780319,
 0.14499947494438437,0.22191523196980145,0.78386196503089,0.6374484359266305,
 0.9580662631789592,0.9802739980938093,0.29486323230946987,0.5483687825285035,
 0.46786105363397257,0.9621817422974125,0.37963233236221794,
 0.6747800245612463,0.7535158954427591,0.8905258574663834,0.2817224801160866,
 0.8512535973436506,0.4634575909411629,0.5897730804768541,0.9168455934495431,
 0.7117395921763217,0.5125013633699174,0.6118378193815095,0.4580539940009605,
 0.8748758721009422,0.5017450782518513,0.14094078152857403,0.9497750204462645,
 0.3695746639731935,0.04724472824519288,0.37514603253588996,
 0.04990547959892666,0.27270059678604064,0.04048459290567785,
 0.6751679292674476,0.0033693194949802052,0.5681835241840693,
 0.6048439198743151,0.3393350595780471,0.5078819886799902,0.36547648872397265,
 0.9620129696250461,0.09935231358906982,0.5052504620777102,0.5815321452176578,
 0.4961715405240481,0.8328343521656576,0.82568341188135,0.1543809286308031,
 0.11301797742047037,0.6976033250248515,0.007607904286462186,
 0.4971519453155916,0.9306517634135535,0.9653971900084201,0.8905519006615984,
 0.4628896318173574,0.7938061049133416,0.6865363814854997,0.3400442145970173,
 0.8812843538164433,0.40726370371242415,0.3282784125767889,
 0.39599976699403383,0.7147350263599693,0.740890246725188,0.4425585369990378,
 0.44508248389656724,0.5547155200136274,0.2009447741675211,0.7511141103141611,
 0.7422708767611397,0.7766772161090503,0.8362153519321006,0.5437455150857143,
 0.651235351156674,0.8487804801107175,0.39193771966209634,0.10966094377468416,
 0.33103840629436054,0.9825065233489758,0.24602993109322313,
 0.06454659871288526,0.18194405677294212,0.30977664371383684,
 0.885555271150385,0.4990568417854093,0.20620546015766783,0.6931660032185453,
 0.21646908183151836,0.295574718991913,0.3175774380496855,0.3621941420892233,
 0.19424086376282146,0.6481026109721414,0.20222245869228594,
 0.9859698182446265,0.4036215401583808,0.5405200725695861,0.6360804304998873,
 0.09447608949826025,0.6282394199593729,0.8078508831098129,0.550481928607315,
 0.9130581514543523,0.7054592801131832,0.7017566616901928,0.03379504391809185,
 0.40183639366429635,0.5351826659044134,0.0821738339499386,0.6294771243248506,
 0.5747122706845276,0.5948625920210046,0.1964163036696419,0.9850359009680919,
 0.2889952631831746,0.1659508903609067,0.012051396235216227,
 0.43980624048321537,0.6631067851845854,0.19087463970588647,
 0.5658824001162481,0.787218668476616,0.7352885377097432,0.6287389048134842,
 0.6742641739014908,0.30105151308461775,0.8858593768819256,0.8543323408510556,
 0.4561839578695208,0.5707382480087507,0.17848379573162487,0.1003699194170471,
 0.09649803313392435,0.2479000380518961,0.35187145967644784,
 0.19605060287923737,0.34709165134057274,0.12970697917829832,
 0.29984193891223365,0.4395911470624776,0.0487083452662479,0.4028719283285881,
 0.30362669913968665,0.018743641354555862,0.6135353544689214,
 0.39827559283655833,0.6735162439247095,0.8267708894874042,
 0.32149338046471787,0.9190209179449588,0.5382942738735157,0.5066988960136485,
 0.33010479397893866,0.5060481277305344,0.10302029191977002,
 0.3593369913728327,0.9497453476576168,0.9677478652249615,0.6191745357728499,
 0.5252508217033927,0.9180262770458693,0.6973701027142634,0.9813472887550629,
 0.4137107208902948,0.5556010470048492,0.9511410182733294,0.375195252519409,
 0.42845560696827223,0.9670549626511264,0.3740464643128759,0.9547554810922928,
 0.27627241351038967,0.884956354559679,0.5168528049965009,0.5822527571202173,
 0.6145127490657857,0.8720842248058663,0.33377837094983986,0.2990662995323349,
 0.5877917136748249,0.42013644714885356,0.9516306124358201,
 0.44378796573782675,0.8076888539748036,0.3674037444709892,0.9455737885724566,
 0.170559740628333,0.45880935145820345,0.5001136785042724,0.7855206175227851,
 0.5515041125957918,0.70042482317048,0.22938730827336795,0.9708045677479994,
 0.029513376549711978,0.48072086085832344,0.30521712281239777,
 0.16945224328610864,0.3489286685429942,0.7283285003316948,0.0659274935673757,
 0.5819157505087424,0.30775669712563847,0.9948903279621292,0.9954451707236389,
 0.628038008757361,0.4545717825812484,0.5953327477888469,0.26658144404675954,
 0.8407204679166107,0.0681971275684079,0.3522794813114649,0.31831413829251287,
 0.0077681363866481945,0.5148938413214942,0.3789349063213594,
 0.7566256478173434,0.5592147607259114,0.06490376104967432,
 0.20659968174314725,0.19156891771249596,0.26119034609250513,
 0.13687535222162972,0.6181129257993644,0.33402500485198006,
 0.11389481617961372,0.9599302858631951,0.7845260227484591,
 0.06868493489656113,0.2139605186636635,0.7041488270529855,
 0.37353747881959665,0.988001030427488,0.026243529273252908,0.83578465750853,
 0.22350017061238192,0.8397743826506291,0.9698801810473576,0.3184239570943824,
 0.7050311789838584,0.13055904287875375,0.6092232897693298,0.4103403735863673,
 0.34641119837416545,0.08730727522393789,0.20178596029788365,
 0.07396712954086482,0.13983657691938822,0.3823768688899596,
 0.2280600370042024,0.6359932579923566,0.7340595440460137,0.17975762293926922,
 0.9968823437452641,0.5841681945503296,0.03751070802516776,0.4237695600035082,
 0.8838280798178281,0.7492113058894028,0.3037158952599781,0.9974932547044164,
 0.4812306089095595,0.35923903608503793,0.40228364345088363,
 0.9788072467383919,0.5572183675170991,0.22862616284896697,0.4138101479686608,
 0.15507765424264908,0.7808985471963494,0.9090159297358718,0.7245238706041952,
 0.7230645207992662,0.5716480510717082,0.31153959018393107,0.5583633897938517,
 0.54694903042119,0.3446371394769937,0.2702785206652606,0.46473044616351467,
 0.2626761797535233,0.13732974488568062,0.2377524944702737,
 0.05486466298118087,0.2543926190884084,0.9352144776580802,
 0.018054830810443456,0.9302765114309569,0.16593019449249014,
 0.6486427920533246,0.3485076605429187,0.8013545944871513,0.2390315924973373,
 0.6684349576244877,0.49940874336535646,0.9982891832644822,0.4661397517062781,
 0.7656987156702577,0.7406854398972931,0.5814241629298713,0.3254635539867474,
 0.023836807821827843,0.6559470732372423,0.9266326112144284,
 0.8679666482681747,0.6932924963586091,0.07841261403821831,
 0.23720150783267802,0.6631322446728887,0.15704006193459852,
 0.4006961618635143,0.6832586790062553,0.9746622010008892,0.9307562794840768,
 0.40486344779307804,0.6451507969344267,0.2002363597467134,0.4909390449844382,
 0.3594112907105911,0.8625137788306837,0.5841812451574728,0.4535973515607823,
 0.6131288479183422,0.5542462487168026,0.004278453008717431,
 0.0726476197278747,0.2973690027520476,0.6664913322899406,0.5075513385649693,
 0.27676078807900417,0.3156477009345898,0.22129561533820385,
 0.17882422102668916,0.1833205825171511,0.33638569846712063,
 0.3163028038474911,0.32135480828779106,0.6314188825434497,0.6069531497131435,
 0.34314098486194355,0.4960542372319473,0.26411346941346536,
 0.7086432829193564,0.5412953346314449,0.21790392262794456,0.5412476817452029,
 0.7401090596828097,0.8315538473938627,0.47445537568646157,0.5155451502698137,
 0.9470396239732028,0.4052859898193113,0.16488853845174223,0.5459198431694166,
 0.46037603351478773,0.9505001230304788,0.010130218534143198,
 0.0759249992300246,0.007940762512488875,0.6246825759631205,
 0.6383052634497287,0.9739855196493741,0.9037023484976225,0.43158083392803015,
 0.7966318873064643,0.005864437643037235,0.7204344129062952,
 0.1914332049648264,0.8672109946312335,0.8378576088207107,0.10643657570299525,
 0.574475290535045,0.8283147469310889,0.593038358560116,0.17218137445620985,
 0.6295443722527564,0.10854078760805808,0.9828889938563297,0.518559468729463,
 0.8306552765438243,0.8134021048485964,0.401477659044592,0.44502983701685805,
 0.9517408512336731,0.9619219657712595,0.8887253450840906,0.8204686356218017,
 0.6338636422381273,0.8923374800191136,0.8701528280978907,0.9024392221601139,
 0.8512280612478555,0.8906562599361513,0.29625385748126043,
 0.17034428378341782,0.2850701773187255,0.05423290817776105,
 0.10286357083790432,0.7988730350069742,0.5825925347734329,
 0.41623447624883547,0.39595538011980136,0.8685989906410072,
 0.2210594599493355,0.807780721875294,0.9681228879150746,0.21113091207204493,
 0.36124638773613105,0.03422265324898077,0.24856288876468569,
 0.0919986766655565,0.8844032796452173,0.6795358613296687,0.7738966389007128,
 0.6820697132495327,0.0773673542502461,0.37506935920560236,
 0.16685299772410522,0.30171524915243286,0.1419558139004411,
 0.9690997461762526,0.727419526196951,0.6864000645901102,0.9259309785298658,
 0.865576449854406,0.9211108781416733,0.3606490087243769,0.41728465478948606,
 0.8169945644969282,0.8994895384440551,0.2174548785310466,0.582082331838968,
 0.07981771987995911,0.3411945598094558,0.11560050472070205,
 0.07455277085271261,0.6528789062030182,0.5049436036103355,
 0.35023613612192595,0.2955965905818463,0.5653416333150054,0.5200808680538862,
 0.402706845463749,0.8860919624705995,0.632157910305784,0.4970666073585288,
 0.7590293878632879,0.07404336932641387,0.7849713326250405,0.8164899388493256,
 0.2726964827255678,0.6907424403076508,0.6111453334704549,0.7925255823491858,
 0.30933957134429824,0.8458814324012985,0.29057536671168427,
 0.7249302308528609,0.41541678315033337,0.7097405417903417,
 0.23227842291801726,0.6367241197860528,0.8200805154178517,
 0.21047182541024012,0.15679859675915941,0.6030165089037599,
 0.5386389337516945,0.3963095258773497,0.20718809088448897,0.9562558142488452,
 0.7311409717417519,0.7983070805998621,0.003518748289550011,
 0.5814947154278182,0.4456402780922979,0.498537525605776,0.8317851093998805,
 0.10395682131655537,0.15809000269516282,0.21233851285595284,
 0.6160341429344208,0.6937688460781525,0.01303756735909456,0.8428088808748113,
 0.9364192340227306,0.9694235746697696,0.08581412133651989,0.6438850233257901,
 0.3560497175112931,0.6320265450860576,0.8333035697846709,0.6234046251759375,
 0.41195922406331686,0.8482888813951478,0.03428996310809995,
 0.028158470596599705,0.7639880782657398,0.7954681813198146,
 0.6760054351010536,0.2639244305652113,0.906354456826975,0.9532673998293216,
 0.27343750193680405,0.6167058375180855,0.41701099516378304,
 0.4380387189122451,0.532080229497831,0.5153050760418464,0.9212239099131823,
 0.13116998061982743,0.8620817759316018,0.5668323754028501,0.9398749833714868,
 0.14565261616707792,0.8809124554614154,0.6905586158089586,
 0.012019276915456123,0.7003800475031084,0.5430349290499404,
 0.2767099741613064,0.6815789003992014,0.4317730230951339,0.8044465191195344,
 0.26826816295976885,0.37690029376790746,0.37360951579029644,
 0.7569938370462872,0.8502039271778352,0.5391688663404817,0.892083509795108,
 0.458937951694139,0.7003031661143674,0.20126956409484775,0.9007271360164416,
 0.49449772009325155,0.9288813825567995,0.705032052033541,0.09668804119214247,
 0.3856597686518679,0.10121082032154016,0.4400285335098987,0.7679359888027562,
 0.05514020254852203,0.4420159848491497,0.5048752461594689,0.3831221510413856,
 0.9768009081707666,0.5065459303141113,0.440795841510191,0.34828840992336774,
 0.8925438108665441,0.19389707058654926,0.7235033631367196,0.7273129216470576,
 0.29841782387489735,0.5411547610640594,0.16388168292137817,
 0.7500353678591719,0.3251535364359215,0.7912415847416604,0.359910453541576,
 0.8560748193460126,0.13092778757923973,0.8007759029267975,0.366152213978312,
 0.07427707686146812,0.037520355619862444,0.44992373939354957,
 0.6145016340631898,0.4387816777975142,0.8469562518620487,0.4978074348053849,
 0.4065252200791971,0.6021426532857528,0.04016973038851945,0.3154442493708991,
 0.17225414799359373,0.7626756620287477,0.5667327496019121,0.3382126732851557,
 0.03220345084702736,0.3925843545390535,0.9843773452882143,0.3380313653638456,
 0.2688542162636748,0.3865366731958675,0.5134813321620457,0.21052451213479728,
 0.872614974762693,0.6581086295820361,0.7230662831551478,0.7589667052189013,
 0.48714206854854136,0.42205146476693856,0.2620575036565391,
 0.47109074415329943,0.912178903709779,0.48434661638560617,0.7740754384908668,
 0.5941974430115525,0.34410075755729985,0.30711976977963895,
 0.5844501717599043,0.31750094480307345,0.8886451531914307,
 0.04544204248477668,0.36276889071428997,0.26306831254718155,
 0.6277018122343172,0.5851980074670523,0.4442209087452925,0.5318523071259651,
 0.253669197361884,0.7350496697652363,0.6314849013255,0.29140619931873435,
 0.6311457994016849,0.03259146679351066,0.09824506737565797,
 0.37075270645593295,0.29892345963082134,0.581100982889289,0.0832550556753433,
 0.609073296843986,0.1578959183644706,0.6101636731176514,0.41718045078389054,
 0.7374131016742345,0.6730428203855623,0.7256681933056376,0.04483524512235704,
 0.4125175914896152,0.7759271936490648,0.5535203104429808,0.5206043056450945,
 0.9075796743896244,0.38886776629154896,0.22509720985872883,
 0.29088566155195583,0.18283570569111285,0.46637098086187645,
 0.39159119764159134,0.10838164598356836,0.8611483369915864,
 0.7231654980418359,0.62056385403545,0.9052700694014789,0.9723728310899489,
 0.9374257714407973,0.3678153045151171,0.6059107053733142,0.9999766563802139,
 0.002606183943762108,0.676656129873073,0.6330584697385486,0.2968922155200624,
 0.73160065814816,0.41339582613190196,0.3838639312076193,0.27071392091741686,
 0.17106027490859754,0.48538889201390867,0.5942005461044767,
 0.6356904285976489,0.4387584446115881,0.06747953426520514,0.6219620178938554,
 0.965646189725765,0.7325499541494206,0.12249363270568336,0.6884962309261276,
 0.32663718379186457,0.9266293252536258,0.7245521803722783,0.9951759170838542,
 0.7979648726490938,0.7194748180065729,0.9646912168795644,0.4560991231626992,
 0.4179259534338715,0.5784695743592612,0.4312371345471999,0.9655562873734407,
 0.08555646216651214,0.9817506482764041,0.47510682888754974,
 0.8707894821587817,0.4691183929156546,0.544037363570068,0.864687604471429,
 0.5822736310075178,0.81918077944716,0.47994511987649635,0.9406373593409838,
 0.2680507275150654,0.9049257968806118,0.471630376051938,0.7376203823833755,
 0.8844973311988465,0.11313040596695101,0.6160963202796927,0.2640834232303446,
 0.30331038538062394,0.26661582452465693,0.09292150337571237,
 0.5336589936513819,0.38300681530744574,0.8087095100759029,0.5175015021032715,
 0.46469887648924657,0.7977045196423291,0.20055123773882244,
 0.08002351609773362,0.6523378993924733,0.5106471183357721,0.2545689989413915,
 0.111796635173525,0.353029834722022,0.918371808127618,0.7192148246793387,
 0.49692841006185484,0.6105155604689324,0.9815856681623563,
 0.45876417841083583,0.18117323373539096,0.5273077668563999,
 0.5016551543936412,0.8959006580550302,0.2643944318228285,0.0688938733891129,
 0.3688001212757053,0.36891176457912833,0.6593960534067389,0.5975543623249289,
 0.414969172168702,0.8981568948956362,0.7557011153404604,0.5002648615316069,
 0.04417212055352637,0.16043031728432444,0.17709889681969448,
 0.17234200985581605,0.8517532271357449,0.9373802879932996,0.7026198503739303,
 0.6897706791552732,0.5227974789374114,0.1685933646271771,0.6123856757309539,
 0.0958526299180672,0.5232195507729757,0.3676332210736064,0.8262174273505043,
 0.5568599789688012,0.806836180189784,0.13054572188546765,0.5840029252972798,
 0.47569306652164167,0.7292189279848651,0.33584321986537624,
 0.8358385479254844,0.33202689570051735,0.8521811668022721,0.4955125541754213,
 0.9007252067522385,0.4601735253568151,0.4076805141524812,0.5383085746707903,
 0.6465172262791486,0.7738682938787955,0.5018889924796588,0.1529675513702513,
 0.8181220771585827,0.6982363151042935,0.7590683836608543,0.9952842073355094,
 0.356481521293354,0.6578341891745301,0.32308347911912694,0.14741052405302657,
 0.7048425120488262,0.8709766533740521,0.5913081390167547,0.2264934354462783,
 0.28442461002984576,0.7554515718340297,0.9406757587132524,0.9639350065962573,
 0.6752245045655463,0.6601157613720788,0.48349235844456173,0.8114188848298857,
 0.3470182932277891,0.1589817317662945,0.7207921101976129,0.4905226513721751,
 0.8708760467627489,0.06127850591225237,0.6046663260024396,0.7100709260015664,
 0.12179898395853139,0.6646838475249685,0.6830349510438555,0.8795651186607947,
 0.25649319985082486,0.4488398418900019,0.8824657276039619,0.9392112428414668,
 0.8351910112264236,0.9719491155770252,0.2532608446964808,0.11577450146822477,
 0.42070414508287457,0.21081991271364575,0.6671867680126231,
 0.6013458355740827,0.08293918913262344,0.06467694783620126,
 0.29083528571564554,0.4181746395444462,0.8095685092582797,0.8937293015891893,
 0.8967914729150267,0.8081593246087876,0.6851301120067711,0.8749203016969241,
 0.21438987733493964,0.6157965671306924,0.6116216084217596,
 0.013565921243402057,0.29968121487502686,0.46656476367048205,
 0.8904416244981491,0.10719633994624933,0.3903343284351788,0.6951339892364512,
 0.9484232601435575,0.2996640018528205,0.5081755122011891,0.28852924520308376,
 0.6788132640238349,0.9318645684858597,0.5162172530796947,0.5588417769566804,
 0.7323694921684124,0.6800752378105717,0.1818685329790497,0.08764722490288246,
 0.8228847084478819,0.13649801272475937,0.9640287131738596,0.7802067378893143,
 0.6377492477548061,0.07394482123994728,0.898471795236516,0.19409833030038015,
 0.5650416834250503,0.4074268262335098,0.7771977469013394,0.39677587672150716,
 0.2467208779451998,0.6906342645696363,0.2673372272472817,
 0.006331607088159608,0.7830950721220881,0.007815484696863328,
 0.36419073456088413,0.5687111976463742,0.6729403273660731,
 0.38782547553477564,0.4812745220876644,0.4844311797424038,0.5400965283656529,
 0.8558149211838667,0.878305461288516,0.8963893154677934,0.6433169623640671,
 0.0802679184545384,0.7827517935412334,0.582054818285829,0.6022486765573657,
 0.6624617982156334,0.093528938559164,0.0657331961177029,0.6672482722840029,
 0.9909887073706259,0.3221766547738002,0.02298461295286025,
 0.025974497494268833,0.7466959811116198,0.5206584327058499,
 0.6786401872267025,0.053993650490196,0.3506280780017478,0.8308704815266943,
 0.44841581379079154,0.980109671518186,0.6552472920875275,0.6974941345291273,
 0.3107345805793731,0.020710516406213797,0.5344763774414584,
 0.8910839916486151,0.0541076339513461,0.41036747213632274,0.7991132392105451,
 0.9087160795949227,0.8323278225813514,0.2792252769069431,0.32005460667944907,
 0.39850843275770653,0.10285581866568827,0.5577443664061136,
 0.7578502261119402,0.4704916568259132,0.18799570037979207,0.8148168667577257,
 0.671717099552419,0.02088852894839921,0.0756600694349222,0.9925295765291934,
 0.7485971751534108,0.5355626863805499,0.9801688264473505,0.6608088837509464,
 0.5073330428473422,0.8803795831960658,0.5284965659343194,0.5141855271549055,
 0.28581814321559373,0.498562094219001,0.11326766260951682,
 0.13027862566220672,0.21518214081860898,0.33004752169597973,
 0.7101456861200379,0.9415298512200712,0.84537596825475,0.676404918864714,
 0.31122651605059315,0.8061444293193032,0.051312518611360436,
 0.3665279445998705,0.6489331640986253,0.3551553805535208,0.5686735543788356,
 0.3233910248716868,0.07176618300696991,0.004895606166244737,
 0.6178959810162037,0.2119930037725185,0.670503022934811,0.6769134427947316,
 0.5289115241995921,0.3386035491218212,0.8146910526041206,0.34519508687063283,
 0.8439155764996681,0.914183365526186,0.9799770365375232,0.20995395274453732,
 0.39228075497648274,0.2764649016798105,0.5143572732597506,
 0.12373221815892366,0.8562077477913357,0.3965202202339808,0.2923639411151504,
 0.14334872065365578,0.05514423202721197,0.24813877329115885,
 0.1769191360377519,0.055350489750850884,0.11727404813081288,
 0.30473738910343595,0.7430162027640042,0.9132255714642705,
 0.16997424453522225,0.9286308903045133,0.9686683356529344,0.9613066942329901,
 0.3964182131466344,0.8813368589431667,0.4984935468108104,0.7240543068986252,
 0.2805878611406445,0.9709402140328118,0.44806704575759193,0.3718098757823679,
 0.5025571107328863,0.011280126325812101,0.22314645271998534,
 0.9780959911150557,0.9556196211936242,0.7720612789949128,0.29328074417183037,
 0.8129364435673252,0.663574108751538,0.4440760470153611,0.33807798704968506,
 0.7589961964121338,0.0023733333755368013,0.9675734209249665,
 0.6198901708270341,0.257132862611011,0.3311131519189827,0.5312465794082902,
 0.1003965788020953,0.49999145034346326,0.8865437667275401,0.4276515614177341,
 0.9737952085726428,0.8393600966134948,0.09680554819341058,0.7731647340470149,
 0.720089882412868,0.8645940840439057,0.581241429210683,0.6261641119155105,
 0.7720801376137926,0.8342718040492131,0.7972024544217748,0.08137468224750877,
 0.943690644861755,0.5730333956803236,0.9399889804526839,0.5552628030418854,
 0.40764463671235085,0.8021321933366791,0.9093901564543734,0.8279328459742652,
 0.38549280169603917,0.641366075785714,0.7490907129601566,
 0.058726441932923645,0.012470381921994345,0.06484738750323049,
 0.5107812648196178,0.47856901755914283,0.7004865739769549,0.3086884614001373,
 0.1905777693793077,0.03208673499351711,0.3956079909201481,
 0.10844879023712328,0.33954123507517986,0.81660374654097,0.15830813490491724,
 0.4867136406782495,0.6631882909567189,0.8990272828751101,0.7192767367498631,
 0.18813521974151093,0.9717711646618894,0.12507514433825295,
 0.07060861735791257,0.7125753057592401,0.18019473303865963,
 0.5244464609418389,0.8749924195104477,0.11262660573359473,0.9172316692907636,
 0.8570299790581913,0.003455063046472584,0.2639093094778956,
 0.6048860516214143,0.3066694706484947,0.7767500938947466,0.5790996420761005,
 0.6465863605417901,0.2253526712808137,0.5600940764821273,0.26025984149942916,
 0.28460842913967377,0.41397558445313587,0.8615836202992893,0.886140694463565,
 0.7300817219493787,0.5941201905072795,0.8343363460204584,0.9256987110365787,
 0.4209514064304789,0.920812074697866,0.887984411395517,0.010469215038815816,
 0.22196344633374365,0.36598355698384455,0.038709038309439836,
 0.8998706342908532,0.7358890612136788,0.014172090679502314,
 0.3378567923963469,0.14741501899063714,0.6807546621680928,0.9957967540741572,
 0.04236296262338879,0.8719739021155408,0.6849811119923739,0.5224010712610407,
 0.19569890461292494,0.7767039741218753,0.29477881996185173,
 0.8826571830089928,0.5116349694294031,0.295802162524888,0.5207306451972725,
 0.22556349983373036,0.8815947513650363,0.8305125601074664,
 0.34910198725326747,0.254113531113052,0.8233453590806145,0.04978160458811909,
 0.4273850243327719,0.37539361844219854,0.8943670775493392,
 0.05938213185507324,0.9845670352867348,0.19697676925631746,
 0.20446508575911393,0.5824758207544374,0.3018821215638978,0.9094498954296345,
 0.9455796613183397,0.005004724213648504,0.7032465086482843,
 0.10443648134282657,0.8907000168116574,0.9683025395377385,0.7093437356970107,
 0.48780353171569446,0.04549437387030819,0.9642024412391813,
 0.9380917706970748,0.8580622455477506,0.927182325292086,0.07369033450771767,
 0.7171363130795441,0.442158809107038,0.25649946252777633,0.5637186337230872,
 0.02713251021986396,0.5333140156526016,0.31019177336102755,
 0.4289643088160463,0.3939633351648444,0.9366534120856959,0.7906211909348009,
 0.1974827962496598,0.0718581876062293,0.6490598204372598,
 0.059550101626109964,0.9113093588456223,0.24291971776628707,
 0.650817697007446,0.2670417770709592,0.7112038944391085,0.36870834286612597,
 0.24061421775574865,0.261493380156435,0.3530400933067499,0.8244230920903798,
 0.3321702261447106,0.33274017349568474,0.6404808788742857,0.7454107870279273,
 0.5672314762950351,0.5642842943403917,0.25822482325496254,0.609496292099996,
 0.043131630017055,0.10943739572833744,0.09709810193616852,0.8622335139764125,
 0.8730863509701206,0.5769172740764169,0.5435754902229601,
 0.006448766236410108,0.981635662427099,0.13945271699066286,
 0.5706171658807914,0.8359300862348257,0.6587328782290001,0.6574519571165416,
 0.7795662360284926,0.7244567577728402,0.6158701641475242,0.06409600587774167,
 0.5189735451397871,0.3877191011993931,0.49194794691763966,0.7460914403906761,
 0.6146259899734166,0.39658105744522787,0.36392616414295686,
 0.9703203047499399,0.6214283356679169,0.8152183433220763,0.2399640957504665,
 0.33874011514922564,0.9078604604140155,0.5977318636505072,0.8924460951604509,
 0.9896709574421698,0.8635818393384705,0.8215108041833967,0.30868847432973223,
 0.5005548144724546,0.45919950990098757,0.638154251730487,0.30245135082349994,
 0.006223276780597198,0.6754137334545403,0.3118749472292689,
 0.20434438635452334,0.034027587248198765,0.3391441048044749,
 0.12196335934377112,0.36302329225384444,0.16957907667240846,
 0.9172717975443688,0.9451981483211084,0.21163523550660435,0.1994533311482236,
 0.1991057426063949,0.6053391219009585,0.3380685594478128,0.38324496164930966,
 0.5556048034604669,0.3712520583273218,0.862792060888305,0.5764203951167903,
 0.7312960343804261,0.21283497724893174,0.6655480270248988,0.4090298998632852,
 0.8301330921856581,0.3372336701940827,0.7975491681432596,0.27255338056595146,
 0.26728609032209705,0.3438784683609014,0.9703185611796847,0.808101126707803,
 0.18204812241761625,0.8417537063486035,0.14861460660268888,
 0.40608753915537477,0.4994279233054635,0.7638118386194548,0.6907075901630403,
 0.23127291672018802,0.3475768141720459,0.17149303570181784,
 0.26539334066304276,0.5538983109639102,0.580568808884171,0.705585966280287,
 0.2775051923240255,0.05607155704829514,0.016817871517182614,
 0.8225471259711242,0.774116662098819,0.7356410853064733,0.46933014985186905,
 0.31221536565869723,0.24926558720732972,0.8339798787063791,0.760654250945012,
 0.22560906239463263,0.25953915936436,0.7521018772533776,0.31930921184782457,
 0.02634133734030597,0.869718072953018,0.3060696552348182,0.7088148644857583,
 0.8733301464076162,0.6226473891518847,0.9676796160450929,0.3355346388646844,
 0.5367982127715389,0.9386669619181481,0.4354083434677367,0.15036416318416135,
 0.5668400276921977,0.37707608457332165,0.6447064205815755,
 0.018267442533374867,0.3055642646703576,0.49701944841217705,
 0.8639563876849984,0.29975375822115446,0.1909961417015812,0.4807636108805894,
 0.5088775611146584,0.4128754691211455,0.010430536495141363,
 0.22936565653233587,0.6714396401527423,0.22242833168365428,
 0.4552652128605845,0.75519832280215,0.8945892411996351,0.5542428474411274,
 0.6251996656842607,0.24165842939510374,0.5884908787699186,0.1609620928630915,
 0.718156712251825,0.7620130890008536,0.11543446114209033,0.20644514401722758,
 0.3525706171748585,0.2658093386576845,0.0781195644526348,0.10983292152441537,
 0.0419378860548858,0.14941643555457218,0.8926647912694379,0.9333725176623034,
 0.7944785243704717,0.13665650473842805,0.7785701976047403,0.8397110366869052,
 0.813983055639173,0.6213002499392603,0.704733417523915,0.4401835472978306,
 0.6787793671698903,0.4175085907767325,0.007133497878701434,
 0.49274131389561926,0.7210263040341666,0.5038658048435023,0.6830903666278102,
 0.00654179528702048,0.6779688537287322,0.768769368365429,0.9375682957860094,
 0.7233072961596985,0.5313774081450183,0.14532026229300266,0.4601848426394096,
 0.7764645749936239,0.9546977228053424,0.9335922457260318,0.31533921235887374,
 0.15610386958570988,0.22495572169246003,0.8349305947742024,
 0.22463993465662901,0.08127779407646696,0.9163614956435023,
 0.4494366978484403,0.09098521554897276,0.42961173779738315,
 0.40696441757964286,0.1426978957840036,0.0665748690082213,0.9171067034788203,
 0.5062029439372194,0.8532793621124316,0.08396805665830565,0.8051729112756288,
 0.25395025804139715,0.04787355547076233,0.8339097023734189,
 0.07117944634726137,0.5382268708443022,0.6753114950248822,0.7106863496936587,
 0.3694361769576967,0.33963807124401,0.654448894176441,0.07549711817684102,
 0.2730396920221537,0.2834159106216976,0.6573990172543116,0.6864952063178849,
 0.49341907058526036,0.3505633692490333,0.8155543409661203,0.7689311546002838,
 0.059947502122224106,0.2920990936930461,0.4711603369536065,
 0.4234431218869489,0.011864723202110072,0.9444024643317281,0.560664445596237,
 0.5946771854583285,0.35330756709927735,0.8263670147255912,0.5628263093150199,
 0.1932121650122406,0.10841837465240178,0.7113913366451565,0.4177654743527006,
 0.050388451469715156,0.08736958804110806,0.3036628654891538,
 0.7263764706897872,0.405804983447022,0.8595869076104895,0.43935835274778745,
 0.1889044396254247,0.21171461324733143,0.6834170727861921,0.3121068855206891,
 0.12089254406563288,0.5902018388153423,0.8931142338119362,
 0.47706662125717714,0.9558843630947451,0.007499453939876832,
 0.06055064113290842,0.12672806386380608,0.15410908254178723,
 0.39307821675083865,0.7252772319622715,0.12327990403376432,
 0.7219161078834807,0.5618202637114678,0.32355052585843613,0.8923412961860712,
 0.4561825233379653,0.8866015632705517,0.9973593858047496,0.6670158719996596,
 0.4495249191578796,0.0015257357104787594,0.046064173079363835,
 0.31037793747079057,0.8919746536924873,0.4467729434201789,0.6960634896622921,
 0.8117139261391791,0.2479115493256815,0.9806598399946053,0.4990061990422604,
 0.06081231013436261,0.8798782426928944,0.3907556582744589,
 0.07548947572492781,0.6414343862775437,0.08577394276913364,
 0.8240734892519727,0.028376925484642768,0.3221292872258853,0.218071414725718,
 0.36689914265046564,0.6249021051261385,0.12448549696025535,
 0.36755774232999905,0.16598881186946735,0.6748014121713304,0.272324644730079,
 0.13309352343365366,0.9610563069795419,0.7634463993818065,
 0.19954180919253706,0.9186228733325964,0.23239557593974447,
 0.5108212484991018,0.5087850954829716,0.44580699568993687,0.0494570998734325,
 0.8902643184895599,0.015739375263243538,0.7969909929410018,
 0.6510118074126452,0.6330504124148308,0.33768232602865056,
 0.46038254794800404,0.02723221867129677,0.22076343395958908,
 0.28137676715173254,0.9460927897247855,0.8260136583168293,
 0.45796861403977396,0.023763193709805397,0.706898943045715,
 0.3515439657435343,0.8473082839628999,0.6430839244024233,0.19685691788500093,
 0.42998363581545085,0.5652205655632083,0.5727216116905145,0.6480726250607999,
 0.3436319394214824,0.10344368236301571,0.5422936995921878,0.6037485317925277,
 0.7115669532029458,0.3985082211334716,0.3710260826333036,0.1964774261078751,
 0.14885761698880406,0.30787928458575364,0.2590112234668659,
 0.39240253893542243,0.4115796899417542,0.5603129892636795,0.9847135352693566,
 0.1493699747701771,0.9537588441318927,0.4633278731652938,0.04991691707821111,
 0.37699620805619827,0.7577189381262303,0.2207369972788824,0.8231100086226012,
 0.3262837774254157,0.7295080001422142,0.6359579375030145,0.13013339967886173,
 0.9398059721985835,0.902264815643018,0.7615925802499289,0.2698092818125012,
 0.05239831608770429,0.394688033173589,0.12273576888115023,0.681279076646348,
 0.7147139227088494,0.2028982119490732,0.7281521783760777,0.0794602190327025,
 0.7816177295313962,0.49776815900912796,0.21931370656554006,
 0.8424204524692797,0.5235084988292763,0.46385120183950734,
 0.24913948926303697,0.5676280200932952,0.8422663812633746,0.9817966233368407,
 0.6565815212135644,0.2803886329091402,0.14952908422142286,0.4673054131436538,
 0.4000818980415253,0.9404986494781701,0.3741055758939721,0.8553767754408387,
 0.559106928092619,0.4741171549822699,0.8660289442074792,0.13488799778272298,
 0.06920218062096972,0.817720019986407,0.41731570253514505,
 0.31099312601572626,0.7918391586299091,0.9394721092003195,0.5934985123535014,
 0.9723955303513193,0.6742380605422686,0.44907143891373236,
 0.07894786481213245,0.18790278004126604,0.5651276098304925,
 0.26087948211209844,0.08471014255426423,0.6917119828281111,
 0.9728772522505615,0.9268703821835448,0.05889536238324533,0.5664866444393718,
 0.350217173641471,0.1090493194955039,0.8046252751596499,0.22898657131775846,
 0.43144772142093846,0.6831380336491785,0.7227153222866913,0.7676703991077154,
 0.5164925058268064,0.1619312771549043,0.10169284687949554,0.2534688877229706,
 0.14759322808499586,0.6604578156428257,0.3934345106898751,
 0.19787111769760035,0.2429635141479476,0.19485564193339722,
 0.6550441758676795,0.8982492480050339,0.3281717473360123,0.7193074574465977,
 0.6554869088530692,0.704855653453682,0.6397243672384493,0.6517972729344021,
 0.03634102851761267,0.3014341478053677,0.2412001995555264,0.4400240193634075,
 0.416588438948583,0.11339007798203449,0.8477295305049342,0.7089828834151474,
 0.37930560958209214,0.8729754116711619,0.2584593194118804,
 0.09798907628188014,0.10869918284893565,0.024814655617382164,
 0.8922884321956754,0.4597764222387186,0.014626695447455984,
 0.09399021929891893,0.04122477911585376,0.5340842548219926,
 0.3974067896730371,0.4006621487230525,0.3141235877987072,0.8705694705154636,
 0.9154705827944118,0.9396979114212466,0.17799167921000814,0.792790046407944,
 0.966063210221541,0.3425913807003702,0.7407233759483103,0.6686619231910184,
 0.9239173412786067,0.9820150343266134,0.9833570072453302,0.48492342193746873,
 0.37037281761205154,0.6199604299020182,5.408245768131481e-4,
 0.0735032968512126,0.2822035704432754,0.0980068593484652,0.5158165718860375,
 0.13247337478169707,0.03735326569367692,0.9069772029432634,
 0.3865443543670821,0.584287887355048,0.8934328548626593,0.9237549015144804,
 0.3121619555078794,0.9011544006742656,0.6300381271363266,0.03741891122267549,
 0.12829829018674632,0.5594699994579562,0.41334324012150425,
 0.7800331574742272,0.019510636732485542,0.9007137434328087,
 0.7501802699418368,0.8717179428101327,0.04371694717375796,0.9803025651320364,
 0.7280326019618112,0.30641027772722274,0.9476587508074494,0.3368109857692051,
 0.465720622119701,0.39305761220837776,0.9998454274725823,0.7901632982583807,
 0.6293765049427056,0.719679324643579,0.9397654014010168,0.16107408516803878,
 0.575049497022606,0.650101915911452,0.3513354576741706,0.8949783277256985,
 0.3866588580463147,0.43541144543234445,0.8955974523393655,
 0.17482064042597878,0.48745338857449405,0.812497692288654,0.8041830839095978,
 0.9629895210456202,0.7501541024065411,0.7397385395519289,0.03067963668419793,
 0.32788032027839675,0.8731927118343766,0.7294927147159518,0.5199310132703063,
 0.4923828826410952,0.8274713986877409,0.02670402942865069,0.7737764122740013,
 0.5209118454498566,0.5827290672311009,0.8835169541050198,0.7624981979230813,
 0.6095781271850141,0.21289042505162603,0.5254149492647606,0.3856784247252659,
 0.6074289926642543,0.7021632916854497,0.025865603105514,0.7022467770386075,
 0.6412542750817093,0.7483580973850261,0.7505902916368749,0.960645665609978,
 0.3177265411904221,0.947104323538732,0.9009499411367065,0.7821387822495588,
 0.14133671058342034,0.4167854230602194,0.3117126789536898,0.8479311881785482,
 0.7708546650988918,0.714285590676679,0.47434868907513805,0.16457031296443025,
 0.13460811539048856,0.9142558540533211,0.43466302320289585,
 0.39630216671632046,0.6752108537116107,0.3222937837023151,0.0566092928817622,
 0.592492370027148,0.5135723647642317,0.012084433719703802,0.5879766599277181,
 0.017294871567521408,0.15406416449052807,0.5308803063786596,
 0.6755526926237522,0.07298142948170827,0.6030043510391838,0.8752584220611059,
 0.018444196334971408,0.07056008217514043,0.6526793951139892,
 0.5183097363228918,0.4097670060866402,0.07387731486223759,0.5198664016511875,
 0.6189948413482476,0.9082153574466334,0.02707512770597542,0.9452131496570905,
 0.7451621676820458,0.16034033171216455,0.1503092880097071,0.8201781746861164,
 0.5149609920057243,0.527755258977742,0.5843800957003382,0.13139531131209403,
 0.07576518956626499,0.769643526795278,0.806608525315961,0.8870413926004758,
 0.6257593343972963,0.042238108500177396,0.10255087253714712,
 0.5798866740657921,0.6278528462081532,0.2195600754417929,0.9355932158981917,
 0.7682432195551083,0.5589723949105768,0.8320416129852757,0.21528057548556634,
 0.7171423495228868,0.4313078089664142,0.5442764727567471,0.6513474583715639,
 0.27266750827199815,0.9746464764592352,0.5514385664972128,0.9051134566927461,
 0.010045798091895985,0.07106575751098732,0.7312571822903436,
 0.8195857625776243,0.1454579999392579,0.5381430095934073,0.7291717398583778,
 0.32464081250454757,0.4630008664245785,0.4749083628881343,0.9479156872449307,
 0.2823835186179592,0.3210914327906538,0.8980191502457376,0.4859171773500893,
 0.7665281938738174,0.4667432591582593,0.6826359882544267,0.5359750296841046,
 0.7200767169438891,0.24644922944163705,0.800911308564828,0.4814992458587689,
 0.7509044456257824,0.671493553801149,0.1507624689100031,0.533901120841522,
 0.4667818075537692,0.5691148351576116,0.7731009347232549,0.5499298477695256,
 0.5329869042639216,0.2933550618264894,0.8864793397457261,0.9221463390526174,
 0.9146448914339593,0.43997454944157444,0.6690062555513682,
 0.02672924663850451,0.48240533250147954,0.08098240090020414,
 0.8628057313201176,0.5708458806905764,0.9179610335456895,0.6221538655769553,
 0.13620475717222824,0.7878426046053387,0.1491087863605205,0.7840448416364415,
 0.49505953470518327,0.18954802114249925,0.9984263439814326,
 0.6759263804768209,0.16775447890754902,0.8011820940584861,0.7819496098069133,
 0.560302970756302,0.11895464200464778,0.40420379825470176,0.310912023222659,
 0.47275496075501766,0.3528982431002866,0.5690721205021348,0.5929308894328091,
 0.9051420545929142,0.1380100882225429,0.8890332532491483,0.21578594332701861,
 0.7402269474799691,0.9283999744234064,0.03676686126513151,
 0.06342289095946096,0.40715913702905393,0.6753980104274044,
 0.33694435285932167,0.009947854576458859,0.959122415604736,
 0.5734230507451864,0.8988885342339715,0.7001654288585689,0.8409282401574885,
 0.5733497041982084,0.12892998159875813,0.7023757203191877,0.7508462765194281,
 0.12388271686791774,0.30526826369180315,0.9035065348720949,
 0.3328885888552905,0.6466708746123918,0.21882696939457658,0.9690274500781328,
 0.2715514460479115,0.938707081819462,0.46284210914545,0.2396659149365895,
 0.9515638000195223,0.42309030538798154,0.49481199985771807,
 0.28412905929027077,0.8010730683191482,0.6834613994215938,0.6202570921755285,
 0.8500571264920874,0.5540560284297138,0.8930909847685733,
 0.027808364631450755,0.6257487792214688,0.7850661130901623,
 0.7406241539337642,0.5042182630747306,0.43684199938500523,0.954329601104155,
 0.6364950926206508,0.9179780042754522,0.51172183002644,0.9644340425120994,
 0.5111246762793316,0.6929183467683502,0.2111666388443314,0.3106050334453355,
 0.6268909008038315,0.9817753569553681,0.024413367077828485,
 0.3437917754418194,0.08955974078711448,0.8641534974200713,
 0.21160347234244725,0.8672784485913505,0.32374483180101454,
 0.9114248752965588,0.5439975285369382,0.0897458030356375,0.3618902012056161,
 0.8096164894351717,0.616293995302458,0.3185362218950146,0.8237755405254283,
 0.8901217686530086,0.5128620231440488,0.9651602696240491,0.24462964798356435,
 0.442197560842594,0.4448981496473883,0.3058425258923626,0.878714016061582,
 0.015309862655037976,0.30677172744975045,0.38170499491203214,
 0.9851263546962463,0.8214827762491681,0.045419937961688306,
 0.34041906191199534,0.29864764534246746,0.3124992484291875,0.897465557404438,
 0.5387961258288375,0.5855445339180027,0.43994051935962086,0.5674535156014882,
 0.3430221185782997,0.010428276069567866,0.24299405895195914,
 0.09014067716441998,0.7571036209203607,0.16039329003412706,
 0.36241767493087984,0.8473066193884576,0.5765658142531095,0.9774645525370206,
 0.3535122952684132,0.5985988054572953,0.43700479595351216,
 0.17183852824378365,0.603091954627983,0.700945897934793,0.04452114816065722,
 0.5394859457316802,0.5475395682236851,0.9341733269738102,0.4506932443881736,
 0.29086951968215013,0.8573364153749675,0.41948764370054503,
 0.44224894357008304,0.09161787004014943,0.9375312242525168,
 0.9675771379471455,0.6590358838120668,0.865981761390423,0.11598588667074583,
 0.044140160187351896,0.14609332552940535,0.47638264217118875,
 0.12361578031716769,0.01796139206976821,0.20646265687261212,
 0.09220371349356205,0.7408660037880581,0.2851845858450127,0.6512681286233457,
 0.06168819322437247,0.2974654537285518,0.5689675107105542,0.3032503502672489,
 0.6441486754943712,0.902699736692245,0.11804426321701111,0.44931020771441543,
 0.5542369965289883,0.024475522419562967,0.6869821211526365,
 0.07080748937202497,0.08126487386371273,0.7921077504020186,
 0.26696265733760904,0.6783250791938957,0.5566820230716484,0.6893404507873341,
 0.9855837370214002,0.04554054320731549,0.16537435946451973,
 0.8765283247071136,0.4972917087302895,0.6320052023777482,0.7942121355422224,
 0.6053420605775751,0.9622970844959825,0.7839602270409647,0.679321131251815,
 0.6449175978595618,0.6744464280157058,0.982892944186729,0.5337373233446666,
 0.5093084988655022,0.6396108833722742,0.029811423485776145,
 0.24762705665216056,0.860041596487639,0.3317468586909341,0.9544405503985565,
 0.7535993956626982,0.9655385257921076,0.92382088203053,0.14841015104607225,
 0.030313525572527866,0.2675062388838513,0.40983964410575224,
 0.5393484420426835,0.2526759305176747,0.2970948987685854,0.5555112711969594,
 0.018093366315547033,0.8638730266781514,0.4462994207173031,
 0.6668114387505749,0.5148361822704343,0.9315074057147389,
 0.040597357208320806,0.4697641160912982,0.5487618636152296,
 0.3986851531648661,0.5359737054462411,0.9454800822582906,0.5358879936176191,
 0.18464275024956456,0.6386029067411403,0.49476988950766454,
 0.5951694124830873,0.08694438330419685,0.9388731978157911,0.1892773425755382,
 0.33480667366369665,0.9277078404046742,0.5352106743466288,
 0.06256460291240584,0.4654278750792774,0.1790890996526795,
 0.12873129998196176,0.6046927056748034,0.41885537561988384,
 0.08196370466053737,0.8264228790545298,0.3525912374263963,
 0.11208981481968949,0.7942512648743937,0.18865532820757014,
 0.46692493552189673,0.7144539579570292,0.18344180961553114,
 0.10328049177708598,0.8425699522819077,0.4746200977480358,0.6489556698771607,
 0.22659681050099267,0.25594270572410216,0.9601962758792808,
 0.8925467109245524,0.7763904309161467,0.6471846507009282,0.21019113804069356,
 0.14736106755427258,0.25535155146785804,0.023360328876619985,
 0.5376289973213337,0.008607975312131644,0.7258298047397524,
 0.6995882490190379,0.3661625347780497,0.3198248620663776,0.8215727715194313,
 0.9243515238230777,0.49879710804994115,0.2802019597068619,0.5290852160562058,
 0.6502921429188776,0.6992786105801861,0.09131341491456113,0.7863583641059503,
 0.7526323720431178,0.13116285234029823,0.8477410187940028,
 0.24814418585036024,0.4472083648737991,0.15403694848176686,
 0.23866648964490889,0.491671817239413,0.7069436493942934,0.848888821201995,
 0.4234389765102603,0.8503129345190895,0.42874010024169673,
 0.22726689814560608,0.7322275028825964,0.3630805424612389,0.2967410222871639,
 0.8566594457344482,6.682708611706056e-4,0.8135274137210985,
 0.9521024363638333,0.6109494769788231,0.307359102804864,0.8034498769633058,
 0.27432412736943335,0.3677673499385232,0.7579702228771665,0.514113014191572,
 0.7338435973728584,0.9713916934680633,0.06017784877518628,
 0.25471010718961007,0.47015435688607987,0.7623237165081284,
 0.48270109333007294,0.158217866131677,0.8875410109353145,0.8327645172456134,
 0.5804720214922011,0.9600792037561653,0.16710205489429297,0.2933244666714784,
 0.21343440462904595,0.3843337342973565,0.8609547092704157,
 0.45365502684169334,0.036989423074957006,0.8168973023923038,
 0.3859352837021024,0.3580968401152541,0.1708573515713273,0.08620428725270957,
 0.6672183228955518,0.8400652086895835,0.5799906995984639,0.6155695361302751,
 0.5452603423766239,0.9417062228912743,0.328000953208895,0.19542418110370463,
 0.32363534173382413,0.7593233543986564,0.31529887557269565,
 0.29413461422009757,0.40694869911119946,0.24238419168899905,
 0.34106034720318257,0.6889163176305568,0.19736427219413855,
 0.2995238181374178,0.17332420752153477,0.9911677956067952,0.8681108453551282,
 0.6154717722779449,0.07656546301645117,0.13805617799644465,
 0.5634834951991214,0.29677262825368944,0.3438148904235375,0.4664847618301663,
 0.9514079546408295,0.3636035192830017,0.3087228505228381,
 0.016435214804161435,0.4251071414853692,0.8394872934139034,
 0.12242564149209056,0.6425059477116809,0.05872764660353835,
 0.8062046756460565,0.05195552860795427,0.20911398132124548,
 0.8233707473629609,0.9861508755125719,0.1272746330030271,0.22374997227717985,
 0.6397643634497605,0.47089583741575525,0.22142525626631404,
 0.6946860878055338,0.4026624894412917,0.6767680282506162,0.4332117906809685,
 0.5870121388676732,0.37221369122182246,0.10722638066160806,
 0.6515277904801065,0.8530864704948219,0.8095445972236441,0.254435124778022,
 0.6914073344484807,0.35780387944712755,0.33887413475152406,
 0.26657404845069843,0.9207784172976272,0.1076299795464063,0.6238478132138844,
 0.4266406088143211,0.7559557477476909,0.817976881921866,0.1835907509939675,
 0.8970676317673859,0.9330115539326985,0.7286293377787267,0.2597137833319243,
 0.36697639559392636,0.9186174515421983,0.058270503787248185,
 0.030181922282967122,0.9308661642414475,0.22731895599758134,
 0.9715742918337457,0.9420326882835127,0.3726455408063275,0.6962105167786158,
 0.7654901151132273,0.37362304611437813,0.8135970982351023,
 0.24597291510356545,0.7319417785449805,0.2716685664902375,0.797413053183142,
 0.7229210014773538,0.5241774043681553,0.7534938230805932,0.5091194077055855,
 0.15358088648877466,0.9085184494590625,0.5843644021364405,0.8334483863941002,
 0.9626476187353114,0.2549291827021334,0.8216917337209901,0.9179503625138754,
 0.0751038819088492,0.7722533495654351,0.7552776997693907,0.9126492648078447,
 0.30789352930853464,0.999438436869386,0.8499858539681432,0.30710310213470926,
 0.5090467020584988,0.6799295799172111,0.9046367813251635,0.8498442063591877,
 0.26386832754177936,0.42642240774173223,0.0736150642481389,
 0.7161477829125485,0.5006369605534841,0.04570490516547787,0.6157551221822372,
 0.5077757925216435,0.2906812254393767,0.18492461885224887,0.6602756796580006,
 0.4838497241290901,0.2899306095998806,0.2756374779506051,0.6637975553189613,
 0.105703237060216,0.44894890135241106,0.7089419241451395,0.1703261546380902,
 0.03734961032668871,0.4633685187645973,0.1960393360707382,0.2264692608308897,
 0.5993841067033424,0.22445965465643292,0.7002605934740562,0.1861820054358898,
 0.7447449967114679,0.44807227526777904,0.7212540655650825,0.6940232557155106,
 0.9012729444629578,0.012838777358254672,0.7665960092841891,
 0.7311759440914564,0.3401756911128472,0.0901761826867693,0.6035894428270503,
 0.1353929160247176,0.2950335672822283,0.7919633250452193,0.7183918596716603,
 0.14485900165308707,0.02250580045876671,0.1859423056516606,
 0.5947406699506319,0.27610496635284887,0.2749619487600179,0.6300976636217769,
 0.9751238909993772,0.9396392802596812,0.179332165554968,0.01565340070535315,
 0.6876496749115588,0.6999357534254765,0.3468658764756626,0.47395674632832363,
 0.38550393948659367,0.3525011683610504,0.4251865577679489,0.8629321425172002,
 0.35089205098023357,0.6529493640880698,0.10536368380400996,
 0.37766321760186317,0.9311742055697303,0.7506974595591503,0.9266808753060737,
 0.5453785427578578,0.6734993071317018,0.3064711659660927,0.07032309453964047,
 0.37466332687674975,0.5137711304721478,0.4000197583464138,
 0.26858413693947125,0.5056753445520223,0.7396366091021168,
 0.14207269053866034,0.9244911003097103,0.8529924205862651,0.510821959025622,
 0.38398319597176855,0.45193600735549044,0.4980591677204763,
 0.8660954077665729,0.9094266833236592,0.9978017208347398,
 0.028013907107354075,0.9778991936104616,0.5038350494066025,
 0.3566370796687337,0.9191923735079242,0.22381743331378212,0.8517657997597353,
 0.873925122810917,0.5460448266211707,0.3569022967646027,0.2387907268527496,
 0.11491029721194634,0.9808268592301129,0.45490881948337847,
 0.3125410345673346,0.9195342125554596,0.2313064147654491,
 0.022348378044926842,0.22695253297514606,0.12689489767518314,
 0.7374803879007152,0.03127399358843874,0.393784570228601,0.8556009554852643,
 0.7535445507173533,0.8819843365878635,0.0745901607103705,0.45809641579896,
 0.43595921660338643,0.9492989644165721,0.907188657693891,0.9760816174783663,
 0.634222410280612,0.8709443167748933,0.19174737471209702,0.6406819707143206,
 0.033474464150993644,0.5182419559885977,0.05168012873056238,
 0.6174733523909368,0.29855735985808396,0.4838577780762181,0.427191997094754,
 0.9994958002385315,0.6841959348871325,0.08231807181555928,0.7947725061912612,
 0.7410401901144543,0.9458897220160116,0.9078656505966348,0.790902782690769,
 0.5086322260045256,0.5382270848596393,0.16435577625301745,
 0.014032711425776001,0.7871044829798312,0.8699821885686472,
 0.9231163592661035,0.5388609141052174,0.20266127335401363,
 0.04863944567324119,0.0973188855046272,0.11726778672385785,
 0.7004017675137486,0.6156023854298284,0.5460066164391287,0.6934246262590344,
 0.6491964762285174,0.31947459581617954,0.7106605887064026,0.4295807472592589,
 0.4900468371968262,0.5307863479341467,0.9838668137868015,0.04243081275129712,
 0.5806555062301317,0.170255369932223,0.15639456993012002,0.2695828399601822,
 0.553635753228945,0.0910439033277679,0.16813316879497653,0.42554087966646204,
 0.12005028342025714,0.7080405536808625,0.9894574606887231,0.5742160693893054,
 0.5681549707591358,0.41052679911205847,0.34034241777479135,
 0.21270190295308788,0.23179993689440836,0.29167984809128455,
 0.22981318455466393,0.7895214672004176,0.14962615949174474,
 0.4007892423136541,0.9416638915354525,0.6067065930410013,0.7910284490168563,
 0.8425204020550419,0.5284869555838823,0.3160035392362798,0.5706553041862564,
 0.08045736174359552,0.07850250870657349,0.013521693862383555,
 0.7693912867349467,0.6387474910193747,0.04982726985375063,0.6938329355101542,
 0.55583714770114,0.4864293777577624,0.49411306136109445,0.71619437438663,
 0.960281653387966,0.06705462228337189,0.17215778999617215,
 0.04367223339272153,0.03737851022757688,0.13497965857295835,
 0.5420538837719577,0.7120319083482612,0.3442052505672165,0.20253735276894114,
 0.26331251728340166,0.04175864635798032,0.09431482682940451,
 0.02626378538257723,0.4166526803408951,0.36383987804525675,
 0.8544563542679929,0.35863272370525934,0.5971251526628044,
 0.35638042615527366,0.20783018846108725,0.3964546718329458,
 0.8718090056999505,0.014832805739004473,0.956554850360138,0.8944925219392942,
 0.4759294354749041,0.20162978980115198,0.387304755598907,0.04635326454878219,
 0.8931172419796957,0.9703701289011815,0.20123721627641777,
 0.11499198975012326,0.42283742205120556,0.4646404273312269,
 0.48396769990272537,0.6708138284823149,0.23034219013449198,0.404030430408054,
 0.8907447002870137,0.3046130997537627,0.8420413631765754,0.5607095848773082,
 0.7388521798525126,0.2621981105126211,0.3213921035955807,0.17942987223429196,
 0.2025333636533292,0.8141883507522687,0.8035248116143892,0.6096353324539789,
 0.15489231409877635,0.25800419960513155,0.26208447203307583,
 0.8194461795737037,0.15587643791957928,0.8427224248759082,0.7183209076091923,
 0.26423411937302976,0.979619149262211,0.05118249790963936,0.322307007235794,
 0.006577373510375684,0.04833304901229318,0.08470574780388784,
 0.28531830704612804,0.4931362337173504,0.8224192050712438,0.7378320722902312,
 0.5238697962938844,0.9658978236595493,0.47210408664109793,0.5925437232936657,
 0.10738901259261657,0.6763098007310336,0.18697762793837658,
 0.4281599787206095,0.5966845039610308,0.4506937831646257,
 0.033554355419681436,0.23916181498701528,0.8390325363754563,
 0.09008424638777213,0.7669086511144706,0.37508960729582586,0.065369299606713,
 0.6381607469261245,0.9713872220396941,0.49183597286226133,0.2666700369868561,
 0.7933572412122472,0.8513405645375189,0.5371132813589005,0.6835014969880453,
 0.9457172089767133,0.7465911837119156,0.6587332932191237,0.5427031903920551,
 0.8019165083700971,0.604727998966975,0.6850768799438989,0.3280490979004752,
 0.24714547427535438,0.8287134567566455,0.7969174989350857,
 0.04823559938203448,0.08284277814643193,0.4319168263022535,
 0.7853873743547832,0.4862734219884488,0.1634248462167145,0.9768467076328498,
 0.33510799968844107,0.4723808038525793,0.6612068608273396,0.8833534061080364,
 0.1881365653221777,0.23341924939502423,0.33681973017614186,
 0.36397191677898366,0.7698263284643365,0.6226645723341375,
 0.17595959882278078,0.9103055659031452,0.2511176560181102,
 0.22921150387888534,0.6389407852997646,0.8482397104119321,0.6800854406629884,
 0.3986325482515074,0.00450489227776063,0.8468122624208407,
 0.17609552496457004,0.8106454080777837,0.11419787707803963,
 0.2655741079610281,0.155012307664389,0.6418593554554191,0.6434892813271123,
 0.019709797420252806,0.37064041063430464,0.10976152136284334,
 0.4128289663592224,0.5733683874187884,0.7107777819267881,0.12062573223313233,
 0.5904547321350501,0.6169974829111508,0.971689906447572,0.9792843307202719,
 0.23712040341683926,0.4994319220914567,0.5693618487486403,
 0.48053732465442023,0.77997766574968,0.6396666206989502,0.6627623549532924,
 0.6029308015281379,0.8414330832566179,0.15571048149999422,0.2667053445234939,
 0.14062665900534677,0.5182392957854166,0.17612332936729536,
 0.8203726819449201,0.8197742684075303,0.09168743837646687,
 0.18205691846701122,0.28642870309744584,0.43948466620188886,
 0.8178996340357668,0.3105632945238791,0.03171300518436426,
 0.26896037053371225,0.9466131568718813,0.7745625550790789,
 0.11501516764531228,0.408586133237265,0.5443850611852762,0.2629987873651588,
 0.372655884526387,0.7233293269135328,0.04885818318258861,0.0457766281803621,
 0.6935542097111644,0.3318379784445227,0.9938511737206374,0.6841433214643187,
 0.5922536550913167,0.3732801084261379,0.8937842416468926,0.6911550232448904,
 0.7168026231300655,0.32422182212905615,0.4151267143186005,
 0.36298816031640413,2.975812691565771e-4,0.7411038298777175,
 0.15316884909208273,0.7657385909362249,0.5938277001762383,0.5923620397374189,
 0.15526365664351638,0.3201307805462007,0.3078400230294316,0.6633346100120033,
 0.6809975731297303,0.010415727192260249,0.8663980634118812,
 0.43853959130204145,0.20383937105280225,0.023452716614572933,
 0.5480309600154465,0.45578239316813096,0.5487448552129798,0.4667401819832294,
 0.2965721820366747,0.28045582463150276,0.043758297438977745,
 0.5951089933373405,0.9719139752737613,0.11742411723087254,0.1026471286164542,
 0.934311686356171,0.9294880145038894,0.9473137196260609,0.3121153318004166,
 0.620822636774366,0.34543485850943956,0.6657286079535534,0.288042499042251,
 0.4067232564282336,0.28274982544623306,0.5880240463139624,
 0.04612493314116917,0.8032401484238069,0.008608683967461728,
 0.35942113488000027,0.9312940140026896,0.20171089841758105,
 0.9415010471072298,0.3198328987365009,0.7046254749759089,0.6659971623633304,
 0.07073868450166809,0.10071519029723308,0.2631822228914422,
 0.6816229391191847,0.7448926278291411,0.40765182353546314,0.7853142098747151,
 0.21587385133772752,0.024634772545909622,0.6720219985698528,
 0.747744777144426,0.65419873410649,0.43497504078770277,0.40121645359690594,
 0.762576226765039,0.1380885826124818,0.29157295220858315,0.8288328734669962,
 0.9149449879662298,0.31945544131559533,0.22688783584969574,
 0.9440174977099313,0.20754160351325213,0.0360696498460914,0.1709082590828387,
 0.4924462047171324,0.7698750826797567,0.7971444412373915,0.1315825347432168,
 0.6438158393552308,0.03085965943894131,0.6403319978519355,0.9095567959638112,
 0.685532235452575,0.5645688077875428,0.5306491882732344,0.37267662023012005,
 0.663607011253316,0.41529914898701015,0.879248915679196,0.9614954515016657,
 0.16124250692836428,0.23758039412436183,0.9165359490439863,
 0.08986377657504285,0.6617730218000825,0.6838136267276606,0.9449217462550134,
 0.009118777383115728,0.11512529152140061,0.2975009947180618,
 0.3722982509665238,0.9120660455326716,0.04833535090809937,0.6757232319433273,
 0.29499238152193474,0.3967116334736974,0.356938535982104,0.9611133433332029,
 0.4164661708752839,0.3278104934038395,0.16299632998735136,0.4980619345951677,
 0.8654644444430802,0.33031979538401557,0.8543052727389713,
 0.34475907196038835,0.9016289778854676,0.3873789812894899,0.9199801646800781,
 0.23694862345323386,0.2662840908350995,0.009909068803571008,
 0.37873540917057746,0.3489904716384965,0.9481470431878143,0.6455351894846596,
 0.7307726984698505,0.10477535607907829,0.15137614414391787,
 0.6733261186841945,0.9266930945651863,0.37577351209291476,0.8150277285906836,
 0.3551660779327167,0.3764275951374205,0.4966953823756145,0.6148580142216716,
 0.9801066583375129,0.8243564678508866,0.7489099897328922,0.7819343576641754,
 0.9998695458874945,0.5391839643651808,0.9280434013294111,0.8085452911607873,
 0.5086921138655717,0.6296357505191824,0.5400436089113032,0.6415746117537325,
 0.5730992166157373,0.7303813203445909,0.47654812916559464,
 0.13958811919941216,0.02470542732591896,0.5644250745916753,
 0.6750365423448128,0.27967295913719925,0.24191913935781373,
 0.13899268902277706,0.39434039146072875,0.15968951441563495,
 0.3341169256213542,0.13868021320166923,0.5978972083125796,
 0.13359906464048832,0.019324056390718547,0.3739301054134244,
 0.23373358032743252,0.20098772770597462,0.5426342790676929,
 0.48195686058694986,0.2930915038843156,0.7878933817732966,0.8057574927838069,
 0.28857679239580536,0.8739207933741967,0.8930712444561477,
 0.001956562753332114,0.6843021951932469,0.8006646699399633,0.799324389884392,
 0.8514601512164586,0.6183538943857646,0.7446474107105069,0.40293665129172607,
 0.8445013371618277,0.10321030569660217,0.1768091675525727,0.3899231445205167,
 0.4984355822546539,0.048443449041439934,0.9724124222848988,
 0.2356795669209406,0.015040410808297766,0.34080019289338626,
 0.3099953520033839,0.5600827054910233,0.42098493558524297,0.9193207694382646,
 0.790989780269747,0.9572247936709967,0.7659707967260411,0.6227405373907223,
 0.5141698524938829,0.25067800151457686,0.9436968570422161,0.7641471176570672,
 0.7655822237945692,0.49059860988635695,0.09615135660892427,
 0.05835405589606257,0.1940558484746746,0.614601446625661,0.6334668697000839,
 0.2266570778575696,0.6373752552539406,0.8882954748694714,
 0.033527125076657005,0.9052001874307728,0.5762083451607727,
 0.5381153405994491,0.28385435221717736,0.1942851189042858,0.4718246941415909,
 0.7807121261765784,0.5625771313962264,0.7283883825791967,0.22412545657579752,
 0.025264909467276442,0.9212934008500502,0.28613107756580847,
 0.07301100950704364,0.11954976372439519,0.170049506494098,
 0.04851465614612627,0.12385921964691482,0.25161822543560963,
 0.8273640005751641,0.20079395392530763,0.6808546669471982,0.9989153190954999,
 0.7416919096460113,0.8800435754632262,0.9634472834058093,0.5651616106878388,
 0.9729534591195393,0.27717861350752027,0.6660275138950446,0.8356130655499439,
 0.14554452716504473,0.881532670653922,0.5022796316607284,0.7256576001570907,
 0.3719234830197561,0.8337467633140775,0.27336575338289026,
 0.11567028787624933,0.5539837797117735,0.45822862280525933,
 0.7724302476425453,0.42742307057323237,0.6349561936205221,
 0.36336378217330634,0.5616505614636855,0.43919604910673604,
 0.5003678564234355,0.27943600627357157,0.17684897999726112,
 0.22011446582072813,0.276856578914477,0.11196490668664649,0.6033011243152532,
 0.0936273000940091,0.4852189744134944,0.24769281830978862,0.8875885899681514,
 0.8668928052720769,0.20557014877706614,0.5579603475354961,0.5978610728986339,
 0.1383916987424073,0.3493356642041996,0.3965084358649733,0.1326082023983406,
 0.08117003259069777,0.2837724028814319,0.9741268426042412,0.4652319431198625,
 0.8442135686493962,0.01866572748774664,0.10429961166061674,
 0.8137364071808305,0.09688225230258174,0.2160551657324773,0.5750967888060373,
 0.8643679499708888,0.1622038258120323,0.6584553286122599,0.8674425261956801,
 0.6745574175049256,0.736027308422638,0.34712744522432293,0.3545352575659326,
 0.7464606295089748,0.6114023690641399,0.984485254161381,0.9711842510334646,
 0.1488353270432391,0.8568574373984478,0.7935472094121594,0.3919332800434767,
 0.44389063801926854,0.9700809961064192,0.9645272438696831,
 0.20333507286637442,0.9299226860074689,0.23537237135408506,
 0.2696019052028986,0.6555205130620053,0.012168571756052793,
 0.12886750675979797,0.5699173396784109,0.5178307634055402,
 0.26712114120048336,0.6513577551779757,0.769340714580374,0.14521967610100495,
 0.6314773223334426,0.019219592160277688,0.19486122289747987,
 0.5615423837834898,0.40801254064319925,0.8796990478674186,0.7777503751375123,
 0.3010404805569553,0.4787645072879937,0.29353753725192966,0.7336766921421544,
 0.7345335408713448,0.3907323368190687,0.20514929150783445,0.5231577029587196,
 0.41397554389764357,0.3212237430220859,0.34627722000028105,
 0.9977605169883379,0.40973813281872473,0.006648993722290664,
 0.765671925450559,0.3068916908607492,0.1820094919195867,0.8779556594749438,
 0.11731588032530438,0.4874948032827954,0.9938941134545285,0.6096089652613663,
 0.7648139264246034,0.3394129046596439,0.42339820497935365,
 0.34123560630817207,0.2749749646948082,0.2512704709420138,
 0.02166632733796181,0.5687997869185004,0.8440418626591186,0.5246571024439576,
 0.23620090072454258,0.23592150253519906,0.2971814531688862,
 0.3614837254611756,0.6641423659858475,0.8784838781462266,0.6586688113936083,
 0.4705055734319008,0.8270001709968186,0.5638164054850133,0.4557876419221619,
 0.158977183479729,0.5833385975969624,0.5404998770071318,0.8172204097701969,
 0.6376183553996385,0.31340733001458787,0.26707627090093633,
 0.04585026876538334,0.6456629995898836,0.4209282250397972,0.4314791515482719,
 0.39811371912466276,0.299689268286509,0.3848192221486396,0.08973302026829366,
 0.6019560548597881,0.9671190281073666,0.3191373272322504,0.7418753874315303,
 0.11842880048442317,0.8339802483658686,0.40014496094733165,
 0.34920531124669196,0.6360354742465666,0.7594826307870433,0.8628598002337138,
 0.7790297631010098,0.34105453258541696,0.6139297239910364,0.3472615723562198,
 0.6083069142410719,0.15179168341519056,0.25039094050718735,
 0.2468978593727491,0.5312712991957191,0.31379495446900973,0.707939878609448,
 0.6101782907271889,0.0036842275450617024,0.8391836166641948,
 0.29533054700699013,0.5992461836118077,0.1515125933489767,
 0.31788935138676333,0.07440702907258767,0.15531980106205467,
 0.7379773492802855,0.4081352830551964,0.9455370410379227,0.1928815897711602,
 0.4156243092209716,0.5989220616169135,0.6505936512549053,0.1608997585287999,
 0.9268128955890708,0.4730449995559688,0.22525030102557708,0.6882027224680867,
 0.6516854509884573,0.6882066734211213,0.14833793096128645,0.8368819182731321,
 0.9760845636531753,0.8742763190452779,0.8600526653426958,0.654359004815502,
 0.885148671598178,0.7422389165253314,0.7673838007558788,0.3743032042389731,
 0.471123361654593,0.4521457216832723,0.2847157308698566,0.6757350079169017,
 0.9828604364044428,0.16874226973456685,0.8261997248384052,0.9934126407731693,
 0.03300868253962197,0.8729572785657105,0.4243101723747075,0.6442662758671773,
 0.415240489735046,0.5086673504052005,0.6800418393467089,0.682135534764476,
 0.8212565418931422,0.5804970850575462,0.17610276826102678,0.7677792506442863,
 0.30945688392091864,0.03068067502013605,0.93447047522212,
 7.020533433520493e-4,0.35710245519561334,0.5879233518434412,
 0.36406923815012204,0.4389772836378132,0.33077137532977174,
 0.1432862329502519,0.17770604885368046,0.07258210779985641,0.757130107904459,
 0.8109448888569759,0.6787998217434043,0.6005664227058344,0.1114296862209343,
 0.2084626242865143,0.8013778309291488,0.24936687728456686,0.8922217243876202,
 0.8248913511248749,0.04138065463553586,0.34131790094173675,
 0.3172200943429102,0.5137661129055032,0.4979258819031088,0.8093039721121728,
 0.8629143059566544,0.40716997032496405,0.21365530166588376,
 0.8505734086790776,0.14409923809431258,0.8875364567992897,0.5530187813388845,
 0.7531179435272649,0.9818777701857279,0.8724856586089298,0.6288475545846088,
 0.28349189276196207,0.438995735099578,0.6777865707088934,0.6163477787557763,
 0.649980469199589,0.6726430330316361,0.543029271126473,0.8395730688133537,
 0.47374663397250494,0.9126207689731248,0.632587474918424,0.7373120421116074,
 0.9092822844733655,0.8305552575343877,0.374006516453659,0.45132730594678583,
 0.5433599742214483,0.3210768522322689,0.447736966987088,0.5016597895629343,
 0.17765715299655116,0.31057122578426954,0.431102541614362,0.7339476323750198,
 0.5330511299413876,0.2555507575129534,0.6754369443160156,0.05743458727289841,
 0.37959972533191577,0.3596577707826796,0.8268238615837331,0.5485655383758191,
 0.9451788649031857,0.18371643255321501,0.44456683567892186,0.726122986756243,
 0.26775311843185,0.07431417349419178,0.7644608547420289,0.5509270379860034,
 0.34360090228801043,0.3131074552891254,0.3467007382178191,
 0.23492058028860907,0.9852575383314252,0.24156277346572474,
 0.18688000289819928,0.3721647114618083,0.7913061770449324,0.8775550542473189,
 0.8115855192304248,0.3382215969110738,0.06445260438474831,0.1898221659072391,
 0.4935118902976522,0.49861676828772716,0.042607879563229645,
 0.25756380433858617,0.017580983465350752,0.37029279663275405,
 0.4359600090018294,0.2127547790397577,0.0946276247265515,0.2667252064129998,
 0.3115730683130614,0.24406315194259975,0.04012841295338598,
 0.8265214225690868,0.5996373954650751,0.5717254172140592,0.2104098619555338,
 0.16652866171625658,0.7953104611171411,0.20410614238703118,
 0.5856346019909123,0.3000710643317025,0.4882934685463074,0.6878536954160498,
 0.15617400191070274,0.023424224649392755,0.7366904785028882,
 0.4325724971862829,0.8568058388450717,0.38901381752253017,0.7437932210890252,
 0.4506683493213324,0.8553236828046632,0.8748137408513823,0.5994140538909007,
 0.21029085698575267,0.11890598363738614,0.10664904888383431,
 0.11527867174883755,0.4900737166536785,0.9814022892006281,0.709666796179692,
 0.4083349755736633,0.11243428583037163,0.01710698881265882,
 0.5515995459962159,0.01648664762306873,0.5574483956914272,
 0.45467130471183026,0.1200999223131497,0.7178153323128252,
 0.04110199304307416,0.7081147196132391,0.9306228182903135,0.082454461806164,
 0.4754516093390867,0.725823171862102,0.636403419891596,0.9071838177723031,
 0.21376248925374897,0.21539565096100632,0.9380837681785494,0.647104320569536,
 0.588017302424907,0.5160499600308462,0.8436517425219845,0.1298442006591075,
 0.07811095140025304,0.649418622421852,0.033784695284731825,
 0.8565768287322624,0.5178684874849009,0.5556975402854243,0.3090719501089281,
 0.586327198199585,0.38624027585223675,0.9931456821452243,0.651796991938825,
 0.8417818484618955,0.2361122728204128,0.15715918724544398,0.7108748777126036,
 0.04762505881089618,0.05365136296180162,0.8512866134513944,
 0.22361537014902422,0.6391862898104772,0.7140719447064438,0.458025141198225,
 0.46253945719098233,0.0015532657926833338,0.541757070541897,
 0.45096164552930584,0.11509574299602843,0.7351343722240026,0.335885975555818,
 0.7819055482040232,0.5300552300886716,0.24286426415901596,0.123366532942635,
 0.7492382877705822,0.840344593803897,0.0334667079651777,0.526709189965571,
 0.25401851432955735,0.009814494029266485,0.218953477645772,
 0.2660778044231895,0.18355602566989415,0.6650334262555949,0.7604858674574461,
 0.45273629132686855,0.2272068319789643,0.5577501038807031,
 0.12144466837885559,0.7935238439274472,0.6414360036410147,0.7167587624756299,
 0.789237252202258,0.7863934525247768,0.4967957589596699,0.11985981609213536,
 0.7008969464968842,0.05290240909109811,0.45243383162220085,
 0.7728867509517762,0.2015998336842767,0.6708157482885807,
 0.010539659215347497,0.24477177231352232,0.7120128204729416,
 0.8844081751292248,0.08412626692283165,0.3988138753027126,
 0.21968324070385825,0.5959515700836022,0.3892681965851652,
 0.028923396597124396,0.37133027010670405,0.20650825692518116,
 0.7467709634503438,0.5174960156661671,0.9624590225409841,0.5311669453638326,
 0.3258452972131467,0.5703736281259127,0.4889429393847178,0.6092276006941237,
 0.014121606344136839,0.8629032005743251,0.35341236557246347,
 0.297274800990998,0.5442059643103576,0.7799672522338255,0.6763283204423788,
 0.5572954940239669,0.6639348802929624,0.5317127206166485,0.9973098482183922,
 0.028307664211738248,0.9708627620620885,0.7878069244079673,
 0.9774864092751889,0.333696208429507,0.6783408204409839,0.34799754600305655,
 0.8644315972657817,0.8909483224183032,0.09929897788936715,0.6751798479533293,
 0.3282627681313576,0.031221848578587896,0.33670590831676006,
 0.6851022758482771,0.7766770762567513,0.8860901783214806,0.7528922974463772,
 0.20568051056347203,0.6923694499257469,0.5522878584325681,0.9392084146473118,
 0.9677062175614878,0.3293833927406792,0.32876428827121085,0.3384601415153272,
 0.04230376838443195,0.6787343444486204,0.8906265670838935,0.5156034024108862,
 0.8577352460695956,0.28893151035399633,0.2699303500955679,0.8105803155247241,
 0.3878738562306496,0.7140089201477688,0.6603466250189804,0.5676307971814936,
 0.4352713474554223,0.5445413138360911,0.1379748905227267,0.33598208754469383,
 0.30023475083279816,0.8029374815222067,0.8902725236306595,
 0.05283909199987047,0.7234883755844451,0.07079400765625221,
 0.45443725704757854,0.6925504769529087,0.600255667550583,0.20221978941087793,
 0.2693406943961434,0.4362847054726041,0.05066216463438966,0.7522537563811333,
 0.13215671811543805,0.7009844768018633,0.747138330301342,0.3606505349372018,
 0.6435213620990617,0.34405799606966536,0.2943486964395551,0.3535132975876707,
 0.4984124257077702,0.45195252994434465,0.24636724635571938,
 0.17487758925688968,0.033989323213090894,0.43437943856072725,
 0.06388269740208308,0.6806719049344189,0.9253514726399634,0.9883555482265661,
 0.23102551039226238,0.22961370891686683,0.26441397268689226,
 0.2063301954595561,0.030436568608352643,0.323705257632391,0.7134217659164331,
 0.4993609061331541,0.8516594024408298,0.27243738346486723,0.5380715440002968,
 0.48616776262659167,0.04987314450725122,0.5361104425473296,
 0.8236811422989778,0.07843802488929608,0.5122609636874664,
 0.24243716957845696,0.9471763175187424,0.16845226994989904,
 0.014134036691768048,0.8519018955964756,0.5260494546094571,
 0.04913390656482086,0.9912185413101551,0.9093817181077587,0.7675023028679292,
 0.2979401025754187,0.5024451766861997,0.3443037715109496,0.6375609175909469,
 0.9325921196473215,0.6243179552651744,0.38393068185115553,0.1838702016475383,
 0.009982311689844203,0.33132715269011326,0.4809609960369565,
 0.16423279472899188,0.2439169088451949,0.3595754960352936,
 0.21318418953111695,0.83899320028973,0.7452423932983355,0.6499194255176222,
 0.6466062726482918,0.4784501058435737,0.5000098971000732,0.6210247640817427,
 0.2831376819092033,0.13193348874372135,0.3674830694846498,0.8317054385693903,
 0.061885269267524357,0.0041800498604246394,0.1832917234321192,
 0.40617840391018056,0.1805807393029243,0.8667752081305554,0.5910279632679075,
 0.9890377036002089,0.6502375256226218,0.017135037527481023,
 0.01030935515540432,0.2002843204109993,0.2057484198171613,0.6043061956542725,
 0.1277151155139321,0.5797973032957416,0.6841444690158727,0.5197349776765248,
 0.00853851434342856,0.9753671276656792,0.33327076424927726,0.77666051572355,
 0.9117613631113287,0.17434428886132147,0.39443473431057563,
 0.5013191656941366,0.6573982369447582,0.27577784149307005,
 0.21853947996547418,0.37448993455104884,0.11738542276873842,
 0.5856813396238401,0.9214479548237614,0.7345676009503738,0.6196294388525352,
 0.7502101638053912,0.730535764921567,0.15199928801899476,0.7600469551571662,
 0.6561434001088682,0.20924628419392644,0.20735093402136318,0.319444034550994,
 0.03374462656624799,0.6344431276307342,0.9905672139220074,
 0.22064234081228862,0.4395166432343647,0.5273586714725954,
 0.07769121574954363,0.9596225718881157,0.4299711420194895,
 0.32024771420682596,0.10018912858279205,0.5661506242907243,
 0.09872107501881233,0.9410720721094069,0.8250680788700415,0.9340659467987393,
 0.13868892997082372,0.13344214846016478,0.7015017595023797,
 0.9430228764620068,0.5714166364689254,0.19739978007706216,
 0.012569034289539815,0.6689184631809808,0.3436895449654317,
 0.03707938895768326,0.7320849767685079,0.2989529159280101,
 0.46841304436488174,0.23423392705698753,0.5762321406201573,
 0.5344743690842151,0.9547769601406809,0.253002926273143,0.9709710652812122,
 0.5781310212864158,0.4381124775088022,0.651817987468351,0.21262295828320388,
 0.9963691767070859,0.39258517934117754,0.13856610805796765,
 0.13074707357388815,0.2554543645607539,0.40822961252859713,
 0.5901734939324857,0.33057009292847683,0.8537388480881248,0.2631832995126697,
 0.37415097231837646,0.21987996674784616,0.2054527422251602,
 0.9108916517939711,0.07244936693049275,0.967258249897775,0.7028640002690538,
 0.32853000970107893,0.3552872426008935,0.926177037210911,0.6262687612988678,
 0.38368798641708624,0.31729899774398507,0.9960609424195873,
 0.6001874095653783,0.48570689734431427,0.6470097117393276,
 0.044909298125500374,0.3583549746115958,0.5344856550571997,0.773848228538773,
 0.07905991136021195,0.11966163483653602,0.5015305660288778,
 0.21933633979908373,0.014024958367093276,0.4714186838152332,
 0.8246383110839767,0.9614162484472271,0.972849293135232,0.4262894191988885,
 0.47515498229090747,0.13109810140660383,0.7734579271441095,
 0.6010667394878673,0.23477425822234743,0.03215836813219275,
 0.8095094130678906,0.7984330137148423,0.5302516652519927,0.6980000977938845,
 0.15697040778246085,0.7915938249662224,0.7609095303168933,0.5353776037031767,
 0.5155888734841948,0.6359648048427277,0.8197405675235798,0.4314159844361185,
 0.42483833086788636,0.9583861265887079,0.6367818564533269,
 0.42872636807533304,0.3508244118573868,0.6282058411883383,
 0.10716524023186802,0.3036508127172969,0.2930408405239845,0.6237305595638954,
 0.9683730644038029,0.06576057853810602,0.5782110505572143,
 0.030095156065292006,0.5626960147945462,0.9478636670331212,
 0.9635413802237055,0.0881067423817955,0.4051836905196802,0.10139299756519538,
 0.46192368959952,0.025517916117497874,0.3382043097721057,0.3155257378978621,
 0.9233856373973433,0.11712498275736016,0.921235393237391,0.08106758623594701,
 0.006421577831361991,0.9055389755465459,0.8836381080669821,
 0.30244710580864687,0.9407754531988655,0.6097153823228023,
 0.23047893824756827,0.06484871024599803,0.14690603222759768,
 0.4499225339360603,0.24073682518144712,0.41285115789775373,
 0.4745808996163754,0.32449423091458174,0.04975265750605451,
 0.6417908184800035,0.47049859871316424,0.3862031324350248,0.7139098738549956,
 0.8583708751559734,0.6853830617033421,0.3467631014809489,0.2677031960020718,
 0.008127191193286532,0.391361228582931,0.017508947710444867,
 0.45591600389836884,0.40195540254680584,0.37125624963871784,
 0.008447534373922316,0.01724446315354511,0.356528536900957,
 0.10803225146916573,0.7001084969734839,0.5945660223640701,
 0.17869373783272402,0.5295931382810936,0.8970726431961619,0.1641990011472949,
 0.6978992805834675,0.7010968313868708,0.760878101648864,0.21348881976624767,
 0.07306892330640746,0.3177340869850661,0.27552879001722697,
 0.09031560645678294,0.17522871718481303,0.9337037563555025,
 0.9465643159656335,0.8868746423517103,0.04144539650447243,
 0.18279399726966106,0.6879808208015578,0.5818051435290149,0.5752052001488251,
 0.6941507615907452,0.5085036420294108,0.7374835931255529,0.19458565304183018,
 0.7659178650276219,0.056167641252231726,0.3083445912669397,
 0.30310926348294087,0.26391052336869014,0.17841326412066083,
 0.31641026271673256,0.2527859040817101,0.8852711572409991,
 0.36592588084961797,0.7002201073953865,0.4198966311935778,0.7808944080846367,
 0.7834091581465307,0.7308188361998357,0.8831112408055046,0.8322110144069662,
 0.41855071153963475,0.682800285956067,0.00789528585025634,0.6687479468275013,
 0.5665232460338677,0.03944310620972813,0.163996085544018,0.6889009311373352,
 0.35053989834282184,0.6531226321135492,0.40483587486582295,
 0.7536709822055301,0.4577349647591954,0.01507722335297379,0.7847433189611115,
 0.19152296223500698,0.3671149794088182,0.3280610332299069,0.8237726195888075,
 0.33377006125481357,0.6356680564526355,0.3368547215698948,0.3870921313008666,
 0.703871126915711,0.9348599936487341,0.3406498696307112,0.7083973816740208,
 0.05113050590265733,0.8266557433917442,0.009335274530684168,
 0.6707102733223014,0.40659240015410636,0.4814941388280247,0.4113072076130665,
 0.701859992103974,0.8637905034460591,0.3376185235589757,0.48413230658660855,
 0.4811688707021662,6.536541001466167e-4,0.009872640447160963,
 0.6675155119869968,0.6700866032173043,0.8518515256985644,0.8715305706602023,
 0.9907989003187283,0.2213619479508332,0.6467952325357819,
 0.022187207151955324,0.11804370782445239,0.5223580745006969,
 0.8574425726199628,0.6725667196857399,0.45034277346861107,0.6622166453776269,
 0.5239531228095854,0.8510870955582878,0.09125753408363824,0.6343428928807815,
 0.708344231755251,0.024323737003007384,0.5933684050943135,
 0.36959882744651595,0.18932006122241918,0.9265442871462898,
 0.22386417204499443,0.8422316933237589,0.6266992758743203,0.9839375514164135,
 0.6391485919759722,0.19164311507110332,0.19995214718929244,
 0.016285242444369352,0.14142300611653658,0.20150432320502165,
 0.27547707583671543,0.6982522243858972,0.47098087500155916,
 0.022756592524765806,0.36497220136480013,0.9348960158242687,
 0.798241713646524,0.9888793892440518,0.6366312505471321,0.9540054114198361,
 0.8759284350132227,0.8668276255268567,0.43342221265950986,0.1828749595879542,
 0.035678635742653286,0.5361861193963244,0.45804813763418883,
 0.7828015747722886,0.2068394753629217,0.31467709600587856,0.6936265249299385,
 0.45298753044463613,0.07501724803592824,0.15377156267974168,
 0.4496072766718573,0.19262774458304732,0.48926110255122846,0.829375381939886,
 0.3071543786176284,0.044153486852173685,0.18727534233442222,
 0.7603079558266974,0.8569309014489969,0.23229755038323485,0.5578215499897895,
 0.8012009834829467,0.7248845485659672,0.7143359129492219,0.2329464070365337,
 0.7311076827441072,0.8202458572976266,0.6887837398151355,0.9849388433076827,
 0.11919176406725451,0.17087113941669185,0.4481893133238164,
 0.7429773542364208,0.9750757783742929,0.708537817262364,0.08702026086865666,
 0.9766176962420619,0.7928360561829207,0.6047222821360944,0.920182783379099,
 0.09574772657921748,0.7697736913515864,0.9832858151960604,0.5447523573505295,
 0.0832301009379588,0.42070347730816293,0.6811827576650973,0.95716944534505,
 0.2909336427609962,0.6962275488044743,0.5976404057504647,0.3014673468417943,
 0.03877885147127602,0.46206130026839864,0.9399922033125196,
 0.23015095610499092,0.4104020995134492,0.2693704701348617,0.0329353145406589,
 0.20731837908210626,0.8958961965327655,0.08634281229538365,
 0.8225759799378896,0.40270203425152173,0.9625201107002508,0.8621710865015499,
 0.18825971556844012,0.1992427802844463,0.05775850289142137,
 0.19582066961851996,0.30183145059137284,0.05593538964754852,
 0.3719132153898136,0.3288190960408237,0.7194982491805685,0.06145958006824381,
 0.18121252773029684,0.6559258695700337,0.21429973505330802,
 0.30189437610859216,0.24630249350284428,0.05671793450152973,
 0.5269705061297729,0.9570250826284084,0.33598617284205223,0.564527021463114,
 0.6599444920625959,0.5860170278557976,0.6224276655760385,0.4371006046374454,
 0.11591995546947775,0.24858248085091705,0.23234753131281283,
 0.571007938073234,0.035734617986585726,0.6530166817951193,0.8191038771330832,
 0.5208299011937016,0.9666668020728294,0.3690695795013612,0.47590725767989606,
 0.8897472894734048,0.10632292804420207,0.49777054313430635,
 0.6309729076888594,0.998886398014494,0.8973109830700214,0.02192969761391339,
 0.7989149085695617,0.34184805395846507,0.7589875847947364,0.4217041502804475,
 0.5204469789190291,0.976029806661395,0.47249274973291855,0.7539139435369592,
 0.888041008192354,0.3600944381178143,0.27224149769669526,0.3893933166930851,
 0.5081374171424718,0.5065675640169595,0.2169095891593611,0.9208562248978477,
 0.5214649302080536,0.03712192927446534,0.6008659717626562,
 0.08701540876402192,0.9428668917386274,0.9769625537667751,0.8229810144668572,
 0.4864689754562399,0.44888532863067876,0.9734904001744905,
 0.26916667441269215,0.7725489538974221,0.871672483255128,0.24110632493863082,
 0.48174188040448096,0.7444140913035091,0.2643233291191438,
 0.14445715518069246,0.05324453682763064,0.04726966828132739,
 0.40066059216248007,0.10177300349452334,0.4422248332611607,
 0.7386879164160992,0.5868907071295326,0.3960877349368269,0.4453119882401334,
 0.41250014843841176,0.19671901465289565,0.3779223861116392,
 0.46992710141786076,0.5931477142253876,0.879111892347324,0.03831113907355832,
 0.5415205546288788,0.2340716632197637,0.5127277510185191,0.263450011690048,
 0.3822608284973257,0.07947454586600522,0.018651278165314977,
 0.16509021189757866,0.2773634299844163,0.46900266523325074,
 0.11002223701358416,0.07427627735099906,0.7312779156464553,0.949854448034118,
 0.20238544543804993,0.9786279296250454,0.4575164360472527,0.7184196006942191,
 0.2630469469328782,0.12283749004348021,0.9569127774351619,0.6766197934040032,
 0.34876816548982326,0.39262666403813995,0.03182948872330016,
 0.09140107458990965,0.937952643743223,0.9154764316716424,0.7164156139688699,
 0.7651920692044478,0.3263204492144075,0.7403964535437192,0.35410318009416375,
 0.617765594253276,0.9160299692750076,0.3567564791460549,0.19689305635368826,
 0.48910888146480214,0.024288131907041377,0.7075247441740902,
 0.571365153738273,0.20871672314946166,0.8974558489212869,
 0.015113450643078341,0.35208392388182896,0.12487645126201008,
 0.5054705676384429,0.9027391526924761,0.7602812336271952,0.8408365096242101,
 0.6206820637692257,0.48984518773901164,0.47156918165073836,
 0.8060835679978386,0.8058924210274879,0.3455293591776277,0.7510148708771843,
 0.12365146083699119,0.9876875101299658,0.4591622234251306,
 0.06626096145059834,0.09322236621611935,0.034636754777526035,
 0.72443955253699,0.29271568451997554,0.30870714551497014,0.8585843919536212,
 0.007401352970849984,0.9426683714198854,0.6996990759255504,
 0.2882210093349189,0.41015287730654926,0.5922131007088093,0.1483940866013541,
 0.5281942852583565,0.4915083040439646,0.32355735450115763,0.500066301000406,
 0.20563149910504697,0.20857463807462662,0.20401170320268647,
 0.9968469783828733,0.8027241975145875,0.035643076961782594,0.807455209718598,
 0.6110609391600926,0.6569768980306698,0.7961007842364141,0.9018624343769662,
 0.6551253004083937,0.4486270919852716,0.7184823187525721,0.9961711317362111,
 0.08690832329323472,0.9778134031690238,0.08795215168137516,
 0.14359900567695605,0.7023658032994305,0.832044753957383,0.34825206320293145,
 0.8623985894043641,0.25132225634187033,0.7842009705942737,0.9485851715764972,
 0.05197934770771773,0.5167509126720178,0.4620132846850562,
 0.45718456763441595,0.5709748919934954,0.8558548162932434,0.5358034430275171,
 0.9280808455438462,0.9693692671026966,0.9781581738550037,0.9145060863681642,
 0.5700870647644625,0.7147037196712139,0.07622761211224005,0.5450459664147352,
 0.8859853951598133,0.5543495421184518,0.14837175051720486,0.6986691078553026,
 0.8739904480408367,0.020050650251837387,0.925806507548835,0.5908074931419003,
 0.3055160322832109,0.01187359707116642,0.14193418220180132,
 0.9311208787641099,0.9102384374500525,0.1728993587237826,0.3882308865647254,
 0.6771502853270431,0.910822811645296,0.3385930858939147,0.5560607930779441,
 0.00899374172067402,0.1448717013404306,0.28434106266433057,
 0.03355781918027345,0.08742776742139902,0.9744631327471972,
 0.33302986654718236,0.8195997071141436,0.8645274662911159,
 0.22197650248336565,0.17260094956532002,0.8570213634656488,
 0.20167904477719156,0.41829239960309406,0.5226145788288798,
 0.09896647968471661,0.7088576998413562,0.7024447723620496,0.6428955356380996,
 0.781426211302592,0.2044898186793861,0.5667039473113941,0.19504078055551255,
 0.03811128731100244,0.33721470674323795,0.706500020200676,0.9209665983038839,
 0.6078357861855955,0.20660573188110787,0.014319862990466725,
 0.541332493567445,0.9225813903163071,0.728679687966628,0.9621473427208989,
 0.2216073567062009,0.3575702985731336,0.2543899214436268,0.19743424403662935,
 0.07672824802446698,0.766511393460606,0.7006702105630728,0.2352398959598011,
 0.017142162106779768,0.10213398051750211,0.9258742855737677,
 0.4430295777345352,0.9894327948305728,0.040400789897332534,
 0.28985792856443127,0.6025947920822463,0.23236004782072717,
 0.2231333967986231,0.6483776146431892,0.8503711882763448,0.753063345354147,
 0.3065293847498841,0.6514448710245082,0.9676919556344875,0.822421956435508,
 0.8246137215431815,0.8342259048282349,0.7218077244194183,0.8919429082846515,
 0.27120623704446134,0.148203390776944,0.2474224183604986,0.7695848518989478,
 0.19459582186654,0.7161328077282039,0.7303506171350035,0.7483269371380852,
 0.824295864643988,0.011406110742875453,0.722705376593827,0.45181708144782695,
 0.9552979305717478,0.5284847177129307,0.9621023951720071,0.5031673868914753,
 0.5373172929090999,0.9742536185507188,0.4028020022738823,0.12015469912966481,
 0.23691940695806857,0.09617818105453912,0.1125831481140106,
 0.7463167414395842,0.26572433071782353,0.6050729530954303,0.6037175722083122,
 0.9282178660954717,0.4754925808543595,0.6912784960209545,0.17706768812373574,
 0.5418738084258822,0.33942212737101407,0.653979367179407,0.6250224154016726,
 0.536183827254221,0.4150591804451592,0.2771963761988454,0.05116919309557799,
 0.7711312183984603,0.4488986342664727,0.5887625628436148,0.2475287159733207,
 0.278337328558391,0.5554881729905372,0.5076163648419728,0.24009291187032034,
 0.3695296736827218,0.7749608201897317,0.2772324839142384,0.2339462699731607,
 0.3912597265221316,0.9071658508535397,0.10366866401072128,0.6301852036744814,
 0.49789657219886796,0.21015778939647944,0.34355527056240187,
 0.3759184976553197,0.04713399689483044,0.9563570542191251,
 0.04863606041727386,0.8387118092369721,0.8008764857716866,0.3111249792345986,
 0.9528685057861312,0.45616990774978494,0.8891763747892512,0.817866242573622,
 0.30605416031340804,0.365309086835349,0.29641416487398775,
 0.19311320351387862,0.04862700902125994,0.7193096642255203,
 0.4714959788565961,0.7457521296582332,0.024800872394966422,
 0.9400486783815485,0.8439498097381872,0.2550986603349398,0.2833222283096628,
 0.05101725889602926,0.3684601799934559,0.4251233770767462,0.6390141047882191,
 0.9765698933038557,0.5991169714942353,0.956732079643527,0.7708603611577587,
 0.9655398006103769,0.5999069748938541,0.12761133947393133,
 0.10843082193120868,0.4476081960855427,0.6025535938505568,0.953921362536733,
 0.8878693202830071,0.22524170233682705,0.5792767861292192,0.2921014774915539,
 0.9619324168321913,0.753497283273859,0.9860797129176675,0.6818834084801566,
 0.6150018246071454,0.4501628620246966,0.8605305875536858,0.3597035488497331,
 0.0635183671884949,0.5557517603177689,0.2225436355316469,0.01576358578668957,
 0.9844907373717356,0.44546010427597316,0.1909287683562918,0.1604988343927748,
 0.3249470615176778,0.7718369291432905,0.4287689213009097,0.6853938232845551,
 0.02042140777254353,0.08192504550325741,0.9937997020393818,
 0.6806617446279858,0.20374382192789198,0.206474230292356,0.1797180522460411,
 0.34353490698800215,0.9699449021598612,0.6481869949141428,0.299734537166336,
 0.3041548618738662,0.732394798495042,0.40722923374589337,
 0.052819231945918976,0.4079981371651511,0.1345975144676963,
 0.20005562517906939,0.12128848714264095,0.7120443832922224,
 0.5433181993741154,0.6842800408060055,0.9233779512089926,0.8661244668218452,
 0.30201965380175133,0.7732745905795197,0.4149271970605335,0.4378925274875305,
 0.2847758863713514,0.15668087645551887,0.5645045046166316,0.8578223643920085,
 0.37685970667937196,0.048163242598082245,0.9567506313941397,
 0.7726167996557631,0.7594377264188404,0.4156796179361326,0.05827111834469345,
 0.3398299742612323,0.6017085403021102,0.8509961428239048,0.02595773471152063,
 0.036409191971630506,0.9461272367925636,0.013239748595667788,
 0.36254024638782056,0.12887276775278333,0.5245882284631878,
 0.7352600013788423,0.17616415513065842,0.9084718631798703,0.7445020285293088,
 0.36565288747538516,0.949165438009716,0.2961127349633962,0.60628221622839,
 0.08555877486315921,0.5380665215908905,0.8179274126744523,0.5667912042560443,
 0.22074155212576785,0.9340797788587778,0.22965324368075457,
 0.19838006801909625,0.40310314319624907,0.9271808285700894,
 0.4381647140359436,0.44810847652289043,0.015375061190997297,
 0.9954408906043597,0.27772039363699486,0.3664884385468108,0.3983385157432425,
 0.16102850310100236,0.6642355364272918,0.36314745938774395,
 0.2057017166474049,0.6405165214318947,0.5280614666709875,0.6126921542022011,
 0.27977255967068637,0.8946042291672339,0.8921514765088681,
 0.17085973207315808,0.8654342050891095,0.162741184849154,0.08653993187149034,
 0.014145849710740621,0.00233897985463849,0.531724693212226,
 0.11059761708303117,0.7628053086921133,0.6159320088396744,
 0.24959376869233119,0.21269994882620957,0.7159223066884444,0.737221741921879,
 0.2407324748689147,0.4865200712280663,0.9493065951018042,0.5384033179003443,
 0.9183149448926572,0.6451823194768112,0.5765040764463047,0.6553647188015311,
 0.1631780186673023,0.8800780663500505,0.749614917550501,0.9953862363724987,
 0.2258664402938625,0.9132605941974128,0.2806256416260209,0.5701657714830941,
 0.4129697747595992,0.15887850176078033,0.2637077972164621,0.9425300524835588,
 0.8555457203892247,0.9252084006514604,0.9983938982826734,0.3207466618481283,
 0.41204210087719784,0.5960592177971116,0.48286068340303123,
 0.20524708880993736,0.03966987483942552,0.5574683611791778,
 0.7884499103704404,0.9417452316243177,0.2848238056202569,0.2087206429129912,
 0.8698699202466391,0.033993211812565316,0.17906345052942974,
 0.21832134983062434,0.35106709964138416,0.8901722269123108,
 0.6102046427684373,0.8237038756958766,0.5179818686922188,0.44452266388631423,
 0.7485268202406258,0.04688465854448809,0.49274684319393935,
 0.17408257806455307,0.8567561062734936,0.22638485116916884,
 0.6114499713855609,0.21528534697410784,0.6507277015194542,
 0.03538897281201425,0.7364457368501913,0.5457445547394296,0.6293907672537673,
 0.5038429723408502,0.6853704386927446,0.07719258386509387,0.3188901698894795,
 0.7498808795588011,0.3919650234137477,0.3529459666341309,0.20465292913129296,
 0.686606718640139,0.6055574379384,0.21386035524637292,0.5155504084711775,
 0.4581342704983318,0.7282723938977499,0.44379483112346263,0.5288557133347084,
 0.3621651894328587,0.5368925043169008,0.949009006902884,0.37779893393712427,
 0.25203409214697514,0.19917145570690797,0.015254377698492916,
 0.7176685177515376,0.5917245751650467,0.2506642906739005,0.9353141647654747,
 0.2734500454122808,0.14732127919092286,0.3167981091011105,0.6914081457399069,
 0.5225878816502139,0.6204748031346895,0.5882172360623162,0.10430492098244859,
 0.43587966683826773,0.20413576057121918,0.9613459495206401,
 0.9813720000016002,0.9023441601804248,0.8146490110811756,0.3802342390125295,
 0.9652859735927111,0.4927463788048274,0.6759826444064222,0.7159778547504335,
 0.01651159184551443,0.6123275610475442,0.7582749193521408,
 0.44352958959807054,0.8443586497787093,0.7976780650991017,0.7785507169437966,
 0.745539551974659,0.8675755791137022,0.5528814789805186,0.5268216643193058,
 0.6771765080213895,0.9897182056431784,0.2811434010300269,0.40101548213054783,
 0.17535952594546633,0.0917029907277559,0.8317024265766159,0.5054808168106415,
 0.06101823624828773,0.41552679476320353,0.11818720782572423,
 0.13864376946447432,0.6412247207622883,0.7151576817828853,0.940712771130273,
 0.9620086128436977,0.7778974060428921,0.6488719857047633,0.37588729895597706,
 0.8724576618690165,0.2658572327525657,0.8491139156845713,0.5292638145242687,
 0.339477832984142,0.06889886218167152,0.8226506730059765,0.02980138221716444,
 0.19283855632205682,0.569626862388029,0.03870874498488075,
 0.33173032979037564,0.5054446385044638,0.7024276580896617,0.5365253397769512,
 0.5429423316936944,0.9148859968021628,0.4697719387632572,0.31053414320812767,
 0.8540619822217996,0.6365733471553426,0.7033514793468905,0.8951892352798996,
 0.05048875224115801,0.14889538366028232,0.5693097794580004,
 0.1448138720822365,0.6514479882043825,0.9681012443204312,0.6756690678916006,
 0.6032951508700757,0.9629944018938753,0.47193473668966845,0.315509770375338,
 0.9783167491916677,0.9924845146357604,0.7948232941251437,0.9575954596333003,
 0.5141734858396975,0.169704319442051,0.3594897197405281,0.07741615223424558,
 0.3994637483620942,0.9879462864047326,0.10822662173411168,0.43663317297364,
 0.006327867432711409,0.053418909426332095,0.012647078608323814,
 0.595964299008793,0.538179124702946,0.7655258184780448,0.20552885031991497,
 0.539751367733163,0.00902845266743757,0.031320640280070755,
 0.020605817398126725,0.8551359518238195,0.5213144023106102,0.622330969082608,
 0.8949209324163482,7.236454235381995e-4,0.10419519806974198,
 0.23105447503814736,0.6941395904538201,0.46718655279431165,
 0.7641653440754355,0.1432059088985408,0.40924549414368105,
 0.44875755278121354,0.6956250343757454,0.8363374512969193,
 0.19970450209923418,0.3616875691128848,0.5401838618736778,0.6105391533238541,
 0.16687592164594978,0.36083705069345673,0.32814470323199696,
 0.15602355856722805,0.1193325958681758,0.5981350054645038,0.5583449839722417,
 0.6356444473133681,0.7575189035717727,0.48368972319835335,
 0.21959069247685326,0.4439695343823954,0.27691140254696855,
 0.18149507644167595,0.38511940461588445,0.9538975061757017,
 0.14864472585390764,0.5299007892623856,0.39688968631076404,
 0.3903773107181957,0.5685564772644357,0.7400023106846587,0.796070122948707,
 0.6150870143097132,0.6652761539080535,0.835744989615278,0.7740814063297844,
 0.45919513792591093,0.23938760851925023,0.9904337337752501,
 0.3592154869088884,0.7425804680829312,0.2878273215994135,0.6444761171937561,
 0.848181460403762,0.8677049139893498,0.1654057927599022,0.39131249565524573,
 0.27557323457125216,0.720382791076964,0.09035547950340839,0.9670600409240954,
 0.6976298889126938,0.5676398628119045,0.02848912168722706,0.9553433000635088,
 0.48475743818625006,0.6158991319059618,0.9369918659462906,0.989651812702688,
 0.1809983034652124,0.4511537431518686,0.512907617258614,0.14220534092994708,
 0.7506006732450756,0.7461627582701128,0.2500940164034824,0.16894241942796806,
 0.404126686434816,0.6503982802146504,0.6178658611331056,0.9240716596716767,
 0.6180909980024656,0.15017781858976897,0.17329963449933805,
 0.8449451715682131,0.828234204489769,0.6867250840964437,0.4872984593713571,
 0.3742353223624698,0.5338239312149059,0.8176366873055096,0.427649074765619,
 0.6249331553741531,0.48497926400130265,0.23594263448283814,
 0.6893279055580226,0.8832505938758767,0.31306208884312003,
 0.41962906668176414,0.26262435623280356,0.5550980174883319,
 0.35896412066105754,0.3420518902629618,0.41981543747680583,0.267300059144968,
 0.4620596067022057,0.01742066037254475,0.3301519359623988,0.7608576162178335,
 0.24304565483488094,0.9157209558416377,0.32298193848509016,
 0.5396053699912914,0.6069237656977072,0.11135188587218992,0.6073526811487772,
 0.24277715398191457,0.4346391059805619,0.4315550711067001,
 0.23525859238629865,0.25600335503159344,0.5424778504938393,
 0.6464235815474595,0.8535217439759127,0.3014257111148053,0.35120300022457585,
 0.8757986540833657,0.6774883845517964,0.8954126360463364,0.40441228723918665,
 0.8212626995872119,0.10768924359298349,0.21188211631470955,
 0.6802578963029317,0.22866305507595597,0.7667303888428747,
 0.30646917392155626,0.2307421575702453,0.17655986052528205,
 0.7066938506040592,0.004721126877038628,0.19181343173615084,
 0.8548872267716351,0.3264547037116996,0.3854347974890906,0.25751214771143294,
 0.7772066442481057,0.5906928144280371,0.6459703044504487,0.17246259288944432,
 0.6827236890722923,0.4922644434750666,0.5655460439012399,0.8271507105016589,
 0.9467515873202843,0.018952245856695082,0.283189676540231,0.9222921418743029,
 0.1628101635966802,0.289701258268529,0.7496596733386752,0.3873719448574898,
 0.6970285584383006,0.9490221999285054,0.007585739685562198,
 0.34148583450667425,0.13459508653332675,0.9162010217918731,
 0.9677108366341444,0.9205873948931462,0.5203046181248885,0.5022427718356471,
 0.2849310536404436,0.1958557211018711,0.9174778666709672,0.39927057575452074,
 0.6346245678484747,0.6765722598383115,0.92875494658817,0.831099746833402,
 0.21678331099531634,0.606141274467818,0.17650228091944287,0.8228354460033749,
 0.4782595040402603,0.9324331264547046,0.5772533154188266,0.0923436290444638,
 0.9032490968162048,0.722478987771102,0.42785867694055446,0.2765148511239929,
 0.08175085288824624,0.1848481696619544,0.800596163994879,0.43150515596181216,
 0.44140965981629343,0.6626327451387464,0.6642570122473075,0.796199199945857,
 0.08429524916180342,0.26699779761218234,0.8895206113352845,
 0.7368373832379117,0.8871202429066563,0.7085308960050428,0.8136494282258989,
 0.3692333459189763,0.19533524862816842,0.9667364271746671,
 0.04247798602034247,0.7902661044467727,0.7060888577558312,
 0.22973315331582322,0.6894507355759848,0.5699259631957563,
 0.33782078846823804,0.9009598854436267,0.40960048344601074,
 0.3819669685939592,0.9208333815812489,0.039729095296048844,
 0.7279077894014253,0.469940036875945,0.15007983779259082,0.5160867653426693,
 0.47685914591031386,0.1897516961439698,0.43453885843802287,
 0.9495986728210074,0.7796914296442596,0.3370658803964248,0.9626861253174468,
 0.6616920639189618,0.07106071952612836,0.9602083188145878,0.8530886231440792,
 0.14560260744836317,0.7864986429749431,0.07542358353923839,
 9.214262966851239e-4,0.3073877387459204,0.5696580112187152,
 0.6055586878795455,0.3487536846293724,0.5779190962881555,0.8236114954566968,
 0.06761799383714484,0.4989213485968095,0.5394574019520497,0.9500565952854607,
 0.2605936885110075,0.7707496332808266,0.48931744322255777,
 0.02940947720620324,0.9456431723516948,0.08223113420803996,
 0.5407947288637543,0.5921914322214983,0.6368126464443851,
 0.029545530256788854,0.5313705777811801,0.8025061502601798,
 0.25961358791979594,0.6293365257658733,0.2257525300710923,0.5686880994354124,
 0.6968687147184214,0.03135406338788238,0.2525415192802938,0.9321512117000209,
 0.6867405554378259,0.23335123969260363,0.6829180945964112,0.5154870623714102,
 0.8271170039142214,0.5783175045522876,0.5481695550425157,0.2820733240582821,
 0.698009506550488,0.38247701232644404,0.09571906830553956,0.7324505449862972,
 0.8463235502548666,0.5623502585347019,0.389385171326184,0.04077334553833012,
 0.8316541599518958,0.7644592236611781,0.2907534421235065,0.8444237425549311,
 0.2889412378014268,0.3308388894760137,0.5793212417836814,0.28647725051268436,
 0.5163542967758922,0.2519868369567242,0.8742995538726789,0.3538992702963787,
 0.537531665163247,0.2691210764151495,0.5392564834661382,0.27997715632334175,
 0.11444297522555735,0.5773497150775121,0.9179148527042996,0.4619911534294979,
 0.7395773228157558,0.8555028450307962,0.2178148089686902,0.45917610228539096,
 0.2374354405785557,0.9411614037945037,0.06512098077894679,0.6374270746425086,
 0.06360117112535624,0.4901625043170684,0.8165760601143871,
 0.22979392923414077,0.6873785802297459,0.4766845262777468,0.2971830262372407,
 0.8282716084012809,0.5941519240710803,0.823958286708492,0.29365546176611845,
 0.6017641377688714,0.5661272810987611,0.6316844993767214,0.13091870622992718,
 0.26122865951682495,0.9941303841230427,0.44113535160130735,
 0.2379839720718112,0.1465380906450251,0.14585594243010935,
 0.24185092715032575,0.9296799995075178,0.3936415289962256,0.3725122436059154,
 0.6556145643259015,0.45195495875752545,0.4405319958639812,
 0.39111927430881743,0.4753020293715323,0.11380681759902922,
 0.12379008512129364,0.4322652727672006,0.9775204020302272,
 0.18768988453302504,0.32390547371385736,0.19177735620790526,
 0.5321181041218632,0.15411694247793672,0.34144641458699365,
 0.6380290967703978,0.24888796001951463,0.1471320228453069,
 0.007567392134800177,0.6890300148996487,0.3946602924688706,
 0.7514304033981007,0.9226887566858865,0.9941178144022527,0.6581220152689266,
 0.0013721346668715828,0.2792875454756979,0.7379773352213268,
 0.9601416979588634,0.24702239610837795,0.15669090149857012,
 0.8184571198712931,0.2459541617761425,0.48960006481093865,0.2834994538812978,
 0.04699643891057259,0.46380088599385483,0.8426401662255291,
 0.22742178429677218,0.9335879964565718,0.23533676040753582,
 0.6868917771449257,0.15653900433935997,0.5279785417435742,0.7988586540285585,
 0.4042570099983247,0.8929540693752762,0.9149950644613053,0.6032265945719054,
 0.5285989286616084,0.8186740415252898,0.8973434862472653,0.6303606344011371,
 0.5595807650937228,0.23708443520297084,0.48739499508763995,
 0.6384708535634446,0.10747143162815265,0.8493466916330839,0.1368507513766758,
 0.8293830960852993,0.39029262037667956,0.03746527722345294,
 0.7438983724920442,0.254185475102917,0.661582095125973,0.6765278403554156,
 0.7463019056271598,0.0407350804504204,0.9123449521156284,0.9306533110247952,
 0.7213904075249951,0.22204397785286578,0.20962915516333247,
 0.7858853333386752,0.26936630695316544,0.6826469002176682,0.4790950976789713,
 0.5147313203127151,0.7886449088471856,0.7111990780055567,0.5922869452728355,
 0.19586795981085126,0.11513959637505233,0.4352594896320573,
 0.47348141684797473,0.09533739101685046,0.33559126227889746,
 0.708420278586294,0.3308300775863984,0.9738343429395022,0.6892044487992653,
 0.9550691119202903,0.3351346261533953,0.11603080617077044,0.842969566816025,
 0.6600380509278306,0.06251481162619699,0.15582605092350332,
 0.039347077536951325,0.44106344641403483,0.2809125839153138,
 0.18345315813601815,0.9741280895937994,0.14074026341773038,
 0.3401100729646349,0.9835004683806459,0.13281626094423227,0.7854603503484284,
 0.6161509843891075,0.5232781821870519,0.9715925195908375,0.8302266716011992,
 0.8080144454807406,0.8928077127085784,0.38638748025385117,0.5714815260106932,
 0.04048331324494814,0.9685412126753309,0.22804443325235668,
 0.9310795000084112,0.004983037107991861,0.9985010320263692,
 0.09288873566436417,0.43932176900546127,0.11091479123046544,
 0.5440848356567365,0.24446726718332001,0.13734270868477116,
 0.35109812620005487,0.8620418425441719,0.5723904996885335,0.7264259783753793,
 0.5875580077927992,0.3043449783726385,0.927354774985625,0.4636967946663415,
 0.9292839359061865,0.9985173408007675,0.9982147206704742,0.25448448500004706,
 0.2355180825347355,0.5159068059205008,0.5772685244988127,0.8943671961078585,
 0.12677063738038008,0.2029698377659398,0.9983244003496148,
 0.38672204597998383,0.3899242912522317,0.1874409960965242,0.8209867363498495,
 0.4547690100734736,0.28745065750627663,0.9792282280981975,0.4398112879022762,
 0.035287951986222366,0.7411069233796379,0.9460489668331322,
 0.4890870384352859,0.9046440723720206,0.6737561106301685,0.3411059430149479,
 0.6393292479141218,0.09886255855300619,0.838095682266576,0.8422520900098265,
 0.8269391978362087,0.02817713817648837,0.011839237375803946,
 0.2620883308962497,0.06054279707187027,0.4736911608309996,0.9742140885396747,
 0.6220081919249033,0.9528649058453293,0.4934862810242002,0.20134003332908645,
 0.519394194192591,0.4882804094460633,0.6301392393388777,0.22824839256806595,
 0.8507146411095089,0.7274075092688974,0.7656081281654046,0.8908832925258231,
 0.8283924172778675,0.6175885029485286,0.03156934348000662,
 0.003170260478760567,0.2791661446845285,0.35438934380410037,
 0.38777450636902633,0.5398386851715555,0.8460064044398372,0.9306458145685058,
 0.04146957489870129,0.5121953760086962,0.7120125935702107,
 0.16913899844763036,0.5459529886937495,0.10927808763721392,
 0.4422677084023746,0.17615534816981326,0.5980580715158119,0.3553115197019623,
 0.4287480102566458,0.19836461475917644,0.10955736023561258,
 0.4813554174963526,0.7927858328045062,0.9384552767928407,
 0.048560770580069246,0.6630811320875922,0.22816721915227,0.5651711029561979,
 0.1398577684948501,0.23007878698032924,0.25820465320020836,
 0.04195097418735394,0.07745058494290258,0.1704656306581448,
 0.7166609382125516,0.8714161203526012,0.5936583834254435,0.12768487827786656,
 0.4074821759471374,0.3665745467489736,0.5968470931702783,0.1693201263014299,
 0.9456713479852811,0.6875154782979253,0.09980156070999546,0.193945970685926,
 0.44512194173529596,0.8517191100334537,0.9190094306559741,0.6167808872114717,
 0.5657563290676138,0.9219558343730934,0.22180793526364362,0.8617498872037199,
 0.8088966264129843,0.5055497924786273,0.736899330828771,0.9257643176705127,
 0.5439705419213883,0.7569079205856256,0.7032114845729751,0.3132427587748773,
 0.3563207288824367,0.6321563780143322,0.6540496114560591,0.7462834197145346,
 0.609071651215885,0.7631518517731967,0.9556815913500748,0.1284105849834155,
 0.22507869311816575,0.43304410535151727,0.5064084888135622,
 0.6664995427638462,0.42213714430203375,0.966988552709134,0.8763834521217295,
 0.46325223699622464,0.125635903900158,0.544420136674279,0.25313623099222715,
 0.2893928900194743,0.6188576511023423,0.6590931104058625,0.9916935519711174,
 0.48811200478201866,0.10186734801173036,0.6837235008588696,
 0.25684863714122763,0.7405558532227091,0.8366112341811429,
 0.10763351747864391,0.6553506531631941,0.5150893040393341,
 0.17438648241904364,0.3501947685758441,0.14610843794245132,
 0.5224249321553649,0.9407216508626629,0.3143762223259915,0.747963265408636,
 0.9250523212049562,0.963029368396962,0.6571144711646997,0.1535509188752544,
 0.32425896027349377,0.8547102002477927,0.4591384980624995,0.6362498615497221,
 0.5533753396187395,0.5238483134824472,0.32957520855854416,0.7947733655968784,
 0.6931815172114384,0.868177609802717,0.35671864009620935,0.37931959551381034,
 0.4144113011230449,0.9572286072377358,0.3126650365511543,0.9687340208328158,
 0.8700494187719425,0.752092237555673,0.39027870321329516,0.5886550479408641,
 0.9774824159963822,0.7806566644633811,0.07301721253433924,
 0.22906782196926312,0.8155494511078378,0.5229546184279191,0.5625789722840889,
 0.5146975555834364,0.14234843027324273,0.19020139937564828,
 0.24120922159595815,0.07255933018062999,0.047634823325133024,
 0.6030406817462924,0.6895798481286771,0.7547540601635682,0.2931051100209604,
 0.4450990474547486,0.625471125413399,0.8117709651607065,0.05892777338607891,
 0.9924173471437722,0.9356803541371371,0.11619474831864007,0.4971619579546719,
 0.027769144476188723,0.34494049234344926,0.385517884731437,
 0.33353788396969875,0.9050061674184853,0.35937061131801173,
 0.3619421376644937,0.11359905893345201,0.6777292998354103,0.913721282648109,
 0.27783469697870344,0.07939118019349967,0.08251369600126501,
 0.5954707984547857,0.18252772554306718,0.25407941384495625,
 0.4672480320892599,0.897784829920138,0.2884502427720669,0.028180315261433386,
 0.38240410802599023,0.4199511614955618,0.574406320616897,0.7339917082327632,
 0.8077100991920088,0.39052944590392125,0.4510414966279527,0.8982962389741622,
 0.44390191495400033,0.7420167408685758,0.20246150017337317,
 0.6015628672894651,0.4658331649443146,0.7828415346369307,0.577914389018203,
 0.465638359504277,0.4003207501245547,0.09289163556783697,0.5887364450462355,
 0.8992374705093064,0.2585552614809401,0.7961570523906839,0.7822347764281046,
 0.04368888949032368,0.6722696070750362,0.5432326707965289,
 0.13158957700263096,0.7696831965142468,0.5850056594846954,0.7445960722466503,
 0.004407459371496114,0.19364257864858747,0.9048376914275018,
 0.47101378318216924,0.2964719633885393,0.5881328540677853,
 0.33404322574958245,0.7641964116890386,0.5714822970539877,
 0.27285848939439017,0.6168537920576923,0.07000296447692367,
 0.9199224642023396,0.4355516595048271,0.4470855950607435,0.29994927098601787,
 0.1651353479309362,0.5670381522986138,0.3856551573144731,0.1881175940805826,
 0.4775274460200636,0.7945945333235307,0.06630359546518527,0.3557537912524096,
 0.9200271504519211,0.766472516162212,0.6700367912255275,0.7013969264165655,
 0.9968445156231689,0.0664115284009602,0.7100329028491981,0.5394792028254894,
 0.9732354708270573,0.07473386777386148,0.5344628108983761,0.94098918124165,
 0.7707941373146523,0.213540013237434,0.1946263198090492,0.3044140280389027,
 0.032211249996175195,0.5329614130030064,0.7992164169567354,
 0.8555328668541018,0.9061493355062793,0.13106053375119298,0.5574213939536885,
 0.9144856190738807,0.5814805856975578,0.9920922215116688,0.1482159374627703,
 0.13065301255280848,0.6754625282106689,0.8116665706140129,
 0.32742362846840267,0.5173206930154024,0.8391690565694846,0.8848455504559395,
 0.8488381453249725,0.23681539245948824,0.6635044121331002,
 0.07649767235486471,0.8566475583600928,0.6122152395131307,0.5996309622172832,
 0.7682062206103241,0.8327102460109616,0.013892567230076258,
 0.3082020551239948,0.41592921818666384,0.8143872750702004,
 0.10962389544160334,0.38416071105940897,0.49111028559532066,
 0.20448861758870618,0.12588981268707655,0.13309043598630277,
 0.5590348598521009,0.8841441376871022,0.22969934702242623,0.6503950191962592,
 0.9364884157071325,0.15664727628733566,0.7114803323843613,
 0.27033410340164543,0.9097464001531997,0.3485500718491271,0.969117702738816,
 0.23076402844243138,0.12857905401824343,0.13564851464299488,
 0.6468066277270198,0.5276006923954281,0.5039225124200544,0.6467071093566495,
 0.897140182599194,0.5635699148312615,0.3495671024540967,0.7002947483601254,
 0.7215282175319806,0.7417710796627758,0.2598474925259766,0.28607174924153844,
 0.21902137075792927,0.23760203273939762,0.735053456932969,0.7748816540786438,
 0.24093830807877015,0.806012762374164,0.33518062286528183,0.4594924177987558,
 0.7888532738244645,0.8776707550092712,0.8960376932414134,0.916420320587605,
 0.8690987106402487,0.16195068046771532,0.01145389044369749,
 0.8536732227005399,0.6438744481468905,0.08435032545090015,0.2200913985889983,
 0.4209810024164725,0.9802430046818862,0.4000748276296091,0.7291679442354082,
 0.5728726188678748,0.49977391861093434,0.5146327170073932,0.5299425957160101,
 0.47753660621879157,0.007797869202722962,0.26150444536787554,
 0.718225179048737,0.24769456775225773,0.7262294533018911,0.6777233577684507,
 0.13505303303980365,0.3358884823422579,0.7718163874994621,
 0.45945396236687563,8.503284604258754e-4,0.2377838978818786,
 0.1170311478841336,0.5367152413073745,0.15120265299871405,
 0.18216925338884804,0.4580984766864845,0.983616635130129,0.21573595465034756,
 0.7375093083741493,0.2356769358517199,0.7088931153203812,0.5985030216155329,
 0.15306171667568358,0.5404479045916715,0.7214613998581844,0.494211251438317,
 0.7796106562583752,0.5926622585212424,0.38879552201392575,0.8295910356852575,
 0.405327243994675,0.5207772055254567,0.6565902576597558,0.9123910937609259,
 0.3819381770199861,0.7640072861715473,0.3626259500091368,0.3482791127424538,
 0.9730241596452387,0.9477850082554753,0.8717148271198032,0.2109492753510891,
 0.40473303824635654,0.013255535365464866,0.7455931686394822,
 0.6793715592491243,0.3599788628237125,0.9682730598643997,0.8208044076658991,
 0.7181810971051901,0.3451659032434513,0.569710204291454,0.9913171562540413,
 0.21355304733767044,0.8726982066003787,0.7313723729003777,0.6500410022682453,
 0.26628754352672845,0.28518819930568906,0.1277858615683195,
 0.5273566590127391,0.8846509931165707,0.5715779986653382,0.00838009832665676,
 0.8218082048339579,0.5617981839451001,0.07016539041766445,0.6539038853274587,
 0.7419397952689911,0.4213619228844493,0.47651115745882744,0.183747973800094,
 0.4883121995917421,0.7303218819843823,0.43354026043715754,
 0.11068107520702908,0.43442560257009244,0.694418184857223,
 0.27180040930850946,0.787777005796295,0.26819606680844066,0.8075360472532811,
 0.16619324077720765,0.5292804073448475,0.2060318849981566,0.3699158141217117,
 0.6911361476421439,0.7993300775963857,0.3877364191180461,0.08697260414452856,
 0.35890569095666536,0.8621966656668936,0.3293913290090864,0.7322739866575511,
 0.975329909494644,0.30474639981062157,0.16009916090103393,0.433147255869917,
 0.3396574860532827,0.9908844461050531,0.5984030069833396,0.5753367235697969,
 0.5391951753993015,0.5010692133614358,0.029522862393658866,
 0.002818048285733177,0.4346600411791105,0.6221361865022793,
 0.5084125511857879,0.5721268028093124,0.21036345472651696,0.8928862015868622,
 0.24906992811795214,0.00346392097706949,0.19416377311855937,
 0.6332498450889288,0.7864337857175961,0.3667599610513347,0.6768167411932989,
 0.12456221965733072,0.5915359428194112,0.6769943457523032,
 0.09722947740840004,0.04542089215860323,0.0012104082281224615,
 0.27472108229991576,0.8908805161156835,0.7694370928619517,0.8297273482094897,
 0.341721679738447,0.9563763645306325,0.07902577954125656,0.9099245142051596,
 0.5744225777636773,0.3853159617260208,0.15098164829976246,0.8448810164646878,
 0.9799763480909147,0.8699427083761065,0.157205080323635,0.03717272910763947,
 0.3530689767758911,0.24208237068273397,0.806328993501992,0.7688877261038591,
 0.5645312192203433,0.21521359445358446,0.082486561817248,0.5506232323328825,
 0.8327609686105869,0.44778103083030385,0.31375683591507164,
 0.8774537840720711,0.4158391473142913,0.17139962633749628,0.7732270906347051,
 0.45303630750409574,0.06091341490527813,0.6046493697554,0.8507410472026231,
 0.6558166039138632,0.027898911083494182,0.79861188665832,0.10378679095716459,
 0.4188710217285735,0.26191797686429885,0.5697140443992053,
 0.07565305209398443,0.4226580983023759,0.13810887364260038,
 0.09455414183265975,0.40195192677672664,0.5863183570855789,
 0.06633878265483895,0.25709095451936315,0.1396335695082972,
 0.08748455995518223,0.1634700813327965,0.7708678636174559,0.5662530953454414,
 0.015533813750538195,0.6173761710271762,0.8876419372381303,
 0.8059578675900336,0.6585114097777816,0.542717402526849,0.08555421330146984,
 0.07100231999300755,0.5140580520783585,0.17881178474259352,
 0.13008895298937562,0.6261863023649356,0.1797026359230246,0.3914992483878602,
 0.20329056139569046,0.7481007405773705,0.30713532271917954,
 0.6296744675680427,0.9712392517273158,0.9338261050340785,0.17250175216978147,
 0.05823109643012969,0.9567599651409713,0.5174207009447184,0.8547355316144392,
 0.3012057455706074,0.09764345908004468,6.05261577722227e-4,
 0.9152938567931775,0.6865742470934588,0.278531825506813,0.4056527163023659,
 0.6059760724166853,0.024632064721161973,0.043840204937993676,
 0.3100712789667417,0.7743901185455484,0.2671612628880463,0.5840178475877348,
 0.5047394064992817,0.769232172923763,0.406596025968196,0.585419284830979,
 0.055297124447290624,0.07412035459265143,0.5884091427466157,
 0.555730493987254,0.3546602545308505,0.8882099823851543,0.11697894668897313,
 0.23789373348913911,0.002036131264091079,0.4211947389414946,
 0.1623720365038619,0.16139745859342958,0.99163188654488,0.7454099806698637,
 0.4307955716796451,0.5353331512076043,0.2504325422637008,0.8699599922496232,
 0.850055912211144,0.1689019135885823,0.37928499058890164,0.8232079415884597,
 0.6766291884550052,0.05307781635698028,0.18360224429309902,
 0.25216990770374526,0.48283244616545873,0.4874976935214598,
 0.025001636666847404,0.5498348031953003,0.2565372442845735,
 0.08550205185576787,0.6171797979994882,0.2519792033846985,0.3017212535516052,
 0.3719900133001157,0.07716471313468565,0.29079317196279963,
 0.21249087448984627,0.7407924180345551,0.9427037436198566,
 0.23592027284273465,0.654962120110237,0.48081839431573514,0.6757101727184895,
 0.30591446507666387,0.5006808636432487,0.7107839412387308,0.428079659004733,
 0.16961208081321844,0.5975749429746452,0.18353508866489332,
 0.19265953693542315,0.6069865158445132,0.2609636889537331,0.8122797129380993,
 0.7072091243641617,0.7477522564178355,0.6620465395723728,0.15883526396635128,
 0.4992639900063187,0.8019209993988642,0.8314169659404141,0.9724894890878757,
 0.9922577799440799,0.8543962094221669,0.4155705655548214,0.7732294000067593,
 0.7079683783380659,0.6248773994156881,0.1731253130882845,0.6582353543415027,
 0.34437844475593415,0.4934200024408463,0.7316681670901442,0.6830614109455146,
 0.3071935783101297,0.7931980824118685,0.7214949187485006,0.3593423530856188,
 0.4583801741127649,0.9039954483236305,0.5140460388640663,0.48930706172006166,
 0.7477217838007276,0.974049000430698,0.4462540314745289,0.5265321544110458,
 0.19993908578236286,0.4508812113845616,0.23939467349727284,0.898313285426105,
 0.25440995886432427,0.39650923465034105,0.6778978293902281,
 0.5048204902311931,0.2310309001366888,0.30359108696827086,0.5980653514724935,
 0.2184326226240254,0.2457439992677778,0.22020466709686393,0.0861310945236865,
 0.6627703097130939,0.7902663814845945,0.12909665656219793,0.9260484997685337,
 0.8386646548793437,0.4635489068904287,0.06725731815951108,
 0.19577138138619987,0.08647352228456773,0.6169663115029738,
 0.7444027722414668,0.8571642321063667,0.4263512455094184,0.8397789486000922,
 0.996593887761239,0.9208307976716781,0.4536763408942527,0.1320810532741019,
 0.142170572500522,0.7924710519633902,0.7313097408452451,0.4856483132903746,
 0.8668911771103095,0.7078319554479237,0.24805891495891352,0.8332987759822649,
 0.7342606423206258,0.9678412030681561,0.8443293692305106,0.2992211325855426,
 0.45733910616456985,0.17895587431837834,0.1516175134415435,
 0.6119494407327647,0.7044662152415275,0.4994399250741979,0.9008458151824655,
 0.6297149399337597,0.8503700849174574,0.6936234538122779,0.12560417612803787,
 0.7497766880830812,0.4149694334622456,0.5212413025869566,0.2580694932883979,
 0.918042067551197,0.3061727586979245,0.6041571085744074,0.20268566839468394,
 0.759325129860381,0.8270104094659991,0.9022647882764867,0.4272018155406835,
 0.907615265323025,0.7633320668423642,0.4024593367374637,0.2804276862829115,
 0.15482099747452105,0.8010540183277514,0.5025441269314068,0.7262386510095353,
 0.9959600057052278,0.8052947787338284,0.7968896211036698,0.8691781798649688,
 0.6101148840465567,0.13718831992648695,0.7612783932370082,0.4019581219637893,
 0.052682156402103275,0.11334458899673605,0.6755305634622015,
 0.879600788439214,0.43155745113712285,0.47521206473004085,0.9983716441617246,
 0.3862757498869711,0.1348997271369814,0.8991727100616348,0.05594436382184109,
 0.22291018686228303,0.5510056207295124,0.7531631535694672,0.8464988388495357,
 0.45432158898009234,0.6002441789024988,0.08827371184836685,
 0.5773074362671687,0.9178181099432839,0.19194543985699897,
 0.11899555977541643,0.9763408369212797,0.38059469882496577,
 0.4036346651351752,0.7799194692254745,0.6259933803641912,0.5566331361591319,
 0.8104896652816871,0.08500321040323311,0.9474311172404487,
 0.13412902769875645,0.1632797988288187,0.017202808333459108,
 0.6072151094815474,0.2283244771769929,0.5162353720682749,0.9703045568440279,
 0.06096877833063086,0.30869374491211576,0.17112282576734672,
 0.3027189536485747,0.36170725353561295,0.7281795331518232,0.9067142113494433,
 0.7558982359513221,0.993225524699859,0.38710449152340665,0.2913512009518475,
 0.4240168842273979,0.8037070684828992,0.21792293095161674,0.7323293290920024,
 0.3617893406410396,0.0803277665651364,0.636635538981954,0.17250020378775,
 0.1767583385916205,0.9612029396068759,0.7328240334516599,0.46755156219841354,
 0.24058523305264046,0.11485590568813087,0.6123892331514431,
 0.7130042188265728,0.8010692960986814,0.998618285185227,0.43587528334692094,
 0.5270048773240416,0.019047566764488155,0.6313642211927633,
 0.1412138092184485,0.48184440511131443,0.04263502014029108,
 0.31839615069547256,0.02409297456999293,0.2278965730988457,
 0.6599816967415871,0.8317651198308729,0.2911259477878023,0.1827473082987251,
 0.6352100928407706,0.6037264578774721,0.9590911819357688,0.9254210091394869,
 0.9582080411910118,0.42763306227072284,0.6190930367744284,
 0.04925043466483925,0.9407895461415197,0.8985464090805491,0.9660900909088389,
 0.9774780481297536,0.25742660197525635,0.111996967679307,0.6453103537541355,
 0.7916120761631036,0.4562583440869292,0.6637711367118015,
 0.045699866488891794,0.2017290663474075,0.14818575422394542,
 0.45384094012414566,0.8844900777907441,0.6266063344897399,0.9788708869685033,
 0.1749665689318891,0.7213354693221561,0.09795499979454925,0.9574039350317477,
 0.21502085459778952,0.9843266668230324,0.5264060812753826,
 0.04884420536486678,0.04373713700644677,0.798016884110476,0.6936701620715346,
 0.14738385510176633,0.6210910424357423,0.8816165365810505,
 0.42060872372716973,0.680215481617975,0.03319700928280689,0.9108255823490776,
 0.13426930536056647,0.9763985060496227,0.0646530701501955,
 0.20797103480287304,0.2568428440814904,0.8718757347157433,0.3029135752398473,
 0.004779533388851065,0.12486205363729963,0.041019383035800216,
 0.3904215286985616,0.06945850587878344,0.41570320437328045,
 0.045752969181478864,0.8408042188694173,0.6914630814751466,
 0.1097344201700875,0.8200079777149037,0.8446919463822268,0.19522059398114733,
 0.7278967817971613,0.0064144650277186965,0.25408756235809604,
 0.9476308523788078,0.87047057799897,0.2558216932134668,0.17219979325374168,
 0.7093397566135716,0.44060553539210934,0.49057509330992066,
 0.4508235557399307,0.3169273977951169,0.9627159140522333,0.3647883898632147,
 0.772226333032788,0.46386204062897307,0.4252274636136496,0.49824898376820514,
 0.7674674690216098,0.7345285166012789,0.20072298599055083,0.9744941595592346,
 0.5041531550306323,0.7644510293791942,0.8391514476143893,0.9671222904227235,
 0.14128221198080348,0.8784007113103254,0.4102586816432007,0.65409880926504,
 0.6809334546575779,0.30687331790616035,0.9354268041467287,0.4443545031684216,
 0.6958428509821981,0.9674314882468656,0.3629723622083295,0.4674682126501002,
 0.4634684042174071,0.3563886620844253,0.22718305986632126,
 0.40016960271448543,0.21648585666876996,0.23206028053147165,
 0.8210832826735763,0.07292201650143049,0.8306069484587636,0.8424402587868789,
 0.27321415650985226,0.1402287310750252,0.8303466166396767,
 0.18094521740601632,0.39802163670731217,0.5754306771731483,
 0.059838559531150026,0.5967002712568147,0.22122134000006244,
 0.7198080496514291,0.6700756354798215,0.8040082194662631,0.2943881556321374,
 0.020323232949854497,0.36995032704733755,0.6889609507719855,
 0.8349359273768076,0.5416304153326587,0.2839503801345846,0.98599190696835,
 0.5216036835207966,0.17210592783832313,0.19655470962086097,
 0.9337306405225126,0.6148607986211555,0.9150311892988552,0.32431657009125503,
 0.4559583531817014,0.4074060101609618,0.44723600725997226,0.5722561858149963,
 0.8349319735332258,0.6000239695753404,0.6928851954102381,0.34929962522830194,
 0.13748264448325642,0.22763453777629206,0.019860551549342187,
 0.7814414907123108,0.6074309670694755,0.9496125595622182,0.5471464303969132,
 0.8873031150434061,0.9120780664038998,0.5281137315266563,0.3711303275722042,
 0.8511901872798405,0.3729916144429065,0.8374218431653218,0.23563702417853827,
 0.10135671656151812,0.006370792800053149,0.3899743708178208,
 0.7422032021547565,0.2067199547318459,0.03719715862999773,
 0.32740582016092556,0.5011055830532143,0.4569404557401848,0.5079999006740489,
 0.9672278076089834,0.800768998377807,0.3943032535632751,0.5115615430777428,
 0.02015510697214884,0.21042111245215855,0.5201200279745354,
 0.3488335880888157,0.6365914227278882,0.11667988853919742,0.6692182591170288,
 0.6073778670029606,0.8728809326372499,0.03625070590336299,0.4958515693207717,
 0.26947319054925867,0.651834177461106,0.9074972226479825,0.0827772125972418,
 0.5960002877579453,0.7742750700802052,0.28392098680150624,0.5819473739455349,
 0.48061374314822314,0.5720390818998329,0.38513729739453384,
 0.8412802345245729,0.3809275419390493,0.2719297083349467,0.8465147301000611,
 0.17596963100979557,0.22997048982853296,0.7980095516243059,
 0.7219613894988076,0.8999209567061973,0.6985248144093725,0.4688869802464516,
 0.14872997326395676,0.7999624393524429,0.157928482708938,0.7231849404821257,
 0.614391080409085,0.35998140075938967,0.683881351084041,0.3993518581805815,
 0.5111664523657551,0.34793766324561015,0.8853347517730658,0.4494623166825077,
 0.9702489739720661,0.6788677131630455,0.25764367798679544,0.7996497441540587,
 0.03098895812825564,0.2269878584538224,0.676484057272759,0.43661016923389506,
 0.0930187146450131,0.7309783900989797,0.7560725417821617,0.5853520721746482,
 0.4769314213989658,0.5843734792699881,0.5586926564862585,0.4437663086844279,
 0.5322207248362514,0.7421289270573836,0.1829989751991311,0.6802788688084225,
 0.05677153704799065,0.551109675701417,2.537582840478336e-4,
 0.7179519582385607,0.5006057482171256,0.9839449934036842,0.5365073941351919,
 0.5657248403400109,0.3998248826828964,0.1563966665704679,0.12919951002728614,
 0.4910342673894006,0.39725442927374977,0.3479184831691242,
 0.22051007629575015,0.7752442174048355,0.38537802805811716,
 0.6666462679109454,0.986674698909106,0.5992728485806084,0.644341421099094,
 0.9114997475568085,0.356118611770162,0.7232607071652029,0.19998844708470398,
 0.7913061317895217,0.5841584264034311,0.6537787586248738,0.8858051942371411,
 0.7984393951119512,0.8875652914613077,0.187807252017267,0.797400090329022,
 0.2458221707602497,0.3221615115247167,0.7943032690050917,0.8733123521590247,
 0.11750529356326556,0.9465243536955219,0.39974079048326705,
 0.23843131895801295,0.30099890677484253,0.12175359536136643,
 0.11875583809327284,0.34445451308728703,0.6191214882282503,
 0.029105158315736124,0.2965629341862157,0.9571163619507685,
 0.7155072984436242,0.6263302734770251,0.6928728094534231,0.39436848599126517,
 0.9751759864679469,0.8955595222900007,0.7686573695047283,0.5779787642904517,
 0.45417678854126453,0.05532534155879931,0.5182676552787324,
 0.9411636058198317,0.9515406965236344,0.02870870704748507,0.5719705223272213,
 0.37312795144660127,0.3488190496195811,0.8055735703054591,
 0.44260946141892554,0.36009170941684243,0.3468482289341983,
 0.8243029264781385,0.45845020413932824,0.44593136372256303,
 0.6762720230396861,0.6508883563024959,0.5586908712071402,0.9251728432121177,
 0.3259154815758326,0.9582332155495622,0.6327739763945361,0.3579368623903716,
 0.44054830570983994,0.541732605279745,0.35127235556605285,
 0.05933858401241199,0.4957631787525624,0.5019042599130042,
 0.15521260412115268,0.627620963922024,0.2598324531818088,0.62919276027856,
 0.9255355797697477,0.21284486686441706,0.8282730774005084,0.2900791869880184,
 0.5791475877211146,0.9517347437114472,0.5713850800683576,0.7919108854628224,
 0.9908377606293258,0.3852167650409599,0.9830135298245146,0.3770159764778641,
 0.7829489623718069,0.3834881682947606,0.8210316581663182,0.5322630625606641,
 0.14485016202993195,0.1825858358833944,0.2348032762840082,
 0.23903955393835352,0.3350281343369098,0.29818941514507025,
 0.37197231761303684,0.7690018769460408,0.3588734686528524,0.7781495501567781,
 0.2778173343933448,0.5575950442541208,0.1599697788700043,0.7787218048606025,
 0.24534373462124337,0.3277322962628677,0.6278978477436786,0.7884051182193882,
 0.4074354500668942,0.5530269452962797,0.18135443089860503,0.3177308156404337,
 0.6659080632117891,0.47182718818920366,0.6884530703701865,
 0.48140784661027025,0.7236201574300721,0.12038791091057943,
 0.6286439274768184,0.15279260451356746,0.7500057764407282,0.7387360786157833,
 0.5072932795916228,0.8558068210706344,0.06856491219234329,
 0.29090780894552926,0.5690252309064012,0.4478788074622312,0.7219234720271536,
 0.8966066451903808,0.26904391047265896,0.5256585972282398,
 0.32651878171139614,0.3273674855123119,0.11172294062039345,
 0.7423358924497506,0.3258841508302206,0.3074204867706438,0.6584795069854373,
 0.5841580175408003,0.9819506164474762,0.4584516655446449,0.7351772150458122,
 0.4717713139823452,0.08332233249843302,0.38918578562078465,
 0.9910567542079245,0.10613437556992145,0.9906300112221631,0.3371556339080304,
 0.3650528274745509,0.8284116443380183,0.5364542233260177,0.7349121506355302,
 0.8298008899451793,0.18461559918435422,0.788374359576012,0.5154125518676633,
 0.817384981710608,0.1449086328774013,0.23211729132001824,0.2381132374668271,
 0.5700925474543201,0.7907861040986743,0.4788719327044464,0.10591414661536147,
 0.5766628778498548,0.7797367045554952,0.7779453052041324,0.2400445234858699,
 0.8082290073088236,0.623544471891305,0.18590443669365564,0.47178529278930836,
 0.6277620698535518,0.005696964373250735,0.40779800585859705,
 0.1798186168272473,0.5720403300857493,0.08364017911028243,
 0.42804369594388203,0.7011221959957102,0.6671115115235196,0.6934488382064831,
 0.9964485305223374,0.6450023104424467,0.4453358359858952,0.9723781973939937,
 0.0405397466929609,0.6886502227082603,0.1527105170430888,0.02722710299376696,
 0.19970973081096777,0.2457732924788265,0.7415588989519312,
 0.12846432000559316,0.6334223128922692,0.9581406232400026,0.3498507775596267,
 0.35606029265719774,0.2679435684397794,0.43998214895503707,
 0.3197771439431043,0.1682377735076105,0.6774782192059503,0.36496061475234565,
 0.30743164968841263,0.332135432482932,0.9382584504747399,0.7922939319693764,
 0.4750978271118411,0.5280339511333908,0.38615757809064244,0.7248551249305227,
 0.6777962196393083,0.44213881209376216,0.08123894841345491,0.7894005979601,
 0.37861954648788354,0.04282045707941895,0.0038257302716502295,
 0.2559281009503044,0.5537371655098187,0.02685064578768992,0.4181327246591484,
 0.48429315465691347,0.41109241596133717,0.3101324016945529,
 0.5854553427995769,0.8598713304389722,0.45328616811144395,
 0.021178155954778677,0.6039936653828821,0.8074361618261731,
 0.19327209993737504,0.5029105519561679,0.562281410294145,0.7331493528551103,
 0.3215971419710809,0.22793790629641297,0.5018449222540253,0.7092207172698413,
 0.18221463979531816,0.6568740947860235,0.9613226930814991,0.9825164443395573,
 0.09845323996588862,0.5737154278290986,0.7704060647543866,0.4743106605906151,
 0.911047655837365,0.1497450112153368,0.9190480030985292,0.8452585177435987,
 0.2458114634704497,0.9428758213695655,0.07796275744070236,0.9637134344983551,
 0.8831643195782812,0.8859769278259277,0.9332087472142495,0.7886330310317664,
 0.817254273649429,0.800387819425774,0.4081069258215143,0.9831056298343782,
 0.16205188115614066,0.2292897552306894,0.8799441699954788,0.367490361314713,
 0.13096974857322496,0.9384249280836795,0.34069558902627484,
 0.004942222869101953,0.8790329276423837,0.05703685605417752,
 0.9097700491201255,0.7634844399035614,0.031357036181378306,
 0.037202562906630154,0.5937571447896923,0.9052391436027803,
 0.34951951487924293,0.9781032971561737,0.3051997468090025,
 0.06666238210750641,0.14893748733111634,0.756032342813102,0.9194943005907886,
 0.30584666599246035,0.36381862442817803,0.160569717783421,
 0.45065798003874225,0.10490735351435698,0.4309048833223649,
 0.8686935547625423,0.17818052430364106,0.16468479131596714,
 0.13923374803894906,0.460382496629387,0.23143479721362004,0.7104278678822931,
 0.47196590873006006,0.25280063362839467,0.7559362563863329,
 0.052677568277964215,0.3601571353582632,0.4287754554201213,
 0.9201600012279806,0.020188677692087764,0.8216776427326186,
 0.7438712087996314,0.710779943045546,0.333667897761572,0.4964315008346203,
 0.23398141783284754,0.3644700022959275,0.5366056966460607,0.9021763412720152,
 0.26493226676797366,0.935276812775434,0.9400081162873506,
 0.053559051532136914,0.6981714173862607,0.6519934603620735,0.600205812156104,
 0.7284951331533361,0.9135953993113549,0.040041781584819636,
 0.05814207939967431,0.5607430565748835,0.8757400788957225,
 0.17069383884910183,0.7358871221829385,0.41421369608435044,
 0.8862813191771043,0.9812006652256957,0.9340492144898311,0.11494248399851548,
 0.056040486764426634,0.08393009739310842,0.8601004687910156,
 0.09734509495541144,0.4629956179863317,0.5337108141807705,
 0.12598543788102778,0.4274230867663116,0.6149866399587134,0.5471616017937101,
 0.09485095756505435,0.5531559676185216,0.06523112963950484,
 0.15578562609085334,0.7354446822425557,0.22165420032371186,
 0.046119391142619426,0.5793250407573546,0.4247810830195148,
 0.9639667871893022,0.4895103292306865,0.6003669677339619,0.274288197846014,
 0.19976534272062196,0.07422453499319748,0.7465639037941314,
 0.9050053353630877,0.24623553787464214,0.29219544254653407,
 0.5651764051116032,0.30843579808936483,0.743436158395097,0.9249618183760385,
 0.19443092109487115,0.5965046191184221,0.2149971712276123,0.9487856759437995,
 0.8431175734994631,0.6841734377311675,0.355951968621799,0.5308639503662963,
 0.6994670432506902,0.5199335455980572,0.9464588092276484,0.4924684469948537,
 0.797138238143317,0.15047334226791242,0.17096170571585312,
 0.01611869664581944,0.0013109788828357116,0.40815466813295054,
 0.8471127603582094,0.5876723684698255,0.9359212670760371,0.92930468607235,
 0.767852194758467,0.9870203288255657,0.36460726201664395,0.190903372105511,
 0.7983088261777133,0.7670226946012356,0.6470639651713811,0.2906825847579453,
 0.7463514245915154,0.7865841943217088,0.7552031359543006,0.6424888139096555,
 0.03691503644319993,0.781985433075818,0.5151969323677931,0.02015112319528445,
 0.6902168860282244,0.7805084010044887,0.7714341211585954,0.5622065588064561,
 0.6352506919307407,0.9375030987346273,0.3796018794320819,0.28264509760320866,
 0.12262441234986976,0.380898205851968,0.6277175672497098,0.6303585186366534,
 0.4071432540660551,0.3772322925766203,0.7051965457720738,0.22721716817755044,
 0.8380163112568721,0.9450816503316608,0.1958163990752131,0.05515231628422912,
 0.014849815156889545,0.22481297473175366,0.7293089600713835,
 0.8526821254161796,0.17211856933823866,0.5811205337076599,0.8912742796793022,
 0.9570601511448463,0.66195566663124,0.5492203268616048,0.6404951435620476,
 0.3426093190316787,0.20136417348440616,0.8027320669641445,
 0.03020278888823147,0.9392815518798742,0.2828682613661666,
 0.46930811329169464,0.75016088598459,0.7599909588581706,0.6640423012541511,
 0.5434942356458643,0.015740449535858292,0.18960777965813858,
 0.995987943402433,0.8373084062504234,0.8099721813099015,0.9789061662827641,
 0.29200222802232156,0.14760777964668192,0.8187340835652982,
 0.5813800214376439,0.3994966123289848,0.08817211752135623,0.7863896179228752,
 0.8149659521286736,0.14495617807822236,0.7232962372398705,
 0.28123167662242876,0.2291950853238055,0.9288870543034179,0.8179912795138538,
 0.5352973434950463,0.6999814791857234,0.9147152995596354,0.19897367556699885,
 0.7709809351872399,0.6700152675641702,0.17955311294569443,0.7875316392796797,
 0.789871013717345,0.26493611905384495,0.5138908381662465,0.18122842221464708,
 0.96676964765009,0.5523676112514102,0.5282829265297695,0.7539673193075308,
 0.6130149827520689,0.7727450891521833,0.5980035592561879,0.8617475320648398,
 0.17982155134055167,0.10110899212590985,0.71776388521222,0.533522306783113,
 0.6696371350644817,0.5066526857415236,0.7616859697103213,0.1093396722168426,
 0.3528368103555878,0.5822956357490681,0.25729033826411696,0.424453360923578,
 0.2972151783127809,0.45415581211751044,0.7222569046525992,0.6740613344455815,
 0.008644631358248223,0.1846178579703528,0.4126062103835403,
 0.9267374383119059,0.12581315131305315,0.8634375137757239,0.4575692666540341,
 0.5029215775964174,0.7397335150557874,0.525809457197209,0.9647125241827116,
 0.6457341811207895,0.6444782248074589,0.8875406788441624,0.6924121289549661,
 0.011467891137482367,0.32760294754367136,0.8594466023933088,
 0.35179480221790294,0.6489926511435393,0.9316070494519655,0.5447387580888869,
 0.024259231164098338,0.16814219464408897,0.5381906994336026,0.95301959959085,
 0.269385293947077,0.8547397853455455,0.01730594312708289,0.8985003496611691,
 0.5708048560567814,0.7898966456894001,0.7060217140840903,0.5545233624620017,
 0.051091729173276756,0.9523740296354906,0.9116225687248058,
 0.25340617488215533,0.9720578783041411,0.6829653931273424,0.7428254362203361,
 0.5364103789093038,0.8692044703709392,0.7882212514562695,0.32601604213307067,
 0.9652707922183791,0.583683802834653,0.3789497378099176,0.7296254734773977,
 0.3769964495084266,0.5938409896432927,0.31422868054677866,0.5114448964629114,
 0.3596003240548119,0.7872286125870759,0.94056653224983,0.7558014683769901,
 0.6979809483145604,0.9363926986064601,0.9804952784290419,0.07157442964579053,
 0.2177753302680967,0.9806675575632475,0.32270921347950854,0.066860721582782,
 0.5303120443378642,0.20265298028601963,0.1648072277699767,0.5967245450483986,
 0.07314485625461631,0.23414151170307218,0.6508657251970251,
 0.6462853208023256,0.002353623865880672,0.9647176730590362,
 0.4464057113766424,0.7375850352148973,0.18518593711151798,
 0.27766581620164354,0.693879869216563,0.7260869844150022,
 0.021952407962584353,0.6301820727036503,0.7112064279461846,
 0.43237852795842413,0.6048952712507057,0.06649415599646835,
 0.4414795420808768,0.4936518848284662,0.8897134926059912,0.9800059936721723,
 0.2998001566473445,0.5257960440361069,0.2726247959346114,0.24613440410986887,
 0.697922671657448,0.19289362563821866,0.11566475473520232,
 0.18627974248613666,0.5390857622503779,0.9039772664222345,0.8492680220786334,
 0.008216916096334348,0.19035035539362044,0.9674004315973103,
 0.174946284274893,0.26279684707134665,0.8821949930077223,0.602196444591349,
 0.49772586546467235,0.08867563311458804,0.07665520287389105,
 0.5179373293768703,0.34549604777480636,0.09102409476647522,
 0.9621807291217886,0.9757333838098329,0.6911578612219997,0.6323424879205283,
 0.3700016972575739,0.6686875365719792,0.232725680412888,0.5590171859175603,
 0.5782769801929574,0.5266265671574721,0.9012993852432033,0.693180468117644,
 0.28854035109692244,0.18456825661681897,0.5432295110846005,
 0.7321182404881819,0.5278636625952116,0.47737736133332576,0.3993363104854206,
 0.17771562327035595,0.4475414207152272,0.7447349069803189,0.8743159886451343,
 0.3953023713328615,0.9065100320915357,0.7176028688340074,0.6807847345997112,
 0.7562165233592668,0.20938307020976288,0.8951879164410836,0.7692004981866679,
 0.33854593485123485,0.17165152765705993,0.9803231484198828,
 0.9019887782329885,0.17409802251225126,0.7617470314330501,
 0.07403794047722134,0.11490825611518107,0.4233800054586101,
 0.35860745498020974,0.5867685045379607,0.6034544204901682,0.7939026054508019,
 0.12103743927250254,0.2850191151218412,0.013466560334073785,
 0.2068607246175066,0.22457998316223282,0.46506843917676044,
 0.11710857055905544,0.8600430746852377,0.012441227116762987,
 0.8942681904267682,0.6753055501482553,0.423312487727576,0.24343129009609976,
 0.9634645305461064,0.46302184890515574,0.3216326254723233,0.9875093676249838,
 0.08830326262125876,0.5650316891026272,0.22217619597941318,
 0.06339454670608824,0.0026050696976458454,0.5327370812298279,
 0.6381769678873039,0.9351940627437507,0.5308156505426519,0.5255116355329548,
 0.5362585755346823,0.4755029027094445,0.24605542757685972,0.4855448824662296,
 0.17594166949420664,0.04583503443085102,0.19368214394136873,
 0.7728786774017676,0.8351015956856612,0.7505453791987878,0.5985899169594152,
 0.3886985849341632,0.9086190287139685,0.6462438362249306,0.9482185455355514,
 0.5691644220831145,0.39073684919416063,0.5480279762111369,
 0.28249961847421967,0.1340302242438245,0.28186446341360916,
 0.19284215295594287,0.5653159183694978,0.4665492155498545,
 0.16292079178745555,0.7773470812509091,0.45988419765801003,
 0.8329830122139968,0.5930178591714813,0.2891190866852129,0.31672469177470863,
 0.5875488417333319,0.5103858005534971,0.6986143308291992,0.9543822677659355,
 0.4374742002103419,0.5536691705830845,0.190252427620914,0.7725310527941525,
 0.6412284979844249,0.1256451914000437,0.9801731726332946,0.7217204857889955,
 0.463537122403046,0.9316150069130633,0.8502435084340662,0.46867533991303467,
 0.42064518267197537,0.00875858461254464,0.5710774635924548,
 0.35687008295242917,0.866723017992532,0.9553776496710075,0.2475310024601235,
 0.8303731359505434,0.5529827433707283,0.8070260311562987,0.6715243465124835,
 0.7375761558733399,0.4737200626533391,0.552797407979782,0.0640538947343039,
 0.8919788263743006,0.0328886165325023,0.9152005954044456,0.6006828889611706,
 0.9415059302800619,0.6354698851829388,0.3162889105079649,0.1620188949075152,
 0.7368646996676977,0.6196694066624501,0.5553074412812771,0.40787604584300763,
 0.4817378634884055,0.4559010867951443,0.6129199313833068,0.7265785082692475,
 0.6249875975346911,0.905889505113376,0.8924573334405992,0.7869291213583083,
 0.6797255300351898,0.08104314974196059,0.12620842112215025,
 0.47945275243972496,0.17486997993369835,0.39247008520177096,
 0.13355314020747566,0.6256141742564765,0.9868855351800169,
 0.35998346078852594,0.18597524386683906,0.8580540584045271,
 0.25700726589151657,0.6330488837624461,0.5052495062620025,0.3600963831761602,
 0.07171927809489986,0.003216718205386604,0.3918177497549855,
 0.48080373540244137,0.7696306324025374,0.06477943625095783,
 0.38954030115367655,0.6059618639030457,0.8674603919760344,0.9903670809448735,
 0.9713320579201363,0.8191154732759721,0.986552894453365,0.912894464448814,
 0.6815063708921154,0.09280253326337473,0.09053127849798281,
 0.13652912766052172,0.4837963279090499,0.5951085561332354,
 0.13397938256470976,0.9824189930000492,0.21048552783078567,
 0.11985464801697843,0.4429388295680401,0.27442460395685897,
 0.07112897690030395,0.979759353377608,0.799446572833478,0.03864846127674004,
 0.45320144181829625,0.6743925899443461,0.10969394857415438,
 0.9493668780664389,0.8388451445841665,0.787364603258808,0.8475903656307001,
 0.4812217771377565,0.8541031105030414,0.7941808742056929,0.2015322718272482,
 0.16352883017936626,0.17483169549297184,0.6112087032572647,
 0.6173176847799293,0.4983221127161649,0.4029526704399685,0.7393971144891358,
 0.7744812151313989,0.27876855585774496,0.06258942432066117,
 0.5425102641369832,0.3565202228030102,0.39609844583529474,
 0.35583038871156825,0.24184020509468274,0.6712987726964148,
 0.8063042302390396,0.9419526189147452,0.7584387105098823,0.23450818368786563,
 0.19780759631798606,0.7308141133373726,0.8930117771947855,
 0.45019983153236687,0.6548113069063582,0.5280141811204768,0.7297087215026383,
 0.2306920690703813,0.17602726797832013,0.5175482633409131,
 0.33589195437388863,0.052445202668802815,0.08249133876669867,
 0.4077290024959014,0.6403000518009494,0.9624162138738628,0.22089036101928994,
 0.2214595371306498,0.23149006876819422,0.8252279728135435,
 0.01917608604983312,0.13047972262233742,0.28591528578914593,
 0.587257613856624,0.24345036752108984,0.7605741180597682,0.8358639644869574,
 0.9745712776160858,0.003936655181607129,0.21362272271618488,
 0.790534152388763,0.9182194208814958,0.8892302573205184,0.7159053355899324,
 0.802737580920212,0.11427531588148232,0.029382298914861216,
 0.32203434002249265,0.026291693470656963,0.583696839133754,
 0.5891216411729087,0.47319499465176906,0.9737622931368808,
 0.07691477674067837,0.13900841327253055,0.02759573087940448,
 0.3085736517131691,0.8559735295844411,0.13811683570336541,0.8637071148891413,
 0.11804741425191756,0.9657412226095198,0.9548642681925973,
 0.45091149347808146,0.5959153244146885,0.37058717674500263,
 0.6192579218792098,0.9982086470328446,0.25191346342926824,0.309538159051164,
 0.0604802429235044,0.9299895407882258,0.14119564742346635,
 0.13514690273146213,0.16629644753686001,0.2688958507577127,0.755211049311409,
 0.9182306306917877,0.6491746920770983,0.31414029958242184,0.8572397081216591,
 0.4466037686765103,0.6277684539187284,0.7248607273793338,0.9057211397552124,
 0.10411876139615583,0.19349566415963182,0.8503431525976365,
 0.6792310819435426,0.4176260730769352,0.21554731982335007,0.6459491381810073,
 0.1554362548627558,0.13608645874425362,0.5711675105880345,0.2609575532906072,
 0.7755126562238487,0.5670249088293181,0.15520810107336924,
 0.014826414761776707,0.771020589679959,0.2800386312483787,0.5560667041569806,
 0.8002467286960425,0.8716031669699906,0.6231586674968741,0.20287858597125963,
 0.636495015958326,0.8011680080021397,0.7912489250551623,0.7351530794114772,
 0.8998036410662205,0.36396692399184727,0.9825573399247376,0.8401544534459324,
 0.3884287955113034,0.24231756146731742,0.44946588087785266,
 0.24993921608114245,0.02333028418321237,0.30064487999491707,
 0.819937032374855,0.9855889830284309,0.39907150710333994,0.23030253111007704,
 0.19346443579982475,0.4896674415758303,0.5417866280001761,0.9246485739908007,
 0.5395491180221881,0.18890111248335828,0.9328881756513172,0.6050955283173736,
 0.28786707726918204,0.6686681822727419,0.9434165917125078,
 0.48726343771487246,0.8440211381273779,0.6883128877448443,
 0.10195487223559252,0.18372677280547256,0.28838429520131736,
 0.35699140920497574,0.09677959837097239,0.32871689097201284,
 0.8671204109291861,0.6314962229422191,0.9858981663516432,0.5562662132988238,
 0.87134277601481,0.23554114480107147,0.12908385314700377,0.29408583433352187,
 0.2389738894428015,0.4322012392659166,0.8178720315394503,0.7111606767488338,
 0.671944812481398,0.9723518739807789,0.36063444406278544,0.6119280142352571,
 0.25062913328077574,0.3477877137920584,0.2306351410338603,
 0.009760554966723123,0.7314829322455798,0.5630048479520674,
 0.2745187665548088,0.2936681376118775,0.4950575790544748,0.23547075508009618,
 0.9433733348209594,0.13316655654460752,0.9287257979216291,
 0.39018722430705544,0.5928203142895009,0.29897134737450237,
 0.2918465137588798,0.517966810676495,0.7787968771496285,0.48016752222537973,
 0.09725469087291949,0.34826174118082165,0.6095176489094073,
 0.3300156958379522,0.5697746348165725,0.4674088456363502,0.24587463183797875,
 0.9165333208239284,0.5534431784699967,0.8208463484031346,0.3700697153873411,
 0.4208460225128249,0.3323801334421277,0.35453717381811545,0.8900466436551904,
 0.8931785772336,0.14746727978229834,0.900608159998949,0.3121149607680843,
 0.437804429857068,0.2673807802161102,0.17267341933990743,0.9783570216330268,
 0.18444977869172186,0.9316929578292955,0.7446819654917015,0.715964547961378,
 0.7466117508796957,0.32667367440979667,0.9124659120827674,
 0.026927486470510797,0.8550408729067028,0.9532316624440836,
 0.08898093670649931,0.4601185456352779,0.5586516415225675,
 0.38949986824026617,0.3829226315309263,0.33772446748512663,
 0.6026201965046003,0.4693025669528561,0.7957306928659643,0.03058105374317366,
 0.7298893786536264,0.03699556724996589,0.09871914434641393,
 0.10417344131966433,0.8959204498336428,0.11223920010346855,
 0.05764873219287647,0.10791419693969939,0.39220371837800605,
 0.9989439468024129,0.14742532133177444,0.39056672138431026,
 0.15114019152712777,0.8661837160483251,0.45325877803340064,
 0.36543643109158064,0.32302055524080886,0.2756713530664442,
 0.6104748281932149,0.15423067821639902,0.1297887801787002,0.8464924058523055,
 0.897297598729498,0.9777513139379786,0.6816752980876312,0.26465782682824557,
 0.6365836621699339,0.5554936081873069,0.5105205866232533,0.46839763962174663,
 0.23018613787548126,0.7408037421527937,0.45748947777289817,
 0.9572952670056248,0.6970010009834593,0.6834058641100437,0.4692787782394139,
 0.3831904646410914,0.27276345128703505,0.4042819431968633,0.8062373451475116,
 0.5848424310137793,0.5355570136431551,0.14744045650022608,
 0.43246662475562925,0.6988175643819619,0.5306894581358773,0.917084820349191,
 0.3559852313539137,0.9575110736511342,0.45741043598646713,0.4370988103980541,
 0.8859232730251292,0.7712715887569973,0.6542307317925666,0.3746677220078469,
 0.1409175566328188,0.36393690696808945,0.2881789956584494,
 0.46547350082117056,0.7881274393479937,0.40565362583912545,
 0.04891575442200624,0.2571051958885644,0.20843514026185905,
 0.8291461992081451,0.383689242500602,0.27790121331977735,0.22419369482566986,
 0.6024743789000092,0.49152684701061844,0.4008640485916667,0.8933112026060989,
 0.7695614640008988,0.3916542625610223,0.9548969345518511,0.41502231042074955,
 0.606338158144697,0.7723492651975647,0.016078824106695477,0.5677357288404594,
 0.25906204643955877,0.9641456879863117,0.7729119837221778,0.5743714691541851,
 0.008502937864372062,0.15024855068192833,0.705801284369338,
 0.09180651789845429,0.08001957128688086,0.8437957095209779,0.948081749913091,
 0.02246682579337078,0.520922502948068,0.1289480764096952,0.5140168240333196,
 0.7710017876005738,0.3249271654546979,0.5259865486716555,0.6917097658242615,
 0.19277963464314896,0.6223950396119231,0.946304172087761,0.3412364301713844,
 0.5960585510507083,0.20811847876067924,0.01637167261981698,
 0.03739418547150519,0.013908040148672418,0.18955505312582832,
 0.1185646509804048,0.30982067829337456,0.3205311932351569,
 0.0036271407987754634,0.4551086102164188,0.3683375066352781,
 0.8379777539416751,0.7912886549049373,0.3888883049945434,0.7401903112657733,
 0.13320245753666393,0.6386102753981828,0.09689531471937463,
 0.5690041419978972,0.07246503047196118,0.002549993223508129,
 0.47674647623395017,0.8125827083268237,0.5022840739484083,0.9625076414685364,
 0.7617644587589231,0.622822674291989,0.7946742492928941,0.3546603452090513,
 0.414068450493029,0.5106248143182085,0.1516572728733817,0.6065745501762603,
 0.8793196976362148,0.6402671934043083,0.18720499219672393,0.61536316646535,
 0.6310225660869864,0.8614496298673869,0.4009031385549622,0.4749283779653365,
 0.9051540109401134,0.8904298562074049,0.8950656334844431,0.8284578227070026,
 0.4538598395322563,0.008617430459549436,0.16311491336575845,
 0.07161431864159751,0.14143941404244254,0.5222780383219696,0.230973915134425,
 0.5902154876778666,0.43279458719079256,0.9497320601853154,0.7893331992148022,
 0.2185098473307331,0.632037322086583,0.9202847441384823,0.005698487729225987,
 0.9144123906976742,0.6705550386813357,0.11600142566842842,
 0.41513840920160927,0.38163479881857487,0.445016825791007,
 0.09399985072599981,0.9315129519887231,0.27355443045662675,
 0.8461302097745901,0.3786030170904293,0.04056833189391962,
 0.22890256646750795,0.6666866773694794,0.40565430043828654,
 0.9812933661149558,0.4189913204897511,0.16901040060443062,0.7854872634105298,
 0.3187093453270402,0.4097205009787508,0.10068287621044891,0.6099331297836976,
 0.35895838710842565,0.8092811985203114,0.2973356510265658,
 0.30519797882306543,0.09783726949926974,0.28433017498354696,
 0.6732772366161391,0.9380188665329598,0.5526505458750087,0.7912791805235011,
 0.6572006521981517,0.1042850343602384,0.036604299350567304,
 0.5791322992495336,0.5363219906009847,0.2589645907330391,0.6474436271137138,
 0.745276918594981,8.837711441670448e-4,0.2903343102076895,
 0.10682001914662742,0.8329816196236735,0.3488167959817956,0.3727787294810214,
 0.5770599749949517,0.06008736203851395,0.9891693361879709,0.8812886751141024,
 0.5542777207145964,0.6964055413600622,0.43702420550931387,0.5348035300999696,
 0.21670544198474972,0.6746503805236819,0.12674974342930645,
 0.9499781813490035,0.011453779850158718,0.05879540802783456,
 0.20272015626704465,0.857913821737007,0.07875590556684076,
 0.027566062720129514,0.4511827354983371,0.8439059284141491,
 0.24822142842469685,0.8726181815925912,0.3455477727827565,
 0.06579139453702654,0.8174882173774033,0.4585068451532276,
 0.44788287898953827,0.15898994997799676,0.2859528319515089,
 0.8860734173651323,0.3171510778242631,0.33046842070973803,0.3559956143537347,
 0.7467316349515122,0.9100338335492129,0.4624871568884207,0.8420389471392529,
 0.19102862974756407,0.8490982958214738,0.13811348609339213,
 0.008822820806605569,0.7720265651683358,0.3904412637569056,
 0.4664637772979199,0.00633964120311703,0.05839187609014973,
 0.9136856648936648,0.8477241645005291,0.9492501395107243,
 0.026442087798775327,0.026429749897526778,0.7602299123850386,
 0.4080157990592548,0.4934512076705664,0.8469695334818182,
 0.048525715744729414,0.6120849289328311,0.906484866700616,0.9778532092321363,
 0.4981386580173972,0.7000706503625387,0.8267261601667143,0.5125729887467276,
 0.8376530222214055,0.2946859189098877,0.9407392417454429,0.1701654973093807,
 0.7917405490850798,0.8417029971294869,0.4681828279707183,0.3885092770244679,
 0.4066049043913047,0.6574432507524527,0.8333749896784354,0.6324701896084686,
 0.9300003705263291,0.236623505761685,0.6953388977611299,0.5815598841074836,
 0.3909307440110916,0.6757935389095918,0.22932961635736926,0.5396952926437568,
 0.3064083104599846,0.20744350030665681,0.8753466752023725,0.9770070264964917,
 0.09492933411304116,0.9287698307531208,0.10628010712644831,
 0.7603226017178809,0.08739579496908345,0.1399496008073534,
 0.10199185080761819,0.6153563271952086,0.30687204706565696,
 0.3247846945868209,0.5272878942429263,0.6631067299131854,0.15770923859824793,
 0.773655747746235,0.6005283929160716,0.7871866015475565,0.14592324968277115,
 0.5880086718448932,0.03336060512856087,0.05290226107457274,
 0.13048577357820945,0.36776561547169107,0.8837839280335041,
 0.9653826747537815,0.8910659876261205,0.8028181756517876,0.9070088982504638,
 0.9604340724747302,0.19216607367194039,0.4048480564922369,0.9546371389267799,
 0.9072691440065721,0.38914750536282283,0.5556791574467298,0.9665699782259776,
 0.34684485292862366,0.6353703444018927,0.46189164258659743,
 0.6586985983459637,0.7564169059034724,0.047840252819857376,
 0.32416485245230364,0.9644881538665117,0.9281329050285759,
 0.15225042798856048,0.9808518167016009,0.9581617020042661,0.2238127496380179,
 0.09335103812357703,0.8203563695547829,0.041006441939265725,
 0.7564719103609662,0.7163470896665867,0.6164564800428852,0.2055193836123813,
 0.6905485553542614,0.8159543731082162,0.962046368195414,0.2917478280757063,
 0.011742027242795494,0.7204191807060595,0.3216746998164669,
 0.030360917752631944,0.8543163799778596,0.08020656575504903,
 0.6593713628341515,0.21117082187937308,0.34005107188429895,
 0.31709488253538987,0.2357514693691749,0.3345185897018075,0.669122687414808,
 0.9957427630930693,0.37844899150197453,0.08814163161732491,
 0.8196153027910499,0.06271581630998779,0.4966335640168593,
 0.41412956157319813,0.5728378515507109,0.2719130645244032,0.5790284665651664,
 0.21886389091791936,0.7887384732395732,0.9127625763984213,0.7501164133974845,
 0.9538207379945433,0.11205727589360037,0.8306058856208549,0.4520955873305158,
 0.29439589272060984,0.3808426078175208,0.3882615300287625,
 0.08487046073919502,0.35556164446495475,0.14696785159846004,
 0.6646143562376545,0.23771573168455917,0.3567503993132295,
 0.06680722154031526,0.7014185456808418,0.017540000329716365,
 0.4791668441248923,0.35860731944997626,0.061366174309612886,
 0.3177242741237445,0.9953124272019728,0.8131238513531774,0.25773625934880395,
 0.37250384668924175,0.30317617355981186,0.5592112549078788,0.94359939558833,
 0.5191413291705891,0.48878805365570366,0.9534501504295568,
 0.015368085895269168,0.9276810285984732,0.9863448671831844,
 0.8083059269982011,0.3671266949425517,0.22652663169008413,0.7500827093148883,
 0.4601613753290167,0.18811969723269606,0.6392408500351304,0.9800436521841681,
 0.588813390867269,0.8051788458905453,0.10050088644603028,0.10012028149201235,
 0.7301747806630039,0.28783254750115794,0.9290123877055861,
 0.27595373082526164,0.8222018386408386,0.10677090293817582,
 0.6806497909550352,0.0625614118596608,0.7783405718476318,0.09746766173701582,
 0.699395585676057,0.13988023209462885,0.8952145220287483,0.908535219748291,
 0.7143685362853465,0.224979173297142,0.18165271688845297,0.48467250795400774,
 0.9617284052086189,0.61473409641434,0.8653941684458628,0.2082099072667899,
 0.00854959443227421,0.29790530539930626,0.2118260136605823,0.597057069332211,
 0.048605227334002965,0.9012013260311253,0.803719129360019,0.8192302128492189,
 0.5132047618444304,0.4114910305767028,0.5801644661648392,0.12273577017580095,
 0.8164169510381988,0.8778996055953391,0.10515906992884838,0.5568632058673404,
 0.026854684213974078,0.06389502541040082,0.45492631160532804,
 0.9492601483572332,0.25738532402736525,0.351785479357844,0.2352254778694004,
 0.6072516052284715,0.11171847908039512,0.45687845363418345,
 0.12714553694529673,0.7953340908383815,0.7279648990454014,0.7074250673787721,
 0.3215220222712736,0.27355566174002954,0.9743822696669084,0.9777548307323732,
 0.2672671561871305,0.14042521307679356,0.9352313335321503,
 0.25344430177570176,0.6843330362584958,0.2688148956773879,
 0.07087464016629597,0.6730143398969767,0.10364903608978349,
 0.8798700496082231,0.36916843671745825,0.06520421321586045,
 0.7942241319200596,0.7035112853802955,0.8218781144763995,0.32318099900383657,
 0.08778097812071461,0.6604249783522869,0.35385477926980513,
 0.4954945361620344,0.5194035635740204,0.9651680557992511,0.7958461745807871,
 0.22389847741635283,0.15286043086667966,0.19594916507283755,
 0.7724351189299283,0.5942245599432674,0.5008331747634914,0.6560403279174846,
 0.5443169909764745,0.5829968145293016,0.5081806447306124,0.6757176103694099,
 0.10521638606570016,0.8369493055952038,0.37197215961879515,0.959159780216201,
 0.3688554765415133,0.3129738138488902,0.20325397662837108,0.1512269554018495,
 0.7062764523807828,0.784253248297305,0.8317355198441023,0.9403770983989114,
 0.6203672434186399,0.2583395505470968,0.3964160414113449,0.30941195803840327,
 0.7698247829039405,0.9973335057573716,0.9261083674219184,0.9472495394925123,
 0.08531650814919156,0.192335750872902,0.05081737916011453,
 0.31784133281392535,0.02839067817034513,0.5360671329399116,0.959042434692785,
 0.059536703182239914,0.593363167352319,0.07836440692706947,
 0.18114463283075033,0.5986229779028563,0.030566143936330503,
 0.4658197552178398,0.4900218288449357,0.9054657518892506,0.5615103890357647,
 0.8875712755210059,0.9034784661829017,0.2952792315158109,0.35190529938174586,
 0.0017431219045226376,0.5817605915292948,0.6991796926662708,
 0.9537417095310238,0.34497451645395183,0.7749013424761907,0.659458345947088,
 0.592273338391208,0.416035971679812,0.775815341301743,0.7272397428861717,
 0.9860462184299028,0.02408592238098528,0.228595548995413,0.730378672094381,
 0.8838111450157885,0.1258865392277153,0.7203742726456653,0.939366472201368,
 0.5426152293907457,0.6100378319348174,0.8393061727083121,0.4729041366757648,
 0.14657292385295684,0.6431558543537437,0.020241543881778457,
 0.9729998341310985,0.8427131569193894,0.7316720370007198,0.4336070604929124,
 0.016055694882407434,0.6985362887229528,0.9781569012125029,
 0.5232027036013247,0.6773858093206617,0.9954417701570084,0.44711157153581604,
 0.6936642086985986,0.900598582813009,0.951315930193064,0.29983641478653067,
 0.8325413285760245,0.9040711907361622,0.4189225437852435,0.1220237836521727,
 0.7860457103561738,0.6329512463774133,0.9390704292183583,0.8544394458829527,
 0.2894077873676183,0.21117576546755568,0.6358887639819067,0.5007165616412733,
 0.9824835676068163,0.5591510567696675,0.99699262496909,0.9474999072021227,
 0.9482101495749237,0.07945830841794532,0.6431461100400496,
 0.046444940835787984,0.6739356670377793,0.8474538843960595,
 0.6350281437444121,0.7731426721713573,0.14148048927292434,
 0.45362710623119684,0.6517613468514964,0.24194335735522765,
 0.43058469141480515,0.8051124740632654,0.9272683672909946,
 0.22277769694421112,0.5792710757128986,0.5558664532968293,0.3368397503096101,
 0.03320905409717778,0.41718073011773116,0.3026883679502913,
 0.9542383240255619,0.7412638640677381,0.3127017412357449,0.09377362527606647,
 0.9311617331493979,0.22018967742397832,0.5125961480224821,
 0.32025248880931845,0.5974292669377861,0.44562471496482015,
 0.27993433709989624,0.8193932981566338,0.7365140192157806,
 0.29915357109782403,0.6860480126699922,0.671410124629415,0.5773537948035115,
 0.5580050551704637,0.7631357784218106,0.8086502273620374,0.5519879399290064,
 0.13009036677123076,0.1271966383831642,0.27784667956263587,
 0.3002965054445019,0.9262790677642708,0.5696452193142874,0.4417647367172791,
 0.6985145367843769,0.03825303257966606,0.9673164674260625,0.8957863370942156,
 0.2903438351003421,0.7992456579438274,0.870856742965804,0.931494872652942,
 0.08477064278233204,0.7745707244903797,0.9999266156831959,0.5916243767920164,
 0.9031981784123281,0.6481420156102895,0.7934238532222828,0.40858123918886735,
 0.9679375159721773,0.2009089599170496,0.9060389213654656,0.769645917365363,
 0.738805390862838,0.6001395848599075,0.21525060858487577,0.4309050280763844,
 0.40664587665647767,0.08920846129306659,0.5614448355487157,
 0.7017437191468975,0.47517385409653456,0.26663568861675724,
 0.4020124502078112,0.6585269994892715,0.021610822118739104,
 0.9197634848490123,0.65398196506591,0.9593154512959959,0.05214053239820715,
 0.8744257120196416,0.17265670043614956,0.9292010936587808,0.0840630695873581,
 0.522953443925319,0.6336816032889768,0.10286112374135192,0.8112135029734082,
 0.9174337398908634,0.9250684407393014,0.5605502035174609,0.6080216517622854,
 0.02455446505243375,0.25725152011046115,0.6324869765213231,
 0.5390802040684635,0.22587823817581,0.5304975862243965,0.22864925196509867,
 0.8087987678636372,0.49617199755035524,0.6772875443382393,0.0329199838540164,
 0.8573870200499395,0.08297201503874951,0.03582038041123958,
 0.20443529906415492,0.5298273989261647,0.44044590624185953,
 0.8362146403754334,0.41392457510649217,0.24700748451244559,0.867490668923352,
 0.8030614415637449,0.9635868624102234,0.5091952177285667,
 0.030198493957491968,0.312335828483283,0.664482306807384,0.9822997207790904,
 0.6312176911995678,0.24514208849843777,0.8822234923713631,
 0.015363677729367087,0.004018527729770849,0.8923971148180211,
 0.015660890260396854,0.9140332870379967,0.7715015205729385,
 0.39019986945426943,0.6744338161765022,0.1649687300732824,0.9249554688583425,
 0.007332274328087807,0.5499369118139297,0.33918621479549904,
 0.9401488717341792,0.36650836748331794,0.1766648970002016,0.6743721905524765,
 0.80985281972493,0.21531101368518113,0.3176833654664518,0.0797173209951958,
 0.7569472921743705,0.20184389273747338,0.9827614309775566,0.7610038308424787,
 0.9975599651268364,0.9979321758838539,0.8140113855493961,0.284360188462244,
 0.5296657301285369,0.4574144397259904,0.9834765786765527,0.8821601823873148,
 0.8207420685384909,0.7807161536686855,0.6552941584569558,0.3014572866138938,
 0.8459396477004233,0.9891594923878375,0.13170410092623586,
 0.26034077396962396,0.3651707374304717,0.12401614977872644,
 0.1162992436871344,0.972564960941153,0.6802119799570899,0.5826444648945406,
 0.5416791233703624,0.6984137358666743,0.5143385868984384,0.40713259560020165,
 0.9960703918461595,0.8187280209690262,0.22278294130006626,
 0.41972837950515574,0.3910163357803247,0.8369647848645071,0.8929417971560829,
 0.5015843089970586,0.6825573493298879,0.6715451608242897,0.6480656238276286,
 0.6606066392559942,0.78264368332104,0.3325253038869088,0.9956232805921917,
 0.891411575773986,0.7025276416434754,0.6375495599941552,0.3126311409933066,
 0.59185941710905,0.17023899928035624,0.05089198469320433,0.765486425764984,
 0.9603811863942444,0.9160822609929528,0.7988417240498804,0.2814893151186031,
 0.872815117222088,0.8472359357441409,0.4423719487980423,0.9841627400491588,
 0.3266544399537624,0.682490965855138,0.739759328117954,0.4713047569879255,
 0.2116808337788566,0.030174770964767506,0.18763925963327333,
 0.5669381482627548,0.3937038035652447,0.5956424237884544,0.557386390639094,
 0.513059705277066,0.9524924529077603,0.9340275381062901,0.05956862321678935,
 0.42626547941715554,0.9798242838826237,0.1634598744799155,0.7240591825190015,
 0.890823386993986,0.9781871354850814,0.08659488217176858,0.466577217528773,
 0.28697212600799404,0.9842234003639868,0.1994869108178442,0.9778712219131988,
 0.6434417433964276,0.124777933657222,0.5711325907244209,0.5290379378069399,
 0.30351350168310454,0.3422451702714122,0.6841801946833729,0.1309948644868335,
 0.18407879679457018,0.5182900692600352,0.24512987907594852,0.581048718820696,
 0.2032991170978985,0.9818025208261434,0.8853650527991084,0.2540151882900238,
 0.06600115266631446,0.66781054198673,0.8921358527523153,0.9670684023359621,
 0.4093228878924361,0.4672652947553442,0.5935446402998223,0.21374596381968025,
 0.7891824202557143,0.2852492709913075,0.1968314161130803,0.7269352917717165,
 0.9018172814736497,0.2932517282160291,0.9028359411988852,0.26162016566527024,
 0.16385478371129503,0.051603577647281185,0.03713751083455474,
 0.3265149568786657,0.7637951502107688,0.8348354352402143,0.35680824380278864,
 0.4110870442199286,0.6409905882854322,0.20826283055332606,0.7079270782878152,
 0.34708816623912364,0.4614880396634744,0.9897792605720324,
 0.45315848892687427,0.4103546409673179,0.4664077057775431,0.6546332016771096,
 0.13136145814898192,0.5493409140281198,0.5162379278154503,0.3625802434160589,
 0.5457609826968712,0.875373280913463,0.2189906378240577,0.7775020727611947,
 0.580864912788546,0.10778148987909753,0.3316172052199189,0.4745906160456217,
 0.3966292304309841,0.35464078291217727,0.5837848069283667,0.5442439984662137,
 0.13454499022001254,0.006959644707595136,0.2908725460158166,
 0.08166083354155507,0.7610769239104923,0.45584430026362466,
 0.38510839138030306,0.87578528446243,0.8628904489837598,0.35809345785857205,
 0.9881741686534777,0.5139485973017799,0.24704535789041449,0.3497305770843604,
 0.9991913591064858,0.5926263336139891,0.4159347298637762,0.5490190590935656,
 0.8551274005218595,0.9980196447147287,0.23610166243778008,0.3703292683335576,
 0.5237608917382648,0.7131203199010931,0.007791336363979218,
 0.13600363747119737,0.4450427671896806,0.6860265814760995,0.9648483128709598,
 0.751160462748517,0.2787886481931051,0.12247355804311255,0.13597025078585168,
 0.2867184874698552,0.9400627658140658,0.31374924185517217,0.8682442338059206,
 0.8147993136773317,0.24386522511647302,0.777194610935849,0.6576426750115769,
 0.8022334169477371,0.8745573402528427,0.4725672352966763,0.2269795305465261,
 0.17419624903316633,0.04470304961875138,0.5175750378856909,
 0.06850089485630417,0.44133868452051406,0.46499448222167916,
 0.7842679794294383,0.8628210171227889,0.34026884919080713,0.6976464329627382,
 0.32052657665030637,0.09613419454125793,0.16514856444345516,
 0.0151093924866319,0.25038042533803795,0.6759284400829564,
 0.21834289226330883,0.7259020166657429,0.9705098226307445,
 0.19077510752723514,0.21765359015539487,0.9486115816885997,0.126493969015123,
 0.05215517519778323,0.2228063606426619,0.394373041199791,
 0.025535604395821032,0.14108948117644182,0.3195346801265162,
 0.6693404824709761,0.8422428653363617,0.835348674112901,0.5519664431796837,
 0.009161917116919716,0.4465158931120259,0.09565757586959311,
 0.4042574281227378,0.38615824327612414,0.7677427078766863,
 0.34557811919516934,0.0615663400756119,0.2494277407681056,0.8812366397127915,
 0.6736376820227232,0.9546669483736216,0.5115820986279442,0.8156850670354204,
 0.7974472861533991,0.6049414794063803,0.7827284352675614,0.7578208679034243,
 0.921882204099961,0.20337346680058932,0.2809704610799013,0.7590130782045045,
 0.6537217911986547,0.9313484667722287,0.15682266635281972,
 0.021707263661289056,0.7324065627448971,0.9693150737028519,
 0.6435862581314894,0.5768223330177886,0.02617659553171159,0.91962970429892,
 0.17417806211114673,0.5811629645504679,0.4978049132477905,0.5361257691141588,
 0.08065664989193577,0.8735376136054357,0.010914941586305793,
 0.8971298139345389,0.018139159039834207,0.6041644565405474,
 0.8235409034102433,0.08134191112219014,0.8362272595860479,0.5100327932155804,
 0.22132444270701257,0.22583060888357143,0.7161248356466051,
 0.21086103360218234,0.9251442801653104,0.32645186132146176,
 0.9148825378561308,0.08821499727074467,0.4049534813042489,0.4618716785392091,
 0.43623932162992063,0.4658992246161785,0.22990210523546228,
 0.6117016200001759,0.766058573244692,0.34848691976749857,0.6856750772531528,
 0.9156411467609862,0.3943161093088171,0.0740941859086397,0.09298009357911952,
 0.7076398246208722,0.5762280919544984,0.5598057767861894,0.0356395274596617,
 0.050016030280197965,0.32136764107865945,0.7883248785999348,
 0.4665237256881545,0.5626510209818651,0.8785460128184466,0.816595673438015,
 0.146461029269496,0.3076300243053227,0.6111670022580404,0.8371753169026895,
 0.6079545819526124,0.6953852663064044,0.9935014677403808,0.5915032201091533,
 0.02979038699008074,0.42983798508318416,0.5547630309862317,
 0.5406469239369496,0.035420148942374996,0.9134275393560374,
 0.0868394943909917,0.42342557325141295,0.7269417366790801,0.6821013798107384,
 0.3408928637079417,0.6573124304501985,0.9334084595625985,0.45302820550740297,
 0.4584666643069114,0.498894478016478,0.4367813257066254,0.047153608177204376,
 0.9213102681435744,0.038135567361360256,0.466295062439845,0.4319737202264101,
 0.9480362083508725,0.026904273644649424,0.044295117772117476,
 0.4882718231546951,0.20156467037732095,0.32880641123292953,
 0.4870509440689633,0.42045753669427743,0.17005253007896481,
 0.9320157535369428,0.3452311250676079,0.33508054145232546,0.6488003310969159,
 0.7450916480834038,0.2820478883117671,0.6522404882727786,0.5216374961242363,
 0.8177854968054358,0.9281990886310956,0.6956465303730581,0.5940663670074154,
 0.7373911800966056,0.37336904660882797,0.30789280168919875,
 0.5735024259582377,0.9395611168331086,0.6945172818686337,0.6275765583084645,
 0.12974513912985108,0.4764786003264502,0.3499748313024543,
 0.05905228135811735,0.1404044662727323,0.1655937343365841,0.4008744453796368,
 0.5351874311805793,0.5498348637853421,0.21380030558977603,
 0.29712261331352974,0.8772762810593354,0.8124945061461305,0.458052916754374,
 0.4724035802935187,0.2779445368921376,0.7221749359967484,0.1771390875015716,
 0.5685890837357873,0.09833818913837489,0.29593219518841873,
 0.9316312093097556,0.7429588386294287,0.772007065248278,0.14972665881072555,
 0.5930955984017912,0.7668841758784223,0.8170701419092108,0.27027396593231434,
 0.17930873637505296,0.8307882391651965,0.13577020064807688,
 0.4659474749722442,0.23853613517566774,0.2855371441037081,0.8025513754972401,
 0.6875728122688006,0.7689637920914847,0.6340945813483991,0.9354105333915202,
 0.12978972634553887,0.18397746352619448,0.9100828186579841,
 0.8633898209212336,0.5452116749343118,0.3991899514630679,0.5324572341893783,
 0.26120279970019367,0.28177940798126955,0.8486570197678789,
 0.8614339551587693,0.4739714499309149,0.12052859541183247,0.0394913477619343,
 0.12138526653443527,0.039227602359013125,0.6164192988879195,
 0.7544072122325849,0.12273582267132288,0.778639912246497,0.950252916106505,
 0.6636912157195815,0.597981594854024,0.10563723654122092,0.646561040668161,
 0.1526105131275517,0.848704834298668,0.7899271045741886,
 0.0039268116538019315,0.21719980402426875,0.11982955951506935,
 0.4893334427601148,0.6477087139264486,0.8033713609760877,0.777175438235929,
 0.18776858859386614,0.26829499692816494,0.5694971856230788,
 0.7667236808396539,0.9601357896181727,0.4416361448750221,0.31033850623134085,
 0.6316363207075485,0.5378785160456917,0.8438086386440579,0.6438377942763985,
 0.6913464024500254,0.12055915116952876,0.9096812212880936,0.9158422369233262,
 0.7312386822094357,0.2867936226985951,0.7757893157808706,0.6735032398644328,
 0.31747488528616974,0.5397457204653159,0.3762503627498699,0.9727302476688635,
 0.6475521136032871,0.8616176606869979,0.7008176111062291,0.27065869980075785,
 0.539560967337045,0.7771546275681833,0.6516399394520035,0.9329205110865081,
 0.6043032102199162,0.7779915894675273,0.8186062112459107,
 0.0025656261233953925,0.6120095337033307,0.3380580824866125,
 0.6847855894694564,0.33870670038565076,0.7581633961844993,0.9792593012353895,
 0.8198168232202945,0.8501785959875123,0.8170793004625035,0.8632256108975349,
 0.015970523654844637,0.4255511380012088,0.18967183249943043,
 0.5220997130057037,0.6487296726126818,0.48108666401865063,0.8723679929782218,
 0.9531787799347563,0.43330106765604126,0.07056803858810912,
 0.3256530716065891,0.7137074541093806,0.9594611432860004,0.9992364042367693,
 0.8687169655106308,0.6768110607625901,0.7361196124692487,0.8493641242148965,
 0.602492551715571,0.039114370791441644,0.3236105147336792,0.1654362641539623,
 0.9950656657886757,0.9447632161519863,0.4854574678896115,0.4092342068289174,
 0.845621484988114,0.9198827023097518,0.9247398912359784,0.40627924748227306,
 0.47645730028396616,0.5330052051214672,0.08201932237604626,
 0.5931857580449688,0.16248031049166145,0.7167737155708057,
 0.05369711523454779,0.2109863802632267,0.5367304144941402,
 0.30437438901512603,0.9585184736478173,0.989501074370464,0.46944977870291615,
 0.6196965826164429,0.11961347517533549,0.6684051866948861,0.5148638629721998,
 0.3426996693174884,0.025288811766647545,0.015015508282092527,
 0.7214578523868834,0.5685009383179381,0.3228928681778447,0.47891173290396083,
 0.08243726346609848,0.219262756072223,0.9586634565075202,0.6575168265511718,
 0.36934409942820623,0.386348707105039,0.6274042636912404,0.35704850082711825,
 0.828233327237131,0.6488797061632632,0.20524593168619287,0.6019143038998834,
 0.9612975986465195,0.23457016703831002,0.9307531230474406,
 0.12013377132473857,0.409477731509975,0.7298592838775275,0.9444321201324722,
 0.32845596058611115,0.8455553565687257,0.8362330477384917,0.8468988380432093,
 0.40209663305591414,0.9243056962897095,0.47837664877839825,
 0.8239956244446593,0.5513299897584604,0.4553590056801403,0.29409214185072785,
 0.60231363832668,0.01272687414195095,0.5863586357718321,0.1315882732281659,
 0.3961625287035556,0.9186012151287639,0.035055593067872426,
 0.015982486547967523,0.3093122488859369,0.5710468433854706,
 0.3617984198211248,0.7304553277581283,0.516408392753544,0.8770334137031807,
 0.8666399871272447,0.5882873261355279,0.8757141503096945,0.9624309359792749,
 0.16016964630380048,0.9130097617123107,0.8855550690175762,0.9343284804052875,
 0.12839501502912232,0.4641478329968076,0.07508762123591473,
 0.4589427128306488,0.41575441708477767,0.08245881826929358,0.712326229773125,
 0.31352377655198393,0.6685520300621839,0.14999920957153223,
 0.5577793565124125,0.6674207771824372,0.3326735833082701,0.5576644303007315,
 0.3246185812234177,0.17244526253088943,0.5655585108579818,0.1126791549875451,
 0.7181960057229376,0.38333409817899433,0.9090120537429478,0.93571762333861,
 0.8209614950860934,0.19362290418587103,0.7590576162369398,0.1821417028282728,
 0.5712946610292301,0.5011544997563868,0.03850795304954402,0.8348517728881936,
 0.851273052824217,0.695976152308629,0.9242138891784452,0.17644549469829118,
 0.6120215958051995,0.7308345135531467,0.19793131484572157,0.759996853618595,
 0.9278124821851544,0.9023403818620326,0.06380509279741586,0.7635761699494202,
 0.5213024722977416,0.693397056740169,0.8370549062107239,0.8313039021235203,
 0.05122018159741737,0.2913007830521721,0.8936498925617221,0.7839629657146892,
 0.523435152444335,0.15179369132475473,0.9534324444430724,0.7018395090242019,
 0.07660374198000097,0.7400874251511453,0.7919035553853938,0.5745483526861628,
 0.5795020095654415,0.07644561026704966,0.8302169137993085,0.4697353930987719,
 0.9781079884862396,0.3468995196104592,0.7415811217580146,0.7011791441757813,
 0.8552972186368595,0.07545089544441308,0.5218568565726706,0.5252989071434038,
 0.320042052818447,0.7942274578029791,0.5170591828200297,0.9650351496598515,
 0.8502991763154036,0.37326043577026713,0.5559982887647869,0.768149213231759,
 0.06918202216157576,0.37681646638563626,0.9984123524600884,0.569794726495282,
 0.6839876993380352,0.03321410290072202,0.26064235409272485,
 0.29528670818543734,0.5555938924762274,0.462066237140886,0.7389669185069454,
 0.223053052243257,0.05944981858173304,0.21905857795115924,0.8052363117027755,
 0.46121477446172887,0.033651225884693936,0.554610912500425,
 0.06928413483806262,0.9125630976928539,0.24053931182319976,
 0.19489564511283575,0.41769307274080314,0.9382553545402668,
 0.7382706342102764,0.25624122932492166,0.595597987243005,0.8014700215331224,
 0.9447464622136729,0.5000760705141025,0.3536759509166081,0.14117849592500198,
 0.12347314016433897,0.21419962048645802,0.999013215574122,0.7677533705662141,
 0.7559566589015957,0.93123024373786,0.5083146538220776,0.7257725565765256,
 0.5792194322337243,0.5601273866237181,0.46137713032001504,0.7803332698980849,
 0.7246818007681803,0.8267585545145656,0.42243460041824465,0.3600935042444393,
 0.5668520709531603,0.4059144566176083,0.8259307457640438,0.3366765556659447,
 0.9273675072346698,0.6268529521327544,0.21025785974342404,0.9936428236534209,
 0.8339824605982705,0.6368365567402314,0.07491151206089841,0.5840348760504019,
 0.061470653199131364,0.4621673271510821,0.5380344829181456,
 0.5759353327163689,0.5295492172058198,0.43769702718136105,0.9506854881806595,
 0.7946197021222621,0.8967112998585103,0.32048684739888056,
 0.32749499826095896,0.9268154349738139,0.6009757340125128,0.4711949516090286,
 0.5347304533821923,0.140231291017188,0.003599046449500709,0.7903063781999342,
 0.5949259021899306,0.3654670199533868,0.29728226414599346,
 0.31198189434537216,0.9511790008695348,0.3962765309214926,0.9754170155213855,
 0.22023204430644128,0.8318831859630608,0.3028572870608144,0.6807870848370466,
 0.771886535338482,0.8817702841521476,0.23779841744323504,0.9991190605450909,
 0.537099007215025,0.6837892791382967,0.9408516704835816,0.21387972331317373,
 0.1181887126673501,0.8305670245962176,0.19135939367894328,0.5065831957607774,
 0.12701188410921016,0.05629720340408384,0.5037635679346839,
 0.2704657310716718,0.1654464383408194,0.3753968523895732,0.04932589189423153,
 0.7460172195094341,0.6143041258979902,0.37824786526878484,0.7999610037945392,
 0.9428354497212761,0.8378348178527075,0.42701781692996343,0.9008705652442677,
 0.06885044248383654,0.27217364487049034,0.49242568817134846,
 0.8896663897835482,0.5552905752702683,0.32021851096282117,0.4432814242463934,
 0.403394793684948,0.9604153595587986,0.90348032024496,0.852603832285824,
 0.30671900902235394,0.9556658402449785,0.7721941983198599,
 0.031459084070896925,0.1151758070081188,0.13790747286335714,
 0.11155388949400002,0.8211314023821769,0.19482822474622663,
 0.9996383856721792,0.6446140109058718,0.9694984894164207,0.5798307475288319,
 0.06895307008257612,0.09294975292034713,0.14555602457351824,
 0.3583438809665406,0.1362675827422838,0.4204798513697974,0.6689187413625384,
 0.3267014966333187,0.001668147401954423,0.898692165137209,
 0.015168903654411259,0.6177522555532743,0.7001023625261932,
 0.13872740924032145,0.2823002894621145,0.9825256462433309,
 0.24632957167223024,0.45602577952600154,0.3555204690689959,
 0.22174088092782362,0.2514293668495281,0.15127573732855093,
 0.06329568343789993,0.7815458877126682,0.6328000762472191,0.4718053327764976,
 0.7749502625184375,0.8970182259586619,0.14994834106653243,
 0.40084510119413996,0.1872965039906499,0.1972328244678827,0.8838681255015396,
 0.013248942599948466,0.2780315776138106,0.8916602853677923,
 0.0205167411730246,0.3870564845032962,0.7199646358788531,0.0679117823155222,
 0.07916361168895225,0.8448514481329117,0.30282990574786073,0.872924938646999,
 0.7664325322011812,0.49702018141097004,0.028827781780223383,
 0.11059963657887495,0.215449506974269,0.057061083433225135,
 0.6707159400780002,0.9376023820150501,0.7275372887653543,0.22388694043578805,
 0.03874967243641492,0.8290230854468216,0.28134243848240315,
 0.38151835238861365,0.183887495282837,0.5922356868396792,0.9185935648937895,
 0.17726723418031387,0.7724659572806117,0.832505253386145,0.2497096531399462,
 0.8750657540425888,0.28496804935518993,0.40330798607766494,
 0.6693227659117814,0.5373193892599825,0.5116797028164237,0.9661070600160772,
 0.5716942945418968,0.0789866295742242,0.9997130095323047,0.3909920153296663,
 0.7977791894434171,0.022166048176607833,0.43163733479529987,
 0.5440425652885895,0.740254817664724,0.6946790748341625,0.16032543673103705,
 0.6406927991663078,0.8698772809437104,0.14063440239429192,0.3493005367475771,
 0.6292214518691563,0.7472945624035788,0.8696922416474595,0.9940648913518508,
 0.9363349953333435,0.9609408832763714,0.8472250368975053,0.08716608881845289,
 0.03745699576640338,0.3380155511745009,0.7250831930109809,0.6279130294317833,
 0.7037261551132437,0.6223465499277795,0.6432873884847372,0.22197017381654816,
 0.3359617624197908,0.4104250482250533,0.9536129145970147,0.7727768378837773,
 0.29310806726806604,0.607465234319993,0.35540194536648817,0.364774978013216,
 0.8647847127420343,0.8558705192568422,0.4004918116986351,0.39960762528345295,
 0.534005851648991,0.7878166040969261,0.8047810592554634,0.9569914039856493,
 0.6461372263208313,0.5218094695521877,0.8351586363679213,0.5056750868506171,
 0.47678498744577646,0.06519831553841904,0.6129093957228,0.28825109511445235,
 0.579072003376141,0.7200578472795401,0.2953086923842648,0.056178710306487134,
 0.8859287117001914,0.996036199744998,0.6661201441506571,0.52607964847817,
 0.3477930354888157,0.6082726481494313,0.4864930230444484,0.3049422751261983,
 0.16533650529288335,0.4770567929589129,0.4805322463121148,0.9351439262795629,
 0.9011665935650958,0.7972408005536834,0.8716654620993056,0.24290358660485067,
 0.42752238933147946,0.989094851553817,0.830632311408136,0.4471623849439238,
 0.4820050170435147,0.5903895122552998,0.852234040278704,0.17664315368952,
 0.5326871336339364,0.6774852034221501,0.4608294241878109,0.48804366923902065,
 0.9629983576148937,0.7048146532663366,0.683234105603274,0.1461947292789011,
 0.39159400591948357,0.9329757492499321,0.2593908210009086,0.1804012892048843,
 0.47944129995187457,0.730134953259978,0.9570877833038876,0.5559672683844303,
 0.44992114907176095,0.5111703149210834,0.3179651297940391,0.8161293156739228,
 0.8949541498723733,0.3300858238407872,0.9027137144592823,0.5435467939662438,
 0.8329495852295679,0.12664077963839282,0.9703732999217022,0.6857763330138242,
 0.15075589581177118,0.5417270058317107,0.2555432228117732,
 0.10245976577180527,0.9015738370830425,0.020326785973064143,
 0.31570461991847965,0.2075740520777607,0.7616092001389283,
 0.10652074738820994,0.7648875072021218,0.009868039299277598,
 0.0606576261141093,0.4661827357913978,0.013390083777243267,
 0.5858057814134793,0.8226743767344679,0.7781203890704984,0.04566157065999865,
 0.3030574619598094,0.9019081124764038,0.244893419218135,0.6098842450535864,
 0.4441477003098966,0.35760902232674585,0.6432001443221909,
 0.17221312488057383,0.8331193581987377,0.6400138095750498,
 0.06669776925535931,0.6807623826138854,0.45832554464657027,
 0.40529044423277105,0.5661755184604702,0.7727857990809941,0.9482370492692349,
 0.880622722901176,0.616632869339404,0.2640432613511816,0.46560777992292335,
 0.2577632088878795,0.3899425261505236,0.6618356597808849,0.6047558291715638,
 0.14342018850930593,0.0025484519976941566,0.9644251504372455,
 0.5926767192902958,0.5067990447801696,0.9854868098073984,0.4248048378827576,
 0.44720633991306225,0.977499703109729,0.33946075131882647,
 0.11643670736065881,0.45700410313870565,0.6552814490251571,
 0.2881503538308756,0.7114187758210364,0.02233547652937229,0.8366173192579268,
 0.25187155514962933,0.8763021937363444,0.5645685709922482,0.267198156380707,
 0.5901194442096711,0.5463630297241144,0.09697293995563394,
 0.17266595627699122,0.881368211098297,0.01920273419312224,0.3352453993945257,
 0.2132146897309115,0.8426610311660517,0.05093280224373831,0.8689097662757909,
 0.6329837576602524,0.4620409290208397,0.20095758792642016,
 0.0036553000194754492,0.6978790367448614,0.7574027832814136,0.42194245691597,
 0.4439028638221083,0.4414736485219629,0.4245735004244082,0.8735361421774218,
 0.6806387206802139,0.6891831418958367,0.5932641467868952,0.91136534299736,
 0.8049400514846063,0.3718326579256199,0.590877529144306,0.39536808466944606,
 0.09733810017376532,0.803762863639528,0.33401863421482814,0.9047598696572056,
 0.6547073066934168,0.4679316567826731,0.46335226650613837,
 0.31113769445179507,0.6431373470215729,0.8476750885374502,0.7187983391140556,
 0.2153683855111045,0.44092378227968376,0.46426232270361534,
 0.6340071176625478,0.46209072089623493,0.6662240863362913,0.8218027509881186,
 0.8386296719771434,0.5040870759010129,0.7189748154150575,0.8009116443876947,
 0.9783029668312588,0.8734499717373265,0.43384135643812705,0.5579366089685525,
 0.3532208790151101,0.580536857200988,0.9007208253135729,0.9027224022534664,
 0.45190724024549134,0.13148152150139847,0.32627477032820684,
 0.22537726225498733,0.7725159386360433,0.9134411208372681,0.9557223573982868,
 0.07655875327365624,0.5456784902301136,0.43728112259981156,
 0.12298455485289472,0.1563090447306602,0.17575034975626302,
 0.6401633499650916,0.9163680335200471,0.4499499801401874,0.12498751805629005,
 2.7038089067921156e-4,0.13092000482521904,0.11405776705329496,
 0.9426942579718705,0.8893646933224021,0.45943761451776355,0.7730941900547853,
 0.531660597884005,0.6822567289779456,0.7333181701580306,0.6422410844491252,
 0.4209946069354643,0.7816327272895595,0.43682410526945226,0.7930260306474303,
 0.8296297125545041,0.6350580383552845,0.20724162926222633,0.607540127676069,
 0.6499095830725058,0.45310405694557143,0.18830893659893044,
 0.7978723052101571,0.41401455807966103,0.6885238293501065,
 0.0035380654218595264,0.5030670413689843,0.11697964462765686,
 0.6840386132152984,0.21300902665144772,0.10235449902747262,
 0.7226864463344824,0.43201866986363224,0.05072330407980896,
 0.8507318661167061,0.47062694997452326,0.4521283787380368,0.335728670654748,
 0.5193249849619394,0.6649774662870644,0.43321377144577866,0.6059679697585323,
 0.6952184598879969,0.2668057271530043,0.7956499147703215,0.6143359377245616,
 0.37519997227399404,0.691919053361746,0.10368887649227465,0.5378036052958137,
 0.028845308389353064,0.6929517300067357,0.8343421307844485,
 0.10747921445732311,0.19592122183513894,0.6553620261752803,
 0.4623493334311378,0.9633795584493077,0.07832806694097194,0.5025277347814736,
 0.5699427377260435,0.08522122989228376,0.7144360168343367,0.4202475211116191,
 0.7680084834004443,0.6390765558328546,0.4897015369101384,0.9689654056498113,
 0.43243909601333275,0.7666568264932262,0.5325120975456835,0.6197345227193399,
 0.27073554825667867,0.8756991979994962,0.09699505350701998,
 0.5445642655239205,0.39459429352156994,0.7674644387090813,0.3193489580805289,
 0.9396613858930438,0.5174092178572147,0.5162620290027896,0.8497246157691472,
 0.517353068202072,0.588925037012072,0.6576946832831008,0.33914347908365916,
 0.336266664574565,0.21030202944607645,0.5962359896876962,0.8069054615637076,
 0.6391292330620936,0.9153532869476395,0.20561804961268482,0.9814722188315049,
 0.9789551445976357,0.452516531085733,0.6174785430358941,0.6784953086673101,
 0.8256035670341061,0.8276931574737161,0.543573945556148,0.43673779674405255,
 0.035900919900481054,0.7975228794835824,0.039926708664717814,
 0.7855162011947332,0.4787012999849669,0.31626484155594126,0.3138066769577106,
 0.3913914642426711,0.7194215441323485,0.12934019887933812,0.7250669572597888,
 0.24304426936000778,0.23623589175683457,0.15884067519307565,
 0.8622158895800873,0.23816342967638104,0.8678678391290251,0.7386019962378504,
 0.18268612178201593,0.6620066555652574,0.7504039389994467,0.2640974635162825,
 0.3729672047092122,0.26102037666920497,0.7163157198121413,0.5677988658488184,
 0.2974971776523909,0.7793427459344593,0.006166899685190463,
 0.8779675385015004,0.5412183541687868,0.9643295993631823,0.14692071838030685,
 0.026687824453767606,0.20727098713772651,0.6403448908347575,
 0.9330727339311609,0.29845401819085726,0.27601785232323295,
 0.9916504094442211,0.6818843948414725,0.9643164344334604,0.4388182497662104,
 0.576347346619794,0.9588795149442595,0.24487599664903437,
 0.047120489654555064,0.6134278691220356,0.6424594221895729,
 0.4375772513119319,0.36070388788585217,0.7848290716002628,
 0.059305334087880635,0.7578856279398641,0.23997218743946092,
 0.3517495537387735,0.6575922838005058,0.03118683154180557,0.9266625142405069,
 0.2379384621135605,0.4770927071698482,0.6708756255700091,0.8411345398576243,
 0.9970704697500224,0.9456752947754078,0.6793048653371551,0.9384808516293414,
 0.8846545905006281,0.10724584265272519,0.8004947086866278,0.3853249385464186,
 0.42364408069968906,0.8077478304241045,0.15579759172683594,
 0.7311390089848477,0.9708168721179682,0.14801053112258933,0.2644018201044246,
 0.035052757297055925,0.9812601738457444,0.9569327366903103,
 0.5208518808751845,0.6297090199485533,0.629543833307926,0.1521398490177811,
 0.5524218187022202,0.6652441064554188,0.1944409476012352,0.43969916659716723,
 0.6053079707865023,0.6349977682189746,0.2127061094751611,0.9499723859772004,
 0.17691066395508281,0.8203807434974051,0.16044141675829693,
 0.5767184039528246,0.2801818646066656,0.5359165374276346,0.35071455039415456,
 0.3578615990327422,0.34158330217059074,0.5680670987899308,
 0.36598850439991215,0.5216961189743362,0.018459405356310188,
 0.5217906275203121,0.9830185831910497,0.6989699566363272,0.21266553664823706,
 0.6049603918511068,0.28864597067146625,0.7680565417247592,
 0.14020222934384607,0.4884478684847614,0.5800740891513745,
 0.19064388229148466,0.42559415051301197,0.2909191387649933,
 0.9093008398416064,0.48459532231164437,0.39177913586580804,
 0.3038851829812341,0.9741807398128972,0.3796171544654292,0.9650324480006794,
 0.6540968056096983,0.5272440570255343,0.5381079617716192,0.39408398963477975,
 0.48955658498322385,0.46900069111283205,0.21724673163334596,
 0.5146379100508878,0.23299556038620217,0.6443577565027446,0.3071824043548834,
 0.4416382346536607,0.3812379836015768,0.37762837144607797,0.9140002166866721,
 0.905501603856067,0.6472107267891289,0.36816671976759063,0.763532247727152,
 0.03891753179828816,0.16425620861528645,0.2489491312857709,
 0.6422673699820652,0.8448412430204872,0.007010925041965543,
 0.8018882625337125,0.06836092940204486,0.9713768692448185,
 0.017077003239494015,0.04499628338766737,0.15864258268176978,
 0.17889465036887386,0.3150695729359102,0.7833298192811641,
 0.14007602731475288,0.33747168163342633,0.9973135622989802,
 0.11352276696659458,0.4669872394548278,0.26638273971291515,
 0.07592156924528526,0.2943819756128915,0.785813803261343,0.9204363218597331,
 0.7462461095701935,0.4645200271904103,0.5115296275783249,0.15300847281701024,
 0.7173880817361853,0.2793404588618459,0.24884886819368457,0.7965636707197388,
 0.49222002234899975,0.49120020043733503,0.9621817333482947,
 0.4716231601020836,0.745171174942507,0.47443400946930925,0.28296360345886307,
 0.602301613548049,0.3269707047538851,0.8174074151468393,0.2373765514341284,
 0.3749529266092728,0.3041550917949436,0.8384955142072323,0.04207538840335745,
 0.5325128673319802,0.2472231647419647,0.27229853876711463,
 0.16893722530419708,0.18902580057965934,0.45047559273159765,
 0.18582855451746172,0.6050856367965887,0.6603804919954618,0.4092617780489841,
 0.8527046710122587,0.5598519538263245,0.6837573914275915,0.6313119762015368,
 0.1536315248792216,0.3198653720790954,0.40620682710646283,0.3871430521136111,
 0.789974628776933,0.3168364603098136,0.06352398883331811,0.8269593076248482,
 0.6968466598417224,0.8444405134947272,0.06400174691786575,
 0.41470870250010416,0.9745392103204469,0.5503394197107784,0.919695597089355,
 0.6935097882424033,0.4741806953668286,0.8666160979580196,0.5445955899391199,
 0.24083978833606368,0.02195405481576662,0.4775867489159531,
 0.4447170351957416,0.44051360272484263,0.7093175393145156,
 0.19472448570509882,0.9935105917761001,0.9003548860386141,
 0.06177677381864921,0.24426090181029558,0.7205422351452532,
 0.6477043492591861,0.6370452006987215,0.1767235393428498,0.22905766922955262,
 0.8723665262893668,0.26474913249413234,0.936773565372488,0.8092592923103934,
 0.9285850236466632,0.3636026943994507,0.7086000586562214,0.05742822265313241,
 0.806182596750869,0.17079847495563172,0.37922545666248053,0.9088663132779622,
 0.2471818491355947,0.9130157842140295,0.4525926301641796,0.31933034021826634,
 0.7477676219977791,0.5388235138676923,0.13297562613626057,0.5938131694995016,
 0.5376873734805964,0.023954546465609794,0.8653372408295533,
 0.8871657417624516,0.2735974014527842,0.6858915269470466,0.8427361892202886,
 0.6278965422860606,0.008015842858012823,0.576056277122091,
 0.12696497597367096,0.2377392066119105,0.5544054447412788,0.6556078791007731,
 0.2675078306861881,0.8477054328114068,0.31640475652804123,0.8625602415347207,
 0.6312287349865606,0.932252824163655,0.2609683017903254,0.7532033652958047,
 0.9513210618077581,0.4513122691907798,0.013982100053286373,
 0.7565065633206217,0.7779077397263472,0.0020028882078790833,
 0.6876869698251946,0.23982737416891864,0.27993992236374665,
 0.9109306226815801,0.726767613209699,0.21562634157571758,0.5173325730164295,
 0.41562024429464917,0.3553444818949112,0.8396190138202697,0.9817719743328218,
 0.7195952648745692,0.8778283764014329,0.03785867828645317,
 0.15904281234787288,0.8334998748504395,0.9987540835686609,0.7602296607947441,
 0.85110896600768,0.8470071180560254,0.04636069314333102,0.16810078775733195,
 0.7592195978150316,0.4028807277051305,0.9196462746109278,0.5299545368118014,
 0.39315079760534277,0.1432143013936954,0.016534638088287146,
 0.582149153688191,0.06661530311522401,0.027141066697888627,
 0.14921553694819234,0.6218064952346486,0.3257303535607785,
 0.17228664108454717,0.6253494738965153,0.8417513372646582,0.7247507316174308,
 0.3954915288834602,0.730045581166847,0.5057781035481902,0.12982232252217596,
 0.8159221133732435,0.55285177596462,0.49312202922434467,0.3310248541674201,
 0.4907754751968768,0.9085460656458835,0.5705373469782317,0.24100559011260492,
 0.24821924745011292,0.26161967094777794,0.8608787362724087,0.785300857984917,
 0.09148970995523575,0.8005559805875152,0.7708922756213241,0.2886854796822811,
 0.9011903599458952,0.6805125965030724,0.8994843759820212,0.28186699031023654,
 0.6972965241585736,0.5273272447986848,0.329871065759193,0.006034445656776466,
 0.07829388343454935,0.38395216483572103,0.7112464107874118,
 0.11471377804036142,0.5619114834494059,0.8982938108307447,0.5275715850093884,
 0.7952993963437794,0.3749965070972612,0.5381514875698981,0.5056052846129172,
 0.28242485383193416,0.22003326958140124,0.798119764872962,
 0.015223092322235066,0.9836447541169799,0.07287723563711157,
 0.9147933179100968,0.37553046014979197,0.9401015866827096,0.6969231463350923,
 0.2772693237662951,0.18815444145058735,0.706964542517383,0.38068059943348853,
 0.6009243506417946,0.666847464652113,0.6326493554215923,0.6404101511370922,
 0.568761112185111,0.3412231380329137,0.10262249965022718,0.664625210669058,
 0.4323049143527691,0.8836908418529174,0.44909896566059215,
 0.37507838642440683,0.43939452654338806,0.39634220294643985,
 0.5067640042218268,0.9005804308852634,0.29581877961674563,0.6470159772654028,
 0.8601856954136443,0.6444300695674816,0.2447333133970213,0.8382852501187139,
 0.5428644599173307,0.5143293171923538,0.3506911580665555,0.24729623164418535,
 0.3946619020701392,0.9066809790033842,0.7195470265749311,0.7475014179202252,
 0.6826627560182378,0.9944577632552225,0.7364568322684398,0.5367896757124814,
 0.4576980897496944,0.1150221923983149,0.2640394170824564,0.7645845492914883,
 0.2680204238361251,0.03649814964743425,0.13914633550270583,
 0.7545355392547531,0.6464550929183238,0.6476468643691562,
 0.015022202698223741,0.7317462359962537,0.4316442888580585,
 0.9666431832419681,0.9916934335805015,0.10969884598711088,
 0.15551039233491837,0.5985922588773402,0.9218766046816608,
 0.22753449179122665,0.8391777768481339,0.6069632084321298,0.7123767584751839,
 0.7685011162206847,0.9996689072650212,0.31312180900786435,
 0.37662487912392706,0.3949502359536271,0.6872337976489251,0.4421004305621523,
 0.9570110720660335,0.5852836421104481,0.3936832587507222,0.5245297003972561,
 0.8145724892564874,0.8399282253509734,0.5254530717317609,0.42456562913340246,
 0.6708251113478979,0.16376187047817292,0.3764598570943398,
 0.10827107074052444,0.6226712053798727,0.23797338267386026,0.564124528667484,
 0.14563703485288615,0.19715414258592767,0.1523097700699505,
 0.2550994110841358,0.024374236968609964,0.6240356485412363,
 0.7235271048812113,0.07315947829971481,0.32502629786632575,
 0.31302815905050985,0.055881213488308545,0.5635821605622793,
 0.3912857155890006,0.8524078362620082,0.09841557194604356,
 0.28467173760519393,0.2420985473581928,0.6850728711734305,
 0.42093073004240056,0.5389677768900699,0.08088697680239099,
 0.04994738912387697,0.49809404076974984,0.0066063798307901545,
 0.17191112076875603,0.004973470015026349,0.27976372273470584,
 0.13902288001764274,0.15588628630769907,0.7528443161501641,
 0.0674306781338142,0.9686578808830506,0.23069588143041297,0.2811229933047542,
 0.12488081248216487,0.06675583532953211,0.7041577602210002,
 0.2012226235586787,0.02095335329889747,0.05559218931116183,
 0.6777026300551661,0.975775731644612,0.7426721736296978,0.9050122644296257,
 0.32935615290467746,0.6533126669642434,0.6223721382814625,
 0.11960851396364014,0.0670805053508915,0.8177636796226884,0.7820018326587131,
 0.38816489567912393,0.48433177650560577,0.36773348281245055,
 0.14116225221610645,0.6419800539744449,0.6052038343079512,0.7967930486333883,
 0.6233580851233046,0.9433707194611944,0.03352736830109393,0.6768187390471812,
 0.1397823559304553,0.061456471353906084,0.2409191239996472,
 0.5327187336683123,0.16323051994483517,0.37690394005039307,
 0.9422778709453492,0.7075860366260657,0.8922312679610805,0.4054063836242322,
 0.4813922588306506,0.642027725890009,0.42851783468308147,0.8537121890492999,
 0.6061373375847786,0.33337767952949326,0.16236576306733608,
 0.4247160461299485,0.9288502122034643,0.5588521920165759,0.9689207642695553,
 0.6654492303575454,0.5890922151633605,0.1139411423907164,0.8754381499676055,
 0.08265082003587765,0.5683175803319627,0.0012365243645782176,
 0.010271320542484252,0.8963712722230657,0.7488031451870969,
 0.3522647358777503,0.9844296887647552,0.5158619025029569,0.05111612363219675,
 0.3053700856329127,0.316594756519502,0.5072888783820273,0.17093509910321303,
 0.9597961845444933,0.3745414220476817,0.7548640112760536,
 0.051177180174883485,0.8746928128121758,0.7346526307009231,0.642482887768251,
 0.43997569054875896,0.684270799232037,0.0852768605931784,0.21959051669214658,
 0.4943398303019707,0.9602897754314378,0.6240042437152895,0.6887875831883069,
 0.4218385652112425,0.9521988622922894,0.51135009542768,0.3243134670319665,
 0.49022706530744464,0.6640367963801764,0.6899247387015475,0.5965808532909949,
 0.915423822976194,0.8263166442515639,0.5549838079186078,0.5851799123492463,
 0.9411320625561156,0.4668213440001764,0.6529162180234005,0.6986923590300164,
 0.18465871621389784,0.46610534999739917,0.42684620288438246,
 0.2625837434091114,0.7451899409450292,0.6796510062459531,0.3666168797092819,
 0.5299521544198574,0.29416576170551867,0.3015827736794312,
 0.04387967353962696,0.7551011798509066,0.9620032960832514,
 0.24379568007464092,0.31305720306347684,0.720832046444783,0.3054969700956536,
 0.58094594345321,0.0762729519191544,0.14771904462150065,0.46265095268603085,
 0.47126055477205375,0.2965869003978573,0.14402708463585412,
 0.6568013031941906,0.21644170427725384,0.34823934790469546,
 0.7865068299576075,0.3050778100271524,0.8890590309993416,0.16185583987499785,
 0.2077520943956993,0.243034399515498,0.08515166517134576,0.5665255228030454,
 0.4929132022019953,0.7125296290073023,0.02719823744561245,0.8924689172593042,
 0.9934285017621934,0.592193101787684,0.4254317410228531,0.5589456679897552,
 0.7571348007527985,0.698372125242461,0.44848874406666706,0.854076923149832,
 0.36423469177789003,0.8718672603951234,0.6940341773549698,0.8299882265933062,
 0.6025111472845861,0.48576694787522223,0.9618849807781642,
 0.12801003557608082,0.3984437951259664,0.8892770467572344,0.6267043942775448,
 0.12375277766125303,0.817125793782094,0.21890930345042237,0.7048518374648326,
 0.03394042220649918,0.252250432941459,0.6526591423384983,0.6707334445889926,
 0.2358831999187032,0.1764545006627921,0.0823588538364538,0.6845078878535318,
 0.0023383604326254748,0.5702790015195438,0.15511552610953316,
 0.4091232687789521,0.21709671586031465,0.6616251002611827,0.3625000954715665,
 0.3109658264171479,0.37431447874460744,0.8814016404367524,
 0.06893622739672367,0.16327553854458066,0.9544867090472221,
 0.46520962399009713,0.43387695297375617,0.937907426238901,0.6556773176190801,
 0.8048473117133756,0.7138184614968978,0.21459767797151863,0.7839249223926794,
 0.7367587008596764,0.4226152492239249,0.39549232653013,0.4421623908041944,
 0.8435584730082496,0.28665638392455184,0.3021872721625165,
 0.22909681454077846,0.5766453904478618,0.5009324005733915,0.4011142701545807,
 0.20289621203496278,0.7351055186356824,0.09251062540419364,
 0.5043131710332553,0.21129413407638153,0.20752520512645145,
 0.38770911965027954,0.04157541238301854,0.9114342828313602,
 0.30640023902762525,0.790803434716483,0.6020699358419825,0.9671046912456425,
 0.6433689235416697,0.842573350004174,0.9036667647948509,0.3880398929433687,
 0.4280264597712965,0.27186626268670133,0.47714789099609456,
 0.45727251617054854,0.6813780113855942,0.9416569045245569,
 0.008128546858567676,0.0806494210890113,0.06676752984871281,
 0.24105244940381532,0.1351488074718622,0.23291091496380867,
 0.8809952816134469,0.971635777334991,0.2843061262010623,0.41969574345462,
 0.7488097789067908,0.5184708748967288,0.3756530668468041,0.8694090008942532,
 0.8284753654469504,0.01950517532877427,0.26279673788340907,
 0.9231323459504138,0.17705215725162415,0.1844657656280807,
 0.05293604379218653,0.9789415663393687,0.318364958491281,0.9647344199656683,
 0.6764710220851857,0.7749846924812646,0.8123762327276749,0.125832869501197,
 0.7916356651001051,0.5501384558882089,0.900040279565587,0.767087537265237,
 0.9284545748171903,0.4510332652649317,0.23286415467578214,0.3143236425038386,
 0.6251411947370537,0.8206597968961385,0.22111411209389897,0.99380365209278,
 0.3545754710527239,0.15379681057465588,0.1319888407130979,0.4409844712441596,
 0.7932759753927204,0.9002068434602856,0.846531534227696,0.26617506766370913,
 0.7927763544118934,0.4734840843403786,0.12470002137086245,
 0.020003776596062228,0.11405385313377392,0.3000105619486919,
 0.29609961565871645,0.05074879141305455,0.238159801552129,0.8420430114846724,
 0.4317433374271741,0.7345610487928962,0.366271870185392,0.915460674433069,
 0.909752210015947,0.5569914164661005,0.06691238245140307,0.22294246675356577,
 0.43892516265415116,0.28387535087662275,0.6209543758252567,
 0.8991182673842021,0.44070811249675657,0.6671241352248908,0.8900000105007683,
 0.3093252755728483,0.16400911132811669,0.1607910895415966,0.972809646013365,
 0.6612522341010364,0.9927767370698218,0.9058076879861322,0.48273559384968945,
 0.5711864034823726,0.2613156244628544,0.9462355435985479,0.5071610489213545,
 0.3683536853949415,0.6398443360708029,0.8307103859730599,0.48065833756178666,
 0.13372749476402745,0.6141193372437411,0.7828848165100171,0.9787668132337661,
 0.37117732796574343,0.6063886659544908,0.48568634143539957,
 0.5081414633653318,0.8217267896957696,0.8941359855575619,0.9020962688263084,
 0.5880907005846204,0.4771087582239306,0.1183086757482622,0.9851871156546376,
 0.8125605841513126,0.21660402227605202,0.34108181533190596,
 0.8209931171107576,0.911519584831755,0.6210960586304334,0.4721443895934876,
 0.18764066861906048,0.8571482705880296,0.9504176170205973,0.4869427941833614,
 0.8796778719281036,0.09751553550018088,0.5472249196336243,
 0.36375987819037037,0.4859840209956283,0.48816026296990944,
 0.7331033832269297,0.26388258030700296,0.3739062029073632,0.5170364319790255,
 0.5512568961589656,0.045412058341010586,0.2824812924890354,
 0.7244713250779017,0.4098344692746356,0.580030807650603,0.29351973077985005,
 0.9798175566529669,0.3534672905081969,0.3923590731230404,0.8572514044312234,
 0.0421483647228762,0.9296345930707737,0.4207113911319912,0.4633620576480848,
 0.9131836099374033,0.7750614449791201,0.1479973619261996,0.1167814514394907,
 0.4295678632260056,0.7502479021298226,0.38770781316478664,
 0.32865029190881534,0.2817313763746899,0.270148696322029,0.10853641640040024,
 0.7279006359032264,0.5346575615209805,0.3451681092624729,0.219958859103782,
 0.78095248661355,0.4381632251937163,0.7000900750444262,0.7251777559262592,
 0.04418517184617121,0.615495926443336,0.6770592553988327,0.04521327584575763,
 0.093227070609602,0.7626284784398499,0.8875631996222171,0.36986251428722416,
 0.23121932195310402,0.9986303317290727,0.2910475830659295,0.2409435631003476,
 0.9244481654745781,0.637039127143949,0.7476616355853077,0.946705703974089,
 0.13180341288899144,0.9141784567239761,0.26627588434873417,
 0.6474046082383068,0.6306078048062589,0.6709911858141483,0.9405530409861143,
 0.1710470454073425,0.18850474128248862,0.19291514137642796,
 0.4554427082077832,0.2367684096520739,0.6920472106913853,0.08221945503655181,
 0.8330642653689991,0.3156367241861504,0.2855202848529045,0.3209658253880452,
 0.6097546938875346,0.4243320051617836,0.29212347985502085,
 0.11960821129441723,0.7080489097015636,0.11340146871345269,
 0.3943298241062958,0.3135385113749971,0.660905673696171,0.08417571307855665,
 0.9017566202245221,0.018382610642766806,0.18006325450648142,
 0.9893431178828433,0.1065348160425319,0.3464734989368202,0.2675519719123187,
 0.15441990162773844,0.49684214709295105,0.14967398049660374,
 0.44489030341390323,0.09218872566100222,0.9589890357990423,
 0.5160385115141691,0.9770969849645658,0.4164627283371629,0.5323353827279687,
 0.8199907900833374,0.42604568956964606,0.3885617619618965,0.2634316499948506,
 0.906836224065691,0.46917178171365104,0.7071239466828405,0.07216697130293981,
 0.9746969304481026,0.08223357421758981,0.29609167968334116,
 0.10343434747173019,0.35328214532667146,0.37313414923775545,
 0.6078415977268532,0.20141066068119828,0.684302567904467,0.9167221052035022,
 0.011655660193273132,0.5111048356336492,0.8099184063358119,
 0.2870228676177269,0.49062151225496176,0.3174922489706341,0.5062549579412515,
 0.4840502069659123,0.11273290150024928,0.7279368237702701,0.9384377078978784,
 0.8237010615531934,0.9679639578473539,0.6961484911439872,0.48863435848990644,
 0.14794641123042163,0.9255966631234239,0.5884291522157444,0.9714160358360932,
 0.8545139559193274,0.8422604063723116,0.39335709243026795,0.3939975075157207,
 0.49683303614381413,0.10861136802647042,0.011438711790591749,
 0.8457686338600268,0.7078838631902111,0.08697704879217005,0.8583141319343135,
 0.9229531449238293,0.5725092355001936,0.28656067488288584,0.9389604390409215,
 0.9455636397042172,0.05745009170957238,0.38293702777012617,
 0.2534311842792002,0.2471323752020903,0.29189158129687864,0.8712117082482289,
 0.21535550888043953,0.3757126572263708,0.4061064562261547,
 0.15086743316718887,0.0950091642346953,0.34972577460473075,
 0.24739034698397466,0.04568864899150826,0.8681890566250258,
 0.5391881599276455,0.8075990777715262,0.3646837549289025,0.6717968640249985,
 0.5158957218104874,0.475262182161443,0.1669411040474753,0.2277970849302775,
 0.816679601679051,0.08324744770394366,0.9818959290264707,0.8301250018718264,
 0.45880414339258857,0.8966505002302896,0.47907028564311693,
 0.8116553741146851,0.7845767778960347,0.4106366050318593,0.4197443808042589,
 0.697205747479739,0.31090948098335214,0.819000875753687,0.5990429982064711,
 0.9753282589238486,0.6161882813394869,0.9968616149217997,0.07305548685098706,
 0.8269654832910146,0.8382604113206358,6.644604543126764e-5,
 0.10966079802951179,0.7341943204477606,0.7186672385892156,0.9261332038407715,
 0.8105681305620621,0.46215079040081264,0.7607187284619631,0.8769136574785623,
 0.776258435924901,0.7091309254125866,0.6130535756781365,0.15429440716523257,
 0.8941604923427126,0.5997755068815103,0.4844174362356626,
 0.060888426696548814,0.5659247217100895,0.16044757114717456,
 0.9057376990106981,0.036273572941451615,0.16612266461291836,
 0.10763638121037566,0.8807907658503249,0.8821542102087412,0.5793930747578739,
 0.7491956178621132,0.20542294375241676,0.10680588694574378,
 0.4880809444955314,0.6357823007183002,0.560672223281109,0.565577617626531,
 0.3929198490781268,0.7832156539680468,0.4870206343004493,0.9416520288317344,
 0.5117532044594229,0.17598908597466933,0.6962669353888307,
 0.029181765583324748,0.5550495722243699,0.47353692652031343,
 0.3463484406992462,0.6971725480903538,0.5128500314626836,0.19844077052263798,
 0.6355603049241958,0.7229863785707814,0.5653667808809039,0.5060244018264808,
 0.6679555303277104,0.22702408113453698,0.8101196737188893,
 0.37303061115743363,0.11709170107801847,0.3862027928799743,
 0.10902310893622325,0.9043026821331187,0.663159069508623,0.6877594637336044,
 0.45730082574771935,0.7321927236807589,0.5830469356421306,
 0.29785549069186124,0.4238926035117264,0.052501569204892684,
 0.0621637373552395,0.7573670937062067,0.5810959593240776,0.3721919111524843,
 0.4149145746938431,0.5964195466252388,0.8614419621261895,0.5687993867198452,
 0.09109805883487132,0.822240541570911,0.13030800422694022,
 0.19840878314485078,0.8548882557467254,0.6612559987120752,0.399017012380004,
 0.5075354554589042,0.7857879760253407,0.04094274331154801,0.7285882565883275,
 0.3175331862607689,0.27232641477492536,0.9431712003878392,0.7959112492953082,
 0.4585482100786331,0.2640826233673327,0.2635430558230266,0.45735888425720017,
 0.5025014520833828,0.4883082108463148,0.8487079189801965,0.1352764955878607,
 0.39416041597988194,0.693771607287257,0.3208000985773988,0.22354518991731742,
 0.08563193177071682,0.0566674687882025,0.9005652068237892,0.2085681682457623,
 0.46916335485226623,0.5007161398969233,0.6752604017724189,0.7122533442705951,
 0.2931839379135558,0.7217154113947077,0.011727672954997881,
 0.8163315103412274,0.3179087053978935,0.43441270133623555,
 0.29661292938634676,0.7544893770208684,0.011339577046307614,
 0.7948081395431628,0.9393465686825166,0.05243600673036386,0.5760880811826414,
 0.2676174537965764,0.12567561515192016,0.19957249835425006,
 0.3770551442683202,0.09156055306635569,0.4946349110178263,0.7768055097145176,
 0.5744259218778827,0.49197826174891013,0.8698875013994133,0.8858348512435926,
 0.2447678840318921,0.3858623665729013,0.03535888296701839,
 0.03512112881934604,0.3337448466527372,0.6729754990017365,0.95695888302331,
 0.11494005274916985,0.31872411639750375,0.5526267243774053,
 0.3162908138279744,0.5878010362687124,0.5881287299267309,0.8269345809222303,
 0.7502953149283412,0.8251494993931513,0.16002544711121403,
 0.017940666416991324,0.8659515894762764,0.6787531505358795,
 0.5848706434459618,0.24285153400924264,0.2501323189018635,0.799893198960515,
 0.8670831203007979,0.8493475579996134,0.6891324638924186,0.3474954329294534,
 0.26976616975592294,0.08677496951065322,0.27568292480154877,
 0.4893396916946072,0.26028497268413775,0.9283160435387473,
 0.47024471913016885,0.08384612494912314,0.3523686826736748,
 0.02693515472457042,0.6315123412076801,0.333542537663846,0.7864742599447152,
 0.29136548767029324,0.039794134664935754,0.7617743719277303,
 0.9443619013251607,0.21606706150848032,0.7312618590868479,0.4914406475948936,
 0.5924464362364997,0.5171446268696953,0.16801129333576026,
 0.36240097673738114,0.6255506020377597,0.6968315664517106,0.7926862581177139,
 0.5918188530747436,0.1019798428482267,0.06756361649505438,0.5390156808592758,
 0.09658669748962945,0.7287813713758714,0.15783875200103814,
 0.4591835155242021,0.5245368601316887,0.6867392865064649,0.3930019358305463,
 0.9147314164978032,0.7369535438721799,0.05400080685069453,0.2586841419376066,
 0.37590807755415445,0.4175388951031276,0.13429833903354027,
 0.2869287906267617,0.6920094899883958,0.07480228524135879,0.6115613125225754,
 0.47825089706942203,0.011131775998628357,0.2093595160805225,
 0.8970721124652694,0.09792734210411624,0.5033381349911608,
 0.47676934508249746,0.28914307835906583,0.9737501656994556,
 0.5238904727039243,0.7344990316348365,0.06542162127151074,
 0.37794368163152026,0.9801204695814489,0.9494158897128516,0.5354211451955526,
 0.29889094614653444,0.07499490450781043,0.15166723474281119,
 0.12102067842240449,0.5484972700881732,0.6980216577613679,0.920012984774452,
 0.3430294918637756,0.5011434186761662,0.130184447709072,0.27745751226241755,
 0.7769670941278624,0.629746811635682,0.3776211853843321,0.8124086195581779,
 0.929714334107866,0.7374768024423098,0.46738018138678894,0.8412034680668126,
 0.067601748596548,0.32631098962181393,0.7614587671598333,0.5159730231295344,
 0.8514710255015205,0.7797420708230421,0.6962768342959608,0.970040220463076,
 0.6476491541341645,0.38976871716778994,0.7102460136500496,0.3074004500206565,
 0.8291932351815776,0.020798925054269835,0.65312017922063,0.5442045900889517,
 0.8074188331918002,0.8816024744019745,0.34459948609496305,0.7619414742534071,
 0.9240801714851521,0.21384875268256587,0.8594540694327871,0.6697249185685781,
 0.7023299857793438,0.4901133955823652,0.4463901361674899,0.17612181130332472,
 0.48570987768067164,0.22714973153790363,0.3195444241610108,
 0.8872046366027257,0.10006009352749179,0.8219647096497745,0.8456553126432885,
 5.49422219067508e-4,0.4316190219845423,0.8646202102771646,
 0.06914787060679306,0.9002801527864915,0.11753199450613061,0.772899274533348,
 0.2000633897794022,0.49096839270377735,0.4708785855402824,
 0.21616127032136456,0.5901169082984233,0.4491969180682297,
 0.15605355356735107,0.7744038032688103,0.12591175538834232,
 0.6344040705178051,0.16888432853116164,0.8048563219063332,
 0.32064253390515085,0.7394025099638293,0.14775577716029398,
 0.4019424273953063,0.2952699896893033,0.5420982575508762,0.5398947601392468,
 0.2835144117789129,0.5041349296276412,0.5782967820073835,0.10925870801936577,
 0.6890346672014106,0.7572603594113694,0.7083040551677956,0.8392137828647509,
 0.7505620191948095,0.11154796863636707,0.8126331232264001,
 0.012016389009024797,0.49297993323715295,0.9037024271819327,
 0.6171651729918518,0.16385174478818532,0.42139223416631055,
 0.05073407901648952,0.640339263017357,0.11308913212454619,0.7203362580638742,
 0.7919285324453134,0.6913628013827025,0.5230403500170937,
 0.024491267110795634,0.2876128394077575,0.5817844638713917,
 0.24316330139751408,0.7256846545820145,0.19751276038159782,
 0.7919485820158513,0.6421111547512847,0.6971778044809982,0.31703363237348925,
 0.40053066479797395,0.5192210634136433,0.11888301340275875,
 0.23161222870508347,0.857501978092615,0.5004311820643887,0.4285842280283536,
 0.9041254265449246,0.9903552366655582,0.45666612956676567,0.662155454362544,
 0.582553435635655,0.6371776304318855,0.043840916563556376,0.5500435551635775,
 0.17015245685557434,0.5618293153830987,0.6702916464416379,0.9606935202354039,
 0.5959555194422239,0.9130449276393255,0.11405816567804761,0.4139803101023687,
 0.7979722664133455,0.7877269628393329,0.2881478233549255,0.2941588758206094,
 0.8512084327491263,0.652294640640755,0.7706423971797152,0.6257565394534944,
 0.8845217414234873,0.5857107380631801,0.8455440095314145,0.31766217839050237,
 0.6953990849878909,0.25412782389489297,0.32460608461451324,
 0.5472131499764765,0.9627266877875078,0.7765209255343717,0.15540403249979096,
 0.33749884801988894,0.5176085379467319,0.051890016259901306,
 0.8939855936705247,0.908162150812202,0.27952196383568595,0.840933457933729,
 0.7831348221481368,0.34975037327945235,0.35424631279254104,
 0.7955847420550352,0.08189049706124751,0.9316125591039991,0.7860804050121246,
 0.7213854772210896,0.695963345032408,0.13946661013247486,0.9746189534025783,
 0.008802199047207693,0.636058164362776,0.7791212999623305,
 0.17822291819428493,0.7614418942962856,0.5138655941926705,0.6684659281516788,
 0.08736871567257143,0.4354689100051151,0.14800025263382088,
 0.31724120629651986,0.38890758583538265,0.5145748946758637,
 0.8277911910075685,0.6505420924625698,0.5845690054816692,0.04946050036454752,
 0.4054707675677899,0.5931417299241392,0.06987562634919608,
 0.37499499851166307,0.7942281374697515,0.38831922485655535,0.912852292481456,
 0.3767956663212969,0.6995112043485826,0.09862136942287059,
 0.13237414463730945,0.6405808825619526,0.4481980296380764,0.6109449887892873,
 0.7603547368739035,0.45872868508374454,0.335063517203261,0.4733147264796016,
 0.6249545498030896,0.4763239163529269,0.05013058616127064,0.6920451272410335,
 0.059781664279476066,0.3808357338567042,0.8694763421473389,
 0.10984202274632993,0.8228422844752554,0.15258366469850393,
 0.6816254112729785,0.8890888952143075,0.15824140101644046,0.5294296606424675,
 0.6499750070624821,0.4591189633388698,0.2603952061561815,0.3475870219752901,
 0.16628054088290367,0.09103281252054585,0.6567672697896457,
 0.5169304277011745,0.38442272957052337,0.2946431063396574,0.3658346878098707,
 0.27808661450445094,0.08887096527644289,0.8971360238095787,
 0.9832286690090842,0.05603550780284394,0.03387773016169704,
 0.37969648840900183,0.9583421633557198,0.6612650085740704,0.4184055447913504,
 0.09983887815090886,0.8012734871013548,0.2854345215558135,0.3921762436916556,
 0.23824775226199246,0.4385203320781331,0.3762328000217765,0.6290916358856,
 0.00844360808971778,0.03403707193368355,0.07837393915892565,
 0.148690946638669,0.7569639415782914,0.6538196760662611,0.0837887844642401,
 0.7364650041054429,0.5984059218462414,0.8971412270648491,0.19913676690387205,
 0.1189140847397564,0.5464018400519383,0.14523341807209644,
 0.42277499811973285,0.4903293298526018,0.6784300621862158,
 0.07521249620361081,0.1280255623851646,0.915509244457039,0.8995735636127056,
 0.8796209120858205,0.7918077257422086,0.7960013641095502,0.6191040591723604,
 0.8146268307621533,0.9705603372559238,0.1850077200854151,0.7614957172328025,
 0.3426919243939647,0.21080249011726349,0.24712166323512463,
 0.16498125888285275,0.3557969798777525,0.2355062953356042,
 0.42040244696535445,0.9270647409877281,0.44977194351958727,
 0.9387572504793841,0.3417917135760442,0.06413218362186146,0.2729985903698926,
 0.13078487346184886,0.6949469513447977,0.5484145271358019,
 0.29727584279412334,0.853004321139597,0.8063827271879982,0.08035349398467184,
 0.25721414755961813,0.9864605431737349,0.4790879534202649,
 0.49236395325961524,0.3943664971952686,0.8759980830157381,
 0.27985855368448675,0.9456146835038682,0.4785856736090228,0.1802044299509682,
 0.682713378371262,0.47045938136154963,0.044722518172980785,
 0.45882739141927575,0.9188338590931765,0.9289022266940596,
 0.0014211066099236724,0.9718413981542573,0.12911655534206345,
 0.40255435108849147,0.39965886520504734,0.2727357318633614,
 0.2454132795083659,0.4690948099272516,0.4128645317857371,0.5016391761910416,
 0.2637095003028307,0.6368893550382132,0.3471390319480684,0.8301932920900003,
 0.7960311354498842,0.47578620493364554,0.6594007389642181,
 0.02831189116523014,0.015193314691014148,0.4838162888172005,
 0.25262939449293265,0.5725507202136944,0.6498900859432473,0.5833571754562692,
 0.8997644290772497,0.8326294375593412,0.7412659002825054,0.6236422422234086,
 0.6099308455368713,0.6747696770812326,0.6778209340881329,0.9341488415640069,
 0.7783620427900498,0.10490190077790773,0.5488918486316029,
 0.48199409583968167,0.03695540065620795,0.021669292352210934,
 0.1548126558173466,0.20450844325700812,0.6882547217478471,
 0.23629739935566663,0.1554534245762077,0.13079888772087567,
 0.14042672534718537,0.3559300630730685,0.5531356177173723,0.588478294754496,
 0.7566758018462796,0.25266042663085464,0.5827423675542832,
 0.21976896266123025,0.12672987004567404,0.41863717237182596,
 0.47018380442347874,0.9781364683702288,0.05650303530086531,
 0.8548632857332016,0.4389998531577678,0.9758676620098259,0.34204455404142964,
 0.17645157551917567,0.17980655274028245,0.5816561630063635,
 0.3273714310898641,0.5842653541149891,0.6824245815614417,0.32115763925793184,
 0.7684880725410627,0.2530869072050286,0.11892066213316055,0.3819473358875709,
 0.940392537240196,0.6051485165510483,0.6048890635453117,0.1546418313145438,
 0.8248271831497123,0.6089320245314169,0.9536567490014312,0.40277814206672047,
 0.3805362173220471,0.2109074831933,0.2958414059567206,0.17366610604462995,
 0.3816394794214697,0.2614472970891407,0.8648560881444634,0.40811625623406167,
 0.6450172723664518,0.39945532116833726,0.07784104605647202,
 0.5417474739738743,0.42322914835858405,0.6808559727841775,
 0.049428588190812084,0.6220991292992621,0.6476618472641232,
 0.30757712457630737,0.3704987837272452,0.5893517610417169,0.7015240455539592,
 0.35239441803890226,0.8422593551174344,0.7267810207516703,0.7553047857676957,
 0.9801257902382622,0.06776896455303283,0.5719330532206114,0.2559850160093001,
 0.9913302657256841,0.5298686604229641,0.6567091926755131,0.5222213500845523,
 0.6251417176932599,0.8550074325092507,0.2233494294950682,0.9521168324471778,
 0.32937292650082917,0.020813508305934425,0.2646989360349745,
 0.43884484887011177,0.8503186662536886,0.8257640876346583,
 0.16239351333134522,0.6508319215606911,0.6070824914639751,
 0.26320163291741683,0.3121177141142427,0.6776877140293845,
 0.42519893702566225,0.1825237243878689,0.2548732714832873,0.7137959231281225,
 0.9968049285840402,0.6285971815901601,0.046385271864551325,
 0.21393641856820722,0.7040649854726775,0.16318244447403119,
 0.5904161325107677,0.3717356000496236,0.9964351648783395,0.9366646745871942,
 0.38418265703424237,0.8620984188716904,0.8030781847470375,
 0.10065241396495594,0.7133921845742759,0.2742542761492237,0.3865122217634247,
 0.7422522523114972,0.29672776643555987,0.44218451333618647,
 0.17068228175149536,0.9450728698949266,0.5333648621090592,0.4769898225648339,
 0.1333519718057029,0.2120878190649027,0.7277718608992033,0.2044165455368484,
 0.8800860973254918,0.48392332382699155,0.25510865169162233,
 0.2973040110553701,0.024736115370143752,0.542105543181794,0.6780113514420091,
 0.5905887772480812,0.4293783061002442,0.5366618517247812,0.7940438754353061,
 0.829435664479788,0.4479778512377717,0.4875843666397073,0.5226689010013674,
 0.5504261259857461,0.901218585051175,0.12037220596021192,0.567693910127122,
 0.5668527224363181,0.05919599295042188,0.8306633033307201,0.9836676803021709,
 0.6138034772279094,0.5410528467581893,0.19140822036539673,0.1659388440346149,
 0.8272530782903957,0.0263486959938255,0.3661644105669517,0.8618615616889593,
 0.10654672966410239,0.4963798781831186,0.8675410268721362,0.1939934551551734,
 0.010871671517284742,0.9429508129565832,0.826200326667925,0.5316270283129044,
 0.8882354899305368,0.4331729358019786,0.8490778762530059,0.4615612239093285,
 0.1904979652714166,0.829346650135041,0.43133082357536,0.605247378793881,
 0.6458341092014876,0.5431110587112906,0.8486765824299943,0.8084727585604166,
 0.4949255021190512,0.4988264594277805,0.29428347715807035,
 0.041360243812883724,0.7055635010558958,0.6308510718455248,0.898982955033682,
 0.9361199421305033,0.6251297185389331,0.2447597920737199,0.2672793908505309,
 0.7070125635976903,0.8583501386621541,0.07958654059503567,
 0.10422967050562937,0.615845569665014,0.6570510808133189,0.5553807369782268,
 0.6600682109844307,0.7269184904146937,0.5971567111348777,0.41401845551241423,
 0.7001819062802869,0.24113545076576048,0.6266218500165457,
 0.34197027941693336,0.8334499484132213,0.12204656965681904,
 0.8991994504069989,0.9940821322426158,0.29965244626596643,0.5918982970998011,
 0.8710215906706913,0.6804259779036996,0.10202248301000805,0.7924948045893999,
 0.7065948921832863,0.9584271164304672,0.3341071936116665,0.8949059348455968,
 0.8375420641173166,0.6124802416924824,0.1800444874134004,0.9116887678727084,
 0.5474622214938734,0.5501754994553966,0.5247116499217863,0.10782493096222268,
 0.39582805075845773,0.6246308258200249,0.8403122853570422,
 0.05009894503220291,0.12365382236853084,0.1363819233332343,
 0.3706225632826272,0.45209284523940085,0.11778098674127746,
 0.6696489987172214,0.7493379027097746,0.26223775086159173,0.9881184737895882,
 0.23956839891192516,0.9782459421406402,0.3573671405748432,0.7515263892665054,
 0.1278538445409293,0.22920145831070138,0.17996544875622866,
 0.5118369277831296,0.002330534364252901,0.1710779577474586,
 0.5507970129002887,0.24941853190564234,0.9595191164513681,0.9613618302789237,
 0.4826249448317933,0.7460959675204657,0.37406676256699367,0.7993745940185639,
 0.7256016226167755,0.657767110427611,0.5006628534024252,0.587084442602114,
 0.8017618905209366,0.4410301891731182,0.30911635789448155,
 9.202649939028529e-4,0.5183388123458352,0.7799895978814058,
 0.3461062109871268,0.726623251130043,0.9937097018942321,0.20661415143809403,
 0.7477945534701655,0.16466225627591358,0.13044515615546226,
 0.4462848566955364,0.5031599122884728,0.8266138889797613,0.1135385913620709,
 0.13865058384370466,0.9535026830436204,0.6470543885546547,0.7691318932658655,
 0.6028496205411943,0.5672589454064385,0.9853834238536471,0.32333622425205955,
 0.23004429679481342,0.8041977624093162,0.676575413976116,0.8708144932799717,
 0.9457438137343898,0.012880186118332526,0.4729830155946394,
 0.9077778734344863,0.7868607491131501,0.15610630332562658,
 0.29906099734186375,0.5195372957590131,0.22947646208088335,
 0.5358933321061309,0.6499623102898774,0.41793781806370167,0.9298496818697356,
 0.341164713091089,0.0013214733949988844,0.7038582538533064,
 0.2765545945122082,0.4225624121917506,0.29884298526089037,0.6368856440382468,
 0.5097907327087979,0.7756616403782783,0.5857308454276615,0.6747857983354308,
 0.09480058225151944,0.19734473621457638,0.4327963614463317,
 0.8191886230201916,0.2749533401165809,0.8080064493291749,0.8117257335925824,
 0.023306089337802982,0.10462989522690769,0.8003830371133029,
 0.6854382614851079,0.9353756755116419,0.025525845494534516,
 4.7677455556405235e-4,0.7335918702936283,0.6610118514210833,
 0.7759482037261463,0.4498652881482499,0.5204559708252846,
 0.005016525509196734,0.042954020296861906,0.46136549500367174,
 0.052752934591981604,0.15512453399535286,0.6907542145467054,
 0.4234825380214714,0.11741442603318952,0.36738758662534077,0.538306115464821,
 0.18142643807893077,0.4807345262469207,0.6307881538933539,0.3905703077170204,
 0.9418618238978612,0.5138300237151942,0.9208722540820861,0.886402467951922,
 0.33459376040510225,0.7484424024535207,0.37224499029279823,
 0.38463273392109476,0.2723245761705534,0.5906198495430997,0.7811063837411448,
 0.8278722299834755,0.6388420181221822,0.25851775349834316,
 0.23746289219466465,0.3537472008851793,0.4441157914341949,0.8701853309484233,
 0.471919140572195,0.7054813028417312,0.8620284182847298,0.7028806051004005,
 0.23198969270681824,0.04024738270398176,0.9696729344217919,
 0.2864093093356498,0.771777536862269,0.1434231946637613,0.19895215813826983,
 0.37426812562659617,0.9226068655449167,0.08285869022022074,
 0.7245714319519735,0.3541833451803973,0.2140032836816681,0.1967476884982251,
 0.17292539942738572,0.7585051351297939,0.4122098558274758,
 0.24219579972308092,0.5168761764575761,0.8071142028153513,0.7367486508584203,
 0.7165802903457252,0.20274569930268938,0.4942017391672552,0.4566800018736338,
 0.21569617825002896,0.3566598253518076,0.17485554265609515,
 0.4193066089286477,0.18532648501773175,0.4056309450503497,
 0.40780500758986826,0.4468913317303116,0.6623821794736925,0.5903582152871577,
 0.22064389660445782,0.3593289432023279,0.3921595565623146,0.8606264978563135,
 0.7180675080979042,0.7003529466266967,0.42723386746113434,
 0.46796292258136196,0.11944549477861444,0.1091393083697586,
 0.7293657584249269,0.2618701223200208,0.42197199425351994,
 0.004829977645261052,0.10314035190370507,0.9179758508577059,
 0.9947340402963454,0.19848309130710673,0.9559557071302612,0.5134473363195025,
 0.27865423657238353,0.8627554843948758,0.7152796071312704,0.9172338427263252,
 0.09267066750370923,0.15224134748348095,0.8193297107321534,
 0.6963192865125275,0.2920116666481887,0.16434745145790197,
 0.07396075120016077,0.6407544844988328,0.9494759142685323,0.2529242264827136,
 0.7817792576776336,0.5754253463840108,0.3808501164089638,0.02645187321232878,
 0.2588962511209689,0.07069672346250411,0.1957281993530553,0.4068869780747222,
 0.7974876695217291,0.17098823157913678,0.4821163076438435,0.7855052432330697,
 0.5050084328810394,0.3584227868997627,0.8509427947194549,
 0.005854920892802484,0.532354881478707,0.2672708008131103,0.7093644400258609,
 0.6682509888837914,0.6766053076310833,0.6380268278281783,0.21511683434909,
 0.9717444627498164,0.18635776480945188,0.7014232469591571,
 0.015444845246087913,0.5179294374524348,0.8494944835226947,
 0.6550793758625834,0.3652647382844918,0.065668897654245,0.620276934010229,
 0.11894550872413667,0.3254449520465629,0.07007820557982669,
 0.24032868359888926,0.28373972278717363,0.03935456100498724,
 0.8732934808974967,0.211187994568115,0.6220009863434706,0.08878918652661916,
 0.95561311439783,0.6162075516209986,0.7572864963050328,0.36752263629399307,
 0.07198584407735897,0.3985475441275532,0.4359312527987892,
 0.13257568588709345,0.5348250944937101,0.6397836205243781,0.9955684889495633,
 0.8039218023265362,0.642432153440756,0.5758648511304136,0.4052833678176624,
 0.05426886409410714,0.1064294472213334,0.9317715491594883,
 0.13701659210590944,0.5834351983855887,0.13319341613586255,
 0.5856214444085845,0.1370742812338217,0.39873804279770564,0.6418161508654949,
 0.9707105830740406,0.7460176300155033,0.9161195698628382,0.35463654346880746,
 0.5817653242016816,0.4681969704135599,0.6318247170265258,0.3415378746659977,
 0.35928896730643156,0.7375338022154386,0.9270848260873075,0.3337150888431888,
 0.26699454318116567,0.5959205662782325,0.5146465501962993,
 0.024097900162139863,0.9698224595859207,0.26788587333297165,
 0.010963464322048755,0.6437533459572953,0.3050070492994771,
 0.3932177992001832,0.6700485734926961,0.1492007405097805,0.6155012524721579,
 0.4049845252916302,0.033092088533707376,0.5205280211671014,
 0.26258782994627194,0.35897489379912273,0.6353334743289373,
 0.051903541009495346,0.9991865546293983,0.6375202715412106,
 0.3645803146367478,0.39553927008979806,0.4109012476763608,
 0.03578414497814286,0.9663829101851819,0.7505118942444873,0.7703557830696038,
 0.8821596424566124,0.3611963695764242,0.45432240190416406,0.3366474910369043,
 0.9751016911937099,0.35131337094346304,0.7888802198840107,0.1604758541274569,
 0.7090705757489156,0.5348311810817656,0.43891934934505494,
 0.15830635555223738,0.2183282067541774,0.6369756448971083,
 0.11836362521278931,0.3856570741163886,0.27360467937916066,
 0.06120894856787451,0.8280164031122919,0.21260055375550235,
 0.4439815600225042,0.2713760746723042,0.7133547791723814,0.1708300634391713,
 0.5233284692281138,0.9175383458449302,0.22236054668506078,
 0.04837309577344506,0.9007312973814812,0.04106605226914162,
 0.9346162849545828,0.27848813931900296,0.1655807079909013,
 0.19309038052477112,0.9772079074623302,0.9207424982724162,0.9310199400452819,
 0.8172191129039794,0.7295086728256089,0.15169369650743625,0.9179205879479637,
 0.33263416135289314,0.9620875259317089,0.45371263013980445,
 0.7915224687740597,0.6851267309376974,0.6709002906347816,0.2996021944560967,
 0.39913599527673305,0.27655878627576314,0.04609776518609798,
 0.25350070511990674,0.21286847652912833,0.6631806983501737,
 0.050884173048539205,0.9565684300240913,0.5486147980175636,
 0.03468031770727098,0.7746734505999107,0.9058498665078947,0.6633166812752841,
 0.08279529372667616,0.5123771711431014,0.9273314082578926,
 0.19832793402674076,0.1604618254725424,0.947528586551165,0.7176687456603723,
 0.4363695396478048,0.8180233334971144,0.09068431273338107,
 0.44537925467429174,0.7740186612917684,0.5655428054591608,
 0.24117107674062055,0.7934434723555984,0.10043298560024805,
 0.23309772652447702,0.7252292890105934,0.5240771844550556,0.9549713385900409,
 0.273468298066726,0.7171279959142349,0.8861045374176378,0.7746578275943257,
 0.17922392593091085,0.4795059610348802,0.342499678328525,0.664688366728207,
 0.9811333727634287,0.4585325524201669,0.7601279029735218,0.3237157143663383,
 0.2754114194876013,0.5380924348502714,0.7307140462445632,0.6789194249094509,
 0.8029885808729879,0.8078850170180242,0.3754333343917342,0.7863261050921558,
 0.6575768474259441,0.911700679966354,0.5051308538201833,0.8123608628647377,
 0.04569634089895458,0.014454034723695774,0.8203990438299691,
 0.3656503065835788,0.5372799724553037,0.3871722377882929,0.4637016314612329,
 0.6193254720914115,0.5429868828226266,0.939401336567314,0.7451671333871052,
 0.46419324124506856,0.950735683853785,0.3183152512423878,0.05705918910762853,
 0.11072554984759464,0.34986995360731643,0.137939106744686,
 0.11762089360319372,0.757092671480409,0.5888062282911861,0.5914454024624817,
 0.6333447794060937,0.2754321890421503,0.41114983106090075,0.8774742739806451,
 0.2639792269105551,0.41829385796447327,0.47883938492427836,
 0.4487117515804352,0.8461817378961592,0.671756722431032,0.19220435481228737,
 0.15357055312542878,0.6661307579283999,0.25504436310515977,
 0.14702200957082967,0.6746232930752823,0.6905053452250768,
 0.43790896451905414,0.902258457950679,0.04657833728040495,
 0.04504455765401971,0.42448537981764356,0.5538687858728266,
 0.07537615697000088,0.6459177757534237,0.10053873453652662,
 0.7187218442222976,0.46838327474285246,0.03401045775386191,
 0.25167656944748173,0.0028943968416400567,0.7085764594895602,
 0.3169086086131462,0.8900048799464509,0.9508073888183833,0.21876313351056964,
 0.9230894342351712,0.4505645215646281,0.6198629461835248,0.330340895282472,
 0.5317402457480582,0.5820650466102676,0.43678609959412285,
 0.011155494781426856,0.7575088199627205,0.5175396873841877,
 0.46946242854032594,0.128383241794768,0.549769860273256,0.6290229900373986,
 0.6991745287930564,0.6363510833445728,0.9844467390927087,0.3065712232755069,
 0.06813929038278521,0.46633448714403136,0.20426453129767908,
 0.28398642160211873,0.382746783700465,0.32839309410981765,
 0.15763564817621456,0.32798017949201763,0.1265253351935295,
 0.41442088233658747,0.8741210229653107,0.14860128362973346,
 0.9015960130657394,0.5303604521538572,0.671488942065957,0.7664754468128588,
 0.01005435544202149,0.6427087972716491,0.32819623924090113,
 0.9376745649201434,0.6758820733319267,0.776293410617798,0.43896933149695594,
 0.6452287266536482,0.5487493680952013,0.6001977520537002,0.23359363135147848,
 0.10749994648464323,0.3221603899743043,0.23248221250304824,
 0.2610828565424457,0.06972900841234442,0.6501981639247484,0.9631058987428828,
 0.7881956888753802,0.6003425023736684,0.6643463164531013,0.8136687374204473,
 0.47216044405011726,0.5438527204084772,0.11897797034122526,
 0.4961604621381953,0.30384999427745385,0.1948394414502136,0.9014721334011161,
 0.07750593668608097,0.6129883545543252,0.7795105287178778,0.800413768006168,
 0.558429129389097,0.39295698798384016,0.3052100178851319,0.6084479046765984,
 0.8709697628859404,0.2840535708367874,0.8917942995176976,0.4778631542352616,
 0.8982095060264832,0.2863320679207466,0.8961547276351016,0.7172954202738183,
 0.0543969897299591,0.8950556496502432,0.12101679711728441,
 0.02304138544413603,0.859183544427812,0.12942821850690733,0.9869079144629703,
 0.6673730375733115,0.2092748589320652,0.6424168128096988,0.5255066973197211,
 0.7527797813405671,0.5440620756436709,0.7693842181743042,0.5683339330280796,
 0.2206625795534508,0.3471423507511143,0.6200226320831855,0.7726002560915967,
 0.3328374726749477,0.4532522006430897,0.7842151108005202,0.07800922145092581,
 0.8888241617563273,0.5436185260383437,0.2871640993563789,0.4796261689513197,
 0.5224262654288281,0.47952852880115937,0.00797158553819377,0.700730374901018,
 0.04388875519222635,0.7931960563509702,0.3855443412792581,0.8762710262934468,
 0.10942795409757966,0.1786873785725045,0.2588700394865766,0.7058116006290274,
 0.03215103393456298,0.5876240818510479,0.2146393213474126,
 0.48282944907634984,0.8588564699329564,0.6911614395668381,0.7035167020385029,
 0.9950652109829106,0.010589915035270514,0.339667919964621,0.4231685507792744,
 0.8078949133707635,0.5230640217680784,0.31911532038886137,0.5089004640337871,
 0.0035107185835893118,0.9414639589051025,0.6229327440634052,
 0.172021677429957,0.10468367620121222,0.2779208409026377,0.9955503939176374,
 0.7598929548865434,0.6309549782853937,0.7396767128391413,0.07988591008185231,
 0.8012837843204298,0.019746917036499112,0.5109552949843951,
 0.9964008879068637,0.8836607778859746,0.59502997930271,0.4194897383018227,
 0.15940558246064918,0.3951095933114309,0.737794898240037,0.9921574277244186,
 0.24056834663878413,0.6075719398868291,0.30327275806566656,
 0.5905332922281991,0.6353363520242039,0.7035254924598934,0.2889408400487715,
 0.7001295880086581,0.2844334376062929,0.4802841677077052,0.04422469562788234,
 0.11612049553331172,0.1665775087733583,0.9560823053164101,0.77223958791006,
 0.9505515340515048,0.8576476667647156,0.8395521801206223,0.30681628738680566,
 0.8133197987366175,0.4084185007090456,0.4549444285217312,0.23563624744135403,
 0.873686377060768,0.590011562944859,0.763381342766008,0.6499356955314726,
 0.548649478993121,0.7536578416846466,0.910517263039038,0.6706136058015811,
 0.09228236157657876,0.44304965368160587,0.5983234112644065,
 0.46132704008359704,0.8065613439590615,0.04238147400457892,
 0.10370494758496729,0.9552871748934593,0.7435311116443594,
 0.19947391687384253,0.45207194318315647,0.5223507868204208,
 0.2568421770884698,0.1507691052462916,0.6323041411187772,0.684002631308209,
 0.8553953882226747,0.5511095708022661,0.6847849877135819,0.8681513532797883,
 0.7299047683610483,0.14257150891848336,0.4862599659245398,
 0.27979005568552673,0.6759301068870613,0.03421059420787753,
 0.34027564102054064,0.654232307453506,0.3752181996913988,0.3564529758436942,
 0.09387895606850127,0.7456359563916208,0.9801025019477109,0.7264443496472349,
 0.925904703074626,0.7881387162647873,0.33336683697362,0.8343556131355385,
 0.9649982347289895,0.45296698278690584,0.7989836745809158,
 0.20264716841032493,0.7319698526146105,0.7987532354380761,0.5384284709653259,
 0.8671415215011334,0.3687294658735275,0.060031214436648495,
 0.3183594186420382,0.3686336969805968,0.9543862423829628,0.05202290220335459,
 0.7326943911475615,0.30172981495565676,0.35021377071614834,
 0.4841492611789484,0.7561565488713684,0.8204274218217663,0.8125522501123775,
 0.8164287354041062,0.18238363702778826,0.8150316238743734,0.4411015208903426,
 0.2046696465406065,0.9922921281450113,0.7820993606537725,0.3593401138248613,
 0.5871809760209251,0.8954971015164769,0.6488205539946246,0.3413911288340019,
 0.7301271816363099,0.5386639646603096,0.35288548368638395,
 0.25495721449440834,0.9788878799382436,0.287279257134354,0.23228507917194874,
 0.482835269246348,0.022786987147382254,0.937387090018842,0.0297062995887869,
 0.9161020177927489,0.5639523793452835,0.4927261513040848,0.6413779766490647,
 0.558200261900201,0.32957797500609054,0.14048182422118805,0.9393573874451532,
 0.020755195375532054,0.8300573975785666,0.334145606841589,0.7661340443225783,
 0.02719005447224676,0.3005024714531803,0.9361177353876928,0.587215049616388,
 0.0912428458557718,0.3786248116208456,0.9160718363181719,0.9442912943655951,
 0.35965747038981444,0.6113334902480552,0.9868950397878732,0.3717630525037716,
 0.7921489625412445,0.7362716713919883,0.18689853042151328,0.5392872715410152,
 0.9162324214750037,0.8189794728068769,0.3259902049295703,0.9596532429591715,
 0.48947148629421344,0.2548905254800796,0.4437356685365952,0.4639552320424356,
 0.10224674880315665,0.08638600947693831,0.5787060075065544,
 0.8893502108908674,0.32850534338307547,0.5543148653481762,
 0.12316774020318577,0.24451260321596968,0.7036102703324114,
 0.9855086238721826,0.32976159130095895,0.7232893285684274,0.2826571418171695,
 0.28182803306433346,0.6584883500239382,0.8567164121562234,0.5554884641480656,
 0.20105937266336538,0.12976006222802594,0.4327807632510342,
 0.08820218501244326,0.37353106896134913,0.41088581710294214,
 0.7460527783003845,0.2525950457939753,0.8254565779754826,0.26102279248661575,
 0.07493189589605842,0.5739560461308552,0.3975529499541012,0.5475448836465091,
 0.20094763561770357,0.33237226973526934,0.01790536150280353,
 0.9214973317072126,0.6948718912373554,0.06308285689544446,0.6573437224161901,
 0.24643482670884043,0.629067048626472,0.732850488217846,0.1703398448397503,
 0.7743633329708552,0.9177793633829217,0.6007690856167979,0.8507556182278231,
 0.9209766225753385,0.9009648797878538,0.9232124898607518,0.27898574227256523,
 0.48930170472286705,0.28166131185839927,0.48842302725538345,
 0.8794631403809543,0.4100218366086351,0.7430512378625208,0.7224074487935639,
 0.02150188779848461,0.7521682778151639,0.40234787324390286,
 0.7124952464878505,0.8220386650522659,0.5435657507175027,0.7660577624607652,
 0.9775864157593569,0.770318915748468,0.30380696072606184,0.5066645186666384,
 0.4835815467599405,0.45208810383585674,0.9372256489635287,0.7642331544024642,
 0.46829947458991683,0.6897753739311951,0.44129029568983,0.9236301271881475,
 0.16154552836057234,0.8398307871815874,0.49631369297226813,
 0.8998189665057426,0.6048369254379501,0.9362717866580578,0.29813889667373694,
 0.5223477369899088,0.40741862384227534,0.3950820523322619,0.8621516351740062,
 0.1808399976320465,0.8968665952057029,0.874480198605532,0.34328043758314064,
 0.6347713947601673,0.6678931525092374,0.4376405602699348,0.10459023922659816,
 0.9361952481269953,0.9799431693675346,0.5854218639877388,0.955833415252635,
 0.7597263360802253,0.8924198860434099,0.6465135056054757,0.1689563027313018,
 0.3205926443145418,0.5014482615140456,0.5908712496693165,0.22202673382979976,
 0.8901784681646412,0.15222657239757165,0.888621529591905,
 0.061922326841879816,0.09151114394255977,0.6417012378651458,
 0.2655626512115543,0.25238205333103325,0.6539592356748863,0.5668419801602332,
 0.43408804068205453,0.13967952988403098,0.8772957658244891,
 0.9507256969413115,0.2109405706142108,0.4592516965230521,0.823147808244271,
 0.981768437186659,0.8583328011394568,0.4942205030444152,0.7487590497675385,
 0.4166236553433973,0.8961394938446743,0.38464307294649747,0.8918525758369202,
 0.679913939048019,0.9763342421809615,0.914807022038829,0.2466623635766373,
 0.7094901085105564,0.9283036806976999,0.9930297710501678,0.751466744840472,
 0.3099313917180062,0.1766211136744491,0.3875297307345471,0.7869846405580186,
 0.8931695063500674,0.8284364245895354,0.7601193341286515,0.1602133870393172,
 0.6052122011483086,0.6114477280345189,0.881287624554418,0.7996778360319887,
 0.6937217028134276,0.19669341982893418,0.14627959165823867,
 0.7785499270715776,0.14498031607208328,0.4990756058133099,0.519072881823489,
 0.9475521137700187,0.6546726014181169,0.2896462740131076,0.5386552845108865,
 0.30038226682715563,0.6843227123375794,0.7771700695804362,0.5811773379212211,
 0.5924938773661608,0.1950331586372832,0.3411243115285495,0.7087555917654794,
 0.004722329230990252,0.4903702458110335,0.17069926397911717,
 0.4191829172294461,0.07828832689453158,0.5779273821371596,0.6900458623333112,
 0.5005925619396462,0.007509581788648467,0.7655067739491251,
 0.4091208502541863,0.35721297796832896,0.47205556226049883,
 0.1770428151514183,0.22287476220654,0.44599248787579315,0.696755321884965,
 0.2240474633720687,0.6181067348115347,0.04401489354937038,
 0.04739997367124249,0.3993326448063448,0.23898824429607812,
 0.6006185153605077,0.17111684374555713,0.5402445965924753,0.7169127897849443,
 0.8702498325030484,0.09717959348714267,0.30646553246837815,0.66208113870655,
 0.9124309374873946,0.04961358602693278,0.559170795306947,0.14019019953688272,
 0.45894707449494154,0.5084377426181417,0.4507437236163674,0.7051223515411601,
 0.4727876561344432,0.76356734775281,0.8466506476291014,0.9924755302414661,
 0.07389876231146464,0.6961097883823969,0.6511287808898027,0.7614873060810561,
 0.42736603861345135,0.39030442869460313,0.6584548549196795,
 0.09515340864562871,0.9921421710601804,0.013999096931608168,
 0.4377659083108675,0.0313407541884726,0.2569269383835059,0.7765856901239393,
 0.508393399729354,0.35165069179264286,0.5938394632527081,0.7121976662166086,
 0.8667876663075066,0.009914729193877592,0.2980408224674399,
 0.32428942602361543,0.08190965751876611,0.8157460075688032,
 0.3447931004862014,0.8453954971745463,0.36539149850374264,
 0.10097368801047157,0.9901445672775967,0.8203580047381983,0.7108488382843625,
 0.3609183742918489,0.8817848928979983,0.8022759839198423,0.5585472646803099,
 0.4806326611937579,0.7235028928220871,0.34054146242425476,0.7833759532054756,
 0.027488783352377144,0.09575867561411466,0.2553342430447978,
 0.3202991429611832,0.38123115832771015,0.08589956774145535,0.920079726968318,
 0.8135053702486781,0.6686810191689805,0.8610326073551157,0.538911681291696,
 0.7846085922680659,0.346959857307243,0.4674067908111974,0.8307454768325729,
 0.42436984588172466,0.4729241831942712,0.44651824928728967,
 0.019828985480332073,0.7337601559266594,0.4328149562928987,
 0.6073755780705447,0.9172630469831292,0.4902853320502063,0.47632357958795557,
 0.4551731209870198,0.570846036888787,0.6227965288366382,0.42546418202375347,
 0.2042314602304779,0.2759811187258522,0.5402010151300329,0.9320410085142052,
 0.4517552975884833,0.695905709731433,0.4781266187789218,0.20819993964423145,
 0.8570546671538672,0.7262065142202739,0.9563289913700544,0.7909867827704615,
 0.43770774999891837,0.0011337187448829944,0.6984162187697369,
 0.8089232240768804,0.2613636297381329,0.14027630892302145,
 0.15973388049549364,0.5671506156445892,0.18418162908381275,
 0.024664416365805963,0.21986458524519992,0.9576538474490477,
 0.7014482704128395,0.87316177176313,0.32735937979237273,0.6850605746993729,
 0.8680074165522895,0.7416744659009967,0.9876439083512659,0.9940329791940965,
 0.06312490022502026,0.9867524080219622,0.7765543677798505,0.5201445111399637,
 0.3400420646261929,0.7205152413292761,0.9829305814374079,0.3357597486354553,
 0.68974641891431,0.4470578313464828,0.9730815867972012,0.6287178381166688,
 0.6596961700901551,0.8074331009430936,0.7939883545743764,0.697384276730305,
 0.7810872308209791,0.9098624571550551,0.06231552067461199,
 0.32814700103823646,0.13742863811551087,0.06509962021113436,
 0.0877885558646474,0.6897024421752447,0.34088926608172,0.7884539924315662,
 0.4411553631158265,0.36321359943949283,0.3205021270931383,0.9576523027013734,
 0.6954518688436281,0.01904071678444985,0.9182622166761003,
 0.21041272018393808,0.36705645485825167,0.50098244287957,0.7466849816079575,
 0.3489126903391243,0.5124714745430501,0.0315967839695287,0.1825969589677644,
 0.3992504282083583,0.6372459655555232,0.05030590412164193,0.7489479094039356,
 0.6096966092442397,0.5845068309787391,0.11787722701696979,0.7273412491917843,
 0.8057572843265198,0.8680967624423722,0.9382578732512904,0.09408489797413888,
 0.9846378544600315,0.7930875504740914,0.611568720098801,0.9647755074367458,
 0.9740170162579036,0.5907167206433188,0.7811581080203611,0.5803534236638562,
 0.9795112824544255,0.48246596211552994,0.10220848718789444,
 0.01703858041093176,0.7446760514145734,0.4449530352519777,
 0.38244808385430784,0.22954508170101362,0.9066825023411836,
 0.005627664909429786,0.03860530344929325,0.5906796461562452,
 0.8232064319029863,0.005530286535349038,0.9679964196907832,
 0.004758923900554657,0.03153384132662296,0.49852750564649784,
 0.7372876696779003,0.6356567899525536,0.17707291580674278,0.7936845599315834,
 0.7376894938527374,0.671936586861649,0.33132232904343084,0.8158677858150029,
 0.599729041861735,0.22390034828332506,0.42169834423660024,0.7857828510942483,
 0.6486110565835187,0.26594494014268855,0.24736356451166908,
 0.09930102391829732,0.5948728082796628,0.40785409156943886,
 0.21183958644360434,0.5361584461819331,0.3249217640725439,
 0.24424069428110362,0.7496622063362763,0.4992480416708295,
 0.29938537591759373,0.9984254498350605,0.3509776328029418,
 0.35792852227339933,0.5089977203667512,0.06646325893281113,
 0.19029364964907614,0.6071583053669707,0.465573328786139,0.6547216493368668,
 0.3659527090727842,0.15017657068687662,0.9346834896577032,0.6940193729379486,
 0.12466550413879707,0.6880135758053788,0.12964357157381579,
 0.3822322104630279,0.5385243212051026,0.19266356236913595,0.9015583594128955,
 0.4599199224817434,0.8270227856899688,0.12210090530245754,0.3770638939922929,
 0.30281905068207526,0.6614629695586608,0.007008841469836713,
 0.8276591197363393,0.4487946458137325,0.21932824916781002,0.0428226883462125,
 0.31051104344580305,0.548970367679949,0.6482071516500806,0.1923652395379427,
 0.6546052120738076,0.8405022731269779,0.6206857766005038,0.5636543962907481,
 0.5811387942815402,0.11389139799252368,0.4912706696405934,
 0.003016165291523598,0.9840997943803655,0.25707545156457634,
 0.9155957364668208,0.08090653060624908,0.767200056782837,0.2194215226682985,
 0.5643830073652519,0.8689779790131893,0.8683901094798692,0.47720242165239,
 0.6672645494384434,0.4439694763272223,0.18485136315542006,
 0.07890033702196786,0.11001013029688866,0.4352808413955015,
 0.32639468760381063,0.26626661488954095,0.3838366187940351,
 0.4363344845932592,0.5385107251794743,0.9873151431269436,0.768325795148814,
 0.4478581358522956,0.2612407743305103,0.331380008658698,0.11771902174117377,
 0.7609938953736293,0.1268032567528511,0.4546962034338903,0.9205477115636005,
 0.785034266167697,0.4850124937118766,0.2237496989757115,0.8390340578197142,
 0.7846372259340559,0.8087667213644514,0.2434486913988987,0.9737386249069365,
 0.02436439881479957,0.3913421404181825,0.8815752751968732,0.9414287242757607,
 0.04791398768192001,0.7021664393759856,0.09391993207309202,
 0.1420663995734004,0.6956938591403127,0.8886047560209829,0.9286662231677829,
 0.03386876258911631,0.268961669561512,0.7160053664998228,0.0587424454008757,
 0.8918383996203144,0.7664662492309883,0.761620927912332,0.2724278914116882,
 0.4207656333276171,0.061174604836860325,0.058398567277050395,
 0.9487047406229607,0.10858735133703223,0.24733796410618414,
 0.0018037185107981912,0.164796326340364,0.8766635265819827,
 0.30563728610186436,0.8920267274743424,0.7197780103490554,
 0.30015901233617115,0.4468255573428499,0.46498634310328857,
 0.6445976424317565,0.9161670251663898,0.5781278650092366,0.05154356615232336,
 0.39321425032936064,0.5874230935386849,0.8756438099416345,0.7017316177412333,
 0.4850756368934268,0.19226084605397098,0.8278410538990454,0.8816316018151298,
 0.0930699208381971,0.8109192033033078,0.24382670216448787,0.4322462373586884,
 0.5368585854731558,0.8307218682280954,0.9331465230001882,0.1600782472689145,
 0.8944972583305352,0.36572703027994335,0.6169495082609051,0.8978768918443308,
 0.28995983828286354,0.13894577448829226,0.3373276740528407,
 0.23814850205106164,0.5177028161422007,0.39977175191311365,
 0.22489257763735193,0.9779141040374557,0.842024431241876,0.2619980293491959,
 0.4875727740425604,0.7824797588178044,0.6472624234801924,0.882311950396727,
 0.9397751260065943,0.7803153203357811,0.006383308790246289,
 0.9723957853081003,0.39044564196894527,0.8453085921667864,0.2842773400789351,
 0.03984370959296202,0.34600483820274963,0.024919872645233765,
 0.12698775236308435,0.8006204187102868,0.0077681358022128055,
 0.04266834717413914,0.6301845082813299,0.615888317407956,0.6871888016785412,
 0.2219448123400185,0.6104600851463136,0.719426160537902,0.92002536822357,
 0.029522073528443604,0.14314073011726836,0.7513407102322673,
 0.4953991491782249,0.33017839785388725,0.6269615487321084,0.9818680367591495,
 0.3480259968510001,0.784149063142348,0.11307606862166864,0.07481505047189618,
 0.2159069227491922,0.5388301151852027,0.26065635344966176,
 0.40821432721189355,0.94690500751478,0.13836164427734488,0.716398040664685,
 0.6159721348900764,0.15032384220132966,0.10735371671370719,
 0.9240724391987946,0.025135678854541932,0.6115993506327106,
 0.13940183059195754,0.516044806783618,0.8431030725542645,0.9562945080749463,
 0.5218728911756989,0.9606312538891322,0.14992781153536439,0.8307756028645407,
 0.3165027196893321,0.9970807902802122,0.9202581276338698,0.6803500782395129,
 0.24366442032965996,0.8394692863529476,0.5537606853089443,0.8401882994646311,
 0.6636636678140942,0.34367209439747937,0.9138039447418367,0.8113626553066381,
 0.5856894918486626,0.8005237052359828,0.21085762749040038,0.5263817026963968,
 0.6047140522652514,0.26225070735729683,0.7604519175916731,
 0.17410908456183316,0.6114352588726635,0.011107714265201296,
 0.3136274025813004,0.4929601095379166,0.6344775839514859,0.7509863839492508,
 0.5038848105397212,0.7774546321769462,0.673155968910257,0.12089189283802848,
 0.49977314347612234,0.06057812050341693,0.5419153181007719,
 0.11975565242300046,0.6217633422172086,0.8579015450987588,
 0.11957914635786193,0.2146213419470926,0.9381983082290808,
 0.22388413465254964,0.9559452920394204,0.12715297414744176,
 0.00810307934367216,0.605940154204467,0.8949266268048219,0.1623344487725159,
 0.5181376399846345,0.718413699754171,0.07130418904815783,0.10576103953841243,
 0.4963987160689578,0.2198820686452997,0.7512750952646392,
 0.022576767695692723,0.14430735025636143,0.6815897964849252,
 0.5435664468647831,0.057689194207285244,0.47232074064883267,
 0.7324477942427418,0.20097448782775917,0.7322022729328495,
 0.06842681455102406,0.06431528207680237,0.9393767502628834,
 0.5073288656263454,0.06897085961951144,0.18590146379648864,
 0.9686315243584904,0.4606775335926663,0.40393990852107986,0.5857626399903958,
 0.6402388114876676,0.5901679152549377,0.5696971634186148,
 0.012561893662820456,0.9159856111580975,0.7969185487884604,
 0.8188471194327824,0.046919558937297046,0.971272183604979,
 0.15252487198955778,0.7736191992834432,0.589145438492902,
 0.020220882172343302,0.3366296323532205,0.373777424535661,0.3842448557587701,
 0.30636616568559927,0.5393817294086972,0.7475370194998655,0.1732652018020906,
 0.8935516914839648,0.12549349194291937,0.24944679095725364,
 0.6273982260950475,0.43344945980642546,0.9665693947157992,
 0.21147283983223086,0.7532244648853087,0.08173369268167807,
 0.05138006530853256,0.1109050500924409,0.5101919367312103,0.4682936562342137,
 0.5947072763104161,0.7722397411597375,0.16368763750097592,
 0.17508245736060535,0.5821918457565058,0.051700553950508654,
 0.47686337640650167,0.05920480804037265,0.74338673941652,0.980383738882055,
 0.8495426227475501,0.12682074493811468,0.9512609714418908,
 0.006013967058863368,0.582762652043576,0.5928148699750445,0.8283865808682105,
 0.997360100814637,0.6442633058531453,0.058487929058536725,0.6770645090390593,
 0.9144420653737201,0.4518554604002055,0.30106566081298003,0.710604096920229,
 0.1524559841652222,0.3913344986245162,0.15225928221842544,0.8626253815239928,
 0.5683790605113119,0.5225045648164585,0.5617488238875372,0.38504112850220307,
 0.5304592435512473,0.028605941137884505,0.941457023247255,
 0.37291615891906205,0.18694682441285826,0.8861044606867708,
 0.2394191494560559,0.6898986470944188,0.2923093629234311,0.8571590947318709,
 0.0040074847899655985,0.03955915330857629,0.6746549550787586,
 0.1503097084207864,0.25394435686579797,0.10349172587494215,
 0.0655571588453846,0.7381818916646478,0.08828868729588801,0.9801502945736225,
 0.8846792098598254,0.6945156473151972,0.6745727439550737,0.34916143423587487,
 0.47352261282431474,0.2555943720253684,0.40989665652954355,0.657294655801146,
 0.95253418744016,0.8610751300627111,0.2968432811123177,0.11185885593916689,
 0.7239563248909444,0.5387604272665864,0.010015548491575266,
 0.3459878509808667,0.5291967290778183,0.6563950413549655,0.8611659231665277,
 0.2862664031101432,0.921812896225019,0.7520585105098374,0.36952888261561834,
 0.6488717939246363,0.5541397582982481,0.06341352819637236,0.7218397539974636,
 0.003523038325741079,0.802106695806147,0.32679390611723247,
 0.1506999335516962,0.8548499905611253,0.21526912256676067,0.6009913946205945,
 0.14599549518729493,0.7108094922515777,0.7934187432296929,0.9329169213891784,
 0.06912028661922065,0.10256584585143236,0.46068394711295657,
 0.6230168530381248,0.5252398778546402,0.3396543432437288,0.3478706561973509,
 0.015818503619635216,0.7240181390794946,0.3254240202872306,
 0.45426977169830907,0.48930483522232493,0.01729198479961802,
 0.6803985985541372,0.3723235596591684,0.33635357084048545,
 0.29948018706597823,0.9727672030378985,0.3398285478599079,0.1335378563955123,
 0.03036805216362326,0.3462420577039452,0.2760132012922677,0.9260190462568177,
 0.8265401506791711,0.11943701631690395,0.9391363093593035,0.3831583926961407,
 0.8993849907637409,0.10344145680453964,0.5764070612363016,
 0.13562229334197573,0.10684009214148626,0.33654386067773234,
 0.7973824073400505,0.6887443531644398,0.4454964285277935,0.2108452863795729,
 0.08817912686337093,0.22217449696741964,0.9185559073265139,
 0.7814050788080293,0.9463862651492527,0.3972157348585792,0.43317677114745123,
 0.4951184077925237,0.45814510176437795,0.5575242714704446,
 0.46671038467411796,0.5947384898410103,0.5374280858076282,0.9695023359663235,
 0.995659064344788,0.17665688638787835,0.8879147984616067,0.19670264701134776,
 0.799072098673544,0.24216713684248437,0.5008970971392736,0.6012148909122401,
 0.4163812376488538,0.8335797849038308,0.951086761680992,0.9888020420605763,
 0.14667943461825061,0.03564449887684917,0.005293350502022842,
 0.43920921378256395,0.43024457012873085,0.23157129270314636,
 0.38468743505840985,0.9114666935756756,0.39601144438468683,
 0.19018644043154742,0.41010059447982106,0.8799790945191269,
 0.2145591594352102,0.8913508763795015,0.5715669521348635,0.5652847212039368,
 0.4284967908666947,0.18080254883471936,0.6503179212824358,0.9700354954830933,
 0.7321524277232944,0.8880692135739403,0.665666082877334,0.5223437996218306,
 0.3973834973063559,0.9751579257578429,0.2887005172754429,0.8196059606776412,
 0.2745529084390328,0.8976276168481025,0.22827052724991426,0.7620803931821823,
 0.3644857774085475,0.49695418549265835,0.721987181029947,0.18968788643228995,
 0.1540572055120244,0.16411021642004942,0.6463330437822138,
 0.26445538388216927,0.9336025274981188,0.15607984681701692,
 0.4228118717206748,0.6088661043292641,0.37352128638391546,0.9003279665711869,
 0.27200218816308674,0.4933490028880789,0.30403120495223734,
 0.37279678066438127,0.679977572807577,0.5403681707723006,0.6324353893196788,
 0.08905003407310808,0.7064239538680368,0.2527510620623439,0.3304928604887576,
 0.1310861475695675,0.10782914848061254,0.015047443829497631,
 0.9862216751933894,0.8662289364243253,0.6769287203757358,0.6115385266655322,
 0.43739296506066716,0.944739506556088,0.40139676523058443,0.4087966304290612,
 0.5835588331192623,0.3063736098385048,0.2505465734357739,0.39845801171537276,
 0.5972774640802098,0.25953992144170734,0.6925840056051007,0.1695461583738066,
 0.4092540082850603,0.023450409140677753,0.7406693599995626,
 0.9613036624075618,0.6457811075714034,0.607389258649233,0.38360187231765563,
 0.8420178973669824,0.7863355742331398,0.9078526609708206,0.29616108765638505,
 0.18709799782922898,0.9911531759921799,0.7814562012056899,
 0.11689440260261186,0.7809508154176475,0.3947803563910609,0.6929627220882812,
 0.22954739717905226,0.144764597729239,0.8919764397342773,0.03905595401324813,
 0.26685659832570274,0.3314613417074821,0.16144886700721983,
 0.30631256918853544,0.24981103861633103,0.3898201364962397,
 0.8278453562730852,0.8376987324719872,0.6079748399412574,0.4145222735507216,
 0.7273642431364837,0.882985177876239,0.059748776910813484,0.3842230375073158,
 0.002808262580366594,0.27369161684226784,0.07911772694762553,
 0.16141330757791827,0.8041004545214285,0.9745618286846253,0.616033768318157,
 0.724853080538872,0.8504974752837562,0.7364920750413488,0.844893425565209,
 0.6935475147586864,0.03348369170233978,0.49827971856081943,
 0.5830451050644351,0.7144660973181112,0.782475305840102,0.5673232790888793,
 0.8226619460512041,0.21196115640558277,0.07880219355254325,
 0.03325835426643342,0.22670100824200035,0.921318373995376,
 0.16814105792373057,0.9869937938743808,0.5876396125338613,0.0843525870857923,
 0.4310517993630909,0.4408491621529891,0.6022594915225162,0.11508973041481785,
 0.0030266559077002375,0.5244515401948617,0.4318947636207755,
 0.10278638519754713,0.3087586440464257,0.2595759816995332,0.3743918475130591,
 0.8433806416080921,0.186676314614995,0.31824437777120096,0.49386130864268485,
 0.4544986030503363,0.9301419150682292,0.9981878850719962,0.8770867950519582,
 0.8363027762678097,0.6677702723265138,0.3933441030209134,0.10482301829320917,
 0.05449740312106599,0.5336262727184115,0.2473527166402294,0.7443970550221005,
 0.9412274011665308,0.7010943599706868,0.29569390466117096,
 0.48257310052675684,0.7061554913756398,0.5216574980961425,0.8357472938379157,
 0.4219914494017908,0.298204438764496,0.08207354824754964,0.5480159516492522,
 0.43278226981290213,0.28534461665713384,0.13380449058737665,
 0.6408710318617284,0.8723643060190142,0.14508994955121668,
 0.03352118746512878,0.4162928544589478,0.13892486130559079,
 0.5390707892113247,0.6064714077347655,0.7253600693000475,0.19822333860540242,
 0.5613894792918176,0.8433618504952921,0.4141718001249196,
 0.006712766258775593,0.6353522055153705,0.468785053905898,
 0.30469720263432376,0.27875987544424574,0.7968979525967161,
 0.6675588952740293,0.9831453942571927,0.6317247838623258,0.5074941216714604,
 0.4843840610631447,0.8637648544485275,0.5562314460753084,0.785303783504178,
 0.3400914247583233,0.5440471508186017,0.326854837983506,0.19923580536228092,
 0.5858527034419984,0.5393117385353808,0.4464956048467773,0.736206960540694,
 0.735231839537581,0.5686365880888622,0.48765687537616254,0.23142098536960176,
 0.5015041783193591,0.9592073033835057,0.5250149795297891,0.8000845672820234,
 0.7454113398561178,0.1729910853354053,0.05343935319990023,0.741488669718248,
 0.7670682461214839,0.18680744114408698,0.5615105671548374,
 0.08251462169123647,0.8158953865895239,0.46864360131347027,
 0.7041346038874718,0.4382715985996959,0.04746477113063219,0.6550056031597238,
 0.13996964047785887,0.3850293245958787,0.45670867317168773,
 0.9426503487064937,0.0036232192080072334,0.47927597047456094,
 0.7943849011680014,0.6035103535571463,0.5810605405574004,0.2919514643810375,
 0.7426481441106254,0.6171545420509299,0.8379496017787902,0.6132116113388143,
 0.26692731903992084,0.6629235526783459,0.9011761221256706,
 0.18697398143244426,0.31218801558677545,0.08412194757595048,
 0.08938835460534689,0.2885248882906286,0.5200517885945319,0.6988315757253629,
 0.5779338308008349,0.5102854666584079,0.6730352437903498,0.18080290370281427,
 0.20605432603814622,0.09162173087510306,0.05854156264909072,
 0.5713187716789472,0.5021186881435629,0.9618401031299378,0.31828017843128964,
 0.999750980885037,0.701373696949382,0.7519135783260813,0.6506345213161563,
 0.5128368307854537,0.9938640141535334,0.08596751897467048,0.9866870433017675,
 0.12057056093518304,0.04114233513395171,0.5119837915699794,
 0.3765257781704703,0.8855089222006246,0.25750506566047193,0.4024260841638809,
 0.9223141468427015,0.7367489234973216,0.6073053023154551,0.994533759975586,
 0.4950391246844952,0.8138852466557445,0.8783366790277586,0.08064161213234156,
 0.19705587346694375,0.2128613477873298,0.5064149045214235,0.8436117978193574,
 0.5423196768205177,0.2612497963314606,0.07055798625147192,0.5220981633023278,
 0.01646971129279995,0.7318328446575513,0.2836465126435481,
 0.005547584493860946,0.42802640099763045,0.09852358337351474,
 0.38889115072615044,0.6927904406080797,0.8692695623760321,0.7353344453017879,
 0.9289110414854114,0.04623900973291717,0.8711291019644429,
 0.014577808356915689,0.927752254153161,0.7654658393141813,0.7578137022458509,
 0.015923200729674458,0.11978986204302189,0.5161921011795111,
 0.45215778825119224,0.8863230065000327,0.5857208829724567,0.8980050151574867,
 0.005005295853345859,0.48672626244095474,0.0029620469860678345,
 0.26286961018798827,0.2441746677222979,0.7636561522720589,0.8072696117877401,
 0.08488577755970672,0.018552082599329167,0.29932125118955644,
 0.2106132329640399,0.4102488713504564,0.26893758867271467,0.6518473340311336,
 0.9196368877282337,0.5504898276321146,0.6302356369325032,0.8705518337100638,
 0.9379041502582739,0.7392831395694531,0.570082255230689,0.31125030956221,
 0.06593140329549474,0.9266575510433733,0.3121001979364565,0.5098743416922248,
 0.38819794724558987,0.06656097559400709,0.19824255909038424,
 0.7673381710411017,0.3646800709527901,0.4643847850051205,0.6473692254641469,
 0.3122940671443061,0.08143051291020309,0.8949429947377258,0.8598185750024175,
 0.21076194719208352,0.15210176398112607,0.6297454118982105,
 0.8044048625359496,0.35415389187609714,0.8134495233467162,
 0.07644399479089148,0.06487250430455005,0.4885241927822245,
 0.00171617685586134,0.8972874687441403,0.371376023245494,0.8275680301650156,
 0.9346570066101771,0.42633033164569634,0.08810816188002635,
 0.02500093094429412,0.9298161783632661,0.617995264636614,0.34206806760242736,
 0.44563339582412465,0.3795622003963832,0.27798042956516633,0.811205899165564,
 0.15046746589183146,0.30747994592046657,0.2085936455758295,
 0.9213651150637676,0.5060405534614202,0.1602012506357301,0.06266630645462024,
 0.2917550061213873,0.4632712038155997,0.4646539950953479,0.0979182382814563,
 0.05362934008055986,0.3777804036309438,0.5675072032720215,
 0.37962254694764663,0.7718922035426106,0.37665527359826134,
 0.4798495624697814,0.009339040714476665,0.16953039031919093,
 0.9161102004907944,0.49207714949365733,0.35101414304855616,0.798698436598491,
 0.07721567490087544,0.22167969449190272,0.6938088409158191,
 0.38079897555281594,0.8267648041702356,0.23794628987005484,
 0.3085297876513362,0.20994486966169834,0.8944579328045741,0.3166285681953449,
 0.20837611704246128,0.8569490432734707,0.05589097146983957,
 0.21736651745062785,0.9039746851649608,0.5688977097669694,
 0.09629891533640145,0.5320563290637953,0.5301540519181448,
 0.23732196410808593,0.80096787664167,0.12625876081703502,0.6330752730828924,
 0.4695133272975245,0.3610526331793056,0.21466465381280764,0.8316524824561994,
 0.74146570079756,0.051781505336623845,0.7506960335166392,0.23162196669272195,
 0.15244830604784332,0.4527303309748243,0.22222243255257546,
 0.8380567112656918,0.2164930694886109,0.6744593713839712,0.10955874278054611,
 0.8516692425905769,0.6658252466863154,0.14795011279523318,0.3634877798933509,
 0.090689018494297,0.8001203251821134,0.8506774678389875,0.16209182103922914,
 0.7516201530494505,0.3881803281720647,0.8433372885827932,0.2111523110416167,
 0.8072985172614762,0.47896025737697256,0.2389413893785246,0.9198767016095396,
 0.6116839817930797,0.5248814124989317,0.16943789800540943,0.8958422370921559,
 0.0075922988119151436,0.4582540733774634,0.5209705826328537,
 0.7946659353489581,0.19548560762200273,0.6878186133188345,0.7761224009435678,
 0.9556742407006941,0.8095277756465874,0.4729226315547289,0.22655515692998207,
 0.7821434334971661,0.8007874228974884,0.7608322934423437,0.9616463231988467,
 0.3715360677934796,0.9905505347225099,0.41993969474203086,0.5028503411775266,
 0.637312063147323,0.6680869762126905,0.8234340006704863,0.27126143870935826,
 0.5483410427655602,0.7636299837412744,0.4858512109237635,0.6699494998358134,
 0.5016200130084127,0.7557871816853098,0.6032872064816123,0.10209263936137425,
 0.5617695814047581,0.7189899323173833,0.8778331999185384,0.06641148752831705,
 0.09639246569508952,0.04784121711476641,0.4017226109021662,
 0.7562581787200205,0.7372445682312606,0.1334703262051179,0.2509791321220851,
 0.7683932459998701,0.21325771334258425,0.3133199287459556,
 0.08987692894698474,0.26773366241532826,0.5885596595893854,
 0.8247690556348919,0.6065951149837776,0.7588345817677543,0.8471934404149786,
 0.8871077557006861,0.6623813715293078,0.7322207409885892,0.3041924942153432,
 0.35086638879277476,0.18197047495892393,0.24662005393933129,
 0.09441264665081728,0.21421367729702379,0.817300102958819,0.4792834866913426,
 0.5104226337642891,0.625726031591756,0.5987664957240127,0.004113062762807873,
 0.16396048891838122,0.33240522322194754,0.15951784008550696,
 0.028589728687801186,0.7651001061383305,0.9053986618114392,
 0.7054227253277346,0.4106826661428802,0.020544400762217196,
 0.5156893371789131,0.42859910595889295,0.1446981364326161,0.3095935005682625,
 0.39171914656043805,0.04909957162534995,0.4473476250777073,
 0.8684439556995397,0.9212806388629757,0.26686282718115506,0.5361900924480933,
 0.3523968568880379,0.8271583526802095,0.9017608893904613,0.8007794125982921,
 0.37164604067290496,0.03412502887652391,0.4294399283590512,
 0.8398006697195064,0.8604471663263775,0.08884193387716999,0.9472201246718537,
 0.27413533164404214,0.010150495733199483,0.5700366648538336,
 0.1824990157913976,0.28188406934885823,0.23327225941324803,
 0.9402156950414846,0.6364189205430908,0.8030725064596389,0.6999414662818898,
 0.9593110382705168,0.7142324244122403,0.09685897114893827,0.9901302900259251,
 0.6918601196572005,0.9917868374785042,0.5742675415614287,
 0.053941673231172715,0.8736358766812298,0.019634934201545717,
 0.6216638313579332,0.0727075796892136,0.4350358086581968,0.5498852358597599,
 0.5419384088802983,0.5432025584368776,0.2413356753785939,0.9072149831935991,
 0.2117757322746261,0.6389593210192633,0.8034862789453068,0.6343559604772637,
 0.10002167966549236,0.6976108534723764,0.4501259384418438,0.5902372511985755,
 0.9647890659188147,0.7589034030705698,0.9656979166892686,0.47453752471055766,
 0.049486058228712104,0.34851581070412785,0.5517012623328525,
 0.6743129412523128,0.6037617995615533,0.34936851726362683,0.2915982492445546,
 0.5420845433790575,0.3192074373023932,0.058009106631037644,
 0.7363640854913737,0.8252452028357605,0.9520101294432265,0.5011579721409546,
 0.007127836781231522,0.6721154814379748,0.5272865326858558,
 0.5422989959544385,0.29405613960550514,0.796945453015039,0.7910861027739823,
 0.26381159631671725,0.24763070106505713,0.6907110909347077,
 0.45250804174503345,0.33081423081581096,0.9330489272861453,
 0.32802890135116725,0.6842104904311513,0.47240269713269334,
 0.2533072738574128,0.7880212618578253,0.5684720918474101,0.5659479006110666,
 0.42835477416133094,0.9652792011093279,0.4884117252240261,0.8612000396805145,
 0.6181811573694015,0.20802383080659137,0.03528029250236853,
 0.8419398285643276,0.6332294805058863,0.9194102079385781,0.5089917968237084,
 0.11408950652001304,0.3475757119434707,0.039382222068442774,
 0.9360102643243103,0.19227597619399672,0.11632037024517938,
 0.8098734269533542,0.6226129950406033,0.5839423587419296,0.603603702561197,
 0.7672064084845635,0.004628478167183481,0.2921329941746118,
 0.45238437714945656,0.8542619862536159,0.23190053412333,0.5305717881383523,
 0.6859202681227771,0.5693947575765986,0.43434532800902403,0.620664412829759,
 0.7621455584886747,0.20000696069145862,0.9759220576155473,
 0.48555790965115075,0.9360158426197396,0.27336880578647116,
 0.6091657231019212,0.46893669250071557,0.9561058122665896,
 0.12605384588211832,0.9270253147463713,0.14744790421896892,
 0.22089480871282674,0.7359129408848688,0.034498004459213694,
 0.20270000813210898,0.5118446963939827,0.7290353127099249,
 0.04669383938308236,0.11910850267207629,0.9108713205868928,
 0.8318010833255095,0.4721624815410731,0.9636289021217445,0.46899287834388403,
 0.6536486430231384,0.63053441303091,0.03354260424106026,0.05053653796822799,
 0.40013967328577327,0.5671351180103072,0.18087078875087959,
 0.8686300302239177,0.04856859431209648,0.21083178583000073,
 0.12672202986382364,0.0826675768313212,0.3197989778631062,
 0.010313758613954338,0.3363830534782958,0.01749226284062788,
 0.6343381305974329,0.4922225105459105,0.7162519565682494,0.07737855951116668,
 0.4087023788034104,0.03216105873699204,0.06755353221465843,0.989649980167224,
 0.24216096866761827,0.1316445959419681,0.8122841911811505,0.5871148436934663,
 0.3843494619181139,0.09509859105283647,0.9918486163449368,0.1490917833138531,
 0.055633308012113414,0.5333207680721518,0.4770796271059077,
 0.3859642034581645,0.3785621290323482,0.09007638911319082,0.9398124732806598,
 0.009433010529426133,0.9737291727022024,0.36466567429041497,
 0.8512058934445622,0.3638237620613122,0.1222478047241996,0.6920402665582968,
 0.9962935595834191,0.5342967953146385,0.5566766532933213,0.06917479917191571,
 0.8197874725798426,0.7222553521927119,0.6211270021051181,0.04801476128501536,
 0.67071263140362,0.8434418285333095,0.8466978647724939,0.6024442989088061,
 0.9697350571673592,0.9002952198188374,0.4676524273563668,0.20131682677367346,
 0.6472869436570851,0.36962751194921517,0.3237360867668089,0.3847827360496092,
 0.0703396207303324,0.6413682198166003,0.9343639371060872,0.2647915031921737,
 0.35837130112011517,0.9578481693522064,0.9378019118324523,0.8384576921177239,
 0.4796285335120525,0.545106759633923,0.9301811910255066,0.8549101004114763,
 0.2407871885745443,0.9169538040287006,0.8696431830068758,
 0.009239483963263018,0.1654518939035029,0.5077859183306245,
 0.5737769755317396,0.2704167022066366,0.24703618981422026,0.3337084750053554,
 0.385075248510917,0.7184661009648586,0.7015797221325829,0.2117423236588617,
 0.6147460571789225,0.47109793621675,0.4600908041116327,0.06320151040519162,
 0.1786334926463754,0.019299447678376214,0.8478637360865027,0.442081301837426,
 0.0948444659418074,0.6981437873474697,0.9710935781696719,0.6145663233921709,
 0.774114022682933,0.22382189824042242,0.2548447512714609,0.8577526941531026,
 0.11425316466030344,0.14462887145564218,0.46126922917524216,
 0.1349427403408895,0.5267630155609313,0.9600873357672932,0.22602489439178264,
 0.1343880449148509,0.6387295735206718,0.10298156933338853,0.7210054584958995,
 0.28970183999193777,0.23396435892766654,0.26376126698993285,
 0.13393833125143773,0.625736224753925,0.8053298556225671,0.23441078065383825,
 0.2681786623648581,0.9287259324820775,0.6270213209741091,0.41685782005254735,
 0.37028666305936686,0.3986209060961776,0.30026314678361243,
 0.9991026774927281,0.7536310527000469,0.2293514340639129,0.24366178464700083,
 0.7848663451740322,0.4605988842659797,0.7791633381953893,0.8653995864079246,
 0.8107734446160655,0.6016129204811965,0.15029236023979642,0.7152531730532018,
 0.768268986030148,0.48394653983356417,0.10596680079268506,0.4087586312379262,
 0.5295140106285728,0.38698498650754176,0.9035282548187386,
 0.14386367294589708,0.9968074182043998,0.685158854160659,0.10473082123985988,
 0.5261759900098546,0.9137559887781466,0.6793100312691887,0.7287522824249337,
 0.5014634486230412,0.2018056560513608,0.22138593937035467,0.8815576444536483,
 0.017329946252562234,0.43183007462792067,0.9041226725919994,
 0.5008306441781443,0.6877367775824152,0.30151363988918223,0.1539888680192325,
 0.06724600348381016,0.16305988821478234,0.1383196198837755,
 0.6645721558389303,0.24349601473748173,0.18698527708678814,
 0.17212889194514225,0.8360998047343722,0.4176073325880766,
 0.015347739561576423,0.8071573215066055,0.2905530950312726,
 0.21218654150484262,0.9078975581862916,0.20658360749942795,
 0.7015168735771329,0.25875650879247336,0.49789627983768747,
 0.4201924943680606,0.06608222604547198,0.8792441776606686,0.5283652794735774,
 0.5383813989687614,0.16348103669587954,0.19290049185367608,
 0.5946566509578997,0.6557701688518524,0.7102669711847948,0.7675845235695063,
 0.20877908491152386,0.6502320532264593,0.19725749254548575,
 0.9292448674507644,0.8218165213629227,0.08054543324875107,0.5662215602177422,
 0.6841841482755692,0.02329905106058927,0.471463702118724,0.7678907905640839,
 0.2078043397716769,0.37904211378839636,0.734986266126635,
 0.0034277865190632806,0.9824646184930883,0.7267520193929106,
 0.056918975293378304,0.5650099061027727,0.8320893972295585,
 0.3312076331245841,0.8701085755120976,0.8832231753906614,0.22576535982982882,
 0.04408647604609517,0.7243063920054976,0.20482961903280272,
 0.23683532890237857,0.006209339346253784,0.03202962669196274,
 0.008315753251607494,0.48601048331926566,0.9114889978158395,
 0.8725263444325844,0.47010333497608836,0.6344040792520935,0.8046567495693875,
 0.05452919710966908,0.24522087427599126,0.9733815366502584,
 0.29734975075137116,0.21423948431688422,0.9221189600465707,0.681141742550951,
 0.03780446039934671,0.03290443941977017,0.7651048768871841,
 0.9524215188898766,0.5417378135427291,0.4340362891521796,
 0.017358520863180682,0.8766825014774792,0.9749307457973707,
 0.31653142832823744,0.4174804304943349,0.4363827450292477,
 0.05452128343163387,0.24749017438379783,0.25848751578919316,
 0.390031370799115,0.9290366876392605,0.7207077885132511,0.8132057429024994,
 0.35065187110306495,0.4294497385894146,0.20081094733355542,
 0.4269460710987244,0.9605244891331903,0.16944341273923025,0.717600340374784,
 0.7156090780843325,0.5284200978780194,0.7405934834392245,0.9476417692731806,
 0.73466752158539,0.1279161594778926,0.9263783855303722,0.5609956256702061,
 0.253562337262246,0.6308532518906724,0.8867196430257851,0.6535660798149228,
 0.12041396843224417,0.5888894850172012,0.12028194677157644,
 0.003761439197289884,0.8124046067518184,0.06234063800842493,
 0.454430986243185,0.9827522361839991,0.585776605046581,0.030228270560526238,
 0.7783674755732098,0.840212649389363,0.15716699333558637,0.3055997512169968,
 0.33835549436480583,0.03869987407703046,0.47760564200313815,
 0.2897931008597756,0.17126277665387524,0.13892620223590046,
 0.42945822070763773,0.7218102554794362,0.404859767526467,0.536802211213165,
 0.20963720314591783,0.1665326999834693,0.9877363557293536,
 0.44877270983771234,0.23990684202535972,0.5127396549862554,
 0.07427222378276921,0.04247278233979079,0.9478025341988765,
 0.04373934049461414,0.9316833868612228,0.7444159528117515,0.6072592397256646,
 0.3254387600863965,0.2809826254760198,0.26447751239941253,
 0.49218950392963345,0.10845703087953884,0.8805720304413692,
 0.6401922941750559,0.9231007749377044,0.6723650963594145,0.777728499484101,
 0.7768506204870922,0.5752752582038729,0.40566618552345124,0.4344428675281523,
 0.49188623103309703,0.10204932058783212,0.2792808400668795,
 0.08482680334938997,0.23200998308700138,0.8333600521250593,
 0.35756574330344426,0.9140763934296082,0.7188202188593817,0.2508396219654958,
 0.8663063154979052,0.3426193448465339,0.5073349500646062,0.43460951701092343,
 0.7570889225592481,0.952086313182478,0.21169973333634307,0.6528310361435344,
 0.6314684421516913,0.8995312683047197,0.4294530345655888,0.6161926434952113,
 0.04929112933203039,0.868349503544851,0.3526118352287717,0.3009686688333545,
 0.18985227586576225,0.37607144332599296,0.11681537406932252,
 0.8581950466567897,0.4715559315073886,0.55801427220431,0.45963112997303834,
 0.8118951976487145,0.6573426459843624,0.9915570206905475,0.538136312950895,
 0.3207051564072346,0.3019156370718131,0.5494110500545506,0.47212350705999506,
 0.9388604852080302,0.735514630923792,0.8763635844741589,0.747337640527182,
 0.0635083516247843,0.7333136699108135,0.1340451234542024,0.10123024967832328,
 0.683671784109329,0.12346342254922216,0.5704740527778922,0.1349545274160251,
 0.504943210793436,6.876634243528912e-4,0.817276300089122,0.18263278646096415,
 0.18054180343553172,0.7156971956734256,0.6797873744599039,0.7867834387762022,
 0.34364439251512235,0.7030703350150074,0.7522855287858216,
 0.25779741364917674,0.2943451087348494,0.2207656418372249,0.1443842440198133,
 0.7668188362127324,0.4653675544522159,0.010748983028564196,
 0.9358579545406389,0.739726020833424,0.3910219669205732,0.6848576606850987,
 0.08119942776082056,0.008066472505633526,0.9156020196863656,
 0.1438217826698227,0.09235369755454148,0.5427596869819876,
 0.44922539014506324,0.7293716574047361,0.5588345247486808,0.7797549714062001,
 0.1122655417279923,0.47975025706627417,0.8653687848443243,
 0.10554242920490153,0.6082811473972828,0.18583496658692567,
 0.8735702250279046,0.7635396895516019,0.3204756671852591,0.7497296339577298,
 0.8687757909486366,0.980126282377456,0.5699323978380649,0.5869456299984174,
 0.7102600364069521,0.14886463136241646,0.6111018614877343,
 0.46846676678286947,0.3988269446120096,0.8881277943984822,0.3166675678623686,
 0.9473997202240529,0.5891651318482902,0.944916537992521,0.8072948491640448,
 0.8852875752057077,0.6154981827434333,0.04682607607696665,0.5932000889993041,
 0.23054417077748124,0.09613194707125383,0.006157058179993902,
 0.38660784842017304,0.2979841907165398,0.7381012126516519,
 0.36368235770742796,0.28453696174162646,0.0823253698085542,
 0.6089952898603822,0.8986972835325127,0.48799754177159493,0.7379217624028482,
 0.10520412126819845,0.7770460975466261,0.6767873549393928,
 0.31847342503467013,0.9770913897245912,0.30343982781182455,
 0.8896127481545834,0.9530935861052106,0.12825366854452114,0.2876902431267123,
 0.31530408016696354,0.3762247057538163,0.06754592791047698,
 0.9337932116051635,0.882469579129335,1.6991019034739985e-5,
 0.6647423706509491,0.06476647721138573,0.1486375272184397,0.8172625253913964,
 0.29245285751121464,0.5778667948786025,0.6804652787440305,0.5893244998584499,
 0.3140147309962059,0.021928207583881032,0.2568322355288626,
 0.31686553211753155,0.37022003495579536,0.2785565395283791,
 0.4518181357458817,0.1418814460382547,0.694892248567902,0.12190792677061624,
 0.39419561567621453,0.5833200231980581,0.5784959952831886,0.3738497479459064,
 0.027183708290206665,0.054455917627550865,0.8281455940733942,
 0.8479039872453212,0.21563629190988665,0.7735547032974449,0.5813267071487709,
 0.2553226480865094,0.3552629962680116,0.015428212132048591,
 0.20237248777116346,0.2973589733236299,0.6964918056897665,0.8440438417981155,
 0.7109701065900259,0.6749816053690664,0.37590860741856535,0.4785045928034497,
 0.6615795651628242,0.5707617685290604,0.5083065649766103,0.7081830614097749,
 0.9750183822070568,0.4967091151173859,0.09545713768838193,0.3875618814916765,
 0.2758492373438637,0.6664075089717987,0.7388259600089608,0.6786505400252785,
 0.23887463396784514,0.378237691972548,0.9667144528233991,0.5433217246367041,
 0.5294746202328895,0.013597110554957381,0.6294629720841868,
 0.04085349309485942,0.9253828934645724,0.0471559098184815,0.3019528738992938,
 0.019745262414332276,0.37449841439816867,0.655740281737489,
 0.8126445339105411,0.6679785516641384,0.8246011164922888,0.5626910920419784,
 0.6515737306617735,0.5588563006199001,0.8662442612680024,0.3356285334017811,
 0.0653621541497803,0.0749409541686169,0.13654227980243594,
 0.45838210001108615,0.2786641608436582,0.9206500307840664,0.484826957243342,
 0.3732563367427847,0.8877127009642874,0.6998205416052303,0.5331281423614386,
 0.5494951783870525,0.5681805690072126,0.5049055522379331,0.6068928137685111,
 0.5935121138773986,0.1760587867483857,0.03590364339860341,0.8334078522694847,
 0.025453032063220138,0.6875045597942601,0.8502124489689198,
 0.23740345985572708,0.6418192003981137,0.7084910281927354,0.7113489002887461,
 0.9949845114704947,0.9379150492117247,0.5543169751622236,0.15731364020016758,
 0.9506361604523617,0.5934169245522052,0.07125935347612611,
 0.055234140299823875,0.9602351882741348,0.758270800720803,0.377799825350124,
 0.02224803211912152,0.07030942304561649,0.655160783375735,
 0.48159375818820005,0.4681767096972306,0.6013738522996951,
 0.07172757398485108,0.3074935622716283,0.8273569722760664,
 0.40014621063165845,0.430137519621972,0.758422066198767,0.6013146957138105,
 0.05265598339817168,0.7753861939619595,0.8140800839743071,
 0.035639969292580065,0.14250353836977314,0.0030831045287236236,
 0.31421691260011375,0.6928796416741587,0.4880481097778747,0.7492506933891927,
 0.15623121788988314,0.6229751017271844,0.4713417309340807,0.7489420289497437,
 0.16096926657108612,0.5390019186304026,0.46094185168260093,
 0.5074746324770224,0.45852086779694956,0.17504461181074438,
 0.7781414289598034,0.02809937906244897,0.4758706330530923,
 0.24587198621898798,0.34096834453480973,0.40478864107645796,
 0.046562694039042896,0.748609664269666,0.4535088741011959,0.3124822592451696,
 0.02474065048288354,0.03347463602502543,0.47266406830106666,
 0.2835331435925692,0.32971018174856803,0.36939883067018875,
 0.5355915335623865,0.2959302976040473,0.8743715487604596,0.42981012847377276,
 0.9471126306233034,0.9541969869332014,0.27281662936826867,
 0.051894038020004585,0.2707255237394883,0.24528359452329107,
 0.46649800438326305,0.5558197218480512,0.9590311754856927,
 0.41350964193433803,0.5532251378368566,0.7633870928223851,
 0.48382039433530954,0.6372237514828931,0.32468525995501574,
 0.4644837348108378,0.1444508960171439,0.43491301948592875,0.4301511991250092,
 0.1012044500827135,0.929339568011326,0.0035996879836852624,
 0.9839802174896114,0.2276375387263181,0.4993457463524654,0.7750741519506249,
 0.29490535950186314,0.5593742857862428,0.9498517612024678,0.8722018377422092,
 0.2368298891879923,0.28684154304389564,0.964654510459861,0.7283762764601889,
 0.47371849973535285,0.7009580568616222,0.3373561164402713,0.8478507283403487,
 0.8911344694745963,0.6299619555396535,0.13515966630283793,0.5345327060238481,
 0.26711549871917706,0.7410730313596039,0.2514540922611471,0.3760007529387167,
 0.8664403586138335,0.30017570475259836,0.2646235156113178,0.2098117170866769,
 0.9559320325717937,0.3717595368930362,0.5639089796166119,0.0770842794523362,
 0.4006502010140446,0.37889969436017834,0.4331070587758914,0.491367497711409,
 0.4366993268850755,0.9457378257422598,0.486245462639043,0.7115762154622272,
 0.28909468079402423,0.21418157515881653,0.6167589433188756,
 0.7868791721087183,0.9076881090116699,0.3169103714668693,
 0.040497049147279585,0.0607183498353413,0.3214642755631614,
 0.44155501444514766,0.34786520653911546,0.83906477672623,0.5752806960687441,
 0.8659535365021029,0.978288146980404,0.19783980107106292,0.9796927109910708,
 0.09880437536949116,0.5081114910052735,0.041894813328140046,
 0.2342167088800935,0.7553352556961678,0.652021807588065,0.368320786308607,
 0.5275111292730483,0.537941101050052,0.4038014547359696,0.14698672776614286,
 0.8514750133113859,0.009207397298704234,0.5332977730512343,
 0.4589002628406388,0.7292518336501973,0.9979621461265549,0.8977203536206205,
 0.9498017628376105,0.9170663945803924,0.2037230272613182,0.0764217374147157,
 0.9927342288778176,0.23785469440976492,0.8494028036143721,
 0.31949374682890386,0.08245868611568974,0.7864662341873725,
 0.8324951405742669,0.5166337890032744,0.167478743079569,0.6081948864664216,
 0.44596855861062,0.793443173599673,0.3713783630524067,0.9748678257516078,
 0.22411497736983566,0.2848425635995667,0.5620091626420701,
 0.09482270167451534,0.048404289278308865,0.8730526628955161,
 0.868289748038261,0.21630117205014776,0.579749289143982,0.23488370366045697,
 0.9391420394974511,0.6822013399759084,0.3794261639876917,0.48698905666795567,
 0.658864596781771,0.8596918073931031,0.7225496053730225,0.3785442017678071,
 0.13148757995335703,0.05331910144520602,0.6365545647709886,
 0.1945342705683909,0.6006106323637875,0.7061336038666073,0.9934837576334677,
 0.5660207415217497,0.6868787979813984,0.594120889018138,0.8855060355481942,
 0.12708866039226663,0.9819091910426015,0.777328970321205,0.6819345351843707,
 0.46174740981529006,0.012810833560627843,0.14356850682495392,
 0.4370958568610861,0.18768015488857748,0.4858130022114441,0.4313221411341601,
 0.467117242082853,0.28136489502467243,0.5455844207851036,0.888149877089174,
 0.597312961718127,0.2722167612197445,0.3378743362771025,0.2774707828229296,
 0.016014521779651014,0.5799657769120666,0.8378222193439466,
 0.24115837633897863,0.5303655364863663,0.19619133198856165,
 0.5837657538030514,0.5311893986242497,0.9317069051227946,0.9292711250732146,
 0.0273164827698128,0.6602163862325015,0.2558704441371603,0.8703452196649384,
 0.8449933061993451,0.17289271829520614,0.9249158458168465,0.7481680138982729,
 0.5214150759826943,0.30838894661244043,0.999439945799308,0.490337690548277,
 0.8654016029202979,0.6332569233033826,0.9233582303858641,0.8978022352773105,
 0.2633762201958012,0.3792419013360355,0.16536134059299878,
 0.08954990384611372,0.3282432667635181,0.37974358014411635,
 0.8915052462365944,0.5113339908813548,0.6766973859923323,0.7759137862964841,
 0.6656977175158838,0.5028536315827898,0.7259391979192344,0.380913085478748,
 0.11465179244668999,0.5495438171361601,0.3250742892287585,0.539183368288432,
 0.42650401502945057,0.3820327108136644,0.015242562008108806,
 0.6111047814663876,0.5729930920634462,0.6849564927859785,0.5109704981134777,
 0.8266726105348371,0.06216487967248896,0.0917719610775003,
 0.32415416114483464,0.06318277238727499,0.5726378468389184,
 0.5799894671576838,0.2797275776287005,0.9198730118837681,0.9857430182904693,
 0.0847794615966343,0.11948522689441698,0.8829929971665281,0.7046884362172952,
 0.2156162200786763,0.389942709653877,0.1540227307371368,0.4689938813596283,
 0.07977817027129275,0.5409511805740244,0.3850866834823262,
 0.09847152215863686,0.41031290124876907,0.9311626763288996,
 0.9780364162855095,0.6794888744793701,0.6460379489385396,0.8038265641303176,
 0.28965020458233126,0.3350789170369628,0.0871941994136396,0.9991981869250778,
 0.3076800004063177,0.7075613720966298,0.18204376120883414,0.4847619658642133,
 0.925711140953971,0.44197437357456937,0.19846036771105235,0.8827490955505936,
 0.9479747510119494,0.15335750370948753,0.5635066735984583,0.788786758953192,
 0.30865264165286665,0.38498662972092845,0.8374534834502688,
 0.9198854499158085,0.41826615405054135,0.3896983619680261,0.9631366729700499,
 0.06579883281458743,0.2632420603695569,0.32204720279165966,0.19503817136098,
 0.7369624889667888,0.5556321887914953,0.6220506968162591,0.34893256439429654,
 0.26457377258544035,0.32265714605905493,0.4192742794807667,
 0.8513916940816147,0.9298751097429743,0.8273358544071707,
 0.0052192526279557505,0.8051809607010572,0.8887525664118533,
 0.7119336885050691,0.21262262081582495,0.9119610759182722,0.7697051091557294,
 0.45500125156262283,0.9619053507590483,0.1957423731654987,
 0.23712607734288427,0.41720754959260686,0.6354987864340437,
 0.12628416667779563,0.4996264451700909,0.20388963211893218,
 0.3077106994490042,0.9565563164693016,0.90043176715922,0.17818249651881723,
 0.9700243731579072,0.7996843684021688,0.6285216882866568,0.6756104831389743,
 0.4489493798129015,0.4662292196794181,0.08252255818797549,0.7300525945070464,
 0.20384826353923258,0.6366678709505066,0.385640902233487,0.27912331962760895,
 0.8280361064017439,0.28334944289330866,0.625160430596758,0.5375947564633741,
 0.5393807718374743,0.468611852286686,0.656446877972462,0.6812583560293415,
 0.11272588098298353,0.7712182864789769,0.6340663168796722,0.738875263590324,
 0.986740040059144,0.34405803967558246,0.06044340797482883,0.4277458301498235,
 0.0021915659129477794,0.2708876846262017,0.5800946744209822,
 0.5198879771403313,0.5362073122109063,0.8414161653991259,0.8519041417932949,
 0.17655273323780618,0.889233107709038,0.20765066098681573,0.4824679218706063,
 0.5306072356725786,0.07636649170567322,0.7836982041097438,0.7442124120655312,
 0.8409002924949263,0.8732703739224557,0.19587466233582573,0.4698532835080481,
 0.5227127679533085,0.7604846767055747,0.5228608119531222,0.7502495356433293,
 0.7619283377852747,0.39295501900661955,0.05854405337167756,
 0.5803509643993584,0.8355863672364603,0.5416104612518337,0.8092996118981406,
 0.3253281326938421,0.20936739821943684,0.18541659556998424,
 0.25029186234600864,0.22401136261672794,0.394194569135474,0.9250355346098695,
 0.7134722213857316,0.4956323353679126,0.9332407443300889,0.7266101509272809,
 0.06178950640396552,0.019192203122946006,0.5547842824808353,
 0.611943793304245,0.18886505425650113,0.20735330580872813,0.4135876386032926,
 0.3932255521960284,0.2968511904823984,0.13357639934035903,0.3464427545384501,
 0.9916947445446008,0.6002691709480239,0.17716122145196667,0.8151472758497119,
 0.36117007414250046,0.11567447710453804,0.06393702417717517,
 0.42168121169345607,0.9203073970871442,0.5804495336130895,
 0.22910577085918826,0.27613379997947907,0.8572819836943202,
 0.2088570858007638,0.6535179444923318,0.9602807931157175,0.14887222346574958,
 0.10765723535329974,0.9915301223381151,0.05629957834718202,0.765872931571204,
 0.8130250834556978,0.5021951910661488,0.1627385072176163,0.9373004506761358,
 0.958970450797131,0.10518343608664094,0.38042775091661896,
 0.17429996720587304,0.4838260798020916,0.5919298443931527,
 0.20767715370840645,0.08504779379570093,0.9563436074775078,
 0.05491202584780841,0.7687189410828986,0.38735298615232094,
 0.6095898350334488,0.28180750075522676,0.8358748636646521,0.5465990679007657,
 0.9463172343587055,0.3743596254856897,0.8089247544662456,
 0.019954609584503902,0.10098286452087157,0.8242908966102042,
 0.010889872666185796,0.8456009613407023,0.03406962647772327,
 0.7429360601207263,0.6606939828624413,0.8362616654424565,0.5909170248962478,
 0.5935419429361661,0.060519129400996796,0.5985868016197966,
 0.007755193308726205,0.5734194115515359,0.24096143967511774,
 0.1899165582172444,0.010997125838362765,0.668980015427424,0.6780840781465098,
 0.45725291058554984,0.05135662676872399,0.6556533670004394,
 0.3280271661390145,0.05375242301804317,0.35741315730208,0.44698014252589735,
 0.44116664991043664,0.667346773145915,0.7086708183483451,0.8347406919792107,
 0.05779773614668593,0.7222710305331096,0.7191007949193727,0.5644452573360014,
 0.724648656465023,0.7309060420794291,0.07708440804978822,0.7016828114492735,
 0.7271551593025027,0.5588233598990693,0.6011731862961242,
 0.038427600326474964,0.21214440780613597,0.1923196001170736,
 0.9788873277335137,0.6258594198187967,0.39817981418577775,0.6006364156515043,
 0.13408888043534173,0.37763485999977175,0.4952622652024934,
 0.7430565340317872,0.3405882846323395,0.8181441365096258,0.46529212217858884,
 0.005801007326191043,0.1257621128484292,0.5077657024253917,
 0.3392458594046972,0.07636789686808854,0.5547042876823898,0.9148102842432798,
 0.9090588263563049,0.3776434601314864,0.7928097033272694,0.3377310201578252,
 0.596392127269551,0.5287252063612008,0.734016901198511,0.5070252096818548,
 0.7897539978564783,0.3474805395039786,0.41398072196940483,
 0.33883232642222494,0.005744279602292401,0.6967871195461173,
 0.30062796613647413,0.18189676581338865,0.4566668440917687,
 0.21913251428260838,0.03304291955039129,0.885698458902148,
 0.011903441316205021,0.3547150753444256,0.14910782148534507,
 0.733139230281548,0.6666718454585939,0.8596156583837571,0.8018891742289432,
 0.29575256275191164,0.8677740587258074,0.6576820650525212,0.898668009725456,
 0.25380114096180517,0.8656987247908206,0.5328394182842769,
 0.23598448496636137,0.7267862422652333,0.46747807728001023,
 0.7836408509960162,0.08651288636080834,0.6142720526716617,
 0.20978381568681415,0.25007631468920677,0.8695775852006326,
 0.22094868400433842,0.10426824474342267,0.808854629550365,0.9846601135892282,
 0.13904045726864467,0.419183058847019,0.787293362931952,0.8676226745933647,
 0.8964128811557682,0.20868863588231357,0.5659157311986773,0.7043286466455105,
 0.5325349856674504,0.7855795102191868,0.38965881667006064,0.7982399544930965,
 0.9554872858056461,0.1250113080365124,0.7591176554189379,0.7564640692627942,
 0.42338624766581157,0.91329551882477,0.5235812509389168,0.4224645669224131,
 0.8813059742367004,0.7442748219615964,0.15260365779787988,0.7005861241102335,
 0.46970397150647436,0.013134210911018185,0.6873189586778785,
 0.6467761645095438,0.9670853747633289,0.7295464021143392,0.6364504186556981,
 0.18979938976621002,0.7266676763328076,0.463119500053209,0.7663357283741249,
 0.9203664443113155,0.22526292534351544,0.6937801386676151,
 0.24365893934909288,0.9140498995135405,0.8518466372396798,
 0.28247715898974424,0.7461726178893396,0.929181978897184,0.7204911044293506,
 0.6879864605820434,0.8623799314939062,0.1422386579395395,0.39345664847062134,
 0.31236741674198953,0.28150856990901063,0.22767972461220753,
 0.17012304241840548,0.3186224600197455,0.4917175214337358,
 0.09781299665225907,0.8238805029276786,0.7536458826759065,0.6430997485188921,
 0.7685261085012334,0.0488173345193722,0.5588914745304037,0.9298722736354571,
 0.0716819277502525,0.47494174871295514,0.036174908778410675,
 0.08246427854839766,0.7185813604041785,0.36090731289804134,
 0.7011007000908265,0.7419423012389518,0.8825622802592076,0.0939553646802147,
 0.8526661773062795,0.23545276290964579,0.3600777033543967,0.7241337231065385,
 0.7834354959725005,0.47780090202116565,0.8841192272928399,0.8840900593072323,
 0.7635488255518744,0.6101519814050708,0.5994587154145505,0.4434114715075339,
 0.3286200947957383,0.43660266820594207,0.4994002580948339,
 0.11215612424239207,0.13447237319328442,0.9379884451072125,
 0.4159205385624659,0.2078007467263795,0.39228298170345655,0.849189317837743,
 0.6360219646708394,0.5845375836346717,0.028374199410661216,
 0.9554159599616423,0.5103674018494443,0.49139310835880334,0.9426430581930988,
 0.4141912168855,0.2679723683347546,0.724591823170851,0.44413739041423006,
 0.8523037337105768,0.19158404664685524,0.4270925715870534,0.7769657636103767,
 0.5215908431586371,0.7102735739699461,0.4327357396051019,0.508398357604141,
 0.7925402796507478,0.8202984466340888,0.30942298959800874,0.5762311206551519,
 0.32979168649569846,0.5274617338612337,0.2661392052923386,0.4163268073028531,
 0.595173471160662,0.8060351403203514,0.5583361478998973,0.19821793671909982,
 0.7780566485266637,0.7678532645359728,0.5762331930418637,0.9785849932757129,
 0.8586452910992284,0.42005763117153105,0.3045795444759152,0.4031869164203924,
 0.9250031404798056,0.5950781076621895,0.16049377272436338,0.4490401908275543,
 0.7914040452644237,0.29053858059131854,0.1469901663187908,
 0.27049390910693816,0.22597967378945738,0.541912742879177,0.0930928695933918,
 0.8623603481820309,0.9227613505299428,0.8096710204599618,0.7722310330609549,
 0.6237860792414537,0.9640438248025728,0.30916651098889714,0.5643758289074804,
 0.12791181164943277,0.22795370606032972,0.037618431618375414,
 0.4703634987938071,0.33296534154995416,0.027862836096422705,
 0.7455475373340593,0.5176153303116368,0.6880284111527207,0.6750157436885252,
 0.9366703940358987,0.6930759777137874,0.741374980507403,0.3908423333752872,
 0.2188039002945682,0.37979642362152477,0.38675048248142474,
 0.4143472012826519,0.5063284207352828,0.3070826136489795,0.9841318713671895,
 0.3356354651054725,0.4382879842321754,0.9777142442708114,0.3567494166837334,
 0.6038895198155795,0.13445446495459912,0.6282675369662973,
 0.21527080095542683,0.27606824063936086,0.44089902640646783,
 0.5230467512153061,0.2474184591662647,0.21464358322900357,
 0.28220367757449716,0.9364668188868254,0.24203832916540546,
 0.8979847585992846,0.18134499603932608,0.6030361742185849,0.8641064601550412,
 0.7677956926491815,0.4995907510063743,0.12203253097315991,
 0.023850533440243904,0.7726321465828779,0.6873248239146003,
 0.05938615112718959,0.8643053399702673,0.84693500199089,0.33655731697156754,
 0.9891202751760759,0.13815160840339136,0.8240825204690073,
 0.11578303619396302,0.7380845997458565,0.09962522263514273,
 0.6402932361805517,0.3634799919348408,0.7848417199667136,0.5240974928384333,
 0.48943907536758613,0.37236369001361647,0.9342728815932098,
 0.18902059539099014,0.8915285547403207,0.5777273478364826,
 0.05714679859634619,0.7295783597890958,0.40800794128256934,
 0.5379749882871399,0.70733816422209,0.22261765161460023,0.4079208361251164,
 0.44006621993884465,0.04749476413785636,0.6452235855445974,
 0.8037125452563418,0.6483545459927913,0.07323146341114062,
 0.16243538066115315,0.9260713650752445,0.9650273196413465,0.98260895162831,
 0.45215503304904936,0.9740498403428092,0.6367660896317446,0.5738424049045165,
 0.49976221380300534,0.763461116285837,0.20668840326758364,0.554521443527766,
 0.5009507644752378,0.5095093531205207,0.9612247446219601,0.6707805692917685,
 0.6371216769500007,0.21995637920882394,0.030981576720072623,
 0.7036696319702277,0.18191338157067527,0.4371333373901344,0.8304158158552751,
 0.14245648515370601,0.5054651013597279,0.43146794399819866,
 0.3937423511911886,0.18352577989952423,0.027632697086539593,
 0.6544010469848427,0.02032194999876258,0.8165392553365711,0.7958599920044365,
 0.8907841410783258,0.6908057210009043,0.17879294889582797,0.3110949266485705,
 0.7138694388459618,0.9034475758063895,0.268880421904801,0.5747238354213363,
 0.1012921959776183,0.9145852598537025,0.6417624972015722,0.5360092668879132,
 0.8905793608821008,0.45371874720511995,0.37626491025145736,
 0.03420817654197528,0.48123529065562776,0.8825670657231324,
 0.9101518752599743,0.4412693660862851,0.04576785062960331,0.4056841685524021,
 0.6873179699877447,0.12702915745074272,0.5192207830411854,
 0.18094795600577573,0.12411838156907129,0.9812797393896151,
 0.6311028042677616,0.2219867684133272,0.5364339151351377,0.6233941011586586,
 0.9887873071942871,0.22756048101636217,0.9014072980849628,0.2877111334339335,
 0.004077061566488349,0.5025632511255439,0.019935982430116517,
 0.28259189736398427,0.5612572974626414,0.7214106376762789,
 0.35164592079048584,0.5775057267569438,0.1829138807178028,
 0.08798900164728707,0.5978661652039541,0.449170599712576,0.47659303919027585,
 0.7693782623318288,0.36063160338417766,0.31234390453568406,
 0.8676119927468565,0.5710208064461917,0.6507983302667648,0.1079821523310307,
 0.20868705573430069,0.9821877249283009,0.40269322275518626,0.767205806769885,
 0.873030111794093,0.11511426405801883,0.08252742766783883,0.480295671240187,
 0.014933422218425418,0.6357753090731033,0.5720022567310177,
 0.6234629967894868,0.7282146510897759,0.7096751941027242,0.5581928766096572,
 0.9723080805896036,0.3129376576380949,0.510514425307814,0.4497061989661475,
 0.45912883987280295,0.31954481572998883,0.6585653504831973,
 0.7119800367964912,0.2585139042032376,0.13388819854062062,0.7732409423831615,
 0.7265168381961942,0.24752954431834762,0.6405525579477679,0.977344311569838,
 0.6492150298598448,0.7988505637850206,0.7712899890663828,0.4221046428425673,
 0.6585043132606684,0.9767854869275748,0.9515268634240416,0.8422747337531923,
 0.7712724898903502,0.7147010329976413,0.04473263882435263,0.1819330771352985,
 0.7437629593846692,0.1639856267494597,0.7881368756211451,0.8302066269500677,
 0.12859953989268502,0.014778829735901167,0.9585976422833212,
 0.33976761214492357,0.4451966539282065,0.7101576451412209,
 0.14060060112304906,0.16317264478856486,0.9867195494976897,
 0.4293905442042889,0.07083338254198523,0.6450919813405029,0.6273046418041899,
 0.1583757651634693,0.12414389772384415,0.5654552647846077,0.7127407013546586,
 0.4390962558189888,0.04075368710385929,0.5765684453924372,0.8282552524069797,
 0.13612351960435054,0.7469736665649265,0.5658287229298267,
 0.35926370195968704,0.22348713877126825,0.41296762574441515,
 0.381017473444056,0.9076015089953199,0.5196346513330387,0.1314676518643184,
 0.37170678778405486,0.45519796971531123,0.5466300830223915,0.884340829841992,
 0.7098154964221416,0.29371780946497517,2.688084211607267e-4,
 0.7755613974823813,0.9747121651389667,0.42058559385982586,
 0.19128923300318257,0.8019120110642253,0.8976038021959301,0.5485871274967866,
 0.45511943916566444,0.7821702912780932,0.12911085984178727,
 0.13850962664643396,0.6788039743898268,0.03887204585176618,
 0.4545451221782728,0.03262840968462011,0.44969832041415847,
 0.30970925037453223,0.389798468405553,0.18531803872384256,0.6066018020427733,
 0.374935124195812,0.4220201155808241,0.78162670003332,0.3581016062170538,
 0.2829008590395672,0.8172786175267124,0.6000573631250742,0.778714706392872,
 0.06177879739480319,0.31501048670914533,0.3793725439934803,
 0.9371383470803545,0.11716322164889281,0.529453487702205,0.51715513730711,
 0.7753909262545269,0.5205370912437296,0.4179782333043891,0.8027806682695624,
 0.06185482646500873,0.6094447134259252,0.9119540614455139,
 0.35995252697541913,0.7648852835600448,0.9892425097457576,0.7254778742438943,
 0.8404512834470691,0.9311209725017946,0.7141180881142463,0.4912144825997379,
 0.17432670268293393,0.17927767011793294,0.23681541833315511,
 0.5283570800280821,0.769842100625213,0.03769068030414391,0.7898849754638552,
 0.9768821583371616,0.48355163405150203,0.6891345394767074,
 0.24759305907451346,0.6176031767508655,0.2886089316426268,0.3889990499989677,
 0.24712357425919795,0.43741473450830093,0.7637630865060463,
 0.8081505361037833,0.49018302729073304,0.04277180544687509,0.060530812596436,
 0.4478542012662262,0.7863756562954497,0.07483112933184122,
 0.19749315607220608,0.32069978812264055,0.15092643088224234,
 0.5637759721551115,0.2797331796142192,0.5050847473648322,0.9352195730859347,
 0.09748029845177109,0.9988002578047213,0.42066148995604546,0.546412018419834,
 0.4791371607578049,0.6450149136045463,0.18222145160523973,
 0.25066550930093645,0.9903598206312971,0.34868231533690897,
 0.14986628583623762,0.6282071201340125,0.08520761672854416,
 0.8471087624108475,0.8060931958316508,0.5552108422809692,0.8368659713773917,
 0.9162990187462565,0.36488930419801857,0.8720816090244475,
 0.028832305164227745,0.39303165337428825,0.9547131168686219,
 0.9968668662848567,0.21061860531089616,0.7583219911529908,0.779166106807386,
 0.9772676412667789,0.9564618385969305,0.9995071092224042,0.23560777399959443,
 0.44298024044174156,0.7588680996658759,0.3363009987345388,0.6675810057283621,
 0.2218189587103757,0.9829250906395604,0.06801226991829312,0.3172944256392257,
 0.9808044854914444,0.5343255306101187,0.2279387631611478,0.7784062531929814,
 0.436602668663825,0.4960897609028848,0.14289773139665396,0.8844437530687209,
 0.5026976195116173,0.43204827209351016,0.7302959841014957,0.8858287300117516,
 0.5238801720462807,0.010821701086212387,0.4063423015332911,
 0.6421680763228461,0.07530858305278676,0.13001252226166993,
 0.22609055015169677,0.9000514701329544,0.4283445756463694,
 0.11268819485074233,0.5131941894767622,0.005394178441413211,
 0.4947026340640214,0.4848156696065069,0.3653930401822649,0.8004605555606624,
 0.7098581341321728,0.38545869193860083,0.47760589026438627,
 0.6842397277153367,0.7638599232513483,0.9734401876363796,0.7402223441416574,
 0.760997079344775,0.6216763885949773,0.2881024844514164,0.2874580100845532,
 0.9651438631978692,0.8289861577577643,0.8681518687345579,0.1960631598094822,
 0.1892716038744403,0.7890799866622842,0.824990846127428,0.40253428212607323,
 0.5274448436900738,0.9189213242782158,0.06776728884385996,0.7187093274285548,
 0.7915983302569977,0.2843337411977216,0.05597021735005481,0.5882237423435821,
 0.5938288846242713,0.7082693221108235,0.7481015467393336,0.5389068102423745,
 0.5045492118425439,0.555120582715975,0.918065721902729,0.6663735210802598,
 0.577956943384114,0.8499553788564082,0.3604531824040872,0.9562159926019718,
 0.3125363992600394,0.35853761449281396,0.5212758657042736,0.8077167418491673,
 0.31658042504199324,0.9379812405315271,0.6531454570054037,
 0.06732770134210597,0.9598028678679661,0.022626715779181295,
 0.5686140621852779,0.30369670973286345,0.6106687335278081,0.7225912153607769,
 0.004267833860504444,0.09784297575093037,0.3096504957792967,
 0.4694200906936674,0.8716820932160789,0.6465150381667323,0.1405300046105541,
 0.167078140384736,0.3527772115824038,0.8898291201111952,0.8873908014525191,
 0.3376341688995753,0.9381694138257255,0.005929931033325175,
 0.5521666143348415,0.3252261262095,0.47368871297789794,0.04694856058549701,
 0.21065093497460485,0.6646308895320405,0.27794516769995203,
 0.23119717454622746,0.7991481447272175,0.782263706619096,0.8447946892576537,
 0.16630224021814632,0.3144662390552727,0.934450532642289,0.46888794524141597,
 0.8456989663483657,0.591006691477119,0.45386621611982614,0.6268239404094923,
 0.7205744741941861,0.6195414233153718,0.8343898004291121,0.15615712129279102,
 0.6981590074215567,0.7867270635991969,0.8334288946938865,0.9179804215051561,
 0.8023229887322209,0.5363721715722714,0.5416859483993846,0.05066855617039878,
 0.42025670541664895,0.32506721367144786,0.3649539317563887,
 0.49516324827153857,0.5851901669192459,0.3866756346928476,
 0.22632761606685203,0.4598509550983817,0.5100493930971098,
 0.09982001158896103,0.13269512643748937,0.3012537337489678,
 0.11219724413251997,0.37491245466799894,0.7116246675995085,
 0.036984581721073884,0.6187823705393469,0.8847031107377429,
 0.11489186983397848,0.6072228273553707,0.28936627379391355,
 0.17315364003832934,0.6163561411014629,0.0953540277896997,0.636099241863784,
 0.6599484359533718,0.8143698237986071,0.9624790884143933,0.17173598516444577,
 0.7484838958372999,0.13817328031816345,0.7100718934870224,
 0.08333118125793759,0.7891699704150269,0.45786094918086484,0.822407082788439,
 0.9676976563317459,0.4742428462361783,0.16192768957333192,0.6020549188487063,
 0.34725433043355025,0.33239502295146783,0.4632407736847113,
 0.13378304791903828,0.4852431206315416,0.032785808734932975,
 0.9710786021308455,0.9172765422602126,0.4453046616061067,0.48825698626168934,
 0.006750178731164569,0.6802352378026947,0.9623003447286835,
 0.4045071216927827,0.22169319381379027,0.41420662753129833,
 0.7943119613639255,0.01712565624661566,0.7614545152755753,
 0.33327300435385365,0.730772004543763,0.5014859343282664,
 0.031792800912512575,0.04819853773252292,0.03276592826069413,
 0.6497156602286527,0.42269804042471426,0.00749569999442079,
 0.42972122276625935,0.8399346886260564,0.42892821996573915,
 0.9021310856400484,0.7235253333923285,0.9012271984481636,0.3721745377471539,
 0.8560097755566028,0.41414366576084105,0.6267392027103946,0.1882517402294115,
 0.31308009916825275,0.34866472063979714,0.08688943487744805,
 0.7631385314844917,0.8881035052250591,0.29180894804060975,0.9335875154560079,
 0.23404697853502088,0.4989281650413492,0.6912835738773222,0.4533461666784785,
 0.6834413078911509,0.19792282238258907,0.41347085456009547,
 0.41161987930643773,0.8020502911797127,0.7020170018618995,0.6051950892886483,
 0.501244773801522,0.3695238726569665,0.5515136205139395,0.0804359816825132,
 0.3389514693067328,0.7996024288107487,0.2591421250156952,0.3836449913028833,
 0.28418914474400025,0.970778778333945,0.02596791732597925,0.5163622578041667,
 0.5774248238557824,0.5161330905164013,0.559318515579384,0.8383702998050142,
 0.690776194220152,0.9746124997415174,0.6431582931136104,0.843614947730318,
 0.6342797375843913,0.15303252441385684,0.8337230179924768,0.5477065209104929,
 0.9453472987951443,0.5857396313657861,0.1252916631160168,0.18852523206214644,
 0.6892182057076197,0.6113012794069148,0.4766618289679543,0.04810697477845982,
 0.4364783193143791,0.7665251674306595,0.6504247294094758,0.920536569066944,
 0.9010424684726831,0.9971061763434066,0.3535345376332215,0.09291614321008801,
 0.04680630760240445,0.8348406319740488,0.6777656145553251,
 0.07990522743513251,0.8022001963118388,0.01313266603960761,
 0.8184310587858383,0.2781136653534514,0.29782686523510404,
 0.38965008519263455,0.6121268154131262,0.8079410697569942,
 0.13223381164982118,0.5259239241058337,0.7044171956622151,0.7210505417947453,
 0.6348454909379607,0.19247028844486258,0.4172785658619571,0.6384094241039041,
 0.3625998205686817,0.9912412633161362,0.6580271068374615,0.5815722880581297,
 0.31099970574334546,0.09681027274618392,0.8748741860052135,
 0.6736624193127462,0.645959529277919,0.6967600128480934,
 0.0073383296813267496,0.04312037304382399,0.7645880865953313,
 0.979501786305463,0.813659387671395,0.32461379278881286,0.8912991932628527,
 0.7243672837555168,0.25922145429483523,0.997943367094421,0.4829897995758352,
 0.4460921609952818,0.7017768436541523,0.7740269726252478,0.22313308425268874,
 0.7555248824929691,0.05073230636133741,0.30476878185853984,
 0.19880728819755666,0.8437201829155666,0.8403622446662616,
 0.48068430370451687,0.5514190957163805,0.25914377111157405,
 0.22974195295757516,0.8896273263594521,0.5831871818464749,0.9302180552480016,
 0.28597395359599465,0.18862108126788812,0.6133934973780916,
 0.42128083922961257,0.8291508049409156,0.17260312172451608,
 0.5397849788347254,0.41407811965462926,0.6867803427423661,0.6439987580588125,
 0.12061930727242831,0.31076296669152903,0.6027889017578285,
 0.13630687857645418,0.4874309477944414,0.2724648599562689,0.8232816915516568,
 0.934012644378923,0.6511808104041779,0.9303883838573439,0.6762002906075186,
 0.47005896592077545,0.8483996164334587,0.7411983985831032,0.9937443936953511,
 0.6378611043242618,0.7109389302214038,0.39581681604117946,0.4640187894963971,
 0.2937376565000973,0.14814367939633077,0.8256185596998994,0.8403504937966689,
 0.1483977698868436,0.4311560571213645,0.376092939150102,0.8513776986868462,
 0.7869775384019104,0.5083040173879579,0.17181764969548374,0.8014430289496896,
 0.34248395920347297,0.6367137612137467,0.017923836751641775,
 0.6545863943147953,0.4323472740152283,0.5535759018638733,0.6211035579712361,
 0.6939617995893497,0.3273518838613907,0.9302417587287662,0.3739048209297038,
 0.6487462851296604,0.9912085951123205,0.7471836320459387,0.8536739767040259,
 0.9676907839839576,0.601747028326016,0.44733326886003955,0.3034006700701205,
 0.6857701816352936,0.6080944243492175,0.8886998657373258,0.8923074703331532,
 0.03067795189751743,0.7241739245393045,0.6025041175501435,0.4464792229274098,
 0.2277133218042957,0.24393265603470038,0.2158679252529141,0.5844058957349313,
 0.39844426503287744,0.20279197507414315,0.9494052728122003,0.623820718504496,
 0.14901663093898443,0.8052663559840358,0.017188107899773764,
 0.5797823127863926,0.6164117570606943,0.4783508330422934,0.5092264385079224,
 0.7358176223703536,0.7857872357199526,0.34214773657549447,0.983994911033927,
 0.7851460078912668,0.669016483779162,0.11896885087328968,0.9745840726123354,
 0.46507378230616003,0.9843660831425024,0.4592820082597664,
 0.47590378370614717,0.7914682485579047,0.24782741842044242,
 0.2923490539141326,0.45601791063178077,0.9583386537820993,0.7620649802450622,
 0.006546495639764283,0.8907051245132217,0.049435746943040204,
 0.9924806030589945,0.03815781337626012,0.1292365899530351,0.3685104612375171,
 0.4264075393779334,0.13412545201437642,0.9637890370737818,0.174909511257208,
 0.360615690799189,0.3096108627432417,0.6726694111895177,0.8857467224274447,
 0.9422104040606336,0.16294520566813198,0.4055554337342573,
 0.27845221977652934,0.29844185522144895,0.72482800825142,0.20031689856913715,
 0.020945292705454754,0.24495476326267018,0.8268188677861709,
 0.0624173704995542,0.04682025052932148,0.6850062612543104,0.5694247031182015,
 0.6691590667989573,0.04744034260677776,0.4562258453876916,0.1749038481196754,
 0.33159238128592605,0.9566703678140668,0.7819493187935753,
 0.23560655702035416,0.5647192563932305,0.44508437462326433,
 0.4846800619947802,0.6065209742184337,0.292110770429556,0.6042092543633829,
 0.09649278531276773,0.2991209610989568,0.04961116147208533,0.689084325849257,
 0.6035893082753778,0.1924598958873467,0.15535948360071572,0.7769370770420179,
 0.8615771229501483,0.6541082734311668,0.13605955107791579,0.5146983543347,
 0.24853860094336833,0.7655766154817891,0.7806120419562729,
 0.09426876808674611,0.39391606534179735,0.9264534403082723,
 0.0679972463466132,0.43433666299429163,0.8917860806507032,0.6759301832244837,
 0.42436908322048716,0.5407834756403522,0.4903129936625099,
 0.02511173699728264,0.865644751891802,0.9823067715401488,0.2501675690631341,
 0.6311079343586932,0.7116030218001351,0.7481693807293748,0.5218975128894893,
 0.8412563483415478,0.8231846260460481,0.4551157250774477,0.03513125767341041,
 0.018445864530580947,0.6592640987906152,0.5054611750347991,
 0.4075117919012021,0.5764260340683345,0.6275866986792036,0.5392998909508613,
 0.3365498898157011,0.8640174971929733,0.1304264903590986,0.8018978469984489,
 0.2978747169757542,0.15239719254699102,0.3728624716100738,0.5758959215961373,
 0.6849377727252619,0.40896387292641956,0.7488107865633014,
 0.48087473353683596,0.12157469359373019,0.11200967447489607,
 0.451370198719411,0.30097385912008856,0.2856503483128433,0.7603458957586322,
 0.6045833340643159,0.8021270056586679,0.47987148302087834,0.6920400348093004,
 0.777253005149283,0.43191839396012144,0.20849247369978896,0.5923251738210603,
 0.5283234761617287,0.19841866560581267,0.28360470534349536,
 0.08548714638666477,0.9288892690097821,0.6488888151743609,0.790775928600528,
 0.3421626177435724,0.20457819067366279,0.7381886932811745,
 0.28470352565725476,0.5876591209208919,0.21448804417121992,
 0.02411199230030381,0.40770202067434824,0.8478016431373031,
 0.6908069405062927,0.08360725599786256,0.9283151488576744,0.7991983241607017,
 0.5968085914296365,0.005647922943924244,0.8239543707946299,
 0.42061149692864486,0.9466282829695694,0.4484650195637089,
 0.16415100611587785,0.8942024519801128,0.042871397199774175,
 0.3134579609773279,0.6272197784194034,0.9492569156173573,0.5560363234125134,
 0.00849648867180619,0.6072133393606816,0.06050940078008227,
 0.8594574916152702,0.775591308718038,0.6105753282032221,0.6935785605300784,
 0.8084941433957631,0.3439611698189283,0.6424316300947194,0.14452111276896928,
 0.07343852782956417,0.6330207341459176,0.6162349201534485,
 0.31987616347459535,0.714310617706619,0.6470422083078031,0.6662763406494673,
 0.28492133606119363,0.40432312614951305,0.4336815494773605,
 0.9446277618266187,0.7215482799781986,0.19942671726725447,
 0.17131798412624666,0.8304537220999464,0.3480935913888714,
 0.07690941351977165,0.00703450415671425,0.834036953894731,
 0.36291465256384936,0.3022116524975422,0.25384083368896704,
 0.14398795520722207,0.12027406967033671,0.12107383907899716,
 0.9914182870032189,0.13637634157170586,0.3795652257515051,0.4478177937721305,
 0.26817174211411765,0.277754746359112,0.30220030115349594,0.2188664208853962,
 0.7459463825563577,0.48039678313159595,0.7571816044089594,
 0.14086395479645408,0.8703545120529161,0.36914180565635557,
 0.5948521864650995,0.682771271379286,0.10830652716051559,0.4184846919394304,
 0.33484745778720126,0.6735244573601586,0.42074464624307195,
 0.7363329570811229,0.4463591539018641,0.6507299649914637,0.9547791910184131,
 0.9655978010593131,0.4219516117946046,0.6345860459075233,0.8737723183658709,
 0.44777367526543066,0.7056790331696421,0.9552724192249636,0.5002785265359292,
 0.9433935897071511,0.5137672871677119,0.4019978342191901,0.7020005602544729,
 0.5950254466298959,0.8545379977959426,0.9482972550189517,0.6066565543464604,
 0.04325098432205843,0.04071519428995196,0.2731699525047774,
 0.8835506489510949,0.6605838871044281,0.9436301660132735,0.26806806055977306,
 0.710476767136512,0.5424492633034553,0.5174465443043647,0.03516256928409012,
 0.22025731222167266,0.5827129009631127,0.17575292365075657,
 0.9513999752754462,0.19400445919140086,0.7670685053438334,0.9056538109905132,
 0.3538858149871167,0.930081978535233,0.5819913173216986,0.17795091514744676,
 0.5040573499934544,0.8247115643318583,0.3596186351618208,0.40652949330210186,
 0.05104112834624008,0.9773204195696891,0.0898872934270627,0.2162864081549971,
 0.6410814593473972,0.020822800976285194,0.5741008650304991,
 0.28244030767553696,0.3085629225087837,0.38145985899507207,
 0.1264037106477064,0.9760527814008265,0.34330711968726857,0.9812445233817114,
 0.7425223196050303,0.7199174896413189,0.7784273519164099,0.6547032447344181,
 0.6329395363496298,0.6648226603913941,0.23998261756344785,0.6558420987919305,
 0.7148701701849511,0.6825821663327353,0.829575605834038,0.47289134220427487,
 0.6507505274467323,0.9688078021836111,0.16011948169404444,0.9572791123307358,
 0.20004282123459882,0.2761078000549677,0.013761075286685243,
 0.8781780212969941,0.014775438690480147,0.7044372301866553,
 0.6657859193588269,0.24799477258788383,0.3633591109067965,0.9491404337794098,
 0.24539865835813301,0.04576335214224703,0.10983699866556018,
 0.16251939982917307,0.8259024425196358,0.11599944697409814,
 0.15871046266796818,0.3734635971042164,0.7051145579934558,0.6259563395137293,
 0.5996714293272273,0.5239577138966254,0.2857517648484508,
 0.014988756745079601,0.15822047293719899,0.25954945345162095,
 0.12064128439729549,0.1840621733572645,0.1283957632184436,0.5028719947467676,
 0.05395096116981801,0.2694235330271417,0.5068018425765117,
 0.36276540424325765,0.12429844709686044,0.011892182730556033,
 0.26008335963561446,0.33873686357562016,0.5651786223357627,0.541030950741201,
 0.7949971737428826,0.4172743171995654,0.22543357511986073,0.967577940077444,
 0.778534183632329,0.7082800834391575,0.7500280698569339,0.2397751743547566,
 0.5179824461854062,0.1776726057829351,0.968961392614203,0.11001829201150715,
 0.9394518296134772,0.9164405698776092,0.45485010239103063,0.4453934134816968,
 0.07793280110734002,0.7698366786727412,0.40111520826143343,
 0.18334122890111226,0.3342684205060883,0.14081919752410232,0.932894464312756,
 0.9620200161366054,0.07709456702749773,0.28671135960530414,
 0.9996435786338829,0.2229040327720423,0.00859697845004459,
 0.31817158564113446,0.6770319181948447,0.7537977163137071,
 0.32426035172734213,0.9054273123787058,0.5069816488531075,0.5215263595193633,
 0.16691943466478099,0.7425663304093176,0.776956225559229,0.15842061825834075,
 0.28496920758638344,0.27709719128424004,0.3254348578060118,
 0.008152968739531863,0.9538346172058529,0.5461634546247464,
 0.15830306117374637,0.8748144861565628,0.49104095299453576,
 0.49361761879768007,0.5510529523782416,0.6068626064166652,0.3415571618797737,
 0.46903709973359553,0.7748947519067149,0.41911411587422787,
 0.2967566417446562,0.9367442002346299,0.4513236240111329,0.23896876848465964,
 0.11100720166836786,0.24688998490383995,0.366216464395556,0.8415700614482546,
 0.6221214192342691,0.3950162586740886,0.26437866593229997,0.1059976098500568,
 0.6043914218833598,0.8667230773904241,0.26129489172152265,
 0.46751851545847223,0.2477466073507768,0.715236909693454,0.9000534497751139,
 0.6571209645476535,0.4986746255070188,0.04043770252375767,0.3756628912942749,
 0.5870778522112193,0.289281260765888,0.05584459130160735,0.3423852474996957,
 0.2886497319002417,0.5954277172099957,0.09290716487262074,
 0.07158841343835165,0.39069699898716104,0.6730884033889946,
 0.09670959492652431,0.46981473402493124,0.3886513017902111,
 0.5966839778018471,0.44037496733427595,0.7878659335213009,
 0.03984679309844796,0.6264491216071414,0.5087013596707118,0.3804997496744661,
 0.5267581809877075,0.7310807450540024,0.8721400035374176,0.46635196229658527,
 0.5061530982557125,0.2443774533377332,0.5986904514542474,0.9409301264260754,
 0.8020986031122354,0.37293254180291724,0.23317272923152788,
 0.18828440567795235,0.5769611722241326,0.892136556583459,
 0.030781975093637892,0.4563880079334788,0.7523165379181895,
 0.7302076642388987,0.26836977601746703,0.23719528700041448,
 0.7094356986184914,0.9857377905600213,0.5279266527206193,0.7798971029495516,
 0.16435943154717414,0.881749159377363,0.5995441084419959,0.6507570437907617,
 0.7639681196566296,0.49889676976956954,0.4379337031378695,0.6933272784685122,
 0.6207160379641179,0.7778966714668281,0.879925932309273,0.9992073600093511,
 0.3714484777119793,0.2469557766284075,0.5824085760969296,0.4762806635468386,
 0.8871428466439755,0.14688546293006,0.09051930573215139,0.825728764022814,
 0.4604168578026182,0.7921215008614026,0.05482139287462595,0.2587780435335242,
 0.6400693533801315,0.46778628446575676,0.040714711166917494,
 0.8584290527458278,0.29046482866237555,0.8764683493849571,
 0.011379416113862728,0.38954833018850565,0.5859197219302542,
 0.07836058526691048,0.07499515685029745,0.011304670180065646,
 0.7060008509420227,0.8086341683958379,0.3663201718724275,0.7270843204926375,
 0.4055055066562947,0.016812809422755093,0.9431612687633211,
 0.08851502818215762,0.14641501818348657,0.9806969144094256,
 0.2778202538406205,0.525116774696321,0.5095789439995357,0.08923208283714301,
 0.6247528147624885,0.3752268498899277,0.7081141965852138,0.4720186159512044,
 0.07866961314915244,0.33865700693427925,0.07760815229783868,
 0.888134587294078,0.497336871016134,0.6565174194557646,0.7569348469139597,
 0.6856975096417426,0.6603472464958173,0.42539576618587194,0.8647420266130691,
 0.46294482692261696,0.5907197477518485,0.8759371023374554,
 0.09167921235426701,0.34744258833278263,0.4768700468118102,
 0.4358540517404116,0.945041516740626,0.3986082407842575,0.35875369904512744,
 0.08465095176175164,0.6674461430392411,0.46494354355685263,
 0.09974828685785664,0.8342627360102357,0.0988131929350653,0.903318550261229,
 0.5078335629041275,0.49249954947356744,0.32199340774649,0.4836042430682995,
 0.751034354792313,0.7814360578858572,0.16978698936844427,
 0.012162161894927581,0.051548643368005465,0.2313859835418831,
 0.7583804619279393,0.06344105307934966,0.7035600738536114,0.9079899711726649,
 0.10501886076261757,0.8205351294259223,0.3163716502948589,0.7181306339295993,
 0.40034991452865387,0.6408556042861107,0.6782418979147005,
 0.27384231186646024,0.6623753245132805,0.6673104762867744,
 0.11345373375504375,0.5018484766185191,0.6730479446895548,
 0.43329236182062436,0.5212660970639978,0.5368225920523956,0.6177315966389346,
 0.028547393404405708,0.8171886190637574,0.37745055448876497,
 0.35835388140897606,0.11773728183926546,0.9724933939762608,
 0.7138671109564606,0.5789996637580299,0.1466004532331685,0.9083453435507759,
 0.7056474245912,0.04793851216400047,0.14958716455816834,0.5868537875489558,
 0.13830812409019966,0.45474973076451264,0.9537753588034514,
 0.6283426956917939,0.18975796163739278,0.7411613770271798,0.2913250669345978,
 0.9709378517878674,0.13054005823452797,0.5797438267026804,0.0812199443068976,
 0.2236138190571868,0.5384715319681692,0.6469003441035297,
 0.044634444024245656,0.05997434030039539,0.8534216869176942,
 0.4668138609478091,0.7916420778765462,0.2231998764147608,0.3121715603961518,
 0.12450882608902714,0.621035890086874,0.03307141181556561,0.753648610923866,
 0.7314331447260638,0.7929200386572469,0.26519895657574033,0.3976151069543983,
 0.4623867770752237,0.1122897193163418,0.684519818879857,0.7062067211523191,
 0.8234263990282132,0.30320721630633607,0.33687013304229896,
 0.2547753630968089,0.6975378129263273,0.0020079337413335674,
 0.8905702075315619,0.40915763122466364,0.2863534600183574,0.8353973980121645,
 0.09417412645719148,0.4481686660727603,0.8711910699438546,
 0.11101051278314134,0.5355151064447925,0.5128721403135315,0.5754400006896914,
 0.07356172945701389,0.00678953265229143,0.6631473056968654,
 0.6799330964591933,0.2721200531443777,0.5088107702421403,0.7711806693431775,
 0.768613450504725,0.9554687146784817,0.22898939592538337,0.6623099905978226,
 0.2618715651856711,0.721311769239224,0.48640635618873784,0.1124724786162058,
 0.02039309844467052,0.7652935002190618,0.6594607316553462,
 0.28998657579625065,0.7638449770541761,0.7408725408180208,
 0.20278163433106033,0.4380666712661546,0.2890230023057988,
 0.11147177382990348,0.2865538651343277,0.5059412665600694,
 0.44177546530342504,0.8512572734371463,0.7538786526752644,0.8775510076814873,
 0.659615249452998,0.10390272248384402,0.31979099909714126,0.9191984757488907,
 0.562936934858964,0.5836592810417028,0.845127307189407,0.5348779013460587,
 0.5074058779878248,0.24669243384370132,0.017704854514833168,
 0.45460318889310647,0.7386016952881372,0.33797110509895356,
 0.09635659398366914,0.9096930707247903,0.362319687492032,0.9780628027775046,
 0.944809508891797,0.2563546773551305,0.5661106269248968,0.4915502655204278,
 0.2880993817493681,0.45648368819428686,0.7114317105966776,0.9868601327582648,
 0.026348985349847198,0.8453800876973487,0.17228394865092378,
 0.9284565016654797,0.17346864338083845,0.6527384314969737,0.7132034801694845,
 0.14066259219627253,0.11669177971665734,0.9310146739487446,
 0.6607978740368948,0.0010105379114658808,0.4436969468364038,
 0.47721585092599106,0.404115165597003,0.7615921199221615,0.989677141040354,
 0.30667498058552456,0.5865891506556133,0.3848766352062284,
 0.19444224838164337,0.2543505021955963,0.44822484920492534,
 0.29136096502982267,0.7849620368944968,0.13275415036860083,0.816165374295561,
 0.33382560715685916,0.7026490900750346,0.8805089497380705,0.6028791340085696,
 0.6437320307616123,0.8250754068396413,0.5438165967334796,0.3767251922182223,
 0.8612884538013015,0.02274871781859633,0.8961070034336627,0.6873699407477433,
 0.49561852115375593,0.436755029514316,0.3400007741876019,0.627899173894146,
 0.4094776885579876,0.9475291275544627,0.7447483799339868,0.6089562582431594,
 0.8296536360748775,0.022812885999534904,0.7383306552215261,
 0.8909006273040008,0.18667165935254182,0.8312041204721111,
 0.38081595593306306,0.5733467458173842,0.447086691084457,0.8015999331097392,
 0.6566780830622717,0.5688510781350412,0.41439893237107717,
 0.28299457877683853,0.961459987146197,0.26339757444692524,
 0.22282365962580064,0.1383978083962003,0.4051451676602277,0.3409690714535121,
 0.7133520254866144,0.6247069152286548,0.8423784682382371,0.9856276951813014,
 0.07708509680147357,0.4942710680983826,0.6100292280113272,
 0.03583430567275636,0.7720477210368812,0.7787974975299115,0.7907876593829258,
 0.9040324405455491,0.8145214060403956,0.5103091568625349,0.07885331271757468,
 0.2556700549974289,0.9096370064355428,0.9559043774160997,0.4942740792263117,
 0.540597937679284,0.8034417114499788,0.21481540087636042,0.8284413631099216,
 0.16098879041458114,0.490329012359497,0.32244149795997146,0.9908518537185823,
 0.8314253965614283,0.35316153523209015,0.3627094159755426,
 0.38770989534586675,0.47835845185658754,0.6468346792207397,
 0.1082096977197271,0.2169828296119668,0.7054015116895475,0.08707837079251546,
 0.24226332578011278,0.3302538017557848,0.07474629871245331,0.424267536235441,
 0.8167582321607739,0.4406731025781633,0.743035047476155,0.02079349473819747,
 0.033160629985620194,0.09196970980077401,0.7794429602557288,
 0.5424223000902938,0.4517972479750686,0.07444670510123141,0.5904306505923125,
 0.09453357029239484,0.2964769076889403,0.9287401374438716,0.4722159526281754,
 0.4109315879003974,0.24691841863701713,0.554269656281211,0.514031243429395,
 0.9520581413633938,0.09845539032848005,0.6277028787674962,0.4471997721162847,
 0.5027543236166307,0.3820866914693336,0.3522155107263454,0.10741838899374856,
 0.7902436552240697,0.5716593993085572,0.2463683944604046,0.11760402175472906,
 0.7484524206950739,0.711874123015224,0.12284397280621329,0.5516824304999084,
 0.6129510883263929,0.9154710659296948,0.25325657576763005,0.524565242827208,
 0.4692203130250886,0.1911982517082853,0.7844229322730158,0.6627229046315233,
 0.30148288575423854,0.6556557112683579,0.8910891331031197,0.570119206933136,
 0.011744819313808885,0.30646519468792577,0.8651577235312363,
 0.9531051666482678,0.5157343923014269,0.8067313890886743,0.3737195589996559,
 0.3312362766927742,0.17855094901324242,0.18587177226720342,
 0.6852540000337309,0.5307448509285426,0.9042463912167508,0.21660917601697172,
 0.927899282197354,0.03807076663100151,0.7181395268296934,0.1392616614492279,
 0.2667722497281436,0.5058238010104957,0.1270639689576879,0.4136257700076058,
 0.6403916510693799,0.6385671854162013,0.12793612074977623,
 0.02882776249622998,0.09486448823438742,0.7093545973413075,
 0.5495136208175366,0.5063891034576987,0.9753765153007984,0.3756483721957634,
 0.2732395308731529,0.9559698778433782,0.3749416496929153,0.30397605472609657,
 0.07864475554529515,0.8453204293748503,0.5771533925261115,0.8527664415382024,
 0.5152551921659513,0.19613059116681753,0.048774844092387526,
 0.7735258036720245,0.28196688192628505,0.8621779212804326,0.6539892584509839,
 0.8303912976911465,0.3052364735214609,0.029929318816155015,
 0.3046934314819536,0.7141529718429553,0.7694426448007939,0.1422556408923229,
 0.7472336561793014,0.02475102885597924,0.9479022434173067,0.1517737367994697,
 0.21095952455557898,0.7191919489850691,0.1467896902515483,
 0.022848095482896458,0.853596502281281,0.46173900650994004,0.787746273628532,
 0.1297722277925858,0.3664020948439406,0.9658191271460165,0.768668002686696,
 0.2708444269016391,0.12158473408680626,0.2774557393187239,0.4166475080747898,
 0.8362340953048629,0.7829910619855964,0.9365686441106487,0.7609474530829206,
 0.39885221858747666,0.11905090176729782,0.7556592182931855,
 0.6121397584145041,0.02745956563143348,0.7274462040276432,
 0.11678112160277165,0.7074268830639,0.2329951763577648,0.12724041039953315,
 0.3936854212706228,0.3866307833589142,0.38355530799929594,0.7369332565903495,
 0.05247563241300823,0.4035056252591974,0.010650809413707774,
 0.6106926603127449,0.9130655642495262,0.4676126427104339,0.3430920749621216,
 0.9404837315389112,0.30460368077681677,0.4803605940615606,0.9672936749589525,
 0.09425045005044375,0.781878252131158,0.6924803207634219,0.15856496773853024,
 0.8354169008904785,0.9243763370124753,0.017535004439510504,
 0.6110306030339123,0.9269284180873986,0.5086907592326719,0.39682649412608617,
 0.1744293525748638,0.911228554983546,0.9048404928070408,0.30382056544514013,
 0.15215035181458036,0.6055313203380325,0.6601722060637949,
 0.01934836375742588,0.20971468344479205,0.9065114134181619,
 0.8251076433193875,0.5326452978224143,0.11328387967698883,0.5558917040885548,
 0.6641591602040436,0.7562031760210107,0.6709743790609287,0.7863453822582812,
 0.0951652154631919,0.31311115142347823,0.31514976661192584,0.772860559995217,
 0.7194390162951129,0.6977379172851165,0.09411259138678307,
 0.38986402672709464,0.7249580753569307,0.3971282776212588,
 0.08629598957374185,0.38653669686498937,0.7330027077397551,
 0.8273272180825594,0.4676470869790126,0.30805017855820893,
 0.41093507526911854,0.6272150888845941,0.38422025363443646,
 0.9574313241139458,0.6434737452952151,0.82298699172036,0.2473538304542524,
 0.2558384994640226,0.48946006246849905,0.8773444200288497,0.9093087468894668,
 0.4603023170884998,0.48856595349792753,0.28495119007339775,
 0.3152168865261673,0.69035823933964,0.0794169368552553,0.2801935236771549,
 0.3038316945905999,0.7650936568022436,0.1079152278342217,0.5812241627272433,
 0.529292335616733,0.9891672168984982,0.4355604661653575,0.5083078420516678,
 0.4200937360220913,0.761560991802845,0.7557078530019473,0.5210381121427471,
 0.440950248264823,0.32168681134319865,0.23859355902831114,
 0.042841059143382876,0.32361736679879116,0.37195342051537594,
 0.14085237124554006,0.8147228406849685,0.6915368003934805,
 0.02276731667159526,0.19604939861019255,0.1827630504530997,
 0.7738988048860067,0.9475539799519348,0.6800722227417918,0.11985455053703609,
 0.990256703565016,0.15753692208175618,0.7065414964774525,0.6352464722823831,
 0.43053538635209876,0.5984040747479584,0.10988006925329152,
 0.35947695973073857,0.9279642840418484,0.5164853730537419,
 0.09373707470007642,0.8139340023034097,0.8053399291885398,0.6937700355940366,
 0.9565832470369664,0.7202002584264788,0.8874078553006151,0.06686442692897776,
 0.18439768388195255,0.7572828023769503,0.4437274728510181,0.8794594382812323,
 0.2665633146995493,0.9680490281972276,0.23355824976519846,
 0.22088392705204507,0.5529143296457009,0.8067723026562751,0.6238233259890775,
 0.3751183672052252,0.656560467628674,0.25486160870154184,0.22909886613631092,
 0.6469766486395667,0.07392762801215236,0.20055360323020988,
 0.6322309872446559,0.7117474054854068,0.8775778616688102,0.7659308234678612,
 0.2414795267814056,0.6561445115084551,0.009938973768581216,
 0.5050028542185031,0.15426254120078076,0.9384950410631847,0.9894857679684045,
 0.11687587729338211,0.6649821652653386,0.9087709136663445,0.9733008513057579,
 0.9862341397016747,0.016009658862362275,0.21314495121843746,
 0.00905716360221187,0.5209871544476471,0.8644222172922437,
 0.27540268116824695,0.7340197172612493,0.6406384454465478,0.2705604512765605,
 0.9335081372563017,0.9085438168968961,0.028673007469989686,
 0.2040548691488696,0.40816641032595324,0.9632974513552238,0.3340618546880856,
 0.7487355387791141,0.7149501451501589,0.5882333208446423,0.5243776433032318,
 0.4421094053873196,0.5641103869452911,0.08512481020201335,0.5194090885460205,
 0.1809255872484119,0.6258192979705379,0.15676134353548743,0.5908506749003942,
 0.41438322828401475,0.19846851704476776,0.7369056155121164,
 0.1921279603848386,0.2932054987448469,0.8237417674215523,0.5175158624058401,
 0.4538637090698394,0.7962856327542015,0.4612301192190994,0.08273304548749816,
 0.7258479815077227,0.11527883789025506,0.7568675267833479,0.3533005280792837,
 0.05464859865330429,0.5342819221228406,0.431060494763452,0.28357684180639275,
 0.7975275297907738,0.9572093694853044,0.8816957264677799,0.08425972050207464,
 0.7758411731652406,0.1897558763543864,0.9850000735060936,0.7357476184062847,
 0.978609504903823,0.11500832534755567,0.9299169021647864,0.21668120812120806,
 0.7890411545533811,0.33977794172716136,0.5516078018893362,0.3524779676620512,
 0.3958479407926274,0.5172009998681464,0.578744084388478,0.4422290321380139,
 0.7102041634388976,0.25787193969520117,0.415596075869761,0.8116848570613341,
 0.09225334305607447,0.7190460129428549,0.6685390851244962,0.8519998030954535,
 0.6363525661526119,0.7178398380621708,0.3523740415736504,0.6526373116557165,
 0.7443975050062354,0.2698430269951586,0.40442274820463,0.8246240711549879,
 0.3352881764931297,0.42748775156808505,0.9142944685568479,0.9092778751887449,
 0.6007515649563433,0.27071673576866395,0.8763081297217782,
 0.39038282613747666,0.02132114601199475,0.544080349677934,0.7000839148951997,
 0.15711765516986886,0.27547235984515384,0.0373129210668266,
 0.9636547026472316,0.4411516116423583,0.030135037590468983,
 0.28472435967469556,0.14910316872699425,0.6979059434449583,
 0.3368380104077301,0.7531427893682073,0.8824419134736992,0.02159514229600623,
 0.19082443307068297,0.2735896657783,0.5996214991437477,0.7976092041703341,
 0.7552793600308196,0.16702539433134,0.7379907154733972,0.25467725713587186,
 0.11456930534196708,0.23043570014079728,0.4113847396888778,
 0.8182182970005727,0.8797022910721011,0.9736101284155037,0.6139392992755702,
 0.019975509413781678,0.09251448105173699,0.9126822115720943,
 0.011996232834753262,0.29939618553461855,0.7770984446528045,
 0.9642714351311042,0.9517142237495511,0.565795309401662,0.8310588463827162,
 0.46359575929517227,0.6527177292747532,0.6726930191010367,
 0.13438332953468068,0.486113678984877,0.062461133903378485,
 0.9926538162971474,0.9820585843187775,0.3051426183769713,0.4229853751606836,
 0.006676387136527739,0.33969542218308124,0.3350775797968678,
 0.8825654580988312,0.6812115923201018,0.48872006641528487,
 0.05002258237529966,0.10758772698390828,0.29857787975521977,
 0.9981654133952947,0.6940682054648953,0.4041456247376336,0.4279683717119199,
 0.46890160568561867,0.1884706270147013,0.2235151652510009,
 0.13806152455631027,0.6060150700416291,0.35031214671804467,
 0.6918008281400391,0.7150917642872522,0.5329421988236647,0.9032764434014782,
 0.40941485258738863,0.1505254006721829,0.7347888793797832,0.417177037798239,
 0.4251613259482503,0.9815452753397853,0.1302037048342365,0.9232032312538307,
 0.7171107754672202,0.8951672422891871,0.3553700287936311,0.77581838208945,
 0.1997436339574753,0.76413193975396,0.7659524369332635,0.1363839330399459,
 0.015801074845132823,0.27243122071489234,0.29201782301092827,
 0.03968658731201535,0.6043166807035824,0.9866230681145415,
 0.19560751624997896,0.7644854946762388,0.23355267043908046,
 0.7465278929477974,0.4536745968034076,0.3199197526404204,0.43727686871069804,
 0.8149895197742083,0.43144785743804626,0.6606791673066289,0.4264260151767839,
 0.5588988115941049,0.4291652446506462,0.4351297911422811,0.7605997626169076,
 0.4778047100867717,0.17671627863335992,0.35801043307549396,
 0.2310521834672189,0.8492207127691975,0.36477691901427267,0.6530124811688688,
 0.7311989911596307,0.5340901220584503,0.5619841350911843,0.506631260270084,
 0.619367281397285,0.002953352910504692,0.48052038351320314,0.693141191705256,
 0.3532085279775198,0.19300699818257339,0.859997087002932,0.6441196293518472,
 0.40508430166652254,0.13634511062389287,0.049082654899700984,
 0.9011519492011787,0.6435867283102304,0.4417124761125334,0.21186582481634053,
 0.4529758254797106,0.15863341499948425,0.36985772477471657,
 0.7507660660888471,0.688765221331894,0.778689728184924,0.7873607783440901,
 0.5847125615420472,0.1654901334731862,0.03926098212693785,0.6520295317668152,
 0.6318977711842421,0.39886459864454027,0.10414313277660803,
 0.16385497599805743,0.8180222097772546,0.786535770249228,0.7184683438352295,
 0.6222944877307519,0.924732641909414,0.6091293949831726,0.07526319981963492,
 0.8197444755116213,0.4785999996330079,0.8349231782897434,0.6490612765516166,
 0.8967167717920821,0.7312302148448477,0.8672523235308791,0.6127034919445044,
 0.5114971631657245,0.39585853171763774,0.4030973192899143,0.3515498622850207,
 0.6759778859172815,0.4577103243693501,0.5177719058685369,0.6695646588652071,
 0.4321021481892332,0.09912931594329799,0.8553908861204754,0.7910613691148007,
 0.3151210041721193,0.13394026167801987,0.6692865134345571,
 0.44241648216293017,0.017096328843805186,0.5383490735674163,
 0.9210670286197403,0.8270444876326597,0.7287409762329276,0.2962909295889862,
 0.7239340939074378,0.6819786931537409,0.1973906620255561,0.075133292207189,
 0.23614711746137307,0.7786882923816864,0.8235157628491276,0.7745729770222892,
 0.8371110978399523,0.18365484840922353,0.2645396250383937,0.7261441245655792,
 0.682472545533706,0.9363382627454235,0.8983486914227187,0.28588069556620843,
 0.9093029297240826,0.19297833265711506,0.06617440019160237,
 0.5035404666976963,0.6027939457797014,0.9461232935816256,0.789235243281347,
 0.912324141019961,0.6504797666813623,0.6454684894897676,0.49804952189894824,
 0.594223774924223,0.619896384379765,0.647299683714367,0.10404906224455501,
 0.7237766396273297,0.21174275188587566,0.6568057221972451,0.311830843360538,
 0.9498623552244001,0.6630319059451322,0.5718497843931479,0.7442364063580392,
 0.9320385694236921,0.7513037946504517,0.9976257781085575,0.8916622105914496,
 0.19406479967603896,0.18642995662372241,0.6921029985988377,
 0.8800061863373557,0.29006768395326055,0.2754522799503887,0.2692777140898559,
 0.5590483496601144,0.9047122252629592,0.4767008784053398,0.2828198472627834,
 0.3363648944136055,0.09549697463085671,0.6648815057782573,0.5679820930156864,
 0.35891041049728334,0.5730184124545096,0.8678468888900148,0.5529975499380534,
 0.17973752567530044,0.41525259099726197,0.5912588048365182,
 0.1110697450432454,0.45168895670890263,0.023689991308194824,
 0.6856711078939659,0.7021648732497666,0.14080138364196604,0.8473139687973608,
 0.7246776526505977,0.2460387997465494,0.37561783459351616,0.6744527847185862,
 0.465629356547543,0.7195256283346848,0.8251593027496786,0.750634134393729,
 0.6689603337266762,0.17572075272367438,0.07727919526987925,
 0.08944899078130275,0.5393618841305671,0.9033526863656443,0.8649510807795886,
 0.14813962529232017,0.08576114367947729,0.1202903411469668,
 0.7215976891694806,0.1382439802296217,0.06700695726711459,
 0.09147239975557164,0.2967291633972773,0.7939813227791248,
 0.35200400747424665,0.3352188823889809,0.191228864759304,0.8540108940268184,
 0.805965140389983,0.07803322355425746,0.05189771308493141,0.5217517414428031,
 0.18379865330417466,0.5416844834880773,0.6435288587704326,0.3508977328465922,
 0.5135930612823317,0.6918412779712273,0.10834333551489528,0.679724788075726,
 0.6057165524584845,0.24338075607542786,0.27758739210533523,0.405331123769202,
 0.7917069259028359,0.7876483998280295,0.9488664609188567,0.6101666130309628,
 0.9722173038200043,0.8830735387286541,0.6249485945121206,0.5740471972327738,
 0.31041471227896145,0.165785661861636,0.08574275476421822,
 0.08482304666913798,0.9557849634736268,0.6042293340732682,0.1445874374045858,
 0.8984421905795513,0.9153685552577536,0.23715757879678123,0.9126389154477363,
 0.6582412864886854,0.05810333277452695,0.6929045125622296,
 0.37181283494792605,0.27011094850480744,0.04480547578329119,
 0.32306425157120244,0.5541230027374625,0.02740774386678524,
 0.8965950869059024,0.16292420017479892,0.17216934680485763,
 0.4843789182474869,0.9744912382735724,0.49243995963618015,0.4152383538046355,
 0.7256492743554027,0.2814421567932823,0.3714960646882288,0.7097436481706871,
 0.1704420372843658,0.5585798112039116,0.3999325377537738,0.7222344371274247,
 0.3988009515723864,0.17711841359602676,0.9693934987594789,0.5102019787150429,
 0.8471009311825485,0.5994872688837064,0.2076150364149537,0.6069249967364123,
 0.7428574088328611,0.003538539686001041,0.9809440997302765,
 0.8412845344118026,0.8767073830458445,0.9084990463177162,0.08551555173398251,
 0.11385852708953492,0.6640073883439869,0.24605194605867553,
 0.3030217308005395,0.5073890736656411,0.4651434306755503,0.20389799095226024,
 0.15008145526211558,0.16327282108664365,0.23920527054883944,
 0.5864544651283491,0.4040068701214312,0.49348203715883177,0.8777046015771904,
 0.5938600247150231,0.758252704789928,0.5196234002438975,0.852938795287746,
 0.0903171951409214,0.6618916700639231,0.29005312966366925,
 0.07719249842072606,0.2452029461412495,0.9221219939423562,
 0.10100757380139525,0.3668488175145791,0.8675579630337774,0.8564661001916162,
 0.3558039759637034,0.2999438157746839,0.11014975431830754,
 0.09507382181536794,0.09499817395178023,0.747214719822591,0.8068107409994663,
 0.2567971636101536,0.13776224404752002,0.17262369435882063,
 0.4453973254536251,0.0637754678306458,0.009737895068616775,
 0.6098314059161962,0.9455518148421592,0.2979772795358696,0.5069838749392515,
 0.5765089166318537,0.8995347826763155,0.43851880146337474,
 0.24880158655945794,0.6914111522437071,0.7062918688001314,0.5262684670868403,
 0.20677548946503155,0.5984709078896044,0.39778012414526165,
 0.25775119064834806,0.6964523907000212,0.616267458100943,0.06614179022562106,
 0.21269838423484422,0.5826185666126203,0.4196500431783122,
 0.13463846355539133,0.5803678052234563,0.7543553626797185,0.696970539517833,
 0.1091919858858168,0.30825182612124913,0.030038700351842085,
 0.7339342385779174,0.10422388736545918,0.698392120348018,0.15046035251653322,
 0.5165674753445398,0.21839270212861717,0.025799549642765762,
 0.020722002867928913,0.9514232260645277,0.0298180946044434,
 0.8970776139765015,0.056133058704158745,0.6455226374058457,
 0.1352082781169276,0.29320292759470645,0.8896755140246811,0.4237919660766073,
 0.5991947310409298,0.4482264892667496,0.4239342070763562,0.7739366136926368,
 0.5344354630125294,0.13650111647736185,0.9570040147034011,0.5446191781485079,
 0.5656058002026236,0.3115665746224199,0.6474309376192228,0.7529031993564758,
 0.24248177689917594,0.9025214161982943,0.8033746297086164,0.3346729015964768,
 0.5522535044780348,0.7915085495235501,0.9286789887558538,0.6809588914065379,
 0.41034673366631635,0.7487597978077594,0.9444767334093652,
 0.33051481744658284,0.18239973043719526,0.09911679586363509,
 0.19620771874143716,0.21547928042362474,0.333228895636577,0.6113157871458056,
 0.39681095462101146,0.6671048617163507,0.506112099835963,0.2881484390910949,
 0.5782312933688806,0.376709794177421,0.6713569275164365,0.926935458856041,
 0.1317669876614086,0.38675678324619556,0.457128669065927,0.3499319473611874,
 0.7351496467123627,0.11982913465692258,0.5262913329604638,0.7568264919356708,
 0.3025905054459205,0.29435954128689434,0.6502316054137344,0.1595697004819342,
 0.8538523734107661,0.37564313047375364,0.7912276848795352,0.8724766056830244,
 0.8802340052020425,0.1931147700321707,0.5709909467447453,0.5843984671471416,
 0.1350150460554823,0.9372205566173594,0.2897798716525777,0.9010279234077925,
 0.11773692037651184,0.6026315564385981,0.48184090855044404,
 0.47930479434879825,0.030457628565232375,0.587275779349168,
 0.01926210605954004,0.9070626101097784,0.5436843374494587,0.9338889736667397,
 0.369073677668897,0.9779610313944398,0.9171579366632339,0.2547635814063708,
 0.3735442212900111,0.8817605765195136,0.8413970704798993,0.9448187904816044,
 0.4473107334861085,0.3016516599412691,0.205517816093439,0.22763253964204078,
 0.64665438025276,0.49759034497164456,0.020261358791674233,0.9270019480359923,
 0.9745969521104088,0.984422583687306,0.5637081313081972,0.11924619942576986,
 0.8352058242794823,0.9085386438595401,0.9144224732017896,0.41202834521812226,
 0.9643645758336149,0.017770479073817902,0.037123022930902305,
 0.3438362461746023,0.26634743027809793,0.8965066155798942,0.6004722274147826,
 0.5786648438018003,0.2565628719786698,0.9486041425326297,0.822340895776598,
 0.404366763740831,0.8197946931490446,0.0034380887835250284,
 0.5365780566093624,0.621864358795675,0.6803343203058922,0.8568209325829662,
 0.15037409774883403,0.8474282636160825,0.28911582737336583,
 0.6237998713221695,0.9719033100786668,0.6103770254825349,0.9481878632731722,
 0.5245208600851475,0.26743520115304353,0.3927913039004488,0.5304415622352322,
 0.5952845663832658,0.3158133476542099,0.3080910539729679,0.6781046203301938,
 0.2831282315151118,0.5507977919943609,0.07353843632619128,0.6554019678601571,
 0.6574531733278587,0.9371518655702078,0.47160538452290923,
 0.39791820140474865,0.36969065736258366,0.4111952244554631,
 0.8836326179087491,0.9092122617547289,0.6916964416384962,0.5435658525473243,
 0.23192880911757408,0.5158971506275092,0.08697024107146756,
 0.5542245066587324,0.9542777061241065,0.22037126166085863,0.3712950302014407,
 0.6290922232475017,0.11460363548811792,0.4834847696240304,0.560117502472031,
 0.04418369612235029,0.8943235648260957,0.6319891613371985,0.3157724967638167,
 0.11982873490392243,0.7048257206259394,0.972799454306068,0.8740786377352391,
 0.9446163439942702,0.5279441424088919,0.7324845380317312,0.7147587274729529,
 0.06974426219338781,0.01623107702480564,0.7834466241342942,
 0.09145633284560761,0.7965218353073414,0.1647864483360898,0.7039344319679787,
 0.7935268544491085,0.6865094436986312,0.08207060589580262,0.9605497202467302,
 0.7296066002546422,0.7091859459874807,0.9424979804176918,0.8288077500423635,
 0.44736627089255787,0.7658273637813461,0.9240762337581382,0.1288230411052097,
 0.4327651447936016,0.02386732886637899,0.5276159152302341,0.6051555762573292,
 0.6983053195716726,0.40975117954006024,0.9216420769823365,
 0.17881806995878202,0.12494590312745768,0.07488807567563327,
 0.7371752306597886,0.4976026539070849,0.9969987539845091,0.8038436854661677,
 0.24896762519835958,0.3239415184981117,0.9322391920309816,0.764322305013225,
 0.25099619188073485,0.45367087970891884,0.7166451857832223,
 0.6524246428012198,0.952367139124839,0.2137287398658927,0.19579916755882731,
 0.3382811298784153,0.06128240423781528,0.8655387056767931,0.4854044384917573,
 0.8101886927600368,0.6008293106747121,0.8121618790516002,0.3382420054993206,
 0.7514136912407828,0.8875733647926808,0.8968878120957706,0.9969542684785401,
 0.8402274298257684,0.8370546303267058,0.33080220413040673,0.2559336037864851,
 0.8028522902097828,0.7162567268757131,0.7663506720964336,0.7793068422108295,
 0.41477057659898064,0.6411289488480487,0.32095538476774965,
 0.4593495208249484,0.7771644482153016,0.07676314919019811,0.7894199415496977,
 0.7700322085042023,0.6903614337676169,0.5591911900410265,
 0.046717949676681414,0.6517004413407866,0.5101354693906349,
 0.023107369904970865,0.04659913857039477,0.45365399907397685,
 0.5349295716531322,0.8640000422305485,0.3192970309074131,0.6956370389299803,
 0.8071326584986205,0.4736837724593771,0.9854096164875839,0.46392574073910564,
 0.46058956355659086,0.36218857261491866,0.4967398343371475,
 0.5360045954491907,0.6120010076338291,0.44437350474301684,0.5165716562929821,
 0.45758033794569697,0.7663873538717051,0.5617333045903719,0.6377999539713912,
 0.6240751044314325,0.5498974230691203,0.4655556425684235,0.1587908556011962,
 0.6021166918956657,0.3579614957823751,0.17225588994452712,0.6442618235305985,
 0.3041849938153144,0.3639539225860041,0.2810001878888567,0.5143034452570764,
 0.025720304142004435,0.8278322959779609,0.9145199232756087,
 0.4025804555677762,0.1398724410083788,0.15228548733716585,
 0.024709320436898108,0.40088391157913483,0.5130299554026253,
 0.9349618314673731,0.4645674688353232,0.6837905672021933,0.895032911387149,
 0.7306415367100261,0.5580463273081965,0.7907389447151694,0.22620339355278762,
 0.6054380634561386,0.12418332714820945,0.8365021516091629,0.812362658936396,
 0.10631893507735268,0.051093312370849686,0.6689079039215357,
 0.4063684105080936,0.6339944341522764,0.08160774484919753,0.6976799201550588,
 0.49122735617011204,0.274880049090508,0.9802872009930274,0.8510755889808316,
 0.4642167518987885,0.22888164673499578,0.36146382774194485,
 0.06234129723336579,0.8141893893828386,0.3032256043175987,
 0.04385096456175952,0.9968133066751318,0.6238724037454169,0.4953941378419394,
 0.7649010295232509,0.25021832916512343,0.665087194651397,0.7340803080843524,
 0.37337952399128227,0.5647105283323647,0.5743293493846935,0.392743791965638,
 0.2559940384318915,0.6970815956901746,0.20969212554111005,0.4600421683684517,
 0.45848441419424923,0.8467802294560994,0.9457089348971957,0.1450331638081831,
 0.10257581647076186,0.026224299977074983,0.8262148795287708,
 0.022823209722850124,0.3421713128643371,0.5051375827899041,
 0.4747295066470848,0.46557547479104444,0.5987595571416914,0.5706892231824393,
 0.870540021909865,0.9913177244363391,0.47570575337376864,0.03700852315770653,
 0.6649377031183562,0.45615328234924624,0.4659953899033691,0.3321449976824967,
 0.2029955321331648,0.8315892048705114,0.37483744305346034,
 0.30090264132325495,0.3824733708589544,0.9065562615271473,0.2530155931601351,
 0.29960569485860034,0.028261515483869903,0.806770424894905,0.483323975862777,
 0.8608027226406358,0.7651151200143014,0.1461448417110336,0.9834957018508332,
 0.29031272510870576,0.9179589054796068,0.577703943423326,0.2645831015423543,
 0.03182998826669048,0.8586020460702102,0.5464922435433606,0.7123601536129699,
 0.630088840905147,0.548754398056411,0.9557339289480853,0.14699384096000612,
 0.8737546467455114,0.6405358147795774,0.2767450349495292,0.3475779032087961,
 0.147177142266797,0.5317098833755428,0.8991851946992554,0.19450876128495143,
 0.6845666122162232,0.2982689930082476,0.7548015652341097,0.264217428141608,
 0.5417856596995348,0.521555043372543,0.932472373879472,0.4948738584447312,
 0.8634743940370886,0.981947133332707,0.6159313814645312,0.7210901154570897,
 0.9751257834960967,0.04873850951366121,0.00936953359357695,
 0.16891905372947658,0.08187239904022625,0.22832315148763815,
 0.31210570677401717,0.9210845886153289,0.0490147868189601,
 0.14885595812058983,0.8845556370463769,0.9895527478056807,
 0.39223907101245226,0.14714632352758494,0.3512625168667852,
 0.7171745175282846,0.999120142193329,0.5421730910067719,0.9266281947055003,
 0.4044502610084978,0.06768180286659431,0.23481533377657382,
 0.7732727236858978,0.8238955678441837,0.8625301389488792,0.7958011382342618,
 0.49730290669266586,0.4794871260372101,0.08973324562350082,
 0.05909539165540467,0.7104145811494638,0.7965717447255986,
 0.20023006791146103,0.9786118164803934,0.0951770808006902,
 0.46497862847102445,0.7525498751480213,0.33406877935387447,
 0.6783562517472094,0.28947582629796065,0.9731909819673944,
 0.24234321647999146,0.3604133058549306,0.7516607304232596,0.996295815915871,
 0.025775794687588505,0.03590449735816392,0.8985515748752607,
 0.002537165646839923,0.030440430682531217,0.4203043859684562,
 0.3751577458053039,0.9725772089840938,0.4675894269896501,0.8477158457990015,
 0.7262194041538846,0.5226690337246683,0.15742886866687478,
 0.39858662976154746,0.07128949458904044,0.6123464998168678,0.788880772901334,
 0.3063940405197758,0.9251507572991465,0.25795706122471695,
 0.43517555469329716,0.5280830667781485,0.26565305326117816,
 0.10994449461398947,0.3020325291281747,0.9220007657080522,0.9667812478024547,
 0.8581365403829823,0.3772883400937519,0.03832363826296814,0.8770557997811201,
 0.4633044446995599,0.23061212825903554,0.743588693150852,0.2797206114624894,
 0.30749804480378373,0.32565227579886746,0.4198496429162446,
 0.01867510253398663,0.8182207440776967,0.28245196827443353,
 0.5941498937551808,0.04328130952890008,0.25943304530395206,
 0.9654355148508235,0.3148584071492706,0.6626161373992667,0.636713739312748,
 0.4138727051957334,0.4862790711893491,0.6879238105040104,0.687944305946976,
 0.9885481657215062,0.2132391823592955,0.09057474220834405,0.5802714985970376,
 0.6826298542153939,0.7786520779372996,0.1539122438421655,0.822366567178221,
 0.4283698537198909,0.3507065200765661,0.46240943016475355,
 0.45564945349315367,0.5117290591486572,0.6608642290459386,0.5456855436707883,
 0.8870050153690053,0.2984343818734122,0.033263712556973135,
 0.12491315589833785,0.1860123043606985,0.9562537240750968,
 0.31892284689584405,0.9665318594799289,0.10005596548136686,
 0.6784892084581199,0.21653587731524748,0.9159602933391245,0.9956080297571941,
 0.9263087140876954,0.6955873176391385,0.697483475927702,0.11153760677901192,
 0.9210399329141086,0.1981190506900512,0.5853097411168654,0.5051894173518668,
 0.6610993894447931,0.056598850932510936,0.8796086090386058,
 0.0961129565740998,0.23056892929783435,0.5066312899588986,0.5819815743250268,
 0.45480615870566155,0.7535570670432983,0.74537287401588,0.3753401307946167,
 0.5936678350432764,0.769742039692896,0.2398035703019955,0.05783631898894781,
 0.27071503809408193,0.1734426798650217,0.7874397173908388,0.9800229591024662,
 0.8753080811153329,0.7463800263105469,0.45181745260377704,0.5783594665840643,
 0.2729642057337084,0.12587401850697089,0.8653978398328699,0.2976059355364955,
 0.0035099433449562456,0.622108893098131,0.3041367150616593,
 0.9992558502403507,0.591824793461245,0.8997253802577143,0.19317672437093647,
 0.4351114224753796,0.6266860863965544,0.9546676449470101,0.7434411083083021,
 0.20207962234452104,0.26888516117724326,0.6898288082303877,
 0.1823452200658109,0.6536629729461224,0.0028257625838952194,
 0.7344780523691659,0.45173249747834876,0.9200408109827949,0.224902406903303,
 0.23674517499988346,0.06403789466902932,0.777741139424403,
 0.04022991295527756,0.8203767686672478,0.4028243369800284,0.1490532592007554,
 0.5502965187041376,0.6410523581595033,0.634984844958559,0.6978635916279342,
 0.5046877000819466,0.49639209961191666,0.9095881350890372,0.6045224499791524,
 0.01799911269479304,0.2095571819157025,0.08592668880574927,0.238954834212176,
 0.4717329326600962,0.9665306514308423,0.14491832223719092,
 0.39540897213417936,0.9104242823815152,0.696628312685109,0.7461490850861053,
 0.9977587105483788,0.7432116723406855,0.7238148869343959,0.9124459255509554,
 0.6110848996253104,0.21514613188007026,0.343524491402427,0.29711169664052983,
 0.5374942921480642,0.15103090267800534,0.7058229516832731,
 0.19750962446640086,0.8186833659509939,0.2270329280428106,
 0.24652139869983714,0.9410025849202702,0.11978394405170156,
 0.5992650936302178,0.9548638080506777,0.7694891088977978,0.962625734411291,
 0.5913939437396382,0.10540322199391805,0.4203063242821077,0.9039629790548407,
 0.6926038598496536,0.15428299410099977,0.7906871518444009,0.4170153743089471,
 0.9391719153060917,0.8998360879622535,0.6379731451766657,0.9369234961838747,
 0.8608564155891926,0.30640526038962257,0.7362683998168978,
 0.22662440486985203,0.770726946831835,0.42867119245105856,
 0.06401243790163835,0.7546672041184779,0.3558613645016149,0.3218500344803201,
 0.7841878469765271,0.4971895127521948,0.3140903260725598,0.6505590845928609,
 0.22101688542225506,0.6735796072415892,0.05518489276156502,
 0.06407354337155935,0.7745211780700446,0.29837928721547746,
 0.1824746696171301,0.022627094121619562,0.5985967463219645,
 0.12534222253310723,0.46480914845622234,0.578849484013903,0.5823579975929549,
 0.2533050947901768,0.5406909438890174,0.5236541006392968,0.43965590644932195,
 0.35884738436934416,0.05820411975258277,0.8210764581799749,
 0.27779378198218874,0.6238443688052745,0.578628614809991,0.801607884018699,
 0.620824025841013,0.9300294233949542,0.597088330193786,0.593498986313785,
 0.2516102885347766,0.6983646425883845,0.17616152327254997,
 0.32772564097306467,0.7117044249924134,0.7494605643246743,0.9373876301881223,
 0.6491705738771517,0.31795021803892964,0.09062373482163455,
 0.03865946461774217,0.16084335661355964,0.4169853853718536,
 0.46333464636282207,0.8685422258190794,0.35491133119342,0.8002729491596475,
 0.6571639648507611,0.6798807471263694,0.1879802914091958,0.34806484647044067,
 0.19833168140786006,0.8114103844115138,0.1511663094022394,
 0.10232343985692427,0.27943966496759287,0.3055526458077277,
 0.5792817258744943,0.8862373166285861,0.92877490296061,0.034737792016958036,
 0.48596526936607776,0.4742177982727567,0.16952025296045625,
 0.8737956893098294,0.5993084986015353,0.8853893157962968,0.29712312917960915,
 0.12187162217645897,0.1647720989544943,0.2653095653688887,0.737077110511166,
 0.9244868985207497,0.19139879790469605,0.371220115049537,0.8898045769839334,
 0.38385995463148415,0.59496674023532,0.6076662513782494,0.8167313136871815,
 0.7578627855762847,0.014694658812053785,0.6260903493417654,
 0.8155368735131413,0.5391965686884382,0.7775681888634355,0.33214625440455103,
 0.5762809776666957,0.954788965983749,0.8311827173147015,0.15017012404904395,
 0.3442853629843762,0.33081112115568,0.5049439477264457,0.5390689756916627,
 0.41239877234533606,0.7699582443174894,0.8049252355140633,0.3328099557220564,
 0.6935542317413457,0.08145638598138727,0.5249923374364833,
 0.31730803111594963,0.4523495580329283,0.8533225334873576,
 0.0011947857218223001,0.2566226321908466,0.3679197203542667,
 0.9655118246014077,0.6112516294958533,0.11873268052844355,
 0.10383144522793963,0.4565574145554545,0.7054069867858829,0.9878826588680348,
 0.5648186129876751,0.17013686336314116,0.9356950217508294,0.8192665017780945,
 0.10760616932126399,0.42191013455648274,0.7018841744228569,
 0.4392513116788995,0.3589550996229799,0.9027413908229817,0.27101385876696094,
 0.10390999127751654,0.5361379420451081,0.31508035920885735,
 0.3124800272595275,0.8199178897527027,0.08865239961656302,0.8596998039361252,
 0.9009228529927703,0.8967333255872932,0.943483359942135,0.94990553587911,
 0.19780773726062972,0.07589810167598243,0.30715886381878077,
 0.6194571565820375,0.7075753937009287,0.8037602784956881,0.37957222055230666,
 0.35035735149045144,0.9123563214609833,0.5985996117397594,
 0.48021288069627943,0.5646991647825275,0.04993401035401401,
 0.9297375404963811,0.7561250648282072,0.5467893088807524,0.10835017935884828,
 0.7717967792544228,0.33122404562315466,0.7807302113478732,0.9665081309458483,
 0.549464262118024,0.4408501134037164,0.8875584780789816,0.536348124381886,
 0.9918277477764361,0.951635138124282,0.40981555725927654,0.9863690182121422,
 0.8834992526608467,0.08070803911474145,0.072693685130779,0.31606647611633043,
 0.6913008156785462,0.3626885087961237,0.923471011941707,0.04330844162896996,
 0.07600793032287789,0.08935570212836386,0.5147319119490575,
 0.8684920576378853,0.025813820664903897,0.5297337999160995,
 0.2314627729490425,0.1382397511639899,0.3542619727829964,0.16846121499736555,
 0.5951606836965262,0.6931646088814627,0.7158739818613137,
 0.024591091935694975,0.018936175542299538,0.11749824085972915,
 0.9224138154852077,0.5446101708571495,0.23805137661328124,
 0.26844811733625196,0.6969069741678798,0.27356234500794496,
 0.9864439639765088,0.02112903249059561,0.5721317242428275,0.6005396662797744,
 0.3791988897856531,0.5175857923167483,0.8359804552360729,0.15159651096569482,
 0.5727679596340508,0.15518850035274157,0.648395451678132,0.11200115321445658,
 0.8451908184513688,0.7237659224285485,0.18420285058815522,0.7581917606224695,
 0.449702512557856,0.2681004254057553,0.07328263679665659,0.1506373418988779,
 0.5296558159212115,0.7485604945913068,0.2692203152567032,0.9250886345213096,
 0.14270383602478376,0.11741915817510817,0.7429193475945732,
 0.8700264230201713,0.3920779470788138,0.565581722901749,0.017134846716860275,
 0.23825514973150674,0.6203325748174104,0.9765999599521058,
 0.33190423538003244,0.6372733080288854,0.19468031427217936,
 0.48275900805669714,0.37595823047526356,0.005612101230623723,
 0.5315453384958382,0.2445792842869885,0.8565759023791832,0.6390629774557886,
 0.7283763594107978,0.9823064586476222,0.3109836253725462,0.3837143329563618,
 0.4725686787101179,0.4233791635330555,0.40504665059197487,0.9503806500677284,
 0.7153521117261631,0.9445046194375621,0.841982243678534,0.12680144780300429,
 0.15773936282878764,0.8514114310364316,0.4045240309817061,0.5441794917417859,
 0.7133556534747789,0.7122951674114394,0.9989273078306031,0.12649017576617905,
 0.5298591855240662,0.43513637756457957,0.16204127123828527,
 0.6470470201816205,0.6923751294469017,0.36457970554577834,0.7486079641358949,
 0.9756846592691536,0.9111023228387521,0.31519516540104386,0.3018660976041777,
 0.058915868117295656,0.83434494865341,0.7737785745996358,0.8218719829848091,
 0.29730508499317965,0.3266791776887802,0.9651891677444182,
 0.060133594688840386,0.6772207406990004,0.3438165933827654,0.870207111610217,
 0.896662898319274,0.7982850226589511,0.9417470387570643,0.6701571238359344,
 0.27858813230005086,0.36616836099146166,0.6535514154166291,
 0.9354897968193157,0.4640810124980308,0.825700874349213,0.7589241585572397,
 0.6069912733908454,0.342028763055097,0.0769889424329625,0.6013192159294961,
 0.6727294974106576,0.8017029125854532,0.14339932823590573,0.6359630896750473,
 0.7344949705485885,0.5029034854504425,0.2684560573895657,0.151173036303067,
 0.9398185863589413,0.6366808165272279,0.3285899167154982,0.0744840832772301,
 0.7004280731675153,0.557439441907302,0.6002706997696323,0.3616026747802129,
 0.8856779466773232,0.12389214776952351,0.8441851401950616,
 0.02880516696943436,0.24485553380894542,0.9854724737412173,
 0.14905135251486656,0.32052702625972207,0.21691324373340826,
 0.3364861029155637,0.7342361936451928,0.14489233932825674,0.5897121084041171,
 0.5977822807254412,0.08720880619569737,0.21719638001450536,
 0.6522063183971677,0.016315386205873028,0.973667009097916,
 0.33714016802749125,0.9377219250564345,0.3434005695819654,0.5905029271707419,
 0.4322420511586268,0.8305946053636826,0.6323073756041042,0.8095734822289283,
 0.7865326012735051,0.5002575944622278,0.8722813749692172,0.37977109433135003,
 0.579536978212623,0.635667874272138,0.7855178832271872,0.3028357236855467,
 0.1878706944072207,0.044301753732916116,0.32079764074032957,
 0.17752395447816027,0.07217320144569705,0.993750895785583,
 0.21203782809778082,0.21323438248362603,0.6863339797888492,0.420094481336341,
 0.5381204042569199,0.2787964096473282,0.13959776532627255,0.9626429474587181,
 0.16991738781175425,0.31237309230815147,0.6835138487774789,
 0.2110739869991105,0.7416191016300844,0.7606996860597675,0.8144347656008335,
 0.4031833883751266,0.2065946605476301,0.855368890686121,0.9175415079520424,
 0.8876489527868398,0.37072522822458154,0.9175946353165966,
 0.017205795890406717,0.23339850787048855,0.07138662887242297,
 0.7929068260571124,0.1593292186247911,0.8167547873954284,0.9209356682052221,
 0.1468812859589964,0.03290996910455957,0.6235321473340093,0.9158784589409965,
 0.7967462163689442,0.1550338824861407,0.7738836759974609,0.9171674224078914,
 0.23827871027749914,0.041702926306955046,0.3005177876609383,
 0.4161505307022819,0.9747498351518509,0.4147814135083656,0.8061499710921257,
 0.5392580175163298,0.6885887409358948,0.38111694482261593,0.5788397821710697,
 0.7083764498683031,0.38972117041705134,0.24636560031641797,
 0.05052670139412263,0.012076978905981806,0.18961294823439712,
 0.9405284843139565,0.8131644247734182,0.7410802450309948,0.8570640504714726,
 0.12671625034362455,0.7049201577217356,0.5393154630636701,
 0.41117000439425355,0.3725115157190234,0.3069841989658523,0.8678353145900267,
 0.8278224813549423,0.7733251125312992,0.548234719687026,0.504182911311294,
 0.33594593068799283,0.14899531816601086,0.028634956550310342,
 0.2566737924915453,0.6797172042191243,0.8250824223291687,0.2122739249840866,
 0.6007834872897206,0.8630969344458467,0.16568082122634864,0.6373776884303237,
 0.7378062684601738,0.66100508043045,0.9683664511274866,0.9611149613649482,
 0.3415642960400098,0.9851635550090629,0.7317154007373511,0.6898896766976967,
 0.16382263396180963,0.4056620488396828,0.3847731339488225,0.5879224061067364,
 0.021942656970760255,0.11238745669357753,0.6670038876206004,
 0.671797362982703,0.6273965518349337,0.7329851439458359,0.4908907957307387,
 0.287868515179872,0.7046946174542279,0.732802841624177,0.7577656431949441,
 0.8183659048161135,0.3063910674274192,0.1929724349674573,0.6893133242610402,
 0.2775655604873991,0.9731126598469444,0.956468483729627,
 2.2988454816486037e-4,0.6095515571743365,0.39500467447483634,
 0.2567465213812635,0.1112871422764029,0.678558864324802,0.36184249573264327,
 0.32600861567431605,0.0674574350050764,0.5190209214595088,0.8734788913631817,
 0.6538985400806767,0.9046508870508522,0.46521979903639277,0.8179950679817263,
 0.4461629389281774,0.3691391035206323,0.3207291021941969,0.25411417797656455,
 0.3276832793042487,0.43141695239957234,0.6021753042248095,0.8957739772058533,
 0.5901046528466128,0.7116938165079909,0.8459786374695974,0.8170187678790796,
 0.9228305414630328,0.06762931547568773,0.22947954259734826,
 0.18563475274880237,0.9686743363836803,0.2375888510305484,0.4140693012578677,
 0.7731463083024219,0.8212905277465161,0.5425365952910333,0.1145979259184089,
 0.38089209313457195,0.1095710629628116,0.09290968400249267,
 0.3369049420617366,0.670584791752725,0.9177241371886898,0.16623791860521875,
 0.17012116660548615,0.04020554682186395,0.6223213809880577,
 0.08163311884129576,0.27149466249589405,0.22317594939630103,
 0.005195600499185771,0.8995575248915635,0.8267059709007073,0.389064588242659,
 0.6437399171815398,0.5853217460673436,0.9302658101838177,0.5006820921399435,
 0.29870398801931963,0.5173623473361919,0.027062507306410466,
 0.01650123910256407,0.7740264721725024,0.5237174020120333,0.5719025641518881,
 0.3814153398726783,0.19805538068052164,0.00933458735724002,
 0.12064482461948611,0.3097783867055568,0.9526437438847144,0.7401993321723429,
 0.3966230677395607,0.9407157932818524,0.7063252565568754,0.21544541488421498,
 0.620567973621282,0.033956550935038465,0.021257823309351,0.328887031159495,
 0.16094489309706683,0.602238672865909,0.7039803705709456,0.9646522743959578,
 0.07855571420450747,0.6226163350413719,0.17020604269521877,
 0.40082615033714997,0.8854005999534834,0.06084528171226555,
 0.34503871444580736,0.9018112466043662,0.419775980814434,0.8265332588204215,
 0.9831769414509816,0.4418061949786667,0.4171174334565616,0.9698773461240031,
 0.3565360909144114,0.6381808325118479,0.8877273154306173,0.6119057703222781,
 0.7181861105366469,0.7841165069309822,0.7000083070990931,0.7544556332597243,
 0.47295433253725927,0.005359984004103313,0.8608244521528713,
 0.4300403626810805,0.0013041063793421959,0.8834255700201478,
 0.23045538483877148,0.49056936916664984,0.5718803052560447,
 0.05376967128265031,0.4143279115254157,0.1415779803881766,
 0.30539871299449006,0.8057986790661595,0.3773261697955811,0.7141200072308239,
 0.28739591046445856,0.8876670537029776,0.7912798518240555,
 0.06077529182158037,0.5876808828934562,0.555327854543115,0.5004332751031674,
 0.03532068111545028,0.4244578618670831,0.2627958317534579,
 0.40323477952445763,0.0889599200490746,0.4737153421101885,
 0.11770058177610498,0.09278329110130634,0.36259374063419514,
 0.6562121088207872,0.054466000098906564,0.14751430677234456,
 0.32341695026718265,0.236754713276094,0.0831745236141197,0.7975531940096482,
 0.01479610213496918,0.18164295570428657,0.15020480260199698,
 0.6476744329587202,0.9741031094707848,0.9246474526670037,0.525127171762066,
 0.721591470532157,0.48574544543731113,0.9416996276218473,0.7393993823138564,
 0.1841045370976191,0.21915203547266215,0.405418766920429,0.2634265529275467,
 0.060481204177115666,0.26469657405730473,0.4247014183662624,
 0.38398405337785957,0.04398220943667275,0.20749016765204154,
 0.8624315583569473,0.6899580343473442,0.11299062630001577,0.2853229353985819,
 0.9882098154619012,0.718385145622623,0.23010948656655883,0.8503973928163966,
 0.43101667671241783,0.5595305757738256,0.1307530427432727,0.3181920957019658,
 0.7983354363261895,0.33740799300750157,0.0744245846653675,
 0.21047039084041863,0.7942356411738611,0.36371557856161996,
 0.9938266344945739,0.9570610509072716,0.49886621958953314,0.9521824455633422,
 0.5224470524340554,0.3768330153601268,0.23796156853941464,0.3823448866390432,
 0.580862588942916,0.2337894294666123,0.16218596622922543,0.6837580201986279,
 0.06636438841494541,0.5222270918006431,0.5997993204380713,0.1759804936767263,
 0.44287853356104856,0.18329345699449195,0.7128490966958466,
 0.002390791271776993,0.9890752106059851,0.24055697112004693,
 0.7229431664755491,0.8638225042672737,0.4221832226956047,0.7289462073713281,
 0.33680988274881796,0.8723854856850496,0.5820545625254402,0.6639155207139369,
 0.03753272274109043,0.4354761367651673,0.8106967157433088,0.5474058023614911,
 0.6178841328629957,0.8924584680282334,0.13482991570772362,
 0.49042077177646426,0.9937217178131198,0.245185294104993,0.7338944382947238,
 0.753282427442719,0.9748049697768754,0.13681594983776668,0.712464616948696,
 0.0830454752335531,0.3175048716741635,0.5926113699067737,0.3303233928933933,
 0.14002322610668572,0.2537431817440411,0.5903173219818454,0.2900927690326801,
 0.36638285994139463,0.5692774542939957,0.6082038465062646,0.3439403643381841,
 0.19023380278371305,0.13689117079448443,0.2301136789973005,
 0.9547655995397506,0.22396448830413695,0.8814828351091802,0.6204809978902163,
 0.35576099816842277,0.17031511134026167,0.30051789512121296,
 0.30181528431541904,0.12081968431908996,0.48610381208036113,
 0.5108427578338293,0.3949235540088728,0.898980156152169,0.8815691875348801,
 0.6295455637707621,0.7326965767368828,0.49214637942161676,0.7840752582734327,
 0.9261709127634229,0.10190083765343372,0.172817742943707,0.8188857087963052,
 0.4240147064540062,0.9762911651350452,0.9437331287210275,0.3573956887780416,
 0.6093596225372526,0.018787765115203636,0.5799559596143665,
 0.5154438576931352,0.17216941176020162,0.6188169604883359,0.4747943552505197,
 0.46661389882663395,0.12810423722494124,0.7790938044316261,
 0.9300320618907624,0.5206276194324297,0.0701276408052095,0.23004997412939587,
 0.6209927761583929,0.4674338638370187,0.06407989289731875,
 0.13137699612450804,0.10669877097391944,0.540461546698458,0.791142035798095,
 0.9232829255563931,0.8803047670911238,0.6615785991033026,
 0.043469782319177974,0.23462466649403668,0.6200467110559839,
 0.6306705568622555,0.18223027179022877,0.8641332699712221,0.7170317059757775,
 0.4613033525315846,0.14424978468576044,0.2827381569029195,0.1358081459519659,
 0.5107468405385109,0.35382435860158723,0.41396557306529413,
 0.045701955679408046,0.854348271496826,0.15894681333215044,
 0.8789855377890375,0.718344204203568,0.3695366278587018,0.296045443320229,
 0.28665767348540905,0.5398615637908206,0.9794753267562448,
 0.007183739210530282,0.587138416253723,0.2016222876917848,
 0.07209444490173122,0.8715311496208318,0.4465140025872927,
 0.19977795504236373,0.8362051977393359,3.764619281864112e-4,
 0.07828619905800072,0.7497222091395254,0.7903141081227067,
 0.03035257850769857,0.6770623249621895,0.987992954385524,0.9336989289819687,
 0.4002408693419708,0.2299349147351738,0.39849328195572675,0.9435174836635022,
 0.26102577342911315,0.055721213983703666,0.641436567038779,
 0.036641038688297156,0.08485740719766455,0.5698685741066833,
 0.8589416513453177,0.35585077406902166,0.08090681495425422,
 0.9304991268319093,0.9938433675786167,0.25269620343863675,0.9856640071606252,
 0.08682803766775427,0.16763372293779022,0.09642812196277517,
 0.3892233130091147,0.13474508453362022,0.15585298985064155,
 0.5081846912045039,0.6572140354272598,0.11312706822157403,0.5738914208653636,
 0.3354097366673858,0.8450933580681887,0.803788460210601,0.8534611767423604,
 0.8752634142158583,0.04291372552162975,0.2350148814749059,0.7891313773964677,
 0.23317568528203503,0.5457647339560767,0.014516123136615322,
 0.8568631948216506,0.9239902558988298,0.23094506729464903,0.7501943408041056,
 0.1740210631110507,0.11816302701448111,0.4236351179596285,0.9350967923434124,
 0.2093522111815247,0.03121135005395581,0.11921779493807871,
 0.5328056787522005,0.8662330939398217,0.14728396819097267,
 0.10363839766078642,0.13977467540818678,0.5940557175815302,
 0.7792857550344298,0.2193438596577615,0.09661152632981052,0.7006585684432328,
 0.7878134387308368,0.055373837648815316,0.6621109369619138,
 0.6748121298635221,0.01694863647512801,0.8817359582224752,0.387637467177932,
 0.8447428652402167,0.2537621447903078,0.9228758844084131,0.4909312502333709,
 0.19974940177491707,0.09852833608483136,0.1365479387520112,
 0.6249159159084188,0.43468673036681593,0.27512202634123495,
 0.8555942264295681,0.9135426888157662,0.48991899756374857,0.4216652033180286,
 0.32419150230012905,0.8065528963236521,0.6317852266497909,0.4070560125201387,
 0.09520232564469133,0.5472113041809828,0.8291980196611299,0.4180911105704955,
 0.6086096639706771,0.23551704577132937,0.8666019705239095,0.9362476538111941,
 0.4611432879087356,0.389477218259773,0.8405418173504645,0.36822482882390384,
 0.2815976312566816,0.969196188196026,0.6528474608118161,0.30926670606533513,
 0.6454824329560895,0.7709631012236373,0.5032515103407875,0.6838542726757812,
 0.9230612735291951,0.017532847900477044,0.7470682581667047,0.644266260293596,
 0.7134460388947995,0.47142177024519516,0.7039966141305731,0.953003027162317,
 0.9050700792000929,0.7996596270547914,0.9572991572301506,0.04877077921382589,
 0.8655819707390457,0.6523394891460615,0.1707239676840986,0.35174715875077156,
 0.45798740921403136,0.7198218769398627,0.4858471958793531,
 0.20317283251696416,0.1275642584068326,0.9462147687225171,0.9803409017034385,
 0.38404964894750415,0.03521463370111533,0.29923599357657027,
 0.35659906314530954,0.846987040376306,0.10821512086708607,0.3488639969981676,
 0.23902742806500898,0.8775029420385028,0.6918633697577327,
 0.38644632848103244,0.7649913350112721,0.5480847781339345,0.1656332195655461,
 0.2738879158665649,0.2628666645292107,0.7814030909698654,0.6478946570377081,
 0.5744494810521551,0.09795951690298788,0.07956251270241788,
 0.1991015615283106,0.9815013599523046,0.610896530663322,0.424636250387169,
 0.4212331111297186,0.04840772659458126,0.32176683151732777,
 0.8881722643064329,0.638674758703607,0.7802050589234504,0.1876363368620162,
 0.2943294372380637,0.8603217759206204,0.4468170912645225,0.6420659695755478,
 0.0183458891122934,0.7461848695214455,0.5353712318607978,0.6911532074117195,
 0.8649476737052469,0.4029394874654466,0.8775365982875901,0.7195832590887296,
 0.13154743043585626,0.6595634598882956,0.43437138256397845,
 0.04529190943617267,0.8112581905109459,0.22770764563984547,
 0.6125052378552984,0.9773999247622763,0.39294135688241916,
 0.27050339176587457,0.81063513303795,0.8498052911501138,0.5255058289959056,
 0.5599978748330201,0.135712310515406,0.06696664462368651,0.5901459256615167,
 0.2790990757404904,0.8009326769096534,0.9050132668410882,0.9271479478133123,
 0.054601195785850876,0.9879603530933215,0.47629913905377097,
 0.19680575955213864,0.1663910154038517,0.4797228315351151,0.9894095193615207,
 0.19664284241566,0.5113348002676582,0.7900909898255377,0.7627046464522557,
 0.13702251024352952,0.2997055167622872,0.5495435467787282,0.7285275159709776,
 0.985958362400471,0.9834469069279558,0.30674626214118916,0.25155930938839755,
 0.3742112556863544,0.5586937383476892,0.6412757258411811,0.5084764591024441,
 0.9459349898507942,0.9672755496996382,0.7011734659225862,0.06843812921684866,
 0.5202466297078259,0.17801160230476443,0.8407232908309081,0.6500716212396108,
 0.8408588000748698,0.75993329669735,0.40845566881421813,0.09506702408664647,
 0.1494288016478481,0.36049683904449503,0.913967141632205,0.5991724884634515,
 0.20204961235239627,0.841320966294425,0.15983838680924523,0.7447745381440591,
 0.3062802931270846,0.08158127285519268,0.654994753002601,0.39850964036497916,
 0.1295795448656476,0.06973245763278768,0.5598975729064366,0.7087548750436868,
 0.2850996539536873,0.1491387747631302,0.4775911610487301,0.9340213330459435,
 0.09680850873291846,0.43327871144297636,0.7383934131466732,
 0.7247151507130245,0.8826639822693109,0.16490092480299257,
 0.06771249514135835,0.19813480974585918,0.8918763671375887,
 0.7890612891967725,0.037281977912617004,0.2445409346686176,
 0.9065096718826956,0.5022531255429893,0.934249153210535,0.3372001559633052,
 0.7126608652689296,0.7749264006576766,0.8335073778985943,0.7501317930267939,
 0.4418088300335631,0.32477780107764764,0.365454616230632,0.6065593652886948,
 0.9262808083854486,0.5319349290448625,0.2116546666581467,0.934507320539326,
 0.7660993306259082,0.14989408601838838,0.6876670048981156,0.6105962961889463,
 0.16156666130394992,0.9569385504057263,0.8823674652365348,
 0.31130984496394876,0.1837762318142484,0.9420158845047708,0.793485229788678,
 0.6094049486788047,0.5729306534794999,0.17953106536737828,0.8240981761522767,
 0.14110144556222004,0.4525082228628181,9.13198342827326e-4,
 0.9034251168333848,0.37139317771452407,0.11440132751741872,
 0.29361867316776435,0.5931033629915337,0.5947072783708919,0.2849937913946996,
 0.6430313079983655,0.3510633451991316,0.5620755369869019,0.18638462714790915,
 0.23783687915589202,0.04413528474635664,0.6897912621382367,
 0.9135671800976243,0.14736568081899748,0.9777328470972382,0.7738157841345348,
 0.4994777551919217,0.39904670510098483,0.5202273876324769,
 0.011038970166811102,0.038259028953397944,0.6261128816715589,
 0.7370354292463324,0.37956215350368727,0.31230788775499685,
 0.5248178273941324,0.7527475445892025,0.27496759194870357,0.699456393618468,
 0.964466803260746,0.82136496060498,0.9103360217184417,0.4229611281314696,
 0.2717765850681618,0.09849354239043095,0.07419024755497061,
 0.3677748582521623,0.2586580359979942,0.7320976106092515,0.7530308255909063,
 0.09382872760451377,0.09360325684069659,0.6307618928016582,
 0.12734473976234073,0.6168349166669623,0.1998220523889036,0.5139880339272477,
 0.058541646578646245,0.9541288507854673,0.5882319312389522,
 0.08452885128624432,0.267934199969899,0.2871781154298446,0.5867636980914834,
 0.19138308151214312,0.5618351260821927,0.6278170634656305,0.5978457076740983,
 0.801853071811655,0.6460886647223825,0.8586242536269579,0.0981219833893654,
 0.7421918997972583,0.9523532806592089,0.15735840122555067,0.5597630178291049,
 0.10782195445048148,0.12145683557653952,0.41680834142208567,
 0.6670917535593252,0.5795305931818237,0.6592648865689317,0.8226184408466313,
 0.6308849167527084,0.492231963115052,0.11143333388708487,0.2478762526432775,
 0.8798427086473579,0.8019447665142863,0.9725714134215526,0.14574954655024897,
 0.786449222801776,0.8847274753837314,0.4539160900741449,0.28690956028543935,
 0.2153594743758469,0.9095992394839725,0.3627310992082534,0.9500805890899187,
 0.3053660937323913,0.6474816414169917,0.23006103521702348,0.7674886789579016,
 0.11044081153541796,0.6879612807230367,0.5631144101535691,0.5375724588703567,
 0.38508763770238874,0.8578037910045073,0.9811453212275201,
 0.07466267783952885,0.33413455802770275,0.4463950995739676,
 0.17917835210483757,0.8382738406095277,0.5461085232110793,
 0.035546703161660975,0.49438539832989126,0.1772375813776308,
 0.21471482204445325,0.8292445925824985,0.22781761173618864,0.959461542758442,
 0.24766215465782948,0.7409384953926713,0.47344451144076627,
 0.7534858457109063,0.052758089241753625,0.8309967405266339,
 0.23686815832954755,0.2187956339492121,0.26191820729171944,
 0.9399339536109071,0.8387269151642658,0.9565677613815797,0.6455866432014633,
 0.3375862262876631,0.24607300835858348,0.5983024428491942,0.4118018683541632,
 0.2698193537169148,0.637163556842539,0.9357530702057688,0.7647328815624854,
 0.06559236300560645,0.6695146226541047,0.5085974429354317,0.6071229100098771,
 0.14098344548143726,0.8374042753913499,0.22771431535900521,
 0.04219967685291226,0.3170162382352052,0.4386600153894664,
 0.04272083394283033,0.6820693077150825,0.8384713874914512,0.9060043222261553,
 0.9065177861961784,0.45880956087696034,0.21708154504039312,
 0.6283304646836124,0.27621188460718493,0.32806044612210483,
 0.7392884795253342,0.43449155432484554,0.10367291327578332,
 0.3342515318631004,0.017760831050793624,0.4871084332967025,
 0.03343752923254817,0.3196570591219803,0.1364813412371153,
 0.39465259791224516,0.17389138487208577,0.6430872858100224,
 0.7659437295122941,0.5689947229195719,0.8679130177739424,0.7053194182137078,
 0.6650403084747631,0.9678229206013453,0.5769241713174822,0.43334426044381724,
 0.9783329017855491,0.08764050246697752,0.7735831949062403,0.6972814119300813,
 0.8754808844407437,0.8644678583551404,0.6432726157685192,0.9900694145815596,
 0.9988744376749836,0.2910564868291212,0.14987292103711658,
 0.13679165705952812,0.44792151990995466,0.24474866892288627,
 0.7305447028978189,0.21083467034558023,0.8531166912857888,0.8193591631142644,
 0.3358698942740189,0.18572276609845018,0.4255133235451073,0.5252494334350355,
 0.7224870208984133,0.09816074887592119,0.5189833697693181,0.6484402823437329,
 0.7649083662722845,0.3807048124017415,0.8266912393079289,0.14996284190765374,
 0.22871625979598054,0.6158591940090961,0.46633940848126576,
 0.0847958464037426,0.18718307890036323,0.5925653527749652,0.7467892196591516,
 0.7642585781138992,0.006145829397410041,0.7665980189011257,
 0.025086159162005184,0.09309113742086406,0.6792420683704555,
 0.1144297613244698,0.02269714591538352,0.3362438189265845,0.3819886399186482,
 0.2576644692644052,0.8671584931713059,0.12568618613016602,0.5282454982864092,
 0.3908993126145054,0.7361465183235617,0.5457671195412109,0.8083305495039167,
 0.7354481801223476,0.8627657468942729,0.6093257533061031,0.0802035724431478,
 0.8173411792950882,0.17543467217731257,0.20751079440762743,
 0.16743972485898695,0.6601557865937376,0.7708245344027879,
 0.39741779793722465,0.9697725390788363,0.8810774177697382,0.9102131887228064,
 0.7177791185970972,0.6859512146967686,0.27974063361769463,0.5352409801317674,
 0.4398190951677282,0.3009785969880464,0.7230071128395673,0.7216420383095804,
 0.35660874279098553,0.33970111930979163,0.9178151915131493,
 0.4415036891210312,0.7733057261105704,0.5342333280620857,0.6067121129799434,
 0.7634030689001517,0.9967410578249171,0.46364070723955375,0.3099439518829161,
 0.019960873409323066,0.6725819691960209,0.7588466436890875,
 0.48379388278745283,0.6338366758001662,0.8447465824034167,0.5729412661349226,
 0.9883603921787818,0.5069695474062842,0.9792279135531089,0.2089927587718945,
 0.8556287205768354,0.8731468908202968,0.34244059090477796,
 0.39547525988710186,0.6614092356368869,0.6369315451084056,0.3567265218283021,
 0.14395067248519822,0.21181194625247535,0.822744652384069,0.7761486237915322,
 0.7375127126293811,0.9814253028043896,0.5359744387375832,0.13019213376405414,
 0.13911644890848562,0.7636898209135338,0.9113544470311398,0.4095203753629083,
 0.07863186116849874,0.6783554733355138,0.9938498609939811,
 0.02848110044990526,0.848894355005955,0.06383564278561193,0.745761362960687,
 0.17781716197666708,0.1917234472531466,0.03715508501502185,
 0.2789146888403181,0.9429825291077201,0.9663236815665195,0.14564969013278373,
 0.4567246728591078,0.9659554285204794,0.0954977063481608,0.8618334271025486,
 0.8631477975945578,0.38358550318083595,0.06236774886060836,
 0.1643073796987411,0.33899375455432534,0.2964310094340049,
 0.48711138189872893,0.12739614337288008,0.9671160324595648,
 0.7120772825193433,0.9293510000925593,0.43138397756369695,0.7944340833817009,
 0.873003285562828,0.34066769665409236,0.8296492308932223,0.5880873027280188,
 0.7003685191829009,0.30214006663324955,0.6681170226875816,
 0.02120097727987158,0.1547244213428015,0.6203494765358732,0.5411150869143194,
 0.7052966225323851,0.27841562606665937,0.39337383375673163,
 0.9431660515189145,0.9325575813764241,0.7464753584674542,0.6442436887029569,
 0.698870609237855,0.5063583809850646,0.09802634615981343,0.6338290710899775,
 0.39849287584497406,0.9930058350499174,0.974536752534215,0.7263344816365949,
 0.7989968882962201,0.9230754872195465,0.271558329129044,0.593829501288216,
 0.20791389380419656,0.7078985862429137,0.7369877874408673,0.6340563502998114,
 0.5861934736176053,0.9263970907423391,0.1412232054028777,0.33868097122905205,
 0.6340852648791532,0.8411511794326263,0.4568282683122751,0.28106244147959547,
 0.061664915114353014,0.9493319593336308,0.8892426542642305,
 0.22666107541430824,0.7947530091095963,0.8053573458165324,0.956338494596545,
 0.5196665682941873,0.2234498687165657,0.08993859064327137,0.6693076532466479,
 0.20706103565422884,0.457264145100323,0.5622325129181067,0.3240757589060127,
 0.5021464424866644,0.40579362821455467,0.7026160511180408,0.952731751968147,
 0.15905817490607554,0.3931510293911784,0.545315459805612,0.6178011526525911,
 0.8628039243527795,0.29116328741455877,0.5531551138236055,0.7966352857880092,
 0.23929440508302546,0.511251840678248,0.6671503981354951,0.7841636503026439,
 0.0867361668847012,0.7293160742996628,0.539290171746851,0.7756717096649761,
 0.043526634610089765,0.8582731376387367,0.07106217525359759,
 0.3889887080654314,0.8954788601978222,0.9027620797444853,0.8328768319678174,
 0.6971955700790304,0.9137471752214861,0.25865227070358243,0.528994552037014,
 0.28487702186490804,0.1527623842014536,0.34605089683133894,
 0.19108056828518993,0.754666893862634,0.5720144531035032,0.6519070532224666,
 0.9481566620076968,0.04239205462516227,0.5214241267505458,0.4398027097451612,
 0.1093700018499123,0.2915273323030054,0.13199063594734128,0.9943280641315031,
 0.6650460976411964,0.8460466773694985,0.4648671088863836,0.10942650626801376,
 0.5520567335058323,0.9165005859190951,0.6863286809577762,0.7016563229620358,
 0.4959872817092009,0.34095854716474927,0.1897154573459776,0.7515991594027095,
 0.8445791348715879,0.9308460354584921,0.785303783316811,0.2044467229703557,
 0.37039765054839635,0.44634053713014354,0.6601825252034327,
 0.29642822507671396,0.48650014465090463,0.6879795965042869,
 0.7336945730233985,0.994074635845422,0.3834632086895867,0.8106655963833874,
 0.2542021335330025,0.07371918807963429,0.45286858694032184,
 0.4903994749580052,0.6914271590545089,0.05163405509121688,0.2901383258791178,
 0.9719834008236193,0.9967797149940628,0.7502702645285648,
 0.013788260353203685,0.7296371395751176,0.15542631692247433,
 0.40904900339810857,0.3066371218043149,0.11107529037408681,
 0.9107351471113155,0.051263982228407,0.4313218508763784,0.2939134320022916,
 0.18347709977911675,0.8514462480255872,0.91318730069673,0.8903151450763567,
 0.6461336654476504,0.6924983159717406,0.2621547130885794,0.11009415720581528,
 0.32485845807993585,0.5624593416876472,0.26009133471078183,
 0.2901560573594385,0.5823787040348317,0.1554580511358601,0.42700688720732416,
 0.8934936845896083,0.40188655096743287,0.8258909671030336,0.9948560095500598,
 0.9011967119896999,0.9514610266812595,0.26421128351910195,
 0.30463786873555543,0.2908559440415084,0.9519161915048509,0.8032591373245265,
 0.05837678792051926,0.593018549753282,0.34501432389580067,0.9201749134531761,
 0.5596995965180798,0.9822804261357531,0.684018422324736,0.6207750535689235,
 0.9268586896793147,0.11780590894697351,0.14160493123128115,
 0.20561984502399178,0.6539320594838547,0.7336437240891176,0.3235624439203838,
 0.3719289994002112,0.2535139546436791,0.6456842111009286,0.6191776178190681,
 0.21268074597332154,0.7285210310203648,0.11675961174605975,
 0.7700543857962061,0.11931611697508149,0.3368178169413647,0.5602781372178062,
 0.5765730676286116,0.9028283720180567,0.8437686110012029,0.1978651464106631,
 0.13260735475724483,0.16316697581089,0.9351096726326332,0.8815113825154008,
 0.2019221331668667,0.7745634534419149,0.5344320131516241,0.2168847003615001,
 0.9172832263963231,0.1251540617551743,0.46235135957995677,0.5164206286815964,
 0.2877905173816028,0.5882613129148705,0.5333854795681207,0.8366020502151373,
 0.2792563189290376,0.3001562714962386,0.6732540638857423,0.2960234920041416,
 0.0757139984961267,0.21155658104349107,0.10652347335756374,
 0.14775240407806145,0.9507069438120458,0.13816866054556987,
 0.9734039336460704,0.36184218074638785,0.7024647642659524,
 0.15111545137998583,0.03474729721266778,0.7880190644575755,
 0.7914587254877089,0.04745063819943651,0.7924994982353288,0.7579921686410868,
 0.7342014903601716,0.13121121550428383,0.49843662246091414,
 0.5371805222759928,0.180638050232105,0.9443220348079205,0.3505840438000871,
 0.1901517029757206,0.2098329052774801,0.4677786429258296,0.9430822170176593,
 0.14952560054299813,0.7967052621892531,0.2230962616801313,
 0.029595776500489723,0.39447266057848585,0.35961191404733384,
 0.20728656299185122,0.5629233699134087,0.8703452584101279,0.8326581493596164,
 0.6246957094619077,0.05112071219632097,0.5627832742808353,0.2241499693620791,
 0.256522747895641,0.6579319579093509,0.5018684146402486,0.9164637977670445,
 0.41745685604163907,0.5543084784476803,0.549913586002078,0.9189622707387648,
 0.14073538474227598,0.26655825178404013,0.6758285685063067,
 0.15076816602446474,0.2167012139385358,0.38993094819194374,
 0.2816344098296909,0.9676372488711805,0.09788723143696587,
 0.31257574244310615,0.14774674181200886,0.19912008080809362,
 0.9203065829637219,0.34068948573930724,0.1372084603363446,0.6090794979429613,
 0.38516861247857004,0.4721677806176686,0.11797032486539549,
 0.8069544779232005,0.7114939440022444,0.8321671153300939,0.5988616898013907,
 0.646274471626219,0.5472813012942943,0.06152064563136339,0.4258144397731052,
 0.8860527380552072,0.4965371984564775,0.8479102243654801,0.2883412835165735,
 0.11088605835629217,0.684356496783179,0.006889349553233393,
 0.45926508627721185,0.10088143283294027,0.611257104623823,0.8120352090431885,
 0.7966623116266249,0.8703123597927084,0.3758655142492924,0.7768515798775485,
 0.1700588367474115,0.698461887769522,0.5869685825358735,0.5496036269182357,
 0.828740581904053,0.3911310140617269,0.1733638966542438,0.3643831670213299,
 0.5377496082440256,0.9172380316374573,0.6338272665064021,0.49399230706307873,
 0.290581889478316,0.6452901377877167,0.9855078243548137,0.40493513441346707,
 0.3262197346325275,0.9943448486994114,0.695192423318618,0.8457092137039985,
 0.9856752761438515,0.19917537064331126,0.43572551917107627,
 0.4148812339523268,0.6009904743660224,0.707149149324329,0.33262994476727337,
 0.3365346649983951,0.1856197873301798,0.8516684228307954,0.1860218338062356,
 0.13598429696014036,0.6571881627138261,0.824310790723509,0.25668689017725466,
 0.7653293078311048,0.5469436346474448,0.436582098856807,0.7217172315945937,
 0.027262844596758207,0.6404826358235065,0.9997841579171596,
 0.5593921220903215,0.18208898290252706,0.6340521184239124,0.5185005236140726,
 0.11947368298067584,0.3996874646614036,0.984549755574355,0.45209320008307796,
 0.8437386331014851,0.8694798441910395,0.8281039783768037,0.19599198902581805,
 0.2663729580294406,0.5243106468221611,0.4771579912562529,0.1961974070661291,
 0.7394670406922778,0.13754241583239024,0.5577193561859564,
 0.41995247608969266,0.8575305039015075,0.7314600583244217,
 0.04683833038595264,0.8366844797853402,0.5444218030499202,0.503853794280365,
 0.6495934905191599,0.7585536984815868,9.963213767842927e-4,
 0.6370658162259737,0.8756369132652764,0.004721745881284711,
 0.060519763186323905,0.2723575267684403,0.46455001141467445,
 0.5115680388104369,0.8001748279836784,0.5279384509485842,0.20566848840165242,
 0.5466720478845866,0.5512854930123382,0.6439386375379272,0.7191422058039001,
 0.761101018334594,0.03079890000744767,0.5070976646190217,0.2948906235129123,
 0.2215661163197835,0.7921969485959881,0.46469140626224237,0.9176802958977212,
 0.11043754288121121,0.39129873669929616,0.10590021087731849,
 0.9246288046703295,0.5555402929354074,0.31554428333259754,
 0.11681789361284434,0.4821401761143256,0.5091245757775641,0.6119174824045391,
 0.10039005115068245,0.08119211890730238,0.5294938941368527,0.665587387186029,
 0.21968501727466438,0.4947972264885039,0.5217026140782686,0.9505893274240166,
 0.27476449439130013,0.4626128899038843,0.47060434791729533,
 0.6758313683935505,0.8490139755701636,0.6580414346426124,0.8502614873349224,
 0.31669425821092523,0.6118123781540874,0.6122067063657384,0.7254854817056527,
 0.9116940095024361,0.6009808981928174,0.5642617877136344,0.9517109004671941,
 0.21909621216610498,0.5532468033290675,0.8549882466370899,0.588801279740839,
 0.14165395429944116,0.6230132378216481,0.2662258896815882,0.6799530169167645,
 0.6576650032807738,0.034720992451072186,0.29019052755465147,
 0.5811184999589794,0.6868604439400855,0.5510869388298144,0.19783626540236632,
 0.35757483544973045,0.2820977656109853,0.1932796253659843,0.451945002460973,
 0.1901755377094444,0.855444925923113,0.3655421184619936,0.23271839311500386,
 0.8623033637186468,0.42092667079095003,0.9952210902494865,
 0.46888218491457734,0.2571865094526945,0.7214913482310257,0.6574038847895434,
 0.7614983670484006,0.501358148986635,0.3567305498813783,0.5186823346189137,
 0.6223042535542685,0.12289650449426237,0.9166258324748111,
 0.31927589039951054,0.1999199544601301,0.573469519444487,0.7866635109025076,
 0.6216299362885366,0.33757264965987954,0.3439964720621931,
 0.47672340602693486,0.9276265062171515,0.5561041827882596,0.893153550271612,
 0.4612066406390174,0.9513259304061182,0.6341222326003036,0.32240782960912684,
 0.03772623190270907,0.2618265440741967,0.7137414955047005,
 0.028932943721636573,0.3932818149449546,0.7196871104699866,
 0.40264940190417553,0.9389934768694189,0.9140837128223596,0.8619584243232485,
 0.6016416939144316,0.5559804682348662,0.6823418043440532,0.9606992823825096,
 0.24550541672378534,0.945761286617917,0.4091554559185355,0.8308671595360876,
 0.4079761205926373,0.9858786424833259,0.9649490574181367,0.1295404352348395,
 0.8905015772733894,0.6179452238375025,0.5641715694925004,0.4647240901448856,
 0.8365158403315966,0.34553050932155605,0.8652350025974159,0.6150246191924239,
 0.852687181154185,0.5758356507697854,0.8367628337541291,0.35361202585673723,
 0.3393720491942037,0.7378211773341313,0.8930756693840897,0.24571326884171873,
 0.9250365959418344,0.6061869843151968,0.8792500120582198,0.4499635816340406,
 0.6433347674455454,0.8418203024346276,0.005328637544844694,
 0.3278622141850587,0.3384192804664432,0.15383802744010333,0.7550425279645652,
 0.12846794264044492,0.9070760776486703,0.8608459005249297,
 0.22808335070883234,0.03830932568590495,0.8417103122322258,
 0.9688277560300167,0.6614234080682355,0.2514887771468941,0.4815367638559914,
 0.581606133269997,0.6168073930029141,0.9891063674640809,0.735541960815931,
 0.359204328016148,0.5271790521720678,0.11317074370710101,0.3532775281763706,
 0.7666722752025614,0.8501122885200827,0.4600878006787643,0.5858569587513847,
 0.7100271537432296,0.34879555235759596,0.3041394288847018,
 0.24824333128069576,0.5905955604846692,0.10571106669505437,
 0.9312164696699129,0.03447934154626997,0.5989341318949456,0.659836312764495,
 0.7917451199937274,0.9470304021667064,0.05200569187233983,0.866580433421206,
 0.9478603379634015,0.24484165455393225,0.9192801955625319,
 0.41230735433450216,0.29225149838191766,0.4467244403385633,
 0.29193614089098396,0.2866007465326229,0.6380599948542127,0.4474081263744141,
 0.6198481095501062,0.5510099255932681,0.14060606622089766,
 0.13109331706404204,0.1863493572640038,0.05085292701352684,
 0.07530399904574747,0.48623663408581663,0.8112056200682981,
 0.0057554940032524815,0.7774270553723214,0.13494065925922116,
 0.28434916594649706,0.6431980993020253,0.43638415322191937,
 0.8489131776686826,0.2347270019159775,0.015156561574329874,
 0.08272507671268703,0.04887276830648157,0.9398161239522962,
 0.13160495432769415,0.8031188916054633,0.9579727488241321,0.5423216875812562,
 0.875329509803962,0.1744377061040553,0.7727712155915147,0.38033473654906413,
 0.8238292755115375,0.17470175166831037,0.33490762702822663,
 0.5961484177842372,0.7434038235732882,0.1402731899354176,0.9954150246403637,
 0.4101087940993634,0.5808308466993517,0.6295619081266447,0.5282289410726446,
 0.32994151062557087,0.9974069967923779,0.13350154173514406,0.521498845655782,
 0.36116944588601274,0.804704701791588,0.7700222522050632,0.10340951500325868,
 0.5862912261101776,0.6535207838665368,0.49815241843405933,0.6857362305786334,
 0.9498610886773214,0.9415073970850961,0.7490203223935228,0.4701211046279794,
 0.08461980189636664,0.6734207334265464,0.07564179463010301,
 0.9587564917260619,0.22879020671550487,0.26902223505080136,
 0.6602852121163618,0.7438477437358284,0.42243932237996074,0.4111875400465762,
 0.6175825342131266,0.003538206176769587,0.4284341151224448,
 0.7034975744574947,0.7929515249342585,0.021570103029772136,
 0.21312866809145659,0.3459995945059269,0.6275473758654314,0.8891813062317353,
 0.5648193043163567,0.2760083736384622,0.35510176738825605,0.7452645968269158,
 0.6827643763892078,0.14280909114048046,0.5593209697185779,
 0.24720755874931977,0.4445785804528557,0.6833150085571769,0.3294161419824889,
 0.07423487652642913,0.24881169286771077,0.6038873379091731,
 0.5335683303881548,0.19176467789610296,0.5513526013844053,
 0.11513387733011404,0.928870443239226,0.3902061856988389,
 0.036688422208924654,0.2963167770494939,0.31029498857538707,
 0.4232778658447798,0.3789380598611063,0.4393586292413355,0.9963519701296637,
 0.41737059352260897,0.6384394346803756,0.7031219004238256,0.6145964323116762,
 0.8846142741307215,0.3943329838569307,0.23983376609767215,0.537242799613274,
 0.183090482675202,0.2005885901378962,0.6561938968375947,0.7677587455942905,
 0.35595009776643516,0.9826858833620363,0.5248944565172862,0.5244010905849036,
 0.90209093037453,0.5966303875222252,0.5665935758460142,0.07495705425958632,
 0.37521261933901995,0.13189326475905372,0.4006083954366608,
 0.8631385395836606,0.5408626999746172,0.9641326555278498,0.4400046420107975,
 0.8488272362746372,0.6958627820162094,0.0891322289940325,0.47942495645959926,
 0.9694397774921489,0.10175496008025964,0.3645041447737293,
 0.34779560645122976,0.16920861089945227,0.864755575928225,0.3537474392594747,
 0.2977081788350796,0.41499308936337775,0.5880961454336078,0.9594885137799555,
 0.4437145943288022,0.12699559114529957,0.3708479688670355,0.5645971703456277,
 0.8440575341310729,0.8456852407365132,0.3358674894556295,0.501345667659081,
 0.1171041184880065,0.8954182748544753,0.45813782176598394,0.7955101552544368,
 0.9099582008545957,0.19656482718956247,0.7573867051035432,0.9705728692340841,
 0.06830458523337835,0.37820924816341905,0.665479122264379,
 0.06171127249321373,0.999495450212329,0.07997974182273027,0.856967684145658,
 0.7279313365388221,0.9537997315633224,0.760778986520995,0.13658240255042964,
 0.14645438700865554,0.7542590962883741,0.7120741475863206,0.8936230979050551,
 0.46619148897126983,0.7643763189153596,0.045907032374047696,
 0.4345796132488817,0.5139016003424732,0.28277068387568494,
 0.40870431161683773,0.7850396270161956,0.6127368248576112,
 0.34684331976883875,0.005058755010027838,0.19921412855372778,
 0.5257470400334286,0.6003453767443853,0.8005875417586981,0.1277077651352383,
 0.06515252871577859,0.931311439435611,0.4534455973188364,0.5668198117952444,
 0.5988152971428345,0.9946466988866612,0.8891750795643591,0.11097085909493842,
 0.8600323707584782,0.9345967267620893,0.550448738403356,0.8392387516450128,
 0.790105467077958,0.18811068364061323,0.28551251823396817,0.9491360704702512,
 0.9883391857488938,0.4611323118258217,0.9673280744296308,0.21284042929182767,
 0.9627863257782662,0.4634278473345228,0.6391062143408219,0.4405743307409855,
 0.45742172025740724,0.7533018918912101,0.9776680601695068,0.5660740083165079,
 0.4454305535685068,0.409943063196337,0.22125630165492838,0.5642834428966232,
 0.8604526326676121,0.616657630157759,0.4532908644729181,0.8090840391829405,
 0.5625933565695538,0.21265011435660186,0.8793738792474557,0.8868849249600489,
 0.34518152988716444,0.04029187235978138,0.4929341886640938,
 0.7991581053757306,0.9493658288179815,0.3794451284603255,0.10846393538521448,
 0.347818655258587,0.9086839085154992,0.10513235049929781,0.20338143839165512,
 0.5489852684443993,0.810656153498066,0.08037015770641776,0.0335731238603757,
 0.19851462149956345,0.6695982356570946,0.6542907778434279,0.8532234764937798,
 0.4005299929898206,0.24154720069344215,0.5357430242451944,0.6018460876773571,
 0.8088516599203902,0.16144197445315522,0.5177406096207526,0.7493984861471258,
 0.2541983197641918,0.561642483694441,0.34124586736641505,0.2593350868122919,
 0.3586661248415276,0.3365758495604845,0.798208967506881,0.5280144511136621,
 0.7465064486690215,0.03003051184565153,0.6541640664758346,0.5647968662754299,
 0.7283767111129381,0.8733020587524915,0.8631762656516073,0.6894475679387895,
 0.355305215942879,0.5640007551746047,0.21543688437939035,0.8229270221302502,
 0.9405004300958069,0.054017760819589644,0.7565645262340172,
 0.2621176613773296,0.8724469781321073,0.17680761848993443,0.793532268457475,
 0.5195502004376081,0.9509555511463337,0.8042067576265958,0.42008122968561457,
 0.6125368936786177,0.01649061974102417,0.48130968412977226,
 0.6765719142454506,0.7168549010662749,0.07044881779730272,0.7883640444010243,
 0.4143241871342519,0.42889514797198647,0.5882134169024532,0.148825065858467,
 0.35823612567006535,0.359204255996715,0.9293526053461049,0.7276975258502117,
 0.9459994988909761,0.20213875958741612,0.3153556865470498,0.4102655621126895,
 0.1671281912331578,0.5765740102578981,0.9240215010154611,0.6254107757816537,
 0.3290191612094715,0.7661673173253155,0.9169624104314765,0.6572323546775898,
 0.9078017156875808,0.23868778641499033,0.026665209328900596,
 0.8433719623080852,0.3042622423259873,0.6474870748302216,0.1516719972315197,
 0.23958952255769828,0.18512767568593091,0.40076327381818966,
 0.4155384798681848,0.04495569021549761,0.4468788236493354,0.924726454028027,
 0.7890248210820724,0.2500791906711963,0.46434136307874063,0.3621409464476266,
 0.04080425570832791,0.6019792423016568,0.713398385394323,
 0.026810485657346894,0.778872416749896,0.9117712017366046,0.6771400421097133,
 0.398212972497508,0.1236755358242575,0.5845092171355845,0.026226206052324175,
 0.8931263962976077,0.9100491315776499,0.7011862539773693,0.7304466134502845,
 0.5760118040455207,0.3190309145928383,0.41561387181745246,0.8051544494222336,
 0.07170557738520322,0.7509377673569703,0.7395990066919139,0.5689942251687564,
 0.7689574465438218,0.625634319995613,0.9198288818897907,0.8193554296201906,
 0.4228967179218348,0.13142626982022199,0.29562103582732435,
 0.6579563674759042,0.866080981476535,0.7902759917789339,0.33048649021322296,
 0.5346471100298581,0.28285525141623913,0.6642182893284403,0.7712243310464227,
 0.021437501145301407,0.8423274675264703,0.8880136949463117,
 0.8119396115468124,0.5272336873419659,0.2364904280900666,0.5462041719201866,
 0.052889652425297085,0.8813361662377348,0.09477152547781764,
 0.741804245250888,0.6614752490582525,0.5353443928244171,0.744583872760428,
 0.5104496513485517,0.32188730947086863,0.22832880576244496,
 0.9161686846354503,0.4869560012020152,0.35610438272045464,0.2693305841513701,
 0.607396153773865,0.4232056853800986,0.29065789975631684,0.1935569032085117,
 0.8192346340798908,0.5895072499083306,0.7716742486887244,0.757871271945562,
 0.6180409874235447,0.48456159943557164,0.8865223814241708,0.9728663706672159,
 0.21718581764678324,0.8872516305361845,0.780390638011648,0.5410062664238737,
 0.1667520532659028,0.536925831789137,0.03570501020228001,0.8802966542721726,
 0.4779168899916203,0.027820756756407494,0.023618987481977438,
 0.629223921219315,0.5959976390133886,0.2558197031655828,0.07473596965134843,
 0.61178213132474,0.41315025419528273,0.1700353740995837,0.8137074898250204,
 0.8449282671646553,0.03478626147568631,0.82935078963481,0.5701193515596108,
 0.9773358038000157,0.9424803953703486,0.2098837315854285,0.6826896503961679,
 0.975679972399792,0.8072246755976699,0.4556435323504455,0.15810250151321836,
 0.49034799148893304,0.9018447126141579,0.7941964301343845,
 0.45852786109201105,0.6403088558878464,0.1479678934621551,
 0.07331934114946492,0.5464763924276419,0.038103810616645584,
 0.07208116763126182,0.30255607294261855,0.912709702526674,0.2529392548800097,
 0.9328515519897179,0.9699920300533105,0.8624733125771518,0.6379266983867004,
 0.8611586243609476,0.047589748204746374,0.5337277687189851,
 0.6952625103205187,0.6852890626635322,0.38745577237341355,0.8727065994873737,
 0.689153367605759,0.932284396398023,0.3408367186493848,0.8878932377039763,
 0.9816088130675734,0.1752987156432143,0.5223196757351981,0.712885863012214,
 0.2692768517888363,0.17074867241602365,0.48812349794763965,
 0.7357571273697896,0.5077694582570559,0.10478900635834287,0.6211353514353115,
 0.597655466691776,0.7244473441440169,0.8675797896779949,0.11259434571297322,
 0.2721360815961026,0.15168264382423025,0.9980305079809085,0.9429425130032006,
 0.9477963954214964,0.06344899292296147,0.2914420271845304,0.5892710636812069,
 0.8624860619317057,0.27543096881578855,0.1923089131954686,0.4701219269974125,
 0.9448416083525266,0.9026132376943302,0.8994626747346884,0.01093940202652477,
 0.9367030417372134,0.44104332792132617,0.3630478940335772,0.4860630545662633,
 0.9077110834603496,0.3034339190408546,0.7864142972925419,0.9742690515377718,
 0.07640066158499659,0.7831920841031231,0.38975300558261095,
 0.5506775431813746,0.9540418870714009,0.1873452925756941,0.6884459568334157,
 0.8340785325678826,0.46127384855321985,0.8079662358954451,0.6403250137113871,
 0.4352697980173176,0.333887905249732,0.32606328099219584,0.5409821094408748,
 0.5186979643742359,0.5791801389409948,0.8622247133755749,0.17070288105264952,
 0.5695374606024277,0.5957979306249481,0.30165101783971116,0.2841364312044815,
 0.9092426501445735,0.9227321625899234,0.23738826996794915,
 0.11922786144908393,0.5868401539971216,0.44254748234572816,
 0.29628419027484953,0.8793741849301268,0.30956966109530004,
 0.5151378740957266,0.14566520013917972,0.593980841887065,0.5197030725908713,
 0.6257924698698194,0.02273686120589713,0.4121195755827405,0.9347353545002421,
 0.5712900746096463,0.4839902377761094,0.3241166329730685,0.47921093187019026,
 0.7608491191648408,0.6980166648057363,0.7363740985818383,
 0.014458882574759468,0.03014094753868557,0.9644080031999754,
 0.04994995842340977,0.2367927433664636,0.03089362536384288,
 0.9289852003085587,0.6540925650103107,0.5006269018964888,0.9072200014602178,
 0.7951193722344786,0.9552689366494892,0.2634909069584116,0.3325635868296428,
 0.015645501144919693,0.31078809221204373,0.41804185732465593,
 0.41795094689762025,0.40285055768859634,0.23645027833687637,
 0.2660139358995892,0.9053947554678021,0.6334986097986605,0.46854217590588443,
 0.5071759820995324,0.6529308423388802,0.2950959752434513,0.8973209236648955,
 0.7669453260715225,0.47696777630605636,0.8755287492300089,0.590634363852502,
 0.32553451065792316,0.4155474421124664,0.7775332744264888,0.8986230394918122,
 0.19239516629389464,0.605921283869239,0.5225589553447825,0.25529366697042244,
 0.26846502411060036,0.14413479611526891,0.26573035726013516,
 0.5076554311522067,0.3253544492381073,0.8598748396293343,0.3392440792626308,
 0.16472106895760563,0.6251593508354454,0.9997071702457637,0.0573311240299037,
 0.4425980634756421,0.2656827097654032,0.7981082597706615,0.08692629101890603,
 0.63027359687011,0.22287040536200164,0.18622915582351895,0.6922291801221211,
 0.0865297168921142,0.10836803228000902,0.8222040355513112,0.0690759018526299,
 0.8279071441173793,0.8428426764087678,0.16494744204880785,0.2568688640634499,
 0.9368499863633097,0.46345417789423105,0.3987606823287597,0.8936313746548143,
 0.9220963406068133,0.5921293215245766,0.49002146309305106,0.9365590796645675,
 0.095111002975514,0.04727034231848015,0.06596027168125174,0.4175654218265127,
 0.8571518743063611,0.18883158134442946,0.2711293300523886,0.5672546499731862,
 0.03393890816948408,0.3985700403374969,0.02497916978938297,
 0.48036464773096366,0.3951198791034962,0.6989699943044797,0.6447817344361134,
 0.7372818940012879,0.9816124324753891,0.7166214765751456,0.9304547866113895,
 0.14992192184143005,0.8836594909602465,0.08287916173048027,
 0.5887254224370817,0.7512604867946607,0.5522697014397601,0.39086298502129135,
 0.7914614294066201,0.5027197101570668,0.6546442334393624,0.6570195485285015,
 0.8267795243024141,0.40709240533185875,0.008964401926225274,
 0.6227136155760948,0.9181475994187228,0.05132072608483362,
 0.21154111651110286,0.3767250833239746,0.15853297155881674,
 0.35531861164861855,0.28709003665636945,0.05182372207832542,
 0.6428770493000349,0.26243426786937263,0.31847664358537153,
 0.37594844673929295,0.10671103204373777,0.6535011963848993,
 0.8140801486622055,0.7429127929206027,0.8899494540777242,0.4447556591484585,
 0.25214476609753145,0.9690539991602685,0.044143591978096475,
 0.5280723984164922,0.565212866653979,0.7928139118766626,0.19123569314780064,
 0.981727330939997,0.8324076118723608,0.2901024339245081,0.42337254821873227,
 0.657307118596437,0.11527866368605377,0.095750899541308,0.06811767359763277,
 0.467500437913033,0.908077611605178,0.4992748537708187,0.9111085585696234,
 0.3630879695032555,0.5965025843209905,0.5324044666726927,0.3804886731603948,
 0.35896235889543127,0.7476237176267478,0.8565441914479262,0.8849491416860614,
 0.6518765498247584,0.8500522772556283,0.6140967419559072,0.26415749427485,
 0.610650365103663,0.2394285497166948,0.46628217783422043,0.3816772972886462,
 0.029464513326936892,0.8063623643449123,0.6668177206738036,
 0.7833627850183748,0.3618483455372393,0.18985278319117416,0.5584413353353282,
 0.36926978021641543,0.34456422978821977,0.8054928751936254,
 0.8935836258629016,0.8425669628714854,0.973054854320405,0.2586509445866255,
 0.30024521622889366,0.9726001680089711,0.8222687945329532,0.5341418535444683,
 0.1873021649262696,0.9011676060239934,0.1858025532205041,0.848684332750328,
 0.30128965733317115,0.332571680810817,0.12635300367918645,
 0.28730313706121646,0.9837499100005888,0.8895372276920079,
 0.07618700961678804,0.5210336572315322,0.7612254435748927,0.8676220342457842,
 0.2808189933970291,0.3019577648501045,0.9046078058548159,0.9351860882437594,
 0.8228435361506086,0.2938266583227902,0.2561955414536461,0.9944588170631369,
 0.531465786144605,0.9900497594598108,0.9397851608517364,0.9212649807961562,
 0.4197813663566463,0.6777889334031478,0.3175767476487663,0.1976790082262,
 0.0973645966528931,0.3562535423070188,0.30376497848666695,
 0.04834994240596702,0.9094649800049456,0.4638816832187517,
 0.25069286040850414,0.4264326243878154,0.16678695324033088,
 0.5247894923483505,0.35070501683622224,0.5892599086721151,0.989480930488605,
 0.18868032394800194,0.06917431737044355,0.017265550892137882,
 0.7380946936117623,0.37204265292501004,0.4174119560124502,0.5437303754160812,
 0.09038035255234866,0.27120647544478516,0.17861387872518597,
 0.6097940418156976,0.6705759247031656,0.9519875243218522,0.5595981918308828,
 0.7583340378461656,0.5636167270779895,0.5553206418190203,
 0.048705392804938685,0.1439971238326213,0.08804763114113745,
 0.15907485586771342,0.5371326516631326,0.6445635178657627,
 0.40717836117323725,0.2730162553370754,0.02803533391696833,
 0.9725434274147987,0.7833857275409526,0.05142649939078858,
 0.03677732033741754,0.33470340361184436,0.220640130210912,0.798975882127907,
 0.1336532608083162,0.9781069543436834,0.6464680767232013,0.3780108165982137,
 0.8071248621712149,0.2169602802541123,0.6074876432009525,0.7795734067045568,
 0.6216698419788429,0.03214159388418081,0.3027678368660368,0.210219347409365,
 0.028531163497067857,0.8282916207967884,0.3334190663900086,
 0.9359582575027401,0.13301918577238936,0.7512180318420241,0.3514313222055737,
 0.4755046098999117,0.46341471478471163,0.21690793456434854,
 0.8026905568396671,0.27810790254085205,0.5882729761900984,
 0.014172856389506605,0.7432876436890605,0.8184562342455738,
 0.8325852101165805,0.016746690530223818,0.5082255961747054,
 0.24403782368542581,0.9909527343398004,0.4133603614490765,
 0.19454612251273562,0.9310827424528593,0.5156689837189031,0.9606294415074841,
 0.40658387525716855,0.3116731319388293,0.30667389696308245,
 0.9607067742572664,0.9604647641383217,0.8691538655638135,0.10886769966686194,
 0.4818121549747,0.2740002138372647,0.7906165829441558,0.9000586276368622,
 0.619624566591599,0.5825118590957276,0.8067934723374152,0.5886114969165195,
 0.882379798461153,0.6545072767127763,0.8149076143933071,0.4710805184200241,
 0.1845660843898307,0.8345738958418021,0.3560921986490525,
 0.0018246305512692462,0.7409615683528719,0.45874108641408784,
 0.8427042884417174,0.7511746700511387,0.22188285355370851,0.9281947573691909,
 0.6203880609421547,0.045301990015875426,0.955875097263649,0.4948106550810868,
 0.8872008000397839,0.652998548390519,0.9909070134734237,0.010029849348226039,
 0.4478907481397262,0.48747167543470893,0.7248999797994228,0.9180215832116179,
 0.6942303178803205,0.4704811814555122,0.9469350792157254,0.47005890728305233,
 0.5473339315243333,0.0530362482078135,0.917958488427226,0.6818733169832588,
 0.6040146329287082,0.15148387303464983,0.43172371149280564,0.959117658051297,
 0.03562088790580975,0.28733765783333187,0.9040053520331038,
 0.15377628169128754,0.0906724722649701,0.5383101106728213,0.8493699598684028,
 0.8080522507401929,0.4738468084141281,0.01864955484245423,0.6080481046188217,
 0.7404344646333609,0.24594954406830316,0.35319793464358007,
 0.42797927305598393,0.20005147858301242,0.9479302606077273,
 0.7744866989218583,0.4455434225371636,0.8125246259230553,0.04930585668628451,
 0.3168171988865265,0.9731484711358013,0.3490153711085141,0.686824355499492,
 0.8081015563390044,0.1943350006630188,0.1666749754455733,0.23975237471822486,
 0.7400175899352227,0.4811576857149281,0.8609592059253264,0.1091024043006128,
 0.4581557466654207,0.049130157967430144,0.6741261432986159,
 0.06649588477216328,0.47346792850682196,0.49191762329930466,
 0.27338074814606206,0.34577276338632956,0.09469301024073196,
 0.5881379535932231,0.38815055783120944,0.6301337135635825,0.1921508407409772,
 0.021429379661347836,0.8220681941278769,0.09029864131378917,
 0.14884674908035378,0.6581651607174911,0.15246519144583903,
 0.7757334658350108,0.27730567167084996,0.047846653063011724,
 0.3848531047590682,0.45983397681773286,0.05640175269304115,
 0.07151268454344417,0.13424757241027918,0.986365731346991,0.8175386666127832,
 0.7672292575916342,0.3710498555438271,0.09080257544702619,0.3169188427850397,
 0.9032123579891874,0.8737784410441877,0.24225508015017783,0.6151523246384774,
 0.9668885729775439,0.7272109319538952,0.4020519917628964,0.737730340103893,
 0.670629785449184,0.22735643170220554,0.9200900374390582,0.49810458490784226,
 0.884017965138606,0.38869132063651435,0.2528747471338957,0.2503318987354859,
 0.40623523931529637,0.9571428823648608,0.8440820080621885,0.39357492087498,
 0.1874643783169646,0.499035688477924,0.26184886396507645,0.47832762547410557,
 0.021388703982327062,0.47582240659461794,0.04148183437588848,
 0.5654035120828497,0.13680375218095653,0.055737120354568015,
 0.8860896593016258,0.8224214916472898,0.48888338578660173,0.3371355202246624,
 0.14074174053094046,0.16575466039197684,0.5472014630548295,
 0.058792728822871565,0.7802215740878804,0.10976606018102586,
 0.8233283163348747,0.9365534920711942,0.28328771172335787,
 0.018030731710461767,0.3216118151121412,0.5694535186581297,
 0.5786152360833688,0.5194908365665194,0.8734965607164206,0.6789019986229581,
 0.011240634496208712,0.6673049831950053,0.6773229881383519,
 0.6679575751810549,0.7375584802237853,0.7135024843205462,0.6044151090319411,
 0.25203304613709676,0.7704369705066837,0.8936200061954004,
 0.15683712106340597,0.4186524325336598,0.9066097594170945,0.4282997215662504,
 0.899610566172677,0.8501375444754569,0.540610743046391,0.9001285755936314,
 0.25352788815589045,0.06729758760720106,0.3990932487772715,
 0.5445338068278184,0.9092488602334239,0.47515017444603,0.0506623194152791,
 0.39484023160625314,0.6506379179126401,0.30508710144914497,
 0.7020282656268183,0.35149395266011796,0.7189955051168073,0.7965734983891284,
 0.5649168239708523,0.04634413728972264,0.9994375682105966,0.4523668826234226,
 0.5809164442525175,0.4842236995882896,0.2537537501909044,0.4288204887121654,
 0.23310250605713534,0.9363201647196443,0.3446819349607563,0.875478131348215,
 0.32808648598288026,0.8893910565371569,0.08062452848605617,
 0.047379246149851784,0.3017363814246812,0.5683944527524777,
 0.38106518835521486,0.45714026507735284,0.05288880877764357,
 0.7943730147392016,0.23773188741189832,0.807635012298928,0.00867703970307121,
 0.16974921400644538,0.512336545420003,0.27311772176646487,0.8447606299057396,
 0.7146888167552374,0.14759861749685987,0.5201345271807083,0.7928698840382518,
 0.6721447087236444,0.5295504731315688,0.8917109435324759,0.8068862034494753,
 0.041457095753292594,0.510249321944685,0.892333256947586,0.1466480006262626,
 0.48346809891574916,0.6052086780113601,0.6588310842480789,0.8259222533551247,
 0.8985506590777628,0.7110409975242822,0.6481700958385263,0.6489753539270675,
 0.7732612523073322,0.561250111899175,0.5544063505924817,0.1807799563894904,
 0.4309364369249895,0.4003144806453087,0.511572831765667,0.7586933502578194,
 0.40083642057399893,0.0707449825451063,0.7420068465712817,0.3548601916978773,
 0.3099717632115202,0.8046555628649738,0.283889800613332,0.6545888387328611,
 0.1421247920968952,0.9846147206900212,0.5094338997483603,0.07031446353116866,
 0.673157954735147,0.02293752672017002,0.12335518929239175,
 0.34354441001258906,0.11483741018540361,0.5844489855690124,
 0.7965656605678494,0.8726142862354569,0.027186810762686253,
 0.8903362161248225,0.5309750742843391,0.21217982515404354,
 0.36524770781750715,0.394799131170424,0.630349494494261,0.05933613445354469,
 0.48088849464183825,0.24274057721418174,0.8795633829905886,
 0.6156456318879296,0.976387225269624,0.5987653172336961,0.8697866944537465,
 0.4807953780241607,0.7481239142234196,0.346496167532506,0.9875732969918984,
 0.0683196189629659,0.17345759776892822,0.2546914713127796,
 0.11065897707522698,0.9369455274844836,0.5945462049879618,0.2966735738271913,
 0.38545065625751385,0.03085595358331883,0.7846541983163384,
 0.7060942372995929,0.18416267464020297,0.679821720027768,0.15784227054047717,
 0.8126300799306927,0.17677618226780512,0.008133600286281628,
 0.19151748810915004,0.8377547433518437,0.4722289302962166,
 0.24156451144212088,0.8502880498267092,0.13669604787081668,
 0.6121645932864823,0.6762400279222991,0.20524376683868018,0.5679440228279491,
 0.8868461034134749,0.5182270422033921,0.7310268607766516,0.4938809379472321,
 0.962372643840299,0.2123299238804014,0.025551797321947456,0.7215221907573849,
 0.4377347690045823,0.8354545206294308,0.6109960551555396,0.8368621476318623,
 0.4984116834986473,0.45792429146499747,0.20222715456174056,
 0.17714659891138052,0.9545734393554538,0.9595621702230596,0.9487138586515295,
 0.7438808527472229,0.013109142532972129,0.7869126383053574,
 0.5426897932785685,0.6113222362439803,0.7943957244854579,0.1886327385959332,
 0.5042979573686255,0.44606869576146835,0.2684567314476266,0.5846408025107135,
 0.38431480594803713,0.35266082159192746,0.2566891094713333,
 0.4309770015588431,0.12443699125491092,0.11565754261529904,
 0.8673797223739843,0.6986773477909577,0.08858217338726404,
 0.24058086750611263,0.11920318598800106,0.5397663030160108,
 0.9993027485898746,0.599822008137359,0.7355810868471454,0.15760547891087018,
 0.6522426101992755,0.6572373587676307,0.4555574185920346,0.4205625469064932,
 0.05541361240486098,0.22016048414121003,0.7867612960646191,
 0.8689516414411788,0.5225900709687882,0.2551809888079718,0.522270535283669,
 0.3801044269217322,0.93110168508341,0.7667236114627134,0.6766901727010753,
 0.04396126628710506,0.7974818512245282,0.5610916516036228,
 0.25081015020916564,0.02433833377929151,0.10978860366939858,
 0.5796390285632131,0.7363833297741753,0.1335456781670763,0.7843124469173377,
 0.15966877330773832,0.8309585037012057,0.016122653108390783,
 0.9110763277679031,0.9871456114861757,0.8566980638546059,0.3072754972466304,
 0.5915431540719093,0.7191157851080767,0.09610051843906353,0.9116154696195069,
 0.6112489960106777,0.7904953982909945,0.3222109871986375,0.9776259214824022,
 0.6924848119598636,0.35653174986341807,0.10970149699377085,
 0.25917513187802665,0.30791736555927784,0.7237795315578087,
 0.29665938359647503,0.8120785937259014,0.9617394020984822,0.8732713357633611,
 0.24158933505986602,0.5177658198393185,0.8738559008845322,0.7640877251065308,
 0.8414421485763923,0.5472537586998801,0.6415548943750161,0.22097763581788343,
 0.42159242318523116,0.13899241859576017,0.03761262672650023,
 0.34637165295156347,0.8621500597546358,0.5878583748040277,0.4770326661750015,
 0.12461996774007833,0.26375339382852503,0.6888475854222076,
 0.8850890574613829,0.1529396261925361,0.0032563963398561935,
 0.13080245052726414,0.36309942830452613,0.7538249845098751,
 0.7492079996952213,0.15513580216381684,0.7906855031318704,0.9694053169986772,
 0.7371446495956864,0.9129976294601533,0.7052950606004732,0.8590081859135051,
 0.37395486638560493,0.9951393928458543,0.2640992429264547,
 0.07428093612652764,0.7305969828398055,0.22852854693094216,
 0.7534799300052533,0.9815171209628668,0.5508188602589462,0.3649357929162289,
 0.8125179377649772,0.7693975416526398,0.6089307728923506,0.11327673476343492,
 0.2819173281781894,0.8651495970642844,0.18221258835085885,
 0.31770434739838915,0.22060808587565772,0.7579645063473225,
 0.8552434219412669,0.5654101961598007,6.487189494089884e-4,
 0.7062844497480938,0.16282975277725087,0.3404597622632397,0.869627530461769,
 0.23724200889754432,0.23413882358695104,0.3444060416082869,
 0.16598462602509745,0.9887499530582309,0.01196185361417057,0.894590444626812,
 0.519083315095474,0.9451522958140715,0.01694346310297179,0.46684575315161736,
 0.7803083945259377,0.9042692518855922,0.952650698051567,0.3865176409357668,
 0.7972112446096487,0.16451825270575615,0.056869790953369526,
 0.19454257447646117,0.14974257838023686,0.5247632481350175,
 0.054849342986134864,0.033714087286962235,0.26982676586891174,
 0.027227526259846035,0.02862436721393502,0.5271713667514646,
 0.41986730721054033,0.8717778132442616,0.7858256338829137,0.976845468837088,
 0.9066150379148898,0.010704023526753703,0.603321525812082,0.2503750869185395,
 0.21112017487012225,0.7973414033220447,0.78425473283378,0.3304370896921136,
 0.40267161352434544,0.6321444470573147,0.751146687847084,0.7432760071435636,
 0.941694078399628,0.34742834089958685,0.4055661915931379,0.2734891053486289,
 0.14371389412293745,0.7508777191521788,0.0780791342145174,
 0.22260278277735912,0.7519317406569866,0.007730748209471838,
 0.9563823158428215,0.7589857714661741,0.06986292805310967,0.8089793327652594,
 0.5581527152634418,0.38934005756409695,0.2118854984197608,
 0.40138599787032736,0.34203913610278625,0.9096952187485615,0.366539988400419,
 0.6018228320107383,0.7758452887893651,0.9598766140273856,0.6113351533889504,
 0.9826753669049049,0.8030175758866558,0.7902991974998187,0.39200865173866983,
 0.20233885789952066,0.015475690126762642,0.49077998981085647,
 0.9351809435813573,0.06825896563644718,0.3708151696873848,
 0.012211496691118762,0.9078001398812521,0.21956580514103652,
 0.4428726808886403,0.7776826578075356,0.6215516412673044,0.3063745370008756,
 0.04056574118514955,0.14659175494816523,0.19403517897620903,0.82845671687693,
 0.6262148273947697,0.13897982560016586,0.10433668348648772,
 0.7487042822345291,0.020441295556906214,0.740059405394122,
 0.12584214086577916,0.10097127338214296,0.20503952687696925,
 0.5284368530706212,0.8712151021143192,0.9666863963067258,0.38476861170805543,
 0.9525952436776169,0.16602102914525796,0.29554189622740423,
 0.08750824197982365,0.43025086590169637,0.13111711534942805,
 0.3558783673669841,0.9900489344991006,0.9747089410593848,0.21823478795880358,
 0.5978611698161451,0.6495845955991323,0.3798777349374871,0.11755718275535154,
 0.5407022536971341,0.333011709273216,0.9525272892439476,0.1761616878187351,
 0.3339775839903397,0.9981377650300015,0.18727815427628403,0.531409905721417,
 0.21429946479578033,0.13088755164260812,0.9063002647774834,
 0.18125416296410657,0.3354575141727488,0.5038189770756863,0.683764845565578,
 0.8956338256146107,0.23621273678052235,0.9930638476539685,0.5685636634456849,
 0.5687587032603058,0.9059417171943706,0.8155832015827571,0.6307421144857212,
 0.43918867649688786,0.6431308671957232,0.9980305342757307,
 0.011862942683997457,0.007147022142673598,0.3459300394869216,
 0.20904821017323238,0.5277986966425146,0.5495943003556008,0.256433149266742,
 0.19178849161161937,0.3488541965113636,0.5350390571405753,0.6174852193895916,
 0.12859699184612505,0.3402851636235227,0.27441175592449474,
 0.8615115188968657,0.06630893806500293,0.501390622726139,0.8904320841885269,
 0.8225116079267298,0.6826590380777264,0.013037223713686985,
 0.4059599103336189,0.2381023109031607,0.9314618377770844,0.2650022652183859,
 0.032792203665995334,0.2184114473387052,0.8676597896852376,
 0.8196568683931991,0.05641864946667541,0.2634738062982367,0.3490602440274333,
 0.016942377575503875,0.2947891289584543,0.7943383788829431,
 0.9990260638851707,0.15203091038283967,0.41186252304171134,
 0.40731669596606945,0.7374812365359279,0.4254251372388197,0.750110675540089,
 0.9693183192197582,0.8090781459323205,0.6959428582943373,0.6476360051217352,
 0.9279350832170602,0.37998794228179666,0.04765954811693496,
 0.6395787799336157,0.4961701559322248,0.6431942283542241,0.759550773477813,
 0.9665918133129845,0.9651913127830744,0.33506424086555775,0.6995823901253551,
 0.2096804663131232,0.5657358962187827,0.02670729709866193,0.9957690911050822,
 0.7914286517028324,0.08872262908656903,0.630863634345401,0.2628422601678063,
 0.9984240684493947,0.8716245348996651,0.10258548021659708,0.3107267968765184,
 0.2518349657445058,0.9012829026723369,0.6198589391267093,0.48567941292945493,
 0.4777202634710157,0.007754376183381262,0.8192503898458552,
 0.9253692884852235,0.09559259390617769,0.5397165135587156,0.4014069606094167,
 0.15442422195720928,0.5509910133206289,0.1335837719148254,0.5488970944370188,
 0.3986217259038174,0.6892627755707528,0.014014574965796611,0.691566044386849,
 0.6666567493403476,0.658438934047796,0.49542762758461834,
 0.017870583494405845,0.17491446443366232,0.47263877308510605,
 0.8378578450754535,0.3287784581574821,0.6037622606156801,
 4.588953542510321e-4,0.04473316542393235,0.5407503205054572,
 0.7979622888926051,0.35615525836618056,0.49536911468377354,
 0.31803249481914553,0.20907510192399714,0.0861782779413478,
 0.5943376581487704,0.6478917204082466,0.033090053083801196,0.617092442753689,
 0.596340482612471,0.6999177578241992,0.630353168902246,0.5520738866599836,
 0.6357483420429126,0.9826888873172261,0.979381165847691,0.05280330334549099,
 0.5159212875386899,0.206892916008254,0.3867674808038283,0.7735019147543057,
 0.4375468154103068,0.4661023150935728,0.9158681269309985,0.3545038296757339,
 0.7825476826431441,0.7442850465957705,0.9850577386152588,0.5131742839151238,
 0.877859633695757,0.5633822462457235,0.5271533173653001,0.12467641490323478,
 0.4548583706408893,0.6826176800181017,0.8097008765047515,0.5991456804852222,
 0.28840830838737597,0.2802349765360239,0.07972178618387227,
 0.4178374144847329,0.5145668833566818,0.4858770134357938,0.5331704057369142,
 0.8125790273329778,0.8724018925547378,0.5844725902517097,0.4947894226901548,
 0.7981139045380986,0.3711857048352627,0.25195630344913167,
 0.49896102646993734,0.05072605786110218,0.01264123473397638,
 0.42627128027977157,0.6468418111045526,0.41827894961676637,
 0.7957660383514943,0.7795808518485909,0.5387785873479503,0.8101394235953968,
 0.5490332830744008,0.7160042704276167,0.4534278019493606,0.7515914849157175,
 0.5418474445489115,0.573579318099973,0.3451749567104688,0.014488940255729443,
 0.8132035000589978,0.386119947445701,0.7299065732015375,0.41250390712797114,
 0.15598129672388072,0.5453361593018835,0.6921890850579769,0.5950220327813387,
 0.28425935030289495,0.3485343661250778,0.27032851529847424,
 0.6280132140284878,0.8127556240755743,0.11327850860537858,0.7165823262746039,
 0.6367572612105498,0.6447561664389893,0.6188464050095946,0.7116903638784803,
 0.33133977853620766,0.8144130863792847,0.046592285700503555,
 0.08214987669912699,0.26358426540011715,0.3641184482240336,
 0.04280881241088308,0.8865361593388684,0.17950279397129232,
 0.7552619624554486,0.20715661374329963,0.3810310234663078,0.9478019591154553,
 0.5055294140438891,0.9237537679202483,0.7126540316597743,0.19967176390277297,
 0.47586109822822364,0.5130621719364651,0.7554681130695098,0.783866125376307,
 0.04441815765718515,0.897680311188331,0.5257141197205449,0.16672606027482906,
 0.9869467432684944,0.09858192183449288,0.4994883892350883,
 0.16660736477657545,0.08470147995583577,0.31430647843724524,
 0.30260105340688503,0.8134956102368619,0.9391272140178161,0.74577236147223,
 0.3883427817318955,0.04641768012150371,0.8651694753789873,0.819974683268989,
 0.3786009502464649,0.5605100613543914,0.8976565419702138,0.55096640498928,
 0.959755275230605,0.5443992194506329,0.11667436235803141,0.6698058630986565,
 0.6441012936365135,0.9556641238520927,0.5201238593399499,0.8365840056883274,
 0.09632425907399145,0.16956560598264359,0.5698575206297781,
 0.9173605640854008,0.1412765076165251,0.15508078973926553,0.9322186225023666,
 0.7874542864748382,0.5691368145419742,0.5350479643443423,
 0.055611087359399014,0.5500436073941649,0.491789823925314,
 0.35484015716561834,0.6461762571108131,0.9938826894451104,
 0.21318707552198013,0.9431960981874694,0.9902562757772571,0.7194121586771762,
 0.9666976875906582,0.356251927139585,0.03689820524814269,0.14835497096824202,
 0.7728127380478717,0.31751012265720846,0.34182839644478114,
 0.15509668115855435,0.3716052070495627,0.8437308758862987,0.3802159952348637,
 0.7386374576200767,0.45275351897585303,0.7240060606188953,0.5050956063033132,
 0.9834624671536404,0.7247206257378874,0.3709615745013739,0.7465360926404031,
 0.8449894543173809,0.521996992933256,0.080786770825781,0.5661215337134236,
 0.44148816571766436,0.6126989066347706,0.0825603517254817,0.980798388758503,
 0.9958287721779169,0.21859626447750435,0.24921964962464183,
 0.5133858244318881,0.46804601099315646,0.6904950808931157,
 0.0022398861787233937,0.7423296506655038,0.7882629744936062,
 0.34170937714171157,0.21298660529498736,0.8122517103235991,
 0.6062351241702768,0.4855955320413675,0.5960766866914657,0.2160329096439625,
 0.19019752475055407,0.5030990350502068,0.10965395987067572,
 0.3299140958015968,0.3242741378643086,0.8459281994208109,0.4937146717519826,
 0.5786339272056092,0.24116752547328424,0.49034686948241735,0.956747070087372,
 0.46004042728934313,0.2465539980621323,0.2436380476894272,
 0.038161545442617184,0.2633471855513342,0.028642119981880376,
 0.6024700187150586,0.2290247825096593,0.18561150442608731,0.7601815148116381,
 0.013011390059234351,0.4315262049724632,0.47294191300005595,
 0.6669932362793068,0.8152827074509251,0.5020694572969372,0.02175814704147816,
 0.3842436397060829,0.681946442252554,0.3927889641310539,0.9005355193272444,
 0.7314974330843599,0.15872311015438045,0.21798382376652659,
 0.6425818882182044,0.40615722938267973,0.4564477160100555,
 0.19049159699782914,0.24088875420277,0.06528034895847867,0.5028770089925636,
 0.5021877054767357,0.42197924199555936,0.3577396039452875,0.4566160819421409,
 0.7496073226778306,0.3311023080982116,0.417574310728279,0.2695961681115987,
 0.9563305564781545,0.6747713482169809,0.07948508965011802,
 0.31328526658265476,0.15215847329635035,0.11408860339253946,
 0.06628008025619059,0.4799972919111819,0.6892757017639694,0.3854786458756846,
 0.4859580432643924,0.8010186384462676,0.48925478410236223,
 0.06286555429627472,0.8817293137342905,0.26013957756103434,
 0.2928721822071845,0.053492310470209525,0.001267292540372722,
 0.7367754812853347,0.11542660260568383,0.7811544205196806,0.7251399402251887,
 0.578591788459409,0.3444179116961834,0.8051472326652558,0.33269282734550165,
 0.6011309636531257,0.9062160132885348,0.8686072506337259,0.8952689232389622,
 0.17624918747518148,0.674044726536899,0.6201251580329246,0.7825713187050642,
 0.24407801862479261,0.8317078336055859,0.3380868153285237,0.0631481456882741,
 0.5352445064279487,0.5413654285985423,0.44963951821760806,0.2836939659154648,
 0.547970105347318,0.9748186192097836,0.5803706202002652,0.9375991247748765,
 0.46328120437286846,0.46733923473116257,0.5994482364217348,
 0.41491664734015277,0.3925022086687874,0.11229219969833171,
 0.5769404993493767,0.18235173380484304,0.9714702376393672,
 0.004486224310941478,0.6597024199581955,0.5025847897022468,
 0.5685472428654698,0.6868311569833158,0.5165197828973361,0.3537795596847526,
 0.16233041377360535,0.984011883988264,0.2949845715034063,0.7720388211885929,
 0.289385072651147,0.4989576651134141,0.7194732733319652,0.3170875555357865,
 0.7921060398876524,0.7360474089861976,0.9209723778471915,0.9306999268710549,
 0.032233286162783314,0.24356357618924052,0.17036134345294673,
 0.47217403694092386,0.5015427362122732,0.8983690396952377,0.6691693757368762,
 0.40292449002993636,0.6098712830996189,0.021768781687898198,
 0.7695614505914325,0.24097057375804043,0.05835807046825392,
 0.49929626572362595,0.003251347048217612,0.5514076136468353,
 0.10683715660256188,0.5968600740670992,0.8489851084907186,0.6197177257929793,
 0.01284868052759991,0.689456914808346,0.8991199407223488,0.3808403381148393,
 0.5873221845433342,0.9096415939679497,0.1269819602373159,0.8263169749748983,
 0.5749947639201829,0.15640934750476887,0.8681794342460868,0.9597247387567978,
 0.9521796996145503,0.28161974460768346,0.0015735247235695304,
 0.556393082207392,0.2799054604512057,0.08212015486548996,0.4905037660659394,
 0.18206639027494087,0.49063932301382984,0.3690780842259046,
 0.8313455275646716,0.922849317338267,0.005029178860378347,0.5156273374722546,
 0.3737427517323816,0.08143199095499964,0.47083913838650227,
 0.6315940649121587,0.1885402253192876,0.7370454457660652,0.0342795198321213,
 0.3637137367967157,0.94312116183228,0.7168106449872017,0.871761686524287,
 0.10476337804640212,0.8043069260652778,0.548368204522654,0.5373016627490373,
 0.7642861920465891,0.30884752396819815,0.18535819039760004,
 0.8997597626754638,0.574795667876398,0.47267560642427553,0.7706099389128531,
 0.9984979569990997,0.02138405105498109,0.7830850166782137,
 0.04571150271145352,0.24775787849072883,0.13212564320553333,
 0.2691487800092711,0.9206484012697078,0.401100002668784,0.43782448924980644,
 0.9464288015133459,0.9742021089958712,0.3007789266569314,0.6367925884199962,
 0.46283127775436594,0.9744963489709183,0.23652987012414706,
 0.1459708403474037,0.7313114979925985,0.27571992559767367,0.7114422235344349,
 0.3600450615441584,0.01308486110946161,0.908879986729556,0.8542044809771744,
 0.946961530755258,0.8179307546162158,0.15686408087264148,0.5427657413846131,
 0.8284593529382863,0.29852626526712855,0.7230145956678738,0.7687923216489964,
 0.3783141416675846,0.20198747646515547,0.5171997060787684,0.1849427035650374,
 0.9459225727947111,0.36463293743051073,0.18797088801737705,
 0.6310031523793803,0.5679605618416007,0.752077653633974,0.5166665545340449,
 0.04159684903870087,0.18681263237731138,0.36885938957248254,
 0.8086218080152908,0.5390580992173212,0.7147275633282238,0.141966529437772,
 0.039475416357498516,0.6738966890959417,0.603375731115833,0.5839143161662259,
 0.22052553269392694,0.5606221663101881,0.5472330515037346,0.869750886746078,
 0.40191863016682094,0.9337258838252006,0.9635495313872665,0.7852342795382798,
 0.4401305472056667,0.38569020275924204,0.6656781261423445,0.4345504680219696,
 0.43307304405491365,0.6959899893159609,0.002705224092906189,
 0.7729350198346526,0.9337576352480481,0.37510613318000263,
 0.14509636197047493,0.7112031485812967,0.03790255139320953,0.590067446899589,
 0.5102045900192032,0.4319385805536249,0.3958306713375226,0.4521553868547693,
 0.7981341415616583,0.873271868874511,0.3206961430596712,0.9043902905268681,
 0.9595843924673267,0.3027874828202046,0.29189568368234486,0.8428808481021527,
 0.04128827213619135,0.30661217110778183,0.5291430678568319,
 0.01657233447543105,0.3013270837807853,0.29060680785323745,
 0.04108771208083961,0.7570352505997656,0.6752406945111407,0.7665402622130979,
 0.762019232925895,0.15255081188934427,0.16821857006672714,0.2457902658679071,
 0.7581797771618906,0.1262835461988827,0.4032494145264873,0.942828593576229,
 0.8802355356126565,0.5123871949710204,0.8302325584513772,0.5124335558706664,
 0.8468575762475835,0.13367023490450614,0.5128821561648502,0.4177785149268137,
 0.09625370863291416,0.7636552093722484,0.5115823572527836,
 0.06373981694654374,0.7336415331275559,0.7414360681557848,
 0.24524474453422807,0.7687621827329174,0.3047978101827312,
 0.16975917556331888,0.7252101865946603,0.23947657048535453,
 0.27930436954154314,0.03811189013342464,0.7322618519766881,0.626739333003907,
 0.6890574519655683,0.6690537325650702,0.8597464258757274,0.607829533971721,
 0.89734857856602,0.33851746342304645,0.5253126796884449,0.6487624613526779,
 0.1595419793591114,0.5196564069771639,0.6690136092398109,0.06479249549518673,
 0.13785581242651057,0.3374712999979843,0.31127338393911763,
 0.46921529505343373,0.26340534542815,0.6376068018756396,0.6814069623460979,
 0.45716487123938954,0.6310249075928194,0.4116263011124208,
 0.15609886523032812,0.9519689263836097,0.7035131794030722,0.0820955085502848,
 0.2520829369387674,0.8972977647621392,0.5930160758343265,0.34232953543564537,
 0.9387640013191393,0.13140956427117723,0.4308956029222355,
 0.47112090082161007,0.9984453926298507,0.44513031728012764,0.712812734086199,
 0.5882160914434128,0.20325429535223993,0.627276433484766,0.6366795672407887,
 0.49783089914218037,0.3894908345961219,0.43044913554781394,
 0.07215525643867693,0.5864958156445188,0.21237747685538622,
 0.4696644285234791,0.06099442042262415,0.975494717179046,0.6281396847133406,
 0.6288046401725617,0.8111090957658155,0.4169267241184118,0.8509479413739333,
 0.28497601264457206,0.033062436477295676,0.8090686379080494,
 0.452697271058206,0.5421540155319797,0.9618395392937344,0.0811973751166638,
 0.387995088147148,0.6329612861036009,0.2572338256183423,0.78142045390554,
 0.07301222612455993,0.32968410131546644,0.35120849042088154,
 0.22825405721507108,0.022736044635023323,0.0759792599545388,
 0.4134752729192501,0.21583645397617057,0.005153756866691395,
 0.20001512859985204,0.5010950639325542,0.16682898642971256,
 0.22743168153866467,0.23442373408058714,0.25414986867049916,
 0.2988298579730996,0.21974088718135443,0.4079387942365814,0.9509865267511857,
 0.002018144888988993,0.8415753637185736,0.3544686148266364,
 0.08134333730289212,0.9409305923429598,0.03507392661169473,
 0.5200888288337194,0.6481384611248033,0.9507693905145758,0.576349064471882,
 0.5303429966149877,0.4898196558210387,0.2700254157074483,0.6847660584403403,
 0.292111050892793,0.06800346651719702,0.027824072294750812,
 0.44753037952022157,0.585293908846516,0.5450306620178992,0.6970009662307847,
 0.0474196702207893,0.32573616011371176,0.49011720483978205,
 0.7867409902710467,0.6118990219080075,0.675070338544625,0.5475586451622845,
 0.9230780016031725,0.6927650658049246,0.23457212349714074,0.6369497064752859,
 0.9681117457108102,0.7429004345380061,0.8929718588543275,0.7574544331147093,
 0.46442568489651603,0.8181515934575335,0.30978369228825064,
 0.1148411044330726,0.41648012031248915,0.7644463987535167,0.5272909845673437,
 0.04932442851436192,0.9724231322315845,0.9642259289391114,
 8.622595688840828e-4,0.07783686398558076,0.9525110024507617,
 0.06918072703311384,0.2827046457498491,0.620674702127253,0.301903337944913,
 0.7690554629703987,0.12380192055263484,0.23487267142919144,
 0.8819434805270381,0.45337189380085086,0.7534752616650024,0.8716539300318251,
 0.40443709871302047,0.459506501114894,0.3834659118071504,
 0.011148600083426885,0.7435013417676133,0.16242610653549328,
 0.7894284718888225,0.8635403126395698,0.6753620655946577,0.9838343132406264,
 0.5231228910778785,0.7218849527197874,0.40991795415225507,0.0796185525107268,
 0.3807668433326292,0.03300762618762698,0.8907408050323935,0.72063112888772,
 0.41159666915267,0.11883670235985078,0.02218236345343927,0.13548879423895344,
 0.40555183529924976,0.6143264366937524,0.7646032518273174,
 0.23001886258434057,0.9321727179600212,0.35377855226543187,0.259661561504248,
 0.7387160526211576,0.4548971060152055,0.24492471438001606,
 0.40059076435529695,0.7800530765429757,0.9691969437062374,
 0.13071249082935477,0.5781262217865191,0.7622554799610953,0.5938744312862796,
 0.8905762511102326,0.6590890063583822,0.4548749471107647,0.3701275390763745,
 0.22442140293295731,0.19123890643829133,0.1821992586975052,
 0.2893373080135425,0.6251063319987822,0.7493017719357079,0.43144123609190954,
 0.27708853476257833,0.9325109111333358,0.35672004409975966,
 0.23426467898484338,0.11066596159127018,0.06831856888865906,
 0.6272282148985798,0.6270809191746418,0.5518428241906157,0.38019140492240544,
 0.6583637878882294,0.6405926647589038,0.6460394496857583,0.31436090931862504,
 0.13656283395382984,0.27426011140371365,0.1724769921555055,
 0.30817290619629734,0.6361979573715099,0.005195974611501675,
 0.9333779129464319,0.050051466752411344,0.6187432611392871,
 0.3878336164814642,0.9714898121217057,0.540754924563891,0.33143770014846086,
 0.907781819518486,0.28967950715849977,0.3801845277120891,0.3211288512921291,
 0.33645662688189615,0.6066109082860688,0.28371577703904016,
 0.9576334280281176,0.6982009315401349,0.6049903487216808,0.16152945869835744,
 0.6960462683695292,0.415566174811699,0.4715133374735456,0.8163434550734495,
 0.83636752671177,0.6215032730118281,0.511298698710088,0.9501869360158748,
 0.8801729040526942,0.4244614460293865,0.9363751768932393,0.7519049263540447,
 0.049313738481528535,0.1525580964637373,0.8555933683158831,
 0.036398523004749994,0.2638703571648179,0.7728482716384139,
 0.48768850797152463,0.6607582968999295,0.6862524010241037,0.6500336354341975,
 0.8487996757296579,0.8114509661653724,0.0852977659368026,0.741762074498103,
 0.6122632369761627,0.6767408799903807,0.3641771140135077,0.9936644636883815,
 0.8303022398700324,0.3642450367678096,0.2465117237688853,
 0.025443259146324948,0.20528544902272294,0.5829411535907595,
 0.2050328490335821,0.18941394393070332,0.020924614290815713,
 0.6945036065355522,0.5785612995649332,0.943043636404939,0.678129542547927,
 0.667983330907487,0.8849829750398643,0.3756478284797097,0.4504308733397966,
 0.6589547593247591,0.004407212204683986,0.30850585705963296,
 0.40062878498767995,0.379752111004518,0.3040794312356361,0.4486162495910735,
 0.5154518107770861,0.4501727969734288,0.5360603377557431,0.2550707801438976,
 0.07250669697263568,0.18073015321519337,0.21877589863632407,
 0.42481666286645825,0.4958882537892375,0.46898542395706977,
 0.9221980167973018,0.03984289183377032,0.641780956372806,0.8938454191026795,
 0.4078939978518642,0.5902067449302995,0.8187703556945387,0.7251104445374548,
 0.8273802634919379,0.7589335026389384,0.2005731348994536,0.18099833737368165,
 0.13035357475895648,0.9682917028852788,0.2306367942107972,0.2891817213163955,
 0.3361251440917822,0.9208515341430691,0.5740384003491381,0.4164838602687243,
 0.9857185641769229,0.7209954874514077,0.3864312435290984,0.06826105107434322,
 0.07259727831230611,0.5711764338554991,0.3446705140388955,0.3499215576012976,
 0.02349261634924893,0.2688227080469344,0.01492462815603579,
 0.8359548634901298,0.8457011721645683,0.9865604170215572,0.9100631952507647,
 0.17884292778422295,0.6179190766447699,0.5812632613949115,0.5242798996113088,
 0.7865004291413206,0.6539871307810952,0.07978943305696196,
 0.29063742122085445,0.5441128945461001,0.06358587049178144,
 0.5982729106516971,0.009354883968434646,0.10219364548697141,
 0.1986171658168583,0.23410897325410684,0.28904297487180675,
 0.49727501347862013,0.7616944615908424,0.5081640489121856,
 0.28518170012692556,0.7339701353044433,0.45763945227699887,
 0.43134247232556144,0.26346468687387525,0.21354028969703853,
 0.7391904822857106,0.981885489063884,0.5715682006733279,0.5911245551341491,
 0.5524325821550794,0.29200737680693223,0.8212948883441039,0.6731518477600101,
 0.8491782254510913,0.7109447277616456,0.2643071447192159,0.528386303922286,
 0.42654509101715377,0.908453092465774,0.11716735657322219,
 0.22288351696975695,0.6320141392500939,0.0034831606886759348,
 0.7666566392634603,0.6178219109351001,0.19526344330212742,0.653821174343305,
 0.6674885869013595,0.18549316336179755,0.8692959611396319,
 0.019479966724038844,0.7189686096576492,0.5623604762147866,
 0.24432558737030619,0.6679926841110215,0.9938136587167703,0.4539084385438088,
 0.8184104815052895,0.8963568187185144,0.22996553466034886,0.7344461170451395,
 0.5471166246015091,0.955568017658459,0.5302544415152779,0.5266802639610588,
 0.6196083266330685,0.7140579324278595,0.3473224412861158,0.7820086402393209,
 0.9959517979211745,0.4152955184367557,0.9543776035052858,0.5955958377159241,
 0.0442412463243107,0.7399229215369297,0.5097255511962936,0.7971278020095002,
 0.1264617684721019,0.8186117238775983,0.9027754905246302,0.2549953220127923,
 0.7879538344210762,0.8348157030887203,0.8471911041028336,0.8111594304440777,
 0.9261692050034812,0.8275997878505932,0.12584320230487767,0.1686518269406514,
 0.0533209587609067,0.15386259807302338,0.3847841362832003,0.6925552120572105,
 0.16707961085937295,0.6151750964662027,0.6312522007020674,
 0.10403390027762038,0.8427176405134376,0.9293392297954847,0.4751051974356225,
 0.3765620801627071,0.3047858618209931,0.6360191081459191,0.13237565945562357,
 0.779578861941297,0.9230495775955584,0.9467544270551708,0.8959346349773523,
 0.361857957156171,0.23679166027342258,0.4673515494788708,0.23025481849272478,
 0.7887007196351622,0.6121891848275657,0.026106200168847793,
 0.5346289002941012,0.28933092895255097,0.3239652758371152,
 0.25296267850737997,0.8239440489142963,0.6835463124271378,
 0.17561964923763507,0.3374087857700827,0.11389240481415697,
 0.42694227541445784,0.5193562532264043,0.3914220718633785,
 0.11320800999835501,0.5530851178039665,0.0609973785878104,0.8057981316734508,
 0.8013000079105709,0.04449009181000063,0.6808954888212687,0.2557664150749075,
 0.49863125160766153,0.7804362903453748,0.6553350431449431,0.7011232349582439,
 0.2553343599824719,0.868712035300141,0.44534045917417364,0.10763627006901344,
 0.20339996973495145,0.1945024137752448,0.39769100512873956,
 0.8990697905179375,0.1282848377754252,0.0500804138362122,0.4332332067664374,
 0.49092691492271556,0.527457616312747,0.38763582694394216,0.8479009622159378,
 0.06759393541944936,0.5215740878923235,0.5151137778540732,
 0.12003523274707328,0.39285106522023727,0.2837736997088295,
 0.4873511652626418,0.029115721237473524,0.11486961294818199,
 0.6443171068445064,0.4302892337049621,0.9565263330001232,0.6090265641115731,
 0.8715804841884669,0.10231533695867712,0.9644356250253344,0.4762252338763546,
 0.10403498940588607,0.625353805863749,0.2805803958313473,0.40727352020052543,
 0.8847772847930213,0.014984334628819762,0.21527487365319153,
 0.786754936860717,0.5382747926407888,0.4595944242587713,0.6235115787424527,
 0.36063137599038153,0.4329892613835785,0.7007806707987836,0.4901146271449165,
 0.5545790996941075,0.6900201610217652,0.6593031889511674,0.4397412934374214,
 0.8292464681118803,0.03568757666699285,0.34036314562454806,
 0.48638872039164927,0.8729872105339691,0.9864177902370093,0.9762205668479871,
 0.657867457345818,0.9282135250539827,0.033974661321136246,
 0.10941233170386622,0.4047620873321156,0.8889615974344401,0.9716438394201424,
 0.622650906027071,0.47581205046873654,0.37959520567218563,0.7428232445918396,
 0.7475439393598771,0.7107646036639784,0.42312258009301473,0.7781382089515492,
 0.6762046478457305,0.4898334782898487,0.5217517135786272,0.8582918842490903,
 0.8015758118202505,0.6476456400087006,0.010661005144633195,
 0.29913139499248764,0.2508448482100619,0.7179085290393414,0.6282855171549079,
 0.747453260743525,0.9890334646466371,0.3770983866706984,0.3514409097257244,
 0.38246885498336436,0.8807726607168115,0.13267357880112818,
 0.9188406642227126,0.31723659210519506,0.5151788504661359,0.4144141860692302,
 0.11481210602034597,0.5135702839753953,0.9181471362120718,
 0.21302412221619016,0.9815843520959541,0.16154316424651427,
 0.39749895431295723,0.9501751172492969,0.3137903538090938,0.8149043614384073,
 0.721286220610318,0.9409375222401966,0.6587845993082924,0.6769968804286101,
 0.06329858673469813,0.9598809136231168,0.13681344743070134,
 0.7693116255994841,0.2564295655480391,0.7756296077290403,0.975975578747531,
 0.33313842243925895,0.9304995515135053,0.6083013184797678,0.4322708533568236,
 0.9162500802749909,0.5682345399111256,0.08966889705997605,0.6101820517613279,
 0.40660283190607016,0.12099972122319536,0.7519606419916998,
 0.07105482149963083,0.6366506123991386,0.2840726607500792,
 0.14230714773380126,0.8272660379844874,0.013564437086850845,
 0.427862283441069,0.3229894080138258,0.6392164755400271,0.0329734830573446,
 0.2545423978922323,0.3364416154216605,0.6823313775553993,
 0.005390519500956037,0.44336124488548356,0.45300964481816697,
 0.5849152173932364,0.8049690498059004,0.7316205281749492,0.420645440516793,
 0.8097561407316891,0.56490868329401,0.6106774594011629,0.15617092495579998,
 0.9191773969491632,0.11461974254663243,0.6406274048118169,0.8486606576519841,
 0.5285223660554828,0.4738234753843753,0.12019528547427072,0.8177510229819198,
 0.27704669892106537,0.5696698565951208,0.36962112178159134,
 0.3432889008149703,0.7165119853316789,0.1425253974488654,0.3494509307427105,
 0.10853208913628087,0.8365768369023947,0.2779533025031147,
 0.22463524959353431,0.6567044052364736,0.19643736501716635,
 0.6460601032951069,0.6221539181482332,0.8382566313774051,0.18561850656518364,
 0.5458389876408384,0.18603198608100957,0.8929503800114138,0.9372078627776291,
 0.9930459933542813,0.60565972308354,0.22244685499003236,0.7448475429640845,
 0.01974824198917491,0.04382337744164466,0.2633031272742564,
 0.3834648600898428,0.17444502384154947,0.2127870623423348,0.5070796539835448,
 0.8300761680029446,0.2703920424269327,0.42690827436160683,0.7845081789499976,
 0.45630854725999204,0.5369623618076127,0.6709365549269883,
 0.40274542673715064,0.01257870577814646,0.007817409434437761,
 0.508384889909609,0.9123998626217171,0.24357746381178247,0.5621176214609778,
 0.2738198200746441,0.8394971581993382,0.3143030550932322,0.822277654161429,
 0.8164150190280905,0.943955198120518,0.1886387959008592,0.9527396878862555,
 0.43656890553252914,0.06424017970392915,0.9272994085596312,
 0.6070497737974171,0.8107220761445778,0.8579948062314443,0.60369372745482,
 0.4845738688134219,0.0655315636974424,0.4942123423757284,
 0.052330516129864724,0.2799396148460108,0.6411750383549346,0.94999063832021,
 0.44572581268946265,0.7941478067940294,0.5726496724138272,0.5544667277292213,
 0.21910241240682016,0.335797610290514,0.5664640720319062,0.3972124196821545,
 0.6713457528563106,0.482007778647094,0.1616584761213511,0.2706019053908414,
 0.06010595928483631,0.06486600794825692,0.005586545100515594,
 0.13315795547064713,0.8019224896495164,0.38238280088821175,
 0.5815424107082809,0.19972237506869084,0.19382909941372073,
 0.5802742934862702,0.35384774074521896,0.3440936742070253,
 0.20330313629101238,0.536949485237965,0.9525573376229358,0.6684453799015277,
 0.9848435561468715,0.8418968640001826,0.7446831755617621,0.4929324101760514,
 0.3177582617610396,0.7755418286370329,0.9040356226222023,0.6750293939801846,
 0.20287623591686188,0.8165164643609991,0.24764304589742503,
 0.09256827042222593,0.97519054112996,0.8436246593029681,0.5404684646276916,
 0.08618445885822112,0.710723519255292,0.9345435926442779,0.8649099222006044,
 0.5701679389023866,0.6787538137454829,0.21556157527240527,
 0.48568304031144005,0.21934455800014485,0.48146148603760486,
 0.3169728268146256,0.05583807047696354,0.6493850016940512,0.8455821259939017,
 0.07307205420567686,0.9821615639160618,0.971835618281453,0.09542713039346384,
 0.15239265576634697,0.12010276348046878,0.16089541515915817,
 0.46153384450104296,0.717201382272792,0.9843709243367375,0.8653197367841345,
 0.9000011837951999,0.16610850698210544,0.2664650419189578,
 0.13533995535838783,0.7481166335314512,0.8159902353430708,
 0.028352006967400478,0.5570482588424794,0.39319638090186226,
 0.6687312733400252,0.9775254689877236,0.08876830701287475,0.7245391975115016,
 0.6852718738570729,0.7667514886793958,0.5469526701909632,0.20888097369830338,
 0.45241617678167945,0.4459896171413352,0.8858105184596947,0.6315241789898677,
 0.3688342471267052,0.22425546786733475,0.11353323897425194,
 0.15930259711484285,0.7675326110032182,0.13853245244978285,
 0.6696732408927167,0.5721612542627914,0.7444553002009324,0.28423925638310443,
 0.31175408913241587,0.7771567452903548,0.13922186388883373,
 0.8780005969807778,0.7528454207429927,0.9657441846998623,0.3147095709625405,
 0.7926700493615055,0.7825706682756606,0.2681389726784431,0.3149352904233802,
 0.6564289190204722,0.2029156594363528,0.5962419196252436,0.7637040209448606,
 0.4956488321712882,0.5258786423453954,0.21697698058240533,
 0.21691709747829147,0.24230586342247984,0.29779138701993024,
 0.6403933584576776,0.8473296508265762,0.1980301833484419,0.608219380810073,
 0.5450962839938849,0.8105676203207043,0.18609517607777404,0.957355956150892,
 0.8684147404840109,0.31390574948342453,0.8788553261066242,0.5092205419355164,
 0.9075198420011836,0.784806193301608,0.41475761186631377,0.8881999052548395,
 0.06380446535666762,0.5718467234159974,0.46296028255323585,
 0.9001923610449962,0.39537107009123573,0.052140297078859366,
 0.26600693030982503,0.6426040008427332,0.287122049558994,0.25385955226877055,
 0.273028791013211,0.030197394291595514,0.08774158165631918,
 0.008552191120789754,0.3743072479321584,0.16889295571596108,
 0.12132322818197205,0.07881873400515782,0.806264363248266,0.5521504026922136,
 0.9541248924219595,0.7380995737208458,0.5166528816540701,0.6005266610923494,
 0.9347137488699759,0.8264233507079365,0.16862888212064064,0.1507400259795495,
 0.3562578834753919,0.03768955429389331,0.663754991718457,0.821568986547604,
 0.852593405670908,0.34486549268433164,0.35428368157473833,
 0.14133292608002657,0.11024964900477319,0.25058411642947953,
 0.5711291688963882,0.644636834045169,0.8190204853234878,0.10783119177548461,
 0.21175652395184663,0.14797518373563567,0.7116796241941474,
 0.6690672557964648,0.4414172794281336,0.1716161720389805,0.5687389216774472,
 0.07001727956201076,0.1422206890055464,0.5759787713728737,0.6482456200857006,
 0.44381758742885313,0.6111604894162083,0.049073841296857657,
 0.6356689761071148,0.9772091984820597,0.6125144114881874,0.984590195390015,
 0.6228646344079156,0.58009153661285,0.28462907944573435,0.5800538753722871,
 0.036324724699463684,0.12550079974495432,0.8994274864709602,
 0.5507434553105156,0.8048372549820564,0.05467750108172831,0.7986569531550658,
 0.9294397200646579,0.12523058759095707,0.5631982281172474,
 0.07815780388326266,0.8591556652664956,0.5706657776159815,0.9846646491649319,
 0.00727564550659654,0.888742549410247,0.7720829694886561,0.9111240472666641,
 0.696397480916397,0.7072513074433097,0.09989727057128284,0.5855431854988052,
 0.6925053727428747,0.06850002150999435,0.8827120039498595,0.8277831058124897,
 0.6416942107472098,0.6789997080393833,0.7786111990912271,0.7643097801640425,
 0.1393001096342361,0.5448859086171098,0.8249996767574944,0.3500193924065327,
 0.47844988753359063,0.907717015877232,0.1441841955706995,0.5382223625294109,
 0.4255337611976677,0.9876520582104161,0.1221720001252522,0.6006258941267532,
 0.32145017347418936,0.5651543136481737,0.9112006216525851,
 0.19135114201025516,0.021012166877866556,0.2717087381466655,
 0.5529129865022634,0.14497579502382874,0.7271816619021121,0.6511392884176461,
 0.04306959328571869,0.5905995335926868,0.7394617002718087,
 0.15420201632475872,0.24209866065832997,0.20677636666289534,
 0.7999015678619364,0.6075996620707189,0.45609069120699197,
 0.05816215525724089,0.7722215538194821,0.012676678961128007,
 0.21978680758425606,0.8873383346575986,0.7284676026775024,0.885680466702918,
 0.8643381588941281,0.28985953291972544,0.698080506972897,0.07861130091070778,
 0.1326140108360223,0.44978108713630516,0.3621534368371111,
 0.46518239747127776,0.9217606644607962,0.7421187007263995,0.8493084725902471,
 0.1422368468084212,0.8336295906403423,0.1928151612624892,0.9422035890175975,
 0.1918435604465275,0.24940656049745713,0.5908661762677806,
 0.28203775989136637,0.14476781975522757,0.6092804171889044,
 0.2880001965772957,0.8300869263726597,0.3427088947178537,0.16770881274028104,
 0.29472897644643137,0.13931433509712288,0.05224807041701407,
 0.6848439510841875,0.29856154220732467,0.18495265945464934,
 0.9379744298695469,0.5913085662944221,0.5266693533621031,0.7409679201290366,
 0.4787776845620044,0.20470153525610302,0.21650320359010355,
 0.16641847575002033,0.8275058241143399,0.4530083043881049,0.5549579101315316,
 0.0034417228496469443,0.4409262646705411,0.4746825440967074,
 0.8132643484410664,0.4399536756023589,0.38065269738898744,0.1542391599353531,
 0.8334642594921216,0.4134363991839618,0.4153406678976548,0.4620910838738449,
 0.20078532626630896,0.2942368985864122,0.4067867624399951,
 0.13051082615191667,0.8831119350646068,0.7816319187698065,0.6618531655341947,
 0.7219039786851162,0.6044959222375769,0.2746341126839207,0.4885550398103018,
 0.18793725019156915,0.7814136314418311,0.8136181401568342,0.5381910702047192,
 0.39337872718323164,0.9411584714696327,0.1995369777681243,0.5480482998435029,
 0.18835704854464774,0.3411788002050846,0.32380580235563405,
 0.27354503547678866,0.546377661454258,0.13292957338414246,
 0.03375608704815258,0.12316397036991156,0.3479336018464301,
 0.03316865858881246,0.7253215849873614,0.9738148297046778,
 0.08559068991676222,0.9285505384164888,0.06805688596409709,
 0.31963821985654417,0.194254420527895,0.966657533269487,0.6665548084491588,
 0.3455660235402638,0.29989233991160946,0.06647416609544399,
 0.040780281712571864,0.5431067244435835,0.9887332433508332,
 0.09058632038012371,0.5912589140776108,0.40768581794529457,
 0.29327263172567786,0.08187609466132073,0.17222275890277583,
 0.6565269178814618,0.8174984677495525,0.9476976899556768,0.35215934203950583,
 0.8497028934415651,0.45571006458002916,0.6363672902707042,0.1416717517966839,
 0.015816852623144362,0.32403859535557256,0.7091683738345772,
 0.17099186688896495,0.9389086868469265,0.3501519505529913,0.339024470530488,
 0.3911181732648322,0.7960184528209545,0.039935877018968724,
 0.6786211034728283,0.13609316303076224,0.4082976302416351,0.1172466777770983,
 0.41293814459993694,0.6913319449674382,0.5446016530421316,0.7820386630602295,
 0.04053120467067617,0.05937165574289127,0.926392115058834,0.4828133677958649,
 0.6712336284914242,0.22453316624897102,0.9167801753191126,0.1738044955356416,
 0.48990337995194444,0.8550686691578054,0.23409310841085373,
 0.6513444138488597,0.42180214525988313,0.6057284954690673,
 0.34710352575265857,0.2990615673475532,0.21724765999876838,0.209620461945208,
 0.32391234606859964,0.4263283119217821,0.3722911815834318,
 0.20174652647753577,0.6967858588773841,0.4901101348208985,0.6822168394601732,
 0.37817820469936114,0.35402278324015146,0.4720147732867359,
 0.3071873215572285,0.7415604879636817,0.684859257287709,0.8926605952843775,
 0.16515234330106965,0.06018399287471898,0.8580439088571632,
 0.3020610542788368,0.8804991062516007,0.9356971154077529,0.8998976503507915,
 0.19391927292661704,0.6751031321568259,0.7800342467153338,
 0.06253291000551642,0.019454805663202346,0.35779731181039853,
 0.9576468850904155,0.12468609068860692,0.2952498620764683,0.386944817854345,
 0.3869820376621398,0.052862652477951966,0.7586676087872027,
 0.9429080966690497,0.5115453310155161,0.8259888710987712,0.10049695912422951,
 0.866477796044405,0.18922064687221574,0.7846837547439197,0.935451489044017,
 0.10921237071052181,0.9819406429279659,0.4569441275135204,0.1590644131198915,
 0.8723222328574644,0.951563824888747,0.3100217896921508,0.5264566072526453,
 0.6211898886269055,0.48958322567342716,0.19543862547526658,
 0.8229374201606058,0.5072489006909364,0.6218781392191964,0.5879755867917441,
 0.9685139572426174,0.6355539587095432,0.585463537232762,0.35939639424990844,
 0.5839138459903933,0.08347995994732005,0.31202828728853504,
 0.5358831914183849,0.38270898685088595,0.8718471977257308,0.6126343021442899,
 0.913955736446265,0.2080589827740743,0.9919009998793656,0.887466094681067,
 0.20109524571894477,0.2512601256546789,0.5913170252784584,0.6416313191740832,
 0.15045729097751115,0.11237446052822753,0.7854460819516841,
 0.03485314831346564,0.5277959995518281,0.41466242108238605,
 0.26212171968458753,0.4022076973713328,0.587090890288624,0.5253027099554827,
 0.9018423680877856,0.21090819084331736,0.9711830667504926,
 0.35897075008752377,0.56149343458782,0.8444884944070477,0.8851978132151962,
 0.4128856682703902,0.7778104185586939,0.9636132604950493,0.7232659861727121,
 0.9303612252976196,0.1889135729144531,0.024050731326012098,
 0.21598213253311482,0.05312649731365937,0.7837201599605368,
 0.44602928218583404,0.10458387196508334,0.9940592049533521,
 0.05808833080557996,0.29409891581613135,0.823052569013286,0.8893155793098078,
 0.13632683353956154,0.8805257181798255,0.9882479201698755,
 0.39351510181656957,0.1167887424383347,0.5919480624833263,0.7714261240479103,
 0.2265338028859324,0.9358385552482704,0.7530952538603071,0.8405732166824281,
 0.944560119542402,0.43720141124829226,0.6604674901505191,0.6238941318709178,
 0.2385669379194697,0.3085426894637111,0.24462950812706086,0.2363431379921459,
 0.8533072414357954,0.8811967053150895,0.9537496552441082,0.7484561394647511,
 0.10149924649989117,0.6664637160503237,0.5047773565513333,0.6226651458289628,
 0.21996434165255663,0.9147187315952665,0.9347344825360185,0.5376795364952761,
 0.8069628676864269,0.8112479656776365,0.19727519426437634,0.5447734360186777,
 0.9792593707830153,0.7740690957290574,0.16826785968682045,
 0.11636446504435993,0.986427540804722,0.3571717303053612,0.5620774103497884,
 0.2256968561787448,0.5490930897457982,0.11599745392313576,0.8375910358519864,
 0.6222832345295599,0.4824401423930518,0.12326345208308498,0.2088167878277314,
 0.5545878119064565,0.36176332498909525,0.43558346541313664,
 0.9786081023522565,0.19048654752537075,0.8720554940427823,0.8561960682724069,
 0.1133170307404826,0.06747600930525188,0.8460767616842387,0.8283451398544129,
 0.12586120137188372,0.24575582425692466,0.8461261274833197,
 0.03529437323001272,0.6235687033340495,0.2575387434373373,0.698139259172749,
 0.11282090968229463,0.5081834285029094,0.9660409656229827,
 0.34363596674282393,0.44578692947175125,0.9530068772453076,
 0.42112065262584253,0.6184239751493521,0.5176871710172072,
 0.45127588623794435,0.518970154285722,0.4601188530657916,0.4875835300657161,
 0.9650446238849737,0.7582197209362297,0.6438152287243568,0.15719136890762464,
 0.21958736844464966,0.7985076481921477,0.497647380707699,0.7711504611853631,
 0.38025413759525883,0.43380301655097964,0.013579995125495237,
 0.642302060376516,0.6413132863790757,0.0785411070262183,0.5443296765794737,
 0.8975402167460521,0.12968407442448238,0.029345937693746493,
 0.1948777490417095,0.266000852696202,0.6448166832799676,0.6870593194193332,
 0.6448802128329552,0.968135383916629,0.060291219341809144,
 0.01194117428234931,0.49975522293538277,0.5853449564101386,
 0.08767604268174378,0.8388202833798611,0.044792429258389554,
 0.8209561484202244,0.8758046302806555,0.021274121529085477,
 0.5300922635027273,0.11002500790814618,0.6950730895252111,0.963714857138205,
 0.26100477679929535,0.7896648336757913,0.22641693400550256,
 0.3579503018999257,0.7316431137847629,0.21926788756465365,
 0.17523947198365297,0.21370658284567967,0.19493694067660772,
 0.4157785223718391,0.5942615925821619,0.1937990880445709,0.24222482514249544,
 0.44770127208755806,0.864957523624025,0.16167196211635626,0.8664312176710482,
 0.09965624117454197,0.07811871294474138,0.4348853545857194,
 0.9048251427708061,0.9075257738019904,0.19365898004549909,0.1250230938446203,
 0.8874103449520219,0.6040269747354103,0.19236692894981544,0.2540431904344681,
 0.33533033387218647,0.672056783746129,0.09137676699249753,
 0.27924594235848266,0.2050835180489473,0.9490723348460041,0.8471451670848729,
 0.5348744968082152,0.9151482203916603,0.19507402551111597,
 0.45257856000327723,0.2917265805005934,0.39085074866482783,
 0.43547914544319255,0.43855182308274443,0.39427335587748435,
 0.3019110642786881,0.5711711982977333,0.37958163216779106,0.4891485853147728,
 0.34557893594908295,0.9265772691200775,0.02810040390435964,
 0.6837191755280116,0.0884076668457483,0.25187243102317824,0.2069167872521963,
 0.11531026109293951,0.5788167968440657,0.6021052401094036,0.855986687343944,
 0.0046730699014034816,0.4409665224460141,0.2007980662474832,
 0.6135943643413582,0.4598789355463889,0.6628273957135863,0.5711426056067386,
 0.13653804458329788,0.3371420444580333,0.45425189016276646,
 0.5897222928137927,0.9644445536007431,0.3125138815189934,0.20838408520195073,
 0.04946255939003841,0.531198952515142,0.5718125950997336,0.8566679968843243,
 0.8809941489453705,0.06930306374331874,0.25936703590605703,
 0.5732664928234471,0.5920278055693995,0.44150905146485553,0.8166327089419386,
 0.42792676331515456,0.1687506233212095,0.043979777877756865,
 0.4926115625091698,0.27282058237087026,0.03731360429366437,
 0.11644717025982643,0.8537134574309381,0.5467047181158619,0.6228469251400425,
 0.3069942411138187,0.16356719482457582,0.7023447916866623,0.5258117312862053,
 0.9188521768847053,0.2622637380948719,0.7249415431116178,0.598625521784161,
 0.02247407556643377,0.9268257910037806,0.6009761905769804,0.6126937284546602,
 0.30028550641590923,0.6826107324701872,0.682806999727359,0.3629740994342412,
 0.536440400151364,0.7165382301795519,0.38331361170994627,0.5197250695728152,
 0.8184267988689092,0.2335331483919998,0.6207709134065198,0.1881671076531919,
 0.14111550787512128,0.20159495415231754,0.7109296055055282,
 0.45723697221092774,0.9676894878553175,0.4266917321742665,0.5625673192658174,
 0.8917431189903118,0.722073702331476,0.40807087504751804,0.8288846875548156,
 0.6303147767725564,0.36055255645692585,0.8008085619975007,0.8689393384464019,
 0.5253860013560183,0.7690400572523451,0.17005144683202844,0.5710101088245193,
 0.4771334604908628,0.29303040788961443,0.4716475138445573,0.2027423113077892,
 0.8969229370171324,0.7643159724560903,0.683084554295804,0.23070332334136745,
 0.8870576698216355,0.7039519780027667,0.914098911499616,0.23301666223466033,
 0.08429213922679801,0.5651512068909346,0.8513494394970468,
 0.34900360716997425,0.32542773467823793,0.16767660317003807,
 0.6044361513167046,0.32973428213062994,0.27998814749881407,
 0.07231858104033062,0.5774855429612994,0.24407798635408762,
 0.7488710024139574,0.026891575340879936,0.4423043101826103,
 0.15153448022984062,0.3732700610282986,0.5105088097527996,0.7689239203318947,
 0.7616620460159458,0.9469969043385806,0.9467155951683839,0.428165442579426,
 0.03072700312352339,0.7348090372361172,0.5960662300847289,0.8065164399298093,
 0.019772483826124287,0.9224624518233233,0.311053428803723,0.5723629316776829,
 0.6599357805754795,0.2724176878629774,0.31369190068549657,0.3192038355922632,
 0.8392789351234271,0.6549647806416586,0.682912018710629,0.6378142479460214,
 0.8507736126739913,0.7894080035385083,0.9141269227739385,0.9494231781880897,
 0.9446254796719196,0.5006566315099885,0.07037886037128349,0.6502868308716478,
 0.7450533121462434,0.5065064620463677,0.5288676138726033,0.09400034839747584,
 0.9871185019556638,1.3570352218850168e-4,0.9923092290750977,
 0.6694956574922077,0.8077612473307412,0.3482887159169368,0.6563283429048911,
 0.22125880191736091,0.4076466662322549,0.7514497816062614,0.7280255098107453,
 0.7686277339744114,0.17099099386886651,0.14132401850133713,
 0.3434367330042899,0.016546469919058504,0.5368262624905721,
 0.039582507403894684,0.330153640032422,0.5100877711352367,
 0.43551159804745354,0.9235494658714573,0.42826674989375924,
 0.15880721395465658,0.6086956473186856,0.9108678452096348,
 0.03839782533279823,0.750022407844891,0.30483447045575396,0.7660263646484228,
 0.28163536562914493,0.1695907437454096,0.16649532228185304,
 0.9048886069902637,0.8963961422940256,0.26334912171315106,0.5658778402896837,
 0.5413908568671547,0.27546003039782196,0.8702016836240657,0.5942350929250354,
 0.6892243862510208,0.10764433386459316,0.5222868997235207,0.3950591613120905,
 0.824813426858093,0.7301492675342475,0.3273265207208442,0.37870140926867435,
 0.4033444346193519,0.40690913798118866,0.27549324843018874,0.532700921985994,
 0.9552058548841555,0.6393254482649048,0.7214546166102525,0.28387005259200804,
 0.8388262299125173,0.045884188710846874,0.7935721243043812,
 0.13090100756240697,0.07493044850174402,0.9506103157028912,
 0.5494511209736817,0.18633016549615544,0.4945365417688282,0.5874113238805504,
 0.2935468509194288,0.17512484514752147,0.1750013802338879,0.9508207822850139,
 0.9543396079815434,0.024630958783265378,0.30495526215215163,
 0.31698304507302355,0.835062705128091,0.7345067778711981,0.4165880328138414,
 0.56835659870841,0.40534149642838224,0.84970847087267,0.7087427539308128,
 0.5523761149881329,0.6571661181922335,0.5405007946251377,0.6727446972229129,
 0.3455033455568557,0.2138381783693547,0.8522600987498259,
 0.032167966452299246,0.4435623502300655,0.09895884370868291,
 0.9265987139527557,0.3548089563297514,0.7670762045353343,0.2443688966362073,
 0.19682543647853534,0.24175636197695904,0.5621393530093439,
 0.7099662315975662,0.7034912390469575,0.9830635683371429,0.9853705007118673,
 0.49066313341606227,0.9730899306913354,0.6984999184321432,0.6372202429726669,
 0.9020641895907832,0.5569111589629745,0.5586545250986238,0.38527806123607355,
 0.7819861849706189,0.9130711262821214,0.5497799594506099,0.2546925137947339,
 0.24062653307778403,0.9564089349644421,0.07759035964271521,
 0.46618449850206256,0.5399725385926047,0.023113922320957436,
 0.4234336841414903,0.6382577088159664,0.45410452312404737,
 0.44266621889273394,0.0664980430459231,0.5480269186281501,0.83228238706236,
 0.46500154313682474,0.2114421923345179,0.4346195243730251,
 0.45833824140033297,0.42132157865570774,0.37616998056245954,
 0.8203808676840377,0.840053905193102,0.8116619258545459,0.23463641284234527,
 0.16669526237860588,0.5220040449298404,0.10417782154758204,
 0.07450382695492896,0.08218430960269041,0.5567652608549077,
 0.33155607960642963,0.827389500673979,0.318628068670874,0.42221695008624927,
 0.13601194912854608,0.6031289069956197,0.9685244057585214,0.5173247014894188,
 0.12366398848969329,0.007993851691510123,0.04531944778772967,
 0.06785815223440783,0.3130669465195288,0.1523698110528372,0.3159917618892566,
 0.9057000528750967,0.20193948167916975,0.14483289582201986,
 0.6236138636506289,0.28719657758680917,0.9376509164234204,
 0.12200000424313628,0.0872168529728864,0.3205621372172508,0.7995856273448735,
 0.4802796516405692,0.7781761527848517,0.6121417711092996,0.3222046561767886,
 0.29986666451838795,0.6743171602026277,0.7678053304227822,0.647884156216167,
 0.19987765100904198,0.9935115507940104,0.9258161469999503,0.9481650484020485,
 0.09308494726175764,0.8363346042521814,0.9483522856130241,0.9510069223274459,
 0.7819107971322478,0.6107970122795263,0.9062636896518684,0.605769668986621,
 0.43793904450769916,0.7202142448133486,0.4340968519633449,0.273604887559901,
 0.8725678289607521,0.6054301555330277,0.4924531903839884,0.29984196136134134,
 0.8782218941972489,0.4107262131233742,0.6926215366218983,0.24855315015658053,
 0.8882999037202679,0.8701674743141641,0.3188365761962694,0.43533520791598124,
 0.6114865472635933,0.9088029030905308,0.03708461704726984,0.851204271596094,
 0.9309585110864552,0.929206728998297,0.7330451486953143,0.794898208977295,
 0.30254775202209494,0.5252940355569058,0.7414154690902774,0.2295677270483233,
 0.5863893554898945,0.8046642401878621,0.12114469455048982,
 0.12440431643281091,0.9343252242841737,0.46615891731791914,
 0.21500447684230628,0.7337258678089792,0.5824492279873758,
 0.0016345576581801025,0.531125537010903,0.5226487281736042,
 0.08857350146061138,0.9189179119236348,0.4662191894050567,0.7450269391238111,
 0.23061128576663792,0.3476172559738241,0.29186906026705584,
 0.9759875063994579,0.4786204014953964,0.22771921941224527,0.7245867627147551,
 0.5104087866879855,0.6657791157697649,0.8530696301325845,0.7623401191668127,
 0.6057504881507086,0.08064667081652244,0.5938016501999992,
 0.10809861502971785,0.9883105134255834,0.31847960153319843,
 0.9066981163844653,0.8829428681067257,0.9361867514789178,0.22844189918671454,
 0.5850366105243963,0.39834596807666467,0.04644267673368041,
 0.5147714024202262,0.982803996112728,0.6137111369639539,0.17702962615278206,
 0.2274759863321052,0.39305573066647703,0.11182472928746245,
 0.7216170962923745,0.6850785147732308,0.3900078377686199,0.7301207399935948,
 0.3058023679675358,0.7677492853368867,0.9186100811272442,0.6533459253541705,
 0.6375259164383204,0.8416785274404011,0.47248689937827504,0.4288616617271537,
 0.7207940513453681,0.33348662970347365,0.33694813593579087,
 0.5515994317176869,0.8518296658352749,0.4001616123986106,0.1597051569214598,
 0.9357956898808607,0.059920978020945426,0.977655462691778,
 0.016171279657516968,0.4613238619475061,0.3714515100357666,
 0.6354224955254874,0.7588332723994095,0.07836274547989563,
 0.04581223656442468,0.02528715117666913,0.9014807205824368,
 0.9727918853459956,0.4495630134289098,0.6149284239872339,0.37509043547782017,
 0.873330416144115,0.5329900111601151,0.5313219889065223,0.984861818737488,
 0.158236771932314,0.8575916978581907,0.585156522990657,0.6620547992886582,
 0.03516819095112689,0.5528949234324277,0.33601166590893705,
 0.3516988427582609,0.6249880041558216,0.8348205088756762,0.5641235604628129,
 0.8653575724104354,0.8841585698642996,0.054621230376431185,0.922501293291636,
 0.3145518220222838,0.8603807682449024,0.31227837380062784,0.5820188970406591,
 0.4619536291545553,0.4594945260531954,0.3895700431842011,0.3720942108541142,
 0.7267024982823139,0.35128495272664784,0.5803990783362429,0.7600050634304505,
 0.942930460551297,0.4661632764229152,0.9071439664026069,0.9998258742319299,
 0.2727463287033144,0.42756681563421806,0.4916040567492084,0.3434772227792484,
 0.3463412653299275,0.1047102425326869,0.3796851677207651,0.05347645727405981,
 0.21629060548131585,0.048670132268085475,0.5777530754217106,
 0.48232956458544185,0.9629540863912589,0.04711534460302891,
 0.4339375560748411,0.5809144476093795,0.023472405177927036,
 0.8516470035967031,0.6290476847655813,0.6745791885240502,0.33542717586962567,
 0.9217376236840061,0.09250075253447232,0.4930004274229369,
 0.20676458987882873,0.5387187701643578,0.42025845678057316,0.842520903831535,
 0.22370080641171186,0.6766514305724285,0.4604151970909812,0.7167499118264294,
 0.7460953340670189,0.16671563748050233,0.2756444591527193,0.3503485172461084,
 0.301032957813839,0.943462392977867,0.28706145313498743,0.10799461688409506,
 0.4697629536124005,0.33875647948597554,0.6781343321606557,0.8536949578295528,
 0.26201567872453524,0.1949175869793347,0.8174623854949343,0.650086926327855,
 0.38814834749327587,0.4779735466360545,0.09376692345371185,
 0.4645359643545428,0.18092290122228838,0.34935218704135895,
 0.4946527015397535,0.9976653987698567,0.43609881841165077,0.928632561970755,
 0.5300534551617064,0.9871863811789773,0.021320945188412654,
 0.7797900633241088,0.8481521449950061,0.7167926458549891,0.3315508841679945,
 0.3134754070600251,0.17754551203851965,0.2615483021073767,0.7817376729861268,
 0.9270023615391196,0.9821635112119252,0.08560652579881867,0.4043294477681564,
 0.6291106528100993,0.730016936114711,0.6221051796085065,0.7929969213534408,
 0.9218545009883322,0.9594106967308244,0.9093235299409409,0.13733577319252588,
 0.11972277341547821,0.9496650166153864,0.09527542744390638,
 0.1474160056523347,0.9027886478822499,0.6432851916105853,0.13483899676324973,
 0.8049544365587412,0.3678290961029098,0.08174409241018388,
 0.15951613707633883,0.21813609780822418,0.07067821130443919,
 0.6075158181425137,0.33173755048717324,0.8078859736451007,0.5381552498833291,
 0.5140779404087004,0.41303640536699393,0.4471393159994427,0.7542705838342019,
 0.9174762281930313,0.3062050679781828,0.3983767689574913,0.7224248951566871,
 0.5631122661870158,0.21554774001580757,0.6298312128544709,0.424316030902246,
 0.061965039913079445,0.5155814564102406,0.8728376758823028,
 0.7285641207082127,0.09715873681728826,0.06315304316524528,
 0.6185940536197977,0.9769970620862349,0.882797805562851,0.16822988791657933,
 0.7502694957742873,0.09974344540788393,0.3385263084605784,0.7506627737649759,
 0.17173127169366642,0.19078244870827166,0.5943395955922203,
 0.9339656738852531,0.6833374858083243,0.03976874434721378,0.6360876255622674,
 0.5592762854796475,0.30945551580646136,0.9096162366104631,0.3646878336914342,
 0.4336121168726844,0.8031306723787024,0.10897090597549053,0.9106809711060274,
 0.7347421397032184,0.9762100603686382,0.8519628150321176,0.8484658602614237,
 0.14248799342276863,0.06812321770466567,0.5715755477403132,
 0.3275218738098302,0.3366037040065788,0.07130650780752457,0.8189339953355491,
 0.416596579528135,0.8853304558578012,0.5888076994800083,0.6657462142910685,
 0.01162492265309123,0.5817603794426731,0.3301290416119499,0.3849681887550387,
 0.5343481546974002,0.554159925072787,0.3290700224206373,0.18863074965761262,
 0.34532590189957846,0.6076906483256315,0.31692434471793307,
 0.46864107970413993,0.28233706901136835,0.3914865102632845,
 0.8205463982856421,0.4793760772277167,0.14534413538739366,
 0.12263402750052832,0.8311369531997478,0.4024935776690497,0.7574936219095347,
 0.4054169316858156,0.528954477455187,0.3919211324668641,0.12342215831190406,
 0.59320017682055,0.25307642834565347,0.2023553195880603,0.7619144855609659,
 0.5683139562168816,0.7258232812212514,0.04558595308307589,
 0.37856306366710957,0.4271027498382164,0.7559116783519593,0.2945886959050329,
 0.7600325527918196,0.7570799211360406,0.9542049658984539,0.08195050335641296,
 0.7830519496170815,0.6932761871001518,0.3657686046806856,0.08970375392780128,
 0.7479302761975031,0.5568005901288382,0.4649404871055469,0.04661472317860915,
 0.6002362684227693,0.2896052621471026,0.12765149032604617,0.7564234625602121,
 0.9122337831204625,0.4028232921148722,0.3756208544481954,0.17226382862499334,
 0.8284559102198783,0.19247867168936506,0.9502753282056949,0.3477191530288076,
 0.017067990519942344,0.09762949295307388,0.8093730111512886,
 0.5970200721641841,0.9840544246915621,0.4140002466950268,0.3321054385435247,
 0.40899576197899445,0.8847243004758888,0.30353272976480294,
 0.8384660247036684,0.8917646594601911,0.06304517666294207,
 0.49782406343338426,0.016187350749320206,0.38569682213462075,
 0.9177190573081128,0.21679998615789353,0.8034936105299875,
 0.11664566670548171,0.4186722077978877,0.008747935814260033,
 0.16492296753471491,0.20062393584699556,0.3870969015884492,
 0.7376063321452276,0.3147127950985572,0.08497954448578404,
 0.07921114392526685,0.6857657295888502,0.5219427888446129,0.5127683505767724,
 0.5100178736527677,0.07189994690411983,0.654606805626513,0.0747844501629703,
 0.3558225863539449,0.1616200272640621,0.28141968771098513,0.6454022647100689,
 0.3850119803287251,0.7295648270301874,0.9641589202457056,0.8334251275101224,
 0.7672094799126807,0.23071656586144362,0.862456888103446,0.9265452792359526,
 0.3661162505254092,0.9412089261623312,0.8343650347060986,0.5824253053738842,
 0.36885887404223927,0.35485368853272836,0.9753043891124296,
 0.5639722027564442,0.8127606054207757,0.7842500665338299,0.8640416287434072,
 0.0801589460776766,0.41157407183570804,0.5323944526093487,0.8235943455112015,
 0.24577110593685514,0.6585157963939463,0.8892067093420247,0.8730463119876098,
 0.7719113236664596,0.08805769859531654,0.9524637184664243,0.5278319728901703,
 0.9816663517634027,0.046509340159406665,0.6926451022120741,
 0.8104372432504107,0.4325567143142913,0.8869579030604318,0.3680859260096988,
 0.0602998394198766,0.4040608111500329,0.579472983200115,0.6565211825809449,
 0.3715777944097137,0.12387755857358385,0.9690197870715368,0.3311104569952973,
 0.49498438254195976,0.0026492016883991054,0.7107190687232051,
 0.882480141606666,0.0115114186279901,0.4917337441974948,0.5540406990585858,
 0.2748815719001735,0.4211166752286085,0.6230161008761278,0.47423870008677604,
 0.2635920252049432,0.8030608900345059,0.8349111364682105,0.7068724161852883,
 0.6779339853448226,0.8101852243515933,0.24476136855384778,0.647917582340484,
 0.9707490295606437,0.8230279080150801,0.31816830547077846,0.9410191761836424,
 0.9144093285912009,0.697132100909423,0.4538967230712603,0.4921967061792709,
 0.5321556513433331,0.9683660845535703,0.7568232155277659,0.5983852909560627,
 0.5887074740290519,0.18927518066439264,0.1438462135989088,0.579492570390041,
 0.9237183459242497,0.034673968206889616,0.6897551474903377,
 0.7675347745972136,0.6818773120280139,0.23189394861655277,0.5971151383627763,
 0.11994498911182494,0.6198479031260278,0.4189998700455705,0.9704882007201913,
 0.31370634932998853,0.24121651229398888,0.030512623855787124,
 0.016620860197393617,0.7857882715833324,0.6294765943818973,
 0.03415413163101466,0.3841147816298196,0.2162733804103414,0.7937393148106118,
 0.28059498542922456,0.7081089546907846,0.5811535171763658,0.7136784449391902,
 0.44192890735912016,0.37420687091484417,0.7465918419373567,
 0.8235036602121121,0.8740931643959832,0.2178632170965802,
 0.004396186480282949,0.7701134272201017,0.9989182728936259,
 0.13215785744199915,0.017522247565342153,0.0897324800528988,
 0.6093369755900265,0.528236610481031,0.5655928657508033,0.7186311216910771,
 0.976430155228179,0.2690759155250446,0.111321064635729,0.22976397123108505,
 0.09931913434289574,0.04336322588970021,0.9989519684236746,
 0.07870996674750819,0.754096284055048,0.5652572973194544,0.19735765362959024,
 0.4237487079165553,0.9072674564632993,0.7206593956335946,0.22645375685689284,
 0.6116227822465271,0.9161825818930635,0.4923594988946861,0.23658029879060893,
 0.06272051470278817,0.8560193283091144,0.5151392499661962,0.5339605370975998,
 0.25701558478678377,0.2904023380181918,0.7659686606621383,0.8091233694586626,
 0.009195642173356777,0.4632668574528459,0.7661709993059318,
 0.9198342390536486,0.3783813095697902,0.6481277357241079,0.0326104032564164,
 0.33829450781084847,0.7011256252738769,0.9326582775715191,0.6523123180273814,
 0.8046270417999817,0.962895198656053,0.22062124068700384,0.2336113276143138,
 0.5444665341693945,0.9380412575442066,0.6063045203275452,0.43692510103261306,
 0.7089763951488913,0.30391175856368535,0.5688673085010997,0.973328996482975,
 0.05559525203998006,0.34931075048273763,0.9086177119086666,
 0.6588223528057123,0.802860877490195,0.2769369294844133,0.11158252270995084,
 0.5654157026835981,0.343374649741515,0.7295395272252415,0.605138694861293,
 0.4087082499992928,0.10015886353409142,0.7486202825044036,0.4162089236647619,
 0.9527312386665577,0.7062685737748216,0.8326577431149493,0.9539447861368727,
 0.24909814005584852,0.7801103817467916,0.7956200774043624,0.2105113566929191,
 0.4609225518338853,0.8144691119608646,0.4259838616214465,0.8431576948437941,
 0.047563347376445186,0.04897861975256179,0.53719545675663,
 0.26869989201890865,0.31234457719691133,0.09615569201069274,
 0.5054611314962847,0.3846891969174544,0.05601696145890789,0.7519952242115584,
 0.4458172917452834,0.970742273286737,0.5247445337444413,0.7563477024943519,
 0.5374093820536684,0.7583600383815328,0.42429172936110415,0.9690871552699949,
 0.10416034330491208,0.4112622126533725,0.2468855031059598,0.9049283020816269,
 0.6832101640509536,0.38087441398880817,0.9184206476916377,0.5934233143502077,
 0.034365188296785476,0.34181645288601736,0.24576130751926417,
 0.062198920226770094,0.05366084510846292,0.8420940709659865,
 0.8509792429741738,0.6889569966213233,0.4379217364919987,0.17227245956531734,
 0.20124983001003882,0.4902408156215079,0.6633021465700177,0.2259110363008613,
 0.7434246604680758,0.43895749341987045,0.9463812121220352,0.7675549974793545,
 1.2160674667671678e-4,0.20208086349513033,0.5292990868798885,
 0.09372908282851056,0.5484456465064323,0.6417939147237943,
 0.46813250562517406,0.9681188944225073,0.8838416983647212,0.6206846476498129,
 0.8067590288275737,0.02138577055910762,0.8023188001357147,0.6471509077967927,
 0.8194007855858785,0.3130360201426977,0.613903289620299,0.23925538813415193,
 0.4395991414559003,0.9700548965421114,0.6959828293483858,0.2940146495080166,
 0.002147668274686043,0.12643522601829427,0.9123774827490261,
 0.07795454896223952,0.2504440950298995,0.08849051365355831,
 0.8995786291817653,0.3761392289020732,0.18173525794470125,0.3602580404729955,
 0.6747566157523495,0.302674543465969,0.25765343808554897,0.789356644955399,
 0.0974259775958799,0.06662683833432903,0.793553184410189,
 0.048307534328160706,0.09947396665702202,0.7492391357554503,
 0.9971579175013632,0.5529455591094278,0.7571968758251558,0.48506548774487823,
 0.984263538634278,0.3112435303032248,0.8603400268707168,0.9568109874527658,
 0.07045950360964215,0.8832060552047636,0.4258712142068085,0.774031318414452,
 0.8911672024794439,0.2720355717825015,0.2810255256940448,0.16302407981282585,
 0.41291226820682425,0.5376501896634561,0.48383129835261107,
 0.5251939183959647,0.3410946351998345,0.21031033411380484,0.8540375431105144,
 0.6482803690214448,0.14559354350259324,0.3914153540419947,0.8800100231377171,
 0.7060984324989434,0.9144126455962736,0.6637274741154368,0.34503075100238956,
 0.2733267892675634,0.2992378367304752,0.1226091718763298,0.7432808034023282,
 0.926503208932839,0.43058812331152185,0.37378881739396774,0.7356563494567878,
 0.152103616075834,0.2075934113014143,0.6194123642788566,0.32334327543948127,
 0.43669470073587646,0.8538152717131184,0.022039664976317708,
 0.9021445921707275,0.08859917119089833,0.7503223157106618,
 0.15974056554789795,0.7319923170999908,0.776696238312363,0.24650790910238374,
 0.5222220381396004,0.5278701859617163,0.6032870828821962,
 0.012707042113528466,0.6358024417060575,0.5368755909288495,
 0.6652091338118795,0.6668129059496717,0.3071561426329561,0.9997122407231727,
 0.4486764479303953,0.2736748265564475,0.5494396843859486,0.6696140260741467,
 0.5683613651876906,0.9169080095957935,0.33514597802652935,0.593341647528672,
 0.05016037301796272,0.7011980054241769,0.8480413244242034,
 0.40960007596789827,0.91300751990238,0.6116242747299686,0.5344877318214513,
 0.9922384449355032,0.9950821018216784,0.029926657367077336,
 0.5236550446268043,0.869668531483959,0.6211198092098286,0.8466085423222649,
 0.30065003715713023,0.5021551441152454,0.08881743508536444,
 0.6631089561814247,0.6895990268204567,0.6765042955697717,0.8232179187389846,
 0.24921283865936583,0.7653490883910519,0.7261932598685322,
 0.008370491950381642,0.9873194959781939,0.16761724399256572,
 0.5395378288951571,0.017717148891777224,0.9759728410125836,
 0.6286460571038447,0.124311582620933,0.4225154212934471,0.6208597561860983,
 0.7678185343194363,0.2880864442107677,0.260452668743956,0.41130708091225443,
 0.8040807305062065,0.18847142177258014,0.11138896967307388,
 0.28946589058104566,0.4358499743888523,0.7915210251444735,0.96840512942849,
 0.4475446370385895,0.08765914872538527,0.30982032634674417,
 0.22510140784950694,0.5473964779602724,0.036275231964056465,
 0.655599087425091,0.4813858752312854,0.05735874063638746,0.573081022335077,
 0.4525388214943048,0.5688010401339135,0.24265248835035413,0.180161814265838,
 0.6538610850970481,0.16510680034764502,0.0056381385250064975,
 0.3229337661128946,0.8188484411230185,0.6644550571942235,0.7874414899935877,
 0.6777079086140678,0.9366671930833829,0.517132468048544,0.5934916162443398,
 0.7791476300318115,0.4081087191634518,0.5643324825642065,0.8438968715387521,
 0.05295320252776348,0.287527133129289,0.30576993639924055,0.5075745746727055,
 0.6755454388434964,0.11445139805917448,0.19639911096828078,
 0.11873017801191255,0.23444521991525735,0.06823343020106265,
 0.48975254826774506,0.9026019891887032,0.7031101846637493,0.2776042230439544,
 0.689359515630087,0.40577431857938673,0.035433959130484194,
 0.8711107201305863,0.5812671678384241,0.5376188700065861,0.2789786960180318,
 0.3561287168553319,0.4392011472761518,0.5125658033217302,0.08361114643105172,
 0.09661798321258819,0.30112477413743144,0.23677082650820214,
 0.10933702226459963,0.8398781326225898,0.40959983241384423,
 0.48031030566060295,0.13777986518712704,0.21001534747112416,
 0.338909329515075,0.4927318658816614,0.36560083798090814,0.9048489112061705,
 0.6994389143884535,0.6795656279488875,0.4287024256398857,0.28100456873027324,
 0.924232576738501,0.6187949535360662,0.28726309553466967,0.6101399083879293,
 0.2845988511739531,0.22076845220039187,0.5425651404008345,
 0.38322154320811475,0.5919956873132599,0.8923950806107783,0.361763584327081,
 0.010079594364834743,0.19627940918783304,0.07824687756779247,
 0.5625961697096858,0.3932208795733484,0.6894922030141414,0.9175447613191322,
 0.41306161043881906,0.8904949952640535,0.7042264634568581,0.7970047697316301,
 0.07632008937878754,0.10525418744721426,0.8808534264482414,
 0.42587956140233196,0.7275482090453339,0.9672851923496247,0.7469143661361333,
 0.9381221119157829,0.15355228027242207,0.23327369488448513,
 0.16840094676341266,0.8828822682667368,0.22543404460789895,
 0.20653166685752256,0.3048008312106184,0.7128906691619677,0.9258920251527921,
 0.943475740807775,0.8534060653116912,0.782318160465555,0.015454630492459098,
 0.9612427343136523,0.28869769318974625,0.6165014290219353,0.2603825510812834,
 0.44006812681245056,0.357901698989852,0.4905660613572077,
 0.016244258890697805,0.19261006679267878,0.6901222274315003,
 0.048925643653056294,0.2090539727634798,0.4354603538740941,
 0.27258333116302447,0.14887640892767862,0.6046372558900809,
 0.8772876073563697,0.21319768656712634,0.3036153570481379,0.8158012796165287,
 0.9463543892035864,0.8247214535739269,0.22498070846764118,0.6717982010134296,
 0.6073908488093129,0.12648559126155157,0.7070686868069078,0.7485501356099181,
 0.9682245773453579,0.5427459041880057,0.06769202784187722,0.750375606630932,
 0.8519586094123552,0.684112913502194,0.5867171727683905,0.9478050861352658,
 0.47172890340546325,0.6641355262212212,0.9000075532631627,0.928468483005417,
 0.9523450491514046,0.64310804664136,0.7520217966261328,0.5491731349447114,
 0.024829557017368886,0.6656822081337264,0.09421652514148726,
 0.11695208754671871,0.37787597112314986,0.49837684637556534,
 0.9227741924077947,0.31414051096837947,0.7476478709170302,0.6905342604336108,
 0.6487570035007072,0.04329638216495901,0.10587269136366695,
 0.054525077796646504,0.02621393411028683,0.08249558564400106,
 0.08638812030250731,0.19890501563076624,0.461847230957663,0.6677057873856577,
 0.7374589775224281,0.7026623719081256,0.508975054271256,0.20387916597920297,
 0.44867206447688546,0.7631839847032502,0.80840900202756,0.3220687592153548,
 0.9733346990402266,0.39578094705341793,0.24060967837788572,
 0.8144433424883992,0.2739157018965026,0.7626979909759999,0.5279847781973526,
 0.24199552943864433,0.3078374885209302,0.02012405787362681,0.657702585936243,
 0.6186530127678066,0.8483727925768189,0.8848043845793108,0.553265068595486,
 0.4508743247097631,0.06300067881840077,0.30158288445602666,
 0.44681001286261324,0.842703297883477,0.7011906299886639,0.5421446000827984,
 0.5336389817600804,0.9338713393632266,0.40942895742249963,0.640852290678527,
 0.09509828329605319,0.43033030811776474,0.39036279595308754,
 0.5746874729750492,0.7067168221885829,0.4871009801734374,0.7118854715388677,
 0.023496569481262553,0.766649705799886,0.8358112020766209,0.9293893603099054,
 0.6671572371556382,0.39159469823397686,0.8207036880671885,0.161594956053023,
 0.04130627833899625,0.3370979516146997,0.3813598840620934,
 0.49879113187455815,0.16518053777963315,0.7395477606665556,
 0.016926948072045978,0.9761769511420787,0.32985749373828865,
 0.46021774935527227,0.7740308405618216,0.09738652376116252,0.819211890825372,
 0.9363563402505324,0.4599535000398879,0.2481298628411276,0.711848252937061,
 0.14979867773749056,0.9382225268313188,0.7116648055451977,0.7586577281101834,
 0.9377185090647916,0.23331465159593168,0.44805709856483233,
 0.0733825022864294,0.8206632042808469,0.12806871343508142,
 0.08739792673704472,0.20599637688419992,0.5339460158497655,
 0.1473431302452053,0.3683148747759952,0.6311380164725234,0.7801945108754501,
 0.21600405604071493,0.5852390624546064,0.4426709485612893,0.6260190405316649,
 0.7724744633681486,0.7017623218736851,0.1681991941398603,0.9752321099983223,
 0.326661685154953,0.01680783599340241,0.1309015309045095,0.42139267674750236,
 0.13485014585743282,0.8125963718648512,0.9950674376627058,0.5882146146318863,
 0.21447038608503743,0.3538760051126286,0.6481727528019746,
 0.38353241101859903,0.6286766537081364,0.24957721603899574,
 0.16223443676536442,0.8594225957271239,0.04251078983616219,
 0.7062439165235336,0.9637101287626515,0.3635552960790206,0.28266651441499824,
 0.5486051308867294,0.5837112805432949,0.45722087000073564,0.3324657903768373,
 0.857614435241711,0.580541765170282,0.9021636309258559,0.2862504502837172,
 0.3345621852641135,0.5860410766003847,0.4469220875519746,0.7183102279871301,
 0.051879483018169914,0.5607688630705547,0.15929539132071624,
 0.8179509193901021,0.15069771559575984,0.031097420052207503,
 0.8370968353272368,0.5998166382559809,0.412344860697041,0.4034926454166744,
 0.19939342003619498,0.3302582637488076,0.11910010773053925,
 0.33892688879638533,0.9082988026495525,0.8899819542539228,0.5998638849821102,
 0.027642555813959913,0.8559360104960252,0.921305067419075,0.8647770210632091,
 0.5271290704703985,0.7772454575889544,0.8171998641464224,0.38170755802555345,
 0.8526244642390838,0.4896175887583484,0.4102653445261566,
 0.0027076415362868866,0.956864260612984,0.030321282474352285,
 0.2785635187977067,0.3274039732362456,0.06182858393667057,
 0.18019441507685519,0.4236688235365407,0.8496098947767791,0.2796130713857454,
 0.3166168992241808,0.1403696515098367,0.46005347085926673,0.5898517652625394,
 0.5289085502801466,0.7944083683297604,0.165142547672537,0.9062202012778635,
 0.9619188052333838,0.6359139052840768,0.6064867933999255,0.8159908339805271,
 0.2928888997510978,0.9946618157513617,0.27896603861021896,
 0.16338218178793262,0.5225717599370759,0.01637046577840129,
 0.35747734436439205,0.18089103600389278,0.16677882265034216,
 0.11224841791388684,0.8815778385478684,0.8562450678378241,0.0345479233712781,
 0.8626145640488903,0.3833199568255766,0.9084954718536182,0.8711630931771952,
 0.6968946678726992,0.9905183868198003,0.8628682752127828,0.1341760957303808,
 0.8715224624192336,0.39617550671437696,0.8920546434273954,
 0.48356056383854673,0.26133506314556576,0.6615505842250954,
 0.6133611955168518,0.032140701577255903,0.050396922160088065,
 0.12517250687724935,0.3721224779614253,0.054670949243451616,
 0.4020748133820483,0.3768657792227954,0.2321094647913966,0.8086594660001212,
 0.5816906520436842,0.335753870784926,0.9503266262928021,0.17547721635302949,
 0.31250364089901783,0.8936771004169438,0.6207136743636406,0.9549697765696876,
 0.9882729814288835,0.6976599332787531,0.40851011838460205,0.1030853616190166,
 0.37057192291995666,0.6803671480571364,0.6560350498094224,0.8247443857525011,
 0.7258448487290675,0.21577827887232337,0.8901197649529816,0.4019263660518322,
 0.01398076880943755,0.9178075205989015,0.876974621007192,0.573298994335818,
 0.7970347121297517,0.45874233080327087,0.6553131095295341,0.33695162458351,
 0.09456341694839199,0.8272236523480507,0.9689446812435811,0.6656612147600532,
 0.4745611343531626,0.17876519806347135,0.316842770258432,0.8819324617838111,
 0.18836168981454204,0.46765831369509736,0.44950751708478287,
 0.5960848040135549,0.4977648235155282,0.1740636156608396,0.6970606079735665,
 0.5088872020603188,0.44478860192700986,0.35403577395990904,
 0.06296401820856012,0.8863870578709154,0.03815225479861162,
 0.4481213913180391,0.44131314456102666,0.920923097890216,0.4363845893091437,
 0.9161770755862625,0.9860001099425348,0.9601848792323519,0.6682627924762736,
 0.6006921929668456,0.20631181593165449,0.20722536693611238,
 0.6738789061880779,0.8725091736396268,0.693040193763694,0.8150198106388179,
 0.44082691556189046,0.44686933768460824,0.24215387030454316,
 0.9566056141808149,0.26234534993108505,0.45108973011852926,
 0.0652133413116992,0.3036076658870247,0.28531648366546036,
 0.10245981003497384,0.1624175900019681,0.8165580627335594,0.5470453952221576,
 0.9565419934006869,0.09260854378141936,0.2719741971202,0.7890739008192894,
 0.5408761584501711,0.09154007918817242,0.4347426794256589,0.3845397949698778,
 0.9965916903700197,0.2906026975370416,0.5366738723188148,0.8312282297907423,
 0.5076860454698526,0.40605335327586334,0.11841083167446587,
 0.12069223569805665,0.4915934033894229,0.27334036274301354,
 0.3575944287176769,0.3494163492459772,0.012964666553505255,
 0.9746580184080544,0.20365280648436823,0.14381679303770578,
 0.9885566587422623,0.8794903399255153,0.7577215636566619,0.759947966511524,
 0.40241418359456615,0.394435498436163,0.8217992711778527,0.546870250986324,
 0.03239573125306516,0.20615716253061356,0.7030186974734542,
 0.3212729435091164,0.7861635681689432,0.20837017105864508,
 0.49867364001624837,0.39271278094583406,0.22299528180363026,
 0.17490442675264306,0.9661254578753485,0.06260285720830339,
 0.7917072985084241,0.13816910543527983,0.34520787815373133,
 0.16297573148675493,0.5474594589345811,0.16096628978169503,
 0.8394010823955033,0.6431658572214732,0.02806464209850601,0.7685610969504233,
 0.8386462145617181,0.4948304061708617,0.23870336722833263,0.2459286500920026,
 0.3049259624191344,0.25126091460411737,0.7368965530647695,
 0.21935196441630722,0.393145554946681,0.7461167791313968,0.42001291583473577,
 0.7084071512110277,0.47188373173766307,0.6364628090936173,0.4051049030449997,
 0.971864199885244,0.14577209698143667,0.9931973434880981,0.3909914253880157,
 0.7590058747983539,0.39684200232857936,0.9967904032855868,0.7205901675124268,
 0.4167272157477627,0.4533487492992331,0.8531440313641717,0.41981809534637526,
 0.6924307495613545,0.7529799431391807,0.5615603811805085,0.4338573315539289,
 0.26947340850106527,0.9842592880152069,0.9614547427081438,
 0.14882182747668637,0.5888871965522465,0.0012918164008539623,
 0.46496348039057056,0.7446303136586263,0.20148701288063098,
 0.37997903803119015,0.4724078164782355,0.7932598110043401,0.8056670144309236,
 0.42790068013946736,0.1222290344025273,0.9002115240729933,0.8590771542903782,
 0.054646777935492996,0.8763268279211766,0.37203446756529635,
 0.6121113564769051,0.08742366593705309,0.3222047746549407,0.9044700950990494,
 0.12246358141429392,0.4639790986242376,0.24645288993759673,
 0.3300075003898394,0.38048421520961206,0.04971297611170511,
 0.7495157412852131,0.14327403264072658,0.22651733515877637,
 0.6482561988843412,0.007786791567776397,0.3649574129098263,
 0.8138731618540329,0.35254722044352194,0.09875847521958991,
 0.9869769754112508,0.03863816786599683,0.8602535097090289,
 0.47138538938987207,0.4840496711549467,0.13425995421550585,
 0.38548982398816245,0.38572512539752735,0.5172325294838878,
 0.9608671062917558,0.555655363847162,0.8526199599782835,0.26317678459291194,
 0.2762373050629532,0.8571649428576714,0.709460329026641,0.46114474887253865,
 0.12754611614744404,0.7981680247450406,0.6892490136763785,0.1199669272899816,
 0.1870751918311716,0.3322311783531259,0.8410595623131403,0.14196810930289794,
 0.18490151961184764,0.6941738784227833,0.06963681218428686,
 0.5113328871059358,0.8774950817154034,0.9902566422165818,0.4653619533191262,
 0.5110099694653465,0.7059023973512589,0.36074862357039406,
 0.29131396745193194,0.93699805246553,0.6187283006881671,0.007737953820517229,
 0.6762238887642118,0.4930574275113815,0.23221801346615767,0.9942713002056864,
 0.947174218918702,0.06433713538773422,0.648147003837935,0.1083291827145555,
 0.15611018688324052,0.3564636209483596,0.08175383112908208,
 0.16890111238517402,0.3453735420688304,0.5396983307424287,0.7254487034332007,
 0.471293418579536,0.5636096841348316,0.04651351831208528,0.400105547027677,
 0.4244428788553525,0.375738827939323,0.3419796902042682,0.7012994579343867,
 0.4415640639543855,0.14214388726147598,0.24207476533760675,
 0.39605882845708607,0.09133197501280543,0.2386760342028066,0.159698801133206,
 0.5802931423593115,0.10658827330674703,0.08953373289783917,
 0.2763854409765971,0.08044833241112725,0.43112938232770004,
 0.38874050816780836,0.5544128576807758,0.9455715977296459,0.3702609887915389,
 0.9340589187895229,0.6544302871193977,0.02761192831350745,0.9498630021722171,
 0.8771288902527085,0.3960857921254102,0.35426086374427856,0.2527814698905505,
 0.4974580695725864,0.16392654825179542,0.7079243485629215,0.954322765013997,
 0.51291170976171,0.41420068326813264,0.08096945795283272,0.10443997202270494,
 0.8991696606885667,0.8283908374797571,0.2578602646114334,0.7448312776669969,
 0.9856989243218464,0.4899712747708307,0.31205303025602094,0.5464778377374131,
 0.5863844071932542,0.34888218014490824,0.7564298558699125,
 0.42766675773745144,0.16067566734239047,0.5493737953516322,
 0.6615905198943492,0.5998654038339666,0.23228908826218586,0.6839330818690417,
 0.8489319013939731,0.1211531851131411,0.012724446893948471,
 0.7679832752905226,0.5806233957219677,0.41074363747061715,0.9078823139373998,
 0.14828602123554424,0.19363510138116435,0.21710031697854715,
 0.7838828184112859,0.3233398224303504,0.2728258221405506,0.6965596415863156,
 0.8393022450318738,0.9913616785775604,0.5584848542697612,0.626257324786962,
 0.4206010113025941,0.13417764521448539,0.7334000133108066,
 0.03204966169939416,0.10081841994257823,0.7506760170010168,
 0.6189981443730157,0.35973307147973554,0.3546024272041166,0.3378641814836164,
 0.7992048139042192,0.5877896094229309,0.671987790457042,0.7195050075910283,
 0.8399516791989571,0.7284907824168684,0.18579005528987935,
 0.10560767674299942,0.8390882641815318,0.5795405880647972,0.2570494959923434,
 0.5463402497367226,0.02297142387265627,0.4339726683217269,
 0.17227264217684213,0.18354449777837034,0.21529821033665253,
 0.22641696046316706,0.859877226912886,0.22071789379932094,0.5081036383708797,
 0.48455923358172237,0.42208864743038066,0.8824064400510921,
 0.2206295967630567,0.8339769649763225,0.8406679338273924,0.285206223592091,
 0.33086879429077565,0.7228603209696354,0.9967532458374597,
 0.23083936401842287,0.30976219555897144,0.37693491816276337,
 0.504295303733353,0.5174866945543936,0.3721263529739057,0.8065700754586982,
 0.3088568323928521,0.9522755937388604,0.49517637068686426,0.5600968385838306,
 0.8222089697907151,0.19588295566377179,0.9605277099648095,
 0.46630137757212375,0.7843612972464633,0.2536166344166051,
 0.27158943441480066,0.5885346853624975,0.24641260707749835,
 0.7917706402990873,0.28170512392598035,0.6564864104567848,0.944026789818353,
 0.9717498439121505,0.7761503668246676,0.40325757300552,0.19018341467226674,
 0.21488656991716726,0.9258316297260486,0.0337732118054912,
 0.47968562537433446,0.4258301160054698,0.7726375915113377,
 0.15377182901742437,0.10475798187408314,0.5505544629864609,
 0.44861319871818384,0.6406677533799074,0.5471336587194293,
 0.09378000496240702,0.422076564575241,0.7741345212364635,0.18382709493240812,
 0.28443084474028146,0.8061791821265762,0.8004217396367579,0.8759685003815096,
 0.91920000189382,0.9761994986909298,0.5427193335796309,0.4383423830971489,
 0.5789208004413271,0.8127899214033243,0.7854300517170053,0.21295703032593538,
 0.34593892443246743,0.1902743883668958,0.5080064074432208,0.2091994159530185,
 0.2733393233306367,0.8690796315624284,0.13244899740423266,
 0.08068327127972996,0.8128315709144129,0.24585454908816495,
 0.8168500686455491,0.7895264549146904,0.12406446487038136,
 0.05353547223627553,0.5525126243544036,0.09100066657334271,
 0.7878952141108848,0.04007968084391411,0.8268175030141738,0.951024663703816,
 0.23742705183619228,0.54484521757815,0.36349678657194195,0.9231810461645549,
 0.7279722974887577,0.2508775528088385,0.5009450819015631,0.07957312442614484,
 0.5713838948286689,0.9096277776855856,0.05276759751693805,
 0.07107074162274851,0.8633136832696052,0.7786928658668061,0.1874020405573521,
 0.15367949803729253,0.012293844816661181,0.563056818144462,
 0.8224100315191665,0.7570796182539669,0.43577756011531743,
 0.17334487362924733,0.8148864599478781,0.5424060042168116,0.8648799468779504,
 0.8830530935836143,0.9724573276586838,0.9442277059012194,0.32312525281833127,
 0.023153732520648784,0.7571486239486702,0.6103793830743015,
 0.2440769651851108,0.3446355830120388,0.8030409725911984,
 0.018433905893902924,0.9643220380568212,0.12112946179955408,
 0.5780773791616041,0.2983251853483808,0.4069365098875215,0.7066158245371187,
 0.45242242674145516,0.46898436210132166,0.6202924211094721,
 0.014098524054969364,0.9262408877419607,0.42943881445289744,
 0.18243935607953166,0.9157071246467176,0.23394460039745368,
 0.6247241512474733,0.97685243946808,0.5702942244022725,0.9330630706049574,
 0.0477869217929332,0.41689855047994917,0.17574990748811214,
 0.5699079653673864,0.07890627561593178,0.28331461989498574,
 0.8514160901645592,0.6256379416724512,0.992865443577983,0.006028784167655177,
 0.33010756679040154,0.18179278834872215,0.3257664262362393,
 0.5420267720645161,0.10257119771653367,0.8490612651062392,0.7268893524719009,
 0.922555315350146,0.4334977140181314,0.3389072230547868,0.2731904535918874,
 0.1703454488895486,0.11831812007965392,0.05502434739651585,
 0.8457232181417825,0.7697959796358081,0.6572973787986182,0.1815271824130278,
 0.8329764635946715,0.1488396999053333,0.6733734598958669,0.8359824656350854,
 0.17071212707012395,0.7525182513390494,0.43122275610978933,
 0.19587868892946358,0.705313710683285,0.5811970909611159,0.8833469081469784,
 0.9020683352770231,0.24504765515964277,0.4655312590888605,0.6820405097977174,
 0.23914998421157074,0.6013954457675466,0.9603986978435538,0.8661886216400527,
 0.1860019508477111,0.6316553813863982,0.097269402447823,0.8261758401531989,
 0.25247601435015044,0.12213480902824458,0.8837999736592284,
 0.5624381018027036,0.9690390353341823,0.773925922851833,0.6790526480019169,
 0.23081365844148005,0.021769781508496733,0.07506730627244307,
 0.4079176169451211,0.19753737548532668,0.02736184776929418,
 0.7919178453622265,0.845626248931808,0.7139932228001973,0.3951439319518695,
 0.6168484694362643,0.06040641026474525,0.18943827922219614,
 0.2210029966473096,0.27765693723225304,0.05923650687031323,
 0.2793716149474502,0.6554465308549899,0.19645484844395777,0.9575629993654229,
 0.2470418141224986,0.5032324580124896,0.7676393985778405,0.9315683861342503,
 0.06565524882301377,0.9502468066147293,0.13574154208735156,
 0.6777242979004447,0.4234696835455122,0.8142172672869047,0.32183389181534994,
 0.7614867726048361,0.9799617012698055,0.08485230261039467,0.5855463512926719,
 0.2626071935343517,0.11316081734364947,0.2538041837233378,0.9715483625704915,
 0.854086834721866,0.8871557630441573,0.27836553884076687,0.3754455638021754,
 0.4630604327901323,0.06420686968773304,0.9053242717327246,
 0.07439632425886544,0.277912274314869,0.1399822509855313,0.3558366434264951,
 0.2341684244840817,0.4798161480383024,0.2127743326324808,0.22395960423322725,
 0.799413399592233,0.22468540282950134,0.1413086290676453,0.6366778153178303,
 0.4747428661132864,0.29659308531759854,0.201605175401806,0.45470664310740927,
 0.6468327885355423,0.4328247255352209,0.613993570122755,0.8776729758353614,
 0.6686999396261617,0.29710686770948747,0.6441696120253007,0.5431249663170159,
 0.94003516314466,0.9322656162125628,0.3673806253548151,0.0857294798950784,
 0.3667881959888194,0.8287862561955046,0.93512546499871,0.31064923193236393,
 0.26054075499334506,0.0535269826287772,0.5627185804791746,0.7661297715769193,
 0.5531543499742697,0.4705090255816031,0.6960262081960854,0.7547320992962141,
 0.27714029399161944,0.23862937590168154,0.8247447087005051,
 0.6449605530610023,0.40633532659148663,0.5385383118497266,0.1785445988472023,
 0.8160985267402122,0.6328725035104392,0.7151343308635864,0.7632135573570282,
 0.34910187893477096,0.8835328112043849,0.38974208548149325,
 0.44732704483525076,0.5626323098959807,0.3451707775940752,0.6694062544712508,
 0.5805888103910423,0.45396582916576045,0.07842014584341639,
 0.9630086306505766,0.48163132737687586,0.08674013667050329,
 0.1201889722378362,0.40030181366916584,0.21965581215079066,
 0.5319336863517867,0.253702346110804,0.5904883813606377,0.8650773446969362,
 0.44900888348159873,0.8359533859620939,0.3050548994509349,0.996499180300452,
 0.8293443043731155,0.3407846713454634,0.2350937628449723,
 0.028081249608682968,0.8841239498274931,0.5684368248419891,
 0.08757030849913139,0.677963248611744,0.35828028676000623,0.2156142329018932,
 0.3778502577064834,0.24380783458994704,0.6867048295965056,0.8163934566863156,
 0.2383416431775348,0.9316131309142411,0.3303617197642099,0.7194876972793915,
 0.1853540828500827,0.947691457495173,0.030619651220878952,0.920375920719244,
 0.6730573464076766,0.4043855074766849,0.06899061204126788,0.6959150269491697,
 0.9779031771217985,0.5993708775765071,0.10410268159970104,0.7131489489953663,
 0.27818996612478475,0.9844567387806521,0.1232397261577951,0.3498370398558295,
 0.40238845306005666,0.28517135108017033,0.3782533452071598,
 0.3241378750523024,0.055921983492069005,0.7747270576482038,
 0.4157236148616853,0.9557439218288257,0.5311629182334201,0.41367930485269433,
 0.49836988287720485,0.011701494854836736,0.14789196901333912,
 0.16556895281630346,0.4824021041644371,0.4940371776040865,0.3784833022207299,
 0.648695091987628,0.6034002210927029,0.961252947678411,0.3384615581075251,
 0.7980880718688742,0.5885253666526505,0.8669085921393935,0.08996904855414967,
 0.4374449504648892,0.3713164851392251,0.9907011488972763,0.4237692245957764,
 0.08322900575093617,0.08006559117134993,0.7907602230447376,
 0.7280879576319429,0.9165632831584581,0.6394251792110622,0.44315179115946624,
 0.1601884426107899,0.09416888852281924,0.8552417468434403,0.9412852646527595,
 0.1276224461064177,0.9396290636206744,0.6815069510996382,0.7626612256012403,
 0.7064858199547576,0.7867468588922318,0.08535751588098017,
 0.15772537176050616,0.24534698159400525,0.8902367576044382,
 0.4541186266209283,0.2519173063035046,0.32255773204858695,
 0.17851980849993676,0.3807243224733633,0.9507976460218588,0.3954545018171354,
 0.9728077580067888,0.14870276294149798,0.7010548276192585,0.8102076910297158,
 0.8810877194903722,0.8206039901715704,0.9689994947833804,0.5083360062529053,
 0.8109363123710955,0.8622713926626389,0.5876789019658696,0.13612272193780994,
 0.8285886204194816,0.04445583808839815,0.9982200616737634,
 0.15524084763081558,0.5820648683783074,0.8591465001604243,0.1829576104738957,
 0.6780116831679063,0.701411108995173,0.06705301821803067,0.3177133420325756,
 0.2441486928737956,0.11396027764474892,0.1603298310333,0.5141512434195985,
 0.44811314545083347,0.5692350753311134,0.6060780887852593,
 0.14183160500928815,0.4890310919736818,0.028263020459121768,
 0.3229827750353934,0.3760380387762716,0.1631503609337437,0.6618593725001682,
 0.7177371878479806,0.5065435068008944,0.45458539172139834,
 0.40913557297190994,0.18553867491461773,0.7013690232225516,
 0.5242344174098055,0.5447202699180537,0.609336596663771,0.3372487521529097,
 0.3834917250609302,0.5459617049050711,0.34578223332110547,
 0.49415724832217167,0.246220284182182,0.6373569133745477,0.19358155137319943,
 0.5021513188936086,0.930743054036537,0.3965961694132102,0.5789946158043687,
 0.3810194292964497,0.5122712718873997,0.7225766582275187,0.9957043061851276,
 0.15150329559526465,0.49684735376835887,0.44511626929961234,
 0.5046061970109943,0.32219564931806344,0.2898836262483585,0.1637614570766206,
 0.4343292339826862,0.4758008164343994,0.51960209105793,0.7594672126771508,
 0.03911628077574936,0.4276467876159272,0.36835138895956476,
 0.13511707497130088,0.4336730522793528,0.986813501686591,0.09371136021206328,
 0.1915180975655102,0.2816248280556275,0.39531038878694136,0.5302336193535159,
 0.6039396268604045,0.11222208681214274,0.19400816427426204,
 0.8319582890739778,0.18260217810281387,0.7463185932637001,0.3252607752681095,
 0.23192623767018172,0.37678583497010776,0.1359574465092266,
 0.5333662964024576,0.2457631398069583,0.8479047163697723,0.7898791746386503,
 0.9735214772503802,0.29463965450351015,0.8214936594832788,0.7111298732605451,
 0.30384934846546763,0.04327970528703651,0.3641357847283536,
 0.5505054898430274,0.5742294974789215,0.9141576606315407,0.3528302143129346,
 0.020141147974738316,0.3148706740417381,0.254515953542123,
 0.12120997098809294,0.6936800732027055,0.7897360400106704,0.5785548958393162,
 0.8404523291229202,0.8817666819949053,0.6874444197167076,0.3706919842929157,
 0.5595987840327405,0.5294848685380615,0.6977109187189884,0.8253367448185048,
 0.9278007034660365,0.11879285865838951,0.3753452865858796,0.2513172969590296,
 0.3646751462380393,0.2840468981843841,0.0579258910900351,0.7897532079587638,
 0.6099434376290989,0.9986276825136925,0.12623173218644335,0.4036703233415395,
 0.4934720764534465,0.16085729260666692,0.9994141408332804,
 0.03606250981493042,0.15862302760676017,0.5076803165301589,
 0.39652241108453257,0.6916308668879493,0.31969771808349867,
 0.6062757133310142,0.9219367999568235,0.2643284729679346,0.8059917596780245,
 0.2926364177117913,0.03957841991539757,0.27951985955792225,
 0.3852653703464135,0.983533987958623,0.7763073167229213,0.034022562847435656,
 0.30329437481130883,0.36805802829553824,0.11425064215715297,
 0.3921130017999578,0.7924849150052206,0.6479471316512553,0.7005465506585143,
 0.7390390335077213,0.23917311835023114,0.25760198968655734,
 0.5382174979013005,0.22625254461498767,0.8108336571106783,
 0.06622162570726176,0.761300728868818,0.19313594586332306,0.7572649301194707,
 0.8448593059804621,0.4012282788703352,0.42302911955169886,0.2502004398854689,
 0.44006175911003553,0.9605854571070932,0.06182412426139938,
 0.7243749928530171,0.6780402933370324,0.35057279188978185,
 0.49112315202860923,0.37566679954883087,0.7789616213201112,
 0.23007580437957342,0.2014169244537829,0.37177772614430593,
 0.9699310764729281,0.9693416598821044,0.3606094067883764,0.5409568179447389,
 0.756049552430745,0.4308569099421504,0.4283648594608682,0.875025277226344,
 0.6347579492228625,0.7782500885286723,0.4077654855189099,0.5976186739407058,
 0.14415415760769446,0.7839346724414913,0.8887041319874358,0.4783713852594902,
 0.24286278281818763,0.09823973850654444,0.9237436199235347,
 0.8581885081740327,0.3365295682242775,0.3664428538091723,0.46576894160645066,
 0.17856920609188243,0.34662665160460593,0.09498367368717586,
 0.09197557104139453,0.24105420793725174,0.7031373928146543,
 0.02152688285005533,0.14025833630584517,0.5651875200323111,
 0.8693983602117968,0.9274285856320734,0.47091365678028707,
 0.09744387966520296,0.4310896199161971,0.8267029087423956,0.2243629186418612,
 0.2120687091818786,0.26629808953281875,0.5961735351685318,0.8528018507281901,
 0.9135413749503507,0.6600451530355385,0.3379440763178444,0.3737578280363604,
 0.7119713724950805,0.0752374566805184,0.5139695044760312,0.22628462617563616,
 0.062139156927179084,0.2568544115079727,0.6990943245874002,
 0.9525582851322748,0.5639617649881121,0.25806182715020154,
 0.12114793095491128,0.6642488784237335,0.9311778002099731,0.2579190639983918,
 0.9656410348836814,0.7213307594604459,0.6438728570037006,0.6148115385035751,
 0.23664668389918975,0.12834169740669,0.42048647245285375,0.23973882014270498,
 0.06814122978737036,0.9685116921876785,0.9998478708843128,0.8859486796080352,
 0.15133402301379073,0.296720545602727,0.15000567219369,0.1742359266644804,
 0.8421131157285838,0.823879494997579,0.996868866873865,0.42135082823709646,
 0.6306757252244862,0.5634615388668324,0.03530208547650093,0.8629272032824264,
 0.02251599153986461,0.9562068194437651,0.03269241478649032,
 0.5514331095000179,0.6498807412380994,0.3472061662987277,0.7767773190960934,
 0.6334748699140642,0.4666980276497745,0.12196014300588964,
 0.08389821750664661,0.04307472327629458,0.9676000112238464,
 0.8266133598430181,0.81784433325823,0.07641139359776272,0.19302650971339264,
 0.024730769615792014,0.5232794076134735,0.3877676339041143,
 0.034477603344158836,0.64820733546092,0.6305078357742688,0.971944442874932,
 0.8822982498182063,0.6211785082290605,0.896860350527562,0.9935011893195816,
 0.24724272075494935,0.813779958137235,0.7724191246064578,0.43933242910357306,
 0.5016183091575878,0.694534916075954,0.2987586170214911,0.47353961130670064,
 0.5607448065256528,0.475659409347145,0.5558139534566992,0.8085857201074469,
 0.16872386961188401,0.7412925910154675,0.9201964311748195,0.4706094558355798,
 0.7739222780015786,0.850940737798748,0.696685663079891,0.9074622530085696,
 0.9484557004812729,0.30600393473028453,0.4759653279839402,
 0.11227853887036121,0.1223344749226386,0.22582391136591806,
 0.39983181903013176,0.5505659850137481,0.4116844980769936,0.9775297144374692,
 0.9272431163692268,0.8765363925273295,0.16975791199949763,0.7535094215847885,
 0.6551185571581496,0.14080809937316152,0.9969258494536807,0.451110254319115,
 0.5557937083902471,0.8625284812886342,0.8965889222324506,0.156689121362922,
 0.5078324761275075,0.012228889557863498,0.5631004024768073,0.808174302649824,
 0.017513378416598613,0.4588320632441797,0.9526083463087028,
 0.07990876960330184,0.512298414863783,0.38599964153096966,0.7887051979338316,
 0.1616879572938702,0.6666782000142788,0.6396014261021995,0.32888239838491784,
 0.8611569555754874,0.2563229188023156,0.4609838303956557,0.46264124983132515,
 0.5419401045868276,0.30876882607359957,0.9555195821209772,0.9019075425574226,
 0.15020175664807411,0.645450421652406,0.1427468556464757,0.8691812001553897,
 0.6042259347790853,0.8749547692849862,0.34262498575372524,0.9326868524411691,
 0.1186320973973547,0.991008989711088,0.49497377898375794,0.452507014782232,
 0.8683320843843131,0.7082614776599613,0.9595157121906424,0.6357568861247834,
 0.8213390345722192,0.06428086467315841,0.30010258465460093,
 0.17469665555800795,0.4100227995425978,0.5366396482954962,0.4374292461830641,
 0.21485103171250752,0.970391844783941,0.1445901043120743,0.8395197488206511,
 0.2605030546620837,0.7911080847272771,0.25455715936251977,
 0.06046795933417348,0.6807591399723222,0.7966874902797584,0.364133830264032,
 0.63368649999215,0.3603485717888255,0.19004585562131027,0.9097018060906306,
 0.13531788875116202,0.9041353943969438,0.3835915076509311,0.1062881010507174,
 0.6569250213001624,0.8959058834512519,0.8253393276327438,0.3347502606547036,
 0.24474082618848791,0.7093922288007257,0.13290317032093124,
 0.18552107325772493,0.5063699520470475,0.7026909934157294,0.6874017855894735,
 0.7737399928278118,0.9712607154084862,0.1614329867950004,0.3369686583556105,
 0.7690714295966901,0.47124580590628806,0.48648582467212476,
 0.4805276987841509,0.2774969078346806,0.39987733647988244,0.5005872638939405,
 0.1689487206875916,0.3302295371007893,0.2486667034675708,0.9320556003534959,
 0.11684232897906899,0.07840781036530664,0.9930429479202402,
 0.5622862877107543,0.9199501848275614,0.36413091325104363,0.6038052507702254,
 0.5899144537454597,0.7062839657001816,0.28269935753985187,0.5345532937955672,
 0.8934573418225571,0.24063089152512573,0.34446362804452935,0.585336146372366,
 0.24418667901632807,0.07489725977341932,0.5938304781873399,0.927870681356892,
 0.965089089891635,0.30767791711047177,0.679487363456183,0.9695454568122162,
 0.8614442531343004,0.9889942612876349,0.7207559025818967,0.4991102621110315,
 0.24165403245397532,0.3951188186769832,0.6950048487547108,0.7445651095867647,
 0.1545412004802733,0.5219803958326062,0.8448535144684926,0.32546561744988356,
 0.6726994701135637,0.44703467850128653,0.22879792323264148,0.250730563732966,
 0.8699965665129789,0.5825946432720076,0.25919059497588426,0.7599004775751593,
 0.6784389528501855,0.629659247931412,0.6312354205693866,0.2227215324682803,
 0.49371071542464473,0.5246984646742856,0.11665329933968238,
 0.37693792985927643,0.1636001121124745,0.8675081084556837,0.7497196346860607,
 0.07716610404092095,0.4589161028470241,0.61272211612712,0.11412075848294212,
 0.7867539815747128,0.4800318082478622,0.04531279572149871,0.6920051655913045,
 0.13424063615579296,0.9178163616887953,0.506657692954979,0.3031248001290572,
 0.86730804080194,0.2977950951723284,0.8993854985583807,0.6247651308613357,
 0.9518424485752275,0.8307546476333946,0.36006080336223145,0.760547672083944,
 0.814269100427225,0.8514901284129269,0.30465381129235547,0.3490663406342863,
 0.6710408055542946,0.5406788698802077,0.992114046921527,0.12601672575200473,
 0.07572168630147647,0.6240145056505014,0.5339781990307414,0.9221795565466226,
 0.4375949271308448,0.8294001172361998,0.88880402329961,0.3956777685962277,
 0.273212260029135,0.4902128504354264,0.9302096372748272,0.2504632518788952,
 0.342717660687748,0.5515807788019171,0.4463131330628409,0.7204121826904952,
 0.7681544466133553,0.9462615051718041,0.5569506249478109,0.8922936962087006,
 0.019181014400081575,0.9335340586029662,0.02555407676852306,
 0.12100744970931032,0.6374401369412922,0.9043209738293468,0.9982375308467811,
 0.8210745414299349,0.28644179737475717,0.09156561588984857,
 0.061957722447564145,0.10150939978450146,0.7063804471263618,
 0.5463778108184959,0.3496940873789063,0.9402693527676054,0.12864184930796896,
 0.6353737607843017,0.16730381275618966,0.7596389354009858,0.2255514646244663,
 0.8292030487234953,0.8940096680230932,0.5426874051893912,0.6582227990951368,
 0.2105002337643176,0.8350008553765164,0.4182824626055399,0.4062655016167611,
 0.25137984199536545,0.9613264919669272,0.7616329418896389,
 0.12850542056977587,0.15054689084726758,0.010100623360601912,
 0.11365364377650644,0.2727005142092902,0.3520146131202292,0.6723504294133316,
 0.3483685122897716,0.8369507332672439,0.5296330935549521,0.8297262588818747,
 0.19900586310991875,0.3441230084000859,0.21431785953351368,0.350682899071761,
 0.5341991780036475,0.9190531299858783,0.18592515342209515,0.75119377873507,
 0.7110449763827904,0.10096394748424153,0.05009360748586689,0.939647240794901,
 0.3627155778057578,0.05074956550579579,0.3348139115020985,
 0.0016421377689302963,0.34749241223703276,0.3168873749525103,
 0.8254917882410368,0.39296983237324834,0.6136819755669127,
 0.03381389391712364,0.89446841627125,0.2718051953304741,0.8152217314874124,
 0.4042489669120828,0.22340733118412348,0.07562809947402149,
 0.26554433244439024,0.6979874886741981,0.05769146174798945,0.609183706798307,
 0.6026101886802522,0.31992831876550143,0.4120309701542004,0.9543509369363578,
 0.2627653763447684,0.26034004474329275,0.44874046353495833,0.836240485750505,
 0.05075802320221234,0.5009901659404097,0.4669121460756057,0.8577639052764249,
 0.7742857817805912,0.24254594223494885,0.7001721711312391,
 0.13780436870250867,0.6859594906416845,0.45911512607090654,
 0.2543700354664121,0.6718787296209562,0.8573421717963099,0.9818903984786089,
 0.515097919563426,0.5494102382720698,0.20127239333077318,0.05403464651985679,
 0.35382507013383413,0.1636706621162104,0.029931080885746808,
 0.8500750276520602,0.037963215118246074,0.7418834232172768,
 0.9559889483637487,0.18793171714115786,0.7040149945022024,0.9554062153195776,
 0.6842001882055154,0.6726948183030514,0.9408640029028355,0.30841055460586264,
 0.7841793882615449,0.5447748846967867,0.5342194427981252,0.41304339417999936,
 0.2039838469435432,0.7672857590570997,0.36755894019189606,0.6743303728102825,
 0.5747982368564485,0.5212556147407426,0.767428775153743,0.13900612077833396,
 0.013117640056830027,0.3303894215416481,0.8554329677023454,
 0.9342324834683967,0.9253915996253526,0.8479091755704286,0.33110776112711693,
 0.017564481162849366,0.5344283255084532,0.07921998213685033,
 0.814963587328813,0.17789272049380433,0.9191195196589371,0.08884123281746048,
 0.8897459634276687,0.7397621777187949,0.5281255591604381,0.882619818871945,
 0.5157285791246117,0.5093488307043338,0.7028367189264235,0.8828167720247696,
 0.8259320882319064,0.68735059364835,0.03479707770763674,0.17286494969735677,
 0.1990633959264395,0.22218479405295455,0.6766341522940538,
 0.35052092177293426,0.527514112939947,0.20710803101541408,0.5178071590440192,
 0.14398568220986052,0.42927424974068884,0.3586853751906779,
 0.46648326037694454,0.5345434661492909,0.44663211151098414,0.500607863746705,
 0.13956861298105616,0.9568284947370037,0.09143412828962427,
 0.0018000012105613417,0.4063362912222217,0.0334626085552423,
 0.5014673399398492,0.8338814983654957,0.9227613633331412,0.16234358324288134,
 0.6771624686593867,0.15279369893170902,0.6923869393693504,
 0.11861074444515007,0.07638028446749068,0.9827090900742591,
 0.35254431242787354,0.7503579209715528,0.055772321665917834,
 0.02703379075584489,0.21458957406597223,0.5565063104119986,
 0.5104232424870514,0.546355962510781,0.3681500319339652,0.9061605718916288,
 0.7459311273804418,0.13366765489756427,0.7847150053123969,0.8363009667400836,
 0.17138564429410408,0.3572029590560314,0.05603853589075847,
 0.5746444910011541,0.19356437503812995,0.5272349281745278,
 0.48564316584826894,0.4588099857699699,0.8794453490238077,
 0.03162699255382673,0.05925138960092391,0.9158828832429675,
 0.3222836693507358,0.13760400337939538,0.9427000151714112,0.7443160206166898,
 0.7087808136290795,0.46809360590673255,0.36340622986861704,
 0.2668275947741475,0.06990642103023337,0.7730102700542523,0.4143075428337136,
 0.6648183668124406,0.408019453371283,0.590003791181704,0.32512156415518634,
 0.2697940418001198,0.46410303986211876,0.39148388463978745,
 0.18608818690552198,0.47369041730664474,0.3535114947352831,
 0.4029212435681254,0.934968335574902,0.29000766203715866,0.7290198311935695,
 0.2888225223154377,0.2672083368043344,0.12560957610840284,
 0.022101972462946762,0.4008839697037935,0.9854121910397576,
 0.8600515623934769,0.9963440001005793,0.8582513478708511,0.10068992736281435,
 0.5643478854633933,0.9918340285952336,0.9229151543398108,0.6226861818624545,
 0.6089129458046926,0.4376082082231252,0.09048798044508244,
 0.05982099693883913,0.7523613288353465,0.6431688342720465,0.1682785380400993,
 0.31665301282695113,0.12174996582210917,0.2646745217099087,
 0.7301443422502155,0.09449240947527371,0.8327754482079803,0.8085665516346278,
 0.36785983696038327,0.6333836974873209,0.9185312397062206,0.3253769204932846,
 0.9639408840153434,0.5138706596351061,0.7105452030686167,0.6568279614270759,
 0.6968660091453268,0.9785762320574977,0.8194937612047527,0.8111245242909709,
 0.8307602304693529,0.14385634905962275,0.9916630002041629,0.0698541783294111,
 0.26421106152350926,0.2682839838608905,0.04112017674239743,
 0.8911779130857566,0.6874523340090417,0.5454133525837508,0.42465781024334837,
 0.9778682979059613,0.8182857841983292,0.4860019323141628,
 0.026551688225822212,0.2953043791060892,0.19214311664522454,
 0.30959657231304205,0.5607013504575227,0.6272994489335987,0.8719117693597807,
 0.7038104622254164,0.9091983753258431,0.28662895813963674,0.2944621013869535,
 0.39998955099085953,0.581094729762249,0.386063653497126,0.9699100213500536,
 0.9058582945200641,0.32375411926536546,0.4665794566888265,0.2116691175690153,
 0.6949731863153178,0.9357589959724332,0.7180523114899882,
 0.002543514961411475,0.8983879875167926,0.11869658535013594,
 0.9864675604553476,0.10180511559892813,0.2741879156510767,0.6725608000493146,
 0.6747126119923711,0.4655226484458226,0.5270636776235266,0.9351927904524984,
 0.6882158142701125,0.28148562157786994,0.6006851810224478,
 0.12165935738255051,0.6187229550193213,0.7190682171196736,0.6004351346509156,
 0.23731181612188024,0.8379090546036811,0.29771974578536575,
 0.9202512049167074,0.9420921118060365,0.04147800626380782,0.9682646168920488,
 0.8711814107107511,0.31929676618579517,0.3889082110655482,0.1653291358995561,
 0.6292186748347004,0.07126817549400855,0.5584727441853979,
 0.015001658335964896,0.4123319052856341,0.5853986780434732,
 0.9134544656640712,0.4958665313730559,0.551101355102841,0.5470875709623774,
 0.7041593068592371,0.9494471924489389,0.3034191920839593,0.5554758523581019,
 0.3734967024326512,0.08141125159861018,0.6165112350359242,0.555305197679545,
 0.6352298726882042,0.21195659208399353,0.05620939854098539,
 0.21358555441788674,0.21084073964909933,0.43025791989945006,
 0.8449878676969278,0.7853886764944814,0.6659482442301161,0.6332928043272875,
 0.8818694891844567,0.9648210547408307,0.0665266387752611,0.799340375185425,
 0.885138329439908,0.46019404055784474,0.6518100156313461,0.3797071169185754,
 0.008291173213441305,0.41934270116867634,0.5528456196110542,
 0.49481890211102364,0.5945040281226154,0.08500140625120911,
 0.5653356848293316,0.1734957526448383,0.37827227914006867,0.7622696007334115,
 0.5335054299106579,0.08782564026371875,0.4539032885871741,
 0.011520960717324336,0.41030228020995285,0.33460499260210197,
 0.4219957845959489,0.002277202170050341,0.3238845513259394,
 0.9125503185177909,0.7528497346255985,0.5879157369006248,0.2849973510427597,
 0.6420597460373397,0.1744734510181044,0.8621310405527627,0.310861053438959,
 0.27198732966234374,0.8798245672051117,0.8702919878819334,0.7758572758110942,
 0.5215846276645497,0.05348838860718874,0.49692463011114196,
 0.6532223752976773,0.10114470898634165,0.9977896567443914,0.2841720741800693,
 0.20920521753616828,0.9824930095129045,0.7702660722791915,0.923059203129862,
 0.6019195606908889,0.8479365988729691,0.904431179140793,0.45308621891683276,
 0.5344209917304077,0.812214468703562,0.4861864284258809,0.5508512091854841,
 0.685308649627483,0.19284728767461568,0.8155572981198521,0.6921705823779315,
 0.40168196747471396,0.7067040602894629,0.058875989482039204,
 0.5368035998134064,0.2953460755104711,0.6089231399709557,0.8798627043613113,
 0.9320605987284818,0.7152903075048236,0.4132743529856666,
 1.7034806862770147e-4,0.30781546255407655,0.9679753902842189,
 0.04355787081812312,0.7072361694986584,0.7920866048380069,0.538571647130733,
 0.1048501480002978,0.7781072704250669,0.46363637236742794,
 0.48643125687903455,0.9682883284474099,0.9939742579928674,
 0.13762854102512612,0.7129329573734464,0.4014812395334202,
 0.21908677315891256,0.3432872310227617,0.4754848111178478,0.3971690994905682,
 0.4859840407251278,0.6171750667203068,0.7234497589947563,0.20929056150967984,
 0.8075439563204395,0.6957094521936479,0.9579338527477402,0.6815943702041145,
 0.7070391011453274,0.4967759662203868,0.7384499437129401,0.9687463528453006,
 0.23174641578825772,0.07490033635712812,0.35535814697519874,
 0.6084260636347527,0.33936685828590274,0.5691511751077785,
 0.001099884876232804,0.08679053672404191,0.045785744263785566,
 0.4047309205234245,0.5201946003107898,0.7168036449811077,0.15225452728217825,
 0.580677225776182,0.3817288487877486,0.30888874363479224,0.4939415531268352,
 0.22175745800508628,0.7254297555568742,0.20729131256449762,
 0.5779960590696419,0.5024547111036363,0.7301598161572352,0.642559158231139,
 0.46347924208664637,0.6789067752560975,0.9897615442798338,0.9466929347256386,
 0.7554334960579164,0.4622759386045976,0.05466672327268918,0.7701314898205557,
 0.7436461191422771,0.5024286837055448,0.8617486909354448,0.711101772789287,
 0.8775121307397683,0.7383483824517924,0.9877331450341634,0.91046165336839,
 0.0911397493048004,0.062402766196203,0.0537233723091356,0.22380995218081112,
 0.033769076618045624,0.4874111003921846,0.9951475867904338,
 0.8972477851140482,0.07927363554036382,0.5969189867545427,0.8659695753732858,
 0.8959561760038396,0.19649192701497398,0.38010926988313454,
 0.5157634279979222,0.11631293713800372,0.15737222265433604,
 0.33546321165144444,0.9797719758829182,0.584124665956773,0.867172932060602,
 0.6257623901576421,0.7580837855469497,0.5177969260114944,0.42454182770013804,
 0.72373016557257,0.2951006212942082,0.6702813023924757,0.4824460697202724,
 0.6170023908333384,0.31189660057696256,0.396211548320595,0.4310996181713589,
 0.3306503602362527,0.04267303525317412,0.6660248521459742,0.902339994701204,
 0.37153259169765407,0.5812439999351718,0.7588335696854579,
 0.42803828060322036,0.01990331585120586,0.16817902124158213,
 0.03860488004539986,0.43285170770451065,0.00415926067248884,
 0.5863310394538048,0.9243574229205542,0.6309610720966142,0.7437572996728865,
 0.001654675393737759,0.31174547387171014,0.2172135533217373,
 0.7129841132900827,0.649738941011884,0.4797011036959711,0.9435357398866319,
 0.7171173872214098,0.19385143653239856,0.8608711917600431,0.272473930257643,
 0.25735253598757724,0.8810115936931738,0.12271961973783518,
 0.8729090376239302,0.3566945318067618,0.0024986805213893515,
 0.38466077483934225,0.6906607384995135,0.6888911330752647,0.9765434957649356,
 0.19910301265836783,0.5834131794183807,0.18083876560284495,
 0.08296381808530784,0.9320547489133527,0.28195173337651214,
 0.5504206363416042,0.4268981241038119,0.23881957809094057,0.5209435158500506,
 0.07138957392920164,0.6527088234827955,0.8599124707464603,0.8756699075833767,
 0.6481147729831519,0.39016055438143116,0.5632007498836069,
 0.05224265573204123,0.9760119431790586,0.7982139913583173,0.6098992668131449,
 0.8689804310006931,0.02731636331414644,0.27683503626457273,
 0.4524385677836652,0.4109371174198627,0.7420133838479037,0.3501441289920175,
 0.23789389798430183,0.9832962960279918,0.610324823503343,0.8246436400391141,
 0.8354317951407957,0.834625129448063,0.009981167788024492,0.61411243480603,
 0.40658255614881855,0.19875599896818863,0.09719937865215866,
 0.32540024001820966,0.4919347603136891,0.9967353141320592,0.9855442016429916,
 0.22867035245373357,0.5215360658984939,0.48119592010160495,
 0.8997830006116228,0.586890311238408,0.8417086547434594,0.9172616458758848,
 0.7609292849730047,0.2194845285618342,0.24290888201445315,0.7297184421840726,
 0.9761551927906074,0.5123053278096255,0.8397391535431735,0.11311718200514509,
 0.4011019697890955,0.3288336841113313,0.859804761914099,0.6771343888383046,
 0.11818717710064819,0.5744658443175575,0.9323151332713442,0.7354788364608144,
 0.7503904536963393,0.9243308656148116,0.26165295269133704,0.3932141274803378,
 0.8088347921926755,0.16130788905284765,0.09341065850293617,
 0.5100376781406875,0.7594919987562514,0.5725912289292154,0.48444481357185976,
 0.8229191701218236,0.6032407480856957,0.47308751794245407,
 0.07310256656038583,0.8773758702100135,0.5706362936735858,
 0.12134421178604893,0.8349143026007998,0.9792763152038825,0.93970883764986,
 0.5401952180846699,0.557620607317244,0.3928172757673032,0.06229003272150624,
 0.7679233787972589,0.9099794693340012,0.3127118624121926,0.4688459158199645,
 0.5841884987562466,0.47559897260381234,0.7505545542260699,0.900814888392743,
 0.14470622791208332,0.18394224690190275,0.5593297380441373,
 0.44713462626129163,0.2518670908070737,0.5525745254810249,0.8176833509689897,
 0.7754909155440985,0.6013796700830878,0.9938756358936938,0.8439631295973403,
 0.9793464296612046,0.4119536662742157,0.8719431967824608,0.9138572913896472,
 0.9713138700250841,0.9645941352247414,0.5598397005656228,0.9589597787938158,
 0.8918786401322373,0.6005442741704092,0.05969386365808593,
 0.12050815591485953,0.7773557650025309,0.9851432381700022,0.6578936850233018,
 0.10971642316378061,0.5138855083992513,0.356757225571101,0.29949421170012736,
 0.8693521800336518,0.1286757158623747,0.07502431102834206,0.9501461597232786,
 0.5536271438008327,0.10861648496177045,0.9608244333648517,
 0.14472075897102443,0.6727012207421176,0.6905100315321346,0.7150205423352696,
 0.5503773129076291,0.4861683712095146,0.25996883706925966,0.7828326691768076,
 0.7778816026640524,0.673060509033085,0.783840144843825,0.4987549724221344,
 0.785901562639667,0.25906157522918916,0.10689867134307451,0.4669649438355863,
 0.2465913589197225,0.810418932127856,0.4021057019915226,0.9442364524997762,
 0.475740896275625,0.8270214272565933,0.1344244584707175,0.6029560449253477,
 0.8707669546099471,0.2245370849163193,0.6716316588973159,0.4870516991585435,
 0.4484703411669203,0.5886915324294919,0.3656343571348901,0.22680846128162124,
 0.6015997972769984,0.9511979824273834,0.6438749129625232,0.5042449284082836,
 0.8384384211797524,0.342741672474958,0.25574834370007815,0.09870792973944853,
 0.2338849485502077,0.640924858822641,0.013413113666141396,0.2691268993744278,
 0.8037080078612714,0.3359515369988312,0.1399793904613421,0.6787908814226074,
 0.26179279484916873,0.41209632866600676,0.7095022557383223,
 0.02414697023279133,0.053108489946881,0.2031796654111785,0.11225536528182245,
 0.6228248422902439,0.6527644232381156,0.16062323972728676,
 0.027064891800244406,0.29873401626792195,0.4899350663193701,
 0.6749364293878021,0.0523091615913025,0.4486308239743929,0.0703575918541337,
 0.8412185489237463,0.27730315056401833,0.16582809465675785,
 0.8441312390236915,0.05117593147398236,0.4218957689261239,
 0.35693278572587595,0.44093588790222815,0.1670119184160832,
 0.2353754802222061,0.8069072490223385,0.19982396033169947,0.6343719313557024,
 0.45902975826488157,0.6962645729642106,0.9214905820282541,
 0.20163687690681542,0.33487716088572017,0.7886291732903932,
 0.4802291389133214,0.7603935908968812,0.22869822519551586,0.1887413646342987,
 0.22952844223877988,0.18308571615327685,0.4360605647727772,
 0.5839141929783995,0.6283725407177678,0.6600117461586607,0.14475969886803353,
 0.5617089065863056,0.3686064264940997,0.4762359361531957,0.1986689844984283,
 0.2798654591183166,0.5720414407655973,0.4910656702391223,
 0.0023724881216060645,0.8393528141243862,0.9330047590260189,
 0.31998414298537037,0.939612151645385,0.8544884253877705,0.8114214872643813,
 0.10960833653846824,0.932198086184238,0.951719456496996,0.20060648324377606,
 0.6532152459622301,0.01958087792982266,0.6385915065497829,
 0.46733909597335543,0.15023110163564746,0.08895484597956349,
 0.17254650631983026,0.3745458445424552,0.641576624429787,0.11790568210024832,
 0.21469130233013145,0.21215396077923843,0.6440680443456417,
 0.8578576268137365,0.6563488485316381,0.0030093260511748765,
 0.5012021715208149,0.05124185060393671,0.16565983593140665,
 0.7141796206372952,0.7956887619474694,0.407606377143342,0.9706772539331726,
 0.15335755556852249,0.6153772826345496,0.5428293346865499,0.3717931404713287,
 0.46660572076049467,0.7842310325787478,0.1823331093258309,0.6187140742338416,
 0.650946706921959,0.8655181023647597,0.08691535961495789,0.5394872089522856,
 0.015331706884761487,0.14501973214814945,0.16080332875379577,
 0.9160954458727325,0.6598523371474913,0.634763082378808,0.03372743347875651,
 0.35817055594053604,0.7256604454731012,0.8947698961563019,0.7763029571054886,
 0.5043862658179701,0.12447344963490692,0.025559662347295342,
 0.6460961517053335,0.3600996847632506,0.826660417279454,0.10230600520218502,
 0.9213105377872056,0.6144831505651518,0.4570897114369483,0.9843711263367557,
 0.7707702561446507,0.7254342615270737,0.11336673418418519,
 0.36529739703440034,0.8357936655468945,0.8072408816086787,0.4126937033706325,
 0.18590601087622793,0.3244688844079272,0.3260003872691887,
 0.24466866929505182,0.6751671172308898,0.7862701288168018,
 0.16220196396224107,0.7378403769519613,0.27986195238504263,
 0.6067843500958179,0.8058162586580301,0.5176282822949718,0.3967693030507883,
 0.06299135685579393,0.3152186229862063,0.9139880435826218,
 0.05349434160618882,0.8023761527262961,0.8407263034997914,
 0.37522956032209753,0.2767630009411425,0.15108378982654946,
 0.49645763634104223,0.22108084002142214,0.5455219025679466,
 0.0893311139688201,0.3093955438340117,0.6790759077555185,0.18127625470362718,
 0.5763704408799676,0.3397804569073303,0.6746374195746795,0.703377753112736,
 0.5806158589619913,0.49655104182088194,0.2842150215683483,0.9164578346299487,
 0.7742640713466661,0.8724327904398184,0.13907747848760832,
 0.31890277509700915,0.300695010055981,0.9079350755746453,0.2789234045388459,
 0.31468523600177944,0.6221251343708368,0.3240041178588585,0.2647965933978772,
 0.4035157160790269,0.5985460732344797,0.06660432794423787,0.8830080503308385,
 0.5113496190103554,0.266603480981056,0.33500849996547233,
 0.032263908686566145,0.7553278929963017,0.658258148651376,0.4164488494922315,
 0.7636620289192786,0.2619557120495838,0.2795802469511576,0.7104155030494306,
 0.7163074625275488,0.16499290609429107,0.5378697437393556,0.2713806900401149,
 0.046211598937374476,0.35795402374075624,0.41181048895485384,
 0.11308785251241482,0.8708101962892323,0.4493332717433455,0.4528568426755376,
 6.571374937495156e-4,0.41029730947536536,0.7738134773143296,
 0.4253967876955439,0.9366416329309946,0.13556136674241825,0.8515114403218228,
 0.007350579896727605,0.19240284586720524,0.5175813485425302,
 0.8984589384852955,0.6890075223398903,0.16734442919998904,0.4109359010081186,
 0.8658982621596029,0.36370970874008257,0.2791135771789346,0.4308019107927703,
 0.3984115935673114,0.6805959944190545,0.9097029355310655,0.7939422046693145,
 0.3661976204680659,0.02141490379887423,0.4827745546347304,0.3591556208360652,
 0.515125526051654,0.2643993422121991,0.40223820562894175,0.16236538513585252,
 0.22605664639807277,0.45068137831605937,0.033589717226286786,
 0.739255501850266,0.5758481280151746,0.5769860228814023,0.016502714184837286,
 0.500810485507888,0.7087897340444756,0.9613526370937058,0.6966473583352959,
 0.623251360257165,0.39906290111850307,0.03972564692785652,0.38187207163279,
 0.6071729547961483,0.9472198727958049,0.5928919632231224,0.5420564042446199,
 0.11606772259712239,0.6095368782877358,0.7724410665410764,0.6410127754907569,
 0.6431330794133504,0.7735701869314671,0.3928711966512205,0.17448757295800443,
 0.29729951536550814,0.2740386484229934,0.6458500615880169,
 0.33692577165803594,0.5116078278213568,0.7051558570779906,0.6587309908243649,
 0.28002433658324843,0.6639013793211669,0.06927553506132733,
 0.5369523445850357,0.18146821130151014,0.5723674536738925,
 0.37219481342997807,0.26147402455597013,0.08610339103486242,
 0.15310173488144962,0.4546412526247495,0.7078822298721706,
 0.17086718691326186,0.6383418245137285,0.6728005333621419,0.8344159767357582,
 0.19811735283071208,0.05993035196564678,0.544802575720686,0.7764263617135363,
 0.5369392052254853,0.9878666185823344,0.41190332191448675,
 0.39497467753377213,0.1495861175432902,0.6771723385023034,0.7138749174717554,
 0.2897526794104879,0.25503124422217016,0.4276907063582819,0.5801266194335415,
 0.28437262947453434,0.3367336092312263,0.2835614121174074,0.7713713625250743,
 0.5625251057667298,0.1603501801708349,0.8761998557288068,0.6287875252890813,
 0.9496376972524062,0.973648446736945,0.6886162363287566,0.4040605422029394,
 0.3919993330208712,0.56959537056488,0.3414059150748987,0.9838995840932085,
 0.46688004405881,0.08427337812106983,0.6903800852970536,0.8717502345107415,
 0.3214162873905203,0.3580833527695316,0.41931538346980657,0.5689585609108069,
 0.8441296920051368,0.1962383952659027,0.020982161611887484,
 0.6960997674844605,0.23042547429444016,0.0100311300568523,
 0.09271452478038833,0.883603609503128,0.6641427582545658,0.06038706937701721,
 0.46447070555019887,0.38164790921634595,0.48531208526692793,
 0.42471703936432526,0.9703742765870026,0.24803160819882875,
 0.4860070143001981,0.9361684695288865,0.504189994107934,0.7383527769166056,
 0.5298841937796794,0.06486248639008829,0.4247992540226585,
 0.49928206128820796,0.8411031980427864,0.8969832224037104,0.7911260864746246,
 0.8581365023719131,0.08223041915437035,0.4681454487436719,
 0.37313178299615835,0.15567933073808204,0.8176495708338432,
 0.7611783921144137,0.28046880075731306,0.05363516424184689,
 0.02936004873138487,0.24821643575491492,0.5437895996821757,
 0.18052986024863027,0.6646482054330243,0.4879164636058275,0.5846423283473643,
 0.22065511394266202,0.31132398506879144,0.9542163238550304,
 0.45198976832353543,0.4730018713747782,0.3554990257189682,
 0.39576834657728366,0.3067577635656693,0.5744009689652982,0.3731457044630142,
 0.2450662412442487,0.9620129665679553,0.848009496477574,0.10651423712466057,
 0.1120232722994956,0.35262974466427544,0.4368916958260125,0.5281028156768104,
 0.8966020010149857,0.07580450250480464,0.6252748236574454,0.9628949211750208,
 0.39579211147080773,0.5628797242186111,0.0018163745001857912,
 0.7169511327030251,0.6454100003120109,0.9165239279579294,0.2069369111935334,
 0.42365372788630473,0.3034078418612891,0.9934908502239663,
 0.01935039720337539,0.06473673638312605,0.8972645322898535,
 0.5980469068195984,0.9604022502525852,0.8626938605330933,0.3807912134674488,
 0.10905162298258464,0.41681523990564207,0.5652935476917406,
 0.34234244166157957,0.6721724185849068,0.7672987952521632,0.6093652081430967,
 0.20063538511057655,0.0010759094142402947,0.6829664136467357,
 0.19765405276305303,0.10450576024369707,0.06857725992738162,
 0.9274662939881486,0.748822921831065,0.9834210337741146,0.7163489248877463,
 0.43080743207086514,0.42604934930436456,0.6787271759162777,
 0.07817485086347886,0.5573890427598083,0.2162683531072549,
 0.02551515650512093,0.7388265416980843,0.8348481704280574,0.7369806497407886,
 0.8261386066433691,0.04719043956504665,0.7509372018752015,0.1958072588297025,
 0.5290839607349084,0.4166711086402419,0.06588372215011751,0.9860517587845956,
 0.7441314269857365,0.03666352203662193,0.15878511660738415,
 0.29355445073800535,0.6768475570442898,0.20807423711027062,
 0.5801646734621372,0.9070480726291671,0.17454750267471963,
 0.07780760696732758,0.42533559363153506,0.7236730124751513,
 0.1363463422005723,0.28645785600666673,0.8650425469631294,0.249664462704694,
 0.2867949220740724,0.7679283114784036,0.18548179878566995,0.658398882909927,
 0.8035649671241011,0.8106041466474421,0.8111571615100023,0.22904688786768546,
 0.6617945059547277,0.8370364531297244,0.1350702373017929,0.9811049050165402,
 0.8861720718777439,0.5125279741196269,0.9010623513888429,0.42593270372536574,
 0.8182203536583071,0.0938052057514448,0.3725719524984693,0.5536800520959869,
 0.05400459165494742,0.07965145874408441,0.5739255368422347,
 0.20433229216483673,0.5987075088803914,0.7934910521949687,
 0.24088331374389904,0.35737708756091846,0.18624384290610152,
 0.2483194385796067,0.4012254746650218,0.9595622627398893,0.26280735011547984,
 0.31499558558394025,0.9841624325051705,0.9999393488607956,0.5737664502587381,
 0.7219333914538174,0.08547707434598384,0.8721003716667041,
 0.07213415275325052,0.6088103676199961,0.8132040386353011,
 0.09378621829705425,0.8590979746776384,0.6467058022702976,0.4502282210357501,
 0.31315211677904353,0.09255637802472694,0.2040379736903407,
 0.10307465462202725,0.34408668052018565,0.7651532010073007,
 0.5584659202040336,0.9286000423686912,0.4481074478329016,0.49717099215441474,
 0.4646392674343539,0.20261882572583834,0.416260886998437,0.6272478427889836,
 0.47176500776360997,0.7519002332701362,0.06022281028874776,
 0.10222587345938927,0.0590614275241218,0.7137696546428322,0.5705605813854698,
 0.6254682944517638,0.22441603125433784,0.8691902283179,0.933535478847287,
 0.6363899600983546,0.3207555878517119,0.0980830522683902,0.8893748371388726,
 0.6364108013047811,0.03894358866588621,0.9865276726800423,0.8634893463909609,
 0.7859759873486217,0.9355197660971144,0.9805592309850335,0.6914932318635639,
 0.9566993828304009,0.4518539067514592,0.020927358445715027,
 0.7837867492035525,0.2293644982702897,0.8224871767906858,0.800455347086151,
 0.9075791798477122,0.35710613124338964,0.9704251059774814,0.7478464414546967,
 0.7936419505353667,0.15384297328770158,0.14344566847855345,
 0.3252046995406761,0.9895436645502257,0.06168413465648759,0.4066247344264746,
 0.6538556989895348,0.5358924087326821,0.7046938885948609,0.7886472460317331,
 0.2545476874601945,0.5731681805905879,0.2879613510157313,0.3100151730710575,
 0.6656191464936283,0.6452490728826342,0.4922253517698171,0.8223404352188569,
 0.43518125237201044,0.7680028776151357,0.8886609423030759,0.1426461092781809,
 0.17806545104775218,0.17004970864037294,0.9804542932160377,0.960316432200046,
 0.17012405172114675,0.7991507813864566,0.8713770971537623,
 0.31709844436456003,0.9149371014419225,0.47346061302756937,
 0.06450619364189114,0.24617752448115593,0.016817277669857078,
 0.8786805140195053,0.917976918028821,0.03204010516626565,0.7409300267719013,
 0.27520968362513254,0.7592142269333197,0.35140973827311917,
 0.9009793883733298,0.9516307604087204,0.7792782482030469,0.48253026032378754,
 0.16107533765295157,0.19037441126776378,0.5024151567629801,
 0.4295793078089919,0.7099106023354109,0.42590485685799784,0.9300117573860182,
 0.19000224815486644,0.08509600752626889,0.7405659106543383,
 0.15221292753697735,0.18168210555380837,0.275923376652879,
 0.24637563980180377,0.3354118899834655,0.023149110793520555,
 0.46325945367068677,0.44510480710025424,0.731705971064895,0.9291891005570215,
 0.8719448766206765,0.5157628195272321,0.11334926916906984,0.6553343350468367,
 0.4422436190805782,0.7067079195534832,0.3543120378355036,0.2680464769130495,
 0.8577859225285257,0.18186649636859498,0.5516034608340034,0.9687096384499165,
 0.689906078511067,0.45908467109144546,0.9040214045887253,
 0.0019553818249419264,0.5774000264120216,0.25882835395367754,
 0.7665809017848477,0.9510034003579637,0.9250310133425936,0.33890085688718286,
 0.8516001795559665,0.5238986900715735,0.12755907634439034,0.9040032632989974,
 0.046404299880360544,0.9137278317496569,0.531259099733707,0.6929310270694287,
 0.5634315206432544,0.32267097318613447,0.48873088814427146,
 0.4256801789909911,0.585086970062134,0.9177350739733166,0.2367017193579849,
 0.395199592411722,0.942767084602062,0.043219394745807604,0.9117705389821195,
 0.19534408757962218,0.1133179958119016,0.6105608474001213,0.9459459731224357,
 0.8229037982866274,0.9554985577989358,0.9100122537726425,0.03196225795061236,
 0.81263272084611,0.21466788712982665,0.714625263795396,0.4962336549715901,
 0.440633074522501,0.6766226748705462,0.6363950190044617,0.4928229591317971,
 0.13812681609903044,0.14071156991871658,0.04600843994755155,
 0.7656959785053604,0.4880389102514514,0.8884909017652182,0.2066978464778113,
 0.15247663062299577,0.42357418286040904,0.37221992778431434,
 0.010286193151127665,0.7120794294794996,0.05007248550682608,
 0.4962580538911662,0.17705540259454744,0.22535574711644812,0.934083977373905,
 0.2594953374249237,0.2699845700339546,0.02095802334190433,0.8746904807357376,
 0.7872459013459012,0.8975360789405051,0.1909288151747104,0.4266425237596586,
 0.6729526414326811,0.3689741966083758,0.23782528807284986,0.3764982635024392,
 0.41041210388752314,0.24642124564910395,0.9770289058510119,
 0.48354711013589524,0.20806514920819552,0.98663896815742,0.9666648898163278,
 0.24495647867991266,0.025483275861011956,0.2582320260064179,
 0.45561917196287194,0.24394441406849277,0.09539795072203772,
 0.10510191223369114,0.6009186426722201,0.5295925689307872,0.6852903954979546,
 0.1202528578966735,0.46289214766182274,0.6695040560265159,
 0.031502751986357325,0.3111309756239323,0.6182771295456625,
 0.5932454556618207,0.7400392529954489,0.21985428742671798,0.8722323818306696,
 0.2945314538658883,0.17268057165128015,0.10595230063035643,
 0.6265591225191363,0.03578777718695214,0.39257425475851393,
 0.8142487765530645,0.150831621622866,0.9202671669727148,0.5212176162590662,
 0.8269164134303467,0.7280421327256967,0.16579466256494413,
 0.22074610250192805,0.3260958865692646,0.42455189967947815,
 0.15899704547248483,0.7569724560557178,0.16127600441804102,
 0.7102925507815763,0.47944555096688435,0.6881258167430985,0.8880541655579143,
 0.9162853302102808,0.41740311648299055,0.12320946805938049,
 0.3064823780931799,0.915367050850591,0.010734971136954319,
 0.18915443812135502,0.3509095406928412,0.6903403394331004,
 0.31560394956400784,0.8870855299430552,0.5922438439779263,
 0.13278208687951565,0.42580034099413744,0.09061317140941572,
 0.9069349272338887,0.8724890727376714,0.08554909070425576,0.549246563084341,
 0.7520744955630143,0.7459084190616374,0.32694434431603314,
 0.20122361691386725,0.5301455503893721,0.9714717915445594,0.6870563481192578,
 0.3278455691426021,0.8541867050409545,0.6667884805878485,0.47614585498018924,
 0.530419559226761,0.4328382872660609,0.8522955015105944,0.8068228696364328,
 0.40052508364332995,0.7715759932928378,0.8171051906088711,0.8258503625362037,
 0.7159206054606254,0.16532783336629175,0.4508338580291239,0.9498800630043974,
 0.41025653521460226,0.529937430914322,0.5645235393652684,0.23617805106539747,
 0.7615980632213571,0.4296123092674805,0.8193065364091994,0.3370351868847745,
 0.2916198468539243,0.0943566165250912,0.754787596972823,0.5198722683471997,
 0.8937073119117572,0.22970632807088476,0.5405138067533849,0.288414899299298,
 0.5561447386273299,0.22208870665345204,0.4489297323679047,0.6311545061836281,
 0.0804859522427388,0.2247389838227838,0.03760207066598131,0.4967564287190406,
 0.7501322616080124,0.46547401251154685,0.09083812506387334,
 0.3277245404012048,0.8206268556786647,0.39342523820754893,
 0.10357793206801269,0.16426548226973814,0.013799412274223988,
 0.19412865603896023,0.8445702373220035,0.6384108586904206,
 0.08997184702824612,0.3817426955785427,0.5275013945806523,
 0.09536379697273523,0.6102071283438615,0.4332253977642202,0.5975884642790854,
 0.8318496077042201,0.6607737138249432,0.042312798046867495,
 0.9700323006155243,0.9861769699802345,0.9302591526476551,0.12048967561623725,
 0.6701700703117122,0.9742447366786636,0.9073979520427605,0.7451213773914449,
 0.3889174734482246,0.5132365395757205,0.17412015020663396,0.8988217700781542,
 0.21974062918537207,0.3658935391747481,0.7791271566047892,0.7057983763655127,
 0.44063972521271333,0.9328069683338313,0.32116143134435826,
 0.14307458520294292,0.048745191119836107,0.556868068696563,
 0.3331187482577437,0.19243040254962307,0.5832960604781746,
 0.17254778067526289,0.844218455480477,0.15722555909650682,
 0.04772826886778114,0.33247955563284504,0.46517640955993234,
 0.49315652037952085,0.5332904484251832,0.7123172875637405,0.5175855067820558,
 0.5961451827969322,0.7232589882948833,0.8196939135515401,0.6931863794167494,
 0.9894290243491259,0.6752417159887638,0.03177602470605678,0.849830672560684,
 0.8384286758557229,0.7413794862940625,0.5264511401891813,0.6000726376642921,
 0.3409406104102024,0.7655772685886962,0.6299417148871085,0.26378819796688324,
 0.17113777442231826,0.644505618084568,0.31281975398764095,0.8052103084240424,
 0.7497672263902916,0.3060393005892218,0.09465331018906531,0.3274529546354339,
 0.8136756633700858,0.15581767632951637,0.3959600895214843,0.9609175830678799,
 0.21073299025306924,0.09742008274676373,0.04849252466826548,
 0.5745463033348756,0.4215341702101807,0.04858917503215243,0.7577179935343018,
 0.869616843871794,0.43510889776613426,0.9862871921642318,0.7478600372867126,
 0.5728552265505211,0.5695711008520693,0.22705569639528167,0.6561609592956389,
 0.4639376108279578,0.6488881330803877,0.051706037733005195,
 0.6629512870954082,0.13383402223580299,0.04314949973090254,
 0.6521655834074822,0.6791654494615794,0.5955315432715558,0.7184583583986179,
 0.5583048599499267,0.3060119172219984,0.3474374098186864,0.2877089169728775,
 0.1204968198809675,0.10058317377017534,0.8980777308813155,0.6616589391596923,
 0.43818919930424904,0.2167314010479232,0.2783625806494723,0.8243791289709093,
 0.2999962629235774,0.5862626824296692,0.34456907460363806,0.7996340165525903,
 0.6387761397121443,0.9662550215576101,0.20808300568528937,
 0.18234392402999977,0.6792742695656381,0.5216588928484086,0.0998432936115039,
 0.13539986508138413,0.5346370056879657,0.5604461761903736,0.7657086590667511,
 0.2520652752217677,0.9863233991256681,0.05776682299035063,
 0.24691361970729075,0.8910593295310182,0.33024326738707344,0.252669154915805,
 0.8517587970249771,0.2216098842862887,0.6781713745275162,0.9923352800747888,
 0.6151542467428848,0.374592575242101,0.48965063664715114,0.5185334360197261,
 0.46970476365476943,0.7824115955784119,0.9004482293299616,0.6819385825500934,
 0.35662153215887726,0.7216494165213746,0.3758110707808413,0.7622434760847709,
 0.984855435885997,0.10599196158342983,0.872695574545771,0.2939176922726139,
 0.7662080879422423,0.7718689467857949,0.2605498934225725,0.7321743722715002,
 0.12000236275500531,0.12263713920939234,0.020740447853199284,
 0.32514118521454727,0.19898156787318022,0.48707730776578195,
 0.04778813035808693,0.13989331043339737,0.43706796406537163,
 0.6626298543921726,0.8119370085202972,0.42456335708703985,0.7190853117321288,
 0.4923457727666496,0.4755741320189927,0.5753574110659315,0.22117739191867491,
 0.2988545173563939,0.6992887845040419,0.17195481067967544,
 0.13034903503712214,0.9040771336164902,0.506172133324223,0.6140757654135293,
 0.3548901876563959,0.26227721924535063,0.7284940467828094,
 0.38510123044338596,0.35651679592844543,0.8535317120992888,
 0.9777875721014031,0.2501518979642203,0.3044177622510733,0.2701748146100805,
 0.5521184599828763,0.31815036570088173,0.03948817002243776,
 0.6256766236731621,0.8904246486884086,0.41981326945978326,
 0.09012038112926146,0.8189477738723385,0.6632274529218014,0.4964302969852571,
 0.8128580366941442,0.190702518355354,0.25081279599718564,0.2936025244742613,
 0.07849943922041913,0.4116437703210478,0.21845940587729318,
 0.5437416104860707,0.5543611751362931,0.35617449563296344,0.81727105329275,
 0.8839449717668287,0.9624330204817492,0.3329221109635744,0.31167657831886975,
 0.8702517921628964,0.9735878515352833,0.4303682856988663,0.5888887034145581,
 0.9180470417521285,0.015456320640942156,0.14327974951680622,
 0.5344818637548324,0.7533618061673456,0.2723747791114465,0.2614175690570052,
 0.9202307304778454,0.23919109912479175,0.4747861194575046,0.8704280853030764,
 0.4037583787736905,0.43097517860677126,0.20161577000192288,
 0.4005529433899716,0.7933064053076562,0.027676169730237232,
 0.2336862980811496,0.7909853884815443,0.17045982067754673,0.960602433942475,
 0.15969350214613964,0.18375669710150366,0.5599980724802426,
 0.0982275727916424,0.3809859549630491,0.9035608102212546,0.7119436984784706,
 0.7930110199913767,0.557680201917388,0.9252629253353739,0.8902708235050388,
 0.8175379860946541,0.22329369298283153,0.9951446483368845,
 0.48490886960902857,0.6770531361405426,0.7428848799071086,
 0.37667144948922093,0.25735446556814245,0.9333166169467966,
 0.7835240173858252,0.17064552390923082,0.3901105281231574,
 0.12198758405361676,0.3988268955342469,0.9275057595892141,0.8018909482460551,
 0.21108554839250993,0.701853907258694,0.4124303247035299,0.7279607349557868,
 0.8664656993182384,0.09679724089889563,0.6342251312032315,0.3981962319019773,
 0.09690648142742608,0.46809587149539955,0.7850444724468464,
 0.5033619477584379,0.576353393650491,0.8097571054939162,0.9168301834975683,
 0.33644695878762854,0.032110707096642876,0.5981079682507808,
 0.7101229082381344,0.06649558446682069,0.005379960476190715,
 0.6263418525197592,0.39023992012435205,0.08774030917234332,
 0.06885451703290002,0.3714009730401481,0.9313796319160323,
 0.06607365562890544,0.6877688521976032,0.04405669791022904,
 0.9353462946608015,0.8880279628049889,0.5370916362579132,0.03881619791140889,
 0.5561231382050243,0.9725518742908971,0.8433645516340136,0.1712197779582395,
 0.9079502041905745,0.9204229830200298,0.8420970229467081,0.8354872908260143,
 0.5860774373738507,0.4175295188872479,0.11823652134867457,
 0.05743328652648927,0.6979367383779057,0.5742068424221363,
 0.04512599333931866,0.20932229958144,0.38762418313780456,0.5205522887382964,
 0.9908305336107646,0.7505841478716339,0.6052685378190058,0.34671027073190575,
 0.4870731633315222,0.9907977401192771,0.4959964008347999,0.25409228131343786,
 0.31408570975892514,0.21826943605216376,0.5616305246626665,
 0.9422265395746128,0.6847478671375535,0.5217033972853702,0.8945218243682357,
 0.023366806334977896,0.06838993488707712,0.17823880229027,
 0.26035101759489987,0.06589404051072967,0.5507911480218953,
 0.3838211701260803,0.39784921051186284,0.8079519319989356,0.4561245826536572,
 0.8014393472560359,0.8228533012963923,0.20203745284517782,0.3459529304025707,
 0.009900405802924395,0.2852338148595627,0.6689377501460105,
 0.7650591231370606,0.8197863004754146,0.36169365822965804,
 0.48582813915442247,0.46580547198317823,0.0501116972862119,
 0.29011556061306365,0.2605437067017542,0.07303557147005657,
 0.2363012899400606,0.0686551401476293,0.2852348639026976,0.29877001084730714,
 0.29694347788404674,0.2893045584658923,0.002241429898966696,
 0.5527871776039048,0.11835876522903277,0.25988051709895266,
 0.6024889283747429,0.5539534136799946,0.011427115307982016,
 0.07807089962484393,0.5518946455723442,0.6783477310539758,
 0.43446316062601653,0.18360568132965305,0.20849856843496972,
 0.9456965471536567,0.7224120503106898,0.9760419484523473,0.7064507457957371,
 0.1279796611211298,0.034571558378032274,0.5013604069391909,
 0.39101505913673895,0.6549342256863048,0.23534885556890817,
 0.3151964954953419,0.7500366563836524,0.16455129169252747,
 0.23345167650944232,0.1435730156952797,0.7685324781179969,0.7262498567950789,
 0.18270002444286682,0.9244767983156389,0.5390546482431061,0.264723057721882,
 0.4552722553553956,0.09265861927572527,0.46827049554757494,
 0.044049198107421805,0.6885811325070095,0.41439786874481566,
 0.14988745396034897,0.9003262152367136,0.7632428167568939,
 0.11319319654579174,0.3068844023630446,0.8598737075249572,0.373698404941762,
 0.4860329229384338,0.5078161300209757,0.7524947920843825,0.30360778060840454,
 0.9643912802613356,0.2539972231420937,0.37173427552478244,0.5730636104426616,
 0.6593016679966981,0.08360635878511569,0.850350721539581,0.5911801732842077,
 0.6521800813557811,0.6704449021300795,0.408913763909599,0.8314122853323116,
 0.3341334554094698,0.9587833417500284,0.6659339030686882,0.0683797485927975,
 0.9980273364947109,0.3384850602745404,0.40720967383671536,0.4446043552843679,
 0.3194227395366177,0.7416281780473121,0.4134360654083541,0.8770236690484595,
 0.9583800371213975,0.059897662378677,0.38593240048339517,
 0.024208091869505677,0.41981086398207657,0.5537311232079187,
 0.5846756380514807,0.5673078677246949,0.16176938219475356,0.409378078069384,
 0.8978750385795653,0.95673218167976,0.9556536945081063,0.2437870431232756,
 0.4172370713435851,0.2776364354236468,0.22060684736767855,0.7173416423153467,
 0.3598213864124755,0.31526224204326125,0.3913213369668048,0.6722170647775233,
 0.9448087397530138,0.35565029925324176,0.5084668532311646,0.8844124119321642,
 0.0049887204022383,0.6306804608016816,0.25866426888221183,
 0.027189931946222212,0.558782919603875,0.4691388318260047,0.5079028555151724,
 0.7374712194320265,0.19652206697340713,0.624715885753455,0.5975539894969503,
 0.9593498800023594,0.15989045125295842,0.4147070136272615,0.4435485230062086,
 0.41296006459635215,0.1473774678382761,0.3948609764407359,0.3037724483637878,
 0.9403998410602376,0.6640590920709264,0.4485694341867139,0.9236163225705102,
 0.18422050951071278,0.4835237587017218,0.3617846702432529,0.32745773121774,
 0.9230863775146294,0.818275795274775,0.5678310239405523,0.6763110102537864,
 0.9152211595803565,0.2604778452608666,0.3656804383727422,
 0.048075452894592585,0.8371342748300805,0.32307264038619643,
 0.7165621340028778,0.5332959004036381,0.3975288439397727,0.3978117628986806,
 0.22478469428106262,0.37917495181676786,0.3813578415831458,
 0.23818403382417774,0.26122151319461906,0.5275180112501469,
 0.6365950564757937,0.6814684810700031,0.8748725678128518,0.23211491083241442,
 0.8009013701362357,0.8552410090839517,0.752162211910525,0.9883540977327591,
 0.7808585447424335,0.0950202809134284,0.3716176519799884,0.4515372243114276,
 0.8518842508217235,0.29292015091975265,0.20601332086940916,
 0.5999195307421914,0.24226969142223354,0.7671743196349763,0.3845468328169759,
 0.8854304540549734,0.8292623244161469,0.038449724614999625,
 0.010413634641252001,0.0729469286771448,0.4398610528768614,
 0.6096189105385624,0.8429465551325684,0.44213133303095964,0.577077609325817,
 0.6207904411437793,0.18900976077207343,0.09810937211474591,
 0.4586215996140448,0.3093438277586962,0.5752744913827514,0.2442431924154953,
 0.3369799667177916,0.6767586502798786,0.25698315287660634,0.5744226943660842,
 0.17642963033928416,0.411172004876188,0.8682274877228308,0.24204147364665185,
 0.30398930832776383,0.8173692223290062,0.08315678577288999,
 0.29868247662555003,0.007972986006551941,0.37559404186221446,
 0.1645479734069294,0.6553355184187333,0.08189962881991952,0.3337164069944727,
 0.7937849539667976,0.31835707854355366,0.5444749490350993,0.6924740859368608,
 0.03433258176811793,0.23322708260923186,0.16033270393937094,
 0.22145623639088086,0.5188984559532293,0.9273244611195599,0.1592316442570364,
 0.34394879496285224,0.319628735528665,0.6839207117029906,0.6805566812408848,
 0.11565861000979716,0.08690171772909028,0.2364931439885144,
 0.06002690330570537,0.2744207300632986,0.0030909039578121966,
 0.7134058510747956,0.41098109015227713,0.10385502769928934,0.818653801261024,
 0.1871617709810739,0.5300895379854118,0.16806278620725823,
 0.24638621225227186,0.030594373114414153,0.6308811208163269,
 0.7615871867167816,0.477462107681085,0.9935670097482787,0.9090938803630001,
 0.8397441790211987,0.3307635887934606,0.6452316589734874,0.6737787139586388,
 0.42785960884153373,0.5912471780077777,0.5409504160934505,
 0.18483659565750343,0.38591386845410813,0.8488877644084221,0.813977554869064,
 0.934852572482772,0.35280199879313046,0.3579820389910493,0.8681070109519687,
 0.741127411262657,0.4826486733584765,0.5086594511331981,0.8002082216327953,
 0.9866557327564427,0.6859714590561677,0.9276220134620683,0.3794392217420235,
 0.9614184666630392,0.41203016314295837,0.805032456335501,0.8361904794705182,
 0.9630084450980689,0.15823858371721145,0.19353013404896258,
 0.6314137593370243,0.09793536490928101,0.25308077077280255,
 0.8907663363670774,0.4785080333074214,0.22975723634721512,0.7434109532826135,
 0.35343019111204466,0.327809737719682,0.30551243994882027,0.4761505904029394,
 0.6090890415276813,0.5508215293432246,0.4744240978225528,0.8480616775421406,
 0.26076513755942077,0.5220115368488157,0.5499415211927826,0.620319923352571,
 0.346449917669041,0.5837460311097129,0.5782657691331548,0.27461700015946344,
 0.3340368165089841,0.8051145489952213,0.18691126943148906,0.3893993219520516,
 0.4963904489380786,0.915659144976669,0.23113588497313697,0.7008239175771155,
 0.07587442547691003,0.5893350159905912,0.6205520318728533,0.636744985780443,
 0.9849363627492114,0.40366074987035405,0.35059507108195165,0.84027697066252,
 0.9401659784901826,0.5593615425357318,0.060246431131485645,
 0.5608528116519428,0.17224556371123279,0.9151324009895871,
 0.10933154879794071,0.09168932828885856,0.4250806049049113,0.667826055132711,
 0.5522349121340367,0.8929665765572893,0.045783238943647664,
 0.26774072232368484,0.35446638083950455,0.6028673090544543,
 0.4710348673539494,0.08450220801896124,0.6073791966184715,0.9802314212667761,
 0.2561437375549689,0.027704252053402323,0.9562174400788195,
 0.8505722578693373,0.5536752446393887,0.998339882064566,0.957031055866979,
 0.12100147354721724,0.6011186030822155,0.21832263748398684,
 0.39008260815801965,0.9925990810482227,0.9347903015687086,0.4912941831710371,
 0.8164310446833594,0.9713139668546977,0.9951162185058174,0.706673511491908,
 0.9062855265899836,0.6074806482725918,0.3309241720974788,0.9782873683267995,
 0.560570584027551,0.9142580498336561,0.48244603540200603,
 0.060882246792943295,0.6605347431684495,0.11064805311720494,
 0.9564386586160369,0.36291500146451927,0.7674033997099619,0.5014093111753556,
 0.6731789702964365,0.23057321886543447,0.08106890583190995,
 0.23472995110296235,0.854429542537317,0.6235951664234429,0.8902395302908768,
 0.6622244108318392,0.9466939608149942,0.6163214882890823,0.20202286185176044,
 0.40030656411363796,0.9538539197811566,0.013422646045437858,
 0.6675198112183962,0.6928463157239311,0.37272942799378694,0.7042648847541904,
 0.8011197291378875,0.14070303282490537,0.5689014330612109,0.626584793729661,
 0.5541926185342961,0.38409627227251586,0.6603355973243865,
 0.30190249543091796,0.14775638313516903,0.5780693357184685,
 0.6696323642954564,0.011365650330721455,0.6624841575388083,
 0.8613682867077127,0.3564823425187934,0.21855707700359084,0.9504613835942362,
 0.9624376062881427,0.04327079369874487,0.23490924837059035,
 0.7011971439336512,0.11362601765362768,0.5315033073474312,0.7680761862770974,
 0.1030333168355193,0.3916632978774873,0.3751704383306611,0.16734187695685743,
 0.2797670087974182,0.7079873097797488,0.8261748981146524,0.9795575382992598,
 0.750369396138229,0.2806846047611561,0.4158030980198991,0.39568310782785754,
 0.1397161082679833,0.5346662928465061,0.5908768728105143,0.900361921742555,
 0.40751007272128026,0.5710443982334034,0.795217964406185,0.9278576948173115,
 0.9682396935616662,0.785175223524664,0.7849024487311539,0.575344383594929,
 0.4663070002778513,0.31442201810512715,0.5798068644930463,0.7616174197567176,
 0.4165498478065407,0.7901266976518615,0.3347354267552821,0.2618099221151339,
 0.0647481102646168,0.6133007065498848,0.10427157018074273,
 0.15353941339481036,0.9006738024547047,0.2974614532374116,0.6330260310035659,
 0.8125598728905761,0.12072489047678392,0.7683243840212465,0.704465148373652,
 0.6932202602901644,0.8260200745808491,0.24479879192969833,
 0.015333840883967276,0.5644479311479902,0.5277524133120232,
 0.19782008841144316,0.5732637135639657,0.1404447884665534,0.7575234682683322,
 0.24692770614303527,0.6095951798414024,0.008166979247527295,
 0.8681204387220866,0.24680131299022035,0.9795380454561479,
 0.04028017285603047,0.8266297198734364,0.06217522598169656,
 0.8355891629499134,0.3551751627412383,0.0041603079724967795,
 0.405697320461445,0.28755119740393853,0.256786812690566,0.2912087852025981,
 0.21494899925460254,0.9366766707789702,0.1397096649990971,0.4879193402252725,
 0.47114951770274416,0.22674819199140372,0.03025957296576942,
 0.6791635670456417,0.7878396300565126,0.41757499253863295,0.6427015874450175,
 0.7527879122844677,0.2180386585974876,0.4496389764150369,0.2254940608490834,
 0.4808169538920666,0.613059704016172,0.11427331822346454,0.06117132468606412,
 0.5099229454766997,0.12604821903301777,0.8010654465794297,
 0.13966591664316097,0.5087922713464295,0.7605486581882815,0.8528984325142812,
 0.26700475002624047,0.906478916128704,0.8966732126606182,0.15664478857373854,
 0.9273643878340165,0.16168408282261348,0.9647441330140925,0.9801999876720756,
 0.31858036294348946,0.3815102552261641,0.9805672403148127,0.8829963628046836,
 0.3629424612039409,0.0921250230953875,0.5408740541087185,0.27100061730597214,
 0.34160870881099015,0.09720409863036172,0.28580182937309495,
 0.21475563577832835,0.3898258462583709,0.26339477289255275,
 0.14836900979860723,0.5265780367302634,0.5433124021488929,
 0.14624602361080918,0.8883302384764374,0.2540601665331673,0.5055253994012404,
 0.45240043852418454,0.6416573138263897,0.5661086881433512,0.5825271694150881,
 0.7121148574400409,0.962357868625561,0.8355358195271794,0.20780585819522113,
 0.943495159070598,0.9760478748819011,0.2596395283989035,0.9290782120950419,
 0.5599671763938108,0.3765867207586354,0.5301218560719788,0.18557862093304867,
 0.2294817517374791,0.8461205610355571,0.024306140922234154,
 0.39159013584355695,0.8090979640917801,0.7279483341884627,
 0.19360053779989728,0.6926039676634064,0.15972913005583467,
 0.4594335847563632,0.8734533160229885,0.6337650385934817,0.6333283348350747,
 0.9275950598847521,0.7929632184998995,0.7183892049820552,0.6858161613592865,
 0.17489766490373704,0.5810704335539009,0.3749531626829291,0.719635397577559,
 0.621535042344167,0.7670850185819283,0.8272817442873787,0.26893142246912094,
 0.5673568335880055,0.8341771194879943,0.49513180888214015,
 0.02607621287630868,0.7789923240694526,0.5971639258098811,
 0.45276262486892815,0.1974927470816502,0.3594901428966215,0.6110256818534292,
 0.3746412365921602,0.09868898825351136,0.7868174714004109,0.088954187060715,
 0.9546334706420048,0.5114277405853376,0.7504516996526058,0.09593623660787498,
 0.5308674884554736,0.9902426629424054,0.9054357109713518,0.5577981138497528,
 0.41297088328438103,0.34030836482831883,0.038075158810663146,
 0.3802101375695983,0.22493747265526098,0.6797080465297382,
 0.027764600943134043,0.5952964155082543,0.6118202634995773,
 0.31338119208986015,0.9948369568245493,0.3984167797147664,0.9468417826963559,
 0.6451633857955317,0.7319742619469558,0.7901502733461276,0.6611340757500914,
 0.4705825380108204,0.46860816849139963,0.7399839086811927,
 0.19278340482819134,0.3669010804849884,0.13185914897682127,
 0.10401525747363949,0.1822950255704774,0.9920948334945685,0.3385980139619641,
 0.27136941049440555,0.5789291848225201,0.7094270122416508,
 0.43842987241056364,0.15654460431076023,0.46561404264813566,
 0.1298146094238074,0.9751159584423147,0.15782840164485457,0.8785783058404418,
 0.3080571056171605,0.7447610338702816,0.004131440108122053,
 0.14956220291608346,0.7585845051452043,0.17931033640230065,
 0.7514132364687254,0.9595978125069075,0.9245546375833029,0.856925203100874,
 0.940614771494883,0.7848753421555343,0.7417000631204904,0.020114736983285475,
 0.35094140927275586,0.6740229375066318,0.7505100765116857,0.6670260194595516,
 0.39871568050629824,0.04011720060539892,0.9691884095946798,
 0.8785675601766906,0.16773967386527555,0.8700446936801308,
 0.27520378728861483,0.31566740389820136,0.01850184048253045,
 0.03325034830140294,0.008509112755989756,0.3904182437032121,
 0.6258735631372441,0.9462564560139266,0.8840763953881561,0.9030804666545459,
 0.3116421456458838,0.41069224980643726,0.3020498597508233,0.7495089131913585,
 0.9715470785508004,0.351438597199347,0.04883119470893882,0.16577195496251507,
 0.38623198422241933,0.32259890309363903,0.4779829987379949,
 0.6771403398962785,0.08515798352445281,0.6565580440951521,0.7570073275864493,
 0.8649492736649071,0.778208786326889,0.7667231654363457,0.958761995290372,
 0.008498675227715413,0.5449464827028662,0.33529307327437174,
 4.817853686001072e-5,0.4218630795405661,0.6585432569862935,
 0.8624753214164973,0.17016412646757872,0.9299070926557184,0.2563074750314187,
 0.8798423018082179,0.3915590755265783,0.5470209142853506,0.6634542947422516,
 0.9972863918601751,0.7160905938640689,0.7385555603166312,0.5273814623301533,
 0.6973891101777664,0.21923410480017214,0.8826455560160771,0.7553517074030813,
 0.8316302319489031,0.400113043687902,0.5171577973416623,0.6008343887264173,
 0.9396918558220331,0.34589463026404565,0.5721009840504493,0.8147361786101924,
 0.3657726650414115,0.9806943763801952,0.44540267970202463,0.7001330624276848,
 0.4699973406627147,0.6960311726843066,0.9633362554338705,0.4711025334209089,
 0.015478680798274858,0.912648921112379,0.3463426830052827,0.114965817921869,
 0.9289539831821835,0.557555876573117,0.12441300858203186,0.12824905796106167,
 0.35963882053215657,0.8875010032718346,0.8830156528661807,
 0.46821430955052945,0.7097924488868583,0.4411003381368783,0.6896066672723927,
 0.036122515440607406,0.1526388145126839,0.8143169709109049,
 0.043206483623832836,0.05813716341329822,0.08444074846398031,
 0.26178254153258096,0.43909281682564805,0.06448538239185497,
 0.6152509659296559,0.8142944013530777,0.3089544454786103,0.4670458702160736,
 0.07564539676398196,0.7820843978509103,0.7679136390168505,0.997911658868061,
 0.510550939755911,0.025607746591757796,0.1292482071417994,0.3185615359775603,
 0.9806676437980093,0.4699205291601012,0.06113531447183418,0.7522417916343163,
 0.44626072260394256,0.42970092342320443,0.2680269438554478,
 0.19082944197885476,0.5671231101941641,0.6133348090388043,0.8996211108502645,
 0.4252711145662187,0.08298410997373473,0.8725492021783369,0.8617113512659452,
 0.09091654921291747,0.3284405941471925,0.6827161637102184,0.3269137412876063,
 0.6988670934729688,0.9382027453335573,0.29563375112436674,0.5595650813202229,
 0.8977345188594238,0.5005935458350868,0.3349118443918825,0.7117551169637663,
 0.7120862672398299,0.06512703267559672,0.09345045475208391,
 0.27258973857548807,0.006995851481949034,0.5846473407028886,
 0.8367348222354551,0.24253495341136533,0.9125934747360581,0.5602874219707451,
 0.5607544500020082,0.41452447249785274,0.4180824319715529,
 0.011408437097067958,0.6461018427913124,0.21306708744580038,
 0.3829105985055823,0.8546232822794008,0.6127652051047408,0.6628251075126096,
 0.9851007567435881,0.18728247840766443,0.7947414650816008,
 0.08875266011467153,0.5528157613664952,0.2661580479450133,0.8293302810133462,
 0.38095031083893993,0.16987128369396787,0.3849196797110852,
 0.9216014431464404,0.9412593972213747,0.9691270825445567,0.5579501489125368,
 0.46267197595364795,0.9986171253777285,0.9908214565101626,0.5411704542069357,
 0.07852214746161745,0.19314962192645868,0.6912388600124788,
 0.5939194074423132,0.12882536432939928,0.6015725535292182,0.274950287475079,
 0.8049171500846869,0.0037278193713201624,0.9457447387866013,
 0.9699683729573327,0.6859728780371657,0.3143378816792848,0.16068008997648153,
 0.04917048231039689,0.010356226072191665,0.8750638400260604,
 0.9697105161594551,0.7752005186092717,0.3751858331684602,0.6117217126545118,
 0.6769035629949999,0.973701183119067,0.4105540807342847,0.8535129136736427,
 0.8973798237831087,0.7472141045674277,0.10860727165888873,0.6728382765514971,
 0.3426165304639024,0.41732691264569355,0.592720334585483,0.6880478376542524,
 0.4265106644288702,0.30293939324327634,0.7314802908208342,
 0.001417686485710834,0.971926130734346,0.42361454113872055,0.622425211677582,
 0.9385093177066619,0.5587293212597702,0.9754834352481749,0.5822755921788482,
 0.10204652731866093,0.9983412725614801,0.12323084090451919,
 0.9632440221126579,0.40383276383269273,0.17522982123880493,
 0.8438343510887096,0.2186780951884666,0.18991414260109563,0.7079283507960605,
 0.1895252984464153,0.5212667949494455,0.9786877063083748,0.9905814665127263,
 0.36494349138857385,0.8981526659964616,0.28711028607759204,
 0.35302149400685323,0.11019461613701464,0.2308005768679977,
 0.6974955098111966,0.6182641333918174,0.7464227031990366,0.13289596934907089,
 0.14869511325943374,0.8234923849294781,0.7561579301488464,0.7367956994284816,
 0.8574833725838653,0.8086896198221531,0.3442851773252341,0.7337162827311597,
 0.7990950543752824,0.21661858851230154,0.48917202966436035,
 0.26525230394280674,0.7550808596364458,0.031481472313299896,
 0.6424977698447799,0.9883845754658622,0.5972219987181517,0.21467279369422965,
 0.5653036881605026,0.17121625939574792,0.2647515615353795,0.3312478482238924,
 0.23338797808754586,0.21898151364079155,0.3821089740996403,
 0.8672677889442612,0.011332675686156257,0.7855050014457482,
 0.7182795528591172,0.002133579859907586,0.6867211976459685,
 0.21616795697298963,0.9322897062263013,0.052949145258397445,
 0.19887895457906568,0.5950739829995702,0.7927336412755415,
 0.16929982035331126,0.804509780245331,0.14159388572834608,
 0.051302584382771355,0.6939806765177234,0.08990537127705711,
 0.9746099664690445,0.4627338878102131,0.9084354154071669,0.9965868324687515,
 0.8157503690751252,0.45464811333118493,0.3613058880330825,
 0.18235052227202875,0.7129758693367114,0.5107186634736764,0.5760104866520639,
 0.5398309205569689,0.3761308116600769,0.20210092125677037,0.6283303136653752,
 0.07739404732203914,0.35424762242173513,0.6992088968277803,
 0.5372235249618775,0.6587962965242168,0.614739901230861,0.20487365242701394,
 0.23049800909278984,0.6360885684669165,0.0898970821131232,
 0.15400812336135505,0.5128428825390124,0.5797580172175826,
 0.10262509179994672,0.16853918801232948,0.42841204006972666,
 0.3370997827813227,0.860946437727886,0.942562796663462,0.7398402533133671,
 0.5405644906425131,0.7303642265322193,0.14494775558016304,0.6013860718804964,
 0.40086867839842144,0.12224646441869114,0.4272085290975248,
 0.8113408404771475,0.27857845045627283,0.7089900137620105,0.647121652032114,
 0.2669878233154891,0.44039877855873244,0.27408836543965165,
 0.0031402583026411968,0.844384212123792,0.7137339062206408,0.607345490567566,
 0.7386454047949728,0.2950023160485514,0.16703031597237417,0.5200032416714482,
 0.47821411760331745,0.2319468354656845,0.13816516257547629,
 0.1750998738821431,0.18713589010046938,0.8876127627631613,0.8686952562335696,
 0.9999329026034666,0.1876754390726164,0.47898658464565924,0.859293183231347,
 0.20787795304011625,0.6629918796172104,0.9201782570019654,0.4589796181131728,
 0.027235814293152583,0.7623140774159923,0.013403517268065546,
 0.7803708462169237,0.08080914893734958,0.47648119747072626,
 0.6323333845981555,0.622304754119901,0.8358955458326583,0.16138277440587656,
 0.6768692203857676,0.5364731214817471,0.7495666407810851,0.07399977034905592,
 0.6559141606727701,0.8269058734412508,0.4569674970254445,0.8352997916872276,
 0.819165061323589,0.9050549735327833,0.722997761043019,0.2805397837777267,
 0.5476897940288366,0.5806852662346631,0.40340263980906366,0.4557017437683374,
 0.9608196465874496,0.17714203605411638,0.3434322182080257,0.7889179922779737,
 0.8630962075069418,0.15756034553978038,0.2568397929099665,0.5561808607548253,
 0.8727585112120123,0.5722043919990971,0.1866862264979684,0.20230494861619108,
 0.34000155050106706,0.9466177705943495,0.5392366690225509,0.2463761516752484,
 0.7022124733209713,0.8219467724610168,0.031175072461008813,
 0.7104395379784396,0.7798543868260476,0.16788692711654618,
 0.09908151948473576,0.9268112667079752,0.5243958493046923,0.8501366231874654,
 0.9730822939532248,0.5060865434680342,0.19436046522812434,0.7922105698875211,
 0.10722086388676366,0.412798704027591,0.5414450794131216,0.8832904311386597,
 0.7806640432307692,0.16256360640003298,0.5813639840048178,0.8301769966579918,
 0.4716266752721626,0.07994690368286506,0.8767179748646455,0.6359735782153645,
 0.5860687178926424,0.6296378737934498,0.03142957342694441,0.429782700142203,
 0.039151811985553286,0.9768795530808476,0.3886949591848817,
 0.09321535822968396,0.4892545660760306,0.9918142700625044,0.918589794783238,
 0.6156447785276782,0.11580111043062158,0.0491818393946416,0.7838599404516706,
 0.15124521570113758,0.8881046787488748,0.8811768078144427,0.5071027893768484,
 0.09709746073198233,0.25461572538891375,0.5718647271973476,0.548043024456362,
 0.23859313499633705,0.49716788494842634,0.9195721181800249,
 0.35056191265375003,0.8962837628774911,0.07276650373830029,
 0.11368400328510209,0.6454903830101135,0.244300805703888,0.9050803310878477,
 0.46971298725292887,0.022842114807690383,0.10573048365196402,
 0.03783020400653092,0.9159468645192748,0.37437210023033596,
 0.07758558886851041,0.6117892432581704,6.941393044230804e-4,
 0.25333639759382376,0.4405524092862705,0.4668067044414387,0.6794893066682911,
 0.5708702830383042,0.9698134814846133,0.8504529115678694,0.6500204100690172,
 0.743619593168317,0.1632323801109855,0.9596027193526557,0.9440826150453775,
 0.27383809465981157,0.8023675064330988,0.028056469625969638,
 0.48902431969126314,0.6515480664910054,0.03534107761926175,
 0.4396615104035526,0.0291220083351319,0.35217666459418406,0.8431830540908762,
 0.4005250840532354,0.5090455527785895,0.8696199355453386,0.039091404926898,
 0.9974418819265972,0.8951409290082647,0.6692410583822833,0.5702881125408492,
 0.5986597548211169,0.29637582083364833,0.5320855910111867,
 0.44399491880668085,0.09362796547232277,0.3877754299948506,
 0.5958740344203205,0.9870517411775568,0.9549784654449014,
 0.014214852954142487,0.5508835827053604,0.11408618672452164,
 0.7481681310953323,0.8527922120016594,0.2699173027579189,0.04881203348818175,
 0.899097176646416,0.4106070166335899,0.6891040852838438,0.8388683949566105,
 0.38730878820507775,0.15407950038701357,0.5535846767938429,
 0.10777970520455693,0.07107741833634007,0.9432773358265767,0.798143833635582,
 0.5344117940736354,0.550127341495504,0.0608663034782716,0.8957046677598103,
 0.015061378551980198,0.46831543363530626,0.9872108630827732,
 0.18395885057678651,0.42896854492359293,0.4997547645196565,
 0.3409634889673474,0.030165364803512174,0.14554605515557728,
 0.9245583226851022,0.7794541463085914,0.4749337717304335,0.8167613902069276,
 0.40942195115153845,0.07338690312853924,0.8016791811316634,
 0.9286125564832559,0.9339123301369938,0.6976048415993382,0.5763790005227738,
 0.8175809173037399,0.7011815954667022,0.3668396450990127,
 0.012249281207707852,0.9940054008214155,0.6223559808112398,
 0.48414608866183273,0.40289512817076023,0.4350692018185802,
 0.5158654338602322,0.46924595631234967,0.3533950393135079,0.6168375587460978,
 0.4987909404702645,0.7972460959323086,0.16624256294652673,0.2522208084890819,
 0.04266767511982161,0.02210574465899673,0.06116355968078202,
 0.7772215601513944,0.7943977453622678,0.706521321118297,0.1990337422209869,
 0.039034512255274434,0.7767069171519851,0.6084241794570572,
 0.3924675767421314,0.21945910198004293,0.23369442522304973,
 0.27742539791342014,0.19587958812488904,0.6839066838715941,
 0.22683764935761264,0.023656326282127216,0.1663961658412273,
 0.7019061584697802,0.11603333394384219,0.4627389525131015,0.6634205317697343,
 0.022694484846857876,0.7556084154104221,0.8204222349609219,
 0.9453972380280868,0.36131111451901254,0.30855598376157245,
 0.4317747328145206,0.9812296255287315,0.0046367407381043835,
 0.44440863179457113,0.6743658136822906,0.5192549420597101,0.849701774247401,
 0.6206909174815198,0.9414783785107212,0.5100751810881011,0.18392975953129687,
 0.08508929161902423,0.5858936353307257,0.7236648822740787,
 0.02116793619061097,0.6331345514800436,0.8913305191620772,0.5236822887437553,
 0.1531357245542222,0.725836340983256,0.44107426193000365,0.7040560558187882,
 0.946928083579326,0.7660149783831853,0.48999559463515996,0.8383269128109452,
 0.8629386953619433,0.3731279336541793,0.22283013668621265,0.2004524386644998,
 0.5222091407344236,0.8881840747920768,0.6207546051250534,
 0.024006191522626397,0.5101045183092395,0.5763322211822381,
 0.5064096643991199,0.24536489431102826,0.12984726441935812,
 0.5681883680813662,0.0200334962069042,0.7608604555978089,0.2559090588191768,
 0.5811165563520476,0.7810965864243599,0.9102974026544882,0.5138783655286808,
 0.5395681076891765,0.31619459192777266,0.815499713060433,0.4511552497664493,
 0.9199407039358872,0.6309412952921505,0.024871142289581982,0.859142201272137,
 0.36591845153024716,0.7884329497204756,0.1063646156492597,
 0.31484924453690133,0.75427477763273,0.41788569070879944,0.16257539347891126,
 0.03240293463325117,0.16639952058955798,0.36121647201450235,
 0.22349821124080815,0.6896535402896347,0.8422483543083455,
 0.48343963225555153,0.40538722646881675,0.4251323725430396,
 0.7454558680931563,0.6451319943173653,0.4888445281856263,0.04495177550484619,
 0.08454669437314155,0.4711078444689981,0.37506502080385107,
 0.03697609990440831,0.05513317738364465,0.009971939454727852,
 0.45264543688344705,0.16472377784445313,0.3768736243299067,
 0.07517737895505316,0.6726672501296064,0.5097203585401979,0.6284646230819968,
 0.6595914734306725,0.3979212894568852,0.7971476990736752,0.8632819310995438,
 0.5177995040775042,0.6926825350932817,0.5921243514414318,0.8929349290309501,
 0.34139682337008437,0.6606869491860081,0.05229614387003545,
 0.4792967279174094,0.468484300846966,0.33412744814209105,0.5740102500259596,
 0.39605998114392305,0.18665140896115795,0.42275104769368066,
 0.04975761224003572,0.8702017349459457,0.39928394856628324,
 0.3857487809923298,0.3281363614106493,0.6589396631978743,0.8971167462427198,
 0.6624756191605127,0.3375625196985489,0.3154347772764473,0.7942086589607296,
 0.052590570091088584,0.9604130433531759,0.3577669139009375,
 0.5649374929999357,0.3879063972068778,0.6623269651246242,0.3859776055049259,
 0.3330464958916729,0.9599408108310818,0.3972844455355612,0.4460249134847829,
 0.6527109706088369,0.14296394080849728,0.876678972064965,0.9426453675489583,
 0.9299180315363281,0.7059283962904888,0.20337157895382296,0.9287831423258871,
 0.2603094776882555,0.34991699219513006,0.2509269285227905,
 0.30963768554797555,0.32790680730719957,0.8232717033739401,
 0.44205184600870195,0.3776233186155815,0.6908863802284352,0.689514895146706,
 0.016149716286366633,0.13376423507816593,0.6447371649548972,
 0.43287464209815163,0.470973622252316,0.963233453913982,0.42625501839926394,
 0.9654029889527046,0.396375165992053,0.44183678349828526,0.987986435454145,
 0.057171500351286375,0.2794736276738923,0.4184982889341736,
 0.43604841442757136,0.02112657136637952,0.6137385132462136,
 0.4354915577477638,0.19045274501537413,0.2155360346109818,
 0.00614278016893377,0.0055368254443094456,0.7419375318561899,
 0.0873348453394237,0.8750213562638707,0.30304567359077916,0.8155888781749291,
 0.8596088635466018,0.5336995832448049,0.43772083274132045,0.336141351706265,
 0.47510160937721957,0.3880574873752365,0.5562494798194801,
 0.32013899196200657,0.3238239368734145,0.6611436610283036,
 0.32692715882066936,0.10902040654734091,0.4694370864855384,
 0.0020852103941455535,0.9995004853276296,0.5368873912587017,
 0.28391404733483416,0.2485292208187585,0.1883672890998579,
 0.10447377358381083,0.7411852253805885,0.006530896318098001,
 0.16597564030981748,0.12460521428475146,0.5120660019307156,
 0.6328851758004552,0.44522918242854415,0.556594976234605,0.4541992501871803,
 0.5176012550791507,0.7362835231691617,0.4003771703184029,0.15356813641810563,
 0.5068654821315328,0.7600435200676118,0.8477060549467113,0.0465025411812966,
 0.29315673148598365,0.8573865185433183,0.024358163249534437,
 0.5447764265773318,0.7832048427780788,0.32030377974736723,0.3654514824241425,
 0.037203847825394876,0.8159088859762084,0.1788064308001226,
 0.8648416972091724,0.5497356919409323,0.6626306635979914,0.614415782210444,
 0.7362238918087755,0.5188639293708351,0.31534097669649563,0.5980136945287305,
 0.6807620654885435,0.12973890399489374,0.6009734669236887,0.1477777752632664,
 0.5610996097781077,0.5296636681001987,0.8255201648809997,0.973562078892332,
 0.1177631751244127,0.2802534867465396,0.6103662524201935,0.5404830996402494,
 0.2751924903089884,0.7159637607961217,0.17777071584407134,0.9200415174155172,
 0.0799847853325315,0.13566593476459032,0.8641890574457122,0.1006446033035796,
 0.07364608207876511,0.7418937434060078,0.9700030675045997,0.3173569954574562,
 0.20599829867070218,0.6753659910567684,0.23722168018774736,
 0.3435805781499356,0.8011264885242124,0.8105410714778812,0.8695415018005811,
 0.23111760048685548,0.3384914022163149,0.9563232808589084,
 0.03916364183785792,0.0837342843132719,0.943427876352501,0.42670089916758724,
 0.6284338978746442,0.891444345131179,0.761083330522263,0.7085800745758357,
 0.8873425575697188,0.35179573237761697,0.7731029810054488,0.9287698654805143,
 0.3085806373369203,0.9166392680925936,0.32134484022847487,0.2354618961711945,
 0.5508512752343855,0.3087184284082295,0.011541214027644298,
 0.47942526249977546,0.7779252460299524,0.8625166154557495,0.8675474114383651,
 0.8787270189845557,0.8035228180517532,0.8256377531080542,0.12386649975090802,
 0.8177493486836336,0.08234128205680025,0.8847596304105492,0.8767074783810196,
 0.12880899441373128,0.3528119222756396,0.9645264419315527,0.4417999972343998,
 0.8560812573784126,0.7047566545044281,0.8999322821827233,0.7968274741200828,
 0.025069180806590707,0.4128780904752414,0.5202170608532484,
 0.0637234233993127,0.34507854757572765,0.1276055356492234,
 0.41627954590683847,0.7134607208727816,0.4446919627948829,0.5481686707452532,
 0.04994841469157851,0.34932524856568337,0.49898201361027417,
 0.5923092554283557,0.7370469155527604,0.4141989322182569,0.7128035040118772,
 0.8158249794909394,0.4197625860161106,0.08529904462666438,0.2761941830778796,
 0.657006548317957,0.9524695787719407,0.38568011798465174,0.8847411697583896,
 0.007351178264252689,0.3550888164432239,0.3801350261534565,
 0.7156635314833005,0.321860741847366,0.2102561784101371,0.6800080753734533,
 0.1447556138261813,0.22740954557425785,0.020146009893448502,
 0.1029105540762838,0.6783246672975081,0.5888010416876202,0.9392738364849468,
 0.8586619186553928,0.2238135225359329,0.6700595892394342,
 0.059351827443556826,0.6425522265017234,0.7331727434528652,
 0.10864966758795469,0.7177655005502548,0.07348341239619294,
 0.8541677826637896,0.0816351407943372,0.6025644893274437,0.7684203458092741,
 0.7462970894999111,0.6675180522153504,0.8559184454452697,0.08287546858618822,
 0.7325983989746448,0.450286653840379,0.49211871540546226,0.4396059270628402,
 0.6159216636561918,0.1573979989044656,0.40928778976355185,
 0.007719482435520675,0.06174257188728072,0.4743635886961124,
 0.212663158256454,0.018341216546240657,0.5544277413897561,0.776159774972643,
 0.5312832244350375,0.9225800562395523,0.926913327969134,0.140338685754115,
 0.23037480047889203,0.41152480664097346,0.793900014480809,0.3584613767500848,
 0.364136679979988,0.8841964684202835,0.36604770575812473,0.7403365850876923,
 0.18846154279304306,0.884688644306606,0.12947827960251668,0.534855665918893,
 0.023005390210125487,0.22211760767801292,0.8399442760431375,
 0.8113620996137766,0.10573401609674427,0.7580189773342253,
 0.13186945127315852,0.6561646368961922,0.7705159489002837,0.9460825421408435,
 0.1687296975988839,0.3218499794083647,0.4963602905306317,0.8696518047119547,
 0.25104517249384584,0.3184250507766576,0.6219692911242666,0.6708127231790726,
 0.7616135080805086,0.25806651831721095,0.7010112804351392,
 0.25901954864248133,0.7597890800190317,0.6951989177920684,
 0.07347879356630382,0.9076175524357175,0.584718315267747,0.3781614083696633,
 0.5207964576222828,0.8881797200295646,0.6837388650148628,0.4086459345941087,
 0.07659691530372725,0.10015299887794693,0.45412898515538624,
 0.0749507540822002,0.4929804426585115,0.7756499522983634,0.3543319523100432,
 0.23751441890253944,6.73274683015479e-4,0.4179918895440736,
 0.7468996401284503,0.9821711493722253,0.8350658290536954,0.08545296992858364,
 0.983554027182446,0.5213372192950547,0.09847101446496487,0.8560754538034251,
 0.47546746080462976,0.18335398271184855,0.4216772043470701,
 0.2804266711864243,0.9100131039308184,0.1034675538619505,0.6286295450175017,
 0.9790968612241706,0.006741014873528628,0.6169945228468431,
 0.6732794492610068,0.5308597837884672,0.7686397279039398,0.4234790344875282,
 0.4242040948621808,0.9400805062647608,0.17063078165267753,
 0.09109076744383215,0.42008305673450375,0.9351187268977939,
 0.4255136725993556,0.6373938575744057,0.7532368303489292,0.21924230325011684,
 0.9441798229746406,0.7888366982820187,0.41511964042162,0.7629283482218581,
 0.8325934818373587,0.9902495188176598,0.4436951570432909,0.2942663338559739,
 0.7052593681402632,0.6759404574804814,0.3731679169826785,0.761500946126699,
 0.7379918003978986,0.7128039971646638,0.025259886268022003,
 0.0679377300386994,0.3438587115411139,0.46007239631254804,
 0.10154177962260291,0.15227790765083138,0.9214748033853063,
 0.7904305222532176,0.9996010390724029,0.368901520990805,0.018871690663659124,
 0.811846061223066,0.3108695765896736,0.03910413978425553,0.3903029487573939,
 0.0737107491487643,0.1489532569992228,0.3646668129624976,0.02862623466811931,
 0.7233674397744221,0.15559410672650575,0.5668805802418988,
 0.49501861951640935,0.881123302285742,0.36557331547462246,0.9702106577790593,
 0.11395578321578836,0.6998963575458446,0.006761883810353031,
 0.9296253276787565,0.028831003802709643,0.7719617537264934,
 0.10201650511804738,0.9052546885223941,0.1706204849764813,0.9934827612552809,
 0.05648711780641602,0.13409699458044844,0.38952698434992505,
 0.6147329321920771,0.7389363717075099,0.8173516751632385,0.3820276156413974,
 0.3291055921419481,0.8133063116544302,0.0749423575452991,0.21916434661592943,
 0.8495226937254207,0.3090048738793403,0.9047355388464074,0.3414247283971623,
 0.24119435036443404,0.6647674679848173,0.9024318247405544,0.6991471897732331,
 0.15196510475254266,0.3604953446189736,0.779131478883116,0.47475781181799503,
 0.4290489689673107,0.1617557869896986,0.8281896509896062,0.9409371028286743,
 0.10380552362267581,0.16265126561276788,0.09817755461985489,
 0.7870628784865493,0.6901127110112042,0.035683131447071026,
 0.21029711236118476,0.33624242256648296,0.016442126522699718,
 0.24362006421238558,0.41074626935450187,0.6284636255980001,
 0.43465336277096644,0.7504026792107676,0.5191530422816547,0.5590894774154103,
 0.9123173440754206,0.09661623053348634,0.20741772541174497,
 0.1075725174512896,0.9538187819294794,0.7020339401080076,0.726451632806845,
 0.3834317891055573,0.24061012416792038,0.5941489659411262,
 0.36841131237707003,0.24709931128046736,0.3520534903806166,
 0.08436575760231946,0.6364631875880508,0.10220127853879002,
 0.34905302339820565,0.8692476518192324,0.8390298268589914,0.501379004053307,
 0.7464655141533522,0.9925048584150576,0.9414033312778008,0.8552423317921867,
 0.5459969721035489,0.18050659501268285,0.13024957070395538,
 0.07619767022399337,0.26754756214264797,0.37774243611696656,
 0.20793245441635877,0.7222469335541065,0.6257582143720091,
 0.37667647438625373,0.07483334899263294,0.6135981986584542,
 0.8303607230720956,0.5605517295644451,0.4618427452665539,0.7227978854678188,
 0.27376345402601576,0.024563899900568065,0.7656221589083124,
 0.7315511285922711,0.9949432082353308,0.315678433262791,0.1074585896465885,
 0.7367914604061072,0.5732827513172819,0.8137373206406643,0.1297905503225113,
 0.38432381032560337,0.7300116341221748,0.7159364235303617,
 0.16077935169769475,0.8854931963198414,0.4525438037431233,
 0.14315289482898863,0.14987154573589256,0.8958416576873385,
 0.1907327144752945,0.8411737935783462,0.9265451018792346,0.38331173939708196,
 0.006555149325142895,0.3955753400709988,0.16738130935028006,
 0.531381933295525,0.2175937441374466,0.2644434884243463,0.023974621450803535,
 0.8080999917725689,0.5332154133408431,0.2785584419965328,0.07158082487655681,
 0.921484902551508,0.9752327685602996,0.09371780553192877,0.7473421423015997,
 0.6514923492087338,0.5483401350679314,0.5139490507649768,0.6875992278890866,
 0.8963131494518078,0.14525007577114457,0.22898249463879006,
 0.8727594965589911,0.1446445487676693,0.37259758564568013,0.89084339725756,
 0.8113104396735462,0.6761015940857237,0.7653647203482273,0.7540725417816653,
 0.984717455823033,0.4929163592136737,0.49386399640580425,0.6086043972467969,
 0.8398980171201881,0.3045488246291388,0.8215169490554332,0.6402063089315428,
 0.5248791382396306,0.25810112464467794,0.39128410401268576,
 0.1971933980383761,0.34312929770371803,0.878742560490381,0.4239881290195967,
 0.6637883787689051,0.2961293601940762,0.6657030081134468,0.9972748829707476,
 0.5539489304615144,0.9372906618517121,0.21201792826903232,
 0.010467729583633623,0.8336046140055874,0.18372333799942742,
 0.16708162335075794,0.8057419875042808,0.7646912467993682,0.5992727613946759,
 0.025270161927676504,0.05893836115680351,0.30696292917626355,
 0.7796439932057152,0.912838213331252,0.35250692064403133,0.39648156284969915,
 0.8160807635579383,0.6614784615543493,0.25529832361479776,
 0.14426323923651752,0.8396328973350491,0.18072572591385505,
 0.8544279838709894,0.14336192713125762,0.6937325668206564,0.7421809725275104,
 0.34203017955333903,0.715902950681659,0.3470894704529417,0.4875037913626332,
 0.9864391255650351,0.09699225280302015,0.57231607654534,0.08935567551543788,
 0.3479356726769518,0.2200719343812333,0.42651467134448906,0.8756389234245857,
 0.9521261915343823,0.31160304100885505,0.02538762216156676,
 0.3976327296865594,0.7901223138959225,0.7514742957706577,0.2218240686805606,
 0.3093646987188834,0.6812899676591094,0.39068100473014666,
 0.30935380495150966,0.21667604910598148,0.11172176307666881,
 0.2511122336139351,0.5133042707684685,0.6621564721488111,0.47342142740603754,
 0.22012614201010638,0.3885586197568114,0.6969437227193007,0.938523044494953,
 0.886250294776265,0.3524844295755123,0.9441310848164388,0.5714481341860517,
 0.84096192430474,0.9968717356140194,0.04838380843169354,0.16714038893576322,
 0.1962173172121091,0.5922205569422956,0.5291601115284577,0.04433247321085454,
 0.11956390111547588,0.9974585225550817,0.9443085067030644,0.6133089776917304,
 0.7925453497107752,0.28225846916240016,0.8908885786922118,0.9000678870622828,
 0.32757747917470526,0.2936872668887699,0.06887839702191723,
 0.06428709844981007,0.7743381589295393,0.7256881626252991,0.7826609505676474,
 0.8388607879908792,0.07115229493043429,0.06544268323171565,
 0.7546739784216702,0.22412440063614625,0.7510476903049281,0.5429765846830092,
 0.6910477049857011,0.6305088297959973,0.29315916902380457,
 0.47670162943867267,0.1439532510134054,0.07175700220124515,0.349584123819926,
 0.4001889859540345,0.4885554391093425,0.05908359163599375,
 0.38445609613314025,0.27618250836379765,0.038185332571399,0.638043779153677,
 0.737260583572075,0.9165321902532293,0.44261732691755595,0.8920763618188333,
 0.7456683571822261,0.4363236413114704,0.04421557537392984,0.3454480418196548,
 0.5429200899097786,0.18475413960179288,0.7471272889562431,
 0.027507211644922425,0.9600688006361755,0.3626411087507657,0.282085414470885,
 0.11365591738972991,0.44649577019544917,0.6022807875151797,
 0.5596913554141332,0.09814113344405284,0.8217510174804605,0.4667290344047301,
 0.3823773328743053,0.8195107783271476,0.7200598844452255,0.5105499904888346,
 0.19689339989187993,0.1833199510699599,0.20531103640553383,
 0.004734428770957866,0.5625331881978999,0.7145332648305485,
 0.30916638145680864,0.5460242956015312,0.08598684431995418,
 0.5742539355935646,0.1206324589522958,0.8698858188958932,0.877438137604031,
 0.4955447482764177,0.7979473180090659,0.9124029864850475,0.4182765505845044,
 0.3347229603437579,0.6406610309868868,0.7969955713286785,
 0.007537017793160072,0.5775202926713172,0.40719899136237436,
 0.08910298101439817,0.9445848839448786,0.9413771047169885,0.3363444764799274,
 0.3787364026034039,0.10815152245348014,0.5130002538883357,0.6223812680411518,
 0.3751755457016358,0.7381962299883421,0.37078359727241794,
 0.10432480150812351,0.8480642227307221,0.36732211823911365,
 0.9122803675139478,0.416836746203907,0.8271455821918358,0.15696412869286835,
 0.6422816433299728,0.6805109755119485,0.16365673576708806,
 0.14885148345715726,0.7000007781545704,0.5643724543009118,0.2586432993288357,
 0.10092288532813976,0.17527726655296838,0.8653993588550998,
 0.2555618386770475,0.9141901365493194,0.5015332098202784,0.6593624708226967,
 0.5486020887619453,0.48044762756299475,0.5538045328784008,0.8913975584806003,
 0.028398556977737632,0.5026695372231856,0.1652442087325241,
 0.15419196061113483,0.2542104787030709,0.8268785558305956,0.8523995155397506,
 0.02328757487648636,0.19102116051193407,0.6883052488490512,
 0.5466921676026633,0.5549557916124647,0.454947979284703,0.05580166518191865,
 0.34934240855351184,0.8671738055921253,0.27333776817104793,
 0.2140730678585756,0.6623112515840963,0.5213087607361454,0.8646092336241799,
 0.39834881501725217,0.8459408266152331,0.45121751355085393,
 0.6295754643945095,0.9454262790888758,0.6069074663771847,0.3218118602385378,
 0.3394692578367233,0.4564629865808486,0.6911845036884195,0.9253013356384181,
 0.6608340792999314,0.8690514676935394,0.8706137688997996,0.8208390497875084,
 0.8327324436052765,0.24666688932147407,0.9703305183584311,0.8789302830006528,
 0.15608543036042766,0.5674858610035097,0.5463127476610725,0.8870119016896904,
 0.1678442691219375,0.1480394900387848,0.22478153597096429,0.8791191971513002,
 0.8366627338255479,0.8677113640016977,0.007726880713642359,
 0.3744885094085082,0.1810070747623217,0.3526730789643706,0.3611138161382421,
 0.7323354380911961,0.8654564418856451,0.14155051875432545,0.8037361251633466,
 0.9062329514117192,0.6661147063280791,0.012138574028298321,
 0.6221374040292214,0.14851041987802072,0.6895802487821305,0.7386781013256207,
 0.7594431616588307,0.6697210987768636,0.23481333850793362,0.2996740081331657,
 0.7749175822657529,0.7118280652878336,0.5913817614778325,0.5784131352662673,
 0.5055951325300068,0.8667010234209227,0.35164900791634923,0.7075014087464309,
 0.9953643138001506,0.030338935370309073,0.8522084536904936,
 0.8195844378065298,0.5896050427302424,0.551280066658829,0.2418227743748963,
 0.16367754247978705,0.30230344986134317,0.1971110408193827,
 0.8029278375698404,0.9240177853029662,0.9551179896011782,0.6395794712767034,
 0.4404874255109277,0.10117257532544144,0.6504736733504871,0.9852673000354115,
 0.8445092130303077,0.27782720893244406,0.5127504669917883,0.4928787922790506,
 0.8302989955450139,0.20681853800921812,0.22339995692034376,
 0.8003317182024543,0.09093579371836846,0.2118090170025162,0.8832576400622858,
 0.5893235072486374,0.7215607998798517,0.06758925149681694,0.5136049487013399,
 0.7798728197154661,0.07352231029679135,0.291462232218497,0.5950654551866636,
 0.65258735556184,0.1979333996670144,0.09528357336586668,0.8638747803525718,
 0.3411048977811121,0.4304716864953998,0.8471611013992876,0.1675373035774732,
 0.5191280962072761,0.7488620307094651,0.6545262610806536,0.6121538713049526,
 0.5671059088959242,0.4362371955033041,0.8441765153972747,0.4779192787034331,
 0.9146462881955676,0.32406076666849726,0.2528987083712705,
 0.33862692287060525,0.7380138972699305,0.8299836531016704,0.5694562173255371,
 6.103126263033998e-4,0.5562075713868101,0.7756358848144456,
 0.25055029708152854,0.8103549263159463,0.7040768827919245,
 0.07613438138900885,0.26024489143312,0.09103573157409961,0.7418896323474016,
 0.32223686506801585,0.15495739576380985,0.38816816226368456,
 0.2299437796659115,0.02049162793606729,0.760443548861673,0.73442788725034,
 0.3597631864219496,0.3064557453632837,0.12941261488801337,0.6818276357890729,
 0.10185509394024372,0.7451391042512665,0.7396173682231516,0.7052514485978246,
 0.2075062210124019,0.6091557325992285,0.7129567988502095,0.6170910549422782,
 0.21771461484943266,0.23491982987670812,0.8294410773437703,
 0.9008336742843163,0.2771411536797094,0.7321121882273027,0.6619023063406381,
 0.5351475436775323,0.09016353185216208,0.5687435347598109,
 0.11219667509674269,0.22926662315100355,0.7556784146871349,
 0.0010272740701416927,0.4757029342625678,0.5342399966917749,
 0.5152885904282682,0.29048992218913405,0.9954427977997004,0.5926740075653406,
 0.140356032693162,0.6557948267668549,0.2803142743327205,0.7123175849678152,
 0.5850566821053499,0.6173703548510222,0.3407177182950162,0.20575523014671404,
 0.013073064111138177,0.8383004703102407,0.5645709705143556,
 0.36465824967051885,0.6053292232028568,0.5137132627696533,0.3698468099053407,
 0.6762277578235083,0.43101544475193065,0.7819258506775834,0.8955859751422024,
 0.7000295078917631,0.11105350121619939,0.4393245525386309,
 0.08665171197007582,0.2822581226628612,0.3233228934184057,0.6631705998314937,
 0.8912395684148074,0.22406231397745546,0.6457510151682166,0.6260212790094186,
 0.5339070544436093,0.7862475425158306,0.40159349407705414,0.8133326609357424,
 0.6870754249146613,0.13200559233779008,0.8797039946289666,0.1617177883887163,
 0.7776310800608898,0.8743392023520676,0.3214074008590664,0.8027516234760765,
 0.32537709079736676,0.16472383267543,0.3397511726103368,0.5177738110999784,
 0.8826981724568216,0.6928833415432897,0.26836400666279403,0.7331169872744406,
 0.1614492755519943,0.1623021274220654,0.5992893260153845,0.4859890972107329,
 0.43999786597956503,0.8956122849076334,0.782779586573394,0.9757876713197496,
 0.9020776374207912,0.8509263735620013,0.47172727978080875,0.9450058126930129,
 0.5706514809164286,0.9343407636780305,0.045477392940646344,
 0.7528905443282038,0.8761742583291987,0.874278095868202,0.37621013525405167,
 0.8485766169181311,0.7349369407497,0.5054861856012572,0.2791485719418566,
 0.8832110353866992,0.19478229712613104,0.6174812015989823,0.9120964963343539,
 0.4367172873892078,0.32811422174620664,0.676776928845134,0.9825849662302043,
 0.7652186480469451,0.9632931604202852,0.8630973916437737,0.8875972905981007,
 0.6636737650413229,0.16459336658773993,0.32022973533453436,
 0.11942326602358067,0.16349524434543228,0.5908279436294916,
 0.04530447291119577,0.3669670054967562,0.0683837446832668,0.1198446231251411,
 0.9095023233125511,0.9091423507034487,0.7667226174955817,0.0939683356788158,
 0.39707853362726964,0.22743480370536506,0.6128670024684522,
 0.7145028103376172,0.5795663589930836,0.5422229975760562,0.4871731817607833,
 0.1305030342772937,0.9941495223097965,0.36154833238692385,0.6506625079734257,
 0.8758517151118321,0.19995447762421803,0.46435333467751516,
 0.8596848625331998,0.8284114361700698,0.7413286821522617,0.391503061842284,
 0.4395240423318628,0.562525721851155,0.4682783181573783,0.9048385769678123,
 0.76463850216348,0.5322052536366585,0.8519906661399679,0.9849772861705992,
 0.5839834439609266,0.30524888530009164,0.2360827887722472,0.6357384320001194,
 0.7419200828514688,0.8588208462816704,0.5954768958055372,0.08084043072692892,
 0.3686992559009289,0.11694872720094596,0.27968666111759444,
 0.6804966667447321,0.9246819265350771,0.9975247370093664,0.8800077991347062,
 0.8678320267599575,0.2268603303600938,0.12415777388666283,0.7241892556853015,
 0.044505247563673356,0.09528866087801102,0.32398879879800146,
 0.34394762329855777,0.43587809915220677,0.6282898407854259,
 0.6345313573323329,0.7914341193289256,0.7335561607827673,0.5652223801795982,
 0.17453972484858826,0.5679629773142577,0.44618619106803936,
 0.8054524386381616,0.07971819435074301,0.545953631833917,0.39084791640323524,
 0.7084957441088874,0.11322897992964853,0.023021303187008013,
 0.920796198439439,0.2586176987578511,0.9031376815446003,0.9012366012126449,
 0.5545071082171351,0.8322394978567421,0.2776175829247497,0.3805404930016105,
 0.9032214973113418,0.43988612960398665,0.5385031593342748,0.9684255951686875,
 0.21315836316928338,0.15240178324409281,0.5554054323570886,
 0.5528278143744038,0.23986462600755099,0.33611742827594937,
 0.1451631205466145,0.36566665698030654,0.9368122765624125,0.7127039284184489,
 0.3667999646097426,0.319043226929002,0.8743891071431903,0.3101710520946086,
 0.34268791634279316,0.30120205384336307,0.4867805303177891,
 0.5273233916407726,0.8296482544038343,0.058392024093947836,
 0.33592348330936583,0.293734635409324,0.7884749537441159,0.2597252950156199,
 0.02446803700180822,0.9314998770282712,0.4262418875733791,0.8779058673044007,
 0.06579299652665105,0.4513680576403727,0.8716219188523817,0.8375621434348489,
 0.8921883799596461,0.14173117726399043,0.008195785327881677,
 0.21012148363867988,0.14476044041414915,0.8212800025325286,
 0.2665102726230355,0.9401548011546489,0.6214341083117823,0.8839732922857381,
 0.027902935805444118,0.4597606229234742,0.4453923912375336,
 0.8212663936627079,0.007499851535417323,0.741214835174338,0.6844387520466793,
 0.7751712190932669,0.7579396809923127,0.3516326880143108,0.47556776532390377,
 0.37149258520284967,0.39928243614742,0.1674299733621306,0.9222704906382645,
 0.5683064520946708,0.7070161451176539,0.5265158056850942,0.988109364480325,
 0.5556178636257783,0.720437995220187,0.8085791390133896,0.2690575608582684,
 0.7452894770537607,0.570217485412474,0.23157036692533017,0.4462810022107272,
 0.587447098760966,0.5812115737514876,0.9675615334546916,0.12224155205426779,
 0.9516529452829962,0.9582447274125663,0.4510479797097976,0.10970606690115514,
 0.4475008360346757,0.8809479034361203,0.17945634654792486,
 0.45846070507594894,0.5335678481186568,0.3272327695870082,0.1813661566570507,
 0.40662457295384424,0.36113231155818015,0.33775887785221137,
 0.31559937389308057,0.7327704702045433,0.08912340400519048,
 0.9352192229377306,0.45218661267682847,0.8508146134082964,0.6312880729201765,
 0.32937750610395167,0.20670916549200602,0.9955104289162306,0.430165352519709,
 0.1131696184402331,0.2167472435689619,0.5441822024147189,0.8115544663020754,
 0.7615788377007552,0.02731055688844486,0.5722523321317394,0.7300887791842556,
 0.6835354575558465,0.3865661449794231,0.6089219377417296,0.2813844715890855,
 0.46603079924537394,0.7118538322791933,0.3231868317456196,
 0.36507540161809415,0.592189495908344,0.1736797219855365,0.3901341602759718,
 0.1467550539852156,0.13451335834116018,0.4979248739384097,0.6179760410641505,
 0.44344300438860074,0.6861423165249747,0.8918837335357278,0.1637192408656346,
 0.6301064449428576,0.18903540409785025,0.9382209954006595,0.7831953656934256,
 0.7149725710645135,0.22468917829228363,0.6416165177704677,0.8315394440823395,
 0.6271076511824293,0.2798560653359813,0.9002072013680156,0.9354026947897306,
 0.5656328345716141,0.06952765199678101,0.8466237046081088,
 0.051108513763313046,0.48146003471577803,0.846240987200456,
 0.9279391083788759,0.018487552355258785,0.7434014003040195,
 0.08069244922075469,0.6645500271339094,0.3322327392771507,0.2345447114947239,
 0.781438127745361,0.42284371931679954,0.10941408890847348,0.7666860524778515,
 0.0817895549698755,0.008906763103948445,0.587570726849458,
 0.12536354767991365,0.430969962070059,0.8369271735362676,0.8117769415088946,
 0.7689811695684721,0.8312074708630517,0.899234224737437,0.31355083334258393,
 0.47876279251448706,0.3728566954843715,0.08752052232574437,
 0.8929987784401081,0.8950648579225076,0.15256766810760958,0.902790300702548,
 0.07378426021978501,0.5412488757151216,0.8256482670791421,0.0923539211769846,
 0.049050720221578636,0.41728930766439554,0.04807086741581612,
 0.06383443772709652,0.6248945240911905,0.8990556362487058,0.2135807272602528,
 0.9037389037029937,0.5509963402646081,0.8356483840465139,0.8956351821406133,
 0.03488826934719924,0.5657704156466511,0.5390730107364325,0.6148095828758361,
 0.26936982648223606,0.0891247873936194,0.8445880287017498,0.9486251743587727,
 0.12009021055923008,0.5821741744717529,0.1961576589089825,0.8260084762765761,
 0.5445751528015645,0.03603983345752593,0.4670145005202584,0.2475480926301864,
 0.914876265198666,0.9435648969339938,0.8039139810845211,0.4084159606584058,
 0.9333360629358949,0.622106988172464,0.5048158073767038,0.5060528728090904,
 0.7270845694712751,0.02927933198358923,0.9600491390792899,0.2847805653168993,
 0.967535936845022,0.4498868063851107,0.5855206615990056,0.6544216948183337,
 0.8608687997195354,0.3853774063980817,0.1569020074708234,
 0.020271153425128974,0.7870518769791139,0.36570983821766,0.6730404354323382,
 0.9129452043447913,0.18055120530570423,0.6094527847470379,0.3133934026571321,
 0.787216624483302,0.34433130829993985,0.15874513599174778,0.331910223476076,
 0.8283201287651376,0.8681393469961326,0.32929843368338074,0.917264878960878,
 0.3585795858737505,0.631035054781578,0.340386968644912,0.10631584867357557,
 0.2259466486840883,0.5244287938542871,0.5020602333728024,0.9827433953099196,
 0.8820907500587559,0.38520343388005773,0.8419454851432872,0.6965133154257097,
 0.8029216643714899,0.21358571823566797,0.8515002432432178,0.9743224166793313,
 0.7065062721209494,0.8632368958235025,0.05790059680483417,0.4217950831933842,
 0.3760994447342232,0.21026399044693855,0.1786878976433387,0.9999224099472391,
 0.7594581922024994,0.48984273028505354,0.870499496652393,0.32604771492382745,
 0.5668157295828291,0.2802953828198179,0.33782016167552475,0.9012451488694109,
 0.31200444591421883,0.3172474363181029,0.8063648429143718,0.7186131117900071,
 0.8243750148601697,0.13825226461593765,0.48642802086610126,
 0.8506340236811636,0.6406099879506677,0.331833606858433,0.7004680837386115,
 0.08450402244575272,0.3220110359808521,0.11436922629594881,
 0.8487909214991624,0.22278389066312287,0.751930200287736,0.563836658812427,
 0.21329664013393423,0.4119027763098345,0.4769379451747029,0.9125178595646428,
 0.9079958709534186,0.11262912867731956,0.31550852758447956,0.77957326334613,
 0.4850047683226709,0.9251846765469597,0.7481907167366968,0.7813729785636054,
 0.18870493235080799,0.9826439248958958,0.9656211562135028,0.9834469797092691,
 0.8433945220904315,0.28951167600456773,0.553006601882544,0.8797174003297505,
 0.3204362879243535,0.7143624301379267,0.4905769686440753,0.6146554142040176,
 0.4691243622905299,0.9664550349514596,0.20053905238900294,
 0.19645478947298134,0.1194948824777573,0.8144385169378179,0.6909070443270904,
 0.03556687679354509,0.5792842804385305,0.053940449205832075,
 0.8299856518532656,0.026046308415604735,0.3429588054491678,
 0.29455772372081956,0.9772741101988902,0.06600853617905289,
 0.4636998808649192,0.4853123957022142,0.24507945359970096,0.817942283484514,
 0.7532313521917648,0.4067528632509887,0.444060515601131,0.48781001031226146,
 0.8963864046509885,0.6049914317874219,0.21253018626154185,
 0.07786560377263463,0.2782890268886139,0.733347976180901,0.9910786155753256,
 0.4675826452168318,0.5896014928867841,0.08880506420044121,0.4694182861159837,
 0.022418977295539322,0.7684101536593556,0.4905554400118004,
 0.5814193457487716,0.19495769278393815,0.4068986230373056,0.7960240913825154,
 0.4040968374674442,0.8860594157872222,0.12447386037620278,0.9857281836509505,
 0.2997826736405702,0.374446113528101,0.8794289625832632,0.7577558670279623,
 0.8909617300099286,0.23807301903508216,0.9353801411070392,0.9871837925009512,
 0.16699387923330766,0.6144405160151934,0.9355101275205868,0.5400953159904219,
 0.4134154993145598,0.6745361716260762,0.9907685107051938,0.10416692595961607,
 0.318467726678507,0.322620175305956,0.7261948685337831,0.7335407826458332,
 0.7529384946465392,0.9925259067502328,0.7137993432162772,0.5485550960546091,
 0.12422691658532137,0.1724691285775507,0.09299740659113376,
 0.9797054688896131,0.305132757462544,0.2502458116547306,0.8370594809571632,
 0.8762481586892982,0.7431197213940415,0.08003224276534981,0.9916695190514818,
 0.902173421558578,0.9378127752185815,0.33185442875090154,0.28456802253649993,
 0.8920390569196971,0.8828438107596002,0.8707874437343378,0.26752430779262193,
 0.51567343040243,0.04238468770766579,0.09210062089713766,0.06592430319131737,
 0.048548695936045494,0.10626301478874745,0.13828150281482776,
 0.09350668215443769,0.4628283091287557,0.6440175506900522,0.2931152030584173,
 0.01705005697731965,0.16359787497954337,0.013002208117105951,
 0.8445138246167597,0.6464275269704166,0.9159550065520791,0.19496771153192982,
 0.46611336982746,0.741732377385367,0.4556102686199536,0.5464084920413694,
 0.40114524246821115,0.9434306406267114,0.5833534808219413,
 0.15156651959768896,0.019358760106689687,0.07250545862941671,
 0.5201553434428564,0.7189059017723708,0.9778874537365121,0.8696431413438688,
 0.604398259949351,0.1367029621472925,0.014823410872029985,
 0.47475847802743254,0.7631330403593869,0.17862295596392863,
 0.8155591756794492,0.5008450680751443,0.9485556200247882,0.19189595887224797,
 0.7946516419647428,0.8224077374941675,0.9807291911127081,0.6877583284188042,
 0.7539750294054295,0.7128361055251857,0.47110921374564163,0.906149123442159,
 0.8218636870093621,0.32588487207123196,0.024757571095069153,
 0.49431579247420177,0.3985160558662747,0.37304843864182546,
 0.48951872696511567,0.8577503388627866,0.2590410690368803,0.7479678711395359,
 0.8367335476101717,0.3426777203197736,0.8423601301293018,0.31043786852988675,
 0.9532372149116166,0.8252768256861274,0.6231420273666646,0.8896023673239876,
 0.19643554970926524,0.5053622710987241,0.9277492821270024,0.6714797887273293,
 0.2800160534543318,0.8722791310977973,0.8395734359153045,0.6086615317447099,
 0.1288636283659238,0.07596141973873838,0.4649078620971926,
 0.26659194745113224,0.7576895007688733,0.369817733854544,0.676781266225691,
 0.17174453405620693,0.984126755942092,0.5320883454145857,0.3816632385966404,
 0.007330964195141898,0.8609205544083823,0.8851716181285008,
 0.8277331685962496,0.12172021657575871,0.06848152088675974,
 0.20698815105432566,0.24506996717274776,0.30839303264706,0.4728960386244613,
 0.8903968774192902,0.2652999753460693,0.9904087765486038,0.7714008046956965,
 0.4494018937068889,0.49408667965131536,0.4943577342011486,
 9.051777825963558e-5,0.9775528675668683,0.969492046731538,0.8590928237971798,
 0.36097708210056334,0.49116504490502244,0.07816581840813086,
 0.5467077362811288,0.4004145067176549,0.5710987019246453,0.4533546222501106,
 0.4932263036010478,0.9266799352055941,0.41954171449567923,0.4026336498521892,
 0.05625459429290691,0.27878570825066706,0.8437158901445868,
 0.3073331121882509,0.7080345967922472,0.770018199891019,0.07939949231801069,
 0.986452190804245,0.9723489389297653,0.9593017437948579,0.12771959720935122,
 0.3325519220133555,0.8030663458072049,0.5562018636478951,0.14099504677106778,
 0.420929566290847,0.5889137952477932,0.03515074800549023,
 0.017775270041491043,0.6395100279972287,0.655061301334228,0.3404995225492288,
 0.022806719757115457,0.5490366229329009,0.11893939852391744,
 0.8388834121444799,0.3147007136747557,0.6472327556019073,0.11217508866359605,
 0.9685340518428022,0.4815485163317893,0.6555192686557256,0.62455643075209,
 0.629621203266562,0.9995821830687855,0.537853026414929,0.2186757033277883,
 0.8900709201304856,0.06281028824439783,0.5680545177848412,0.9655308024688377,
 0.6110299383658111,0.4814774138662755,0.06278475814836482,0.9406845845075038,
 0.5064865821267046,0.9515842454898888,0.43677790477654466,0.5721170344860798,
 0.48985959918806676,0.7978658069021697,0.2109452769155653,
 0.30699900546637693,0.5910125112968687,0.30699205179243017,
 0.4815038016587353,0.08630856867248904,0.30252611834033316,
 0.9646167595069807,0.18757371417496915,0.5752984918526471,
 0.38907943212567697,0.8709093489181685,0.1444805008041159,0.9474144113940356,
 0.04915539046980222,0.8348828497917207,0.5792185652880663,0.3401249609369431,
 0.5185806247355077,0.7849806138717065,0.44167794276713745,0.8393692331182937,
 0.7234485818500235,0.39530449432051384,0.28937856017353203,
 0.47921007537349214,0.2910605400780937,0.8099944878663415,
 0.24462207639913358,0.40920423409539775,0.7200981966778177,
 0.6605931407042257,0.22119514340443358,0.23769619846250067,
 0.37709912731888673,0.5537237834937488,0.16021272576561085,
 0.8930845089489421,0.6009770397891122,0.7032215386688884,0.7872612092911417,
 0.15968493009477724,0.8793909699319074,0.8397999043042954,0.7230093598526105,
 0.18904208877147588,0.6805768568181555,0.9351421753133351,0.725493249938553,
 0.7660952240229121,0.871674693270013,0.14559403314823116,0.7875461021286407,
 0.39278709521100263,0.38768344428370716,0.20021981713543646,
 0.2942840412868585,0.5623870604434613,0.6977365103078937,0.8082676696667107,
 0.9706174578267779,0.09798700101630797,0.7743959743981491,0.8445617543664683,
 0.4737234862725008,0.7655351970996527,0.15621601624419057,0.9327058107672728,
 0.17126378625554306,0.172196819669441,0.6844600058478442,0.6039953816755096,
 0.246098085156635,0.8850599841676833,0.5487720372968375,0.3893055119277943,
 0.33034211911260336,0.9276947555556159,0.0562848538649835,0.7530721942459511,
 0.2793863218201246,0.7705781405023275,0.5721157430015951,0.7487860258955384,
 0.903897416930361,0.4666367478011091,0.08045680734493121,0.20924718928647634,
 0.4103851076478131,0.15109791282455998,0.671221656399555,0.9096076623492513,
 0.26449374574219,0.07679354390508641,0.6489244485194893,0.7372160685831957,
 0.6799839826112061,0.41412449043050914,0.9495288142126218,0.3559012276690985,
 0.9407870700608549,0.04822550992461372,0.40349641132246783,
 0.5789755248951627,0.4046622468768636,0.47350907627402816,0.7410337850815106,
 0.14211685352793801,0.15695834555203503,0.10803879533377769,
 0.14004570593013144,0.2567358875654637,0.08219815691052301,0.427289881195863,
 0.985718228421422,0.1828761188821425,0.29909636209549595,0.4920113783082589,
 0.2955398231467006,0.6708235283012527,0.9833855733361712,0.3196494555621705,
 0.5423645938137721,0.26850588864849934,0.08950827253057725,
 0.8802379865224116,0.7318396821831732,0.166346701956264,0.6619958927671529,
 0.543168505246347,0.961062095424085,0.8426791600348718,0.697048727514199,
 0.7905887958580566,0.030945368959562458,0.5691531940869486,0.867731308728339,
 0.2715784506743315,0.2056225869874022,0.13949695064706258,0.9400468278463298,
 0.8180132045684139,0.32136240439078645,0.400021835897451,0.565943742488185,
 0.6558069255416809,0.4550521510182436,0.8954024809295437,0.5311228605552824,
 0.06658388163158624,0.6527457167754783,0.2417366361251354,0.7247114356383287,
 0.6850177395766706,0.9771354419927459,0.5235564243035188,
 0.033232085906237274,0.9506871956661931,0.7057386993974624,
 0.2965564720859726,0.849150838793969,0.5056988690930482,0.9627983250348547,
 0.3382670825833003,0.03506426786459926,0.9227963952302541,0.4287983010099373,
 0.3597146297870494,0.5786618539189885,0.6347153513588568,0.6062659207253762,
 0.01976589662093753,0.34203083316305316,0.5252249560932121,
 0.7612099358375118,0.011942513545256772,0.23216603299001015,
 0.5035575160019461,0.3686013342657255,0.80003174697985,0.8403518201169922,
 0.13054377466763556,0.8776141065516426,0.5505191020920527,0.1035055435224157,
 0.5282819621708865,0.6469292497911836,0.6599466352577047,0.5113323568510463,
 0.21293863008793923,0.553507681441133,0.6931752238612668,0.740645416514916,
 0.14599233515341914,0.45535714462077403,0.26088634296307944,
 0.9924208966339731,0.6551696558175866,0.3325106929248627,0.40780819599051654,
 0.8575093864569578,0.8431458987279218,0.2853535403085612,0.6530169000420071,
 0.6932230435939823,0.4127688495652865,0.17966840297535835,0.5061257369860293,
 0.8340858016201578,0.8234314954781039,0.4037691082153583,0.6135522716410858,
 0.08333944230703993,0.20896492008966816,0.5310319067318002,
 0.28511637189511674,0.26129627652652243,0.4113266906243701,
 0.29515258743015804,0.6956459425239527,0.7367917459228811,0.6964427500066013,
 0.2178775251614734,0.6844446349906256,0.5357801732770604,0.04880176077995557,
 0.07142171060016667,0.8100211744548222,0.578678472075162,0.07953981222966489,
 0.43162394137602755,0.9948374214057666,0.7275269459510638,
 0.20080154669544403,0.6065443811875931,0.7408320837905068,
 0.44148228441672455,0.019965114024307895,0.4181943382609449,0.70918783568145,
 0.595957178945535,0.7102184212709681,0.10023958571432434,0.4294693638109496,
 0.11399628769142445,0.8231095284774421,0.570982098735907,0.4961051871446611,
 0.6533127907663214,0.10658738261406397,0.3176887300508654,0.9999537923883254,
 0.26050352155569134,0.21204866083883667,0.8495434598936276,
 0.9535297336953483,0.07755476432786568,0.2156267337090375,0.9545829709579642,
 0.9904691171004366,0.714456684227545,0.5185009028504304,0.7845974084465964,
 0.011485226100476531,0.3090681054075023,0.241297774679052,0.8179291376581891,
 0.27401056640434607,0.20087212629388473,0.5625041674880237,
 0.9610946567597373,0.26974826278770436,0.19863534992224585,
 0.9964337377482726,0.06360162760109811,0.7813672663581962,0.9218221009740785,
 0.19062693461406244,0.7223707942676102,0.4853254190240456,0.3284412862425584,
 0.8651023743775511,0.7961779996088947,0.20759361496888817,0.6556952540788622,
 0.7306916352555659,0.6933139581366662,0.009025356587968547,
 0.3114827328003459,0.5317267201752516,0.08292923712308531,
 0.19983198373493083,0.7053977844729146,0.14605633264656603,
 0.3452401754169202,0.41902310067457815,0.09530495062992239,
 0.7847615527105487,0.4963127520751698,0.13045732172690883,
 0.008000030152704007,0.7064291557621797,0.5554439647281589,
 0.9524782927258097,0.8182466810879728,0.3755241960266775,0.3136634523090204,
 0.6418906319417889,0.0644032636778119,0.7530176169089896,0.08093937331566614,
 0.05193285079631882,0.44427203890475253,0.3067437393354835,
 0.15382946747391146,0.816040740500421,0.9703818965965618,0.9871680396865083,
 0.7738153501312405,0.07617490943484106,0.45565542797104164,
 0.1416325491197168,0.5136087907454473,0.8606104030809756,0.721120638830479,
 0.7442630080639959,0.7761747134913045,0.3062988422056754,0.10693739576198003,
 0.6234564314394397,0.3058363154137216,0.1099661514696153,0.23403848987453102,
 0.5472925712176822,0.8702881843362411,0.18065477348763004,0.4354102652026678,
 0.6573154668666595,0.5298866231112913,0.7912525769324916,0.8564348414036791,
 0.7036736790463252,0.18874066567512116,0.18993896294247092,
 0.6141944735118047,0.5992711694381675,0.723499430134572,0.6142625989274546,
 0.2531522885308607,0.3993499130807211,0.14029376749506983,0.4873225872117296,
 0.7608404615935375,0.532053190234331,0.8594241314955352,0.383562507701924,
 0.041025883631514826,0.09232872894729893,0.22095315825741935,
 0.06245844152817126,0.571870484540356,0.8063585932080685,0.9790970653344657,
 0.4547946834274501,0.9766738100066595,0.7358138325258424,0.6426774324681583,
 0.69785349638714,0.3459714441887468,0.07522726542981817,0.21674790208374228,
 0.5977912044372253,0.3051608622520965,0.48941289194009663,0.9610298572971566,
 0.9203276292879946,0.8857390054158693,0.7083317027220224,0.41493018939367976,
 0.20785426431562037,0.31273439208358433,0.020182592513943054,
 0.572862453220446,0.029020807483143396,0.737257704524939,0.29163302623802656,
 0.5667932858961549,0.5004897416078699,0.3228210521159356,0.2296834671353637,
 0.973732239126023,0.544523667569668,0.007901562961229902,0.19104861366281622,
 0.5359834743810269,0.46440063516095986,0.0808245109282959,0.5572815916112902,
 0.7649666798370182,0.24767635895096263,0.22906734150198926,
 0.5505402250376176,0.8736084421396448,0.15854022355076647,0.3773137122695429,
 0.9827520299535711,0.16872409265087385,0.03423289159367382,
 0.04009539835219389,0.7702249352600995,0.7649848502846431,0.5774576278836566,
 0.4394055253099506,0.4280546447685897,0.2973964013406498,0.15740342691733422,
 0.3627578004397468,0.897195062800749,0.08094740353388497,0.9989753958092251,
 0.928223511436643,0.0630650510008155,0.9010424749048985,0.11437756962984447,
 0.9915608475551937,0.5855386524007841,0.7523491059145283,0.4082261878021627,
 0.6242599340643652,0.8648823449798494,0.6338189343798499,0.489488831454276,
 0.6849752078303964,0.8400149050622805,0.4606521853271204,0.07608316625066314,
 0.21878753296727338,0.6562565609936848,0.5315599135279009,0.9869123407967919,
 0.28634632038928054,0.8823178594074971,0.03157410341359679,
 0.18466624846803859,0.6956802517162847,0.180019088663117,0.4414586552696155,
 0.05480925523802005,0.544606075013783,0.13784338065539203,0.4239305467504506,
 0.3431438113395149,0.2368929564082155,0.16907792413010825,
 0.21828767089376688,0.12484125672624469,0.14138811784624217,
 0.3078267099825678,0.643467469034168,0.25531073739187005,0.6866372516148329,
 0.6380498076268106,0.16708771353309715,0.30542347687402027,0.353003242586894,
 0.7318590909591391,0.5881450601905894,0.3554790647505237,0.16384287167481082,
 0.10723205435380889,0.6088555523101643,0.21570953578744279,
 0.048867662158531555,0.4745967319393256,0.8075314543802395,
 0.49332313735205346,0.11575093295539662,0.023617306574583496,
 0.7306504087671719,0.9675471616615221,0.8761173442454182,0.6488085333364979,
 0.5873963358235631,0.6117453215672054,0.09302756935091949,0.89757980491316,
 0.10221745131716775,0.43181195312767295,0.7768714505383799,
 0.8148012686142628,0.2501306731756201,0.29788853519852976,0.7904122711273239,
 0.3922772271725733,0.4782095182723104,0.603534680393874,0.1976616635633618,
 0.62589096884613,0.015159799532081464,0.2791307644156338,0.7115682088645721,
 0.9011858772257157,0.31086214379786603,0.9489963413125632,0.4923582632959471,
 0.4403230197199697,0.9796108333287297,0.3804964976474292,0.2965644269009431,
 0.8334443903259534,0.6022980841749557,0.3801920224715635,0.9782021934622133,
 0.7041509533897599,0.5226961313135401,0.7409597289880221,0.07972369483424702,
 0.804936296687571,0.6027148762413828,0.181837463104358,0.8314250585764091,
 0.866349469487786,0.41065827015615997,0.7784054391534787,0.4465655869801164,
 0.10049741556195779,0.2710666300160447,0.3350051300190001,
 0.26782526548178665,0.05684040579523986,0.034253941699223356,
 0.17677511690528258,0.6479842691272572,0.251989440327975,0.6845685425384501,
 0.1654423903743929,0.8238069944793469,0.519263701432594,0.6229803857098153,
 0.46252180358234485,0.26693209471950496,0.798552994104637,0.6007880405059234,
 0.2673617901147517,0.055007813937339445,0.07319630673627686,
 0.9340934876100726,0.6357359305562144,0.29475422505799165,
 0.49627057444365497,0.4626301330609366,0.3125237611859836,0.6454357232255288,
 0.8756441097334823,0.7791334484153896,0.9255269607471124,0.5503408607607602,
 0.2942278053771745,0.7355869850341133,0.08486531249501139,
 0.09560110467848659,0.5157556906142883,0.2517002223297906,
 0.07631586192090445,0.8005886522798428,0.7129109457311745,0.1743223003740051,
 0.33144303437191946,0.8814271442067085,0.04724882946694364,
 0.12342371677808672,0.4236986986422482,0.2800676005272207,
 0.017215865651715734,0.30507600771512633,0.6145817786798515,
 0.7900208564425241,0.569679056306085,0.23602664092610615,0.6446808990899309,
 0.0906670010239431,0.3611355692211059,0.4302458688157734,0.11107269702333189,
 0.7031430764043655,0.804075426507592,0.4787869910891508,0.5006045601485933,
 0.40778994267133195,0.05376559677198667,0.3343616692567123,
 0.42148893740731763,0.7314686450136106,0.9778956180500854,0.3266323482432144,
 0.416364495888618,0.21514167255579986,0.5078148628466915,0.7534092859183137,
 0.25186329009855113,0.7512974076264186,0.27433882931477294,
 0.1751202145715861,0.670172250854689,0.3115767447615513,0.23096998079460562,
 0.7928214969312766,0.5621840907056739,0.1734991899726872,0.40734695545758615,
 0.7000911753992257,0.267820415330413,0.25272365468908486,0.30808525625601957,
 0.9404974670891272,0.11277389029052476,0.26140767236292073,
 0.43394148057950943,0.08050824049810157,0.18675205478564694,
 0.23251483973813347,0.010854564156949742,0.9346809892389817,
 0.7937975928866375,0.12026638598874706,0.4039632325550634,0.6690132961373967,
 0.7602728054913157,0.2049380084484359,0.09710970653654272,
 0.20827180342272245,0.4688541191749136,0.02659699136259419,
 0.8933966924875503,0.8813284454840783,0.6817165638797772,
 0.025566740144904232,0.0035824908384236576,0.6927625553764349,
 0.6431192645592327,0.87320456102581,0.7636106395793028,0.48661302734204026,
 0.541738853306199,0.503511159353327,0.12383477662534859,0.4797298392417567,
 0.6871796530530456,0.614972860750785,0.9406771771640452,0.9445816730905948,
 0.8908028636376862,0.8155467891877116,0.6630768517719468,0.6303431573767132,
 0.1689276999099918,0.6502811409431182,0.05326711219210256,0.9330879680163384,
 0.8535040313401613,0.9602046027660243,0.7302501270900112,0.20780383944641212,
 0.4238692087976288,0.5762661788853168,0.9014035479642813,0.30697529577476446,
 0.10802982962636154,0.9896164739155521,0.45143035429562084,
 0.07040380663762136,0.12881754238002996,0.030494279452021766,
 0.3634706470066521,0.06580612397673213,0.25000956192292634,
 0.06459386471676898,0.28471471283111716,0.3315765571129574,
 0.5917053716630032,0.977057521195613,0.007384146867496111,
 0.10058350827470453,0.8028705552405375,0.9471537048961588,
 0.03603561551446155,0.5166677395946404,0.2840985205898384,0.2601594896997754,
 0.4784548274585412,0.06049523810878499,0.9413490937864055,0.7659458456349669,
 0.3181055680965622,0.8948265711312192,0.32331625636152095,
 0.43909864402416754,0.3875687845737428,0.43798916374631536,
 0.7278450283219007,0.11815285885085192,0.42207195716988166,
 0.053034161453918816,0.8374238624135253,0.09510291001329785,
 0.4581043974993193,0.5702426279000172,0.9351723137644501,0.27733314407906007,
 0.4945081466987119,0.6892884846223599,0.19050595661743008,
 0.19387182695149163,0.8466463781090966,0.8509635190734388,0.6230715940950493,
 0.4965054932453865,0.8285832281548507,0.019997923239843174,
 0.38125294110622243,0.528915988183272,0.031310162437397926,
 0.40625281199591456,0.4575279631634448,0.432874532695137,0.19640667692847236,
 0.11512535460596107,0.13873402110615451,0.07642283853116372,
 0.08803337589567317,0.5783732198304581,0.8875677247154103,
 0.013733440945389419,0.565741117341208,0.6874371450225074,0.5398122788388552,
 0.2062258528291696,0.959067511528056,0.31119421760216925,0.9883070194678423,
 0.933916930362699,0.8369898269182781,0.6053005239057863,0.7772201917330602,
 0.45785577927241083,0.4669588327168446,0.8218062038320914,0.3208970825252443,
 0.03316218432397344,0.9835893510471202,0.11822520215936994,
 0.9969292354494351,0.6530260457133054,0.24243077962347137,
 0.39247325441853287,0.9976367234644008,0.2331241594013711,
 0.11240240495651022,0.3863713359838591,0.14355247161773665,
 0.9351032403207709,0.31490862602081915,0.34459349415963825,
 0.6586091575058277,0.07770459022082266,0.9188564814252417,0.4299639331236603,
 0.35867411904629976,0.3489968503565629,0.16029620038397896,
 0.9615481539127986,0.13886905965914642,0.26974751943928554,
 0.8470883286619076,0.6040563351383832,0.4379589621149361,0.08298905567223813,
 0.19860896898661662,0.3041701973959442,0.07646307103104566,
 0.7979214171292234,0.548944250957537,0.28740390225642654,0.20098284019100798,
 0.9146583998208918,0.5206420766640806,0.34363995638180067,0.6841183643282379,
 0.5779631737302848,0.556619818644112,0.9424476709340956,0.5811008360033465,
 0.8953638498060147,0.8045970531434572,0.8141883315068387,0.05667904580137495,
 0.8990595437348957,0.329834848896158,0.7249249121649293,0.4959236289469402,
 0.37904895589959264,0.313369432538815,0.8438761855745995,0.6743862845584799,
 0.6565839479874594,0.38634524040323637,0.5237872595812618,0.4193344693355897,
 0.6747256344788284,0.6012600338371854,0.7944359663115068,0.4114019746817601,
 0.799713789378952,0.5820063124314075,0.6728171036194978,0.25255258743296116,
 0.6594555425521892,0.06037928939076509,0.24527098058777796,
 0.7504845688873889,0.276481687928545,0.00836089103571791,0.3870510114619585,
 0.8450310675597476,0.4265065870260424,0.6786852428769476,0.8776565338588336,
 0.7169574330404364,0.8156929928641113,0.6598550289606788,0.4022513450444192,
 0.47543162089748703,0.1916624648675379,0.5246391595665667,0.9126613803419669,
 0.3951455247492429,0.2487520693584454,0.5170322809824598,0.39378276168635207,
 0.9629050992137794,0.04124832050053373,0.005184386377190209,
 0.7037402246283933,0.3965532691052458,0.45360565672888187,
 0.16174377776739857,0.2881372646963408,0.5606375658862499,0.897386264470138,
 0.8645529258513542,0.7863038947156892,0.4580390486734657,0.8595381207455247,
 0.6364267781158145,0.8736346903587635,0.10982782020396753,
 0.39787139790244397,0.726267596724414,0.7350425654582433,0.2715716510344238,
 0.31712293353093135,0.26199342214073507,0.4042039152499183,
 0.7018775967447933,0.09912125781799874,0.400525623686683,0.4694541886016629,
 0.4639520782703781,0.24301098538599508,0.8191381605444608,0.6738524129156472,
 0.01089716166555521,0.9339722530050525,0.5543875857927227,
 0.18960810380357807,0.18796436376549108,0.6283759629519989,
 0.42424837162157714,0.4414387155454419,0.2859927545113148,0.5228834080225001,
 0.4857410240083593,0.28985333169018224,0.18155470967430842,
 0.5636113274579053,0.31346739414215785,0.1947921938722078,
 0.47415590097779936,0.6635949250887797,0.8856619944519402,
 0.10139878911818467,0.693051572714706,0.4574209027466748,0.27712182265768104,
 0.7958088253173711,0.9441796139385865,0.8193164485904817,0.03044423432231591,
 0.88030020258877,0.8051657293220305,0.38441545663190124,0.6618062630745856,
 0.6681850784160003,0.6369548518117794,0.21363616799392005,
 0.30588538796718234,0.5228430060341367,0.01968271735946403,
 0.43635806069014915,0.8262838903397959,0.4941693300886828,0.3172361671671462,
 0.3197783207745387,0.19460745152579662,0.6821094563282085,
 0.04827505400312837,0.5761458226840364,0.032132038482445946,
 0.5827759603969869,0.35427516654669244,0.18105339514600371,
 1.451586854109843e-4,0.4971149750523681,0.7402104245418145,
 0.5382186670324269,0.08541023427715211,0.6837062622872214,0.3650743172835126,
 0.8216667458789013,0.4529813610598793,0.17741543754966926,0.4102430124480527,
 0.35778435207704995,0.3217025392755566,0.7531414676060693,0.9420225313459824,
 0.8587252056458315,0.6461366545862527,0.5845781112200222,0.4390067047730989,
 0.5915999404001538,0.4803136124569425,0.9356238058632089,0.9501532761776126,
 0.7104107532879564,0.38916712152200617,0.0805941297436007,0.8059325976791682,
 0.3244268453837533,0.8796798694919669,0.6697060043591757,0.6023131894890048,
 0.4365488892127677,0.7836868407566513,0.28495777908207076,
 0.21925637278051457,0.19488529350349804,0.2421932848117524,
 0.11235755148298032,0.6337940267917946,0.03945581054382463,
 0.49829604646221637,0.4096436774873208,0.2856975253873423,0.7860007897970193,
 0.15914514462973406,0.8097180347689399,0.3652649883987994,0.6026895078408655,
 0.9164848440532939,0.14415784025300293,0.5493332005276196,0.256538474072858,
 0.05451615751285299,0.22996216233637568,0.9023473099034253,
 0.11008212197270517,0.5561833796408943,0.7302761479704314,0.9921644462622085,
 0.6162949504114695,0.17977409645903641,0.7996336144730286,0.4457969292063556,
 0.5155356441576614,0.04971121191520167,0.9689428753438669,0.6951276758658856,
 0.7943219842109736,0.131212972520258,0.4852345402999224,0.20764743179840905,
 0.1702024985228483,0.5346215780725724,0.33333734943448157,
 0.21191591855018332,0.982664617463203,0.7706207066984181,0.8005064805851835,
 0.7934460615883935,0.06512021220302477,0.02532460432293615,
 0.3326191715333875,0.8756218037300049,0.2370043784913185,0.36009116329722635,
 0.8770842025484828,0.35097653966700193,0.9505535862850851,0.4459434877654155,
 0.4797792802036025,0.31414652075724603,0.7029368617013908,0.7798079777833715,
 0.18826846158557153,0.20922842957879728,0.6759962526417705,
 0.6518033397868497,0.13663710528809614,0.4075143933867541,0.3151077891418059,
 0.9276389066744513,0.28465220166093486,0.8939885151453904,
 0.10015077633999059,0.6998631891304131,0.17372699731649377,
 0.4155564365664203,0.7400955475714927,0.2630241036314763,0.9942548757466874,
 0.9609149439385851,0.7590388047221827,0.3798967223224382,0.03059461502356775,
 0.3867409540382154,0.01976327760927854,0.06558361539614332,
 0.9587243286076934,0.3378853357862601,0.8055953177529351,0.8500511558340086,
 0.4389779387063194,0.606776622876149,0.06817152814462013,0.10221642272409592,
 0.5758196949836197,5.051838548033594e-4,0.2145060193699877,
 0.0572859791397482,0.24665511407112295,0.46201072221352757,
 0.3850967814492108,0.3862201981206774,0.8232991547350375,0.7899746323968174,
 0.5414637870149664,0.5901738917968334,0.9798611357676424,0.0685285489757228,
 0.6664853368719833,0.08810079564177564,0.6247826519235102,0.7497542569265625,
 0.6223327921739054,0.02325917350178719,0.5468136747983561,0.3535978800733218,
 0.8141664265772931,0.17179081483723702,0.03141470899252907,
 0.7783497021608687,0.4604170258620941,0.3660117801327436,0.16266545955274325,
 0.22423663948624206,0.3245985286360852,0.3333762981095597,0.5185459750860608,
 0.23472823674181453,0.5032819765222284,0.09710245132199447,
 0.9776596215757868,0.2630313971247813,0.3982988277447327,0.42351655650634523,
 0.7287127220430711,0.7300460868779383,0.814332666044402,0.06412811404735697,
 0.12029138524978777,0.5460208949576922,0.7144434744431051,0.6625910940270546,
 0.5875336399944528,0.2931853775101517,0.3905453736133918,0.1322496128175804,
 0.04273880593317747,0.02832998496814465,0.9212322773489361,
 0.7586536372921301,0.34605719187349737,0.08505303265669906,
 0.05961713822163994,0.6684180179211893,0.05477367377816844,
 0.07586678989269224,0.03242793603025407,0.5947586597179492,
 0.3001221377477117,0.9891283149122324,0.5803912596751295,0.7074003614615374,
 0.45867532376466946,0.13461618917648754,0.3441196318702412,
 0.6355749214315014,0.4426580492222786,0.6370393588416509,0.0888149210793503,
 0.44714545096764424,0.9447755094244876,0.13790529943767715,
 0.0341723181760788,0.4465920423457881,0.6379552545268482,0.16341664095798603,
 0.21836048179151124,0.5358700838783812,0.9007337070837595,0.5063894360465785,
 0.318777047509877,0.8641360261570227,0.3974247196908902,0.9378008477146234,
 0.7955981131102956,0.7196336471079832,0.5310403505194897,0.8146215938760285,
 0.9545475837542808,0.5959467471179506,0.41830059332490943,0.4192952877283207,
 0.8969677944083363,0.5336882119933044,0.3321262193687501,0.35067566881598733,
 0.8967347411904973,0.5542380756292441,0.10398025066318273,
 0.38819944268872353,0.2623277935353129,0.04700378947443051,
 0.8556499016558461,0.40392749857721333,0.1478827770983283,0.2590751634677203,
 0.001127568752383401,0.9362356117690346,0.5207928892285502,
 0.8602312459573274,0.13662281080576988,0.3299940988993645,0.5361754084571964,
 0.16730378051048478,0.7267920382599712,0.40907789949115325,
 0.05485744005631854,0.5816525082347628,0.5844997339978946,0.3349544966193453,
 0.4842101031785869,0.18701736927627288,0.21323139990329687,
 0.9798433343933388,0.13809462315275522,0.6528535752804115,0.5410161577243857,
 0.9143556798880631,0.7590298707169381,0.5378212382983252,0.00693446714661583,
 0.3909881857267634,0.42132873593098563,0.7049208838578159,0.5946434977081376,
 0.18382068581229682,0.06680705256288866,0.947999535655895,0.6655987113188575,
 0.2855311942675586,0.4240294003477322,0.475783603684101,0.870676476626859,
 0.8067183243552,0.4030118877589174,0.39437237225899213,0.954538130036053,
 0.14712053264260394,0.5679478533003938,0.6051912972047365,0.3968535184336318,
 0.5958778537857421,0.2251772562368316,0.7046937194060567,0.595138213622857,
 0.06561682392066748,0.08745215441649679,0.7498768380871904,0.871132288658501,
 0.5707275535426735,0.46033016810738336,0.7475539703733148,0.3933188808997361,
 0.8522206254421827,0.14683041090431526,0.6691575117059237,0.8730028305058147,
 0.9784320078514643,0.7710308429218811,0.17642149871550916,0.5308221611701631,
 0.9795934592517044,0.32600051417274667,0.6363073324354835,0.5115621863052444,
 0.7985228404096604,0.9181144506205946,0.6246368035386161,0.3918433542126888,
 0.8279508361657261,0.911385894371833,0.7514726712872943,0.653054498341875,
 0.007330894809782884,0.7216466085066944,0.7517968312296514,
 0.7732534954511167,0.1540562476916285,0.6677510854062951,0.39654724069656166,
 0.920168814680005,0.9980771952884033,0.6206823941741582,0.191869440077854,
 0.7384632848675606,0.35019215634133916,0.4081540145452188,0.51617732366205,
 0.276391743199344,0.7138395022355928,0.4204963509633828,0.4900450960726517,
 0.16584425028089078,0.4188848745299062,0.6304398833727746,0.1321798149424852,
 0.7034328364816091,0.27148905301634796,0.47578452253495385,
 0.8999208225899051,0.7389605027229208,0.8456629468004667,0.11759320289021391,
 0.9664570927181931,0.052307353216839214,0.5289904618406246,0.372783502352965,
 0.9836839351437039,0.5489879608832707,0.5808333266096497,0.05555727644155717,
 0.8865677277095921,0.448596685426907,0.9044450000541046,0.863021619270734,
 0.48899401980187096,0.059589841533436605,0.8880187516068085,
 0.6911962764807724,0.37842912680493257,0.866984467874804,0.8239351632224959,
 0.6714181034572507,0.03524643752345902,0.2869290876489268,0.3550268965493716,
 0.658668370489947,0.51654844997693,0.6446061210374344,0.2559771719891356,
 0.8237642100098153,0.4517223412981981,0.6227168876500379,0.6173953786394804,
 0.8957634357857057,0.13854423756745815,0.41587068534131455,
 0.35851680482956505,0.30052767592456364,0.4793306742689516,
 0.7414603197170961,0.5056652697232641,0.7594898032309316,0.5520233203047485,
 0.9438467853407575,0.5625560997993866,0.1615987237508043,0.7464821097608464,
 0.4948780674473966,0.3420544178751195,0.9230629546929929,0.4493367696419768,
 0.7495864280849432,0.7668851227965673,0.4567175554203641,0.8674584137416269,
 0.36885995196908683,0.8927423915243972,0.7968797916135512,0.8354199691553409,
 0.5226976403342973,0.1642895498920569,0.47323216873400276,0.8758040920626242,
 0.7596056391455787,0.3409484198661499,0.9449385488729256,0.44335329375223653,
 0.8719000346961429,3.25348861903052e-4,0.29741097250289605,
 0.23985269021618494,0.9295712063479649,0.3163282749378371,0.8158756639596293,
 0.47952070174805894,0.6065562821128146,0.768090806913606,0.3565203746749498,
 0.7722929836604506,0.8985986953090794,0.3071638007327566,0.8667017053734697,
 0.40958964207760584,0.06331840955476875,0.6650111256435117,0.688040196274145,
 0.5713445623442643,0.8613249536799387,0.8874887148564372,0.1891159197005745,
 0.7637620939321069,0.673107503688002,0.6525108732643906,0.6116464067110641,
 0.9981049281796558,0.9469309857971544,0.5218477927166775,0.9350253730397492,
 0.5951865782140772,0.6243341628778473,0.49410222626727685,
 0.22795579620674555,0.3537535038797269,0.617981564241813,0.05398749246271306,
 0.03996496099301772,0.1982802011560948,0.5266022899881012,
 0.46614879559759914,0.6750183078936978,0.19517313554247429,
 0.6311249384130483,0.5374131546200859,0.5978806680842945,0.04910860695434449,
 0.5746740946055858,0.2683384761773714,0.3037329161386495,0.4802310636067769,
 0.8671434034035375,0.1116285023570105,0.15167873243835117,
 0.44550924954634663,0.5516676629769219,0.012115486338053927,
 0.20148086678480293,0.5293855729604867,0.8300188893855418,
 0.035824724074720926,0.3170675493932201,0.6732583101558149,
 0.37360322067672247,0.3524323307202629,0.740016214033145,0.2237297071391331,
 0.45968588291855417,0.03558516964962888,0.5134175709272741,0.170155022844229,
 0.4785148195401008,0.17133319828285287,0.17369062928200663,
 0.37871654337313354,0.21894911695919905,0.3956846338564959,
 0.3633273820690086,0.49717060897504695,0.13324000314242246,
 0.7640795820723757,0.3482212502579496,0.8464359417611336,0.4672783804145261,
 0.6367446351808028,0.8293168191254402,0.5162377214921966,0.14918041361333922,
 0.2763725788452954,0.67899062020814,0.9617281815662482,0.3055071326258588,
 0.2919970701380764,0.6009417949860394,0.17169478484441392,0.2472930143397486,
 0.3726960235856974,0.7903505290089645,0.9603054105473565,0.9527646170599602,
 0.8895181816541835,0.5646034564030404,0.993625850173703,0.3960372702212699,
 0.7433243722351306,0.6105468264032105,0.5660121630305811,0.06343939492481265,
 0.7182026349640327,0.18536932583866517,0.5694688071461641,
 0.11839905682639773,0.5765806028338879,0.5605831822475361,0.1681381892140308,
 0.5225127085017847,0.9545577339508918,0.4976024068445166,0.287203088844283,
 0.3634277054299019,0.39549597487869725,0.5721374045783624,0.8731686124968036,
 0.4204599598139911,0.26487478867915804,0.41988873899516277,
 0.3161145737034283,0.33903774027644595,0.8313489762435922,0.4535909454430276,
 0.21468405762967235,0.8092567909863303,0.13835604590571493,
 0.6118001849245799,0.5000352456336655,0.623098969476254,0.7757488314664378,
 0.039411444947887064,0.5205669827828988,0.1362189077465995,
 0.8394381773169881,0.22542197720488866,0.2463492881467484,0.9240778915950856,
 0.24098879823536756,0.7934610213875727,0.8128266946136931,0.5703745081764295,
 0.24151544996633567,0.5096413275164529,0.4364860398869047,0.6967942934861245,
 0.23112295345516798,0.7941860055594916,0.959735954534111,0.4617364379579537,
 0.12812666705007003,0.16930802479022578,0.8626606591370635,
 0.7641528057527819,0.9575019131611116,0.855097882894192,0.2211552785618549,
 0.5118402321537112,0.6158805437788692,0.6511886451867932,0.16872806355398562,
 0.0360390824603779,0.7027990981120669,0.4772118843255423,0.40223920998036433,
 0.27868905774991937,0.8093867659096688,0.9835738785796934,
 0.27401359134003345,0.8395331406904667,0.6239360613564219,
 0.21365583408006616,0.865339108437419,0.701069242691059,0.010261771470743253,
 0.6209472770634843,0.9454238339479242,0.9098665360664618,0.9237814659841498,
 0.22380637981991347,0.15688560428679443,0.6025295198338009,
 0.8384722190175671,0.554882378089953,0.05960542091804988,0.6683310529447798,
 0.6834249761532024,0.2337212690611542,0.7643696055085385,0.5215404005022615,
 0.8145476603518963,0.7443537478711623,0.02747577617833441,
 0.09452185346395492,0.6953863671984774,0.900754178274296,0.8342578062421838,
 0.22835685224265745,0.3186796872316152,0.22515547985050577,
 0.05774637745172451,0.33980215172517414,0.8415951813137299,
 0.5485631631552887,0.1865096376652693,0.6006827328779661,0.85494428985505,
 0.6136969381650923,0.08124248924295197,0.70199727706502,0.3784939637577518,
 0.33614642670692685,0.1922961654280303,0.2174557386801056,0.8171037531546725,
 0.874841373982727,0.16673837580133632,0.5469029874812652,0.31626541125264296,
 0.5850651502320946,0.23811875549564354,0.7858121109607891,0.8165291532991258,
 0.6984889685946619,0.1509355598614064,0.6029744157904803,0.2977519201309722,
 0.40550986727733584,0.3801279342803683,0.48366776866005334,
 0.49163056579813347,0.20778001057123197,0.6482983189784145,
 0.7312161331161816,0.4484068333696658,0.22581233485641983,
 0.13772352923554498,0.0039573674472490605,0.8416455245585138,
 0.6739310259327855,0.9174592179007443,0.8388303964576185,0.7340689034715709,
 0.058973290280051494,0.9547822869121732,0.14186774950010916,
 0.9958994587829614,0.06748464486292438,0.24874446886803414,
 0.7752375074575493,0.41224339452826686,0.5924977601596776,
 0.15763116668779176,0.7885612224361007,0.10751751944790255,
 0.010599159884086218,0.7534197092759276,0.08605750180960114,
 0.888249926374025,0.1439508949641064,0.767577747126024,0.88630548058535,
 0.8369927237694428,0.38899116530350897,0.09198274437911502,
 0.18999467386698943,0.9933150999725395,0.8157276727756643,
 0.36643755703714354,0.17008861077445547,0.646268703293666,0.5017117334251818,
 0.8333147251904386,0.8025624373667307,0.17620582864204237,0.7702446696936667,
 0.17173979958255758,0.5256267810152698,0.7119554566779532,
 0.06142980348895766,0.1681634553301231,0.31070718702791367,
 0.45315047093792826,0.9510569999213496,0.6521339114077691,0.5988393274431072,
 0.39461023998219824,0.3353369700167944,0.19103190896887512,
 0.6637872819573659,0.3958291273202317,0.8309415812849799,0.6038328386059926,
 0.9845000775067476,0.7834067161334045,0.16788886363629618,0.7245544031429904,
 0.8148641885537455,0.5526501905251753,0.9837447273305953,0.3158346558605004,
 0.7450023774672565,0.9933296905017179,0.611854610989144,0.9044612870749877,
 0.6441223932943391,0.04633973788765988,0.32231396544594204,
 0.10195608480458984,0.9633200629734537,0.12458364015667234,
 0.9440242688912388,0.4456815126005387,0.4081973532647949,0.31261988120959283,
 0.22743806944319866,0.5649750517568509,0.418056349443066,0.6802933170840336,
 0.37797522973312225,0.3633964224974664,0.052962454723586294,
 0.7712420959054648,0.10162756878481272,0.3518555121719348,
 0.28109889735074656,0.851222761018443,0.8850912495790197,0.23031141434914326,
 0.15816986275827605,0.1344914685598172,0.9198887736159594,0.2519640223971926,
 0.9561990595311165,0.19729218833080597,0.36439010423727536,
 0.5481242280924896,0.6686466856363781,0.9728020256822123,0.36842590223595817,
 0.8910849423165587,0.3422528910881034,0.733022079899933,0.393135937791664,
 0.6927344290718885,0.053683385048506116,0.9574260278739272,
 0.1463932169737574,0.14407136582157198,0.949525822037077,0.6206798753210068,
 0.7689633264129565,0.15556977280272066,0.2054454228921132,
 0.10345576746757312,0.3240076491063012,0.2973828951211098,0.8654211352144069,
 0.0755452867820805,0.1714800344420081,0.32462859092192475,0.0657254783896557,
 0.4500844965440157,0.5860002606638727,0.7074356948459796,0.9848297983800969,
 0.5629091404697634,0.931452276306892,0.8214684156139311,0.9471248363360043,
 0.7966556713608737,0.561640753167094,0.8358414976504588,0.6784792618573664,
 0.02924040951161111,0.21940568571197727,0.6440768973659166,
 0.06801488535846922,0.4083824540222838,0.5158552961876224,0.9020210024219097,
 0.21996984300848865,0.47823066804134606,0.563332169669081,0.5363625870965328,
 0.030192171368690346,0.44951448597687627,0.6098742816813219,
 0.7487457799566025,0.2881466127760153,0.48479428699701343,
 0.13644522800159675,0.023138965198462924,0.13979279980534476,
 0.7330785618652134,0.6183289618769752,0.48908114022529514,0.6433213155608524,
 0.0030116310750933906,0.8089183365223601,0.14877191249884358,
 0.7701264387505022,0.36339049777309196,0.9195383006562707,
 0.16428038736061523,0.24388820702313352,0.5563213821000821,
 0.5115435594973725,0.7593421004703185,0.3193304960804262,0.15650317866517027,
 0.9439930310256021,0.27561794438059006,0.5964777201660731,0.7326421986423745,
 0.14829088754213138,0.3613925220087071,0.2705043870616687,
 0.03085538111505448,0.3223808136382347,0.5369919101904197,0.9013520762936564,
 0.5656321958561864,0.9210189856693717,0.2950065269869173,0.7821065584275881,
 0.6267324708078006,0.8968520394993293,0.5096507612779149,0.1391907055919529,
 0.835582417549928,0.5587864001365614,0.6614027429006588,0.8931610851976814,
 0.15609982205841155,0.6287562208351448,0.31215156602975525,
 0.017828808429086074,0.48708823376681465,0.2279794366378588,
 0.5345828410685393,0.5975221821424822,0.8892348897670775,0.3365442954744323,
 0.2469298316616394,0.7516946702369097,0.16064854059620437,
 0.49544438869209284,0.5161918233494758,0.4765177231877419,
 0.02394727511564132,0.9733944134917497,0.19621970839302727,
 0.0789377736800474,0.9366060956588909,0.08740519804926006,
 0.22800934234068793,0.4512573770756201,0.07950373136597655,
 0.4115483502243177,0.12025388297671435,0.22390000769827834,
 0.8488599373844081,0.022964886690144626,0.5535840313182168,
 0.03014462724108813,0.908114980358433,0.5913622410726442,0.3820316616704349,
 0.6543942280840256,0.9510821273323027,0.8457876514710794,0.8552333086222019,
 0.09549033071612156,0.22205240334216614,0.26777955924602115,
 0.9539715599878669,0.7449124643901232,0.592145907403788,0.27067220242256684,
 0.8756546785638595,0.6966641208839358,0.8040641095321809,0.9126330585580247,
 0.5032742513364181,0.7152322887799171,0.22088366351784006,0.4362883046618896,
 0.5818280277739034,0.3711985278653015,0.5238667207835251,0.36656841903303505,
 0.6012447432538156,0.0076524171638693295,0.381492394575071,
 0.028418571447464624,0.6830970056892934,0.2594706744893984,
 0.9677338007480231,0.16927302678310474,0.2841996284543309,0.720432441079011,
 0.4855062535567445,0.17329616324349661,0.42038343497230524,
 0.7728073603561677,0.939645041189384,0.8068390380339522,0.28051118141327547,
 0.8536796520701391,0.07181176099722164,0.048877676350675636,
 0.9291055868557585,0.8575737046950913,0.27902252186282483,0.8367301115850513,
 0.1100374234957433,0.7140689668788566,0.4210315306581457,0.6791828259488399,
 0.18339810928071665,0.19871934194499152,0.19685338372571148,
 0.03586822609242524,0.8849015645884237,0.4065854852665447,0.7484711887078317,
 0.17314212945631247,0.5547263846407338,0.8572729210751038,0.7085170306303238,
 0.5284894455248875,0.4166085154598844,0.7452837795068847,0.8587733470990209,
 0.27223265813652997,0.32740054328658086,0.6109893435635135,
 0.7980787979290718,0.5726433629953022,0.7367398566764277,0.5190614091535761,
 0.7986738311613857,0.7059760897170607,0.819862994779747,0.07844551389889698,
 0.37420656656743845,0.9679293053997406,0.790950146188393,0.1085511296062448,
 0.3683916267006395,0.8137627524291077,0.5126323135369958,0.3438790379068116,
 0.4233218242581076,0.43929757304658,0.35974662445692507,0.9206345111357676,
 0.7966452100425281,0.9697157792286827,0.12149100066953455,0.4213786458125489,
 0.8287542369958941,0.5695362386276538,0.41050444643335604,0.8004860683281821,
 0.6602898693850038,0.5018867314905469,0.2227973886194613,0.7965001811031193,
 0.37429970293792447,0.5277103177428049,0.8979286669578856,0.4747509756264304,
 0.3186821584775523,0.6909494730225044,0.312814990338394,0.9388995702409104,
 0.9140911300028547,0.9295855847368922,0.32231987050717514,0.2598084108332718,
 0.5386689341872384,0.8751167801482932,0.036019484324420925,
 0.1993991772038386,0.3469788627127477,0.10723750139742916,
 0.26811191065875817,0.19444835017463713,0.45223695823411714,
 0.9649708899291413,0.15425208165104354,0.4826334256374176,0.5489962442236143,
 0.5929972823315769,0.24283700969947863,0.8700726484542605,
 0.45760527826029485,0.3402255159740496,0.7164734726175978,0.4823068847490479,
 0.9543518298576699,0.9697672253687301,0.4898298886687099,0.7944367907785271,
 0.2499637025755035,0.07064756747254208,0.3045225084454064,0.6285731000736942,
 0.5041014543797583,0.7383832266836625,0.5234314897244451,0.8170987527750841,
 0.8289350413601954,0.15425783445678254,0.4307309066534568,
 0.13064962369177902,0.47617033878183224,0.1365688533641729,0.406073847162445,
 0.30731691152272067,0.7113436146390101,0.05214567072952847,
 0.7308024257441799,0.9264766186746725,0.19351192330714861,0.9086799070548993,
 0.13614633406315313,0.7046183356799057,0.09803906406511675,
 0.3637079200747292,0.7874609272543522,0.619094402823696,0.2957233150813626,
 0.05127604109152395,0.2770635518861162,0.49759323987877657,
 0.4208248024219903,0.018756440627851756,0.7666204534643684,
 0.07056725403050867,0.12391657624364694,0.22704910984375193,
 0.24383328454117048,0.7316597929331734,0.9728607699855001,0.991584530295132,
 0.5514210457484436,0.702269575075339,0.14030649708005993,0.21447778173935106,
 0.35169695961675906,0.8421562056089551,0.3209235989238788,0.9911587801549958,
 0.717145840173085,0.5810823016501045,0.4067579535787369,0.2047409921924277,
 0.9871831097614525,0.21606739753765392,0.5820528969107683,0.6021597004679388,
 0.7193346773331151,0.028796778127254563,0.06499977964157855,
 0.7853964999677723,0.9724828089397548,0.3426489775288012,0.37201116152889746,
 0.878257323155597,0.15817095795186087,0.2202221628250438,0.512845599273318,
 0.6353149539979132,0.9105491770420446,0.4328531743528279,0.1906832955104144,
 0.2946318478797857,0.3276615178878006,0.6839997481762635,0.9626798776184758,
 0.9261250697659558,0.8070316156322285,0.24924815033141112,0.8187471767700043,
 0.3638681695392779,0.7560628308282459,0.006809965369610871,
 0.7537674814185901,0.980226866398694,0.6680844629425772,0.5834881998124032,
 0.16754427976790265,0.02257967865304278,0.6565122167129598,
 0.3953050555655293,0.6266457380823002,0.749538325860705,0.03634161781759948,
 0.3600422135958923,0.16496026809457687,0.7522692024034398,0.693176274535762,
 0.18454138963328404,0.7585005381447191,0.8681127387660976,0.2596149214854999,
 0.21502699674592096,0.9167906452227353,0.3676096473064181,
 0.28420964538548743,0.9567378061590488,0.1883519968532288,0.7833980556377345,
 0.3908414712661008,0.5685839564104644,0.5906293775299977,0.08303544942437724,
 0.8944848663291276,0.649128271225676,0.2779752763606762,0.33720710749267346,
 0.028616904752535532,0.6202788230658804,0.662190368279729,0.497214056968602,
 0.5659801194593747,0.19494016081057042,0.33991125731920946,
 0.12508607923595638,0.04819006593509656,0.6524907739806494,
 0.16970141153618068,0.4444390091962307,0.08964315669125966,
 0.4372836551360214,0.1304569256708622,0.12825793557482157,
 0.14917121813770384,0.2007471076929509,0.5173943821500054,0.1731247387988727,
 0.33743062665554646,0.602284304720238,0.017176171585367106,
 0.4959936149733416,0.12718683635961392,0.9164532362989669,
 0.06240500443178687,0.3159153381675768,0.44756450938383463,
 0.014468187958314949,0.032663357431031104,0.5038445622067473,
 0.9502802127032687,0.5282931514815096,0.11268763788457692,0.9314955195483722,
 0.6538096710233102,0.13949500440414275,0.8358355962827175,0.3669022496616593,
 0.8077359133083937,0.7763717778192588,0.7903258932453382,0.5114544692505423,
 0.3269148390559229,0.8932382666372979,0.7784613356985489,0.8545618819537849,
 0.15001338686763088,0.6010462122715504,0.40527888674979584,
 0.7777978964050537,0.25179339914670096,0.9961907679561395,0.6521547392647972,
 0.5375015979171792,0.417534808985098,0.8822477532645924,0.94858399175723,
 0.976003341252018,0.3571080391700696,0.14012875998475638,0.6256115663085422,
 0.7556073727619174,0.7164835651589597,0.9794532491195163,0.6066446757412975,
 0.6240474566492433,0.6393757208294262,0.10793741687611269,0.5328274795843522,
 0.16702061218140274,0.004438987900296287,0.8555369519841537,
 0.1287234081329922,0.5222688431471502,0.8657199853064446,0.8587756934358683,
 0.2307729805371408,0.7053168682374724,0.11231371666957723,0.4736269427415769,
 0.5083190261618915,0.6140698130354766,0.5213021665744811,0.8894990720483144,
 0.34243932767742624,0.9317683275662179,0.7084679388685333,
 0.13267398367178718,0.38494403318655634,0.04477326763981959,
 3.9740634733542723e-4,0.4485206654645698,0.01936912292607673,
 0.7209128577554865,0.2289007628393126,0.4593513708752349,0.7553940217882528,
 0.020508110423587467,0.1938310030289535,0.08128691639821573,
 0.4819230150615621,0.004017178314019043,0.9922615892030816,
 0.6332460088435593,0.6380833577103099,0.012895664331283418,
 0.009223186456435273,0.4437807186391256,0.7173916256021229,
 0.15712688747303338,0.6152873372980633,0.11298156531544967,
 0.8818941855096131,0.3312425553846132,0.22279848883095787,0.9653497318343515,
 0.9142331344819296,0.8082395421685653,0.46031822063546823,
 0.17509367983733826,0.8497654821469153,0.10635334886266423,
 0.42497018793097185,0.09612195435658177,0.725231458541986,0.7140687181215996,
 0.26185837525475497,0.4046519883954268,0.3808480003397303,0.8698125015533784,
 0.03577867069865148,0.25129105216285486,0.2965457315492045,0.371214709991186,
 0.6330968105643277,0.7527491648884425,0.47897159640217835,0.6251652278496099,
 0.03603540498291902,0.8930952787752403,0.058386764678986625,
 0.5249552397314686,0.8258056625593432,0.1988145857620962,0.48563743197479037,
 0.40899529354794495,0.5582898614606341,0.7815638682522064,0.8928074507068775,
 0.3279231826752951,0.1482375905463552,0.9566920826948188,0.13136229285602818,
 0.40583783832051257,0.5932139435885885,0.9288729043483986,0.7215093232580627,
 0.6700998593269066,0.323157850847523,0.8365986519063118,0.846400746785241,
 0.0343386820437378,0.40205724985937485,0.15959054901254266,
 0.42963636382398274,0.30274593118291837,0.8720035790995704,
 0.9886983652023124,0.8220521633624215,0.8445685347589437,0.7585771783004467,
 0.0976754926905129,0.4856081778213147,0.47089424094291144,
 0.01164454782794433,0.45930250687690566,0.6124031113981565,
 0.9726607004639434,0.07981265258286696,0.5235590012540201,0.7643018881182518,
 0.4868008639870458,0.9905071108109051,0.24166810432504238,0.9067791556597202,
 0.6521026737726258,0.7869764000690702,0.780790085012813,0.9928057519048447,
 0.06663172623526614,0.1730399788967253,0.05780848010476092,0.500939110728974,
 0.8501711688571517,0.8429121609577279,0.8750175979466489,0.8179562388067446,
 0.2100020687221743,0.34373021539751514,0.719414690262553,0.5013865514373222,
 0.1368610329381329,0.8791985908463188,0.4855919914902279,0.5755463092029781,
 0.4509913143433193,0.7833011624567178,0.08180456090692312,
 0.06257756060715192,0.5129386050499506,0.4067525035660242,
 0.14249199713785698,0.7251088115234683,0.3196961106368529,0.5644507637451206,
 0.8436874006952033,0.024509835667093638,0.5071991481085399,
 0.9795669879728641,0.6250621552709172,0.41555665330992586,0.3699633851854274,
 0.31818503246181296,0.9160716074897505,0.0402255041426699,0.5718377822826543,
 0.32307768655354674,0.07758410805738292,0.046933926624872724,
 0.2487450719735017,0.05081957451374863,0.5703040785527996,0.2961359762173982,
 0.5028424957099851,0.8214574399387761,0.6479913600982268,0.7388604622306043,
 0.8638012624052794,0.8315723207547321,0.6612086055278257,0.2699772687077795,
 0.6205489586394511,0.05146140654307818,0.6779454689673534,
 0.21022446393284688,0.35744536669097116,0.9141640233114379,
 0.5865997475316149,0.6037015286539643,0.6772873506821484,0.18547734613420985,
 0.8185913964068205,0.11338283533640148,0.31240741374292214,0.548433306905893,
 0.4899076513562718,0.7018215689562516,0.41659629544107146,
 0.43556195710172607,0.5903539435853036,0.18502174818060801,
 0.8916771585413863,0.0719579091931477,0.27836342873387854,
 0.13098890628408444,0.5465651254946791,0.47463317811674544,
 0.5939512026904199,0.12559436040852434,0.2592170307829993,0.7762774957488288,
 0.7456478309403143,0.3851308857998651,0.616799123026753,0.9481125061225866,
 0.648963961293515,0.43806454857365884,0.6514330422785488,0.16412520762233362,
 0.4163029865869986,0.9528579093355612,0.6376127163731702,0.3711554879800363,
 0.6012206535508704,0.7235950391505479,0.37505170592627013,0.5585030683324234,
 0.9470912869252072,0.6664654716841035,0.34546319892453015,0.5371138091171321,
 0.933631408342843,0.8297510817116254,0.6127015449803963,0.18938576540837837,
 0.4863775386346868,0.25523968787988005,0.8003444705468519,0.7121576061853303,
 0.16870423059041684,0.24704362458938622,0.600771143844383,0.8321699229298638,
 0.4101483749344922,0.136322601206883,0.8681042664309856,0.06990170582171618,
 0.7873040793736259,0.6194834710557385,0.05939230949117602,0.1704408648682244,
 0.49362370425776825,0.09803011811681328,0.5319661924652317,
 0.12172219374994997,0.9091020097751858,0.9099411449118802,0.5131929257945418,
 0.13969664906444612,0.8608027851099112,0.6268330066615055,0.7442102164675133,
 0.07280304465090992,0.22028663189083142,0.38033854402484957,
 0.7895097699524029,0.7259936456064912,0.8947734994936363,0.5177200608051858,
 0.23882080156687313,0.903031408286782,0.323544468851361,0.9413430194085608,
 0.24967426113182967,0.031238197923967093,0.33436368856017695,
 0.25576136063868904,0.8737659164027166,0.9468482497336271,
 0.11859897345934722,0.6219081750431198,0.8548260446196307,
 0.28057624586982377,0.7882975435281486,0.7963138069806722,0.5768971310720745,
 0.6432721122075972,0.46310872553780646,0.4962996008549597,0.8147579311971931,
 0.31995239669702047,0.10371988138399957,0.5363870624621547,
 0.6999376172080191,0.730726810214008,0.13705297723001408,0.466204393677762,
 0.7598242699286281,0.0911387049465574,0.9436753119410031,0.9219298054938516,
 0.1881151007147025,0.7473222292218659,0.15227210363770183,
 0.035766578034344465,0.5198328231120348,0.7957668164297611,
 0.7622736935305561,0.5797981031801628,0.9552587017215572,0.0732551847313141,
 0.8626102078981637,0.5698943788380102,0.18769026509014508,0.5457373204802507,
 0.8675936561995806,0.6777734048521272,0.17931005621397922,
 0.36795626269541604,0.13449343102878164,0.4794230076220942,
 0.2397318900097939,0.9455276401504196,0.28710514953355026,0.5364801113319027,
 0.3568233531571612,0.6788179448811275,0.5964849561887904,0.8878227061719125,
 0.4476807780995088,0.03438303658258324,0.2120308284595115,
 0.24878520391479375,0.6340258028656146,0.375430075884872,0.08126001192365462,
 0.30611220999539834,0.7112617863926235,0.9363111871871304,0.2402390955716277,
 0.7928897545245164,0.12478890698866607,0.15142030465446654,
 0.8124883618918911,0.4957613064588563,0.7107881562114898,0.6910516254768035,
 0.37977467172682644,0.08800308762747144,0.77281448403445,0.05152847063414723,
 0.3937313010485839,0.6124278216500628,0.7456827333270606,0.6049937446258902,
 0.32493827537432407,0.8717601044909841,0.5661367082456593,
 0.45945115809383785,0.872203153758755,0.37547537012755083,0.949309145065503,
 0.08741481729708844,0.9891402656592484,0.22405866182360112,0.65650468293316,
 0.7557528787949157,0.32387511850525497,0.7697620833253268,0.6435444818980907,
 0.6526143720927856,0.15682748097824128,0.21786571670082833,
 0.4538503631952614,0.667985959883721,0.44475498082496123,0.345193509593566,
 0.2582231794850768,0.290161360741797,0.612884703772249,0.8706494371508513,
 0.9928429564973418,0.7879132703583633,0.43485648183899106,
 0.14667370269480795,0.8562573505917331,0.6340047348457687,0.9650538231491097,
 0.08991429388824834,0.7434194016914092,0.3581001057886577,0.8726891456512667,
 0.8644798982536933,0.898801717231335,0.11365018843918662,0.1767352890855417,
 0.9639638050735119,0.3689653809411968,0.658242633229925,0.1357435388306656,
 0.782567706119932,0.20508908038258356,0.08691627458137519,
 0.06350333576492773,0.6019153223612848,0.41044154507878094,
 0.5452480189785894,0.8269427039233861,0.07416773225445072,0.3154267333174925,
 0.8894695400009122,0.43755215373534007,0.9459939220448649,0.7583682447600234,
 0.9419970275907212,0.9550851390303068,0.1917224787633921,0.12475689909548304,
 0.06378950294344954,0.07094121712627155,0.6916315633507246,
 0.6549687035766407,0.9542185368735084,0.9771821871122924,0.766793400880252,
 0.6568667686987292,0.0838118897980481,0.9949132136035116,0.5808050474992092,
 0.9508094352484532,0.00844274722487437,0.8717377363632243,0.5385810714860595,
 0.25061678215973415,0.678995575594108,0.34224625394184527,
 0.10194466308851147,0.4735135968145979,0.9822543993852354,0.9418390418665715,
 0.857678033076593,0.04848998332984189,0.6585426277954474,0.733603100837682,
 0.7863057760602135,0.3192899861464902,0.7477018350946115,0.05640750665590766,
 0.8353618964967676,0.8988385630220732,0.14705734003091997,0.5392219316784156,
 0.8318436657684721,0.515162460176658,0.4329668362808218,0.8662318942338841,
 0.06825414765605853,0.2920838634099554,0.13071161673523846,
 0.3018118967576212,0.9418559068740158,0.41517559597432563,0.8649848810518932,
 0.12959672832144475,0.1929177892981,0.008308701241451555,0.18653004553011754,
 0.03289334767550167,0.06626221452799697,0.5725742133144747,
 0.8308537822877038,0.7837004368221997,0.4437847078819499,0.6351847544356868,
 0.4505701414867841,0.13144125885963576,0.6296997984581003,
 0.49835000885177627,0.267845217600002,0.13158640103995767,0.5786803074237992,
 0.7260638290808784,0.24108794436916503,0.4555598072759762,
 0.26927510876236593,0.7715506969430894,0.4547794102421727,0.8841482325052568,
 0.07622749900594128,0.19258900673394375,0.05596143964210054,
 0.372709178388017,0.9505335065552564,0.6730480446622729,0.4991053381040985,
 0.5173874665619891,0.045640883030786505,0.1246974189159995,
 0.7930750601683165,0.7211806639978235,0.2883998589894695,0.9950799906732024,
 0.24583591534113702,0.5294887992663118,0.3344363406570773,
 0.13865356077346058,0.24449843495789714,0.7585489633234035,
 0.6524159495299233,0.07458969241592828,0.26026333249850375,
 0.7274060248830703,0.6340835837341288,0.07218431427116223,0.1180053285228797,
 0.6736105075710634,0.026701853904123052,0.7295478861440018,
 0.4031615065444901,0.5706358056372879,0.8948245549411938,0.8191070446040515,
 0.6066738383558995,0.13207843519853413,0.8397724186520766,0.8835016651270523,
 0.5205230406692333,0.9643514604768031,0.6565010011088824,0.18226069982583704,
 0.7178596304455238,0.2873977958725179,0.26689922616496997,0.948757808675437,
 0.598199273904722,0.7038960498376828,0.9820131770369167,0.41980799940857416,
 0.41162980683547146,0.7255375517294412,0.7053918123319171,0.721982115495734,
 0.4986991876345541,0.01773236546850354,0.2027177617644741,0.4948329025497141,
 0.013415628724778417,0.8735547625255319,0.4408043332209616,
 0.7682498050798603,0.026950632483733195,0.2366221487695831,
 0.8773075399322486,0.04589021908704549,0.8479117464040233,0.8230927326331153,
 0.9172558788199654,0.11046276693020673,0.6545382397962883,0.669495846241541,
 0.3036761061128105,0.7940351434790134,0.2547513916914673,0.2762359064299721,
 0.5052276298347076,0.2899996719790112,0.05002358256010542,
 0.48204634785524514,0.4546181306112649,0.5931062346067933,
 0.43073438468251635,0.4581419685539322,0.189179705258674,0.10673868731334812,
 0.014499193463159976,0.2453857584863688,0.45032310949901977,
 0.5474747822000061,0.9385060505366111,0.9001381797819881,0.841645000827128,
 0.9168982570213179,0.4125514603309033,0.2723290807299069,0.3217375284594546,
 0.11599319626004201,0.22483076289929427,0.5647862713783562,
 0.9228670650877573,0.19525107580206214,0.9865044843674309,
 0.04456605879975406,0.4422594844625056,0.8313347816125228,0.5720756867782241,
 0.27771512840319534,0.6517493285828213,0.8254148045319698,
 0.014688557136484759,0.3431039029043319,0.9437107705135307,
 0.20264105821754597,0.031565050258044125,0.7510278206800295,
 0.5797933835314318,0.46190849598945594,0.8110972441874744,
 0.13383990275268098,0.7470236569767722,0.9334849514766738,
 0.06288008813802581,0.5017025616746686,0.2893478444287917,0.96012838152614,
 0.5071961359464671,0.47415498380864773,0.4140147420870893,0.9392948315612802,
 0.4616165642223534,0.17836601824015963,0.7094394604902714,
 0.24872104926529492,0.1534182968231732,0.6163505378970875,0.7263409970485375,
 0.5225392509378719,0.9498170877914898,0.3193209018229384,0.8956973645642365,
 0.21922406970827069,0.7449643952205535,0.4362575350725363,0.1624113802830759,
 0.8133950575711564,0.9156731056199304,0.7210794181632814,0.07333497290351965,
 0.027417436156993547,0.4062795087610128,0.4155108738838633,
 0.9649533976928054,0.39694699094136177,0.6155215821770925,0.8930869429950962,
 0.2085117348598775,0.6697418250641013,0.9410582351261859,0.8103731591594097,
 0.41521084613339065,0.6311771396165735,0.9102299693953778,0.3438364593894341,
 0.5108049287987275,0.3756278682561729,0.5064760029439591,0.4481498785173068,
 0.016296044024039502,0.428242002053559,0.3989044447683724,0.8772432799430586,
 0.3907907367733846,0.7506204554060705,0.02958669788095436,
 0.27744518772942683,0.17455296773696094,0.1096628655565336,
 0.2448067547400372,0.002980257339278314,0.395665812375274,0.9474528230064446,
 0.4839425423123557,0.09070803243425518,0.033216517837383996,
 0.7121908543485463,0.9620157583256012,0.6654253927571296,0.36071292270459265,
 0.6087567955335634,0.6728727480309665,0.08745129831691167,0.3753662816278204,
 0.5857483607186488,0.6167204867867593,0.49416366338732187,0.3295902478266637,
 0.08757687317715801,0.9175082939925164,0.4464480852552377,
 0.27779950748335025,0.3418312085153302,0.7484268541045375,
 0.029654549362227445,0.1680751091750683,0.47147337674063716,
 0.21706825091868998,0.11343192088469645,0.36632027951012525,
 0.677071741443174,0.04967287830534517,0.17188320186608697,
 0.47588251596832665,0.1189278737363052,0.19194279557352878,
 0.9154941110215984,0.48133212604647746,0.03501609220533919,
 0.6990823215738906,0.8031084553674094,0.8207640579782778,0.631695088242749,
 0.23650026184766526,0.6689014584500347,0.04796726043486088,
 0.3186843816965166,0.7528154756841149,0.25837620447861176,0.4591783205986314,
 0.7561202628117598,0.5332353777485425,0.02094750124285716,0.2616140798275486,
 0.7556103054606872,0.8140822163783799,0.5597453770849811,0.10377448965846225,
 0.10206697494515293,0.8482549541084805,0.3994221188607813,0.7916520853204623,
 0.010687624270190987,0.43305332306709887,0.12462767188712054,
 0.2888787307909433,0.01748946231880688,0.8710599128619324,0.5807928820285808,
 0.9231098367272239,0.6474001727153755,0.43448727464450876,0.5420497103828409,
 0.5930246159914798,0.08015545490408149,0.6171944330009731,0.7739633042880742,
 0.2709606665272958,0.34637527262082,0.7815087917055805,0.18239802250464598,
 0.5917111504822736,0.2297585554709196,0.2303278973640045,0.1921417120232275,
 0.9795771457749386,0.7171552386045305,0.43571757625334784,0.8666894767497935,
 0.6185718333218513,0.6704726291729272,0.7890920096464599,0.8914353998775975,
 0.8218458077315733,0.32777087680820116,0.9359631038217125,0.9475253874210338,
 0.3347960633033308,0.6076413619309629,0.7937516645361918,0.6933312512822953,
 0.7430854044018277,0.8538126114908401,0.4006828418332332,0.8638166324983336,
 0.7163824680781327,0.09246109750561371,0.6032411565765216,0.8736245882717896,
 0.46643551400147165,0.9837278606146664,0.016335544424437742,
 0.23151848250656182,0.8526710625917058,0.5545958350993011,0.4684079915537793,
 0.05481258843286685,0.7063858653164897,0.49402816925035853,
 0.3688533142548296,0.2865963485887817,0.00151375222650163,
 0.28373422853879027,0.26995764571687797,0.7347023001726375,
 0.9131877000390993,0.7503553912965697,0.14508387852875115,
 0.41705572015498216,0.8122294604858288,0.5868216162506162,
 0.20510483453914075,0.6408031398407774,0.7357421228407567,0.7825319024043873,
 0.5541502476133537,0.9949379703019817,0.31104028623986524,0.3725484548559964,
 0.9487772709240233,0.3946498830797456,0.4900517899715855,0.46609407722528906,
 0.6267161060245823,0.8839598520524778,0.36823769988454835,0.4553613807726543,
 0.7406331948992118,0.28016015414333373,0.7775977028671566,0.8683740282419835,
 0.7875286268429708,0.2072724461785561,0.6587626254500455,0.819087946807854,
 0.5040578240504806,0.5004071212946615,0.6666825504878521,0.6608129677597346,
 0.8317981813705186,0.1566814052583323,0.4595699765536082,0.5696684180972081,
 0.7301635411002498,0.8119279025256173,0.4235093796964208,0.10337332652817266,
 0.4224690621998436,0.9709357832571976,0.8741987298079992,0.7955952217686006,
 0.5398268373092473,0.9368428802023656,0.15867639191012195,0.7610461809439822,
 0.47990880218059406,0.14967363394960864,0.658640712043553,0.983178729616914,
 0.6176393519081119,0.331298677674559,0.04319221641889048,0.7847181635208731,
 0.2518524334875508,0.6250306340916912,0.363013572948122,0.5662862413825867,
 0.943613481180813,0.17802546058510593,0.24445369343612555,0.9446451430202607,
 0.07679763741478429,0.3220569700729432,0.6669689622249012,0.8528208292079793,
 0.17214014839081482,0.3616358111251182,0.26353201579356966,
 0.6140484222124605,0.7221172787681475,0.7674331630921941,0.813505572079348,
 0.8241095223743673,0.6444887055580213,0.6369859896800164,0.8735020957490205,
 0.18706220718094546,0.46343403894550783,0.5107726806420398,
 0.9764851328178576,0.80803434857449,0.8395758474444127,0.32989409457772745,
 0.2894376361712382,0.06386286534092778,0.17222232612814392,
 0.32022256506471725,0.9458522642373107,0.37737019624191004,
 0.9264793589304618,0.4150290063642612,0.804674082609777,0.5719401763549837,
 0.9921035431429668,0.11630047014645051,0.7712625061612524,
 0.036726637874648915,0.6961423772206817,0.23528031698762586,
 0.7608685363938619,0.3601717810961713,0.5004903714679675,0.518769755000743,
 0.3840917922028467,0.9068796918424836,0.7225786491244957,0.416095601361377,
 0.2546032457645786,0.8679170839478954,0.41345788150679197,0.7576791537925175,
 0.9292734026455146,0.2744327699608369,0.16128778270827815,0.5102347935871632,
 0.00922085008169049,0.14976811208034757,0.9938288809988826,
 0.45737807212588333,0.2879777283457984,0.8436242750978944,0.6703541886989178,
 0.7650034588528118,0.5661242806743456,0.467792401634347,0.6570815924829617,
 0.41188582263771123,0.4896149022810933,0.1913603752557651,
 0.09726367691701587,0.5054723186540264,0.1941385862962446,
 0.09972541920606304,0.853776713717977,0.4134845648995974,0.3196713782280667,
 0.3568720023799087,0.5642485970031732,0.5091968901553838,0.7125055292156786,
 0.3363406350789365,0.9799864781236604,0.34491823149601675,
 0.25371876289172357,0.8699329055145572,0.5860768863010909,0.9446800588814802,
 0.709966862984456,0.8510002652183687,0.06188491687707609,0.11150678346213705,
 0.019570277229149724,0.3112697691256132,0.8231260947280524,
 0.7045903782799957,0.6448305248182317,0.3526520683259261,0.794576816877636,
 0.15864371959703905,0.2268572693984643,0.4912196816698098,0.0383052941048333,
 0.9334134602161521,0.7868825258480977,0.5700763876913038,0.32598916173151626,
 0.5338019221289669,0.8926987974182624,0.8385360127258212,0.23202297481614031,
 0.9249704871900023,0.11657508717727416,0.15356151883520874,
 0.6618999027866124,0.12444973690528538,0.20230021577922075,
 0.43170898309619277,0.9853677744507924,0.21000131680409695,
 0.01804980753380392,0.5988515127559497,0.022238278343049145,
 0.37072667972746043,0.7042798934617284,0.9977871116818384,
 0.46140007401365457,0.985972535830984,0.6218262508654158,0.16456124260125915,
 0.08528274320749019,0.4313282231079183,0.640078946719333,0.16957490059366132,
 0.9442204287179103,0.37452352341970163,0.04174703127441415,
 0.7936131716060018,0.6453498986731401,0.9500851812956214,0.5625660114459394,
 0.9848834380760778,0.9354523369444917,0.4255408520898435,0.22396741596930736,
 0.6830419565063259,0.34030595070444447,0.21889902730067146,
 0.11848609192216775,0.6735535983036613,0.835269237347312,0.9496608368220957,
 0.3956386089516034,0.8555290607341476,0.30457903943336206,0.5588063311106413,
 0.7826953342854852,0.03457224245912416,0.8775838378661627,
 0.34800126372305884,0.7931982498421207,0.011260067633704574,
 0.4858691601425613,0.467814472839316,0.9927716046053886,0.7523791620744749,
 0.19182090193021362,0.8706287318223485,0.2548060567578233,0.5531871471192935,
 0.9795290646676715,0.8409016319471112,0.7847127495799481,0.5558398838445573,
 0.8589211415408757,0.8777434320299059,0.9547489457858108,0.9854523954558542,
 0.5705887655461878,0.03022683262942827,0.2268314275280897,0.2155464276794471,
 0.2165345786484878,0.707875859633912,0.6991338851888482,0.791614288691372,
 0.9333308189977294,0.946455937737412,0.7413685986341312,0.26949114909257876,
 0.5670376455905974,0.3206841777070468,0.5038763240263602,
 0.023961965555577835,0.33704685664250067,0.13317551429812124,
 0.7343855804359181,0.501838269685104,0.8534066736424597,0.08208630743322387,
 0.1637171411614724,0.3959433815932579,0.8657364739627296,0.08206305328419183,
 0.33848377202642543,0.48650325960611696,0.7459139304567014,
 0.6941773069089271,0.07715025359863803,0.6730762175614213,
 0.26520316111202336,0.8800434889896886,0.3516493632306975,0.699889271275049,
 0.4128946293869652,0.6967769059475288,0.885175899211587,0.7339477261277254,
 0.8015408710083066,0.4957683349749952,0.570083527072262,0.6584544633055918,
 0.7787265769151709,0.3836723299114482,0.31149520008762754,0.8586173065595355,
 0.6269637931853226,0.2669404140165902,0.09919502001898772,0.2894275385701387,
 0.8852855371091207,0.5065230623448338,0.4235930986530726,0.0313488744829431,
 0.7409187085795279,0.35385537820144475,0.30868937134469987,0.992568257883119,
 0.5203671046518383,0.5402365049393206,0.05741612846677535,
 0.47756402133214026,0.8494000595642559,0.3080269162429098,0.952171948649217,
 0.10398432316887152,0.9892173681281219,0.8292815752620862,
 0.08625697879992755,0.8581530597639179,0.5991355600961317,0.4834463868872503,
 0.8050822436311617,0.5739950942121537,0.031495846366400126,
 0.7429893730765293,0.19928417908981677,0.045414427362439946,
 0.3333670908304749,0.5900389873248847,0.17047297282608087,0.8530609634741931,
 0.9039762931439959,0.06441869596352945,0.9289611919079157,0.5703814164951058,
 0.44026564743648666,0.200360933279081,0.3250537088578521,0.34231382090056073,
 0.6727328482181343,0.07825826957802895,0.14095174262698174,
 0.6383443631369654,0.41789538385094405,0.5264573861560761,0.873832311983281,
 0.05304074537108516,0.8537141888436635,0.1646884242524067,0.6159692947191008,
 0.8772797444943712,0.6699243315756124,0.6356925297084951,0.346774152524013,
 0.23193569984084184,0.6573469761834089,0.5967339271826555,0.2441323429883081,
 0.46790707775003515,0.6802977984542242,0.9393600653141446,0.9483861465335055,
 0.057960801451808264,0.046752702583142125,0.7689079859224062,
 0.8483524113665194,0.02767146291304934,0.9540445248931124,0.8050784224842014,
 0.016546889798719322,0.7894101594097129,0.5465598626004174,
 0.2199083551286828,0.8810472540431447,0.8946483165271332,0.6709404791952089,
 0.2757622559488675,0.9932471883356954,0.23307491751414866,0.9435128649715832,
 0.32482953006107873,0.3148623723191898,0.40842910154093715,
 0.5791473537219247,0.19470580912897395,0.3947076194567628,0.7243376513029796,
 0.6862588816762222,0.26801486510608474,0.5013495495117599,0.840689944506931,
 0.6269007550434141,0.006933794353361167,0.18567462439405336,
 0.6720297258537528,0.6481729810888698,0.5756416410307604,0.7517069062317254,
 0.1665165150172605,0.2352932900593906,0.6654341058216224,0.8379268122198954,
 0.9569358381246874,0.6149368850522154,0.49538552119092827,0.5074056485445946,
 0.7881190642473409,0.7942075981240031,0.12737323878647067,0.633002016925007,
 0.7968948034466228,0.5284173930339131,0.5636943923105265,0.21515513753674975,
 0.26336177741626887,0.8537488873870913,0.7621890958141958,
 0.12865802189415032,0.23378579687344714,0.8507055392010994,0.734278506615822,
 0.95907824369251,0.4122648495557315,0.3776771903831593,0.7486252997526358,
 0.006547111832046815,0.4808141089159568,0.7779935257428072,0.458544281762318,
 0.7900859168462582,0.8817579231467071,0.7546224005346129,0.6412870989435561,
 0.8554568672588154,0.523981583157282,0.7623309418085941,0.8258462381074729,
 0.12647042415953869,0.7485148825837646,0.2704344705148831,0.750697041464872,
 0.36138470726908667,0.6723530923296049,0.4749625192856026,
 0.20492729362752693,0.28416791548846543,0.6845602935770346,
 0.6794268050584723,0.4903297065416463,0.41296398938656353,
 0.37949967370160675,0.7908040471619707,0.3301501577591047,
 0.17241242248344224,0.07651183660179128,0.8018862792472558,
 0.7223967937131542,0.03928090325803568,0.023704815811691482,
 0.43127175710185095,0.4352991670136963,0.3182704451152657,0.2860090486331789,
 0.7408518182663647,0.21912751581590773,0.047418637534476615,
 0.22723477600254915,0.8640990995229009,0.8958468158222298,0.2746598293208514,
 0.13645305312198996,0.3945032408413178,0.1259191975859485,0.3776782786577687,
 0.5582670291518339,0.86288392706119,0.25611344698315597,0.8642929451180763,
 0.8394260612608061,0.22390487621559718,0.9809588853652712,0.8538750114030151,
 0.060263249788008455,0.3212285247980966,0.7145169925516538,0.612651157790871,
 0.8149202363300536,0.07141465439274353,0.8666241354527064,
 0.36388551031734184,0.562070161444406,0.521514682601869,0.923957315959334,
 0.7036131810153093,0.35117949356558364,0.6887057441246814,
 0.10182229949865351,0.09428284942653231,0.8382923821917924,
 0.4641900411825036,0.4415585493626857,0.4241045631142071,0.05108215847908504,
 0.24395349800699795,0.6219761274574623,0.8341785963622566,0.7316727082942741,
 0.4805879060569306,0.07357248403458172,0.6893635779058237,0.9284184270233105,
 0.7582858291652925,0.11168909777112401,0.14441481368064657,
 0.9115941679758311,0.6983180994646572,0.01945810876596854,0.9918234419355464,
 0.38555000154742125,0.8840790377221289,0.26021526543756424,
 0.20438355936461972,0.5052368758947862,0.14017440560135053,
 0.011150620646530607,0.8649198221679916,0.4793352103043944,
 0.09412999989560245,0.35890782385562114,0.9796992773819049,
 0.7440260609676163,0.1456684109854285,0.04719104626628556,0.1357066350912305,
 0.49058906722220486,0.7202950003269681,0.2982110163695647,
 0.17264987807827026,0.729576570790919,0.6613388820532137,0.5129651312029294,
 0.6142924088033559,0.6756680260854309,0.5132727849506958,0.4706569702209632,
 0.3023527457022973,0.6396469365010508,0.395505954226673,0.32710578749984087,
 0.16328263744294158,0.9763262624133328,0.3413098215364332,0.1775542891667139,
 0.18625765499001257,0.9188539046959653,0.4889500095199051,
 0.011319307922241895,0.5995821355495683,0.22301103562903613,
 0.2866319301063509,0.8866388765671331,0.07824451290520273,0.7511939331097495,
 0.8229209761177828,0.16444390344509896,0.8409613603304424,0.633306480110836,
 0.433109752742744,0.8525833076914762,0.03919402348181544,0.47196093923497273,
 0.7179520202225896,0.12441348997136314,0.9488392929467012,0.4727804658524347,
 0.3219296051621505,0.35719108492272644,0.8153089095685796,0.5379324095767117,
 0.21560249669221543,0.5426438648810828,0.4802572432515225,0.7500733766427113,
 0.2040412235147807,0.8716399312147954,0.4102166411443987,0.2976425347494486,
 0.16406005867179252,0.018023367015119418,0.7973504457478233,
 0.4184092837025968,0.57992196248011,0.06616483130910977,0.6751453795724339,
 0.20070379624361923,0.7767296425900333,0.8172628080115376,0.2358059738994528,
 0.2533063412316533,0.6744853289639711,0.5348459244040841,0.34601038856499966,
 0.763308244178819,0.9560210419183574,0.8413120501809911,0.6059860065814626,
 0.7132352553988833,0.5007974120195623,0.004018410823282359,
 0.3994045834891128,0.38559483559207897,0.40632059432201073,
 0.9468630346856239,0.21175199427986713,0.7277491665225343,
 0.46504428044282253,0.7686062946986947,0.4783914156558098,0.7325722365917021,
 0.1418000657895938,0.2621499972500698,0.4961033895998086,0.10820664992915663,
 0.37401023739400774,0.6873842826494552,0.008047185932611112,
 0.8722950349409634,0.5440121456792767,0.7858630484866462,0.2877466525622491,
 0.16360122743826744,0.32368279715429316,0.8596672320281189,
 0.8261508697473811,0.12855233801929133,0.08778857225515002,
 0.9939888634871448,0.4804038367103822,0.5272359588013688,0.4743045422239883,
 0.8056654226310473,0.22184097541833003,0.33624279776510546,
 0.8834849110721791,0.16023281383472698,0.5747025754052474,0.1347087498106978,
 0.5834339646464517,0.7400612111579206,0.1177531048595376,0.9094374794344716,
 0.13051092118447483,0.75369486457724,0.07564884835747243,0.6666189198291621,
 0.6475085899950502,0.35693147721310337,0.6948156455640624,0.8729498301294657,
 0.5929165530712384,0.578663082386241,0.21861270067322702,
 0.013610525884134916,0.2603810783792768,0.3142293352626766,
 0.4778896756399066,0.27930192725036285,0.5234731731262654,
 0.27406823011517134,0.1868295364202921,0.09977793634581578,
 0.9233464834324552,0.10101073564476015,0.3408682165625787,0.5538881329284314,
 0.8006729075842784,0.31556641721026446,0.7830247895097877,0.8498874231069274,
 0.4186439502554745,0.6334529350272997,0.3072877626097569,0.5388749543980625,
 0.3370962007577689,0.34233863637583806,0.9470396278975515,0.8691422270680423,
 0.2695042031466538,0.757937199936517,0.01033188313794653,0.47913594979058394,
 0.7313916867196792,0.7210438897537844,0.04336638235792989,0.9853473162900355,
 0.23966359511811708,0.12509778208349154,0.37899625917121715,0.44571365502311,
 0.612763034314493,0.9877922142840981,0.3172234004285127,0.9588662349115368,
 0.1345046184722003,0.6758340152574174,0.041850458951797176,
 0.19754034011256183,0.3057311770892268,0.3249912021665208,0.2291360699514473,
 0.09431076299727092,0.09874072134574818,0.7884358065133791,0.402933610735729,
 0.06054671304496484,0.4688773803437848,0.8822017595318753,0.4974030584782011,
 0.19867819662239494,0.09859295718003147,0.39679460248093246,
 0.9393025601535866,0.8365275964170764,0.6015355240321999,
 0.017028345420033508,0.6574008703783437,0.9089266034149258,
 0.7274964715967476,0.5275532796428961,0.32940841864627957,0.9782041075295576,
 0.11207222447524134,0.42618468446996316,0.817370302040043,
 0.20208766698311298,0.49541999635719414,0.03800022766019795,
 0.2972568452751725,0.5168544611207017,0.3656386626637459,0.5592245735775447,
 0.24004787786053716,0.2782919919795508,0.3340980156734472,0.9546149531232748,
 0.05620979519451452,0.6567423253990016,0.3141270988985392,0.7651279396338742,
 0.6341293442948469,0.4709625383393925,0.05258029052887381,
 0.05485846792979299,0.2728215252481716,0.8402458241695994,0.6037085041517599,
 0.9270439926359887,0.38797314304560004,0.5836084485106603,0.6556449950290245,
 0.41149695626076754,0.051981757592358235,0.7167139166619635,
 0.6276932568382939,0.4397687411577482,0.6752455226304709,0.8382848067542015,
 0.7481574073285036,0.9761922923353608,0.5134402945626001,0.6382751376144652,
 0.8848576995888534,0.72948627005837,0.18260915658099364,0.3351647275792137,
 0.7381169213920358,0.9374394045737897,0.643685542373949,0.6073708864583302,
 0.6305969473872773,0.45439879394183114,0.5859477961074004,0.52228069232117,
 0.804172679890014,0.7187761379394746,0.9383837795436315,0.1477378495719337,
 0.047893475074037406,0.12875757740107052,0.22464179447659394,
 0.15827607138133304,0.7166488612963551,0.6106266252779573,0.3264640504063173,
 0.13071607770094051,0.5979347172393593,0.6119897778784905,
 0.005497176218986377,0.9207741550408844,0.979202184455496,
 0.24825037462597033,0.8729112414763736,0.031559033371472145,
 0.9657348166369509,0.030756578382771793,0.3542275299724864,
 0.1705762803926485,0.20286652363115332,0.14385060995116505,
 0.6631911721808407,0.6393802788338777,0.4445448996075725,0.9764475968354321,
 0.6461828137853914,0.11201430475356022,0.6431251629935615,0.4061244941819234,
 0.2181086084626378,0.08506542268467165,0.26001731926365046,
 0.5360878321768854,0.6354422414792957,0.4867494322061472,0.2527756433470174,
 0.7147698015786637,0.9868106868716922,0.73339607280242,0.46826218049055823,
 0.7321291201930005,0.8161472992806341,0.0799139909637061,0.7512746850633799,
 0.6907344067169838,0.7372907633750734,0.4338774467336839,
 0.061860715838298774,0.1604203972688102,0.9024749517616718,
 0.25196219190550373,0.21182397282218435,0.0038325251198656485,
 0.24960273586775383,0.0231962603446616,0.17440891837526618,
 0.9908069472776683,0.7342909593613032,0.9859008294780488,0.47085964204758335,
 0.7877219603261454,0.9795093800823255,0.5011049253579103,0.7648340786162836,
 0.0325921430965217,0.5495033713205317,0.3294906048866638,0.7122123640977072,
 0.35661162150920733,0.5985522405038255,0.26322834948324025,
 0.2026695090414513,0.734418587512668,0.27549911016414264,0.5122805419121662,
 0.04480973144793676,0.21810942974515246,0.5260070609410642,
 0.9238491255669946,0.561640627236341,0.10636722047146141,0.45582946097681565,
 0.21651786289683073,0.9208731325175624,0.3644280950571076,0.9975222356973228,
 0.6249727017091478,0.4565336665586789,0.26136771550003124,0.780320864169231,
 0.5673577852694014,0.7428449535993803,0.023746960287590158,
 0.9727760959508633,0.2367990375827339,0.6112240567937435,0.6280526064649484,
 0.38255536240439136,0.3405209510977043,0.5609728078009086,0.5234522182229391,
 0.6549444987796602,0.11034120545454229,0.7696869580017462,0.9074209694513448,
 0.41995947142951306,0.8265892458050512,0.003365301154845457,
 0.5019545122735015,0.5834780865411396,0.014549363194862108,
 0.14169103567742225,0.706411450880777,0.15137421074887591,0.7065367013203754,
 0.14653087864928338,0.8740042154831021,0.0014071241308293292,
 0.7670903789938889,0.4356129635847813,0.11370254590398265,0.6719010351024872,
 0.24270060298509588,0.48632084535216924,0.12615990469219918,
 0.8250749657378373,0.07151336354762794,0.659300760400857,0.5052072695330705,
 0.9871215568684224,0.6463731153128005,0.7124986231892362,0.867874796977903,
 0.2974954916568312,0.9723513826453534,0.32190905144192006,
 0.22828994165619287,0.45245504288793437,0.8955992857220236,
 0.27177886028384934,0.6940077839919067,0.05732398713205289,
 0.9426626801877461,0.311025743528492,0.09934585376592808,0.48404534502346996,
 0.8091908687065976,0.19143927666432803,0.7479853851336462,0.6822447858429035,
 0.7667691184157863,0.5982262195678194,0.34424808175092103,0.8881254885775689,
 0.6372839161646239,0.9333964969013334,0.4864027965042894,0.28956361336868164,
 0.6452439475785801,0.043335720069089256,0.4464312958088873,
 0.5348733961498912,0.7830804778127408,0.338176825185713,0.6477303396364161,
 0.2837507833490749,0.0794109327856275,0.7465892604478677,0.09733611424345112,
 0.6519633691848769,0.12236431835755024,0.37919786675630984,
 0.9081686107150597,0.33545946848069463,0.7737277723629099,0.2797183270490933,
 0.3964627724638201,0.2635108852530238,0.3559486535241506,0.24950154460137297,
 0.28416638175299236,0.3580129533243378,0.7731808909390379,0.0878478919999024,
 0.6353960097197021,0.769653836569512,0.6464608772920512,0.05292849962513646,
 0.20732819820542459,0.3320898698026984,0.29779918897708013,
 0.11313659226757256,0.11636770821206777,0.7571812005769125,
 0.1637303286386862,0.6479272505318157,0.8176951882055705,0.5285960968273887,
 0.36162291423472237,0.6143564311713018,0.09069018419997077,
 0.21165468008396426,0.15709662257141,0.4821853744792811,0.44015934829914805,
 0.3041630960560222,0.9661850200689436,0.9144473031756523,0.4820462565705407,
 0.3401930454831197,0.1210656330962755,0.37946345833059214,0.9475910927646912,
 0.7872713022621977,0.6705115989730482,0.2972851155478434,0.4750037753187113,
 0.7690389170255328,0.9818481170300876,0.7874619544332383,0.4172840022423946,
 0.02371850543294385,0.906008355559341,0.9955156413932817,0.02842060837437721,
 0.8971311649442995,0.36635236183515196,0.9881104610062079,0.4190539561699129,
 0.9005251247782464,0.5307129192866089,0.27945543105642145,0.6918816008348195,
 0.11353705986400697,0.8205837260498753,0.3420357765501131,0.6170485834516517,
 0.7688184113134938,0.15938518213997444,0.056600003410334665,
 0.6607958790707164,0.1266473542856168,0.6630658044163023,0.4066463688989067,
 0.001085386152725465,0.19543628387745549,0.7267193162802443,
 0.7364195874348909,0.33614044731943105,0.8921455963045966,
 0.05650597966743276,0.6724672975985572,0.5705015152163522,0.6016773475899804,
 0.8691671990097608,0.5368936245908172,0.32671470207437303,0.6654479556094954,
 0.6773787760627552,0.317886811686135,0.3038465719347194,0.5081594097809848,
 0.1736090166133163,0.26021873608683666,0.5395677455966503,0.3321307276823875,
 0.09911678412095659,0.9559429990395841,0.0662427121479876,
 0.0074895101529326435,0.26330811988833935,0.8993972878016726,
 0.07762772353723901,0.07599716447124338,0.8761184315253893,
 0.11218416469358339,0.8233055675631139,0.8293423383713836,0.5308552538957774,
 0.3969661522106741,0.2726566524984916,0.6020625278344467,0.2284947220606981,
 0.6680658862247333,0.6420893979400522,0.9797641470373545,0.28467276414424514,
 0.7386471136109309,0.8562997344371176,0.4260182199240822,0.30925662902811824,
 0.41207754821308673,0.4855367659722346,0.8100891827725181,0.8981718730283769,
 0.11358402494556574,0.5169320960577187,0.01944409042229789,
 0.5073536690776822,0.542866135367042,0.4332823545440505,0.7594897742370521,
 0.30153600647244905,0.4815474667630587,0.9192445688451498,
 0.23847326817610248,0.26930165998882394,0.7326099297964739,
 0.43491106305273675,0.5972033787332445,0.40135235390424806,
 0.9443430000630255,0.2970410596389509,0.4501078487736031,
 0.0077704132900447975,0.14796974564187015,0.8553526849378927,
 0.21173059931299576,0.2460436887142723,0.6403613277223249,0.9218155315232349,
 0.2744814785863402,0.7890940817560702,0.3108734235178535,0.5623373928397798,
 0.6920428332455355,0.2995835367637225,0.9958799237517163,0.40325178700872855,
 0.8626692048879089,0.7263979962436787,0.7188762219403526,0.6114690039488982,
 0.2661626138029294,0.9193284892771058,0.6858828357378413,0.6294185093251636,
 0.3502322412048138,0.5136220597640857,0.9121449726325374,0.7054687777902987,
 0.2552239934241072,0.41630700511053687,0.41411624592943463,
 0.7942881528636558,0.8786179866886309,0.285905557935064,0.4583556166907711,
 0.05547298607786688,0.8491643768930421,0.02302994300692085,
 0.5539055276641436,0.6796936244013556,0.11524216445311264,0.7935797233271091,
 0.2114645168189906,0.46676165512023726,0.13874988906836194,
 0.5886732596774594,0.1296862834454715,0.5868295683954636,0.4831927920004998,
 0.8510758052408528,0.7082832497300937,0.8775221086432372,0.01824342546361646,
 0.8043340025363552,0.2600509742918292,0.10014168113476052,0.6341950766790219,
 0.5114172752941194,0.8944836589221519,0.37610187430430186,
 0.45607550003011166,0.14667484149559906,0.5996975180811712,
 0.5779971403256994,0.8893420450616765,0.0732262949975252,0.6431739502919357,
 0.24388913404601786,0.219851091283496,0.3261772816127866,0.902374227603364,
 0.13809035905973,0.16154588201830322,0.5317496230522697,0.21204217861589447,
 0.541727494812057,0.6751317690781546,0.004519638958348393,
 0.012111416714732481,0.6460436223312813,0.08210809863636392,
 0.22062863723677928,0.24726077471268326,0.4570960546223237,
 0.9249902521097606,0.13303400714029157,0.535766932685124,0.6495627358655638,
 0.3422352349888945,0.9749243861869634,0.6472612609926588,0.15172022658649542,
 0.9704694225993165,0.6072695672588668,0.9003305562129046,0.21637408224020238,
 0.056469859347603624,0.9223672018556197,0.2304790595640609,
 0.7102082814713011,0.4604366925933322,0.3539777375152866,
 0.027299995745730388,0.6077541134251268,0.9135148250286964,
 0.8480948075746537,0.01128585255075587,0.9148845765501,0.30535399189293344,
 0.32711199726507023,0.7274439972054844,0.5686384420390719,0.948535775938151,
 0.21299269793586162,0.2648336033939013,0.8963013927811172,0.7947278126742959,
 0.44323461162662803,0.84749336923312,0.09628217573082254,0.6975641698417334,
 0.32128066646710796,0.3379406521949142,0.8920115102412477,
 0.054377241498121065,0.8691075394551858,0.2235801914628187,
 0.49161878731917596,0.08654838071257198,0.07159067887645887,
 0.7987070755486009,0.2826709728109452,0.07971217992880608,0.758903407567978,
 0.01759310091692634,0.6533214248417257,0.5708764668845354,
 0.31412609431704086,0.9781683684176181,0.9145783870598416,
 0.47494072482945826,0.5323281801248326,0.9394114801861185,0.6021199916281759,
 0.7366455261611247,0.4173999232730814,0.4515132479239524,0.3858831640659579,
 0.5665666710318013,0.44222159190398314,0.22278553528487655,
 0.6514084067217503,0.4761621616133258,0.010088686457214058,0.586028015323687,
 0.21950272079046085,0.3234214727530441,0.8778929030073339,
 0.05050478268900338,0.23358258676957022,0.5056791130842906,
 0.8261762796070962,0.7166240669436594,0.7630593053507781,0.5536359042213456,
 0.9058607431324122,0.4397221388506355,0.48951147697356234,
 0.44388578784032573,0.8446618287866844,0.06328906422914737,0.840910906844634,
 0.15908296816071887,0.09448665152284808,0.8549660086030032,
 0.1837154429889355,0.8381979560599307,0.9253582108588101,0.5560182794654811,
 0.9145746857109165,0.2037777509658688,0.3384074554096643,0.5574813993607249,
 0.5624036265230671,0.2955903907614378,0.21937666286900115,0.8830528903834278,
 0.15964792903503333,0.8815587388740126,0.9269778480698586,0.6953192233521728,
 0.059726879600049854,0.7705371996634175,0.5469620701149405,
 0.6976913341113015,0.5562162174369711,0.036329484636707154,
 0.6834153692944419,0.010898423530668833,0.19488504740579093,
 0.6941836109068995,0.9962795112967042,0.5690198000225881,0.01159584060578367,
 0.2750707884665171,0.022690145451985555,0.7832974948326038,
 0.24189961701361284,0.5894040243390267,0.6832625676659134,0.4102558696580413,
 0.6481314430649697,0.3007655902370173,0.44548813709179824,0.6851556102771742,
 0.37271007838286174,0.6516669131797832,0.04554342977298642,
 0.41067477108207695,0.5814225266710196,0.15519074689206724,
 0.7351954218545721,0.04568833082537149,0.13582935138474117,
 0.05605520139857101,0.7650214633470367,0.4037244023830123,0.8450616656778549,
 0.6097689079692848,0.9845023281311278,0.5810283722957248,0.510842680670534,
 0.5965403010682316,0.49524347520477807,0.10775687676810308,
 0.38013447240074605,0.8933605675632069,0.16420970259144263,
 0.7948729326335917,0.9401325651112032,0.6699760537335815,0.9101307845419343,
 0.5107125582099321,0.3528310888844276,0.9053503600818285,0.4042466842528095,
 0.14194803554245616,0.100870688845491,0.3453734930897341,0.24745071702686927,
 0.8456166359692519,0.37125543434009445,0.28035604498801714,
 0.7361677176564462,0.9738844140677535,0.7301109489369816,0.6825170426235844,
 0.2064238113354362,0.3835219463095215,0.4562875722457558,0.5389550545332615,
 0.6901573711050559,0.10144951689669024,0.7542170035021349,0.171466938180994,
 0.5261242928482267,0.2032992084819979,0.9327172871654275,0.6845717501395113,
 0.6482498583512016,0.5526026681426467,0.5657854055917698,0.7729023286058853,
 0.6317096016300758,0.432026014075537,0.1708193389881174,0.3744126293334824,
 0.02736063105675124,0.12610587202310308,0.6675786284851029,
 0.7488586964453648,0.2311527969250372,0.2246912422527525,0.7932143065422445,
 0.9160035117237628,0.7879224442045653,0.042746514344243924,
 0.6063263575262832,0.056822259050715695,0.8092816083905384,0.827481515746983,
 0.005242290078366296,0.8892020429438089,0.12778710239067936,
 0.2659424142487361,0.5025269156722999,0.20374776940847406,
 0.43696004311335057,0.8714734868802945,0.4766366112019864,0.5476218485312021,
 0.0637333718289731,0.2748181779028017,0.6150069430986845,0.9539924897741658,
 0.37946839330029736,0.2647262486804708,0.15732206607472687,
 0.37708303117623476,0.8383970330015159,0.035750806786774114,
 0.4679646769933944,0.8674769909926453,0.7772742354767556,0.35630150206984457,
 0.7001198940575247,0.17019451252347872,0.5081882662357126,0.822976785535026,
 0.5458274624634594,0.7570748876442335,0.7514440197507354,0.6754819419462307,
 0.8730479962440488,0.5765066815057315,0.04823272445187021,0.790560018685983,
 0.16847824106060527,0.5983218809819858,0.3674083020014631,0.7110811229667097,
 0.2987558064052461,0.4605088600144659,0.2141329456192551,0.0949241077124352,
 0.45892608246694633,0.17387093307672286,0.3477600423669698,
 0.2397439095619578,0.2526551437414467,0.4354934557867185,0.9227826562135314,
 0.5257013360118621,0.5893193482116087,0.7256919579595049,0.37224551133484796,
 0.3777676698407516,0.1079584739402979,0.42680484844567756,0.8236283029955147,
 0.1867782837614118,0.27863054145161026,0.5707101585472135,
 0.06790960479324148,0.5790253645437566,0.3238902212438174,0.7092895220867983,
 0.9762440104996941,0.013274184548064505,0.2776554842432568,
 0.5431205931516253,0.7018810307326826,0.31740159674257884,
 0.03750294935358578,0.566161147909827,0.8981040483819956,0.5120530080711116,
 0.2850424240914873,0.41973081820253255,0.152686010294951,0.781466751560114,
 0.5757155198121355,0.8815153122453366,0.2124863778623024,0.2009504761074421,
 0.33119181965103306,0.1319275178621041,0.6595487363216087,
 0.12279551195239702,0.9118664323055001,0.7781625614483327,
 0.021573796826972913,0.12175528239414624,0.38845224945678425,
 0.8550680296453413,0.9238171194885397,0.9405607505383852,0.3676130189392154,
 0.16240320929480523,0.1338810417964722,0.7013255202764486,0.7646912183932616,
 0.6466814415995569,0.2547305691003866,0.6683210728810878,0.5424240116906298,
 0.4433053856122505,0.3855837256788964,0.8967898393128487,0.9393266879409726,
 0.3033063194203448,0.5725439338443248,0.8108549932663603,0.324195033504318,
 0.636044249349328,0.8760677621938863,0.6088080558795457,0.7475158485192062,
 0.8956623735459889,0.0351928203405687,0.5973424882910091,0.9055319348661509,
 0.9479954314642817,0.8799247997415693,0.05107059113324541,0.79573979639224,
 0.6001609791694854,0.8611417964156368,0.06177880062295105,0.4266115059224239,
 0.3953211645692415,0.6715470869712524,0.7665897538675651,0.366480321754295,
 0.4463783470252515,0.7501386345671154,0.11383057692068321,0.501725561643985,
 0.5263683298419803,0.8865377304760332,0.40771569141322483,0.8129298275895702,
 0.5340153650230955,0.45593805911161756,0.8603256120288165,
 0.36541494674788355,0.8749637569377977,0.584275557703732,0.6534937360215306,
 0.09708497848388009,0.5008083700968423,0.668882669427655,0.26690627103277864,
 0.06840720470485395,0.7204172428759783,0.9495992443657084,
 0.42090264099016617,0.22483542248401933,0.1281756377981007,
 0.30223799042582034,0.44977789619312947,0.5435368118347086,
 0.8108280662425037,0.2979887240998915,0.8480523150924075,0.30697361913926735,
 0.28642217078649834,0.7928744508750365,0.39676724370144334,
 0.38133440818835573,0.6244367232569357,0.5727745913225695,0.8745034361951838,
 0.3455259168174998,0.7519465627262567,0.9142078465565999,0.9757196165375379,
 0.6660488301512766,0.7210160129464915,0.9732551125449741,0.489657513731909,
 0.6977723067604419,0.45377521915969576,0.12459515977758207,
 0.012114484806515113,0.6892728262441682,0.11704099178783611,
 0.2974236098368076,0.689538540201069,0.48774305568820064,0.9414942103878069,
 0.6557955177543772,0.2897876953951486,0.5867802649870214,0.5117893213349226,
 0.3284767379273884,0.7445271500186896,0.7636028180914807,0.31066955522169504,
 0.6521110422144983,0.6117620409653044,0.9269310673225978,0.3171923902944739,
 0.4436066855810612,0.3633175673600362,0.051949408390699836,
 0.3121733759463612,0.7161255266073248,0.07692423938494342,
 0.21787816024022755,0.10413374650769758,0.7623084409626779,
 0.47474692064934154,0.5237009809100758,0.7467175325286942,
 0.48925635586251304,0.949066820663302,0.891258098981283,0.15590062671671745,
 0.11864659176360648,0.13572061083239984,0.24308181623157687,
 0.21318370088105298,0.5497993064513604,0.7205467982515716,0.1593034405471555,
 0.5384666931150656,0.40867520749996866,0.030082460793291355,
 0.3697199999231734,0.7258930730691074,0.46375230587224214,
 0.25089594025217465,0.745915279476646,0.4817708610381508,0.18901097518997134,
 0.037418395883221045,0.7275165526906024,0.6116467092552955,0.765830549801277,
 0.9266595431929087,0.16016287990475975,0.2894902021043042,0.8848059500750601,
 0.6945495167546359,0.5646259910839829,0.5641624321797689,
 0.011710225897747728,0.8422358248020994,0.5769996971561364,
 0.5999817386492939,0.46445944337116674,0.028665771129093076,
 0.26113364954653706,0.6903652414309155,0.6112933198187179,0.5385353222467956,
 0.12606964732389137,0.09189332255720317,0.04161410091424955,
 0.8725584779180287,0.8034099548046539,0.44154398110317183,
 0.05681516272606013,0.20725338112498115,0.5199013780940799,
 0.4587997147549381,0.6183870559385596,0.010502158482251556,
 0.009714050627029014,0.22822243340975445,0.38955916505299537,
 0.46087008598525125,0.9197725511296364,0.6566148076083503,
 0.38077265474967437,0.442134288445717,0.4033553747973877,0.7584319259739795,
 0.6250489946087416,0.9720400820378876,0.7730887011529374,0.6583205644420289,
 0.088817481252905,0.603928662244425,0.0399031272865763,0.30255676081366223,
 0.09646536494503188,0.45026422748878214,0.6886028839919747,
 0.32357656538570434,0.436537764083617,0.9353741632744995,0.40659538134453166,
 0.15462637711922078,0.8010404739977797,0.14399710274488664,
 0.8293508566904062,0.22948969747978587,0.06271017048219973,
 0.4738533858004599,0.10834241928785615,0.22906496803856513,
 0.9513313755499183,0.428556956358196,0.9089878764482048,0.6778297105236935,
 0.7714536562563025,0.9904731556808366,0.6663691338354776,0.39011366743374853,
 0.4396935084927408,0.9110249901495071,0.7385851723498095,0.5956681937808987,
 0.8689196293430398,0.8570771991757097,0.8407630502105559,0.7369056684751855,
 0.09105400850877199,0.38930733803585627,0.5930262167321086,
 0.06850242379558369,0.7460442841458185,0.38799377379307454,0.626504962452227,
 0.6155780239627335,0.7826060002278922,0.3118856719760146,0.28408198164240084,
 0.5775233270041477,0.30546141648051706,0.5038942595940755,0.5974759448851663,
 0.13041792042744593,0.3335801499401603,0.630844502486426,0.10482095545204506,
 0.2654478604142808,0.6398031072341693,0.4898709784788877,0.6387528070163372,
 0.9378515008862056,0.51088822673408,0.1311760811413314,0.20661659622714823,
 0.0026161555457193675,0.937528460523107,0.3435414256081315,
 0.9638807585611324,0.2333779544070671,0.8559409647984774,0.10999846670471602,
 0.8755713143844148,0.9656214754497368,0.42456650435125076,0.7312585166248949,
 0.12474350065066753,0.4565421024840426,0.8667299287416316,0.7154205072578399,
 0.9874579685341541,0.22729934356743742,0.4152573452620296,0.3856799024147013,
 0.6985825321696375,0.7584629529230802,0.237283746908317,0.36831049332557175,
 0.8681186343693607,0.9825452512829973,0.151402395518343,0.8125768698857553,
 0.7777041498214315,0.7393073233375969,0.9877053897375248,0.22441242310957465,
 0.8747505649084655,0.4600849167655455,0.5012483079110684,0.8088010786081802,
 0.6672669336960823,0.9270829027626205,0.01933184114192177,0.7363754526879641,
 0.6590319299475839,0.4836769138769108,0.787590756082121,0.6877655687738371,
 0.70099798407563,0.902500663691701,0.9678053267994702,0.6161465063502973,
 0.1502371554622468,0.9961403484127311,0.8718894366761583,0.7172888611461556,
 0.18844986914814488,0.005398868776318566,0.4898073486214678,
 0.06802980539042303,0.30363043879400575,0.8527291621652131,
 0.8586093758705402,0.14526331974440376,0.9200198998969351,
 0.26499594822523675,0.3947231055403503,0.13601837646840287,
 0.26623803798259005,0.19215296519441105,0.10197381451537857,
 0.580881082098629,0.7384042799003574,0.38992210973234265,0.502910370454654,
 0.21485210635992247,0.9218627780388852,0.26317732585168374,
 0.9296425300193472,0.6783518982133009,0.48698232793811147,0.9237566736432505,
 0.8182340312793696,0.8141166583678848,0.11719131843250674,0.576920666030848,
 0.18537276834495264,0.35974881419118965,0.5534641322923539,
 0.6048560148353854,0.40562586979389137,1.9496106052518947e-4,
 0.9937815395161127,0.4792889170289226,0.37109359221372584,0.1735718903651513,
 0.32904283950130075,0.9236509421211372,0.5881063304868084,0.4512714737527499,
 0.2908369987411541,0.1503900561392978,0.546345103126414,0.6085316872319282,
 0.45996965711977056,0.05726394758683828,0.16920477276657497,
 0.2662286563464228,0.32809894000941264,0.46545670191592825,
 0.35562558626058616,0.5992339439650973,0.12827039077353863,
 0.2744316380381804,0.3743670723586755,0.3816463783787075,0.42243444285647724,
 0.038817192852490634,0.7174149117721953,0.28433913988635406,
 0.9289103601911188,0.8361546834632083,0.021397924036760774,
 0.6961574991950624,0.7626232254570019,0.32524034749862374,0.3595021278109324,
 0.6430285781826931,0.7732528801896229,0.9212777817444984,0.8219166747299999,
 0.8115045340312139,0.38700866831433756,0.019182142161145777,
 0.9620122220166486,0.4467919839413895,0.3269351509902567,0.2153439563139009,
 0.09703748483813301,0.6807270385586576,0.22553270971848316,
 0.7620261702817045,0.7927823115169783,0.06747862666474369,0.949312462573011,
 0.17939994402269321,0.38155086773269886,0.9800962673679744,
 0.16984443279011707,0.9322386618765163,0.9776992824217402,0.1219166371803525,
 0.39317515559897,0.005257047483867017,0.32934156671543424,0.8384622529084428,
 0.7856194007320733,0.8572341513922769,0.8387092152616927,0.45996681340635515,
 0.6807830850644065,0.9028792395311469,0.7195679060474897,0.52588604535344,
 0.19577263427408897,0.5079383959422297,0.27119210174018027,
 0.37086522321272797,0.24431835692000647,0.0015258282964109027,
 0.06345395668246834,0.6445154220506146,0.27580297071937276,
 0.7390657886747655,0.36864795767711,0.06743734893092757,0.8939566106838863,
 0.574078571099864,0.9014467674127687,0.8337643665796739,0.06231152138539453,
 0.34479407061703005,0.36462482098375204,0.44508582021845344,
 0.5316803128503942,0.15966240025483636,0.3153949389727514,
 0.12165326374449936,0.4436453853144867,0.8388554218726731,
 0.10386663877153324,0.4863558451522445,0.32642911161996646,
 0.14505434554693009,0.3091382753590204,0.20127070513913325,
 0.10078746105435932,0.5039505480368427,0.9611620303235208,0.3238753983622753,
 0.363701300407947,0.46865619729505836,0.7588807074823598,0.4157849259824432,
 0.4119034458778753,0.9433442662875702,0.812916335706265,0.04284098760967414,
 0.6500565049647895,0.5424877952010566,0.8863443502458835,0.23560344211454565,
 0.7447768511798677,0.27990897280997606,0.27571841917605533,
 0.47164202333907934,0.9091033959850869,0.448132799359223,0.3647195202560891,
 0.011391359854180605,0.24365021649382335,0.5272119605694638,
 0.43765243059966763,0.8509162799475845,0.11608820390503904,
 0.37233588497302106,0.2130709655566697,0.9089250310021278,0.5750219003109194,
 0.20983927530566404,0.2822328708951284,0.18788919986099062,
 0.29804091724205484,0.22829186553261094,0.9501729749853548,
 0.4657169163218374,0.41003228809413694,0.04697495700600174,
 0.9040057092852287,0.8593070902967829,0.6063346853692528,0.04102425316698832,
 0.3674409090178611,0.5129018861017317,0.23936742948951262,
 0.043009625597939194,0.3969583591781758,0.7658327704226382,
 0.16587546465728054,0.33091183529775303,0.6304521335872728,
 0.6180314356605415,0.05736768531737413,0.6988141734391814,0.7567782226842686,
 0.17916008656985827,0.13465184117623297,0.18649083810167788,
 0.8382639139282926,0.9983565035120968,0.18412179886847224,
 0.10875347471777241,0.2420343103905177,0.7701519592251047,0.5642729340117708,
 0.26059901548042264,0.9404326355781103,0.9341872929443702,0.2795668875182348,
 0.9347854552038803,0.2663267180056865,0.5335897662471514,0.6471248020702383,
 0.6049624254985491,0.7933731354527671,0.3518477097148336,0.6353260746278787,
 0.19988839786768908,0.41476082051079166,0.7898218054993164,
 0.9376736467817363,0.6851496461426143,0.9312603402415559,0.4433341544697139,
 0.2772257228856557,0.27406123798966464,0.029691363784122404,
 0.13816988474724845,0.12441573614462165,0.34397753912622175,
 0.7202511545741217,0.01918771531309138,0.6393800493768517,0.9789238673543288,
 0.3435787584253194,0.5804656341967102,0.8730638035005945,
 0.021900935534961308,0.39258496821588895,0.012245890082126287,
 0.19141974474886225,0.45261044911902215,0.9997058338774215,
 0.6504515627691049,0.36477521790348666,0.7890010492439501,0.629099549421331,
 0.5436019510169425,0.0924545353126287,0.3471267493895459,0.7564734875536148,
 0.15206158521870805,0.29735233162152264,0.05914107028660187,
 0.06163048646985969,0.2487356167291742,0.5659108859012651,
 0.006804066460820657,0.6411713067588181,0.8462453833533062,
 0.12719459768615993,0.9063837585794354,0.9665081918390191,
 0.015486000485428364,0.7202773490280862,0.49656098855266473,
 0.2331672003861609,0.3321095759230953,0.8454661345808034,0.9449338288981776,
 0.304064143355119,0.5319911001737103,0.255517495022516,0.8421986637104959,
 0.9175473315020344,0.21406059799145183,0.3433584803793338,
 0.22983200361513412,0.9615251578393487,0.02377296100970705,
 0.4700416231695419,0.9711874652853922,0.9058711844745373,0.12313640605551812,
 0.999905051857083,0.7726307259807813,0.5102102046601522,0.7455188897162444,
 0.6630026459110118,0.2640417088147937,0.2503704172482374,0.39625241978934156,
 0.6710195905238333,0.3927190242451153,0.2820341327815623,0.5748029592201389,
 0.580293700637115,0.39702504276322004,0.7022327002402071,0.44067854180727983,
 0.08182996469569248,0.37532990837097246,0.22236377564519927,
 0.8824121729242564,0.42720677406249075,0.5648230386565387,0.8446306866895181,
 0.09169027252324324,0.8758364892160709,0.7653255481891577,0.7495002046010116,
 0.49701321186531655,0.2847728253663756,0.04124102058762613,
 0.6452777524066762,0.08715851759042115,0.24324347892826714,
 0.4893623561122733,0.6141438783771334,0.04303577662192981,0.7239283300341933,
 0.7823199933575955,0.3797657992648752,0.9603078312129243,0.23079728707591707,
 0.15366920893847802,0.8862358802638652,0.4429805780296143,0.8210432808712647,
 0.5367052898534572,0.4046584147087846,0.3327200393133106,0.5658016453175908,
 0.38572374982567204,0.4604728048106006,0.24243837993029005,
 0.24744594343206716,0.8403745045811359,0.22307917252899478,
 0.3041664092243539,0.45741650512859344,0.9036140580238431,0.87729954736497,
 0.4516371742967672,0.1999824905456825,0.17991364690510847,0.3569442904914826,
 0.8198018251057223,0.9316064087091531,0.2101129539884844,0.2370853281591081,
 0.09511171376872651,0.46473395618943114,0.6758605181638085,
 0.9758551161103799,0.8456496888680828,0.720955696658669,0.9680546019398135,
 0.48454125840265205,0.35533531997208745,0.32254902470417024,
 0.06702235109729071,0.04261890002042512,0.3246241076819063,
 0.25966654671923006,0.07057131324147514,0.959332673763542,
 0.27316572526509486,0.557810550765148,0.0666662077462884,0.26900762103333886,
 0.3242849555441979,0.5605674264414134,0.23367239813161067,
 0.19997571647795376,0.8709105864963633,0.8840331709556487,
 0.14429391433728778,0.10506445646977935,0.5084024118744218,
 0.06615205904652322,0.7575369293712967,0.2918805725827617,
 0.44960202791178305,0.010248161241137588,0.0815914034106966,
 0.4467020863948906,0.7895956989175744,0.3576256099226045,0.5335218283264143,
 0.49518349103557935,0.3391675605487807,0.49004486408897696,
 0.48631108337515294,0.4532110411645648,0.6895676118881295,0.5458737301978996,
 0.27933213435531057,0.8908329322333999,0.24035067630652063,
 0.8142692801940669,0.18721577551710533,0.28081091840352135,
 0.8256849216090445,0.3329821756648089,0.3269827604739588,0.2486285285918456,
 0.3583910853538672,0.9795124231497251,0.2628049268084385,0.269952524877338,
 0.306804455593618,0.46395245251691863,0.6788272618242113,0.3458777786214252,
 0.4483820043066221,0.9056981155623193,0.21240821326225245,0.5391009737306107,
 0.9109282733208721,0.854345030361578,0.621996422918189,0.6512286454289531,
 0.26519550488299015,0.18716787852341232,0.6987666148881289,
 0.9372206059249354,0.9169480971474695,0.3861954192070757,0.2963930718653869,
 0.5463050279424453,0.17799557845862934,0.16054474592859724,
 0.09440114575830294,0.12124825792369698,0.6595320286644168,
 0.0342501743257253,0.9346378171711516,0.09815889397471922,0.3795502147957024,
 0.914122840152533,0.1214864053942385,0.5561435154784354,0.2807608185593612,
 0.8443188703391318,0.5308554789499992,0.5785004237362702,0.2537164163561517,
 0.04127469397810135,0.4053209082735546,0.9553856550282592,0.2450157244561424,
 0.2793137404713628,0.5841322550052285,0.5123031351736022,0.8232639819627359,
 0.7310700594944818,0.15399667923593086,0.4795839107101503,0.7134979029023368,
 0.01337484954009649,0.6044656648509884,0.6216938017195597,0.977011509555948,
 0.22540719655812747,0.727615223137614,0.8621261289364661,0.5289977749978516,
 0.9965877514520075,0.10636673931305762,0.7774296509449041,0.9969419805890151,
 0.028740293300849373,0.40436822840508035,0.717264557737966,
 0.9845035357217552,0.8237500325998104,0.8219146578556802,0.4721977148886469,
 0.2101987188405009,0.503115906645397,0.49196175268758213,0.809056729749488,
 0.460997081774263,0.3892997404981401,0.2833787540592374,0.3122846407459988,
 0.9993299238209356,0.4084431032322763,0.6069284083465469,0.5784742461592124,
 0.8927911490214191,0.7585266941052693,0.2237772167742612,0.8622100547503899,
 0.06432931277853426,0.11505299926442036,0.7131036206675128,
 0.7675845605293291,0.2709545788118657,0.2892771556030487,0.20334321490204632,
 0.827570739383757,0.49405415431146193,0.5466417900620408,0.2944114769112013,
 0.32093793032085116,0.12415618088002978,0.14158406850201144,
 0.1682153356141779,0.4227197672345684,0.08507558718037178,0.4321299980870401,
 0.5325814955162177,0.20392324711100174,0.4361549597735196,0.7843579490126174,
 0.5859780114700913,0.7464424451835721,0.8540206676153886,0.3924525901337095,
 0.42136072692611315,0.4497052097274342,0.6517257650496044,
 4.101292950815516e-4,0.954035151494411,0.4674427651853108,0.3649101311813756,
 0.5056089640236026,0.2094040049979221,0.581908485174669,0.03449689171149606,
 0.932733255154641,0.5026979006160514,0.3277503950270956,0.3771788990356707,
 0.25341684187602265,0.2247466671367393,0.04839406322276585,
 0.6656234260414021,0.5886711629275244,0.9717562999738303,0.9539932277895394,
 0.8310157659495722,0.16178835257885604,0.4329313136705021,0.681159472021778,
 0.24060001378051588,0.8918774063691968,0.715602909021626,0.377111993885917,
 0.04775192248254645,0.9021155646573711,0.8618299666734244,0.6348833969275349,
 0.26804931733481363,0.02178552276212964,0.8314511542158772,
 0.3420055764057386,0.7792113886409862,0.47779548818548234,0.9627552097153671,
 0.8899298439871913,0.7227444513488919,0.5830172805850018,0.09320171926565024,
 0.9287685561509739,0.7176345205675745,0.19679839538493216,0.7418405814539255,
 0.8687580792858421,0.11902712043584807,0.059833169101976114,
 0.46239937637934525,0.6058849049136867,0.7754493124382479,0.6051352317346795,
 0.09128931100530935,0.6507950583630034,0.6002352819691131,
 0.18153836228169962,0.21406139566188698,0.8133902635921203,
 0.07304298885405813,0.6300629674887205,0.29424408588160544,
 0.7775318778127489,0.013892630959067453,0.8478339701626193,
 0.3165907310127275,0.4577531572828699,0.7704496230056275,0.9403050665659076,
 0.7210589325821495,0.07331527633564427,0.6128727307841086,
 0.023579579124411554,0.5637606189377626,0.4473702825832814,
 0.17363759451015148,0.5913964564130402,0.1566529651487123,0.2804689694299407,
 0.06989334999798291,0.26455311552308436,0.7973652449223749,0.69492155627108,
 0.07100091083117976,0.5150077360728949,0.12240261363755467,
 0.14229677069016466,0.7087431946389626,0.14202163314054927,
 0.36514041396537433,0.19147012931103546,0.4144142094871377,
 0.2833415491517055,0.21937169825070846,0.35178284869934384,
 0.9934191026766144,0.0996685958418051,0.21722236819736895,0.5690549386272945,
 0.3454295721066255,0.5817115820014712,0.08059281175440536,
 0.30827403025265465,0.018822416336916836,0.4322591530334938,
 0.4738097535227265,0.6637779716055014,0.1559380917550054,0.21960444675829816,
 0.3260570375869526,0.7806169106049108,0.6024026679201714,0.5368509075945485,
 0.5903898736461201,0.9699405896853133,0.09865056746830236,
 0.14202974221823206,0.3837782092038756,0.30627370218213223,
 0.18759204699400944,0.8081419134190674,0.14360069805872988,
 0.7686908873369456,0.9732311462163586,0.1856085534290145,0.8846400656370882,
 0.9793163226372156,0.5146830913731377,0.894199167204561,0.5473704791723492,
 0.7061628165477309,0.8440158212751586,0.6216326764606301,0.8546187482511679,
 0.4556867602351554,0.4198973865643899,0.6563284928818641,0.2734654388175959,
 0.7703573943930305,0.6000777903186919,0.5762008739151676,0.5467260140773432,
 0.7527875659589814,0.2171188575619779,0.5386268671630936,0.3996120119296682,
 0.7054323121984476,0.11094523135572543,0.05265017477675593,
 0.9995646281551632,0.002562911750594865,0.39826211769812336,
 0.2751865096360703,0.07920087249258712,0.23126781467960433,0.444968765849852,
 0.6389020275779006,0.3764189890001085,0.2212660791843103,0.16564235713996733,
 0.46327853755526505,0.35437692627445405,0.6046332504016646,
 0.6767959470279123,0.5615895668705416,0.5224439334284185,0.7110255674407333,
 0.8330785587125256,0.06355878076718946,0.532411212147274,0.2458655090743146,
 0.32485898151085035,0.5651812595091026,0.4579782007915111,0.6445271670015502,
 0.014872112080379418,0.6429534671773237,0.4775766500006906,
 0.04764158692313514,0.45049225224669887,0.12793588711346182,
 0.5226586137381422,0.739130348826873,0.23092949042689281,0.4173311158148335,
 0.32384900304369635,0.9881619873361648,0.7058854701836007,
 0.07832834676479772,0.21911128855970896,0.21022940605054796,
 0.2654734331753985,0.4020727937871752,0.915090221143714,0.8020154255436776,
 0.38259895789199405,0.7692142400008639,0.7630701483758678,
 0.11683974153220222,0.12939963005287036,0.9199096516253848,0.338890081693326,
 0.4154376927370673,0.020890758385472608,0.6321668242199017,
 0.8171706475088667,0.8016108507554455,0.8283887533307056,0.8403308240587287,
 0.9880378206811331,0.05917248906051986,0.05221191494870481,
 0.5950405627356675,0.37240881118325764,0.4620039124139319,0.3725014822373064,
 0.3388828270941636,0.6715259751493453,0.6169502219489176,0.7456906925114132,
 0.5672026264049351,0.4073418789425902,0.7816328237277219,0.5667723386717872,
 0.6371645691695736,0.6061056852920035,0.23415662303422435,0.8833317235281727,
 0.45526709240709173,0.2112113103851201,0.9504225547315661,0.6269098417892585,
 0.26309285884437017,0.38145848269201366,0.4706105397770226,
 0.5743817470915685,0.276047852512335,0.6879289844946676,0.8127408970974679,
 0.16292529241836318,0.4651765633724304,0.246965746200553,0.26864429826505587,
 0.6748569414882557,0.06931875214704952,0.4027473510782599,
 0.030767237497326683,0.20438957301093996,0.8143631992778325,
 0.4105343335516294,0.8014139692476394,0.03050882861544768,
 4.0354700462874593e-4,0.6956751157072095,0.0596858929493187,
 0.3402783159204045,0.007467898314634791,0.9547937303630458,
 0.19000008934227985,0.19202585011614848,0.048778473845823456,
 0.005116712651855204,0.36471699873922114,0.0682923309684258,
 0.9844898622720775,0.20968385141179025,0.6730784523965092,0.2646480608533287,
 0.0986702419220219,0.9354211874253475,0.1970864720071568,0.45192594318953616,
 0.9102662912345233,0.8867560591292598,0.7437982977436723,0.8006181557120957,
 0.5916541055305617,0.977680647367718,0.9990113124110942,0.3035001755465234,
 0.6556277921992287,0.9369024438087877,0.6490535124344942,0.22871690958048585,
 0.8245595332304901,0.7925714718190795,0.8074427824883497,0.9292350620498573,
 0.8125673033050085,0.8476226999097188,0.4236188043715091,0.4681452506120809,
 0.292630108472256,0.44496925752711114,0.7504238221642338,0.38659114838200015,
 0.9034192645299863,0.046975876980242526,0.505566049654272,
 0.20832550597075683,0.2368242606534452,0.672972986510201,0.48816891226191184,
 0.2688415576995238,0.112423777924189,0.9565398863170909,0.4776681304027566,
 0.37210060336231365,0.20578109710103143,0.7165319265406302,0.920221991194732,
 0.5340906297285896,0.2559316872494437,0.5771189686941793,0.11179864339098522,
 0.5557095870746553,0.15840961115719143,0.3742463262567388,0.400426248813021,
 0.440908203229347,0.7058855396423283,0.9419436940052466,0.5096244269791319,
 0.7380517172142099,0.5542144976093788,0.6037638920735106,0.07704556871949564,
 0.23095688817159266,0.6638969296470365,0.9874382777030964,0.9086878320954741,
 0.8120746568124335,0.0115515398525331,0.547401580252184,0.04973948469788947,
 0.4184030441318396,0.7122365795471333,0.05358275984517369,0.7270871451474573,
 0.8053190021243871,0.04681502556172901,0.6997436099766272,0.7901928147533002,
 0.44188436689968846,0.6281063042651615,0.5846015302735255,0.6341326441368499,
 0.7988607275032455,0.4647023980296454,0.6774366208336451,0.4359271627233833,
 0.9097877588606543,0.7408456411667537,0.1763697120476171,0.12492879551765412,
 0.411372074186129,0.6319953397668124,0.30609219940168586,0.6330917499482239,
 0.32518977788014936,0.7891185588698089,0.6610537484077538,
 0.19442938605616733,0.5064207639291353,0.29564760775455345,
 0.3088685426302765,0.7463635575533998,0.24714893944024774,0.6199815517116923,
 0.3698246730491762,0.4295505175560643,0.8429053930403275,0.37371507870273946,
 0.5098942241336353,0.02738141409834083,0.07149514892015296,
 0.8969494272387739,0.576036357002911,0.6017255945875959,0.48573674605962935,
 0.44937603270161,0.1605337044141254,0.14045026028177787,0.6476812279672599,
 0.8441167942598209,0.3146253078685375,0.37331533172263054,0.2869032877197457,
 0.5061836050543815,0.17732581157945204,0.030415126669410264,
 0.631310514149809,0.8452110206925796,0.5038934519408561,0.06619909496128518,
 0.6510052765172862,0.7717852764517659,0.06560873559008151,0.4958395334215179,
 0.1593619204114881,0.869824973595254,0.3292118507537408,0.9064672868582834,
 0.4322179468675671,0.5751228143458992,0.9049654902017389,0.9673540425681317,
 0.6893136548833411,0.2717325973523401,0.25338808643706534,0.5168942691693763,
 0.7678357965327328,0.8030927380091712,0.5926379298583135,0.11104699895823389,
 0.141120988419831,0.8252056199999699,0.693480334383055,0.15650252776724005,
 0.9428989528519486,0.4451334802277679,0.935901934894575,0.8163454601127021,
 0.2230593041812463,0.6173063015680075,0.05957583313437287,0.0667079036159558,
 0.6209752071123376,2.0359782724233e-4,0.04954207130191446,
 0.08204154492100235,0.47139028310210696,0.04137937565822847,
 0.7236177640110917,0.9585492489965385,0.8998702565297463,0.14223856616297892,
 0.01765023616752004,0.7746995912920491,0.16631677952045443,0.645191077641696,
 0.26108133509862674,0.8095006840532014,0.47463542299241057,
 0.6022217729273127,0.5821845944437717,0.9848877832081171,0.7921605702764338,
 0.8191922239378094,0.5559697737783973,0.22877418930982907,0.8223095972599513,
 0.15082046988260334,0.7363349342378748,0.989274111028379,0.766559052641307,
 0.46453278209013105,0.1837469415711085,0.9093803109091064,0.963589521964832,
 0.08200952639637482,0.3210232810314484,0.9697264437338642,0.9950227314358976,
 0.09571217557157796,0.25645714367198347,0.3906373546814398,
 0.004650376944904666,0.7769925257231197,0.6971260210396125,
 0.3844850487457746,0.41352994219414985,0.6675201254142739,
 0.11455904600587075,0.866260926805416,0.7930652840738875,0.31650749592783833,
 0.8777700758288262,0.5141235522486688,0.9008550200802385,0.8903005026475783,
 0.8741266394069381,0.9409838353637332,0.39879166732536697,0.8362442172317238,
 0.42891475997859363,0.12430994056934297,0.41469837073677507,
 0.9041509384062145,0.5654341124150586,0.35426730441102816,0.5693123579213714,
 0.5434027792067114,0.4312783133642043,0.3628607986178045,0.22269342712743956,
 0.3503083977194148,0.02958811274773221,0.9869073641042553,
 0.07156306766732679,0.8186051640865819,0.8740836709002064,
 0.13685208298408902,0.40351460996933664,0.17014089654323783,
 0.27066690962116746,0.7696920184464486,0.5499732034642296,0.6742446566836472,
 0.8460977590899728,0.48066905951867867,0.14514988945673712,
 0.44331270709342785,0.08747773647872503,0.22213635763615813,
 0.7501500465658905,0.8287573613360518,0.8952940440418341,0.3618057898222071,
 0.7164448127817821,0.5801185972650523,0.0727895063309687,0.6428831240659282,
 0.5329210529059468,0.1645082240157243,0.6384006291661822,0.9774269479368413,
 0.4577558353969131,0.1782878274378672,0.4854252926530277,0.7587424888106126,
 0.9142265757392926,0.6830894610464567,0.9642905028343681,0.13947279206797958,
 0.023404194299537884,0.9123800654533186,0.9733658624059787,
 0.11427682037575226,0.7723169051968107,0.6647312496644168,0.4856249699775076,
 0.09780469213287335,0.2691909139101504,0.6094991553019573,0.2877983654334314,
 0.33306354780233494,0.00558608053137033,0.27963191486804273,
 0.3987274183136562,0.9741870147969932,0.9313773424594944,
 0.026792489439738665,0.7591202123490226,0.15274880999781748,0.63831318950644,
 0.13148385442705168,0.026406742337566236,0.8038462997735597,
 0.160918233240949,0.531591302779151,0.4818117975474764,0.4589784255538599,
 0.36361237071784425,0.16983973490097504,0.7151771010607609,
 0.6065397658434662,0.8397622921054709,0.77416477408689,0.11343412507751405,
 0.27210846199582306,0.6965947029845561,0.5570614595275118,0.4603856520457906,
 0.9265884133446884,0.13771119835579848,0.898980927761133,0.3999947867899688,
 0.12706890415581407,0.37833680423311267,0.9413525997864166,
 0.5792906955250523,0.053476214053765014,0.18307132350928657,
 0.07465813879308314,0.5067757401230373,0.4434682142144627,0.1958690243114748,
 0.7600507126983628,0.843402423499588,0.4234652702620424,0.7692053297845832,
 0.8658510008962604,0.760886698543881,0.04017145811642331,0.756443344305307,
 0.22287567282530052,0.7820953880761249,0.5412002390930701,
 0.16207681237836735,0.6856238255811359,0.7165590259618024,0.5768014095271602,
 0.09314092135130503,0.02304490001639392,0.7522059401054986,
 0.8161701815111239,0.3920489603823907,0.6114984980770691,0.27763007274924956,
 0.7463626919901952,0.20211905773604188,0.9743614946104033,
 0.017406229151759156,0.6484575077273831,0.6072648071364263,
 0.6679518809817881,0.4431052130058818,0.702661186235095,0.8491046082262704,
 0.26702665987843666,0.1690934704401701,0.0065666015304673975,
 0.40455111381571884,0.6132200216571089,0.8158379801213396,0.6982179765435806,
 0.029096343639246625,0.2923488607217333,0.6886663303546783,
 0.5193699096239691,0.08145575250375547,0.30355661880233265,
 0.4974304558827859,0.34527282263279235,0.7537523716914952,0.3529519877644238,
 0.9057944279816945,0.7643592454162184,0.36658989425011923,0.7506744368398535,
 0.5888638411169719,0.7335073892613773,0.5801506227205739,0.17311642295092733,
 0.7935528364872986,0.3869066585071871,0.1774187319623357,0.42080341764129336,
 0.2637080914023421,0.48131198313512236,0.3298785131529943,
 0.37696300764233936,0.26447417556537156,0.009782006140574118,
 0.30268973461341186,0.983966821609402,0.5877970921930894,0.747525220144985,
 0.6844856859214044,0.5974667745430634,0.6698544072001045,0.08027154849807339,
 0.3289384041336576,0.59883408704737,0.9148140618158649,0.6410516523570694,
 0.06697408822095718,0.7297869594376056,0.9647672598624443,
 0.44768325573388923,0.06861477205476108,0.674006080284974,0.1389276488802872,
 0.683360896695956,0.14584608193392212,0.7221141969068132,0.9767211666261513,
 0.43255077507558615,0.8855583714585118,0.7381007655012041,0.4661977623546252,
 0.27632779688670117,0.870476581477564,0.7054405354580998,0.3069260325602774,
 0.3395841769705912,0.7814504761613035,0.8608502656632311,0.46419986460091667,
 0.6075506503134761,0.5539492147231463,0.3260964421328445,0.6437556958720322,
 0.5086587439182935,0.8579790245230252,0.1838585797465293,0.5374422175854842,
 0.923016205442087,0.8771049471535964,0.012764871925615595,0.8901458274534434,
 0.0916659508533918,0.8082476280784585,0.9037788123510678,0.9471371365641262,
 0.5458606373087608,0.9766366840046381,0.9836889409299215,0.3339399403330574,
 0.010412678545319576,0.20709401163527252,0.07018260909140872,
 0.03819969459714745,0.5174930558424147,0.699175334907205,0.2389988632923613,
 0.06061653601777073,0.20427870098733203,0.717646678871255,
 0.10062433888258548,0.7524514914478875,0.8693112713533383,0.4485066442013097,
 0.8478262720902602,0.9095916462039239,0.6466881761731033,0.8472615524448065,
 0.5748865183479053,0.09110464207966762,0.72153995677334,0.15085373108951994,
 0.7833405135593811,0.47786768338083585,0.9156284486261845,0.5993013424572063,
 0.7679593651544487,0.9162391906844269,0.4338115074123913,0.1444969400907481,
 0.9578583640370375,0.7038455834185293,0.5846716649447086,0.22848747213630238,
 0.9208809756732262,0.38168381751646974,0.4961668160564995,
 0.25292721528916917,0.5258135472763263,0.8131645809084334,
 0.20605132423184536,0.20684154765439944,0.7039317779304209,
 0.3911272253989194,0.5992296176602945,0.6393574752671289,0.16697622909361898,
 0.19723000881641672,0.2295661940512621,0.41430600746700086,
 0.5294727386974816,0.37743437849905037,0.46616898990122924,
 0.2104644220873253,0.07105363611171611,0.685793588134467,0.8635598195934915,
 0.9305189206568383,0.7341141190714047,0.03799155710367841,
 0.14147767925184107,0.9883366701425149,0.6331873704570328,
 0.23107507212411138,0.7458468243403683,0.6398892798570123,0.2638222711031508,
 0.5403078771444045,0.09772337366022643,0.04588642895396866,
 0.8606903519347378,0.9365139801934514,0.9665462686109529,0.19304984734862096,
 0.16564590257686151,0.3788716040709289,0.2551879371907658,0.778422139358455,
 0.10143231900117622,0.9734572952347886,0.628623012743923,0.6631569293034001,
 0.4805982780902518,0.25895470745708626,0.7703635774095696,0.7076765140965956,
 0.08053527892178713,0.9656740210747241,0.2729847636991003,0.572204925093994,
 0.17612138790491438,0.985968884696893,0.12399242159433588,0.7462363874837081,
 0.7770842407570842,0.9121760582719441,0.9483318712304069,0.9071422279163444,
 0.7381784771621593,0.4689718117494077,0.44913278037205695,0.947253928523696,
 0.588857803396725,0.9827844023520675,0.43558788556328765,0.10875608588344188,
 0.9288292039098089,0.9243548976787469,0.2847771568938974,0.48726074129451613,
 0.1369780561040892,0.8610415868831491,0.6874514953270638,0.47166590391936736,
 0.16136056595654447,0.7668172015394616,0.08058486040112867,
 0.028209745538455344,0.4661630544057894,0.43301459618880966,
 0.9902114347641904,0.6980561018611332,0.8566332794441311,0.4724799790651768,
 0.9395454993387842,0.943412503996989,0.607802314968793,0.0915971848672873,
 0.5961270237962786,0.5247394387026675,0.9128809717208495,0.3692260756721959,
 0.7931286984692045,0.29509990004494313,0.4853177419774686,
 0.46835756763045144,0.6300587653409198,0.30590049752728365,
 0.061542446505248205,0.9609920103746641,0.48576928156434906,
 0.6450974067464805,0.8155473479178335,0.28141846576878105,
 0.17304042736805458,0.18887585367160054,0.8743513505417824,
 0.4958771315327698,0.025374534940468374,0.9431961802728203,
 0.3371733289077483,0.332288985534711,0.14141529602987257,0.4336868814281861,
 0.9131419786148126,0.687145901420085,0.04039824415655313,0.9102682855606394,
 0.9248124015086839,0.6415862964827749,0.8870710080127628,0.4889138512368364,
 0.36686021491077847,0.8712858875065631,0.07451309930074346,
 0.6738935918157072,0.45378661195263237,0.1711113122424479,0.6159773812068293,
 0.703981652041901,0.6463737774694379,0.6248351444982816,0.3888862690178595,
 0.4577691325022386,0.353799008765326,0.3867862824367787,0.19337700310506112,
 0.887260329847352,0.49330090563947726,0.3928193862500762,0.4215424515561428,
 0.9983622536653556,0.05286034803636808,0.34819195826760285,
 0.10458207134296638,0.875940125781334,0.07264370407493714,0.3981146283622451,
 0.5363293242630498,0.0412617000497123,0.12286880167802128,
 0.40172762464850365,0.2648743350939031,0.3872890097729673,0.233634027790231,
 0.793586836908544,0.8018557145610995,0.6283576327133051,0.7443819049607847,
 0.5162581560022557,0.4334601419065134,0.45787999966312365,0.4684459052944414,
 0.8637299994849414,0.9873908170458123,0.5462399982316828,0.2883462564774367,
 0.3985774468087707,0.8413621200818839,0.5093434251383364,0.48142422698188847,
 0.47033930550171854,0.38468596327063675,0.9340661306053754,
 0.057485461010339245,0.8315617493052088,0.5208645786918883,
 0.2258945322082495,0.650267441414292,0.31825866421731874,
 0.007709684981355602,0.559035963702297,0.29684518743294974,
 0.8482416166200589,0.6410616499437501,0.7960207933716372,0.318444513036538,
 0.7251686004102708,0.7323206195236075,0.1547467213790925,0.9912847879401402,
 0.6648765776556912,0.3777376747411374,0.05184505199018363,0.7815772313127679,
 0.3476034651779787,0.20161052307158278,0.14623144835958424,
 0.32486966851263466,0.2483537759084562,0.046683067484959095,
 0.35602353079292937,0.9097922122699847,0.30695512371945666,
 0.5098910094741604,0.1852300457550924,0.9318548907305138,0.5727267070933729,
 0.8739326505810951,0.6025650561843884,0.5361255490420591,0.9993700242950571,
 0.1323039689800578,0.4412225554560585,0.3624247795294381,0.7123424514170382,
 0.9826015233311703,0.48909190789879053,0.537875382825587,0.6994100712176535,
 0.02210978075225667,0.49205753871054925,0.32396650358437473,
 0.3564680652813704,0.6256766618244652,0.7306157958021545,0.95837854637591,
 0.41733419113291415,0.2571244770701121,0.043810149786226105,
 0.20592892878006452,0.14010932741109916,0.08734801580501117,
 0.9164692280784082,0.5031757575005146,0.08583707329461876,0.6123559150247572,
 0.03272222115256673,0.9564083673169375,0.8100357550019239,0.5428629543598879,
 0.47081183118257,0.3079197248108282,0.2667000798477899,0.15190938897777562,
 0.3261796493360659,0.6699862049054044,0.9573133423455534,
 0.042183673184944204,0.8997475962506007,0.9159059373146473,
 0.4289945162993474,0.14893639360485322,0.09199676145417301,
 0.1172375882072243,0.9819168345006043,0.6477160748805794,0.6624131236401217,
 0.4095185989864085,0.04267396979277427,0.7821153354243402,0.5613441529237655,
 0.34925077323055365,0.8631131221613351,0.35292622546094576,
 0.3563326085244366,0.6720597881603839,0.4102962898948096,0.11483901642412009,
 0.9807217236356361,0.6845645808615048,0.6044032178988713,0.7129348086353818,
 0.8497914911898894,0.36551656236589825,0.4342749472368401,0.6764022058221224,
 0.90357401625482,0.36973647009412913,0.4190847707795414,0.893847460401492,
 0.1957425602273256,0.3071737659456557,0.7104917070350402,0.27598061860199263,
 0.8302678160366123,0.3733543831001037,0.6236253754294467,0.9344983775981149,
 0.21089597322402265,0.7616863195551836,0.4146686649582312,0.4183753210136848,
 0.2810623595692745,0.5607088144660342,0.32147627331666184,0.4427801198670367,
 0.027398074986410337,0.5774872131955935,0.3322122687744531,
 0.3182676474227729,0.7234531473523708,0.20129923745096812,0.9455188409667551,
 0.9775704285364959,0.11178149787486202,0.7484719792664225,0.2709606942724201,
 0.5709002627443797,0.7641846867288781,0.7543456513988394,0.5556179385605122,
 0.5778230672892681,0.3752089569351488,0.4186148713963772,0.47043066177382464,
 0.9813447614197304,0.321646140213355,0.20171818082931092,0.7989029494592542,
 0.3019921752934911,0.1820121717919524,0.8663638498896999,0.04401610662006805,
 0.9427936865893192,0.5821027612095735,0.7691769055922908,0.490246307731796,
 0.0331869772962049,0.25090685938797075,0.3373177581290894,0.900881529041385,
 0.20461617611572536,0.8239582102594959,0.8404110217899838,0.8096420961106964,
 0.052698449384519064,0.1621078361852949,0.2097223426412702,
 0.14348868502102274,0.6203773207533204,0.14812291328447214,
 0.26955165607303155,0.7152154941016453,0.17333333062060818,
 0.17986394646090342,0.5084008919416408,0.9355956647655566,0.4783760892583594,
 0.7809884637114664,0.35224607492458193,0.3366504798776444,0.9155597109534468,
 0.7765451605775694,0.40878743607803625,0.744777540959128,0.8946914610171377,
 0.329882889309014,0.08771280578941076,0.7185673457253544,0.14442779160507913,
 0.06659108182154316,0.5925210194376791,0.17838758355782114,
 0.9690123445417392,0.654158124169278,0.1362253451774167,0.6060472134803109,
 0.5304492211182917,0.35441242617662294,0.7889344168281567,0.5031102856515732,
 0.7896971848621346,0.11464596300271379,0.9804542469238667,0.1780803919739259,
 0.9929173755136369,0.017124943552954686,0.20533518454338684,
 0.7208064469273072,0.8708920452994686,0.4610754996043599,0.46572454715540945,
 0.3371685860975,0.4519841092690411,0.8408104110928141,0.980609802755168,
 0.9886065321006616,0.717684276058816,0.7164854056903553,
 0.0019357465941003404,0.45859003774989493,0.21868243724785386,
 0.19353950837751477,0.9802993019690819,0.740227672223635,0.28934308600483094,
 0.9670527974405562,0.44752753201159134,0.28904067936059585,
 0.5282820009531336,0.02602645444725049,0.018788222788092446,
 0.10857917690330565,0.636206430433993,0.45683898268990075,0.5127923801197236,
 0.5079584123068226,0.005236111649740027,0.21756309059206635,
 0.17837600078811877,0.04084082370332953,0.44327009031060394,
 0.44986719365889294,0.2642676221396737,0.3764239399470334,0.1783961864147967,
 0.8599657432382607,0.23074809061995438,0.536185462568957,0.9698321854370262,
 0.4392052089552897,0.901911979142214,0.22360566230037393,0.7849422790994676,
 0.35515656557860265,0.4479146058290814,0.619571526610903,0.480667945711438,
 0.012568548970756654,0.6082874746421393,0.6483373940896708,
 0.9868221929128599,0.49143410797864373,0.6645752246311128,0.9793372942585722,
 0.31101431629322707,0.6189222208716831,0.9764996744808683,
 0.44308231795994213,0.511467417081871,0.42606388163512854,
 0.08121740964129753,0.2634985461619196,0.6831732172773172,0.7533846780647052,
 0.18700087814049993,0.5190623805934117,0.5197856058999366,0.5314999173706824,
 0.4528067638700213,0.5981827748278271,0.5825700804797362,0.3598778812952932,
 0.4724667307462863,0.24536584300186381,0.47050614927130896,
 0.3813266051560269,0.25898017276381324,0.3350051430241141,0.8712532171790178,
 0.2992725162158595,0.24570210753373045,0.7913705003582638,0.2760889187380363,
 0.0604902654108731,0.6967007198848902,0.9866270819873229,0.3275831952070156,
 0.7757098310521134,0.3286682751584167,0.17680317365813658,0.6274012779636325,
 0.2408200542366009,0.6363539276696903,0.6251851211224618,0.3129185064367844,
 0.5104608325615108,0.6134685649032436,0.04339956417210333,0.9274432698285047,
 0.001145314163454203,0.13017727584497918,0.786272615412915,
 0.023766635401358482,0.30542473354298016,0.2345298561550564,
 0.3754675275587761,0.3545105075033582,0.6797837062638225,0.5220829810521467,
 0.5206687907103813,0.49712511599759,0.009289940743520075,0.24554297979283723,
 0.7116042182479829,0.3859074096516102,0.8384024758619522,0.934041689138412,
 0.473814052301896,0.9676134550838029,0.49210484962989953,0.1426259756520033,
 0.9132322634322089,0.5771481834998147,0.3125399816643841,0.8115454175139929,
 0.14029660675339262,0.20420927219046225,0.48497596348290073,
 0.5151457969003042,0.5120306554612053,0.7727544683697893,0.8768717921927073,
 0.8518583593077437,0.18506773583881575,0.6996499841146792,0.6199547156399519,
 0.2995518686359768,0.23900190878975036,0.2210180733272472,
 0.060805517342715665,0.5777103599766061,0.7029476790297466,
 0.4633002914620671,0.1037839864915342,0.5953194961962034,0.829528972289431,
 0.39618534333959965,0.11956062940560275,0.1385749970647535,
 0.8925453571846561,0.4309227197851949,0.6777412260080595,0.40957726274022566,
 0.6140481684140217,0.6258613167244196,0.7282979079800411,0.06521085503613644,
 0.5271491118535151,0.26889183341742195,0.5196174785929965,0.4375815798135201,
 0.6969393244107537,0.920795368355317,0.3338840222346161,0.6721405516029838,
 0.5411062890924011,0.8269441965577042,0.14347517627283868,
 0.20972790465407798,0.011049546413940292,0.2297560793502723,
 0.7930683227221439,0.7410158599573125,0.21031296814168032,0.3757395037747312,
 0.8121251361015371,0.4779686992125918,0.22428812864502623,0.7512664576358058,
 0.8788266121274478,0.5838108645599351,0.33671714200215597,0.9575106865852081,
 0.3144096345373455,0.33840854745042803,0.6971744790882888,0.8779695294598012,
 0.3076171293629465,0.04538240721330078,0.13582781207626982,
 0.5514656311628154,0.4339581950389104,0.38354431717208415,
 0.29917367838701947,0.336591283076618,0.96785955483114,0.15611330149244762,
 0.9758772642978527,0.6780954936119292,0.603121898866074,0.8281325451684034,
 0.05547348802723995,0.4379891364294548,0.23617125814265771,
 0.7639050965489603,0.5725217001222018,0.23488019726136788,0.7603446434694843,
 0.8652558843363263,0.6254625574128017,0.8187342546630051,0.4490856960977635,
 0.5560667706394441,0.42342638006931044,0.04121039790741321,
 0.7314887519500909,0.14437035365206807,0.9791149891163177,0.9291701001337173,
 0.7799273347951168,0.24985309270178024,0.5934820321464723,0.9245485166213515,
 0.34598210257931017,0.5750002470318157,0.3720033590954087,0.4065876986900069,
 0.9811343238453069,0.7153397919248822,0.9145529242970146,0.6479231349878227,
 0.8329477904018374,0.616571542073996,0.8213334894278816,0.42414231715068307,
 0.26231322502357424,0.9109317146014574,0.4002635292052005,
 0.28220040602630614,0.06073255631075836,0.994759550804577,0.6481874784806428,
 0.4850614908293207,0.29600334462424627,0.8067610508049774,0.1389949162987757,
 0.3129014118256227,0.8852824066553435,0.9049020858644984,0.7861765820688368,
 0.6230386351678094,0.29155917818458443,0.5066986631937426,0.4423732909292758,
 0.6611659407985252,0.07389569825246312,0.12608173608480677,
 0.02070982069566385,0.6023569857588802,0.7079808299968715,0.1675147344629223,
 0.8892029355332235,0.7281352413329103,0.8469984956429488,0.29297759597403406,
 0.06265964827432646,0.9560312810821279,0.6460216081581196,0.6041447035404813,
 0.3692533117519565,0.5719410613454906,0.046009431551231494,0.146812927585676,
 0.7398292223210103,0.9664162383581363,0.7764853692746154,0.5619586299967894,
 0.8558743257926029,0.7596392998675852,0.1715299117755088,
 0.006818765044038333,0.0695246092201921,0.4475693342746605,
 0.5841636647855276,0.07406032295218168,0.7165953080452694,0.6477837267830231,
 0.0943701690583133,0.8401324603013792,0.448408496831495,0.6971019481682188,
 0.11512624365897017,0.9681586833831141,0.7959195244900829,
 0.46226281821294735,0.6180920580611357,0.8506982855102947,0.794314771462287,
 0.9633791520318005,0.029715689888604846,0.06511292807391023,
 0.27206403450652683,0.8520650720220949,0.04874883652055051,
 0.8353783942165106,0.025992330987526402,0.4586062416635944,
 0.7370800282286984,0.6920396606834538,0.5406650220461665,0.2537420741738967,
 0.6259123092038633,0.7941142842236673,0.8295235597695243,0.1983627464043759,
 0.02533347243199613,0.707012413941316,0.12445752883590655,
 0.17057054065364985,0.4661092544125337,0.41154500328845667,
 0.9426184937820208,0.32599334073956343,0.059060359862673883,
 0.6354652827795335,0.6067356579071425,0.9495519491332429,0.11515611861450736,
 0.41156570214599486,0.4008350606793193,0.057055875537810774,
 0.8906130106503067,0.5576213822360787,0.07830905301514579,
 0.19398017867019623,0.07908179794647174,0.2339573231090455,
 0.5146481913751277,0.871189119350018,0.659909558485864,0.650004924741817,
 0.5531866711568311,0.9690707667709534,0.9693386559590547,0.737414893064943,
 0.7849741365032643,0.5621492191707012,0.9391516015025113,0.6600672923671125,
 0.9264420587997982,0.6108251669943427,0.36473664699561736,0.1995709008181732,
 0.9418203830269273,0.8613632194256291,0.435668803863951,0.39052774542563473,
 0.3715680541774149,0.5368542886130557,0.027832880119407766,
 0.9063555081026334,0.9112742490381359,0.9862558293993986,0.5319991500197193,
 0.05025451402454184,0.3462362756334705,0.07191997557293661,
 0.3203319010783695,0.4835082812483391,0.2875891944075114,0.16144425499573056,
 0.8009271890671336,0.2788798817599749,0.29614330436562264,0.8590789253295281,
 0.8070209678827883,0.9859805342277087,0.8177954711163463,0.7524798660913542,
 0.3857737662817914,0.3145547557642556,0.7282702967892696,0.3914144339458808,
 0.41516789646343466,0.9345129584530807,0.04820924616146638,
 0.7234419125895284,0.3506880502649077,0.4071804838343617,0.7698229254316512,
 0.08823578476672189,0.2901928574763133,0.14280125669694732,
 0.5241662978585149,0.8111498750248254,0.342135063783615,0.39941450710901627,
 0.7395146704087173,0.5134014996698918,0.06309740526171836,
 0.33424012456871144,0.7972456893212267,0.7212063564038622,0.8476733293480996,
 0.6140157363762376,0.42409856478952057,0.18621773178674295,
 0.4194184929703072,0.8358685036498165,0.45542303140973506,0.7988727789515252,
 0.23640219033030052,0.8267805711908692,0.19114320031385312,0.55237362120116,
 0.24001652159296527,0.4065843483106878,0.28951749767856494,
 0.6907941785596047,0.3934783137100346,0.8206616691570883,0.9370981973559359,
 0.7465194294243085,0.3182648348334475,0.7978189952452966,0.2734677578353537,
 0.5306424085003972,0.46093443578604987,0.5623399171915762,
 0.19514611177147978,0.37370283136239246,0.6010005660049577,
 0.6385125001031147,0.2245713665707264,0.599694373986795,0.5481409446409385,
 0.07795477994954925,0.48921547906601615,0.5825469370460307,
 0.9174664212868467,0.3084719444467582,0.38592258924137257,0.5067296735331012,
 0.3087654181541193,0.2505984319422013,0.21625075619562695,0.8028578578726252,
 0.711346684717637,0.052850240123794245,0.4360300988891208,0.9103093841048803,
 0.3175093725708359,0.47967409863822574,0.17377410614817135,
 0.08402269268549667,0.6586557157140265,0.2518222325345857,
 0.08010074536830047,0.36016662396748134,0.6378699107349594,
 0.1572662777867625,0.505346765495673,0.36114783802241446,0.7093782216921853,
 0.7318409556252605,0.3749414469223623,0.9384320725207214,0.5229442453913858,
 0.7051507957849928,0.33775981279354517,0.7657836012127449,
 0.31560477965020495,0.22941208320479323,0.4044112693959434,
 0.3711870951453542,0.34416144298772977,0.08715264170125592,
 0.7681371249344269,0.9482591138767,0.5596950763246605,0.8577340676584542,
 0.8925360723430051,0.41656352876531644,0.5123536266916016,0.8298122028830986,
 0.6359269103736581,0.06545050545453934,0.25172590545569284,
 0.9138601217279979,0.14154889085105304,0.4187025958050765,
 0.25119391283732595,0.5647750660497768,0.6217804265210795,0.830677338787391,
 0.7545631607386976,0.01971686547517737,0.48929469193514097,
 0.05879850619925708,0.598919117662818,0.5238227360474406,0.29226576689190076,
 0.3359606080846457,0.3044681688376455,0.8243066303554328,
 0.003806535942782019,0.4033938392322164,0.06612091087127514,
 0.1441782464908833,0.6604037512254795,0.45950413347927865,0.7269599586186457,
 0.6597863529170744,0.8755567667909013,0.5087285581289669,0.3307681804241378,
 0.7760532138612812,0.12957494502783373,0.06115024058571106,
 0.1028966114748231,0.2040579956805345,0.9418631734895602,
 0.003520304655557749,0.6972132563369974,0.6808045148679976,0.661745857606881,
 0.270122348762581,0.5545556963549887,0.33319817076980074,0.02288657125595439,
 0.7582116411050689,0.3931087759182379,0.5093626552517515,0.07809176769628734,
 0.19145238232904438,0.6987674152467671,0.8291076620526894,0.6065038549738067,
 0.14998878798287163,0.3200810247193706,0.02962942595271767,
 0.07359660344543317,0.5103202534003513,0.9206240955124695,0.3598842850366065,
 0.8492228452964842,0.7440477043480938,0.1653211721635094,0.10102480070260755,
 0.6934981123400623,0.6407601448092032,0.37780043792826723,
 0.12554269888552427,0.27422287704196857,0.43347961674662183,
 0.5576922255783302,0.9050790242995402,0.009880825666920723,
 0.12629765941787408,0.9454335682110007,0.8677913367892767,0.788807159929988,
 0.847501342930877,0.5925210652849879,0.18383798895327907,0.6800396668410971,
 0.6919884475398792,0.3431593457804426,0.8589411675920732,0.30745851595804374,
 0.44935644246269846,0.7059926455397963,0.003293199878914188,
 0.1542901658385769,0.08932510713844344,0.2198599333470974,0.4797302112975488,
 0.2516259415590125,0.05306329318109093,0.4148265431468039,0.8572790672190687,
 0.47992938816392017,0.34014542698237804,0.41454351510229637,
 0.6523912706145191,0.925246762668163,0.5471335205571304,0.10196970538559225,
 0.4359831547752908,0.7040354847124719,0.0790050147388267,0.0944697996375633,
 0.12674228671339538,0.48083258238945603,0.6841091245366354,0.624630441896897,
 0.4779856078014324,0.8025157712253383,0.7888975725473723,0.8522538268702664,
 0.30401902872219544,0.02706370710129624,0.7988610137008315,
 0.3421723997230437,0.5319790866403918,0.3165042506085898,0.8035098721359987,
 0.8629107913754576,0.10090844943540966,0.7190281768544464,0.6882984283862754,
 0.17245371196614412,0.37332319602626074,0.5495109292893826,
 0.09625614658400183,0.5069443330290782,0.6235273985113989,0.7646032947289757,
 0.5846010478320565,0.8468455021104638,0.8005860731508228,0.8149194840226275,
 0.38967660157810236,0.3665050277434687,0.8310509736647794,
 0.024596635669951517,0.715140905814577,0.335395457433576,0.6750428180087426,
 0.21997187570934706,0.7350352942442377,0.2168414598148003,0.7858383068534075,
 0.5723906925636053,0.09782764668968658,0.8106278187501181,
 0.20461839970188467,0.3025665538060339,0.6758218810239073,
 0.27708050693911557,0.029969101879032345,0.1453162415174668,
 0.9539009912621496,0.9009924738723377,0.3361160717345505,0.5884137402108816,
 0.8693682465915431,0.8899521850636885,0.1583892813155786,0.29015846304832593,
 0.3775866671706102,0.08248679472977172,0.019244687241340097,
 0.8178809167288392,0.24733396089100224,0.33663983407771736,
 0.4585425394499282,0.9037259340047592,0.8945222792420805,0.4249972648894077,
 0.030246301024776878,0.43585429109309315,0.07811999136820391,
 0.23879566769150617,0.2308650903314089,0.1340522014740616,0.346173403244037,
 0.6008144462440537,0.3208624304174611,0.6780158867042818,0.1609713183367678,
 0.9550125491701394,0.759380409427859,0.6209627643149527,0.2740174967808223,
 0.9301726959966118,0.06240723761358158,0.9196339128306484,0.9788918120072643,
 0.7010450420477498,0.36585042957614355,0.49292785389892935,
 0.9438675012579367,0.7938943346656933,0.311546560156282,0.8566608818961995,
 0.8626776302040188,0.2962995569035387,0.2776986184985859,0.7514061004058656,
 0.1621174865261702,0.8003249348446411,0.1629190846934918,0.7982918938497754,
 0.07086524010822592,0.03453283828761955,0.665818676135832,0.8907872832493278,
 0.24670435235104216,0.17419689785279735,0.7682726693645185,
 0.17583251897928553,0.35632549069909814,0.5231178627365675,
 0.5780325087086338,0.4437528005442781,0.3431100589971695,0.0989017970174807,
 0.5328046766811332,0.7365446660991537,0.537216425520442,0.935315776565315,
 0.8394205200900249,0.8849106255622563,0.31324976215267497,
 0.17146771582393816,0.15521314602825476,0.35574088086988453,
 0.8774423185822134,0.9655069449908172,0.9441730214307038,0.6641675456469214,
 0.20486960359374518,0.5921474728404965,0.5722040903081802,0.9448994423782456,
 0.05374661345188558,0.8521903814817852,0.05230558220716719,
 0.06853587445962472,0.859691806351545,0.3916604732722143,0.39853558604496264,
 0.33084384799324307,0.6328836058226774,0.020920950356293933,
 0.917414847956309,0.8346126765617718,0.4383313953449651,0.5467023340993722,
 0.6315581321531933,0.5844681875658433,0.9435589545421823,0.8794870949558722,
 0.929328818944363,0.8705115886406526,0.2291192650101328,0.040967538670726555,
 0.4159490862174563,0.048226874213162074,0.6171603782558186,
 0.4316627876276041,0.4963942457415105,0.4576899468693083,0.4022510467295286,
 0.5330979979988484,0.06994311190451541,0.11901868732517018,
 0.2075871979946925,0.7715848689196914,0.498258398043829,0.7728006318113068,
 0.8729931024996841,0.5491658719543402,0.7270518845549737,0.13056616798169163,
 0.11340879565545081,0.1859049615499645,0.19282140717909912,0.576854111176703,
 0.5054726767072384,0.45920629002191493,0.25433200804235934,
 0.16962348218992862,0.93867879917689,0.27384955221492113,0.1287051173004643,
 0.3181415296888144,0.6064586590146279,0.8771634575062611,0.5313795476178953,
 0.5227060388089528,0.7013275170495551,0.7556665382031011,0.10466301621528662,
 0.866556650056564,0.7965140837386823,0.23012980648561965,
 0.036952357986520634,0.577215875682158,0.813718331112069,0.3899618158609537,
 0.6159648434437375,0.5225222828015201,0.7790514145214309,0.4107488132328133,
 0.6010615976667821,0.7597356918584407,0.28490612397426884,0.5505045609839345,
 0.38630602712686013,0.6052150281317406,0.04915024697782089,
 0.5720857148562972,0.1558617021903761,0.4793307575495134,0.02248725715329425,
 0.7898990324677254,0.7067626023237386,0.566384950223855,0.7797466650581543,
 0.34847196400007197,0.4238194521939307,0.9586997225954688,
 0.22555516724995173,0.3215329896753307,0.9992275816838718,0.9914167359055088,
 0.047247533759604,0.5147583158147149,0.1626650422364151,0.2912851048597561,
 0.14838235545018486,0.6246139747335563,0.14128164343574057,
 0.37498712756838937,0.8154914500644125,0.8712190526470012,
 0.09976073915659145,0.044289442445543115,0.35617618553171315,
 0.9027766963506125,0.7805640230096537,0.28911808142163575,0.3340910225199464,
 0.5159644917190647,0.3071958447995736,0.6740264325714453,0.36822595761632493,
 0.7127169090593641,0.3262082596331214,0.6284049512231441,0.8370318198767082,
 0.26464750102686374,0.11871964093045873,0.0678108579276645,0.268321627949225,
 0.49223576519055057,0.8666466318984044,0.37453573039413623,
 0.05325263889283338,0.358162189581371,0.2079801230453986,0.7829444368845002,
 0.05170701078247231,0.9221875159035343,0.2876548002681083,0.5970758959390944,
 0.7042314650908535,0.006505369754649881,0.7529618837833942,
 0.5321658281315561,0.061618216283243266,0.843445000743754,0.8123699351342564,
 0.4113443573892498,0.8785532525007438,0.7015445720018847,0.9043206808097096,
 0.9511720426774706,0.16741617233457207,0.3392040451541467,
 0.22359861412150162,0.8712866557265525,0.4191248718669466,0.5413295656203696,
 0.9305449671110015,0.796161873062495,0.5890485619284868,0.09392798345214814,
 0.48566000283784183,0.5785215212702286,0.2823883671915356,0.7754496787245131,
 0.553858122892826,0.3422848406834571,0.11968373481665762,0.5543018156811259,
 0.49246985282842515,0.147881135399906,0.40678563328174244,0.7940363562973309,
 0.16514332445746682,0.658121469583306,0.45897849443281047,0.610867127745901,
 0.13720093958907753,0.9026256703144686,0.14451884353881905,
 0.6338936175314563,0.29053569025507797,0.7292731112419041,0.5463130542158244,
 0.6074660372786412,0.3014439665437819,0.7750345652645262,0.22715699985158588,
 0.22777846349770658,0.057783411784595895,0.03728286551346105,
 0.06972210442139004,0.9827474676077871,0.4328019409281413,
 0.24815502650221732,0.14752497820147947,0.9676636049477534,
 0.26264712438796056,0.8842114250325477,0.8898522633948506,0.5394251769346989,
 0.8884342429404983,0.45563897914467777,0.4756498668694976,0.3077568459345761,
 0.4251177503988789,0.5723477388377683,0.3478307049185614,0.2632032216749164,
 0.8767787733840975,0.8779069787603535,0.4009255763082429,0.16691729245770393,
 0.5345493990073881,0.9692414360252692,0.590092033868216,0.4362447710283929,
 0.16021113119163222,0.7043081364947961,0.9445478423129371,
 0.18284413415144063,0.8985861952708183,0.5152976827466564,
 0.011151831202556584,0.2796019098369854,0.8798112788303453,
 0.14616827090105733,0.3338976733940565,0.6460495661754039,0.5408640897549035,
 0.7665463445512326,0.8152752683992456,0.23215827491626462,0.7051737570159551,
 0.9134600901267133,0.6308424529461795,0.47705426487776703,
 0.06344583993857233,0.9238222871566634,0.8611214674897034,
 0.03974843224814051,0.7488948119905324,0.09985340300547974,
 0.9002449861365455,0.5596759364369035,0.49790494580586875,0.4235306102635964,
 0.14970100312744739,0.4977685379369735,0.28703739232686654,
 0.44139254262872063,0.3162027854080468,0.7652432427045213,0.6487203898974483,
 0.4517192250253407,0.9547307284071287,0.10765464550393855,
 0.019631348749477584,0.47527770963439187,0.8855311407784306,
 0.9146036629519703,0.10158571808277805,0.637250799338465,0.9980700748924599,
 0.9204201052726112,0.47099590913327005,0.9215583235692489,0.3567339179870177,
 0.7661843796321338,0.3993145397316614,0.11135945088244881,
 0.14960232035120968,0.464981157790873,0.44890976030319774,0.1893826240080806,
 0.015617767238816915,0.5644746830986618,0.9131055129483033,
 0.12384563461690334,0.39475778187153077,0.048818097980855124,
 0.5209593219544764,0.3396849881377271,0.07940658226564468,0.6642213376119828,
 0.7982283349162507,0.701901189514532,0.3121575258154248,0.47873617335584084,
 0.4106560170990986,0.1952445220182052,0.08694578318312651,0.3644042219384097,
 0.07436435731162228,0.7390751689169066,0.2916196312432806,
 0.20789671083418393,0.2922036068298298,0.718074258910776,0.7634415775348066,
 0.9259842401702223,0.39325053127111254,0.2565474683636414,0.9741641747761006,
 0.635867658177474,0.5743160570980524,0.4488136582119161,0.5290327267163505,
 0.23633635885259585,0.6883036559735989,0.7980835688442461,0.8792129599445746,
 0.8069984242714329,0.18776271103456965,0.5131002843380634,0.6615076074963201,
 0.38769200071447996,0.5446586680629982,0.16481235650631887,
 0.4692109995048317,0.5995929410595786,0.26843216305636874,0.4075895910093523,
 0.5019998701268066,0.6935124695348774,0.3756274897641213,0.7364827431708066,
 0.9320320926310395,0.824760192860861,0.5276925269686301,0.7167040177786592,
 0.6779233918278081,0.12827053506523534,0.9047971088102905,0.333097575244201,
 0.1355183772526044,0.40431561420311857,0.7678670395733451,
 0.16821260448301334,0.6154658906376325,0.0474713293924558,0.9150595281615836,
 0.2869042928855876,0.6995638837567624,0.8428156199956237,0.5344106686405168,
 0.6279808260972322,0.4501835047603683,0.5616290907161007,0.49130390528498025,
 0.4266670987031733,0.10595802477909166,0.08553587342429592,
 0.11742631537505388,0.15453422601115852,0.8941397380402973,
 0.38936181883874776,0.7915927995224057,0.24029743755245114,
 0.48320838262160637,0.5799060897033136,0.5183165808928073,
 0.25170824622866284,0.4745508658336951,0.7102416496883301,0.5070693952433556,
 0.05658067388526944,0.43822070880699815,0.2624636287319392,
 0.33881101362903066,0.8278874085115129,0.7733338326340292,0.2919039653517772,
 0.7333966778602,0.5848722075560243,0.41013990432921643,0.14958167947133494,
 0.07229332695382717,0.812686004343172,0.43144084289973617,
 0.08423005594671162,0.378543882517539,0.8772836992934431,0.6191112713800606,
 0.6724096227895385,0.6617056443016496,0.8192881820522145,0.8293697286891438,
 0.09629694520378329,0.9359391652356047,0.26539725863023744,
 0.2695091785592041,0.04757996762630323,0.7691195844752904,0.385503370663284,
 0.5018496414924641,0.35278358441035707,0.39281520969870243,
 0.7909230427805982,0.6690514183651974,0.9341570562440549,0.4162316321904279,
 0.4213666231175798,0.4413944378629603,0.14247654405297883,0.5672159429732021,
 0.13859852140387086,0.3658345112851559,0.1566789166909901,
 0.33156025367908815,0.6267196186563683,0.8496979310996111,0.5690060539367893,
 0.09070981504007714,0.01062940180410954,0.4548810070708458,
 0.8615442833644014,0.6548455675839042,0.06645787629188027,
 0.021901610554205253,0.9475917679189324,0.5620157348370382,
 0.07349437338598197,0.17950688894130074,0.13999912597194386,
 0.7061670931128677,0.5983970874003179,0.5886327745096283,0.1115630498299034,
 0.83621180646875,0.2275794512957272,0.19294550783646114,0.17361134746160012,
 0.5879260118199735,0.5037862587509977,0.1608911588966091,0.41251644937142373,
 0.16718993388661962,0.472105670308805,0.4582855335622872,0.15267067261890388,
 0.01902280844270654,0.03228459606963796,0.6247196014037381,
 0.7516756617644876,0.49270091942284255,0.33262977301510066,
 0.7705614303823236,0.3481451188337743,0.6533540574165333,0.6423523299603432,
 0.37698844022064915,0.7133418505697257,0.9664460757062596,
 0.43158269670176785,0.24019474282617725,0.8525071846090492,
 0.4374872230413128,0.7249015518323288,0.9835326475758231,0.26409213681598787,
 0.531328874124018,0.3621199258331681,0.2859931685839515,0.11510285043532087,
 0.503734758894403,0.5347715934166316,0.49821011259999803,0.9380844243194065,
 0.896498950654299,0.6473222722934798,0.40336605990727314,0.5123903469931941,
 0.6725650232875136,0.3430993548773429,0.09845473929185311,
 0.41502821006972146,0.4710891927272416,0.19268157749094061,
 0.44988666067980154,0.9594749544152421,0.7878330802243438,0.4994176944478931,
 0.6244438825217834,0.04368704239144661,0.030713483869465263,
 0.6686479996259329,0.8511389672033715,0.09766701353281726,0.5065967883975938,
 0.04956343845790945,0.6136199410240828,0.22103698698081609,
 0.9694402738526297,0.1777477603728933,0.5488166822847366,0.402878124078327,
 0.5539745912793578,0.008066470385706404,0.25483407475463926,
 0.5541879759554417,0.7695932339736407,0.6905401105531497,0.41396154913977146,
 0.7587060081651439,0.9771953321740721,0.32696101933594035,0.7571309215318346,
 0.12385936770453576,0.9591174206300724,0.21079676820970616,
 0.6461860325082593,0.5432917645817125,0.17878826024079109,0.9896782378622321,
 0.7442944774463092,0.45171106669526606,0.4293565210894197,0.8942273702865265,
 0.5735833582577367,0.23489541379739287,0.3045619169811389,
 0.052146296107484225,0.6492280709567138,0.6309783149529788,
 0.42715391560829075,0.6933039947701313,0.7030274878737481,0.6669112205673476,
 0.9808987030101453,0.46076220215792385,0.26120220391990334,
 0.25295772669075534,0.2920421547908797,0.8795557223169637,0.8231289872076832,
 0.2751780374457933,0.036150043968184,0.7154571137557417,0.00950664009202018,
 0.6810777838316204,0.9720477117327433,0.92784055661821,0.6319783297539256,
 0.8755821957130042,0.27244041839368927,0.4421540076029443,
 0.12819624670470786,0.2509427436688023,0.7044446824843774,0.8780108802426196,
 0.08372745926312719,0.6741182864831751,0.3885968118508055,
 0.14256247418177825,0.00934734163405826,0.7447473754293795,
 0.9788670514309804,0.9828258038329531,0.6536426929968123,0.4496621675483736,
 0.98802464208537,0.27762421196588727,0.23060930459787876,0.8392829626253444,
 0.0956788130639542,0.6112488975957086,0.10025738175163212,0.6769015228014484,
 0.8570246944529716,0.7940788028717568,0.17256442363642566,
 0.46576584305842694,0.9097096885781315,0.5907813444290402,
 0.15206558087259747,0.2721824914707214,0.024039772228848455,
 0.9385970406612436,0.4211636289244651,0.592450794426527,0.6087911620497506,
 0.3664079429976963,0.8556372939923963,0.4385051577621054,0.7280182033179555,
 0.7117386550217827,0.46807432451496134,0.04242494233617411,
 0.42220691118180775,0.5658072095111403,0.5877014075588052,0.5140847761817341,
 0.32454693931413603,0.2214483987040291,0.5294867380319485,0.2515042793198805,
 0.6255517576621059,0.5520422291436291,0.4816873880373268,0.7797624096998944,
 0.13899995660678754,0.26416893118816887,0.7722694980753392,
 0.6052263500821677,0.5437398461457672,0.7165392708996939,0.7139708733892581,
 0.7721919507763112,0.144468368922124,0.5281090356821819,0.327600942568295,
 0.9244797887053491,0.7977484019851422,0.17728004981587675,0.9266808422140547,
 0.4981114903540076,0.5107981769794956,0.4196792694498419,0.6391316777183886,
 0.8815570522388686,0.5549561756008391,0.14562563934855965,0.1017175603785957,
 0.2860585567872431,0.13256444300563308,0.21864397925601498,
 0.21544350127973233,0.7926575240031798,0.3505294820085102,0.8661352053876474,
 0.49206500641587014,0.9603349827209999,0.3785188469056373,0.978652704960929,
 0.6077866282050145,0.5514759044618479,0.4424522874601988,0.6852112053500543,
 0.5171348212730281,0.3256497599539334,0.3001621235733134,0.19243077909939077,
 0.22243538403847318,0.3848131619817937,0.40317136825608746,0.486028740079813,
 0.06936091081788542,0.015086088109657503,0.9313502172822535,
 0.5222250733406876,0.9113960930762165,0.2828515898558809,0.33597123813829677,
 0.5915989236946382,0.3050033523040714,0.20684679201680378,0.4758822830316676,
 0.23739425066904685,0.7626153987790214,0.7556045858059234,0.6688193264448916,
 0.7967294506474183,0.40020531678500293,0.7681908064249732,0.9451852138450151,
 0.5910942484112311,0.6546375018619416,0.5105607919857553,0.563093542438813,
 0.7546568079756376,0.651420852052795,0.405950581634008,0.4089983182862653,
 0.727085991982339,0.7738045729625671,0.1544649080840248,0.8165176016056688,
 0.46718942604176406,0.8554779041047734,0.6460214240921809,0.2726149718950054,
 0.6872939358704621,0.250805770830422,0.6491069593107223,0.25949824733499094,
 0.7269278581644878,0.8485756078221325,0.9857364173832704,0.11160408271128175,
 0.05696481794519037,0.910694186355241,0.3623821783273262,0.12626196854597205,
 0.6390989616077898,0.22608641144206076,0.08327730334172834,
 0.12309913848400178,0.552087862544813,0.05803847388993766,0.6473329970637913,
 0.00938032736850647,0.24876782433158529,0.464259582191866,0.7213222269284171,
 0.06457751942291956,0.6670145472480715,0.2818225006502709,0.3583254714218498,
 0.25974260994105025,0.36683628052558637,0.5541004056907085,
 0.40052059721183797,0.5102418882579851,0.4856672390013632,
 0.40896750074749244,0.5178881968655975,0.7750468780124212,0.7355208893128539,
 0.9610690505670039,0.5993304322029602,0.4875409997871766,0.6664073372264184,
 0.6361558600461266,0.6334021847564094,0.865037492119205,0.09731542283808325,
 0.5611935438587519,0.1062557954229506,0.2653039854846805,0.42615101008352696,
 0.03730701623106303,0.771869741866806,0.419132929781992,0.6833561616201574,
 0.17774507812535245,0.5204697177431642,0.16012477080684184,
 0.6517965634054491,0.6072267693117188,0.8483293410369752,0.7164279572892751,
 0.8628371964735823,0.9481941568012437,0.6928788402179585,0.2608944853615981,
 0.6817246986854364,0.7566017519229985,0.6713747426034513,0.4741348629057305,
 0.3813750075493413,0.06465276247005125,0.8001892553904095,0.886714102039962,
 0.4869449615762542,0.9025039275508244,0.9067968998995526,0.6038051828036095,
 0.668289805530598,0.556383478510324,0.9105229147047424,0.5600156522287555,
 0.43234586059241953,0.2319248066236539,0.18595001054108928,
 0.8004094294363913,0.32051510925633864,0.09543061954318344,
 0.5098788701160009,0.7770752492878099,0.6355150194835524,0.3436272169631699,
 0.8758651728832847,0.9346163968721589,0.885722256074444,0.34298571567188063,
 0.3695887298249241,0.4173114611029358,0.8854051788890187,0.5902841296245371,
 0.7224111267151981,0.3716885307955946,0.1314283308890326,0.6789143026411987,
 0.8704737686958446,0.027740005111517974,0.0919663275765421,
 0.6799083132234953,0.6710479460529082,0.3447116453828917,0.6057519738771776,
 0.23821884764589984,0.6991914848330929,0.3600324940602013,
 0.29784375584697886,0.815447116084689,0.8033072358901504,0.5974048101796413,
 0.8158997898167435,0.2983982632612948,0.002241414643304296,
 0.7100619542176659,0.5872486332917037,0.43059199780006674,0.8962039619581408,
 0.011039122597149231,0.4088875688262231,0.7957403020710043,
 0.4840584103456762,0.24276631442345442,0.5059658622794803,0.6112161051629684,
 0.18812462764080196,0.4858698571244936,0.21848353491226602,
 0.8779677180214583,0.24676959673646737,0.039154246313178764,
 0.9403766892121717,0.03152352857463303,0.5816645932322759,
 0.17737512117900467,0.001411573808193567,0.8528949593269324,
 0.2963874996182506,0.9054017476565834,0.3844763401918302,0.7687315747139518,
 0.13158897704571082,0.6606460263921805,0.7332049670036264,
 0.11521441486243333,0.9460055297794147,0.011599768082477313,
 0.6586287182190901,0.8961017391584412,0.8432133741574017,0.78279179826886,
 0.8291977816025612,0.09876781208814045,0.6039629268001793,0.4501490971484945,
 0.9697952983708567,0.05588357964438473,0.8088448566435428,0.8233714429036357,
 0.8083538304132936,0.4350577192058346,0.48947572363408554,0.9650884600057155,
 0.8264161608300439,0.5722121321643606,0.5680648530569422,0.4960845224678525,
 0.98996414045896,0.28212937193680665,0.7213215751852468,0.817053990591231,
 0.06453241671164833,0.6296897031720965,0.3678424577309022,
 0.24208671572948803,0.36504111451411847,0.13728667565475616,
 0.5428695714508255,0.3359027100828432,0.4536105266723418,0.45006235236604675,
 0.1512751150968238,0.08099226812034455,0.7532797312957258,0.4080370731887708,
 0.40956594370114874,0.40148893446770717,0.7570468655184873,
 0.7997117886661107,0.2783525733892296,0.19436311133523498,0.355528475185379,
 0.7989397267885846,0.8295941326401015,0.11541423327145861,0.7643144217609649,
 0.7705208695548652,0.47766560155186666,0.6387931210802089,
 0.19843742607978065,0.5085237498436934,0.3254555834721544,
 0.056580932689514474,0.10407158270114891,0.3978893721923281,
 0.45935813293352756,0.009541858962635708,0.17077350278658965,
 0.6622139948794485,0.7627987655781738,0.598090631695975,0.3477405551153675,
 0.1110537837341592,0.033710061293969806,0.2641583555432446,
 0.09245893074985523,0.7572056810385669,0.969453462520191,0.9144689123034602,
 0.7316524521969969,0.7220313664375633,0.42232451652916525,0.2835693962731687,
 0.7043165144689922,0.324336011107892,0.4655903309206809,0.3322219087294124,
 0.45150355989007274,0.8943560296626331,0.09403210208090473,
 0.5507751828661346,0.20202748546225635,0.9190263719318776,
 0.16826504356096061,0.8431880252405332,0.48618012594301996,
 0.5270026165364523,0.15453028352862297,0.5972780919860308,0.6989849264328958,
 0.23652832061503604,0.4126307354870853,0.849374710361782,0.01707896883773774,
 0.3801485987168133,0.6789059565177983,0.25335396958262313,0.9623676162103446,
 0.40264651673900065,0.5369789381866907,0.0482568659667757,0.2825088934781179,
 0.8690967634140091,0.5970846328143316,0.8613486212548128,0.09170766987720325,
 0.890912718858422,0.6606644454390802,0.46498097808366823,0.7966083378556692,
 0.6295291459480558,0.6117453651330341,0.458084084835912,0.5425626249435848,
 0.9864879890656705,0.17833106256656972,0.9403483834596885,0.4656164056531834,
 0.6957768323658684,0.733719304112893,0.0467288335462539,0.8746980446346796,
 0.7806573866379702,0.23011060082271162,0.3641002415357686,0.6537330009049711,
 0.43876486902285683,0.7274711815726087,0.6263251200003802,0.8032435929901627,
 0.7175584575459681,0.21893475700596254,0.8092781394312385,
 0.11656069174471106,0.8047216067904461,0.16230474136216289,0.765795560249784,
 0.06409721572402616,0.09326496741312573,0.8608016338236716,
 0.6567191094965891,0.5666268278324248,0.405374215073405,0.38558983033581584,
 0.799662818393726,0.21920576291988492,0.3592779256603994,0.36923775637117007,
 0.8606202608822907,0.035716507334468295,0.9616018808296442,
 0.3609031502846699,0.13057934332028664,0.03228378636535745,
 0.31071973408478926,0.588687386505108,0.20372120002556704,0.5688864642218123,
 0.7319313440827985,0.2629054218926725,0.37601288803716093,
 0.024717484694639524,0.8262028538758439,0.32439249649723134,
 0.9757819450096152,0.8814021615250023,0.5299538950458867,0.1026187982336354,
 0.5924146799382252,0.07948615602326836,0.17348628403091226,
 0.8976942450382346,0.765432821503145,0.06952140906146931,0.3636365880300515,
 0.4820200191447188,0.1942285475085752,0.7481772630186591,0.5752308699474539,
 0.03843673988721652,0.8327416294226225,0.8537015179563952,0.4954556206419266,
 0.05713001497747516,0.4633434899316594,0.3197281390354153,0.5377644146874752,
 0.13399599324688272,0.5683463249807916,0.5524761276769936,
 0.16714176033658812,0.16163038224013804,0.5447023353918072,
 0.8346616496367177,0.15375600991209315,0.7155360234327239,
 0.45983887551154123,0.014744007690982053,0.4786951517025315,
 0.3030570455305781,0.027368639533597583,0.3606206026312242,
 0.14650925097237177,0.06165569733123699,0.037182372943649455,
 0.6948363663023586,0.5275142561772239,0.7854232401162531,0.5580246554697158,
 0.4957347657537179,0.6868745540200889,0.16379371486440597,
 0.05771414037856337,0.6770880458810765,0.389114561967262,0.454842739580114,
 0.11583339710386298,0.9534261801136242,0.5327565776857153,
 0.16674257245306678,0.4208433422298242,0.5174851520737279,
 0.01429251057289449,0.4162856804531653,0.9019088525050358,0.5910639641975325,
 0.7157479807319176,0.8453887104072566,0.6058426001859564,0.40600274919719714,
 0.777921472360969,0.7060245459231224,0.5540043830005349,0.9596763101463112,
 0.30850926719882,0.29488198126451126,0.41341126360483194,0.8082650876109165,
 0.7576730904221719,0.023707611627637237,0.5591430970266741,
 0.15531029643804828,0.4768041921754843,0.8142542405596274,0.779886473294591,
 0.009135498232670236,0.8432973617240541,0.4492062386825435,
 0.23489418935455253,0.9158482207909896,0.6494390345390435,0.5979764161729326,
 0.048342377595121055,0.049797926021688976,0.5271785405478701,
 0.031048564409925095,0.48871310723993355,0.8210965797824125,
 0.22102685982424197,0.47887357157447585,0.23705924659913769,
 0.6883293595984993,0.07156968328508573,0.34548951266854144,
 0.7826875081054268,0.2541480009390924,0.22738432262437536,
 0.46997609175068555,0.7529323999746512,0.1787039049506456,0.9911246294399557,
 0.9939008411783982,0.09996551764605166,0.16420406741136317,
 0.2284325776635765,0.23192071024618244,0.5168363368549589,
 0.05556266770547014,0.3746381334637783,0.7643240378111871,0.5071610878484858,
 0.33464111161473165,0.2860405029012476,0.4363043991490264,0.3325335751517937,
 0.4511293916550875,0.1802802842752227,0.48795546115120103,0.3142842284770695,
 0.2226207370309663,0.3303624266777889,0.1056752028110417,0.077355796094007,
 0.39950296603987845,0.00831878974902267,0.27583057752005524,
 0.8471455729520198,0.5012246223898916,0.06963473973540968,0.219781068525832,
 0.39215965855898105,0.5379582535657054,0.10867993714642044,
 0.020864344073248775,0.9346837229056875,0.9353903886573527,
 0.8152607323695387,0.06180972261569362,0.577365384344536,0.10739843879892519,
 0.4965244512080407,0.37986806980769494,0.7229865788192494,0.660923158715957,
 0.8519031436961214,0.7617747653574577,0.5090132585896022,0.5493976215748603,
 0.06442136457037151,0.5174284948398662,0.6659974538963236,0.6095380707731228,
 0.9403182543556604,0.9272733362548445,0.632234684106678,0.18228798872096674,
 0.3649102601511629,0.7052290638350258,0.6538398989668646,0.5970529752842485,
 0.35109630402686975,0.19448418660325362,0.4731942815117711,
 0.06461736750948288,0.5771733412283353,0.6451029104443373,0.1226757255644415,
 0.4723832824951604,0.23837419740355437,0.137715930943205,0.7000356768017512,
 0.9875666409478024,0.42280779351533215,0.10445411416143224,0.92368970189252,
 0.5458198953472024,0.5662153617640586,0.7952257303970645,0.6425121641034091,
 0.298238120881831,0.5656073654065407,0.22323296967369188,0.4492622968764687,
 0.8757595928481999,0.17366000012282679,0.8963795911518894,0.5255592798705123,
 0.23934486897731888,0.6552854162450173,0.8872272864541749,
 0.20207415111062943,0.5158739417312315,0.47090199811470923,
 0.9363884185740201,0.19128369929531086,0.6790692009272883,0.3184114026203396,
 0.8718170222372457,0.032748224046940666,0.49799863309341874,
 0.7589998405450117,0.8721972156801997,0.31191698348922814,0.4489332888349731,
 0.39210219453916895,0.7294078948442949,0.8275192862248215,0.7783467962538501,
 0.3354008812051361,0.06057340000556921,0.5915553029500098,
 0.16762367159822789,0.24615452792628445,0.735281534703459,
 0.040417408531448706,0.3854838628667704,0.03470855685601526,
 0.2709693016496949,0.4491886541535789,0.48681247463169197,0.8872600041052591,
 0.18381795823932234,0.4798862498719427,0.5216929596424602,0.7676943007975217,
 0.24156595823964755,0.0678120461030689,0.2704191744319302,0.4804607293212687,
 0.08871523164260187,0.8226252558759709,0.6918884138627341,
 0.30924269265790494,0.41723484276135314,0.6412574715710546,0.85505839719289,
 0.523866531042393,0.13943350557842793,0.6895343629054024,0.9531036525535437,
 0.16383124388750026,0.8259140110108472,0.6930823422612148,
 0.06043222074067178,0.6242198064064373,0.495549256007772,0.6852171242135446,
 0.5550575321344117,0.1279818530226403,0.16170221113883698,
 0.46754196145177396,0.3104224313123194,0.2277712154176469,
 0.09511325015511252,0.9334859509002654,0.9901603911696505,0.7760861576036525,
 0.013792245738583508,0.5815130566455554,0.035970974513488296,
 0.4419903013595281,0.6454851583502406,0.6514263752535236,0.72806021079698,
 0.5723967881169952,0.6133551401477544,0.8493149768865406,0.4998338139670033,
 0.49396274980537713,0.6207429118993284,0.7844075117125264,0.9569167882364819,
 0.9465104190413836,0.510140945549738,0.39286669122911777,
 0.016393615947758344,0.9690830585196425,0.17436316413679498,
 0.931436736115573,0.7335954427591469,0.4460153191686046,0.2570561706650034,
 0.5806055832974516,0.3402424878729662,0.860799140349563,0.30662247153603095,
 0.46105493529015584,0.8542960458476699,0.7679991734201578,0.6511860942986353,
 0.837130467431356,0.5754949778648614,0.9965302731741421,0.3955639006046763,
 0.7957764422226051,0.020126126182407722,0.32050360905466413,
 0.9853741477820807,0.06490494256509327,0.8552000132377093,
 0.30529523384527835,0.82717930591436,0.5786364236736519,0.9664262407018317,
 0.48120285725151857,0.9989934435871735,0.643009569835556,0.8338883484682617,
 0.7103331585616472,0.7025011877680476,0.767966157513137,0.3613969906060329,
 0.06638737160025954,0.04103761094577918,0.34540945485775665,
 0.5628367936638995,0.6041998778514747,0.91359363867079,0.9499506191421528,
 0.40852433024102386,0.2733867405105943,0.11069603652007731,
 0.8955193142421887,0.9959338631659156,0.18467407823586546,
 0.24488895377205044,0.034464193960640155,0.43883134243598154,
 0.2692028798272472,0.6065750215864482,0.42533218210963075,
 0.16534587076965357,0.8875467173102964,0.37823335180692474,
 0.42501044540248356,0.05035399011409181,0.10632854063569275,
 0.7188505207495609,0.2838699356345312,0.5667936939054408,0.2628581539672087,
 0.6595601220110446,0.9886339155500798,0.34233170942332014,0.9625469363070935,
 0.2133734818800408,0.4324379732593632,0.01671728196853639,0.5633999149464968,
 0.6838889605821142,0.6061089746494327,0.3532820631606691,0.7110095726486123,
 0.8076562999967518,0.27171825476354505,0.9795581361824565,0.4712590705927969,
 0.15693889986551135,0.005663816508107655,0.13033190433212416,
 0.8751138785551353,0.5900742904605318,0.6919830224585407,0.24786519150346442,
 0.8968473853751644,0.3037781557422111,0.031753279936404644,
 0.07348251228939129,0.5029951650646723,0.7084533203189188,0.6024984118846008,
 0.5088541947488191,0.7239318309581548,0.11814326888859195,0.7816551210217096,
 0.2820166888677774,0.0652359034304133,0.5830342148258169,0.8586978801816865,
 0.007594484148391811,0.5624699416474941,0.6996680149087108,
 0.24153180772315574,0.957187371823102,0.11501328435928004,
 0.17256221348371215,0.682545013818925,0.3055497430566906,0.06142693018428458,
 0.7835615141153338,0.7584770550442579,0.26855128605534906,0.6768540112359361,
 0.3012415369675415,0.5117026779923447,0.5419683373883943,0.8801706104978682,
 0.5960711494733654,0.7459602990610866,0.6690062256284028,0.24317710166207374,
 0.5226378258807087,0.10453472198772418,0.5945239673803107,0.3537896990819387,
 0.2618118396291226,0.5889554874625986,0.1711231680262878,0.6855662067617623,
 0.7814624741368522,0.6983669130436763,0.8511824498975333,0.6300608408118272,
 0.5781807934817071,0.4221593890698504,0.803423422886096,0.9689584673047966,
 0.2893772679639982,0.2857268167505581,0.9400930272207225,0.65664121112323,
 0.5086935062286982,0.7875564254941323,0.09400170448940948,0.4002076699771885,
 0.8707697693050824,0.6215937389087465,0.8040850008676433,0.668128730508764,
 0.8381788584412951,0.18369891763610013,0.19028262635252235,
 0.00788320790680086,0.8404705498474891,0.3858685582587358,0.7790786315143177,
 0.637301488190109,0.09573747764841434,0.4903959504872849,0.9915802215325101,
 0.07466333707537876,0.6996905856912091,0.9938709874954456,
 0.48611535600299527,0.8751005620710013,0.7785894390699895,
 0.14355249044132745,0.8227154325210881,0.41419163044376284,
 0.9854660875524178,0.06232984116125706,0.172641417531035,0.5395335261440364,
 0.8551118973586479,0.04436567029219707,6.874800642202583e-4,
 0.4656005195183416,0.5361389461023448,0.16032237039059072,0.2115593822496642,
 0.48337819098168455,0.5338636508085206,0.9801199486101657,
 0.32720811119056137,0.9167219508120449,0.9831578063607506,
 0.34576791266840834,0.10424052648224447,0.7877447024931541,
 0.09414454231476144,0.9437922524368881,0.46587765076596943,
 0.05891927765143967,0.5201138820041451,0.5668215027498609,0.4615768753054733,
 0.2092658847181209,0.18420548349196597,0.6071376285533818,0.6720944253201389,
 0.7814835574100655,0.571277225175539,0.47905831084729555,0.03243373103672831,
 0.6128063505208062,0.30369919937198053,0.21765097008271983,
 0.4281401983761214,0.7665222076402507,0.39160891155157684,0.7141957189982593,
 0.08143969286991692,0.4547197728163693,0.43353207965601137,
 0.8555382502339964,0.6019349126806263,0.7571376377586648,0.8000277256016762,
 0.23562051389154426,0.5987695087896548,0.627871702694607,0.8517373941855132,
 0.34120623625365765,0.6891641143175667,0.6561251557291721,0.5287074284184443,
 0.8764978079572197,0.9231778400707858,0.2983014400377393,0.38624645613072195,
 0.9900512607609037,0.7220385396364558,0.9910847305552569,
 0.052982496522439915,0.8743998371340482,0.9868247092388183,
 0.6459683710196554,0.023878777082360636,0.22052580722586412,
 0.1984468112624912,0.6164281224777296,0.9089432570599741,0.14144617176182495,
 0.6989436824503096,0.5815604189119965,0.7360625073741506,0.8563909989452682,
 0.9073943615840091,0.4544339018691067,0.9895448380770178,0.459728836033656,
 0.0668747998884286,0.24922658445313572,0.07797396706450122,
 0.6910764137614196,0.45550455919134414,0.3349027429844488,0.5359249716220368,
 0.6510960805215896,0.003527812887061943,0.9559894958798769,
 0.04746320920765956,0.1849808397045425,0.14537411391121413,
 0.5463031458257072,0.6667575656409858,0.367383714598842,0.2978985916011003,
 0.21663522605549868,0.912918389843631,0.49796483227871646,0.7962301962062371,
 0.2335301290234557,0.7426526653679962,0.6359768090952322,0.9537642001408582,
 0.31354702425214476,0.7634976823590517,0.35903856693165415,
 0.26667674468063995,0.1836900678661234,0.8305007884762716,
 0.20594870450207248,0.6601418964259704,0.6225204351388566,0.4609823299274508,
 0.2208911177930426,0.14539448886500084,0.6969803283692914,
 0.20321697941098493,0.5672264218699494,0.36914183890724006,
 0.31846932027082964,0.4753009893152915,0.39131086418129835,
 0.6036865883416331,0.12187927131550746,0.4572819412834538,0.6289603823832289,
 0.7367560574841077,0.2621562445234754,0.15011807560603385,
 0.012764857833838184,0.7993887028228689,0.19101673127588703,
 0.5312895145936994,0.3872701326242075,0.5516742927787494,0.9331488621542353,
 0.9840385169120868,0.8854453838114205,0.5401190987084913,0.28430525236540927,
 0.7747936071446269,0.7149250481895888,0.595152046803464,0.13087711096381283,
 0.30947469933813887,0.38063152286069946,0.4040331990369799,
 0.8479905168759031,0.048127972374493955,0.04001868728967428,
 0.3174765498394384,0.05960077386170737,0.42102666873848527,
 0.45879847008814056,0.7295364184546073,0.6148051304271619,0.5459118081787928,
 0.13241973193616297,0.024220866027760524,0.08452901650364819,
 0.8288294625266098,0.9902862431341086,0.46514902831673877,
 0.09970343659997816,0.6566527719652551,0.10778616854835943,
 0.21930041409779755,0.37953735494516905,0.9071640353982817,
 0.3372930375291695,0.2591457414464087,0.9689514004029706,
 0.023458984947900108,0.9521887166918843,0.816132650410205,0.8854451270384613,
 0.6747029090820865,0.2639640121768959,0.00675185414352808,
 0.13688859690513577,0.4345955226489432,0.29183815871422913,
 0.9444784643175227,0.37686258519774485,0.8432662800699219,
 0.41812682313501404,0.14026053558029772,0.7621297752375213,
 0.9870377009551226,0.8356410646068456,0.8744516191351898,0.6848891994883898,
 0.2539123112611126,0.5285257832846854,3.2691285550767724e-4,
 0.26571877726244164,0.23440413982872765,0.45106642954862164,
 0.21138884890029164,0.9583416894034245,0.42770026339735134,
 0.9314161382810232,0.5605358517661287,0.9590734492108642,
 0.004950430713996812,0.6907887225882958,0.9125370064914211,
 0.6777605906802744,0.9808434845753125,0.3244942202452019,0.1644596680639967,
 0.28724402760861145,0.6941967249504997,0.8737804376457262,
 0.32420872208462903,0.0442764669064204,0.6051284215462135,
 0.38389476758900576,0.8417282950247659,0.689152175373402,0.2918776336645198,
 0.7966274760132076,0.5257765608445968,0.17192408958859595,0.9321943082784279,
 0.6222625872709677,0.6696580485488368,0.6918423099618999,0.541627235458101,
 0.3473296537888343,0.08624372244239153,0.6389380423441969,0.4090404096707353,
 0.05157724583983925,0.7408145061675901,0.3177070439048152,0.5944586234811451,
 0.6534008597058043,0.7228300976661595,0.14464908889505945,0.9726992541054729,
 0.31569553615707846,0.3485148294830632,0.9689952741063584,
 0.06197957312096325,0.35326081934848497,0.9420840487604275,
 0.8913562936173729,0.22565720423020375,0.9876362358249191,0.9270696868060488,
 0.5677632800085758,0.22144710403805534,0.31728535246587364,
 0.8671347585987322,0.968539191054867,0.634962909608466,0.22262895811118488,
 0.05868954149670513,0.7816072686291166,0.5191364672472228,0.1788274930332644,
 0.48257104396416084,0.45571147402651635,0.4855742775869015,
 0.41537921993162663,0.9479436629871321,0.6225995631217834,
 0.25277771082342504,0.09290254858357716,0.040502054310198154,
 0.02948468483436073,0.7469695991578582,0.5555715734776833,
 0.26064781298667294,0.23849745442034653,0.7390877284123664,0.523553831394978,
 0.013427652874982043,0.7095629356263674,0.9221491241391742,
 0.3176738912456133,0.8839811817651958,0.031128049523316936,
 0.015780995372090212,0.8740894508516012,0.7319836171259868,
 0.9672926163335466,0.24134441733121847,0.7322525586661282,0.5782041872227538,
 0.7895279785439506,0.3111936541722652,0.07480574864772227,
 0.03772213715274464,0.13783439438152933,0.06420534893197649,
 0.9418102539969792,0.20988989474257358,0.08564792666787724,
 0.20181388527606903,0.3198087397180819,0.3778652665723259,0.6907369330678599,
 0.6972293704540033,0.05466398735576716,0.8992309795190282,0.3840948943065359,
 0.8784949133784643,0.743853894676924,0.9557879964416925,0.11141828069948423,
 0.10497454562256125,0.2187806775191774,0.9027057486929553,0.4960637961998422,
 0.798755701005989,0.6848478531758769,0.3749912000989928,0.415904800002193,
 0.7897400776899052,0.31650685909442267,0.34232163251194925,
 0.05328660816540132,0.4857197829332086,0.4262722282464112,0.3669173687084244,
 0.4816858486397473,0.2897756945123585,0.6980029062914936,
 0.0071102077303765165,0.9246219173273664,0.5180696682163086,
 0.6494147911192717,0.6266050113945489,0.5757120215220578,0.6071071163283159,
 0.8577566099942278,0.028837819991905933,0.4334676278490772,
 0.15537265404931977,0.20836571320319275,0.018821310719105866,
 0.22431174325371983,0.49807437528888,0.5531960027089547,0.12333353872912467,
 0.30500408143972435,0.18468877953630347,0.883692640855925,0.687242671233271,
 0.968370890261046,0.15555996390058813,0.4654100775689052,0.9641386086143496,
 0.11421069280707608,0.028058938459233462,0.35390338817304356,
 0.6754929530487505,0.10613269372276046,0.43452178096151983,
 0.4330435023919367,0.2614227199077861,0.27026193806666576,
 0.39000133683271365,0.7704154048691936,0.6274807297047177,0.6225206968177449,
 0.9653175103222275,0.27394893081152105,0.2938876645558558,0.9584883987518595,
 0.7140690798445151,0.6862138229025969,0.8297194985398776,0.3884544312183297,
 0.7409822575793923,0.5852655027377461,0.6046005678430262,0.8987491945164433,
 0.22326382109401932,0.9998012325519026,0.20501566288821038,
 0.7626916221575648,0.4266677389907607,0.9561030003156714,0.3879370236862122,
 0.9140234940839782,0.3963458729956112,0.7174319031405001,0.8106714484964556,
 0.3619487919941967,0.8561149117915988,0.13469135062652815,0.5814166303978311,
 0.7344817380063557,0.4394930974617599,0.9864954898369477,0.23482694871512344,
 0.7652250491391007,0.7957231080042335,0.046998602406919,0.7937456543451157,
 0.8914184902024792,0.8368848532691215,0.024976185308739796,
 0.39024890326295314,0.6876053052487403,0.7043420172149024,0.9826488650954038,
 0.13043927694170399,0.30273791306840503,0.9301650054111945,
 0.7856238460464279,0.9995037904749023,0.33906644652578144,0.415907935521418,
 0.7846202941945789,0.1324270110151109,0.5329798782272792,0.7192537587067193,
 0.3921939490111638,0.5633189139858077,0.2796184730574057,0.779744146712261,
 0.8134133461571427,0.35748937695029426,0.29098996901198293,
 0.32081022954113325,0.4508291222688249,0.21688657606661355,
 0.04600388531662358,0.13301083047050843,0.04636463845595795,
 0.10670661380647439,0.19369643523187996,0.27289846388237904,
 0.7557107534643374,0.14207530263366142,0.7616863454900222,
 0.19175094423166994,0.6433026833404961,0.5357907451243794,0.2543589222348377,
 0.884486960188914,0.6313733099801009,0.4190115579211666,0.333284132076457,
 0.5513480554311235,0.4167949060716083,0.01090286239325211,0.2250708345769944,
 0.6670780227972136,0.6511870550855585,0.6881155761381033,
 0.004265043519866785,0.6308586658610023,0.9602175204700013,
 0.5547494856674544,0.11122040318141935,0.7333042972314121,0.7735133603676125,
 0.5291673435450069,0.4498858093814738,0.8863715810741335,0.34692401248295257,
 0.4680944762998236,0.2920057648857012,0.7098952289708608,0.1263553868023941,
 0.7175351979609818,0.8998358318563413,0.5127543729291502,0.7335186709474067,
 0.08821456812430062,0.2240597939870208,0.2755306690939545,
 0.12870552216027908,0.017516113564455038,0.7026538988907707,
 0.7233747112242062,0.815663320088378,0.17628736757456998,
 0.056187366120925786,0.6004304844480527,0.43519410098029476,
 0.30561752468785475,0.1716838033997964,0.8057523400442808,0.8068976914919195,
 0.2972383463460152,0.09521166726702934,0.8267534533846599,
 0.07934246281259583,0.8618808475770786,0.3738280008125646,0.6267022864850743,
 0.6471676553184915,0.08591229553702706,0.3209188252662706,0.9246909569937979,
 0.4828413046968174,0.0636475234138173,0.1647799811321815,0.6876998066354164,
 0.027564564656119006,0.38047273501176726,0.466831442613021,
 0.04653584001840638,0.3540121312576201,0.04423050305565912,
 0.2146334045699978,0.8447138699691498,0.17071093341471633,0.748930670828094,
 0.03310964277673689,0.12876790333829902,0.9496213770987608,
 0.9711822990956998,0.17081920206178824,0.7016188261578695,0.1287272650397312,
 0.6577935874394646,0.04712784060917963,0.449256974583631,0.5684585927366271,
 0.4432200818225627,0.33554298697905194,0.23529103161000364,
 0.42244787571527487,0.0650794095713525,0.4909903720916905,0.3066877447578529,
 0.43907989005483683,0.4237619866470199,0.6381796323039007,0.7058681272089932,
 0.46139063078623044,0.8476370209043648,0.2942374350900142,
 0.010883107424791483,0.3114584142593534,0.8755618338502703,
 0.08698777663921686,0.19865214689960164,0.5875081287097585,
 0.7399249642034003,0.8535082601110712,0.8785650811210292,0.5594804271230235,
 0.6199927139682678,0.912941751207812,0.026930235928937618,0.79053335274937,
 0.8480959066550301,0.9779244615503235,0.8338033661991104,0.18018194238713614,
 0.6175707193933434,0.04356608633570158,0.22728401320844904,
 0.7157589152927251,0.5297969082795945,0.25032074953671035,0.8629284269653404,
 0.9556948785025068,0.8583896116259556,0.025029983774005005,
 0.8630519697728387,0.7644576182002973,0.25401853521969486,
 0.03015787929923497,0.7187377488854396,0.6082984566875981,
 0.49303955755197393,0.21603599839017595,0.9494906575826799,
 0.04481224913613058,0.6506377014505117,0.06430899287829162,
 0.45176156551171875,0.03798858787440196,0.4528669464486983,
 0.7142204741488528,0.43024809421897814,0.24224144320714647,
 0.7912311513860006,0.38402985535079337,0.7983089373488457,
 0.31504111492610654,0.578324535298616,0.1411597169233333,0.8100218551733107,
 0.3452502084829042,0.9664502098550174,0.007646490246802973,
 0.47437959700444854,0.9120805726903648,0.9206254640082246,0.638985648811913,
 0.4246847316704464,0.7280089220772521,0.535729504284492,0.8619348437666646,
 0.26339317632351267,0.9818125646713018,0.9086004131782643,
 0.23011568093324408,0.9221690741782567,0.741971791679948,0.6992643074435124,
 0.3373896575017217,0.6839856313868332,0.0656068779400919,0.8827764968832005,
 0.33814105814908524,0.7852697032093608,0.5078618509529693,0.5177450684391942,
 0.24284262770742338,0.831017501925311,0.6689087452813528,0.7242894330336683,
 0.860812750744555,0.044267329744391626,0.9199659545444778,
 0.46187362162300594,0.6682949802225371,0.6310919566867544,
 0.07200635963091562,0.36687272791549863,0.834889469661486,0.0804232748725151,
 0.4686411747450212,0.024408444202830504,0.14259291132772156,
 0.9049183777921824,0.3426169614657425,0.5446196641765916,0.6505393269474383,
 0.06066276209435917,0.7806443135250304,0.905873869960718,0.36051273668702766,
 0.7303814428496186,0.23514533340114785,0.8842055842035259,
 0.06489814002209993,0.8716652030352452,0.7524564884664887,0.2138993994256686,
 0.7479182904529946,0.9315769966549566,0.8674090553818897,0.4350861582572674,
 0.6726656849511116,0.5191213738508613,0.5505042426825346,0.1820247928734543,
 0.479175014916001,0.473185524907778,0.8966551005824777,0.019971846249135727,
 0.15764122462202135,0.7946910203246622,0.11463817453660796,
 0.6410088089101765,0.4254326918240968,0.21732169784230082,
 0.33629674243738594,0.20113951738284486,0.7869405381992086,
 0.8436699423530478,0.7128160270592511,0.9868951359841981,0.9322110910114272,
 0.5668067906609757,0.5247994966000472,0.7476815317370907,0.6526435088903306,
 0.6458993077782925,0.9971492395488775,0.42857182931431326,
 0.20496063800721365,0.90354925944225,0.9515417403934957,0.14650769346427572,
 0.10436860950031246,0.41349835473536545,0.4611874944036175,
 0.23454892274235273,0.670745452456008,0.4546386787432344,0.22100381208320186,
 0.34271097035703835,0.24657379184771666,0.2912697295964518,
 0.20598244835990243,0.005739443041323011,0.007578726883081699,
 0.09812705678738431,0.005454585782413424,0.13774564725027139,
 0.5993586625597369,0.9233160119437205,0.8446126388122975,0.8642045038935344,
 0.8906304431724968,0.1682719896139434,0.5965933153327991,0.6390159366958441,
 0.7542989473136976,0.6693189571422937,0.028123782053194213,
 0.22026610924792422,0.8506347658182642,0.35146638626515125,
 0.45147280759655173,0.4292667692393086,0.1459330895076374,0.5050338051915626,
 0.9147682219031799,0.08940542239885407,0.23898205991189236,
 0.11310960570293249,0.7871003885810938,0.028688229844719304,
 0.3853257969249676,0.23925351257822192,0.6953602667682429,0.4412181086644319,
 0.4676358875627755,0.3305246768152198,0.0573068610021229,0.3881830657527602,
 0.6344159562456939,0.5054152097573942,0.05555260269904405,0.1729747953084223,
 0.09428100744077628,0.7128579051600364,0.36005540273644343,
 0.9151746837126459,0.4007288488423426,0.34652674319512033,0.5042064967171291,
 0.2811135506503857,0.774736732324894,0.01545261851043811,0.6331395718877191,
 0.9849367118672345,0.2050445643372436,0.4793343781907383,0.16329708165897572,
 0.9964986886616103,0.2643981646105258,0.33733802325116224,0.8295095719812391,
 0.6910644259824376,0.6102265980874698,0.7033718766570303,0.8363665893832436,
 0.7410949504079867,0.20109897340328986,0.6580207438728066,0.3862674504754011,
 0.5379589421778807,0.7294161760924555,0.4024065331918626,0.609843200407034,
 0.9587218771200878,0.2009045307411752,0.743239863654092,0.6224436326323354,
 0.9963826834759594,0.22561368449556019,0.3220108442600095,0.7139584678197838,
 0.8682924773189263,0.17245048388030293,0.1275153491718094,0.3861235579130735,
 0.13758749391976444,0.23440771419689144,0.7007333037568585,0.630178176514586,
 0.5562461436664037,0.0779121333865459,0.9373906963180226,0.41931732143980927,
 0.967784569694228,0.7461249702244052,0.27320921240613283,0.875081456239631,
 0.9060566693078974,0.3054032009373848,0.933144540726351,0.9959734085892309,
 0.041738620912878055,0.6183998556776551,0.17417795584919316,
 0.8971264598892497,0.9748129938231912,0.7002310442247037,0.2740443363213201,
 9.021975421336093e-4,0.18681236170735682,0.09366158374956379,
 0.030674437468485416,0.5884294364409324,0.16936997399650444,
 0.00989317572938686,0.545213573494217,0.6595183334127367,0.22582040964031336,
 0.8234131856316619,0.6253072797131053,0.9784724188287144,0.12843444450073582,
 0.8466383093579216,0.6214539839754731,0.6427758444604262,0.30544583717487384,
 0.8149632766835718,0.6729393504037724,0.15878455909965994,0.1266330290763784,
 0.6681900436408881,0.16930358840176463,0.05077761983673457,
 0.5618725003853253,0.6138344666398561,0.7072681994758448,0.7162084689511794,
 0.1156947284341504,0.7261396802790114,0.769403138683217,0.95712283593822,
 0.16444812891295357,0.10360949277959763,0.983637585858321,
 0.015564988656084133,0.7013435058021893,0.771936327582719,
 0.48318334917444483,0.1445118153768019,0.5822806033721897,
 0.27921567123639757,0.7026381531165088,0.5305049018630263,0.9294800246237763,
 0.13062324662803926,0.6054313843272947,0.41438159220160853,
 0.8979313329017853,0.5962094675870789,0.9376715026821892,0.5157689240839047,
 0.8641442659066958,0.40399805680078327,0.4780882787022819,
 0.08026219360390374,0.29184168622066453,0.6972560190066832,
 0.6788405275505216,0.645521558500334,0.8818722564788121,0.670524600716243,
 0.4120466577409657,0.9153937914611118,0.8998082846423703,0.07436138193590613,
 0.6341126962657997,0.09692919160611946,0.7137255728153954,
 0.062366789122658695,0.651835560513153,0.5601650725478475,0.9855861180558112,
 0.6434516660190716,0.3150998972953174,0.1855340036498182,0.9608545155707799,
 0.5513278498218376,0.19261426001934157,0.5579361450942799,0.8123108202682398,
 0.26422658398397203,0.9907618638759301,0.24886907129565894,
 0.5604350810340567,0.5615770190643592,0.13679876999401097,0.9248877741670934,
 0.8413601968521005,0.3254562326448819,0.52805790005737,0.09915555241254814,
 0.9451956637545631,0.8895849319615016,0.1918831599951092,0.7784282761950969,
 0.480637349496574,0.45356046282930684,0.505128519007311,0.9980582828278319,
 0.6146765939290209,0.8230210296629347,0.37261906663816324,0.5046206779509502,
 0.5825206774593392,0.8505135677434701,0.7493615485615162,0.6128946155753134,
 0.5567285074370868,0.6291672478215248,0.6331866399962252,0.25977865349870155,
 0.4732034288280711,0.22208004623962496,0.39097510397587765,
 0.5457957671932062,0.3718950602769979,0.005164639970746787,
 0.5732216398171809,0.7866471285227333,0.27746745581611953,
 0.02303223810445687,0.38184658898016177,0.1301680474654091,
 0.5606524146759019,0.951408529539797,0.8389963331994651,0.6694243721429226,
 0.5080041230177273,0.28756306388300323,0.2125094945270446,0.9772167718915188,
 0.2819271664876263,0.11839681407885172,0.3093958490444374,
 0.29739958345914574,0.4846732211645334,0.9203671812455613,0.4174683404686781,
 0.7759996736862181,0.7106646568060612,0.42715451447315544,
 0.06986911268715346,0.9139738114612014,0.6204472643824341,
 0.35798944512585784,0.5424886877878237,0.2061293561005858,0.7461699976521715,
 0.7500715380485138,0.2832564009926005,0.7265133788745471,0.8632786809232735,
 0.08209570233480501,0.8555804880678282,0.289912299695565,0.7124484087029255,
 0.5580068006909688,0.8826023542216691,0.4393413094855597,0.4775193505880958,
 0.7743081756780948,0.6462599817931568,0.4531806375534082,0.18925826389172806,
 0.6087815739518093,0.797844315696109,0.21854160424898927,0.701790341242942,
 0.10872002345745169,0.3565014153179351,0.11741461314071044,
 0.2552722582710659,0.14731551451583946,0.10570152187799009,
 0.5984878709053102,0.6985267744477819,0.43760672211608465,
 0.28886622710664533,0.8574054255301853,0.397825486273363,
 0.021848119605380023,0.5892782797696063,0.501494740040548,
 0.20779545456038317,0.6310923493078824,0.05881255327458168,
 0.2540329977236211,0.08199106438157822,0.4721716023015471,0.2734865966732012,
 0.7240863154320641,0.5013675473007383,0.059500463944769044,
 0.8693251737529832,0.38307177078171284,0.8287843756116116,0.1822355865286892,
 0.33154340980161545,0.6508866881817065,0.19409037053438827,
 0.8992022776994728,0.32978294673237807,0.8065339894071933,0.3705727623118472,
 0.09626903791898911,0.2107129784924624,0.9931718875422976,0.395834709367222,
 0.24581223586466683,0.30926804520993667,0.06974809992008746,
 0.8419308894725781,0.5631798510997796,0.22527357116649405,0.9194444423130274,
 0.12474924993908598,0.4808014452816234,0.3300202489947215,0.0308225393044117,
 0.31200902553269305,0.161348843664737,0.026873102505889368,
 0.9324931721762751,0.002348839776637912,0.9771637819168204,
 0.3869018506161155,0.04639193913987383,0.17061199908672708,
 0.3342398963054376,0.12816200900079944,0.05451462603221424,
 0.21580148939795862,0.9606700221106087,0.5184334472824601,0.5769879466678591,
 0.7498198530984832,0.8535556091295726,0.2885702307440816,0.9099877487078429,
 0.672823085549937,0.7521781288537375,0.17195177916636462,0.7025423219504723,
 0.634392273985372,0.4805709265456931,0.25685744472095773,0.07338289015163568,
 0.5501251795120586,0.7961985299423167,0.14447842132210975,0.9529695052404789,
 0.2483692839820879,0.15734116376126694,0.2433027849522411,0.7053066434735235,
 0.3470814409944952,0.21250857718498817,0.3532228186959693,
 0.14174713117835358,0.6696746074104832,0.12417947734098644,
 0.8843845709336622,0.531265750485339,0.858609596340509,0.7983236361673743,
 0.017511726316738252,0.04035075517482711,0.9756974582632698,
 0.9905802631562949,0.9817048288220535,0.38357800710843204,0.1797950769224378,
 0.5918085774105406,0.8866108329397021,0.4441861193218012,0.5896369957264339,
 0.8267739131503815,0.11139474923402282,0.39039027603836374,
 0.6568216983930143,0.8285966906413483,0.21763395523150386,0.5506793426514031,
 0.28340643477600314,0.27048124093835924,0.9685445127648175,
 0.7589773442198284,0.6506110640946713,0.14239458568156904,0.870809203028318,
 0.5387912732534673,0.9012358272149086,0.5024004651811947,0.1209062371690055,
 0.1822603836560137,0.6610915115487312,0.6048514942661153,0.47798560899239906,
 0.4276586384702389,0.4596035791983797,0.698681779540461,0.5384576983784773,
 0.34171398378260653,0.2924589030047213,0.5030319676380006,0.8150764229923831,
 0.3678923786477498,0.9092853914220048,0.36077381978278833,
 0.22315281068349668,0.2402535558865777,0.661955432655858,0.20441323520563337,
 0.7424095934895992,0.014325706438696129,0.45752814252035834,
 0.3195817783806816,0.7869175566789787,0.46356226827923086,
 0.28098975372632207,0.01747729200192305,0.06728653570110421,
 0.9580625847704466,0.41255502989757,0.9628025657862483,0.5094514506665346,
 0.8075277372324791,0.7498874891680551,0.8317196635194831,0.4843655414069238,
 0.4957792951291068,0.5986312840410074,0.3749750690591306,0.23013509533785204,
 0.22374341559061428,0.5039701627711994,0.45284139270761437,
 0.4638442628514482,0.04654615307930876,0.5363674792378413,0.1023656005297866,
 0.8290045700109987,0.7106049296122499,0.4690302505616786,0.8566606592647987,
 0.4788204008427098,0.7048971046923808,0.12084181444373687,0.4025510817398833,
 0.07421908976160752,0.4266042431787441,0.5914737660412569,0.5253940077885316,
 0.48701204991318947,0.547889502985342,0.6417360617413979,0.9180154055084614,
 0.03113372233844358,0.7367966158673996,0.7905731208471654,0.4391562706610277,
 0.8687788756120428,0.2170084875545092,0.31761658120831926,
 0.07635090375489961,0.6132767127411999,0.7898958292630454,
 0.27531953026843226,0.9476000858846485,0.38545169513724087,
 0.2971684156693738,0.32565058674183733,0.5011993037402479,0.2068939466847871,
 0.6578755474495963,0.6997231549730372,0.40141371259461855,0.8018758580144676,
 0.4304141066761844,0.21174340061048114,0.836813552543023,0.29997117577202137,
 0.6382967005448207,0.7550958999043491,0.9890745755937405,0.7355952578472963,
 0.2694644913514246,0.6034283926898587,0.6416334286203312,0.03732412476567348,
 0.6438843526405857,0.8713337733232799,0.1579014894629368,0.6036444135043437,
 0.08721823532366979,0.16611032946045257,0.008180623449915103,
 0.4296375766456343,0.5956026516598028,0.43808592906320143,
 0.44576346936996136,0.06983117484422896,0.3358079001399529,0.894742304465323,
 0.47767911183077905,0.12134543702914646,0.2519144579823257,0.985988737059523,
 0.4099569956839111,0.44528711269137533,0.27869786402263574,
 0.5499312083515077,0.3840626710860361,0.4885134012862561,
 0.006864624011566733,0.5813602878876697,0.3499730992595417,
 0.32316979727504747,0.6149769950840579,0.6851034095752281,0.770184908063424,
 0.17749093125713433,0.14397339796986575,0.8459745874109811,
 0.01923050527974335,0.8146991643108182,0.45362653210555437,
 0.2571556882968521,0.9083237866448056,0.5171720724807787,0.4202831476837555,
 0.11885515062531282,0.6133985312704382,0.24940618907666423,
 0.18546815139760175,0.30283889102897654,0.684049129268322,0.5829432029731969,
 0.7247943126259514,0.28243731167569675,0.4687228413754192,0.5602907562152275,
 0.5117494628279579,0.12043920821256426,0.5242720083506811,0.6485462613510662,
 0.7106268649863622,0.9164925348840846,0.2522843554223041,0.29472957062976957,
 0.4991029778968318,0.2985544452364107,0.12002240226755334,0.7773999909748689,
 0.6057946817117247,0.2680670646013019,0.9657142701664873,0.5888942791829029,
 0.34016222832374354,0.4404311736833115,0.030577771973196666,
 0.49754826284563225,0.9540137392790808,0.05428378765337194,
 0.9977595813743564,0.21570319539663008,0.2843509197042011,0.8129234577788242,
 0.09230799951542124,0.05535923964432321,0.15338702436715645,
 0.9294698573508325,0.18301094506819948,0.42847815729314764,
 0.17820788697262835,0.5300585334691471,0.4558536998592291,
 0.12146842993593854,0.521906437357097,0.8800220188543771,0.3180070232973047,
 0.6125484560762214,0.41655734473910944,0.40210921544396583,
 0.0895910279993557,0.728090890373526,0.3766889581242473,0.9107988187626594,
 0.8270361341424555,0.07551610434066014,0.8909017885619577,0.9984798448991741,
 0.07026536326382593,0.6112590108470314,0.46411453044367745,
 0.6632340933544946,0.921735566739402,0.3056215358230884,0.4954992303341731,
 0.3364027780852896,0.4999702373903532,0.24540663971347,0.5058612463134248,
 0.6724736745337567,0.04679991181222043,0.9271973137876993,0.3273402630843202,
 0.14689172682526708,0.2038707904189434,0.2877973425013517,0.7528536611123823,
 0.7540591227964077,0.9762470134006584,0.06752420991154728,0.5882427589516303,
 0.9526225679919259,0.1816889232535287,0.05049062239277535,0.4446239687777176,
 0.9704624205901867,0.849867365011574,0.6767328165130513,0.8296517198302003,
 0.9304940884993088,0.7051315000583298,0.4245907679417894,0.3639681268648096,
 0.994281498575976,0.47182688383131777,0.7606313646953619,0.7117278606509237,
 0.20473882620823547,0.683676229315973,0.2946241894806121,0.12680536975486756,
 0.8280149467834254,0.18008134421150268,0.49224615777586234,
 0.5163600233997906,0.49732276472746406,0.4125772464224877,0.7808777843534351,
 0.8338411383878598,0.0026340086706931887,0.7136062972462065,
 0.14796947012151263,0.03296573446414497,0.7826384646783324,
 0.36137956231374013,0.02022489958758278,0.585565251501978,0.7768004298356596,
 0.4437160546463118,0.6258287155099337,0.7975250278492115,0.311778024035427,
 0.36805430563913955,0.8010808353763554,0.012266494722452781,
 0.8055593188798025,0.07312398412445875,0.6480177584212653,
 0.013112792216372848,0.8606710127794062,0.7598735583725821,0.630079047657248,
 0.38663602028266864,0.2955575765190255,0.5744180255647321,0.1470152829930763,
 0.7175170426477314,0.06963994497584802,0.6317932686463685,
 0.33933619967897655,0.8697196862124855,0.731358288071409,0.2604775655567513,
 0.33147149515000673,0.44726308137905246,0.9224703267578297,
 0.007146849092758534,0.9759511268648132,0.6827626442777319,
 0.19813102972691354,0.5763979595864864,0.324703584087609,0.1237064371224329,
 0.24770197019132645,0.49173953544893967,0.38797701077036784,
 0.08107793003876884,0.553629648435837,0.6921847566829044,0.426560877820397,
 0.1968386571212003,0.5592563982178483,0.5668849122749129,0.6622992660581728,
 0.2234367150412585,0.44874944346859325,0.915851672304742,0.8528125248625671,
 0.27547985052314605,0.6898857782698448,0.7324926700310155,0.366895760915674,
 0.5317574358825258,0.4112912739810345,0.9924565852165769,0.3294394008128003,
 0.5906071335106856,0.7947327283799657,0.40856799320758186,0.9899800915157176,
 0.6829296040678277,0.20714317812260796,0.525529894696872,0.7364489527332384,
 0.45684405928488336,0.7837776898403109,0.44540266323918143,
 0.14970284520328492,0.33959156642572397,0.16394957017879497,
 0.7654600290147031,0.4658073515046689,0.8772488860675474,0.6794272006853534,
 0.8760599301431088,0.2848152605954024,0.7998496956400627,0.15764262641352056,
 0.9963258945747494,0.901898305207177,0.8075879161574003,0.9792810918298018,
 0.21570517983016957,0.8987417853487901,0.8388653850140724,0.5917619435580728,
 0.05955773902674777,0.9485064579174707,0.40030904728957695,
 0.6118752001975648,0.3521532056197978,0.21126461160418786,0.7149003766287263,
 0.2712374589234162,0.9661998353671335,0.7838830176176069,0.9470999280036551,
 0.10387949931586316,0.6001783191074218,0.6624330451213489,0.6451088525806007,
 0.21835893857631306,0.28323417175102494,0.7177071495551395,
 0.47516936431627577,0.44794212827283075,0.30860510478443215,
 0.005508253710551303,0.7422423036806511,0.41156614782389755,
 0.7689129070339306,0.8579169370765201,0.8903747315748011,0.9033867644147602,
 0.9948808284505234,0.5497482050709118,0.6323686216398777,0.5269980541181727,
 0.048698258898650204,0.589004061444881,0.9489452403442092,0.5631674493523586,
 0.004767212848903912,0.3216517038890825,0.25470391475401855,
 0.9533489439511565,0.5817821551383007,0.8936764769757353,0.45674039696503177,
 0.9452432598738438,0.11778342355758042,0.45249956342825737,
 0.6616903260692525,0.6975175602692901,0.9551622267763081,0.9483337343901392,
 0.09167541992428496,0.9376760938602332,0.09012799326034826,
 0.8156371670140218,0.7823050319342566,0.7862144076988035,0.7906024077434572,
 0.4467398280584123,0.40918039440581844,0.7526141520893325,
 0.05239690226496141,0.024419317363359228,0.6160903374043997,
 0.16786269494715422,0.6943321780450622,0.4877459858476073,0.8482371281924004,
 0.5161277180837414,0.04981229232181761,0.8835087475536595,0.8492439202788299,
 0.48379489911342977,0.9921895535445093,0.5507582273515433,0.3528966776980049,
 0.9953306016545594,0.3897084992660482,0.6182834515212622,0.4309807684853818,
 0.09539984799252244,0.7772907599329394,0.2676231461552929,
 0.49758490458140525,0.8875940985374662,0.03679976184766498,
 0.5730178414397393,0.0037574833122588913,0.16120209735917068,
 0.14147492034624087,0.899545345164008,0.8042326050973511,0.521459446079348,
 0.29287162017499124,0.18890753859662424,0.4476038338905748,
 0.36547992709761123,0.5714658827029969,0.41249063545427644,
 0.8831501080248256,0.8038117648858893,0.37126920152654974,0.5613581279068662,
 0.8898336568580032,0.010112872966296371,0.009221754890894496,
 0.4562341161262192,0.2947105512031414,0.24330314263088237,
 0.03339815151484582,0.566488177996026,0.0480200731388426,0.3815010940014525,
 0.42655143379638205,0.6999957085126859,0.03770872253583224,
 0.10682508445883354,0.4386602006088012,0.7629481086505467,0.8964762021967776,
 0.37616836273922594,0.5285875331261393,0.40589617048646653,
 0.2926927471741627,0.8754782171090957,0.7153839928184985,0.6234010838836095,
 0.8714687580392173,0.6731123119326596,0.7203060959983827,0.3155944391770831,
 0.9347739566900886,0.45623548048961904,0.6526400690761437,0.4499154525915252,
 0.5721818812205819,0.7482514267283801,0.41012962415236065,
 0.17026806807740502,0.5346456795412995,0.33406242860762214,
 0.09005915404355691,0.1996070886503556,0.8612712013910289,
 0.38127599271877555,0.4470971213713666,0.8863866161068554,0.0560825829968814,
 0.027394896688639148,0.6930210906352176,0.42903574064992,0.02507343389250072,
 0.8444061380990001,0.30097744324447917,0.5061141971336838,0.9170606072701813,
 0.7504660418044856,0.5544069662630664,0.7010960705469739,0.5101840644223973,
 0.6115494964824252,0.967446256527826,0.0574838367974807,0.6201641093005508,
 0.9743245704828694,0.41402040121416483,0.44462813320716055,
 0.8928546629525375,0.7847671555456033,0.857279230909988,0.5964895419116164,
 0.9897288491717341,0.02114628098496374,0.7453170440756407,0.8453297365079268,
 0.7919653400265532,0.13806414959155333,0.6223187126121154,0.6278262036127352,
 0.8032968367050263,0.56544465535302,0.6090570090983486,0.4762915683534078,
 0.6160437963393136,0.8046701408866359,0.8539870167084377,0.7780525414640176,
 0.48221527910880413,0.4587456518812,0.9845973797686156,0.33899669951536426,
 0.45252830010963274,0.09764607930651881,0.5733046657853882,
 0.9645917933553532,0.6004105376399247,0.7308637472411765,0.3487247341961802,
 0.048353765478705846,0.30480281673478193,0.9044317667331191,
 0.863127357677504,0.8789503449992824,0.08078717303941851,0.9601706713064405,
 0.8817354873432238,0.2675888885516373,0.9227314059343775,0.3230120749347519,
 0.34581447010396327,0.44602660758561175,0.015611734501614793,
 0.36164549348631314,0.6316111269384878,0.2919802557817075,0.9428478803602962,
 0.1238821261087717,0.4723898384869305,0.6272084522930987,0.30196355795739394,
 0.5350570016605164,0.4370383506782076,0.16324997984083778,0.8391222752039511,
 0.6728615257271507,0.8053035487997589,0.7603959879614097,0.3926304600040026,
 0.30261031262468496,0.4591401382286143,0.13182644541887967,
 0.8609962349307007,0.6070529903281793,0.12425633040630446,0.9090205129335771,
 0.30741242315858064,0.073664278673226,0.19742586927838757,0.8770800276232368,
 0.199235260814825,0.07655761218462453,9.587769191191642e-4,
 0.6281217259227392,0.8063994718041132,0.8912106115151597,0.12925510466498702,
 0.694498774858605,0.4683263480661052,0.01614954967583815,0.813715768239785,
 0.6027848084193395,0.2113628576367227,0.3777933502244857,
 0.0036168776177191164,0.26288950432473746,0.144740869345246,
 0.5301368283756341,0.0732444664054257,0.4287795974823517,0.1150958167109607,
 0.02853351787999525,0.14832937204541685,0.3861346349211724,0.773498138262579,
 0.81686835549301,0.7541925662344848,0.11715428111295478,0.910103110013175,
 0.7734014222679053,0.16112657566689392,0.9643499299072044,
 0.07253823675269533,0.834255958144776,0.8596245875448261,0.20878007550782984,
 0.4831659031335468,0.6785466677418557,0.4612547680598381,0.7454902334098573,
 0.9683731662504769,0.9504924122905769,0.7801885376176674,0.4178052650559124,
 0.8999916238595271,0.22272836901476245,0.4553252658508916,0.3820778221241019,
 0.46588027484517824,0.23752129460379012,0.057636241048802606,
 0.5569260922910914,0.526353942084279,0.09852600459609617,0.6856928421434584,
 0.49155009180230413,0.21448980034333998,0.2697711899704024,
 0.9124165589822566,0.4075492030869974,0.5896894404773235,
 0.028906993638018408,0.14760374418397382,0.7954668530392389,
 0.5233736591395463,0.6796122646388615,0.01516566717258816,
 0.44692710703031624,0.6217584587094386,0.7540728613503748,0.4304381685159998,
 0.4729326628009154,0.7705689843339932,0.4691208596929868,0.5044625220197677,
 0.8377132933793429,0.1268992313791344,0.7554762181149102,0.2980580865720581,
 0.7610210908946782,0.3942387379638491,0.6039865943062054,0.8923322539975241,
 0.12952553273473777,0.3506607356742769,0.9164048673582563,
 0.48897392337142487,0.3888024179060233,0.8038337999438717,
 0.03504543789918835,0.21916582741091006,0.7085336337945474,
 0.7794223458019152,0.996130855906694,0.14349694698038895,0.8286784068800975,
 0.06601595551965556,0.35702759502351245,0.10550779125463372,
 0.5366431939598744,0.9835391246414555,0.5697231314674744,0.23453653401209196,
 0.8913968309051731,0.6604228342920789,0.6853897211671995,0.1811904737477823,
 0.20900754000255872,0.6498514271040505,0.5708451847919702,0.2838161928162021,
 0.38282370649708075,0.5546517175721198,0.03207115745246547,
 0.9373674572333596,0.8260251958715614,0.7093578729887265,0.30691941083921714,
 0.20252655795726893,0.29445805493136634,0.6095943303955572,
 0.3206703663401417,0.769339523931436,0.9670423947765244,0.1622955284636064,
 0.7300424278748839,0.9967255439579318,0.9528621526396154,0.2056206015408173,
 0.2548813754838415,0.3719259986679433,0.2074649725814306,0.3954465145141267,
 0.930640129312103,0.335898958986225,0.23128661560740882,0.4628826962608662,
 0.27708392011628047,0.9232342462796214,0.3766802964103585,0.4139882614439493,
 0.31309673371780344,0.288649847150413,0.586592126241078,0.6361551589380345,
 0.5544646186711455,0.19975837750744196,0.7764025308168436,
 0.008536357989629595,0.13143146790418836,0.6637554773103463,
 0.8847145206999601,0.744945971152095,0.06851289394523985,0.3057206839802886,
 0.10220097269341688,0.6344587893134054,0.5994486944735472,
 0.028867579155287704,0.49600506944489975,0.4694517416663009,
 0.12476289740135793,0.5124495860685645,0.03998746629963723,
 0.8820508922626554,0.6606867367255791,0.5110884783897905,0.7941789365500483,
 0.774117150087666,0.4365335797599387,0.7048962007570746,0.10219279461061792,
 0.4373700889936054,0.6944055669346126,0.9800938643285337,0.3956681472844443,
 0.5511545371527156,0.576926399575964,0.49557173844838265,0.6331054387676578,
 0.8262471847985905,0.20363494608005328,0.4570221746948462,0.9232277454177903,
 0.8025417804764865,0.28858656778689684,0.6061569978832461,0.7041843397650367,
 0.6807212590513771,0.4136679456750757,0.3141980940049016,0.5255535468414126,
 0.43094791950561473,0.18693991242453822,0.2435482557366926,
 0.36807986627025646,0.01773426862050953,0.16951121918206546,
 0.6076891289448096,0.33675175872858787,0.158314690337088,0.3424559857803613,
 0.8092770620354587,0.20911379526923213,0.3628021713197962,0.7875938987766005,
 0.5618989817111876,0.856694791224525,0.6148973483443232,0.5427026253328995,
 0.2621928863644213,0.4072893561917995,0.646223123755072,0.14391457371789063,
 0.7104894385886547,0.5876808583481617,0.17796524614170273,
 0.13958742967002347,0.20983817482811595,0.2658832081877974,
 0.9792178547821232,0.7909482271559177,0.8664482802245633,0.34644594441437704,
 0.8369786921130014,0.9575382382787457,0.5530787214758159,0.1787935295490557,
 0.5161734470763908,0.8578532405418872,0.9470541727953972,0.2013793841094883,
 0.28616056792473843,0.5543602134931624,0.9380034215062917,0.4788860862447941,
 0.2916599194892473,0.7233355943011001,0.3305383239490699,0.7760702219652915,
 0.45381063215083917,0.5432818748345147,0.6961754132834721,0.9692139268088661,
 0.13712454633882942,0.40945047739742724,0.4897043994035972,
 0.016712690413920095,0.8076581776976987,0.4410524451712895,
 0.8203822016397181,0.9314639921101029,0.5723589468760027,0.7276305873977236,
 0.10216321424417196,0.5320951460819089,0.5226501259774508,0.6226879807268961,
 0.3110779234787744,0.09266293776151935,0.8516425797273263,0.9935331975560086,
 0.010699452683201272,0.5453690438504168,0.2928915090475628,
 0.032628270079577915,0.0583258742199928,0.8379222552453248,0.954002640052113,
 0.14755645240559778,0.06793145355522734,0.03743030863898211,
 0.16866563663092127,0.9173856112205837,0.7850069487108798,
 0.33060868676308686,0.1351663297590242,0.07897540312360407,
 0.9919294673414096,0.4855271373243748,0.32551825204179796,
 0.14808235663481195,0.4789729407423011,0.3139186032248322,0.0767876863486211,
 0.034663882011768954,0.8973065966669458,0.051010520743246635,
 0.4005777640440489,0.9799827120920228,0.42213124191888174,0.7575308886230531,
 0.6684940570957408,0.4906350661090608,0.1953156756758524,0.5384514504306771,
 0.30213988104048606,0.39464059213141367,0.6635361188478126,
 0.9513275394480727,0.5586705874175552,0.7626214432780017,0.42461106906562673,
 0.49979284488347364,0.36805255887550925,0.5303636916589866,
 0.7533945993647682,0.5591866219019248,0.4072343509181553,0.5780452137719382,
 0.8907821777905024,0.35427358933808883,0.8165747982220551,0.4882830967994297,
 0.9600613127927782,0.34196263267148796,0.04506378980367065,
 0.6447098326237566,0.6303930323940796,0.2397280004529162,0.9641571879213939,
 0.49233065759315475,0.3635489157261208,0.4550342264889701,0.6445236396356783,
 0.06938251608969037,0.18476911769296012,0.7564550888324824,
 0.5161501829898018,0.6220294668470366,0.3328733504878105,0.6297741645192045,
 0.9023691528599723,0.26587934651399026,0.8140907735552255,0.3706760019273496,
 0.8619598836747361,0.06334486860654087,0.8605219805598556,0.1799947773295021,
 0.8895601609256302,0.10021457681592683,0.9752870542130683,0.6686548836604201,
 0.8482120531412752,0.4746234962956054,0.10930373068832266,0.289272895079991,
 0.2554704428962392,0.5854055149443876,0.9820408966524226,0.07583355321027674,
 0.8574896211164893,0.2954225293012257,0.2250200334500404,0.4794512292139811,
 0.110972039292075,0.9446591372711346,0.963917781645501,0.6452126704305203,
 0.9996275682976271,0.24350712984989709,0.2092772014483949,0.9485773605580886,
 0.2896648039822617,0.7489420975861575,0.06768730586255145,0.861864783960808,
 0.6540938131764615,0.9334114583792014,0.6881424091862204,0.47516103109101027,
 0.10316143686409318,0.17992854971323657,0.42545188311518867,
 0.8817317316575217,0.7543336697983278,0.021071455784641424,
 0.9478024524896055,0.5901975646804092,0.6540699468890043,0.1660876516188643,
 0.5459847766006842,0.015341349106931412,0.6208826442776735,
 0.39920728717523746,0.6682105847066745,0.19101807139890803,
 0.9339613978941785,0.6012558053318872,0.9899041188572009,0.16216132990809373,
 0.7228623690902048,0.8678781230985656,0.09007832776707025,0.7596699455718862,
 0.8691329075379466,0.13209789081411805,0.5102692770004821,
 0.23283327360076178,0.8484405922741969,0.16036875677116735,
 0.3047253654390283,0.261214496274607,0.4428087436905014,0.878763953819177,
 0.8891753460475498,0.4168705794148494,0.6431189561057209,0.20737618776423372,
 0.2734227760151684,0.33325774219857596,0.09250124820617978,
 0.35737526817853493,0.6687428991917964,0.16574657627930833,
 0.9966551676092944,0.7236355827644642,0.8907003171338901,0.07261356274007291,
 0.5761737048701596,0.3165305414282038,0.6385931377113203,0.9548602809212632,
 0.5718840249165823,0.15550534649730763,0.37167641470657997,
 0.2795295940801389,0.11678993010218053,0.4738819097019026,0.3875171436060647,
 0.9276614178878954,0.7535845822872341,0.49343425961228027,0.3699906628859053,
 0.5563442072393773,0.7323204301919395,0.7760359760525376,0.6790301336863287,
 0.23753796498188617,0.7013259930417817,0.9487436299471986,0.4441748618523146,
 0.5592273948605244,0.17900444145448813,0.20849652815525554,
 0.2933620634414512,0.7496167375085433,0.6580041433914099,0.26124314785552727,
 0.8081856505029508,0.7490892515838586,0.6907965234226465,0.9735042972340234,
 0.03020981300402681,0.4915233414620468,0.3614779030059695,
 0.21810387646191298,0.3534149689578011,0.8511711247002461,
 0.003985436510736351,0.27576967126779917,0.8626353045448311,
 0.7078277060933723,0.3578034317083191,0.06902212891995485,0.8355641640610225,
 0.2910490250701582,0.40024274422201545,0.135916214604624,0.08731495891496843,
 0.5941487033815684,0.7696687759134344,0.20032884923236716,
 0.15739941438854443,0.5467941602629641,0.9039188450396891,0.2002495827827815,
 0.7401797444280682,0.6429047418522501,0.3334250245826912,0.48317440689267377,
 0.3014552096572237,0.877043174841126,0.9587267312320283,0.7265057272420745,
 0.5432825966469603,0.9463924437470701,0.4505916799821924,0.5690870082804996,
 0.7436842149177219,0.3894626327697742,0.5097974439746707,0.427831128753426,
 0.22109737320967915,0.6553819128003466,0.1569160462688265,
 0.43545603633084695,0.7560139057534668,0.8788672309147386,0.8668861953090593,
 0.6894096924678088,0.7119085627893007,0.7693152406798799,0.5791890009550354,
 0.8569926138259679,0.739780945628456,0.5621937601940608,0.137525616094764,
 0.25490177156440097,0.6470737844073301,0.424137230860678,0.35131619875626263,
 0.730994411622971,0.7976954115597398,0.2915595234044285,0.6219569347622328,
 0.06206496837406972,0.244056871933089,0.11213233363623554,0.7065350151624229,
 0.823211278551946,0.5182016286072868,0.02526803959738022,0.540791998625765,
 0.6349349412236207,0.617928368438361,0.3927913807940945,0.6055884885832571,
 0.2387328204870115,0.8127400785051438,0.6964724779011097,0.12465399063816895,
 0.7860227911848997,0.4210014246219518,0.5070136845091935,0.46878817533121664,
 0.07365475235904961,0.9545075846708155,0.535213213313671,0.9643638953674233,
 0.9633544327541893,0.42481706951345155,0.5548980687415842,
 0.21090988177512715,0.5830667907720205,0.7358807154428245,0.8837674722504933,
 0.0812386555473964,0.5495484934761408,0.3621027600709694,0.278286115905805,
 0.4472956160770124,0.3298186780944409,0.4558869443220527,0.19152614502444876,
 0.8941435952004874,0.6220086954783218,0.5172700820737752,0.7498790965561775,
 0.9625568892066197,0.3521574553497353,0.27222725479899723,0.8204299713825737,
 0.6660048083402664,0.8641512639198005,0.3068019120423985,0.7747313647202739,
 0.8105787400644944,0.31168451410372067,0.325960379190112,0.7114146329619103,
 0.1298338129567005,0.21984534693669489,0.6662117348615569,
 0.23805945550029772,0.6505835210933625,0.8604514169658943,0.879890710625726,
 0.22029283813628397,0.4830662284649784,0.16961973517355255,
 0.36323461172842286,0.2879648435259017,0.4392825920063388,0.9325008692746554,
 0.654823748431724,0.6389746999595144,0.4633254497095066,0.27613608238843534,
 0.839959833691576,0.6604243973543023,0.2606155082624433,0.8210244834913205,
 0.2688126680502516,0.44113667381646093,0.22977213669901486,
 0.8089189033215543,0.4345263191830042,0.11667907877607153,
 0.22933257570590804,0.9084218480141554,0.9116252547967623,0.7410052946590453,
 0.5143551211809579,0.22934819561322828,0.38735046500748316,
 0.17362991912283188,0.8904492011555785,0.19946850596089727,
 0.5312395100163231,0.3063707172823127,0.02627569584189937,
 0.30599678550791287,0.008437513294838439,0.9272693118747368,
 0.6183020328811428,0.9767309742828384,0.5515627227504442,0.8183874581998477,
 0.1562481605376872,0.906999413809864,0.729965500131369,0.46858379270909356,
 0.6994648727806777,0.5883145900254874,0.3579465784851955,
 0.012108393351389468,0.6206068500436673,0.327820099580336,
 0.09687039494972915,0.651059659462772,0.9636755925373695,0.9066630862945628,
 0.0428795116331282,0.16207051742389778,0.17172510982077016,
 0.9877894476319774,0.4077382949580314,0.36515084444797663,0.6500163780583343,
 0.11395642045193899,0.3760687798887634,8.868682740088563e-4,
 0.23894649427185854,0.07061666585200188,0.9707672003838695,
 0.8339875617353687,0.7480361470227052,0.9209400246990221,0.18379817264262743,
 0.9210415038132318,0.8993373400939256,0.6381145134403224,0.8921807986826552,
 0.7552410276154917,0.30325273323610435,0.7338478717096436,0.4618072938246599,
 0.14651945484469442,0.8238250158930468,0.46216869404978234,
 0.9831961262783205,0.3741437407027832,0.8427048159153719,0.1508833144321805,
 0.7670130697187074,0.9454323211179709,0.8063282520265871,0.6940103350333418,
 0.8411106122242746,0.34694340852056704,0.6668140814516875,0.7111837780520077,
 0.2580258877999225,0.9336123216656085,0.5171098128711098,0.35474615587884106,
 0.8499356238864022,0.4040131872738145,0.877014328648707,0.12309202033911326,
 0.18709043944529902,0.3950874893553946,0.4866068825004497,
 0.02759523595404434,0.9206168340901603,0.0699301171794785,0.5825095306075864,
 0.19587803035142526,0.8216689899594702,0.020442412221667206,
 0.06994717275887252,0.7970414699822468,0.5876047098322651,0.6999794761631606,
 0.7811417532280849,0.7998782648900663,0.2294192693540691,0.42031166350496585,
 0.5850060005052181,0.5988851056699404,0.940858898400004,0.8049576543940203,
 0.18009706011884785,0.6217733347580733,0.6336388413389086,0.9346213249936748,
 0.8498343927164812,0.08233182718910248,0.6283813866120072,0.4590714584327742,
 0.9208547628987007,0.3510055836812138,0.5817509533027232,0.6283651451215113,
 0.25935324436890905,0.6017328777612229,0.06557936437617506,
 0.042298396343554034,0.18825767936739468,0.8064452309345752,
 0.8853407828785238,0.7179409095135165,0.20439761063050743,0.3598380039837157,
 0.10205245216102077,0.7375358588950783,0.3416761964980264,0.3396201431165885,
 0.11601545536700142,0.5345521124671317,0.7667847657217212,
 0.26760402636555214,0.45127759491566766,0.3604122594834368,
 0.8480874636214553,0.2813022218956932,0.7182325741012906,0.1552218988029519,
 0.7374755144816345,0.364709844211594,0.9891223034581322,0.8593629647652623,
 0.20582162733423748,0.29003428483814775,0.710678842609046,0.1848585294054349,
 0.17868012304706626,0.5153628833042182,0.16312823563597334,
 0.3520562275005247,0.5500414688196659,0.6791936024107506,0.26837099605219716,
 0.8025852323870062,0.9672384253213787,0.8445271043619383,0.7481767773195468,
 0.47703147575228844,0.34206859076817964,0.4400546472283471,
 0.5133387947516126,0.0072852536527214795,0.7025142245291522,
 0.22750402316757967,0.5390678803308023,0.9061700247859097,0.3254796297283249,
 0.5426684627745262,0.5110193511638449,0.6178484792985421,0.3133383375569685,
 0.7130232633022462,0.8061524025654447,0.06761237044709101,0.8187383595120247,
 0.23186034873676087,0.05946675888261499,0.3104368742599375,
 0.9696635154938384,0.3347212275723107,0.17068541782047453,0.642169897356901,
 0.5878233425214483,0.4215455697949928,0.7150352120380961,0.3525313684875955,
 0.3057715597603434,0.08277061795671792,0.023066998124663174,
 0.028149338958494585,0.7852217918345286,0.48558626803906935,
 0.16811868429859522,0.909486000405104,0.6738651971653815,0.20550576450274793,
 0.7653477836436675,0.7584048858643185,0.3075335370635657,0.45226342186494706,
 0.5202115596622621,0.38117441826001497,0.3671302838592987,0.9148096020713996,
 0.05724947115140833,0.20524273629780498,0.5397328391318121,
 0.8168881549023719,0.37783206013573856,0.09351230565701507,
 0.2687987545207242,0.36651799935994844,0.9242166688659259,0.5084219266722071,
 0.6105705025707545,0.10924020234108478,0.936087289277897,0.1759321731958794,
 0.07280613882221265,0.5084161259007741,0.47930455082648304,
 0.29021676981729805,0.7720494788655261,0.7917701182876705,0.6142005407953215,
 0.5520703683933329,0.2938153864122607,0.42182352342370033,0.8914842963231213,
 0.8999672142883538,0.9121996577711559,0.7106050675397844,0.6910362516092663,
 0.3197655343267495,0.6248609662266081,0.9761230910436374,0.01851413540806468,
 0.3390900397475941,0.19203069742421208,0.22037373984406095,
 0.9336923627269642,0.39292145793504396,0.999792441238508,
 0.024517781089506197,0.6110117274507592,0.4364123386213803,
 0.8282082108580321,0.2633024940369971,0.7097519374739327,0.3156208497995715,
 0.8008367897279313,0.03230313934255913,0.6015749978523977,0.6089043825533696,
 0.810200160642645,0.09193229110330425,0.42664106333359475,0.5673504337006321,
 0.6652421668284584,0.7182801068937645,0.12118402245552196,0.5459707636040665,
 0.7244267236091508,0.8707506184611761,0.7496256058652034,0.04352335077439562,
 0.0441865260354235,0.9959033534169967,0.11757877813030349,0.8766813149370769,
 0.33045635131896367,0.12246536184344026,0.8861553063409563,
 0.24676258481643554,0.283033588739808,0.5186465703636627,0.9005783164284656,
 0.7659488067740232,0.13557378725211044,0.4400123498310913,0.9143776365310291,
 0.022845360945858095,0.7659011412811899,0.15094075719542666,
 0.35398974884092216,0.1290492230285395,0.5259656345597519,0.2706803753637793,
 0.5312502871694406,0.3530678940709475,0.17193059881110162,0.7450294120054669,
 0.12477894164426018,0.9576477470412498,0.41191761784737,0.21077437104186292,
 0.1590184703210128,0.5523824866353865,0.5984156580330939,0.3463852822203919,
 0.6875762628906728,0.7237597712614352,0.7609241865539751,0.27502616716343065,
 0.19328558489482361,0.1831865896019308,0.8267170237184359,0.8990631114834577,
 0.2335100987698957,0.15427257230764246,0.31973409775390904,
 0.5163223679458586,0.3121664379708984,0.8689237343656058,
 0.048840355983389205,0.13722418523215896,0.8515849485787725,
 0.9313565694585101,0.3617778545320808,0.6356575175255054,0.8289969011901929,
 0.9606528036386733,0.019852008510825025,0.6409086434527095,
 0.6473500431885314,0.9854788965400628,0.4840601971083822,0.3747875888906812,
 0.06169209199013059,0.15391274912899855,0.13754483807221163,
 0.5129165950281789,0.16380079862525243,0.6225527273498657,0.5056408553008904,
 0.7025718565168373,0.10350918711961288,0.08694717139455788,
 0.21371391456716982,0.2269519559354023,0.7656580195925138,0.5303567829838431,
 0.7782188557108116,0.33715892571415096,0.4504907922023499,0.6366452338631859,
 0.200915833837776,0.40713614597956815,0.6807687463092783,0.5911758312569965,
 0.03833970611764781,0.069732361765112,0.07561029028553945,0.3700437453717278,
 0.09018910365847899,0.6361885760272018,0.7085496795245254,
 0.39640791425646393,0.9656637456870816,0.48569398547879283,
 0.7910110618610339,0.7158359940728451,0.03252799249672256,0.7697951416022343,
 0.37486620932307124,0.2926080382201576,0.7251162797287397,0.62874001240831,
 0.23628077100169265,0.11214556662914643,0.7414925617126469,
 0.8660039990746152,0.6811433649310031,0.5840228258679546,0.815588219206111,
 0.54252898684738,0.3199428322753335,0.789098594290649,0.43524288039200054,
 0.12435992641430504,0.5320260023352945,0.2136681215274301,0.1260428295592606,
 0.12418060983527224,0.3743143357515051,0.18626910758063076,
 0.2559327916218008,0.06798053486303823,0.22032750959278324,
 0.05593577149829365,0.641530321608198,0.6150628898453318,0.21062943827144398,
 0.6909159422765297,0.7828693527473753,0.23694585446349625,
 0.36622926375517273,0.7650498496192218,0.43289036243031775,
 0.26749740530409694,0.4387620845138215,0.19842977167589315,0.575722413585839,
 0.7506158953132296,0.13420575872639873,0.20778199680191645,
 0.25906962346270934,0.9792414530400403,0.6370122495088988,0.6655526493502293,
 0.4115447391570204,0.5624505118490519,0.09246666857793451,0.4787881410013831,
 0.4067316866487145,0.497516586745243,0.379227938329471,0.5122857865766562,
 0.2576100403404765,0.7412776392054699,0.18495987743867348,
 0.35676255899637155,0.6077387897556203,0.3210917685417094,0.9141295129578948,
 0.8038841901035909,0.839236910142128,0.24739595190741503,0.9216027592198334,
 0.5352647291273236,0.7276613428477671,0.9927905415401727,0.8347733277542798,
 0.010179600753971951,0.30758798368744167,0.17081638614800332,
 0.22622117443232082,0.7419383850335589,0.8420648319751821,0.6739286588873784,
 0.15849568664744662,0.2569754263198445,0.8762917804962447,0.2724186251065067,
 0.02075051458308963,0.9745330997727488,0.5122083181094927,0.1723395440763703,
 0.9986288523967252,0.04665252935876818,0.046299604592541344,
 0.7075766860332984,0.5488415650295076,0.683951320488097,0.028671440641723556,
 0.8492884786260694,0.08978680073334966,0.3659308551240005,0.4845038485080043,
 0.7158248990393691,0.8096250520423556,0.49090574566038614,0.6724458896015786,
 0.2056562456115052,0.49104022819160054,0.5116485729053574,0.595470633494491,
 0.7197908791656892,0.8339748121405687,0.33945659899600145,0.5577275302196439,
 0.13441163047060933,0.6080057185343215,0.9584699807683104,
 0.08265491125124369,0.11154971891790644,0.7697418328587627,
 0.9125237817807641,0.4279782748217241,0.813060463283521,0.99954692506419,
 0.7053863734116514,0.7126890488953785,0.3110390632509734,0.8742888430641775,
 0.3051664197140215,0.31929080813586286,0.4775723351732555,0.5139135261317462,
 0.9313766255428669,0.3202269083709811,0.12271733583680156,0.5304875873999922,
 0.1919193110813896,0.10795121840173705,0.6226984835590798,0.430404093099757,
 0.2291517537709734,0.7180203432150698,0.954579735506694,0.8080606002212274,
 0.9326830214008661,0.4706074237564162,0.8322390018772767,0.15508077273625698,
 0.3057939799217163,0.6135945558895742,0.4716773006407631,0.9803107957614255,
 0.5254492938048917,0.6272503391656488,0.3974997783566093,0.9923317993867937,
 0.27079615071275664,0.5290432079111467,0.4286717354917703,0.9355643530398587,
 0.08327240582392859,0.7115321506255127,0.1368456978998286,0.4809918346099361,
 0.24464787473823357,0.22797588975334337,0.9160963791639894,
 0.11776083873031395,0.25261996056907055,0.519264314185444,0.9098035389588917,
 0.04244052806063736,0.09846553081460296,0.5034479995464843,
 0.7735137768727993,0.7807973351523672,0.297916134289276,0.41869874326010226,
 0.3640331262727299,0.07025441637010355,0.5891533918719235,0.5575530456510798,
 0.6814667985658972,0.4752054997975025,0.6908728565734876,0.7189128619030534,
 0.21484345523936588,0.8520422179918914,0.5142797621420653,0.77875469117485,
 0.09851318774674656,0.43894498601492127,0.02655685224139326,
 0.020422822289592246,0.16438398174594993,0.3392621799248353,
 0.4032062537908121,0.22414417539979403,0.8925156823579332,
 0.19807200386521717,0.09544832679011672,0.02061760170871163,
 0.06174780841146854,0.6256683947076184,0.13656336727629848,
 0.9939027725571653,0.08519990533778732,0.1329689562303784,
 0.25007179483718034,0.46464845649958,0.6257168469883304,0.711257216237526,
 0.295882255922011,0.45228107920460414,0.5915896392890867,0.8238626439569201,
 0.41702784854547714,0.34546480086741216,0.055859976322182314,
 0.232917650557406,0.3659475484769751,0.6467812695483729,0.8974115013358311,
 0.8764331289172481,0.05072401685548544,0.6507942773930977,0.5791970151823826,
 0.5883144266612492,0.2734005081657178,0.13111774412772148,0.3799071457936821,
 0.3474630475064653,0.7621260428983329,0.21594676687544023,0.5866748979129104,
 0.7244181876203359,0.8933539852532696,0.9354363660126104,0.35486336212470937,
 0.6303107414060678,0.36528083271207523,0.7940471805838658,0.3335283078159944,
 0.0749052761915997,0.41899674469302295,0.2798686730213482,
 0.47812559023729206,0.10751311689556953,0.324306696200261,0.6293597842642352,
 0.7113592017191337,0.7666262135785664,0.3411033291121345,0.3592216014965025,
 0.0044114296107757855,0.01733830958922522,0.6590075011682858,
 0.819518523998594,0.6830589333707495,0.4907737422057715,0.5542027425739717,
 0.5681649361557011,0.8425346628188226,0.39569804596608493,0.7709395857043072,
 0.1340727002147979,0.17702811172595168,0.7658418298815614,
 0.38785861977977976,0.2705064131614411,0.21583953897949426,
 0.8479970002752621,0.059206153668768025,0.7754081578706586,
 0.19206135650696377,0.44724795986117627,0.7146875218122306,
 0.8388686292351777,0.690241567585111,0.8949669054721863,0.5146583360942619,
 0.24350905243142051,0.23407029487674136,0.35481644399039225,
 0.9813814104452236,0.6058107078163946,0.391004787495282,0.4508412157777194,
 0.1254391326664045,0.9126425505451374,0.4151750280305232,0.09808184178652057,
 0.2688803961332964,0.31665533465026285,0.9742550972631039,0.6391480559369966,
 0.006316138842260233,0.8772963090228048,0.9072464732466945,
 0.4508017871526119,0.8111742275857537,0.1596237161251962,0.4659068637892638,
 0.36419024259167676,0.6883256494397137,0.17954858271663754,
 0.7621685344795465,0.2894045908321863,0.6427077011361709,0.7193992526370492,
 0.4405896040399161,0.13113569247518353,0.8843315835486818,0.4466294694770404,
 0.7313397161477424,0.2153354369869408,0.08280889432098182,0.4528941609851198,
 0.8691988310808774,0.48068102770243804,0.0691169376488513,0.219715841672673,
 0.9143802599756851,0.21540104232233392,0.91993376633431,0.9527200726315308,
 0.41036993589350246,0.06506142041985608,0.9317323948137506,
 0.7281427649795488,0.2512373664694675,0.31987668386431,0.8250025853901211,
 0.8299595196216123,0.15878799463098037,0.3538022315284248,0.429209521910616,
 0.07736497836105727,0.42346109253279385,0.4331051666653589,
 0.03868755767151266,0.9511442285534062,0.9862816659354646,0.6793669791753887,
 0.9017598180933546,0.3861290081105856,0.15250312999796978,0.559392992474299,
 0.4456442886486013,0.49726837479910924,0.46820549362239783,
 0.8453377116261664,0.25504673774572395,0.22201618072429963,
 0.6155724342208757,0.70767125264573,0.5824967116129264,0.0777022249495749,
 0.5095950237614469,0.5216490930647691,0.8608865993484714,0.5070095386423876,
 0.015765702627791978,0.017267431994146376,0.01066368544332974,
 0.48677425753753856,0.45865158632429126,0.5518090718491753,
 0.22002308846153862,0.036255324313160586,0.45800656948885754,
 0.2822540167524462,0.031206188304146565,0.5264277598831065,
 0.7954492097284694,0.34405283196899783,0.5363863668517758,
 0.45576964659938146,0.3649909896285122,0.1639677667585342,0.9034638679025507,
 0.7500437605176071,0.602292773780956,0.15935066922491603,0.36310088801345586,
 0.6397750112436347,0.16573177714309273,0.6115093110237161,
 0.35023006115751765,0.9790305143294226,0.6132530877194243,0.0202392720743515,
 0.6907565190566655,0.4729501162367765,0.5447306563993259,0.3761354163869757,
 0.6098741036152229,0.7077124618103198,0.40175742712870366,0.621047264503955,
 0.15553207322173046,0.1690217552854152,0.11176592031668142,
 0.5669864494809802,0.14729657774584215,0.7185204788852824,0.3306310955906159,
 0.9237339793934023,0.5827676097818434,0.616941453964189,0.4524094598100439,
 0.8953360131101903,0.22448101533230524,0.1751919351750959,0.9182781986760457,
 0.7437562420245651,0.8570378657919513,0.940108239109753,0.7889430278531239,
 0.8453894879112406,0.18824608626628703,0.4561325181590834,0.515443437678528,
 0.3121114376814713,0.2800393167288905,0.4251747102912271,0.978679949467622,
 0.618439476216184,0.36262769444984344,0.6262024794685958,0.9620938163757841,
 0.8378307079616444,0.7668579267383171,0.5233438850235707,0.30642282249582653,
 0.8530656213464904,0.9536676933936189,0.7202447935373796,0.5227420207874611,
 0.18306693796732842,0.17011432144199823,0.7185152496027317,0.22044393741264,
 0.9274898502188639,0.5406032375616383,0.7999274276449285,0.32896821901929396,
 0.6857969718705297,0.6316691599412327,0.06551551528060084,
 0.017081764903853047,0.7222707599327854,0.7362964441141441,
 0.9032722532028348,0.3511430909425052,0.9591956054185702,0.20484448634396257,
 0.20585442800728848,0.8529676550657239,0.8093506953508313,
 0.08127690999604464,0.7533902480221207,0.9511245728015009,0.695120938157693,
 0.37905836938444837,0.640982406398749,0.5097700555983636,0.7445193221287401,
 0.6197010011098244,0.30632119052032947,0.10526762635603204,
 0.15621245893246039,0.41388822854300156,0.5758542263482922,
 0.38667945109999835,0.5748467202059245,0.612328403041874,0.2877923611874206,
 0.370155352237697,0.121702087907684,0.2231526693202216,0.9311517848117175,
 0.15582519884733825,0.9672853839927383,0.4868687531418898,0.7675102431747107,
 0.6920455704884322,0.788037803383236,0.7347059940089579,0.4925961194669607,
 0.48441291293801947,0.5722508568217801,0.46571238906822954,
 0.6608772344826919,0.29878627478161823,0.5530906880454302,
 0.018062094957041253,0.8426691584831951,0.28221893770642903,
 0.4197242449334254,0.6358460397673894,0.8325215119755768,0.44461899757548484,
 0.0405249458231699,0.7571475749629032,0.22212722263142481,
 0.06965040920868582,0.3779326211410512,0.028541702564388705,
 0.9918113990092454,0.8102133201257029,0.5104822315978135,0.6508540638447355,
 0.0851880453487166,0.8885081181617075,0.28991735872937696,
 0.27440344327592037,0.3096222905908027,0.5366335574779214,
 0.44036677195210927,0.9241598556741872,0.15647227347419668,
 0.5005985915143518,0.7473687723767678,0.861360343786381,0.5452435207429229,
 0.26243324084784947,0.013200518184117382,0.8904450662486124,
 0.3691038075341231,0.5105950907068971,0.023346258929739516,
 0.029468775871631103,0.28502349183035436,0.5160542020796097,
 0.8390097187641671,0.6714936704520376,0.8782803709825369,0.9419186192273225,
 0.2485055274324538,0.9329642493809688,0.36898360634509875,0.9752649867911846,
 0.8604663472170992,0.723607537762788,0.9886709093763735,0.4781709123583706,
 0.6801334886567192,0.16496109668538905,0.5194949492761867,0.4423917531583972,
 0.47385994326802383,0.024643254883203958,0.6141462918784512,
 0.7884504145622533,0.7923274061523478,0.6658299527212309,0.6997263438386413,
 0.8854379841261647,0.24812838671657977,0.7274023155644456,0.607918013796417,
 0.202076375112213,0.9515280175784759,0.7194508045606192,0.7170027922767916,
 0.5333780159670989,0.5011248682612169,0.2171776888693644,0.8959708191103186,
 0.6844530283053022,0.24951195549951644,0.7081948449546563,0.9542662022420328,
 0.28623890505611993,0.5691341440707214,0.42400035039209616,
 0.17811503653833338,0.6892633008873434,0.6367125155747947,0.9692383561307357,
 0.7255379157785703,0.508854016731951,0.261951207045805,0.12039438728887641,
 0.10114119294399226,0.24464739923221024,0.20885722582795418,
 0.7968442913510103,0.38868347018349847,0.1055223476754249,0.3351553422705861,
 0.30077093690663403,0.04371938669259834,0.6667053498314917,
 0.8117926603008639,0.1390013812802,0.7873285726918239,0.7590866721001442,
 0.8694352094833534,0.6703022981566371,0.10548025841680708,0.8000049300004838,
 0.7139921360200011,0.7973972081066851,0.51390724949495,0.04495307357633105,
 0.19082530006591636,0.2018402329197072,0.6526583625897036,0.9029885216920199,
 0.9367395500813727,0.963649429543286,0.9382663872007049,0.6949407789280424,
 0.20703257580404677,0.11466781393186187,0.2743014359293816,
 0.6875897003595143,0.7958309266796346,0.8383854994197741,0.8485444840556968,
 0.476191420011539,0.724283659126316,0.47006809862824395,0.025337360618443494,
 0.46384511715195953,0.6247608445909273,0.09835641204841072,
 0.4982163575060723,0.9716967347619981,0.9745491005555125,0.3144340549571806,
 0.07909135876918671,0.4603462324653733,0.6394331138071718,0.7039733321657703,
 0.3324754835302026,0.5619937467852122,0.7160502103191903,0.09938241976986872,
 0.982032530693598,0.1118043502215389,0.5114325754970047,0.9427601808577368,
 0.021435067010993736,0.5590857729395171,0.6992741276647781,
 0.6401497435202507,0.09579997538793306,0.3226315248248236,
 0.13505762706937074,0.8509490112680899,0.611616890939231,0.2898035863789141,
 0.004870408351406397,0.8463766884931305,0.38850453925857753,
 0.008054505735774709,0.6630618171086715,0.3124295067610242,
 0.2839944221357036,0.0041782170448037,0.7152925546786452,0.520204004843635,
 0.32089110499327134,0.20305030383299383,0.9255937134720977,
 0.4170835501365846,0.8146767113132166,0.13465834502290086,
 0.34726390120188766,0.15932736288662863,0.6047693404655047,
 0.2599329096528581,0.5496752281795829,0.8448593537906611,0.9189673551424242,
 0.07411719520913862,0.6616506851697396,0.1589472425049192,0.9943618208633377,
 0.782538686515517,0.8809331864410646,0.012153173835959041,
 0.034799865717230705,0.18673563261550452,0.905471420876341,
 0.7251933328943034,0.35920338716671796,0.161564630877701,0.9757823973409423,
 0.9952965271894592,0.7185699770446149,0.039967352073860485,
 0.1403800452391506,0.5370343247858962,0.933049229691889,0.2302997961400648,
 0.47380842434908765,0.9871103129348757,0.07354795385002377,
 0.7086516483635351,0.7514367982226208,0.32881972017477257,0.8093494588206989,
 0.8912764437710532,0.6606902272836152,0.6692421325132712,0.9104820118360225,
 0.5160056156453044,0.5976618930693651,0.7568682357648738,0.2373751972216689,
 0.4274064216724377,0.5689654999284085,0.5798034757375963,0.21216685079343012,
 0.09588489799373512,0.27959771540679756,0.9794182337147416,
 0.08830901280117787,0.25535708188101924,0.1678829526262664,
 0.19641018803425392,0.3352801267576838,0.8294542999370255,
 0.23698639601315286,0.20313567165778812,0.19646747666707276,
 0.2627590717111937,0.19638449967565852,0.19268468663754723,0.419785769065144,
 0.3364126808999732,0.7429395957799088,0.364079335462155,0.961576614772248,
 0.6868227030675437,0.18766510849877505,0.7435322285098084,0.4792994985522421,
 0.8383659103859753,0.32143302448483346,0.4655551773406471,
 0.15226282488659548,0.06910736895896641,0.7953872601443206,
 0.26896960113067103,0.29060676437518773,0.43605456642871776,
 0.4700153609829224,0.6743565830852404,0.6163997354848932,0.5334266489437782,
 0.565509437877787,0.6948230020273429,0.6851448808660106,0.39883316019605264,
 0.59333543732956,0.729755381400723,0.27596427688145797,0.14709637127997677,
 0.8841002901497825,0.20872793914849097,0.870612190539204,0.9886805155349389,
 0.1503968291194836,0.48255607996559524,0.4637545216694945,
 0.045374725520574644,0.9395646996643277,0.8965957046110608,
 0.029215483338889126,0.19536566820814227,0.33419634843409196,
 0.43318383380806014,0.5144028433749578,0.09594929236982697,
 0.8084027117826319,0.8198152788648685,0.46863001959636796,0.7261688660091117,
 0.10634713795683526,0.5276605799701309,0.8186761388374296,0.5938272336013799,
 0.9640693157355543,0.14610383499583657,0.9401401512239518,
 0.23830137665513396,0.9232185476338555,0.6841161873448445,0.8925966436123584,
 0.49832335737673594,0.8816823676567416,0.1888373767854702,
 0.10401407232293103,0.927434414974976,0.3696092680031433,0.7241393117708856,
 0.11319496602365797,0.8244080189184866,0.6791459719959411,
 0.38455821871311635,0.23490463762229696,0.13078222715880028,
 0.03541025034229561,0.7868920353164874,0.9266215912089,0.6998692105584043,
 0.037318525780724365,0.06501936014405096,0.05812259072951442,
 0.15363923627828668,0.6411040094101642,0.42256001053437275,
 0.7365180225531824,0.3634934576893414,0.7408293671609076,0.8614012050780577,
 0.26933118085916996,0.40582887734986883,0.8817855405820096,
 0.8302945220802579,0.26417465090128744,0.3849879156666688,
 0.055471384558539505,0.658467352925174,0.7660692045994243,0.4122694973488553,
 0.7829679775850713,0.8274491740520199,0.7654067420728113,0.20719809716066306,
 0.33676869521529273,0.04702632962954956,0.5516511594708289,
 0.19153269758028402,0.7644317831585906,0.0034917022280754884,
 0.5915254618173521,0.2666475080520898,0.6276777454513947,0.834249178505189,
 0.07556853103164729,0.3972474422687421,0.7177864044601893,0.6582212653519764,
 0.49308872686704763,0.029205668861161604,0.612016437235726,
 0.10286294601474166,0.35294495189023567,0.7498091722712673,
 0.7833126717109865,0.6846619666458527,0.29772437410681185,0.7561009346502525,
 0.2856242270685958,0.16310347809516512,0.2646811153110682,
 0.15313852801728234,0.6830225779716951,0.5307872540870119,0.7273806798038184,
 0.08796374935863294,0.04699027232673236,0.3748654795143117,
 0.9779193859429128,0.6644384065723378,0.6501021146767493,0.8286437789777626,
 0.30412564542860854,0.328974779479972,0.9296571302948134,0.6595887322987456,
 0.7971300554633707,0.17434750051776193,0.03500162182369926,
 0.6638750837966185,0.07550918509902493,0.9993609171522759,
 1.2592795334087636e-4,0.6189257322234005,0.12320712315462545,
 0.02948189641132759,0.47824283223312347,0.08583362627175817,
 0.933322302128892,0.9971875734997554,0.5198560555505946,0.3435695118430425,
 0.5904559698500862,0.8370476616458296,0.8218507185290572,0.1049967321937797,
 0.4600571075178046,0.8059681394044684,0.03990985437395489,0.6511612830409383,
 0.5835905279067748,0.5600590615804102,0.4814144199736696,0.5910466456157574,
 0.3617324146851344,0.16509489425326818,0.6166764739605657,0.8066775249583102,
 0.5794478515316491,0.5248584080364708,0.4717133508609468,0.9114542835716387,
 0.1913016428066715,0.5710538886883872,0.6702037825036764,0.07928704989313773,
 0.7325570466393142,0.5922915610514623,0.4337592176519369,0.8522473968088129,
 0.05082150224087578,0.8853222984790836,0.815106810230006,0.29984122438150396,
 0.4867627223683947,0.7158800116006724,0.9579838442310206,0.5728921299064423,
 0.8755132679964212,0.5372567835400659,0.973909066147928,0.6974599287598927,
 0.42272630008574774,0.1203966284426764,0.9664859943636014,0.2009628044694911,
 0.7717257240985965,0.8010404278021447,0.7769518404481192,0.31009707104144235,
 0.5103877806411232,0.3174771289178582,0.4736692353309391,0.2772014461918402,
 0.11477496717224023,0.9992811354654197,0.7043668704877937,0.6497903894077672,
 0.09329990138473065,0.6494171161365788,0.12791408687076133,
 0.47933273895145767,0.38041905131252474,0.5787680729104512,
 0.9824088761133569,0.7544003698118958,0.23604375647687925,
 0.25563803599030877,0.3611554378889892,0.4886904431143385,
 0.025155424736103216,0.7927834645969718,0.7490203816892299,
 0.7400636060604899,0.7209411384148176,0.783755355888736,0.868359634048957,
 0.9739394970914654,0.8793033288819704,0.7604144959001393,0.977367484386687,
 0.36449567073655686,0.38127484599586703,0.19500733470381337,
 0.7560631365536948,0.815292535693569,0.5584205542547391,0.07897329957735533,
 0.336085620611535,0.9102564092314154,0.6958229197748659,0.09869949576018722,
 0.5324836283509781,0.2718541663202426,0.34618578970430824,0.4677813643457085,
 0.3485911780088139,0.17236100046736036,0.4029508721444477,
 0.08643925604129288,0.35817521414811404,0.3713774144081964,
 0.009536511235551215,0.5745498021052229,0.41608004512270313,
 0.6635899531656603,0.4674998178436375,0.7342791776657112,0.2901197513127579,
 0.6800906630708909,0.27933765872152017,0.914181086795735,0.8908441128298962,
 0.7684160210712508,0.7139587327847624,0.7204619493050856,0.9553954205573184,
 0.9112004612024026,0.9854488071960065,0.6834843583777418,0.5167211000308913,
 0.33609302041623745,0.38398049905598286,0.10080407172656758,
 0.9915807019970018,0.46302425516201284,0.3315245769380599,0.7133511266952128,
 0.23582527225195682,0.07620096177345936,0.01456453922946066,
 0.3995213947204641,0.01852142439671245,0.2366084444447395,
 0.33791928541982785,0.6123730038694956,0.7746652735386352,0.570021264677853,
 0.5988073741279933,0.8756011385489302,0.3093265102206161,0.2679233973487768,
 0.07500256019610707,0.874841627640687,0.03446263225716639,0.5965024015289743,
 0.45177345377820277,0.9913065919502743,0.2919675667921253,0.4574640306377158,
 0.9779279596602,0.6388916611158136,0.5663628342121585,0.4373084504727974,
 0.4153121087801821,0.8090476130579932,0.23461569694781248,
 0.15879241012175005,0.9502654424199909,0.3937736017744169,0.8188567318626624,
 0.7531730596735386,0.7461661538000568,0.8865258614056033,0.7819347905022366,
 0.8793843270465991,0.6043257711616028,0.9553227470706696,
 0.006663979065997694,0.7119814671658675,0.37196468968630914,
 0.3723188505481596,0.03848323168465184,0.2656431512212475,0.7999111523838328,
 0.7464264808808769,0.715121035300338,0.12006535491967585,0.1122120717030346,
 0.20249125625810593,0.17777689799384533,0.6393996707161725,
 0.23945870084885412,0.7773975643290589,0.971203322290339,0.9135381385519823,
 0.12466532355581783,0.1350237575271862,0.09064084825939289,
 0.3422020926312489,0.04804225592719824,0.43225911816690077,
 0.23959820741958038,0.5049872555886986,0.12963909087795011,
 0.8775560001972961,0.08496519804328906,0.8437524975113188,
 0.24152731354326729,0.9382501958686505,0.7947452586005805,0.7480029981452804,
 0.34208408239854216,0.13569025640628163,0.34870618955208965,
 0.6055819415662329,0.4110257856625523,0.5276693118133603,0.7908134299570417,
 0.7144226484199304,0.08270258536347552,0.755242721713969,0.6959920712550552,
 0.8797246094530811,0.5912293243025759,0.4799168842636278,0.23819926346520726,
 0.8605251002127613,0.6835124464664483,0.2564041460117079,0.18052023858509703,
 0.015240461651113257,0.6507343627107631,0.11796683947885189,
 0.1734927848944925,0.5691455314426188,0.9586976785821235,0.9131750387389377,
 0.946492528081403,0.8565056100744254,0.5718978956438409,0.6799547048521983,
 0.013749023216128142,0.34745417339491047,0.2565076917071296,
 0.3611321219811561,0.9809527977063035,0.9740463566807016,0.18339892873339558,
 0.4037202156983608,0.6284629476160877,0.4517054617201265,0.761274470637227,
 0.5049420080980399,0.3927354039560689,0.8768663422018466,0.6039085816272218,
 0.8775681350221489,0.14573929629517157,0.7942305228050327,0.5577841413785725,
 0.5662715046207276,0.6011987763606688,0.5380854698189171,0.8335711161252538,
 0.4291181717587349,0.09078686301725636,0.552316266906766,0.4030423034066961,
 0.6385168628150746,0.9272428662549634,0.9177864459069688,0.05835239665869607,
 0.7619084351638313,0.17783620958655033,0.760673775041359,0.06760544743046504,
 0.03308634446427172,0.7785777863907289,0.20381652432635466,
 0.9658754318411762,0.5423456115493576,0.23933196138335344,0.7816859058686186,
 0.34798457443652064,0.22673825043135376,0.22176806066964183,
 0.2468719205923673,0.1856105268042958,0.29472150262649244,0.3658344829613148,
 0.4636035641799304,0.46125618043499417,0.5263141739840473,
 0.43874927812001496,0.45255089299277973,0.6707419582377926,
 0.8134196503333585,0.05778378355926139,0.4082926510538003,0.3277356067004248,
 0.34913984933461195,0.18376616230618958,0.5770313573904806,
 3.7291780323989165e-4,0.2260782086590798,0.039748550247829084,
 0.2499391282052812,0.9321837626604793,0.7076022812209686,0.7586350517721238,
 0.1412307276665914,0.6799107977885349,0.90286271481606,0.8003337569039968,
 0.859452871404701,0.07822770497477238,0.3416427175290443,0.4307795908688937,
 0.10332606483547102,0.6481694635894527,0.15925421064949763,
 0.4943873386338097,0.7441268162101479,0.9583536455312356,0.3233880004065316,
 0.578842719347351,0.35003407021313937,0.712934348736248,0.11227393763154103,
 0.5982288017726411,0.04490741579460822,0.7975161026931554,
 0.08278398380660612,0.1490997322968367,0.692887241692422,0.9888005136309349,
 0.5423242361621072,0.6143464318843149,0.7086818606729484,0.5000617279813264,
 0.17883876104580532,0.47613292547617847,0.7658968790715263,
 0.27669643115385056,0.7229329308979401,0.011180381576951448,
 0.48401015178965057,0.568082757960282,0.6640246174063577,0.8253237130648261,
 0.44442769854284747,0.41857370017697826,0.6056523196706173,
 0.6842909240363192,0.14385357544265598,0.9575388536385614,0.5960464604957925,
 0.17594519550877474,0.8252411267530104,0.3605986229378114,
 0.057604934764233784,0.2958750646267716,0.01751036742714729,
 0.4632326248037667,0.5982083296333194,0.2139397023308055,0.7014553804986894,
 0.1125421277605625,0.46442432843603865,0.49783931582153906,
 0.8070677744855499,0.09717614440374533,0.1948760327507577,
 0.18843680692092868,0.31267565955240106,0.6706015306632357,
 0.8977325199942932,0.5297960687330593,0.23450783359519722,
 0.35417330433532446,0.08215541893938927,0.9887687445924029,
 0.7636995299662427,0.7592324187624944,0.9593914020528578,0.22030882874212177,
 0.3114315833517456,0.030846944247937103,0.7916631094960499,
 0.7575776329179278,0.2828192924506806,0.6447318015408463,0.3592275888259847,
 0.45635753708295046,0.047298216642728375,0.24453325214599886,
 0.9676282410455765,0.5977222934080803,0.06426502514955978,
 0.35761720991316903,0.08857001069168768,0.4793639156885212,
 0.4324678857235633,0.615034023377726,0.8279089525478422,0.29847279080338684,
 0.8073454938832954,0.4770448821244345,0.5960817188549982,0.9396095394031295,
 0.31993862364058445,0.8126660773592406,0.010051485469254384,
 0.1907795126653361,0.5072374757290029,0.7324715545975422,0.0930979229961344,
 0.825845638530453,0.1813196297298435,0.07671035304574914,0.43221417353719904,
 0.23213859279216376,0.5868813729636293,0.7688040054534032,0.2574696063201205,
 0.5485246438337851,0.7053125214996987,0.7785172504721877,0.41229379110356756,
 0.3517553701671676,0.28614956661594815,0.46934916978501373,
 0.19711413555266644,0.5968860686163449,0.5478045375213854,
 0.11442015446941467,0.19682400382902954,0.8658636238757002,
 0.9268417246103546,0.4745683254936811,0.8425176156963489,0.5603627013244772,
 0.09695553988041827,0.7358874994142468,0.9342257968823257,
 0.20329905289799655,0.49891646190567296,0.05215040928554626,
 0.30130185436876655,0.24804318192886032,0.6593087018636813,
 0.49794203838574447,0.5560429623650923,0.36919202739652734,
 0.07032552746825194,0.9247855464815232,0.7409692039621972,0.6862838798514197,
 0.8609674362864013,0.5505388514769894,0.42794697542331517,0.2635635113499736,
 0.9821005894866948,0.47237204120333565,0.6590341758296735,0.5476769941799183,
 0.18733288292272876,0.786440397995715,0.5442229275875528,0.6536386480436078,
 0.5919184573183329,0.9704182839928784,0.09145879843177473,
 0.37879272352205295,0.42348290784093123,0.9007013888231044,
 0.3239777592466633,0.05575558712663997,0.5553671805287006,
 0.13778884013745663,0.6326073460602177,0.37620632983528424,
 0.3407429360129506,0.7678689112106846,0.038584784293085095,
 0.13770491739853652,0.9298574667641659,0.08493220806684854,
 0.4088428983594148,0.5761313588021688,0.14122752218550694,0.7002876974801509,
 0.6217865498455271,0.9956882214502678,0.5027667230315461,0.6169806389225896,
 0.8035394917767894,0.3844410646737645,0.36036854556159237,
 0.013721423923828024,0.8683489438188587,0.24789143119770096,
 0.4356970765273832,0.5929095910197112,0.8243983807445375,0.365940385613602,
 0.2932736206932772,0.26623774127604327,0.7528163925887225,
 0.22643664799966923,0.7355577396001687,0.6281998533373974,
 0.15168411358379275,0.6115087989185573,0.5207474410682575,0.2698699588421404,
 0.5900550538307139,0.4102656249808342,0.7421161629938341,0.19632103587989258,
 0.7722499819539337,0.8672956446268532,0.7261288279354202,0.19225521914622978,
 0.8891394999645132,0.8740127984823769,0.05820050069416827,0.7065233436666699,
 0.7306399814318384,0.16309439138141546,0.6425491472380657,0.777781658708152,
 0.8207711540499145,0.8172442095936419,0.4014087496213644,0.5312032999192653,
 0.6856854602505645,0.8817193220281674,0.8464910657731942,0.7848245592954415,
 0.02265799892218434,0.7419006142769169,0.941026216474737,0.7642678790471625,
 0.31559102506834624,0.09716388052226566,0.5986560932576087,
 0.42387863098546896,0.21950722108941256,0.11596913767256134,
 0.8486707974561085,0.781147963815722,0.21552617816494457,0.2660674908046696,
 0.39072858476740135,0.9756376308686895,0.8059424084001847,0.1576789348627916,
 0.9630799435857159,0.07819609396867189,0.9959302151174383,
 0.12698320166030774,0.3439892661267451,0.13440635990528083,
 0.16814588616664006,0.4726319450645057,0.8000759343718771,
 0.25010921595852054,0.8701462180857378,0.6570900781531763,
 0.015869230179615612,0.7514242350403865,0.6272556085477475,
 0.22705837257632233,0.5426439846077302,0.3740605676155264,0.8081755195068522,
 0.3633185237934611,0.537858566773183,0.85816081176906,0.8783606643006713,
 0.1554310668266492,0.9334596858746839,0.4528165812631304,0.5584139778166034,
 0.28159961294160896,0.2388872458786072,0.9036523689885421,0.8143017008689183,
 0.03507873517529936,0.47325519710471,0.5015810103935122,0.9682629244828367,
 0.4180955644489911,0.06638520485813904,0.9346432544988256,
 0.049461231797255034,0.7478340274257801,0.2561267503353364,
 0.3063485327821944,0.3197047410994409,0.7050586220809077,0.9354938804537196,
 0.6786805886222328,0.11282588330221377,0.6587444369710967,0.9464413929399017,
 0.6585847911979076,0.21298651209733888,0.4781073050817539,0.799054807908414,
 0.42735225367637963,0.9004576902207229,0.3002910837203153,
 0.11913114320057372,0.9134459826956485,0.49012571601117205,
 0.23507453355138375,0.06632002594420061,0.3738648227754676,
 0.07745647199181604,0.8845123193592234,0.401102315410075,0.8170308899467313,
 0.34887739237630844,0.34114124540728796,0.22024150910920037,
 0.4270081830123693,0.9497094096800378,0.6522259782022963,0.9476857445777672,
 0.7083619763272507,0.43164310410561335,0.8627077002399942,0.1914993899712194,
 0.5030469060219552,0.9752208348939181,0.16555042505781048,0.6535654341057929,
 0.9656918829125772,0.14346928899121525,0.6201447517568459,
 0.04692785765753582,0.6180053882891002,0.7547936732388595,0.6834292038133498,
 0.8472076014960436,0.17748988288169432,0.18595881590428187,
 0.1598075084729924,0.7607261695250817,0.8995161646722359,0.5555069582165262,
 0.6987152582098918,0.7335504611510166,0.9244736197615113,0.8534239640456124,
 0.06157270077444521,0.2616895771353269,0.522795552044389,0.06592611523704228,
 0.4866661834693905,0.8701786893354465,0.7273404877613482,0.5776006540761318,
 0.2610657981405672,0.07257641158086992,0.3254148346407981,0.7693201442877214,
 0.5564230101945677,0.8873549227325587,0.9433580429836252,0.7790371250310654,
 0.4725615392032917,0.25862191222313446,0.9286713644818614,
 0.31562895913488387,0.5705777320322749,0.4688651152219374,0.4039144264811547,
 0.924978803311796,0.5470266963820601,0.20993855556080143,0.33116688267074634,
 0.5121502764148678,0.38801583961206154,0.6565610589949401,0.6836219851818176,
 0.5687462057808592,0.42750892502941296,0.37660520660776753,
 0.4308406852684381,0.8447734763872324,0.35419097870142136,0.4694675522592574,
 0.24291474348430708,0.34295240322496445,0.10573429354144503,
 0.7599229608972291,0.9209836520283952,0.7606362933501281,0.3846827132326084,
 0.035417890303983235,0.8101685192082247,0.5643878281592905,
 0.9291717409519691,0.3775739296745313,0.4661909144102463,0.666850563031289,
 0.728040809858937,0.7268305356781514,0.18651446851961584,0.8035402097436695,
 0.6048622813005994,0.8052574791692865,0.06119356512646812,0.5911609804738491,
 0.23630466387780724,0.5024666424042178,0.2201268697212142,0.8852293309335296,
 0.39155299407905975,0.0677985337898559,0.22749028376407832,
 0.6911781987181567,0.5717479254088181,0.15310638593305503,0.2373499769956583,
 0.0856504916221289,0.9221398364131151,0.6144497552666865,0.9647673579028601,
 0.629790174437131,0.4995040798820076,0.316798765885527,0.7966943931817645,
 0.29773616239147116,0.1215893985728469,0.7342182247201701,
 0.16696847655121905,0.681989910999395,0.9780368841978391,0.9206152852848015,
 0.8500601190720696,0.11450857657699942,0.8932838006502707,
 0.040783468351048624,0.2936100505332192,0.7229333229944077,
 0.1411232997779135,0.16491984924313652,0.8464920841009189,
 0.16304686509991528,0.3053169410861365,0.7034187023534985,
 0.01411124242648587,0.687348556637628,0.4602826093685273,
 0.027286113409873636,0.4080035907245819,0.4511580855842716,
 0.5192900587162893,0.8446505822460798,0.20116381477452427,0.6609170163825745,
 0.8697689288405627,0.8092758641386675,0.23953917387900037,0.2613934321743415,
 0.0748836896888786,0.20987378354749164,0.5254338785493542,0.6022075913330502,
 0.560119950480466,0.5350179089633897,0.8917222328347463,0.6878892031319538,
 0.32235773383491406,0.041005741950052355,0.9488037503838508,
 0.8402495581183339,0.901068193917268,0.28481406776466933,0.9500432706131243,
 0.22702951063282406,0.6682040292628576,0.09152802268724602,
 0.5945870120194279,0.0155577178232027,0.5884887693768142,0.684944975988864,
 0.9758653573503318,0.07441383059770201,0.25019395095941876,
 0.6528685561727546,0.5019397850454461,0.10255299850673849,
 0.40433210726025837,0.885565016008706,0.7018370179583627,0.2511761760194877,
 0.1902494955168602,0.7005361289454743,0.3056090198314324,0.8534376903789659,
 0.9048446640970279,0.3291591953611206,0.5465748160256793,0.3109934765390745,
 0.8624323297908999,0.41049578693576094,0.3314844690370884,0.8820354464884652,
 0.47073481520203087,0.8788495758624323,0.18164128331581175,
 0.3421144083392539,0.01233708342994777,0.20888688418256873,
 0.3755062932254004,0.37945480958508204,0.2825248310580287,0.3375257189023634,
 0.6555247856847849,0.07731980250618786,0.9895323771027905,
 0.14488983520119347,0.8071601772824715,0.16838640733381416,
 0.6008258036994565,0.9108612797528208,0.21991286910814645,0.806505889712913,
 0.3569170055119253,0.631969227765086,0.02604588247457984,0.2377571593863882,
 0.6980590970621329,0.6492924960225321,0.6483831001096846,0.7942273823214432,
 0.02067392187309025,0.2341630342852158,0.611499244534764,0.5106986736085118,
 0.2364682769967965,0.9343719001427909,0.29733228633639286,
 0.07625361694947419,0.5687787659044025,0.8207731528337754,
 0.46693668387923104,0.5608441608776673,0.10569701259957709,
 0.5411181864287843,0.6568438527137328,0.6016784117065027,0.5303348133643927,
 0.3121267148568546,0.7352198361627491,0.016003503361139537,
 0.7654434471144361,0.9867545621666138,0.7855032377686609,0.6558822985097772,
 0.7831942883264442,0.9223100352817853,0.43490202205505324,
 0.16996567084725778,0.985577551297801,0.6019618395444765,0.1747039378002102,
 0.5632191850619198,0.016226821009777237,0.2962732879869454,
 0.002240441898181622,0.7189922399986391,0.4644086148958906,
 0.8103561298987186,0.42373326370930997,0.2278695742792115,0.7953581376437507,
 0.01368600850649282,0.6693526598708665,0.7117959159477963,0.312035743962887,
 0.4411300543534131,0.1190516480136492,0.07598079704211735,0.4086937773661007,
 0.5400463073853284,0.3449930697804455,0.7712736414356474,0.38881508734477466,
 0.3394439605256264,0.18710818297492804,0.3382028237963659,0.7623986808514203,
 0.015253975868271374,0.7804536882939563,0.24342168430507882,
 0.5214609762221863,0.006879150243739196,0.6292618756180947,
 0.9911832349228731,0.3295173041440673,0.8381943862408835,0.5139875223861797,
 0.9208244169035057,0.9438943273425182,0.544904074830443,0.44359653300043966,
 0.28109990902204296,0.7017406026974542,0.766437725421678,0.9042511382132539,
 0.1577134262581431,0.41176929251386707,0.28599416890186924,
 0.22523302505887988,0.5551075147154921,0.8806963462845142,0.0578494345088143,
 0.32786280715957616,0.6214274231705251,0.6545897987056425,0.7649104173078807,
 0.03632302896153916,0.21776999538554054,0.5955482757271873,
 0.5249661396152618,0.47460899747995666,0.9297804091296149,0.9636525982317111,
 0.4673300603335473,0.219892349445443,0.8419154897827932,0.8527586128154233,
 0.024434689248700292,0.13613654403198683,0.23942308914810462,
 0.597144547390062,0.07943164895599542,0.6726052530317417,0.5318487137119225,
 0.4826628561146179,0.7773312199521205,0.5746367718240778,0.11093881578356557,
 0.10951148190886517,0.41653010280586533,0.9179216709204591,
 0.3531718214689825,0.250850150112804,0.680682233781448,0.9339938482829003,
 0.6493311324460899,0.7977291628125707,0.27023706296484606,0.9165978097235287,
 0.09999179329277597,0.3209884079254268,0.8100110594910349,0.9108572351405486,
 0.36396592917553205,0.24704368236989205,0.3947566120605773,
 0.43695254802112504,0.20950493169119122,0.27941166880954416,
 0.7684613175606559,0.42293581913528877,0.5475454242862934,0.282864184560365,
 0.8992641197116578,0.6695154667284465,0.47684691065797546,
 0.016472717476866405,0.579480906373801,0.23051176086902547,
 0.13749102721376705,0.7339593594213956,0.18485715942777425,
 0.6663355315270317,0.13493070354772962,0.23645809071023505,
 0.8432120227976638,0.45008699320865475,0.6704096125589397,0.9815621697828966,
 0.4817109424172291,0.6938865143029577,0.9177359922129198,0.6906406826303022,
 0.5084888134974427,0.035140662905520426,0.7429657604038916,
 0.1720314938849441,0.12782993970372725,0.5889160829993904,0.1334668822403624,
 0.30185529753700324,0.4987298524716586,0.4374823147382053,0.8247314637888944,
 0.46951099587240797,0.6194779986917811,0.7288794403781093,0.5897631511897199,
 0.42619946712855317,0.48494525030669067,0.7909771161827122,
 0.7694949901475734,0.2477025440504319,0.47138798577285457,
 0.046882230423492266,0.9599519815119169,0.16793560627759407,
 0.019793000273494288,0.1791882358653646,0.2711693775006798,0.967213452847661,
 0.01977966038947465,0.871896137837126,0.873060637795335,0.12584159218680102,
 0.009798871265799036,0.5676884312844047,0.8636141539322367,
 0.14444130388656717,0.1961567678015418,0.6836834971538241,0.3398103127014509,
 0.6276406091748706,0.3041543046648575,0.2202565572099796,0.8413420013387884,
 0.8538331616255652,0.9430885110176211,0.577146525890382,0.6762232282009862,
 0.8290503287543137,0.35519060302519334,0.5402439083438907,0.9732621623938545,
 0.46442134584656025,0.730011287842193,0.6128274342343134,0.9091488142991248,
 0.4729332498420806,0.15238137840053956,0.7940792137525308,0.8092346462156846,
 0.5049809086239795,0.46067645127728385,0.8252940571261876,0.6527680319069304,
 0.716568116868039,0.39092834260149756,0.8734514932713617,0.7420672668659022,
 0.6365065668509727,0.25013427867842286,0.7752406963930438,0.6001967903765352,
 0.9851636799408926,0.6849460123670881,0.7489472231208872,0.06679761313848431,
 0.43744074786390086,0.6523722747482538,0.5757992974662021,
 0.11828328087577789,0.4545038941281664,0.4051291663089427,
 0.15801977577661286,0.06833851648266043,0.8896744849797085,
 0.5383868754551772,0.8637040719152346,0.6584347378812427,0.9914130387679431,
 0.8384452745431885,0.7624696231764484,0.0645978635790786,0.8077646767464123,
 0.20780155355492613,0.22917005874118546,0.8505150787240764,
 0.1877170524218723,0.6059727218477307,0.7382635431619279,0.4429710145667656,
 0.2218311433533786,0.4304340424023133,0.08567104604777476,0.7603064492317652,
 0.3403732118305677,0.8614016646123922,0.3762996091965498,0.5529961483135835,
 0.2455194770579403,0.3630799444556224,0.5427877779028241,0.20825856960940436,
 0.40249048501351137,0.3490795180277013,0.8285731569321952,0.5148458500183355,
 0.2601675810847601,0.8619568166130495,0.3394783515433716,0.8369541337706893,
 0.1345869102522701,0.4079686412565965,0.5493480028431083,0.6732893535372436,
 0.9949896203999493,0.010401660425962023,0.13728115230180327,
 0.4600574605236992,0.7569682698393665,0.8352030547016066,0.2977378417230184,
 0.01451564927710658,0.805627906392432,0.9661106934180395,0.18779244406375228,
 0.6349341075096852,0.23736100035306407,0.5385776543111802,0.734864375674104,
 0.2381848406771846,0.4072611068360801,0.5168301972675197,0.7823955390710355,
 0.1813322679467152,0.5127515092837398,0.08394503442319201,0.3021135114542718,
 0.7615572716370883,0.009167980265831321,0.40211872618549127,
 0.7236540846511649,0.12904524408959406,0.8011396899484904,0.7674942736242087,
 0.9696479769233156,0.8121919536395249,0.32223127958435727,
 0.005817472761128539,0.04565144298244583,0.7696479850875309,
 0.3882554296662101,0.08187060916340494,0.4164663485176532,0.9317418963826216,
 0.2065845556020044,0.13225896126965853,0.862233311621276,0.8042311106711709,
 0.5697521949667963,0.04273696427647211,0.4802551903910066,0.6752741564896811,
 0.9731520264446327,0.07094743493937072,0.6498674742204349,0.7553188739585152,
 0.7822422407856056,0.8840487547273158,0.423542223282252,0.16421523244795233,
 0.9711471843551299,0.41605599724653375,0.2865721814495551,0.7041253385706668,
 0.9960132505696436,0.41895393007153015,0.639958476110391,0.7407690212319605,
 0.44212103877349307,0.8840631777579233,0.6072796023787778,
 0.36833486926264425,0.20398523913024946,0.924373606195118,0.7422856948765094,
 0.09946857917236662,0.6422685719271635,0.6204886494489327,
 0.04555807056191119,0.7743620565690036,0.2842078473793199,0.666546653162983,
 0.9034645411397857,0.21440961366092126,0.27302346047142634,
 0.37226229008851897,0.48874561351335344,0.10782011319975415,
 0.6752952852277164,0.7794695967124157,0.13689119534994754,0.8542052108000767,
 0.22046173675526326,0.18402983935706962,0.8392906442096121,
 0.1579611873594562,0.6759431558506714,0.049812971832544406,0.969213593273735,
 0.01859718759659934,0.4241669970564027,0.7486387914704968,0.3767444925590675,
 0.3221385110711258,0.06366601950360129,0.7143371369177076,0.8918121725089522,
 0.09829002624895622,0.2700892091066447,0.2464517650758501,0.7120552349297292,
 0.7777421348968013,0.6251430261461316,0.8665073818423524,0.5677218534248664,
 0.9381734897751604,0.9942636868177176,0.7717914752103393,0.8568569627297595,
 0.8766354554636124,0.06860601443393466,0.48558357870919144,
 0.7463276930925123,0.6372055264126384,0.6987921504918477,0.10357237925946716,
 0.9718964459811524,0.1505797724730289,0.8633249794581644,0.8459889931215753,
 0.9053147713509312,0.9713941334851367,0.6090710874882975,0.7823754655585802,
 0.45256859824520645,0.4205679568888784,0.6444068113872938,0.6532201410589278,
 0.5287118324744147,0.9639024811566438,0.8720548150459855,0.16244938798921027,
 0.9550503500445323,0.6421558216391673,0.9116297914734599,0.08366010110092248,
 0.18143634262617692,0.058840443332679015,0.5185283519343875,
 0.20540084309932882,0.17974296396566358,0.4071286734545997,
 0.27332057891717787,0.8062160540358256,0.780885063257026,0.8260760006157426,
 0.21061636682281537,0.6994244263635012,0.03242670693524374,
 0.7384898741289899,0.25760862689087805,0.5016426446022998,0.9298908514387693,
 0.30311978159499964,0.28638068903430125,0.1399755317490432,
 0.6976452008736549,0.031069171486068914,0.6259546252687135,
 0.33939504703905543,0.6019315829811424,0.2622384542842091,0.2699450439910036,
 0.634440780868407,0.035900685478411964,0.35478735547137097,
 0.19836858091608367,0.8466816571162634,0.1504794484402755,0.7866024608782601,
 0.5636571104409549,0.8855422848683346,0.2549541857703248,0.32167030734341606,
 0.5208598009343394,0.987915759245497,0.7351244064342396,0.1574866504891198,
 0.3634397841234359,0.8765332948010071,0.2642544299782038,0.5532279656184463,
 0.41271934374339814,0.9970540483694061,0.15176487567858876,
 0.29560589591519615,0.8405127375115011,0.5676910915329055,0.2537905506759093,
 0.13012387925505575,0.8396461418799301,0.6530833652041317,0.8937545942913738,
 0.679305867067708,0.6350452759704089,0.4366803027202081,0.2349090165181965,
 0.6587785745077689,0.5435464099935097,0.927173680023803,0.6059975046947976,
 0.29361524630471325,0.006459164210833146,0.6707231137782633,
 0.3467524114412567,0.3943921754944487,0.5368146439907125,0.6589204306532345,
 0.39222705342251296,0.5760569602670517,0.9400079605745665,0.9740010998145477,
 0.0408990899017081,0.2717662638831241,0.013057213338181661,
 0.31186291538841315,0.20041815923359374,0.36367826718748364,
 0.3395712059255285,0.03694602562418481,0.1600769446407856,0.7798871281567993,
 0.5812921866271189,0.54799955822423,0.7696725633973064,0.6839119222917974,
 0.6530203096867515,0.9415454894696467,0.9331784210597504,0.6299504322996965,
 0.1265486755306492,0.6578631749608541,0.38274410459841235,
 0.20923215498212056,0.5982906988053572,0.4837416459219488,0.4168468335259594,
 0.41633083976354257,0.17037851956654582,0.9539449230808368,
 0.8242084637130943,0.08567084425693239,0.9772565235463495,
 0.31479410374208583,0.15531536647016952,0.8293763859187424,
 0.5744802624651978,0.26609883097463516,0.2234681179468686,0.477561676598656,
 0.46326904539655,0.9715549635876206,0.6384739374792135,0.5077147295281366,
 0.20613764773006227,0.15971018494947375,0.8159178629299346,
 0.6063991492170003,0.703381115120659,0.4258806296588499,0.6004210954763893,
 0.19726602052926667,0.8504013535886434,0.12917906669391033,
 0.7520621415830654,0.3423917643829656,0.19189125638130689,0.5016561231023684,
 0.9101776073004842,0.630397871780094,0.35959389386473084,0.03312914797898392,
 0.7376202366934367,0.3346040144788911,0.520137720192327,0.5982583964483627,
 0.2938785578060752,0.8230974414621142,0.7396045491115442,0.4361458444063617,
 0.9431743581460772,0.40635693691989716,0.4953624833218473,0.6120603063565635,
 0.4036790753993099,0.967909089707211,0.3686964059305301,0.6676090791533831,
 0.8937910413744109,0.4956526599637544,0.5935281938411819,0.1149970659380063,
 0.7215106508283797,0.3726865315866441,0.8608887637985121,0.9295847061656137,
 0.9612609120063254,0.6170570891165179,0.03280429890882974,
 0.010501709256969027,0.2557066028596753,0.27630386469273127,
 0.6221498458092191,0.5994181022956944,0.9149710126425333,0.16957823102959146,
 0.976883786814171,0.5773495706135119,0.8668801134457209,0.21715967314041085,
 0.7556911709338859,0.9156603606964936,0.29664769192311413,0.8444799082815113,
 0.48507563592168057,0.7608699819484236,0.5801958092674206,0.194991365975706,
 0.9062197146394272,0.044845332932964554,0.5952988046701233,
 0.9892189079548791,0.5094418818643942,0.15012024826867054,0.3596086876651048,
 0.39197690146665654,0.44131490797314354,0.1939548064633364,
 0.4233955392266444,0.42390756014570086,0.02483546285352156,0.414442987925313,
 0.4412062608786198,0.44436190328370906,0.3439247917449303,
 0.005507897338572376,0.08158801549030104,0.21405375730676968,
 0.4960380202896417,0.6509193616276218,0.29788438932969474,
 0.10162829407818774,0.8450682697821574,0.8353601703049909,0.3861883200592524,
 0.4915554820022836,0.24917621065602935,0.6244918076774392,
 0.05812163519868885,0.8029711036977761,0.46272839895390716,
 0.9989618567814604,0.6049814364593709,0.10584089860600865,
 0.29084832336208877,0.9876159163537094,0.3377141166269102,0.5104243560985523,
 0.83719961565861,0.4813825179957283,0.6427605751217305,0.2606819557913532,
 0.6365322466791785,0.8379000530404355,0.965228671941965,0.05063295839182769,
 0.7610697940426858,0.1768309851810359,0.7943355288528562,0.14490681299738128,
 0.49000641803941924,0.1091973156935736,0.3097120077214478,0.3704275933911071,
 0.324799732417303,0.38734361435676123,0.6362082531288583,0.6741826151124022,
 0.5007626060748742,0.685219786584946,0.11722600040112907,0.8673067061642526,
 0.8685277093120657,0.6834519202519177,0.4189178274233678,0.03392606983329738,
 0.6003846843595501,0.24305971783502422,0.7777392464681587,0.8989811176891181,
 0.6302416198776505,0.6633195073987525,0.06083104046028487,0.7804488205394577,
 0.3861823015823944,0.10821695172658152,0.26418249000331806,
 0.08081724699422421,0.4464037659719722,0.08000487179092253,
 0.9764967515587032,0.7572141730238416,0.05215977917714831,0.9899061341145542,
 0.832688558349698,0.01229465107789518,0.5750141368543397,0.7873086154678526,
 0.5416703715824749,0.06204878095977795,0.05446976288323502,
 0.6743050026060585,0.8162335529115194,0.3446424724449053,0.8598482678335141,
 0.29757963980969704,0.9194711086392956,0.022965394485271462,
 0.5427195262145079,0.08837863941387936,0.5648735508170581,0.8793773659888244,
 0.5964217069071258,0.8778189448666665,0.8138500184253881,0.2838641207978223,
 0.6588477713347064,0.9304187664072465,0.11030322845487983,0.9413213415906903,
 0.9968519656718873,0.09309299332910481,0.7679262919130945,0.3975989902736409,
 0.5335647714434194,0.8492167680674885,0.23199523919179876,
 0.46763581368828233,0.8231358566747505,0.011394496226812656,
 0.19184146976614513,0.9997555315761388,0.17595096699649038,
 0.10356302190858846,0.33588982807980516,0.7426113641030839,
 0.6842537820396823,0.15844895745382193,0.7852400217714378,0.3665169230804147,
 0.4759942114467225,0.36603267781144777,0.26475567817105583,
 0.3280408945892509,0.8811920489112386,0.8856602899105993,0.4373485862003601,
 0.6983458337808578,0.8417686962089083,0.7153846023924372,
 0.035860738424351224,0.9947326116777987,0.756260762892117,0.5730895928503286,
 0.18852708792231854,0.5512432683569253,0.7136966416622873,0.5701045208146575,
 0.878869691857162,0.14726556279627823,0.7829803254445518,0.03690470160986459,
 0.6421975579333672,0.830648243923094,0.017208553668179416,0.3643113969811286,
 0.3737935908290495,0.593685528449998,0.4703189883519322,0.3993847862281962,
 0.09878551280929859,0.010478853191947635,0.9494157018925854,
 0.6460639497290647,0.8319199049398794,0.6356966060241489,0.05997559928710794,
 0.6185467723813792,0.6847292129149204,0.9709000041499799,0.5729805220738321,
 0.3679637042353876,0.5565499235418497,0.12171634781126817,0.5312519756814429,
 0.8231499874322566,0.8475656697342832,0.20087312407245583,0.7420736953525766,
 0.6036566723506298,0.9694195262872198,0.9553306358878744,0.8754071294552003,
 0.1068357347460358,0.9541915218357979,0.3033223273853104,0.5131645638535762,
 0.44758891505877807,0.2647508699599047,0.5942846953204843,0.7247122633566437,
 0.962963093719565,0.34033090583572134,0.4685729303239804,0.7118601761834622,
 0.2607839718610986,0.1971945994273978,0.40574337064422394,
 0.02463154263903089,0.3976101386920521,0.026055663423787512,
 0.09595963124754636,0.09471516939161384,0.5526452061473454,
 0.3085045245126391,0.2007838822841892,0.3503868225646527,0.48914542701994856,
 0.21713101539963897,0.3185151882879571,0.872357236291073,
 0.037931516488438954,0.45969637928050633,0.8073252543509335,
 0.07389307506943177,0.12879613618323338,0.007327138025504221,
 0.6398045971880151,0.29874026272182697,0.27347523601388635,
 0.002797119224087319,0.7812001547094422,0.9009637266658724,
 0.9575385473353879,0.6245388553731949,0.8601203491128017,0.686147082849402,
 0.8705083154199841,0.2884406661971435,0.2465591299519635,0.9998316341731983,
 0.23781079274830008,0.9673464425913322,0.41395267979168904,
 0.21020447884517235,0.587837643068155,0.7592381840304231,0.5921584810707611,
 0.7682289315370594,0.11369002191562827,0.5548241063739814,0.1945356230491283,
 0.8900832933200193,0.522474748600149,0.9181623810533318,0.13050761458968851,
 0.8827986754378135,0.4677336763969716,0.6151164614108453,0.19083336213436408,
 0.5511883793317408,0.327083286341849,0.04123750772594781,0.6994724661104968,
 0.3269858518256923,0.7502416321058843,0.2569865557704355,0.6010134668137208,
 0.7521834879250995,0.26762609413860894,0.26057443845499306,
 0.8450092398242982,0.5525463610869844,0.3936614172286671,0.42814782787897354,
 0.4417502170522657,0.38628857192986465,0.4080448116793599,0.2260189679013891,
 0.20886469360571358,0.5209286197434719,0.35140631107283726,
 0.3081029647325918,0.6902484632127562,0.05099969806008897,
 0.33792220130220707,0.32538870265092323,0.1998577641894418,
 0.34881193984028325,0.2623662906591657,0.7710594148603401,0.2493609379258539,
 0.5476880911901716,0.49786848549612983,0.941975256550045,
 0.026059771956477573,0.11924869378663427,0.49730733787907433,
 0.6436118724896602,0.02062666825397408,0.9566966438804252,
 0.35883711903340176,0.6929826530309455,0.41376571807295215,
 0.13644505917981375,0.015490646865223057,0.8322600225196477,
 0.5459546289913575,0.5737809903764814,0.4808771329770717,0.6954591723234723,
 0.8117937371671087,0.8487635791144693,0.1740456551494125,0.11289232758742695,
 0.01615221054746241,0.7644602439786427,0.5033391713054024,0.7523969758495186,
 0.25772166124260354,0.09541622998008137,0.3018142693799637,
 0.1380191206070519,0.7573405199634471,0.5889392259441815,0.785971194510251,
 0.9415153807779133,0.010749162670407597,0.2802294468911116,
 0.06390228282572563,0.48855290910369875,0.7817199559876097,
 0.046938753117443666,0.5121903422984324,0.47963018070874297,
 0.9126476480043227,0.35242079220514355,0.06842056966825083,
 0.7522985792937089,0.6206164404372273,0.5140060679413786,0.8318178598822422,
 0.4937353286146118,0.017271199807827764,0.4700099347626443,
 0.6748520632429699,0.303792520260754,0.09614770747636447,0.5869649029268299,
 0.7567657621084796,0.5560478506220976,0.12191535306311907,
 0.39225650319226557,0.3287082569730171,0.8717369009510207,0.8180099659601748,
 0.08872539698713511,0.5729311706408874,0.4882894577956123,
 0.17578748635103492,0.7709229204018128,0.061808976143643446,
 0.7673041054115143,0.10113566367212512,0.2641199926695209,
 0.22264182177397696,0.35759482909301377,0.1572348707157868,
 0.7925229424750948,0.7982056275604607,0.5312096982128918,0.2558337450523567,
 0.6822581218377488,0.7359026840003101,0.516584132833452,0.8798019479211647,
 0.012920185510638893,0.20085098291272274,0.47313437484468346,
 0.10249672767486928,0.7471200533054593,0.19704242828793705,
 0.5232134113322342,0.3375993720667855,0.5137589399739176,0.2817958419055513,
 0.4954246438144072,0.9167984151535363,0.6080953838237325,0.6879832917105102,
 0.3796144217394646,0.6579028599458725,0.5389076289320516,0.8959305293116034,
 0.03635744716963185,0.792965756488601,0.9473120864436555,0.756071700841291,
 0.6235046510850536,0.9014758782281403,0.9946253487262018,0.2824559009075348,
 0.8073652294734565,0.04606631884046386,0.8107964344747061,0.0348249028100347,
 0.17135373004868926,0.9140145166933702,0.861500725371316,0.21454061501664112,
 0.09855249147144829,0.41280095349942547,0.4814060342216194,
 0.20709203611761873,0.8643129383595122,0.5385142039306725,0.5054673376325244,
 0.31600340079924694,0.19947006998172134,0.175757166040859,0.1287773637231614,
 0.19343119723937863,0.24106774107480455,0.3431496431223471,
 0.8379618566398876,0.45407124776659236,0.04205545723521853,
 0.1914735031337943,0.4533530406634009,0.9262837269533551,0.7042479427036151,
 0.9164597797248577,0.6907724309572717,0.8973459541820025,
 0.018040734780172762,0.7923818739238282,0.3428792801133581,
 0.13756516861993595,0.6971071548054657,0.08042558033754188,
 0.5082426095823948,0.9386704331816088,0.5457910060079902,0.7876120054633555,
 0.4023662034403962,0.9612130566057286,0.8381708825761978,0.1894170312152852,
 0.47846457770912476,0.6787287691784976,0.543892682772395,0.8854292136825657,
 0.15688630017926408,0.5475186858372892,0.6800198773417044,0.6440130726823567,
 0.909335572114367,0.8579820227538434,0.45701154471148664,
 0.007685128906005589,0.7492770185810833,0.4648225883740853,
 0.7800249847760473,0.4400165650105248,0.42276519421396797,0.3973353798943712,
 0.20501367194009223,0.8426882250251133,0.5726635013573352,
 0.13286647131637408,0.6093737994900438,0.8205622818295859,0.8143790355357043,
 0.1211378315180176,0.45316981085220476,0.7721629020500171,
 0.08234489378530163,0.8280246914937326,0.5090346376828734,
 0.12355787436157817,0.7335578760756698,0.6049736618447961,0.4136334897648132,
 0.7124163918907461,0.9248516941751397,0.9529652917385524,0.8758121052596366,
 0.8755734126856273,0.5930821534955282,0.57076515129646,0.3245698818686874,
 0.22282579831412486,0.33428990764723565,0.08479835328650331,
 0.5172798570471546,0.29966329493838073,0.1118503012873372,0.6175317980447036,
 0.4410792966444699,0.33290156833205486,0.050479367137800546,
 0.3596671596649732,0.09969900989932756,0.11643019603669358,
 0.08810208565922928,0.32826718631399654,0.1863430375240882,
 0.02585067585452505,0.9173210900967549,0.19729722815112838,
 0.38791567928037274,0.16447998312356227,0.5525797426458938,
 0.6209707188159965,0.33928256402663304,0.9759714914143273,0.2864996721186226,
 0.060222471528373456,0.8222623008395352,0.9093916543109117,
 0.8058941975204853,0.4296399642969132,0.34128430891991146,
 0.024786918410972025,0.24397723881566402,0.4881966844601977,0.11555450601452,
 0.44478565831867256,0.8948291063342593,0.4777648664748213,0.7602536121504727,
 0.9800205971643516,0.08288350379993104,0.6144082503944239,0.211477488887589,
 0.40051891985838906,0.6222724028157104,0.4304995414010412,
 0.28791335587852784,0.3998943520495992,0.6125431586291643,0.9654666263594313,
 0.7559032100345687,0.5945889418885419,0.7152557086656276,0.7747461095001046,
 0.4441505824554165,0.37804136353764406,0.09104300711366486,
 0.12030200699969895,0.11315992807638853,0.1049716234626279,
 0.024475693136497667,0.476982634806688,0.3022965062900824,
 0.19391497780281297,0.572391477038488,0.5966167366558177,0.9921440121354428,
 0.9182626416250631,0.5409979393695276,0.5145484270868014,0.7778786382445279,
 0.320723126785451,0.8334505888352143,0.8038165726861621,0.12206172021623907,
 0.8735280447842632,0.8382964038647758,0.9696525202831443,0.11713078543627031,
 0.7513712000516642,0.1854562079179648,0.9503467297818556,0.6455975354935936,
 0.9277837735396885,0.7126471553903788,0.3283291364760259,0.6458683035853995,
 0.8147704453982545,0.6179072163132079,0.6277422736620042,0.18608665040382388,
 0.06534442345042235,0.600507084756714,0.2925458579104482,0.7965884789378334,
 0.5805691436386338,0.9128707038824544,0.8514194169887455,0.7974187438627012,
 0.22597217513062606,0.2573143226534339,0.24367608942162244,
 0.3094075191193719,0.9053065899407036,0.5846275027638854,0.9112769028910687,
 0.3619355335616623,0.42657018647431655,0.21394166833388095,
 0.14152459521559269,0.9188532037666421,0.3930990777158534,0.5985122288974167,
 0.28491086891555173,0.13325177246664732,0.8045795051938586,
 0.5910325577748956,0.8249932261074108,0.5803266416545281,0.6269488954036169,
 0.09132196780396029,0.5134213349179653,0.13184280445287966,
 0.7091732940229095,0.6407693220674701,0.6949217337630952,0.6842237853558669,
 0.4585709358447396,0.45999390293036424,0.011332111742137219,
 0.9263245376341792,0.6111381970584869,0.29523426064858904,0.9002830626467718,
 0.05310746932595989,0.6960389363089435,0.7391529843892923,
 0.33226951451722886,0.15739563933320433,0.4478641856159167,
 0.9761475326027361,0.7857461304694945,0.16084960599759324,0.7950927409951488,
 0.39275946201862744,0.6206923680788016,0.9925347042101187,0.3863663083326164,
 0.05549132624989017,0.21546845601108555,0.6021258915422885,
 0.8952257559240433,0.8646559214887183,0.8789590751293046,0.4325896149631575,
 0.9762889850655794,0.9063448119944142,0.7291375536525474,0.8816017056796127,
 0.3476611228687465,0.7222732751675643,0.7970263629087607,0.8016585349484089,
 0.4852752122965728,0.4640934931548335,0.14968469727305234,
 0.34816160088852655,0.8588202290156713,0.5107222543417601,0.2722971539872592,
 0.9212834129188027,0.20671445035183855,0.27324578129541277,0.799574257088372,
 0.17577542687954628,0.579108276061862,0.6079692343747396,0.7729992249782234,
 0.4111824247009388,0.8044545747543401,0.2858474461335576,0.4695646861147067,
 0.4715213004217834,0.4267546343208394,0.17675707739864088,0.8473714265760199,
 0.733308113523305,0.1674953441811533,0.4341609352709265,0.8066473177909991,
 0.7610508275385184,0.35513110951720717,0.209246799117491,0.5141192246374718,
 0.6824060952627113,0.19497376123103471,0.6103317859725884,
 0.30603024960155734,0.4271208910440113,0.728316166896402,0.3100111858389022,
 0.1355338702200798,0.6441799155304923,0.28963173123059327,
 0.02471567221020554,0.10334168949291489,0.35108809170947664,
 0.3269355305021132,0.8400422275906562,0.017532243703594408,
 0.4028492547898842,0.3935692139305126,0.7350440306328483,0.2659971619090208,
 0.8032256486985828,0.5702825118770022,0.7949836711110267,0.8795387235595451,
 0.41001099548850106,0.8997955867587072,0.631368657208914,0.2417521638151795,
 0.4301593650590255,0.6074646869932958,0.20202901405855023,0.2401208348598024,
 0.8489637146627599,0.9094582784715645,0.7509103530129344,0.7659612333977048,
 0.5269680976366198,0.039926261901093385,0.6407228130861723,
 0.5245790706943455,0.24012038078083242,0.5751628970350819,
 0.41149008683473487,0.7921010355576134,0.7064710955653055,0.0267594185546447,
 0.985460096612499,0.18743981697890333,0.023210296154896892,
 0.35276522090702145,0.3037939254553913,0.41489676978645496,
 0.8028357757725545,0.5854734017588967,0.2975763680259519,0.3939840721628537,
 0.43705910828898675,0.7455415919505847,0.2524597347382824,0.5553626913881252,
 0.4062607875958183,0.8074863411911514,0.07647084537703697,0.631111821354391,
 0.2306358490910715,0.6916784056414383,0.9964735277092319,0.8378320836327571,
 0.23530150803498862,0.4009876099525307,0.9808252367495666,
 0.30371513718941756,0.47974197993923207,0.30564534006069966,
 0.9205145769733047,0.9593188087067172,0.6783123987628725,0.5513751307023708,
 0.7374920404047747,0.20429709881096836,0.9322468582226886,0.805145376901069,
 0.24169397053031627,0.9479836523269897,0.3565628882071208,0.8831540238416542,
 0.8833349727933137,0.9826825471002217,0.04474688573130026,0.5388799881495518,
 0.21259725149619624,0.9800024169710282,0.3641542046696056,0.282852597249888,
 0.9564344693207296,0.9172095292289008,0.014617544654784154,
 0.3609016496539631,0.7270664715695223,0.4946763402333425,0.13486558255120662,
 0.6913638348620043,0.5041153087603106,0.3122741708357769,0.1651545964034674,
 0.5826661109671979,0.34969527053878846,0.14052505557933748,
 0.6990326949165249,0.07660066271654109,0.5287715407807733,
 0.25552565297260144,0.4319644499110965,0.3935045900058204,0.8794421756975574,
 0.08643325723715678,0.5878952758994346,0.32215797729169027,
 0.10258704683347852,0.24688567310025344,0.3345585783527596,0.631292221246737,
 0.8356750155321376,0.010452788200341567,0.5805434930961856,
 0.17401389166992376,0.20877922897769907,0.03046845873462467,
 0.7260955115440013,0.013187393770862421,0.5424772162292195,
 0.8976061941575313,0.9806458496306211,0.813711162882322,0.056775607244391346,
 0.2892652711198985,0.4773841996949987,0.9662750165526928,0.20953684512184523,
 0.15639882967109575,0.11544383826387572,0.3235491025447562,
 0.3817302573488055,0.6989754667802054,0.11597483247331053,
 0.11629451555369541,0.4597839612985457,0.7472157738432385,
 0.006797114132595761,0.20645582452539657,0.8841764617076413,
 0.3000309662196694,0.012169654371690597,0.7911430191833984,0.453605458249249,
 0.7142078100450442,0.7192665676390193,0.21790885838446006,0.8853450397415584,
 0.6414413230461936,0.2827195995690578,0.8374781284876309,0.07951904077472138,
 0.6639851380889635,0.5948846239878463,0.02706996732066269,0.3511793596185534,
 0.319057604965455,0.9737918566606368,0.10666298809049668,0.6569277969782588,
 0.5546813096096777,0.7823808072703965,0.18883192109035762,0.5259244128452925,
 0.49977967897382347,0.2835778097113262,0.37236633898338467,
 0.3683681744339493,0.6591646394996181,0.49464228247111497,0.909103333939004,
 0.8371665194816695,0.526590306313716,0.19473803399814327,0.7791656454761126,
 0.07987099605481751,0.8706120071359887,0.571584124365137,0.7027350234973029,
 0.5879953517385718,0.5351685935723762,0.11025136109871969,0.2929690597069208,
 0.26382497404849614,0.5534229684709132,0.10697685471657992,
 0.5221580944116225,0.7347312017839568,0.827959565985078,0.17228589064561362,
 0.9956984897806114,0.4076379834064473,0.932439048815914,0.09777341168865417,
 0.7490383784705321,0.2806164458116962,0.3565510385213284,0.7010636048193852,
 0.7612467790669989,0.2649890744849497,0.877886572495507,0.806157702870159,
 0.7242571743749739,0.5585508612231336,0.8007866143789398,0.4571884935270859,
 0.6945849948581881,0.36853407892048995,0.6043024081180032,
 0.23821522941904405,0.7327172790120943,0.157221238476691,0.9092201104481805,
 0.8311146513413056,0.348361096661571,0.7766196686989507,0.5914673544847471,
 0.3706768405934515,0.2648700970215093,0.7473190656515145,0.19508620580124403,
 0.9613377473884069,0.6972387578139395,0.3884092723850656,0.4705851520091735,
 0.20807063860254627,0.3867055993476042,0.6282097968305216,0.8665087803620445,
 0.8535228750690655,0.9884312985205579,0.028196848300351074,
 0.45837195720935053,0.2115314899405547,0.5083815524333908,
 0.17139069863387957,0.6195887433925895,0.7083101619048887,0.238466851022613,
 0.5121498407393518,0.47128586574993747,0.11161140909916745,
 0.8616442207779295,0.6758033593993977,0.995366644502599,0.7733886010939944,
 0.17136018800111663,0.14734109167737897,0.3042780770870125,
 0.7271200115532996,0.7609846581380374,0.8272670359725022,0.7099769633799017,
 0.14184085277816538,0.16232738685370907,0.7095065118635417,
 0.053133916657856606,0.9869357301405745,0.04349268203182677,
 0.5360885012828337,0.8718406442800541,0.22728242405803645,0.5841261062653735,
 0.23519847819483075,0.8668680735623207,0.26991548008483,0.42333399268888927,
 0.5942634858090058,0.5834149863812459,0.41317316932182513,
 0.28745944842236204,0.45412658365590897,0.8694922801401311,
 0.8853286112253371,0.8798298095860123,0.08326163573960421,0.8217651483433058,
 0.5462478480645595,0.33614467030467776,0.3152363991820559,0.984160273981407,
 0.9415235014306876,0.6932573848601177,0.7748761550972131,0.08507488316792144,
 0.9627635709970671,0.0598204487235795,0.3332370201073682,0.20223291872442595,
 0.4622567177796145,0.8002294109966235,0.9368626360701409,0.7174869670585058,
 0.3571374427596795,0.4419237267203495,0.2600936145937406,0.7403102922839998,
 0.13722243176832727,0.8634134423492155,0.33647031812364636,
 0.5242691596435706,0.7002997569316273,0.8898920456258846,
 0.025721590952478723,0.9313572196088926,0.4307518606644145,
 0.5202817499160589,0.044788919021738094,0.27091534734455147,
 0.295746911105516,0.3925702607261573,0.2502671728486867,0.9179800274998922,
 0.1762830167759416,0.12100953614840781,0.8177424360775912,0.3334503857601916,
 0.09219085928344728,0.41127855108022715,0.9791336644326326,
 0.16671153035719022,0.7726301145602513,0.15070864283743712,
 0.4253314399188711,0.19085612321322842,0.45224350468413843,
 0.2218107539554216,0.10120734651480179,0.7914209205037278,
 0.10786219907283856,0.6244995260016272,0.7703153701232834,0.8829277695237338,
 0.7858875931140104,0.16302779567635228,0.5422132826645273,
 0.03371076807848361,0.32449118316986925,0.911851829128829,
 0.48618890477978915,0.5222511623772261,0.5073615385932674,
 0.019824045345321784,0.3238757473268119,0.988840138009245,0.9811547572319816,
 0.02498870204312653,0.9624685579842331,0.746233959096095,0.4340982729278201,
 0.28556185205553264,0.3504551484322237,0.0044940489337726985,
 0.4488371712798054,0.8077385295390705,0.5512945886538467,0.26916939090376557,
 0.4673550284651795,0.9370559334979413,0.4269008302258891,0.10337732711254322,
 0.5054093480643207,0.03683037750768736,0.06694030145475616,
 0.41000908449511764,0.017954029285606854,0.005138743899441134,
 0.5357595187814242,0.12839347512419153,0.3572255985166217,0.5355695155272515,
 0.25659117015769795,0.38065962991570523,0.10536545554619203,
 0.9463546001217646,0.2915600574179734,0.6799535508451803,0.7700514310806009,
 0.9555154075654899,0.9825211855118483,0.2867709900963,0.9147294364498852,
 0.7904371801941934,0.21416798588515662,0.3575853902903158,
 0.02159872758295245,0.5377521253269824,0.009802738192283389,
 0.4998221113648398,0.35593039396277737,0.581688363057748,
 0.060045882838015485,0.7718742115766096,0.6843519034275329,
 0.5753012044901561,0.5706436542403622,0.17174945960876686,
 0.23470843282766274,0.7070368474617317,0.4655967179506608,0.9391133157876692,
 0.23338124609856425,0.8983544422271922,0.6396181137825834,0.880862803507103,
 0.9928635761774156,0.12442921759856718,0.4940151131200632,
 0.08997545583426869,0.8451732428306296,0.18656143420765092,
 0.5090039464181143,0.36043388487904615,0.38508490660540995,
 0.9750604595016709,0.7316909575859203,0.6476335180687817,0.2307717779615972,
 0.446883074001426,0.8509503266671806,0.5842404666476297,0.493389322763889,
 0.22463623721838022,0.28700376975592057,0.9390917115592963,
 0.8949966836950127,0.02186387910397858,0.929082546384008,0.8866728249747539,
 0.783250148121712,0.6838671720639651,0.1432565626077953,0.7175179595802589,
 0.1285875981453667,0.9415193985438318,0.4844883943740097,0.7742065909338367,
 0.1951451410037156,0.3336525665502095,0.5330660233659725,0.16266799613969762,
 0.6172297912053121,0.07643250652303357,0.7363059696087648,
 0.47763523586531964,0.26760193802203136,0.4549129920654055,0.598977910548968,
 0.8441727243829652,0.09763649426965149,0.8666271706883473,0.921625866876893,
 0.2016208710273113,0.2527704172451739,0.45110818128477614,0.6250774726977013,
 0.7218318022477238,0.8935704161927828,0.24324776347381483,
 0.38303153399467016,0.6361770862189804,0.6020297234227903,0.3167643688764984,
 0.6469558063075821,0.5555903609267281,0.8787936115823206,0.20325848960131077,
 0.8104283900861626,0.25167722703405593,0.7801484735310282,0.469392654635964,
 0.6624998371488342,0.06943134348923496,0.30338792218474686,
 0.8922463589468751,0.5158960755866006,0.082338317225636,0.48559762101814574,
 0.6621329499000308,0.7087781224859478,0.7936433864135914,0.363209008028089,
 0.4472386410895296,0.8381029886657482,0.6841576516724821,0.9830334703503188,
 0.321602951115847,0.9945603345831651,0.5995042530975154,0.20955473492406929,
 0.0557565383186458,0.8433092471027217,0.3866658610191288,0.7115415280114509,
 0.6777031020197958,0.568556295973905,0.9778382432928113,0.194163577885058,
 0.3880827368035563,0.3425255136103339,0.19919524310590087,0.2709088601444676,
 0.9687927399934302,0.7219297424699965,0.4815950302126977,0.5657974591562054,
 0.9896540299386885,0.8320143232716769,0.32020571259867425,0.79996624462885,
 0.6128097120980374,0.05216416714212058,0.7984753915184287,0.9043896223049155,
 0.6121681009641928,0.4320099300322605,0.601518627676891,0.37741935540721416,
 0.4236148771623012,0.8650493961652475,0.1483319989116464,0.6122488386493333,
 0.8040065380109822,0.8771926006151903,0.13197506550214166,
 0.02038766865026942,0.3335136300545205,0.5564806001190782,0.1365887921233475,
 0.5228060972586498,0.7604739203801976,0.5570234885409877,0.8066644391753175,
 0.8498855625647309,0.5544153798282714,0.42145380384020825,
 0.17554263378946844,0.7262151614575422,0.036485968708611316,
 0.9753784863508828,0.8232994712939812,0.4205265726871952,0.9395994977487231,
 0.5185932504294911,0.23170234325485373,0.23502633210765655,
 0.7533534318175121,0.6993133744479145,0.8382013857946387,0.9047200045155135,
 0.20737727957718888,0.3284969343915505,0.9032038826336108,0.5907618240587024,
 0.2838718878984815,0.6274055139237018,0.44908745191545685,0.9474041447215702,
 0.22215574323828147,0.9949618802574128,0.14373852266562648,
 0.1508057058500074,0.5331111202810068,0.7143167791857322,0.7886825421753474,
 0.8483524725688896,0.8484889242703678,0.6422767819025085,0.4810424604159339,
 0.6263225873287291,0.8700209228321685,0.5709018071578896,0.31834030389774237,
 0.8309870018686693,0.8627657896929737,0.717581693823219,0.5735702489003502,
 0.7952375735957469,0.8224442646809336,0.9812532538714412,0.8415150552800689,
 0.08062622514944884,0.07108280951533563,0.30330997435899754,
 0.17671842141890715,0.019861766892263377,0.22453235640770353,
 0.18091498995750244,0.7506218800920088,0.28708943678876453,
 0.8498833478369343,0.2384606906881961,0.6686008269928682,0.7956985299176303,
 0.8458281889118688,0.08586263812180184,0.5549293671925356,0.6802206129787067,
 0.9477639532481915,0.9559093918071515,0.6220121301352116,0.8048770236978781,
 0.06164964291282382,0.5614643485976201,0.8933108201898866,0.7397185610933388,
 0.29343650471610017,0.5665318958264851,0.4074762402242903,
 0.046904744578790014,0.9188439395332727,0.6734778379364278,
 0.043788864904620306,0.9757925671153641,0.755925453605847,
 0.15381127941167005,0.014441000816366789,0.5750651792604738,
 0.7446438526775914,0.9292144504929588,0.013637245882607396,
 0.16163421118126098,0.27366078043525766,0.5519514153077575,
 0.8886831436718005,0.40607923265392243,0.4470515146354439,0.235491556264942,
 0.9596228988635865,0.41201183657279306,0.21251072991167563,0.667245586856516,
 0.02026419175584393,0.09072189861176194,0.06940880569151808,
 0.5631062192374495,0.9526759937818061,0.18981881232079545,0.0800524544340192,
 0.3403900561584473,0.2352395509634747,0.8828016744588365,0.8816871058620652,
 0.013657542790165733,0.34599955165330165,0.9860948399759467,
 0.5308604552359173,0.46422328778096333,0.1785843149835762,0.3467408633428929,
 0.49486408830417283,0.27082882182276335,0.49398304535827964,
 0.5070331114651284,0.16871272917644553,0.5853062211801792,0.915108403723782,
 0.9761290115385319,0.7431295575626544,0.8978084577690078,0.5365765130664579,
 0.06503118637264615,0.7716727789231398,0.12144285687850409,
 0.5630017298282269,0.9652757738874178,0.31218627345353867,0.762336367044418,
 0.5719814466996289,0.4475425754564488,0.6932226747575425,0.8536734490404881,
 0.8949353518042243,0.5635864277466287,0.6543359763032005,0.988549257980805,
 0.7628884133225698,0.3437579882335161,0.7675395758465183,0.9472075076958795,
 0.81195525137983,0.07534417621004252,0.11432512989865162,0.9012525186716418,
 0.32680759213776356,0.2396904386639831,0.6938051610462386,
 0.11391570991345124,0.5432787576176654,0.14803345933352297,
 0.2628071064120139,0.3637613439206949,0.010959555165786061,
 0.5096754820366326,0.6682926580966265,0.39242737979891396,
 0.23564817835160612,0.9853302886106988,0.2982056799732986,0.4182758271369367,
 0.5794872493749333,3.0737038759598434e-4,0.11229027629946153,
 0.4140160819862325,0.16461909552823062,0.288519067890907,0.4375783197237946,
 0.24068734570328731,0.23403661671003406,0.6785739168254169,
 0.7937260617165799,0.6349979943714547,0.015830280672655683,
 0.7115114114468399,0.48454550244327976,0.05071322464680483,
 0.5277563886024321,0.9325271911983031,0.6660008096245285,0.3399635199459071,
 0.7939634205248192,0.5145070602071733,0.5966736427019236,0.11664570639448812,
 0.20228085268783003,0.12586822584687252,0.14655258988196662,
 0.8038157847944573,0.15128670120087018,0.4728320508459445,0.2372103919397247,
 0.9388717940458333,0.22290781205813093,0.29785677867305793,
 0.9735025641649716,0.047083427252779364,0.9432205022624675,
 0.3253286966929022,0.21339409271765353,0.6630784662987149,0.5528667681377245,
 0.9182264421454512,0.5049833188339543,0.07462721621944157,
 0.06712747383744355,0.8374331435707731,0.839347524942073,0.903573288898718,
 0.9824396195557838,0.9777285731438909,0.6022603694478478,0.11154285860405444,
 0.0048123423766001094,0.5875059695485586,0.857620978553779,
 0.3420330106248022,0.24938532792570123,0.5918987087302295,
 0.32047198629418383,0.9080306645714125,0.4678227523968398,0.9139414696720228,
 0.07992645598534165,0.6526926889871526,0.8765995289843488,0.5160127204685271,
 0.5052598258350074,0.6776551317886326,0.9286704995537276,0.08919318536818521,
 0.4546134417918709,0.19016050026240128,0.20959977818969022,
 0.1890099223425662,0.23854972181441902,0.42361160815164756,
 0.18148402463171076,0.3837561665177097,0.44452578930523035,
 0.7202189775511225,0.13405548716576365,0.6120152471626458,0.6860029517726335,
 0.3327414524633141,0.29992467142486645,0.6281877016638401,0.2568541125986543,
 0.04548544886466199,0.6487790221184216,0.014388586155386851,
 0.014260178942556823,0.9070245724538688,0.2686047032939156,
 0.6974022103909775,0.5704308606832643,0.37207717568063614,0.4322858547356764,
 0.15816091177780933,0.07588492650723899,0.7546850752783384,
 0.5077313542070442,0.7884414144430327,0.47842678934997396,0.8947140937047781,
 0.9091758925888836,0.8992157861977047,0.7857899355334195,0.05072391596271475,
 0.3539020313667325,0.0060092474641250115,0.4295798510987672,
 0.8745678972233778,0.975128337826682,0.7002323195329524,0.6785308622334002,
 0.1556311925865852,0.12110870134288554,0.43631147073754173,
 0.2897982227259769,0.7715696659868794,0.1947659879387188,0.6955579421193903,
 0.37382882794901784,0.19461276657613313,0.1461283185793426,
 0.9987757427205085,0.672616331380328,0.7295124417123964,0.9606408596965464,
 0.1123376102310446,0.45179705341082954,0.8606269699483288,0.7721000017044778,
 0.15656703859535692,0.12887220294097446,0.9567170592906739,
 0.6432544054763704,0.22171286183349748,0.545735528543583,0.735736957007351,
 0.7585598272880398,0.30705659486734327,0.6635835069371334,
 0.32064145046725245,0.19545611963877685,0.9554399523635011,
 0.7164174402941486,0.4385488486665554,0.7409923313771144,0.5523238118335415,
 0.07961687985769661,0.6854232975667263,0.46712371836270905,
 0.7135009393509544,0.9337422813766791,0.40036665677139316,0.7840369489992625,
 0.48686289234012226,0.9770801022642289,0.16067194196257084,
 0.8345675093352829,0.5002287455901792,0.5116581473261982,0.0996689726615515,
 0.5867593737644708,0.190974337853397,0.10615585037200481,0.9224503509568782,
 0.5223112511494239,0.2997890562192369,0.598662906140763,0.5231057211083617,
 0.6069487077366633,0.625601959220256,0.43917980160642334,0.03358129191161785,
 0.23860838685703023,0.8323629060009425,0.2528407901033538,
 0.017549610013170236,0.9641829536624049,0.2866002850277911,
 0.5513782721519028,0.1529635788306679,0.39473191936922647,
 0.36824006394583586,0.5129966891900067,0.7909174820512263,0.7872569547727949,
 0.5865611180380387,0.12547000204887715,0.4041198929168881,0.2021485749886025,
 0.7253488210041328,0.50132770711699,0.2860705272609084,0.6492689842744599,
 0.195078662463275,0.7038032443567226,0.07328447397302895,0.952665025460748,
 0.9124486637740423,0.4183058958960014,0.5688405892715415,0.47149377210580157,
 0.6440250856942182,0.8213077783160526,0.07806029793425151,
 0.10627021436262007,0.5825114830501343,0.9084283678105692,0.798528723152405,
 0.5689314878391463,0.31355875264894495,0.8698303593180389,0.7646764972856193,
 0.15626503212508547,0.2313894813054751,0.9164088453248311,0.9635994823269712,
 0.3760855550205684,0.504364394572398,0.9062074115862321,0.15950695562267547,
 0.4466097221774765,0.6455062255509927,0.3886991495037402,0.6904162901690867,
 0.2802438298317702,0.45152009267329485,0.340527245452411,0.70571590069734,
 0.6116359349716272,0.7278465867102397,0.4745503082883379,0.29543478521975663,
 0.3354912317403693,0.3432736316981776,0.4425300655844091,0.23297779781183037,
 0.6744427650214004,0.9505132945653875,0.43914872763097423,
 0.05787122058287708,0.1196485725443528,0.20023471631897916,
 0.4186332720542121,0.5605657537753307,0.2493975315799104,0.10840323818901287,
 0.16126410837643734,0.5542498170796345,0.5837677926250413,0.2410296652231514,
 0.04704831729997472,0.6541900562071785,0.7453440613182701,0.2222555616127938,
 0.6606031090021192,0.37524929581447686,0.7452115766984938,0.5615904068850411,
 0.7358092741121506,0.8436646361181355,0.11092761472700285,0.6557880266884131,
 0.8772524979472976,0.19800925955816928,0.34404743419856976,
 0.7671083354208895,0.7963586186047666,0.9859975861332152,0.27041722437850113,
 0.09551580177212804,0.9811384289065446,0.70184301391945,0.7870852283524544,
 0.7157648644820949,0.03462614300412348,0.524893986770011,0.00900573786659753,
 0.9528892291030541,0.9102879917811879,0.571896942754277,0.8896830568859058,
 0.6781964660610966,0.9450969862830886,0.0630474389708705,0.9950236121070031,
 0.26332528631374574,0.4444005025422304,0.04968197807034502,
 0.7021089871937012,0.8634494391781447,0.41592930969440356,0.9782471049763308,
 0.991729315316514,0.25216363613960235,0.3438660936064699,0.4477674370902822,
 0.4818434314739679,0.7558258146175633,0.05254111078600965,0.9412463201155478,
 0.24062774568057765,0.1171202249950799,0.13997251829096802,0.582664752286651,
 0.011180435478797435,0.5021259027863119,0.8002849285876095,0.609243761897801,
 0.3563988382973823,0.2548972658513864,0.5289050141541045,0.4535510314723229,
 0.18377456781651924,0.9079373919465947,0.8798757627386412,
 0.10728000688534456,0.08737872470886332,0.7336080040781409,
 0.18940804116152854,0.35888053753150795,0.7346828353302011,
 0.7883072056676859,0.34318147720353254,0.06516179763522001,
 0.17225442561231452,0.7652140669118832,0.11305867191917485,
 0.1499876879646992,0.3689834379753447,0.9251112741078775,0.5556259363839734,
 0.8338379491127705,0.12413898162877057,0.6434122784984941,0.8764540137684905,
 0.273082477404597,0.5473618414945238,0.03256273081385763,0.301823907540159,
 0.2519997157000232,0.6087582653921382,0.4205141463579869,0.9966503709632322,
 0.6237937580417895,0.4984563652643277,0.8756598622125661,0.3930784705647591,
 0.3074999153547845,0.44620747062638266,0.0943890318545133,0.5450143432257641,
 0.9903600739576363,0.21602133999782924,0.28986329217071694,
 0.6526076249261956,0.4712202432289253,0.27550856654964795,0.4823294010286924,
 0.07875932600289759,0.6177712653079666,0.18036290927029064,
 0.1955430673662384,0.5827127379142394,0.9624453593517451,0.7781270672132076,
 0.6272325662142522,0.8231151092119111,0.9045282392462124,0.32178196671169546,
 0.07770697482238187,0.42822338459887344,0.3900313349641856,
 0.7437628497298098,0.41967904418290625,0.5750920208634844,0.6394602623442975,
 0.7923779283194565,0.8349187669682095,0.9701391787199545,0.3803440658240357,
 0.9227949009930793,0.6253499190489237,0.555195274906335,0.5247163830999083,
 0.6174403918741881,0.0027459103966560594,0.2053035716638214,
 0.9406088935545891,0.32500741361589724,8.693086632559632e-4,
 0.5241764903337808,0.9807331151615453,0.8634182484582961,0.32995609768661327,
 0.5517824107113137,0.3816936472122996,0.5035363852239699,0.5165883124155863,
 0.7619353668171944,0.37183735360175785,0.1456739181799903,0.1701292472590945,
 0.7356271994802163,0.23403891952885658,0.8215279000958152,0.5011540350926799,
 0.2102321260492036,0.220960228885289,0.06036961461091772,0.22989788173884418,
 0.276961846863953,0.7107013303566013,0.0956668176644847,0.7165906233961267,
 0.5243640516627365,0.825601959917099,0.3336397029589597,0.8472558986828943,
 0.041391245044287595,0.4221525554420109,0.9520210230274688,
 0.6776003981580826,0.5452272162888466,0.8696793149100807,0.554798765602433,
 0.3933212582415472,0.08298842704245002,0.2858065847019696,0.6431641962990664,
 0.3149171738286364,0.7343906680107994,0.3889233947966474,0.8369584444591183,
 0.45272756547309745,0.3453410391019802,0.3420668040931871,0.9955433186874539,
 0.1659776838484378,0.5553463558606364,0.028015513117526814,
 0.7676111757213896,0.8169589978671085,0.4572229260416173,0.1695522710865659,
 0.04351742749502374,0.9134325295091604,0.5731160087117123,
 0.20941923974060395,0.7368471362098692,0.017897412987249894,
 0.4888284236788385,0.060388625444433464,0.5252857387565697,
 0.9139483445810873,0.3845716790059468,0.5604723661285199,0.22578182651539414,
 0.405716380785225,0.9785715091836787,0.9201854213985479,0.7685935304312157,
 0.6869277602107812,0.3461059927100084,0.214633279419438,0.5078636726448661,
 0.8911262679792641,0.32504623685094247,0.020378789959120458,
 0.5228153314979851,0.19979126372200617,0.48972242658870924,
 0.09981331811210992,0.7614896179708786,0.5918646971883051,0.9991496133474248,
 0.5388337309141799,0.9951639598263347,0.9008534257768519,0.2640435656399903,
 0.9600289786180797,0.8424304311535651,0.4632052395882391,0.39920768909907756,
 0.6283861352901428,0.3886923967984033,0.9537372252859624,0.49566618557428754,
 0.9536312496915693,0.45709508205288363,0.1291412750066676,0.2070084960843137,
 0.8315900624349427,0.6668356578818241,0.39766767020986926,0.6320931646268327,
 0.5605042622722036,0.47849009313727764,0.7861033210643775,
 0.46602446567264877,0.899226758295985,0.18834496314820048,0.9104239077862266,
 0.09093821233621013,0.6395002916893748,0.623657126884515,0.39105253164515863,
 0.41942161694982205,0.7872772688509957,0.5177051911707434,0.9416419058460982,
 0.25941174616758444,0.9070979149407344,0.9268851508332128,0.9820500833133936,
 0.8610170111960398,0.024024092404935193,0.06958701056311711,
 0.585225390999078,0.3539016164715787,0.7744723547745501,0.7092891634998186,
 0.1956280972863329,0.9924408126322025,0.6605127361917617,0.46809558906813087,
 0.5991245638889586,0.9172683452045677,0.7754063795650922,0.4300977505716692,
 0.9580782982027968,0.4328697832164238,0.6215405145473302,0.09933063003404652,
 0.638952996524362,0.8717954447679384,0.6607740054633893,0.9702810933929503,
 0.9290897324574201,0.7498275933961305,0.34389302639269614,0.6917450562070274,
 0.44040410750043724,0.877188950851897,0.5590334477040428,0.7172055824252028,
 0.8270748611347315,0.260205725263985,0.8626438349136549,0.18733924021011694,
 0.32536753959673237,0.44193938777808783,0.9390991970350396,
 0.3499296054821516,0.3602866197104858,0.5423410055851184,0.14901481065273559,
 0.22792591560045794,0.5753293714689597,0.32371239003587926,
 0.5520724796405104,0.8250165432095219,0.4631480755312126,0.662367549850958,
 0.4326130230280443,0.7354022067289246,0.07693016064748737,0.0675032968992415,
 0.630773357136218,0.18903853933190973,0.32901322945917866,0.9597910142724034,
 0.514357416548701,0.8425387782588662,0.22158841896890102,0.40402085805415944,
 0.18549476156637779,0.7579931846240051,0.1745946858420654,0.776091887401827,
 0.5541958501420863,0.25003222149700666,0.1604058242480335,0.6235441577803633,
 0.7356637501458851,0.7094354713280234,0.309579719692327,0.5555373244214223,
 0.09299301248081782,0.7710489017094052,0.6502477821577571,
 0.05126248719201476,0.4369944160438056,0.5856084534187008,0.5591494352979961,
 0.49731215002194284,0.5649536021382712,0.3709673371886134,0.5807668880350881,
 0.9433352158200025,0.6187877605630736,0.21792997032930372,0.8073908475046977,
 0.6290614126170628,0.1099106254922293,0.4638883513011848,0.6026786196999746,
 0.6305286048658276,0.25410336424610214,0.42053588398975206,
 0.5173377237382524,0.3684679383901237,0.8828827561762115,0.9315148614310171,
 0.9661732005324503,0.23890242123145145,0.9036682651406853,0.5480887047445385,
 0.8258616912784157,0.6266329097598355,0.8196489685884908,0.9362399702016195,
 0.6628726342726252,0.07044569906891118,0.2462727771259362,
 0.06114921736531198,0.6360937810101254,0.06342046750192754,0.306726065696343,
 0.9079010131380965,0.5438425179285247,0.1391002627607515,0.11306792768089491,
 0.9760615747069175,0.169193801165599,0.6427946225102372,0.07332602366098795,
 0.9191603324241705,0.13279004897072555,0.20944044665206707,0.810739871526805,
 0.034234116706440054,0.2515298095057942,0.7485176813482656,
 0.46303243437173924,0.47438034857795996,0.9059821865287345,
 0.5730919133486061,0.824411464482859,0.422786640317102,0.5484832691983924,
 0.2612033020641109,0.5126859795158553,0.6372317669361203,0.37846330968802977,
 0.3875863961069488,0.7005344662035342,0.24098491695127777,0.1926548289495268,
 0.772488995600565,0.6125948457054413,0.38931116501790175,0.7437492589392044,
 0.06739072834836246,0.6058792354937269,0.9518926470082049,
 0.14166117142117818,0.6534705212410961,0.7035188097783837,
 0.40823798715195037,0.1691214782764927,0.07080572797815543,
 0.44468085074880015,0.28169742516580387,0.030996797136667253,
 0.5571749129283843,0.14257622639102907,0.321700019089875,0.11911201550623773,
 0.4317268844091723,0.31137638105567944,0.5408916045372693,0.8268304176991819,
 0.6703236174568666,0.6367682459119034,0.20662403490984582,0.5535523223764947,
 0.7186365973997823,0.5886539446942779,0.5909953377247225,0.48920255148584046,
 0.6879168360114215,0.6342951523726118,0.8019669847470423,0.19830187710910596,
 0.38270408616925333,0.6957757390908563,0.8485947284682324,
 0.15224539531910464,0.5894507890507518,0.7652473237778464,
 0.34318400211910205,0.9890102107245056,0.7719691214931189,0.6456427502068478,
 0.42847712020054707,0.53053962957949,0.5735366241621939,0.8028572568733616,
 0.08510431802148444,0.37927430537764406,0.9338870297806163,0.024474208609505,
 0.7261914039853468,0.24116827893477366,0.598136200115402,0.9821505351492164,
 0.5025226285556552,0.6993147066951967,0.9099696761776592,0.7974604093607431,
 0.3584000949917041,0.8959090936379117,0.5381015681912714,0.24330689693113228,
 0.38247697520786006,0.3972207049227988,0.357257489081643,0.1968353274206528,
 0.6729682301802784,0.435633704054293,0.6928669952771274,0.17475549564238335,
 0.49961080224449883,0.9426815974552966,0.7364217375433308,0.3262165116541531,
 0.24889657582423386,0.5128407693107141,0.77557134373753,0.6519044930014258,
 0.2905807733821779,0.6269979499598879,0.1422430847003946,0.1329438399707462,
 0.2644043493000131,0.8133849506010733,0.6959887186085668,0.7468451386530356,
 0.12171728917483415,0.05875605933955452,0.817474788075438,0.5539203109461853,
 0.009129590639874952,0.10041063099422742,0.279360465797163,0.99557416318286,
 0.5096808911245471,0.7591147400983461,0.6947452305549182,0.5408962560796349,
 0.5475243137186973,0.39678820737666,0.8067352437509889,0.6704500919171369,
 0.05414186750187566,0.17002547523873368,0.7101947111753932,
 0.7672636879410554,0.3041249060947899,0.49052517808315665,0.6133775727807219,
 0.06911918356501623,0.6675440314323704,0.08234889655507294,
 0.25004752988390533,0.102949856583617,0.48159542408118694,0.2205213921574759,
 0.8019694848428623,0.2765539220778157,0.3890996931551379,0.4037464617781835,
 0.06587256972643518,0.1946291531918226,0.44230432559943433,
 0.35954534535081295,0.5153622159807908,0.14093612350094042,
 0.6120795129844926,0.06406884994177364,0.669979707288242,0.8835047525794778,
 0.26521653453305316,0.9215190661489747,0.24374854100574928,
 0.5427969248815802,0.2188567500928188,0.3011647325642166,0.9620294571059407,
 0.6445493868855157,0.08291892280977065,0.7649257831407175,0.7136942441956194,
 0.4527425382699519,0.5237387433464695,0.5228452093573513,0.8694235387330952,
 0.40567176820590267,0.7390385017784649,0.5780866921706228,0.2544382477806796,
 0.2526867160806884,0.3974220351431724,0.22250471425885,0.25865803124154163,
 0.9500275148980477,0.5460547722798514,0.9074128342422454,0.4224476943374512,
 0.6853667452247486,0.9063906989610818,0.0692219917686776,0.4345098882779327,
 0.23306315511070252,0.5329440249483912,0.8407117351150468,
 0.46098368450817784,0.09539189799287973,0.9313080493068532,
 0.6630530664863062,0.46075843751965817,0.6730854994862767,0.7940883756824011,
 0.4005634370290977,0.875275865203088,0.6594544562507217,0.937037350767063,
 0.4522244514232897,0.2440197762705717,0.8571462632833845,0.6870208107283493,
 0.11083128818322774,0.10273828299231624,0.6883250772304862,
 0.1676977118384062,0.6054927725344763,0.23005830069910083,0.2955063846391134,
 0.2803053598494094,0.11093912589436439,0.26530741347957043,0.175240837267731,
 0.9185020139145902,0.6708005468976647,0.9219623513229697,0.5569204562796479,
 0.7864672247407382,0.8223025427172943,0.2325646783066776,0.7736473161907843,
 0.9177573472792636,0.28719386597126006,0.5175721060431397,
 0.14695411047268447,0.5436330486664097,0.855675919510168,0.19043957818654433,
 0.30844096585088343,0.18933340653407038,0.46009921487602523,
 0.6954103667262237,0.36855619861292355,0.638864451295505,0.9973982112919841,
 0.6093947205740214,0.5466201249506649,0.2230557570945908,0.42586777032073386,
 0.21833513340731903,0.7747390455926304,0.3691645954005772,
 0.20282382374213082,0.07766550643295722,0.07204462909911247,
 0.4741862822894324,0.7551609324012227,0.061377549071911286,
 0.32507294691878674,0.4720878490685735,0.6458334661424773,0.1469333908879883,
 0.04260804606779556,0.19483874090429898,0.3114056864411133,
 0.5088788382404024,0.2735247186756573,0.07774325280841954,0.5599929758191803,
 0.9163893165867608,0.3038404466112067,0.15269950705923607,
 0.09575079406991227,0.21434237963910074,0.43477941992464453,
 0.07431275038604868,0.4891332810795733,0.44267549207482937,
 0.16991412059266786,0.37850514050831296,0.7256921103868234,
 0.9380903589412165,0.7298250593474689,0.4085891309933518,0.1354961325356212,
 0.3849524148690039,0.9613757870974895,0.043343546205534045,
 0.2095835058703195,0.5428169232278641,0.6208776541972404,0.26902156203764205,
 0.16336678654705195,0.884255896503173,0.6859030739029273,0.8457884410683404,
 0.87791795510943,0.20355050298962352,0.6081484203014138,0.478964440688164,
 0.19026777505352632,0.5945009435197627,0.2195703730943518,0.4113058756516974,
 0.8858481677362884,0.35971333632512714,0.8517645138313005,0.9853550779259933,
 0.6845772312739025,0.10137960430162896,0.20192251031602892,
 0.37304334849870424,0.3060834287164039,0.7519476026293189,0.5344079164865991,
 0.8957019707666162,0.5923168381947813,0.2591763687075316,0.5253101403852574,
 0.1710579114922094,0.17588134752145868,0.10488135819991873,
 0.7471116069506766,0.6757673966706501,0.5539631937258666,0.11065558393297625,
 0.864137388345392,0.7970780074915488,0.47295991756085476,0.9408808607426982,
 0.39651452739114845,0.78151233908775,0.9510610986925566,0.7366768047553175,
 0.257796136942291,0.5118822927956915,0.21226136860993028,0.13526660363511878,
 0.078248299543076,0.2824419378744607,0.2533793764754071,0.05188761628881411,
 0.7041865037529671,0.8940358246469692,0.9681958940653939,0.8999436543394389,
 0.19612500317557646,0.44278696655343996,0.9104163386275331,
 0.8534863412034145,0.02436402156915851,0.5807379950843964,0.5236214957603522,
 0.029296211109816617,0.3849495487973944,0.7609121009924722,
 0.029021914612561317,0.6381862585512874,0.004639927114139519,
 0.18802765289349022,0.2509660557070439,0.7995511644113463,
 0.32306663565908267,0.08369521910657018,0.38948430417151014,
 0.5210512509103419,0.0606614799408155,0.8661212965614208,0.06960421548045492,
 0.7590397305010252,0.8617485798095962,0.027070578433210923,0.926589261430761,
 0.3949146935154344,0.5362169479973029,0.7121533189094494,0.9602706185624621,
 0.4649327200130995,0.30164922821727147,0.19634293928964408,
 0.6843598448881095,0.6487298772919093,0.9044262477082637,0.6356671936201419,
 0.42197495637967886,0.5252705074752497,0.3464476630671218,0.4509840453691605,
 0.48041757201014357,0.6645296132000528,0.11919875528714974,
 0.6041927188902969,0.2594305115480874,0.6143763147085743,0.5271182459136045,
 0.041831629942363,0.8143987118540452,0.4675825891596449,0.527218318198771,
 0.13496745915444308,0.9691321491843378,0.2574111384231984,
 0.12180489090744484,0.32688276475911837,0.20254577749471325,
 0.03549770961759813,0.37916285472753863,0.49840218717803875,
 0.15212695486424432,0.3165384631294751,0.8337671051264371,
 0.48087747915224544,0.817581955966691,0.5928539622938775,0.7718081630123541,
 0.7554269405642154,0.2150848444592688,0.27513905916341297,
 0.25811257993697256,0.23659318389705286,0.6363751522681592,
 0.5408762999280354,0.45316052763706804,0.5637230495287755,0.9451392760521906,
 0.5139497493426424,0.839489964347913,0.3516669845810828,0.05008454758114422,
 0.17717549014453038,0.07255553915186586,0.8913776818876779,0.292985406236961,
 0.6927572358774213,0.822669962926395,0.11862866595973143,0.19987458764531685,
 0.5709488386643656,0.5235359595552112,0.2599055143440919,0.5010650065126258,
 0.2159379282262759,0.3562864430903765,0.6149227405954147,0.33820594877610766,
 0.13287085367134244,0.8994421587156648,0.7835095390377781,
 0.06241272618662119,0.03635063535780181,0.26023887519512523,
 0.6981549526500681,0.40302124558532704,0.21274371025238403,
 0.7564198211390835,0.7080872864472881,0.30955993282427574,0.590699659874685,
 0.7908560471490464,0.5135299496151599,0.3685915289353564,0.33751471670910904,
 0.44112059559044714,0.5687373354093975,0.7955992889146863,0.352480070869156,
 0.3275933586135913,0.6578756406524581,0.999021434459493,0.9017740700495235,
 0.713629172612049,0.3628942341001816,0.4773797446949015,0.4527777605526533,
 0.5032049589843106,0.8461788995549798,0.033125031559553486,
 0.6511516935905461,0.3088492181459557,0.5103246230751627,0.14583208696732142,
 0.5343664266546324,0.7891878847485359,0.6973598304515322,0.400105019159413,
 0.1800136480702318,0.29135437208202597,0.08488695624420628,
 0.3762245776588773,0.5267378408096721,0.3473764636036145,0.8703181604717726,
 0.5276703874251303,0.8497247096364849,0.603345953753895,0.9115554312068221,
 0.8402871812001016,0.6813550844699809,0.812715639146534,0.9860998596916501,
 0.018522139810782434,0.9433767198261233,0.30480583915520243,
 0.22551583656596996,0.003358857726571185,0.06993447129130093,
 0.3169824441556005,0.7583598565152542,0.8363474307460185,
 0.003937250996215225,0.16690846660998115,0.1168670148401596,
 0.6458784730718747,0.7471628119398263,0.36433664626885387,0.4805422078358764,
 0.7368187454333401,0.014301151182500416,0.2519226763565311,
 0.7950887336012393,0.7232771563970717,0.803308945593577,0.5347514314600258,
 0.666986569767534,0.6891847983910372,0.1296219571102375,0.38710104970252923,
 0.7752393427093043,0.7492762617244395,0.38600135423975235,
 0.39578890736545835,0.45183160981546167,0.50748756504168,0.22209999898454136,
 0.6111522787050903,0.5261685973773101,0.1289408226570702,0.6475588936095665,
 0.9789534087967167,0.7288496842810434,0.3055641900747522,
 0.038337656233709994,0.045773153203314876,0.2087118547574689,
 0.3441704175386737,0.4161868083470658,0.2154212100421662,0.5395078245565903,
 0.5481840500419053,0.8504573077840029,0.2076006240133348,0.22130259869539448,
 0.009504238391536135,0.4291945958053106,0.40160898827755787,
 0.17036768797743662,0.03507946006158136,0.13056075212077856,
 0.9000055622788629,0.9553324653955834,0.7957515633400534,
 0.007957107389246154,0.5759270189549743,0.9008828618569353,
 0.5234302565167778,0.3093990482769683,0.4687678776164589,0.16964489533408345,
 0.7026828625979329,0.4033928992673319,0.11829115898541598,0.3736469676555344,
 0.18661906799271188,0.08577414275720541,0.08184355696390233,
 0.4355696462297316,0.3940099254970735,0.8040178502994934,0.6763587921356197,
 0.7964547903204883,0.769477150788976,0.08546968077593264,0.4739264247894198,
 0.04577876965064587,0.169090480586491,0.16614559609161683,0.934268472057307,
 0.4791298867967113,0.12750851148591624,0.4179999325414707,0.907012723702082,
 0.08687445238695446,0.8551435385602452,0.28982327294083965,
 3.335997664632995e-5,0.06252284560501531,0.6359672904694259,
 0.3416946162263401,0.715815586289523,0.6509844879728964,0.22551305391079912,
 0.11299703726370502,0.5534364168777436,0.837428406837708,0.8439709437802726,
 0.7647252967873993,0.4672623118443473,0.02148503587440065,0.7306239889120504,
 0.029425836861997623,0.31107966808070975,0.10144255292081428,
 0.7637851849738899,0.10193785830291469,0.9495345290414129,0.2557821075722604,
 0.49678423618599776,0.7745866845555222,0.09526498192442734,
 0.10575914957918664,0.9608057615062435,0.5005847876458025,
 0.13015235138897274,0.2895217694439822,0.35310917308188394,
 0.7952661821723329,0.10201156607242123,0.1580794915410516,0.2537968814351549,
 0.646143868552521,0.7012709143550659,0.6095152565938714,0.16233956766717905,
 0.6427892530050022,0.3723524619868521,0.10164292333571223,0.8860879955665926,
 0.3608315162749852,0.6459702440540661,0.808417419205578,0.863638438757727,
 0.29419086870834965,0.021256156614440913,0.1751361132268604,
 0.7305549180777671,0.15674569486520562,0.618328240686967,0.5533142673805767,
 0.7106128862273623,0.45246370309508444,0.8577479011258127,0.9321606300585281,
 0.7011578770232288,0.9897655991712543,0.8483572115133295,
 0.005498277864730894,0.7897214474275281,0.38879412869583796,
 0.3853422691784679,0.10814293959582455,0.3666761145296096,
 0.13739816733455057,0.1490206391039417,0.12807072059207536,
 0.9208501148645208,0.9645843015217596,0.8968591744767457,0.5398225261937688,
 0.5449788901993952,0.3560354004058366,0.6102488288379528,0.7187593483416692,
 0.027707556066675854,0.5646060519785224,0.8070637764357113,
 0.12326512950825907,0.5871118276069164,0.9552219649970861,0.825111718002639,
 0.8670923547943856,0.19031743989936212,0.5598935944454801,0.3881871964810637,
 0.8781843783621391,0.4877147414443743,0.9127580784790964,0.6588152712847093,
 0.9036874613957376,0.3463347380762788,0.8600789045067878,0.8627949453845415,
 0.732285100158252,0.1340633642569301,0.7552730029789773,0.5831612256015642,
 0.37960441364200337,0.187595718705719,0.5385199626689893,0.49230547056574625,
 0.5835078060592889,0.7769652166511286,0.15639602568908684,0.9682284217652635,
 0.8812439597394596,0.9722275083112836,0.1870001686649475,0.20885752419139003,
 0.6448147221501047,0.9694003392738114,0.15631038353642035,0.7711357971888659,
 0.05933231668066952,0.04923433906698205,0.7940260349598771,
 0.7853905103516654,0.29588304165668255,0.20709691555270682,
 0.9816913013471563,0.6217971740001118,0.9976186830840594,
 0.008458924193109452,0.49543369649445435,0.8476698579698239,
 0.2671651686336717,0.8028226148775633,0.01988767037097605,0.5701962762340497,
 0.3248769504657264,0.9523084334778588,0.23375887134790818,0.6317741430309995,
 0.48252550213180523,0.7620279574051825,0.5949217980650434,0.784612125469493,
 0.42861591636521745,0.35362523192599626,0.17222221917570157,
 0.35298229254699465,0.5241088652248835,0.13271635847044905,
 0.9520603134478036,0.8183470329663468,0.4955164808397554,0.7669346185360586,
 0.2912961033721526,0.3697576294127244,0.12271364729552214,
 0.39770631711565674,0.47737882799028486,0.08222970996831336,
 0.8008722524626002,0.5607486180668859,0.7613029360654534,0.2356280048198507,
 0.5339806564585984,0.42900681753934644,0.24210590327929038,
 0.7557351068274776,0.5600143732003586,0.890338691676599,0.7742023342105666,
 0.23029960067306066,0.6302625829596913,0.6313521955258353,0.4549060465177235,
 0.8172211524417587,0.33123495494549204,0.6332409331536488,0.6498667866646562,
 0.7852747966555084,0.1163243818356382,0.49640991003226154,
 0.07631384704127275,0.6211943726901944,0.9250444114005719,0.4102887200240548,
 0.9910330284877388,0.025741958665939202,0.9365719167170176,
 0.8104737243201778,0.4394751229239068,0.7221356832201763,0.11058378468523344,
 0.9824410788174359,0.6244765997251965,0.5968176653582893,0.5150572046350803,
 0.37751547835141097,0.3178247871711163,0.7625737855635767,0.4767813177662875,
 0.7085509752137618,0.9954594325764584,0.06107650196391923,0.2262785056309461,
 0.5881310996185248,0.5817567177491656,0.2151303206326407,0.7924400141505368,
 0.9098663422379125,0.273666527360073,0.6791823267334635,0.29194181288057797,
 0.5870325376748691,0.6781245637637565,0.26404800987720267,
 0.19558538820969895,0.5954922830489477,0.7951820539811361,0.9163662980873836,
 0.6147703558613048,0.5770939686272273,0.4338404449910944,0.5211683611147914,
 0.9664816434258179,0.7784662299794692,0.363028504933099,0.3607044173607299,
 0.7448891361190286,0.3575721326254755,0.9005166759298623,0.5833480674250426,
 0.49382067207117997,0.5971672045001805,0.8927559487599438,
 0.10734401554106388,0.3537838325683085,0.9764851718287225,0.8737129360656836,
 0.8906726159393482,0.1358025126999095,0.3679170576046399,0.8561654554573407,
 0.9161181009622102,0.030851466795831284,0.6396237532099067,
 0.2582918336571006,0.9611916671106923,0.9573753288115641,0.8361309069992464,
 0.7257651410713686,0.013929077850630578,0.3688455368474296,
 0.18225175797512616,0.8576756026907083,0.32077392625837464,
 0.03692114541088043,0.2614364741244749,0.8675332829935708,0.742697977185543,
 0.3100028819622874,0.13700120275023053,0.66498227175301,0.7648523809556407,
 0.8952254434626368,0.23653459671445332,0.46601479528676026,
 0.8779719830829826,0.32908195255115813,0.8977655394026363,0.7478040910780396,
 0.5582326880671704,0.697049271512412,0.6704124839341349,0.8999653027777708,
 0.7113942193127376,0.5188446014605474,0.4447197890344343,0.5702898220826731,
 0.34573666364871014,0.8476190179535564,0.5630292564398024,0.919849123829109,
 0.5335444988563638,0.06609731306863631,0.8579053351151614,
 0.11677003755396209,0.013536685914689928,0.1405798664859499,
 0.2384542351674791,0.614063239752431,0.9986051663478227,0.3109697815423207,
 0.16989673532882,0.7432200374421798,0.4577086850223073,7.220513368511039e-5,
 0.24203147223140375,0.8957798903793404,0.23098190224991388,
 0.9050756138833211,0.7590003508731166,0.13211631782026034,0.5190785487023014,
 0.5794190366176758,0.01411922189881476,0.9290463279580728,0.5163879790290956,
 0.538353856572958,0.04605098799394802,0.4725951993072517,0.6020264646132878,
 0.15924686964100876,0.026524150882229813,0.3566351127769124,
 0.4660659147702293,0.3253308116374596,0.5706842498918089,0.8709653879954683,
 0.3675133542024548,0.1820920967731563,0.20741367164650248,0.7268240731245984,
 0.896343531253383,0.05917710196489023,0.6236511553319477,0.9489928478002067,
 0.5105608991770317,0.7714469700911528,0.8714229306958715,0.5919440767963196,
 0.6668753060477551,0.681695161054866,0.11610617917441812,0.8403738536272545,
 0.6672131901240623,0.6009822602862247,0.9367180169819027,0.21648094042345067,
 0.3965960601194617,0.4716775519006863,0.9666476713279661,0.43745501835284895,
 0.7859149568965883,0.832573358663589,0.0639362486977394,0.5010635722255588,
 0.18782012736274067,0.9785582746770919,0.3963444510582409,0.3632708199562422,
 0.39544059595281533,0.2536776529738455,0.06441377564273776,
 0.27053918307831815,0.12867100136980825,0.11424475299318038,
 0.396398896215166,0.13344164879934362,0.2610654176842807,0.7256486701049752,
 0.3570665586104437,0.17701968322147033,0.5799696411874493,
 0.052982208402320374,0.8116611218615561,0.8133222580516464,
 0.2664829665120034,0.4635206658670512,0.9045647076925091,0.6373958937498447,
 0.4991156923342508,0.721181672431108,0.7232166245283571,0.47829602149085215,
 0.11313039760533561,0.0458275770186507,0.03149791220868725,
 0.9433479262220956,0.8281391704815975,0.5739321806786952,0.24056095988354653,
 0.16983675298041656,0.9745659227867304,0.8686895594365961,
 0.36133985934418333,0.44404299853932794,0.07696906975431705,
 0.20862842600464981,0.21341538011965533,0.6492577333241062,
 0.10114972825715862,0.5613219463496091,0.8339478228143518,
 0.41136652097787485,0.29366377566906665,0.8228533057076084,
 0.8742437749430513,0.198166030536975,0.39332783663969884,0.9047574328641634,
 0.3668074422613432,0.5603732477697622,0.675314567224525,0.06377000858584736,
 0.34214286825798756,0.4265090980240146,0.4076945478268049,0.8784679936130436,
 0.7758965265799234,0.30295515516468585,0.19350687370650288,
 0.3166193833249601,0.7993636534897255,0.11896164261443087,0.8685192601347216,
 0.9671888969880196,0.8235196299372826,0.27819697262944965,0.926903268410368,
 0.39182152814406424,0.650223819175817,0.28942726718931877,0.632479890951718,
 0.7224130107245514,0.13105801794951788,0.46472583860566696,0.549429214688254,
 0.954465647761551,0.5118008970803691,0.07755974809547472,0.8166832879472956,
 0.8570873558009516,0.9802793797314023,0.6243378536630128,0.7013460036602615,
 0.34494122569779084,0.9877080893444777,0.31173475328005806,
 0.5610706554220557,0.17093917526797198,0.5445120248985867,
 0.41270145113254453,0.844825124158636,0.772116182054105,0.10708853741024427,
 0.0683296820880872,0.9594475193308935,0.9790317774806231,0.2664405926040371,
 0.33451000221982885,0.10924679841432272,0.6770412777806559,
 0.9788418703713414,0.6872791266340612,0.5679102175863554,0.2411402484975953,
 0.9579562650101812,0.583544332181601,0.1878293081049376,0.34301812103370377,
 0.7609429886533922,0.834955332897197,0.9476299166649862,0.5846984014984827,
 0.015897946274223362,0.9042457053861845,0.18259999912601677,
 0.3641485811130196,0.6716399445272736,0.13141741829717457,0.5832518830044423,
 0.35516657558489184,0.20396415011637803,0.636652231751355,0.6710412723017106,
 0.4280484600649327,0.10723666829555611,0.8684708144123908,0.831739740009975,
 0.4272777054968352,0.36092637676555817,0.18811924257333823,
 0.9621600888337905,0.6253799983073072,0.7669839532671788,0.9046958654566507,
 0.07870593320845187,0.3802559221331797,0.8480065942578008,0.9950310603707694,
 0.9843945372314108,0.7305066878208996,0.5551388727636754,0.39791562141927184,
 0.07720467574564138,0.10189396170995457,0.8125073052293217,
 0.6009745349755325,0.9598023381044272,0.9559812729220593,0.9466102257883374,
 0.17755906401133592,0.7950833122699901,0.5143054459064381,
 0.0016866399997973458,0.9868628049012924,0.9201806969238363,
 0.5981638565493054,0.6800780473095206,0.45083505276725677,0.7263947196100174,
 0.377508333088579,0.11338057967089799,0.5411599174908818,0.5776596804670563,
 0.2682508243977265,0.08060191227908498,0.32920379177799153,
 0.7756565243490358,0.5594418258445584,0.0236227696972936,0.1443353759409085,
 0.13315567019739172,0.9920818132286584,0.5155504725475705,0.9667199342788085,
 0.38806024237569436,0.0937013131920521,0.34024138836033857,
 0.16175977524725582,0.29893093455242736,0.18915260841556336,
 0.7726646486222408,0.29400381639086104,0.16144472866694348,
 0.9160839838273396,0.2477849834584196,0.5269370735682621,0.46531657623168765,
 0.6025377012786621,0.19911206490128952,0.626643379896252,0.3790057961824539,
 0.32588959105886517,0.13644604244882086,0.5963708880679915,
 0.19840290625499368,0.06717709372616432,0.627000056743666,0.6503091419025471,
 0.9334696360990529,0.8062275252312188,0.7834123578700748,0.3116279714637912,
 0.4482928748324265,0.25470832680744804,0.9867068262097071,
 0.41011114935278925,0.586035013191371,0.012440036240746188,
 0.5751721839779432,0.13584421297855864,0.3912780060404042,
 0.011730716139793707,0.6224625518465442,0.4217280766558327,
 0.5898522261145666,0.5763976319313069,0.5407782403840864,0.10142791796540873,
 0.7184992460784267,0.6391191545448374,0.6727663926772989,0.6364849364954428,
 0.1600902588900147,0.8026459560888727,0.7873097455605603,0.7321601852976021,
 0.07097825806344005,0.9698468234842095,0.3118835562621428,0.7157655462454233,
 0.3015365150302023,0.8207355385767248,0.6928808427806794,0.8742524449382523,
 0.4241092203934258,0.30583860730441503,0.20832482231983818,
 0.9087527948505785,0.880567809973547,0.6005179186596116,0.34720644148347024,
 0.7563982309315386,0.336923230368927,0.9108548815217575,0.2445550833313277,
 0.5217322113260854,0.3586089881366733,0.31962558654987805,0.975670783298968,
 0.2733948350083133,0.7403537277188058,0.19473843352769982,0.878235346472843,
 0.18601532515029995,0.3346646562279524,0.32406709490775965,
 0.5197785710577646,0.9085512910915804,0.13607427944712658,0.5487370021887839,
 0.0376542363314164,0.21250464945930037,0.7681015744869265,0.6098169810010865,
 0.7565200113711251,0.7619224387309429,0.5495867486047574,0.3005813891915262,
 0.299381971710021,0.6777095435463658,0.40408248805075064,0.6994377952052979,
 0.46098077626601175,0.0868717233105426,0.20237066144445526,
 0.6657468503291218,0.19733209515361194,0.9469895483982975,0.1914247361229987,
 0.189054451422889,0.8295267816451906,0.5597417713388595,0.533428718078874,
 0.03272048064302546,0.14189781983059424,0.2638430153247875,
 0.5135582286965722,0.04961871118099759,0.26872147748251274,
 0.5101709239410995,0.4233269971343393,0.9680711271136481,0.18088992171605134,
 0.4166092823230404,0.7153730137652569,0.12973498399920946,0.8542406661656042,
 0.5274620977074327,0.770827795552054,0.2222178310699856,0.4429276327298327,
 0.35470695851616374,0.3174152222932829,0.22321655913511584,
 0.7122175158639543,0.6080848431887579,0.08760527848417288,
 0.33199386320572255,0.18182739503971157,0.13972686639376186,
 0.7809521565086742,0.9843590145645822,0.11213331152296624,0.6139720441558292,
 0.9477056065909091,0.43447219989805286,0.9412327803314504,0.7976883740527337,
 0.9863461430381614,0.20048387666120426,0.3755930971421264,
 0.19226060669691747,0.1201052168978003,0.18512480162321843,
 0.5577668990572284,0.6698893756008776,0.34397679655373714,0.5803339578591432,
 0.7459270295883199,0.7913212497952474,0.8612556214262757,0.1006832074853079,
 0.42735865326938105,0.659153238796907,0.2193376080113466,0.1747207184000601,
 0.6654154252327973,0.7645561862504118,0.9900608117492739,0.15642454630281555,
 0.34705612599912405,0.8260937756523431,0.11969586464992898,
 0.6238611474398489,0.6609258420438989,0.22043113246149215,
 0.15996221926055387,0.9917867464243741,0.7919158269160244,0.5500402425356067,
 0.022634515185369075,0.31933798420712645,0.4829443079868462,
 0.9631139163622878,0.4288548625908729,0.964887805648452,0.14859382376516472,
 0.5602248794122062,0.3071928634533517,0.48023209904002284,
 0.45968226765167064,0.4352273694074169,0.4942789945934094,0.7956860394940858,
 0.833551277414898,0.9963195334033305,0.026741737650008757,0.5464418609482059,
 0.4274389278890346,0.3147966372784671,0.06264686340443248,0.5827973123244661,
 0.5734417343719159,0.5350225099363343,0.7986023425199441,0.31372311161875477,
 0.7486094389837997,0.5159402718534241,0.10987635719802746,0.1442215814874317,
 0.537090475034112,0.8216339416408235,0.9686104811930628,0.18381697890172966,
 0.34328345399622107,0.3882625824243875,0.4869465582329907,
 0.40941382881068245,0.7028425733887433,0.6467565469153227,0.7960593287408542,
 0.12409417759462227,0.6415533281607085,0.10454934627982504,
 0.3741946298346299,0.21920962495617313,0.49501444850082543,
 0.6251863127486749,0.9686990215633323,0.8766531974604972,0.9774013236807996,
 0.5484045087978151,0.8487921670489059,0.30566348185889103,0.7966943170495017,
 0.5950304018009938,0.16999546053679815,0.5891786485053332,0.3317125710297939,
 0.799896269558446,0.687837008108342,0.21720296236045422,0.7963623646074752,
 0.7628183068644113,0.10686718438888665,0.6095127586880796,
 0.11237414626463427,0.05923513483792009,0.061051379379081894,
 0.9500307025338808,0.25349708686688,0.7359444909422056,0.2162296774836232,
 0.5630741376588405,0.3319418824254259,0.7234229073656486,0.5959106487027022,
 0.6002412567963362,0.09395318509654071,0.8727545434969871,0.7254518810178819,
 0.32660490304477385,0.4447080358967237,0.38440288712478987,
 0.06852471207738342,0.5132306004173485,0.8811028344627866,0.8007236503905518,
 0.9189303593070272,0.503783921956868,0.6738894633519135,0.9507497265634905,
 0.7591663215617189,0.17751800096484627,0.8397440762364256,0.2056592442439722,
 0.3589801242779731,0.6648005838499307,0.19807686611804942,0.2392456051857076,
 0.8827787347355524,0.9160865652057172,0.7637548083658439,0.9231670749778358,
 0.5470063595794842,0.464752445762606,0.1499726716455373,0.7646378971700405,
 0.06801941921554666,0.11050094047714731,0.9920162223109529,
 0.47625721144451416,0.6453854415652311,0.7652605718222248,0.8957639847168835,
 0.31398702628399744,0.07746139050757239,0.6317244509463242,
 0.2898634938329484,0.46163888205339043,0.8184222067795232,
 0.13220498329717645,0.09476693886756082,0.0920851622242328,
 0.09324508896344508,0.03477144110222263,0.6673126647385504,
 0.2317257283229236,0.656893399312139,0.8281970975291171,0.8319716603932829,
 0.5169268618842058,0.5165681143702767,0.44935166889732825,0.8391747714454926,
 0.100605276788305,0.9440362555785895,0.17067209056638988,0.24326042005787674,
 0.22369429443797506,0.9324848194681438,0.6130276651515666,0.8020630404983018,
 0.2676824589652541,0.7394170715814228,0.31393989918337384,
 0.28851694165199815,0.25128545821488746,0.6397155821455084,
 0.06810214121478098,0.8759606092102793,0.5206936293624325,0.7094913228395249,
 0.8999967248729088,0.6355215583399189,0.40109368442298365,
 0.29221582264842894,0.7790261429432273,0.7815481747022787,0.7621499737718523,
 0.6499039186095306,0.5555098137706169,0.006017203933667004,
 0.4761306492588614,0.9371545946360637,0.8379920384641247,0.7944717855655675,
 0.9383817975078047,0.9160598240824952,0.530493319870083,0.9638546791628884,
 0.6774477431473558,0.17464159729437279,0.23763530403030442,
 0.19418877045015615,0.2868529331959323,0.3543206713280054,
 0.33835508871146214,0.39295945210957184,0.7362449086736421,
 0.6497264217619041,0.36009700271393497,0.9908774297967035,
 0.020156189619177423,0.18258050381254054,0.8672752890345384,0.80109283636659,
 0.15035062887864814,0.2053692292495345,0.596893245814043,0.6916761616237089,
 0.9073219776850956,0.5658896842512839,0.6970993834874745,0.37443036138134655,
 0.041789624703040706,0.2090765795291718,0.6939634610880506,
 0.9997591295228383,0.698237795514434,0.7106452680509348,0.5170024023709873,
 0.11408312289517819,0.6686468712727894,0.3460416097822586,
 0.00477762651747482,0.12070434741447511,0.0343261929829608,
 0.8501728190262968,0.6305513384456949,0.48479689434374784,
 0.12601375951437932,0.12611819805952218,0.5080242878353778,
 0.6584907867121443,0.923195670810656,0.8291582714121235,0.3467915487826966,
 0.6310645670821561,0.43164013933578904,0.7431347190722297,0.3465104537230532,
 0.21113717110438202,0.4153006648142975,0.1795501822091472,0.9665023336570051,
 0.3232738195256216,0.4609187283422257,0.1400779257498781,0.2957933353495713,
 0.1864121698195862,0.20836626305666617,0.684274577605438,0.8132346817604743,
 0.3818727525083473,0.20984905618849992,0.3829894793078096,0.7646516046835472,
 0.40205606638217073,0.455594258145365,0.14069988347251228,
 0.33872071754040844,0.2932125466116745,0.8612925933460932,0.7458684580604504,
 0.09019221546989042,0.06333797404743646,0.3826226704476744,0.495717169027285,
 0.07937336851022547,0.8402022651527332,0.8627728143002531,0.2980243752980295,
 0.24363764281829803,0.9149126962926909,0.7189015169540101,
 0.024095038201540464,0.551040550118961,0.04422346451962167,
 0.20182333730438362,0.12991566320559167,0.3078069777863809,
 0.2729272106767733,0.6490117061334814,0.9059792649827014,0.6202226190904627,
 0.5885474862150379,0.9254168487845594,0.8097799152940132,0.7621848122761087,
 0.7435523050842868,0.8976838008602677,0.37199890542120295,0.6474575720364165,
 0.26923714341293936,0.7232681640741778,0.7553285241032541,0.1679587506043868,
 0.967108798949788,0.5875471555849772,0.556697584630748,0.5271899061411117,
 0.8723814033616116,0.2690310098596477,0.8537962938180967,0.49148968869666,
 0.9628168903074174,0.8727441162733515,0.8675452583451122,0.8806127051064613,
 0.9794954618517364,0.6338228199608097,0.47928028617550233,0.434628769609977,
 0.9094453175241957,0.7073551454886663,0.46320527051597193,0.6953797535984044,
 0.761309427438185,0.80553052780897,0.38338688240049934,0.17571537482340593,
 0.7143140375065289,0.509358321241652,0.9195171486508292,0.5232388059385032,
 0.8868877593348566,0.3388249128859132,0.04153889861409177,
 0.08055024085466012,0.4420533722707546,0.6395678829258377,
 0.15272232562174715,0.38164286831208627,0.4110444227459691,
 0.8199431705002641,0.09366438223357054,0.7528990558370523,0.5786561417786344,
 0.96377905084067,0.34847154903207034,0.2083712501849907,0.40380362311780704,
 0.0800235843105096,0.7195970775641868,0.5619333806489657,0.6643795369150154,
 0.46867391074598064,0.5415159120073981,0.3668282001897123,0.9716851832716846,
 0.580442725968542,0.5734027631794745,0.7565949718003886,0.9678712262589918,
 0.9452715563959045,0.4366530420751077,0.19899407363078692,
 0.40782233206412766,0.011068952370241858,0.6322310459288205,
 0.5347235888564809,0.678947697811086,0.18358904168325574,0.46249848715218744,
 0.8423837237260371,0.6857750306336046,0.19729176598691578,
 0.11727191765335299,0.6773822492447007,0.3514410740805388,0.7919904666714018,
 0.2681642816518849,0.5144196347520923,0.42439375546759917,
 0.35236513974127515,0.6975987498494828,0.12413921487150925,
 0.22123177290442408,0.14607246512696304,0.5364277446688925,0.517579858248877,
 0.3899750597947612,0.31634344074502596,0.782055707069221,0.748906732358026,
 0.7356621207070896,0.009474340993372965,0.7634532529428386,
 0.9796327189695565,0.2631357132842421,0.7659900270857797,0.8138534659214768,
 0.8516883522421183,0.32766757193969864,0.6326194220031622,0.7070467695298975,
 0.45088380550185714,0.8716493343619242,0.7678872135857899,
 0.11421939384624957,0.30162796966907723,0.6921189825010761,
 0.1656760488150547,0.05719117589628786,0.042853767424950506,
 0.09788477413765873,0.9634370829571856,0.3957289080848091,0.8921903039156303,
 0.6874522789279821,0.11476110241725856,0.12950730658747123,
 0.18416370784404212,0.9861787139350547,0.15100245391712996,
 0.32779137866025543,0.3447814098442178,0.7388505034477718,0.887613202511207,
 0.7298032705445558,0.3404157106609941,0.9251222944055917,0.47035414862646474,
 0.4640593785187195,0.12739332584995722,0.1607521094336567,0.5956080966525854,
 0.5905406921324355,0.47331229706905875,0.06396629310004187,
 0.2380980197927558,0.8656128388506109,0.21219906077796802,
 0.10044199730796044,0.08673233665123592,0.6567425105327239,
 0.5030778008241023,0.957087841309444,0.151115813297504,0.35108475269209183,
 0.9331437820512158,0.7996757288071392,0.36453005628228163,
 0.43063749673062945,0.9623904760252635,0.30735114700659905,
 0.09191843437664071,0.2889582836903557,0.7914784346115206,0.64950682490781,
 0.6027578424512094,0.3100069710360943,0.20752330454327572,
 1.4702035532698776e-4,0.42520244491454673,0.7742082854732916,
 0.2851763227234537,0.2333728822462936,0.5033229223623651,0.9721232605618028,
 0.4946794898909467,0.500591194935597,0.4545972993371533,0.8771300309784976,
 0.9030525461916132,0.5307658840014956,0.5532978571326348,0.2298692520283987,
 0.7094438384309973,0.1666274012093346,0.746310874572774,0.7872991420062702,
 0.2602390266867811,0.3654956137060783,0.8323784000047292,0.1552112268583613,
 0.9925693864474043,0.06321787512644061,0.6934042975148672,
 0.15503951354774648,0.7472412119833168,0.14109122896749637,0.565659044128598,
 0.825042991729353,0.24539929732163257,0.9856558048715401,0.43930229003898424,
 0.8662777260214504,0.8336729142468869,0.7678626773855586,0.7156904313340843,
 0.3432549819544144,0.9394527479540807,0.03759541720329962,0.8199253190911959,
 0.4006602040620354,0.029943998029236196,0.6853655592600263,0.848562582204301,
 0.4969015523727246,0.5544546715409043,0.7590212238442938,0.34245971616494897,
 0.48346522694541005,0.4602532252386584,0.17783424594822583,
 0.9119179557930723,0.5706593018800973,0.27775139724165254,0.8930930783913635,
 0.9153119272376632,0.9251526483901642,0.7379997003187367,0.176724600783566,
 0.5565712957795671,0.08527036913560548,0.15457193503461486,
 0.10741460188230623,0.17161085342915827,0.4671017714467195,
 0.18731134458431853,0.7292597061900851,0.052453904008559205,
 0.5325124354500707,0.7811590184316912,0.04264540732800137,
 0.11238241729557485,0.2426461786430435,0.7954588016620758,
 0.04823238740837521,0.9414635139135435,0.9673013222338555,0.2830406343297607,
 0.5755145511090167,0.09409529537967809,0.2954751283325131,0.4016417321852739,
 0.12501391867387535,0.010256243383490293,0.12238079647687583,
 0.1522927859500176,0.4391263230787573,0.10842934980831775,
 0.07180645945820097,0.869394925689914,0.7886871699461651,
 0.062239705932651046,0.19658241018648592,0.9143274733325703,
 0.846743147561291,0.18467355662621765,0.9493990065603709,0.9550587999282476,
 0.6431178094741636,0.004503878478358314,0.7949452721049377,
 0.8892646910656367,0.8200690052022336,0.798446896520794,0.5262041916110834,
 0.5662009917422259,0.3760321610162024,0.039452984514153355,
 0.3204100848609761,0.1538427535915129,0.8773261251703972,0.7354685866429548,
 0.6132246569194995,0.030656684894794783,0.1200860601231516,
 0.9240977246716163,0.8110827059403329,0.30371931034547417,0.5926835262001094,
 0.13214700519221623,0.3528665197423625,0.33365579164706016,
 0.14872227381366376,0.021594946747169552,0.12998640639433312,
 0.19405329945552452,0.8363270810242076,0.7643043699332528,0.6995167089305658,
 0.1540333795942126,0.7429653930074073,0.2330549881816535,0.49659891495329,
 0.42366944148377006,0.4179161934722553,0.09146046142344999,
 0.6424774551253882,0.8532480978723332,0.07792478022241967,
 0.47175135813932423,0.7244880593880697,0.5274494078494144,0.4490222316834479,
 0.980503254463026,0.8939761374708407,0.8911389237839993,0.641290392790511,
 0.06257838015225137,0.3583627781980371,0.3670130708388798,
 0.11881317331423458,0.7788794660802607,0.014718357179127728,
 0.7280210432892594,0.9869743704037341,0.10702367278169267,0.9048056370305376,
 0.8626311424228125,0.29564652847900974,0.7444530044044693,
 0.40374246969465655,0.8210056341403786,0.9197763011806683,0.7897711502979636,
 0.01564441455226162,0.8329203637360475,0.9736205638706287,0.8682695997803436,
 0.5236457565820903,0.45490717503768463,0.6349263261605181,
 0.15294871139121835,0.16534098712227707,0.12999757355057062,
 0.02608848140316644,0.21060775435267842,0.559322692045785,0.5012888233694132,
 0.8274944205227014,0.8249880544890135,0.8866802288966331,0.20256585182987064,
 0.49120144794166165,0.9804946346659182,0.7627943703703034,0.7076251293747211,
 0.2083496115207133,0.6647018775153855,0.1102897808449621,0.2511978951298638,
 0.18051992570994246,0.022760803681500685,0.3979116361585011,
 0.29733551873457065,0.7444400443028585,0.2782428653434218,0.5437182349570516,
 0.90040882786433,0.07308822933073988,0.22862369148919948,0.7995654021036124,
 0.37683101321568524,0.43405797139476976,0.6575520047345593,
 0.06894167215441194,0.17008504140024516,0.40879480710705673,
 0.674298903184454,0.41839976809303703,0.24402406142656496,0.9489307238424702,
 0.5145897196248972,0.7811185105462911,0.9272858128519256,0.25725143757072877,
 0.9911080511975908,0.9962246840178214,0.3688391476348166,0.25323792246632826,
 0.6193305884308198,0.19418747527889635,0.5295084844804999,0.7767879255795762,
 0.39668018346132194,0.3730106419186675,0.35795176149563546,
 0.6456496866765615,0.9883263095988696,0.9653327150036686,0.6319822723370133,
 0.7092823462189212,0.5797701362807357,0.0043524385982441904,
 0.4196449189103719,0.9183079340534765,0.4977079953404653,0.650802322440969,
 0.8384163983988631,0.41452244908403746,0.5716858817354844,
 0.27962873148688505,0.5477739098355304,0.20256014562657276,
 0.6182377911663977,0.7533014438356838,0.8825851412777077,0.2957383713321541,
 0.22886173545298016,0.49784722694903016,0.2271345321221121,
 0.24637319297845028,0.013328200418284752,0.72747378700797,0.2488622874454618,
 0.5881357743441331,0.09325239048899125,0.36583381560216155,
 0.5275914354638254,0.9029080932205454,0.6880914277898834,0.6194612724264514,
 0.5857539301058328,0.17395873641015536,0.5212786009747472,0.7964575476669962,
 0.13066468198941472,0.28643058271040145,0.48402835965100977,
 0.48898777057962317,0.8702243636883634,0.17169267248591613,
 0.6445460055023139,0.2897834266597604,0.6536241508161413,0.602799606369696,
 0.3791235958277346,0.34669543465984076,0.8534103314339984,0.8752779810351043,
 7.052717051920787e-4,0.9399166706743466,0.4468789880845607,
 0.9597768549409688,0.43399139659492436,0.6085092626621674,0.6567097439824107,
 0.43311958128369366,0.010236752164524132,0.9146441633231102,
 0.6026618391745316,0.30323325740767015,0.06506940376363313,
 0.38441769020042893,0.9839183343393302,0.05823752214505795,
 0.7702298673579979,0.6654517452937455,0.5976271231314999,0.6796660033401656,
 0.31024065551405156,0.9928983563442533,0.7614438638808538,0.2330510419216284,
 0.8091055072034516,0.6902206114297116,0.5861603684712909,0.23236936094940097,
 0.20199336237964494,0.654142908770013,0.42747996242656794,
 0.05131022383128725,0.6268168883094949,0.9255520710458881,0.6749416101560757,
 0.5522156666499892,0.12549924166778614,0.7221536328719765,
 0.29990134439623484,0.8795412465000921,0.3680887239092182,
 0.35032513159686074,0.026621694929911954,0.4070532415623882,
 0.8935691069223334,0.6444601862585945,0.9538895064409336,0.6243135244121257,
 0.5285767913251583,0.8107428286880722,0.8877800724849992,0.8876849797911599,
 0.4315098892832018,0.1348471814858173,0.2525912360125926,0.9770841917537534,
 0.4694163816051906,0.24770935881737077,0.2647986926032445,
 0.24980703756046363,0.2640199860862056,0.35657082270869944,
 0.7288453062630481,0.6283909258539148,0.29768469402167774,0.600275100275182,
 0.7270463818165027,0.6846614677866711,0.12661714922776712,0.3457093296619913,
 0.8776145120843604,0.8826808834784764,0.34456398069641303,
 0.15642047438457385,0.20069236255722278,0.5201838017821414,
 6.506481090607918e-4,0.2953701625604521,0.9307459667773004,
 0.10240494079079954,0.7452064410084455,0.4906277242975582,0.7737906287885505,
 0.9419456446664423,0.6931675849897978,0.2000401257988642,0.9082306213591875,
 0.7448043443737566,0.8850212906554286,0.4376375172160871,0.8289616480550617,
 0.40825157907415544,0.7943825013684253,0.9093304342288517,0.5469495349128507,
 0.5567145987177975,0.41412350824620714,0.04090909861938974,
 0.2913767240762992,0.2366102297809498,0.9687980339887974,0.8908342309192979,
 0.8170344540521217,0.036379555180777845,0.6884125975157183,
 0.6280664992051953,0.6122850041887673,0.6575047798936331,0.08662625768802479,
 0.8308891773025088,0.7776681026148041,0.08669547684516277,0.6625788098890342,
 0.4236312274760079,0.8527946040719299,0.5579283794841895,0.18556734789266394,
 0.852084635937071,0.2521771344982511,0.32400043878563756,0.7359171349298125,
 0.5226565484142428,0.8792985655402777,0.771435491379448,0.8584247384649424,
 0.23589346010026535,0.4768184686851721,0.9505022260442496,0.6229598105432864,
 0.9380366853041927,0.8418964947307508,0.35047266227960827,0.4603563941030788,
 0.3154814057948363,0.49233636893928523,0.0455547422477387,
 0.051439943839716284,0.49264478728626226,0.004350521237622806,
 0.672257045508143,0.08040776610174827,0.21250089465263022,
 0.04872994582138057,0.11921766273226841,0.18529305440247523,
 0.4585378592259297,0.6727125196755913,0.013495366976016276,
 0.4492922671099552,0.29477892807743256,0.2248901756931635,0.9657158385864728,
 0.7347859554717531,0.7805410549244681,0.6928003811032648,0.16316466396928087,
 0.6634027149445515,0.7552322767426679,0.9368031870994361,0.12390808917688223,
 0.6623451640047504,0.5544928299112751,0.5536372341811298,0.2252106047137017,
 0.30875827304505143,0.06619433303322575,0.2848943000028503,
 0.7405542638724177,0.3598884186143132,0.7648697790173332,0.5941129599498423,
 0.22307692952578984,0.010659676733696943,0.19161297211196648,
 0.4863490100806632,0.36370639691998574,0.589904721669469,
 0.015366196823911049,0.46348393089675577,0.811553059429338,
 0.6997053082609507,0.1830070220090234,0.5801862046092944,0.8491380945714473,
 0.2582732955560233,0.966315419959206,0.14699904631987126,0.9609615617208526,
 0.7117195020281426,0.5001740822509052,0.9457710057508991,0.17769563519276466,
 0.7479538486102155,0.3328838939874629,0.910907037017743,0.8574980305820576,
 0.6863772288027508,0.3392821808584213,0.6390970079965695,0.2739976064883498,
 0.10151325946965928,0.029960094582750618,0.07555812444522758,
 0.8405155902469474,0.6702211909559898,0.1524515740307697,0.06628751559935753,
 0.20791291357195218,0.14023390944027803,0.895140049757281,0.6380840887172501,
 0.506140797429541,0.4656278466078706,0.15737456527938276,0.2015700218532105,
 0.41526657961585434,0.07838348594540112,0.774180581812378,0.5431136377246212,
 0.5905889192158517,0.3665793977675861,0.5327413186688935,0.33421836699810803,
 0.11058099301626367,0.12431169769647565,0.5857696270199989,
 0.9381160608809085,0.5152857618980283,0.936142784806991,0.9544395708123516,
 0.10077300290337321,0.11300206880206842,0.6801616121517295,
 0.05372199821337709,0.880891875541087,0.316199368584707,0.34920970419923725,
 0.3495708007109689,0.4782991007028986,0.8743808552445244,0.518073314876504,
 0.47794147122724184,0.6393790814779565,0.10731741523550087,
 0.10004356289527727,0.16228312625242403,0.3531698107398753,0.151777564874659,
 0.06728616228886053,0.7825081615086715,0.09336040171182314,
 0.3273427547406865,0.2908869120717028,0.16420890447573977,0.1327184568279357,
 0.12335953911951625,0.8771843663719581,0.3867992907698944,0.5914465544163399,
 0.5723783044142474,0.5750936147886394,0.5692623382889612,0.6730142731305875,
 0.3709048002539681,0.7072223095534306,0.8243908948124026,0.6844778358050684,
 0.615176407958722,0.3365622288229879,0.9537675988897351,0.6012807244580499,
 0.19744228207930836,0.5647032328226246,0.6479565376924502,
 0.38829679018915153,0.691514018845566,0.5068357837470163,0.3597269358920081,
 0.9465229805992974,0.10822987408300544,0.9684392361848646,0.8405974352178658,
 0.390437131088486,0.2803175227232704,0.4811738955477223,0.7859219223311443,
 0.6265755412556581,0.7807874732002363,0.46781862327182566,
 0.40307850307218285,0.2504038522190336,0.4589361612795336,
 0.14961306164389687,0.7497330861854565,0.36790834847964937,
 0.7598337755402172,0.703126783903426,0.02307413000046854,0.8833871731946713,
 0.47575342536545195,0.927558731066723,0.9409885790699483,0.5348480402270785,
 0.6618184569847293,0.6348226799102025,0.07873863552004301,
 0.04860294630959605,0.05509904089811701,0.32076824044335805,0.28362248822434,
 0.5798827286478977,0.3608157466868076,0.15294015111631554,0.7684730119724512,
 0.15416010684900494,0.969422335881497,0.9855671563522832,
 0.011028406684364334,0.1835752981281249,0.9349481425080401,
 0.3287066378345045,0.10023330650041296,0.10606966087877412,
 0.7772902775287645,0.9257005437915462,0.2670064291523748,0.8634648226122053,
 0.19117271723093743,0.9382218520089082,0.9970273436017804,
 0.18509244496827315,0.13843179222455027,0.7322669440847485,
 0.6688018718997518,0.046121527258145445,0.4670104042671711,
 0.9374353166766102,0.9536471177150195,0.5213775204354824,0.601582233682173,
 0.5343934600207126,0.7503537221498021,0.8519091654860201,0.5585372886872225,
 0.40482373702422125,0.1997589998495013,0.8582196099654256,0.9507582534933616,
 0.014222348820748909,0.9474519270743348,0.9205773917125915,
 0.9003836022576404,0.09952884801495432,0.1873373203625157,
 0.24167155963086673,0.8708352059176387,0.49335115873705593,
 0.046095258984707455,0.051588440126233426,0.10830342969366424,
 0.2801193420695698,0.9789500682655023,0.2631733957933988,0.8734887602283088,
 0.17816752176809625,0.5455263866253787,0.9123010890336593,
 0.45471711969135853,0.393291190316294,0.7709973624508537,0.9844542226520692,
 0.597178556764107,0.69195668828884,0.12085564753552591,0.351370755528422,
 0.8204863516462404,0.8005826715428509,0.5232393535114761,0.550696497604585,
 0.964041025310815,0.974559077793709,0.17625072049770396,0.9375650015230416,
 0.9739155896210029,0.9086516932028617,0.8867254360689012,0.6476741169384962,
 0.4448337089054911,0.85620293837979,0.5122775658448707,0.13720226269556945,
 0.36912583943145494,0.1423477888662481,0.8445258900969563,0.5904812285203598,
 0.6133345295366006,0.34768467434437156,0.8884998825892416,0.6697361382092716,
 0.7756319043728098,0.3680904959586422,0.9525885969805893,0.9623723941667599,
 0.43748835101230354,0.32318799000520515,0.9917913335951761,
 0.9111238169179672,0.5735512849980167,0.18458790000288883,
 0.006517221281024277,0.11270767152487093,0.5801769528423757,
 0.6021110905498428,0.850916916921948,0.8140847346273491,0.500939783101613,
 0.7695351506750407,0.45093089170240097,0.11284596748365527,
 0.6090427091326808,0.6854209573120403,0.45723580023099963,0.3226201985131104,
 0.049389558930269306,0.732207068560026,0.4721576885073687,
 0.33268114767712353,0.6990307746988539,0.937664253453351,0.9039879845975725,
 0.02894045344497642,0.74873813057252,0.16555461093537294,0.7358551632475085,
 0.579631876019475,0.8712502329024474,0.9915333206991173,0.33406610159091377,
 0.25668750858555,0.458645963490142,0.0977867641414466,0.2488626485815102,
 0.3308347512054921,0.6784090737036158,0.754013950660374,0.7197309404508204,
 0.5911564790986972,0.5542882079280043,0.8913242259419404,0.7724171949950714,
 0.6276590006583702,0.44078955586193014,0.1901740212768258,0.720932683898142,
 0.7649331013738965,0.789393488155575,0.9468867763863771,0.8188775561057257,
 0.41545109878750497,0.6816289739241662,0.701092633848409,0.5378424026896084,
 0.48512597866189344,0.03788271717124092,0.4838415495002837,
 0.7493305551132057,0.47664440986864354,0.08666282044697782,
 0.5636941576336932,0.06941272128010145,0.989529179288126,0.7640710764975729,
 0.7938721637190178,0.9015562235116467,0.06889142260123582,0.8173998526459956,
 0.892063297121875,0.04482863714045149,0.7613804912352736,0.6002905314657868,
 0.13830617057152694,0.08097852438241693,0.08145930540935276,
 0.8666423900654506,0.25833215690542577,0.4264815088670115,0.5976124040803125,
 0.15517139355576737,0.7622446916656497,0.67558537403823,0.7939393482112787,
 0.23322326398328974,0.5295786471694166,0.25534487264008754,0.838657343513433,
 0.1934215324861981,0.5256560834307047,0.36448627124482524,
 0.33982985227162943,0.7373123323465463,0.6836133264999075,0.6393346779857373,
 0.23957822024878594,0.1706102763152697,0.09159348712950477,
 0.1748243415546309,0.0674379065736459,0.05061474231838747,
 0.11744633003452443,0.8520517661826216,0.21605666540307333,
 0.018594601411411782,0.10748743021853435,0.40482385270683885,
 0.5447306390781224,0.2562691769169121,0.1528753280168773,0.365009841024564,
 0.30158490370553515,0.18891524369197832,0.5356171322097105,
 0.2656710850837909,0.2731729948783639,0.7810469270763032,0.6724175856944121,
 0.34714237693053895,0.29490674418979224,0.6459149551052434,
 0.034565546312200435,0.2572523243457484,0.8276844719407639,
 0.8119111391462436,0.8938187417609245,0.9190212368919232,0.3493342761232133,
 0.35303558642413324,0.8000207546643848,0.3172554219108866,
 0.07343232838182512,0.6447976810682575,0.17231490674210892,
 0.1355089132117152,0.9869510558157156,0.20689536405230657,0.934990502775159,
 0.7124127604697964,0.397586155462065,0.8443938644417893,0.8955069054913652,
 0.11414343638715596,0.010768539554006473,0.700323888116878,
 0.7484964429625438,0.3355973634487581,0.1576041180453085,0.42735376488374577,
 0.18131332142500023,0.6868716640944421,0.687178471722492,0.2682243844590426,
 0.014251560528552032,0.6189925228545515,0.9066522361706003,
 0.4983785624986289,0.2203389469483399,0.840417109154337,0.7364296001128711,
 0.09180407215998176,0.8984468737528867,0.9648694916380552,0.9450852402182011,
 0.7939114800258734,0.5315335568761757,0.35871081764300283,
 0.02414135457081723,0.7535615627445263,0.4957651472603912,0.2548841313716361,
 0.5718763639194726,0.441719048555099,0.11970898288014176,0.40378751288605597,
 0.5181515829461505,0.11245764100500866,0.9466650808432674,0.4882371604718426,
 0.7096429603690652,0.8454987039735882,0.8369924077012564,0.2507369621352131,
 0.5526399872092556,0.05639949324005378,0.8527633985276205,
 0.07853898882574817,0.1075218848426217,0.46071835975034126,
 0.4757252506131534,0.8683943267634238,0.9467958584590934,0.08999961374031062,
 0.9980976731035566,0.13278123469669778,0.5398511098777024,
 0.17879215826597206,0.09212751364577176,0.04451503248946809,
 0.06441681173554703,0.826097459116601,0.3144742993969688,0.05248347004691878,
 0.5817932221024528,0.22959521146717465,0.7191606557755554,0.5712142285740266,
 0.5423041595976048,0.717043027462776,0.8196346186257163,0.15047710207256215,
 0.9332813103396846,0.8784486089310688,0.6477314262858183,0.7856876590292878,
 0.35233616083758557,0.3474927642198167,0.462177957051926,0.8413203059902782,
 0.9151333111195374,0.6126012408338783,0.28090032447431257,
 0.22827613724330564,0.830757039287461,0.13828274896553205,
 0.17881251948975263,0.02601663583774716,0.04936467177950066,
 0.2044582022532725,0.07838715752544834,0.9676736979467637,0.8594654886190988,
 0.996427725763066,0.20420863563065028,0.4881788346176552,0.04696802727716509,
 0.03229187107289777,0.12265703890158974,0.4717802837928762,
 0.12249362744058756,0.3550240477145,0.12886654812570497,0.9447396682841875,
 0.7776964981106571,0.7838063004999899,0.4269329773945595,0.5442107081521357,
 0.23937847958783365,0.16615824501028398,0.8741781872300578,
 0.8079101144472048,0.3594572064391748,0.9795766110539862,0.9292414549392439,
 0.041755579041018454,0.44447166094802126,0.26478248855507347,
 0.47160773049246474,0.28544402061467344,0.26112942800305805,
 0.4883728031279888,0.9535435530272837,0.5382243361744727,0.819736458789233,
 0.060230006741361475,0.8850516052066394,0.8019794881527557,
 0.038088801189975596,0.6183997797371775,0.29689969922662196,
 0.9031007062159775,0.5033923989102599,0.33063960945101134,0.503145356001871,
 0.7295211785020207,0.3793826942393981,0.5787184629612172,0.794255476789951,
 0.430703234797165,0.5930162786591706,0.22293535883769822,0.7948889425422107,
 0.955142838934507,0.4797859952887702,0.8470634905788422,0.9600210613625184,
 0.2485270786472108,0.4184912399330254,0.3876513868650979,0.5912327873949234,
 0.9934921062596834,0.843831929111914,0.1933293785258482,0.2688236664832433,
 0.48215714502560536,0.9236172829699217,0.5726223969092736,
 0.07825132614280461,0.40143799627865295,0.17080874873442708,
 0.013539259392704395,0.022917108553155097,0.49700953196790254,
 0.9812669699782682,0.6277012330198224,0.31247164494240554,0.7146205306879665,
 0.1612503706293541,0.4668371764232008,0.7194685270674657,0.38018203126935934,
 0.5957768866978231,0.49944608710874416,0.49633637639443684,
 0.0922992005493849,0.7806599109296634,0.7776092059262774,0.8413327361699348,
 0.4577254600123085,0.3653523145596338,0.5445751745115934,0.13164441620094247,
 0.6042056461576657,0.5478544024598265,0.7434170459177916,0.347322221353461,
 0.546262401177862,0.7234560452040886,0.12428971953071,0.7569503510456953,
 0.6845822151905632,0.23769031390638407,0.7798314220509719,
 0.13719402769144773,0.8558054178736277,0.8943782427991251,
 0.09410426522376414,0.3261863259719432,0.02577081310960727,0.552441274956665,
 0.8224228319878604,0.07375892613571244,0.37321766202799733,
 0.3866119509929986,0.37337421432529716,0.36529889390809234,
 0.14045062175521872,0.6613088468527506,0.031103252687381677,0.50286040000209,
 0.2114881253053148,0.9902057271805873,0.6695159685506069,0.6043917740140204,
 0.5636984784381872,0.8101855735992513,0.6101281251449956,0.5630877279818127,
 0.3616057012334051,0.35398214068927114,0.2212605513132263,
 0.022377195352222934,0.1420214564633514,0.523406051420975,0.5178646048055625,
 0.45137806151207605,0.9513817238948264,0.48697228601703824,
 0.3544078088197029,0.4875666594101742,0.5448792881723655,0.19971205961831373,
 0.038672142178834745,0.9202858932382876,0.20607530686287912,
 0.1450732273595552,0.6181442452883213,0.5062125396533785,0.32444451322630485,
 0.31070305291789513,0.938450715630022,0.8411560472987948,0.7425926680743904,
 0.1977831798070785,0.7906070605133483,0.4386994250631244,0.4557133874282595,
 0.029321291473727307,0.8981604317662005,0.35711905878278916,
 0.1533696151376026,0.2080981391245541,0.13093500185163742,0.7836159678261255,
 0.8465943062411702,0.4970872770714144,0.15772893950334754,0.797187193692727,
 0.07443394976456075,0.3503814325387913,0.38378855814606017,
 0.4568091004086722,0.6865855639323724,0.6126917314717527,0.6589709129625327,
 0.631700130852865,0.15988485187351376,0.3831737963310443,0.391975113979634,
 0.794526166080869,0.37920360675991893,0.8058013600846096,0.877372918489489,
 0.08298415802637082,0.5877579594968297,0.09552367862005695,
 0.5748862843783138,0.47437016761243533,0.44829111727809223,
 0.2863060726496107,0.6579426995569265,0.8170231796070423,0.7626674335412603,
 0.9204210863585058,0.6370576527107231,0.5260890819253298,0.34164977620274084,
 0.6252752423736044,0.8246807535391935,0.9339999802251955,0.2623002986998155,
 0.9624545477203581,0.6712435055833811,0.4023496373743115,0.9130001835927111,
 0.6047682216502432,0.7890414621627773,0.5978643415819327,0.2199945638247114,
 0.969131736951773,0.696158917895644,0.2643980596404557,0.29721087837167925,
 0.728187989532999,0.8577848987293384,0.4000053304301644,0.9958031507203624,
 0.5167136623918409,0.66615837917169,0.5512911985640254,0.15544956447438263,
 0.5161973996187069,0.17008832344911384,0.7947262741646764,0.9426934334986593,
 0.5790580340863505,0.9932144181199861,0.8551294831060282,0.8393030575134063,
 0.3349977865588787,0.6619955909956746,0.8196117842955498,0.0204880484756379,
 0.841006330471563,0.91156013852031,0.528841403811096,0.28108140611197063,
 0.6670412523219738,0.5490929613603857,0.48037449655116615,
 0.017672653993340393,0.47304923311596014,0.5274898249221837,
 0.38611518212246165,0.7192064298543016,0.9352832932466691,0.7658161660536635,
 0.6001045901966552,0.2861215628506074,0.03810433252998524,0.6003741194310845,
 0.4207316253581409,0.5695485876871913,0.5818380214369508,0.2896285871391614,
 0.048417682358816494,0.06580723999540494,0.09545455455155216,
 0.620652260646379,0.4191688116756829,0.7853865027986712,0.8470522739827543,
 0.07390387750475425,0.773126085575446,0.686264369713722,0.8520124939169941,
 0.30889985379281626,0.9694799800513288,0.47619817422372057,
 0.7483208258829588,0.42285404924590386,0.7694532254055465,0.6245910467973492,
 0.7757230840966122,0.8680112835752494,0.2661981998120444,0.12237354405247636,
 0.035992511681817385,0.37704353116275147,0.6531639343915538,
 0.34683183575821863,0.03634022151453653,0.709419001536869,
 0.37771889853489904,0.6682941433592264,0.9835861350593714,
 0.07507143871788302,0.3147802614376204,0.08431063673283334,
 0.11195499103278561,0.8796538479715587,0.24482372353890314,
 0.7678377247402272,0.21216457747178352,0.2848912933276144,
 0.31377024216312677,0.10581250888927685,0.9710045385413475,
 0.7367466737016799,0.4116916272783796,0.7752912786275848,0.6886848673312385,
 0.20138170946198508,0.6142092430781686,0.4661699985712364,0.635266336548932,
 0.4286209759295687,0.17863253245684418,0.7835273111400672,0.7685480737369368,
 0.6714384721570381,0.05913269483541528,0.839506144188617,0.37066176230361125,
 0.5502418197274315,0.8917777872267475,0.47344655092883503,0.695804840940086,
 0.42121359896817834,0.8361345109049899,0.5801289906978366,
 0.16259369377292487,0.013538334538535257,0.4947652827870824,
 0.3506751978549588,0.022255038041431252,0.7090128483978181,
 0.10848322164163882,0.17591838605513654,0.40790812420475386,
 0.07340474965592358,0.929011697057844,0.46189196451893444,0.7812328753476748,
 0.8771247435184102,0.6272798824345147,0.8772850853663233,0.5292688176134442,
 0.7805117627972942,0.6818481589649216,0.7132034742794123,0.9368936500109449,
 0.8576805462283639,0.7403410133571069,0.5033432766655316,0.9542096439774368,
 0.4000095750824195,0.4399421186386926,0.4589491966256348,0.32285688568947357,
 0.4465717196688488,0.7077439974624008,0.05726232714318791,0.4923741604547882,
 0.7528582293979267,0.03839815121599244,0.8156201048220328,0.8972481064536673,
 0.5261998533249014,0.4831605263522255,0.3017393836186417,0.7205588594360818,
 0.21896884956641982,0.9575990343984155,0.9389460590812025,
 0.03216344810115668,0.42009562657308697,0.898736753256208,0.734971255619579,
 0.8449738131916904,0.6646909069893397,0.524764266726218,0.9275300308725519,
 0.11052401774530396,0.7415605191302612,0.29735213603912813,
 0.1067116288304576,0.21649430506191714,0.554397021348093,0.8426532080520398,
 0.618809381773779,0.01498606713365358,0.08926475924550614,0.5168359609002379,
 0.5688736192947822,0.7418077061152328,0.031325592170700745,
 0.6142350067839347,0.9304638898545419,0.47441373443746904,0.7746483113950712,
 0.3743691112054077,0.9250460060050887,0.1193199829629249,0.23887779928387476,
 0.8935211725795715,0.7933175258259115,0.9010331006171127,0.45343515641409116,
 0.34602747995828653,0.3935952880482412,0.4251425715748107,
 0.16519469393431496,0.7652975759926303,0.8973577450897388,
 0.41551500945719877,0.7747257515069305,0.3867801424814903,
 0.17411798122731015,0.2794628070816877,0.6317616165208325,0.4046237796611267,
 0.3293072048778014,0.8336549869124803,0.26705701141839666,
 0.11899796368793103,0.6555597654338096,0.5133283903976289,0.6142955425351603,
 0.47918240877379403,0.04742783222446345,0.7999701788780966,
 0.6071861579722224,0.20594629039844115,0.63216213439754,0.5026116043522937,
 0.23181781423769376,0.46545324507180785,0.3385707485102937,
 0.8686585747778555,0.7594594134364269,0.23855864303969798,0.2764920024227059,
 0.05358242889032949,0.3848232215944213,0.31900858156682443,
 0.2907517744096144,0.5416906895021283,0.5471885934547288,0.3357352761490707,
 0.42745723273406067,0.9509177932263116,0.4381297790502531,0.4387972977875396,
 0.9437716480813997,0.32800646544145295,0.38460861240533684,
 0.32579860579121234,0.6991749739340575,0.4077285454651898,
 0.061515205250251226,0.5233954451192951,0.15480791734324317,
 0.5242494820770002,0.9111732764478193,0.8192133478288373,0.14284074232270805,
 0.7908197435758941,0.16442051699980798,0.34347246852603097,
 0.07068639342797867,0.6677403067951084,0.4095561896895127,0.7472064152353993,
 0.44252452633800865,0.5700329794831858,0.15571467719669607,
 0.33967530931689116,0.8225337353166352,0.322216445692181,0.2788697206299374,
 0.18365022845640622,0.31226597173575454,0.9283540426221065,
 0.04475855051887756,0.5380355532206745,0.9485812640659607,0.2457924650069383,
 0.02936575913561401,0.2421752305446392,0.7340426077519283,0.368540408481402,
 0.12676284612552058,0.6356573615688507,0.6437986660967713,0.7832977220939192,
 0.8075022613382459,0.5500526733114611,0.003145494211826083,
 0.10752235614908578,0.9650935193778093,0.6983277548077833,0.8099357459754917,
 0.852110590768365,0.31174848252656706,0.8280885230868106,0.9573949589147107,
 0.622761155899317,0.20443426211783056,0.6690914795799985,0.9231092478280556,
 0.9797892103110122,0.8758112199042049,0.148744224273193,0.4539456013549795,
 0.7449815023867938,0.871146093194164,0.6345000945623761,0.3628727508564502,
 0.6488259040815801,0.16832679827601693,0.5351835766689598,0.9468532877017627,
 0.9987765435909872,0.26836624718101554,0.30743790817476013,
 0.3700687003798596,0.7181750991739804,0.6888595369109922,0.9123774538414222,
 0.11732857141568043,0.446120115482168,0.5417189085280553,0.6389000171070831,
 0.7246399592639456,0.4544833369510197,0.665536448743987,0.17021165167545327,
 0.8010866771799957,0.02968348621097916,0.6099082996465732,
 0.42050570191560377,0.9304344428669948,0.6130040649172591,0.6410203743542575,
 0.6850383916861101,0.6255363503953801,0.30888624860027825,0.8529822707298943,
 0.060260778364843004,0.43555714622363007,0.3222998443274253,
 0.6679383503159415,0.9512242530117425,0.03893664990861068,0.6035633604053284,
 0.6752331854315143,0.2241850789738722,0.5098228654280227,0.2085997175989297,
 0.5960863217397043,0.47405547321924046,0.2110101514050271,0.7624949967455613,
 0.23935560304077796,0.6720780439712363,0.025087657397619423,
 0.006748938784985725,0.25649231264174954,0.4535281792543192,
 0.46628751598097373,0.600832555156125,0.6763469473446878,0.7554807990181859,
 0.30585563599212673,0.5450667865485055,0.6270886243094946,
 0.01908451424592761,0.6482367076128049,0.0011439374305417527,
 0.45895147142642434,0.9044781109547528,0.1325537543149976,0.917819345765336,
 0.5673737064330809,0.43713737965651167,0.25338850104055044,
 0.33593857367074986,0.31989392680126094,0.9153072105380319,0.705956224574789,
 0.8994002114334234,0.8714255931741236,0.06657390186326917,0.4060274612048894,
 0.895171165093689,0.5973017861052887,0.3785671071725222,0.5732126571624047,
 0.6781078882149654,0.03878716853073483,0.1952177667036008,0.3117625314517476,
 0.11099354437853948,0.44827048173594763,0.06696046681140877,
 0.48363101546539555,0.1115193329050076,0.3148546813931794,0.4500011323922565,
 0.9419119518429615,0.7719211047617661,0.43757327816552116,0.7752719950057607,
 0.8881263856891712,0.4645932742703772,0.46347403242652485,0.2533390157730213,
 0.025994654670432826,0.9876574676957767,0.3481306226139351,
 0.13050042651229943,0.2694192908358779,0.8515346120837448,0.5528822963162037,
 0.7723466633722889,0.6051113205762046,0.5707932147149124,0.7390379147711689,
 0.2154487640017405,0.974215771367149,0.09628096500095484,0.7392436203537867,
 0.2779377691332885,0.22554764917229908,0.6426105107976706,0.7668779327430945,
 0.10024125113398008,0.9180191930419092,0.09996695733974592,
 0.4743914128475326,0.06788714810329766,0.1415099515042283,0.4968698484705625,
 0.48289015711042893,0.147750633241025,0.5865614104358634,0.4105699929242924,
 0.8379854973060491,0.051078553099957236,0.11512039541973917,
 0.14062097291709175,0.22776725153048538,0.16253994465773813,
 0.1781723778429921,0.5263711214498314,0.03947948651104649,
 0.02234548225164168,0.28421239357322436,0.6217675784947775,
 0.6074339877545485,0.9331997347214623,0.4515572563717424,0.25497582741240965,
 0.6851482739049994,0.04250199501179508,0.2666143617195671,0.981287635390737,
 0.3699753101985227,0.4992964807936534,0.8094815779954599,0.5922030663886549,
 0.33760869993767695,0.25504720757845467,0.4694060463560342,
 0.4256953572620287,0.36533112042986393,0.7160013965836357,0.7121457129386243,
 0.7361829105872713,0.25569015214628465,0.8327652036809519,0.4025121064174111,
 0.9820262468761417,0.16225931724014098,0.45329533794147137,
 0.9500290845975304,0.794283081563847,0.5453124306971727,0.14051510878944373,
 0.7601480159964976,0.6539585771134641,0.03461035887731245,0.7012226697892359,
 0.11655007177061438,0.9098070863910773,0.4426975754620487,
 0.43846916231488364,0.15873904113235981,0.6168023820756952,
 0.2024533893940088,0.2464190758593212,0.7150737311130738,0.12919910436164472,
 0.6510908088103596,0.20234981834850463,0.7901154531873115,0.7513692472670955,
 0.66453263340923,0.585142285190012,0.6407245639053782,0.549209558549127,
 0.1885101174070528,0.7610766841711943,0.4547405828366258,0.4656635564778473,
 0.4915271354546047,0.8435703158413329,0.8584677360837064,0.9263353890928013,
 0.07344479411026339,0.5321346791216451,0.47845759610421335,
 0.1710882703718073,0.0013049039615974367,0.43541264640741906,
 0.34603786672674963,0.046986014253673325,0.7291320155574714,
 0.20401935219088685,0.8173195544343705,0.42573854628199315,
 0.6929616689446119,0.10142574922584602,0.3210501001941153,0.758445418225015,
 0.9486654245032666,0.4830287312215995,0.8288384411184521,0.08558801439816477,
 0.8236573370382716,0.9155680930895498,0.21626020655699174,
 0.43574808645708685,0.16062505131076976,0.08719494617722767,
 0.3654442660419509,0.8727959378334703,0.22572335858011838,0.5248855573666819,
 0.7406990433904176,0.4091290222385512,0.8996368482193752,0.7686266798070323,
 0.2792823732137234,0.37008955222355644,0.2405535898573572,
 0.23897085524423956,0.6900992264216689,0.08164486411011496,
 0.8053170190601477,0.03528731935681695,0.862158506066132,0.02849842944364034,
 0.5705175131776987,0.7093813604111607,0.37571860298666193,0.6597398075586763,
 0.6531533155745123,0.7550133663303014,0.04133564096691511,0.5595910508262352,
 0.41524857271994664,0.5644642629548678,0.3600290766064854,0.5940938623961592,
 0.20740099709023174,0.485948138952465,0.9261259807595147,0.3064239911592873,
 0.2466367141525586,0.9501633176586036,0.8682129599226829,0.4901210298282034,
 0.9937894458466043,0.9419936951555001,0.9631263369360648,0.08727796520301989,
 0.4590869118825567,0.24611125875884066,0.23125781129360656,
 0.4186680805678811,0.8220242989181962,0.36973551220235246,
 0.39408805960415805,0.9699539523672269,0.4883791490451782,0.6165696249876897,
 0.7142617156862773,0.5864652386223749,0.36216318715555895,0.501174997747166,
 0.7850296932712088,0.13492530355663812,0.8030267854599805,
 0.07378637080547357,0.7814754262563111,0.42898809822766204,
 0.011478011168849722,0.5283172306685969,0.043053404284571606,
 0.0038533997172175427,0.7365472321210516,0.18828863538480178,
 0.7983170914321064,0.2577154623857021,0.84084933633933,0.706729770020164,
 0.6774467407381324,0.012140506820543018,0.34332933772880103,
 0.2038407247252081,0.7676976186426034,0.2912269205241368,0.46415457693695417,
 0.4382157839494386,0.0940835595734264,0.7459646381798536,0.20178085996149364,
 0.7454989549323494,0.5616011117194448,0.44259599413051953,
 0.29896382316682457,0.4951447062764599,0.9233600847609643,0.58614003369816,
 0.21220055521444225,0.016383269992522043,0.22743026173750203,
 0.7731081996445311,0.7646991380377257,0.6335165228301651,0.9095044628245609,
 0.5275840341241503,0.9192143922176274,0.5974076765249311,0.4354618568774813,
 0.5465681931303066,0.8365681426158231,0.45532320691479833,0.3783870756547275,
 0.8367451820413246,0.5412174610259735,0.7955269327706337,0.6771629364601435,
 0.43078680108789724,0.7609896138177029,0.8046790857913787,
 0.44640937594094743,0.6604328574911678,0.4152606847742717,
 0.16066061495180206,0.10151285977698055,0.6646061242331197,
 0.4714754065034884,0.15086589583341958,0.8177460666527,0.3596012346909173,
 0.03699941573773702,0.37028672430851217,0.6343836508753152,
 0.7160782339686385,0.5062367017368772,0.44328082069434904,0.4722262240811892,
 0.6301622102860591,0.0884311054426623,0.3191930859557217,0.41465603172117593,
 0.8734993973910471,0.15993615718139442,0.11307458260447834,
 0.10481627938184057,0.8582577141805613,0.5748749112931565,0.5075007277664921,
 0.7549310107551686,0.5786057772837623,0.2391105788383765,0.6083501507590534,
 0.08843341118708148,0.019134476824532398,0.15689626425336278,
 0.7289074866058698,0.1770823718674146,0.33208650304579823,0.1755634887544565,
 0.83982274972163,0.01830568428994528,0.7505714567923417,0.7417031316855641,
 0.08988619829234623,0.33952865954739564,0.0968668424625111,
 0.5170273643824824,0.5451473169267873,0.8125070028770637,0.4274788772751156,
 0.9402082999297401,0.4538087984526987,0.2864494744632796,0.4911043477173914,
 0.8666675357303273,0.9860539817601232,0.18312469042801593,0.1375485431171578,
 0.4880862689607637,0.9778568649106015,0.6902178173938472,0.5153141502969443,
 0.9965237733023022,0.568858475013795,0.1656495824315788,0.7495217295379731,
 0.6217041129864357,0.7757812173568774,0.768382500143945,0.4923039113626174,
 0.8251518321361548,0.11804591872588588,0.3001682390788415,0.9710395823597842,
 0.08134320927237193,0.39214983629998246,0.3796273983587468,
 0.8994955329250898,0.5706343393173192,0.30331661653727804,
 0.22242993017075674,0.7061069605599373,0.38855626907066787,
 0.10976213610391583,0.2515250273845461,0.7452812447961077,0.7339977587924114,
 0.9371165985132389,0.3135583622648399,0.3048094972550127,0.6915875673387459,
 0.8885499054981058,0.3397371528673887,0.795365744241009,0.48839298013425736,
 0.9025308246807354,0.3791110937315413,0.47727701177970583,0.6981139021860904,
 0.6527425194241796,0.018117077691578,0.933289643780738,0.3640297216897441,
 0.9468228995633191,0.33358116670980753,0.38971941284093536,
 0.4704572559270668,0.5323963766267349,0.6654513574011359,
 6.201647625605489e-4,0.583592634850491,0.5489344728841837,
 0.05727028326956762,0.8877552071401857,0.2952949290356296,0.7450926885053122,
 0.8992014073557898,0.8337507001751999,0.12348650923792714,
 0.0027512301872382183,0.39165819630760956,0.28677406598859956,
 0.47290708852722063,0.6885206301091169,0.41245597812898105,
 0.7531049683483132,0.777634347382436,0.4678544057695553,0.406188214497206,
 0.3091206661358342,0.08128955740370392,0.07326166623779162,
 0.19273507431708814,0.5473143447787369,0.2197839985643395,
 0.19059579171380847,0.022608480430566624,0.27481617165710637,
 0.22196805498325645,0.906407673137056,0.604171752371241,0.018032370393701314,
 0.9370629656838182,0.3815010794767628,0.45305542254500075,0.3696091039811725,
 0.9185869016701165,0.4789452788918649,0.5989691966349183,0.26198962811882387,
 0.89290638585033,0.2963851182798206,0.7674050293988139,0.27996357878325884,
 0.07579608848303854,0.5152261940790402,0.6395227121425922,0.517970714864412,
 0.23290288915757484,0.6298932983468631,0.6284487730333649,
 0.17302676065937528,0.8254082783295338,0.7216743292198134,
 0.27165568252604444,0.4957143058302488,0.6605894043705989,0.8967753222419327,
 0.9532888722295122,0.30982278506609384,0.21327041591586804,
 0.6131845019178479,0.36176326025300076,0.1918294317528928,
 0.35977805692324183,0.6646243790715551,0.3972500483249979,
 0.22800145014037265,0.06409967524954374,0.5959872225585995,
 0.6573876714740524,0.3753909642769563,0.33742189965754443,0.3403037974871146,
 0.7043524979146683,0.8599023763842858,0.6110739695910327,0.5491413839622488,
 0.24715374846314653,0.47646387225468045,0.14125357533315497,
 0.21893149801435519,0.8142859158500813,0.9542647189811547,0.4513042768568858,
 0.8187257502839319,0.9818706382018246,0.9835567592727024,0.8675389501935815,
 0.25022158365518943,0.44329397434543183,0.9886217356572009,
 0.6457228979045557,0.46207136244748404,0.18393610892159684,0.814630586938471,
 0.19275001731244756,0.6478887722162279,0.6561102050275842,0.8074576555028176,
 0.21377295332692814,0.5515200628648708,0.523186724079924,0.19914296366401085,
 0.9338104455676353,0.9108207490707643,0.08236452767505753,0.4058847648970322,
 0.08118621394968517,0.1932873321456664,0.661501176366336,0.31289678914712116,
 0.20301102889456057,0.29772661127454847,0.6012846914832799,
 0.8203428692689582,0.3377071168529371,0.8476816629821344,0.4555729859443459,
 0.1358303524747716,0.8218659672280068,0.9874723886753644,0.12201688514747522,
 0.49459062658380415,0.22789979534376859,0.5907524365656805,
 0.0759207702318132,0.7770081360976685,0.8909777580738463,0.6767440508967987,
 0.538867031004949,0.18360916381470216,0.5074916551678837,0.22618160618396133,
 0.34522223655802864,0.3466741924194773,0.7821690417890866,
 0.23563679989592956,0.7869271688114217,0.26788275941527295,
 0.26611253782365885,0.3352288152184806,0.3376022242463561,0.3669632388547841,
 0.6740862784872889,0.47756808792930716,0.6921494748185657,
 0.19109795398581886,0.6830588427749335,0.957399987073262,0.21398581734063815,
 0.577406088216252,0.787632360851089,0.22529606342709319,0.4495545800407812,
 0.7059568619753481,0.729009287010108,0.7201056263836959,0.032222473139425944,
 0.1964276591091535,0.49579539220637436,0.7373084560979367,0.8723495023739409,
 0.41945754572051885,0.9784260159490894,0.7645301991851445,
 0.45093862679811236,0.1872407122834968,0.8817940837918642,
 0.27786027468398133,0.6873956998477364,0.9680481197898066,0.5478535453880289,
 0.3055322402326177,0.060027714660870535,0.7898199769780162,
 0.7745875006435379,0.9890947997911007,0.5302011613550583,0.9596567888758691,
 0.7572469160049733,0.76976418260262,0.009095870148921215,0.7182536970751261,
 0.3164274430467797,0.7365207563088344,0.07252126712484319,
 0.22367248833849074,0.3459063180473264,0.8324548981944083,0.5352228937324329,
 0.9291396905925364,0.19071217242574734,0.37654623443896884,
 0.20163846665241336,0.02709474718369531,0.05299136891429923,
 0.06587366169873998,0.24516306182894854,0.5775928202527952,
 0.7636223197180915,0.6804656468359295,0.4539880788159526,0.4453400577756632,
 0.9537748825214345,0.7557937256905373,0.298741461337665,0.5053129836970118,
 0.3329274813337294,0.49986142785618715,0.310863573273479,0.8294840568547218,
 0.35240873734442746,0.37800186672209657,0.01405965411130461,
 0.8175163883797031,0.27653503756368014,0.17438557742968164,
 0.2864531721953487,0.9179034915243234,0.11725354146924127,0.8053795298416054,
 0.32491111202193157,0.9210308940253674,0.1914343477267978,
 0.21327184634419305,0.3114867162272277,0.4284126382817852,
 0.0018577861244010174,0.20532168330735778,0.7284370250935557,
 0.4067669165834321,0.2801981649469084,0.5116764534901016,0.8627716901873177,
 0.7202073998457426,0.01658355875891593,0.46777342935016575,0.444421868857762,
 0.02240119446729294,0.5005408560779441,0.46120055397717385,
 0.3713620245092977,0.5231462173443175,0.6565651293951227,0.7144727257058258,
 0.026787455207814337,0.08288813196926259,0.9627323733962433,
 0.8844684297128796,0.13093005257939816,0.8110948794767328,0.633458333947018,
 0.15232738077840402,0.9368018787641622,0.44513518435407484,
 0.9799733766876471,0.4970231066249011,0.56993292821015,0.6502368143939606,
 0.7553589092393175,0.12384456985248626,0.10984605610308351,
 0.49037243828587684,0.7244512104220592,0.6212973722613706,
 0.09910910074491974,0.5293620233940294,0.0646255375030459,0.869005788935155,
 0.01842562810267967,0.6089579223333992,0.44931823121422443,
 0.6921508132253766,0.1425842441537365,0.461313174610006,0.2024231366326983,
 0.2744777921033421,0.34446314270688005,0.33001581477057407,
 0.4187491547855161,0.3419458745150715,0.08367567364579487,
 0.21189296116485823,0.12764292909296138,0.7050732632127257,
 0.3530203999099525,0.7034957870165184,0.6723743922892631,0.8518888011084012,
 0.37905177170049076,0.7953250226308803,0.6054234774224908,
 0.056114675016153304,0.21242074186977078,0.5942712228741571,
 0.8670545915573368,0.6997503923960753,0.7051970182111575,0.4255067232571922,
 0.6025679467343563,0.4418131009030005,0.31930060977263064,
 0.10070186063472886,0.7923458117076628,0.9981751837223604,0.8175747665780788,
 0.2596169778406505,0.9287514551265821,0.8521088546860882,0.5399443718889706,
 0.6335079650247727,0.1361780996034745,0.3910659595215178,0.24488528505840668,
 0.20447353500091814,0.7166451433135396,0.9151932234890012,0.8266301705022412,
 0.4138100680216412,0.47632886034761135,0.1344086793184548,0.5795949996270846,
 0.8355288394477722,0.7084522883418527,0.5926660154423284,0.3891896843328364,
 0.4671956026153977,0.12605877019436484,0.833515868223913,0.6979878735558017,
 0.8554117279936476,0.8569733877184563,0.9852239044894304,0.16134014816913733,
 0.17849495736242882,0.20906823483726067,0.3585051424411937,
 0.2963106923699732,0.18632980420216028,0.3785956744808048,
 0.12162865311154347,0.2394994244783224,0.8124615847252781,0.5918880110326178,
 0.941481121680796,0.899782968380788,0.2425191254451199,0.05154235287708109,
 0.7611362529377735,0.15259825628968005,0.4353903759409167,
 0.48132637201965445,0.7674229478753528,0.47655302009906353,
 0.6900672663218331,0.5034606221081402,0.21298218322390772,0.6577886069318246,
 0.41864589423088416,0.8294345910200556,0.702346390803013,0.17776888281941572,
 0.2341507036145899,0.9626529042021713,0.6455290591678854,0.01931354014416342,
 0.03912650840769194,0.034365254195629524,0.2593723345416141,
 0.14977181508149495,0.7637475159934459,0.12889377301689342,
 0.4179220819860615,0.38492616845084204,0.5109258754290841,0.5638429915617736,
 0.41097955497877037,0.2460000893223454,0.9935107281014705,0.5814455643628078,
 0.991144844924021,0.28865301031356005,0.26652117181970536,0.8273694853618648,
 0.33373303867507953,0.8114562719936562,0.20142640702620784,
 0.24575401611277625,0.3623760146350963,0.9404053290740235,
 0.17880917784487516,0.7518274356312251,0.4350685098891859,
 0.06963343198589378,0.16454747872855224,0.9157881704124691,
 0.6041353322157946,0.7894117122362034,0.049944840135520585,
 0.3928373786694067,0.7794755813781905,0.7435581787987733,0.3956689066801594,
 0.5859020943477234,0.47545767934596916,0.8858610080080089,0.0768816437265557,
 0.08272669722416748,0.8012729311688814,0.7127057827286142,0.5572111791052962,
 0.35311094170889434,0.7393185910238749,0.6650376930166698,0.5388253075389139,
 0.2450853015825185,0.8561694248001771,0.8026000569131508,0.5282515137710908,
 0.7627517217151455,0.19308977553736795,0.1623914760113504,
 0.39948787601918834,0.5358655197107672,0.3588974475962,0.6698114598306688,
 0.9248628244106003,0.3845181765375485,0.8763877618226485,0.16287177568991007,
 0.39705740489221586,0.07752726867673598,0.16295242376975938,
 0.06722503070667873,0.9148754050880807,0.700869611379122,0.22685696187608595,
 0.8433611574767641,0.14999464718753075,0.12314610736601539,
 0.9031147484822147,0.8826737362008943,0.5697947767380886,0.5379201465876433,
 0.7109511461294148,0.9286218791890707,0.7209405603003879,0.13719238871373518,
 0.5968251031320794,0.6597864581115594,0.3225686065502358,0.242028557152278,
 0.7385166525412211,0.5667232505946906,0.13048365878552515,0.5016739278001591,
 0.31060519602591175,0.799092455240917,0.8173369748364071,0.40963892514006284,
 0.5003594841301192,0.07753284149017903,0.1040626110202395,0.5277539293994593,
 0.9161477476946048,0.32625754933301965,0.39452646150950754,
 0.20862848653678112,0.12890548285154046,0.5246017905809116,
 0.40524676410894167,0.9692370794173937,0.865239826254208,0.09928943327848605,
 0.15108569295716123,0.40935207107596283,0.5371038507895153,
 0.33402701025798354,0.14402060243073767,0.6657067790558268,
 0.6262897788470528,0.9714410484739253,0.565707105635087,0.12708217064287486,
 0.07856842873819514,0.604596908147992,0.3937751803144105,0.16232592833530834,
 0.6192656708629312,0.052313650217006735,0.08363346661698146,
 0.11251655597007648,0.10140777294152448,0.19089976226552785,
 0.2678762042615598,0.3208465391514703,0.685602895508112,0.5728086981722255,
 0.12653399982858715,0.1986845835312009,0.3255452383236366,0.9699343371820611,
 0.13888719510425585,0.5218834962797843,0.6308276340976974,
 0.16121241669676223,0.8843558428736468,0.4773309260292389,0.9434983151136644,
 0.28012933041890054,0.3283719050649059,0.048869962001627854,
 0.261719431984056,0.764729693388121,0.8104210100398246,0.7872910582950489,
 0.9965967109863869,0.5795727940574049,0.8048152371802123,0.17924469366680018,
 0.6982272455700331,0.7463433241036884,0.5021164950598379,0.9711056039674182,
 0.6858096436109133,0.39405679186438414,0.8319692539063115,
 0.17126668401917988,0.15524256086436505,0.6819442795911166,
 0.5434834120206233,0.9790332795892465,0.9152295778397371,0.7187121479528988,
 0.3042777772130847,0.5801918736887748,0.45679307829031757,
 0.17109393418553087,0.13138692585386846,0.5557019567978415,
 0.6341579778983928,0.43524526781146444,0.29717426453506646,
 0.7687645148658274,0.09016059372393281,0.19213326009906084,
 0.9337452092655469,0.9091763281153497,0.024198097468615676,
 0.7775866202525045,0.27352586341395724,0.5412994536970233,0.7044406412981224,
 0.6592410372259144,0.09537758483357484,0.553783947330238,0.529843830286719,
 0.5062866161594363,0.08010065523356447,0.3624236833221839,
 0.04650616245152017,0.3941071059501835,0.13360334933898343,
 0.30521132248748994,0.2364577889916526,0.20750613987630329,
 0.6886669853841091,0.19298950189635422,0.9998355101253259,
 0.07438089438002149,0.3425892020062562,0.3848505057631194,0.4666211305093597,
 0.11773395752201254,0.43280186497905204,0.34206973357651216,
 0.5216926221724436,0.38618748715710116,0.8370901745794005,0.4601108286075736,
 0.43370456663920975,0.36590440952628633,0.871573166519581,0.9246568947526448,
 0.0859274678372377,0.8574166942353822,0.17540695001744422,0.5100295691864821,
 0.8634422862672712,0.1060288464677056,0.8473926547612813,0.4978856592101949,
 0.8197208331645502,0.41848121949594463,0.586153358697647,0.09415232999101342,
 0.9619084024279172,0.18031038998407267,0.8936550550909049,0.7877872709604448,
 0.43944507660318144,0.8054571502075198,0.5357784783154076,0.8733907772860752,
 0.7046926851708176,0.8299688956213765,0.2897486827445992,0.4504977821376065,
 0.8239995174764184,0.7513648816668058,0.48898607243693004,0.9612740582318648,
 0.9196268175698558,0.24276425508191402,0.7696406229465158,0.1347006795853647,
 0.6301508988951438,0.9786022006570517,0.9576008006725091,0.3144180002251378,
 0.1395660229127157,0.5869987421638954,0.665442788372975,0.9402240853086921,
 0.251423778114665,0.4612532468635736,0.18597163918211712,0.9117857748393272,
 0.9353031656331363,0.8043827622699953,0.3456679388645343,0.8474741520633274,
 0.4629089661445497,0.6175011369669908,0.12255973001084086,0.6215102310482603,
 0.15277186724977776,0.5097695725903941,0.016627298502634558,
 0.26278020743459085,0.6825582115867476,0.9945693480881997,0.3827813666685327,
 0.7812740561391919,0.3038616330434384,0.054568991061082706,
 0.9168022326497713,0.6821711046421979,0.16233985753517555,0.8049105756796432,
 0.8057880918627929,0.10612734176570726,0.5197089040655418,0.8238103442047069,
 0.539529157495738,0.8044238907214335,0.5825651662894974,0.7596416443636995,
 0.575360211008488,0.4295103114049902,0.09438083370568429,0.6960193501996653,
 0.11887180256955876,0.6937079699878226,0.5725601938026383,
 0.18532719894299465,0.778176796441017,0.8098218534208093,0.3426726872268707,
 0.09014821611599233,0.23456524178690052,0.9617590771566537,
 0.39531061317431904,0.8693768929203307,0.6383549092997417,0.6335037439664544,
 0.3763891827636954,0.5235327116569934,0.30663813677198914,0.995667866533882,
 0.052585099569394056,0.77106323593561,0.5362196503729783,0.315369464016269,
 0.9773840793672119,0.2960398713274751,0.8186913542299105,0.9751195025176448,
 0.2447352736777555,0.9171204840496082,0.16125953495797907,
 0.11515000886259819,0.8011412714706945,0.0714551798512354,0.8888453562883356,
 0.7325293641952363,0.6968504344048356,0.4887627846905541,0.37281586056239735,
 0.3169412614142658,0.3291478854187736,0.6852406458542751,0.15302248458246048,
 0.5374190957457889,0.9280403906281084,0.7878949856388221,0.24466384928236484,
 0.1150988541469754,0.6293411242648211,0.9431943345726695,0.808463904251775,
 0.006047139252488898,0.5060109943425383,0.38558141796494694,
 0.2721639384991429,0.43130439514348673,0.19614012194332908,
 0.8672432366569404,0.9128957694838755,0.7316336127318809,0.715071409437476,
 0.4513441707202297,0.8931258855066219,0.7156350485836149,0.3317447947953087,
 0.07862537548372894,0.38516906756291824,0.4351596464974994,
 0.9282988311018608,0.8542372144642493,0.11789393017792316,
 0.47915268445501347,0.7619174258155259,0.2891381992611838,0.8631777873560194,
 0.13048381043750634,0.29335785478797904,0.9974186180091429,0.179337309314888,
 0.5669185183754692,0.15330703895629239,0.7111350262929211,0.8162884432411682,
 0.45969437044103334,0.0359629897345185,0.22857339090983075,
 0.4244953108294789,0.24430948260254493,0.44906590011645553,
 0.6866586213332384,0.7743207550941559,0.6100676193387216,0.37348570154178673,
 0.15848118284910517,0.40667495480137184,0.7107742260093923,
 0.028251559836948825,0.8841863367445038,0.142713575479539,
 0.27031136281220014,0.8292041624171828,0.5402853624283339,0.2279597762136889,
 0.8824832430497656,0.0512629168303782,0.5272728601460738,
 0.019313628520543302,0.7610686252447739,0.08414793627821748,
 0.011616319075895554,0.08619272452850668,0.6781349620310069,
 0.8882766877099274,0.7885752772397664,0.07299357875671852,
 0.49490287981649594,0.39618634933806274,0.46964239834572585,
 0.00599047988353929,0.7593556921456988,0.7036949882699209,0.6211544618938745,
 0.4290272967240225,0.5232083209322511,0.9347009139938385,0.12566476197342402,
 0.6331485076767843,0.5631368703612896,0.6911059343055416,0.8873175656051011,
 0.3012356439744909,0.36994201465883647,0.6339428851713859,0.5965886524838866,
 0.6452967945900362,0.6949624549050631,0.268053893274278,0.641417699719772,
 0.21737610662505302,0.41637360891194786,0.33904374257106507,
 0.708251563824618,0.8693452920432583,0.785859140793851,0.4711573854928873,
 0.7458789656501159,0.5955032470771228,0.6567526174434888,0.6133394276476194,
 0.8150848264418041,0.5550227137916689,0.1603717158721969,0.5206990533110623,
 0.19814567129408167,0.17801352063624354,0.11111574103701138,
 0.5454436704980848,0.18509133671069988,0.7970283701827269,0.5439339361891549,
 0.2417540339630364,0.2301775912827646,0.5735020901231781,0.5184412528806045,
 0.8518791165099074,0.15379323817285062,0.8615555308499,0.12103524130127474,
 0.5318640776675457,0.7587372675554886,0.5940517186607152,0.9308608798617238,
 0.9884085511486819,0.043621129165507866,0.4285406781248088,
 0.6041897715273947,0.9188428035945259,0.3988333791430405,0.41938229433406193,
 0.5057535692025851,0.23539158700212737,0.7717533023683575,
 0.31910998572900817,0.8173593494592972,0.7234475829588859,0.5458072175929424,
 0.14803254693150103,0.7562843160733382,0.3038216264989677,0.9080410944827055,
 0.3540209648684878,0.7789035886055484,0.5537187613389466,0.897081768977037,
 0.41571727847071893,0.43140020272820756,0.39710018760604227,
 0.31277399732057076,0.46540040527980553,0.019083499476941768,
 0.7761248459418592,0.38704086563203766,0.2813160093799081,0.8841697640711118,
 0.09812597838314752,0.8675167656571844,0.3032609317535583,0.5270623182481199,
 0.6163497056385201,0.3970975525410545,0.6908212517578225,0.08077808024608868,
 0.2735190740072361,0.9316026422333342,0.6784450524392711,0.8537246881575037,
 0.0718332694939604,0.10510026886153823,0.9292752103265673,0.7146651760719752,
 0.7982315108615289,0.9297729684042023,0.004442376207453469,0.828468206803785,
 0.10078013123319329,0.6100452050749561,0.3612977766335632,
 0.18505828234789223,0.4650778092335448,0.7881733823703231,
 0.05791330835199526,0.8226828123314245,0.9705308003138128,0.7399903131225547,
 0.956563683823746,0.6266284761700943,0.7073302439755002,0.20813749249356595,
 0.6166664544146739,0.21119886328955184,0.14167732300477232,
 0.24136333819122135,0.20514795158384636,0.13336544147692775,
 0.15793878937773598,0.8553251606579495,0.9520362198368523,0.8435548425857992,
 0.013427981404415479,0.6266960773365063,0.46782393977160563,
 0.0144255946227696,0.0484092146136923,0.501562765021216,0.7001244408568759,
 0.05181564523186066,0.08112987496852453,0.6513680808240258,
 0.7994906491292159,0.804057255440562,0.24792352751000357,0.488947567832933,
 0.42303972989405925,0.7193709353711085,0.13253472262601118,
 0.6283500654850859,0.9918567849099873,0.01624413777149447,0.2567640413334802,
 0.9185268723193618,0.31892298025764065,0.017399377614795597,
 0.7837629531000129,0.17769632525141843,0.472290293569292,0.7960846816633956,
 0.9286631247989197,0.9841661074051922,0.19910662555679437,0.323515578218463,
 0.2614627306440499,0.7650265073449991,0.07672850820623534,0.6571353673013198,
 0.15869449218132925,0.6393891054012599,0.16565911372787445,
 0.015461549951150011,0.305084397027679,0.1428851578015734,0.9396924900463269,
 0.4560044690935159,0.646840981746303,0.709641341683862,0.4376687833561108,
 0.21539207882653089,0.16626513335725512,0.46044748292441073,
 0.2079191727876425,0.05265201831137123,0.7670463441976865,
 0.19019382425775777,0.5529476629802148,0.7430558643445666,0.7848982894886725,
 0.22333534572737224,0.273850442339161,0.4123102909206956,0.22113383165406786,
 0.2798298054637298,0.4447713932899452,0.0767588967767644,0.6991547862425899,
 0.19470412986104413,0.17416911881601715,0.8982498713213514,
 0.5551423938051644,0.9027851386713283,0.4638601204040832,0.6697032052884649,
 0.24323857475772825,0.5863753638463075,0.5971638789934542,0.4516245293591685,
 0.3035815007952386,0.31728058929481273,0.7135224106300592,
 0.34747331217212307,0.6727329131802551,0.45287084107348896,
 0.37851141566541124,0.6872224386365452,0.062494262069831774,
 0.5514036936708528,0.4822087418127968,0.580413002821519,0.22159705484072545,
 0.2229507320825187,0.11634630521659939,0.13600244144162654,
 0.2289206009818443,0.5617407819980703,0.35436679901564916,0.5145701941820203,
 0.12019227022298351,0.33413285914611635,0.15256103339934346,
 0.8204058113723882,0.6401218099917528,0.7996725880281096,0.45562716065898845,
 0.6286580644610615,0.1537978207129207,0.8659699618185281,0.7028427000829438,
 0.19361269383664348,0.7467173995693317,0.17974469634226853,
 0.6001184323592217,0.355369021896761,0.7690291328266706,0.6422500064012331,
 0.007995883372883483,0.31291359409196196,0.34633229676882715,
 0.34172365943130023,0.534412287486671,0.5192953911367775,0.9847505866385272,
 0.13043459314179728,0.7589563753797836,0.3522501630656578,0.7284669221264384,
 0.4477053321888853,0.13148254351157462,0.8724873947463416,0.8481377055193723,
 0.46291976696251336,0.3285831398755057,0.1231205984178787,0.2796143828556754,
 0.9910019975753687,0.7201187679151851,0.9638618209592744,0.21244585973660435,
 0.9170982710375318,0.21139934728020404,0.8171761553309793,
 0.02661218156420031,0.5788304786661813,0.4363090308708568,0.7665258751021934,
 0.745081560564483,0.8294827263100496,0.16654572398905498,0.04703023628938274,
 0.566294613665985,0.4588035607640544,0.6086927760473113,0.7945601383754167,
 0.8319259848787323,0.8523472051249465,0.7272781042073239,0.7716232246605917,
 0.11496174561944905,0.324493871670132,0.27856610899603906,0.5774783157847191,
 0.9145120015863752,0.2766042798481295,0.6238675931391396,0.2754681973802171,
 0.12957220480587284,0.5126605246655637,0.023919037872079385,
 0.19916282792060624,0.8813083951424795,0.8357930580801199,
 0.30309957035623436,0.9420868034563651,0.8789927041363484,
 0.41661670914994775,0.07183908058132693,0.6845387538118755,
 0.6324143487761894,0.20246028376865055,0.8679428631967008,0.09629068687872,
 0.2958028152236778,0.6580137447346667,0.5805056763064272,0.8862172122298356,
 0.4572435296969394,0.6996625981135304,0.738485074046289,
 0.0017825511608582367,0.6019481757032631,0.7636330972977119,
 0.43525123056117376,0.9716083395627588,0.5254299850874659,
 0.23971949137202142,0.5443759889873814,0.29505188511520186,0.63939782790135,
 0.13278404662692256,0.27110411735427753,0.5943133022558738,
 0.2694766997794972,0.01459005769286148,0.9414824174579153,0.8088026907170551,
 0.11345776618822256,0.7122082811165724,0.30551515759351044,
 0.0981921229387297,0.884668564663212,0.5716906981291964,0.8402496311070153,
 0.6256112162678604,0.44010674754013446,0.9927918847322602,
 0.014420924440161742,0.2672671104499853,0.6727038047742788,
 0.7642245319477199,0.8055863389208412,0.18226092442335295,0.5847931815918911,
 0.2787060139562384,0.39521157084413405,0.8072190932975636,0.5480698600897874,
 0.8087987136671572,0.9652869972309275,0.6237510196977929,0.6247925637753245,
 0.7063342637399598,0.012765893253723237,0.15521848372636005,
 0.11312742438458723,0.24302177482513798,0.7387027350834755,
 0.7870501088419929,0.36567398331131806,0.2895164191795816,0.3104178510194753,
 0.054648700031630915,0.6575209956243857,0.054522516047271985,
 0.8694174917508612,0.1216616984076746,0.2600519546074379,0.04388185599082384,
 0.30226033811058306,0.14815281929412893,0.10211221977548268,
 0.30872382187604086,0.46016330830590313,0.06518178316114165,
 0.9442597575426543,0.23262657674764253,0.12926575041436295,
 0.24237477217956416,0.9020721322483063,0.882813047140108,0.6850753627373593,
 0.5371399495083032,0.3175068282869199,0.30618176091283433,0.576314960616397,
 0.802513120719532,0.741229144548088,0.13711456248798326,0.5836102391717566,
 0.45188475313112464,0.1045492310508751,0.5432623184726653,0.999998712326579,
 0.777332755017992,0.1400173042495877,0.8335999607116124,0.6290204960029734,
 0.3921321748766279,0.43509083877536103,0.610104879670431,0.8601082775643404,
 0.33682483519426776,0.5168349366910948,0.1417671082144456,0.6081110086294969,
 0.8756713308696362,0.26167610047712564,0.31042782617963927,
 0.49360458865369494,0.8925472918508367,0.5205287568917194,0.4753492330255077,
 0.35143641715046225,0.7140225909345702,0.9564226607199698,0.2155862603825579,
 0.0928054265634276,0.5350859121276506,0.24830165448448105,
 0.04305648039808674,0.5670321979176933,0.9792535843062362,0.4627092331685623,
 0.696653236302534,0.6962241583111051,0.8567016742084768,0.27356201783647693,
 0.10019949244043103,0.48433310556076126,0.005445379688657548,
 0.6590763051683979,0.6782659310490927,0.19894666297330643,
 0.23072343815487095,0.9649510397341807,0.10157194156362759,
 0.8957532747458579,0.8579490154377362,0.19869374172631993,
 0.06983829095004312,0.8747874422737427,0.44941667764904736,
 0.7350313451970562,0.12678155688811632,0.49357507852687943,
 0.9393585958722369,0.3807110252698037,0.3331819559739172,0.7188107207704131,
 0.06354564036387056,0.544941707521859,0.5191024635942543,0.21422070997898135,
 0.366853707069715,0.6449467367377006,0.47621967349477035,0.17861396823980902,
 0.11245085160675106,0.7906109861877297,0.28814989939910607,
 0.21108937995253652,0.0638157176526355,0.2674484138134239,0.3428516590002698,
 0.8766979042790821,0.1371047305465305,0.48257484432469566,0.6908878217927646,
 0.08441724884329482,0.4773168149978726,0.6832232108525236,0.9541884821420467,
 0.4345631341700096,0.6407850149714694,0.037387036211769975,
 0.13585500209492318,0.3867436155593047,0.8554373248277816,0.7608848214012162,
 0.5148045436970752,0.9734190524151151,0.8730524482660171,0.21779867487460391,
 0.9588081257223079,0.9079700507734962,0.4995028732479263,0.2407092859634743,
 0.907825792077979,0.5010423829249172,0.7909413635694318,0.3539244944184211,
 0.5482876563472807,0.06590377787438106,0.692125148042913,0.34824402275111,
 0.9827743444911774,0.40635344921705174,0.25622705175375726,
 0.8376793453339152,0.626370102420931,0.4388094575245477,0.23782554864430705,
 0.2653388817195672,0.23538109232820914,0.19118827610705524,
 0.7888553803700626,0.9575007306353225,0.35858715714421396,
 0.09323011149944616,0.010867424109451851,0.41015624108513904,
 0.5117193056971361,0.88719417680726,0.9544935082226941,0.13660652925242567,
 0.33332259170218126,0.9951785684009261,0.9847169374994295,0.9366366041391555,
 0.7073458653065566,0.039929922741613555,0.23997071552634175,
 0.983660142844755,0.5412498219430394,0.20911652376579792,0.6582599287101708,
 0.8860191756309208,0.638123151648877,0.7630295513147621,0.8545825434927878,
 0.6023740432149057,0.3165362076723427,0.013472975539264298,
 0.28684725419575785,0.3784931230451105,0.030176637204686996,
 0.2789963742632783,0.035762075202271926,0.9328075080733984,
 0.30015384953256763,0.7468831709841385,0.6305517328813023,0.070611362126483,
 0.8696667553122809,0.1553767198495063,0.07124155080314898,0.9614281353574767,
 0.27013744547559904,0.8735677684523472,0.6389899110098884,
 0.06662865239730975,0.882176281227815,0.657489635798988,0.3598396204419909,
 0.2221760168232696,0.1809006818184965,0.11645449511263228,0.5614055117816019,
 0.3700662309492315,0.20694812690446318,0.6936554123450859,0.8190903862389203,
 0.707308570133039,0.07298945563622239,0.5816928001700455,0.5568025167461755,
 0.49211793493998157,0.3939142773738813,0.6645696727726726,0.1765261446576598,
 0.5572123641709467,0.49268919528085897,0.8807130390394278,0.6033458760219338,
 0.5806982053653402,0.018673779268954638,0.11668114966799192,
 0.6401388406023281,0.9662258304797864,0.4502517781848142,0.6514568394754665,
 0.4295799448720261,0.12164249202520616,0.4242291154505492,
 0.17194977802512357,0.4309522610544425,0.3492168034056792,
 0.10086957199341595,0.45726064122009824,0.03534475495312028,
 0.1016822154029553,0.4107255636144731,0.1666026697950873,0.537151651456048,
 0.6398511738084728,0.3611927371947308,0.8473199699366714,0.5913403156460295,
 0.9249688446680173,0.6251699589683869,0.8969957659140582,0.7247270098891505,
 0.3277840205183076,0.8368053178313306,0.8809152759703034,0.5285371508030894,
 0.7606070374835211,0.4176354642858229,0.813809772843872,0.13395304332774405,
 0.9869311659216524,0.8270345401885199,0.4279277338609322,0.5254863151205714,
 0.41577914141805483,0.44937127111207653,0.4303771164456712,
 0.012137062921103947,0.10205976229373537,0.9853432145891992,
 0.8162909738409166,0.8387129385911394,0.22757639238857907,0.2045775548047596,
 0.6055356568741045,0.10330334152916998,0.8874552807569388,0.6833118416826914,
 0.2987124471294763,0.1835592141653184,0.1396471145998045,0.9335802716574978,
 0.02848232033928355,0.33626224747995215,0.2402756362921844,
 0.9728426277622945,0.5468935866875424,0.7124665149116022,0.30034592882418076,
 0.5407082410909395,0.8994692478324348,0.24470456084642056,0.668175606115684,
 0.07074463053066304,0.6082264431470256,0.7992812205433368,0.895616493500691,
 0.7943346103984854,0.6771659527219525,0.8239677167294318,0.39549090413425525,
 0.008485334555061175,0.19271571953292654,0.8059049134615246,
 0.03462831871871941,0.8339767914569717,0.2856738060887032,
 0.016945633782638803,0.27033539634255566,0.26234617108092595,
 0.9070649457140326,0.3447316555976032,0.028750350297495175,
 0.6503075399926216,0.9460766201349412,0.6014960738559141,0.9205119240112736,
 0.3275257043765746,0.08748985032739942,0.19195348701670856,
 0.8046276032295302,0.68031967048387,0.8022456124428077,0.36740410441650373,
 0.895825708610832,0.058200173981778014,0.6022419663507226,
 0.25263788457900327,0.6895334928351851,0.8338259469193069,
 0.32638647989440506,0.12860559392880322,0.19934718347651337,
 0.7084870896762011,0.7662852264671525,0.9882039544244139,0.2580680983513508,
 0.7786234120270623,0.33779984849876854,0.37340053027602094,
 0.4881913906938229,0.7890774529820133,0.9239182543747128,
 0.017535049785080803,0.16284971885581123,0.8216575710999983,
 0.14739156959132083,0.48798473076311255,0.03131733566605621,
 0.0231427281490878,0.00821314820766017,0.5892047595617351,
 0.20357603492716858,0.7906222493048491,0.35241118795410653,
 0.09815378833598676,0.7062968250304343,0.4872968837219087,0.9345139552798322,
 0.43624977972729273,0.9985354117722203,0.5398921636399852,0.3691403823260865,
 0.9257118995368353,0.5865411570840132,0.7719107373292763,
 0.017251548283996332,0.6326679928059405,0.8870805601054425,
 0.8097424038698124,0.8740251804455341,0.6094425185109714,0.4828173535133089,
 0.6646118549629317,0.33949426833959484,0.23623891150006338,
 0.6712192094790743,0.42780676082646873,0.1330193594392779,0.2803682469704625,
 0.33521193539671146,0.1477574918293919,0.893664948158671,0.8511496611249549,
 0.4393836374017688,0.19524413376508487,0.5285052413915547,0.1794877339343115,
 0.9888393926493823,0.9240100147431833,0.15519155109485216,0.5830311638281648,
 0.2224283983537323,0.8764028657144323,0.8089340042945872,0.5378776495440164,
 0.27377589121208556,0.7016764644916871,0.2860982079426666,
 0.14462779446054452,0.6679907716161657,0.15080285363737533,
 0.37586969730218267,0.713681535745102,0.10840750254906395,0.1309904398360433,
 0.6718926080708209,0.7426352052108072,0.3212416158336513,0.8893275729764208,
 0.6019805401705375,0.893579350442042,0.18351890157648787,0.1479372284118765,
 0.6682162710864654,0.38909755953217284,0.18396250651002788,
 0.2142833314608883,0.7491362933735307,0.9527764967223286,0.8954558693004239,
 0.08803914532019941,0.23153617090484802,0.13702137905397016,
 0.44390067021968127,0.26371804852205105,0.8221552223457632,
 0.6327336399393753,0.6797968772364378,0.17291957386377,0.8302681441635519,
 0.09875934387390906,0.8555160160497007,0.4972778768199654,
 0.16921391507255534,0.31553831522880027,0.830830579607618,0.823678053035395,
 0.7336748942050797,0.6750867571733894,0.9761283174626922,0.6460281171536363,
 0.047609863386827156,0.8238260096244333,0.6783348455795531,
 0.47600849519307986,0.8219092771852697,0.6157908227595538,
 0.09235050400304701,0.9076369259498123,0.9635031536551459,
 0.11054022763424409,0.8504943092531683,0.33764567026593895,
 0.20144073530748174,0.5473776313462391,0.38085565422378587,
 0.06872087380081715,0.5467057879656833,0.4770963196514638,
 0.48215550682801933,0.03825895055105277,0.6512170691314058,
 0.7293274855056124,0.7874585464665775,0.6168297176055288,0.42743477110822936,
 0.6958577112639738,0.3290798181006288,0.7126125629384251,0.8938273269591271,
 0.7692256455452037,0.8527373838302001,0.3606329205824771,0.8012576524346091,
 0.3038669310667008,0.4187377694226042,0.3253128983381294,0.5926760250629582,
 0.753960665291706,0.5831591589706823,0.5169045935613057,0.9945101392527252,
 0.11069498316711923,0.5045887293616461,0.9851664388202495,0.8903144269830323,
 0.40086475245299447,0.9941815116919676,0.8976839762892769,0.3257187845164805,
 0.2994957112499216,0.5120446487458654,0.3492879103617086,0.13542499204042358,
 0.5560492826117286,0.12769487329727136,0.6077595297293573,0.4013875594274561,
 0.8320866648247934,0.35185541514491636,0.709903666483463,0.17550793986085955,
 0.9329176431232652,0.21630345326894485,0.5188097494270123,0.491709465075046,
 0.9954677861499074,0.858082679683857,0.9037007668047244,0.8537682849902894,
 0.2142888140255732,0.2217950632123955,0.6094624282708029,0.24952610766861583,
 0.4718847498650631,0.573561501966146,0.11040744573708483,0.2009578939822908,
 0.06273654908490234,0.06498147750700756,0.7807031571078902,
 0.18976281124067618,0.10404196709189117,0.13481289023203158,
 0.790004637067176,0.9818071802683821,0.09305468867116518,0.6838783335816783,
 0.29583335078988626,0.28092841374053834,0.01838757376222122,
 0.13828353573939522,0.26070189151624545,0.8105000291124532,0.36324160537549,
 0.9354106743689174,0.9068213096104671,0.15251435151400428,0.5255647608994423,
 0.32175683976880665,0.054857837847413826,0.3707072890814471,
 0.27564817505808803,0.8365773505397992,0.40106781854383633,
 0.1924248734577838,0.07336965306108845,0.12167885804755885,
 0.1113280145114307,0.8765090726470129,0.23434460655673184,
 0.04172303021515955,0.821573898075423,0.8122700621960417,0.08719285309883629,
 0.4049947017001553,0.4739891231868889,0.030437691286351543,
 0.47804738227000665,0.9274742630339795,0.11478921319833213,
 0.10686327028899822,0.729906176068877,0.16488481460390259,0.332697406911558,
 0.7355883174940578,0.45247876940890464,0.4966886385404482,
 0.10844010729314713,0.1302378924189863,0.4268089838838379,0.7241056400714433,
 0.6788886464758938,0.6777217970925605,0.6315847678816113,0.3385214524180651,
 0.7105482984759717,0.5605935462990355,0.14025577057497063,0.509540023452744,
 0.07696298235189192,0.013288084081755303,0.3435680514160314,
 0.8575081468111045,0.3666818467132167,0.9650041703481665,0.9999538651414077,
 0.5663664449899,0.9052005293140665,0.8837941991248234,0.1737120714987641,
 0.41777885671979087,0.36457774981632984,0.5360518545491033,
 0.9620952928996063,0.5523170062369589,0.9487995245784228,0.8496387367837839,
 0.29730464607522555,0.4286320374395257,0.8749780309469488,0.3603062262637019,
 0.9885581815602599,0.9214448870214345,0.45649096758798313,0.7645178501704871,
 0.7531982444373252,0.3107788594436768,0.7122453276049195,0.20123860905389956,
 0.6844826247932018,0.3572249286587439,0.5716696404507681,0.8638091196604439,
 0.24048380914313572,0.3397207462853552,0.8768917683575106,
 0.40664100401368464,0.18397675659171153,0.109482679232911,0.9596138507852632,
 0.8557419328508331,0.6135290434873552,0.9602702561503639,0.8014557990831613,
 0.41837391410145475,0.8832141549566888,0.32859034454943936,
 0.8442216735341487,0.36219892609175075,0.6388632414835972,0.730496120347004,
 0.20024929178120598,0.14674471919724208,0.7929356888415116,
 0.2925385577208818,0.18502063603116925,0.6776223688597602,0.5466700115705537,
 0.6920990840297971,0.8169819562189207,0.8665378933389095,0.35345893328155287,
 0.7997281480610143,0.6545421285394963,0.4278596422689105,0.8672870064023849,
 0.9947689090258308,0.7974784983641945,0.5277823199469663,0.33814401588427645,
 0.9705747813567656,0.19664606321015887,0.03633309251488792,
 0.8947861547511607,0.17508535740564934,0.8746298427498582,
 0.04923875360245311,0.41917448873053287,0.37387983807794756,
 0.259173220316142,0.32554438829409227,0.20753787890735598,
 0.37799836093019157,0.9044539733064375,0.2517738029610621,0.7058871369381036,
 0.6521536658820338,0.21763588902836994,0.34395045498027865,0.527305149639324,
 0.34131446933062115,0.9437951764857664,0.367400844831989,0.38217309450978676,
 0.16728564175756488,0.6917045006465157,0.7946244996103514,0.9576442209597243,
 0.3841610113803722,0.4315996659873337,0.44577794067316323,
 0.06508797186083493,0.553558453742695,0.2738035076119265,0.824035556155375,
 0.18652131031578445,0.3278210856244341,0.5930402070697387,0.8001784324093973,
 0.7986896053624912,0.980301312799707,0.721197027902667,0.9484043488806664,
 0.288521901678227,0.5261397682537411,0.6385420113175493,0.670931879086053,
 0.2646801741561704,0.33958045420144833,0.2251717724624932,0.1605120049179073,
 0.016697025537754095,0.35246411247837073,0.9327157772798049,
 0.8851407896679473,0.20713111820066465,0.5113111319553264,0.9599162547689256,
 0.08765372853902986,0.6339829372589664,0.6561202117360972,0.8438465127895408,
 0.8514450241834826,0.772145098324756,0.8976500786323443,0.02272385738241156,
 0.47839237715165117,0.8097556349277616,0.9983957854137098,0.3298423583660087,
 0.42229261850450317,0.3346124780795968,0.6772258208872524,0.885490796638662,
 0.9045907786614168,0.7623004322947321,0.14079374084030594,0.8572927666146629,
 0.35714071753659404,0.31115439163821723,0.18151282422639337,
 0.315090235113087,0.6085143484221046,0.9802065844448166,0.5910320883637159,
 0.9458764980647283,0.41099880604387395,0.25476137297016443,
 0.5258710160687345,0.16845632088369245,0.035099773182748484,
 0.5023882789071173,0.12058608616260369,0.8387384341369875,
 0.010606487704942436,0.6540883945098763,0.13453570658078762,
 0.29648359692462245,0.7970459949944791,0.15733245175974497,0.522728777153483,
 0.7666225898729926,0.749269088242318,0.8260920258578456,0.17782124633726215,
 0.1523703377449922,0.7205006512669715,0.20169780090996603,0.0325056121887386,
 0.32011979516483047,0.5255335665033228,0.7655478483106637,0.6422579192438395,
 0.5863016952299489,0.08279935926446147,0.9072796643533304,0.7962907945550555,
 0.8556400036768981,0.6172561349493734,0.07124537881205628,0.974511116517123,
 0.12838617187201917,0.7942098861272772,0.9741799808317864,0.9049660779482782,
 0.45606029915082913,0.22076821879308195,0.05368812076980456,
 0.6845863097157866,0.5987813971955053,0.21994667300487158,0.9082126956144282,
 0.15610593990041288,0.10396982504923535,0.12368603998006766,
 0.27405118265286865,0.0924120249888446,0.6864641691580967,
 0.05678778311889021,0.17873262587063354,0.997430333296348,0.7469039393685509,
 0.5810978620438806,0.7049743103027077,0.4968663168208217,0.9579344252994707,
 0.9213879815594732,0.5944464728636685,0.31788494742813445,
 0.09805027671152167,0.18504347053535264,0.2810550057130804,
 0.4848764830702642,0.693577709405421,0.4313838335595285,0.5192879162461852,
 0.41785206612918013,0.582437621901959,0.0630463679276072,0.8493735633513461,
 0.392918227184869,0.744291881757518,0.10607229531226303,0.5923132104980509,
 0.3516722736162211,0.4377319259185194,0.18330035438406422,0.973626790507622,
 0.6520211344753246,0.9670824297249948,0.4573113065383484,0.9675471662700127,
 0.470136150286916,0.7290866758349007,0.31169819073670635,0.7954712026027673,
 0.9053810179505066,0.20111587416449517,0.7365558178296338,0.3664630661857665,
 0.5829962591381804,0.20046849059479666,0.6830538141138653,
 0.21256902776436826,0.6586565558735119,0.5896525828591906,0.6169517480536904,
 0.24811193534852027,0.7444598135159103,0.01446360436163352,
 0.7344310140715283,0.5932133331585656,0.9181539534106007,0.8186770081874416,
 0.03738404943048301,0.5464233998105172,0.7052001898438081,
 0.17071934485993945,0.14178859803238464,0.37061754427620686,
 0.9223754173325657,0.49592444152012805,0.5316022574318597,0.2867788216029137,
 0.0027531597576605504,0.7802147513916219,0.157007817498777,
 0.6193135281123388,0.6001802073437789,0.10285497374441965,0.2556304702255483,
 0.22963342227505068,0.24851807968728612,0.8663465324265832,
 0.005043594424697373,0.7628535118904027,0.7522271715029127,0.279220184071313,
 0.9437249942887996,0.7211725398018802,0.1983643010489995,0.19773812897225007,
 0.3728124423731122,0.03890821959459134,0.43348608021393853,
 0.32031469384330935,0.7054724208590948,0.7697739182730929,0.6591158418163263,
 0.4958783031780577,0.6835459906472207,0.9148118140302773,0.5139818073456521,
 0.36059022013214503,0.4390899723662335,0.20896082905117475,
 0.9758668823819913,0.3356841883933005,0.4805505375337652,0.811337061510571,
 0.002751523361839703,0.9972188720546256,0.06800277865901538,
 0.47141627200376013,0.926851076239467,0.4974658480961258,0.6849273839252566,
 0.013959441639914583,0.5752545623933155,0.036470060199620624,
 0.9707972251710386,0.11496412726802052,0.7837392497535748,0.5053717046285854,
 0.06285062169327704,0.07364768709354719,0.24331717505046213,
 0.09541200177246412,0.6895391329886182,0.133779587948768,0.4117174261393316,
 0.24465120642897265,0.9166477021205143,0.3404211306432856,0.1363749240434161,
 0.4215213004253596,0.9446487496627889,0.5776132297398692,0.24220083123446245,
 0.7510235689948287,0.2912993621437797,0.6794407244180795,0.3872208167326393,
 0.5582680963362627,0.3900715985257047,0.8900781307645225,0.6166050868591614,
 0.0865964972170783,0.5485360137407207,0.6745245466597636,0.8957785502382068,
 0.21128657271048246,0.13781192281585675,0.5547663567643027,
 0.7194988447513536,0.8105003025423144,0.9981583549256025,0.6633068402340294,
 0.12811404244193858,0.7746311686984443,0.36351681834738203,
 0.8852232201708127,0.304896824369715,0.019203959572406615,0.7043224490166444,
 0.7840397747732114,0.9306009725013226,0.15356283281807803,0.524013842431726,
 0.9913504320918809,0.5533102332336405,0.9497014464858982,0.7440957650767379,
 0.034930797849827,0.5236259221576248,0.4108262679795136,0.9562331156847169,
 0.7663340620419692,0.6238993169100178,0.3173120671829255,0.3999440043385558,
 0.10334274699191326,0.2177545584594337,0.8539236098563651,0.8687416314428492,
 0.431108236643874,0.07481914493252695,0.6007338883623542,0.1822292661142474,
 0.341493978004907,0.7938606532484871,0.3169394222324504,0.9769252173711195,
 0.8256114076179686,0.21373820574397795,0.2379178523329556,0.7147419222898006,
 0.7239679794776696,0.737103428371978,0.9418669644702045,0.8105827855653138,
 0.4154295930500358,0.8987087458880465,0.6290246428883655,0.30281620892412264,
 0.6151041160586149,0.8681617889698635,0.6712013067685852,0.2051648580518204,
 0.16881077324061522,0.04188801960923538,0.7533501888722791,
 0.05988257588417356,0.3937657321771243,0.7318154681290829,0.4112724677657601,
 0.322605204008119,0.7832428244123721,0.5930020501909578,0.21692865838053121,
 0.36672132438699345,0.5854051073308587,0.20797177363860164,
 0.7970238621103299,0.5071608286008721,0.7909918457600132,0.12635811962799304,
 0.45323028964927925,0.3612794237797817,0.18537505034999002,0.455852713885721,
 0.43023069200236475,0.15191647008409837,0.25980469772222414,
 0.2782346595889209,0.652760315867972,0.9161683825104365,0.6758766070588785,
 0.7837297349966441,0.5233014416182514,0.39526691924033186,
 0.43636681681102973,0.6617463723946313,0.17738898588083696,
 0.9422639853802659,0.28429774993687573,0.04489913040585514,
 0.6621482318971504,0.24941361095868642,0.24210756326493688,
 0.8814195004832079,0.16720443965115828,0.8799955179061221,0.8602255354803268,
 0.19476238646692223,0.290028897191674,0.7322056132878276,0.5974863455158066,
 0.10499580616025916,0.6970678172965308,0.6734991395780556,
 0.36504200107352247,0.6263733117049056,0.038831297452612046,
 0.35929780961148183,0.8443533647784901,0.9449730884568189,
 0.43296484464128837,0.4606839027969962,0.5685392463189302,0.9628060305650875,
 0.9483368662773359,0.34287411802866086,0.9331063493469092,0.3565962521913114,
 0.8991640928673343,0.01294499126456472,0.7948761621004263,
 0.08114905159482877,0.5522579379723148,0.2076205562786333,
 0.19672859141879595,0.6243550978596601,0.3114899567295639,0.8079218798455539,
 0.6559768003125923,0.7134884179994526,0.9731927225553374,0.45176499177912144,
 0.8111870959862109,0.034019332534278024,0.5044517901235228,
 0.9337384981527475,0.7827114441285457,0.9864736588397306,0.2431907490935803,
 0.5741015267642154,0.8224699552500115,0.21503466829442996,0.6397709680709254,
 0.7960339022343125,0.4003350641734399,0.4349858198710742,0.6893115716699678,
 0.3428352472220969,0.20225935396777428,0.325405349967669,0.9238188219693513,
 0.24235053910139315,0.36521018174374986,0.1834257611411636,
 0.7516870255658553,0.8818583246317351,0.5478492900980771,0.20796556794221965,
 0.11710936049298726,0.2419598937961086,0.7740851801601023,
 0.13457301170977853,0.4385588157294884,0.7064737952061866,0.8054793657390086,
 0.565847578010973,0.8342112964676244,0.49228847022985145,0.5377358889761614,
 0.6184165802390259,0.6461228949460514,0.7423636365329004,0.5000919017560661,
 0.8547605488518809,0.14114637188586665,0.926926693954387,0.49430768133594205,
 0.06936242738036724,0.4383432521392221,0.20705068524898362,
 0.6875145293664487,0.9813959443503407,0.3737791057230929,0.8586580912883753,
 0.7470067934297813,0.5214000985178399,0.46186868956492966,0.3025940537703632,
 0.8603197640495647,0.8358502363850696,0.4961397357167965,0.5755729651849487,
 0.2917159978847427,0.6232268198598943,0.1322323356999755,0.31373269105419066,
 0.8604643581303903,0.6221317320192866,0.9123213308229552,0.27815738864873085,
 0.3250314476909387,0.040582593063355654,0.6550302649862578,
 0.4725490792198138,0.9525261790753319,0.6178882854962211,0.516363669700825,
 0.9563753619731169,0.7418209727284241,0.75952209797243,0.8898073362866533,
 0.002072271543556714,0.5296673921230529,0.11384765236749395,
 0.18644253796917898,0.941530481706472,0.9638792555992781,0.8881555746135678,
 0.11324860719235985,0.9176420248304104,0.5927286433747646,0.3802156518114521,
 0.6036735750448847,0.1579807876240158,0.8498760414445342,0.21610607368385804,
 0.7073644836983126,0.6941885370750236,0.25002380405969726,0.7394248870701262,
 0.2819687270838993,0.22878886610077687,0.5465325982490374,0.6447496185724543,
 0.35016415998275063,0.12528949733670414,0.99183385955998,0.813478932717931,
 0.662290216792409,0.28699992878462866,0.6824883000943489,0.2659577108025807,
 0.810418828542115,0.5755498619388747,0.9498118711425395,0.9976700536257878,
 0.086864778961118,0.12849333073398428,0.20891172506265443,0.6495859200971273,
 0.12413282161433836,0.5477600576793156,0.46466001785731614,
 0.9321560141563715,0.7932603079193786,0.32829060354822737,0.2412543573157464,
 0.5999699808011388,0.04316856380158929,0.31571919453574615,
 0.020372992460030548,0.014651312902509872,0.7870346508698152,
 0.593655052524317,0.8291829146141318,0.22797915947709402,0.5451858491054115,
 0.7305210679128732,0.46255167832691324,0.07336435560090016,
 0.24146984921272596,0.14307891661671712,0.4248970334506351,
 0.5997146453221276,0.8867541238828842,0.9872574702814072,0.43037092528941867,
 0.9212791777470164,0.46543308159978547,0.632340470702101,0.48694217491198877,
 0.6725183424757225,0.43039133117113215,0.9686618272326909,
 0.43240220399239815,0.07122259919005314,0.7276947550927032,
 0.3034221779173998,0.24434349838755598,0.7396598099750388,
 0.07801991453749701,0.258488083220322,0.9875587209973022,0.29648496039524375,
 0.5274905338737061,0.10649244674010983,0.2199106905809065,0.6548548633280774,
 0.03900681353149715,0.1444000773369214,0.636543957868122,0.04648157950953469,
 0.04230569049603883,0.3784019299662692,0.26519770192753855,
 0.8397376608373972,0.641403880634892,0.8202652797888225,0.5833289778909982,
 0.6831119575548743,0.3757291625337058,0.4648061319617053,0.8859567341851143,
 0.023326482392218906,0.8348555217474541,0.4482494882842627,
 0.7440417488259135,0.935985959556868,0.1960978050783556,0.7182961121460454,
 0.6566610066540215,0.827882056292633,0.1477391563613203,0.4530931142206781,
 0.44255290750716825,0.5703393424307681,0.28602854004969236,
 0.8906365700257807,0.6739290976334282,0.9217275677123278,0.18743917408282407,
 0.07001108043365756,0.5576036102718808,0.549790387818386,
 2.9210952900060594e-4,0.8397832048945447,0.15322240570576617,
 0.031707404230811975,0.47894046446579064,0.8830346331027895,
 0.8880392206589132,0.5846221755126576,0.4898492745242281,0.3923641244894075,
 0.7052854932414002,0.37277782142912774,0.07586482939574801,
 0.1200145156751461,0.06918968266618508,0.018727286434016666,
 0.24685276999121686,0.7091783223322266,0.8516371440517353,
 0.27597439078392716,0.18697463202357156,0.8553502381230919,
 0.2349766813384785,0.8094036684095777,0.9391372139455157,0.7148637286765288,
 0.19007360653274719,0.33135282524017573,0.8699158778337622,
 0.3100132640573019,0.8386410160718903,0.17928932188809077,0.6137285481832326,
 0.8772856458481245,0.16267119262552243,0.9901561308753136,0.4908823485189141,
 0.2924631772580709,0.7662139752951006,0.012874152471226585,
 0.10046508965817713,0.07907626097006337,0.11416117700167339,
 0.07015041456801929,0.5131121704615589,0.0973855605744759,0.5922959212273909,
 0.6383785280106159,0.5830736453628445,0.6981325917094615,0.1487237396602965,
 0.9945299526622432,0.6346981051824566,0.8826117140557145,0.6758527038979527,
 0.9108622293643267,0.7911457196373274,0.5098113440164079,0.3163547673801642,
 0.4023113393966199,0.14993875895488384,0.6986345100071457,0.8586364057613628,
 0.03639972849154294,0.4995559857011318,0.3979540478219501,0.8252047099125756,
 0.6904837426871598,0.36056421067546385,0.9679966535730568,
 0.17299315302504636,0.5194976825964708,0.3644733714913312,
 0.17485065214976891,0.4803771762599237,0.7094863928933645,
 0.48274693869469953,0.8250816042620153,0.3961522696744182,0.9674284938901425,
 0.9530001656064966,0.383888332078425,0.49417450698103016,0.6285650026564565,
 0.32498436759534044,0.6742127779297573,0.9662598307450889,0.6189367769014502,
 0.03456951110977824,0.11525248151958245,0.5720772876346703,
 0.9768233585813244,0.9122596494387212,0.6729491458528661,0.7595681242249077,
 0.15927946098146295,0.9785451020187342,0.9644801679464567,
 0.11695331822377442,0.45937998883449194,0.69084098369094,0.9805855027206495,
 0.6728179319025932,0.29510752567682574,0.052737868535393306,
 0.4395323504307276,0.10336357032163601,0.19395151630231877,
 0.42614784870054756,0.2499075375158306,0.944545407833038,0.7077281251854277,
 0.7671564036663927,0.6603090433896556,0.14335876422803984,
 0.09121761449411125,0.8225195520371004,0.30027626971286236,
 0.29072056676326963,0.08767230716317409,0.09217390933102554,
 0.1387352140140805,0.251852900354668,0.1824920377516237,
 0.0021977134792801056,0.4829443239920421,0.3014285101115085,
 0.34725124875578806,0.19072739702080088,0.9421722799000054,
 0.5047219485254306,0.07844841399275804,0.0038800933422914463,
 0.6531442389302162,0.6863502766300078,0.01654780825122759,0.7400171192714358,
 0.597024086537851,0.6591348363301778,0.06656636132860316,0.0736811824298117,
 0.8509555230435024,0.5633520879454407,0.6550664496931599,0.2913749903685202,
 0.030784925842550437,0.5541770337137861,0.7347971401919751,
 0.32503623855577146,0.6885487753487911,0.3189243108603357,0.4979742508099533,
 0.48682099396428835,0.21722692863981885,0.9771989457118974,
 0.10613951502704244,0.5962457115951592,0.8979017363283198,
 0.030633040711332882,0.22851633293367712,0.4685066526990501,
 0.3366818312661807,0.8826104277111658,0.6773771989177103,0.11605370712910823,
 0.6165382129156725,0.6964738029705986,0.038345568947063424,
 0.10769152713808039,0.1743962944441111,0.7539463059186501,0.8145249089559405,
 0.6909232780417489,0.653351588258037,0.8717635638693065,0.8554129983894214,
 0.8838092548421673,0.9351277807969235,0.44503843977794044,
 0.41333377788055803,0.6359591816398551,0.03772428929256355,
 0.35050570800268366,0.6495668155989476,0.13221204476577908,
 0.6160217027097512,0.331635382365543,0.8277546730828481,0.34571981153349174,
 0.7573139251243167,0.3815540727138328,0.6569280151548692,0.642498304474527,
 0.18676447389738426,0.8939147838516682,0.6763893427550526,0.662206841905714,
 0.8666563686833669,0.20739822621253956,0.027122827699236685,
 0.36372074500419327,0.7739448859252981,0.531225892283025,
 0.007094982897194235,0.1060593770705549,0.22095130549287822,
 0.010282787952141992,0.5786581822282404,0.12869310375050913,
 0.4780962117719112,0.5866154078617342,0.704021817743912,0.5187389797974379,
 0.46264059545024383,0.9887466475015944,0.31919622687225524,
 0.4953860269337321,0.9760050829047419,0.5448938799853162,0.48657918098280006,
 0.422794884159049,0.5592873982722005,0.5964071624104427,0.5257030256817757,
 0.6003734937578803,0.772016310816986,0.37299998426119874,0.48347284039380645,
 0.2516148229214308,0.1480227610209477,0.5301578420889237,
 0.0028132228295614237,0.7249363309418402,0.6200933093174057,
 0.47812419008520723,0.3261549751044889,0.4280010679647953,
 0.02167386681823147,0.07568388533326131,0.20546297781578082,
 0.3795199699730297,0.5365349675613995,0.30456402802482707,
 0.07326933138034863,0.00210454928837267,0.17649663907694835,
 0.32092944451594363,0.47288430775293055,0.5798592733631345,
 0.6302666984607215,0.9882613107860795,0.5302389658306309,0.2633573817198205,
 0.5474901426710588,0.33366251595990515,0.583443611490463,0.5024901987491104,
 0.9051658075410643,0.7765962320243054,0.07507643934831898,0.3441935002783346,
 0.286133763849891,0.6679497105429326,0.029847150040619774,0.1710897407637535,
 0.681594175411677,0.3440516156652088,0.015108055418773358,0.5451648482990054,
 0.12008880941480782,0.5142815169791497,0.3655396088021279,0.8813079999265262,
 0.17487270853492065,0.017332020784614932,0.625623310819567,
 0.3477542486184174,0.6657246287829754,0.9959615943546526,0.4223208211990839,
 0.6653902272679394,0.1049334656041534,0.8622078078484418,
 0.060069391198745326,0.7110596980267826,0.1630265220156335,
 0.17996026051270075,0.12378868773333618,0.42832660085897956,
 0.5459727451512201,0.2713860220212093,0.23284393924106894,0.3446938567705923,
 0.9463906424889543,0.3469811154718936,0.9742916082149466,0.9901870952028107,
 0.7499499038381592,0.2674455046602675,0.4934359940630699,0.6078181342746791,
 0.5335171622213059,0.7699884299454773,0.599256791294139,0.21590867778054934,
 0.7360131004139293,0.3827885883701946,0.387170254690433,0.2991568236949127,
 0.4779277750922516,0.9528957590769417,0.5530437240772925,0.35495032861566134,
 0.9232113257543075,0.1189488011123534,0.23770993791416417,0.622714479376272,
 0.24698976155765906,0.9321574865899723,0.5895655261448763,0.1688754717232186,
 0.11633729003250115,0.6876107242688878,0.7153375812349058,0.8702445059386521,
 0.6384846252820301,0.7375380015686062,0.06023813916879872,0.6299848728059163,
 0.1662817029791408,0.6313755028371836,0.3562068721769228,0.9992688241483021,
 0.6662625060954832,0.5886064302541818,0.6652857935557759,0.6368898550598063,
 0.23984037666231384,0.04585161411879879,0.9215125827248629,
 0.9155537804083892,0.0654697449263848,0.5559121466265755,0.8267570799019186,
 0.006530000082460852,0.31592013191047785,0.2935014533702305,
 0.8605920995658197,0.419596384361144,0.6931730928024374,0.3029775849888875,
 0.7193584518012426,0.45443181723278014,0.2855659423092952,0.7409929946654018,
 0.10210237198489347,0.3987730618644515,0.8300577841293132,0.285167096193477,
 0.7592452026905763,0.8898131470158714,0.3109806662744955,0.4761334884814219,
 0.6673282063382794,0.6594415271581711,0.8779246230051032,0.699082931591438,
 0.17755876588474606,0.8302736448070794,0.7655430066450849,0.5303181656445576,
 0.6793237380181782,0.7527813188476492,0.04153272514140616,
 0.20686773033443406,0.5245640097040313,0.4087914153029961,
 0.42369566335737263,0.8840125275062597,0.20015820216072555,
 0.14119528752042487,0.17288517599826037,0.1941260906256801,
 0.7926280438458398,0.0671072781688965,0.7333810968060139,0.04432692754843026,
 0.40735144991352235,0.8141629113255431,0.6330394318428563,0.8697407249930041,
 0.1766228946925057,0.795239713974785,0.5844714446031275,0.2541069058169455,
 0.980035702048793,0.5413644885952112,0.5078807743024658,0.9248738877836944,
 0.3263461241632817,0.9192930759862715,0.9131977532712297,0.649385238686644,
 0.6200834893965758,0.053990632336264555,0.9272064650065448,
 0.6771063441123993,0.17947274215993092,0.7833175168747915,0.5732827299683515,
 0.38164413900934113,0.3138445197674765,0.6768627506049585,0.8757781387414721,
 0.9705901519584503,0.5092422164654322,0.2126223136582328,0.7408416540626258,
 0.2114368166167755,0.5068972967587415,0.7187486556965355,0.39734046950057045,
 0.6306159189317955,0.8105057902651295,0.8853425522897175,0.27082623351830326,
 0.8961388835469811,0.3586400636452608,0.6928955570127985,0.6296596205688453,
 0.22747806317017227,0.9447903518928751,0.942289946045088,0.14716930558999963,
 0.8605819458019772,0.7201631468154382,0.155398830506415,0.025812347181704887,
 0.38515000331496474,0.5199947428068631,0.27661761724659595,
 0.17459567840286416,0.3845417653286578,0.43845917320942096,
 0.8385293244441132,0.609436785741925,0.9287619880928317,0.618073689884715,
 0.42503982885759717,0.12910670414874215,0.6034020239753616,
 0.04805504094599955,0.0684117716467667,0.637231867377262,0.7869442817380484,
 0.13611773240275482,0.028911133063063366,0.20870139394913334,
 0.04369196788002838,0.4761345791009962,0.12295011149945823,
 0.6861242242399421,0.7657380462586848,0.5930629202907032,0.5910620709771384,
 0.5699984461379177,0.4843498003083695,0.5572949340349899,0.3140504067351355,
 0.31727608063318957,0.8619913027180437,0.1890058655703425,
 0.29582743565347547,0.5646420007426949,0.7067143106899376,0.352945092022001,
 0.8667558866905873,0.970681800289128,0.08433372696646646,0.6407741645294168,
 0.3725035589931103,0.7690697343959652,0.9290613232992874,0.3632898009656813,
 0.06914185429112929,0.8089319643318098,0.5571586086898446,0.6344260944892495,
 0.486106672016699,0.23665066840613191,0.8099377140930226,0.4825567868258194,
 0.5708222850193374,0.6288114418759164,0.6240545561210469,
 0.012143369573475615,0.2853299690410994,0.8003094673527323,
 0.08306767345266486,0.19636631210524513,0.8606740732706699,
 0.2543613330650534,0.8946013918898299,0.6106296099332599,0.7291863761962446,
 0.9953486237252103,0.11855122900448001,0.23957481831085192,0.750298204367565,
 0.20259673721601845,0.11253087749452018,0.06280821280794657,
 0.17272202609127585,0.8224703835444079,0.7218407385263444,0.21458110912229,
 0.13168128261147538,0.7026616536719021,0.06546309014856666,
 0.4042871438991458,0.164003044814656,0.14077401935045908,0.5292662721347378,
 0.5295436394972656,0.29173277612334336,0.6873489051677553,0.7511114674294771,
 0.7580346307056618,0.6422186101758777,0.5823599853768372,0.041815327373524,
 0.34020398568359567,0.9395352314052355,0.2135630159203712,
 0.24193496534592596,0.10996273415103741,0.3783692424334325,
 0.6944571241254668,0.05874934709141166,0.22646862397968381,
 0.22132131998146254,0.16378312241075998,0.3228581504913449,
 0.08038579209851282,0.5290078262761617,0.13557516076411158,
 0.0356126283507443,0.15357616442408117,0.4895909244784751,0.9509182344467376,
 0.5002107265378365,0.8510978505442262,0.13163009945331938,
 0.029335611751645008,0.7916190325107685,0.12469060501933205,
 0.5875484067795455,0.30048545196683074,0.4262664178506903,
 0.16674449939954294,0.41177093302032464,0.9609231936238208,
 0.09219352348155219,0.40316866499043247,0.952703081346565,0.7986939402270355,
 0.6442562067086619,0.7494267010365708,0.04287555064330306,
 0.06985092075090782,0.7213129928535453,0.8548542676037101,0.6263401509684189,
 0.6635965594743531,0.20949382176347264,0.3944711376420035,0.678443736434085,
 0.9620943192919532,0.36351625186187086,0.6724226297514113,0.5053845745406081,
 0.3351714263985748,0.0757847874739449,0.12579717657154077,0.4727315550323833,
 0.29509681614356387,0.42492390441833583,0.04736335780082457,
 0.3517912292298966,0.9999010576399257,0.11180066565518054,0.4060348313348081,
 0.637799366805278,0.45339348553209713,0.14327159920147425,
 0.22146219679548673,0.18934527168886672,0.28794019037388996,
 0.7778325930891922,0.8158557301955017,0.5345972823029482,0.5995198498791964,
 0.8242556760858601,0.5881219086972582,0.6482521729911762,0.3868544485478457,
 0.07205859506095658,0.9790887215968506,0.7630815201844181,0.6065165108732791,
 0.16134172184119977,0.6317151303044859,0.7168755409696024,0.5797547037212615,
 0.24830146454018498,0.11036461413231091,0.5502616390424737,
 0.25139565699913147,0.9370546860096953,0.0781830440979221,0.8136081832195073,
 0.9772554518101617,0.13014182353795656,0.6042904399564897,
 0.04471253335020253,0.6936800651889551,0.6189838508623551,0.5462664331025144,
 0.7453743762330804,0.010364322026086059,0.35034874630935153,
 0.32795120594570293,0.9362591358670351,0.3849355087183659,0.7252172395833328,
 0.7436595350682285,0.14308283484835072,0.8000292456609185,
 0.17800693492606534,0.4762035234676153,0.25621375540582314,
 0.10829716071165707,0.8222435353656905,0.9381696384197015,0.9589958897586013,
 0.6211605016344769,0.006802984759639585,0.2811697465459234,
 0.5025503299631969,0.15592679588673186,0.09150109311495791,
 0.7359827958402445,0.350939798949788,0.7924689275451803,0.6698657379978558,
 0.6348210141422047,0.7695454997946045,0.6239154301590755,0.1057527816015239,
 0.5510847754424298,0.877814312915683,0.5967398717143683,0.9806098891808811,
 0.6656747668887956,0.26426690951905407,0.36025396254702857,
 0.4902447785390578,0.6990590805735635,0.6824144656367919,0.07099712546063797,
 0.8457144911174491,0.18763001253070688,0.23470751584539384,
 0.6928020441610511,0.8512022997336551,0.22324757064755818,0.6393713261460917,
 0.38270967622196184,0.8220031107909225,0.8543424204099465,0.0993705925021734,
 0.3919529683002544,0.9030397325267332,0.11731527964427224,0.5954360445838183,
 0.5317554345527642,0.1461078414072845,0.9797497045358109,0.1626341185078226,
 0.9895815163407394,0.8923673478979576,0.760841074212098,0.3364203673425751,
 0.5262404166979713,0.039039247116731124,0.3675740919449658,
 0.9392752696700596,0.41282250930580866,0.902319099231068,0.5492292864479712,
 0.992972087487241,0.8455264268387523,0.7046613279352572,0.5957975654955369,
 0.19091850971669744,0.5625446853976169,0.07799146955261549,
 0.13705800097879017,0.5024073596710684,0.8345297924898444,0.2175774863303288,
 0.812192407728998,0.5960497628109546,0.3277282735163898,0.6100877623471633,
 0.11891821744227005,0.9271532075110248,0.4421919106582959,
 0.35790189781272874,0.6041452210489839,0.33125440033138,0.8804411798667175,
 0.971105592720569,0.8723992624511462,0.46829613484826027,0.8960580040516176,
 0.24559024420824238,0.5398959343109472,0.2741520940839446,0.581551613463059,
 0.030131124231857376,0.20604144359694032,0.6566013139941831,
 0.5968415393910088,0.6804807070644834,0.3310394096466417,0.389006604773646,
 0.2213789075760927,0.449417546675821,0.7182972595877213,0.35315626255795773,
 0.9438082075216432,0.3346518337210702,0.41992093572410083,0.3125971939466118,
 0.6871719626809977,0.7984303235489081,0.4594523437884743,0.7921986056967353,
 0.8431123920643239,0.37530883064670884,0.028586324061955493,
 0.29697428459082065,0.9210109196178875,0.8465056907514809,0.980670877044588,
 0.7697362409857238,0.04350197242005438,0.5921695666505672,0.6568407970179003,
 0.13173100798366266,0.29842448467195637,0.891672848204095,
 0.018074435895918617,0.25561840287000404,0.7757678103478256,
 0.6943397167207249,0.37491644642931155,0.28432032364680915,0.242695896009263,
 0.7486812542122421,0.5117543112715631,0.8970387360065579,0.12172445235980756,
 0.4052106278742301,0.7687827915517551,0.798326276534329,0.21938564725934062,
 0.23815540135823587,0.274484345188479,0.4413163673763415,0.8213708232277553,
 0.4953254802898578,0.2251466517803974,0.5821680650816274,0.6505076119107178,
 0.8477440112994081,0.8266667735294945,0.7205851785615244,0.1311445732429284,
 0.8342804264544544,0.7546935361991558,0.733570782977226,0.5229255826609591,
 0.34773745410444856,0.9562884385941594,0.548678858404872,0.42002467305377467,
 0.2195258745792844,0.22557754771915706,0.6988859219922572,
 0.0017789061772713888,0.834571121697788,0.6334238879700065,
 0.13296962844912874,0.2923350471098609,0.3987789413627112,
 0.40652613950009386,0.24897102016007244,0.264839806209736,0.3014293493561363,
 0.8691970936019984,0.515457421290322,0.04333058699681225,0.2259280081260615,
 0.15178898180477085,0.3109433515400206,0.6740308293731416,0.2858342412928937,
 0.3844578677348609,0.5036663720703051,0.39300039990041635,0.9448471127688147,
 0.8917698554751798,0.16043142684840395,0.8333910529292917,
 0.013490961190922746,0.9557739914150808,0.6967750323473746,
 0.13320721766172205,0.6496501724485974,0.5449475571810991,0.5968515618426936,
 0.5444352900395156,0.5002194121750462,0.7316088013429154,0.19224330795455435,
 0.3994507009124775,0.03367304797668891,0.025836965810742574,
 0.14475694997538835,0.5728470004096097,0.7102845774663317,
 0.39121048458218033,0.9495807262534423,0.06810525445681315,
 0.1415666828495734,0.8252373798930913,0.5515430413712692,0.8706557335716416,
 0.22176477825573482,0.18885159325270828,0.283776709167501,0.3736723080711781,
 0.12786006702013097,0.7054572617440207,0.9908002475962275,0.8041785275545884,
 0.2719754700393141,0.17896106116713417,0.19934871251339714,
 0.5076886402428077,0.259254077499368,0.4207527968497665,0.4283123265847858,
 0.1050711248256827,0.8891374100054978,0.8801338406524688,0.24626577291446572,
 0.8631478932557811,0.09601483846575398,0.037413657547311296,
 0.9290148549992315,0.44888036719309776,0.8874634530837276,0.8711958371726369,
 0.2908081147041416,0.5918278504446242,0.3520437039064075,0.6744359609306487,
 0.08295988423757217,0.279857655543572,0.8369419595252755,0.194178237395908,
 0.22419250349291542,0.7165863355093972,0.47961728019475336,
 0.9504676987912244,0.4878326244505742,0.18858821954490068,0.8597753696840635,
 0.720755065859846,0.7779325902381642,0.773385348617121,0.7342623466231618,
 0.6945304187941815,0.07649338071716949,0.7713428956028042,
 0.10335744303404204,0.6609293120442963,0.35982154314019743,
 0.8398655255678751,0.7872072563444987,0.726414392946728,0.3060349682533846,
 0.4179044848328576,0.8044180099845913,0.9909279472610966,0.8489212055434939,
 0.2548225171868834,0.5011758556669543,0.4153400100346273,0.9257400436592301,
 0.5232426325707009,0.5790793431187687,0.574213651846466,0.29593623415630843,
 0.3477815152727983,0.2144236234531982,0.22888834985240125,
 0.44366940114912623,0.5184968980504642,0.07169841920971842,
 0.7126737335216162,0.36063891240727275,0.3944519939248128,0.7219446302167885,
 0.19080287544017838,0.8314532482954498,0.24657253730548667,
 0.3073807442154035,0.1782596018187651,0.6518150341145499,0.2421657405357912,
 0.7269257391537397,0.8097521412122632,0.20195722759209545,0.9984198504013125,
 0.39800159574670235,0.8384715756936654,0.8641974755143956,0.5248322549171665,
 0.2195785027985231,0.9386577580457862,0.5067400860984721,0.2344892604135863,
 0.026958497365591083,0.7355626937622243,0.11770191830678434,
 0.8721510611585582,0.9895057387550582,0.3987371209415026,0.5454134240869173,
 0.7579983587572945,0.9515228303723716,0.802622824389169,0.9641214762613207,
 0.15711743760153107,0.45041322149872054,0.3522101515859858,
 0.3800182452483509,0.6836123911529782,0.9819698484304477,0.31360873820777435,
 0.026582636180437103,0.6427834979540279,0.6060889922902679,
 0.2366975110167252,0.813077346285979,0.8015708182401278,0.9713883801288065,
 0.8304175849752625,0.9644387317656266,0.3338320919498762,0.4976217381406445,
 0.25946073581116624,0.6370185616298084,0.6636808852892412,0.1893181421241269,
 0.8322141857914129,0.06766756921611128,0.9942071989624234,0.6053260226283498,
 0.5601696334024319,0.13646753835658454,0.575078042378343,0.5416647093528874,
 0.5106000774406438,0.04751841382552191,0.9593292015798563,0.899380225961771,
 0.8419250728525048,0.4656634306170071,0.36293940693964455,
 0.24249649430317688,0.020835380145645455,0.45804340593928694,
 0.2167495500547575,0.3587605568552412,0.19063202913751498,0.8830034850504264,
 0.5092529254146694,0.7537747948738454,0.7581291277396269,0.31517548754487545,
 0.8626984480323145,0.10350083878683536,0.010278634264879072,
 0.861670670341924,0.2737932897420944,0.4247995461203765,0.20340786241338937,
 0.4626321994824867,0.7077393361896567,0.037304703058349875,
 0.4971844248152011,0.5400247695359689,0.4099368832165333,0.5865154465618556,
 0.22953964585848396,0.9504729819726206,0.6074816689755518,0.9801159668541273,
 0.6603746935776347,0.15140792951210102,0.6389866586849559,
 0.21938649162665746,0.9720842228397745,0.19715433828227347,
 0.1415537286387788,0.09635783961983235,0.36283866788471897,
 0.5270984515464168,0.18321418824629676,0.19413547452820068,
 0.3259861681594882,0.5449058168862139,0.6961617542266451,0.1752812761275302,
 0.9504758463392244,0.17309937964709854,0.19981566419032515,
 0.14091394520934386,0.16263404651931979,0.4699309706190711,
 0.011807416906147394,0.25280044098611687,0.5030384492888251,
 0.4832685195278954,0.824767902879804,0.6929859502338198,0.7529417253027336,
 0.7235317602684453,0.1424465571306719,0.8396703671069824,0.7722884134230226,
 0.12968070160746037,0.6474681285033409,0.4682574133665195,0.5625319480282596,
 0.7378465333016107,0.9068291288067942,0.5339090516377571,0.1891818461035697,
 0.5267918396048268,0.10255957429855544,0.7831936662326706,0.9024801863183177,
 0.13313938518258128,0.9792540070917535,0.2308956724341269,0.5636633043269599,
 0.9532279995935021,0.8867278266638848,0.0684623708962,0.21418567264903243,
 0.20726269927881635,0.8735563772290401,0.6929356969663973,
 0.09868043647691827,0.19281172449197714,0.9889350908058133,
 0.8967893198103815,0.3453723201049149,0.6216501365652729,0.7289283330238521,
 0.4697743235851589,0.004276377643581242,0.893101206642328,0.4979930249080816,
 0.340298014494432,0.8806486955062766,0.1000122459756867,0.9164756933264386,
 0.8840585449646652,0.04644126811155447,0.5846780156190391,
 0.20801279227218672,0.49728914227779486,0.12994128029007224,
 0.20077076977975072,0.9298605959358792,0.9211007206801466,
 0.17310968161701668,0.6743188997245783,0.5012587756517051,0.6651280547978582,
 0.005208803690105568,0.3969852268314622,0.9918850203624294,
 0.010130700446645102,0.6571418799492306,0.08853609569593168,
 0.7684151803834193,0.21470632389870337,0.20247350374227457,0.583948253310348,
 0.8442150405325396,0.06694066905774998,0.1449699104207347,
 0.31558997642834175,0.05686930987686456,0.438396276787244,0.8408560434978998,
 0.6169846094916348,0.7284248791294823,0.6118965647890705,0.3567466324374524,
 0.6691802719926394,0.24902507024424236,0.5728832899221719,0.7387932337906766,
 0.8701224864160466,0.46763872200177137,0.7462988251029278,0.6183582791293983,
 0.6726153351142538,0.7895683224510224,0.35887466110960753,0.6178144933641894,
 0.13848513223976278,0.8844195531566772,0.7710499940576636,0.7501914214135196,
 0.6147237410412951,0.1074575782137992,0.34503292366064187,
 0.05068187677079483,0.8715947096447308,0.5815948124644297,0.9415683750219258,
 0.5412365924735332,0.7223412212200074,0.14544050248761198,0.7348423895418393,
 0.2147116963219602,0.5262066651156042,0.8004176021066648,0.30446547265863955,
 0.04948425810048729,0.34206223922825396,0.545597670996462,
 0.24541538538336138,0.5971830114688794,0.19441874276518667,
 0.3972658724616521,0.5326272474786075,0.8344784192003238,0.7361640588599333,
 0.3687814793464368,0.13125458977471904,0.5068601986647548,0.8391076466665085,
 0.8606451718377028,0.4866267879389323,0.6049207853444054,0.10202360961921475,
 0.1295577136346867,0.7757166189473713,0.7605803590919268,0.34553203390149567,
 0.984529873307446,0.24768901863510462,0.7576578667923262,0.18254143201361517,
 0.8594032345413035,0.389660344291874,0.01525953612165798,0.9420097268557595,
 0.8291289556294377,0.07065449014450298,0.6872920025074327,0.502802418396211,
 0.3297836734719527,0.7337773777510915,0.9284458088724729,0.878155129905761,
 0.03911594441324562,0.3144516228790232,0.8879943701088806,0.9429265795763497,
 0.10602479109727547,0.2890095811485698,0.8913263113359561,
 0.041544403124451734,0.8086141011583505,0.5467310055995978,
 0.08034488196616674,0.5338605815913257,0.013465824879038357,
 0.5023442999097473,0.956586122700477,0.6760572712539903,0.7466337197242278,
 0.555669178187111,0.23743409482601197,0.1309632174009272,0.8125088322027492,
 0.29613484486461894,0.6783496586417481,0.39275680243124345,
 0.2760402124954514,0.01983922596272647,0.6467493242815056,0.9569813694537782,
 0.842836145901522,0.7494818563714447,0.49324579133233126,0.4732753412303604,
 0.7328509619674163,0.04416122319175708,0.5575710006500834,0.3455205090072424,
 0.2584272634193754,0.2996328436803717,0.8434427392104992,0.18070060685644096,
 0.3213524876219157,0.4543715877597524,0.09962546002493577,0.5640615806178547,
 0.8877348108357006,0.8914345811311887,0.1702540052456416,0.12930551463488993,
 0.6194021631262039,0.15620242639305482,0.48810723661720357,
 0.46357860847155197,0.3133665106845772,0.46742617321853763,
 0.40975648591979674,0.8744474214583492,0.5468144573058368,0.9266396682083781,
 0.6377997831116304,0.4592652040235137,0.7220844839304401,0.9929976207410129,
 0.07736305392706899,0.985267773944482,0.1800216797402513,0.07558408400142258,
 0.040343826184476184,0.9555548210588463,0.25070436784288186,
 0.048371363183493,0.9744163891793567,0.19471384974050188,0.6025702663382928,
 0.30572463607557065,0.04717699685063681,0.11886631875873299,
 0.7949917170410797,0.0388643164058885,0.7447223258303299,0.6334598994409496,
 0.5618799868281301,0.47060912527587495,0.8462439071570663,0.3673293130180165,
 0.24909322348002494,0.019720720225051003,0.8296523798597281,
 0.5139546057727731,0.8609322502406849,0.8709033431159403,0.6390012644174786,
 0.6546194860949183,0.7752423523677836,0.031543956458746525,
 0.3558678057043676,0.5585059076377537,0.4957860852215352,0.35098031891027304,
 0.9170323571142348,0.7600959512713026,0.1846925243123737,0.7656119878446227,
 0.552066436814157,0.9897637634273887,0.6807428994760396,0.99176385035622,
 0.2670022459592444,0.5409450506167948,0.04818983599326443,0.2636592584890839,
 0.7004888872014707,0.5365026347694977,0.23074728493264351,0.0400113125980891,
 0.5133935875870022,0.23087933046669096,0.416018951922536,0.45523907229411675,
 0.6230904927788005,0.8244887622809164,0.44508523406785994,0.9758922270424334,
 0.8219670536546597,0.9751472348307155,0.11451388249975447,0.2768363290281973,
 0.6860717563488975,0.15017655008285824,0.2760247862634426,0.7628348039063272,
 0.7087968155737847,0.8453749561310961,0.9393621024744293,0.7026708196089881,
 0.49778967381502537,0.8437779741663871,0.18227048463809858,
 0.21083960965919257,0.39872395342880185,0.5938184726481354,
 0.12097814634310355,0.29987913002592514,0.45026191123533454,
 0.24438544500501047,0.9595536234558748,0.22390303867190509,
 0.15498922522895597,0.5562561400293276,0.6680117543501697,0.5462686429157387,
 0.05458435585205845,0.3344381747263554,0.49585669761224205,
 0.28872634000088415,0.5936000122263818,0.4967136714527953,0.8573540113477784,
 0.8044444786538435,0.011363339389489857,0.7898683642155974,0.947728968290668,
 0.3010388859948261,0.6224222628420467,0.5228738076427073,0.1140728147227813,
 0.16053780264753337,0.5801609503398296,0.4171524038029226,
 0.061614402770614474,0.4297953180005206,0.4819111762489944,
 0.29489970500256035,0.5025960189106585,0.06728228736191033,
 0.20652938254739928,0.7936090757745627,0.25106637642373897,
 0.9458760349630035,0.5971278852058471,0.2775221081497068,0.6394681586643752,
 0.8872499274685949,0.894550499973121,0.5696492804405835,0.3535762298442796,
 0.2876859477850361,0.591589562485855,0.9928839687903496,0.6869045267488547,
 0.15731658924968417,0.314201452036895,0.965301053803239,0.8707307295293015,
 0.7931209320475936,0.3961992888893322,0.2815516476823019,0.7893128002741943,
 0.04209914215527144,0.8562681923828309,0.5854727826763393,0.7871219123302221,
 0.6408726214650444,0.9281988463091855,0.6908689381766283,0.29139301966520437,
 0.6054475166195907,0.32360921404481013,0.20652704408843303,
 0.7616593404328271,0.713948262749876,0.041207512924192335,0.8079089091650076,
 0.6110676243287442,0.8574742221300826,0.8096156584544616,0.6784331771149661,
 0.03578819844402403,0.004464009137200686,0.8183833938281151,
 0.649704079277522,0.3366947445813765,0.8786191213776657,0.28562182937520686,
 0.40044517541571656,0.6771358329266723,0.9364579860478761,
 0.24210208088994412,0.35589552206822717,0.5809687199439155,
 0.9963121594438261,0.04097238349808563,0.5813002934657816,
 0.013513810122400272,0.06803309438398353,0.10990198523667827,
 0.8418576710372309,0.7969304155834351,0.05663197720316604,
 0.31357827531417526,0.21502557459070681,0.5740421673055389,
 0.3555246192454926,0.43604212332920944,0.4029815941303436,0.8827572286551941,
 0.5576474661479397,0.32811461485832716,0.12233316096781877,
 0.30213611208997027,0.02763507750355909,0.2209773164407708,
 0.10674349615701129,0.7139244782330245,0.20329775237140835,
 0.3304915738812241,0.47432302493601,0.4776810793353564,0.2711279065540123,
 0.9071836991420339,0.5536275983598631,0.32013372621884906,
 0.19262813124620137,0.4433017846561339,0.5707646079945374,
 0.011667146726532684,0.15766461183896385,0.4938257210972554,
 0.760660821952244,0.24243281470423428,0.07570856737779563,
 0.21188729025219444,0.8937886628624885,0.9535079743880006,0.798773247900299,
 0.4065501971165376,0.048645500138465136,0.958365224107782,0.7757113451028284,
 0.3879318877873881,0.7970865712789015,0.80835875579779,0.6374103250254778,
 0.3145711979004988,0.35485641837901616,0.9219166405724555,0.7298190658707185,
 0.8910011772601879,0.8311757520204905,0.2571013127859254,0.13693449767349775,
 0.3435795665281399,0.2809171529521556,0.6667161208002854,0.7974110779339328,
 0.4665439862001889,0.9846783556074027,0.9700831774050723,0.8201668676659667,
 0.23825534055658626,0.7995141592235006,0.7214226548426416,0.3685624286711632,
 0.6989192455284203,0.2788554871183859,0.984430383563796,0.7630926814563853,
 0.3313877291767299,0.39208434818456495,0.9279865280807232,0.9550660440863906,
 0.5257001268574105,0.07009771774608331,0.9613587038131038,0.5692694507298631,
 0.4639099056828312,0.08522964458402682,0.17249705973401996,
 0.9108336408733116,0.22287816269817307,0.46425838478482495,0.518906301628879,
 0.2996361954757182,0.14130875064801862,0.7076244596902524,0.6684925598875036,
 0.26779674415520305,0.7123393902155257,0.898680226238959,0.746695748167547,
 0.2421412403011326,0.5487268157739829,0.5416885824429826,0.7891680345585026,
 0.019710800456594568,0.40744374714721454,0.28752873754160946,
 0.40623590792261066,0.1755213585762201,0.45429939094985883,
 0.9029191031232231,0.17180493695652066,0.5659603263732306,0.8724346410204095,
 0.7577270653483122,0.46939558760326283,0.5542442961156732,0.6205546849371836,
 0.8968438853034324,0.4881496921685541,0.236253406851213,0.08767731375598697,
 0.22913036138300813,0.05475952267444684,0.24153241812692228,
 0.9334693109942628,0.6776294453267625,0.555773861106843,0.3740731775748888,
 0.30489321928713686,0.9172602960839662,0.10174978397637935,
 0.25968900856063215,0.8872706563554027,0.053568206055101375,
 0.7366158806123284,0.8342656164927557,0.8890342667355384,0.05004470575639297,
 0.34108207714541217,0.46149740591823674,0.0026877534052592855,
 0.05420308254475792,0.5249926970991826,0.3601047103070444,
 0.21030149762497286,0.7049436011607485,0.29583358209042976,
 0.7098172964751107,0.6444153302542439,0.16215142204879573,0.3871148791224618,
 0.6126403531575711,0.13838700444740792,0.46065440433840554,
 0.5288713495887837,0.7952336797424013,0.1065988712507745,0.04599028079220269,
 0.27949584088856894,0.6876510460409332,0.3184084663351716,0.4846681464978634,
 0.7877095284984871,0.6105950228755505,0.7955445059931111,0.22783766479419,
 0.23341238261865216,0.01940544819689094,0.4763226674643265,
 0.34988367177261304,0.7674742740416356,0.16989409163823144,
 0.6149605464937398,0.8384377027129531,0.8547863066298302,0.9243206791755404,
 0.6525679713898007,0.8746555802798702,0.4125698882964207,0.45773497040851674,
 0.7095997550617232,0.18056336077950164,0.09308777811628455,
 0.16565482053344893,0.3544395225461555,0.9588697341123803,0.7890238647761039,
 0.5668575324814387,0.3117627567918114,0.7476262266722156,0.2771811823748508,
 0.8083768539952545,0.47905311145459817,0.7298827954164634,0.8780767767273296,
 0.9553789794303814,0.5340417696564516,0.1816485456791379,0.9622997673943969,
 0.9992931863955504,0.23893182972152616,0.2342844706569014,
 0.49873576583889556,0.6490085051295463,0.19400968199438906,
 0.3679648650633338,0.30705109091550464,0.11170765812476124,
 0.0052045512122544935,0.5847046025472864,0.3037138651555761,
 0.42144999972888364,0.8733190322372693,0.7130208671282309,0.9462987400540598,
 0.15573179566534856,0.7452949551098376,0.4368956269393095,0.3389098151843294,
 0.4485233308020873,0.7511677835387663,0.41212195975854726,
 0.08559651263639123,0.8601075922133081,0.24223129499848217,
 0.14826444587646348,0.5161378761749719,0.5947197778046655,
 0.005651893854559997,0.23138385859205113,0.7270921911115624,
 0.8181931184053572,0.77420427147291,0.09646110504550387,0.34296177068299205,
 0.31223028915182205,0.07663130348634817,0.46575168685108914,
 0.6173357671986339,0.43574160042300136,0.2698743847938122,
 0.15258346111506493,0.7603446118487039,0.8488837641764319,0.9346354591826489,
 0.3824507081387519,0.7089455249286764,0.559629252270145,0.6588625042800029,
 0.38853752853229306,0.6831252902957325,0.181179393666418,0.2875806499787825,
 0.787134503934151,0.14741260010532697,0.9495349134434181,0.1314718865336767,
 0.5429174643753895,0.11547409055130875,0.06686591104258688,
 0.11801279958595579,0.5953797229158944,0.3512075955460603,0.8498917283368151,
 0.6778371159141456,0.4378518096954279,0.641908326364403,0.7629451435065054,
 0.9645113549859277,0.9555356069435263,0.3112822703920808,0.10025550157542418,
 0.4852516494734027,0.010167896611282146,0.7406321889511063,
 0.3840771581709813,0.3609611180276212,0.9395151582329139,0.40696511630475707,
 0.9972890387970736,0.23682623572194572,0.9149697754211504,0.6783184576575653,
 0.7935135446629439,0.30050928322802284,0.5491419685280476,
 0.24470785436609455,0.6432877175591296,0.9427528757378518,0.6078964993063841,
 0.4656678500455357,0.042194904954886536,0.17031463436386973,
 0.11961132921348483,0.26788082235890653,0.20599612798301425,
 0.8613850441867674,0.99010909840378,0.7024859426719067,0.19650363277491234,
 0.46989686722252877,0.6861397707174846,0.9645813599896589,0.1705294708360523,
 0.8541535952134974,0.8537290718221605,0.48720234152436026,0.2877238225598948,
 0.6591733699158401,0.9316177478528125,0.565234279311802,0.44033779745942103,
 0.4307750186418635,0.09982367987062535,0.12686027312265513,
 0.5836669052721536,0.9132038608810411,0.3559735144742866,0.549439097065994,
 0.9446704520292822,0.4577345226732987,0.8558198909852796,0.9885097733961606,
 0.22199194520783017,0.039081363408815406,0.9477478589719325,
 0.8388761024871692,0.382806622648589,0.45275058253176004,0.6035891536857063,
 0.5590420594919454,0.9344764523935971,0.21930236403434045,0.7059123331539899,
 0.3754631789582463,0.08525556349745722,0.2886872899081535,
 0.46122647497059044,0.7298332775235669,0.6137848927031444,0.7153218021035768,
 0.8292753858244505,0.299761014068302,0.07436065622843557,0.19130809365191725,
 0.007958944672712742,0.8087524369383114,0.35930020280646335,
 0.5220402907482941,0.048584383808431886,0.29370348458463047,
 0.9969884661425553,0.38106694168519384,0.3321868965160004,0.7762410699412914,
 0.16022081195998406,0.588956266703979,0.5513397977797757,0.09767781037832368,
 0.5533921998169544,0.36876882911710596,0.20186962169360623,
 0.5618820628554837,0.7555023090348187,0.747071484680663,0.4237772684549387,
 0.3391446611046407,0.7475077039026403,0.5076466504386244,0.918718592955789,
 0.6233933634247834,0.07495602865891404,0.336556250322096,0.7038065222688148,
 0.5488567795019255,0.5141900090044248,0.5590286346662295,0.24554127176050533,
 0.36235647281316263,0.6833486177828251,0.3494480963809826,0.5414874453398559,
 0.1088964628082677,0.9046699178072812,0.06908454775039008,0.9773275815935322,
 0.4234794353096414,0.7853108061398006,0.9942283203365997,0.1327042369764666,
 0.779871031171573,0.3558095402743151,0.2794815893944965,0.03598664347475511,
 0.7671652043527922,0.7855245726401497,0.6955613037479517,0.6291229981430557,
 0.8907558866965009,0.35820530396502137,0.939054757768201,0.4973376935289644,
 0.150537179237629,0.4502109970288315,0.5702330137145202,0.5541248890905869,
 0.615651562942976,0.6773677022239438,0.5951392256638997,0.4544191068224881,
 0.1346126622116901,0.8675671530792067,0.3747313002363165,0.5809346489730994,
 0.8282917001075037,0.8771347157900764,0.3191706092211224,0.8703054510498229,
 0.5597613563750159,0.24648051578833352,0.25276841714433607,
 0.9934622858997023,0.23331772357833347,0.20308930127220448,
 0.17471434584258216,0.43847746450619407,0.698748500816872,
 0.011659233223380472,0.9861432794506643,0.017026248100935892,
 0.38782189203002226,0.4524869904894745,0.11571242286951189,
 0.8278879815713774,0.5293368733845104,0.4382754956877484,0.7048553393810217,
 0.7279244765000566,0.7559221072630611,0.6648970475401669,0.02352368265711069,
 0.9549262061125335,0.7550666664257537,0.6023728290020929,0.22342320557502116,
 0.6805854614261636,0.20903240317185223,0.020651986697775016,
 0.450861216810284,0.8096289076802577,0.9174589514083552,0.9779116798882047,
 0.4421055056332446,0.8540375799428626,0.41504463080381915,0.5577639583862355,
 0.8061469220361176,0.23619818815220528,0.5860763000179509,0.51284751070154,
 0.5231497215123676,0.968972671971977,0.09976191177253102,0.711841265561926,
 0.9511259763337824,0.9693570635938141,0.8847530290170335,0.5944836188815852,
 0.45503674225262125,0.46320971292765756,0.7598936200563678,
 0.9410541231655807,0.6391476080316857,0.23260183776908516,0.9181757924637686,
 0.4618207254461354,0.532620679361537,0.23668773714727898,0.6563109280040335,
 0.6452070470462435,0.899784658110331,0.35085685980439296,0.07494225142848576,
 0.3410832741547136,0.5241356081867288,0.12738950453146158,0.4622252311730799,
 0.5291387454650844,0.3957563379054865,0.28069993340939947,0.7749799103574785,
 0.8044064046821456,0.5276556238365506,0.10964278489509216,0.2824005635016056,
 0.13283138737313893,0.6220771430904063,0.06873517264285933,
 0.9624903219696513,0.5875596810314304,0.6092454919894408,0.5477423192029983,
 0.427581461912373,0.6061348060110765,0.14174134723092358,0.8584355791818991,
 0.7400777716904784,0.1909862408051396,0.6664990263158453,0.45334615997260197,
 0.834236658106539,0.11603795806523687,0.9434867260405416,0.13308087874766084,
 0.5910922177807983,0.5931168366068256,0.5154307009006163,0.21479146293304885,
 0.16774889712368002,0.7310172518941596,0.5226376810017204,0.6948263803952491,
 0.02850394489705066,0.334933495392966,0.6434171628494811,0.00906256492945079,
 0.48319173091189715,0.3337193567755885,0.20579924825252593,
 0.22615512951070194,0.9408130709496801,0.7834917178405504,0.9633200623053625,
 0.2976405702061682,0.7880015422518175,0.29957673371766835,0.6333502140269781,
 0.06759594340233588,0.5041662858104744,0.17366780380214497,
 0.40383000451284434,0.3443195720628678,0.6797014991999568,0.3474138716550773,
 0.9731843502435472,0.5830598436140076,0.9353684455987423,0.8470204239262142,
 0.662938712869832,0.9312662073633559,0.13469809951453582,0.9107864256745983,
 0.06685556918183022,0.7906846831144134,0.10103257382448261,
 0.2659673591535503,0.3723219185558164,0.9544652978845959,
 0.013117787910967516,0.43769617315267506,0.34429235810946046,
 0.9627519149164054,0.6997508888181102,0.9991145626078738,0.9248506607002355,
 0.38050215071205384,0.3071835317960313,0.993864363915141,0.4976371373165369,
 0.2694427244232179,0.2940171320104571,0.7604159636189527,0.9094856676740586,
 0.05116202807791159,0.042236680645382885,0.5950746057751759,
 0.23354510381952864,0.5414178012755375,0.041913733291177535,
 0.1661715107951902,0.3897895541108787,0.9421472811891052,0.7352905523016251,
 0.3888341463048043,0.1876959604283499,0.16990462963376518,0.4410473531026263,
 0.3464475471113344,0.48366949266643133,0.8048392406690439,
 0.023607304850234323,0.8286351942551018,0.4572021777485026,
 0.059212364765331205,0.08226174523646412,0.3924336996398502,
 0.17463691273704762,0.19187285364959616,0.3155109977722289,
 0.9055813524462312,0.6726380212168445,0.48873473884300056,0.7682810477745007,
 0.5868605475937753,0.14006803835432102,0.10593190531623231,
 0.2758728051637427,0.9802083017157978,0.12760893831349207,0.6112750820391385,
 0.02602649969122972,0.6840727014550341,0.4030157241093947,0.8891801518919917,
 0.27663678787597656,0.5887896814765667,0.14960082124890095,
 0.3874665846803088,0.7251344734616418,0.9300119835023096,0.3547066565476784,
 0.007504522142546932,0.47165701609289545,0.4306674991390129,
 0.1491468395268839,0.32820596748134556,0.858013724973418,0.6319981901570937,
 0.7301052852865098,0.7160633389336216,0.09355679683827345,0.5796158371840037,
 0.1583478105015862,0.36394595678499486,0.40057898166797334,
 0.4461157638572484,0.9926811871610024,0.15400825683908048,0.7428354257663128,
 0.3041790730030167,0.4262105063798899,0.7928050915944436,0.501708679273532,
 0.07600769177599553,0.2861484254906459,0.11342673640461376,
 0.48612457905821715,0.532943472726224,0.12963248751605572,0.2419766790918848,
 0.22559248198720416,0.7716897738463091,0.8741625990596795,0.6864284570062007,
 0.18312017490365706,0.11647371111952498,0.0225652221417888,
 0.6253284475758775,0.09778226189543204,0.2878356917690139,0.6314942351449907,
 0.22103298682957384,0.4642063602233314,0.8440585407696509,0.7697481828471426,
 0.3565337483075215,0.9856703848438665,0.9695497802801814,0.7593431203679557,
 0.23092872510256135,0.031094686462320276,0.7236371393174027,
 0.23126931566748654,0.8354078973856444,0.5736365995727478,0.8362459138686695,
 0.9659212461472853,0.3265664298131905,0.5808579818786305,0.5797626530792442,
 0.7894612727218412,0.7918445057669838,0.7429301240476444,0.3606810423250476,
 0.3631882569019329,0.7935357320448473,0.9969740478420733,0.6977603144170451,
 0.9672520398595451,0.25354448627362436,0.9626682242887107,
 0.021965515236675448,0.12104370124953645,0.14559374492401034,
 0.08241016766517462,0.5158737153767141,0.00347463466871345,
 0.3146803855464496,0.2166530075060219,0.1446804132885,0.36636672625259337,
 0.6029200272011768,0.27320918825595764,0.7420743964861692,0.9993230024189176,
 0.0215452056642792,0.6049977444890775,0.5981913730654008,0.748555324124408,
 0.6750701129059988,0.2722954912734248,0.7010812020127243,0.34837956749883103,
 0.7981703539588461,0.8891057323302237,0.6092013589215134,0.2522681188728876,
 0.07350750061033795,0.43712812488110586,0.3567715214438183,
 0.040355836945799806,0.5622150668387319,0.8919857716690127,
 0.11824475009700908,0.35512679151585425,0.55174138518214,0.06458089848164061,
 0.6090187315849938,0.7059483267805503,0.6172852656903494,
 0.023633856650998264,0.6987340226930995,0.7364580813919931,
 0.9870655917268364,0.8589355604346999,0.7727204339551064,0.08815672931346263,
 0.5665106556454257,0.9979961825892777,0.41634257122555984,
 0.38850913397633247,0.7053513334319393,0.1668386082847202,0.5391650910144837,
 0.16877307679512954,0.9444450232631277,0.15012022129082625,0.480213890199332,
 0.6724172986282452,0.970610831565358,0.9254373908206748,0.5382582523128603,
 0.9437517469776269,0.7009791727062131,0.9299107162839856,0.7332454320493541,
 0.8827633251581688,0.21395908770950767,0.8752951777157838,0.8587998317481238,
 0.15888328056842227,0.5995476119000811,0.3135596979208082,0.6004461299691515,
 0.09330913888616399,0.2429725482506656,0.9213550633688883,0.2462234901929261,
 0.5975523862351051,0.1580128152166379,0.5814095706554732,0.37271768160160634,
 0.9299331692398122,0.5068651798983536,0.6461496876082531,0.30433687293751366,
 0.03681167520909012,0.09115275095328967,0.821200870038894,0.6685543778086029,
 0.3558458207451016,0.2311238616495852,0.7364155917654454,0.31793878010116916,
 0.5826508433899498,0.7607513242765698,0.145685090083586,0.19488025564055933,
 0.6834246966398324,0.5821335814517633,0.4105135323473973,
 0.005592015599171329,0.5159473920155466,0.19636940525823654,
 0.4647293416891356,0.3328393307447046,0.4159809371733053,0.3246007591732375,
 0.933280871889908,0.4855800999652612,0.8303685903218059,0.7992942355538277,
 0.7049200665917636,0.6755710227652658,0.6247011044178524,0.2739904546838985,
 0.16466438196214328,0.36081753412468687,0.1177950633106637,0.978800026124709,
 0.6566748233138375,0.2911308837636323,0.5388716424588653,0.66633351468431,
 0.4792155869881033,0.24528465748355233,0.27078438720377407,
 0.21531459572950118,0.6167295756656315,0.3973245806231095,0.8432111805803981,
 0.413978241175073,0.6192055190858103,0.05004851063005389,0.37062792274726253,
 0.21808358847769416,0.2991122006228213,0.21555743412232586,
 0.3778322437747208,0.24319084078578102,0.9835241321804924,0.2366878629440572,
 0.5214635677916921,0.994307097166391,0.6135031719961319,0.3550844156050229,
 0.9282512666995937,0.15811915666785048,0.8792616182943063,0.5796253305942508,
 0.30105366463732497,0.6476761849766048,0.9765565199345209,
 0.14393641859803608,0.8802656831374036,0.07734720774204218,0.732915806860931,
 0.4955452200059143,0.9133003835885142,0.48337355622323597,0.5613744265080837,
 0.016159234466782357,0.2916472237546178,0.08839927350998056,
 0.6657440163694419,0.7171912305334952,0.6162299239586579,0.9191466646716907,
 0.5212052362285461,0.56388564417788,0.5157308147945993,0.8862464711727062,
 0.39079866848251066,0.6216174818698289,0.8698433724847873,
 0.11446518310788134,0.7991750426194292,0.25474383178235716,
 0.9072141082755268,0.18376735320352877,0.799234587433112,0.0636256112465513,
 0.4030222837583737,0.8576517080752543,0.8471461982801712,0.826378189137796,
 0.6088630090532332,0.4461430976949672,0.33023102038736796,
 0.029607360712992303,0.37231300239795695,0.2829304251381497,
 0.8818260132141205,0.4771215021929027,0.6284670968173339,0.3738494500927674,
 0.3438246179579092,0.45066461155512005,0.8055890905178663,0.1078475294361183,
 0.20683164618375782,0.13749483893592573,0.9548057113476737,
 0.7370062029173875,0.19002874793925095,0.914597817873489,0.08205906460542411,
 0.48986022495709924,0.06399759994001819,0.5252699845549621,
 0.11997198754567462,0.43628566784226375,0.33104043567020836,
 0.38258545705890057,0.033180858528574775,0.3616114548454401,
 0.7829298528280421,0.3351789846978027,0.6874286116628997,0.6886113628386755,
 0.46198557484642966,0.38857264399018,0.2620085776494774,0.13296416662920751,
 0.022826465359515713,0.7808852875911827,0.32356703046229907,
 0.3413261568602516,0.42004816110075627,0.6175942016526321,0.9237029508398416,
 0.6555629902295287,0.13730698048273304,0.5698303952779282,0.4700638528058625,
 0.7237935669824795,0.28895624289797894,0.5061347869663003,0.5725878106082221,
 0.2215106914507584,0.7126636743242131,0.3673192500178424,0.9150584440240315,
 0.6626460506702532,0.25703773297510013,0.7861135203496781,0.9139565230441488,
 0.4492129019335286,0.9183344935907323,0.5184169387810653,0.16887041959205884,
 0.5081754575258071,0.35172412136985765,0.23043847823256502,
 0.43497051708671153,0.553685283176801,0.9146969135940862,0.9238707589775618,
 0.7825426298632974,0.35721287068774,0.3852519793672623,0.7280559393684253,
 0.47745555406775975,0.19114472543083305,0.9974552269748148,
 0.3572011809188237,0.6404707626221999,0.27125217265789536,0.8570203606058047,
 0.23093569513496792,0.3157551186828298,0.851374936171726,0.6611360628294503,
 0.7280912136494321,0.18938583190738079,0.8469716700435408,0.8382161329978874,
 0.6312677866250468,0.3408024590877401,0.5771211187309297,0.6046592464582345,
 0.19885731259801553,0.017298983032153825,0.36093974345832003,
 0.22437462081186266,0.9968055123878476,0.2732610165242383,0.4999075188095664,
 0.9715392111805232,0.8090516974112361,0.3991255148680971,0.7261826668206508,
 0.8081567884766137,0.11199940594848745,0.2079443539374165,
 0.25536810280810873,0.880603572575602,0.056350455017057266,
 0.14574374387323108,0.5632754226848306,0.7138919114862643,0.5619785989179689,
 0.3559202028008941,0.11021273195686099,0.7617265322178536,0.428931289971185,
 0.6619153981500147,0.5720339101796973,0.1571197005812376,0.34733678276356117,
 0.040171005880609845,0.360850739750458,0.13785604527001383,
 0.6523798582826837,0.46643553226550105,0.951312261751546,0.9263002924693158,
 0.47045869961794073,0.6583797122775327,0.8279618536346821,0.2801760153935453,
 0.36366462732333504,0.5516426479473014,0.47277575667912675,
 0.9723673323711967,0.91057644323553,0.39548436255426256,0.7456325698364699,
 0.7664240793433326,0.647253123275102,0.5112952466411099,0.004247251354677406,
 0.8995905941092746,0.8450022119836614,0.4001045812839633,0.7655520218617016,
 0.514529562030711,0.7888242548121943,0.49260132021186886,0.7356118208007426,
 0.7799627089975197,0.9540279840180226,0.3221350367224486,0.2130176255563564,
 0.5115637453357251,0.28832653809057307,0.6237196477858762,0.6538170362278797,
 0.8785174532118608,0.6219939054262874,0.1464288650232879,0.8981901204843471,
 0.6146421353589087,0.3895378821487797,0.6374972386207051,0.9363177849385008,
 0.713050018298543,0.1443782840446668,0.8525956171357137,0.14959037884087645,
 0.26287862828669983,0.7371919486164956,0.30427651184903703,
 0.5937735465447144,0.6529508347334874,0.9815285919370527,0.5496640361176267,
 0.055073490498395716,0.7651439838252683,0.7167744570186335,
 0.8025983467721449,0.8388715537775124,0.04584075684051758,0.4456164647447214,
 0.7472408652817234,0.8089928964558665,0.8749080026614983,0.7664065522032384,
 0.19349158212901751,0.6229594562630497,0.1748382167879452,0.3052989361373085,
 0.9435887977077424,0.2165188130534733,0.020456281688539968,
 0.28882877297264975,0.9454364470944451,0.8865729135300796,0.984142086708105,
 0.518241192187754,0.7146747863336772,0.5307148453817987,0.3336340859144853,
 0.11645561056799414,0.36575848773878894,0.624903308591888,0.4604567483256503,
 0.6352263874617148,0.4563791071516411,0.44461924354279336,0.7622823879391301,
 0.49138313325466065,0.6648239031110904,0.6101470514067348,0.7996986345068414,
 0.1473544770791555,0.6890227705138141,0.059511694905819645,
 0.42264553731911925,0.5423873690931141,0.8682458825597723,0.1280984940439911,
 0.9697206471597575,0.7582715879282418,0.9430329918895374,0.9995064379870171,
 0.26085128618766973,0.4475906355288237,0.979573897930802,0.10962226160172606,
 0.21903396875190206,0.6542070130091495,0.8730495583101209,0.5516263601475607,
 0.4299958994278009,0.03542578195892809,0.6713307070446489,0.4827703506071994,
 0.4332925067408958,0.13912168591544216,0.7701918197827055,
 0.39446067598518997,0.4149423043964392,0.1393777549228986,
 0.009597487726021092,0.1782918554127958,0.8998737904861178,
 0.6492956878189717,0.7439755260627159,0.11957587064722308,0.5638562276198726,
 0.4096224436679996,0.8262973214098895,0.9740093431340606,0.9798638189985693,
 0.34305047138160905,0.8288006041456185,0.992138185582296,0.04795758420847174,
 0.15476259524235492,0.3335991529544341,0.9120693482477209,0.8127148955259105,
 0.9041140515294617,0.03570947065508656,0.044653111715990956,
 0.5432837363022456,0.5645884119803433,0.5686159076619637,0.5180040813300641,
 0.38057308583002425,0.24361315287714913,0.14576292048307649,
 0.27015606519597646,0.1952718721655381,0.629273804869337,0.7600251804136385,
 0.4878962554251559,0.4613669486659445,0.8215710214110665,0.7826866506267467,
 0.7478592757826614,0.5879265424444057,0.3347950134938956,0.8767531937411303,
 0.6701931649196292,0.5536665386512438,0.18663028107395452,
 0.12749466680175103,0.2537499754407938,0.18209899850631872,0.801069456339516,
 0.5653382684635213,0.4425110871075081,0.9709027546186859,0.3230436413066672,
 0.8880649665574076,0.3933596169638215,0.8273029863439023,0.7594697078037785,
 0.3479151335438675,0.6795990285589166,0.1765956707824463,0.22062521043545202,
 0.6154320262841897,0.9982324143171925,0.9574465550381966,0.9639039026375716,
 0.10821120216711222,0.1317308760834801,0.19343401591647935,
 0.8634919425371401,0.25202630781812363,0.9625786654196701,0.30675115336533,
 0.655240315819993,0.875285024523405,0.3798090525068275,0.45578769466188906,
 0.5581296113922853,0.5466834651214132,0.5894348477025683,0.19255431683658342,
 0.7834153851667093,0.5301138134982388,0.20992367776583465,0.8211963424306136,
 0.8839228983279208,0.1901479593460832,0.5381092045925969,0.02333434413393809,
 0.9232498964556874,0.5022009534885431,0.38049669813197584,0.3885070142534701,
 0.11456170154264866,0.8996012604697068,0.8579665956252525,0.7706602187099486,
 0.031017066651421454,0.2800683244255737,0.3206050160531172,
 0.08534433791715157,0.3086993576649085,0.6987626826146566,
 0.09880263242645515,0.2350978362326076,0.8722952141680226,
 0.14748111720196377,0.06765731329025315,0.23881972757760972,
 0.4684222119326762,0.28028876700093874,0.9684666393975382,0.6512955444474187,
 0.2722624411917077,0.9101585974785791,0.8643152367632184,0.2828987421645146,
 0.889422225399295,0.4927491373739703,0.13280407385856274,0.13795976469274795,
 0.5468166715069234,0.8616759382011814,0.8181532320117092,0.7624071764774516,
 0.11941576617240524,0.9208800434665938,0.7551428545299861,0.8375999413426636,
 0.7556669332971687,0.4576965087432099,0.4629755468608031,0.601244924172711,
 0.534088694334913,0.9780845797279961,0.3891395247366728,0.3016993704412393,
 0.7182800614152952,0.68408659715188,0.8893678694856227,0.7607663245100038,
 0.08680655897538625,0.24781687385955975,0.5399655500746197,
 0.2853397650815249,0.8314731823261313,0.3136183586457567,0.08097726604968714,
 0.5195640475139583,0.23477881651299215,0.6503563570066107,0.91019162593368,
 0.4065614189189395,0.27691654922783815,0.6205199851316866,0.5649115991065029,
 0.20710383440208058,0.21790918306986273,0.8818941173139976,
 0.4244004068807109,0.9419950308674083,0.6145213929399691,0.4982463090420832,
 0.8848259439191575,0.7694062479718669,0.4556134096009249,0.13197175014099527,
 0.7166213248069745,0.629116523695067,0.8317496360534661,0.9516922001802259,
 0.37857732067176475,0.6598737022127348,0.6164074958437681,
 0.19065986397063672,0.3349610486903871,0.5994055564182702,0.5648690830715746,
 0.12766388846769294,0.8594204006163606,0.7966403373529981,
 0.46744026124138527,0.3212269208759264,0.6154079354532512,0.1770759884062263,
 0.8029791374161677,0.9505194181115786,0.6190930871158017,0.8260133016944141,
 0.5230748675213293,0.7018086927543246,0.5127070128648339,0.963354410455062,
 0.6829533682435198,0.22241304818315877,0.6507294410293205,
 0.16121626722422977,0.08045560604449797,0.25710995302360296,
 0.8295564097909075,0.1830038006852246,0.5285243493408969,0.4494050801476973,
 0.8600801525379738,0.8594275968485777,0.4226900516712532,0.8918701202250068,
 0.7126660077219161,0.4727234459827877,0.5739921318590875,0.5673948236737086,
 0.0862598025708845,0.6653693846182283,0.501880033924827,0.40863241856247723,
 0.8038992238988429,0.802593361209933,0.8017350237199367,0.3629358211757364,
 0.6236668342816475,0.29663721192546555,0.3660322894233482,0.4451160631570702,
 0.082405343173512,0.2235785557496922,0.5132174851778809,0.027560713963331374,
 0.44129505894779275,0.9165543729554606,0.22388361384109778,
 0.18214147017734572,0.4621724520477255,0.24163424663161392,
 0.5132828114901815,0.9333633566846952,0.531246139242517,0.03209444958680685,
 0.11263346788422468,0.7586318585676055,0.09412661449604665,
 0.7807729415652161,0.6625486930924479,0.9962032147064419,0.7677936324427151,
 0.10794275156858257,0.2033868066675817,0.6603250917386465,0.9989226113807128,
 0.6443619939945162,0.045911387973376905,0.21770756653293488,
 0.3355719552374684,0.6656666160473019,0.7038595994323992,0.1288172207812599,
 0.39854344536679454,0.2985869218290136,0.0761837723550961,0.9772856496654924,
 0.5123534180908698,0.2166282753713511,0.4066073144043607,0.5235885582653195,
 0.5984130935350687,0.9222620195793912,0.011315620792568759,
 0.3333289052678984,0.621163058058723,0.8027771039693135,0.650332938255088,
 0.8813072505964076,0.8191406473963445,0.3031682358373553,0.0466302963878894,
 0.21153994990513514,0.354418176756776,0.28424902959354115,
 0.42739621626651003,0.5749078708833019,0.18023191107775094,
 0.6298220591029056,0.2614659291452449,0.5053254303696142,0.5947037490311938,
 0.14049801553305818,0.15143755227235,0.9983121151899322,0.006803160800080144,
 0.4723430862877034,0.6821983711152336,0.7396348521405511,0.1485724117892191,
 0.6393723873805054,0.9276009733945554,0.6024693530622912,0.44412729061238365,
 0.9293642558510571,0.8691030671132309,0.7196593038093579,0.8961025866574623,
 0.9246124217072117,0.0255905081839789,0.6495487694737521,0.4118845111909972,
 0.9164743035583243,0.5197044967697741,0.7114787002504206,0.5896793551323397,
 0.4103291639946801,0.16721280836757546,0.10099026048213522,
 0.11412406529886421,0.2258482609220076,0.7856453448548045,0.454162334045651,
 0.6396661973732893,0.36335675891380137,0.06450960750650958,
 0.6827242676180325,0.6064176108509428,0.846973018260628,0.6787868257776655,
 0.688280305557065,0.9048943919116744,0.3678999649949083,0.1944069693040572,
 0.9138259745669143,0.7471608520401465,0.8368657973713146,0.31598620366285457,
 0.19431412670514725,0.3613719305493217,0.7091011949262165,
 0.015870706663982714,0.8458160655649547,0.6756509325712263,
 0.5849644900022525,0.15068938116523334,0.45038347909036436,
 0.9023510149161588,0.3647932929234857,0.7463193403260043,0.13927582262704896,
 0.36663536674786323,0.8249586848122578,0.8687749854859752,0.5575397735552587,
 0.8018350419826308,0.07125301572289666,0.39847728116583936,
 0.36707564751572175,0.1525046585297647,0.6916716112451112,
 0.13334219500217948,0.12563317944452246,0.20598648420704924,
 0.26846909509896033,0.7036001563724166,0.15907048049010009,
 0.12052826689313689,0.7217390556957868,0.6747556378783748,0.9274593797481474,
 0.498089300609846,0.633256504410213,0.5357534739793108,0.5668657668303959,
 0.11306029405146201,0.36575662398231223,0.24666354324660278,
 0.9363591244867815,0.019002557333890446,0.4168979034545155,
 0.7129711698549124,0.06764244557844834,0.9582123669006418,0.3841545485520389,
 0.41660948513692775,0.49111192448941854,0.7968529251939482,
 0.7038585246695669,0.03770394930889853,0.11882783497843497,
 0.5340522021171394,0.8964444758175203,0.14232003870374044,0.9637938228687077,
 0.16241074060945404,0.6865711962127069,0.9537931048376378,
 0.06794721575299933,0.48686711799221427,0.03888655301553667,
 0.7006274709576642,0.5995916731612085,0.6258194033746386,0.367979558129494,
 0.8950254436237968,0.026498519510257057,0.5963206354608424,
 0.6608523978416491,0.2906122075995625,0.49700543314872236,
 0.042229516101527254,0.13280124604180898,0.7373566876158417,
 0.8196598610837701,0.4223307333227605,0.3453094134289989,0.592380556424942,
 0.8909611477678462,0.6363397452334774,0.7271105927685949,0.22449165580548158,
 0.9971727104213962,0.5245421472517325,0.2360757529302766,0.06786218868415284,
 0.009736704624804693,0.01671800310368099,0.02782687160324271,
 0.14231113562069908,0.07830558351791317,0.6630038397246577,
 0.9511546790250585,0.8514735498412104,0.6378620551296802,0.44315826603298236,
 0.3706508280465416,0.23029271131827878,0.33865371389783117,
 0.20068447017013769,0.9409910893482627,0.5790843246059406,
 0.18981581186001995,0.6424892674080768,0.7173985932926774,
 0.14265319584769998,0.20243090180612244,0.08105695922565814,
 0.43897764499040515,0.6809783314654753,0.9050296339014415,0.8433579929346239,
 0.9030292983487278,0.6132614653093188,0.39323966981161484,0.9957212481009654,
 0.045943755300696054,0.7906334544318356,0.4925759925194977,
 0.7501878094123084,0.9962024159417381,0.16397896131126943,0.6456961328349644,
 0.3059946304340885,0.29295825877772896,0.10852225596310894,
 0.18261040871459122,0.25410363761979105,0.9971766921515218,0.29254053851451,
 0.1007511718053048,0.00824494467331549,0.18292420979653645,
 0.9265853581840808,0.04616343941854539,0.6450150637667982,0.7410291931931943,
 0.22720033378785298,0.9049338077486512,0.17345076267285053,
 0.45969605845263783,0.8706807450980053,0.8827804085738058,0.5600445285142831,
 0.37933489003725873,0.386912022109286,0.8502835426646035,0.13448357564777702,
 0.48626701811782314,0.8886104959573556,0.9937178649995452,0.8983992914129072,
 0.9910501369558977,0.01281809875494333,0.8868966613171457,
 0.0038577883938845092,0.6999175451205134,0.0063924475002158365,
 0.347625415599924,0.6063492421118352,0.39885524136001593,0.1679099850153316,
 0.8235001809262482,0.6474370937728212,0.7525550157226917,0.414659096287774,
 0.18178782715150588,0.8526620738054941,0.5085680806603657,
 0.06521922263288626,0.38591826914006844,0.7601120734420783,
 0.6711961377137812,0.31402221460724333,0.7503804388472477,0.2947671246788197,
 0.8370980095174744,0.260152587610078,0.7719471635673469,0.20145484587648566,
 0.5351335692972379,0.025688294247099508,0.6267248385427622,
 0.6012687996780338,0.24740469728906778,0.23960944232961157,
 0.4753257678046938,0.0811157148963052,0.12304079094782172,0.3662729456795045,
 0.2845173898164447,0.18189908487656026,0.3947030430823223,0.4813880995571016,
 0.1422788771353768,0.5866690001963077,0.028570960359087683,
 0.6356308566344806,0.9795554695578793,0.11760623219079502,
 0.43362101114277773,0.836404323910868,0.5636064003304782,0.7870672025194282,
 0.6427888191543051,0.31783233353329543,0.06975595550808988,
 0.4748999587097017,0.5423466808987282,0.41688355113897657,0.728029168795254,
 0.13873778101680512,0.7269245864374714,0.11030442688812703,
 0.08302143874157875,0.5583186887595964,0.16476500894338175,
 0.5256952482337973,0.23618338535286687,0.9090947767830396,
 0.31122969468042794,0.6587770224358416,0.010862367505061643,
 0.13259095994011305,0.4632172281746534,0.17560601776584228,
 0.7724628739121246,0.7504336943196378,0.24998939466557923,
 0.14552251148058926,0.3120342404918319,0.6926913105098498,0.4661933171871937,
 0.12336670547143169,0.03555242793382707,0.991829697209444,0.671837601877164,
 0.10367397502232623,0.02198575203677222,0.28402782126868265,
 0.3366892133983379,0.3321121546628396,0.906183903012047,0.8762772227817937,
 0.9923368829052063,0.0725297940926839,0.2364031631844905,0.43009211402713654,
 0.7404483286958726,0.3054221498724803,0.8204043402333641,0.21130221983541553,
 0.32933614861735705,0.5638897762168307,0.3175508479850222,
 0.24647070453231967,0.5762565123108998,0.7346314020785735,0.9992350707850699,
 0.6316202405422319,0.4524117975879566,0.3014624051270226,0.23913064718179544,
 0.12575113120630532,0.5866994806463031,0.28754611274407804,
 0.8900382912726293,0.7193259480698249,0.33796768006199573,0.355440631470446,
 0.15325224158675876,0.6128892948446841,0.045718886742199594,
 0.1737655366059565,0.17223896144198003,0.14792513072183178,
 0.2864276589397494,0.5774024467945282,0.6488085117810561,0.6004073857103919,
 0.5278770488119765,0.532393829170013,0.3282945258848907,0.8964989241117481,
 0.7544300351475404,0.3878227265587918,0.8091977372461548,0.9310600935287089,
 0.39975435301289797,0.6945854518420762,0.6845440695624204,0.2597448656644318,
 0.5274449240232659,0.7739642762821286,0.9441033104982659,0.7477826951274282,
 0.6289894083439086,0.5695128813378345,0.6163850012919939,0.5134563552785434,
 0.6765757711360976,0.8072700341488802,0.35310688754673014,0.6707232808578177,
 0.7412346864118949,0.9149490527665041,0.28551922910482885,0.6620104190818846,
 0.9453644353028243,0.8416269119956482,0.6231762473992162,0.2354118852622079,
 0.6302647682692,0.5622991907944905,0.8207930604051951,0.3880344463189994,
 0.001673476060328638,0.3042635105900642,0.6949148303669783,
 0.5877857254002241,0.5360832680837948,0.42396403309319364,0.732841058739188,
 0.1777919630851461,0.6019347241038558,0.6457849424582108,0.9937371499157973,
 0.3282239223905732,0.7600581444210943,0.9191295505524094,0.8804142932898483,
 0.3499101445130892,0.8936351592644742,0.2009146723448525,0.5390194183545545,
 0.9838881614987751,0.8138462590665637,0.7891979059346267,0.24390809369269406,
 0.6811325104567044,0.7202784065079384,0.26161627186271064,0.9082883220193448,
 0.38203852370713554,0.9518946022717756,0.6732888217829097,0.9242529768879072,
 0.006794837566497458,0.3203635070296168,0.29094890177239174,
 0.07236255339417774,0.3615518271104753,0.6300635190915763,0.3667983697618521,
 0.5629369740783796,0.8033781563346664,0.5611872474329002,0.7257688700103816,
 0.6925590389192875,0.6649170097405575,0.6522945696602455,0.434743771948654,
 0.9843467849095376,0.3405643640896061,0.006964589054500481,
 0.13635776019325296,0.8470216151221741,0.9481352797119603,0.4325029431106242,
 0.5582167651997754,0.0657701257259311,0.7822114675082267,
 0.0017800044982156926,0.41316790484877475,0.8561599355375636,
 0.0010631169385513228,0.9604859312439317,0.15729642501150376,
 0.12237428368008973,0.40124708346748494,0.24899441380955956,
 0.3974779811701544,0.5430774272945836,0.22204382978670822,0.9293847577271414,
 0.6449971508644688,0.03339665572309458,0.41307455627327316,
 0.48609845825580134,8.019910819790921e-4,0.5190261746231153,
 0.5618509740282523,0.9334036843856842,0.2832000553571191,0.5296677741279662,
 0.7451623360954374,0.3092989125638794,0.7467570574352789,0.4756838923857315,
 0.8763411485586358,0.6862101301754127,0.2140451691136741,0.5450527379094844,
 0.7004864672103701,0.7669168957176528,0.07769974063037033,
 0.14953020494549119,0.4358657151785785,0.506116625914539,0.1484831676424705,
 0.558390659951464,0.7073721185623264,0.6576581951154811,0.9252593518761039,
 0.3783080628049904,0.19643840351437936,0.0676170996833585,0.6229626338906646,
 0.47760392070135604,0.6721411369360684,0.3911909863551726,0.5197786671546251,
 0.14586639586139716,0.032956312239957763,0.08175702811436747,
 0.97889671756861,0.537857629426452,0.262679262319797,0.6220447205304358,
 0.6981471489659427,0.8894119414928927,0.30245212284140044,0.9076609243118747,
 0.7949599107433206,0.7339484002235319,0.548117281822321,0.9691267344479501,
 0.9660705934951671,0.9779478712374358,0.789912787002938,0.6121430036795836,
 0.1591551446749584,0.04970533410893907,0.19074704611337223,
 0.2221732338748854,0.16160555018344513,0.5786681105573253,0.572239486831176,
 0.07090471643022644,0.4446245410036549,0.8696011361680008,0.4672492525055376,
 0.7217488320270624,0.6942956750616122,0.3302301278485551,0.7255967717124239,
 0.5182331622870054,0.4749918081247285,0.7337261148298876,0.4934296569335892,
 0.07889157203991104,0.7763657861419239,0.21165389839431148,
 0.3904293684842566,0.784253540672173,0.12357740444773346,0.6668998122672384,
 0.06260931749238441,0.7773000594536131,0.7251207904596052,0.4212948201409472,
 0.7209311795159645,0.5026977922906888,0.43802074872572927,0.4673602026038405,
 0.4271617385963412,0.5531855595801515,0.9539668730654599,0.9328394543815862,
 0.9890009799593147,0.802612466459889,0.7636968737335539,0.5513826827075494,
 0.8534963870297969,0.2754034612621703,0.7610164115860789,0.1642244781239275,
 0.795762291553126,0.8011623138227294,0.3886446376667966,0.23862223566208385,
 0.6503486022988835,0.2405351272493904,0.6536227237978708,0.48567839217871833,
 0.18691080344338717,0.41843743570554426,0.8406854112522869,
 0.34492338436582926,0.7477212997356101,0.02080652456718346,
 0.6400823452735716,0.19955857857206805,0.48158984172345454,
 0.2965435542057022,0.22307513811356805,0.17008678032742996,
 0.7412211310753111,0.527469389228492,0.2712064252008174,0.7685299127115028,
 0.33366866508898974,0.7952373550816121,0.5168378318119006,0.9433779205070792,
 0.4308715489131677,0.8730430870181426,0.04178037863339068,0.2384156562964984,
 0.15657171715444806,0.22528681481310797,0.4126975690878275,
 0.23832762728916834,0.6205732487007561,0.4884262372355318,0.9224469526430179,
 0.28194988013932354,0.7494274030366609,0.746071566838114,0.8698465577008161,
 0.04721197860230064,0.22951131372784817,0.9170890894276231,
 0.7426076539395217,0.26440380441810185,0.10860368624573435,
 0.7643657350985269,0.8476226888900955,0.07354603719400388,
 0.20149560793493337,0.0842881834462419,0.049546597198540376,
 0.08895054204688435,0.9946287746223579,0.21365263131811063,
 0.38802332659477645,0.7607744789899606,0.22665739774441607,
 0.5066724371888357,0.7009783526818204,0.18913405936739736,0.7381784465076663,
 0.027011513289833977,0.5306112081736586,0.49932492074888724,
 0.7326986936247324,0.8824562375075145,0.9886010171598879,0.5474694080238245,
 0.7579527663556563,0.19978571814111068,0.38111879334729626,
 0.8750169823245649,0.41598934323343506,0.09840736004477968,
 0.2717841694331238,0.5609185727917861,0.38553558666587995,
 0.46132958134218216,0.25165753897648124,0.36283400972108604,
 0.9368003489973864,0.9613602856726511,0.1956734138426729,0.10374705948132124,
 0.75709297090853,0.6571388089032064,0.6910273390287467,0.2014745517458605,
 0.9977834309314051,0.1990006115772861,0.3552470925750334,0.2384561435188075,
 0.006556407535431519,0.8776314873827863,0.8974502187688876,
 0.18912127386398192,0.6724039275823355,0.5903215271709249,
 0.42564354925143844,0.42320793268306156,0.22055399071391002,
 0.9135831166450707,0.15816656393368445,0.7154469641424578,0.9934996587858093,
 0.7565059529059657,0.22239088279704167,0.8906487615396116,0.8800297821254153,
 0.0556594323524916,0.8006961915678212,0.20144413166003672,0.9521627081112417,
 0.7918509756514912,0.10957863429722348,0.7022740993571407,
 0.24736213750372915,0.6305395530289134,0.20923292145363914,
 0.10109632677640978,0.40227088808667844,0.03460639484896877,
 0.9617353678541518,0.9203744483973274,0.6726412693416294,0.719925386135569,
 0.6570257106894931,0.6929426652304035,0.6853892267457232,0.8127972794059531,
 0.8323324378706777,0.61767906984215,0.06199423661803416,0.38000920399446436,
 0.3037608311653913,0.4084311577295019,0.7028063862913179,0.0374194735256379,
 0.8133774682455754,0.2181473806337475,0.4010819416655269,0.7620217230243825,
 0.11241319589067389,0.48479081270925717,0.641235981586247,0.9915094462088598,
 0.3258498600067331,0.2617506697506292,0.6178488953524758,0.9041685512430524,
 0.9460635463345706,0.3437977675330466,0.5093840366896323,0.766301497731404,
 0.8974523388301452,0.3991774130073621,0.10053865009902874,0.5460115568046783,
 0.5007551657166686,0.1929511728896055,0.29912044072986055,0.2933649533737932,
 0.34565542611076217,0.9036245979145787,0.4636889522257678,0.7223964531069643,
 0.4856531979012769,0.19678267585885867,0.19616846234274166,
 0.21304194446598768,0.7716137422872297,0.01816916226578158,
 0.7198340724240515,0.45783744553811856,0.43482861541626916,
 0.7687760070992748,0.5501595165934865,0.3230821343291259,0.6999740495273418,
 0.7842740698762951,0.28020316444054416,0.5913804406451391,
 0.42000886665770976,0.9456809979062676,0.9690687884298761,0.7755358113867165,
 0.4667645446361123,0.6705890174835414,0.6750748589255213,0.8515121875226046,
 0.4413690052695022,0.7980361656996992,0.13139637211894029,0.6895604851730711,
 0.9065581299731603,0.2637732702861858,0.9235615470666637,0.6362895327801336,
 0.7333782891005662,0.18782928254950226,0.6223695217922668,
 0.05866590191687027,0.5936733948330815,0.6573853245966288,0.7858296290760601,
 0.4356252007828876,0.0018353956989640707,0.9791034134886637,
 0.22427908549155984,0.7017965745563888,0.7151932491686799,0.8782345024297302,
 0.27015175006155456,0.14921010093812515,0.785051646854191,0.8680091206008822,
 0.38038046346522647,0.9745979428584469,0.8558468253024977,
 0.44320629954168245,0.5959644110806321,0.318912513602692,0.15159614395856702,
 0.17709246162272874,0.07543520644833324,0.5404927119210275,
 0.46347856949196664,0.9066848075633742,0.28494708216569675,
 0.05732768344427952,0.10883238919133975,0.6692639306014031,
 0.08360034420726536,0.9709639297255338,0.2897280064092077,0.3442877884008042,
 0.9570571641060212,0.6627374326915443,0.6115349117324927,0.24150438103002514,
 0.7722785549257366,0.1775554037135485,0.1821215765890143,0.3737155749229899,
 0.7085185828704219,0.5401447320758189,0.5220561623974629,0.9571197450061975,
 0.46759420655566464,0.3432239450744611,0.2447240475543414,
 0.19306555690516858,0.18228148986327486,0.8253935609881309,
 0.17674936712948408,0.6059212415587787,0.23720739741488195,0.819357511368773,
 0.6207026224208675,0.42058367406012276,0.535107751220741,0.03528435490593462,
 0.7662519183720798,1.5814280447423723e-4,0.3997474639857377,
 0.4792236156779982,0.9669974076528606,0.6831911731069173,0.03176084601817508,
 0.17826885824163674,0.5063315534747855,0.8897385540973481,
 0.09698408251446722,0.9390248867967284,0.9456454191778068,0.3636566598486787,
 0.794216708614557,0.35129531194163,0.08727534404432669,0.45384596327459015,
 0.13379968918094887,0.00833084401159745,0.85282750993843,0.7335516476166952,
 0.9723846003005143,0.7052562296539444,0.32181028864691563,0.1549468413435391,
 0.24231467141242524,0.8704521299382038,0.23380818832920658,
 0.4136832166529322,0.4416570544469629,0.02606230667047682,0.2444290657612993,
 0.06306854856800759,0.24932979449635362,0.9243942483439072,
 0.9366477021449305,0.7537601161664517,0.20652482080980483,0.3698989772590444,
 0.4882519335449236,0.9006280020722021,0.9474416725857148,0.3443596303148848,
 0.14360518527567112,0.5902655418467168,0.0885700429024231,0.3637428107291737,
 0.6245326710365495,0.22346130855987467,0.6302794754931158,0.890193134015367,
 0.9769361104724745,0.7379421384677808,0.48971556758309454,0.699276620691434,
 0.711273851072167,0.45923403246152483,0.5614631136615599,0.9783774864758884,
 0.3746392334205184,0.6768985580764608,0.8113053352909892,0.0584411237012169,
 0.4121743850852778,0.26506494884675713,0.43444120959717114,
 0.6755435764506103,0.5579381181085512,0.5997769322415009,0.5763776341864146,
 0.9199770167529296,0.11955959040562203,0.24105683311549464,
 0.0391084061767033,0.16379207286984232,0.5493681992003351,0.9371220422754789,
 0.8711966918765328,0.20493936974813964,0.16046735572199533,0.746905471294384,
 0.877228602648289,0.44373197314470525,0.03863469894660565,0.9781890318622294,
 0.732765361774748,0.5283294628850379,0.28950118561482907,0.9044211271123145,
 0.6992442458691269,0.5922588091358079,0.2555908002925322,0.27037221043265824,
 0.804467488858874,0.4329250882807316,0.5877718701460051,0.8180229561926928,
 0.684554936062418,0.6777711172163589,0.37206834764004415,0.3125043542276139,
 0.8901929052286031,0.6977142959287311,0.6161310287717792,0.3660224647115239,
 0.28877373214854496,0.5019903432079501,0.29492648914850017,
 0.5567474853291484,0.4977814801518481,0.7960725526402136,0.50257697979944,
 0.08998094056372719,0.8183226861859382,0.38273884906061983,
 0.7296243409764537,0.05903890290225522,0.04921998260307614,0.600986984153384,
 0.5262661740693954,0.11152486733426004,0.427565915471414,0.3607185704857221,
 0.3311153540946593,0.008187843316860954,0.23020981713827815,
 0.1185773802288127,0.8120265831818435,0.9910558299723798,0.6819357605679486,
 0.2857709401757692,0.8423037114094533,0.4563738346098385,0.3415268869281043,
 0.5692152786802573,0.8742496918409806,0.8580556213093691,0.9982950021116367,
 0.15552481046379352,0.3434910497504683,0.6155045958016335,
 0.29324466543801475,0.812825219136397,0.18086099174465897,0.4235677004976808,
 0.6957194104121048,0.33043710927336634,0.18065150713381206,
 0.9002302565841107,0.7225633477342439,0.02026428995579932,0.9068413524894697,
 0.5848815459605162,0.9046179565583051,0.2757005709239613,0.9611001721851552,
 0.3638032519480603,0.9367848288728966,0.174943790841257,0.8252532230545493,
 0.12216601526820137,0.5406541525447955,0.7739320674590052,
 0.46701231666149146,0.006750236950670785,0.8923829961193371,
 0.6315399248356699,0.44693247563737115,0.07323636367644226,
 0.7181759611916663,0.3862676212865477,0.42211251168554464,0.5398291181686963,
 0.7771290169028084,0.6551918218304603,0.6953417686607931,0.9387105802382187,
 0.028232372464420008,0.7727991423490503,0.8355284522972939,
 0.2135674017387751,0.9216481669002933,0.8523080996866075,0.5071320024299967,
 0.32776278243430346,0.8262808880585517,0.8695764576353718,
 0.14004461440737814,0.3800332643820754,0.8193908642304392,0.7937976760189676,
 0.7371639713819933,0.4725298636220455,0.25678743247622093,0.3836953882180404,
 0.3061174730184246,0.8161784552524012,0.1688925841074682,0.08206590243812217,
 0.47892248626323286,0.9027016307847094,0.6007361959016948,0.801856437122153,
 0.5306751215516232,0.055140983141313216,0.4109121428408209,
 0.9416312078517848,0.9689139462808986,0.959604284722503,0.03186185537793973,
 0.15884760097808748,0.7414240567483206,0.09340638955138314,
 0.49063060148665405,0.5916021256097664,0.4149928668193974,0.6495507969941752,
 0.42046615462040815,0.8289508983192768,0.379685028655925,0.20308898139774345,
 0.6766031357495169,0.6564141121102679,0.8037877338555364,0.40124328761933126,
 0.19688459840433403,0.6124867709094122,0.05886624322335321,
 0.8686944682502483,0.14253889963921196,0.24745524700386534,
 0.03939168723987618,0.09439636838391141,0.5768341371772329,
 0.6581291242326547,0.5594021827833883,0.14447536496688007,
 0.003993073002116487,0.8304255773628494,0.4726961041769894,
 0.3364901369914324,0.2197214547816302,0.2892366360560704,0.5805294814129999,
 0.02251479646703225,0.7756243124124015,0.10244296916378792,
 0.9917691710800848,0.6030130612471505,0.2642679551519862,0.34753688607777644,
 0.8301399802935262,0.8796755706700773,0.6534189378800638,0.43500863747659413,
 0.04416735165207397,0.5269179983087514,0.007802270285587887,
 0.6575014639562694,0.8820565610344145,0.6224479764401403,0.8545194845706554,
 0.2837034701879817,0.5683222522649818,0.8116349639674332,0.6707704942118226,
 0.8688602321940028,0.9061986499484227,0.2920692281701538,0.6965789082360042,
 0.49803258290031427,0.3513223113802455,0.4269811313088211,
 0.18498457559713222,0.6701638343304117,0.42865487118366996,
 0.5568072226505524,0.9054567022536202,0.30418284027510245,0.7440952319331438,
 0.4575200277539062,0.5161053373072816,0.6271183669872248,0.9674467513231617,
 0.08325515506491721,0.8968820063353738,0.9386210402632069,
 0.17644007940264506,0.614209589936367,0.6567126188677362,0.8254616975453337,
 0.9779558000285312,0.16354900185065913,0.9703285126845761,0.8836395860909374,
 0.5968909668940688,0.27544009008349524,0.24605519887074667,
 0.4418530215958997,0.8239044647997016,0.856345208482301,0.6769191432619343,
 0.4429748884851402,0.6257065425973849,0.6925340256689244,0.11282293256256004,
 0.11195688079838939,0.48010472505903135,0.39138171688619816,
 0.36590593149251993,0.5106030822507878,0.7248771356016517,
 0.14993327244634636,0.4401144070169215,0.9632818663353727,0.5099405128621214,
 0.2534691218510793,0.24801576665363112,0.8555882692293761,0.6386553365640687,
 0.7188921346852288,0.0011581393467714474,0.11977513905173165,
 0.6953657658622613,0.42503043947954355,0.22182221233546418,
 0.1444407684475586,0.9649195218229276,0.8523389544621773,0.2972043162240824,
 0.7113576918529512,0.07063705420923694,0.9951616472222408,0.9831515225389822,
 0.6529308250190509,0.6530429227324872,0.06610121095567478,0.7893791772014312,
 0.19275183734957257,0.2668793365691047,0.34328927799028564,0.861672323997551,
 0.14591299733106589,0.30631611768487954,0.71680867155879,0.4790229418488344,
 0.04428020569061952,0.8371118460096247,0.4571838606823918,0.357066125226267,
 0.6621905436163972,0.443358186988148,0.9280463018166466,0.25946999000973503,
 0.3910528683222121,0.7975545511223524,0.4927464281464049,0.9920171721091197,
 0.4933469842071815,0.07267828311492308,0.8946189474989806,0.5927032570178585,
 0.23316408486394957,0.5874851409556272,0.43622099840985495,
 0.1692205488732459,0.016995400759187262,0.5441670456278641,
 0.1453011529469893,0.8262215450429418,0.3565824202639494,0.28624605445997453,
 0.8670846085767013,0.7586423746755766,0.646909258701625,0.7316473489135725,
 0.9672416204629777,0.2891875636896232,0.32547844824264005,0.6872488148518562,
 0.3384504746733956,0.9506384953731383,0.5107266215247412,0.20930195860562018,
 0.47189934711622916,0.8308706095118433,0.7550382956857776,0.7933000846605862,
 0.7639485472213992,0.24245406312979423,0.03550542127860945,
 0.8414456118531108,0.29303606649797964,0.49949922284824644,
 0.19435385117174553,0.2634818786380295,0.6633197237485379,0.6811574479918222,
 0.949876396384401,0.36105042894457684,0.3534360586181384,0.7619948982614697,
 0.404273916051081,0.1943463819120007,0.7144557603565213,0.9768556047216643,
 0.2631366225892021,0.8712768706676084,0.5982803720801276,0.7499673490338616,
 0.3349623997570982,0.27769749812355293,0.6129990338472228,
 0.33666172365606173,0.4762663127462886,0.6461131790319511,0.5564153740842678,
 0.8989638476428854,0.6752819297460723,0.9952955800803229,0.9830199241294287,
 0.25942872064104705,0.5062838282565784,0.033406273211986526,
 0.4828251796858678,0.7189745267027461,0.8744398650403373,0.09725836477417815,
 0.1463789130029134,0.4669255868845299,0.5276987398635714,0.5566077185716825,
 0.06839862165496058,0.6595075319096709,0.023390602547897044,
 0.7187247822668577,0.7926820932413099,0.7992974479319964,0.8160751485856421,
 0.6246383255634019,0.5558161754962364,0.728424665527532,0.12318344668278491,
 0.7834373780245764,0.6380798816932782,0.7349857281761888,0.27990770048612135,
 0.2067306741171282,0.43852911926657123,0.16488004343054496,
 0.19906582077890178,0.021313866618817068,0.7185986545256604,
 0.1808499580413675,0.9168862647085095,0.8736260652497034,0.6241870459157042,
 0.03553156336645369,0.5897170287656397,0.9894822679361487,0.9881828119404321,
 0.5309285594216936,0.08274264815296695,0.8884357042941847,0.3039666768007002,
 0.013730852930624371,0.12754259186845307,0.5323109831080818,
 0.46733751435027215,0.6801993365759671,0.5886645625935547,0.5259426622803915,
 0.9777189384602418,0.20840086995095375,0.6109341511992037,0.298529284806068,
 0.05561804645512669,0.5888890433040702,0.770528496068297,
 0.020950120389330862,0.7776302611199829,0.47022815020938247,
 0.05181018231067003,0.7736723525429011,0.13678709577909032,
 0.34609661569597483,0.1929803438657608,0.08096141232636223,
 0.5440275928753153,0.7929247368410618,0.1699442581447118,0.627915748644058,
 0.9628252104927814,0.5559327109391609,0.8626768019676238,0.35532387874508686,
 0.23297303615602538,0.909037367828993,0.8720838249508256,0.5157753357526282,
 0.036092695599970526,0.3171955354254665,0.7978092047420453,
 0.9971496254701185,0.7349979157300341,0.40213869996390983,
 0.10194550650777634,0.549176742231298,0.01393457884432403,
 0.14697193197184732,0.04640675504357594,0.22845549847244428,
 0.1940152103494832,0.34174603131263503,0.24927648925702728,
 0.4329089007990925,0.002358701008235875,0.5544458845750081,
 0.5497830246170524,0.7754819185825862,0.1463425470514016,0.10851778675826518,
 0.04413801209648338,0.19948965895202153,0.3516704668062688,
 0.0911330678633191,0.5505477265641427,0.9685517635463885,0.6487391297727703,
 0.29902125580328043,0.5537713775815651,0.5272145937435877,0.3881726463137265,
 0.6803329915773647,0.39472887489955455,0.2238239634522421,0.2664743923083397,
 0.3634134722929817,0.39953009293128594,0.29181373783322984,
 0.8574068127216242,0.5405577424114727,0.6985664305740713,0.1096011188968794,
 0.29857760496823527,0.47368246023456406,0.20780852759509927,
 0.4378631508660238,0.971274865396962,0.605520462624569,0.36279337419906077,
 0.3989524615215786,0.8823927840851981,0.6563121769947416,0.15288186282701544,
 0.08286249186357375,0.028814613871709027,0.35628929587651503,
 0.7229984818371996,0.9485631346613576,0.004437807036827612,
 0.8973582186452704,0.2768622618454256,0.7955428827919473,0.5671773696255901,
 0.2276028456019591,0.8038278731735609,0.14276704839296506,
 0.49227219440997527,0.06072807327789165,0.11282334022186857,
 0.4090760920399693,0.19947465344088233,0.8277542792118815,0.9791722661047586,
 0.4751921256553162,0.39338157522107886,0.17499389058477122,
 0.9774508786653744,0.9722118960738458,0.5197997744489751,0.5765778266142435,
 0.6227737990553814,0.7178119914206753,0.22523821046078507,
 0.37915712205817853,0.31204856315458895,0.7454394990757781,
 0.1385601264368912,0.4511703765447649,0.6026085500694444,0.05359843439005563,
 0.38904084668136363,0.20487515118050903,0.10119100783225798,
 0.21524902830057246,0.3635030613018204,0.6045261336240924,0.5358865506190355,
 0.2715333958571985,0.06415111933460538,0.12333060209337066,
 0.5757169883290713,0.622403329978629,0.265078816827634,0.3056754017102532,
 0.9552527630916391,0.2337136268610862,0.18492225354835135,0.8451481340859044,
 0.5868012838506336,0.9543891792475505,0.6033415879818775,0.6315973972133229,
 0.6980458646541153,0.3419369346926473,0.6266674928405364,0.9320412791621713,
 0.6069970453661538,0.8340633157989529,0.5820659991928855,0.7539607508398187,
 0.7502138511845997,0.46491794196307945,0.6518582001739976,0.9247595981208289,
 0.7777870309914032,0.8024945950001604,0.4060969209252955,0.35962075834288343,
 0.6617739509612445,0.6491711008461751,0.38951604023896913,
 0.10946092773889904,0.5661512356657277,0.035897150895260355,
 0.8373942184737967,0.43925788180742675,0.4274027773781701,0.9424944570100477,
 0.6777119367705495,0.9871956763175029,0.6907438806110642,0.9826312985576029,
 0.5920826163102882,0.7612146553580621,0.9369991091079006,0.186958187679802,
 0.9200344101310403,0.9340879421399348,0.11159109444200033,
 0.03806617156819725,0.745930048975949,0.511048081494162,0.9481059195535398,
 0.16996015319215296,0.0988818006796448,0.7354823324006644,0.3056908194451652,
 0.630517795314857,0.1894592488760224,0.5728910780513772,0.8561968234651294,
 0.09735154266546608,0.9063857477741023,0.4153058509485361,
 0.43878486703138975,0.26082974525275704,0.16713990267641055,
 0.00216465419189249,0.6385430279641845,0.7070706133214149,0.7343533099385303,
 0.107050303024815,0.5336694907967331,0.2701261556104171,0.31248909682725146,
 0.11948967383820674,0.12200426030754385,0.8720657979834783,
 0.5973476239269317,0.3878511696610224,0.5495327255660902,0.47466486247857964,
 0.6346978722789971,0.7938036808755847,0.8914123506786878,0.4590202109371495,
 0.7086282752609103,0.6769926657677493,0.6770709832557215,0.2713846338984547,
 0.5190546232452755,0.3523805954044422,0.12831318535164138,
 0.43243166025143565,0.10800097299257683,0.3804028571943895,
 0.3003543701968351,0.6711645999028284,0.7308641981592883,0.46087508352542683,
 0.09908187216729791,0.994390695199889,0.6365578253570539,0.5954426368686685,
 0.4216279359343952,0.5033225460526272,0.019737252183496645,
 0.04228954964470644,0.7422758440411471,0.3745307326930668,
 0.16681054873175027,0.5533363235955913,0.6713854234408944,0.4582708886053749,
 0.44090615508381414,0.4905997844371188,0.5803246765894362,0.6737835504078667,
 0.12884290897131423,0.6804420570588343,0.25232430125311667,
 0.007845608004698468,0.5864961436019921,0.5315822427736707,
 0.8496421603757316,0.2321021540868241,0.4758714167568071,0.08221914764467075,
 0.3706805049878761,0.5517507633480437,0.6770456319015198,0.6290982437380996,
 0.9254414328577322,0.3368621562691074,0.9718520873495585,0.5081857311232445,
 0.7180738016114729,0.42254209003411836,0.16072422187916136,
 0.7025681856749258,0.11018551274884203,0.18533295547019546,
 0.9637023471776757,0.4128923738133944,0.23293843541539305,0.640294225088978,
 0.1883399401701802,0.051415259219359855,0.6505332330099021,
 0.16134858663218676,0.2678901038622339,0.6659489575123219,0.8139313654346126,
 0.8010559490299067,0.6637788366508428,0.05310785400254925,0.3235625344711349,
 0.9044793950532105,0.2605295895786741,0.3508017432425994,0.5365876397678406,
 0.5424736143482012,0.13884103781345225,0.3522074234866197,0.7664396588194553,
 0.914959785861341,0.22087079762266404,0.4630564546051644,0.6068975708235584,
 0.9990674080796147,0.6868071166990097,0.4981032461274424,0.5318760160465325,
 0.6586973797367695,0.7290761492094134,0.4688754069687191,0.3191961831605792,
 0.4753712716784778,0.6765566384412492,0.31175389886957416,0.9139134486031062,
 0.816982813235691,0.6853363125626868,0.2610195635024699,0.5276036606609195,
 0.8130687554731919,0.9085014533993482,0.6885122136761678,0.36876152059381373,
 0.29593155935106985,0.3509096656762004,0.3088313095638504,0.628657198465009,
 0.3370301718515849,0.4190438873165221,0.985350240457683,0.09190561466187053,
 0.7085700183432201,0.7750067550136557,0.9045102405712842,0.12626590207437638,
 0.8059969259017264,0.8569878604484985,0.5170109914295468,0.2394400576180522,
 0.6627906445536244,0.3988676425376121,0.21569622101765518,0.390202977575555,
 0.18109942715311345,0.13729256693999492,0.00821198858053196,
 0.6961006934151036,0.6615059402338966,0.7484935305854492,0.7949389201522021,
 0.3233212376745511,0.33051607714236464,0.4228063090772666,0.7879520434664775,
 0.1280363391105197,0.5652222001992866,0.926577058457329,0.7083364550074884,
 0.4801247752999984,0.2924105864745876,0.7703737894165626,0.3280867751183991,
 0.3048873508831259,0.9092329136044728,0.6971963933889448,0.5148445237022209,
 0.8441095356511165,0.12943753625784649,0.08033188842004657,
 0.3935457666504879,0.013081704428374086,0.6250735260424186,
 0.7876851637295278,0.7988691070571778,0.39803396208140485,0.7486397603509023,
 0.17227161342302788,0.8527534996547153,0.46165054332206024,
 0.3575902452645864,0.36558863964224386,0.8719846079664373,0.6767068868238293,
 0.8297857731913986,0.7536294208217555,0.87246245732399,0.8978802833499342,
 0.5246790652050329,0.09980946439479066,0.4791703448051029,0.8710628865048182,
 0.31033970349278583,0.764314749614822,0.3945214391778661,0.8886006534106157,
 0.6968475354314407,0.5887160838917063,0.2316143633499067,0.9973630895833112,
 0.24453430852511404,0.704474921542083,0.8743835870427035,0.8827885701646667,
 0.058366293154266335,0.6851154931554111,0.4606754057869783,
 0.8627784236734457,0.46928029693172424,0.8392614473805208,
 0.19677953643956736,0.12449619207383511,0.27824422835305995,
 0.7095554690482806,0.057554430207992335,0.37237999578518366,
 0.5901485542510158,0.5580263187157549,0.1880883094969732,0.6788737513968381,
 0.7606748854317031,0.12813834391793222,0.7464419945695197,
 0.06521073626383844,0.4684901445966617,0.1558999539861986,0.5880965643483819,
 0.9277739203510936,0.6334757544686288,0.031023596936177955,0.613998433489183,
 0.6626186985323836,0.6025175734921782,0.21077590676819702,0.8889507931080262,
 0.8280673864458505,0.8042292633600583,0.6811973968286341,0.23189455259817426,
 0.699811325434324,0.3580752707691558,0.4565099534194824,0.06739947702118143,
 0.5054659315886978,0.9925562094009245,0.5583040491834257,0.9958270521116614,
 0.236838471525741,0.7514575650118451,0.749203678634344,0.9700212249720406,
 0.8547358211363305,0.0099466475578317,0.2944571156448337,0.35901815395198355,
 0.9905548008455247,0.6988135778478627,0.45962487342610625,
 0.49871134756824576,0.41310596031969715,0.5207998788508353,
 0.005406376033748828,0.9444505901489655,0.12423965080676924,
 0.7806739849984007,0.5249260687951871,0.46754448833828355,
 0.45817406608034605,0.9573769610332648,0.3652486416623014,0.6710755310327063,
 0.13599845999241222,0.11602967069550785,0.16208325374911448,
 0.8294496446664007,0.747882247069463,0.5840157639897969,0.6885950733498876,
 0.05568652051292222,0.8466554958501756,0.824703606565065,0.6008483928904006,
 0.602175354342348,0.12520750197772224,0.21533542400135353,
 0.30367386090876947,0.19867327115046762,0.8803615969759663,0.994289014650753,
 0.04086305168942228,0.563916845122013,0.5520183537285368,0.5789352189143082,
 0.9782209582866983,0.9077779293969909,0.10301122987458322,0.8917289254190941,
 0.49663288716863807,0.5137432776217078,0.6096569024746,0.678740165964675,
 0.17643816986801608,0.6960885677848158,0.6143208489184268,0.8278030196654241,
 0.5464836094091529,0.9279321670412897,0.32226180054036724,
 0.26474661304626546,0.5342997543766593,0.9495354543493003,0.7581966521650016,
 0.14093346582534927,0.6581817323569957,0.07883903192603892,
 0.9392758026853326,0.7619596826649379,0.24763582155383368,
 0.17830108893173469,0.2638568482104944,0.32687085551615025,
 0.6756013348235801,0.21602724194029654,0.6983874261151861,
 0.14385760723687568,0.06985936587101982,0.0689578093753348,
 0.7066100558210818,0.2358259446120139,0.8314353044496499,0.8128521437016815,
 0.059596518134583,0.28010636457393323,0.5802072043762683,0.5403266089729504,
 0.6763053855527575,0.8484132267890291,0.9242887096000938,0.2817829499603728,
 0.1362296167367556,0.769652127380039,0.11919105618238202,0.8676972540932336,
 0.7442429077031243,0.6103255310649535,0.8123021175121523,0.8617777682860627,
 0.9531173371616481,0.3307707989400188,0.3596958842727773,0.8306084514054172,
 0.4263407110169768,0.8697486469019525,0.3746502510809663,0.5701666872809104,
 0.06424257772852893,0.8372242979600283,0.6502145754354383,
 0.05680319157998637,0.2005009163707504,0.771323316041993,0.19653987570459797,
 0.3037587203432537,0.28863860908932215,0.20573476583865724,
 0.9565786469112993,0.9471639010972435,0.9951081375101339,0.09151438553026192,
 0.6831632105940867,0.3540244042798031,0.2192809019578178,0.5313411332857191,
 0.5954427191694285,0.7691459142408978,0.5584277647478031,0.4498911254684136,
 0.7063988375428922,0.6325001542621703,0.17523476518561987,0.0914432151912068,
 0.4044068916983389,0.18830868074851492,0.3116950413049905,0.698645697478048,
 0.3873121692217435,0.36114068014850664,0.5504064628362909,
 0.18707685278443442,0.37551880349394173,0.7521550597600987,
 0.21413020724391918,0.3086578340917082,0.9054267616775883,0.3819595216583507,
 0.9444329010818768,0.7996166264117882,0.043785793857259114,
 0.9474379134650437,0.9419432309255631,0.23368227108731132,
 0.06893223512096225,0.7378150756158135,0.559706051501255,0.21559670096447592,
 0.27974882990981187,0.8072297442844273,0.23553338292702797,
 0.45884900111537774,0.2473231841498219,0.6005863224249588,
 0.00969765039568371,0.9592158842824897,0.9141705475877744,0.8666044826717039,
 0.7738240062587212,0.2215812460949359,0.7024644772031294,0.27643482220945115,
 0.5274505528682051,0.8500593617280264,0.15324860402718954,0.1664385627823396,
 0.8932455985864901,0.19209662868515598,0.11503726814308035,0.722479789054093,
 0.9720246003189055,0.42617077794650027,0.9218446060899546,0.7984235595260223,
 0.5251351770573639,0.15920783986324383,0.5547353976374813,
 0.035786761758353824,0.5570459541650123,0.6563801064800914,
 0.8675359896341411,0.963395501982375,0.8696244259836314,0.3317321349467802,
 0.13691597232636532,0.859647583118032,0.6920651752360956,0.6353293994202831,
 0.09788417289196527,0.26423695373432454,0.5742088515866067,
 0.0689016787220772,0.3587359769795412,0.3145533218608547,0.7116877431152527,
 0.28854297233796344,0.1733188606442463,0.4842284371394924,0.9364445512719657,
 0.15024384302578309,0.7204776120225121,0.2144511982032823,0.8157262756514024,
 0.8978757397026624,0.35873656352525096,0.40847834846028874,
 0.6515150478094113,0.31780341059112405,0.8208030663073321,0.3935165712501032,
 0.4171734642859516,0.8024302497992143,0.275752676266111,0.19237023145669985,
 2.9159030234615635e-4,0.49080391187421657,0.5070340315921981,
 0.4532042024971221,0.7135124154803844,0.7074748905329125,0.830294795203783,
 0.4821050981051319,0.8416831870725765,0.340822013512999,0.024338196423439262,
 0.4334355795013055,0.09531723082342836,0.12326496366552764,
 0.7025868510244667,0.39890303072451927,0.6949760714850728,0.4136395232283948,
 0.7046539165300028,0.05132619572318986,0.39185950480455023,
 0.21353740764925433,0.499153985624719,0.5322383755103138,0.6873311538264303,
 0.575657943639949,0.354782395656818,0.003631255061336214,
 0.028341525937145673,0.2969297449536147,0.028563505249450616,
 0.9674710394085527,0.640307603571804,0.6777015059061168,0.9856162894999777,
 0.43513986093002543,0.7963759821388182,0.21649909047012172,
 0.6281212978409324,0.3098657687855684,0.007402018978499569,
 0.32832890565950335,0.5635545995214354,0.6865535968007332,0.127465025466579,
 0.3540406017396326,0.8138950709254952,0.562372389076194,6.477645578710778e-4,
 0.4743984766335376,0.5034877016421485,0.34712440370966413,0.2401249836071806,
 0.7780045159796174,0.7755490258997884,0.7254176294994612,0.4561212729566968,
 0.6275238947980075,0.0362448279510561,0.35562458657684726,0.3292194918880461,
 0.6347043271187837,0.35980466981302006,0.8693799878598221,
 0.26148661937641504,0.6861761994553741,0.324076200630404,0.603110166614304,
 0.8401005059496778,0.6024001695592962,0.5612481711362843,0.9095886390904085,
 0.62448913308679,0.6519703107629033,0.8904752368674561,0.8287941296190462,
 0.6130086626248691,0.8922664795628258,0.24831882988895515,
 0.05923211738546941,0.5387491649101978,0.8599728340144879,0.5388823747990805,
 0.6435204747792083,0.3571000954494744,0.1575343221891894,0.1397461711031367,
 0.9489362465489727,0.5375051420262309,0.9712033485311422,0.42077385760164787,
 0.8645594483508144,0.7069487037189899,0.9005459498677255,0.3990453660286073,
 0.32998260546753444,0.9700246453972594,0.38567343790217334,
 0.12251365880757714,0.8426173274232716,0.0917161961100198,
 0.26122803892549773,0.840648762091332,0.6989167197900779,0.5815242027046438,
 0.2457441455708994,0.9509560748530418,0.5122131626045886,0.9540964511267326,
 0.41837013306825577,0.0037548811747086797,0.5315759877760177,
 0.8364424262567093,0.9940127552712836,0.3975061142302897,0.9163658186366539,
 0.7611444942677597,0.6437819390177321,0.574746468124277,0.9379914193793782,
 0.38987644078258177,0.030833093566906333,0.9659024971465068,
 0.28762286466332354,0.37291739440979166,0.6063987442432033,0.498285072846433,
 0.8039792152920671,0.7902814410341726,0.964739012883277,0.45596103492604956,
 0.6340360830131213,0.8769847009861707,0.7207392036645477,0.49805530433912837,
 0.5383044885734029,0.858699630375871,0.4784471875674844,0.64305879729431,
 0.6957120312349577,0.12518969249928424,0.4530327656488434,
 0.012110583105438555,0.5690555584461852,0.319358710684981,0.6761812026005622,
 0.08195199242968532,0.4842612254512577,0.3927010669248756,
 0.24425775805898076,0.5426654106231165,0.5525922888428998,0.9644930433341343,
 0.16336456522527398,0.05879135946195324,0.8456942988760673,
 0.24704040181951026,0.30990311966489115,0.21728708079846337,
 0.19241502111419784,0.6718508098990014,0.44448396419723124,
 0.5978779255361147,0.8128678561652631,0.5289823278222832,0.9326665012329647,
 0.17945996632498662,0.3814977707413114,0.9295921606589075,0.3327056452424868,
 0.23035986490911797,0.7655789893498638,0.9359515084608786,0.6404442622826485,
 0.9673373108432326,0.06174780361655907,0.6817788826373377,0.807853324103772,
 0.542900751924191,0.1356181172645512,0.7804415284832265,0.3505727446815241,
 0.9419561371978696,0.11681029029703205,0.8814574358552185,0.8112044979744407,
 0.6102726235566984,0.2510806424057632,0.052819511577624656,
 0.13262238884951227,0.7296119080091945,0.5161455018376847,
 0.008590029809980582,0.4099173165227339,0.8087400007174665,
 0.2464766933860415,0.4770912046079905,0.9165247872493897,0.872952718094819,
 0.48834274009179957,0.8895302701751637,0.7001969577866755,0.7643747384330767,
 0.4803787163481813,0.13183213026960006,0.02507234045378115,
 0.5253310814172625,0.29163100686222243,0.41810716238750634,
 0.37237193837272065,0.18600190643880554,0.10210141409150164,
 0.9249970807125825,0.14295985383213705,0.5539142681413709,
 0.22073935266296107,0.801837750412513,0.8745742620084596,0.10612231243709436,
 0.14461931798214778,0.3490833025513904,0.1704450526211031,0.5796044287915798,
 0.6284148838441342,0.5672122499534937,0.3885387377387758,0.2825044540270326,
 0.6651304203205783,0.34787285865987316,0.09451093181728976,
 0.9254522207945174,0.26863432688155053,0.44081610321203957,
 0.7879608839347774,0.9932848457187513,0.3639547845079898,0.9907525237819275,
 0.8518836526408653,0.7429137046511896,0.5098815012286967,0.5297085844164404,
 0.4757622143860759,0.08098881047058604,0.04096554198766422,
 0.4711317633563219,0.18687680176270993,0.05409989598437104,
 0.004119389819166042,0.00959008112412607,0.0034287611974979093,
 0.4092909739399778,0.46081671940172675,0.9633414399481797,0.5090338151540901,
 0.41661022626223865,0.9054799997964449,0.4399276720538652,0.5287878755202752,
 0.536125090963555,0.8797233038324554,0.1564826046239567,0.7372121616885605,
 0.26523831905498485,0.11395728736427135,0.1521429195764954,
 0.9745594010757302,0.5160948123971025,0.42534627995511576,0.4581508036338513,
 0.8486341594563404,0.7706697079206568,0.5841915421325881,0.6165687742507869,
 0.01585687342545039,0.052073329387685496,0.960799786989659,
 0.14236884642519876,0.31655668747619026,0.729020382362961,0.8830465244454864,
 0.8511365025988487,0.43318383287018813,0.8499922014755078,0.9377772277481425,
 0.5756737381757895,0.6288705387471184,0.039363986518769156,
 0.9046852863653534,0.012031050680461841,0.7747661068227373,
 0.6570191182273625,0.14445829460308768,0.6977518173687458,
 0.07401060455199937,0.03372808381694736,0.25867535792306373,
 0.029725303666866543,0.6893481824680459,0.5471044528277643,
 0.5112001229895088,0.3804088175468665,0.9626037559028531,0.7913573285040318,
 0.8339332308706836,0.3332125100126607,0.8653088360453041,0.4358327107828762,
 0.010233193531435036,0.24175345540263038,0.8657067055711583,
 0.7211324442582262,0.48699861713399784,0.22564129727416837,
 0.0421496566900571,0.8258916643246108,0.20189139662516575,0.6785381302467228,
 0.3912561627244322,0.19059298116818646,0.2720187384443681,0.701347602155953,
 0.1626120927703948,0.17171149399139396,0.9502325765458058,0.1732486360415063,
 0.0015093301876423215,0.1251819903775847,0.22971165924597636,
 0.2908260379480917,0.5724834427928002,0.3359147488293349,0.28141701557725907,
 0.16149514658080721,0.5674691830332009,0.8870224697964209,
 0.19422675302767045,0.9110824115782139,0.013488899364934248,
 0.3208988998341329,0.5681782778594463,0.5048824551986361,0.23440209911807952,
 0.013417159047193983,0.4349106440130235,0.6005692770274007,
 0.05128872825990127,0.8003276719588461,0.1414148097805451,0.5101038085627135,
 0.20343990924114574,0.7126062168625936,0.063250141508709,0.7801883783184604,
 0.8789554709470981,0.420533341626678,0.8795091759570021,0.7401165229855254,
 0.766049033671194,0.05871810366129149,0.15747548299460767,0.2014121432857201,
 0.8675174764779845,0.3443236874754586,0.8628689980390437,0.5229065486451101,
 0.3987010341040348,0.9038654238325066,0.3836764516220683,0.6698859501139016,
 0.3123174029467576,0.5589735154142653,0.9007107597640006,0.7637142499840672,
 0.46096589614580397,0.4617918633195581,0.005060495730063952,
 0.6553661066187783,0.2114528895729949,0.16793908443189054,0.8613872508462355,
 0.73924155074351,0.10137439753032862,0.5009636582296004,0.4482148846944696,
 0.9538843127180867,0.24663794532773753,0.5643661910555937,
 0.044917427708094815,0.998560697616646,0.9145608313911529,0.9991255525781798,
 0.4072636617390627,0.3312047532229483,0.08523247967172609,0.6220327920310219,
 0.9405338031587634,0.4157988840347856,0.9613385121316241,0.6698531021875206,
 0.008948640440125333,0.3079419769163859,0.7753178709197667,
 0.6540456397157733,0.5049858298304131,0.1928592096559446,0.7513446363850401,
 0.21360284193585188,0.1288586833000328,0.4817039539115273,0.9042079724279493,
 0.35398805713466186,0.8730122093882324,0.1219449817452849,0.6707063408242278,
 0.8400051814578373,0.18243020485068917,0.9905289829548352,0.5490401683907207,
 0.6629077198587108,0.25205832563423636,0.9102659320879547,
 0.18939039369833421,0.5905063751002394,0.471918352325698,0.12066016007828284,
 0.3792495255750592,0.17073358766797708,0.29806056521284696,
 0.17982054098405353,0.3419915326159729,0.02877798652778263,
 0.5308374553881966,0.09000651770778834,0.4279363322182681,0.7682581174181874,
 0.7724839682807223,0.7122424611457032,0.4803521040635943,0.3382776209068492,
 0.5724076717396166,0.606282106595511,0.8064193137050285,0.5220357579042079,
 0.9234830530170932,0.4193426709154733,0.9050605422363435,0.11093644068060693,
 0.7864427571551573,0.0142690563005754,0.07159110538498548,0.3840614670593786,
 0.5386778056588954,0.425570601461533,0.016944902349575686,0.9355069796454436,
 0.2435722341829365,0.44353779445557273,0.8266954639268692,
 0.10320551331002836,0.14760606388786657,0.29237353601717,0.6308478237680502,
 0.466939787017419,0.9453006387003478,0.5230049764796572,0.8861712805071085,
 0.5643656885119275,0.7795923260873875,0.1629491156982834,0.8740382082692133,
 0.07405294935851803,0.8358312261681158,0.5619517932095744,0.3428640151982104,
 0.27379055258209994,0.2001985825814644,0.30484247726978675,
 0.06584048255313668,0.9669516080290615,0.42433534327647926,
 0.4612924639845494,0.6131049465170425,0.3128056439799849,0.7394032835396392,
 0.9163354388060795,0.2999426581953837,0.8430017712468985,0.8759603402003094,
 0.41738208000829924,0.9277609180467742,0.6168294214898731,0.5748601965478678,
 0.07755752151122508,0.8494269990223419,0.8416442534319519,0.9954544563851382,
 0.23843036150211594,0.7785474984034444,0.8669404186732205,0.6799022486151945,
 0.3667503101741747,0.15217513620847412,0.28110556034802703,
 0.11731791230399846,0.5245032569276902,0.6910530164318279,0.8545954983083082,
 0.29143261024186473,0.9774150634057829,0.8983681761418463,0.3729534072866839,
 0.5892961531744773,0.7720068826608023,0.6021152220901744,0.739474601335257,
 0.7772393076074844,0.08649510246274028,0.8520863457715999,
 0.33093760082013923,0.5471721553951,0.16369927259569061,0.8251189350519544,
 0.8768909226941861,0.2913968634641999,0.7034427893908088,0.09757701865005186,
 0.817639321072052,0.3899079035451458,0.9781909977808919,0.5823329254922986,
 0.6255989470851889,0.5652239059835875,0.3214757978283074,0.49755938471563954,
 0.26198887534153303,0.06855873393238698,0.900513590810354,
 0.06754528692057404,0.9672484267536897,0.7158916075641095,0.7232350297557393,
 0.20581192373648138,0.06303801025384626,0.44195798664173624,
 0.5325897650584231,0.9812988605915909,0.6749010309386094,0.0331932888804678,
 0.7376190598184719,0.0714071642220716,0.48359889630875563,0.7397701554318933,
 0.42221147046766205,0.7515387990872991,0.2340142666584466,
 0.04593612699350269,0.5008525205164378,0.4730171147444613,0.9079198545190117,
 0.6449799207720742,0.014626243362464653,0.41098790016760056,
 0.2053845533996601,0.2856897478910799,0.3322590841376257,0.28221834206946905,
 0.3070472445137322,0.16092552343782263,0.3674318884904362,0.880628973936073,
 0.22571872657588266,0.48667461198864315,0.3324173123766354,
 0.6156517245933019,0.7491476993192113,0.9849409414396613,0.0414096580769614,
 0.8043800880635718,0.2078678254013605,0.9089786620890682,0.8478886780084584,
 0.9676419815687534,0.8743838390237291,0.437323488090237,0.2734831996974767,
 0.9569070173635179,0.03610445781143812,0.3643779634877736,
 0.17984887864525878,0.7438096327874849,0.3331292550315116,
 0.44615728229205265,0.7386966933058755,0.29974393518324105,
 0.9432179915995192,0.743307589253087,0.1693554521655436,0.8969302225309277,
 0.4494109217185348,0.5939732999692043,0.9971678705177099,0.25924688855813127,
 0.6335570217742987,0.7595291604827357,0.5900063054945089,0.14296910462805335,
 0.08673110383659854,0.47202090299858246,0.5472636648982822,
 0.9350527070198115,0.4343560241999114,0.2538955009918551,0.08540196755302154,
 0.681622372318879,0.5799679470052348,0.8296797657648213,0.7067782893149315,
 0.46189173130924555,0.17023115074306916,0.6312437244786628,
 0.13191221599626757,0.2528564175463144,0.04706061858590982,
 0.7849062044941406,0.6273232209926005,0.10194780068453502,0.5858667263071493,
 0.0714818299447928,0.16425827878496346,0.9557094807746078,0.7522284154525303,
 0.08899870202614502,0.9960351304944788,0.21051772720953543,
 0.42501720681869715,0.7514862822800781,0.8652969081548348,
 0.40110425494601265,0.41518704491314695,0.9393597103016668,
 0.31007604307486614,0.505708822715319,0.9474482626680849,0.39684185351419377,
 0.6767598925132149,0.8223634392793235,0.06972384381667551,0.8816195807078039,
 0.11016895501613688,0.9648250077383684,0.08705525420363747,
 0.5066378781371281,0.9445941223642816,0.6450401221183409,0.0487703200056141,
 0.9454617049375583,0.8084323068347248,0.8558767125070889,0.3896266442326356,
 0.6022070019612012,0.4003386959152918,0.5894285907883009,0.8484821976168361,
 0.6811706760991086,0.7236019317647683,0.4411590614038826,0.11693850286679441,
 0.5897356095752689,0.547705973622248,0.486690670916215,0.4506069558985579,
 0.6404644199551067,0.9940525596498085,0.832414583470976,0.6327539568722143,
 0.079062430577438,0.4362871328039589,0.8939969415321951,0.6703803506688375,
 0.21344605152771934,0.2793257254851076,0.371008796951964,0.37578543325087554,
 0.5946631741119365,0.28724727166231434,0.7642943162702,0.9238065371721356,
 0.31949301279113296,0.8853348672936805,0.042323826007889864,
 0.1704860730535197,0.45608434467741754,0.31910028970055904,
 0.4865998099058746,0.5980059742179367,0.7676225638627048,0.7166437621106574,
 0.2660878002335473,0.5976188528777326,0.38310978653828,0.01757691438993858,
 0.1000259007341171,0.3142574149784946,0.2881320554057544,0.30375410790127966,
 0.7663315414013313,0.5059537445742369,0.7971770860741397,0.26798398113819033,
 0.7925122639215619,0.305170194241262,0.9851411381076697,0.6192697351346268,
 0.37872829058664204,0.43282379068675636,0.9667471117434684,
 0.9488263343706268,0.19049365981682054,0.009993473384293239,
 0.5201552555951787,0.562553128639521,0.8481472240910296,0.028324295337100236,
 0.4348643611237587,0.717678427021938,0.5143586198711093,0.27684656561635324,
 1.9603407844170917e-4,0.464491748993332,0.5054581973957064,
 0.7978776784498235,0.7887936058012414,0.5630047761241849,0.5079740646501316,
 0.18886478997650802,0.17929888149290163,0.8428488480841351,
 0.33377503148808463,0.4446916205516276,0.07234721637058417,
 0.008868493885401119,0.028493192885419405,0.5318151650932189,
 0.2729302035341381,0.9725628987497938,0.16409012876770768,0.8825500134240567,
 0.1443155762742102,0.73544208101581,0.23889235166374068,0.25106292390084817,
 0.3960380827042287,0.15224022297274675,0.41708707295112024,
 0.8826214143781465,0.41114475832136954,0.233322650210962,0.07544555436652511,
 0.33572955541565885,0.7844817104283274,0.016116544511265984,
 0.8674734597132665,0.3008289215764274,0.5812719422903456,0.4634994588661232,
 0.11356030347675594,0.38080333881065453,0.6772530831781424,
 0.6528925740861954,0.4889205213209856,0.553634811757438,0.5934481859866841,
 0.22557841261500733,0.40691830028624787,0.009857355432123627,
 0.6092517288656991,0.9896897528721648,0.7285110464992098,0.5185651194710439,
 0.8950359724320176,0.2371088943609958,0.1307858351503557,0.767644705912604,
 0.01566666292870278,0.5476863004614548,0.10097030485511482,
 0.08116108743438677,0.0951504337725636,0.2716241335225439,0.468059720758666,
 0.19609769984767622,0.07212145328496089,0.3025933669296469,
 0.19140691155255385,0.6782319466196252,0.238517272003257,0.8369113237144458,
 0.1085864747054871,0.178212527860227,0.034981405767115614,
 0.08660518432870579,0.20818888896126408,0.600227294619369,0.318069723329079,
 0.839050137196864,0.45944942663716704,0.30360329050631707,0.1806667448631435,
 0.4891031279677083,0.24184217167716326,0.0792280386717904,0.4049925148415363,
 0.9589247708324863,0.3656803025783265,0.11421705703219875,0.8323524710809614,
 0.5741652252373576,0.9556142081807255,0.7659170994658331,0.7254126383790228,
 0.36608933938507493,0.06329182987574966,0.5103354275058835,
 0.9622993420234345,0.10167154608582718,0.974935627070579,0.7010714627496382,
 0.853261237310353,0.7225735329801858,0.20637144051902334,0.6925287706541235,
 0.46010983998076516,0.09986658766038214,0.7939702042381683,
 0.9817113143648605,0.2599976255558978,0.2905890164240379,0.2158108031928908,
 0.8769159520147171,0.6465675468067694,0.7436994205922528,0.37926704484656726,
 0.4780210651751515,0.5574733355467542,0.7843698231319951,0.4539522710812741,
 0.8688388542239096,0.9350779453801729,0.3127402263682333,0.47843956592312087,
 0.44484006608384896,0.47462159322967523,0.5329957541248838,
 0.4561971486844598,0.5167340766316024,0.9181108588733362,0.5450144852318565,
 0.8327911442208147,0.3242313372050457,0.36679361320357384,
 0.10476009753872317,0.08770555739334318,0.2616123202481573,
 0.1718746770347973,0.06006630438814764,0.9700118297355709,
 0.32259984972399036,0.6084838728489947,0.37204418938366157,
 0.1437316317128392,0.33342649268676117,0.2042688356272926,0.1984551921522646,
 0.5311672135513573,0.2338922760936306,0.7846195349755569,0.3279688436393874,
 0.6646310276456947,0.3933509118215682,0.15896996772052863,
 0.21716627287328805,0.8732090868738472,0.30351677140984346,
 0.49109082858767616,0.35718010366797504,0.35979587328920326,
 0.4661557433771385,0.2771341613788968,0.8228310269135274,0.932625183816812,
 0.7200327522862886,0.5747433040393508,0.8486007821096037,0.22287329624596597,
 0.9961801547972959,0.9642252278469509,0.4304745148489195,0.1443721609470675,
 0.37865492592515615,0.03727012224690851,0.9607020698153682,
 0.8439174407876383,0.7266616149895629,0.7442319762884341,0.15411252015831278,
 0.7476027139456505,0.8730909365484846,0.5607649163034906,0.9193383705221073,
 0.38402632080227905,0.7450860940939888,0.5158872647604082,
 0.36949551051775653,0.41297144519412554,0.7736009226010918,
 0.8981756947989629,0.9126736001884563,0.6621447051481867,0.9097110088744857,
 0.8508650039162994,0.7827914106764836,0.47179679031490196,0.1087964162772117,
 0.5631708448625451,0.3406551805016952,0.8821828352030158,0.24988135918826038,
 0.49241383645795656,0.5195684890696372,0.36277839783936594,0.504382900032011,
 0.12338995926797225,0.16919095389010597,0.022911520033290778,
 0.370003399070304,0.2121341758415084,0.8119293547671369,0.4269668061845123,
 0.522455529030876,0.8994252084207096,0.8476778780921472,0.9120465534644643,
 0.9991585788511046,0.031046367097249727,0.31199248520111256,
 0.9181997094874279,0.07543262235566806,0.9230550965594007,0.5351511515538084,
 0.3340073495552798,0.8584738471140787,0.8945035020674955,0.6462220605439755,
 0.9413852407227115,0.07655806604066406,0.4648930052221325,0.704917501932689,
 0.49593818038945536,0.8534263775471084,0.7005153450913808,
 0.42493839386262333,0.17066702639638542,0.9244416380568465,0.7681405285861,
 0.8044784661677382,0.7208189140412289,0.9696691128691749,0.13720593710224682,
 0.6525231580027518,0.07916252286634029,0.7531762207152203,0.8212754528655424,
 0.023382160013335818,0.36374932859435327,0.6114199864444584,
 0.46571775750317934,0.5602300754705132,0.6284246676850052,0.3201029343134354,
 0.12952463015269755,0.9979627335280844,0.9064548718510013,0.8506503576772961,
 0.2475373999894872,0.7773247520277191,0.9626140027152612,0.939059806229094,
 0.4292103123323239,0.2610094598444126,0.10696311550283877,
 0.14756875065315922,0.16203551763840063,0.0026136047728471112,
 0.04991493195281349,0.283970480757605,0.7583926637102131,0.28179530545529874,
 0.4381678167433898,0.16332705733062536,0.5851261405092822,0.3433855443442969,
 0.536302472038875,0.33893154315300766,0.04387689498792202,
 0.07207526672840281,0.895781636157763,0.47478651669022853,
 0.16708911650092406,0.026641063270277265,0.46929960672770055,
 0.6584678866222523,0.17023649608034264,0.993178722417673,0.6869660109006666,
 0.703446794109308,0.5047700565232721,0.5557850234591767,0.05163509000643618,
 0.8140572248049154,0.8075552222534443,0.9586405093056901,0.9968036489460994,
 0.44346450667850834,0.8805992512629401,0.6257831783563896,0.9900569554612394,
 0.6444249002313919,0.36422657466854813,0.3652179091901555,
 0.07558804001710129,0.5454048582305145,0.16272322076643397,
 0.3485412392012338,0.3580728250898466,0.6299266978703455,0.49257447516766173,
 0.08534648930401956,0.06734680916822611,0.0649810607673793,0.532048160204105,
 0.15253163497556532,0.6090780613082148,0.1691580365392351,0.9209991915800494,
 0.4195956303235513,0.7762549477488209,0.6053050367834963,0.2695575075367279,
 0.8098269042789509,0.2045264632254581,0.5267078197554943,0.9153391198157018,
 0.19045533101872592,0.4636703802229104,0.3990178280783885,0.4129239235688037,
 0.23007286148627082,0.37284827835231893,0.8993597206017996,
 0.029482070238634805,0.788529200022304,0.48895198708333176,
 0.40812818286454633,0.06657764528131027,0.8507444864120972,0.665496129762253,
 0.5466451872054862,2.868519397081215e-5,0.838409053357004,
 0.025640216158466123,0.31276161149831305,0.8927326219546379,
 0.5589834205154286,0.951923893720823,0.488629285414928,0.7369915116735837,
 0.8941065047534464,0.6080616664543786,0.7764985646821283,0.8328039385901767,
 0.005891682579687529,0.8577541995435475,0.1356094441468989,
 0.5377058638068173,0.0860015598670647,0.7569580349222956,0.6866732709152973,
 0.1287468881210896,0.36260979627577095,0.918530537481453,0.48234750225495515,
 0.05535952557777046,0.30083848263197255,0.4161314482482692,
 0.6778962439517856,0.42740770040873965,0.0075482492921912225,
 0.17520247209751194,0.18903648328103917,0.3499122279428808,
 0.9569862538723279,0.3540672463051747,0.38571977706426497,0.6035636313893922,
 0.9574628605557862,0.3736041573069314,0.6703314571815371,0.8258568248665754,
 0.48413885295278636,0.5965730747525234,0.9281634391839592,
 0.41429138476344496,0.24839267760632777,0.03901719748560717,
 0.0262748352168245,0.520068849170257,0.965807665481684,0.6012517423897729,
 0.25215178301724117,0.32497493195246463,0.04651215243614504,
 0.7545079876999925,0.4048776064780192,0.7838356004213948,0.2023270966321573,
 0.7356124225773228,0.17518606554891958,0.8597692580415266,0.44460074619642,
 0.88071058636391,0.8024755911732488,0.9451252844984774,0.5108045963433879,
 0.5306256670630692,0.9778082944631711,0.6347371770493356,0.28757389556069923,
 0.2528814405344112,0.01283596427895295,0.6108147202850006,
 0.07010200977461478,0.3476060407325482,0.8520540122763416,
 0.07412265171434318,0.010228369989597752,0.729639008426779,
 0.6678308935138247,0.9862643068067957,0.5895671655397434,0.764461199789071,
 0.4590322567792632,0.04591947664966001,0.04181097558023339,
 0.4131479366249623,0.5015848235756413,0.6451832007156125,0.4463729339519332,
 0.4674181115909902,0.843265276947071,0.4373301978725812,0.14654942322533382,
 0.9353085313719699,0.37649891304567185,0.9034686245679151,0.2190565949018618,
 0.368385263093453,0.8899973049950991,0.3744892870792915,0.8483428311669966,
 0.20021416760892974,0.9625383930064109,0.00467443760515851,
 0.5295074675906793,0.20874181617011112,0.013742791525600095,
 0.9476150033287847,0.6855393863524979,0.5045609295923489,0.03738486019611653,
 0.3436596358995916,0.05439304146758672,0.9665587413131043,0.6660791810027598,
 0.7254769209892277,0.48045022324709996,0.19986818813518004,
 0.7259510949665586,0.7536165401275235,0.2619231725401945,0.11176516427975303,
 0.932554026264604,0.3472991705808304,0.39089716334143565,0.6536961975563299,
 0.7241819327206017,0.1153147080844803,0.8295283480743898,0.600182247794595,
 0.4975112859712961,0.8925911264759743,0.08998785839912427,0.778753463506604,
 0.19662604828914199,0.4800971991551307,0.58333730450408,0.009254456597576177,
 0.5924145889359763,0.8507949728899348,0.38368311292112267,0.1883488647706053,
 0.6326848531691092,0.6346199414089252,0.9496451086656418,0.9583149173886134,
 0.9686333850291932,0.3628858831496844,0.312051726362626,0.7903724755519184,
 0.953029857666063,0.8263546005276732,0.8068444246558483,0.9296690242015152,
 0.19434468617819733,0.7631921000705129,0.6506497374224973,0.9051692714128456,
 0.8744251337169701,0.7176806999757982,0.23072041210975391,0.5959244730582584,
 0.4753662744766184,0.9141087285186779,0.2219138842997831,0.5960712451582804,
 0.9905404067824587,0.7996636750952865,0.6799573925636774,0.6348102019653293,
 0.23705733957181496,0.9163745341889713,0.7878670682512006,0.2569518274090583,
 0.3512439304449194,0.8843306864330501,0.05025102159338524,0.7774029238499003,
 0.6408328633533261,0.036830786422205386,0.37207323178319207,
 0.053686083476871094,0.5424755694376645,0.6523178340051301,
 0.8876037906152162,0.5968376102688892,0.5659892120511039,0.9751343021393808,
 0.4830659743009037,0.9460459047274996,0.18034935378617845,
 0.44550659639426937,0.7103522815966749,0.6109936919680976,
 0.013909557726370703,0.6018404696455364,0.15817791778099,
 0.018494244960288397,0.09831564774343882,0.12698735836499908,
 0.1014766971256551,0.5098051526734358,0.11411392350281002,0.4758489927510591,
 0.5900421309076557,0.6386808604182421,0.4439710890264619,0.16214945060487485,
 0.9123639016297937,0.458767779711334,0.20874802143207294,0.8483939786878613,
 0.852400212331198,0.9551726927187109,0.5445604260447253,0.24115580486674804,
 0.9410750289912689,0.6003517073075852,0.266572171808268,0.762842601810482,
 0.7447196983669858,0.9389426458423606,0.49606332339991244,
 0.04944431654657944,0.9048053195518911,0.8087751542204068,0.6213273919516935,
 0.150872928747798,0.6208216234566101,0.313984649053892,0.04934387959660791,
 0.015697948897426595,0.5497139214550306,0.668283667070582,
 0.15201215336015184,0.9497689143264013,0.81904784123822,0.24581594934415918,
 0.374200721995525,0.4608409445166184,0.10906425370711315,0.6731701227998576,
 0.8414202632403793,0.2940018140794707,0.3312698725026264,0.7524684705873519,
 0.34680403409482197,0.5750719354328973,0.7110744484680641,
 0.32053848717819344,0.3363589842193504,0.4901752532739303,0.7739249331963626,
 0.6315485713253175,0.20127261092113746,0.8055318125341576,
 0.47169769992074784,0.6795562270291355,0.7055807087878997,0.3329480250789104,
 0.7560687484311901,0.9190758420894927,0.5520237233187042,0.1051606977618108,
 0.5821253061972506,0.20456182464866268,0.9028654084958285,0.4557218753074871,
 0.7079811940462608,0.8182851745222404,0.4373751198991003,0.0695258007454781,
 0.6216595446278292,0.32353044783878315,0.06483462080808555,
 0.19530233410599984,0.5580567384341403,0.4410996769178872,0.8999488828286963,
 0.5679549596946787,0.178724040088857,0.4969174746112096,0.7333140422288001,
 0.7499536679476935,0.6382539647289993,0.6779193618193686,0.9213373847044026,
 0.1335701891432448,0.20610478833735812,0.6859024249297639,0.2486115602747455,
 0.7369523693837129,0.2621286514157233,0.5821053816291454,0.19629969569923023,
 0.6966746638054708,0.7112016232235516,0.23878182736811326,
 0.035671468948423435,0.43330327471592023,0.7790328285319749,
 0.15609076308385517,0.08631948851057347,0.9752607199215994,
 0.5503677788183163,0.22605198363617052,0.4583751473612071,0.1586588718265759,
 0.48496709560029716,0.16590647814645587,0.10368847780056378,
 0.3519570374681573,0.8145254393969451,0.312190806628893,0.16868018484148106,
 0.330108878482681,0.7779128517599905,0.3058260523867318,0.6965164149313321,
 0.6150330582993387,0.14746683704384989,0.22824955104249656,
 0.7412503408114715,0.3682424666692691,0.2980626644576786,0.08167293753227389,
 0.314205392622783,0.7194324457266931,0.4339506262669409,0.9908828902109996,
 0.7107160631343835,0.02752044347097704,0.5254919687032634,
 0.12956722950444144,0.8299311998230673,0.35255706459054525,
 0.6401636490305838,0.13326241163930563,0.23451596445739176,
 0.6504067168414756,0.44142276449707085,0.3793524639366186,0.4695045093210449,
 0.56420448003194,0.947262944023209,0.13871582727634013,0.005930454097501592,
 0.5355219468040662,0.960181483212827,0.338525623772722,0.7987409260972916,
 0.861083896022159,0.9694772236476131,0.3955280331594333,0.7668990578569276,
 0.3366951590624865,0.6724413588011362,0.850247561046214,0.9243279565076188,
 0.09884373170066174,0.46441827385271905,0.45350438577473406,
 0.036765304571700574,0.3518093453517014,0.966335222114814,0.5052213699546539,
 0.807765379095098,0.7747543741099471,0.4179243576460361,0.7839723836916059,
 0.08933010746494618,0.27963891999155477,0.3923147926307903,
 0.41596281872924434,0.739955844185051,0.8124307937081413,0.43740068031308665,
 0.812303169188814,0.9214430874861944,0.7150724659627961,0.5542999618499823,
 0.6528318248613387,0.05952040056408536,0.592802385024811,0.2235379765732065,
 0.6677066863971108,0.10933505373594865,0.08499921997106707,
 0.21448240945860664,0.16062325123440924,0.673249151116397,0.9184011944581489,
 0.3484722748639849,0.6061872580285748,0.9134658220010373,0.09852519834054962,
 0.871261645041583,0.07689418807768544,0.002712070227209762,
 0.9094570548798844,0.273242527826691,0.32562978044666535,0.04952867712926601,
 0.9271424057499275,0.614796339198442,0.4609321397492272,0.9945189823581109,
 0.16704877068937896,0.9415728497408207,0.3020992756870199,
 0.21878064511633788,0.1799864652507862,0.7116444036441016,0.9426681037069001,
 0.5687689577349224,0.7256697722455411,0.8599144310941221,0.40560823514454614,
 0.8410734414468023,0.7557561485907478,0.022848399644013195,
 0.8732649387260082,0.2310536052285086,0.7561637073685044,0.05892672486944173,
 0.9233269445126672,0.7216108422114816,0.02829492960970381,0.8604069433853233,
 0.7814888909268466,0.7920597526537503,0.7585291760397006,0.42222540449998236,
 0.36824385818985284,0.7711844629462753,0.18495535935548757,
 0.2240712982914711,0.83129223461018,0.6980183324685088,0.31171341396390084,
 0.40055524522786035,0.06547620510484564,0.26766677710829345,
 0.31724816807003553,0.5555467454422304,0.8118805794904762,0.2677452339836478,
 0.6239314118495726,0.38118635960974434,0.7122771169813928,0.547837431280187,
 0.09948025141033279,0.48025325669275054,0.17636986734642657,
 0.312826859143992,0.23191866214409096,0.055288129508227146,
 0.6743790104191663,0.8310200195069278,0.4360376268980741,0.31154616510932054,
 0.7047030035676999,0.31420796036349496,0.9998250695545882,
 0.14761208763950295,0.18800427451349155,0.48268359211980716,
 0.2729317624009877,0.8795408488090163,0.07097727001578535,
 0.07200088398501636,0.17631745792716724,0.4309725457102427,
 0.7008345598195314,0.3590251116011508,0.46484986084966506,0.2791468130105361,
 0.9736638077786524,0.536879275575012,0.181684618998335,0.9014809660883445,
 0.11395403116658054,0.5974863794063255,0.8153128609878233,
 0.32927472360492593,0.7260085534713087,0.7880827479848677,0.2084531869357985,
 0.525355681982249,0.5702897068940068,0.13643897079499112,0.3751318578329361,
 0.5840441185052899,0.4260688600370297,0.2088294172412577,0.37905631105269755,
 0.5455116319768767,0.06999597887403308,0.4045721103966827,0.6621372524045588,
 0.6089714859763683,0.35100565231010017,0.1021160045140449,0.7004521963636015,
 0.6264054460709342,0.7867924097085744,0.3959054447648098,0.44246832463016716,
 0.49385566163339145,0.746177085668363,0.48783892496957837,0.6258174809841397,
 0.3352875178142015,0.7473785398911486,0.7478179289164429,0.44996770309530487,
 0.03134041448226432,0.9979825181766118,0.8780915164652361,0.3590632312330311,
 0.7602560274842364,0.88214830241716,0.6305833543653742,0.33405534128569947,
 0.7936910475381375,0.38585766057727455,0.7510246527799002,
 0.45948629561033805,0.37112746146907805,0.6596976871205332,
 0.2773349572713685,0.4345601998395021,0.9962851353623461,0.6518257210001676,
 0.18406723011449355,0.21716264731248835,0.30915894536462585,
 0.2569154119303666,0.5547379879479553,0.5435662643221499,0.5135920781338457,
 0.05993392044158685,0.48577632441489416,0.25187759653845154,
 0.8732357151931844,0.18449064899180267,0.1861221337479051,
 0.09229684802394489,0.14024835672448566,0.8988837874970869,
 0.6894923883738762,0.9165473384095689,0.44967675705974897,
 0.38868104920236934,0.21316843190761214,0.6070854345556012,
 0.42616155314308835,0.5589170422732926,0.5148087561065884,
 0.32317610269534003,0.9266904006375638,0.4630406400609578,0.7626551325179124,
 0.19130224507518667,0.7586138853654916,0.2150467934809832,0.5266352712475995,
 0.7670087685520146,0.3260690822582203,0.4596436789916617,
 0.026638357275137303,0.2766987158418419,0.7971702990908065,
 0.3375930887803439,0.3183212480130504,0.9652173972461199,0.9239522523071201,
 0.07980118152824167,0.18315231725240944,0.33222867458939576,
 0.5905291545765268,0.7851650750889089,0.43143450349431955,0.2935540718186468,
 0.9255048664788914,0.8822383299508075,0.16665835465160483,0.9998416682439818,
 0.6516778423954979,0.8157623121821747,0.23596200736544848,0.3242244331659965,
 0.667750647197024,0.7937459057965263,0.7086915983335673,0.2505855506855069,
 0.7664781724982558,0.6905177023683229,0.5517858534992326,0.7341136312908311,
 0.029941656849251208,0.8470150148451652,0.06372267917716568,
 0.13822049410085957,0.6904438437623106,0.31177234478112337,
 0.28307686859238224,0.9833167535968608,0.5959613655128982,
 0.033432126977746335,0.9598319119775903,0.2900908874866779,
 0.8266771585209848,0.36338288004522556,0.0749515952548323,0.5460348639591411,
 0.8785545329331158,0.36863522278634586,0.7842883265333191,
 0.31907208042385016,0.2052226151715102,0.10793232893576077,
 0.20127373435167362,0.7143023141506086,0.8054364127352268,
 0.051554697084855716,0.40448600999985374,0.8315135227924355,
 0.08560304990321999,0.1570716148281326,0.7827472390810221,0.0551030853913389,
 0.39764396831784565,0.2719821366685504,0.21355213969934095,
 0.36190242483245993,0.16738490526581795,0.4781091209345847,
 0.17175841153237836,0.4774127354011005,0.7226909290676954,0.5105713436833587,
 0.6193509491331067,0.11962226666843268,0.38560409246758665,
 0.7873505010466919,0.07920435485869037,0.9254634825241482,0.8359178202483053,
 0.3088938118787734,0.08826724182600443,0.8609704359598032,0.837626164594068,
 0.44720344937804246,0.08402304957436257,0.6332786717859218,0.51663984546621,
 0.8207766709335031,0.5695689191048816,0.9750196274402033,0.12968255455914401,
 0.7832213953234521,0.23409372085145486,0.11119279853077724,
 0.11326431799911252,0.8361722042917213,0.6075296949233969,0.8509782091953385,
 0.5215638718260038,0.6440103933277037,0.734554146472046,0.10009262087057869,
 0.6192213493401296,0.9144055308214636,0.1315295941467951,0.4296974831580034,
 0.02911969181597851,0.6330045085637239,0.38934322055273496,
 0.9229624316686122,0.08032750724988913,0.6608036833713735,
 0.08828721527094319,0.3039057108485933,0.43730522047460996,
 0.48710913663302324,0.3378589989554902,0.008947224316738112,
 0.15929586781544702,0.10524824870539451,0.5354257578598098,
 0.4907169541600579,0.429988466577677,0.8432546981366074,0.11955331176530115,
 0.4810611871730719,0.9774338546770713,0.5312143932733416,0.9444250244318243,
 0.524819653853315,0.7873650457905117,0.1477280059003525,0.8475744363905247,
 0.4775012914068273,0.7754586347500829,0.7364144962009495,0.987420526931247,
 0.00631456561606325,0.9587934966312662,0.41991090850456836,
 0.7413663205203626,0.12366929159610784,0.7694806173446501,0.8643024265887801,
 0.4875420295258308,0.18499107103546808,0.7883154296279119,
 0.43011258572220623,0.08609127189471255,0.21858414502976797,
 0.8444377516960008,0.3962141534976933,0.6756384256314111,0.70909020099775,
 0.9568345244676141,0.9087523883844775,0.257089945653726,0.9755393718969171,
 0.5089628385853656,0.29733011980107793,0.8684423987894232,
 0.18416054560312967,0.5192925895387046,0.9108357975872918,0.5650909393921245,
 0.25492694047372166,0.8110032333620064,0.9481298201348113,0.1945009915315583,
 0.1937661701892961,0.9055799279901563,0.14831843871666794,0.156149929113386,
 0.15300250037391405,0.7196292200847404,0.43278572810461613,
 0.013923570421324749,0.6797688037615482,0.47578765976622384,
 0.5815992682268051,0.7190995237792739,0.36363662121566265,0.6913089542739267,
 0.9606222930051573,0.6152598995187909,0.9356908886519989,0.7166587597616403,
 0.343450902062833,0.20428809806576864,0.33808115018398655,
 0.37473224981110786,0.26534163796916677,0.6621153374669517,
 0.9144065288131805,0.1417113125049544,0.2494183917335473,0.6557890236353202,
 0.8025001721832384,0.9036590079550855,0.006407114520073143,
 0.28228890565651354,0.22289517525474078,0.29209996587216636,
 0.1514983452653631,0.41653669134799753,0.06414057869673484,
 0.8465272908777909,0.44949751822194384,0.051480335245231434,
 0.2705554730411124,0.29719790675446,0.4149141267209482,0.4618235298600536,
 0.42355406941036633,0.8319402695463941,0.8127624141738001,0.6566511026411744,
 0.09004684338284519,0.9124153721538936,0.815775361335866,0.269681166310343,
 0.6150443667053817,0.18540108815609413,0.6431070275214463,0.6904331030673572,
 0.5697174920093011,0.21389341582121002,0.4923367215988147,0.5994724772853381,
 0.7369729463717561,0.18346501597453713,0.24422527124423388,
 0.11005568151234135,0.8159790745899526,0.9814935765030044,
 0.36689356969315445,0.7331619425168507,0.04324915450165989,
 0.6609547863057373,0.05063427736117365,0.4826013097089741,0.875676481388369,
 0.7601925242937542,0.18080565172487884,0.9803558876181921,0.6855750136991086,
 0.26837026731307967,0.8781693776797249,0.2442682589943883,0.7483053817327161,
 0.5638338980667446,0.45253610191300386,0.40051174015603797,
 0.8750978113985167,0.7352280202072471,0.17731278273657847,0.43964563380107,
 0.7814270304529862,0.3311105677957471,0.06463739989824735,0.5837205434861832,
 0.018084188341841334,0.086399695042916,0.6329652734689658,
 0.020199332883819587,0.2654343483539434,0.27058449292993547,
 0.6528025524052905,0.16099164276207267,0.2502270364671122,
 0.20324401705793627,0.43525081288074075,0.18143242216835131,
 0.8384706679062379,0.3137816709187128,0.10018435917191237,0.5191624411760514,
 0.737273994778214,0.5169069028728152,0.4639704141998334,0.4758370661046494,
 0.19664687196409703,0.8708245996347717,0.6315547104220243,
 0.16252467183318753,0.3237605837740217,0.5646203771332923,
 0.48174301716745727,0.8484268805426701,0.41566647084556485,
 0.8898548053686532,0.3520380907947077,0.2337138292634654,
 0.035460014709899235,0.3877114384686573,0.39795960377912887,
 0.5621439903660136,0.46196891921920946,0.8776824923653758,0.7898476823559626,
 0.8576995339021811,0.2536440608133186,0.40427065937299433,0.4071428867673954,
 0.5409630065211652,0.2896967756758808,0.5547387168505571,0.46006180391169726,
 0.10482340898353248,0.8491026375256603,0.5115222101051152,0.620147746621962,
 0.39312127005603903,0.4599787271851733,0.25415046527176166,
 0.6562785754543511,0.5667596489043125,0.95336856616516,0.42953691465431787,
 0.033322871582217695,0.6341561698643655,0.9914185358663926,
 0.8535734165152966,0.014339190765026655,0.16000586721325183,
 0.47788429375878905,0.8147032251344147,0.21614565388243856,0.910299147834253,
 0.8860330485447081,0.570445413656044,0.7888298279965331,0.34589483387190856,
 0.2727608118824192,0.7872225732261378,0.8293395221161876,0.03756516581110925,
 0.6837920928826763,0.8095883998919171,0.8514068296211805,0.06925361036281119,
 0.9985416511506598,0.20531616734838454,0.020573945633762913,
 0.043825851457420684,0.18559001028854338,0.8792091147007524,
 0.6464456587857482,0.5547973304444201,0.39995096825635723,
 0.43816452469084666,0.1709272933923015,0.6671463761468629,0.916865104375178,
 0.48136278316870684,0.26300159506736265,0.9455906724812367,
 0.7675689700726869,0.06599041986462717,0.9681454604995607,
 0.03954450284398381,0.8461997002131556,0.5291942553935448,0.490139880278438,
 0.15701114384058912,2.2989177878407485e-4,0.5444656884516915,
 0.8898579154828682,0.38295844704490245,0.2658805131719597,0.1668245240887426,
 0.9972054357942407,0.7775363030892489,0.5245890399846288,0.6432464796701096,
 0.525256197805024,0.9095522081032851,0.8968632593469823,0.17522933232951876,
 0.7290078094875879,0.6726831646870663,0.25538327433513364,
 0.37953998146236867,0.6039805018971239,0.7933674356819085,
 0.05656229635527055,0.2262089402648796,0.2874642828511489,0.112183581898473,
 0.7734098271027592,0.1461372423855305,0.6914333229661979,0.3984950835836437,
 0.7509688572154437,0.4846130581856263,0.5892128578031157,0.40608151514432844,
 0.21805690966802316,0.43762555149958826,0.6542896275446982,
 0.46372156130905096,0.7780965352435372,0.5906127057595364,0.6550892816425182,
 0.45620565251107303,0.38130014221814357,0.5216359753494205,
 0.18168488769779145,0.7400907812102608,0.15333208098644624,
 0.5673891175075408,0.9799685858588126,0.4568450542775512,0.8013096971041103,
 0.8260354991353464,0.010101475958170791,0.3388628364198931,
 0.9095579534463648,0.7369496602114165,0.11533736013728213,
 0.06818672510408907,0.9726696303059694,0.7912965858190525,
 0.046156016989120086,0.4773537066647917,0.3092272537029803,
 0.6555715759079482,0.17175954190879095,0.9220165504678312,0.9450710675565323,
 0.8645705027430488,0.8808973829138724,0.5358850044505827,0.1348832074238817,
 0.995223136219633,0.5416663236903151,0.32005158912548504,
 0.054102995328950465,0.54687158963675,0.7677021745630314,0.9388775440522658,
 0.03543768327452157,0.05537070847238035,0.7496823386283145,
 0.07901160502797544,0.40579250282788837,0.1256149591130986,
 0.8586307381282697,0.40904666908315834,0.046793478739251326,
 0.2909902459530431,0.8133689736968619,0.4961688727038919,0.7803929356355603,
 0.3334164452779703,0.41275745398737185,0.6835217820727526,0.615175622626382,
 0.37368767113424484,0.66248689235828,0.21699930475809692,0.18619956260378867,
 0.21588767135417286,0.9181937915077617,0.4406118283147973,0.7557852785229394,
 0.6183699648989609,0.9322883819437497,0.2895912866380279,0.44997189260792925,
 0.4194694617239795,0.12235910080727253,0.8581844550385516,0.6967405346590945,
 0.9894862436597873,0.6845865894635904,0.6690917759931445,
 0.019131598301119546,0.5476913109495201,0.8403498072223662,
 0.8288906368420049,0.7033179110045944,0.8032996469561989,0.7319846497684026,
 0.24171654040359625,0.7765681111280336,0.026802106861540498,
 0.6913870733230605,0.11266406076395152,0.35986252566690646,0.783805724123153,
 0.9877026793503314,0.964904335903871,0.6495791284976251,0.9599130521958112,
 0.10629559912726227,0.17437030244627805,0.7293359062245531,
 0.8891558951419595,0.7111816434744723,0.3844883028345276,0.3503247365395791,
 0.8376362774115678,0.15743960158851633,0.5766749359964346,0.3640357623347963,
 0.5465823498624159,0.13987209314083116,0.6475148556257024,0.7886739400512797,
 0.8756653862828218,0.5474602095692565,0.1621478600258509,0.6408137933035039,
 0.8466538823804097,0.08147953788482198,0.35066909164269544,0.173044237275645,
 0.7690651931312972,0.04226503636507195,0.8152400771846214,
 0.46196205118435296,0.7012979507509904,0.35683453635535145,0.419690099352127,
 0.7751292188813227,0.16471477031777548,0.5710204045167896,
 0.04295799644874476,0.1170811495353884,0.277726612444546,0.6825958757311984,
 0.7380981082233411,0.3498782385663888,0.17821317756688082,0.846642009770058,
 0.7496872719783381,0.32133291024179045,0.0071103548502001734,
 0.02661493558892125,0.1149985212819904,0.8356800860208279,0.9905478276488542,
 0.00456338009913082,0.6638374494653155,0.32713040584254327,
 0.29904847634387477,0.2334678299415014,0.9287441436428809,0.8576223229999331,
 0.4901566339727037,0.2968075380635643,0.7879282603553415,0.6680771284795883,
 0.4788889300065612,0.4370948235009532,0.7968184966169636,0.9270573493615472,
 0.5747225326940292,0.934592675918718,0.45115682117416955,0.6869246150744176,
 0.20888389744534208,0.13510547670359685,0.9501014942807253,
 0.6843724259293635,0.9607507326817865,0.23183373172810462,0.4614828383568472,
 0.4250272933794327,0.20023043876310775,0.5740592575288178,0.771781607589354,
 0.3919336099917001,0.8268713933940527,0.5139991405366158,0.24126233491611637,
 0.048666194444515254,0.5643725489574911,0.9631596459538531,
 0.10069961928642979,0.4397212333188254,0.5498797496587999,0.5460706559855266,
 0.655373409491205,0.8537617746588797,0.029012619053754296,
 0.49582228906866754,0.16784808998301137,0.9354425940968422,
 0.2083298342705142,0.3607922087593054,0.09865701407679406,
 0.11577323039464427,0.738166453873791,0.10941046439694213,
 0.02293632060423878,0.4545769484684813,0.11621585947060942,
 0.6833296989247987,0.6319305452061295,0.8758999854242646,0.33717263898847327,
 0.22395518684195748,0.6795066082834549,0.04852582700876562,
 0.6734047059895663,0.6620583844389585,0.22462735824640656,0.957503322617681,
 0.09640091957335684,0.10469953346730632,0.8302041934078046,
 0.4302986669356177,0.7480661186664538,0.8722073354438015,0.7774375424280141,
 0.7681062498446258,0.40512014703663624,0.3934870708609035,0.6212538797340322,
 0.8912320564998492,0.18892762956890907,0.27566165434748113,
 0.5674732419953814,0.3910116557505968,0.7986752126537615,0.0831093086442446,
 0.6747491258751747,0.18568051772597616,0.8289165292638219,0.9336723977115556,
 0.7287464205192373,0.3223181710138535,0.5291255605314638,0.8516264101075347,
 0.017656903268792323,0.5823164949300539,0.47511231647132945,
 0.7767849670225218,0.40492612282566043,0.7426517160025412,
 0.47677828811684153,0.17892281049569492,0.3020848869831936,
 0.8781314032629924,0.22597616348278138,0.4505791331954748,0.3047241157094297,
 0.24464985200938605,0.8286281647821112,0.2685835414601636,
 0.07230476589911938,0.6649157855631411,0.7825734730669538,
 0.05008046766105223,0.5525573307612786,0.2492902885285041,0.9875039312238645,
 0.7355304105802751,0.008270850675388464,0.18953565401134242,
 0.3495878044641064,0.15633618385815784,0.4412196810334088,
 0.13726440360185643,0.4114236732528874,0.0967351017543282,
 0.14764887358511958,0.10673785772910782,0.42160593224816045,
 0.9343627997580927,0.6231444395601221,0.1685637739013357,0.12956555201621933,
 0.6325695258768962,0.4013999797818256,0.13372499203625043,0.5375574457421067,
 0.5097233432020002,0.9294817628023869,0.5081751432140157,0.1207349558868317,
 0.21249338764235404,0.45643288413948213,0.09910384427933538,
 0.07498302769882725,0.6072517327999032,0.0805956714316014,
 0.02416015556957729,0.02727077847223236,0.514656076243786,
 0.40225110772363926,0.9322281656201399,0.3133111150782165,0.5026020400270257,
 0.8719084576907489,0.020376036313041856,0.7457118768390835,
 0.1642586331754955,0.5503855117545365,0.9576974611511121,0.14650848177068432,
 0.9636736274684702,0.6193906686664302,0.7391950871805482,0.3602178064248911,
 0.7750737268208006,0.7146847480035808,0.5951539443611298,0.5396731367766726,
 0.5436419171878445,0.7098264657818958,0.517229385752747,0.5110240630903111,
 0.6542745392548864,0.44871963370127055,0.5999815094072172,0.65759048483534,
 0.9771350090321209,0.26958432624882844,0.6870519504307767,
 0.37937672057038896,0.11634005649945156,0.9422573694500124,
 0.6701366568264486,0.7092958662924506,0.3120429409899492,0.06790190729878387,
 0.9812998538494082,0.48495354662904977,0.5376285394613278,0.3230348752413995,
 0.2944112490033567,0.5107161198701405,0.3751792033315834,0.35033357924110264,
 0.2229723860353715,0.6655253299715487,0.9750310107736648,0.5795785031506975,
 0.18906104161497295,0.73090157761183,0.561672796474787,0.4131692758473944,
 0.24293373427760856,0.33951276498013805,0.236136182538062,0.9403484802871336,
 0.33308100717628564,0.3084453314498047,0.23426148023859883,
 0.36185715468467394,0.056947811041486274,0.4860157300624237,
 0.8136600549395839,0.015461657727979472,0.7399308379378328,
 0.3073493630628613,0.2254510001699458,0.9790965609573524,0.23285397131761765,
 0.6864940211024538,0.1072856064007418,0.5721547711569198,0.04742550101383236,
 0.3904500229239306,0.5326247597755246,0.9895755528529927,0.02785822421715589,
 0.32639532202869803,0.03753227042476426,0.19839326405447066,
 0.8665904202768447,0.5684077337202712,0.880130076436495,0.12882747705750808,
 0.8804812032763503,0.8938295820243131,0.5556259118755007,0.11762429071607661,
 0.9896742691499243,0.4213429410549243,0.09285611911987124,0.8000643753115301,
 0.6795452514145194,0.1667547629724082,0.6006665613644991,0.723602588241845,
 0.7163234925634994,0.7494759839894264,0.8775327559403047,0.12397166624493461,
 0.07640405999548316,0.18087420463488113,0.5930077581731539,
 0.8944200018348116,0.06601820647100765,0.6936177632150327,0.7998074196680447,
 0.43206003963892126,0.3334181364679968,0.3152509647320587,0.755752367395319,
 0.0850221183949299,0.5823393107512287,0.36567741804698417,0.9730313303354621,
 0.5201474085434873,0.3579878348287604,0.3285933647079097,0.14626106763343383,
 0.34806417194188366,0.7218618414165604,0.4605329994305589,0.607804210908156,
 0.2541123986365754,0.1324551783063893,0.15174472408349404,0.9566738455284466,
 0.12906711279391214,0.24747122898706309,0.10857141291380046,
 0.9696554626744096,0.7202881313321436,0.038717568610298514,
 0.08394811541538982,0.8395846586314619,0.9484874511882332,
 0.14622242306973288,0.3808668883949933,0.736186437704687,0.35047261838527033,
 0.1798577073374532,0.38221893329330503,0.43326240987126163,
 0.19543112651152106,0.23159064735192103,0.4358682730776108,
 0.9907543798238746,0.8912579995957812,0.7708392354121821,0.18504653306138463,
 0.7569487653828926,0.9413167012860778,0.3713665379487202,0.5229960546280806,
 0.39143766126896296,0.08903646075256477,0.7676073846754607,
 0.7867672889538379,0.08735485216911831,0.4894085396412683,0.7252451980095054,
 0.7981670870525103,0.9702914925945973,0.8232754424385905,0.4835319263542286,
 0.49429361200759137,0.5916450096390294,0.09515087916371534,
 0.9655977124979352,0.608664640129314,0.3734473150647648,0.4363425850063115,
 0.9647413263806359,0.47450366654423926,0.29345620575042264,
 0.8574863016585422,0.32546299446368476,0.9482625063042796,
 0.44853288028092275,0.07525839260803024,0.8004761166857535,
 0.5610763351074302,0.26772046572202024,0.7944023084188433,0.7385348794279413,
 0.21092579784925425,0.2582822120956719,0.09419268941127845,
 0.9015550747137462,0.09530253366082986,0.7125489033960966,
 0.24486631384519675,0.8644006803340751,0.7267060828918521,0.8137930595240741,
 0.04294766328490973,0.18347673911887097,0.16666473178373553,
 0.23388108408239294,0.7138274641478737,0.05348008979988228,
 0.6804192396463358,0.8951193158155557,0.6627513878161599,0.7501341851655622,
 0.9524888002787539,0.9775269590103807,0.12304427191334077,0.2810917123422696,
 0.07692466072608528,0.7914442427098445,0.029822083592693227,
 0.8677541194791349,0.7522292400979611,0.5416782616981083,0.16546213932967513,
 0.6628175515530866,0.2333071310427982,0.6001707035210273,0.7218515737323666,
 0.24415404367404425,0.06501879265493637,0.22579664397472,0.9869898534432358,
 0.5570179168530838,0.7428833760865692,0.8543630076987645,0.6811324815358936,
 0.525860172266062,0.9416253776527369,0.05078694271795592,0.8452701404741073,
 0.7611454921060731,0.9697237315976991,0.4319029255506842,0.7253696410671704,
 0.4861835898718676,0.5846514534846785,0.9224408515181248,0.7166149334325896,
 0.8261972343613779,0.006227033854681707,0.7264625150717762,
 0.8349896361874346,0.21978532364188386,0.3223715584786604,0.8613799255092462,
 0.025690991266823837,0.7561690785955064,0.3149800979831726,
 0.08477069409144744,0.4942745785939854,0.26150037540360904,
 0.31587475316655156,0.6890336481665301,0.05081208422154104,
 0.2116005871290696,0.010255291917479537,0.0074565956583679815,
 0.9264009541337205,0.04668555091434934,0.31243452415370876,
 0.7679411533852772,0.583074151190594,0.6504338508875062,0.06275537190031066,
 0.21389500527648253,0.26518048420222795,0.42305067184614986,
 0.18566035537643333,0.3655117857006269,0.4497765524027283,0.6525144285426734,
 0.0924016741205701,0.7633388413849349,0.7076153818498526,0.6060581760794403,
 0.9659275755315511,0.10216152976628867,0.32297748475656685,
 0.017705174279764924,0.9250807950595055,0.8432591202571305,0.622160267157531,
 0.4406992441066917,0.6205029729699285,0.9440115842986065,0.449242580522889,
 0.9467320374587246,0.27751329235656297,0.3290275904566847,0.6393701497632567,
 0.5986051060376926,0.723671102096878,0.7849087942234209,0.5810662241862312,
 0.802646890298676,0.06107275384596034,0.42688834589542435,0.7704053583245675,
 0.8372091439198807,0.4820424253223936,0.6807847583507743,0.2989873683287251,
 0.18953044326047996,0.9120120269014501,0.7062884245653912,0.4585003277704569,
 0.3079611467980333,0.8623965995061771,0.5192916511256298,0.5376482740855881,
 0.023496095969536612,0.9660928324261969,0.027665934129363023,
 0.5604984567262233,0.3783124604709662,0.7298188482113537,0.5420885114180325,
 0.879519855793554,0.7891948239329668,0.5434322920857141,0.07662150825063974,
 0.9972520140499961,0.5617227321313234,0.12052731151498108,
 0.11933184094705063,0.5079140151262527,0.8394432907356992,0.8791625029385631,
 0.909867786162177,0.28808312986834683,0.8254367677487091,0.4281902129484332,
 0.06554240486941909,0.6383636189322301,0.12171932826522991,
 0.8978465583383254,0.3832466738897202,0.7218899455254919,0.7951954356570521,
 0.2817204923618818,0.26259510494950855,0.6030269198461624,0.6402452929908466,
 0.9648231225826933,0.9003846129679564,0.9670630462474469,0.7072311135092637,
 0.60463944954417,0.6164260191977851,0.7865924021634433,0.922489916913364,
 0.46289457806661327,0.23750562289847632,0.3943779801542616,
 0.7884613995214703,0.4550385366781835,0.3924040875181347,0.8687514802809162,
 0.291195593421531,0.2677390689427601,0.7489369912424257,0.010693517801632413,
 0.9858205241929159,0.25911704502779187,0.7785554679358302,0.6781449171706841,
 0.31800222240215636,0.3170646624341118,0.2489291890029447,0.5830336399492594,
 0.9399149558480805,0.19173991903514354,0.3156862512488716,0.5147945724828882,
 0.9859068754255207,0.761367278345421,0.34188403708348225,0.5402184332021729,
 0.12054315450402053,0.10536219649659817,0.3149443185177031,
 0.4775106714897286,0.4142492550714648,0.7664075209842786,0.11166809666672395,
 0.704395726281141,0.013246355105363206,0.3571597401772396,
 0.20136385036660287,0.8889281624439177,0.8054602896595716,0.7683480291261336,
 0.9670870678027335,0.24961494220326008,0.09593681635960583,
 0.5701710195964239,0.6578704047143686,0.33312086158567056,0.2608254974258515,
 0.397078870066637,0.5704700268394374,0.8933841843952218,0.8151621540140935,
 0.6417615226498528,0.682711791389236,0.7147461589859255,0.2679714165511179,
 0.11851361893605272,0.9787978823293424,0.4639715356601716,0.7538434867645329,
 0.4516841015177432,0.10506556456302885,0.950070364371334,0.5459958543798545,
 0.3701624098633812,0.5974714409184875,0.9084101003469771,0.8457626735141395,
 0.45953393642246965,0.13340184774743524,0.8209848822100838,
 0.6136562381523439,0.9613424931098967,0.5164875275442979,0.9015980513339871,
 0.52699791414807,0.9039433426909056,0.3438550073393013,0.6394349869639436,
 0.04195633070062188,0.933061801338343,0.8231410353029612,0.46355813374084365,
 0.3397217394910609,0.41276983252671084,0.8646199973489636,0.8237524516527811,
 0.6584127505066268,0.8053420671809399,0.5749918197268157,0.8776626786244242,
 0.7078437804985778,0.32131176159586583,0.10806752995388447,
 0.33421394744263166,0.6724522566207644,0.2914684193688235,0.8898601026352128,
 0.4866933085495918,0.7277556837342043,0.3618582618031344,0.3376588901779972,
 0.028974425389064695,0.7438594803697917,0.3175644251768781,
 0.9325599644945771,0.9536248152821092,0.9357110298081157,0.5546063521783441,
 0.026494429947589992,0.8874903695522038,0.6896694432934751,0.485599396933841,
 0.10974385446968149,0.9927192757538051,0.1843957198846613,0.3845958106787637,
 0.8774094071888207,0.8179023070135971,0.4109327814964342,0.1543887323177311,
 0.7168750330599802,0.6233430940630211,0.907228257692455,0.6744314732549391,
 0.9677516528916912,0.4389835537105917,0.1880826495131691,0.7523752632636992,
 0.051017044719231186,0.0371239166563635,0.699929144875523,0.5913743566880242,
 0.6380570535828067,0.04318388350229363,0.44646366892670286,
 0.6657775552043685,0.523114366311218,0.4416128769645834,0.2037982492639463,
 0.722990216630541,0.7140425066491984,0.14167121015761563,0.33352205579308214,
 0.49667564979094747,0.026304744051527273,0.06690859463399201,
 0.3335181686647396,0.7091771122498081,0.45994002202416073,0.6835833658856779,
 0.9749131667860601,0.4988013547321307,0.891575987602819,0.5839906065665543,
 0.08936369820018353,0.222614926705351,0.7345561715388481,0.6837014897165873,
 0.543313687952246,0.89016580818868,0.8299384833658682,0.8953512997005856,
 0.009776644514745914,0.49584298432003227,0.28379181387578667,
 0.11961473629674269,0.31409179580781665,0.47152448022082594,
 0.15289453849457746,0.5464019187413571,0.863342772612264,0.9717160720120718,
 0.22262356276055262,0.33813526703683006,0.08092999887509644,
 0.4016523076299445,0.29929386149572423,0.3658338893314167,
 0.43948047174750393,0.08934623299634104,0.6246257806344985,
 0.4154033923764294,0.4829568225491201,0.9670924772332748,0.2226413646297115,
 0.8536979060890513,0.6530193726199282,0.34515248666891707,0.0902442771851737,
 0.6053953670090635,0.06464851848879505,0.9971961952601163,0.7074501989962153,
 0.5389687998559514,0.512915744127976,0.30573585133641423,0.33012615103117116,
 0.7532567211842883,0.935470061736265,0.5917878576470272,0.7457502638840241,
 0.9625589421727451,0.7324188408146933,0.2583314020058778,0.15763694900532943,
 0.7303170862317728,0.34433115800975256,0.4468522015471399,0.4694456533223297,
 0.8722919683338224,0.5899262878416331,0.6669560600979682,0.6612501438876093,
 0.10942754701698709,0.8183798321744137,0.9844336004411103,0.804811158914602,
 0.5717630020862321,0.5791236128541253,0.05580947070950015,0.7053290750843595,
 0.16017365794718974,0.1346295617157801,0.7066787704439975,0.3822193312901153,
 0.4113444238060773,0.11109120300014208,0.3016452240491736,0.9455304470515359,
 0.2399501572607451,0.7627556212338291,0.8818371166191022,0.26638861553558724,
 0.8982961305014231,0.20575072027105468,0.8514622540056536,
 0.33180176763624936,0.960353101929949,0.9554247315763159,0.08373040056021819,
 0.9102727552408333,0.6450741970409279,0.3123948404194081,0.37244297928241776,
 0.8353816578916164,0.3056472716764028,0.8845253237665612,0.9380545364188985,
 0.24908011393615737,0.13528402451385946,0.3397856064358441,0.993264368137025,
 0.6846664944713885,0.02280834617271532,0.2497482202407718,
 0.04311488076966363,0.8452929854992273,0.6846013178834401,
 0.24521131669582652,0.5974984505000454,0.6726457833419482,0.7590317680868383,
 0.08055476016130991,0.4584983957580473,0.776567281641666,0.908156581963198,
 0.9046180081318385,0.980804409542209,0.7934344942391346,0.25267380212398405,
 0.7534647971749907,0.07393278828463246,0.6329151328492475,0.9812323044386198,
 0.8448854245240335,0.2466454975667891,0.6058018189486706,0.2655845463226576,
 0.7197065573595183,0.6703889048279494,0.9112506085179992,0.8792800524771938,
 0.9576015680368304,0.5478080275753593,0.5098723754106013,0.8221880948322522,
 0.5454239108370832,0.058978284604713105,0.3019696295969886,
 0.9983035036104774,0.4095394814535176,0.3234240665461423,0.5006651311769754,
 0.16368184396169094,0.40254370714712606,0.7435967492775333,
 0.6827251625244508,0.943039871932299,0.8016023750317747,0.7631056106579723,
 0.8335380843422666,0.7488229948661703,0.6355278337748655,0.6593692847200615,
 0.6068718717598242,0.16261444649232604,0.48086026110808255,
 0.36349663964080736,0.46133283785110757,0.8654247344865389,
 0.6506926756808877,0.9281898258033642,0.7800328450307707,0.5069026802415864,
 0.7426620308971191,0.6731719992227889,0.14865932314189945,0.5996763444579838,
 0.19938189710763576,0.5820332526900586,0.624671806095178,0.93944693176204,
 0.9017677537659129,0.8185531197898933,0.6110823694279063,0.38087275316189273,
 0.6668067616103952,0.9420276878610065,0.48424180284887797,0.9439406872552519,
 0.5887707451874922,0.12065470486780983,0.3829310467292343,0.9866012468078594,
 0.9875907045236006,0.0592675392524743,0.3490731466290985,0.8464265524501136,
 0.6282614682655618,0.9165890315408842,0.5767914024142051,
 0.006045133008759196,0.6137058301051461,0.7435308950354607,
 0.8769210827289126,0.8073774083000761,0.06210163134159341,0.4844277648221611,
 0.9223149409179738,0.6672576139058233,0.4229689045470957,0.6414341130901892,
 0.08871281490513516,0.9605181531633742,0.517443856884728,0.13107920362614878,
 0.9633370542135327,0.8789775492659785,0.05246567909022226,0.3635958286728427,
 0.6827106399427266,0.5756181172047692,0.4361171188701245,0.07136662992345677,
 0.4289702425517864,0.7611523703949937,0.11979088887290312,0.8418054546816481,
 0.24031142028522035,0.34416355713773605,0.8051715645592092,
 0.07972157782103617,0.16023276056681568,0.024964518160378923,
 0.9385501390140156,0.37071142893597653,0.31604139439103496,
 0.20927883529709002,0.4070644465333857,0.2192198809917395,
 0.25613507745986586,0.12440224497448282,0.2243364257101219,
 0.8045748222929837,0.9872484965285979,0.5197122764729433,0.9480360276629051,
 0.691107669240047,0.32512236201607725,0.6626075805322,0.1577008235599051,
 0.8055850164319037,0.12174081648140656,0.6340614811951969,0.7402595559979561,
 0.33762937710296903,0.5136177556075692,0.2543647328816907,
 0.17890692663689078,0.6648776231891378,0.8184569203936541,0.8139871758697997,
 0.279704409605241,0.5941825902389258,0.2559434071401012,0.959201676271658,
 0.9308749414386982,0.1102867452096149,0.07552382877213837,
 0.40607761204497894,0.35567525921530974,0.9448437745098586,
 0.3009345193773283,0.858868857232924,0.9059302709976808,0.37439166704225735,
 0.6016848754439933,0.3678806629889291,0.42694143975259913,0.8339500400633632,
 0.546262892620478,0.7047385392437333,0.8491152706782341,0.5227110349318662,
 0.9038537533719097,0.6806191283751433,0.4244231762168562,0.6308393024039268,
 0.767244667388118,0.1495668137331655,0.13416947848923022,0.2997819319215058,
 0.9558803729157792,0.6829126168614608,0.7945162261028584,0.4410366672191457,
 0.9557144362365925,0.15072233653511657,0.7766734325256563,0.5977333263713398,
 0.6618995648129142,0.01167979027168553,0.4807001968311865,0.8786118626713284,
 0.6421718355155943,0.8618091415988602,0.15524474921509834,
 0.04744134407980549,0.9461563259488156,0.8308040908222895,
 0.17252141557367384,0.004445382079063198,0.9116435665907222,
 0.5088193673911106,0.9937408954101545,0.6562681922496512,0.4625801907242184,
 0.6373908020211136,0.534643365525735,0.8892388557836108,0.8476112700193651,
 0.9802451848894634,0.43779582307565645,0.7523632714212488,0.7689897453811234,
 0.6712958881424583,0.4060561725645948,0.3491309841254122,0.20652461285176182,
 0.26155294696785014,0.7201015515387931,0.48870626705306675,
 0.6360203328525654,0.9310671616324556,0.2726157237636214,0.43784456683036943,
 0.029800209526479993,0.44971906179165566,0.6156094607035889,
 0.3679870871143339,0.6842698772854736,0.8389306500362222,0.2736830054923376,
 0.7547075826609143,0.15329858680113417,0.30573643219980773,
 0.3414654386052477,0.7813281707330366,0.21701807189004474,
 0.41522611904792517,0.858379616300154,0.48833598227541564,0.7330704886240692,
 0.603723653476228,0.6178597719505257,0.4741350970801652,0.5132675305712551,
 0.011187486148388825,0.15089544819759126,0.7723063422950148,
 0.7909710529207964,0.682612832306534,0.6043491169881514,0.2036167665240748,
 0.1334876113531327,0.48079433333154653,0.5481804885892411,0.8896152010338519,
 0.6758418170957097,0.9963645589642963,0.9034619156470076,0.27500173579553655,
 0.31529573333609395,0.7442044636757302,0.41339530254941903,
 0.6466704843624524,0.760302714629019,0.2151851898136229,0.009678864842689205,
 0.06097690066369654,0.12187332007914731,0.9967471493536751,
 0.1448344648231602,0.22897489204834642,0.7602656205405521,0.9902766301516436,
 0.9769624501411573,0.7519799082127773,0.5655222655092399,0.3534007531369894,
 0.6399537014777635,0.6760414098554302,0.22619421207479706,0.5936083321832949,
 0.5292438097007428,0.4134743984747071,0.8754712510191169,0.7535195683623134,
 0.6903544560601529,0.33005066381678205,0.5440110359568441,0.6551209307705247,
 0.6289668069411104,0.32106110234599705,0.3449560463437211,
 0.48019240810983754,0.2551949387857455,0.03681675366550052,
 0.8173305995715728,0.6440170309742156,0.4622736388281262,0.4423405992195314,
 0.2091926365248341,0.6788624360892964,0.16828683989211346,0.4784463768963244,
 0.3625822052193082,0.19521381763800028,0.9095757909644367,0.4328191548983893,
 0.38859586766917287,0.7328048268548709,0.3948761536441735,0.6805722920993871,
 0.8901779886620298,0.26184777944669013,0.18955624867938403,
 0.45549966205010006,0.21346695322038922,0.18394704635169834,
 0.46034362423050434,0.37785815679885004,0.7764605047695106,
 0.09076412933802658,0.47877735800574117,0.7283607392609919,
 0.6633718657910532,0.5229608820323776,0.9940862778511138,0.6684114898419811,
 0.9209108671169055,0.8633544448077263,0.22004588863622243,0.5947561516373009,
 0.9387736149842775,0.8208186905415593,0.43728019652672356,0.9601579248913401,
 0.8184112305564697,0.7948909851824009,0.4000571735152444,0.13909509021890087,
 0.7733327563679412,0.320307022929337,0.33184875163467775,0.9682943371206446,
 0.12174650437687728,0.44291999553594497,0.7017289996432213,
 0.8427746984377861,0.9426402367118697,0.7598765797399196,0.6203491147300664,
 0.36092985080017703,0.7070745087308842,0.8706301888403093,
 0.45011885440690746,0.643177201207836,0.19333819633431382,0.9579370903816947,
 0.5250193520728433,0.06442050432011337,0.8082730928431471,0.0842319045273271,
 0.9827421980970674,0.8050773513881586,0.43738852761901814,0.5341996189978149,
 0.8254877426379815,0.676738653857489,0.45612031037220513,0.9822576584230068,
 0.30565575598980876,0.6930126755576762,0.3337827101344014,0.6371849898068708,
 0.4314211604258015,0.694824139778314,0.6239524242033557,0.6535400785696867,
 0.7891961256946344,0.4723825683371957,0.795638093521906,0.21080559014354072,
 0.4071889768455287,0.7230386800716249,0.6935159726684288,0.11494525148573409,
 0.4735387621184478,0.26690781447329526,0.49530105956663195,
 0.4723006821231981,0.8825451354749604,0.8288399967926342,0.6812419730096274,
 0.4959076352319651,0.18888563605107545,0.6593792305165936,
 0.49556915741236207,0.8079809659172801,0.2938854480127391,
 0.02255917256687634,0.24092814922370032,0.7369525768032814,
 0.5264955853093697,0.2985956950312503,0.4799438405894312,0.5986856894323108,
 0.21948146617481534,0.5136420320761687,0.4856714867998958,0.3457092681136148,
 0.20808660842910964,0.663426978061234,0.24829299689477757,0.9214218697394426,
 0.9568532951198445,0.5861672908117644,0.6464431153850233,0.23255411832880912,
 0.5453309628228583,0.9994992029789458,0.28608282932209106,0.7001723266360123,
 0.43916467604818377,0.39160978325524354,0.6399713487057269,
 0.6219306808954107,0.2263619997231232,0.4603390608283736,0.6005952161009783,
 0.1465055218287904,0.47495236627495707,0.7121640555241379,0.2765855460353772,
 0.39074985915425386,0.26630690949167357,0.3223124123697969,
 0.19142813890645394,0.6966449679576281,0.3345023138355834,
 0.32247619490120194,0.23150656869139796,0.02740822077343763,
 0.1501095694178608,0.5432856008198356,0.8116235716511758,0.5556808395124015,
 0.03097414331205639,0.19222622286864055,0.6297770146342389,0.966771451875153,
 0.18101764292035694,0.18577338833682633,0.026476633066300104,
 0.6988888617913052,0.7951653626633849,0.19669903780295483,0.9343657497328473,
 0.13074449218181905,0.9561671084143042,0.16947726935404828,
 0.46583915250696895,0.9526241687293321,0.40436911167349376,
 0.26051890732713934,0.5093059659097284,0.3000464317455521,
 0.45745083503886275,0.49105895915642694,0.4809015143157489,
 0.9098983880828333,0.8338996623084354,0.905354302494485,0.7220256043397709,
 0.3162491794305039,0.9400544243728466,0.37804247808597446,0.9306262179910381,
 0.6091776670310203,0.5187914662978542,0.8076086004332321,
 0.056919753750302116,0.7516533678295421,0.44588092522059153,0.18162065804729,
 0.7821926940253583,0.6609226998304714,0.7430432075027652,0.6140602701479657,
 0.764976212453809,0.9873003798458508,0.5879199844123738,0.8504143434976907,
 0.4021155058682506,0.24625472385259606,0.698459241754793,0.09663782657928521,
 0.19292786581149568,0.46098936720650907,0.0874535325192195,
 0.34845886033390017,0.9832366632131314,0.18647522387451942,
 0.9028803502474062,0.322614084380517,0.8332852120681793,0.7196648540296403,
 0.9685337637812372,0.8682263905751115,0.928719613021877,0.00910770639347025,
 0.2425333501411292,0.8531972797676131,0.14974892636086534,
 0.23792474709624567,0.05565489953196301,0.4250074198952867,
 0.06181176259768073,0.8782775021590067,0.4751332650554454,0.5696291317040184,
 0.7106693007579307,0.2037346170896186,0.7910213179240271,0.9162974424957232,
 0.5334574999141737,0.981590646092906,0.0828943016810002,0.9003016836254965,
 0.7495668753031651,0.8272462499321478,0.5999650272403932,0.13729842568985062,
 0.2946273953552109,0.18053825959829084,0.8491633429580878,
 0.07036756037797809,0.7656280288234985,0.013461332095173084,
 0.3367193099187691,0.21072191098676762,0.1998178909293108,
 0.11465301632513869,0.44970535780777765,0.014896060656261145,
 0.18951736218105553,0.9073319234373741,0.518317528774339,0.7616334888676979,
 0.8337428234574926,0.33047549342172766,0.8268248878462995,0.9240678432084145,
 0.16460253086312882,0.09876336275493225,0.8038293432850171,0.327164182047608,
 0.8752279663207374,0.8666633590655333,0.17450383405541625,
 0.05724299582079195,0.3829354798823319,0.1141344282252591,0.2768850376653741,
 0.9361371557042131,0.328573310088885,0.32791219021335294,0.19850044188255545,
 0.28549725875430665,0.8538060167397753,0.04314321045484826,
 0.29289620692502905,0.729555115939446,0.470051497642923,0.4135870157776298,
 0.18801425813821382,0.5155372660401777,0.9868118416616661,0.8912881152081554,
 0.24588577236167697,0.33118682153313317,0.7865986203281213,
 0.2677196747219822,0.4268414097593686,0.9816067252628615,0.1417913719834606,
 0.25312321270841454,0.5686334583661893,0.5148277227095465,0.4388253102213515,
 0.5403852431083067,0.6241609438846075,0.8829919032846597,0.388531370662774,
 0.5149036025765557,0.2548311579349075,0.4556409545963179,0.26453132943083757,
 0.6279614714928747,0.7513390888315842,0.8932689765176758,0.7750684970733885,
 0.26621008855432804,0.8516979763447626,0.017186150566474656,
 0.0124440248660449,0.44194779291688546,0.7542339687551579,0.2869035486240491,
 0.17009180039209115,0.629591533203286,0.9185573342860457,0.08728864314799856,
 0.3239846921449536,0.04011993150913429,0.445127983965359,0.11218471518813611,
 0.8605457541418494,0.4158328008020632,0.5832366083513693,0.6896672111346168,
 0.037596129252960964,0.9212143781963039,0.4124676813922409,
 0.39519331666028257,0.3257407627607669,0.2865376513596344,0.68068930219734,
 0.14898541052335834,0.9682949722238451,0.901229011772815,0.18661822898458613,
 0.1405072608249055,0.12782703901902925,0.4553114043203986,0.7879004353342082,
 0.3716950295878345,0.07897167568601027,0.5211045678863052,0.7582726758105999,
 0.5537346271607946,0.7765495598726146,0.9935812007650593,0.03842271934366037,
 0.3338980103611223,0.4525164527914286,0.031974714488917866,
 0.35109698213826357,0.5422243163542768,0.8999727096043985,
 0.17833903542542928,0.478033418706448,0.38444424039276204,0.7072130898374721,
 0.7140543811211518,0.660842728610263,0.5981473849908653,0.2405597357698861,
 0.6713848525531596,0.10228760616590471,0.8582381917720864,0.9954978766997618,
 0.4904724184299545,0.2623485286350109,0.24522476176168184,0.8982246303215272,
 0.774425728507466,0.4723453540391289,0.8559873327508746,0.6825146693041034,
 0.2654642785926493,0.3339026525528066,0.4244637806133116,0.5987485134839494,
 0.9761250544054978,0.08719528329644799,0.9133941206356414,0.5491662619677307,
 0.7674622014684729,0.10431956671729692,0.7618458089852566,0.5525838245589632,
 0.1662840141219626,0.9697786382615887,0.5049497281913359,0.5675797390298257,
 0.7422956805738606,0.7007041369181274,0.7278232037639624,0.8654768460988732,
 0.03143679743956318,0.7133274158238223,0.541080513327442,0.8890783907977197,
 0.20361233616229568,0.8356512261401977,0.45455305021518533,
 0.012274004439703057,0.33184485460964763,0.7524870700534841,
 0.14864463405653783,0.9217924707119346,0.01808558949481709,
 0.9991684297637436,0.14815082324628093,0.0492730552540771,0.3605228426111866,
 0.23196518842515323,0.007462898910404747,0.18549199517502024,
 0.022645993838677825,0.0011296225610483157,0.19530829243989412,
 0.6134562574447457,0.42379146434019743,0.41629114883661855,
 0.8246924074064792,0.6870826670666496,0.8977983980224823,0.5395864928063895,
 0.2730521294427064,0.42419289218508216,0.1748680595815909,0.7911377557754995,
 0.014688335593452284,0.46503862265557694,0.0646608188427028,
 0.8491167834094959,0.5657913487751911,0.8052002200626671,0.5836304116551473,
 0.6033428875401663,0.6369564057356258,0.11308386366414558,0.4931876348729859,
 0.22124653894790924,0.399033080450959,0.5833381964978155,0.2197996289064158,
 0.44740387381500235,0.10749958384627423,0.8215915037979254,
 0.8678557940590662,0.9333664056952101,0.5194891314954783,0.1898817437034669,
 0.5588062640759011,0.5165562044008865,0.4515715205824723,0.8621478564486776,
 0.1022369241962976,0.8085916990340595,0.8470509538892476,0.6195883221959706,
 0.10417816266999447,0.7270270392076086,0.5851985458342696,0.435678855390941,
 0.9603461412162596,0.09720755006009962,0.7216253366871336,0.6805135660984052,
 0.10735646530924192,0.6657713159652232,0.09544568972439438,
 0.31734560661424316,0.09716185037440384,0.6211934908120457,
 0.5296825979790882,0.6937477500799132,0.9715479366695443,0.3289276742709073,
 0.7835479496582801,0.3027104427034297,0.3565213657569616,0.7531228452498256,
 0.1653912117590839,0.469241454846854,0.7673908051790084,0.4894327306222863,
 0.8803608965056847,0.226503930445639,0.451266003037775,0.17054824633677967,
 0.19761374427020184,0.3523050188969362,0.8583380256852083,0.2344207896137156,
 0.15023108093092918,0.9563007840415559,0.1680306427384538,0.557788703396922,
 0.18098083557881328,0.4870294856553544,0.4354834968514928,
 0.07231156282193152,0.834754537847503,0.22942092134633685,0.768920607275443,
 0.9699557148940323,0.9153069358928891,0.08956921038137267,
 0.03364130674575316,0.9527553541332916,0.5662648373921211,0.2229565443292999,
 0.46561858254740773,0.4207638680899991,0.010836743429792017,
 0.6167600647731571,0.6366714920078494,0.5403186514281836,0.04896357154071018,
 0.7627410286116085,0.36109185408015154,0.16927482253164494,
 0.5216286223269938,0.8957078404039114,0.8528492531923493,0.9939194681967771,
 0.9311167757328791,0.3851687808205928,0.0651272571994187,0.2160253112983761,
 0.32100157382361794,0.018143641735628213,0.26720126267253774,
 0.5974034592222358,0.13698931270533432,0.8681355393046494,0.7455208547991877,
 0.8326256740311846,0.7888549257936877,0.37874464509828176,0.556285000739282,
 0.03343689695650609,0.9627100486369832,0.6496501905134276,0.8799018073369969,
 0.2660404592315,0.43860851889986197,0.6280244171585471,0.3837621875422911,
 0.11230856558481062,0.012234943089949768,0.8602482396585758,
 0.5676347708176382,0.2531613671954138,0.5291186386409528,0.2356929000811816,
 0.9911688078339754,0.0654189755508332,0.4169315794825682,0.5255571528588142,
 0.3380272725263169,0.30663359894614517,0.4999281951552961,0.7980818444888542,
 0.6431023309596653,0.207754655695494,0.9347265869336424,0.6847415761317188,
 0.25216775578619344,0.15786136509884208,0.686230309315274,0.3263646049256409,
 0.7840135936106879,0.35293092146495575,0.44075101518440907,
 0.7562987409850972,0.32767612816265734,0.3774348507212242,
 0.32551119274445794,0.45093378549403573,0.5555894658858963,
 0.9256893126553077,0.9798578958575653,0.36211711092398535,0.1278951864340756,
 0.30876040009158234,0.26795675910198,0.31975203681566644,0.34492661601388996,
 0.38421023181760594,0.8429713835730884,0.173979636663672,0.6507822980716066,
 0.8341513791738521,0.019523024028200275,0.05091474605825319,
 0.8459815265968218,0.3958139381919389,0.943381357222789,0.3943476389797441,
 0.42823430699882825,0.693144526393334,0.624089251542273,0.2943575336450296,
 0.014924741845712686,0.3615648210465777,0.749978335882796,0.954906464114934,
 0.5158873532949714,0.4239997407068252,0.22547460742114822,0.6444851948663363,
 0.558438683137611,0.4019551671209145,0.32477986499480394,0.4914625616845075,
 0.8553400430196407,0.6207626907564663,0.9800177556409432,0.6616101873570578,
 0.800754865362886,0.7344017588732661,0.3900840832651362,0.8939863958913165,
 0.5310099197804066,0.527733711083168,0.5329712726143828,0.13601915250138596,
 0.3042472582178488,0.876126278169555,0.657553636733556,0.30742696880604514,
 0.9173405103409263,0.22127143360975732,0.1907896073777694,0.6042222690501673,
 0.5516578292640082,0.14844451617583143,0.9547052153368812,0.7528067807936526,
 0.4006690440895778,0.4649395440091302,0.7085761316580396,0.04299364801844019,
 0.12430801308176243,0.5810703350867022,0.07453232215607553,
 0.31591385789873083,0.9172181278738829,0.5931114582481882,0.6045439749305975,
 0.34730419022062264,0.02138992393926542,0.3716637989587296,
 0.40912307691113914,0.6870204268039877,0.7873212374344174,
 0.23936905069450698,0.7202306563747137,0.1111838326524639,0.5537163237532535,
 0.10428561954502613,0.531951261251287,0.207902990473794,0.24764460970713464,
 0.5496554512046454,0.473084509806209,0.9664100485291338,0.9917336773788611,
 0.7287571015617087,0.8949097079930513,0.5308290607307661,0.4232505574308262,
 0.7873752256629478,0.7440628269020624,0.8910812294675768,0.09501237863752454,
 0.9224055252367449,0.6682110104509323,0.3201700802618721,0.772709921580865,
 0.2577226184222775,0.4432317724485544,0.8666889649599397,0.6339911494099141,
 0.2558270662716412,0.7790057384400455,0.4346552289088189,0.5442115787563624,
 0.690359854920666,0.017863425174075465,0.3840488011010711,0.3217195838314372,
 9.220592047488596e-4,0.49039153629757426,0.9383156586556869,
 0.03880037724859475,0.9584754942068138,0.03257935581524096,
 0.5602195850650307,0.9067730253916998,0.3964639651467925,0.3516580383257214,
 0.8442909421651655,0.9940971188524039,0.7032623874352762,0.4324897701307375,
 0.9270644419632206,0.9864975970007418,0.37286320747537505,0.5175865035846681,
 0.9190785009230616,0.16301600991666754,0.2989328914420649,0.5464355376439605,
 0.5611374317279861,0.49819330534829476,0.2055391924971206,0.1571009972837908,
 0.5442650524750934,0.14066843303268306,0.6340946973625369,
 0.32555665239202003,0.844989094435697,0.18046922226552498,0.9158071346576218,
 0.6884695315834997,0.7692770291396777,0.09775972988439197,0.848956632708908,
 0.9689597010490094,0.17961216176496664,0.4844966540500526,0.8789688916080137,
 0.996416105381345,0.8426023473601435,0.5894388476256947,0.2927889192198875,
 0.9260758388261136,0.8578889068014336,0.23713098534701516,0.3515261072871104,
 0.5519011207424412,0.7862304849861469,0.9017406074590723,0.3121576978835592,
 0.8669829502627296,0.08808724341221386,0.026794138287370983,
 0.37858586699382646,0.4111332999561176,0.5438051087509419,0.7541458239039763,
 0.25811479479079313,0.2154857535587722,0.512888220390111,0.22438291614483374,
 0.46954578793292323,0.9770593220373821,0.368090353220797,0.7896705921731463,
 0.7722170559957446,0.3737114343023853,0.8919563920127649,0.31800095966251807,
 0.4921283500876473,0.6039771517307627,0.5478263791775239,0.8838497328096487,
 0.9543400237652128,0.8589763477263052,0.5516917890944185,0.8996482423771293,
 0.3522103452709764,0.614612572588356,0.6997740083047854,0.4443052112787187,
 0.40162872336918776,0.9326567458824603,0.6894603338486709,0.1432329516171531,
 0.6469776036640664,0.2965532185032149,0.9446003155663588,0.35030590877797985,
 0.5230322547708769,0.03457604836400374,0.4064056816988396,0.9411261485029967,
 0.9531223472185663,0.7297002356376259,0.41442791668778334,0.9551219698899556,
 0.9392353257751401,0.1899376306160787,0.9129516301951099,0.9844726040384821,
 0.0896528058488324,0.8590915204567202,0.34074818416356045,0.181791723166438,
 0.16800946867134126,0.9485624057775126,0.4057242438308567,0.9840444115382349,
 0.5805246362740083,0.33274080897459823,0.5528647110953102,0.7281164035544674,
 0.4563743516655584,0.9002345650206312,0.5301319668871112,0.5146862463189809,
 0.905665740916484,0.786255962384174,0.7276096646866727,0.21830756157318354,
 0.28166619848231744,0.24564246172461335,0.9158292317425859,
 0.21179102254941828,0.33112425365889664,0.6964139775874326,
 0.7736023666682804,0.447001330951593,0.17108429824486793,0.8200374766851831,
 0.8258077639527973,0.5332656453291964,0.14698905012294516,
 0.07146267719821331,0.16571021090402838,0.8369561292314173,
 0.3981415346082018,0.9426848366766545,0.3147758881453606,0.03491076018861583,
 0.5220030559320298,0.07354153396688423,0.8032284734084538,0.375809883175226,
 0.7293849430678487,0.703974895389233,0.7858989134146485,0.8479208059015364,
 0.7890998859216932,0.16761696321594566,0.06323302942484088,
 0.5139042626761906,0.400446884751845,0.2687698073831788,0.7079189289335198,
 0.9641818168762877,0.9655954257140709,0.3721064919421968,0.5434736167089168,
 0.05238320755498005,0.46783368031207173,0.6864951958281853,
 0.3669845468066568,0.1772330760466393,0.014086327610948524,
 0.3621510334301725,0.8918173511699178,0.844474869588727,0.5482061236737157,
 0.3064149390089679,0.740343464212771,0.8556056081134201,0.4435562530002146,
 0.8935983572554255,0.9914931053314859,0.7926119502672112,0.8898799312350191,
 0.9708505461735814,0.7509301862668017,0.2235684382776406,0.9628015141003966,
 0.01905674023269055,0.7876885705566233,0.8695404979511768,0.6876843739281352,
 0.002547883552486141,0.3864154163595299,0.8534891884413132,
 0.4954328510756716,0.3828097083408486,0.986127889479636,0.41916363483149155,
 0.10248504871475239,0.35697176973675204,0.29759890376763065,
 0.2083790404630096,0.34221649217534567,0.2984888010496459,0.5204462329822088,
 0.5671421075086319,0.871298993277382,0.19488977859761425,0.8731632226495203,
 0.7305009196668577,0.9494483909547398,0.10387752399282035,0.799853917380563,
 0.06603467331812274,0.4623447652152528,0.388878209301601,0.8184509988450555,
 0.6499372086511919,0.005830646898668079,0.990370484519952,
 0.05608797286519962,0.013150690701167589,0.8077486097560347,
 0.9913444351207743,0.6218105508616969,0.9133199517956252,0.2630099854606479,
 0.1499241149960614,0.34785796003880964,0.3297441173363548,0.7524543393980443,
 0.3537566264506239,0.7884987877660032,0.5881097918695828,0.13133138202211658,
 1.9684168589217954e-4,0.33349095656088956,0.49772178168018266,
 0.7266280268156076,0.6858018818562545,0.9318869212760386,0.6783036606623776,
 0.041254836518127336,0.5039708555490555,0.6784234107862659,
 0.7050420288409772,0.07771712146270815,0.6265460759554475,
 0.16302042599339828,0.20109336643355924,0.31974016264404803,
 0.38699117936495364,0.44538749857963733,0.7062138546387494,
 0.9696413471351363,0.6493961024417133,0.06564043191167412,0.6583876249161121,
 0.7204416112170079,0.8103101411881567,0.9792133932005368,0.7761733950569148,
 0.9745565944546842,0.5056495182855416,0.14759957042732974,0.7515855321941033,
 0.5040845912251799,0.3016063668378328,0.7619119102818073,0.25319848708590187,
 0.6952980712452309,0.7224053941129783,0.15492227689068294,0.9148723000177105,
 0.9783572107868898,0.39837101697128696,0.4389502227172509,
 0.26439622654614925,0.4970152962391179,0.25625379686981953,
 0.9528782592092454,0.6247181307463058,0.14417498919273952,0.7541011734653479,
 0.31295155782771733,0.13799405424554312,0.2950991308953961,
 0.7599147189664799,0.20768746528749993,0.7955902410253751,0.5904839402488585,
 0.07060377060389933,0.7641005426611163,0.8056032803488175,0.2470698384306801,
 0.3042588352070774,0.018994727983798865,0.9483584597529151,
 0.15866040180863605,0.27964732979750884,0.2840933257509869,
 0.7874626904545792,0.31275753803909034,0.8725408836427946,
 0.12225602540361635,0.6514972890519032,0.9186530672623505,0.4361206035586631,
 0.5304303366220601,0.20524162260195766,0.01730877205656023,0.701551848625908,
 0.3725321425578363,0.10328347971358975,0.8735420385249959,0.988806819434736,
 0.7774633682761348,0.3720010923883923,0.06874144440122343,
 0.45388997450388235,0.47367038394351735,0.35145062950406036,
 0.7772786761521197,0.524958085476364,0.10299071868547283,0.22008112659432633,
 0.5435857168592751,0.8736446763493941,0.35405421190019715,0.8841867098555702,
 0.811490918027453,0.3523169450217114,0.4324360731495075,0.4496552158766872,
 0.06780881322653753,0.26317521600665905,0.01133063227205755,
 0.4025352904703998,0.3671646255464688,0.6399622721308783,0.5033424821118472,
 0.6774640721422932,0.17852292726349828,0.4197137774666967,0.6212313243259253,
 0.7678770046157023,0.5435239086482748,0.6966812991222269,0.1248859787936859,
 0.707690384391042,0.5702433296423395,0.6278666578890483,0.7627378967295817,
 0.6363127677232759,0.23222772361155086,0.9179735146389981,
 0.13523457058919852,0.9197136306835496,0.5300144253702199,0.8435065191085751,
 0.6492704848025586,0.054599716093084893,0.0744458129277854,
 0.8678697848652006,0.4475643222740635,0.05835330519630033,0.8136531691689468,
 0.7902141848060283,0.7183734452796345,0.6121293249406072,0.9266620300010341,
 0.5154322430624457,0.9734003882978711,0.1977317581338116,0.7328742368578616,
 0.4998032038707131,0.48030548728717093,0.4211946627678659,0.3683671035149281,
 0.4756727774439452,0.13238277942217258,0.30552445080120116,
 0.02446379729851289,0.4321864389485859,0.6928190049398479,
 5.771591216507943e-4,0.7104637190563347,0.230920276455865,0.3306406947014834,
 0.11207031716190385,0.6442364373541148,0.48461811950865563,
 0.5774622328566965,0.4509079815153113,0.49740897242385274,0.6972125304617086,
 0.7268713728184273,0.8422151445938821,0.29918416071311615,0.6246996484059847,
 0.6747486518787932,0.774637126982108,0.391040311541735,0.7410743914943164,
 0.28446868092621647,0.21892220434326481,0.5996828648019018,
 0.8217633592493915,0.10172276867030616,0.30089405420400894,
 0.5777402607641174,0.0013265912572574257,0.26478557413152015,
 0.7610275768246683,0.4973967351452774,0.9549538059549944,0.33667619119934056,
 0.5851379446790881,0.09099082726656604,0.5014895947565209,0.3888170436389605,
 0.4975269303604447,0.8324736064563487,0.09468756716138627,0.9864677265775187,
 0.6044824857443761,0.8506436225740189,0.4837392154421687,0.4302014975863835,
 0.20295798497408768,0.7099337528674601,0.9704778820473764,0.9785688539123064,
 0.8872528323564696,0.5037146090050862,0.7771987269264576,0.04771782764529531,
 0.23837903695366247,0.6023256919779848,0.022326155310927742,
 0.3031450794491516,0.050468122470161414,0.162725499827565,0.1763349990381411,
 0.9388684815470989,0.5034715029796311,0.6074073583460204,0.2890865265039324,
 0.20634432988114026,0.44547223300541394,0.5278104538671932,
 0.4550343808709323,0.616456858872924,0.6754367863425671,0.35641236156142675,
 0.13107382445416282,0.04945729161596901,0.21255600288358467,
 0.4628357664453904,0.9647640584135073,0.364820810473647,0.42377689115447337,
 0.8495014940429857,0.8990002175016539,0.25996965940639405,0.1292785523613016,
 0.6316206579799716,0.8967456346900261,0.3711582408141191,0.6747904830886678,
 0.8695519304803486,0.8378793090994147,0.8494240341271352,0.7825487634757182,
 0.8905469630556722,0.47443116289013454,0.07148150992135083,
 0.7709897970871274,0.20358863752602674,0.10094352907624904,
 0.26805847753066814,0.25046862394295244,0.2931209714546943,
 0.8705779473329505,0.7971674001400804,0.28150657249724853,0.6135702787470245,
 0.9265977102236318,0.0884137336534091,0.36158112481401017,0.6203540313220399,
 0.5176162058284798,0.21167909969441245,0.39051532463701444,
 0.6467310031002294,0.5432843702488365,0.31301689317979,0.11737461976731878,
 0.1175039317936416,0.2512511844745162,0.9594961009137426,0.583942957112452,
 0.4119221220676812,0.010086538521777877,0.36575749587712947,
 0.45711348150808817,0.9080573322930967,0.6651851830097512,0.9188162274128081,
 0.103678377909044,0.544538844735293,0.526282823589014,0.5435505132104669,
 0.593714521854872,0.4302500021631912,0.32285939809140374,0.5625424380007241,
 0.06912796328753745,0.9241598756927494,0.4429754889039458,
 0.26587865099185737,0.11986697441960836,0.47467490056343953,
 0.6336528404602338,0.4083366198223397,0.8475693734107403,0.4683404449704607,
 0.1393295072273052,0.03353609758263676,0.1575985888043716,0.5141604570726388,
 0.37310663803819155,0.7275410612889964,0.44215446186658935,
 0.4192902523430452,0.3082506426264582,0.35755738674758,0.6944728902845914,
 0.7140778072191458,0.5352208132394378,0.42886445623203695,0.9985538888430885,
 0.2780289495930258,0.566220428823474,0.007644105656910316,
 0.08046752516122957,0.474514581279572,0.8942692263005073,0.0819645062306299,
 0.5054591635174439,0.7687079840018101,0.6357375601008892,0.4811208013352376,
 0.26616573626479845,0.9289882095087656,0.6107644329975194,0.8517293680739034,
 0.990177565441108,0.8386381544568997,0.679146823594558,0.618491840253062,
 0.08256956967259887,0.9914373237427359,0.7789156081068419,0.2782915070948401,
 0.9842877062853822,0.07838510403544408,0.7344880404779967,0.7734057970329942,
 0.2391277684279468,0.3448912593463582,0.9996856163200747,0.7580583799753883,
 0.6398475672381747,0.8718346310258654,0.8203306727050756,0.3519908446551152,
 0.799705070461741,0.8262881533365364,0.02385947821987422,0.06659269129702539,
 0.1918402505809177,0.7200786008243208,0.7039995942974792,0.06457465391144412,
 0.36088602286959237,0.7853715430678473,0.187510585059543,0.32330824036668704,
 0.007998178525027821,0.7253164720965865,0.8356470513430858,
 0.27467369505509776,0.9273399423997721,0.47803721970645563,
 0.16785868789663239,0.6802804895012668,0.6826500986043531,0.7833548755395998,
 0.4425085951529971,0.8313600939911242,0.11773425089935785,
 0.17273018252991112,0.4112682875305048,0.6657368302547071,0.2891397695349185,
 0.8114032128933397,0.6131140241801729,0.9033655681361983,0.5947395202288498,
 0.8033492357817676,0.8812518127592,0.1549295510820391,0.9002450600948064,
 0.5884100031220311,0.1967066018781225,0.7852750072188672,0.6801605400494521,
 0.755490420140804,0.9055128709878656,0.08208538929524534,0.1877982605149019,
 0.6677561138679962,0.5938580013713188,0.8325192403176378,0.6127274886885624,
 0.7623535777253394,0.8209091687965498,0.995566147745285,0.07020461542504852,
 0.8324539993830556,0.1924984955898641,0.6095831810317653,0.00844155347439246,
 0.4260181230505071,0.15500805721769395,0.470349203017405,0.14983505154272636,
 0.6883345688539844,0.9309180641943597,0.49098119852820865,0.5980701704486429,
 0.07175485576016238,0.5139196667708978,0.3140288589395265,0.7148705963098152,
 0.2643997936329491,0.29295567986498616,0.695594668469332,0.20663651684976037,
 0.7916760861643972,0.4173172390154244,0.1560722348813588,0.5182790033375421,
 0.6598145154916124,0.2348419251687338,0.4126065289188455,0.6977671663766921,
 0.015253841378505273,0.7219996601179944,0.8820676339612293,
 0.07213644659929808,0.06955591463666433,0.13524895548966598,
 0.5830404530488185,0.7572494883692547,0.769210655073978,0.004531187014167437,
 0.8901275342362577,0.2750130960624997,0.920301555689987,0.06819885152739003,
 0.48205425468594676,0.43466629345128804,0.9677883560435563,
 0.32261254635946457,0.9321014267750645,0.8813545854594805,
 0.21116593474634404,0.35165125282638887,0.5695838752034481,
 0.6282757301573553,0.6833562037298591,0.44849287297585183,0.9989692368760305,
 0.7805225899380859,0.38879077024835007,0.2403237555795229,0.9726227183112275,
 0.7971703276962004,0.9279335608258732,0.6790317928056948,0.3552305885811977,
 0.1159118305044814,0.2324018501864623,0.6052162145828692,0.47170110673647747,
 0.8935198001889748,0.954670837915703,0.23621964459176314,0.6230394787157361,
 0.9143744756470084,0.5203508758339522,0.37666638666547403,
 0.34140026563148607,0.05968254929425265,0.2624341649280112,
 0.9503500011558091,0.019561073127710493,0.8911818054166154,
 0.09862815600264407,0.7405102163800554,0.8243628309112239,0.3025985644050787,
 0.6367492379221171,0.8159790124204194,0.29884210838473624,0.8601667641192718,
 0.18176635960499865,0.25442330836997895,0.025861882295222238,
 0.784426820527261,0.5182757183914817,0.9131408025213359,0.183770339431515,
 0.2461811830730326,0.37018710679639666,0.7614948002120254,0.7816787324120003,
 0.06145687870947092,0.4613220800751213,0.7047606822862893,0.6001690434873118,
 0.3219295556423071,0.6960135826055307,0.6787042328780113,0.48968294059947326,
 0.5103236598497334,0.4947733760586812,0.7739939866566525,0.07278397371948919,
 0.0011398630421540767,0.33328229462412495,0.5858076983603574,
 0.7331473931134355,0.0902063674480289,0.2729477259301255,0.7595086787918545,
 0.46916998231315776,0.8458608416573693,0.1119467421510878,
 0.02365469826940958,0.43426299647957767,0.47591270936687147,
 0.31901895361751464,0.8381457130118145,0.35689140828866506,
 0.8607999891079843,0.06578805765325169,0.7897036484906466,
 0.17038281135767108,0.34511447140122486,0.8040248564178519,
 0.16581545750590254,0.23260313363840168,0.16687631279284554,
 0.5603388629446662,0.41731599027998345,0.748074951036353,0.7741397849258155,
 0.80226403516162,0.5424331780081498,0.13635412409529224,0.1972903588756577,
 0.6578886463131566,0.4274907419601588,0.18557230607346786,0.3194769334501254,
 0.4868505837180289,0.14795460953592865,0.5836113534868836,0.3619140958738497,
 0.6648121157102508,0.605460614417835,0.2587655792780664,0.08625773468800346,
 0.7600909535767941,0.4269585170171577,0.7639935459226421,0.5802777941064439,
 0.8239091193932372,0.878732207205577,0.9287071693777157,0.8186554866909954,
 0.4093345725544316,0.8571264322807113,0.2936014159683017,0.36382337641157436,
 0.8374057114650755,0.8990983968435189,0.2458557524329833,0.38994149520296606,
 0.23965311166450454,0.06015362332672791,0.010551027295977455,
 0.03959351660049315,0.4592478838137207,0.6166839721020138,0.5351376247156645,
 0.24136777614077987,0.4038062894571173,0.9309060137023374,0.9441841763342922,
 0.8801757594190928,0.8032711224611443,0.10445206675080865,0.0109928736257775,
 0.7113136508392145,0.3638891475016206,0.36858738949428393,
 0.04285125485984864,0.23008498810206368,0.22386544480448056,
 0.5859521394011559,0.0624738410828487,0.6913338968003485,0.4274807224418582,
 0.5757887255772572,0.128636755209954,0.3371153513142968,0.3873148362769152,
 0.6846381327424622,0.25983591756177526,0.2545706595276903,
 0.25843820467340206,0.06053429824996559,0.17299444645037565,
 0.31451193559058355,0.6876741251297142,0.9945635883418427,0.8001635514352163,
 0.7110838722204984,0.5084185293044434,0.4055700233522622,0.16951164118257167,
 0.19062329082216656,0.042758240431786776,0.18257776417909788,
 0.42890255480925155,0.11449533884654561,0.37171738716072944,
 0.09009286251219928,0.688690284361425,0.6430732176312173,0.7777335090347061,
 0.8605810484671745,0.3959892146302507,0.8964876188238718,0.11944296928872378,
 0.8090550241748077,0.15269983888131955,0.3197933593814569,0.644815023652043,
 0.9088833781043147,0.6453897830433681,0.7529506010077307,0.3553646370853726,
 0.6231780220369877,0.4833700336331793,0.967128843494018,0.7525950508905351,
 0.7682109467938203,0.581997732154568,0.8314143452656728,0.2602184780262822,
 0.580850879556483,0.010638461251395781,0.727772597247778,0.9762441519795334,
 0.9613922606222486,0.09060180669848106,0.9459137602888009,
 0.23504220050458424,0.487216537365766,0.8843859800946506,0.5849110182668871,
 0.6186222409823338,0.5841651520455435,0.5452765570308827,0.14397968972007924,
 0.15473417514215715,0.6020686285473817,0.558864500689069,
 0.061277403706534006,0.5781588524669862,0.7033118228141158,
 0.13486028910219827,0.2851404830075167,0.3489929428469113,
 0.09002948592977711,0.6536353178749168,0.44073913677423093,
 0.17826107036919836,0.7919553899323025,0.003178746552247147,
 0.9388151135738266,0.08061231540117597,0.5453258917270984,
 0.48956396949439307,0.3458147006970479,0.34084741946029307,
 0.9746595496879091,0.2162805760064509,0.005321324516949888,
 0.41897147380337196,0.8721804459296356,0.6957951866345219,
 0.25790346188180313,0.8320315800652857,0.7715940583606122,0.5618011748577135,
 0.4167751808551857,0.570370526903739,0.14751438786637716,0.8570821932016486,
 0.806883880470812,0.42044364101465304,0.31507821945647896,0.9546600983193693,
 0.019596653431867117,0.48933303925941984,0.28079588716301873,
 0.5143894693546652,0.6258922366384523,0.1563181614994874,0.48771757820044337,
 0.5193929645867958,0.8064896110974772,0.7721341528923252,0.22940467651162189,
 0.7086225000145208,0.5774154868274465,0.7238750029093705,0.818100566993324,
 0.18844919164785046,0.05946472439732298,0.9218950142907834,
 0.9862307821760453,0.002322515808476977,0.40506429050509973,
 0.8855106037030503,0.36555515022725804,0.43374810611226167,
 0.8460797383087769,0.14339923068570437,0.386682487400263,0.9057566953917844,
 0.5748456954415714,0.8487187035618662,0.8442456236657533,0.5165319919369176,
 0.6134051952248536,0.23834771522574671,0.7300516080650543,0.985704833059304,
 0.3542636021782999,0.6317549672779854,0.7808802429678369,0.21216786578977775,
 0.18765786482250935,0.972717857639085,0.5106192266046304,0.29992778852315627,
 0.410895836052104,0.6046016944457895,0.6240503507873435,0.3499116211642401,
 0.13002959495437483,0.38958524872821165,0.8015324392734532,0.578378533778156,
 0.6157810010629604,0.34542809279000863,0.3783065601207666,
 0.29132214576681015,0.49519122856474174,0.4034347998328659,
 0.8470196636461365,0.8269856571088403,0.6342693237286825,
 0.011169911287213563,0.8236067743705137,0.12878493568721405,
 0.0678708327851758,0.11345795079180476,0.5727251669868129,
 0.07441793583711809,0.8589486366152921,0.7984409871128513,0.5044712447112907,
 0.6434970886118714,0.19485599405296228,0.4101965753705973,0.2575636300764377,
 0.4624438833706769,0.518359481283665,0.45895045587263394,0.4680053739070217,
 0.6292629835359327,0.10554819948203764,0.2932231033139143,
 0.016430444616378437,0.6521724923703227,0.9349438858580182,0.986356288047195,
 0.9854783730630259,0.8347028742422543,0.512445090893372,0.018682584875249875,
 0.07140224231538261,0.05443002741577452,0.8518626251914692,
 0.49402598769603157,0.6290488444695769,0.1496076023536329,0.7996888693484079,
 0.408729409412115,0.6174097884396945,0.6551645255595836,0.1679105835029715,
 0.771403239939025,0.7723757283696161,0.8049889740679043,0.3282758651999651,
 0.28321820386447416,0.718981394408159,0.26760677354778006,0.8785704871675377,
 0.42072878982238915,0.4782448584536002,0.01580556820400192,
 0.8476973972512771,0.46104815431001533,0.7686117525965059,
 0.26311155179032963,0.6477847356782966,0.7269201628119949,
 0.03158382691549633,0.9658954362738239,0.623782458328626,0.6674240197383062,
 0.6201895113570675,0.5441397048505257,0.18536427155901258,
 0.38170420285439965,0.42391734237677947,0.20575826826255927,
 0.20675143411924957,0.12122316694780144,0.6654243986281463,
 0.16556975425305254,0.9318122523392025,0.7729345630882754,0.406190188163008,
 0.012152530848400622,0.07383737977047267,0.4555454789028197,
 0.6159276841039177,0.9449912106027188,0.3449155151632073,0.1464872867999324,
 0.8976401304972284,0.6987547539558779,0.08204955334854536,0.3185059556896498,
 0.26698378013013846,0.8737858414005226,0.0618567995004059,0.7433780228416721,
 0.5670558907997237,0.20648989080881253,0.7176188424221703,
 0.06888617060551683,0.9543556506453063,0.3853989951195106,0.5916085848540524,
 0.553835145011206,0.6433446768944666,0.010274899564296835,0.8295295467837438,
 0.8303556429578398,0.623751337338031,0.6066769146041839,0.8839355204483003,
 0.08621985732487047,0.3511432011690031,0.402679653421923,0.38447107559867455,
 0.9403257575479883,0.25055551540998544,0.6469830940152054,
 0.30686425279686147,0.34363243593184756,0.9522331603214709,
 0.8006920148048078,0.8033211410026796,0.7392362326459808,0.6720418049409722,
 0.8552046103512907,0.6211476410943211,0.44535367775024504,
 0.32184590502617316,0.6338858395205518,0.5799377753233907,0.7172149686064689,
 0.08759356198048085,0.7684561180997328,0.10803463454152107,
 0.8950884759653078,0.9399826622010035,0.8689365432306801,0.8142407152836467,
 0.8389218019338475,0.7472148683861095,0.38172676940879546,0.8744505194984623,
 0.7254848502115225,0.4855690776027528,0.5239950483449167,0.868646077333303,
 0.5489849756887565,0.9416949951524599,0.1388837378562665,0.45817953355785845,
 0.37137943938703244,0.1598854193049951,0.423278546366296,
 0.047924105877038436,0.6953504262383169,0.4441644165566192,
 0.36575502547064276,0.6188059596634019,0.2740029487718937,0.7834598444683109,
 0.6723417836614587,0.26989739168412075,0.028500735142361844,
 0.9617164537523352,0.4193756525482306,0.05900004119446167,0.7211024528931983,
 0.9973306293678178,0.5510851481779251,0.38237899153292254,0.4683462817345039,
 0.8515653072922983,0.30787514098624813,0.5993461605620685,0.9895133473037463,
 0.4122179729218889,0.0424560464939967,0.417534216897691,0.2614558997137625,
 0.25355693724661055,0.6073773990856778,0.30554380941425285,0.858229976889841,
 0.002409905443928828,0.43202905377749456,0.7616610598852643,
 0.5794623767905291,0.7956586821565418,0.6031036222689575,0.7238807262038766,
 0.6513374857388609,0.437129841711527,0.9685439494656789,0.18522849380559203,
 0.919509541059021,0.20711657595314925,0.03989804207287706,
 0.061471991453125985,0.08619090033191845,0.9691871699143352,
 0.9572029967417991,0.601545819259313,0.2017304157859069,0.9985842633653689,
 0.3988430237256809,0.6733446929197267,0.02809138225801744,
 0.11900029107928534,0.6369858183579213,0.5792747405692756,0.9254737153567573,
 0.2739855889681404,0.9852439413173227,0.5374851114846535,0.9629455971551889,
 0.6564544699787732,0.8081664942004954,0.4143020207460545,0.6607669303375568,
 0.06464055306742977,0.4288588388545136,0.07829494139356874,
 0.5669521859720102,0.9813044199485332,0.1721317844837048,0.5999077658089037,
 0.5692830261597035,0.9914407137174515,0.08150888944091406,0.8837911153132336,
 0.03574606658283441,0.31908897901920363,0.715662815628772,
 0.39766819473269677,0.6873298712971462,0.5779571194731081,0.1751370953658039,
 0.030459852445725244,0.88097846586844,0.03594546564174872,0.7454470600450379,
 0.6252455190568421,0.8719736618681859,0.8694595746862066,
 0.014263049574837927,0.7969019788280166,0.636914567032883,
 0.03777693116096481,0.24198085412991344,0.22413128457968223,
 0.11261036712572858,0.10237174481255751,0.5865942538657006,
 0.7200602477465667,0.6553655715991368,0.25902035230709064,0.9484385154561494,
 0.9563537872020265,0.6016058863589004,0.4075554090796133,0.3918635736100431,
 0.5849128220576718,0.6191464940110918,0.06545507997070099,
 0.24382534350563978,0.8082795040559895,0.4476494801202827,
 0.10254274225336779,0.1718225651512869,0.695087980269916,0.9842791381332971,
 0.41582423755878917,0.11846370761132341,0.15795563568617177,
 0.017092387721518998,0.2875726280016292,0.9391113680436132,
 0.26092586128626394,0.9704999811538442,0.6752391102866543,0.5297470485449776,
 0.7742189909695145,0.4688532998565199,0.796475058102963,0.033334015143160145,
 0.5973533865079624,0.9314922048620734,0.7143639726829004,0.24952517653398343,
 0.6324714819169929,0.06288139494396805,0.41766729674753034,
 0.5145889444869397,0.10695367309369708,0.9751962489490705,
 0.29322421797716536,0.8434898131664373,0.7134435817769728,
 0.06018512576188695,0.9624154061094391,0.18192689315154442,
 0.16941364710345064,0.992090969684544,0.9456358733617634,0.7341992459122424,
 0.6208104390744118,0.26069655206951414,0.2493985040645793,0.3578924619315923,
 0.7428790269874329,0.44255848510269724,0.24944483064479583,
 0.38641412025109934,0.44385943806096195,0.5481424016960639,
 0.2103211189093308,0.14463425550646059,0.8453382296964811,0.8731385748781225,
 0.07959595526563679,0.46603399593010364,0.9904688476947189,
 0.4834710176771312,0.5210201367904509,0.5419068081798541,0.17833183973464306,
 0.5203220320039618,0.16687161821082996,0.7045297796556063,
 0.44499853636277364,0.9824001278363723,0.30466586777359017,
 0.6470707687401683,0.6537301718318587,0.7610123476075614,0.20011517360059328,
 0.41330123143959796,0.6737978452988334,0.8362874788128691,
 0.06902572695546261,0.03291313994536549,0.25443939288861106,
 0.21329417300154863,0.5683203727130485,0.5239466730606496,
 0.12462723594233671,0.23349605446469157,0.20634826773665804,
 0.29557250057956586,0.6101671743728861,0.8575211414026562,
 0.12561809674262392,0.7636447621634085,0.7293455626357469,
 0.024744113457943362,0.37840489286811807,0.09929539632017481,
 0.09980626354853361,0.10774612458902755,0.08928959389492064,
 0.8290627433301243,0.5887593883033058,0.6285809070355829,0.9492591921561243,
 0.33666987161311,0.3314823747992264,0.38448969138359956,0.048281707292697096,
 0.757205099416772,0.8783619554223763,0.20985064681325782,0.3096980055587969,
 0.5513596672513795,0.601699411529661,0.41844006441791104,0.6446370901540492,
 0.9812002136505421,0.565074998311235,0.5914291787129524,0.43768829816322974,
 0.9890466797511309,0.34700405175714033,0.7196073675073147,
 0.42812259686148335,0.43266033987013064,0.04818591482181833,
 0.6049450323687865,0.8260052383611391,0.8938723135491892,0.7935344257175475,
 0.8956418130910452,0.3390443966740436,0.6573444366633143,0.40453745208302827,
 0.9463208122301436,0.9814649193230639,0.19357534218606076,0.2294312547245766,
 0.8944790279487991,0.4539118188227802,0.2104333213693067,0.28814653595893613,
 0.7093861490427967,0.22131924673225878,0.12586386013401096,
 0.2721566423510098,0.9393921280653862,0.010232582832071402,
 0.7532168071354904,0.2542688128399998,0.37197050722345804,0.5922408131361949,
 0.014968650893348867,0.44510194974861683,0.4615260299970838,
 0.242986020574545,0.34299680305882685,0.7110436770047872,0.5288694929119816,
 0.06492892225874325,0.2986951606103394,0.41866789738453536,
 0.6425220759286093,0.9358246502590634,0.9654959495860175,0.5212156927311012,
 0.9944946435369234,0.07431319875980746,0.9082940794824681,0.9100168203571549,
 0.16907793438601937,0.6294885322775351,0.43361475672015637,
 0.48088201300115974,0.5720291243758121,0.5321237650766572,0.6107707519682974,
 0.6415239482544476,0.728066753524898,0.7917336634514605,0.8709582846226671,
 0.24893697802272619,0.4262382538560614,0.8595993199398041,0.1924192661613432,
 0.8751432859566931,0.15458734878907898,0.057100974916053926,
 0.772655661486369,0.8970263887171657,0.6801834682193878,0.5960515694489834,
 0.12202235120162441,0.9633360094134717,0.2704878472924186,
 0.03729518200294768,0.02798334802957836,0.07901446068975782,
 0.005812395597737607,0.15029637009726204,0.00122001347901457,
 0.9483075542531018,0.33260753094829176,0.32943674727133787,
 0.5589499776697017,0.34937766033933026,0.8217703125987659,0.3970220135092596,
 0.8927389666367486,0.59944239423462,0.9646474464091421,0.6658409436443462,
 0.815611756256847,0.3298301692937555,0.5974957256666704,0.7957739620340311,
 0.988814264452366,0.20896410536827492,0.5298054356503752,
 0.021460153405881255,0.4906920606318852,0.22200828506209613,
 0.6344876601237922,0.3035784620640335,0.3520941664464746,0.37300492957517084,
 0.605989287940296,0.859287591194226,0.5665539928496224,0.2935576970356901,
 0.9979200365314165,0.625499818228871,0.20874107738658898,0.8415562048218991,
 0.22587034884105817,0.1353139400186768,0.04070250835600142,
 0.2194494974830028,0.9527887039880065,0.5103249224153533,0.5682243735121397,
 0.4913167037075723,0.1969501664277361,0.41978248949367647,0.4078113433366767,
 0.7911424564148054,0.6589959743723863,0.5810688152960898,0.21408471081465452,
 0.5172783156018607,0.2866069951207716,0.12988123851790956,0.4248393015551507,
 0.09846276187890535,0.9680143915895383,0.705100622763919,0.574759031374501,
 0.08142019254503574,0.5285375045794863,0.5655725876249045,0.5872970568247113,
 0.35347922370737916,0.505735433259603,0.8555840704016996,0.5536447789144272,
 0.8618071267523834,0.1430553556746761,0.14414847043051104,
 0.27675159653816395,0.42006219387612886,0.7194102064599744,
 0.30478650418446707,0.15222483564065892,0.843478071344266,0.6132577735167951,
 0.6293013409627517,0.3629960041457413,0.019399697651459658,
 0.5763137982855294,0.009895948570010082,0.08595216984545495,
 0.3570154738540037,0.652008062690086,0.6505798194848011,0.01070490436368643,
 0.4486959001910992,0.03530500764976674,0.5652961779244196,0.8031113218161179,
 0.1129828129434699,0.8834996809743574,0.9507135271276608,0.32020880521321327,
 0.5178466839627711,0.5127249607237511,0.23435710284068156,0.0301424149045042,
 0.3943068572945232,0.3691982310568003,0.09087540229194069,
 0.11511205211037323,0.5253676402684633,0.5567316052485808,
 0.34452692368765114,0.08087299265295078,0.43822599355963987,
 0.9630107922605459,0.24634854205899592,0.18154825733953084,0.955218402916166,
 0.6297814805006678,0.8922773793617127,0.2445560731509987,0.272122852047448,
 0.7499476559304395,0.07754853926193594,0.748648439563061,0.7029076401999982,
 0.49685512194086423,0.7704726761593794,0.41595435471926545,
 0.6088474039641468,0.015743061560431926,0.1097183894104583,
 0.4507078273709533,0.8030376558106366,0.8043943415072259,0.11493259089316643,
 0.39786402536200494,0.27564961505678176,0.9534451234554853,
 0.24843739420744448,0.8153676124064322,0.25272128727460386,
 0.5107883460961937,0.42219264663960854,0.7355055991250485,
 0.39960153492383255,0.3539957749622189,0.11076637113718757,0.40397969945156,
 0.5536349555237897,0.19961892524497715,0.9297532023827982,
 0.15473440261332494,0.23626196739664296,0.04812952381987645,
 0.2989010618271073,0.27894336801398334,0.41735882364902355,
 0.6172758942404648,0.08376183301665363,0.4761239744876081,
 0.057129072330147634,0.7264539291494543,0.7354502767192076,
 0.9942301144234327,0.6716894899934069,0.980423754825643,0.7065162590175513,
 0.48897690123478144,0.1917809448036064,0.7216661835054702,0.4631153095905053,
 0.5720742475663367,0.02121889353913564,0.1725764979778709,0.8283955757841865,
 0.6175664981714866,0.049502771543808555,0.42622945002230916,
 0.5660790176833588,0.5590544599804103,0.5790468923698505,0.16121718667128604,
 0.7738614015278562,0.9909980064958727,0.5573217626505761,0.9073097270808687,
 0.5192239117897106,0.9428567469178541,0.0910361239800418,0.8652516863447187,
 0.28511996996260214,0.7132287487055468,0.41823163817811637,
 0.9692280178469206,0.8985372659684725,0.22108138123250864,0.1347079621712165,
 0.4377107553875951,0.8626815544366595,0.13484023945298917,0.3244083004074465,
 0.026808847338524955,0.5236766641506957,0.34437239974819356,
 0.9403551494200619,0.7679987673908881,0.18158817409265715,0.3737088652982782,
 0.43500717982937964,0.29958204020611434,0.6206164388151278,
 0.2653396646319225,0.9827941595875784,0.8621989339139352,0.4632671835707719,
 0.27857582672319026,0.3892273715000071,0.3880353698543547,0.5480163866460139,
 0.5886494046067414,0.11554731527300643,0.46960444338421115,
 0.5238364928634776,0.7347818661847497,0.8850644583969631,0.5478542976105195,
 0.9686239205067074,0.6326108369019482,0.996350619746073,0.9477937994981752,
 0.8666338769602864,0.3092218936564174,0.6504158231623418,0.1471444437684133,
 0.9862070959561318,0.8040247304450326,0.20613995774271987,
 0.35590319996036346,0.3783221405442033,0.07563400164183554,
 0.6842058717632494,0.7749903384963817,0.44358376234562436,
 0.06810929215931805,0.41363304924271604,0.5719317666278213,
 0.2290269607287143,0.3851232902766081,0.576127017669565,0.2693401419563325,
 0.3148109336247389,0.8013209936729091,0.3359943532298063,0.5959284450618688,
 0.08515560935690059,0.32684462984522167,0.05397904336313297,
 0.9500806933513414,0.9323514022397127,0.02905796301536301,0.7986112509513545,
 0.222956336252481,0.6851205428418329,0.8882303284732314,0.6689231929997701,
 0.2790314235698985,0.8498322805902088,0.727061193102772,0.4028794258829568,
 0.25254862542918943,0.532132657494321,0.3094000327191708,0.5195245589784107,
 0.6201982737642027,0.6561934246126803,0.49240559795204086,0.4943444274247455,
 0.9774687744733541,0.3639040141512472,0.02034239373241442,0.5141366265905254,
 0.1940938799907903,0.15494836822515623,0.6278932049302489,0.9752177879122798,
 0.5152809526657921,0.07923511461865773,0.3711192168146561,
 0.23759673167884965,0.3217487605432451,0.7273157058197764,
 0.16362334484155872,0.33751668026923876,0.18222818213934389,
 0.6345730713887529,0.463694385642754,0.9540421856817561,0.7424713941173369,
 0.5225972375965744,0.07443033301403523,0.8096056163994572,0.9561421886652148,
 0.06776749755787737,0.3528334609914816,0.5261679099135854,0.8583680608216954,
 0.07179770757291892,0.9109641792335997,0.8342084441865676,0.685467542259853,
 0.36678079775139893,0.5846994262133,0.09044821914802137,0.7549644450451335,
 0.1671346376379177,0.9251029938667802,0.47894787955620766,
 0.10764690022379142,0.5573969347306611,0.45814073110786047,
 0.6368107573121343,0.518951564135115,0.8049770529129215,0.4338215431641611,
 0.0037711011226707614,0.9171398414717026,0.24512841893957216,
 0.4301558170219978,0.3701798844585962,0.3932641512215914,0.4222472259109047,
 0.9933156961914689,0.44975547163553986,0.792139893273273,0.9024572913022405,
 0.6410585420023965,0.7445572715399496,0.03798275106093407,
 0.30766758518499093,0.4951975337957226,0.6307481876260559,0.5732618374126637,
 0.210691494506004,0.9586182985916885,0.375914316838173,0.020979689891470366,
 0.8510380030413456,0.6856858496085265,0.5518686268556587,0.7262669881684907,
 0.3088904639723601,0.1562077193175464,0.9200054345370511,0.5501226721699928,
 0.519009002428214,0.9639168954875901,0.7044267612497414,0.3178307331616588,
 0.06570956846646392,0.6099638574745055,0.6378794363239076,0.8734627817442663,
 0.6702162958397321,0.44752361483359904,0.3965205559584746,0.6548143719000936,
 0.7093223098596659,0.13078091014637894,0.11522572644367968,
 0.5389442606452745,0.06944337284072744,0.8723890648070761,0.8914240120777147,
 0.7197044804448711,0.8102510515856362,0.521149480796671,0.35781428574175855,
 0.6710520795157042,0.24616211993524162,0.7679087076306056,0.9518636674309772,
 0.2529259448792114,0.24313850086364985,0.5529303432377513,0.2085959858741414,
 0.14298225007232057,0.19046432581249306,0.31368009474699665,
 0.9115247464714417,0.6714533594968063,0.8994188661827754,0.8114128705364599,
 0.28297877136189353,0.0011833727096022262,0.7736870516582286,
 0.46015336445307964,0.9009765619429106,0.5707720048809038,
 0.03737872446114998,0.9430013548030567,0.776342404378841,0.2311407596825963,
 0.40844066649025645,0.35586793994490684,0.3758929540251258,0.633191039253569,
 0.6212126423711913,0.3280759042052317,0.7203417908591407,0.3241219339017817,
 0.3282480992932624,0.8801101626747936,0.7187091767052921,0.8977533557422457,
 0.9966527332258184,0.13632512688623766,0.11580666548742735,
 0.6617095222563956,0.4207829668119145,0.28129458583898637,0.7251067872460863,
 0.05080074800780565,0.22146024380899165,0.8978498643540095,
 0.14504953187590441,0.11394316831464846,0.11945935148817022,
 0.35317967784030335,0.7551089742380032,0.5468169831536133,0.3720526577512405,
 0.6220696438742457,0.31895561496698455,0.2593562529392589,0.1885958692289622,
 0.10784005191295831,0.6723278695062608,0.9965461914183613,0.4697961590078964,
 0.3135146950772443,0.6507762571715611,0.23669418956513788,
 0.09114438976193351,0.5509923629844206,0.29324415640138746,
 0.27555672092004657,0.23074554331756802,0.6462700848567223,
 0.6381580838954835,0.7538639338967563,0.2054994078219312,0.16708937773017607,
 0.7812041699931845,0.45432085127020694,0.756767855718828,0.866516855326739,
 0.04807114144824576,0.6506176017662876,0.5467969340305792,
 0.18356217593396962,0.7168737861082043,0.9474459520846488,0.4182467983261917,
 0.7294986122895736,0.33612102041673597,0.18439066173123242,
 0.37526687861278707,0.6061400089862432,0.11961368038346698,
 0.6258414686295192,0.7459664702152431,0.7458869649689828,0.8330391247103274,
 0.020198365662845053,0.49160410790956943,0.18068680125934844,
 0.6463392872925461,0.42448059908156477,0.5795046121701036,
 0.24308200790116452,0.3113902531404338,0.44055610536770995,
 0.3020884249960588,0.4392632757514232,0.8707984162728598,0.06879549089764958,
 0.8295589582132308,0.8631675066800657,0.17932379397713394,
 0.16336198949673786,0.1820890022921271,0.01675153557344622,
 0.2172135836461886,0.6575864180387784,0.6450470110513277,0.781552560092015,
 0.31434593059274085,0.983957753437138,0.07246272880022575,
 0.43396779081599957,0.7761171366554006,0.17276603088143172,
 0.5034849185575517,0.6207219415524914,0.7638962799991664,0.5614686660592723,
 0.6038542713620436,0.46775662372047233,0.705480449619642,0.9961419962994523,
 0.5238220000215944,0.8308158789853491,0.06914305520962039,
 0.26077576747099196,0.002688818504553492,0.3755261516984284,
 0.7522618776319466,0.12252752895247732,0.16581083741633318,
 0.6235733370856743,0.2959109621054198,0.9561154163275313,0.40166092139962206,
 0.9764750217795303,0.12265252191171117,0.54218624037607,0.6739442088445577,
 0.48229616423506527,0.24374933457212733,0.11066007197466787,
 0.6235444442661658,0.3282437832517131,0.15795978750783224,
 0.14553826398779812,0.044331471954155544,0.8665111426895397,
 0.8516021424722651,0.42551126502342607,0.5397657780596133,0.9749226423577316,
 0.16151281064895806,0.06353078624191566,0.10768212703313562,
 0.18005907516665887,0.9026871879659696,0.707750618555859,0.5106867510587937,
 0.9787653584385175,0.32680013132530494,0.4848324184421089,0.8762683842927744,
 0.16225628855821017,0.5938025223459182,0.4395697517108441,0.8261985106902361,
 0.8635488460304426,0.7916331869224281,0.4065837012224287,0.3749676183025258,
 0.7101513897017108,0.0817314184831982,0.1924045505356704,0.37254288666743707,
 0.1939019555494652,0.41155113241652597,0.48409975313989284,
 0.5633788943853939,0.08453667925693764,0.9331099727823504,0.1133389507039414,
 0.8055272150053228,0.18651306744665497,0.7042048452522288,0.9341648007707661,
 0.11240443707537118,0.6614816811817084,0.4087700228745377,
 0.29552039101734706,0.7146227943931684,0.841811842434498,0.06729729273497909,
 0.2553590396250559,0.9982883241237429,0.2608225651372317,0.9984065609529268,
 0.7301279289448266,0.27772994559385944,0.7977463176166392,
 0.28058268562583155,0.03309619724177504,0.7446562499112106,
 0.44811053174536253,0.49487419909837016,0.8448360138333056,
 0.7506850964235284,0.9678044458179644,0.7633585152520771,0.4154355181902455,
 0.23298518926128142,0.2697178376485334,0.10388750253025969,0.299287846389351,
 0.3968468817860804,0.056385032876604346,0.5060720954121216,
 0.9796649872960097,0.6167497591727299,0.5869818317081507,0.6872320294489527,
 0.9047923847804182,0.32378619424636423,0.1410034833618743,
 0.39509402177408703,0.551608024441369,0.36032735255460135,0.8953424881545637,
 0.966770378389869,0.3818192306446532,0.7329530257131557,0.823057827264554,
 0.1843213146869358,0.8016717745595696,0.26631234429561557,
 0.21638750859043487,0.2272913889577054,0.6277532031718221,0.4329348210317917,
 0.6557134905088754,0.7557783251184559,0.29895242908153186,0.8584071529904556,
 0.03678649284582802,0.025859152562012877,0.6858164240428863,
 0.7824454037602848,0.0017862532426909938,0.35436341108685565,
 0.34397274898066876,0.8349542857323775,0.43367548845966464,
 0.7099225357266602,0.7015760512652861,0.3645887872757946,0.427462737752496,
 0.6771737304309453,0.6620891394713545,0.017257711937825326,
 0.48662292866896484,0.10004186399490311,0.06578982790648269,
 0.5194694188696892,0.36269584614742745,0.22191526082277668,
 0.14742006846529498,0.2650063317874698,0.2826983466376307,
 0.31317599301631016,0.5379763934051417,0.5017887805818037,
 0.20792008874725998,0.18245430947994723,0.20166714098669758,
 0.35030053116507864,0.916333420882821,0.6717217601877479,0.7542530608400255,
 0.6470214781665827,0.6102579657494989,0.36204365588956255,0.6214213117027207,
 0.3046503638380469,0.576812551256616,0.05906869127677261,0.5939023318320625,
 0.8778662007850605,0.9009944221822922,0.07655481767170513,
 1.040331799011529e-4,0.6016465198790555,0.6442647939579355,
 0.2505546047027022,0.7853110206927438,0.03507213593138869,0.3075656042583492,
 0.5142227730995336,0.06513829646355374,0.2882204182059789,0.5010916191403609,
 0.7489136048293257,0.184192070983619,0.9809974297683461,0.5628381308881898,
 0.6688664388053432,0.1805768119307486,0.11339119724445779,0.6282120328450864,
 0.1951959330041272,0.1924031165786947,0.7379743394848519,0.7144195179175923,
 0.02422646790521754,0.7168721610217063,0.46880350767875956,
 0.27907017950688484,0.4199225519273938,0.07627997398917596,
 0.21055295999074009,0.8369852823336368,0.5915552103677979,
 0.40879223261449393,0.781879278963247,0.7817331303215898,0.5427356276798548,
 0.9063730173346243,0.2266178384546932,0.6895928149467325,0.45416130258729503,
 0.3300522997183307,0.5185636944774171,0.4173681373385315,0.3537704657875368,
 0.3324116300676102,0.6878018386240659,0.3545558999781303,0.3103856829625542,
 0.6809650890499683,0.7263626121962304,0.16474014790246738,0.9388244600027082,
 0.1561816208529473,0.4333533985852147,0.8232501616198277,0.1883232215034608,
 0.6704542969251727,0.5551940649378224,0.9967500223355721,0.7764110816267371,
 0.861692077669988,0.7407738127325635,0.9966119185805152,0.04466125100096707,
 0.07622149231900022,0.19767787008640747,0.7888025695498715,
 0.5355594124700492,0.3685206864846102,0.49445495611456547,0.8184199009005984,
 0.9028442111406613,0.20705195399213594,0.5373640154130982,0.3566491830545204,
 0.592201875404883,0.8917959786440779,0.3025548146076783,0.7227843685648324,
 0.8225370477221353,0.15233550804020557,0.055917972516996484,
 0.6237902564083968,0.6494260394226238,0.5288818269014607,0.8494920630032498,
 0.13180228878166667,0.8860984635148632,0.3630131573320513,0.8636543691187776,
 0.009021458831934948,0.7703509775789663,0.6612818580700578,
 0.03459151154151263,0.7399310734911916,0.008936355217244785,
 0.5947492613927942,0.9016073845174807,0.5028160401131032,0.09012627305232634,
 0.9290048718635296,0.4444783702132311,0.4585976429894878,0.8816642295232832,
 0.5956605714644762,0.6459951122633026,0.18883551849462976,0.3132654216269841,
 0.019310114450252547,0.04014535581091161,0.6305700055425892,
 0.8362397993060411,0.06313714585238361,0.5139731926212505,
 0.05727127389978026,0.8953882696086588,0.533462196428775,0.5681488726764775,
 0.2049445191564344,0.8914648210824364,0.9052066639009487,0.6725522332358165,
 0.5549249459074627,0.3941651553253862,0.7838908497208119,0.8765572294590867,
 0.6796940671915116,0.29266592750554676,0.463813432103819,0.5917120189634721,
 0.19371309715683094,0.8144490388652691,0.4017247110493868,0.9591096397914293,
 0.3355109546005284,0.4473741194619554,0.8027926543945689,0.9750880584479068,
 0.4896980645104413,0.950385750753489,0.6623028165252617,0.007152605292189862,
 0.16379767803323497,0.6580732918163883,0.24237454913406975,
 0.3930773237017484,0.3750587002405099,0.6499249115834894,0.7208458324283237,
 0.6575173383545738,0.49708122656295517,0.3772510526379107,
 0.12913254754865489,0.65088025083795,0.09808392432682422,0.5416649089358095,
 0.10384192521413138,0.3226801311347469,0.5376275665379592,0.9902208309830729,
 0.05259145710161861,0.3430465812164014,0.5623020133228736,
 0.12856286128387662,0.09951688850841633,0.09758671470426183,
 0.19140292467354436,0.0341227237619981,0.5159104612914767,
 0.23519642250246353,0.11562001035568792,0.5937895966971759,
 0.4664960198844593,0.09416863468656378,0.2702390360066691,0.9327251566945577,
 0.3778555480601754,0.4401133410043193,0.7488581801940395,0.811869835330795,
 0.07933498846751741,0.9566937754036746,0.605496876544057,0.28453314624753856,
 0.2430544415770003,0.3406341590332288,0.9063787583845743,0.9947099711297386,
 0.8311620219290088,0.3758007290747034,0.40625699677005767,
 0.16896302014074505,0.7366626381400624,0.9929540826808851,0.324601437175851,
 0.584180044771215,0.6347709127343995,0.5116868536763357,0.5512391810908339,
 0.47290014986193785,0.5976385822364148,0.7778219582761398,
 0.22402234271315835,0.6059329692692765,0.9587856309900938,0.9303194702083786,
 0.027211197877072157,0.05423086215811079,0.5221608795509842,
 0.42019351922798054,0.22875610232253418,0.8565486596176309,
 0.4855160210689671,0.9195618031364146,0.6803349618073808,0.552079318751691,
 0.05647978126323616,0.744946728510365,0.1532698904904497,0.5984319309479206,
 0.34024282920302606,0.1721819288776637,0.6807344274929723,0.7937656838525585,
 0.1430672482023737,0.6647956489876751,0.5328728532176051,0.6926611355746719,
 0.44469570603807995,0.7257624273959857,0.9170677531957039,
 0.35242294495979853,0.9250340942192323,0.885217368843131,0.48939163133377106,
 0.3664951024502354,0.04243722622528601,0.9482492177103911,
 0.04175327428478215,0.6122242965286002,0.842604948065506,0.9290001248066028,
 0.289693415338781,0.654510947343378,0.27196607785032045,0.3917871478659042,
 0.8523393719597006,0.02848595216329919,0.6273738314081684,0.978022707580956,
 0.7265585077041117,0.2191548884372848,0.23996639521215135,0.8222649218471879,
 0.880514277542501,0.4475587255891983,0.5104673548303831,0.8498791973138312,
 0.10967279570981858,0.4000366599147297,0.3883421447752843,0.5902106292260406,
 0.37273532169910817,0.2943441785855897,0.2385135444833476,0.9901038002496613,
 0.1671696765502071,0.9442343069889064,0.33058503862740496,0.7289944628138896,
 0.9908307740504043,0.15031176720839012,0.7622810177999202,0.7081305203403403,
 0.45422771546747087,0.5312550074477107,0.16229331001752545,
 0.11634019643612148,0.8518770720121288,0.8055023325327951,0.7136173386527292,
 0.31494692631527066,0.8464099327668213,0.32463987296697183,
 0.3039664809413577,0.4783811851492552,0.09219025902305822,0.8448051245943393,
 0.5562544930347479,0.9973805487348713,0.7293687032553079,0.3902035818463352,
 0.6101028853066875,0.626989238315623,0.6406248319056451,0.4398261762814658,
 0.1730546338632578,0.5664567763782032,0.5517211495902661,0.16578272584966003,
 0.8236870142326783,0.8760734842376244,0.6861800437956223,0.11642667204247537,
 0.48608887722018534,0.8783584560292893,0.518140150537541,0.33536325673178424,
 0.24726851165930674,0.5030345180725766,0.7054858964792259,
 0.03255097110877192,0.21236702097952687,0.9274418452829911,
 0.8328860230573232,0.5249588560634288,0.6506710430943519,0.09403194338393117,
 0.9747361065590137,0.24855529566935075,0.016376063587631773,
 0.17198759070533276,0.986769956737878,0.1657929385509317,0.5015326603523076,
 0.6394458029276667,0.5140507739574472,0.42925063845779066,
 0.35341313760868065,0.4897504569275707,0.18856708665477018,
 0.6398934522121152,0.3983363489640389,0.35398447182045123,0.5980214904732243,
 0.6050072251977091,0.1843077823324648,0.9270939893665883,0.9119149971050446,
 0.754810417043336,0.6463652634890696,0.6838871181474154,0.5703043208068883,
 0.927630544540655,0.11019449957533245,0.10924283990395756,0.6852636680819129,
 0.8316816127332932,0.6098786055366926,0.058843114105814553,
 0.8352749569608352,0.6051191197868897,0.1812557527153461,0.5102633088986397,
 0.3746562714261463,0.40308902290336457,0.8404574976756143,0.7454628753811848,
 0.7790577395448102,0.31491624319634204,0.42966093097668345,
 0.5459223615196671,0.46472760333839325,0.7973964455189232,0.3757462089627749,
 0.486722823214473,0.7137594848877679,0.6522823356474514,0.782455637181477,
 0.7198311205470538,0.47172230928368697,0.23338186625641844,
 0.7420939398713392,0.670323788693882,0.3392026745878193,0.36044071357935736,
 0.7865169126006061,0.3235359103072286,0.5531307600812031,0.33162465856146917,
 0.6704558478885333,0.40237482756315024,0.23734516692641172,
 0.4984022233255514,0.6507950352689111,0.44627891049201585,0.7135780487550873,
 0.1750113336702257,0.2797330618176187,0.9082789603537823,0.2013219192882888,
 0.9295811063745227,0.46044170784352534,0.4852432224889368,0.3671659471229056,
 0.19961838883135785,0.9378621259362908,0.24909399917341402,
 0.15247694681156532,0.6950778511161041,0.8391847507374901,0.3976097296962853,
 0.8875615181175402,0.5830953139895629,0.965411180480157,0.39425416727789064,
 0.39236028588871574,0.7548475828065837,0.2684074598508098,0.3453650872434392,
 0.9022356248229946,0.8119012738933988,0.8371099548531775,0.8263152966668851,
 0.779110294936154,0.7032980072634518,0.7583889326521623,0.528569918728073,
 0.41702183471956367,0.012759856374019551,0.8047984837261917,
 0.2409466045665386,0.01071900071271803,0.8975086700414322,0.86484296528323,
 0.2614570846612484,0.460289898885599,0.8522580123119935,0.6341508675615245,
 0.966987673155079,0.9527236352368816,0.6081867317038866,0.1251141946927492,
 0.8791034839566572,0.6573862892567218,0.33249688062956995,
 0.013536378126519644,0.8559145845577861,0.3313881942747152,
 0.17492343678853017,0.8063985415900553,0.39119743297194287,
 0.3228068321871066,0.5784910416481172,0.3413088012181067,0.43588010510766306,
 0.46266617886594497,0.23882010416067934,0.26650947536328873,
 0.8347332923465237,0.3553209610418009,0.8299160779066441,0.9611088109915475,
 0.9612090374585947,0.8043390385606105,0.0076413912995607625,
 0.10688550806416641,0.8408996907715403,0.8200944853862644,0.6220604756932281,
 0.09710887558111492,0.8664575144520289,0.8282817030010655,0.5840600542762204,
 0.7951736201766173,0.3259215673918552,0.21183600552761417,
 0.10511494413763889,0.9972035794301752,0.10146922735928088,
 0.2737565743296442,0.15041207769763065,0.5909333398809147,
 0.02100370086708092,0.8583755334773752,0.37349807334403895,
 0.21020505827255787,0.07256971516592525,0.24401920707962665,
 0.10630641079720216,0.5944262931951111,0.13426056564563982,
 0.0029168204017239185,0.24360466909758505,0.9909598879714286,
 0.6374939362923345,0.8859573089948398,0.32315712162928345,
 0.16798911731714516,0.9140901093577848,0.04138022302925992,
 0.3542524029102231,0.38709813546712235,0.7178499115516036,
 0.05277564508697785,0.3896032543085306,0.254720110052467,0.1539485518048478,
 0.4292981330291361,0.5256742852793704,0.2779466456767461,0.7656470101227748,
 0.6388687593378468,0.664347854625885,0.9326016447654636,0.007671557334937185,
 0.039624191590592384,0.8701685922186235,0.041400297161668154,
 0.9491172173921198,0.7749266352875704,0.17376054447990497,0.2837006866425926,
 0.36013256361281387,0.9142037499655178,0.4034246237950754,
 0.30833527489022816,0.9906426046725341,0.20536969177970876,
 0.8941777889569202,0.6962136139432278,0.17175277892139051,0.6086063161120319,
 0.05265057716583876,0.6047293111161364,0.33999277648452575,
 0.20425754823657083,0.545196055019481,0.4110088839749606,0.7376989409238489,
 0.45627592806111983,0.13019761324628554,0.8023821282490325,0.878899411401387,
 0.393791185210995,0.16020083867757662,0.14363145721634318,0.3929026183985881,
 0.4478562245705693,0.3785971167587806,0.9424617212621378,0.5823456322780706,
 0.4317918683411528,0.7287943855741553,0.7984436162339871,0.10523579994555654,
 0.9963467840012415,0.29705317125005604,0.6470586874286033,
 0.21312630801450994,0.946521707598345,0.3264478316863618,0.19656200284963254,
 0.6950385503919265,0.8431309595245192,0.6526470138682235,0.0690645630127833,
 0.8129708919193481,0.14259183550256183,0.6599485580585662,0.8186169954656598,
 0.23379162093256367,0.9088791919612046,0.7832489069319728,0.8482030064056516,
 0.6857915841169533,0.1390887529341719,0.6148956237467302,0.8004199932364092,
 0.9584364852824916,0.3023884678387909,0.45039433321713573,0.5411991052912507,
 0.06705844988751408,0.6135674556936687,0.6489716711877294,0.7161671825050961,
 0.44622889584406755,0.19544018334080415,0.5176518643894884,
 0.8829985485215128,0.8623643170068983,0.6807664693718201,0.3144772655845225,
 0.29488382231268195,0.2878422234445004,0.1107948000949962,0.263593113634914,
 0.4351495544615269,0.7526225447691294,0.07740912730259264,0.7370641340746537,
 0.5960431864300018,0.8543830779030108,0.27745967076551215,0.7418313059651999,
 0.9547077913455009,0.34747961980700837,0.18943316840069424,
 0.9055981278523041,0.42420822238088474,0.20800656114703708,
 0.1057144624976587,0.5272695763455962,0.12374598983638596,0.780315410622164,
 0.07701714121410397,0.2540135785316059,0.5194261674748276,
 0.26346829282114737,0.5144564958514752,0.2852110009511515,0.8102321632983531,
 0.5916425307505764,0.4206123142100686,0.16740085016565465,
 0.016644686718903068,0.9159250608617049,0.8036526460099965,
 0.7336351592589481,0.11063239312861561,0.5666363861085627,0.8545963164961612,
 0.5957091497474507,0.38874711346624236,0.8942059290043609,0.7523942264171115,
 0.23559399301293027,0.89563087583781,0.9478362273300683,0.8914319232757156,
 0.6253068652484717,0.2744457439174375,0.9383579775403154,0.45324218420732754,
 0.35371544024209156,0.465332914246099,0.020054523496525256,0.801183965126238,
 0.358047195122178,0.9694806554257278,0.91384818735201,0.7516904721381364,
 0.5356674638284976,0.2798034696224623,0.44112061077124987,0.9459443052591172,
 0.800312667074044,0.42763868081275536,0.21151611709256357,
 0.013977675731724815,0.42670954763837643,0.7628822973437859,
 0.44018936888899096,0.9219446769119829,0.9026185878289738,0.7199924607599609,
 0.173519121340873,0.6171977459697122,0.18931364750191748,0.5875313407207606,
 0.8854912338565207,0.5450889304488927,0.3340089463710426,0.643873481393057,
 0.26296390668453296,0.40413545773184767,0.3975363426944184,
 0.6173743584032092,0.05674619200525188,0.6142542564654649,
 0.04727480593119149,0.45524540951507275,0.5090789897980788,
 0.42770438616408657,0.6581038458488628,0.21944454065240537,
 0.3798337093781774,0.15373476886822568,0.01688186330007968,
 0.8955135567966461,0.7022922491517819,0.9051535205024014,0.9161378064518153,
 0.9440153156119601,0.9102834957928067,0.37570405907091875,0.6568015910761625,
 0.6553426931435512,0.7181088667990785,0.9563639049148057,0.7545427347111866,
 0.03915944747682487,0.3128843890627486,0.23806284629507113,
 0.3927921777000001,0.3911912319280251,0.4377156888174588,0.38129961749069563,
 0.8585805244249872,0.31733475009571344,0.40376234479239037,
 0.9925178872861857,0.520529133726023,0.6280364737317271,0.8569917665595373,
 0.7303424483755038,0.1794249882485074,0.22518337193176363,
 0.14585682708947767,0.22016902391476045,0.2261211047281747,
 0.8852746156266634,0.8335402541180011,0.467239470055273,0.6938135929563336,
 0.1807683541800169,0.2008833285143602,0.9447028900240663,0.12360462939048678,
 0.15044786697203416,0.572365007481197,0.15334946354431533,0.4497181270995446,
 0.6761378770859182,0.32112750308984084,0.20725171969901401,0.727595044228209,
 0.6687504413819357,0.4791011863307201,0.19189785357421452,0.6291082849199998,
 0.02091428735800127,0.7088162636823306,0.6278097212787215,0.341264409869225,
 0.9943064749439966,0.60409103649326,0.3226147765513748,0.8048964219629973,
 0.37556912181439184,0.28130530358235906,0.32819361356915266,
 0.7751558797596743,0.7847735750208138,0.07400888413206219,0.6720891353152354,
 0.4950132523022519,0.7320030709868066,0.6001918136040235,0.34481336746575875,
 0.8615977636207117,0.7064311902504349,0.9371989596846595,0.2348745467401685,
 0.9356192196563395,0.32416808758274585,0.8782753645647746,0.6154299536286967,
 0.9448855607075578,0.16698192093310116,0.021285167601992616,
 0.6663237751894286,0.02200007903691792,0.5174990235827854,
 0.06013402039929949,0.058035199559437034,0.7854019095629038,
 0.7711588509998216,0.3147662687642384,0.29931063838424665,
 0.13617538368608006,0.8889903940195126,0.963560584647179,0.8574319276293608,
 0.7140297203463974,0.9013258046037811,0.22591527517046317,
 0.32440529381060446,0.2168673679937443,0.9789012247014098,0.9607964945908488,
 0.6092950165851522,0.7751578719088769,0.6963066522490731,0.5666303674670137,
 0.40113352911794986,0.97768582146616,0.41761445055039115,0.45682982393091376,
 0.47421807396473037,0.343999334114107,0.04131229059967956,0.5854422518621576,
 0.8360727718894523,0.18335939150411906,0.823402992157063,0.42073383872580983,
 0.0747498840705163,0.429335052336832,0.3357369426903828,0.6885244137902969,
 0.5529249973774469,0.3831778902158063,0.06746466693076847,0.1941254484758893,
 0.4682679056291912,0.6797114872304867,0.383403966049051,0.9404094679348323,
 0.145610873644773,0.7954747012193227,0.5146186491663381,0.8191057931676407,
 0.14117928964332116,0.7240922780058088,0.6102327100824386,0.3076867406310334,
 0.6446913598023029,0.21526908214153717,0.3715611019416035,0.8942309822162936,
 0.37267782543894934,0.4973265086197304,0.5537439371075977,0.3092274897395507,
 0.7108956252927284,0.10704091054459852,0.7767107648736729,0.8963800321071386,
 0.3654339054939526,0.7411952483734913,0.12314561953140335,
 0.07050739359743363,0.42210472318706715,0.8374764297576223,
 0.2987059344927613,0.7492105392953354,0.7767397358404474,0.33011240044284285,
 0.06019008266710102,0.9831553931177466,0.5387967598358707,
 0.49793246837288874,0.16313220136628648,0.3824442435764458,
 0.5747971957141242,0.24284963320682273,0.4823829210790781,0.9426775085692444,
 0.6550285080537456,0.08870801600153555,0.2151598260406551,0.630546221469398,
 0.3258504983921847,0.6351571657917571,0.5232570962926534,0.8595343210147184,
 0.44992189503993585,0.8029188692283349,0.8185170507535012,0.757856139139477,
 0.2566292502697446,0.14861800048449092,0.4906288833393856,0.4019954470994579,
 0.9482181339823778,0.7719614648442115,0.8334268361297553,0.2181482229566445,
 0.9136543470406843,0.13978517362434917,0.46331406363751737,
 0.9854796548681289,0.8749827220074033,0.312519042197006,0.8459192201393352,
 0.3531656920184222,0.7614554810942645,0.1627882508583981,0.26555621466101076,
 0.4957981510193282,0.06974087237894999,0.3475448481113521,0.8470271147472926,
 0.4517920031816546,0.8967244129705185,0.12378931657639358,0.7373925811335426,
 0.3992272601313748,0.45807764514697213,0.9145637793666279,0.5090959255336508,
 0.7259432535578227,0.6027594274183294,0.3560973289638957,0.9662119951213611,
 0.9344178186771018,0.244520586578011,0.7972327550176633,0.2957175411669336,
 0.4253268091542305,0.5428611662812426,0.3192493996994199,0.9251831765047941,
 0.8553776819861074,0.4707590091169159,0.8617815429008694,0.8369933464051501,
 0.09179881497045395,0.9200791623934685,0.3358546497382725,
 0.04249497286068116,0.12564486573493283,0.6173275657900521,0.520564313680187,
 0.3448548426099004,0.5583993635957237,0.6957078209309806,0.2310662851803782,
 0.36780263500044574,0.3852169219014411,0.33435413635663136,
 0.25088927676353434,0.970719478053707,0.7347101403073185,0.790866871469553,
 0.023597799180970713,0.14398121104204153,0.5847734480032,0.24188384376380867,
 0.92673285669982,0.5059221652474933,0.5198681836884212,0.36003692764152107,
 0.46385308988171114,0.597830236901695,0.8617353715183895,0.23953638176615677,
 0.24269372330461558,0.24731868371505739,0.5982320092409885,
 0.43914358992006874,0.4443886415807736,0.8617947234868387,
 0.49809991498203265,0.54918266419277,0.28420568067747753,0.5452107514088366,
 0.791823081781324,0.6962230081576467,0.7900759050914531,0.3426251822202758,
 0.09777029451060348,0.8970503154361321,0.05010790199032189,
 0.7441258647627024,0.901731096419081,0.6322020215033317,0.5222463028923625,
 0.9348857217054687,0.6585939586110434,0.06620797952359714,0.7161132592856648,
 0.7386844919086569,0.6473121995608883,0.9645069344883523,0.08282559670717926,
 0.485076452278585,0.07720232022231055,0.916043661350625,0.18207044392062877,
 0.7568086428935826,0.7778534195695386,0.3353158363829619,0.726884896834721,
 0.348883786546371,0.5688644114004335,0.4033215102616028,0.46897787839871596,
 0.6376637763481987,0.9324712463647534,0.23707244607873323,
 0.007229902972479718,0.10429839674207309,0.7223053522881947,
 0.03212089072406532,0.8216778096916648,0.9150455700810193,0.9497588714620399,
 0.07499906219554331,0.05422539366259016,0.6126603235331214,
 0.20246652804930543,0.6205499310587563,0.31765595674360236,
 0.9776428936120116,0.5602721807606672,0.726997188255789,0.1757601752304272,
 0.31364149406894515,0.5714248480739593,0.981723600400337,0.774744054751366,
 0.20242501386302025,0.02185907747727467,0.165263513286517,0.7700312470947621,
 0.31862272753977305,0.07453475991197878,0.48575344755055405,
 0.985027606241201,0.8221688496651237,0.2660598071471386,0.0766621238299462,
 0.3225792098158151,0.8196678808633897,0.850758341249654,0.3050921601294553,
 0.5018363490983023,0.8948102563394537,0.5589227876760491,0.9068220179694345,
 0.5112556094977732,0.9769578111620086,0.4416420497802229,0.782863337642531,
 0.4807741559917805,0.823911404975445,0.5142174224042179,0.5622648480519004,
 0.9286704159193501,0.43478870588099705,0.8547308557709822,0.9587893220731448,
 0.9851039329292608,0.26454138779651304,0.8122964584143347,0.8883744014407524,
 0.9822408421073924,0.3097905988648657,0.390326511329,0.8281446023265109,
 0.9535344087663996,0.7170841435625686,0.9125186387642411,
 0.026940821461033937,0.3925487064923028,0.024038114470726124,
 0.22150656323002849,0.901783257806016,0.8256184318624596,0.9532019024330588,
 0.21054864586134492,0.5404013869612453,0.6308021120084228,
 0.25500902057026453,0.005291889113339998,0.5211064287368984,
 0.05871881781192012,0.4939103153226112,0.9519132390763446,0.9973475104342489,
 0.8720904819421923,0.9113349577879586,0.6784920917336191,0.8104804035510774,
 0.9296844152920168,0.9585024146200282,0.025233334148741182,
 0.005558423837947135,0.0945385707298112,0.574172426788674,
 0.016684126638116803,0.4447971817953046,0.637307296987376,0.5415838933945237,
 0.20122836237737884,0.8622398135497195,0.3951271183544941,0.4010091067836966,
 0.9543727562054459,0.3745107589182868,0.9135466320426561,0.760306773622703,
 0.8636348844956833,0.23807504658387446,0.7800251775434927,0.9145876903894271,
 0.19310953502464656,0.46261008401137416,0.5645068130195072,
 0.6534585125569897,0.9852912676977146,0.0893066662239217,0.7101793136244378,
 0.049330346575670925,0.1522855869443297,0.43317750216252804,
 0.42565760262673014,0.4149850671457711,0.12349408903236148,
 0.7832353579244185,0.30998914215767637,0.7227252227234859,0.349395779327764,
 0.7483159549948042,0.8056065176473612,0.41414574832964746,0.5723629720998341,
 0.6257296329125203,0.9099611140565425,0.19046334115158403,
 0.017333620412556372,0.18381083180470936,0.032988657500281016,
 0.4794058942465055,0.050645471422084976,0.8646797272595906,
 0.4375165055211768,0.7596782578401404,0.8257192139090694,0.6138699688414522,
 0.6162737646250962,0.05868557362777238,0.6794444736525335,0.1769116684646902,
 0.9331934837226492,0.09616754867135469,0.20304759949672602,
 0.16853198370429268,0.554656986357548,0.7077006332720505,0.16035279915325273,
 0.36375112728911196,0.27214218648759836,0.513980486252241,0.6973593539585922,
 0.7847693398438157,0.9191233474511302,0.9946577499486144,0.26955860446423086,
 0.30971813067456955,0.17533248840414906,0.2910142128562744,
 0.040041137802463544,0.5814916604315419,0.265314975285197,0.7112645507863193,
 0.7858284085613964,0.6542541084800771,0.4748495762234095,0.9615612260595453,
 0.3772164134655491,0.6518684320488227,0.28229899096036304,0.4203826435473579,
 0.09481226641581053,0.6877655265268171,0.8436587321519062,
 0.07351889185351235,0.7376109554608083,0.4555151506069923,0.7622189619004032,
 0.25624295537927744,0.32584722988424153,0.8164520793951787,
 0.31771327057675747,0.38567969964542215,0.4206771149210562,
 0.13171068075980352,0.42346388810880264,0.5643931115763062,
 0.5727216783558722,0.31746429008122423,0.357642366345425,0.8588535500400251,
 0.473965582003698,0.07696053622003896,0.44787692039517935,0.6048458242261074,
 0.877403354143246,0.2105346703690828,0.2707353269895566,0.6395899788666087,
 0.3694461379237568,0.318418734039121,0.20640932710702609,0.29750940481917065,
 0.3992935178065399,0.032889489983180975,0.4476063060028762,
 0.5068784882622948,0.07489040468591379,0.17287848386167526,
 0.3803971061315363,0.7847373667186874,0.004507810715928651,0.192963352688307,
 0.09744985009558293,0.9042821790278414,0.6950610057077096,0.9760729176093785,
 0.7712014846145386,0.14579413539976094,0.01044218648075601,0.875388076388289,
 0.7870938602318877,0.9514583020650462,0.6295937963149133,0.2495083435401767,
 0.5076557220499748,0.9435221957716811,0.7832241944764471,0.10827720040207722,
 0.6672520220393496,0.25159093922671216,0.8464875591743424,0.603242346709622,
 0.4279072553202603,0.45531575721469386,0.04517659926924111,
 0.4972857243866625,0.11802806328106763,0.14329400911028867,
 0.017676825455890555,0.1006520284774618,0.8187879629515071,
 0.3981286607603203,0.5332176893712102,0.5504843825448998,0.39411407724107084,
 0.747411487675636,0.24217067425995364,0.1266354640491979,0.18823212216596485,
 0.18876565524001243,0.613744869032015,0.01851747534349857,
 0.46632456196120553,0.37167213952488387,0.06726895784769549,
 0.04037273007467368,0.8589937993362855,0.3585206669363974,
 0.06905896541056722,0.0346705523962052,0.3261910881796303,0.951033005126816,
 0.6295618139434377,0.39855811003637553,0.4500874646519507,0.6922499931266011,
 0.8362383939261697,0.985452742084874,0.7039836300111634,0.8587656615207387,
 0.15174104656803156,0.5143460290179359,0.9146568477908268,
 0.09265923654289976,0.2751222350094261,0.8223827989358019,0.7814436746971504,
 0.1511364405441088,0.036165554317901694,0.5323942334304499,
 0.5494996190792697,0.5503058539007064,0.7754677565106353,0.3369862715132368,
 0.03013614661626729,0.46465907479919233,0.8301320209027445,
 0.3942165068740624,0.05542934283988088,0.13659638349315428,
 0.2818866778008612,0.5686842458763345,0.16848364952853379,
 0.42445154239106275,0.5307907698066318,0.5224096306407967,0.8476838137449354,
 0.9506518769890062,0.9234953894773875,0.8202289157779912,0.9699463849647272,
 0.23157610978706333,0.4467586605982391,0.8216695819870212,0.9990460849297709,
 0.0774184309556214,0.9387297186553742,0.2656743200081835,0.7932736747770169,
 0.22595593310212392,0.03377305595145652,0.8875431610662363,
 0.25100925536213303,0.4180578912558349,0.7858980982821542,
 0.04739522900564841,0.7615459540245545,0.00850626755515238,0.94863113115893,
 0.6170933009350433,0.34354527128776025,0.7163200711331231,
 0.38893374252643476,0.29818025109640234,0.8667506660998513,
 0.6612391687610236,0.880183025673571,0.6337027917530113,0.5997112990943533,
 0.11613523132306358,0.971454319783934,0.040497723113045225,
 0.014466292059504782,0.5309587347135581,0.1733382807718553,
 0.3139452166519774,0.42483321210853364,0.6067623349983257,
 0.08260675313782928,0.047170284168784216,0.4280502387043381,
 0.11477228300818976,0.2931402369063132,0.5794114936132371,
 0.16257849698545523,0.6823636381328962,0.23642952034020304,
 0.8729871967316255,0.5110889408142506,0.4233047231244852,0.2742679100410893,
 0.10339824749355109,0.3534137605859815,0.11361432562453,0.20516451927068236,
 0.08330504419081597,0.331990551695415,0.017466943653632994,
 0.1039785795720648,0.6611584057528382,0.2941180805713848,0.1966753059221974,
 0.7829476915911888,0.7318256029148908,0.2592179811049544,0.3911737412225088,
 0.261007995912205,0.9317642524390223,0.6676418521674023,0.21004772867968424,
 0.052567078320165384,0.13511509045165515,0.4360644998389289,
 0.5343793950992841,0.22340731190559115,0.42209453233773375,
 0.48320201885114333,0.5133484949950543,0.6921322655783249,0.146205536028143,
 0.9855620150137951,0.38343905993297045,0.22014095936728528,
 0.04422300595630679,0.9207216515594152,0.8643666620381993,0.8326895148306256,
 0.2591003525276311,0.6554309861395876,0.2745534263116456,0.9337516557409024,
 0.27765734196595715,0.9832815497231391,0.8482337973425238,0.523785359141642,
 0.6855459704560196,0.2717937595468487,0.2784048353682922,0.7190976061994948,
 0.6560302393354461,0.6751242102502226,0.6173181519826284,0.4866022867142038,
 0.6634566655282991,0.3362993422653411,0.7485316661492518,0.9371697452084409,
 0.32421195302235073,0.030072636485364734,0.6656105881938891,
 0.1944270334486109,0.8129745009734445,0.5313058113488984,0.18289623835339452,
 0.015111309834965003,0.8489936507791511,0.8268246141575574,
 0.5280971763724416,0.9903567324234513,0.471366983361617,0.12299595744016201,
 0.8659160863575167,0.33346357710093155,0.9659938477038038,0.2594452137405461,
 0.3220485480649228,0.5303399191855356,0.12152790292684901,0.9996964771821263,
 0.40154187813159203,0.358670488252407,0.4525339272218971,0.23525101602732823,
 0.358877201641181,0.8131225138469774,0.8543547279289997,0.03941444616029055,
 0.794073031590879,0.46605705226976524,0.9430541228432895,0.9656833218104582,
 0.992901985993348,0.7843621006424495,0.13798186160210069,0.43627999934092987,
 0.7205338901061695,0.009408174284953441,0.4066898269004494,
 0.23398575824916057,0.42084952332508285,0.5211171292132619,
 0.0026908170905864948,0.050870495157027795,0.39698512337675684,
 0.14304321527530028,0.40436345992808764,0.1437380975087834,
 0.2600765911902949,0.9611034367047678,0.6097598602759768,0.04074894798853235,
 0.16197203369292035,0.25670119491599697,0.4707755772819071,
 0.5611508088363869,0.17672689947749398,0.3556506159098819,0.7912275791846295,
 0.10492893557856675,0.20166479741062815,0.6899310480635724,
 0.4768921693203514,0.02911506448171719,0.2292482696829896,0.9150436417267971,
 0.8880042338306592,0.4241882303780783,0.322074504898637,0.23919324748941795,
 0.4253297262629522,0.8333543076414134,0.42853171595715955,
 0.26250564945828403,0.9816989803909342,0.6827032548528191,0.5506327216228438,
 0.6314352150652321,0.5933052767666841,0.05679119237370722,
 0.07182716310004789,0.20201054410460062,0.4795062896745401,
 0.8393248150497349,0.6522061507720898,0.9181835954859752,0.5660267801814453,
 0.11293997778377562,0.5012301485936637,0.16318518207979815,
 0.5849929171704115,0.39734761248874806,0.19570311610809188,0.608692068701767,
 0.3337424490566079,0.2641278842881999,0.8013292100073589,
 0.006889185636346995,0.6092236249600604,0.42314888239073767,
 0.05772161996519909,0.1130487226307173,0.3095765952777523,0.2839217347018397,
 0.1107662674665103,0.8149420960649894,0.8519366732054545,0.1328531783107948,
 0.8712708607135806,0.5939841255645334,0.6870742684675027,0.1100744357641451,
 0.44901348409871367,0.648733047909522,0.9467800935754118,0.8003111105778006,
 0.42573285887274637,0.813179961956754,0.9247734177488007,0.4887103993717137,
 0.9448461224862751,0.3320293161930886,0.2859718854845812,0.7397434624679329,
 0.13089696816437468,0.552917137433825,0.0773714033703028,0.697114570898037,
 0.00477794339465476,0.8591029666545121,0.3009353117343201,
 0.07518414451297972,0.254255942863982,0.47575090896878014,0.8753798082973789,
 0.7560246541646752,0.5174999379718738,0.27371291382015706,0.2550167759253099,
 0.09008344697959869,0.29038431733837233,0.2147967263886661,
 0.2687564260554286,0.6816436983047519,0.08145142414112916,0.6828828299563338,
 0.5554513388178088,0.5429598111167875,0.2590281209179104,0.06010937766427782,
 0.08485557709662772,0.005279902208827791,0.585457350020688,
 0.3455590712937997,0.35741826695664136,0.3574439858521503,0.2623112322810228,
 0.23626985823103575,0.26891421241236513,0.7316914986962375,
 0.5740561898343347,0.4319227973141848,0.4872234537482646,
 0.019070829888750862,0.14245535007418186,0.1277384467774354,
 0.14622672021368588,0.9462798059327295,0.20058189281626437,
 0.07568201153664345,0.6261771099533776,0.08873000919383112,
 0.032021136951595475,0.9465079574960293,0.7411771034377189,
 0.38190728361475157,0.2137373635713361,0.40376428952642907,
 0.04629106479822753,0.26345207670346016,0.23088341209808083,
 0.7944213488992323,0.38670113554728414,0.5411381589145582,
 0.19204500909633737,0.025003996086079505,0.537049949132545,
 0.9881319319687165,0.9041646015488255,0.22602182314118668,0.3425591442340781,
 0.3336635467962653,0.37675897366068667,0.40965198878307074,
 0.13961071175779927,0.03229526268029659,0.963725042848439,0.7633008743983801,
 0.9085158101326392,0.8250543993888199,0.5546633534540857,
 0.005143484102284379,0.4278892724792023,0.6308688662833575,
 0.1019507318294891,0.4041993848934279,0.2535052691313383,0.8793416410954515,
 0.38364386299174147,0.1703906370968985,0.45531308300773454,
 0.4034145315772245,0.509839748333221,0.8294616043324458,0.5090732020570553,
 0.45801287059562235,0.9431379369769528,0.8934576270082701,
 0.052997475355507806,0.14807571498370842,0.8006649188111659,
 0.3863524094649631,0.9677926832669851,0.8498293637990741,0.5828710972753471,
 0.7712028396686723,0.669364997372114,0.6377285647699625,0.2944056821754365,
 0.542516998549143,0.5558137174029121,0.23856864292848223,0.582037461174008,
 0.7244028529536384,0.5900858733326205,0.4163415791521191,0.3489329090110397,
 0.657225790232629,0.3191059580789267,0.5773361882446061,0.1623956187978367,
 0.5703575310505792,0.23632613491258447,0.9584233260818821,
 0.34778722848029187,0.962171994284424,0.9162802396850928,0.13241040649122693,
 0.2899174034149463,0.9279151464961425,0.08842743328948166,0.813432153439233,
 0.7945242313634054,0.5812455276462789,0.275876022703591,0.6192059012232172,
 0.9774932379779889,0.6204614528353384,0.6592302308458146,0.4444932233205928,
 0.38008148681589426,0.9318803556503411,0.4018001296118826,0.7909640764247974,
 0.9456933366045511,0.5282741501778149,0.1180753278619775,0.20786113282973584,
 0.18135795063355398,0.3192991506672397,0.6576756125301952,0.9050908376825224,
 0.46755521206592277,0.586094255494371,0.15846739716098102,
 0.09195995158433956,0.5826134339539981,0.638022425774682,0.7911737050126907,
 0.8381145094352395,0.021597777040040955,0.5928422250786516,
 0.4495535813417404,0.5557610946605748,0.02451803061206692,0.7335591278865394,
 0.13766703895967458,0.5783083863950584,0.015882427331064486,
 0.6981875407362468,0.2485672464072396,0.8586754524483856,0.8095494518748827,
 0.6627893395218661,0.34798757103050637,0.9146866371977403,
 0.33722363662187504,0.09857351602608433,0.704521001379507,0.9326470537952717,
 0.7061826615315141,0.33425003837837963,0.9733825739612982,
 0.19243977332869566,0.6891882431322974,0.23716225291382187,
 0.8565334780577889,0.55414720060818,0.7099875037987144,0.573683517156953,
 0.7961215053717409,0.8195080355935779,0.9192881161746832,0.4633977455138051,
 0.20184676386226674,0.857123709896488,0.6307165071280277,0.3368234885332094,
 0.05827822654931547,0.42012624842434454,0.11705811829485335,
 0.9405447671643546,0.4316797992633826,0.2179402814070921,0.8321262779433112,
 0.6212828574773235,0.045816817843964186,0.9726386843704874,
 0.6019519720187657,0.1555660468580532,0.6456465092801877,0.7283212858649696,
 0.4265984502370824,0.5197843238987415,0.30836845300562277,0.740487975906236,
 0.12342292091135909,0.16751193218261118,0.8025467238605739,
 0.3241579629776935,0.21002867415118742,0.7233792631097165,
 0.47973639760083686,0.29046787454830914,0.46608576752478925,
 0.018923679987144726,0.8102334311320871,0.5094361932931226,
 0.4782200666725115,0.49164291137266125,0.6833303795297943,
 0.16867350783926738,0.03545163240576965,0.7675793484181838,
 0.8151921662401982,0.09975485686002117,0.4403766878446098,0.2621314668574224,
 0.5301339163460675,0.019526162761761334,0.12033692232728599,
 0.5708984916657567,0.4841971108946036,0.4535115231207373,0.4408286572486071,
 0.21886897665390292,0.07406714305199424,0.3463058959069044,
 0.4807531241673384,0.010210514209481714,0.01182732339263215,
 0.12524307304918647,0.8554963536670147,0.048502170682381496,
 0.7593945526183326,0.7338158568361016,0.6788151997990174,0.7192895044085121,
 0.5841588200867204,0.33535057749753017,0.04581173436866148,
 0.7111397141745269,0.5921290742005135,0.9740988315933357,0.8724331892243625,
 0.899532978667279,0.8292710043957919,0.42179937814007384,0.29553561463637135,
 0.8578969627518851,0.9415462323617809,0.6598947127011024,0.10347635409707157,
 0.4619810291770461,0.9470590119475113,0.9573344779297195,0.7506110767924763,
 0.6098067368355624,0.08712543307688536,0.8875850639515297,
 0.17663982738938222,0.2186725657465982,0.6450538225302007,0.4399369371655921,
 0.613923862137145,0.4159789071211619,0.2938505473656491,0.22211200153638933,
 0.5212839938796214,0.8134102580352022,0.705148667850529,0.9994528877618511,
 0.9482100070418195,0.6832041431721929,0.6715051954461334,0.27717806085994956,
 0.6253882997341703,0.6394040168274011,0.8466413172387757,0.8955251884369393,
 0.10768363370642009,0.6795425496005529,0.4112526761672759,0.56292616604548,
 0.6995870217811319,0.6582754915926344,0.6856599282437182,0.6297886848387627,
 0.8005380654534836,0.6783903978082707,0.8351390033973978,0.8717628880792623,
 0.6188428510886128,0.5899971507599667,0.8242458573570383,0.287946977013008,
 0.6950494976276322,0.18817685554613728,0.6909403688983585,
 0.22548038186487607,0.1231206729062061,0.6648748311556733,
 0.22770675780457916,0.804900991418699,0.6167386483224233,0.168900755978854,
 0.02659944569357353,0.16079247027654364,0.03118198791865856,
 0.6413431539078542,0.5798619561440295,0.2698188886121373,0.2440537316290846,
 0.10008806373107881,0.41209131868456184,0.10590305607523831,
 0.305455973856247,0.5418477368933025,0.5762809443093163,0.5349821698759551,
 0.16842894537377728,0.5238204447096904,0.7379807523334194,0.3183425368611559,
 0.09003011533672789,0.14083218326083502,0.35546288395523873,
 0.7404997938979423,0.9297449092425077,0.5675023026218263,0.6032887558570753,
 0.48716846093133903,0.5003197907411085,0.5221958305763897,0.8036410422284986,
 0.9609904840888416,0.03484346272787742,0.6119276558559411,0.7668960312565314,
 0.2037437805197151,0.2028088609428832,0.8202905379338017,0.2523957633251682,
 0.6941776882111825,0.61286578603858,0.3197744030413404,0.8747511776248791,
 0.9705016892435415,0.8596896641713039,0.840183303453392,0.011830220789767143,
 0.9933520075908221,0.33542645179224917,0.4919839233287099,
 0.30968318134991646,0.9394044665594148,0.8555456155611374,0.8295657263361906,
 0.8964704127606812,0.4515622209594009,0.9083860942384484,0.7344451240092722,
 0.8994289311244241,0.8278538556991722,0.23727014019400583,0.5509975940515166,
 0.36363592145590584,0.4649183974008275,0.6075227535098602,0.92733655716013,
 0.4137866319700465,0.04114826232612756,0.40269271345579083,
 0.8075523140993276,0.2559052495540337,0.3673123566010308,0.9443471111945876,
 0.9317614684487967,0.6823543666299338,0.8876005769311434,0.35175451352956755,
 0.07925350256678043,0.5742206498112766,0.6902796919457911,0.8533627187020351,
 0.40211990982327084,0.019778860327317238,0.9876647304866393,
 0.5993255312816265,0.7691155534505583,0.14664775475953795,
 0.022689297080893955,0.3914608651841448,0.8719752430474647,
 0.9106905297609005,0.7966939057939526,0.7278044630265715,0.5607267363815599,
 0.19369946298280993,0.3361677022564098,0.24134340017345923,
 0.5391029316420595,0.32984899330801,0.9168930732894232,0.9183854602907746,
 0.2961395598283457,0.5494166035181214,0.4741679321603325,0.35666422898932604,
 0.5617598990742154,0.7342796176972262,0.6283838373114071,0.686601376534599,
 0.303989591627118,0.055070788766993006,0.5856322733405824,0.5402184620670841,
 0.9993238498002446,0.48220438706091207,0.5260872446233309,
 0.11740718406861528,0.5520081442420922,0.8653383855627779,0.1842838689647559,
 0.7308952447433184,0.46147775893308596,0.3080285710406321,0.6652839033793411,
 0.37429763736919364,0.7197810937436724,0.973395671148748,0.27495266480568004,
 0.6093836061757252,0.8673133205655689,0.49417692642024214,
 0.06491861255916831,0.7103218784884723,0.3341593150141874,0.7944907661833838,
 0.5815423240572866,0.1740944227876423,0.2695508978221066,0.8879181380271397,
 0.3442400521789529,0.7168635519622775,0.5401301184824411,0.9835192614371024,
 0.8412978504745336,0.4611238439154456,0.5356344942775858,0.8405692806586373,
 0.22105377355515188,0.8165636675314445,0.46842306471755524,
 0.8494815773689885,0.1529996725463546,0.47460185811299715,0.5273750816071563,
 0.9398283851308634,0.7287413987396871,0.5755122407327921,0.145779619850957,
 0.342007796352238,0.7300073950499839,0.7223267798633444,0.5133805959994072,
 0.40715266457648047,0.9356136470618095,0.3512182022286183,0.4906833158315841,
 0.3461410955370272,0.43056035525125114,0.5620197982020365,
 0.10530717669944689,0.9019918988356308,0.29370891043683667,
 0.6944703448720346,0.7164119494632374,0.07127419782484623,0.8818537206298839,
 0.9772042947278358,0.4110238077314926,0.326537454306411,0.467030679838344,
 0.2576610910482007,0.7366307000141153,0.32852709794647594,0.8051690383400968,
 0.9686797126804663,0.43591361662375794,0.36258279866655485,
 0.5281184871119927,0.6855680176486406,0.8827787057761394,0.1202631587515558,
 0.6913782201206118,0.243701856733339,0.449074483217372,0.568825375511364,
 0.7739273219295231,0.2846974982192563,0.6487233441919855,0.7215651500296484,
 0.5878663503832494,0.24073254386534493,7.846755276506023e-4,
 0.2365445890950787,0.41864933614553235,0.8918112543956147,0.6540540924785724,
 0.8962332302513625,0.4993180736819308,0.2891329870222,0.3372788932304849,
 0.7307973152539713,0.1876908526652712,0.01884908252560813,
 0.24639930582999892,0.28714571374157327,0.3776865964866801,
 0.41494491323684646,0.3620348371908938,0.6191158118151332,0.8941923148275277,
 0.40247019399575135,0.4532857285854516,0.27121728647078414,
 0.10795542990364704,0.8721007503222565,0.9179520881868044,0.8430513391235763,
 0.24952075893661652,0.6930731036277201,0.7436209098414636,0.5211774408790586,
 0.9731179259629257,0.1613363068442979,0.8444782449836558,0.48218194113005663,
 0.5307155995312037,0.4519185256883803,0.002406724900300583,
 0.9597511412872561,0.8521529827925813,0.5054692167257403,0.5642088451302563,
 0.7665789661811145,0.5539101325329087,0.7303786770045199,0.23397579556372705,
 0.45677435860667537,0.7959809142454828,0.349536566953351,
 0.059263188899256414,0.13771829561088023,0.4976857716264963,
 0.47694913261019545,0.7050885106097072,0.33065231373874915,
 0.43090165495786614,0.00976916465916089,0.17100946961891572,
 0.6842630626580466,0.7550872812372127,0.4248708674195205,0.2202222715891714,
 0.9675648443060133,0.8773771038007974,0.7439889881500791,0.8452108826592346,
 0.17032463185177482,0.5335912117897208,0.9299284722021266,0.9623678756293379,
 0.9542527760107951,0.2697030132617072,0.8726731710214112,0.3466096099734237,
 0.8793438855558994,0.3976516558924845,0.23061747921279596,
 0.37468049681796234,0.4169020753095245,0.37703208945569044,
 0.6420701796010768,0.9665868206800258,0.4509930950382335,0.42937952206506336,
 0.24437296651587037,0.23203343379813124,0.658248150667751,0.3212674970521793,
 0.6924958820224351,0.30103596818372025,0.418091211418387,0.9731941794404436,
 0.7281918226590709,0.43784998670205744,0.795473104311078,0.4257571826719885,
 0.08248064475824046,0.5233474731668887,0.4214352091110579,0.7863427261460917,
 0.06163738240327998,0.18570128382063777,0.9340128866384448,
 0.48431031527597745,0.46696519670388836,0.41177878189414074,
 0.8552595476714764,0.5539307497431146,0.33200782640057636,
 0.48016702190917304,0.25762366407560333,0.6364254386172089,
 0.15722441140952337,0.9444830799430416,0.816545594142271,0.26169159978616263,
 0.5015936153705561,0.6078787098455433,0.3844048378105879,
 0.0036163032146104834,0.15400018695544337,0.3205927503159768,
 0.3817340610605404,0.8131771529130554,0.16149991891019955,
 0.40228185408992534,0.12424182740851153,0.11988226486347553,
 0.09776124439340705,0.9970194335348469,0.4164490591823701,
 0.16643948836838818,0.5256633683773306,0.13652437755325952,
 0.11405147543790262,0.4619530826742715,0.526624141178139,0.7084459757361283,
 0.7627244946325872,0.3024432655495123,0.7684549617083535,0.6179822398190389,
 0.7358027531590406,0.6394842058412484,0.05371520789598527,0.762440240808053,
 0.11134760234635721,0.9555104181539104,0.6015365287559675,0.9167600465557346,
 0.1436591443917149,0.33567672835140094,0.7249879794924148,
 0.47902762669853693,0.3769847607207375,0.1363086430176511,0.3299101089444043,
 0.09896011728825294,0.6404771766581108,0.00990219561892125,
 0.4006057557445296,0.1898388032277133,0.706584154361217,0.5617241534543134,
 0.5611122810953565,0.7013282008057797,0.6502604328084651,0.704099973205954,
 0.35500398020406054,0.9831755478793058,0.08319784641668126,
 0.031351181371111925,0.5955715525056132,0.22014269765181627,
 0.3961001798541832,0.24922500096397182,0.4556842175718624,
 0.005718125747252167,0.9080620851652802,0.2573421087281117,
 0.1730955465614844,0.4995245223252369,0.6640132652607212,0.07325862175251396,
 0.6262011461857713,0.9819303554913084,0.7430570139885693,0.22365469837978347,
 0.6887053179280493,0.9607887327244989,0.8951987911728234,0.38043520727167723,
 0.9325766333071762,0.49799258283554093,0.35138062181238383,
 0.5406365195511336,0.18376861682490597,0.48153745574489204,0.908482695130353,
 0.31146629242518886,0.7012873342334871,0.8241228713752178,0.4896373947245318,
 0.16004175931570752,0.7946282761680699,0.10332065594201156,
 0.8400250078471199,0.4472263037254124,0.19441850693582996,
 0.16149416468259947,0.16222465063210512,0.45683419192312,0.8562452665490272,
 0.4288117078727991,0.3134673682979764,0.6606384308326747,0.8946340602958894,
 0.7029731609318254,0.3257220645610722,0.34563302533973783,
 0.24043300951204372,0.9070996565144106,0.41462592143004096,
 0.5159482627552603,0.1917116260230879,0.3403977222267002,0.5088140501716616,
 0.5198241772959008,0.5827075923474612,0.9074418917666003,0.3594487779661367,
 0.5050162429133043,0.9370639499355669,0.24590898653455417,
 0.40665858063721827,0.9086386423126577,0.013338236254472058,
 0.13698016721270967,0.5104172683935853,0.19592079726395084,
 0.3781514908385888,0.33815755775143286,0.41221410464736763,
 0.7321912638952927,0.6342291073321425,0.6955905232476027,0.9782929828665656,
 0.24131987660477316,0.6076891885231455,0.5860115957574784,0.2880408053265122,
 0.4661740658589032,0.23068891233468158,0.6094388140181054,0.444512478891841,
 0.8938710511166875,0.46654089123665043,0.6454411787301175,0.7650231311523886,
 0.9326728012153247,0.3227626099347768,0.930679571168696,0.676018229661839,
 0.5569978291986492,0.2627781794064665,0.7268564728913207,0.8179516375441135,
 0.2554146672500758,0.67193631571809,0.9791209940680207,0.7607957788630267,
 0.49297737863672786,0.05046813785825677,0.18371343815246366,
 0.9896433386117244,0.29064316753136343,0.03834356368152303,
 0.6705965790430575,0.22377861942399113,0.7010363402800016,0.911046172792934,
 0.4194596587832452,0.3895914896565472,0.4459607142791324,0.36233424329923336,
 0.32222898286370094,0.7070248848174802,0.5376623699971675,
 0.19844538179634652,0.2603787856136187,0.5066036204177298,0.9199903022403111,
 0.7126373902533599,0.4093026763044001,0.26389925689086646,
 0.17725724998870696,0.2769204321724268,0.19761114807914648,
 0.39775560482639083,0.6556275203478612,0.6851780647206173,0.2593084961291079,
 0.9829252716574288,0.2381213970375592,0.7622856568636979,0.06266185045408001,
 0.22957144256438167,0.02662169511563539,0.4570755462731061,
 0.4148818901568484,0.42154493989108344,0.24886462556470734,
 0.7286155883695062,0.9389081415971929,0.7829106926153209,0.3345130623182009,
 0.570349812131551,0.10853843170298016,0.6601262292710341,0.11578475385177067,
 0.4295387107069917,0.12736709809308477,0.11001530008383953,
 0.23089625652096968,0.09048699474046429,0.996103183637776,0.6537583253508691,
 0.33237396070763336,0.19349772018556333,0.7015057956326789,
 0.19584290987413566,0.9346006399844431,0.5267307352684195,
 0.03419673134876655,0.8328421387190558,0.3994656626133921,0.6719010643168785,
 0.6816787031949382,0.8008747262858186,0.8910795224574305,0.10285100137480796,
 0.5181081482892803,0.3712124637418375,0.468804740455655,0.29924861205052045,
 0.17528212579093605,0.395149659786457,0.7345615331232301,
 0.008750327882003361,0.23570717980137468,0.5249860454815714,
 0.035443191178119005,0.5663689029847571,0.5562076121010726,0.980460238748334,
 0.6448907567717509,0.8237352984316028,0.7590254136791637,0.5644065697096468,
 0.12308392819683434,0.6720021643242459,0.07667000927324008,
 0.9383136558085036,0.9351081977146443,0.5581674610765652,0.3656974690782756,
 0.5406305327544436,0.41944719003347497,0.16475800321583522,
 0.6670991499106889,0.14176033584170566,0.9332503127619141,0.8706128880812469,
 0.4755089157564961,0.0936933111404693,0.07334160553757196,0.8386265579990728,
 0.3675380594016845,0.5116071641947016,0.4608194308383118,0.8708091332717945,
 0.7574081940632228,0.6968560993440318,0.8234868106623854,0.3738753140414822,
 0.8315188556196609,0.2677259297363387,0.38962269230438007,
 0.11125363787190934,0.27851621116018754,0.589976909185847,
 0.06148752763159937,0.028752553241457313,0.729350878145963,
 0.19352550987723127,0.17275253719356143,0.7773526742934745,
 0.16242478369162683,0.7147025612198745,0.6804721268603566,0.646248347520981,
 0.8807630548768146,0.9935087169449143,0.33103490548559134,
 0.46781198139483815,0.15034192608164443,0.5184156133358853,
 0.18259988245247072,0.7488478934492317,0.4693230800998027,0.7564984673510565,
 0.018042457253772626,0.5254453847927456,0.816182453002437,0.8901713446866144,
 0.04478848953371417,0.589697102949921,0.34078157993209135,0.5703725950824586,
 0.4865785178445391,0.7809446766789933,0.06853926388417442,0.3348540055200264,
 0.334224990741568,0.8578644218971194,0.39635254535074826,
 0.033000037522954995,0.2323988513062658,0.5796193098903221,
 0.0927018481973052,0.7630958095434132,0.8258078898789857,0.07353315859726739,
 0.4979708295879851,0.1858281296951838,0.44198705106571057,
 0.06553525395123683,0.9182098193938586,0.6187428073939006,
 0.47393239467949333,0.9976122794916528,0.502647628005611,0.5919824676435242,
 0.1439407993756615,0.4512500138640799,0.386043763837951,0.4406385798446468,
 0.8579953851293394,0.4408179359774893,0.5256435595464177,0.05299578773997626,
 0.1737110062278876,0.45320507985810443,0.029062628129514323,
 0.8238467819251539,0.7902399513999747,0.035804740960921455,
 0.12050018480803382,0.18026131967008396,0.1971670598094466,
 0.7435020879916525,0.13089694348085723,0.9779347460335572,0.632652935776073,
 0.9687572231207303,0.7442301031498315,0.09553865723345212,
 0.10518272478586366,0.9450307849260433,0.6051137293310491,0.9337170123078862,
 0.3538243474656837,0.7721945053474909,0.3917769715197804,0.4188168354766696,
 0.38917158200280133,0.7190877785763075,0.003960240394066283,
 0.6239733479229623,0.4116848790340232,0.42721420399125554,0.8522792438625078,
 0.028509977123888364,0.6314500089965861,0.30084855710074976,
 0.6734850732067821,0.5214225143767844,0.8606652394538944,0.6385383912398642,
 0.13620193716592832,0.7379702140757276,0.5869925375142229,
 0.14607331326206208,0.8243208426727461,0.1200284895007776,
 0.34125561362549073,0.46338940816514607,0.5441788436805519,
 0.4211797247328275,0.03713565856352219,0.6355714052204279,0.2839144206077422,
 0.23388279085837538,0.7613218216063162,0.8417039381700079,0.7012006057678491,
 0.806276842640321,0.3640416957068826,0.3625469319987662,0.7402034771469906,
 0.2308263715165767,0.10458032521730487,0.245745056415832,0.5293189759392954,
 0.11194176296198666,0.3729114739597178,0.4494109694862519,
 0.14328853915152784,0.7886575232097264,0.38207519949523194,
 0.25667947528837487,0.7484331232289954,0.027246291565866554,
 0.5475889771720732,0.38798174141815456,0.9550467349651719,0.818130923362182,
 0.4899295817355481,0.3774511698111155,0.48257913288632515,
 0.37872205954132865,0.9983253599601087,0.706823000242624,0.7343161697583539,
 0.9739965803473322,0.568588244334691,0.8169906501397998,0.030337532997274597,
 0.5852798110861024,0.5166196740600477,0.906224425287311,0.12579288218141615,
 0.6387920632994672,0.3465326394245082,0.4549964029203285,0.4736561312648664,
 0.8367375725502455,0.33319403707298034,0.9415961737822157,0.8956409888951268,
 0.7270813214651585,0.2828880525328382,0.7389783609555296,0.658499262472714,
 0.31589723997181407,0.3215642493369343,0.7511479201710014,0.296055390500926,
 0.42207419557521386,0.22692940039237353,0.2610876303190588,
 0.5583974885071447,0.7751958126313984,0.3346616838938138,0.7292357044006381,
 0.8093644397919814,0.8771748602118532,0.21774618673226365,0.52666364568575,
 0.6456675384840976,0.8769594042582042,0.3676625799401728,0.1309247887172007,
 0.30675637510098375,0.45602407422274427,0.648907389258777,0.447083762797447,
 0.34818174042166583,0.2539816787953111,0.3741127828159899,
 0.11171853556696121,0.5655493409862526,0.38083646689254236,
 0.4486993165471598,0.04805968425979057,0.2942766155212453,0.557933782792262,
 0.5298281099587585,0.024176779368748513,0.4752281037154089,
 0.7186129002014714,0.6617676880327383,0.3645756892992036,0.971101719632472,
 0.5497988859561974,0.9472448766140277,0.042325531059376864,
 0.07599139204475847,0.520161195875242,0.9709657178876488,0.38824044875642627,
 0.9700007763736898,0.6828375268287361,0.8320324529940373,0.45291377915010933,
 0.17462560742228284,0.05472737635377789,0.7041303088270942,
 0.4695367271570918,0.015807504113192428,0.42136845115736343,
 0.04710402294940008,0.3539049112193472,0.8628072570561012,0.9466098377964073,
 0.19126672335472672,0.40781749011791957,0.6222853956228603,
 0.9965755267176412,0.8845241823860986,0.049369794052178784,
 0.16059617391498504,0.21310877795573258,0.5202634997353529,
 0.7589298999350516,0.533953612749746,0.6919454708937454,0.568805141243158,
 0.8899250719010022,0.38166184835923933,0.9709626855557774,0.6015950770434272,
 0.07144333120035606,0.015371867231395386,0.07233405775312574,
 0.9390057192347079,0.6060990895903415,0.476543156749764,0.8345589629219285,
 0.3354432512668204,0.6574222747423458,0.9788495988300291,0.16371169603359514,
 0.7230294955707962,0.61885385020441,0.5785310415467304,0.9969625666825809,
 0.3790104644735697,0.46499967012938126,0.694806484281801,0.5581103748450661,
 0.70381493896536,0.6911924383799238,0.7154883603607607,0.8353892865470491,
 0.14262206430227087,0.25916731057888576,0.8733373865820653,
 0.12243775868659312,0.6324841814304307,0.05972645550553091,
 0.16800064179041274,0.32443535063603546,0.5470119406021297,
 0.9903244871372708,0.36598835559920184,0.34840738796698323,
 0.4460732978996993,0.6257893469887775,0.9694285044984847,0.6932209893018608,
 0.03342764198904413,0.38755385702680945,0.919408326838678,0.9966557808436993,
 0.60333660831673,0.593585028806415,0.43669053279961456,0.3374624944866357,
 0.6975619031791733,0.5742275085670648,0.6760536656598358,0.28406708257210667,
 0.26273404301664405,0.9324248954631892,0.1129546909542708,0.1809595706391971,
 0.4213155734119771,0.5851831808530106,0.6184442109675321,0.2001362695687786,
 0.9703340255642937,0.1886042616708744,0.49546816732021604,0.8885871134309826,
 0.43578270927631435,0.041704998111596336,0.5403198950834833,
 0.21320139770793023,0.36416538316531033,0.14986383609962417,
 0.5098898740953411,0.37510413546547194,0.1019162923006861,
 0.00683887316793641,0.14169634148438015,0.838096918885507,0.2919383862692313,
 0.7280670023176034,0.5870959285242461,0.6499856367313369,0.3044804927459678,
 0.07286802910008527,0.2322242825089652,0.1043331660162764,0.9184141994684051,
 0.4845184299574581,0.4139123355640911,0.19212134312777385,
 0.16568879231956157,0.06506000902321696,0.6812178023825867,
 0.9230454401861677,0.6818130229705451,0.86063705955972,0.7606131520999714,
 0.15389157609690274,0.5655464132131774,0.35412245638733886,
 0.7298295288163278,0.03264630317108752,0.5936694013382551,0.8434008478577242,
 0.1547083111465427,0.48981579036256717,0.7523504550928395,
 0.027606157761385752,0.10556561943420162,0.611348359506873,
 0.23943516626608896,0.7385575997499835,0.5661915732988569,0.5572663878500913,
 0.8053047874969136,0.28688283534373293,0.14918337910569268,
 0.10110968324093006,0.677765193669753,0.5052203496935709,0.9421891264091866,
 0.1915818861206966,0.7957131961608647,0.03984397749253765,0.5423133704319185,
 0.3392528964657986,0.6987340424739718,0.5149520819679503,0.7607801660356877,
 0.7785241054936543,0.379325163762121,0.2243723684216552,0.5164087013648059,
 0.7385990737695552,0.7012916296863382,0.205282149884229,0.06742797020572211,
 0.27021982749239404,0.11387432316578239,0.12188340825718724,
 0.5793564459406533,0.3786535728864978,0.25930222487121757,0.0346538854203563,
 0.32944122817681876,0.5691878676461528,0.5531613023078924,0.4409243707513687,
 0.2171312615929133,0.509290076078043,0.2565804967054366,0.9550824985690392,
 0.9249874383831953,0.15533037898801483,0.16758443030546744,
 0.46180915851064497,0.41110897277481095,0.6347031045742445,
 0.4115409581481748,0.7942398952394669,0.8489403162660702,0.5934919367733995,
 0.28501842010707523,0.6817293464222577,0.3827011190994456,0.5583962008877278,
 0.29828572209056103,0.04749114789512743,0.5941417073279205,
 0.6760944805539808,0.36068773077610317,0.07696123341256433,
 0.27627660586784186,0.7993084942746226,0.25383373710871593,
 0.09683058346568629,0.8101661940999318,0.725772557745023,0.23765852850834368,
 0.3004359925182811,0.7438222339065045,0.24563676271640578,0.1377498518939898,
 0.7357644213638592,0.5152630272794735,0.4080232943921782,0.8180309242346905,
 0.23405138990194108,0.45374510186412564,0.9870561801543278,
 0.3524281406355684,0.0141996329139491,0.10550606665110884,
 0.32880413791415564,0.8304185225243821,0.7815462884225026,0.6482909117813505,
 0.9918489747681573,0.1763331873207178,0.8295248413133409,0.9076397050649039,
 0.061737780131909004,0.5184650584701174,0.5766824024037214,
 0.4165068134882144,0.9156342203442921,0.4292785135936993,0.4869478032931701,
 0.16221293418600602,0.9173139915113044,0.12662666659226163,
 0.31559896029892875,0.03968972958802819,0.6924307024241889,
 0.7230495142053052,0.6531127465080786,0.4540914471788662,
 0.060078282167018604,0.2744014245343813,0.11827114761627355,
 0.3557102185705203,0.4525753143380079,0.755863538687781,0.7368921352131321,
 0.3616929669460812,0.0018094541854620783,0.5370149466887734,
 0.8034338043665183,0.31439127962890656,0.0025986451249451203,
 0.19383199691173592,0.8788562871168315,0.17482537289666777,
 0.45931471965576853,0.2467241077180009,0.933253714327682,0.1295019798471233,
 0.2789219390311747,0.1682679770952794,0.3740620012058604,0.3394765949518683,
 0.7422972876706055,0.0419712540837045,0.4326521607382787,0.2991175323603399,
 0.3880609414634967,0.8586141187487539,0.4738147622744945,0.9648603574216332,
 0.37681951823552706,0.3819796272384327,0.7726761339622124,0.2673522668449575,
 0.5857712882361724,0.6165348020787638,0.6444389727290925,
 0.049675057753062024,0.3370705467611206,0.06538223658104048,
 0.5020216236818862,0.4114813647567216,0.5000568950643738,0.3536837209708321,
 0.3481789720947084,0.5896326322581261,0.17533721689108128,
 0.25367474977819215,0.8818290670402883,0.9512713615907132,
 0.17566462175811082,0.12443117529291492,0.7818454186561407,
 0.8731293207819115,0.6038120418318385,0.8872256085665717,0.9546185594861865,
 0.2343031272080125,0.55696378209776,0.46050682405782817,0.44234774255400944,
 0.87452931471372,0.6611419746026093,0.24736187613718807,0.13223531186115356,
 0.8165281984081602,0.44509165276728724,0.4720361071846785,0.400154905635028,
 0.9318262973893321,0.9228617120078499,0.6251378464831766,0.5860797230604167,
 0.3371241367318286,0.19792933733486284,0.2644715600222758,
 0.34943719644097326,0.14760494421147952,0.2320027492830269,0.789660601571168,
 0.03151194274582492,0.8912254037019509,0.9950770315788302,0.2441634336570322,
 0.9153141755505049,0.13028471397991614,0.4470276952340825,
 0.12644160603048027,0.7860273420841719,0.22462122393189787,
 0.46369864780631975,0.35068180269624916,0.1215514110430086,
 0.1068496920674119,0.6239885477855518,0.948334208250951,0.9818385567109178,
 0.820521210686727,0.5240671102172246,0.6283442271152683,0.6078140207187853,
 0.26380990816845573,0.8471561400155956,0.35889173966977483,
 0.7502979608429101,0.9070756656464584,0.9789889264558247,0.07175091873904305,
 0.7423248507889425,0.026344948266823542,0.46472110946004197,
 0.6804891810449896,0.6528875708294546,0.15404339370948394,
 0.32696236438968795,0.35897942478882583,0.8108753357918269,
 0.41042441558005893,0.4671711755533178,0.977309305375827,0.6580687534705318,
 0.19352843433400757,0.35693912315165566,0.9161482022852567,
 0.10192399653987416,0.8243302966646319,0.2504994376857105,
 0.21725231552553248,0.800404611546863,0.13613163600521572,0.7515257706062992,
 0.36375007503701884,0.35232121154202534,0.6194562952848705,
 0.14545843130747382,0.11818119984783182,0.9806691902538351,
 0.9048171891594677,0.7065154161096772,0.7375636534887385,0.592332464019747,
 0.7712934327997206,0.20483938813564573,0.6773627824452947,
 0.05611366768452197,0.4976721569692798,0.3677724330953772,0.0357071586264478,
 0.9789888447020006,0.13540617511747632,0.5854200585984771,0.6294161298313667,
 0.6404701932538729,0.0972329529135545,0.253741485296082,0.8837356056758671,
 0.46464962628492557,0.32217761681436474,0.7309874053412547,
 0.4861777416997516,0.5410221617685477,0.8701696592334249,0.5703429740099645,
 0.10461995008915859,0.6495973670909624,0.46899112723355685,
 0.42359866299629423,0.38979924243169106,0.1260010457767453,
 0.3015810988724339,0.8513863588651576,0.015420783688517292,
 0.29120634446697535,0.3359903720977996,0.25185372294190456,
 0.5691625216887285,0.7815467556871889,0.24974757815382576,
 0.24716473730333954,0.8862912497170292,0.11521391116213742,
 0.37640011822942965,0.9033040275066111,0.5276476210590975,
 0.30728979685296454,0.7666671012355425,0.5053804464654081,
 0.16354413073504404,0.190772225411166,0.3750588867817215,0.1840227147238772,
 0.8948668991892582,0.630093857608007,0.14165197385705708,0.334407889246768,
 0.6636241494551203,0.001492334015368968,0.2596288253409922,
 0.8179521817397158,0.056353369699983946,0.7245826966825573,
 0.6076462838410163,0.18557127463814804,0.6004942515273269,0.5658452502074933,
 0.3836057538849851,0.9355250807342257,0.07018611621496473,
 0.09878579609543203,0.19512297963955594,0.8042986900757652,
 0.7661758542850665,0.18744178004726975,0.5617174869062644,0.9206532964280606,
 0.2808248569659808,0.8054722401352584,0.2961711040473698,0.589232135022701,
 0.33003993268395904,0.7145003330916533,0.47581424764024016,
 0.05367997559747717,0.4108690550784776,0.16541955359021276,
 0.6723390886091967,0.05802012275151747,0.523088957357602,0.23721269339567075,
 0.9825033391746237,0.9109157802163268,0.24918813700118347,0.8574890870100008,
 0.9119793559945474,0.20271899926324777,0.6827434350703705,0.422048727453208,
 0.4488082704757661,0.664829569471232,0.5721984323450257,0.8050194668218201,
 0.28191714458774875,0.5695038832041756,0.40163089116799766,
 0.5332168903433183,0.10330742185549113,0.8486218010649215,
 0.14982603945226503,0.6865857886669895,0.6668513912062386,
 0.34969916055736905,0.7032823159831109,0.280242625265649,0.9941478070770007,
 0.9884925743108885,0.17227820865660093,0.3772251991976442,0.6675085898332145,
 0.7037554981981076,0.3055208623042751,0.7464097406636296,0.6189458369344829,
 0.880765852799467,0.3292366385899156,0.9176187318535439,0.9986206405922924,
 0.12638651699949,0.06580854907054767,0.39200743125125725,0.4564056994833905,
 0.7717624237350018,0.7546601635275881,0.9799749982688234,0.06513549370230387,
 0.12442401011036719,0.8733600100148621,0.6374257614976928,0.8172129259595906,
 0.5952010726059964,0.6235129218531732,0.9273868260981224,0.6058583598004752,
 0.13883696272129398,0.22625761160651758,0.7030493189864436,
 0.6635638716758959,0.03591185045827161,0.8966293660480897,0.6892795042959095,
 0.8002148681572001,0.6256404165927858,0.8753745379161391,0.6601989626290081,
 0.7440744242056365,0.13532033983885305,0.6157762972751069,0.8521511995065543,
 0.8596168280763821,0.43031667334330115,0.317364138479174,0.49787175936144623,
 0.6612030461332585,0.6769169807943295,0.7365998679864548,0.8219089290704551,
 0.08178124661819408,0.32750408303494805,0.6254928321685366,
 0.22043842243573786,0.4393182359301383,0.33942544227982074,
 0.4706216139391286,0.8478659052468137,0.16877249442241893,0.4511371325059489,
 0.09328305123524694,0.7250544423957861,0.07024357940193948,
 0.6393020810341988,0.18467533173520867,0.1841924061391268,0.7519064697653582,
 0.9046513229420229,0.22272492087389273,0.590225977000198,0.06387104351007944,
 0.06324854505208144,0.035134464472052596,0.16177215546194212,
 0.5860890922382235,0.5679750405237397,0.0336524105673951,0.879729835795283,
 0.5750614118460966,0.3667612065567174,0.7366721883968443,0.08761534368900648,
 0.48889543216650655,0.49571905214923206,0.8659967161538492,
 0.5781897890869256,0.8079487892129473,0.20909126453819848,0.1943766911952769,
 0.2531686239368989,0.19309367302746483,0.5561216177234485,
 0.24322303479908458,0.48125121435547,0.7356145400027667,0.793043683853458,
 0.608705134961124,0.12029956420831767,0.584846909384927,0.9311248828438708,
 0.6668962013506936,0.6414667751385055,0.7522892382424637,0.26294741941599287,
 0.6969867670277111,0.05735266324604371,0.9022222427777126,
 0.27998606865577835,0.5994558622050606,0.38027457417728816,
 0.6483729021762226,0.9343074791425201,0.5952297875530136,0.8368409478235932,
 0.48564340486876434,0.5882867519028774,0.2494685015258803,
 0.042378847135459585,0.028155697295052073,0.8813227452165914,
 0.4063772170921892,0.40948567100274424,0.7265985776441165,0.4455891171775763,
 0.7654954746863881,0.9622768261139648,0.6239493529058505,0.7319358612974693,
 0.1754787439928298,0.11920103538177818,0.32001542589932097,
 0.9761078549288816,0.33489020576735495,0.4050621971291626,
 0.44959098596984703,0.8547094371389166,0.12043549711609713,
 0.22243857602894523,0.27925997342336584,0.4839518886853692,
 0.8892093218153793,0.17350433693931133,0.9098574858323321,0.9773282054956779,
 0.40630228769474486,0.6904252014974335,0.10488883234448465,0.727884788909013,
 0.33449258404252136,0.626269235450377,0.0042747762405473955,
 0.16456217015792507,0.5591539571457121,0.12264619393656029,
 0.9427262705404829,0.5188102588172244,0.6345400459069456,0.8946877149456217,
 0.7929958408914287,0.4423174845456931,0.7364058598565055,0.2592791101355555,
 0.8975549460161213,0.6930864408820455,0.8679042443773081,0.8294142264296716,
 0.46862984406439734,0.1423931632146601,0.2536391648890006,0.6222501026082471,
 0.20406747996640506,0.4264585116309617,0.5668391134788086,
 0.21474371638206713,0.357326161012976,0.050206862842535505,
 0.8042045663015303,0.7107180394405764,0.22582308145227015,0.7415776743022151,
 0.2612269823709874,0.8028287651051541,0.2070072144402082,0.34704289791490006,
 0.7157455636169368,0.5844122874880042,0.41338501897424496,
 0.13368593768499704,0.6244105559585782,0.9671985384262177,0.5566204189448645,
 0.9826928534019899,0.5339340737338256,0.09673138832225403,0.8408438853227901,
 0.5881827544305656,0.08187312004420144,0.20450522871389498,
 0.13362378446704115,0.5015653376347042,0.08109210741089756,
 0.6150092308036658,0.42930828036280544,0.13303421221508926,
 0.27845693936881444,0.16586911400596382,0.8159214083682057,
 0.5598554277631402,0.27715883017841003,0.7517848344432282,0.2033540602821282,
 0.7838763745122118,0.49994250258737516,0.40750832332040965,
 0.2802967386196539,0.07897915534552635,0.9205399474801493,
 0.44712022735487056,0.3803727252270266,0.9188089836752432,0.3005621985413951,
 0.5401439403946939,0.8381561731257674,0.5560554439634284,0.4158132148097591,
 0.012263585672263044,0.8833415693394817,0.173784658396974,0.9920005214966928,
 0.14320754091465882,0.2747572155883081,0.08657596871116957,
 0.6098778488574197,0.7847081604913839,0.687581631225028,0.5903659781687772,
 0.8042626579529948,0.6728797602749186,0.6287159925920942,0.8825321586400626,
 0.7003094175904102,0.026666570359593744,0.2740367644658974,
 0.7790604025772141,0.4221663103740224,0.07545757579088697,0.4426482924745079,
 0.8743531204957854,0.45820727746686174,0.6448082290303148,0.4334701214089156,
 0.836007515452594,0.044426103813934525,0.9868991717839932,0.5793795548994047,
 0.5719309094249244,0.8159301444754177,0.021424773110593764,
 0.6053764072121205,0.33717493778842633,0.3953898082822441,0.9958275156473617,
 0.9908827511335048,0.7629464040600409,0.6820889014801299,
 0.025669814077248887,0.7439644830161319,0.24329183590642478,
 0.03729351122102775,0.4298650280152654,0.10945555757758119,
 0.5045463571770933,0.7000429570712986,0.9550446967687984,0.23678431733589034,
 0.27582714457585455,0.9506735467276292,0.29338583102547366,
 0.5498356517429336,0.2056605529829506,0.8314549496272283,0.06275407989284254,
 0.7632328455698002,0.8046139351116536,0.6435097074777948,0.7422174870154921,
 0.9345952964509524,0.5563460324816633,0.13253002387753532,
 0.37306281465283553,0.7140266724324824,0.3450923044470193,
 0.003147478903840817,0.5281769568272846,0.8131081899568933,
 0.8253041385056885,0.2403110800105136,0.6818193779774467,0.661537020811224,
 0.11968863719433598,0.3300762027363202,0.5890377499896536,0.5213607156696973,
 0.5954538858685533,0.841401622122408,0.053767686715905594,0.7448467920207291,
 0.2912062965303941,0.819020811036481,0.7910265991180625,0.7323359679382947,
 0.6060864643121041,0.17101565845779965,0.6774548053996013,0.897158592217388,
 0.858520368999324,0.6212086477627222,0.5248093208641329,0.2613044204370685,
 0.4393614385518262,0.5415142096211271,0.3151142444997307,0.5160516816115504,
 0.9798273528142016,0.4215407928041416,0.61005099358627,0.9539289145429294,
 0.5321407105834303,0.9284844034152562,0.43211077035531753,
 0.29092211697553205,0.7451217599942537,0.40448302608885567,
 0.44249501574640293,0.761767829446144,0.24743233922054353,
 0.010216489467418599,0.18731495131023113,0.17901974937849185,
 0.322496536954885,0.3139031481185891,0.18665542250547518,0.819272988166325,
 0.999364865340465,0.9265739610051467,0.07250551016289575,0.9199393013190311,
 0.17660709603174185,0.41982712904051533,0.116684154500025,0.4144653617341535,
 0.9963590086772217,0.8249678153692254,0.8541706210385203,0.5348070230874614,
 0.926798192671852,0.4011242046757887,0.5277335856454053,0.9764315281355478,
 0.20548255710479768,0.5024411592605701,0.08900171851249505,
 0.6336956016816584,0.04538562560700721,0.10893740877195135,
 0.9576707499455455,0.0737042070821009,0.7192627190644985,0.3426888494389464,
 0.5164097980112182,0.7463704178827456,0.38553904212084555,0.9699665529368676,
 0.5253232643888933,0.7029508498245299,0.43102518970407566,0.5744534939111852,
 0.8519948945346529,0.5695576939129611,0.9586630668057067,0.21027075871105505,
 0.7230788113834864,0.3444767564067157,0.954176046384932,0.8688080459711289,
 0.6832546467749137,0.39264367350532614,0.1212418023473738,0.9676863277754754,
 0.8627791426724482,0.20203460759706282,0.7514384865242192,0.7595284706960854,
 0.09729166649242271,0.011268423686783713,0.3413778784285073,
 0.6954977370339596,0.3829071073834859,0.30413696432439896,
 0.27996655974063134,0.8063003140266388,0.9414888459984756,0.183706024544672,
 0.31553332154818203,0.10662681414956265,0.6133300905343924,
 0.16343485596647123,0.5691862213539077,7.184347085438247e-4,
 0.7779520886612628,0.3147165236537728,0.4395425802995988,0.37525935747445216,
 0.3857908164116458,0.028394019566229667,0.5288576175236134,
 0.0364194467287432,0.5359495031661564,0.5303756954279961,0.7766379659690754,
 0.6309668862876988,0.36489124468106626,0.4527819494902734,
 0.47920601341075497,0.31022816732701863,0.12445992966608332,
 0.29990101037369954,0.32600184694982315,0.12475491170137887,
 0.8289435979728035,0.9804558224739763,0.09099190515956779,0.8232045136808381,
 0.029222758147494865,0.9094579260187865,0.3739074943462495,
 0.8629172354695109,0.9604698546825302,0.6206302528598356,0.5639354732438445,
 0.583802150338571,0.7895787078656633,0.1571564526178465,0.7720085809047325,
 0.4803149791819018,0.5210502204092162,0.04253397972067563,0.629645552230681,
 0.8757753576367063,0.540760278627703,0.4398756851673119,0.11330604273414235,
 0.23385714880522657,0.7612440606365516,0.9337926879110539,0.7499202229697561,
 0.5406813609141659,0.43179925510732,0.1412248883367796,0.8035758895483014,
 0.196210391108101,0.5242324812779818,0.5745387070508327,0.6278352734100787,
 0.5216654727575307,0.07224340660056017,0.15331074141886292,
 0.8066033901716435,0.21692697721535759,0.24406029003513707,
 0.5751434526653956,0.6046373050554086,0.8291047481166243,0.26479266833906123,
 0.9346848216516523,0.49850828775507827,0.5862165063510821,0.8927356985767378,
 0.5594245828270893,0.22698353470763022,0.2818716284927786,0.521648407314731,
 0.9691447842454838,0.17679932391996045,0.9883200374038194,
 0.002858845710207336,0.6688224024730645,0.6548754714041399,
 0.7013006303061671,0.9382081484415699,0.9733669289939542,0.2376266530256459,
 0.7031992704807207,0.6400911733808071,0.1965200329791027,0.3715794425031903,
 0.7067146501948806,0.6266836027832924,0.09100889522583677,0.8766505888768977,
 0.8916761006550956,0.9277908726256665,0.17584931726597297,0.8080543826379718,
 0.9212529528766628,0.43780737559685223,0.21169139009283322,0.605478665728888,
 0.37574559264357354,0.8596343379736248,0.6335593401542794,0.9332606481087031,
 0.5214235371946627,0.18940502422876904,0.3068417781313658,0.6583170381627005,
 0.17930766156331646,0.21500279154005253,0.4778524573502223,
 0.017457160071788325,0.7881297455685037,0.305729739182628,
 0.09346981404655463,0.6003290955034812,0.568260964028835,0.9591043806780287,
 0.9708754405178297,0.40115469236893375,0.23708588505742023,
 0.5513747991336296,0.7916287721068122,0.7472909123751382,0.9130584011551865,
 0.958483123168933,0.5509365731648588,0.9004864285630478,0.7710235550419329,
 0.4530605076929146,0.8625511584450931,0.5114688313924245,0.33424464953309496,
 0.5233981717355302,0.9114934091369596,0.5655161552223581,0.07510553741320014,
 0.5409265037358978,0.4115004523038823,0.29400398864865296,0.8331963708461391,
 0.2333364011908955,0.1868241356874748,0.44993321620241633,0.8027070434368768,
 0.012519859601300087,0.22821610825028882,0.9046391301242688,
 0.2036165651157288,0.7348534944254194,0.7709761192459141,0.44964613289180466,
 0.26797300054235906,0.02162571841649419,0.886246264618188,
 0.16480274795723981,0.5474438227787094,0.18508801256566632,
 0.19317311838700146,0.6189958790340863,0.1520942999180056,0.3859312879043538,
 0.4639828707787633,0.14109567250519195,0.23507997390953472,
 0.5146207954857004,0.6187243591790957,0.22346092505749293,
 0.15579593834751715,0.8736813524220325,0.8307183712976051,
 0.39181918682073424,0.9760377494353087,0.44797733344345514,
 0.7784392133267161,0.4762106992937055,0.18233353620435988,0.3005750841439806,
 0.5666356588942955,0.8646660762771593,0.3030771526576077,0.28939416203651636,
 0.7347993703069836,0.021482144136367176,0.9621350224672237,
 0.5231706544164736,0.02153273073446127,0.7159025799359667,0.8815468564202519,
 0.8979540353122374,0.53558337631047,0.23108575345076876,0.3538058942965867,
 0.8980459520945719,0.6731431319884217,0.5091151824829542,0.1164726876513752,
 0.27333832516395007,0.961940273126195,0.8118647494900294,0.5792992443366521,
 0.21381122712222578,0.4304993102450111,0.49339082269358336,
 0.4418922606475194,0.32673320922731797,0.5638990294896808,0.4396667516424855,
 0.48400729711828716,0.730961545830508,0.5815132174497653,0.17357748351368663,
 0.1139305690691621,0.4221016851783467,0.5796482929898408,0.9886215301856723,
 0.3217930754485817,0.8302814628507117,0.12126457125597434,
 0.12851192150202728,0.7766508790127722,0.4534642328897357,0.5859538669608629,
 0.2404061698736204,0.44389944974269013,0.573120350297257,0.4267571672667603,
 0.2692285273353886,0.8562770512222349,0.5003731301168876,0.2703872527534099,
 0.08797660082799952,0.6942097870521007,0.055905091759500425,
 0.2373483067177956,0.02607036289432796,0.5968537529237035,0.7009106207435517,
 0.5432817270446316,0.44963287095162874,0.6355548591316103,0.9262644393924736,
 0.0701713694835645,0.857896750796044,0.7693956866081278,0.08682824480988593,
 0.8115083280804349,0.7487765623645327,0.08413891149089381,0.7579204330585982,
 0.4403822755035435,0.3143038122494335,0.9555844463785322,0.8628401552966829,
 0.12797517224317456,0.15354904110223988,0.8555538001668206,
 0.5974447364438337,0.8139235607364987,0.3031400190014909,0.3562474433906545,
 0.09031237304534678,0.33152485323755965,0.6388520714888627,
 0.6823575738055437,0.04551262593847549,0.9869168723717445,0.3827990148255187,
 0.42894779063376676,0.9507138203878234,0.5334955171143348,0.3816271824969615,
 0.7728775737895734,0.5684989207343164,0.817409114786261,0.6150335168386836,
 0.8489416101442825,0.3287682779026967,0.5655528199709116,0.3548653255104476,
 0.14061311873039506,0.5506265612175584,0.857787507055215,0.8499270444331236,
 0.7655569726396462,0.03925193188154552,0.6318924605292751,0.1705906321120535,
 0.6981483225035333,0.3531092026364352,0.8128771285470133,0.5932748792921136,
 0.9143876532628435,0.24157385209560311,0.31080422603280544,
 0.2674822301790686,0.4729913282762849,0.1415759438789338,0.10736688045312881,
 0.9490962435980936,0.7917836542199851,0.14889364822531614,0.5980077735883815,
 0.189637090057858,0.5125521263407256,0.20173088351492108,0.6478045461735341,
 0.8761937277022789,0.26766686947746543,0.7895513518658966,
 0.12094870773442057,0.16156754517750005,0.9363310178647409,
 0.4175967916491474,0.25871551144187865,0.4810048463090678,0.75756699783383,
 0.8517927491573491,0.8752291112279886,0.5242908900925753,0.4009281085922495,
 0.5862495823137461,0.12789553196589976,0.9216140461572265,0.888031293881733,
 0.7029566466961965,0.6607853700019566,0.44048447358990794,
 0.39428388675507164,0.817267757257491,0.6960548355625369,0.7757870771873724,
 0.20138391518000032,0.34705412485085096,0.6829944109048082,
 0.6316256201706454,0.21906445333241764,0.6266066448277178,0.287498018053411,
 0.7903104473649892,0.7237569683454982,0.6100237335817491,0.9247221120086092,
 0.41819428643195145,0.4476019586592327,0.5657835448681767,
 0.11970111043910969,0.6014184157978864,0.3190732178983724,0.8427163577487846,
 0.2506881762797769,0.27748117250172666,0.31012368433099446,
 0.04628135403671485,0.41571566240959,0.06909704274393758,0.9013587496190203,
 0.8495168020801827,0.699864377740061,0.31474229186915925,0.8867125456604168,
 0.4052982861150224,0.7772400449230524,0.27836599046859223,
 0.15496733263833373,0.5158637753435298,0.44394643747788687,
 0.9148618894864683,0.690838413084393,0.6602854801341399,0.855728939656992,
 0.4247672830383473,0.9720666440326955,0.47466501664214755,
 0.0024132381163041172,0.5444341237443983,0.32763482685144973,
 0.9005029654016415,0.2861034143151512,0.815372769187032,0.8331923313380654,
 0.5523370606779353,0.2676035249177331,0.661600764458768,0.36075863973119093,
 0.051121112295855586,0.39810494872282465,0.19912805082747054,
 0.024464212693695697,0.7094988378967284,0.3825787869641616,
 0.14658724591215855,0.478058065023236,0.19041055562478004,0.5378253986183027,
 0.13433974018103245,0.48462558925349797,0.5427139450107071,
 0.27938780460256185,0.6905445710311796,0.9891688544556498,
 0.38323567084267207,0.1477672459824103,0.28616720586045463,0.518911843664745,
 0.16106696324103897,0.7790504059583292,0.30445176823276143,
 0.9159423636883472,0.8599126746661436,0.44646482725995007,
 0.43552961558329284,0.7522781452436634,0.0744957238941999,0.3437939706482651,
 0.35092267367053365,0.38063507996341805,0.9412261885386543,
 0.4341543217531544,0.9930892190952297,0.3609665382626016,0.3023095939021395,
 0.5453530472179452,0.1856390455785475,0.566603515153179,0.3138946806002192,
 0.12013775508615232,0.8631821271359182,0.2298312155391874,
 0.19092952931270168,0.36545777731049334,0.9210101250640824,
 0.2981325680158624,0.9784664388019085,0.6379025575542184,0.7030094213934084,
 0.7390374154723305,0.692601468534094,0.47934906413426215,0.1362172111693012,
 0.8226467565199748,0.407780041432926,0.14255461998071617,0.2341486008044238,
 0.9299883873617975,0.8400407008627102,0.6749461299818353,0.8091182751136393,
 0.4959529901126857,0.6208950380046461,0.4242118715158948,0.2981900626353128,
 0.5868238438718252,0.5096934205775394,0.08193567010096836,
 0.20982265122555055,0.525634279878343,0.9587946118821223,
 2.4390382173544722e-4,0.07119235592708262,0.3017097250105657,
 0.9476498702373269,0.9553729515667725,0.27357781307620366,0.6468380891280577,
 0.5318106419405709,0.3191824371593096,0.9642899883047218,0.11883739915413227,
 0.15987134605754105,0.07223182616296875,0.1889781962739866,
 0.9225866584739512,0.34877847610967994,0.011908217932789733,
 0.27296759087184086,0.8708445355546448,0.6721546705535866,0.8208573681391397,
 0.9221533095456452,0.9527219964784027,0.46997380824257384,0.8758078963804178,
 0.8890546400651307,0.023036791755398967,0.7854425403315457,
 0.4154080389582855,0.8383160029326571,0.014034218779151786,
 0.8364119905108172,0.09926488584083981,0.39332621832548287,
 0.3375692162700893,0.8545925312968484,0.4305994675328405,0.06957455380487454,
 0.6058253249100487,0.7085493331854796,0.09631672768273236,0.8624963503699654,
 0.8126045635527896,0.23465775937223654,0.4698214342038437,0.5813483424270939,
 0.6720153316645222,0.35181744001806226,0.9572124309576879,0.8580624909972971,
 0.008489857552185232,0.8909467376937493,0.637109101010245,0.5676810005021518,
 0.20770032539652517,0.4442160810516641,0.7119877700797592,0.4277024597437239,
 0.18907721063482563,0.9406817346997417,0.8287985201567345,
 0.43398167934545606,0.7872160761568533,0.6398003687532875,
 0.18889736505556776,0.5491353268317863,0.026483921906590324,
 0.12034458629373046,0.9337872986447969,0.08720629100698774,
 0.5597611410149106,0.18502089325907423,0.03293296739282903,
 0.4601166359194855,0.08786742590695273,0.7292075945782748,0.5132210310744274,
 0.19446219555320954,0.4942751159789508,0.976568885564181,0.282840002496752,
 0.9814814395034747,0.27701312448286186,0.5381342972434592,0.4988078621176619,
 0.33655950945501145,0.6531420420955188,0.5247469711738197,0.7849525632189057,
 0.46811925564713797,0.6590295946184543,0.1177278670468781,0.9943881395421382,
 0.1835676416264288,0.23705548572392,0.6510166665542761,0.230600447062013,
 0.5029846367194662,0.015882496488492603,0.3574326972279588,
 0.8215680954205421,0.07907019148426353,0.6413064197208663,
 0.39280846613989495,0.25229925720666646,0.35874089517263974,
 0.45057259223116297,0.7662072322023485,0.6902488723780005,0.5398660097423,
 0.19097754319325155,0.15423193498414067,0.8161385100914959,
 0.8983236237415947,0.24222814899885425,0.3668186974834545,0.2585511089076844,
 0.3289713543714473,0.3328223905386045,0.4859042903889075,0.5995781189145185,
 0.08863316455976011,0.08358499982976475,0.6406126603379328,
 0.9856644346796437,0.7151132977418987,0.7502390080115988,0.2773044929140188,
 0.16271480286195028,0.09367029917956671,0.8969968679464451,
 0.9428663690778426,0.26311370730414196,0.0441965203112058,0.859910833379513,
 0.3135637557602422,0.43739875446703413,0.9100867370526408,
 0.25077539387978653,0.15527028499209294,0.24866601460263027,
 0.7746806600907652,0.5706010342085144,0.04831713249443359,
 0.14608093348991802,0.6486081595664197,0.12431541875560137,
 0.10481502185508129,0.20737956885564846,0.3356531234777712,
 0.3963028989826485,0.08937244173786052,0.6212192854025327,0.951095500804821,
 0.866243349254235,0.16046827082940807,0.6844569560981423,0.5839709477807967,
 0.24176457711466093,0.6787145048738722,0.6642149313020717,0.477822392280028,
 0.8513618213779168,0.8948453365385763,0.9706466029897743,0.32359598297026093,
 0.6589480653581521,0.2735800863838548,0.8090895984151487,0.10395337155874618,
 0.8044001803627143,0.8100427069957564,0.5090467739227933,0.7024336185140793,
 0.959352883185897,0.7518971519219574,0.1258086282667854,0.6369932591616476,
 0.8133209541558646,0.5041306919299152,0.4336639604659437,0.9086898958002235,
 0.05501610303354687,0.3035828817432502,0.3423428872762283,
 0.16247400436637505,0.022130611876684192,0.7266366143851309,
 0.04876379493670935,0.9692074344793236,0.4928906584416394,0.5015748872772723,
 0.8167157928578903,0.6406778042202069,0.13469206925516775,0.5473798512409086,
 0.7838313834037619,0.49847534725681253,0.6168830759865125,
 0.38717656848198967,0.9048673251391436,0.88479220105034,0.27805472073677684,
 0.40833164642276076,0.24670578265644993,0.5027561021333661,
 0.9550799376528116,0.3122121487466928,0.44179423001283824,0.5931204159490329,
 0.21645511280712415,0.4741630557142782,0.31213262004608566,
 0.9740575911461684,0.7694008158441729,0.12410380589648207,0.5756304194589006,
 0.7252822488500404,0.4722327413176217,0.846861650351758,0.05876902801866224,
 0.2054485528312726,0.4673149465341683,0.9782802390334238,0.27246983368754973,
 0.7032425197513578,0.847727605345364,0.7238001188173727,0.5975373587699132,
 0.04002516039535786,0.1779151900563749,0.9127201785977773,0.9835720224180639,
 0.5816680489462381,0.9207240262132912,1.1614398676351723e-4,
 0.44560642156570784,0.31540603169748427,0.8542642831383729,
 0.2442124533054345,0.9633869507223931,0.2145638603449096,0.6140766542850665,
 0.9081034312646116,0.10720352626659779,0.28049355358898487,
 0.041667437307003174,0.8505296101650092,0.8426155448547528,0.675608257497423,
 0.17934978613235164,0.2954256142588486,0.9714470431988288,
 0.30698526278087934,0.9312478216032836,5.265420540815358e-4,
 0.9786271248523546,0.19961606659632058,0.5062791915418364,
 0.05349823400207565,0.8958109584191161,0.2311209449658631,0.2932941692764113,
 0.480776203306558,0.05246600088112374,0.681875512596531,0.6646685074005929,
 0.6431349483375983,0.7571298339539438,0.4835070313901094,0.47984760950839733,
 0.3345738040209044,0.995832342143916,0.30852609416193355,0.4739729673958417,
 0.8350836315644572,0.54147146000983,0.4014206441132179,0.24641236931002475,
 0.4517070584873779,0.5153852615506408,0.09678325898520812,0.6001279603336653,
 0.09790187892143376,0.6097148172329522,0.8367180658127391,
 0.27227075381077914,0.13316111105085093,0.22818354852341505,
 0.17744338366840695,0.1778982802198269,0.9371946688074221,0.994069707320758,
 0.11882787911459525,0.3024025892764024,0.014563046038458038,
 0.5129770289314244,0.02944104622793009,0.421508987631521,0.9175757100786218,
 0.976446212450696,0.13830867455252172,0.8940907783055583,0.18230714761064992,
 0.1838815630273971,0.2377286428525308,0.10203979997188417,
 0.14126670896108307,0.4859837604454408,0.32741001760136945,
 0.10966197836799518,0.37040966388921326,0.9101673456560315,
 0.4560702287574846,0.8806278554533032,0.9949164194526877,0.9771251096366244,
 0.8465188192131078,0.7515697762655866,0.5295798048390489,0.6874937383363594,
 0.38695046463657246,0.6870960358801125,0.11865004281584368,
 0.8672363318914194,0.041979864469686934,0.670091589105614,0.7984099402629073,
 0.05993890305926586,0.04315204698785302,0.7974394451770309,
 0.7632060123685238,0.8852997897576991,0.6721583120352738,0.6061406774365783,
 0.6066001523511992,0.9066182862029384,0.7782754700668865,0.7969657256505707,
 0.2926426076335302,0.8583034265399562,0.8991268941230637,0.3326792551578356,
 0.778301174799176,0.99604851059896,0.986332893447741,0.4572539291415356,
 0.24888376072700558,0.16034762021110105,0.08631459541413278,
 0.01852140988552442,0.3047171586506049,0.9138421559572094,0.9348253365696833,
 0.004238792966018368,0.7365004128500177,0.17972838244864753,
 0.050005695234157876,0.864616382606196,0.6462042676939861,0.8506160917545375,
 0.7020781903197242,0.8272887621909456,0.5645447678732921,0.6950166849299465,
 0.8266076865509623,0.32386533144574203,0.7260120006374196,0.4621176392181785,
 0.4760248049856335,0.1554787161940071,0.4208403713356198,0.08438091890497945,
 0.11565116897706362,0.9094419687751939,0.8015604756738162,0.7919953108534075,
 0.02051497660186008,0.3342521963110088,9.870779485660064e-4,
 0.19715752715167745,0.7359186481149207,0.052331207193611395,
 0.8862501659905747,0.9359150629785153,0.33465846669050203,0.8017532982179775,
 0.45149249418315074,0.23324989487408776,0.26963199226336254,
 0.8989802354002027,0.7082743846379478,0.41697579117225436,
 0.32256422659634765,0.3012885259989022,0.28250313622879464,
 0.9385791962823513,0.371702194312049,0.17360408774979064,0.05851409011105724,
 0.930418849666134,0.6930485678256966,0.4566890518382031,0.9497074182050549,
 0.9561385190013932,0.5476508016551388,0.195458823835294,0.619140558122177,
 0.11014893592420938,0.11327564036579574,0.30363735646554146,
 0.9969886839005468,0.6379328021516906,0.8401522217078808,0.5144911571736369,
 0.6300709931019729,0.902965491173903,0.828893167055716,0.7758308728000649,
 0.3328291439737522,0.037319857358454,0.06446615655851107,0.8409994313212911,
 0.08409422635914166,0.8465255323310547,0.14496106709066248,
 0.6536920428172401,0.3471881790323884,0.5518980048054484,0.282800876400365,
 0.9927667960203569,0.3947882734365815,0.25019312044869724,0.9211042415367964,
 0.5727729917383283,0.7604706702670672,0.8290406567416876,0.6566561443697274,
 0.9294316999400352,0.07416392373666447,0.09051214322152978,
 0.9280135708418135,0.8156422798206174,0.6427888411481428,0.7566232934804492,
 0.15305249545017052,0.4156851043332711,0.48216259208248835,
 0.0016801862096117448,0.192652959244495,0.8246140469507052,
 0.13532077988168878,0.1584282052066719,0.5632357904590553,0.3067290246641404,
 0.2666678384824902,0.4807067687665266,0.6852303415947085,0.6975605774291592,
 0.5881385078827089,0.6817012181725695,0.016542906197654306,
 0.49884935912407935,0.5775054619440209,0.7609022640392787,
 0.46742728810747536,0.8376585808590631,0.45133509181166764,
 0.39048272033518516,0.1920353995837576,0.02736792765322149,
 0.5948898696340112,0.14500772123032668,0.7596784434952003,0.6007495290293631,
 0.7687124301479491,0.8530618481764154,0.07653342501741023,
 0.49680976778814734,0.08823896394234154,0.13437179677001143,
 0.8322509871841235,0.7364345240384689,0.8206845007824363,
 0.050236008454799475,0.9598961949864725,0.3335675028985766,
 0.3256563173880731,0.24081225191382793,0.9820917254867039,
 0.08543302076843962,0.19220093914058256,0.960148809930269,0.5941249810324825,
 0.7386339232890573,0.9716091837501667,0.47577930246371913,
 0.19447366239740194,0.14841462972640296,0.5996480381820988,0.289545350992339,
 0.8323167951575213,0.7708594386674178,0.9520961148478683,0.21543616327502957,
 0.34937883429405703,0.4038986059765435,0.38204589668587485,
 0.5218169109957262,0.11023585782702794,0.9950784002200878,
 0.28757172709699585,0.3848403837501403,0.8372764763979661,
 0.15845031633013162,0.41847574471507987,0.8516202867026622,
 0.5498917549047582,0.3269943108480069,0.3282393282246189,0.4888087796586904,
 0.6406696132938653,0.1247436378260981,0.8952469489763202,0.38343202644383867,
 0.028027505943692077,0.3729698951227165,0.13512829043731367,
 0.24894841644760202,0.2844109931794587,0.7540223329102984,
 0.41814203856909926,0.3038285526432667,0.9177141893985582,
 0.47957264861338744,0.5002927303147215,0.5896802072592154,
 0.27782830529833913,0.34677875499888844,0.35330568055655076,
 0.5930287902987416,0.659799848304376,0.2737709916754075,0.9145989630747714,
 0.03376187629351368,0.48468698046787184,0.3217424888677184,
 0.6682649694971674,0.4942618473663716,0.7796276701891316,0.320738397831681,
 0.9481913937136301,0.14876644720341792,0.8739594712918445,
 0.034292636112644326,0.6865729804263773,0.9960816323380137,
 0.5412514601468779,0.002853524286337117,0.7040113595435843,
 0.36258036584065645,0.29233886144863064,0.9893222224222681,
 0.14161806853817227,0.6945820738881694,0.8413800939814338,0.8364421824420715,
 0.4381247592928268,0.0874391185169685,0.8753801443425329,0.7254374416755365,
 0.15038637876144556,0.9665608205893486,0.5873159307114494,0.9054562928478462,
 0.45529638091509694,0.03147113257275991,0.51654202738366,0.39715499792894604,
 9.69650177904624e-4,0.3309311221674399,0.38251548315827333,0.434914828951971,
 0.29832463324967995,0.9249565732646642,0.8116570979780289,0.2697257103875439,
 0.7764576169233046,0.5130892793781511,0.1025247456961027,0.13026852742055917,
 0.15217030285721,0.21531647685266309,0.4859974220084812,0.04296330507709778,
 0.7883504286457683,0.21104306295909014,0.22586382305222585,
 0.7341146277209271,0.39608106651601416,0.7011195178580676,0.5098451131596845,
 0.9508780652299362,0.16453177380148143,0.7410387885099636,0.8614651977848654,
 0.16320610761922982,0.9862047586665377,0.5269140200205308,0.3666523379953871,
 0.8897217958518141,0.5305422476317831,0.5663000836923542,0.9206802614197128,
 0.7359181913720849,0.25578034824523854,0.2958231525787848,0.7975491288074665,
 0.2045058802158991,0.9531925112734361,0.7360674775613418,0.2829973705640989,
 0.5074216776944432,0.7993568032097128,0.9900951021806025,0.9493022632341807,
 0.2390494652377716,0.34725231842284765,0.05913339809276452,
 0.9397361415317914,0.7460112501504588,0.13871221101195397,
 0.03865707838447241,0.538033506098639,0.6350146662887476,0.45607316811287735,
 0.8936433968149136,0.6587903879918603,0.21210282370395017,0.2804999006188609,
 0.8890643618220203,0.8811577895571535,0.5431082521009878,0.5332123253824047,
 0.3575135321741377,0.39446781251060115,0.08116449896638989,
 0.7713355997586514,0.0937774458001428,0.6250372265642979,0.4791755394254472,
 0.8409781794152381,0.5800226948410246,0.9479849706699222,0.6350265893274241,
 0.15110058020838812,0.047925393681459205,0.5371943318820601,
 0.9332038053237364,0.37022567635846804,0.2690758349070227,0.7210993694197307,
 0.3121201608606208,0.07866110427134798,0.7179880443894104,0.905121674924326,
 0.7708482985261036,0.8824476643601511,0.4854017796376934,0.8975211111827978,
 0.640525026635169,0.4588177123586249,0.2369620240677459,0.543645737059194,
 0.17249486917608103,0.8920058175592627,0.6617055487197446,0.7062014187788421,
 0.3384660756794424,0.36904691396807543,0.8233809658913918,0.3104599073320745,
 0.9571677126158992,0.6614021549872074,0.03140165423755181,
 0.21222180915027067,0.8587874645828862,0.968204325905167,0.16252211558419927,
 0.43279083486388836,0.17844335829617197,0.044057282371361084,
 0.08732884459004142,0.8741349693869489,0.3734181721136709,0.6808259175699611,
 0.8500215068470586,0.9240424323620867,0.46008275144615673,0.940051759967493,
 0.9992452611542635,0.704508935826885,0.18044808803115098,0.5881075611207123,
 0.35207801086148105,0.4381535902378412,0.7600753987153532,
 0.0076696857467184465,0.13052291388833515,0.2747806997759058,
 0.024068601801752854,0.679923187979982,0.6704640923978812,
 0.20718495524205915,0.038736004661205636,0.12263491260415182,
 0.7867611319159153,0.4633606375619195,0.9791959073483049,0.9038655241910039,
 0.839096461126353,0.014789347277205156,0.8774401807915565,0.6955185616571884,
 0.6363019235101257,0.8834489790977009,0.6514441733113395,0.451206410684697,
 0.29527270800600014,0.02938818793404796,0.23105265080511606,
 0.6641784353735756,0.9690959439731665,0.5211209771175455,0.11421840305919917,
 0.35876473205804704,0.7586055380017444,0.8288492718537286,0.8279980666385824,
 0.9477572830317658,0.5781920597016075,0.37593406006992525,0.7014691015108082,
 0.21197713769426707,0.3186830597606234,0.9700647031389258,0.845034976949622,
 0.8287034951474721,0.5337069062169648,0.4833359736958843,0.07047485800676312,
 0.2819506156490621,0.6215643430337997,0.4336721374729292,0.6945758161104738,
 0.9978482422937878,0.2830701146068987,0.021722964845301895,
 0.8620924168576917,0.6066560528848308,0.7358426644433277,0.10223876837667045,
 0.5345002420881613,0.6070745299697956,0.023664893222295258,
 0.6640689296684784,0.010656415967098765,0.45328958980057166,
 0.959370054845177,0.13790173337266665,0.2452053132090457,0.19799047379545665,
 0.09328173663079686,0.9012420376897732,0.5990395296276662,0.5990877339708687,
 0.9367132476110779,0.08480572316224844,0.5452151382335776,0.8861371397384326,
 0.5240599786796055,0.6341221294023364,0.8374793123659713,0.0929748536838666,
 0.34878220417658923,0.5754955061070471,0.7252219765308481,
 0.22783125643743496,0.5434395754617894,0.12521080920530125,
 0.039664734105992916,0.5913040741834659,0.06594113073355856,
 0.5027278465819602,0.9918626520127445,0.07907772511233468,0.3145324518167777,
 0.5561508385336275,0.811364671669282,0.8676313717499139,0.845792615359412,
 0.13938643769974757,0.28760413546462205,0.6999315235595064,
 0.5206129665466435,0.7519157295031796,0.40655981310858047,0.7652570093360904,
 0.9423739443267143,0.8660615139309499,0.587700796567479,0.7972029766132073,
 0.2944672888030453,0.8520298311012566,0.6700089420308343,0.9581759243437986,
 0.9089132415684295,0.871864404731951,0.5911101742839725,0.9727296286227108,
 0.2148987397480937,0.1066738352084311,0.6585778682670164,0.5600320229724831,
 0.7577932917832041,0.5625199489454973,0.22109917759479714,
 0.24015262369086865,0.9465380316673451,0.5193105405273878,0.3554902379124141,
 0.519817758226717,0.007251572225919989,0.9647724894518338,0.8710128227184537,
 0.03458943927427105,0.8554413151412098,0.021886644127791288,
 0.30737459236819487,0.27523288644401445,0.9021686976396558,
 0.9239199416096676,0.7348963011347601,0.4143720895672034,0.6461056704567425,
 0.1638830576318807,0.12322969212260115,0.4124874338536495,0.1483032462392524,
 0.3237775208254433,0.703780682172173,0.9129818280278017,0.8991147889132662,
 0.07717610981868339,0.9965851108116892,0.6033773619248173,
 0.011971453560293388,0.5679265135406201,0.693590868392987,0.9268405723447556,
 0.08675456204156706,0.11171830019889284,0.25703802053761327,0.77398842888848,
 0.18948887908817347,0.8150514827404719,0.04217075898532585,
 0.1516266611845999,0.21124904522414367,0.9616782931793805,0.1955653968940173,
 0.12205956992454736,0.22248183328333182,0.1566786294967677,
 0.2956238182108808,0.8553055904664665,0.9618798715748197,0.641868639450112,
 0.6485732768292827,0.4259914148110969,0.34837464000096385,
 0.28118700405770325,0.6673715951540666,0.05060370305777673,
 0.7539958660153314,0.9699412596565061,0.7345874437740414,0.451797693768097,
 0.9640710279326042,0.2767664926945904,0.1764282881272704,0.0827606906334456,
 0.21964620581940242,0.4858133888554341,0.0676943262462113,0.9276940942225207,
 0.41305905370085316,0.2890013635478492,0.18740445283087082,
 0.9275724300336576,0.3053665323497635,0.31187510090729764,
 0.45692613634397217,0.4140616139440517,0.9519039940885176,0.8036084392456624,
 0.8947033556180131,0.35467173969574683,0.8140759197349952,0.6265132143732808,
 0.6506392497189437,0.44539356412340836,0.11763477197999705,
 0.6079361043649737,0.4875453073497672,0.8480406891632684,0.8276772695564804,
 0.7561996794860798,0.8213948844748445,0.09652145101256726,
 0.18306743963280891,0.376415895226049,0.8212274430276143,0.09769851892900583,
 0.005304980929627612,0.274229518740032,0.3413796324436532,
 0.15467010351417998,0.08566987352413102,0.14436791663252968,
 0.4162823366698195,0.5997352636608199,0.6554836453559345,
 0.041825990452616923,0.915637714332133,0.996754367732989,0.35960722645601007,
 0.3140685073852101,0.7021757797151865,0.37564131099247433,0.809111728972148,
 0.20354633255273624,0.9429339363041155,0.7685508972084404,
 0.22873858975722594,0.8752146065261012,0.38161054845691045,
 0.4316958000111051,0.9940395576735739,0.8507422677241175,0.9435680514486342,
 0.7413193944333953,0.22120927335073315,0.27614570692014184,
 0.8648245699146859,0.23864169526352608,0.2874217829108955,0.2550616823365315,
 0.752248400869209,0.19311361923377346,0.4488760619464538,
 0.008418145263968668,0.13391114901066958,0.7108909227665875,
 0.22531308524817795,0.7898225389948863,0.9876638367407121,
 0.0055601075357391005,0.5158964393763159,0.24164225852139276,
 0.5927906263939848,0.9098125984215968,0.2118871134478888,0.9013086162847791,
 0.7785946137536728,0.9690670084236984,0.7934434601839094,0.44598560681105703,
 0.24690224804175553,0.04392268797305143,0.535684768218581,0.6099358926781291,
 0.5358322118217448,0.03507202317875846,0.8752456029392884,0.8251988928225643,
 0.12865926267357242,0.45053438949168756,0.5016510870610664,0.855130841411152,
 0.5144166443350197,0.03781744090841355,0.9675414408594413,
 0.26654729200838834,0.4109650814738467,0.4601671866992746,0.5675343069503237,
 0.226602200667,0.6562500843905942,0.4077034214634392,0.45800241064338376,
 0.200022369671374,0.35985112983451306,0.3494164902909298,0.11559972053753587,
 0.985957854989445,0.04821830826093276,0.6836692072828122,0.6326123394539769,
 0.3925116137620027,0.6631163420852133,0.9727108438686962,0.7734298742444923,
 0.5142103214054123,0.12416334081395286,0.3648902006276521,
 0.04256639658098871,0.6471318936062496,0.8142690376263905,0.9770982079114906,
 0.03957632618071072,0.3749456596757694,0.8022330064835056,
 7.877527642501381e-4,0.562983159071696,0.40259890544698995,
 0.22192974770908447,0.3461390462557641,0.8074062636778605,0.8843620411864955,
 0.14061544008690285,0.9902120337664572,0.08656912500671499,
 0.6929180704932172,0.5327605939822941,0.6813054939060155,0.43395091920746465,
 0.8760244708266186,0.6377017986339287,0.44237346058432336,0.907189176262204,
 0.5757543189006303,0.19718297646805216,0.5321713020693244,0.3897500095520372,
 0.8120511846322941,0.6617297011724826,0.060181955327026415,
 0.022762281467014756,0.6902059212902204,0.04602784809929006,
 0.33757934210739116,0.3353796058874181,0.8402389860582298,0.9017107613059038,
 0.44590397546640326,0.3404023986102962,0.4238679790758717,0.6896829546588992,
 0.24504439642843023,0.7833975949118619,0.6069719976675287,
 0.04407370515423059,0.5554421764799198,0.04603247266509902,
 0.4077346286826955,0.6389805327346163,0.706482834306254,0.04406677371161272,
 0.019226578827731267,0.9289997322382836,0.9986536269797874,
 0.6234974275283611,0.053221898036841964,0.2646380408442589,
 0.019829543030027375,0.9367034098209259,0.8488588971382127,
 0.3667098511204031,0.21510179255289708,0.7923035511614158,0.2052553206927057,
 0.1989019817622637,0.2586709827090059,0.7584051750781516,0.4388972455407123,
 0.475861644607803,0.4670109746280783,0.7892485763863822,0.9826786483761819,
 0.20471877081495526,0.7185262872301759,0.6283336729292918,0.7795360269464571,
 0.5478324508951083,0.6025857369308972,0.8675621892079943,0.7772735972620538,
 0.9604492680869378,0.4750437928408915,0.5128310905087842,0.4630278928104238,
 0.3992700637729468,0.4565390046184099,0.21858627658104013,0.7837521968136101,
 0.1427895083273103,0.9593643684655543,0.5730501144972799,0.9900572163942787,
 0.22775223279369494,0.31537534203971296,0.020447391227020772,
 0.7183406149370074,0.8229333805612005,0.15416383693930935,
 0.27341471013300966,0.8857610318701958,0.09823273038493696,
 0.38697392358129845,0.8143904528423445,0.22767191664805864,
 0.3247363293916752,0.6184546758234328,0.279222020310426,0.3417960782098739,
 0.987040001420312,0.8289515522437322,0.40620856601172817,0.7446828608137598,
 0.855454506532659,0.185359770399099,0.512320588749951,0.27578047347671175,
 0.7377809058352839,0.7778593231343512,0.029820322922653797,
 0.5166858324673863,0.03772094081907118,0.7405039817872505,
 0.13389629923956325,0.6779670717106684,0.2624836369378476,0.2716957790236867,
 0.681419692603934,0.8998147972688895,0.8271291030410074,0.206327304999128,
 0.44816378540778457,0.5262977884560046,0.9647116129158813,0.374194094904567,
 0.1505691539175259,0.8527012396442228,0.14798358692574454,0.3132919183674103,
 0.35584942254637697,0.6634764339373465,0.34553891433987405,
 0.1628260814368354,0.9738053174958057,0.19581828681412417,0.5474017086793445,
 0.8401920240684418,0.7811926097527983,0.8299885255343679,0.641499423979863,
 0.7740725126622796,0.8356269885057731,0.29681971624121095,0.5231026120146275,
 0.11452331865406917,0.05044780582917663,0.7447790942533576,
 0.14059378564141267,0.9751495165328277,0.8928831455761651,0.4514386964134599,
 0.2610274330883262,0.5854756313534638,0.24913478278228252,
 0.05367304108904103,0.44980556488891166,0.7336634146325526,
 0.6145151949160234,0.2792184188995579,0.6251434866112772,0.8060246293946586,
 0.8676659605233681,0.2776468151413055,0.5246346172857778,0.2399354579034333,
 0.25107118927219063,0.629362961715465,0.8922218769178692,0.4073367541562821,
 0.4200217875917359,0.7712764625381642,0.6885025083838587,0.02394424796680461,
 0.10362851779307825,0.44089357740655943,0.7965643141948341,
 0.4529195325372376,0.23302442651177546,0.2061953148525607,0.6911611299971889,
 0.473220427639627,0.5781252363602302,0.8473398133368061,0.7750220172372602,
 0.16793521687688973,0.36573839386953255,0.3229251678078887,0.257098116598386,
 0.4488330031063371,0.2340469150485618,0.043703941248247724,
 0.2754184502419058,0.2986905029771646,0.7159461811779079,0.08616607383478125,
 0.7009919973267693,0.5094532594923163,0.604699422337891,0.66692031306738,
 0.4837125584550651,0.8266789990210865,0.20461549955045488,0.8461456780827767,
 0.605188266587271,0.6221911543054539,0.3951775947166758,0.9024531311440485,
 0.48773907876359746,0.16584225412267828,0.4239312335272868,
 0.9628796432778309,0.8390296257623514,0.1009042723895004,0.11441520912681402,
 0.44251453305614574,0.03474444286868594,0.5594155797837098,
 0.6218879228928517,0.020488656024722718,0.8947770756447888,
 0.3826733588970016,0.8839011920855983,0.8228192956561009,0.14342278650100604,
 0.06552534825500134,0.34026462974614247,0.20592027442830918,
 0.19907454127252378,0.3656457438259547,0.7272964063789844,0.80035202133073,
 0.1794516033206056,0.7452964887018672,0.6398075529660275,0.77250281592548,
 0.22806287508074752,0.037854916024213536,0.1236238128365742,
 0.1633354650372455,0.1730614117453989,0.8447464986696771,0.1334924421685787,
 0.4586441620662348,0.29233692009165,0.7217679911016265,0.7785965329090037,
 0.5887363611927733,0.11506941284968242,0.3457463424870135,0.7278885161938348,
 0.7363880789831849,0.19967302117387842,0.651649497466797,0.9439643819555649,
 0.05688544173433663,0.31654992787630865,0.23025516990098704,
 0.4502917835775794,0.5449052606876672,0.2600457864956114,0.41308151127849857,
 0.7710132682739737,0.7096886103249411,0.9391307318749378,0.4049867242248979,
 0.6094027790414931,0.6713609865189025,0.6441886831544453,0.09174621351969625,
 0.5974898490291585,0.36957564937369813,0.48462369408683015,
 0.9248080254951093,0.9663895106973655,0.5458912879484341,0.5995437575574649,
 0.3422532900653916,0.16271823894444004,0.05593147390002606,
 0.2278802882925911,0.23539408367052583,0.6010063313579401,
 0.08603226907750106,0.2576336611192628,0.171187649879053,0.3804146967682258,
 0.5606605250919422,0.8480874562760626,0.02820916385149974,0.9375211857533717,
 0.8878296835594415,0.5176738707308374,0.5111922792946153,0.5783287262116109,
 0.15291507141493532,0.4410068438132895,0.4020246190740847,
 0.022558521110825458,0.31395032215968977,0.9304699790284454,
 0.04621978515640479,0.6120560322768003,0.5920766055431805,0.804229775969408,
 0.4729284998420523,0.2504369114745777,0.10787395063829808,0.3681712692369813,
 0.5704008642283221,0.2768232438751661,0.48946445832948615,0.8045758245019298,
 0.32588387570798305,0.1044642537327174,0.04329043033744817,
 0.13390580114721506,0.9953491720559241,0.44678813171175424,
 0.1870757066740747,0.6983460282384935,0.5908116093981142,0.09504427562920137,
 0.7116324885882193,0.4391564411283254,0.8219858981546744,0.27331334064156554,
 0.8349993308733449,0.8177061464966545,0.26894289826074247,0.9306196094731531,
 0.17830818010082627,0.08011868200426342,0.8929038284273563,
 0.4151008733895256,0.6301380940395025,0.4754280182603877,0.37962804734470224,
 0.4157903849581088,0.44676678946644277,0.10826606158461627,
 0.8451906770539219,0.24047364542853455,0.610677318557887,0.8556740536991624,
 0.8070320266801,0.9017538059509378,0.9710526723519172,0.10699594305284776,
 0.3800282825106005,0.6285552394582389,0.39649835799681,0.8160652452501682,
 0.49029518543358075,0.8393848424899271,0.8898860420193218,0.9956139837539523,
 0.327901565546465,0.36997105621161164,0.6474630698751572,0.7357525016641491,
 0.5705246970970728,0.8539600905884828,0.8355321267494302,0.959330302296609,
 0.7733463502020969,0.7910547517577554,0.6502853405449986,0.734764182695301,
 0.9401485474046742,0.9725258467939576,0.31760899167201084,
 0.18913783895228642,0.3132471856682709,0.48033401000557474,0.683634566129556,
 0.059781373923237124,0.7215952417688118,0.33172385114514813,
 0.8610733365757965,0.028570856961489755,0.8305843805519887,
 0.8639309531940768,0.13327852657876926,0.14338191537582756,
 0.8484229161654986,0.764181693603028,0.9352660516332225,0.35771364914019255,
 0.4219483372369037,0.7951060914777034,0.2261304219370105,0.6891465945263211,
 0.8296063192778351,0.8590787942462323,0.5870572372933963,0.6371514058769936,
 0.9890940580455823,0.8356436900255237,0.2069606160649522,0.381951770480204,
 0.8248088696637164,0.30857712665965753,0.6273211377626465,0.5208828549487124,
 0.045744603169532105,0.4818220755142598,0.13785440703316232,
 0.9411827758626756,0.7706549374480116,0.27108046515806317,0.0251998285424615,
 0.7952261612014402,0.0698759075433919,0.5130103801819011,0.7189632715870984,
 0.18142668595229794,0.26336847053043355,0.4520754062142148,
 0.28741432376746434,0.9206419927030006,0.8697606367254075,0.7286428882816613,
 0.7969232837849551,0.4287026191644585,0.9814070822168686,0.5502460585587389,
 0.12302179548521308,0.4199136169595815,0.9854114277358372,
 0.042894823350467015,0.6177664163928881,0.09010273351135378,
 0.299966471986998,0.7300039101382236,0.1909476282447392,0.7381414409284597,
 0.9803727800672347,0.8276663834977146,0.1583235876076673,
 0.051026490422722115,0.006041801788525758,0.8670678490213973,
 0.039434112995821446,0.4156814500141789,0.706152441196986,
 0.057615206613076064,0.19972118402680827,0.7448745738501835,
 0.07583968089282533,0.4776240167803776,0.7705203866924113,0.7259176496121461,
 0.5701179603327655,0.327725563996638,0.0935860907608328,0.7116447848114446,
 0.19603737324091142,0.6443084932477438,0.9043715321607392,0.9384824917303919,
 0.07715634369987034,0.1486842517520075,0.14566449406973392,
 0.3972876529002277,0.6289721331983571,0.6893911126316995,0.25985554010850254,
 0.49318756032817745,0.9343844862028191,0.5956289106124607,0.9873309308909088,
 0.43582207019287256,0.026579822618370885,0.08069499507135369,
 0.11968698400024891,0.2335936433542214,0.66158043645062,0.4642029096474469,
 0.1537132455143705,0.7770385726656003,0.938242030926799,0.23048193579124165,
 0.943752285096877,0.8992171674977882,0.30961666284478473,0.8041806214034102,
 0.09139805274114743,0.27963746627647945,0.07802755717402077,
 0.3398176812676815,0.1916030751499046,0.9641555150624961,0.46210919178112553,
 0.4030959837882532,0.32281207443573556,0.884067176037243,0.8133871456970059,
 0.20676228032431276,0.9839711724423855,0.2943194874145678,
 0.21480910479018345,0.1111447869785177,0.6271579702957926,
 0.009662491528242612,0.47301846494074906,0.5937188665925868,
 0.932769130861377,0.3736962252639632,0.7454477487927872,0.32992330027647454,
 0.9611224819808114,0.8592549666651309,0.3758285391672147,0.7460498063417497,
 0.5915732178370642,0.9366340884625133,0.886802567293852,0.7077833174506551,
 0.915179331618711,0.6725150384182846,0.7101792601136412,0.28656433124996283,
 0.8039474246844952,0.42212218356062303,0.6239048479263448,
 0.49801704351035836,0.5880186827805142,0.7422225031535687,0.1521535147525782,
 0.7354833410160427,0.6424487172491965,0.8785770569340772,0.9805516535088245,
 0.14989034255816625,0.0015619511075549708,0.9808168084735217,
 0.2425306644421219,0.1261037327116784,0.533094533724445,0.6858219047704328,
 0.9416374101030941,0.9421628788208231,0.3775063015476241,0.5885319568548811,
 0.14577519817186768,0.7667860638318388,0.264211641859021,0.969735705201909,
 0.2313114096084774,0.7862132935393372,0.07814902775849375,0.2318849900610801,
 0.6863732419042091,0.3791337816780962,0.35378857331278624,0.690103873021274,
 0.5226618106872213,0.5576281465198513,0.0032322827343512772,
 0.734795629763417,0.8676698676392102,0.06454993925362695,0.4536360067826155,
 0.9975692835154015,0.720078961516734,0.13723015724594856,0.6024529061412689,
 0.6373987042271241,0.31835624605346546,0.36647223851147837,
 0.25862804596045064,0.11561063535497162,0.5151376280603137,0.927957128091486,
 0.9785461899379031,0.47522630305070845,0.31752955460437604,
 0.21974606357483417,0.9112422450636107,0.9975161619717786,0.7586865647048654,
 0.777645194729372,0.019154972286862804,0.03743264085960818,
 0.7006217913293349,0.08245324166479617,0.15065839223011845,
 0.9108514375009199,0.16269918458241162,0.5672057630811688,0.8109939581078245,
 0.01457047474203399,0.9961968348279879,0.42656862832590836,
 0.6759694354812044,0.21489470870860816,0.8759933144611791,0.8703468964265277,
 0.8479681100855776,0.09740283355269874,0.011827247869657276,
 0.2469221397299579,0.7205562529840169,0.1333806934380809,
 0.042468474187198124,0.6362985409296127,0.49946627003248856,
 0.38861714370034606,0.8203936683559079,0.6927133068068795,0.911153352857744,
 0.9121291786132042,0.6001609135587461,0.1567788639565092,0.29150358508812424,
 0.6216345223150601,0.8593235006345324,0.6191742627821094,0.7851570262728284,
 0.333890513604837,0.3839228661975409,0.543946181577802,0.8547442001057448,
 0.16262046857900514,0.6097654384686392,0.448923993605723,0.7965546410610709,
 0.4762088187936455,0.259043489550038,0.017495035359815003,0.3044302819573975,
 0.4383968453375858,0.8255249376642378,0.38986987332979983,0.4378518178635269,
 0.12272091711347621,0.8267699105734081,0.8358859766645855,
 0.07731673603630995,0.9849754881871011,0.8583458931896553,0.694754698693814,
 0.7018334859683649,0.2940442831131742,0.43968584022006973,0.7695686637526236,
 0.23878391280624123,0.7060727599553143,0.40361907355414084,
 0.14122685887426156,0.5793305740377017,0.6988669244555684,0.9499839365836699,
 0.4216522131704993,0.9873684781896188,0.79667190888626,0.08742725893427039,
 0.3246120057429165,0.7974802989757138,0.42326824596499435,0.1517238443788037,
 0.9024947832283113,0.10750561833501171,0.7236250824016739,0.707563731236966,
 0.8687341723784301,0.7496646569890633,0.12989534170752415,
 0.18654966540051865,0.3658885210527393,0.2753137256766609,0.9334551621148122,
 0.6917870160230428,0.8099409564104534,0.1255402881148352,
 0.008981592789301551,0.9584201361142214,0.23927878723776697,
 0.7785857762244526,0.6971835709320529,0.12867170646628956,0.1492823605176602,
 0.3165827877962276,0.2664747872791139,0.4939848909845883,0.9980220473046948,
 0.06609749846275703,0.7395025737716594,0.26735420809000676,
 0.07774134305437064,0.22214404208342953,0.19633658156851708,
 0.7329825022677948,0.8494799391133133,0.7894532408108562,0.8960918888543801,
 0.5622545849921744,0.057291161139950364,0.8233187936457746,
 0.9495076261211152,0.6582347863665848,0.39158828540477786,0.7792375531359778,
 0.32426727513719045,0.5416769858816896,0.6619206662875282,0.6234499316242442,
 0.23059470709188967,0.3203213116529282,0.21058002862623104,
 0.35751998718015865,0.18814990579445467,0.1331863638288333,
 0.6868877048805058,0.1232371278039025,0.754939924944136,0.9776561991084688,
 0.45042691516195865,0.6148176662440927,0.33843601927930544,
 0.049237829059376015,0.2589077366482768,0.2582682285569343,
 0.30526933054584604,0.20808906660839765,0.43249815917920054,
 0.7786870667701656,0.37495815803340937,0.9931599339756385,0.5619251832963899,
 0.8385006563254929,0.3488671061258031,0.4555344401431134,0.33021433988289606,
 0.017802313167302297,0.6816826136498113,0.36192096923875716,
 0.062239230433096426,0.6726737299688653,0.023154414018624947,
 0.7856156197126111,0.5009063088389845,0.6157972100548859,0.9995472909594671,
 0.8157950419109291,0.9982284648183859,0.6037079881985834,0.33316769211001296,
 0.534215046910715,0.04935820265557156,0.8085248049057755,0.6881991297378932,
 0.6276540908701268,0.6464642050399967,0.7367208765685446,0.329050049430196,
 0.04538133941669231,0.6175903036379553,0.8589298712876623,0.1637305056574292,
 0.7478625279389131,0.5204326531165895,0.8115425039447011,0.39282799800036117,
 0.9437387456097944,0.3722788209588048,0.5629661687764169,0.5925462066627447,
 0.3104817303431112,0.33087546720415295,0.4896449885532286,0.8708005784772433,
 0.06389041256525096,0.5764597785355493,0.9713396059853732,0.8041680585758351,
 0.9740745043797352,0.6043527701105462,0.5953182458277059,0.1924741157514216,
 0.8666284981532042,0.8348375727453373,0.29037891263265436,
 0.024690656432391656,0.18101197275379044,0.904940687012665,
 0.6051222067391815,0.7481871369192494,0.6255959582571283,0.2262913308694947,
 0.5309996865447165,0.16461619251434978,0.023184932229045274,
 0.5458710310865587,0.6786116248774292,0.49147702502781554,0.4941861820365899,
 0.4255263305959427,0.8440192862590344,0.6108673558628113,0.6206554680124173,
 0.701725374064364,0.3551498784099837,0.6691261241434474,0.6730745407879892,
 0.3657817701470657,0.3157740429070821,0.7601272520673379,0.7739424816880085,
 0.17162626487875476,0.6532069847348274,0.15017988397399396,
 0.08752503961748237,0.38381973230334576,0.29004868718723953,
 0.36250403226350947,0.8459192481764373,0.5627389747964717,0.7621591329059294,
 0.17582765271757694,0.5109368925909217,0.3575818718564624,0.7910494441040412,
 0.4976116306591848,0.12139869214259058,0.18313517927288336,
 0.39153341490512794,0.6345929495803148,0.23949319005735537,
 0.14372151078619555,0.09965720215929907,0.3134365567687529,0.68502868835687,
 0.4825874442699838,0.6520071034955627,0.7505525558008199,0.7801772175563767,
 0.2678013106890804,0.028009518219086926,0.40215490741744575,
 0.9908951350546193,0.21994103948705868,0.6969557739481528,
 0.02509171829226653,0.6825932619885974,0.24815040020852763,
 0.7755521375033794,0.7575006708215088,0.3827076904114821,0.7698032246007271,
 0.4539631992847939,0.2944464390692324,0.05064492183882796,0.5112946325593065,
 0.15038253742804597,0.5992030570623206,0.9265893254114097,0.8782041312757333,
 0.6629517525030637,0.49731840054244625,0.4066389762770136,
 0.14431185560731707,0.3003119203088307,0.11625715968901129,
 0.34951470511284366,0.5512137961307813,0.32753003409925463,
 0.5339562484934118,0.8666968579006273,0.6608347084157032,0.22359036303824797,
 0.9019377443780663,0.28484888519802365,0.9863434256754307,
 0.14219084702544682,0.4123393049377996,0.1792373299531489,
 0.006644484744092849,0.8675745926970904,0.7085872687130821,
 0.2083241075081993,0.7191081377742066,0.19704856315886143,0.3200098909068677,
 0.26520460905279275,0.16415430186385338,0.7635596256813854,
 0.17534178064978767,0.4157166917927013,0.5660513350212457,
 0.26385150415564296,0.5408475888986045,0.06356630579706213,
 0.28787973908097464,0.2948216955825922,0.16183708387036633,
 0.6036224618022553,0.4721582328698807,0.848234397144894,0.7630432203218394,
 0.7208628854633322,0.3223698757904295,0.013066962716041575,
 0.7170812716657757,0.3077485678505617,0.017908849474449928,
 0.7142460575253206,0.4042251479212302,0.7853651395462057,0.3003455588986277,
 0.35465178545729703,0.3578250454984515,0.9247362480246673,
 0.026534532862245608,0.325794255224478,0.28533360407914354,
 0.5312542651593429,0.03245019944158323,0.33362603193462537,
 0.6142507419106388,0.2792200300425989,0.21760360434422288,0.9685075298444297,
 0.4098820837669366,0.9962622741945752,0.7105456882719152,0.35733718293243166,
 0.12435246335312689,0.5733849504092943,0.9722534791993418,
 0.017872092308426013,0.8227999200405031,0.7090908743103606,
 0.11395607436849686,0.8149139114345508,0.7360648404499291,
 0.34798659231375595,0.2759421510134623,0.7544116914141512,
 0.25332562867779895,0.13401339409845536,0.577831961875144,0.9056999378262005,
 0.8919484587381062,0.8193897995255319,0.17659626081119417,0.5258773522030291,
 0.8668807628396759,0.7293711169036633,0.5285111073720645,0.5968736725291591,
 0.11395003144511229,0.5580143451792892,0.4343248201967215,0.0449346002270028,
 0.3531371851242854,0.9464634294217529,0.3706053609714447,0.2110570769621768,
 0.44316266529953896,0.7646836125356352,0.927427222559134,0.4468720786334087,
 0.08158100365252274,0.43013638054284464,0.4466272198278325,
 0.7406077653025217,0.0926110602933421,0.7347482255424962,0.5950199435406744,
 0.09979356151213192,0.8428446959247383,0.7815390214179683,
 0.06575372634010845,0.6440989829208351,0.7604282489174158,0.3351044348004898,
 0.5339722080744673,0.35751480514481904,0.4818078652900394,0.7963082427832152,
 0.5350483496609533,0.9097962394376848,0.6600388313704647,0.5859959520261603,
 0.9893569143115002,0.5426752657725401,0.7336797437976191,0.7424828476435863,
 0.1983250816800709,0.13482085093137586,0.09567933360808256,
 0.7422229489636798,0.4386641685101541,0.17377512449664756,0.8499295627854614,
 0.6933007217119211,0.9583775508632894,0.3296641395227037,0.05485536609096342,
 0.33573186453443216,0.8709273569779786,0.27627130820052226,
 0.12736479967509384,0.8730203247536058,0.24767779597948925,
 0.02526696896795566,0.8123451581588669,0.5627596521947141,
 0.18286222332076574,0.7302424157768288,0.35446925507236027,
 0.01641176163323743,0.9039556727035823,0.879133157154718,0.9598481264671991,
 0.3728376439737422,0.7521147458819797,0.8244599231262024,0.15721058802654586,
 0.262124441331014,0.9330900296256524,0.5025787582196601,0.37240996970046536,
 0.4336698444547148,0.4121764036183365,0.22084704389072163,
 0.25651819784486807,0.3792948739515455,0.8394949978341413,0.4858152109804067,
 0.09663953301309292,0.4718921500882236,0.9097031227210904,0.8938863361935305,
 0.5847730533716544,0.7377671336678929,0.9507193274032494,
 0.013638859501331702,0.23246328412871364,0.7131134154747617,
 0.7694379739518431,0.7245534953340279,0.007064488351036569,
 0.8295497858409918,0.27837645261553634,0.5391707349963976,
 0.08475406210717906,0.13106462989306888,0.41345801682494143,
 0.6200923686987356,0.23366315222668965,0.011752994100304504,
 0.5488712773160886,0.7689609537020236,0.5421478430417378,0.5235689995380302,
 0.9631497235176716,0.8342994255918841,0.7020366613990497,0.298757430834497,
 0.8395597616661918,0.9872844931173601,0.3370877946645343,0.1415405262254028,
 0.12072216724171347,0.10439674723645098,0.6437935865541003,
 0.8303102154317137,0.5216911742839405,0.6030497498704012,0.15557113709248915,
 0.17201844550795586,0.7174845449683889,0.03837408226393024,
 0.0864742009868652,0.3210480622568087,0.6217623383899581,0.44773540626784925,
 0.8484113932595275,0.4439568406741381,0.8812720485252277,0.6227598066110125,
 0.2834502540983368,0.21203864159561214,0.9351062127589347,
 0.14906938036144024,0.21289697326099066,0.22936684855821488,
 0.7883362380749482,0.01745350353887498,0.8000777895460708,0.7432001363303752,
 0.5478408579431747,0.8413683306990436,0.33293890473837995,
 0.36666221152726064,0.6118140294175313,0.7492968247012968,0.7523783108787979,
 0.34574111415253095,0.5706080195890396,0.2662980723570153,
 0.07037042142424654,0.8736850191626753,0.8716495710671837,0.3453690842263287,
 0.18330903929040643,0.017618284725005817,0.07115197696656428,
 0.19393433021500728,0.5639007308224993,0.3385687776583637,0.2891244258345035,
 0.21952431126915928,0.31243118538272485,0.8721268992689899,
 0.4832946868746364,0.7499672374049378,0.5846007224281449,0.08663848393328655,
 0.5474621532934908,0.9325524977951735,0.5356004139817458,0.5472882189097951,
 0.18325172529412892,0.5080233848177047,0.81844771905434,0.7447724238665259,
 0.5330836623219015,0.43250219123002953,0.47664835389231763,
 0.4471856475731516,0.17599611496710466,0.5487179588067936,
 0.031021990492809626,0.8783470464558892,0.9472687973914539,
 0.19895558915055733,0.5617665724962336,0.782522410638796,0.4984333312310665,
 0.5404993006545581,0.17582168635298645,0.10803084361619475,
 0.5804597920951597,0.048340363967731226,0.35457037749034925,
 0.2594258284580002,0.058631646216670386,0.5925826865524353,
 0.5845176093715582,0.4163166279219246,0.4330077231069689,0.39387166354853376,
 0.7939228081192218,0.33246703514954956,0.5655450798464372,0.6353982996889709,
 0.8196423828174075,0.3932500269051922,0.3003032406956361,0.29424004442072604,
 0.6544482983737127,0.8421848137815485,0.9354365769891858,0.15240160768789413,
 0.699738180398748,0.33598603262541893,0.360929229431993,0.4730134323590677,
 0.7497321679971171,0.2797129687889056,8.869490450258422e-4,
 0.9608307879912215,0.4551056138845435,0.35411474055439707,
 0.04393245322129413,0.7757762202205477,0.34646403851021124,0.903592928915954,
 0.8488547533382957,0.33913500720571965,0.6863819819869181,0.3380147078831953,
 0.035822798193721694,0.6067415985122182,0.898728219019579,
 0.40451282339364514,0.38119438926654003,0.4825737471129542,
 0.5032914484532958,0.2199712427433539,0.08206744193806281,0.5436082995121817,
 0.06951378419177845,0.12901209745675857,0.35320573542145084,
 0.41269084017288904,0.014519074931399367,0.46343527193768186,
 0.7294778174661296,0.8412372751189321,0.19722586893777216,0.3925669086508543,
 0.086592315270131,0.090926361927261,0.5641240440014386,0.28815477451653726,
 0.7222493065123273,0.14038782376051795,0.6062911917497178,
 0.03117261198617749,0.7048908027795837,0.19529974607931322,
 0.042534691021118065,0.8603156407521305,0.6576224852621158,
 0.7611194248914281,0.7388692406178424,0.5439193018753854,0.1665095916945083,
 0.4723407231531984,0.46006531814918405,0.7276075915077653,
 0.015333723818995892,0.7861132206578065,0.4530899625517678,
 0.044576378424842966,0.5825162475528552,0.14689094513372258,
 0.13566901146237798,0.526913299808849,0.3204804024903165,0.5929269294251733,
 0.004563307909815517,0.5078782839837893,0.9102982033658352,
 0.9634251075439441,0.2601937128989741,0.6257425102459748,0.46049545387723767,
 0.0426957311802193,0.12236299009036355,0.2135319060165658,
 0.26450460345394566,0.30478316453737087,0.5337744939812434,
 0.8677053281152397,0.9032679138599211,0.26297258252097533,0.6160689939160965,
 0.37454058436448545,0.27022677700269326,0.8223239689084786,
 0.8485747678417828,0.6404777954741709,0.6572187385217874,0.2979514595871552,
 0.21957352768306637,0.5267482176236847,0.08086782250024949,
 0.47682030834237654,0.45481380147848727,0.17234222182197978,
 0.14010779054840683,0.07348068202587976,0.19445532024097323,
 0.9392186469950222,0.9611663535679095,0.3007313338398865,0.38099774723537894,
 0.9645364091579478,0.3585152959017369,0.2908070814295429,0.8917965646223012,
 0.34386044442397723,0.0665367896382395,0.04459040016748361,
 0.2483145372552703,0.6548466519421796,0.056689937391669276,
 0.16289452339105626,0.4654694674304143,0.25915416023530713,
 0.8402774835562066,0.8123298167746219,0.3825847890446411,0.8669652834676114,
 0.647915637431971,0.855933936955969,0.23315244204594776,0.8435084415856349,
 0.20385445537286495,0.9217513011584179,0.7271359033913777,0.6152168677339748,
 0.5030279027746499,0.7250655626103637,0.8205978221839059,0.5191678882209633,
 0.14640820658675946,0.3277683341454748,0.7046796058965596,
 0.34014232793850285,0.4904029835635779,0.753881429183587,0.2625312284263557,
 0.9801008896029727,0.1982436260861663,0.5582239366963395,0.7370922358082348,
 0.011603992982993194,0.09450774500669579,0.994220507110476,
 0.5412276292973802,0.010508428756803356,0.6592545938783406,
 0.29832874305564827,3.876728494593973e-4,0.5861054549352587,
 0.9803260211572437,0.4015380090768256,0.6099251667671414,0.1931509888396501,
 0.1221007945762923,0.7406677106564757,0.22849784025948494,0.7660232546086774,
 0.3416978137029826,0.007173111714762537,0.42595763362610806,
 0.31814415416821273,0.12944793628752438,0.28161920242223415,
 0.20097493671685374,0.49713585550482686,0.28591558071110545,
 0.2135232956563673,0.9122758434510171,0.6606898614448327,0.8603798689212506,
 0.14429388967207246,0.5083322379382119,0.8298361441293566,0.6606174481332836,
 0.14590135729644726,0.5351351756414826,0.5222526067338564,
 0.010741663122725287,0.32150407164907857,0.2637952317537057,
 0.9158749482418342,0.11582537105872315,0.8077054790151595,
 0.09371009339832892,0.7680427354982309,0.3674121157835696,0.755042458835136,
 0.8422600868584923,0.3739858521881244,0.4096378832424624,0.6883284545152438,
 0.2274184143350687,0.5739145670440087,0.7721591566962205,0.4473105361675056,
 0.847152509378988,0.1825926358136244,0.41210930770464405,
 0.030590877203684586,0.8656559454077263,0.5673513609793416,
 0.1267383201549176,0.5241586080881516,0.5767873483366962,0.9939402645123395,
 0.6408917541172465,0.43432543160005777,0.46337030702631754,
 0.8008435696084952,0.23490979260892564,0.4340658241297056,0.7498613526006636,
 0.4077006446108793,0.10908767563920119,0.8385239312617156,0.6027525330850224,
 0.27237468105688967,0.8997940330532428,0.18352282146540166,0.860579199859354,
 0.06618947725957725,0.6787107587518589,0.11714877443688576,
 0.6654716858773266,0.508628981281281,0.39593335510836525,0.9319202878008318,
 0.19656071503964256,0.8523403467998565,0.9504030087969547,0.8511355006080795,
 0.9070745705491322,0.7452999034437571,0.14480182729761504,
 0.23663913339585052,0.07242490985228389,0.019724439267682747,
 0.3542287284915122,0.561997145839193,0.1878602026637901,0.6243608048190307,
 0.9476603112754348,0.5410316566960763,0.10457772695962031,0.8011688558574841,
 0.8848625373062289,0.08302664120830716,0.8417013019572563,
 0.46400501344113576,0.9669327880669728,0.3915178614936611,0.7765074274494784,
 0.7820700515843693,0.5148068548034339,0.4794981540051002,0.19426026021816822,
 0.23411732858699663,0.46653312584387097,0.3299373367991235,
 0.8269907563344052,0.058937259678292,0.3386823764192931,0.18118559836000936,
 0.6797790530555008,0.11817327332890881,0.4871884397270215,
 0.34922263227739747,0.40677329612482027,0.5079434808381181,
 0.9544912691284811,0.2958477409171111,0.826569271690565,0.3916905093332991,
 0.3814713937061348,0.9469712552971998,0.9124375167864858,0.46400695668815684,
 0.9071304377601604,0.5617428808267992,0.4153445516236479,0.68502067531382,
 0.37832640445637233,0.18156406151283255,0.16959130376494624,
 0.6810101519034016,0.8000016896532162,0.6656533802941822,0.9853746542439347,
 0.8756694775122831,0.7646139076034572,0.2210611044794596,0.0674831344710688,
 0.4736497998574618,0.7989853097386312,0.46366414375022436,0.554450259677806,
 0.457907578697452,0.047224622564502994,0.2389788395604,0.8935726730926901,
 0.46399509824954976,0.5525186183652757,0.3993643308796384,0.9321348060706072,
 0.5405687559157133,0.583148253300414,0.5617931396032623,0.49089593178181634,
 0.4056608175909141,0.49241719758833224,0.10373519306686285,
 0.9906988767601762,0.2427066262405032,0.2053328937982315,0.35796216419591154,
 0.4776683654376085,0.05406622704035424,0.40858897316955156,
 0.37263176556870636,0.5075570175137252,0.016924817593877872,
 0.1869525163584531,0.6603977031332838,0.5619990718416793,0.6617317387753773,
 0.45169707889288646,0.09724625174139978,0.7506671461309515,
 0.8468659437671455,0.8939959736781297,0.21644777530111425,
 0.06821985044193946,0.8260737756245735,0.9216187514980532,
 0.24044079124028228,0.8944699158391681,0.2922425943347302,0.6303109503829822,
 0.30663084979383315,0.10224779403501771,0.409958244738848,0.7226331426798817,
 0.6630860030247677,0.9867575789913255,0.9849912222109861,0.7983999632948252,
 0.3265744829710334,0.14877839487005273,0.43966760641450964,
 0.9094939992363131,0.917323334195141,0.044019721556026736,0.8814914879873429,
 0.17174574944584353,0.4056206785500075,0.7859033341117124,0.6993806678235489,
 0.7655586891418664,0.8156205130743532,0.10041974822746835,
 0.028318128514007945,0.6470440506348123,0.37991012414623737,
 0.9814132574160135,0.9740977357931606,0.8806966384035497,0.07321877856862702,
 0.08017745811043298,0.7310895896290063,0.28429370241923557,
 0.3727537750031865,0.7933039025402996,0.6022607548643946,0.05869740560200898,
 0.5705039957085182,0.7152571468986693,0.28362734881386276,
 0.11855087709696654,0.559274536726735,0.8963102984463771,0.40770314581632383,
 0.7871636686811931,0.22454290136595234,0.37365250146765605,
 0.44704459322740386,0.3671708234316926,0.9474787073546854,
 0.31106649142577825,0.8272380306455496,0.1803265856572176,0.6065031658269269,
 0.003726531904988728,0.025746868476988483,0.26436793474998876,
 0.41283845969081223,0.894673094327539,0.148244269628246,0.8470420904057616,
 0.050219305795249136,0.6532019739745633,0.4512171170272292,
 0.09875925281067355,0.08728460382114245,0.7572251480555172,
 0.17989452212704016,0.6085582289934249,0.08565156108126182,
 0.43967641298403204,0.7286835434553965,0.5118411809565528,
 0.10227264099987532,0.10542332938800725,0.7209630197727426,
 0.9288631213313407,0.5574909176434025,0.5530036119020836,0.17627086335577735,
 0.24702358061452778,0.30801855061346173,0.39919861828678804,
 0.08514717330272903,0.4666179395558978,0.9546860282213451,0.379983322654052,
 0.5436347153541051,0.7902268287050318,0.8470918608995628,
 0.016015232088228526,0.5387061490959422,0.7851318671220331,
 0.38056701333893284,0.2821124766592993,0.03830036210837584,
 0.6640087790050757,0.5503303915273635,0.3884013939093869,0.0980588415416701,
 0.09634987736705036,0.626849228444327,0.7185110302768782,0.593791086914839,
 0.8698265281886929,0.3945317261663094,0.13189330315348446,0.9597463805260439,
 0.10683715727692611,0.34302483768540526,0.6505394758853491,
 0.39757301243594223,0.9926037923625127,0.8620747313480893,0.8496034898064961,
 0.39572322572150487,0.5302237810052635,0.9177262067520479,0.7622341366017255,
 0.08503841031705917,0.06257796375870239,0.9560891794778685,
 0.3121440503398406,0.13851443222763737,0.4481135314948351,0.6725486586213004,
 0.999138161518941,0.8284852147613395,0.09662618745549201,0.42319458060766324,
 0.3352834228843625,0.6523879514879186,0.04694809118522203,0.7797674636023875,
 0.6435721894012957,0.4202015804539416,0.14277465635412534,0.645128662861013,
 0.7363439119062304,0.4099330055635514,0.9214084405276224,0.1989656625701237,
 0.9933926600360771,0.9843187358299729,0.6724235396083806,0.20596647397279133,
 0.7663320293249054,0.1671286367703192,0.4815426850225205,0.49936117599044805,
 0.038093309152712584,0.7612525967766459,0.9893611002160013,
 0.6280092307653209,0.6749119118714532,0.28297907521749555,0.6574829239604716,
 0.6324590565309106,0.05929076968762259,0.17891945200018955,
 0.09923515537115257,0.48133969611391547,0.7415633725318694,
 0.24302677909666126,0.5649016408425142,0.08351611186393781,
 0.49159905067402676,0.7064695492490178,0.47674748267589373,
 0.5697036077545113,0.0797063879855413,0.7601420449630312,0.7577426254862123,
 0.8655820805547706,0.42652198224274473,0.22498654542990093,
 0.5821031930002788,0.9901956250704602,0.14573007152552941,0.9012944016629585,
 0.7508639010605206,0.20435913095655278,0.28296032258342474,
 0.10103940379707166,0.39879612064205205,0.11773997456540775,
 0.34987015836763335,0.7084917398257005,0.013167461765901223,
 0.8778808372589194,0.2891894957801604,0.09854121124009751,0.64157743547085,
 0.7305302106434267,0.07127091142593156,0.7381514288873948,0.8889486524183083,
 0.2177832602553822,0.24993024068653968,0.7651063529032376,0.549639794763996,
 0.30989075508627195,0.6109792094521642,0.1781845075839159,0.9903295829780617,
 0.550183670902142,0.9685379265413311,0.740470025417717,0.8310345802574585,
 0.36647598530895464,0.6879050298371078,0.8974995103813499,0.690513638710375,
 0.8255250398091657,0.7742502583192299,0.6252036643540788,0.4660294538204358,
 0.8688232187818157,0.07555860062122577,0.6400679234911791,0.5265768503920225,
 0.7842975988600931,0.6570164344918531,0.9446701108187885,0.36332988478360606,
 0.9232058973098829,0.8401272531482632,0.2745766356097018,0.3136736149322057,
 0.6107208204798049,0.5270379757026672,0.7485853541331082,0.5723582100238415,
 0.18502191793470313,0.10269278664837111,0.4849416203792277,
 0.3282935946742439,0.3376090854644068,0.08899858029060592,0.7285322992542062,
 0.9246829822091859,0.3926802875492783,0.26185818004299743,
 0.008409868144037769,0.3635714291830563,0.25379280677461624,
 0.3379869665724442,0.9199060268136097,0.8246766841425412,0.1507843355084586,
 0.49448774591616185,0.2037629647763648,0.18513331520130394,
 0.9338626031097027,0.5136746321056231,0.4732198086698013,0.8251327217261921,
 0.30681277776973204,0.9143065904630348,0.27663382426008554,
 0.22540057181594253,0.8992357560971014,0.6179466799185552,0.4777886005644141,
 0.458841352961635,0.3355123385979045,0.7128707543857936,0.08588161807710148,
 0.06712973520321652,0.7533784759718416,0.6552352114278623,
 4.65852624248142e-4,0.23151387928846012,0.034612803381322554,
 0.5591588092315007,0.7862167066496454,0.8759650771422755,0.05580578717479712,
 0.9067716671689752,0.2486126945826619,0.003567301045331095,
 0.7589068291045237,0.4387450416547184,0.6684680819838953,0.3184638285429101,
 0.06364213021128173,0.9708653245233048,0.48992126764012367,
 0.19181484480608946,0.649107012009885,0.2678014755659849,0.7552105293658855,
 0.9799421676721537,0.8393277832348714,0.9991146595008391,0.9306602990536537,
 0.07666961175011766,0.898057362004898,0.14128546992839253,0.336383158003415,
 0.022191355334758245,0.2423515744804522,0.14960254483006563,
 0.9289796831512108,0.38073138995123745,0.5039481820066316,0.3425967060945575,
 0.39069791086463523,0.6288946021652527,0.3805601462387309,0.8542024424579774,
 0.3752799364084968,0.19480413521841833,0.35558653858533373,
 0.3896443463443654,0.053919884523758066,0.08147941065889841,
 0.08269352996998236,0.5367686952827041,0.8024897259167026,0.9664753471997535,
 0.48921424029980976,0.6402862888568297,0.2504657612935093,0.9268211390567784,
 0.6007226296634468,0.5112004408731827,0.39972126426656707,
 0.40132663402467017,0.21189300065111716,0.04434619063508949,
 0.874411874823622,0.4155322485727274,0.6614379803752596,0.41581291837922985,
 0.929655277924041,0.8247597029806094,0.09970586645060464,0.5123267782121571,
 0.6884663654956609,0.21122698626394754,0.49352989932608204,
 0.34052856476686844,0.634648465717596,0.24242911024102443,0.2699417327115038,
 0.6072414569083051,0.234974724806157,0.9307429331365551,0.4099856668990659,
 0.3636123961394828,0.05588454708495105,0.7601662373838858,0.7496747320991319,
 0.3213900065180302,0.14635962527815838,0.588254950038533,0.8563702659386674,
 0.8593258535368825,0.11701089972245127,0.3758700018414386,0.5940391952276335,
 0.8417695338898682,0.671614521817375,0.5413172810329263,0.8759887883146592,
 0.24930759597579155,0.32212486324364265,0.7194328022884424,
 0.5294665884474405,0.08823087451280887,0.037588970095790875,
 0.8404240564847539,0.20158777209923395,0.14947881620440506,
 0.28913832717183074,0.14776612573179748,0.7175551594155776,
 0.13070971638633366,0.7875897761348526,0.21591453427406682,
 0.7059809498936764,0.656884001511342,0.25095403874678845,0.19528461095126082,
 0.3048698750482356,0.033947775015792936,0.5441384478897088,
 0.01042667858157209,0.5089867571302086,0.2279297817200664,
 0.45600324928140246,0.19616173018890715,0.7596879853550585,
 0.010833522815269947,0.45544768461550156,0.8578875557705011,
 0.7108495408128022,0.37049633803354975,0.1620768839840625,
 0.10633228348369406,0.3489613207419717,0.25222550466199545,
 0.8705339920534682,0.9534323602415391,0.30848631978182595,
 0.09079314810937167,0.39971264800822404,0.7749456546339155,
 0.8455824326917849,0.9768571921148854,0.17070563586587217,0.7535676606958385,
 0.5457648353856127,0.9229535049233663,0.05304770935040226,0.6957644998050105,
 0.9065687387144771,0.9136403257645692,0.21659388604195817,0.8515201508031398,
 0.4914475140464185,0.3140109244337802,0.9130596472190191,0.6168344154249991,
 0.06716431627749042,0.6786301063630898,0.6788671860697774,
 0.33702061773984415,0.6320815105822496,0.5407230847019392,
 0.42696442586392214,0.33143643232165243,0.4797472285935026,
 0.5039240761704774,0.22660165211853445,0.7359267703337004,0.1987856331333414,
 0.3666123155090151,0.6351492040935566,0.8702820764022995,0.11154625848090938,
 0.15200589902991346,0.23916808809646706,0.7452183453180645,
 0.6384593354370334,0.6392277039035276,0.01891948016049194,0.1888446578247146,
 0.3007042631026712,0.4908701176272884,0.6607501538332621,0.4713607244969943,
 0.5984733813167501,0.22558830965289556,0.8642426506046673,0.1756495873017242,
 0.19914653629695256,0.19278232756097702,0.5697211502370971,
 0.5905471483925652,0.9183760590155019,0.6909283796018793,0.8692026278206392,
 0.3864105035857549,0.7513569554900393,0.9900852509091502,0.8114042708540216,
 0.6528297818010506,0.3035716235999508,0.36570956100776464,0.7411023060410682,
 0.05418230145793701,0.23610280743087486,0.5031484571163551,
 0.6169975855025277,0.5657414713391548,0.01660957997531709,0.4838550206186153,
 0.33599585658331355,0.8350526368571949,0.864846936342487,0.11367272814347917,
 0.42087079979597475,0.7345872952597714,0.6975302873261993,0.3103485959019422,
 0.24734099882911742,0.8716966611928368,0.05062439291711707,
 0.7967522839797251,0.7192461381218592,0.23885018800963354,0.2024806726552999,
 0.9709638161627453,0.6527086982861889,0.9619983581956395,0.7872282613221722,
 0.5870922137434833,0.7014909460701924,0.20138637136737036,0.4423157722502107,
 0.9970770647861832,0.4327649213535598,0.6905888637292931,0.5595777462493823,
 0.9011992615562727,0.03173997161989217,0.9643800292785382,
 0.22211978140406508,0.011573409029238535,0.6040637347090196,
 0.02961652153094385,0.9350604883575665,0.027431157721022448,
 0.8198980373332982,0.8564112683005661,0.9569200393590598,0.9130276701556926,
 0.018525402169767946,0.9402777554011608,0.21477341784782245,
 0.6162193626807875,0.11883136919137349,0.7664164202034763,0.502606556907213,
 0.1649940913254887,0.47125596343727494,0.8767077106065361,
 0.24673721844676655,0.16775572116290016,0.8327804623269658,
 0.16663056248929375,0.5129977339234787,0.6212985629003941,0.1873308914946198,
 0.3513083817622794,0.7725914707195978,0.19231066196330604,
 0.34630468950091486,0.9184879246998678,0.46872832376476414,
 0.8844487373513851,0.7136882221185318,0.6004478723301815,0.75318938137778,
 0.8959734193269352,0.33565851933169066,0.5603145119097666,
 0.43005245998189134,0.7230979611387788,0.9484957989221118,0.8935852501931238,
 0.42175796060801163,0.8357303156755748,0.3124176647565067,0.7454084314483719,
 0.47917540340539044,0.3012689255701423,0.17385574146082616,
 0.17131974802682903,0.7848389067509396,0.9277148349923365,
 0.03864671267066744,0.06214246410440871,0.5535574431328054,
 0.7575552291493061,0.4659493066230743,0.5017769799664181,0.8503603388540751,
 0.14141006847531734,0.3788344422006934,0.34013399476005635,
 0.43536532337491796,0.002979650741565476,0.43363152618649126,
 0.33670566760305876,0.13227581670883815,0.33388143757007527,
 0.18317078179029755,0.2527216525392264,0.8601840342207173,0.1313652925016794,
 0.7020250155843283,0.0709139633012077,0.06972889425449535,0.276100490683863,
 0.8877981709997932,0.95572526033646,0.17841401560977543,0.24964878378551236,
 0.728386526459312,0.24815387195011074,0.3030355531018625,0.42822056017458765,
 0.6049175797162162,0.6301188351799832,0.980563128496425,0.8934301434045794,
 0.717280102040556,0.8775218053028676,0.41824578133476753,0.8980378773674873,
 0.6713217288767434,0.7028581458832596,0.41751904499125636,0.9611474965405259,
 0.766705427500556,0.9188650199076919,0.8240251069107061,0.38042059538644535,
 0.0702630657134079,0.26739353146713596,0.7303375050230637,0.7047975603381519,
 0.2689537282644916,0.8940239968241384,0.5114936175543883,0.8652430255897648,
 0.4859508685650621,0.6953788669900647,0.09824014780784385,0.311050533945896,
 0.5154409283865257,0.1256445016133333,0.10414807786465896,
 0.06414708968280225,0.40512635459773305,0.07294358757285835,
 0.8086797971358308,0.25383315746569857,0.924875465825086,0.7074833096898682,
 0.5108798240195586,0.00600251516305772,0.7019346195968885,0.1658630813802371,
 0.30457139078903284,0.08338331973328617,0.03827457974244197,
 0.25189265274072026,0.8558035877770601,0.9505349752344795,0.676803490478095,
 0.5761629113358822,0.9039287194586074,0.6220304975545944,0.657892080767808,
 0.19618655793118578,0.5024064237888344,0.4786138384984644,0.266866302518145,
 0.13772401162853476,0.8177959877031681,0.21445081535251975,
 0.8074080672167352,0.2698020998049353,0.6755373929016719,0.6562099709899307,
 0.6281079915515054,0.2450350445463806,0.060828345410073315,
 0.028430671483650638,0.3997531288764071,0.5994418555479558,
 0.3895846990370323,0.44464829736157996,0.4001767531797874,0.3162028304756269,
 0.014167924143454802,0.3683516417331999,0.23475900791990456,
 0.5364647647587816,0.571835203100495,0.40720853538580837,0.805986768099401,
 0.4459355083082761,0.802147789887355,0.16772491864054895,0.3790813138112187,
 0.2218296175054687,0.9989742845446764,0.9714663154855272,0.09033861276206823,
 0.635741053713778,0.9867990031133757,0.5510367494592029,0.9523333748352895,
 0.8037955499748122,0.6666815881843456,0.8044024539521729,0.8903236674220181,
 0.8542265419656228,0.8194641444055993,0.12063477581151383,0.6814617224834151,
 0.024319747492864474,0.22267415677146796,0.7022718547600321,
 0.35591257558487277,0.7357701370642918,0.9599979150684834,
 0.026778542639605107,0.6344746993700198,0.12541019943980158,
 0.6598362232649512,0.16368843319621174,0.21016715064284275,
 0.3223585708557969,0.28979412522747305,0.11187307194287832,
 0.35809531716695486,0.09520350401946032,0.9910386738269517,
 0.05934238425852689,0.9625606668332093,0.48675818058738163,
 0.8149512291047452,0.9070087470618862,0.8474166192915108,0.11017700409441578,
 0.541338458301176,0.004682757830807849,0.025729402362253517,
 0.7081650645865322,0.6617644553765573,0.9025195144403914,0.5336982009316878,
 0.7601233489722757,0.9239057853262412,0.20431352569020333,0.3578749124639161,
 0.031238770119276338,0.9819655025002901,0.4508463849520358,
 0.11292083248108464,0.9357922324085972,0.917002851318277,0.6237610396135126,
 0.9182848108211663,0.24803464873684666,0.1400468887130054,0.6246728741296822,
 0.049115230459634396,0.8004441293684291,0.9452969203104123,
 0.45338943610236393,0.21830894196831507,0.6094422416805033,
 0.6029974093275133,0.45140431840499506,0.5639852905349272,0.5676966686196854,
 0.41582380169794186,0.40175093211616864,0.3173692663712352,
 0.6440924442154634,0.559868942505548,0.44744201734738254,0.6476103143507667,
 0.6838371127352298,0.31755792894264845,0.8853531224142718,0.7017312773913165,
 0.0023938618248781074,0.9745474577165821,0.947990727264753,
 0.8042505644745648,0.16483656809049851,0.033137571702519186,
 0.007813363916170157,0.7926519385259488,0.8091811845146814,
 0.029853169672623747,0.5652989922152177,0.933698937257137,0.6373068246246106,
 0.6999168193247551,0.4914684368881941,0.033866063313067185,
 0.38302733118444543,0.7704779482657026,0.2329816655829463,0.732457604569837,
 0.7667695627424768,0.2529752805407822,0.5275793412196145,0.23905858639635302,
 0.37609692899735014,0.22444147805055026,0.12419959434763805,
 0.9360273247761455,0.018540337842483212,0.7250498337852941,
 0.7083771300361263,0.8770921850521092,0.5134741310193913,0.5779129131661569,
 0.36217451870624984,0.8582222450443789,0.706204085698904,0.7106646796123736,
 0.06561326617673124,0.4775789115182685,0.2525118052405625,0.3539189979565167,
 0.4475707057643633,0.9103683718652775,0.47842279284571476,0.6128680953133051,
 0.9856762981462972,0.9970852900223376,0.17684005307865247,
 0.31089580693445695,0.5384494055894552,0.9454590496744147,0.2879092774164864,
 0.6596491782040381,0.0949450465600391,0.5922204205681215,0.6333534508104165,
 0.9608383762050949,0.613486798937579,0.3240443244431178,0.012280968676582238,
 0.15440196788081728,0.03198556178412271,0.8307704723336757,
 0.3619370891678261,0.9556830745298909,0.3926010788937604,0.9372163333569038,
 0.26217779176890965,0.911907523379814,0.9418791459477804,0.7723586835847549,
 0.7362903203449058,0.9661156438162252,0.952274271570106,0.5836068785627393,
 0.25264560719220286,0.42686215165874886,0.663151579741155,0.9429763030363825,
 0.22157022931411452,0.21408099429315264,0.2735411528368017,
 0.27039600864498203,0.4863978360436536,0.9197640962036033,0.6821205186141633,
 0.08115634830144602,0.26197472599908367,0.6664765910177158,
 0.9680280780581216,0.9146563407684841,0.9053319811581679,0.8422314126598648,
 0.5043680462831017,0.7967682940097668,0.41542301341464594,0.1475551138069764,
 0.6623660175249576,0.773925829157134,0.2744771385943314,0.2573906933262472,
 0.4497368879834299,0.45303321236569816,0.10177823442268197,
 0.07119501851947863,0.7268803882575537,0.9127482358645606,0.5610971408721647,
 0.7563217381128055,0.14189786826323625,0.4088040478335423,
 0.03569886022850799,0.8590721195380143,0.6548051838888442,
 0.24852209442983486,0.2913661730466759,0.8200261766629939,
 0.09812367939143207,0.8762688183656784,0.7678130584163325,0.2820987064726256,
 0.035719802990192506,0.2529879052022438,0.6703190954303617,
 0.46381653850475335,0.3457926914424101,0.09212134170522701,
 0.052281391423052326,0.4632941953240657,0.4043135667760711,
 0.18432496673641208,0.6298520425118871,0.3245533838078599,0.7779374422027036,
 0.6311342430781616,0.31874938610822035,0.8824292275944804,0.8403966479969895,
 0.12722975392034763,0.8001754345030797,0.39726468845406493,
 0.4409516790850967,0.8247596302874863,0.6544949853319662,0.324527351750744,
 0.534567309652466,0.6454888553685265,0.10276460664364107,0.25964338003826504,
 0.03097698541424665,0.5977293611280469,0.4225675428953374,
 0.33212441175901175,0.4441663926364925,0.85399688310867,0.471501378835948,
 0.29720790763713945,0.595025372328151,0.053888567130981335,
 0.6126241875244532,0.019959605780436207,0.2811532621621857,
 0.5479907573604484,0.08897067741056708,0.18796796648714764,0.763056218981823,
 0.44010237938752583,0.5680253519082447,0.4207856121179341,0.851883989010491,
 0.28041718814739836,0.5837454282750377,0.7322815065029716,
 0.48846867762095125,0.5333463543582265,0.9454928926696694,0.3500365041812672,
 0.6012526274746142,0.9783155955414418,0.7531349134602413,0.41343333162061113,
 0.999280028696978,0.04341381884602058,0.4620515139390904,0.01527912648464913,
 0.027316687864706757,0.8100354121672755,0.42451464355704505,
 0.6362168171245539,0.2852287787464536,0.08274069707850362,
 0.08150768157211419,0.6245472816812123,0.5058311948562454,
 0.40313193306705886,0.7565957762024333,0.2792340440843587,
 0.17760538151674954,0.07780631103101965,0.775008774668097,
 0.26531805803429453,0.14301844323002755,0.577907912674871,0.8502085872716144,
 0.976460998789667,0.362461325906298,0.4932043389516293,0.8356741740835729,
 0.18938136770904346,0.7184400238548388,0.9565163039302118,
 0.17039635876085524,0.8214510618173542,0.6551346717989496,
 0.009331821291348241,0.8732929880836172,0.477853194778879,0.7850581259255754,
 0.08005131302319513,0.9365340262271115,0.46505084271500696,
 0.8725682483604363,0.2027037197363506,0.1153019693016808,0.16590843692591528,
 0.020349607949180837,0.4556586195069694,0.10558017254568575,
 0.5287201354810928,0.4487302433015964,0.8904079480900905,0.47990984819835825,
 0.20727213124144872,0.8686902743249404,0.12710691217506498,
 0.7380849018938116,0.32406257850241993,0.95152087748972,0.6552421253073608,
 0.8192070637318689,0.05852565365956164,0.01800581645726007,
 0.3271831801336875,0.24000023371102197,0.5594689527027603,
 0.09425202203839711,0.5538734427189811,0.7783460699415052,0.1433799406228471,
 0.1314908339978509,0.11727116816527416,0.24506468574520823,
 9.524590123031196e-4,0.009564176904457322,0.34732649589006936,
 0.5372826642435158,0.591026804788062,0.16231607768091827,0.9742606311750853,
 0.6779680471222086,0.20617714393014164,0.8352426537627767,0.1960174860577183,
 0.3386333809159813,0.6901939377254909,0.967713305369558,0.8596944256420812,
 0.532661099284315,0.3960338232079563,0.589150627670392,0.6058634314792791,
 0.5909185358502969,0.8826035831737306,0.47756137963847256,0.8337375159015576,
 0.12084718549268669,0.49744734956045544,0.5053816302714982,
 0.3979424990974151,0.11240993596433457,0.9302144307350269,0.6870465583594642,
 0.11089251844429049,0.4338765481127873,0.10410750496959076,
 0.9829649077917749,0.7295510589619898,0.71881407338574,0.29438754535706857,
 0.1346806393014287,0.17853015359202073,0.8369482667330919,
 0.28999225902387005,0.14764779659430816,0.940618751311648,0.8583661947271609,
 0.702300543236095,0.884069449449534,0.3414283864946087,0.6585567270478674,
 0.43930334012020256,0.15348690294315637,0.653696921295364,0.6902510878676564,
 0.7270687785895648,0.46939349351362425,0.950698734801258,0.2865566313009106,
 0.4446171899344594,0.8803105120581168,0.17490041668203682,
 0.05023773479895011,0.8773226725116526,0.1342082362311533,
 0.08960504687189541,0.9733247292205526,0.2567335229423896,0.2630134429451233,
 0.24716303623591018,0.5443208967191233,0.39202018467732147,
 0.6422519494006059,0.4303015186993838,0.7515014180379191,0.9545090246894187,
 0.9845333385537655,0.5436374577520522,0.676285857688879,0.09400906125942543,
 0.22684362555904802,0.8511127403407215,0.9924199396882774,
 0.22253340176677894,0.34591890485391064,0.571928560765985,0.4172471529409729,
 0.5034790527653946,0.8016670515598056,0.6223394498223531,0.18343156702917274,
 0.457115070961288,0.9234420647150906,0.8783565532578091,0.8621410657687361,
 0.20845284771636718,0.4409138816543654,0.5144642198227264,0.7969636178426696,
 0.1473737642160996,0.9013686667990599,0.5746258644322995,0.6530833921880042,
 0.1432196602899164,0.6364885028091916,0.1797459706791521,0.27839817825444946,
 0.8548517789136443,0.557431716580302,0.5250862555400756,0.9207028328918296,
 0.11266536393491211,0.38911917564367604,0.676544225474992,0.8251886464361989,
 0.9589589632276586,0.3133462938063164,0.7003261190871848,0.1569828278543246,
 0.37409551042752454,0.9664012226179076,0.036856083413336016,
 0.6825744104605089,0.5286684438653159,0.8433996460734674,0.43230442859474316,
 0.9852444178606966,0.43604018276464185,0.5743580820012844,0.7363026268422473,
 0.216908162987288,0.23359709969550047,0.6932753212600069,0.27988306160363363,
 0.34420952407049765,0.8978771163320566,0.6662062511850851,0.7135827803775069,
 0.7019620960237463,0.5584771984830763,0.7414432890947743,0.5961674156364215,
 0.9181272742918872,0.2775736474298083,0.7377992541646392,0.8263399990038404,
 0.8196663656105918,0.22026402822988533,0.997536687923973,0.20260048116994933,
 0.747461441335807,0.5806868919271498,0.20508354008728924,0.7486111654844917,
 0.1529521976371233,0.5524439033137135,0.7666008244842549,0.41263301891565807,
 0.21989932205659812,0.7351130204921257,0.23070455280204627,
 0.8481657759724205,0.10742674672630648,0.9926073387855214,0.57161641428981,
 0.6721662685487308,0.5622517062265465,0.43810770531426824,0.5610679282760799,
 0.8453089119649789,0.904549934796166,0.8639782096153497,0.8225925328918844,
 0.5860338237700585,0.6019388675825845,0.6096960748783694,0.693265608572749,
 0.21046651976866082,0.15185378721688658,0.7647857367702695,0.598087861163775,
 0.38748071954641405,0.756768618313804,0.957495272095707,0.7522380521945616,
 0.43160807269488255,0.7468906995631783,0.5362815103787296,0.4302064356929012,
 0.28710265443197336,0.3597085077430736,0.31745380622563735,
 0.5254533864060802,0.4078653438985661,0.44876868196109865,0.7895751767602673,
 0.17428959552272139,0.3243575327744461,0.27951611072629534,
 0.6612998173980694,0.0845307351393716,0.8244120460538485,0.06805669998656394,
 0.6142042623113906,0.7423528009148831,6.269159622853682e-4,
 0.40214224899770923,0.18363209884178133,0.19781478178931144,
 0.7242253283817285,0.26922285627845954,0.558100372056799,0.4537022831530657,
 0.6586374112805179,0.5829760414046844,0.13125832862921571,0.6999184112831606,
 0.9614261975571368,0.48957038266669395,0.23722795964023102,
 0.9070724968212284,0.14743519457046217,0.09665659950343375,
 0.5753182514790611,0.09341796201808439,0.17153706631475674,
 0.9387700266800265,0.6726040065591047,0.758150281313474,0.43762991873756174,
 0.11174130927581538,0.15362438296958214,0.12642498498028487,
 0.5967349087561991,0.29357135943200097,0.5833866967436389,0.9550310104915513,
 0.91571303461456,0.6295313471222007,0.18786834719735546,0.1346270883564752,
 0.07716583174279323,0.02827085483928027,0.6682788925922611,
 0.5507771565332289,0.7953292599153565,0.08291259987289712,0.7904512507508252,
 0.005672376466039242,0.8525041909728198,0.6553278661013902,
 0.3379612255423696,0.22123941467925645,0.795221327509381,0.18896415451267512,
 0.6645154142517387,0.5285751165543119,0.7336829196217965,0.9702800679489378,
 0.9618567677656821,0.510382477461275,0.14122460006262427,0.2859021774660816,
 0.8624647095665113,0.8607668229602153,0.5644942627856955,0.4678859299950613,
 0.16269363681258353,0.9498604934479671,0.8788261526675554,0.944583558668937,
 0.3361981018072463,0.9065242852079107,0.061232018432816226,
 0.0873539870682819,0.7307687608203013,0.7049613350262577,0.6148650545478901,
 0.8646029653238704,0.5020633055523431,0.3418045391126925,0.6436379477587424,
 0.6429139045979669,0.5883205971894296,0.9757020292309699,0.7896152312913906,
 0.9768879356566492,0.43836222387451684,0.8742865207632784,0.7673076023166707,
 0.7115836178232499,0.2597668039560811,0.2542081604097508,0.25734232483336716,
 0.20485785612229312,0.3133883710339205,0.7055875935595025,
 0.23171536000628767,0.9131604956571217,0.09415267784526793,
 0.47364099034630736,0.2994657484668539,0.6177249995265089,0.8808151779466727,
 0.42634294173424503,0.823995665155977,0.06475536766317025,
 0.11984727568304487,0.9210323159624662,0.6261450297888879,
 0.35923770476679384,0.8633140555464848,0.970244181023465,0.2850379437371411,
 0.6435393334871801,0.9861888910175772,0.01529562517103189,0.6276106613675327,
 0.8557168042700379,0.3228495727150441,0.25778529000400296,0.3537307356156253,
 0.4222924243934062,0.2880951807876866,0.7562748365300113,0.6856368465305601,
 0.10445506126167092,0.8381255007641835,0.5644799009713939,0.2605148035446819,
 0.4913717426529127,0.11573583812382537,0.8683907076179905,0.423461630495804,
 0.8471895884453893,0.621839918623481,0.48466910484261405,0.819331871214097,
 0.5892332030986639,0.6717377762412482,0.6332310563143826,0.5777455117973727,
 0.5063128439833624,0.2782780839594745,0.9255029658720361,
 0.018508674960173943,0.686341242898482,0.2816510189224538,0.3076561312474333,
 0.7956054643914464,0.4585595815174013,0.1975715540454892,0.39474536909640245,
 0.8086775288134579,0.18286994620337482,0.49304381825200105,
 0.24745269210690035,0.527834724604926,0.7074127251641493,0.23952842708482858,
 0.3061489465788826,0.43353668793514544,0.25470360240103496,
 0.3159918725090112,0.04353071928509322,0.37881188343616345,0.765684740757873,
 0.44123718131877054,0.461277681077491,0.8512868328146906,0.7854447157326753,
 0.04306307594666359,0.4011666414738254,0.2254954979329361,0.3928346298523687,
 0.3563220335283259,0.9271370625954617,0.22621141673362022,0.6888981369346757,
 0.34071182432508174,0.6258971089740742,0.13851517962012572,
 0.22063246079563403,0.46716476933302087,0.8078268107842878,
 0.13526726780342024,0.8774676431422398,0.5832267761715368,0.2137950097517055,
 0.02644444113412625,0.19546965320463572,0.1516555394225495,
 0.16169133497892652,0.9361908999089624,0.8577065275782603,
 0.21214975438437722,0.677349948233631,0.258043520252233,0.46388238885912947,
 0.8756479763470219,0.7287884993726916,0.6567073577545759,0.9878617965159957,
 0.2916544211613713,0.0061139211920121506,0.9826260396997722,
 0.4457177889177817,0.04751980468541772,0.02245913354285889,0.861470476895512,
 0.5339898805784753,0.7515177801938205,0.1763618730739973,0.7552055412879048,
 0.451927646274887,0.6070587361305072,0.8928256273671086,0.948385205079375,
 0.8760001206546386,0.492766364604347,0.31317897669780725,0.05254744618980545,
 0.8126095880219346,0.7878526669666773,0.7685000319584344,0.1869161681322119,
 0.32191216194618744,0.9672783315242959,0.10430650946816489,
 0.12899985427913307,0.8142784969205372,0.9457114308797135,0.4229659568963138,
 0.06610834254328324,0.12880127739176794,0.09548667468847571,
 0.041426310892066054,0.9771037686365456,0.8969150875908225,0.398101103808824,
 0.16741182575881308,0.006002868295428954,0.29529308117979736,
 0.9621957888844921,0.9377010307088417,0.8466210029456949,0.8682196483477274,
 0.9903582479631635,0.8084074052493673,0.7469407739932697,0.9094490370283895,
 0.1875929989084657,0.38182839691167647,0.33852739657146047,
 0.5054743686066548,0.30899162480273246,0.793362351874534,0.753200558983597,
 0.6100038036746985,0.16220103203425262,0.0224599604141279,0.4482166122402875,
 0.25667813497881387,0.49847889121249844,0.3456157132520843,
 0.6578603527618943,0.19717077613008882,0.7937731706295967,0.3575932334503267,
 0.5860970799098201,0.9064728475098186,0.3084269248604047,0.721688739951261,
 0.15351630991487264,0.03642113325492835,0.18068397445598972,
 0.014027255931385318,0.5340343982729919,0.9022011772347016,0.823528894185022,
 0.6113844574315591,0.8455718709049725,0.6294015167205306,0.9476874489025486,
 0.01943540217602291,0.42796894265293817,0.20019188294756163,
 0.37375262161109835,0.7626494179599385,0.7744382725910928,0.7305492480224771,
 0.07493854262687327,0.38656603837893067,0.627729947130076,0.301048092646545,
 0.6202469757114534,0.5168330325186709,0.5691327482758212,
 0.009637458518813813,0.6330712837358433,0.8999761977475467,
 0.9081121361221752,0.8427286074293538,0.4863255833474591,0.8289696090266404,
 0.6085364061502413,0.8349883165221603,0.0066907020668385275,
 0.595403657369165,0.06800287492969381,0.8862264806622877,0.7066630441440668,
 0.9279312003816347,0.42860609319176235,0.7607383350575997,
 0.06751963884967782,0.5716833379200813,0.7832141919036101,
 0.05937262472829097,0.7946080397412936,0.7170228330810106,
 0.02149370204948553,0.3564416218090335,0.22459258036206353,
 0.5514860619653409,0.39306875007996545,0.5711946683431431,
 0.10405486271383135,0.07863658925104777,0.04100804449209616,
 0.643724431477487,0.6122643966549383,0.5488656881239278,0.8722799802135328,
 0.6418802245181765,0.0862563217760175,0.6563985490590714,0.891040157533057,
 0.47721257772499526,0.5546456426813854,0.3629077752130878,
 0.04480683342553449,0.5916139408409267,0.8886631445254743,0.6669376076074328,
 0.015658311544310854,0.08404046603183613,0.3268169558632118,
 0.5558906559661299,0.7200901433472585,0.04563001947390466,0.6308703675644315,
 0.21185944584642302,0.8044352623796079,0.8333152855267588,0.9476449072681659,
 0.2785743520472893,0.04732520590700551,0.6443999377258554,
 0.21655251093626127,0.431066173132277,0.6597273605087879,0.2969148757032959,
 0.46161359897645937,0.8002393612762153,0.34359774681107647,0.543810255280799,
 0.7003177192578063,0.9676189623600573,0.992813325320548,0.9627529227319858,
 0.17197734194041958,0.8575640553688845,0.9985641644807917,0.7027278990693597,
 0.09463427529782797,0.6381963745789397,0.08820253281239487,
 0.4030034988313649,0.7576931918977852,0.5369717283940476,0.7629231894163078,
 0.37791079258804483,0.5215112693682118,0.5410164488837252,
 0.35552925610431396,0.5725517017449582,0.16874398537912427,0.510484182936373,
 0.7836736519549747,0.4376668625170691,0.10817856390729486,0.9872590600304716,
 0.48919581067195017,0.3650219415031427,0.39032014081085287,
 0.8852099040903514,0.2337309540705994,0.04009677529163724,0.1132705276111019,
 0.4654367488102039,0.8034320356328201,0.02371548673221191,
 0.08183585633664459,0.8301956794216003,0.35641446455923886,
 0.08499688596390031,0.42787564539375267,0.9619897511706935,
 0.5815015768469154,0.32822223347659496,0.6007202628055438,0.3427841098529151,
 0.8318802774688917,0.8620545069862457,0.22313203179258234,
 0.18665460365002373,0.7814121702889959,0.6327227245839002,0.1847438466936855,
 0.8619567083475214,0.11123020918803306,0.10110068650105852,
 0.9533391970789603,0.11434191739736965,0.9235221305327216,0.3865769677188452,
 0.4460189580347337,0.5216445228092528,0.6600200103446169,0.2036855223238847,
 0.48762026457068863,0.813293869954715,0.884331603165957,0.8595817921064002,
 0.05756824850315323,0.4759107503989093,0.46547268928511487,
 0.33740532607481355,0.41214605035487306,0.6949918528514072,
 0.039599161601640676,0.06826441056771493,0.11789078470581682,
 0.3347983796291367,0.13170324295048952,0.2629388664063166,0.3144348243678313,
 0.948722814725366,0.43666612043535913,0.5925068874931754,0.20654547662419032,
 0.40841744681461456,0.5042109537452809,0.6011235491131122,0.9761906712721343,
 0.24967026668254078,0.28590405326094515,0.9944161432893228,
 0.1964178216811613,0.2713172937530074,0.6442350157049547,0.8458298582476346,
 0.4622564262112271,0.13404006932064372,0.16010441401182807,
 0.7287552948594543,0.2534066565088029,0.7255219676016166,0.7373295508739182,
 0.4292919631445403,0.8134046842271272,0.2547668393124255,0.8000736984249168,
 0.5559381042877654,0.005804975765929643,0.1643392641975333,
 0.7109711683643876,0.22006545934477761,0.2934604303679097,0.942401096783458,
 0.3690623826133447,0.1984869022849448,0.09254715178256656,0.8520416808979308,
 0.23481098786229482,0.7351552909605595,0.5348678227929642,0.6781153154771317,
 0.9082185148168807,0.3747337340941853,0.5867115575186209,0.8618624781534601,
 0.578856936986086,0.8638220478740364,0.18833450377854144,0.20948392138091676,
 0.011653063843171885,0.789993032791919,0.7918767211091206,0.6204263492519972,
 0.32162006850908764,0.32076754979468125,0.17696230537832214,
 0.030438789215891604,0.6499679451089461,0.9738109585233227,
 0.29817429688081676,0.4197136602058257,0.9288415446900191,
 0.013024186736949717,0.5196109546617905,0.9068616675641747,
 0.23449966918507714,0.6456935337423477,0.48729953597631004,
 0.2785568523976778,0.381572207186446,0.46734510122729667,0.22709011086889352,
 0.23224726588480538,0.7253732797080016,0.7392003627393586,0.8098354565476328,
 0.10525259361141559,0.4608497474812744,0.889359799951373,0.3118020990085847,
 0.6198176152738941,0.16809855391330697,0.533762200421504,0.2935336509094981,
 0.04697053768108683,0.4367996003257034,0.28243778627663696,
 0.41519416560931266,0.8647669668398811,0.4424590712474348,0.2514989568828767,
 0.765745739559055,0.8388765322943348,0.1844212640631363,0.24666270553006542,
 0.13323142957281753,0.22678002310387146,0.00962220025319227,0.45922236198775,
 0.14420813728335602,0.8085677526243635,0.4707741888301217,0.8676016600298682,
 0.1785461575156162,0.8130865803233922,0.2365140963475184,0.03373962158096799,
 0.6404784733310013,0.17714102717970204,0.5383708455010319,
 0.06307199061533231,0.030972330610091525,0.13052322528716598,
 0.30994475499715135,0.7054771259623348,0.15605813763344034,
 0.7806742683812795,0.446745441242939,0.8424480673070652,0.8503079974382155,
 0.3946473478593804,0.03340979658998877,0.0030003255236678017,
 0.6215612865152902,0.714784878061496,0.4979472320472531,0.4977746102480334,
 0.6685662614188049,0.3338855067609314,0.36715813894371696,0.8809992153751336,
 0.10308596051790153,0.04854511102743175,0.03746204177748125,
 0.0947848778774587,0.49199465699218403,0.36657504285927767,
 0.2568170158181029,0.9038056770548445,0.8210589579675117,0.8053456640289934,
 0.17778117941403615,0.2113400069340673,0.7373270390771435,
 0.16823428299970122,0.2256794977851947,0.06607174890157774,
 0.7624006043164986,0.910537840388248,0.9215800924969278,0.2963411568996569,
 0.7232812774664218,0.875397425104672,0.4930820063042337,0.1797430633917133,
 0.009892805298257201,0.21326486026468516,0.9090863289121642,
 0.5515807719792362,0.03368209603272165,0.30307375098108524,0.78419338203708,
 0.9982567312020452,0.5837784043295738,0.21413613225837302,0.9319106936953545,
 0.8132187727715925,0.5246761930335687,0.2308651642249273,0.24781705100737905,
 0.7371036411143973,0.44429464644537353,0.07644545854172324,
 0.09748443378586291,0.05223442500181186,0.5044577714638542,
 0.11325709232973713,0.6146200190485801,0.09065059470963677,
 0.2972603536835916,0.2030342345967453,0.6276913503518382,0.30729497345019297,
 0.7146241323129632,0.39185938752886507,0.15114149097049578,
 0.9194193066573154,0.22935115527490768,0.43921969146359363,
 0.12689187096413201,0.567735182291253,0.05286081937801934,0.8082488733826773,
 0.2236807511135841,0.6226460995221117,0.021516706561609533,
 0.36406231650700727,0.3373135896917273,0.749052137393343,0.39311237265648247,
 0.5108954123372528,0.3949441692708563,0.37611656299493146,
 0.002257335474037614,0.16515361542828266,0.09992978692646404,
 0.24818369568457566,0.8699713166768065,0.6465944055444626,
 0.44700540087635143,0.4221289198090713,0.26951471631423773,
 0.6966719722575782,0.8653339022221342,0.8673795509459665,0.03263900720296564,
 0.3724880545338748,0.023721333019606328,0.79829347059215,0.21360042286234426,
 0.08237226385287177,0.7855227002274915,0.025310905656131477,
 0.5247126513528768,0.8636327688722373,0.9007976300226485,0.36086882917248475,
 0.763658889018761,0.6152108972076583,0.07537276695147854,0.26964890139879805,
 0.44389017981311607,0.029364691607284055,0.5698259336286603,
 0.6488991135415534,0.7476899721993335,0.8445228588217852,0.43153108998496437,
 0.43446955884648175,0.17154049273172678,0.6312556837862641,
 0.7574652555932162,0.6791294857236714,0.018001208081805165,
 0.005742007033226804,0.9252287701638852,0.059561771929493634,
 0.739145222871957,0.7269849726449445,0.5762515524685309,0.0699203611774597,
 0.9848583833293395,0.4298560511570101,0.44371025794730934,
 0.42213878941520666,0.3135008924176651,0.755864369035662,0.03621164197801374,
 0.647774156392732,0.05494316795634446,0.31903656298296235,0.3648985420849804,
 0.09222536361593148,0.23061977442577064,0.17280482875528436,
 0.26002114404066057,0.7593176151693761,0.5011937551144108,0.513016704531437,
 0.5114893739366059,0.7916057095021576,0.5748469939041958,0.4336579113716418,
 0.10041674652116106,0.7214351511158037,0.1116733007155466,0.7145402734970833,
 0.799452614765155,0.6810814180860831,0.33153922531727553,0.7529584863776958,
 0.6364998006573384,0.3532755380116296,0.018238393243531603,
 0.42526003074560537,0.6275082016465765,0.5703375289959316,0.5771359635485206,
 0.36284741486094185,0.7766490073605029,0.8086943500100563,0.601769134722356,
 0.6065740410075883,0.016323612632741624,0.8638217318700178,
 0.4109559899898666,0.7187524033826393,0.060405589427032336,
 0.5048919710922103,0.8025007068015861,0.41266756832898976,0.6148021779673645,
 0.7239760411815763,0.1100043905137732,0.8262386382694463,0.7527056174010578,
 0.5240552534541226,0.2731150979846988,0.7562089810941973,0.9559915674000211,
 0.3759387601780798,0.9502311891977874,0.08142378425053032,0.3588113207252125,
 0.8270178940347316,0.9741645958712957,0.89505337358535,0.30161985752698817,
 0.5960038914559762,0.8428249211546079,0.08541960003640114,
 0.055575760969886256,0.596732715518412,0.6823908038781716,
 0.24688760067876347,0.3928174281435245,0.9929859410781995,0.0387557423420708,
 0.8611869420982026,0.9731819557317556,0.6267383134206255,0.1916343666660565,
 0.12926780511000002,0.9171675336061323,0.6025863573113531,0.8856638538527165,
 0.016596711549073606,0.6984712128375666,0.03849998545142408,
 0.09305685395449081,0.9495142117772203,0.810119239538864,0.43853284666299663,
 0.5849179047891807,0.47331761672913175,0.7176848536470514,0.4358571678147991,
 0.8870722920732341,0.4945330668197031,0.7747808865217797,0.6350400685814985,
 0.5091808525461201,0.5639926862091464,0.7132705236431001,0.3144562486530914,
 0.22480245955701483,0.6086216959321806,0.07348135108812459,
 0.5471878062118632,0.6198645235271774,0.6712160147762182,0.6691478908827727,
 0.6644204710756162,0.9825149364278618,0.13979484422205035,
 0.19706549010955599,0.04913343739789178,0.007292370165730033,
 0.9753620227545128,0.38112445444573906,0.480383529302042,0.9766683745232636,
 0.017329516933604405,0.37183459535444385,0.30149279877999435,
 0.8694873359019053,0.6622964029100864,0.6138130351435477,0.3848964102949528,
 0.1585854486789866,0.23479697547545264,0.03285215227278038,
 0.21671654389660244,0.4301858681470063,0.9929001848291675,
 0.17299449952334467,0.7953857076995499,0.9544681972269289,0.987181057982709,
 0.744743552621805,0.8529655848706622,0.43713439101657336,0.8085627889552489,
 0.0837270724561936,0.6007257517243578,0.5791439913601725,0.29542519987966265,
 0.5975309541609203,0.30022178229505947,0.5717544765847417,0.6846085527374165,
 0.8661753982313015,0.7469924194497002,0.04625771908051435,0.7326755918751521,
 0.8085728464812166,0.4720424149263466,0.11845118576823277,0.8880733295279982,
 0.303767118662554,0.5290538470457962,0.3715133133256445,0.5743844426466561,
 0.4530101076083066,0.4934519413705243,0.564040119964001,0.5532388458134035,
 0.4853418024344852,0.9218247362461217,0.20902257443799288,0.6328923240215091,
 0.5086130547200997,0.912723464485303,0.6466901807712604,0.9345664169477723,
 0.4288332828004392,0.4635664103410586,0.5704563486310739,0.6585549629538862,
 0.05557857637181629,0.06500268126484732,0.5650140598057163,
 0.5568208612991391,0.4790047949311428,0.6055420792634361,0.47339531940410107,
 0.44245922066624954,0.3778899636839217,0.962611740636631,0.56547787525339,
 0.2676179694177522,0.2788045580006826,0.2669694449585954,0.14147279106410116,
 0.29367602926896386,0.589345976447268,0.4645899311724975,0.11682567226011487,
 0.1510320183701006,0.46296806748015085,0.899173492181348,0.9113337456715302,
 0.4231453574834715,0.15373572872887975,0.5316940281631872,0.9059294913764262,
 0.391850522738858,0.7747514228086512,0.4969822379056026,0.3136485102981592,
 0.1345216752284879,0.11531330609381862,0.9288727034462808,
 0.030572686504958813,0.11997145186179448,0.7266984140677093,
 0.2835680921593535,0.39122097610925166,0.4645195971791227,0.69975366182857,
 0.4076618206651742,0.646403284747161,0.9083457931980873,0.02189122956711076,
 0.5495531263836749,0.5158327226859525,0.28039316583564255,0.5036258024235287,
 0.14256328942841667,0.8539877647817126,0.17118818762778742,
 0.02991009597811889,0.0648154106992459,0.1183942018120383,0.6056275114046314,
 0.27997116316802273,0.9944656398968542,0.7118359988341075,0.7511493869337978,
 0.8400660773091048,0.8004234561385956,0.3067787685082216,0.6198163110143361,
 0.9498761668573407,0.4249602024808694,0.88794607826358,0.5000387460725273,
 0.8652820029376839,0.7253508837331695,0.7394322651743102,0.2329790068704578,
 0.4603094572593259,0.46712559756974414,0.3103439416801541,
 0.26086357300927165,0.182146767578386,0.42081472457104896,
 0.03398820569335903,0.6781553656987551,0.6410538722810107,0.4266315404961193,
 0.716951998678447,0.4556285757577765,0.31318018627996436,0.03578659675720908,
 0.8441960037714095,0.33892479354822636,0.40255815918389093,
 0.5715327906434281,0.6553730888482339,0.20873048940699868,0.6385003503767949,
 0.07311132205975268,0.9099997690919276,0.5642659167737016,0.655671458063808,
 0.7771297227713999,0.07223566944552984,0.8844523950242226,
 0.29365448114786563,0.7294132784277763,0.16999883723142328,
 0.15169753647087325,0.5927772759340527,0.2337303300587632,
 0.08205091563041933,0.06077835496870598,0.1910471890283577,
 0.6089987735981327,0.8222171577194101,0.5775789816320026,0.21130683976520948,
 0.47547438826295485,0.4704920599328034,0.7439913366957007,
 0.11635019667536106,0.02144346179446377,0.961811531303563,0.1615364084635177,
 0.5343095270765414,0.731229156666255,0.6877952233651768,0.6027338219926716,
 0.6926946698512932,0.5298448862304582,0.3468207267260557,0.9694677572686263,
 0.32300212788710436,0.5100350588715109,0.09531054140500728,
 0.6102595997936479,0.3321466434484843,0.5518039316857765,0.7781714186832349,
 0.1523302450083106,0.5137778975809634,0.368993578916742,0.7100457602748628,
 0.624763256279185,0.6593968025317012,0.19483207376264744,0.12028484725763622,
 0.2541145447787767,0.9493630684018198,0.3826005199525693,0.6079347861406688,
 0.6150281818127192,0.822940878012195,0.7704400699520378,0.40663747327046484,
 0.3466980836427317,0.09805861856301923,0.8559196181975618,0.7784734208371834,
 0.32010303659734207,0.1306628285690994,0.7204168300102133,0.6277759862887556,
 0.6312083599304287,0.05010329171747829,0.3713359114170086,0.5534552551884181,
 0.4339571733358152,0.7978902888792446,0.5643681027469915,0.24551366847245393,
 0.4590710531313863,0.7561378614636942,0.6964877655616104,0.5721368709213147,
 0.0741265461864884,0.9640160733923168,0.6791325026231659,0.8843565626308889,
 0.6323473093370691,0.1369346059582972,0.09269256423610805,0.6294668205803428,
 0.5340177182512144,0.8926916850253165,0.04171895223112554,0.2185561631510533,
 0.9686101087515673,0.7991670213078971,0.2302124628933574,0.9788245058231002,
 0.8043254654507533,0.8623515672122569,0.6875603114749124,0.04833796823025227,
 0.8766120901096007,0.22011972659083234,0.6994418960751344,0.7650751450621907,
 0.8054307311983074,0.644607805074368,0.8182099048466076,0.3047804856811096,
 0.8923070040375769,0.6601927129234739,0.36500114012687723,0.3158498011897585,
 0.662197738297674,0.43094035151627974,0.159554962949915,0.5448509048066099,
 0.628662733613844,0.5495002840353511,0.32804196316879586,0.575985111907868,
 0.8937415337932882,0.9563088768775615,0.5542196974245971,0.2938283506647248,
 0.031238401766350954,0.6718454161467162,0.9785777358851733,
 0.4716029732102899,0.19578439523366753,0.11741088262284693,
 0.2044293579756643,0.03402899611612176,0.6390490845907133,0.9548710836396508,
 0.9798424861475493,0.03952424674613397,0.047112614836557154,
 0.3175955308910312,0.07302721722043182,0.8715147443908723,
 0.04206251321831467,0.8039736659181057,0.7871066019623777,0.7091491683534491,
 0.14838045902045338,0.38836420134955485,0.17711362259878305,
 0.517178383878772,0.9352655695234265,0.12104753509307908,0.5779810590844858,
 0.06151278660961956,0.7952994645765664,0.29383122404455153,
 0.3570955495919693,0.869947655310548,0.8732471765203674,0.966902319605545,
 0.16060165529186665,0.822324079945153,0.07448844687890277,0.5561080033844504,
 0.02354988777920175,0.39436437530332014,0.3024767556512872,
 0.10765008626541506,0.22180909029017049,0.356126541342367,0.8138456233780746,
 0.8958469540783973,0.7146669001262058,0.6580348761785847,0.01758998650546506,
 0.3522187042871694,0.06529522127608911,0.13788676124889965,0.718678420606102,
 0.8274681901462038,0.037108427791719856,0.9976810034077922,
 0.3113267091251243,0.34668038518734345,0.3749690725885182,
 0.07944235456541782,0.6788145910418024,0.013304192593109443,
 0.9736835761835209,0.08528874531926423,0.45926026199623526,
 0.8122053025690886,0.670531782970868,0.4796998658870619,0.3059207983136374,
 0.1463897819257467,0.2770761555418756,0.47378195952345636,0.9761991339777981,
 0.1950906986716241,0.5581907173298303,0.5251517465085133,0.715010190086644,
 0.010499767250892633,0.7821193405648665,0.11427246360704757,
 0.3661812567787537,0.3824581214499685,0.08700751763011161,0.7735637252243908,
 0.8673189582910192,0.9136594396560458,0.5633449833487979,
 0.021356558102058165,0.5391009293760898,0.34150692366772706,
 0.39429873930654125,0.36483023391815794,0.42069380088302755,
 0.9689580844617092,0.958228455165717,0.8369910329422532,0.8634777213206181,
 0.4808436686091665,0.585940930500757,0.23547187502548428,0.936823836839731,
 0.056920175864494715,0.0014607555053345322,0.5797986234581409,
 0.6351681992510658,0.21446486223415306,0.7355401625129414,
 0.41460823111895295,0.3026392794179854,0.2514636121639413,
 0.33176617846251455,0.7269956095933864,0.3279504623965819,
 0.07550644570504472,0.1084929272655153,0.5026555638386935,
 0.31962192927019295,0.08094776600472242,0.6305672715030027,
 0.7918529326940249,0.03434343207309154,0.35423622362643525,
 0.5218379350779551,0.01711224395724431,0.742516159985563,0.42450653813173367,
 0.2968322948196478,0.886311876606733,0.8761521169912626,0.00436978351761852,
 0.3079423050680864,0.3059807324958793,0.06459167694350465,0.790553201055599,
 0.6845151731723318,0.8937580493542265,0.8093797466642919,0.15581707975826742,
 0.7416020095816945,0.23041268338036236,0.4624605544738999,0.5468030111602828,
 0.3419072360701372,0.6654991365480779,0.05658892660928172,
 0.23189836262689145,0.5991698001936208,0.8331765895651401,0.7728377642721378,
 0.10785937515391386,0.6013449950414705,0.5069925695587784,0.8586466419857934,
 0.61102235454783,0.32252794858920353,0.9816835693427635,0.21236161106281026,
 0.39675533951634623,0.7804958420275654,0.5450336605371656,
 0.22124342493032345,0.9608515524349259,0.9507724596353226,0.7101380053850275,
 0.4254497895052062,0.9753612687126013,0.9501352163317777,0.21527253375110178,
 0.5978269675464686,0.918279338023017,0.5593633270492927,0.5022341045113073,
 0.893043828120501,0.9387709161817407,0.5460539299593101,0.6831871030501258,
 0.44421447963580374,0.7755344982653357,0.692787065544284,0.4899634671620018,
 0.09008082591074107,0.6322151862515653,0.2120194938310429,0.4082754779650972,
 0.23475572215711693,0.08268331630011327,0.0579556573806117,
 0.5611567307258174,0.8527276058010671,0.33292524067626594,0.6274290799758965,
 0.22046411560601753,0.026567642793125512,0.41725271301326305,
 0.17908314237882383,0.6676967269114789,0.8146113842035094,0.2735387662530283,
 0.6322248148895124,0.05780538008398495,0.6450337442885177,
 0.44252276455252526,0.04637669170603331,0.14073949613502046,
 0.3721732517997176,0.2889265399743117,0.0866492663009435,0.22860466367245236,
 0.5177754984394614,0.7036225215687181,0.8044085581382819,0.10920059182080943,
 0.4551853822736627,0.5564987744877458,0.7196799276350472,0.8449243375020018,
 0.5118757902837017,0.7286442283773636,0.46684637086840464,0.9098514415815264,
 0.7834844134275918,0.6964260743175369,0.16656918142330046,0.5737300981181016,
 0.7375119905215034,0.9578432190868873,0.03569147155548247,0.98351230013528,
 0.9364217560541186,0.06667145122353291,0.05275039683778848,
 0.14358880009608554,0.4522029640844625,0.8063195359700548,
 0.47455100055755794,0.22675025985741615,0.6083640980864818,
 0.014803388138337814,0.3895456695446494,0.7067921157704794,
 0.5862011742688843,0.9460468754145643,0.12025608400459453,
 0.09665949259786322,0.12208429961702927,0.5138068874003685,
 0.1625695815442766,0.5689685588082637,0.7073363740253119,0.0204943955122594,
 0.3601928606127016,0.4984010826605175,0.4257240367744388,0.2651946667087769,
 0.6424501422405339,0.26508794513179346,0.7722272259036471,0.231942577659617,
 0.568927695413117,0.6826832956628789,0.5036582353939565,0.3791355646094736,
 0.8651232025498974,0.3569950490425279,0.8002009815556018,0.11141256279020306,
 0.9827113287674031,0.18475958556199212,0.20052345785497938,
 0.04980240946627379,0.35229989139720097,0.8581401049769033,0.630693472013307,
 0.9982589503859841,0.7902393016321799,0.9836148173401893,0.5010443891145637,
 0.405729638228869,0.03607914142748547,0.47005688991572325,0.80545838831726,
 0.4433285459142833,0.26295837646688824,0.6276370684005027,
 0.028748297394713962,0.27760411259793405,0.5099569331607653,
 0.9663085808108807,0.796444909215905,0.8816199112568983,0.40966508888936226,
 0.3483628067449773,0.27784152398111617,0.1929316365171181,0.4946802484638423,
 0.46557240753457685,0.10101536962778612,0.39408605076104053,
 0.6017397801099638,0.4448966403822898,0.41287689027326213,0.800496235525353,
 0.5137457361148564,0.4098450389110033,0.5224573303634887,0.41923132238866956,
 0.9509317340643219,0.020762084112207257,0.8145439154666799,
 0.006913932142887824,0.352382593191671,0.823939526599159,0.08879143650797072,
 0.8997873436285282,0.5960653572766693,0.712493447194997,0.7398820548425797,
 0.32830217841536813,0.07096842576498674,0.5615677162754428,
 0.1678454247633855,0.5897359741289584,0.08185932956019482,0.3284262950359723,
 0.7822069087087082,0.4455921092703309,0.27586864941337896,
 0.45977613968016695,0.5429186255832533,0.15724603224941902,
 0.7860134304650883,0.5070860144714431,0.8033642868757114,0.8376359525832404,
 0.6790124826519438,0.840938451024179,0.6931995072120327,0.9562170108290735,
 0.09061076729928053,0.2767266668437265,0.8184288747955162,
 0.30175577665318554,0.48410578710599517,0.35718136597502914,
 0.0842540107655192,0.1436797072991146,0.9680443055855751,0.4318922244076938,
 0.26590938310494705,0.15541679776427852,0.8015897684201092,
 0.5742996821433699,0.7079810752509575,0.2520804010661699,0.18310928351467415,
 0.8251270226716191,0.33952187589526617,0.10526173355475821,
 0.7761134167480197,0.41534587727047634,0.6014727754313123,0.5674505849524125,
 0.010412912207156677,0.9257815068938068,0.6596013687402713,
 0.4222920596945703,0.8460167232249503,0.864461467949184,0.6796876056668928,
 0.6294259272233783,0.8361336277835081,0.898588518858512,0.671176711430563,
 0.1911627128337201,0.8439545226424846,0.5760705993738297,0.7402070955474789,
 0.04005140070736435,0.4630032320325801,0.5956931573176836,0.3389384514384036,
 0.043016297084076505,0.6624485603416242,0.4819069991909083,
 0.6416929947941646,0.15379767673143663,0.6628540920427083,
 0.10296810007015567,0.49684772246977815,0.7613653046688821,
 0.15485708160501233,0.005979438880428178,0.5505303372835131,
 0.5659473560059075,0.25690997273425653,0.8266762836979641,0.669092328916765,
 0.2157479181602835,0.8510025631482139,0.6502143687999898,0.43406872362902327,
 0.9146286813780892,0.619500998569529,0.325358691573169,0.8616730697987404,
 0.47649691029416164,0.001304016571206601,0.8400545662765009,
 0.33205570695980446,0.6926615095861874,0.33730395239895816,
 0.42200026815359604,0.9490733883996649,0.8466096826228973,0.9613831027243469,
 0.2765753463228503,0.33792211584889376,0.8282559100851595,0.4938636083611152,
 0.8110756354820734,0.014694567619613186,0.3072898655946992,
 0.06539118933656407,0.022075483033420218,0.9467506597540536,
 0.7435345995609195,0.8492954656862961,0.8670499028251523,0.7922800300726647,
 0.7567440226063818,0.9786300958532278,0.30199650688152335,0.6004531993963708,
 0.3566051464204706,0.3906839077766212,0.6693458872269697,0.40940671743807844,
 0.686630859492022,0.9194361979683655,0.352361686040237,0.5586430920076157,
 0.10500925571729225,0.2735205331600683,0.7837467277245778,0.6778521452365277,
 0.4586389797403446,0.9631516362361233,0.062374896821483894,
 0.36160379922401276,0.994263744668114,0.9330788507293852,0.03822465265688102,
 0.9905590628604408,0.8203579231129136,0.18569609484297822,0.9912868256857417,
 0.4074245399887605,0.2409833693759026,0.44471976735531327,
 0.12045071470888524,0.21653954582989754,0.018622078209256432,
 0.22078387870775695,0.07669450002121425,0.14428359197080753,
 0.28976112412538657,0.7101804175605264,0.06563579060988622,
 0.9923882645925123,0.16881724375544682,0.07113688536425133,
 0.4264392352231342,0.7410372968523061,0.3075511187454447,0.9833570938260024,
 0.0748688464979329,0.26271283153602953,0.09917753653837946,
 0.10561092342616685,0.0404049687408885,0.8156957177161981,0.3578134184381654,
 0.0038151005683840777,0.07816902202198317,0.4617539621145418,
 0.5639888486314588,0.756830222573927,0.9981691705922865,0.1214852031780006,
 0.6961234288999453,0.17609292608977456,0.11518158833985148,
 0.7644887350959723,0.6720915240652832,0.11761291255298945,
 0.08517387970840573,0.5463495242728418,0.24791186687724942,
 0.7471528208464041,0.152115041909062,0.05817741976647817,0.08020132769228949,
 0.9689302901623918,0.4053193643052484,0.049707207987393476,
 0.006972809363425814,0.7358506841799557,0.24768569539380336,
 0.8084102821285185,0.5307636755333132,0.41411254558035804,0.8981855972440664,
 0.5794205648030133,0.606322834824919,0.47572492183684956,0.3527645054400117,
 0.6289021509587193,0.95996027508203,0.6583929062742715,0.5481800167747577,
 0.22835409726323908,0.9512234458880384,0.4392400799713122,0.849403186171536,
 0.9895532464844581,0.4701857752255847,0.2718515198830931,0.9108675250254464,
 0.5746839694365256,0.7918454463040565,0.9019811858132828,0.8633115516871974,
 0.7402938488878571,0.45459272615250235,0.3708114147388223,0.8735971654042303,
 0.7995793459212007,0.1448062238070138,0.5536406054584864,0.463694772603526,
 0.364788068943364,0.20765872589578294,0.9792436070188268,0.8844053779130324,
 0.20424737462910558,0.9271360890502596,0.4427670985630099,0.8048609246415124,
 0.725816404384596,0.641398627412429,0.8225994011276654,0.4182237303747164,
 0.8592914557447477,0.27567625606144164,0.5356315928370268,0.8895261759240971,
 0.2834653238328956,0.4496513694835045,0.869880483799921,0.43484012053584054,
 0.9805401889024025,0.3987059889639789,0.8837837264168699,0.5247818753142091,
 0.5077034599823836,0.09009131607640342,0.5549948024051853,0.8838909846590293,
 0.8713265142603204,0.19014772472116015,0.19926186843274507,
 0.014596179364502748,0.6704423486853182,0.05611671368479576,
 0.9078891280745242,0.2724961645635421,0.8077290296471668,0.8533333642377902,
 0.6365168838852566,0.483769661622768,0.08915741113152198,0.3148961040388538,
 0.2986138596490825,0.8245103008258938,0.08075314189918237,0.4496375723245285,
 0.2509612412278702,0.9445452694628947,0.9241025435943522,0.4995427357573037,
 0.5616606127330714,0.8606715604034212,0.1087738775877013,0.6200231982033526,
 0.22901307621987044,0.08900534743693744,0.3259246523815825,
 0.9816486776093752,0.9746385523053145,0.4387084751436382,0.42919088004946304,
 0.38316329404097704,0.7466513106963416,0.9743530308392908,0.8012407319996437,
 0.21683705946179388,0.14593249393462415,0.2769815484025404,
 0.5732338325437653,0.8522330490601623,0.11104732524873984,0.6994208250851883,
 0.5680661059023678,0.13057969977311135,0.24519646686525587,
 0.8324443969101382,0.06428608853926909,0.23064942613615247,
 0.8431044624097513,0.39498518337064814,0.5694800758432801,0.3837777012437077,
 0.74418081258182,0.18496805999653698,0.09934017786617311,0.45581887994904924,
 0.6812930854532406,0.18475786032100638,0.7571914101242853,0.3623188104417495,
 0.14400824885575292,0.9453915497843094,0.5379935928588179,0.3701854492923232,
 0.6975037472453721,0.015492662149664937,0.3414283466987731,
 0.19953356633718733,0.32865072139435725,0.20449210878407342,
 0.47201946423576746,0.5569242652322098,0.7580296278699931,0.6938790278839786,
 0.2357478353555129,0.018018516436989596,0.15406264801933678,
 0.5792221848993041,0.52932895308323,0.7795552271977418,0.35721861039172986,
 0.8285372587427964,0.20295788990658115,0.796230931164422,0.2828079951360256,
 0.11601820588446499,0.546517642284371,0.15114728422748713,0.5662028790863339,
 0.12423170285587837,0.30552347908927,0.432433664224915,0.9630288619151629,
 0.3904245964262876,0.8235840018619335,0.804037935357783,0.014724066611985331,
 0.4568681936823178,0.0714049591303868,0.25551084120195067,0.4209303757514391,
 0.8646401352616114,0.13812067909242265,0.8793957085060989,0.8974200345546866,
 0.9598526725406668,0.20890946035882507,0.699516956848679,0.6727922527847188,
 0.28194458864595695,0.004088131766989855,0.29177323881333705,
 0.03412515501908753,0.373194814229612,0.45258106734709314,0.8815532465568632,
 0.6778904532354006,0.4999644709780384,0.7054074544441336,0.9223361827731281,
 0.45212494442917284,0.7435461323436583,0.3520581454800109,
 0.46615600414493763,0.022527919278016784,0.3619226058649049,
 0.8814506276172547,0.7753001438718254,0.3149427710967282,0.8803928049106059,
 0.6586802803180195,0.20463891012281876,0.26957937887268546,
 0.6196980987893288,0.15699313902954504,0.4913902690527856,
 0.36234579735466466,0.3517354428201821,0.4804769312709686,0.2561552029369599,
 0.7667091061126713,0.25626638348169006,0.7186657348333619,0.6758722507136082,
 0.5144470115835086,0.8205080497961699,0.6299138630025967,0.9934812547869927,
 0.43015942344743396,0.6419697273133372,0.05820247530655909,
 0.3518560524477139,0.9212364499399501,0.8928718600316345,0.4427174303093283,
 0.866425540243142,0.016818225220332828,0.3978811021292986,0.4214305481978795,
 0.07799038437737715,0.06925446976602401,0.18776319281895248,
 0.2891390316755066,0.9975001411671489,0.4828817128210927,0.9840824605443336,
 0.7464938317086117,0.6883118101561819,0.36229714202763696,
 0.32972020512686373,0.3853404804957995,0.130769757792222,0.6239233298303807,
 0.9103358182930659,0.5505901696949616,0.8663953567991645,0.880793270902887,
 0.6385610480212085,0.2981460175095152,0.3687602785623938,
 0.007562969987810142,0.8659200732585811,0.0937737717252396,
 0.09878684000454185,0.6031704921223656,0.7404486778150816,0.6364233381063117,
 0.655268468762818,0.5868980107676007,0.9233870787739938,0.6457020958680759,
 0.5429638168733757,0.07223868340935258,0.8499092404483577,0.5412106414476205,
 0.018775365410712785,0.45027214073708544,0.6742465021489367,
 0.20262422160013682,0.7084554172028169,0.8120570970140726,0.8676738316533293,
 0.6259981077840656,0.3328556601769437,0.6418024500444716,0.6236523983614393,
 0.8458221152110343,0.08858647437767042,0.3719736160602657,0.5727146822823679,
 0.15694303002304222,0.03647254468779193,0.6233219697029382,
 0.9329519417054778,0.5568727157616654,0.19779419059743897,0.9161668510338138,
 0.05171798350178192,0.9730625561670311,0.19100096770760178,
 0.43973365668524234,0.9136813996670133,0.7264633042328149,0.6187747321019282,
 0.8785479055333443,0.7300399462128992,0.8891860856451447,0.2505064137158759,
 0.492701653338112,0.15493910653595233,0.3245600303269902,0.05872204356326716,
 0.9939307281729121,0.19213849939804017,0.31755456700540063,
 0.6424726506917595,0.27543169972778037,0.5327417623550222,0.4892093959394318,
 0.011429832523881522,0.4666687858479981,0.5587438852546565,
 0.13604796605124414,0.8063539596249631,0.9538459250640088,
 0.28588930534253487,0.5636613800763117,0.6600059505402409,0.56526484296405,
 0.9697516020476933,0.08312230205985505,0.8169658821063535,
 0.35612398786801325,0.41766686170986034,0.8074239625549262,
 0.9773878505712215,0.4731561791877682,0.8095169625365354,0.47658800044284166,
 0.3066409906937193,0.8085863154014019,0.527548009428004,0.38400140701595287,
 0.6970024783590537,0.3228944284459462,0.41961860021906805,
 0.16798961876494622,0.4210225780106218,0.9935720406351433,0.3345374581207664,
 0.031357327622588804,0.36297771175249605,0.4605188996555576,
 0.3621048946572054,0.9968781789714717,0.5379676728304301,0.913269288609437,
 0.8061751554559423,0.07520830994531225,0.2822235658137692,0.7367703937747443,
 0.35987083091236516,0.05986210517997359,0.8613112944520888,
 0.6943392228701184,0.5131843962243865,0.16714258550575323,0.6199278884255492,
 0.30845222931157323,0.64460300988811,0.39136519171258155,0.1562231525883181,
 0.9788028630584548,0.22473479746928404,0.20720614987821817,0.905222423174895,
 0.34775125696881215,0.7099641684737052,0.4829141164332652,0.6012840796695437,
 0.6700572373724167,0.9768612265727875,0.24563126461259355,0.7612100203426332,
 0.15044505365890493,0.8016046153279123,0.10324308097743506,
 0.5982555051154751,0.6086250461827689,0.2922019644215924,0.5271547984924752,
 0.7142804408425732,0.32657074136938224,0.09943330024738928,
 0.4455142246692382,0.24393111588869665,0.43664957987264297,0.527005556912785,
 0.8750819938966224,0.3694215796402649,0.038286750876078024,
 0.5059122324650049,0.12625179804676956,0.4274695333681847,0.6297297208367377,
 0.8175820356531089,0.029988717297279255,0.603123887282555,
 0.036401388646938315,0.7697979364189668,0.5194937841392279,
 0.7621258089700187,0.2828944684056699,0.1275824853270836,0.81237967491982,
 0.17146232372647385,0.8263378728090228,0.1211823007702113,0.5360894074417505,
 0.9694997680085227,0.5621444122080523,0.7419994126989422,0.6753481199217893,
 0.7082597006951072,0.5665109580024759,0.06166298876753973,0.3287163755632615,
 0.31193325152872486,0.6909273623015592,0.9663682692870877,
 0.46327935674172704,0.8857966174565816,0.542624534042484,0.2598262038313378,
 0.6437667715353064,0.868332020235721,0.6051263125633327,0.3711667652620181,
 0.0436835687193301,0.6691533849171456,0.48316686786298757,
 0.07151142928058662,0.7185111357155883,0.13522218987679457,
 0.3236688255534791,0.14332487416707973,0.7038998858521716,0.1405691425706025,
 0.543410811403212,0.056577226413377035,0.5420867116428072,
 0.19489124374169453,0.35940558250111354,0.9804957094428829,
 0.0563925839979843,0.40960642963084204,0.30440734039206196,
 0.39945364415899554,0.16414597599244263,0.6936682733918097,
 0.9696912816955234,0.7820199635929008,0.9704002615932342,0.5990581568950816,
 0.5995699476888533,0.45418750725650814,0.9656361195225697,0.6524027862261667,
 0.5268816473124628,0.9088923249704068,0.38030670062577554,
 0.023592543374749697,0.43664681890587187,0.5856300925785911,
 0.605504111038744,0.8722150961042894,0.1458306693023541,0.6674210805982215,
 0.012829790764389837,0.7480573409736484,0.5609023091536096,
 0.20443769544596846,0.3047753175673823,0.03842301680172677,
 0.9683149873852255,0.058056784025967034,0.7837137263857716,
 0.4100267414917488,0.33426502590432006,0.9297652034703353,
 0.15844476402893526,0.4146118014596768,0.2431156858522412,0.7296167937506017,
 0.8234266342175263,0.9744805255564088,0.07399680405059528,0.499773047730496,
 0.29053046675220906,0.33095270006060173,0.4147044489974371,
 0.9083545234117308,0.3251503095213041,0.6100959451087885,0.3040017101523438,
 0.07254438876344305,0.832782254985039,0.1591880451716049,0.8974284762158714,
 0.4568831008008286,0.31458101244925873,0.1461065429326266,0.6920612126207609,
 0.73199728114399,0.7370595687167464,0.13631197228029002,0.2190342261528252,
 0.42468712431239786,0.34457837761635657,0.1905464887495234,
 0.09981475572611587,0.37391883579249896,0.19493453841524389,
 0.20038410991447464,0.8747269657885572,0.18243453859064818,
 0.8376837825496195,0.6718066103379945,0.2599400513342368,0.518228249307321,
 0.9308234713142611,0.6241329360273382,0.10048448223953654,0.6386152669219701,
 0.6778732861051601,0.7337665550486717,0.24948352228184056,0.9344397487179428,
 0.11929800192154116,0.8303114165123322,0.7535945462621103,
 0.21192497920174824,0.6237551937362659,0.02129012297817079,
 0.31240998498758443,0.08653870280845766,0.5374897539352792,0.979458298154767,
 0.23763846199521876,0.09800965945121132,0.8447831701628963,
 0.041138144784720665,0.6770868853229073,0.3250862615352742,
 0.02557053776103313,0.5295562538154339,0.09245999073108757,
 0.06379531498405655,0.1814269042886516,0.44338064086311746,
 0.8555804947839918,0.21464042633942948,0.873557952711119,0.7606289937774963,
 0.8877180584513482,0.6594275043004383,0.03494785150023316,0.8467697609485467,
 0.7067984926686084,0.2172368827329313,0.260607724739345,0.34639784056020306,
 0.14821987260523173,0.29492468912234593,0.3360086116174382,
 0.6844225053668459,0.0819676073732144,0.8376810930940919,0.17231149092155162,
 0.7950277872956405,0.5371245498902372,0.4003092262290867,0.02293374495846323,
 0.6559803108851867,0.08693347725806022,0.24648347369273005,
 0.7003258348417261,0.8888819077163257,0.9115207813551589,0.3574518604428293,
 0.23525588083661508,0.24262369581431353,0.3366614749366754,
 0.23395313818883434,0.9230359248369353,0.5652264730807537,
 0.23552596656022973,0.31098152738800966,0.651595557970109,
 0.49189937582116416,0.3788290041429335,0.20310547837413084,
 0.11408055747259338,0.18299612975053148,0.20931207998103196,
 0.33000953589405324,0.5542493884662693,0.7378613429058636,0.1780797286521325,
 0.9691552925018614,0.46614612454029736,0.019081504367030155,
 0.10842765203194238,0.0997808603186564,0.4666878954426157,0.6011075215355689,
 0.12618740282873198,0.28663906679689766,0.03838162230283837,
 0.29322774567276144,0.4543805291697923,0.13540185081288614,
 0.3086765234425115,0.5272494225591178,0.6931671034124326,0.11225348385525713,
 0.4008344724500157,0.612685512127923,0.30990384836956764,0.2767802543504524,
 0.1822842231292452,0.8375024771918652,0.8679445390328893,0.4707138390844643,
 0.4401496969813462,0.05116687931995145,0.2116149504716055,0.6674048483635789,
 0.8735712602870074,0.7619485625788702,0.31070673930213866,
 0.024123275894493923,0.12017135842775972,0.39139673904931127,
 0.21235556388846932,0.7613548453482184,0.5218290855002387,0.9044572395267503,
 0.37795986291999817,0.638192560312056,0.8218892999521745,0.9112926633283435,
 0.7095798483578843,0.6712531948148195,0.6770934569587208,0.1880313999014238,
 0.882763618546046,0.23271319685991743,0.2851727176390889,0.31628546486335896,
 0.41030773611382587,0.968938167628908,0.1627459143167982,0.20277164778417844,
 0.5445242690391133,0.46827888917380844,0.1448363226656073,0.4321325098056865,
 0.8470472637392434,0.14974895383807507,0.36269858043714653,
 0.3281710449700225,0.4601039187570459,0.6231047220950316,0.15872576266466143,
 0.7911476404860186,0.9532361943545434,0.2905952996432356,0.2978585557357252,
 0.4846124160107822,0.20057572250276767,0.6892573570324965,0.8588965399047427,
 0.7765957761933469,0.29101181671888643,0.8712069408058222,
 0.25019943894015995,0.5362843592916727,0.9600618992829346,
 0.27978254976396943,0.7362319803327341,0.4636533110197627,0.1654969159307299,
 0.9671215379441387,0.1397968885083385,0.8285055989539543,0.20521948789150457,
 0.40506112686788476,0.44896803242819305,0.49690220584169087,
 0.10491832413554247,0.0673621755507674,0.2708902058217669,
 0.013003251377409653,0.7233603197330222,0.6099383490810665,
 0.4474812533482375,0.684948005669751,0.26789454947509583,0.10304207740775007,
 0.3005702224039528,0.7535027264060083,0.7323976927820537,0.611377810106312,
 0.4834379136501984,0.8873595754919772,0.7967197402705399,0.6175673531936925,
 0.7022535546100432,0.43759910249207545,0.45234975813597733,
 0.9091805018373309,0.2921153217140192,0.06942283831553664,0.5041741679262577,
 0.8087545600849034,0.45718549963198263,0.48471003803855195,
 0.5993976794629623,0.4594358260848701,0.13662014192426575,0.7488037333938106,
 0.06676870901785092,0.34749178833306993,0.983771303355811,0.4191647505267131,
 0.11734697684832662,0.7617651417472358,0.552103492486277,0.6446590858655625,
 0.9319826362828214,0.6543850993546516,0.5607033742387331,0.9622781226104327,
 0.21486947390726707,0.6940424570023612,0.1765103506693696,0.6485207808113633,
 0.7283055011800251,0.13931717677522482,0.5772197514488172,
 0.07829569964770078,0.800584380262737,0.7247586192168726,0.3732372217019174,
 0.41632943640274056,0.7588653907473861,0.9727795472931122,0.8850960829853134,
 0.44453011035691625,0.4128189449568299,0.05227186324835609,
 0.7665132761689224,0.3375199349868343,0.007862311897628538,
 0.3781620321329604,0.08143486448423398,0.31317213544843603,
 0.5884661805197146,0.7519447129008765,0.06621530968941958,
 0.34995429728889793,0.5671847966842363,0.6790168972099148,0.6854891700791641,
 0.6319411203786295,0.29798105575145073,0.751077616276194,0.3848148662208859,
 0.602705839093423,0.5231816537842269,0.9305593892749103,0.7503968981319309,
 0.7511892183001289,0.03425773834157064,0.179672454306269,0.5123536729196098,
 0.8314815688522531,0.683166853015094,0.670927090512164,0.5201249888358207,
 0.5626241699476608,0.6150894656431677,0.3469950131776822,0.8115636704458361,
 0.8978750224766331,0.8848616183524796,0.8992868595098149,0.3064691313999728,
 0.22867980728356152,0.68126796646604,0.6267133202475171,0.7911877337979409,
 0.2583429242278894,0.8026977330871181,0.7465158879855067,0.7576571055254112,
 0.5955738967721123,0.17142567250289376,0.6290868595499569,
 0.40820726229406046,0.7498408297580795,0.485861007577136,0.44529584674411693,
 0.5843177917719485,0.8545537030700988,0.5621931555468354,0.8566634191360466,
 0.6002591544254334,0.49423832417747693,0.11676527651906288,
 0.7371043795925563,0.7004734452406295,0.6802929109603004,0.7906689006298143,
 0.6826092995781338,0.6726818398952384,0.9900132742525773,0.6244188384010563,
 0.5256930662956112,0.6465111873885552,0.28342197444902684,0.8994817963173733,
 0.30419073290135434,0.4655165767135091,0.5403678029353194,0.8908295309927664,
 0.2367308022725435,0.24497506979633143,0.4646932360392877,0.7946340481628134,
 0.2655443326402086,0.7107861105274661,0.3503112592266796,0.5899738808591777,
 0.5797726710196701,0.33259783373602403,0.42545915457948935,
 0.05618571431870789,0.10961025661001667,0.709827776794459,0.8658372119939545,
 0.35060654465191354,0.5136628400151633,0.1255401843040168,
 0.07803455532951065,0.5887019093297142,0.21046251650892223,0.866274859562373,
 0.380255736744018,0.25844439873692726,0.9022805029645572,0.2927967688235735,
 0.6630549695681516,0.8689529523887564,0.4323654429023771,0.4858365474059041,
 0.5457417304946051,0.7521380837727422,0.47235541498302913,0.397097684727325,
 0.5927073541818797,0.5879428852205374,0.4175935883267321,0.7038691913537007,
 0.5994446725908675,0.6563949195540864,0.9562266374010253,0.8796686999385077,
 0.4251822476830215,0.9683199437172605,0.07182922149010029,0.8031816588634668,
 0.8176997783558008,0.8500252190541439,0.6360058722909214,0.19095003845672587,
 0.2542974391876931,0.7329218674669794,0.09466628286453171,0.7098026200623776,
 0.94275670033191,0.20394708470386902,0.6779788689736633,0.33637755620050147,
 0.4267283137071679,0.05055800265301402,0.7355235316208159,0.3232544459855953,
 0.10700476541669168,0.2708911683617643,0.5101123254372522,0.6971678170227852,
 0.7800413461999508,0.3755705215019791,0.5858938114456614,0.27062405188593286,
 0.8185487803473578,0.4618252493630297,0.6282598590069028,0.6105097029815922,
 0.8480333454623962,0.8140082530875029,0.1314550556134828,0.5520735133272459,
 0.9482679345244233,0.7832727505771951,0.6359294948481883,0.4963373854370874,
 0.015364811576518367,0.9544872936296206,0.5110818270368613,
 0.11487042248356749,0.45871958433167104,0.958068950685237,0.4186241458952702,
 0.34459223494982094,0.21097440178575733,0.5166398610987863,0.291289230967103,
 0.09402909801791937,0.05291543279216804,0.4867650618858098,
 0.45089391445220983,0.6197475248240604,0.23912127957093854,
 0.7507752453563725,0.36904897731673636,0.9372956331887512,0.6588066139281499,
 0.5535854294077662,0.3849201882656186,0.05321027294043856,0.734618956524732,
 0.08464985871403052,0.6977097402044453,0.2463808242047737,
 0.20754185027874283,0.17733930324723168,0.06768266924198407,
 0.3520690019161966,0.9650569344299792,0.5109620913075374,0.5669782865009719,
 0.21987095664519163,0.7722948457971136,0.06561440282437392,
 0.32386450892577456,0.3554408027612108,0.2947022111329707,0.6470317863765025,
 0.9542867296525942,0.8017856876709829,0.33355168155657655,
 0.15227415294206947,0.8129880188110745,0.6074541210062541,0.9939913681695328,
 0.5005285802936091,0.2102081074722767,0.36928068766828726,
 0.30010659737715306,0.3691794518346114,0.28249520181562116,
 0.3674936351663306,0.4726971904326329,0.4208432492045566,
 0.015218289161899157,0.6507980507778607,0.0663442697149389,
 0.6102876660336365,0.4623079702200522,0.9872190045311391,0.12888322107420258,
 0.23181730454617688,0.48709705183588925,0.4340759370357117,
 0.35322605279488806,0.17264157351894427,0.7752515803733419,0.678376156889727,
 0.6040515487825261,0.008200478691018365,0.26795885838743194,
 0.4824238073262972,0.7154598039954023,0.533004526426002,0.9524316836865224,
 0.3515350582600467,0.7823447396689593,0.3089213890154153,0.39127511393423586,
 0.9974992938391564,0.9872541639147134,0.011770577832563967,
 0.49696511224647033,0.7953139980613889,0.4882155764873264,0.7690116434622247,
 0.3648101236351956,0.4230869076914724,0.08380811869475024,
 0.19257758841416917,0.8829700318759952,0.5890201890644354,0.8978532971862587,
 0.37700498216909095,0.937845372988732,0.7939101968792214,0.3726098434570495,
 0.46384980653092656,0.021503161945247817,0.39532341513330316,
 0.6092006096034265,0.6035835191757707,0.4463270425608673,0.11646422030976922,
 0.31969939656422974,0.3512319530307728,0.27592332649783335,
 0.41239451117347703,0.1215593126953235,0.01119429551855855,
 0.6989885585877835,0.8370349507710684,0.3575334083590186,0.5407551611702933,
 0.5407634206348356,0.5888300299327451,0.9593263740203434,0.7753244270130477,
 0.40561672398645643,0.9351335685751301,0.43750299242929913,
 0.3541698695393264,0.8260458909910503,0.7396083019299046,0.7798184994158246,
 0.05050223241818452,0.5121272194157442,0.27085724047769055,
 0.3277268399011066,0.1835089107418023,0.5436014092681913,0.2296095610057528,
 0.6593159748094478,0.9399909086884579,0.716204367054505,0.9861267620353129,
 0.636828574674039,0.6830135428246945,0.6808899670791492,0.41513854528290883,
 0.16944437717432115,0.04823480500680288,0.8101837671929799,
 0.22163479468078606,0.8367760985112047,0.805368236595185,0.06570405328078044,
 0.2893190635743469,0.9594297407505225,0.499182361796911,0.6454830299571173,
 0.05861221629198232,0.6949323320271512,0.9570507447749002,
 0.07258471225353547,0.42110091928745197,0.8855147836219759,
 0.8813836293889395,0.5041440063395617,0.38128924385654794,0.2192353901614602,
 0.2983452028068059,0.018809544271609968,0.43861523754692977,
 0.015787023131782374,0.842117847397309,0.09180051742416528,
 0.3490389124939052,0.22317631514188963,0.5606734376006647,0.5548920926091093,
 0.6443438921420148,0.9143424237310404,0.8357204113367359,0.41689817557782893,
 0.46453757138153007,0.752406898374076,0.036435040327585155,
 0.7486472635508867,0.4077241475099832,0.7917568214917278,0.6164423822977986,
 0.15767138353476562,0.5360021093901014,0.4196322052438525,0.9527114704544986,
 0.5958210602740567,0.6218130351053953,0.8436826515998568,0.5527912192762139,
 0.9759217183947406,0.41163454653637166,0.029171418497690227,
 0.2630793091408078,0.6689954782621299,0.704607247640086,0.49956504192457774,
 0.44361221830013786,0.51117365484601,0.6802172932907802,0.5115895021026425,
 0.7295482469641561,0.6254969365339598,0.11028604033574207,0.7413809967923423,
 0.40280467326639446,0.4699039262214749,0.4229183199422504,0.8527725582479979,
 0.1640337698388009,0.9194996786068023,0.2892733159306482,0.20526461284925213,
 0.5169728867073065,0.44798456181427426,0.5508221200774807,0.733595954202491,
 0.74197001537553,0.22339287488899529,0.024532053301992818,0.8322437105335851,
 0.28359568224295795,0.7777014017395562,0.9288434953264807,0.814522621333187,
 0.8130491684868564,0.979818340457779,0.41092802062200495,0.8589158006943337,
 0.17676876605654046,0.9768880255653611,0.12868471441489338,
 0.9930543226194652,0.1894755319546898,0.6915961333264807,
 0.048433173371772775,0.35041388751631386,0.5283362752661211,
 0.5617582574303494,0.4151718059938406,0.36881183269312245,0.8627508145660369,
 0.8439240005302269,0.9983867656038696,0.7455644479715089,0.24869300388373972,
 0.7601518036385481,0.5088018704574786,0.02405845829351405,
 0.33060673410512387,0.3872371797872941,0.3719120845981748,0.9814995884447851,
 0.8811501349254299,0.5160587684535057,0.12477878168056289,0.6806757215637497,
 0.440195016146799,0.44929303735757264,0.5226404504588981,0.6220206335019992,
 0.5901328343879103,0.724308624403134,0.4356795924624274,0.07437438288996889,
 0.9984655873478374,0.5479095579048333,0.7593379642202513,0.2808824532244334,
 0.5301312530470744,0.07221896596835675,0.4170914626880319,
 0.09503188501458937,0.0256539060511366,0.7074204971055829,0.9580575607469595,
 0.48308397877480813,0.07153229752020174,0.30589218948046415,
 0.8554830088593052,0.984871288039205,0.6193778024784429,0.3866490473563525,
 0.8223641302468039,0.5467035212376288,0.0785336304666775,0.8351164879635008,
 0.5598818293197131,0.15656239446561915,0.8128255762657615,
 0.45426201751885287,0.18329741696458968,0.6092509958329168,
 0.3322013447391684,0.2493000060652708,0.5744492646910018,0.9896693302953832,
 0.6736202360641936,0.7223424609434941,0.9135917690579233,0.9215789206354174,
 0.6921506232669771,0.29678451830590225,0.7844788020128928,0.8726855928148347,
 0.5175167590275824,0.9231190256920777,0.8040699458862555,0.4061006529596942,
 0.9038363735238983,0.6738898840676035,0.2728700603486067,0.5138886440927213,
 0.10546842360050923,0.21816556274152277,0.1162831170772669,
 0.4889278882218663,0.108420115881535,0.1710062013831889,0.3856002645144355,
 0.6842437644390177,0.7964891171973959,0.4694502544391005,0.15218360309740275,
 0.2463454976195053,0.06940804347166663,0.6366462227885429,
 0.20046227460603683,0.5785737146394409,0.8773194258909119,0.4833873537266097,
 0.8980230427072357,0.743811856499218,0.04313212132795141,0.1893034365197337,
 0.8188266373895138,0.8946838037818626,0.36971868039293676,0.5150319568665258,
 0.5694579104816909,0.748685748662592,0.8932409689531754,0.8719246117878271,
 0.7805268642750658,0.25361944677609305,0.8472743052231149,0.4918298810454216,
 0.166426616004954,0.8089726715037786,0.7424544336828889,0.47789931061490754,
 0.40566684442242706,0.9438174926048155,0.559976843239975,0.3287894788278545,
 0.5228044835000648,0.3963147682295495,0.060816086847941264,
 0.6550163105239668,0.5063743186648031,0.40689748160693673,0.8681313545726909,
 0.647223348399487,0.07593712094312222,0.8421892033612064,0.31830441815199084,
 0.3097762946792797,0.09827810254527503,0.5150776408434742,
 0.10436963094399376,0.5405188844979486,0.13275662662013898,
 0.20906532932120037,0.9718940822119415,0.5711365541873822,0.3015097911648177,
 0.34664152274465687,0.21541923281523423,0.08846599542648448,
 0.37061344678516805,0.16696745456573714,0.8545849290026677,
 0.9453397642335742,0.8226206727031876,0.6951647496790516,0.7842382275776653,
 0.044599998788051876,0.3414851383384515,0.2202641972963103,
 0.8036268620444503,0.98543120427783,0.9753069750258234,0.6329125918266834,
 0.782103608926989,0.5249450610056132,0.6424809537198721,0.5390521984747767,
 0.38520014471251274,0.24778448162502864,0.8229823144910975,
 0.45533447819154804,0.8487584855242609,0.011846899585989834,
 0.15115512856201263,0.902277912466053,0.0781206904351226,0.8644351085493783,
 0.30836612232592,0.9802224883365085,0.4533253056735598,0.226163025551267,
 0.17443883264250837,0.24780056541310225,0.03315338663444234,
 0.09623199974685204,0.3210376916545977,0.07837986361251359,
 0.07312520009637624,0.9600144708399908,0.5645559856221,0.06632702868999207,
 0.18010316883196287,0.3373478562127832,0.750399526558748,0.7620176924057402,
 0.26416965779417023,0.48493930244736183,0.07307538134789304,
 0.7420746782528085,0.9017534999409061,0.7272181952113035,0.8656708906909538,
 0.4263414334047775,0.3179667750563874,0.4930973245964959,0.3201031650791384,
 0.44638056279852134,0.1656594113982386,0.41108399778873195,
 0.2577583597775165,0.6470399283755295,0.400067048316866,0.7799365745157896,
 0.4314627274080792,0.7571869045421865,0.2899899164187857,0.7799413738086168,
 0.48392212458234773,0.7131835593537168,0.21555852182833413,
 0.36784534628189913,0.17102587376901468,0.2648128456740738,
 0.028786090168663403,0.42813590861580153,0.9824265332119974,
 0.8440270515962748,0.7860444300080505,0.3122588918418735,0.0829261606073004,
 0.5324508467679013,0.9072529268507489,0.7884106767650352,0.4246385906649923,
 0.3437948442942548,0.9417556148332767,0.035913578724700135,
 0.5082257658714544,0.6869993343088591,0.6586926029350428,0.3710142606478666,
 0.5527147967794166,0.9533038382040687,0.21897747891391428,
 0.36246671703569533,0.5506263995751284,0.25919461289240386,
 0.1457088193080971,0.35801522339252045,0.45415620267529133,
 0.11725923748179634,0.6299055127548079,0.0784223318970112,
 0.45691806660266243,0.7307135977312874,0.20630763523513185,
 0.4579943851359263,0.14834715478760485,0.19771934338518693,
 0.7759978758419999,0.8912252012362795,0.24060783816287667,
 0.11038792003858044,0.745895624253777,0.8545486072763678,0.6474117161183824,
 0.11228106055841003,0.614430177306891,0.5658268378845389,0.9445936514149172,
 0.3731463942578266,0.5132398670489342,0.18672636191043823,
 0.18632552809550373,0.9218139920988226,0.6150262453197208,0.7201533333058897,
 0.9466311384338402,0.36670814852504574,0.08490303312133618,
 0.9575864587322189,0.7239563582274853,0.8133886630855196,0.5594268426155372,
 0.1352266979118828,0.018943655321444597,0.7901850524017373,
 0.8968330532278032,0.3409386717659322,0.4716113055895149,0.3808578416825853,
 0.6975323911081295,0.1535215486173741,0.7007376479915655,0.4065088658098971,
 0.5334487394147392,0.89655447556173,0.9145017675937376,0.45571389893103575,
 0.7258860434487782,0.5315852475034407,0.16984057324107005,0.2447004372231123,
 0.2728650673692955,0.9767625159521417,0.6326086995104941,0.06598903912238496,
 0.1521406084371031,0.5759135934894569,0.6096164065112758,0.5303885567049252,
 0.20297016347507812,0.42115940746123315,0.3630495399892745,
 0.5679127423296488,0.6449914869149778,0.12870512242946142,0.3389908348218078,
 0.7536180352614638,0.14299685345731772,0.3973316912254613,0.8758369706366786,
 0.4753190181533886,0.6954824494084629,0.7497240974779411,0.5520641018811197,
 0.9225263177919937,0.5810485382688286,0.31111645836086765,
 0.020753907103716718,0.20533035783022302,0.8209034407554019,
 0.2690670536027673,0.05546104289005738,0.3150539698191851,0.8080536425604844,
 0.4518270917765157,0.9768170830033113,0.7513425136497963,0.4350324980589797,
 0.34142323732671265,0.8920982212776589,0.9335974112770318,0.3319075210321599,
 0.798524070888764,0.8232443545957902,0.7036234490116038,0.23545892221025877,
 0.6715841116655146,0.26007850580727954,0.3754455457231757,
 0.28671679780708537,0.2865938345782477,0.5397446677812707,0.7801991967126571,
 0.3800306330747413,0.7233220066949652,0.2560617074169823,0.9905963393575362,
 0.6014771009225944,0.295393822018699,0.5465724001260812,0.18695963880349997,
 0.15587908203326595,0.08196214467075913,0.5666866687388982,
 0.6220282528839082,0.0665101790666599,0.2979653681169162,0.5259656676541429,
 0.1690623038861162,0.7640530234083863,0.5539853395130596,0.589003905283992,
 0.7840717497910435,0.7367477558186274,0.0355945389007537,0.9143564059132432,
 0.09727702263055038,0.21374448865662843,0.29470380828763765,0.94708219410103,
 0.0983877010519012,0.9674659686359206,0.5151950692958756,0.6150876342018614,
 0.8227179253373971,0.6071187169395126,0.5168599819856894,0.03852534093258042,
 0.35799712840807907,0.8561379952669708,0.3381031133895279,0.9051391590523693,
 0.5666011685050207,0.28504043040066784,0.23934225367288797,
 0.7198981206173006,0.283308976866682,0.3692986778062365,0.526055500585551,
 0.5951798751608617,0.1858094610326757,0.034893044439921916,
 0.0751912905784502,0.9448312966923382,0.8418770747404108,0.8033620697125826,
 0.013182771964899476,0.07627843138937251,0.048619607476062976,
 0.02979996091533421,0.03179304790607029,0.6690745144207326,
 0.22968028027329257,0.25801303369755424,0.043275022091207704,
 0.1554351517407535,0.07149007970145282,0.20940954646586263,
 0.2494761021294407,0.7302737218442639,0.5269530384802299,0.9068364878848391,
 0.9287208560318836,0.4303849941259049,0.8198138600076931,0.7421661720499358,
 0.959541235089201,0.11512620258695228,0.05431391180074363,
 0.25575675943858966,0.7437071126282463,0.40656998721699933,
 0.6107991070677662,0.009635075384040293,0.9174018294599297,
 0.05730130527319821,0.12339485737296463,0.5072591746944932,
 0.32116785906351386,0.5329917784853269,0.7599319984626641,
 0.05421262195518306,0.19745732885691525,0.616081625488969,0.6093719825797194,
 0.18181779203050652,0.56315485282846,0.7924480370149195,0.39368791606532116,
 0.36622756744764295,0.5637530234359927,0.595646996981283,0.9854165609090708,
 0.8418573555822279,0.6942962996446503,0.41549423965181465,
 0.30360554611838597,0.7918111323267606,0.3921767794183333,0.5678238444621033,
 0.40760883177703244,0.5704716519997597,0.11670614999432416,
 0.44289063693915487,0.20004135886015717,0.8333956437137682,
 0.8685250691455699,0.03516680559331109,0.9192351150403268,0.6183272395230922,
 0.8566059789168375,0.5420473855313237,0.34398212536396533,0.4530326560279787,
 0.6992476021049776,0.8892338048435794,0.11950489124732655,0.9958029458896487,
 0.044077950583827796,0.7119753714682354,0.6741442709638856,
 0.01838254744302259,0.9267361130487091,0.08789778422637273,
 0.22816464387798796,0.06612469041794977,0.38295510097569163,
 0.24796279458923065,0.8711445573833985,0.7579827920229485,0.5655394947898293,
 0.22499246211258972,0.6995854633820887,0.6438869830551578,
 0.31417015938739445,0.0892700124944249,0.7948114901223655,
 0.24413577641934925,0.20109335512324078,0.45205915908275773,
 0.1993778473106228,0.7282689364978417,0.18558193656102095,0.6272454904924166,
 0.2588468048625657,0.7202441278085118,0.5638795361590692,0.906866551531252,
 0.16328132996366318,0.06508191990469603,0.8159020760544585,
 0.5731042049447611,0.20380083909391555,0.7344862221416263,0.6852496312482516,
 0.8614089042109625,0.9012112868808917,0.27027939929783873,0.7616454819338393,
 0.5310771958282847,0.7656740996007654,0.8485225411663803,0.2449729613011019,
 0.15433083003335712,0.9860910542922976,0.5013190502918233,0.9001813356482394,
 0.22602035939322263,0.720793662201551,0.845925003760589,0.9625615074191622,
 0.9784023708642948,0.2889412796734925,0.548286056022179,0.8191786073987319,
 0.46673552374459015,0.5441313367388079,0.16869938077782265,
 0.24153896031705335,0.11254504601125792,0.17799469932595857,
 0.09903710288572509,0.7899257690642459,0.07416128331126493,
 0.4745780360028753,0.2028056311908596,0.269460076142532,0.8528093204369978,
 0.8275892534137811,0.02381358470536743,0.6607601937968618,0.72146960088076,
 0.7716134097131213,0.4874033459739511,0.4170408810589191,0.569139956206449,
 0.47824549754496914,0.29271316647199874,0.6769536448265469,
 0.16626359123765455,0.8428264527264755,0.9925845564857403,0.5210172300739616,
 0.11536670099079982,0.2953858624037571,0.6556636874741785,
 0.10861936221093049,0.31000135674458107,0.7834147287217469,
 0.9415737621061014,0.27635844937232745,0.5883879659391802,
 0.39236870074719865,0.885979273969429,0.7271424017548144,0.6104723559100383,
 0.3573719629573415,0.47709563001040656,0.9408091673238397,0.3882345802861812,
 0.1702717139576344,0.7903866142134154,0.03160662588042962,
 0.48202108426157664,0.3933155061127064,0.3606287542955199,0.179077878059703,
 0.3110592315381844,0.6828793590385333,0.7240036796735493,0.8204632222769318,
 0.2843320167227945,0.30408534252421615,0.8135712846628181,
 0.021537880084603334,0.8610458284998898,0.6691763284341892,
 0.7701753713424638,0.23469290183486846,0.30332304927924747,
 0.1000413739763244,0.19672306249105387,0.33746425606427544,
 0.2432639225759099,0.5646735233413589,0.11888289461984936,0.6861072196420143,
 0.0022802295251334037,0.6025264415823022,0.668400594150179,
 0.6352622296360977,0.16547558914653315,0.655841699503738,0.7921049848554211,
 0.5413272491401677,0.6358597734655697,0.19871065346941097,0.3586615241507176,
 0.7844639664248788,0.34662271396502986,0.6711020459984697,0.610676078514574,
 0.5337134269102443,0.46524128956802757,0.2556931175677103,0.8105689953092716,
 0.5364637752216823,0.971299937635065,0.962416585687456,0.7623748615349026,
 0.6743893945493817,0.11855905022701718,0.2824956863207647,0.9949803756985186,
 0.610796283257529,0.22144933704425496,0.6945330190033387,0.3485231553881455,
 0.012538480566933852,0.39296688067802943,0.6879069555322048,
 0.20211969011072917,0.1140869982554431,0.43977927705883746,
 0.5471042588284415,0.1643072182994183,0.2157323875133198,0.6626726523374628,
 0.11728124441264398,0.023359485380080702,0.5490364944897987,
 0.016036470378954615,0.6947542556449382,0.9615990890498712,
 0.1736836074411141,0.264627541559715,0.7480193686372796,0.22424351841077417,
 0.890424794000165,0.7894092456422988,0.3358375084287958,0.30395697762503315,
 0.5270032679302719,0.6845366757610752,0.9116240880970503,0.18487068959812192,
 0.6098081858508103,0.6240839833453757,0.7840041562405045,0.9398541703828812,
 0.12857577380739804,0.17304933463275496,0.44842361228777994,
 0.772880111940373,0.09981082344728998,0.661641603274355,0.8593211611281073,
 0.5664766609687707,0.34205162210274265,0.4972739316673884,
 0.16781487786668525,0.7119368909450194,0.8517781185902751,0.7569539034293807,
 0.4143363366735806,0.9071921406205156,0.2429662230228682,0.7059839625223565,
 0.18769562800319617,0.29341926050956313,0.4205176356915792,
 0.09332884436596034,0.389111527303194,0.5259814527783694,0.484992096826778,
 0.10878474278122907,0.6138223324083738,0.13081355461486233,
 0.23503166339024095,0.08778369347208181,0.0819119262410577,
 0.6299631106818043,0.5041790119411027,0.1351924503463151,0.5572882585882714,
 0.8951312249132038,0.8201499383567489,0.2749127224038882,0.25893429563371473,
 0.06151892194029318,0.30819757512011936,0.9864854053608023,
 0.06412808345946508,0.6929617770521599,0.8824261653887665,0.74562698094274,
 0.8090029563133477,0.9371192729491375,0.6568284887305118,0.8256668866816065,
 0.7426752493303264,0.5508343538324287,0.4477848160188421,0.0730723121938428,
 0.5211265194228478,0.5152982330881459,0.8023226233067509,0.2934470628121809,
 0.35237144591343683,0.6616273657084581,0.4412481659903208,0.6132978113512166,
 0.568165835870486,0.06103702446896686,0.1615671286981406,0.04921211738467002,
 0.12552411903196348,0.9288987906616051,0.011993279600538531,
 0.8257341867669246,0.7786603758935999,0.926645297990422,0.232156811300963,
 0.03856752213167902,0.27004973109311037,0.5766658466451426,
 0.9986533717329332,0.25472710353439,0.131023835406613,0.5897065144898269,
 0.3663431925712842,0.5534271058630473,0.4062783970667201,0.2937009259455867,
 0.10892433943958357,0.6061418687503543,0.4041279844826744,
 0.014680385734218593,0.4044839932707335,0.21050730835705078,
 0.7755745021636211,0.7220650845763864,0.5467677637459023,0.7716606019400306,
 0.6821978323335162,0.8870588415717281,0.7488814904509098,0.7492166787254158,
 0.5718619467352282,0.05388019541731348,0.875033271075341,0.35158262743066937,
 0.22524370013338701,0.8932572542954457,0.5269463950821456,0.4656382315895724,
 0.6826184923160507,0.7575146097005461,0.1929253341438858,0.8981067940618395,
 0.7268140753115292,0.6393095610378088,0.6097944402868725,
 0.008644071169171985,0.7189872755740908,0.6671364350347805,
 0.9405605949705708,0.570911199049714,0.1953027967429316,0.8599966144486378,
 0.8378783346750018,0.44265494473013534,0.8254688539592594,0.7760035523458101,
 0.2104558485031991,0.521556073878412,0.9671701170499456,0.5427328613839227,
 0.3373794271781403,0.7510704234465257,0.16431600039600802,0.6706735442502734,
 0.640923608508511,0.6989985758960402,0.6248140278567956,0.6157953137211336,
 0.7374912447756898,0.07471413258713422,0.11232231360467826,
 0.5157060690907077,0.12835967412212934,0.26816685361297676,
 0.4701500267757215,0.07487960308319308,0.8666840487668758,0.552384197802272,
 0.0595569420044233,0.8991988703349763,0.8277396812545031,0.9360165515173637,
 0.41021106995715095,0.5615024353343225,0.7395124421753297,0.5985085572310553,
 0.9505720527226897,0.6864527814360357,0.8515754098346408,
 0.007461758776080796,0.8836114842946956,0.06566619299133958,
 0.3112459081536283,0.7320276926248479,0.6398618745479522,0.21836117431592017,
 0.4048121417227587,0.7579444127726884,0.04449249258645094,
 0.08876086439749131,0.9029247752923459,0.5820082997201065,
 0.22158277112049374,0.807389188218846,0.6979504873026514,0.6651296667271275,
 0.10935797956238902,0.13133662218174103,0.8498201194498441,
 0.9086276117587734,0.4389219475457611,0.6323580712228949,0.4752716084630898,
 0.8865500420894827,0.4499394041425069,0.17769741281608487,0.9759693973434403,
 0.8157403203067731,0.7699452898204082,0.558311385938953,0.4210024268634731,
 0.436844355367244,0.8160976286929684,0.3136631936086396,0.5565886462878042,
 0.9937950765903883,0.8383750080075062,0.7846009002997942,0.16004020714111566,
 0.03331691583775631,0.969618982616216,0.6246353426268696,0.5692189159897233,
 0.6228688390460738,0.29852287609441575,0.10928138015022604,
 0.21609628323521646,0.24847270839646107,0.7562424965671378,
 0.7528392361759652,0.542414036661629,0.6062793349490602,0.629771170488646,
 0.27376739418174567,0.5391600761497001,0.8533974754758056,0.4595284690182978,
 0.7893414065988927,0.955633529122328,0.41494829375601616,0.37380886974098093,
 0.130713533521984,0.7786274500207995,0.34404793327269834,0.9715805455225092,
 0.5516181183083452,0.07323761821337482,0.9301814491849658,0.7577692260953337,
 0.17153352581926506,0.7988422482773342,0.13783319562763086,
 0.3404198114938415,0.10327393216014491,0.4291255332841112,0.7151152925688845,
 0.29770158387275025,0.36699814499186134,0.30255484446313863,
 0.11302368242202798,0.22507154167922183,0.2200617085596751,
 0.6937739068465281,0.4392347094218719,0.5964124250204086,0.5997683015465719,
 0.16234864780909553,0.607522213819192,0.4081714452477557,0.49347871091758466,
 0.3126089666839291,0.08372086053202843,0.4643186599272019,0.4861567091559894,
 0.4672873378105482,0.8711481136563722,0.6927161200747854,0.682716418040574,
 0.08805584958674306,0.5412929252299182,0.08245647679604085,
 0.8680097163141622,0.04073508462144604,0.731008827107272,0.894162401101308,
 0.2919563674466319,0.7293950209213764,0.29057543799914587,0.6397061396111143,
 0.4018010832227227,0.5978213229010219,0.7850315580690191,0.21764966156146714,
 0.5492281158807529,0.5868094117662769,0.5285018394593249,0.7106436435516654,
 0.8807011552356996,0.10014416360670775,0.05058958230484745,
 0.21450822834239247,0.6123373978884139,0.6934246995057975,0.5384769919067147,
 0.7962641858700527,0.0015191917222174833,0.5791394715477514,
 0.23639327671810284,0.6287777394520241,0.6571191262690399,0.8659120312533868,
 0.6144960136101286,0.907147816223355,0.02400059288751033,0.3182573562264498,
 0.575556983104896,0.12530047363606545,0.3192262388737147,0.09348205840051338,
 0.27323346555315364,0.8493987567397037,0.27613959821808676,0.60784412460256,
 0.26679951809898084,0.4666049338208408,0.6071832402338194,0.9781072581880201,
 0.9056589288441701,0.18361908066526955,0.7240720556204377,0.3048711266140963,
 0.9379582984855636,0.9037812812484186,0.8671233678374352,
 0.051555714927369656,0.9464645952792493,0.6618402786908087,
 0.7563996460938587,0.7860826333422062,0.360146325304019,0.5888148908352986,
 0.6336632844184918,0.500996108576376,0.370206060766507,0.056947548635882694,
 0.6626983146979178,0.9094583117535294,0.6937587664951312,0.8764219131686224,
 0.9410777464573229,0.3268646271755359,0.30941795857655174,0.8405191035646683,
 0.7287403806093622,0.08326004707622581,0.579484895597442,0.20545455929007805,
 0.08795786845595444,0.24745392474828032,0.6705342749266793,
 0.5179181561628682,0.8362880510175935,0.9109546089239347,0.3525660029162435,
 0.21294897167969395,0.762259504477778,0.5403343206080766,0.6741785586174497,
 0.2301729120956424,0.007755308711175912,0.5025433425310564,
 0.23658452220928616,0.18187371092853621,0.7412925051789581,
 0.08747758291959085,0.635237805532815,0.22545464257987935,0.8888319594704516,
 0.4449705051521684,0.09118602517868091,0.8861395737138559,0.8428462536992508,
 0.2207539256965645,0.7170432721131489,0.2725473709202332,0.5754382044610457,
 0.24599011156763317,0.8429985595942099,0.49818937408444375,
 0.7484379031088391,0.9009793544532755,0.660407259042117,0.5493625738074808,
 0.6378747122282298,0.8181265224446714,0.16938306570946715,
 0.010572918736550974,0.3013073749885262,0.3064042942932056,
 0.2730948537279584,0.5808821603838974,0.016561226637390347,
 0.17154526689350846,0.9237648253270668,0.40528976134409134,
 0.11103915326140645,0.5269126090141685,0.14143847032820278,
 0.1381526228039448,0.8665438195075167,0.4646768102626957,0.2152747055816977,
 0.5905930626508271,0.07225123137324063,0.562233358511468,0.24603213339672036,
 0.06737448109034594,0.8576343935148507,0.85805876896172,0.7048651624972754,
 0.4602515463569158,0.07738616506629326,0.35281390293344106,
 0.4616641244293872,0.7156582295482519,0.4554213921772752,0.10160644869544277,
 0.48575407096768053,0.311991339128908,0.1463326661660499,0.9847687796338458,
 0.09781030108350275,0.33835987286992075,0.03460049240496199,
 0.9582235717995086,0.5452327537420971,0.6308039775800501,0.35200022954394017,
 0.6882093398793571,0.35815414210802243,0.10054891278131639,
 0.5204112241740682,0.9415412672026698,0.35907310178206386,0.5485390406945696,
 0.32652142694919817,0.7008636606424672,0.11382740157604332,
 0.9203102538750259,0.4126785763178711,0.45827140219176543,
 0.27590561363600385,0.9044491428367635,0.08917144954610612,
 0.5732131853383946,0.4418569068206438,0.6639859032606064,0.26516895639619076,
 0.2625065710209675,0.22051048660306471,0.824152476249497,0.7859354084591761,
 0.9284532337816641,0.7885588967977435,0.011287863166993484,
 0.45510358793986505,0.11801169933787925,0.7361569045651197,
 0.8896255478274631,0.6866430227709562,0.8974208946180275,0.7743429131337587,
 0.9607271529250533,0.5000609548641569,0.05980584207236994,0.8299141722675493,
 0.41954671445654834,0.7749861328976001,0.9230278125213544,0.8627794832634894,
 0.8797221586508349,0.6650149148822676,0.43695400399022477,
 0.45181833862729626,0.105153317008715,0.7185500037038565,0.936760493484464,
 0.454083705678179,0.4887262280396797,0.3126414320635804,0.0450471658224485,
 0.3869866109727095,0.831406450945649,0.5912868039791583,0.9251117542146821,
 0.9592172822982804,0.9756704054437898,0.7025087630761165,0.3380735972135145,
 0.15948351241090342,0.4912661239963412,0.9354148337728191,0.5994315421703548,
 0.4437628590410385,0.10828776927960737,0.30678970113455606,
 0.2926012476432155,0.43081630454688047,0.8148405739440532,0.949326175668014,
 0.9971953405822093,0.2652587978722811,0.051174236398667916,
 0.7667513971506611,0.8630526248489013,0.1161706641396627,0.7469770859112439,
 0.110131184181671,0.07414698517925411,0.8452551034032828,0.3121846390456783,
 0.49200931208007304,0.4196630030208889,0.9740491614508008,0.6257951105085244,
 0.19826283021654345,0.0032414036247658817,0.9039177587052089,
 0.607930756886209,0.8756565441801707,0.8683886521079224,0.5394598891542889,
 0.36400048594279444,0.5926362536487182,0.33379024270416746,
 0.3822223014130902,0.9269241312996837,0.6493568544406871,0.8793827325552612,
 0.7755298573693031,0.6517418530527981,0.5796321723026855,0.1996254405739316,
 0.37268385646622204,0.4019663139882268,0.2521029206111216,
 0.08871120341775474,0.9283018413661585,0.20666675725210049,
 0.1506974459903253,0.8311837161318018,0.09908964319776914,
 0.25120049166038017,0.8621296664202911,0.21984726736926208,
 0.3129485096926463,0.047105171710238536,0.11747435181634691,
 0.9341062807371836,0.3252952489083172,0.07019640334945021,0.5580206262119236,
 0.8268167416137648,0.052971318049344074,0.4310998891830701,
 0.7575866052743492,0.5041358078195664,0.0551412376872199,0.09130513833649989,
 0.964166887061718,0.09132287632372615,0.2672540667310701,0.47197076973773167,
 0.211529188575996,0.15514231729196937,0.1596535951498359,0.9517929327325237,
 0.3435551854486798,0.4366079688181923,0.6127224061893145,0.2826279615419074,
 0.9083445256598202,0.5457921353568078,0.005978452710232562,
 0.8637038165936586,0.6284279147609722,0.32384157209374553,0.8693954655161922,
 0.5058314636611381,0.6382840670016842,0.9539810786128039,0.6221116815543758,
 0.79444119578146,0.0018771911109796147,0.20786037961216353,
 0.7778268323318579,0.7311897726544418,0.3139964587461431,0.19315090043649252,
 0.48961247643248684,0.9481826628443067,0.8872462152476333,0.8956492373739141,
 0.8669276168046086,0.6708719811369847,0.2638837838596165,0.6551153833706697,
 0.37196135906330985,0.5347159551462766,0.06331323995710259,
 0.5750618468364491,0.38773063261253027,0.26133480577879575,0.746753460035301,
 0.6331559197639287,0.9968556065538494,0.34662361769433336,0.5681287650903076,
 0.5761739735879502,0.7868888071894744,0.2531470938879037,0.18226224442475392,
 0.7212947088783935,0.40110036549869854,0.5806227648345307,0.9633436887052694,
 0.8473990915218608,0.8052772729531079,0.8416279371611564,0.31600786490783206,
 0.8864888524691187,0.673740008613301,0.9062909506682041,0.7071939680811403,
 0.44711990910301436,0.30294864889905215,0.9950290114765415,
 0.9589419022822252,0.310559206129289,0.9819194804299359,0.006714018663687704,
 0.8912453123527609,0.28937239356470124,0.06991725139618787,
 0.6196678673481887,0.45217535891051563,0.21990776129169376,
 0.6344649430753609,0.13236477512103517,0.8417820730133082,0.6447038470299011,
 0.19364874871496918,0.9733623589279151,0.9074522874727289,0.8519467772941497,
 0.9454652123369386,0.7835528832171377,0.32405963305067176,
 0.09719485940470252,0.5027946147632013,0.19645010387460915,
 0.5922511875134668,0.3104963168298084,0.5393658279169071,0.9833586318208032,
 0.9117647933918999,0.456953068810538,0.4207171184207613,0.19067130456222214,
 0.8599579375891829,0.9473123189699276,0.5097435265277591,0.43304487703770567,
 0.03146519324382646,0.4379700294285561,0.2793500821750162,0.7909664371241572,
 0.36699536894322815,0.39000271841968814,0.32101453125529456,
 0.5919186123785335,0.2613744984532843,0.19187691171604349,
 0.06986214038613547,0.9209325082036415,0.5327213490617978,0.6392078431586191,
 0.9115051543058346,0.9020828907851224,0.1200831986167391,0.3770033416794163,
 0.8131512986177992,0.5817953847815603,0.8049579520320493,0.433005424421645,
 0.012304617180004973,0.47117203622806525,0.5751761987810595,
 0.3340877782167423,0.05835043011397889,0.29313381694661456,
 0.8419976091641594,0.3489289048036136,0.7328930419946587,0.5557950600855247,
 0.4056851861903058,0.8377137374596413,0.0588644555704404,0.8159328683482446,
 0.30030655623502556,0.509721246507874,0.7790411866961724,0.9288592244262299,
 0.9303533900890646,0.17863903680675852,0.23868606134104553,
 0.1083327643193217,0.5347312928969805,0.2366229861534228,0.226347522418478,
 0.9339570328538502,0.11977990391690085,0.7405634165090211,0.6339372938684219,
 0.5698538004166732,0.8759384441520116,0.19907281610324123,0.5698568068803666,
 0.7771540002548321,0.6308384181450949,0.5944515717566397,0.3691174399486753,
 0.1611184428274225,0.9096458778317393,0.11493935187628379,0.5900710962017192,
 0.26511124931000507,0.7243957390576742,0.3142831985844379,0.7972492935625758,
 0.7473648888666414,0.6314277054630815,0.5836055932446083,0.8483113905834714,
 0.0029420583682974177,0.24793717335614718,0.21007523162712893,
 0.891432157784344,0.5819024892098622,0.21733547056709646,0.3732336870157714,
 0.22652122657434282,0.3214521240363939,0.4630279071027379,0.5527616933357083,
 0.3213675520866549,0.38890607069114214,0.28928014136189195,
 0.5524716071896341,0.8484784894322717,0.717210230592194,0.552574310568814,
 0.7503568057857666,0.21948491215785337,0.8069824034265359,
 0.40764024681207534,0.6032611874142262,0.741186332922436,0.49831704461667903,
 0.5608003090550272,0.3761517849496504,0.04251813913514657,
 0.16012444654031244,0.31449950856182196,0.6841358452168848,
 0.6413782438418785,0.21762963566414384,0.534635482518749,0.8478058893800607,
 0.9407650108567338,0.8605133557428686,0.19948000934611443,
 0.38217020199438956,0.4939647418692117,0.12369576124315262,
 0.5708770972098195,0.9458178454069783,0.1654012754938553,0.17080353811287652,
 0.31317923329954267,0.9593469372215498,0.9479687737636406,0.8819762467983736,
 0.26092257064010216,0.6605187199227597,0.3583308872534503,
 0.19464314358075252,0.8809276175584074,0.3495386804742361,0.6192442884168506,
 0.30829921382781655,0.6293753439503206,0.31874294990470264,
 0.02182187382085321,0.09807061556591623,0.583651210524389,
 0.19313646145691765,0.981557249292295,0.9388663450062968,0.17889559387888254,
 0.5830285098171457,0.2349916903960051,0.21275945849549383,
 0.43668874850838535,0.6768590906241151,0.29511555420713464,0.770475587591945,
 0.7522380122460524,0.8712074340197385,0.776499384537271,0.8973201223445502,
 0.6178184508001131,0.5049988804344269,0.3105716176671037,0.6429636800782041,
 0.8244153152962369,0.016149562544477525,0.6101863685532503,
 0.14634412306604394,0.48936378142560144,0.5421294890065685,
 0.3793243469178885,0.6448237098576981,0.5442714753519632,0.32336290861206696,
 0.36905331927469454,0.6340163245186342,0.2698240137951222,0.7653657363363431,
 0.2704732640165336,0.14671214811407263,0.3704210760022466,0.6935085003400676,
 0.8848473710987241,0.451283009558358,0.9941575241095589,0.9856097409644399,
 0.4164703235456473,0.0680361790501749,0.252472910362898,0.05143130997528145,
 0.9111760834710785,0.1574986719027861,0.1620095075655098,0.3478671185545795,
 0.5573872335001648,0.7392188704081424,0.6671877855245879,0.6790479488351188,
 0.7521758222821862,0.5235146405035024,0.38927473924337186,0.249767170794446,
 0.26658238075520013,0.12340678875966615,0.8866003139343699,
 0.9815332422849128,0.27013026197849366,0.8299410163960049,
 0.061179333439694394,0.7106957094050053,0.500806583441734,0.5844147738032397,
 0.5055422712092882,0.31398535065828104,0.47341713258309515,
 0.7578421753197145,0.056559999176550524,0.5598530412766094,
 0.6531823419026059,0.3205219974890472,0.7825482523061564,0.8782107916355396,
 0.14738585885831168,0.10288362204569979,0.04235370145845263,
 0.8670528888797745,0.23333749983125007,0.4811351555787632,0.6668522543026142,
 0.15835751288554922,0.6295298988262203,0.6418254604645849,0.832074011502931,
 0.9565223216740173,0.6173032568423387,0.3188453145009993,0.2546727607611572,
 0.8307176325116529,0.4559014069370424,0.19795504211166204,0.543554959294583,
 0.10976615047744587,0.45969772758640914,0.26501404603538203,
 0.19548416922863432,0.020694912200723503,0.6693632934515078,
 0.047063095804158794,0.2003506177372204,0.5990978272679983,
 0.42684698260882725,0.6660598495781002,0.6114548443036524,
 0.22014975393896175,0.7128943825228463,0.30725250281839134,
 0.04615948659078328,0.45774918505490825,0.11943726442776204,
 0.7581832414477157,0.048340792790135256,0.36132975208829043,
 0.6259767912058274,0.20945281508656788,0.21831304282449826,
 0.7327273955983735,0.19477778570104798,0.7314688382587174,
 0.19715313357331343,0.6736882953568355,0.39523448561951346,
 0.40733959683495746,0.5010706847143438,0.030016346425283058,
 0.6396094958037952,0.10889597862772638,0.3013484308456553,
 0.49539860757103416,0.5501338616374578,0.09347271122703937,
 0.32538076112191094,0.8890009519799399,0.9819556715030292,0.8654676824891107,
 0.9806719328850999,0.721520011296628,0.9494094789825288,0.5483460100902119,
 0.9428559316438141,0.7184112712552964,0.6264865145064212,0.4197976679535713,
 0.661606281029763,0.5882216040835794,0.726378790064601,0.45984362557027614,
 0.9128566338947014,0.9404404540466802,0.9537353342297447,0.8501176356761649,
 0.5026328506459475,0.7380370360550882,0.13493894008325735,
 0.08580175940672496,0.544034071681625,0.5615784866286482,0.49579511036561386,
 0.17725262120237617,0.5660881617869641,0.7450515963947323,
 0.36270154903870333,0.4213273301496211,0.5951849675270459,
 0.19001815120797227,0.913369723641634,0.7624806853057299,0.44734325251247287,
 0.321106389211656,0.23823910925776037,0.6101231671709438,0.17678580723423465,
 0.3432682608703109,0.37858379922836394,0.4320717030326715,0.5028433870428376,
 0.8102771154860047,0.6240482034228858,0.9295015940692142,0.30593886762117384,
 0.26311281287110533,0.889985474897251,0.7832007825431689,0.6778017779789012,
 0.43800272182789246,0.40831235021291623,0.46582490114385,0.05617725188564704,
 0.6200427747157371,0.22247987381630363,0.582681793472889,0.4243905874907805,
 0.31713251309906887,0.8795018974150021,0.6622858798279805,
 0.27388948952805015,0.6597845036050358,0.4465813731036615,0.7316811711814698,
 0.3222376581894303,0.741169621654063,0.3527759733798701,0.988297338768723,
 0.9347184624706714,0.09856843995979214,0.6627992779421618,
 0.33902446379623774,0.0761546617539206,0.6723087574230228,0.8497281115073395,
 0.8859449711649136,0.6428211586792614,0.12241274577621608,0.110847778352957,
 0.7230606590999868,0.5013706962459188,0.7797122654800348,0.8897710621474877,
 0.4072026781176121,0.7245354310173323,0.3909779593747156,0.42302937370138016,
 0.6236146635060443,0.050331442720370934,0.002214676816519545,
 0.835933609098806,0.8211584394653634,0.90224769174434,0.7274706249720864,
 0.7967406610136231,0.6648990010538485,0.36991321664370125,0.9637552881793017,
 0.37605468219392446,0.7310927428963806,0.11971169801215042,
 0.22617774820527126,0.2475261309038933,0.4989377246003541,
 0.019077953548089388,0.7551035240159821,0.9221122797505376,
 0.9848997710717888,0.2752550120353964,0.05888536064038752,
 0.09618894309393655,0.8617246937391083,0.2688084151370179,0.3070421935806086,
 0.33837573238669383,0.9796712639073838,0.45084460747522337,
 0.21056068076214152,0.9222567579202959,0.03888821090368133,
 0.8453021995708421,0.6265388299434601,0.10768204245194579,0.6027488418633321,
 0.43981640045334813,0.6495449513356601,0.7301345793097807,0.5876166270152569,
 0.1479266976481547,0.6042598229556221,0.07708973304983924,
 0.06440518070606915,0.31303364349894947,0.4005480688630929,0.761051142890599,
 0.6996674602965159,0.6851716178886019,0.04177245320660539,
 0.38016861372697064,0.0381624960982444,0.44753663311518477,
 0.3469778850652083,0.9906622600266108,0.9888106235782037,0.750835964940191,
 0.21235970914917546,0.39742488372223717,0.49748948326288067,
 0.20429154066962774,0.7944327100464852,0.27345743893427166,
 0.5931249932772968,0.3059145838723881,0.8261592580240826,0.6515157178505695,
 0.2706362281216639,0.8448620385669013,0.5749770229611102,
 0.018392146655226815,0.7365226977965879,0.33138041981456245,
 0.8264970244878481,0.6826204433213842,0.27581541999448056,0.2702087952297564,
 0.42513490127211107,0.7184338501859711,0.1176908726991247,0.9703631026596993,
 0.8676468186764937,0.23625920400224687,0.7942879378818662,0.9741940055794005,
 0.153494331410147,0.6032229267751108,0.5744039679298245,0.4719062664095788,
 0.29255231348525246,0.11758079262563981,0.21773343669670053,
 0.2617647345082913,0.1830381747935904,0.729868582948186,0.5625637142936095,
 0.6535751930394591,0.34564740684148343,0.649293385308831,0.8184842458118864,
 0.03484035525718365,0.6019990455171842,0.36493626892885833,
 0.9432103248837638,0.6373604597523227,0.9265572479916715,0.6914957565985663,
 0.03671797885407424,0.2924932709258984,0.6765522150947941,0.8955940298023426,
 0.10561936556707374,0.5368153282185906,0.40134396224400715,
 0.4569524016445454,0.22881877867851508,0.1362218679636612,
 0.15229777047363158,0.30039861744984653,0.19370964902568177,
 0.37891030955685956,0.37200554322172297,0.5430724862701056,
 0.6021298452320876,0.11776934633571157,0.8348475012600749,0.5212345090219519,
 0.14804507635865294,0.24122393577675605,0.7818831941971492,
 0.9393035952672962,0.850405984498405,0.06982124042074933,0.6112074394505597,
 0.2781223839985426,0.46604796386465586,0.5581037699575491,
 0.041633465291881055,0.47938163563459346,0.09988391006225072,
 0.5326209335214006,0.23295066382087803,0.4460223658132767,
 0.14499432580265803,0.727422783988965,0.13074911368284514,
 0.17601433375491593,0.7919758980642133,0.023055248825683772,
 0.3752793311828202,0.5651667900475154,0.29798562068724743,0.2539712354142385,
 0.1004719486615151,0.7513303381083412,0.049161754795244095,
 0.9848502295121903,0.032724922592036076,0.7108272895637859,
 0.32962139972133264,0.15927070064926185,0.3139624762182458,
 0.9306355899432054,0.6803854950985815,0.8654697485672508,0.1840143591564506,
 0.07411388907740601,0.5001481630780218,0.7610393655439321,0.8160850516261019,
 0.39497592757757904,0.4593684698955969,0.901824593057948,0.21492384511877793,
 0.6523892261248008,0.22129095308234914,0.06609922922919464,
 0.5083918106178094,0.7689816548396932,0.6209876702662773,0.46082948894596354,
 0.15680854653650989,0.8110577532115014,0.11159741831742409,
 0.3541133334639439,0.4576300418759336,0.42254262937780884,
 0.12096255998823191,0.3832610505340174,0.3069425279635718,0.9482103637456726,
 0.3838226300425954,0.508137686383606,0.9363919544106917,0.7036286342883125,
 0.8082494606036967,0.6257639587420565,0.2046823235373716,0.6013005829002867,
 0.14161834195374778,0.6536151314303082,0.8601861414805612,0.543964963220402,
 0.4322100079530835,0.9232999224946513,0.7233511117754758,0.8722071325582434,
 0.982739820489359,0.4881257005464035,0.7785731853110244,0.8215942625199976,
 0.9983070724843064,0.3813947433148007,0.7914578440510406,0.3250818673615573,
 0.2992773971786318,0.18344888530433257,0.9459230168213107,
 0.013448524943188489,0.6720048225523301,0.1476817081259163,
 0.4126071755581959,0.9920154806095578,0.011922424072443594,
 0.9765194742441906,0.41618078410355364,0.046882453683305236,
 0.28341088723369534,0.04337737245900919,0.7814426211173497,0.986726863748133,
 0.39104518395002685,0.8199345555172276,0.8040199190730895,
 0.40505959840345584,0.41972030143249395,0.4373654681107406,
 0.2756783536259284,0.060315192251394145,0.5082465163477795,
 0.22524152461227054,0.7099559873485607,0.43000893752991387,
 0.26477840158384547,0.38956561460496564,0.7844656582266315,
 0.9437837035847383,0.8675177930939821,0.07748845754645228,
 0.18399205516774642,0.6224211694217352,0.30069157732960194,
 0.4194474753690586,0.7523746028381018,0.8842705976115983,0.18085389599310298,
 0.3512675489595807,0.16381247394315046,0.5129413408432348,
 0.07492325893284157,0.39806040101037077,0.7622705071334993,
 0.21937718554922858,0.08329961477930226,0.9078886821503768,
 0.8501375431615067,0.22638932446143611,0.040395365552904794,
 0.5539585216790647,0.33577686760319003,0.6218644683411781,
 0.46960253764094617,0.5720139106566939,0.38454306565865803,
 0.13454577066988438,0.7212234725313313,0.29092474658875345,
 0.9465230424756812,0.35045360643913814,0.35648992974358207,
 0.0935941179832589,0.9528081751297068,0.7264927751569009,0.5349135090315793,
 0.37833164248060247,0.9624492279575454,0.9655307351131411,
 0.48781844373468275,0.1483998708449561,0.6539357985433347,0.9404508392997025,
 0.6223297852423241,0.11349084407671384,0.11889094471911787,
 0.7120914582975165,0.5043279180048902,0.32399325292796055,
 0.22856499332928482,0.5423562503339655,0.7873958514728648,0.9764927295084181,
 0.11422722010333375,0.4070512460114797,0.5995431241394767,0.6487735296006971,
 0.29367468214718795,0.49624921977549197,0.680807186668047,0.3695237478738945,
 0.17164599325091623,0.8872639031564697,0.9827502977177793,
 0.13960804834798224,0.5586098897066771,0.6751184837538831,0.9205103050097815,
 0.9816608268828254,0.46799773353117446,0.3731500818594712,0.5446368871109432,
 0.26630141213745506,0.9738774705545354,0.20210255553328227,
 0.5968867855720087,0.9254188929090663,0.3088372664778345,0.23171851474860972,
 0.5091149310604938,0.8168647931611281,0.37900931767632084,0.2739659609617363,
 0.08618980944898602,0.8052592231500632,0.04494000598316039,
 0.2600106843605726,0.6086542781810705,0.5950092374296286,0.19296530240568677,
 0.39746891928635153,0.6213967624225916,0.7880584840142817,0.9113528657763543,
 0.6069649550152115,0.8831425002817591,0.8625724803040165,0.36183893007171886,
 0.9471473601884223,0.3772841744488774,0.6270855626147906,0.77330958677896,
 0.3068583791705449,0.7550262514881637,0.7434922352781461,0.8960278470068825,
 0.33206464530898194,0.16542184554470274,0.42750463344503964,
 0.39793683449847295,0.013426185688397041,0.17154915289396155,
 0.1818890403563692,0.547069500223685,0.8294420740771135,0.25439255129202776,
 0.52082270631069,0.9128760596758122,0.3860429194149374,0.06048557473578986,
 0.703823247646474,0.7462607373685566,0.2071702490016678,0.5919195069498571,
 0.07492639099845022,0.5578243320986475,0.4131393160402994,0.527982772658523,
 0.9309048383989076,0.9183993029191316,0.41689296178662993,0.2629439463362174,
 0.10010332287389767,0.25052516865482866,0.9920098568374296,
 0.6953804783213569,0.6583590266518362,0.11384217521831053,0.8586972268239831,
 0.3132355648699381,0.17359452471282633,0.884597198067343,0.6349408147394103,
 0.10284987973583282,0.7695255805433793,0.24251478196570542,
 0.9956140923902459,0.34469316827197605,0.6472390830896534,
 0.32974729271421643,0.9189488743841445,0.2618484644470811,0.9095690354947268,
 0.38593316683271406,0.9619003457200307,0.13073156318593926,
 0.7328797105420843,0.7778106740262596,0.058006937622420374,
 0.5032416153677577,0.24021689206325703,0.08454978478301578,
 0.4779145732964528,0.9661588856590818,0.008687762641175767,
 0.5584090324955687,0.4214830119820938,0.4342724247881826,0.07559339336875281,
 0.45268777833211116,0.5507061773972491,0.003175843956816027,
 0.279192259859922,0.061268139742227556,0.2459141376066537,0.3072832034473334,
 0.4973800687403528,0.07726135304280968,0.3825343752478382,0.9961664324578737,
 0.506008726739562,0.4332728520340797,0.30072165573458365,0.5579082281592098,
 0.9026647805238949,0.7576118719075333,0.05687037976280518,
 0.18410739439645862,0.7122002726231832,0.8385624625624867,0.77098271921624,
 0.9234037195437302,0.4660184624789041,0.7957044000766246,
 0.023602636744316063,0.13193384283382703,0.9535760036926153,
 0.22974938201374628,0.8855623434487822,0.724616583905946,0.29985828409807347,
 0.28717375758790087,0.6330407253125286,0.7975767749234679,0.6716879670959173,
 0.5909769485160908,0.7434214544675696,0.9447615956890688,0.08933607121252807,
 0.7994862041817845,0.36426442125910274,0.842969591618231,0.40262547424720463,
 0.34976356017941157,0.7507043834353886,0.9540570132934041,
 0.11541384651234432,0.4182315098689422,0.9059095434664599,
 0.45203184902631155,0.9918155743867516,0.9130579124994189,0.5640913041919646,
 0.9328901999489906,0.5542767751067789,0.06238177937092482,
 0.34237458991466896,0.09322569901982891,0.1874298214085932,
 0.4240077286629319,0.27354656787413145,0.39719656376134815,
 0.3491860007015035,0.1601489771990584,0.07213253468202807,0.3143575681677834,
 0.46698299035437485,0.8085678196685391,0.890379237434414,0.2726850174425519,
 0.4454652320865866,0.5572703831531022,0.1445659286488047,0.6222259265548269,
 0.5516276691496402,0.5395947103855965,0.5635398037167596,0.9934117168948192,
 0.1481660475899177,0.2546343738525012,0.47236774537798465,0.0753272090470064,
 0.4355201422885955,0.9240544836062738,0.777273058619938,0.5172993735723865,
 0.12723905312189077,0.9688882629643847,0.7530093814192598,
 0.39463920283886844,0.5912915965814832,0.45721205368808093,
 0.1334622735204377,0.8403547057934676,0.5981916968680004,0.4883534412992846,
 0.355694593762409,0.5681879829028695,0.3501334809754568,0.7906996266907182,
 0.9998885141352071,0.8110721509296019,0.6124973407233161,0.872796288112731,
 0.3456983037151813,0.8688722478769844,0.5521332764668035,0.2044799168728073,
 0.7945350622223758,0.49460065364066197,0.21512312576852377,
 0.39336091605093115,0.45162799946016363,0.45169801319897407,
 0.7955574026718208,0.2910950699943812,0.6932502843347053,0.6940663380002664,
 0.4152643433657739,0.8486739008437446,0.6227255662629095,0.5013920446727382,
 0.3985810152597047,0.03531479981256247,0.04608681431377171,
 0.18684465011442442,0.4147732813686449,0.8769359566813812,0.8000871313818795,
 0.15709379175770488,0.42243721079774865,0.18506202878728362,
 0.7741403770264227,0.8010992042081423,0.14506412636733523,0.7638030437980967,
 0.05034431533078465,0.9709148287111498,0.22618116098051633,
 0.31122794422354794,0.6595213436869493,0.4156714935073659,0.9127827116798033,
 0.4291238862568273,0.09975726548823871,0.38598177408016876,
 0.8180834884614471,0.7346865280495942,0.4374508314385157,0.6297256159521929,
 0.6246122224898527,0.7727630226724918,0.09755873469691079,0.3082197769784042,
 0.7149363084800991,0.33078791124796003,0.4125547049690321,0.6915759632246505,
 0.09224055818478238,0.06647542392166661,0.8129988867701186,
 0.6067402096828424,0.7389043680998152,0.3611573632392602,0.276440114651915,
 0.6020245026592527,0.8719852056108491,0.6984103098136799,0.4834672209699087,
 0.9350514628173743,0.747054562228213,0.26248655511696206,0.9952285210632723,
 0.0179798235994284,0.46718566233039327,0.8360093038240055,0.1882767038766593,
 0.7388538201512416,0.05573541602361065,0.18786108637052035,
 0.5261925262984911,0.3286954864684376,0.5532355209572767,0.2443783274469823,
 0.14858219978813125,0.2679240755684229,0.7129142357282318,0.3754973498782872,
 0.7487593770406318,0.8161118446613792,0.9078851201990095,0.3937020803100315,
 0.7161001779796426,0.7448071920954203,0.6249126915589706,0.2942881836573551,
 0.4106788841045941,0.6140022315112766,0.5179483539167375,0.07767251666379937,
 0.8590937604075151,0.0540255978419687,0.13333589797491563,0.4620557319050801,
 0.4966603995649549,0.26503265263891196,0.11323545715445782,0.697375887312326,
 0.19531349280825594,0.04972800727355786,0.15825908394445076,
 0.02093085204509315,0.3133110290057233,0.933578083896692,0.21434287463868362,
 0.5406916172609739,0.16870705405891762,0.3799267635733484,0.4402151870457237,
 0.45610129122533005,0.9273690398852317,0.07647657855313128,
 0.9898900424104748,0.932062831860189,0.6949818343608628,0.935845166647876,
 0.7730832833316017,0.5351719338596596,0.2516736301393455,0.7101507658736066,
 0.8935976900693849,0.7849368980751659,0.08980281290832492,
 0.10794866352545385,0.7087712916623886,0.9553881689303645,0.9989286636509503,
 0.5454278176284652,0.5786916566930331,0.20688528801369088,0.5779513542998503,
 0.7240381181330289,0.7721315022609546,0.1167078215178482,0.97765971101096,
 0.9333038458363369,0.38494571195887006,0.4132367031447288,0.5275699013375311,
 0.8919305871037604,0.5046454616858,0.9457815600220054,0.07735830434738344,
 0.6615385857519563,0.4230350585348882,0.04956393606414555,0.9288970614107915,
 0.3742044679211365,0.14927291082972727,0.14615046490126704,
 0.3572443766785014,0.5256936228237284,0.9047189133287341,0.5639721441949602,
 0.07583669945608151,0.739212225942275,0.779667566380359,0.7916096818625088,
 0.6139156183484484,0.7538482292310299,0.2397937956771573,0.27602871205709034,
 0.3034270043884143,0.10073338455059844,0.5797299384112141,0.637837413356325,
 0.6491577859077808,0.8879100761382837,0.9025292016307828,0.14928885695967564,
 0.9552365635090736,0.955188973429626,0.20819284265440385,0.7376033362687625,
 0.8476140798147782,0.43492968785814334,0.10359376672621357,
 0.6247719671620585,0.35200957135136157,0.7613837891048102,
 0.33105777226893196,0.486433513091863,0.3683249805721629,0.5169159923374052,
 0.6410218424337221,0.6265763610615416,0.27711444586275547,0.4733395700229426,
 0.5479563104389776,0.3462641976576445,0.4023602621676854,0.6709027987380085,
 0.32759316643072023,0.573323251441722,0.4655660807442783,0.8410276583427587,
 0.8295333761053314,0.7740116339452916,0.42735724983879475,
 0.31642832321494496,0.0308303017255942,0.7075872848426411,0.328912946942573,
 0.23641633515332072,0.9010528782265175,0.5616006471943664,0.8750793989196871,
 0.46770160672622596,0.7075920486138269,0.6374041091930782,
 0.47487302748584903,0.5640567003359694,0.5213344574287178,0.6547989322067688,
 0.2534003248933766,0.6067570629423656,0.8391358930678874,0.20544460050569058,
 0.5573397887494678,0.3404382556278315,0.35992840837221063,
 0.11475543136425181,0.001496560227425281,0.06980250730078918,
 0.9863194743711032,0.08653008945944995,0.9428028417416178,
 0.44155811831554925,0.8507997583338958,0.019685848501065473,
 0.2903556901494956,0.6507965039790072,0.7824858331641659,0.6365463161660899,
 0.8319358663109846,0.0780885257255759,0.7889759661527803,0.12764656895833337,
 0.0801300493886552,0.13050870426715255,0.9742645527701982,0.8869935415587429,
 0.6649467037167613,0.17527401794407982,0.6586388855159521,0.4809480766915235,
 0.31030177647953205,0.194761932521463,0.8988286991343208,0.5382054700886942,
 0.7298824543118394,0.11762515826402431,0.7349619536103953,0.3264113133279515,
 0.5560477039681132,0.252697080726908,0.2166700839080249,0.6103473280069062,
 0.46529589667354043,0.025156780109282684,0.8665869062678283,
 0.687217871113909,0.07153409654688891,0.7900923882830196,0.4979044792223649,
 0.815881343598358,0.9230814642577924,0.2633519944395186,0.28905219415207917,
 0.06237319861523993,0.5730824518045642,0.6138867764812364,0.988340901156923,
 0.9637909732369542,0.9733979816376536,0.2092352783786352,0.4822291336680342,
 0.7717316101524144,0.46414812141045014,0.30032113877547484,
 0.6099578619792844,0.8470813446101422,0.11552575716588975,0.6467456435514725,
 0.32644246333988136,0.1498979165315476,0.2156704198926337,
 0.39045152284522167,0.034802265320589276,0.08893947797573165,
 0.5830280296380445,0.1325500257368949,0.5604119762916104,0.9822225610986469,
 0.13346943480313533,0.7140549828432209,0.1551621923850406,0.8516225856323141,
 0.2814061943540216,0.65920271723352,0.2334076616834485,0.7609377170186349,
 0.941791863923275,0.39350062164996213,0.5107758868507625,0.717083684513804,
 0.46102511205769203,0.4081205757853029,0.8276220043388096,
 0.25118727755785986,0.8266663764107991,0.4846550327764406,
 0.35173187363912084,0.6826420550829938,0.7033034262433296,0.4113357828473484,
 0.633577757212627,0.04385028992282414,0.8230434799997077,0.12351157917617661,
 0.8258665814421364,0.8998873520175434,0.42092878742276685,
 0.41233502621876394,0.124138762984638,0.913116418505947,0.7332171948954486,
 0.7778637118629709,0.1298334170908404,0.15607142104191563,0.6335263525994794,
 0.5005532125312182,0.2561692557880011,0.5576133479830065,
 0.006259920812183406,0.6514469303994792,0.22962879957162585,
 0.5269969792190395,0.6723897380713494,0.6166646412916496,0.31982174145781717,
 0.376864301686735,0.3337093189769633,0.9146745117009409,0.2658241483883167,
 0.20607681853597115,0.8902974352478338,0.28367605340646385,
 0.3701018732811474,0.4213791646514813,0.31389270079150333,0.9449168388808296,
 0.8242698225036789,0.031930008472419,0.31532962821869337,
 0.019295722957187178,0.14248296138088534,0.5115516062667473,
 0.619181458733609,0.6327065776003147,0.848811388048863,0.29189604097560806,
 0.605986615521378,0.8033964322793542,0.6617562592942767,0.81499421475876,
 0.6590548035431951,0.3583797590592793,0.9718052007336428,0.35507739395975646,
 0.4093866740187103,0.10021541616329355,0.5754334858266635,0.8138223268061182,
 0.07007540648261812,0.830649745485184,0.20164915657807514,0.5312705604277659,
 0.557983730114091,0.39796096820419813,0.014867683817443633,
 0.12659390429410133,0.5336047080637616,0.41482447351315443,
 0.7620821323392108,0.012963243010625902,0.230217265200642,0.9227921761195489,
 0.9041244561863456,0.623393979538823,0.3870480554695507,0.08946299007705616,
 0.3865917055839071,0.27204149878763406,0.4372658381265466,0.6095761622580518,
 0.18427843425209434,0.3017577310508537,0.6826590644341514,
 0.20322957084995763,0.23541818125823133,0.8587496077797557,
 0.5402899386455657,0.34252845141622534,0.9804672001031199,0.7223932685799115,
 0.8286557152753216,0.7693978243766342,0.39446850077584694,0.5702495446242613,
 0.9159706457654645,0.23917142265565627,0.3897263088563856,
 0.14749383459819077,0.7030798395030401,0.7681606515270826,0.6644609981698788,
 0.37927100977236405,0.46038302911879947,0.2263782944127153,
 0.7373195251926632,0.9283277895561681,0.14918510334426005,0.8388937351009196,
 0.13770225500339162,0.5216333138444129,0.9426013398746724,0.5436979909075058,
 0.800577323438212,0.56635906137122,0.44788893532429663,0.7597705384809732,
 0.18710369575706076,0.796984258487922,0.799963698320647,0.6592046092297396,
 0.17114485789465783,0.552959818750636,0.645815187765928,0.7673017169542768,
 0.37406522949855725,0.5339182568730964,0.9461311274819018,0.5364062859061198,
 0.932555063667346,0.004409547177219975,0.2862650823701537,0.7389388360440114,
 0.4081717316906892,0.9508118556169122,0.35839268487647946,
 0.03945950632217665,0.6207636290821107,0.4636239759706887,0.5756703829810323,
 0.728252745242612,0.8912863318187654,0.5654477381985757,0.2583944988818637,
 0.11240613271895561,0.44008488299037396,0.5882384057848443,
 0.7126188038020878,0.3549271406310074,0.40989464416453103,
 0.006523123613917381,0.6557925099514506,0.9289950494374599,0.646941277258087,
 0.5472089918791876,0.49476519873825,0.9780471198969954,0.14264602829931128,
 0.11709773346554264,0.05636652921557728,0.8408187144571099,
 0.8780304365678947,0.8147565664943893,0.755796393833393,0.3369889707449988,
 0.6167173027257075,0.5797614378600366,0.45249345857869816,0.9858570187256047,
 0.9220109431868507,0.4582169411229534,0.005372874533766225,
 0.13861297393292438,0.23014185167815426,0.42189375778822136,
 0.4543066739758628,0.19299802966519575,0.003778846172158068,
 0.9059604554999343,0.38826531515845975,0.8798574770732793,0.7179410405291409,
 0.7512401356709046,0.9616802646237697,0.12582775162467297,
 0.22375902952354554,0.17737975527949068,0.38730931451288175,
 0.9746642099639378,0.73501038523927,0.00444114290964015,0.6996312100194376,
 0.9414823621846329,0.9388315779428992,0.232788036812988,0.7589854013863471,
 0.4406683226666168,0.9143197318450018,0.3334277475742423,0.08822283012856713,
 0.8106354012271296,0.49174668243171515,0.030454166771413216,
 0.7821640135506934,0.23460361328184676,0.6611753145346849,0.4345293076424579,
 0.8205473035678079,0.5934408709169733,0.19246016821026546,0.6991820537303486,
 0.7070283548762558,0.8260818021443038,0.35927524368266295,0.981001336678307,
 0.03671480000629046,0.2856238818614628,0.9856249044390907,0.5265223483312629,
 0.2803276271951044,0.013162820111705775,0.743539346560907,0.9979966625026655,
 0.32688950439100495,0.5385944861419478,0.2778691375597906,0.9901989816643105,
 0.8218855503609107,0.7202697826763931,0.605733030512027,0.8982378182312424,
 0.11815271128860949,0.09041060244244481,0.653003007479052,0.6039322100964206,
 0.29386578255674856,0.8976618858609919,0.3743349774572058,0.836714423819442,
 0.21571610388432938,0.014546164586850319,0.6457540689384546,
 0.2228698133647945,0.5203540653117715,0.4316158482238657,
 0.039540999750226846,0.6308757049894272,0.9374935701837352,
 0.49093272546657785,0.8730089347194826,0.8723956171368586,0.8629087093834752,
 0.7620607700990774,0.3818633421049181,0.9297293145315184,0.3763660162604253,
 0.8420752617860848,0.2976822891558677,0.7344554208474456,0.47943570114603595,
 0.04746268697354883,0.5464248368080455,0.3521640910146079,0.6736011539912127,
 0.004043536562838179,0.33918223808254033,0.7176229341847995,
 0.31479692292577144,0.7154744533652259,0.7990537765901128,0.9116249976306116,
 0.8216988769319021,0.36103922219476736,0.8859366434770294,0.5668964948791029,
 0.2495054733842723,0.21189169569837363,0.788791672466778,0.4139314911975579,
 0.3131400462832028,0.7599850636772971,0.7700004201943271,0.8441381738812046,
 0.6709720609542096,0.8662515240825412,0.4908548074862631,0.9745008260823894,
 0.5323836113022803,0.8003986088561872,0.9945575020073804,0.24596453607058988,
 0.8068879095590851,0.7285458436310506,0.778773755001608,0.7438829116754131,
 0.7502340442144988,0.31223514358403304,0.3615115487258085,0.976598169927585,
 0.18292477518003947,0.4023569079793403,0.16810543226791408,
 0.20328501147982003,0.42360777028500973,0.7145664514295877,
 0.39689553204549666,0.805142960819754,0.4606374431489636,0.7656955915109185,
 0.5634789179109991,0.23276600331388442,0.19213221883260456,
 0.6041598196890604,0.9711708090523226,0.4802923026485221,0.6409413846164667,
 0.1362024906517153,0.38386525404440075,0.06662670408372706,0.762344546050031,
 0.5595470112349876,0.8333968228077537,0.03643063288540316,0.2773574194962195,
 0.8418010773792566,0.6443149708387977,0.800165130004405,0.04935063026257369,
 0.7474436212292221,0.6587663485176967,0.4297270239614651,0.4852592838131907,
 0.9474381079040467,0.6435524568513002,0.5953115077636314,0.5335699080084283,
 0.33229466081364156,0.6905586128043116,0.7667793486905636,0.387918158893072,
 0.14364587533238038,0.7392876266767835,0.967762014602083,0.05537580626531147,
 0.08169137800713733,0.9534311498057688,0.28362238362376535,
 0.14073494966326916,0.8143257414227156,0.6525206237067804,
 0.31534115206088453,0.2916600868922159,0.3622478358169974,0.8184861913767378,
 0.7861482748278393,0.5392886253872589,0.4484298272253271,0.8367520380406518,
 0.3341624786111901,0.09261885513528823,0.9515788643251688,0.8052524942054845,
 0.35513236797904535,0.49887758671827576,0.38494230272235586,
 0.4938309008688325,0.7413061067534783,0.9781127844489974,0.19570376122976973,
 0.27186956885249414,0.6076803538977371,0.46276308643589215,
 0.03693757971348477,0.543592317729387,0.19786109624563974,0.440712312943195,
 0.019056839699229777,0.890727139827339,0.7119495039994675,0.5228307316585625,
 0.9194059981852544,0.6492945928723984,0.8750604938641928,0.8556194483572379,
 0.018847417997196803,0.16463314662831108,0.8958809790231068,0.40920629995838,
 0.06438914820473407,0.8358391207385548,0.5564767527957817,
 0.46395566010143685,0.836648141337316,0.7907885337546413,0.3203314119895393,
 0.17983803332248516,0.12413236860663224,0.5170978490553526,
 0.3783812048719337,0.8220341481749441,0.8255759166546861,0.09615249221325983,
 0.8460915204249032,0.7653281967871446,0.5267449578648105,0.6747772478370048,
 0.06918898462745515,0.5323437999360326,0.07622837177109776,
 0.15286407333320085,0.9227588133348128,0.17122292521811433,0.675246181619959,
 0.3543937307374183,0.7623346768188455,0.6418317757828524,
 0.005394810275487272,0.5266219866888235,0.08589006747781058,
 0.5815533155388424,0.20246690946420443,0.2898392983938315,0.9095113824595129,
 0.7834788654329441,0.353343858553576,0.06290448742556087,0.5681247531175213,
 0.9477437043658591,0.8006920225894668,0.4234782082678271,0.39980066674727033,
 0.697251439176668,0.3290913319176946,0.6088086497003425,0.18848615048001682,
 0.6912264226708602,0.3602697556823067,0.42866236362173904,
 0.07326242721182352,0.8010438671010611,0.42231706224868204,
 0.08670323220064091,0.19542691482104246,0.06589328724095955,
 0.4055650896403936,0.798291871068265,0.4006829188365829,0.9690006311787874,
 0.11367478668439046,0.2225968167678989,0.566620923642814,0.5020536180278157,
 0.21273816918605137,0.9577931243687169,0.18424523955032568,
 0.14048870711603234,0.06856186692249966,0.14513794262232393,
 0.07612944911385577,0.8310636697517712,0.11666965141137753,0.491285224206093,
 0.5004575106026137,0.5973196700419032,0.10955929791103003,0.5478266671720948,
 0.4822616619849046,0.7058417666594834,0.6585150205311235,0.5394459835289979,
 0.9811993080261052,0.37819269482139495,0.11848382249332368,
 0.5710083013249687,0.5903681064951768,0.888967469003999,0.27290395112202637,
 0.6579350654679821,0.6969285006334125,0.12301957982452993,0.9612965571740719,
 0.9974451136870713,0.03674438550594783,0.8176835621009574,0.8429988459945511,
 0.46226486113712395,0.7084835338529403,0.621981645594248,0.2001707767308466,
 0.33103627382909895,0.09170880639157841,0.9855978605264687,
 0.13721625393006776,0.1338622750498396,0.2676201590444012,0.3772855227151737,
 0.5773533550387675,0.6228135647085427,0.04906917224472085,
 0.037971467375901624,0.05238597477839824,0.634676653242758,
 0.2682858173444128,0.6165666534159052,0.6706784122215393,0.935319903431217,
 0.39444987285652267,0.9058429438472355,0.079785638053817,0.4440818971561833,
 0.09719437039112344,0.9509349396655415,0.8776845831218771,0.4690397509693174,
 0.815734873274105,0.3207011080823552,0.27826653818725955,0.352838099309709,
 0.8482455344596163,0.18146864647946293,0.07472969074899649,
 0.22606126502141355,0.2678481489030369,0.988252674715679,
 0.027029607654521448,0.08722229664547643,0.3813538993839426,
 0.7699900670211494,0.2140343067861873,0.4676625141599636,0.8729319895596973,
 0.9343724456782172,0.29459911327858046,0.9545153780319924,
 0.028822303633060042,0.6583468564460541,0.8630479099338257,
 0.022699462376178148,0.6051270336294344,0.43833110408244935,
 0.9545184981911177,0.18294621289102475,0.03179019955672857,
 0.43890974957066586,0.4250341143128862,0.8808374535227914,0.47296309982925,
 0.7433115235331833,0.9912639361752696,0.7055896961609668,0.3252255378903739,
 0.30464249730550175,0.1107676528230801,0.5342147388098887,0.1996930516305513,
 0.1960178694510578,0.8019237883442114,0.6979722965772127,0.2953363783183117,
 0.21321987331839665,0.05151318654570025,0.5021685823667518,
 0.3468153091674935,0.38942964773104227,0.6353177644374196,
 0.05088233907085937,0.08484234574890792,0.8210222377837987,
 0.2826312826501195,0.04037748772176375,0.028919183804014037,
 0.07480088785786698,0.5592600244022093,0.6640043229646237,0.2679893983300099,
 0.1482989460254014,0.6388510527132066,0.33248652765509124,0.6248426212357236,
 0.5011326348704667,0.020284660304231927,0.024526324947601763,
 0.8199773973696428,0.5940417407750977,0.4238240028417688,0.5710453031481075,
 0.17232607591078053,0.15944292549578365,0.7055008715787133,
 0.40935201508770824,0.3371881478710945,0.18432988128194838,
 0.7637672676108567,0.611392234958713,0.08885689958383969,0.7173588899580025,
 0.3659158280469973,0.28296824180377167,0.6757935031395585,0.3924663629099525,
 0.2983576161877869,0.9635264587144898,0.7378652493356602,0.23638102460340527,
 0.8820210152510015,0.486070355279288,0.6991275952050588,0.24674713883715815,
 0.3320659684378088,0.6368314562441652,0.2673091461475543,0.3401899102990402,
 0.5122958950399773,0.7268812570564064,0.8534493680501839,
 0.021206899524471967,0.6583068982081778,0.9359911819909997,
 0.6330035474847918,0.9744226361012758,0.34726080350403055,0.8513584554020143,
 0.8543754308805342,0.5063558005643257,0.4572189540888466,0.46426874891818404,
 0.27253186553439046,0.26913077419990017,0.6562217436047744,
 0.06508225023206049,0.9630696377995043,0.2035777112909234,0.9684308520412985,
 0.030959459627612707,0.598427081514452,0.8635132300053523,0.4138515100274591,
 0.9630031203850133,0.4603853643409619,0.5122719179015368,0.12220568974544777,
 0.6582883964451339,0.05546642555649317,0.1660131115060809,0.4246884608088184,
 0.41060808237757673,0.32886243764419865,0.18703963537023593,
 0.46323731506711174,0.0352963656005435,0.601043003586411,0.4952817065392814,
 0.9032942690431608,0.846365199891067,0.9417718360550839,0.09779303512817372,
 0.7918921705528903,0.3911767710513475,0.10018898445411839,0.6162159429842824,
 0.9262858800199089,0.9617766543020472,0.7316601577885481,0.6454519864329721,
 0.2819164920177798,0.6663446755848547,0.384941896737914,0.9507193486880534,
 0.9218892134652691,0.7536980575320278,0.9065388276935701,0.622319204132032,
 0.45340335615941196,0.4432326368290662,0.6878316738289646,0.5842668402355411,
 0.08387447280777316,0.6927757137459928,0.6975330070911946,0.3751404665097291,
 0.914054416520896,0.09773930914270179,0.777583947610582,0.9699773340728113,
 0.21625728020341473,0.15627341198470157,0.9688783031967528,
 0.6857006102247044,0.5664287795311652,0.6965157392758554,0.6037624210309553,
 0.7039246988218976,0.3355356542985233,0.24518677882898798,0.2805817275905229,
 0.3965495872845356,0.17004891908412412,0.8517114861709935,0.215957094863775,
 0.7387526943672386,0.1433278616787238,0.6139420304457779,0.5997759957502039,
 0.28808374369650125,0.4372301585647086,0.38757097534856677,
 0.11309010974386147,0.29715308121892015,0.6303982435393459,
 0.8537225666148148,0.9507825600381286,0.9915887104210442,0.3577902740803407,
 0.7052556212095439,0.2540578099034625,0.04945448943598263,0.6058366054713124,
 0.10218828096754373,0.2505272020013988,0.19178040475515368,
 0.5811490587392226,0.4176297181676907,0.7584133582074376,0.33276516030614467,
 0.616649392613785,0.8010492133176893,0.006811882510436629,0.7534377156145124,
 0.9245097071302395,0.43084139952772893,0.6179240893003635,0.6971679620057167,
 0.16970078787384757,0.8415361027240758,0.41666819996142956,
 0.12968265723876682,0.09248049280183623,0.3723031212502086,
 0.036276685781417295,0.8187889111329524,0.41325752545676453,
 0.08901301144335827,0.08290296477424108,0.3723639445140938,
 0.7311217528198851,0.2960268503629866,0.6631582177407462,0.9542874034037924,
 0.8024567703015113,0.38655714783030204,0.7264800235467381,0.8775255667512181,
 0.7586048949609218,0.7302040267409936,0.22543301249902248,0.3925103872641973,
 0.8981819258337259,0.7497308368143838,0.00350724337214392,0.3900899458732433,
 0.671803781281832,0.6857581006171118,0.26729706001774944,0.5581934789811817,
 0.7551618232926438,0.6449567822385451,0.9135126974399094,0.23934689466815195,
 0.0018999384565537847,0.8395427619697907,0.9487305378755919,
 0.5297260222265475,0.11048678925588928,0.358824546812642,0.18065389523768227,
 0.293629963614361,0.47820352766134255,0.4323593465542006,0.11671915792987642,
 0.5332907671346682,0.6340118381638999,0.7523503929193686,0.9551657737749215,
 0.31856402690093133,0.31053482767398044,0.0910720658095483,
 0.2936880217089233,0.5574582868974178,0.18976756626743685,
 0.27443749176625076,0.8923722354972304,0.6839708707873128,
 0.41772306883699073,0.7977208675138112,0.9569924201226909,
 0.17448628613577233,0.8238546227843863,0.06912044901519543,
 0.09170996500629769,0.6510219870827471,0.21953631332415324,
 0.21057984489521175,0.342565390267501,0.5714802187545137,0.1830736707524694,
 0.9388294317113433,0.9981252950674688,0.48513095320529986,0.7391923438729241,
 0.9436901299340998,0.7356123208575801,0.2803701261109416,0.32386191992010394,
 0.2490243664059424,0.46734071636184626,0.9103323657198937,0.6340832297734539,
 0.8117212441951808,0.5597810656940692,0.26077825818191414,0.7007102727144174,
 0.09346461473350665,0.5325424667774199,0.5082891567343502,
 0.37771254139943844,0.5736325384385487,0.45575797781093286,
 0.7010919656652677,0.6534233829951521,0.202474426418914,0.7291728585364564,
 0.45296439291590085,0.17124761485719942,0.9532865999777659,
 0.0776855264134928,0.5519206725832205,0.28854002736878326,
 0.29491078793368375,0.39988811357153486,0.38328181130909456,
 0.8835335600275558,0.8069722611178681,0.49057418958124543,0.6174893609180859,
 0.3291610215244416,0.12856354242906165,0.4690789254283917,0.6461041985320282,
 0.6118936400456592,0.6505835108069605,0.895817513670555,0.20514131460934126,
 0.31407017829411565,0.36864461780373103,0.5470510865602746,
 0.09601238045211513,0.2580139060908564,0.23958180548842778,
 0.6590023343317788,0.996797236406822,0.956581016625321,0.15573803060895197,
 0.4723818421050857,0.4364942218299057,0.8532779733153146,0.590794038224743,
 0.4150105127317576,0.07949891451287361,0.9990030988383318,
 0.16495980194654591,0.3305167964608038,0.5614007836888648,0.9033455958186764,
 0.19252844240475175,0.6429196615235083,0.7835530921039079,0.3225201489475793,
 0.5871336086129263,0.4870674937553885,0.8221327265092904,0.22429151644558054,
 0.9881841854449221,0.5573526568286484,0.6358987608365372,0.9105709903371257,
 0.2806398439080846,0.9304542819001912,0.9743456975812373,0.632847446623896,
 0.28855714934277843,0.35235132027381266,0.28506528519248997,
 0.3063582410484883,0.7375707477248064,0.2532913941311412,0.441041034996415,
 0.09222884520237629,0.9069951503536302,0.1803806970261429,0.5686683550801981,
 0.9476593405420232,0.03560662502178402,0.17048051049116852,
 0.7016623685182629,0.7225480239009675,0.8557225904374016,0.8989917631105784,
 0.35495368745817646,0.49000635978435403,0.7926696474650481,
 0.23977558183553493,0.2958380685156544,0.8501284246315205,0.601575413356237,
 0.0869032783782655,0.41613999105461585,0.12615630478441997,
 0.5127411855804849,0.32501958553526855,0.7771285927032155,
 0.23747775385571845,0.8701338394724939,0.7433431093487457,0.7487646007576524,
 0.7326877695604826,0.0024487010061795544,0.2001938191829068,
 0.6734687623584229,0.3373000325067622,0.32254068771167355,0.1736524030166724,
 0.2731957210070701,0.9542329826413446,0.8856639715048605,0.15987129752261975,
 0.3089901419206683,0.9934513819195873,0.5911704368541512,0.940103309205072,
 0.6589207326353113,0.5277575418827978,0.9554485662578324,0.19432834240031016,
 0.668509339861218,0.7161496937973127,0.8797898038439673,0.7546594961359511,
 0.6387487081344435,0.12496175708679957,0.731533283285185,0.12313341633122288,
 0.16122379477533222,0.0572607906892274,0.0880218151864085,
 0.07575800615808048,0.981857406715083,0.9979772906124306,0.053041788352714,
 0.050405309958210465,0.4573234778650218,0.04422167113937914,
 0.06029402185230959,0.26677739698758984,0.7813069623175433,
 0.3371152656040878,0.5443811574947636,0.13468660057309378,0.3998222771261942,
 0.29169527091947844,0.6449386123518279,0.499267577849429,0.890769540382562,
 0.8107986260046297,0.33609359903076375,0.911249469202373,0.7251221596991072,
 0.5040965260469907,0.5959078228387472,0.621950121530419,0.2274576477204423,
 0.5497058012427611,0.7904777564699151,0.7754261580964095,0.5120019010949043,
 0.4265727009813155,0.13752472431299878,0.7157024170875523,0.3772407369256965,
 0.3170424036764282,0.7629431665759032,0.3838341037932633,0.7233341401945469,
 0.39859230525641287,0.21648083113254635,0.9535066462884596,
 0.2197458600376494,0.9161671035635361,0.8460949800424384,0.5355760057900438,
 0.8827308058337547,0.840348522728857,0.2611716058971515,0.762411227456011,
 0.17430248054252628,0.6244118348023889,0.13577786476069797,
 0.2986511111539004,0.3231037873623419,0.943248351018916,0.9192115167372577,
 0.8097156504671713,0.6269293950898239,0.7572097452931241,0.9100171187635591,
 0.5345622172913185,0.3184784249447443,0.7677740950781221,0.4881860049210137,
 0.6218540697804322,0.8257983475143378,0.0821957639462767,0.14242847070330844,
 0.9128305911114447,0.4462268631384886,0.009815801446607209,
 0.46350618938683186,0.47074636119567836,0.8733339461721537,
 0.5391610562756574,0.6909920874376566,0.6828132793274578,0.7164458380896956,
 0.8636180460718821,0.5599057788801506,0.41230354027159044,0.9278169999505487,
 0.6005274619444149,0.5425144673749147,0.5915155212274917,0.21543429900562006,
 0.39037066725231484,0.276063571224634,0.11837007787626597,0.4086195526852294,
 0.6028324501316145,0.636793851334334,0.05916817249056783,0.8804518984139978,
 0.6173194522259061,0.2765752418077497,0.2084909077805508,0.10220338554160102,
 0.9394469724410577,0.8728381823568605,0.1673410984618785,0.761831608800861,
 0.5846151118152018,0.8465840068021285,0.7132603102572193,0.9990497123340611,
 0.8581965332000272,0.5022537994332762,0.9742605578993169,0.7790967468844394,
 0.2602719340335804,0.8212683085907184,0.04884239463634499,0.4396029481696835,
 0.713097832234537,0.8743465395605667,0.501330433748521,0.8135087126798074,
 0.14380940221756655,0.3793275319783591,0.5597896657342852,0.7504350987685066,
 0.053768605228702304,0.410473673718995,0.9342321476935468,
 0.48582439715218684,0.00807224329003653,0.5053038766621709,
 0.34506309825132897,0.1657092247259918,0.9354426061984092,0.2813566849263516,
 0.5250642225959897,0.5446880780454824,0.28057342414786235,
 0.13127153183191365,0.016278185166974435,0.4132862831723134,
 0.9224111985351736,0.3556516517133954,0.14421996490087485,0.5350724890398668,
 0.6758231058532315,0.8543094154920158,0.33691469799005014,0.3836013569757817,
 0.19695909585833138,0.9454794151509247,0.48424213187638787,0.867240176755268,
 0.43884504808789804,0.026999254399526218,0.5124456635177269,
 0.595203204699332,0.5841019906231284,0.8116203488974705,0.50210227902996,
 0.6695471769103601,0.7046261170576874,0.22474734878833857,0.4520052023234632,
 0.8330185516150207,0.5883724615799911,0.06324829428245238,
 0.16971465191961965,0.8523939913332558,0.7084004939214368,0.7441050200638288,
 0.7022885250548443,0.3625470075834891,0.5694001792671707,0.5392791043474493,
 0.6343679100046249,0.37668177700112127,0.22943508162747062,
 0.8259082613335181,0.5502960202418847,0.2569719954367953,0.8201042166632043,
 0.6826822915748731,0.8785750090410068,0.3085777480051336,0.34065569213710745,
 0.4331330054343314,0.9886805586946505,0.3318625019040047,0.9169616741163962,
 0.07895609469450793,0.791145541467666,0.7625666314229052,0.885314389902615,
 0.18730854877769476,0.9855283040003305,0.19663726178294172,
 0.9324054806497014,0.40524835567862516,0.9422324622776074,0.1764955228056967,
 0.6181363320362795,0.9341586759877099,0.7309545217342882,
 0.041495060205764545,0.7299506010187263,0.15533960852280815,
 0.4124834822970317,0.17170494945094228,0.3629835632128917,0.7505449343446533,
 0.3871537209929281,0.4253738640594992,0.23250554761856523,0.3491445070543234,
 0.7318667330551918,0.6398350414669388,0.7262485865587,0.4358991036399005,
 0.8392605694558583,0.7257513490255717,0.9368517794811341,0.06650437552998834,
 0.6185726427426019,0.07397691900819547,0.08191012852119739,
 0.2812444172135806,0.7390894721632363,0.1648627013578724,0.44726845883612576,
 0.44979610045389684,0.49286517233228166,0.7179205265427032,
 0.30788337456716486,0.6798729988499971,0.32486567078303996,
 0.052101762450278866,0.27559793273047994,0.6161788791455388,
 0.9800582456165449,0.6201887120959804,0.25352562388923294,0.5789685693591428,
 0.21205062947003772,0.9797288766544142,0.3473243063256529,
 0.38573511336258326,0.09024003543655867,0.553329073068786,
 0.44929926318492597,0.15373388821418688,0.02852472029548847,
 0.9804254218541177,0.7644126062842308,0.7417408592133504,0.9780031493762985,
 0.7836722912929853,0.41845937323060167,5.739310344199389e-5,
 0.06541396270687216,0.8387500959056972,0.5117838085078875,0.4059086182984184,
 0.7284166033515885,0.31540294695998905,0.5348585265955026,0.6750760712598276,
 0.9496344970626464,0.13329450735835546,0.26562844530630747,
 0.5802325310053552,0.6620267635465148,0.36293857630354764,0.3377638702780634,
 0.969027244713514,0.8677586570270842,0.5596305290287764,0.48906270697785914,
 0.7505597137595404,0.9972097128607369,0.6443026250001185,0.9023606808019643,
 0.02860957545161691,0.7826527052607408,0.32650719442190546,
 0.21859764872937093,0.18311453410243117,0.475946093613028,
 0.024353057093363772,0.3872076117495826,0.46226849906868284,
 0.8367325455932495,0.7528817001412119,0.03074310016034787,0.5794466458497106,
 0.491128253434285,0.340357428281305,0.05686828806583388,0.8351659891988724,
 0.6233953015302882,0.7262213643394164,0.5407971324841361,0.3406298112163365,
 0.5717535271290981,0.03251728195557857,0.5213955744007955,0.7290912632363092,
 0.5985286518822105,0.13370727033225016,0.5602169118704761,
 0.31026634401001296,0.5163680986210851,0.9917262167983695,0.8961721942682681,
 0.4680032582238839,0.46344590125846286,0.8418095155240545,0.9766846367273172,
 0.16122727557682104,0.34808869987716073,0.26929850377952935,
 0.5656346655389604,0.9344442707991544,0.3137163981200317,0.9414081407301607,
 0.9024793311127102,0.19370171481881748,0.6741253771445085,
 0.29416975268547696,0.8879804204450173,0.24735859041922037,
 0.5735525764297461,0.5033472293766742,0.8138041009270367,0.136607059234503,
 0.3061759680828201,0.2740218495963047,0.7340262833624,0.07085695947107806,
 0.8083670845731107,0.9567512111920924,0.42186879445514713,0.8773470437734299,
 0.05273381332871818,0.7307421916006183,0.34296198919422305,
 0.40909129085124407,0.2981479408684017,0.6160546512954439,
 0.21560772603401757,0.6911673665076393,0.5790664833058932,
 0.40465896491049724,0.5557440289411363,0.3244015850933484,
 0.27992134680395675,0.7929283346227576,0.31724843391258617,
 0.0745950393441549,0.2864221116355714,0.7970887639977787,0.03485324714840832,
 0.8695076863025497,0.9022340346363296,0.6082871324813973,0.1086573038710531,
 0.2289424675159708,0.9344603706049329,0.5133555672142446,0.6410655505124083,
 0.41942873435348693,0.7499211427818525,0.20383900440648828,
 0.7260251071941813,0.5128341408564515,0.7935976039050614,0.5887947473057581,
 0.5390191104127345,0.7303668180967897,0.03889321097632936,0.3793050712071818,
 0.6008790454559809,0.646850070934462,0.34390437319758904,0.5054541887530489,
 0.49466174219401304,0.0731415588044455,0.4330171316613416,0.9153466307139102,
 0.5952935518882814,0.3226763019579021,0.014523529444183558,
 0.9776336797870824,0.9834395477366127,0.9083758973793996,0.06638600449574494,
 0.5550334266868778,0.44040256814463274,0.9589780052108202,0.7077061943110086,
 0.16207305576193853,0.9293486344873795,0.1775180599318693,0.5066372202079121,
 0.8687521845017605,0.3036099445604211,0.3564963349951594,0.2147259298377856,
 0.06909260924130012,0.606427627762704,0.0992223758803853,0.2262954137175801,
 0.9264743442861493,0.2201763639109262,0.4534610367167681,0.40480692095258053,
 0.25450726809053803,0.6441496764922598,0.07971143374845135,
 0.0018461216374167089,0.9480908575630577,0.8868524782691517,
 0.5773258577114325,0.694804241688304,0.4699964968124972,0.8393234383196301,
 0.22376169986762295,0.09741585018065901,0.8854628462146149,
 0.2023703994960082,0.11597596890341844,0.13820558926206716,
 0.1744972777804551,0.6579764131825901,0.1658202046261804,0.2160953898133673,
 0.9291656119058772,0.6747909979706368,0.4350716133505478,0.2603170901304045,
 0.9301205122885305,0.3700985890455235,0.012831115531647286,
 0.5360234142807837,0.48171643039225154,0.9898168849886547,0.684129312912408,
 0.37647565673020655,0.6531139271158874,0.22247521058791264,
 0.14682525965785498,0.8353832086564542,0.9365044095070415,
 0.03853892001201875,0.10163827669361047,0.6952454704610521,
 0.19033837436452739,0.7002118362952263,0.18587410299602702,0.284201399210392,
 0.6472124348455156,0.8871989971856271,0.18086021762090398,0.5518921492674833,
 0.32357094831339506,0.1835071715839659,0.6394398606891001,0.8977777758945205,
 0.6020927376393468,0.12137190751246107,0.3731075679468352,0.8139221423255867,
 0.3024467540596619,0.8419046159727318,0.03714067745609839,0.5445266894597213,
 0.8037177441840452,0.01828081857569941,0.652953839945244,0.7010767721897174,
 0.4411013161872198,0.1319725609572887,0.909021749332998,0.7728691100868733,
 0.16988123939028954,0.18573693085807164,0.2294328315069325,
 0.034950593733008684,0.8631989008208899,0.755618010176677,0.4218633701082539,
 0.36516037918522415,0.14340432107714052,0.9845336774589329,
 0.5222958278120728,0.6006170422226645,0.710366710409613,0.861494379688523,
 0.16186444152174628,0.25412841567997957,0.761376589399068,
 0.45502217519934296,0.5228719023100524,0.7860355582640233,0.2932838690680093,
 0.25893095775121566,0.9550562993379912,0.23500482258461153,0.991656173179411,
 0.5914953117479309,0.7867650916583844,0.2402211479827433,0.9768039500438517,
 0.5034763803614228,0.9606376490066502,0.8659858080918817,0.6623141185444671,
 0.7309777466908646,0.565758163989653,0.5563604435591895,0.4157836611861848,
 0.7583286284820637,0.5118209626643784,0.48791737185891504,
 0.11497449260685577,0.8985257549151882,0.10688466958034104,
 0.7584832339691142,0.61049184590144,0.10836667424795854,0.8549969168231495,
 0.821881673375522,0.37401905007614167,0.9848228445887781,0.49118301167900025,
 0.23653061822871735,0.6177398732669069,0.2149949905381624,
 0.05668780646694083,0.1963682969809054,0.18677361781543045,
 0.34449592418574393,0.32146863075825527,0.3094303365335678,
 0.14745551494642517,0.3327453940160312,0.8616935667963458,
 0.07166017056706986,0.6232086690276551,0.7217375012274813,0.7484249270978582,
 0.14410074242004123,0.8288130044063898,0.8434357889515676,
 0.13014777931960708,0.24418229945876713,0.45356726235409106,
 0.40680951344301997,0.22998902432218016,0.9436785431370746,
 0.1770422524122891,0.360474154790437,0.8391389372537839,0.8784788529611858,
 0.022573162765860966,0.9832815388065034,0.3824536942911223,
 0.1375605101188172,0.11223720625744105,0.6913345409626973,0.8936199666990293,
 0.9410349318900935,0.557670704007664,0.027582233767992737,0.2683912517889734,
 0.7532951563132625,0.9165883095921197,0.2496165892384964,0.11994983624482014,
 0.5802480470414635,0.1191781475176078,0.7436880975411615,
 0.016418267222644833,0.3356923205567801,0.6206288410382855,
 0.42768697216234675,0.7573174126057467,0.5967687548339475,
 0.19813841875602323,0.5946565622173388,0.3855438429030531,
 0.49537034034282446,0.969793663856662,0.33155911059812326,0.6608286036751027,
 0.2967880589841402,0.01975228368005444,0.35513090988326457,
 0.3620450634976785,0.9901912222764676,0.4006435827553071,0.37717907775388704,
 0.033150380200265506,0.17089620134418693,0.10977470301096925,
 0.7096826251165057,0.9442308804461754,0.260939486748863,0.2997287057086764,
 0.08298316476116951,0.1277356699706611,0.1575596986459602,
 0.35411918894794203,0.8193475729706958,0.581978143254379,0.9994096294676631,
 0.3886116540909932,0.16386864109021526,0.18557699885868062,
 0.5990174605281053,0.32500747867130186,0.29926641107666296,
 0.8655840502523116,0.3536658831427073,0.4319630541692504,0.09136316373772224,
 0.7456594586640898,0.004546880856521329,0.3341086387694807,
 0.8349316651600671,0.17324987210936804,0.3194390420865819,0.7692349983185534,
 0.550102590457881,0.16315950719123218,0.5362861373411001,0.25347538296455485,
 0.6253018558614469,0.293252654091015,0.5399026745172458,0.9319665818789445,
 0.5843441189379911,0.25325074201552245,0.8970746302119676,0.5557325110827289,
 0.34228133590465404,0.6807715923806696,0.8631046601788186,0.6781379418493996,
 0.6715563186927925,0.14251517090359578,0.8269518649840533,0.439926160058678,
 0.06443021043842467,0.5605092414907005,0.20849150079796275,
 0.2572930196942128,0.7769520566929385,0.8638699216372231,0.9820810028999691,
 0.7973182925564931,0.15689530610804592,0.744602956023657,0.18400798014009068,
 0.8310399844496748,0.6535507675381796,0.4235677766837771,0.4913011689789727,
 0.9035425604142941,0.23384647490447596,0.3615291325665657,0.7737981845409391,
 0.22315832410235437,0.9044145034465723,0.24803308460810114,
 0.9096046668718683,0.42640553893556155,0.30588257363356053,
 0.8701337120993361,0.8444574490602061,0.12303979305671087,
 0.08514969092599656,0.8330667191047773,0.5337998414430373,0.8095094337164379,
 0.8997395879674371,0.9735837459909027,0.6226914297560424,0.3478671241765502,
 0.20679674331123243,0.6693321669491817,0.8589452444133523,0.0179907770199339,
 0.2799322276569006,0.21263519009961418,0.23140349268460592,
 0.8684432224060724,0.23132176403937188,0.6457892537507922,0.9923500568088416,
 0.869285656574113,0.9910868835098867,0.9302578231282636,0.36462254109804415,
 0.060006384531711365,0.7543410525674377,0.1101585537584635,
 0.7202652465328939,0.5356850939439397,0.2630886912495376,0.3436995160338182,
 0.18231881799156513,0.29905363088971393,0.4178687823909307,
 0.2287905788004001,0.15883286434579313,0.52239093330968,0.9290390827722209,
 0.24444702443895672,0.8907100609827289,0.5138909448840423,
 0.23013432517636834,0.9545865991341189,0.3238990070466995,
 0.28291866730612036,0.8501957514663788,0.8212696588153195,
 0.49716645109401203,0.4555445910189043,0.6842972392841734,
 0.045886116992405745,0.8767416376067824,0.9095597184830697,
 0.9593909672333467,0.4625006992807835,0.9911697401176379,0.6798802245690323,
 0.9751739242023716,0.37864544508259934,0.8951194128098623,0.7632903212334997,
 0.44915850603853524,0.5337444951410075,0.46427445150164204,
 0.9590935335426625,0.4618631592727197,0.5567248364642112,0.7357064605272916,
 0.3538631128927824,0.8367680666387911,0.01710097297203661,
 0.03124602541795085,0.7224835786484411,0.9655045363334562,0.6447365442283832,
 0.27132233151748086,0.1878331290298052,0.9787396616450968,0.5235658417944635,
 0.4971062868829842,0.16840574819345255,0.8467372240745742,
 0.23412262127184513,0.9176834673007253,0.8702638190962595,0.502806613099372,
 0.8483392352684248,0.7622304563613062,0.2816952227306989,
 0.045153348330271514,0.8637901508648234,0.5233393283769351,
 0.2624000729245246,0.12148641028402996,0.7965525012239242,
 0.02043256516512515,0.7241754751112788,0.9477993893840432,0.8371850791017945,
 0.8171076911632478,0.6340647427611857,0.13278859235184803,0.9191090957172873,
 0.5727374357847548,0.9962215817888341,0.03527583420923419,
 0.17878259282704168,0.021361412263954382,0.9258147235993512,
 0.3437765088786229,0.09717601750041105,0.3528985522335102,0.4397218292739218,
 0.182671042246358,0.007552941473860653,0.09996081586324457,
 0.9842385314389678,0.6492806235639554,0.07632951075827932,
 0.37541525460773684,0.40729416569343124,0.910962632745002,0.9697532726405261,
 0.8576924775760102,0.9111116803024318,0.5048787299743104,0.8368078907346422,
 0.45023937874940556,0.5517228104436218,0.6974967341807221,0.522840890293529,
 0.6164385561245675,0.47288266456524064,0.44620141753636244,
 0.46549083153039794,0.7395967692339089,0.9158741551433316,0.5115811643583279,
 0.6009489455636661,0.6350974515789629,0.7922975271680889,0.1931814709652696,
 0.3660928275358103,0.20119110724163236,0.3653167068688278,0.9208136232548831,
 0.4580415958700682,0.6125058917427909,0.5036671917146645,0.9795250476778254,
 0.35091947193041007,0.42324409546339825,0.09504591047181832,
 0.6310825198981675,0.2495667411270236,0.19710023748483663,
 0.19041349844779476,0.43105927848309444,0.8624596817922353,
 0.33926972214273543,0.2911280738219717,0.9658084187684258,
 0.42496176192701185,0.8442683557806892,0.23729414875816812,
 0.9416555574492387,0.22531943936151855,0.6995487569921075,0.5591621486151469,
 0.3394819280564969,0.050478707077204366,0.356260471636752,0.3933038867948975,
 0.3989385548402363,0.8166667950055131,0.652565004573617,0.3079099134225227,
 0.24771911060366836,0.26850204601836825,0.8540722746963186,
 0.5387577378685525,0.9014941338588398,0.37791187619340283,0.258587517809693,
 0.7192293695273866,0.32362033822545844,0.8995779446458332,0.6116122780594602,
 0.6665252735379079,0.30048417134140926,0.522030850571661,0.657872974287745,
 0.7507330541351971,0.8339144198385893,0.7537439746651337,0.4834456336019559,
 0.9763344674021661,0.7274271402254145,0.9320473361269204,0.5066671357051359,
 0.9453810579647841,0.5080014908791818,0.15903487176178954,
 0.22859982524541655,0.4308264712776866,0.10777814620321746,
 0.5318180184443739,0.5597004764821346,0.05984494602733281,0.2957442025124527,
 0.7308787575387643,0.4288835585152133,0.685658273405886,0.9694062291766371,
 0.9398413914137973,0.6054062730428904,0.033225411573419295,
 0.8012702582773159,0.22912878533451764,0.3681662829047122,0.5709799447216357,
 0.3848612731024147,0.9469586360432845,0.614102505188332,0.06259954372248988,
 0.07303400087294287,0.6062593826647968,0.7074132365426304,0.6842989239501595,
 0.2873211722356025,0.6964864199320112,0.3537500428915228,0.550921295054092,
 0.5142100400574552,0.38899366137817837,0.9693647246355392,0.8335429895700259,
 0.32182391841107005,0.5025224597927964,0.9203776646869033,
 0.27461611725806356,0.8567309186181321,0.29296361721913533,
 0.8950248829321361,0.70006106371935,0.6302619661632696,0.539675886493531,
 0.27697455700555096,0.8924608257832607,0.03959201761864617,
 0.5885731263149008,0.23138481207954542,0.9405622507561826,0.9777719429950298,
 0.08446008646352932,0.9989338861613823,0.03597783181881309,
 0.9447870656056129,0.9124038910497262,0.7499266262043737,0.18964632317851415,
 0.5570666829842821,0.6667906579442492,0.6175150506959991,0.8433673782270237,
 0.36837841384021974,0.486177157032857,0.5442650286481454,0.16594070366845892,
 0.7969225850277096,0.9551439905431875,0.4385519404962026,0.8888093692363652,
 0.010377645688286252,0.11738949455537617,0.21991689959231653,
 0.8953403187335174,0.114578394995005,0.11616984702141087,0.7889066144228998,
 0.416668648434777,0.9274112778344354,0.6102444829221249,0.5889116269175012,
 0.9381006929849935,0.07227104225663072,0.916869103055014,0.16862142138956382,
 0.4212857615299853,0.34036258841623834,0.31266998594541084,
 0.003107900901814009,0.7445220025621233,0.8903521745519984,
 0.5696218103890075,0.7669022619229754,0.33428400191885677,0.9923013817609855,
 0.5453350004759436,0.0018871558518005482,0.6711594170904849,
 0.9073236396499278,0.2425493519554459,0.7000608721819518,0.8131322081880575,
 0.2980604646992333,0.9400263395462578,0.07451169105159972,
 0.34823194536628854,0.24329700027440015,0.8113444467744917,
 0.9729030703549468,0.3043753687191829,0.1474235764748021,0.09104953959238238,
 0.05298909510595562,0.21673073917527463,0.9861297898901016,
 0.28375606398140407,0.15640147866616805,0.9272300642322976,
 0.31803099595917794,0.04151354690394793,0.8944643863904762,
 0.5684286570160212,0.26896153887491336,0.2727606884346526,0.8354514008019108,
 0.9541611786446947,0.8101365490962716,0.512187041425223,0.7437220508751337,
 0.8457555349351681,0.0036667082262525907,0.5256667076858589,
 0.1558372723867777,0.010908612542207208,0.008823216841362558,
 0.8758408093316418,0.9815344897891166,0.5925861981892285,0.7905697999691973,
 0.6965454894169534,0.0534217323728261,0.2795678845635402,0.2126595159472663,
 0.3604769925206508,0.9753797174346206,0.008353671106187677,
 0.23507102156912063,0.7955095086847612,0.12385446899530095,
 0.48877590610759725,0.5808316844968391,0.20537093876937418,
 0.1031651371190957,0.7856496228837959,0.19585103755764943,0.8140741656545699,
 0.5668359947092954,0.5596912069957387,0.21904762166174907,0.7276069513304627,
 0.06928772367515323,0.774765330521957,0.940693246105816,0.13395321963594198,
 0.7959005816861418,0.8171336870838842,0.560337945747762,0.9434774697646857,
 0.3146480438734033,0.42337892310232295,0.2571282552363866,0.8111619568299222,
 0.5978855544366838,0.122954752730975,0.4379485780884286,0.281127634704058,
 0.550609530218676,0.8577449304611929,0.9131997983472174,0.419899383688235,
 0.7265846611361175,0.6740770873642736,0.529142922938464,0.8302773834080299,
 0.6596894940817437,0.8398488809969668,0.16264713434130418,0.4146819745881589,
 0.7188857161524244,0.7872718388196134,0.13627853967301518,0.2226635597427684,
 0.5234730700387991,0.5059416741254538,0.6965381584921995,0.33930720947075044,
 0.708900604184552,0.01066783574153174,0.31088954753071063,0.9259483581551016,
 0.7596135840650411,0.4129549115472835,0.3816869507057975,0.19412290984782665,
 0.8199064211957339,0.7644051887359026,0.8699641203822411,0.4168794537398065,
 0.28162277713195794,0.07857654009445147,0.2999222426471815,
 0.12766377887259617,0.34468552006106257,0.36734494348867064,
 0.1112342741371084,0.8642233970153502,0.66519627589958,0.38985787785318593,
 0.003390705740469957,0.8741267660729436,0.22760513828998552,
 0.9177633259953423,0.49523853479997937,0.5001812091373106,0.7357765756227441,
 0.007095533690829381,0.04087298446251553,0.22058611060647337,
 0.8704292078019805,0.15802976725378282,0.9084472004947105,0.9216202196218783,
 0.6604049967383581,0.021018316428686235,0.609679424703849,0.2724807859330781,
 0.40401791562569556,0.3330600958856389,0.9886519848282369,0.9149155853939859,
 0.05645005471901898,0.8394172585365328,0.19864505931750953,
 0.07442029610604128,0.733110636631591,0.9169513362356843,0.678369657820506,
 0.3266371759623685,0.6805994710431348,0.020798607176239048,
 0.5044082053386406,0.8971863138377261,0.32134093927416296,0.5658456469128561,
 0.08130936508598252,0.7743886696243509,0.15118070803355865,
 0.7807798855483922,0.49732572787193896,0.5201102018982905,0.9916324555818036,
 0.14131450741345786,0.7758135525284672,0.5240528775061346,0.7447157452125683,
 0.7064448357036529,0.4120647842535423,0.919370477183084,0.9805573347044737,
 0.19565252790669607,0.2160810571000088,0.12479432078879338,
 0.1649423889445989,0.5468044230149257,0.7809415013435355,
 0.008467158195180513,0.9722107626469948,0.01593764920183094,
 0.5223748182589619,0.9645402599629751,0.9483585861569004,0.5507643423518296,
 0.5432398329335291,0.3641188263012871,0.7091883480410095,0.44642252046984865,
 0.5631305835581013,0.8989211249652624,0.6036310694386106,0.9840473161712738,
 0.6266679689548175,0.35902185189463376,0.4513302391536701,
 0.36446531870020393,0.6194329825058024,0.7248391725587798,
 0.024513883157604832,0.9660613993244017,0.7099896187351441,
 0.40744084189779883,0.09772351368006194,0.710846988851768,0.5428544056438445,
 0.8042907034988372,0.9782787718762419,0.4454225486137767,0.18555026319570145,
 0.021090437965255338,0.5720388980421378,0.9415442942700087,
 0.8063186919244208,0.3036895580658928,0.038429080089144474,
 0.8822643129508574,0.5246172032250175,0.6066372513524597,0.5219532812232127,
 0.31369007785304137,0.2802563760808645,0.481423086447488,0.7255663711627339,
 0.3621890896140343,0.07566370795779931,0.4260116470258404,0.6171048043603118,
 0.8621292761542199,0.6509367869776008,0.8624439858640992,0.3006168781422761,
 0.9241221299367662,0.26560470525711066,0.9959495999733328,0.9883252287948128,
 0.8577865701071471,0.5501349376906526,0.3649462372489991,0.6190576770286536,
 0.40998295313310307,0.7286397239946769,0.9351449308465776,0.6450131814158934,
 0.43424124457584135,0.007181555094783465,0.7837218516844585,
 0.40235802235956275,0.40916705225996697,0.4905169835403267,
 0.5592697334263765,0.4110189257147421,0.3833079442128511,0.4804058181382629,
 0.8116373935436031,0.5459148445431159,0.554000593041713,0.5682180996255402,
 0.44183630413221975,0.16470537982990752,0.7894122530604317,
 0.6615598607958261,0.2505636490283444,0.21685400206824912,0.5294174219660803,
 0.9078573701066548,0.7973608691472264,0.9262163445505263,0.31244618000632385,
 0.5458226313964668,0.8637614389345244,0.17903815875235995,0.6522120754512581,
 0.6097150718283141,0.6748481410374854,0.2491319572713966,0.2419698122526166,
 0.0022400911415154656,0.4156375356107974,0.8743895973831771,
 0.8707681064091646,0.7623260470900619,0.09001200870983972,
 0.060592942066109434,0.2806739348771201,0.7353721772265217,
 0.7535542988899379,0.5010481233004943,0.8870442026243663,0.7083774552146415,
 0.5306890656912762,0.6024625106940272,0.8974472335760862,0.04504793590423384,
 0.7077953720456023,0.3646035789721349,0.7469552674091319,0.9111206221607329,
 0.6897779190945543,0.09707577245393162,0.9656093032639839,0.8507753778031808,
 0.94978303861998,0.5825664524406423,0.771298273264578,0.35791736654291206,
 0.6211576711229867,0.2792528196002907,0.8516139112111035,0.8661988069612228,
 0.9814554942406729,0.17267664882556522,0.5980244383207806,
 1.5893824205526208e-5,0.8294062561516219,0.7954969021132063,
 0.17651507690316282,0.5734339028163918,0.8421244530476039,0.9207783904845002,
 0.1587836442074677,0.8216781763460536,0.24955558374608777,0.599331918409544,
 0.5943524061452927,0.7605294641611994,0.280570245886514,0.5784388828454549,
 0.839601313403766,0.63219172753397,0.2557642611642583,0.01366287098100949,
 0.48204130656276556,0.4887199379368081,0.10338288800271223,
 0.5010765618639079,0.19097996924608762,0.3217832239314986,
 0.43934605051749176,0.3281637389184784,0.008714030839200748,
 0.40039648988603616,0.0030235230070886354,0.672992608743027,
 0.548750480367726,0.600914548027516,0.746350083742322,0.157543396701318,
 0.6697399396535499,0.8161290370198424,0.6392621763888631,0.23759589514512247,
 0.5904224732395184,0.045283351317168985,0.8323866702016578,
 0.8598770237625182,0.04508025001154703,0.45048815875802795,
 0.6785189413316858,0.08947516462157079,0.9039679809144404,0.3758596986803522,
 0.3016445348079001,0.9535849630950723,0.2147771245525376,0.2996528547047439,
 0.6554670148110988,0.6577729429797964,0.9912721890190148,0.38441059815691847,
 0.5780486686688255,0.03180915944215157,0.11710036178800054,
 0.2502950637850244,0.6615272734660548,0.8322620430556057,0.42773378907251924,
 0.1721032911286855,0.727411274311889,0.6182193516598263,0.4518466461662495,
 0.11636151361274161,0.8979414824236066,0.5812735122493335,
 9.911218381480902e-4,0.10448966963770978,0.3941379951170947,
 0.5796141294692204,0.7841118005412308,0.22965619307178897,0.4875901496333579,
 0.29955647749276937,0.012749615712520201,0.9496931857274064,
 0.07072818050786223,0.7316807058578974,0.9965770271195632,0.4683802218775831,
 0.17473413241583258,0.3562779306692171,0.4574311479754307,0.8300982280115616,
 0.9040720907193288,0.8017644518398985,0.6209611577914801,0.8254592684539915,
 0.6739708598789723,0.40946941263163694,0.972870241372143,0.12414628570227626,
 0.716179679596348,0.45151031386669505,0.13456663477921227,
 0.42822747004784245,0.14014699008570286,0.5913835652692414,
 0.3163451844462358,0.8194800603847561,0.641607163068759,0.6676262891007193,
 0.06621299443717976,0.1504934054338729,0.3875252361791093,0.9948880589971144,
 0.930614680319603,0.3755453166175524,0.5696536895941428,0.655372068902049,
 0.17442059860807058,0.2961386684654578,0.15963951691313794,
 0.5817291365877693,0.5459910411120595,0.027285493194888155,0.646160419248665,
 0.6573853702985195,0.3749830029760678,0.793935980397202,0.7467762936800305,
 0.6800448944530303,0.45919556320569854,0.6406152383614284,0.6124485661522728,
 0.6256273340897074,0.39327087395227645,0.5058807822477421,0.3481046782028644,
 0.45629461549308403,0.86660073283538,0.9667706261090402,0.33493982676186995,
 0.3712210132167113,0.6742900816786709,0.6042273493182412,0.6704925067366279,
 0.62662188485992,0.4026374127710881,0.9815055565308385,0.37656099057293113,
 0.8471211061759575,0.87432620617513,0.3638921591934938,0.6475939506316689,
 0.748265170865236,0.9905456877179296,0.8130573235898668,0.37008403618037156,
 0.6002316472437397,0.8610138397915171,0.7028550769307864,0.031411834241869,
 0.45904713671558295,0.9297992526483336,0.6041394629286605,0.4189061975777141,
 0.39857085453976326,0.03935750963098361,0.2880615692457612,
 0.3022215458328549,0.44993749979525655,0.9723233527247683,0.7160915236838984,
 0.51058944233941,0.863439542087639,0.7975991471810413,0.5593545575860825,
 0.5736867896861352,0.7733066731885502,0.0880884152245297,0.5024350972546957,
 0.1906181606852707,0.358646581116582,0.9898846757106433,0.24254160035486705,
 0.3530723541193542,0.46308157735478084,0.10020579777282312,
 0.18123684958776398,0.760929611833176,0.8959974819565895,0.2735353826692034,
 0.8567995874415196,0.8533100231371359,0.6791325218920188,0.6235824319065316,
 0.029502522638921058,0.046192830238989124,0.9024188836216609,
 0.6928416167574817,0.2652870435628497,0.2981449050491235,0.9830404902475456,
 0.6657245255451376,0.1966355526840312,0.6450793731732795,0.36057960567380576,
 0.098914126202031,0.3196961023447509,0.589988797535676,0.6596535297401241,
 0.4735851550382919,0.5401910014574266,0.3911564843220394,0.5433122580592619,
 0.6075196723993999,0.43374636649218745,0.8533737087019149,0.4520132115468358,
 0.39585211324176384,0.908814586530274,0.6500781775806241,0.11412497580795022,
 0.14885956212576357,0.7966322642201484,0.8942823474366014,0.8446560680258441,
 0.6633691577111133,0.24741878737868295,0.5559708443021678,0.80816381812172,
 0.9406993936588053,0.28933037413632545,0.9755874590336376,0.9019707408755773,
 0.49682262179649506,0.15306500463764905,0.5950430175908619,
 0.5155190708540052,0.356309122976448,0.060986967704275474,0.3054950001106441,
 0.654182585427585,0.21587828513892704,0.896103188334395,0.5606112705013504,
 0.612242776171083,0.3952089378375121,0.15754136322746481,0.06972650910098077,
 0.8076860085308146,0.023060057863489525,0.9886026342010393,
 0.2989958471123906,0.6151400183748438,0.4109641508692734,0.8661041433918918,
 0.5052739912344706,0.8337368377250266,0.8739224908913772,
 0.0019335295220118676,0.5463652706103681,0.5909190977583618,
 0.9174820615830575,0.7536176241392636,0.3980735586588591,
 0.006804815792493923,0.4595179285263993,0.5705119842470985,
 0.24165777062037053,0.08293853088242842,0.12708231799497482,
 0.13088381887069522,0.7783774821306423,0.3229155910698336,0.3512863512162713,
 0.7468774955600812,0.357719308881277,0.3367489356131006,0.28837446461497784,
 0.7687568093173962,0.34074706170411906,0.6219359507631715,0.3874116374861899,
 0.4699563181576043,0.5545279780821336,0.36685730730685984,
 0.0043913374270745376,0.1514881634796983,0.3639569759206063,
 0.4750259562044624,0.6712523080270282,0.5871415354812726,0.6359068934464381,
 0.4237439461865079,0.45678252927944796,0.7293629769907564,
 0.16257881844558697,0.47901622799267485,0.8817781241793972,
 0.9620812401924225,0.6407301354323204,0.711668058252723,0.481794493548388,
 0.43475009701343403,0.4296840799695304,0.3157434189399153,
 0.32638599699846993,0.782849127130522,0.5769103598056159,0.4717364825919621,
 0.3765983672372891,0.48819844625580444,0.7170734945845362,0.6196469690582825,
 0.05040758559833347,0.6749358130397572,0.3417382209328268,
 0.15920273063304924,0.8647669762976948,0.19584879562875201,
 0.6928600827266114,0.5880780780576403,0.5200717077961896,0.35502793838155766,
 0.42371643217955823,0.17564225806406597,0.8733752040141779,0.718444458929371,
 0.2447479348269822,0.6224482434322338,0.5372036932566461,0.10152830195931295,
 0.18814882163602775,0.26637002297370316,0.27652681684203206,
 0.11264968934111663,0.48275544257658376,0.7305923462410848,
 0.12360495056527943,0.6052136115815381,0.8614567302710255,0.7390372194328096,
 0.8191241946472833,0.7294969265165249,0.31014276504185734,0.6435732118696321,
 0.3980560901062258,0.6065588566578967,0.2930994905496185,0.5658831569013444,
 0.3029549365344656,0.12676264519034985,0.7045646805202975,
 0.06283270273838737,0.7176946869708278,0.5301999205343816,
 0.08418004065986784,0.2012694830279379,0.26321470418520976,
 0.049908888244466265,0.18183303404459583,0.5360858913583527,
 0.9418169651823349,0.5328129956224884,0.8880453944913771,0.03850109224745757,
 0.43046924371239303,0.493440763037285,0.25843058622206294,0.6533349640419237,
 0.4230731046830556,0.745402048781492,0.9685138912131301,0.9941246794558822,
 0.866444146843286,0.9206801133861547,0.8246917959897213,0.07491761564831667,
 0.7269547339963134,0.6893931598301034,0.5212932889524222,0.2814917386686928,
 0.2698428796882012,0.18047933658279314,0.17662324940841656,
 0.09895541551454368,0.8054525796368288,0.8641975256168467,0.9660178760319993,
 0.8141639350240283,0.6535848845441767,0.1110132025241044,0.20067762452000582,
 0.4825325848453923,0.5349248448688906,0.8360252300783392,0.7221501210148195,
 0.0565824951060232,0.40149391454509675,0.2641389985139214,0.5601046864923488,
 0.2258627738756207,0.37179169150848834,0.9658524562821178,0.4235881602557092,
 0.8662447739759856,0.6422411731533353,0.41776815209273277,0.8567618018388591,
 0.9915109475522423,0.05107146348753078,0.7324329055702002,0.6541420419405699,
 0.016448430913658996,0.6568870573061747,0.16223620763515867,
 0.6201737519532488,0.5376929872387756,0.8664155181826874,0.9492798205864397,
 0.3963230167942783,0.804837025094433,0.29036256686291384,0.04403420081679821,
 0.937532835058615,0.8117130766466855,0.006749455923374725,0.9037518747137432,
 0.39475634529650494,0.22819461271019814,0.921641526176365,
 0.42808628879153454,0.9763404972936096,0.8482629728048302,0.7696837604053897,
 0.35403467270331457,0.22907076050897734,0.18842252050826813,
 0.4844729931073464,0.3727955764894316,0.05868184927850628,0.748725395930469,
 0.1733435866528522,0.8432603873050943,0.8625522703553983,0.0711274348541292,
 0.0612479282067423,0.6670096337939091,0.3594813443814002,0.02755976350800582,
 0.109106831523472,0.28427841500324835,0.40534445214032555,0.8660546995415437,
 0.7368573641367048,0.1447923840980745,0.22017769975661938,
 0.41741434720838777,0.30065131035894155,0.5596621539657466,
 0.2746807414641611,0.6003451730905413,0.08699945997220737,0.5653576757194884,
 0.16009633656195654,0.1561120659234908,0.47866025679798274,
 0.4949436778680423,0.3752168496592475,0.3035809549070315,0.06369537209508236,
 0.7810809407826905,0.17552501181873337,0.17647507968749876,
 0.3389460807838285,0.523803661745866,0.6472352641843931,0.14255550045581056,
 0.6118434367275384,0.4838064975727159,0.7644716837052937,0.2412013625985151,
 0.31522224335750093,0.6207013403183949,0.45860756559981686,
 0.7195910851460505,0.55487339314874,0.3548208965408396,0.12007027285099281,
 0.680236141539253,0.7468975487905267,0.17182121582196674,0.34033360411651026,
 0.8339661005434568,0.8217908070940437,0.05994103489610403,0.9431193587335505,
 0.9008072378547204,0.8581278181201107,0.04301295023786322,0.1972982633508542,
 0.19848717089079382,0.9743778985513161,0.4878904336778469,0.3129808388949661,
 0.48985098270751104,0.2596626318637103,0.22689076840006206,
 0.4301544843909455,0.1837074385738503,0.6302959118383556,0.2357577442581259,
 0.44727468284944427,0.13569650392682475,0.6771069027140966,
 0.3086770196094253,0.8125872038299959,0.48181859859719267,0.1516891729133213,
 0.5709030014310906,0.5466148130654829,0.3535148854260364,0.5180525989197324,
 0.23710293731194965,0.6309220957153612,0.9691477598233402,
 0.34706022482298415,0.11987058159714503,0.04805954362883613,
 0.4719941870548978,0.21466069380510322,0.9240143428112103,0.1080482336229377,
 0.9577628475077437,0.37188435866890834,0.08539374588365956,
 0.17683947904834874,0.664048369473677,0.46231227903090866,
 0.30910495572072105,0.42884060354628795,0.9335683355337959,
 0.3656535801556764,0.6786939821858691,0.8187470181625945,
 0.043813263380944156,0.6302404789791307,0.9102332150890541,
 0.6551076050315596,0.2976646244326937,0.047514209217609404,
 0.24277572155021976,0.6158799543238908,0.544271972768807,0.9795352459318609,
 0.531400854704733,0.8515388568773866,0.7771650030154286,0.6315762110450874,
 0.7410646103147914,0.137240162300152,0.7825091985109621,0.801828509082089,
 0.5185386382305778,0.6977440812929507,0.422433204817376,0.4309214850618963,
 0.48354296370102423,0.9842170180215647,0.17276849475879175,
 0.5795848260174519,0.7341993205831802,0.6877412799293006,0.14672286940646806,
 0.5664281989258122,0.5243904143325079,0.38035116425570714,0.8596731555658123,
 0.9981157974579979,0.09865985793641208,0.7094394555242338,0.7862562593570591,
 0.417070192096065,0.11370273817984156,0.14525486823715594,0.9275576025177042,
 0.825025847578954,0.24321206010397523,0.9357154534140197,0.8320094504049546,
 0.975127021503106,0.9365240796400156,0.06649868554669647,0.22968794732339315,
 0.7222914894240408,0.9751288304025675,0.12135060578011303,0.34784767261684,
 0.48838046915890054,0.9130558778723544,0.7017051941112913,0.7601710047618992,
 0.37485376747305554,0.7425121869109064,0.1428128667635732,0.4951748754099706,
 0.5115173438557806,0.4436415702155182,0.11707034077142975,0.9710944191397273,
 0.544955696006887,0.6354594176310858,0.3185638635019138,0.7434033121014253,
 0.8771785500967618,0.49148763656767724,0.5944615641760711,
 0.004398579056157925,0.5558394366440723,0.0311072087386286,
 0.8921433158946943,0.1684672043456229,0.7583280276339324,0.06508737177953117,
 0.336678405596029,0.9112284581919641,0.4985376440398097,0.927148799193064,
 0.420345975254815,0.03928910955133447,0.6108871526233511,0.5737246368716479,
 0.22510548335196057,0.4588744575352487,0.11608779026832472,
 0.6416593148711395,0.5016861233326168,0.2853114799347122,0.9047721727344149,
 0.789188469091227,0.3158245476965974,0.32333913667777625,0.0493148417953595,
 0.435611996131268,0.38192143750609187,0.8703729831156063,0.5076941115720401,
 0.8668383241687525,0.7429741907750694,0.6098900281085693,0.8261835040257041,
 0.20465484220203078,0.43281234335591323,0.5779478547377329,
 0.1822328564948854,0.4068928416687373,0.4952784251198007,0.7873849135347833,
 0.42924705653271455,0.298407714121381,0.1804407188242524,0.36192780345536424,
 0.9799413633134869,0.12624533586054643,0.2878604168610913,0.9077846872095984,
 0.04635693464151902,0.09583155462552784,0.5916172276669973,
 0.8140934945385878,0.6597746914697478,0.4722057503555922,0.33458302162650766,
 0.5179962805157964,0.4109360154047361,0.7932762224985447,0.02861165180792813,
 0.5206436047404982,0.5020951814797067,0.4156836198218362,0.885271612819222,
 0.27026061701739246,0.741840496278576,0.5795614584958495,0.6993117702071988,
 0.22036907400839056,0.3236195572848495,0.22649594514753557,
 0.6797635210123867,0.8611754110616973,0.9200807616456512,0.9406724006311735,
 0.7834925209820028,0.9232367828275178,0.5832838731451643,0.35780411586504224,
 0.3385784927189003,0.7706390778145265,0.7894015756225284,0.4934932828341083,
 0.9403146867881216,0.27741498368564277,0.6743184506970499,
 0.061315290712316006,0.7307285895613231,0.7691684832227305,
 0.3999060132501966,0.5945297006728181,0.6082620835722949,0.05314541101959147,
 0.680808114025449,0.31596865417941566,0.8487693431821977,0.754198013050718,
 0.3554667377150862,0.8509452631394453,0.8013134881202205,0.889463006713968,
 0.8239430457712837,0.6693336285806111,0.7186627718757339,0.24757443886274733,
 0.9434981951810264,0.966499285497558,0.2729702061144077,0.5408015717111812,
 0.3723786606257645,0.5128752848294501,0.3068731136274163,0.2508977580820284,
 0.033698628491283134,0.604012117000209,0.23945489796946307,
 0.1560816556877328,0.4753112863818231,0.35117634642003226,
 0.019881698872718734,0.6366924374065923,0.3793008501978036,
 0.8633832880177209,0.23645712286439913,0.9985734940118633,0.3076800134243457,
 0.4346051289433258,0.4234692821005889,0.26053525655826393,0.7192162559385988,
 0.3084000008330079,0.08383681966086565,0.8929641886544721,
 0.048027675957982296,0.9662183207166785,0.5115174402346436,
 0.22240141944237102,0.1670594868332892,0.8229584741094431,
 0.15256050897213624,0.7451553562663669,0.87949151495243,0.9438673988610013,
 0.7144358376958566,0.1608090343085281,0.19572207053136348,0.9044771213406738,
 0.21452891554951214,0.2744049641960361,0.2847911224159818,0.5545174299477188,
 0.32295283785729223,0.09282239199282638,0.8516608864719145,
 0.9739189835148803,0.4360372077137068,0.29989558181244746,0.2876564398173196,
 0.04330409618488218,0.5939666066031216,0.6942405384879374,0.6068093968918173,
 0.1242718834841956,0.5048125997406551,0.12379487965484459,0.6886967763625028,
 0.2100106624819743,0.8933043935232486,0.6663241412962349,0.9539626072574298,
 0.9033961106449092,0.4670881625889015,0.05622270702337362,0.9185862936343886,
 0.9113248400314196,0.9896735405484036,0.9838277687349635,0.3229950078231383,
 0.3090462214958085,0.4408632220516777,0.19413392079765013,0.8069012664494666,
 0.5733183463266661,0.09338337934268237,0.6462026923126616,0.1479954880393164,
 0.2873761886933104,0.5357195565346086,0.40875508435758745,0.8074418402233356,
 0.09682054562994069,0.35146824969059653,0.5376467084962703,
 0.5775379696132098,0.9212369763142179,0.5274257627235543,0.5620943404509733,
 0.05380706360920351,0.8934443330430535,0.9957186541467661,0.5203664641018416,
 0.4638215587358838,0.3564958031671337,0.4370219045111273,0.4096210509267699,
 0.12195108386866804,0.2429814249034966,0.7007467295334189,
 0.020028252349663433,0.946634459254573,0.9976494656843016,0.6698414584468148,
 0.4927735375807303,0.4783503530745451,0.30351185855326934,0.9091309606958261,
 0.947786197650819,0.8027238431637411,0.5273902890409227,0.06182347976041802,
 0.5832831364318236,0.1991391669910204,0.3172473603579946,0.9057944366681353,
 0.7362259357823362,0.6397284274409345,0.5914177216850066,0.6814596865200545,
 0.8213924633507772,0.04324864157618724,0.7491060973634931,0.8141703577369477,
 0.05720276850249384,0.138643765279583,0.5785696983704568,0.15494679951007306,
 0.5120404023220094,0.5286452566780843,0.9040290036118915,0.8080029848116475,
 0.5093864909038803,0.694906070210501,0.7353590371755964,0.033147911784442985,
 0.9422104891684056,0.40115655615267753,0.6299684377954943,0.4727694829911959,
 0.3006185652120382,0.7871653238329156,0.023356130914648943,
 0.6704301432975724,0.07971595718929336,0.32861987244299384,
 0.06838328673674965,0.9283358505452123,0.6717947284466923,0.8232035015720806,
 0.865691696210239,0.3705295538060076,0.2734287746358115,0.2800130396021464,
 0.7417537815208186,0.9883267543062048,0.4048318983481206,0.48447289725777365,
 0.6649215151810215,0.7622175940067822,0.6637026801123262,0.3839603240276971,
 0.13909006600537854,0.3046790426992928,0.3422092372248041,
 0.19047793862784457,0.11786975636767338,0.1576932171825005,
 0.7641406814842038,0.5389307330480754,0.332634256766538,0.6091712070442776,
 0.5688993785864622,0.9740426357494273,0.572080848462372,0.14740100395735256,
 0.6482919486092551,0.10053979877752706,0.07904624879176603,
 0.5783650777022652,0.17586272910754197,0.35603566088668237,
 0.08227308960203228,0.10452626181071223,0.283635941554815,
 0.49744913309653327,0.9851030636834306,0.33943201369038456,
 0.10062151889743043,0.14517160841520593,0.9884810746975118,
 0.9730104216409187,0.12190484465873164,0.5359490013215193,0.7779619955066019,
 0.6093859340139676,0.6253303621281224,0.17764518616864,0.5436483215001175,
 0.631172920899837,0.05322998693921288,0.41107476074438387,
 0.36928680539410663,0.15104973676305877,0.4985228815940802,
 0.9454117483086515,0.7657933320943143,0.6508467344667029,0.7676970270399179,
 0.3620981862639143,0.7984691663214747,0.39077034411553935,
 0.18050801152421436,0.03617925802087163,0.33960164862569986,
 0.6036372704005296,0.6484153623217705,0.043566140601790204,
 0.3130643825070889,0.37148933747038937,0.7450619336454604,0.8259876683615094,
 0.26356913543342764,0.6647737299394724,0.9639885183547063,0.2710073211241927,
 0.7111723004421578,0.034418326990663384,0.5080217461926866,
 0.8921009656525577,0.9365735027493485,0.7661519636054227,0.7307493602677653,
 0.35961585869460433,0.09743094615652137,0.04688517017750782,
 0.463623218911012,0.4054365266028107,0.5779330978346398,0.7281540244287881,
 0.472406115359455,0.9374578293563375,0.5142750360342195,0.6688719794557524,
 0.31357492111347374,0.3905698269628588,0.039319185280941804,
 0.6689387696106309,0.8751369660350568,0.8910596373907838,0.8784718536610614,
 0.8722977876200925,0.018272842652718446,0.9031966275993024,
 0.6314778264256053,0.34255764456504767,0.7608229971175967,0.4714465306810507,
 0.6935684839384583,0.5448559634826918,0.3737429917519616,0.43113040839348016,
 0.5201758690632252,0.44134389823703124,0.8217178435526495,0.5434108492276255,
 0.9003877293187313,0.22607448395803875,0.5804391252802295,0.8569443413918496,
 0.803920907090224,0.3666538821745331,0.9860856892774469,0.38760826595995135,
 0.5444288061139293,0.7624170889367788,0.16016083342070042,0.735936949543108,
 0.21352924610996438,0.5010057810596561,0.4445084365193175,0.8211780825237958,
 0.3997500377096307,0.2080388902361413,0.6693050874289732,0.7469854204251326,
 0.19130347035525264,0.5841956804523892,0.8823117434628811,0.3233553377589351,
 0.08644548043637668,0.7497259726532186,0.8431268499838732,0.8355531568466535,
 0.8107226420977021,0.5148744358271835,0.13741143437663594,
 0.19556531548419842,0.47345293319743575,0.44059714373337033,
 0.02676759346901869,0.534796566050008,0.8901683214726228,0.5079093471394762,
 0.980740782665469,0.23596181710313224,0.07987021220823098,
 0.46914027948058223,0.11567383869772474,0.4948618582311408,
 0.9457934022848306,0.5227076728657004,0.5259179683168146,0.9121034816924584,
 0.33987728199180856,0.7987577132580684,0.5296057145261478,
 0.44585969953827986,0.26799997851726276,0.9173516907476541,
 0.8686915689622556,0.5689677153143885,0.6036192786307294,0.5612760048292261,
 0.9647638585898122,0.3849199882395259,0.821372152367069,0.4391643429577078,
 0.5752155816489323,0.9232757347795353,0.6402381336431519,0.45855010436518984,
 0.044281565044989146,0.45820354365289173,0.5384599487465791,
 0.4706500627890391,0.18294051521504295,0.07921138463722288,
 0.6192160392502277,0.43095089984088963,0.7082533017086563,
 0.05419721563899471,0.15256913446737852,0.6925542695380681,
 0.22856262924723236,0.5799595251255563,0.5043476116375003,0.8730988431323706,
 0.6986609843315769,0.273991227013495,0.8002888769289245,0.6881261224974222,
 0.29713608458848006,0.27987018453006884,0.6874925417475853,
 0.9528603605752426,0.26685501651284815,0.6625194407075752,
 0.022259475588588096,0.01993069151774851,0.12669044882939284,
 0.2779864388313047,0.6961934162109537,0.25181796804588563,
 0.08026541517912045,0.11700061740443979,0.6803529863479612,
 0.7773465890280232,0.7132758971951862,0.10329868047126745,
 0.11817495365532671,0.3706451460016156,0.22751372901235745,
 0.06068680436895446,0.4733408062914616,0.3803761133295285,0.6118200704347815,
 0.3523227050447528,0.29433505316526254,0.057685810252206426,
 0.5139024680074387,0.752693992605569,0.12261829088725928,0.3149874661408947,
 0.6724297606878273,0.48198795105914727,0.5367322671637366,0.449977597272619,
 0.07432224898823181,0.6377414754908565,0.19130393503011356,
 0.9940758732240969,0.24068307849534143,0.5447304620027537,0.6123511980796199,
 0.7832622453647152,0.9308308070401914,0.2920168017693231,0.5714750490268835,
 0.3596607602250491,0.915319760906117,0.43755376701405424,0.5710488814420356,
 0.9230796654044802,0.006287955437681569,0.8245238755829469,0.937631768705855,
 0.5005003247295304,0.8181440851892353,0.12331388193258963,0.5153103697375944,
 0.7733157179407151,0.1328426769105121,0.31604929382295754,0.8864269344288626,
 0.03260838133542732,0.18197600662842106,0.13525769356557127,
 0.4547878272614585,0.8530575617303888,0.34268758556824097,0.7038806147690735,
 0.5702332049052238,0.9145071543807897,0.7249119075501294,
 0.0065067838091784624,0.5736834096037113,0.5453432193493836,
 0.5472709326715504,0.5282329247089521,0.18936948003021103,
 0.31647913621789436,0.9670282481330793,0.8313377455853941,0.610162381224775,
 0.21417758571275536,0.028013087726499575,0.15931331428128215,
 0.45904572650606845,0.27777294933338603,0.9529225688337275,
 0.032571021500849784,0.3769651772166467,0.7676488405821176,
 0.6146711645482704,0.5731523030625929,0.1841540891462533,0.4503070037087634,
 0.10216156538843935,0.3223160843709044,0.3859676591643769,0.6068018533093713,
 0.9657627557347065,0.3043076398805431,0.7023106317307785,0.514421188165395,
 0.641431625784959,0.6707118950740174,0.342282873164099,0.517004458129263,
 0.28436000820829654,0.8252216916265128,0.39011184483364525,
 0.08465490426534761,0.43844043885821005,0.8136629124715574,
 0.4494491165877944,0.5847817279410766,0.6185072646690539,0.6456809334602577,
 0.44531378298949276,0.39265244015896783,0.2719839473535527,
 0.1628921564483392,0.11213236892576073,0.655236081135504,0.6189783693850113,
 0.2067139208442056,0.287649134537048,0.2065501224439199,0.15423879911816551,
 0.2995928654587945,0.7703656939281915,0.9056836056680232,0.7989731766023265,
 0.1858036335126263,0.38098737575895236,0.7593315235870727,
 0.02934838170986276,0.5099040764499884,0.07756788011052262,
 0.047529453473532435,0.6440859619151331,0.6544584469296675,
 0.6595365436693026,0.04450311031692844,0.12674820816193488,0.76071914597323,
 0.06623974624714313,0.007085089861298521,0.3629603820935442,
 0.694615187909967,0.12001156187175233,0.026040014117244548,
 0.7347187479729061,0.7586795767999696,0.32170231803765514,0.6357309288655237,
 0.05168913463704705,0.2681505004488951,0.5097768974018344,
 0.42222374340963265,0.21057320053116102,0.7615752719043835,
 0.7965958267639145,0.6792739824744259,0.6199675584717952,0.11247184212817274,
 0.9494204023623928,0.39759999063196494,0.03864583903124763,
 0.1100978513978551,0.7308097736000548,0.774523918236771,0.5432361637777154,
 0.23368164609983444,0.8614004984416959,0.11467042894443114,0.782673349909332,
 0.5166390134127156,0.7739643621072693,0.756685015279308,0.20137831663307681,
 0.19671773290156758,0.6798892228198864,0.40849077466746675,
 0.1991028438543183,0.6152346870539984,0.47588423948670666,0.8718168375480302,
 0.38877901267902126,0.3688530494971429,0.88225018706105,0.42407119833375306,
 0.8298225156735693,0.37324027266546644,0.8762385044408404,
 0.33697411660555754,0.2399684526040179,0.518491799571556,0.781366332033078,
 0.7283896553107209,0.22145527439220825,0.042915234970891336,
 0.19481182776105288,0.6884422646281183,0.7950331683584864,
 0.00734971500588022,0.1857168895590614,0.6740547095575457,0.2105585379233601,
 0.6786515366039413,0.6040403416068099,0.4442750371592461,0.7670455292162572,
 0.0991476981642141,0.1967988195676531,0.6095121482587862,0.28311624131557567,
 0.001333322854032315,0.7278576741054473,0.033923214385661815,
 0.4413186372032568,0.17684867953810823,0.16420942862201404,
 0.9391940178409809,0.16807464393599958,0.9400731514686306,0.8331079476276474,
 0.41154420157954275,0.601189801425458,0.7435557826838198,0.5753499787477923,
 0.6502529348633007,0.7226692121634838,0.4687924127410418,0.7897958363693045,
 0.4373243814564213,0.9457490869616424,0.012626307065727538,
 0.12641640431120937,0.7009689543823405,0.01571879518424857,
 0.10377468103206344,0.5281823957676881,0.4847199940273864,0.9720030098168917,
 0.6107311008891061,0.9249455904839056,0.9778413107979932,0.23166482378939657,
 0.8951896308323889,0.3449176436454173,0.9938771184333409,0.7291555614445184,
 0.8975450517674061,0.40287035210960376,0.9004853845193312,0.9511923815804437,
 0.5346554758478166,0.33645110164712666,0.6611116635841776,0.7645643627821818,
 0.2298602668777292,0.007073949756893461,0.20021069237076217,
 0.07648310311298157,0.12984738451858568,0.9509286108848093,
 0.8393536226599726,0.8905746578153176,0.5779581235487177,0.37728202032782177,
 0.46397329572814516,0.038413151988950234,0.8275677755876715,
 0.6725327843077574,0.025772444887434798,0.1094124247158943,
 0.43387866747029247,0.5260062513996627,0.7651021894965765,0.7122162555442165,
 0.3514630857092912,0.37007709251301646,0.914294976337667,0.7891024508565686,
 0.2832594111945007,0.777502240621765,0.6933668107785353,0.44707328446990346,
 0.5499607310873806,0.6758055274904917,0.16968650405966268,0.601814187921714,
 0.8324617424522942,0.9429436648808396,0.6040067100494139,0.9800270814190575,
 0.9216254990130204,0.2243558069920757,0.21611814879268343,0.8024021703029903,
 0.8075805829473994,0.4594322699942438,0.7435968691270931,0.581863805843289,
 0.9734793059240542,0.584060136959899,0.057500301693220957,0.9990273613338271,
 0.7160439760283489,0.64372706073758,0.39576802050323734,0.8530773462646408,
 0.9885518853956555,0.8580465440290029,0.3288461109376928,0.22835149563102808,
 0.9211104846972843,0.17988769144535666,0.7894373108378758,
 0.22990439542760333,0.757987985124444,0.18879153474451937,
 0.38639562205149824,0.41941221043692045,0.9524280280796921,
 0.48553904479741095,0.4324508609459998,0.24310949959825623,
 0.4510497205379914,0.8303440135582733,0.36778894029570164,0.7995245689662746,
 0.49707369359320874,0.6708129723662004,0.1468872483257213,0.9898268683216134,
 0.5312385556535434,0.9276320277065251,0.7521706868245909,
 0.006089439571109523,0.1749862248436571,0.8274187885859466,
 0.3211754341652431,0.41044665900309174,0.7301169751582753,
 0.04678080987441102,0.15019444163113382,0.7489417958056301,
 0.5438832422989033,0.31481456411379416,0.5699311284144706,
 0.36968287546379974,0.5345207567694858,0.9627248994155826,
 0.24010186499709718,0.18347057419901314,0.8738990913170082,
 0.26618990468626147,0.6183353094359046,0.42398491981709086,0.063218516823077,
 0.014338213730080374,0.6796629144840247,0.7091271194788297,
 0.04805120828371723,0.4934480793582229,0.21982981697135073,
 0.3949235488086438,0.7964084319885976,0.6890857201111524,0.5485221620769157,
 0.9368584055440028,0.7597907211785705,0.7725794518852271,0.8738697380567687,
 0.471167809242389,0.17668191290497737,0.8046815732720245,0.26839343575990704,
 0.07153703611491857,0.5701236447412642,0.8725695207597279,0.5338273689806514,
 0.12719271580662816,0.8622541690147912,0.14519652280565087,
 0.7881754042531396,0.8759640896542216,0.36238288599034696,
 0.41433962153762716,0.04195726211215334,0.2670631268755135,
 0.9696230163748494,0.6253455943067631,0.43132760870137854,
 0.27845811517513996,0.49934966610567266,0.4966564649477352,
 0.8435691147301692,0.06741029321523517,0.3267954616657053,0.6314947093606775,
 0.9908484486618462,0.9501156397114918,0.7084792310541862,0.4727808630772309,
 0.4552431985934927,0.0548300888774077,0.5846553240327664,0.9522697111324976,
 0.8428815069041095,0.7309564036911641,0.7859939963006823,0.9539531139643843,
 0.041618027217843245,0.003347813269717115,0.3866852511440545,
 0.7013484461936105,0.30179680905398043,0.40257899056737556,
 0.09012721891756337,0.4500453069806014,0.07990129370987165,
 0.14120811506024689,0.2666444364799553,0.037475935221517265,
 0.9556789765795077,0.5683075158970095,0.37285817269821564,
 0.19405258783033963,0.5343597136014585,0.07749879321344932,
 0.9889388378644363,0.8777219705141485,0.6698603782467367,0.827476642110013,
 0.1179913154156692,0.5101769838208859,0.6897657893524902,0.2252959716530365,
 0.23208223728632282,0.8872859284363741,0.5327707697534866,0.86025060487467,
 0.7645448512853603,0.3501471530274507,0.015270627776678314,
 0.8829213588552844,0.6896193452492502,0.058638980902341986,
 0.6294365552205353,0.27768179233092294,0.028805386888748674,
 0.6415050876841726,0.5232321304679094,0.5852817362939872,0.2963971343689381,
 0.5084879479271974,0.6880400292661447,0.19017090358444377,0.4189272369154502,
 0.30857610919698797,0.8221276832573309,0.9516898427798011,0.5246639084234312,
 0.9633146696111871,0.23171815803467277,0.7801406753475851,
 0.42907376053170865,0.7111743339497193,0.9943047163726584,0.7530382556580174,
 0.008061079934640003,0.7858114112723341,0.2839607775729316,
 0.42896735843377565,0.8271279031306417,0.44401337524388773,
 0.28521521883187406,0.21438846522197952,0.9103325935204254,
 0.23316051695384943,0.7375459897351553,0.9352624913664297,0.452780894212079,
 0.2526183539584962,0.5646994862402964,0.024931322926882715,
 0.6176202945152746,0.03749286303152832,0.8574060163928916,0.7855791423151596,
 0.6015888217358533,0.6111059615205067,0.6530030564731076,0.5536954071570582,
 0.11377320754552822,0.010861620459964971,0.1183479569380238,
 0.6652564519169395,0.5468154610396629,0.5616849861476454,0.03721945434685314,
 0.2728677722246067,0.7511700495017913,0.6187239599941958,0.4690264670103558,
 0.9117037176094828,0.5378575892948253,0.33169977089447555,0.4987263000724379,
 0.9789486579437678,0.4357515178169431,0.6127794935547594,0.35624589519145067,
 0.34069547407224643,0.4803094362743948,0.3212637969581974,0.5556701345017702,
 0.3636364164659729,0.5119184767540166,0.104120412897027,0.5409777363822752,
 0.5730427165507973,0.4951111271737574,0.019867653229039828,
 0.07962521543960455,0.14031613757663242,0.9965219139290195,
 0.17489419517200466,0.28092267192277487,0.40208223280207034,
 0.4752631214947982,0.3225427379767184,0.5961107900392595,0.125405490328379,
 0.6310681864157827,0.4015401220411964,0.6363263510919538,0.2803793852656611,
 0.18647752315468713,0.007556665820709352,0.9220721902084394,
 0.7744515951470855,0.175506871457487,0.5137785739148507,0.8863551019557615,
 0.44225625485100917,0.9696347853460929,0.20416587408348164,
 0.5463080275042413,0.16960512267606997,0.1257056529227819,0.5023444142744354,
 0.3762118558268349,0.6740387786651502,0.45971743279274724,
 0.07473721408556888,0.9913396017406176,0.6325337717450951,
 0.40116443495144516,0.5488008244842237,0.7905071756293964,
 0.08825697062402449,0.7231435006190359,0.9231731321210217,0.6308770086925025,
 0.822877054023532,0.9213195726410349,0.014945191489018628,0.7217640922507196,
 0.8603170899191198,0.1821097458348504,0.0043155958098159974,
 0.9181045140727372,0.6996258801081876,0.03843087713117521,0.6456538556018796,
 0.14164959246505782,0.9640872336431969,0.2526859670358006,0.2406496489032346,
 0.018427912811930347,0.7532232143225375,0.5049923380009024,
 0.26842217865451723,0.1320384480521305,0.5042559096322469,0.9552570687422812,
 0.6204685050078235,0.2395720591675582,0.12939647059675308,0.9710800188064885,
 0.9553638063612875,0.023526435119768596,0.026888508050027182,
 0.0019165635482348842,0.3274347290499696,0.6897622663801747,
 0.06477954922028828,0.6885457773179505,0.2035369136921681,
 0.30124156101337984,0.9627982489522143,0.25649515539383294,
 0.4195215691065902,0.7029729606192378,0.14121282203004437,0.7837247353454443,
 0.11716847460845581,0.021751094946272254,0.9308796480951632,
 0.17217331142896075,0.31149927114720244,0.33594739141221674,
 0.44556550575545883,0.09800060849151493,0.9731855758891601,
 0.6162318090691665,0.5554476478294432,0.7792907538970034,0.35197063473051293,
 0.3825084010857237,0.16142674894371678,0.1004523538342521,0.9311821966195927,
 0.3604293111878113,0.24016817788651645,0.6376518452414595,0.8345775019987666,
 0.3568781409804547,0.217322870115205,0.8212336024920387,0.6384859581506036,
 0.9271066963653741,0.8336044598057989,0.8500225384422998,0.45423651312221924,
 0.7880606197839914,0.03262341094773469,0.20484454659880724,
 0.1208803580126877,0.03883925051015524,0.9733205411547478,
 0.007825953871189895,0.41031576581190465,0.14169304494451918,
 0.15706479863821787,0.11752841198784336,0.6178104906173519,
 0.6134253665642652,0.38919827267015206,0.2714953770410262,0.5525071137301126,
 0.08735592331006725,0.8258034144503564,0.3807498960325192,0.7396089791832408,
 0.27234571448382083,0.883494131015216,0.18822582351185857,0.7214370428435444,
 0.06772625605590399,8.862268332262557e-4,0.4578757755878282,
 0.5263975842430622,0.3594889868296073,0.3592957514306002,0.7661804088836718,
 0.5148767706936735,0.6283469267568281,0.4584542824499652,0.6589562007825803,
 0.8290968537204968,0.6129453482811481,0.2746682372472299,0.9610476666745839,
 0.903164095308845,0.2794369288430357,0.6311348688743925,0.6132283420237253,
 0.8822144108934749,0.07347563383004418,0.6968715336682478,
 0.18118683358581633,0.9233953815371582,0.9348590572724333,0.8786179438450848,
 0.48296873417463215,0.20666605800445315,0.9699851603666133,
 0.9633564628039775,0.1608568972048222,0.6467877062028494,0.5260114209209974,
 0.15634718248425217,0.04804267698225828,0.10127526924623065,
 0.9566408062230463,0.7374566146509763,0.2575899966260975,0.6901235295075069,
 0.8372597296155987,0.26836265914385415,0.07756370194300088,0.440990931023876,
 0.19573034450834248,0.6497986491920624,0.38256123475504167,
 0.8914041002752215,0.4592705210854636,0.24118615265262622,0.7834693959708718,
 0.6294572970360243,0.4066525710944775,0.1871483019648329,0.8707273543664438,
 0.744890956198667,0.81685791872759,0.38968712779894443,0.7567451460481143,
 0.5354848585498186,0.789689895846305,0.4815869071858151,0.024415240292079976,
 0.11690242463935452,0.21130929203294002,0.4141868587805716,0.962447659202836,
 0.35630269566502304,0.5432458509949224,0.8682544212221049,0.7867218194447206,
 0.359142558914777,0.8923664545963885,0.1311114805034559,0.7023370413978454,
 0.03682652201179493,0.21579927951463818,0.3538375605051953,
 0.8486772508187386,0.41681660606013293,0.5857891981955653,0.4374851178171698,
 0.6533061173508253,0.4048045921813337,0.7995894033877593,0.5259550520651081,
 0.3310240043138173,0.11211269278562497,0.8438682187177988,0.8092485394396098,
 0.9820403308058012,0.350171246511196,0.2725578652882782,0.6764232575219385,
 0.784355824485561,0.3483014884021963,0.36166211173678864,0.9629990175858869,
 0.16054326907073224,0.9223558029726703,0.42348789876012727,
 0.13141349177379524,0.879750820675164,0.9085852897137401,0.5194360839338595,
 0.7282045058614002,0.35780876049393395,0.14866093551866344,
 0.6047303753832575,0.7762704763042679,0.7746730172347,0.9743332280049088,
 0.724226240926332,0.628225792019778,0.1747369554933993,0.07694561747105133,
 0.7545167790707922,0.26836946539817574,0.7786624053617732,0.6370763302815855,
 0.39805352773512004,0.9924596527900882,0.9528688772774019,0.5171496189808734,
 0.9090900857785669,0.08082941845858038,0.41056976493715625,
 0.10123983459220054,0.8014183251660463,0.5653972637779313,0.7777624747587878,
 0.7529256739571822,0.12638113622007208,0.9098968133438646,0.3768576422190013,
 0.14092616066352315,0.6868872064902457,0.45265229245280225,
 0.27789311758637814,0.1970394250373001,0.28465355197080644,
 0.7257113811454002,0.9671530060826556,0.034460661535501314,
 0.05484476955157258,0.9857464226013969,0.6765307805173763,0.8169182533801027,
 0.8036398204952379,0.0675469384092604,0.049145660665169344,
 0.12667260305013017,0.0325618249046733,0.6861607216229143,0.7804208703519436,
 0.07317998916755464,0.3132706138066331,0.38526889008557885,
 0.13291820915147867,0.4023484747169622,0.0019146751016502206,
 0.593799592136797,0.0030450757201443412,0.136345014037778,0.7944599381287378,
 0.664211557136996,0.6504046147878492,0.24946662498383776,0.799828968697653,
 0.884844500669594,0.3388499381696517,0.6796496293623816,0.3711191708934274,
 0.9230601457530319,0.7529654644934344,0.545809043149582,0.6570004010150289,
 0.9850320802722569,0.3291355987584773,0.29040598383942884,0.8843593518643977,
 0.6701815845979313,0.4104767608145149,0.5595140448278887,0.09820451650655515,
 0.27409702151306903,0.5260754103793891,0.05126985980808163,
 0.6867594160966296,0.09848265871147599,0.42162509722467734,
 0.05600050519417865,0.9752288741700326,0.7580382208636586,0.2140189105609177,
 0.8844966347567722,0.9112831295092212,0.36914146349155774,0.6439740148095131,
 0.0648326342584381,0.35899491188396737,0.9236895296609116,0.6500574309868068,
 0.9358875867612033,0.9813212929386685,0.06802327156283028,
 0.35133905057217807,0.4769460774973574,0.32713293553238887,
 0.4136245837790167,0.3683265760908395,0.010271146740957748,
 0.4254130354413749,0.9784816869890746,0.6540406357987459,0.32569872119492604,
 0.3188488530761846,0.29021258054539634,0.12914962639749716,
 0.9111033129939933,0.2736149051647402,0.7032842822245526,0.2330365901048843,
 0.3008456422807596,0.38742852933322725,0.579750725976131,0.20081071017061425,
 0.7348847812850058,0.43802004384080895,0.7520448432922602,0.740536884755268,
 0.04584793206760018,0.9026578484298028,0.6229096518823977,0.9747454052649821,
 0.8697722222911449,0.2217728851580547,0.5851653017784946,0.8620594926164987,
 0.4076325706844417,0.5588826246460451,0.32228207640982454,0.907097966432852,
 0.7094395167882948,0.5196631271303649,0.6594543716733818,0.15025896962660923,
 0.5468843381421021,0.8140476801984734,0.7581722327065719,0.6770437517975422,
 0.6381245021217363,0.2623816686430359,0.6670047696867262,0.7352898144089142,
 0.7543083144488664,0.4889557610724322,0.6078435060944333,0.19738142250588986,
 0.6481650382254022,0.9285801523761645,0.20950375764389106,
 0.44500994279530026,0.9651293999245597,0.18609901478248325,
 0.8391144426579995,0.26346590486258037,0.8363828534912106,0.5060338028088275,
 0.7108085775543893,0.802669062778131,0.7203819375702526,0.9672150098925009,
 0.4754552592344218,0.3517885612941264,0.30488329190565266,0.2752482776086761,
 0.5761661297155422,0.05371111785820404,0.008031750237130875,
 0.5757901817362402,0.8555072417207862,0.7100250239438644,0.5617428378087476,
 0.16882026920607485,0.7661025881095964,0.7626216387650795,0.5796892214922887,
 0.06817899579518971,0.11397077097037245,0.05986082671246595,
 0.6082945280286991,0.7879826214179961,0.26954776608056263,0.8896285834588372,
 0.5555320535495488,0.14938115637135596,0.12653931526690254,
 0.6451855522558881,0.36839782547906863,0.07010231551352741,
 0.5041237781309178,0.5531408387913723,0.9730612296154941,0.7901239676337101,
 0.8380487854895939,0.36826239388784043,0.885200611723404,0.4557453133720464,
 0.644298025255263,0.7577494038319782,0.8883651934515447,0.6731154119727462,
 0.2931201781655557,0.4253022533075962,0.15700720894891373,0.8726406559953411,
 0.4945682659185817,0.5146527245946493,0.9546741916306586,0.27065871294384947,
 0.04678428608510987,0.4935531983611411,0.7548631414028772,
 0.07739554209886601,0.470557677764991,0.5797630029151507,0.24397989790521846,
 0.04081232782428734,0.6052402918629701,0.1116869324322165,0.3900257121011035,
 0.8374610754217789,0.12495806557183164,0.47308531281048294,
 0.6640474335171626,0.6433000721689829,0.29912304341379536,0.6007792463970228,
 0.8766932818875548,0.7601158700513662,0.2808534611032666,0.7807944650334372,
 0.6118026583844279,0.36687114120148345,0.11529734705480865,
 0.7983772124278729,0.5185577128679422,0.3992990635591075,0.6307721790481601,
 0.3825579049536074,0.546300587898179,0.9049449674202645,0.919418511231821,
 0.3086522499179585,0.06512764980140562,0.5230842379910745,
 0.19141669914563997,0.3044621516474755,0.9493753413735586,0.5173072193093505,
 0.8262521481324863,0.7168976271535301,0.26109233760219663,0.5613747698634367,
 0.5470966304782623,0.20913339934447817,0.3407996533311498,
 0.20885948841530344,0.19486050902785867,0.24511982253379438,
 0.669817803576776,0.8625364472265036,0.753468573816527,0.5577345997465347,
 0.28568519650942215,0.4167218685110654,0.2221783357174547,0.5051025192175412,
 0.23764169442597327,0.9111510254564361,0.0048735040751859415,
 0.32910465713596926,0.37334788120824636,0.0989625093527885,
 0.22050890609768137,0.9070450917872683,0.5666486249548632,
 0.21688155811896226,0.7607478483015812,0.4410746441432085,
 0.010062906606941202,0.12765104202745647,0.9560850183267314,
 0.48817081129647233,0.007572142660004277,0.7405247597373833,
 0.8134119471578677,0.10692078469197852,0.5148105203828446,0.4861744460647647,
 0.4825349323180521,0.10941326956132025,0.5622180811212992,0.4468693935794672,
 0.7852137802963471,0.09907679111080359,0.45559787372558624,
 0.9334515803583239,0.018412267440811703,0.8237076534008055,
 0.6233621745467643,0.6295947518424675,0.13430695311586272,0.5995690770481994,
 0.765811345450189,0.3072624982703631,0.24556744140565456,0.19569073158350847,
 0.09799262730825942,0.132615598436577,0.8606179916738897,0.7346269117458654,
 0.23770915252891256,0.49003456066006235,0.4749695839458232,
 0.37092150904856436,0.9640758394088306,0.40709637927782394,
 0.23951131530493974,0.6794796678404249,0.8186839907256773,0.4543182113804921,
 0.579814496382608,0.3759421333103278,0.753661631859756,0.7730218286821582,
 0.8190196006562154,0.24632572830317034,0.9908345308090202,0.9395606740760276,
 0.35109131379457414,0.7127154469109569,0.6404072511004376,
 0.26169327207113047,0.2442413172283754,0.9703713481290986,0.8497597709289679,
 0.34909833063276685,0.27411469267043254,0.9650140774647664,
 0.17026080589302628,0.3151658993554124,0.6775450507168059,
 0.48026021878190206,0.699709386230257,0.9336433027720528,0.42696800631431364,
 0.9983395186801551,0.9408741251683075,0.8399256586868037,0.3563616609363245,
 0.6360729205692136,0.38473444358055564,0.534606084518676,0.08542196292462434,
 0.40026087123736,0.9489959314174545,0.08967897402614822,0.9501473612551488,
 0.6120965405356852,0.7196179426537017,0.4148700132290233,0.7942242915761651,
 0.7984506510312833,0.23635414984075476,0.7731616255039889,0.645978754377537,
 0.5769017348785951,0.2575609739234618,0.5173866448287434,0.06535315880562342,
 0.8146327924283167,0.06536965054591781,0.3016552051503649,0.5150267052777919,
 0.07002213433851623,0.8974652901832206,0.3342149509512653,0.7524221637412702,
 0.259624644685175,0.6458723664451801,0.24577853611554146,0.10101785081807113,
 0.8606678715691883,0.39407157354339617,0.41276172425441493,
 0.2719278762339513,0.9799331979911675,0.4629184557844992,0.986406601502063,
 0.8420470635436845,0.4339394965921931,0.3557421402251424,0.9366335108908462,
 0.9930234605517434,0.7842929201331683,0.8667471179393049,0.9273034265687877,
 0.9584521803772232,0.10079804841235895,0.19643846314023572,
 0.7755913777507122,0.9170460684452394,0.6504955053980733,0.16610797614618578,
 0.48796418813261677,0.301827016195354,0.8133686935534272,0.6915220668749695,
 0.6784023995308623,0.26198822067618366,0.276768792308127,0.8942688054657792,
 0.6381330055334774,0.4259498085732327,0.3088854963654002,0.39972758167282096,
 0.9945416233474862,0.2252764717859651,0.3850448381449776,0.5475999013049938,
 0.7031840457614076,0.5527055023809534,0.16154102824146188,0.9302791397390204,
 0.7729374932810479,0.2104298251133867,0.47064260782399236,0.3456171126180104,
 0.8141959798890508,0.07260810579498878,0.547204485538709,0.37312405830813233,
 0.9276695609142964,0.2849650266205547,0.4867390769092126,0.48279420074396273,
 0.3989652492703484,0.3123280588126207,0.4940876606267067,0.25898319373660506,
 0.29420272532256275,0.5205525478254021,0.5083450648666229,0.7786507937374016,
 0.07830108495256483,0.18723819075788473,0.3834050584829892,
 0.9279828930363034,0.6341136338845281,0.5133374272734303,0.1751860791171661,
 0.4429744376953787,0.9620095268383515,0.534398913691862,0.788614144421469,
 0.9653404349216316,0.7307138024873499,0.8034750163615854,0.652828636376298,
 0.3751511232539423,0.43396748675621166,0.023314086759601893,
 0.5464236199247346,0.04912121022250604,0.8730843518372716,
 0.06371171261715791,0.8170596772114462,0.03548599097754035,
 0.1941684086693285,0.9491369422353922,0.45461501205350907,0.9601394364210165,
 0.6043855383673666,0.656240249325247,0.848724483108918,0.4590812324678739,
 0.01117500602760435,0.8331527481715096,0.10959925249670266,
 0.7406608375019792,0.8547862461065165,0.20354490640770095,0.8757092884381767,
 0.5625097725800668,0.19977511039012452,0.6616766109618861,
 0.013769335446111808,0.9281844866647165,0.5688800235378487,
 0.8871122399789124,0.14495354143070482,0.22216261131137427,
 0.09290598063804612,0.6264728024232731,0.6326666342009036,0.470263051261787,
 0.25808996278821206,0.13337075652456132,0.7088334159678283,
 0.10769227895822708,0.40919299541443477,0.2184229666300691,
 0.30407125301400884,0.4397732352402325,0.9704505771320242,0.2626479117130205,
 0.6526283070309135,0.49295930240902,0.05632901628058917,0.8562328297333796,
 0.9364920297505516,0.20268650515774278,0.23520512834538798,
 0.5948976296657355,0.6484306259841418,0.4019696818934819,0.9465532816503806,
 0.7000666964289796,0.17325540340933532,0.06412654402649143,
 0.1828446860463223,0.09541327544051903,0.3882009930465079,0.2311250031945834,
 0.2608566060015699,0.9088895738245184,0.1158873341992348,0.5117916408099461,
 0.7384490265231645,0.5726912327819688,0.5703713933835857,0.7390547252266089,
 0.4838078485823212,0.4256263030475185,0.020488725049054968,
 0.9658624845027476,0.7471389667888,0.11455265191546116,0.9835978371936904,
 0.9115874259281507,0.8474826129812187,0.8661231795214205,0.7697007942735296,
 0.05595704507163457,0.18728699092623702,0.9900571110823962,
 0.7059799129946348,0.20691670269883122,0.37200366339937263,
 0.8199887490495202,0.9474303754308936,0.6564574937349406,0.8816352276713573,
 0.29174618836220845,0.6924285302601704,0.7640961980517629,0.3174346632859597,
 0.06847338160331662,0.9603588708009557,0.5329456577701697,
 0.24257280157284855,0.27709943115931046,0.24497768420632404,
 0.42247255972265574,0.5605295243279993,0.6244708654875349,0.4637665322780711,
 0.24671452772324276,0.5313687387427001,0.7623826668733749,0.7558026559648456,
 0.03801733670612539,0.6742343661276,0.33057100620335134,0.5269751255696973,
 0.2155546922778927,0.6993562864790934,0.7640625419738275,
 0.057855105137405105,0.049750302578007055,0.9589963643733681,
 0.8649063977480592,0.7785610252833409,0.9007364532234414,0.13483324492715398,
 0.7260226302517938,0.9663346389037042,0.16660356618629824,0.7729677556795025,
 0.35099928426425475,0.21399389131238133,0.378568303828143,0.4033963007621053,
 0.5409563319903403,0.5093569561394721,0.7576780744079438,0.3651052931768225,
 0.4196454692023964,0.2709307433847886,0.2621097846212034,0.3566745590138354,
 0.8955625372412214,0.2742019120531317,0.9852972279295527,0.8748329294681089,
 0.29578290135009,0.5272124087391666,0.26846640503368047,0.3891406647384955,
 0.7148693431925998,0.3794131321804679,0.6497588094314993,0.37739192097200314,
 0.9971639071185637,0.0981108981015204,0.2664745236861995,0.36838853827064466,
 0.03710550989901196,0.37960301054092715,0.9010480940102816,
 0.8311085294023813,0.9841140520580124,0.3066378202827925,0.8189950053254017,
 0.004740045456475772,0.5916655050075288,0.757705512119171,0.6516187227773883,
 0.6006871057666936,0.7815961796620867,0.16893314376917257,0.7224615026415516,
 0.19209275825647076,0.7166289715322306,0.805223999919598,0.1999828844150836,
 0.6189679547847726,0.41971253687622667,0.8672066108506734,
 0.27957857328645574,0.9472562378959984,0.9808184306470995,0.5489759971189248,
 0.9538341272200416,0.038045446536106975,0.03308948016147084,
 0.635934029595895,0.3943370809101463,0.9940131239142235,0.22048330082820444,
 0.3809134540363047,0.06491276859302686,0.2891213465372834,0.5244850384816867,
 0.9479489161407325,0.326037464719652,0.05410023031194733,0.7439315709453016,
 0.9577604509838357,0.7212073245574531,0.47050780849959783,0.1584812066361937,
 0.5566219804386355,0.07328448466217619,0.09850188344289168,
 0.8120637741065813,0.25432183546569187,0.919922520165414,0.9352895618996485,
 0.9120970561375173,0.20985655461892583,0.7701598251238162,0.9120384816104987,
 0.2659780101494157,0.8521799413019718,0.7731206910232697,0.34982175956781525,
 0.7224073186568204,0.47439791991618396,0.2861880169458634,0.9981452318746669,
 0.663959012953205,0.502272187333238,0.947801813022443,0.3520968865326629,
 0.6557197160921896,0.04953437630322477,0.06524625020245323,0.765365764998818,
 0.42255689406561225,0.7542562648543178,0.7848699421417795,0.9033214622338549,
 0.8551717743666742,0.9507173240625206,0.02569188559203095,0.8822419129718517,
 0.7306436354790518,0.1240528608285324,0.0459826701327809,0.405794439310151,
 0.9018135746683354,0.6695808174976785,0.6457105281667739,0.8898797929956147,
 0.007832174732863906,0.4987059854824072,0.510455355597441,0.9828379315861275,
 0.17374428787004947,0.3591238714020091,0.9631142439679828,0.1326777213126148,
 0.1366071212562473,0.5663557580188965,0.2277377987473025,0.29332973587805267,
 0.5177870649157256,0.0727948614127415,0.029441708556397383,
 0.8246284309831751,0.5077769911406844,0.8493464282702471,0.44233397693254206,
 0.6295421052633992,0.09246339513642854,0.29004315385240775,
 0.3160027016700495,0.4932170519513037,0.22418055228852418,0.8141909356422742,
 0.6061404156078974,0.39605327950694824,0.1343250062635475,0.908119201122109,
 0.40702530020834127,0.8132559693082972,0.3001695724998331,0.6551280767612602,
 0.2879492044088139,0.7262585307388879,0.1546842278182171,0.5275842077267568,
 0.485681192822781,0.1328382190221602,0.6367063303562671,0.8060084486530121,
 0.20114628146011815,0.018746238606230925,0.8583039118270798,
 0.172340132356383,0.85077575955838,0.5718352287084648,0.8408100877041468,
 4.312174083448417e-4,0.5087283585659211,0.3954901842041604,
 0.9551992431025785,0.8430386715496083,0.7452761508845416,0.31874804246689825,
 0.38299358570019393,0.2932566693804788,0.43663678482203827,
 0.25284594690164697,0.7658605698482922,0.15947873984441907,
 0.9240942038386091,0.44542943851066596,0.40479286936745285,
 0.46112955982906545,0.17838074766919065,0.3720717931368265,
 0.8864994768219528,0.5566562187023041,0.6142344043631374,
 0.030210775201885642,0.9362843860733248,0.4524094545053745,
 0.9028002414083472,0.49073184205569453,0.33743328397199057,
 0.7704759193671094,0.09779429283404195,0.9275275309599806,0.4260169148248679,
 0.7658468776964223,0.6819235816379792,0.4783318589490545,0.10836809447120443,
 0.47960084621811916,0.25351970278794633,0.02196790257797887,
 0.6384563961533605,0.9104090415832928,0.5821023089767102,0.4944398462758143,
 0.5295112277467907,0.8243705396442624,0.1807569827440325,0.04668848521062552,
 0.0672364452056522,0.9083713150075006,0.6567309951098683,0.4341921413314289,
 0.9448033930241956,0.9322316501892202,0.6920803372647268,0.7554939560334927,
 0.7063810846772953,0.3148601614968001,0.17294069021235425,0.7590994750072827,
 0.8571599158609274,0.6341878051794725,0.4820001455261411,0.7751670214801156,
 0.634208226381107,0.8925965474460198,0.549002613361799,0.40102838266236773,
 0.7469308370387555,0.4039741888910361,0.22704235048961652,0.7628032254267002,
 0.26020770262644377,0.024661210510323328,0.2388686610363958,
 0.029731840942719745,0.31769363716835985,0.7695311076439614,
 0.9032444065411627,0.5678646452715457,0.04614913906983109,
 0.18939695248015342,0.23722104625952056,0.694195530533385,0.4174889960980186,
 0.2522417612025636,0.8365537321334376,0.8315150397228268,0.16034825071637304,
 0.6062428168596934,0.8081397080305165,0.1029398073748311,0.5468623396913762,
 0.6143124055695619,0.2824362053177598,0.8477083937766338,0.8433312503628827,
 0.09151098052444107,0.478607023038192,0.07756492870275489,0.5466200046213536,
 0.35534408880523494,0.30606425333552745,0.7217878704741812,
 0.6942168807321649,0.2544752599639677,0.8904138730838006,0.9308427140555654,
 0.44552164369885694,0.3553209002919221,0.6569627243794214,
 0.014802989213456352,0.5489291559650581,0.836897298881547,0.9339046890485527,
 0.8153848683479985,0.30731059715959486,0.24058080824368266,
 0.8561081104362409,0.9238208831134046,0.008460388735485713,
 0.8194828874254017,0.38823882376744956,0.0805205788847676,0.3411215107488025,
 0.721788705366101,0.9050625461822224,0.5735050718417267,0.9923395258408181,
 0.6482464166283136,0.5164519275992107,0.02502328302385004,
 0.11374918939125334,0.16299274939804298,0.9552894822312017,0.262194683497216,
 0.8898148999458637,0.7975799169851081,0.6363069161957919,0.7492790402563656,
 0.8777789337553417,0.5152401610487193,0.716805636316681,0.7671435721990557,
 0.67836548625705,0.8088812032212962,0.41973414364982253,0.79713486423739,
 0.6726598730241614,0.04745837288963761,0.8043982792150717,
 0.28704216456450915,0.7055104249800481,0.7496766768458709,0.7638977773367206,
 0.54852639224397,0.6110668478782042,0.7167677659195348,0.24578336686989133,
 0.3707945068884476,0.8975236065639075,0.00427946072830232,0.6092587201998164,
 0.4552060444350179,0.6734957534301264,0.23754024170826016,
 0.16714148780820948,0.8105936463795208,0.7223138405251983,
 0.09016089721239906,0.27218408758497314,0.6567311510657611,
 0.7697524739447479,0.45940729451563433,0.9303180199091429,0.7366483652555722,
 0.8542443535518313,0.34680594069247284,0.7722651715145421,
 0.11258122779425639,0.7987591049817546,0.8527193847929259,0.6031386663795822,
 0.11620225655667715,0.984950731911996,0.953280107322778,0.3431323296006039,
 0.8957542518508526,0.12237274131842169,0.49877581730670706,
 0.12841043397193852,0.9345213703597103,0.052271220323323764,
 0.7130677454144478,0.5195124762475216,0.6835608487915663,0.2146186855952129,
 0.24673733517266416,0.4087310236282169,0.5444344183679217,0.7835469441914057,
 0.13236068678002888,0.1268170609765128,0.05854291899646713,
 0.8324733459224308,0.12980646515574834,0.058670750994510246,
 0.15254835507823372,0.15806734193596172,0.4111827213490762,
 0.4765991902558555,0.31636862421743184,0.6734963850739352,0.9004523946943744,
 0.2838582916976824,0.6579817881206221,0.3090497637495796,0.19346102951604316,
 0.9754051761524468,0.07458903016110241,0.8192053737429409,0.9176019953987882,
 0.06989562574890229,0.4540054178921017,0.49220188069906323,
 0.19902323142175682,0.04952064205161344,0.43315171366222427,
 0.9877073684536358,0.017298285752148868,0.958612021213772,
 0.28824348116782816,0.9726601269253745,0.46013659128149276,
 0.6872526436406612,0.6178828175206208,0.611085501751144,0.4818556587046484,
 0.8800551350989079,0.5589930667817589,0.27639737024355016,
 0.037973422164395365,0.822369538859128,0.9327295497952678,0.9211984288688636,
 0.600293624220582,0.2837168143081894,0.42183253898671624,0.5662875326981082,
 0.3428703940034906,0.26607150911937183,0.5049878541913433,0.6440437806503166,
 0.22912679791491009,0.9357712425027674,0.9916835441415774,0.6237942291920454,
 0.3450517807661527,0.689097320050312,0.5108564242405172,0.5510120394911886,
 0.05726768524489245,0.14882993467702532,0.15262929532258873,
 0.2851872579436694,0.23501816000194564,0.5820191679554038,0.7106708729888559,
 0.14161223918119903,0.9511947348534641,0.5989262924203315,0.8160302663369734,
 0.4655155529552303,0.07678311984095043,0.8350844081567136,0.6017535042372449,
 0.3171236174004013,0.014787747277025254,0.005407898482370999,
 0.6993666272393639,0.2977826413418676,0.7744506737083566,0.688410569638307,
 0.6640182334115201,0.4964322618116036,0.9721692729122325,0.09147277721354241,
 0.7199152708003379,0.3884059747852032,0.5137542062705718,0.7825055615106657,
 0.24490902101012102,0.8124186044531485,0.41196845000288396,
 0.8694992623172406,0.0537079274582144,0.7886996842395243,0.9638997354117239,
 0.9245957755954157,0.5866407504394495,0.302382966476207,0.8141458817774189,
 0.5588627428015569,0.27846739688632927,0.4923252300544635,0.8241311588489317,
 0.953427203793681,0.9346291945867254,0.9436287172868726,0.024641272502436196,
 0.7523762525676115,0.1918445064358677,0.6695985501064982,0.8878674626881163,
 0.5132578658045294,0.16820905059337188,0.02530005760803533,
 0.9121932679862184,0.2077289527234466,0.24726729271156056,
 0.22843002222162734,0.1388080017603356,0.31822184260244846,
 0.8444910988177726,0.7479715022960154,0.49011721221494997,0.534555592033148,
 0.3113397646727343,0.909379009607243,0.6249796994289887,0.2936817216693919,
 0.6743780077802912,0.6631432306111247,0.2288486876446172,0.6094798765441513,
 0.5989642819939933,0.61683086800659,0.7009952182917017,0.9672821728213445,
 0.004525855419037383,0.6970446175541344,0.23170814978540666,
 0.9132067226570758,0.29676279068869194,0.646990121111668,0.1244981957012884,
 0.6620238879327393,0.15053678486112498,0.5519288544579494,0.742824894880808,
 0.5527517164125263,0.06128713383636131,0.5486312723977886,
 0.47581544371383644,0.7009575033327857,0.5063866485873025,0.1949772415540335,
 0.7556812005039601,0.577251750203165,0.44783482802551255,0.911429968834173,
 0.06627398368582993,0.6460299654310474,0.9043869079215922,0.8404427409203125,
 0.6198512222019757,0.8363890463886746,0.48969507509893195,
 0.07462470385227205,0.03260736030483358,0.5053751762832941,
 0.6668337291849346,0.6804582429324835,0.24715155938376676,0.0824533368703323,
 0.03249156977018142,0.5803472644351366,0.5042691526404111,
 0.07314079185980482,0.9406129292464676,0.7313916532245411,
 0.34121024059441685,0.3170186759626352,0.15553131675763565,
 0.18979257825742002,0.2875436152195673,0.05671162636457072,
 0.14108536642031166,0.8169786846592167,0.2588784661652105,0.4001755463316795,
 0.3637260701732099,0.5022293160773443,0.5223221967240634,0.4762887294828436,
 0.7713114045506952,0.8156815268842823,0.028380045478370697,
 0.6108351044799383,0.3370588433753884,0.7050648530245407,0.8608169656459251,
 0.26124640036372726,0.6242121286134371,0.4992057626934805,0.1096588438640842,
 0.8687671017009414,0.40085070186258465,0.011950688602578863,
 0.2364619917383537,0.47121469875866784,0.08469666293608624,
 0.29736042133723495,0.08490585024515662,0.7813700876094926,
 0.5441471576595989,0.9638183424190087,0.5511501826361995,0.7607107912238356,
 0.05632219945027861,0.07312808679749838,0.8020978460441498,
 0.15728008535722804,0.9870609734699608,0.902759500068602,0.8961416150907611,
 0.957139163322525,0.66238919363247,0.33537724009499825,0.9282256527771243,
 0.2825235866036806,0.9477916232311273,0.5474436854278738,0.6828714350373155,
 0.8021225966937147,0.09523328899514472,0.2396959857788663,
 0.19319422729473312,0.0244331770737386,0.09110297472038775,
 0.02403644106459657,0.3871468632540327,0.0027314702618914666,
 0.48980151745167233,0.31778483805271707,0.45290759608956144,
 0.9754332623716021,0.5675747593156292,0.6271847274953994,0.6355046091525043,
 0.8204200442372382,0.8502586292941743,0.046281636060492404,
 0.8256778829013327,0.06644823762034091,0.1357550856673574,
 0.17152829847618611,0.029021495819269383,0.5103966285859242,
 0.8643532237740145,0.5601892841504987,0.5567271320691406,0.15341835662304337,
 0.2111526409884057,0.9276084918815958,0.23512839734027258,0.6778922567838078,
 0.7068770724632887,0.9742087424444372,0.08192450589470635,0.7594510615167139,
 0.5423096073267664,0.28450803881156006,0.5987865850522567,
 0.22719571999260735,0.7992556159574058,0.4711926765629394,0.9371023029752082,
 0.20441607288305275,0.06267560847874276,0.12561419106926974,
 0.41604707528225005,0.7868137030279901,0.897731221716666,0.3338301466510485,
 0.6463684315863895,0.6780690983402724,0.29793332750528023,
 0.21724676710408708,0.7730924680448366,0.6401821084561494,
 0.49128779753173224,0.9897763930646072,0.40110714167015393,0.827471068134576,
 0.17805180017723554,0.06147610558399208,0.43594076746616794,
 0.3388747007642121,0.2895054923987088,0.2086018170711902,0.4107182246376475,
 0.01997270951906549,0.6968041558345727,0.43115097025168614,
 0.6323395497223251,0.11202799141219033,0.4347114503317918,
 0.26682064090971647,0.3572566978735482,0.4467986570023741,0.454479650854541,
 0.9429969503851935,0.990029752052303,0.10856325894390761,0.786936407542998,
 0.7820583337773875,0.9698746362318639,0.68506894807084,0.439342216212411,
 0.526525439551022,0.8414037673413068,0.9270133619877237,0.1944784879804583,
 0.6618380806448676,0.3235675720035124,0.5745136847119285,0.07089558679508978,
 0.05915961253792146,0.8644565524680802,0.13290439714084523,
 0.12890685484391085,0.7590207561612186,0.02605557138063197,
 0.9665772009973194,0.8887139454022719,0.28345523385878013,
 0.26446443680140297,0.8192426692471753,0.8771142403392144,
 0.12457838810368027,0.2099296113862119,0.6784971541935978,0.5873985537667081,
 0.20196950317490858,0.49033273750021444,0.2965143707484251,
 0.5807471480231493,0.6911801450987759,0.7100217316966303,0.8422929491185684,
 0.2117393030241813,0.6666832120534145,0.4586930581667904,0.5525124920482674,
 0.9329067458668131,0.3243591946484097,0.8264334514641358,0.5620027501748149,
 0.9050951240951326,0.2947121482354922,0.05608147653540929,0.7480054905498877,
 0.4953351607328983,0.6234758177712775,0.5198126716328877,0.9964082847642055,
 0.6653931784760158,0.7333512754057621,0.42173395188558627,0.5424464867284489,
 0.4045408008025747,0.6657741449904112,0.5600110843272186,0.34028171539723173,
 0.13722492646975326,0.6150990142491489,0.8462643917429725,0.8971692014358079,
 0.4397641604569553,0.5312997922781242,0.6749858081281338,0.9865040037521321,
 0.8116498738869716,0.5257038670710737,0.85376114987837,0.9051607560730375,
 0.9903474540593844,0.7850249612457812,0.6863865823432985,0.5288077670177789,
 0.4787538092675536,0.682241843926218,0.5043394909109312,0.9732427883296209,
 0.9755611947801819,0.09637428273534443,0.5973463938077903,0.4791741554519877,
 0.9348277175935565,0.9618169739229703,0.582790613966834,0.9406025603930434,
 0.9802419546543668,0.06821035311397972,0.8718872894546431,0.3510651287281076,
 0.7184410134909335,0.9144610874304813,0.7374408970154452,0.08067805865161626,
 0.6261722305801619,0.7170824833837128,0.7185161778855135,0.5930044592410483,
 0.47771565606678457,0.3998144254694763,0.3493380308464482,0.8184419672346958,
 0.08638531419629947,0.5650158413147662,0.21497118987108665,
 0.006040857761198892,0.6899593406506989,0.05020095194197338,
 0.8187164582775925,0.012083819381967564,0.26747294431029145,
 0.7460640956783211,0.675772097500488,0.7694889879944894,0.3619063322577907,
 0.9635413197431625,0.6851315663161971,0.8267832266436614,0.06800418554141319,
 0.5000550048493209,0.6029964088993848,0.7513833782180059,0.3608878252790516,
 0.1498564862710866,0.6855633420674817,0.9527490462897801,
 0.017244859226207332,0.23936563289157686,0.6488709450845176,
 0.5754344887191217,0.7856101027228679,0.06032018333872968,0.1543052444012526,
 0.2915281977617731,0.517532726893965,0.11789452850100568,0.5004151260034193,
 0.5637913852223594,0.9689054295299674,0.762379373482318,0.25032527940100635,
 0.44376272775554293,0.8496900328767424,0.7901955294231618,
 0.16611150099482164,0.5809392841793806,0.8047293009295496,0.830208367103886,
 0.8100858884200999,0.5576816920114648,0.768033979602632,0.9625878850075714,
 0.25668702716748926,0.9186894876876069,0.42514439405921756,
 0.27454814329683863,0.6417681679467642,0.15245419987164843,
 0.33203701542732333,0.4744820784465217,0.7298086095384008,0.7565458523094761,
 0.64012380124775,0.6510517677625957,0.7700524673700527,0.34019440636480924,
 0.9990883010003602,0.10966889736866514,0.995108260810983,0.7802078774057806,
 0.6178003817215036,0.9069797622827478,0.6714978874961536,0.43015886991677044,
 0.4959654708255897,0.5999591742036818,0.03457233432191009,
 0.19745721366582436,0.1468239234142803,0.3952005406445831,0.3483081259623072,
 0.9355009695538185,0.8867551231269972,0.9104567904851933,0.8012808201261495,
 0.1130788605741524,0.6148043608774261,0.7247725296308923,0.5668010045989613,
 0.4899131724023864,0.9250023394068282,0.430433580611726,0.27670166136905583,
 0.47671319457058037,0.41979953315225715,0.13222546350916964,
 0.16674318805154642,0.6477068452220592,0.9401050406816993,0.3915368346642535,
 0.030080541514074466,0.8986494156887497,0.5681447003591339,
 0.9725071384177331,0.751944823377213,0.38704640882052976,0.6213887528416442,
 0.24197597917962144,0.8339088361742418,0.8475367193696518,0.0940840689468605,
 0.9156457199095365,0.5518527841460186,0.9524238068138144,0.746205843723475,
 0.5302247358920615,0.09592032964435482,0.22569659595249103,
 0.12772042011766993,0.9979423602010232,0.6672059350128076,
 0.28382559797475637,0.6125596380324662,0.3207828436584006,0.2864021287721077,
 0.7475064248401271,0.19305173535719422,0.6131063435402608,
 0.05838386633940018,0.9382835641655113,0.20338182462858212,
 0.9346844564938056,0.07343466375284424,0.21315741828878654,
 0.2782301082332461,0.08560177049571138,0.029224461413231206,
 0.316696651970897,0.2457197934643126,0.3485347045802347,0.5872971544817135,
 0.5772351568650455,0.13823057999341604,0.9752774997584709,0.4014677156360572,
 0.6870700828447288,0.05218856791762472,0.4405696181159563,0.9680840653160855,
 0.10208964782668706,0.7119426894362073,0.5270127030524168,0.5795910538405011,
 0.8423259177136669,0.13075716262063253,0.7609349120675453,0.4340471434040307,
 0.6161410517844226,0.7929571346030494,0.8856937822612971,0.8730901780513289,
 0.27790673960935763,0.5649071611915614,0.6566865525244961,0.407500490437939,
 0.6159081500374934,0.6434139309224831,0.4126881556619868,
 0.033901452623433914,0.6421255109236362,0.5281208584167618,
 0.9277226278664298,0.12891274897800697,0.16206251396674864,
 0.7663670856766798,0.3995770044578093,0.060388319650634625,
 0.025997809708914077,0.2474095952718952,0.8744105657200096,
 0.2625051224942453,0.28617364058911665,0.8318451193308224,0.940307914643382,
 0.9959301500016514,0.23524745200680597,0.9170625691461669,
 0.33795691691135377,0.40610029881044996,0.14571714992423168,
 0.45830282280968004,0.37780371313514605,0.9922039584697844,
 0.5439706084627396,0.2753367009497836,0.5832342738577,0.6966253372667756,
 0.9772355354739255,0.023502274078840868,0.22546466193791515,
 0.7446800949623299,0.5382925463672743,0.35905735017941787,
 0.027265194559165984,0.04823951644314106,0.15098856011362727,
 0.9514294054155358,0.06129586976454293,0.7911364485343321,0.6979054634937895,
 0.17953293230995304,0.42678572701717943,0.014109862725831923,
 0.8463480883099055,0.4039335117141134,0.04704859899203351,0.9431781117838736,
 0.18696962818236196,0.2553625420182164,0.43127881052228856,
 0.2512210313966692,0.39843137112881033,0.7228221133027951,0.976119709474117,
 0.9196839427949157,0.2519643898169197,0.5991208582835958,0.44800589826850734,
 0.07637782257134651,0.6571651149328877,0.9882821451252217,
 0.17008091112479284,0.3373232059655441,0.7958807502645551,
 0.23956259507979238,0.3828938289007646,0.5506125347821826,
 0.47889018542098327,0.4803947218796847,0.16466919383289147,
 0.37263504764046396,0.3539285572486227,0.7295163432732451,
 0.09398587746103126,0.7073682937236552,0.3972757732321366,
 0.21012743288798785,0.25386925682148564,0.38716051873759083,
 0.4553182310135705,0.8819954803521927,0.5601739360940954,0.7292940504099016,
 0.6324380513468219,0.8952985298530579,0.5069133858651104,0.8583868212739,
 0.2007540954172614,0.272645506958504,0.8681221741857486,0.7132275529256816,
 0.6369419551994999,0.6993161160351349,0.10846993338427746,0.4206835169856795,
 0.3814355508887488,0.6121934785678891,0.05366359071684479,0.5294684900893055,
 0.012702638768577579,0.4519533346246076,0.5258772139218288,
 0.13614425448505374,0.07686116813110822,0.8301990653042309,
 0.7986054512511065,0.5466802116653963,0.05335008898623461,
 0.28746414103445184,0.35064377214297004,0.24637202536579028,
 0.6357025544322861,0.27852608944579904,0.9867013320087863,0.983628998889525,
 0.3791332897105484,0.6430557191920426,0.6525285129032682,0.26545268360970253,
 0.7824878231571402,0.22585922747623133,0.8688663151424648,0.9922302586485188,
 0.6983454798621032,0.6635002343608691,0.3467011235228,0.5492462451682328,
 0.7828888196592364,0.8905068241580736,0.448238964639218,0.6724233807751367,
 0.27805766813897115,0.7437950826209961,0.20842383495521677,
 0.8986729982893346,0.5832535668851289,0.782579419773597,0.5938593510695898,
 0.07682058764080646,0.7386442138325175,0.12703081693546558,
 0.6453226053970273,0.701216815319059,0.7401817356416992,0.013285863947439247,
 0.20506731864819072,0.871761735376486,0.5276014260260288,0.3366296518115006,
 0.5278198958832742,0.11753732892258362,0.6472357500271045,0.8349080480291711,
 0.03903671213109228,0.5042493838196226,0.18471202660526465,
 0.44601549663557916,0.3373263744573297,0.6515921805778933,
 0.18341583172360942,0.11860827105596394,0.7771988815012071,
 0.3681525441926865,0.4563993087430406,0.9637629923821358,0.6182248509280395,
 0.4635180547995287,0.6726535515042069,0.8135767050054007,0.6653463942988431,
 0.03277569228303112,0.8597391741957412,0.32035514263611775,
 0.07810476727619686,0.9225288573857494,0.8379845424943931,0.6951127214172724,
 0.27095737228918393,0.9796459814851782,0.3930284975923416,
 0.47168520344741005,0.9789941406267085,0.7832310715291814,
 0.23664572141571594,0.6364134563367169,0.03962432167040242,
 0.7156278498969326,0.6719936448913532,0.5745623391896982,0.7177109816165508,
 0.21466574677593153,0.9205287160247934,0.1878025166033892,0.8955402754364367,
 0.32579895994593067,0.6042317777419676,0.4595747277991775,0.7759105212080714,
 0.02864172892602157,0.4732841913243635,0.07036478583004824,
 0.30531715272484816,0.12131863078633298,0.04641292735645752,
 0.2009492383311623,0.15417329145728798,0.007074691050164228,
 0.6914608593118255,0.9609609457579853,0.7551676898267055,0.8652431103312225,
 0.29472509493177357,0.06125937307937479,0.3973502754634153,
 0.9635579282452508,0.5349462119212668,0.008753783408462557,
 0.4441757951629699,0.9618768383991814,0.06735831494552158,0.6713810730846347,
 0.44924440642906505,0.8090456711238648,0.33479032581578105,
 0.3258398458538613,0.9649258428125215,0.5102364713876794,0.925715520526867,
 0.06264024498635157,0.8044335729222798,0.2381548127422226,0.1757409758128372,
 0.8120229216467698,0.2915171866153805,0.9082429549220684,0.12377976646774924,
 0.9615597895088662,0.6514162045943592,0.06528672796749402,
 0.041563471748038316,0.8498879613678846,0.013417513804936942,
 0.0365121459149943,0.3461162099193855,0.7897610848182877,0.4217600655933005,
 0.8083100546845542,0.17192029281379595,0.8798942750211429,0.7284646315703814,
 0.5321991821695097,0.6204031210820179,0.45615683805487817,
 0.32067349932837497,0.3379732380759596,0.7268580146494348,0.5139858024526273,
 0.773230447487839,0.9598596456650755,0.9458865246389366,0.865136690438675,
 0.11697346584809909,0.7460066459794362,0.2798976524493627,
 0.03653531655635256,0.9504308691859285,0.28190068930365564,
 0.27126010458820005,0.07693315221665487,0.2167537049176117,
 0.48002055092082685,0.5482890375571069,0.896910173165408,
 0.018871765136382446,0.6537570740689843,0.04122932295227588,
 0.4135286675062573,0.9226841857980396,0.3885926539758139,0.12793774202167085,
 0.16068473721149867,0.6368574934073234,0.1930244553783469,0.3197673083175665,
 0.18154411878815202,0.4803686972029654,0.02338686445935767,0.79246259827429,
 0.3859663453130566,0.9488738924761813,0.6986723993753936,0.2234087652091279,
 0.8487841239347838,0.4327785585081263,0.4578384079798927,0.5154441150181185,
 0.6883124035113778,0.5187321310774107,0.27835754453946104,0.1926290710783476,
 0.21934612287331134,0.912753471614971,0.8404393421401045,0.8743597272433279,
 0.63927535921807,0.23219722482940497,0.4111591040898289,0.6297006366405683,
 0.14303880815218084,0.019903693584901183,0.1691774300078115,
 0.9516232122298132,0.01989168555880161,0.01667558198972663,
 0.8821471235723837,0.47892415627699325,0.07535858869525958,
 0.48191458850812185,0.22828791927970082,0.5198002068591017,
 0.0016298426031556845,0.3637997234656415,0.6962518839482901,
 0.6246416184078227,0.7917150059804487,0.13483017634688377,
 0.36148411676016823,0.10940979475509915,0.31854780458661625,
 0.28457716543192135,0.5084761588056782,0.4642413563371237,0.9049046961906497,
 0.17360469292894087,0.13467118395846545,0.6107149180029668,
 0.2341654553820255,0.11530995147239886,0.4917661175458119,0.497364632882062,
 0.17593804255101786,0.4112128591364096,0.2670809250669698,0.3891175893173068,
 0.47016305105477985,0.6674183232621081,0.11136685332702234,
 0.48209822125003776,0.5640433039060416,0.32209151631594146,
 0.4194027150786942,0.8968014982011104,0.2677560887176559,0.07297218391194538,
 0.6248597921068614,0.274241921334033,0.3419383574163245,0.7332003200025451,
 0.4753460987950455,0.001641708568461131,0.42983458371868766,
 0.04844188403107963,0.34074308962067257,0.8468156690955535,0.982697908856362,
 0.9590443602743262,0.9926694831089777,0.5417631584358924,0.6797560367147417,
 0.07348400653878262,0.23980940924405902,0.20778280934871396,
 0.5535777465046089,0.987213881659301,0.2585631603376599,0.9474160861939341,
 0.07171738394533167,0.5957836602545273,0.6517012619988305,0.684184702405012,
 0.0853238605152109,0.8575251637386935,0.3946654434422092,0.3277145749362025,
 0.007727154590971508,0.12514751012563052,0.9093513468332561,
 0.4188936848453292,0.5239044876476213,0.6543414402856924,0.04101629210791802,
 0.310213290276525,0.15719705142813978,0.3678966744559582,0.11334057572057954,
 0.45815211046390014,0.17012270056729606,0.14882913265421494,
 0.40102156140045675,0.42708453202990004,0.03024681483252334,
 0.6927529564850885,0.10242337991572814,0.458509144763075,0.2848303856328157,
 0.330155915638606,0.6888227876000297,0.675525383127519,0.9682520313109151,
 0.9801284600226285,0.4063421416195655,0.8533247109932381,0.3388209656160853,
 0.6751602295189171,0.06462394137964889,0.5948556128217235,
 0.16281001005928997,0.01980424347852816,0.5738375483536534,
 0.11486377453471441,0.6661710602719827,0.18288345218342383,
 0.9981979455738448,0.5102626489008681,0.7832529341608176,0.00336955695944674,
 0.04111649682887242,0.965322137910274,0.47242506614519386,
 0.16428267695607823,0.9271785410422282,0.9355365449429562,0.9619049741592777,
 0.49101178889347996,0.24861045656646275,0.07445736504890288,
 0.038552849318579385,0.130633481562499,0.5576810413203995,0.906663637528077,
 0.29069176599115176,0.023291066240328173,0.9523479213380132,
 0.011988161819763588,0.7589008399169842,0.8333924821735527,
 0.5079277886494936,0.5067742662238978,0.9493447722002376,0.0835777605899809,
 0.9363966209118404,0.3828226663075823,0.7236134778106792,0.5805412493442682,
 0.6842528420246061,0.26241802681383775,0.9756872650214886,0.8839422371499096,
 0.4520543170406779,0.13014224904134186,0.29754392800112117,
 0.1752719746018021,0.9045195923291564,0.37643002050179,0.27571212804221834,
 0.48131374937956983,0.7527623572202407,0.9900557227189608,
 0.28778817391843436,0.16905858473561008,0.3101692251536805,
 0.032661749844294596,0.11224722953201671,0.5649738753710215,
 0.9958447379097574,0.5247610086008931,0.481315716986607,0.38015439268890616,
 0.04446511573574974,0.642785487016552,0.5891586522302962,0.5372449471272395,
 0.874253468304963,0.8571104653956457,0.31297119850767874,0.7601636405419339,
 0.8175735135041147,0.2147206488478124,0.1639224423350888,0.2367646566135887,
 0.533282884991861,0.9698895348195774,0.3871578349366994,0.43825084149025906,
 0.30134746030393755,0.634680915367124,0.28395651007291745,0.1148876851738333,
 0.35135258074711595,0.9861874127138993,0.5027538047020089,0.5578350572930597,
 0.8805483946112226,0.3690723441501603,0.225377125382472,0.092677703831334,
 0.44473048794699466,0.09162092299797853,0.7637366457313508,
 0.7237722061568621,0.13585205378118037,0.22176285429853648,
 0.3987139326710336,0.6774794075257701,0.5755529041904159,0.19303566117581528,
 0.14640129121438838,0.9183490305370677,0.5863933767590407,0.8792141887125277,
 0.0950696860476743,0.07812718456038004,0.5520889161684495,0.888660223294705,
 0.3769354084489235,0.5963229798523295,0.5671460048114061,0.8323302854496597,
 0.5854488557438035,0.7973138875975376,0.7393068700225729,0.7007568232521014,
 0.8383942499401372,0.9309470531602949,0.34017482478866967,0.8468657234902759,
 0.697430211510712,0.9764286414600799,0.010995190899019902,0.646286211555934,
 0.7573043221360897,0.6111267456936977,0.04483707423353556,0.1981751367000848,
 0.24565965734091322,0.15418231316937336,0.20730096351299465,
 0.014393440511257571,0.9784082137117739,0.24729388983428713,
 0.6745397187372119,0.42265901417429264,0.4643994053234377,
 0.31147677387271366,0.438339714903659,0.93873122752871,0.7431171603385048,
 0.538238807863331,0.6551819340204735,0.21910259464120463,0.33724659653259037,
 0.6584306507970359,0.6398029698935075,0.2827938809166496,0.8741447297714041,
 0.9359998053032796,0.3707806970032611,0.054165159172200994,
 0.3578376078875586,0.23559091563270906,0.5804443713544081,0.9674205310395657,
 0.4754014074947809,0.17232464387263446,0.2811824801554206,
 0.19964531886180148,0.7823932971625503,0.7198664957846165,
 0.17697173163492275,0.596958964394703,0.700830884326708,0.9828995606222226,
 0.9476163880390941,0.4313371055312544,0.8530337608273789,0.4687963146691215,
 0.42528873797315314,0.3920689951321561,0.3214246785365329,0.5269163004717875,
 0.7790918201194037,0.3236284000646875,0.044629667188608946,0.438905518688514,
 0.20873578075344823,0.8165888824720738,0.20184164034625773,
 0.5302774037971995,0.45747660433457404,0.014355588969105049,
 0.8108186166883778,0.3043900019020396,0.2932757655274759,0.2766276185200133,
 0.00972345609644909,0.10894265575316653,0.618587523986199,0.9957684859116883,
 0.5355226795829156,0.915567726418104,0.2524915553177025,0.40635048281429487,
 0.7501338380421446,0.670054136379323,0.1936600911702696,0.9719783752839523,
 0.5432175426249146,0.7553610993591771,0.37259187706011476,0.8845270350806542,
 0.9488391042818809,0.6125658154316966,0.3014723593875933,0.13878281765776557,
 0.584874241525842,0.3269529947297396,0.7252534477557896,0.6839330487204348,
 0.784813486984295,0.4134691886427023,0.4947114903334279,0.7727783056376432,
 0.9646996131653848,0.5429889483637418,0.8320289941031298,0.9238638081091207,
 0.016505168691820615,0.46398998678627246,0.2822815262463413,
 0.15906053132582254,0.7523562511269875,0.8524546107860616,0.7128008693125562,
 0.006079277693669116,0.3177552835089932,0.2895273527200297,
 0.6800526051050585,0.9921800642722786,0.7500940609326472,0.39744921738089856,
 0.4670134264286676,0.4916431967682685,0.48404171870211166,0.8982385265890565,
 0.9326707775196945,0.823268378503556,6.043213107593814e-4,0.667471868334038,
 0.5555875742017631,0.19165317012809646,0.9991074872138259,0.3079816769866788,
 0.8513785247693615,0.42142431160775806,0.20407882987495918,
 0.12293468993096057,0.7617632685669551,0.23610118123080925,
 0.05836120690207247,0.9231368728045284,0.9286720525247785,
 0.40026767293941456,0.39652259415151225,0.5288017188654659,
 0.8509008053435124,0.41196086730459947,0.35022425276549074,
 0.21899575601223265,0.7261211853636981,0.5052596910248431,0.5744128209831209,
 0.8952687422008405,0.616091591771937,0.13573708045261978,0.4275817304732198,
 0.591053285104147,0.5817147818208289,0.9161134664861196,0.7145267654744101,
 0.07019640357419554,0.8490773257064861,0.4812616899322839,0.7743791821556677,
 0.7578500969411438,0.34013374991988066,0.9080161057189793,
 0.19932491154815224,0.5158154320010918,0.3800002113270555,0.4182235310468805,
 0.1171862361359578,0.8522897261991421,0.18755584564990158,0.2842600180777841,
 0.8751901547092257,0.2420683315966713,0.5638555265766693,0.4207844480024622,
 0.17564588067598674,0.11934222096255709,0.8661425526335556,
 0.6083991717996974,0.3768523182017264,0.16270948446476363,0.7999857334183884,
 0.5372647812766924,0.8301381865212684,0.9207941454668259,0.9058493634295076,
 0.7451377529001859,0.4747915917104144,0.5447872928060987,0.2469183000078382,
 0.5819217065832636,0.0639889548703243,0.060440118906888496,
 0.5212028803506505,0.04494732688448744,0.21517728282716742,
 0.23704735474364347,0.8362173382225132,0.48489083478298056,
 0.1543693540694503,0.7325909696140062,0.09679035905629285,0.9511498248189683,
 0.8426338863869269,0.5415316270780707,0.8511862184958241,0.08001503755143169,
 0.7168248520624652,0.8944911471791609,0.22316940348360104,0.2966596745589589,
 0.06731112159415009,0.7679873386148948,0.0013619679340607371,
 0.43499122318034433,0.8685000870283869,0.7099219106855216,0.7836288499658224,
 0.13513045167604887,0.4190554209974965,0.036365433611270315,
 0.2337670753347325,0.24532643437354396,0.7584279617152053,0.3513689172142481,
 0.4795400909409433,0.8487830942236517,0.2429270858687168,0.33599035326852045,
 0.6516819608331593,0.7861683970438711,0.686776742893979,0.23518782334587873,
 0.6546214628704999,0.8077808160687467,0.3375515643416416,0.5064470011888725,
 0.08018957570978791,0.9642126970128364,0.6464443337769407,0.793376528479456,
 0.16819325815352504,0.8761664466380467,0.7609900613127312,0.5694690928120696,
 0.2106173796933175,0.7412846288679412,0.692632703124314,0.38614006086561037,
 0.6143107925108089,0.9236646664408071,0.8712047250017964,0.677369464166362,
 0.0894318110648985,0.3259478318862442,6.475459447313758e-4,
 0.2681069855390982,0.5944409861006618,0.4463573913989157,0.5332820828418807,
 0.31927347720566757,0.838501023625145,0.22163616924719864,
 0.07620691102809185,0.777725463338626,0.8053178334392578,0.6701617554965555,
 0.34974340549100935,0.5935362807373296,0.44961790945325375,
 0.018764697569049305,0.10236819565415345,0.2399779364485598,
 0.18596125952610176,0.8475861681004274,0.3014091000844843,0.5599996980861219,
 0.2142723166504923,0.050603099651096106,0.5367712879551936,
 0.2295204751120884,0.5889526953178859,0.6166307414682491,0.5067534912117488,
 0.11020517387504547,0.7115193219545124,0.8293795364553113,0.8754065533348034,
 0.9013988488539468,0.6570335704131857,0.0014326603240956137,
 0.4029312921034047,0.4807532029507462,0.8854974698487504,0.47410487294845227,
 0.7666822362498459,0.21681742904131507,0.3530931531686402,0.1878929279569197,
 0.7802638719804754,0.3845135950518185,0.5691455341024502,
 0.016034411450011943,0.9338022602068954,0.9998081631493896,
 0.8489319906072684,0.8193014170329231,0.8252078769131002,0.4222490110158731,
 0.46410261568642386,0.6568647493676261,0.3533224599421928,
 0.03414433053129551,0.7187451739114076,0.2274793836244552,0.4045046819319478,
 0.09818420372223757,0.5233687950230286,0.09875263262169598,
 0.9679477955324087,0.9437337836681094,0.5460735900645582,0.1664733190189761,
 0.7847991397838165,0.9573434307140237,0.7189573636386901,0.7615847690769927,
 0.8085492617093079,0.5800229444853451,0.5622608080960849,0.3019832669755881,
 0.40988883212243854,0.6756538473597631,0.43474300509855346,
 0.40035612938159193,0.2172631284622164,0.7947797219244548,0.3621433578898411,
 0.3309224334391244,0.8933783479195113,0.01074588253799369,
 0.33844368325505503,0.43140113580378714,0.059030956064661244,
 0.5211761101763088,0.8100665366218146,0.10204901039934988,
 0.23644089850278316,0.4015341873088828,0.5768358023144223,0.5569213590639303,
 0.11058331894703333,0.19394303763877963,0.9059903506753523,
 0.39989995694080527,0.6169544627078801,0.9399395679949101,0.8721112034475582,
 0.1443882098547059,0.646868532840831,0.34407334360419806,0.19099104516843057,
 0.5784975950033435,0.3104068207851116,0.6877747572024246,0.959520304325819,
 0.22136077168999424,0.7957424206772532,0.7095467790915991,0.857494416609072,
 0.6582249322317743,0.8644486661494794,0.7763047691894688,0.41485145524360567,
 0.32237880586490386,0.12098898874178798,0.2576766738193841,
 0.4379858249591446,0.8665803826911986,0.16342940559794783,
 0.35047395829635475,0.8979681590921889,0.8146005013990671,0.8485633764859891,
 0.6975997714190592,0.10698682148457361,6.089811900545916e-4,
 0.8847815705729135,0.7614796957757134,0.9312504882547918,0.7542222567934006,
 0.7771473200227816,0.785475376535707,0.21320547901170595,0.6971158781442741,
 0.8317644271551465,0.6607757475047087,0.0327821661132619,0.29816864130020393,
 0.09374069249762684,0.779074806696038,0.21317487931871693,0.576473965429203,
 0.022749867423177306,0.4018731108479068,0.2638504150410047,
 0.024126604168629218,0.8042587561535638,0.7952494385931828,
 0.7824107918225532,0.5883494643640509,0.8820399921663827,0.6821143381191344,
 0.0329421523415393,0.21580160069570198,0.5391560758973624,0.7220113506816377,
 0.9889644554068567,0.9089668359578699,0.7546896584826795,0.5840157687291961,
 0.7374859048681216,0.2768772602058541,0.1301837702794284,0.8108319771028081,
 0.41803292842777395,0.43264881311152126,0.8079157037520289,
 0.7673181053985134,0.2265533598962033,0.34297316934201616,
 0.34432611099562505,0.3180865550798925,0.7170281098876142,0.659340408551965,
 0.27799700852905895,0.8473299550005354,0.5551446744189995,0.6970286161974235,
 0.40483700603140893,0.9046128256631627,0.6928607129227472,0.9755337964990216,
 0.6481205811098609,0.4017199519915009,0.1282366738776709,0.4153605737357755,
 0.19966352067557736,0.11656837823107902,0.19280055451801725,
 0.14908595575608385,0.671943984146375,0.30570005039679093,
 0.00882464186977916,0.6532664305214642,0.8307806844734179,0.9877161139921504,
 0.6141311749908516,0.7941548903093245,0.44107249213684685,0.7417502368490819,
 0.26610534373501704,0.8737653611271312,0.5389084532493,0.8041931586643744,
 0.7461277539343278,0.6928706068065299,0.9302862208773095,0.7108693970269019,
 0.04156734482683966,0.4931524170964856,0.2651288235658894,0.9539391756761577,
 0.894170387153899,0.5066140038696737,0.6930116637642264,0.7633022545724419,
 0.7429402767640899,0.2935430099526003,0.07105260886244003,0.6532659443451703,
 0.9406189074187656,0.7004252318897223,0.6767656392639669,0.1247287691930622,
 0.9005114894290629,0.09044643269603081,0.7358955532791686,0.6645645965086978,
 0.40052580873358423,0.041586138669200357,0.6553571693013138,
 0.4604425093974811,0.3760707400249451,0.14268490472849304,
 0.03240950584702207,0.9021873018327133,0.0674248366776159,0.2628861425612785,
 0.9313682925328203,0.9415168867728212,0.8075493870017072,0.4611118944867255,
 0.10054524567360223,0.7548170943069907,0.4535454629971215,0.6752488772588171,
 0.23834978829784403,0.4013743981860891,0.2904413927626597,0.295242509625946,
 0.45983043979894866,0.1427736736773445,0.035568377468534695,
 0.3239995098474815,0.6210275793900871,0.708745546207597,0.42239879377352085,
 0.6574958795065757,0.8818138599893853,0.11230808510647483,0.5521499014179762,
 0.5326130783360085,0.9979526860006126,0.485527716664937,0.24355901162206073,
 0.4673829926001464,0.9804739915860553,0.8655999235693966,0.8954530925197721,
 0.8457488999973556,0.2839069231066501,0.9454130323090786,0.8067911614857343,
 0.5867128506835629,0.930408434085851,0.1386027085185647,0.9853411346490312,
 0.13741847198855162,0.8836941811981065,0.4911824257389287,0.7484841397064613,
 0.438272529746787,0.664189133826306,0.37488895740496264,0.8625547335891625,
 0.6065018137716519,0.3811403646395448,0.4965396558652905,0.40801718884804516,
 0.1898780498447331,0.7384722332827314,0.21075032297849638,
 0.47569078636747186,0.9682897054417144,0.8264592909031268,
 0.13448477710290818,0.4860597883165232,0.3491870579145264,0.5256478398979758,
 0.6674325821256859,0.11651991823519847,0.6216144372500519,
 0.30113231913107996,0.05722882671703433,0.4795437498888123,
 0.37109905907987284,0.31039734486703663,0.2489176717424335,
 0.8937828700724804,0.9865052115566143,0.5430481650140985,0.8616813774811513,
 0.9213916939572929,0.2745436686079178,0.8165849366130065,0.7478784729347289,
 0.9364725912845361,0.04524222554155466,0.592456372555284,0.20020107737302106,
 0.058652415179626716,0.19721375690121734,0.04862674511142617,
 0.9072881642047695,0.8397170070477386,0.2718306367139022,0.4820976993404694,
 0.49617679853146335,0.04032968316904384,0.28195512849118054,
 0.728398741956371,0.2255323997664237,0.3675699564700561,0.8836428449580297,
 0.3303078531569268,0.7708217997639981,0.09815377484021692,
 0.07477730351128808,0.9930481635189539,0.34743890439151837,
 0.25892406350331043,0.059070610965471615,0.17025999289633076,
 0.06977897625129725,0.43956906025892484,0.06898545622101593,
 0.8679054203933847,0.4903000666629952,0.562910742539667,0.4753127132872902,
 0.3787790132669592,0.052299482691090415,0.04101090749911207,
 0.30124721265333976,0.9041696477185224,0.8992850408781669,0.7477142564919544,
 0.5608403217383782,0.7793618170346202,0.2587082709413808,0.6428078271527701,
 0.9669762820548915,0.5753434628802628,0.6806397841620155,0.33565071624704246,
 0.8105407667060953,0.25212793825420443,0.9891005137769708,
 0.23848729547152392,0.30715179082196853,0.6442844031728686,
 0.38426671622269326,0.6699167968343532,0.8096438909022935,
 0.15312035763529852,0.4315635518109705,0.6942127467465253,
 0.35290031645175457,0.30722372557475475,0.3408502266830361,
 0.44289258103982376,0.5182846205988804,0.04850456908445733,
 0.9433211502542086,0.8577766823199487,0.5676511835850286,0.7790474186367965,
 0.7072418605838797,0.8330650640466744,0.9429556046234433,0.7332195538328081,
 0.8251757696475249,0.2602169651757251,0.3728264668554113,0.6921727176649138,
 0.7267235455714416,0.2269569433922224,0.5067045380628118,0.26365682133340984,
 0.998448044330233,0.7877682446402434,0.45348466947303445,0.41648015814979655,
 0.3601870343956858,0.3659466723855267,0.763889637646663,0.2625717938277715,
 0.035005188271290955,0.3953197809065858,0.6658294380411364,
 0.7438891718810188,0.7508535474220712,0.5490412617283307,0.3939296590732787,
 0.7912866917587599,0.7992037199842965,0.8740267084594344,0.4524072497862547,
 0.12553740384132006,0.3950915594960055,0.7213467777872282,0.506728709399076,
 0.4717852319890493,0.08360148504024278,0.5462952229497309,
 0.47295202339154807,0.3301004290600248,0.5731888267155422,0.856881118135785,
 0.5360093030584729,0.9755838494099954,0.29066997280801643,0.5210410874852721,
 0.8799435901801294,0.3649198199568824,0.09894729602582286,
 0.14130974755210546,0.021213435281399518,0.6617625221741209,
 0.6569529084387794,0.016762359069124044,0.1025350527470481,
 0.30528151805356685,0.11049003093988219,0.4280972753338792,
 0.07517019753191878,0.16790803092469186,0.011660836514816308,
 0.43766538592048343,0.381214692575255,0.23768726682989172,0.5199529435570736,
 0.6878654816788576,0.7792599685022353,0.7923243231632984,0.8187422741667199,
 0.19135752031289732,0.6751931694617661,0.30457335741510394,
 0.6504001454035995,0.2743850835022039,0.49575164396543525,0.9079405200498789,
 0.2545833552081852,0.5461800259184107,0.18936231105725432,0.7954315395264291,
 0.385181437274032,0.7836058457558637,0.6354795818844016,0.32746821886631794,
 0.6973068125230084,0.7288010695583403,0.8980306074158622,0.47834230021708446,
 0.630340440154884,0.926379755582321,0.14959796911586376,0.8583313274027793,
 0.7739581738858358,0.4528231967011481,0.8706362442701564,0.05938058060112017,
 0.8130823802284215,0.8782509202347593,0.04454931571545184,0.575823441275719,
 0.3354292728327639,0.40387204856990133,0.3489825544742172,0.3941790530595336,
 0.40904833424405773,0.1215835929144301,0.01083132496945971,
 0.7284465266122573,0.21930378002922735,0.4180121132681417,0.5382868185247687,
 0.7089453523032827,0.11564681228004625,0.32852872089048246,
 0.9557576400173937,0.6436843787856366,0.0434405717493237,0.6556213861812326,
 0.24959193295779736,0.12862659745919136,0.626027160407336,0.8261069051580213,
 0.04330342189933978,0.32199837295219436,0.03721933731653959,
 0.05443515639599916,0.6053714983873336,0.6850314770371403,0.6583510725530568,
 0.2625944107906495,0.06859512666899437,0.5478181267624893,0.2892525796283235,
 0.6302091518663753,0.7711394207765031,0.26911361272509327,0.6597829145324288,
 0.5119407789977863,0.7186847057716995,0.3453361084918123,0.38192239489429214,
 0.23665101352167528,0.973407254008594,0.21121540980274311,0.5154173568302503,
 0.7434900783105481,0.9212021817261955,0.10203912046338826,0.9808121955891225,
 0.23193679529158395,0.8645445116953383,0.4806453019359922,0.7318497680947156,
 0.8487351037445298,0.7525841552618941,0.9775296511374895,0.8206789577545763,
 0.5380240917034331,0.2686889099670977,0.002597583623764388,
 0.6750821265324427,0.6076093879324047,0.7321497043133656,0.6426794256456327,
 0.03378655307604106,0.5310522405603912,0.9647248915066019,0.639357298208662,
 0.6115264371935516,0.4580251186199682,0.9595568689809951,0.05594788613384705,
 0.858141982914044,0.4339342199621631,0.5307672968736303,0.5561739052087336,
 0.9707856532235788,0.633856951178705,0.20503078421784293,0.7094046395284734,
 0.8647176245844331,0.48618331795208936,0.9136061754237085,
 0.02836655894281781,0.193983138111816,0.2781075715585306,0.7626521066430841,
 0.9664187826801819,0.6935569892021634,0.9365092154371408,0.6969483536851973,
 0.3016174205700547,0.2967926378728407,0.5038499561148616,0.09039815435391862,
 0.01188467393641468,0.5428313449536741,0.7566979617399144,0.8112164728509882,
 0.4017061024029629,0.07266848969559936,0.17402226413634148,
 0.16894498354807808,0.245024138399493,0.19052019514966834,0.2960846811077684,
 0.0675416971641688,0.1053048601111235,0.5997708453553386,0.7516398466468824,
 0.22155204780768945,0.6639365026531692,0.8415858526129129,
 0.42042528956532144,0.6600561780866299,0.5736736971172678,
 0.24297557217362942,0.6695361411582668,0.3467983927890643,
 0.10836829117229807,0.46084937913254764,0.8637563444839401,
 0.8160937515959337,0.9511468342862802,0.13954108421345768,0.6082110351686093,
 0.8484718508090006,0.5225955593222156,0.6673938663524898,0.19940398737451326,
 0.4541368485579681,0.5648209086576897,0.73884737495975,0.8958478841895915,
 0.38949519668836385,0.8526641520038167,0.9518023569082483,
 0.17259431227349942,0.17510128872536013,0.8256770821355524,
 0.25898467013535187,0.16862187408499918,0.3224751463760327,0.704338538644064,
 0.34112103659657367,0.06456284359345921,0.6602659124010041,
 0.6773509762996158,0.7666393231650117,0.9800148562754559,0.3119197958673613,
 0.7699991170085037,0.13445303464272906,0.9609438713435435,0.6847976416462936,
 0.22557178201013284,0.7734349442645992,0.9322901050633056,0.9205950431478577,
 0.6809341826109042,0.067309243844072,0.7917804797422155,0.28184722112572447,
 0.4820601564252852,0.03076140177288389,0.7031083229005837,
 0.32277566796962975,0.11335116694812086,0.8330239300300069,
 0.5478538615084471,0.9492905753132164,0.49709231677676713,
 0.42162303073537544,0.4842879657050607,0.7500941161991315,0.5552999573696097,
 0.9938879474794556,0.011381678333224232,0.35371195533087907,
 0.7996676035643637,0.7104447018513487,0.17994010335768496,0.8228602187651977,
 0.9287219819732048,0.6008926716654925,0.29893031732634234,0.1236055400004138,
 0.20325384934218116,0.048691648782779984,0.09942141979755081,
 0.3763774608669477,0.6127657790213781,0.4877855768136238,0.7730021258601881,
 0.3795732419212212,0.24035513944845577,0.8385753226892588,0.34097616761656,
 0.12952083658119884,0.40751175953334406,0.39759863164321596,
 0.28367591793703595,0.6058218854984813,0.460776349891715,0.42088580850718726,
 0.43636950546031805,0.22062569595027282,0.38433462103056104,
 0.33928904375318036,0.7727923868847208,0.3000908566625037,0.8365689183203443,
 0.16387702460965314,0.7875489285659154,0.9836278024197852,
 0.0018240497676890177,0.24557190523801853,0.5015956203666222,
 0.9609471320892974,0.6889307473548759,0.12737472333134336,0.9417446875599275,
 0.117058294407688,0.6575493102749238,0.5725213081180553,0.2817180719636272,
 0.2205168496109008,0.14644922400441962,0.9067957406022982,
 0.23042550883701995,0.8250676065948637,0.24096794653089149,
 0.8353924949975585,0.7596359334624315,0.957608311191307,0.7168905626233494,
 0.6005790767184568,0.8535035149607852,0.9775241405507111,0.4097308919157119,
 0.16830423487289725,0.19102477254367567,0.7384326102950912,
 0.15757922492764098,0.5298655430733709,0.9209447205807113,
 0.21312335241267366,0.2980661171056026,0.8186187450134268,0.4914894698175045,
 0.3230091627328061,0.4698569665898785,0.8202252131598843,0.24310451988664394,
 0.873334312804929,0.7285521991787173,0.39153756946892604,0.5858168619480147,
 0.9627889134605825,0.48991525680123227,0.8674419298510758,0.5784526749914404,
 0.8706007773161049,0.44695739402191703,0.07912462869880854,
 0.5863424328654273,0.7260131051021164,0.5945475755030946,0.6964603938151999,
 0.5369892577013237,0.6087703059463431,0.010818265610757605,
 0.3544000560908722,0.6737856160109544,0.5820001770108851,0.6218274601410341,
 0.30310293448219816,0.7410326193667195,0.13956064783656297,
 0.8854459426425199,0.7920361617152786,0.01517571691742936,
 0.39746859023290537,0.27191979507258357,0.512427415733729,0.8014862652551464,
 0.8698913737191798,0.8854083426586482,0.5243400934002052,0.9409718257691383,
 0.42312220105583254,0.6475217708244886,0.3952234811916089,0.8377255783613766,
 0.44283356228217985,0.560017317904517,0.5266985714802305,0.05628436511659263,
 0.2074184347344996,0.9379159545188299,0.9597554157698747,0.9033146902909186,
 0.2135535480236066,0.774900857961961,0.9770906848543245,0.6178354801988876,
 0.6230568031242821,0.278555227115478,0.44118441363540917,0.440330938246436,
 0.42217348251490283,0.30833312945520397,0.534173963527777,0.5596049800450388,
 0.2247700451255381,0.22195717423074124,0.04524923492916022,
 0.6190477935457027,0.18062014333700105,0.7610522802408957,0.6270617078706548,
 0.49802936794876584,0.31958851893770923,0.02274604994110363,
 0.5794881129654476,0.45128570980418414,0.1443352744578701,0.8936247907710254,
 0.40980225916471613,0.14079146854601254,0.8363594815591936,0.850163189473282,
 0.4598556105618491,0.8115246110250348,0.5210286227917124,
 4.6614704034797505e-4,0.8543765602704445,0.9188097739852732,
 0.005676779624740669,0.8304297639953423,0.24716333854845685,
 0.7953990642094084,0.8184625056331156,0.9832980756398177,0.7542896355576392,
 0.6921428780101678,0.048144322084068825,0.20351332359160257,
 0.519732497773328,0.21084832967583034,0.4986169275941039,0.5834035816298301,
 0.5670606530322603,0.0027316180707792626,0.7039255781700586,
 0.5717187964093118,0.31472041499378745,0.03263260082049457,
 0.1521543490492656,0.5957779329289595,0.8286803681499241,0.1906221383426393,
 0.22812469676942593,0.09587872702888645,0.9953418674117285,
 0.9848707040541617,0.5372296485645841,0.9557503615258628,0.1929287332705092,
 0.2492886687287199,0.13639856098873038,0.5089271944764339,
 0.046720620701613313,0.19705538033696512,0.16977717733322695,
 0.45894082126447344,0.6044342565331168,0.25665774070047576,
 0.6910172184049007,0.3350454926513794,0.5742966655997691,0.31098160852605716,
 0.7588809541641431,0.3307143643069901,0.98983102220873,0.5663071132100044,
 0.5396696104687922,0.6015853214525609,0.05919489780088116,
 0.08886859721379659,0.9731157933360183,0.546141025308896,0.8436580385357708,
 0.17888841331205352,0.7408811664687466,0.7029108880109074,0.756265815767917,
 0.8363706460992216,0.15255618645199986,0.12791191954110837,
 0.8214543378582687,0.7818168383822655,0.20967236777521547,0.5107896516995374,
 0.029850627303633304,0.03676816098133695,0.48874515662267626,
 0.1679383373625719,0.4264674666101337,0.16135561662585407,0.7716877661095851,
 0.9786166516736572,0.5443560945443128,0.18362771291741964,
 0.06915200730668458,0.10235528284108808,0.5506264490993169,
 0.09688760764616644,0.8653016041677355,0.26881673436013565,
 0.7893346322027056,0.9142436498814069,0.3946188957349418,0.27831147306155424,
 0.8258702178959432,0.7135344920681003,0.506319996629751,0.7629266474459414,
 0.3304514140195668,0.42376687605330543,0.4261804114798011,
 0.06852559022397142,0.1797830665397493,0.8573949523271929,0.838488050570736,
 0.5295607520119576,0.8284443007923004,0.971657087987689,0.21406208699847873,
 0.13855625964483353,0.4220239474895586,0.29241122458797475,
 0.9612720972945562,0.7931417957009848,0.0655837786057939,0.17240411604938366,
 0.47044079473129186,0.18870135175402258,0.9904003867114543,
 0.6408185059159993,0.02981763617249711,0.9207729441071605,0.7993845835174895,
 0.6407025822461602,0.03960173986403792,0.22161953333926365,
 0.015584644404377901,0.07142963850905637,0.7822176949396558,
 0.47633324774983465,0.49839341353724764,0.8063021824999319,
 0.6600566386553137,0.2769684141451798,0.2906730262743661,0.39777425185649506,
 0.8985690281308313,0.1076275199174902,0.8307664645361938,0.5338082468668508,
 0.5474401599128059,0.9457062204738848,0.8456586757131206,0.928154362856782,
 0.6618526275444357,0.08144459664384596,0.5463841524127699,
 0.02555330862417704,0.9188303837053683,0.984968160081007,0.5478334578032419,
 0.7129088724515248,0.5488376027002766,0.9135452857977627,0.681723322352822,
 0.6814498943104803,0.19499541961559141,0.8235164700580798,0.5830369650048484,
 0.19973294015520404,0.9326217963132856,0.19428788024371613,
 0.29998743564193875,0.6219425260228266,0.7614563596449894,
 0.36870587011866895,0.7066703921686643,0.12102490986445857,
 0.5434639928920919,0.4750965521352398,0.03273032590285152,0.2510952962727867,
 0.05467870116058349,0.9142752016393282,0.947090259764619,0.4092756527212036,
 0.4129444743350966,0.40003196418736897,0.9854523261767412,0.5485994411833415,
 0.5877639959455512,0.019140000518646394,0.8577245347623956,
 0.4837306723480206,0.7351354620907904,0.27336275858658865,0.8783343182054142,
 0.4232517462851991,0.7835799101859816,0.8695105514696754,0.12025761434556959,
 0.681376090082975,0.9308318643261388,0.4358658509989579,0.5269624524650397,
 0.10654792557674919,0.7541323333414041,0.9232182609083546,
 0.030852259587312303,0.6634757517713157,0.7611728501300251,
 0.21784220558472356,0.27774507728172815,0.23545092878860618,
 0.8174593836246243,0.5326052159035329,0.6762795414727476,0.9586100528301399,
 0.01346403288187381,0.7001851214566135,0.7388305209632275,0.6052022271848736,
 0.7057622040969108,0.16734672931531847,0.18051861116100731,
 0.6032642257701308,0.4501516303680182,0.6497857645395385,0.8980732940754914,
 0.876727510632404,0.5210922297738192,0.21973081341582934,0.1429762503725165,
 0.6460009763750956,0.17406158229590485,0.7526823225744723,
 0.24310596389058503,0.9838737830959489,0.23459920446236482,
 0.5471116213785773,0.4084549326019742,0.06926520232423916,0.4730881230299453,
 0.8035610212735452,0.09570880983941821,0.8152643753852871,0.3352530267684133,
 0.28653673981390604,0.36334308895507983,0.5745632070962501,
 0.4102117709797576,0.719649255857579,0.4797541035794425,0.7885657559634602,
 0.4383980596152385,0.49009212187915385,0.103645016156798,0.9314189347151245,
 0.8360555577997629,0.9918514352958865,0.2395845707332478,0.18078290912744277,
 0.32111335322336976,0.6775757425146567,0.7222003862314873,
 0.12133885633240815,0.2926120500202808,0.04326219129804787,
 0.6046167425801582,0.49298675092260713,0.8274412941063262,0.9956646471119184,
 0.7801383472851011,0.9859194154316432,0.08737260020897264,0.9418675306135468,
 0.8616371541329872,0.2911027533466789,0.6588898001631383,0.1608608847048828,
 0.504125578762182,0.028099600638843203,0.7226320547339251,0.5705563963797662,
 0.18334822762601743,0.43795263320947764,0.7192362588487257,
 0.7891975572927832,0.6605277171371937,0.33829279445074467,
 0.16488984040035792,0.8072655361525565,0.6962005911985667,0.5945972320259002,
 0.40778871793686766,0.4223228394972003,0.7637024208787484,
 0.36461400101637564,0.2797043502356855,0.364979842870228,0.53293665213332,
 0.7705098509842779,0.5121320976233111,0.7731050808212936,0.18032542231196547,
 0.696225110029808,0.7904222974691384,0.9384622642229707,0.19804177985255955,
 0.5365825739248883,0.142781574738841,0.1676326882291579,0.6712991219974521,
 0.13658667155469117,0.39270727336050126,0.12340106146683161,
 0.4323946085440795,0.926849199245353,0.6539452832727135,0.7363847358771503,
 0.8470662585445193,0.9820554215319133,0.5036019652013833,0.4356498984320428,
 0.6597999029853812,0.3496680802605272,0.6693340932536629,0.24560750503660334,
 0.9304611364441808,0.5579162253895731,0.19053326609119492,
 0.0027693147134809637,0.9014604412075843,0.7145806075014517,
 0.18749335714052895,0.8021006101687648,0.44466813373009284,
 0.8703192418734017,0.8395126411950662,0.8939161987604448,0.20344950648044346,
 0.48808344233281764,0.6277926915306332,0.582189992992883,0.8887892716571412,
 0.1564663183820839,0.33903412057696247,0.44968312969084145,
 0.8826498051080662,0.9068098795721916,0.8726726444365316,0.09073716610919935,
 0.7876460945479004,0.3082663213135154,0.9293428410552736,0.5150654063030617,
 0.4562245541940816,0.9811654883387345,0.30142209714366197,
 0.45966153159966416,0.3838428250655085,0.02701859437812726,
 0.6119866423619305,0.9353641354257461,0.8867343362179594,
 0.015339731857511807,0.34363340956533184,0.5442601668991227,
 0.6351624173685358,0.8775223985779742,0.18185132180931973,
 0.09914573967769047,0.4226421377371531,0.9388582570142047,
 0.39884609485857725,0.9735011469244875,0.0316487682482508,0.7772900214721693,
 0.09454664132836155,0.9953625489530167,0.3958601014450829,0.3921945734031992,
 0.25099806938391844,0.14056413331132245,0.025604242181921144,
 0.8367143370131199,0.9499893817213779,0.5147820222389996,0.7626268034917032,
 0.3763647316924895,0.2997593575282286,0.6079878090336025,0.15243732847703972,
 0.038506294883076,0.505690284415109,0.7406155715423921,0.37545443771466747,
 0.40101140463466,0.8036490634588466,0.7827996239335515,0.6554585372012529,
 0.7584199260369817,0.3260729640719473,0.42192628390275355,0.6142924590486772,
 0.5071811001184758,0.5735134734568285,0.719735125131031,0.7139127214096254,
 0.4122144178997673,0.9816640344106401,0.35150675845840496,0.800549524149164,
 0.24060913968834807,0.2542760532200974,0.23850308575921753,
 0.25745988197818037,0.9457544753671319,0.14173072307942236,0.736286946611267,
 0.14724348429045575,0.7879041212465131,0.8409453011314624,0.8339633039825027,
 0.8154123195996575,0.36862145904151566,0.3261650974594039,0.8501825492791669,
 0.39475411724042153,0.03232498064082445,0.3964362771301744,0.9115916982642,
 0.9230736127522843,0.9012728580074739,0.0923374989079393,0.69640562200563,
 0.3666774662694967,0.10066093664509412,0.8555846107134687,
 0.048088833721588786,0.7292814562726281,0.34448215707024676,
 0.7728807823351981,0.9088208211201563,0.40555040151498645,0.6108021587042962,
 0.2520396149771372,0.5724863650665587,0.4128409910685369,0.9955429322636871,
 0.2970952407042956,0.7155616043430076,0.8427710858924327,0.576084289253881,
 0.6284350373505357,0.19384148004978363,0.47652811327260447,
 0.1839543507109359,0.9961289230442915,0.19330444188395401,
 0.33165308685729644,0.2910748230461946,0.8369750972247525,0.8845898244713148,
 0.23750226310958877,0.7141103058177779,0.7037212213469187,0.8483977927064328,
 0.9790796427658524,0.02819717080414963,0.20333597652995272,
 0.6024886260340365,0.2823780953456031,0.29125485077090985,0.3639913921606823,
 0.17595795638948197,0.006782479746805126,0.8600245299105137,
 0.18978689201914545,0.8769530203261435,0.9765607623760552,
 0.45324358927075825,0.9237124181335782,0.7621059894813484,
 0.04685298911581415,0.7538557019153989,0.1043769740486139,
 0.24942799425018436,0.28153947881343333,0.7896593578432709,
 0.5885439834207694,0.03013327013076994,0.36948498101415483,
 0.8649693130269316,0.9412656647638253,0.4934441214735994,0.6244223201987108,
 0.3350782451000548,0.5037943481754876,0.08973728903531497,0.9595366435705643,
 0.7862484480181355,0.4510275247320473,0.6094363699229688,0.3037240102385934,
 0.03588992509981048,0.23354013493974124,0.7884568415183059,0.975793042256926,
 0.4491217792142652,0.9807571703008277,0.7272275872293663,0.9507870450644238,
 0.057629597778428465,0.15535169800188586,0.2914009748938462,
 0.693624542191815,0.4566193981417218,0.6293494770171508,0.30674366522021845,
 0.5990025561807086,0.44589865165478015,0.29195018610645973,
 0.9060422638356795,0.9069999496600051,0.12736464127421976,
 0.38255094634553566,0.24240256678242156,0.23568265663095345,
 0.6772971304709873,0.5244519135939845,0.29855090694776054,0.1368117963751574,
 0.031766649965964566,0.015581366391146156,0.8871927238479002,
 0.1753891034809194,0.5085042694978946,0.5789001981727084,0.9106622745173009,
 0.624410053983761,0.023931023216874014,0.4116315154636426,0.914276735422515,
 0.9474095918230254,0.006553735893360613,0.5609395939271059,
 0.8925620392977288,0.16072980766482192,0.04194044559237209,
 0.35205824656610496,0.37220473813484534,0.0022774002039436247,
 0.4574615379374145,0.8878760838521785,0.24600343759453602,0.7064172764001226,
 0.9364590126262234,0.15130635515899316,0.6198620315536963,0.955888723716362,
 0.3950075408410879,0.4325674035170963,0.6783253473710269,0.6712916021601032,
 0.22978933065689877,0.40437658180720004,0.9940793377947523,
 0.3020859464258223,0.8872709248481367,0.6097825012936928,0.2140678659934495,
 0.9141169923188558,0.47220039282345194,0.5750383340809847,0.8916317846606119,
 0.5677200970894042,0.4876176509588551,0.6902614672163727,0.7705576771745514,
 0.5113909717411947,0.6567429572982655,0.9854976096149971,0.7683318978839118,
 0.784759506919095,0.48233474058403447,0.4724316221045102,0.5870597433337016,
 0.9880479158623505,0.08123334605651256,0.22591412521243415,
 0.5768623671294316,0.48137431520891916,0.9087029634240763,0.5812207622202181,
 0.9696148908196575,0.13544506232015352,0.8100381535982114,0.1201245036144043,
 0.3105518243954538,0.8431308823241533,0.0649654246591036,0.943918234580018,
 0.7952911609956118,0.5113882508200303,0.8179901616509612,0.9915784266481555,
 0.7233119808964767,0.5937922159985449,0.8600435759908696,0.9199060086241294,
 0.8059650033675343,0.9994350070508232,0.957685769796876,0.5966459061568482,
 0.3987044091928512,0.09600049274715772,0.4522178434899027,
 0.015385453805611848,0.9674282504643383,0.1286868438738692,
 0.7313372341519286,0.9913592419581804,0.38880510825162395,0.9654181443918958,
 0.6566651715997396,0.891337828667373,0.4577846916330852,0.47652914755060116,
 0.11738475532802828,0.10868767771490984,0.9313600366631335,0.408819588157291,
 0.42835490795195774,0.3477683275130161,0.5563317018294154,0.7256816424888591,
 0.36833604417149657,0.8803135603918127,0.31384504666319346,
 0.8104420545844038,0.9699113860325799,0.5109941633359921,0.31196430050288726,
 0.44776520423838817,0.5301248533391453,0.6869974718223661,0.9998095812140663,
 0.4387116666249581,0.7231746494465836,0.5705637743869407,0.7656410809839878,
 0.5501862029899811,0.9958125698079916,0.6876622344689493,0.14761368023313626,
 0.5403957655709353,0.25071546152945334,0.35102022306306935,
 0.4607974212934677,0.17681725395367764,0.5337873232565103,
 0.24512481833021227,0.40332067781521197,0.6316447265162851,
 0.1107810843113366,0.6983088746699854,0.054774697827836616,0.832841530907142,
 0.20833014610038902,0.34998977670505493,0.8713500921220105,
 0.07227689799819603,0.8966342693564706,0.1596572491771583,
 0.03688667188833539,0.4199079973610397,0.3114109704431268,0.9951854635015378,
 0.34404926043723527,0.3489701764670523,0.1845644715849618,
 0.04093501140736666,0.7718532032731736,0.27214136201539774,
 0.16062769868733096,0.2650111515434268,0.5071746390360725,
 0.20018762546841207,0.8189940620205964,0.4773864394541706,0.6075587896568853,
 0.9877753585259119,0.009166496113137912,0.7321811106456604,
 0.05977117760017525,0.06509234755362969,0.28413299376461887,
 0.39101340406178997,0.6010444614018202,0.6019011914623715,0.3323306819972375,
 0.1398130954174749,0.8703148766573567,0.17528737915324044,
 0.35341010519327465,0.62036477220389,0.12473629468228731,0.3239424288902818,
 0.6535835583300689,0.11457769588944977,0.012222129584176344,
 0.05864603788198497,0.3251638003666487,0.7458989563221347,
 0.20033147884759084,0.18593242249534925,0.20506509555950803,
 0.8576206931023415,0.8522426971638857,0.3480423304617728,0.142893026391808,
 0.37356432565581343,0.6053173005118755,0.007657663335379095,
 0.1828868776112058,0.8102296900344481,0.27952125064608435,0.3461915162662147,
 0.5805132835664342,0.11592140227035552,0.04558938810118196,0.876241385101752,
 0.44544208656461104,0.41427786421432566,0.9310780473949369,
 0.3173691906258027,0.5707308743439777,0.7223038669183798,0.5333017658632063,
 0.46099100691706296,0.8772747808597746,0.8347623337716819,0.8013636191044198,
 0.8548679038191377,0.4760904011278094,0.3786783150273464,0.3749324005283403,
 0.8538466930467342,0.8449675849204625,0.745022852472557,0.6588154443221718,
 0.8623904711369575,0.331793306158237,0.9906469508211017,0.14981529756008605,
 0.5585072209462736,0.6903753336737735,0.5470782483686571,0.00749220496616998,
 0.6957920777299558,0.7124700100281744,0.03019630199846568,0.3153216231352579,
 0.2896944339434365,0.9722945618753033,0.27661926296649464,
 0.06966858556653321,0.9289496294576874,0.678865520875343,0.6711307935534824,
 0.17831777924631398,0.9352794494648908,0.302573377314569,0.7628108004718738,
 0.6965128687759319,0.5380245320892247,0.8987258465937609,0.9242358665927289,
 0.9697350570719815,0.3916722807839076,0.4723598874867999,0.42665524258224063,
 0.25927758795143907,0.9359009039588235,0.005725513208187483,
 0.28697879589196873,0.11421344137149136,0.35078540768970734,
 0.8078996409578553,0.048135656336340205,0.9828445919613322,
 0.16715385932794047,0.36278545866014034,0.11069380333860268,
 0.7236746156977373,0.25776416522271395,0.10373621027763402,
 0.6193091878632808,0.6834904791715285,0.768632354239365,0.5960206136480793,
 0.18776458390399553,0.4928634669130446,0.6802796903184868,
 0.20293339178996117,0.929699454761094,0.4346407250069617,0.395508539268133,
 0.6458623868316825,0.7136123309849123,0.43168926153697484,0.5533998310626431,
 0.5185194906814049,0.5929905865464802,0.2270074896930342,0.7788708578798444,
 0.005589259845626415,0.3835126060735883,0.02317730347218494,
 0.6266099883062476,0.7507155601213044,0.27286973725239694,0.2936143486301821,
 0.0363774651256652,0.5639911243198283,0.46446367016216683,0.3918998545344484,
 0.168267385283855,0.36096637427865585,0.24159527375799927,
 0.04067777637971348,0.4227896472503322,0.7666008519252417,0.9259856328891551,
 0.661051260793222,0.5636505122030143,0.8195259215165233,0.34470400253533295,
 0.7470868845461679,0.16777967050410925,0.8314826483882013,0.0347847456124758,
 0.6988692602052973,0.6052255800401624,0.6554746238888529,0.4052794793050628,
 0.2892749177637358,0.24213503766782374,0.8689495560506318,0.3734766032559774,
 0.21694071607483278,0.49190905211280733,0.9070182626391952,
 0.7840032645973202,0.7992803216733234,0.3914283817075731,0.04598574608003125,
 0.0034305412640776023,0.03591725084857045,0.8374425400896875,
 0.8737431896028588,0.5193262502812159,0.3769499892236249,0.590488930368683,
 0.6409875247626263,0.0796016975259175,0.6811296266066548,0.38743085062929417,
 0.7944138760249888,0.5079553014025182,0.7154563354121849,0.5140724616498181,
 0.6519126433047847,0.3388895620504475,0.2520269445205313,0.6962509893440435,
 0.5411966785319571,0.5537287626298678,0.3595417366887619,0.7776183868257439,
 0.7137306642185819,0.8507644448107252,0.21057468625668707,0.3487084222226058,
 0.3969308833354225,0.7432263047733921,0.3616704699696939,0.3374918186862417,
 0.7267233131449211,0.6445366720039063,0.33533229578836743,
 0.39036387009383655,0.6198058445583523,0.3723218026755952,
 0.21318449418126617,0.1500577677926126,0.06353734525652821,
 0.8409992407383343,0.07553631506283609,0.6245104797409304,0.8370991123587241,
 0.33383154918373315,0.47180995805441006,0.39494338788943517,
 0.2284516151692395,0.9367008748064771,0.3477353357774464,0.4878165695138261,
 0.7495408248044423,0.20982205430031797,0.37511498946277055,
 0.18360509609301268,0.9484916521379196,0.026065694312048415,
 0.34867560449715396,0.7168591575264318,0.03494288590894523,0.581439147427401,
 0.8655385477340394,0.07910628633531325,0.9254581354455784,
 0.06174690294806151,0.8531262924309185,0.9143355362216276,0.5457865003153447,
 0.3009427149631676,0.27631304772059184,0.08822641817473886,
 0.6570794429213102,6.867608312756124e-4,0.33099444335141726,
 0.052576665650062915,0.0927215180666634,0.3428886324294982,
 0.7431652197479423,0.41937446425462077,0.9358455905951308,
 0.24661279275946058,0.6156990136460729,0.9381866705614014,
 0.25912113885560295,0.4615983003351666,0.02072354292945744,
 0.09326015247747133,0.32917018079190496,0.6207308646294135,
 0.07192083100231994,0.6271872068048545,0.20382676100832875,
 0.3390296484316955,0.3170730850842636,0.8257453843555291,0.5167469634969868,
 0.6084061106753156,0.049606881486501386,0.055851381527380894,
 0.7957655315910637,0.00700584018066297,0.14577642787847145,
 0.056263882154110934,0.15746021538434407,0.4690182377413681,
 0.017795014722490565,0.1725041182646132,0.6279898455534181,
 0.18212888099652247,0.10066811719144764,0.798110594356596,0.9224336377028113,
 0.3380997097135643,0.6770297659316284,0.29170163052375353,
 0.045371889909256446,0.4782521427599423,0.03485522286588483,
 0.48902086390865995,0.4753967486994589,0.06565866894203032,
 0.7678195018351178,0.6295787247731833,0.8997543427564471,0.358722311102316,
 0.526096261648093,0.3536720626929636,0.4332960382884967,0.4013218048916647,
 0.980147442215131,0.36970305991075203,0.5216139654754551,0.11924295596408419,
 0.5326410167032503,0.4952804235466498,0.8016914254121885,0.5604622210267229,
 0.6029026534473101,0.49117142809602543,0.6264396555625643,0.1929346078771359,
 0.383698504758898,0.41787227943762906,0.9624990494269625,0.95998643721339,
 0.3813338642401921,0.45730176413603874,0.5167150335470725,
 0.038957334613264116,0.7669450536685115,0.2986885153568084,
 0.8859458549757179,0.5846297192902271,0.5417786041657922,0.8796785495205526,
 0.7418362106793507,0.15794188945928078,0.13553373100502208,
 0.1171931601579268,0.7820174551295889,0.5605388521563298,0.34199706794154694,
 0.5746474095724634,0.2556103072542357,0.056197973008383695,
 0.8141606785589202,0.2071616112385215,0.6155188725271092,0.5718731898897005,
 0.38730270461162386,0.31580226251170573,0.1233310721983294,
 0.2493111405621865,0.12550514266213375,0.15060432216179787,
 0.6674551416914768,0.14346788667701083,0.9425406970158176,
 0.015119391135608784,0.6556109419242895,0.8969879756215873,
 0.47438313575999647,0.3640702934443677,0.9255737804023154,0.8150968037003666,
 0.68566171825024,0.5182057772480934,0.9027582385086669,0.9775839585038376,
 0.10698662387104607,0.29241436726110714,0.4791073558016393,
 0.04376004477121331,0.8380499021360615,0.044568589293902616,
 0.5875900237323285,0.728282334645189,0.18693649590408423,0.41702091207156455,
 0.11359902139436595,0.16427146307480478,0.9374130798292923,
 0.6504574851293806,0.5442064062213942,0.4848022735303633,0.2246273476585554,
 0.27187047674717146,0.35380128487635687,0.2242790375655892,
 0.7445379387051482,0.06218605505279173,0.5714361521054949,0.7463281544212536,
 0.7327104560186755,0.13893146859400596,0.5862393525925823,0.4915041701837607,
 0.32361767900787375,0.6738817946621567,0.21475543576714673,
 0.2686262113068165,0.769752673443546,0.18226641709118052,0.8433894875791792,
 0.005292136980400541,0.369959120450855,0.9236537527174519,0.7375799138256532,
 0.12162017655163337,0.816685324320223,0.9255060663841442,0.4172525562513716,
 0.6080133798908076,0.2084765279979266,0.13912007650517055,0.9575780805701966,
 0.20301499372017384,0.9680093447810455,0.06327446981194218,
 0.20632471012608067,0.9374163978890995,0.8675037152544658,
 0.09820632629758452,0.5595344197879496,0.9376250714607572,
 0.09141869868133579,0.6619598568879557,0.09423593618348824,
 0.6262669734699403,0.2650766737114658,0.7678720814034717,0.43606563617446303,
 0.6656832084007956,0.5140677190783589,0.3078454040185088,0.5989428373880432,
 0.006999940562358531,0.025625674229093454,0.9002484813427317,
 0.08213741520694184,0.6575797039796119,0.9420191759170434,
 0.12420926498479712,0.4025248199785737,0.022707845870101773,
 0.813650227367965,0.8344582442003629,0.19017408858270135,0.46121179394644707,
 0.6663461576851674,0.42615026391303434,0.6123039591446987,
 0.15551981891750988,0.3822382426598453,0.02360107065551853,0.434862020824511,
 0.30824696743761804,0.6740962475605357,0.44452469868635724,
 0.9947206333461062,0.5163847343523256,0.7075112411221083,0.44215726763650876,
 0.402723507328442,0.7749533688322306,0.9273198215937337,0.14789096996992757,
 0.6112154139160682,0.20068625430822007,0.536081589420991,0.5681806481259588,
 0.7717321531934058,0.5750438473807591,0.2358592784144966,0.92222532446057,
 0.18828598793699503,0.38659202114029967,0.7956087584328535,
 0.4753389748955642,0.3757492058734869,0.30228255942508353,
 0.19544742359331058,0.5560246631016703,0.09477637688166063,
 0.2675354131969969,0.9006469792742958,0.9000182608521609,0.16440370399464732,
 0.4197489769141953,0.3299398308433551,0.8180742016560629,0.23764714159288547,
 0.8705088751370317,0.19303652058915355,0.4511602903100359,0.6947854154676733,
 0.273889732751186,0.2824104049403532,0.5685970853046233,0.9593522152611782,
 0.2373473430910914,0.9153752429279202,0.38285236999971106,0.4231610097225841,
 0.2837812831926573,0.30789695422202557,0.980730825444177,0.5224081092953599,
 0.9134596427411825,0.9267950610579498,0.1027253163253522,0.9931444803192179,
 0.9716169637141395,0.8365975179579569,0.3389666930506776,0.8055641779570576,
 0.1209712249331425,0.7786583776455998,0.3249168422526155,0.5425129358555997,
 0.5790239437044726,0.1704763502881479,0.22143913009441474,
 0.06409888498382077,0.24894019705489234,0.9785083824410598,
 0.9453131738548711,0.86126637240943,0.8020349609104954,0.6830460326902313,
 8.815396697773314e-4,0.40880408836247883,0.7424091007218067,
 0.8781873280956657,0.35922620985432396,0.15063126213256084,
 0.5959775388458537,0.430699524176938,0.9690339098879338,0.26347503194981237,
 0.1761527004534802,0.8299201117399218,0.9572157804454896,0.6429499399190262,
 0.14837907774652193,0.20729585216937574,0.2400303397804031,
 0.5455360234084384,0.9934070898451931,0.9275021876103092,0.36592744401182686,
 0.9007563801729643,0.15410505594828305,0.4432949351818811,0.8388425224894069,
 0.5623802800727762,0.19263625167160592,0.05915577302366615,
 0.6498922950493293,0.5724403159156712,0.7896730308778053,0.7070489826649503,
 0.2679085188173954,0.9885647468210121,0.9077326871559455,0.7921978365604536,
 0.8552184321968247,0.93734487842971,0.20254095123653404,0.05158497486464253,
 0.920180541743316,0.97672633457151,0.713351804061499,0.4942685556647841,
 0.7088978029830837,0.4864089176350105,0.3131939365794446,0.20034050713241125,
 0.4809127465009756,0.8477454969850068,0.2844575578597611,0.8692641290987247,
 0.14957322240216286,0.3423126614435641,0.8990709852110431,0.910942815505054,
 0.03118923189026268,0.8408544413675481,0.6951303703183598,
 0.08589671012863043,0.1887823466949854,0.9169589757769461,0.8343881270721851,
 0.8463725778108273,0.09697790704351794,0.5939166530996648,
 0.12819755243421693,0.3822759677225872,0.3769078094109237,0.8881944492507587,
 0.5739078742160355,0.8947483582652588,0.30451479244056057,0.499035214573892,
 0.3094834680603473,0.6173453271278931,0.2779165310412479,0.6964284857607224,
 0.2348355854233487,0.07726908747706951,0.2230726419571053,0.5257008752554412,
 0.3760942708403492,0.28255975353793794,0.7096562122138588,0.9699552571566963,
 0.060892661948336535,0.9002903298491387,0.09178848822884034,
 0.5800139271899782,0.566000681484661,0.051446616426892744,0.8154093081701763,
 0.13779632198857505,0.12966953856568852,0.5223989025018851,
 0.5300636986369004,0.1479953370439101,0.5909910137424345,0.5869291128586599,
 0.5841579823872054,0.733672235988435,0.25379773984237486,0.10607038812587011,
 0.12404099905100074,0.9574136903130483,0.6873057144142535,
 0.008256925053972686,0.7438784391447841,0.012467885861120354,
 0.14451670597007837,0.016806171545970194,0.3058428155522882,
 0.3206342695349964,0.08017953457342908,0.30657555078484144,
 0.6879757995427095,0.20938156895852722,0.05954015678801228,
 0.6049346814208217,0.9750660231560201,0.9001983995989471,0.16819598738820207,
 0.2691678138129625,0.9701173437445051,0.7746129465144719,0.18533140811877014,
 0.48886456006828005,0.8542524013331486,0.6726595154241516,
 0.019245594977673974,0.9124283108889714,0.45375729627483385,
 0.5008844324798407,0.5617783761509034,0.829473906919932,0.6419103205321486,
 0.5026437488701783,0.5049081391020077,0.9804041048554725,0.8742305056431924,
 0.4446855206795257,0.07909542209072451,0.6299284193686343,
 0.44546099852906507,0.8730713789151257,0.8248326807632969,0.7801487540385682,
 0.7731847649617895,0.15941656481084487,0.5334630141936253,0.6337765278466702,
 0.42401993061680976,0.08369417668160073,0.2288492653035329,
 0.7826222739672836,0.2045768605133036,0.757335425920511,0.1207580921565834,
 0.379462851979702,0.2673348143633545,0.06799802411590117,
 0.059888104365579786,0.9268395889866612,0.14669169943372362,
 0.6259099229323736,0.2855388644391058,0.12231161272139357,
 0.03636002556851459,0.8539610312697954,0.7639554302170652,
 0.44167088532846577,0.08218612050659047,0.13130254058336188,
 0.3669816509862611,0.7639703524513667,0.31015984566555854,
 0.03244816947410967,0.19686103919191789,0.22235968282842367,
 0.39063176920741904,0.9844904131430775,0.3840387039816223,0.4794495341534053,
 0.05761401441846503,0.26763827656139305,0.5632300975004281,
 0.22041395970719113,0.6650892865239579,0.34930578721064076,
 0.06898205706617921,0.32271007060801793,0.28475458360575967,
 0.08409314801556356,0.056192452437443086,0.7783831085478324,
 0.29440370970819196,0.38516544475588144,0.15541250291324193,
 0.441179586096632,0.21732313196037967,0.161758037618142,0.2355230663676975,
 0.5112279717069296,0.5507962518674223,0.4047518066181899,0.9734684595696294,
 0.7306688022324637,0.5464503733672383,0.8231606831259511,0.5912783140212222,
 0.4114347718460297,0.7684753444624409,0.3984587991389482,0.8252954307475604,
 0.9641060176127487,0.4591294772596637,0.1500983658750793,0.2370571050207697,
 0.586029304250714,0.06400477758847123,0.43119871553549105,
 0.19336911997946027,0.4436597776564364,0.7813277459517081,0.5297225803804037,
 0.41896335530304374,0.16782530112798977,0.23017189991990117,
 0.12453008119475895,0.8885153058425934,0.8174657098310031,
 0.13945443021562287,0.13500375081409954,0.3577851985485456,
 0.12213688721905225,0.9491290753307534,0.3355023233549407,0.8379497367295803,
 0.3173189655917291,0.2832148280473453,0.3035799685891093,0.13102153970840424,
 0.13082194690115512,0.625014645850862,0.8817233457358986,0.95756810369315,
 0.14139707649915512,0.4540706027397762,0.3964923497715903,
 0.47670779659348217,0.2651564144418409,0.7972540326842533,0.8140133389064884,
 0.3962079421397944,0.9781012103094123,0.8352525412487085,0.8827710204270982,
 0.8128677805213449,0.45756814174324734,0.1684628428868471,
 0.43563933020667023,0.14935284380525127,0.463142498896454,0.580445927163708,
 0.8216175036150475,0.777933739158873,0.18294066665761088,0.4081105362346479,
 0.5801993184315991,0.6648326680435211,0.8099151607892835,0.8500757593043081,
 0.5991798082503199,0.2713038521644867,0.9190175511199591,0.4057310526876875,
 0.8180659177944418,0.29535965257234764,0.45338164960910254,0.776871596686342,
 0.43479778555687687,0.6129688122961481,0.4925256708441835,0.6527573562949618,
 0.14713534315293753,0.02959765796668279,0.9875824285071492,0.624602088048854,
 0.6863693701625955,0.8392944314070993,0.8454683128636666,0.4229591067790892,
 0.707021688012405,0.4148733962388672,0.4900031417449425,0.9024300366934894,
 0.06478388139376023,0.11034505824524932,0.24502729446293858,
 0.20188496505830666,0.5911807462508244,0.24090528415493428,
 0.0603255336478864,0.09043530594750149,0.018260392271265746,
 0.07316884697930925,0.6521763264935798,0.40322859791081855,
 0.27071766302428923,0.6583256328762017,0.20255924400629854,
 0.9374780649586505,0.13995965936317556,0.09281810671985846,
 0.04973423201074134,0.3140839669282024,0.17537934275981826,
 0.26334043079589275,0.19340331100867236,0.6797831418959581,
 0.14454154490674898,0.6953385907212088,0.4757755874494852,0.9320930891387623,
 0.45103434848382284,0.7945157206320198,0.5778668296442394,0.8577185041582489,
 0.0028404604207827866,0.726823435127461,0.47853690314736597,
 0.9053253210409113,0.766171554937207,0.6605649088530868,0.6604448106955163,
 0.2536611982269692,0.8086398606792272,0.5352138483421682,0.3385089787675929,
 0.5578786374642585,0.6552515238284156,0.13847912068778923,0.7519811984292621,
 0.12786058737160988,0.25356053445258797,0.7158588180127543,
 0.8841531172402923,0.5382961332170452,0.6951223808806063,0.4888273769173158,
 0.8192631453639924,0.2702002029221593,0.12876932786392947,0.547162272959528,
 0.6071462884347334,0.3470662257428012,0.8142011072644287,0.5581111281102169,
 0.6606892758117127,0.4726906169105831,0.7766335135275945,0.1053517707520365,
 0.3306721797080677,0.47723858605403213,0.8619772315394224,0.1775936082179399,
 0.40997506980051657,0.411367397009865,0.9229091246272505,0.564162774199428,
 0.7461161991638816,0.6151611062245261,0.015016805711427161,
 0.5765046782748926,0.6578872155773124,0.6893261644664226,0.4637415217011214,
 0.3127660083916558,0.33772868239324905,0.24621714141580453,
 0.3382713320325036,0.8460799707905744,0.17408558542420616,0.5305483116570006,
 0.5075424967934123,0.5599231711397125,0.5114355150617671,0.37107708274981566,
 0.8580656274764027,0.20041641691382361,0.28719131272011844,
 0.13456432799722773,0.1626516684490018,0.46787377896628346,
 0.45131903328410994,0.6330709962474053,0.2868393903882196,
 0.14029798231404667,0.21255327772552857,0.7059019765061979,0.997639316581283,
 0.23303330767794694,0.48415642271721415,0.7434116302749572,
 0.36369717473910457,0.016234528503570322,0.3798564610892301,
 0.2775884195002615,0.638899979287132,0.262867039204276,0.05169718127123524,
 0.37271664360727796,0.10127889772256349,0.7625938813995754,
 0.3095140908739712,0.884983717897074,0.5354710565695902,0.4651929375584889,
 0.8750800755134903,0.718373827508262,0.4441212797797578,0.9852765767105331,
 0.6166425525307698,0.9809964894878936,0.6583777829012525,0.18565136268977245,
 0.26914574282803094,0.7292156477985533,0.7377263136215036,
 0.08860004331389149,0.8749415872238929,0.7970697738109408,
 0.11445416539859932,0.9048499601095212,0.5959844003165539,0.9749775480369305,
 0.8370006352441138,0.7320588369481464,0.7737778079861313,0.31310349164774776,
 0.44391588151596606,0.6205439950863172,0.6586624984951903,
 0.10619847498704471,0.48404321942209816,0.21174391460340558,
 0.2893172569735726,0.24570192442452854,0.8606421011171937,0.073872360329026,
 0.5803323034562865,0.1113661934071003,0.8167462997167271,0.8002365444169655,
 0.5216257893327548,0.13739482130522984,0.36908458203783645,
 0.14523367784173802,0.7742963037921244,0.5846663942440474,0.365499518732067,
 0.7734368259964669,0.5509466919040932,0.5815479044768859,0.5163840514480523,
 0.8438363808050493,0.9756517157491208,0.8558549594892176,0.862247104798197,
 0.999931387498655,0.7786984016937967,0.025782452198691352,0.3594239566496775,
 0.7093994395417671,0.8893668588704008,0.8851567364111703,0.6042852560047987,
 0.6806447026110296,0.23559681072162497,0.5614243122713528,0.5172633098199949,
 0.12279917735207468,0.882160559788565,0.10255068037495019,0.3600447112630494,
 0.09703606481691218,0.26870165609527963,0.5912659082297879,
 0.9510690832562048,0.5119705115903597,0.9317034385427816,0.5418868617352975,
 0.13737221983332826,0.4750039259568585,0.31241259495120566,
 0.9761523640654817,0.8291160884036336,0.8901386894440243,0.5431831022679917,
 0.47475439498448935,0.5706689171431063,0.13749050210612956,
 0.8829767638579816,0.6163866618771303,0.9805764611315878,0.17901646026963147,
 0.3684768082795412,0.9358522492227858,0.3706356318419892,0.1968901948102566,
 0.8063636450308421,0.12587187631113472,0.6196792481945267,0.9471300981916054,
 0.5207905113129045,0.8378668527757206,0.3856979768346187,0.07406845216746938,
 0.3981218020715692,0.20372992233155318,0.8422456460668788,0.6214141180065511,
 0.32434655454173944,0.05633321078720788,0.5743244281541533,
 0.6358882524717543,0.4510364489177925,0.21151270031166247,0.4912186625906064,
 0.8353992569977781,0.745510001646184,0.013984730630100084,0.3227331043165057,
 0.9620539315970776,0.35071554514272296,0.9087858016402951,0.1953173395843525,
 0.6425409813196994,0.8751604174108146,0.15061671360513784,
 0.15171451668832048,0.1060009895112477,0.9285271032276574,
 0.20722507087737796,0.08412070546809458,0.5615966986410397,
 0.8913706876331565,0.6638012641568769,0.7220931034977816,0.100567491929198,
 0.45371616268412884,0.05773034270571287,0.8807663386003162,
 0.46567751666509305,0.8767386552734585,0.7587606254175541,0.3485367045503579,
 0.3231687339591115,0.35722514765139113,0.0774159019003755,
 0.04099077735487633,0.06599500626425203,0.7035496938636273,
 0.003373492195286043,0.49662157636594284,0.3552944360534297,
 0.9049138654615917,0.28873495948856054,0.19695230885626513,
 0.9976856021878088,0.23970435380017385,0.6153527964971802,0.8373892866008972,
 0.18847746950749522,0.5416254063042095,0.471028385953551,0.8573943480116071,
 0.3117989868196278,0.08986734252197792,0.7602211127909313,0.5651814077283454,
 0.1690251966743438,0.884577351736428,0.14462539509384997,0.22460133719209674,
 0.5103314027751783,0.7247111811139608,0.567573227309883,0.2978934206543027,
 0.46787757392525153,0.39189931870746275,0.15173738651542412,
 0.23383780465398096,0.9778170829943239,0.8234685358941307,0.7645055834295535,
 0.462739153809851,0.18871157687029505,0.47135749563310725,0.5174818793948627,
 0.2591043875931338,0.39545420400955844,0.47378706557782446,
 0.45935432360275874,0.1009820014088314,0.8531123588365029,0.5674553339298107,
 0.7001214985159554,0.5507952116553415,0.061273394019702554,
 0.6778328667894242,0.1282232381106796,0.6502794158377408,0.4703400508144818,
 0.3654899191529746,0.5489810235224453,0.8107641981721945,0.6684225708023601,
 0.8526584005801304,0.7966469551459396,0.6933710082486735,0.20409216521302254,
 0.1546752887915862,0.4768769268458586,0.5182577328975554,0.7820572753576578,
 0.418302366734579,0.6550710549519184,0.7936106148687123,0.3051970416879045,
 0.055889738656381804,0.339152110046662,0.07141641901316453,
 0.5487109670571367,0.9909086147676056,0.7667922932109025,0.3700503171737578,
 0.11205708762766953,0.23823903880150776,0.5430937380491696,
 0.9613771379917078,0.7186137827678145,0.0960256320509002,0.5721784062070733,
 0.049336536129362396,0.4734634621474455,0.07003422744192567,
 0.9314739020768522,0.8288997414185745,0.8806360303229235,0.24194184318060785,
 0.7109060956677917,0.45005058291557454,0.8586071185790716,
 0.34792690764838663,0.14774419864661026,0.9891481263284462,
 0.1430488269128911,0.8249675324840301,0.7015424509753289,0.18234199059630996,
 0.19131235333621022,0.8737005141899106,0.4098252440073792,
 0.13039714079261078,0.900595707295188,0.31774855629253895,
 0.07427282040224459,0.39321620970093185,0.2214167061937211,
 0.8465094148608933,0.2334693268436303,0.6818241455888888,0.9781421884824992,
 0.28842053694478265,0.1132719748861768,0.1368800675015852,0.4302260158824167,
 0.5705421554235737,0.40828299827561754,0.5960406154712506,0.5686232122305647,
 0.5969803917556201,0.1634160085734544,0.10490843670595529,0.6599139940463131,
 0.6680334636179643,0.9425625161888068,0.5759115861030344,0.3815717207854734,
 0.8262653336947375,0.11634566082029174,0.2795676466384148,
 0.22439343541458723,0.047237067126798205,0.169661161606127,
 0.29066045351242276,0.8614494283755123,0.4238085658549756,0.9278706666934828,
 0.4287616587322066,0.8327671057886548,0.03394656810143859,
 0.45384980464571245,0.3637924795456935,0.6987959625571352,0.7270462832891955,
 0.5039020250595525,0.2567134904443402,0.8154172720402117,0.21292944500731614,
 0.783405670909215,0.026238728247553023,0.6092889818270464,0.7629602959391172,
 0.6575822239375566,0.9623708387206862,0.4441144888351667,0.7776238126811506,
 0.5395979422877594,0.5717071752264357,0.1592905152787195,0.6795452662129294,
 0.8207960108083393,0.05025966797909298,0.056592820259182774,
 0.9215505003939093,0.3462277661635772,0.6190507259036266,0.40931738444523014,
 0.7015312554119382,0.7773527516585214,0.03356433552706728,0.6697009595927524,
 0.39003908583690394,0.24793106741116977,0.6996251311237183,
 0.008860698424686397,0.21585433421093325,0.8101626231336688,
 0.3348251906501464,0.5685897430197002,0.7819594260500746,0.3098274722680069,
 0.31284892493777305,0.4715275657533764,0.17083733424585046,
 0.9142621980309839,0.444618985447358,0.9943835252600797,0.9994420178695455,
 0.6154669098330807,0.9069971798461575,0.5968445248060558,0.5226279478858216,
 0.6249594235210322,0.4211561651619937,0.7059407409210238,0.7676512394520716,
 0.3521679738545258,0.9582251938130135,0.06238573089539967,
 0.15426436676335065,0.5842720393521779,0.48584032651062636,
 0.8848097679713068,0.8028712022315372,0.6352284123272297,0.6133081405186923,
 0.7706237938798185,0.15027578379830053,0.6180170494222844,0.9470312819786083,
 0.27658139614966604,0.37718016121969977,0.826624713210995,
 0.20336790461682241,0.8559377010059226,0.605027590268124,
 0.025861196781568285,0.11462891088578087,0.25372764518678426,
 0.5556689193851971,0.03371724161579903,0.2717759972474765,
 0.41391754226665667,0.8871504611944987,0.5046640511204701,
 0.0061027746129570515,0.9514079215167709,0.7279636284675678,
 0.6289086303213484,0.17245133128283696,0.08065823875740652,
 0.5034430711817077,0.25398026397150686,0.581020806130127,0.24970320036525928,
 0.05833170379582764,0.14696770708908413,0.3992595233811065,0.390160088056143,
 0.3093603002454124,0.8642838632249578,0.617650800258134,0.8490246816347714,
 0.9151825571644974,0.04260044759432424,0.6196791716788157,0.2224924470753128,
 0.8699681133672628,0.4158356015696737,0.11120161229652237,
 0.11718089892727102,0.24891869790323928,0.08724206164390136,
 0.09950194794991574,0.6719501812380473,0.7642822204890162,
 0.31628710571792873,0.41805972289908533,0.30975127589710494,
 0.4822719504954286,0.4941312476032542,0.6911294040317423,0.48138526710279905,
 0.3152350595390293,0.7337019595535126,0.15770584125248588,0.3153870504293953,
 0.6490942903889135,0.7500544434819394,0.5647844146886702,0.06722112391536217,
 0.3409510882893678,0.5648486280711089,0.23582935069310262,0.7223315370763265,
 0.67469265107726,0.7862040216998454,0.7123888906807827,0.802930391328594,
 0.7697597585006657,0.4709374675988234,0.591187630240525,0.58604377020788,
 0.023851275419781492,0.7531289096372273,0.4334743860277248,0.820601729544049,
 0.3399435157509798,0.5577240358702685,0.5498227267759518,0.21747614056764708,
 0.9658102243141384,0.3423309395587024,0.2742954540003504,0.5920583478623592,
 0.9746804258046936,0.4940018032327933,0.6074366408105143,0.31836684260082904,
 0.21173176615525346,0.286886941937726,0.6274528859699915,0.23259904543362542,
 0.6364097047591634,0.5774035852223519,0.13826052063562022,0.6604123829809838,
 0.725435388894547,0.9271383191118673,0.46566308667407097,0.761416623299487,
 0.08920391687948093,0.350702979241444,0.12658135450355956,0.6868133363397306,
 0.4101196193065302,0.7066234448871178,0.05785483746603959,0.7780258638574123,
 0.6184226544425733,0.21175341161205663,0.7805013682969238,0.8578491224322463,
 0.47947566413556475,0.9770491920522423,0.4638342607608169,0.5470865490723054,
 0.7478165618003308,0.13533073111467298,0.6331579966474945,
 0.33227373135973903,0.866799641386885,0.2751861366783419,0.16025614022652435,
 0.9611986460904067,0.20793609178973915,0.3637991552181702,
 0.04752620869555879,0.6908354906862291,0.8583303932293268,0.5098265265681919,
 0.9006230406612475,0.9598669010318808,0.009791661780374272,
 0.7763177251675649,0.7831668030078688,0.4292693333261758,
 0.050335336976088274,0.3445980060759598,0.41352097666866827,
 0.03375117820104023,0.2162381037056449,0.249443881337304,0.8430134629686254,
 0.1226206885319806,0.30969785548801687,0.266250216267614,0.8245110288228639,
 0.1800130836864533,0.8078365348057868,0.6492170031912516,0.3203063956033231,
 0.26745116334644436,0.9446175153341259,0.791835297340686,0.9649660383114967,
 0.9351968109013078,0.7558415355885159,0.2564032142065539,0.26055154242413425,
 0.3724879335168252,0.898498087255176,0.4913519270096617,0.1509188176384877,
 0.5318234799456846,0.9141222669267098,0.3881154276407469,0.8818147740120823,
 0.9659100628885056,0.4469302982882196,0.676603271430519,0.8713745714895087,
 0.7776198667665433,0.0649721600841584,0.41960620357804546,0.9026124283554235,
 0.3875930216978636,0.7601156936131871,0.6305872467628234,
 0.014275826940904679,0.7863472757332102,0.6000749815984973,
 0.1992907036791043,0.8466086978022664,0.250292352295308,0.43424717937434487,
 0.8201399898101139,0.969667886843225,0.19155004304363532,0.6680569773764896,
 0.42220115930981583,0.5611239771936336,0.8133686802051862,0.6413844369847355,
 0.6999590886211768,0.9359557754343292,0.969480517468769,0.6298677381212662,
 0.29699291023406504,0.23280966848398577,0.9161700548016892,
 0.5546594381988196,0.6334297897816998,0.6212177049393074,0.992821953412933,
 0.5608319340644523,0.08684933501310843,0.7267424245572565,0.3947912271573004,
 0.9894419158639571,0.9929796184714774,0.7661761207902575,0.6489797171753001,
 0.5666922825164771,0.07325537666826698,0.5867970197757857,0.194409565665416,
 0.5207628539059521,0.11494164158086395,0.6758978710329869,
 0.42390854417053203,0.36245651400350287,0.5609081024322379,
 0.18381994720296047,7.272611220888736e-4,0.06169675111420592,
 0.004541148654109595,0.22271440311978163,0.2353837961207943,
 0.5342945898562286,0.13563936410389754,0.5968953750911226,
 0.05550431464764638,0.2613961262746338,0.3573944313938583,
 0.044563964030567615,0.250902177091916,0.5839320375999768,0.6540903085050808,
 0.36085368971168463,0.4093521562495137,0.6844280063630956,0.8576859093847368,
 0.29760945448918297,0.05471709589886564,0.9164980325753823,
 0.7845462115603512,0.7220137991830103,0.6551204620323487,0.790998910472932,
 0.3252684512747315,0.4441781983095313,0.8405376238968686,0.9320456670885038,
 0.6233011341061986,0.5171705037109923,0.19269947619972583,0.2985113050541599,
 0.27430054949423066,0.18089622622177637,0.09140769774639634,
 0.06580568927871311,0.6358117505204413,0.1928573553992653,
 0.38000792243148884,0.10741327027316228,0.6398098540216269,
 0.7014134017611354,0.695614947041607,0.9667203089134392,0.9154880635577718,
 0.9601009415462148,0.6543505731255754,0.1378538503601835,0.9459204581725621,
 0.5133249817248683,0.8842113747314042,0.017112647668033,0.9887996033025984,
 0.561431944001829,0.9616082236038194,0.7712609137878826,0.4442784897215697,
 0.9819428890511728,0.9769506693551464,0.9588257109736946,0.2544554286800401,
 0.2537409177958101,0.1983516073153001,0.8443159566472265,0.11330864719028577,
 0.15464162275394888,0.5405614624247647,0.05193403638335392,
 0.8585735594847174,0.3100538263412981,0.39015033632333296,0.4808715388107919,
 0.10619791518926602,0.08668966136570255,0.510535690092604,0.5601627296713044,
 0.8847314794672303,0.8853352133867496,0.3376870458666359,0.6103054169934508,
 0.39299210695940057,0.3941539007505488,0.036454886323906255,
 0.26923555668796306,0.6090879242623289,0.7178900732053605,
 0.21335355498007158,0.5221042589990859,0.4948126389886963,
 0.22912875786269682,0.7817996557833997,0.04117352470061275,
 0.9160144369152387,0.425580953914121,0.00937567150401697,0.8422788587379788,
 0.9446903503043157,0.4889522271096066,0.18241106724120937,
 0.06010589392081034,0.19695827475411232,0.7011048490490183,
 0.3215348974709412,0.8343156523888526,0.8528041571778442,0.5751208708011047,
 0.16788087509684124,0.037343127745252946,0.8708379688443287,
 0.6528819870133984,0.05096193786059844,0.40135811290339485,
 0.7251216397126217,0.42500254931131853,0.027792618409299807,
 0.8361675949954093,0.6185587545393965,0.5440080115535728,0.7278992793762897,
 0.7731343517951438,0.44532199251889604,0.6992966427390177,0.1631964395834844,
 0.9254698948697357,0.8993370062170742,0.10139254493171446,
 0.05724432791800316,0.4983836578278166,0.8764020029100752,0.9141646678636766,
 0.46837367253435913,0.787278644813012,0.7168562972406755,0.3054120677503297,
 0.3685661366214713,0.04984245400296272,0.943154734569253,0.5943063726064847,
 0.6923286677587466,0.6314433641905463,0.9437349354354955,0.6280523406208538,
 0.6144559185520309,0.11099919732117924,0.6303467930524871,
 0.13295375385637442,0.19435756317750374,0.2755592808055243,
 0.17260142438475934,0.8407700027705172,0.14393080254751112,
 0.7217569393812328,0.2939663196422784,0.5142036832153045,0.37183600401946537,
 0.3934354644089517,0.9094815163843826,0.13204706937584254,
 0.46867627928755073,0.4978486377815968,0.908600534700193,0.33391962944978704,
 0.11807925278316822,0.24305690465644014,0.5873579119691775,
 0.46582313507246575,0.6031113052156809,0.9141021161120063,0.9152893474232514,
 0.6228503345106996,0.23913517890080094,0.414952916911024,0.9925999555226834,
 0.5147643407156282,0.49788227083100556,0.2821294895958424,
 0.29416673014301553,0.6591116875786778,0.8927292517169756,0.2563268478256987,
 0.15363075813615157,0.7751685225546656,0.1930843378726741,0.8375727620167869,
 0.8924188647132719,0.6682913037264993,0.266579689054363,0.10705792036418083,
 0.22461197650394515,0.5876428899768327,0.371616509420182,0.8314670977048189,
 0.924540786054133,0.8660411731164629,0.03567831073972905,0.1867608162008526,
 0.47613410879191986,0.10109063369760785,0.5819095951310236,0.454037959869367,
 0.5989639071782061,0.49314225108108756,0.8461103624614944,
 0.25897392468055425,0.5537445350631398,0.5607234405718741,0.9827971774259294,
 0.6921760020427755,0.18586401159516652,0.6772175129480266,0.7852374205120292,
 0.21584108257558166,0.950892927011965,0.3096202948090562,0.16942152377439668,
 0.07234885250411527,0.4240216959311388,0.35982659767045455,
 0.7516678481847437,0.9479668417090377,0.6905210025005671,0.9020636968457038,
 0.8998902590724767,0.13679399856608776,0.24847218502665158,
 0.2125894823919392,0.3383593795100064,0.84788142077791,0.7483252127334905,
 0.8462582470311818,0.9199321781365122,0.2938103947923082,0.9271362036142454,
 0.5664730583692813,0.6553168877065461,0.22541378564354164,
 0.18522442578515252,0.9738303690209443,0.45732525887921005,
 0.14136555821763153,0.29043540137358603,0.6891018909448621,
 0.7949244469547125,0.014097371515005563,0.32945973024642006,
 0.24357479511111513,0.3092726103010822,0.503362957591889,0.30933189857808674,
 0.6856716837717224,0.31279109755484347,0.6030946966005244,
 0.19537966384326588,0.0824866366421475,0.8298272716161826,0.529573379589724,
 0.1634765686668156,0.9728116104817387,0.974635017512888,0.6726300143346265,
 0.22800765709605786,0.8272667140994411,0.04841038220248006,
 0.18238471116100552,0.3896053658427179,0.5286190419208019,0.636829656657929,
 0.48208837690465955,0.21714669264879594,0.99473019070772,0.2575887289169171,
 0.6045846563862838,0.5234759513780908,0.2899314255821245,0.5145671108999474,
 0.4591452222043001,0.17771161562651117,0.48245546525928773,
 0.5080074006925934,0.4888315491397165,0.03594294388396191,0.7939970784569623,
 0.34031141736230597,0.20732464843909637,0.2660950678707563,
 0.06219909100531429,0.5349593791136037,0.006692207890451107,
 0.3524295911298633,0.6246292123439061,0.9086966559659007,0.7317739312056114,
 0.9467224422921537,0.6689841687476046,0.35205044844112143,0.5656602785970747,
 0.2990377641088161,0.10636034251488347,0.5801182351825517,0.6690353188987846,
 0.6167437733092525,0.38377872325763396,0.34328586901744784,
 0.7265790871700345,0.66780968404052,0.2301193188334021,0.49247742542347495,
 0.649122260131796,0.3962700605156597,0.40743804690427143,0.42894822225236884,
 0.8001649444845457,0.33772267414787804,0.9548586396714431,0.2106965041099238,
 0.6970537811066405,0.478451602515235,0.5469263704807955,0.7055230119854216,
 0.5216561832100772,0.7876817279554569,0.3208484098397525,0.9685637617023182,
 0.9413808880940089,0.9680388669722308,0.3060805864369843,0.3321062484998627,
 0.09716751274801849,0.15525084643965115,0.6153466062900141,0.646219625699479,
 0.23280411391300448,0.0779913515252727,0.15956806449275762,
 0.3733844068059575,0.21352365224446068,0.2650128753025507,0.8195494130157812,
 0.9235688790623576,0.8343196171564339,0.04026627157162388,0.990325910820114,
 0.5079449873180368,0.11749395837796317,0.5177619039298664,0.2858531235664157,
 0.5672160656934324,0.17385336668852203,0.9695406669202257,0.701346181276922,
 0.10987356947330973,0.3273315327299785,0.01602149489477489,
 0.8025690711947929,0.5525354674400912,0.5948119359568962,0.31175701607409734,
 0.13691511427560243,0.8159378278664575,0.47500215995427797,
 0.38850371689261076,0.3609358869430749,0.1437838228345647,
 0.26677463413709923,0.5865318186438886,0.5899655983426648,0.7828364042249543,
 0.36371205793447636,0.10259183532108751,0.007215167828466984,
 0.09432013476330137,0.047058473499643205,0.5540035162344403,
 0.06875743023230318,0.0793380147271634,0.6159961478460422,
 0.19838409327490625,0.01648423567083035,0.9632546105900169,
 0.11708314576266643,0.35349457977899934,0.0900435048714876,0.762740313710525,
 0.2841478847636798,0.30504003062572027,0.8972961712072827,0.5655600505202837,
 0.4479099208714873,0.6998714520560079,0.3142264158711462,0.49400949191084975,
 0.1556643060293973,0.8314109922479685,0.6539786837624258,0.00912555539246418,
 0.6774528991893554,0.4347517515527538,0.36679206154710653,0.4487841540374794,
 0.8297455585179873,0.537475607674789,0.25741961011235004,0.8817440179220788,
 0.9641198933298463,0.5350403416046925,0.7580063855906117,0.9162722155790239,
 0.05287867641940214,0.3166260942383238,0.3921932088178167,0.7631988083426434,
 0.7169987238657043,0.41561556963185564,0.7462173342323112,0.5737135530004259,
 0.14012205592439664,0.9580281250226944,0.1711708979934422,0.628757920380143,
 0.26179905034682016,0.7507967105588925,0.8576121804032326,0.8272254877984636,
 0.33191400875079136,0.4507642425263997,0.06858308209876507,
 0.5275500370914363,0.8199942012338474,0.8447748582141369,0.8007400719264326,
 0.4485496285740246,0.6296575490978388,0.1786942565342846,0.03988893815716632,
 0.6443903611617334,0.35800338160467726,0.5706826077765709,0.5399955023843778,
 0.11004478455469802,0.1452182766296135,0.4003974289884652,0.4765230521443379,
 0.8478697542475911,0.24626216230893738,0.3405192143046514,0.6443347600204739,
 0.22018064969975293,0.05034949939096822,0.3461091278289242,
 0.48308754377393215,0.4686626998193013,0.4842944289478317,0.4423182655771215,
 0.1630125621453692,0.44163596258995275,0.5317722731929118,0.5319526977045254,
 0.1490258436312798,0.6201210798884942,0.11403790890959054,0.9083534767443522,
 0.7586681350311345,0.07372572127152843,0.5191987271802833,
 0.34607074476863575,0.48419854503610493,0.35423721600912583,
 0.5854112303878384,0.29379787444136407,0.9570929640840338,0.9774233233283195,
 0.5615988833630519,0.9054868385396766,0.45919981601064097,0.4894869486811937,
 0.2365031591237189,0.9992514774275867,0.830489270764209,0.3599821802766058,
 0.4954916854245017,0.1623603300563209,0.011356723115937828,
 0.1400686414417215,0.6970369381959305,0.48772293748592477,0.2872504968338858,
 0.389810478403147,0.32376632607560074,0.3168206470051298,0.06396057004533062,
 0.7039296368656932,0.4497266149623894,0.9205573186197553,
 0.029826199069232895,0.9367598518792335,0.8976867297335754,
 0.5455797838806414,0.020024118838552152,0.160476704215045,
 0.16980498894130203,0.19008504432213535,0.7276176035637908,
 0.5922437626326671,0.4295932113887102,0.44361213462195936,0.5852326830873371,
 0.6100098561242371,0.3888272266079429,0.9131847528792509,0.7513449754551443,
 0.8654820525592736,0.6536362522441812,0.809528102539567,0.11489779964690527,
 0.10600986148581604,0.7999288908550165,0.08680182568703843,
 0.20819675614887956,0.7312025532952414,0.44555581201859384,
 0.11819692222243794,0.08890639621935914,0.7188163184044716,
 0.33076119166475415,0.7809286525606038,0.18452104543962067,
 0.1281907775364599,0.7041170469941747,0.6879463409583935,0.16833168850461844,
 0.4184206687513301,0.2756009748354943,0.5257166257217272,0.9106935346864501,
 0.33601134134824373,0.20397473266003552,0.5184716081910029,
 0.5434171680346139,0.7677389590768444,0.3580756919565645,0.7052177023126049,
 0.008551559956996857,0.9892805462521173,0.6061820929899676,
 0.8116482304856054,0.8917005761613794,0.03252900909218248,0.3636458043170142,
 0.4129866334952085,0.4711373083050395,0.724552857118137,0.7431252161907872,
 0.2875292983475526,0.28755821626619504,0.004940889042443941,
 0.7421254634792323,0.42816074529746917,0.0032504893292083503,
 0.841901797574026,0.8928949500842984,0.0020675089971867733,
 0.6278698087387143,0.26447437500039295,0.9597255023053989,
 0.07730889712154054,0.9882121086831167,0.9910283431065342,0.2912356283399128,
 0.5147092765391946,0.7610257003442761,0.0066050285532891495,
 0.4499924184929358,0.08925455012877548,0.05045956476585578,
 0.40029491764349967,0.8767833572845416,0.45676417870946484,
 0.8718343900024619,0.43009563360143555,0.8930845042002656,0.5086156003154492,
 0.4448659218337174,0.5942730595697461,0.2505566245472016,0.9207015588150247,
 0.37638359040826064,0.7258290510119664,0.009102898561520356,
 0.13569500418511937,0.482332495360846,0.9208939810372182,0.7007776342413796,
 0.2472220324601615,0.5019181650399533,0.08547112554422465,0.1531539111570681,
 0.42723187074056024,0.29511274047926994,0.3553247525130847,
 0.4316726575782798,0.9315836381224425,0.7950194543507532,0.6208690837514755,
 0.7746058030632671,0.729429354515433,0.8368736323684347,0.9596548298904422,
 0.04327990661360548,0.1113324742761479,0.3285471446156283,
 0.21357596505544296,0.9723892377283545,0.8628062460589568,0.6432372275174567,
 0.2167292426411389,0.18522181035499197,0.27484473696951484,
 0.24923122306847223,0.8101380723935311,0.0013877711033540852,
 0.5061236122288548,0.920283756255835,0.4403080378757579,0.9866951577478974,
 0.8886385186106152,0.8250493233354701,0.3642160058851006,0.35957786014829596,
 0.016297652332658896,0.6906274139154653,0.7930134355514714,
 0.2366153245378082,0.4653546651535645,0.30459795529078093,0.5141103398621709,
 0.040122519482218455,0.09282093616588605,0.643630519197147,
 0.26016923500027755,0.9246417246931432,0.14162470929017623,
 0.9467673071137468,0.39695497639567323,0.887914079500708,0.19709449103105725,
 0.7417500399436984,0.6546464256969119,0.4592706863524665,0.5572065618887823,
 0.939468748597287,0.20483061544419312,0.9612710327556627,0.9786344577866514,
 0.5493792889639358,0.0835754667430888,0.9518388807626835,0.13891637287117808,
 0.06561914034174254,0.6311754008238523,0.6333278876318081,0.3225377675596167,
 0.6437006580672798,0.35768273521564375,0.2546389571251657,0.1532678289016558,
 0.08474897091331723,0.38962682669232795,0.3580774183483786,
 0.2480924117752734,0.37100565527173834,0.9316850374763722,0.3104437754317211,
 0.21295837522232042,0.5041447848620508,0.023427841096125834,
 0.4387771082586266,0.42579498522795056,0.8958093989311166,0.4983965481926593,
 0.07753859621751225,0.3983924317074914,0.07539194189844,0.42238048434892794,
 0.8722969299739738,0.8252127999295131,0.5862748233418913,0.4838910196904619,
 0.7101704160855261,0.15135944835254467,0.8618693867282402,
 0.11947317240849642,0.9580583613061192,0.08086823230384677,
 0.2397680164534931,0.8291423703507337,0.6936940769134734,0.8333431513098065,
 0.6543095335898136,0.007288107375523101,0.40950407198146743,
 0.5034764189980826,0.6766087071184819,1.4223209388375402e-4,
 0.01787280742142716,0.913183599041474,0.5100361413782559,0.9423649966048747,
 0.5227864942051185,0.4354410247510949,0.21684963696387982,
 0.15460338927072692,0.8780452278757167,0.20965751884910266,0.848698683210725,
 0.3979853919959879,0.23102430488408487,0.8559672607196848,0.8760829888327383,
 0.6160868035415987,0.0076039192305471515,0.5306731969785877,
 0.8046936493570435,0.39903602147866457,0.6349698539049792,
 0.49030149360744724,0.5855617273495278,0.4285645928202191,0.8701662726767528,
 0.5656358185124425,0.617544541168441,0.9576445465240966,0.22949727785597274,
 0.4461653144904124,0.6771588972899045,0.3932206013409193,0.9038588528184571,
 0.31158819721194586,0.8421535579252939,0.0034687839024329925,
 0.8433187844643699,0.4055114684286514,0.4245498331485372,
 0.017086010422838527,0.6074561056274205,0.6284348221983733,
 0.9303121285782943,0.4211539445642152,0.23849413447882906,
 0.40926175475707227,0.16367343747033936,0.16087712832035717,
 0.7458351542117253,0.7243048504106908,0.359529055029012,0.36975194645492404,
 0.4723763982952712,0.36416983744598275,0.9878998118084764,
 0.37845670238277174,0.7613065348185686,0.5690964072797892,
 0.13937191637071944,0.5668607159401566,0.7875323065226062,0.854782460399273,
 0.43229404110165226,0.22967046873104602,0.3571128098607269,0.202235313579182,
 0.5607650121830433,0.46955273789485397,0.17957802240245613,
 0.07422714455605972,0.5384940578314474,0.10340598083992558,
 0.9037967956881302,0.5581939473973497,0.36675263838520755,0.9790570526646268,
 0.4455826528907434,0.945455163306071,0.4778434106867816,0.05551190470827727,
 0.14250000860403467,0.7906393907912417,0.7548615529731708,0.4244458985442694,
 0.6627897346863418,0.8800790543134452,0.40080624731068537,
 0.22371856572955162,0.03400701265553818,0.2864287239701223,
 0.9055673943189948,0.7273822262282836,0.9177479420057503,0.2046557702770394,
 0.4211014378200072,0.8181054106751106,0.5535100658505652,0.5950663874749857,
 0.029444362849917582,0.17417553598093638,0.5443419725456291,
 0.9071876528149437,0.15509817781845792,0.7087408655153378,0.8989919801446753,
 0.20370336074044904,0.8299789372211597,0.3278198145469562,0.8167451840151393,
 0.3104173168286555,0.5293553320369487,0.7503131950299101,0.04372795127543805,
 0.6176950116362618,0.31781172241598854,0.45427675545402313,
 0.3304264702062143,0.7493176686034819,0.15187941182759046,0.8290109028829102,
 0.9727544829379297,0.2687350418872607,0.7769128944889054,0.7069197302319236,
 0.2131422834842538,0.8760501891153177,0.6610392583920963,0.3124636593888799,
 0.6602734745957644,0.7141707927645846,0.41574843189556665,0.6959332760743036,
 0.36084400130145367,0.6403923226907832,0.5266849796149565,0.7391335598240807,
 0.7349973177609189,0.7794845955002877,0.4787491297589025,0.10666741008638425,
 0.6882003730282118,0.4911603655068185,0.1967856075424259,0.22213306413427025,
 0.9692870777019671,0.8408560465452458,0.6887390566940236,0.5557021005311336,
 0.36788759453961783,0.32942353931882773,0.5379012112394134,
 0.37036432146655685,0.7085848035862534,0.2714106986936349,0.7174690758683901,
 0.2453432122231014,0.4949407105976845,0.5067079816479727,0.8552878938253521,
 0.04796394088492506,0.10620058326498438,0.9784188658611156,
 0.2292996091439723,0.6280815984826233,0.22015618906202616,0.3367040623013282,
 0.8151172949051088,0.25623749083615155,0.8398374453188096,0.5986293714385298,
 0.3240749949500594,0.27899418020930455,0.6159512862077368,0.7235453195509377,
 0.2089239473204696,0.7782652639678282,0.8188090096019232,0.9730628606328948,
 0.2234167853608373,0.5688707501424994,0.40951532606895136,0.8773237732123851,
 0.5435191123647634,0.870907726835899,0.9536579627172265,0.09706837876703667,
 0.011518520714296265,0.378450188572649,0.09394422079751286,
 0.10557583670828752,0.021878227969154174,0.15031584378652518,
 0.07642489269418684,0.44461108674968575,0.5353674803641683,
 0.9367108318145979,0.7616724800354487,0.1908682883489815,0.27708208835234904,
 0.1733860536971965,0.9110473409614509,0.04920867216693503,
 0.24723984673623378,0.8462055667607753,0.09793887858727157,
 0.3757038973089739,0.732781443497425,0.4668904179727764,0.1842773297820185,
 0.3252113331997446,0.3200704895863722,0.9104193259995541,0.6051895207772875,
 0.4608918550626362,0.6492900491454577,0.33293526393134176,0.8458631864701323,
 0.66102581289795,0.9927624606516874,0.02010991249470928,0.3808335231733533,
 0.3186187022627711,0.6097305739144799,0.19655458687552474,0.4810344296087238,
 0.3145095268422736,0.44293277303745027,0.371952697498398,0.10708175285114185,
 0.27734840772046154,0.32330755043233905,0.7752326411958076,
 0.9175900401388715,0.0015902710069732873,0.9991453114769936,
 0.08612402050247914,0.7016134586342542,0.9677237320846344,
 0.35196213198422777,0.23295465908594948,0.6908940489208955,
 0.47838210647373025,0.3516415636016764,0.7321336561182286,
 0.06321261449466542,0.8083457428452236,0.4691304393568727,0.9216151138194969,
 0.38352477654435535,0.6023553317845987,0.14235980784257662,
 0.3143419321683868,0.9990022272195885,0.334394307812805,0.8862969071398823,
 0.23999383253160034,0.527016758135775,0.3958952705941501,0.9563946459074306,
 0.5825013623827812,0.7102491404217552,0.08523750379835837,
 0.31900781222972907,0.6223263964906797,0.4351878895896538,
 0.46961340303200205,0.603028546820096,0.8584763986824603,0.7951344584156519,
 0.8086772782476073,0.8827667787139265,0.4401962008056125,0.7432314988950802,
 0.8640521288943783,0.33955185978351754,0.22133527714855883,
 0.23696601431546938,0.5145092021581351,0.5065339676520321,0.5781783410992775,
 0.9192509232677845,0.858097891050642,0.030462951900316337,
 0.022448828576029367,0.9918095631598595,0.2748176429923964,
 0.3618513950601414,0.26854986144026327,0.854869763788011,0.6457179146872962,
 0.3912743250102837,0.3346972651744048,0.390924859595733,0.5029998875148334,
 0.10249110192596877,0.4217229296246582,0.32721164916706236,
 0.2647926422818605,0.7961148789378935,0.3633906103266039,0.5413536237821422,
 0.08955772783196125,0.11415328034653394,0.43412188251726425,
 0.17927286263511144,0.5466090236606931,0.7796653178727939,0.6955831506875584,
 0.47196387165178555,0.8095062139355866,0.00885891339834588,
 0.043512366424741544,0.8006684872245552,0.9653034525058637,
 0.4713294733703438,0.137706153700894,0.8293012913898161,0.8465255936436918,
 0.5403759077492739,0.5001380648685695,0.9684996336263363,0.41852286570553265,
 0.06145100159606387,0.19016848404601316,0.3571766200727845,
 0.47597045945068805,0.345779982056333,0.05319876395545886,0.7686281389420244,
 0.5445983978844403,0.7163223211050092,0.9133987451442681,0.8764122421539072,
 0.8969474184770271,0.8116329649796423,0.782003564641719,0.35289791828470607,
 0.0011519647337958627,0.3845484646648145,0.7309018569247664,
 0.40420114593696355,0.31790395566724816,0.3282318698082495,
 0.1410892140668315,0.9993357380261557,0.3776512387598606,0.2677394678942532,
 0.16401604245206258,0.48637883426897144,0.9129705771143937,
 0.4022287601213349,0.7218461963137665,0.5974578383473981,0.5503845913719441,
 0.601292979736385,0.8042359718616762,0.368043298920815,0.7634572105219048,
 0.7395544930621637,0.13671681326476803,0.5483097082699514,
 0.25475150519249523,0.8377134308280356,0.48436014937536687,
 0.24862935213082282,0.8923311775401546,0.621740434111905,0.899545618563949,
 0.10850636879536146,0.527311796696442,0.6752400507844656,0.2853680071258845,
 0.5076993950814213,0.47091668494911876,0.14615729502964503,
 0.048139630531371935,0.6563103633933037,0.8869504901831838,
 0.10992880135582117,0.591204864116925,0.7213737431477509,0.21846040681376666,
 0.6260176416792113,0.8286496505304972,0.0362860761896725,0.5049800516043905,
 0.3248502027402874,0.7444602341727821,0.5850580953840795,0.6749710636125518,
 0.32393272466800693,0.9201908366635692,0.5589673289765908,0.7437740988199415,
 0.8560534022668312,5.274017023390032e-4,0.5668178762806869,
 0.33694134732505476,0.6497734432629856,0.21687239393744262,0.138087387504078,
 0.11443845395638674,0.1801241481092578,0.2935077415388391,0.2779564201905842,
 0.6537607453496894,0.9339244878219628,0.005104640606239275,
 0.022303867326488036,0.4699020712491965,0.8603251437917968,
 0.6495946204007064,0.698517980054364,0.5676410583901478,0.5059458158619871,
 0.791029993900835,0.025575964151294373,0.3371607350119218,
 0.32495968646020923,0.10368803892660594,0.8861928424001922,
 0.004217089823570719,0.7272376911082002,0.23419814517238546,0.44462170940194,
 0.794572893452614,0.8638843502182051,0.3444182581083771,0.20061754023787215,
 0.8666648500751338,0.8763978960877331,0.508661473469264,0.8088139723995396,
 0.8561321111167741,0.5203193381858278,0.11751191160910646,0.6195821632431251,
 0.5586749982178275,0.27821282234572386,0.03699771959324549,
 0.012154991990815156,0.7302343266162122,0.6023797054379563,
 0.26138900258027664,0.5648323753120287,0.7399694628414828,
 0.29310642033293477,0.8524188748142953,0.6237493921228736,0.8874806236043048,
 0.3154306863251829,0.3375684460909739,0.23686226621181372,0.436374346407735,
 0.3937781576813556,0.6011276593811312,0.8604672501976973,0.5395880372224285,
 0.7468642165061945,0.6413153678733001,0.7496826517017411,0.1405953786638543,
 0.19495716328403445,0.7593505110056484,0.9021909523419844,
 0.43581030688876266,0.4083047099456123,0.41594286048484785,
 0.11715335434898932,0.5416382784839732,0.2453839186946406,
 0.34564602283716506,0.9805445174556101,0.945027270953095,0.8066393572209761,
 0.805774525179724,0.39856197163662765,0.7643261352279008,0.11685946721167995,
 0.5024811532360456,0.0035418779496418917,0.7773310089141906,
 0.5558288850620294,0.8374821515983437,0.4782533520845329,0.6857173310790114,
 0.42452011328532624,0.6353851198647024,0.8768343050423051,0.3030157773129518,
 0.8363695261710354,0.17562052230371694,0.3651093935300028,
 0.47234450997806654,0.0958481798711196,0.7504782007075701,0.8709660105702948,
 0.15373298421029047,0.6719189769575954,0.8367936992006262,0.2473088454966451,
 0.08563575948191338,0.8657168535729739,0.4926744122115022,0.985921290070335,
 0.38626335107980003,0.47607089331979047,0.5975241842520566,
 0.6526511530195751,0.4416320866581538,0.6254000945234895,0.11525016880953132,
 0.7562756072446218,0.6978708963673529,0.6006739043758585,0.7980278255709801,
 0.7342179747576838,0.6787010533007385,0.35634832316174936,0.4063415001531101,
 0.5908391180638375,0.8836559019157644,0.6756312461595836,0.766369123341338,
 0.054363719503283114,0.5270817531948949,0.9114224291158163,
 0.6137708636089741,0.14793472032249189,0.9712927891438788,
 0.10854346108183854,0.28189378981921553,0.4787954054271376,
 0.8724983579450551,0.19124975085522833,0.5045874576495617,0.5878750036756542,
 0.7539287707587607,0.2666218554322468,0.07978579583626533,0.7133574730701309,
 0.8220706725530426,0.7109659974385057,0.6062174877889096,0.24548986488047542,
 0.48111601657214664,0.7484383560774484,0.5325967238186641,0.8325538680403212,
 0.8406090622571322,0.457066836480147,0.4541815590332834,0.6002321609231098,
 0.7408158615383618,0.7752294887491182,0.06383133118188011,0.8136717694878214,
 0.5221385476713729,0.08966289904662772,0.4260925866266432,0.9255947982869914,
 0.15730997320710705,0.9023032948021106,0.7069258222924448,
 0.41834018983776733,0.3533154280925006,0.711643422480203,0.4886510165177671,
 0.11417552369857276,0.3786896384774254,0.7079917616647868,0.1854575143789854,
 0.40416112992339526,0.9254355872226487,0.8071421180071111,0.4850891397521436,
 0.8600804269583846,0.5564013262959701,0.1338676170319678,0.4694992273675964,
 0.9726144558409257,0.9399810647988444,0.606465641218295,0.9620799993093873,
 0.41173176596712246,0.21811900198720702,0.8736142026164488,
 0.15009103979282323,0.5767796287297853,0.8222384209132777,0.8675616241336009,
 0.9198364581998376,0.16343056758037755,0.8107194753999489,0.936673326307612,
 0.4078270952975246,0.5423156300383742,0.7904407400079669,
 0.010038351803500989,0.8397505383378117,0.958873766659707,0.3367948463296263,
 0.4432232519950733,0.4751290485781535,0.5526643097507952,0.6601831720588349,
 0.8100262989312547,0.552171768861457,0.3803148858162696,0.2477589356715889,
 0.018909900379519784,0.782154542104305,0.013985884887995792,
 0.31566465754546613,0.8389294070525418,0.9544373795143242,0.9223530635229228,
 0.5066768795342361,0.6567007086330645,0.2386517025223569,0.5799329840798098,
 0.7404387316972898,0.1999422296470723,0.12403616285799512,0.4898193701487009,
 0.808303106255655,0.008711363153426133,0.07331124872729877,
 0.4982727712882532,0.8753693595555712,0.5150478149136379,
 0.002041972209481324,0.6403145295964521,0.8868238278071315,
 0.7250331660702927,0.2020678223544936,0.6055082499479892,0.8232691207177354,
 0.7602743896641047,0.5691617211445378,0.17802263643284455,0.3901188720748964,
 0.7204835629197851,0.9881400381691817,0.8422551121404934,0.1904627144567208,
 0.7707351935642378,0.5459487727608462,0.11847121229541602,0.2530121667104944,
 0.01612007769915902,0.16817012489490135,0.7652498411049303,
 0.7703557309211859,0.4707707926785699,0.534540995406788,0.1314184273622664,
 0.5990862979831868,0.5765873191943588,0.5817089632137029,0.9878413411639129,
 0.8990207047473852,0.6483467433514215,0.24186296937258334,0.8005970928689103,
 0.8790578921419895,0.6663033855138086,0.9470096669560335,0.14753561715693675,
 0.8076684402482032,0.3207379614417548,0.1989633832499259,0.7274052277653368,
 0.30083534546543467,0.03664313526113783,0.06662593714158727,
 0.5118482662862573,0.678801023493107,0.8315209669163244,0.2621301637406015,
 0.12827700472604564,0.8702495541094585,0.07243735855701328,
 0.48006658684640946,0.25230820434212975,0.7427098765930952,
 0.8610183907275435,0.7781727585903568,0.8636353858572685,0.5380843239628135,
 0.06007605534665206,0.18904709582049684,0.41691764709268986,
 0.5620302806639386,0.0784441243576609,0.557901899143713,0.8313843606673307,
 0.8379658842167192,0.4598548100061626,0.25297765423664054,0.3155409521296111,
 0.24941403784205363,0.39626339850331793,0.2874180526600768,
 0.8731718441584451,0.5181248534395038,0.3913330151141865,0.7990364603917449,
 0.6702888107117255,0.9797013982543818,0.4001725276370125,0.02500517116740264,
 0.006845575209373456,0.660913142463774,0.32022534027671445,0.567858977900939,
 0.2152083254617465,0.394943947990101,0.051495551375139725,
 0.03313805757433008,0.6118728615912971,0.18197331908359127,
 0.6119055777373427,0.28418348675040284,0.29242048385414654,
 0.10324469949560278,0.5713001844442072,0.7832839764239002,
 0.32569867626990046,0.9769328228427329,0.2283813577463496,
 0.0037932763603912845,0.46739178641029633,0.9018001007707539,
 0.9467850781830097,0.48766050243605485,0.3997879924346127,
 0.42156857673614423,0.10965441272938237,0.7115545752578223,
 0.9218055809339427,0.9008146391015938,0.7147696545760958,0.6475767537837388,
 0.11448710913315496,0.8882409880746677,0.7837418171621454,0.7016683201739189,
 0.4833025059268662,0.17925334825560602,0.39400139930462896,
 0.6461233232927668,0.8257375151224684,0.1090784945637171,0.24630536211387666,
 0.39390112728611415,0.5842831528385843,0.21033216559380896,
 0.7935583615287275,0.3193844305625788,0.3083935383363062,0.928779830557188,
 0.794436468378513,0.7407409590524885,0.0426141670757727,0.035813693196531604,
 0.06654310153089726,0.4259570910027576,0.28241781740562255,
 0.07890556275066463,0.7006269654785641,0.5749353719955587,0.9015127412928813,
 0.5376032551019635,0.18440007485569987,0.4336991656214484,
 0.25753430418428724,0.5176924850088521,0.855317643315837,0.00813764841178366,
 0.5226429236889181,0.04891086328062266,0.6906550911158441,0.954322644554297,
 0.46726782522972166,0.4316716729996598,0.6354176076040935,0.572219971513049,
 0.7140863738935432,0.23255015247514665,0.06970198145975304,
 0.0017956643496749303,0.9576519742873661,0.4761072758485323,
 0.4802025984873457,0.18920638915630805,0.41530262185891087,
 0.6409585291075679,0.4353821903845303,0.7961103704124469,0.34921331877480255,
 0.7907004206369126,0.5567199520488167,0.6516992942660516,0.8174923822487618,
 0.7818345060516777,0.6003782746961177,0.5853071125568087,0.7700739350224677,
 0.6052934058526366,0.09115005151391797,0.3731659723927069,
 0.29901881804764097,0.5519509956460991,0.8647536823027688,0.6528294900741164,
 0.30300873285009,0.6551723730925578,0.41040946423591096,0.9906070280855718,
 0.6974751879666539,0.6187103061521506,0.1691621444844772,0.11998702603024436,
 0.5674541879792558,0.9818177619465107,0.8466701032396062,0.8424208624237431,
 0.4285635710970095,0.43713341139096673,0.425594211848359,0.8008927049279057,
 0.3538318203660543,0.18976962181846158,0.5131007847829262,0.5567963645496742,
 0.5244647178301894,0.8675946369865442,0.18998473747998568,0.637142170067036,
 0.6442845969358296,0.8462700431907378,0.7612425020322109,0.5271596841074896,
 0.6129646158021265,0.22068299890796006,0.05726899761610815,
 0.07496893543696292,0.11651940432766605,0.05917371199623145,
 0.03156568189897002,0.2811995746974161,0.4623196522069215,
 0.027444183643413167,0.9286470230276286,0.2527164592303829,
 0.8808432552251948,0.002009113704384502,0.2166910239034463,
 0.7864639495367101,0.7611764887758619,0.45686052789465836,0.3219479109594279,
 0.5807225375519527,0.4292345438006929,0.7827889413171708,0.10291650187852175,
 0.22932109856080807,0.5093459358985389,0.30173998843820915,
 0.16250047799909773,0.5132066657371526,0.9520320246752773,
 0.16953982801789602,0.42418825873987864,0.8900035717934707,
 0.023288908326371427,0.603234008134249,0.013144471952591386,
 0.418093936110957,0.7473185356738575,0.25657207435689366,0.5755729230537996,
 0.8853675542701329,0.4282654244966042,0.5542758521972795,0.47442848094713075,
 0.6224854074074033,0.39498210157102487,0.05190714873079727,
 0.9002892253893658,0.4266165848814858,0.4205927970628981,0.4322005573533434,
 0.047254993133787115,0.18048507849254114,0.8737911216364541,
 0.450336935133122,0.8441612927921889,0.8276032652014579,0.08092993162688111,
 0.8487743364272894,0.3668695195640068,0.06076549972019851,0.9763974471070458,
 0.8217911587726796,0.3871953882809761,0.3469385437494761,0.23563063447434662,
 0.3314023831914322,0.26554223102112795,0.3487203521588116,0.5093382875613837,
 0.477526238654465,0.2860522698489396,0.11294277645115824,0.29086643719228644,
 0.7851387425887677,0.7302010733192666,0.6301159859162102,0.9954710021466271,
 0.7745394150368146,0.9265859106573928,0.09727414639664189,0.6047055577277789,
 0.21261033906881033,0.4274506629908451,0.8923193184991213,0.9463950628102503,
 0.005146560887035356,0.684114282556309,0.33698294313783794,
 0.13516032518750387,0.7171366127583478,0.5868704574666271,0.6744135038072869,
 0.3084731977817762,0.7332949742375929,0.8211248044230682,0.31341517338000213,
 0.7841129738453251,0.7695830956814994,0.1750633119117848,0.19876276044595076,
 0.9300312711065263,0.9759881729730259,0.20981046972231476,
 0.43298859591470795,0.7677766791654279,0.23436268151555384,
 0.34185445608930687,0.29201326254098114,0.5358990077023696,
 0.7765715343872068,0.9303272753958678,0.44112051370737404,0.6894687125060145,
 0.890086140039435,0.1712007730829923,0.48970400248573975,0.7654846679951647,
 0.9553484718978552,0.576908451561015,0.8334871067386235,0.2635726410639422,
 0.27594561094798364,0.6577458257183088,0.604602813568057,0.12953374136712414,
 0.10528339793051555,0.7312990289733996,0.5040907541858709,
 0.04246272630517944,0.7122909977642564,0.3959027987364986,0.1738248667574508,
 0.1360351320490477,0.41525487194414046,0.30679650955600335,
 0.05460345035886294,0.20559060926016537,0.33848714291706605,
 0.5536385647837349,0.026287654853046405,0.6798413617866755,
 0.6676075233301488,0.43286235751866997,0.45937049617578374,0.886050308761668,
 0.5324465053357872,0.4154945831169866,0.7663159335508891,0.16775441632874422,
 0.290214857564155,0.8341194048513099,0.15916991423017823,0.6325686216054026,
 0.5778934684902853,0.07342563488970166,0.577491259818567,0.04701112861833445,
 0.31575278304873244,0.778271893321471,0.13737739776702063,0.3303339012900094,
 0.7928705551651662,0.8180703361610214,0.12882328095048834,0.6149653640709518,
 0.950623882524265,0.4503769618863933,0.45512077087986014,0.8534740865007099,
 0.03153221613896118,0.7748837998631395,0.5789980403233129,
 0.04092820800108654,0.6583548180489109,0.5886642057493372,
 0.18475048035345842,0.42273936915357646,0.7836139108699431,
 0.47118301491206116,0.7285197149067469,0.18313847873514444,
 0.4064479693307328,0.24948687270285963,0.5515614764845954,0.4091815569391951,
 0.4355530591636565,0.3957933323615501,0.7695379314011814,
 0.0032238136865012645,0.4160080383983701,0.3134179605177877,
 0.8948784899661626,0.8691361180138664,0.5062315412651242,0.07681061749606066,
 0.2239061400722152,0.5310214134641994,0.5643673175239428,0.27144761788254024,
 0.9834894482250859,0.908652571168097,0.26167083591882767,0.517537858677709,
 0.19917350112138887,0.52634646103709,0.4651704780188194,0.6314827881981765,
 0.9602724040544061,0.4196434566866386,0.6394503429168538,0.42759452384351937,
 0.5932524669025889,0.7976472222718819,0.9983994663756374,0.21106217580770692,
 0.26874044497471716,0.039692791006356254,0.3406633322080612,
 0.8718918695312046,0.6132320232682147,0.6389335941217061,0.46603257814723253,
 0.35694436615895186,0.47035178289804913,0.5008659574321449,
 0.006480479969789621,0.7543811387354327,0.9595060467950421,
 0.09359487575238068,0.4931104174753108,0.8822965925687403,0.1370353879013867,
 0.9865279566644836,0.6160442988855622,0.6665664480568227,0.507258135694321,
 0.7439105322259165,0.8823546192492554,0.29050680534347983,0.1272393581976392,
 0.680399186489056,0.8173790782192055,0.7363762941844353,0.1561319681986244,
 0.16482231134588388,0.11700385430274651,0.2102534119113394,
 0.9300416144321758,0.8277605321340413,0.41027973014815444,
 0.06241682586264874,0.10473125459166854,0.6154445942710246,0.938715582054537,
 0.8945066846659964,0.015443202235984388,0.7660903656854723,
 0.3731107334927031,0.4447099175156668,0.4064848130486579,0.4223035637995781,
 0.5767061854948387,0.9524064333441262,0.7674567801267338,0.5228689641594175,
 0.4958387675403336,0.4456225251543807,0.46918131394147855,0.4812906977636342,
 0.6036200504755564,0.5561852819006559,0.6958437674899854,0.29566547292182976,
 0.9879956153753152,0.6743337534933553,0.182372795666343,0.7776192965710662,
 0.9202514164197585,0.8449549330025805,0.1784026417305945,0.895940650828428,
 0.5020388130982021,0.21428069473467182,0.5857110052871067,
 0.32417924224517747,0.28739837404591506,0.04244141434832904,
 0.04301928113338249,0.004872752325454477,0.2278524494445482,
 0.040939089721888644,0.0058950273282927235,0.6825186817706634,
 0.7785924308610332,0.09934689408651298,0.5577341791702396,0.859841783702491,
 0.4278341992725019,0.6504861848003847,0.5232200429553868,0.7054189748676755,
 0.09321360917603039,0.18446760489367897,0.4393480628560368,
 0.17015151559479513,0.6095790832647651,0.28688115447337603,
 0.3712215408254337,0.9111247477569275,0.5548717114459477,0.9200632218051982,
 0.46120851770670357,0.5099117179719871,0.7029913549343341,0.6853271665155625,
 0.2948262251037084,0.04978917898195234,0.7353103979703882,0.7812697800845793,
 0.06937188870311817,0.06092735618663603,0.2651608899417446,
 0.09589774390631767,0.12630592623962222,0.40605882791845216,
 0.9893714398602492,0.9501424130378199,0.13026058364596782,0.8241144168158939,
 0.5028478567486758,0.8020709337452214,0.4986964000820958,0.6636885617003548,
 0.47492506525965716,0.10993107745751063,0.1815521097449314,0.411864850308608,
 0.48941721038703234,0.09889069505805126,0.6439319132399423,
 0.04259735114697527,0.8821803052995328,0.3923848154283318,
 0.09935362851318175,0.45918637303239573,0.7430609678710769,
 0.9123438991699677,0.2554195236960869,0.4703495744648968,0.8148933574645186,
 0.9471004542137461,0.08804003232724511,0.7114305223427515,0.5189845028084541,
 0.14012041827847033,0.6737657777704151,0.851419227759555,0.5071287028795965,
 0.6086587625969292,0.006022688557963507,0.5245937091157513,
 0.3514799255435117,0.13839081759353067,0.7454211163554351,0.5844848699448134,
 0.5389956006351118,0.6732822753258799,0.4923442705919323,0.7176926266386536,
 0.9951086851302099,0.9430565579295412,0.6571598803901139,0.9596642706515762,
 0.2880679592963018,0.6226369608936808,0.18521187041256582,0.9033634314348536,
 0.42208232168618376,0.5223739067318156,0.6048746496645373,0.6198933533246509,
 0.16008980193775169,0.7918072043774873,0.910750996980073,
 0.059789584695218445,0.6296011698206786,0.7425264883989121,0.938128140790522,
 0.17924194987248732,0.7250401826623476,0.982616993434443,0.20766166759511162,
 0.7695093750374568,0.40915841302347455,0.44963023532960067,
 0.27116055862959854,0.13670156997303795,0.4885304597040535,
 0.8915977712419636,0.3634746171672506,0.9710299726244241,0.08026282211180913,
 0.7659967019345374,0.5341969117928952,0.3347845310580172,0.8495008436570862,
 0.751366057098447,0.2735095836932383,0.9342909566329352,0.34352571922113306,
 0.9426760384508674,0.9469287531629513,0.573281988683978,0.7096286681633197,
 0.9209697556587264,0.5644669439538336,0.033219365959770264,
 0.7544696601227752,0.6680275577040896,0.24441187088408,0.7383601025434285,
 0.677999167771414,0.3573044960116736,0.16057607594252632,0.7897853628574573,
 0.9350756783988577,0.19675855852289292,0.9079414799809514,0.6831454675919393,
 0.9557248079490924,0.7631226557757654,0.4729958989585292,0.28753679360063633,
 0.5823817650198294,0.2756733719727942,0.1993853306753739,0.151376132051512,
 0.9572634753670446,0.4063668642356931,0.15011801290064408,0.7315937020579304,
 0.23129729690945167,0.7277276794860357,0.5930208743396364,0.6584235751480129,
 0.27482241463825074,0.009742308766804486,0.1953197602742064,
 0.13355357568906556,0.5288084728913693,0.5266533775790351,
 0.43406362236681884,0.7790089649788752,0.8983114277133469,0.2438705132071235,
 0.10318294263608407,0.31829293184168206,0.12111765039081634,
 0.9329871215346465,0.037843215414321674,0.060105295198924824,
 0.6485599890455951,0.8237191210339505,0.8003812820127225,0.41849070643017816,
 0.7488397234318954,0.9486764863638704,0.8022250911786425,0.592350304210042,
 0.07016518154890938,0.5179575101467703,0.4625980791613935,
 0.43023947641136306,0.834861906156944,0.5496353735087919,0.3008658214433947,
 0.7770645512949337,0.7187219855066194,0.7977405479347247,0.5984945759387967,
 0.09025231552373358,0.4710885423251896,0.8809465762210125,0.9981745252117994,
 0.8350487009605897,0.3757749299286043,0.3888998751615348,0.8474130890278339,
 0.9355938754970408,0.5116894785855731,0.4177535748025196,
 0.020172321438479912,0.907156031453128,0.9851748078564448,0.9780711323733029,
 0.12814649965918157,0.1425358398184826,0.5113177889722161,
 0.18465585815458607,0.3634602881357418,0.4221681430022072,0.7465845264159087,
 0.40529756833819375,0.2579083086298912,0.22651339016071803,
 0.8951758697265236,0.6643001137074233,0.6146820440344091,0.5437160204117885,
 0.9873934889559758,0.964119354569475,0.7319473705442859,0.4170320647913509,
 0.4670829425729175,0.8527186014944734,0.5360355425854098,0.14748703391012463,
 0.11668959094978248,0.34632888905460657,0.34238559354154785,
 0.2548242980129489,0.1294632599823199,0.7296915603690459,0.10311858013499009,
 0.6159277754273477,0.632988692886192,0.33117041452912965,0.7588482308134037,
 0.8249090270762491,0.9648833094129617,0.2571269840237004,0.3006039490010993,
 0.30057493633332366,0.4343180231249034,0.05347475646566424,
 0.07721392447741837,0.8425495119804889,0.2755248155716705,
 0.36961164253545586,0.495917162908481,0.25469600533932024,
 0.24064930731206602,0.5095661886400938,0.4160985521838334,
 0.12175534132735177,0.40436612225325697,0.22322113513247965,
 0.3112422220425072,0.33937920416340606,0.31752076086750713,
 0.6586930298175481,0.42870744213971146,0.4850259734723079,0.7929909970996718,
 0.25501254720924416,0.8850559893593717,0.943627234764785,0.33274989791210197,
 0.4848410221769004,0.07157895884843635,0.10814347800831625,
 0.3964185848554096,0.10790687844793267,0.5143228398168582,
 0.23685460727122531,0.4679571668120861,0.01905961817354984,0.514703459998838,
 0.06045458445982832,0.17289697200142917,0.9968660349055068,0.783885906449472,
 0.6181692212694845,0.6413463357272172,0.1822259768251907,0.3504896557294799,
 0.3357427316758437,0.34102279301734384,0.35630612211502044,
 0.5378055627115796,0.5918994485911937,0.11790546980869843,0.8005065184649012,
 0.19874614734294538,0.2004309522019675,0.733654976564434,0.7808655651191516,
 0.9466545445119312,0.7492133099459612,0.26702382437929917,0.4415633505507409,
 0.15211067221969166,0.6758509553260673,0.9247312007297833,
 0.49956737412157004,0.9449930193743262,0.10570038753087119,0.462991961840577,
 0.19679967066756543,0.8175718999038785,0.07099766359856752,
 0.3692342772141717,0.9547701591377952,0.8110605592759468,0.4547380116173917,
 0.1815915181814307,0.9624268686877012,0.1425781616055405,0.0546229664783251,
 0.6199543525658109,0.12598479100112459,0.8908648749439219,0.4512925306084855,
 0.6101293705347357,0.17953296884788394,0.6764146751527783,
 0.44798784508190004,0.7358057309530726,0.9653576822481873,0.6319575781596515,
 0.42220038396297055,0.013213318679202946,0.19388309542314186,
 0.24183943175285705,0.2954556153768447,0.42711284860305376,
 0.36245192048359465,0.9926976293586243,0.7608760019472407,0.3474901978020034,
 0.737540305124236,0.2200493310677586,0.7564733772735232,0.6543474424936637,
 0.18889308484951228,0.04047747066147411,0.7794475931122804,
 0.7530494272603756,0.8369075184148106,0.8963393280732556,0.8386771806554048,
 0.8700175097264844,0.40273704928287724,0.2058857197865922,0.8473167430576587,
 0.5271223542159137,0.9183153033494126,0.976663044538978,0.444111685308199,
 0.9406490564051437,0.959107335838593,0.7944590287331347,0.6369730859309565,
 0.4649245563490899,0.8540926430870796,0.14036577328197597,0.7647653355996095,
 0.30662816998645126,0.6030235059487156,0.819080655827688,0.9852636700822278,
 0.8442891621594966,0.5451373823465894,0.2656141019207804,0.5620768135508776,
 0.5342258171014407,0.5196147708670498,0.9745317420303201,0.46205230852187695,
 0.8692553560746942,0.008400873136801379,0.59620920658958,0.6926242487390875,
 0.08243821989872335,0.9094286385071267,0.2307058837315582,
 0.059616903156216816,0.420442782629198,0.8182551735915624,0.7186308799285369,
 0.21809201805982037,0.7426610413041802,0.042717309127990544,
 0.5855886402664218,0.07430307827296867,0.9650469127594365,
 0.04526672480344229,0.8104409813538899,0.19382449301511362,
 0.19086239780706815,0.5140493085457121,0.9387279332118994,0.5072228262804501,
 0.438780682076484,0.6461700942192834,0.021024787828091696,
 0.07504176920128891,0.3416638638259508,0.9367806499814941,0.7996089591853683,
 0.6059916436337396,0.7364054163693048,0.6836017802630173,0.4252256268385394,
 0.3355949125635562,0.6812046465635682,0.35673845028499396,
 0.40825286617644996,0.20741914589584487,0.06576435291434679,
 0.05460445691340543,0.7332144741996481,0.3969367987165826,0.3197708269183167,
 0.4089110862880384,0.5128169988523389,0.8731915352627921,0.17503286497054615,
 0.9755001983846271,0.8607208316734594,0.24935548169982058,
 0.11537076193323237,0.02850180838569727,0.33059367524427297,0.6067268367363,
 0.8420270952264036,0.5100277029891033,0.8221850486688683,0.744050130478473,
 0.44368254934952056,0.31915756142448726,0.42181753835308033,
 0.5954165842510212,0.9633013801159329,0.37535624141015445,0.8159621167341495,
 0.17253655261590217,0.3417008925820362,0.08993257957005008,
 0.21688899560780073,0.13379805647950027,0.07975712881480235,
 0.8674359070783516,0.6986128003178179,0.5105183402498142,0.2792182437976568,
 0.7239740825036711,0.6095269431388873,0.836814464604447,0.24999015298689709,
 0.21841608541523083,0.2689829967148939,0.9992798943141502,0.4552218854902843,
 0.4774796189165651,0.8026456457658261,0.03706725438731673,0.4564788960585865,
 0.734924226028192,0.38650597530024644,0.8285330381329008,
 0.040636602548432266,0.6951615925707691,0.29492092825054894,
 0.5657436587563835,0.16657431021956048,0.2897761150614989,0.341384012329381,
 0.09032525266559333,0.2956352726747171,0.8328972237318099,
 0.002685117296591044,0.08607339442518991,0.47780404284243483,
 0.3257899106679899,0.3686030298246261,0.22474528137623595,0.5479285773716642,
 0.8617099892425281,0.5120056830782538,0.0049788797136619545,
 0.06866272621591107,0.6631865393877173,0.7097768383115777,0.5489006077869175,
 0.13015233745045307,0.9085217790661575,0.8732784456715406,0.8208864900248568,
 0.2554790302393799,0.3603554738217699,0.3355888928723384,0.3591392494510923,
 0.19424336901641226,0.6752519975358745,0.0453181045632155,0.9757086799940764,
 0.3096324523822158,0.20593839149049464,0.6329153976868863,0.0849778386116884,
 0.8810738237867503,0.5854759060737086,0.9530992744435917,0.13558587325585136,
 0.7497076743785592,0.0674842122662227,0.177463798054725,0.40250901139305406,
 0.6233742163858716,0.9573182094186095,0.6965422031124203,0.5862364301125899,
 0.5502987066555396,0.7562908870680936,0.953675941226464,0.3124059523115402,
 0.6121002888876996,0.6255555452068307,0.5531821173320006,0.7142317787376103,
 0.3888973178090074,0.8357417270305201,0.9992033585819166,0.6882014068854607,
 0.4867488804858633,0.1799996816033509,0.9264170302626177,0.20724167351264433,
 0.654863898845721,0.7383738182972616,0.41824606661636954,0.8843252264485354,
 0.9058818452305167,0.6147076249250534,0.4556799893071103,0.6626114235226161,
 0.8096326777761494,0.009158622644723002,0.0241509123323721,
 0.3300529083925464,0.6426937989657509,0.8578420563382795,0.995568785200737,
 0.2422707666164744,0.8094510619204519,0.1197719531871656,0.0864982251214762,
 0.8090297109975244,0.49002791603967844,0.5929623785930138,0.6580390321583063,
 0.2849059333430197,0.4852413495238126,0.7083565514932034,0.24178727517909404,
 0.40025551070789644,0.4017798342666552,0.5750623644570672,
 0.09555052745636472,0.19073630829551758,0.9587825460535047,
 0.9349182889795101,0.6955216866394567,0.08951981913922324,0.7284970888523414,
 0.40579174074175306,0.09051438657033717,0.11661575113020672,
 0.2829235021683433,0.8142075386368122,0.040246210874429145,
 0.5166890398663934,0.926148031305554,0.7500784648188311,0.8226385824781374,
 0.5390725688799791,0.5034540835113526,0.09968808490958492,0.7400110041181147,
 0.7107075982245499,0.4223784488880099,0.47953825954897944,0.2655235745864196,
 0.7491980242592196,0.8674120060584627,0.49229349698520697,0.0960317872015154,
 0.5087897774903181,0.2748636263343386,0.588340733531661,0.08983280263977012,
 0.3522229248611406,0.9867812001322723,0.8238002862948626,0.5148356321462785,
 0.5207337305123886,0.001451805298110398,0.03011261367444662,
 0.8145144560489541,0.5274824708854919,0.3326700591685944,0.7493300186799298,
 0.9864694673828098,0.02232159184793625,0.28853484025133735,
 0.3301081070925105,0.25411654569667386,0.3690317700072816,0.9313297775434006,
 0.6957236957477653,0.7650890855262997,0.7297460894416388,0.9925198122331026,
 0.5416798943294143,0.45366225464934784,0.6866550239280694,0.7933914509074399,
 0.5699511117621421,0.8931984996301537,0.24441346120873197,
 0.27549801095776105,0.6346507557672966,0.12765579158023976,
 0.10907158548543605,0.06498309186630791,0.9514989699641648,
 0.4400392551238742,0.025853035742123387,0.5936302833129959,0.703254332108244,
 0.3924689969848192,0.11206411352957546,0.6373632044430811,0.5709220656762546,
 0.7737871534526602,0.28724220850984006,0.2591950031058704,
 0.08697645053824632,0.5946695541820418,0.8974011699186636,0.973669190002489,
 0.9592032998340444,0.6698231151468084,0.018294934629164983,
 0.2677153217498903,0.9496933809569205,0.9018209874049334,0.524781190191085,
 0.042083302182692184,0.6357478130630531,0.3579044372172817,
 0.5722442653173061,0.8265571775117563,0.33206030519340357,0.4835884212858732,
 0.047429305198578464,0.035554417978695074,0.937996766504488,
 0.4736059131880138,0.011422528218387629,0.24744669413452192,
 0.12490761161946884,0.08799540046143917,0.8794311875688728,
 0.11229833123455246,0.38566277150995987,0.23112722552157372,
 0.6298820710755548,0.537445529219485,0.4477401155770093,0.5579751718831978,
 0.18168843171535687,0.7177566057303719,0.7636163389365966,0.7232695229359841,
 0.39372984689075663,0.7960498903948553,0.34968173082118503,
 0.41107542176445944,0.5954254897348891,0.45097826931800955,
 0.6328468182830793,0.7367084518939457,0.007476923712996353,
 0.6305078096250214,0.3309166182484211,0.7721234037425829,0.8154799226365552,
 0.9380512534552358,0.634541383304554,0.00804884188082211,0.6117936079752282,
 0.36371560355226495,0.4060803902008365,0.7580703142259881,0.6119295402964574,
 0.41359744728683123,0.46810012802809564,0.6120477001612439,
 0.10080608136336322,0.7230380928552318,0.46708938427530844,
 0.20196280513316545,0.7629511188181094,0.5244887057637011,0.8126816055713261,
 0.1471671272783619,0.743328298247974,0.5504150485002476,0.6053351167137309,
 0.12062878065303828,0.8127673316679485,0.009037572541739536,
 0.10984541628594646,0.9506903099137356,0.4930805644012459,0.7377265236914905,
 0.5618885667517741,0.74367949697875,0.7236270501122183,0.8668653993519099,
 0.9647578545895864,0.6341065523853457,0.2715389879451727,0.9996768630883305,
 0.8515689866768064,0.33828975503670355,0.49272149451952385,
 0.6832078393294487,0.08234173718877558,0.23150122047368582,
 0.8247007083648293,0.40282167861465057,0.08609161878051785,
 0.05765637648668909,0.569352750556895,0.12812124599590935,0.6472591792387701,
 0.8909488308955749,0.7849016384974221,0.7376511425402191,0.932501297108763,
 0.1284633752652109,0.3121898432459329,0.5695791625350828,0.04172645921065232,
 0.391864527534446,0.6421620289783943,0.5600290638060885,0.05378465128874532,
 0.7438831797237142,0.47821302981503766,0.06203616126266087,
 0.7762099449753738,0.5449028636233783,0.18049391930399172,
 0.42861254167935636,0.8573241077171846,0.4990952574874419,0.947115177236723,
 0.6393611234201155,0.8654931518159455,0.37176531836506366,0.52128779540781,
 0.7994950781562357,0.3845127186797552,0.7435032444855065,0.01932478139540461,
 0.8782279679753581,0.34670296320509486,0.2544737189280738,0.9008311426581244,
 0.5276577655293778,0.14913935509132792,0.12527201685586586,
 0.07743852045396471,0.06271309995694718,0.4901591317969475,
 0.47333818496351465,0.32077793556204437,0.4918501390664375,
 0.6581947445785239,0.06410912420772508,0.5706788405619985,
 0.45240046094112774,0.571549753667251,0.6429313240566469,0.2788725661024103,
 0.031283045955802224,0.2843465381761753,0.12983972669122368,
 0.22799039973695712,0.052002181313091755,0.8211961255187795,
 0.8385261958159314,0.7885285545939885,0.003000700263894207,0.972742193553208,
 0.6187197325607704,0.031077799677677875,0.6031797835817965,
 0.8056195124081661,0.5388253412659741,0.25636107782382145,
 0.031521260757770175,0.9431289877628606,0.002797756147934982,
 0.4607669049751981,0.95621917842406,0.39022637466517385,0.8216206323900063,
 0.17627847336875213,0.009905150899681137,0.2056594868263133,
 0.17517360803712523,0.6977529763250201,0.33319554195039136,
 0.5889444476658229,0.8363474891867695,0.5480633684969445,0.6501454063387511,
 0.07847933516054661,0.8496939876921993,0.5927040252344352,0.5889124400310265,
 0.1512177367581793,0.2870871620729625,0.27469587845047516,
 0.24177551396901464,0.5826106783963201,0.25459695887767353,
 0.8761365644788548,0.05248388387371905,0.5453929912665687,0.0430985469844527,
 0.1930537614814194,0.013830975068448215,0.3991628468306162,
 0.6325795610077279,0.5827728523696195,0.7967740258117575,0.892026360365292,
 0.030195075093611967,0.2734949208897426,0.004099100386817045,
 0.9664151470321609,0.19937365657543316,0.2909429746708473,0.4359946028651325,
 0.8164115284472376,0.7926948596817719,0.02770106782072368,0.8458615337896611,
 0.658243323961788,0.4891676832109436,0.5031194621917224,0.9868509706834191,
 0.0943209916383182,0.48007099445561185,0.6101027819169812,0.5599221349264504,
 0.5900396271834636,0.271386354132655,0.6948887600531285,0.49292490430481306,
 0.6504097229908736,0.5874720389651974,0.11997882247353597,0.6600086634477029,
 0.24877389959302953,0.6669425611240862,0.5834331114320532,
 0.16742615795775384,0.9632825158172407,0.5141815296618304,0.8803454166601111,
 0.5862917871584856,0.5459766329570684,0.6066732705757162,0.15765391329289946,
 0.7276773724130253,0.8864902067795952,0.04985254449698928,0.9833245524893702,
 0.21708167648195653,0.6759937681024657,0.8642528944396835,0.7497164494228266,
 0.741844093839814,0.9522536872244569,0.6967997981803026,0.9727487534957738,
 0.1322248919349559,0.11630027074963456,0.0018891042090605215,
 0.3949634996188991,0.2818213027621993,0.4022058240733398,0.8311259394396139,
 0.39060048628074096,0.5335735744755259,0.04233156737426491,0.217640918396378,
 0.6211679233618602,0.936618519011297,0.03842229849031109,0.11740782594229116,
 0.8334243106909404,0.9274736865257686,0.18943145984041732,0.0709852720872789,
 0.018733207376447902,0.7256690808581309,0.11759573507712184,
 0.44761153145696264,0.45638933639499046,0.9855803597639293,
 0.46686438743736935,0.3036334852395701,0.6483198905891636,0.9353077639204206,
 0.6181186581964788,0.0924297164196971,0.7289900814815711,0.4230279510695324,
 0.3512585279041244,0.3420999493748773,0.6625159296258002,0.10786706592507844,
 0.710862483568338,0.2484969461994958,0.6304363325899033,0.06097208191646075,
 0.3389027145170055,0.9263720046436702,0.540193783538597,0.4211099628990135,
 0.42207752056416115,0.5717758411441322,0.6042032550159651,
 0.22480818202111263,0.4869419441586712,0.8243910495481828,0.7685029466445843,
 0.41466855711952955,0.24549178439565356,0.932934066905117,0.9805517231409244,
 0.4482753827818676,0.7138927100457627,0.36161303769374176,0.4612353662236012,
 0.7118132027964612,0.29652079043644153,0.630796514982098,0.7162941390118052,
 0.8910404284630032,0.07653969534596383,0.8437854898900136,0.8719730346588134,
 0.8439949226636141,0.8625183542062345,0.43100310597800406,0.7387971648259872,
 0.8907258286376272,0.8836458911258608,0.09056867088314491,0.6663633146089287,
 0.1687553689619291,0.9964650195758757,0.7124823235270487,0.48977414411432785,
 0.6952246556304768,0.4525196721501319,0.46202732704279326,
 0.45081658687373105,0.3877260488447952,0.3189836098345139,
 0.07227527568831826,0.030720799085385164,0.7453176415465451,
 0.8162115685864202,0.8717943314240915,0.5970125931250453,0.07791724939551248,
 0.41011560625671295,0.04291478233350687,0.8819985042559167,
 0.2043852334952585,0.05665745439444447,0.5169914163540189,0.5255752700970413,
 0.4933693671672241,0.11738261301661546,0.629169893577114,0.6495785085038431,
 0.20255360813711143,0.01564227072348434,0.9309906748327001,0.821020632371211,
 0.5120506339837303,0.9452157194587247,0.7188871510471964,0.9459295712151033,
 0.22316656636362897,0.05174809500735178,0.7912642040184285,
 0.8135445071687065,0.17017579517356551,0.027443788906226896,
 0.17680452609982222,0.085382489972189,0.6792764507957427,0.9034420205077304,
 0.049178754981697104,0.08105497173937104,0.5066873117445542,
 0.6718579765741188,0.9443977038200073,0.6448283413100994,0.35669266045358805,
 0.12804016045062494,0.786742617316709,0.24936313718173608,0.5556464485822115,
 0.7751879153272814,0.8160988548575608,0.8697004505439155,0.14221095122368088,
 0.9498434621046566,0.1979471057845198,0.1400541651871463,0.09139218316495734,
 0.9053125302007313,0.18508662769076678,0.04615423742611591,
 0.6184190551196561,0.22137503441588657,0.8854337164580461,0.7642065631633284,
 0.789889478645525,0.5647775352997729,0.1776727143416369,0.2406197468472323,
 0.8269027636783373,0.9566158893810033,0.6631662891356891,
 0.047515638679078265,0.014023964182563797,0.2480679158404575,
 0.23082980200446923,0.7295346394048934,0.35205065031392435,
 0.9586628866759956,0.3669655198830658,0.09539010719388041,
 0.48381130712081866,0.4311140693523966,0.3797085867744072,0.2545861894354945,
 0.4109267437342279,0.28424607994222173,0.7677854425118631,0.8817870747844725,
 0.3839986053251092,0.6524614399505213,0.6789243677041849,0.1325470192591891,
 0.7185553722074617,0.25777175097396476,0.3188290816241752,
 0.06441346213301513,0.47240483505792985,0.2728861324223376,
 0.33497484664073274,0.007159468693336946,0.5142368661281069,
 0.49001649805208425,0.7136360664435273,0.24467122889888804,
 0.42735712937224235,0.7694370017924469,0.5736338905115603,0.7239896401076122,
 0.6485784030678073,0.936798348191647,0.3631453977783293,0.248059904539512,
 0.7294251644114325,0.9660011189521032,0.043662483024294985,
 0.08267092474498927,0.7130818513049458,0.8858668780613295,0.8764759214329072,
 0.5918453282826573,0.6916616083347187,0.1544429193605038,0.3762005714863822,
 0.7732445858149747,0.6805648135258969,0.7799127896719992,0.4731305004798376,
 0.04309488416434548,0.34356949672344106,0.8742062719835775,
 0.7200433907681597,0.9758658361274694,0.5019784643368591,0.9426941100295658,
 0.06474640325237069,0.42076430495285533,0.17686214917201215,
 0.9045465462549762,0.03365336860275292,0.0010993944619562823,
 0.8032898254308781,0.8661540626097518,0.6217443154845683,0.2660840792418214,
 0.3442389296740872,0.41533869600516926,0.9245792983315503,0.9737064129567126,
 0.5749349767389027,0.4486435115832643,0.551948517404098,0.17535246410959016,
 0.3131644875068954,0.09079630132377148,0.27914651019085046,
 0.2651418011399048,0.6689472784671984,0.9899638346483078,0.6460667141837705,
 0.20981418604143776,0.7202395586897096,0.04349970380424495,
 0.9882084664900367,0.17536343730710957,0.11405980606606614,
 0.23977659957177389,0.9145900799170995,0.4773745519485487,0.5161940928798141,
 0.6844624132357707,0.07484851035820272,0.5214712944181038,0.9235837824894235,
 0.5633595136710388,0.0995940259944541,0.19715082152200247,0.9571292495070028,
 0.48100542004700364,0.4266830284811448,0.22521045427725972,
 0.6447303312364021,0.8982662681975924,0.10207898074597255,0.9141931250556062,
 0.9212352304519289,0.8078531775647552,0.8875998636358342,0.13947452947340178,
 0.500626525451634,0.5539892871467114,0.29674001779035786,0.6018077688704911,
 0.7937353422908053,0.9800689649224612,0.08587741598999321,0.7935050025024059,
 0.018391710581939225,0.9175341042679683,0.23572912822289482,
 0.04372702884900892,0.30313950189765626,0.5768320497680307,
 0.9367838397619747,0.5886586772235168,0.6498063951987768,0.6762101614308084,
 0.5624273521635573,0.020416139473999673,0.49459803156313875,
 0.8876569971272794,0.7787474936894644,0.16930829755551213,0.5058259355251349,
 0.42732796425866093,0.6752462623991174,0.5144722483321278,0.8241408164262247,
 0.5357834278600794,0.007798626152497046,0.8852105160634076,
 0.8841825506330974,0.9237034788309142,0.6488366455908133,0.5491182402680477,
 0.7897205725886673,0.8102372830742902,2.4621848486261477e-4,
 0.7140027289346748,0.10775736953952075,0.021009440202579444,
 0.8280604269623977,0.7097375587175296,0.9968881500268008,0.13187077266317127,
 0.314398100484669,0.6103804300329957,0.07274200843970902,0.06925887920507923,
 0.011375486329897155,0.11597144636342582,0.1305102735193131,
 0.5653512288056018,0.9382142472541576,0.05652722250817399,0.8719058495484251,
 0.6606903017135871,0.9602116450815378,0.6334024256318506,0.5220925533063133,
 0.3641628863402905,0.8416807815840917,0.2550017640099139,0.9562698919153024,
 0.4942714298878459,0.33742747106823956,0.8187715184834022,0.9560769625394786,
 0.44700131854357916,0.24574192788614913,0.7201967549974734,
 0.8313156396712758,0.5396654468956286,0.3740108789587758,0.7473605329332477,
 0.7764424263306048,0.7397372038880875,0.7265010266055126,0.609348189176705,
 0.9358960791672996,0.5556434539586365,0.3255194340987352,0.8562693983370314,
 0.20691485866509074,0.5807399383984858,0.061991046090842494,
 0.9803366249533614,0.692408172741045,0.18463515356953453,0.29856283233206393,
 0.009284696820190952,0.37555920906307105,0.1294580123244189,
 0.08585466782398243,0.8044433060362336,0.755166034329883,0.6014361705484885,
 0.010729205538604347,0.6858679666758705,0.7286078829308982,
 0.7769202097911465,0.48799645815681103,0.6996738805538951,0.6718455707851352,
 0.7261207393797279,0.7582533742705928,0.6451800596474038,0.3745095709892391,
 0.7767268951183469,0.14285727476753685,0.7914950524463932,
 0.17164976773813012,0.820596188644628,0.39924808600029493,
 0.25247914071606326,0.7726309422136081,0.5438251200168536,
 0.20186663057854715,0.5350866902463844,0.2050275486182016,
 0.30732189494820217,0.10331349863689077,0.39850981707556066,
 0.0690466731481002,0.30810346467676486,0.004009102069095771,
 0.288115480535005,0.9523429915315064,0.4349984435925084,0.2938138063131922,
 0.4949955735364864,0.8752294409438285,0.18214625021637987,0.681981723786627,
 0.13493688133906723,0.5232525634549995,0.73851501111131,0.8737329843414519,
 0.7195688871196021,0.998847244036387,0.04948739423260895,0.5098289453743429,
 0.08965508500000585,0.5367710394803866,0.2458696764890712,0.7871756775085379,
 0.03315355575586698,0.8053672283913053,0.01950353237011515,
 0.3929672647027491,0.9095299241386007,0.1366110364552101,0.7320044846723703,
 0.016706767708347003,0.9834299482304965,0.44375632219573624,
 0.2880399652467204,0.02702506349861411,0.3597328120888539,
 0.13445628135481058,0.9019218243964154,0.7583231086886465,0.5796754295812204,
 0.239316514904091,0.3430141784182057,0.4570444277511796,0.886912363390659,
 0.12291923045583208,0.49531886629744826,0.5907608607554797,
 0.9562442902120186,0.8016368183062835,0.36963278509246855,0.6173429127699174,
 0.58380157764936,0.5327121194271236,0.09663154774264426,0.5867915513745274,
 0.03256246099932214,0.7481954930707805,0.7688935035636405,
 7.038523861361279e-4,0.2699889204355208,0.5018580730773207,
 0.6383024386481007,0.018082829158515557,0.3808701614440334,
 0.14232429426186832,0.2566064003882004,0.17607971719929982,
 0.5213508132041675,0.4588279838488387,0.9531708282957355,0.6652394253677691,
 0.6926280664548258,0.21103955525807105,0.6587980648666519,
 0.38343381206357174,0.5984452556651417,0.9986714439364254,0.685548442661893,
 0.029489045462114838,0.553141076486205,0.8596388962065649,
 0.35153649890193084,0.9355151830188602,0.2122680342476274,0.8655745344661324,
 0.48810989839721053,0.16776840890937061,0.9018946205718167,
 0.5638548445281617,0.4491939228443578,0.1569215421854554,0.6877333924633724,
 0.5021747722988386,0.8593365703822425,0.653486690772531,0.1657198856052189,
 0.009703641136314611,0.011930966227519102,0.5951399711780511,
 0.44669421766679773,0.21093953866418869,0.7914521771857135,
 0.7024642482490088,0.1318606465675478,0.8792531171239406,0.9375671410430078,
 0.7112917910689878,0.9271679691528857,0.08854756250964935,
 0.38343400830397223,0.009789895925675696,0.7299347179792549,
 0.8139283674715954,8.535838888223957e-4,0.3921902289481417,
 0.9863805948271921,0.776878300667349,0.936373419347489,0.8011338341723458,
 0.48198931022982683,0.13558476725506363,0.3915157613414233,
 0.6735351797379819,0.7290133160353232,0.18450159230457208,0.8240100402080051,
 0.3777149084153084,0.026827391591271454,0.4070458765892725,
 0.6975743981325045,0.6046653235629463,0.823054830398605,0.45746302448808984,
 0.6920173275244802,0.4746498326375551,0.10420129446737192,0.3839287262733969,
 0.1000094816469711,0.7665045843726999,0.3642491697095659,0.4623727849981055,
 0.8052729740182386,0.5569742601191059,0.1586811791560433,0.6856377092020556,
 0.6642682571132287,0.0863230302024045,0.9903882257742388,
 0.009366079894710277,0.9803639613996945,0.1546091218600063,
 0.019440734444617136,0.8432630430555497,0.9533753437497484,
 0.16140053505418672,0.3643205881098345,0.03902970751261847,
 0.3146432287457229,0.5142447960040365,0.5483426667631337,0.7465701679735267,
 0.11296089246248786,0.7689244944871272,0.8015560880925375,0.026610663261043,
 0.6670818634480518,0.6795755852363404,0.8993845333891599,0.06268805407677558,
 0.4069154250119874,0.28601614064313663,0.011298862849321467,
 0.8449119135505212,0.8772064028579666,0.9631282791519382,0.13079815189219957,
 0.7286430847322662,0.9410928271650946,0.23762117302668706,
 0.20843832336784807,0.08824766225666214,0.1751695765289224,
 0.5145509013629798,0.2772077671830627,0.8465740052055494,0.07978433766812199,
 0.867747432844546,0.5619223850175759,0.733773406987217,0.9630782403086457,
 0.3530921243666747,0.7653709825527396,0.8004863580645406,0.06223902652694824,
 0.8892907906687961,0.41428080455171323,0.13814077561217297,
 0.6737460533759294,0.484556225142728,0.995917542443975,0.29691560245468807,
 0.1495457309022442,0.7239679360568698,0.8465337044297767,0.7754792812840482,
 0.15945132803249473,0.6063900410103761,0.822218856879256,
 0.033269006277334756,0.7165841486475109,0.4423527421295854,
 0.6621285900850444,0.8109222554897787,0.8462452916570844,0.3703844089858416,
 0.7643253725145277,0.3621061719573272,0.8707462563726115,0.8971358523225561,
 0.7799692776947343,0.16018489266139824,0.04926979846766044,
 0.08542663963816277,0.08153723241928801,0.3571595742597907,
 0.6189088216698884,0.09894316153833838,0.9450801566292172,
 0.14094471211797888,0.4870225329549829,0.28112887030998246,
 0.34607744728789425,0.20759099142017867,0.5300742870949617,
 0.32386666971603206,0.564742391661128,0.7511411774000363,0.8904593584527669,
 0.6249610148117055,0.47762043632063134,0.9968022904497696,0.3863278680518635,
 0.14014444423889882,0.32311654870176265,0.08997480850521034,
 0.2821155398612325,0.5778061429447429,0.7341506868555662,0.08658212443955227,
 0.277391701303064,0.2554940749102522,0.25802444516631096,0.09795515423454415,
 0.6590036029503745,0.299198044635414,0.01676631035630205,0.12960622301714175,
 0.9467543735493799,0.1083958115735304,0.7623840639345505,0.34321144281820004,
 0.30014594454197463,0.5839770529073294,0.4359498524072627,0.6121080486657338,
 0.8337003587462202,0.40785074605959204,0.8139927795322685,0.6676253559585388,
 0.9897073239759114,0.3928615468813931,0.1946296150985436,0.6315315681559257,
 0.5873049579315863,0.9767441959810589,0.232114109169412,
 0.0049717190100230235,0.7474180523343523,0.04898537459662011,
 0.8343906535846938,0.007941598123913263,0.38995143223520345,
 0.34397319782276015,0.7799428992514619,0.5091898447815717,0.3844915799005121,
 0.6352051760672311,0.3112915024196583,0.8463972361428136,0.24644751913107055,
 0.262272131660088,0.547195250871464,0.45933544750741395,0.49821954163273774,
 0.2999090719291111,0.30654790693685885,0.25560340644040436,
 0.5275897011991512,0.3010395308170358,0.5860849406810528,0.7919105486145401,
 0.802540618494962,0.7800104178021615,0.049435963063952926,0.6198527267640159,
 0.9876153979910026,0.3164567172751508,0.11799218863592043,
 0.10522057731935641,0.21090549529255231,0.9569647118599692,
 0.8131068916378698,0.3637379420254475,0.8514239145816895,0.6752210943939795,
 0.4713408220061863,0.3578125551585336,0.7862230243161993,0.8166430530430692,
 0.8025332689456028,0.6963722869206513,0.009743730300434739,
 0.25630709363276805,0.15272818278437184,0.37098174593190314,
 0.9438885837730673,0.9699502974167192,0.8757230484317614,
 0.031066892814327618,0.6169931836891777,0.8715868486573113,
 0.08473068124245386,0.09705079540923689,0.2400359501935475,
 0.48690062429220227,0.47070317957656704,0.7547728915744387,
 0.3705166600945069,0.631469623408087,0.28246698491420763,0.9074584947459235,
 0.9252032569678382,0.6314444959053692,0.9138130013764867,0.05455866844297397,
 0.8220319686635079,0.008674370600347259,0.3247845251762551,
 0.5019371652014581,0.6706985579670542,0.2812522884998887,0.979665836845928,
 0.09446126784110254,0.8769808056876914,0.31422418807803154,
 0.01930633263795245,0.9918179057856202,0.21201762787255762,
 0.0626379750240248,0.6138114868518023,0.8114636139289696,0.7365262052498007,
 0.42334571348086747,0.8018880629536187,0.5670014851336546,0.4976098641547695,
 0.007176142767313376,0.03770700515016134,0.8537148141258104,
 0.4985689943213597,0.6630684474334775,0.4738389841716787,0.09273425343134356,
 0.7187233609998177,0.7551684730534833,0.3888715607913126,0.8666346426773263,
 0.3102425516507008,0.8263272661545515,0.4716901642082868,0.41243656868208933,
 0.32842597308557075,0.6969277811160496,0.35775296574696325,
 0.4717202511980654,0.42039190935279613,0.6791114206718184,0.646807295368444,
 0.1888746703122829,0.9995941690403416,0.9897641655076241,0.8464267571008547,
 0.7432613794706008,0.275555990664924,0.3006012864687925,0.6709447601785442,
 0.4478527621314188,0.7787559951513968,0.6171538680421081,0.4803956688565728,
 0.426821026749189,0.9996938086274215,0.8743974091340548,0.08311129781466287,
 0.9215918796498841,0.446965568136689,0.9367759799667311,0.42747053921566525,
 0.526941088410163,0.534373372067107,0.5611513841666147,0.45111017421454724,
 0.499816172472595,0.06066899994599828,0.7228816353094019,0.9548644905066592,
 0.8591097131334229,0.1679885513510011,0.7020601770060898,0.9109902997313116,
 0.5856661643880686,0.9176622327665416,0.0933876764444388,0.858046133573374,
 0.7036407528318135,0.19139374707159473,0.8421805503156642,0.6962496000275673,
 0.6884285825552136,0.2807378537013775,0.24999250088665603,
 0.30630822643482836,0.3858911961040675,0.6828292028478284,0.514561177393224,
 0.9636359637419342,0.5391046111716165,0.2562723146888828,0.4506622718782245,
 0.8852248311109876,0.058343552972329626,0.8835747677023559,
 0.30527783279488174,0.41024629117370714,0.48711250118669636,
 0.11664646474104512,0.2420210441618994,0.924410130808681,0.5527851142499852,
 0.8868669446203787,0.38920814111588586,0.6106243728055561,0.6085309744791527,
 0.49575343103816927,0.12466858907127532,0.24320170524820062,
 0.03690496377794328,0.32930673619401396,0.30203772764781267,
 0.08664939030716345,0.2785943879953646,0.05914642456350472,0.844822947099188,
 0.8204727130059504,0.7037650192305477,0.054339866245685764,
 0.6726111141983269,0.2721247515263415,0.7303950918645218,0.5053208921712893,
 0.11040919784939507,0.12828196276084358,0.8265588603858509,
 0.2806752291758633,0.4464475381716194,0.7775691031352318,0.21389810736653136,
 0.7762763876292978,0.037902333077561745,4.1355814996113516e-4,
 0.5293294111296531,0.7104705604363636,0.45773179285788834,0.9399128914276154,
 0.16723157241135667,0.41593752785910887,0.19098225861181495,
 0.39749825810379524,0.03454231827121734,0.896159973744804,0.5575900340082217,
 0.3813985741477727,0.609523568246054,0.5368280336151533,0.6401485405148479,
 0.20574088431323367,0.5857242809043558,0.8057173277229909,
 0.18622072632932152,0.48498771748445113,0.16824409775635674,
 0.9079675137179725,0.2930295018741743,0.9218941544033102,0.22020181981389886,
 0.805532514980001,0.5104838773316059,0.005284492017385012,0.949564254984976,
 0.14248436570737455,0.9988178241897522,0.8862591632675731,
 0.47523162485264736,0.0721219814245494,0.12292641389451242,
 0.7570867727328182,0.9508372015986399,0.23563112104711237,0.6909039166231243,
 0.67314847207929,0.19221956743236512,0.612123278840283,0.3746370796380212,
 0.10639325841428304,0.39910380343608143,0.7354944225129529,0.617789477969166,
 0.7986589139381618,0.642250797696399,0.6285516592320003,
 0.0023403645042499033,0.1824642788240638,0.38166711721774205,
 0.36490363634818757,0.4615071147395188,0.9168120606729956,
 0.03609403641164488,0.8363072235418735,0.5891717477472905,0.7787634516350316,
 0.9597435373844418,0.5865202349484736,0.9811158277009825,0.43845504115130485,
 0.4241967500757077,0.8081687473552887,0.8852413293807596,0.7284445145689808,
 0.6892520108121161,0.957756233084275,0.5398842036413644,0.8486350786627428,
 0.4969150834335405,0.6905423922372416,0.47959491483718497,0.8281732496049745,
 0.47272817294060454,0.31953605685588005,0.8959033856122945,
 0.31641302405072735,0.7822015155710341,0.7980042816640116,
 0.38739684547007247,0.393641749063371,0.5717837323213173,0.6013060207301021,
 0.1579631357243223,0.2816737435620008,0.16949170782380918,0.7444443715101956,
 0.35229714616478003,0.15959347370905763,0.7958762347790616,
 0.15498989380043815,0.8930396676815889,0.021915548573814614,
 0.9880474750745084,0.8107190512369025,0.8929014283380351,0.8917181917925212,
 0.46550197339994615,0.8079801218374334,0.8015535129550133,0.7169439269072578,
 0.6279924818903262,0.2218997776446877,0.29341153422216903,0.0492986985835171,
 0.9556134905821131,0.8212136110752325,0.6355317503210206,0.522360438034707,
 0.16837621842666972,0.6782091123694804,0.37733053060435284,
 0.3280480714525752,0.8766252898287095,0.07178973242965714,0.5853465093622541,
 0.20958872755776126,0.33264093302541153,0.6925032742228077,
 0.9049091310403771,0.8295442945415676,0.4593105323520028,0.6249639806870421,
 0.7869308759137051,0.8155662872334374,0.12151345468479369,
 0.18570739274169035,0.5434969165824223,0.7491601278288196,0.4133536386818162,
 0.4248966494090425,0.498780302766908,0.2205570716985168,0.6000923405846115,
 0.4718784641527045,0.6151646959061277,0.012142306648014967,0.842674950737905,
 0.9898891874503197,0.9042312758075067,0.30703109238412885,0.6926979726048714,
 0.059026545594073276,0.7332620511951884,0.44123379322679324,
 0.5214876153615156,0.8382784119059586,0.5264946788983632,0.396714920663495,
 0.9480046394218127,0.31548948191837534,0.300871944650279,0.5311110795146323,
 0.7413790651683145,0.74611467591392,0.767244103004987,0.38179260696317674,
 0.5493531405645043,0.3327010492193233,0.9035657266190925,
 0.008273274729623137,0.2582470237997244,0.6445605503560413,
 0.11888925597277611,0.4633332469973669,0.48256833804442856,
 0.9427538935164007,0.02651131028193665,0.4462827416559264,0.6893898628511425,
 0.7520282451654734,0.053643212750060676,0.9511070665427898,
 0.12382825521734031,0.029433528454918312,0.6115001727927374,
 0.5549900720745696,0.5086666461383191,0.7572417884688278,0.9097021275628548,
 0.5249998362193619,0.11476483108914748,0.8833443292043806,0.1852210401423311,
 0.14761567508426143,0.5007459894003203,0.7158954460660324,0.7011001251797666,
 0.9820255392281352,0.9414789518323101,0.9629752435209585,0.917469381523006,
 0.7133220637368418,0.22357826394574243,0.18823460630152467,
 0.5324249847327918,0.7965069403273332,0.12340057737763888,0.5126982626621694,
 0.4725178199832979,0.6427087022397685,0.42928612005365996,
 0.18318071198220465,0.6971961816232729,0.32460371562672996,0.645856378809673,
 0.18685625903486736,0.9232103691237861,0.514959578453007,0.8211795081145441,
 0.9233383118532799,0.2386343911843115,0.8927636966191832,0.3738147577145964,
 0.3490172618491627,0.014467802089935033,0.3084951033050046,
 0.4667735414191849,0.9215936709644051,0.1918614144164157,0.634767285899746,
 0.3590912236476965,0.7025471566941872,0.9296788991411621,0.823428449471204,
 0.7374975491432877,0.2252121754556945,0.3747930830491911,0.07501277783703186,
 0.25583972687883394,0.5179942952061788,0.8151332166662327,0.9616641189835058,
 0.8100180312994847,0.019936529501964895,0.9591911266045084,
 0.08932367274324027,0.3087122167881726,0.9817889934682941,
 0.16694552176303223,0.6826245304042394,0.6062160356190747,
 0.31286537585372887,0.3431279444552432,0.5610742909919672,0.9349043405075169,
 0.15817262829675993,0.6863992096588454,0.6987486585492573,0.56272572554395,
 0.4722334494861774,0.7901002826076513,0.8542569016897076,0.2875199370315329,
 0.8528606367867886,0.2516641816000281,0.6275903116527263,0.987736115451702,
 0.4893401550654223,0.1327838833561199,0.5383655062218761,0.47138144637885626,
 0.3651239347201155,0.47056076634181154,0.8903657109301262,0.998131070120121,
 0.7100026670038679,0.22267986702732423,0.07556330047463233,
 0.11276457602481804,0.1569714898300565,0.2963221881365222,
 0.38324706717518764,0.4797717931182909,0.23348673489386595,
 0.2401284075870409,0.7858971680156286,0.38841547859769543,0.6521906987333952,
 0.3409405577306239,0.32168312480182903,0.20552796618751745,
 0.7303891064541911,0.16974395756459082,0.40110670677089366,
 0.46989597797449456,0.25402876197820756,0.39235307769792027,
 0.845804195839552,0.45306571580056654,0.12371062032391889,0.6472726910246209,
 0.8069118907726597,0.28932453263502644,0.385646968701997,0.26234047081759493,
 0.8452044937817524,0.7531753126914669,0.23694856108960205,
 0.28483751662849655,0.5931115575972173,0.4854224980250115,
 0.23857100468538917,0.32208238334638406,0.7284056658047022,
 0.5719533084511775,0.5204559463923385,0.5674732080639027,0.11586356816387278,
 0.6869520403353419,0.7358100106908176,0.6790115891158613,0.4154476948331678,
 0.22026969477216618,0.6208289189181301,0.7551258188627905,
 0.002148064854567311,0.9997129872395936,0.7759920914181909,
 0.5540802285274079,0.42573630274449636,0.7002057496685529,0.5871830284635872,
 0.6892584813865056,0.9915527929226902,0.9044945658516338,0.9599374459251827,
 0.9706167869761753,0.2968554458325786,0.05080649256151726,
 0.23766213368745503,0.5902207502464508,0.059655834022144805,
 0.3103766316895641,0.4042031946006457,0.548179594220296,0.9841600593856343,
 0.8611098058948763,0.5122356523669432,0.4684052357108246,0.4311798965533993,
 0.619370326016564,0.18775888175908895,0.0800159566715577,0.6944247908045901,
 0.5611526025238935,0.0811842142087087,0.8506856691588276,0.02600476153555642,
 0.06776037340192209,0.858055069569265,0.4035107378939853,0.8266583424445294,
 0.3630720223041045,0.09019696424121137,0.055141307144678686,
 0.2805157590881786,0.6202434289612445,0.5633038131269773,0.41750516257855985,
 0.5019501386071141,0.7033286426307146,0.014412002058596718,
 0.9592025034350377,0.47518034745349946,0.6041389449938754,0.7989069344355412,
 0.11146100385898461,0.19225216721064553,0.328602992392405,0.1360115299169803,
 0.5530407502562001,0.3685397343370227,0.6465621776040624,0.5190339448049599,
 0.8898575310636849,0.6566972737216268,0.6750759267358521,0.5235996309668505,
 0.8100229780853375,0.8392329532507731,0.8618266895646483,0.5098336331987984,
 0.9349072611665452,0.24252961335070256,0.8221763092195705,0.7669973706350753,
 0.11997473463512665,0.46704020077723596,0.06458424412602826,
 0.9364351168825112,0.8283614294182826,0.33744581646811056,
 0.07796790045993907,0.9666070152457485,0.7130305016724285,
 0.21738883158769873,0.5670893020163394,0.07727392299851044,
 0.0214577481582664,0.07998865701735891,0.0063477972845895625,
 0.4346817425085163,0.884100355209169,0.2433440991123319,0.6122858740240725,
 0.6130918096344462,0.04678856447142232,0.0022221345589855668,
 0.5917673227806992,0.15831305618105762,0.6787724252444027,0.5037549849664619,
 0.5653749844127097,0.814596851552542,0.9494259264286257,0.8206428717201195,
 0.23513298292634122,0.8831819867809365,0.31142594948199465,
 0.8983803569068934,0.038365022865012754,0.7520211585644297,
 0.09657788361982389,0.057678512002502624,0.9013097182695362,
 0.45734704405051274,0.9513229417230109,0.36729214396422827,
 0.5070441279535454,0.5948008825012199,0.6049050867628364,0.2168283703980045,
 0.1562178623404683,0.9347002616300358,0.870739924176589,0.4880146779487988,
 0.6934613299948537,0.7153907368529774,0.9698041121660016,0.9029131023346326,
 0.07653668488903254,0.7091127160240971,0.11081546098604267,
 0.11693324633366153,0.5202228207645923,0.8461935354471496,0.9209014062598123,
 0.8159885644043896,0.6315921779488849,0.6001920851302598,0.42980586015694966,
 0.8462053289979394,0.23902659220030031,0.19352532456782223,
 0.13176984846729722,0.2749707622736407,0.21640909901695893,
 0.20352570056042962,0.8065490614413129,0.5324791324962356,0.3912446677774897,
 0.7736941758822086,0.35838631541373966,0.5512099277148808,0.6919147592745705,
 0.5660664875088108,0.44250156756615355,0.9740744050090718,
 0.40133899158265307,0.42879378580304717,0.22845230916423964,
 0.08358226807713853,0.23997444353581443,0.7410893531043174,
 0.23850049840478404,0.8062605172840708,0.730363685487086,0.5999060073729512,
 0.5694442686731606,0.41305562772423166,0.6620404898558244,0.5365260957916496,
 0.5078845773615948,0.2786714094269447,0.5698418103133388,0.26323253801205937,
 0.6310201418671796,0.7413644030654923,0.349517489551866,0.583315593524755,
 0.5195377337984142,0.36139437825846377,0.3986354419671342,0.3207474409193749,
 0.09684599277656936,0.030833671857184664,0.31504004025466015,
 0.42766645665284353,0.5519720561484347,0.41046908876003807,
 0.5376347003086919,0.1187294855698865,0.5564141057263219,0.19937373590309804,
 0.055311473287980584,0.11305535906400765,0.6939700181503784,
 0.3604285437563842,0.7232132780558864,0.4681749896940608,0.11589307140628091,
 0.10253393765159913,0.8207988467454355,0.5492600311016713,
 0.16377805400180978,0.4657268739693913,0.9996591571045493,
 0.39159163083853255,0.2891929541211741,0.26216006680083126,
 0.8752304830008836,0.4157173555086997,0.6081163659613025,0.8197479638281645,
 0.6748345009361,0.18158347686760212,0.7757277105025221,0.2292264116448679,
 0.677747660866886,0.5147231168137787,0.5369406919611185,0.9836789156763497,
 0.8402520129345488,0.5219277904151414,0.070057012421195,0.2569622331571182,
 0.4053701394443536,0.31043654099241025,0.002124666606927117,
 0.501568782536204,0.9692374389696329,0.49378264549886386,0.8278215450225034,
 0.9185604367958198,0.5344375471574732,0.7406935592619244,0.19833050142720243,
 0.21105316071989022,0.40564197114097,0.8782538653239174,0.6047992106537584,
 0.05065997282367174,0.023121313347170713,0.8086076020188269,
 0.43723289271084254,0.568650473983705,0.8207879376226368,0.39945384741154943,
 0.6029359205463506,0.5786544657533271,0.3759628027147379,0.1824482247135677,
 0.26146881443447934,0.015261159123506007,0.6758089718298823,
 0.7840656143154453,0.10465250836629403,0.5346596909381021,0.5365176537209051,
 0.9236425371931464,0.2694564176163793,0.9517718813104862,0.9159008236300484,
 0.871602153782275,0.23211223419036764,0.005109976124616278,
 0.9039529993207441,0.5380980830901344,0.6040156936757647,0.10071917618680704,
 0.4751788288022095,0.41395653137764055,0.6019580407785213,0.922952277327015,
 0.7745150796011131,0.8145413070544189,0.44025045675524443,0.3216903268117046,
 0.6398149743803209,0.22245089175416632,0.02303530672205989,
 0.24724501106940722,0.28727070181249104,0.7819954557600155,
 0.6516277407726493,0.6420101393753797,0.4009813777187011,0.836554152524444,
 0.957275078358431,0.7468349481481651,0.9966947556886363,0.8993459779264522,
 0.10771143845816489,0.8403519878196986,0.4735403933808636,
 0.32836087311245843,0.7751193726501324,0.20398035786821178,
 0.8160008155823917,0.7607110298154971,0.777130468076582,0.39251678068860996,
 0.7977945558923965,0.5173771913890498,0.05460178344512823,
 0.33913575285325004,0.6052389132407556,0.21819931975372975,
 0.9560214633356993,0.9759080994605022,0.6690571696730705,0.7312688098836146,
 0.5507250760003147,0.15597551039163493,0.7098385935837975,0.8313052323920911,
 0.11940917204948254,0.3321898808446737,0.13571383652443447,
 0.9473799062414118,0.35969282769325284,0.36646320549224676,
 0.2724487604503547,0.4499628205228674,0.604763246539497,0.9482010398618775,
 0.5511110755488373,0.6804330054463449,0.1006818804992019,0.5616434632552717,
 0.27748727158285,0.14931917029944253,0.5459791492720965,0.11248189237327699,
 0.10807944887685872,0.7716247819629029,0.5199753135073105,0.5210573424666289,
 0.2237154910089365,0.25899536288111646,0.4425058197053535,0.8436811556158619,
 0.1930442968125139,0.9833276193792321,0.4866166499657204,0.5753312100406055,
 0.6058492148638615,0.7353669111647663,0.7945373932195293,0.6442504799026547,
 0.07217190753989622,0.4197810719878393,0.17486576419016575,
 0.32333080027765015,0.7844355918864316,0.09221192603087691,
 0.6764671376889946,0.08643119557252865,0.4309495908472607,0.2899650786921687,
 0.13410041324357136,0.6804096777066472,0.3209296639024545,
 0.08068273542871074,0.7119051021087436,0.021500114096067557,
 0.030866654579428388,0.741893740013654,0.26069812378261703,
 0.5799568902137316,0.3337244703741875,0.9139990599775926,0.29643485142977055,
 0.9567235939817359,0.4616578440964023,0.1864460244505557,0.3819304504438489,
 0.4270423205944456,0.7417572790351469,0.15978127933294906,0.3440000708172337,
 0.3766269563719683,0.33348978682398456,0.10026973753118307,
 0.4824420912563908,0.6274032989470165,0.7073308285300993,0.4032725964497107,
 0.652251264182885,0.21816367375495926,0.036846612701176285,
 0.27630680061319735,0.3704047190919679,0.39650225630332125,
 0.22256961723313096,0.981279264490751,0.32770820398666967,
 0.06540573576448283,0.7121664735779757,0.0187962458725206,0.510172913558459,
 0.07355720926383569,0.6601007743390069,0.365890967248382,0.47044697894737963,
 0.6273215868935653,0.8960854673443478,0.8877611746898271,0.8506205404036872,
 0.7880976996489248,0.9008570747847097,0.6691703540828473,0.8384931210551715,
 0.8049497306714519,0.8151963933020605,0.7291116411297993,0.19425243048182705,
 0.984988396244639,0.5317598772748314,0.6884735923431005,0.8432041436279064,
 0.06952661389235071,0.49104320147742053,0.7520463905778743,
 0.5286729780469355,0.2157591071415622,0.9928803828789496,0.5655733305757666,
 0.6722898419804954,0.03304062352059023,0.1327915709357631,
 0.44007551983627835,0.6472590811844282,0.6280579402523769,0.5518912073211207,
 0.5172727931902554,0.9830098550945064,0.5026882260478618,0.7996137417862738,
 0.34765392909195714,0.7112713350332913,0.15962394228618626,
 0.9860900551132118,0.11869331218876222,0.73139130893883,0.8769288252146606,
 0.7922563980035603,0.29208037282240007,0.47384996122078193,
 0.7345902086654514,0.5592712872702924,0.2716311856738328,0.42006500862497753,
 0.6165700125633792,0.13621573852600755,0.5268169020967478,
 0.054502603550488526,0.013555055335747301,0.10516120399446571,
 0.9690737484967915,0.8958329171048751,0.4695864026939005,0.2530033165614175,
 0.9031777917547421,0.737475470602441,0.21151845528407454,0.6690161669782565,
 0.6954460721210488,0.5211357672498244,0.022384598364054153,
 0.6209582368052087,0.9646752854168918,0.6183286318988082,0.2806035444311745,
 0.033154415272039994,0.043417523204820396,0.751825313996978,
 0.7409791431439405,0.16574661044123218,0.49482864210025335,
 0.7731137948038652,0.5597973281867927,0.9643204966968899,0.522144175113489,
 0.7422245125175073,0.6943573190008048,0.23616524833347252,0.7251148267032177,
 0.9852853627149701,0.2491109673612406,0.8802975272822763,0.4287654833050202,
 0.7786361437742793,0.18099001126509062,0.9089498617381394,0.8217017820244249,
 0.04048620468194519,0.061741235865941047,0.9667654161146637,
 0.2579059913882513,0.661331647310621,0.8675856746351919,0.5801861552798817,
 0.6843999742653947,0.8260690528112048,0.9505485593049992,0.6972281250993806,
 0.16164952477977712,0.16498965763916773,0.8695361664256231,
 0.5144377863155264,0.45761499465914834,0.8326712463285337,0.5002859549206919,
 0.8100487025405992,0.6922468492867108,0.6789496703683507,0.8774112110794184,
 0.14891792193345355,0.5154676790435886,0.1413472344157296,
 0.42854756430796237,0.2503173977034806,0.40611388514411106,
 0.10731773561552371,0.1335004756719388,0.3529801965202659,0.5959527245552538,
 0.6720534373929071,0.30429203994575427,0.8100289344403462,
 0.35815863834262196,0.05965488677432795,0.6883128900419107,0.566687648389472,
 0.464503628058053,0.7884940792940837,0.8510732428349235,0.5414757951303961,
 0.17770453194795055,0.3571003048203212,0.0639020030088926,0.4774290101495675,
 0.10944924591849403,0.8083917026377399,0.7237541718549297,0.9634505052535971,
 0.8100955823332955,0.03477314271277088,0.5281295192990959,0.6569864495512179,
 0.6046392054398535,0.5454480480980131,0.6684103321784141,0.24738108549229298,
 0.7146769606375563,0.5067046073869637,0.10841366455362156,0.1066140820973438,
 0.001364428590306166,0.28091158365931235,0.9952275400818508,
 0.22958451744534702,0.31760658475935977,0.4774782437955073,
 0.9174655174898968,0.08962956639507746,0.28128856746542874,
 0.6125672840060474,0.47011758025381023,0.15958482480560066,
 0.5725623596317269,0.8730443199440444,0.43091150330342454,
 0.26038585061063146,0.012445712080054405,0.25762249886287836,
 0.9714537289861755,0.5448367396754954,0.9210658795862633,0.2817835673500534,
 0.9160225769068506,0.7399160199409212,0.6707295401889204,0.7112592301170687,
 0.3960293582528439,0.9400677776246564,0.47394187367580587,0.7717712904576831,
 0.03596113474134843,0.018639920370783436,0.22275558308299148,
 0.6614448228482019,0.3715159095718532,0.12068685239124177,0.8697558746070718,
 0.7981424795221334,0.7963036280562585,0.630633724904093,0.09532942973569125,
 0.07036929875368403,0.28337903777138906,0.5292059944403107,
 0.9379586669334073,0.6946144563941358,0.27904197261600616,
 0.16205748095890882,0.6326228937294278,0.4256480768640567,0.4136781396467838,
 0.7926716705857633,0.2971457748818024,0.1591746837672,0.9776481300370965,
 0.6873206199568627,0.8749712354600715,0.42999560240158263,0.4651036140183813,
 0.4772967107928118,0.7428859298114625,0.35925271725908003,
 0.26188082970170534,0.8978166287556663,0.051599076184454584,
 0.1643581233417839,0.18130112252563413,0.5167072478259818,
 0.21041447054621498,0.11744135002253453,0.04545420108167897,
 0.01931856136864385,0.012255862647611915,0.8288988339042843,
 0.6696613702104176,0.20566333172593654,0.944770228883484,0.6628064917615494,
 0.8830533796438005,0.225818021806389,0.20861234280269514,0.9551099569615421,
 0.4935610190138998,0.766996271847716,0.49509488004083524,0.3015144790725153,
 0.5047279422508867,0.38543162716193,0.40548302124183544,0.30049805853580835,
 0.012153736387751524,0.5258489890151401,0.8689678079975178,
 0.49777264249155995,0.47454215072946115,0.6721400478187073,
 0.5516751194488687,0.8271818704248094,0.8114211111917365,0.7828751678427559,
 0.41512130123680424,0.06534697889850405,0.8224787493344122,
 0.7437881738179091,0.8645939394084278,0.819800773803073,0.0402876600245351,
 0.9309083759947334,0.9986234973134703,0.1008418696420681,0.11522916504912373,
 0.8807488601894299,0.33734967008194505,0.07007194872324551,
 0.7036582842603618,0.391604755962522,0.13329508212146313,0.7409185247451927,
 0.8834025849356927,0.47173579438323154,0.64206114114862,0.7503628634821328,
 0.08354549139204404,0.0026314264477390603,0.7208358324832806,
 0.5234417204814472,0.22474463951535306,0.761967252236007,0.4970036492303249,
 0.9683804278453744,0.5967979725061008,0.9390761872038332,0.9739491172284285,
 0.5348288535130833,0.2349907798520412,0.21261908303339028,0.1603900794994252,
 0.5255523346162171,0.4376987274894759,0.878745516934472,0.848241261508983,
 0.08975682989850264,0.7520712418743641,0.72289128795216,0.7863506571344174,
 0.49590291284783783,0.9380837275763435,0.6286273225176485,0.3678671614329436,
 0.8457912504103445,0.32135896943855946,0.20670193619714927,
 0.4803228533945093,0.7755339079173746,0.16295681004777018,
 0.30351943007677074,0.6813975224553547,0.8139801747190771,0.9072618926895806,
 0.5500942431683835,0.16192543428081108,0.2595350196323072,0.9548637352761848,
 0.7257258420001319,0.2522162711967676,0.9500211957741252,0.6507834316601623,
 0.02603794681514049,0.9403863611755117,0.1747762031771687,0.5733506237459114,
 0.33396306481968074,0.8296025749524067,0.4027228854263727,0.3100736258444513,
 0.5781123908383874,0.1744247664762013,0.7084089306116078,0.18695580232734288,
 0.33205125160578575,0.22976246350237894,0.2924528459748238,
 0.14748857122605086,0.8513252321287169,0.13611689293676388,
 0.9412742318721004,0.07388246325330172,0.8810772438391137,0.201042829055731,
 0.5337950105404463,0.08114697828344253,0.8200268894104966,0.9333373313884438,
 0.6122469331800456,0.20902623729813785,0.8608831064636913,
 0.015167612867221236,0.9740267454627878,0.6249954355813419,
 0.6026950705083697,0.741129515260504,0.829842900174058,0.5435474367470214,
 0.17659375217675166,0.33159922348260373,0.7443549419510179,
 0.9827153052905939,0.7701640774071022,0.49598400662785325,0.53488268690086,
 0.7745929019219675,0.4310203791704733,0.3428670215646823,0.5050770689524575,
 0.15555814953705127,0.059483773577395205,0.7452854803815893,
 0.1614145138205848,0.994141550169316,0.43282437337567503,0.8803418341111507,
 0.3232995350955754,0.6401603387723757,0.7283475289434072,0.5010081856086707,
 0.4141107375032862,0.3681162146376117,0.9078135097496249,0.04283995668309071,
 0.01623160642431365,0.22019852901449288,0.8023541351148464,
 0.6086287855700663,0.1482407098447257,0.8274135000919681,0.9133299810357127,
 0.9590921001444165,0.7777565311847279,0.21478712148337042,0.62983628004064,
 0.7645721179963897,0.557226644567848,0.3310467098569354,0.19417205986926456,
 0.8537027536790462,0.8274696327802601,0.7929251025837079,0.1547668367998729,
 0.9829040790040151,0.03306750148488813,0.7221937074090888,0.5640053216007193,
 0.0878888509964918,0.2741941347235126,0.03332267487146412,0.550386239305643,
 0.6413256881124461,0.35438996238688203,0.1501929518509849,
 0.23525397936676407,0.2448708995874428,0.6430650033657493,
 0.37029729989660787,0.8876296920798564,0.8774277133173497,0.7963820468347359,
 0.2834750382360135,0.14065418016326936,0.5200589586986375,
 0.29045755402080586,0.6941434179318997,0.12827281975414018,
 0.13611775912806656,0.1801745437731257,0.13792071212976276,
 0.8760246923428077,0.14243013905543034,0.7438029535669654,0.7374487373371702,
 0.8130954442403864,0.9794343818375708,0.5385735181985731,0.6264600171455,
 0.3909920139137557,0.795878146341992,0.6558125799363371,0.30227916522329856,
 0.2929014748085713,0.4344473868146105,0.18017278517256763,0.4477812199330963,
 0.41332219752768184,0.7547601021161061,0.14218212961715815,0.750258615085555,
 0.45075870470632906,0.5216427200419171,0.8849700278650887,
 0.08638952679350154,0.4373635109206865,0.7356909318891678,0.3092579536942307,
 0.45977003561237306,0.4445230863940818,0.23121281829917972,
 0.19716939593356064,0.6543325206296139,0.6549141412970856,
 0.0027093863014785446,0.8081569526801015,0.7537409814041287,
 0.9509697517934285,0.414237974990493,0.004599444539658992,
 0.39292192373252144,0.8470866035809643,0.5375693701332714,0.6702762126324429,
 0.10092476837669384,0.40761949540514275,0.6760070123126359,
 0.6565074121995469,0.47059670932483966,0.8214422980025079,0.7014028653452393,
 0.4485571446562586,0.19343796244479206,0.3926425575828052,
 0.09920102777299689,0.59434744482787,0.0207074941405061,0.5951732915661325,
 0.31342823809454023,0.4196911938535747,0.9195359625763118,
 0.17237061207739768,0.22929287874393878,0.9005841759027579,
 0.11220068021944207,0.4787449268926346,0.751436894711985,0.2315771455081448,
 0.2250617402538131,0.36974248551827793,0.9970616902387095,
 0.45009283262760524,0.4368724004800325,0.8033790885478063,0.6101590373880628,
 0.9320526441870868,0.40794874816815707,0.009171185023512995,
 0.4674462534309096,0.7193075611155397,0.7666965876787161,0.8247494522544413,
 0.9036203989470424,0.0452654398071175,0.303540085634428,0.33533685377674083,
 0.961460967449761,0.6789245109547767,0.8096237129946532,0.456451424178937,
 0.20719008394981286,0.12978968878418506,0.37464301025304825,
 0.37967565162347594,0.41492322632853307,0.5816495765826202,
 0.7299151597937125,0.9615081001392973,0.055134940755896755,
 0.9133480996016484,0.09624778072992324,0.8445013249400606,0.8323430043872527,
 0.5016710518849347,0.5170513497843789,0.31896036216639756,0.6773498632286868,
 0.6239020636006736,0.8001673763493306,0.09286728369262387,0.3382966228914853,
 0.7528409316580572,0.17978051301883036,0.9159650603677373,
 0.11256737649798887,0.2011313062369341,0.5176547091484494,
 0.12159919177380984,0.5148986464204195,0.0812227110123247,
 0.07248348377057678,0.9058569886928779,0.5549583047636044,0.4245144520722538,
 0.44416781272578554,0.870203233577014,0.01973767047559316,0.5555806661730925,
 0.6603115403401805,0.11742229222590217,0.7349813548601773,0.87674307157808,
 0.02594697137013513,0.07989362123178712,0.4917551504287456,
 0.0341864531530196,0.9638465368207212,0.22560230737181153,
 0.39887285220273727,0.2930506733709418,0.15560275421644842,
 0.9021778758068069,0.15351616117698375,0.14456784339211193,
 0.36118640227942833,0.35134906265161936,0.04795199119476612,
 0.2924918116951719,0.5444407907494221,0.8170825070654993,0.8085594151183513,
 0.0659934581553534,0.9189006292823166,0.5006313797886562,0.5030952215801531,
 0.14883791868438834,0.4332244577220966,0.034008931348839644,
 0.8893215082518048,0.08688512889237565,0.9230710498424616,0.7523271457657802,
 0.0806340810967443,0.833684593677926,0.6235811205173534,0.5635588881794554,
 0.09397510928519415,0.6797263476933485,0.521172487786402,0.45784065074799263,
 0.07517110149871464,0.26330246099627463,0.6594908094425467,
 0.15771472399843223,0.8947460498166396,0.973941500639962,0.550979920801219,
 0.006467977658626833,0.474585313370235,0.7481269909522673,
 0.11436911021216434,0.6188652967459337,0.81101147646014,0.5924465069637482,
 0.19225595386971217,0.7824365645907768,0.3261080259653688,0.9709761221192765,
 0.7206528926153586,0.5412834403902791,0.5122722336924399,0.8799855769592895,
 0.7194375073405959,0.10033322150338353,0.5318201812005839,
 0.25833841993542705,0.5165974615654985,0.36691864138422037,
 0.7079092796778808,0.05332544450129528,0.0743955241475307,0.5968578312492546,
 0.6960605738142002,0.9192262701340618,0.4819607155089417,0.4817732363567895,
 0.5483213449953777,0.39643481928070756,0.6567161620089474,0.5762945226848034,
 0.5372308192978901,0.6862666500927018,0.6754096956420264,
 0.026591797724182742,0.45919004001361985,0.059434568608632055,
 0.33379097615851594,0.7714968695320265,0.07352853379418156,
 0.09105506867516011,0.5467768594751413,0.18222602350767525,
 0.5884422560641668,0.8366953323395224,0.5286312571984224,0.3244445774747894,
 0.28529090080397457,0.8643068139885035,0.987899471942915,0.8282724936486752,
 0.37157031405529506,0.09500066650033834,0.10725859026424012,
 0.8847962674866238,0.4576040636230504,0.31945783907821734,
 0.47761836041400163,0.5741633634209294,0.7798466338934098,
 0.13613917588012536,0.8790176374519053,0.5697746123699742,0.5548674011350377,
 0.48138842046582264,0.4686569226201338,0.5203898716690393,
 0.36352575412403043,0.4593081702045395,0.6288188084237055,
 0.24852341147327528,0.47772142992767785,0.16044157024925043,
 0.9745519427416753,0.9841929619976639,0.306112539762442,0.9824502317881051,
 0.11672207290984327,0.24820901096831383,0.44627479863185926,
 0.1671330771465025,0.9062503416272127,0.5650283989140565,0.2712105499102757,
 0.9055262786405017,0.7768372168674423,0.5874588440597759,0.3936567201351927,
 0.6027364096972704,0.8552799454896834,0.544009695956369,0.6225190675572772,
 0.06873760592919986,0.17652289582976322,0.20349869782554053,
 0.9956667297460317,0.7210716678035957,0.22424533751286357,0.220967223131864,
 0.7982109378282765,0.9105096685700527,0.9929102782472492,0.14418417232141156,
 0.18532745352538949,0.6683634365397341,0.06831983252760576,
 0.6340184928550255,0.17993600392028397,0.18659288533700147,
 0.9859323294542752,0.4550988886514292,0.47417449393234856,0.3719382220911386,
 0.7082138146227127,0.6887975885172009,0.7590500568052032,0.9093230539821845,
 0.020863974024205145,0.4215086425886896,0.7742565337175031,
 0.19732588606975343,0.21296193009236464,0.5470095482339726,0.638003818821039,
 0.22500594564713494,0.680781314485682,0.09433718779020772,0.7822892232731324,
 0.42944550694591,0.905846717163003,0.574182093544798,0.8142611791582532,
 0.4217179234212872,0.8411471909805064,0.2282153891214852,0.07456011994722811,
 0.9597278049222643,0.3851864230224187,0.4659824104732643,0.5349035153808469,
 0.7287070223788993,0.09064135888549274,0.5044488372925154,
 0.031465533200889406,0.21713823474870209,0.8749815174030161,
 0.9221252491526793,0.5996451595424496,0.1379568585084291,0.07835770137139464,
 0.9147049295992666,0.81115739492825,0.896590682510911,0.18748215733490559,
 0.37935140185397265,0.3808448875580478,0.9700513400235649,0.4011668481926195,
 0.18355129078059895,0.8750838927108109,0.41707488597717224,
 0.7067854512073217,0.43843701671251456,0.06617292206341452,
 0.05268089542233312,0.9592021741804194,0.9483199539242582,0.8534932315557797,
 0.014994920238665443,0.8634491101510207,0.4785118341287786,
 0.9234739143827849,0.46396492514388743,0.5762665830936415,0.4120200222757959,
 0.5330527112997575,0.2649747951601048,0.13782092790876732,0.5406188822358784,
 0.7783836447357452,0.7018490592884806,0.5746796732914108,0.965759121741339,
 0.6465897320929135,0.2676574701432767,0.40657272001350275,0.2383556960802221,
 0.9622092395668657,0.9975342337593734,0.939189372590304,0.592100491814256,
 0.3477822913814711,0.01661000674208657,0.40716083363388156,
 0.46358339038083163,0.050363089674368755,0.2628762187913525,
 0.5138353470561676,0.7498275367919767,0.004608469254934078,
 0.3419943905555701,0.7859862199347771,0.03723538546624072,
 0.24335682296082872,0.4855979812389202,0.48411030935455646,
 0.07571944182546719,0.747473663824668,0.04941631919945,0.11093145893673129,
 0.6036086919758216,0.9836671767140823,0.20520234223649636,0.9084601310120645,
 0.6466659745150871,0.0227184415941577,0.712877119818331,0.9107407307176091,
 0.6361496674835639,0.3996377455998864,0.9418452437969729,0.5353163001312098,
 0.4029955178488611,0.9879056846941797,0.6855164002687602,0.32593800314105226,
 0.4707089127887646,0.07638462548542402,0.22733752708289168,
 0.11434708391411386,0.10967874649418474,0.7950947814740797,
 0.21867370620002513,0.15473179827956773,0.5552347743939889,
 0.3791607784765031,0.18018267821938827,0.9089191450793372,0.7414547025591436,
 0.20673090448187648,0.3402534987970609,0.5305294516164725,0.7203124272791294,
 0.7573727163143156,0.8812235770624661,0.6050143125851455,0.7264722237709113,
 0.13325977032826675,0.9651255297830404,0.794265965363324,0.18138366213489931,
 0.5853942031772879,0.9653487493901916,0.8855202970343742,0.90716452785564,
 0.7079159065110995,0.5760192598107706,0.9852932207373758,0.8925642609910658,
 0.8041654621904013,0.4059519876643556,0.4899571782503227,0.133302931736479,
 0.36146207897259097,0.5335022851893253,0.7652215914522388,
 0.24014128419159997,0.23334768107856507,0.5434958172757276,
 0.7220329127200654,0.478770702023975,0.24899806338424857,0.20349095204678902,
 0.6989423666416743,0.1437333216736827,0.43985189697518723,0.4370106367532557,
 0.008768917719052416,0.8517734195367561,0.6660042686112208,
 0.7945192064565517,0.7546144835977828,0.7686384121047615,0.0816983271365983,
 0.2234293206871406,0.12796616583735243,0.30575416914169495,
 0.6833594969101373,0.816678994425366,0.20863661787581533,0.1034859754100681,
 0.8031885057417677,0.04552631836201648,0.7308583015085338,0.7359481720921909,
 0.9128009090918279,0.018363305921230255,0.1621014390041705,
 0.24330376855257918,0.0310881343199495,0.33663901526885853,0.755114071366654,
 0.8124532034492927,0.33086559332055854,0.12446948603743546,
 0.33712007156783697,0.22641996536043796,0.7403389655237731,
 0.5626446111015981,0.8745375352955337,0.531683226607452,
 0.0043260838296153015,0.005138012813534876,0.5362666566841316,
 0.2602694648072328,0.8494427486968008,0.695966941553355,0.6193144555559448,
 0.5621399550304048,0.7799685547007418,0.5492384569247357,0.9926041480721083,
 0.9217438042774052,0.3705226929701205,0.6076235631615703,0.9504708052975064,
 0.09087313548196874,0.32398121191110385,0.6737039813614133,
 0.20700756286227628,0.7099147197546989,0.927783834042031,0.9121044497754811,
 0.2960026722620037,0.08272459994453607,0.3322815807947299,
 0.35241354809342074,0.5787292869264308,0.17999049430862768,
 0.5385782028113896,0.555023347230047,0.362698902375745,0.19620444778991508,
 0.46754950038003895,0.41567492026972586,0.26050455281651164,
 0.4326932599318414,0.2878421541901709,0.23402459087643512,0.5091422143649651,
 0.4728694438524481,0.6811079788116418,0.8867407747404414,0.413142206532781,
 0.1980531731714943,0.6849951287186324,0.2988700659765572,0.20324533217751117,
 0.8651974778259932,0.406707397845376,0.2122397235801713,0.23035537850700505,
 0.21182038552500926,0.07437107712080893,0.7253700957634281,
 0.21666596109881908,0.36933172074976717,0.9125232814752088,
 0.25787273050520365,0.0975826172139731,0.13270040275024853,
 0.6830731449815194,0.24211610457682342,0.7204031388193187,
 0.06522276847562147,0.3034857722708846,0.8203102074862735,0.1725595228175103,
 0.9757332290005991,0.19774457713403926,0.678218596084852,0.520682726735389,
 0.6969073868356265,0.6625864886305999,0.3206202700988472,0.07456218967845374,
 0.6370824042756187,0.43438771129317955,0.15499732676197442,
 0.4474661055591803,0.658348692240814,0.07311403896038349,0.8955373278171334,
 0.23136370269560014,0.44597199129162235,0.7336656207302997,
 0.1250491376218399,0.5592611556800706,0.40593544288381966,
 0.12297917329307262,0.8514920769118782,0.1845779684548038,0.3852720215501704,
 0.23135286861155546,0.6592422028544012,0.8688505484856713,
 0.010091974507901669,0.3201386510347043,0.1662312521456547,
 0.2843318256281041,0.5823209104833709,0.3898578956028417,0.23843032040561596,
 0.8862393710522406,0.4756116466542325,0.9662227906190307,0.8686386206935602,
 0.9722352317146151,0.4208628814535432,0.6466001822954227,0.6781725120579853,
 0.9562971222528215,0.841895941622185,0.8907072154571326,0.0588741440467323,
 0.1308469422637073,0.624846623648157,0.6010444982001739,0.5539519981468679,
 0.9526783084986322,0.2571975570957483,0.3627934677548178,0.7181857323257953,
 0.31872669158616307,0.8212593105812968,0.7983761674446722,0.8043399589725021,
 0.36118396909431105,0.5853427447130987,0.4337746780863537,0.9999054777383553,
 0.5658981533202067,0.08665985342224414,0.8213322646374628,
 0.29338576942565897,0.10861061490867852,0.5329805023763615,0.985015328043904,
 0.4497044397049681,0.9216924023526933,0.4782875682036971,0.9959938222487359,
 0.6317881997214987,0.9754332480177894,0.7087880622463806,0.04361406362241327,
 0.6324869561351257,0.550576715649003,0.31973147828072634,0.3488233796682727,
 0.4096415238254456,0.40245947632855794,0.3067313845704138,
 0.05138005668394385,0.8955021160026528,0.6368003627048484,0.3325411782392057,
 0.5657291042582218,0.6628123351816612,0.743592197177839,0.14384449901094964,
 0.1288694332090461,0.9061395419841864,0.3500820878308447,0.22505569923628,
 0.32186582257440655,0.8155386316807378,0.7120053608531577,
 0.33769208983965715,0.6180213137034829,0.6387389065436695,0.6708014996536299,
 0.897742402077301,0.6346842232552006,0.26671891288704397,0.8195581038612549,
 0.7996627131204906,0.8048336911950439,0.17762920728013842,0.9384616052269771,
 0.6550840696093073,0.9463434416190853,0.9314855025173431,0.7709365587747361,
 0.36709162849472154,0.19843839305635602,0.6177721497012129,0.264350935551143,
 0.4867583067093859,0.1244500922739521,0.433331115962757,0.35811251193652427,
 0.026930711619287617,0.9388870663635507,0.9444358636244411,
 0.9018261509087466,0.7183163560252495,0.39572800340828773,0.336544928815499,
 0.12645296551670904,0.5642144922594289,0.6955209297384004,0.6301363349024408,
 0.029625808997057135,0.5708035260695017,0.7345264259122104,
 0.31537595093846904,0.5250232248699496,0.08501686332825786,
 0.9020021909807637,0.5205270752114308,0.6081792914604519,0.7298228425769068,
 0.9684587986801223,0.16930654497628306,0.612361432367242,0.24332599128287857,
 0.03394183422537622,0.5873978719827926,0.4421115723490222,
 0.20065924009985536,0.7054336547569322,0.10200555744764306,
 0.5937384500521073,0.4071749731767388,0.17345870275926356,0.4607050109115205,
 0.04473585304283645,0.47164740745473477,0.4762967934125447,
 0.8816665670153021,0.8272503639908662,0.7827772113344158,0.645860300389508,
 0.030923088801021903,0.9066369751609041,0.569272086740934,
 0.027354272927712642,0.5385030556876161,0.44326669458513135,
 0.8948755469125784,0.5735356746338209,0.3776806652719532,0.7122336930264933,
 0.694660678959246,0.6447487856902733,0.6500351368169389,0.4827823916196202,
 0.20783514378203072,0.36288033837964984,0.32656645870879375,
 0.5233735041911871,0.47762929450284497,0.058864072576751525,
 0.8848655003129089,0.9480782489201286,0.7692298818858918,0.5827021565180036,
 0.7035424203799064,0.16977210615579896,0.20835001601415537,0.51267100732999,
 0.3088660462902544,0.7954993255789877,0.0031816530582148284,
 0.6842421614611547,0.6191941834421205,0.14862686718216878,0.8836095806860464,
 0.18609479189061273,0.653769556409826,0.9266315388280465,0.4542191726021628,
 0.03331876123926225,0.11053079503927732,0.9600602402545021,
 0.5466276521654405,0.7749803261834824,0.18450818957114612,0.2753571805676056,
 0.3958486229903906,0.7979791544252939,0.31291279253334026,
 0.07024431618532712,0.31749853589206145,0.9803732737576157,
 0.8828436518736362,0.17895264067882788,0.07995160049031169,
 0.18570781830075433,0.17217064003045357,0.45230016177980437,
 0.2061468667107259,0.10530400531918149,0.8693759453850746,
 0.15010430969224564,0.578179315784621,0.3192589962652346,0.26091098821143777,
 0.6794495619189953,0.8463386664107994,0.8151145963836307,0.7708572044750495,
 0.660406143686813,0.9717071161882728,0.17455241110513553,0.9285799661584988,
 0.9370153044875297,0.1513852237581027,0.05209693941898663,
 0.20630329369172706,0.8668825227508397,0.0823276737977966,0.7693449134368263,
 0.115247505229507,0.932378502425976,0.4640134356744723,0.07660761161013063,
 0.3880118661707014,0.6612598791817982,0.7821838304860113,0.33649756836041256,
 0.5627108866608119,0.2972920965657906,0.9470954691545052,
 0.013859657839414896,0.8288671987518856,0.126091322483744,0.5961152404645648,
 0.2949381078275227,0.8557604578633407,0.0717959418911267,0.8786802676704217,
 0.6514485777214141,0.23399914296953728,0.25244754384827217,
 0.27325869907147304,0.624525678206971,0.1503941973132824,0.7521325763047887,
 0.6548109119713671,0.3392943085595168,0.03077294501489114,
 0.36678700533676933,0.4318748631738465,0.7741498266576894,
 0.42551323816729947,0.8088538443525971,0.497680856480883,0.06313780997106433,
 0.826748930472977,0.3133442947457532,0.01995038257773718,0.10113136090598474,
 0.7315069327737984,0.32966683682876563,0.10418070846522776,
 0.04418731550381805,0.8155645903419008,0.5573823801806213,
 0.04790809463092638,0.2925570495972296,0.18441713542044658,
 0.7187775987323636,0.3391251990054738,0.7397890040661269,0.9990335864504151,
 0.7439471080986956,0.5495486928348341,0.5128418909719934,0.06317441271285029,
 0.10374680780038936,0.11649677025158445,0.9196401451164878,
 0.39663608236288495,0.27264752820584104,0.7253101828759502,0.696643920576809,
 0.28361951605077873,0.6791969240661005,0.9637916173914927,0.4728799150234765,
 0.20061884605974967,0.22390983390179975,0.8776359921555944,0.931818737259398,
 0.8793283867769977,0.6083121415987904,0.1544362309703725,0.9875058144809534,
 0.990955857822559,0.059038010199027546,0.03525564432135231,
 0.8669610106895671,0.5498040286508394,0.5271976166159702,0.5057313581631708,
 0.7795980630065773,0.3443779316914969,0.13909041037993053,0.1294107752689091,
 0.7369529900521314,0.7613017009966846,0.13297536976592983,0.8641710682574216,
 0.03858980185663452,0.933522162264886,0.9958066244349358,0.06444508650543956,
 0.14823693107467673,0.4296287375077088,0.26993082461880447,
 0.22085274089547602,0.9720301227948513,0.1797454679098398,0.803756381384826,
 0.19653183812401387,0.7520834798470106,0.8814189912178408,
 0.15729646056871482,0.9047094809923414,0.8432677067550202,0.991227220969493,
 0.34760418583036157,0.07565530837712453,0.9599358450858718,0.831871409603756,
 0.10821741295820464,0.15435949129802695,0.8061885638362888,
 0.018118984851023145,0.32642890933374824,0.5610353552136097,
 0.1278888791291869,0.13043532164530525,0.3092849943118774,0.2596559903937161,
 0.7752424320069298,0.6413100675358568,0.566044976226601,0.7499380664666639,
 0.9984168346925859,0.6309935073119188,0.5666445457985951,
 0.005177222448491059,0.679322309319163,0.2632137066217459,0.5737312863144648,
 0.9945711046489781,0.8348157865270153,0.22590740651809327,
 0.14581639082385722,0.8089586811731934,0.28377765301107205,
 0.02588889673902328,0.550794420834399,0.0935524277273363,0.7929754100087393,
 0.20264583226180966,0.9137588158925579,0.6619922668150522,0.5137333527099082,
 0.6553980793962648,0.7196601509199806,0.16887689695461283,
 0.01780433211703969,0.7750634997321209,0.814190846378406,0.7367593126797897,
 0.38564005435475956,0.616508960405074,0.7586403619363852,0.5864447201949001,
 0.6959860909699994,0.8064263969002511,0.5036804749150778,0.7084937373178757,
 0.5761170796592803,0.2894833413926683,0.8340297238739552,0.10781316946824504,
 0.6230405512137459,0.5285982644467717,0.533833115871656,0.21019754470762797,
 0.24906890801992987,0.01640465954304049,0.8722281071106758,
 0.5595568213732152,0.9617497569069949,0.5878969605824397,0.308721613961944,
 0.3323286452223747,0.15908827216770383,0.962433328901262,0.7734128133351111,
 0.08718203701252669,0.19986773340124275,0.7940723672612399,
 0.7725740087854313,0.6619470887703494,0.7100250467562137,0.6310161879397808,
 0.9758035675139687,0.6586562368563308,0.29970348562830484,0.4951881903119514,
 0.6983832298917836,0.35102670450453866,0.9270610779067642,
 0.10557252749802948,0.38132978178471066,0.9488989148267436,
 0.2974974988866481,0.3664374919901231,0.23751470513490824,0.234184794514806,
 0.6360773780486269,0.4091764029644861,0.9629003465634427,0.14908925371148063,
 0.6032153784121513,0.07064659027697884,0.7123431326207648,0.5025497519753243,
 0.6527713946624546,0.6343133303821024,0.751384763431595,0.39140458842753956,
 0.9940978061408314,0.03601297457407082,0.08941912933987395,
 0.8574749419387918,0.6503792647393088,0.8025764572400644,
 0.052116677702795156,0.6151851840931025,0.6361430563610868,
 0.35257346189895655,0.6485185205547328,0.18018458982602303,
 0.19776559175907527,0.06774278001310774,0.5508437752066125,
 0.9969698107239593,0.41592794050270476,0.29553969215046694,
 0.2599819671528122,0.09405418736347837,0.5542375414714247,
 0.38414944106881355,0.15110737736963942,0.3802745685638462,
 0.4885789625347582,0.5077777055713363,0.4750617213461966,0.08072221394360579,
 0.3381384705441862,0.7483387824597506,0.8937165609062013,0.6260429263722855,
 0.23671311398400785,0.11430317338556084,0.9813984344658402,
 0.9133036639922035,0.8878431052509972,0.04930678228971874,0.6010931330972007,
 0.8434263833292514,0.9001376750008289,0.05958236549695817,0.6858446773927412,
 0.6603816521912632,0.25310469238661004,0.5390919192410931,0.5662073325867233,
 0.7555725884576286,0.9269914294626714,0.18121134896556745,
 0.33471255267917766,0.3300059172226011,0.5015347746871262,0.999427404107285,
 0.2670894244274582,0.07047492936219824,0.6026804916538385,0.802475533409181,
 0.9858613044369926,0.6173762158163187,0.05708712929814652,0.2906235431948194,
 0.0036613584169987723,0.5589772136049105,0.022754920854247906,
 0.9005090309161763,0.8137682806840808,0.9250623021226225,0.5631473824224964,
 0.01779417706577946,0.8508960404664258,0.6286110250440632,
 0.07013957234444701,0.803096497419443,0.36951336927906087,0.7269619714179727,
 0.4932871914118817,0.27431522933352737,0.3172750367854633,0.8664505858093411,
 0.27258210039672515,0.871452772397608,0.9035089684207569,0.0508760799739395,
 0.7679064625094735,0.8299789580054662,0.22906544565492826,0.686956980359648,
 0.984430974465926,0.00666679197419473,0.023100808864191413,
 0.5156374547275866,0.6145514495424059,0.4640721640517622,0.2533662279369975,
 0.39723476149255577,0.9588805988490571,0.030082057501156556,
 0.32087603653614205,0.6076692772045067,0.48376195107348463,
 0.5684838135081602,0.6869682657044919,0.45231748312626563,0.9734333185635784,
 0.4447980757823763,0.6531907565030963,0.8513866920809798,0.26183692931165026,
 0.2903227410265019,0.48871756325018634,0.3619352159750543,0.6489197370147692,
 0.7131324017379663,0.8749695444565834,0.390048119095041,0.23975171662846884,
 0.4733897120406414,0.4829520135739812,0.3516243709794288,0.5214892718373393,
 0.04414305252909212,0.6728730758871763,0.2632398696554714,0.7678629772465371,
 0.24448280013767099,0.31434128507412995,0.6214175979734411,
 0.33622524147322763,0.40940591114631397,0.3891930709707694,
 0.7348355088595594,0.4333605534167493,0.11982774957114206,
 0.10656922268140623,0.5742248265836177,0.706211037523717,0.4551021154178079,
 0.25556574484539374,0.07610094376149523,0.9934427425843817,
 0.9327910893894348,0.8048406800145009,0.20725184067346936,
 0.27647371925564523,0.4328020788157322,0.16000974677972668,
 0.8054051627030002,0.7671375223929813,0.6128192973020596,0.8683338252112257,
 0.4785320765265504,0.9437045340977911,0.1499106374064082,0.5703833735361687,
 0.009670844494527708,0.7209469818929013,0.4583446449387416,
 0.3863026551010129,0.5954495020993059,0.0770373575012282,0.9373692179925532,
 0.06786312974469721,0.7802359176139869,0.4315942624478445,0.4504955601300118,
 0.6635403351784666,0.6319359112360792,0.7692468781711476,0.79452158336519,
 0.8118468272625684,0.43566564818786857,0.11615179534175735,
 0.7538011260012543,0.9768664793739804,0.04323373238838646,0.2342868447166797,
 0.653098745254659,0.8288134620790975,0.09851232859018988,0.10413034717884806,
 0.5162253984506913,0.7249543010085113,0.15663667930221037,0.9539834836434948,
 0.4348678357809763,0.9387688466483961,0.7824772520683969,0.9538699562993647,
 0.45518951007627384,0.1909428345896319,0.014159342122521013,
 0.08241190912259377,0.9352177371214869,0.7952512181250267,0.9605432767563459,
 0.9065320116887284,0.7974532414413424,0.8703683617212682,0.2563694726132688,
 0.234674182378648,0.9134770440631924,0.33451709766607385,0.8282922238945625,
 0.24236687032911486,0.7659154488764071,0.718168538751347,0.21760647395057997,
 0.2557376649677884,0.45265954906332617,0.5008038887499383,0.3343958660191326,
 0.5759171833542626,0.5547252383917084,0.0504826540137725,0.9097978981348505,
 0.20617661257681352,0.09118649443059468,0.5994678765573056,
 0.7156592312029173,0.20143527759397217,0.4151674637479914,
 0.11168141845743551,0.3958662912087497,0.23048761074408874,
 0.18739378764104653,0.13714845990247904,0.9498012597272683,
 0.48646710558866546,0.5473916808453982,0.0031931537563972823,
 0.23264414454659021,0.15408734973210758,0.685656585498496,0.6425425542130753,
 0.45448705641254894,0.4859954904394016,0.9895315522936603,0.431798575119106,
 0.30821014170224714,0.4012473050881997,0.12372643222299995,
 0.32550887807368756,0.5137860841871945,0.20125492193739158,
 0.29280462058397694,0.405389809484032,0.3218456819491835,0.9830587083503162,
 0.4300227497987529,0.01956281096776724,0.08026486310775616,
 0.4603645820532932,0.5207851429543441,0.9272849311994501,
 0.002563872181746696,0.9949118668290516,0.698867146626614,0.5413345616189081,
 0.6127357762544721,0.02799488167260611,0.4723064379058376,
 0.17338072262148563,0.04388557059180476,0.5694386575367633,
 0.6087105523692472,0.6702584013312107,0.8379716456438306,0.5956049470091491,
 0.8360596808547684,0.6158097921758047,0.930226687437658,0.016177372040150906,
 0.9134551045440206,0.37383251653610505,0.6307475271202349,0.9351153827062767,
 0.6576060620818771,0.6334031284912279,0.37669725089070627,0.8057536069665138,
 0.11430790410806058,0.4585341296760219,0.0895667859590128,0.9393525823127075,
 0.21424830250868843,0.36945668069895654,0.3240831475806545,0.202555542764271,
 0.19269354360505642,0.4465365926088929,0.028747402067667904,
 0.983884227743725,0.9048790994443968,0.5969638562492267,0.5640035317349983,
 0.6800212312226029,0.8820061895540765,0.18849409985849708,0.2002100018046452,
 0.1732571757839011,0.27190544846628684,0.4991747644658576,0.7991841209838849,
 0.3442515599208028,0.47016548996803176,0.1865905422774401,0.8092179444659947,
 0.20725891823831444,0.5104796560737579,0.9248692040110007,0.4761849812654916,
 0.8580611961316252,0.29024187556287506,0.5877866450227792,
 0.04113922531737557,0.12310923477273406,0.1111731237302298,
 0.37582784132604474,0.3398820868541528,0.8643674931813079,0.8901446723795957,
 0.05240791831735303,0.5408915248645432,0.9035161383642952,0.629457354901139,
 0.691903833139468,0.703308426520309,0.9707788759946454,0.8536837249296227,
 0.925724057209651,0.30377069189865924,0.28454784682529133,
 0.31688918840924796,0.038691202891376286,0.7272208132619572,
 0.014862483283949457,0.9238362121487942,0.23728062199985211,
 0.9905145247880989,0.44682393269722587,0.9709328058149834,
 0.21291442722992826,0.23223405219103654,0.5389286556246883,
 0.6208917747884636,0.3897065422771421,0.4729878592997573,0.9456706667491739,
 0.6927380157233641,0.3102077577963762,0.4730986204609857,0.08145894015273158,
 0.5974690312506418,0.9407536165313296,0.5200921541409931,0.3815906283148577,
 0.49074416585908265,0.48172422518366487,0.8293665924036244,
 0.9162222520890136,0.8632799523231417,0.6513692662813694,0.09437421661651213,
 0.8558563355975068,0.1399097746473078,0.12341591640841698,
 0.08935329780386725,0.9747500286442538,0.46641735811074214,
 0.6844252292638882,0.4888248088689042,0.8752887305406404,0.7811080260102716,
 0.7426362675538914,0.16166049121564996,0.7804268515063907,
 0.20863941831537325,0.6017290066853762,0.8874964202295843,0.0992458693958429,
 0.9845258271861703,0.3822004683312211,0.9185217874270647,0.8117119489224085,
 0.008173188907250939,0.24283563659114393,0.13160139935966786,
 0.26955558697447135,0.6779176132490948,0.6685449896833462,
 0.040020463267219974,0.89855007208493,0.670728378460077,0.9758231851014709,
 0.9031043871598936,0.8698556651736251,0.4337610644305636,0.11227044380617202,
 0.7448798365815001,0.2077177785113089,0.5575647805717953,
 0.031113141225230012,0.07436816140546076,0.7294296430994561,
 0.03140124757230445,0.6084075271399653,0.6231634136108763,0.3087735719294332,
 0.6204249948786035,0.726354455784381,0.8913982651044724,0.7706310572872149,
 0.3736824717978835,0.6846969392852026,0.7205399671305164,0.2510842061728188,
 0.7516131845216409,0.15504314209811043,0.30414787672149446,
 0.7973825539054945,0.6816234666922192,0.186725545225628,0.2860298499110465,
 0.6390221022526124,0.9383484657012678,0.5248362586885961,0.981370824307383,
 0.10933431478640587,0.7479739309402635,0.46601478808740693,
 0.7812110775399437,0.7636706521671147,0.11295282917957739,0.7807561872195053,
 0.06783608127184981,0.6799759745290724,0.17860612083715832,
 0.5076888350437859,0.2576448069418741,0.6327413153337418,0.3124442478567011,
 0.22254914801705483,0.05295166113129346,0.6410354674795901,
 0.1023558692652321,0.25518521230504887,0.48876877624239434,
 0.7169819698846391,0.24962391904863268,0.43555113693920133,
 0.1507277843238981,0.6852985812670999,0.11976324274786931,0.6837751778229495,
 0.30214579837937916,0.253492974496776,0.44206756908177014,0.8130498939293723,
 0.3378786642474254,0.5733069269144494,0.040924673349632346,0.95870517244917,
 0.7234608420132566,0.5878990249740661,0.24711628544091024,
 0.43081618365685403,0.0719335834433289,0.09010615558798463,
 0.4793329041183676,0.12919789702249196,0.18726253882446542,
 0.17996100576978435,0.42118601237748743,0.1499340737528012,
 0.3069483296062524,0.6877224701280313,0.1971156699505544,0.9326988169348864,
 0.36863896579426125,0.6852201782157314,0.7625726446599943,0.6683249528962479,
 0.04681039573950241,0.8006794812897647,0.7118271863892826,0.6004722379339934,
 0.09605517225896698,0.30756632327375266,0.3135256096974345,
 0.22728815258011614,0.6418122592301257,0.5417134335136706,0.8599013746246408,
 0.5971598148437895,0.32181058934244156,0.2229267626226552,0.5181382509539486,
 0.759280745137927,0.8554675580926081,0.7578069329208914,0.937445649137663,
 0.9452939862040098,0.8780012231590404,0.6241785173035922,0.2227546463370249,
 0.7056531239338916,0.2364601727116452,0.9833624218942568,0.4972549683388473,
 0.90588654979737,0.6491550280649805,0.08468159410369791,0.28365687829779673,
 0.9398281002627514,0.5736529533482255,0.4983511679867507,0.8977428547480001,
 0.9645342560609289,0.6030330032131466,0.12853103310896952,
 0.26550150373690085,0.4565141659297435,0.7385772545545687,
 0.025653174119992128,0.691095286039458,0.8362642973376329,0.6862718244452574,
 0.188152284220506,0.2051007485162637,0.21075057847532186,0.8990274531566187,
 0.5489770358392615,0.8961136368597853,0.3985187480586425,
 0.034689954055569405,0.13550515818265751,0.8731813273829273,
 0.449857038704641,0.5884554877213898,0.3523741128272836,0.9072310957120531,
 0.37869747864580594,0.157546472012267,0.8979035274500656,0.17050987990896527,
 0.026461109696125695,0.5550974520709445,0.04683726073370742,
 0.38893659262167424,0.49482789588816567,0.35092323690406424,
 0.062446130586858994,0.2125065986139093,0.4660885059785984,
 0.3126808452686134,0.7907972625670523,0.6255866134919785,0.6671123262661277,
 0.13207703766472678,0.2651790470774088,0.20084103469442793,
 0.9283329590857614,0.1477089647840879,0.5424082945891707,0.8096575690192618,
 0.8068345482418322,0.8621864968817534,0.39881979445475324,
 0.015446747238151759,0.9891354135837964,0.18673685049209965,
 0.3814165446207076,0.4830839043347146,0.4561313311192243,0.23674668401797705,
 0.19333349020682056,0.29800703495347747,0.5444830602139552,
 0.4557544774710707,0.09221856028025388,0.06841155458776638,
 0.5870720844921317,0.4190810315233613,0.8682543976233245,0.8736934881164857,
 0.8158421425534256,0.1895187699331583,0.18468126713582533,
 0.25030529178473493,0.4416102369053615,0.5180719274356339,0.7238755403639952,
 0.19889826050004245,0.07227312054844881,0.49729458435632345,
 0.9598934297603703,0.41087915503249006,0.0578401966874047,0.8802944203355223,
 0.5817351821069054,0.7998613128561403,0.17343060869801286,
 0.007079253467440649,0.4884580215446307,0.0999701674458594,
 0.5805353629565612,0.756514805414076,0.46889029154965356,0.7119977174861797,
 0.927576362897159,0.9903964709795504,0.49611935354901515,0.40009183558498185,
 0.012171783123613222,0.2959501475885463,0.8775533066779619,
 0.23444877359931693,0.7772357805057393,0.4518204298690516,0.4421570823208836,
 0.6883676210706062,0.23928618543077895,0.43289751691038436,
 0.14970003137220367,0.3606743052887702,0.38766369900868214,
 0.6119283209098404,0.4459991203803667,0.3273265879766898,0.43967359620540236,
 0.14541587046611437,0.9092459320815572,0.35330225446978136,0.878463114361348,
 0.3156125204030328,0.9424554875463154,0.0750433092267464,0.3912047632674437,
 0.45653108904338235,0.6562134904063969,0.08975405791539393,
 0.14736584310189704,0.09179054913932627,0.4019489970131698,
 0.8736086574122669,0.6705134859460935,0.9794478135677157,0.03454992785882727,
 0.96457703393111,0.12508081291944384,0.8599211157659374,0.8847171728500965,
 0.9169301762317658,0.4169871668417069,0.7760072258858667,0.3336959158009811,
 0.4880168794324843,0.5657583012668194,0.27071875998113903,
 0.37461296530768096,0.89975646417557,0.5007201216902167,0.9843289505166173,
 0.08416651374862316,0.906982970327124,0.6093335684751049,0.866960182596812,
 0.7745266071633842,0.4112377514409312,0.8499758159148232,0.9078260450039035,
 0.906554944058974,0.712969191338772,0.084514020987001,0.23798359188173646,
 0.5216429013223335,0.2980215762390461,0.6887781708171904,0.6098658370006756,
 0.09900765048880533,0.5725956365361555,0.9917217380798135,0.2911908407113575,
 0.6788016126960466,0.9733673124238222,0.9392919031203512,0.4857564207631351,
 0.05367767822848024,0.2723848567082321,0.7405330870001723,0.6478781744870202,
 0.7627977614606485,0.7864643678337968,0.8368438850561276,0.8038497213744588,
 0.18766088858917607,0.27082672536852037,0.19992235893087196,
 0.3781742679430887,0.828816878439823,0.3037390291106634,0.9769131315548512,
 0.5989811878296448,0.0904867317506397,0.5798190644186697,0.9339233760309292,
 0.7149944208927906,0.735567569144492,0.4339165479107858,0.44935671797753074,
 0.17595353775781986,0.8685414882210143,0.23827263892245498,0.628322665773118,
 0.661727725700566,0.922950692384815,0.5349274274497086,0.38399231492580943,
 0.9354016069941399,0.34274132214803166,0.44196706685259024,
 0.25723658121735593,0.5686672885576372,0.5012176713960762,
 0.38075371352185317,0.7141691997199906,0.6378779995428515,0.5775674827794856,
 0.518999276483495,0.13874924423011148,0.9706885684733827,0.31553205637987325,
 0.3154360905507918,0.09257885076620442,0.7722444743875643,0.9676416248275925,
 0.6817744666044565,0.4940489519369071,0.9624069314558776,0.4062336472190605,
 0.41690496900974405,0.5241036175308849,0.08682463650248529,
 0.6269731476826468,0.15835491464484552,0.560596409643496,0.15366277099270476,
 0.3141537254858586,0.3944005805864518,0.023271048937673067,
 0.6454192429388674,0.3318184631050536,0.5244374586931484,0.03583733706802994,
 0.9610812276363064,0.07497291038317833,0.7318966782634545,0.0880775648491623,
 0.8938120408877301,0.656502661999106,0.6436093266741721,0.8580098786636146,
 0.47119342323237445,0.21044677850902194,0.5930932743792063,0.941068557582792,
 0.8365911529331798,0.5997559727802492,0.6521540494238476,0.27650111728728033,
 0.5700223841726985,0.44643238305406374,0.3537117030203034,0.8927069897667227,
 0.028289800896355133,0.3907637602943371,0.4511468732034134,
 0.9458633539949102,0.8926170932932181,0.11968886567124881,0.4138153087894909,
 0.9443898585832813,0.809339204683581,0.5300424235886069,0.18232645022585015,
 0.16235271922391037,0.4615336187206318,0.09937903386895719,
 0.6834031523809216,0.5904374720816559,0.19073119598202548,0.575705596194711,
 0.0759633431435115,0.5829491908692153,0.0860746390044258,0.159069506473129,
 0.7859984253194683,0.7375803488060039,0.4070318659340816,0.9983728426207134,
 0.4177304794495035,0.6433760387666895,0.7397467515623677,0.7111797757611235,
 0.463604412108104,0.4712430129837255,0.9771448930516873,0.37409830742954686,
 0.3027313552028238,0.5342995651068763,0.785996403850516,0.3450141729276487,
 0.9948338314450995,0.6097029309068825,0.7587923438169349,0.7554081616078672,
 0.2061603395370073,0.27888595079053635,0.7514187133334174,0.6937701803996933,
 0.642239597037183,0.3034575842889552,0.5487565475486601,0.3111305662087158,
 0.6874898442064035,0.39702777538866674,0.6565494614943943,0.9553490116980701,
 0.9758887287964008,0.8068945258367813,0.7173620937916875,0.5516869764807865,
 0.784003505993159,0.6265432481234934,0.7998260837806923,0.40740318545310084,
 0.27186815414698773,0.7430817848057828,0.002402815093311639,
 0.46297167448630006,0.3846319174267997,0.47850597367122594,0.881698960671826,
 0.7337773652926273,0.08248878208308952,0.2485914881755762,
 0.09134740224862703,0.25565993114457886,0.8630887957423913,
 0.7875705030041845,0.20979271412814604,0.29019881513042867,
 0.8704759973132244,0.21004797947292542,0.2617452148594145,0.9477842087903512,
 0.26259986665031887,0.01984258883017409,0.16305971614230186,
 0.39527796168532536,0.6303122329155086,0.8386496886422578,
 0.13684763227777863,0.833054430138071,0.026100194927601295,
 0.2354631474977571,0.9063265656306198,0.4074926747690337,0.8924377653427709,
 0.9237682481689147,0.20872682390633324,0.9507428679707615,0.364898965235124,
 0.4878276525087035,0.34191316221280044,0.2012022152468309,0.7605677576018599,
 0.3408121801974615,0.4852064079657006,0.5320848417023041,0.2359464704537757,
 0.3734199599604666,0.6552217950919998,0.11000108271749087,
 0.41324111694775234,0.3885511445957043,0.3274881881872005,0.79008953264126,
 0.7625676707994412,0.49647236028397357,0.005359912773650333,
 0.9196357637240169,0.9657673945275885,0.6281800914541484,0.7615324803526351,
 0.009947592869309907,0.33418350338377445,0.5405547600797352,
 0.9679224792105128,0.4121910051005768,0.7339434458430051,0.9155417918302575,
 0.6036653284216987,0.8283114814882857,0.37595558633773773,0.7071093528190887,
 0.03771422683371339,0.885102649119713,0.8484301982776795,0.09089906010454496,
 0.32903741232341766,0.7774942227245478,0.6702054786192273,0.1163144172670022,
 0.2719779616739215,0.2591191200057946,0.1597825026981753,0.6563028075603156,
 0.5146759872720741,0.8701905420159157,0.6119187031415737,0.19742953000150865,
 0.6098715464875987,0.11288664041761098,0.580974934156878,0.1207475389140844,
 0.7310636014831358,0.3774515527458736,0.7107668322841987,0.6615831596149618,
 0.8621664443100725,0.5593831392963815,0.20013022790320012,
 0.11938582975537104,0.24433819970284165,0.1891530888455788,0.673959414683525,
 0.5611923358463459,0.013397445876128677,0.7118390799668496,
 0.6224234392323389,0.4436606137359713,0.29956015449071804,0.6494006016483104,
 0.8050980463534889,0.28392743991055447,0.7224675727412755,0.7468013628309995,
 0.6717525038166073,0.7944133261832301,0.5542577501068812,0.5561297744788761,
 0.5516941508335858,0.8889500625751875,0.638677054196322,0.5101219229350846,
 0.47513502177395706,0.386206437292691,0.2598928440469419,0.6407329140734412,
 0.17388867213137993,0.7517477250294345,0.978119943763174,0.32297548796386266,
 0.673709568701035,0.6315928128978744,0.197907383486989,0.1831516742516045,
 0.6358184947282388,0.34699315271258935,0.4296480495394752,0.6280180920552795,
 0.2201395994085943,0.5878440132403941,0.45586909344005766,0.8899479849834078,
 0.31534955805397125,7.867772070948575e-4,0.8054199794139872,
 0.24725502387672094,0.20926415333266613,0.5496393268769815,
 0.14784371087533366,0.1221664656592899,0.8531814760073776,0.6335605891004977,
 0.6380169895088313,0.7534007277595571,0.7496912331905052,0.38839808126004693,
 0.6354184757954318,0.18971414065373748,0.9303527724099871,
 0.45810127075769314,0.9470673887969546,0.13149607054752854,
 0.08277706761773718,0.22646322392708695,0.5778584882913909,
 0.6359959881676165,0.2684380037191594,0.724233898231195,0.9152915190944753,
 0.559545400624779,0.03173850312248638,0.1802585630859711,0.6127223166653555,
 0.06277333208174662,0.7020798639826338,0.24834310275485294,
 0.6476297013846449,0.9266157435118518,0.38026355197229444,0.7402374917135883,
 0.5435683338262025,0.9896371884489237,0.7654471314986502,0.3589479055957808,
 0.2646828503436368,0.7474974907506354,0.5781296902724979,0.16782037955129836,
 0.9082087529635562,0.30854097133324543,0.5484659016217186,
 0.22475525740737234,0.4226997374289817,0.10965680242871068,
 0.8964902350004715,0.35612764737761804,0.011104706435342004,
 0.5272448789216118,0.0897347308070322,0.05632964473255908,
 0.10543241500925948,0.3123012017007276,0.524074636207608,0.4708207631221051,
 0.10910924487703522,0.4784132269718754,0.2619805695528161,0.4007537835787429,
 0.08638114306324074,0.668002449050704,0.4936037211093439,0.6196490466819953,
 0.6333740405596169,0.012774561813992613,0.43545483258206985,
 0.7258443934945915,0.3632717460701399,0.6161794325773926,0.05606886358152596,
 0.3955407973956988,0.28898242300272603,0.6338967239776135,
 0.48692551301493026,0.7003486163442078,0.2624088013954108,
 0.27775721298484224,0.6674426826753452,0.8827070940705597,0.5517964091802124,
 0.6324918037367581,0.3760257845027014,0.8983806152052298,0.1306690773763559,
 0.3020483405617823,0.5076552764449653,0.0475231027467502,0.9989835973082105,
 0.5416478450257972,0.6830033884342844,0.769733389050018,0.9362465079421924,
 0.9520351488528722,0.30538172976795563,0.7566572235059585,0.3688613543027307,
 0.9752793968787996,0.18903072212285155,0.6057064467670602,0.6742478464824899,
 0.4003468050355089,0.19719096078914922,0.5932178138376026,0.3760557900863909,
 0.9487563668379597,0.46020122599208346,0.8913649317360618,0.7100817580423391,
 0.12378279871446507,0.04952254711443549,0.3453555200640992,
 0.5558050377561246,0.5491305669248254,0.28733669227261216,0.2526188689720972,
 0.35945819580910165,0.04248356158292754,0.4441493095571686,
 0.10379810630608644,0.23297233504901882,0.46862655444642065,
 0.05952983880419149,0.37485130667409583,0.9829604373423261,
 0.8964812696276918,0.53081351790494,0.7327312080697789,0.46799216671618715,
 0.17722838746578273,0.7390099226371645,0.14170389395591232,
 0.3161003993820196,0.537174550482132,0.5043348486383981,0.23459758530743013,
 0.6432069168726693,0.4374442486717043,0.9346205162544324,0.20211789105531797,
 0.14476829518173873,0.49142595139374157,0.708456043898174,
 0.38405490218056926,0.0017251670663234187,0.962914909450725,
 0.5759238531143815,0.8728372525120177,0.9713462952055733,0.4454573665756245,
 0.44289057313882807,0.9794584467577152,0.4310343163358761,
 0.05804663154762402,0.7470909496253102,0.10260968207122223,
 0.3311438253310033,0.934279228966879,0.5473044698309506,0.5216834982116848,
 0.7467259849507799,0.6859184707514183,0.057245980401314456,
 0.5289442357726435,0.06312319360266305,0.27370250110506333,
 0.24797660844242264,0.8325291236581726,0.8287884415990233,0.6484574499144199,
 0.3244494687609033,0.47018516518773235,0.8368018826086059,0.8747093445043514,
 0.08860561196611494,0.25801509465685124,0.48907599764977117,
 0.8781956547644649,0.7480410637060749,0.9282369207719333,0.3817228150611667,
 0.820190434800042,0.7543887575641997,0.8535923094966237,0.3981571337390656,
 0.5325570755090849,0.6956735311029135,0.4953688153001248,0.343988306239712,
 0.28283520184176336,0.025998913348802777,0.5002196640757905,
 0.3131496661380814,0.42136522523666164,0.5710075440715394,0.440536024716605,
 0.6847836686057895,0.9075389611046895,0.33611494280756826,0.2602604203133516,
 0.29080709043834374,0.8331806836641722,0.35685795960338995,
 0.8298437895090482,0.6307170877343754,0.8615592258975174,0.19143089950570458,
 0.7120605887680282,0.6816570888422464,0.7079762120023122,
 0.034677279885760526,0.39969799335322387,0.6009063185458945,
 0.8201546400898297,0.021611879890492403,0.36117397725127853,
 0.2843817409970769,0.8103386532556679,0.9796973436389882,0.2463214007875909,
 0.7547860278920833,0.7167461903673222,0.1752972446903448,0.7039585549133213,
 0.4957396049723086,0.48862301679663545,0.7693087407353659,
 0.014777291975393791,0.49190973359265977,0.6142874256748188,
 0.2266748430992913,0.9428791637208178,0.8298372501782001,0.16600283923919912,
 0.5582464667855273,0.6747392532314698,0.4044272726987578,0.49539128774276675,
 0.6294300463237817,0.26448519829301986,0.7199479418569188,0.7116309023708338,
 0.08623896264968112,0.5669865885220842,0.31905322079657883,
 0.23017721179941686,0.4558393564104769,0.46354731499439694,
 0.020433669886838368,0.4000107885460443,0.35273131441418404,
 0.8074013757344157,0.9122248254706731,0.8647576155449039,0.45014435449106605,
 0.6409654124545484,0.9968193554334691,0.6709983307553995,0.8589225128671212,
 0.25424149920187145,0.3205929589589348,0.33359844446763764,
 0.8923449368399456,0.46141223114439023,0.11484623439995845,
 0.7225412959107241,0.7897954649110209,0.4482217699828003,0.725354378895,
 0.19939942695531532,0.7021262754758527,0.2186878553972561,0.2592365587612324,
 0.7656222904487051,0.8228391145042266,0.020868973350255948,
 0.25221160855025226,0.47000765864339256,0.9036283215385901,
 0.12946104069457187,0.5455326638166945,0.002671108384403209,
 0.06158300129561678,0.831377557530656,0.47519390572871967,0.8412656493993026,
 0.9798494535615487,0.8261722575287092,0.7141081811435239,0.5598561975324453,
 0.2973260617464728,0.8795013094680062,0.9402600573466229,0.6410281652119838,
 0.10971183621056835,0.8535162225279,0.5869963967103389,0.8506289382102539,
 0.10789225192858709,0.08917780165169198,0.35753321447722697,
 0.8070873027765019,0.8922193309266515,0.5423193065688172,0.14812935817006867,
 0.7303329303911985,0.21713174939158852,0.8538453489162489,
 0.13972522250618769,0.13703080053937933,0.9646726646261744,
 0.7712344935959786,0.5150863043311052,0.6924997562372831,0.5206949603305556,
 0.4605236636789305,0.569298090564371,0.8071351067986988,0.7904520075923945,
 0.10166443068225695,0.389253006617869,0.600537193982016,0.9596718928552035,
 0.7748595134512128,0.031894651748241354,0.9847394033180833,
 0.8391309031847044,0.013269645011199871,0.6409241852353376,
 0.1177465664176146,0.3814438589494554,0.04368407383437478,0.764629543575234,
 0.12522803638392688,0.011241406708310864,0.6174267847202095,
 0.4447405361551072,0.8482950511163813,0.9623680052518749,0.6701691283635783,
 0.2069201107184384,0.7550492618689898,0.31330172653551136,0.8051083717208607,
 0.6024264996009508,0.045975904716192506,0.25651738793668344,
 0.6124786288135534,0.9991112161263307,0.031699309212213866,
 0.47334292285897406,0.6198836842400686,0.9433469036145958,
 0.33679823733525593,0.896951885611395,0.1565249912558716,0.9315496526065606,
 0.30390715260532475,0.278057036135012,0.7836766396252093,0.5825461321710189,
 0.04017499449258943,0.9299023898707204,0.7669338598370874,0.6795446701723039,
 0.7896001600654013,0.3312964228480284,0.9955796168305379,
 0.0033896135018336437,0.006544036385741414,0.5529660472943141,
 0.3073240524841546,0.9854491732880183,0.65637257555203,0.899266379765995,
 0.011878937847795146,0.6526201766742039,0.8001361327229279,
 0.2624872090885926,0.9912249769434547,0.7172503689879456,0.8007074260032807,
 0.5752280500569122,0.3070327103868068,0.8147913498982731,0.6525587946945924,
 0.47422323590157966,0.5850383165719895,0.967803258257989,0.4724456460910038,
 0.4733604443893069,0.7584119915259379,0.9205069659961492,0.6616032047964723,
 0.05874662761328764,0.5466107384877187,0.09193774715339187,
 0.4198032778950711,0.41140476154681127,0.879294551275123,0.1959749688594965,
 0.5741361090784646,0.19745974189859528,0.03502916505926379,
 0.8668169546227786,0.691607338799486,0.6228516561729386,0.26812571437408794,
 0.6960256702187839,0.857548883325594,0.316539710609123,0.7398602049408755,
 0.67519835009397,0.5984576731901188,0.22129555361033781,0.9977560218833403,
 0.7339584705212001,0.11682099272965885,0.5888738209941282,0.7182183708394821,
 0.5251132776182564,0.575946117175557,0.06278276258920756,0.15422128122177559,
 0.9896634286192247,0.8587250172983161,0.5620819891393799,0.6304025689293362,
 0.21844048037449237,0.7828885432696194,0.10844114604833399,0.378515239496399,
 0.3451159975806055,0.9823285356760828,0.27229464968139305,0.6553607138313471,
 0.4944543465869191,0.2614394827310983,0.1128850471173315,0.9958014432137567,
 0.10023668779198214,0.34792992013982205,0.03901874802842109,
 0.08604098827914275,0.7686194997479304,0.5031315958822027,0.9333187241875258,
 0.038086000169202716,0.7145888781874086,0.2642452758880647,
 0.4185894204644991,0.185434595917914,0.7629314886741511,0.6667613411723174,
 0.8595999875223653,0.6024570593424059,0.7601359557620442,0.10968144734560137,
 0.9200338311669956,0.917451971602922,0.902549927243661,0.9108859441535823,
 0.2798941259100488,0.14208698756417193,0.1523108916008118,0.5538835817984025,
 0.23698011199380775,0.02328456204956386,0.5586970083624457,
 0.1547836426692618,0.39152188696978896,0.58188714443705,0.34410777979949747,
 0.8258714248542612,0.9729989087545343,0.5385759743519507,
 0.012526877032670969,0.7670680168387896,0.6438030802845034,
 0.043918418973816764,0.40458172346583987,0.5103600029331029,
 0.9532960146381472,0.2281143833329342,0.023533924186501354,
 0.07954203574585583,0.14164289742928726,0.5073737216846039,
 0.8944830452531527,0.8377273014447523,0.09464483080460462,0.6602608990404998,
 0.6226523175813583,0.42248689063264444,0.21580346499165748,
 0.3873659198254831,0.5405522474217299,0.6350530300767574,0.23243477333795837,
 0.7240895630226359,0.7595550813952548,0.20867012791621675,0.6191925556172088,
 0.38478619249094514,0.30006930306499247,0.03800207136495126,
 0.8972103076292485,0.09847216868353315,0.9046765051806519,0.8022185178404228,
 0.05805781575305591,0.5401071771452309,0.9628780535115924,0.8366841277938072,
 0.8317139133311853,0.07522073421744191,0.3756955044281828,0.2664003812583532,
 0.3632794001080688,0.6221607689635923,0.016941852664132395,
 0.5626910605430449,0.2516697247619687,0.16501023079124522,
 0.41517328291874966,0.521737468492638,0.19946964572848824,0.8320632422386961,
 0.5498415572994546,0.7297008329353739,0.6622936545858982,0.7086230064800452,
 0.02788802899293874,0.8484941227389431,0.5477122739968202,0.634129926675962,
 0.7096357115566311,0.9390115770153981,0.7392860544321223,0.34807803464796294,
 0.46614744210815173,0.6221836012611589,0.20391870353890862,0.285189118592037,
 0.4783152337976103,0.9469337863086137,0.8463248785808453,0.21054653601265072,
 0.12119302864267179,0.9589243482803335,0.9161695907528786,0.9327819692162246,
 0.7288629378696736,0.31640364342189076,0.7686013900130728,0.553287361846944,
 0.37966021275730255,0.2697452391142393,0.8594475540033002,0.8305565436335708,
 0.23164200177338845,0.22379419732669925,0.6418829255052059,
 0.5592880652185087,0.8963697351900035,0.32387861644038407,0.4339599123878306,
 0.6148235880706887,0.36903345146147326,0.5636166757694937,0.4285865464722841,
 0.5219696218838371,0.6705259303031542,0.06194066393563413,0.7817799957372679,
 0.243420979149219,0.38595179261745116,0.15654335235507533,
 0.23131613416521934,0.4341445707968823,0.848567294959464,0.5541684352853429,
 0.5181584558423236,0.04946100538024656,0.019357998590555536,
 0.716299227585961,0.42530935045420937,0.6117757691251158,0.9966959274543897,
 0.2706947675361012,0.6939313624958423,0.028548861738622144,
 0.26265599667971273,0.36786899191564904,0.7213942592017916,
 0.6367786535693047,0.8350651757044779,0.08855564566181195,
 0.11131641136279014,0.1566304096632235,0.8570965183128865,0.1233622144892017,
 0.6500211710213281,0.9496743476524997,0.7801898952563168,0.20853653124745242,
 0.3451394625613553,0.3479523476018602,0.4661891607112786,0.8288750667278026,
 0.9471812173571598,0.08752007643704418,0.5764354735136455,0.9053384580306426,
 0.731042734813734,0.2475431965449436,0.8237939739509023,0.7461312713328917,
 0.9130885788360743,0.6731711507341925,0.7194453620558746,0.23676296608351377,
 0.5436958205034156,0.09261050846021601,3.7049409693401536e-5,
 0.2075322552317438,0.14868115508505375,0.01259780142447342,
 0.9570456817980819,0.5238931702160987,0.9537781064965588,0.7965913968645009,
 0.9412586365006165,0.3940419623800615,0.9353285150575209,0.3205299327729514,
 0.0015824612570790597,0.31389432629902103,1.9728568986066009e-4,
 0.0829017275480265,0.08535258360313325,0.30733607748705916,
 0.4359056255089625,0.9059544812515328,0.6067995811733102,0.0952108073226664,
 0.5005107505819506,0.0916183648627571,0.7915450571991297,0.28156824401328007,
 0.17544498508381168,0.47651998206210344,0.8642984608429423,
 0.9667295915363829,0.18641814426811154,0.5613896027058747,
 0.41394984790916867,0.027559015256962915,0.5933796133970517,
 0.6194171586861799,0.9594889117448163,0.8944664970644834,0.5981912659477031,
 0.877803036076693,0.2059794223073812,0.8656526117643305,0.6032402841442787,
 0.355556027366317,0.237213996307988,0.88837611069653,0.06737934225400077,
 0.659330768356062,0.06709008218792822,0.8260257720831934,0.7810539406805866,
 0.9191063717025717,0.32616963416721123,0.4860988136884822,0.9523943297278619,
 0.9983830071870443,0.800121677196305,0.08594030472847125,0.13489126900615234,
 0.031094252615440854,0.234712252007089,0.4374634714385802,
 0.06680555843316704,0.8549266391937506,0.04335730556956696,
 0.1301786669228553,0.5306556057881157,0.9662197589777826,0.27748666362540986,
 0.18603866292377902,0.3887118431419351,0.978762338718869,0.22527058419870238,
 0.9803725412047533,0.5815961580359242,0.2751709039929594,0.5667645858279319,
 0.918741362969855,0.6610628668501808,0.02567281280106215,0.42111804220272986,
 0.37183170948829836,0.48135392603393856,0.41702065855719317,
 0.6999283790584464,0.9013874523209968,0.6447080634697558,0.5633757275314015,
 0.51020414108711,0.9489277146163412,0.5731937017551573,0.7878451139055467,
 0.4290527192138677,0.40638672709578616,0.2860567448759286,0.5456172763540268,
 0.29903492785248,0.9612767108397131,0.8903104637329745,0.0054054974268044065,
 0.5078480154840762,0.16261809880588818,0.874409201613803,0.5510700320604325,
 0.5167116030236206,0.12082117735649689,0.13176080102923338,
 0.6327951582445741,0.39127969967330634,0.9961018619343094,0.5647151194630181,
 0.8669962711557898,0.08786798670229312,0.9869340515053823,
 0.053579523986075794,0.9923497463552163,0.6818395815779769,
 0.6381595866888969,0.4959250342076428,0.07829862120984155,
 0.35989045961704935,0.6625520197856686,0.555423481287214,0.25725865818128624,
 0.2332975788448829,0.08070040692320457,0.63746280395,0.9264094960975975,
 0.6552686236225481,0.3369951216218833,0.577397334017409,0.6707429715303879,
 0.8574105043569359,0.013027721787552515,0.5910838021121051,
 0.6890237047556828,0.881272468305686,0.9160387352889604,0.016541512230819788,
 0.030120590302279426,0.235467985587537,0.5486984315808319,0.9885278937795529,
 0.1686012379391152,0.7859836053497361,0.033684088933619405,
 0.7845683223564656,0.9291882018434148,0.11587327263499003,
 0.21964194049914376,0.9954973611010668,0.5448876441072827,0.6741734056527438,
 0.5009823202751291,0.6925721396978237,0.46210279627064677,0.8673386723732501,
 0.8537657830299636,0.507196337251874,0.4890445384239879,0.7249704685090541,
 0.0582125610629427,0.7129310276636713,0.854564431899695,0.8905278916933961,
 0.5024365323978255,0.44966288390230935,0.59049156794666,0.9399744689485294,
 0.3674884466406778,0.5891480649788265,0.9503111800061754,0.8279766307500239,
 0.5463074334827813,0.6552591867959817,0.8922805365687161,0.5163075109332184,
 0.3481651113673533,0.3103001296896133,0.8061520082163113,0.9940002172076571,
 0.44571912025808746,0.9186478817122304,0.03542112983795498,
 0.6152513337475073,0.02371119356228879,0.7941940118122033,0.968064184661852,
 0.6156452396420752,0.7643218159562307,0.22433266794820272,0.7819116804717378,
 0.13005224474471855,0.44618165102485663,0.26035495504585704,
 0.4569576282613834,0.8361291428800056,0.6287619392082645,0.7795936987256296,
 0.9906189641314704,0.08770173306154427,0.48644902409398494,
 0.14619182879157544,0.299994856283857,0.5360875382212751,0.08622966536439691,
 0.13209899092180966,0.1582221335119024,0.3424685183542554,0.256610739959764,
 0.06913871086773815,0.1855106506548163,0.814450166139786,0.26719933632239157,
 0.4263542591917392,0.8020717492937202,0.3772275803310603,0.9571492102433652,
 0.31048489944878943,0.7706531053159278,0.7582319186802382,0.933222262365957,
 0.870037349262229,0.2886758103436404,0.4851156850622683,0.3435905827265484,
 0.758365950467267,0.7641304078769374,0.6004852231002066,0.5762748252837449,
 0.9335201360057377,0.35817034512381263,0.7219780621173196,0.8880618984354344,
 0.38184770272331825,0.324964750034917,0.11083396695139136,
 0.38331198537087796,0.9222998600134767,0.2625233217404118,0.9699840217800453,
 0.20509102599901752,0.566689789943693,0.5233187486945697,0.8617063231050238,
 0.7223362423571656,0.9789052282815764,0.8131983790517442,0.4688613971223514,
 0.6283185246000088,0.42623419921384775,0.27922839470370864,
 0.8330537160138992,0.2920648769147435,0.7823884088847055,0.05986376501141788,
 0.9830449330368431,0.9192543302082891,0.5772702130521407,0.9922841996255573,
 0.7372910438519291,0.6382234489511861,0.1973547907761205,0.5944445665489138,
 0.06625104899423961,0.14980330036517542,0.3573509259329193,
 0.4650628846886762,0.7701371425170426,0.5445146675360393,0.35550410411894906,
 0.5444196094634943,0.15637305644944943,0.6814266550478216,0.9821834624341463,
 0.9320931771977233,0.8291295214566299,0.8231223939951496,0.542990425278561,
 0.10534472633751846,0.8905290346110214,0.9727284219905596,0.6079055397112585,
 0.42088071035770347,0.2738756519459886,0.8211488690276288,
 0.056887318933783626,0.1765185577738554,0.6240715113253936,0.632325764810334,
 0.7547772763103058,0.6933979872280028,0.7429211566419809,0.602178045157921,
 0.4515407523619859,0.39585613798885844,0.7566239267290139,0.3451323998765943,
 0.173980506384535,0.3866902312716072,0.3616549474040961,0.0961655083937436,
 0.101811131318292,0.14573743673611872,0.6281492093368554,0.5419890035232515,
 0.7715414794535098,0.41929188767870096,0.42245333307111965,
 0.9193093445537522,0.7674948117594895,0.7397348429281926,0.09139942312144855,
 0.08883480583887515,0.35202876980697384,0.7388439213469367,
 0.49218841824542636,0.04787979178879864,0.5251460575362175,
 0.7784015630836442,0.8876986858181347,0.8040151037993051,
 0.051786549708131235,0.9183616577089463,0.9597420162085161,
 0.5630896008418087,0.6064227083369336,0.43249190151847194,0.833016105546835,
 0.6874663978254557,0.05137641248738922,0.94382780629684,0.5799360756243038,
 0.2287763540498845,0.144116136196349,0.8612769537040519,0.5754974131967692,
 0.34712621030117774,0.11312596471655278,0.9312800873266214,
 0.8875312728732656,0.1490965262370989,0.2860342922418634,0.15271846656250543,
 0.523205279744791,0.8574568776615611,0.3936054101797999,0.5237183391097039,
 0.22748700036230463,0.4548062608207384,0.6091357357469667,0.7660398511110809,
 0.1428417367660859,0.671878652106441,0.2987283866411363,0.7960519005504778,
 0.13753110137048097,0.15272606364297603,0.047317125538028804,
 0.38370222756198225,0.6249134359121394,0.2688850620664206,
 0.34338109664606853,0.8662967620871618,0.16556978135366685,
 0.5154889104961532,0.5215237182580481,0.021078178764733724,
 0.9708470403408317,0.9626174602508506,0.34679059341862384,0.2587556268529829,
 0.4351591365162414,0.8048635384024615,0.921606260381762,0.9446174862260032,
 0.10645635909575168,0.22750504361386892,0.8087272776758396,
 0.012437700596694135,0.5678247355544888,0.1629858920779963,
 0.6129189983218462,0.47012960233911105,0.4946976502887004,
 0.05775236428296815,0.5498043553110502,0.5498321610862441,0.991013883034065,
 0.42305335621076456,0.5175693525359089,0.029251408079356556,
 0.7846227868263738,0.16319970506561876,0.20052324646675224,
 0.2614744601014731,0.5713275128284717,0.28731735789671575,0.5687416929907848,
 0.3654656313131279,0.31389838499831846,0.6954495035261706,0.5289196458861789,
 0.9132669431732908,0.9444020424301668,0.7037656766993473,0.12025290651315501,
 0.28176803901925407,0.5747098216102928,0.13726339367451668,0.453830578135902,
 0.6137350994566078,0.5602902106220319,0.9981460093116065,0.5319185955096812,
 0.07527685563694264,0.39634499060139194,0.3904054879666238,
 0.030985284349362652,0.28366723940777594,0.294801784555756,
 0.8012871932248595,0.7482744175780365,0.9167623588806929,0.8622359725536646,
 0.7868609623009616,0.9142166110497391,0.6887970539738636,0.6585014911081964,
 0.5902394828349072,0.37354388265909755,0.1295498389997617,0.6437788559562108,
 0.19408626243421,0.8286480982480265,0.5193337855520965,0.8382024042316529,
 0.2463267909641328,0.5694498773328746,0.04920580482007075,0.6141925042559931,
 0.5910810776600157,0.9889391442703401,0.49858849692872265,0.7978490831735083,
 0.005236289379472847,0.8507501999454845,0.7263958002703993,
 0.09244959424321975,0.9894435048862968,0.8553651267353568,
 0.37760738277745687,0.3173091169618161,0.5279215753562273,0.4331808828950885,
 0.46516137769930843,0.3383942477105073,0.48100460854395233,
 0.24121683948180783,0.25263410930894303,0.822997457233791,
 0.25302878456531164,0.883038961328766,0.3615750587646166,0.6022731663085505,
 0.6440685151521327,0.8820432494461985,0.7652353032657025,0.40671893866941955,
 0.40057751201183556,0.2531128722163358,0.0660735018572094,
 0.36646258287015776,0.23775022893690168,0.8723993914925016,
 0.8290765419161311,0.5184077006218816,0.018290022998230238,
 0.22107967456078792,0.5552540664085589,0.7952614745792708,
 0.36806651040488125,0.31235555448949925,0.35090499429397337,
 0.4372236131450117,0.22505837981102994,0.22778589743680633,
 0.45256789318389057,0.7405520398790121,0.9314882346238081,
 0.49330155028068823,0.5544449462959106,0.9777276577555395,0.3981407626114649,
 0.819313413701376,0.24936158076782422,0.17438244271693049,
 0.25056920735855015,0.8071144702668249,0.18770551973215832,
 0.6520698221868493,0.3586676719592622,0.9804449083069728,0.6855049946375392,
 0.9659118021636047,0.5681801647318582,0.22029250015601276,0.7289354293339142,
 0.4346009592734691,0.2618005846017837,0.04476843086398896,0.9611664268437937,
 0.779450932333476,0.7910137664138511,0.9849972207829815,0.8131017088645446,
 0.3487677679207474,0.12722012672969107,0.22061751918396566,
 0.5541000342214016,0.04523031133222433,0.5588468685223336,
 0.12993470005900554,0.1050567119969511,0.949146481293806,0.7867140723028746,
 0.7142908233097635,0.8880346793713678,0.8876349384541766,0.14179464956939591,
 0.0353446548288181,0.6023846672025084,0.8809529687549329,0.5553743433072212,
 0.06452328901728221,0.23848898793518747,0.1798677563796771,
 0.5963965757566386,0.6317869935123419,0.5927175910101345,0.19403816234640336,
 0.8381008851561127,0.09997091474534348,0.2704916192523986,0.8175278811348837,
 0.29465044772607796,0.7066876855696367,0.6448606622345447,
 0.03989321030392734,0.7097864226176445,0.6616531379125583,0.8764595853652007,
 0.41557231856446963,0.09983487250128253,0.6426362164066253,
 0.13251948565656158,0.925818112214514,0.45942062885315715,0.6911322719733484,
 0.7690553512314172,0.567275597372155,0.44939964458900383,0.6801952060025425,
 0.10455172829788917,0.033956245213118685,0.6762196217525149,
 0.08062219203775489,0.9735920427408327,1.5454875232190357e-4,
 0.050549221782285936,0.1900822170638654,0.08399263250756461,
 0.7373763145350426,0.9033536718862158,0.956659316515882,0.8139019333642139,
 0.4527110557775703,0.8390092161479804,0.32233165109115103,0.9349307247361571,
 0.6478564195776899,0.24734140761219936,0.22625914575456452,
 0.37064327441907663,0.05038483606848121,0.5818887762260557,
 0.3245536961309742,0.019277066963251155,0.4003575184711472,
 0.24086122602851123,0.35035240432542214,0.6928496641156356,
 0.9554259900617517,0.48685770838442655,0.9136939935326155,0.9531749591173404,
 0.7333979668051799,0.23082853016753224,0.5761321997563497,0.3283154782682671,
 0.7194993983485398,0.47828333513599575,0.567214511858575,0.19220138210236404,
 0.9855629866616535,0.17199590798277586,0.9081002348655255,
 0.05350746609830159,0.7962122827105234,0.9592024000390998,0.9408558217127129,
 0.23568634956624068,0.999427527482797,0.23709601269283176,
 0.22176141603880495,0.913195282131833,0.6706587360051895,0.8980207446418237,
 0.33717325893160977,0.8305324952796073,0.25998399432679586,
 0.34881227288993055,0.7516809211962121,0.49658425624661506,
 0.8527435682346205,0.4956679710502133,0.09344514839808282,
 0.05543593424250892,0.6252942710326985,0.6894741684111015,0.3787195826089632,
 0.33676875600703726,0.6983527804679002,0.19059629876266126,
 0.7890749202156477,0.8913129363705251,0.2350409554625652,0.11301031137608719,
 0.17577641508052433,0.8178900464986777,0.41683967139512335,
 0.6263445374317674,0.3486802501129267,0.7066938448747083,0.521592555716076,
 0.333752733265288,0.4585674417194181,0.6306281745826394,0.37854538537583715,
 0.09607464496297402,0.7772665795081141,0.40247503701078247,
 0.32400096761718866,0.78259879935337,0.6692629348803422,0.7586997738846164,
 0.5923843485248489,0.13109584747385572,0.5411926905912179,0.735529961171767,
 0.5720619028114197,0.6464744956330736,0.16228894118335957,0.6257640841961359,
 0.1891258920091503,0.17834656178625286,0.6137017553864073,
 0.48675113830910766,0.614129132333865,0.31417191604001726,
 0.022668211932272975,0.3909901284179983,0.9811923630312783,0.904704762767383,
 0.9348947742291582,0.9799107766232376,0.055047392127451134,
 0.7905851445561263,0.1130407634549957,0.12088058754033337,
 0.28547625948371147,0.8432438133542052,0.14767630772216567,
 0.8887452058536119,0.1650908769744378,0.7631086595316918,0.7950194981911556,
 0.3379581397224808,0.5330105955254484,0.8209556935465979,0.4729087713137513,
 0.8216583633823441,0.12439063265673878,0.23705055956990562,
 0.3631879557274561,0.18400869320947721,0.4354993021354657,0.5146633403737826,
 0.05141200751417596,0.4407873612412738,0.7434468013729367,0.1378769387370271,
 0.1518683038444777,0.5604052505253496,0.6472772374516539,0.9012068316650376,
 0.20685072489878276,0.4321091510077617,0.9952284429756735,
 0.37673704493616766,0.4417492600120835,0.14351329531598112,
 0.6431340997988354,0.5182878454656945,0.11529059514249695,
 0.10084977176822163,0.37064082460087466,0.5731345979667254,
 0.8687242195479468,0.5886686570032922,0.6753245355296742,0.5250027785894255,
 0.4042061846410837,0.28952787864888574,0.872390650560637,0.6225658424944698,
 0.39515583398262266,0.212532147239202,0.3187607260748879,0.19957538925663165,
 0.8719011589944887,0.23224040302902438,0.83499353182714,0.6472762955596313,
 0.4765916792728232,0.18752920592123257,0.2766605202980552,0.6299811277103371,
 0.7662319202793244,0.7368533637297597,0.32475827230570253,
 0.27317217790285653,0.5229687998067868,0.4801337324930951,0.8175618871659933,
 0.20970170721170145,0.5139998237001185,0.8679424666044937,
 0.16269186404952507,0.9836706745525547,0.34306386274745093,
 0.027706927084548028,0.8666921690490502,0.9374266179170223,
 0.17130743472641452,0.460973559254652,0.6516615579376838,0.0870090566988998,
 0.34478431229192197,0.8048192191183775,0.28741052432185177,
 0.9283974146951922,0.0462475652857941,0.6826559340869048,0.2249519793320751,
 0.24945539397618643,0.48893957363310536,0.9860461618080335,
 0.1178110333970197,0.9842212961166433,0.9004531426191297,0.48080018842481964,
 0.4834785741536488,0.02150004728873922,0.6578877191782984,0.7722731800554783,
 0.8225053769250612,0.48442675208377206,0.5059861878331723,
 0.10326584126286165,0.8216606830222698,0.1221500503513826,0.2470823594392123,
 0.6061996174161363,0.2615134133366499,0.3361993106602644,0.07790440032229429,
 0.09261416707635362,0.22582643128217206,0.5158298857924977,
 0.2968747064036197,0.9601032761921592,0.38577555301810684,0.7945423062396136,
 0.9664383136479034,0.9131433446788029,0.680223389160766,0.41434450176676785,
 0.15337274715202842,0.9177361621652147,0.9165592525296837,0.23116769324914,
 0.7318694289148248,0.6616000288377935,0.318551502553416,0.7916587578743266,
 0.24792494473055082,0.5366160610912574,0.8577401046302818,0.5726731596144022,
 0.5553870245089778,0.20450981425642623,0.15313028127814965,
 0.2372117602852759,0.8732438997980033,0.9647464829370527,0.15844203429670278,
 0.33625060794479045,0.005139705776352255,0.1539095614624424,
 0.5090312111938837,0.357496972090632,0.5641667250588018,0.3188177965846345,
 0.615519156605117,0.0738625264369489,0.2489110604931949,0.6559918154138028,
 0.9990504295315418,0.1551305508255506,0.45650861295740497,
 0.01012357658702645,0.9748100609490729,0.008195056847421056,
 0.43882469387373957,0.6359254531124126,0.7375397449648453,0.6957380042385598,
 0.33277248931492187,0.8538757525019339,0.8077833581338316,0.7610579569219394,
 0.5004554443817816,0.0631159978150938,0.37024575194595355,0.7829358376399701,
 0.5419112084325282,0.0713654980778099,0.7692033646555738,0.8140813235039666,
 0.12745056360625917,0.1659636794782252,0.006425324046637759,
 0.5958794159892947,0.15354906943626045,0.13522073850066252,
 0.9366519080138933,0.6413499091526255,0.6367183830407329,0.9875106296066312,
 0.668402671916637,0.03545418612847484,0.667758016832233,0.06666041515653764,
 0.792403058953266,0.3085321110900887,0.8573593506150226,0.6288708949332528,
 0.4858045551100486,0.5553328048842088,0.7913680144078701,0.5355037280219636,
 0.370360330731087,0.8071372808882699,0.07714779154730556,0.47976518753989195,
 0.11083766110796267,0.6722138621159649,0.40430211150335116,
 0.5664320749326387,0.2939978071342081,0.0905707445118703,0.6651832899117524,
 0.8286971161789419,0.3200754340532934,0.9544271371590825,0.5245031757893595,
 0.19663598033794527,0.8571594959654981,0.7988259606423307,
 0.17931162982801785,0.55318195596027,0.616205533718385,0.8794724376642492,
 0.46538301495250267,0.6118267993411933,0.09983328198947705,
 0.07422285806511908,0.141453018575337,0.38631217544421914,
 0.17261516561946588,0.8577005378629808,0.9294343336071037,0.9844194425674408,
 0.05086002830981595,0.2363197601750815,0.16745130733012814,
 0.9337540437475589,0.3185315972577665,0.53225973043236,0.26211075589150157,
 0.3033247184560647,0.6431287454595417,0.31456960288975255,0.8753369678361357,
 0.19849223737519395,0.10398433201414004,0.025216642583616444,
 0.5977675503975753,0.1262193844877617,0.7077317113977225,0.0945567377499047,
 0.15875549370651765,0.26825084541409483,0.5546244638233232,
 0.6703577868517425,0.7311266275817956,0.11587660104782882,0.9255122221659323,
 0.6875939007310796,0.3831971880039524,0.22275693489446957,
 0.20654593849954228,0.3025205969979662,0.27921458284194345,
 0.04649236526543865,0.9221408776358089,0.6791052591759039,0.6736289100060748,
 0.6277849837847058,0.026457542688301472,0.5888409574020683,
 0.17106846074669102,0.017369399656184115,0.7169077379059116,
 0.16708679328865628,0.19534287595658129,0.8818056342259808,
 0.7034770581052908,0.9611306358388451,0.45685889411929137,0.4206032080690829,
 0.6606044149243052,0.25437865836396,0.8464790232256677,0.09444717193987573,
 0.022764977397410235,0.3306220016455108,0.5374813641588617,
 0.9307773127376209,0.9334102729609233,0.7208191015013369,0.15131377482639385,
 0.8543898147764506,0.07628495379847688,0.6554876199290822,0.5309821450913029,
 0.5182109129478244,0.6633773818371527,0.5593955324603641,0.8980128806041965,
 0.3338920976240817,0.6219634343464827,0.18444731840437367,
 0.31092045926279266,0.8534977304363394,0.7097363445287654,
 0.48836601938696456,0.8893783877302107,0.5550637344188415,0.993113332405494,
 0.27660711652466596,0.645513254540487,0.7032729605258583,0.21203592739644705,
 0.4311181617823976,0.39335431924485187,0.5786690117929236,0.6602512954064662,
 0.1623369551164825,0.1110313359825218,0.632117108908996,0.9415399161798956,
 0.9572430272980483,0.9311066967734057,0.9734208939431466,0.14882362412777173,
 0.08399068701932544,0.9981419040642318,0.6224155824663713,
 0.42043538410665315,0.2309693916303983,0.531698574064916,0.14641134885517548,
 0.3863559555535738,0.10457820113999994,0.08866140514528209,
 0.5420087997758228,0.9746244206308663,0.8638208268135027,0.45229019584597985,
 0.17253440019799227,0.6235042262264504,0.5536629562379837,0.6504751022201399,
 0.6643892573939167,0.46722946982551383,0.07197501884463131,
 0.18383632746154355,0.3524142915521429,0.13652921746937774,
 0.26415603111035235,0.6683925105707619,0.6609978260911226,0.549887398704614,
 0.8547140664400916,0.9885149576781116,0.47647337487404195,
 0.15959924964506222,0.6725924334976832,0.6511427311111462,0.5951529413796758,
 0.9394878540327634,0.8463309732449501,0.47859838579005887,
 0.41263305341253964,0.2923664219451907,0.22220896535859969,
 0.6154464209827147,0.12480523188409576,0.5557789335174863,0.6319469815085641,
 0.3962528292791142,0.7085382327307158,0.49562028455108753,0.857619480054251,
 0.6331896918542672,0.44470300752843006,0.700661350757956,0.3598296044625938,
 0.9500739859292631,0.3291910134490056,0.9135276422635088,0.42401669196835756,
 0.6353244244613716,0.24815432252991665,0.8889733930297257,0.9461640451113378,
 0.7699029249614715,0.8842537940335699,0.649888973136938,0.32188172955688177,
 0.27058129748885207,0.5550662245527229,0.708307145533752,0.09248747388731904,
 0.5295743007926685,0.8357098290025374,0.270699108513458,0.9732888731531889,
 0.1309917437165623,0.7367618991840365,0.2549430386721173,0.9022752375471859,
 0.8113701801480316,0.22455680770052966,0.9361567406425741,
 0.18451118003010758,0.3165735320543831,0.9319506936366758,
 0.07052785725566324,0.05709499697419662,0.12839252581989646,
 0.9793672816190545,0.061967123539387936,0.7677356806324219,
 0.9954368216631663,0.42834235404207943,0.45609812352791634,
 0.3677293935263115,0.6076770091864285,0.23414207079449945,0.8370676770282915,
 0.05268164363000627,0.026687453220226143,0.5702859635670768,
 0.7412910283936401,0.23217817469655355,0.5970175328207326,0.5288146802224932,
 0.3973032643589238,0.8780039363101061,0.7624118715627287,0.9195322033634644,
 0.5186362648579076,0.8327913904092342,0.5636683592806465,0.3905872159914683,
 0.9220145190382955,0.6223970070560207,0.15550447798201283,0.2079364699512063,
 0.13694691448580076,0.4983210680573321,0.3872643336238669,0.8496353834935656,
 0.2489759730365998,0.6556292899953059,0.2708049790555689,0.3731465117084435,
 0.5149603007700665,0.9722668917387136,0.7097090699984436,0.388652376331437,
 0.2650366130608077,0.4512926349333506,0.5217564391127896,0.23832730850710593,
 0.5802156901647855,0.3859933408427424,0.48122778776293507,0.7159703534114047,
 0.04201963343656678,0.8540950094831969,0.3939566656409291,
 0.47121291673883836,0.7917104498376611,0.45994806582269854,
 0.8290513537570945,0.43041969868040486,0.40455835351110603,
 0.7502122516550535,0.9683782822791983,0.3079672694223494,0.7164272278477168,
 0.6406235340964366,0.9292714687781403,0.6887813478644932,0.9428665382835204,
 0.27530508884061977,0.7371415113485335,0.2550164092090167,0.8528199636356437,
 0.3876380552620483,0.16683076301573008,0.927213400365544,0.4811750134558844,
 0.2902037568499657,0.5207394067475363,0.44559428556086944,0.6963975742206068,
 0.492766390443081,0.5497879838626573,0.6247971613991414,0.10292923743087512,
 0.8940728903586502,0.2729496068860662,0.16966368077528493,0.0696344161808945,
 0.2578185551687603,0.7375632675077255,0.3190219568594206,0.7820624303797237,
 0.577246481019784,0.9239388039292875,0.22786464842975673,0.27437937721316663,
 0.13896397232710345,0.32487360377010677,0.24532633922816405,
 0.6097705559325955,0.6598744778689094,0.5781709676561826,0.31426025695810755,
 0.19723254312835725,0.620596632751438,0.8635790427260965,0.3292339779912582,
 0.2938353499647217,0.44525410217534356,0.4456186066416875,0.2752394062134025,
 0.9545073496675107,0.585383770496017,0.63039820743821,0.18177646120495972,
 0.742845742762862,0.40735228844799565,0.1993116437109569,0.7223629946721634,
 0.47242374546804455,0.7510274331900314,0.5357054851525486,
 0.47849123886367306,0.7504256056290943,0.995752943099015,0.3550950507036772,
 0.8138068791198397,0.7090373784579996,0.6872254606426346,0.14704997198132852,
 0.5489441190694998,0.7050189746170996,0.5504196513051385,0.22003667206413302,
 0.16192101005528303,0.5292890786512809,0.4823674795361026,0.511938080162709,
 0.8644480674471972,0.03398525564968946,0.3065069948554138,0.9240837444165912,
 0.800288923827008,0.5629340855764741,0.9620761622888294,0.5170308607029183,
 0.9327145946819859,0.3557161267194925,0.5303046194512049,0.7054967904241283,
 0.5564144890742613,0.008480358360464235,0.4230888604752332,
 0.8308898966448913,0.9310993987636602,0.4109033835117568,0.7949058750566003,
 0.31653503598462085,0.43519115954186516,0.9553710901749338,
 0.3743688493527917,0.6369197001077053,0.8125413549113639,0.17455692299342362,
 0.006584298327842131,0.1924246751581674,0.11482505369127116,
 0.7758628195210732,0.1460048623066198,0.6661477278847191,0.5942140617892775,
 0.09843116211983383,0.3484405221455835,0.1141989881310369,
 0.041694289704904186,0.41690610908856174,0.7349175228407606,
 0.6086920689946133,0.19530527707703937,0.07221030408466722,
 0.037063228499408574,0.07875896202953425,0.2903384047441233,
 0.4060577413376857,0.5043366057259133,0.4194028726497838,0.7515624255997789,
 0.6797001378179625,0.4359135027448532,0.20979159294071603,0.9872654776655853,
 0.48394706559546485,0.19244510789183633,0.4506940445087807,
 0.33585676308092505,0.5028263318162995,0.3205468512708727,0.8747873504148849,
 0.6811329440842282,0.9392647158082349,0.5629819063095827,0.5903498658388743,
 0.5257819018162931,0.9913034399153954,0.24481948935675157,0.5062313210001047,
 0.8100795896928552,0.3472471683844014,0.6768270141215527,0.3717417075874737,
 0.39961690153701857,0.7202898496634369,0.4862463207042692,0.7583647587719191,
 0.2921799272526844,0.4782666232072401,0.5074330057350617,0.9731069533367849,
 0.005539861046078287,0.8291570418291796,0.033921628319922514,
 0.676243199330404,0.9511814721462876,0.861342097797451,0.409837189182513,
 0.7261961646987568,0.3004934083849323,0.28385957932324746,
 0.11059659465540528,0.28151003911550443,0.4988821717383535,
 0.5259723404087739,0.4399206619342445,0.4699686944359146,0.6744768876177073,
 0.3140527392280633,0.28584558519023073,0.6043673699948954,0.5615253599921656,
 0.7444988190590902,0.9521010631203782,0.424632801679357,0.11585476859907673,
 0.39120379695232543,0.47035991195406,0.6695631215819587,0.1465554288374542,
 0.1744377817684658,0.9381067251682693,0.1927997601179413,0.2764549945520721,
 0.6534810675028061,0.9344674787998004,0.009088112326965603,
 0.6322266829850373,0.1937977229673623,0.131103407053479,0.050763589870285286,
 0.668261806080738,0.5655621301261271,0.6534584340613062,0.33499486980837545,
 0.04837348009040365,0.936473818052368,0.5628665043311911,0.4358948488400948,
 0.6618444910244785,0.8958204068806976,0.11947029200015347,0.8166633109357493,
 0.7061088203215418,0.39568605612297414,0.02816918360797882,
 0.3178712564508611,0.22957425641406382,0.3098869647034779,
 0.19585224847450666,0.4629565983108045,0.181889753600744,0.13954915170117177,
 0.0889359881930909,0.4682747282258338,0.9350519103265175,0.49110882468851125,
 0.9429201325873484,0.9738078685969177,0.1784548924809921,0.145586736575531,
 0.46581662807328517,0.9840974819012691,0.05644090281737768,
 0.13356439975207923,0.2260306443837028,0.05601205095972417,
 0.36753901183812465,0.415809115911142,0.7612412085631837,0.9515913006301836,
 0.01405359777205828,0.29754121634571384,0.22541667977713908,
 0.9564522551027159,0.9944377020714472,0.31852429366514656,0.8660611759653177,
 0.4454805571230357,0.27521369148974506,0.7982449480536836,0.6839226845693784,
 0.2154943955626869,0.20200317103102816,0.49808115888094584,
 0.2682763753653742,0.3947597274721766,0.13895318415994407,0.6815826952425217,
 0.2770349133844562,0.5901916079505707,0.2037129849628695,0.01882080791134244,
 0.018795860828942246,0.239499411235909,0.02037519409066435,
 0.7770258080488177,0.9309988538087087,0.6472189205845884,0.6771012700925968,
 0.28781421481751246,0.7306739401620385,0.9522667207046374,0.6606691537601365,
 0.3787346441102738,0.8852000879609883,0.2509997290468763,0.9952071672354633,
 0.7198480125749243,0.3328041422695601,0.058034622909611144,
 0.7650194864367618,0.11491029099278771,0.48278048390524286,
 0.7620103557739444,0.01548771975709129,0.9194765535925762,
 0.03638916157338523,0.3848400387156198,0.30484981224147345,
 0.28817895595705434,0.23687926682871363,0.20753422679166955,
 0.34771192837659326,0.09501091496783776,0.9191956189792329,
 0.06440905659819329,0.4413558776532859,0.9210356754222605,
 0.008266899963993879,0.4774180126914097,0.20535745352778978,
 0.6848611876176429,0.5018534503849305,0.15983607286135781,
 0.06634671851438201,0.8547514476161487,0.4999099707674173,0.6838012176200471,
 0.6648135021189265,0.8843207175750123,0.38287578889183405,
 0.47909467791001137,0.26403750869565856,0.9065314991592525,
 0.43423327587346305,0.5781639700501013,0.21393498150900747,
 0.47469529272427224,0.1482749313649835,0.07350478182704367,0.85468489582599,
 0.23782665168658057,0.27961152813948353,0.9862824767155904,
 0.9693974167387432,0.30002319929314736,0.8110602682071044,
 0.18340709434054536,0.047205622007033154,0.1570000633637828,
 0.2325687507607852,0.6685216498748877,0.9563101625549734,0.26753964126860197,
 0.9359518442225314,0.11848608639021341,0.5827968194976009,0.4838795073673443,
 0.10305494537469417,0.180141052165129,0.7187061546742911,0.16384172388984197,
 0.7666197462808045,0.6920296904671592,0.12216381618591177,0.9063751274785073,
 0.5670326580953172,0.8582020907911002,0.8824739838500434,0.4507173903206566,
 0.8411222296398565,0.9535210372654301,0.7051974313047903,0.283220574047653,
 0.12598122395799982,0.064689351758354,0.5959842696905585,0.275815151865794,
 0.9272556125092102,0.768818155484166,0.12424998017837163,0.4719639404225946,
 0.3338486420122432,0.43087228422269486,0.6592195229045389,0.1428061251306082,
 0.8329305693011196,0.3990194466120509,0.9499733425964799,0.897118245911802,
 0.14396212349887316,0.6967784160679436,0.17406630536173762,
 0.6406036137002608,0.8944217909480813,0.9616518209563397,0.3670017165863172,
 0.09701689932701396,0.17374441098136018,0.3010930882840479,
 0.49950147132208966,0.6960961314455069,0.30060137123128094,
 0.04831501569534702,0.3408374781546326,0.2797525461693494,0.0116196406702056,
 0.41445785663777046,0.3072906826433476,0.737673968056409,0.33248441259211037,
 0.224844845618122,0.8958944325266457,0.3421783336355143,0.6595388986977346,
 0.9285128780920356,0.5554392177122737,0.7574732396763556,0.882679466185506,
 0.7640116430589381,0.7253345333401454,0.7311130807066816,0.8956117116080784,
 0.35026931746211365,0.019153287215703863,0.20195650805925847,
 0.7030624803010705,0.6283158220922058,0.20860432143279395,0.6982824058285918,
 0.847099920058792,0.3982718903904594,0.009995764989536138,0.4955412999876564,
 0.534208599349829,0.851917192590761,0.7140804047700146,0.20519579278655053,
 0.48800450772221404,0.5696067163623469,0.27302278150950476,
 0.5039693690916093,0.3810478813419803,6.079759725315936e-4,
 0.02981236255597397,0.3385068117960245,0.3917521417687455,
 0.30375081832367845,0.2942730759614486,0.31402482911948726,0.576053386357258,
 0.7117290463350676,0.2509260684167123,0.4497191265151895,0.23649447395100354,
 0.6109221496435971,0.5766380812638953,0.9222918697600921,0.8415926133180249,
 0.9111300517445948,0.9764990733927319,0.1890436708070613,0.9814417465248143,
 0.5982007397753898,0.12403556272372518,0.3906357379621188,0.6832078967789749,
 0.4887838594607543,0.3530286074504132,0.45931719844168706,0.9148386984918412,
 0.47879815446529894,0.306060834891547,0.3961762417172532,0.09388214743077639,
 0.777001963093918,0.4245563170994382,0.2416526204560846,0.23996680426717298,
 0.24238525671689293,0.5893738957170898,0.29888715173422387,
 0.2165579643381369,0.697985160833546,0.9513184913203738,0.8238608993860655,
 0.7448419351554316,0.28563094018836,0.35941724263138486,0.990802520519687,
 0.5398322437370001,0.040939190108301826,0.7270782788969141,
 0.5092563033828005,0.49522132543515873,0.6248540935949531,
 0.49023776141136144,0.20831281633402798,0.16833592333110048,
 0.9690810603870725,0.22898718988969224,0.8884768052506338,0.820854523386996,
 0.7254082807783823,0.29585866095786506,0.6403994351044959,0.2535717372063193,
 0.7361022868632123,0.7241335073682538,0.4678637870616833,0.46658156411637497,
 0.7389817947643542,0.10331924968903738,0.2016215955536571,
 0.49618272962269394,0.7249381675169663,0.008149245938692151,
 0.2604550794599054,0.9684454130995622,0.39948602450089676,
 0.28209401394386246,0.1948057932722631,0.7760182948749493,0.8254196610884205,
 0.7904484309991663,0.27526596172031725,0.7446653861098752,0.8075216767402722,
 0.689666004126436,0.39547013250961716,0.1875423329833813,0.5322137692205198,
 0.36649479420579245,0.3155392786757293,0.6438394547189532,
 0.07064529102496941,0.7175087887513671,0.7746636310126479,0.9492142324035437,
 0.9679066225236005,0.06677184481706655,0.9905515841061802,
 0.08664040538470863,0.9650639603158492,0.8758071049210241,0.2849008274302893,
 0.32680087405701974,0.6882156155771011,0.07057493222913525,
 0.1081189477392639,0.5712385146198834,0.3745439932632141,0.35474696478500556,
 0.37747894111825453,0.9757251558224291,0.0025920471927939337,
 0.3028121849089489,0.3936216278435041,0.41006295378673907,
 0.25923662413054227,0.7245835070459253,0.5133427708827389,
 0.41348287029833786,0.9317879238919543,0.4927526230428979,0.3114137069122701,
 0.2115943812270138,0.38619594539053326,0.7782847692861329,0.9306036513035031,
 0.11183256983992251,0.14870108024900153,0.17962696085330287,
 0.2395321317980279,0.07074852697936473,0.4697861633128846,0.7239170807254757,
 0.9708688810071522,0.09480408425665865,0.21573378663956677,
 0.8849517465738368,0.6423015326254831,0.17156914888044894,0.7413616416111706,
 0.8132062306527557,0.9515611263124237,0.09598817898039247,
 0.33772876007891317,0.041180096930595544,0.9167896163230319,
 0.5832692319288143,0.9182993375179488,0.2867064480862713,0.9170033095505706,
 0.14299989818973136,0.0025656962682171214,0.7967517832322921,
 0.7481722853219203,0.8156689467998125,0.5658902761269301,0.7083462743159542,
 0.3346566820820357,0.6664685046880232,0.9612511378705708,0.319291829513054,
 0.27517127704285094,0.38070616789541034,0.7179374084191819,
 0.24125270779397878,0.8424196212516983,0.9627574916423635,0.8915835645494453,
 0.5283424246541468,0.531119240498995,0.5032396282048964,0.36153579151019666,
 0.1260881433091433,0.2973094696855534,0.5133656633559791,0.9251311395878299,
 0.8087992580494426,0.5679380167563532,0.8330519655407176,0.5657725050047313,
 0.6267027296873622,0.06455029198568174,0.9427788938634942,
 0.15511677240216204,0.43976622024198364,0.6099072357577962,
 0.9857052077085992,0.8889676147064741,0.46001992965271443,
 0.22913947292159742,0.3335691512119856,0.6707482409849633,
 0.03635142241210243,0.06856286836765957,0.7790585088945373,0.52181375189132,
 0.6489979752975285,0.7911823501926994,0.9124355052530835,0.5586528215911225,
 0.4949667787306664,0.9242276360303219,0.5604394783455802,0.6273117640487543,
 0.8655417268302845,0.7183441831771074,0.686730154204231,0.6808829005982535,
 0.42045531702737315,0.024545559238562276,0.4257523136682335,
 0.14232582355356516,0.5056691408767949,0.045630483933865706,
 0.7533584621186522,0.38974884958740186,0.34091312947398433,
 0.3488911823276061,0.12525636427101194,0.594811822787059,0.2085764803813528,
 0.24636037509751685,0.5415529326901164,0.0855001655877029,0.9962035510778174,
 0.6659311058892726,0.727378165918032,0.5929205547183861,0.001091162448314531,
 0.5287739162499787,0.9105008404020792,0.8652139750077681,0.6290090991408667,
 0.2566427049514787,0.4922446882109124,0.8238528111386337,0.9617629128825561,
 0.26736817237539623,0.5345001463562035,0.957442797154603,0.2988373437368773,
 0.9858091471824569,0.8015631775383688,0.18072569149570494,0.622234915096779,
 0.059556854580104734,0.2752524146708214,0.6705776423010394,
 0.5411049523247149,0.8600302626765498,0.8907178530626398,0.9607042758561448,
 0.7953458656701916,0.8354236937143414,0.12281171617741371,
 0.11805206356244369,0.167243648439937,0.749291204431167,0.5505894812293024,
 0.7213328246540615,0.12339422110524922,0.10015337552335368,
 0.03239369759595423,0.49311367363381975,0.38593367936370093,
 0.2349275791402341,0.9761812374098653,0.7355871459632926,0.06360102069241924,
 0.15227407402319038,0.4458914367804272,0.11248195408808792,
 0.5812434998763637,0.4554872919871167,0.29799320385702543,0.7146272905671376,
 0.3153889564676685,0.4597117704861411,0.5204417172167143,0.28526598289769645,
 0.650571794505939,0.19855710401735938,0.905940062434866,0.7340131957546193,
 0.8007945033774678,0.31713943348769014,0.20865726967282316,
 0.8042583744914383,0.5276538296741139,0.8185324390767661,0.8998730632829143,
 0.39274827650906563,0.2720676068002129,0.3044569616025591,0.7522499001497219,
 0.19754257532813768,0.02651623528792113,0.24878092427778198,
 0.9233339301160015,0.6199685594801269,0.6225752530871282,0.698180677900013,
 0.4015199468608638,0.6144498987597422,0.33456209328618336,0.8910027371055823,
 0.5838027008992384,0.16659578746943227,0.43101992323126126,
 0.4005558505281307,0.5444005337967948,0.27511994801995465,0.7113037461747687,
 0.48887214611823027,0.6435307909214593,0.9660195638236753,0.5306023414023762,
 0.23415173236353426,0.18123401533679784,0.8577633430593428,0.789401284756679,
 0.9644785267443795,0.039304166630264925,0.34804206615790134,
 0.2601660810178076,0.24749209713898834,0.08387378390609612,
 0.18448190754535632,0.5141960821228185,0.5431777936650736,0.9482510929692515,
 0.8790829448221411,0.6561469274233231,0.1435904119785777,0.19482049790140799,
 0.6350439198487041,0.3043703117905452,0.19909298743882853,
 0.21213150096609912,0.2557988043612853,0.8737205895340647,0.8837839750758675,
 0.7629338196055333,0.3880064636025,0.42833201664089404,0.12115822977543234,
 0.865860075219313,0.46308457209428977,0.6898644586558713,0.8128581534903131,
 0.490434003172184,0.5443181389771403,0.992487531744782,0.04870726871785158,
 0.6582714760852756,0.18836987013349782,0.6827015233003418,0.5692488248512144,
 0.2683864722560866,0.4816442836985978,0.8572060724143676,0.16477172568861076,
 0.7581220682424027,0.6024096943936528,0.5884049265658426,0.955527915034916,
 0.14914898954117473,0.3855238325955819,0.8934695604867526,0.6924918606743268,
 0.5946946939339229,0.48900999207199025,0.38436109630554205,
 0.13022977479659403,0.16197157395626438,0.7289498711225342,
 0.6593502394346036,0.10014439696635824,0.24454811611562266,
 0.027168215078460567,0.535524541709449,0.03878861905203279,
 0.5779492736109693,0.5838082110243447,0.46681732843752655,
 0.43833482186649486,0.27992594115694525,0.07992424319669522,
 0.6924664447538774,0.9140010279860579,0.048865774104784876,
 0.5576657713484663,0.4129911903081156,0.685606877245348,0.17284430590142086,
 0.6691093572189848,0.43830872647165153,0.23871960733667752,
 0.5641060302615822,0.288393296704738,0.3251236061360797,0.16145810621456547,
 0.2902116169996536,0.9532136808065137,0.6632780549061501,
 0.006850718427214297,0.21659933119237929,0.862970024138525,
 0.17417769362593838,0.5762714789079989,0.5311420471878884,0.8997753888980962,
 0.4722587638614939,0.17393796270433515,0.6281160523956253,
 0.21793103275873404,0.19049304778835296,0.9237506711182323,
 0.9695010893084498,0.8144670916373412,0.9946754340254041,0.7160345246415876,
 0.5101795841862149,0.46918512840973725,0.13527343951867787,
 0.8941872102284886,0.004524901951461269,0.3416130857751458,
 0.9369509225649287,0.38764468627406357,0.8976330255855067,
 0.39442831990203886,0.9049486380792822,0.8350455949865484,0.6629530311701187,
 0.9219852851192353,0.1469159007495897,0.6156232507531407,0.24789342580546803,
 0.9634212432525874,0.18846425039792225,0.5314386365155345,0.4712224518971738,
 0.2014749741006182,0.46169642604178396,0.7061631707961828,
 0.45142093866086896,0.3319983576068538,0.17680535802150232,
 0.39458454365481554,0.3116820795018618,0.4646771837974809,
 0.013274802613259196,0.6009355655920564,0.7911767007538403,
 0.4556821912143556,0.7960101881148949,0.8922201260462657,0.15705540427107834,
 0.4922520165073273,0.5266036712310294,0.07394468968139567,0.8835916967316712,
 0.3461172038851502,0.5120576676316811,0.9498557557227503,0.09305008547469684,
 0.17661572680176207,0.9704012615854831,0.13419629228426522,
 0.04785974878200949,0.08310707386485605,0.0583018265251638,
 0.4247656834035337,0.037495102302759764,0.5832744645940302,
 0.10077358876022591,0.584367270338147,0.8337593529012179,0.5912749703205068,
 0.5025240659988475,0.7566271797434531,0.11995015393856212,0.6791354193802255,
 0.20473095007826636,0.7933914377142096,0.2893315360584616,0.8667546462004433,
 0.5115292389388304,0.8785186751438515,0.16983174891269792,0.6601823128666896,
 0.6387828200808158,0.04164763027711604,0.06519078325256034,
 0.08165104142433122,0.7005301900506434,0.4677253516957298,0.9917394902220622,
 0.05426762361649318,0.2474509514118337,0.33314736363656894,
 0.1858710145613487,0.08459254630104851,0.9742420012680592,
 0.04350704535846406,0.9126602817722765,0.8053722174118249,
 0.029711473446983105,0.5507068023576899,0.0973386572457523,
 0.10918368972278636,0.737450670365456,0.2965835978708502,0.6522942603272047,
 0.234519533611539,0.9076103336375391,0.5199076654071934,0.5988874593827725,
 0.5988236683950872,0.19964734158915176,0.4024358071985077,0.2904712740704031,
 0.7033488087345867,0.5872795696409241,0.4344035866357153,0.5550936661090787,
 0.8870560442160351,0.1855488760222922,0.07640887157236076,0.2005048461003761,
 0.3721242349785757,0.4280228274264919,0.45141895437563906,
 0.03237149126717065,0.5979206565071259,0.7389822132307933,
 0.43846120154666635,0.01984808763854573,0.40415247608955884,
 0.9329456308411613,0.9579179921276524,0.5346524917182609,0.3983798121244233,
 0.5822925854739611,0.4392278816837675,0.20484983112144906,
 0.06416651150207486,0.568362760915212,0.9182985020280429,0.9192854950081879,
 0.6153761011528761,0.4121355397903973,0.18603432192449643,0.6832450058503512,
 0.9998373252400978,0.6748021583673902,0.3149502594440644,0.27710521070065386,
 0.5294620011573578,0.6757699679852158,0.05485222823476121,0.8644848205551807,
 0.46017071696809053,0.21485931450088847,0.3953155165870089,
 0.5805136582742125,0.009086553027437105,0.44526818981075755,
 0.292500534407491,0.6205952264825565,0.4545313849995368,0.850200370527548,
 0.20904458532163994,0.18332196325362005,0.3741891947485284,
 0.5856528932794155,0.06935661715548536,0.02203144059448514,
 0.6481303597852011,0.08765825410089645,0.24802402620913655,
 0.40155377202927656,0.05227599675255601,0.7804365707988418,0.953622381113654,
 0.07541524816525669,0.9587739259100115,0.6365953662673979,0.6886294254834727,
 0.8232842848281272,0.8379591240997523,0.04922967207590878,0.60789070009893,
 0.6001031793858828,0.6756121712673231,0.6069796235219791,0.03389196790427218,
 0.1651274673699492,0.8986710014122247,0.8982711013061662,0.5141238683177277,
 0.05747487924057115,0.3397145131774174,0.8179840626377451,0.6622824465419086,
 0.8127199987531679,0.8093201627717904,0.1337859796665184,0.5715292333698867,
 0.2016526860988581,0.05973971848863002,0.3764655238976079,0.6892039942259042,
 0.06833338923395971,0.5595711997435937,0.9842142992738779,
 0.17416928111949115,0.5514611769626846,0.5424601224768439,0.5167189618340642,
 0.14907358668606374,0.38952421791968983,0.5434568591194253,
 0.013419518785574347,0.2992132933028929,0.9512659255103459,
 0.6740805585777201,0.08125147869047089,0.04432204235372117,
 0.8757037987327514,0.9239688245613598,0.5381926437549083,0.15219311186311935,
 0.2966278976013442,0.2728464264314894,0.5547060935969376,0.3995215830939861,
 0.5872238500938741,0.768205972034485,0.6120742187079288,0.12268207783011498,
 0.8895058260414938,0.7588736855104173,0.07647132623845643,0.4874523993023804,
 0.9320887072990649,0.8422069975970402,0.2430098616710633,0.9365250630717648,
 0.6956935718482447,0.8047307225132562,0.9535377841944148,0.5246522507123645,
 0.8517842898855468,0.5613827125171277,0.2835238397716029,0.5890040287306726,
 0.8810460791307007,0.11019997318512931,0.9304992687271301,0.4797798093905208,
 0.43453971333013075,0.7829271196015368,0.8168852579397714,0.4015914386169369,
 0.46704203913391373,0.15685882013646424,0.05734633717983206,
 0.7077327552711101,0.7050760114956375,0.18119449415909994,
 0.04767392747493837,0.2694475117526661,0.3191312896375209,
 0.36981713482304635,0.001458263283650063,0.15877299869507122,
 0.7787038552152388,0.49444281390960454,0.1919827034849635,0.8962801072423177,
 0.7170619635303186,0.6555516390671222,0.38394005357125705,0.8556610256261337,
 0.9622064871579693,0.3174908374956156,0.3974660272756928,0.6728184983353664,
 0.30237113479709377,0.6105694367110411,0.5987137487482512,0.3628236732688511,
 0.40661808937633226,0.7089236540873902,0.9493116972618734,0.1538394708586278,
 0.6400216495885654,0.2488737511595146,0.9396518652816002,0.6479314812120944,
 0.2578983699321259,0.14115246407678328,0.7823529438334476,0.8720041345271353,
 0.34197372596252495,0.4023993373785837,0.4645612771716797,0.4651309853530986,
 0.39207601903899303,0.7225914065692569,0.975131612222868,0.8411244319478658,
 0.9352650579258562,9.130435404680615e-4,0.5487602329207354,
 0.37065915889822265,0.00750729595549382,0.3335334295129867,
 0.9927242164369241,0.4964474948614994,0.18409816030663073,0.9654083937614534,
 0.13068358635474187,0.09078234753667302,0.8226030483150544,
 0.2820978571459788,0.4525615761484676,0.8036684786294729,0.8381922579217954,
 0.1118556907525955,0.592752812689729,0.05181965106210429,0.6745059876796592,
 0.11512515286405312,0.37110352861116935,0.5529549147172165,
 0.10455763148984865,0.43261570230574287,0.24177512939974666,
 0.6645543671554501,0.8145049894281177,0.8307177723771462,0.5743271548567481,
 0.7977761042142322,0.273260573826577,0.8122795973828925,0.4252303151646738,
 0.8306988498288236,0.019676565939243607,0.8659435985043333,
 0.43903799654933906,0.2307095175508065,0.7337976514053149,0.592191413956992,
 0.28812447547728737,0.15888751434448456,0.7404235099163433,
 0.21976222185308958,0.18687232498753004,0.2557704639038537,
 0.9609311074863729,0.19699947845486032,0.6388114234728685,0.7909935757677656,
 3.743436939207534e-4,0.4767086728717612,0.3286564734259262,
 0.3089890571713193,0.05570729361647042,0.4378006913929211,0.6770984244380519,
 0.5056438781905239,0.07081685737456611,0.23498372626616304,0.512111889077157,
 0.6038601487616404,0.04603037675033561,0.9331965729343505,0.7992416961649962,
 0.9285361964428241,0.8327044878184452,0.6861405260939276,0.43309679926858136,
 0.258561656408272,0.28808001167529307,0.42109876514086697,
 0.49784718423506014,0.5133233433664667,0.04960132564077069,0.388979186203924,
 0.6015036450423229,0.8488892822318663,0.611594992840735,0.6071810279608468,
 0.309920426188495,0.11849579782380326,0.9561583708396864,0.4159403769499226,
 0.31843439221673997,0.34098754541791676,0.312217482675736,
 0.12064284331658603,0.2613372966683335,0.9671975571590606,
 0.07528333688929978,0.7894267736637912,0.4963873514968826,
 0.23015280309307673,0.3478695534649703,0.15974976052217782,
 0.7977984123719201,0.21438006943379329,0.6290696795621479,0.4970873457153947,
 0.9603115736135356,0.7289834702301995,0.8264853815314261,0.9790984795980466,
 0.8263799260436925,0.2175044219251685,0.6294362486812475,0.11121234799970192,
 0.5993887678552979,0.21726913129936776,0.28165584412223,0.5442062980276583,
 0.4702725993063024,0.8350985199728183,0.3055159258704304,0.6504427112677201,
 0.49757043531164125,0.8633456735461934,0.37420698759230087,
 0.20131060684631075,0.07608910818451009,0.19057000774236532,
 0.3708185682394215,0.4347009055807437,0.6421006416437599,0.8727975963727226,
 0.9669395383255145,0.6618489947724183,0.4853325883932016,0.7480145592024847,
 0.440030453493933,0.18175006437053787,0.1664148651112275,0.7361735936676598,
 0.4435564559527384,0.589539034089011,0.09030632980631803,0.43978961797355354,
 0.5827524909858162,0.3615469395635508,0.4346164558051666,0.9433601891585238,
 0.7407927140519749,0.00639251453113332,0.9147071173968344,
 0.05219608319100899,0.0040607502180535215,0.5080458629796547,
 0.7644462515283457,0.5097256616004565,0.5786251940083655,0.7172097107663589,
 0.998847939311645,0.2289044135510353,0.3461269994895896,0.2694137857992688,
 0.4328220513857971,0.7269396376175722,0.25605465866904953,
 0.09591547775365616,0.27073559146780446,0.3523968573437708,
 0.9706195737655449,0.04089479287333486,0.29961793294112726,
 0.5960642507727454,0.6219434977957584,0.07717720747437407,
 0.26882682576278394,0.8871442970155343,0.5416673078528835,0.9545113790490334,
 0.5824084367964307,0.9742640574685061,0.9669421715586675,0.508916117698316,
 0.7169387754778171,0.14884498955028713,0.7466350538079372,0.7173901831946761,
 0.12778377262628038,0.7539177822911719,0.3007395497885028,0.7730227146729425,
 0.2121798999486585,0.9457393315932209,0.5267122592471504,0.09013993178889834,
 0.7083070212491835,0.7608964646968112,0.6137197405193789,0.7199547698906166,
 0.7340024422312087,0.3265141819732331,0.2536085354990586,0.9624818337527636,
 0.11906261824106512,0.13922187202913716,0.8274913915906343,
 0.2539241676258135,0.20939572921911354,0.16446796559054122,
 0.06708480739757916,0.340515693469027,0.9771490699513486,
 0.058077713742360704,0.09749823737739138,0.5722371165252923,
 0.10922810886494727,0.28662944869016116,0.989023957349561,0.918176001198709,
 0.9118305367927707,0.09883218557304918,0.6800248006956561,
 0.11654484228769424,0.6581466985091102,0.4033538966828246,0.8046770268859984,
 0.56531582001069,0.08964767293890208,0.20046011475083825,0.891994992491409,
 0.7565936855785806,0.32701573916186066,0.34414921894343475,
 0.10552826671811077,0.2268414602367893,0.7556898793834467,0.6056749922078642,
 0.34989721729312995,0.5555889600420348,0.1552795070230677,
 0.14925712907437583,0.5457001183985233,0.2161437131642736,0.4509209735384795,
 0.5408789347675111,0.28596725784602783,0.5636420804112738,0.8816708091183906,
 0.4128895450515544,0.6944208028846152,0.14047804753387438,0.7314760987338148,
 0.5279065531315089,0.11961689663450759,0.5465422725513434,
 0.18258503138452786,0.21365360692373836,0.25275442028507966,
 0.03268436294716137,0.4402048203544917,0.03988371405900182,
 0.7437745029858592,0.15563492994267325,0.263755736720901,0.17420080414376016,
 0.9818602901896843,0.18271203265493152,0.6448031424735684,0.5452317815746424,
 0.009213437462782137,0.3437204714163803,0.7291638613849224,
 0.06513492498465101,0.13862199903580263,0.30998928451397323,
 0.4770997795729752,0.7042054209169786,0.31638967720409805,0.6669135696393175,
 0.07548224054505825,0.042500977441239596,0.7651714270213316,
 0.5861817257859027,0.23979755745914355,0.695243196896702,0.1395740842431259,
 0.040522919301234195,0.4047588055538056,0.08584497666097135,
 0.42276966647476866,0.1306622122222877,0.9731120077511108,0.9974665714376061,
 0.1283482436639518,0.7356185854916784,0.3191535129672285,0.4815493823742951,
 0.27748238335937947,0.8632368148399299,0.6065787269456585,0.9500667385161984,
 0.718948633230601,0.3961604654099732,0.1140968288398112,0.30365266216011755,
 0.5441894225396834,0.9867017680790712,0.5700598025757945,0.9259778489645359,
 0.7953072383561871,0.9684374333771246,0.53186510190069,0.7357640657694017,
 0.9446736256334198,0.08676843980001081,0.2753610125760291,0.9250165277877348,
 0.4735834696678569,0.11066782499071992,0.43633484449226556,
 0.7327964855264655,0.10306694713876074,0.1842501420470345,0.9703563099582327,
 0.8108967046518929,0.6275282692084242,0.3597321431596828,0.0756861997153242,
 0.6665280311272478,0.7634062478207118,0.5675421914270965,0.5450923335044269,
 0.6597194620160673,0.41814618959708694,0.04255171269514568,
 0.5320667894023406,0.48087454148431097,0.740620609551303,0.5482393392891673,
 0.09271049417806987,0.6129127055939524,0.05906365509021805,
 0.9050869922187829,0.3329931043808576,0.34460143096111917,
 0.03947723931502356,0.7282765693613859,0.2081482375138518,
 0.05227566862658661,0.4113447619698586,0.7256890128621712,0.8571274470494689,
 0.42727021608161375,0.2869814050829341,0.15537829388472757,
 0.8548139729598481,0.9231630580506666,0.11725090365989199,0.5858938680016632,
 0.7027004989148404,0.5807731203513986,0.359434179780754,0.9695327369712893,
 0.22248396113919433,0.1839069599640868,0.7933813429519697,
 0.11648232210414378,0.7703645145727038,0.1696884785845486,0.8945763057201476,
 0.25575956205332573,0.07051764745702549,0.5280878619479384,
 0.23806049792270378,0.414293777274957,0.7619492176676285,
 0.030476402666331892,0.9842443354675074,0.8864980683190212,0.838512148438319,
 0.3202328929022922,0.9723449700892667,0.4291505047611519,0.2459234896099426,
 0.6224738765211275,0.10236605188380421,0.4870979173238369,
 0.12129521773315421,0.7337340147206395,0.06347241222776812,
 0.11931196679264455,0.3933992962695805,0.0939181335260042,0.5994843080283606,
 0.15081703015728898,0.6595403788148134,0.5797206302469358,0.8187021781075632,
 0.37989536513239963,0.643387800722268,0.17003387075845444,0.4367523247400187,
 0.729280832803397,0.5073306821113515,0.759071139182182,0.7913935244119206,
 0.6284305415390845,0.5751145062189363,0.5503754238711788,0.8590652315689542,
 0.2919395983729416,0.8394500896259037,0.5951926989228609,0.9682983604121016,
 0.46906785551584373,0.2740247853124427,0.11025277297862335,
 0.15340218722341548,0.6844380046032212,0.5943418590684494,
 0.16697248175336454,0.12208188461577096,0.6779471569602114,
 0.6875449035269652,0.5608469011307543,0.5375694093409301,0.33778905927465697,
 0.7676821251251403,0.941419275365467,0.9251022005056526,0.8867422963409898,
 0.2389976128717004,0.30982167941577277,0.7083872336782595,
 0.32173204110458375,0.884958824954371,0.8967082087733047,0.40512268710628974,
 0.912152245865228,0.42921769991414016,0.14457434529110047,0.9637896235332982,
 0.037024234703224135,0.7832683444246717,0.5227127403259502,
 0.5156084932159435,0.04805282540213507,0.8073272841616135,
 0.45065400521011345,0.9248742474763091,0.05073345708548649,
 0.7294141716303575,0.9742898449778338,0.20076193230823502,0.7317746809243141,
 0.1702398332315842,0.23413905052664397,0.2555302314593527,0.3957227328290708,
 0.84048996304088,0.9103579302456796,0.27814301041982215,0.1389467299677528,
 0.220582623860078,0.11038635235253147,0.2885993616836522,0.5409930876068108,
 0.8654574878057811,0.8631110752170111,0.4310437179209884,0.7077404976154572,
 0.9396379215580708,0.4097346600800641,0.4723470327918414,
 0.0013321729916390135,0.09353258731668057,0.8894119854930685,
 0.7219097905728229,0.5280735871374991,0.28553050310344097,0.3182311993940541,
 0.27757277428136284,0.706009277897517,0.20559401501085373,0.5098686644160446,
 0.8851212125657538,0.9042097541484249,0.8394633885867635,0.737472057387319,
 0.17650990580124948,0.11938919538278248,0.6117227900114813,
 0.5327597299624985,0.7290056564974089,0.6979624893471614,0.6751008018720694,
 0.908245383554546,0.8998645550795843,0.8754298024599556,0.5341169436281539,
 0.710280597471626,0.4236487037581108,0.7025147482277929,0.510837864920632,
 0.10155260536374167,0.36248203530882006,0.11640276329448107,
 0.7354232069826014,0.35742268298213786,0.08654544472209247,
 0.037938452460670624,0.9148635867794304,0.8665112988881569,0.226174979279257,
 0.3338099917843069,0.3333317871388619,0.6750408982112323,0.4921954900519663,
 0.8141748592153911,0.1642395076414166,0.39676468831066736,0.9288350575124483,
 0.1498812538912312,0.8631708049327779,0.2797675742067214,0.1724105975754997,
 0.31145884635201637,0.8850386268571441,0.10618628921945683,
 0.31390886589439737,0.28309150936489136,0.6707339753550471,
 0.7365469968745562,0.11852509126015531,0.3686329593118891,0.9624901520257326,
 0.27585430436918024,0.3920433757853732,0.5206379986306067,0.4402837631785471,
 0.8665063640087639,0.4950192708408139,0.21333474929613605,0.335948527289475,
 0.5055509911511764,0.09312166703282587,0.23030127811493906,
 0.12041865761713955,0.7462619828199555,0.6243954150187486,0.5237735326243987,
 0.9724313976362872,0.993974519216632,0.013469850186831,0.13337435686898924,
 0.9265950278433679,0.2881606897587624,0.5977164314201769,0.8420373499782564,
 0.8960590062275929,0.2690908727690069,0.3476236371718717,0.13703703146142943,
 0.09244456492477315,0.7755437497419846,0.6205980771865042,
 0.020017798912893348,0.4436868670663504,0.09510142794715404,
 0.8808717957468883,0.5005576391766293,0.9006690486666193,0.13215829573836102,
 0.28463114570192294,0.7022422432164867,0.7456870894166767,0.3437652932000864,
 0.19913857855363248,0.36874548276753694,0.19185416604222738,
 0.9638338884112758,0.6918620499891126,0.45369375819443025,
 0.04838226257270539,0.7928924379611078,0.5148793502958682,
 0.49593555184119154,0.8887094596665982,0.48783895420109147,
 0.8936499949369858,0.42943994566690624,0.14965388431535365,
 0.2397126853117718,0.0046077575231255885,0.8243160452333278,
 0.2533155198055006,0.7382336415906408,0.016702002756664336,0.658725784164711,
 0.6595541104321389,0.6619107863892186,0.08134509757913833,0.8062951729766762,
 0.6453607368500575,0.06601847234390512,0.5749264304952018,0.4390479503644542,
 0.5948248983225629,0.21675823754807577,0.5098864394607792,0.6926375247409869,
 0.34968180200242993,0.24436848582003634,0.9723767822098952,
 0.9836077710398216,0.8507916328129465,0.3991823764005986,0.5036230691524749,
 0.6394988378177939,0.32555592651548804,0.5979809363962824,0.5968102673580473,
 0.08113183382832012,0.3808864831935658,0.9484143337933442,0.626516491548194,
 0.560855539593981,0.0033379119760499343,0.6494527738829847,
 0.07614124447908888,0.773835596910775,0.8006729000463058,0.6931371025032816,
 0.38355762990039044,0.4507273051096221,0.6986865680700194,0.5035644740882629,
 0.3007832448746296,0.688255487473712,0.16512628397990925,0.08821291150082289,
 0.1148147966690446,0.1146713726490235,0.7636532889502452,0.31746389509687567,
 0.6187061837200292,0.16237988337484,0.7209097254076948,0.8782043082448439,
 0.9152143127100155,0.7625617091596322,0.19890642753216037,
 0.32689983227166763,0.6308757751480263,0.5972141591713902,0.833615127212526,
 0.2969224219490124,0.5452084647724971,0.5330034849204196,0.41388482249352354,
 0.7265358963519734,0.7201014596284909,0.8175518456046108,0.34315043599065953,
 0.5287069752491103,0.7246882445769091,0.49806284516381927,
 0.11471667754154546,0.05905928182186926,0.834729779714291,0.1220328027435138,
 0.21925944116860996,0.48456430787863325,0.2842597800207526,
 0.9133284889556476,0.9838269084390817,0.7538656820097009,0.5546025153390665,
 0.0752945203613622,0.677031428050489,0.6945988575824914,0.8031016655408993,
 0.7919884711732453,0.6817184803761991,0.4094304305209022,0.5423911216564159,
 0.6927784865017806,0.15624222224348772,0.27413219581253867,
 0.03454134762065619,0.9445327561069586,0.8821793848445253,0.9099603216574036,
 0.00883524368548605,0.04014174682666938,0.21204104068770901,
 0.909985634125568,0.7606943230676522,0.5088276757814638,0.13816616780616842,
 0.18335959701302262,0.1938091576465879,0.5515897188033089,0.9242939773551979,
 0.34713485433833735,0.594229019496676,0.45765451924410994,
 0.44642974170786676,0.6819206737826473,0.9475867496001265,
 0.022256013025309773,0.2068726992639236,0.7800293030608191,
 0.12930978892005318,0.4811318091509017,0.9976493627378121,
 0.48680355639777706,0.8230886621716061,0.7965046173701712,0.7426197320125008,
 0.3526276118367776,0.7483148311879361,0.9780075089759935,0.9238412769826262,
 0.6455811925827929,0.3867410495115121,0.5412633640870043,0.9084306683150272,
 0.9952535290802128,0.6008630250465012,0.17921587125210503,0.7171947379079197,
 0.9735300424511807,0.9922668333825239,0.8829445271926984,
 0.017945155285681302,0.4465521661385845,0.8250616872279175,
 0.7394506392884819,0.8102641311682697,0.6428808695227978,0.141253022103624,
 0.43943744169132437,0.5611755192865266,0.7261761936933544,
 0.16131600968315218,0.6167160003302192,0.05706434632047985,
 0.1428072261005099,0.4052985720188612,0.36145791240395375,0.6610588276938679,
 0.43380399708710227,0.9380671913841048,0.2995380361558855,0.5299212639652301,
 0.31600661373410566,0.7371584223238455,0.6657210774937042,0.8439508323440804,
 0.7698268435127754,0.4642497483044543,0.365234311847041,0.5358558140774718,
 0.5772083032297651,0.5229465160897281,0.6919749466511232,
 0.001162552062300759,0.9861609587155422,0.3601562859278027,
 0.6719096317951472,0.1384937919372934,0.7025154548391498,0.1326810578876183,
 0.37661626142328086,0.9680074567402301,0.9754606060031071,
 0.07211608905265654,0.8258168639201194,0.10945686972822433,
 0.1565380778993205,0.9644908337272781,0.8063715290671944,0.9518078008670358,
 0.9116367121051847,0.794204194951945,0.4855182904068629,0.8437690808958661,
 0.10545773010757897,0.14555408018308535,0.12368876181815858,
 0.46095521848860843,0.643331867662435,0.41641225240893054,
 0.14899477801074168,0.1514158910940383,0.3215744542015311,0.5360623351211,
 0.39251933325889743,0.14885715915876696,0.33355445616810075,
 0.5174307557405173,0.4733653632802687,0.471074102074447,0.41743177062316794,
 0.34163809176087134,0.965534771528319,0.1354074513269028,0.3144078321900732,
 0.9262076045241417,0.46590175392510835,0.5609042738149315,0.7742261662195942,
 0.6323326863758116,0.9794212130093409,0.402629386108431,0.1148031957165383,
 0.7315875711194452,0.2610427990357632,0.4769034074299815,0.9744121948231605,
 0.18406432198710532,0.588412665344013,0.08828967989761083,0.9171158226983945,
 0.4833918861355493,0.4602192740287544,0.5823714920185236,0.8211216510285194,
 0.14386877687510746,0.11107284127786277,0.385503002550722,0.6994858041527414,
 0.04994426591372614,0.6030712165608787,0.164605900984657,0.984280390681875,
 0.21546680847269317,0.4705314458936062,0.16592599192635604,
 0.33113177527041626,0.41190098244271367,0.05986685815407855,
 0.9403807260916154,0.5423980456864442,0.39587956529982327,
 0.36037451114697994,0.2835423817258268,0.34185374103440314,
 0.6125288801977715,0.22180977027730242,0.7051489083055528,0.6207625234100318,
 0.9537556399528104,0.845992173573434,0.34598456202926475,0.25991912939060535,
 0.425092567757822,0.9792456838525644,0.5660137799388351,0.2167075077396733,
 0.20286024476250808,0.19740261498526712,0.7980566740427695,
 0.8061514569755504,0.18035678893822105,0.36801328357688234,
 0.7492123462703525,0.5935349384096859,0.3247936975452257,0.16138182443346927,
 0.18015386559062774,0.6175049722641507,0.18342501648874043,
 0.7930347977540443,0.9324597812186124,0.528887993240261,0.5117496091935877,
 0.4037213433371185,0.00765618724013184,0.8565928222710865,0.7180694881023277,
 0.3682401198545262,0.4373547514986331,0.2440629803435499,0.4609161202530675,
 0.03795985221157527,0.18265705790728193,0.23519874044332512,
 0.30376231059206693,0.8710152641286996,0.08905898434435666,
 0.16900331056607776,0.20974009074299638,0.7301823237739842,
 0.28670912095724954,0.5076833358025051,0.3932509562365666,
 0.22392578191372148,0.4573355932487262,0.9286481475094502,0.8721613485117229,
 0.34200392381851563,0.1643282299246609,0.8123353135606639,0.6376127909209248,
 0.3475188923840038,0.5491622790983807,0.7254487998470978,0.17977698311343704,
 0.6195876924475519,0.26120476593508934,0.8894409227451981,0.6006816710356885,
 0.1335475501219171,0.5612928748902867,0.19626975062935348,0.3840306978542153,
 0.5362543354331305,0.6552325387122148,0.1294161662621438,0.9313561815847925,
 0.9479111857313377,0.6465058970015654,0.2134059806551193,0.12293582579238638,
 0.14927597689099104,0.1328582432215336,0.5841406052940165,0.8812219216194632,
 0.8820275761283055,0.2554278568506516,0.29626162390102917,0.6465861728260768,
 0.4390273848952789,0.13801693818510752,0.5211338816610551,
 0.14862874117920444,0.44907630995275205,0.0167563215733032,
 0.5357767266605464,0.7417446899561775,0.8054823452552011,0.7057176429084909,
 0.19890170736624213,0.7496422973308795,0.15094291216422095,
 0.6383926889143606,0.44644807478363857,0.5874282444424459,0.3334546084156088,
 0.7361804003857831,0.3369767109375893,0.9082526904541691,0.7032159313486098,
 0.5828496865580759,0.33741419680002727,0.9852840918075011,0.7370700998524821,
 0.6455934819074696,0.8841557675380322,0.11944421384518944,0.9366778946528272,
 0.5908900046783487,0.760032836220485,0.8527217530150315,0.48716570982257845,
 0.7161284326595746,0.6723619979573685,0.657206833007354,0.37925146331399906,
 0.6711964290086851,0.9595661368025368,0.3359849798090726,0.9943070371886109,
 0.2185736992958962,0.18186145348461913,0.4401128932083047,0.1510677167389889,
 0.8925803242354245,0.6766371541715106,0.9032903728955469,0.6765260299569275,
 0.2888466131785683,0.9815416393073562,0.053303133418072024,
 0.7600343925034589,0.934622498311571,0.8290694468713088,0.7712253519942497,
 0.3795469639201514,0.7702358938940151,0.5131193565109016,0.0919047388209503,
 0.8336996966188539,0.7873213501689678,0.3808337035099765,0.627171102678071,
 0.9559786982347214,0.8916971158094485,0.03036986089706184,0.4839416676935354,
 0.010333813613302545,0.5578055131206803,0.740544302467072,0.8730107299068056,
 0.14394513031197054,0.13908765964894343,0.25775061329571247,
 0.344394309463668,0.9486037877604322,0.9342216000236693,0.5039715339010181,
 0.37277567632702524,0.4490322887574488,0.7773055638931328,0.6096870183780893,
 0.3576473429927369,0.506865904797523,0.924825471308703,0.13341585650167376,
 0.7263964488102617,0.7838360815827317,0.7281559457227795,0.36564557831038846,
 0.007766030693379555,0.7251882776090819,0.22715397160165318,
 0.6552996023886856,0.500395919021685,0.3959214677537607,0.597295886532754,
 0.6389991136460231,0.18839906090632952,0.12032101063081446,
 0.27868522516242233,0.07995378783130036,0.665346138588132,
 0.22117462361632523,0.7066888161974395,0.776081193867444,0.17514782346350755,
 0.7473213386896344,0.25240390024517156,0.5982175814198756,0.7161738922968226,
 0.7926624407818701,0.9807752815404255,0.7525958858686451,0.7026296095327378,
 0.4300664908543229,0.7316201760514114,0.10647725720236101,
 0.11869989118042412,0.035109376848689644,0.5436175297596486,
 0.12395274522465782,0.5194160927099167,0.6076729045194506,
 0.14493815106516406,0.8931251318954764,0.7679027865697567,0.2044422466217617,
 0.35631246895706414,0.6706725342776106,0.9548913611459285,0.3678493701546468,
 0.5620180346097001,0.7793982683848781,0.29870588117561925,0.9852643078614096,
 0.23120335083401566,0.8845031110537502,0.10080491912701794,
 0.2552509873889911,0.00537013117884555,0.13388758462552097,
 0.18111916392730887,0.6523380615463762,0.15387109979609892,
 0.7920623176713993,9.90082099835532e-4,0.397855915205486,0.6792182752376429,
 0.7925702601697783,0.21484453667342074,0.9785662850635941,0.6628975734799983,
 0.018490351085110657,0.3231356778504936,0.6414006053732094,
 0.6405040229348591,0.750683093975203,0.7833286282913677,0.22352927777558473,
 0.9314078852485427,0.7859946547059599,0.6483899948698603,0.43896630915134704,
 0.20060828348019588,0.745318766346089,0.7228098114986199,0.49752035957735363,
 0.20486006903006126,0.21031221691757562,0.0455681460427424,
 0.8261678565341958,0.7348058820486982,0.1913144888305427,0.3234055013890308,
 0.9911547588118963,0.6886582805388857,0.15064287651391028,0.9875293790085313,
 0.7038063525816819,0.33086312350783453,0.5069154629992233,
 0.027889595451769056,0.44025928248883006,0.005572017850203226,
 0.5275973283992915,0.10380745068868324,0.12906181690552976,
 0.6323429828370855,0.11037770555705706,0.019638558784048632,
 0.9939703789833554,0.0818619464406144,0.5502055055474183,0.5567020976659852,
 0.8450004250958955,0.21679293078230588,0.4558027580162268,
 0.05516143691029529,0.5324528880408539,0.5945397157073828,
 0.47489785954354347,0.7538301009767255,0.7571997971852117,
 0.01767375500929036,0.6493315839866759,0.22064654344968504,
 0.7670256948517759,0.29159037423438217,0.7893748140158531,0.2316999755794571,
 0.060548352412444606,0.5557323943380496,0.5908259777359498,0.294318524823345,
 0.4445718650024487,0.8651837340145186,0.6131349585161585,0.16153548216647395,
 0.5321390811090997,0.806851055371071,0.3996525664104166,0.8544189889220107,
 0.8165166719510332,0.09339936637328528,0.425783070246326,0.5734256336427095,
 0.7930032775693842,0.13526826415649318,0.09442758335298795,
 0.4213417213138362,0.9388466597286802,0.9175988148886961,0.874296173070416,
 0.67470969087881,0.7591941060900094,0.5750402258533887,0.16081393417226286,
 0.8683216069981589,0.22847066832075968,0.02160761575509429,
 0.3907062180764037,0.5681953089652207,0.22850145860051407,0.4928982352996488,
 0.46082807935253534,0.8523977419331781,0.381731302623911,0.7186924731065989,
 0.5343630494864269,0.2868991822942544,0.7622269360346273,0.37195814897086077,
 0.928131195030124,0.6508567209828585,0.8327446870294299,0.7790216313705913,
 0.5955580231652742,0.06461916943079282,0.8886839798081426,0.7891376580711346,
 0.5934359481766971,0.6528003634359677,0.10081898878902174,
 0.19996456840236587,0.9441809065968575,0.898478654645576,0.9944760637214153,
 0.9516726495436065,0.40526261231347593,0.3932601148246786,0.7736566944004974,
 0.699775035300261,0.44415874739726957,0.9249118352501716,0.09965029597836617,
 0.16344268700701514,0.6339446482949599,0.9002223025506001,0.7447670682795823,
 0.7706105986727056,0.25172382405345095,0.1995202759378012,0.530790818703407,
 0.09842697415296287,0.9058450950203173,0.8401124455183404,
 0.018846916531160995,0.7554707209891164,0.4595578584798039,
 0.8394524026816166,0.8020011210819777,0.6098795003695675,0.4760860912388719,
 0.5408415604953728,0.2397853116791897,0.9448768216002998,0.4188129904650171,
 0.7916042441914211,0.935874322466905,0.633863494281542,0.9058965170612553,
 0.5705377886274788,0.01955025124747667,0.6795076835500681,0.3474541755036664,
 0.8479289213181717,0.4534755662891953,0.6064278915185204,0.2174887055079573,
 0.4349262719820828,0.8213805137505924,0.06584647889133499,
 0.33618672766552016,0.20838066943658606,0.5980797553096658,
 0.9151810560184289,0.9310930188065987,0.8771586029151588,
 3.482036322740578e-4,0.29676962973035215,0.8775647931570315,
 0.0858554275921164,0.6102055462020725,0.4431126743202576,0.8132601215902718,
 0.5025738071871432,0.2834955409684565,0.7539090624951006,0.6361951224283494,
 0.9607182158519643,0.23534762336084258,0.1501935046615538,0.7628383105842217,
 0.22336280581132573,0.7734190912443777,0.5429362870087899,0.8979103800324773,
 0.9743381274398816,0.05326175352134377,0.8326105826269166,0.537849426425529,
 0.43576141477486097,0.826358782762847,0.8160809192908356,0.34147568197580314,
 0.010185766508877814,0.35761533658215516,0.11525416356482032,
 0.5980562990519724,0.18968923352023448,0.8309328657914723,
 0.04031750033416337,0.3787823074901133,0.9091692206319931,0.8023199687331195,
 0.08504597327204533,0.4748025497576296,0.40262973902022847,
 0.8769619199506983,0.06506095934259493,0.6637881786372529,0.9637364508439527,
 0.23119782159963265,0.8519551481907071,0.8925444368098111,0.5644372565091011,
 0.411559874747628,0.5908527268805254,0.5148178811583527,0.4262572127207763,
 0.8921017701053277,0.27631946009138586,0.5848946438786077,
 0.020830143028855108,0.7253730043687477,0.9493036174071049,
 0.6696615919056103,0.8480196908414112,0.2005160238378657,0.9308178178176472,
 0.8122924620149141,0.2519069766390266,0.37287951475953374,0.3649838198870984,
 0.4384462310373707,0.6517622568699177,0.0375472407423969,0.4663721192694401,
 0.8439352446227903,0.9244094965971225,0.3612721102701747,0.5442651399931105,
 0.8304667305813339,0.5678407972294669,0.7842326967464366,
 2.551852809853372e-4,0.5667637242059737,0.2628328107511013,
 0.20316034549110962,0.18571295354706252,0.5728775284186974,
 0.10387078006442696,0.8203448146062361,0.54573861962944,0.8139939341846674,
 0.9057337417235025,0.6790946560727517,0.4694617886278559,0.7114030622454798,
 0.8636551844320154,0.2976576688844028,0.795972101156005,0.5904156945188581,
 0.23410608898713703,0.06321029227900188,0.701833655395605,
 0.00678699376540326,0.5986070626839944,0.1710026987041634,0.91067835228252,
 0.32456640152107963,0.5559178962276271,0.19887643669770794,
 0.02022560439593335,0.9797488539013083,0.5671962361712588,
 0.009977238260942545,0.9302177222192881,0.40933485253328117,
 0.406630516446976,0.1946144269561032,0.5171828569456371,0.6922665717972178,
 0.10986486148092922,0.3522544359322216,0.8119310987718759,0.7455372775967631,
 0.6233372286074824,0.11511000076186484,0.4294025363098056,0.3267377739363516,
 0.38139175985596707,0.2791178981994582,0.2897797288828594,0.9963823361626695,
 0.9538527820992173,0.2814014118872328,0.5789052398397502,0.24631055480049313,
 0.38498021685936257,0.748188068959181,0.6695311728540442,0.08947690975432987,
 0.8280174042052809,0.22381394581191916,0.5853321488106339,
 0.08566171159306202,0.44039081885078346,0.5519297336117936,
 0.4542890829143572,0.09269600827536528,0.1197925027877963,0.2195074139098343,
 0.4773943804493094,0.02623338449781243,0.9155543449541097,
 0.05565224597708052,0.5732859830245208,0.40709344473456044,
 0.6538749993862264,0.42197904222613747,0.9110857132763708,
 0.15365984211306027,0.7706675296488414,0.4389698480361245,0.8970005730642843,
 0.05266879895733534,0.48356421163987273,0.5953924265936292,0.287290979800941,
 0.6976934995052338,0.09961008671141969,0.617235959042442,0.2825470592754583,
 0.096494394989042,0.5246267785807323,0.5999498960847875,0.40830395578573986,
 0.5316725555114237,0.868743492685455,0.4068292285358712,0.31307295950224134,
 0.9403827174224587,0.3568984556725129,0.9690010509226493,0.1311430356591322,
 0.6649678183897642,0.6077339373524582,0.10382429848298469,0.4600689540471645,
 0.5885313766532589,0.0780387973077763,0.5806401177915799,0.5297919889562382,
 0.07807525961943185,0.45122247397460036,0.39407581053329455,
 0.7943474307447326,0.7088434381728088,0.3618395633419682,0.582181015164728,
 0.3001998096370919,0.39474869469889695,0.8436797274818844,
 0.19421926200557704,0.7804206600196462,0.43287567595331833,
 0.3735006473637469,0.07483550363367797,0.5824872673178006,
 0.24885528871459672,0.9535233629111607,0.7573470459014163,0.2010014987988732,
 0.7546452287348662,0.9827601257501186,0.446900699318064,0.774700806315073,
 0.6060489161374437,0.49206951248379804,0.8780753962683296,0.9541648795158133,
 0.3474970790919967,0.06190737603851437,0.09226798280727522,
 0.2828725015464719,0.23781969993031116,0.5831426632240683,0.3917177370959867,
 0.7884563773412626,0.9722194507275502,0.8394384773661882,0.9902559863449948,
 0.11721716243172076,0.715317877574937,0.015205396976634478,
 0.8548918144111934,0.8439208497002918,0.727234740181623,0.6208459396702621,
 0.7335040623374818,0.46685190165386337,0.635537412235266,0.9332734388111329,
 0.2900091612179374,0.3787840250587703,0.6353848127283194,0.604101969405601,
 0.1989723491396913,0.728681829811498,0.3840225597280499,0.6786021160500759,
 0.7253792961316049,0.03345041247957359,0.006314742684928709,
 0.6372495830033721,0.7945181433919982,0.09431435425229917,0.7149531990007539,
 0.8217063260792216,0.45374059797216293,0.35982119224901454,
 0.6054371642767332,0.6662364947048836,0.753933714798531,0.4435705359324319,
 0.4918481400115704,0.5326344239930484,0.7294150536695645,0.5314277623601272,
 0.41066246556237906,0.7514095770914491,0.01868551475006819,
 0.10286515379037298,0.2938821889623906,0.34542106304077036,
 0.008283373472724787,0.005989026132842135,0.9575597981827629,
 0.948277523324654,0.6802638049334472,0.5055660810344187,0.8158703056883201,
 0.3236794930439295,0.8969236187974918,0.5554011991698667,0.11818253685048041,
 0.16862678497327888,0.5942646828718655,0.08454759075492724,
 0.8770144484259325,0.8930745155485612,0.7172463467288168,0.513945430131933,
 0.41328778675755073,0.09468616699904586,0.9990921425064536,
 0.8892209868983599,0.19745658657112697,0.5051987151374969,0.6766207337384318,
 0.10956172747150816,0.5993984940883726,0.12135736642839479,
 0.5182928954837153,0.7931369714849131,0.189247828357323,0.8739107962475482,
 0.17081477653910193,0.3911329378623124,0.7649281799820462,0.7850922306746854,
 0.7566686827709916,0.010326966554495054,0.05925139089796949,
 0.17762075076713857,0.8539617765441693,0.15930053941198508,
 0.7943735383613082,0.35300285093041184,0.37013221807083185,
 0.48998004651802185,0.48208643909400317,0.024915220296600094,
 0.40582625320177745,0.08058454220802569,0.8383897573557977,
 0.2957007897485294,0.4977739880420393,0.6977650832094081,0.9461666063371443,
 0.687872046449595,0.7923984076734352,0.7150180239286898,0.7629729117518279,
 0.5448055621678216,0.021675468846880985,0.14287825551325473,
 0.27716572284451235,0.07211266255432303,0.8476203843932196,0.97858823230631,
 0.4473810754309452,0.8188023834464904,0.6027388873180193,0.03627634319577533,
 0.9585657067017862,0.529375262124092,0.9651913276259717,0.8480599322485941,
 0.8870726833434063,0.11745310001655629,0.14990192981402783,
 0.6077236915563639,0.3471862811461871,0.42270093482343685,
 0.12424829433454299,0.5947240131551192,0.8426528405268119,0.585930967082703,
 0.7316516802391569,0.23737667010688368,0.5446794406552546,0.5155199466573401,
 0.2989353647027173,0.41462687364923323,0.6969713715401094,0.7771118675317279,
 0.014740084068371462,0.16268726005678835,0.1663915172004532,
 0.8699650283946709,0.7355110903969619,0.46377541802110933,
 0.40043250776659867,0.6868105000457418,0.4220452856386865,
 0.24661845924123993,0.8932299646087096,0.31006052318833366,0.256593435609211,
 0.005952786831430634,0.218087007734174,0.6534517247128817,0.4340422012919689,
 0.9066541214592954,0.4568070548747335,0.8700676337390753,0.9096626753370295,
 0.4976301529475823,0.25876291772345583,0.3743349243280176,0.7166902540371369,
 0.9215596035126625,0.436096522073654,0.4011478893577526,0.8232947478269131,
 0.09188419542133708,0.12671262578217168,0.6851558659083418,
 0.9002340740521679,0.4522748553631085,0.07214675352860189,0.098828841071955,
 0.857099415159423,0.8321001983034366,0.46343403712440634,
 0.058929894486298706,0.3780108644721005,0.4854125391659031,
 0.18678549198625827,0.5121907728940558,0.09113425959377897,
 0.8147973410351295,0.8304694765204781,0.7722570372374084,0.1839098700799212,
 0.3416127185773524,0.03250844232019412,0.9183918123495999,0.8428973256678003,
 0.9728818216534642,0.13333558582468252,0.11710951779024492,
 0.09225975259780794,0.09129971628860434,0.5658768579328096,
 0.6985020169713994,0.05587871342112438,0.865023881082286,0.5286620037191586,
 0.24884852645327482,0.4783434727487659,0.45996367367937285,
 0.8781258499081291,0.2835916038554075,0.13265177123030492,0.6058217100952408,
 0.23792480696672746,0.7907896772568535,0.8668576563710036,0.6025983688535403,
 0.32993651604774776,0.04330913995881347,0.963307637170072,0.949335262042225,
 0.26766454235858683,0.9764115353574178,0.04641211618402186,
 0.3779738341772667,0.6368609736930589,0.8064789933722154,0.5824874220524064,
 0.36908219667939146,0.49966707271734845,0.8397140967520014,
 0.5009605389281717,0.8935263502532828,0.551909974715454,0.14018736973890167,
 0.4016835591775372,0.21057413560801985,0.2570111406929282,0.7286948402974227,
 0.4208830463975032,0.3974074678023418,0.8804093344621482,0.0981723993749053,
 0.3534120445806671,0.6536761912886122,0.29375904055409896,0.4353044565113209,
 0.6139144255616764,0.5480089681949178,0.7634213949648206,0.23832590038455448,
 0.4254762916377922,0.09291216138613989,0.15858406760096788,
 0.7190544082055302,0.8968398743138224,0.2658396532950478,0.08205023960865032,
 0.17020998772937324,0.1773215780812053,0.26435143419956353,0.679875269638172,
 0.8785145366604717,0.9583849237435896,0.026740623571224642,
 0.7241318189260069,0.25650151903729856,0.9325775545749808,0.4803010132817448,
 0.36093958634683787,0.10780013057545945,0.02346960021390543,
 0.664189606843097,0.949250572666773,0.18844478099730377,0.2862062223627365,
 0.6847746242497261,0.8284839913548874,0.8871930654913898,0.06241640944270854,
 0.6285814514326703,0.21217328860343299,0.011013814849823444,
 0.6008958097403251,0.7609782679370912,0.2389346549411966,0.5929403533821617,
 0.9700235312111838,0.8340268396227711,0.9389511258092275,0.8358333705383278,
 0.9692396519014115,0.46447802940593186,0.011347876339378504,
 0.5456208953110209,0.5064556881947603,0.6933583929142665,0.3544914343865012,
 0.30509393276545604,0.8686202979744992,0.050589979523806394,
 0.9076921528129253,0.6411688456345375,0.5374833639477944,0.403036200841962,
 0.60068984345708,0.6238234004925673,0.27227964446897146,0.4665417723462274,
 0.7017735815855262,0.492005969951526,0.10871859540591255,0.9070340358333787,
 0.003609090726475306,0.6870819392071508,0.06752636997262196,
 0.7192271722933313,0.7754156090115131,0.3377325938060547,0.7114385057601581,
 0.5514961343751392,0.7102557193266124,0.7297008769580574,0.4971708804045041,
 0.3399780598626023,0.4121465428429044,0.30024486910570547,0.9162719862896114,
 0.2742258992284593,0.1671229928305762,0.9544169483285435,0.8584947186983024,
 0.8180759189374465,0.8982170081623437,0.46483534189433384,0.8517987614146265,
 0.8804377317858091,0.7466990212626097,0.6263451304847,0.30023698928059916,
 0.8781214307747751,0.8353349303652694,0.2522257988101073,0.6727013505349545,
 0.7231576011692106,0.4330206892231795,0.07431094331492583,0.3774228670507993,
 0.1387388633097011,0.43934327892471325,0.7385135107736894,
 0.39253763473443926,0.23426457140083412,0.2858552681670319,
 0.2956000146410611,0.0045253505833601215,0.3449675403447825,
 0.1353179721338954,0.8867236644740423,0.9527562973351277,0.3107748278713619,
 0.9672315543456186,0.9833242842350531,0.6090608314984249,0.7355387225202494,
 0.29245445654686,0.143546793972109,0.9403690618452579,0.10618931946314292,
 0.919886302525204,0.5844093130640273,0.5795751217970673,0.8568403370987686,
 0.8695532389787699,0.2771073971837854,0.4894120371898051,0.6520694373964562,
 0.6576927214667985,0.8234310528911968,0.3184029466367423,0.7248524962044668,
 0.18087279213962904,0.7664587899370178,0.7175385076287073,0.35146342248289,
 0.22757812020098012,0.3385510004283927,0.16920006048177372,
 0.6367194326966402,0.034914577184128714,0.5329014911084268,
 0.007102566675368349,0.6163572904345332,0.5605575642368044,
 0.6749210400495047,0.4573929800139914,0.43772610557430713,0.363806505445879,
 0.8407226810585313,0.6582769644094701,0.2946225875103722,
 0.046761460853545156,0.7955667934419183,0.1865584732007427,
 0.4466987951161109,0.11896017317545604,0.8098289448854677,
 0.46115163470396925,0.9647026699047432,0.7504621873143978,0.5144563026375396,
 0.25576655767724243,0.0899704893086064,0.8888296412524062,0.7222734748639602,
 0.2772861821851905,0.31778323653827856,0.5462808983746483,
 0.12172872632069176,0.47853066466810235,0.19655090650659224,
 0.22160609109787976,0.646355948508173,0.8918198201345805,0.47501462779728465,
 0.463977946285128,0.7037256684460181,0.7545987182388856,0.6343005262180244,
 0.5044583521599441,0.185708842029925,0.9033546808691137,0.6598098334414073,
 0.6581804823574047,0.38148060033019304,0.1645083660509763,0.7154174406848715,
 0.449990609602429,0.1630660605951172,0.7067086246802793,0.35006199277562444,
 0.30666000039755836,0.1665786172420397,0.06146194611832523,
 0.5622061754881996,0.46959016406403253,0.5311799441516986,
 0.06596438034653251,0.07076518691770672,0.9541402691612398,
 0.8787437302930954,0.1514220828458137,0.9998616941559448,0.7299075151680757,
 0.5570393368471109,0.7186765113791934,0.754736973443435,0.4819596301989427,
 0.16289380081860017,0.9667823913535151,0.5479530091826728,0.6558128494646225,
 0.08542514141903979,0.5100882138598288,0.2641466411009483,
 0.04369686150258412,0.09353641483455577,0.5900289433278718,
 0.2331782742352313,0.7752965727403842,0.8620523244881398,0.9643218483212955,
 0.6545825387152903,0.2703249630901634,0.2573130463984593,0.3007246248673504,
 0.9183510561914581,0.8203901263131821,0.35084499005754166,0.5475473384812423,
 0.8698604283186585,0.7822220571586819,0.88881434291292,0.3755876107934618,
 0.16328473154355783,0.3289606252848005,0.5920023889673487,0.5488373439101272,
 0.08861367625250394,0.15461888894682385,0.4848498120394139,
 0.4595227433850647,0.9074879625806577,0.20960792340816825,0.3400662988649438,
 0.945888438884567,0.25005322330710944,0.35967560770777773,0.4933415871273067,
 0.5804179258212804,0.4052406254647756,0.5882008796388232,0.2951880172415895,
 0.19085385131575217,0.39487247716017204,0.8171816079369932,
 0.3135492671276585,0.2514602794956726,0.5324263958049109,0.26050371596474675,
 0.3597062752028305,0.35958868928242305,0.7858290333249627,
 0.46796135392532356,0.4163104831325797,0.11729581390636368,
 0.7334509178254746,0.7615792083273255,0.5434061639883777,0.18901152211444328,
 0.8313688185377057,0.10345698480198418,0.5519899198596078,0.9661802051890594,
 0.9105149413007663,0.5665413050861777,0.1929409011295966,0.6234864971224733,
 0.967232549791327,0.7006364105192755,0.51493800231175,0.9116978021915129,
 0.9304192985228358,0.5137097669904765,0.8539168745043988,0.8545524326941253,
 0.9049391898735766,0.10634793484246818,0.02784338393693786,
 0.35291398686381537,0.6119753909417494,0.477367684907817,0.9194482568300264,
 0.5460665088517959,0.49882608211011403,0.14814403451743352,
 0.7814694484835478,0.9441002231116968,0.32958551851340034,0.5452331647385663,
 0.47400032056152475,0.12403245098872273,0.26146763834691344,
 0.5403390763086731,0.6005139026580077,0.5886950266094887,0.6980971353805491,
 0.3344651406332949,0.7312842998014356,0.09560940719143185,
 0.21086262902336228,0.5965942141544652,0.7758129295567388,0.4604863494393303,
 0.017231856466125794,0.20212179869265479,0.48085958232179304,
 0.6965307862499202,0.7158747575259764,0.5350359253550677,0.19518471387596303,
 0.9659654638936958,0.18062734316459494,0.6005314122974731,0.8379072603679995,
 0.19314235935855617,0.6334368018298602,0.6081365270159028,0.5617511031611669,
 0.003837758114052736,0.22394682626279305,0.3954539568296187,
 0.6281396529136736,0.2746181642368133,0.07034039501478162,0.6378484199185354,
 0.8501337719415232,0.8730126684361847,0.37603674838971135,
 0.47376893406445164,0.7105056174900103,0.7723326374326402,0.0316219538342184,
 0.6206504641077369,0.8527162601185287,0.2934903420638033,0.9112766163490686,
 0.6731900577527064,0.9463721127308292,0.5381269241051396,0.2590475303820067,
 0.36925517103333616,0.5209883244579315,0.28090093346016254,
 0.2956264459253799,0.7048634117568042,0.42028882597444284,0.7447155742785204,
 0.007066990565008036,0.9008582501210829,0.7858102513647337,
 0.6681462071847902,0.1352562572936744,0.8003091554198684,0.9428105627306277,
 0.566282261477724,0.3208585745970266,0.8770134485975625,0.682529004100548,
 0.30077709492459637,0.7723770635179518,0.03461843720649882,
 0.3558938799195134,0.5822004566825696,0.9837088005044415,0.9030099066667796,
 0.8613252228646384,0.4914134591670566,0.7751128358763466,0.3846145250315611,
 0.6767323273731438,0.46471315971641947,0.6007653885306157,0.4097457774861335,
 0.7645793529178097,0.9221324391850223,0.5963965405788562,0.08322660642037827,
 0.8789773802304638,0.753236653941557,0.3165844884836202,0.2156800880869406,
 0.3043543061812619,0.6862059309758832,0.2919951094439095,0.42293790589038327,
 0.2906592536624534,0.3738889106130481,0.1977834120615346,0.11276369767854877,
 0.5733339998343576,0.3769031156402515,0.025033999138783236,
 0.43190246868585114,0.23966839879957647,0.8669283135516154,
 0.7570414762579645,0.6070015325474081,0.2400584114143255,0.4357028357084658,
 0.16856065065209336,0.6979411574326452,0.3901326230993736,
 0.12225425578932902,0.990303889090223,0.11115409961126133,
 0.15796719846659235,0.7821469679677923,0.17416348398256198,
 0.18222504945576057,0.7364097894843903,0.46581464337383816,
 0.7933172416134835,0.420752672657033,0.7484083173014338,0.5792375985742192,
 0.18442059754566276,0.008759557614520186,0.04411242420277628,
 0.9346083442158398,0.4774518316664179,0.301618223305412,0.39039588006310677,
 0.98462394286513,0.9989208592413207,0.9677740831046063,0.34654474509195865,
 0.763777196964463,0.893907562533565,0.1657585745802208,0.5161507204179672,
 0.7782126894667333,0.6058484687159202,0.2922453044616301,0.17466466966628513,
 0.4665777465604316,0.6403164107571935,0.08365848290293398,0.9202653932474476,
 0.6382096495717428,0.3789380782488432,0.2753329106331639,0.9559251508332464,
 0.6123752326870672,0.8531078304519957,0.7231396312053358,0.4203250705961816,
 0.15770040051303913,0.9689549731734499,0.7477870883586514,
 0.03638308466375284,0.0322622709327518,0.9569806560908289,0.5554951453783173,
 0.9998613234027744,0.40341463848487424,0.9971714694998062,0.7038097358054165,
 0.8748743910087191,0.5319836945962775,0.45546967547085737,0.898913137638029,
 0.6501261960209161,0.6038738716088723,0.24728532008232518,0.5655772811643174,
 0.6647309711713514,0.9064798758677468,0.42673001600441884,0.0829036360950991,
 0.7446251080873163,0.34097710558098526,0.799038078992194,0.21297247104550643,
 0.49011430124965827,0.024866561330383896,0.1078718415181259,
 0.06498405749113767,0.5796584079043944,0.5849131267121601,0.5713878737226625,
 0.46770601217425367,0.8256637576003527,0.895027268209363,0.44209051089635354,
 0.06370243202709924,0.8310316354462859,0.14605355692729327,
 0.0014545500890886798,0.1440522921233761,0.580105171053152,
 0.2617036240385049,0.5281709639216073,0.27250402224466086,0.6804862403449835,
 0.568850385864553,0.7985548243081702,0.5102729235402381,0.7651746511571921,
 0.7754851909256958,0.2892070694516512,0.047803858468367766,
 0.038884011067224944,0.4466069513346338,0.6239131064922019,
 0.2735526610109448,0.6807215701147292,0.45204619136364343,0.437219452123046,
 0.5851660336569671,0.8280996106829261,0.3563703698252102,0.5581588112464879,
 0.1277448529868228,0.1420156477309944,0.8619838683929447,0.13055912380918888,
 0.7304170823235415,0.038570105245646635,0.019028248322437213,
 0.050956379349354375,0.4216651014963867,0.3959827864920572,
 0.26034955530759285,0.8856022572724424,0.5318663419644888,0.9671766820456756,
 0.5249006507423285,0.6517472740982843,0.24940784622508272,
 0.14593326945498908,0.9736135482524069,0.24970918134281062,
 0.5688656546103901,0.3942691733341256,0.8633081522531038,0.39614452943864653,
 0.7574372183242034,0.5938367218688062,0.5694602250101607,0.46437138392442523,
 0.4150727123936918,0.6915140312562978,0.12931515082566902,0.5993420491640915,
 0.8448052673749848,0.8778229965001041,0.8593487091067559,0.01702393066078023,
 0.1822123770832661,0.3005136716091453,0.1716697493739301,0.559985816019049,
 0.3244005303248435,0.4189200893376811,0.8343322367471481,0.5030294099719824,
 0.6149644748175982,0.4576581039365034,0.5297460301230634,0.7640870075416227,
 0.2094258418231636,0.13951672426753037,0.023769121691551987,
 0.12312956351362381,0.4786754884931156,0.3649723916574529,0.8582495170553955,
 0.42671722680074065,0.4198736701717084,0.28744079540306633,
 0.022252645127179504,0.5518424912414973,0.36150102545907514,
 0.14318863616348243,0.8959184969221677,0.8887674293172351,
 0.19233384829101174,0.8216120763866657,0.6424222094184057,
 0.03641194783683843,0.22081846000153704,0.4859263421949943,
 0.8403836248325941,0.22081273115377142,0.7311192749063231,0.8060763500931617,
 0.03733208868433413,0.7433691819299842,0.7680150959982894,0.519456793075347,
 0.41818027856441453,0.7537902344370346,0.7428763832426315,
 0.28138490779836645,0.2126941740563434,0.04959396173540287,
 0.8517452785295984,0.7307578640959832,0.6018174652864627,0.48692398204765075,
 0.8094322192616938,0.9951089886764822,0.5431858795833719,0.9995731748807072,
 0.8884671664448414,0.3302474167382934,0.940997673574677,0.09063789834607372,
 0.13214725946798433,0.6157196462870105,0.314421401165372,0.48084537414877304,
 0.09695465012856885,0.07541753746199598,0.11469689990005694,
 0.9183267378373987,0.4002968688531916,0.7040606985586397,0.07783565118783065,
 0.2852146442344531,0.4193948415732113,0.4461978492348083,0.9859158494140448,
 0.3849836709146093,0.010543965251249454,0.6507329529698024,
 0.08532707784635363,0.8644208716221837,0.5746995668939641,0.3847863257239843,
 0.2605951626028038,0.7036712799608477,0.1981006439006454,0.29933480779822463,
 0.8809728004586577,0.7029362318132012,0.30333200707374475,0.2589407425558987,
 0.3614449720593985,0.31682979540211764,0.16056467496065752,
 0.9519855619557671,0.9462245382268617,0.09382387344207876,0.6903714260208098,
 0.5728835466071424,0.2565763081324036,0.10997257359339907,0.7408351864059025,
 0.7023422731070421,0.6433273654682566,0.12501326273421087,
 0.10640384245604584,0.43111608182674177,0.8046172135231623,0.805348130653697,
 0.4268904122583559,0.5435538479178678,0.5071044287764084,0.991595687161163,
 0.47539036132029855,0.21547242158256075,0.9184049148919167,
 0.3582075727197451,0.1999690839560022,0.5287237804631427,
 0.0030009502054797466,0.2290052497777999,0.9641285058049825,
 0.37844737820852514,0.5313594724093379,0.8606216114504341,0.6558665407642,
 0.08206646148835228,0.9406191155361218,0.21929458503414612,
 0.05688379790070375,0.046009333139629405,0.061454318482183634,
 0.618604546784847,0.04640898077875233,0.5968350777296664,0.31401436176880604,
 0.021659302859271867,0.23696958731787543,0.5007449454307452,
 0.7588423099402695,0.38193340300129885,0.8538907138921289,
 0.36654533019999813,0.8971824981121557,0.05567846502196594,
 0.24913306441059935,0.9419589616060429,0.038956219523030966,
 0.34340463848520153,0.1543425330805458,0.4690775604562184,0.7818100363207489,
 0.5077997979593348,0.5055800146761775,0.38607869844754217,0.3941838456345289,
 0.4441102321194196,0.5873071929729103,0.02329446954424208,0.7138269495161123,
 0.35809657379576054,0.5837584977644339,0.8801280883264216,0.7550391964380052,
 0.8017678237554836,0.017293156652681385,0.8308435814379678,
 0.7977656922584315,0.06211348748538925,0.4154170903695029,0.4589639742771986,
 0.8533095154688534,0.5749849138369657,0.5333010394492114,0.14792461747507946,
 0.8125861990707977,0.5180741581737797,0.18343696260546172,0.8616650622521049,
 0.7666232111297479,0.6956777636454781,0.7168260623031133,0.8091518625061129,
 0.9662008120501864,0.04014473926651163,0.2774332613914601,
 0.12406866151778784,0.9091440405000877,0.4832971678155552,0.1350695265349655,
 0.4805708167587922,0.17565988737830707,0.7584395915533769,0.2901440594614966,
 0.39688656360761565,0.7258846018003641,0.153430266763801,0.25227807856859163,
 0.2236092123394684,0.09132166780400075,0.6543490375457269,
 0.09145992562923577,0.0036407020874184015,0.7442286716372114,
 0.976784329449705,0.6538582010622775,0.7899626548327278,0.8917962970303234,
 0.587412471843782,0.21271287211898537,0.28471369985371786,
 0.49963993193605105,0.7339787736890947,0.16444518288786225,
 0.5120136028866753,0.27964937475514495,0.3615395822762979,
 0.43996473032814953,0.3313572587647593,0.03422678323415418,0.722484432487375,
 0.3216106092463592,0.4658060546903323,0.39655175390888286,0.5341902715441642,
 0.7433399049738532,0.3740226855764317,0.7308086438576136,0.32727837751831823,
 0.9900611977533321,0.6291617423468489,0.7966777028427261,0.5644104889581018,
 0.9185534973819472,0.9972856662381205,0.15204357512200772,
 0.09161098084632646,0.015719289320227592,0.012077980362968366,
 0.3858145028325497,0.3404422764834135,0.2407247425336294,0.06760633346210487,
 0.13433367351890535,0.70076340390787,0.6768770387549579,0.7044627504867409,
 0.17702712332422454,0.20844955024377365,0.46355280196785664,
 0.4490312963327763,0.2739504452925331,0.3708523250581722,0.5883836349436105,
 0.12931727857656272,0.7402107929732558,0.8956576068857707,
 0.05560362979218414,0.6402077984964526,0.5187537682358394,
 0.22101912008650493,0.9589929729986792,0.6057616656543766,
 0.18598228822958096,0.5465334844370394,0.9444602369458194,
 0.22808891169381917,0.009844453494917582,0.9343126244815454,
 0.9936375037087242,0.09080463644452597,0.10759321786644038,
 0.21070649405729924,0.38282148223132195,0.09541897675562405,
 0.43013008431852273,0.6219721000206332,0.13867933526146514,
 0.5460567099920299,0.3887682361275562,0.9341287419231106,0.6756660584112434,
 0.5046801089087017,0.29548020550870624,0.7659006940782773,
 0.28601693985356524,0.00413842085574756,0.4237352012808885,
 0.29699411545876186,0.2724200478765677,0.25135037798149606,
 0.04447754458931508,0.7468516235952297,0.14615951614106393,
 0.026451897473470987,0.25430247990228794,0.6204166384886955,
 0.4072903531381249,0.4216632929194063,0.3440731352934989,0.25640152765899404,
 0.500585738840163,0.3134130842203643,0.5781421411146169,0.9593835352847808,
 0.9122712017985448,0.9507760167925781,0.9410892650227245,0.47166893432416546,
 0.20516916108387462,0.25809234078717413,0.20897324149033614,
 0.5267607196881934,0.7062588566182182,0.9471285102587055,0.514401815334785,
 0.8054483220267188,0.8669353682105574,0.8757774007650008,0.09316683937703185,
 0.5731647270477205,0.6926968416011459,0.9322044381954988,0.3537629247867279,
 0.6870273541954885,0.21597171495559986,0.5456836412516926,
 0.0016070057320589637,0.037610653794905025,0.6837046421336218,
 0.48174955976842027,0.8462390196973262,0.11785255356467805,
 0.15911862527107778,0.009039312410471623,0.18011180209654187,
 0.08774054905825446,0.26569018008488954,0.8454483873695695,
 0.23738006896154218,0.7637520813745596,0.45072340589388116,0.633354937990189,
 0.6464731857835465,0.5861467631799505,0.9398878940474933,0.4813442607888688,
 0.6888259079270398,0.7280345640780351,0.8555533280021609,0.9547835373284663,
 0.07335220451080204,0.18759410117778685,0.8931909723403119,
 0.5473721592785767,0.9035599805492189,0.7488785051847275,0.4257743073884708,
 0.865956066236888,0.7717379918987705,0.4694526863099452,0.47833539725230734,
 0.1868141221080979,0.7930250102254915,0.05188123658263011,0.5840931391829972,
 0.6644560173857931,0.4398854438893244,0.904566661758819,0.8581212492171427,
 0.788521632318556,0.4282930446333544,0.49427469062163065,0.47557842015228946,
 0.002984712879025575,0.26526133691523845,0.8667748258634089,
 0.7191449650879693,0.3960912107277649,0.8573168589248448,0.3304099640231588,
 0.6193640371972572,0.7066297691229135,0.15618153718477557,0.8295438838381897,
 0.4655850343222253,0.24539401035694586,0.5075596732455921,0.8419838277082109,
 0.15150854946271308,0.9015045770569867,0.4809377785229,0.8239934421219565,
 0.8679901031903757,0.7674896481649714,0.3375335815635525,
 0.034897685796321354,0.6687246925822568,0.7152789572697593,
 0.2173753316255942,0.6067298043698702,0.26534163223834484,0.5907279196118352,
 0.9071088160843637,0.8664932662364162,0.8056292820909792,0.14239218010298282,
 0.9184633224980753,0.1539696378972124,0.6902249171244589,0.2685996881614897,
 0.7363277827284498,0.08850224630747139,0.6530124030134592,0.6167186729774456,
 0.33254628151093624,0.43212784558576944,0.6152416358146219,
 0.06553352088315112,0.3615497833343764,0.9116960023398586,0.4353004744335246,
 0.8942990054517598,0.7380597765979997,0.4477439207775066,0.7741647988902908,
 0.33178480816110145,0.6234139674738788,0.3567226044403454,
 0.17628436720491203,0.35426371482082797,0.6425049315483359,
 0.30718525517937967,0.2630234245631471,0.08958622362129987,
 0.8967674139802204,0.28506940441729045,0.5210593287724515,0.7933883825923127,
 0.6827862513395747,0.08667292128752502,0.2500726723740291,0.8240749463812433,
 0.5640912027020234,0.32269481368612096,0.6767277004698541,
 0.03038022408829133,0.821873359122995,0.09816495201964237,0.6548488803724498,
 0.5541922632663396,0.6262607593295707,0.9757742336418962,0.13727919889745743,
 0.35606790605223493,0.00772280409684567,0.5578244857892497,0.250824190540207,
 0.9769937547115943,0.31960276865606374,0.008872765230702462,
 0.5298074717676909,0.37380173124329674,0.39992449912286354,
 0.02990401804146836,0.6581609357605052,0.2763421052967646,
 0.22781555176284252,0.5371150026921526,0.2967269710876229,0.7946849688217863,
 0.06240044098108599,0.594279716251208,0.06602646613574803,0.691143512313518,
 0.8363010957299922,0.4104713791103586,0.8426215795794684,0.341773646266073,
 0.26585807743304213,0.6978348263592813,0.624540548752329,0.01684239683125599,
 0.7641758878341206,0.9246799796011774,0.5804830624807477,0.4719499322350843,
 0.2577918669235535,0.26294768410940694,0.4393194925291476,0.5590252748085238,
 0.6796534816347477,0.12083248975293126,0.8992963277017648,0.6656263383621429,
 0.19225352497729553,0.7363970094407714,0.5778240610808524,0.7665374583887978,
 0.03819013867358656,0.6116415157670849,0.5746397277478756,
 0.29250277075037057,0.05566507908279561,0.676354273950187,0.530461661939834,
 0.19193553588293033,0.7334578898151068,0.9487591164628523,
 0.10035897777157221,0.7200338480382587,0.041899087330299456,
 0.44244498854052305,0.04892376515966068,0.5862793100610506,0.640983346012445,
 0.30657622336322077,0.28208455937699994,0.804029390381607,0.9130439863587613,
 0.9146234951141865,0.48638207353117924,0.9498802685782635,0.7844756934363496,
 0.2824958512971557,0.9778672683131379,0.6513853682250903,0.21076361441853808,
 0.9554839007068129,0.390138655715383,0.2842234369620601,0.19271403634690853,
 0.2041113662220988,0.8562543922141491,0.5280889768012952,0.5998014193237513,
 0.871996907283432,0.07013195082288326,0.22056470976795617,
 0.30009011456839796,0.41475402484244117,0.28557108901810235,
 0.4714175207134855,0.3535744408303102,0.24129605923966746,0.7343762059650921,
 0.1766452238802555,0.28827165861871196,0.6606063978869858,0.6538965811722143,
 0.9585233450769949,0.5912620190014097,0.7641416478470417,0.7716572169162733,
 0.560183704256358,0.6728569503651929,0.13572162079202466,0.999779108342892,
 0.2378250741212522,0.6601599547841424,0.8696078875562092,0.9672362779619516,
 0.3218647996329499,0.2578391727874726,0.2992304746169998,0.3898100149042476,
 0.1455269836406754,0.22793465210333863,0.07143962763205003,
 0.7075478924749947,0.23785115402657175,0.4684164351192831,0.8380066470389458,
 0.24041063882199876,0.07344723292491773,0.47276415135934724,
 0.21720540510451714,0.6672873150174109,0.17236846267721273,
 0.8580825728979636,0.7289581605065228,0.511590185285813,0.6757261269602554,
 0.3265526179445719,0.9592448222597663,0.9958108124982266,0.9004544619752001,
 0.19074013861022143,0.45427292124625396,0.6080940075854325,
 0.2887378135197487,0.985974346245783,0.5975323180658282,0.040296711425680476,
 0.08381165001844137,0.9186274092838641,0.5675334420366229,0.7749188999515161,
 0.4865001957266496,0.9741686872973515,0.07525634080848342,
 0.21290129318816087,0.6667559199976155,0.6681327732114285,0.6338523077293501,
 0.9051601337893649,0.9007944945055689,0.0880840664636302,0.8077330291613529,
 0.930871955238629,0.6999406463183394,0.47835531773493567,0.6170883910613237,
 0.40153193661428,0.9302036191710245,0.6982231743113612,0.7746640340545876,
 0.18903451902836033,0.8606089043159396,0.9174450182689177,0.1628282876246795,
 0.16733811091837025,0.7005652593605995,0.29004077401518114,
 0.6699061899623301,0.21214847232322587,0.408235521989166,0.4025587798481235,
 0.20917369605214986,0.6778580760677769,0.15407068493991605,
 0.7066571030760763,0.9229463526101584,0.1426071226188962,0.6362771587315823,
 0.7458396814805818,0.21011463290293497,0.9285348918471468,
 0.0059885697397568105,0.7935363879211006,0.7394416461958128,
 0.5281622624446229,0.09292452550982311,0.3179656617126323,0.868030097552588,
 0.24653724146094624,0.9844572024321747,0.2223744412590879,0.8944211073447723,
 0.21133993780732963,0.514287663593164,0.5874677100453625,0.41450539552185917,
 0.7364256278509796,0.297408829659654,0.9445215560676122,0.9415635618919741,
 0.6894426320185236,0.5713461063401554,0.31526887502338863,0.4631393993598225,
 0.1454839723096809,0.6996761195263773,0.6626808864091402,0.7519604329177335,
 0.6808497423576987,0.38846857767052134,0.18494058980723882,
 0.2858740677998537,0.8405302902277302,0.664049921791768,0.7644606413859445,
 0.29738679655412603,0.6531096171066005,0.22905069043467896,
 0.39167873487899174,0.6835858824299766,0.708879223377474,0.38959334701163584,
 0.6407013478125916,0.6970290022986927,0.209185072624398,0.8777940693684001,
 0.9628283810696078,0.39172547941636693,0.40084520100140086,
 0.5281323342305392,0.5260060903362437,0.5698489804205953,0.14809149115119224,
 0.48291083169471816,0.330801720097448,0.7589779190957929,0.9653460826248392,
 0.0756019816924487,0.4758173401430611,0.06300914117892553,0.8688800701347658,
 0.3873421425821917,0.513117730535964,0.4200053247643255,0.4725644250822819,
 0.8248244399557527,0.15567958781060176,0.4705675186333649,0.1848806667891587,
 0.36815574126108497,0.5369906864610592,0.8272429499019411,0.7189122246975466,
 0.5590328681806677,0.4836801218154785,0.6387151892594319,0.4359434865753642,
 0.6810674611608362,0.2804582257945373,0.2208425904795125,0.49652427013059186,
 0.47165264947284524,0.87319765453277,0.3394911818302536,0.7293216384447744,
 0.3237327514444004,0.29175609510239386,0.7617533190446314,
 0.14718180985198082,0.25479360757218195,0.15472168420719623,
 0.42756642400521794,0.11667830305539029,0.9124456802745626,
 0.5897939588918226,0.9428711017136672,0.19085708797164802,0.4507340780227793,
 0.8608899324047116,0.6607317038273581,0.625354396786028,0.00292113342698741,
 0.9520213399049308,0.42215592204061814,0.6056117346834492,
 0.21953556438689803,0.6918126490910217,0.7830446433040726,
 0.11816579757760104,0.9231516173092956,0.9616359322205199,0.4731703873776757,
 0.340369590573805,0.5073132265956187,0.8187697063082399,0.062113371481778756,
 0.6058056272052602,0.1877437857338552,0.8446605120004581,0.20118758694258787,
 0.01389163912515834,0.9229010537123675,0.3945837065150819,
 0.06086026522668764,0.31819093556957867,0.5976460095058082,
 0.7696509517586014,0.5668578366078363,0.5372241287854397,0.35055927276107945,
 0.8311444276191726,0.1422899727297845,0.5447987440137746,0.13774693156672047,
 0.35515650552971256,0.5566660336516668,0.202089672534703,0.5186603645504122,
 0.10337549424200376,0.6791094277970894,0.33945074614484994,
 0.4271572118644784,0.5334633579519016,0.7666831387041512,0.30269472749903914,
 0.4105957335994267,0.2293111803677217,0.6182966679598529,0.39948949056988403,
 0.6470585487174809,0.02189047279329126,0.2468499722046401,0.6027758675284273,
 0.3961311082875345,0.29432662870684667,0.10793991049742524,
 0.41706253467116505,0.48043068700746916,0.28315794981312226,
 0.3467737345622931,0.03808389395771028,0.006628954904898232,
 0.09381148566789055,0.6687792346858736,0.9346730987217591,0.3221510005887843,
 0.9167654279341291,0.4580666742308125,0.5204613033680634,0.46505734093190054,
 0.824745436624334,0.11253250053982922,0.7342589003195474,0.942475423747544,
 0.75700723674921,0.41698898502241966,0.46651886467813863,0.44849724884082,
 0.33251125395602044,0.8559763302087876,0.41721769910534057,0.852038048775745,
 0.9815346332277614,0.2084527662941541,0.4250492271410462,0.11954247162472775,
 0.5178910632530882,0.28996553680048665,0.6952122810038567,
 0.17615917267366987,0.8761463376087258,0.11439866226849915,
 0.9506876050973404,0.943374012488877,0.40765280915888646,
 0.050782233619442896,0.04043494354221766,0.8325917887124658,
 0.4017257016209471,0.12524451570508388,0.328496922440881,0.9734747860838492,
 0.3988679462954974,0.9362034048034515,0.394020056566615,0.6775529305264278,
 0.8546180578698512,0.8801415919745182,0.15498597036557016,
 0.010538675450331492,0.019115598932724853,0.9414267447535444,
 0.06408927670011255,0.7771963608561633,0.6418635067176597,0.4813923878625219,
 0.7500844418235273,0.39187602894398954,0.2259355006217355,
 0.07217966250325614,0.853573773011241,0.02087231540363854,0.3706601318780418,
 0.5443387992853633,0.935509261083456,0.32350286098747694,0.385491751051514,
 0.6879536552668631,0.34185464201588656,0.6902744505663345,0.9457188354044607,
 0.2636484956345245,0.8576420525020637,0.8551765456604697,0.29588313756463125,
 0.026579698608857893,0.36936551552740715,0.9109922307036771,
 0.9591460861077277,0.8406732753225687,0.7144817899178825,0.7355544334710712,
 0.47397983339631133,0.6890882321409221,0.8229542281627249,
 0.007392060097412001,0.42334044785079905,0.27871562315781995,
 0.8411140724842429,0.169790373000123,0.02071236428998191,0.7024118178833522,
 0.40078237242872317,0.5413130075062288,0.8174226708350038,0.873220725790212,
 0.3034565342455706,0.6018205774017609,0.26315350257230696,0.6263614165227138,
 0.9739042864549133,0.4110279385252599,0.3325463638149915,0.3324362501706224,
 0.402612025704394,0.13190753645183462,0.7784632558323716,0.5296614752155255,
 0.2152496738791445,0.4472599018388487,0.1857000676550714,0.6607964122340901,
 0.3808494760076937,0.17340890097831907,0.6661799920593646,
 0.47117206653638577,0.5582812791904291,0.20857692949292805,
 0.3981663785234354,0.24610034426948624,0.007850289618472783,
 0.6405433175865053,0.3088076826569265,0.3890115751432438,0.65415147486453,
 0.9095242823214841,0.11182371327825302,0.5006099625932484,0.9563373652465844,
 0.5717692575163213,0.6952050336527642,0.5647665450367026,0.7854708647630231,
 0.6456830787147669,0.473818280490234,0.542507050301116,0.44764458360362214,
 0.9704228744672915,0.6678223036994639,0.589172281946441,0.511219508652036,
 0.06204932290214438,0.7710183364564396,0.704713181951518,0.8955668155511729,
 0.4172439757577311,0.3924902374525566,0.04819282368122879,0.8185364814313779,
 0.7126621388053852,0.3163816108357088,0.07997769062050541,
 0.32063221065933745,0.7453233540809898,0.2562471895923325,0.695525656148534,
 0.72411538757505,0.3545872958394294,0.4082460937217548,0.801710719178317,
 7.007202727667972e-4,0.040773618430270275,0.9922493633283814,
 0.94355832280257,0.5664599383587571,0.09282544273253379,0.6196289042447287,
 0.29025067262712856,0.5334948543984883,0.029381222684199737,
 0.6297526059270118,0.025216816343942128,0.8929025679441656,
 0.20123815568445158,0.3793506105142568,0.6442813448224189,0.6405128610167641,
 0.6578695856281771,0.3789074512249371,0.6332034082514252,0.495771981198172,
 0.5841853849043416,0.8792120550653375,0.6696594824403966,0.687423691252767,
 0.36260517590342733,0.43365792163258443,0.21824837300266886,
 0.7767118182280287,0.057887616412197196,0.3374885284813156,
 0.025993644627727353,0.7409908702152215,0.2003873251476368,
 0.7752454866227756,0.21056355203348887,0.2893604225584243,
 0.37704137832458495,0.4206253990155807,0.5253678015231891,0.5979860483579352,
 0.21800112738835853,0.6013623920943209,0.5242541193718449,
 0.11760168462561515,0.851220277465867,0.6170025023260082,0.536998282403084,
 0.9027244885640777,0.9488066582308337,0.40479692800924794,0.6031116239007259,
 0.037703621805487586,0.7024847625498644,0.37838654488803136,
 0.48293951807071667,0.5880891793003545,0.02218508427929733,
 0.5071035609176263,0.8336515358154555,0.6422640187539072,0.3925165809751895,
 0.5143861991962406,0.020417409057284508,0.12278587261524221,
 0.5038994776972067,0.9600472276682598,0.3370770742192929,0.564382758196869,
 0.18473206915439855,0.07647746556000112,0.13245629253872881,
 0.9844290265437801,0.24011597332435197,0.8332394877499415,
 0.42588687673624825,0.19590759431477,0.748001493918839,0.626808528078383,
 0.9437905017075146,0.7448132382819374,0.5438834247862003,0.1138986724045179,
 0.42560385017023505,0.4320479952044559,0.762911167480856,0.334968920413414,
 0.10221939250750045,0.7859193270492462,0.5151845044061142,
 0.11981410943818138,0.3748343024394557,0.015152933028357074,
 0.689327742276191,0.6556244765121809,0.7564877018740266,0.5666218876555267,
 0.6515647463943174,0.1435935978710341,0.9463808854929345,0.5484874836753985,
 0.3034202366401596,0.552896344113473,0.031355359362339796,0.7475323816171434,
 0.6659446013061682,0.2100380989117111,0.09476489583089776,0.3476270602782443,
 0.5088555019073181,0.9640139779346444,0.09305001877100849,0.8773296478774196,
 0.8387801721443835,0.3181803424080273,0.3153677952208055,0.8741022095381338,
 0.4859303095891162,0.23690040197771767,0.9680241065695349,0.7318649279323061,
 0.36136218755296245,0.18033403363737477,0.1302677246094277,
 0.35620835576353704,0.6969383761302773,0.5053219643284576,0.249367507888563,
 0.19490936839475814,0.28718319999772446,0.7403662763193947,
 0.34816913322201715,0.6128507991520923,0.5563379765145771,
 0.24687145174027847,0.7930837048731282,0.4779046294359658,0.9643827179104132,
 0.7034628151862631,0.865807131000628,0.8641921560811968,0.4528139419975783,
 0.8135460154113473,0.31070208926144705,0.4751648986226287,
 0.13539160023239594,0.5916803188470837,0.8661133551897554,0.973820868252137,
 0.6173392008855141,0.9310630577821153,0.4319757381602731,0.13388788190634182,
 0.4799265889266926,0.87742686962056,0.9343229901103653,0.3559929072815863,
 0.9924474251545448,0.893437886742524,0.5386246760183846,0.23247814916476317,
 0.7251277184331232,0.04501075271459665,0.5931600382311323,
 0.30011196812671637,0.33358208553417446,0.03299459506105462,
 0.8550516506137885,0.061677490337523344,0.32618682211738803,
 0.2640132911777501,0.18958784369355586,0.521199383464306,0.9039829350076556,
 0.6944477420026848,0.7337491978824542,0.7120707025781049,0.15821575341384964,
 0.597451818167628,0.37123515583866695,0.8518071565333308,0.47582479579381987,
 0.31908063046490565,0.30674297286746266,0.25738845534038135,
 0.004006561984276291,0.06914778030847968,0.2134889514752597,
 0.012391296153141496,0.14918045837819538,0.6221860140949953,
 0.48182074057054347,0.05413671391891339,0.6801165371552126,
 0.09033719209969215,0.7070237572224185,0.6350470368163362,0.6710356539598582,
 0.8234514332695165,0.48387685998856034,0.40240705655441356,
 0.7686177416374124,0.34380247421953536,0.24300085316552023,
 0.5318233830090793,0.32687730870104303,0.17776589180598246,
 0.9132427518068953,0.5306902514286014,0.8663725227836344,0.5644814406007322,
 0.019191654691272486,0.9550268276933732,0.7903233838770889,
 0.8127356999576474,0.4338889550764422,0.17570024906846582,
 0.017880683031576194,0.863065090740677,0.34249786916574665,
 0.9234383890958809,0.39294180336959084,0.3865023245812437,0.6134875682660452,
 0.1445548878459908,0.5761648126080228,0.8545138869871988,0.05452740038085058,
 0.9418073363686351,0.9850789070389613,0.9063349633523545,0.8315565923591877,
 0.010111583532526947,0.9825641638777187,0.1401821504979004,0.681600960188351,
 0.28478030300423196,0.5405360618119648,0.8393737621072725,0.4316363856993115,
 0.5328108097895448,0.6479042757473356,0.016581983348963325,
 0.5258103919879127,0.4399871656986839,0.8135054527931032,0.9825306930219926,
 0.13807311597023986,0.02685617781071603,0.7852084171889807,
 0.7069893961685976,0.11455663502503655,0.4808906084057182,
 0.36982928283024774,0.11091217864151126,0.9110739106877869,
 0.4837871167114083,0.7297148030757579,0.6672147364102826,0.8765981556350038,
 0.199967174422767,0.7058749523928652,0.97000171524232,0.534421588784844,
 0.27929434458182645,0.9252774455112387,0.18161900809021492,0.617817419322639,
 0.4368699187200702,0.5737245987891819,0.9400026018086978,0.4155550873256917,
 0.117658889013456,0.7301205652822403,0.013635218228179147,0.7056104885495005,
 0.7398028740697273,0.7639812398677667,0.43158621993842505,
 0.12558924738198485,0.8815585002605412,0.35920421099698097,
 0.8779607457416925,0.38590647836508496,0.5442885904843501,0.2626826704600298,
 0.11393464431765044,0.5901621924561513,0.33169512435824045,
 0.018538002998758518,0.06374998154794476,0.17647464406176627,
 0.3613113970260915,0.7654984875607136,0.11690277009621597,0.6172818343959756,
 0.30182525585423914,0.4774321975979636,0.7273833768454392,
 0.13682442684240292,0.7161094563774905,0.46116515287000936,
 0.9533511833982099,0.7932358534134922,0.7493286033592543,0.01415239092244125,
 0.3969386536107389,0.7170125849173086,0.44377804918245956,0.8348061867403836,
 0.48307827427915373,0.2322523344765517,0.9762734304190442,0.9602459285099035,
 0.7352013264324314,0.5426098813827801,0.24598748552896266,0.368647115164214,
 0.13229729862366946,0.7474245401784354,0.5600005478679373,0.3426557634951204,
 0.7042370262803045,0.9106924636741632,0.4088575420588664,0.9174868051854581,
 0.11291907000657708,0.20310088449929453,0.23474679193920678,
 0.04470094664671542,0.4078261522185289,0.12005846886070615,0.989092567549478,
 0.3892576223510884,0.18583601219291168,0.724864019352751,0.8966230815113697,
 0.8195561551292033,0.34650117636849287,0.1335110258560549,0.6353700336817296,
 0.5366616507501634,0.7611118028442033,0.08346239553947399,0.8711280541765266,
 0.86618328047755,0.05799676475326365,0.5778689046926295,0.1833256941791206,
 0.7685179330495985,0.19151147517871758,0.476767376670842,0.2868422332266696,
 0.8770015156362948,0.4991187307118493,0.3930341435199137,0.6400987494968445,
 0.12424355762138672,0.26820457778180484,0.3913032532140902,
 0.8438196528018551,0.7991374453691908,0.2887387563139192,
 0.028989767190173876,0.5616974444142301,0.11774961273220164,
 0.8326375985475151,0.7515183367132932,0.914817224273216,0.1626573152796319,
 0.22671961280956876,0.08607577318628679,0.07928275991046574,
 0.2690692200262339,0.9452524317142811,0.25630863103321133,
 0.39207302962858637,0.0733857198096487,0.47435563076368026,
 0.4291025128029955,0.16092137908292325,0.4148829820919866,
 0.48429336833477143,0.8521912732362307,0.035230976841835404,
 0.5463007501190135,0.876984212169557,0.1002865520788917,0.9972313425219943,
 0.881840342645857,0.3751141223399642,0.1932984798551971,0.12032701691886405,
 0.022662406917943856,0.18942665922420177,0.3080675974408861,
 0.6652568977183477,0.6991088566701652,0.6002285525786155,0.5410067214075645,
 0.9434809892178546,0.7866519980686719,0.24715164398570866,
 0.27728995168096193,0.7085566750658301,0.1699675512667368,0.806533410908069,
 0.3645729636828584,0.9587269653988875,0.6387292582595023,0.819303591416712,
 0.31296678884558515,0.3566442938798353,0.8180091041709501,0.6896896494959089,
 0.31211720735094706,0.7652345916776592,0.5145842782151535,0.406792291345348,
 0.8881067084282106,0.40082000791910544,0.85019009781864,0.43101122020197813,
 0.6233193971033602,0.7690146674176694,0.551964304549076,0.5001335384434746,
 0.5333832245520206,0.4082895819674336,0.09542167709069582,0.6599128277273042,
 0.5421443153473521,0.29401866280563527,0.0798967526277663,0.964319619822909,
 0.5162834659695619,0.4187283680302559,0.46280046896971394,
 0.06848015240403438,0.8932930388065726,0.5611584793791715,
 0.31022946660809214,0.3199905143356686,0.08059593358727934,
 0.5775199671391182,0.49046773181098013,0.4944738417578788,0.8162197326319239,
 0.8131264565066293,0.8822640672181138,0.37510272046299553,0.5115594533705756,
 0.7407959849801969,0.15447993415784078,0.50565837095949,0.30131734847738456,
 0.5453920326347697,0.1354490980178018,0.8401296376023186,0.7756689635933725,
 0.6149781099209913,0.5602542747218832,0.1720627143408353,0.9891852175409243,
 0.19528042814426583,0.4793611983467321,0.96314122338625,0.46704551848702003,
 0.3859413928901604,0.2696424620700768,0.4800054285455898,0.7917378655327668,
 0.9211705798194302,0.04742152190350857,0.7022481692660787,0.8405891402756424,
 0.7592124587612562,0.004394805339786423,0.8405000996360288,
 0.17595629826307924,0.48148723663097626,0.08555314009202009,
 0.16948514427277706,0.9098348415958974,0.3983510064298602,0.875381484479537,
 0.5483585157646029,0.9059796211147615,0.9959350101533392,0.12078511243770851,
 0.696546061169175,0.09828428017633728,0.9358954947138535,0.8596219453554073,
 0.9258895022306393,0.97987384463505,0.152638003812121,0.030854428592395178,
 0.37536075093592225,0.5055258045501314,0.37976369268043175,
 0.7770823143649205,0.1203908741789339,0.02715880866457754,
 0.25821160089469464,0.5899850107935348,0.817029892961316,0.0955016756260505,
 0.78724605479306,0.04839840432856679,0.04679272687369451,0.03501159203428361,
 0.5988282365638845,0.4016145335427048,0.06806124314428863,0.7801633722189129,
 0.6930135179386254,0.8680389604799045,0.827510674105318,0.9925084624945423,
 0.27528006263034777,0.7318619610287362,0.7949250601785545,
 0.47804083134304587,0.1380175679684168,0.817132386834513,0.4551741175745796,
 0.6442901621588454,0.9955222276744825,0.5756968501220195,0.6214577818894024,
 0.04937512572488767,0.6381837500965863,0.45154487898758344,
 0.6600372030915598,0.7677388714129477,0.8601924686306035,0.05511584101361433,
 0.8382655223120581,0.5348807257018169,0.9414606117412545,0.5843140662738493,
 0.36069249848904594,0.8899596669254602,0.7824689111176886,0.7509837389042928,
 0.6322700776532928,0.7873904902208537,0.9662634253603583,0.1365215313342828,
 0.7294442892384883,0.12010128891697014,0.45707732124430744,
 0.6637700993381036,0.10988315608543509,0.7356489868040158,0.7067192170532194,
 0.5990698000438952,0.8050897442425636,0.6518525630998336,0.2838929938313458,
 0.38700367025016946,0.1807178358598749,0.055175072680873405,
 0.44017129427885204,0.5698295016811961,0.15577211228141985,
 0.7232690468578795,0.5705206713366306,0.6873458926848522,0.3248853763549342,
 0.12682031067478117,0.6711479371871738,0.5926279972940868,0.2657613653518164,
 0.1873049300067532,0.12977383273239385,0.9876265337905289,0.8158489609377859,
 0.12054366027620877,0.11569421902775268,0.8573747529054767,
 0.05268023976345848,0.3484611057513516,0.12196704771411337,
 0.9812005960594766,0.9748821352002778,0.9128583316693643,0.21117854123896196,
 0.2506685082605645,0.7103342545993742,0.36381312861764414,0.3223467431983642,
 0.0378195429512187,0.3598890808555544,0.22073046456052958,0.9410599216100798,
 0.5721805388477463,0.746443126788301,0.09921581460686413,0.7629969305584647,
 0.9043109567105218,0.45484503454500214,0.9740259784109037,0.2975370488645126,
 0.20015694373266957,0.4188797249027246,0.725553345153873,0.6475063470880887,
 0.16058100604928993,0.5290729567783341,0.7562240006807451,0.7499788143152417,
 0.909279211736696,0.3339899760208487,0.428275984247927,0.34712221263110143,
 0.05715821786847397,0.8972576011467518,0.6482265833047233,0.7870191375392187,
 0.7472898200822554,0.9339180722249527,2.1047272350638302e-4,
 0.1761812519877477,0.9385378415029795,0.26316874717392424,0.1840145014869431,
 0.05731096433132565,0.13147656236124694,0.9081537022508587,
 0.09462957157847096,0.8652370202833631,0.026700109736640165,
 0.3066870560457432,0.7421154314074028,0.6881123349469227,0.41372897015369003,
 0.9177949645811699,0.11927359459586029,0.49155862680054796,
 0.7598436614436743,0.15346028341727225,0.625574508933127,0.717566994041875,
 0.28215181196258154,0.19059856897782312,0.3969540844464342,
 0.9234581926381042,0.5745046442887234,0.8158874452412782,0.9583373842954551,
 0.7406287194432306,0.6181994537720804,0.6730022494454753,0.41680914114858747,
 0.20973863470813447,0.09052142475849911,0.36644193406691705,
 0.727321641742712,0.17232940597026225,0.6718780068316955,0.6808021916913986,
 0.44700792477495144,0.2199699761763525,0.8659042637286908,0.4607182338039031,
 0.7211517940117964,0.3443221346346629,0.18410374015645325,0.3349852819019178,
 0.058726476579085474,0.32802633866660236,0.49244232054513826,
 0.5493046553860497,0.9149869169904498,0.5165617638110128,0.8255090131582552,
 0.8316554621256553,0.8043842186912632,0.28851440561922337,0.5184822177712634,
 0.45858724108158233,0.6682416687869326,0.1175883852366495,
 0.11121629692555535,0.5939772972102569,0.2880677737226218,0.2644869162054446,
 0.06020128235315325,0.5789219640232818,0.6983874705322831,0.7573156403333889,
 0.44573417073507904,0.6068408487862411,0.7339921042256157,0.7726273068644568,
 0.86657777646306,0.3552271384613165,0.14747074947626926,0.1777229644798901,
 0.7896166815842434,0.10046404206658588,0.13038809538891072,0.748097619536511,
 0.749001573674956,0.5482401584318951,0.03687393245998105,0.36126159851099293,
 0.27589851342761706,0.46615899490628243,0.8986364358689027,
 0.8601552326921117,0.8325385778227471,0.827117116372595,0.3735900418829522,
 0.1389709001495819,0.9412886278486844,0.14230602139257098,
 0.40471666303325116,0.7635747771608661,0.9532334636444474,
 0.10599694197557952,0.1329144884548803,0.6862836608973566,
 0.44421150441331037,0.22897403423791896,0.6655580710851042,
 0.5995912880758315,0.4147025988482036,0.3612193343784301,0.2405812878248932,
 0.2265276638398448,0.6061908933998621,0.5877128548322521,0.9908191162017271,
 0.9105537505566397,0.25676261493859853,0.9878121808471615,0.22629563860587,
 0.04350802888048377,0.3089645258024716,0.5974006195314672,0.5673124016970995,
 0.9722118393614042,0.6219873671802999,0.1494341121891264,0.5938229189810371,
 0.07526276044329339,0.7010731709173855,0.08595840020204892,
 0.3351713751747124,0.8771415625491313,0.24156928951598078,
 0.026325421143544414,0.023595297539423243,0.22218387985061872,
 0.28090028664936806,0.2092437162294417,0.07914605750899051,
 0.5108192086702934,0.4557504427089617,0.694768861878078,0.366036418085377,
 0.5319767563349227,0.5139024566036483,0.25603949273921456,
 0.019987202381001268,0.14004217164385868,0.13174803828894033,
 0.21690487383224188,0.06590636718078169,0.8970035995033026,
 0.7487267611192711,0.4000673908287562,0.5784450551191497,0.03132601280818592,
 0.9246952825738215,0.39851694081368993,0.8697313391899038,0.623822190602873,
 0.49931844018467963,0.3438260618156921,0.021832003631849006,
 0.6572464623724226,0.23213692871746605,0.3032547872673038,0.0205401849686907,
 0.2611821847271355,0.26514710988117907,0.9976424657087064,0.8446643356825349,
 0.7642735738993318,0.8069514713244399,0.27074963208095837,0.7541830211196167,
 0.015515497454952398,0.7410189280310423,0.8511859507568381,
 0.5756071896776612,0.5910428279576851,0.6611326309182379,0.4588280213108642,
 0.7473502298246341,0.7451647042602196,0.6109659688761533,0.47910752021898795,
 0.4218639458544231,0.6772828193875327,0.668863334319427,0.5303389913780732,
 0.47465250567068396,0.6278063766031674,0.06424648198360083,
 0.18214615573921822,0.17155047609833818,0.6046617264643471,
 0.9003019226789158,0.5508801180638306,0.8356958113665374,0.3814515686493496,
 0.6731567852819271,0.21173925352364154,0.3451968274613182,
 0.42041769852578614,0.2559533637852113,0.7690766411718339,0.7174404321151395,
 0.20016001578335996,0.5048029143320037,0.6978168778927797,0.6582910713187458,
 0.8158094588863393,0.22010519564238118,0.6664964027291427,
 0.08323816734668077,0.31348168527226217,0.6926388895873599,0.361102653107916,
 0.16668140167048073,0.6253754379199663,0.5315003868359355,0.9740470812354729,
 0.5361818482334819,0.5271649225203205,0.6727152058765291,0.23456662778351955,
 0.046307378503795404,0.7029273064344328,0.29570208340844295,
 0.10314158795530537,0.8869296420984394,0.9023410837246694,0.5139990643476913,
 0.059575934960569676,0.04617938081722406,0.36852465163427206,
 0.7394506390049254,0.18994118464048715,0.13377298707877383,
 0.05534817577357998,0.5666191331155851,0.6530016109803628,
 0.22138988424812311,0.3567122213561631,0.24123097885243472,
 0.23904788744368566,0.2972872156671036,0.20210365597500968,
 0.5435685157686724,0.6572316518094033,0.04695276120450398,0.8693889673581375,
 0.1537446506654896,0.7261267545377841,0.8830292267058604,0.7290756514451894,
 0.8944209689300868,0.5330004966205235,0.4167201361073519,0.9048294037765788,
 0.6288792751224848,0.05047774780453551,0.5906814752207612,0.8754269191595346,
 0.310192478377177,0.5276326347691521,9.465004071603778e-4,
 0.22926672190324338,0.26469434715304885,0.13838006222994748,
 0.42835573619316936,0.0543213767330476,0.249483576636933,0.6995347117604074,
 0.3064355667487054,0.5846192414980503,0.2719922829050049,0.9386097864967421,
 0.023727672975341618,0.24306297450095848,0.34140822093982304,
 0.22943785005053685,0.9516219159752981,0.13660809255864226,
 0.11500887178125241,0.35572888824369797,0.3433443468961117,
 0.5126977450525281,0.8074324530910347,0.6424119399132286,
 0.018655895785955945,0.5393666188592162,0.8415836570935401,
 0.26416852041398686,0.3919514206675163,0.5019772494601012,
 0.07967880463875843,0.5644687116851586,0.06768513562982292,
 0.8766491007489472,0.5821669931892878,0.6527588982246247,0.5688640208750676,
 0.42784559900203545,0.03285691519349876,0.0779846560045091,
 0.4888540444439684,0.7671247326204078,0.1639462262960094,0.2982764863671239,
 0.08190692503989427,0.357725573217724,0.19488591025181035,0.6497566354310411,
 0.1271410374189591,0.9841646000274522,0.5919464473877601,0.9698014103936946,
 0.40475209353112473,0.1739625752214199,0.32783632508872107,
 0.9741852110198232,0.8190901891260827,0.46615317742696305,0.6011529458784909,
 0.18030222853312095,0.9087092362662953,0.23364468263669647,
 0.6085905679315573,0.9946580305564331,0.12948591718678182,0.9188670109091902,
 0.1370890914701668,0.0010712042162298596,0.49155217998063105,
 0.5984267104537752,0.5419933401668121,0.4930825383323476,0.1090027702677745,
 0.43440283175749594,0.6493899285071463,0.980663011899073,0.08180527687221573,
 0.582678383950214,0.28505264647439577,0.7005889375912129,0.8323702846918266,
 0.9968463570861562,0.4215045136699096,0.17894243468978221,
 0.024062871123230356,0.4560937577035167,0.1226564904285774,
 0.5443500808221703,0.8718750845772818,0.011096783839130708,
 0.4911683593244347,0.7334956443604098,0.9544382435870965,0.9286272795191853,
 0.6902595300552418,0.675700775348343,0.4708236101594704,0.5020728844561598,
 0.2975925023993887,0.7232831540230348,0.1491944899278237,0.18933718415088652,
 0.34155409645707,0.15188729529425526,0.18827723465451474,0.7317058823242862,
 0.6616489286887411,0.2625402839209253,0.6538883549071662,0.6840475164006381,
 0.07567947021934929,0.6171290747037126,0.5152340806319494,0.8033215788747633,
 0.8750350434197488,0.9732864410917628,0.08219928866060389,
 0.21523096172126854,0.9486372921887731,0.979022896973186,0.8483434516612768,
 0.3964272000009833,0.44657137983222905,0.09185090883445457,
 0.6107236783414605,0.706264347070583,0.2206936408029947,0.6720196633792443,
 0.04136897847894405,0.5594756657084102,0.7612407835616453,0.9886658138459026,
 0.5570100847442261,0.1484475253772849,0.6131066585271379,0.7316701962603104,
 0.09285902580398686,0.5950777689718302,0.202444512544518,0.4064314485871213,
 0.21550066343139962,0.9656008724173036,0.3170516048873724,0.9510066412804555,
 0.6731830436640521,0.8075660488430714,0.39265636817014227,0.2615253980288448,
 0.8273829716656252,0.17448665209176728,0.06054523363600439,
 0.6356853851723733,0.4502070352463443,0.5412781960280385,0.4421511868920712,
 0.6898274874239173,0.42151515449028454,0.703611680981775,0.3219284110295456,
 0.8042130004102875,0.7190789011732485,0.4805205289362743,0.27842433556855595,
 0.9665814451576193,0.7784868719359542,0.30582627280939645,0.644646002893606,
 0.7055714068323301,0.14913543168301313,0.19214414747318653,
 0.8372803572754989,0.281222835172414,0.6591177619530727,0.2844709157226547,
 0.18991109233083747,0.6334032957004426,0.5489192630774808,0.8894747135763694,
 0.4353840362587572,0.7779941867350996,0.7370274308069389,0.6802473601637604,
 0.13083440413845282,0.03243629040085294,0.6716381689090095,
 0.2566128137558118,0.337045108395458,0.8296460267600597,0.2442041919656469,
 0.17815393557169046,0.09813339463840887,0.6323444721491235,
 0.8139813598674313,0.8296555782533609,0.8234589200659385,0.7213288766473023,
 0.3299464115331543,0.3606408169572537,0.7213116976309671,0.9624019103506314,
 0.47768620399420747,0.9570867251477867,0.5535612575294959,0.7026361171351311,
 0.0017099757978085162,0.8540598356835134,0.5781220036676743,
 0.3990652725841015,0.9208768806687422,0.6282060427039985,0.5975435095870398,
 0.46135779709521874,0.39334784174252946,0.6817734119413928,
 0.22429711659019058,0.4963619165159723,0.5231773349911263,0.9515841824503023,
 0.9072682623668987,0.043366158816752653,0.23629865463788713,
 0.6076587195375475,0.019291950883625897,0.3551038291013101,
 0.8269318554498954,0.5551893455579284,0.9958187347222597,0.27933213869609996,
 0.5734985966828685,0.827206833946523,0.3359523796836078,0.5530021620388983,
 0.9457602765526991,0.14251226090616065,0.6705483512466752,0.5261879679048771,
 0.49771292370418396,0.4461283332499252,0.8843482656120873,0.5530453506015186,
 0.9954243537825843,0.3121567018133029,0.6446386851030472,0.835501529488967,
 0.21668825447867757,0.9361800279403768,0.9952901333007379,0.7945936523289691,
 0.6535625692168144,0.5341903626053692,0.5535752335549328,0.17069558265987061,
 0.4257179710780289,0.9074218173101851,0.04940022208589245,
 0.39223569571118877,0.8657931925242198,0.08933813773724042,
 0.13202018876041244,0.12217291653575324,0.8273904224557462,
 0.29510922625623726,0.8584675514662772,0.3784391763241143,0.6418718295550282,
 0.2680683939009306,0.7289368198896011,0.042131497373042115,
 0.1166684465618707,0.027574038793470645,0.7133652417047353,
 0.43001999006379166,0.8778662079604445,0.9520975212297786,
 0.07314680201323409,0.9276818875266033,0.35802891895880373,
 0.5808743157943409,0.837377837276562,0.4487520733761242,0.47285015645351125,
 0.8040596393774379,0.1496898297779672,0.7561509777101593,0.7545478758367992,
 0.2902204840725692,0.9919351187352592,0.43386865302842237,0.5446973282995126,
 0.8845893448771446,0.8436481661438183,0.6936181146148307,0.2466801011058114,
 0.5238172863434546,0.4422279229232249,0.44171988362481995,
 0.05151895642317439,0.7790591205112937,0.6486881035181878,
 0.22323610395712412,0.30528374624609456,0.09217623176023382,
 0.27139148527551915,0.03140660751315094,0.16901148878243644,
 0.5175265054925099,0.7117309080458816,0.7291977436758854,0.08921266901298897,
 0.3361929434196238,0.4743639542479964,0.12331667702175397,0.4667842668661407,
 0.7460915449383781,0.5881891319820556,0.16441261236230476,0.2120780964618424,
 0.22265138326069545,0.8829262799567921,0.11022910158640076,
 0.8884781170329379,0.7440427988724422,0.5687094737220532,0.21911453956325377,
 0.6539730011810165,0.31037586845460985,0.04865641447633284,
 0.40318819887525237,0.8337514550090405,0.3244119364892557,0.5479971135215568,
 0.29225147879935065,0.8796466578730445,0.5577334354332305,0.223715174946155,
 0.41910803562268073,0.12983251795203243,0.5561994356482984,
 0.3684878483611762,0.1808701294900048,0.6967473682747155,0.1389738887347841,
 0.004249787652564896,0.06598680153083503,0.41392432363938925,
 0.3390787266445059,0.6530537613002956,0.4975752994721473,0.19460803929804238,
 0.8859083669651061,0.9186015966590863,0.4382767495231915,0.8572021181456537,
 0.3175319873818252,0.6697863008327068,0.3947041405140287,0.7783871352077621,
 0.6897525319431361,0.947426123119988,0.4339785912171723,0.9527448372751358,
 0.855774433804597,0.011967281006416641,0.7909379604057449,0.6159262156103306,
 0.7282711787198141,0.9742223625783712,0.0503565238510274,0.9969168887630906,
 0.9566300196259244,0.39188521952706024,0.05671728327841985,
 0.7118680768882816,0.19572283077865682,0.13639818672537407,
 0.16528243848359414,0.03308678952282196,0.01267866898415937,
 0.7623652569144308,0.8695808103246715,0.49761435671520826,0.6085348316227448,
 0.4750564047978134,0.5018287642492318,0.029778766632011844,
 0.37417722189604063,0.7763796747675473,0.9463044101218588,0.3250913510530338,
 0.9598505561418639,0.0241578831354754,0.548664429983175,0.18076489368571402,
 0.011479982539850053,0.27352923410426533,0.11475154442608715,
 0.13908652685630285,0.4442274105223373,0.07594107951868967,0.650325718749752,
 0.5760915125563877,0.042749014858568635,0.7866356218403172,
 0.7370189249642685,0.4458535850001717,0.8420997886375297,0.10145500494123116,
 0.916632753066669,0.19463820781077534,0.5391064632327474,0.31375600041504437,
 0.25496913307240265,0.47300372429822246,0.8787004912982466,
 0.4783049143682032,0.5009752089699323,0.5158146778044588,0.5376481343649713,
 0.6776855649469361,0.15754887992961553,0.31935736343888055,
 0.31305484820948626,0.9819055286024762,0.15875514716520622,
 0.7932567731555742,0.79755541514858,0.532367151903835,0.8241580819791335,
 0.2624451031425227,0.6596212418976185,0.22037156697498828,0.4032310601740039,
 0.5186451451527533,0.9937272235496204,0.5067565421267904,
 0.028370674246914973,0.20105271732914176,0.7593071685762864,
 0.11084224713167667,0.8767866487566259,0.023878010880717837,0.7547905717415,
 0.050003634534305896,0.1262298914995441,0.738496429456751,0.6821308574084236,
 0.019222237010188548,0.43242107876328795,0.4889594030227956,
 0.7553795425833643,0.4412411634557677,0.028713597035267435,
 0.7019274402004524,0.3075930631199657,0.7766223050273064,0.49174376915572204,
 0.29112593506174766,0.10876687275487096,0.9824932101340814,
 0.6100529307925856,0.12361096246660452,0.5183430221493045,
 0.09502089790197266,0.08476349793373217,0.8730876698858787,
 0.5207683146367954,0.5471184743629355,0.9112023071292561,0.8975456935024957,
 0.7355060281444743,0.86173302870822,0.08368771039835432,0.3905753457851664,
 0.16538876602807184,0.4527833175864674,0.12628611308517124,
 0.3112047961100588,0.9289223263775888,0.5422518642456304,0.17628453480973583,
 0.08969633335841942,0.6073712121105558,0.3512890282739527,0.893937662503468,
 0.217961549158321,0.624775450794872,0.515999540926417,0.867707345959456,
 0.6128701002523643,0.24447058285836099,0.7815547355902553,0.277257058260424,
 0.19625763258334228,0.19586377629545626,0.7974213010578971,
 0.7081785165205301,0.8203117997356955,0.9827770097095402,0.17440612201065153,
 0.8127221055875216,0.1334716551590729,0.022775616148975053,
 0.38204610578074993,0.16151526173653719,0.9115148021583481,
 0.8611369704356145,0.9365141120553873,0.24426572376162436,
 0.07602800488294226,0.9772082080683201,0.06678946726703827,
 0.3664370262318758,0.45954191015805934,0.2943245676405605,
 0.49873024766029295,0.6509591772441786,0.688097891735083,0.5994875681374171,
 0.5439428544657465,0.6869364988840702,0.6302246649088974,0.8252791105934741,
 0.23138275331349845,0.8133042428154625,0.5086685498613631,0.5964507840981125,
 0.5091780138610231,0.24393330998442442,0.749165411244991,0.41187319819077006,
 0.7996800399181351,0.9342393159911808,0.792601855793296,0.0709022199800512,
 0.028267383250456968,0.8947498095544885,0.7572591815495372,
 0.7526861708319401,0.31170586650055854,0.19053606935154443,
 0.8910351189498418,0.7920752123876726,0.29271649858318505,0.5964940283986813,
 0.1402513221281907,0.15844792207168512,0.7622114854232848,
 0.07417170961328656,0.39356597352564937,0.5873564596403049,
 0.060760482504508606,0.7720699989164354,0.48787779588312374,
 0.2660077137684569,0.3077110817143207,0.47210432515050105,0.8057688460433046,
 0.30046843947072677,0.6636245005311262,0.7249382134237472,0.7074589726056366,
 0.9219873465631775,0.16566169096189454,0.005856094827019986,
 0.2213141493683064,0.18006230119331235,0.27665482226217797,
 0.39934032109253403,0.920975258254088,0.21667929174657274,0.6621094421810916,
 0.47982708824865605,0.4977968767077624,0.9901867292141155,0.5360760177343102,
 0.2464784587467027,0.5339175478484766,0.9430054997751568,0.5404755259500019,
 0.012381968694525347,0.7293066796091605,0.9237147140227496,
 0.6445400681615029,0.7016310724164544,0.39133941751427936,0.7430516922811655,
 0.9676810617062659,0.087842828380573,0.37131454425019705,0.28307210077013145,
 0.516403712351451,0.8711534460749963,0.14199127635788455,0.40356170513175404,
 0.21943124345374154,0.5299114500194079,0.32860315952792074,0.63988746881407,
 0.24398716694597178,0.9835619801219865,0.45276160277901445,
 0.07577237441140439,0.9966155948420046,0.14230030103410796,0.783227174689289,
 0.36066332780003085,0.7520165512427928,0.40249541789819077,
 0.7984395669474351,0.8415435886646532,0.6964175338004753,0.6644771849369556,
 0.9982263694009397,0.17590406112652734,0.33149879705025775,
 0.32123994363838937,0.18078642714672566,0.19995651226157884,
 0.5262963205624016,0.795705894363519,0.07235914146916567,0.9883732721363301,
 0.1903107703220962,0.1548155893816643,0.18179899989591886,0.665826533742355,
 0.9142080974943231,0.27956386261809285,0.5341263007836938,
 0.34308993552742106,0.48827642527669446,0.7626434696462367,
 0.3186586303760278,0.16561682201203054,0.9768596339902516,0.7478956953557615,
 0.6739627392522776,0.512849373225731,0.27964508442780156,0.7428840209771157,
 0.972008279571482,0.8808614575868251,0.6028636207241092,0.2408249910692115,
 0.1630627391634495,0.7347728939500432,0.0722597384384508,0.6886816331160349,
 0.2949254430045254,0.5079400286862055,0.8920202699721242,0.8743049915595464,
 0.42479148599420613,0.47348203133921973,0.0465217458265762,
 0.0042483362869726315,0.4617439118712464,0.11745334238763827,
 0.48940154359356813,0.7485951680150816,0.969352043880142,0.6653664980059123,
 0.12727283561922442,0.5863483912008578,0.545331572120489,0.26240655111104816,
 0.8937590892998556,0.5778251643726395,0.7359733067674485,0.2317643072742276,
 0.981374483846069,0.7668964705611536,0.8709325893960274,
 0.0013970684341276796,0.6021705872034682,0.842106241648924,
 0.9799129054680218,0.17554925853093617,0.20583734857797853,
 0.7331106675740523,0.462153242126232,0.9914724379519793,0.051114062117895265,
 0.37350186674002805,0.0664972490104273,0.1869353040699121,0.8571128205098302,
 0.28879388295049735,0.9026842501462639,0.5404705302664837,0.293295367182296,
 0.3649161372122952,0.17383781711940716,0.029358715113336653,
 0.6984270538146498,0.8508239420603356,0.6602061560482809,0.15798975797164438,
 0.39882303816298303,0.016050903438776576,0.4921141837428529,
 0.5547513287509368,0.2885563961940769,0.33557635777718275,
 0.47151112785110794,0.02378208589163222,0.2618372323537871,
 0.7436433939475812,0.8663233460496806,0.4447495574590902,0.1773005280495853,
 0.7830618585903554,0.6802397346984186,0.33562714054733067,0.9185605548566514,
 0.015128249050845866,0.02184279988801574,0.014806838620609053,
 0.42041821603152096,0.44339549065800565,0.7393933187230937,0.571292791450226,
 0.989076618617594,0.7203171720678254,0.5423022260988778,0.3940886636287848,
 0.3543230718058741,0.09163545992996147,0.42304194977078835,
 0.03692715683910541,0.40265322671879544,0.6373296580380996,
 0.7462068424520654,0.2760216433283412,0.8866156261151626,0.39800756509724955,
 0.7313688455558003,0.6664978613607662,0.1485528625689041,0.6303852273790811,
 0.4213293043382432,0.3508662821050629,0.7205434069441753,0.5576834938896497,
 0.2035509335118918,0.5189698952862141,0.3113942520071974,0.4485923015449722,
 0.006614889266637225,0.8427460030116732,0.2773813180123267,
 0.27814246758122085,0.1648112831129711,0.6549618098510096,0.507336311520683,
 0.23057134931035617,0.7609205032136085,0.2491915442977608,0.9193073752329781,
 0.043601943869308823,0.3089472857101567,0.05960631300233965,
 0.0704344780703009,0.02144431284124093,0.07811593821496166,
 0.3607520897981118,0.07465992844614622,0.5826113402977859,0.6327171525725256,
 0.8081724956668803,0.5134767269686306,0.20057209839907075,
 0.32648515921978616,0.7707679919731024,0.4424263551994463,0.4538977968464959,
 0.7846892522091939,0.09530709406348892,0.641388285796004,0.9666686050780785,
 0.2875572056926661,0.030410815280555115,0.31800382107570746,
 0.9754551498129856,0.5529218725854765,0.5523921143666656,0.07034174121510539,
 0.5611482659312839,0.3856107277450067,0.07595970474980429,0.5367783178463273,
 0.8159915369889216,0.9731030232229125,0.3913288127364437,0.08418533843764187,
 0.8434760938637097,0.7081690443092318,0.8626895066664264,0.6876227451836301,
 0.4396949205669326,0.44298599795607296,0.5649673515542945,0.895713139214251,
 0.8411872592464529,0.8336087351664722,0.9215752414403658,0.31852141992052685,
 0.7420886826320912,0.8827564871685997,0.6770855119334477,0.8868573469924077,
 0.3491363813260664,0.4015532987956165,0.5935947088669957,0.28738662524346625,
 0.2141327764695613,0.06874825032599619,0.32031736071566164,0.733528331556026,
 0.41431478987604,0.047010755981134356,0.4919531377921804,0.1559607804919525,
 0.5157366330807222,0.6584678035208492,0.3110360329921483,0.23815482443489344,
 0.6913483423667354,0.5672776932762998,0.34613770415311484,0.6864249069526369,
 0.133379303758289,0.9714122951162966,0.050212643375391,0.6590836992755252,
 0.961591597815104,0.8887110032129166,0.5647028701186187,0.18664746221550843,
 0.6545968748160728,0.17212296654075543,0.578665425313766,0.39795736980946905,
 0.02392523099464583,0.8640620703252461,0.022159269231092904,
 0.13844256822724033,0.8108725587210166,0.9089651283488268,0.9654199484043302,
 0.9088592241385722,0.9938227772152086,0.13876630749828123,0.7416189626794909,
 0.37992113104869596,0.74704910901249,0.1966527234544886,0.3799272966369468,
 0.7865358675065428,0.3961188165851689,0.5851206036905051,0.1325771910057979,
 0.5947389550568588,0.36558561004204726,0.4161228173531384,0.3785550823477781,
 0.3357942560131636,0.06676381422408006,0.7210523618474555,
 0.029992601472245695,0.32681993050903324,0.3697057460506957,
 0.6291304487734545,0.6565974553357401,0.8038557588222957,0.98814352042891,
 9.905061781243507e-4,0.8835209319951087,0.8248400657137578,
 0.7127133334837709,0.7332824189533846,0.43708175700754903,
 0.07532600499632347,0.3420822839760931,0.5417636880227381,0.5940261843231789,
 0.8081326663903406,0.6313604025710038,0.5916561894492577,0.9873444342735465,
 0.0954720851336277,0.6637173513568801,0.9377520490997281,0.5645543078085715,
 0.048001087215034266,0.22601455287004757,0.8928261111816121,
 0.8254565433035878,0.7232809336397438,0.7063476165852011,0.5213171683011818,
 0.38359818541348023,0.5220256568777346,0.6239365641447909,
 0.35722284813216176,0.9584061562471349,0.5140594553146727,0.8228182361031166,
 0.3772448488810639,0.9268882378233944,0.12873691192665004,
 0.33002211849243257,0.4583334122830086,0.516420780473457,0.5790098161666491,
 0.08530988852443955,0.9530247005329677,0.29766954540613144,
 0.9999806774811659,0.3722670146135352,0.8977939648248943,0.30601022070269623,
 0.07764422869895626,0.5541429049110953,0.2674298181157416,0.6099720611518582,
 0.025308848741767642,0.5867142417323118,0.7226472365762542,
 0.7116073569703059,0.304369464786739,0.8695840410833405,0.6224325994958759,
 0.21173099432982712,0.35378389445675695,0.5125621445127374,
 0.008862036608606072,0.648004251177841,0.5778664636644507,0.3362342572914412,
 0.9009427393572993,0.7588351717021538,0.9916277732332782,0.8340809845756105,
 0.6294495315754527,0.44369118782756933,0.6415580354649288,0.4068260857621405,
 0.18980325959923416,0.4950093931250915,0.01300221240933952,
 0.06055728830332374,0.7048388812221157,0.4800457750932079,0.7504770361510713,
 0.13160868369123446,0.011270599699835016,0.583617597675294,
 0.055216380546698196,0.18910463630522623,0.25441453876891273,
 0.6454621233304999,0.2913792515080914,0.14727188667346036,
 0.09610527741666286,0.7713284785111562,0.8951321056320664,0.5006411442131801,
 0.5121766069419359,0.28477400341227155,0.7422631209998264,0.375610381719652,
 0.22596625719865593,0.46690060499435604,0.19536822654512176,
 0.9671497094899364,0.2951138122593583,0.19179348084341852,0.3263117553290018,
 0.07435550297599858,0.6055150507111832,0.24593577522117072,
 0.0166323642506625,0.17873366054152373,0.5683353655529759,
 0.010362142928307283,0.48686990258006546,0.5403036405529271,
 0.5120938405436432,0.1580939275991241,0.56210002069873,0.974635197503305,
 0.8616102935557496,0.22436566953191694,0.42633691666851836,
 0.26854310654567626,0.4263731114778413,0.31728499693168477,
 0.6033630193586428,0.7469718440636355,0.9713188869793101,0.6505084921380868,
 0.4045978421984352,0.8138810114256411,0.16337554575892055,0.5366582660130244,
 0.22692335089242943,0.14652277470824693,0.014756397581124236,
 0.5825784158380313,0.9639624895082212,0.17203590629993137,
 0.36884092867103324,0.9166241555016446,0.3021424576922356,0.6322081847567308,
 0.6151616429731738,0.736840979597107,0.8761715973810945,0.0744582559949537,
 0.03439909204171576,0.9463558226630867,0.3823152353492141,0.2582788290954078,
 0.8314904999302505,0.12185315703353328,0.762512059051081,
 0.018522304598455386,0.026845512993766696,0.5135015716413199,
 0.1490328667395695,0.48183905170066044,0.3000777815529664,0.8963774833951069,
 0.10338164222620905,0.10509108315640081,0.41545936295354235,
 0.42878354373025573,0.9280708048223885,0.6379798714317203,
 0.15282685605945168,0.9495138711426822,0.2592267039782373,
 0.029505429035849406,0.1408297413464228,0.536811486181281,0.7191021836058403,
 0.2857345519610621,0.8289079016490064,0.5876691598231081,0.24073275005636374,
 0.4334228254777197,0.5555762094817462,0.8080316865554745,0.8573503638045593,
 0.11442713964924145,0.6002785222219692,0.43828088590354364,
 0.6449845590368096,0.7295989321691561,0.9870055120763916,0.5100173662435361,
 0.6570636647366007,0.8821926366229064,0.8873714921457075,0.34143510555970946,
 0.04732499585999472,0.9825329270438498,0.9246441964367083,0.5793701200122379,
 0.7823812265880871,0.9279681792966059,0.40727500002762285,0.2522239293150741,
 0.4032400240999445,0.640625823085134,0.8809189998265614,0.4837007232470808,
 0.8690329544179183,0.055752040040937145,0.015441761254700648,
 0.6190697419454866,0.46286813847264663,0.5930020308846736,
 0.45484050591864733,0.9014991608421528,0.9219852710901207,
 0.15656415421534564,0.22894815324977524,0.6782055779870727,
 0.02995910808510993,0.23464299555384227,0.9567586503452387,
 0.6677257679023652,0.8850157271114446,0.2671899517785242,0.7251211430407354,
 0.9452054681138712,0.8005022862330068,0.16878194262805013,0.1618994448719817,
 0.9675517866766076,0.6556179167354917,0.922782064790946,0.8097449243518742,
 0.05810312033524623,0.19642334772730174,0.11241283901691224,
 0.8286407316913299,0.9521812003903021,0.4709348176312267,0.0319579554209003,
 0.529853743700083,0.3455935665849904,0.05124649318787444,
 0.050954994402157006,0.8493993578542031,0.5229775073422779,
 0.8759041041914205,0.5446218105140576,0.5546002421507604,0.2660401109763699,
 0.5549537529166433,0.6609734013325246,0.4335349217427322,0.5918419152987295,
 0.4247792698523448,0.3925982745238774,0.9780859971198695,0.8391369098426885,
 0.12568334257628677,0.43483414230025086,0.8306495865686147,
 0.6696424344236218,0.5059853257615048,0.060896083279150304,
 0.6589326822241107,0.2698467630686604,0.2434946901755458,
 0.003372914452552167,0.5460557313389411,0.9696076102277297,
 0.5725283059561324,0.3267366294418733,0.5766431362430332,0.4212114841441671,
 0.5984446197134806,0.8773601512736247,0.6465379362650672,0.19894433570273162,
 0.36701995642198715,0.12833479877898502,0.39766413982589655,
 0.10141369940705303,0.4218334157421424,0.04231241340640646,
 0.9562529473154686,0.6819690360555473,0.04868792229448504,
 0.20950900895762503,0.3464181378123933,0.3545661696153095,0.9670743637503251,
 0.017733127839505758,0.9176438027847005,0.6184022705425218,
 0.28477013604459034,0.5938578996961716,0.3301898432894632,
 0.16861860320120226,0.9097388559133358,0.33095227993231036,
 0.07185050542171867,0.7377573744109266,0.16693582874855728,
 0.8926780949185241,0.6349012055744107,0.238495313635935,0.4983520135181878,
 0.17156019047044657,0.8041826848904233,0.15885664711168512,
 0.8587214267546668,0.8500349480490808,0.3008851897756014,0.5095389903723682,
 0.4698871267049759,0.48139502522603483,0.4714079042540139,
 0.19733744471483128,0.16780306912025977,0.7786301827381561,
 0.3760323096712529,0.21438069236194313,0.05235065577999587,
 0.9962697673204426,0.6878951771193831,0.17935223220907548,0.5099088731812761,
 0.17667869535171532,0.49913033165106124,0.7796146257500656,
 0.47509672988983687,0.6874456153013535,0.5907144108121927,0.5603729042819187,
 0.4410664265954939,0.014519282888019958,0.1913093595373624,
 0.5214875093421749,0.9522650528888287,0.3725767513850775,0.3488130774151159,
 0.09169248847523048,0.37988555846214644,0.576944388825983,0.4382637823232155,
 0.14150977195603942,0.6073537643473088,0.039473631720373326,
 0.9259319607552442,0.3981310982395465,0.16023369334488258,
 0.37593282166243136,0.7447481576239283,0.6911540595634958,
 0.42142877430676273,0.2696971030610168,0.81816531025941,0.8425181818027629,
 0.39055909677558887,0.2248145952866194,0.9122251014020978,0.6913405855794069,
 0.24436473061864372,0.13377315562483671,0.28196750027770434,
 0.9191981853223689,0.3634392134806278,0.45806145370996776,0.6774431003912222,
 0.30135492129637087,0.35754691580372333,0.4974045089202125,
 0.2740513657270832,0.11579209651682254,0.3800882085453814,0.7105135462330676,
 0.9394949747311737,0.14529708968076627,0.09941714145806002,0.652293083886537,
 0.1102525892062971,0.7344992203712307,0.5611474508382577,0.4078042622882333,
 0.29849835546732717,0.6079078220529136,0.29372136115109315,0.330907586510263,
 0.4521014781824121,0.20923450145827138,0.24647557777759943,
 0.07549946612160496,0.7337128185331752,0.02658281978517607,
 0.4892768135566422,0.4976800688357077,0.14021490102713208,0.5141962459304276,
 0.16322099581167604,0.3892882333281873,0.9065422018940571,
 0.09283031038897205,0.8286421128675863,0.9167142770809247,0.5501629099252764,
 0.3182985046585042,0.10280873851038463,0.5494076705370962,0.4059625696174398,
 0.8719847132826665,0.11604402439401751,0.9458184154833787,0.4416204353286882,
 0.3456870766852469,0.07767141323940574,0.3424347423254611,0.9341160264069919,
 0.5225325930937396,0.10602048337100223,0.47831573054417875,
 0.34332874642465705,0.21190768600051024,0.16428309476549785,
 0.34504691091995665,0.8866005484962995,0.10024040042204929,
 0.6124043609069332,0.818334584585466,0.12913000541300146,0.5090948936889486,
 0.35536553781735003,0.5585963873715695,0.2825776503876204,
 0.03399497173046462,0.9855259663788458,0.16566474257016184,
 0.7047044319016859,0.9638530161278995,0.7363308771106688,0.19286676349320175,
 0.6608378291021357,0.10599297921910222,0.6943875261869958,0.506993604246133,
 0.25316957093087733,0.028669866781609876,0.5823832827298588,
 0.1383456763965072,0.6047562462241768,0.23159969368051525,0.3386080078734963,
 0.11089434428840805,0.15632549496712234,0.9971752396495497,
 0.6694282762431694,0.4690738958005548,0.1650341843866152,0.5101585412597296,
 0.5700039485255592,0.6329622527471528,0.41514013831163,0.8821807438881668,
 0.9633986206700008,0.7178156640144474,0.6098494960253723,0.1409482454911336,
 0.4849576525717305,0.31927045879433225,0.23167566171456777,
 0.07953760130679899,0.7821737304590222,0.07583409071129865,
 0.3959223739874199,0.35045686609885274,0.23043711889980867,
 0.9399596746713107,0.29986736709884565,0.11054815839746013,
 0.6305110218039003,0.5125714327447909,0.84805477192117,0.7869297788496711,
 0.05247868193694427,0.14934967444504477,0.43412086953491924,
 0.43211678168831735,0.13201319430520586,0.9683627549219613,
 0.7761478826383965,0.012859882946944845,0.8990229358453774,
 0.07467805765477864,0.3806017420314166,0.08655197484978183,
 0.8945557223003973,0.5402147117294589,0.044719224419387915,
 0.6604305176136556,0.4773545480027961,0.9429744825240932,0.3579416207912888,
 0.9827205872966891,0.6096937703152818,0.13959759269326477,0.2293957198548564,
 0.8609667506116698,0.8872535552011608,0.2654815348083498,0.7165785881761474,
 0.38917236907314334,0.11063827636999202,0.930126963608332,0.6755323094217801,
 0.23021580825182142,0.6106329739545151,0.8524499173934932,
 0.46555935381793156,0.561050952502912,0.7473093695823253,0.6551330571721138,
 0.16837764977552394,0.44326129177317997,0.818457277930817,0.9636155050988828,
 0.7573617122347147,0.6822116071389116,0.5578977113490546,0.7786012760242506,
 0.9571568422563692,0.10428099437104632,0.5712667423403546,0.2889670490285191,
 0.8995059851855419,0.16323200552703288,0.347163360812042,0.5326310767256143,
 0.2331123531004905,0.6753931704382989,0.8250370156635976,0.862002637180524,
 0.35541365568515215,0.2564393216136389,0.33770125153986785,
 0.6161099962971547,0.7176092400496585,0.8476057187744234,
 0.059293690733380644,0.7043895127320183,0.5039280881069897,
 0.7728405168273121,0.41716742157022424,0.5214537140689542,0.7545225567187864,
 0.8348431222520584,0.38603641553349655,0.08346871013319268,
 0.49519526654246726,0.43845449888757715,0.23247473451498857,
 0.729721383825308,0.9011271130653716,0.8802648040113874,0.04388989584333558,
 0.12114762318635552,0.5599671657825368,0.04637577523149661,
 0.2365503295610314,0.8091968829933037,0.8647065526834756,0.7298777051696915,
 0.30494003401649095,0.1546828929127143,0.9908799164274062,
 0.45331133513301236,0.9532037584563082,0.9291012355630208,0.7590768863916127,
 0.5204519267847287,0.18118640013141418,0.42472725757676555,
 0.15314292735551638,0.19750659962568207,0.402358008548085,0.8100463834729482,
 0.9114254522136808,0.2799177211526497,0.683429580143958,0.4879022029003489,
 0.38464109315756223,0.12614139574499128,0.7768073995276596,
 0.6875711650724469,0.09517912700618292,0.727052053261901,0.3071032292482119,
 0.5333230259868893,0.4348922875354715,0.9368228083995782,0.9006917925866276,
 0.33889144036926744,0.3762589334746891,0.6765085838515017,0.4525256808262348,
 0.9063802391099511,0.011223749956213558,0.46812385499039744,
 0.7092367551710863,0.8005873877763554,0.795033317499935,0.5984755928296976,
 0.3441756254723217,0.49610041253863546,0.926160404725506,0.869756708500933,
 0.26725546212942364,0.4077231781617322,0.355111368115572,0.4289660581784551,
 0.9529082782198763,0.4237502062094345,0.5968177897253188,0.3176102088745715,
 0.11270642473173542,0.4607194419046712,0.35654085026295634,
 0.2961707014670508,0.5154492088536219,0.4270221679560059,0.5614081890320137,
 0.10449773732734391,0.7018024019713838,0.2084013625397132,
 0.21524841210567613,0.4132984650706508,0.040559430466036495,
 0.7687340166782859,0.7240336522478552,0.5143343883088857,0.1750806359759285,
 0.799132632337736,0.20771275735263195,0.8352561265836062,0.5921748027014184,
 0.20998703017771136,0.6544797984547651,0.4235621143787748,
 0.03247953577938589,0.8952319131318269,0.6137858628173872,0.3256336354517885,
 0.8887579588218085,0.5571663713363275,0.6569156115046064,0.4323833278551532,
 0.36836452172855816,0.004249661667221294,0.8907809553879815,
 0.1662272515131007,0.04447543032531731,0.8485598625003191,0.9530966094979489,
 0.4503610778414605,0.044707748290979454,0.962216341177411,
 0.13713258750362434,0.8038729327432923,0.7737000652263822,0.6407824081571214,
 0.4074418618030471,0.035853377801325914,0.2982527893034653,
 0.9908249226977789,0.9038043481956082,0.9593298121011763,0.24481360519012263,
 0.7023765578898715,0.839927758987936,0.43294291811486296,0.5850298020541269,
 0.5292576624610886,0.4894140045704586,0.9024572999499465,0.1835529286270744,
 0.0626591082144693,0.10152253008514756,0.01565227459379781,
 0.37155159233691226,0.2619437982276662,0.3621786669472773,
 0.44400079859075003,0.21840067274391828,0.9430139170982447,
 0.8574849269758753,0.3587201210924922,0.0024037564722909632,
 0.659985596941755,0.916534201315748,0.2874861072653736,0.7430225218043759,
 0.865559453630292,0.28332263780127176,0.7553617947333915,0.6466295121654573,
 0.244285959579974,0.6619378166546824,0.9638257423425268,0.4872864088144908,
 0.1868965408784975,0.6285250884801409,0.6913492121541893,0.08842737086427954,
 0.1522678030923672,0.43336212886287806,0.9530185374790245,0.8064004548544933,
 0.12462375833974193,0.685437411706724,0.17919567338061992,0.7745443336452905,
 0.9224947977081271,0.9522642455310321,0.07768896109917744,0.7592734575699581,
 0.6975902543643444,0.2322323653315892,0.2747849411861797,0.4808964067205025,
 0.2209673233229883,0.37062714233646044,0.9689274643215493,0.39494481437662,
 0.9224813833738379,0.21285453955123013,0.7882279129794494,
 0.27327826989001125,0.2765722197599617,0.7906730482310151,
 0.42793467193036183,0.2922230539637415,0.3736587613935959,0.5391812003490728,
 0.23316386338460626,0.8976977603710037,0.22156080076372808,
 0.5936717790135091,0.23114665413919644,0.4703912362189355,0.6015524640841123,
 0.7780766866580199,0.00605413252455389,0.10652519620056322,
 0.36194192040695894,0.7839623933158575,0.11596956427439054,
 0.11355156610305661,0.570130041707679,0.2739166688422071,0.9422591767329884,
 0.5291236805653569,0.06529702218393052,0.12261757989539923,
 0.9432018549737787,0.5133105243108019,0.8749998510843221,
 0.056177017323742806,0.0844984986947428,0.8881052879783765,
 0.6429738083569209,0.5943293706292461,0.9813099525616003,0.6227529445751216,
 0.6586771582892481,0.07321952060245385,0.17786433193723117,
 0.9642848970255914,0.8187163639054135,0.4334461540588216,0.1597924997914939,
 0.16624943181104102,0.6918670369149872,0.5061525013376913,0.7592516935002486,
 0.988905841882501,0.3534658545581383,0.6112337688511361,0.5053425693453694,
 0.9675085255008723,0.22003245364240087,0.27015433860213567,
 0.31527902644499384,0.28076842809564884,0.8545129389967283,
 0.4172466046157509,0.23988055260892827,0.09579987686104463,
 0.39401268742188356,0.9855883298166972,0.5065788635217288,
 0.34124458942538816,0.30998632219019706,0.48736753320960335,
 0.22401261851033583,0.4095367119036709,0.6922172029358843,
 0.09657085420306011,0.05881424891122233,0.6139301900276449,
 0.04024454945651934,0.8042507288071512,0.7277731436876999,0.6203264500330097,
 0.9920013340419671,0.8463684558573056,0.3850217054563684,0.9188311869926962,
 0.08927632475924119,0.12195881687796628,0.210881437143156,0.7838266023644219,
 0.26587694356912606,0.5841725828597694,0.5535519427092183,
 0.12041217142680383,0.6795465330587721,0.20225612676516258,
 0.6707879337893017,0.572791415468579,0.29137401932338525,0.2716894781986182,
 0.8257410104466115,0.3262295745214051,0.05453060599687898,0.9573888250179331,
 0.4452372070932209,0.218853044334119,0.8356795949640381,0.3876090521755208,
 0.41688129841275756,0.4618106002441673,0.8497461871657003,0.5083272222748957,
 0.4429933296074985,0.30826485151008,0.9960224235438238,0.583881628698088,
 0.05234411127391969,0.16966844924508695,0.08022204482305872,
 0.5844132284299933,0.7057306814132891,0.6595697406044805,0.10505411987443702,
 0.3461016518331246,0.3318467398527154,0.8359812987201234,0.4344260948520393,
 0.16070177527914475,0.4921952464300202,0.014534977472734356,
 0.008790358371040363,0.9096970275459461,0.19214597182774662,
 0.4516376985687981,0.8534935895540228,0.06405198361258657,
 0.11250160027017442,0.265357331921826,0.016353810272217828,
 0.9078144659938213,0.1915008615070951,0.9853710704264446,0.9702975232913026,
 0.13344272289538128,0.7176920126022172,0.7221541670845649,0.5767413422938896,
 0.2448809521272879,0.4116861603819473,0.31108273520550755,
 0.07680966677118184,0.6376953149829804,0.37398037822254215,
 0.44254215328022584,0.23492373944655076,0.9630658625964349,
 0.5044399871971931,0.02754237390066283,0.15416857903943193,0.510458203209689,
 0.3239850863445718,0.16352350059029086,0.8563130537845299,0.7938272000439713,
 0.6173629480281723,0.6505686035753004,0.3821516396350092,0.8569213429347845,
 0.33167709780984034,0.2914447017133711,0.3761863908239459,0.3034066890894409,
 0.10512354067048535,0.2905965596741692,0.3578195684140747,0.6464759579547517,
 0.46331213943309524,0.5477065730945376,0.5287462668929575,0.8516203874606332,
 0.3360978692670098,0.14489346637316247,0.9028778085781801,0.179035982681221,
 0.162378443154492,0.4642098336981003,0.574329063170888,0.39029739311633294,
 0.13459171873526754,0.9347345702480495,0.8660228511913621,
 0.04606857790327534,0.05933665060257565,0.48790720877102856,
 0.5179032774422642,0.4279865401786028,0.05497191116951783,0.6285644922005484,
 0.6626525885594098,0.3315176616113821,0.547550088186265,0.30781076402436536,
 0.8142780665134912,0.3586741353745533,0.2592646059222,0.47318392809174925,
 0.27878626237034454,0.3867666778802006,0.4275461170163599,0.9155564698242699,
 0.3929056398055877,0.7652324187657829,0.7361370805145416,0.6110488765997152,
 0.06506847806337135,0.17798489499339654,0.45093518294920554,
 0.5199697193435543,0.3278939831554223,0.30926595897363784,0.8124717701873156,
 0.4580213758541979,0.9215817321621513,0.08164779005921308,0.7142056265336952,
 0.2232743408999176,0.9956725882469787,0.9691822759775768,0.670970500307579,
 0.6648925692768266,0.4625780863360873,0.1961738695410994,0.456063275904974,
 0.8692097971368568,0.9811182825657153,0.10541652198864226,0.9173519872059095,
 0.12472201837151375,0.760495327221431,0.48502053688638036,0.6425156413727289,
 0.2781071561288393,0.39408781104666835,0.8466930876596295,0.5299233718405625,
 0.9236852058384173,0.5948235055641747,0.32323638719055026,
 0.07984172508672038,0.011181757557231942,0.7468937737338129,
 0.7022806122118866,0.19578719151716661,0.7352155735505085,0.1596171647287915,
 0.11533883261807942,0.9700972441888145,0.10659871161302903,
 0.6887255987916938,0.5677679578003177,0.1034333574466868,0.1086449012591203,
 0.22719929419078055,0.04819567670495317,0.4803817471067625,
 0.6443178084061587,0.6296936579501891,0.0015898933678253346,
 0.5821525800681423,0.48638945214088747,0.4106791724173413,
 0.44255808648686823,0.37783762988620706,0.36084522683690745,
 0.9101061819997806,0.010722733238179627,0.17825852128206465,
 0.17656363994414315,0.3798327216511489,0.016047423243943815,
 0.34548353016164546,0.34683546773722096,0.09971640563617246,
 0.1317389480296841,0.4334316337805567,0.6572552075421796,0.9421034914219582,
 0.6519628687559996,0.5874792255191674,0.2817328061804236,0.7752685321065145,
 0.4252870924860308,0.7283543397618418,0.8405436646470195,0.3887542047383986,
 0.29653912651473946,0.046569991004664635,0.5236910019957504,
 0.43559062863966913,0.9440636899902157,0.35373691785190464,
 0.3571854407437166,0.3424866034570079,0.15801696175638724,0.5677414000965268,
 0.57543960996755,0.9761545484586793,0.7200807443699746,0.2994867874303657,
 0.4641841733840073,0.8415810417349616,0.6778232238489497,0.7294657399697307,
 0.7402466642242163,0.8253537975975196,0.949163145871781,0.9921154711779067,
 0.028502098598540804,0.7813237291247299,0.8030351694567286,
 0.010174610658089556,0.6678743727934221,0.25771058148020975,
 0.251723052714805,0.8823657122731483,0.9538092777821344,0.14861706658798313,
 0.4733651539580863,0.5678469267089203,0.40575713306592576,
 0.40316287627225367,0.35673637287571314,0.6509298710466074,
 0.3782867878768281,0.03492022008183138,0.49866674920138665,
 0.38466389732903905,0.9236326487394138,0.07737981786687986,
 0.3203316547748586,0.9636080120541459,0.5391246196859651,0.38233363959197486,
 0.4618816403074343,0.14653434176481184,0.3111095315371266,0.5931024881100672,
 0.5122671161863906,0.9035890975073809,0.046971577486628946,
 0.43107759804289625,0.5069257835888957,0.7906409225718347,0.2859086171577244,
 0.8672358732552343,0.1367652211200322,0.9499342536355012,0.8695989117747377,
 0.9967222432641538,0.3537391421863578,0.40019006889610764,0.9489585633423008,
 0.22544719132229685,0.7676862034559647,0.8669942993414118,
 0.08116792322277844,0.5414651546728271,0.7853103694772812,
 0.26471065391324977,0.1594868948521242,0.6865000780232785,
 0.026453199635536162,0.18618374134223015,0.31538419576679966,
 0.02986835103673413,0.05825526589143293,0.2997416763993487,
 0.24856962758216383,0.8694206997975774,0.8246053129148022,0.9453717625615922,
 0.1915071007111826,0.5935700028921621,0.6503843057603285,0.4722827770760022,
 0.5005954625981628,0.4245559486031679,0.9427776775197519,0.08437177989515854,
 0.910703371053452,0.9791575228851072,0.09712254599499848,0.2994164351692572,
 0.8886351060593711,0.7382258918900806,0.7172345848244558,0.2199908555895853,
 0.7108832811268943,0.33503606476968484,0.08937930892988688,
 0.2096038653713449,0.14571216367768658,0.6724831021276215,0.8433434809518372,
 0.015350172523280303,0.3734007064336541,0.04251943081034271,
 0.431780593583858,0.5842260855455479,0.006685920463016415,0.5647158264516297,
 0.42879477138364974,0.8596741373044762,0.9139182859141505,
 0.32013352292659103,0.921841349329169,0.9108630109027959,0.06200199907355319,
 0.25790363171504627,0.29250901669944285,0.5976795721240376,
 0.15897808193103147,0.4267290477979033,0.2870319905333838,0.7010175197415848,
 0.9062374645953624,0.12293659689689174,0.603094563933665,0.3450645328702,
 0.023584086563147255,0.004629021771127673,0.2786193487032659,
 0.6072389126839309,0.7696946164367493,0.14325087459461705,0.4612616867900021,
 0.5340850131747017,0.23636078476373934,0.08363420076078831,
 0.9348156306032435,0.6176286932913341,0.22620003764050522,
 0.07060827496863598,0.2765174329648603,0.234730816513242,0.06730846401353219,
 0.429887159979197,0.5629403629650993,0.24872539200085675,0.48069796472122195,
 0.05026857882470759,0.7910675627481683,0.30574421747671754,
 0.1444996457681884,0.6237560246298602,0.05827453987104503,0.9404061536833338,
 0.579150521050042,0.941237449152734,0.606461585607361,0.7316241656590725,
 0.8523069474721873,0.30290879878964416,0.9920211694226928,0.5084280136631832,
 0.8195257235275222,0.3905747963412708,0.007160890461740843,
 0.6836819058145608,0.12126757060870463,0.6380344453930142,0.6762491954592058,
 0.9022076323075906,0.9923827210071117,0.29263492338378094,0.5129814255986997,
 0.20729510034698018,0.7389916224589266,0.10289336232553425,
 0.6709035287120133,0.2565413732666424,0.974837231165101,0.9911176677848663,
 0.4959696148550601,0.8338267905199523,0.7504478530770045,0.31629735240907586,
 0.8730561988225725,0.6740458269151579,0.23526734012225248,
 0.39933118544483237,0.15367267484954272,0.049484777538363445,
 0.8599225202222194,0.07140347041421635,0.9947275886230115,0.2156708357492929,
 0.973988439970787,0.3892148322675859,0.5896578691785224,0.20699142860766062,
 0.9860554764197579,0.34897861473008895,0.6779127513044049,0.811017401619329,
 0.07228778670859848,0.6605257357932538,0.5772398221617724,0.9468528811171563,
 0.1534140811080965,0.2900471953731577,0.9158852594963705,0.9479681349775977,
 0.8375923362523513,0.4629284296562992,0.5224442254399707,0.07766249741954656,
 0.843138028441487,0.4216062459540598,0.35233198353955364,0.41786813915340326,
 0.7683255968249625,0.36543489760847025,0.3433273198556781,0.6689578171389327,
 0.6232696075843893,0.7192644690228763,0.2695326062066903,0.3142578664241398,
 0.26182450044335914,0.00475324180336989,0.021552029225753055,
 0.08585418977226178,0.5649603849516323,0.3430030995244755,
 0.16328425469911945,0.47504177991512186,0.10077879864600092,
 0.46719856832644036,0.39018435570770227,0.10968321580373552,
 0.5683521610710187,0.7662637622012628,0.4002269157667119,0.9674180972020586,
 0.8917441871620736,0.03062411570076784,0.3481172435166847,0.3646736410498008,
 0.6226986827475338,0.8651240306726128,0.7258631856455424,0.2728122381267477,
 0.4937706290963746,0.4539857895584052,0.9547289481781807,0.5453381191342026,
 0.8513855522983154,0.9144364051961312,0.06597807457819527,0.6724421792682357,
 0.288463058810976,0.8887199629810012,0.6294847042783532,0.7324733942916622,
 0.22652225362608047,0.8084633125275109,0.7908639055896058,0.4203520471828349,
 0.6750194183519678,0.03656646535880215,0.5075784670071344,
 0.11615642709527829,0.620693474172462,0.8240821836176366,0.755990378946064,
 0.16550105028315898,0.5303719182779398,0.712311616917854,0.7773585082789328,
 0.12690549441949528,0.5985171385197209,0.08386699186907698,
 0.6123411183906164,0.7989579431297417,0.11827908192896952,0.4118430389084575,
 0.8212149539415667,0.6352200595296736,0.2681232550708498,0.9894047927288145,
 0.9221487693662567,0.968543692456697,0.34391389814383266,0.14708781862000953,
 0.7067300578878657,0.8034096995740574,0.7743279366843017,0.8661269095503397,
 0.1244163820748011,0.2750143894880336,0.3657137802010001,0.23355024066862984,
 0.4067020791783338,0.09539481516288006,0.10921049990158382,
 0.9303879021638437,0.5318492700013703,0.6096637646693961,0.7902183276932222,
 0.05690860683370702,0.5423422697003832,0.749228010573159,0.8030721795826248,
 0.8002374999293502,0.8130056432037536,0.1905885383549336,0.10251497289479516,
 0.3558130222297624,0.12215268005236235,0.7343917794849971,
 0.42518943745261906,0.33858595008568926,0.03522393843931393,
 0.8686486503694569,0.0411899828773008,0.8261311461682277,0.2261121540644413,
 0.547070903868397,0.39943512837135586,0.2423119418303249,0.1892171811287512,
 0.48324614354180684,0.4719865193086611,0.36973053599100003,
 0.27085632285186567,0.9043919009001951,0.89882299356452,0.7095094513617966,
 0.45466109251199527,0.34858036454147523,0.8739995933417528,
 0.7587721945288554,0.013347985669498352,0.5523050681855327,
 0.8141078694068736,0.2549521094475391,0.9502951245937914,0.32495450638749435,
 0.5206345681669011,0.7130859716291114,0.7971223028422666,0.6007051809936839,
 0.545118761820442,0.8772046243007161,0.4460602611277654,0.5385723785793699,
 0.7773158542499998,0.5017548590479676,0.5056208103050622,0.9791382437180534,
 0.42297516220790654,0.7754784970569902,0.8309967745930118,0.1786235791142362,
 0.23696546200785185,0.5609672698341617,0.2654377307387866,0.3743088877158839,
 0.22533649305770764,0.9057119154639456,0.7115842061292339,
 0.05397924330366399,0.7617316330630572,0.7295309314666326,0.7515746773860195,
 0.6976539064604818,0.11076568883553073,0.17240008515600214,
 0.8041381881407932,0.3485561841505913,0.360086013723371,0.9387223469388952,
 0.2036120052153796,0.14915559586921256,0.4631445261649183,0.3880518954699145,
 0.8316009716471197,0.6156029838892586,0.5148194067504459,0.6634757528477113,
 0.8456593209902179,0.26913592690170485,0.2613919456671223,0.7848044299614849,
 0.3590158249570983,0.7853075695237225,0.5096672625243924,0.776123205661546,
 0.24438886328390985,0.6776302969833823,0.33747469387058915,
 0.9284378467062855,0.07756102291691436,0.2672217033244263,
 0.0024812031531832313,0.818983159636973,0.4648540463226736,
 0.7751638372152079,0.18409174247007232,0.2775737483813028,0.293891210500128,
 0.8245797442630968,0.10333514719727577,0.5447110228254752,0.9681196940404492,
 0.9680749111010728,0.8666809583573668,0.24236161506329534,
 0.007506703051842201,0.32426053568941104,0.09592728276983753,
 0.8871847289494434,0.11124784226598805,0.3772388094967156,0.9544722305042299,
 0.9335044823507683,0.4978086488438421,0.025753966847550935,
 0.35294858742660884,0.6314913931897116,0.3599721644309437,0.9653408635007171,
 0.25766056233610846,0.9287921193187656,0.0871950882239858,0.5485111903317964,
 0.7952156624667027,0.6780824428223012,0.00789798794243024,
 0.23792518246696615,0.1850860332771379,0.8634090148460638,0.9342167183615641,
 0.3704495237569797,0.6603996634262637,0.3749808793797309,0.8735072613567825,
 0.9900824670392077,0.8262085619988087,0.5442876819261536,0.04061594506468713,
 0.3568124155636657,0.5770481524053457,0.9280428611506959,0.9595845954307038,
 0.4458966826911468,0.3513975264947019,0.5928800476166872,0.9511714237495636,
 0.2099715078724682,0.38325018759017204,0.06788506064012823,
 0.10692430016431542,0.860051082063788,0.13214674739439403,0.1418359540136218,
 0.29269864735787277,0.4085116499970448,0.3739048339879215,0.4802429980340538,
 0.7918677361400112,0.43652378401898273,0.3224421415831512,
 0.32270690522354295,0.7406757032305531,0.9999223486592151,0.6544126830468027,
 0.7793961851057637,0.01851010041716028,0.8704262914592994,0.2249953149191496,
 0.73541331212699,0.15336529693183898,0.4519937410645284,0.85795333117408,
 0.36184735662271017,0.7948157190079809,0.0543231716909075,0.3693579611919843,
 0.5251323854066051,0.6605237614544532,0.3544871869795936,0.19012072149742898,
 0.15771851444992602,0.3734864148020225,0.15058452394086053,
 0.7307394091732197,0.8102745541257277,0.2901463940070561,0.9856877039352884,
 0.6592669516065026,0.268445501873005,0.8151550810807132,0.6383957847819398,
 0.2515006233271381,0.8290169933848938,0.7674803680581305,0.8693520925339319,
 0.17708112509504026,0.43012349410970074,0.21809369497785658,
 0.624258816354967,0.6203327623781021,0.11529594620792993,0.2186936438242949,
 0.19922512456032793,0.8120881309984447,0.20284109213636703,
 0.5970121463153651,0.8981942534842375,0.23013813665322813,0.5784416486300354,
 0.9149050506846359,0.4241956448440021,0.7093008652280284,0.9152556007108006,
 0.6323499246452986,0.5740089592026218,0.17331609452530594,
 0.04632804875837182,0.983008078692053,0.060818799297403525,
 0.8846864776381642,0.8197877712069996,0.8261723525113593,0.30126036091937847,
 0.8391900459480496,0.5485689856647841,0.21357087432472255,
 0.20654384056879016,0.9041179664126487,0.0282384085502565,0.6975211418853623,
 0.7580486127123592,0.9467260576318586,0.5368369830404877,0.10160385404346939,
 0.6428089488378989,0.9856006736722094,0.7980313723788708,0.6336847115498999,
 0.9564357668708731,0.599478322128803,0.04409462164300737,0.13770177445090526,
 0.6361297010366945,0.5798887781635136,0.09039051128284192,0.9461790261463168,
 0.14647452119633275,0.20728505147010368,0.6832179073776752,
 0.6530290489314359,0.3959139243992009,0.6883770979739338,0.6079227346552126,
 0.32958863598235044,0.18596247848262426,0.5433027983123198,
 0.8704439604020735,0.5805724691690773,0.6280177296624773,0.6501190990445799,
 0.39558661594378364,0.38476557344011364,0.538939487179385,
 0.44498815178761997,0.5886917438272617,0.14309089559458366,
 0.006675725579559533,0.12371258898125936,0.9455717242328725,
 0.7208127366273825,0.5208219000410491,0.8952018182201925,0.5257192932127632,
 0.3519412317249633,0.39570377436404436,0.4100844993319682,0.9865892820424844,
 0.7694338472205074,0.5164347728874517,0.5324704944054032,0.9884419122178575,
 0.6166892753500779,0.9607720180805285,0.5319714841119477,0.8127439379309245,
 0.3808935020416382,0.7077868075171394,0.1770857116448784,0.32600938147473935,
 0.1064705140590001,0.8994707278880112,0.8419207205129299,0.5956012835615906,
 0.9921485974399387,0.2566661810473452,0.17061507745188353,
 0.036842751698664356,0.9779367765036578,0.2855041492029937,
 0.04067814495266675,0.0846069209954724,0.9558149561379923,
 0.42215085352770954,0.0060914820478743525,0.3212044630693588,
 0.5430742513672207,0.11311165372988141,0.2985705230615353,0.6529263994224226,
 0.27655696661599893,0.5362023603550122,0.9594640883953761,0.8084419722928924,
 0.17730773755321694,0.44263403060278117,0.8751068859880861,
 0.17684838543061143,0.4994498717690019,0.009902727494793373,
 0.47698919657446737,0.41537053360674736,0.9667343833053365,
 0.4097573501241213,0.8402112041849183,0.47684876631443607,0.5161436561603345,
 0.8772057861779421,0.19192777002988015,0.027956608487627266,
 0.7929645173017525,0.08495588417994204,0.4516811745963313,
 0.41105239391718873,0.6992435100325309,0.752100360914595,0.12484096927922839,
 0.7001295446710349,0.2601627928454018,0.9665016159958971,0.4284842564461955,
 0.6512105328208763,0.5068599919867701,0.29961083161548363,0.9529356723067208,
 0.5468293146582179,0.3088332200634263,0.14512972771504185,
 0.19433687504757502,0.2575887099089954,0.9219557009707395,0.7654980243494367,
 0.45660956835663236,0.35580955399281966,0.7318941790797548,0.883919026830994,
 0.11477514311976222,0.18621618757821978,0.9965991327522618,
 0.4805117330413873,0.4401467321277224,0.8514397185623411,0.36711826714438067,
 0.9839914931432551,0.6286409369930506,0.11083797177010479,0.7287988563687346,
 0.5774408702562297,0.8993850129446583,0.9406940475285095,0.5196494914089825,
 0.24459816889259078,0.753110402305073,0.778684273309237,0.9772210690195113,
 0.6511834303773548,0.558397122838543,0.9703716417859924,0.9945449493776937,
 0.9937820371605279,0.7746937088562256,0.07799188908371901,0.6156191616834574,
 0.36945245455217746,0.4249758873651326,0.9329126801442762,0.433887273347324,
 0.21206425704455956,0.2997619262247555,0.09253317842495912,
 0.2248223795658172,0.834199294433464,0.9073030590610024,0.9454992067372237,
 0.987300520678868,0.3565891029555427,0.8558228198498821,0.988106782838595,
 0.3142422775578403,0.4009255403334461,0.853810409329795,0.11421563437796123,
 0.7357701658293657,0.8318541330476664,0.1380585759229116,0.5332174382436997,
 0.400896366623014,0.7391850367754803,0.25844733265919473,0.10279887339374572,
 0.5978942275951411,0.8993613904635022,0.8618301276968298,0.37878181632498453,
 0.8930923133071565,0.011718892099324751,0.9914850893309651,
 0.40123799500810853,0.4257792363249666,0.9727818687081735,0.8155462068275569,
 0.8440179155256681,0.43375454136309255,0.6790040052750278,0.6372992588603716,
 0.971056873236914,0.3519057996469296,0.37926979611417033,0.5157585483658252,
 0.1684388553906303,0.37643432793591813,0.563130082520573,0.6832986803107721,
 0.811419763387101,0.5548481944756029,0.07985614552793507,0.3308237927509535,
 0.7154174200124909,0.3885068416650117,0.42104392735025553,
 0.36007467511068325,0.03112291899798003,0.8665120708083109,
 0.9719155791091875,0.3568433874801996,0.5873494074829193,0.17977933499652432,
 0.7515475032384715,0.2822098688334487,0.2617145021701164,0.9515956141371814,
 0.4449960714678076,0.4108542044876018,0.4529517224366215,0.47915563751234946,
 0.9805651914833405,0.5675558701472684,0.34928094195970005,0.7175009437422738,
 0.04078985001933666,0.5191384313718813,0.1985337590900339,0.3271591951475734,
 0.750349827226859,0.032120300532989976,0.28454822405125735,
 0.09568642222646528,0.4974052089483366,0.49357633652523225,
 0.40505485644820327,0.056996175769755686,0.41268160225678996,
 0.4892446580682336,0.7492345517220287,0.4305064262491265,0.2560834951917851,
 0.002812433877794507,0.38440618515417224,0.663123066026819,
 0.1559273308915219,0.6497219116817008,0.7927575421490262,0.24332012388465796,
 0.7513992506236871,0.6917086878156108,0.32120791691036266,0.1613994959995959,
 0.9492514762755304,0.3888432410600774,0.24935901458707777,0.3377839742204315,
 0.05483638478168518,0.5103259282500867,0.2457118546465391,
 0.13273806897119222,0.8603677911851539,0.5271473653796063,
 0.22344994763614812,0.07405787764429428,0.9852551432297817,
 0.4987348155843814,0.9336082530243631,0.27692109065547754,0.8350932628631086,
 0.7640162822288482,0.9586273916046536,0.28912462515266435,0.7389593844519581,
 0.13035571193844242,0.6301626754670702,0.9405876384986241,0.9655812025596606,
 0.3160747255819445,0.7500948582786915,0.33888950812427465,0.7402079718424891,
 0.8535386363994341,0.030225449908510393,0.2012199526106947,
 0.39546575598329303,0.5658263025715152,0.9897110151105524,0.758555816815681,
 0.4961513970872321,0.7574080145657346,0.6256518285661082,0.7887720682904903,
 0.10490837823617394,0.524990339087959,0.6475790371851049,0.05710534851983151,
 0.9454453041787038,0.24076648178729532,0.999480527036175,
 0.0064187203097261225,0.2763326561073063,0.6527423450460611,
 0.0016869169941472473,0.4095079856884024,0.30454381223182025,
 0.34223173946190133,0.6604915975353788,0.4698183863467801,
 0.30168889984837854,0.3467066640720369,0.23358339265928552,
 0.17615977197971766,0.18849263215257772,0.18531827017229008,
 0.7215176827445231,0.0825843199614642,0.9757661115675201,0.3870500991680601,
 0.12617727743863627,0.30328039029753606,0.09555234402627422,
 0.959636558973207,0.9546352008256399,0.9372136025337225,0.21419868600811365,
 0.5107902738343916,0.1291911812259724,0.504209964178136,0.5169253744557909,
 0.036349094232851376,0.32523157783449497,0.7770167159359802,
 0.19846420318319458,0.004100914371186981,0.1279334839004218,
 0.5940244156407327,0.5261522779018726,0.029911946043154947,
 0.5925420090166623,0.9361653095282226,0.638548558601077,0.17264736541980386,
 0.4772259427276301,0.689286927462939,0.5860455937224581,0.8694384805807083,
 0.6226797974109195,0.2743002103621106,0.5983256655331424,0.5906520287604085,
 0.23077803683546183,0.41857612730526716,0.7911643215698154,
 0.5368337675756384,0.2295315622877474,0.8082349150280025,
 0.023677180173241497,0.3515678576771648,0.6855072523707211,
 0.9909229548925833,0.4751773740059222,0.24799074953172773,0.4919974169280825,
 0.39156404031133596,0.12424847620727908,0.2961465478185743,
 0.9408894956202094,0.8841912218615509,0.789155546893401,0.13991040971026802,
 0.21432814835854053,0.21645719093757176,0.8232061236611756,
 0.8550819095670867,0.48932073283049593,0.5537008745294023,
 0.47311394260660944,0.04384881193814916,0.6652135702128986,
 0.3662255945641054,0.27209399973405424,0.34678485051152286,
 0.12656113231149824,0.5197138635258819,0.4751555401061647,
 0.00904503828663683,0.8370350521016745,0.6975978338475625,0.555841998083385,
 0.5996922416460992,0.11195446297052858,0.11417030557719776,
 0.6876582107260676,0.7499582211556198,0.27721490908775803,0.742868605554699,
 0.4595568385787815,0.6251050842987256,0.9255306590246838,0.6598827988436462,
 0.7038806668578622,0.7053962937801337,0.8630006395138669,0.7117024594394714,
 0.7570716545636896,0.1617853908486071,0.39035952103915417,
 0.10920044114940985,0.8769044502808501,0.22455912010050272,0.970594165418139,
 0.24638457029360838,0.9145249751262566,0.2069163755411092,0.675068638763702,
 0.5436045704021834,0.8914649161938192,0.5544545264273547,0.9695274431620896,
 0.2534514717127526,0.39433415489963086,0.3495300642269641,0.7680777548800496,
 0.6658015631046554,0.9582289985312373,0.3338863001118796,0.12316877509263136,
 0.5751263757084771,0.9234067159571678,0.10812401126177962,0.7406654336548941,
 0.40824657150214005,0.11693757849763342,0.48207105073897283,
 0.1110531307404291,0.29005169677183573,0.14909743481126203,
 0.9096335300885703,0.24907048841432,0.07769470165992542,0.759149625051037,
 0.3448270961339105,0.4983510269367849,0.12308085339813701,0.4064954033398672,
 0.6612878829421289,0.7100614938590911,0.7323721168590558,0.8432915514490943,
 0.1939121945502792,0.1192605007135058,0.9748753430235932,0.3416805931194903,
 0.11971968431358715,0.9246525356827746,0.32770134663861317,
 0.5340492898749704,0.39720205156033195,0.639715732669454,0.5291765657300662,
 0.6852564985727267,0.17129297051004688,0.36945813274255357,
 0.5731162171041618,0.8811155361100625,0.5648392864453105,0.8350510025829493,
 0.169146107589478,0.3358588396343173,0.4400649181935975,0.5729848116916594,
 0.3700731233919313,0.9790939817801377,0.08384819726564396,0.2812537533102536,
 0.3906470246659651,0.6892715830576223,0.8055339217652613,0.38937567065884904,
 0.19406716974677496,0.575497986079117,0.29404767754497674,0.382263293629339,
 0.8586919147208815,0.8232422425926244,0.19369812908756223,0.3295729303141194,
 0.28061112423076706,0.4417376754728579,0.09674938565745139,
 0.08968824615966808,0.5833281039734174,0.02056446875807061,
 0.011974792729802841,0.5596993478420786,0.7441821187123218,
 0.5837918463947276,0.9396073118121964,0.09864497301303965,0.3188713070781257,
 0.6501533017516382,0.22998944401149501,0.1923208759047551,0.0741260104099255,
 0.18727092355287667,0.9739843395545694,0.288908624495569,0.285340784381259,
 0.9748033153594016,0.3819726834717859,0.9529235444893502,0.5022524598494293,
 0.3674417871243787,0.2793181274919475,0.9300677071284649,0.50116345517029,
 0.4311425073965409,0.4992163051232107,0.24650175488692305,0.2993105769126506,
 0.17654473164143103,0.1573468622764993,0.7343589684505892,0.6654376215106766,
 0.4587389076328452,0.13959129278187676,0.045908906681629724,
 0.9385715298767096,0.7683874487223712,0.067583654752855,0.07759598000540291,
 0.558012968545347,0.1794390701094335,0.1309408670902097,0.6999763974508726,
 0.15124142476606595,0.789942004532994,0.24228946668929674,0.3720142532859604,
 0.511727952086302,0.910885819272301,0.7639349500112752,0.29944945448335303,
 0.509001412558268,0.07047344818164802,0.99506079219773,0.4047835859119059,
 0.5690505694597284,0.8327241272752832,0.7752315719664296,0.6455890468988628,
 0.45876121204281173,0.9644664479089844,0.8961377454283449,0.539773838776539,
 0.0325795332638138,0.9923725866682535,0.887446751795329,0.040028777402450366,
 0.9108989909477783,0.6294776551090004,0.6434508840298665,0.9417088586249496,
 0.01893686734071054,0.2643678384488495,0.6657519265211348,0.8927758119747986,
 0.8562231179985019,0.0018943003937479386,0.1284636278335839,
 0.6357125477838903,0.8320600490770739,0.19477451606941387,0.8397065413798483,
 0.8749213135471454,0.9327571499063998,0.00861025958653705,
 0.021846368748816514,0.9961420120769724,0.863585311759468,0.5471010843183767,
 0.6486004627910615,0.8232461372966104,0.2395343570469456,
 0.045369845235114004,0.6002070527525475,0.5305464949841688,
 0.9231139894718435,0.4616101662601473,0.9709305758423163,0.06514277520356804,
 0.014236111021415976,0.6814676451256535,0.6063373984061516,0.887471190988364,
 0.6423060508536692,0.9833452354585945,0.8183860740913649,0.4573625750971725,
 0.025136590526642744,0.39353010400689015,0.5292527179756344,
 0.37356143310380985,0.15390557732260002,0.38779055522012795,
 0.15956265813653525,0.0027106043420198622,0.5622019551709865,
 0.8559115114842283,0.864630698630779,0.20770524054056316,0.44918073236650213,
 0.05085871388134122,0.8659842048850344,0.11028489266955732,
 0.2107032022965878,0.29823489228828093,0.9200500734538702,
 0.03483264119455076,0.5801894416976346,0.7061834015594118,0.8173568524021326,
 0.03831182830052626,0.34407488007044384,0.7275169034420264,
 0.9324034867591323,0.4283044747201923,0.1918394462280777,0.9271270855753768,
 0.3067667745443623,0.8445665018183645,0.7419156695545877,0.5990321938130426,
 0.46240193552902153,0.776127215336456,0.2601677186295773,0.485304656397346,
 0.12281746834401519,0.8375031330741354,0.6407968963810388,
 0.35072623937617065,0.8632434419301235,0.5000317197844952,
 0.30838195506424637,0.5012037189492368,0.25891340937358687,
 0.24721765998260226,0.06613426818471124,0.35546361758810985,
 0.018128457142810528,0.006329138197151418,0.8066705623015914,
 0.8026318921169416,0.9969658691696699,0.848265395427782,0.10895725453407357,
 0.8021520454514597,0.6060131215118618,0.06768336535143382,0.5223041431491279,
 0.5831040068493534,0.6112084701944461,0.009050441984613689,
 0.8435655812684024,0.9403159090019914,0.9987108210749549,0.7868748635754191,
 0.5652982423370374,0.8963497108613271,0.24011709151817706,0.3619795441260568,
 0.25543015717371964,0.35098500902017127,0.9789249088062517,
 0.9486332180487745,0.8030063639404333,0.19302771472165037,
 0.10248523917059194,0.45260409608948526,0.03390481961865288,
 0.1657321956144815,0.11593277656542822,0.0646729624247282,
 0.04496093912580701,0.0960144090048507,0.9382074667237311,0.6918218383396849,
 0.5186672540147801,0.14110108867621574,0.4126837676518966,0.346733413117988,
 0.524612687491831,0.07707904062160598,0.3846422209959567,0.03183433669660296,
 0.534790928982829,0.39514081105487153,0.6535756571449804,0.18352138410080188,
 0.5310548915425732,0.13805444371354092,0.7689495749807892,0.9776485533206459,
 0.4550710039237228,0.2894897798121744,0.26327237592905683,0.1573194235268096,
 0.598469752216543,0.22273725166655112,0.32357275241301586,
 0.07366978231701848,0.6637898296071847,0.7490275292165285,0.9410623076049558,
 0.8840706588688381,0.3337507738644432,0.21162461695535906,0.9394617190989554,
 0.2261693871553605,0.461855685842511,0.7954251997915323,0.6262361648400946,
 0.43322264937829846,0.8816475425614843,0.4420013847942883,
 0.20934428029483676,0.7722241021641133,0.9206861493088215,0.3251720499885681,
 0.6197960679393991,0.7204289019908273,0.46037487973025093,0.5845488231863297,
 0.6967214387312214,0.42855961619032334,0.9557136042711593,0.4349745816674111,
 0.30776330782865857,0.9196178314789343,0.03578652958194417,
 0.7505784541337577,0.14564413992277014,0.5319092647180076,0.6823841035295114,
 0.8650284255444722,0.5617700868725389,0.36398553666019484,0.93317742107029,
 0.0922537825201063,0.5601137644785343,0.10127315274492144,0.5529416284061194,
 0.7377997965326012,0.24564766290638762,0.998688410430998,0.33376659220847904,
 0.5946986907063033,0.7973078067835243,0.6149765572495943,0.4535386395663008,
 0.4345309422090202,0.6524755082360357,0.6836046189591666,0.04640685298809488,
 0.7904034709911862,0.23006696822881945,0.1752728481747058,0.3545031540506859,
 0.3104412747919185,0.2273062565520373,0.3397690305013993,0.7793990217883132,
 0.48399225259925327,0.8970479784873783,0.5368653260032801,0.5746939729622998,
 0.3066620391681544,0.33868420527958676,0.28210692731585574,
 0.061396846858176146,0.5690543525599678,0.6174644164286739,
 0.8592912540100851,0.6609657351716681,0.6653905384423322,0.6878019703589651,
 0.5535649944506558,0.5180902944834243,0.6231014264179832,0.35222858550901026,
 0.16043906808649477,0.551525068791723,0.015294081809452909,
 0.3161768860977201,0.6721735091636238,0.8947087303027219,0.25041150427398384,
 0.6100876120389891,0.6346223585977544,0.30911178048953325,0.7423044365927055,
 0.6674426027680743,0.031145868841322866,0.4321629116269322,
 0.5204873474507542,0.05343117179796608,0.2755916011100433,0.7882142222651354,
 0.3335901370043649,0.6190528551861663,0.9887758883508418,0.8167285471355251,
 0.628300199904274,0.7339115220665968,0.45537198571696613,0.09467555582204168,
 0.07344160814043876,0.4301122794197497,0.07110238635764454,
 0.03990678838795225,0.6459977805637309,0.1176989565183949,0.8534988200474124,
 0.7413793451404712,0.30018157393334477,0.2957406624303225,0.9411108780466915,
 0.4256058780454086,0.14034662646365237,0.23943473451281538,0.904873391070918,
 0.5334057256117982,0.8990147226031184,0.7555540852832332,0.7429071991943168,
 0.6043910455085585,0.37849276854988056,0.6674351975257237,0.7839468479033087,
 0.13651174928385645,0.031327429941023555,0.7135627721810361,
 0.18193448201148643,0.33328724734880777,0.48912614181351555,
 0.9452686230558174,0.21687133200533193,0.08602331889421966,
 0.8344443649178294,0.616980789560282,0.7926225111705559,0.36946297081700663,
 0.5452850575104918,0.04405226211601887,0.2817415771173014,0.6149990909314289,
 0.577348988260223,0.5782165515088203,0.7460448565310874,0.2297048620755947,
 0.10840589912132992,0.3481230223809385,0.5511963891249171,
 0.11352765444389146,0.3919282119348628,0.7895628454296366,0.5832660621931467,
 0.9818261500213441,0.8078606126095962,0.7710324900997638,0.7656920899858028,
 0.08093812099350262,0.5866576235275396,0.6880112854618001,0.9790713468387459,
 0.8354678963287011,0.8553330696107619,0.9778757882268614,0.7751472007926821,
 0.8103096064665962,0.17282512790810978,0.13928152635198132,
 0.9210294896876946,0.725890300220995,0.11413305899043125,
 0.0072178167025391815,0.45679374104110326,0.21076645898434676,
 0.7462503249669561,0.3144804183489017,0.43051950683546525,0.6408973116098188,
 0.3493569215795225,0.8449284589078484,0.5954008469321856,0.40145112241867553,
 0.9434865857177144,0.22504750856530786,0.8639551619651363,0.8854054025159023,
 0.46593525264557056,0.6408433235452512,0.21817698696136212,
 0.9116696024606394,0.563557435200535,0.8123552379288842,0.22295710687477566,
 0.4368941807067056,0.6480548502429863,0.08372772128009154,0.5459792260482809,
 0.8883736695929743,0.8275053405409715,0.06574349740151986,
 0.31084399578083477,0.8988029422057396,0.3296390231911077,0.3517517868669774,
 0.2538792364046102,0.41513413970663693,0.9179278005113836,0.0632686533402127,
 0.02667667625866832,0.7525031710893544,0.3345834088533364,0.7770808648334115,
 0.529005610714736,0.7228071129697151,0.5125930035183675,0.13634793577029614,
 0.7491291354317935,0.7335073824815882,0.2760174239824108,0.8481573215451839,
 0.2623745926714769,0.08524981954053124,0.704110932834594,0.5555781521849678,
 0.1376851755554157,0.2796153065398289,0.7847565560750368,0.7483792802146643,
 0.6853861349179761,0.3736149096028395,0.31045610265595436,0.2375665702107439,
 0.0441970015164328,0.4053416852461913,0.32837625927013114,0.561315575966236,
 0.7986792040237912,0.44085536438974104,0.6678996532432948,
 0.40693579111760136,0.6666256002741969,0.40377588855149715,
 0.6071776990446387,0.748901648303859,0.31742700283753966,0.847707872589697,
 0.813195495159216,0.4007013372963213,0.11028016464098644,0.8171854420394494,
 0.040512065816334264,0.6251943960443526,0.9857123229336667,
 0.1791644438052875,0.9719275203695124,0.5044247212109452,0.6879025658605994,
 0.5729312498193065,0.6492310090821725,0.16189561534950525,0.8662207549204797,
 0.4282413908815228,0.4659858766312195,0.5615154268823997,0.9803628062865709,
 0.6539464118312435,0.6324823577964787,0.8799466676955228,
 0.030985170181725863,0.8994725561857752,0.3536277662359081,0.871720649537309,
 0.21880906854592408,0.4223148891482318,0.8635727100357441,0.3731719707133011,
 0.6624301828306995,0.3006154038793336,0.010596766211743103,
 0.5130883624043824,0.6384929644027705,0.046094921887364526,
 0.6783058651160339,0.8763100163370842,0.003671658951071466,0.53570798661863,
 0.8863867010782739,0.21165676091467256,0.8894286385108157,0.4653979699517482,
 0.8232310589090579,0.1286403410742819,0.11321877864007346,
 0.043300834508588126,0.9481224653879718,0.5094200943417696,
 0.7094130245584459,0.8732123471734589,0.22059322238004908,
 0.42002143126380165,0.31819149649764855,0.5023946998561658,0.764732980736898,
 0.914565307629096,0.5916372037532993,0.3447062795232214,0.3931888991109229,
 0.15015213907052793,0.21907599803435773,0.23979754782468543,
 0.12379018681329146,0.0083755141729136,0.3173108785276002,0.2887645694639116,
 0.9629645798074564,0.9680947359702898,0.3467312026005481,0.46315103757506826,
 0.2370496477771562,0.21352160435763623,0.0023450725032291286,
 0.6751742603407238,0.7427565899705462,0.13589902243792062,0.9597874965264659,
 0.04755775237086679,0.5205072655129709,0.2609397559610752,
 0.36895303412640745,0.7169338541874183,0.08730959050543019,
 0.5962892293384694,0.2612366991716143,0.5592378044117089,0.1805332420468908,
 0.8043578916612542,0.04729614073194699,0.20153238571195653,
 0.5739581220958189,0.36213060479757875,0.1072874394682295,0.7436459978454244,
 0.39122082900152844,0.09871381630619491,0.5383754762972089,
 0.3574170635260472,0.5817299680460254,0.9127354302292159,0.12203408148215555,
 0.71133419560619,0.8428144648304173,0.4490393745895799,0.6500038858254134,
 0.876833020206809,0.32955407622667365,0.06530158416473264,0.5723118157492812,
 0.43179336816333946,0.665515461856897,0.8266855415225357,0.7904757506381984,
 0.903891343178676,0.553621880259801,0.43740937945984204,0.9674107538589417,
 0.9309445962521596,0.8965481540308192,0.744224914057659,0.39729778877548416,
 0.7013947816744764,0.7502013866147632,0.8228773083683075,0.5173880452237316,
 0.1582401590248408,0.1760369819526556,0.06671046319466811,
 0.29033157796328757,0.8874427224687352,0.3551594447920884,0.5736842955131999,
 0.9154922610371792,0.8793982954003435,0.7494954522247363,0.8599822018120772,
 0.8851741469830577,0.5130284554215724,0.7589457265776757,0.3760782766288445,
 0.12589966741033765,0.5872719566769611,0.35175023314200227,
 0.25495271007085585,0.16807810861154093,0.47351851695003,0.5941575183244082,
 0.5684975450846199,0.02690933440341603,0.4250277599460841,0.5244003722686356,
 0.1851116192758845,0.6839832075225495,0.8263776320784547,0.39377437116370095,
 0.5937386671199136,0.10720427476031236,0.14444589047846068,
 0.20370132723491075,0.5406350535415023,0.5708128649931108,0.8477048173805838,
 0.2368878826742855,0.7157959501277942,0.11353251110090912,
 0.06171602513092789,0.06394234131436505,0.44519924706412894,
 0.45089527200982826,0.6407186285918071,0.36457901374089974,
 0.062125696544740094,0.128246590626792,0.370505087079822,0.853153743669821,
 0.21453312468515162,0.8766355221978577,0.7161454206915766,
 0.0013874614132551866,0.3230740330547974,0.012187227627238784,
 0.05081485093403859,0.08424418324404703,0.870055271012036,
 0.42094786210629676,0.41607320684457716,0.8275806178253688,
 0.3685341722872725,0.503534083077611,0.953571845506901,0.025850776165399658,
 0.8835154843238642,0.711115880256411,0.7506848544983893,0.8316042985103697,
 0.21921636236932063,0.651341704481561,0.05405333827317427,
 0.13106959468853985,0.8341634522050355,0.5866914679928927,0.6508963506818252,
 0.39627456638059255,0.6983005227093197,0.5355720718184122,
 0.15815305595306928,0.37498773505068805,0.9632891938174714,
 0.7981610146966182,0.9754266533511433,0.19628737927039064,0.6668269115100216,
 0.46848124839521155,0.8772108501796765,0.17658121180214392,
 0.7839201336585374,0.4081823188651029,0.5082488531214842,0.03393117373091137,
 0.26957041774489343,0.439207064292483,0.4595180624198699,0.4426254980595935,
 0.8011206798104585,0.7766561040253669,0.3817714983931586,0.10818960236489539,
 0.1796870641826701,0.8545585280343782,0.4480120416036819,0.9100208117114534,
 0.15420996332405035,0.482891264728856,0.028152548580756,0.4716313376474053,
 0.8200501036778448,0.3749557434560653,0.42794492164493125,
 0.32936770377071223,0.02197405744111869,0.5073881688809828,
 0.5361958756795748,0.2587415772814444,0.5535477615909279,0.6222390800538757,
 0.5139522051065692,0.26982544084540494,0.9830982987923923,0.882319265868358,
 0.8721254347084988,0.6723043725030675,0.285916407192351,0.10364943609608235,
 0.3813048185686868,0.7255040766875877,0.9524348914855937,0.8783939072435298,
 0.4475768284015468,0.9161498891613324,0.10212569637663926,0.6376869070973279,
 0.21097903272988927,0.9512368315914501,0.5257314921770061,0.2102789620914156,
 0.8665065730847971,0.7092761257428368,0.45850991085432036,
 0.22294480729517863,0.3947725395344577,0.33929920342608955,
 0.7794752536916926,0.6616433794366826,0.8825590662562756,0.7661990811898753,
 0.8731699194184,0.5678368288011189,0.12634327461744288,0.4662273026485353,
 0.6525617540502032,0.431437241648409,0.26001352184833637,0.05672358383986409,
 0.11436620440469547,0.14653033076916555,0.6486657075809771,
 0.058217594221472435,0.38686889868850904,0.3967252932492421,
 0.2932489754789491,0.4986717661054284,0.07526836031615036,0.4155458909483789,
 0.0974695815287997,0.6111714646624402,0.3427000285711146,0.8384146914712836,
 0.8206867788668997,0.9478218965444951,0.9844166483376011,0.0499070448982043,
 0.43991599365604084,0.1395332455380357,0.6651126355766408,0.3255655665870728,
 0.07901712070435551,0.8965416309060866,0.2911728148673297,0.7321147066222123,
 0.5070133261946903,0.5626915217780718,0.8341667016495966,0.4827537574216373,
 0.3992369900583834,0.9666183850254133,0.06297258914792792,0.5613893207866826,
 0.7708960178953274,0.7123957017499731,0.9750158000593453,0.5276402104749142,
 0.6663741557849044,0.7615608195376754,0.6707034270532972,0.9820984106491494,
 0.33434103508267254,0.4618804332855049,0.5194604867775084,0.3857296477941511,
 0.20090680359795565,0.41185122785129946,0.29305911307278587,
 0.20081301653908934,0.254165156134442,0.061382662608489125,
 0.2511633929266974,0.30246425923728526,0.31889719160188934,
 0.6422005319759515,0.3010482647857041,0.2421017012035378,0.6985105450743668,
 0.5352772181703931,0.6725798375692866,0.2895847374223678,0.18912862130792008,
 0.2678868049022286,0.3392190568041872,0.2693466299706573,0.9204672817092056,
 0.8074126193058613,0.4694823057824501,0.018268370054646432,
 0.6062298023631029,0.26211049301322964,0.12559139483387327,
 0.6157131555349199,0.9326302269566079,0.7021481050870886,0.6858478465354287,
 0.6745520481918222,0.017504311407306905,0.8922736821444217,
 0.5307041040931317,0.9951425418791666,0.4478463719742166,0.46813823010958044,
 0.33878386340918965,0.12720043192565578,0.2925008842065904,
 0.6234358430863969,0.18658355555524242,0.09335140591388968,
 0.11633115062292587,0.40120435844527025,0.6988926777833483,
 0.9933421279409586,0.7934336783195717,0.7895529620743597,0.2869109526299196,
 0.30507378379506456,0.9412855859075548,0.639294177689485,0.7052676735155682,
 0.36857334051558244,0.9718778286544731,0.20673982860894569,
 0.5410794660676327,0.37279489743175365,0.03126510822895967,
 0.37652039199758025,0.21437485622924013,0.39479747062975146,
 0.40184255426785886,0.500640234449123,0.1337498417985793,0.8112472446066393,
 0.526047431492801,0.08490652069967008,0.4620703208510556,0.5987898811308475,
 0.5387035515914345,0.04126278991531507,0.4782603852925019,0.9282752398948003,
 0.6268638698572574,0.16987022409911157,0.024190677530332105,
 0.5771215120033837,0.15204568429302512,0.09867884212658073,
 0.8306236637117788,0.8464449466281255,0.7141561755084654,0.16308150553898848,
 0.008472112223811279,0.9810792318976006,0.6778038709283971,0.878758667538887,
 0.20332022683789497,0.49876218637648606,0.28340104571178704,
 0.2290234403760908,0.7408135035038093,0.6484999965768178,0.5651999314220411,
 0.6893047176326811,0.7619185493769368,0.6686740867317273,0.8787443121547781,
 0.8059967592634953,0.8089877564477937,0.6057432279845806,0.3236359400155224,
 0.49347139320371325,0.947691688054991,0.03780892759850818,0.9932456763164059,
 0.3277211855104639,0.8614756395010231,0.6333382823784841,0.6540323584372696,
 0.15312570947182158,0.5299809165161364,0.5825201243070645,
 0.07042064300426276,0.6711462982583057,0.6904014199796984,0.2455204445838136,
 0.8204140448203411,0.017987337583720064,0.9434581654018189,
 0.4410527021876862,0.029875646606286388,0.29223864101306074,
 0.4604275552478887,0.3450160322920004,0.8939985962800714,0.7016889984600413,
 0.40127013709687975,0.12065059235594122,0.5579692490506488,
 0.6093949930678849,0.23777594550741887,0.036282656764740184,
 0.07695783879284757,0.3671405163598497,0.2563746826225576,0.9589569016496736,
 0.5859275455767778,0.5745705074432121,0.6322420377232313,0.7316817244698944,
 0.47774350338382254,0.15610144842912232,0.950359011798833,0.7442212557498964,
 0.07251858978818004,0.4464054356561148,0.4307591089467355,0.3433047262192671,
 0.135411078865891,0.37607609231421313,0.9429084249984504,0.9638404229981961,
 0.7811677023026148,0.7924504750910331,0.20026982528103554,0.9448442134819603,
 0.1596260678102177,0.765854579521031,0.14502671660287947,0.8461411177742462,
 0.29512881741124664,0.38476104499364716,0.18444080576367872,
 0.10115799188212204,0.9113881346112814,0.0016212673172266712,
 0.2783856432893963,0.9328295894170959,0.6058136099978045,0.8838842601521545,
 0.3297602774650388,0.7708090412540247,0.5888229726221035,0.3927182406438531,
 0.058754454748798146,0.5239440560817932,0.6753405259618408,
 0.8032971726978958,0.7758529192683425,0.9863945955686287,0.654611289112275,
 0.3989132776851665,0.3672923060329829,0.5063825242296279,0.4907468531086965,
 0.6610705379431654,0.33889708519425465,0.5508634500586742,
 0.05423188092475073,0.5302156840360792,0.716711518265104,0.6778650451700178,
 0.6471158998603597,0.4421042338442718,0.6256086044177844,0.33235565880316587,
 0.11055605119872802,0.17402592335339162,0.44113087733946144,
 0.03205311545176559,0.9301458071001235,0.9324704462345639,0.364466832559745,
 0.3278209666866383,0.2308890485808056,0.7291815741874318,0.7634144726058494,
 0.20795732608803597,0.013744125428179266,0.3293149669807317,
 0.42627193791059703,0.1175586106171913,0.47595728844351837,
 0.8279933324303013,0.1538982100601396,0.49084205246303025,
 0.27435871327603367,0.5486217350781192,0.9450636413309144,0.6699695229563983,
 0.4350932748383666,0.4023391572925098,0.4078801362245389,0.4992809304551604,
 0.09396987125383216,0.4449298811330169,0.6555184099380869,0.2421376261213809,
 0.4338543971408848,0.6692668298399445,0.17545791073351813,0.7886568004693526,
 0.35817387362839836,0.14198230579474191,0.36256360118662234,
 0.8879062094781734,0.06246316371985405,0.8009186634207275,0.8973643884278047,
 0.6093954185429182,0.6624762271677678,0.5466190925373501,0.6027713632831828,
 0.16919607216312837,0.06741643786396811,0.4494848048177067,0.921988322365528,
 0.7849657485929468,0.37294608539902496,0.8488356316982723,
 0.12498369885932226,0.5355346524221394,0.9951753733934812,0.4380306273492761,
 0.011610901162652953,0.7696351978254576,0.9113467171139618,
 0.4169342078144893,0.5003632498385597,0.4394615592293679,0.5481961897647881,
 0.2772259585834571,0.8923386952015797,0.9484342250068029,0.9312577995778335,
 0.23116587565530622,0.9022240776765118,0.8724862450967226,
 0.11252257542538069,0.485989019453109,0.5595014725417468,0.946224031798352,
 0.9526944131717355,0.19949323512289674,0.8640209860191664,0.9659190600323537,
 0.4520098703009219,0.7388005383833967,0.5053788710522279,0.992285086741482,
 0.9093187673475542,0.5358275746089236,0.9204996546351756,0.28042885784024973,
 0.530336753755598,0.09332649862255282,0.43143361281743664,
 0.09626586988264418,0.32160323013809633,0.3981345766438502,
 0.8626474415089587,0.8456798003295699,0.03995447219745485,
 0.39013197436920966,0.42267923634814986,0.7345749728092071,
 0.5044878871826893,0.16095948486929268,0.02768115745267674,
 0.45548194986135104,0.21275364956563703,0.48562195014406884,
 0.7717226419544109,0.06300034198379423,0.5909354079576546,0.3779739786903338,
 0.33193481707545214,0.7530200120438459,0.9134157122928783,0.9105150331595876,
 0.04061792986288937,0.12371723316596195,0.6782296884369292,
 0.052604713567199535,0.4799377781986043,0.3810573121401396,
 0.20949120056962023,0.2504237185885574,0.6731938759346447,0.9253464382043162,
 0.2651196556139942,0.11196367281667496,0.5253019221728688,
 0.18828697357574975,0.43511584522091273,0.7112339473809765,
 0.5082638919497755,0.21749075848356936,0.6297718080938031,0.7893520264688783,
 0.20248448232404437,0.4657625016564335,0.9722398687630011,
 0.20114049240617415,0.9162249451242621,0.13093813780321084,
 0.5961302306377074,0.3612383818033582,0.2516430452506031,0.9159331949489941,
 0.6681363137469346,0.7880786672665666,0.7609291722672156,0.2586148693621846,
 4.7820695813216574e-5,0.4977516672218829,0.1943884971596186,
 0.7491822007914439,0.4049630999558188,0.1028204270967974,0.5148924061549768,
 0.45864912436269945,0.6480082822099719,0.08811212800322343,
 0.8834321151758899,0.24373405006062687,0.5049796666990607,0.6511755384998801,
 0.5699530482437769,0.9433725760784716,0.4232936780866021,0.31445254317244464,
 0.8014762172154017,0.5436381966945223,0.8251036315324655,0.7815224305833127,
 0.26155939618800617,0.7188702683697547,0.4963290232080634,0.8022209199387407,
 0.625318242195137,0.011888628526879219,0.3207971721523535,0.6436746848510649,
 0.8011788219714331,0.431567352292774,0.612406280798546,0.022996528040188147,
 0.6756116698980368,0.36572075988494235,0.40329694574068986,
 0.7682090488412885,0.2625432336356639,0.7873075174151685,0.689798010355035,
 0.46589850710334524,0.8591567798128311,0.051390639149893724,
 0.5329379912133158,0.2947369390099338,0.29066771395402113,
 0.19605374235925432,0.5390241112872518,0.12853488671673252,
 0.07791690688282782,0.9120539951978891,0.18790650111035656,
 0.8446020819923377,0.22694774213584357,0.9387495605144477,0.6674355725848117,
 0.2317655994456601,0.9342046298413041,0.9155599878984942,0.1972512329533418,
 0.37566033231670726,0.484480771398898,0.1657257247721028,0.7029353376397153,
 0.49619450865524295,0.2229677555788585,0.14329489835472586,
 0.42636262574688244,0.3377920525320661,0.9454057509453411,
 0.008141147802396054,0.39838263994403056,0.4883870179639357,
 0.20620344342519503,0.5130124678921952,0.43319150424327857,
 0.9240398685601444,0.3980752024589288,0.8475788486783499,0.6513201664650793,
 0.5340051001961859,0.6686721519582848,0.6393170718495277,0.13110396439422412,
 0.7290814943205406,0.7379365578597196,0.28526387008751564,0.7604036813759665,
 0.6152130466420719,0.10131472580185619,0.4610001252791118,
 0.15003295793935578,0.2000322554895767,0.9558023060895959,0.3753540216325578,
 0.6444140726523786,0.19471108230753575,0.6808867582874057,0.7259318980499356,
 0.8362212114034668,0.3495703510664563,0.7789250350656801,0.5409119746094972,
 0.7191703729704065,0.5014260218952248,0.43461145417940894,0.619418529482501,
 0.31198607345010165,0.37587752807464736,0.07626224117683289,
 0.0452866707255013,0.9028525554863801,0.03012183943011537,0.2005823278845319,
 0.44971603282911876,0.7592766017424585,0.4958243891096368,0.6571231788628085,
 0.4335923832570563,0.08470473817104684,0.4172785327202905,0.3535421823803859,
 0.4635068013710244,0.39313673381927594,0.7809019879151305,0.6469728422918358,
 0.8918197717722826,0.5297205590825864,0.44517340859831045,0.5837217009927732,
 0.11613131189298642,0.6629409348437982,0.44955981154571556,
 0.7425970337351563,0.16560885713703988,0.9640266766079566,0.6748314340516175,
 0.36444073716769976,0.5299033663430949,0.13113525445741625,
 0.5706557321092299,0.9062957313987766,0.08258950508375462,
 0.05204072162119944,0.17955012461258113,0.40841470510997224,
 0.29547630082529097,0.06745482132871494,0.20858258283464925,0.82774614504267,
 0.7274160117113543,0.8997849498644163,0.8812143303137772,0.8156192647569895,
 0.5741784358522926,0.08205037460052711,0.3854088876064836,0.1544237597831093,
 0.9166313172099176,0.053070356461373036,0.034737369356693915,
 0.8060941964736663,0.9147343471593484,0.14338473443117972,0.6965770062556426,
 0.6595801620234627,0.22270162815074934,0.9671898903388412,0.4519272160768676,
 0.010290256918850549,0.8347863904393821,0.6661753511934051,
 0.8649148718271396,0.6477404694015455,0.9052439660691367,0.07010534646354383,
 0.6531551875931444,0.16930957129714552,0.9752796431488762,
 0.40205013165010006,0.7282966441856661,0.9235041584101881,0.5571363170077254,
 0.9662768301933573,0.5886329649747098,0.7316090105116361,0.13985457208652774,
 0.4234770085470557,0.20600648704595037,0.19482260470698043,
 0.1133022752898657,0.4528605532395469,0.436397190234037,0.35605280156438956,
 0.004751770682059564,0.8892404459532458,0.9948930234493691,
 0.2839203990741649,0.3308310073950669,0.5946466465411397,0.2175489798946727,
 0.13499646200749427,0.06546487029544246,0.18241450357984434,
 0.28474309310754975,0.6339456474732856,0.2236847412212335,0.7765381341553936,
 0.5557060341474334,0.1091531934125165,0.42575612208832414,0.9046144453644864,
 0.7764256455291297,0.5385884252246913,0.40297325898640546,0.7328161673734388,
 0.23056239784203314,0.6232355724979586,0.9715897736671059,0.5625882480780506,
 0.11170853473209763,0.9011675589352472,0.6308909808838785,0.8349359247165036,
 0.9249062107172664,0.7851613878775954,0.648433679238639,0.25091137572919087,
 0.17074414590996045,0.7012217108347808,0.12538373562623717,
 0.6623479463205114,0.4514823826566856,0.6603977498326449,0.14590011217068255,
 0.15863912754223042,0.45040393205358276,0.06223035562806978,
 0.036064396237171614,0.929408283541951,0.7318398008342336,
 0.03882641320325364,0.08131029932221434,0.8936564893904685,
 0.7281818482707179,0.7144491786040605,0.6041833963965699,0.4792086374835177,
 0.2646551443929438,0.5502463067928394,0.17334158578533965,0.7448527048306415,
 0.42811369122379905,0.9505296585915788,0.005984159194067251,
 0.04957076902600144,0.5451003527513859,0.7784925032534853,0.780431624500797,
 0.014160508851594478,0.31375076226957455,0.39319439606731477,
 0.5644831319637391,0.4200412061932146,0.3014829584354175,0.9986673580965629,
 0.8497797757914534,0.849472653841552,0.7025386377256545,0.09156523100969571,
 0.4173278716949492,0.9100000796172748,0.6276983288174133,0.3254200111158805,
 0.729999721917648,0.7143995890262236,0.519283148366036,0.8405493604585507,
 0.9639248046822207,0.11506559550256379,0.6536708749503345,
 0.20471446623390022,0.8039240810643542,0.4153596466482993,0.9818497065594691,
 0.6880725891748596,0.16032514985309365,0.19970709332167846,
 0.10666258769065018,0.7613494105538522,0.30286453685546466,
 0.16634121979394412,0.4036674635298999,0.024903112292762652,
 0.7564901830400379,0.4781094058366796,0.284364997042583,0.799005339236239,
 0.6960981977234147,0.12866525077376334,0.8804887428734012,0.7995107158796783,
 0.4664643379240496,0.5150843659708572,0.7325135399813215,0.15089483684533067,
 0.5645419269356622,0.10159075387134175,0.06435002501982656,
 0.8055326585995009,0.060630872922764034,0.28904074210788366,
 0.8152796117276265,0.19971506301569109,0.4392522170450497,0.9165517463320603,
 0.5686230040180701,0.3894776262464468,0.6510063351048174,0.4244494360584482,
 0.7661724150839553,0.04669997094402678,0.09949357373473566,
 0.5056378589014822,0.9126263475401175,0.3559738344216755,0.37772926634265647,
 0.14702563409007652,0.09216570131499835,0.25190720741413686,
 0.3012367730238735,0.3702663833924553,0.24772705165433162,
 0.30719502733478876,0.041332797198091464,0.3778243053664434,
 0.09170917088037023,0.7976159779383082,0.3401711898268116,0.4283347290302044,
 0.03192315393412104,0.027044394370526637,0.4654870417549166,
 0.11265291539225697,0.7450039424372743,0.5599034622565617,0.638695482567027,
 0.21869356242678029,0.7319281424032192,0.8021846712557614,0.6878000618845362,
 0.24671244541995296,0.7248231176037451,0.3630752992907258,0.3991527170502478,
 0.2587829601338153,0.6819969456711278,0.9715958444522548,0.06996022740781205,
 0.459936143231391,0.2690825027517789,0.979604280873152,0.2908023343000635,
 0.5654010610838579,0.3576114965999755,0.22914022470952933,0.9258268579008699,
 0.38036572151318016,0.5294092071940478,0.5517248514436867,0.5652255545428361,
 0.0451722480620349,0.3073596616514839,0.17484494501237602,0.7412790163963674,
 0.9490981352297935,0.4250972359473928,0.7795551983294045,0.3620562717860608,
 0.4489410730732074,0.23994204836048194,0.795794509982149,0.6630317066798103,
 0.16895939166312557,0.8776922689057858,0.9809382459693363,0.2693582650593519,
 0.8226975621480243,0.5377443228606698,0.10964458538036737,0.6477658965855655,
 0.11362179410317486,0.22895226488086884,0.8890117815946663,
 0.14415532439098133,0.5555783035253975,0.21282605324337767,
 0.22135010722998272,0.799611765907847,0.7651475638888361,0.9470602509330558,
 0.7023113861622762,0.04662703433975113,0.12173124771989163,
 0.29528745396468137,0.18041164818204913,0.14888704545018727,
 0.4738114191541958,0.7898110641444758,0.7738992758776844,0.499970712661573,
 0.4780312435901185,0.8898545248747984,0.4821879747576432,0.30469586067642274,
 0.22223902620645464,0.3299409716491223,0.2618685563845228,0.7543626516723092,
 0.1917411165049785,0.34692048084354077,0.7352716252382065,0.6045962610057023,
 0.8993856149410998,0.8600362116957272,0.9163728806113216,0.8257285257419468,
 0.16219380853731835,0.5553063661382632,0.9953130098060541,0.9137659950172347,
 0.368117599364589,0.04510912626525587,0.8401263560823854,0.0629591154798741,
 0.4498070070801762,0.3565027711673846,0.659455095613662,0.17329799248484568,
 0.8872786406562937,0.799588150937727,0.019088991730711213,0.586232326693575,
 0.22751262502240754,0.5631307351538272,0.1280288159191949,
 0.40255019973683437,0.9409495776585146,0.514228621748419,0.6448959322465346,
 0.8886528145927491,0.1239603650887593,0.6969333385325384,0.3672861770595508,
 0.3854743565277393,0.802062774048487,0.44774280697520086,0.08939545583777264,
 0.6279590702349706,0.3396843523631029,0.08222296983432442,0.6661389849138402,
 0.2756656235028616,0.9349388088690471,0.17028691587007438,0.4345622448946006,
 0.13061428836612654,0.19542197639307435,0.9802863300614556,
 0.31802827018221125,0.6426070590831402,0.8888971232508438,0.9975659927442735,
 0.3050967133548631,0.49979489608089145,0.9156949021658819,0.6811376891699487,
 0.15780307326051823,0.7156652983479446,0.48541703171303097,
 0.7076428997587717,0.3526851310676953,0.6504299805508437,0.09244532122554816,
 0.30771318000946923,0.38768322292165447,0.6202414468735051,
 0.9243551175710321,0.2247907872212539,0.9244138230071668,0.33712204968726833,
 0.43249132835446646,0.7927787658423548,0.5470488788776343,0.1649300328438812,
 0.519823327057221,0.6498506305855847,0.4965241666378777,0.18036901482540535,
 0.10842167773376066,0.3415686390600249,0.5707350768695727,
 0.44541370963363824,0.44469048945745515,0.43185682341849896,
 0.5518266462659689,0.5020536583423605,0.09764671574482442,
 0.12954177208947038,0.1253321536991684,0.37782554101153565,
 0.8162681953870455,0.6176972117444796,0.5707729768654479,0.9452319106641526,
 0.5463587250594518,0.5504404158955584,0.2340933819660338,0.2735236603718838,
 0.586612381445125,0.35271690951899026,0.5551873194160407,0.7182378387053971,
 0.9520673744090129,0.9430106148198877,0.6834843569415345,0.17421169462564323,
 0.4518159399103693,0.08565044365716978,0.7982021469092598,0.4008901445514843,
 0.8967929944856403,0.04823052668655314,0.25118785185657855,
 0.5862849359520366,0.8410434797753482,0.3853871675314703,0.3874533498569048,
 0.20574681038414644,0.571947286325094,0.8290485312877099,0.41219552422879835,
 0.485538740316216,0.43886903456119075,0.9928252912379292,0.8887044009415656,
 0.9824490928923155,0.6609826165764526,0.6273911910517854,0.3269905738545229,
 0.37454136877249355,0.6728805307453187,0.17737231574292156,
 0.6030684555338125,0.7606254170635811,0.7694622092629504,0.38548859430015225,
 0.8925589460545958,0.8362358013609759,0.6549104876767481,0.7904844057801045,
 0.7155629051466313,0.9302756251657561,0.3570138679580189,0.378014109264325,
 0.3095066382982432,0.277909502569743,0.7884852268727174,0.757742491630574,
 0.6524343575145394,0.20799313565803113,0.11665430700496315,
 0.1775874647106347,0.5599799147638163,0.4015356844814443,0.5059007338615049,
 0.21441959801028254,0.854914146011352,0.25972492750391596,
 0.002044107319882138,0.16528783774152034,0.7508983405793993,
 0.7194122034259888,0.5754238625392256,0.09223133725662636,
 0.00485976525918308,0.810117488184862,0.8857540820834814,
 0.010078406945332907,0.5448681065534269,0.746531285597253,0.6818825901683018,
 0.7755512038220149,0.4976524050146627,0.21568581565771328,0.4783735777130309,
 0.9043161615046796,0.24436843150738707,0.13465582725092673,
 0.021051515278885824,0.6192329342100193,0.521044744891576,
 0.23136299417596307,0.045147843938278975,0.805006293266106,
 0.8257651125348746,0.35297133153932836,0.08102125202851718,
 0.9736754282416538,0.49418832293648984,0.2646453098096684,
 0.03479474492243062,0.8898506401452322,0.3565002146848901,
 0.20771523906090805,0.8085857134895783,0.5468350526709491,0.5232254459454928,
 0.831439859111362,0.08119327315329727,0.7014431625039337,0.8288426600593483,
 0.9636495059802992,0.5941452418993888,0.9402743162971425,0.1863230465847301,
 0.7499914526073318,0.43611847220779076,0.22626284541815456,
 0.2232528268118994,0.5920704147094248,0.5175279699190733,0.8084999828865912,
 0.33595189958247773,0.7749201993173188,0.4813779479042669,
 0.13003825285233628,0.07334982279738622,0.8627189153125724,
 0.07740115281352156,0.9927092625738755,0.23185598986327005,
 0.4238878968300587,0.8184901768462691,0.788163108458694,0.35123146249438797,
 0.02690270450758403,0.41934175704363363,0.020852306710496316,
 0.14292922963847465,0.9946851684960303,0.546793354882262,0.04976158730269464,
 0.6670421498377457,0.8343394437361239,0.6938479281038368,0.06999457497400174,
 0.883237319185267,0.7418138739243325,0.3678023466132968,0.5510885088684944,
 0.6303392645520969,0.5573932773938992,0.8631199658428926,0.9644673508616632,
 0.09839194596555034,0.7528326333018261,0.5359331692750311,0.784846660095149,
 0.5183301168006594,0.364921795915919,0.2346428744431618,0.4682663818177746,
 0.2992736694104201,0.886028642985044,0.7282792307920194,0.2636464188618767,
 0.5376497129467298,0.21211408471974302,0.969806273280482,0.15972735206852262,
 0.19128534070295533,0.6282182190680818,0.04917368841449676,
 0.6817546361842108,0.7023745697197293,0.0019082455784474872,
 0.27892494678788027,0.562530671307925,0.4530406037600845,0.20623898946727337,
 0.9535983765134516,0.11944784796348085,0.08271444747213974,
 0.16098214163523772,0.7721528463624746,0.20179032609490666,
 0.6350418489865706,0.8036234987982303,0.9547988881493995,0.9392469094014957,
 0.6617109588345733,0.01739666384577232,0.7478025259812066,0.897369260537522,
 0.0822405829142584,0.44862820311722296,0.8147770871965808,
 0.03522856435230948,0.8178830779355462,0.5134568767005089,0.5365680398107504,
 0.5664940597818587,0.7217392302360408,0.9440871883812662,0.4835488940246926,
 0.6153358406051564,0.9406053185348274,0.9205974134175885,0.17430071348766552,
 0.8039163607653583,0.4908745534870983,0.30152192480765894,0.6711689431945433,
 0.37098280430170394,0.6133323853023493,0.93359621111426,0.9201226312742519,
 0.43343315386944803,0.9911423737031051,0.42145710929986857,
 0.9927306240724775,0.05764839504412955,0.07594019136117058,
 0.5551156454702838,0.890254792868777,0.8815483906844099,0.8050078954823031,
 0.38179577300733947,0.7579262996138942,0.8578529064450234,0.5507030949539887,
 0.15014987274111702,0.7013739159706014,0.20069224243318384,
 0.3804703492336481,0.9736317704297975,0.3681652860864031,0.48490251127505746,
 0.4528570638119769,0.9821984987390746,0.3515687004749114,0.3927839962571421,
 0.863618698849073,0.23722063588505948,0.7842295915454985,0.3560918048453693,
 0.6127298408433329,0.9772467849818969,0.551325000662528,0.1171998450110755,
 0.8243674477052142,0.34226791741122764,0.5096077879253089,0.1432421740317188,
 0.2623564771271334,0.20215493890879055,0.018080777150820238,
 0.7653668749594553,0.8565602257220918,0.9081297934155151,0.912987212536895,
 0.6577472852841235,0.32457828271049227,0.23818687610806577,
 0.9746934794996135,0.6005840999750526,0.22537717737444052,0.3281751284932706,
 0.26414806323906126,0.7635047210842624,0.8205436184017801,
 0.22545235883734593,0.9929322405305367,0.1537482924450947,0.4626283248430557,
 0.8815183944612163,0.9660369290200896,0.8133489312045538,0.9405886257916936,
 0.259146094889784,0.3624524581896952,0.5690979778439388,0.3953952581414415,
 0.33716221317109674,0.926629422848531,0.16103111861640107,
 0.02230028785893534,0.9432626352825513,0.8513675634148903,
 0.15271456334988065,0.5286897028846416,0.36406698480146693,
 0.2740730147024125,0.4032043664090219,0.8641177714740652,0.6652223941919848,
 0.6963775335415973,0.17278534302777349,0.6725505238031154,0.9586754836088446,
 0.5030200757404074,0.8433259268405296,0.6197164662712988,0.9447790000817613,
 0.03721401283586845,0.6329594952812728,0.5569480837527345,0.9604968550122566,
 0.06221140684038351,0.6802445601118269,0.9182177181399243,
 0.21888928184052636,0.6920325805816379,0.07715913986733325,
 0.5522678735532159,0.46449430993734386,0.10269289318562314,
 0.9730315152447988,0.2737064487342341,0.2203594165357573,
 0.040761999617734745,0.5126664003423379,0.34821340974410586,
 0.40865049348116156,0.5619430302766639,0.5621357031313929,
 0.36960844606095855,0.10238941751184194,0.17962637005334825,
 0.5147582793336984,0.2744874298664348,0.4428815244365565,0.7119353304632452,
 0.9822265021963132,0.6848050985839651,0.6489107224375161,0.8241756445281213,
 0.1486286926247451,0.6409214353416564,0.9647428253275275,0.9464632211818182,
 0.22515715528709002,0.17620269397686072,0.9367395055473527,
 0.6791547318772772,0.008447163394294965,0.19538220349870228,
 0.23873604889552102,0.013254515672860068,0.724929286679536,
 0.6609294610210421,0.025913493587627845,0.25412233473431034,
 0.9430569801581292,0.33869298830983663,0.4682864884443184,0.4677736345729585,
 0.03536922001534237,1.7548402267353325e-4,0.5834279024692219,
 0.7958955370082512,0.6186411383418953,0.20265201331838067,0.5551173694099035,
 0.4884909278475251,0.01125086744481929,0.18181359331564095,
 0.6991686197565423,0.10005515529944242,0.5955362471742047,0.1890502716636535,
 0.8883588191759325,0.62287789970533,0.44586730277922604,0.7513048619480268,
 0.9817238212286827,9.858471135381075e-4,0.1950704394551952,
 0.9397012655138787,0.5372394377685202,0.06543527653008774,
 0.44502518799995094,0.8276164033943665,0.434400103845783,0.6980545747552722,
 0.8234928993250084,0.056884963797511734,0.8827742078867177,
 0.24626509008330677,0.5198788168768507,0.18298242312276558,
 0.8614013107807297,0.34099284742561,0.903165294443835,0.12503528766723415,
 0.3479406755070359,0.6533377955396403,0.35014570480258933,0.5808093524051822,
 0.9631827345221654,0.7557986931149284,0.29473858897285166,0.6587150379587869,
 0.6378815729973862,0.6421706556667044,0.3987722164948303,0.18789946141376535,
 0.22815584892528173,0.41882444827612886,0.2257977469350182,
 0.017122105949386457,0.5510390441257497,0.02452655588249475,
 0.47831620968058297,0.8767745460441722,0.3138578359914226,0.2640356034379512,
 0.6671345904524276,0.46930266320752734,0.21500318243283567,
 0.8215406450256109,0.5002086703789118,0.008996087651199414,
 0.8908293724703066,0.4425477100729056,0.3409222337771316,0.876110837471956,
 0.5857226353684163,0.839216683801327,0.5579849805730921,0.613729323381779,
 0.7063126410246316,0.03224870925122336,0.2120537431703594,0.7987781340829145,
 0.2625886412454429,0.811483239708536,0.818593925534471,0.05053199143215226,
 0.8025657356814059,0.4971782257624694,0.9192907289654415,0.08721729147648916,
 0.052567351751731106,0.7955086246519059,0.8632456261042188,
 0.6553614013323057,0.8632073145583459,0.13057910121117344,0.6915697550015838,
 0.7993055509106378,0.8594594461043757,0.7514073215617185,0.5192247268056265,
 0.3925552288766926,0.8291067033012587,0.07060890047347645,0.36096652624086,
 0.24382886399149206,0.5247991929764868,0.7888995344777991,0.2279992052300479,
 0.26022827741431,0.7490678215788403,0.5113138882187916,0.8817532749189767,
 0.9044261366829174,0.6277125044959533,0.614075583744332,0.9058366957858834,
 0.17145436980021111,0.5980578614736759,0.2565753815642834,0.9191176079346016,
 0.749686069064826,0.9815470149799254,0.19431842658842324,
 0.016063029954436647,0.9837106966947405,0.180628066805554,
 0.43847096561842747,0.6312527301866252,0.23134685586926773,
 0.8659087098433893,0.949516432214555,0.7696918061149916,0.6578131631969375,
 0.26581624410690874,0.4462885675040208,0.04127779656374131,
 0.23945572918109115,0.7419505949580402,0.5399141487006583,
 0.04538248054464378,0.1165787264728354,0.4875617516326948,0.6711363964770025,
 0.6831713054576409,0.3402119035161152,0.1422340355875915,0.17668909860539284,
 0.515751208372006,0.2562568495880281,0.8199128650983383,0.9146323947833331,
 0.4960088048400437,0.29359736346046217,0.7373136454979673,
 0.29395957428481145,0.48313851547533593,0.825330189309708,
 0.42357133511714906,0.27654529562832253,0.0051671066564481505,
 0.5500272507985114,0.5903078910065147,0.543169325167251,0.49450113269832174,
 0.11105478598038754,0.8725044132183222,0.560348326005505,0.17658086605496925,
 0.039294089633253426,0.9221856396985877,0.2222290661860694,
 0.2505932029921538,0.879003127649219,0.7287435850572251,0.29239524455103805,
 0.7556738299188128,0.00551690629165047,0.9354765431855132,
 0.08097692714967908,0.6601278175826554,0.5278221032022635,0.3403700841593247,
 0.704023310898272,0.492451266679926,0.6290689636101253,0.2672078260705718,
 0.17058737533382007,0.07163118040058358,0.14360603999856947,
 0.39747302702539744,0.4654941651216832,0.12570048639154707,0.58942155953853,
 0.8030785789472275,0.17586838891760048,0.977936713032507,0.2896416788607965,
 0.9951719189505441,0.8597695514328462,0.5457963469856237,0.23316935775691428,
 0.847575553878,0.7085891869394577,0.36559487129949164,0.4500085480443541,
 0.019836552188429946,0.07032750568434976,0.7943428114368359,
 0.21745096017084165,0.3759826497150599,0.3315196831005379,
 0.17574106425028424,0.6528455340186132,0.10578916435645758,
 0.6017657766835827,0.592767031383263,0.09709658237821739,0.23798207810276917,
 0.07331370816968863,0.7592317313334068,0.8355648342033133,0.299235406889078,
 0.4945072717200385,0.16658810120803536,0.34033240846130197,
 0.9803540351099865,0.573996212942853,0.44796380869162156,0.38715442527174193,
 0.494173230109339,0.3546078702622685,0.18500684543000812,
 0.0028191925459136735,0.780212275273697,0.4052380075452374,
 0.41171523905547425,0.6981548987671778,0.8363896997449743,0.9812500325743979,
 0.15361890369720377,0.9656674063065732,0.0851551497395241,0.7148901399523555,
 0.5375285947721836,0.5860100781873308,0.5766962294981453,0.9072825138916116,
 0.7594515534408119,0.41295516290985057,0.9486718777068678,0.9224512876826075,
 0.8818699920670071,0.9152483394830369,0.2958538393249346,0.2245841349410811,
 0.9416720205895466,0.5435521510121152,0.33736315327746236,
 0.09328082033555474,0.29569891282423844,0.3045970827222588,
 0.08609787963970394,0.3330289243372023,0.8988251177985522,0.9821206268722676,
 0.9013834236239944,0.10535421123570776,0.13781081520175809,
 0.23523352185711277,0.8774614786754675,0.49711057944441095,
 0.9593938599824489,0.7502975137327117,0.03547949224006086,0.4822238948832064,
 0.15817870843285542,0.737252774265101,0.4695156431328965,0.6384507547628242,
 0.7981470463904308,0.2810517330543265,0.7912793832368836,0.6288239847725716,
 0.6587574110635462,0.3173149339073732,0.09253352678896754,0.4404695883295434,
 0.002251983196082774,0.10548065676016449,0.3902216786596845,
 0.4411751885346472,0.5018146566787883,0.8726761462997007,0.587261804092906,
 0.8432876014197375,0.2693707530340397,0.6212100967684508,0.5243670607658597,
 0.1106689180858712,0.2614874177037242,0.32961788586373086,0.9736512156127858,
 0.001457120195002659,0.7317456196180011,0.9407901316630587,
 0.1651287807966807,0.7177882258611736,0.09668697019688788,0.7892231681783758,
 0.5563877374656031,0.32673476311041605,0.9094802449908714,0.948373771171091,
 0.49669918105072486,0.3791768848154594,0.512211785577814,0.5557788129253216,
 0.34411989041518043,0.13267451511919903,0.835374237881823,
 0.11188993966604388,0.27559498994369014,0.30671420801621707,
 0.17830442491199094,0.5498386472756045,0.6488581030262196,0.9336715702837571,
 0.22802855465699245,0.3256102555102104,0.41333701484112806,
 0.9709099643080562,0.49500620303432785,0.6307368436799377,0.5087419942057457,
 0.679605986976008,0.12773379165678533,0.37214809625325596,0.7499158780132419,
 0.9593008162386572,0.5547406273452362,0.773178422854486,0.7676017666576075,
 0.4928441234376325,0.10350465549751808,0.4708906282394886,0.2937171050135772,
 0.24961880677911008,0.4925170766505311,0.7053318347389105,0.9267619036553821,
 0.2293561319726235,0.7805457925162306,0.3841373744896712,0.6089235108145674,
 0.8144416441327212,0.6608515343856207,0.5980764604348437,0.7428578638947473,
 0.042793188260130344,0.12003040265270548,0.3967897081611804,
 0.3394388121177917,0.546531765700146,0.3917957327375996,0.8131674258165512,
 0.5660765742784072,0.48142907957490144,0.8647236015166262,
 0.005473228989622325,0.7076721505533791,0.5911322853676533,
 0.9157835808043474,0.20153398302710068,0.9187642787463846,0.7519102497223026,
 0.30798281838220565,0.9722850281826124,0.31088175186398237,
 0.2873218931938035,0.8666813501632314,0.5712772366151045,0.8469975015972154,
 0.3240678212275152,0.22211180060157798,0.4634901994731129,
 0.49481581885865733,0.25734592247931265,0.35567798217789004,
 0.7940658812665163,0.3624663380146522,0.41191080808884506,0.7927593058952395,
 0.8818197760744942,0.7220031857161053,0.7426440591150352,0.8962387420436561,
 0.09878970348032756,0.5140152268343763,0.31023223521093324,
 0.36010918487794985,0.6162422423132976,0.44450204672975424,
 0.7265474163991695,0.8667572655891653,0.05553870372184522,0.1721505504123021,
 0.5970419058036993,0.053573896844764435,0.13266919165550384,
 0.5034262399389631,0.0696563210284249,0.39882545893270716,
 0.06220551858911172,0.2894284745925534,0.6489793222743514,
 0.29770890613617906,0.3965474191191659,0.013114186987651766,
 0.29702554289637106,0.425367084255601,0.8287777838727244,0.39611898418383007,
 0.16677724769431013,0.888707206806419,0.08671310418104128,
 0.07802034776618316,0.24380611368009608,0.5103392354264216,
 0.9948886099404479,0.9015553749848497,0.8618283735912995,0.9927308545222404,
 0.9026437895507423,0.1045283271928027,0.9361058906593279,0.7660373057783266,
 0.9304446666466838,0.44288337171543357,0.5255789583532134,0.2754214454884186,
 0.3721592673776122,0.8147507607644994,0.19032645800543568,
 0.32278347808055596,0.6113462760366936,0.5628785501239932,
 0.13050606518064234,0.5906241946036914,0.5493245422388506,0.6768009218550539,
 0.7286352962250335,0.575341746293861,0.43749910757147425,0.6226139885561962,
 0.41286580779249726,0.4030604991868957,0.5798225278111387,0.9129301007158604,
 0.6886797282162737,0.4233627385786807,0.9205918937107492,0.1877564571532867,
 0.5403395281660568,0.4673064092476894,0.24049480428426462,0.6197175467045323,
 0.46704164700238315,0.6869132688498798,0.6333833460192534,0.7297076481613798,
 0.5421515655229308,0.9048865425867526,0.8067744019551637,0.8919427016201802,
 0.49689074960465707,0.5182607044603809,0.791470293990778,0.954367779453812,
 0.7577677111825634,0.606820141990012,0.1771460792446371,0.2409447773905059,
 0.05858480976892988,0.023398710316657967,0.3494066247930111,
 0.19794239536091685,0.6620649249385923,0.870904476774695,0.5596159553286396,
 0.9662459410606187,0.4493699032870334,0.5631843893322308,0.9484203323429118,
 0.4135787047751275,0.20613629394715272,0.5647880247681479,
 0.19163846889778335,0.9982152329606894,0.55851911325112,0.11909755488683071,
 0.13130583837191,0.15979220025633223,0.10554276518074623,0.2910600087067595,
 0.5877720201243889,0.8591984997206086,0.2170594071650589,0.6119752972293325,
 0.9231055975396334,0.15463336873332567,0.0467296631475006,
 0.021155710744055423,0.022460544749040134,0.04339238796269629,
 0.3505822142369057,0.1343661647972807,0.28955100790881694,0.9121051956877627,
 0.9983498467505217,0.3001136694219241,0.6476496476065933,0.25966308735438026,
 0.21627560444315463,0.7235390474510329,0.8106973058694694,0.2703756897010663,
 0.6102606462585374,0.13751572698879833,0.4215743216037555,0.4792184254741064,
 0.9584812743591345,0.3817520962034925,0.24241192551585122,
 0.06476329709028716,0.5825424581054679,0.3535825657367736,0.2121964900278751,
 0.3987186306187154,0.8777537180941316,0.5830811480171387,
 0.013575991509002128,0.3327447274642924,0.3413812411182652,
 0.9028628606942602,0.4504841401734123,0.3265200454350565,0.49362113929091533,
 0.9785531529783243,0.18791112633189744,0.880031988311656,0.04136572904839453,
 0.20502725740274375,0.45615314620356273,0.1444121687781822,
 0.13773546641705292,0.8684768266129547,0.3284737347230765,
 0.32731025630370936,0.7585107466834456,0.9518859151504229,0.7832295533416593,
 0.540295119433803,0.018152234970808645,0.47794082023560647,
 0.9014963613977357,0.7298296424808906,0.8123314701769728,0.8345364990908833,
 0.7592896433631562,0.3195691989517113,0.4675605107495583,0.20568787439895297,
 0.11333844862324527,0.058615514159104154,0.4015103828709683,
 0.08764478558320066,0.568876623491667,0.6093764767597079,0.9703556428624144,
 0.6155323062468486,0.40129056758103565,0.635566033467056,0.6741063159326395,
 0.02578356246013913,0.132240673511836,0.7868897119830667,0.8257706549369659,
 0.16330479064952108,0.059798816014711154,0.41445542199781915,
 0.1617815073684049,0.8424565185466534,0.8044916419058903,0.5509706986734025,
 0.8656045757092214,0.33135689194084994,0.8870092083368026,0.776423956671356,
 0.7423621897203516,0.2024751792133832,0.808593677897884,0.699180566715618,
 0.12115849110560517,0.441449964838054,0.2856834313453869,0.5257411075839833,
 0.0951070150773855,0.38136154274145806,0.13672468957852568,
 0.7639614153272284,0.03149672760343669,0.08335523752532736,0.592773047198358,
 0.9967662796330323,0.8720849445873878,0.6205780968157802,0.8267450262227996,
 0.6306306981796562,0.4938803766868963,0.8154291051446404,0.900589336811478,
 0.2954941496577542,0.18092175996982574,0.4382961534027968,0.3965403331977905,
 0.04067773849994205,0.8990093098322322,0.6981156230209709,0.7401398493149355,
 0.10863245027240831,0.5473699522595885,0.7261574253603801,0.9638013119449808,
 0.7249892368624994,0.4616388028162043,0.94164094603786,0.7405228480137303,
 0.5463186252749601,0.7376473922353146,0.8468809456918623,0.5324005753271631,
 0.2828482751078262,0.08046761016395976,0.7651452897604327,0.4581539587248509,
 0.3379895007150402,0.36334212655246856,0.9689398420595287,0.6328213493690985,
 0.7533150716376867,0.9588655966164578,0.7135175523970896,0.6715424630413939,
 0.9569351137004202,0.4407956433241993,0.8990410803040397,0.8332737068357806,
 0.15214234644641733,0.5977932290419408,0.7193179216830338,0.5511733677713324,
 0.17345038209768382,0.70718041490557,0.8261601333360643,0.8945612067312743,
 0.2740489887423312,0.15469845395728354,0.0727235630412334,0.5603283306638798,
 0.044424238897577295,0.7604026915401635,0.10758036569675256,
 0.352442917452488,0.11208506374258298,0.8971590907517181,0.717086506876846,
 0.9511347509850708,0.9996660019128513,0.8078507164674436,0.7811432940014477,
 0.519013481431984,0.602974631740086,0.0858258362405665,0.002700519651279043,
 0.4954700580210418,0.12458209036713908,0.4669748239796744,0.4850491241179923,
 0.16497450532734204,0.10549543335014155,0.766582230528027,0.5027226649801487,
 0.7879018570754562,0.10515758626037364,0.6960631968289198,0.8247765214975824,
 0.31257661120448077,0.5119813536276326,0.6716200650469387,
 0.30196040236852184,0.7862937618280627,0.9905074174580101,0.5753728827324837,
 0.11550648087024695,0.2129581554551836,0.11959837306200927,
 0.8254523317317236,0.7937468341244762,0.015286251130116724,
 0.4652506850207906,0.34025487640165164,0.008931419280823416,
 0.8124349997964383,0.8296679844980259,0.6953542619359949,0.39479927324145436,
 0.617362779269913,0.933028038759194,0.17871362784524736,0.6255600285358472,
 0.2783806564021837,0.4503253849464991,0.0016934845958552769,
 0.8918147354416989,0.6656398484750606,0.9802951743810476,0.3535798854766866,
 0.8443760005831843,0.882084124433258,0.47267087845901323,0.34375660335607783,
 0.0342260241054847,0.43151676516603976,0.38759691793234696,
 0.7555213962194711,0.08764681359131365,0.9389144635050494,0.6757090129982822,
 0.7907799937205405,0.7827318258300873,0.6996735447867133,
 0.029366642575871404,0.06403157594389497,0.05040704739616675,
 0.9566859885166263,0.6284393226171437,0.42029563363360267,0.5319217042239202,
 0.8780467382349644,0.12556215559605877,0.406733875133245,0.15695501412822332,
 0.713453963131454,0.13660873627279602,0.4286342752722402,0.1742234216265388,
 0.6420353122788626,0.704313302297884,0.43557112727828506,0.5913200407704444,
 0.3058528769605613,0.6069520051407389,0.7900182437912295,0.5639023551413311,
 0.8337420523189492,0.520966243565336,0.8981386433885923,0.25818522506095354,
 0.9036623496883662,0.7160985709667665,0.5364159172219037,0.32828989809073295,
 0.5088251496783913,0.8120663868999516,0.3026111298487719,0.13294295135219425,
 0.42663569088946307,0.8054913390490405,0.8487021553969404,0.7228964174345016,
 0.7952779209371297,0.39855266871463546,0.6455936623268846,
 0.06008797790648435,0.6206016330094525,0.3719297769429232,
 0.26167358458528367,0.8247783016987706,0.14002325186635334,
 0.5268698034970011,0.3342870633664886,0.8108034715375915,0.8773311577517715,
 0.9620343170976161,0.6488536857600601,0.36841741791126825,0.9295948168372159,
 0.7986617695211428,0.668954095091754,0.05774694139963443,0.03985840403215435,
 0.23238253745071036,0.5037181244124647,0.9614219502434063,0.8161275685543397,
 0.9833171239067213,0.6465816382790527,0.8575782152020807,0.14163701977448173,
 0.762371725686046,0.03803890485884154,0.35266615621510533,0.5180988407509374,
 0.998172314644467,0.9077216952180789,0.857175972758543,0.028319101179576922,
 0.7064220853192383,0.621974077155346,0.5790220140119232,0.14762125224794032,
 0.00756601932905876,0.30182310990603056,0.6610820361552073,
 0.6764702628695568,0.5075575546722146,0.8271513971439253,0.4399646009831345,
 0.3580710023139184,0.8868164578361741,0.7466032906022122,0.747124223490132,
 0.27711963507339166,0.6188803785016466,0.6551177559413602,0.8610842247001727,
 0.36611158484915995,0.09822370358014432,0.6304887684022774,
 0.1295330107480943,0.8617651369996766,0.3885410870063434,0.545655748677456,
 0.833297996310154,0.5075605065369909,0.8404585316023199,0.011282107448425949,
 0.7224613189706623,0.9270250176421693,0.13471747669787648,0.8501957015239971,
 0.8978370844107333,0.09664790669039103,0.12248121076922447,0.85276324949775,
 0.17991040345487608,0.3838302327716645,0.22030913027647825,
 0.2064170271692638,0.7525689956528374,0.3281146501632929,0.8257692983906187,
 0.29159242727546286,0.7524212171897816,0.5646041459243238,
 0.28185692264558226,0.20485801804375348,0.06883570175868181,
 0.6555570604002758,0.47652707949017936,0.8215937926513642,0.8016293209304987,
 0.14796917473171345,0.36606332510624107,0.28952506744941553,
 0.1381915938763425,0.4307068364876059,0.38003388644605085,0.5754885214788203,
 0.23969434080125518,0.09505748142443893,0.16966966575848696,
 0.6184562757408452,0.31826664676795635,0.10994000861592501,
 0.3277735827291438,0.25836872942095346,0.5419265482426201,
 0.45009076105952006,0.9637832003325334,0.8648854249436019,
 0.014799791485545777,0.2514351405228623,0.2685968126523458,
 0.2987045467257442,0.07570624603711484,0.4616732124801055,0.6849121140036367,
 0.9773022391830971,0.34189436696838205,0.7440072900145696,0.9575113102845841,
 0.3971998688605135,0.7754895846317994,0.3077971384740318,0.3233648415800914,
 0.11611282842164816,0.7415810632402815,0.27330755241714344,
 0.3813267856338841,0.5913140758241966,0.5319738379245185,0.3976348595874707,
 0.03168510719645434,0.211518520366205,0.6892783865610107,0.3011780655647067,
 0.041094505340052656,0.8101210625647998,0.6366212765993645,
 0.3229074452268734,0.659861721561992,0.4204662373833148,0.5620675817756261,
 0.021433022202047347,0.43830119112695876,0.5336791891898544,
 0.20160925181094802,0.8482612727985543,0.27263484047599595,
 0.9312157865143831,0.5101451775862926,0.10949165130743577,
 0.36713906650873473,0.37732207981565735,0.22534258991927258,
 0.16203265505552045,0.03969791904193887,0.9472464601582805,
 0.6597585457395381,0.7716133952910592,0.6626338083355138,0.36415519882016356,
 0.24752608182341607,0.1884355240157214,0.4317039686808689,0.75406738495113,
 0.17061366495856423,0.10523263296590435,0.8452193521828446,
 0.05039332702662269,0.32190453629852667,0.11822861740362833,
 0.6606385406497719,0.7445124419346086,0.7767110226252143,0.7494179715483881,
 0.2613584269446907,0.8147274075482125,0.05696556837940048,
 0.36512132346245163,0.9690289451896855,0.5744717281060748,0.4997026294180963,
 0.35708806578872987,0.1817437987556786,0.17655969398967208,
 0.5339364730914705,0.04908215222095291,0.1312901580132032,0.6930811061555431,
 0.6192956286737737,0.5218468162066437,0.9369867701965956,0.5416994839774935,
 0.8792193680206501,0.23413122826130595,0.2883480843529005,0.9567343562726528,
 0.4055758891584942,0.9421366409893159,0.8809283914294461,0.15405760821615788,
 0.26025878371158406,0.9350729402355856,0.2446721962121361,0.3495783953380994,
 0.16001080509236898,0.5649285107840064,0.8588354152437649,
 0.39118208898317663,0.7536517032368628,0.7862778599015843,0.7188610330757311,
 0.8770557340165162,0.7477909244402389,0.1092472609441163,0.8986851318554876,
 0.5828544219629233,0.1251884610560945,0.8205809609818437,0.22490129616514143,
 0.6683047091031363,0.0362767321876144,0.5345271608940872,0.4839939698334609,
 0.6140354108591969,0.5874507355069013,0.8851795339972506,0.6462764946074784,
 0.47641670840807115,0.397549067095871,0.42668702776227385,0.6035175381949313,
 0.10676967818921645,0.38905826109759944,0.11076788681049987,
 0.002218757283553363,0.3829975211748019,0.17381925455791192,
 0.07495954863834431,0.8171910329083616,0.3921583620288416,0.7810661884655364,
 0.9014529498843574,0.683432465077978,0.8272330891589234,0.5986935265538524,
 0.1535522095820282,0.49066701054202755,0.6810815390416829,
 0.046357781338637105,0.18466390845659308,0.0013106747824646048,
 0.6839737267689292,0.9643469079863144,0.026001767660941866,0.955241640014948,
 0.06502400382253415,0.6164576560806254,0.3126399590686615,0.8824301471021041,
 0.36717524085785325,0.3365654513788907,0.6144743936154166,
 0.18993682743018958,0.9840259490671805,0.9786721311414873,0.6100812989291732,
 0.12395784389568676,0.5920531231690778,0.5803708705193709,0.6976387741893173,
 0.26591998287757634,0.5367280873070707,0.904377528938255,0.6835072420200595,
 0.7681357747880175,0.9484724940702681,0.7518904918041974,0.49893641121015897,
 0.7986888704763417,0.3737807672503701,0.7459524693700472,0.6028317034484092,
 0.9905354190969686,0.8383219064962322,0.6534545810264962,0.9426619665998643,
 0.22206904337458688,0.9419508524455199,0.4607430093683692,
 0.22174771019711403,0.9194300531829618,0.2918230076041346,0.7228120843810685,
 0.9351636691898769,0.3167932059107106,0.646642641124238,0.07950294939216485,
 0.5224374416654229,0.5591092818291709,0.4948633333312399,0.20942589187951643,
 0.2546474183287032,0.5237445167201612,0.8686716133394943,0.11375132326627513,
 0.5930581791392743,0.4804533422096571,0.8680701749119205,0.18535156991142843,
 0.8351736702068923,0.14772664864335416,0.440831121206148,0.28782887518744626,
 0.5760915251371327,0.1123485552926975,0.04471720514236621,0.6280775380681017,
 0.4812009695038304,0.9311908718248156,0.17464884053483098,0.993536998528277,
 0.5445561288115948,0.8534798029387904,0.9424231371346997,0.34812117136442855,
 0.5215984472373926,0.8228645391948239,0.33922611514554524,0.1534450936763846,
 0.39535715218976564,0.7586747460245659,0.11886779745515241,
 0.34859902703560264,0.09660042637619681,0.15967838141420843,
 0.6790885552735515,0.31280511290919866,0.04812264413771694,
 0.16300129986030432,0.04774133680146164,0.8174040128228661,
 0.44622019584013084,0.37173811053098704,0.2378972644799887,
 0.18280812790120748,0.8924871055248548,0.213572288131061,0.7814655914464134,
 0.6345004727632926,0.01752181980949341,0.08143266311626352,
 0.9539414732484881,0.44042328877375736,0.8439460090530336,
 0.33726962915626957,0.8491488987998499,0.7401218129925953,
 0.29883650173827536,0.8091059269548513,0.2140886117880093,
 0.0035014153194041775,0.5828902983067777,0.0412122657666113,
 0.15134358961217953,0.6083770174555664,0.36970097721308937,
 0.9481976327672531,0.5143979136730498,0.42604639822481083,
 0.31783485377964527,0.5655651216480047,0.15172681178763137,
 0.8387993705657876,0.2669782911990163,0.9105298481667496,0.8769214319100702,
 0.7788765934207111,0.850703951445026,0.1393097078358564,0.8958768217489032,
 0.8642182271971417,0.017128348407717597,0.5437235707881336,
 0.9633245264434488,0.49950764598053476,0.9848135564763542,0.3743050114765093,
 0.2079036919353494,0.3667052688971675,0.2340553246090714,0.8977229082615218,
 0.5900309727076025,0.8942212949275424,0.7590005510020918,0.9666414367322407,
 0.20281931457804714,0.7817939710033102,0.017496949188569255,
 0.6188531343316297,0.5190047776969721,0.8575699613301658,0.06837254294470241,
 0.9858293438706686,0.956187423531029,0.5892531192924336,0.2806937768996395,
 0.7780682992695066,0.742268490489012,0.3659550638203738,0.18927262991419858,
 0.2785890074718276,0.032951959462829694,0.9685935671175302,
 0.7401781582574174,0.3240217471976852,0.42574928477415785,0.884563228540984,
 0.15546840694053743,0.6201954649829444,0.9742219813200653,0.3199140103010729,
 0.7498764289408943,0.26748163282561144,0.7510396776026234,0.3167529560363652,
 0.2902556299028872,0.32756303189710423,0.06392223915664685,
 0.4313721056555474,0.4555093117721386,0.42797125548621906,
 0.07407769455231017,0.7407783190818953,0.3606022205828837,0.4913889828546888,
 0.020032130566899897,0.2968727564699094,0.30109616449845733,
 0.6109369086378181,0.36119461383813023,0.4959188555696916,
 0.10860915923772263,0.6861049763761233,0.8063880759041427,0.169004438459185,
 0.5284513447796763,0.667241054724935,0.23064677804792433,0.24920491577692894,
 0.6075658774796188,0.06477394820292692,0.6879314980581308,
 0.046244405660729515,0.773476450904837,0.42553798946954635,
 0.36871063710614915,0.9514642006679528,0.6547396361986452,0.3432137567510798,
 0.010628209201699268,0.19993842131719386,0.9726161561594859,
 0.5657061232151902,0.7499134594800707,0.8579570123564898,0.24535110924812298,
 0.5566239456101882,0.10203178703571836,0.009641162233000267,
 0.0934721636579593,0.8134046604366927,0.2522082744872568,0.03390255456234903,
 0.7521629711499122,0.47944324424214657,0.5503981311267405,0.3483729020368457,
 0.2237504418884393,0.7628874181805056,0.0916287386030803,0.578488297505342,
 0.08475485782104353,0.2980450926195587,0.16659792915015215,0.613744295705928,
 0.604343038744378,0.2375530594744346,0.4785353831558824,0.2521795096541738,
 0.5483259755293681,0.7609496626268649,0.9769999423234099,0.9039716628411305,
 0.9403998613392834,0.7998284666095471,0.706844630983144,0.8530629214278557,
 0.2797107695298803,0.3726279878087704,0.904417439786339,0.017103092658785513,
 0.3715340149534492,0.5375796585373046,0.15044333421472644,
 0.24523693827284532,0.3109544995321335,0.20642809272700924,
 0.38813441863345144,0.14451448975970715,0.7746378897505246,
 0.8181422319925746,0.4421375831683566,0.48030233116514276,
 0.40933628767987484,0.2994485351371746,0.34210715692262694,
 0.9827653681686077,0.19609076657239588,0.6471906657876675,0.0709567781840057,
 0.6691932986038382,0.6314616511554743,0.9818612384413374,0.5938340414633544,
 0.42801983876006733,0.27849139244325183,0.1976255052251049,
 0.1569198516628364,0.24631578951951338,0.36188550997355784,
 0.13206164669393738,0.2987825277422467,0.7697153871027369,0.5069540428092265,
 0.8950644582321355,0.17884336521009803,0.5825860205033314,0.6555960665459377,
 0.06546362921280247,0.5199612374742315,0.964007143218612,0.4390383189853536,
 0.5621841774437968,0.037467927030415904,0.3013699667313041,
 0.9648320029577667,0.7301846948197263,0.8567464658850484,0.36755470101619203,
 0.8689480574277637,0.5318361643907876,0.7720755784732025,0.9983970389919629,
 0.023036017772965045,0.25764536770663926,0.6288917313985755,
 0.15081333362524152,0.13119042723830276,0.6788009067607167,
 0.3552639180911019,0.2871631502020684,0.20529104952356758,
 0.18663016182496694,0.9180289434759475,0.6555865543731353,
 0.49200640966418085,0.5701057527817148,0.46410499188743193,
 0.9368423871175972,0.14914751091654666,0.7433551757577506,0.5463055280572506,
 0.428106452494044,0.7338146076364943,0.5391668937083286,0.22441273739209755,
 0.2623130719377956,0.39377206751917326,0.8050055525032477,0.5501661220694529,
 0.7333254053116411,0.3949526361983843,0.467133555400602,0.3823615975963538,
 0.4592578364327041,0.39088420703706117,0.836418024331082,0.5996508876561997,
 0.3075780619264389,0.010933982761442529,0.47116952883320096,
 0.267082668039372,0.08484465823519116,0.18349728969355195,
 0.14125532602878843,0.06319574482326551,0.024445940446123915,
 0.16533708350756948,0.38434856133071094,0.8009588035374038,
 0.23714335275158183,0.050398342259736606,0.9628382339169727,
 0.03956255732632874,0.17302063117269717,0.04870868769468939,
 0.2364191764110828,0.49502525279165377,0.5033656817753414,0.5624476800169794,
 0.5303268600953441,0.7847784294587731,0.07954300304505091,0.9967792452689236,
 0.8118046042949194,0.2941550772337437,0.9020777176427485,0.6793631453228192,
 0.39869239583793536,0.0267372578421603,0.2780408610627342,0.5267911713073317,
 0.9604996560777752,0.5684685908076706,0.29990252841722687,0.636914590566629,
 0.5876117837511454,0.15387771507565995,0.947640023994788,0.6026707175758259,
 0.17191278280397881,0.5820261584562809,0.2825577906439971,
 0.12492686327229874,0.012777208626596925,0.6008589462157959,
 0.49544834774368596,0.7441786960165496,0.3455085780256729,
 0.23477302231297736,0.7346577848926731,0.24078816616971277,
 0.21329390589293695,0.7363019301386904,0.029501696658550625,
 0.8741452522779736,0.9236711846749633,0.7109003249366685,0.853981335980736,
 0.16097414429503165,0.9670547968600383,0.6755273860388631,
 0.14307971153931043,0.8452079472066805,0.22749479292143926,
 0.29367148723140124,0.8379915222040489,0.06836202700852667,
 0.24121825547713394,0.21963556718127653,0.030431860077750117,
 0.7582808661910329,0.3985837601366303,0.41314271725316254,
 0.21831226204297505,0.8659550523224885,0.8606539026463387,0.3267987188377507,
 0.35351008380749427,0.23999665751099575,0.4648540684780458,
 0.7334592192031899,0.3331248942612698,0.4695852391329577,0.34124686759627376,
 0.6700300775371033,0.46439239488502804,0.1690010386277413,
 0.08305657698715951,0.8078230740238276,0.10799882110054704,
 0.14177639610982373,0.23595121552570575,0.6814147788731747,
 0.7041696539521255,0.7268194448308987,0.8917420817783163,0.6686161340107644,
 0.05641406400102533,0.7912649532214768,0.3461751264065043,0.8751495610419345,
 0.35018236265813174,0.7496138013559833,0.6945796126329973,0.7891597684559075,
 0.30300270507628546,0.9528383443938547,0.006485925036595508,
 0.380039776893926,0.05510380241985524,0.8962250273843445,0.5241814459608561,
 0.007648731277742904,0.5592507405839744,0.8639919848792763,
 0.6306464827698003,0.27174708449572804,0.9462271131186177,
 0.19020567300387825,0.7261866081098522,0.7892838230698573,0.3590701673569615,
 0.12480475859459916,0.9405598505771199,0.37454251586020104,
 0.7797000220373068,0.9383373795673634,0.49802111390592574,
 0.16808409006065073,0.076548816231651,0.9035912134821964,0.8895084535339668,
 0.30913168824733217,0.8141340974503047,0.2759853690765981,
 0.09862013359346467,0.019629535030896594,0.5428048291330542,
 0.5174079817458199,0.5803116691539865,0.885919861795136,0.9913286090596257,
 0.4768764038298299,0.17951902377030327,0.8545056771327633,0.6111352840102153,
 0.43226476603573716,0.42487187785671876,0.78137386858795,0.08268024751793102,
 0.8091386460701843,0.13309566589334643,0.2356345041475041,0.3427731785410718,
 0.8348351157485564,0.2075812293793724,0.1624409944515528,0.8383468834561187,
 0.511974091311382,0.9450024177608316,0.9094816171136798,0.7645717648476966,
 0.8027149155755764,0.9351090736983197,0.3859446711822705,0.9877102392836985,
 0.8192374920315888,0.5805572486199873,0.5542919594522369,0.8913473213541755,
 0.16453897149322838,0.4606904691607201,0.920388250160285,0.1668153126457511,
 0.49965925827800917,0.945581495846959,0.517709745442987,0.0932592412706219,
 0.9474270807100252,0.13775638095588139,0.9405307913294844,0.2677068070672677,
 0.31289003833591733,0.7734328598604255,0.11280000505654864,
 0.3754391750212992,0.5699402898146568,0.5875432367509705,0.7544478408372411,
 0.5972764393498742,0.9811852803527019,0.8161980328405486,0.7113293247105804,
 0.8510611835967159,0.43242667432896353,0.2358034197725789,
 0.18689708045756848,0.05790188116062067,0.45140099043980864,
 0.17127940281057963,0.04805728073288884,0.012991516315851048,
 0.2397852837467942,0.24853101827466095,0.5844350733511353,0.3009010353478965,
 0.5888875732044451,0.6278263224541751,0.9471163468602456,0.08928061228568385,
 0.7756104665718758,0.41846739682977396,0.2855575099329296,0.5649964132818299,
 0.39990246908480875,0.17461092620768548,0.0946098716583802,
 0.8434628061641268,0.5799568384850062,0.038121437796794666,
 0.27119168872185084,0.9292278291497899,0.7609245441403615,
 0.016493596992184667,0.3799728642385741,0.9981436429486715,
 0.1378399319343112,0.34932170771881965,0.39878641416235927,
 0.6667659273893709,0.5159416280093964,0.03038358621483106,0.6459641815090628,
 0.42816102191525873,0.31435771400160495,0.4767818589754579,
 0.37468016749150335,0.3617754188489367,0.6883756269826904,
 0.17909878135205504,0.8793628759305686,0.10204937616896848,
 0.31570637265439405,0.6299870605744855,0.7710587154616356,0.8143581907383033,
 0.08313951290419319,0.5453932376330944,0.7467507413729612,0.604667256801426,
 0.5299998227371434,0.12096324422094973,0.2985211035794806,
 0.43784359363717196,0.8651711581473508,0.07606736547727977,
 0.8938188009311064,0.6995041972938174,0.94829498352752,0.44027760753673395,
 0.00995790324394219,0.6568570443063548,0.9313438346863595,0.4053771151013901,
 0.6117601495851676,0.13595622532946927,0.10294986698747399,
 0.8162281837671196,0.5505714142279361,0.9812081189503843,0.5800692986505234,
 0.29233681954858304,0.4970539677009377,0.24536633676787845,
 0.9110434899347202,0.27386663528530275,0.8824214898516002,0.9112674068322546,
 0.07879350749730374,0.3052047752119841,0.8708595717397691,0.8775628565963802,
 0.8990277279021217,0.4573451943808541,0.08626796188140629,
 0.46650402438258043,0.5557474897665129,0.7424742481652447,0.8443714798392764,
 0.6111318581409549,0.7962397569795223,0.36603414033483594,
 0.0031797847893135778,0.06465080951294844,0.771971530472448,
 0.5317370350063036,0.474115770155335,0.31585441449957463,0.6448410248940926,
 0.16199021316002038,0.9466528922018373,0.19908459867920902,
 0.3781930187027106,0.44842468337514596,0.01508706544831906,
 0.2675845844803839,0.6310312870990618,0.8060418420740011,0.514750061544047,
 0.23909694167874107,0.8779362847723026,0.1698112698865012,0.953574543237439,
 0.34586235383053376,0.874648455201688,0.16791612402829192,0.5609922572886064,
 0.6345310968893314,0.5646512579358935,0.21219749864732362,
 0.14962745737064487,0.8416780937873762,0.8099119407864667,0.6527946157619049,
 0.8453568462239296,0.02433783059389627,0.6358362143982245,0.9510054995026304,
 0.32073893769117257,0.8187774458002494,0.33547502153233155,
 0.3492762222530714,0.5784172145780214,0.4270813984250935,0.6401556593470193,
 0.33028772478107393,0.0999524729537008,0.935603904661287,0.45496072073780136,
 0.5190341888523847,0.9155739723787997,0.013499270553829157,0.283371713631688,
 0.5687992013594276,0.6459289785371062,0.3664320922067229,0.07323684321940094,
 0.9814863633731074,0.14440670463771654,0.28973395126835444,
 0.3374927951281851,0.1272771861372879,0.6537311020788132,0.20375273992814347,
 0.3741092433172195,0.33694173565364793,0.3143428435401773,0.7410383465300662,
 0.7610824329414535,0.7705096328017399,0.8230630685505089,0.42517353022961235,
 0.8519545407054692,0.36190924305822325,0.14206435499506798,
 0.7191617367041722,0.3017842889446073,0.31743881483369163,0.8814040623854004,
 0.5292726607168385,0.7203012910228467,0.35843339083170844,0.736935139269656,
 0.8490532518546251,0.4320455426452482,0.9254089562187171,0.7468253477795335,
 0.6890409414957475,0.2097076884541932,0.704574874112996,0.30498943200785333,
 0.9739598680387587,0.5731130756806532,0.30337373949424873,
 0.46106173881714163,0.4014511474075215,0.24847031709400014,
 0.8349610383246778,0.04377035752932601,0.3859473008936054,
 0.28958666761912577,0.6055689660174006,0.3304237805737088,0.8095711511901642,
 0.3837447771406379,0.7074130320018075,0.9870997001388032,0.9066285346446813,
 0.34260476180527755,0.40231856315690173,0.9015981292936146,
 0.12454604149709914,0.5226634262661226,0.6691000922488848,0.8502397985299943,
 0.9440985097768955,0.16192820383348094,0.6153767530219166,0.9094861761900852,
 0.861342538442363,0.4147242289076003,0.8176377548446401,0.1233052693323552,
 0.4543279522193686,0.3161414244639764,0.5421790337925142,0.3435633635362736,
 0.7701192813107083,0.730586019774389,0.5723570198971031,0.6385039733147301,
 0.4431272252782993,0.09917668220593534,0.6020020581757562,0.5902846811685245,
 0.6071966707164229,0.7501177734398811,0.9981261063870099,0.9425916521520353,
 0.3501890876572038,0.014442142929376178,0.19452689257799494,
 0.866142285868411,0.22780286774545289,0.29583463529294707,0.1784831364275119,
 0.4169841826590206,0.5020626576100109,0.800367436094562,0.21723546394921,
 0.7560235363118356,0.2630051727521141,0.5611780952404268,0.3207112768923983,
 0.29955130923913886,0.04783772627430394,0.30594653272468375,
 0.3192623489356793,0.10078778487874984,0.7380483182693256,
 0.40564056778479496,0.06692112580566556,0.1287372042445094,
 0.11075535775634437,0.12894679803823939,0.24362244101210095,
 0.854859820297459,0.7250835904895236,0.18795338962443642,0.4350977621066926,
 0.6996751384106465,0.947391559268708,0.8401313208571395,0.9218371721368671,
 0.5666237178973444,0.8716719326456843,0.1942659190160586,0.6428127389593905,
 0.6593128537181803,0.4508680312807708,0.15417745727475385,0.6759824544734575,
 0.25888679035755535,0.4451590801508115,0.6281831323386318,0.5860451321782333,
 0.49909134168988456,0.933793223991219,0.8920394540914873,0.9950913312338306,
 0.4023694645521677,0.002916582086802788,0.7381302400453554,
 0.6561769660465347,0.8448589899608655,0.13691137167810385,
 0.015584099379656435,0.9357254734383106,0.5690226083673284,
 0.5872513358092448,0.026390143128708687,0.5989930930108759,
 0.26562841639208457,0.18249530483931586,0.8960827879565042,
 0.4132069780337959,0.9568425749809107,0.09300318337999947,
 0.27301587293251306,0.6698713063221601,0.041214054334211325,
 0.48765355755308604,0.8312007973959711,0.7763202674191483,0.6320593960138324,
 0.146190961546241,0.11565936526438825,0.03875586675119169,
 0.30278519031967965,0.42562599951594504,0.8457126597232905,
 0.7399644519317946,0.3989132305621407,0.7381778290274722,0.8822241845060743,
 0.4869765181449204,0.38648951643173834,0.622574060382761,0.47564945602814157,
 0.4323704048716259,0.4357288471538221,0.4845159873408199,0.6434219771206671,
 0.6500006320562718,0.20792096028514317,0.8149278452730719,
 0.22110730094596054,0.36888268163329463,0.06270858764544407,
 0.8274063499238364,0.6221185576726584,0.6355152545745923,0.2025019530191643,
 0.9639007838674674,0.44593649451898176,0.3213997769179736,0.5828923826980623,
 0.6648404198692942,0.1797767968935814,0.196029948689451,0.21010102384656038,
 0.8506235251031202,0.05246397639150169,0.2204160555804766,
 0.19047471513209968,0.5762876683517325,0.0981618245166378,0.8333405580612859,
 0.8215309616270577,0.0442653196897721,0.7882542875570749,0.2956096129361939,
 0.6828037792304225,0.8579330988485444,0.2676565777466152,0.5479830356334353,
 0.4357515838924323,0.19205747908072657,0.5826848616884063,0.9482843617659125,
 0.2386059661132146,0.19899111837213435,0.8258336204772194,
 0.09553579056853456,0.4894694800231494,0.9210308431706257,0.7109350216782522,
 0.6495510107710546,0.07019926911603824,0.7202130696642585,
 0.28687232822283826,0.25691860068450567,0.4838887365807112,
 0.48010079452583165,0.9506078877503986,0.03002542327937363,
 0.8599746074729665,0.3454773396145644,0.45024545132275884,
 0.48908061759609023,0.4730570910481412,0.014517833636709665,
 0.8252917669771566,0.48403327630788184,0.27098063341012624,
 0.6020335610971783,0.43099616169278154,0.4630689462608143,0.3357871145330549,
 0.010591667579241015,0.2381452133431079,0.20718716679696725,
 0.21823945382955778,0.6589784779093733,0.3791918578838612,
 0.41158099834431117,0.039399235319753756,0.2885535464614507,
 0.022334713413335816,0.8918648370589681,0.04205515773185953,
 0.47400551362480714,0.323421889402864,0.42080406009611604,0.7279608159308949,
 0.04048205802276017,0.273867621866,0.5128121523584275,0.022564715046369366,
 0.13041339939909624,0.40360248853235947,0.2934553987697637,
 0.6952314712248874,0.9339088940774938,0.07830067421872844,
 0.13907993817399567,0.5609962044839983,0.968574160289307,0.5903990349371289,
 0.4779665273440882,0.6446993312205013,0.8491254551977044,0.8781231220132758,
 0.6813393343204793,0.5685544123441352,0.9785846847599936,0.3378537463013893,
 0.1602229829718339,0.05900586702390154,0.9982416608629187,0.9204358056687095,
 0.07011216622656802,0.972099815048776,0.24563972767561593,
 0.08669911812060005,0.7110542163898892,0.7434893776980003,0.8210835821359428,
 0.5581530487089804,0.07004981951516154,0.514694296576938,0.8485375865361394,
 0.6378391050993415,0.37471970640939656,0.5131860685237047,0.8612850382868006,
 0.5410469242947525,0.6485020552959152,0.2968322270364957,0.5622913199944353,
 0.7988493627521455,0.6882171729581849,0.45462287457327777,0.8218622203636491,
 0.29785789078941993,0.4001233687551531,0.092882269067604,0.7037013922842337,
 0.1766930890374514,0.4026779681340642,0.9649704828436384,0.16917604325169022,
 0.16368429788371908,0.3675268995757748,0.5253537356341205,0.7021341470760556,
 0.21063511934158097,0.46378714476297866,0.2756305593564269,
 0.11450029752927837,0.15890205374054311,0.42984853047862914,
 0.5130686283169152,0.7305812004070061,0.6189264531777531,0.8152449586735562,
 0.8232218298825236,0.5549917578971413,0.5997852839282614,0.3526612328262957,
 0.1720692223007847,0.011695067713989205,0.9500282321164252,
 0.7412677977515945,0.3171023885297424,0.37912562871138256,
 0.09784788218244445,0.10205770910876888,0.5257757947175319,0.869061988180311,
 0.10193672024784506,0.8153080300898511,0.0876033109129083,
 0.05530443728420875,0.7842747619620671,0.5291828097760315,0.7520687125778808,
 0.604894715620698,0.07352297876802616,0.8516606067440052,0.36719582150157715,
 0.26984853520013274,0.2058404756631541,0.9316078149887582,0.6151363924312645,
 0.879530998878305,0.7584117512869964,0.14640648501970732,0.9693273412094544,
 0.9501834725995941,0.3879237013759791,0.14375731785108692,0.5167996596859656,
 0.7708824238197705,0.1284260003929506,0.8447077858422226,0.4046361328845396,
 0.5449099540567248,0.9100210999870465,0.06901944886605305,
 0.22607305601374483,0.7048369721875998,0.37879603235356096,0.556988864124467,
 0.437995429597405,0.6731047469449906,0.130176547051698,0.8252886651298138,
 0.7916602895701785,0.004980400203174695,0.7375113794084966,
 0.7022497855144019,0.9566087055605769,0.10295990900563912,
 0.24056361624124423,0.8925001576154168,0.07777468405240306,
 0.9742335856834311,0.0864448039036565,0.14954037436560108,0.3568386940503231,
 0.01621014237954488,0.44582442460050997,0.2919520261885128,
 0.7403708085713847,0.781993592637276,0.027910198737317704,
 0.35266849185814186,0.39217947103086914,0.2971606421223151,
 0.14256764873950578,0.2779645968156378,0.029830427614610855,
 0.4689394631919376,0.5697803003104225,0.2644297804721809,0.8334639186015056,
 0.7804315445283376,0.46521687167761894,0.4010021170907807,0.7877518001421513,
 0.7505307771541401,0.9365402123856638,0.6276919364476842,0.672801388649691,
 0.6556533844997683,0.24252248914550734,0.34731586726154307,
 0.4689837775252954,0.04358396285018884,0.7857438989054907,
 0.04632081703507773,0.6896502456747281,0.05672273028000585,
 0.4792250276706498,0.7124887171837144,0.8966379162403398,0.15933107615632058,
 0.5728987171909427,0.44317753143974925,0.9031436494859048,0.6362789031215229,
 0.18958697514548162,0.1640031906747913,0.2054383223405234,0.7618991769646923,
 0.5505288106254883,0.11104184858593913,0.5153454683651759,0.8615932870690182,
 0.8870268453632696,0.5855850587867166,0.5749258723964313,0.7007441475645722,
 0.05398036341623447,0.7315513337960535,0.5695018913462787,
 0.24046630172329664,0.8455184567914813,0.024597182219708635,
 0.20765659304248352,0.5051942742511659,0.18861469270734066,
 0.14197441511699216,0.4918382533313509,0.07475701123384437,
 0.8235471287367755,0.5750595853464727,0.7254177068011861,0.37973608369339085,
 0.21795678534050134,0.05191896320279976,0.6484855550245049,
 0.6247113281727042,0.8748709361314964,0.6516141902487805,0.7285414313163845,
 0.012745165098680289,0.3745000404305522,0.6906178256939114,
 0.34149399445393724,0.7254961833163338,0.9109807382104493,
 0.07347001941870557,0.3088971303745671,0.8976160329582534,
 0.028104204827648416,0.7556553417140915,0.21792732374092905,
 0.6218632655993683,0.44460095994488724,0.06595486888536195,
 0.9661847408806683,0.884832444792713,0.38398316457547277,0.4082574868335068,
 0.8612825135850881,0.018681646747463643,0.5854427353063059,0.878348625235169,
 0.5825457454884813,0.5953427137406758,0.057547127066741144,
 0.7544348635321514,0.43611259439730177,0.9086808122020535,0.7413423123918266,
 0.6631031558030341,0.675765517039964,0.5433905447731695,0.4331815273711268,
 0.6889743983788806,0.39825030860636124,0.25063299867182764,
 0.6974357041025098,0.9895162629689607,0.029091697135750993,
 0.7936934849014033,0.9545865274240032,0.32674417870636274,
 0.24116290957832964,0.7840645604905091,0.7597824019623022,0.3466867448630928,
 0.9387642694147024,0.4610920074613749,0.6184139909002682,0.836452824076415,
 0.24021340956664794,0.1916187032996537,0.03276855077884877,
 0.9345708946234437,0.5478273384620991,4.941895862553558e-5,
 0.6646751032713054,0.6311944070118671,0.3554759871030506,0.2542822661647085,
 0.8894275255125836,0.42153123033053785,0.7522174190527714,0.7502005071239604,
 0.8412870880137329,0.13774928926241137,0.4132521924895839,0.6058148071278834,
 0.24512732048770336,0.9652568536055194,0.6160507234141404,0.6813161337687722,
 0.39049599145906,0.7634265949938457,0.7255338368641056,0.5111370661100453,
 0.6329863893836258,0.2548398754072291,0.304901719821144,0.31874471051078335,
 0.7861411314398755,0.7869791635794342,0.9341569941627873,0.974220487134879,
 0.5070465556932806,0.53149056121312,0.46952865654017695,0.009524252369557118,
 0.7907119687718508,0.05157410739967028,0.4987332150165521,0.8692284783625658,
 0.48066228685710577,0.9119529555005594,0.2925603843643021,0.9329964911789382,
 0.49335891801787335,0.44333805495594625,0.06278047385319152,
 0.14938130935146632,0.039850980030558425,0.41083962476461733,
 0.835056657045645,0.7303826879047404,0.38352169112833534,0.6974933388874931,
 0.7509712138581089,0.926522313684151,0.788624229270433,0.22721574192208782,
 0.8206224975508152,0.2673230738352652,0.3152325947840321,0.15747170290193224,
 0.5334120669749305,0.18412933682061583,0.02083569003904384,
 0.5553602270453356,0.2865226037314842,0.5557786843360392,0.01659152484205073,
 0.4963567457656094,0.8170129832420133,0.3205675528985247,
 0.018034588506136928,0.7421469931463464,0.8721282391682235,
 0.4992748947354477,0.6829161985976039,0.8382268701909601,0.6567993978833891,
 0.4285584509111299,0.25862400751250825,0.35208648936469855,
 0.7796695577230565,0.3423757585124144,0.19961283493567072,0.8644380326943293,
 0.527158666028277,0.12062317127967925,0.6962868181907415,0.9397035005830839,
 0.42160780971123346,0.9010879335135689,0.95457032815095,0.15749064388129552,
 0.8060365152336122,0.9562080491546846,0.11813803313956317,0.833120157168454,
 0.4491203211646596,0.8675643870882184,0.4135623968711556,0.7156946830292604,
 0.35179578341851436,0.17127364374059328,0.571470439207079,0.636329170044278,
 0.9989536465435468,0.7315236637239662,0.40114628522556317,0.504896649529435,
 0.5812993895308175,0.07575806009649999,0.21744057149214757,
 0.2998595283293275,0.03147122666001945,0.6104867703875304,
 0.04878509122217933,0.4550828953756467,0.36804666007543196,
 0.8645130942166042,0.5507232132122631,0.47258020993696714,0.0824660210063215,
 0.006630839598146476,0.5135647836382309,0.8092551247596963,
 0.4548037148010331,0.7658074731184006,0.003475362465574894,
 0.24285846570205072,0.0309493444022767,0.8500348301847357,0.7713243555989099,
 0.27066332158798345,0.2610968424437543,0.6859535475824907,0.4688997927688424,
 0.6031699368788421,0.07755566779569012,0.948406088344952,0.7828005861360545,
 0.0802971604796976,0.09182077293520974,0.031121172667488484,
 0.3835262300881175,0.49483641194431427,0.30605785534413954,
 0.8703728003756579,0.4313274904707137,0.44434704478154097,0.4450629853127156,
 0.7484636358325164,0.09797725515533162,0.792277824392061,0.8318698730491647,
 0.43359316190776487,0.6909847672638221,0.40371121746913574,
 0.5354133722938724,0.8756059820705907,0.36320211670842184,0.2892195750408195,
 0.021007718456909963,0.18948261239080844,0.42627641819213546,
 0.8013568758110508,0.9389339550545479,0.5397894946952545,0.8776529206829886,
 0.41830309676417476,0.3536700412019862,0.7602398102934331,0.6452680667679436,
 0.45158755986375465,0.4092875194991614,0.9348035074553286,0.7891412843481506,
 0.1114581555278491,0.3830353696843991,0.17065899841547338,0.5804970557174965,
 0.3794424500861966,0.5198156851263664,0.7540417438942758,0.7202055044533426,
 0.03969793379651965,0.4505373516753448,0.5058217242096567,
 0.35283163164271425,0.13043607286512682,0.8616915354023833,
 0.09379622686436861,0.534694181775724,0.9590038940556102,0.8739989175385082,
 0.9828610981679745,0.7891710628795021,0.7648477717859608,0.8357398467111105,
 0.7752889500074872,0.8080061916534215,0.5718951407075942,0.26532883646899197,
 0.7313106164046651,0.247889954013071,0.0646913223647626,0.5376773491194742,
 0.03326936334701236,0.03347230951314273,0.7121894667133559,
 0.7965449663410564,0.46426891665308967,0.3553046790576252,
 0.22943109988983001,0.037677676931171855,0.506494150214402,
 0.9070874255017527,0.6948728368631053,0.33005518946781043,0.5420117471363686,
 0.41403296641004017,0.05685943171445884,0.8943682005612494,
 0.08323989149076716,0.83498843853301,0.555379646031145,0.8538212841116745,
 0.4915044064402315,0.9441407032749305,0.8180921555059516,0.6059259415154964,
 0.0394784062927237,0.8103265388751026,0.8682623049678592,0.5985864261402164,
 0.9170378883107863,0.9536294258915656,0.2302149557204064,0.7937516767364404,
 0.9773636151562949,0.6954311427280242,0.7430742673303387,0.8850731818318551,
 0.7585549653576792,0.3308785227731441,0.6089278045157313,0.66700196236972,
 0.8770334075464613,0.5462422839155001,0.09196145728413241,
 0.10466516311449103,0.5488707128987177,0.8148014785350443,
 0.12035906239725713,0.262828770995968,0.6349675708558266,0.455691686987295,
 0.5631142368495321,0.645270500992688,0.9710271932707747,0.43416804112751595,
 0.5456880330243441,0.07553411680857003,0.6366138584994396,0.1671346134585644,
 0.323269098279361,0.9607475106229864,0.411941049729186,0.9610807755690629,
 0.26006987673105464,0.3980001516453825,0.7177444070413339,
 0.25642836460899865,0.1656772034169487,0.5787050496714654,
 0.45272202148974383,0.9694104943440756,0.8546837440002805,0.5667334750881137,
 0.9175113954667095,0.43827784112288815,0.8953777473789947,0.5570477592257633,
 0.2767202156027686,0.3634925118172557,0.7461835118775356,0.8249528658672948,
 0.9779465073537579,0.0392385446407455,0.14268799527910958,0.1502817097505984,
 0.24922794700376405,0.5164216226993129,0.7387869532648487,
 0.35887922246785997,0.1377188904141271,0.7139441066155329,0.9473485443799889,
 0.45456772096733833,0.4515875933401897,0.4077983469032843,0.6442990696903371,
 0.4606450644828598,0.08889748828726607,0.7135020262236773,
 0.34102745784239386,0.30861731170190454,0.9977432213595894,
 0.17309922456220705,0.4046700619447776,0.9841877647563194,
 0.21112190541566234,0.2218009479797185,0.8101999867047273,0.7390090346014114,
 0.4591506054959127,0.2075690349860737,0.9716038616226593,0.6432134490214357,
 0.6004254909194038,0.6126718864189004,0.5059441902546077,0.7881694158204259,
 0.8765801541686302,0.08644109808817668,0.10132405456482618,0.980829461893193,
 0.16394979914702623,0.6347846276938794,0.38256246764458623,
 0.5840679994931472,0.9195855051187196,0.17819782408763774,
 0.13592770989919734,0.753936921983672,0.903585524699019,0.9077443816192412,
 0.6852545990303096,0.6151929150670988,0.7776370414656636,0.9534645249438528,
 0.644809754414334,0.6186301959696229,0.7302757683101262,0.6528159001789917,
 0.42284506985790427,0.23614547559377463,0.1390222932008106,
 0.2794252197494008,0.10180866834001989,0.6105226964129966,0.370433182603367,
 0.6463346172313096,0.9790615881572899,0.9713773187862639,0.7084995792580737,
 0.7005433030788369,0.704058552349794,0.8907875416731459,0.6836518636698881,
 0.2294449753787705,0.5257614079498509,0.34098055889801815,0.8611361963299256,
 0.43096883190474744,0.14351396058965926,0.7990250266002712,
 0.6950777732306496,0.7054895841767768,0.06010206821874331,0.6936711327746627,
 0.6266094345234947,0.9105646289288751,0.646492990871273,0.04052763410886151,
 0.8937637111327896,0.24017597123775902,0.9297028369408594,
 0.35875037102053153,0.8033994773112556,0.7789543870979126,0.8732019268984555,
 0.4885957474413679,0.0690834433553259,0.542308539774004,0.28066061630935135,
 0.2928156849851188,0.45143249434438737,0.2791100103309604,0.672334838260493,
 0.7077170320889704,0.5210133697586643,0.48450909309957235,0.581930488716815,
 0.7195731980341132,0.13490672656745772,0.8197083499765879,0.9564418018351394,
 0.21693656343925627,0.8392534117913175,0.44396913647185177,
 0.5868499221654518,0.01396355773886504,0.9494813869250481,0.9832659322977212,
 0.9869404160159545,0.1575140789542574,0.037320660304840714,
 0.5681285540081713,0.28637080553259975,0.856780596660843,0.9107860567346528,
 0.9879890146321693,0.5372809251595903,0.1893958091643725,0.2909166189808706,
 0.12326932534982349,0.4339118172621572,0.4172313571513876,
 0.047046363329780294,0.8565849395062695,0.328454205927013,
 0.04744916777677488,0.4787774350008528,0.8082085318865443,0.641874031510136,
 0.10239244095787858,0.3219990634320081,0.9924110838872127,
 0.17211110894654247,0.5039788303865814,0.2679256782882691,0.9253949332583169,
 0.5754325032108034,0.32307878371855336,0.8894351838302005,0.5635089322074452,
 0.3820225944796687,0.524100508755273,0.06251690572032165,0.7272464453345844,
 0.1599218519397937,0.08913903032951787,0.7526551458591323,0.8293676617327491,
 0.6680024150551538,0.9143453868767651,0.8111250810791785,0.9534491860070473,
 0.18736475535661135,0.44457067658330385,0.25677774497457717,
 0.30007738327864053,0.7868774692497702,0.8934467006665143,0.636418407808811,
 0.4837851246680369,0.22176470335161957,0.13172042659754846,
 0.29258783371768793,0.605493558252994,0.4927084545119863,0.6609929485165618,
 0.9564965903877234,0.4386400380173062,0.6582186842282475,0.3878016879457624,
 0.7350940146195828,0.7197211382280916,0.40563500428430554,0.855285481153874,
 0.4647950110228869,0.6700128834356742,0.09482924120018588,0.5236854819261966,
 0.3114668259311909,0.8628136806421863,0.8693056395139506,0.7861376654282162,
 0.7822082364942063,0.4079165084428048,0.21635433541632598,0.9858591201796524,
 0.9106605829165848,0.8378588363996984,0.9613991310400859,0.2712657218029474,
 0.4875106826388007,0.36200381806943327,0.27443413107302794,
 0.4469982670359123,0.9297767935315344,0.7623975320428689,0.15114035335913,
 0.4239135817318793,0.3401125665734238,0.015273022290561156,
 0.11857514940667546,0.9171744132509753,0.391130453004343,0.46133590410245007,
 0.789657598290774,0.8759701245832472,0.7242196697125289,0.2598917492201932,
 0.8598821322035247,0.010870778666610348,0.13206807318956826,
 0.5961810541636575,0.1551178846466469,0.8007820050092033,0.8081755553929646,
 0.2216923939277069,0.6747016967749258,0.42405488789759804,0.3891323005367717,
 0.40447557950139523,0.09966852136216064,0.35651678891042937,
 0.23121909552100484,0.7564060083447042,0.32357655044861333,
 0.6563495719811689,0.4821765983864108,0.7087833654910726,0.933566846472662,
 0.41426287670468664,0.8131717733329,0.6023105093774034,0.24109051617507016,
 0.24044012000976978,0.18092508960487186,0.6106705790565722,
 0.0058675469543616465,0.11425566951094146,0.9141776246792548,
 0.5020331003103814,0.1855904332078443,0.013368989434176637,
 0.8428350911650844,0.9505710605303264,0.6411314713587356,0.2830087892416657,
 0.5131745066827659,0.7332046897746648,0.4551060240761474,0.10520373524213578,
 0.5395134337971823,0.017768310628675676,0.30570526053271097,
 0.28181226314619345,0.08230319393474528,0.056334457207568134,
 0.3189707728344734,0.43954213107398876,0.455886047453146,0.3262739138720445,
 0.7463321630940123,0.9447623052160017,0.5061261796380647,0.2021435171843804,
 0.5415849124532598,0.43379476202955913,0.1648225742765479,0.5220303338840105,
 0.7328198771645772,0.49331951337538704,0.49400826508320006,
 0.08041754411216473,0.5417998133534185,0.7847630389621949,0.7290369942053835,
 0.5401604036760272,0.8171649021590419,0.02686311378801176,0.9826005817581661,
 0.061966356905401754,0.6527237011007239,0.7718255372069698,
 0.8791250141317273,0.05700221374203518,0.027036204568179678,
 0.14531533143726194,0.545710221648027,0.2276324661424235,0.30732654407343074,
 0.5842569549493317,0.867183610034184,0.5663720368802629,0.6752082792370253,
 0.33769223878194343,0.49000923826676124,0.6022385295807055,
 0.5651367050833284,0.19756332971145307,0.19697412063112996,
 0.5762088948498724,0.1637463609608495,0.6821573601319009,0.20566981249521576,
 0.3576224231795734,0.23355928617623478,0.5918996732942554,0.9314390372557377,
 0.8170957499093014,0.6643871484931385,0.5026925771850029,0.845107596263425,
 0.5322816205279179,0.31709036446023753,0.2105880029934445,0.8165127621964821,
 0.9082104551589497,0.7345276387779411,0.3170697088417518,0.9420923605344426,
 0.653643157170137,0.17546142593580127,0.3140523419855221,0.5120106018038335,
 0.1443508559309894,0.3032413025541678,0.3055314763666166,0.48320580421138415,
 0.8472084540553037,0.07710152913309976,0.7747462180550189,
 0.33820021794620014,0.7327917977895657,0.2531896558024598,
 0.23143210718246332,0.015477582611067797,0.2631522894971602,
 0.1770874218781392,0.7338817256956945,0.45917173372789855,
 0.26911417489314915,0.6074405445757873,0.5447192003889793,
 0.23007559268292488,0.8720566032200565,0.08964864319194421,
 0.19468140863089833,0.6966730110987678,0.9822176618563118,
 0.03781918854744648,0.7670916154092253,0.5647703856201564,0.3537231871066344,
 0.4679480237964182,0.66205908820999,0.7654658920088845,0.25264763164180737,
 0.8332844700109071,0.9976773027738745,0.9530285585417799,0.6426856474192777,
 0.7325221113092869,0.4871688261103052,0.2845394633094923,0.6301474964462642,
 0.9488947132231506,0.5320172725583081,0.6676322025401349,0.6024663765450142,
 0.5842757436088252,0.38719897475283593,0.2492979695636942,0.7185848044209504,
 0.5636541123974267,0.515981981665787,0.809417410132852,0.1834766656936585,
 0.8700499458189573,0.1588656965721139,0.7596864394221489,0.27647130287716004,
 0.04131757642443368,0.6300054770027146,0.9360917329673281,0.4639293392554632,
 0.4813296588153362,0.8559883527165679,0.33896297853083335,
 0.19648768518509951,0.3014080510361701,0.6758728078896123,0.52269568132095,
 0.23614727390673929,0.8641723315610048,0.6964773606495077,0.4079572895151743,
 0.8703566080752679,0.5723539878117958,0.03372390952579585,0.3485706398138111,
 0.23422042352620054,0.2653742792755117,0.22794758233124024,
 0.9613650985215827,0.24965567744110462,0.26012241823662174,
 0.34180702509007865,0.6337562349633943,0.15514322155393057,
 0.16422260411384937,0.9247543433289868,0.86042778692272,0.12031687642722777,
 0.7144171796675213,0.24757682757498634,0.8984158281055227,0.6439795951441144,
 0.2037624654918483,0.3297287658440292,0.605212930332922,0.10433887212556603,
 0.20426191144282413,0.7130016211354655,0.4615262810336559,0.6238950451568784,
 0.6080666485949073,0.7862901364320518,0.28229699488956683,
 0.09081135759541015,0.6686978882427508,0.5585289591143869,0.5762782231623149,
 0.08247091707459175,0.6092742986162868,0.7916156917603617,0.3907122443535247,
 0.1799816424633789,0.055497272763818106,0.09986299147036737,
 0.27662430874885036,0.9142501823556853,0.9456655618105988,0.2583271172494831,
 0.41015030854586,0.9499357455094302,0.5402650252716995,0.285791561256839,
 0.8964927701763613,0.784377795096529,0.9577608460568462,0.7687412616327646,
 0.7797754926835438,0.5460608651997383,0.4366607670464874,0.9700376809520459,
 0.321179711835173,0.3311219839460041,0.9633155336848015,0.10795266800166992,
 0.1186964181079182,0.029430852280747932,0.009072978374991214,
 0.9133184472197716,0.10943956753537243,0.3627008469520304,
 0.14648470662956825,0.8459167999718714,0.005333962968006478,
 0.7080547481348535,0.8946568552039424,0.6659037697670227,0.6869089488979121,
 0.6680163976106529,0.8286342247841896,0.02807527782867092,0.5843939435059137,
 0.997999986054216,0.7933860265266113,0.9779067269782322,0.6273227808573397,
 0.6050066022626268,0.5286233469295439,0.5920835369810128,0.8206379594326308,
 0.6269956693832271,0.7006640224392977,0.1765206497452827,0.7422285262984715,
 0.9627049908499223,0.9961061720481318,0.21063148135037169,0.2234151163332706,
 0.5028688061327374,0.7676047387554701,0.7884341020021163,0.6283182075427118,
 0.07623481261279375,0.7307955288471122,0.049787776898682745,
 0.7182894377660665,0.973814659028194,0.0904698761019811,0.5313750453343403,
 0.5366105554384221,0.6945490849255976,0.9857078680790164,0.05715734726213351,
 0.7791270832281545,0.13544822633410536,0.8892498320161266,0.7449782231026643,
 0.6884410973246008,0.2510175869387432,0.8512529887193421,0.4955865031025761,
 0.7937482649078736,0.29091586206087117,0.7340636615526822,0.7741938646903935,
 0.9104445653099184,0.22337578815005,0.18075298677946894,0.20642459629739773,
 0.01302680489221586,0.9584469055504937,0.19851074136564295,0.702127098957442,
 0.5466680158419854,0.4430308847821328,0.520644145925601,0.23500461856690769,
 0.5080084357717427,0.23056684049631038,0.6976294408412932,
 0.22703191738256745,0.225349369369368,0.4772882508482016,0.5680178780109517,
 0.7508218040681613,0.004445176301722031,0.7493570078133021,
 0.23998857488724978,0.4143341569270329,0.3442953806889244,0.6599241679509305,
 0.6253267434041346,0.7147469121645655,0.07197074269029802,
 0.056164982311886824,0.20222920342843798,0.5325084482621915,
 0.9676701087527515,0.9739867094721186,0.4532869921308298,0.9359186675314428,
 0.4562080497512748,0.8389341458461133,0.8658694740391001,0.14619155130873063,
 0.6566266827253124,0.327744707239748,0.20171645722419562,0.8375362297727716,
 0.5999980249372934,0.6877219898278304,0.654547463885427,0.9827900836500567,
 0.120271042287874,0.633095237328541,0.2337900978890024,0.678804687607709,
 0.1495495272833257,0.7542388166295175,0.064562779444445,0.11477190746292643,
 0.2094809141403896,0.22883609237067393,0.7990914071190188,0.9194736681261535,
 0.07931777774285564,0.4809037278420977,0.24362194396545367,
 0.6080442917830106,0.9845921067871871,0.5571912167102734,0.27284903075449674,
 0.6063188272273317,0.48204727986357354,0.8779073125536663,
 0.17717924201578183,0.45579690262505923,0.7278655048258948,
 0.21656752301363358,0.49176319684243075,0.7704278640861512,
 0.4565621559589649,0.5971509200795868,0.9164194535059237,0.7601354348835341,
 0.3986552377667336,0.47276092240032686,0.6358801975397204,0.9749871360339288,
 0.9301238624576494,0.6326872325245305,0.02538027940587262,
 0.18749905196543148,0.0749857936776589,0.016588142190269606,
 0.4020313211404961,0.042588329485662935,0.7859663599511096,
 0.8118871712401838,0.3484958088316257,0.2003404199107751,0.174072543542092,
 0.13655668431397305,0.23868795285618072,0.6664670616740912,
 0.9388801503147349,0.8423569939771738,0.24504865748874638,0.3647675543624691,
 0.6369079698974397,0.0886076137962053,0.5295220592922638,0.27392251030666637,
 0.4744947791576831,0.8869277404130927,0.10344961825608612,0.8195776694949216,
 0.9026966553083329,0.04829949710431081,0.7318879349307971,0.854097485782065,
 0.03618818343508634,0.7819281360149801,0.28837460034276674,
 0.27859017900873717,0.8688315584429022,0.04876402634878274,
 0.9766119266634871,0.6780092115570973,0.08504263212107488,0.6110268499881875,
 0.16570128979199428,0.4502528955976124,0.2242422202397747,
 0.01518871947736955,0.06703126489095468,0.8540729020353623,
 0.3653745614266589,0.230183907797058,0.34282931063706945,0.8083959523657891,
 0.2517837249875273,0.3920958203282099,0.6399255239132821,0.7899938908532212,
 0.46670945707097977,0.267128927987651,0.568834397042575,0.8734404438347305,
 0.09854453932942653,0.2344533680599259,0.31069265843518945,
 0.2627551904060207,0.6151415035574149,0.4041846068392755,0.2551579772461514,
 0.5341947673293755,0.17673183598397646,0.581607717951647,0.3422014918638582,
 0.05451122346873982,0.6887272691820144,0.2093504113721114,
 0.17211725964328817,0.4057044503764943,0.969689607030456,0.48707321040197415,
 0.503588404499832,0.929860864541388,0.04952943634011786,0.2459523533473964,
 0.30750612285507795,0.7693577278417653,0.893814570493768,0.762923343596935,
 0.2907593579697434,0.545714690922814,0.6903588789961087,0.11093377381545966,
 0.8445788520673327,0.37112207990567336,0.956675764648812,0.956792488312198,
 0.7600746695984878,0.7354065666891614,0.6576968427193903,0.5359026506025153,
 0.3243984850102377,0.8432733397047469,0.28002661774215387,0.8414948109053817,
 0.17521268969076365,0.4676429436511429,0.5253737098764102,0.8649569024559199,
 0.38370432571118696,0.8673359958943819,0.9898551786591063,
 0.06144899686453431,0.08727231143896119,0.33250510036660164,
 0.05006235842238793,0.31252401287154574,0.07910323477758541,
 0.08429493721304504,0.4492130661983179,0.7082533309254855,0.8100107101363065,
 0.6647528124107288,0.09545187919513431,0.9413210707578281,0.4833920082342843,
 0.8259539542796874,0.7196022737503893,0.05112580855221682,0.282552490234277,
 0.6781075027285203,0.04805353467485307,0.9386399580830314,0.8993344443682605,
 0.02966365455984843,0.6617468357184266,0.29310188007181814,
 0.6380575436733924,0.06576677678035858,0.9115994492121459,0.1325824244646916,
 0.3468919821431503,0.8711597731734189,0.05761895778114301,0.5689232582867295,
 0.5617701372912992,0.8766709124260388,0.39474657844741556,0.8680867499076166,
 0.3606731786831938,0.9633080879176981,0.18954786891507058,0.4309549275734945,
 0.5712205281117799,0.8568193009720813,0.3199546415910677,0.3010352693569107,
 0.4944180133145091,0.27092874361020725,0.9792611314418737,
 0.49369032722756245,0.01463207778365827,0.7709526548559749,
 0.9532404987338772,0.7775169465235073,0.8770104781020824,0.2876249065661063,
 0.1293141473530941,0.6244578119602756,0.7339343589772641,0.8750564602181412,
 0.8219805175087043,0.8646308119144455,0.4240925630573845,0.12317148305106662,
 0.8414596788704561,0.5704541055818054,0.4559832187191811,0.08261532499869972,
 0.3927605740955977,0.8906259586087472,0.43276520161104015,0.7376462339608839,
 0.9199171562771647,0.04190286553570999,0.8089092270831596,0.4308190716362448,
 0.7617457382269074,0.24559898432111038,0.5856795726693782,0.3673996565626546,
 0.27909695437956217,0.08120869697664146,0.6089210769236041,
 0.7130366324439172,0.7378497336569375,0.43564197105503943,0.5224766433314505,
 0.01885531223896708,0.5194172642942442,0.8144108743336145,0.6856152970016135,
 0.43477238405816676,0.04379989963415687,0.5082380513628659,
 0.5351473937001504,0.8722714820653296,0.7847624909256332,0.27653634885036915,
 0.5364665331628242,0.1419245809815114,0.6914463910651489,0.47618216075546216,
 0.03548278655855597,0.9471520860224939,0.14045431826502608,
 0.5258174693192226,0.03602883478457153,0.8199197604322218,0.5119180639116402,
 0.44805716361187153,0.5816475681123991,0.3851178110986977,
 0.05105508591703267,0.8431883578427819,0.6108719713884583,0.8704134750619748,
 0.34830926787490657,0.8209723747421604,0.4141043842010075,0.6881065470110463,
 0.9568213446598948,0.6769177778118913,0.1851809620332837,0.1430682603183111,
 0.659866981440494,0.2105277786609372,0.4190157831860464,0.3396665733511597,
 0.2295139429703188,0.08125774663783814,0.8299304408371547,0.9233620190710996,
 0.7162230338584532,0.34933247772162535,0.018942609537848543,
 0.19786326409409627,0.46325528327871146,0.36400120571148287,
 0.9398866166352774,0.5617334787722775,0.6409600343202336,0.5249671127418253,
 0.36307209985124567,0.15718125891042378,0.8483994522264217,
 0.4814444849060263,0.48051743205402586,0.14142722279525255,
 0.7671085189725059,0.817260551500551,0.8304145064011562,0.10230559690064311,
 0.96912099885762,0.0028484998649458415,0.40203446997185344,
 0.9616856931580289,0.327954664803184,0.1877900328409695,0.26251109570693953,
 0.6469292127230162,0.6421324138185608,0.29182098822051084,
 0.24391900269116917,0.6045881755203706,0.9543009840756577,0.2133075551500857,
 0.25704610613277323,0.7598032514456866,0.6658426212294619,0.7310098182983916,
 0.364152285603221,0.6616213451689311,0.5032487883041566,0.5715112756281691,
 0.04617784783721568,0.7967449656600446,0.5439416153389779,0.6107258292761506,
 0.23865741169685584,0.9272057838576334,0.12854517684246525,
 0.19010146695627483,0.1660685721583507,0.2811762416700262,0.9828130143379491,
 0.9234314686097124,0.2964774205645635,0.7408208335838984,0.44660722683873155,
 0.7402932003103843,0.593411029713822,0.9272261350645935,0.5261730327710106,
 0.39114735642959175,0.05275703851242375,0.38573169914949357,
 0.5802621033976625,0.03885444711704489,0.407187140311022,0.7739559164707798,
 0.7514286525316991,0.923252352315517,0.677395339477617,0.12265684292382628,
 0.2853881731578669,0.18067367245810573,0.970629963416449,0.4034859688169903,
 0.8132845290210642,0.9267262143600078,0.1569604076963892,0.8349135235390581,
 0.8995682682459382,0.8231587736684123,0.5741527525409366,
 0.027524005841241372,0.4331434994187333,0.08451150294790688,
 0.09629237274356806,0.5702229445059165,0.649998712676702,0.2568505343760906,
 0.5752182635195775,0.322126440973884,0.664793446140516,0.8752696948685572,
 0.4904138503813823,0.09682743338204292,0.05730344595804193,
 0.3150800121443623,0.13528166828019694,0.6801065084074809,0.9441543791255452,
 0.7727909312026703,0.46904613230909875,0.4360026850896471,0.8611412784226496,
 0.1913137747054423,0.6259960368519598,0.10424028957697606,0.9013676243742096,
 0.9116176216549863,0.772454352317029,0.4453507025116553,0.04734800024215058,
 0.613691632791191,0.30076290117575033,0.604580080088741,0.5158283434716741,
 0.6328569880925746,0.8743573010656555,0.6157775435149666,0.33627761114130317,
 0.35836162141633854,0.7058771278398316,0.13057413447149901,
 0.12765761107580254,0.2913789904753519,0.2025330183960129,
 0.33763913123659617,0.8823357876858591,0.10902052128194817,
 0.5101181144059237,0.08285588066197591,0.4648382326475483,0.5391952833186394,
 0.11360387338711642,0.4686096539041237,0.9048719546669133,0.5124902835264743,
 0.004077087668642898,0.9509070956567451,0.8736221036490418,
 0.18906320184935077,0.20229778982176883,0.8277804343436268,
 0.3259994047990682,0.3585150233641192,0.8248497703781853,0.1875747209998775,
 0.08734462051252267,0.02174490844193666,0.47174239746211144,0.4934660611063,
 0.644905769478257,0.4538415052840953,0.5834688953844044,0.7341164960358371,
 0.09282195414943895,0.8929577464250007,0.09776476618404284,
 0.14875022902553336,0.056681085107726736,0.49678786357955895,
 0.6825935093433086,0.5863600503909256,0.5064430676470805,0.6678352061784749,
 0.8401201184115363,0.8340643729489899,0.08402830684127927,0.7536920088970588,
 0.20589533685649497,0.15299111339666993,0.6049768821283465,
 0.0479093227270051,0.6772770847111902,0.5733087633066452,
 0.009529884543972522,0.15056161542862923,0.16799788656594805,
 0.36146940982435094,0.6212447829326748,0.26437216098979543,
 0.5186203444975636,0.587097612944901,0.7057863369078661,0.4976295915846001,
 0.36773827413786675,0.15334623633357702,0.1691247211032192,
 0.9182395645453839,0.09005506858718593,0.15420226919824775,
 0.8988875161925548,0.5188037099107974,0.18704482173708048,0.6902254995495405,
 0.9956340361538265,0.4282455033994388,0.6915500054052248,0.730158849725771,
 0.8325006215159686,0.42765126147116506,0.10773033251787889,
 0.09062259735438838,0.21065892578121403,0.3361216222724742,
 0.7286599104590119,0.7652577831097087,0.0768163502650715,0.2586491678252951,
 0.23408291451108043,0.5234441029994247,0.15599996048227127,
 0.3735377327740539,0.5056481891260765,0.31122958958984026,0.8162355294190862,
 0.9147603944434627,0.9406792427825411,0.7187086770838599,0.6107154415828371,
 0.21456750592332452,0.12020675168308081,0.4518019815590736,
 0.23751135927577294,0.45648887851156084,0.09214720367197071,
 0.09103336313866572,0.8070973597058823,0.7046776339638336,0.615632573681469,
 0.46129711961475406,0.3083487128248732,0.04851412216785578,
 0.7753035464606647,0.18703344987246528,0.7478614392637303,0.2711820478456015,
 0.9928551005610726,0.28185853560000074,0.49156006454901346,
 0.9315254176613413,0.35319074324072974,0.32360598994073697,
 0.24790975023237205,0.24670908246365675,0.8455322155655252,
 0.3806948751398218,0.8563278994341983,0.49475557177594365,0.7845275551402711,
 0.6071083930130027,0.567144962034992,0.7529646757912534,0.02115054977716191,
 0.6399871467011339,0.4183045426760398,0.9111568391082954,0.8131773253820472,
 0.569101880110549,0.797385284736331,0.10273971980118235,0.014536601051720055,
 0.3334772827770641,0.8062664759297122,0.47533343882646395,0.5280337849895245,
 0.5180449089806526,0.8918038814600457,0.05494244338279897,0.2318412328022691,
 0.13264960582083263,0.31686836136282115,0.11602368763174553,
 0.8677137456819959,0.4281939555002313,0.6151413582725533,0.6702947354471382,
 0.45193650153957343,0.658718772756451,0.9552448417097752,0.24896430154309013,
 0.5276040664755709,0.14047735987457544,0.06648104342478012,
 0.17134871112998706,0.27910300800365406,0.43154415649030853,
 0.8625569018243747,0.15241916474487516,0.18928450007686348,
 0.7593053169671445,0.022131562516995218,0.642527594075152,
 0.19990230488905691,0.8347499101572047,0.3374027698405717,
 0.29705813612479537,0.331120141943654,0.28781927177099154,0.4510727842323071,
 0.7896095293842404,0.6824579864340248,0.49559627905568027,0.6784075634732316,
 0.25389258112662205,0.12335251199277086,0.6437547366649973,
 0.5541005257006626,0.14869455870589,0.07989701914248581,0.19979820176829066,
 0.31993528583097797,0.2001510776015144,0.06330726810752296,
 0.1344023681158295,0.9247186316677279,0.5012518275572947,0.05660738355211603,
 0.7259605170077873,0.7326747485357507,0.2579342897414366,0.6943388256162824,
 0.7600394035080132,0.49219754424020445,0.40826152216832234,
 0.00337853147842182,0.24921176130346412,0.053359395526258235,
 0.4158502893864714,0.7875376794615915,0.8627602046783241,
 0.021324566189058536,0.5554544596516415,0.18340644636896175,
 0.9940761540184496,0.02986618339900471,0.7806620085105417,
 0.38466378664108214,0.8229828550268333,0.5875518547182176,0.5369836862638196,
 0.44158358523665575,0.21207938882752964,0.5758478468417106,
 0.8210697925395775,0.3627744744946738,0.6984933117752758,0.632918225364589,
 0.8487970456133416,0.05533580574452346,0.40949392456853295,
 0.4633787634218062,0.6466552697765885,0.6497003940391396,0.6061641822204054,
 0.11358561491460095,0.12541301966599938,0.8281884753255643,
 0.6713240996652801,0.9621457477489223,0.7431657378681902,0.616849007310263,
 0.6496575214674957,0.7831011410289737,0.40668859692144044,0.2139734816273282,
 0.21580651319640565,0.6126183701116904,0.11442039462318243,
 0.8412327495730529,0.46144446435409925,0.8466972679352417,
 0.019775688127532387,0.948981110554414,0.9271300561816694,0.3113396239550327,
 0.3668392443163532,0.46857344896505293,0.8166419318868103,0.6681215689224898,
 0.6466668302564645,0.23565864611626408,0.5192942105949278,0.2544174053284265,
 0.4951202893433244,0.1492170145751155,0.45117900912894093,0.6938821206602044,
 0.10689314424450869,0.843743050427684,0.34539498625290177,
 0.20890753025561337,0.43547934443539293,0.5327009513891262,
 0.18182444981740453,0.6414493145189288,0.5397363335367251,0.9943050737443763,
 0.6027082817014624,0.708335389711697,0.4823808547037569,0.8222421389189049,
 0.44811727615581887,0.08097827373062993,0.9166015067272443,
 0.8532750494742124,0.36083462561317936,0.039665611584999816,
 0.2976317439406595,0.43722964500321615,0.8354240748526223,0.4367603814488077,
 0.9719884494643184,0.26008324804978544,0.28687981823769004,
 0.4955691100990899,0.8284775847526141,0.8465042619785895,0.3289501277885982,
 0.46329281949981627,0.05820465725042079,0.419456057709094,0.68861414452356,
 0.6423729180540697,0.1636819387902375,0.9677224401984802,0.22960920615894787,
 0.49751399656140594,0.19697983749578096,0.7603877792338365,
 0.3431476091612109,0.17770487027634485,0.0864032299224251,0.7038062517494236,
 0.42157179500636077,0.8003001117175185,0.6784736463970021,0.6102247040126866,
 0.7585767123185764,0.5743066185655593,0.6965368574439985,0.9723747028406013,
 0.19479672625859512,0.09326812379779237,0.16927877147117476,
 0.6373505424301743,0.7104192387794133,0.6714649175697198,0.13544140590915266,
 0.9025249526027036,0.43524997697929535,0.2118339152840789,0.7697575155142347,
 0.5508852671002016,0.8213892296254115,0.9017884084250728,0.23456846726270697,
 0.6879437201063893,0.49999405227892146,0.8576251663475664,
 0.47150944735910594,0.4565527484469415,0.3627662966809603,0.7106890890642654,
 0.9783713297908712,0.0854995384282129,0.6015041962240196,0.2748357760568392,
 0.43617586088087057,0.1673906781815765,0.8522586786539273,0.9403796557472857,
 0.3786682270936761,0.23935455171769604,0.27694816030541347,
 0.25152229457159236,0.8956484072020237,0.7355678871296616,0.8440732907625935,
 0.7834319990128776,0.47533642232297735,0.6137157129890196,0.5421232530446182,
 0.1538169442394638,0.6122946631907084,0.486869753068788,0.6115445565686307,
 0.11618827437768586,0.2716184334113053,0.117533205610872,0.7233129237181625,
 0.8025072685488125,0.01164934248583227,0.1528873388354577,0.1740857620284778,
 0.4691999519185108,0.00516317056432769,0.9895529941498584,
 0.035456914032283016,0.4877053701201878,0.43521713125648176,
 0.8710370075259762,0.4989766632597462,0.9795333173935155,0.3695831724425791,
 0.520968687540294,0.3959743216714766,0.5580232138668857,0.15971643155320414,
 0.49321118856503066,0.9972483408249024,0.6579789415895457,0.6017837827809525,
 0.7561175186451626,0.37607806915049957,0.5142028278076363,0.8195308123000722,
 0.8148934701726239,0.08431186237751132,0.42279630852159933,
 0.47557257820937326,0.45926059936962926,0.8552860213558767,0.926352803367664,
 0.9662717008598698,0.5746683671415665,0.08267682608920879,0.6042194680969757,
 0.5615397492040092,0.4110783103346849,0.4165217274383046,0.8349951625367434,
 0.5339041057273379,0.8857871824522443,0.3556127245126125,0.9287168065581201,
 0.44097699404327406,0.5154405029078161,0.17936363389205134,
 0.6775384587111525,0.4153865574838219,0.8708838210308252,0.7681234153805983,
 0.2769628504288212,0.641792371021098,0.008612572579112743,
 0.16567590896618445,0.03179622844213714,0.4652630619026511,
 0.2663772556453684,0.9852706488914759,0.7697652136749391,0.42176586291242024,
 0.7486505941822758,0.4686140572978308,0.5785855512991307,0.8626573603776968,
 0.07793891120277152,0.42154798442644714,0.49502443195486956,
 0.12457378906261862,0.15608012564236795,0.6193155300996295,
 0.026081161410318354,0.8232779579235291,0.3494647700073389,
 0.7633514963665596,0.7505546449424668,0.1399994435791898,0.7902303264460565,
 0.865162089914421,0.5241355133575872,0.9436964365342285,0.5832350795473892,
 0.10400905465856036,0.9892993114979436,0.7530627224942837,0.3898532998793359,
 0.07963341137537494,0.758534679676814,0.42121775629662905,0.9185133202834965,
 0.9805202650773502,0.9058281664069194,0.732694659473825,0.5538204120338858,
 0.8455386115049274,0.05096653506404181,0.44717374873806914,
 0.005125378812236625,0.9097976390324661,0.3363822606950426,
 0.9183270116039164,0.26670206297638077,0.9291061444924077,
 0.26769457580246825,0.0563472874583677,0.6112020977535069,
 0.05009110735691791,0.3622309632256385,0.7565113040253105,0.5374356958734853,
 0.675379439095839,0.22349682183824637,0.9369536076326954,0.6587256016769312,
 0.3193624626569831,0.24647162997306848,0.18181170559812387,
 0.8451357063520206,0.24331705678158633,0.09976882633427975,
 0.8255115539759612,0.14109747638062864,0.7523687301594735,0.5566180869306178,
 0.6453923349790868,0.3744153584252681,0.20010439965385707,0.6156725554536306,
 0.817122287222108,0.7590981348285579,0.7565838437740149,0.4678653337567179,
 0.4952614104260876,0.08719702772242444,0.20443176761629744,
 0.6764535095377897,0.006415675757161732,0.6229125301633458,
 0.1102786035433525,0.8137200401204622,0.16420425122348303,
 0.19373714460997116,0.8113321754175337,0.7171184684808326,
 0.008865439983251955,0.22711325582760677,0.2937777289112815,
 0.7568676761831807,0.8473438919221672,0.340810097406791,0.2418193980000407,
 0.6395632028700456,0.47371672676961396,0.8048387455500048,0.9848763247675809,
 0.5232517766375788,0.9566768022323517,0.7578186823615377,0.04968296725811738,
 0.44074687831747017,0.5020894589681536,0.5143075109208776,0.4995520181020755,
 0.5351787684862326,0.9219835675913314,0.8309140218069878,0.5661451579607706,
 0.9846233954154922,0.39033420644420613,0.2976105790001464,0.7277476812754001,
 0.8647673193667367,0.2823041509529908,0.8373345026976271,0.4359036699317389,
 0.24458504800862724,0.5393098437159376,0.1458569529256869,0.7250954594448245,
 0.5922422434974578,0.3753697486182044,0.6376434868792441,0.20451463538111037,
 0.8665307567231654,0.3186226255914507,0.1328942535109574,
 0.012062726151877201,0.4801867880889168,0.7583926564294772,
 0.1409841530153042,0.4333437884364286,0.3543268586637869,0.03283337810051479,
 0.31540877578221993,0.9319934972204218,0.9840086220705996,0.7663113279830372,
 0.4152943036603225,0.5667760956501842,0.5691091960242863,0.43946864376428885,
 0.8771800661894384,0.617419092112991,0.32543215680760573,0.8252319414210901,
 0.8101857947217878,0.421092743261146,0.29933103902445934,0.7534990928794617,
 0.8247457534254263,0.17255779465232768,0.17717382103285362,
 0.12817774133632454,0.641235842353863,0.602107719544105,0.9167916757998328,
 0.25870139123976044,0.31918158348146575,0.8991864819242066,
 0.21800380547278064,0.394448874779074,0.16142476068186562,
 0.36039140013766113,0.8265586166292791,0.3281826858765001,0.9514743844892611,
 0.31639175340400827,0.6400055342648598,0.1678694628723426,0.5981665344678043,
 0.30227627970794035,0.978460241654215,0.621177505182616,0.836927122866863,
 0.9225158109892154,0.3531391132169359,0.7644469138385714,0.36557729778374015,
 0.9239590987921,0.6335767226223765,0.9318266709668094,0.2926479315361292,
 0.2690699856611387,0.04385646030201984,0.2166961386061379,
 0.07077039570625188,0.7363964212820523,0.36369848115813097,
 0.024940390085440578,0.2090621891727258,0.15504629229230626,
 0.6853246830876031,0.7977636286368893,0.7586376022539124,0.20166195231947837,
 0.9558974380762822,0.26324474816053844,0.0797377886370757,0.8031369891067692,
 0.4009764667319694,0.6866767209613791,0.7078999596773556,0.9045807045369061,
 0.9042192989023923,0.8860879973561948,0.1634159160104085,0.04686458322048892,
 0.6644778245441031,0.07685703950215506,0.18371774950359787,0.336199800046702,
 0.12144264508857572,0.19899115437113468,0.745264313061515,0.7805818196705491,
 0.021670153800342984,0.26870863885165563,0.27977133847652724,
 0.8809382398595118,0.6969740137137178,0.2886169869452566,0.4915758904625902,
 0.9893840092978294,0.8869964262337112,0.3171659642619855,0.06769226635172432,
 0.4248366404029358,0.7818705992300969,0.5455524139598706,0.3386901174638679,
 0.2510726233529278,0.43875818321072124,0.07608723523135885,
 0.7210560576751386,0.9823345133672263,0.2627690353814176,0.9103397669817644,
 0.8589925646772654,0.82427688716214,0.6198386258323925,0.47467163100063037,
 0.3252431341666365,0.32363103451667496,0.029630787331212494,
 0.7515983356650976,0.3339280451598283,0.7240403811299468,0.15609873024725873,
 0.8187999671382515,0.7921005564935167,0.4894497975504216,0.9309634735335443,
 0.9013613229056967,0.22678842630817597,0.7664042507413309,0.683288639087064,
 0.8031739569228019,0.4061073303203263,0.4669134295967716,0.8709550815030845,
 0.8877083049207948,0.8621867171225752,0.9429243511487274,0.45574538272811427,
 0.15233951298376303,0.09450092101330376,0.018293934776634546,
 0.48618786903555533,0.24239582504639956,0.1801369171702043,0.6762555747497,
 0.4152446153601127,0.2952650186346226,0.0618248754632611,0.4076270943840301,
 0.2671536271811302,0.6770155477668536,0.555031567411438,0.9166289004182662,
 0.44419576222142265,0.6766320389978677,0.8189375490978034,0.4712163306947603,
 0.21344370337209773,0.9450678517545615,0.5277218114637932,0.8636896480616519,
 0.7754010007505816,0.6726510714767462,0.959637784580707,0.4088864338401852,
 0.8597987415173306,0.8289120007069722,0.13913940126150326,0.7682845367802117,
 0.6193391192742574,0.16200269137776335,0.3007666375058151,0.4374130077714805,
 0.3737552518990651,0.2951142580732933,0.9967542537248426,0.507404207028046,
 0.6922439263234677,0.5145797250704274,0.8947435484388607,0.16250421684783256,
 0.9840545992499238,0.8625154053633054,0.9382157214822178,0.7581793246654845,
 0.6148264144120639,0.35936157898773524,0.721900612725075,0.12653553202304646,
 0.7329004099613489,0.35102148992609894,0.9637282030574715,0.9275363230017692,
 0.7022434381762264,0.9108993389148646,0.6277353822842966,0.4796805213122739,
 0.051194510610313104,0.011124112382033702,0.4716035623428405,
 0.5694863763771671,0.7627349565690951,0.8833966422877825,0.2848641188289387,
 0.452188675752659,0.7217048646407965,0.8586680841215558,0.21615284106697408,
 0.9684435780482348,0.7152346514754053,0.4667565884544115,0.09289087157369036,
 0.19339393049612252,0.9221878522296654,0.1274458202090013,
 0.16215499684716184,0.7945619293759641,0.16616025922755495,
 0.13194565165941818,0.5275573829178805,0.7707782280000111,0.8003311009990035,
 0.39009003866603564,0.06409167183304176,0.2335550855974715,
 0.21986816586448554,0.5361376863982107,0.34387524201738684,
 0.9519377152772637,0.3651142452989177,0.8783118591581198,0.04373412225320672,
 0.6307852519085377,0.2947047965135785,0.6488645170814916,0.1579240308161418,
 0.599914942802661,0.9899955642476388,0.22544578698008455,0.1720490734362239,
 0.00153376060545507,0.9674460762132727,0.8812038244822715,0.9882971946720213,
 0.25101867577748915,0.9971635315056883,0.34412696547725863,
 0.4231573744268451,0.922123855327137,0.1438650472192321,0.7717284534500509,
 0.3868460058281774,0.19249981264291882,0.3324614157906497,
 0.47311868198995655,0.35376953563778346,0.03835757750518576,
 0.3192564263139862,0.2271096097938783,0.2508413061562411,0.3897179227856633,
 0.23171677380613165,0.8150387030613716,0.9831982001963391,0.7776829197547551,
 0.9341929795392299,0.8023617525153085,0.624304127050902,0.2456797917932606,
 0.6188924445577212,0.6959268023975733,0.37135552216342493,0.7948163911212806,
 0.7098343159662963,0.9215409100716013,0.10033548776020995,
 0.30565674885522653,0.3485629112946689,0.1559253578770078,
 0.13320442079873795,0.2802045838635837,0.24156229855630962,
 0.1361988064143087,0.16991657075490751,0.5737013411468024,0.5517802652175312,
 0.06525470330450456,0.6241138591902278,0.5720665560629815,0.7121800466368994,
 0.9103996059602044,0.2908227157845421,0.5613822446604755,0.2914426899580822,
 0.33030426786332834,0.9184551135568881,0.4173961272421748,
 0.41858998755324195,0.7896202916959684,0.425146603211645,0.20920216329192376,
 0.569877929070741,0.3063906532495586,0.5970140492804253,0.772383231120769,
 0.4841221237418465,0.03090944421638986,0.7359627055517441,0.3283122080609099,
 0.09791978905049581,0.8086788217631802,0.7548999241107395,0.584816260118016,
 0.02555781955969627,0.0316254595082186,0.05203862563624573,
 0.48032697165636595,0.6428830005543269,0.2930417791173239,0.947824663948709,
 0.407060111285096,0.9516730359044534,0.6853223072496627,0.4839577453378735,
 0.9666113070296821,0.3519901655116814,0.7037155962233496,0.6437499138002494,
 0.507508545622453,0.30839036612751425,0.694394516809661,0.18115146629510304,
 0.3902974276545236,0.18744734101127902,0.8984926542813874,0.8897348009465266,
 0.5302415880586345,0.38420788846283016,0.8020848703513987,0.5789608671679948,
 0.21216393409665457,0.6774111781865534,0.3867190778375663,0.8103158692379822,
 0.7428902784400869,0.13647018715451087,0.5408044697032814,0.5824380744636063,
 0.8238754775694539,0.20040044655095823,0.7725274732828732,
 0.19761141020056683,0.5295749092244763,0.4728783034544477,0.529697345519641,
 0.9876819986485047,0.8989972943163077,0.737984145802894,0.40864206450987617,
 0.7109358544316345,0.41795754819421316,0.366437050177049,0.5983448287766986,
 0.9097891917334152,0.9884394176439797,0.38555043176997705,0.8401078059951713,
 0.15510959373348177,0.8252158751413671,0.6288920440829789,
 0.004274038356466003,0.4077068232133382,0.86102263146215,0.488950842954057,
 0.8085102575646663,0.8998660985338649,0.7284625974025214,0.9697640201884383,
 0.8814679810824537,0.9789116561145361,0.5036025226365217,0.21005444771126358,
 0.30612077559839557,0.5832730789903715,0.9974313161624581,0.7789661159905171,
 0.2564806553835822,0.06788934649379841,0.02172822586602574,
 0.7424772153462454,0.6227583667948731,0.8467761211661716,
 0.033071578034894245,0.530389099484236,0.5052895736814275,0.3500786717075224,
 0.8940638666981447,0.7202768566352763,0.2803766229627781,0.5923887162226829,
 0.30212495117180205,0.4924045120169547,0.4217722830964932,
 0.43412696314825794,0.0211626196411574,0.8133519873974803,0.8739954597991707,
 0.15581677152000073,0.25556323497673294,0.6692007954424558,
 0.22010989486094767,0.6086341560740804,0.8954801501866062,
 0.20482481433704225,0.023162329678366844,0.803545321028732,
 0.09737128147498564,0.9288177539505316,0.6216997008594967,0.7003578604145365,
 0.15974837054638968,0.5730149120219964,0.1802397732934352,
 0.16746118858275005,0.4546174025851387,0.40151549361366057,
 0.13037063066579813,0.31018221635401244,0.15026907060207584,
 0.42883163705606886,0.5510276351911424,0.05494105270733829,
 0.3315135986672436,0.5959427114740021,0.39110559753933594,0.885223870212867,
 0.6221583917504575,0.41669792941292727,0.1737269844209942,0.8432941124619865,
 0.07840370068312719,0.542563422406416,0.32714850766587555,0.2820211342327872,
 0.09091191586499725,0.1219751076118557,0.7853659986581111,0.7123852615392902,
 0.06076384235795762,0.6459596579565988,0.36420218411626715,
 0.5546663145619899,0.38857893936898213,0.08013322192829575,
 0.1869186627001358,0.7600380807609671,0.263523692034659,0.2609661051512887,
 0.9938179115787584,0.6825518309131503,0.05897977050363923,
 0.24161172878908277,0.9978129088851808,0.9740641301792176,0.5005635022184355,
 0.5243154495662067,0.38507728979555444,0.27408309303953027,
 0.7446658379630664,0.005154343480319512,0.006763213319770811,
 0.18877980535065397,0.9387005862557761,0.26869853077798234,0.799687283894611,
 0.12914211248660568,0.7853818916112302,0.8380545695009087,0.7225258374639532,
 0.6041608486431027,0.017605525160108715,0.5939284979706779,
 0.5385013827074236,0.3227512962000757,0.15419910254535363,0.6664842204658186,
 0.1095186672484082,0.5544574138967331,0.06841696319144441,0.7572996329766318,
 0.7952433009040801,0.16395725173900044,0.8684517787664522,
 0.015048545792005985,0.3421718998019153,0.8251723271871731,
 0.008004621104785725,0.9441678242473674,0.04526529032761073,
 0.39017755120359343,0.2608337905479584,0.5640237577109799,0.6796024061513251,
 0.5776570479458804,0.5355659393236606,0.5499500702795991,0.07280878552959058,
 0.2789161489285843,0.8161702084284116,0.71413062269859,0.49587376427719565,
 0.07698498414027932,0.8324758361416432,0.22633577966403662,
 0.9445092771881692,0.04902691110871249,0.9308982219279747,0.4658135424615526,
 0.005117224020245281,0.871778582220331,0.4115386271283872,0.3315601918232477,
 0.5573521663885913,0.5937402791207564,0.290514900029061,0.8469393215214172,
 0.4918940243532718,0.7483492463545701,0.8897622007513712,0.4812399107400668,
 0.5779076056473398,0.29244675267829545,0.5741819400282537,
 0.20797985002950026,0.49070406059434046,0.14465852192062423,
 0.27073175905513636,0.16782965130229255,0.7288872894657618,
 0.14624436207381553,0.8223919958151364,0.3055631017299004,0.4437619496540197,
 0.3842671895949966,0.37305971096068025,0.5599505786913066,
 0.07332089644029505,0.04081345542694281,0.9742061043141172,
 0.11788873968114111,0.9998604463324248,0.45131000116861386,
 0.13116104126887507,0.6631889864657116,0.3033095689211913,0.196516933873788,
 0.14935261152933843,0.44997150238970796,0.559065919578177,0.811209251490508,
 0.3627134879363535,0.6410623269458806,0.04662903843198152,
 0.25431696107765434,0.6445795840588846,0.7310994606072762,
 6.875075078844883e-4,0.7558114045036999,0.0695718852390641,
 0.002783760115898204,0.2991666900589375,0.7643641036854343,
 0.8823139122021671,0.3346231622893068,0.012506354229275374,
 0.03046236382400913,0.6753545382536107,0.801444998441705,0.5421285483662808,
 0.8061572573384325,0.36364183285281937,0.2721617303842284,
 0.23535747925803951,0.013386915205237382,0.9128440546747155,
 0.4513498749496868,0.301833266160765,0.37804214816061465,0.8958738309356779,
 0.36147459929074843,0.10946209838940046,0.03158230307614285,
 0.4794590173416492,0.031965711276162434,0.6407098880460014,
 0.9152548507299444,0.029943573887192487,0.6248010845481008,
 0.9346781167684983,0.46122242515390044,0.4986521562768358,0.6276112666892559,
 0.8123413852244253,0.22696388504971943,0.35366058411218915,
 0.8264669193948869,0.9095723251664096,0.8579665630761228,0.8931864731297509,
 0.47027503514429003,0.12999279947661813,0.6831774627796579,0.712381346314055,
 0.3644691699133181,0.8350339247666518,0.1384143512479975,
 0.001320027905043819,0.9208871463082855,0.5760252904626416,
 0.6514296189506975,0.24799280735441642,0.6654820278673649,0.6707465139891825,
 0.49129638939065645,0.11768869944584348,0.290894017974652,0.4553814473020622,
 0.11793566741300177,0.18323809203608177,0.7758133229508054,
 0.6738661014809442,0.9057152521910762,0.43648751695664434,
 0.053458518787852016,0.06441023812080837,0.9869918749497675,
 0.9176123237409369,0.1869752985075518,0.3739212618019849,0.9763873678815723,
 0.8146166112333869,0.44306975534820625,0.18025701447214282,
 0.6799572252677751,0.8649271667231916,0.3729187119981626,0.8689877635247995,
 0.023875370286499553,0.8120863304415471,0.9507240457403335,
 0.5636333630841561,0.5982262786976628,0.006726255739436859,
 0.7606663556788901,0.7814621293306244,0.22706510100461585,0.8227143532652926,
 0.3108625295599471,0.9932574218273911,0.5753511899607269,0.3144326418069233,
 0.313902221000501,0.3618715057570484,0.9402791462078159,0.6816850078359327,
 0.8157279261510724,0.37102448729315174,0.8753239410474588,0.4648580980558332,
 0.15280921723118202,0.15340396939270307,0.2095533598433228,
 0.5138824411655801,0.32183497384340976,0.4046118882584686,
 0.013444487036106012,0.7726914896175169,0.8703572300580589,
 0.4611095240189642,0.8161898567717337,0.20387984005587922,
 0.18117265516288528,0.747161535428849,0.22199944573019592,
 0.17712125699094516,0.18702716065831804,0.7146935801294148,
 0.5220299817231397,0.2511142146450731,0.04777526470294413,0.9057085951739096,
 0.43201418525611923,0.20245017977733926,0.21632247389890302,
 0.7782591081338711,0.4265522227596784,0.3871617918586294,0.7098924109049145,
 0.3089597122456892,0.8318766157066791,0.17777029837851255,0.882829207069927,
 0.005339927765274699,0.004332386133390265,0.23146564264236058,
 0.5416942889977127,0.6210548085376364,0.04976269518904752,0.6525252207923705,
 0.08141580150017202,0.19048008552962803,0.017140073933693767,
 0.7829118632697466,0.005853416160708558,0.2457104341544636,
 0.16125139532757027,0.4537379739780891,0.9227898529373912,0.9762687421597296,
 0.8103475072273627,0.339638207112142,0.0869146115485816,0.9665590720502283,
 0.17935475474224793,0.3257475912832992,0.8063941724275754,0.9030940137636121,
 0.09843882686161687,0.9761707103183366,0.5929718115998333,0.4880025792952605,
 0.8719762099400716,0.7912708599443568,0.30990484827703735,0.9516886401700794,
 0.9293102888497956,0.6058700085532436,0.9538109451174818,0.6072212310662717,
 0.2362015755804664,0.10808488015430961,0.0030753628763899865,
 0.5990423518512951,0.7822646915087718,0.5247644810304212,0.2803825822616466,
 0.11521369035689044,0.16385450186936212,0.06389143737237146,
 0.8281378345973158,0.8617580690099034,0.4199000788137661,0.8371244761092685,
 0.8437524724306366,0.3297022888425467,0.15708097164941415,0.9650751170911805,
 0.6926760744189645,0.6976940133372114,0.40577960369148236,0.4441677980797164,
 0.928946743185188,0.9803388149050766,0.9674307800368899,0.6329546197863234,
 0.8839726997299497,0.784839623170579,0.9492305118457386,0.07784588149357019,
 0.1938709198478089,0.9032711765940555,0.8584798773478982,0.2565838561547438,
 0.2521413982067007,0.8757692221724592,0.014075053438952878,0.273720767949279,
 0.695375905144559,0.43599724650387783,0.8815200732026467,0.9340062628555661,
 0.6581656618179227,0.8061155235411483,0.5374088084920725,0.8667181220888823,
 0.6083555373582348,0.31031664771372125,0.9126918927650998,0.4697250293694696,
 0.6957864955213863,0.6948081008979833,0.908694157850856,0.8990058661169009,
 4.3515544932415384e-4,0.4368537410650448,0.04288428374813125,
 0.5407346078311907,0.32037476272316745,0.7986733899075555,0.6384838226147493,
 0.3996834533141047,0.6628909759983331,0.2984955833805323,0.9500735323840865,
 0.5559024888996049,0.7957291053238944,0.5046636965686364,0.6506176697912261,
 0.1261736745842459,0.725840628859719,0.3825636061325325,0.13965657323496006,
 0.9370983680255203,0.022790441782963633,0.4908670906691309,
 0.5941789920324593,0.11197604689511453,0.6157958683385711,0.6305907752819528,
 0.16863986778060946,0.01382405852609958,0.11467623703493457,
 0.2950715430130544,0.009334998067391909,0.3056677904115942,
 0.5878824904799179,0.900346620183075,0.5552931059371617,0.13588637322274244,
 0.7895027199593706,0.23500994058813363,0.1983647479258287,0.9873118056521255,
 0.14703089736944053,0.6160408127096595,0.6650036625637288,
 0.04323557255394461,0.9072821048271911,0.005445409876933982,
 0.6237409346574785,0.2846082887107667,0.2190916454887274,0.8636320946609978,
 0.8520791143892541,0.4697351657208557,0.2553151335965853,0.4303547834356355,
 0.32324192762479576,0.033278340747872504,0.9861971232430106,
 0.10961367168623315,0.3581603114253564,0.0656807311391262,
 0.23648183045587878,0.14662598073551847,0.9523196622375968,
 0.8687224750848315,0.8968429733625496,0.16823807609835562,
 0.40279209339854294,0.34307035405376285,0.326631469207876,0.7682733439461922,
 0.03233947969701467,0.9060861649634032,0.10246020506862275,
 0.07898569770436392,0.840987559356811,0.5712683837904732,0.8874553407063938,
 0.5267228074666135,0.5292356446951336,0.9423241031547906,0.9314356533985613,
 0.14905058756999445,0.4322480857991372,0.985411942511213,0.5453947230975029,
 0.8537020372764188,0.6946435112458028,0.07839914980981788,0.9969385726431507,
 0.44911688666834293,0.6446863481561323,0.7732440288188578,0.8563944002136701,
 0.5646229174602149,0.911326892628055,0.009923907781651087,0.5072162253792158,
 0.33708450956304326,0.1629705931404335,0.22694037199049444,
 0.4283126380729946,0.3408979820962609,0.7553946145575767,0.43113853121630474,
 0.4320142109628803,0.0844159161627076,0.6508377920385086,0.3922109775364503,
 0.6587269735006265,0.6652841926660258,0.9773178635446722,0.386718793062351,
 0.17320051853862994,0.40806055181868883,0.07729797514722825,
 0.8145463765882455,0.5173994572595048,0.955837616266519,0.23697923958085387,
 0.6584241148938843,0.6825899482167642,0.588323836095643,0.6213498163363577,
 0.9887469945338577,0.6880193478160669,0.5588745224011707,0.5115543273612233,
 0.948961281839864,0.14957649503274806,0.4931886777451844,0.1669292188882212,
 0.5018409444068801,0.9033618170153452,0.4563404530002231,0.971594084928896,
 0.5202539881356819,0.8588328387380555,0.6052191970858474,0.8051321042797184,
 0.8342596071289601,0.21231097418818445,0.1945714202353651,0.8772488522203872,
 0.4919746198252578,0.14030620959036055,0.5672466563808812,
 0.42202092460131535,0.38098755822062347,0.4334752292100519,
 0.48243959529110214,0.4007285614454298,0.8728890621272999,
 0.049066954338056634,0.1874461035503009,0.18246315759791243,
 0.3055359696961212,0.25335669145945106,0.18215632446504526,
 0.23817314655410193,0.6647782095343794,0.9100612667780994,0.3810565254175986,
 0.80616879814597,0.019305301818979492,0.17529439251885526,0.8199036813966272,
 0.8058200954019382,0.13137191060341158,0.03167391547586473,
 0.5891040049443637,0.9286236818309109,0.21588817229737756,0.4758443499232994,
 0.9459999164948729,0.7447206623219468,0.525934499988028,0.3257723441693957,
 0.5872487268261005,0.49566206325982964,0.15966303285807337,
 0.1237669006589528,0.3983646876343412,0.6731241527047135,0.7641329508407595,
 0.5003226777071516,0.9440065186171229,0.6260728356904306,0.4263740993118291,
 0.9644432091999909,0.4500465455486483,0.023644720254715867,
 0.5809521386868668,0.18919054932983848,0.38456807719207675,
 0.02554753711361968,0.35595015572961175,0.27250390198935703,
 0.08024459413416607,0.9375415810722066,0.06816429147623315,
 0.3820425980954383,0.12172461209364027,0.5264440096120029,
 0.25568794611525203,0.9334470703726657,0.856875321354624,0.5737822450087378,
 0.9383073196972768,0.8712191414198512,0.8942201409239876,6.71745589028161e-5,
 0.38019572099115406,0.24106599870065526,0.08649478712780934,
 0.12408527650357493,0.13252844396025854,0.2308270341589187,
 0.9491291714119793,0.057846152295479514,0.9912795531614785,
 0.28870508993802413,0.2836038466502999,0.48821703178095677,
 0.10186260632758493,0.2137337023094258,0.3745454188403998,0.9728686049253361,
 0.23655875800649295,0.9575939088799121,0.5043297026725573,0.6624249916848588,
 0.9271435043614406,0.2669693032148306,0.9802682587531102,0.1351893841990467,
 0.6080566627607117,0.41900109770746286,0.6405755605643546,
 0.25417984000997496,0.5658936274113371,0.220714899006609,0.15307240088090301,
 0.04363438942399944,0.45512376919758135,0.39780414838307676,
 0.6863207575170718,0.4463157843986706,0.06440037943466381,0.5797292753391369,
 0.5971440195047064,0.3233752952127501,0.30883611319998505,0.9194189887773503,
 0.7395979423079277,0.8966686625053425,0.8767548508739264,0.230050571094356,
 0.16694104626399475,0.07795807746070693,0.6643548486191615,
 0.22960671335563632,0.8298507203138188,0.528068467963905,0.6777312301451621,
 0.014460139217223111,0.25501363803579724,0.3370528984628467,
 0.053796074845192776,0.9317114238629569,0.9798993504081152,
 0.29970081119470915,0.2425359092534709,0.5689492848099964,0.143413345745243,
 0.17459221979446105,0.5280222213684422,0.8700535832647386,0.527465042949782,
 0.9441729746390407,0.05944953327130864,0.20387241427440017,
 0.6542726757883345,0.15232835226747687,0.5113176028267968,0.5328247338477264,
 0.5106198011214548,0.03562581316943014,0.7262680927399847,0.5748616802740917,
 0.5864469628758872,0.2883633499515905,0.7964418708214729,0.4322714009540306,
 0.09403712857132673,0.5752385761514941,0.6205874936487824,0.6136482374573966,
 0.5490176080869013,0.8503976096934959,0.6649846201967058,0.9188410068229911,
 0.13111803641959352,0.43835436217198565,0.7571621149335552,
 0.6762885451550619,0.36244778699025904,0.056656454737420536,
 0.3646589041213195,0.7648650136574259,0.5228841599901094,0.6883219700856041,
 0.22625025138792154,0.7296089828563613,0.024635665232034976,
 0.49678783173175434,0.8966992286967436,0.2199667202195611,
 0.44716865557587515,0.9715883524749837,0.6570248411319217,0.8775756842098735,
 0.4243848372199811,0.8801018946088521,0.3007881560956541,0.8033950538702833,
 0.8701303209715995,0.3359887432528763,0.043206885005980356,0.961775059218503,
 0.9251325218701981,0.25393339149153227,0.04386910168912839,
 0.6598165928904035,0.893498984653308,0.6470792511992685,0.01141121474829776,
 0.8557380825831289,0.4157147385530987,0.28165242575528004,0.7188297556682833,
 0.13033516740065698,0.5978470422201724,0.20770545445506716,
 0.8836094887996419,0.9776053088724576,0.6934801769583586,0.5646920816312017,
 0.6576169220488266,0.13739412173604415,0.8518354496514253,0.6493055342915588,
 0.8835755375720501,0.5487665674593399,0.6041326538744815,0.27194748021697324,
 0.29329230728480193,0.801502207671889,0.9201718648432817,0.21955381733534152,
 0.24807841615512327,0.26855840700362144,0.15168856958541566,
 0.20410021547267132,0.27355182807430767,0.48507975799700853,
 0.061055915532882477,0.13610446738301496,0.19529090499760982,
 0.6258938183696019,0.7021616476943294,0.8999634599442731,0.5249878632533372,
 0.16214398543348985,0.4579764225233589,0.01230607037028475,
 0.45274588036590036,0.7908718534358781,0.16029912820539782,
 0.38003779454615494,0.09291120741074566,0.16929187823535663,
 0.20486247251536305,0.1021226291823808,0.4587417282827493,0.7452307602974416,
 0.5030464046250198,0.5033007845122626,0.44165246514591316,0.4381007313594658,
 0.4594147276371716,0.1732425488206828,0.1970402292048521,0.7725174728536081,
 0.2583922493784686,0.5249778594124102,0.5728074461484391,0.0600762067609617,
 0.13459626786420464,0.5255700674004629,0.6081774623089833,0.1219872368176419,
 0.06862408421201449,0.035750756297384,0.7596873529711596,0.7911179510137263,
 0.43192236538405215,0.2934477404499637,0.662596675894494,0.16308175136373704,
 0.10279669894151422,0.1743210288824183,0.03837747323543417,0.517720505243046,
 0.22545081397503997,0.09003230478477797,0.5909463196713786,
 0.4079060502193188,0.4829719406429518,0.44951615350557106,0.3072233777905917,
 0.2547695481777875,0.3303516075678281,0.37176976706345577,0.2247499505046765,
 0.6565883339118688,0.6488547896553049,0.322121977578271,0.6028243531414192,
 0.050226886294001405,0.5339225903907806,0.05999292895576247,
 0.4384951245099856,0.10183964806519674,0.34095487529107804,
 0.005890920116921494,0.20505643378170169,0.45263892390466254,
 0.2443912262894573,0.08641282852272403,0.5967802488666817,
 0.16194082533535092,0.8442459561786477,0.614735787108865,0.7343788131501984,
 0.691912575811186,0.623831795854091,0.5609415914631035,0.7227158136245404,
 0.5980679223209555,0.5755503829541835,0.34405351381364757,0.1359544299726505,
 0.0023861458622953524,0.41032291964967016,0.3249134433532984,
 0.96932671921781,0.3484447494049725,0.7764145417606392,0.4728994230212027,
 0.8013706430236732,0.38056509850871933,0.7352690419877418,0.7039592950624091,
 0.7706561690104029,0.6763589485096784,0.20193647755241717,0.5381047864830075,
 0.6665064316472878,0.5137680054764787,0.6458881481418238,0.5659049953511471,
 0.5297425815869,0.8249232476495415,0.8446466409509221,0.7259031365375609,
 0.663972799149207,0.4703888380883481,0.8685090243447156,0.33560378273830627,
 0.11602069081102917,0.8643636980830351,0.9497237915904043,0.6247139671090209,
 0.8006269503294453,0.41199108213904845,0.02331016632340599,
 4.756377567298742e-4,0.8127737382850697,0.34536888369962293,
 0.3775108653234792,0.16303200898089665,0.9474119809381887,
 0.22618921558523464,0.29253607576963625,0.3239477003085325,
 0.47630986405021025,0.5639342934423315,0.17566939999365494,
 0.7562127367251565,0.34137886147906005,0.9980556630545303,0.7733410106308628,
 0.8713956190226861,0.5207304868035743,0.7903543168569985,0.02877318720221711,
 0.4329216072393436,0.2455225085847499,0.40607514378183873,0.1451305946470829,
 0.23615316646414397,0.016472745367304364,0.5961611049601363,
 0.3888872908271672,0.8638796264660387,0.6069322110707396,0.9027748722901611,
 0.898863853402631,0.34765097768156605,0.06056382141053773,0.4212180566267818,
 0.2553576666714996,0.3009531907611801,0.9227435833092692,0.5721026495442971,
 0.38956825024398023,0.4571162870523514,0.20889350335632118,
 0.8475378820355471,0.5701468461738136,0.16377736255502595,
 0.21800487144241087,0.569849372264477,0.13744592141143264,0.6542178133030718,
 0.06185101524748937,0.024785495742649877,0.3467415185957531,
 0.4138666308333011,0.6261223421817202,0.38293483646507254,0.8468431436266683,
 0.5482981498559771,0.5380774847547751,0.9588762481266845,0.5962427515759345,
 0.1706297856022413,0.5421634437943026,0.22047424938328497,0.8592570780541678,
 0.7425007760903701,0.08232564207282689,0.6760292120126797,0.37550212070693,
 0.6449152475496511,0.2861900450968611,0.48998747501481266,
 0.06307106539882268,0.81411751413748,0.23688722657100636,0.6166160124816158,
 0.8140810980526961,0.728957042205439,0.7972391846643921,0.43836839374094605,
 0.8204449693021078,0.5170498920542711,0.4154859700087552,0.581585159396488,
 0.09610348306039329,0.04706711637674266,0.026841745458583532,
 0.9406233385341414,0.21064104871316558,0.30145073174373227,
 0.7664622099100652,0.3203971841733694,0.8533260716833719,
 0.056508714194687615,0.3662896759216332,0.05811239053203954,
 0.5115047107204728,0.31799537003551437,0.09352431604774891,
 0.028918003645366785,0.22336029070370367,0.5618118415682363,
 0.8422793606264193,0.6348917467042197,0.2664917303346054,0.25811059295760164,
 0.9920186971289904,0.7112137054597839,0.7093173305266905,0.8026367343020281,
 0.34177867595386335,0.5491865805467512,0.8092105126916641,0.7532442344426984,
 0.09698414244675513,0.29602321555692634,0.5189916699866148,
 0.17820592158444948,0.9761133360913421,0.09178489121362876,
 0.5392281754621457,0.01724998946857459,0.43437056259801066,
 0.4066701846970112,0.9848314028357038,0.22315245869042166,0.5031078109911831,
 0.40144734785449154,0.24302271580173762,0.1316702156220405,
 0.6083355286330825,0.17251597054439372,0.482521150327165,0.19981283898708635,
 0.7854045688729021,0.5259307492078814,0.6421616164394045,0.5776124628290116,
 0.6354273406276036,0.34195755665701344,0.43452736430331285,
 0.3522036111440502,0.7878093898822273,0.9151169078148291,0.1042043064673095,
 0.49111088538542946,0.7281248516024146,0.7884268480299403,
 0.40444035471820083,0.9636333804033873,0.8046576495634644,
 0.35976467747929985,0.3806396495771148,0.3286126354072051,0.7587006511028258,
 0.1906183085734714,0.27948515348779257,0.4906932230937633,0.9146744245854564,
 0.5780446853533865,0.8259329908610612,0.2637886296580526,0.5949821219143525,
 0.8466310974342335,0.5573300233857902,0.7321894640583436,0.505548875444741,
 0.36426142985323806,0.09127334530174547,0.20767573714705911,
 0.047236179686543034,0.3775269246296813,0.2321522685658326,
 0.37871916101765113,0.2415217146906602,0.03510715485473803,
 0.8423420870757339,0.16793606746558287,0.44014984391942147,
 0.4139411638381102,0.5769807891433979,0.16646460866928803,
 0.24309649664027666,0.4210610296905216,0.7267186442450313,0.6126268244299621,
 0.47962696367093105,0.05510633866686865,0.6558760468093559,
 0.9860575673338321,0.6273832512324933,0.17061166338408096,0.3226024451360765,
 0.4105405638386649,0.03680090020280591,0.309869290226772,0.48286599124347074,
 0.7342893384041415,0.45075902451030503,0.9829999241414807,0.5736519548691289,
 0.07300874280140357,0.7757304082103296,0.6573374671878921,0.5425526777831824,
 0.8345398017653597,0.7921037347359268,0.8403337150547792,0.46738884880393183,
 0.5752912951688105,0.8308155043055513,0.7170191865104873,0.15725385379514656,
 0.3910896528563219,0.5637898094819259,0.31829429777232887,0.4712252271044427,
 0.9092144119863639,0.9196611006395413,0.43145269768776184,0.4228236581217172,
 0.5799636839128772,0.4316524991764561,0.41476472051651603,
 0.056309610740376215,0.7802418161359038,0.5909685818571356,
 0.0011498538418481807,0.40730757574160115,0.06270702766868952,
 0.9570824745307562,4.90296563445014e-4,0.6891353206577329,
 0.14562331717504695,0.14991072938879735,0.9795443674593011,
 0.9836284999592031,0.07381430566492075,0.6706225835612695,
 0.29904193176355554,0.13897479982486816,0.8917877791057173,
 0.46755222018794795,0.32030530930874224,0.4350704258110524,0.580454519090429,
 0.7921220084075898,0.2982311299606112,0.20440139545986225,
 0.40750800647469854,0.8576211685593697,0.14224631807203536,0.950859946879126,
 0.5095372923722655,0.8452487312985401,0.4060033866698268,0.20325660796087275,
 0.33244716161568544,0.2624569270823581,0.23547316679240227,
 0.7787849662237162,0.7658672903539459,0.2998151563840862,0.16250316387503616,
 0.6707042620368947,0.2596002831354983,0.9130693820102431,0.27639381747539915,
 0.3938532622777817,0.1841788341289614,0.4486299071113038,0.9035840428538725,
 0.993829405700282,0.1773319994499789,0.27906039668443305,0.3832620053874566,
 0.784456779028962,0.3149633855790024,0.12523300024187178,0.4976258867530732,
 0.30692892573984776,0.1647199517574418,0.6278849279369391,0.795263939909899,
 0.714256570880547,0.12631301381464377,0.7268087849339535,0.04764530752568508,
 0.5167897740130455,0.3171667509272358,0.4313918166358863,0.7993948739163408,
 0.9641452603660781,0.5879681447169474,0.4694090790639198,
 0.004543823689209359,0.6525776867288609,0.3143735566864432,
 0.03307167220256868,0.3366539669419266,0.45576368609960294,0.647844535906719,
 0.41402087076633154,0.8126792040944472,0.757465846063457,0.6090299401885987,
 0.6315834527311422,0.41397046544916494,0.14941891650041828,
 8.836229083389879e-4,0.9325602438987717,0.9593318625513567,
 0.5301947613215321,0.711983820182925,6.770597915852772e-4,
 0.28159375804981546,0.4255503143329974,0.9834789595309343,0.8559825238501269,
 0.13657210479960225,0.3909416426983221,0.5732960806684417,0.8575237910256107,
 0.6550462290234639,0.43825031067947107,0.5714931744166062,0.6805578349088319,
 0.59922979355854,0.8366528922258358,0.9607389312652861,0.5982316620776084,
 0.711576521940682,0.5928345245886583,0.2528970195532525,0.8389904868380617,
 0.9966921050863944,0.9506611811724494,0.9990248790223475,0.5748331108082315,
 0.13162922573429992,0.40865269683759897,0.14391226332241436,
 0.5976773579195668,0.6578691229438908,0.6701655096313089,0.7623566947407551,
 0.9411473182796791,0.432114546652303,0.6915216585283627,0.45852016611292623,
 0.6493934273483166,0.3580040496091096,0.5018793876803614,0.18193851876278455,
 0.7263833845965733,0.6660501945400072,0.8851174780308141,0.8656373729600787,
 0.753620613450927,0.15619208872337642,0.44587779747905765,
 0.16962581341500016,0.10903941312398058,0.07045589128479213,
 0.029526624130167933,0.7056706369343306,0.9002999972300433,
 0.6959773357037244,0.24424157912284183,0.47353362306422364,
 0.6359461397321459,0.8909088477344711,0.08194222969395792,0.2840894547175611,
 0.377064235121623,0.7973892409236213,0.22130622134481825,0.0672938386872467,
 0.49193544957362956,0.4446335767891789,0.31791522021725704,
 0.1273611276427684,0.23143532996742022,0.41038306384683576,
 0.36897099356829755,0.7540932364269164,0.9832133810243405,0.6874714522392559,
 0.6651351015497151,0.5417528501271027,0.3685332719272333,0.5501182035559775,
 0.03992689282689832,0.6055955233414317,0.16011351632861448,
 0.02813778548760526,0.37554653227797974,0.6626174534925569,
 0.007638626724120412,0.8180393610263408,0.2791439693184361,
 0.5499564268866438,0.2248028242740998,0.6985779373907464,0.11265977155463947,
 0.1452423906415765,0.8965281352769439,0.6760151690288584,0.8413015093089542,
 0.40259716281734503,0.909490226870186,0.9394599344079927,0.47475631723035217,
 0.9215846287898752,0.04345758518341658,0.7361798187560887,
 0.10176331456096988,0.3794427141720029,0.5141961104070737,0.8941569153935524,
 0.6401198831156556,0.4903386290954057,0.03241670797774898,
 0.054981438951364425,0.3926789693633934,0.2561987128119386,
 0.18602439939809545,0.25475454601091607,0.27482444877512824,
 0.8150821617356091,0.4742565102826629,0.5517737156321472,0.7981233052499314,
 0.1791931054749647,0.7147810995854578,0.7196072801999132,0.6200217757357092,
 0.4810705809373529,0.021714800677521406,0.9973456811628024,
 0.7658523898199838,0.5435452577959147,0.7083543745674341,0.11907457307297353,
 0.47617602488155175,0.3810281554754671,0.0668508663619537,0.5444147302470828,
 0.7378152774882484,0.40548876950046897,0.5989446056477001,0.5658486539171331,
 0.9033170119425219,0.09799308001702345,0.5440978639845419,0.721745713548462,
 0.2943175353250662,0.18650079105496942,0.7689198886114705,
 0.12110563386434814,0.8236408630505345,0.43389954891843185,
 0.3201932720574968,0.8170303639175317,0.24157485649300392,0.3757550153963093,
 0.4749215796133255,0.7440130520820238,0.465916065189138,0.12548022349790666,
 0.021750470463218985,0.15659139829310043,0.6467927524246911,
 0.38593626766697187,0.7656077496846556,0.3650009771943842,0.3665879383892564,
 0.3069459584594656,0.5202135311484216,0.896912900241225,0.2317097969315205,
 0.7638925804570689,0.679512930987845,0.8884867911733232,0.5495443991180095,
 0.5678365564738267,0.550697700489043,0.7539498719951202,0.21215213752557927,
 0.25871450546689645,0.520742481355907,0.24145153974761435,0.6338420839361716,
 0.6432604207846997,0.6159504062459221,0.34303057251650293,
 0.30464150319369354,0.6530864728136041,0.10922581680062127,
 0.6857745782029832,0.6284806357438977,0.6765830346548823,0.47390873330764394,
 0.40431226718266333,0.7441873291562316,0.8970911909629781,0.5274721367084122,
 0.018960112199551515,0.1188345217617448,0.5979328007020421,
 0.4310880316045401,0.662844155931831,0.6539193354757499,0.9444434256791004,
 0.4934918740788533,0.8998902519781927,0.8534213986196495,0.48989031997471655,
 0.5628178895983452,0.5407037566067459,0.48021915224185263,
 0.16599125419384686,0.7859102034799227,0.893718076158898,0.40059578086220493,
 0.5565677283624315,0.7468188925382262,0.5904620300215462,0.30181586960618,
 0.2552384866492483,0.7976665220815664,0.751717331917776,0.04179843111021819,
 0.806794549648169,0.5495934482309665,0.6248403698632022,0.19829569544389414,
 0.5702948231925603,0.7947814469065623,0.3133857228993844,0.21203970270144823,
 0.5485433413744363,0.27568244785158236,0.8830278528805767,
 0.08192348733169164,0.2969501875909393,0.8817930348666999,0.8277656694366995,
 0.39015488436059054,0.399985813302829,0.8061651199488457,0.2184503962123766,
 0.6471237746903901,0.6315199985013374,0.467168024664225,0.7474199657456752,
 0.12829432096442317,0.1481947890652704,0.6976409821508734,0.6105408283565751,
 0.1683754935948253,0.8310175696390231,0.9793945466981305,
 0.048406066435300654,0.3405548513147316,0.6614414553516919,
 0.5477795088339081,0.9286430922015957,0.25258282377774477,0.7452360324117373,
 0.8893047436164595,0.19952447293107856,0.4755176956424779,0.4500092337790149,
 0.1996883488102893,0.06809750479940724,0.021130130848637463,
 0.6172638307421949,0.053018948442315805,0.8357477334370539,
 0.1267593409906944,0.36251847678667304,0.09448380838619275,
 0.7655643438990478,0.8387907571016552,0.13182867121638375,0.8774493613849617,
 0.6216593173040904,0.7215520196710241,0.4564822133884341,0.8552085700621804,
 0.2137155719374122,0.7189071000711664,0.6132466612619194,0.5871792369978146,
 0.02529576704669889,0.5667653100823264,0.3101657954207575,0.3630851174519132,
 0.4220184080349947,0.5086513167220059,0.8736759649133305,0.8437594578762035,
 0.2713321361260239,0.4981965151034594,0.19455769799944334,0.7280283506317957,
 0.8715799889581463,0.721773821206176,0.3328563348127467,0.5504736195598106,
 0.6547691644163465,0.5072849862099671,0.3349423212663214,0.3197129446254289,
 0.36787427611017065,0.3623231325144277,0.08440396603865297,
 0.15055097459028488,0.3829747777595539,0.671954377398151,0.5738100598750872,
 0.8143780729785887,0.23257895973421594,0.05479158772523407,
 0.2917487752302951,0.6655624804327073,0.43702100703200797,0.9266547643687347,
 0.600700268309468,0.06960864711914827,0.41721806213543067,0.8727223420287891,
 0.6993997742947439,0.24192663118676516,0.9196084642010145,0.9521974303000914,
 0.9514196386602254,0.8010554466239778,0.7925521888293424,0.7124502003879607,
 0.5014614179896659,0.3404600486204269,0.1420647080344125,0.8676322926831878,
 0.4306181029382431,0.5031202968394358,0.9966101566974309,0.5119809971973985,
 0.07854260184066297,0.6556822662125487,0.15681564309937357,
 0.24782191336656967,0.8567218302095944,0.3831376469445267,0.4072006408204988,
 0.25830459348909507,0.16197367167079268,0.042641641556032006,
 0.7933119501717636,0.488439509963405,0.9677652459450807,0.6195899744615995,
 0.29173820071505874,0.49152668573290204,0.8174660824449088,
 0.9914481692718721,0.5507127919324253,0.10397728913181237,
 0.05948736065895166,0.1362614037718647,0.08175845523033642,
 0.8623225902039772,0.2685770777292914,0.15320179658234778,
 0.21737581322693666,0.5696234320752466,0.23081192713299648,
 0.9716805875286552,0.08726319825755913,0.7064615903439246,0.3403009381211979,
 0.2916717532280295,0.5751927625934086,0.5162519976007345,0.847255514335731,
 0.31678270305753875,0.30758202313042127,0.5352232097803301,
 0.5516235240623425,0.5677228513953703,0.30871741540274533,
 0.14935121328665757,0.4490764945312702,0.6634846212407671,0.4873322091232959,
 0.8369448890675166,0.2247791841971234,0.014665786728577546,
 0.8527687211938086,0.037090931362949386,0.21354279754879335,
 0.46729500749018016,0.36785418494656674,0.8487729898546323,
 0.7378185178381939,0.6327178883540917,0.7702633009269799,0.4805857421473181,
 0.18320176797767773,0.8077575977429325,0.9055086389769365,0.5970341491422163,
 0.19501814441871912,0.9476174355170406,0.12733841768462462,
 0.6074052002386285,0.9234133442906415,0.2005922566798053,0.5729824205582003,
 0.5363002277122565,0.3371284874586985,0.9454965501884033,0.22596567275293566,
 0.5657450132718874,0.8238266722613337,0.3642716072174236,0.3392472264286117,
 0.7851893358155012,0.7897553344920902,0.5862243960815156,0.45732140506136976,
 0.19287843059811338,0.21592998399159535,0.8998128868335243,
 0.6477704765729819,0.9473723885704094,0.9016399986491439,0.7807824250226014,
 0.5815343454618064,0.7000564640161261,0.9708358890046511,0.6456024042680197,
 0.8145411592936941,0.3054897838662035,0.5200941010338201,0.27741323406750984,
 0.9305827582966693,0.7623764569428273,0.048261423636697054,
 0.20566133801260755,0.9206116514444255,0.4766911539882702,0.5604253387043301,
 0.42031444935156936,0.5096724121454732,0.9163924999108379,0.2737841348678479,
 0.11219683011558512,0.8579656747786913,0.7451566209934022,
 0.48745508389199443,0.8430421683346557,0.23463872853591217,
 0.8214725987286648,0.6081628584015748,0.981739953924548,0.37563638609204963,
 0.6184633448538879,0.3723282596345068,0.6644083711105264,0.8950078843886727,
 0.18190163510199098,0.28219138748522066,0.05814018442831037,
 0.9498488992280353,0.403263545325378,0.7507301884345591,0.8686623835867296,
 0.4955156475933602,0.09971139817601049,0.30590580065554984,
 0.9302228423567038,0.5268519476961069,0.1527127644276548,0.6323717354880171,
 0.11684701117092278,0.46454463119160805,0.11552493885369297,
 0.051858541351190834,0.35068685302658487,0.7925221822859783,
 0.20939027676675315,0.7017547972127561,0.12271598884267398,
 0.6098371240293153,0.8897042840408401,0.89328431338605,0.7429194858951389,
 0.30042580922350304,0.3273690610018174,0.9155962625839734,0.6326699070842143,
 0.22678626889115705,0.9453916579668317,0.34216616814391343,
 0.4785162217897474,0.05026562287968983,0.16463597999737167,
 0.17792546745303006,0.45260838835137607,0.7369481494196443,0.291081071484255,
 0.13814833054830444,0.39943011155039465,0.1703572487232835,
 0.20865831918126687,0.7314969408419167,0.03353419252891776,
 0.7045463022927343,0.9155513927947336,0.6810142434325906,0.8001541307177511,
 0.710370801284916,0.9284156374600468,0.10398538628234577,0.2952692997771922,
 0.023746597169331474,0.7007389795520345,0.6452322407047536,
 0.9648585441165407,0.8185039404646751,0.3510965002257085,0.9378380637395023,
 0.06215675998010539,0.46479047275001106,0.14672812681731895,
 0.03342190358580177,0.020067161456882476,0.6008324237735018,
 0.15154816429618456,0.282207393656872,0.14134414860467537,0.5043430706109793,
 0.5588944888157975,0.22228485411882581,0.5606416576176295,0.7524873752001406,
 0.670530285604106,0.9143325077182236,0.6877001784650424,0.10411087948620412,
 0.4176974034964428,0.23809835930847856,8.801682984517223e-4,
 0.522319215433773,0.42418107098197966,0.24897524829939077,0.5027397131843837,
 0.048820641233762285,0.11696966243054652,0.5364019531590107,
 0.03282215003459288,0.7226728756989179,0.2587628801105242,
 0.39586076684178484,0.22971066291415088,0.03876812995560952,
 0.9996134612126427,0.28037060099503175,0.13791761065443742,
 0.8262587136078107,0.7425447369934434,0.08698728610284046,0.7989526060232068,
 0.28367731906140237,0.21261955441496416,0.12244862414830426,
 0.01771063219963731,0.45206341767174596,0.9493402058488256,
 0.2937929786553395,0.9607656841104878,0.6602845881548375,0.2983442617732939,
 0.6876248349491559,0.277330809707406,0.4941484212590185,0.43658126470999137,
 0.21958734284558334,0.5898773739558134,0.8082299984536617,0.9934189674750089,
 0.3760778517296186,0.2848925926764029,0.3303932336416463,0.7225683384042841,
 0.44595285827549414,0.8062667034799648,0.23928006688798686,
 0.11731047683502349,0.6192700994170559,0.2859036913285946,
 0.39279928281374954,0.2354512626111822,0.9276178097879106,0.9029627153670269,
 0.7054005810040131,0.17884756697879023,0.6059523122069934,
 0.31859695779399255,0.8274150071455144,0.008479888828747728,
 0.8727190135990059,0.8421792717761396,0.6255903698724361,0.9376182633986252,
 0.5749184649517408,0.7509853629237238,0.5484297657068555,0.5802603294303753,
 0.714988522140479,0.4918016316737609,0.14938181837828435,0.1875300199664427,
 0.06877315601306155,0.35846186361974774,0.6181096843478439,
 0.8871171177796863,0.8850120820395052,0.3786594508797423,0.05203597553034944,
 0.8041514067940034,0.6738931972447278,0.31275520076156527,
 0.04202985161630446,0.3726891467440976,0.5649722012945547,0.9735751486581478,
 0.4886620655389846,0.3698450147122998,0.4702188268148084,0.2564526111101353,
 0.21070849335144182,0.02074151277626246,0.2755710875897973,0.887892249526837,
 0.5462634390876909,0.8309434686015612,0.4594810845982894,0.6278887477046959,
 0.12384009391176543,0.11061275608824506,0.6193295009092774,
 0.05024311464333908,0.9189932607527302,0.17709778524723907,
 0.5421080472212888,0.8955412333500701,0.4911858976563057,0.9618853797293015,
 0.8464257683775529,0.76336104429446,0.8887993423534841,0.7597845110412457,
 0.18348068944183682,0.3166633588362482,0.09185047377699418,
 0.5505985501444444,0.6358397265164927,0.2777532001398957,0.3867461476236804,
 0.6444981690383336,0.7037196299173551,0.9512384685305796,0.8068087488987297,
 0.37068270336419995,0.4496935907420675,0.29731925739821197,
 0.6378673491098414,0.7264661096710807,0.4239367889582756,0.27185094298637635,
 0.25991925816947137,0.6704246789107587,0.9375885531788939,0.7453377382620211,
 0.8739368278832164,0.6677425890481694,0.01054276188984904,0.8875468891225511,
 0.0612549006029135,0.48755166583138154,0.017199824183594004,
 0.18248750265650848,0.52904535376218,0.7211586107689292,0.24860017610067975,
 0.13947557136589883,0.12184442464473222,0.07567048378356411,
 0.22062477139790782,0.334795164796031,0.46141853755982276,
 0.46403684590123195,0.39786390907887337,0.2735925173667815,
 0.8877170512381816,0.7679405739333963,0.3100458077474255,0.1594100576243559,
 0.5857176831580971,0.05937641408258543,0.6593268782103733,0.8280808380118353,
 0.8588546813236486,0.4557280781737335,0.9255328417878746,0.6225418938236161,
 0.049420560469439656,0.3372856876750592,0.7785910789766222,
 0.24169376206990867,0.6465320792448438,0.9278935651345975,0.8881351101415301,
 0.6508401333620019,0.8483835042269794,0.7449647923367575,0.6475338373825568,
 0.1653333402198376,0.689012448805679,0.14058217520378335,0.15491770744883826,
 0.7537735373165484,0.904321746450751,0.6128116995752506,0.37475893877007027,
 0.3451900001022312,0.3589861867983477,0.30440760818841295,0.802294174588376,
 0.9611003124235608,0.2707103664898248,0.6522842473980189,0.8665798547119834,
 0.35974940651286713,0.8843733792264521,0.8158385526130346,0.259259737640561,
 0.8199249657312033,0.5532478569017644,0.8034445744554255,0.7160666683529517,
 0.5870362112555785,0.7417353567446848,0.17966000904085744,0.9690164988260559,
 0.9411387564455795,0.5422455687510744,0.8587798508393616,0.8177789425413433,
 0.17210955032551534,0.8365356780964092,0.17530813975546078,
 0.6329224570262665,0.02400599839472628,0.8664253764077896,
 0.19438148589912518,0.6760337244628941,0.08652890527171253,
 0.6640798851399781,0.04086756517585266,0.038348516451047865,
 0.9584468852835831,0.3844440036494545,0.5166849829158785,0.9843097454815278,
 0.6644140228142487,0.3024710932259498,0.4543179091369487,0.04409264887650155,
 0.494596620736825,0.7481838284518181,0.07184710767752556,0.7966797744227532,
 0.48226253761044,0.46484235297465315,0.7280420016973679,0.49746499143508893,
 0.6289512029967169,0.802930696237331,0.662769223322917,0.2902339185600096,
 0.4809808573061295,0.46462879757584985,0.572046392240009,0.9397051735219994,
 0.9937983243555453,0.0726320827220186,0.16148249767191547,0.2963165216427347,
 0.24317918736497046,0.39276405385881674,0.18320322555498103,
 0.8785951947441928,0.6152643431213711,0.22189283529435633,0.6703124026657269,
 0.7027549203043317,0.8304964309702658,0.32461994893065593,0.7042308580840388,
 0.17384858503099265,0.1223447370630607,0.502505495756716,0.8417925774554248,
 0.4463399359717827,0.7067844443155391,0.6308873361912067,0.8565946631026847,
 0.24862640831027194,0.9192042752155223,0.24188788524744853,
 0.8862402064356556,0.8373364537098072,0.7043418127099322,0.15624633754438444,
 0.1910143494748583,0.6225233824268902,0.6957427413251223,0.2853098752392319,
 0.3391804172719429,0.12803701161111325,0.3066167075167786,0.7442376856413748,
 0.5316493976351204,0.6480106728336433,0.9274027367518518,0.46100501351360557,
 0.8191402693876731,0.2595233787251454,0.09556196912035908,0.7523168233613817,
 0.4084126862391355,0.6758753313793262,0.06210081643074161,
 0.24273845836012253,0.10707624639935276,0.8952737978916351,
 0.7214759588508338,0.09741734891234255,0.4011745014894028,0.4341340059789388,
 0.5441065402481053,0.6908158535021982,0.9310956662403191,0.4584804342674319,
 0.3660658047408587,0.4931020381581056,0.34936459157011523,0.6853623288641465,
 0.7141697914720012,0.0013402723878270795,0.1451045323809934,
 0.15374396795052014,0.9481249306378301,0.7411108578544767,0.6722518964280578,
 0.21465751840450187,0.2258041818000709,0.3431712313161942,
 0.41463717675325773,0.26760329212456413,0.37632096635066414,
 0.6939990727147558,0.1905457102373186,0.3370336463539436,0.9978532499542356,
 0.32633265276160783,0.09037139686819473,0.0759621390824401,
 0.4333002273945117,0.8585133241358264,0.9590904503377023,0.3824243262787167,
 0.3245166950339018,0.3107935297480071,0.21564162474674742,
 0.34353536293444054,0.909554041324722,0.03338514713553542,
 0.21797634025218982,0.3706037844386353,0.14718822357597205,
 0.30001284470375966,0.4420853850638884,0.7069740074955674,0.882674129939194,
 0.8658952952091404,0.004816405132734403,0.42055700720343026,
 0.8523344459436817,0.8091785684586021,0.15592240676641145,0.8247584321040089,
 0.33052620976100466,0.6678764259047443,0.5733132184244742,0.9935554170728015,
 0.49205710971853744,0.6826088240059477,0.33460477097023267,
 0.8044521235634443,0.32092048837667775,0.6511090062510528,0.2589562081463588,
 0.8432792606786511,0.4855379256649204,0.12193639282683588,0.3652475545285071,
 0.09264541808794169,0.72727469239296,0.9695833059120748,0.6811083849919195,
 0.07758599482669626,0.7876405525116148,0.19427722348319876,
 0.5076251824987924,0.9380127950324982,0.7532293331174691,0.22150665909757117,
 0.570320244114081,0.08345726775891416,0.6894681945849342,0.29397186140127407,
 0.3758035203794715,0.8618023671130897,0.8257499167110994,0.12269705158085853,
 0.9659265352120047,0.7413466029296584,0.03570195851385094,0.3010804296352487,
 0.4005292421745855,0.21238240596403635,0.039778529670983276,
 0.5573967531997123,0.6753549510970501,0.9106524297251785,0.5646312832979102,
 0.6948079760422118,0.285050061433479,0.3959711223147334,0.9778004920267397,
 0.8217465881601796,0.3992967027207005,4.4965131842844563e-4,
 0.1888128278753246,0.8949407776591585,0.2647038361382472,0.8409030046792696,
 0.7746009014223916,0.8687697755853201,0.8009578863503101,0.856937360456256,
 0.9992559900291379,0.6215645315866313,0.07298992445928276,0.261826478668048,
 0.10830947493082999,0.03109149649013454,0.6666768392197026,
 0.7706236110379099,0.1600653432138257,0.9906875219568728,0.09320965762792788,
 0.6184747642959716,0.8660749125536755,0.7589453078021581,0.6379580663981262,
 0.9773873403711553,0.41037452930942986,0.8977791013565527,
 0.17025407863395148,0.4439657453405541,0.1509704939005927,0.3253594513287028,
 0.2538063104579149,0.7529583034347767,0.2789469308728366,0.5820812518106763,
 0.3332471043849693,0.3656632187916601,0.6114090225769706,0.6790543255856845,
 0.14500497020158964,0.011803420079939952,0.3746710436056683,
 0.20622909338565898,0.02046242541065557,0.9152012859211983,
 0.4896099526955813,0.5184272855533973,0.8106320393824308,0.6290698079398275,
 0.8862988042868521,0.4721434754228725,0.7087223913911687,0.9379376553188528,
 0.08281010844065062,0.46475897368703056,0.4009575664354157,
 0.3561477877296447,0.291308041436536,0.7319289896859251,0.9208784475266872,
 0.1955549904451126,0.785854776694945,0.42284975911630895,0.19872556686607967,
 0.4429342810653072,0.5714998989003526,0.5472700202307037,0.11938259860877598,
 0.16560493295201406,0.012915449882703989,0.8744635436807412,0.61800405152618,
 0.5913135724107701,0.295835421835291,0.47418280776216315,0.4118548857128722,
 0.5088596256588156,0.8606286743615231,0.4454046007113722,0.12641740871107532,
 0.19953451801148603,0.38099536710570947,0.5671262500614275,
 0.8026575801721861,0.9479261720557042,0.24553903094724117,0.8266609827103919,
 0.5452691361039175,0.5682730631825011,0.9215101990207387,0.7638607600764901,
 0.47071120049837933,0.10294783694957799,0.0014742943000776165,
 0.6300645325399147,0.7326856253866785,0.8296179571964922,0.34009287987244696,
 0.30833335574330734,0.9755122670984728,0.42417585703544514,
 0.9866777119307176,0.20674961043797768,0.250594796677725,0.13666019482406677,
 0.7367477419467273,0.5943280668707099,0.5746972302577923,0.29598462014294524,
 0.4964860971068573,0.08581149359164253,0.4372570370833987,0.1443329368969053,
 0.3274771504234357,0.9617348198605338,0.29916973115428724,0.787757261313682,
 0.8419698126388345,0.6669709456134305,0.9630801709009883,0.3910179854617405,
 0.23869583114947757,0.014786043740014354,0.44173508766583314,
 0.6462009361434654,0.4051715241852446,0.7585271694557323,0.9597791907298283,
 0.9887783469676175,0.8815328396444672,0.5971186055031131,0.12950258433094342,
 0.7743662975955081,0.4293219840141078,0.519950090510363,0.36205242848372887,
 0.09457596292098724,0.6312032769429476,0.12395169194397848,
 0.054251807944457386,0.9783036234198711,0.9040656792872523,
 0.23137442666618246,0.7561062958503868,0.4781429264922463,
 0.19163460731646498,0.7554251317683052,0.36967946293301845,
 0.04878192420367744,0.3906978746245503,0.6062550373387007,0.6295275253233545,
 0.5991511320340369,0.28154237192212683,0.05624595032688395,
 0.06710195470142688,0.25450155295445387,0.04258334633166794,
 0.5162725594747011,0.6663011419545399,0.9762725382876054,0.42241253858772776,
 0.8907766717708475,0.08360876682076457,0.5873724403903494,
 0.14268377865377557,0.8970950384934646,0.12291537466670344,
 0.4482129994824633,0.598867164270158,0.867330257893603,0.13998944948150605,
 0.8104870131761209,0.6519863896206163,0.06908561763898013,0.7766372653580861,
 0.01687911328098013,0.24358604714227206,0.9487844861459431,
 0.6644709250390121,0.6063555784133275,0.5330694217284678,0.5121699917958105,
 0.5963607098700066,0.5952719058259441,0.44984404991348625,
 0.48022401702361395,0.0717825145481088,0.1534957128548463,0.5549683828850364,
 0.6818857147156003,0.9822876227938817,0.0779691737349959,0.17463592163969244,
 0.30326697310482076,0.687734720741811,0.49049064460690583,0.658539031579323,
 0.5850656399269969,0.22853460815195137,0.3192611455673673,
 0.12290826611371564,0.31569893049378417,0.8163039854670192,
 0.11771380192886038,0.917858448283949,0.3751917724391012,0.10506048781933508,
 0.03098509575420505,0.9495856755125505,0.7337931778327769,0.9624239451334957,
 0.8250933165400729,0.8735732767785805,0.9590180777567118,0.3398845181084653,
 0.8663858713728616,0.07624319835079452,0.7355274618457778,0.6109674416481982,
 0.23099066013582858,0.8892396194980396,0.2938216939845273,0.8771719639709535,
 0.9568951214256785,0.10675101867245429,0.5391119306107743,
 0.03818747155432822,0.656081090744416,0.3202400378315988,0.9976809019274169,
 0.9184270014873845,0.39565376753375825,0.9718017542345141,0.9471939895375038,
 0.20573029579714242,0.38806604608385564,0.0986814763309517,
 0.5127280910674106,0.34707668157301375,0.5531779593626104,0.6659207993714177,
 0.6363318887147512,0.9205008732035744,0.8483595252598074,0.8799280546015328,
 0.804380668240132,0.8306910441213873,0.40987977204158366,0.3217685266651512,
 0.5936178256466279,0.984450283615624,0.8455387267104443,0.7064049080387331,
 0.6732641576891798,0.9399560744090724,0.11732926983206726,0.6906830925857558,
 0.8455620259076904,0.3472382615868672,0.6714911980572064,0.8070394153421241,
 0.04608922289250117,0.24516136144855827,0.08500014456024874,
 0.8473525808190023,0.7969832544409836,0.6693263645448853,0.07622062271607188,
 0.6995175266979875,0.5039616964037397,0.8767417879812456,0.9672400037068142,
 0.6819289035852526,0.47346583926609065,0.7033512864349213,
 0.42224119325730236,0.07468493740170534,0.18686393879637564,
 0.8201319181932315,0.1659220831927336,0.8714181603230655,0.322590230675151,
 0.7395053062886208,0.048578179920052866,0.8378265153278016,0.648263911200958,
 0.40174950298275625,0.5131493647211315,0.8162895474431224,0.9041912249722928,
 0.7064864549377218,0.7580522718335073,0.6389157822090298,0.7612508379271803,
 0.19158446448960376,0.22256699698766047,0.539621008759264,0.758767114786712,
 0.7869769538290614,0.36684207838778593,0.7152941680932812,0.4497839512378322,
 0.7749716030591237,0.6106929119376129,0.4593894961712478,0.4635553073965656,
 0.6055699764886764,0.7054110451126089,0.9902879427308499,0.5888177032783309,
 0.5497115596463393,0.6379814215133907,0.638036956362957,0.40575705462706735,
 0.1411466038266317,0.32893749230462366,0.13980291008256962,
 0.4538423012913022,0.446649129696561,0.8969919368620947,0.4668566376099441,
 0.2048100299695187,0.3029937142150536,0.33333843638805816,
 0.04301920246248703,0.1990988161141911,0.8761180449311369,0.3918682754944456,
 0.6637752051711637,0.5884734869444445,0.22076945465518938,
 0.07050228177139617,0.13610489422964123,0.5479499227104104,
 0.43622542381260554,0.6599109491488719,0.8542960295789346,
 0.19482034899689293,0.2657528874083621,0.46699525594358937,
 0.15029434298100453,0.1446521755203517,0.3603553016220633,0.6435107971745612,
 0.3898882971659945,0.3724500068267713,0.9820529954507039,0.4252253081509553,
 0.5461262862485037,0.6185565408650723,0.8584485496231424,0.9777027113384326,
 0.8868363703308562,0.4249015839390401,0.14864377842465049,0.5577715347117437,
 0.21668805875488084,0.3005967343230068,0.8040297326518471,0.6475247062194439,
 0.8813362677623734,0.2907419487542964,0.6100394748606563,0.351774257216973,
 0.7436980660951624,0.40645356127800636,0.2657697425392398,0.4979001962179085,
 0.363346218998037,0.6753245240327148,0.554293245003313,0.8157327779888898,
 0.8770392723435808,0.04820965166989666,0.8600793498549624,
 0.26327039205774005,0.9611029173951351,0.7347999137303642,0.7719774001849531,
 0.5164999722758539,0.3366421605435067,0.6174043550927116,0.390003015299512,
 0.5022913565779958,0.18066996984049288,0.4807167079165058,0.3493183767650351,
 0.45733742149215284,0.9079959652145222,0.9670013618444778,0.7723107829776553,
 0.8474287411679075,0.2958224845267211,0.05225983117785871,0.7580842205088688,
 0.17666393681171189,0.7042530513057186,0.06890524153705402,
 0.7347824188567703,0.8797128076564741,0.24645460037349387,0.8292469991477753,
 0.6912309622431339,0.578864492245107,0.8230257700829569,0.23201801277426037,
 0.14491465979268925,0.5309636341091913,0.25473612741245,0.1906541102457121,
 0.9364022615556474,0.30794759062202737,0.28048346197938256,
 0.5216741181205333,0.1744172001690587,0.24736323209092925,
 0.27753804082925915,0.3354600221251105,0.043194877217228145,
 0.8537228075545934,0.8133804447432119,0.013373819187247005,
 0.7104672531936809,0.1715629897235751,0.4434581382762791,0.404343382325574,
 0.2524061404090474,0.6871488172686232,0.9093978279710494,
 0.031844471971735455,0.41533683058103976,0.9371551276057773,
 0.765612206877352,0.0675838269411938,0.9106037271347167,0.7076386343368319,
 0.6681995359427033,0.46536937526075406,0.11734995710319251,
 0.11525044530251838,0.810949561812789,0.5629822716428542,0.184521299267967,
 0.1435530542402269,0.5293109201978716,0.03563236203897646,
 0.17605886452549646,0.9331418669345424,0.4401256003796389,0.8630608874260299,
 0.29775259289174705,0.04179103840959131,0.08675335621063918,
 0.05264027700301688,0.242209869461795,0.42382454662290314,0.5208112148773189,
 0.95327788303544,0.19706057194619697,0.7487665635773886,0.086465264272374,
 0.8999994898221054,0.46456872026058527,0.268359549651619,0.23016144019752416,
 0.5378094743091392,0.7199578598727405,0.962228978578399,0.44856159672904194,
 0.37070036604872847,0.5587954568928263,0.43405173451746937,
 0.0330588959651662,0.7711371350433012,0.6114888094710966,0.5304236599695789,
 0.5146123977017,0.6304570425574226,0.125297027130014,0.27216472477210185,
 0.6440632896636252,0.2301149779799626,0.3150995539673095,0.9770688053863353,
 0.1886919738363534,0.9848802822168743,0.9476443217071042,0.03854205373893904,
 0.6625418506591063,0.4713620172590919,0.04645686073595612,0.66602258312594,
 0.8507285814570144,0.11097838424000472,0.9187730021801163,0.4977731735926424,
 0.6969545438222006,8.206167606106263e-5,0.31558240597010134,
 0.14296427414197566,0.4200737409621823,0.1765404412492504,
 0.24410642898949697,0.5849219897201342,0.3500918508363068,0.992997118650421,
 0.7635442691873742,0.5993538585105659,0.3636892498898877,0.516545816619709,
 0.6037207510860791,0.7172847403206265,0.0632937334038769,0.8622462362866194,
 0.4234270022942439,0.48797452207290337,0.2646175189119222,0.962619944033003,
 0.6809303259996504,0.6273164550492767,0.006085735560079186,
 0.33671207458260155,0.007237314123223482,0.21335535621086654,
 0.508734004658702,0.26371603522254283,0.9934865042018332,0.35027247099065706,
 0.09125295435519987,0.4026631426976951,0.06909376352383922,
 0.31212926568607235,0.13037591348295896,0.5766991131162953,
 0.3572210034621459,0.4909890171648589,0.759998562885696,0.70087135338103,
 0.5242048986961637,0.11967506484513624,0.18465441027568763,
 0.2928813262884008,0.06917114987288042,0.5285656740409571,
 0.18304796858731665,0.29920338112222045,0.7560909369123162,
 0.6932708843204222,0.6479632293359253,0.9846967845099767,0.7300826324645445,
 0.4090752865549879,0.8645791929551614,0.6836907116996052,0.8048157939822663,
 0.9346393917818318,0.616279270300954,0.3628073070613481,0.13006588677489095,
 0.5785777010851151,0.6087724695713029,0.0195881385065525,0.29374598120552786,
 0.8069420907109509,0.5442668250279828,0.25741671465184957,
 0.08329626710783211,0.6271401164952239,0.6778305626055271,
 0.14048522471951252,0.6309592103281685,0.8762898643887457,0.9399395664333624,
 0.7912226099511467,0.9474455065728041,0.657239463487857,0.25025032130387714,
 0.7599591605365,0.504321463197022,0.5651683315595396,0.7457893972538969,
 0.2585628198708999,0.2755221713620617,0.0648461842983532,0.6932791867313204,
 0.96102305603723,0.14012002654434308,0.3219383679935518,0.18386694710053764,
 0.5014617010373082,0.29538974465954426,0.317899753354124,0.7910345011771389,
 0.3055834834780663,0.3338002601675437,0.1813809254450377,
 0.015967711279563424,0.7387586433765932,0.47811112625209173,
 0.4004362248686286,0.9920161874811154,0.343712716274016,0.417515993201966,
 0.30323319535307014,0.16900057785996236,0.9480621309455336,
 0.6826323655112952,0.6456469399644758,0.6267944798093813,0.8864218247734603,
 0.006648667838925371,0.28287574807843563,0.08078036333519045,
 0.14100138508620808,0.2603592458248585,0.19414257859496875,
 0.6392512180379344,0.4216179079158662,0.9173745325026031,0.8054502862779732,
 0.8633682596286166,0.1764610763982769,0.2876490715331481,0.06349085649495934,
 0.47508669455411034,0.6810416295652173,0.9970330925722498,0.3428583805595853,
 0.5582627279574609,0.7922185027382453,0.6591748707973875,0.8095325062273825,
 0.16954919298672566,0.14219380975692741,0.43773263435501975,
 0.056594036210620424,0.36563311302243306,0.2033228260241099,
 0.809570885698512,0.26175223924114044,0.6451755262167577,0.16872534073924317,
 0.9900023064346105,0.4298027046678724,0.3162100127508223,0.27444423828964615,
 0.4259001296187078,0.6924467704427631,0.7469210025186251,0.48061735108267456,
 0.6611240299722853,0.8558657832504878,0.7967109519948367,0.9498506885214537,
 0.42758876338861895,0.6024774487275186,0.18336256220760294,
 0.9704138063827256,0.24313807747661498,0.7456986956718579,
 0.26900787258145653,0.30503952968456405,0.21475806477649184,
 0.42781800078054333,0.32342073611801503,0.9976626692008557,
 0.40891012247753955,0.6447555206582913,0.17648611910946232,0.714451279522071,
 0.5783611029619747,0.6617363615083904,0.1705210105833972,
 0.0025256229713062073,0.3970273227548704,0.3508422698462167,
 0.45744798168181766,0.37808102061508364,0.24243053472091491,
 0.0022644159833316024,0.5606654907843284,0.6003033759350027,
 0.013277750787553888,0.5954641880505117,0.9089596518508891,
 0.7195903957971453,0.836192333303312,0.7811761502765213,0.28529181978426943,
 0.2788106979563777,0.2789997480388894,0.101322224412142,0.6568413955893596,
 0.21355849174614028,0.4979885237512187,0.7982765919721927,0.3536989104757208,
 0.6630679171185717,0.8870956768587774,0.2386991318018158,0.26762593892692854,
 0.9015092478909711,0.07693066223662692,0.6757224597550532,0.3518942959380358,
 0.7935270122569423,0.18734985094692902,0.35593448094349023,
 0.048063554036426925,0.8376511858522311,0.8075263396640457,
 0.5852843918296985,0.036189306387332754,0.3264998472773841,
 0.4681952743974964,0.8748083670077479,0.546209701853123,0.43941484031093336,
 0.16509002637144776,0.7971346375109034,0.9229276163575849,0.2688820025934805,
 0.016511128680419263,0.17216309314636546,0.04261418640902026,
 0.1579546480926608,0.13215668621382604,0.17988284717506686,
 0.14429061000768206,0.20410672949276443,0.29216569847538076,
 0.1201003538006794,0.5749486881720174,0.3988226441672611,0.751325208813081,
 0.4012054258240668,0.5714707985667845,0.48056263458258697,0.5369839117388636,
 0.48760133473025924,0.1337137843950671,0.5806187965507748,0.2121120830781874,
 0.2476923838901267,0.02797490936736935,0.8618936679436828,0.9837474981443475,
 0.783509058963133,0.0563762253853356,0.8079041729192664,0.12400978145186659,
 0.3352210359829212,0.5323229105205163,0.9961824999452086,0.6945723354852821,
 0.5120754113896658,0.9247084707212752,0.7713928509325979,0.7571998464054142,
 0.7900300453036184,0.16362361255195546,0.4730912842753284,
 0.45017863780928813,0.47028962928686946,0.1457729548599842,
 0.9317763543040158,0.721078899543954,0.7867174137752903,0.9043733807791203,
 0.35635968599828194,0.6613964927822726,0.11275351440562931,
 0.46965525594986257,0.4766634633661595,0.7745336631090357,
 0.20776608169908617,0.13643009795331607,0.11693372385548295,
 0.20848415258334885,0.3265032956600593,0.8919364051607486,0.5695683844598622,
 0.5748200371246766,0.25671706360923974,0.7727283705097927,
 0.029928809936391132,0.03185725231222647,0.1486032675017086,
 0.11860921202648023,0.9880486119770083,0.47210057195292243,
 0.15041416555514708,0.7072886950611773,0.7601690488724001,0.3260079293671575,
 0.6195310087718094,0.5196722608565796,0.012390532554753309,
 0.5213296877049427,0.15455217868337923,0.6741846647103611,0.0858951169951887,
 0.304308589135326,0.7080337372844219,0.9495233861054779,0.02965739867302497,
 0.9528671631031642,0.6744902795195898,0.9728276697011201,0.5509644442447512,
 0.6783970941165102,0.7154273484110829,0.2686437635276179,0.9110625700071069,
 0.479301076250915,0.028171003277813855,0.5310076203548871,0.8812969660167383,
 0.4820400010888257,0.21181154784364276,0.4014621142771363,
 0.30313240614610026,0.8259898125721946,0.6927652209688921,0.5695063408491079,
 0.527869532850445,0.126089259053626,0.3180492190258587,0.24943534480862983,
 0.9084072614186007,0.5376020039262478,0.11498970163046429,0.9942516135360395,
 0.9589310121769042,0.7857630433118702,0.036440088900645984,
 0.9220985283417616,0.38319477893758025,0.666847567152896,0.8824103024481261,
 0.43525491901084334,0.3759959993646993,0.9857554542867848,
 0.12725627209291912,0.8173746614853057,0.666922135000956,0.7370283109174475,
 0.7520161750773041,0.5144751019401151,0.547814071977004,0.1970001231688543,
 0.49239136696744434,0.17406586647614386,0.1392197994415506,
 0.9768989741391026,0.31957761961527453,0.06716894538557239,
 0.7316712073599925,0.48532792614403997,0.02619953625100646,
 0.5672620365640831,0.6981920836511204,0.4868635788387504,0.08798270106817052,
 0.7990338279471855,0.5041326044833231,0.36947887048462846,0.7340997420047763,
 0.36877699221523874,0.9594861052954988,0.7513474426239023,0.5363345809318601,
 0.7249655491317124,0.5907832778862122,0.5902531755359808,0.05736618270184768,
 0.04703912057406434,0.6121238388774518,0.07735810050829917,
 0.0068745329554815715,0.18642146117095837,0.629053034131067,
 0.39169965759122105,0.6583223323469848,0.24869092015901617,
 0.5926302623577956,0.699449074399619,0.21232642756063758,0.8292334450350028,
 0.5069575950901813,0.5309495964803168,0.24480256270263345,0.5418262677634123,
 0.9214210812464602,0.7834872347667777,0.7796356339001929,0.21804627448282266,
 0.33446230090756823,0.16869702055458458,0.08228187519982433,
 0.17983346164037772,0.47366311114181503,0.5710314766203984,
 0.2497359205061881,0.08821573882798639,0.0876441918402312,0.413020995404207,
 0.4430755220007637,0.5293705508132286,0.08607175434553371,0.5355903116065243,
 0.643906109743708,0.5200067468864126,0.1211218019084691,0.12081822059933156,
 0.823491165944823,0.5423435741143081,0.307771171297306,0.2869675895566083,
 0.9568981566190026,0.10215549582758132,0.8336538062257048,
 0.29989198887796564,0.6142663894013142,0.5150926735953428,0.5510084482628376,
 0.03463001028766499,0.7210894669241261,0.9206827431191278,0.6489442908190478,
 0.8808462655531683,0.2938248511900299,0.40783471517243186,0.822765570179156,
 0.05584592643840158,0.5523502067122308,0.6335824668007417,0.8765133331076821,
 0.12635133917625707,0.5580777768391072,0.7189954499259089,0.5987655588294812,
 0.05716828082689096,0.16561396260642613,0.9822584698762161,
 0.8277644048361342,0.5169188187543821,0.2812887615345693,0.15465298125940286,
 0.024282136723475567,0.8705696615368254,0.37216314215751267,
 0.21560636153401425,0.3154857962935088,0.9641094757425265,0.8186728925082191,
 0.09354415691257423,0.6920822219371817,0.423698842308621,0.42727029378482584,
 0.14550069881902117,0.6018206816866931,0.3711830777780669,0.8802417549851207,
 0.7779243801047893,0.9296364699819728,0.35487731621162033,0.4475269348955033,
 0.519039185810877,0.7497033322633062,0.5839022095675036,0.6082557729604819,
 0.009482629502874929,0.24359088699662712,0.35232418631355933,
 0.25538312973081256,0.1914770285844658,0.8014913216012001,
 0.16285487130665965,0.6372737042614929,0.6645163442667363,0.3607603220202149,
 0.4462598277876586,0.9685166894366993,0.769103180798404,0.9853780791785798,
 0.0728710695455067,0.19236412263096536,0.4103148028637924,
 0.23817425120122582,0.05914240276765781,0.6412374136786158,
 0.2216664146906301,0.487124674680538,0.24050744596483187,0.33477366177586276,
 0.2901919006040472,0.617645063870267,0.22266546460379466,0.3186788947601822,
 0.00864952686470044,0.6049943688086952,0.8805697959124524,
 0.12160832616589934,0.5908856787941432,0.901083815728524,0.27160278596814325,
 0.2400544243318523,0.6025369254159887,0.8532797163971296,0.7739177767058355,
 0.0957174471869664,0.36035200008255186,0.8551290203256616,0.8429258414140285,
 0.03474336623673402,0.3107115240317575,0.33119446402146613,
 0.16279459611891212,0.6368023703663765,0.2724256820846508,0.6020842564902561,
 0.7637771861147097,0.2686024828240017,0.43744184457949886,
 0.26684055843063526,0.32856856001922385,0.35898628499001584,
 0.8324951228415918,0.2636141022026143,0.4195325582595095,0.5449629298204156,
 0.6491850404428046,0.9033070198160165,0.11987223251803036,0.5501903627794389,
 0.771041453475801,0.9011259952179007,0.4778164466672118,0.7547467789909847,
 0.5087782955342101,0.15964177554725245,0.26514779936487587,
 0.6428075275200307,0.7219781737348042,0.9259880755707083,0.9912065943918663,
 0.7016261237456807,0.9830361274745035,0.16642776834294748,0.5725386514065781,
 0.889692343897874,0.14716250815063392,0.5338700457452926,0.3522648807440171,
 0.019102583013815955,0.5725274752077932,0.8500369226186806,
 0.6220175836662425,0.17432537248592272,0.7078284335197159,0.2684020024864384,
 0.8311976352315232,0.6785169899594936,0.20894351828730606,0.9833549502968351,
 0.4636917690387049,0.35284601622489276,0.5514587207166695,0.7845557796686403,
 0.3364758858592475,0.05057739832537933,0.3656569229147879,0.4466000427516702,
 0.2996508739093253,0.1503968292385155,0.604735135041921,0.5013461407341345,
 0.7197295691194696,0.38316165055873985,0.5365069509937281,0.3323721461449113,
 0.13574379084616317,0.4816441324784959,0.6771935355355826,0.3227975103185978,
 0.9531871400149073,0.9473642511004723,0.0422694472535029,0.564659025867686,
 0.4487212413030981,0.7882056387136622,0.23820091862748005,
 0.15258398921237004,0.8928784530862952,0.6376282939843199,
 0.17656594772185663,0.5275266028311557,0.5529052563014183,0.4361974428505565,
 0.24472106715911612,0.1096260953046394,0.5272657574939871,0.6461329263853691,
 0.8879284095195228,0.2775848516298679,0.5568858221684363,0.6553496172738555,
 0.7632191933072802,0.8026638807097397,0.6962681966835127,0.75888614333423,
 0.056231377293367246,0.685332740453546,0.2692518215977334,
 0.40466911230380287,0.6686027153641416,0.22353414983208786,
 0.9195007449682167,0.8512031395126874,0.7434753282088451,0.843545981369499,
 0.8426218193713453,0.32196188277097637,0.013684198499138045,0.58918465038414,
 0.19743683465466644,0.9749678528323965,0.19921690919958157,
 0.5425990106537409,0.889277483738435,0.07503254289875105,0.09825774502970774,
 0.8731148253140888,0.7111472928303795,0.0806424097513494,0.9549944425720582,
 0.6180185993436965,0.9024587625545013,0.9596280920156613,0.5348777102537529,
 0.8446781572072697,0.7231086762110163,0.7409009316280355,0.8725342319499383,
 0.8679703660241666,0.04981355084203487,0.8489468316866268,0.8916090428437418,
 0.6982504086889115,0.6190717226094119,0.0026259830956762986,
 0.09276553412485622,0.5463243886643123,0.004587134150973449,
 0.6690228093886936,0.12930863824205097,0.42357442365623843,
 0.8965005552781659,0.35184168011496764,0.9535930285779906,0.7408670229355975,
 0.6583306879674576,0.6707252621393519,0.13257272878373882,0.4848599865671397,
 0.43634913330595637,0.28672561402942953,0.09589873466504295,
 0.6613037647042751,0.9703915521519938,0.3988836602350695,0.9971928275687029,
 0.0289693595856787,0.4055592840839731,0.14982319590603388,0.6094589861645315,
 0.08832824267344208,0.9195542907840217,0.9574283419671652,0.5631663132089854,
 0.5249960253528402,0.002190073254989877,0.23873076513904878,
 0.2363060489697275,0.29837979724178465,0.5720700723543766,0.644649381595344,
 0.15585300639132138,0.026130449020296487,0.7994132047772434,
 0.3571995348975423,0.12771461198830913,0.028029849341069446,0.54924218051431,
 0.21701195619755365,0.12715400740871408,0.42151882557083786,
 0.2659609544834233,0.41184525948280104,0.9452561423856687,0.7136609314385318,
 0.2699127276999408,0.3401539435036197,0.9211942591193732,0.4672455479935578,
 0.9802667417293919,0.7637348159777534,0.9459837828306128,0.5318337345669568,
 0.38234971265771656,0.6052900932300161,0.7770716534776723,0.9093410668767729,
 0.40569459234316574,0.8056040198021895,0.33846169504097245,
 0.8451674093378968,0.6249345913522123,0.22846217155589355,0.9371245455752503,
 0.32643809865504203,0.9051915129596627,0.05908401714063127,
 0.05396617321298858,0.8180439235673941,0.33504253946649576,
 0.37701456090789986,0.8475268611516432,0.8126938552997733,0.2613701691383713,
 0.23381405324625004,0.2701935664074312,0.15912838610547764,0.302366552919317,
 0.07381044741563936,0.5945740918623529,0.06629597285667421,0.285018042423404,
 0.5108884789982155,0.7540510814541802,0.10820901522906645,
 0.020375533587958072,0.7774505490246023,0.9095415216880366,
 0.8367805734799646,0.11241863405811126,0.9299775346542426,0.9746955194311314,
 0.5345019191893698,0.34458638081065374,0.09727532070025136,
 0.9080044845470138,0.9669081702044662,0.49856612158834324,
 0.48476925325757114,0.08404414446103514,0.32666166867787694,
 0.815663700264512,0.2730377064916887,0.6044396762275485,0.335583026444384,
 0.7357679986877678,0.25393555275875757,0.016269270215920506,
 0.5394961359069974,0.8615427929501461,0.2948636216476975,0.9012349343149713,
 0.564892025436296,0.12159365193437999,0.7365068190636481,0.2901009041815965,
 0.666668920373513,0.3449489388648973,0.0406721443083905,0.4424255057401726,
 0.7566113201031751,0.38928379300151217,0.9249909121512816,0.9868251537272951,
 0.38952905585662556,0.2534982112702927,0.5212673906195268,
 0.47200504422598977,0.5478237687205681,0.09825091408076692,
 0.3490874940636959,0.2402524689960185,0.1750289531693655,0.6177861598521529,
 0.5854642843709923,0.42488559958434713,0.6402082385165608,0.9770721483915747,
 0.7756439542848751,0.561901205521652,0.948038834439175,0.42191078381925506,
 0.7176524166512722,0.15575087862297154,0.8604672346109793,
 0.27244949428375564,0.47247754308196477,0.0033641502817940427,
 0.43722001022655366,0.31794731408780486,0.9549847387973366,
 0.040472455727531775,0.053388789229932065,0.4867020635675303,
 0.04703453354835907,0.8083018830602182,0.1332088816786109,0.4412355301448717,
 0.436460537859544,0.5179890370091841,0.2881599626326591,0.7224887337053233,
 0.004186666810062634,0.7172542986172743,0.6042635731566237,
 0.5567083766935196,0.01815097547555322,0.9213704640691093,0.8073451644499307,
 0.6388789951295393,0.03597084186419397,0.5345074392911604,0.7486843500754378,
 0.4910587679566466,0.9166594063862534,0.14115949989199916,0.7087083930688847,
 0.4481493766777389,0.41525706935124584,0.6049408250709671,0.992519877240419,
 0.030626754510137855,0.9623637149845977,0.24863325783449308,
 0.2335142146853667,0.8067229041189846,0.2616539628686161,0.4009368822744783,
 0.7500242716253355,0.10269605479343868,0.7495814390249324,0.764935237828061,
 0.5657615848585501,0.18230361557270225,0.48221921438171933,
 0.046904912265579024,0.16140858441889194,0.7070623115736889,
 0.752012738276581,0.012453854060297753,0.9384341332862255,
 0.12352768183664986,0.06508790843862322,0.1179433727395992,
 0.5322346967214715,0.11397737215991677,0.12667342770955636,
 0.8788138062505253,0.39946261872339184,0.9725697642875752,0.9221521312888217,
 0.6000187190862898,0.18826765753195185,0.6009971446868307,0.1405605907759797,
 0.5642421205840912,0.4222385356216706,0.7306497664364053,0.9456307498992524,
 0.7343872159119386,0.49742063261033687,0.8659291909763285,0.3784768232962742,
 0.05955806634934058,0.4692211431407378,0.9925486905199434,0.5765058731431749,
 0.2298802848905448,0.1214829473188086,0.1478706343861187,0.0683855029663275,
 0.28973778088586666,0.4830809506305298,0.4979291537418683,
 0.15844621801097047,0.24062549401449318,0.1914455574392382,
 0.1257625635399069,0.09161780516886953,0.7807393840571414,0.6399453776831913,
 0.16458733728669095,0.40736080263863217,0.4433377213948313,
 0.7479696713260514,0.7291629843790217,0.38928690438081714,0.8877508120954927,
 0.48217074633423307,0.09159019943824098,0.4947753235921095,
 0.2703145328422929,0.27978158957722776,0.39622638379128383,
 0.5788801994376016,0.06474541637925013,0.5821713738283607,0.5707909473847294,
 0.4834173353561696,0.3776531840951409,0.21396119481866926,0.5711870044365746,
 0.8074533845229983,0.33924557441995484,0.9346925551325814,
 0.31005770762624185,0.9176047550337282,0.10967542716498035,
 0.5468545241397346,0.9444419273692262,0.68060439882138,0.6177631226150236,
 0.6306178059681488,0.580784677387509,0.8669977361673342,0.3929897680628003,
 0.8178183712290674,0.1954988208474615,0.654172222524748,0.8306955913632684,
 0.791297958703123,0.38732163070099945,0.581305330986882,0.8974382673508683,
 0.9150722014953167,0.16471849297242602,0.5851282238585161,
 0.21876853214244563,0.0660828263812161,0.8754286882471566,
 0.10878360743175641,0.39737542465059406,0.18042261178175412,
 0.06320059965210123,0.15186573433887696,0.45749301722245217,
 0.5336524266498047,0.3745292785103015,0.22770640837409173,
 0.15165676022048513,0.9643298669781678,0.35478117311301705,0.612518416372492,
 0.6615838736342318,0.8355187195435233,0.7029281619125153,0.265053720560142,
 0.9614930452033272,0.9481389888407633,0.3693249457976534,0.8080247271142793,
 0.01899134688876436,0.8581969698021635,0.9302434550054999,
 0.13488799509341054,0.8851882498349376,0.08023066186947314,
 0.48744460050868454,0.16517030901466823,0.8132983977490972,
 0.7616016287752927,0.849689210275505,0.12553564169589781,0.334827514398479,
 0.11650317250643505,0.9217539294113023,0.7674070270287424,0.6931621226603591,
 0.6848127104901445,0.8056096280615266,0.8178432304097029,0.31602160107967825,
 0.78063661464407,0.38742542676813485,0.1374078546188886,0.06527739883611483,
 0.32699882114856327,0.10573383017604088,0.13121402590789188,
 0.11297850607855331,0.4836663419855006,0.6255615603807243,
 0.36326553954338614,0.4581002527383802,0.955957529579331,0.5548206198935519,
 0.734480872756261,0.4441254622754326,0.18702804388393934,0.5947726504574713,
 0.6171180119031702,0.5816809581011483,0.9289432600361168,0.9591936953182854,
 0.5166850101744568,0.3952849585619461,0.15488849720259645,
 0.20435491588754728,0.9350373043351999,0.7586267320118326,0.9838873188933392,
 0.3030033802729616,0.9047897954847958,0.41415699686822205,0.5347142732998474,
 0.28302842866500844,0.7094403842288478,0.22022518710033623,
 0.9884487005999126,0.4134264951257194,0.2362145741452355,0.9718110802346527,
 0.5321822458102853,0.7307107412797206,0.836208375625994,0.9156411989234169,
 0.8435749296266574,0.6324510941145939,0.9736565997020936,0.39003294857727766,
 0.1956115013856512,0.13272444499894043,0.24263158068037582,
 0.3585732896538817,0.47198431137235797,0.41982224001136526,
 0.3995837017851317,0.8017057810048882,0.5272976817349982,0.4920995376497165,
 0.8288699209893674,0.0420481837579022,0.9997201207871954,0.8269979720784607,
 0.9643373748550573,0.21848419980475398,0.2163006731151299,0.8254416794334227,
 0.07310394074784421,0.7546585090874661,0.8176080049556342,0.6755798340021538,
 0.242197916872939,0.33168276796896023,0.9055037988472412,0.604907796610941,
 0.5529328160015757,0.29105733147833823,0.9548627405024237,0.8335279118643438,
 0.20417466769916548,0.8695965047120692,0.04198078009702755,
 0.8099628805677036,0.8236757379705919,0.5787871744381071,0.3388343242369274,
 0.9409413626969539,0.5097050180751456,0.8489607757968256,0.10828616161342275,
 0.30485770475894547,0.9651735864613722,0.41008392488679635,
 0.022260621137490277,0.18728092834473198,0.9799684379904264,
 0.6081564926077583,0.5080208835371706,0.9206962851886783,
 0.021657081871515493,0.7171791994241479,0.07846676714687917,
 0.4270892022706474,0.5229391858807143,0.7019117931701272,0.2732622980273214,
 0.08886481784697808,0.039269333430878794,0.3789922018095254,
 0.30090438567219946,0.4512723263475482,0.3690960153089059,0.3960603396394202,
 0.9550998185817705,0.6700325293496656,0.04178566765437097,0.6908477488733838,
 0.6586357009690901,0.3189101452452987,0.9214501817966981,0.8058289881299612,
 0.3334004989787702,0.3240146680138841,0.35037870012301664,
 0.07660033641722674,0.125482858028788,0.17217589140010514,0.7121391505907227,
 0.6410890515143983,0.7064251258440594,0.8824403283344394,0.13157442729997815,
 0.6583671469255159,0.7310222221857021,0.02822175982288866,0.0793893727088486,
 0.28909368476057995,0.8499655912741012,0.492973178907248,0.9749259465298762,
 0.7173609527226044,0.3223624933668814,0.8782228090789361,0.3943969782613723,
 0.37473229986975976,0.6004524696080404,0.8043672609763468,0.2972952795241742,
 0.27042834487451706,0.28768740405011073,0.8075757907998521,
 0.5653203532210206,0.17137602819808728,0.11468807202979692,
 0.6543515569533215,0.12033927170860925,0.9866410121870035,0.2849428556822827,
 0.08058266447753182,0.6429578160222875,0.6816107644874918,0.960711441374288,
 0.007843922523359192,0.1250764775042419,0.6531086411373387,
 0.5040384280434322,0.24615580887366073,0.058667700444721804,
 0.076962378759729,0.1854225207627339,0.2694573403354191,0.4577981530458731,
 0.03604600841742889,0.9319559156726969,0.13826385773604222,
 0.8178323296531855,0.9657019165331525,0.49325235784433774,0.5581687994764003,
 0.9264141948975801,0.9119687881710117,0.759144141222929,0.2460713241091762,
 0.9033025764528523,0.2099309860316404,0.5882820591425404,0.06307816855355852,
 0.2460070699055772,0.8985340338506549,0.5608473804517711,0.0493996040667114,
 0.9221829086606846,0.4118279652424188,0.6454914983470316,0.9524563473637264,
 0.10309182672550676,0.695767134394035,0.8386879770182635,0.43441022499373816,
 0.26170367481255363,0.06737854390674602,0.9693189562854411,
 0.5955417912847686,0.8029199239318408,0.8119823334108568,0.2909706589226664,
 0.18996420127214064,0.5135659490406984,0.16331428489987454,
 0.06888830888491915,0.7388283946727598,0.5115425641037948,0.6431973831000536,
 0.41268803485291683,0.47599043779992045,0.8253813942382148,
 0.8501499492616054,0.22657278987514984,0.8421021537110781,
 0.25816928311540543,0.8793765638491984,0.410901162998088,0.2289046921234581,
 0.5317195098984744,0.7238796903001462,0.4474160957415262,0.4193714747698558,
 0.44820914638443776,0.28238879266675776,0.25590944855135955,
 0.6788996937607169,0.8433635413291872,0.3889280360561832,0.7554618206843571,
 0.18841164915425768,0.5653787693593849,0.11866560894077882,
 0.21769053583778442,0.733160342049374,0.48554105904764167,
 0.06360059705729582,0.10291053472178036,0.925871545347609,
 0.37475476297331944,0.6736989786194137,0.045981197338881175,
 0.9907325196063732,0.2121959376348853,0.22604481778755892,0.5050704543666853,
 0.20282859441755874,0.9280329984469718,0.6253711652042555,0.639052207682705,
 0.29098177837056505,0.9450228994632434,0.5897280589947421,
 0.008588593310959958,0.6349353658244196,0.9856612110247307,
 0.2856604548645203,0.14217003083722135,0.7420857455230688,
 0.29099010198090747,0.11357055151369166,0.5014568922635607,
 0.5660458282192435,0.921769274982923,0.419471611661551,0.8622273103120175,
 0.16437762179593196,0.9201933761275453,0.1904026811981041,0.4498639974759996,
 0.11173158587801824,0.14978447990739063,0.5911703633597086,
 0.6947707287105153,0.6609314864244914,0.39590979690560923,
 0.057032562521804664,0.44802509936608215,0.24738762197517716,
 0.8695937346228635,0.06922773257662707,0.6442330398770618,0.6599488037805823,
 0.9438613872858737,0.10921286891363613,0.6830036059536017,0.686607789218006,
 0.8833766363385882,0.5754403852918181,0.1689766377221289,0.5929520054882997,
 0.9305664304716018,0.319552687187149,0.4180029205026049,0.25317739290077723,
 0.33786749294115626,0.08955007720249508,0.9451871696706906,
 0.6039383740683473,0.7049026279813158,0.850966096227962,0.24475622871289693,
 0.48325464990711076,0.2187855688786431,0.8760551864774695,0.8907433169194784,
 0.7249146870244525,0.9578117935868444,0.6978443013341217,0.7110459343081551,
 0.9743195732662318,0.6586539292456542,0.6084301396715268,0.8933622796456024,
 0.1735140679378171,0.6351738153573759,0.3517340711019714,0.12852691917382286,
 0.47047197316057787,0.5788760042767882,0.2848169675063461,
 0.028108661191548312,0.3990060613568738,0.39456913516903314,
 0.4167642874694888,0.6189882337845058,0.7008697670348364,0.15525926615211372,
 0.3064966873811962,0.3815433007178709,0.09981510040986874,0.6544558987018086,
 0.1849902467132094,0.31625996721258287,0.13804827312070944,0.516373457220576,
 0.30864960941174946,0.00489565785777879,0.8764044859868618,
 0.3950935369106404,0.3700514366186648,0.08573997909236764,
 0.08848265013965972,0.6066793817380818,0.7121648339726859,
 0.45200089117191533,0.6389780269113903,0.6276780638823727,0.9863466975255831,
 0.6699787152701567,0.693546036427093,0.8643511568022619,0.04095574415132197,
 0.0649787442848676,0.8886574419056397,0.07915698110550529,0.8416538906702442,
 0.05430403870387812,0.18769904450845232,0.8458526810510832,
 0.21005551791943178,0.06749999055868283,0.39291417368772297,
 0.8205633723019816,0.23520084159487187,0.1883834835118886,0.8908750866954679,
 0.2129714475207094,0.6196560170344637,0.5919770325370811,0.9610459311559179,
 0.105559171785932,0.5574889926111009,0.7045197974367702,0.7300087209248901,
 0.8189496514572755,0.7295879707919768,0.38539506396992684,
 0.38819715392774523,0.3679553890415975,0.2532500778755804,0.9760701395228143,
 0.5673881193425491,0.8211559358274095,0.1882192967115972,0.07833246785884074,
 0.7414399517861294,0.34471496462071016,0.39144897700804704,
 0.08971058316808878,0.12526242383658137,0.950020912762076,0.9264790024220015,
 0.15820309824003376,0.3962058490659697,0.6446187865707096,0.2932790180695769,
 0.13776265494433504,0.6705831803033377,0.48158808001289677,
 0.6812328230265541,0.3819664736845926,0.6565556280245209,
 0.011054185699592045,0.563487160325294,0.573260296146888,0.8618389159469142,
 0.6742221093993782,0.6304145213488647,0.9384313295492631,0.25397713582538206,
 0.23141543526525798,0.7219309003657817,0.798098851264113,0.7056124126563637,
 0.30259466801757184,0.990249073369153,0.22516858831880282,
 0.13152789776690432,0.06894740418967782,0.6341130408177542,
 0.9361951595994202,0.9636032530934798,0.47623352493172555,
 0.32091355982648884,0.15881536587281653,0.767863382699639,0.6301168849561124,
 0.9968045356220259,0.11155330778264827,0.7400635659675286,0.931300019989986,
 0.47658673407822605,0.8789391771066124,0.9813360797665566,
 0.30982134138749484,0.6467807003354036,0.18057813485642482,
 0.7054416778306672,0.5602366000389903,0.9209284233517656,0.7060349067856286,
 0.760205864794119,0.4740889439358055,0.14420628790630696,0.7703468766429157,
 0.28399978041053997,0.7064912600348937,0.9494724363145787,
 0.37868305062938423,0.9733947227323687,0.9421352450725301,0.94831012322634,
 0.005091821461116974,0.8788458420892162,0.39201587693461637,
 0.13985655519242268,0.26746143399739797,0.957020367765995,0.8420297133230578,
 0.6442383425404045,0.3471229015931073,0.5794877484270733,0.24231579848651374,
 0.43590995355797535,0.45706846212530505,0.9585271669870498,
 0.5240674868179199,0.8303216690072466,0.6465497457323314,0.7385525085349569,
 0.7392536056386094,0.752264690504103,0.42452151382968273,0.5690021036366749,
 0.8819676691854242,0.775151618690533,0.7401483214879332,0.9139538894839019,
 0.528837717764656,0.353903037105631,0.0325245598249958,0.5912978626388792,
 0.18388381850090063,0.24706534614156594,0.22585018095290144,
 0.7127614861558765,0.31515936312085313,0.3940485218272257,
 0.20018056245344673,0.07482548397554445,0.5525934109670386,
 0.8570638981652414,0.516557598296354,0.16905184543667273,0.2553017883913463,
 0.29051117748529076,0.12026981773912859,0.5946802738797448,
 0.1455086618708703,0.46188637954138845,0.842389029578428,0.44689551872617594,
 0.4414231679615368,0.45337470859969087,0.12790479555860101,
 0.8126843960467324,0.15467950475293302,0.1389823029694639,
 0.22672444462701757,0.8760164573637126,0.7300496497803755,0.6834852584983501,
 0.3585622268579941,0.2791424022457155,0.5732860500416577,0.29286693369041095,
 0.9075132095106601,0.07630905400236432,0.8774913718036551,0.5851825878662977,
 0.019453886128226405,0.7228825673377413,0.03758543791320945,
 0.7361957329953339,0.7074120384436818,0.8841198917794006,0.591291426002293,
 0.4143097592770705,0.31965644628817336,0.06781987145424662,0.678984191057238,
 0.7971312205227106,0.3769802833338872,0.028293708945394735,
 0.1876408749604076,0.09448333524429775,0.6225492134730745,
 0.49588824948650667,0.786064242981369,0.1267439640103344,0.3950439436559905,
 0.6280117802198499,0.7995232246447238,0.21029310541069168,0.3553728708308901,
 0.5481166504667732,0.5200881331027581,0.0067190755392285695,
 0.8562869896677529,0.9875247907361974,0.20335631822383138,0.9723292352510698,
 0.971939548157988,0.4371962766332851,0.5289812929029178,0.8315789534802817,
 0.6669919218380058,0.6154957802842393,0.6398738122563202,0.6114495726403345,
 0.26277106430525543,0.2735187781781334,0.39405236880838845,
 0.32356073323352996,0.934880453486522,0.81011182379486,0.8186465480518552,
 0.06497654215672455,0.7096877830900203,0.42488447523432304,
 0.3951550872140447,2.2293896346070596e-4,0.21707326444426311,
 0.44194689391414477,0.6607498593686569,0.9240549875889446,
 0.42846484091623294,0.3254048064828343,0.41052654121859444,
 0.9939063696954074,0.020562633340040937,0.5104011188735919,
 0.5156337477658287,0.567257784355844,0.1582875168386021,0.5656559187322024,
 0.28819443197515027,0.3376158706644894,0.5687058225912803,
 0.21519668317579865,0.8118148845198508,0.30973710476327465,
 0.028186438365446875,0.8287161031387857,0.22327793078843106,
 0.7110184761582664,0.6148028746292274,0.2297393193240409,0.39174773630308357,
 0.7169679225386268,0.6904252158030684,0.7434050343368799,
 0.009032953695615853,0.22112555830380576,0.07345901494823182,
 0.07203915628267443,0.8389259889086422,0.741939795150587,0.8437949616888424,
 0.11326792586026047,0.87646832069277,0.5491709542880988,0.3596991167993102,
 0.1479470815764269,0.8384560365443597,0.4077864204542765,0.4092190372726856,
 0.18590515871306112,0.8880455769026241,0.37302572636776343,
 0.6800697696641906,0.050004571454949964,0.28073814208556946,
 0.24617787935482505,0.5720300779247963,0.2842725846454045,
 0.42664672551844696,0.27612535627890755,0.9454793876735089,0.695865947336671,
 0.932508895271976,0.9891540746827756,0.300007156706539,0.1983643011366748,
 0.5213917315351091,0.7776989374371347,0.019231504786266784,
 0.5782502075998505,0.5400383049356932,0.22002084606719796,0.5896295597029714,
 0.5080634767034693,0.35143334643411206,0.8068824706463203,0.6546358154042141,
 0.49436115462987984,0.5947025258921611,0.07224218743024757,0.301920142302161,
 0.787824328634819,0.25928753691189843,0.47868464112751563,
 0.16805713798975297,0.8212110995822024,0.12993025041962147,
 0.2067082760622977,0.6096307829164141,0.020096773404373103,
 0.7953627128117849,0.2773380299990401,0.4460571604691541,0.1561693488043151,
 0.4330276450063546,0.7648798160958118,0.17260615989139205,
 0.09078488855122568,0.9776600520505334,0.5472563423382812,0.9458127272688304,
 0.16652984810137805,0.1995299138230734,0.781469257675035,0.7706637833440495,
 0.45575133535034895,0.9506831070708299,0.4325272858872661,0.3906089462584873,
 0.3016089146385025,0.025227860676925618,0.5603962392254883,
 0.0731189256435889,0.27532840118004753,0.969003723168042,0.6499991994106142,
 0.5621849096945658,0.9328035383706517,0.6986410034175611,0.09358018706886417,
 0.026898773034260248,0.023537221656049545,0.5946937794178767,
 0.12297225869871475,0.8085807266052474,0.5598352815841717,
 0.013444150099646612,0.02634772603237645,0.20460963883524885,
 0.0070324982322290674,0.5239571622259668,0.7761651622389385,
 0.43619167548010873,0.07838487065833566,0.807018065752122,0.9899804938789637,
 0.8410554210053449,0.8031476887483441,0.3329700723740946,0.8179525494003789,
 0.4818649284303872,0.2811053191094256,0.11240113339135593,
 0.17594189729765486,0.6507612841076948,0.335121084151799,0.7712017345112149,
 0.6379354343209012,0.454472527422355,0.5256218278536037,0.5408986385376664,
 0.656610014197241,0.1988483715482381,0.4620179978080238,0.44333411656840827,
 0.09845006554211921,0.4895285253798316,0.18531947508614677,
 0.3243445050588548,0.8418246793411475,0.3818334470595284,0.19569164960118435,
 0.30470628062300564,0.9049341610158843,0.5884203220884702,0.6950702066346054,
 0.7983073872339691,0.5831864874595292,0.8773675011231017,0.8946847386710864,
 0.7031174601531056,0.2980337905606425,0.9850697355583934,0.4090980651157897,
 0.951654271456339,0.2977752068661025,0.5303313595963357,0.3279774942071323,
 0.495188638893463,0.1548304408154424,0.55073629500092,0.5535399571805879,
 0.012293528154417976,0.792322693162731,0.5138802453145349,0.8603600114670129,
 0.8151593200840287,0.2199278802376513,0.37370100744685597,0.8981260133745494,
 0.4321997683339278,0.9174276263418747,0.40967476270204917,
 0.25000748857420185,0.37411740691681383,0.5053725940521371,
 0.9911324425205779,0.9499860794969182,0.6287119557625285,0.3942432431100016,
 0.5209904154640155,0.32074012230778237,0.4123446499943062,
 0.16710173320032606,0.9539513990238263,0.6115718431500896,0.9488034442834113,
 0.5473391323931291,0.12310686895667144,0.8432450388035986,0.7418561858309242,
 0.9684361155917657,0.8801122861041221,0.9578724906129863,0.9115122894980525,
 0.2039880181385566,0.023828181653439917,0.3340840397395686,
 0.8782128636113131,0.018480606358683538,0.8417555560341087,
 0.4995149169414256,0.18710610183307785,0.36462603317075315,
 0.5284838095121814,0.8141697088781676,0.8079747021268134,0.6955861942564032,
 0.23868936480125869,0.557732481088313,0.34783510016147934,0.7143243962663199,
 0.24834895553346525,0.14687285066830746,0.6164417015192473,
 0.8171417080878365,0.35389446193579555,0.743697977862999,0.7563624383117442,
 0.8467371543771876,0.12485685191046392,0.348495688956863,0.4835795538409293,
 0.2534722830393983,0.7459522077297635,0.5688273045152533,0.8015136644320072,
 0.33299177109348,0.6981971477945648,0.3309270970149455,0.01597943278831504,
 0.38581381812242643,2.5898016941972024e-4,0.051918528480861426,
 0.7541609536818993,0.7825993821045947,0.01115821557258656,0.7138165805809706,
 0.03551793657930713,0.08663514380855852,0.7798528310813383,
 0.0723362116740266,0.1601840062604064,0.7407317892983258,0.08540942819411679,
 0.935656957058585,0.9579776752901208,0.43080612688398134,0.7510178763004027,
 0.7546822650950314,0.9549904346587816,0.6068174786544263,0.2632611867060579,
 0.24180902911378688,0.9670080127642771,0.42978893268976026,
 0.2911226447155383,0.1979662856133496,0.14436503091005903,
 0.02870589595820683,0.01934235562881359,0.4508320898968641,0.920058532004852,
 0.5391945909067212,0.48013043537986955,0.1814280724441112,0.7182202850191504,
 0.48592280429135704,0.5946128781742795,0.03423102089335661,
 0.7844828600254552,0.8917469693617593,0.02686594656671004,0.991423290114409,
 0.6448340392152448,0.6599582356172472,0.38857801821849747,
 0.04875784849415754,0.8753221184047976,0.19940284305435807,
 0.5608289493090391,0.9986708664376001,0.5433230153256654,
 0.018750606314422713,0.6475621732541155,0.9539712203528784,0.801389706932739,
 0.26501311613925904,0.6214288369748531,0.6741250133375764,
 0.04688816424411979,0.09325704100678411,0.07395180009066604,
 0.056338183089049476,0.8291151948806178,0.6841776636035398,
 0.07484596748945294,0.4860597029614868,0.0449785917616482,0.875344753238507,
 0.9096951720111435,0.3414735881734159,0.8418639443348945,0.4321169395545468,
 0.8197898397775778,0.2468751280906646,0.6084361984449075,0.622946405519133,
 0.8632955478449281,0.18484028259181762,0.009051586084385943,
 0.7967454604198912,0.5559773301626769,0.13203185888785507,0.7235329749436242,
 0.19707455821514908,0.7658754616029269,0.21039633869266594,0.894154790292947,
 0.3824038077318509,0.5054980052726669,0.7457666388329646,0.05964175328775578,
 0.6279057924526534,0.4044191852711193,0.9427588818033311,0.13076211875273946,
 0.2877685009442529,0.959858028932833,0.06205362783583768,0.9403447031883001,
 0.01015258318234813,0.06918540254142314,0.6953094403859359,
 0.4690798448308092,0.5730120727892254,0.0010856389852931247,
 0.8139401514949139,0.8284460367560965,0.08058828947199403,
 0.41570165243416524,0.06986112025234315,0.6970555958111945,
 0.42412766254060186,0.15036714349865943,0.22215906131826435,
 0.8751815692484688,0.011168097837859436,0.8323867289857384,
 0.7083725993345322,0.6435251006911349,0.6244454971730807,0.924318049338646,
 0.6901156520168805,0.9006998514522035,0.29757657816936334,
 0.20462935606171395,0.5790929205699227,0.582357684999751,0.2799273826847538,
 0.2048571998450932,0.07454686894883555,0.48680443959567854,
 0.10621981401042513,0.40128372886861774,0.0035262753296629867,
 0.23978273871303024,0.07280547152722017,0.13636308737167502,
 0.2050918410508351,0.907557684549404,0.256022575310328,0.9562636850132333,
 0.8889381174102681,0.3323231448118885,0.7748091345150917,0.8321511444701062,
 0.7667541966141662,0.3686343483442124,0.20325127383297714,
 0.11287305504400336,0.10667420299323827,0.8479772618377136,
 0.9061797697968847,0.9602999652146595,0.45876633131190026,0.9426263074779306,
 0.17415000075717557,0.174194840118586,0.23338120735266854,
 0.03497392168162883,0.01398397299878551,0.4006703812493929,
 0.35693718400299324,0.5521315088480057,0.6153306249976217,
 0.24448750232900773,0.04262824568785717,0.1508139266757358,
 0.1618165463010216,0.6997003402068969,0.35552079319305685,0.398538778677882,
 0.3202585059212947,0.9896996565339272,0.6795255837380149,0.8849357551147642,
 0.7534901205342823,0.1299462232821793,0.5057089027730953,0.2089399300027509,
 0.6142638380299278,0.4950245381268391,0.6452147519968301,0.875620751231076,
 0.5404904949987802,0.8057777825468033,0.19822743685661182,
 0.14484609628779743,0.1209550120240932,0.5617377857234861,0.6154139785958441,
 0.8568281375900473,0.4380249827411058,0.05630826557757651,0.8902490819822493,
 0.1840995715703203,0.1264931104220789,0.850753850108072,0.2590001755521846,
 0.791578445804542,0.08662244048597245,0.9321176693487209,0.8258538792806358,
 0.537071521798342,0.285501495699365,0.651355630030523,0.7874343948036396,
 0.8424046886184121,0.6992086535164235,0.8582502769741179,0.9764901940882779,
 0.5895807449445866,0.6250039933550213,0.08124067334852447,
 0.25564651553791984,0.2625344598850393,0.8926941177874353,0.9755952408723549,
 0.9745824030635082,0.46078122067308436,0.6504903156032614,0.6451356496888416,
 0.9545884215328989,0.17279980565532027,0.19233062888249775,
 0.7448157593776802,0.8018553101184372,0.3284421706808648,0.952777517164507,
 0.7011715843249812,0.7937173266038491,0.26578828581067915,0.9465683065529649,
 0.7021913537449572,0.3403796594977532,0.8505708730047475,0.4570841974220434,
 0.6000020266871918,0.9648789866444413,0.22438087382607275,
 0.012161307403909971,0.3824159379781691,0.5613573708442996,
 0.6483580032827094,0.13284595857838266,0.6119918017548007,
 0.38089476427103663,0.4221832750991954,0.03949072054395497,0.181055852817539,
 0.7025608514058524,0.3859077643414959,0.49580941660127387,0.6973840573164248,
 0.5101526890047677,0.8297218461538387,0.7794618799356798,0.3810269988669427,
 0.9228373557176587,0.7830700052108853,0.07187556152652719,0.3478025926999212,
 0.6305574183876724,0.8172042896651834,0.323223960353499,0.7501164682551833,
 0.2920528248953693,0.10755867341474645,0.1117146672714483,
 0.20377453242458665,0.8957582203746044,0.6666738958327116,0.1042943264692956,
 0.5294261027713505,0.21739624602131324,0.3010985539983898,
 0.36562809787276995,0.7757958821589469,0.6219474186730857,0.8864127350113613,
 0.1256500952268469,0.09430762784875402,0.6164794048426613,
 0.24306969298703696,0.2399583872576997,0.11352593160000923,
 0.7685951665769573,0.333187841533102,0.47540997492273507,0.9573390680392619,
 0.5272944085381724,0.07761215838657909,0.8042308456678844,0.9099651626057939,
 0.8419868268896811,0.9811927335955786,0.18575596023774033,0.5956503635397606,
 0.05630218233701856,0.2491640490413496,0.9324681413500411,0.3843926723150465,
 0.7919810131312107,0.4252052589184111,0.37718200255336365,
 0.25618752703473824,0.33194034422139385,0.8285257828069199,
 0.5450097420261846,0.7430009013472134,0.19200878619250372,0.4786884567423918,
 0.8751662444927364,0.7506926227145431,0.8630993416147343,0.7215456154675683,
 0.5205290985233451,0.6778532902925183,0.24059101648573922,0.5295269827335758,
 0.7456799844677514,0.9269510506163592,0.9207008070543935,0.3277577355490312,
 0.4351968304380631,0.02587515208393365,0.3619682066340437,0.0469635602284425,
 0.3172561012850126,0.2988224365950094,0.9565229551302293,0.2834596182712983,
 0.24682024372263056,0.9584658166605676,0.14763543746727859,
 0.43787619120152754,0.5664303540220972,0.49563593428641983,
 0.5190940243730016,0.3241620519634687,0.21057227947790735,0.3723222827822652,
 0.27047899103875084,0.3678451825087934,0.20893205438591422,
 0.3539589635738726,0.2527793240460684,0.6296099537525178,
 0.002960613139557422,0.39213991279754934,0.7679951489057902,
 0.2337390747821304,0.6932435878062828,0.14675606413788933,0.4148752259238997,
 0.4068938963824177,0.43007434839122005,0.9571059874529317,0.7789127710719578,
 0.10358661798940172,0.5095810333690414,0.8934641080303518,0.4396719956198537,
 0.3173270957372035,0.11443816659066519,0.9751326638477265,0.8490502942735237,
 0.37575045670392426,0.374382968917795,0.21640826751291187,0.6617682668665016,
 0.546700581769906,0.9513797055761642,0.36339318718995484,0.03716275217765397,
 0.5837729301726869,0.1613956460602446,0.9537220707686832,0.11510973391870716,
 0.6720094616599097,0.9733606677091973,0.7463620005590356,0.24468610808160074,
 0.6790914502215976,0.014847259440480154,0.3678968000838341,
 0.7219816844429982,0.7046128128220753,0.25378103279349973,0.0190523527831572,
 0.5818901819479524,0.5610703817053949,0.7034785616491107,0.8388305166863814,
 0.5638796171327286,0.6927684828731878,0.7311620946820927,0.1061570614988856,
 0.19181184241890326,0.21430965825359405,0.8465802622093466,
 0.4655237118281338,0.798920048728516,0.13110773102090234,0.9327037256805056,
 0.30376938614116,0.34486864221479174,0.2554821813933281,0.45120808488687225,
 0.5122576779182892,0.685983387720436,0.43571950438552587,0.6660914907206904,
 0.42932632659363457,0.7041340460371316,0.5570744421557706,0.6214314670065093,
 0.447160389162923,0.07783590018335718,0.6407224168337119,0.3732949253407235,
 0.6774187752990927,0.8810165862493384,0.09812325645985698,0.7383744427182984,
 0.8378641810148636,0.784342537992237,0.03029179990464781,0.03325570945977474,
 0.06588407505811378,0.8947148681444981,0.8382992825940828,0.8344560155781143,
 0.6437501389952925,0.889730361866389,0.9795195172836784,0.012082449431188413,
 0.06660677641935453,0.9947828144644035,0.20323798063331266,
 0.39296980600528697,0.2382029973901516,0.6038874940885636,
 0.22299356382265412,0.6488852946841351,0.11681404705732734,
 0.4231807912824497,0.8566222702888828,0.617205168319958,0.02696126489443651,
 0.1198591889524735,0.7104446949791855,0.9952365738619685,0.780973165738642,
 0.22408457553790473,0.9386298097439033,0.4361549585989375,
 0.058524054718591456,0.4005031767546601,0.33427321994729886,
 0.6187529653422827,0.35575515213226205,0.7723292657547745,0.8660579648098672,
 0.9729922416269003,0.7851595922865026,0.726764654023201,0.8871357858321821,
 0.4372963551042409,0.9671131562496956,0.1783906104909032,0.16475951707853997,
 0.3422716600470569,0.07503117567793705,0.9387743777671227,0.8176156533463337,
 0.6709852662220419,0.3058167564575801,0.5529045363730597,0.39639703024685535,
 0.6764505492590094,0.38801457345932755,0.4201380992185888,
 0.11110316340825088,0.1380487424336423,0.18070781605774877,
 0.7598783475556903,0.17132384888814844,0.8419123658228324,
 0.007163991960746019,0.2574865630831512,0.413806926311566,0.7709407602663727,
 0.6398971989932257,0.7784660699068373,0.565739154538518,0.9983565717864097,
 0.46613866695654804,0.09972396542863549,0.900159778808808,0.4017612793799399,
 0.8252615011123363,0.7125573957787475,0.34915258591259346,
 0.08015972231546209,0.9717397677245527,0.06418832048325296,0.886088876700465,
 0.8633151032360996,0.6276147451956418,0.2841470732143916,0.4716657761765388,
 0.040468597398559014,0.6592156605701358,0.6776535839113575,
 0.17819237088852335,0.015974802815472144,0.02526640477285924,
 0.35635004449606833,0.5946707455588373,0.9907034336374494,
 0.14851191948718223,0.508734583670442,0.8097893417284739,0.6776580293917208,
 0.9240067710183436,0.08758391289493295,0.8858895906689704,0.9137857531633424,
 0.1442226391236211,0.7544067793459512,0.36114349672069956,0.988937769459605,
 0.2264361062460467,0.8005843863131573,0.99626188096504,0.400497442694574,
 0.6702308603502761,0.917469660005629,0.3103869638513911,0.69700133623062,
 0.5209943217177897,0.9319925453389589,0.07389105391700246,
 0.35213417767321187,0.16422678203869268,0.24747437484367985,
 0.035671236176951004,0.9418959763663288,0.12323844485539903,
 0.698661097363643,0.08664608512406236,0.8535814447812582,0.9304744209646445,
 0.4024804017034992,0.2184052089256845,0.7155535775313884,0.48894532199458296,
 0.17212823090807605,0.7934671510481467,0.3243520933534736,
 0.22024032207452693,0.19257762291392888,0.6909097118050284,
 0.4996539524068887,0.8156558431259963,0.11843194957116532,
 0.27801562441319805,0.48156974463074576,0.713145542581862,0.7832375272346033,
 0.1465985830628418,0.7645280666601938,0.03888657675108875,
 0.26028498220063534,0.8684711746055963,0.24618021495285947,0.589785649514212,
 0.3704133969213177,0.6926644272672466,0.4077804760126187,0.6646343301428863,
 0.8892730257116958,0.6722602995791255,0.31665063198237364,0.6181472929892433,
 0.2829387283293664,0.6646794964291807,0.5656753833170881,0.1747453327423809,
 0.2808063662692043,0.5505795842593316,0.6930863494222522,0.5465158007134705,
 0.845140837083491,0.2622052685178329,0.6444861286849657,0.9823156553417066,
 0.4662105185335854,0.5704230066420386,0.4228375039604573,0.5398317636687195,
 0.0309757581889768,0.9544352147576223,0.6273498680677919,0.12453485621899629,
 0.22181518769872854,0.47152255849397084,0.16450371590193547,
 0.9531050461407823,0.199604165931217,0.11210634728612978,0.8263916081723041,
 0.8741568949820921,0.448687640801372,0.44549779238476395,0.9511833478694058,
 0.2991237062139507,0.5180175038079115,0.17338847343673391,0.8990793551805736,
 0.2731767756806506,0.7798301539056963,0.4635593110438091,0.4746032567469102,
 0.7976572101878211,0.8117813497751342,0.6006275046063034,0.6560591344358397,
 0.66335716791452,0.39255188599305346,0.5122432244304919,0.23280263314825733,
 0.07275525354744083,0.5347981595279798,0.8196440122401439,0.7452705118144718,
 0.9965614909470217,0.7024856652274502,0.3331960736552798,0.20929660320630217,
 0.715087110787808,0.030023578727314093,0.4668369449108992,
 0.10099717973474209,0.891640080014379,0.7449335596981504,0.19484739094476322,
 0.2127650832396727,0.9668822818503768,0.7181965997293769,0.4581072754101245,
 0.15689259931632626,0.952302756843379,0.3540540308020752,0.19131167341593436,
 0.9506380468113435,0.3682698903523177,0.9291743069164938,0.7128423756926359,
 0.6504167081353651,0.5308609730505951,0.11851214011852051,0.3947786068000335,
 0.42070986968336,0.7925095666560927,0.918707859336238,0.3264982250861552,
 0.4031534006198596,0.07036998312823106,0.49763405199909305,
 0.21496482710620468,0.020360517354546248,0.4518181695577774,
 0.974919247555698,0.08630548040408359,0.24409054959079446,
 0.03262507220097732,0.7772480430709441,0.9936924494089194,0.7625767488091049,
 0.8006831703250616,0.47890685648133247,0.6412936972255884,
 0.31417377339663943,0.33871098362993624,0.2521910104299039,
 0.9333302005871383,0.4328802441637698,0.9718569187345026,0.7241171545961538,
 0.28653094460975514,0.8083656809581015,0.49355629012131863,
 0.2924953063224711,0.6264463610595541,0.5330772014740376,0.25537663874543726,
 0.9052607082926643,0.01861985725390447,0.13680788682297806,
 0.6813076809519389,0.18563842380175555,0.8678875718546715,
 0.15929376773392678,0.7136588796531808,0.2163907897647842,0.8532774016448436,
 0.4210654990082696,0.6125650067418217,0.7045619613357108,0.3664649943743088,
 0.6554759810392196,0.0035999650412638085,0.0011426346328625714,
 0.24692192810478486,0.19265653492090618,0.8186506375384632,
 0.6133001612066813,0.9161779303006784,0.9960070674677389,0.12245284882742313,
 0.4795868323442072,0.09001705463404819,0.8807329556212926,0.4410287597859013,
 0.13664599022121104,0.8923288490625749,0.9392455367253428,0.4140213308599412,
 0.977543357474104,0.8510703862574888,0.9631970790928468,0.0244367147321799,
 0.25263096738536606,0.5045561166408712,0.9986625589784263,0.803259123655163,
 0.122516826121576,0.17062177368959464,0.6131107328545119,0.5885661371675545,
 0.23815361324881879,0.6464777990307837,0.7018221391316452,0.6139548024714522,
 0.13054543198022095,0.015270968518380057,0.5980246112423891,
 0.2569112374112551,0.5336612528260869,0.22266913677174183,0.5285485846857452,
 0.2320016863797063,0.46908720313591923,0.7103585086810201,0.4070993293620293,
 0.09577007880984767,0.3761732217888001,0.0956869239663033,0.8079662649327665,
 0.5158369491585163,0.9276728424641699,0.4912453703661679,0.5400328454950941,
 0.45741509401588654,0.5456965036054027,0.4279965268532012,0.6577840405320416,
 0.2890442166878031,0.6725956865931224,0.17185492718618545,
 0.26769610064450633,0.07473921706785791,0.604025062613852,
 0.22563717653542326,0.5386059119881692,0.738356634271178,0.7111934974494278,
 0.5525309411105646,0.9431550506759824,0.6578585640485575,0.7295132857021344,
 0.12213595473480265,0.1346933770326496,0.8187869428243775,0.5002130131772095,
 0.06589292217239728,0.22782324174006374,0.07247945040120529,
 0.8700328724511722,0.6257563047422591,0.6058733946011717,0.08131914012148456,
 0.235403799981048,0.6099511093805307,0.1744253561131649,0.04410260306787306,
 0.8808736317075263,0.48622187638254866,0.36394974670155666,
 0.4883340442092189,0.878374559066079,0.9088096005026228,0.11045180709672531,
 0.9104107075596423,0.5000980971916316,0.3331523228684983,0.7116694647401668,
 0.4353105479057352,0.8354051546659489,0.9887624364210312,0.29893564007823237,
 0.015620488866123883,0.8713908038947904,0.8916126283603851,
 0.3582724647555936,0.20350673657376728,0.8494242933514782,0.7627085128498737,
 0.8952565628877998,0.37817896084898783,0.34055279007298056,
 0.30215351222825526,0.20370925007124707,0.37809665886918953,
 0.24008907617421593,0.5841381191259498,0.8186610483933152,0.7379628776425786,
 0.5149988911330435,0.8226852639567112,0.9912698682659231,0.1202221266366914,
 0.03433374951428847,0.8170705625083475,0.8518662598235006,0.1940682759580633,
 0.273913707094503,0.707089814339111,0.502147513359405,0.412783850275635,
 0.34324242960756823,0.9953292284253157,0.18736356175353663,
 0.16150967065887845,0.027657805846773686,0.48934458625434574,
 0.3200635823248472,0.758012044529549,0.8836963620544896,0.02751672039124875,
 0.20143215840519357,0.9921383127596366,0.15763536880406892,
 0.5192815944692534,0.25286508228763604,0.879548473720938,0.321541917939959,
 0.6889130459791603,0.653020247357185,0.6514240418271752,0.9133966907125591,
 0.30503533228863666,0.365591915101636,0.09190825785037293,0.7070874371806578,
 0.7664396377716265,0.9942133595793193,0.7796532247763308,0.8796449202793909,
 0.8396480248655986,0.25701969902207455,0.03412921990019169,
 0.9310050677915492,0.27214385548549525,0.19102894547249505,
 0.24827913786134392,0.221071308795442,0.8902563868607039,0.5722002602374128,
 0.4272868506944254,0.6420201850421071,0.37909904284445206,
 0.14278048710763858,0.24335713465445485,0.8459519874361597,
 0.07679530179110938,0.1904935289688685,0.7503124635430418,0.9804076054974068,
 0.36257869581358293,0.46913865823743484,0.5376854528072924,
 0.3677961523889398,0.22728229477663087,0.9904411386116097,
 0.46207821431060414,0.5670434001830228,0.1396606818111098,0.1369810767492624,
 0.01824873724022147,0.9103129132617698,0.3758902557291154,0.94048446226592,
 0.32992232965891477,0.2752028708716824,0.9944719882772786,
 0.09828774985698685,0.03985207100788635,0.5830893210782391,
 0.5718345593308821,0.009943673193371527,0.607956033754279,0.9877321499106616,
 0.21555578263807007,0.24026960864201152,0.9857377210745503,
 0.2375261312032112,0.2778435229375167,0.7549303612301297,0.17110241261144565,
 0.34831863026675425,0.9120380669332634,0.11934185660617258,
 0.25984128118972605,0.27777709277990126,0.8738948139162486,
 0.5889727902586503,0.2301545985935347,0.483643863822407,0.38213104171249324,
 0.7264242848272895,0.1380726388450566,0.7912628818114212,0.8139845281199606,
 0.3247795543215297,0.7308307372010402,0.8220860577184408,0.39719887725350744,
 0.37426388357770657,0.5395797499141898,0.02814358488601887,
 0.13444966850502804,0.5741910799752467,0.6782945981971642,
 0.08890401133003911,0.6577566952361934,0.8536583758988288,0.22553745182107,
 0.6158223918000476,0.8708095075885751,0.9992222335866301,0.8349840540675633,
 0.024790689662400278,0.8779036798047449,0.986353472353833,0.5062577093142016,
 0.4871907987891656,0.8861121913762416,0.06866523302237038,
 0.32233901341752236,0.16849714163673435,0.9103941321819742,
 0.5069093913115261,0.7370813033006305,0.9691461505709704,0.5024949497168331,
 0.15865343286759515,0.9765914802078935,0.46309389500540876,
 0.9443443624655765,0.16195784738150842,0.2470401883399418,0.8581189403308159,
 0.9585767204013204,0.47328777880091333,0.9241221255113626,0.604212154745789,
 0.11107742889899352,0.7493738421688514,0.8909708418151819,0.9904252153938187,
 0.658523771054915,0.028228906352782412,0.6518425632525823,0.6263779384914624,
 0.5188157171779356,0.19241800329244818,0.05759166656205661,
 0.7895440040978254,0.08016319326028754,0.3127699916336695,0.7826355521410163,
 0.7558864901585024,0.42801110687002086,0.8990161943383298,0.5371297362332011,
 0.07788546457970225,0.3884607146257226,0.6817721204995971,0.8262438921555613,
 0.13742150436681833,0.9866221301755428,0.5203777982245283,0.5978233997433083,
 0.9348539615144111,0.4398239663072253,0.038519063455155655,0.333568123329823,
 0.6926701535273301,0.5331535675617518,0.4582306540888206,0.4629573713186542,
 0.15049639196761344,0.24411921406870418,0.39743656399538463,
 0.5815743720784267,0.5659068232002663,0.6008279351585648,0.8961043625692047,
 0.5921352163034643,0.7500202799794786,0.3322901509092352,0.4000813131352674,
 0.5373335365663261,0.23537299159545777,0.334322346860482,0.49165768880019245,
 0.7131469159645307,0.5579057433124646,0.13165099857681062,
 0.004000584678663599,0.48169499841028673,0.44268837579187204,
 0.12003182502112852,0.7135219009206807,0.7150824889756884,0.9560362425359501,
 0.9861099857394329,0.719267219436215,0.5503948025665477,0.46950396827146756,
 0.9619617634738591,0.6589302471197798,0.4033161231698348,0.8924450935847915,
 0.2750674286647895,0.0751367611253817,0.3547038476771217,0.6217978666047637,
 0.6926947334182065,0.6015932988163891,0.9655657776987585,0.6003109621890463,
 0.742029439330231,0.9452124025097723,0.6428885461906246,0.32068220956898474,
 0.7947703766685187,0.21030578691548685,0.7689109774605575,
 0.22959550386972816,0.30431164849714987,0.15607986747596447,
 0.39541342414627867,0.23732426104262194,0.902319849365922,0.0706807299294171,
 0.7589781589019461,0.34069656135824844,0.29063325793691797,
 0.5032311289579201,0.025546797808538524,0.8164769956242466,
 0.2258389459984741,0.3416183135589106,0.6155314729953054,0.3076629269826642,
 0.6406182993490238,0.5793947112839671,0.8378930513733553,0.7941125079881886,
 0.7524315289059728,0.8842181872549207,0.621655277604107,0.25939789266402247,
 0.7023205593343874,0.926780737858189,0.9414113360142378,0.49935327632956794,
 0.7248653140836154,0.2296731891674324,0.8399164912972674,0.9697756960559734,
 0.024119956544417342,0.24514393538526513,0.46737163407832916,
 0.20047514635689134,0.35047800749211977,0.3937198750252807,
 0.8245120268322665,0.5882612587440672,0.3135393107488432,0.32972984229511126,
 0.29774819424637033,0.6920970948148518,0.06029449665777897,
 0.34463679445148965,0.9683642189651094,0.13680033321574403,
 0.6787649290818806,0.7942789868338729,0.2753166955227565,0.2508876167111358,
 0.404482044175867,0.20044441765318433,0.48084306146870714,
 0.35215277856878224,0.6256822686305737,0.8001537486488105,0.3709812470696915,
 0.9292162594481619,0.19854031146598405,0.7664956094582334,0.7040839010454251,
 0.5091443578023678,0.8806271613690523,0.3211952299799685,0.5633993794806773,
 0.5223983394633084,0.9493535652421423,0.34413315836646374,0.3670219422770792,
 0.9098432181621071,0.6383339224234059,0.07379567133141685,0.9389858898912331,
 0.3816912521156085,0.9619887736108164,0.825013122866144,0.462966296299155,
 0.730206255239219,0.1885827602640313,0.6070829995166601,0.11981058820963542,
 0.1955827067382675,0.8521555768205418,0.19277461639007698,
 0.024038224627267013,0.27239263906362043,0.06095912344181553,
 0.7276978818685549,0.32080402799764207,0.08220388235255172,
 0.026940418078994943,0.11737846856667344,0.6890299783600604,
 0.3153417474673005,0.23325755080635402,0.8865860277659945,0.5338336479455729,
 0.9774165977927218,0.6061493814017439,0.054270802299165255,
 0.6687169064509524,0.20179828978773917,0.21828234639968358,
 0.4867791539680574,0.8541453384092976,0.49448580615711446,0.885973391170491,
 0.5912739369146883,0.48256574765586757,0.283849689721865,
 0.010132480687029877,0.8258093410525373,0.39737897721926274,
 0.5633256157683202,0.02673619858658216,0.6594784574541541,0.7450017829305802,
 0.13195921180000758,0.51845153577118,0.29717971092343964,0.37491700791101656,
 0.13327044813951572,0.6892925056986406,0.48307929176973363,
 0.06683922520437635,0.8060507148489682,0.2864281521395784,0.4127913115303118,
 0.5094973560188388,0.6185577162269014,0.4272285372887936,
 0.006640165107969165,0.22012632374681262,0.5399890417744381,
 0.9010638695516577,0.32152043231798655,0.0759061784947751,
 0.41940565717820055,0.29584940300319607,0.36396376969164557,
 0.7047783978514272,0.1773497810684822,0.19558121396175598,0.7493571298454871,
 0.3625286145452701,0.28572230410407495,0.936408321052624,0.27067126795523344,
 0.5326360338185294,0.6882730033678289,0.49918270756146976,
 0.30987378452936676,0.5569841897340975,0.6526916125158853,
 0.33381326702463043,0.5391336258685059,0.5710062727755056,
 0.033165436035048734,0.8635503908991747,0.7610274526577737,
 0.7368668743361568,0.45019469073412866,0.4421760620084927,
 0.37305479100143213,0.6548104370651928,0.12876022121078345,
 0.5422675788788023,0.5256331036306096,0.06539679317995939,0.7398740546148688,
 0.13369470040966436,0.1265578545364371,0.679563219742343,0.458055682945552,
 0.454671423314845,0.8546635378254944,0.21087128802094623,0.32366737479979635,
 0.18165475784334095,0.6935854678583554,0.2463931693926149,0.5201024125925563,
 0.41884655293882633,0.6816341747730384,0.43353675790765933,
 0.32008375162400204,0.5188966169512534,0.7474232703778271,
 0.13571122696946314,0.14632231914236904,0.8738404283597588,
 0.12543841946934675,0.015829452024892166,0.12978409956880577,
 0.06582273999065769,0.3390405518168095,0.8079785751890227,
 0.47237883052795393,0.9969428388612873,0.555678602607682,
 0.043960945994523715,0.07724955376528153,0.513523763743765,
 0.49426121350010044,0.26778409369027056,0.9225574610797059,
 0.23693334649626074,0.651755088778522,0.5093419929397837,0.5845347653258458,
 0.5339119269293153,0.8571766285533564,0.24035401889959662,
 0.21420269013122373,0.11205236173833388,0.9884444735740189,
 0.16697862082821824,0.44045899987676385,0.7433827485052873,
 0.5086977425503498,0.5737299758025999,0.16966111337242562,0.6268843863120178,
 0.9619627451581391,0.5040512136693981,0.08818775077851482,
 0.20368214766021664,0.3643864643197433,0.7227233669062166,0.7558763568778705,
 0.31855216961085453,0.011415718358494087,0.5083465578285024,
 0.011757672986399959,0.13354624545215354,0.4572490245255163,
 0.5380811786505256,0.4783043829041441,0.24614892587000392,0.5528098294379774,
 0.7565810841559436,0.7529055431430465,0.16029808099103127,0.5637534629917405,
 0.3509380522006996,0.604696859145893,0.29373392526789266,0.7802555269708662,
 0.28901608591952443,0.7824057645801195,0.5998736620314837,
 0.07951900957437741,0.06094738838899305,0.6718597655718489,
 0.07753439116773075,0.9217429517420648,0.8065741648074625,0.6972350428731827,
 0.9042587750310835,0.9000206624282674,0.46734134983705,0.16876640166714396,
 0.31365261372309683,0.09979418050738909,0.1084850107618176,
 0.4279084892185384,0.2773671589884906,0.17585285089257185,
 0.14702439466073647,0.18652700279975998,0.3720447784936335,
 0.1839881625343769,0.9324910339844329,0.0518716390952606,0.8632606119253188,
 0.07731258483983716,0.7994995355452034,0.4330289730293435,0.1323945476047491,
 0.7834301368645669,0.1964896553426726,0.500007868226171,0.8867291975408989,
 0.8613496546526884,0.5940831075902613,0.4076056491408674,0.6740519809174632,
 0.41265684265798697,0.6057703872251367,0.31000051036898724,
 0.6999516210342634,0.3105188012275244,0.21506543627632924,0.8308274907205156,
 0.5350549920585718,0.8603792811132883,0.6498804800020482,0.3119563981219642,
 0.727112101080078,0.4860716159702084,0.7529892965197346,0.7075266778913678,
 0.14041167087356554,0.6148474152304646,0.9348632357486956,0.6512806471129222,
 0.7900712378675879,0.22923207534125245,0.1638334429345334,
 0.34309584516091585,0.06133273489007607,0.4345770880396522,
 0.9879289549666607,0.4835039005942884,0.6567508703796674,0.9425004858642435,
 0.32763038098232833,0.6705607440848347,0.31705357878477847,
 0.5999712767114734,0.04597123787927604,0.8837373323427442,
 0.34070958827952114,0.5071222687064012,0.40920759998493184,
 0.3285296446266259,0.05650396713102368,0.5153759721191309,0.9853032581239718,
 0.18501352766425516,0.1027688474884787,0.8155102981959771,0.7227618079177195,
 0.9960583368120489,0.11045033077328958,0.9419177862961982,0.680472856361344,
 0.47030719839051305,0.5918802402585023,0.2645359895380963,0.6013316331829932,
 0.5072435843236262,0.3487692559611827,0.8679507759107562,0.9492393992080562,
 0.42394717832229456,0.9160655828683645,0.7688543637809209,0.267684237880893,
 0.6644085812084273,0.16374011658215415,0.6107134025063884,0.304319541221754,
 0.5953835966238563,0.06953762134105357,0.013490222210383096,
 0.47662382211909204,0.49130986088647133,0.41272901713567867,
 0.018371988039338927,0.057943654732595884,0.79756844125505,
 0.6760927162117869,0.5988928268714889,0.6904903252157759,0.6001210674931947,
 0.5305514630081904,0.3833816094309559,0.10420526518188056,0.2526896649936692,
 0.96826955372963,0.0394437633045438,0.31621775465781876,0.4681908248120844,
 0.1689485993147396,0.5737159569451153,0.6050630021081764,0.673549857780874,
 0.7324220820591386,0.8211080409847172,0.1185279098331794,0.5749712922158444,
 0.4812768109687894,0.6711014243565842,0.08811607000798038,
 0.42555409643437536,0.967108444202514,0.6257305147828895,0.8098801997792892,
 0.9275848172732677,0.21840166181037668,0.5626035352925952,0.7642772410480408,
 0.7019642660265712,0.8919228763026911,0.8892320248916098,0.9870578807266635,
 0.22308956755800025,0.29472841518414206,0.9386221543646999,0.733451230312932,
 0.8004365309311002,0.8972485676462412,0.6634852049631011,0.3953894009821076,
 0.38283720856307457,0.8417623764279294,0.742219585615647,0.08781934204269037,
 0.7532419311269136,0.9538841367058946,0.2605212475959837,
 0.009166670578794167,0.3118811793777335,0.10657481188316786,
 0.11104311485138463,0.4595370665695033,0.35522369422336475,
 0.4845826107309965,0.4218746985763068,0.5923116457051152,0.3770463282582087,
 0.46832525531544944,0.3491358840429978,0.5649573445170962,
 0.16341834741557038,0.15449957957025773,0.17758792896415299,
 0.8904171299947499,0.08733224840288367,0.6194592822688695,0.6318364986793201,
 0.5560845790666461,0.27198676013215206,0.5101908695005194,
 0.008857289740454855,0.2057788332368906,0.8818602076199644,
 0.28405150038349314,0.8066618446146786,0.9400541829850865,
 0.14000365306987606,0.8020876457504135,0.5146792423525367,0.6099866381506318,
 0.008751986291080316,0.806085861660299,0.3253175701532509,0.6803935958688948,
 0.5757495538579814,0.14775954268357994,0.6890783850913369,0.1527565436008459,
 0.13930448848241894,0.583234661190178,0.09628021247040608,0.9612053430819687,
 0.6264938191097915,0.370886237629495,0.2622884855201544,0.9245123594306384,
 0.27576677715940034,0.47377535663454795,0.9741023339845835,
 0.8634508372474697,0.9095981250141203,0.11917100370794342,
 0.13888590107977883,0.1392006352790146,0.9866150611144797,0.5058322912394931,
 0.6064303142603819,0.5473100506474565,0.1695043938306937,0.8110883427287865,
 0.421123472503813,0.9952443793799095,0.06707716176610556,0.07365145874788692,
 0.11162494486841001,0.30976946926129667,0.9778912346630452,
 0.36924252321357864,0.7712785225176635,0.6662167256294267,
 0.046391985657323165,0.6409450697860432,0.8228092989989833,
 0.8587982249682968,0.5038375826384125,0.8882792714424027,0.5229352378702463,
 0.4970073060462239,0.43831588377463426,0.6900663073666468,
 0.32085254446280675,0.609273433945724,0.8714231142057796,0.2639167957755595,
 0.3165057984519053,0.06132099812659475,0.09466311108036396,0.648174598247913,
 0.9702237254100093,0.19121295471532773,0.28738024159849507,
 0.8254662402616963,0.3518472580576567,0.9858028469385621,0.7131010755697847,
 0.1342402958831762,0.5169488274065175,0.19886214506414845,0.7338842489662978,
 0.12988072558554264,0.7664716006872381,0.6602490094871151,
 0.07506116552718567,0.31179779849994305,0.03329144876799217,
 0.43454606532864637,0.8728080620300691,0.8869570920505632,0.7103909613648164,
 0.3740010310264721,0.12748541389319223,0.5877334302409417,0.0652171189800671,
 0.06618774959901341,0.8747589434783628,0.14644567756119986,
 0.5988401792751635,0.1887354927551964,0.21662716582429398,0.9124123340115653,
 0.09555054976008726,0.9492146364899825,0.18402132453838205,
 0.8440978723272041,0.8454812527309985,0.10281663499487093,0.9393231705217479,
 0.5666392795738225,0.7713967926704949,0.456091900364479,
 0.0073131019355631155,0.59614666655834,0.40435244364770184,
 0.9025121574841141,0.7892473562776197,0.5567591032037637,0.3197794506500651,
 0.9128107061460311,0.25555668704259804,0.754152833631654,0.4041100614142028,
 0.9295498736463298,0.15775189248957766,0.14580638627876952,
 0.42839630718718325,0.7811983126912705,0.1398448069966668,
 0.33283763289927826,0.41144236755683594,0.9557163098307346,
 0.8230306838980312,0.008321709331526539,0.5253136444063783,
 0.6624095907145661,0.05512616692969008,0.08987453894064845,0.607634499524568,
 0.47053809829144777,0.6278979548138699,0.5115718562653272,
 0.32436483198605326,0.721766624020828,0.6043584945238281,0.6226851581943229,
 0.71065538616222,0.8187748764961474,0.41817804321304575,0.037679442929540286,
 0.19961953434132007,0.9257462552725467,0.2644435911242172,0.8708220104248986,
 0.1463580126857078,0.6232612556903596,0.5370251548522855,0.9253961212232078,
 0.7239083515031413,0.9274321403516107,0.908981884912455,0.9589548828405822,
 0.23167420106221503,0.37979746496044164,0.6393664194506532,
 0.9773615914723635,0.9412954307017187,0.006038711178274259,
 0.06796918773694183,0.9488942324102757,0.3643386741554173,
 0.42242188061450014,0.3384956596874744,0.7817678763137274,0.0963488494291389,
 0.9000980072912231,0.5675011836523322,0.04312999912742155,0.8287843153068899,
 0.3969905173729483,0.726824992476044,0.7790744844685547,0.5785074516501875,
 0.9752095149390998,0.8925740075289195,0.7103613461545096,0.6064719886428392,
 0.8117063480406645,0.18729837602404098,0.2984764208528706,0.1952270803682149,
 0.8221948083711742,0.5051578162590536,0.4641485889687692,0.8275662780953706,
 0.812596337892945,0.7326713095385234,0.04762538162651875,0.07799373177699975,
 0.8458835743417447,0.4747447607215619,0.7640363555226978,0.28850480063149586,
 0.6935204043780309,0.27311915066989145,0.10600103256807025,
 0.12141268051875143,0.534046221348214,0.8819710458465743,0.4103310344177462,
 0.08783401718843664,0.3841030073663396,0.09301522549258046,0.779129587903174,
 0.16158971197032446,0.7343449172843882,0.23721769360285883,
 0.7330162597281626,0.2351803449714447,0.32991928573825613,0.626483665015377,
 0.5409491500057655,0.8055449234772284,0.7131158370528057,0.05830905509710416,
 0.794953467177533,0.3045910242447336,0.7112340897442677,0.6430311897956466,
 0.7279109161014229,0.12565426198683372,0.7861301388167115,0.5050547453065285,
 0.9304881086047813,0.8339325064248624,0.9717086758483611,0.9683043344524169,
 0.1554388664474713,0.13671296674087752,0.8291265748428264,0.9141999794405208,
 0.9726270360852795,0.9613702730867328,0.8547544545374888,0.2913890529623836,
 0.18571925921739274,0.2923329887692214,0.7542095485297684,0.7535520694075494,
 0.1442179280342485,0.2509203205459729,0.5566226107071807,0.2800815401046518,
 0.32850358035080507,0.6031763822873863,0.18260640514718263,
 0.5831683442883842,0.007410088400868986,0.039629282328972004,
 0.08381759491001284,0.558579653512428,0.36346495952932345,
 0.0026100960325505795,0.5166050632977649,0.7006513828836424,
 0.5100790291720946,0.9066484213521617,0.23566877873887826,
 0.007808368423216194,0.8177046691744472,0.7050987097902145,
 0.9221239860046974,0.006140007692186922,0.5483006600535894,
 0.5502420657473055,0.8085258508402976,0.21217608484331807,0.2536737170338139,
 0.26896794274852476,0.28921006054021714,0.4811538168966545,
 0.7875189138296215,0.15627213811970797,0.9179862218812774,0.5196320195396584,
 0.730988379433123,0.6800257808193819,0.5265959841331419,0.9943341678496201,
 0.18466611939989253,0.17906593656721181,0.3818391711717527,0.516178027666379,
 0.022807223437676716,0.8299520080673354,0.421509965775986,0.5285502793910644,
 0.0484657007367858,0.15976456928494864,0.05704427564877046,
 0.10280968878447494,0.7309202853683363,0.9596584108865256,0.2039633562836296,
 0.13271003915280977,0.827783320017744,0.5485244731287914,0.07664433695956774,
 0.19581387996498867,0.08546920733307639,0.6459115939783735,
 0.08884905492887951,0.9954393820715769,0.2280491757064128,
 0.26533857994795396,0.43475950047220047,0.07325700189303808,
 0.9333272736618193,0.00838798458256207,0.6543067307894007,0.6740710445915215,
 0.18316078009055548,0.7498572205303979,0.7594823808952634,0.3848315797863733,
 0.352762176808596,0.8761370439804923,0.8019037357632617,0.8636082898201745,
 0.09596451562707231,0.08175427090205467,0.6629940482766236,
 0.5996156190714887,0.2528965232974152,0.7646993637938854,0.07991857445727624,
 0.586319203407449,0.30219112473281395,0.00888901340742898,0.9156495268612445,
 0.2825305923369533,0.6211789053283899,0.006982602693999862,
 0.9284883936600875,0.6893129131179865,0.47317418266037004,0.3486740705340807,
 0.11963322328354742,0.6649228142885713,0.14073571151111652,
 0.1761637807498937,0.30669385707636454,0.3239573659865227,
 0.49431451519740033,0.7897642431598317,0.4929955212632763,0.3794035780973296,
 0.3124873420551819,0.9107141964221653,0.8829106427274085,0.912302943536395,
 0.42381417718156644,0.2630894671357811,0.6618150180976949,0.7279998947116004,
 0.9050639731858661,0.6369363664847941,0.6216029258199037,0.01783015151549483,
 0.9239759496851517,0.203770847708713,0.07041684314181984,0.7740018953618748,
 0.21152665189367614,0.9322817725618672,0.8461426483087695,
 0.16011671878990497,0.6908185733486004,0.15727151365928282,
 0.33092393314843527,0.8489048593803782,0.6513407980431751,0.4247312227561272,
 0.3774280893041908,0.4288860309534338,0.40073405405655094,0.698649535133151,
 0.32112990387883855,0.8660904613408629,0.3351268878431849,
 0.22656745643674747,0.3470281734751607,0.14611440539045117,
 0.7553252752480561,0.3836459644601752,0.1270087817374721,0.3189767278445834,
 0.4406925095975175,0.9975776293942353,0.9605709106404221,0.2309705400731803,
 0.2581997645140376,0.6255381336289414,0.30582812069356824,
 0.39385493997811594,0.8825277383590201,0.0070151753992845345,
 0.1593241388899449,0.950491413453533,0.7313000343747613,0.5462338569863349,
 0.02846920005941711,0.42702400177694766,0.8048515073601556,
 0.6939002681508631,0.929718442640655,0.22024444420942524,0.5981794238704163,
 0.9159679709760791,0.5156361245780704,0.07217227641122659,0.0591142578447158,
 0.30955785094596344,0.02880569913330422,0.9599734694657726,
 0.4837025061963153,0.12424413804545253,0.2801715674353543,0.3626758187930097,
 0.4146112044562147,0.2748915370871108,0.7302436636631996,0.37107851540652204,
 0.25103983335376256,0.3380360918942652,0.06984545627702099,
 0.7636853928559146,0.8573078778500114,0.9673107693380516,0.9823774657074008,
 0.060673385872589525,0.12604946536231343,0.7680489312522206,
 0.233080020499291,0.5433184551718023,0.24956216819814758,0.13660531973041667,
 0.6496779556550952,0.04725780992279138,0.004278704679423706,
 0.4473381333873998,0.922923142427099,0.8262951103153866,0.7648151487362044,
 0.18762657724632825,0.18392271989143172,0.6266454046503743,
 0.0027810447395923976,0.002632723635227263,0.6815569906670857,
 0.49600020026371916,0.919328444986628,0.2033645204175304,0.7306889520208116,
 0.2925858202339575,0.23393829203568328,0.5864561420849584,0.7347657304515274,
 0.17556140147215793,0.012920848685490371,0.8884773612847687,
 0.9447252249123741,0.07422991877412255,0.49795210081606067,
 0.15617088414607594,0.17087427927156762,0.09383630736255344,
 0.009788211546419179,0.7424505927785107,0.12494500274108089,
 0.2510181840787663,0.46055907831319665,0.2745622061606292,
 0.49821646794474794,0.28888507845766775,0.24585449704965012,
 0.3746931350877051,0.3212713080498417,0.40615851922517954,0.3719503011028489,
 0.720872666118352,0.32002981197570934,0.31032800787138826,
 0.12513858690021262,0.5495070826825857,0.7648260762977955,
 0.17664350514807792,0.9479917440953262,0.9279686091337757,0.7202899877571197,
 0.7701345147787594,0.786524666241686,0.10073859768192617,0.31372344089695436,
 0.043578753018612915,0.18120656756628895,0.5008605029872636,
 0.3078409607980147,0.6326685638790899,0.6251796843166461,0.3224816239956413,
 0.8008386033680621,0.30172930618961136,0.07785470947129391,0.520103695666917,
 0.6951293294765069,0.7822978075928162,0.37111955835008714,0.9445682264099778,
 0.5541229713135497,0.2228522726993053,0.42576857699378046,
 0.33723073942801285,0.8733670960633871,0.2390804075094103,
 0.41214276076156564,0.9365502349329959,0.755438105839137,0.18821764426414,
 0.9791487469029377,0.8520786471596907,0.25244459489569715,0.1738216442143602,
 0.2148814147418746,0.3593293935904205,0.9615416333957381,0.39581331931529284,
 0.8853987754962533,0.29903839228528284,0.26259551168945106,
 0.22287764345589645,0.27672227816618267,0.04829853273917306,
 0.19805331910553348,0.7218331350617704,0.5413254169323864,0.5706248936142585,
 0.3277005773877135,0.44735750921782436,0.5149384483228779,0.9080765580249173,
 0.7453336239583086,0.3584372624487061,0.08049911144669553,0.7820796728430441,
 0.3459561015497812,0.5979920921205033,0.5843678315301729,0.940387385392,
 0.20248785196417207,0.5922883873152764,0.6085813914312573,
 0.49519189477970427,0.7451480293869576,0.8610477408489332,0.32509023449511,
 0.02230361354809074,0.7531044020592897,0.8637333506395471,0.247911082584237,
 0.3432983044640213,0.9296279897339067,0.587341530629364,0.3812428504804477,
 0.7692609388941044,0.13281651915421322,0.5705770952470646,0.3964426304720037,
 0.900463382430992,0.21931713644524442,0.2626207222715742,0.1724694013492436,
 0.9792685948086921,0.9150414292469757,0.840578287198797,0.1014055287302077,
 0.309701659679964,0.9967444104525978,0.3698663666227646,0.3907077521614506,
 0.32125343686228724,0.4142269768979848,0.44510226724764723,
 0.43943584030953575,0.9906536083410284,0.3958669932943557,0.968770719372292,
 0.6918697815569387,0.787589806992406,0.807456716589257,0.566665683233469,
 0.5055244910023503,0.9651032998739071,0.8116124371130675,0.07909040114507881,
 0.39432334904215305,0.3758882557485328,0.41269412766312397,
 0.21368264911591095,0.058835992224560596,0.0104678203565427,
 0.9799580420366518,0.7496431877230318,0.7517475465692438,0.5246569610659031,
 0.1251939769844176,0.9685950266363156,0.9368940149401936,0.03328845713451334,
 0.10846056088399669,0.8849044804118726,0.12834889253094306,
 0.4684626761787205,0.11913389429498666,0.6661255093599247,0.784284632865752,
 0.5536914116883531,0.8674137599517192,0.83902681711462,0.9641952116855818,
 0.37301759557854774,0.6203065957656426,0.2313546906523224,
 0.13679073563937127,0.17712658622571686,0.8597065403409395,
 0.5695921162822977,0.44121377580009924,0.01650756927160535,0.294787632789111,
 0.6467541530472796,0.24707308240848225,0.6744880677016917,
 0.22244631303620443,0.6784844524118091,0.5904782457947633,0.831623306967185,
 0.9251063417212035,0.582925421548274,0.1274660631359713,0.9933306096157338,
 0.007970514633680814,0.6053862942731489,0.7312274241991474,
 0.2312355479964997,0.8776235969935176,0.2887917958135373,0.37891128040016586,
 0.5466073348069791,0.3022491445948553,0.5673501462197701,0.6678835309670119,
 0.42587360717734823,0.20476044608552402,0.15925189630550263,
 0.5129028287626213,0.8993022147688847,0.09769647176944074,
 0.19216925958172315,0.45688562119936793,0.7413998606032747,
 0.9438626565229161,0.8191545286161321,0.5364972554521196,0.5639081724192753,
 0.41573803970263834,0.7306666956296619,0.3148633377005664,0.9202362006647692,
 0.691862268466889,0.6916038171860084,0.7056236920587045,0.5523402960673386,
 0.8708009713881782,0.0968352747378638,0.22562378805116567,0.6530619066034707,
 0.019837824046446406,0.7302159990279022,0.7646742231843967,0.445349657328987,
 0.49726794116769735,0.4059195681589529,0.9444438086509545,
 0.046649703249760144,0.9226226674788863,0.6792709733517945,
 0.5501448254618295,0.47549178999492936,0.806232705781975,0.3987608597766341,
 0.06303389227388034,0.4501650555458785,0.9267301346026242,0.5033196066308991,
 0.3159304186377334,0.4515493054318436,0.25513961708958,0.7234638620150925,
 0.8938307515731443,0.9290082820051532,0.9051428596829542,0.2655911059618774,
 0.8647577855735995,0.6349493382163909,0.8557907925300209,0.52705355895635,
 0.8453838826787345,0.6343125697638439,0.34207578156726703,0.8365949389755885,
 0.6556033690585344,0.11896191180667937,0.21409868367968177,
 0.5558663819277803,0.6343854586399289,0.5538015249789073,
 0.0043515789186873155,0.19230436093861747,0.7479152432125263,
 0.5352683895402953,0.06064165041613223,0.560080094775319,0.3731702892507901,
 0.9715023116495962,0.9837495635356318,0.5254671016607279,0.8843833062372753,
 0.7775856794686004,0.20799226560997608,0.3289837037324428,
 0.050227710138911474,0.16541965332900255,0.01693926287807246,
 0.5862894469068015,0.7937478762367541,0.9964110959986835,0.5541565398079885,
 0.9993529037634545,0.5773741022116945,0.4414413444410692,0.8940991682282359,
 0.8688750308267348,0.9591204705331347,0.6008659730255151,0.03515925260468511,
 0.6219254964354364,0.7135783607392274,0.6505651294361516,0.3861034989337607,
 0.22524903349766334,0.04580805630715834,0.7948752071209381,
 0.5517071064740885,0.3993810291196417,0.4785197737540061,0.7441697538455755,
 0.667357511346422,0.4354644284076389,0.8875635462031681,0.8994367117281077,
 0.8209650233982757,0.6876500097135751,0.492199309848667,0.17608158246487904,
 0.2383800801545498,0.4478005150011064,0.593325516344041,0.9356383657436599,
 0.28569177708721927,0.6498352371814181,0.5492382047501159,
 0.32593312011068776,0.00377274040276232,0.457884567643546,0.6259225796911946,
 0.27946775247000955,0.8324180236135107,0.5262509037430898,
 0.44966464190389777,0.06437482260830119,0.4236457211461725,
 0.6863460480021502,0.978074701724743,0.46774304137131495,0.9220504558956716,
 0.07166447946785293,0.7730176156742049,0.7841010338718164,0.3170508683228792,
 0.10422267132026386,0.41261431485370803,0.7975108566872323,
 0.7414384196032066,0.6904328933818283,0.6602019387723183,0.2683496053615767,
 0.09627496129350988,0.09628152486560859,0.9016169928790277,0.409436166531285,
 0.8233733431029849,0.021275081075329827,0.6896055113175741,
 0.6110522612476132,0.840490671250601,0.9561247566173038,0.7069326149019456,
 0.42399343678134427,0.023601299682553867,0.47268232665445586,
 0.865416528731013,0.4727692258138424,0.7517808980572398,0.5608720320397373,
 0.4680295547656892,0.876105278612231,0.5475176450795498,0.7243185888132275,
 0.39108824632298467,0.7885238482772327,0.4557549209819194,
 0.02496121961676334,0.10908441003774227,0.3994055018676117,
 0.8318256732961068,0.06773813934907869,0.8866983333326441,0.3657498709956942,
 0.02652946513399046,0.9519595109335104,0.4822424336434441,
 0.34590447274963043,0.007150563837289736,0.014495795895971386,
 0.493518626386643,0.33231165540470486,0.6554260998358072,0.8364963000421249,
 0.3016750593426454,0.19442759842989632,0.5920623627211542,0.8845106708911217,
 0.2038670520912097,0.5621203529984559,0.7776042928094992,0.4960400795385368,
 0.02999918810104374,0.44843341731436603,0.05227313474065809,
 0.9199988462000878,0.8205292373029855,0.12104867347543291,0.5187833070480211,
 0.9712596675475433,0.27684776116392606,0.3889547732276222,0.5075540315565722,
 0.2258882982115421,0.2783557298669199,0.14037007422347525,0.7938676002706913,
 0.7909205838732798,0.0220121797538666,0.6998997586251781,0.792440338323534,
 0.9916851920756448,0.32431642288020046,0.2317116011829865,0.3249484537077866,
 0.900075489364655,0.11370110328440008,0.6929104315015054,0.2947040410570684,
 0.16298333365966422,0.9009963424637828,0.3723877101423875,
 0.22864784744430677,0.9188779265800837,0.6245368287402814,0.8146554461259653,
 0.10931530582123039,0.6525263365719184,0.7788557309528245,0.8526714719003449,
 0.782825291032022,0.6531629462023304,0.9767173301308575,0.13120774015328707,
 0.5622620676301463,0.050280856269981244,0.424961884704094,0.3493339990562061,
 0.9744010269939597,0.016049339873595136,0.7670290067025923,
 0.8755906095562649,0.8195182774197088,0.5071983341381124,0.6108750845762609,
 0.5650774787128099,0.8303178586536623,0.30062630612993313,
 0.35481438136994303,0.7656213564365209,0.10065112650035546,
 0.8326468426590831,0.5431495323156638,0.2435495720592733,0.24565176595882798,
 0.3869915428974626,0.6224978727250028,0.5698246548559012,0.26161283763830956,
 0.6282084902752086,0.8001132550464856,0.10929669971192746,
 0.12273016644661017,0.5020565194950986,0.32916413844842984,
 0.43554398942688977,0.8694791992993574,0.6345542337550372,
 0.0016100475573589401,0.5999961685537086,0.824294418990927,
 0.5991479194920053,0.7671529274230098,0.3066625781228408,0.10643030935992637,
 0.09489107434282673,0.19409232220975847,0.05161084652572101,
 0.2808776757355621,0.9192013112537112,0.0047994345370980795,
 0.1375584818688662,0.8716050122623531,0.1370986336011929,0.093663323047106,
 0.2067840113797389,0.9174058445586204,0.011084657919559704,
 0.16298108759836194,0.43216188675876877,0.16018493357586316,
 0.5831707016215845,0.6175017423918563,0.5147752637197912,0.7766512581448665,
 0.8406288857573925,0.033584151118862926,0.8613886626740997,
 0.36727265546254695,0.8437038227186461,0.5419996978420877,0.7594580831781725,
 0.5069649268240737,0.02024164294353914,0.26413283596946513,
 0.27432868443283664,0.10817267327086033,0.15184666577778683,
 0.4317010726676559,0.8959137568095411,0.4637968362612238,0.36142473687649646,
 0.49617890568339584,0.5157496968041266,0.08351044506930727,
 0.7586979601988354,0.44452117541202707,0.8926515291801567,
 0.25097644111179007,0.9923595937622288,0.5638237121544664,
 0.02498679921180247,0.7852556411530304,0.14405748752600156,
 0.4367899005187552,0.395081733367903,0.40732677091334113,0.38183536805523366,
 0.05387057231492787,0.5186105865598091,0.8753541703302214,0.7766811068921136,
 0.8631722760641563,0.2988117113638571,0.6936417731339027,0.5938271737167425,
 0.1452111952266878,0.6126634280513622,0.3254340802124598,0.44537183701698047,
 0.20334894262412462,0.8492514670884782,0.5587817798268551,0.5460847538194287,
 0.2860195614550488,0.6266139967309912,0.13692839723517325,
 0.011601759443572224,0.3151264389632056,0.6939410291224151,
 0.05362149684443296,0.6455748974757225,0.02083057293372259,
 0.7186336367549052,0.5548734718569347,0.8647396577318294,0.6928229223232496,
 0.2400608615933829,0.563524475900658,0.13980436918753292,0.21290567542120342,
 0.07849627201642595,0.01013152010770857,0.7157655498201637,
 0.0016821657970975679,0.3544809046142048,0.4873702465830466,
 0.6835390153371265,0.9998259197852304,0.9542021364832625,0.7860439360832541,
 0.41783608340928646,0.6632162426150306,0.23175584570100172,
 0.14224865401866105,0.9621435065800233,0.12089862048952005,
 0.20765930602090676,0.021957396451389166,0.5255635183759539,
 0.8369589615243289,0.5737007957912829,0.21393812764782216,0.3094110335985809,
 0.6334233795929782,0.4228918225257249,0.08327677609967643,0.4586840743597723,
 0.8400188285593101,0.24081403152772174,0.7772204872238567,0.698268334740603,
 0.312498303670138,0.7957061889902238,0.9883520723448489,0.11461540201210041,
 0.51156026406355,0.019251281742196147,0.23816487714234635,0.4561730497153599,
 0.99791743855245,0.12007837180863778,0.006480413135694696,0.8182779510107752,
 0.8039819786755211,0.6537572546969842,0.04473736732450484,0.9305698644002859,
 0.432204696840814,0.8786497790949477,0.5835402259610835,0.45722071885235627,
 0.603383951824554,0.10454806606113032,0.5712257662277116,0.5633086384891686,
 0.449729463681816,0.6827579690965624,0.7241550191975683,0.1922206788191031,
 0.20138186023035365,0.08824772577899798,0.13422455116608156,
 0.4589120868849047,0.6558296725098258,0.7533736394656493,0.8929965955070382,
 0.16500670587407695,0.4880187273700929,0.11599924550848073,
 0.5907506872991797,0.9470249077312133,0.1694461439069841,0.8847071994220876,
 0.6781653280735235,0.8325067651187549,0.3752349738589551,0.10370926105667333,
 0.7284592564001373,0.15153186099495586,0.07429324999551479,
 0.5518699825827881,0.5931158428420991,0.6713987034075458,0.7751072599443439,
 0.6503898486071328,0.5284403517090476,0.27759612409990997,0.9005132610011755,
 0.8486912652693986,0.4486656018116788,0.5073310148957659,0.8420007285637461,
 0.5626305171112654,0.4813280709727745,0.8627539501368158,0.9315381561067956,
 0.6746253465592374,0.8915075065055434,0.6873376137384277,0.16363278278640525,
 0.28904690989825044,0.05449684954406364,0.2625351226565056,
 0.30302060611882287,0.9622887605001569,0.6781801577394011,0.6033226516712943,
 0.2779218726158871,0.47509414814525797,0.511703199100803,0.29914068933702065,
 0.9025059513843158,0.9158942507431788,0.8962724309224669,0.4786557486643235,
 0.681250219477127,0.9129718349513478,0.843918519450287,0.751825845606579,
 0.11099173842363963,0.07459908695296935,0.043183922744684744,
 0.13269366268911986,0.5028236388974793,0.3540721362964565,0.8385082514219035,
 0.13508411703350087,0.7919282227829435,0.0021543850138463494,
 0.8944215043603346,0.35244951697706806,0.9303079113761845,0.6403073516481865,
 0.4133682317526557,0.036077344615352946,0.3864007025396706,
 0.6476951177376097,0.19234099830090545,0.27152466622564975,
 0.3173356728966793,0.9215653269201574,0.47763608791633205,
 0.08167140890262459,0.196916784470348,0.5829971197818371,0.17376342212087703,
 0.2707697774018868,0.47422445341498,0.4463511810417735,0.27199812010493496,
 0.3614369883301204,0.9182946561521806,0.06833604809977534,0.2943989942198959,
 0.9354968329536404,0.004131283253146489,0.9487389500595842,
 0.9689549093624167,0.32797921542767106,0.38898336639967646,
 0.7454727773921322,0.26298516870200495,0.597482296847308,0.7947945722208655,
 0.8707463199681337,0.9131670044404288,0.7945007979550066,0.6295452784381237,
 0.7989696546228496,0.4873234198260161,0.15450225503726245,0.7001251169075232,
 0.4621540449448446,0.6016054534183528,0.9078796788161647,0.15625112848532663,
 0.13615931280297977,0.1664165552823208,0.2898258794293942,0.5728616338872687,
 0.6058775525249941,0.852400028762209,0.8231744812683365,0.1341496872168234,
 0.5076367889584007,0.821854087852325,0.952780372349851,0.35957800456841527,
 0.8505434334254729,0.9627347933279193,0.8627654011411393,0.8469971558901871,
 0.6956227191069473,0.9460542631426958,0.6235693134957012,0.5886447518248064,
 0.28879908879363536,0.16933958350645317,0.17718045518757708,
 0.5940618643232286,0.0774426883866568,0.4060041824188262,0.1860540575130396,
 0.9745285308379233,0.7061111804922463,0.663258839302446,0.22061773932048545,
 0.2756839831482071,0.8751553286271129,0.9075325127719251,0.16356766907649234,
 0.8215757833561598,0.18997920151846748,0.0050330766052202325,
 0.809858574081629,0.9764937709694914,0.9010835226705256,0.5139537138769619,
 0.17270914039843654,0.6250277915909074,0.23620833312540912,
 0.6273683011498731,0.5571615214122407,0.7131351883746785,0.9669991679897824,
 0.9287693026333987,0.38432398218217045,0.17005131146658348,
 0.2243777533760326,0.6699037396225025,0.15884418164943115,
 0.42314810989703555,0.6643115683782204,0.8543653285164946,0.7774161164193846,
 0.37669528691769794,0.8741907915449326,0.08320489870784253,
 0.5790977360916736,0.6764279731635997,0.09595245146034781,0.9023408168557526,
 0.5462114095473954,0.8032371394752329,0.8007822001012143,0.5703381504705302,
 0.3642288504839979,0.7974759414809608,0.7563284226076535,0.04701186083872877,
 0.4231694081049395,0.9367935920609078,0.7434281936665693,0.828602299336972,
 0.12904313919813726,0.6605320185849052,0.13605227000164866,
 0.7350125378669342,0.7582139591501149,0.7387810877590102,0.3605836946789648,
 0.4899717483009767,0.09259821857292616,0.07435700980502047,
 0.6012794495801788,0.8151203865937752,0.6878856264912554,0.7866025686025273,
 0.3084536461768942,0.5938797201240074,0.18053439265132853,
 0.22500153931152167,0.2736863050381838,0.09604236552512213,
 0.6863095277596951,0.10018837889283916,0.5318440130791683,0.926836286934172,
 0.7152525461634118,0.7277724326157196,0.7250722112122483,0.11495425271976789,
 0.7404120193033454,0.09956505682870886,0.7789553348713592,0.1517171369458522,
 0.20668327904536898,0.5262656131998331,0.21126144554285864,0.529430522118962,
 0.892902158088926,0.2476028363444005,0.6288170701809314,0.39263137729322994,
 0.533970375464796,0.2987422689051853,0.47820488830559815,0.8128044167654253,
 0.6116893052431022,0.4496152175986554,0.6098338850139848,0.5654785788493064,
 0.25781612420008015,0.9923382392534997,0.24493365888760077,
 0.9614616639655462,0.8774417781129629,0.3718098237535179,0.7282500030188077,
 0.09914222584911836,0.05683220387386423,0.008550041972507527,
 0.6349169634943393,0.028514941391573068,0.062079252770861126,
 0.4239680131643645,0.8982150420770381,0.4457191481271421,
 0.009805818548530243,0.9649385391278592,0.5077145035340902,
 0.7988004180645943,0.56316618216672,0.42038497118706597,0.9054613510400173,
 0.12849398912651644,0.08342795699378103,0.2798843681514589,
 0.8980068604593942,0.6017255881172872,0.800008366769104,0.35958951727964383,
 0.023540259590554502,0.39759888260225706,0.6292333880297902,
 0.8187035867321291,0.5874834816774228,0.3477611476759135,0.8234135959433913,
 0.30668590410402863,0.5210857456329041,0.8396539855211469,0.8533456562857835,
 0.6145419652712351,0.9217767457790361,0.035565559920279655,
 0.5101367940132361,0.55191218690781,0.9319383791387936,0.05021605654457395,
 0.022064677214137962,0.3358401826787899,0.9256740455746305,
 0.010393458328650906,0.2002741864281311,0.2946130752500721,
 0.7840891292093503,0.6344645222910914,0.23336044354051289,0.4401066802852017,
 0.7298223722372283,0.6065101663911941,0.05740396401028114,0.8681256084277936,
 0.2704315828602921,0.6352012316929366,0.6284479006427776,0.02214799900013309,
 0.059243211355481606,0.7693288547571817,0.23701750321407766,
 0.6471974054147428,0.5195312179508349,0.6802031674342802,0.3640786663905974,
 0.758202095853958,0.4349857268809715,0.5146737760384635,0.2123497942760718,
 0.07011864278469937,0.4734752299679933,0.2846818960602212,0.3735198964343759,
 0.04629530655959324,0.36461600923703374,0.14226080738550217,
 0.16931490059942766,0.8241934568920487,0.5920400481912187,0.5806023221130523,
 0.5554001778599493,0.630467320038356,0.7014688660576895,0.5358297074875796,
 0.10996280823572424,0.6834194521194411,0.04919724642315404,
 0.8216519054332994,0.08444862442131695,0.8536119965295411,0.2710526620138004,
 0.4956614870689121,0.9316313147760749,0.14390212136162295,0.6761238159320545,
 0.22193797679085014,0.19082072252737403,0.13485753799126843,
 0.3326940162164068,0.7089867284852691,0.8905898056506987,0.6990628145748523,
 0.12873571484018376,0.14010946981906314,0.09957224003454923,
 0.6158124816786714,0.13471650015258818,0.5455056736181052,
 0.13189443831179648,0.1489318074065099,0.28304439555091643,
 0.6897989404108699,0.35403209627951404,0.2245371772806295,
 0.06477324962965725,0.707167192200917,0.7628145473634358,0.5416097160743948,
 0.6076844581554671,0.7156440458613633,0.9277893481902258,0.5094177099533524,
 0.47239136402403525,0.12279482564652688,0.5576672361030517,
 0.42526705178485513,0.2298604877382363,0.2582667529757612,0.9706426723263727,
 0.6114033003510314,0.8709166842397525,0.9577789444662791,0.7901647013449106,
 0.0228166337534208,0.3741475458665998,0.7898226448116032,0.40401804674228115,
 0.21635185003394974,0.5018587019201122,0.5012034140376707,0.3726522188394997,
 0.27307202669635056,0.4017339544617605,0.3954504064812262,
 0.04431876712404159,0.15336987531095547,0.404929806250389,0.4018663398905049,
 0.47782503894048345,0.5143491839907801,0.864882315594067,0.3199383522952237,
 0.5933859455948619,0.6457190891196041,0.4576833869800442,0.06112938777061905,
 0.6092852664916093,0.44052503005349775,0.3597164835040664,0.9202974167976584,
 0.4081321548514616,0.6499583221090742,0.918891112462527,0.7573887837923349,
 0.8989318680103461,0.4424775814110895,0.7202723820120954,0.9470879486988065,
 0.5375014809301829,0.04678220800903865,0.5615794450760937,0.5597111284436087,
 0.12114153524589744,0.1594131333407378,0.20561265683745455,
 0.8195334358179427,0.6807030145048456,0.14629014521369832,0.8812744491407041,
 0.8063245044582702,0.48840201948436845,0.26996196914972825,
 0.5519140936044982,0.8859976014249543,0.3468604182154973,0.07050607122536812,
 0.1823641368152189,0.28016173354139073,0.205273383975918,0.23784088698536798,
 0.15207031691122552,0.7866380476799024,0.4565361215768249,0.7462965197044719,
 0.9247228562682636,0.8623201232636039,0.7769126406038932,0.48726329773318466,
 0.29554792593186463,0.2848675216714579,0.08151670778772335,
 0.7799259063347469,0.957882690792782,0.7971659664492988,0.035890283652566635,
 0.2718831911636165,0.48666844667678955,0.7361019943447079,
 0.022426021705460153,0.1776959564101983,0.36353926625622657,
 0.8657140771416969,0.49996551684104606,0.7375138260408032,0.8188487019011953,
 0.36894691842136074,0.22434668467530916,0.8241208534504902,
 0.7591195519421008,0.6943615284170118,0.3090600823769566,0.39894663441321077,
 0.4048223676520305,0.3763869089317978,0.43518930476389894,0.2227050071613823,
 0.752800419193499,0.406368095159904,0.6728661167221013,0.5315524523802871,
 0.5159402853384171,0.8866136920254347,0.47526079411589406,0.9182957234620621,
 0.4111781853642056,0.3666416935780371,0.8472970680608073,0.31573303352224413,
 0.15985872081123853,0.9489997818275953,0.1801678078318606,
 0.18333877940999055,0.4607110620937158,0.21372284275885445,0.960057983947833,
 0.9553665296148621,0.41767389037509683,0.6457489309714444,
 0.06526290811133073,0.12306721923109543,0.024155642170774705,
 0.15025072406112505,0.32713039165620805,0.3086967237730721,
 0.8917251577193458,0.8235099861346621,0.5776428935358876,0.6234841220726162,
 0.7352993221560372,0.14784801034028128,0.42098350992186073,
 0.2688165841286103,0.7667151866307368,0.6037369272159462,0.3027576647862966,
 0.07282288136828485,0.22655188477963017,0.8109757445224299,
 0.5068269823982294,0.7425920259296713,0.5157401017976161,0.8864415696363509,
 0.9703333814637568,0.9636614476146509,0.797293954930372,0.6470908144808971,
 0.052256560431128474,0.4998233294290372,0.6873892626996048,0.575818123694151,
 0.3561240073249239,0.31122631697113023,0.4285457504260829,0.9358005341082373,
 0.21136406803616192,0.0028120364995326552,0.4855079025204859,
 0.3017209876965061,0.4471294236651179,0.6069017420377943,0.066349284839613,
 0.3087446726698906,0.5799279720188466,0.8689756561354145,0.9500046130753462,
 0.6541739921890497,0.19631609350147494,0.9790803243119243,
 0.13567390803775403,0.8033999898502648,0.6174547525980912,0.9222554305453757,
 0.3982518381058592,0.26176674013538725,0.43286237695393126,0.96741338022908,
 0.9729385461445504,0.6295073523987901,0.5521891832071913,0.4047319083151184,
 0.11312440435876636,0.14691407302415083,0.7693040062007992,
 0.0060623717292873636,0.18398345784181191,0.6983774716006037,
 0.04677085607359144,0.8277378646264106,0.7341745686319584,0.9460916164623504,
 0.3836929041497623,0.6391842612007552,0.20716410997563706,
 0.06020648046609889,0.5075195594005274,0.29888873401640437,
 0.6691870919171695,0.6666136831761855,0.5754872151241408,0.2811088565098038,
 0.812408610518634,1.5275387314717115e-5,0.48818691772604383,
 0.14630359858852449,0.682075698463646,0.37199339707015333,
 0.0038942693093099168,0.8317038797472456,0.9217493667318254,
 0.0045610597735240566,0.3843347092065379,0.7621016006712501,
 0.7430888360043162,0.7621200395341347,0.3070319432313797,0.7754880919316565,
 0.2739000835391522,0.36124049682187176,0.34641789217228036,
 0.4012067434404578,0.6289799498192784,0.1344381734993113,0.726555626501271,
 0.928156073242655,0.8430753136009425,0.6570139296058858,0.9880837124100573,
 0.7084112217612059,0.37201003146455436,0.13322736871414298,
 0.8060666663314082,0.6977068922635097,0.4442879023853753,0.946792234382571,
 0.8645494535611927,0.08607402411497433,0.17101953624911947,
 0.2531996261129714,0.8996543926235532,0.948127184638182,0.3686759477045749,
 0.7428591394745865,0.8863506365077439,0.8227735138659124,0.5631030082691572,
 0.1322253302565055,0.47798200613888875,0.43291435269354084,
 0.6434000644920874,0.990621640066668,0.37570289318249983,0.5076936404170604,
 0.7748012469625798,0.5738709445332235,0.5305748812540122,0.6775599140011808,
 0.47226213789069604,0.864981359511336,0.46652619246410987,0.4075194470737954,
 0.39977258861862786,0.40139777505871077,0.015207292041928722,
 0.5280075011461729,0.5276819421359283,0.5530130078009289,0.63144268565625,
 0.7337162370082341,0.34525171314699565,0.940152761903861,0.0989048984556673,
 0.4265478797321228,0.6459725101111661,0.7248402260981704,0.8038175195010689,
 0.9408472507876298,0.18798468983968564,0.5858055095390108,0.7202281671012761,
 0.2901852272364208,0.24559310614943608,0.8322506802629774,
 0.20113938297051492,0.752216204266361,0.689103158351762,0.872316971453838,
 0.567562597212448,0.9799401323488434,0.5198620707970288,0.8459114782035339,
 0.760377945145058,0.663786360965627,0.15131522819712262,0.8053180746638444,
 0.39909365950434506,0.8170977504896347,0.22734590750163974,
 0.7015449395426583,0.6746173136869934,0.07418343049890208,0.9831261451970805,
 0.3844603816944677,0.3588864360905277,0.8577010674083159,0.9925283245404257,
 0.06222190557304208,0.16413367820753777,0.9838393539573057,
 0.09829283254270293,0.2729498238565462,0.9768331199569696,0.1931470510201977,
 0.11722030267218309,0.08229153042364679,0.013840931987359473,
 0.8132186046100601,0.7491362332922468,0.3902971877070748,0.2886017299160457,
 0.7772958003301458,0.45976387135406327,0.25437851188371285,
 0.8638691924526335,0.685572989526784,0.27480061101766984,0.2629019727136064,
 0.4535719711896914,0.8697937387001978,0.3571154811917854,0.401474717106741,
 0.7460376279531044,0.3350412446574853,0.820830481528889,0.8967809747005987,
 0.5840830547727425,0.9375015862877606,0.26322101157970457,0.6137602625579839,
 0.6709293261700557,0.5648116137478039,0.8159893899189038,0.8221001177577463,
 0.3480770108646716,0.04284688619390886,0.8748406733311136,0.6892100649991532,
 0.8570599493088423,0.046355315225567795,0.1901745828698328,
 0.3128088183706773,0.37871853804901534,0.3315476305485754,0.7820840476575236,
 0.18533522237666944,0.493818322514596,0.3581591637294901,0.3735399409626273,
 0.3971209718414016,0.7813338503084046,0.9619644003943844,0.4642790274963231,
 0.35952693595737384,0.4546485511413453,0.7007866320584168,
 0.03887953030526847,0.30470374088945884,0.761221846369848,
 0.09619625802530518,0.17565971065526098,0.23291299386080033,
 0.3582272485312683,0.45100541651156023,0.16148573332427452,
 0.9413205295107464,0.34594690677820905,0.45114109091003207,0.937401184869614,
 0.8492312651472064,0.8961982278788854,0.1528598672053585,0.46452997869947155,
 0.10696435844723184,0.16772810062455823,0.27597760351338574,
 0.25613574100454317,0.10161203346035919,0.9017990864341026,
 0.9715111991918152,0.8254075272971043,0.09716845525476248,0.5814251636063561,
 0.30743053165861833,0.011260428517618437,0.8888315531730361,
 0.03596639116509237,0.5456221304812447,0.8286334498121602,0.2675228736859734,
 0.5241128847284922,0.23155055458081142,0.49678290717587137,
 0.8078319171151647,0.9233384159114636,0.6752241591516236,0.696580624699352,
 0.6792277752594967,0.9771114452510987,0.9362955597438078,0.20496134182701398,
 0.23510088391748496,0.6602170564558671,0.3165977136834719,
 0.04077922651090293,0.43234206112651263,0.6536541401781536,
 0.9160665676041154,0.28560865788204637,0.8321428094820649,0.9754823729925501,
 0.29253741160035696,0.6491850876081862,0.19789041154858167,
 0.6670794766250517,0.6606436952562722,0.37740468405134675,0.9513899721836753,
 0.5878165768629606,0.8054784762913372,0.7187940646156017,0.6561783548198614,
 0.6164337121531153,0.3781446058480661,0.27175423267250065,0.7906520778406226,
 0.5925445040215782,0.7465906954574397,0.10113867225426543,0.8898475245928057,
 0.9973828938865262,0.2190387816360646,0.37785074555428033,0.8103409470667933,
 0.6963945725388673,0.2747553364340737,0.8962332262689706,0.9873884321850445,
 0.5142763728758681,0.7118593475640519,0.1771281134124406,0.7984451878694808,
 0.35205476235596456,0.8497282401434574,0.8999332314047573,
 0.43269328454299905,0.7909254643500208,0.11615541339265656,
 0.5851195955686724,0.7877997332502311,0.7022685760325497,0.19381441591465032,
 0.3591692063518883,0.9962736679996532,0.1471457461377983,0.5798601742467249,
 0.688720472147695,0.3336293730646491,0.6506967579955721,
 0.0012543748486697481,0.8372377845579171,0.09465726030763122,
 0.16837441174154222,0.37886880403666745,0.4025814111127156,
 0.9152833708043145,0.9598791139733689,0.19831033987501523,0.9575325010666258,
 0.013834698026098824,0.47787889494480496,0.3568691385839665,
 0.7369189876283483,0.5667856110242779,0.0970859152186535,0.03474136590910615,
 0.4897940387636558,0.9137654033333867,0.8594763560872808,0.6405707050141795,
 0.720725626955638,0.8755891262509943,0.8844093801544913,0.7002485878270581,
 0.9014515267874533,0.5823748121019575,0.8423503379708227,0.9172238385896512,
 0.8328194719276097,0.30317020551985707,0.2329533008331348,0.8659364682141406,
 0.23833970425504303,0.9676428798598007,0.9789516004043102,0.8208594764580245,
 0.9402362970841954,0.2850796216611845,0.9128553043548644,0.05487980948712068,
 0.679221153391999,0.5001206713555222,0.8437411216865446,0.028397747764806636,
 0.9177779949207864,0.6909849180807687,0.8366566105604526,0.6711684518266038,
 0.3718004345394491,0.44849218563006266,0.7574160020020583,0.3659615976951691,
 0.5234854047401832,0.10081035427464247,0.27837972540784983,
 0.07236108209290171,0.5921983433502374,0.7370869267872884,0.7864524283842251,
 0.3639221752581472,0.6147358522182012,0.4011121375912621,0.26722179016936,
 0.9640201246658288,0.04776624374015215,0.8080054708652322,
 0.18684876096687464,0.8138744541253473,0.9317328167057514,0.6681667591622857,
 0.25776233589801056,0.7896369476269652,0.9643321347283138,
 0.03016144775020768,0.22131883422319698,0.26633515448766243,
 0.6513465293399513,0.6963053527956174,0.1899513918234106,0.7775457639601475,
 0.33873464027569977,0.6242973620759169,0.187431729689008,0.7281343248530081,
 0.43924565172394603,0.7294608605366664,0.5884573779159085,0.1475003403039996,
 0.4746655985058903,0.05598482432797203,0.9965186254676128,0.9011591372970704,
 0.7911397216615276,0.5917733202709117,0.2058979578216783,0.26326648518152074,
 0.4795533918841308,0.22293538209562203,0.17600585587198958,
 0.6865758883854396,0.9941642533397617,0.6337365953690258,0.7048583454742997,
 0.40207816591714496,0.28744289937750556,0.7151327875332341,
 0.19723615275098472,0.004692393374339576,0.7010494221589626,
 0.41994063657255176,0.664201538254017,0.21712976266979567,
 0.42056116235286856,0.9422929790307641,0.04696719506783098,
 0.6774727355498905,0.29646007965005594,0.0769822770556432,0.6147586626294119,
 0.6960950940385274,0.23941919823544255,0.4920473220245012,0.5883336838995368,
 0.6605773063596039,0.6852775660806891,0.7793051633222677,0.46735454284143485,
 0.3762001082034643,0.5855996742849364,0.9302871251252502,0.10684651607489315,
 0.5500677061305453,0.7479100837012392,0.8550903410392322,0.46290052303271134,
 0.2061088085776258,0.8853744160400896,0.9808086640073519,0.5420084248873452,
 0.23662615730631542,0.5897022008359464,0.9866739488617086,
 0.09846656193913794,0.5147878527507336,0.5934851061040016,0.5072549073075073,
 0.33620329825173434,0.3557102010443668,0.4811513752185752,0.558082054421398,
 0.03573201602645548,0.6207870119652482,0.8354297870572157,0.9491369633196215,
 0.6893186995642522,0.25191130526729943,0.7247511615797464,
 0.15569692461750473,0.27554742624063033,0.9782421904137402,
 0.9582143283013259,0.8273512072629409,0.2651199819603757,0.30237088040114324,
 0.17727543535209234,0.44074527197457036,0.09475151039208785,
 0.9013856786119152,0.9417235535391696,0.9818787817568009,0.38388583985614466,
 0.3509429929626777,0.26964058701833826,0.1891205731207869,0.804003918706948,
 0.7265266073429687,0.5988253352381923,0.5078827041460939,0.2762794461503906,
 0.5340349965034831,0.9850697065848587,0.7484291951623786,0.3443532083923344,
 0.04948447963612801,0.1585926746134747,0.5753052989685269,
 0.38394242058765815,0.5098714541446429,0.2962460443166218,0.9773949496657426,
 0.6632739675087889,0.08412728839440942,0.24873107961396557,
 0.6024591395950681,0.9364207103194804,0.9559494538207479,0.3916497155502221,
 0.7397600009613434,0.47574254958545337,0.44702733649904935,
 0.04930018425567917,0.974524113482038,0.6937405585294963,0.4903919473163326,
 0.504380156927621,0.03625383275722449,0.006407095805409657,
 0.3103992661385112,0.4793788635964161,0.6595694226963421,0.3362412223186022,
 0.32333014292831774,0.9243354129560238,0.48656764360185056,
 0.25532505152979657,0.4571922943350625,0.8982900803992692,0.5828700514836702,
 0.9717123657507267,0.4605097111142046,0.8229623414694367,0.9872417513745053,
 0.6719108151587537,0.04636558690623327,0.2547558318388109,0.7090454411754513,
 0.31463516109097145,0.447591525270722,0.8442380135801226,0.6718092176677759,
 0.4418194654475921,0.9945634243800973,0.028101920093987265,
 0.37600185175890377,0.35733044241871836,0.6163105735251536,
 0.8233949451332252,0.16892802956182196,0.16615902769838964,
 0.28985163576977857,0.5992275386157823,0.6601785285419652,0.7879435061531894,
 0.5985346503329291,0.16765748371559652,0.4212068267337876,0.5689312891587808,
 0.5185174208528177,0.27399336955295106,0.16380490423547722,
 0.9600844624508695,0.2665535229865642,0.027874620836771613,
 0.7198811154005663,0.9486469506129614,0.7797032386674307,0.2438975509188035,
 0.6314863091722289,0.9911746966158759,0.8130159850514668,0.4938268973401899,
 0.1585064656230759,0.09714390260580341,0.1286963499385395,0.6030545558201559,
 0.1417167814631215,0.3219699845583762,0.1525824219223011,0.8807103898351338,
 0.2548876866663147,0.07626465109353564,0.29901860776861255,0.550814541042195,
 0.8585009563893772,0.7010831357678122,0.5309127753163934,0.627797434654853,
 0.6492999000675264,0.31091461277754706,0.6466362612687828,0.7002488278901761,
 0.5077653224295525,0.3143754653532401,0.24883518870329002,
 0.05611914546458996,0.06543165489179792,0.6254751731205923,
 0.9785902266899427,0.8975717555299818,0.12095960270712347,0.9098241879909896,
 0.9460036854655929,0.11392967341253213,0.8783938596119241,0.6015327240208764,
 0.5039623325122919,0.9006898748932392,0.213055797683825,0.5652227665964169,
 0.33736625234475737,0.05396155653395063,0.5933919917353565,0.970503735347695,
 0.9116644757114618,0.4036556725268069,0.8845813814765948,0.5472045579567528,
 0.22445979226844281,0.5906798574512222,0.6701746945047269,0.5899468782514725,
 0.06849556304821114,0.6175639393326188,0.8430389848585171,
 2.742192688690581e-4,0.12418702682626459,0.5483441485176779,
 0.7073483503845052,0.8343326777917096,0.13403557948653233,0.473802584416669,
 0.4070778718365017,0.6467137331602048,0.70641128447063,0.8853587530462432,
 0.06256363378824792,0.9119672351285198,0.4197835964754444,0.215313324295076,
 0.739936215972178,0.6508758277955927,0.010342584154160783,0.856756986300996,
 0.8098467416409891,0.4337833114690778,0.5098400571292989,0.5592711791075923,
 0.3783772830490453,0.2371614465437939,0.9569169785923521,0.4523720958393843,
 0.9089279996488738,0.8731084369737142,0.004671386091485985,
 0.42165466024243414,0.1460126630706754,0.39718988309334236,
 0.6888808504991533,0.07786396860561706,0.8955186912953582,0.5727562632688395,
 0.36008255741550377,0.8662998737434788,0.5147916249167491,0.902316842878787,
 0.6058240438460776,0.5700547576759477,0.9153273176826315,0.09010892024986172,
 0.6941113086250817,0.42768533056613434,0.08761876395633694,
 0.8735991146064164,0.20942145215002217,0.23168468914957197,
 0.4987216245474425,0.5363058122844855,0.10583961657485341,0.8048800745313759,
 0.25073393721231263,0.22873740414457977,0.6931357153208796,0.456085448589042,
 0.36825635413757896,0.8345566488137981,0.6056255350505936,
 0.12565271332927885,0.3083474933169075,0.7154862883102441,
 0.15513196368129778,0.018465316461091286,0.11374080640372375,
 0.6483098375598577,0.24225509021494063,0.8382708998562246,0.7881917813668099,
 0.8269480907978328,0.03619738577401177,0.2804648902001988,0.992931515588039,
 0.46948628130550996,0.394178286879576,0.7488544776254812,0.9022055587499335,
 0.6754230683101246,0.8424601379720024,3.9990168990833297e-4,
 0.016104081310213125,0.015119539258335424,0.6789333066398027,
 0.1517048866057964,0.43250736481059004,0.3820022494650427,0.717335184815818,
 0.7932198213862905,0.2729071625999414,0.9496837902640768,0.4439630364754149,
 0.4363925006865311,0.636208679396595,0.6086156566146133,0.3995737008647181,
 0.48640185754827614,0.058461572516341875,0.07126834118990799,
 0.8405925483818378,0.1938469331375724,0.7229189253632456,0.9389149335169072,
 0.774787078049332,0.44663237667237,0.9452319705476679,0.10309574679851441,
 0.24125401613303488,0.33214895973991465,0.9595844779354519,
 0.7057311449439974,0.6577072408019868,0.22065215662538895,
 0.09857816792483443,0.422478037086506,0.9689349788295432,0.5888145425887727,
 0.7207162649355472,0.8470377446707467,0.7649516906794994,0.653320526034754,
 0.7727169539086257,0.2979580294212484,0.8609527658058643,0.8267863632785281,
 0.11095441880257995,0.0913063035456868,0.25123526850694944,
 0.9836440701532272,0.3120605936554678,0.8009365535832025,0.9332754980251479,
 0.16731544086092875,0.39834760381734,0.2632811065048999,0.7302324151911589,
 0.26522558056005685,0.21530334451598332,0.3278494474767256,
 0.06427289403567449,0.6939860760032083,0.45661617280951006,
 0.25871588528486233,0.03935104700345504,0.12250690780528861,
 0.5894137529189598,0.20644658670147498,0.04534004193395569,
 0.9454902836632022,0.7955502100109048,0.1610845706409847,0.5500266204146711,
 0.4969760156602596,0.7572489154737324,0.14265127308554937,0.4731256331726039,
 0.5266003123763888,0.24666831140836776,0.07166522510297269,
 0.9271488787086981,0.09142603395082816,0.13931822373671776,
 0.4988186951428316,0.7268791439457736,0.30477454705689266,0.5685674515436568,
 0.48514014336449773,0.3658153495594074,0.5051317275217879,0.4129298440094715,
 0.5170369540301664,0.3409899897305019,0.44342913973479015,
 0.26248565412275404,0.4942560884206465,0.4207133047772137,0.379755468867525,
 0.8633121085821311,0.12047892053571518,0.651643027217213,0.896710076675075,
 0.8898856464716574,0.2550228702084486,0.7786621718186862,0.6892002266231647,
 0.8178157309513405,0.56376740896864,0.23670425745181212,0.31803659395744366,
 0.17266303983210118,0.649512104641373,0.6403254256847495,0.09495838622062436,
 0.030788437252795453,0.7704703173511973,0.6069457094762227,0.261162276750157,
 0.2536635297166655,0.05878950675878181,0.7122296758166834,0.812133581491783,
 0.9523703611025949,0.625856697491947,0.9244582217119569,0.4213078318492405,
 0.06149353788334133,0.44125781203640124,0.21873984361393528,
 0.24658215513548987,0.05734485752867835,0.6511281769759818,
 0.08739998717945918,0.19960747453479843,0.9885574230502805,
 0.23875744815957367,0.7675659442058904,0.37887068605511465,
 0.08468328430948913,0.5205475298670434,0.11062509488951644,
 0.9722500342427994,0.5629854227997724,0.29612486934098414,0.5718275948704642,
 0.48187406007666267,0.08415369303613063,0.02818149071780318,
 0.5183876780031177,0.11664184114147813,0.8489598441371695,0.8706247553112435,
 0.12224860276347704,0.6128245155832406,0.11368975923940972,0.994425674412423,
 0.06286972286361359,0.5607610291906876,0.27470142688539223,
 0.5535759151347319,0.03375325932728279,0.11884141822162175,
 0.9859442550411568,0.07827791132618056,0.7127881804393638,0.9919632326986771,
 0.6633039128228723,0.8320913968361106,0.9298565144809305,0.5213377223601002,
 0.9033366253171536,0.8133933782057142,0.10825648940141475,0.4361106310199969,
 0.35793842040068446,0.15451532829435788,0.2572684460264444,
 0.3040316338185537,0.794925424595385,0.8630481470768779,0.46933190628985144,
 0.29639681187498046,0.1029253490087797,0.8628681178025778,
 0.42122568745598354,0.23486421629423515,0.12621150830244265,
 0.07728504113648604,0.10063823116639825,0.4401499684473005,
 0.3233049791246514,0.6106105912531703,0.5685649233275394,
 0.020535293143317923,0.7820728441356579,0.08616631572706157,
 0.3402694268185532,0.320791006229989,0.4568854878201426,0.3604130208478644,
 0.8954787132705476,0.42681821485079197,0.9336056949540223,0.9698235492164788,
 0.2945136632101292,0.16447250991731877,0.9742450780264249,0.355373869796489,
 0.14713202597055108,0.8428330427816723,0.7143087331703879,
 0.003008168051193527,0.2531311022313647,0.3365961255961937,0.387386131574306,
 0.8879535823875734,0.5369527213644959,0.05113981197326245,0.4980518525151624,
 0.3778789916470877,0.5526078257641502,0.986439577058864,0.23152316407291584,
 0.13815743714689044,0.5225541084759682,0.9494678641723314,0.6192018174766554,
 0.33889673875576065,0.4463504866171796,0.31911536316260647,
 0.32726636128870634,0.4285474059412874,0.7301838470748898,0.874090113084091,
 0.573673176267617,0.5555588907131287,0.5648064417619368,0.7845903139413977,
 0.7321980243886118,0.5761691248127607,0.17573179124682003,0.8059230751917973,
 0.8068983840383173,0.05739499881997234,0.5791666723572477,0.6827541714845741,
 0.825811242049663,0.42108031186178907,0.1370063129038862,0.9518088419949478,
 0.5483682085055988,0.5337026224259743,0.5764718492549251,0.5007784713626993,
 0.7028792304968335,0.38885266133308805,0.4352742335242836,0.9717609089864562,
 0.871500902012389,0.7092259268482201,0.9640463780581421,0.46811515960630024,
 0.20049662177406913,0.16350671821989016,0.7564304659601727,
 0.44092340031223565,0.9220343655498815,0.6910613673692133,0.1748436105790424,
 0.8612774221219122,0.07633994191648452,0.07496665735349572,
 0.6010365392727216,0.861546859586563,0.9667451449148617,0.01795780717613804,
 0.8522303639870632,0.43713644390499096,0.5538495858279857,
 0.13174116063170627,0.398427087953378,0.1035665458368541,0.6989314163214343,
 0.5004644682638306,0.5028237411545122,0.24696907833157467,
 0.13989428137965254,0.6065063084288662,0.006605433765812019,
 0.8766182038768275,0.30319147011870806,0.2480424354448051,
 0.023173549725693854,0.08941546671860012,0.6432072840045497,
 0.011514078432173114,0.8596956773877832,0.550047159953408,
 0.04126446062297373,0.9314162166492224,0.865034142771067,0.872064434399122,
 0.185427808818446,0.5762637083907579,0.3549861623739802,0.009812375075511914,
 0.050225406075344425,0.02772480074682182,0.6276699620094933,
 0.8295851612411587,0.40710145892038374,0.10922633315385988,
 0.7643425979055498,0.29073700239428935,0.3943030627761712,
 0.010044793574878197,0.8073658082625785,0.27192348563966673,
 0.9330095264766987,0.810837109080601,0.8424069868139855,0.09982564875389255,
 0.8314626461514132,0.12745012215804807,0.44187994592810986,
 0.7729408899195847,0.23226130245666654,0.033193130346729105,
 0.46050935442491425,0.41693856202779755,0.40634304594217097,
 0.9385410146312786,0.5018789431118104,0.8276628899355319,0.8463505926763704,
 0.4699295704663511,0.33341167057062926,0.49484643138708106,
 0.9381575402461436,0.3652082941954994,0.9582518534790055,0.9581490026779645,
 0.7948733616645934,0.08637916782435173,0.48522051903061936,0.700746287019434,
 0.043908472649135444,0.10287266406661,0.058187068870664604,
 0.8130134903787699,0.47271375673827876,0.3698327306570819,
 0.02773750547676712,0.7237152719725576,0.2981063002361437,0.7487190877444363,
 0.4197405296121711,0.38760502328039514,0.9451912142170984,0.8985939114606787,
 0.22268025322574614,0.3010157128843427,0.08950686274800435,
 0.16622763993114675,0.03340451970210567,0.9979871371186161,0.986244772833461,
 0.8279623681523754,0.24750901357427402,0.2292113557000529,
 0.41179136876907996,0.5059330612471583,0.813001625723922,0.9601869904995246,
 0.28795866184789487,0.8705543870088857,0.15891981669810606,
 0.07529265252577555,0.021382502680617255,0.6266077891084052,
 0.7049692766051429,0.06525976195435634,0.5023367653731476,0.557892558056684,
 0.7105056051317264,0.012573898367840508,0.7992295003983609,
 0.3535585466663832,0.5207130267914614,0.3355275363740404,0.7204629608201308,
 0.7862941410929568,0.05911405520788804,0.38151737816381637,
 0.9332383449451577,0.2516667079987045,0.23154456636472664,0.493620609600411,
 0.43115683149058737,0.35603929089096864,0.7818172877776416,
 0.18952402385842682,0.2705665299917711,0.30924642786337675,
 0.9607454252190312,0.11394807355973735,0.8595797155690158,
 0.012974515821640442,0.6736668582012153,0.24808654484405612,
 0.2357813109084973,0.6042352651932428,0.6615028275205488,0.02973025209606628,
 0.18518493586115237,0.05055264878900034,0.3495301003013065,
 0.8853350041707545,0.9876106140338592,0.20582126333110196,0.7051487238961492,
 0.6209392629585045,0.2787729873927032,0.7652052796679166,0.8034372509510953,
 0.38150701945385257,0.5053878773290515,0.1029741878785907,0.8785829969167854,
 0.2839090009767875,0.6426499458352227,0.810173888275999,0.046419566927536016,
 0.6715041608852322,0.7223054490625743,0.9601765125255882,0.4885982492562403,
 0.4534696344542757,0.38447385079645413,0.8114062533860107,0.7900837927077937,
 0.8036399978197939,0.8439465236185493,0.6620249884803513,0.3092758780790058,
 0.3925533979315894,0.517453058868656,0.08966909256722799,0.7135503633002654,
 0.6035623927287233,0.14385406027003955,0.6951512532444595,0.528690151296597,
 0.3222169644083581,0.037551876563981956,0.2706081288149291,
 0.3272424095800631,0.2895261746759925,0.1871961684024157,0.5179246569066153,
 0.029388742278535607,0.7853225499446046,0.8408989614042888,
 0.08172724803234588,0.39268789026432327,0.4910178009100994,
 0.35394480736774414,0.3630286056899965,0.8637205770191904,0.5682602540508771,
 0.5171586379212468,0.2423288624478348,0.18748228302436898,0.7343052311269184,
 0.6548576102352519,0.9794200442913295,0.4128948763628735,0.3847144963435756,
 0.6153353234587888,0.43337896759461203,0.20605660134373593,
 0.05837558959658096,0.825229222025706,0.8344923311575899,0.15950162296972392,
 0.20296691007711087,0.3777406059280305,0.5942319578027025,
 0.32219605611661173,0.3757147340037845,0.5477892115056603,
 0.08692216213081128,0.846788275526272,0.7490222084543106,0.2771163959301326,
 0.28854435226012654,0.8577030532775936,0.6899605579724494,0.8224135056803867,
 0.4934095878808794,0.29830098157758955,0.6873887136538659,0.8193228357123387,
 0.36210236616056435,0.9990810667054416,0.737016009615017,0.24326742746513097,
 0.057731837519917284,0.09767371390527357,0.28932138413574304,
 0.9131522512092073,0.5878910437266763,0.37405100068434527,
 0.06666845158412982,0.7007358480456674,0.6598541147441931,0.5809459363476435,
 0.7447322926652117,0.5843423836166753,0.7160775371905256,0.03533599389065922,
 0.611441979895092,0.5850768975702333,0.383875353912975,0.6606102011325234,
 0.22580524449825246,0.32541504618613915,0.8924761782522616,
 0.8653792097374239,0.6643633922545736,0.8943491491317704,0.512163406671656,
 0.8640849652928417,0.1493018087729424,0.9510128227845206,0.7191663479065967,
 0.7878043362865456,0.621489230927169,0.5903960226215564,0.0913231378537318,
 0.5544967438316477,0.5177470558866947,0.87807226267138,0.6852760745607509,
 0.5891233756728907,0.05737561804638014,0.6311387781743244,0.7092895757950435,
 0.4462595477114245,0.8299006709581085,0.8061953905904442,
 0.027219716590470844,0.5823632744741382,0.8367303610027533,
 0.5060348570842482,0.4010063038377043,0.3170491634125663,0.8330284234517816,
 0.9473354962488939,0.043878420203252166,0.6706368256082549,
 0.25315440838016345,0.21555126859245843,0.30493301541434725,
 0.4424053268922329,0.4041066577790797,0.2755087054463351,0.872414634063682,
 0.8249642360991658,0.004901646514286928,0.26296107226330234,
 0.7502808263254932,0.24945430471976748,0.9400767529167042,
 0.37652624885941777,0.6052662060906449,0.9896815233526022,0.2499104909914146,
 0.9025380805129232,0.6602740314037981,0.5634239732447754,0.28184490225617187,
 0.7447075877046101,0.29261749538329596,0.4458782819054693,
 0.32357619544841687,0.9970600277308699,0.6754852462834808,0.1622616621697538,
 0.7063855563028361,0.6464070270135985,0.585028002132724,0.5775327691638703,
 0.7917251269511687,0.026721881549005122,0.6651574492010047,
 0.5846167746887385,0.05422747252123128,0.2933789346774759,0.6626320875437729,
 0.2674882386174722,0.9670262561642861,0.7424077296787984,0.4662468517014994,
 0.11719254553824021,0.6470742568705556,0.414256576992017,0.26059451789021204,
 0.30758343052431203,0.8936699987497765,0.48019659045240515,
 0.9371399751284148,0.009132036948565014,0.017511279085601683,
 0.708879925369454,0.9766984466511973,0.711207280766993,0.3577201559998322,
 0.14860667673768957,0.3421701568694837,0.36705639964771297,
 0.4727935301058592,0.6553370211970417,0.03461535217830791,0.8556105627502502,
 0.35406342050269834,0.4486137872881836,0.7951034342112335,
 0.08495637638588804,0.14863273036031366,0.8592322190280017,
 0.9391027314145759,0.6436117445841477,0.9809455871364753,0.8395924153479255,
 0.4059951909766323,0.7499120356839206,0.8614906909688047,0.5276769048941935,
 0.4740133774179358,0.29609477396222617,0.3348583865131003,0.6679292293958692,
 0.8518791545536756,0.6705343243779354,0.09992566927061763,0.9663310046169359,
 0.3820040816237711,0.7254384174115123,0.8443008450227408,0.611802470859272,
 0.1833735848790029,0.8482830504443484,0.8326260338290236,0.6095183105302924,
 0.7949956743348906,0.687475915584017,0.504006772487426,0.7191688094323594,
 0.8363227155081643,0.9995559930552479,0.8952911628479988,0.17782957205286543,
 0.5651834115436474,0.21052488655638701,0.04185003519358399,
 0.23808384531259752,0.10009771751716878,0.924464362309999,0.8783226692665114,
 0.28678016355492053,0.9669164850280599,0.9178487378430512,0.9102185645021886,
 0.9885665532505516,0.5226925491959692,0.7993291294963055,0.1720257388066394,
 0.7145017503781783,0.7072168421802143,0.07131379755521816,0.8948523853814528,
 0.6510395971202702,0.7879319087474324,0.40622883854317826,0.6648758702450022,
 0.24891863932550673,0.24343795596753925,0.42877243750294003,
 0.8294109817027926,0.660369618784429,0.9163134733849843,0.9142899973445775,
 0.1154676926031728,0.5954465742120736,0.9914754716481794,0.2699512566621932,
 0.8851147575701104,0.5844177254027135,0.9773209712650471,0.43240127617096513,
 0.5902916983145304,0.08078866912057114,0.7036839178982667,0.2696417327221907,
 0.9974380877920809,0.7368742447137079,0.5805997227693053,0.3966707253455526,
 0.4754457181498589,0.27204278506511304,0.17887853588001956,
 0.8890404691521968,0.5627067367012888,0.6455804301446674,0.8470739752099555,
 0.06397862976980084,0.5761036790427092,0.24997561369020505,
 0.9659638641937336,0.9899318988711311,0.5968443354568325,0.9791566088772047,
 0.5656007699246423,0.8917730901702631,0.7341949440038884,0.9791782450364017,
 0.6218711864889416,0.46798242856516525,0.5080356757571879,
 0.47836643725724604,0.23398268517246235,0.22833253525490527,
 0.6616079078463717,0.6758837073950869,0.692197654020092,0.4315239543102003,
 0.9085497818322678,0.2248685616742977,0.6209169595653138,0.5352702501284895,
 0.022967880732098656,0.14511448374441416,0.06555636939816234,
 0.6506293126268365,0.6482070995706166,0.8190903472832313,0.763486317895048,
 0.4348060391143337,0.10319096643048442,0.0973402837458508,0.0681366299056444,
 0.7486505014687698,0.7807601106068173,0.22132360792226968,0.4854373624432119,
 0.5991525999359456,0.48276465945981384,0.9214140457843851,
 0.41425369958919045,0.07273928327806822,0.4819363053730106,
 0.08256934295301077,0.2536577116018942,0.5430564257254912,
 0.47116280685555056,0.6409788782861982,0.23771887130316272,
 0.4669630014524133,0.034906059458035044,0.3748559441759105,
 0.16740343137999947,0.48061046063761637,0.6476203631256318,
 0.4637617597139846,0.8927852504803563,0.16290308418887178,0.3824752701473698,
 0.16722817670843226,0.7003023191827351,0.448110726664842,0.46795886670583964,
 0.2278762537718304,0.4256090795537881,0.6643683900828282,0.6378287114942753,
 0.739419661775617,0.2925452517501813,0.9213478802150914,0.5414391011485163,
 0.9981395541550042,0.11621062581258956,0.18841409935213527,
 0.3182902107173283,0.0840477983669331,0.7657076154948568,0.5934434198683641,
 0.7421429417504716,0.600932273492396,0.9407912808355606,0.5549569546902955,
 0.1765130515776574,0.4760896635118448,0.938869537020094,0.07549219727449552,
 0.37814247857936645,0.5713670873848122,0.45043960853442,0.7079225678186737,
 0.993847079690155,0.5191627983955129,0.1681794628611969,0.5139202136558794,
 0.8537767538547097,0.3976862940890902,9.081653670901524e-4,
 0.6900559102930169,0.5306428192208651,0.8164489149048747,0.17227650679423567,
 0.9700540113093228,0.3212208331840125,0.368821548606048,0.8022967123864636,
 0.7188647724675972,0.22989538639354823,0.7300333881363471,0.9086096043246621,
 0.8843449046417947,0.17984796043261908,0.8539795617797192,0.341072601277558,
 0.3780134032110446,0.20128743192547482,0.3808300499005782,0.7607358530290239,
 0.3834550988212294,0.4197196994516409,0.9304529003329637,0.869465034391574,
 0.8195449338336619,0.6796632068261732,0.1252074682703085,0.5992223220618971,
 0.773773025401168,0.7583469684325638,0.5821634343605103,0.07920638453279072,
 0.6269354941735868,0.5693398625436302,0.7071400028065942,0.4058513796089389,
 0.8620511126592525,0.25194875449643883,0.2778743298737528,0.8586309315706753,
 0.9479273905612796,0.9238655780169375,0.7787167032676918,0.6400435108923156,
 0.8854848629097499,0.852636954680093,0.4594824253948151,0.008747709575245177,
 0.46808055907751256,0.4512878579700508,0.737349524653946,0.32133167561594256,
 0.43561272811938045,0.31462908576837856,0.5623823578196767,
 0.07745324875792681,0.023966694836543345,0.34780763905043455,
 0.7786708670729705,0.9145163351107349,0.6416531431747241,
 0.028971516813169007,0.9128587205868391,0.30080182267960653,
 0.8924210840244143,0.9552620648018599,0.4464465187156139,0.5449222722316381,
 0.02067745746599181,0.3610589145622999,0.8036272826395563,
 0.19235943918770093,0.49053712725403187,0.2748991178539374,
 0.08940983502645994,0.17772762667873254,0.9883059804457148,
 0.09061776309597736,0.933969369837541,0.9602771418246338,
 0.033195528688977016,0.14629378829502415,0.23087923262026977,
 0.6126581063530415,0.002788166898151756,0.5718402079144305,
 0.1180397181181595,0.0803882872310756,0.8560994906627599,0.7886729471555801,
 0.06370077576085276,0.8650561216163777,0.07443656855597369,
 0.6757769503434831,0.740992063160855,0.9122398007695501,0.07555601662739841,
 0.9959874180887733,0.9125822554557723,0.8590276765694309,0.21584163546254076,
 0.8513033548048843,0.7555972223887859,0.6412345331944143,0.9679278674592893,
 0.3099161001647074,0.9198905180971951,0.39309159801416216,0.6305737368075609,
 0.4041314314615366,0.7356833097677373,0.163789635613525,0.40809757656019874,
 0.9773426136694193,0.1785117516407464,0.6295889110455802,0.37529564139216154,
 0.8747377190439434,0.4488721628728096,0.0313516211842888,0.3311066317343776,
 0.38115597688477254,0.301104753589321,0.2268547937672145,0.9803755324173197,
 0.215300901187072,0.08040243296410365,0.1553155665863637,0.6084692087159291,
 0.7294915496936988,0.9343211005385402,0.19430206585632848,0.7631628867546237,
 0.25525013391310203,0.8810022152483812,0.2688825747428152,0.5548593279338786,
 0.39007756143263594,0.755552195437744,0.1584077998615555,0.5670127853553872,
 0.310593594068612,0.9944899575848554,0.40139647019114566,0.5719367797573611,
 0.24441478469013778,0.2628228237028962,0.978480315557078,0.5543983140254228,
 0.7150304549450516,0.5952741382962828,0.26760420684302033,0.8631680680921204,
 0.4262809535967721,0.3369531220730395,0.4297010184910047,0.3327287883690737,
 0.16277559763848193,0.9016392297897649,0.055673421945622636,
 0.24816923926834644,0.04187468422232199,0.45724393875404656,
 0.32879394872247714,0.13993154982443312,0.7074646699565076,
 0.8254131907156254,0.4818430606237005,0.2250178953523434,0.14560242629942044,
 0.716052611146085,0.034748582769725034,0.5592301474987886,0.4880260886653405,
 0.6628177917426481,0.722983729660196,0.05043530947741193,0.29471902892750146,
 0.8564936944360468,0.266240011768366,0.699121488966884,0.3818636982575859,
 0.6423063210328486,0.32718668142482965,0.005079252540170831,
 0.5480061989165308,0.6621895216120998,0.005511463066007027,
 0.41961450467972117,1.5312265471134623e-6,0.4588038837681333,
 0.7066176473343885,0.453219897694928,0.6774922204567111,0.9427515138083947,
 0.6139399302920586,0.9782656301719705,0.5044461909227067,0.33253335586424704,
 0.3715482604679221,0.6059136064927219,0.4926219235056367,0.02794506949322828,
 0.2588259599774728,0.6022232136702119,0.13759101847362198,0.8798568076863371,
 0.1660573482182811,0.45223352321507304,0.0021012329914043004,
 0.757350969267743,0.578294338853911,0.8251653299214814,0.10027242864622743,
 0.9215909036552312,0.8522937598241477,0.9040582322099079,0.23978075610119443,
 0.08667592487022857,0.1637275492764919,0.8805393016952009,0.7929515158919064,
 0.6545570026351792,0.3270603411638454,0.30080877195074895,0.3746174920687697,
 0.63305639960772,0.7151829851023381,0.9990762568475072,0.9016330625719554,
 0.1355344148948846,0.4775065503428846,0.9639599584257281,0.7583420847077428,
 0.600154202374515,0.8069057503718112,0.7194787764592251,0.16818061195255118,
 0.7410016758348748,0.512251760529953,0.9713764184491642,0.32161964915400754,
 0.9402959856821697,0.6095983223861984,0.8480144137741641,0.43848901635700577,
 0.916866916747529,0.6570675481226009,0.05634501909830614,
 0.018166905109003317,0.7127845103528021,0.2928659891881127,
 0.3742170382631196,0.6527897563878137,0.020237336123317684,
 0.13516991836844805,0.6342592993266593,0.5707470335278139,0.4001192769919798,
 0.8409489282235915,0.02092311518203216,0.643352487154052,0.7498224848474251,
 0.8801014466833754,0.12857548612923086,0.4276799836613008,0.3649280120548287,
 0.46732787233908024,0.19856229848580376,0.12085633795015827,
 0.6429870988650088,0.253857278221998,0.8231287777015781,0.6533678106194876,
 0.7058050261773114,0.5333502071710965,0.43599785570619376,
 0.06757043839191268,0.832245687786626,0.32238581323822746,0.7884802039866246,
 0.10578866197655734,0.8600817054563468,0.1967280269806282,0.1936341069405353,
 0.834311616995292,0.8277206892673165,0.2708750535026585,0.3387969150709197,
 0.20552665740434095,0.5739696637559843,0.9178064744019796,0.8653291789538117,
 0.17274124039281769,0.6931509309980688,0.26787057795880853,
 0.34522729320620404,0.05213327488839181,0.9285064460455607,
 0.5465638715118587,0.6287592592634639,0.6590186782042645,0.18612851036916034,
 0.6519481970439533,0.8776954552232403,0.03926029927436825,0.9584043771723257,
 0.1179325484898357,0.8935515685187949,0.6396132856824758,0.28661953841163756,
 0.17379583494353046,0.6271990920866044,0.4347378650649635,0.9076964855835269,
 0.8755887590794852,0.18801203044678794,0.16172812316439733,
 0.7276065050028445,0.5663138512913544,0.6022544011151604,0.9566705767664317,
 0.3987591374216891,0.49499716286586115,0.4984353299352763,0.8273911356429963,
 0.04951519437243168,0.13900930843223458,0.6759517013678898,
 0.2921095839959593,0.18290492184330087,0.921815653619523,0.05909161400098495,
 0.6884654698520939,0.9155057822469368,0.05409862576574331,0.724704866717157,
 0.8139567313034664,0.7586387370038258,0.4316152715573873,0.9969441468840516,
 0.5802648167051588,0.7687821309693438,0.2799419766232035,0.6447717190310103,
 0.34678897333539893,0.2825711054494997,0.25095983393150845,
 0.9683968312637518,0.11672229713631643,0.7360339051248119,0.9879736073364132,
 0.5988068953827007,0.35211008441779845,0.58803688006285,0.10477009169526619,
 0.7413481537372969,0.3553702595221282,0.7227022493450617,0.5609662153325273,
 0.7877363598987093,0.8788229237842912,0.16339849395338513,0.4842595579456841,
 0.6705703720701112,0.6467884812796006,0.8557397952477043,0.7339988672419764,
 0.38321644110915143,0.07902326428684858,0.7241208453374254,
 0.6116978039384556,0.06131656440261857,0.48886522737697646,
 0.05832356492516466,0.21729829494292752,0.9169637851608529,
 0.7708659795248572,0.5826993299173487,0.8783826002985003,0.1518811322739324,
 0.2310934226217327,0.2235089842152962,0.3710735433638399,0.4679624756781211,
 0.1545071217644416,0.41617665059106,0.8800540821076368,0.6835486877336133,
 0.8196378153535515,0.22253497420014212,0.969596219814194,0.6412297449996172,
 0.6394135296562479,0.303981719412974,0.32721942920700464,0.6809395558806679,
 0.7062678173901594,0.6414982770248172,0.5799961502320016,0.531408474486147,
 0.4686557290773875,0.2952807892896121,0.41030783221046385,0.9259314470652549,
 0.7117781928379417,0.598838381036721,0.3486508002972897,0.6593397126256169,
 0.7993145901115328,0.33684180507227923,0.8097041274737321,
 0.16382901097996894,0.48321236092688225,0.7615857894836241,
 0.45282898191852305,0.23975027504379454,0.2005674475741348,0.950782283908177,
 0.6353527890201183,0.4373164243783907,0.06196666973379816,0.6006872474348869,
 0.08029913808393108,0.20463994492681314,0.431567989969723,0.4706269686624803,
 0.9322812965625328,0.24447627168378105,0.5036052529814422,
 0.24606138098179875,0.8739640719000459,0.1689991997884177,0.9479245077481708,
 0.27939970241939693,0.6985532823040268,0.10982044348870978,
 0.9246891662940133,0.09605010875100628,0.7445205599611524,0.6162309058646059,
 0.5434923921498677,0.49526616072968077,0.21795033261215846,
 0.4978425141886912,0.666155436686526,0.5805304704435892,0.6968876769570508,
 0.4305298255751724,0.2559171125878683,0.37090345756504306,0.9153007306462279,
 0.09638273747822224,0.44230765215691914,0.45084283978494033,
 0.17812209941416868,0.7227190776473547,0.22607936510588467,
 0.9575300104434662,0.6552895502344426,0.8371402086082004,0.4684285490063983,
 0.6907744445357684,0.1943337869993894,0.9378808299110402,0.9053579093454811,
 0.09598823972962955,0.7856362689190414,0.4002403451772092,
 0.12958973143472297,0.3342526787533604,0.8490480904491884,0.9165715674812066,
 0.7165669052508503,0.10805029552794432,0.7194842137337709,0.9042450491017721,
 0.8066425283062211,0.6534697889964155,0.9746293024768,0.5063311002894698,
 0.8197576020888725,0.42912434986691295,0.40307829754181623,
 0.17547167804618036,0.8132753544257362,0.1195838725144287,0.7024700064896658,
 0.08798548119924354,0.5651722464585279,0.39473091356918033,
 0.6989070106232567,0.7670184394917361,0.3263261419020551,0.496316988557446,
 0.9392763941464277,0.32268356361368156,0.08122857428551877,
 0.9831231583067275,0.9749756835288413,0.11635963204586075,0.7630758005604907,
 0.6095625252740149,0.35542074885844765,0.2500072105896146,
 0.29816285883098104,0.14037905473045464,0.054133321054864014,
 0.6465714256471194,0.7837765033990374,0.7343668723933658,0.17574913715702578,
 0.46233938879431946,0.3090999345915927,0.260438528078905,0.8913300347688575,
 0.780870453517962,0.22132416245374076,0.6144865085087879,0.5698999308730978,
 0.8519005946265367,0.7154094301907427,0.6718664708634716,0.3889293908412754,
 0.39782378025185905,0.056774195071544686,0.6436636159517913,
 0.8834241063365873,0.6823932880514355,0.2692536276785361,0.31661116109294396,
 0.3907968039191553,0.9439930996466642,0.7314896994102579,0.1858831708045401,
 0.8031361763461706,0.5516604030162414,0.4887779868899642,0.8253562750036046,
 0.5886784830325497,0.4253899475801075,0.6706118549674186,
 0.037806767942278996,0.48055260010801315,0.22050803834850663,
 0.7559340416289974,0.09272213521320904,0.997947262077259,0.5232595373560011,
 0.20910353888786914,0.012076680993079902,0.706423161252164,
 0.43552844983155126,0.22948352488639556,0.013182833580849374,
 0.024086061121558644,0.17580687448295917,0.36818675904582376,
 0.30897928454502477,0.29922603554397464,0.29725429698424677,
 0.6105388370332706,0.9017357451974003,0.4045098351032663,0.9789385673483145,
 0.7168488473179836,0.7725549584413177,0.38513311237005565,0.9365907193500016,
 0.7362072063311184,0.023523403056806558,0.4378080165968139,
 0.05728616563005884,0.1369450093132204,0.9256196089580044,0.9342825953625318,
 0.8702610256714344,0.14129380792488344,0.1441130389920181,0.7017815601094638,
 0.979183820313721,0.7730600208192038,0.6822025122035128,0.21767405886485158,
 0.2892647409358461,0.06524941820164532,0.10747860960998723,
 0.5852154798918917,0.48954943095637926,0.5768086685943552,
 0.042992819841006114,0.11561877896749251,0.7347045302219049,
 0.8824870247377683,0.4774653532671992,0.6487888023360313,0.9734660185930746,
 0.11523346059977402,0.5616285446599453,0.15195429563983265,
 0.5277107573041201,0.9700145625038575,0.3054625550027239,0.749297422898566,
 0.5908770391190079,0.19320027097995762,0.06757056983085097,
 0.4431446868905239,0.8663937936116072,0.26904416548089616,0.7685057054506048,
 0.07357816655213711,0.34278550887946047,0.6551701519780228,
 0.6329329004694468,0.22133298271751067,0.6039775207716258,0.5664871472120908,
 0.18925530709348592,0.9612003777598292,0.05892017299446051,
 0.6893539695842829,0.6629654381731654,0.8440013123080462,0.20284060038253626,
 0.2601862882857483,0.4000895338674477,0.5570004447557766,0.8577960397401228,
 0.579330114943634,0.6388395481583294,0.6927009616242047,0.4110635598331418,
 0.3110138582147164,0.0371550328141973,0.6167845750824059,0.1456228161043862,
 0.36270961651983546,0.26042251603676925,0.9128824774187001,0.915767137412707,
 0.014407799291682122,0.8273775085041669,0.004314412377307675,
 0.22481923829140804,0.726447320793922,0.742050631809397,0.8742730161868123,
 0.8191613587704435,0.6720651125767694,0.9042725289067174,0.578253033064348,
 0.38208985011414764,0.5511276440867492,0.7067458205917476,
 0.21027011825718822,0.43315751816718784,0.7199714342997083,
 0.6481090656632391,0.40784392656225876,0.9392188051822103,0.7377173446184343,
 0.6343170025250505,0.5114702351702259,0.14382082524529183,0.8584663259625286,
 0.336912162132709,0.9994933535185824,0.8490733472207426,0.4589668661104427,
 0.2590900309927858,0.7180443859314206,0.05326245744244629,0.6993691985657711,
 0.46800278442012166,0.48943924056685284,0.600132784400361,
 0.29375019995966056,0.8693801563516285,0.6986320931713288,
 0.16859400722537865,0.6511127639335583,0.3479428965230149,0.1291104357664834,
 0.08094649110160423,0.39859078178024365,0.721421903044704,0.7824668393981328,
 0.8221349301729629,0.48380876144814455,0.8603264675740903,
 0.25443015154842175,0.4143898371400244,0.15933496365407496,0.226699815458261,
 0.3426919694917798,0.5772192490759799,0.2835514541643869,0.15096481891294522,
 0.8527737942636885,0.6376151268901619,0.44257365500147394,0.5467671332672401,
 0.728056720396936,0.8853559012556463,0.34416541523428035,0.8140562927983367,
 0.8235876355778965,0.4517660487903771,0.45395637796898214,0.9253668250107158,
 0.43419658522009463,0.8165365729609955,0.6542114159246066,
 0.012879123659680358,0.028248530694256146,0.8454126312695104,
 0.3090289935729428,0.193137947007622,0.7990380783230348,0.329081603539394,
 0.6231624736153357,0.47812163200140323,0.8819878637750231,0.8686505904747444,
 0.1815795067302184,0.9309439227578946,0.7381906077869906,0.09250030728387948,
 0.7336817188190733,0.46754105013143366,0.3910215168897557,
 0.0033953650218747633,0.5885410305651857,0.5776517525459807,
 0.12237596830822206,0.6145153653583553,0.3964670678155031,0.319137672616422,
 0.56250015948456,0.596038626296411,0.4696450275404549,0.33161497236923987,
 0.061479395947364956,0.35131604530194893,0.7391119746906438,
 0.6938152842643039,0.8292298702964861,0.9722901115360499,0.4763463913249911,
 0.6995910244692728,0.9074832398206201,0.22409720240488284,0.2007456529290027,
 0.46352600825051105,0.15677451140725407,0.7839532711461829,
 0.30694351960752053,0.8204117902404872,0.9065516428109224,0.6750658414629381,
 0.48908949347603947,0.39225321713175365,0.12379653674301172,
 0.34792825272783967,0.9286850555394132,0.21451338638672035,
 0.6496924463753774,0.3980755706041451,0.36945781980739234,0.5334114490937849,
 0.0418690658100237,0.43480337921157586,0.733809597269823,
 0.015834567186406634,0.9953394015657119,0.3019012011905269,
 0.15123591577982332,0.10667411123231518,0.4832935218497373,
 0.05639344021535986,0.3215767227939863,0.6333475407536093,0.677908622236723,
 0.018695268445390822,0.6409928436406359,0.3724907613697991,
 0.7248546514784044,0.5187404457675087,0.7263744237198982,0.41203494157628184,
 0.683499727181361,0.14947542183323792,0.9281923034708547,0.17788839304393822,
 0.2108927086697926,0.39056151549492535,0.8444869625374453,0.669829388014357,
 0.651276329722058,0.49763346203761794,0.9524543635157573,0.8225008090719826,
 0.25002522728149823,0.6851643858432698,0.03014879518150948,
 0.7796377847170184,0.5703004934005297,0.9045223452804811,0.10876684934703629,
 0.39438144789427854,0.05783487317698632,0.887198054312899,0.7539516548592011,
 0.31574879805073264,0.509410197195597,0.37661186371858457,0.6712022712018818,
 0.8473401256117693,0.6459835898651775,0.06011174174173284,0.0888075004836828,
 0.22960210274938253,0.3422739731399286,0.8972793226622875,0.724288307927617,
 0.29829432099498565,0.14453624682292943,0.0641197926318775,
 0.04866398001310068,0.4126894465184615,0.3300087646289934,0.3794956121048525,
 0.6782045310943424,0.5818530015780031,0.022786644440836,0.028200608372078384,
 0.16262855298905987,0.17424228241705486,0.16971794628079206,
 0.37567256324311626,0.5212057483106275,0.4294675537533137,
 0.030169952229944563,0.5339462129491822,0.8364474880343974,
 0.7978275238058781,0.34611605528775635,0.044659813980793106,
 0.08092680509296113,0.778089306807229,0.4694363520599558,0.36407734788514323,
 0.4465220067235215,0.9795689927849021,0.06748428606387136,0.3303514788980123,
 0.2374507508664272,0.6711980207289778,0.2473789981702772,0.31541801992412133,
 0.5719413494849848,0.9455386283964108,0.07967409178667673,0.5630421889015148,
 0.720651086573208,0.699537378943216,0.12032958324823817,0.6230783166841437,
 0.3913028009808288,0.20105104400815954,0.6198124901238478,0.9109236539446697,
 0.06609535904869668,0.8289550622901581,0.46713629735429607,
 0.7714988325138279,0.8397836836033281,0.5609437463860929,0.25690093785883883,
 0.7074106237012985,0.2759542101748469,0.5628094299365288,0.09704862508669443,
 0.6170724182349576,0.2548243521165653,0.3558720799772085,0.601079006294153,
 0.8365711563409614,0.3152052066955804,0.16152309086534145,
 0.30633598160497577,0.1497072900240849,0.6617255184175455,0.9769041452193047,
 0.8074931579356506,0.9709194865152506,0.29558105381112587,0.3566941353135118,
 0.3140028474867519,0.06700770203168249,0.15957299194332852,
 0.2081723161464698,0.4347490084452226,0.1387572265561644,0.35760044584860784,
 0.6392949376670023,0.6961312970552356,0.12381582568802885,0.5771885925498959,
 0.12899532751322385,0.08297593864388031,0.5353041604571619,
 0.49438338357517086,0.7795764935737369,0.7282428027081561,0.4962282084941674,
 0.1098299245852884,0.9434745426266409,0.14497351720115825,
 0.22553765366177658,0.9343273699631484,0.10656271912701065,
 0.04859743888348067,0.698335864346298,0.12678118083977952,
 0.10106406198085116,0.7357127895047897,0.8440375744847121,0.5384056039251829,
 0.8848884935260319,0.8717694752125511,0.8961963135108586,0.3357094326708262,
 0.2140390153484616,0.074839144241309,0.010481384137291272,0.8215633611210209,
 0.4013244995162465,0.8935315120696614,0.067258274025641,0.15947488389944509,
 0.8687232148394044,0.48165301341258804,0.6326641919225211,0.686121207703133,
 0.46288089581213177,0.707554881820374,0.9360185117326439,0.2343925068985041,
 0.3082416828451522,0.1544019362852871,0.5803257496766686,0.3517637246226083,
 0.9534944700615116,0.69555575508182,0.7141493745647253,0.5883177684698112,
 0.27978900383607697,0.023800444429338152,0.6095186926300924,
 0.7866602626708701,0.8947875996777535,0.9802407708144693,0.46570277922887227,
 0.07851659453332838,0.31392163824836605,0.10675931636033953,
 0.39514579483920376,0.718964897961048,0.7777373963641228,0.9290458093367331,
 0.8573774074713529,0.7093619644552992,0.6069846676691198,0.3297835973076215,
 0.033603037669364455,0.031087772669704417,0.9817963594708696,
 0.0958905076006632,0.03671605504388742,0.3831836701391843,
 0.10906742546436199,0.23455183457909257,0.4116312081661737,
 0.6021052096515196,0.6151463353927399,0.3018813985143993,0.08260218169916744,
 0.038015478525825874,0.48315764798127137,0.020510043601596695,
 0.4247908273656823,0.8876451406237595,0.39731571167608837,0.8873524592741441,
 0.23619790367149496,0.8805122467518915,0.9545108178935793,
 0.31256111722278856,0.9709075118628424,0.7506428653094591,0.5066953895480875,
 0.6629124394090329,0.9572599141556436,0.9684479841612701,0.7995207200695721,
 0.2706982756507469,0.802114803871119,0.7270816279100619,0.06349797603376561,
 0.8137662736188427,0.2576654260958833,0.5374280132987339,0.06773394866440996,
 0.19896925023462464,0.474052245798831,0.2987106580694783,
 0.034181121114456814,0.49978479258211894,0.25215832511667535,
 0.031933065018121876,0.2978167775810274,0.22288433551657683,
 0.3566188902200361,0.43702754573463565,0.6206754144161986,0.9908926110146621,
 0.11463545627825933,0.16001336845122982,0.19106259128436553,
 0.8132296214038215,0.5872339484550623,0.11981489038491089,0.5542246924055577,
 0.03439502635736813,0.5114495183736502,0.2047213183878307,
 0.32265908349668737,0.171672392527897,0.720755129029637,0.7189934492131964,
 0.7538067472933256,0.6371472782726089,0.5318026416448659,0.15366124685983684,
 0.6997331254873844,0.804753625548317,0.2322752454618936,0.8877765015622898,
 0.5404465395593671,0.2735561902414211,0.6071803116839721,0.4944670940001822,
 0.30457752421481066,0.5974508758865885,0.5875314922311107,
 0.10653995182050013,0.4564874217873379,0.13997441010649592,
 0.027456641430476658,0.6680108802567447,0.08005195334159199,
 0.7791800966032952,0.8605706288581937,0.7806268304740467,0.5162997835563887,
 0.2915786987315867,0.656957522187851,0.4892039723376421,0.07345459618636785,
 0.0579380775595284,0.1048322527556742,0.44749574769573863,0.4347037717732918,
 0.9461047321536515,0.9434434783799354,0.8281950866552696,0.7711158625346068,
 0.49761232481358453,0.7415220239881954,0.5945970486847394,0.5347404447188395,
 0.7905516424946453,0.7256560592366598,0.5741104298481311,0.4935330515095946,
 0.8603921575417122,0.8009130221428253,0.30392745730729775,0.7741600626312164,
 0.012533964167894429,0.7545551324662289,0.9473030816549548,
 0.06854492908779086,0.7334246173792552,0.168905856175392,0.9252124540279087,
 0.6611940515969039,0.15421680635341117,0.270581316555605,0.3233040335133682,
 0.6403158782857764,0.7300099647152988,0.7618689890987845,0.12481201568282196,
 0.35336683581902817,0.833591554469863,0.6182729707349011,0.7270709280108802,
 0.1961379721019092,0.8102868837299602,0.8199640927807559,0.24967048066748543,
 0.9648710987068343,0.8327254451116833,0.8733113205410652,0.8949063737494365,
 0.5550266067425954,0.10174269561358329,0.9897109630753884,0.9180744295386807,
 0.045198869222834004,0.9654045868085246,0.41153052618614594,
 0.902429394443147,0.3564980295236643,0.8575134384557134,0.1527462657277342,
 0.9141608828410515,0.06860291616849101,0.22833367304254548,
 0.0958790698688895,0.7426786956279763,0.6225267513708654,0.20957636770860266,
 0.7511187413661142,0.9423852022312194,0.3017013259989626,0.9241035927249976,
 0.9174405709923813,0.474510588008381,0.50943597596533,0.009142592337665234,
 0.14850972249573824,0.6272823386535448,0.6851550965107478,0.837813791261858,
 0.16053869145272603,0.7334941007705005,0.17840625725560366,
 0.0738730244501482,0.9509864473032515,0.7778896655711945,0.6288419175721203,
 0.4474455315988983,0.398522201069306,0.995708296537816,0.3632860572121389,
 0.7363201002708327,0.22177780470786468,0.7971672595731475,0.3049037159795841,
 0.6061727560016791,0.2239603655970408,0.15837106901574305,
 0.11042479708695085,0.31743817492045334,0.9394772868192146,
 0.14251340483581654,0.8574664816387964,0.9583845090141554,0.616713581703454,
 0.9479455874681181,0.7881989182233133,0.9750194140624392,0.08617154795832793,
 0.618510840985753,0.9977156061221939,0.010514468582969894,
 0.40633837029545516,0.07744677911730358,0.36777388825804747,
 0.4088069214099169,0.5069377153413158,0.09094442595844066,0.9287340851660055,
 0.4789408100962044,0.23182323100667135,0.32069271769713614,
 0.006409733321623312,0.7962551774847931,0.5264123705723615,
 0.7006976174789178,0.9291543358188245,0.33994435590970995,0.0672815802556127,
 0.2030905170427616,0.5606688224168194,0.85167673488305,0.21593219178770873,
 0.37059631777703317,0.9785530808720784,0.8560244250664625,0.3262563037010332,
 0.8771540477592472,0.6640958166173812,0.18213553210787503,0.7577631341642784,
 0.28261771456590334,0.11834938128028316,0.07073901577077957,
 0.7655188755730409,0.368664969526422,0.4437308071254845,0.5792868539967552,
 0.052809587825620286,0.17580991561335646,0.5134314156002513,0.48182827989287,
 0.7185317316844811,0.1050644072704564,0.5787393077901886,0.44776558333381733,
 0.3078796974178589,0.3441663400878804,0.5346242532241281,0.05502537462682733,
 0.5835403294023223,0.2764110972743443,0.4188385426722657,0.40750964218879604,
 0.2644922708220877,0.5069979340315209,0.6680817367225882,0.27142860603869035,
 0.20398835209763333,0.8112391602427299,0.27586504231509656,
 0.08030625464536167,0.9569866023593012,0.7324284029064211,0.3330808601872529,
 0.7464137397010306,0.9331121311922446,0.7271882872789418,0.8399051717591905,
 0.38970017354842856,0.20423726710392798,0.5937194698466732,
 0.5694215180327047,0.692516334013571,0.6075873042821212,0.06068780212748415,
 0.016605680371684972,0.9380096039580944,0.6657222507449115,
 0.5739601929903407,0.8894427432299804,0.3848494657651734,0.09265845003397388,
 0.3992037010372035,0.13955947084769027,0.1682727879361472,0.982465774087636,
 0.5372626746358703,0.5223654232157376,0.0666014594624511,0.9188811284015771,
 0.8093859858016731,0.49226267602265805,0.39122997642690405,
 0.3200414726646079,0.6212164809267201,0.30796177916155243,0.8022447038862726,
 0.7959667898186503,0.15925423243951775,0.11706854692443391,
 0.7358822464067118,0.5568881074239302,0.7992671124341674,0.5104404915325831,
 0.45211961154831726,0.8951985415375335,0.7935777216413379,0.1376619232063141,
 0.4013753285365418,0.9724729323842767,0.5787067606392907,0.21665790772820914,
 0.4978555951614406,0.21771460612008464,0.14771151265081783,0.696505278642933,
 0.30066389142062966,0.8419322235036742,0.2004923955393405,0.6803593705923598,
 0.5522284834690309,0.21115660375536782,0.5865527235899453,0.9335860941163061,
 0.800370648112187,0.47317582862150664,0.9048142818703656,0.5700863126871459,
 0.7555699465502489,0.8298871717453105,0.8211967246382207,0.48753435271782286,
 0.8088042819915948,0.9275023010369894,0.515431747317161,0.8734728603309032,
 0.8113090568267055,0.8124935432362382,0.25352320008368334,0.4064770297868684,
 0.3234646532920218,0.8121641369435734,0.614152863235476,0.1676892574858816,
 0.9456855142066884,0.5548471198463238,0.4276666051782442,0.21726680146199162,
 0.412506333635555,0.8276923321078122,0.9317267616582313,0.4714718494988327,
 0.07596838591454702,0.642075085799989,0.31554174485769415,
 0.30513742627318385,0.09851136884834333,0.8216282249080289,
 0.13879798020339096,0.6002253682155747,0.491987469666638,0.3002973013671191,
 0.9827089417330065,0.9480253874516009,0.37258748547080855,0.6340150304688367,
 0.7756502343867169,0.06173116314999405,0.1603056512678256,0.5706123852063258,
 0.48689723762265047,0.0519510677346503,0.969109746010076,0.7994963102424142,
 0.7666896567341311,0.7614115658361174,0.09552179976111042,0.6705189430841522,
 0.46001369409911785,0.3381955180397739,0.9615381984620867,0.2064579067834743,
 0.7244707577149813,0.27492402867099897,0.34655117927020695,
 0.5426381960713301,0.38515990270399736,0.9870333020871989,0.5040174478946154,
 0.6480814719412047,0.11981139235240579,0.7008833591993229,
 0.46425149475205685,0.35584544133941254,0.2812201121960274,
 0.9641525241407506,0.17065241506105255,0.3513245874278744,0.8350457490042464,
 0.35538524586261333,0.6116083041270048,0.6335665617859323,0.9692765161724103,
 0.0011567679573171175,0.7129708682510411,0.019861260997633745,
 0.7590066572902334,0.6161917951224937,0.07511563441055757,0.8925501097947739,
 0.8255080521899831,0.4260533442629394,0.1950687032130929,0.02625844770699537,
 0.3061835038594998,0.6689172157041889,0.37235817526637904,0.3502203368628082,
 0.2448846749624729,0.2109207130553501,0.2956361672400648,0.16957255409360983,
 0.8168992052160159,0.9624136859034795,0.8687666788965683,0.32648460454495265,
 0.9607511459102259,0.685073565954525,0.7246167499081246,0.935761702147889,
 0.7937498466716568,0.818243053869618,0.2913313606146327,0.38757336255560926,
 0.2146177534720114,0.9647869140136752,0.6839710727982105,0.5869462816816021,
 0.47049217956875955,0.08633720252374588,0.948354134832897,0.5447220115019994,
 0.6896783932332624,0.8601725439634876,0.4112024629337181,0.16457954943303532,
 0.1284113389114867,0.02136631752547935,0.03181296496449204,
 0.24552942182075777,0.8714295244040418,0.05613505360295601,
 0.9045971668046813,0.5893281805989905,0.15392695907641873,0.1857375456650019,
 0.5268590935031127,0.9588007101455271,0.5036922053620404,
 0.062201005390320696,0.06672717888737623,0.60077802610258,0.5957181566413802,
 0.60846205284091,0.4748509292929278,0.0307976132722948,0.5434027160893027,
 0.1022075454938639,0.27243564853123736,0.7542037188225053,
 0.07026911887179665,0.8083709221657209,0.36047452939093905,
 0.8805875982544303,0.4480996036867747,0.05663043834873216,0.4729772877551267,
 0.3329732076563685,0.8694643426365823,0.6540200829443319,0.8261842593835707,
 0.08776879212260624,0.8254462798953611,0.5482649412575383,0.0859712202610643,
 0.5863335265450876,0.9162556718207182,0.6861956373968926,0.6442456692998096,
 0.0606485615486293,0.32310079523293556,0.7213991620147915,0.6133073000826783,
 0.7242294905305321,0.6955272475027703,0.8721062137381346,0.5384511859189687,
 0.8436653226403147,0.10928383792660967,0.6473329981358882,
 0.36673698097716656,0.06806806967435364,0.2060896783814754,
 0.3826033451044065,0.26302630645008285,0.21927462313166424,
 0.11809834420264476,0.47370790335547786,0.659407685461142,0.2597229098931986,
 0.3191127195618376,0.9565719530648732,0.4129737707924961,0.8515007599002028,
 0.07163763496277442,0.19599741738136556,0.3532474050989617,
 0.8007319217745242,0.9562091977311548,0.11874563704054286,0.332298622049788,
 0.6865152795955145,0.738128706706878,0.16310709998006345,0.6126510072284992,
 0.8428107690368469,0.182013168169318,0.5657213257964377,0.9296653864086414,
 0.7334093823656338,0.18136209778003687,0.701696548778116,0.9634770019864157,
 0.838371844648906,0.1548703753703562,0.6414658312299563,0.2707947949706868,
 0.9305683741630671,0.7238794507464898,0.4307120852543229,0.0316721253856318,
 0.6245053095882707,0.009014273660900463,0.9127405569989621,0.719147326047225,
 0.823704947013799,0.3396605631594646,0.14706142600635264,0.9097307863725878,
 0.416728826008656,0.32526420138278234,0.3844310055208662,0.7573456653768246,
 0.8451258005547602,0.8552355019965063,0.004732919599340191,
 0.9258082137290526,0.6323894480279535,0.8076340051689552,0.900072244035547,
 0.053620691333179704,0.302066127644275,0.7816162874102842,
 0.40461604243848914,0.37861975663423464,0.13951712468681599,
 0.933206326137781,0.4087619487928086,0.1746767697001853,0.13533652237226246,
 0.6814163878149785,0.7323310749352403,0.5601839682721382,0.4430354867175267,
 0.25594982238512964,0.2803487050613168,0.23195950391403386,0.916548921750438,
 0.4303779271000482,0.9254845553761736,0.8935849691394715,0.5566920885066445,
 0.21851302858711918,0.8979553970806917,0.8293563906618867,0.8080355005164624,
 0.5500457918912498,0.12841466449857541,0.8184306034312167,0.6842180725595954,
 0.9937158837680222,0.7986448895655065,0.25343342243484634,0.1415917392319801,
 0.25435060773086404,0.8689901736895463,0.6182881954180884,
 0.02091359181446628,0.501417676293997,0.9247486453860341,0.5548954771124534,
 0.28848204792741283,0.7527370615249163,0.553344361161677,0.42743672286350076,
 0.3160263082467759,0.47468623197685156,0.8615936216251427,0.7610518757807921,
 0.1459360340117032,0.3989019968802,0.2621592285900255,0.6028230913513918,
 0.8291833059947945,0.96325838279204,0.13985597634702263,0.29451458133417874,
 0.45696589852470726,0.20521974435692247,0.7651512140050546,
 0.8145020307538524,0.8963154662662715,0.850143547641495,0.7990177085851675,
 0.3570912085336325,0.44058516691750726,0.9099742595703798,0.4136186627818781,
 0.7382919836899953,0.7394477033618141,0.9095634417963316,0.5000808499622006,
 0.25678143197791703,0.5359536576884344,0.6101246218158883,0.7651639297724141,
 0.8916766198476731,0.5479241440938459,0.9836453777843723,0.4680170149892766,
 0.08479652045123842,4.238795010580443e-5,0.7021732788458525,
 0.4635864072546195,0.14732382221769946,0.36880237449813413,
 0.11778994426304257,0.28079655611282317,0.5454313251948931,
 0.4030181658173415,0.3173993319394083,0.7212870948354849,0.6016330728520021,
 0.28337930182333004,0.03317804986211437,0.29063902308329026,
 0.5973222771484252,0.37133718404830407,0.8020285790258583,0.9407783760599882,
 0.319730247725019,0.06749149110723929,0.8548686132101128,0.12191362141194184,
 0.5747015918693429,0.3162759484219576,0.560003155727203,0.10523062875632949,
 0.8562882365552947,0.3310003939791377,0.8549084337722099,0.5407695120273791,
 0.2661039155382796,0.28919830432490967,0.7034315927941085,0.9282571228825922,
 0.26493692984342365,0.020144770369204923,0.8824718341155865,
 0.6381909561646287,0.9315979895013197,0.6640387153855051,0.8994319133490333,
 0.6673791696480632,0.5904683846231644,0.4290627280354502,0.912911830775025,
 0.5116602361385156,0.17383428032104242,0.3049413290973535,
 0.37600245126068155,0.5229059089903596,0.4599593738725798,
 0.031893209106274134,0.9912610777436193,0.13971404054755765,
 0.8418283497160585,0.9430566854079178,0.617583941786591,0.4777266647760712,
 0.11934074276265882,0.3708310313569555,0.9584278578717564,0.8808157995882658,
 0.2769203561114544,0.7265909153927864,0.08477176311948587,0.1268083031138555,
 0.8131578170750854,0.02413014830921212,0.4435753227667816,0.8299446214655387,
 0.05860067305205374,0.41267023890371424,0.34189289103923803,
 0.03892325547945352,0.3049551267255034,0.3500301859065872,0.7807089843524329,
 0.671442092060457,0.641024854410957,0.2819599178790535,0.3796986536313095,
 0.9647781005891314,0.5762906878085539,0.5628481769990652,0.1550197248692594,
 0.3557977698572974,0.05082199903889972,0.4328462419370843,
 0.17492906366248628,0.13943108221081035,0.0534029650814517,
 0.7061783321546168,0.372638676847486,0.23915223418676812,0.41752596426975297,
 0.30832420878310174,0.5604986039470139,0.6558540777291197,0.4002946572628897,
 0.1747968386396147,0.7381605534154839,0.1540485653553484,0.8500428831991433,
 0.8733996648411667,0.7246632456951282,0.07357938122847185,
 0.009501898712912293,0.16279560613106359,0.5203558985230202,
 0.627025744669419,0.47356816622755393,0.26186967575201425,
 0.42348115818527654,0.8592311793804629,0.3503079734795296,0.6045885642713646,
 0.24999957878941803,0.7624679644244976,0.12586271462552867,
 0.2450966372201172,0.6036179329801096,0.5983588733111295,0.33394968829368243,
 0.4552643127834247,0.3268181445040087,0.1591818614778624,0.6070025582072018,
 0.8787515356872585,0.38931229017879165,0.9783612058509773,0.3363518695962875,
 0.6785742805950725,0.9966978074583633,0.5251069225799846,0.790346974957021,
 0.48889758101424863,0.387127198800414,0.027788133251903102,
 0.1965470841820196,0.8383033429061972,0.44683969010513114,0.4275715145724801,
 0.880499107460963,0.8918528947373304,0.9019048993443587,0.9403830958195208,
 0.22630061096939125,0.8988442512727406,0.6085979132378766,0.5362793924333794,
 0.1896061469570114,0.5666937812478305,0.07567978506931505,0.4585414094838902,
 0.4073408142266318,0.8675631175396059,0.3446473839592694,0.855228267024532,
 0.20602899962073862,0.8259170218660807,0.1764735282468346,0.9595102828332323,
 0.5682740022385226,0.2514982336965792,0.33997731096753514,0.262362152317698,
 0.9036632456636688,0.7592937609603073,0.5159802768274372,0.1766961838042268,
 0.6906066548247333,0.9054456085353779,0.2714833391095274,0.07081615329423796,
 0.7590018371058952,0.3092748890165702,0.4664946347996113,0.3049514851633953,
 0.7034247415593937,0.3784228909409887,0.4565696576719611,0.7592063766110965,
 0.9214232889287168,0.27004297712407266,0.7643455588763756,0.6200484592406258,
 0.8055652138707506,0.37730907406310776,0.9829862720134979,
 0.00283584356588773,0.34264600091384756,0.2892344775587484,
 0.6350918046691667,0.6337214862170862,0.5483975522828687,0.6185672473087859,
 0.12992750445840429,0.6273315674739081,0.060577864035459505,
 0.7845389156264547,0.3507536168303449,0.7553781851812114,0.3727197972396308,
 0.9906466327530156,0.6169408862570211,0.25357445669770384,0.9633412465556338,
 0.9515080136431837,0.5813993312263307,0.12151524289813387,0.5770766064052333,
 0.46874476753067085,0.18591946593874253,0.08480479750819214,
 0.6916826880102777,0.26650296140931773,0.35052713145445946,
 0.043187209647992275,0.7344949852747341,0.03563150052061592,
 0.8156418387977379,0.5537616047520526,0.04641988463446434,
 0.044993476856425074,0.28705317193930857,0.4102795976963597,
 0.24148831552888905,0.2030927793200321,0.7654212102586495,
 0.49980426011703294,0.7573530927614829,0.35332919440043187,
 0.9158189365507581,0.6002307679494785,0.3002435587761406,0.7347665495988346,
 0.7022308423106344,0.3074602771164545,0.8447384855016274,0.2832316383031602,
 0.5758364484004305,0.3900219641185394,0.570485891425432,0.25261148987852766,
 0.1466249040465657,0.9316314508872506,0.19357710045507603,0.1415405124808179,
 0.5218526557973833,0.27841778449450794,0.3438845518178445,
 0.17432675040286372,0.8392700678729378,0.9024811661574279,0.7681007045695554,
 0.5495657892185302,0.3806851935117084,0.49266384273953245,0.4142768859089373,
 0.1625385503379888,0.14466310874487298,0.5549498713123101,
 0.10919822023619441,0.5696608557504872,0.826589645710579,0.6692562111395476,
 0.6598105371210178,0.3245717600294743,0.48185500576501883,
 0.30372743900348675,0.7012390806520153,0.34146029389124377,
 0.18716831124321986,0.0730901482186942,0.8942252357705961,
 0.42711896763893553,0.47357640503563136,0.8893258948753023,0.474305685057697,
 0.8747557919474855,0.8797927544381842,0.526128445277281,0.59692668705026,
 0.0699947127310564,0.6085824722484249,0.9090863818768705,0.06981203300663152,
 0.61286403848796,0.9828538755956859,0.5673810299611509,0.2973067763764268,
 0.29777426402711393,0.18145244939778993,0.9090262818963498,
 0.5048354365037903,0.21234875871014214,0.048312939769667884,
 0.9062616664481277,0.8105057501907583,0.024256686735559363,
 0.9938756616045009,0.7908228255821019,0.9491899058253592,0.7199274544317011,
 0.980142200520881,0.1914898436984772,0.7460212716955059,0.5555664062709907,
 0.011193220690049799,0.794334962333556,0.8538010485313196,0.9182223094218838,
 0.3279154947099532,0.7360734257892306,0.1921016524408815,0.8110578465072632,
 0.0765655946269439,0.13347208805975397,0.570091191111175,0.9408082167727779,
 0.023236543761604334,0.15942165186880852,0.039328039978158036,
 0.10297796876195664,0.9595642795245769,0.5691619900991991,0.9261156044833458,
 0.0754287530536829,0.7924824685604324,0.8072095790729952,0.8292070657522457,
 0.12345093740719437,0.07053848380009009,0.24827255486247946,
 0.015207371005023895,0.886441458957475,0.8804033588195541,0.7799845353817378,
 0.9815947047179058,0.382157978938947,0.3455993238769358,0.5277725427391022,
 0.915858862621219,0.2684529891410692,0.5785046037400523,0.4541488102980734,
 0.13909478319401725,0.81510052933742,0.8350359666902695,0.30871390484167593,
 0.5659340355431753,0.4898720526596241,0.8782799979783565,0.9675173603365074,
 0.6614088231323155,0.16558657793314446,0.2028956440121259,0.4903466068272936,
 0.489591171039116,0.06088594718759299,0.9890007889050958,0.562818617895668,
 0.4108553648963029,0.19264296556738925,0.550286686950799,0.6193423960321188,
 0.7950272060038853,0.8659841321165529,0.504062234507784,0.10515306398117041,
 0.32972902196598286,0.6335298754205233,0.5210020618445718,0.9556146747644951,
 0.8117961091438014,0.9207741959278734,0.8526678514483111,0.695593439129873,
 0.7065084047529837,0.26966644476057444,0.5889928650796229,0.3358235241273223,
 0.8089586017557122,0.27703279626525235,0.8718105603281251,
 0.11830499263335226,0.7190393483520328,0.8914151187730335,0.2758493486841077,
 0.9446142677741279,0.40603305424246416,0.7512334246381445,
 0.27493132094543304,0.630662826710397,0.05138255354946997,0.9288334315829317,
 0.3274352493761692,0.16867841149895446,0.34216126695069016,
 0.36773098849805397,0.801623185936604,0.45875458624420506,0.5224322371397208,
 0.31757876053914413,0.5791334404259659,0.44312429037287737,
 0.18877545046293998,0.6024395367677198,0.8075879239971466,0.414185263406588,
 0.08481601066254063,0.22101643795522452,0.3912223952608457,
 0.5666589642842639,0.08355159425926462,0.8926643961265297,0.6204129988455926,
 0.10347410167118398,0.28785980125240507,0.9175229553071103,
 0.6211504788745096,0.1357535761928127,0.6369873627847072,0.9883912192213321,
 0.003523794240735345,0.9636847653631591,0.9620027810244015,
 0.5294271004333218,0.43968734322274716,0.9979521565904544,0.4076675762947245,
 0.27856678905151444,0.7541603340715538,0.04081308821592977,
 0.2520700070144495,0.7482776082219754,0.3545085428824478,0.4538157567821822,
 0.26403353819132724,0.572308898499601,0.4704272559045195,0.34073917161788936,
 0.9119434747464656,0.3510480863137444,0.6330415199687138,0.246202508398221,
 0.11386405077724915,0.34676281467578574,0.19056371668787253,
 0.9753904465615865,0.2755773021478476,0.12724231937992114,0.5815399634459304,
 0.5413071338210774,0.730833050782732,0.7320725359140021,0.837756174532603,
 0.2882426159268112,0.7054693113815176,0.4379521881501155,0.425146377733426,
 0.15027219925669022,0.18634588561249565,0.05183056909781514,
 0.27736865100202723,0.4073998888283068,0.9103520137106897,0.7038418790031968,
 0.4710117937701521,0.9956599569394384,0.3415167444008498,0.45683879694683605,
 0.33983867228149855,0.5987397331034622,0.29149228955579365,
 0.6237374187752579,0.6240554598027935,0.7210472858883142,0.8969427595782749,
 0.8221748868703168,0.5216907057500988,0.4572265671035388,0.2438407719211616,
 0.05146365540440068,0.024179708040864112,0.6764688226555187,
 0.8616891364516466,0.11631145961117484,0.4137637906387378,0.834887896101159,
 0.9155955484191608,0.5153021127359086,0.8631802560136054,0.7615480156918042,
 0.6161720033955157,0.40468915468182176,0.16226730239104903,
 0.006405050336563356,0.8646040911376857,0.48352668662296483,
 0.13124186027431828,0.29123898879603005,0.8699204542525506,
 0.07572799661310581,0.5776993419514377,0.8160198868288989,
 0.10995857666203512,0.2543482248039376,0.8418603805712306,0.8881059750729161,
 0.8830578718125135,0.6045217744333906,0.7790192625848893,0.15278718619746234,
 0.7437618290570478,0.19555076295939555,0.425221246838396,0.5592146305480621,
 0.9523044818058266,0.0901558092126794,0.13597947786121445,0.3431820676291709,
 0.6849803788353866,0.6714374562850526,0.6934533871929545,0.4605365330243828,
 0.22575925712262257,0.19000444063547572,0.795733099723033,0.6639504838774302,
 0.9563135974767647,0.42590992249469495,0.06388782846878693,
 0.9799551513873397,0.586560033988484,0.1301844741824858,0.9837037559417785,
 0.42517105612588146,0.35616980311240787,0.8944407164026886,
 0.9539764344872858,0.4000140615766522,0.2744325959229519,0.3894772725759005,
 0.5817205128676629,0.8137142135553405,0.2420468984097438,0.2128682447716237,
 0.3389835973181893,0.3446436361184685,0.152887004110841,0.3590864899850469,
 0.2955442687210782,0.13311189233116782,0.14957177032914892,
 0.7131787330554368,0.10923995042521173,0.6077190086802151,0.6396738740958956,
 0.7801823434002868,0.05868689190403398,0.25966539991837134,
 0.4227833657800427,0.826993420759494,0.4035825388519677,0.1398785458441857,
 0.12066791818971945,0.7920274426977991,0.5196247588983562,0.0258927501435402,
 0.553784697521505,0.1782972080209102,0.26299455862131793,
 0.046600403681656966,0.8960510973335419,0.7098084975558178,
 0.5742438127103514,0.29111600906291735,0.9091123215244247,
 0.18056903450786244,0.2792956403627891,0.32765491216718645,
 0.29905533669932716,0.5212860217309813,0.850316225835871,0.7398074527405023,
 0.9713588396152535,0.26262858278221135,0.7091952856048956,0.5888102589821845,
 0.5680718854615318,0.7557305536275968,0.3604456710131343,0.7922254961983638,
 0.7789328567964472,0.13203979159363866,0.5685097233679951,0.5074203514931661,
 0.4268538074401367,0.2240004519120531,0.8568758725440003,0.3732357102871242,
 0.5419585827612636,0.13109194411347813,0.6721014345632703,0.9083590598080263,
 0.01696269223983793,0.23983947956711482,0.04089739232019396,
 0.8653155726738019,0.23623554321103846,0.27690469792240036,
 0.027279559996029712,0.6201128849348737,0.19970941855758362,
 0.706613293903106,0.6843916421895209,0.3978186848311992,0.6764988611913847,
 0.06089649794163621,0.3657421416115615,0.443817761154629,0.8329986863979236,
 0.9769787824558647,0.3012364076659575,0.969066562096817,0.5002811372052304,
 0.0391678286211824,0.651078247225732,0.44447748837570655,0.4865672076608242,
 0.42319587359681854,0.2128026762064903,0.39518217743148876,
 0.05246060730061475,0.2630598984345296,0.45767652841038475,
 0.14051471399382742,0.15919262847988547,0.6530702459470572,
 0.47553962099529945,0.6666942032869608,0.6583840001024839,0.7661181861259747,
 0.5316026907640646,0.2545756985076524,0.27710919253872857,0.9968092079724241,
 0.9930581127891763,0.47125356558106357,0.5858306051979424,0.3212771249167222,
 0.6323572710045184,0.4045487262595816,0.46920505486957464,
 0.30981509518521677,0.5801686782357732,0.9677684427161348,0.7407836731992448,
 0.07824101902969116,0.6619822900682969,0.3544267164328502,0.6497732931857716,
 0.6224821341159501,0.5708916902387738,0.7845592967863643,0.8462925905687082,
 0.09615950520462857,0.6189280864184057,0.786476663195924,0.25879329575844234,
 0.9233906651187151,0.6942804404859383,0.4857508963840502,0.3635898342041535,
 0.449753679331531,0.13301316692097853,0.36448458137352735,
 0.31735914224524764,0.8943912522471918,0.5515260234130087,0.6215077555001287,
 0.5441308247353434,0.894758179204352,0.21553362275707477,0.20554487520820253,
 0.5958141163209488,0.7442452754287295,0.9032546713285233,0.1348331024478835,
 0.6793271453733534,0.24537690522227773,0.2684477141648334,0.9608425647659958,
 0.7433513115592324,0.5935274393905056,0.786188470544564,0.036072079619214104,
 0.24699392153105837,0.8005945036100667,0.14658375249109568,
 0.49209661695827855,0.4537304946011691,0.10507764167217226,
 0.5207799111347409,0.737403885752174,0.08048799180500321,0.43875972381821415,
 0.7728574133485815,0.03167186954315815,0.7494001530019023,
 0.41401080969044857,0.05244065371740003,0.26099212011453976,
 0.0922608383045006,0.3139667135346722,0.48273731385783014,0.4570614352566186,
 0.7743264036847366,0.39737700514632923,0.5746793060602199,0.4516451385486929,
 0.23654846924418194,0.06992476917385249,0.4452521061681878,
 0.4591590087343991,0.7513634823929219,0.9250640108373375,0.9014214423181725,
 0.5653458223976908,0.47856384155933485,7.125414566694355e-4,
 0.2116435255265634,0.7287304104435127,0.8049677575428512,0.26552150791233686,
 0.09939005656620792,0.10353430199626656,0.8715378616151003,
 0.018874007429389272,0.8589222959514098,8.389659710417163e-4,
 0.7790261124773531,0.7682312887323096,0.17103152498467333,
 0.24220432734126585,0.9773359354371902,0.15930502504089739,
 0.7670698902965288,0.27438026974838514,0.073502948770098,0.18066096890697603,
 0.4896224922275284,0.3729134566320741,0.3337980370466137,0.04523307752823047,
 0.04632470380241838,0.4443314559933831,0.08148800268443579,
 0.1706538917611199,0.1224742899302601,0.2425204341365791,0.6313934302994375,
 0.39638344538760995,0.8577891550235476,0.25686420400336973,0.625457391924781,
 0.2803795477266443,0.9778040528720487,0.9467609029260496,0.02745895188719194,
 0.25899640127417656,0.49426958203489046,0.38400095264718936,
 0.7968785444207207,0.19955163174856594,0.5246911277147596,0.269410273437942,
 0.491921417176313,0.6348904238827624,0.6906302826674212,0.7200255009399036,
 0.18887688302915828,0.49050795513211953,0.7822738210932239,
 0.6385384975158137,0.2548181664168858,0.4023224703284488,0.35388577611911043,
 0.7838595355149955,0.4475622942936769,0.726759039680342,0.4283866563743608,
 0.09774494880170104,0.2970684696207191,0.34718251379219034,
 0.5823150645805325,0.1984614140766382,0.9309580250642169,0.30789536165730236,
 0.12473660117683105,0.23841898208581025,0.9898042077638397,
 0.45625927332044214,0.9144939783216024,0.16224226633731798,
 0.12171568226112228,0.49615976577380816,0.772503249470545,0.3629673937884085,
 0.8690741892692455,0.586394715177613,0.771100911678997,0.12727425654465863,
 0.9412173509088984,0.9882084479468446,0.3488515547507517,0.03215683612208331,
 0.03440458084339848,0.6514451492984769,0.8690819545690679,0.9964451626113969,
 0.6437471329998188,0.5864671342787864,0.21461191934389856,
 0.46722781684409964,0.03516491271643185,0.3991913246740998,
 0.08466259396021014,0.9440363347419265,0.46229612102864803,
 0.8057378959395342,0.5655048938229139,0.008631579841060888,
 0.23947862026298927,0.1545885119480671,0.7683613813429054,
 0.40235354281674696,0.07001133045799801,0.5912188802487184,
 0.5813582687681691,0.5619557720669548,0.31548491712602,0.5966734934956791,
 0.21672485546128384,0.13833407638881412,0.5897567978122581,
 0.6158999861285458,0.42892528675218133,0.7973819055433271,
 0.24151388966857057,0.14927510365365682,0.4923832600520217,
 0.056755459211668136,0.6160564912355049,0.8362593865630832,0.606534293955404,
 0.2919539048771117,0.9651496868331679,0.7866210589209951,0.7988869537804933,
 0.12934469437983132,0.42031125147789683,0.7297922312673614,
 0.6177036590257114,0.6696338575898896,0.4762598913894919,
 0.0015963364171854844,0.40245734335785255,0.11708295549870285,
 0.8084880249469636,0.36177378761334644,0.24216887128639308,
 0.5772802229570073,0.9184175528001973,0.9735432191393371,0.39190803461182455,
 0.6495013135079459,0.7304664569248671,0.608909430882162,0.211555600627086,
 0.8903909673329811,0.38840238505937696,0.1941525197172237,0.4439519459141468,
 0.7981047714021419,0.7360308508496862,0.47934040185615323,0.8504413997137277,
 0.09794127778269546,0.8631752204289999,0.8850656517487445,0.3243558661613377,
 0.6318033745733329,0.21952091120044914,0.7196102270100733,
 0.16697779232161236,0.9083463460331715,0.37250881126833824,
 0.41909731057807675,0.5944995555785006,0.8663448859653627,
 0.028749482543516724,0.7130152337662904,0.2187427946870284,
 0.056546454483879405,0.31981127487633787,0.3870546568268741,
 0.027883500598485433,0.7470603377167683,0.17560266321963214,
 0.9964954971471032,0.23494137187862218,0.328733492792174,0.7316556941722799,
 0.8506176697518072,0.33884573036703836,0.8703267852509082,
 0.07298985524944168,0.48385678486201633,0.4909422181353619,0.812751895893231,
 0.6241007170743603,0.20280135845359493,0.7531584073448464,
 0.09070259539543157,0.7675970727184571,0.1966013064884975,0.5174834534096302,
 0.08124514805925354,0.9230912060238126,0.9431965617835661,0.4045438910809982,
 0.5242718007057101,0.2806150692696574,0.041695580675683463,0.582884485591423,
 0.6046273829497565,0.5926144715921409,0.46457739348840266,0.7012944607768415,
 0.11650613527275666,0.6127514896160703,0.7720505565919847,0.9039124513120546,
 0.9909655717578815,0.7993812597312774,0.17331104541100595,
 0.31570310809502333,0.2296570722246737,0.9430040874340389,0.684559406613889,
 0.7356595075061304,0.5337432088702505,0.6322703978011179,0.0892069379884447,
 0.6146586671417706,0.90437154285322,0.1590732193503267,0.7849896676409427,
 0.5505607496780531,0.6624579187601786,0.7112739946473805,0.1167069828737175,
 0.9477609910451965,0.11105106362445161,0.5444415743346417,
 0.12759541853590528,0.21003277098488993,0.4613667220489538,
 0.17790599205114976,0.02435459084739744,0.04722567617007245,
 0.7927392034744516,0.5203916176679784,0.0737269214534213,0.966362207650708,
 0.5251582977269333,0.6778214302335107,0.10904113413827832,0.7894651033306552,
 0.18819028023670148,0.14952140341805875,0.660755198233849,
 0.26332421661210725,0.5563583386583021,0.17712454433961,0.4810444648370389,
 0.07132094653089005,0.8361471710603708,0.8105242079580623,
 0.17047360495342012,0.3606184287040627,0.17630766428032407,
 0.8061497990636031,0.4603671745482165,0.9087418603433037,0.5473779702339613,
 0.48533733837213533,0.217976244994323,0.4281885721575899,0.8601197627942129,
 0.4676081591128304,0.11133450287559021,0.8585028024770811,
 0.40449864888997755,0.5017593255851459,0.5271723819387528,
 0.18531258992505306,0.7704034649150538,0.999633877131734,0.8106710519449204,
 0.39779055622677306,0.16963626472594895,0.4779028068118256,
 0.9550918141227098,0.4149224433833514,0.2799070067693027,0.2538050040347809,
 0.9226992362665973,0.29234538629281315,0.453847408747433,0.2615233805258518,
 0.785537035432982,0.36593485519000823,0.008365747973820836,
 0.5014736565904496,0.9894906095455895,0.5872777011554335,0.01866991372526816,
 0.18193864427307693,0.33534396286764045,0.5805197495191726,
 0.15402662055290262,0.276338990250937,0.8516113440887808,0.5669783535375572,
 0.19270389627802442,0.669539038963205,0.05762151444170416,0.5721982065673424,
 0.5329836374805091,0.2973497988624344,0.034677857664543454,
 0.9966681703456234,0.8862264614499609,0.6198807612842872,0.39041400248330116,
 0.7671541334694078,0.5245678461239318,0.6502009097165713,
 0.005937746468602789,0.3239057096259319,0.866793527450594,0.0665899970900583,
 0.07919636429878352,0.8587291150242972,0.3351310895831143,0.2071900141870754,
 0.5226503141372159,0.02865103078034481,0.35500744163577636,
 0.3179668741688487,0.10861272350902851,0.6791060958039428,
 0.32443005724138574,0.3360425026078485,0.768145392101123,0.3570678001024471,
 0.434323333845963,0.8733180166305793,0.6303246439986653,0.3919416102105818,
 0.8034111179600911,0.20078200923007195,0.7435938984229595,0.5996671509153892,
 0.6083197958788856,0.7516339881451415,0.3621784060093962,0.09578441153187478,
 0.26780577382884563,0.11105815847040812,0.4583188600925544,0.628956014994716,
 0.8940118800140053,0.049066325948872946,0.4804518543364188,
 0.3849484116165248,0.5712146648349948,0.33643209246625494,
 0.060482393754122454,0.29185401411012146,0.9981929905127099,
 0.010628210848623865,0.5097148351419701,0.9143196901575134,
 0.6188165875728867,0.7898399246923223,0.8688712191557695,0.9923224841461779,
 0.10310067712188209,0.16253177225347337,0.12478054501923141,
 0.13880551386132967,0.06819792113668899,0.9518955368613654,
 0.12735175248523256,0.09831762169898572,0.0839505481505235,
 0.9197072752459118,0.5790597220179601,0.6125826123865603,0.49626369887099975,
 0.8402154476566006,0.8288726586811979,0.004715711105819698,
 0.36946518574756226,0.35031077446781944,0.5750386680823812,
 0.023384993899221662,0.2613467995709331,0.9142179743671697,
 0.18535855286620562,0.24265856082133752,0.8081258921428028,
 0.20238197284763837,0.7720993070644031,0.4057083334246191,
 0.48417769214985973,0.6127193274635299,0.8029132192046172,0.5384872539429889,
 0.7093806980303762,0.5089500084402789,6.808778891564948e-4,
 0.4056431451854915,0.1618669872247872,0.7436214856669547,0.5287507287799339,
 0.7309605265221637,0.045051994766592496,0.9355274044254674,0.74734335924633,
 0.3626347720361428,0.6143133969249364,0.8536211522685186,0.09671417285101913,
 0.8090287801850362,0.5377499987080573,0.40526538411876567,0.6938955598971712,
 0.4200939971634756,0.4709134067611147,0.997477532230558,0.04986944783737968,
 0.6858796252728245,0.7788474583267008,0.5642487103384215,0.6188204653738685,
 0.34676260800589587,0.4821637336042712,0.7439627975764185,0.2734884239749321,
 0.24101034083227146,0.13636615886976067,0.5366934532116174,
 0.5881068679934889,0.8506317793101077,0.3147355208379866,0.8272836846038011,
 0.2528523284715599,0.23554395659835858,0.5387432246859927,
 0.11555007804245843,0.2235352499896366,0.07649209616551189,
 0.3758721752839107,0.8357429514922176,0.9965089927385462,0.6805651782918698,
 0.804189366175342,0.540469001326484,0.26990698000159496,0.606502078690556,
 0.9418707027126693,0.37392732288405894,0.4904173875756035,
 0.20166051223735038,0.4423818749844268,0.12108347118783458,
 0.21356218602193877,0.0497280358455503,0.7689208177310206,0.902582173102695,
 0.6051461162674654,0.5294379950808906,0.8432819818260232,0.8169139528884345,
 0.06289123438910493,0.2069928943788546,0.7837407181900569,
 0.48770250931177417,0.43965367265496835,0.38510142530667235,
 0.8066194527912974,0.4695874371622377,0.5145247652051906,0.45853070105282423,
 0.4377803447167955,0.8832663888386557,0.14845774791939348,0.7455887550926796,
 0.7670995951089117,0.7659607992037372,0.5733761233169652,0.9145860140795682,
 0.11298625697402853,0.3678628161984556,0.18404383039381522,
 0.46951441811918027,0.15857750574286456,0.2795562851717478,
 0.9124188867893702,0.6342648421210766,0.5903176154504678,0.6420689183722414,
 0.032391901181850646,0.47257290940894225,0.37869032459257324,
 0.3230060747877328,0.286168713134648,0.4964971573931668,0.9883139209551859,
 0.7107005818478861,0.6490382425683634,0.7241698066576122,
 0.003525311186206914,0.3299779501451371,0.9457227272429706,
 0.32239565427087236,0.4959972647925621,0.2810020755073783,0.0164921957705918,
 0.33305692484676797,0.2787039642559971,0.33531100584899565,
 0.4583959003580097,0.7885875294962759,0.9191089886229552,0.4342611576265878,
 0.5435204090189045,0.2826812094935621,0.39094747354065185,0.3925511340254011,
 0.20828091475271748,0.571263727512091,0.6221282371069035,0.9124045624273687,
 0.6229490852585541,0.24151829518407908,0.07656096976665405,
 0.6592107189281868,0.9889577642240934,0.5478901099270906,0.896321460084595,
 0.6041546283972506,0.49644291229515225,0.1958000002923257,0.7042471265380515,
 0.3313593462382597,0.31657018362262834,0.1893572270653271,0.863678551227802,
 0.7003625205968212,0.1397178334858602,0.9849358990552481,0.9281346916070405,
 0.7717332742458197,0.011416286106173401,0.794286949791679,0.7958428209387,
 0.49362820400054663,0.6889208321874928,0.7263868887501836,0.5411084828077984,
 0.10057592896614909,0.8923176461151147,0.21780638341147818,
 0.7911326563958222,0.48551380046989234,0.7158958146059621,0.9828103636263542,
 0.7605148871691854,0.9303323061629184,0.3907571773492504,0.6909840252174948,
 0.8899783128746592,0.9596257883713644,0.8057232025542737,0.9376979314230354,
 0.9621404514896964,0.8093702484862785,0.669359684572273,0.886257958846625,
 0.39770723058709956,0.4659600976874112,0.5671888230420014,
 0.32882995056820097,0.9295866201300012,0.011666857487388937,
 0.37264341651101496,0.7502313130226294,0.8178651743870328,
 0.07788731427962658,0.24642198075836463,0.5653283134234984,
 0.38833237739964355,0.6420692955245602,0.5029226223162996,0.6729347612267049,
 0.7154431519364692,0.5345590409086336,0.5844870748943759,0.7426876870072117,
 0.5142157629311072,0.2459536575390484,0.3321044024875075,0.8015948665954886,
 0.920878850343888,0.31705439153466863,0.08925364706610228,
 0.42009111183286285,0.5674357271045709,0.5816803938418553,0.9128444886935941,
 0.07874062688142891,0.13100323115264412,0.3702788280546895,
 0.20556450313132246,0.37126834122726726,0.8107476008610156,0.302201460041969,
 0.07310919104923608,0.6019719703730104,0.633966469269887,0.06908978945359812,
 0.40525300186890867,0.8551820974310687,0.08665953234862012,
 0.1512704370779442,0.40780218976956883,0.7615881137613858,0.5516215323057274,
 0.42355565026171393,0.3276785736694219,0.7989877251653158,
 0.41463364636170785,0.5850192121963178,0.7035615061030792,0.8634584188719394,
 0.9711591684288067,0.15905823236065542,0.8477847901490081,0.381181905229922,
 0.42580627913825886,0.05199568367176022,0.5892498716078548,
 0.9693512951727377,0.5864076971783085,0.1812057262057083,0.6409526418702604,
 0.5991323881649815,0.6017495716541053,0.9419366995889745,0.9176334245505449,
 0.14954696343093254,0.8114901484439702,0.6712523785814927,0.7284859661859716,
 0.45038120067819454,0.45196052916437235,0.8681792441426409,
 0.07440608595233711,0.9429084531859785,0.09234390022314853,0.141458504371882,
 0.34142854502789954,0.9697527078297102,0.8039819502591368,0.7305294199713672,
 0.1390514685231814,0.9375815526879615,0.22925924260360597,0.7231484561184759,
 0.869393165131345,0.12009518305688538,0.8375989331065505,0.5896226367618321,
 0.9088814283102663,0.8518193843106174,0.09534101140457207,
 0.19827505864052242,0.2556559115818007,0.8713392583593734,0.3198767595964823,
 0.841792934665455,0.23428331772396715,0.08708885910848496,0.7001497151937164,
 0.19738484498779296,0.1291764853391082,0.10781261434479883,
 0.6700813335067817,0.21711450246342168,0.7113013901122086,0.4258566367738781,
 0.2553746381033436,0.39032760005385314,0.6778152095379277,0.8921840798158823,
 0.8806628139335964,0.4475354551213123,0.301452410749683,0.5455197357513787,
 0.8138051307271545,0.9591664116859003,0.493071825819434,0.7006930662043744,
 0.6889168239164962,0.7182666417689039,0.7710255125531564,0.45561864534711993,
 0.39817001782986305,0.7743262627058991,0.40252682202581735,
 0.6535233523399347,0.5226156746279464,0.47586296458763466,0.7977289937327249,
 0.355954742535059,0.2787822084453139,0.1826170337433609,0.8353369177952618,
 0.265078508818158,0.5346523941435621,0.04532981291005356,0.4326455682526379,
 0.5724426547851661,0.4649444376355081,0.362728619103618,0.7579220607671099,
 0.3171692959391781,0.9322170638774046,0.561480618810127,0.8644167653580488,
 0.7175439704152207,0.5895114883782894,0.6367674363680393,0.6921521457357315,
 0.1183012463213029,0.09486366744568075,0.42991385948467176,
 0.7483375295563729,0.9525943487866741,0.8933626312858084,0.4838805931094696,
 0.5601041637250523,0.8486845922307125,0.13364795138939867,0.5933489758661338,
 0.1609890581057717,0.3096164703020501,0.08706564213167023,0.6519384888205687,
 0.8247499460560218,0.8491658101342395,0.18337899736812513,
 0.44225723295338304,0.44896281298176055,0.627286929200712,
 0.16366070700372237,0.3475661256990986,0.6729158368680719,0.6986606482672102,
 0.7168127731527262,0.7672710675293735,0.8859050781804418,0.7037718733673991,
 0.5930283973885985,0.2565675668280497,0.8991081698687271,0.29416998576014064,
 0.7262631417831529,0.7438780010716262,0.08202718252813412,
 0.41141426370182177,0.286565694479875,0.9717964019585493,0.13506779078886,
 0.8720535184532203,0.6736648602495241,0.06182401838327756,0.779089452222232,
 0.6394857790962944,0.1167641470665628,0.36707227693255884,
 0.007692628272935487,0.5155790945406459,0.3454067445531491,
 0.7899461756701401,0.6283881675626526,0.1333609754159113,0.8438574973043359,
 0.04611444975280854,0.7862308005550257,0.5740205108961273,0.9855314497188665,
 0.38722268519868086,0.38279882798655096,0.16951462594608513,
 0.8910848745534877,0.8583257778432898,0.6555359465542265,0.20005076676524525,
 0.1452865117023494,0.9394005644192055,0.6928229124309021,0.21274120153629839,
 0.6837461761030249,0.5006173611026762,0.681786090420653,0.09611538802279984,
 0.2854801934350162,0.6010046076663947,0.7450875757952243,0.878435531324922,
 0.6205853972987032,0.3256322287020669,0.007849505340350893,
 0.6296184598098812,0.6033312182876963,0.9615526654311931,0.5784321452575646,
 0.1417271778199488,0.832674016254486,0.2093613036489318,0.07351890121841864,
 0.5947693795619171,0.9252468473149907,0.2834797408116413,0.7921089865856293,
 0.6311295252845466,0.9943294174022559,0.030674887773663895,
 0.6768168680699153,0.5711931327066764,0.27209128785515047,0.9683280337401087,
 0.03778316278675009,0.9160178081527016,0.8393076963446562,0.7034056509099388,
 0.305231468430869,0.7146226362130226,0.9781613888828441,0.37428680125855585,
 0.16675974965456164,0.648915229941966,0.156244453903972,0.43856510726148845,
 0.7549764285705569,0.10500300791560047,0.4532746390478759,0.6984691948065133,
 0.6359289597050568,0.31909782004661913,0.9253928249124493,
 0.17690178926694788,0.6306864503700427,0.6049218848868526,0.7382118831479395,
 0.10369207941998593,0.755245634114041,0.44472981394364797,0.7725714635982536,
 0.2333475459014126,0.6279382696206881,0.504468641471302,0.9734028684557687,
 0.6027316044185118,0.1533028730127035,0.49736724466680327,
 0.27725547485243107,0.6680997279363704,0.4852251044224545,0.7235677255291265,
 0.3352825138080745,0.09797626842916496,0.9538275243005719,0.4948481080294046,
 0.7458851107516365,6.974730406068552e-4,0.29710495342067766,
 0.19270534158666142,0.5031559437187104,0.13884473435395805,
 0.1513251090538893,0.787105721199273,0.135370384610781,0.4990268966228585,
 0.07992678858944036,0.48027762021645815,0.4055655852495398,
 0.5946655348245481,0.23874716417875774,0.6988264355805537,
 0.25718112942175586,0.0039598061215749825,0.2077709163735766,
 0.8058775377499315,0.8033042318269357,0.20530879681984715,0.3767311857369956,
 0.4552720965214414,0.05750446708675638,0.973879354862219,0.8235275300247992,
 0.5017574047766915,0.4130762744832964,0.7633251802464625,0.5177364241073005,
 0.6890649038957823,0.9349159956101282,0.11224121703218559,
 0.14986673244492388,0.35551906208244344,0.3897377967654325,
 0.9727960453994926,0.08829066847907296,0.607196086222686,0.7059490275909167,
 0.2751668861939841,0.1264519894880085,0.23461795559587018,0.5350043275253782,
 0.10938094994514025,0.10547550154574503,0.49515742717312416,
 0.41612976945625535,0.34704316789871525,0.38128349082621626,
 0.7873544958510097,0.1264951622926962,0.03857411485725004,0.6249848916685843,
 0.594325713100227,0.9163609443824312,0.6745714474194744,0.08323920738701784,
 0.8826490328985725,0.6760838427530709,0.32741665954227617,0.9670838288111766,
 0.6285538130593834,0.082082330688793,0.9002927236345295,0.19035679264987038,
 0.09251970084417371,0.6253538657754507,0.4890263326177473,0.764767243888182,
 0.7152901969373104,0.7288289507570409,0.7969365154008838,
 0.014798324203837754,0.7980535364139487,0.5095794516250101,
 0.8446676056512771,0.4087170379955337,0.49988536101353953,0.9301810529534084,
 0.9114411949792178,0.17449469214608104,0.2018540565182454,0.8475367382373754,
 0.4752961257934917,0.21990360139789344,0.5608833920284095,
 0.12963024161399772,0.2922548509582429,0.1730072388647521,
 0.07487761835894702,0.7144461498531997,0.6594953172693065,0.9661337293229444,
 0.19301176183209312,0.48615447247559473,0.9130714595089663,
 0.22295687491724037,0.3405449412133952,0.5903409283425685,0.3798828552422311,
 0.045413061614501316,0.18905018833908072,0.8803482872803169,
 0.17622396080522518,0.5579146224167811,0.7148572537336115,0.4738428223669141,
 0.11934229158311005,0.9591321479916999,0.5581557188175156,0.6344874751098349,
 0.6473029281960058,0.8851891803797269,0.7497250619983725,0.43706193016861294,
 0.9628469770592459,0.9512713004152425,0.9707350561732858,0.19907239169181767,
 0.865306419386437,0.7603921204793823,0.4696594427577768,0.23218927985277205,
 0.6263997693553617,0.12701383367624464,0.42329386789386547,
 0.45761349572165044,0.15685379408994649,0.25555262996477224,
 0.528500924735176,0.4709569624789036,0.1516438646003343,0.6799459439283897,
 0.516058046185265,0.011936990220774324,0.5694464517591395,0.4982828710036715,
 0.5038416626542033,0.355689355613392,0.26380444644111056,0.4868839671006987,
 0.23369094065384322,0.6024600159621468,0.8110808529699735,0.8868012168077186,
 0.35448318391206735,0.04093240496192796,0.9592621826878119,
 0.9478661000282832,0.7280549359857158,0.835826432011902,0.7826779668702368,
 0.0849282291320741,0.6536536681698789,0.04148797513627933,0.4581512885141118,
 0.8305505506007054,0.16349510331597727,0.8698380696242105,0.6625923185227506,
 0.8855588395607643,0.7605791894869862,0.05798864529056835,
 0.11978891676409331,0.8932923032679647,0.9379259060767817,
 0.12729435175148085,0.08416593185639976,0.28683388922301367,
 0.6022638626977004,0.48566023327744434,0.6694077419967858,
 0.24992502845084108,0.5470879612397469,0.7343631501003677,
 0.16407844120141457,0.6068400280441647,0.3480737665229171,
 0.15849423109108274,0.3615178018425662,0.9717019876974611,0.8562732463761398,
 0.9578728061115015,0.23208153709901336,0.09885760517939646,
 0.9127808637383299,0.9155210973482606,0.7984470697776551,0.19945827080583056,
 0.4142284855024648,0.15389283938640297,0.04744859581875127,
 0.5475162251511215,0.7351237416530076,0.7307301330792236,0.4594850439490057,
 0.23618187924192957,0.12810148968577462,0.922475915115555,0.8422012713457875,
 0.6864971456316638,0.15297898291994638,0.9407336341837161,0.6245849320211434,
 0.27483890617621687,0.011989172864623643,0.28298304051961576,
 0.03233603241199334,0.6842446850674218,0.16193370264788964,
 0.10756526986176007,0.5183753012217238,0.35555911988421174,
 0.5275100786926123,0.17307568119912464,0.7637179480880227,
 0.19748323410902968,0.017024828125338454,0.12354869832408699,
 0.889344391137827,0.007315292178003641,0.3684360075174342,0.0895682006723213,
 0.8176341205211893,0.112000526062355,0.8607778833022546,0.1961932250534032,
 0.4358919506570732,0.1511029128959247,0.36622785440890115,0.655068556160625,
 0.8460370152829496,0.9814330156083079,0.9526041393275533,0.17264385297315177,
 0.49488119354894144,0.6476054818304467,0.6417807696637399,
 0.20060982007993133,0.7341166213233785,0.1456072900229497,
 0.17469763994762666,0.006145165981003786,0.17757676110241793,
 0.0781516118244141,0.07363419307794383,0.641794447371669,0.23261020615526162,
 0.5899525069273805,0.7406492642018475,0.2833489845358197,0.881994427394918,
 0.0920708861391567,0.19638025862451336,0.543337060302839,0.7678054134478586,
 0.37114621585600527,0.1166213598591318,0.8892332260477271,0.127743878781577,
 0.3612349403530497,0.07844732208654115,0.28313417733346247,
 0.005597724651375868,0.8900626781573203,0.07329941091161574,
 0.583835029290898,0.4142852076325284,0.13409775850319305,0.08585997767024434,
 0.98420095116706,0.7009942533036727,0.34474380683061756,0.2160101743507753,
 0.5189286572360128,0.11170690794904226,0.5923005101168757,
 0.49944601275903344,0.32806787562356654,0.9301881000799574,
 0.3510581744078153,0.189046095193391,0.6242824170193578,0.4980348732609112,
 0.08703503158504766,0.08307414507852029,0.9441795154856912,
 0.3104246098711527,0.710907898910146,0.39292395995547436,0.35978034226071165,
 0.16190027693865483,0.003594225614264257,0.7408623713617517,
 0.6880653986960349,0.16071452060410052,0.5643630891260931,
 0.30529134001863023,0.3695971917652665,0.7923268397787532,
 0.011689931935221165,0.23134300605708902,0.45057246276971386,
 0.31615866410330384,0.02782512097600298,0.05489887270700933,
 0.14943918013797808,0.856632800339644,0.37650105404238865,0.7451265843359911,
 0.07179195059692223,0.4207693952724636,0.23759623783823436,
 0.005468336295649134,0.8940641326636415,0.9054566897253137,
 0.3719110282956748,0.04620750219461045,0.6324330040269938,0.919838373661002,
 0.45126163562588995,0.3386488719352687,0.00789168431101539,
 0.9319624427157756,0.16293893529219283,0.5720254490035391,0.7753641248294642,
 0.9472513287660865,0.6907370712356383,0.189710337770849,0.840084720558097,
 0.502636549883088,0.6881089792014761,0.6809317326537075,0.1480332189795972,
 0.7620483291905769,0.9973763546183945,0.34984475912186463,0.5592273846884154,
 0.12361220346258917,0.8790444610749968,0.7291726176572451,0.9196454488230575,
 0.5914763931462781,0.594626707515088,0.7910832009402439,0.6224969849406234,
 0.8561462729839928,0.35601468132966685,0.37612884067602637,
 0.3042507085663373,0.3479722693772874,0.7513439454731338,0.4452249568385752,
 0.18712071648801043,0.6197811029382131,0.25179529696226444,
 0.8413254655535916,0.999910103771117,0.85973616822105,0.9262960803386856,
 0.1378462793883024,0.7555929728966365,0.036367700923594315,
 0.1641722937930422,0.5899048044175362,0.8069639524892976,0.0944586561157682,
 0.32552920491947834,0.7360531262218803,0.46738007910171275,
 0.7079655360780173,0.8595414362797987,0.9731819483663308,0.5699394373337487,
 0.7075707713381875,0.46510339100156783,0.3426759472137235,
 0.23445378222986446,0.15189506029254896,0.17716142912914534,
 0.3205517078997131,0.0068408125609170645,0.3367665183515446,
 0.356273983786201,0.3639312187366035,0.24040839278270942,0.8813148279227291,
 0.8979575331920633,0.7768660142442152,0.04288997634393854,
 0.15062006501284242,0.01699399970366311,0.8294523305413923,
 0.6324760238194409,0.6977890768477488,0.35503320210840994,
 0.21353303468287232,0.26633284450022643,0.6126140854455697,
 0.7135206253828128,0.6627519022967926,0.38556496150163966,0.9238189313675524,
 0.35457269813559145,0.8740367115945168,0.41457054538168825,
 0.2053977568905917,0.45304735726607803,0.8360014696593868,0.9607410382658916,
 0.7651624811019027,0.329869536260483,0.5440909893779818,0.052643593749785444,
 0.5782053476754581,0.43295599766706616,0.43259839139863543,
 0.16444973669807406,0.13752585403162243,0.9397305434031713,
 0.6447499146706437,0.05040288242059665,0.25402515704364803,
 0.5591190052330907,0.49516610964383834,0.7394266877671565,0.5632720647304102,
 0.04210371037679994,0.8245581576524906,0.7019577927714826,0.9256733508045509,
 0.7707415230638884,0.5869920484373784,0.8054073805912352,0.5979800433674015,
 0.6591329417634699,0.8382136166333198,0.6024855867232315,0.45476787770885607,
 0.5172938396323279,0.24812432298821419,0.982395235488242,0.5353758616989234,
 0.02586949583949827,0.3547459664385928,0.4802646608868889,
 0.27900321113952176,0.366902590771357,0.16000564412853735,0.390350552991126,
 0.37548154844259296,0.6271957369560116,0.9340377074694691,0.9054114264036233,
 0.5846777722505891,0.5806479672834374,0.6168884397040947,0.4425544398378264,
 0.8108125066126376,0.566060944271519,0.24134974301400225,0.07981780645646874,
 0.667167146499876,0.0011828076180743219,0.2276975359437673,
 0.0010956991596232646,0.06994698993312698,0.36435458936730336,
 0.6756867099771335,0.7659439175529145,0.21599096749500502,0.843853850784039,
 0.09466673402735903,0.8032576532104644,0.5447277651981572,0.9455185552223798,
 0.4197177880508969,0.4029896322898616,0.5256759093382729,0.9557526139079434,
 0.6551179564657987,0.774174051773156,0.7561495783303531,0.603716876486711,
 0.2013527520949998,0.7650544294426624,0.6109953416105693,0.279888149386434,
 0.2410571422591744,0.10759245350198088,0.006298421419700606,
 0.9333769191029803,0.6874693795394791,0.9639826143115904,0.9312990410692307,
 0.5958852567716806,0.8725044274400027,0.5944071152507787,0.31048850204971723,
 0.221182201039252,0.060522422329618464,0.5386559344204025,
 0.06723685499154086,0.9593070413074549,0.4135757788603859,0.8938093822506126,
 0.11686900200931682,0.4333197064974421,0.7957836593192466,0.575579848402286,
 0.6862228460275113,0.13442139467113834,0.001902820489904089,
 0.514768932069857,0.4473034811571104,0.5552424432322324,0.20906324309093027,
 0.7185301057898412,0.33920047162872236,0.2557305756439755,0.709703094563176,
 0.5153811784596327,0.4907229940322674,0.5330239287693553,0.09499677627995584,
 0.44417361943392564,0.7747311262896031,0.057850589405124775,
 0.2961021403410765,0.06317471043451028,0.8263648871825666,0.4735421798360524,
 0.7442341740589887,0.4613215449424578,0.982764094578369,0.4663202196263294,
 0.07424789922457353,0.2908719805662603,0.75183624504782,0.5694059207719874,
 0.1366987246977771,0.852361220622917,0.7262328704975634,0.7186423047352526,
 0.8396446118885026,0.9310770140896751,0.02652057419902687,0.9814002271643286,
 0.6936994713736054,0.20051827111409237,0.9953478305088842,
 0.21396080178638066,0.6305631661077268,0.390208927766998,0.4192132256317337,
 0.2836988857125451,0.06248027009473389,0.9120638987177272,0.8255147181763229,
 0.5120788936408635,0.782253108057167,0.6228578001542249,0.9050972523420278,
 0.02269939024914014,0.1210180390312271,0.18328460949950198,
 0.3581408924437085,0.513800731675902,0.014439534253345432,0.8300157757451151,
 0.015899456301835757,0.4760854766016589,0.10197042079228025,
 0.6756485896076656,0.13846087324340273,0.8155429935485428,0.6954324565084785,
 0.22374050275921042,0.34050632144754245,0.5426254349036204,
 0.6825937109137605,0.5047511091390708,0.7362941476874497,0.8241524082949512,
 0.8523729017483612,0.6760220639051164,0.2090754352816092,0.9391959548664053,
 0.7452651982111703,0.27545908399721797,0.8990280540348861,0.12974044632157,
 0.35382771362644505,0.25585551558659403,0.39022518397515615,
 0.3937719056042055,0.8355800696508277,0.48246080934725166,
 0.43477341480498066,0.29899435877284974,0.6945026621418203,
 0.8569733226496796,0.11610486827508004,0.2577973793301971,0.606796551689921,
 0.8818595406768699,0.014728841258987102,0.27768397637745057,
 0.4540902390141348,0.531135857063555,0.4459510464447324,0.3587809885719263,
 0.6644113282813842,0.5059645634915432,0.1611090665380437,0.2914667291713944,
 0.17863471658460395,0.7540463248636539,0.18087267253555095,
 0.004868269892619015,0.7111041114858423,0.27197881557821246,
 0.5875638643854486,0.31765731378747775,0.8039297022175976,0.3018830226917013,
 0.6584398496624739,0.6275173335326895,0.4621379422879943,0.5455850679299297,
 0.8002663959090535,0.22953765821386862,0.649652214992302,0.21249467729974492,
 0.1618328304829424,0.5161473814671659,0.24517285274907563,0.9349495873303176,
 0.20618947467411552,0.49695506092158204,0.7725669219991168,0.795842070254527,
 0.7312794610728541,0.5265841484551497,0.20976384810893167,0.1847591215788711,
 0.9413082007177446,0.6297866002514502,0.5737274717366598,0.9190059674876276,
 0.15837234164006952,0.31175671147655715,0.8976449182025247,
 0.028106598492413815,0.29994198233205405,0.16124981783858539,
 0.8961066635658044,0.2191748673534053,0.9475668369708975,0.8767085686876056,
 0.7368891322513613,0.7568203761011955,0.10256449858192385,
 0.40431487788586284,0.9293029641241903,0.6798452008235882,0.9786455984445692,
 0.5447370523699201,0.4046368250964284,0.4220401202880434,0.9122275307232415,
 0.9430937292603471,0.2325601862966763,0.9796946634503824,0.9474022820327249,
 0.604895468814027,0.8469716233201923,0.6064953056474383,0.8499830320698643,
 0.875148417514796,0.4839881210997974,0.49295235990323194,0.8186839086463616,
 0.5241074849885915,0.6202739507372473,0.9314138193945404,0.9718497799948228,
 0.18678933955233856,0.6397381129704073,0.10966321611708774,0.056748581847617,
 0.5975225013539679,0.25830928602210745,0.8780508914203986,0.7063715004942216,
 0.31391690727123533,0.8867410811149529,0.6296351860411576,0.6330971011618012,
 0.11437196266610117,0.2109124348622473,0.31744104986221233,
 0.9897444411439817,0.6318304286687382,0.8238606574520461,0.49810818825431213,
 0.7328287661917482,0.3442207816007481,0.014504274586298549,0.555271739372658,
 0.5077058892436144,0.9580092093242341,0.7041505398966791,0.5621275085914911,
 0.14476376490325915,0.13152356884634808,0.6414269353613671,
 0.5081176313276718,0.7629807532763538,0.2774016439552386,0.4676879467278676,
 0.2820146481399247,0.7937421095759927,0.8001964308865942,0.852104667833227,
 0.1183906931798353,0.7433970122272667,0.7194491529886471,0.8543276626421974,
 0.460422245162749,0.7987246446223698,0.2641746353838732,0.9428577092312499,
 0.7571063864308315,0.3314370885668967,0.36427791587662983,0.8219924980757409,
 0.30980660028159845,0.061758735739326664,0.5993655277759485,
 0.10580049819735637,0.17991196972068557,0.33039514091766253,
 0.12683978356084913,0.5036194950091001,0.9728786195195089,0.5500346189815633,
 0.6487317652858944,0.029867805566238137,0.7671093275582764,
 0.6438166876655514,0.4685738335306927,0.07438428920990525,0.8729095029924402,
 0.8900951291709844,0.47450920660620777,0.6112894582251284,0.9937012933322181,
 0.3822808824922408,0.5676261323933942,0.7447723237125504,
 0.010355095792802693,0.6834004972407199,0.45261020153407716,
 0.7291754177485466,0.03237154582783264,0.3227077819925056,0.7113892314260821,
 0.9996977960756654,0.8690443100484476,0.32523443107310124,
 0.051495786261057885,0.13616449452490442,0.9533063136182215,
 0.7832688506842689,0.9714807061576136,0.5677334365131099,0.767076658527369,
 0.28383649634644603,0.8691000779122582,0.3803233128669118,0.4487080224365094,
 0.8916492322195043,0.20443458192231645,0.8671210724878115,0.773116341249694,
 0.2333651371927007,0.9747655016506915,0.35557149427277923,0.846108001792815,
 0.8155494458549116,0.859941754370213,0.7276327394292965,0.9961801436452683,
 0.7333922888988179,0.21215309606948274,0.15663182713535984,
 0.16357902311455474,0.08779872351071683,0.08674009874296118,
 0.07834786061009424,0.172539872856164,0.31549278016702564,
 0.40952536614278656,0.2914914537837171,0.42653043859370576,
 0.23503902654080144,0.6392546203780852,0.5374788201117007,
 0.21499853835563676,0.028138328656891143,0.7119051504088684,
 0.09062011832170136,8.094074607934987e-4,0.5758528470991601,
 0.12775586041091724,0.007800312245158203,0.6808509298687899,
 0.24748485244526164,0.6912283837723391,0.9891657337471896,0.2084355237343951,
 0.6783648839646346,0.3464302543472153,0.3642447329558507,0.6397690414131646,
 0.5545668004033635,0.8542839239367003,0.870452726058023,0.8238449829542536,
 0.988984761335113,0.5083081214659808,0.8180700151650032,0.3377024032495364,
 0.5014678935766234,0.4652384166016206,0.8774926483135206,0.1931443961475131,
 0.002722798252376646,0.1781744343866869,0.7299608293399683,
 0.9466487947904207,0.051958747220826385,0.7111279700985965,0.640521930110209,
 0.2326990217826551,0.5711002727917508,0.15443138034746307,
 0.44259963578862205,0.7754014737822172,0.5718921217283492,0.3542739762632797,
 0.6171329998210444,0.1945471930268755,0.614995481374166,0.12571564039312233,
 0.5003899656413289,0.48739096386083347,0.8678709414057828,
 5.372729952836108e-4,0.8082656469068814,0.3262019200608305,
 0.8986017508969091,0.47829171302395523,0.18298210463443454,
 0.02324461696769109,0.340484487398904,0.9266063659335962,0.8046785929198974,
 0.9590011296343945,0.23172693945001588,0.6857459849777631,0.7848737176620627,
 0.2633801119482636,0.2692662421273445,0.842989480963853,0.23149616973364184,
 0.20752160966551103,0.9964510958670096,0.29678458232749283,
 0.6960248500073791,0.19064275786485574,0.6770232249149207,0.4673999488274019,
 0.8601053996551473,0.2003996467645397,0.2640129329783627,0.5923129937625822,
 0.5360777682064485,0.6987233329250033,0.9132764679003902,0.9414325158284041,
 0.5201179291932734,0.07863542475298835,0.29448765573300795,
 0.12924801557683818,0.4441603426329547,0.6911854883609775,0.6626778500242635,
 0.08625724611108943,0.2652989587444736,0.36712627773882,0.29131052319516415,
 0.9167583988123511,0.49922512825087906,0.7243173166714489,0.5297604903353856,
 0.7650023569770551,0.27179112669877215,0.3322431972606795,0.3473976272645647,
 0.6334736262854704,0.8267248274775975,0.2794686131328079,0.10098171011711632,
 0.3981317512005996,0.6957030985371407,0.26330564893266895,0.9409695645140166,
 0.8303867574864952,0.9191879884924752,0.8298987398976403,0.6798137225319226,
 0.44065091428196645,0.6146251104190832,0.6449622716366713,
 0.06106516320272637,0.9245096344726679,0.8849506515957611,0.8106019382498489,
 0.2310830751701467,0.8318752416800854,0.9502889428131963,0.12622694773978638,
 0.6318649085907309,0.17934188148759433,0.7023121878914443,0.6662782527076585,
 0.9436371862846819,0.003203176247148165,0.44262651082705584,
 0.2612336421267071,0.9943172215861865,0.11683973678912829,
 0.47530468390275127,0.031206091279762926,0.1339426138117683,
 0.0197352793841844,0.9767296399488,0.15972147324440245,0.8873852591263166,
 0.9139173225985628,0.7718853338425911,0.3951208176393486,0.31658597040122616,
 0.5763981918900465,0.5811388309338785,0.9849309185797464,0.3007538106108004,
 0.9415499099661085,0.43866901905204236,0.4446965290756718,0.4366274881926745,
 0.07435563923315414,0.2654633395062411,0.05561090061531282,
 0.20099225970103607,0.5207564749929393,0.6306140209398464,0.3508874726129819,
 0.743140880119097,0.3774979417492944,0.0429108793637083,0.3162870165257128,
 0.5469145173870728,0.0843743242150885,0.5523200879194206,0.820066239421745,
 0.6842245564126749,0.328539821225427,0.9744500478140666,0.7063520575001541,
 0.07272623402332601,0.8590625008759477,0.36904029842607555,
 0.2738371482445955,0.3988953185501958,0.5458383715028168,0.6240531678400828,
 0.035600995823528714,0.3131157656993445,0.45274978791073395,
 0.2172454775753121,0.8235970119210335,0.3782268542653121,0.03658374602060033,
 0.0033062399625235495,0.8125308329901285,0.8332161754377727,
 0.3964392532304233,0.4935553734743876,0.2651650143350406,0.1162634119604744,
 0.7747513377998748,0.23911435848583495,0.32382692378780487,
 0.5949145163529203,0.62080282341446,0.6010779899962143,0.9762608333005152,
 0.48257940338296235,0.3044295877610743,0.5649461933930051,0.9720649705222725,
 0.4250527484983446,0.7588178444120998,0.03256231466127135,
 0.23871774468069829,0.24759260028385066,0.19944547290594106,
 0.40616349194389745,0.409050805244654,0.039056338891462206,
 0.20096549609453018,0.3896574234288763,0.5076992020922959,
 0.016175081009962677,0.13827574036903634,0.05050123622434832,
 0.7504689186982432,0.5866589476523952,0.9627422471473512,0.09765850426635758,
 0.26899819621629584,0.7901806653897129,0.20095245807716877,
 0.8627179080974322,0.4683706243001863,0.7453975655329572,0.03740969029176722,
 0.19690975344570305,0.8010464933458435,0.6619497445666304,0.5457243416617039,
 0.4862600012230588,0.3969622121874302,0.5257950837035287,0.5626705412846893,
 0.11752375251219749,0.3807776043929847,0.7455650458897825,0.5366304484656909,
 0.7452740334874076,0.7046031534787695,0.8098834132057346,0.8917481871588768,
 0.10015425204099748,0.6037795715708523,0.05211218965875897,
 0.39609769619559376,0.8031432683490811,0.7545574223370974,0.2948640487606866,
 0.27535339967140904,0.0879634319453435,0.8038744588587599,
 0.19538600656527927,0.885258916384429,0.6180626877937709,
 0.0018380175831359535,0.4813727651292623,0.055158340072653744,
 0.9714678296466863,0.9438023992268386,0.32310096919708386,
 0.09450042403281755,0.4099049032752392,0.12492160111115458,0.902352355861494,
 0.6443156055259939,0.3744381003670679,0.4346976786419594,0.5774583759945383,
 0.9470639329805096,0.803712759147066,0.07912300253637816,0.7460376207638064,
 0.19098374204314783,0.4869594002153872,0.4710037086648933,0.3703016649002334,
 0.7084520742654674,0.020505536683611103,0.241333806429037,0.6656322858195278,
 0.7880599923551121,0.2384672543330375,0.9838095240567283,0.49324566988191354,
 0.016579908451735736,0.5217086441199418,0.6735770515759774,
 0.033607643381189156,0.33608122017379294,0.6327921805186206,
 0.788978419316484,0.650043983689134,0.25827489810500415,0.9436665083471898,
 0.5660391895205555,0.4198449432664695,0.6208878785408678,0.3459817260331586,
 0.005914135895779382,0.848755359229305,0.7327311668989619,0.9543189886222947,
 0.7029678177418233,0.17715086676658476,0.53693428290842,0.8824182096348324,
 0.8623872757413474,0.8458386371869936,0.7152393006389908,0.807162923285762,
 0.596552322162223,0.8341441322040852,0.8412922531323271,0.608030121946471,
 0.11338417780414423,0.6646348493482549,0.03448000796930262,
 0.6087407076961366,0.03810220205196979,0.9237101468094188,
 0.12312882776975076,0.9508816660538675,0.02814502420473164,
 0.4278340461662682,0.831464644490783,0.049407507401245776,0.9752451510520069,
 0.6098545626626853,0.4642994219594765,0.9369150747080957,0.6483624387026382,
 0.22576181618122604,0.9329405286328345,0.8744848768507617,
 0.08140843681564314,0.6042453389119163,0.14922830717547908,
 0.05265998933257876,0.08292643037463376,0.8539403974499316,
 0.02581571459007659,0.009169009602020717,0.022975952617290463,
 0.43446855799017947,0.7320848776576433,0.7090002185536435,0.6780585824590514,
 0.16441407185152612,0.8772835477315353,0.28700815770638943,
 0.13005931545911098,0.8621079502446132,0.9266778414438834,0.9120488696611625,
 0.1029198207506603,0.8458729363699905,0.9151340886504855,0.25681265477361903,
 0.9119286713790173,0.9769036754580311,0.7803556821235053,0.8743669384727158,
 0.6998759330553814,0.8309248982541914,0.3151741654480156,0.03321384949210848,
 0.19596552667819567,0.78357743436538,0.48781042204383085,0.5762876076390726,
 0.6022468592188881,0.5978559324373465,0.8648104016003013,0.7694329476547486,
 0.09907792546252192,0.384753484710607,0.2845988612746788,0.2634686722139836,
 0.7256214603091182,0.6562062364086301,0.7588580663346751,0.4730485867706129,
 0.7215758435891851,0.16688715185992975,0.3258366750766559,0.8348922355375089,
 0.07864507969828238,0.8384411305087605,0.6052345890598184,0.906791524028498,
 0.0625578845684357,0.07249362631292355,0.49114072693343824,0.397170816192864,
 0.5895666360698503,0.020572479022758272,0.9793434151914959,
 0.3015829171761477,0.008445738986124773,0.4831148553161375,
 0.12632942745944964,0.2792031129159076,0.612344646786398,0.3703983468749543,
 0.2046709179698899,0.46047111058867907,0.04367865279938721,
 0.5523577808499045,0.9883322700304757,0.3792456577749289,0.8458591471954872,
 0.20708211850544556,0.8980881212007324,0.6308112563838638,0.5928381595481579,
 0.39023431273187226,0.7088666967264261,0.2086441112951265,
 0.45450052128334184,0.4774807498403759,0.22841046661275555,
 0.5516657700031228,0.6662369794462685,0.7489091645885736,0.31518359717014466,
 0.6880014521480582,0.27943333588404273,0.14756052666765396,
 0.45467363178347897,0.3570192289794898,0.9383740397870813,0.2979021203814185,
 0.7760074395076589,0.9566961654526185,0.2191567370543549,0.14310281224419574,
 0.1959835009338301,0.024286513840293678,0.582378853740441,0.1849481432836072,
 0.5958140052619574,0.8780659971407598,0.8123756917735809,0.798595323682741,
 0.5818809622763449,0.025482959880010148,0.0963330435465759,
 0.34321793476743334,0.506307577610008,0.3158636924517232,0.5121385021286459,
 0.7529620146930637,0.6132856345671462,0.7925447940952687,0.8129596163140962,
 0.6785964292503581,0.01132266069015464,0.9783717252899685,
 0.38849072462219747,0.3931544052945892,0.3105205678499452,
 0.40983759136030906,0.6534773770884937,0.43267502407414116,
 0.49253987176344194,0.6463202533972185,0.5937137910227215,0.8633992946929103,
 0.8021881049213923,0.9473904262041265,0.9296764378448078,0.2211696334091523,
 0.1616011700955421,0.3522174128672385,0.8582011609848628,0.39781972823254963,
 0.028079533163250292,0.8503221509886474,0.19183383031776013,
 0.6334079712940507,0.20621410071367086,0.6517511197592517,
 0.17383936567714853,0.598389775009876,0.2146365414281033,0.7780458331170712,
 0.32598368428220525,0.8421875389179376,0.6410445575897007,0.7651870031749578,
 0.07760140477381183,0.7613002234569666,0.7708299022244223,
 0.05296983094355101,0.4581512720201164,0.16979150088428385,0.814303688727639,
 0.1877953953328999,0.6041192334221609,0.456037966768006,0.3446080468582311,
 0.6961310230051575,0.5142287809019181,0.7415099045224947,0.1989036333338714,
 0.4200017431144297,0.352658323323502,0.5766573247413365,0.916768982035257,
 0.22631635308354137,0.5729422129043736,0.727298827601407,0.8639894957421419,
 0.4782397786785779,0.298490099469337,0.0955245335383853,0.6299869399822136,
 0.742974523192089,0.6856731022587694,0.8261785260727108,0.6896733405013873,
 0.25725055694669774,0.05937255087692617,0.36895760963008994,
 0.3297516071536416,0.35193137439184263,0.11743029276723527,
 0.40285074233862084,0.355655865520894,0.6613130864278529,0.6853980125916049,
 0.07612826621899216,0.5902014435418413,0.9867433169633754,
 0.26304417730793017,0.44279529002051876,0.9131928212115227,
 0.31349555570236687,0.12424806231117314,0.36368958198694934,
 0.5327685788095459,0.26782051152685327,0.3758748346221654,0.462029879789005,
 0.08563724780378246,0.5910818764652686,0.07294276215846107,
 0.5359708936302773,0.7670739195860892,0.2039517936979418,0.40267236221261593,
 0.49985552429291213,0.30552896317192624,0.8424535067026414,
 0.05720924821939488,0.03605329887435049,0.10721731180645044,
 0.9350006868280761,0.7677495354416874,0.09321876135389628,0.9591255760488819,
 0.7776130389519573,0.6919501050867394,0.4064028985956627,0.1311608130096733,
 0.6342000149462191,0.5245382825344271,0.9013119096696481,0.8131108394392064,
 0.9604699665711832,0.38136259088037194,0.30506077005910337,0.575763619599397,
 0.7556763925929273,0.17113480528164138,0.5494051664067126,0.6946523327103014,
 0.17425121249789466,0.2716092376435508,0.20075702875308865,
 0.4950141899736825,0.28096235921246226,0.9578817856757623,0.6988989048149998,
 0.45967139044106387,0.905253983464597,0.5384162007370737,0.5796294179027389,
 0.20973335862720854,0.6704303874753124,0.49118796355321825,
 0.8773123355918842,0.3343980571170899,0.19448096435743112,
 0.49531907018942434,0.5536334702411042,0.1824831478657032,0.1336140327762415,
 0.5785764894822194,0.9313139852904571,0.5311823321028307,0.3857602277265907,
 0.34522024796810313,0.7620659833439223,0.28322509543195284,
 0.36848570907634215,0.8960401813833586,0.9639362318859854,0.3574241266787638,
 0.1603254922119417,0.6535871615894029,0.5843775353915746,0.1584416060848024,
 0.9139136618536909,0.11928710816327082,0.3700772214851302,
 0.24224085784747218,0.6595949198439767,0.5082928606508996,0.9932627985860731,
 0.1906527770352362,0.1692522100810544,0.7516418413920767,0.9629439937905685,
 0.08666673564116956,0.5171592827437562,0.2736710116472424,0.6941581378687386,
 0.7867442911626774,0.9746007077353679,0.9234294461300403,0.03755558081429822,
 0.9621559080496449,0.4155092683868893,0.3052155479328965,0.09232840188024183,
 0.7337836613773033,0.9385372927222004,0.8184951162372446,0.13083554326437785,
 0.0723073960898386,0.8229170918170668,0.5683921792135284,0.468311588915007,
 0.7872520283605915,0.8612368774428057,0.369926328494001,0.2102063020899535,
 0.5800739130998483,0.8699299852723519,0.8045086210085,0.5138495933939232,
 0.23782921251617395,0.8666606297719678,0.5562118672816241,0.3985625938790588,
 0.442522432933666,0.8049318270586461,0.2544143210567169,0.5867145714727408,
 0.9130863138995352,0.7577242115339297,0.25482065984913205,0.852572847435857,
 0.0968410109848119,0.21650551938446472,0.18204093732654092,
 0.5679631902084968,0.5171660175644959,0.9692536273360628,0.43111148296250523,
 0.2317707969384133,0.6930861214517245,0.3833335958557814,
 0.054230025042595775,0.5863237802405766,0.05001216138550302,
 0.5738246817536179,0.16475699771806485,0.7928051929810384,0.5887360194222836,
 0.9325038114469556,0.2750293715116172,0.5930274393874778,0.7124629781394753,
 0.8625476815170009,0.05358678933581451,0.3741218575587085,0.9185864397580957,
 0.10272714220140655,0.8359949117717146,0.755068025154356,0.19761880416781996,
 0.8930769240232942,0.0669511927037304,0.8840630709883754,0.90325277322325,
 0.5736084792894226,0.4276131436351239,0.004485938253878552,
 0.7029852500219953,0.41241413261387216,0.5567497117419007,0.3074637507400807,
 0.8888682167025309,0.3669437262687515,0.15729437062397966,0.5434919986169914,
 0.29875907762349774,0.6786778464270457,0.1325267851818256,0.5188018296302086,
 0.3002404252908055,0.3440411803714556,0.0774196444944405,0.9447478756369537,
 0.979074822209224,0.23729241117072364,0.44005422038900655,
 0.14417970475227848,0.6803115923740609,0.42242818868829146,
 0.6302150597348083,0.8035011488039766,0.09404420987122997,0.4780435559373353,
 0.2822089314776789,0.4614491854721302,0.8092940736110805,0.23299447482941116,
 0.02866534012352684,0.09440556679931578,0.28875999660565543,
 0.6136575418361402,0.1819167214531605,0.7782439052095137,0.39715098159411866,
 0.47863709097931106,0.10806774445415934,0.4927246386619667,0.952788865435821,
 0.5059266691801301,0.9706814121394322,0.025029762504785613,
 0.13357206052992976,0.2442490104763222,0.8889227974393087,
 0.42697430048706264,0.08976908086621183,0.24490917692915504,
 0.5532105060498859,0.6931747617945632,0.9562269969144344,0.7394872860139072,
 0.2890829872549928,0.9952213231753081,0.8570551275237341,0.34476387538577624,
 0.08571578239547639,0.024693823954883287,0.6958007883997346,
 0.6128738424890021,0.031822606828191935,0.3305308427409548,
 0.8507711926924217,0.9091114723047813,0.2754281156722497,0.3682131544291748,
 0.21458343302431826,0.12370620755811601,0.7479651035359565,
 0.7527022908796661,0.11919231401961272,0.8029562386864222,0.3832788808305393,
 0.5830336835444367,0.5619816053098325,0.7289962171403793,0.22136147687627772,
 0.3803503296954256,0.36437987206678846,0.031115175696479813,
 0.4982339154523191,0.7904062751367894,0.7955277961332273,0.5262030622567676,
 0.019404228338107288,0.23586019187146645,0.5866791028042085,
 0.351916197300576,0.3763789020821203,0.12211224454950265,0.5640250206675607,
 0.12066776638082943,0.19981629955898506,0.599649088533593,
 0.39533958637873834,0.11861893989372407,0.566071515460618,0.7157078927831949,
 0.9995532534785538,0.25093827648576406,0.3829749677653037,0.9198490008797751,
 0.06159633442900314,0.5291186382170645,0.5631756290645032,0.8412219901128132,
 0.22286035613845911,0.9009409087579084,0.03347073221489483,0.532676124624876,
 0.4382048343352083,0.6379110846983856,0.6287519605844485,0.768623786079993,
 0.7509634867495346,0.6554724397477928,0.25935718112770345,0.815098452904121,
 0.8872738763889598,0.38608438998432604,0.13662892513154157,
 0.8081053311097854,0.01832546646531341,0.4341974569604834,
 0.38264416693473935,0.13921216551274795,0.17093479057083694,
 0.9579534574028017,0.009774727385634785,0.6501819520446316,
 0.3443391022994876,0.2706840089456495,0.024170085282243958,
 0.4202360598800452,0.5690756292844238,0.7900031743220327,0.7951686596480321,
 0.9116991782110271,0.22496424650600622,0.4631506123042666,
 0.41656409880411194,0.4575708964720754,0.4611564312363322,0.5223715216457351,
 0.8027368333775129,0.4051584113104516,0.5499619661499663,0.40112848235043885,
 0.6460627260411255,0.0302631657566107,0.7185808240669764,0.41597850018168203,
 0.15730895056952965,0.36557525577847205,0.2036242918140021,
 0.3205031530766136,0.8494686401217904,0.48672392778013773,0.8208613876773744,
 0.4777896456992353,0.25941437379980925,0.6754668762981448,0.1660370699875967,
 0.977983339187011,0.6066638915681282,0.6182142546728259,0.8099921981712852,
 0.6876038736951053,0.41796809936312707,0.6604279267004621,0.8931147205314923,
 0.491430506086203,0.213817752456673,0.0561651848141278,0.2427665759784403,
 0.14764353577679334,0.6857783586643348,0.6858558178437149,0.8617919096988848,
 0.4148458680916647,0.24141157232998245,0.3313280577808825,
 0.14704263499950054,0.952032992490521,0.3185510044285603,0.892960960168326,
 0.9437145724352353,0.5703640024440905,0.8309114379634122,0.7555999818739788,
 0.22185168247796583,0.06808797987591109,0.8615392814675741,
 0.42125266611320156,0.6590605638615724,0.7159296540727802,0.3500478391415801,
 0.734551357345601,0.5871525730722071,0.5980766998028806,0.7344704617269998,
 0.0923682027853836,0.5025678944583051,0.15214437976952622,0.4189632550512139,
 0.22627285519734375,0.8420495739460727,0.8625004965070656,
 0.35632307799398233,0.40728338982926227,0.6668240306781501,0.515091801518457,
 0.28009993689666945,0.7475895657924503,0.742751292972396,0.05360469567991599,
 0.12859241640894992,0.8958976380307853,0.15678219674324345,
 0.9166498446968906,0.33775267212133,0.801008769724722,0.018850849260738878,
 0.6856820332686071,0.2040169352211898,0.798162088898255,0.4011065073021408,
 0.6570522976128553,0.8078119988063865,0.7579934598397035,0.08305875109556482,
 0.9397804213826573,0.8317846283289123,0.7850370651028873,0.7323794994691213,
 0.17058837762962298,0.6154447827696397,0.2857573684893798,0.5347800831433189,
 0.9280348530151077,0.8802228553036909,0.4824463794709576,0.7387595672549598,
 0.3637912097528234,0.8387871647453709,0.4665835447513651,0.12461831552358171,
 0.010652495197466472,0.6923038852492124,0.4927234505388264,
 0.6220817868883713,0.16629677745582272,0.3644021163098903,
 0.05925804705187965,0.2173278088158438,0.5138159113909044,
 0.35741015343228133,0.7729716209573203,0.9491391664760361,0.7005018716110039,
 0.1683870290274374,0.11345279954011467,0.12932404592453628,
 0.4938011989453148,0.909598450721467,0.5051351843942536,0.7916765652342757,
 0.5441064752307326,0.06638344982863964,0.28787676158621134,
 0.1543958702475553,0.1814362582571304,0.6382558944839207,0.11249348938239168,
 0.9760694724815162,0.17698112739644678,0.7862725682198378,0.8135324299032238,
 0.834098743235103,0.012073259241454082,0.5153874664740176,
 0.49798650909515707,0.6711286155962116,0.8421734720348965,0.8063100615610508,
 0.26075733014646885,0.2513048062288906,0.6431729860727226,0.7871771121986452,
 0.33865624735636546,0.5091080924079303,0.8077957320957373,
 0.41523311952861497,0.03144685423761917,0.6813019293440542,
 0.24140758178032462,0.5978711362579546,0.2166725716935468,
 0.16728052738245147,0.44515047426988774,0.8144364186683999,
 0.12103898010864489,0.27686310979146445,0.6768413646995002,
 0.7587512305358081,0.09440206711158405,0.3723382823750976,0.8099909601382282,
 0.1948011369794399,0.0305278988871851,0.10774354641364803,
 0.39614675488022866,0.7832074890881433,0.8936414049362474,
 0.051502633105768525,0.5539186727571958,0.17409521031788422,
 0.0781916161792734,0.29759269201477967,0.3567795354527734,0.9269190299162453,
 0.9751904949244488,0.7712856782143995,0.3287467675740088,0.6197420089728763,
 0.5050817442142145,0.5340970165797517,0.6302126027394849,0.7783304460020203,
 0.2483855968898866,0.7847255239533063,0.16729181309761154,0.6314940512546318,
 0.7858615504248836,0.3198386163949205,0.9728436957064032,0.6155810530570441,
 0.3819884038259562,0.009338482675577686,0.6735556701822838,
 0.7380555494642239,0.467846356651743,0.2049948680160817,0.5483568192590029,
 0.6273542291308549,0.7067705296423934,0.5045422050858069,0.66641085447738,
 0.4790347539049896,0.6682921134378679,0.2073781512488445,0.8464141456062784,
 0.19530226905204695,0.8336604063408694,0.7817273180685828,0.7699362514104221,
 0.5639749252750392,0.30399682472610534,0.2866699385171584,0.4568600633767703,
 0.23018484251140237,0.32058098837932736,0.2835388966778656,
 0.011511857122062374,0.8485358669915594,0.23134673933127103,
 0.7941512612822562,0.839655010867748,0.3142561382444444,0.676939335375186,
 0.48267144199203416,0.053998795899631746,0.08752021750757388,
 0.909249689879665,0.6381553374095789,0.056781695713082225,
 0.19902716507290796,0.957140886230954,0.2518131924172894,0.5164777197144007,
 0.4143918655488812,0.1072547226831192,0.1392181065940562,0.04977721201413898,
 0.46807142396566626,0.37615998449588495,0.19080641278691646,
 0.5507670886150658,0.20907888031859,0.9662839567874377,0.7293039683954198,
 0.3605451535233365,0.025493355767324255,0.8878321531966296,
 0.22497797900697702,0.7668935387810876,0.4259315829706667,
 0.07630278779406352,0.13781021868182775,0.5079366634353044,0.469491618498329,
 0.9932966400616355,0.9176742327353357,0.2868662853196553,0.8807046000950256,
 0.5709989075811754,0.9918860493812919,0.15455049966685253,
 0.03161409805050641,0.7990697089823182,0.9850364063560036,0.7989330321000409,
 0.36774811614661473,0.5985154646302862,0.8332676029740721,0.1526338499001444,
 0.5766257654964764,0.19308370906667505,0.9394657593782249,0.7511121300321345,
 0.9771720376216643,0.822491860522207,0.020328954472567995,
 0.16891017029589828,0.8732609493199394,0.6666877139791649,0.9075801556027259,
 0.027476439679141373,0.14854577848808403,0.6684625086018859,
 0.6239242725301002,0.5262651592848355,0.17579559039887427,0.8237706225101212,
 0.7997637593462009,0.15046379703743962,0.6237697760930893,0.2910695080680018,
 0.7686395611868205,0.47985560388600434,0.06672555711637795,
 0.18885061171993156,0.7773258501936915,0.8951017722421011,
 0.18987632271195742,0.3619893418246358,0.34358109619534116,
 0.16230044305085423,0.5554054193857351,0.6152813019486041,0.9400130835949789,
 0.61532994270423,0.26708102354309693,0.19800925015438087,0.8458760073260859,
 0.16569380527872413,0.059496957956992946,0.9341845076118969,
 0.35171471408813837,0.20584297064138646,0.9526527738146408,
 0.38698730555274974,0.4540759389305534,0.7446196137553525,0.6138094037475186,
 0.508038866084074,0.7896532778099491,0.9975632732434632,0.10601368119899235,
 0.284178576607104,0.18941612843416378,0.59781950468438,0.2688825399920365,
 0.1937947269559257,0.672138940448478,0.854968891694978,0.058903618232039046,
 0.2713296460251312,0.13317968048362272,0.4162810595046931,0.9506891671259156,
 0.9668467268406717,0.5567147956387662,0.7422795900313143,0.7875637032345164,
 0.12844415130278364,0.5212123310081492,0.9353714497190362,0.503136425166771,
 0.1660349953623974,0.7044949192990863,0.8263053241362971,0.17935043675396944,
 0.9237043584781086,0.8021399670749382,0.5539985854932877,0.04979336799990963,
 0.8344759574428786,0.18575028059616205,0.012569417530725047,
 0.8577941930708398,0.05513547436362021,0.18769086725530326,
 0.5131329103432303,0.3188633344635431,0.12152158315753991,0.9213040567589516,
 0.3083688707947271,0.5566539675501776,0.9368764854468258,0.8471664759108524,
 0.6067925246673354,0.46665547908493155,0.9265243054584767,0.5522135677352538,
 0.8155664344629137,0.5117631636387645,0.48297677125976457,0.5167780793571448,
 0.06271765210653146,0.30719598969804063,0.9666521379374903,
 0.3200638793129117,0.7447633739467029,0.8437152323766547,0.21586336129651085,
 0.10553783770007208,0.7566084319545401,0.5362646914874877,0.8798928174114657,
 0.7065968006302626,0.878758355341575,0.7709802328315344,0.03072736860640879,
 0.3157007714778487,0.5917918523688821,0.8493330491424752,0.17482605996605516,
 0.17076779766201422,0.17609961005349561,0.5965546703963786,
 0.2372490048233742,0.7431333870275186,0.12425142642303189,0.8634029295177671,
 0.5059770213951192,0.38878849727184484,0.1417035503154649,0.6523566975804537,
 0.7128774116747447,0.7136653946089283,0.8327773823487145,0.8601023077360611,
 0.7309662910238325,0.8121360526300769,0.226917512464629,0.4976616750615427,
 0.09813548011641171,0.7975419321717702,0.33102189314292796,
 0.1210804073924786,0.621823009614992,0.1898603346253448,0.062453292992944,
 0.6068806054908361,0.6810701877766773,0.756985778798513,0.5414590310368426,
 0.9371116031716387,0.25175109158788533,0.012864296209269321,
 0.8467780645429044,0.41892308800025135,0.41837451440880424,
 0.11599334564861963,0.26266265230006147,0.7571021800798533,
 0.21349562832805646,0.4954131189620129,0.5890332046376823,0.5704069566596723,
 0.46041463242089475,0.13059420350771977,0.26002142948390905,0.24903054135627,
 0.5903716749178979,0.06263997044990344,0.9551006074645896,0.3341590810115088,
 0.6390179465439484,0.4842613486870475,0.21271514312968787,0.8185956223745546,
 0.07647616364552279,0.23552043545296775,0.10571676652969153,
 0.7546825092481555,0.18114410378000056,0.9372597960141569,0.6205553678779866,
 0.15221597899757588,0.6981028349581124,0.972341168251573,0.30233750878277765,
 0.7131429964741118,0.5698078029109828,0.07513007831961649,
 0.35407236591202895,0.22653211893844438,0.6545216420144767,
 0.9722012722548576,0.9574818802815578,0.1679541485574726,0.920938285852445,
 0.12512642063679047,0.24141121565482437,0.9552709189201501,
 0.3674859815361089,0.3476080052197904,0.7583660632955298,0.3925521351881526,
 0.8035902910894384,0.18867266208937683,0.1595277874382992,0.6960788664743731,
 0.24360119814215797,0.4312847586994857,0.1927908497376506,
 0.40245470435919506,0.5587752927272787,0.23524570995188898,
 0.36704665500121936,0.672481900022743,0.3810231482668953,0.8805752815128128,
 0.20472901027622703,0.25959034389232727,0.1770757000492379,
 0.3360171268393446,0.5632921163099165,0.5495345262879167,0.43385224903941877,
 0.7694565375718734,0.555072203911108,0.8202352728683822,0.8937905668847222,
 0.6010619199863895,0.4176171966338992,0.48821746732248883,0.6230033693691314,
 0.46407066235897143,0.5192328531450635,0.2810090232866084,0.7668060142304512,
 0.13384537425748833,0.7323790268869851,0.030408438429258555,
 0.4258801364294309,0.18254045038845246,0.6989982204288062,0.1392291911475272,
 0.6797457407230487,0.9814503009401566,0.5322455972235594,0.3466887371821633,
 0.5698880806231827,0.8945968902105988,0.579185588845257,0.7156136119303654,
 0.394440616735664,0.771009044351455,0.8151242954431382,0.11097700207470629,
 0.8956658315451453,0.9503851752083117,0.14576286363220015,0.8383734686202143,
 0.32810850726023577,0.341896379406042,0.6149137919359527,0.8408733316630421,
 0.059761576529428595,0.11697101819795341,0.48662832016633917,
 0.6440558631790738,0.019444369043067344,0.7873219599314502,
 0.7087493712356827,0.8075350992809004,0.7822553048277583,0.4363258127118008,
 0.4519170929866958,0.5330161759837168,0.02238966245695062,0.9528623676909727,
 0.4791890971275409,0.43637071764263347,0.9758120730625831,0.5664358414745435,
 0.6170768701341188,0.5497356871581149,0.9544213578694368,0.11525803416096181,
 0.30049597701223496,0.29453935464877135,0.9601434219114319,
 0.8271567761516398,0.2728525826571677,0.8501198040674461,0.4302846870145678,
 0.5414686672466976,0.16556909113367513,0.5831033278202524,
 0.13394406251044289,0.738033963253286,0.06683591654264509,
 0.17978721194949487,0.26880189766957974,0.43760882118522093,
 0.10264767913060258,0.9550135211272542,0.47746426241083917,
 0.4604306284450991,0.6373748778407972,0.561416977948306,0.5707907950024993,
 0.23786938904831012,0.9678348300005846,0.7417976246206462,0.5211908954378159,
 0.32318025173574694,0.7964120609318172,0.6124413277153221,0.3195988308275005,
 0.5981871115220125,0.607095507829144,0.6782949522014526,0.9131043069739876,
 0.6669942906810087,0.32048597901851616,0.31825199685737526,
 0.6598108180788113,0.31754312395558915,0.8718149626685595,
 0.20345296418546022,0.8498630667198471,0.6297823815890401,0.7800296322534878,
 0.6679989622290459,0.7729237000133734,0.17021414191067452,0.2233149903222471,
 0.6043657920520957,0.334319733352636,0.32915053996055,0.5593117791662029,
 0.4084391140416811,0.8831728282989036,0.7111999211587278,0.8805399404805557,
 0.29819429525783336,0.0028514801458267014,0.342505788596545,
 0.37071524382263643,0.3363597442247447,0.2883751151934334,0.892246615259965,
 0.1420871925445527,0.49148705725884456,0.6508712030809439,0.3641182625132532,
 0.4740329698656879,0.14068164674485384,0.6157320182657883,
 0.07616889831195284,0.6747955190145669,0.8303894781441861,0.8062225413356733,
 0.26622135924444323,0.45103395530763957,0.14634910184128336,
 0.6836733051682194,0.5190289966913,0.44304585320731915,0.916601365323155,
 0.23685195551936422,0.23820171379720279,0.7026792450500453,
 0.1955879563635452,0.5032773982305045,0.8985075210095048,0.5295930370811593,
 0.8516830280188724,0.7402991306475766,0.5055275373825782,0.9529856827663602,
 0.10154355146540761,0.40914505693511394,0.3702099856267509,
 0.2897469748970273,0.9829500722576218,0.2789880829664254,
 0.023728246536974473,0.2633340356679419,0.9213555977816675,0.954781041193526,
 0.6255209141271827,0.15778818325126098,0.7591778004366749,0.221586836521389,
 0.4405257128863067,0.22933132643026533,0.4525529040574394,
 0.03800992936510106,0.021971916204356567,0.6291995336309705,
 0.6931009674314629,0.32724896436509043,0.13368222113143602,
 0.2404045723317627,0.4099884223208169,0.38702733710299164,0.8183216693910254,
 0.7563407986935123,0.422918826377923,0.6031647685589926,0.7897899880126107,
 0.9434356754950688,0.1781715083220209,0.23847913435640034,0.4088643433212523,
 0.8360752109861036,0.2675099915722481,0.46062200332045344,0.9313299264093833,
 0.14292882469154666,0.27690581486181864,0.9243346280771193,
 0.19499305249195487,0.780392113323885,0.002556350027335008,
 0.28189574252484095,0.4523173568028639,0.02199878870454941,
 0.1988131542480348,0.5758271287424267,0.4580468062227945,0.11343003574941146,
 0.44364672742463296,0.273554356957989,0.7692714546819857,0.09096444038532703,
 0.22159386542083337,0.6299541837709373,0.726696871777468,0.31125923743990447,
 0.8393215026618633,0.06961174520126978,0.08729788575899988,
 0.4567449836177061,0.4757176685034542,0.02169389720075654,0.6837491648062299,
 0.4825743939561651,0.2738317009960638,0.7896315632213745,0.6342263174271816,
 0.6096727864698523,0.4485651285673893,0.2872473354302504,0.09035977262625383,
 0.49918162837490854,0.4615345393269379,0.13530781098168165,
 0.13748567104024023,0.714862139113591,0.7914185068525654,0.8382750116437084,
 0.01769818478074392,0.11450201783137426,0.04270066798261629,
 0.5247816580438727,0.6559738246999718,0.3770269008769016,0.15085007292613795,
 0.43154732643776217,0.7602517562746474,0.2806858678980957,0.9679999180876405,
 0.6043531791015606,0.816867968204913,0.35216971434033506,0.41865982959439196,
 0.7973716739754355,0.9169513912398997,0.951832907519235,0.5646256691463171,
 0.9234008577454091,0.9292657029809736,0.15038904033071998,0.1731915850473199,
 0.46832401619716957,0.765087912748544,0.1185182822638069,0.40216459638044766,
 0.0020589151663653604,0.2693513994732196,0.3282549106982692,
 0.6208801392757171,0.7039577342041078,0.6565830758251829,0.49115443085907085,
 0.0875477900606878,0.2815811259755552,0.2768436321089781,0.3693838179038438,
 0.7278958267786094,0.8681485928159147,0.48950489584183676,0.5378279715279607,
 0.05856573708438406,0.18150469749827935,0.0363390859777496,
 0.9887518068309594,0.4645278867715067,0.9357011183427324,0.8238339384884998,
 0.6922237702009497,0.3552258050862376,0.35765857067933327,
 0.38928093713343215,0.38653074173258417,0.4350631476251545,
 0.7963127945291231,0.4993961294350593,0.7521464343261881,0.5717799258613577,
 0.8394466598521356,0.9039093378920291,0.3240665231770309,0.34291070126914025,
 0.9857803862143831,0.6271699905565469,0.8851916990405148,0.8501219275556264,
 0.9367070549214299,0.9659544853905491,0.34951218023863895,
 0.23465381977209565,0.21365174248037722,0.32059673810842493,
 0.9132071092597538,0.72741565710396,0.022674225819147686,0.08879123200656869,
 0.23752324799589886,0.2972239347085879,0.3776037354640942,
 0.11925537921520435,0.8827418888892214,0.5971821794134019,
 0.06158082805114484,0.9054987810873283,0.0070717933375156194,
 0.2795349676016502,0.21592624904268598,0.17583019482665196,
 0.9168416237372364,0.0912452938970022,0.8529841513862249,0.38102239964725937,
 0.9403578012029093,0.578582040385097,0.1732753137998182,0.20305010165917392,
 0.7370584601590328,0.4338995935213087,0.9942017318436402,0.7349357967473698,
 0.30164468217336615,0.49178854701061336,0.7647584526877703,
 0.5016205310083808,0.4496511054893312,0.9593765012406514,0.2245920206829619,
 0.2534931181001413,0.5531444836953306,0.9964921455065421,0.7690740507867809,
 0.7390414729882794,0.46916356700310247,0.12697745950636286,
 0.5153303687400685,0.03674307464896176,0.7777736638295191,0.9285115750018325,
 0.7976184105663991,0.10974052881999596,0.25120164586661553,
 0.7776579409192431,0.4109342527807076,0.7580895069133667,0.7497446653545815,
 0.4982276917994144,0.015295276566680949,0.30112894088905784,
 0.005552763718315368,0.8999419576472616,0.23763675510148963,
 0.5037077697441164,0.1768314999364038,0.03813452737349421,0.6371680977122797,
 0.847997762531908,0.4336828014384322,0.9968700180672859,0.7592447666407877,
 0.4403089885939421,0.48335231654456934,0.04689561190445091,0.796442004418785,
 0.38604643325968724,0.4850276109197138,0.46569654194586285,0.865174571783867,
 0.1598617613831692,0.6206454536115718,0.048905801239565116,
 0.9145160191898016,0.6039846204260886,0.6403752632791566,0.47567646742938774,
 0.9404469417229789,0.7660258863557173,0.21530510568899164,
 0.05014083828252036,0.3914183502243995,0.526476553403601,0.3622762181693312,
 0.5866256966489124,0.9126450819154747,0.8813001535979348,0.33021435317918224,
 0.9478234688082661,0.995365437978687,0.02104932111299851,0.46666969062954666,
 0.9808018848512914,0.5483203228217112,0.8343144487783127,0.6154334777774361,
 0.27974931002499204,0.3861795019401909,0.3259784186523884,0.8855928791865658,
 0.24797238146703338,0.6505828577148243,0.44928048945250365,
 0.4114718886905784,0.6251631455003569,0.7632023961967399,0.500945790021326,
 0.033004902502558386,0.7013461285190038,0.19026883021814078,
 0.6617796368304196,0.24572888695830286,0.8345217653810052,0.9225882636146017,
 0.7402116598233752,0.49996755717233077,0.5726550078373287,
 0.015399618559210904,0.405857859179958,0.8779970614036712,0.3271630957095396,
 0.145597669978909,0.8687476323762953,0.5226001606923821,0.31179618469742876,
 0.6197483663946446,0.09313448953098125,0.5422930951526518,
 0.01984246428860681,0.2873231414779148,0.7513685786934914,
 0.05977346416390927,0.8772584794412212,0.5046381102476932,
 0.37502864818402604,0.35923028646894606,0.5991240919778499,
 0.3653864424149751,0.5878887604898893,0.4090664336164618,0.4107005889583166,
 0.6254102763941045,0.09679305086775092,0.9850191783391198,0.1658139833646548,
 0.6984462913887853,0.72664908965076,0.9193421164702502,0.6109114990443877,
 0.38629212068260166,0.9465011382496503,0.7485296643832741,0.6051313114155439,
 0.13291997383405074,0.8837118426508744,0.9427248359873436,0.8203482576498725,
 0.7515317581590767,0.5234102098483803,0.3679598096686604,0.657479711138084,
 0.7040756259707105,0.8882224183123637,0.047943418644307845,
 0.15426557402271512,0.4009137300074529,0.3951627165814888,0.8179475376228373,
 0.3672692466786158,0.10521135358956579,0.6283158349384694,0.682950249316248,
 0.7559496763792974,0.7945065597437545,0.8463589264779315,0.12440001984254068,
 0.6879393504312691,0.9037896101106145,0.9901202516089863,0.9593040409981248,
 0.1032284092818494,0.22013968997222066,0.008757137713445773,
 0.49699374349464465,0.14523395072098622,0.7586857901803832,
 0.5012152933419193,0.9087678567178279,0.8390604546907166,0.06047596885563966,
 0.6793370704334513,0.11581765690865065,0.7548558484471759,0.754614187729684,
 0.6395185528167094,0.45184920975892084,0.008553639001087632,
 0.15527150934222422,0.32887386893798565,0.05481890520857513,
 0.9163788457886819,0.9149571478439749,0.762498667800703,0.7837820729181219,
 0.48316242595241954,0.07477901882297278,0.6936006097277674,
 0.4256673164982654,0.8382422603721126,0.6851371932819347,0.6724299682187181,
 0.1672039287618946,0.036468571069638855,0.28466855586173234,
 0.6756368808981337,0.46478996989063304,0.5329023222626387,
 0.13092867490464344,0.5525682381813166,0.7666020837827086,
 0.49904879968522364,0.8764988258906683,0.5426240215808316,0.4830736400358455,
 0.6143237588498106,0.021696550929673486,0.715933717262564,0.7359082082868131,
 0.5038535128835495,0.029586926039519623,0.049219230279558435,
 0.11352063883293995,0.12287226343104185,0.8155118581750989,
 0.5628743393443139,0.048132196535489946,0.41493931512849014,
 0.27161887165529275,0.44657972327999473,0.7005867263279494,
 0.8274094986812006,0.712246998903822,0.3712344443223812,0.5164704961774643,
 0.10596929266517197,0.5054745123927714,0.875304288046363,0.59314196722495,
 0.5974770878844124,0.07376198898361397,0.2120026989404633,
 0.39007431220314803,0.2965524681026138,0.6413309355553555,0.8300832152996062,
 0.283357107859056,0.8026501842990295,0.012954643452108172,0.2782953830565169,
 0.9957355515204618,0.8677411882962437,0.9226854160886152,0.9119183532950612,
 0.5698091533344386,0.584185392571321,0.8448432364796894,0.44107102431516476,
 0.03921287795252759,0.2724850383202708,0.9376284197036141,0.2828496105079601,
 0.9529606031966319,0.12935958866123154,0.22645145997522265,
 0.8246272134313177,0.1599335322065516,0.3970868082781098,0.2642904618524822,
 0.4260296456409284,0.9744547913934489,0.19767265638304254,0.3294040435476825,
 0.38771702107734685,0.6503913491833816,0.918383699602102,0.9393774172424681,
 0.21191408281709834,0.5573274285948413,0.38139188066741736,
 0.7719373778631732,0.1944843715594654,0.6330760459897441,0.6460052335281706,
 0.8986275845285924,0.3078059367553959,0.9345107068055641,0.36997026477457184,
 0.2839546992582962,0.79556000954428,0.302257229009141,0.9250349339802268,
 0.32461221845403876,0.7613347049263798,0.05860046047530376,
 0.15676949625110792,0.49268076574399133,0.19519934350194945,
 0.24042999132906606,0.384966274032009,0.25461821363605586,0.1325476741520083,
 0.9621783913928902,0.49294073252817316,0.7549837928464679,
 0.44975980137971805,0.14450917325442847,0.6135075059391569,
 0.7972912807763842,0.6592082279503679,0.2407357434662073,0.8606487933468314,
 0.6050496876645115,0.7357311865896706,0.20970571079136535,0.9576119318059519,
 0.8544137746071364,0.42726781961898863,0.3807077888322097,0.6625968111266259,
 0.26037436278490433,0.8074784798115822,0.0965790364721848,0.8121246371863935,
 0.26038926623746705,0.5097324295591662,0.34711110568366055,
 0.9095394605758007,0.04237633552889242,0.07024182895422759,
 0.5271476912527927,0.9435868651158443,0.49536104519941127,0.3820716562150267,
 0.9211191269639832,0.12646189653392526,0.5472597493543925,0.7673322076731133,
 0.17315356067291243,0.6118023287290364,0.021260357769276705,
 0.6363970197690362,0.34026069436075557,0.8657338179170702,
 0.10919244670230999,0.2053893635487436,0.0030096642406143292,
 0.3084143479738638,0.1775330043086334,0.8319096228983516,0.8566595989824327,
 0.7047317436172917,0.9916055938920875,0.3981809886374773,0.9012087641881679,
 0.0666982438153838,0.744630562901363,0.47439813825892196,0.32943210781202537,
 0.13119190325874885,0.03394947388676639,0.09754455702870546,
 0.41603168533538504,0.9718992916094389,0.582063437773325,0.7382837339177113,
 0.2853191495773333,0.8394759360418571,0.27726904339576586,0.945264021311443,
 0.20278062444135525,0.07870025816746373,0.7765089855283304,
 0.2141498485853891,0.08499611226482007,0.7600088343208201,0.5796203534243969,
 0.7400887523803386,0.23019113400618607,0.08436887408270821,
 0.9535720095592528,0.6783703727133099,0.4474193055158401,0.22931434724845068,
 0.13473395315387093,0.3115646935713511,0.7427964932783437,0.8557887830977117,
 0.10571115518631169,0.7730941004723627,0.10069011119147375,
 0.18922322640484868,0.337460331312777,0.1982118026382672,0.5668014170772558,
 0.7023307811820696,0.10706149666667664,0.8236078251663207,0.315562620940143,
 0.6462193547577283,0.7731042623150708,0.24375136737072978,0.9650046819450548,
 0.9743733155819194,0.902747750290512,0.6526404169350459,0.6641129476158166,
 0.7654776647468249,0.01416772751257056,0.8708051102115469,
 0.49327925822342644,0.05866463335305738,0.011931695697570577,
 0.7103928052194444,0.7003605932392806,0.1079574623227707,0.3052370577383362,
 0.6832166080580018,0.7269281627973698,0.42554637691616914,
 0.24842490794939798,0.08722137738660551,0.3701362843991677,
 0.2991297005783158,0.0024421246457417833,0.23297648600883725,
 0.6654480391937381,0.03176564026975337,0.12426955712719323,
 0.5711228614760263,0.704920010643624,0.598750372762754,0.48128765143357544,
 0.2502432258005801,0.07274842502714407,0.14626622044311788,
 0.41549073733188424,0.286308601669245,0.1541331305876943,0.16505661550918005,
 0.09501349613448329,0.3068229006121941,0.35205022373427486,
 0.15198953576266339,0.23807488940811383,0.709138737525827,
 0.07301250023106465,0.5006765989629196,0.7280094118532892,0.5817313303567965,
 0.5795864906447383,0.9854553444327089,0.41869303582404194,0.7841830808502844,
 0.35440460824986475,0.9888940387300393,0.041022250907243984,
 0.8097168950238901,0.6770853553118685,0.42055943491287406,
 0.07482068340109715,0.02847312934897006,0.9278652256329838,
 0.7024933496343808,0.5610026857500631,0.27464667110504637,
 0.47422949826601424,0.2281890025411304,0.5538289688545653,
 0.35319140836659824,0.3311684252278232,0.436495670869923,0.15519662376857468,
 0.1613807587575874,0.32372007657706514,0.8146892579265334,0.8717994981069543,
 0.5536744677868858,0.5767678794755985,0.6231031769582983,0.4455296536889236,
 0.8311865651696213,0.6650619510012424,0.6986363424039759,0.9652609494532675,
 0.12139168467897532,0.9668979716806992,0.7679448732652909,
 0.21560893464947872,0.4752881295519261,0.511536991686574,0.8483795695395324,
 0.7688168501862638,0.7149101354771455,0.9833321200114443,0.7442056915164964,
 0.9477717124561992,0.27133804575872655,0.8963858158655776,0.9754352192207362,
 0.019433777583308443,0.20357890411139012,0.32476727653451276,
 0.22824224267065196,0.8188749673863344,0.27313619050499494,
 0.9235281238203131,0.13525894523082027,0.5841122344517587,0.3009382814452173,
 0.04475778928792207,0.005341479276377514,0.2860353545185004,
 0.2287920752389324,0.07287707584169345,0.03054180702128595,
 0.37057111327482684,0.48557021212944684,0.5821995292744186,
 0.44475963251754114,0.015500655617610382,0.8350489813682592,
 0.4145955514600266,0.3270001089234428,0.6941461776930865,0.9592650811512424,
 0.3809365624755343,0.9827575737023003,0.7111877915933235,0.31262504416620573,
 0.2614352199120147,0.16187965140852534,0.09410959896357007,
 0.8474703107607692,0.6831344493696198,0.5073736576707378,0.5123964414667934,
 0.8363517831910453,0.7856332712846794,0.5052162465850011,0.8532752706288502,
 0.2218755355481684,0.3825976568928602,0.15654276652119647,0.3781329165601497,
 0.6162174953124341,0.6183097805718092,0.6605907549430802,0.7589636248137196,
 0.6042239388731456,0.24886966293245527,0.49777317639541796,
 0.007242678980245887,0.5968011862761172,0.059326993778142545,
 0.40539168973510764,0.7193402211004991,0.3887941810924873,0.4374589960986248,
 0.6389681259641916,0.5946533460876533,0.6959923006152053,0.6332302799052119,
 0.8162122931057597,0.014321138294138391,0.5438571394930248,0.761744231368271,
 0.5720625903210952,0.29958729660689287,0.9719873487912952,
 0.44658693200714805,0.08692489099774026,0.4983641204724081,
 0.7687682413133095,0.9628970506905719,0.8760506044712012,0.8028298799023178,
 0.8649287590881058,0.6259045437976926,0.6496515163875392,0.9170700849253626,
 0.5164654149058622,0.06848934495626358,0.7982358943961805,0.331342085520263,
 0.8716720405601813,0.34006971924695173,0.6463131821952339,
 0.006961297176058978,0.9420194802724499,0.9158691935524182,
 0.6605657731169432,0.040950978273990635,0.7391150856429366,0.708733334589383,
 0.8330303690245067,0.7533721573877536,0.8152572979306215,0.2797409896501424,
 0.572805755379359,0.03542174462145331,0.39398031356606766,0.4035697665624939,
 0.31798234073293463,0.12600181549082679,0.3609376772673285,
 0.6529758191416565,0.2272523386108921,0.8322039977219184,0.3559304947986962,
 0.26172647289253204,0.27008671831823095,0.882549098178099,0.3246276612891309,
 0.4098116677796274,0.9020886533517648,0.38377392449634185,
 0.09299710382270965,0.8306474800896202,0.4752190170884778,0.9407539764786935,
 0.008911770987208634,0.13845452110112855,0.5805154767427374,
 0.01966186301193007,0.44182738975399494,0.5258327731730839,
 0.07800163863141196,0.7751525704466062,0.2973249930088362,0.1432564441864833,
 0.3645995923991312,0.5199714853938044,0.4510002008984074,0.3798952558368992,
 0.44644206309202294,0.9925916748169322,0.43269585589992043,
 0.5366078187972848,0.5982738401700667,0.8268126442179311,0.5796021759099487,
 0.3605378513551907,0.37785889281025087,0.687054668585723,0.5912479624263249,
 0.48084850841681603,0.584132729775032,0.8275328994536297,0.9335940438689092,
 0.4055706137632674,0.5372634021692739,0.03901777492346148,0.5785345322720774,
 0.38440042264579377,0.21468067388687162,0.35860791397902714,
 0.5014271096128912,0.9021290324141,0.4566026309854928,0.31249047783869555,
 0.9790257586151169,0.7013206522278149,0.5177795083981274,0.3830507253429286,
 0.4841265355557062,0.5147697963358614,0.7770995247099605,0.2684811730072276,
 0.38075285065782205,0.5127899317353602,0.672642581473297,0.7285983476781199,
 0.2664272663580506,0.5155857652992977,0.538484156561506,0.7836585173845717,
 0.8598910627479912,0.49693201045837543,0.3523769944410573,0.5542920441199467,
 0.7249941230989662,0.20633885607086555,0.41393232323336515,
 0.19246608850327895,0.038503546523678156,0.3891703750808989,
 0.8646998732201355,0.6842943842932048,0.14458952427350735,0.3948675353162533,
 0.3747903525075603,0.5286258873678323,0.024369703366894813,
 0.5126126997202142,0.627360094012685,0.018052190932797085,0.5771188077496961,
 0.5680911872016292,0.11492095668022184,0.1684984589485825,
 0.028924169197034777,0.8678592402109371,0.08697240351256885,
 0.9575275608023163,0.3842280029456816,0.4406751984431946,0.612991029317355,
 0.6057716005907597,0.6833592137474114,0.039064827709563654,
 0.36817939022604784,0.049753634315251594,0.08899462697685279,
 0.2583306518512296,0.9216005697461507,0.05441289902006041,0.9840632506838514,
 0.5820563060390083,0.2725234017639281,0.5925814412653866,0.6270252321591193,
 0.023865309583495975,0.2965509090068845,0.825321356735156,
 0.17423879198964531,0.9131270007812042,0.21152224736422431,
 0.6574409161188508,0.9417912839262409,0.1551300943115732,0.10168110463312952,
 0.12510882426258652,0.5025063978372346,0.7293415642183434,0.1008975599001688,
 0.25860268078266435,0.4044681281683139,0.6793771756426464,0.8872591109438854,
 0.8289784095209772,0.28257167106639014,0.6784170094711793,0.3304906381810109,
 0.8133798390545282,0.9048020747310613,0.21902113420476477,0.4495305266709164,
 0.5273351285144682,0.15705285417478632,0.301361767178419,0.2720776241506454,
 0.9740520618268386,0.9923450296963038,0.9243845719166702,0.8686051197861681,
 0.8906553437806823,0.1413290633930674,0.168201054975461,0.3637814580163836,
 0.031129862906388417,0.2727463048144183,0.21609055253724763,
 0.13990453677604497,0.7197452479609633,0.7428037094312707,
 0.15348680253148073,0.07655893216088838,0.039743408020874504,
 0.8951910023422173,0.7318582528266133,0.15203677563936957,0.8295587714175692,
 0.18968589265753277,0.23491031714385457,0.5585258569945437,
 0.3940724783396863,0.3598047698649187,0.3668959305876236,0.8532432099496466,
 0.4515591125349412,0.6518703472423825,0.7752466744156855,0.9903539140501005,
 0.4179658681178753,0.43725546339840315,0.9078740747259212,0.3368152412756641,
 0.6256689129618579,0.4774875563190356,0.9483756118754549,0.12569943523822347,
 0.1257957009384546,0.9877600156141882,0.7350982874564115,0.0899036912885931,
 0.6607148040895335,0.29296438735251396,0.20277415600591508,
 0.5493522378957614,0.9177060608027571,0.5471785582804973,
 0.001000892440881307,0.6561556794645846,0.14820417497328287,
 0.06177296627984319,0.8640309980234114,0.7779916187300093,
 0.13132697714985797,0.718072958294439,0.6184979324916989,0.11580753550119083,
 0.6111689066141588,0.02425633723405307,0.8230434698060555,
 0.29392187788038426,0.3112678067369433,0.49481700419274677,
 0.5373178035335582,0.8579386555415569,0.8491392922090929,0.19577884068012497,
 0.9629213548763743,0.654336119814911,0.657990731728021,0.7067249463491556,
 0.9283453841323506,0.013020683523958465,0.41137497785407406,
 0.474401508879182,0.683335041900549,0.6109828225255809,0.26304914260201484,
 0.08816681155274841,0.6738837715219592,0.36442930130579576,
 0.3861277256821595,0.3890585553303877,0.5939349052096401,0.2181572814544792,
 0.6748822357603731,0.9811603235243058,0.05206336674202172,
 0.17753082253620145,0.7617231872202195,0.2903685421879486,
 0.19803154881346652,0.3849172713842357,0.09415303346247694,
 0.9121273172732498,0.7431313860601444,0.9892641695111981,0.7628594465843292,
 0.10008269027095651,0.7213803921825178,0.18903687673403724,
 0.10123340651403179,0.473907453235877,0.25896718442512545,0.3396545983237155,
 0.5541016921501646,0.14901246327920314,0.6100523829883319,
 0.20897189237629887,0.3588900568802834,0.19250713628045402,
 0.029957325979646066,0.4192541595163617,0.11109600214016868,
 0.09487770744342294,0.7715334924839237,0.6961796410388178,0.8234109688365043,
 0.29042094805762353,0.13901549727536666,0.053757880418318305,
 0.8795548150025257,0.3681836221686434,0.1836008225991772,0.701648315239225,
 0.8085324759725733,0.6958644266737437,0.7312989083149849,0.14601637539172674,
 0.8736180037652546,0.5430569926106309,0.7392857705779711,0.13723618088326528,
 0.164929558404722,0.0025709940132996323,0.11793468228425419,
 0.6045136344417414,0.026060867847832903,0.8941759609443166,
 0.7857259696343951,0.7332761512521981,0.6767207868566953,0.2722156517239255,
 0.939269414730086,0.8087798752996331,0.6961048313223397,0.012057445216959728,
 0.46899191885285285,0.025000877406530786,0.8680042199193219,
 0.197654584517565,0.8113817249109705,0.11396827999210457,0.2110300533392706,
 0.9151062656519025,0.2716978623987749,0.8380622051335687,0.26939631147491105,
 0.5570831208782674,0.019978299251560694,0.025450256175964303,
 0.33031297414964855,0.9637575557663851,0.8833868117035909,
 0.09943500755090173,0.7072117010992768,0.27603648513318546,
 0.6914364408183841,0.647175256458626,0.4831929786003343,0.23837809070379,
 0.3476845342195265,0.7424820399987291,0.21359890840010798,0.5140657392711366,
 0.10620551420374691,0.7308595626561654,0.734108001430289,
 0.006705908684099593,0.29742618423708644,0.7608392921655819,
 0.8929823514388242,0.7883918278324905,0.7750225481107454,0.8236384137384627,
 0.9038845136761795,0.3214803653789873,0.9143332150713888,0.5100218430209891,
 0.42193439540789657,0.9788801947261407,0.3412810204390061,0.9820280407757721,
 0.679890762704857,0.752903619458868,0.06535004645387277,0.9108919416801232,
 0.07797043633804834,0.6775935990696962,0.40829168463768206,
 0.34303313447822903,0.15795195169778375,0.7127227195367357,
 0.7066421424905577,0.8498619173860258,0.5066206109133848,0.936634539317507,
 0.9395694773268034,0.47022069186365845,0.5974860868008819,
 0.23534385130864477,0.951138584462864,0.17120823257460183,0.763765407892517,
 0.44743190667118204,0.5639487190389417,0.630175656477827,0.1265022201993019,
 0.7715444160487822,0.07916420480580777,0.533251827158884,0.4728643322926257,
 0.8409426232177498,0.8270091025774812,0.8440254064011021,0.6967751627540933,
 0.22066456012338564,0.8097570823062352,0.6738955578432132,0.6699802873790905,
 0.6770972593001685,0.05446825347385509,0.7868552431397862,0.6157626593315421,
 0.4026377082602203,0.24125924992855197,0.04800104002368766,
 0.43150857704400214,0.5750810767245307,0.6070204800196908,
 0.14776293317591982,0.560802657690811,0.7194627991762548,0.13478103246866546,
 0.9657761567792557,0.8779897207112968,0.702378913418177,0.7304659572475539,
 0.29373507079442285,0.1534037292484477,0.07411573590757725,
 0.14981792222803547,0.4998865888031747,0.2874535355750518,
 0.41204825473448836,0.35016053299481453,0.2716604817802283,
 0.1443899061628785,0.32155840583407436,0.24990344689339705,
 0.29011608526730825,0.82800441023518,0.9888439294029832,0.0706312007044454,
 0.058508410950518064,0.6509314413948162,0.9184910975994641,
 0.6430329630880472,0.24563702303059265,0.08701260679376133,
 0.9396866881131976,0.33539433650875494,0.7824029079945066,
 0.014374422044323487,0.6181483444990696,0.7363074636462836,
 0.9335097357358318,0.5779957340192755,0.4602604354344767,0.24796278087759915,
 0.2196218971536834,0.709660939397472,0.599084232685224,0.3242628198850843,
 0.9941534230714467,0.17681981507030453,0.04283325693128115,
 0.6826336051037093,0.0822478574062222,0.41086720056895076,
 0.16398190753085018,0.21703235551952071,0.8108376392801773,
 0.24873401522877547,0.965206809069846,0.9404661813257494,0.9139675780808664,
 0.09563139204758597,0.04934762374910351,0.7910364247576105,
 0.33644897507996296,0.5834646828143879,0.1574718337601939,
 0.23181827556208123,0.6647393526026475,0.18837729723678265,
 0.7086773297862611,0.36271935562338686,0.46635679554936915,
 0.5440716622694871,0.4690330636095935,0.4743035601138895,0.04875513852862179,
 0.2702067303934482,0.3916322714075421,0.350894037328489,0.8361125698514775,
 0.23631929569669774,0.23105786274020879,0.8598280831477567,
 0.0980100257441534,0.7302670107888355,0.2619970391711466,0.06573517156643138,
 0.4428183488490869,0.6055208220123506,0.8433890213848692,0.6133437299654644,
 0.3382381720778037,0.3843074953604967,0.4721095677374034,0.6665823896980765,
 0.3992776520860515,0.6214901748618131,0.07472314962566262,0.5068131674628751,
 0.5685387048761362,0.30219483934116276,0.9772040429508814,0.1617800337950137,
 0.6226125862268962,0.7263670739144306,0.4017152484165981,0.845314245025103,
 0.5300787012923118,0.30387562431723203,0.0798965533013899,0.1041078716076469,
 0.9067714305925159,0.14425762997106784,0.9594131655379804,0.4916221706863627,
 0.3568524738621013,0.028993294223464305,0.06514914457197574,
 0.2502389345460031,0.5036344791152553,0.07333447031278895,
 0.37536256264151635,0.904127958104124,0.725339223380415,0.3763349791241618,
 0.4677851079841455,0.6560630282848858,0.046322836406055945,
 0.9582318828280495,0.36336347908279254,0.9021189719010756,0.9162745485448128,
 0.7935078157820711,0.4092564013818967,0.7961701067906828,0.888045016683226,
 0.7109025240313825,0.07427532790846902,0.4651092508515766,0.7898682176389852,
 0.9801650373270943,0.40601769160850054,0.25757697885083064,
 0.3019078967098029,0.48209484872111474,0.19665456769785472,
 0.33805074840636506,0.9485528625530601,0.022953319962026764,
 0.6289650938066785,0.7247420630023212,0.5238753669602091,0.9412961934648735,
 0.835057130811371,0.5287468058747956,0.273026288103587,0.856005643111375,
 0.2207528851884034,0.20212166505249574,0.6508362263240369,0.3035386831589343,
 0.1603313726433987,0.8243305091145257,0.10640632360000357,0.7728494081899455,
 0.22761989508538882,0.5524802896060326,0.66596994689911,0.7888191783500615,
 0.8465901003518935,0.37429230497005905,0.2829793435044382,0.1589179024359455,
 0.3796031102276092,0.6402193796166903,0.807756843407902,0.5422713516824779,
 0.4492199322334618,0.23169519166567198,0.875239501475094,0.42277578082173917,
 0.8508340867092804,0.7923147065004487,0.8323259450029518,0.2889833185240992,
 0.5440200555813769,0.558667243193626,0.012953153654204685,0.7909520429544609,
 0.9772823972002008,0.16577720202587554,0.8846159502544491,
 0.06451419087462296,0.5297267462833367,0.396430464783743,0.7800956978800878,
 0.9485466488121879,0.25422188961020076,0.25247146539348486,
 0.04651343826619281,0.3855508110385859,0.10333753513195398,
 0.7003722192992744,0.8419760894371263,0.7579230106093341,0.7279864305927543,
 0.5625527047899294,0.07762395825636881,0.4103561333538762,0.6302429094980034,
 0.8301825858753688,0.1066273807004281,0.27460496441162985,
 0.11532705319470726,0.04316000705496903,0.051190910170242354,
 0.26569101671846074,0.6396738706696027,0.047602706072238266,
 0.24231408848014768,0.8110350257750376,0.24452318828013808,
 0.7493748042218323,0.43441950319990275,0.92038142317792,0.9453513429199752,
 0.15862566409887324,0.41259736395960644,0.011732213249905699,
 0.4764349354567673,0.28648597031789047,0.4067751430310822,
 0.21851804786318896,0.7070617390394984,0.20020815944563086,
 0.34598980774389687,0.4424576426428074,0.9754694061947737,0.6801522274887879,
 0.5317313758666118,0.19984997867495735,0.5647903690746081,0.4978451196946734,
 0.4314387745828787,0.5338254892401342,0.7965669707468925,0.8662041004154019,
 0.5381324830202587,0.7852152523591914,0.34518805278051246,0.8500816372851357,
 0.5567861154133236,0.8488105460490256,0.8048926181753338,0.8616970442755835,
 0.9812016358296525,0.7447031148604794,0.007511574165420476,
 0.8241342701419845,0.968507219517708,0.28473160070803505,0.4111912605427793,
 0.24401969211522834,0.22982151476409984,0.25244807372515476,
 0.5240621837610325,0.37874517673014785,0.3375040388102165,0.9903784713788457,
 0.4054737998265264,0.2638483980458235,0.5495876687231247,0.9296837034494017,
 0.37580454375291994,0.8396824504917174,0.04257333188050183,
 0.34863829382776224,0.24246679347477895,0.7729569230229992,
 0.48416834253544216,0.374398457147735,0.14694741946716983,0.6272328812591236,
 0.7982372718050973,0.7865090680329926,0.8451108985732202,0.5522612230825825,
 0.49014749960862747,0.03311209424978245,0.991383368650441,0.9399673671796169,
 0.7630393281029217,0.06275205469023815,0.3142012959123819,0.3091742734109447,
 0.2995268551957022,0.03874418255998746,0.2032611737273392,0.6950867225210915,
 0.3128044323165202,0.7703552205758465,0.19505502514740036,0.8950643699924801,
 0.31603625953376735,0.13995534808428556,0.39397013647224377,
 0.2874241918618964,0.5497684626958955,0.746004364493068,0.4905383543510077,
 0.16196341412633075,0.503179452062634,0.9607264354204389,0.7520855031310875,
 0.9902424365110659,0.9915645533560933,0.5323102332915297,0.09943707841963656,
 0.3628578253429878,0.1914823461871411,0.8190038739807063,0.9156337437621568,
 0.6130264983406352,0.9754896139502497,0.6375467427080521,0.2728974144899561,
 0.05223305246386756,0.2796660223362277,0.6920818892236901,
 0.30502440255832586,0.22320477492774238,0.8067288201191698,
 0.5363933608238659,0.9158213762214987,0.5244911455615802,0.93192628610007,
 0.975209120652562,0.6078099403482647,0.41071907525781126,0.8469373809196015,
 0.1291105126513048,0.3777942358249342,0.6313013832204993,0.4285997709195297,
 0.25684127556218206,0.07646506822544363,0.9302399515739379,
 0.8465952038030723,0.7459037974749051,0.7100351476055966,0.7631341076925611,
 0.6010112404161263,0.7773997862254564,0.23511874390481924,0.6144242295833612,
 0.5332856326046514,0.8979568225452782,0.7255723664903999,0.2352764328147422,
 0.9407759652139778,0.17706462821827118,0.08407003548643743,
 0.48362416432192623,0.35105996725508315,0.7545050056992477,
 0.8469250762918369,0.5548813664914166,0.023902235626070967,
 0.4496689308880497,0.1063429661266212,0.47190070404760953,0.7997573469849224,
 0.342161097655727,0.16555906263158138,0.543265937044128,0.016641377857464,
 0.7336299645020943,0.3638974348059446,0.1417760602450372,0.7282740357026082,
 0.24558068941584965,0.13766095621700336,0.8931817882358263,
 0.11851633124960548,0.9877136606901136,0.1248593158036777,0.6198273744234521,
 0.895617121370093,0.44706381310633203,0.7377042910514167,0.2566787468351326,
 0.5807716440315996,0.3454169243508818,0.22084956706993708,
 0.32442089715277556,0.10432222491224263,0.10611383284612586,
 0.2732174431159209,0.07958164908609822,0.16442487946802697,
 0.1709044177849608,0.837908465599504,0.18760295351619938,0.7257777114974971,
 0.45499737016266806,0.31101021357666825,0.9465186273902455,
 0.10401836359646832,0.5159132300269645,0.16180516735426964,
 0.040321124070604064,0.45124871757675356,0.4900084422110259,
 0.36862907364471376,0.7430280806819711,0.7599612577348018,0.3495810435310164,
 0.5563092950322969,0.49917133357713117,0.8274932149692458,
 0.045504516625350155,0.42606561521198283,0.296240670413068,
 0.21581119466063092,0.9459258953809907,0.7334180221688982,0.9338239019676388,
 0.39562217234536945,0.5975881288553221,0.4020715209704735,0.4885875718216015,
 0.8063517125042723,0.5744946601640232,0.6217289474626477,0.46353694589634475,
 0.1881778839984155,0.6818327129703202,0.23053684342574288,
 0.07001894562388955,0.990794355693121,0.054578169119705144,
 0.1494800717558118,0.7897699156294272,0.7149420464598685,0.9150422653847531,
 0.7650833806973003,0.2925925999266821,0.8743585525625169,0.34091633561269274,
 0.029690697919036668,0.5793663738392122,0.6999843906361545,
 0.7861704429580187,0.4269376400420629,0.9218835186790288,0.5664534044977869,
 0.00427062794257449,0.09581415551643946,0.6589086447967342,
 0.8831961752202259,0.31555365881088404,0.052882274517326855,
 0.8891737944923268,0.6373000768011012,0.9555982231399638,0.9188884372703422,
 0.7878043645174839,0.5634106957581868,0.6876480298718218,0.2736389329602342,
 0.2704864051468331,0.6365326280004884,0.9455597784047524,0.12189650618609293,
 0.8470768703987281,0.050437656424067834,0.9913190506207513,
 0.8529054785262248,0.6665032771666886,0.6469837817418426,0.44642873039749964,
 0.20505495026173848,0.948377142793428,0.4780751769922781,0.4015929183625073,
 0.42354769114637436,0.8396027497424167,0.9458507455904742,
 0.23589261159095531,0.32774044249726453,0.7109305678216558,
 0.8732549225093016,0.33647293486882845,0.6303570389659048,0.9956387882186691,
 0.79941816219241,0.4851215614903409,0.5126753042002582,0.8392098060927532,
 0.5893407803966468,0.2975253111151729,0.6222711722764706,0.16279072926736549,
 0.9944817406897773,0.7584302487896124,0.9136234154251647,0.7819917155080426,
 0.8326904885965559,0.5421918930261463,0.4803640157182324,0.6155937832984706,
 0.5214367103423707,0.49882982218124305,0.9400130227980023,0.662304030311219,
 0.018181573190527156,0.5064088003853441,0.647371500202702,0.2852439887178845,
 0.6072174338382303,0.7163147317336762,0.8833854509756085,0.6678358864600769,
 0.9713011960869692,0.8914283184374938,0.32953525000305917,
 0.05062740280291911,0.7892484930922109,0.8751632014070428,
 0.13475438149994745,0.09827997929764531,0.6024561481580235,
 0.1548806304533792,0.6475262167957744,0.5512949252818746,0.7175574242442213,
 0.15940765823667602,0.6076493262585644,0.008263425581376271,
 0.584744574694584,0.676086354182588,0.46638832163031285,0.9418467834819424,
 0.2739806989023248,0.07554560595139459,0.14528442418371135,
 0.11862665248057103,0.954282291996568,0.48292781398314133,0.4367964775466734,
 0.269012860402118,0.7419430984089057,0.9087897198013256,0.03820755770357409,
 0.0031048617394855427,0.42671144893730273,0.043430292583349295,
 0.608197071197881,0.865762059122018,0.932570865461361,0.34831614904362385,
 0.259580978956796,0.20577963374114927,0.27674334009239465,
 0.17377654162079492,0.8470339125785886,0.5488929535658987,
 0.06374432135236763,0.6026118244484732,0.5132062968992177,0.2754693577822367,
 0.25723392874086193,0.5630106553262575,0.3769194838988936,0.7519580706669193,
 0.0345589619747253,0.8405334195092556,0.5122618084021822,0.26565275539668554,
 0.5822783422094102,0.12597567262049358,0.20289271377956886,
 0.09192438397578062,0.868989004808961,0.8123446669631019,
 0.030696453144810842,0.5972823636190082,0.04747950054262884,
 0.19520548485124567,0.03722156565393131,0.9517810706743294,
 0.25008447729101346,0.28244714941283333,0.8974770692027157,
 0.7656236668435876,0.6285567204859208,0.4358430760992974,
 0.040671534179711255,0.9971407002990924,0.09972959177185159,
 0.9415895232445401,0.024067595309228307,0.25466007525848267,
 0.578727940479138,0.6347552346787552,0.9578222272060595,0.7740406109262706,
 0.7810043216375396,0.6604041103644828,0.36639260220902325,0.5034023454278684,
 0.6199253517795139,0.771642920639714,0.1738696439334021,0.8145683779408865,
 0.8315311474799175,0.670231719023814,0.35659385003320165,0.6288056571682853,
 0.8354224920883442,0.12487344830127745,0.30264400567039107,
 0.8204806163751797,0.4322720032417898,0.21431842474536644,0.6302024186997807,
 0.6222017000287746,0.520576297556522,0.38919923595492956,0.11841100014924943,
 0.4647460478424573,0.718675564710824,0.19395767790074303,0.8216218334655053,
 0.9305963999362962,0.3438717552710656,0.6820765558237948,0.9232617822686614,
 0.4044394144551553,0.37619568496127054,0.9940031952810766,0.3608451559631839,
 0.9206359666166094,0.40825760510646825,0.07400719731622507,
 0.7600495131670333,0.3832374650752306,0.5125699717658783,0.2670671697968914,
 0.5487249795396134,0.49382316040786634,0.9863336875810751,0.8194991002938773,
 0.3367254480108183,0.9474582541683333,0.34901999114705196,0.7720988681697034,
 0.3277925013849029,0.3792836843257732,0.28500791485860333,
 0.018199064853470337,0.4504594687331278,0.7299900454816937,
 0.5042395292890725,0.828563367132701,0.31142428993154203,0.37711422521697513,
 0.08898776866215796,0.2146388894952992,0.1240966216713969,0.9517429759814715,
 0.11323970181267196,0.11253250928029146,0.009151605641990801,
 0.7409512451259934,0.3612161739647167,0.506169289838669,0.13018106966574072,
 0.4214497988481014,0.8635091970215433,0.5522958022681812,
 0.009888040211039062,0.5813231485214582,0.37704526053214393,
 0.9931535940360583,0.45776362132117276,0.4508586860996274,
 5.650678352631111e-4,0.887259127412707,0.5450970996815784,
 0.024297722176139347,0.3072895830468979,0.5582077575728831,
 0.13602717596126512,0.3097504195245515,0.8068454383944331,0.8799439962833796,
 0.5781692084916406,0.2821550573932924,0.4554733643906118,0.8852948455671622,
 0.93416443393593,0.34467995603235146,0.6464512458088892,0.9685622142077264,
 0.9907472375078761,0.6388664867524905,0.8099482783571911,
 0.016370111925309905,0.6904122672224197,0.5102161119989238,
 0.6583913973814599,0.21085930627848415,0.9504720582687134,0.8908175467383002,
 0.8642786657805392,0.15558303967078158,0.7750772281167504,0.3374111059425142,
 0.0038057654466224777,0.3352039434773152,0.24424829645221724,
 0.07020972625330213,0.30088539587660046,0.8762242186488717,
 0.6782109069517778,0.7468524894037984,0.5049159125973236,0.31382046485480053,
 0.4413743314370425,0.6293698734516361,0.4646046005662561,0.10981388374830914,
 0.8181386465578065,0.5152247640671354,0.08007100429810854,0.1446874386410788,
 0.9765466560584914,0.029632821748732607,0.43975511152309965,
 0.794618063169527,0.6265267005071689,0.7026419069881618,0.6263376690611718,
 0.7467133114820511,0.40598582395551364,0.8054926340534492,
 0.43329606954309086,0.2081216407515648,0.3807373142445446,0.2768812967012091,
 0.7913164598066944,0.7053225794604921,0.08718314318146847,0.6510682156421654,
 0.13190651404069964,0.8899768475139251,0.8442260015892505,0.2790014220176741,
 0.40310615870256294,0.9620543703019152,0.062483455087369855,
 0.10692253393699713,0.6957808642219289,0.8761749520227546,0.2962641991734184,
 0.8284731517551528,0.9135835340875426,0.34348864387936495,0.6541335160971036,
 0.6091701734121266,0.8524207026059665,0.3329162906952524,0.16076668856831933,
 0.35280431006430246,0.7211952578090932,0.28001190501381445,
 0.7687606254908432,0.2181125155298701,0.0616817316029038,0.10729911726033536,
 0.2569346949296937,0.855249633477147,0.33490362166104193,0.05037337922495888,
 0.918056902004536,0.9722327411183631,0.26998083136127726,0.9776907278500421,
 0.05233318324813363,0.9357317990505113,0.5993647701120715,
 0.15754683466995723,0.5883200411377589,0.23892369302576744,
 0.0362087602405512,0.16677541104641558,0.7221196466377766,
 0.18530112228882945,0.2796995362814304,0.4816294212340222,0.2983196142419138,
 0.22278251208417688,0.5458024887950187,0.32258174215007684,
 0.5964627403435353,0.4713599055168465,0.816743491952625,0.6951211585670445,
 0.7057570827347313,0.2861248126826136,0.13867746773811152,0.9195802550562014,
 0.08362993091375825,0.36585990743777086,0.10227117536619934,
 0.8667782467354467,0.7961276161029607,0.7352380811901391,0.2164371711749996,
 0.550238874065248,0.15820546062154028,0.9803837388983717,0.2675227110014815,
 0.5731489860925876,0.19809793942511134,0.329962340572051,0.4139536112211555,
 0.6648460600864003,0.9186783344848577,0.7747937822955386,0.3138108841844446,
 0.3247466115592683,0.11809567333745619,0.08641001734018838,
 0.0687095495917126,0.5983404620749068,0.6468334886428467,0.6003121848678342,
 0.4633121727775902,0.9640032286588304,0.40557501824484987,0.5028188307514567,
 0.012228907184280935,0.9269346679817532,0.5064832116110232,
 0.45698387044931843,0.8407813167716186,0.4320629105347047,0.791293567821251,
 0.04919992811597751,0.5080738458258158,0.5619782117122418,0.2242854905144014,
 0.2972427481809775,0.8267616430492117,0.4008756080053316,0.3580541687373311,
 0.6531568611263358,0.35848394062983047,0.8923712298125974,0.9460820715355588,
 0.4250202515057754,0.9237800334945026,0.06586958355916772,0.9944745499097178,
 0.41979487865917386,0.7625839577276095,0.04135960234329561,
 0.45882085902036507,0.335250329088425,0.26997107621937766,0.7332832505880634,
 0.3387431148709543,0.6551329362116963,0.3217370536585459,
 4.047226009540328e-4,0.7242801326074321,0.09714346389466,0.7907963571484456,
 0.5898525287985238,0.8678529089557698,0.7266611394086128,0.1937766627168409,
 0.4714237591907873,0.8378269600232017,0.8388967657466319,0.615039435202051,
 0.8326171118678731,0.1907165944628696,0.036585470377349494,
 0.2588010700141631,0.17232606673611417,0.5666405741932777,
 0.08940358532390957,0.3256790102565912,0.5844305641816755,
 0.46451292311918735,0.811331038834347,0.4023388613742097,0.9759529889919827,
 0.722525835655019,0.35420214475575584,0.5603592388287848,0.6790412730109917,
 0.8422695324664825,0.38962052053025176,0.7504638165218696,0.3585863548856424,
 0.5360460479549546,0.8611291973905089,0.053721493956621735,
 0.10841730472882527,0.534128991978546,0.844318521785105,0.01225798279324486,
 0.7639784932477883,0.509119493221407,0.3265067174950922,0.7245796309629106,
 0.9199587007121862,0.0954780190625465,0.5827021592138125,0.28012784336158014,
 0.9176340645585265,0.8974359842150121,0.11387615250098648,
 0.45832774488595207,0.25231795819997493,0.7624947012388027,
 0.16352655573208663,0.07876349682515082,0.8543592167448517,
 0.47554887252364286,0.6662189583144312,0.3723289477195597,
 0.15998978793630003,0.2566396624814332,0.15490480050720346,0.664380063527282,
 0.4096325815581553,0.7358365979842445,0.9948950341047069,0.681914789613972,
 0.6900601329898987,0.39257188598585735,0.06834621168360222,
 0.1310729248499476,0.03638772003826507,0.22360435629157205,
 0.37611236051540486,0.9695995185048404,0.17043661942613353,
 0.8239916003950108,0.5348960725981651,0.7503020697705509,0.6739616176108568,
 0.501109114179223,0.7055215991083729,0.12100604865199083,0.05405157290787477,
 0.25694700538526405,0.19058842457354308,0.2576409781996396,
 0.13890581035533067,0.23573998444815425,0.339529895996843,0.6914082699221128,
 0.5492470098735465,0.213660171219336,0.6782280339890674,0.696310791474393,
 0.15866031921464852,0.756945123712954,0.6475963167798883,0.3425246082294571,
 0.6663462443233423,0.04883159514835578,0.5633503405773409,0.1424750288389457,
 0.07851845567732785,0.8162102598979754,0.7536736500917969,0.8598316565823052,
 0.7678925475307293,0.827893725182112,0.7979543654269282,0.8350064211758526,
 0.642540055010835,0.9995608319634861,0.9720117670107634,0.38133461841964744,
 0.6884863404846665,0.27545008561900997,0.5504994454411646,0.5994017458059379,
 0.9789737263368596,0.5579764318701775,0.7530920316018034,0.6867037758458981,
 0.8720766041200854,0.9239818745658204,0.3645818372613109,0.2303697796068369,
 0.3161482512389323,0.4066126017368604,0.7604167338507126,0.5626057436431132,
 0.5974540698424752,0.850636232941117,0.013905021442228827,0.6409881890759433,
 0.7098958076928736,0.32193841360349185,0.004707792034492009,
 0.03150226850718396,0.08798990669969642,0.29688743403825113,
 0.44847650414115325,0.14568182348181735,0.9076929176603694,
 0.052626906952428554,0.34192911438171003,0.46787938011695285,
 0.9328057831423496,0.054445089683228365,0.6206635509143981,
 0.8598659246696823,0.03285396671022778,0.9817886729373727,
 0.053145211575961526,0.2969589112348423,0.34494137362905286,
 0.6329599599473098,0.40141100134462193,0.2480652903548497,0.5685320571335652,
 0.5991268353844522,0.015122960950266662,0.6463730395219808,
 0.2685707991357571,0.7085440519641097,0.1349234215254721,0.3032110894766932,
 0.5847816244723206,0.3960609988109476,0.33323276816169833,0.3592201770947945,
 0.41989468320747836,0.40974025591833585,0.24064722369924796,
 0.46504929730946853,0.5744819568209276,0.3057964797613326,
 0.45909606740510056,0.034666554609883105,0.3557067562053293,
 0.047344818642826136,0.024541791911336652,0.8673990078469325,
 0.19257281175786267,0.10914080810889937,0.8822202950095817,
 0.7274250560784675,0.029446754135866837,0.9199793764856009,0.273310493225408,
 0.7430695361056369,0.7604644407762555,0.09475386556724297,0.6698816012236479,
 0.6180332493619359,0.07612569175839401,0.34815917642665606,
 0.3823561389144249,0.4543080384778224,0.6914390445962784,0.25300139736303573,
 0.02744929361104753,0.1739484898817939,0.21914761525469417,
 0.1452840959219257,0.12665230832839047,0.9039741836696595,0.4475912832295973,
 0.3378911424187966,0.1426485645335741,0.38384211489940523,
 0.20634221637915695,0.9329905553053564,0.2697730850389958,
 0.026751149314470646,0.9161974495764447,0.6752884108733677,
 0.9491978993358884,0.9637585165446704,0.5507339221428298,0.4360666661257877,
 0.562997442462008,0.7854924010520222,0.47335439719410155,0.07124812659915292,
 0.974217773632853,0.4954954576956919,0.8319363517911449,0.853244643122778,
 0.7033556919976796,0.6408475552273111,0.025472279048995516,0.881535861510444,
 0.8523663996598024,0.34027167520146295,0.4147781803443371,0.4813345364866445,
 0.8819819135503828,0.32822031462100765,0.6711150164385346,0.4182421318840467,
 0.5032754013471095,0.017907425475955696,0.5710147119397716,
 0.44158807757852425,0.5754949794384706,0.9754439123541494,0.8223620566629237,
 0.5269042135927475,0.42946783123502086,0.5402143920193403,0.572943580665172,
 0.31279824612749774,0.20748742742124748,0.5415441666598703,
 0.025800224415720385,0.9483875565527791,0.17268612220469037,
 0.7735991880097921,0.9452988435660636,0.6510848741498823,0.9390735156815663,
 0.16157194250524265,0.5365384774128505,0.7478598990776316,0.575402354513529,
 0.14375577613596235,0.1999027111133902,0.7483224109657094,0.8417492519748234,
 0.07760362457514203,0.8592570182534893,0.8811229371904519,
 0.32148208917477905,0.7300249337225357,0.6506892423246174,0.3740552576459638,
 0.591404170525683,0.454542348890684,0.8898889254373908,0.4488185943077716,
 0.4815755914847303,0.8360298498378398,0.719988091943329,0.9743053274629548,
 0.22909571746934687,0.3817870633399023,0.6545120278249164,0.7655737398896448,
 0.15369866916447394,0.9942701042621803,0.14113901089524805,
 0.3047358877614923,0.6523728381755965,0.4827111486571387,0.9784036347374128,
 0.8260374933198761,0.6639094661613403,0.9914277588679705,0.5794984588698402,
 0.20104942480177213,0.3181794530302793,0.39989733845909703,
 0.009314633987920873,0.21784749448188856,0.024289605619394083,
 0.42320375661015075,0.23428618988998562,0.35371784263302897,
 0.43142809747717914,0.02579436877978636,0.6460716118984698,
 0.6402505046032709,0.8906858837420265,0.5526763244194588,0.636186176915888,
 0.6778846354462588,0.4058235706414245,0.2072597363249148,0.7866639653176216,
 0.3260099336693404,0.37087128922223056,0.549696312393261,0.2332601065137534,
 0.705023917321503,0.6078381136960721,0.06434560016959612,0.10656444435141266,
 0.7533578154542656,0.5739679077459012,0.7742602247546441,0.8607593563774698,
 0.08648711306814705,0.6606907775677436,0.4433681279472068,0.2914869986500386,
 0.4563870117836265,0.6704105811576911,0.2479614328275226,0.4238100254523408,
 0.5908051101758407,0.6636002111748643,0.7050086927038957,0.34441783626938904,
 0.35294585414303603,0.12475737371502937,0.4037679604701506,
 0.9874920179242807,0.6971104830749875,0.716767560466584,0.8826163638304106,
 0.5215774060145101,0.26191578262186765,0.5098568975576545,0.9369776970932521,
 0.6191878627499183,0.8161285188838207,0.6386796318815234,0.07982657643978586,
 0.041661582808083186,0.4424898316337684,0.0983310131959293,
 0.0025971578976359266,0.041466388519033504,0.9088069702355344,
 0.14002731573353244,0.48472232861739695,0.957016846314836,0.6723253718261368,
 0.44859446284163595,0.5143286546840657,0.09849594971566311,
 0.3290200460186713,0.1950786457468241,0.8949846547950461,0.46018354971224784,
 0.1886363215069644,0.6330217254256294,0.06730117950200432,
 0.21395317643334344,0.9828468430671533,0.6542084556208623,0.359767219853664,
 0.8024151475976691,0.9808303484668142,0.8812209378619744,0.5657381916711887,
 0.8678041022421819,0.336952977809317,0.9018070710599192,0.49266088871423475,
 0.5250020793378022,0.4399087090830658,0.40197551389344943,0.5629510610811109,
 0.3164468165742277,0.38141507657419904,0.5464561479100914,0.9107902349898261,
 0.592273384031812,0.22815914145491523,0.013816636852266528,
 0.9691548263442117,0.5675976662122567,0.34700167357514844,
 0.39258480243503635,0.8185479608620433,0.43478213216137174,
 0.5059864559733908,0.7867790635019679,0.5113927310598286,0.3142805883239197,
 0.2066718222038033,0.33903470270682556,0.9248744278870471,
 0.10262186811412233,0.8697766266529143,0.16157247713594047,
 0.014451607678080736,0.020519186988266935,0.3118476295556032,
 0.8397841263267948,0.6823363816721321,0.12479066263030836,0.5078439887808335,
 0.16154745244590651,0.4390249995946517,0.9720229407218833,0.7254917931226095,
 0.7836963237829864,0.09050658391042721,0.099425941409783,0.4879570116683585,
 0.04583520588849932,0.6030557359529354,0.34294366832690826,
 0.3088119119044721,0.790680211182763,0.32924639709984926,0.21434460400160682,
 0.6645288438850148,0.7833765609953708,0.8308780105615478,0.6851958589309138,
 0.8281298016392227,0.6212987283508917,0.45765688810788974,
 0.18339405667470032,0.954771944873223,0.22217420769848262,
 0.05537557322983855,0.18685047441823321,0.8465632780331643,0.816620857272243,
 0.9559912187524718,0.867998325161143,0.3346736776333863,0.6355198781214627,
 0.8761761984677868,0.9244195771304808,0.5455262532399897,0.7640588117605177,
 0.712482396479924,0.4122513040384832,0.939230361602985,0.021734790562607187,
 0.6200816107206254,0.4897238550479557,0.3055746900814075,
 0.021071818470214665,0.4047362220321107,0.5182355948287602,
 0.015516677944777402,0.9164460215110461,0.8998705169422783,
 0.1681884291565119,0.14855553818022893,0.39413582504951816,
 0.7066491302258988,0.8890048143510932,0.8166691209852545,0.832495584622575,
 0.17346548494639746,0.6430797474127772,0.7888344202756128,0.649248301394442,
 0.7088726413343934,0.5881454361741565,0.7521482298870097,0.8718610670694628,
 0.04046222034890412,0.749899739098358,0.8406285622100913,0.7399261622893661,
 0.24676415336613133,0.26090065405245566,0.32203647173991334,
 0.659019472252511,0.784999615022137,0.9601813121829886,0.3320719786434202,
 0.7032748410635111,0.8904668001970252,0.13602071276673244,0.3898018491979711,
 0.7129378998765141,0.21529530137415986,0.9238607548087789,0.5267920642858568,
 0.2081169463466288,0.2108279613757218,0.7071431004761985,0.9408223763648782,
 0.7145659495423831,0.6310172286032418,0.20713461636790864,
 0.43823202051556565,0.9619726729581979,0.4874097861711295,0.9338783375165154,
 0.4850827610245647,0.15173385199135148,0.24104787514249848,
 0.5476748733844403,0.482983477500475,0.7818063620603809,0.5179556112160804,
 0.36486890409156414,0.05221525968523011,0.2024707245668954,
 0.2662627410976268,0.5894755070099048,0.06741625280115415,0.874121690344233,
 0.977694015430506,0.31562879428897506,0.04796508599075466,0.9226972105921785,
 0.5349078845753117,0.9159807519161797,0.5786979328550561,0.5077763607159735,
 0.5795869242513015,0.09613131533209218,0.05995970435265874,
 0.9174384282892464,0.5276926868235935,0.2175665469319309,0.4304510782085449,
 0.5576393231201382,0.013043251733028893,0.9002270287560529,
 0.9755834733994934,0.46086618713670346,0.9858377038008918,0.7278681827763789,
 0.6996464047925031,0.8297431163314318,0.5492468785021427,0.36584315544875445,
 0.4849372647583774,0.25847305989254354,0.21754656185552246,
 0.7668818651632555,0.8831519640705999,0.6963268995287637,0.5574397613072786,
 0.42456308168464807,0.6383922974382101,0.9688929924214068,
 0.18676487518453755,0.18747037789089316,0.8431711307399108,
 0.21363363919898437,0.24817441140321495,0.7712915117164288,
 0.5534512565113715,0.976846353605717,0.2813289637467665,0.46387207565262856,
 0.7848572980420987,0.008898444569077069,0.8528157167568547,0.885817027233875,
 0.6688924281789626,0.823687105498578,0.8402597389225425,0.2610462609644415,
 0.028355760003638597,0.5078610578782703,0.7080122610280696,
 0.5952837597282096,0.7406823359635376,0.0839713440141342,0.5772565017327833,
 0.22603714235256744,0.5907845842846857,0.7985005432400943,0.759666604642988,
 0.12248013921159784,0.7459211230775674,0.5277934361048605,
 0.39490792211388226,0.29632984253667305,0.0329014333118004,
 0.04932085803701003,0.5643563876290877,0.09282135633912192,
 0.5763938277404138,0.32796402459040774,0.2135932400327758,
 0.49002091561859296,0.449407190072002,0.004952346873320246,
 0.8219690140555779,0.042744405573004185,0.96863729416222,0.4032755109057634,
 0.8078627155372092,0.41623382380536966,0.8907857771032668,0.5725091584420394,
 0.5269701201923401,0.4005905051127898,0.4822126963941096,0.05104643522836316,
 0.4774467208461275,0.10612675090804169,0.9161535929780524,0.9393632669627084,
 0.20913534502034492,0.7091981940855051,0.4353437586266721,0.5952034079395803,
 0.7770416949851866,0.6844571413563638,0.3459932315594376,0.2042118993421147,
 0.1328268250061907,0.4786065472097838,0.06409034177929784,0.648058698515856,
 0.7691785226397552,0.45815701826273547,0.7587187615782975,0.8386702476240109,
 0.2991700507523247,0.4082603830589756,0.20512415631585845,0.2634253902203604,
 0.46860811424920323,0.6336686930354768,0.5520495637322478,0.7638562349872852,
 0.027798807632093525,0.3367647560464826,0.9951318986054335,
 0.47878186312410065,0.28854296675120616,0.24037980943822068,
 0.386001652966406,0.10375729958793445,0.05049965638514364,0.8492658923752575,
 0.25809477980737683,0.45796730492335824,0.45299802784715215,
 0.49167722879391285,0.5058887030888168,0.26477606280043964,
 0.5748254888957628,0.776774651130788,0.7346296880194336,0.5848726493970444,
 0.0596694950221307,0.7334751383813466,0.8379475060174283,0.4590860373431217,
 0.3304404335135336,0.24307943037812407,0.5448885348101322,
 0.22051506732031867,0.19583988154544418,0.8772868920239416,
 0.7823200344247504,0.5777580910719042,0.8863936076181437,0.3918024992663831,
 0.7035279821891008,0.23108294714104172,0.15954058793412995,
 0.6155049528856366,0.004867293799401984,0.16549639761025403,
 0.040051343497422676,0.7345478697644284,0.4646266421779067,0.629128618062681,
 0.6016334991002735,0.8894959570425895,0.8303198623081859,0.15252782314424773,
 0.8599565405077464,0.01211724059813113,0.836300265953427,0.3337014277057043,
 0.34533380305899786,0.0693370821870275,0.3765118396246142,0.3350454784349115,
 0.9017862373849024,0.213814075598315,0.8407901855342104,0.15289465128902513,
 0.6049782778612134,0.3455492561892861,0.46712042060103964,0.7349313589158761,
 0.9120951452826837,0.5743286819674887,0.8735734607140369,0.4838307578172155,
 0.00567230013295239,0.19727005911831785,0.34686998690332005,
 0.1994087860837228,0.6566372432176961,0.3762891077881285,0.9034678824691463,
 0.20428595543763772,0.23904635594479473,0.010264309271467709,
 0.42060468539003026,0.6194496372875998,0.17338658451506173,
 0.5131717847965549,0.45492812662551074,0.18535515568481709,
 0.4474476677708833,0.3722424316342433,0.11735866313950893,
 0.11686458803869038,0.2542537627234369,0.36471946316357373,
 0.5031553096442949,0.689024324476514,0.5825585804614725,0.10092745824722871,
 0.04037349691163605,0.8943822605295076,0.2718166499855549,0.9946647269652871,
 0.6341919086405917,0.9241139176901785,0.6237047617761512,
 0.009285968887876028,0.16126918064584872,0.8290674143768053,
 0.5527439634012941,0.35478283512791897,0.15127254044453087,
 0.6989894763322118,0.7715533927990453,0.27127240673196373,0.4341035164025021,
 0.7183590880342834,0.5725205676409952,0.9437980590453768,0.06505598674467128,
 0.06539057440196183,0.609546077487434,0.4659866530458898,0.7490096975441359,
 0.00495827942178606,0.16436393330224486,0.4061089235765669,
 0.9491316622920766,0.5193426578535353,0.40561095032647054,0.6703947395337024,
 0.21264116208306094,0.9985003592030895,0.309170949734624,0.6781029438185848,
 0.5300701716206784,0.7760222079446051,0.11151101794576501,
 0.041818313660270334,0.8438259566051672,0.736140748468444,0.5115392064130138,
 0.8503550641156445,0.1491619237306674,0.5348298084119298,0.8080364312738414,
 0.9223004802371981,0.4120974723515709,0.967382558512948,0.6810555988642296,
 0.3661356109517315,0.5706433012339149,0.6949237083962877,0.7518630554740844,
 0.007297082640249752,0.025495998634898265,0.6618442113720873,
 0.6981934992163407,0.20896002281205084,0.560382708052229,0.9112245349152863,
 0.5328448094423518,0.02972785604408401,0.21269385297429144,
 0.6277287634475156,0.17408225764985452,0.9017723979467057,0.5950654709916924,
 0.7373317140936954,0.11598350203584695,0.15992092551899983,
 0.0647045061100675,0.7952736096920727,0.4258644899152425,0.5698751602489369,
 0.6389671260691587,0.04742167286902266,0.8784389854654994,0.6746761399938394,
 0.45288611175545734,0.4480607619208048,0.9916555455863048,
 0.48324324372060534,0.4556398533168933,0.9668458496231425,0.772445144383917,
 0.941775398620296,0.8186417931341167,0.4181340485548096,0.42950632844811065,
 0.20130410462238157,0.8671147759874984,0.9244134585533293,
 0.10478039457716559,0.21594400835909466,0.9402538685371771,
 0.8133204763988056,0.09445288418010811,0.10666291490006208,
 0.9273377902031795,0.9401039754355951,0.33428833815370607,0.7368103415365413,
 0.1106582559306366,0.5288518142182297,0.8469693288837926,0.24502992376049482,
 0.5510529445966892,0.937081657811747,0.4346107249743978,0.38797774378459526,
 0.5997173316258229,0.04355379788056546,0.2904847354244543,0.5049927287848691,
 0.3581937447307977,0.6507109183352692,0.16236199069550494,
 0.17516869458141926,0.016816928294481226,0.6479649714816106,
 0.2922431684404888,0.31798791907902124,0.6624603826757914,0.2862280871852343,
 0.3770970902887878,0.7023634186279202,0.0620082750160017,0.5638678278863469,
 0.6459313855165377,0.9977769474193784,0.774814339687323,0.3510966021867351,
 0.31827715182234195,0.5041594680925172,0.5564399096085385,
 0.40990157427432317,0.20840996581752713,0.8274596960820757,
 0.1858754986827189,0.8760054440385214,0.34405134553493044,
 0.16529514868517325,0.8344252810277956,0.05306428996071533,
 0.9949920940338003,0.15621948626687798,0.765009022672182,0.5654175650690926,
 0.5636658648199508,0.6277468851095777,0.21312664489508493,0.8664520472123888,
 0.33741776708702076,0.9017087108305241,0.46853564667908654,
 0.7945376206534979,0.10038695910699014,0.7020410246393101,0.8411905933965401,
 0.4834403473522282,0.20953396325601992,0.7946563707488598,0.4053614531617288,
 0.6797506744661374,0.758090962756403,0.9680197397639388,0.885385043181572,
 0.009497759210623458,0.9359505691609995,0.00978859722240677,
 0.9561296629607355,0.5076954229576274,0.5792508972206144,0.6801692470819896,
 0.3310644589557098,0.20254451477148194,0.6125188110953933,
 0.10525704955106874,0.09511763131548534,0.10515207661264914,
 0.25651234341530227,0.9817538144537625,0.5745259276435589,0.8128884007623246,
 0.5092598676151803,0.1952843205796848,0.492725327356463,0.28021827646816166,
 0.4887237006650039,0.9673188280769155,0.5727581904701753,0.9647435310045345,
 0.8531935011524888,0.7906833280423344,0.0460151337597664,0.3080679933683076,
 0.8967522058759521,0.18817680257860903,0.7161093937407915,0.902953725995805,
 0.47751662836021147,0.7149859740951906,0.026475131636537808,
 0.8623652779797919,0.695903455640283,0.6999474419419887,0.623638062187376,
 0.9583980221428867,0.6693687737798169,0.925754454640813,0.29963421902379717,
 0.01027081085698478,0.7812048714637997,0.1962176477106391,0.7600412138384518,
 0.22905409339145189,0.33021442570962756,0.20851869455577898,
 0.7526821929563088,0.8110967162183506,0.5880135697769193,0.563977572738507,
 0.3348505064926387,0.4023422805393677,0.7365439730786979,0.26383919539947065,
 0.9833637178419218,0.9383324126976453,0.3107176104249142,0.7510938109899985,
 0.31070166571309354,0.6395090570162896,0.1332331842291966,
 0.43087911663647416,0.3533454440670539,0.22925552345972278,
 0.9177910406850424,0.31475787223481877,0.5353987745229162,0.920388516161479,
 0.6157641735324897,0.3366543208049758,0.18847461372243568,
 0.16316540525875822,0.6802409813123608,0.3285295651043101,
 0.39763711368206933,0.7026965915487619,0.741031427145609,0.2695777588682713,
 0.5419559466919323,0.0468466726994875,0.9753917200133443,0.14164174981944433,
 0.07152722402002887,0.8185304189583499,0.87033735137781,0.1557220903217429,
 0.08598952103283164,0.5478240048586842,0.8904027506234359,
 0.43613871501090773,0.8057375363139498,0.6974158600167475,0.6293211507588522,
 0.10054178964757332,0.43402447458020754,0.8024889356997765,
 0.9469857325977087,0.9783945750604457,0.6816352407293192,0.6321138424956259,
 0.7791387659375459,0.4132321278987078,0.4233868919960009,0.4575167333750886,
 0.09064100758244764,0.6894711674443218,0.655641389136945,0.6065452779759782,
 0.8820665002934989,0.8596782479560581,0.9291446245402263,0.17701555528247548,
 0.026267851771423545,0.9804688736880622,0.3584055374141991,
 0.1831483496451739,0.9272520326305389,0.49514181378568545,0.8415328562085067,
 0.4834773185130209,0.7353527545919575,0.34399104535944514,
 0.07615070450189454,0.5373708990466153,0.566043184937105,0.5184038402025234,
 0.006543004376307615,0.31794676054492976,0.971897077287916,
 0.7706036629466488,0.7292081608503561,0.1508601887135822,0.5321009949746228,
 0.9244021118732707,0.6956541576395786,0.7517093299968994,
 0.014832399694733378,0.5120199690901459,0.46394116138852104,
 0.4283005794506769,0.30257453911666443,0.780177305554726,0.543801278407918,
 0.26187202536711696,0.4807206708731284,0.017341207427224448,
 0.8716494975717876,0.37120508717134804,0.7068722593961779,
 0.11529536556512887,0.42105978021657386,0.6925380263554348,
 0.043742886123899205,0.8387290897776297,0.2635694081934342,0.511229393391754,
 0.6873269436890102,0.05930984125409722,0.6700142202356453,0.4964658216930341,
 0.3922916681660493,0.7830346905314197,0.34028354116220827,0.1902713958547716,
 0.39135627013047813,0.3607839056778509,0.25718430794071334,
 0.6435063444907416,0.5044268545675847,0.30623488423247436,0.6437593756750218,
 0.1943373191553589,0.15140747273493882,0.9561417644255975,
 0.09728662548596734,0.33622525628008015,0.34784851970446073,
 0.19656324565368466,0.24410630709444092,0.10719799824976806,
 0.2721370243505363,0.08897494565678266,0.6088548873804724,0.6512477401870861,
 0.9012939829214507,0.7986082599407976,0.48367566820592434,
 0.11854959158189449,0.9388590803594685,0.3891370095840625,
 0.47451944963129244,0.07996191525883123,0.7847174858932938,
 0.4560029565814696,0.10365901177543102,0.8766931094096995,
 0.29046493167012755,0.9833661751807119,0.13060367618275193,
 0.3072591938553767,0.2911151649002538,0.8844333029857859,0.3529681030101056,
 0.5617502583237798,0.6874930016256317,0.44141597729190996,0.8954308250851383,
 0.9624797987656095,0.2676995393427156,0.46215080057665814,0.896656440078393,
 0.4074075127848791,0.3659793680183512,0.4034857170859789,0.6203008848286683,
 0.39763714900723635,0.529054358296104,0.460958505358249,0.40540506111459373,
 0.17376342483105278,0.636983096711845,0.8194388741233793,0.3463002757936521,
 0.18921564503650612,0.6749244285449895,0.0022193809135842013,
 0.8044413619188566,0.10934506908734054,0.7043367296350109,
 0.07783632969485599,0.32346860622301654,0.9010479087746519,
 0.9631733910816875,0.6129199547703386,0.22634391489850247,
 0.05154131622666158,0.08925186302061272,0.5850486013582,0.18776730983588963,
 0.39097293050290793,0.39682015254702663,0.9889293549929058,
 0.9030173788990832,0.8370599155852413,0.9229978560857638,0.6966106020224326,
 0.8248541487370574,0.8483949724982228,0.10425886153381891,
 0.13973373063276073,0.11471581657064467,0.8366769237693459,
 0.9669226587924343,0.4315473565591581,0.6859296496659633,0.5925039089598262,
 0.2875777656152321,0.4174324315231044,0.5858664817626188,0.7154722702192938,
 0.9621268776425123,0.6921925910337537,0.7115452558128297,0.8209050403338555,
 0.24490152444632995,0.1497627658614863,0.766817866504999,0.7554460247152925,
 0.12522925487725134,0.7987446605225891,0.860014825136329,0.9684751662679547,
 0.8302001237176163,0.2771266151888341,0.9833583001000098,0.5363005193556517,
 0.5808614940423882,0.23782991196237557,0.5427016365569569,0.8269439131850609,
 0.25767945950775906,0.16108353940859343,0.7472523763687765,0.932406866527332,
 0.6425161900554369,0.5730974883332848,0.023340623124008975,
 0.30140300005417675,0.8969560500880445,0.9830817913828686,
 0.49393736986750725,0.7836835831206387,0.010910612234333317,
 0.008094552312736925,0.35688222000972025,0.2332265883005029,
 0.7792345879333424,0.7552842163398653,0.5394839865466852,0.8299749513066417,
 0.07575710857687312,0.09083517355984294,0.48193752032397985,
 0.5008091842432449,0.3814934750150272,0.43926190354515715,0.4035841734916866,
 0.8870472437321242,0.44618879670876577,0.9849096929470865,0.8535986084129147,
 0.9453017137522605,0.029658699633514862,0.45871006149957116,
 0.0443783822346564,0.06027205463651408,0.805163529644456,0.4738673807045871,
 0.5558133007504416,0.0999924904976891,0.12176401519026425,0.1667652108459553,
 0.7397546312727283,0.8823093142000258,0.11861938226720414,
 0.013893181100987806,0.0655951344217196,0.3040004377915628,
 0.8117152094613261,0.038247254432720035,0.11785390794669193,
 0.26385531471271906,0.7440197194166364,0.426255719541884,0.6440235971636186,
 0.5627996824384953,0.25431606404111684,0.22785764736606573,
 0.011174177374832839,0.3029746725771366,0.8932494218483517,
 0.9071949988017982,0.32469299304884514,0.845499091601703,0.7717379543303122,
 0.42986764390588006,0.33465165968560073,0.1366332189543411,
 0.10624032663045646,0.8964891955189659,0.5209009614260413,0.9955240450763342,
 0.8430118623047582,0.21079251369409968,0.9912841578049563,0.5605996791887748,
 0.7504477525430855,0.6744349792882168,0.04086713228110761,
 0.033360626320066755,0.4937870774299262,0.7277661224346138,
 0.9038265849440661,0.8471754280708513,0.5958071285770367,0.3137288207663065,
 0.05363814788917065,0.28314752100140805,0.0397864678232559,
 0.15006627566139452,0.08622719523266742,0.6778914133617273,
 0.6072806294216215,0.7311765176752846,0.8855244612615727,0.2773743318320818,
 0.3372267191669882,0.10594156711765101,0.836664996289385,0.6080771143693892,
 0.8537850114475437,0.05971738578661023,0.08778946707997326,
 0.4046161474286367,0.19064214582081607,0.5800679223527155,0.972135021829174,
 0.17389818564712867,0.34354317415717217,0.14863563107379674,
 0.5700846526702885,0.6981504037270005,0.9790739107838556,0.6342056192013423,
 0.8773939968804114,0.7553933192915948,0.7688995478080642,0.9945409020031953,
 0.22565486211724184,0.17581409442888152,0.3597598692398565,
 0.5837265932245057,0.37020980447117047,0.9544983027075259,
 0.49433839798267176,0.3462280294258143,0.8800401874161192,0.3841079822739757,
 0.8391134119309801,0.2586642301781167,0.778207539316851,0.8493427826937533,
 0.37273930812277656,0.35719476487724167,0.11839992135500133,
 0.2762774399404616,0.37111599277166674,0.5303116577660937,0.3346969433636704,
 0.35522120098723653,0.5211339625952525,0.10801001017211298,
 0.8023943456145224,0.4457488932642073,0.7953750576623472,0.02330690563368021,
 0.3696464614970405,0.16351655038966206,0.5839602084059421,0.9198515944780575,
 0.2873073767516173,0.7076941245631594,0.5823957620035409,0.6552662264383667,
 0.9516333172846599,0.39104273854165783,0.014507413122571089,
 0.22382624626993863,0.34022336961787103,0.6789286086193452,
 0.3343146724580962,0.8491188972217771,0.46407335626545965,0.1590495217434127,
 0.8956657030235734,0.761894570399849,0.10616204534409279,0.8843589875940291,
 0.601185930468094,0.018902390949864567,0.2041385469596747,0.4078128610661357,
 0.3400565040508631,0.034738456207195734,0.519519301959058,
 0.13262700433621555,0.19876464940354577,0.6691673105181768,
 0.6102166491681665,0.3244551510629282,0.8639846691835741,0.13199795813572457,
 0.3267202807288202,0.20106541807083,0.6045999383513376,0.6513658520621077,
 0.6057936845050302,0.3268569556897084,0.17773576599153695,0.5628821896911648,
 0.11899083406174538,0.8100057114338851,0.20605627427477669,
 0.3877743494440876,0.921091112989097,0.2409806209800146,0.40827850876549243,
 0.013434495939712354,0.34025373154733785,0.9825658614617689,
 0.04862941495684736,0.33019060518491217,0.38714397655088917,
 0.9638628521181457,0.5907558314874535,0.2356129144050384,0.27076724085425663,
 0.6390668005535967,0.47535601457237053,0.058368050312689546,
 0.6724436618148094,0.46236224906131973,0.30858218560875506,
 0.24004228855102094,0.5969263301665974,0.7943607848760725,0.4116877443964122,
 0.2693988668562711,0.8598750516702158,0.8535426103560209,0.20613730017352316,
 0.18843268526142276,0.4853892677772378,0.022159482021691934,
 0.7666573796241883,0.19866113110122108,0.8905984560963706,0.8669422923046094,
 0.679760871716101,0.44382830330647116,0.6194500882048676,
 0.016548088869988598,0.19736514443806152,0.1521016520754923,
 0.8522373079057618,0.344229611486742,0.24908648404094358,0.29292439148207905,
 0.705341345744159,0.6475918857352119,0.34293296486356317,0.5079994007260197,
 0.027778790635433537,0.09259999225338178,0.4806591932028972,
 0.19768825877615992,0.2990237462235713,0.5849937188295278,0.5037191958669579,
 0.3301511279786903,0.013297547521124997,0.7552781983351443,
 0.7681263428327451,0.06242134147890166,0.4437944813249891,0.860812276156043,
 0.9160889590620434,0.5661986082361992,0.44604143557812814,
 0.11605182113221568,0.036741763463554844,0.8926435459968052,
 0.15916728966473193,0.6784162425141296,0.46178185142081174,
 0.9874909011481714,0.2373090221566977,0.9991797763794517,0.7283965214656218,
 0.5078910462386812,0.49304150021368076,0.9377339160680962,0.8623980818986293,
 0.5603731287591449,0.755290520025668,0.20408954623854536,0.16469103544531538,
 0.20367894706235024,0.8996517497353437,0.14217382258591316,
 0.6688896813062076,0.46027842142516506,0.739488718910165,0.359375435042931,
 0.4087007198266439,0.6978602325826271,0.9389604567573597,0.29573260747868724,
 0.956683596541317,0.42476269759791185,0.5449298046261524,0.7182415863635523,
 0.40223823230242106,0.7681427406476402,0.8177636973151817,0.6473468492888917,
 0.987690822246552,0.6163183944687035,0.6000812204428798,0.6965360152242238,
 0.7988220161842722,0.19088513256433792,0.7829066689721563,
 0.30183550093310874,0.6224995277217392,0.371707087416303,0.6621105358463569,
 0.9511119063108086,0.9923729593602973,0.04828616781384165,0.8373219488574972,
 0.37199609762215735,0.01560077154449624,0.5797050004087795,
 0.5453793245362782,0.6084247960544981,0.17617739768881324,
 0.18859450950364332,0.4181095272638925,0.8860208320301027,0.545998546172758,
 0.7711739762487537,0.23347931529466637,0.5795354071573191,
 0.17636511272175692,0.41143027802333276,0.39219485417962563,
 0.8682687539494014,0.6846565947681588,0.1507888806581863,
 0.016705948494477596,0.002617826753631425,0.8200617628779105,
 0.734537711832551,0.7632195814391288,0.8360930516396828,0.43578493529270723,
 0.4101031039795522,0.7725754493738938,0.39066397255784635,0.6591673494758538,
 0.3328305720920711,0.5401071413795717,0.6987697221558493,0.531258326462847,
 0.2794843206959169,0.9508485928660418,0.27430881787867245,
 0.09125645660372272,0.8852484310446884,0.7437854040947043,0.9622072249060536,
 0.752541681170042,0.14607376220949042,0.12261998073150471,0.7638255255510188,
 0.28535725441941706,0.9543999470640672,0.9740674740271437,
 0.25029413111796184,0.535429615510293,0.12599455638292678,0.1945254566604373,
 0.4823609103309806,0.8294194913415124,0.679296189798265,0.05602184801945442,
 0.04369973993731502,0.01734176133138754,0.2409873943174402,
 0.5845656268261743,0.482402786281369,0.058760665790691835,0.127423435134852,
 0.8467405249016271,0.6056908951613142,0.4854087903886861,0.6287759357707621,
 0.9979911796018485,0.7381827871242614,0.808886493652067,0.06317389121158401,
 0.5069557205112927,0.023775156134785758,0.07536811461822912,
 0.7239405645798624,0.4666135932919383,0.6054464411212486,0.295789949064576,
 0.02705457764546859,0.08496702237938425,0.4900029143270147,
 0.02808560093963819,0.4190691486634428,0.15384634940024633,
 0.30992289677682794,0.781097743079566,0.28527061343328275,0.635656190680355,
 0.5037635584131135,0.3106790493540319,0.9751329045309531,0.07879970062933772,
 0.620612659388928,0.49495912212786,0.9607019477524195,0.0509206670309964,
 0.12054473339091054,0.6065270312992814,0.18767528849239856,
 0.7320947314388709,0.7891521257552462,0.4741541056097196,0.6203669013829146,
 0.8930689653099494,0.31074845020161557,0.9872128783401378,0.5501874517071261,
 0.8977675814775036,0.7305395987357188,0.5205959363047753,0.23357144911774308,
 0.8089175923388855,0.1265120824854209,0.9056434542677859,0.30243097764031823,
 0.2738661039744723,0.3949278063510867,0.999182379628169,0.6073280641916071,
 0.2729015622566008,0.9020737924288611,0.7093529885817552,0.9905099658043548,
 0.34334663061608495,0.9763347373333701,0.5609662449634689,0.5562900579519267,
 0.12805097442794977,0.3469381984231631,0.17012648263028896,
 0.04981677065033807,0.18382788312091747,0.5882755389611075,
 0.1398560262980253,0.8103824942933128,0.28961808915047993,0.8668506551200814,
 0.270766590167976,0.320881745795913,0.9804197627395866,0.3676688695378333,
 0.35504235707417764,0.4436057379095413,0.38695882714781094,
 0.9201442917510576,0.8151213705443404,0.944640250973411,0.21260986630123335,
 0.4699279186937205,0.03033503736632215,0.020035665437117434,
 0.7582265073578417,0.5142543238665607,0.43546789752039416,0.756328829781961,
 0.25435462928602504,0.8103817540705691,0.6757018360803282,
 0.38914489036462285,0.6879504749388398,0.9038091405665709,0.1576525988530505,
 0.5062555214922356,0.20620401003659317,0.8433802057617921,0.4545177573257426,
 0.9956908475657152,0.7921490065792473,0.9868055280215752,0.34827936431141093,
 0.8308847358554388,0.9781448965307459,0.3090435397362281,0.8761464347384594,
 0.07094332372583889,0.7226251810470111,0.35080386718635415,
 0.7758854897156635,0.25193713613360236,0.6459418830996787,0.3703735592782902,
 0.621546922349807,0.40876257581541897,0.24922206718883777,0.7913819657089419,
 0.9396299682541177,0.7340868913459677,0.22320192322873567,
 0.009734251992378207,0.7705280494539057,3.323644036598017e-4,
 0.2896351626719643,0.04737446821070457,0.5848332337593095,0.6852610502130355,
 0.5914790368296508,0.2932391568324172,0.22753316762711795,0.4633357754092875,
 0.6537566825493599,0.5657606099647984,0.7102729662456754,0.30680542919577014,
 0.17172520114028988,0.5700744816306436,0.6020472634317702,0.1229916028070569,
 0.7415868490295523,0.49244729840337453,0.006334813662443706,
 0.3420441835821949,0.5503773180173129,0.643338429763795,0.20699843185975952,
 0.6521566314753101,0.25765029138093765,0.49168930634814534,
 0.6945605566507187,0.18596258228190543,0.10542030944527281,
 0.5438669033577631,0.4250889014968008,0.05699374581443406,0.809920452779799,
 0.3593847020993485,0.24186703894279826,0.6622515070821686,0.839111979396496,
 0.280083221723306,0.3078749405223643,0.44989706814789787,0.17729188558541542,
 0.8712460445297063,0.9328918677503496,0.02027240831402377,0.2007652211346047,
 0.5227104379206454,0.5633112743975182,0.825570764833083,0.79860460399444,
 0.5446634408578874,0.38947319682018877,0.41775348450763516,
 0.0723022055467839,0.7749506110657562,0.4072806941280769,0.8696573788084937,
 0.4821427859896067,0.9238713387258214,0.6374598644083177,0.6503721410502402,
 0.17478192704743978,0.09828600418234545,0.06589178171432386,
 0.5037324155062917,0.5983233017082314,0.5262205734966335,0.8106369472192179,
 0.17720831324500363,0.8109265071422932,0.2771162594779042,
 0.09894667588354955,0.6724798539080843,0.4477080798125417,0.3766261854544912,
 0.3451079062546959,0.24856876183726795,0.55934993626089,0.19140270350532518,
 0.822266490940837,0.03660027391263454,0.2520814467820398,0.9155137546705666,
 0.14893883399226382,0.9324066303576972,0.4030977149906666,0.9047840989349704,
 0.6295830980267408,0.713160507639607,0.7138814291086995,0.6548579230349698,
 0.2549323547274037,0.6391962882202168,0.06568263193139001,0.5130801809642129,
 0.8911010420443265,0.7815683874509206,0.8784657024197488,0.6281090276569518,
 0.3516472627797925,0.8116881918386826,0.37393227943724416,
 0.01821627777061674,0.2040703519468896,0.7772403365774206,0.6909027903398254,
 0.8434108586861907,0.7821355205812234,0.02939653474578252,
 0.13419268594983402,0.7714430770941559,0.2820025606496962,0.5339226916902247,
 0.8672852187381487,0.48766869453494954,0.42300157840353403,
 0.2898893700145142,0.004933856343494192,0.6966972435592168,
 0.013010816808499026,0.8612651712375984,0.49697124973263906,
 0.5382289794367705,0.5903804075197837,0.7214071806010026,0.7109299529396905,
 0.6226865559164795,0.6503526680364851,0.9356989028349614,0.5278867295063578,
 0.3065537749495707,0.7836880323718586,0.1848467004859078,0.23519356970319727,
 0.11458901179614811,0.4416020805557339,0.29532870760084473,
 0.5576607857130323,0.11657517917409244,0.33734734733119887,
 0.09246184059803264,0.12831493717744236,0.7619039922965014,
 0.09460810789071272,0.6267361784261417,0.75934132875662,0.34536691475637826,
 0.2937639029076936,0.1590231724648986,0.15475130357988887,
 0.012520488433900634,0.7805755569962576,0.3357513093201746,
 0.8096772824713363,0.6191666187992056,0.19650828204256576,0.5305117898674298,
 0.8883194648217279,0.18352898225393555,0.3297911360098089,0.6414777848585531,
 0.6813448650782803,0.07688384203443299,0.7785424223561417,0.3788011502127646,
 0.4086205554331305,0.8835134602655975,0.11435328970854342,0.8418308407887906,
 0.179377375778224,0.5163629530962509,0.9348096972289803,0.1102160828226385,
 0.15246176759416452,0.6313241284881971,0.0889725200763708,
 0.14878792941157237,0.36633838074808456,0.183026421726891,0.9368757626843439,
 0.4293027710357934,0.16946761242756914,0.6039033645084658,0.5413103140356701,
 1.1893315415134609e-4,0.7527649821890275,0.13175965057805483,
 0.37329334363818867,0.3677043142110038,0.045489547780158124,
 0.14870162521325092,0.3496899237916751,0.043598055127683555,
 0.7206814038854601,0.3615278905389999,0.11415042991582103,
 0.07330314016479511,0.8837468095661218,2.1026291509640416e-4,
 0.32929854109097945,0.7954222178420957,0.962514238924507,0.7308142060139292,
 0.3010742253176105,0.6611901998221217,0.6953617331077896,0.9320212239685394,
 0.4230736747573608,0.2391171791494403,0.43772651309876043,0.2941513001023165,
 0.07755361006281258,0.0710970751664951,0.7212552950675605,0.7604413654504598,
 0.7474877060334978,0.46263021560946893,0.014174198572245666,
 0.875697742652537,0.06304993733955588,0.948713357145101,
 3.3434894166550766e-4,0.24831385350973423,0.20550131917651282,
 0.07440221777014,0.6500515157444341,0.23539134829883457,0.5894299331238742,
 0.2973846063915697,0.6488109511864463,0.946685252786673,0.8230931465935916,
 0.047131801748158786,0.19995331950094808,0.21584601781387192,
 0.8344302374827293,0.512758173046588,0.2254462926865728,0.05169734099172374,
 0.8853757169762492,0.043011027116530376,0.14428988726487957,
 0.6881806429167875,0.12016790458208959,0.9312375773481711,0.5283949762144211,
 0.9646148448448274,0.7861212177565579,0.5660670948134847,0.924564808577441,
 0.9864753825951097,0.8465949178849324,0.60336619635857,0.5220847008824986,
 0.303524733952927,0.9785584359301421,0.5889717046282195,0.375683256023151,
 0.5231119127524664,0.8845315356857827,0.7848904888272953,0.3719573374897881,
 0.608190561101212,0.2932561099454293,0.6920131666900523,0.843340053847164,
 0.22719659049113117,0.15176031528864886,0.9007959134641008,
 0.7655604965294162,0.9400143912544549,0.5542153707701148,0.9161904287488881,
 0.3119440500168136,0.8176881856408152,0.5149765783543842,0.30747948012720827,
 0.690035133717373,0.6652831780631525,0.5606137889231254,0.0412642956486069,
 0.6915306860402206,0.2513645485540522,0.731724994191234,0.1522570334861435,
 0.6068822657170735,0.7414798721319897,0.2530002613887016,0.49628415920908675,
 0.16105620356555517,0.8089025529698247,0.32314870771308624,
 0.19453063660538994,0.7132359079088691,0.7144407370590301,0.6408259026820589,
 0.623844881431005,0.4973918802015237,0.3099807777310376,0.3213071382469235,
 0.6205764420547819,0.04678177726597843,0.9336525772228318,0.8170591340253339,
 0.048652877551952356,0.4257295718584658,0.5030288886414774,
 0.32295361583897575,0.6908324839984186,0.1780689447300442,0.692900023291542,
 0.1998926654856733,0.8816180353641516,0.35266561970844656,0.7943489631671209,
 0.2801443838237463,0.4020410881453289,0.9851620427911074,0.67042221072164,
 0.09931402077374107,0.8207145442524071,0.449717895983036,0.10289963670179214,
 0.7235652924875947,0.2948317016291824,0.20203423648181917,0.4048330446389864,
 0.4173360700964177,0.48838144876163425,0.24083604604277742,
 0.11294456706915346,0.08874255706298306,0.06253051040443791,
 0.5510533289627114,0.3534973098490456,0.8984409079597939,0.28857459950916753,
 0.08305198993427532,0.35668857383972585,0.12429831483500176,
 0.608587431163948,0.7767718588051753,0.6841298227267719,0.5247937722496875,
 0.12693420115516496,0.19346563601614664,0.8040505929710416,
 0.39208507909749435,0.15467096681087122,0.4664026995759468,
 0.14423582290145143,0.44179088016829604,0.39657180306203244,
 0.2612067563062095,0.2843588724392615,0.36662795161347483,0.7469794748994499,
 0.14196805263089873,0.9640328630741748,0.9311422340327431,
 0.04277405840118109,0.25353175437628583,0.027791928759847262,
 0.9735489088911063,0.7588432704748299,0.8537584987159773,0.9669855433702306,
 0.8569138572255562,0.44856547238602695,0.12258953303637699,
 0.2982377209180038,0.7038078079687144,0.04508210853985506,0.3564472226382974,
 0.4505471449014672,0.44612976077786737,0.6243446986495456,0.9600406040999416,
 0.09688007819676436,0.823589739150787,0.8366481880326736,0.3449682399828974,
 0.4344057746307909,0.5563862987134974,0.49222886203629557,0.6883597930807511,
 0.1649329594841734,0.96036510355807,0.28059943330234294,0.5680391015138846,
 0.16642053226051234,0.27456687395376034,0.6527732791626775,
 0.3271648415598011,0.6936400851403154,0.9875443006244474,0.7146332839453806,
 0.026745456167747927,0.9956842079287984,0.8931460011446819,
 0.4863488126615375,0.7786836563279842,0.5994735861338709,0.14897079070671315,
 0.6623196161694944,0.5891002998150826,0.8446064017750436,0.4344180376990818,
 0.3101182722194543,0.0322158149304379,0.9163631655746318,0.9254831973949704,
 0.12193842832518476,0.7167946291471472,0.22077395732226002,
 0.10015177853982593,0.9261113343292602,0.8744405406006459,0.2588244559406431,
 0.8775751894591269,0.6789696507465395,0.6374348375143426,0.9461755686818678,
 0.10138032054299584,0.9652003079244971,0.36117121615809145,
 0.25990451482940957,0.05843604343931541,0.0636580764430068,
 0.6312586239672825,0.07248781161156659,0.43642483970830803,
 0.6353526017923578,0.5281375526263647,0.38145984014190826,
 0.09220384105492896,9.908229615684938e-4,0.27423463602851894,
 0.1614667230022062,0.7996032858280138,0.42972540161611117,0.6192142266404438,
 0.8689262608715117,0.2816724605528962,0.10612589838356623,0.9180189843634768,
 0.9472246261918601,0.9466901165821016,0.6240994398849604,0.16697729346498003,
 0.2346874850107351,0.678108792833526,0.3626124950984607,0.8073061515785935,
 0.64620608720817,0.6866645074714613,0.2820351303611428,0.5427394845076771,
 0.20805533113913555,0.05495505805439893,0.9010556462255814,0.149975764510907,
 0.35342503473572584,0.9911006981438556,0.9346116648564622,0.7854916284589362,
 0.6270925824463457,0.1557966680687547,0.3987905726990274,0.32153307997321334,
 0.8100358209625013,0.700616155844519,0.8531934578762146,0.144975407437138,
 0.2652114064968758,0.18526169087507727,0.586379077031768,0.5713613089029412,
 0.8647751685278285,0.6358046293820652,0.8888130178701068,
 3.213096845265939e-4,0.4769229980970142,0.11484350877463578,
 0.10254709588156885,0.4013350315693287,0.44101049871221787,
 0.7024555159919241,0.6179208036688748,0.7999523463661901,0.8332289515349157,
 0.7345338700703624,0.8571168031545006,0.3450073693548875,0.3355244580129205,
 0.5710026465973592,0.22608132240309686,0.17827556426818048,
 0.10090231999989352,0.0011955386607690466,0.725471521851374,
 0.5647083635522868,0.7844708026097863,0.9194539457023103,0.7425254241115022,
 0.04893372090988257,0.603347375825255,0.15835489222877897,0.6744110296312192,
 0.16088932036066173,0.25022872700411547,0.34808143425172944,
 0.05238684452007303,0.7382268839280632,0.2896064907889745,
 0.34970867742609557,0.4698298742324448,0.5207834654209592,0.8459008524919037,
 0.6907211901827744,0.6049179314683912,0.07822886556456776,0.7988314603187322,
 0.6253608901906924,0.4182775608870498,0.9560862155668812,0.13532192892657158,
 0.47159528045543464,0.6506443567604944,0.6618729731621187,0.7401802984375443,
 0.41058450745105257,0.9744401813555086,0.35658360004484846,
 0.6247076131373401,0.3142930815375352,0.029617206679597108,
 0.20655547149142306,0.43428356916882604,0.4336975494057935,
 0.03803440908817757,0.27407095842988527,0.5471698414293473,
 0.3688155470436576,0.9034582253302299,0.7798337941509544,0.6762856687094112,
 0.06377033049095915,0.4859638236988987,0.22553683928785317,
 0.8685298836955568,0.4776359079514534,0.6277527377862635,0.2974487683225838,
 0.34239666425120185,0.52064840786718,0.8489438428234514,0.4462430887236779,
 0.5409797599060224,0.42558481844727536,0.17071086823934822,
 0.8188508471235355,0.39155875992278755,0.2946632155355924,
 0.40140796083351626,0.40338433678740016,0.23267864064799737,
 0.07148275131967219,0.9627466513950923,0.3450573972701605,0.857893553269414,
 0.19285667228818437,0.8054769925410585,0.7463544121593576,0.8150844844038233,
 0.28448834463645,0.1479955827150824,0.10517112892212,0.4018425087164238,
 0.004289578460788146,0.15408143060889234,0.28700710560288833,
 0.9771047968152536,0.9736009229025946,0.26047362150095654,0.2235804626201927,
 0.21015140924059816,0.3618025488763438,0.5481303268648442,0.745868490525915,
 0.09236299201234799,0.6253575343685315,0.7142722866439447,
 0.47454055852794963,0.6866992644307716,0.16417157329830923,
 0.36024235683349515,0.03321204334372707,0.39947808362236126,
 0.1290096273458735,0.1659036768557487,0.050741164628885005,
 0.07159692768101344,0.026647149447318497,0.9304491727452995,
 0.429482224883442,0.7761583272555468,0.9650221606271548,0.8562050729933793,
 0.3368865032233659,0.7470824580135237,0.7813195319542645,0.6191668807183934,
 0.32296554898644614,0.21940529142411336,0.5949791516200289,
 0.16394121863047373,0.4882155093649765,0.37160703254670713,
 0.08634645066615387,0.9770899251538422,0.08263076411279213,
 0.3591289899167591,0.9115072252243972,0.973975724258691,0.21152176944635392,
 0.022313094230469455,0.241493651018012,0.6042029698536336,0.1861009114920783,
 0.179838238242914,0.3636060677840167,0.6881972939630938,0.9082075675258616,
 0.6821139084614802,0.0298119650212344,0.0018998235131248986,
 0.5889577545644871,0.037493516412582806,0.03753734897078698,
 0.194266297565305,0.7727980797022393,0.22525878183987225,0.4016147394059226,
 0.8109518915939286,0.512838275500368,0.08870355294762067,0.06937731351315912,
 0.08586633713734149,0.2214483504738024,0.7702630131377423,
 0.13607236233966857,0.06196630996351138,0.9181179378481428,
 0.3472846116080981,0.30359660908259156,0.2537862363919363,0.8392346029849669,
 0.5986414681238007,0.9747658445278242,0.5126937078519382,0.19444977889512916,
 0.047148018501813715,0.5719579624082993,0.8064623688682312,
 0.6364756191665072,0.34041927090338575,0.3196700321458559,0.4318510058273217,
 0.9640967076569124,0.7148459277988588,0.46179434046684753,
 0.22943052564728084,0.09635672992094146,0.5721514054851772,
 0.12167698348273626,0.010698783642762999,0.9122041053280857,
 0.9108541548645894,0.17867247022002564,0.3824052925032252,0.1904139241544025,
 0.8196836453446018,0.7142688916198804,0.06516129170217999,
 0.44804863812559126,0.8425438462417821,0.02145269031736352,
 0.13675247212616437,0.7494837430460319,0.42189743984585526,
 0.21248728273489215,0.6272465929860765,0.9398044545490285,0.4147870150354006,
 0.6033541827974553,0.21533373582588677,0.395672755821038,0.8422962003517763,
 0.9410448884685043,0.6273298444323657,0.12959140538065528,0.4441511744713056,
 0.43191691503923935,0.20005854433900472,0.33638963667606325,
 0.8894822892285392,0.8197633789475451,0.21163586181766736,0.7617768966325583,
 0.4292398110993816,0.5812813966580705,0.33757828738862816,
 0.20835182875294533,0.15425423609980216,0.34557009541270944,
 0.8255110602647845,0.4175444235493022,0.5802160828170776,0.4247616971603564,
 0.6198612231592355,0.6632230284962588,0.43995623741870715,0.8006298392073046,
 0.263453995938159,0.616050787052597,0.6418269985868413,0.9507091053763237,
 0.16418684061458622,0.011153275246075789,0.059372184819042606,
 0.5921623847282,0.3566162443713706,0.2036192436626083,0.6574973665052168,
 0.3076647605155033,0.7055313747241158,0.9905082313027533,0.20772751924925192,
 0.09015332668499654,0.6175366123574649,0.5073593702731094,0.6468034824258373,
 0.41850686458811515,0.12418494130095559,0.5347042183211296,
 0.18240963093893936,0.5491058369272248,0.8840577430224422,0.8156082801165507,
 0.9544516743848453,0.7293531230794178,0.6295776065571114,0.4793694461829744,
 0.9359338424615884,0.702300422482218,0.9400982119831973,0.20704689227881057,
 0.5505001804735732,0.7580164376792355,0.9251819720543026,0.4374277026000548,
 0.8746326463469128,0.004117778664466609,0.7243008398344916,
 0.6552415502340001,0.23697718698321768,0.6018362362242536,0.4740556590383451,
 0.511307025872019,0.5258423213540829,0.28430883918475924,0.534827812257952,
 0.145854864190177,0.8116575690112461,0.7597529881676308,0.5614144086777423,
 0.14125498927070557,0.9144019539500355,0.01818486822136034,
 0.11379566020364851,0.8702136237308857,0.7417128567488415,
 0.15247546381719423,0.927178915107451,0.7117166046324838,0.9043879121480023,
 0.38202456975438703,0.9471219208246682,0.4788613592996288,
 0.040795426965979376,0.14239600164757538,0.6703711761270233,
 0.050777038694557075,0.6219185661004207,0.8359065293025966,
 0.17557956306147204,0.4040466551903654,0.6125099523532573,
 0.11381383996069383,0.36345106531017524,0.8290384076584792,
 0.5673111351206402,0.5005635494209009,0.22226942686500462,0.6225476739338615,
 0.4868387720293246,0.9957570136232583,0.20797578766522185,
 0.14279645189320977,0.9194806556544938,0.6678427468248753,0.5331772160422216,
 0.5030090432495432,0.16539600647567387,0.23903626026567426,
 0.06673111779242513,0.8611184172414506,0.7460594457619923,0.2550679074803128,
 0.8828584033228485,0.7617151867275038,0.34483110042227627,0.5719518890780411,
 0.8619291239843445,0.44822048264559344,0.8790601745028183,0.2262231254348417,
 0.020674813015859272,0.33093652255593786,0.6475034190145096,
 0.1684629045253747,0.9734203400733261,0.8862708878022901,0.2135622534860947,
 0.5358068555538608,0.8617495721895012,0.576190162437294,0.937670811118507,
 0.28404878052396865,0.009262484001850702,0.6443926355434682,
 0.8097402260635653,0.2548136082892808,0.24973812910546866,0.6902875665075401,
 0.33314886750887474,0.6300020142915309,0.023898645791290862,
 0.6098864598469015,0.5632237505853567,0.6388249373076005,0.6559545407922022,
 0.6150641966121352,0.24180824050995087,0.1416277271904811,0.5890336657405121,
 0.08410507142610446,0.7958059783203649,0.7293633924911771,
 0.13508984500652166,0.12947548981861767,0.5384717254891894,
 0.4760085781139458,0.1170247202864444,0.019002370941329394,
 0.2589974258334786,0.3929451002460853,0.09166474595532426,
 0.45623949311926315,0.3002917063005013,0.8432744013805664,0.6174161628886746,
 0.23506743443687728,0.09483323716362013,0.8403567983120191,
 0.8786749489950622,0.14560323385393936,0.326519119346634,0.7445552373634113,
 0.7873399389408414,0.7058330116431473,0.7252994168013763,0.3938232399105107,
 0.5653305911872971,0.0930239753577281,0.8417159312312626,0.12885510343850248,
 0.37729221355425546,0.7880408967188351,0.37508373938578077,
 0.5298977871519682,0.10225568377206828,0.07451411892364035,
 0.28455303794420606,0.6898249023705623,0.02989402746763503,
 0.3578823111595988,0.23651143647606054,0.3452352508578195,0.4196760321858306,
 0.663851375225309,0.15490818489408265,0.0861002065990829,0.8477649947067537,
 0.49263492010304266,0.961367036687761,0.6202685488835836,0.39245401026361415,
 0.3108339177154271,0.7361431573397814,0.13962209309213158,0.7309815894268955,
 0.8916345401109893,0.8222761537637925,0.8040689825074752,0.597374818202446,
 0.4333918427059821,0.7147563985455491,0.043510803095653805,0.786351845983573,
 0.2945335165776434,0.8947368154491904,0.8182528069550428,0.0893268680742123,
 0.7068462778891703,0.7577904335014918,0.5887988428436227,0.08670996426216593,
 0.7615504920755326,0.43375086596941226,0.5442601653862587,0.3876283994930636,
 0.14749912965189005,0.04548732411338541,0.8747785128957646,
 0.28593604521720506,0.6328123991195851,0.9256485355132749,
 0.08618350505987571,0.8531746052518805,0.7632906307028722,0.6147500288461755,
 0.5865574833230028,0.36552433983155164,0.11491560783885701,
 0.6862682600252924,0.11567035986062602,0.35338680275652146,
 0.6928238704065309,0.007425093242598213,0.15053442428799757,0.93076307656283,
 0.7486409215870161,0.8315355340415261,0.889671695499108,0.77720337885343,
 0.7010330526266252,0.13369500887503194,0.11604261415861883,
 0.9747148692114653,0.7005910898344881,0.6539889580860319,0.3198029005814058,
 0.13008062017370903,0.92697406804433,0.5649449931568353,0.7236198084755207,
 0.1634517367017715,0.8317288553706627,0.391372986656078,0.35009456060820676,
 0.36731354903921287,0.42095893072988655,0.424136274190011,0.7090800735700876,
 0.14654531083638733,0.2449533529994008,0.7378130246617971,0.1607071574547979,
 0.8763896969630208,0.269664783698194,0.8264565268745123,0.7052366057323636,
 0.37329259534399495,0.5979406263516133,0.8297326078909109,0.9774851037662752,
 0.352310636648276,0.2917728030653741,0.049569377247463464,0.6325974705321069,
 0.5804670714733422,0.08510518628262276,0.6606253776844936,
 0.15693898446380394,0.5177843384490819,0.3179852778911266,0.621889202607715,
 0.3923480747540711,0.10597611812790331,0.33291146481365486,
 0.8567712304079577,0.3241365649096909,0.5922286115826384,0.5218530944848918,
 0.5610962069262742,0.19220942185766143,0.10163866189535231,
 0.48109252163748917,0.7610814620539905,0.49088257949602954,
 0.47390111160695714,0.05115456360305892,0.7625464246327841,
 0.7766502886426911,0.6017772632185906,0.2816183585655395,0.6468525976449391,
 0.4939459647965475,0.7710639538085393,0.2656094192275841,0.578744874516618,
 0.9698910095357041,0.7978850366595392,0.17507147226427078,0.9648679252955246,
 0.10990526831894076,0.6478044172680562,0.4325400786909994,0.0511431210776897,
 0.5324166655012137,0.7552176328905045,0.8088962568336548,0.7342666241991708,
 0.19617781357136732,0.5280228059452442,0.8257790472544201,0.3542935290645126,
 0.053081134631490556,0.8062676443948459,0.6098081354486377,
 0.9813919954848999,0.6942156885705619,0.06273934714781981,0.67971201768027,
 0.6293531586753457,0.4861868443253119,0.13372839123311253,0.5606443986547015,
 0.6872148627196155,0.4473261180769188,0.12818159173317578,0.7333113118233144,
 0.5789266244945057,0.5909621405874677,0.9457297075758726,0.5351990460449503,
 0.8400235396705162,0.699419333295779,0.9317644952974661,0.7394658223334095,
 0.6822515761936913,0.9855283062060148,0.89523400539822,0.5189092802511824,
 0.12499332483690573,0.8615349660945568,0.2975106118583073,0.4989801487770066,
 0.5017207471902676,0.46596816707714506,0.7571091123585997,
 0.04018150581341884,0.5390710428558645,0.6780272309188017,0.7287272538418763,
 0.3151204905416135,0.5311344880365912,0.6524969056228627,0.7528400624466605,
 0.15265287884157974,0.8787228226715736,0.9964803020599278,0.4351123551352569,
 0.1002712865284674,0.17217370857374592,0.41836621486247394,
 0.7007792131671611,0.6369267032354315,0.6775205186407858,0.0775601336039986,
 0.4070214207359728,0.7833731732080458,0.2838416178253975,0.92207390945419,
 0.4029017380525465,0.714165370917159,0.8481653327070974,0.7445109897183063,
 0.5662674071782275,0.7685637718002697,0.28363027864384904,0.8970872644540683,
 0.37197621485531673,0.004207051415225149,0.03798942888383694,
 0.5972753204388832,0.9031652384482709,0.48305810717926934,0.6029795248711698,
 0.9128711232263712,0.34510823705715044,0.5851462446306873,0.9843813222821775,
 0.9869740314814965,0.12925702851155463,0.1480285591697016,0.5125866634314665,
 0.43596900770747804,0.09802126505100728,0.790382040942048,0.9028974077677113,
 0.10906358125096804,0.8825183616628375,0.7472956479311421,0.8059833380406789,
 0.14897285725619314,0.9384915674828154,0.3496593407825045,0.5294385326750836,
 0.2725350357766525,0.4943478063511829,0.1976076402283451,0.6370674996339053,
 0.5019999064012357,0.01820203603636883,0.981670405006532,0.5835808896237584,
 0.8272250934770623,0.4804041596198181,0.18818580856685752,
 0.17343521874967516,0.3862142419207585,0.2643265632615005,
 0.24123652992772482,0.6319484978407354,0.2111800438740561,0.5063698366870464,
 0.8955247452421317,0.5545318338042734,0.24763991983957412,0.6925703925742381,
 0.2994560124566753,0.2119090989099276,0.5642442861788886,0.24935276577009324,
 0.6821494667193266,0.6773575230418695,0.8984979747213114,0.7396864032424101,
 0.9860623987104361,0.6428288119276975,0.432186087884344,0.27777374430935586,
 0.3202834030468177,0.2724163999445961,0.7725736852363743,0.4032749297561268,
 0.09348810705934196,0.2525862757646039,0.9698132778799206,0.5362727617347025,
 0.6298789960830742,0.5177923060564615,0.9952918298968749,
 0.048606470887210484,0.33967493144204486,0.8503626582992189,
 0.490375978925918,0.17049035111581745,0.7346839867106911,0.8794639137887321,
 0.5808245220690096,0.7901300290078006,0.6381909414893732,0.489364788950691,
 0.214967377025992,0.010388291157925278,0.7085825892275169,0.4517193174864792,
 0.12383060851372907,0.9123537443095178,0.4141606508321186,0.642373024968403,
 0.8977931256848002,0.788838049031562,0.470095208528577,0.15766752365430758,
 0.8171915514006982,0.20106352511961645,0.97604564521343,0.12859504479280248,
 0.1340410650248649,0.8372192746876091,0.9581300761339827,0.29494386107247306,
 0.6020318020886501,0.8397370425452099,0.24385815985872494,0.7832064730286874,
 0.3987962144611672,0.5380335468847898,0.5839431062442857,0.3158962987715488,
 0.22347072783280697,0.865277135457267,0.7134495948653222,0.31627596333535746,
 0.3926576360593663,0.7593966110071514,0.26386752719953543,0.4705237008875175,
 0.3955385137811148,0.820118087972752,0.4974030333821745,0.6863002284446175,
 0.18847390386867025,0.988896555985091,0.5934559668380917,0.01723582942246049,
 0.5991564154210082,0.20620286969748358,0.3466708654181885,0.7476188120998454,
 0.2488707539889926,0.546875934900234,0.030717010194024952,0.4401065194518611,
 0.3550945069224253,0.19012650563048672,0.15739003436347687,
 0.048437954470236244,0.2903044378763163,0.7894072165779262,
 0.2810670710970614,0.7854954553860564,0.3346698182208181,0.2902291639147493,
 0.07617029291660027,0.5536758895084702,0.8850561404887543,0.6582965379783348,
 0.5668903863234686,0.46981787443911627,0.13436051754829414,
 0.21878876768605993,0.9334344931384879,0.6341469481159144,0.9795056029542517,
 0.012173585747045967,0.45546216116326077,0.6977723706542099,
 0.2332210121326701,0.646203520088024,0.4489071190323438,0.4497319521611163,
 0.08684106588162566,0.7155565887152062,0.6227791610126037,0.6294146816462984,
 0.21245973345642255,0.8422613138618555,0.21436111238894573,
 0.23129335482840585,0.004261659411974983,0.13934892225845608,
 0.33426974576255575,0.8966079967091467,0.699516786757565,0.5324222832787218,
 0.3861478473218001,0.6512358332785856,0.9894220676908408,0.06919187666105187,
 0.3203629462826435,0.24050144794796235,0.5352950586888157,
 0.01825437520907991,0.665086522061786,0.784749876352356,0.6844312894085707,
 0.518791147440354,0.5022872389113413,0.0687427436410315,0.27795976317445437,
 0.6649581711262017,0.21391812314657732,0.7423080657438932,
 0.21797101320033485,0.8911553276508654,0.27135037317572164,
 0.8167203558627292,0.5266978564462855,0.4445309085156143,0.5071704028546996,
 0.4586767217640966,0.9145054835407544,0.7939017924635843,0.07675443752496613,
 0.13614106902549516,0.7282001549501396,0.5448065481633467,0.6648733045570405,
 0.4712567940201805,0.5625293741097045,0.2639300891240136,0.39435423567378425,
 0.6064909127796779,0.11429157692309155,0.4332338921585067,
 0.03189225515868943,0.9380797462500803,0.9316863214488963,
 0.09201281896888891,0.20185263784918928,0.45873478750016083,
 0.038676296937846244,0.07889968615181309,0.6632710714899244,
 0.5138725002229193,0.12491500888496498,0.2392379174865943,0.7589905669169057,
 0.8974078412463798,0.9776703714372803,0.13814373258602708,0.2320552335074213,
 0.5260041306944362,0.9506323276287553,0.614633284647581,0.0804513747662654,
 0.4015379678796127,0.5727202269785948,0.5846387455705202,0.8642065703481994,
 0.34653609490541737,0.301807753937687,0.8736346954181066,0.7124714823826425,
 0.17541744012954108,0.8042828381738664,0.6086199782450442,0.8918592875661691,
 0.5725597228139763,0.4341551217866847,0.11008478948901057,0.9238235065019569,
 0.8996059495694099,0.07645952020494418,0.6679166368482643,0.4845542684010393,
 0.7839399457323021,0.5195566042291365,0.5122970948513406,0.6757892565023871,
 0.34740538075600313,0.24753265209446118,0.3865151093626612,
 0.7430523437486809,0.9893882741078884,0.9265479120863945,0.4649967256690969,
 0.8026060188250804,0.8892926868417994,0.5330155762909712,0.09729134082837676,
 0.4713083835034819,0.44386478854171596,0.18646907878262398,
 0.20428977425107175,0.29503106891041875,0.9785781844701211,0.96491319740731,
 0.8975369547684366,0.5204805893295886,0.7871437811264599,0.7393531687052183,
 0.7812521864106973,0.7215656072340618,0.6418893343175981,0.49925590888462945,
 0.0736011185108637,0.521313580544154,0.4521202687658552,0.3869384089477954,
 0.4092128108046209,0.6103723596684524,0.934604481110785,0.7643504386621275,
 0.9478263431059272,0.9917190791021909,0.09280329656336461,0.7974398262023925,
 0.45123569192176627,0.6059108069182353,0.4262060821793747,
 0.44819992217848525,0.3978480071538364,0.44896957437006413,
 0.07544204167090718,0.5140321463459132,0.3935248055574101,0.8868108752278914,
 0.3358561682505845,0.5520890726658773,0.6090139036268731,0.7713961852343016,
 0.7317766594090175,0.46131639315452566,0.7002518237595065,0.7472972298997695,
 0.7712331913354649,0.3606024424766723,0.8928890933811212,0.5915908687793249,
 0.5243508434130896,0.8669233202994143,0.2901631787734039,0.9017623911054572,
 0.34774167645319587,0.6833307603984506,0.8878736494691444,0.8515154988302226,
 0.5401483444924938,0.22944095235258066,0.37277863702296266,
 0.32492729740824455,0.9502647093370543,0.8767814318721971,0.3955410520993624,
 0.382321143459213,0.47126126410605707,0.23616421425017498,
 0.39819288509682227,0.16480759738676787,0.794220210097095,
 0.19013273729204427,0.3123292810049193,0.9030981350013694,0.3482830821890561,
 0.7446920635503217,0.5432737882536732,1.2412185683474686e-4,
 0.34525447493768546,0.1617794875043952,0.6740145408018026,0.1449323654236061,
 0.8287474076094283,0.3114695793014265,0.6050336718352015,0.9434757500244019,
 0.10990127141639916,0.03574821830380048,0.253176718703336,0.1224618003644472,
 0.266391749876965,0.061527223940284426,0.884628733938565,0.15406961592505286,
 0.7602028871425741,0.9147468688341608,0.6891409812471508,0.6323914134316198,
 0.3151812986934487,0.4000606448170738,0.3768104714661187,0.3716489729517578,
 0.8317551470396374,0.5794837763374676,0.5092989785180237,0.3228383138879205,
 0.1955710514261475,0.4333432728796035,0.5844732437019788,0.7080406446748551,
 0.3531531780875503,0.6816996577652848,0.16475036947490307,
 0.13876222148851358,0.0036572810896992047,0.4612530960835065,
 0.241818622889719,0.6774527968046995,7.49632875420625e-4,0.9073627111837183,
 0.35557247806213677,0.7784888171415946,0.35080322965252686,
 0.27810799126307606,0.9408383137672722,0.7669755764856565,0.5748975285180674,
 0.8285851518454411,0.3517241669909923,0.6348109564393962,0.3823920446889194,
 0.11236210093483923,0.23449940310179074,0.4496167010406198,
 0.3295906951932811,0.5720382903098888,0.7684956224384479,0.5667185077313278,
 0.13670788738273398,0.5490301545854587,0.700781819003944,0.19851596115340342,
 0.6694130283562869,0.7393391735621554,0.8686946770407115,0.20064157743026367,
 0.36898158929242164,0.39967061433154916,0.4161074982452653,
 0.6866322625605898,0.23347461154594684,0.8091135110718145,
 0.13008278770622428,0.16458952732710364,0.3640786778359485,
 0.9644205227800164,0.08930300753104004,0.9561481169656404,0.6708758107551895,
 0.856493378524979,0.5193302899922521,0.7842039622423835,0.7595080629337483,
 0.35596889299523904,0.6237814277229292,0.6583412469314972,
 0.24233309932558067,0.9176775162089025,0.7849623481764273,0.3871784044740253,
 0.6641409698232493,0.28584706869265186,0.5944748485505524,0.351780115075744,
 0.3162259527073936,0.08075384560923493,0.7367038233298142,
 0.033184782450775385,0.48253316038844285,0.761114938155961,
 0.22041302331902324,0.3361119500270421,0.6587544189647482,0.6310762424408899,
 0.2783278204116981,0.9052976921229398,0.2238142565503759,0.8816465343870539,
 0.6328729185599993,0.9229125572303345,0.010954711853452204,
 0.4020493078456505,0.742881449351893,0.3141646400817094,0.38998951741011034,
 0.37507497524094047,0.6060595648883723,0.5977050802595765,
 0.04002248717843493,0.5406595517144126,0.08924493054122817,
 0.27883856302573695,0.2456765734180052,0.07370208350180274,
 0.5411561800960576,0.6801543933463547,0.6650583695303631,0.25217366742197356,
 0.8911761800041071,0.5322342107471995,0.372160617618403,0.2469344697850192,
 0.9883878691850781,0.49641261032490014,0.38266524654347844,
 0.05938800384149878,0.06308955464058763,0.10518885001935407,
 0.12173128587183135,0.06591564361839441,0.7865079256114378,
 0.43127094387454223,0.32587797656505424,0.979047256433602,0.1521712372276861,
 0.2269058637832012,0.27258288557738863,0.5139519822688876,0.549042827408666,
 0.8392883645211805,0.9410842176463261,0.4595929634164537,0.35898652376276274,
 0.9195832931026685,0.5078672099173743,0.5009237693142229,0.9441060612530139,
 0.7292210275844494,0.42923275574176956,0.07660189087280678,0.558173232110309,
 0.7277638942069127,0.2976721349549186,0.20111899373050912,
 0.42544856699274725,0.09979604216885807,0.26349785960516336,
 0.24957551881044693,0.5038936751834499,0.7839591690900803,0.9565776618241214,
 0.8193468980474754,0.5163323496228783,0.2625516222019906,0.8275666632767257,
 0.2965864637902065,0.7825455291259228,0.4249833409152618,0.2804590194808455,
 0.07585174513947357,0.6442869671438111,0.9833267784793834,0.854830711061969,
 0.33139383420146296,0.2185408629438923,0.8850814589301155,0.9008449488244731,
 0.18462135170630056,0.08902614424925404,0.4983852505205294,
 0.4864043310577669,0.08508975405967811,0.8569324297222112,
 0.37892704800500143,0.36447256658191884,0.6289792817339167,
 0.45326607141377595,0.482687229286761,0.3546837588458107,0.19162663520564893,
 0.03128169480970899,0.08393587613907216,0.61062345430785,0.30353394769039366,
 0.4845901375003139,0.6552021993449473,0.99308772836033,0.063240453927778,
 0.3250988691584368,0.318740110619672,0.3909795901379587,0.10531264225060788,
 0.5511354006711393,0.08680046506020944,0.9782428559747893,0.4679209765632628,
 0.044620901433925386,0.938432222754121,0.7905247791959026,0.7339796412018256,
 0.06658749361395111,0.2705740805557173,0.15817307139144765,
 0.9978387806872752,0.4794267133470246,0.2265242138116731,0.7675997468294963,
 0.6631901275300521,0.45792815636017326,0.7843749921176355,0.5995472299056206,
 0.3275921328370085,0.17499173481247954,0.4426384772212706,0.9872007418065729,
 0.5511091506359073,0.9006007770556844,0.08755913268508753,0.3070381247260455,
 0.231715797637831,0.3989556789841573,0.39590417669925215,0.9739369859096825,
 0.9646938300611847,0.8354301930661509,0.05707703356978111,0.5008322763618174,
 0.43761934509878975,0.2086010474812382,0.40571425226078195,
 0.1593767676206833,0.024379487967897018,0.7864590669987263,0.903214014972022,
 0.8236066646662912,0.6390495959352966,0.366316904949727,0.9014188491252312,
 0.22690792170357388,0.709174925430215,0.32710953568336765,0.4987440743137572,
 0.9763140805369867,0.3466709330076311,0.35621312969786467,0.6250818432619801,
 0.6804416735378552,0.6525219152508671,0.39165060800814167,0.6039283303923405,
 0.9862435474489457,0.9785468942509806,0.6041634391696497,0.7896121978430879,
 0.7469733741577435,0.7146827193555206,0.7378267690013072,0.5268611639593986,
 0.8766859854374108,0.8784293045984473,0.9670246668284022,0.9518879513296641,
 0.5693238177113988,0.7753090709950683,0.3944487767289837,0.7140272205914612,
 0.3590465798519271,0.7398314318813302,0.39533845019856284,0.9168029838263153,
 0.9872809321107442,0.577294262511379,0.007565541739335657,0.2779993272036513,
 0.2517751245925276,0.44327355147633674,0.32529783185416483,
 0.24332356082334183,0.5251269011399998,0.44169033889550846,
 0.7516889532957028,0.7630066098182103,0.3840507079839721,0.8750898865836344,
 0.26718932031191345,0.4177299069741611,0.37351716607252683,
 0.8521627188174791,0.9096522459355391,0.9830073850193028,
 0.049954489288867365,0.730970194664051,0.09945529826185129,
 0.5045180862784941,0.5037576421148955,0.7071321422024007,0.3782492743897762,
 0.5546009691830751,0.8667417607155304,0.5554542925091054,0.3408653297979328,
 0.7949631246016242,0.16429446271187131,0.09824011779670672,
 0.4708216114070616,0.48873833999007643,0.14790721043480515,
 0.10907545350919234,0.21930103547457813,0.8493988525160621,
 0.2683832555578518,0.7679204190843141,0.059268212920265295,
 0.9642496056556867,0.5832007120780847,0.6938950960174564,0.7851021409738133,
 0.9766525106988355,0.26652545847709597,0.34290116750911337,
 0.21216974204569017,0.2911485034700323,0.24431442793564806,
 0.33959936910112876,0.8433163577955212,0.31758348527349545,
 0.4557277079816203,0.0201970368777048,0.3677864854772753,0.6275209792530891,
 0.49050252337912315,0.8450237668701166,0.04591367953286496,
 0.4010956589486888,0.9076611930186389,0.4539911404577798,0.587363264025471,
 0.004806787994701467,0.20175864514919073,0.8616144177007018,
 0.19892220163257313,0.24774953597945693,0.07639714442420986,
 0.8866400308216832,0.870470622471957,0.8030196103751257,0.7685059495277313,
 0.9243466191294089,0.25823461911474743,0.48130924777135065,
 0.7215190727095697,0.7532515727639955,0.3015315637068451,0.5547860226555882,
 0.03930233869124011,0.01616205223827294,0.6226512760850043,
 0.8383419478549282,0.6392240467293077,0.9713982933821876,0.376467177466279,
 0.6997298535124984,0.3058621372041527,0.48423325902689207,0.2954700195538844,
 0.9270926090917392,0.3972296960066706,0.3113194695843573,0.7330722937799634,
 0.35759630457915703,0.8870894665422603,0.7852893809294852,
 0.44382123024010867,0.9066800846165753,0.5840059962163602,
 0.0044365874346873335,0.1758598355345562,0.13549566706211036,
 0.4827385221414956,0.5969379641871406,0.31291588584726227,0.4380913349585066,
 0.2953913581866061,0.20114340283392862,0.912523300781296,0.9369106869630061,
 0.42041685556066133,0.5078958140619878,0.2396570364089614,0.9094093775273351,
 0.12195727018989355,0.4832988826569413,0.4046191370141726,
 0.08338649319897795,0.4738960158614638,0.636354259462986,0.20149437234740364,
 0.4577830048435235,0.6783074732676656,0.07738006040888368,0.796491447075337,
 0.590939477540337,0.42059368772664396,0.22644986947236234,0.7950304329240745,
 0.8977855067775377,0.8616644318514265,0.047281760640376036,
 0.4746266450863744,0.41756527117763675,0.4454196086152937,
 7.305829549686393e-5,0.018310322142352287,0.06704813156834322,
 0.7191652257438202,0.20099599310049765,0.032774715213248706,
 0.3858357116517287,0.5301139163808152,0.912723745740698,0.8417068843872704,
 0.7807404287781042,0.37413131485311835,0.012627953196123087,
 0.9221089197012104,0.20426830385524464,0.2965846965078519,0.3553597168962601,
 0.30144889625580096,0.8368982475912601,0.837840780240911,0.3043193648058551,
 0.185751693335934,0.24133225561901872,0.5190780754605056,0.4728575541495277,
 0.08170110336016234,0.5596989884708609,0.15230204281556514,
 0.5297922865585963,0.2701548659678137,0.3400593513371182,0.5776259933833497,
 0.28883973459683976,0.07579472145037969,0.0970114292774531,
 0.21892140107541913,0.3409581140839323,0.5264221766267374,
 0.08743439119830809,0.029536112221753785,0.7977144382612017,
 0.652252779970953,0.9807765739787274,0.6702139317545753,0.09022237501715547,
 0.6325784704463469,0.9551570404933893,0.10129675845647246,
 0.07243157486001184,0.9102680796833041,0.11983603858341829,
 0.006845874284755582,0.6165898541616607,0.8382554436648927,
 0.5865059205904813,0.024927767523821864,0.311209802561087,0.7103296350739673,
 0.34311689662351397,0.5315829802853678,0.1030521779014103,0.7447829319065489,
 0.5241956276478215,0.21725274797129912,0.6791499710576224,0.9483087703738345,
 0.5748639139207703,0.06022704931263223,0.11983023430795092,
 0.7752225108722838,0.3604396242447626,0.2276321576993614,0.5468507304950303,
 0.3948634826408919,0.9432768565065548,0.620227382943695,0.3534261198718808,
 0.3347648245727771,0.3456070021308858,0.28431229057433427,0.4612665439161676,
 0.7892350734954575,0.3316848460307351,0.8278812989720377,0.7218010311482728,
 0.8097443853791537,0.7647693685154965,0.8479970702155598,0.7654483111701961,
 0.21573285032164768,0.593402005541882,0.42622595810605013,0.6600488246067374,
 0.2701927569289573,0.34411570037922234,0.48857516467291473,
 0.47726826234218955,0.01673129151833974,0.9437384439287829,
 0.8799099858289386,0.9797935797620028,0.7449447898006293,0.4566390384831299,
 0.448142710307766,0.011236720526676458,0.8635735998074762,0.3582410110597236,
 0.8743378495685256,0.14252613453600105,0.07133151311023445,
 0.9849636713190077,0.7657853218119648,0.5391010237223623,0.9319669593310534,
 0.8878344924855924,0.5643206263773991,0.6627967807541916,0.6791544426176335,
 0.007541646191403961,0.2606499225086666,0.9663334933097116,0.614920750125506,
 0.2925526292528544,0.5343831703271686,0.23683416871647112,
 0.06916966318879902,0.597890691703692,0.9919400157216369,0.8258573997629342,
 0.36061151451840145,0.8902857196213929,0.7925062288114796,0.4638442939512444,
 0.7571322759724095,0.38474415487433666,0.09276399166349747,
 0.7123596747254889,0.7676941474622424,0.2849133699202656,0.7692703672879923,
 0.028722175580337406,0.026338188625287806,0.08330744114594335,
 0.5265389450428059,0.25976070031656406,0.2169965235844702,0.8439549715967426,
 0.9762739563948813,0.4632130627144757,0.05575044681638275,
 0.40883283104737655,0.4938943419517232,0.7143581678432692,0.8342967468488053,
 0.12812477169685432,0.559717672097273,0.7302798746724788,0.7897251576717584,
 0.7162126375023077,0.4918118217468709,0.42088638389118715,0.8031719745886868,
 0.3499880142101802,0.6769735178779028,0.1030916501995105,0.6455235336388241,
 0.23289800032779517,0.3930102437323383,0.29223320782427287,
 0.18110962508257678,0.3800946668457684,0.34287787047145213,
 0.9690949657429602,0.31601276915439613,0.9624670675282387,0.8893768562048259,
 0.20803935595473677,0.6232076503476626,0.42135529896391166,
 0.7615903131701656,0.9849873686445539,0.1349101939241003,0.7685031716099243,
 0.652802098386774,0.5023098169949402,0.21398566374489303,0.10024028789347783,
 0.5320582433459056,0.20109578309357345,0.30284162783414625,
 0.9458219898132031,0.4463409995644908,0.782024557195958,0.34835332394770124,
 0.8275417056215021,0.154824974655437,0.5567842075998737,0.5486252462069063,
 0.28045143047013055,0.4168980718068207,0.6065024985394776,0.6891416264873708,
 0.22534708173896723,0.6872683591733487,0.9418227517112207,0.5606939083612493,
 0.6987412464830627,0.245501633066613,0.2897800583023302,0.4611165000367896,
 0.9905134225239454,0.5300374911783133,0.1579508144694587,0.91298446351175,
 0.6248203235260181,0.6770947380561673,0.49683931997966135,0.5134802596218112,
 0.6975571013454783,0.9977009923265571,0.37045946291959364,0.3121961333832516,
 0.33529809461995974,0.11870423111052897,0.9312427930453174,
 0.9247728493018316,0.7353988794198055,0.8957810305723147,0.5259183876672636,
 0.5593478860401739,0.5827684358597439,0.8336562623798514,0.26148013637157486,
 0.5393815929592499,0.589994332470634,0.4816711200327176,0.5594224835521933,
 0.8478168554478636,0.8073117805784717,0.7837771662614124,0.6930440932347481,
 0.3243382057814028,0.2887830230968921,0.25150983768818524,
 0.47431592196143635,0.9117266180856174,0.015780837986848928,
 0.6180083524573733,0.5935175829341305,0.5165472612897495,0.6925539185852694,
 0.18572028441951516,0.08580134422625807,0.3050093897802386,
 0.6236773294342336,0.5835873420080923,0.6274762841974854,0.37668958869114455,
 0.6384645533877363,0.07533721199154242,0.07225926029929541,
 0.18544886775702962,0.39648902144004894,0.8204405128753233,
 0.9341049883856852,0.29365361288643654,0.03864189180397237,
 0.5653996799915194,0.6996681741178191,0.9456021030864042,
 0.026270682496994047,0.8136680858464505,0.1327286100288213,
 0.9291977082160877,0.5668048192374977,0.8286140686834877,0.7434209303120376,
 0.9432128927447684,0.861064633698109,0.24932883526585514,
 0.021027629063425368,0.04555701387096078,0.15467599662027998,
 0.7897125736619217,0.7370288274168992,0.18228416050960994,0.9523352462717809,
 0.8099603918300196,0.3983467391757587,0.7043338798069604,0.6387696987666135,
 0.6207698110263207,0.6092398238841709,0.9005172993157413,0.18444436858289182,
 0.621115594458601,0.749412222529905,0.3344203394419889,0.3346733259439061,
 0.4798194521108943,0.3442329780589628,0.41127318834308124,
 0.30992074524383384,0.8673502211449566,0.6308304603719028,0.5257813013024355,
 0.8929375669810851,0.9658652171401202,0.1418341005240643,0.897411032981616,
 0.9457905443459297,0.24420666007865122,0.14463944875932944,
 0.27687010626527264,0.545193986760472,0.5678700464465425,0.05574948124334844,
 0.37552908974357113,0.5828340699983486,0.4297391670899462,0.4558239627622156,
 0.7290295782479692,0.34369394052047353,0.730551674439093,0.618900578935752,
 0.7055556400395397,0.7137860394736049,0.5358564546619243,0.4105268577805835,
 0.7988719287743807,0.16051054646486385,0.856157779103091,0.715095881042666,
 0.9831571896762727,0.9283337102647089,0.5604029545796175,0.8425589497393089,
 0.5004461057611151,0.536045694678772,0.0013717022044736904,
 0.11435180043289461,0.22460483530982744,0.3708596234014677,
 0.6070817604460359,0.525885412614111,0.3683652489821657,0.9425649988592664,
 0.2106911657505247,0.9588564806791031,0.9460066806313826,0.29550233523796665,
 0.1838716913544205,0.016689897860553526,0.20409354209006703,
 0.9803811579464842,0.5935808282358781,0.16086827022720895,
 0.046737320943833205,0.4681219224285498,0.8061825859237008,
 0.0769554708110689,0.5726642123533564,0.7497213254848957,0.48087500987504517,
 0.3093534074895423,0.5416176693239394,0.31576965873143903,
 0.17044667330053054,0.5780558031818844,0.5112856417705303,
 0.023622272014446555,0.503212352915801,0.34667199518786895,
 0.7967110549766369,0.4318929860251224,0.8206308880091975,0.21540720805667135,
 0.07998575907096572,0.8755774423060387,0.5671399333050453,0.8939292800905072,
 0.07166673499197085,0.005959902520737859,0.37192377566421886,
 0.07474326843343027,0.43773982237134124,0.8051469428063603,
 0.5364742234205683,0.0290217800031497,0.3054993372904362,0.29017196118666044,
 0.773911744543148,0.16778316034587182,0.2563658468786546,0.8030967689102724,
 0.6911509387798666,0.3723834600394845,0.8060401872833987,0.1694259409277371,
 0.07950274584096162,0.3563772353698178,0.9145526813631522,0.2371262340840128,
 0.716861977675265,0.1250948362966846,0.34498804357573576,0.8139992310570568,
 0.22242857714974962,0.9208726453936437,0.9164646818579307,0.0396131298757294,
 0.10961390145292904,0.42648168360277483,0.569566733438996,0.1679543258923163,
 0.16151840284958252,0.1745908164333031,0.84975175795337,0.6820027431583108,
 0.8369382564980173,0.5869035377598489,0.013154071774915899,
 0.9835234085851762,0.93955547659892,0.6726664795121429,0.5442229013822094,
 0.13491309671630447,0.29855883652801685,0.1582365708515514,
 0.24348649715049797,0.9708532898593143,0.8803296630002155,
 0.04280101682117787,0.75342816504224,0.023916408319948235,0.0692246014624116,
 0.410827401674819,0.033148231299289344,0.06591807951259765,0.918856931527573,
 0.9490809889802951,0.7970999928666644,0.8457177681143321,0.3082407078597913,
 0.55214270988708,0.735158691015203,0.18742050195008875,0.14795864030730044,
 0.3177458949193014,0.3740113756312563,0.09356414137745828,0.9768194767319567,
 0.07369216237750265,0.051182393944575355,0.6242782207172303,
 0.04416539643897632,0.9940784410010481,0.7999758261671216,0.2377148000892222,
 0.8906367691381806,0.08554102725908719,0.9442770712426667,0.8006405751379013,
 0.4839200153616515,0.2091072491219057,0.22103803886869933,
 0.09261544663700894,0.6433793923251575,0.509842347773615,
 0.035060909723696354,0.026626462742597945,0.08347225894479737,
 0.14780258776634547,0.7036856322812791,0.47742874014065884,
 0.6338197634824324,0.849705391350272,0.8950898735051762,0.67297650592922,
 0.2991497785939,0.008809801939939299,0.35191596921461843,0.7160107336517456,
 0.32818211970565003,0.5526664517702208,0.9573628530210532,0.9339410408434526,
 0.5514581224951512,0.9953482888243388,0.14584213916226973,0.2165815039981559,
 0.03065700987875475,0.903702464491578,0.17638237651973365,0.1670324993570278,
 0.5548322420893594,0.6960824595366053,0.35960327401531234,0.8508537164410259,
 0.9297887918786123,0.3545317924051501,0.8691633820227488,0.9165736972976941,
 0.03014480761492544,0.48158555530901337,0.13643936122603761,
 0.35415106877106906,0.1820904491795775,0.11584476866384197,
 0.8951272675039805,0.28895715161461477,0.4338143381659125,0.9455911724327226,
 0.14094698315919008,0.21628543602820727,0.40486542070263165,
 0.9418605649332559,0.7761934791106114,0.47230311392436364,
 0.30916538417996464,0.837937676971365,0.7987676669997039,0.27428228851291114,
 0.7383796653234019,0.7510665928244717,0.042638300466738555,
 0.24100056341902776,0.16678850789457966,0.3769503738171366,
 0.5545320093488575,0.8092711049324182,0.2660129607168489,0.7348427062234946,
 0.537049682980325,0.0733477061257024,0.5084804204477906,0.31246591947687974,
 0.9350012850527235,0.8150933326987705,0.07942800572137665,
 0.45028344554775246,0.05813069572090923,0.8145519818114085,
 0.6192401699956209,0.017449121416909463,0.9464679076570763,
 0.8992132749897741,0.03391875583700221,0.3310705562139964,
 0.31796046473132666,0.22305826196542422,0.3037958017194209,0.64403809404438,
 0.17317160177487778,0.32262089372108005,0.9264596469774941,
 0.3676037220803874,0.5349214539493848,0.25030864374644357,0.6571933405621757,
 0.9541757214028578,0.5226360797106815,0.7151189461187749,0.9479034076433217,
 0.7146171142096714,0.7817339223859209,0.008433513832841966,
 0.5775610132292642,0.8671905994580547,0.3472040590358201,0.33249975094496464,
 0.8134903686370073,0.5907532078859191,0.06499929793518333,0.4407883699395627,
 0.3691156155232731,0.4654727034363264,0.16739710199066127,
 0.16577390531699776,0.7474074210823409,0.26098990144413425,
 0.3448528515804733,0.21258167500685043,0.09537979766980309,
 0.9666031295457229,0.3711326945747827,0.08868947199794996,
 0.48427505805696347,0.5534015210511238,0.53844734940172,0.6107468330925222,
 0.021454135400714502,0.01742850239345861,0.15946242671958377,
 0.6389864640564622,0.08545239394880644,0.8963379956136652,0.3008863112479727,
 0.4244856606327544,0.9798684726193523,0.42297431242981054,0.6559704339196377,
 0.17972135468484562,0.5728010749968335,0.18630352786668325,
 0.28006439423661245,0.5998775271557515,0.6322009778609906,0.8447969585118561,
 0.789628997567527,0.7049173299602124,0.2362881042128232,0.10570494526418284,
 0.6466757011806941,0.359621766638617,0.669974880413599,0.5776306755994938,
 0.36620648290457003,0.30752680056418225,0.23081042021354747,
 0.840433254363862,0.4590216940261528,0.8624999095411628,0.5669830367312216,
 0.5967858000912983,0.914988818584176,0.08993856318102311,0.3743091224660645,
 0.34997166769799526,0.20652690873348933,0.6967037206028918,
 0.9770296874064411,0.5345741617589621,0.10276436870787953,
 0.20944733075668898,0.07606713655892694,0.4049084117487194,
 0.9141199790575849,0.8137465495804734,0.2881434634112523,0.28876071027166184,
 0.10869719544768841,0.5638874028768611,0.8796607163806649,0.5072721787350924,
 0.1924828533897518,0.16151090784843392,0.2993537271631852,0.6729111765731421,
 0.8452559432006505,0.775741836946591,0.6084581975589929,0.8165550891193791,
 0.5881149975112049,0.42795647371367274,0.3042662136591092,0.9939786064198157,
 0.46800047668399514,0.49624433369453713,0.5848386493006528,
 0.2316334603575948,0.6601687470565812,0.2562464648000443,0.18935781928239903,
 0.2669088835262594,0.7453071156646744,0.3976266173533779,0.37785313215491256,
 0.7075378680596175,0.7832155795306789,0.6163292630246069,0.35966774046052663,
 0.025460606495897542,0.2992059402523296,0.3427262026298902,
 0.30255223766396844,0.3928262649721932,0.9426252872497706,0.8647639862760589,
 0.11804923000767864,0.05617944287421195,0.8152019580280658,
 0.38611814910135267,0.8113597977074686,0.08626853576949067,
 0.3474323833254497,0.2110480637650246,0.4771077933768979,0.2760879436875371,
 0.6433394122068095,0.5249896925534766,0.627157779173459,0.10105520939030288,
 0.028763308660571196,0.906182998287097,0.7333723720077201,0.5947540107852415,
 0.12245439199180486,0.058319207729986666,0.9844620068662905,
 0.8347794134860357,0.3795928388604871,0.8016672821448255,0.12695670177607998,
 0.718693659399867,0.8162418513080215,0.17121902644874953,0.6978788310737155,
 0.1786774762841955,0.6461048194645008,0.054987815186867905,
 0.5844604959839197,0.23902406387807085,0.6484968162612249,0.4984788991074086,
 0.41890028717555317,0.5290430438383453,0.41037390081287584,
 0.24383728969186036,0.25961179947427926,0.9917085559791136,
 0.5175095099926235,0.8010809161461239,0.8469116853390095,0.09299403355569913,
 0.7375463469322113,0.5784797848526692,0.03528667731396273,0.2908203142628917,
 0.030716153958909476,0.46176284317800353,0.14519038709070742,
 0.5127613925128773,0.5030871081906616,0.07035020765156796,0.6499020796535799,
 0.3529701185043834,0.9459092012632532,0.7833526965378332,0.22183122780384068,
 0.9320173498291342,0.19477925567790166,0.6057096941719048,
 0.18217893938518037,0.9242516736159124,0.6729800824728808,0.564335549313381,
 0.23102475315087878,0.8256824309200539,0.10819465010136353,
 0.9347502529170773,0.22698590924044892,0.538168435964321,0.5961538753404425,
 0.3269948723333074,0.03743965879896827,0.9378142527275998,0.4223001157585352,
 0.4237911788174479,0.42323266640634294,0.31641701177822634,
 0.025973152372447905,0.7623911692738201,0.6873449493631781,
 0.38630455964356447,0.2859261058483974,0.07920894173249415,
 0.04890350750547612,0.5825327094400287,0.1864882577128809,0.7755640807513516,
 0.3579556534024917,0.6000197550627299,0.8982521353591786,0.8999998238157902,
 0.29770294385682927,0.1369570919477664,0.7249033523109917,0.7859879383140629,
 0.1025398188156994,0.19904170590867432,0.4306616614920814,0.7477523704136612,
 0.6397409799247754,0.15182741405854128,0.48917402766459483,
 0.7448702768763676,0.8998382802827188,0.10215695771086353,0.631596309882177,
 0.20752915951049478,0.38443409875598444,0.7175526802082511,0.65187242019543,
 0.9388260435133481,0.2787525254252766,0.19115314215204138,0.4837658272004759,
 0.9896096741165256,0.30994945754262926,0.3888451931895861,
 0.12050978182593608,0.8306171390562845,0.9072820636853478,0.37302435601756,
 0.9844555098393376,0.8157938379217645,0.006925790738597337,
 0.7569480552205056,0.8976145431110486,0.49505933831733784,
 0.23161317630480993,0.23266640229456326,0.2386643768009633,
 0.14142274901026108,0.09807574646100492,0.3853788333069925,
 0.7751698408496391,0.745721469437738,0.27353104506862025,0.2659504132733299,
 0.9821116183843464,0.6952923462817229,0.9050108338360909,0.30389487052945774,
 0.13307352447920595,0.010014564062051479,0.2643176143859506,
 0.37838767909036297,0.7835907562265036,0.3115939720181201,0.8563807007408963,
 0.5394521781244875,0.8181259860601002,0.9523997477555713,0.15425180177467612,
 0.6461937825926822,0.8264415418155406,0.04945642153347374,0.4429125301323983,
 0.1113175464742503,0.21216984850584386,0.2887630843063336,0.9019342097841707,
 0.5142403783215672,0.29340054677448313,0.30218454046040255,
 0.7070015420209238,0.1400698746854413,0.5137180077234823,0.5167328444586264,
 0.9339652387400015,0.8773539237885201,0.4582785359289643,0.7334675202753314,
 0.3154570892727899,0.3953303044312705,0.8653440302294064,0.8841494503915914,
 0.49279298328764776,0.8736538034016245,0.4746477149543664,
 0.37329365880524323,0.47732574422581553,0.3957395113061153,
 0.7216707328743603,0.4050797729227248,0.8268979835019741,0.668640487838625,
 0.5404584753521136,0.17632771349340493,0.6116286234401468,
 0.30214171885283847,0.5436638175887015,0.9647313900502921,0.6055259920344698,
 0.5186358407306564,0.2010235055441778,0.19888795412512972,0.2559579756831588,
 0.8755667529335038,0.2963540164995454,0.6168170420094543,0.3331276055782959,
 0.42036354163185363,0.9553792750410339,0.7837575880871214,0.5661158143431767,
 0.8780782465774825,0.3426258536528124,0.7225102611083094,0.20689441761579241,
 0.5363220186671164,0.08052519582254314,0.8589604970115488,
 0.016037422608664254,0.6264905425829689,0.9922947305080854,
 0.3302517670024322,0.5445691926645785,0.9824777617843209,0.9931311837365089,
 0.7529010302722026,0.9300551769957347,0.748201506747395,0.375164729168329,
 0.2970105399770159,0.1613500134715281,0.5893509724244055,0.8366666057942163,
 0.22690058945227143,0.08455246654078252,0.7888990447964569,
 0.6518417612537797,0.8268979332648505,0.010335004089730315,
 0.20858780476388838,0.38965448724946694,0.8491434570023572,
 0.5199356815205809,0.7703952249863346,0.15840014004877934,0.6648386344970862,
 0.7624995812891684,0.09462442690068651,0.6701545309200446,0.1683896551881474,
 0.322091986493533,0.5645339007263316,0.6277455864532075,0.4408432431537136,
 0.9557184726741471,0.9160297898943298,0.7939996310562599,0.2746564319640652,
 0.8477179183761337,0.581366199607167,0.5690450750902338,0.9407528166572113,
 0.5822703287818627,0.34859404385269366,0.32766183463869814,
 0.2599777987676982,0.5484821565174326,0.038955790369549326,
 0.4753632678571331,0.615508913369661,0.6737477083342923,0.16485888203749355,
 0.3457081959783399,0.6313724353975172,0.9978286976600239,0.2737417034005162,
 0.38064157362473094,0.8166674630620679,0.581518525581891,0.6225783685777517,
 0.050239311603042736,0.18387130307725374,0.16177788322120346,
 0.11511643306746722,0.6354596328080618,0.6761350129773643,
 0.06352535936435355,0.6000430477437162,0.11164730621825858,
 0.19578064001968631,0.9330262288860092,0.5140353605104734,0.7083826934419826,
 0.33044272746229464,0.5466798307477532,0.762354127637251,0.7157740956462231,
 0.9993438009865205,0.3766211577226556,0.8424480380890196,0.9834329897828,
 0.015883449406146832,0.07696452321398906,0.9864954301264913,
 0.8242186125377671,0.03333367359253736,0.2140983712902752,0.5655426317078717,
 0.4729208674604545,0.242350497652714,0.20381608863061518,0.998852078548415,
 0.5300619861819788,0.357734783203415,0.8272648338132373,0.1897895547987537,
 0.18667089300529527,0.45818372313206546,0.31797050671682625,
 0.9700012720591628,0.4053048784549371,0.014382088113486291,
 0.38116456450214165,0.3191901130959678,0.03642346434737409,
 0.6830357941762136,0.8583536672367081,0.14227706513708716,
 0.028697340498595514,0.798592027148945,0.03590093185553278,
 0.6794709827898276,0.9851978680454857,0.4751203507451073,0.9349987517277555,
 0.9479345807875645,0.7820098477334754,0.9388132679643497,0.3274760581099484,
 0.6168459190558794,0.8747992824478359,0.6336020738939936,0.8759306660145836,
 0.37371811077724093,0.29121444623240256,0.9651745076282194,
 0.36779261038349365,0.2949706165747038,0.3940354438977596,0.5047972654590819,
 0.33282878501529933,0.19822808637990585,0.434687997335508,0.8475400975251599,
 0.9954954007177284,0.5459945565465117,0.4380933076328508,0.268765563583245,
 0.12301998316404261,0.4078776695021321,0.5422885329802644,0.3146103472011045,
 0.0326415493054244,0.3669462992475231,0.10393836383030763,
 0.15861296983558482,0.4688262222700852,0.9428908670047808,0.4270409441200569,
 0.1991642057839761,0.4023213034114258,0.5346326715172329,0.20515661894668358,
 0.7679108860151682,0.3502689443654473,0.2254571780879494,0.8113222139072732,
 0.13464384262415552,0.13103645891764604,0.023583276526911945,
 0.06210671080284225,0.8439859110510148,0.22528318368189093,
 0.3670939708344745,0.5850839695792791,0.981319215340346,0.39084264273337044,
 0.4586294420871133,0.831001513771816,0.5932836476206661,0.8395845980650882,
 0.009185817992404033,0.43839040659843787,0.4724376794019046,
 0.6016960441583958,0.49099524211404744,0.36786375767471424,
 0.049900159667906174,0.37529732932547866,0.1600860090120252,
 0.5075872241006013,0.7919997265005108,0.031075478610719465,
 0.22729426921966955,0.38284009957393594,0.8043448329010211,
 0.2736536572911843,0.89945405508301,0.152336503034753,0.38921973373184626,
 0.9392454239149933,0.762130486187233,0.8001681962125443,0.6752743675451715,
 0.5922344389489351,0.5720694178975286,0.3504212076672567,0.6442673706721234,
 0.7448408289522408,0.2539808220777734,0.9484483362036137,0.26627899480472694,
 0.34820452356338083,0.2292781127158141,0.1856549619076111,
 0.37844545269264573,0.41301657651269663,0.7676829686958561,
 0.6436259303606121,0.5350272831840884,0.6856717707778925,0.23731726305309842,
 0.4642392580682405,0.5248502100551545,0.6372119042813448,0.3762966156980796,
 0.628354445620161,0.8327325776179888,0.4591326822055821,0.21058208579224247,
 0.827576381386739,0.8971830222165398,0.6416597186284959,0.9328143185584623,
 0.1276103403017318,0.8955507073429625,0.9394730773115655,0.37271096870587606,
 0.7462164076933918,0.9682636198134084,0.5163644435215744,0.07833271204074788,
 0.2427526571871973,0.05459814896563453,0.11723129014607914,
 0.9498212238372583,0.6781527364973196,0.29991982189867583,0.2538720146310095,
 0.8665607280577398,0.34190562763272203,0.6166383805370591,
 0.21333928784963108,0.6453381731781573,0.9097240116885279,
 0.49399795008217484,0.9557983661451339,0.9609209086838502,0.5963501182140871,
 0.3381311002717352,0.06380204980978632,0.4556370805009694,0.8851896525111933,
 0.9201071865178919,0.5181266677797127,0.548721290943837,0.4304778903986668,
 0.466680685782769,0.7404337566023472,0.5584530157871506,0.4646054116433481,
 0.14996412232224077,0.32084153321194875,0.09603091565775912,
 0.9897153140691344,0.025767722976539398,0.35950321808128016,
 0.7199181418086366,0.607838395894467,0.44670220980820696,0.1453411394123143,
 0.4710984116535353,0.824950811328663,0.11946859166053747,0.1292091957407494,
 0.8320590322662884,0.04249423819515297,0.36525406432471097,
 0.7357019662719514,0.12418942571077396,0.06217658572241447,
 0.47146984944216364,0.5123124872497482,0.6372057454363537,0.5746873698412571,
 0.7150317868156754,0.5511061000614605,0.9255749113249659,0.6079965295578412,
 0.2261990051056656,0.7955618096689148,0.18516828502865934,0.9247875210322678,
 0.752736618455918,0.05412318545473216,0.481877363920017,0.8025232058579473,
 0.81355861343459,0.8447886583703983,0.3845624590122567,0.1649141205936122,
 0.4911841423813773,0.5399458859088639,0.26601278923002447,0.7150999310838584,
 0.046857946453464305,0.23996800410933683,0.7549200166934065,
 0.1476598164748133,0.8464613413842648,0.8052742176101408,0.9654735011651313,
 0.5113966495834674,0.028065105787182132,0.2505132243146978,0.877805096779706,
 0.01239841799747532,0.8449785840811705,0.3401174207163139,0.9537923388709758,
 0.5912222482347226,0.19634577596452263,0.5715825933801012,
 0.19270087200235708,0.12039802021990054,0.17906769355604846,
 0.9293660707625285,0.8405223191125214,0.35947597455220737,
 0.07958896968240048,0.25556538236781745,0.8537379736681282,
 0.44290037658505776,0.23165581255313716,0.8735112750527372,
 0.7995760037442866,0.30120381843701627,0.7060535264957718,0.8323242986071091,
 0.7964734371829806,0.9562239218757541,0.48198425100320597,0.6004612551665612,
 0.5545595068309735,0.5998417817807951,0.510833754720202,0.47999814880087177,
 0.5632852289343111,0.8099561258900566,0.1006310294193421,0.7445727289512387,
 0.18415612470905152,0.8313555305973397,0.2108402690205845,0.7150311204660671,
 0.5821314918175461,0.07745768436392098,0.7858477070851462,0.3611249712379475,
 0.5719185081122491,0.12811651934721802,0.82865513933446,0.17546448986556318,
 0.7323684252998292,0.36937729220879567,0.5846825827008549,
 0.48246316652330234,0.21049607345585875,0.4819939609896142,
 0.5064882750317465,0.4611011691274661,0.13242311598942758,
 0.08533289423982515,0.5211368730526993,0.8279941071126262,0.5732476972166259,
 0.6433555356125793,0.2482580866795998,0.16957752062225884,0.945138381365204,
 0.02602480218826564,0.23626306372221673,0.256815124629588,0.543739037283332,
 0.2614629153391589,0.5324491599365861,0.4157486692820592,0.05758811500907046,
 0.9281726795138959,0.12235791733306889,0.6507497667518176,0.9045763573870943,
 0.5493657619708865,0.0333341087054535,0.37715780154652867,0.9054781076242382,
 0.5995459561887967,0.167446264404557,0.8877120942768895,0.6351565539200541,
 0.8472919983422168,0.1296224319169108,0.3995869333184451,0.8263574110388591,
 0.8379455736488183,0.08375938647199588,0.5100406645194243,0.7580986278754172,
 0.2809900519961053,0.6154620318474727,0.0892045294664392,0.8109836451719943,
 0.738954335906584,0.03796128032352286,0.8401578880903094,0.32474784478027374,
 0.47509107634120296,0.8931505932735594,0.10800576758276537,
 0.3144010034230371,0.9539117368169056,0.4269222690698925,0.5200348015932361,
 0.36965336288227435,0.6845924283642145,0.5814390744845415,0.942191429655057,
 0.28050926951777055,0.8419180925433116,0.41982286831133475,
 0.6441169630290507,0.4837206797478558,0.08666982889919961,0.8676340655737925,
 0.5925035634751468,0.17839406455589346,0.9190616505691043,
 0.12160296909173962,0.03979174096364968,0.5881473324662456,
 0.3886631431451799,0.7911137997815832,0.3640865060828655,0.8279685803055905,
 0.08472714426659667,0.8729550237100652,0.019181499271870517,
 0.8408522585392721,0.6715615113964145,0.37041262044858114,0.9900188928811978,
 0.011559557885935323,0.38573713876597726,0.485040048764094,
 0.12660857533084424,0.4634558480498796,0.6214098580454399,0.2333610664206821,
 0.18615588242237402,0.08828729126293156,0.7117715190806755,
 0.48634916306217646,0.616768287033509,0.4465104104748273,0.5343327243351181,
 0.4339102381360921,0.6159722492641045,0.3843838446420689,0.9841690788864779,
 0.02517038925267523,0.48180439647182716,0.5573757868330159,
 0.21123456621332715,0.9128309732333549,0.3969343879302202,0.5073337945434927,
 0.8930604752397064,0.8520309735727386,0.32140337703390653,0.5563034809034121,
 0.8294623395063349,0.12419871882068056,0.7108922012456422,
 0.06057226489454315,0.48909544609529054,0.3344319007046288,
 0.3699900749597118,0.3639757682562016,0.467619738542708,0.8221916341272119,
 0.11027488407647712,0.4314684750384853,0.3326190121069773,0.6819964458734269,
 0.35316775179957405,0.3532669792162053,0.3202245458928229,0.8798613227506146,
 0.9684256660024213,0.14323590599545255,0.7567525862748204,0.9096338436995566,
 0.10708547527134848,0.6425323791625208,0.786885311707426,0.42147387269001757,
 0.4245959989715481,0.8604374218906855,0.6064370789380049,0.5278151394691958,
 0.1718422303652748,0.9224400063111523,0.41930454139889395,0.5483171463249418,
 0.9951883618685398,0.8102604565620882,0.6154887064608825,
 0.002798393780741426,0.5020933794027983,0.7184178319050523,
 0.15372954243286585,0.5905166201320455,0.4219493255222181,
 0.39403120092141053,0.22958260770751826,0.33686392812944566,
 0.47119832767091696,0.9342606292532679,0.5863445631366652,
 0.04173114277631529,0.15282922679634436,0.656674580199914,0.3985701408918306,
 0.5526163579854402,0.5831958163539661,0.9460491147890284,0.8323661730019231,
 0.7294232907626916,0.3297606862469217,0.12667264805773948,
 0.07243683668368894,0.5184746998343761,0.9486647439913218,0.2006979513943451,
 0.6222499233697353,0.5932671218407515,0.6272772506709156,0.2010849052641317,
 0.6023831089513884,0.5702509571272845,0.30389919717024916,0.9498160102048935,
 0.6719030610431966,0.07582742544413668,0.22366478413537227,
 0.1748660498290443,0.3072639406734293,0.2712966454487703,0.5288673745056138,
 0.21219425817980697,0.06836610186141656,0.9254500796435422,0.902872886882391,
 0.57269829404049,0.7852093080317109,0.16157813813288502,0.9017527415844051,
 0.5428280235367167,0.939174223430493,0.09436642139061757,0.7947351242690284,
 0.4534536318307185,0.514297241665931,0.6806450229839414,0.9991752666590547,
 0.7162448244969903,0.08875266669914916,0.3103385208483622,0.0917669835562539,
 0.23366257348606745,0.1891960737292414,0.8734747126497611,0.8578336594537145,
 0.8150081155570509,0.9701735520296428,0.4655505322595507,0.3837360030371526,
 0.17025300510803587,0.136385752617884,0.6019336022991026,0.48246538825961904,
 0.6125676371220523,0.8304491747962314,0.18814787968299407,0.8722743284070662,
 0.12383566447464611,0.9113053490112742,0.5062190510659454,
 0.09237829195638125,0.9691513367715084,0.037387690671670026,
 0.9648592551871036,0.9631977662807474,0.7303292613300508,0.7856397586632538,
 0.457464768937898,0.9536221187279078,0.3300745502595215,0.7252752463831361,
 0.24134073067065953,0.7075063286962795,0.598654702355345,0.6812163202061924,
 0.18482926173348802,0.8989596531814666,0.9318688132653503,0.3302573365485145,
 0.692626162900927,0.17720313127832343,0.9469628873115445,0.1585279972589937,
 0.7221489063139458,0.20101617599066812,0.7832689937968489,0.5490102284096569,
 0.6790345720798886,0.34527813399476237,0.40437292373340306,
 0.8328394940859464,0.2091883331282829,0.6372455782527184,0.10328757597641991,
 0.28703583229910035,0.08286305140331329,0.6115618148056403,
 0.8019745033795638,0.14190055762795384,0.5565971148603069,
 0.10140449890976322,0.25808907726224506,0.6546430342163623,
 0.11881671021716467,0.5901231837902772,0.01314190261849224,
 0.07518874654155583,0.48021236616937635,0.18103118803773122,
 0.6226473707135665,0.5174727148436201,0.184914072781048,0.9838764957027955,
 0.6348597998491486,0.012254886857753222,0.8898372656313563,
 0.6223355989151265,0.4633107781396293,0.3991971825769397,0.1731271269875111,
 0.5130623436724931,0.14315610083918917,0.891981642136642,0.6189511889251196,
 0.9188159027264229,0.3932455203070764,0.5006282430038987,0.46814788861029577,
 0.8133874860959758,0.9164295819043631,0.9068975630585759,0.15979551483532717,
 0.5269874442994562,0.22806057456360174,0.9095931452753474,0.8292762947448349,
 0.2020111076745954,0.823607673255047,0.19806612818760239,0.9892157234731905,
 0.8700497288287495,0.2297474317671685,0.4573270891206084,0.20599421190544198,
 0.2566231770469102,0.6171417836872144,0.33259670276633635,0.4766692805088555,
 0.897683672405569,0.4354398765682087,0.2552863690404318,0.5595613665852606,
 0.5035851138466765,0.8631916612298685,0.46166100076575023,0.6048870582630602,
 0.39021703583291933,0.47322850184530385,0.8885086084719263,
 0.1399655231766569,0.11864711977339448,0.3952100095298603,0.6366909522530534,
 0.9823193486831184,0.002597087956223376,0.33124076398422553,
 0.4453076669388417,0.8195261737463477,0.0301212574641615,0.14764417598806623,
 0.7436220750933971,0.8166851204102764,0.3991427278492191,0.7224853369340097,
 0.6258849005213722,0.12582546317059984,0.06923643710702942,0.198435235042136,
 0.714778619166501,0.6025106299445377,0.9009727225390206,0.6876313107231075,
 0.35457586416918996,0.8436349804532638,0.9469240754402024,0.4789539911999229,
 0.4760077768861297,0.02538967099487255,0.7046523601999357,0.7093950052941596,
 0.5636219989749655,0.027270048567751948,0.8527086396043233,
 0.8538305153610913,0.9773870788238787,0.5226754575041482,0.1746400873182492,
 0.12742661017548063,0.5246802770116614,0.9847712669218439,0.3893104732759636,
 0.7968766130138298,0.12627703023169257,0.6386295258896559,0.5783304414976926,
 0.5615675011111897,0.8523461464372359,0.8876197553303152,0.7670304613676464,
 0.32913765105898296,0.009989574450834482,0.6913214286708365,
 0.28483240758347217,0.22641645108854191,0.7262007545023367,
 0.7798063034951368,0.2729350969216481,0.6799061723681825,0.8530894415147712,
 0.850094651316122,0.41166404507857624,0.3033127270346688,0.6496020245709757,
 0.33489637173650333,7.070586018157465e-4,0.9274994046043328,
 0.7873896440035357,0.5559506485517848,0.8979802507744643,0.5708072164923736,
 0.3331774807527169,0.5556998550888259,0.02757975383370881,0.5896917115667677,
 0.13243422153996764,0.7068965077940553,0.9338390881446228,0.2659814953840627,
 0.3652328638191896,0.8333397259237969,0.23576002757035752,0.8676821379492896,
 0.8788324662175799,0.0657598936252991,0.8602279416296118,0.6302324889556401,
 0.9059418411811625,0.9608091277676033,0.9857217697345968,0.05515286260342944,
 0.5090292271613183,0.13610844472673667,0.3721746859755093,0.960355695881066,
 0.5791695114404927,0.9859067293351954,0.5283103987700036,0.7855721483395084,
 0.9535234705175223,0.9614714777643032,0.12395263942345891,
 0.40269488635411077,0.3941788446425639,0.11555355476955387,
 0.21511356765576728,0.9951630285838701,0.8977202039626445,0.5044648501176886,
 0.041336430198073204,0.6909561790846955,0.753952076732749,
 0.44774851370421276,0.9100772153536109,0.7632759183367901,0.4622751614502174,
 0.15497412159405255,0.36606388334576334,0.1825485566703653,
 0.8852049418674138,0.5441706561502913,0.06190940576490167,0.1364488687698252,
 0.3120475849155815,0.6251684968412291,0.7976514649723613,0.5203518309601977,
 0.4197664490381273,0.9570741114008712,0.3642478343026685,0.5826321578060625,
 0.5110031996317898,0.11694325568685882,0.08543210945466162,
 0.8994523442274511,0.7564718130601384,0.3382524147486238,
 0.057394435982141556,0.9648791525832783,0.875604315123966,
 0.22130103434854187,0.35325956912154477,0.7281107822182777,0.994556280583249,
 0.11611885210632666,0.9030365074803844,0.912331908511375,0.19193723598790724,
 0.2729747016745119,0.9706214261722961,0.30723725543510594,0.3939425352457715,
 0.998115933042528,0.045462455305864236,0.16930783411142458,
 0.18884816364376766,0.3152505269234185,0.16922743555043174,
 0.25210567778030857,0.1120281644667609,0.8073164313059997,
 0.15988309004416013,0.8918090316781919,0.8694707620664063,
 0.13049545591105394,0.6280211118751575,0.04689063439057983,
 0.8634164457577894,0.9574722785856269,0.4514803346008379,0.789406898914764,
 0.4991789313469662,0.7269766917892326,0.06385536961148253,0.2641179435669505,
 0.45188533024197786,0.615354455327131,0.6626390054839322,0.31664538809072496,
 0.8682064717707776,0.4153008105674807,0.7126895006821975,
 0.004138395392018768,0.616719676342284,0.46494167926945373,
 0.01681213966401429,0.2049845972716292,0.4239432242387774,
 0.33100851464984204,0.2789778244213019,0.7576679707856756,0.8149727736410559,
 0.8194033735389336,0.32882361823061357,0.2714479767950446,0.4732654083061112,
 0.4603205208590917,0.27044230147535286,0.6962651869284903,0.6153871466982703,
 0.9626904074353339,0.9398413399991059,0.4418098265290584,0.5091892905540983,
 0.06189018478127206,0.38892695858801685,0.4372608321943735,
 0.10380852183816724,0.18236987368025792,0.8441243898610968,
 0.8695072703337745,0.2757848023488666,0.8915059345350762,0.9450504967258403,
 0.726975537729013,0.7442853715366413,0.5079768958566954,0.885722562119939,
 0.9485172549517661,0.8723530608688651,0.6673280015406863,0.16291166410350066,
 0.5562051857998686,0.40753198847854033,0.9814198447200184,0.2006239266150568,
 0.5438733039739789,0.19508871198957645,0.5768790469685822,0.1711967274381445,
 0.8072475164918829,0.1768272695598211,0.34600756629144525,
 0.19289929610271228,0.5678265333500422,0.6998979272169684,
 0.07585092792526826,0.3610400146826097,0.28299918654035394,
 0.7250078375629518,0.8509781716628014,0.4890594124732902,0.04211109893869991,
 0.5674928685727315,0.8511362709796978,0.3314927242072109,0.43755911354040866,
 8.912815498131055e-4,0.26118653854246754,0.38540749801304275,
 0.3914279669249605,0.341150625967116,0.2863742672657885,0.5063086978168497,
 0.7548887986453767,0.31680091295253454,0.22377958392036967,0.808058270527027,
 0.6348909392610507,0.5044651033576537,0.8889449756768621,0.02967280252921778,
 0.46683507389869583,0.9167764035701909,0.6017048308495677,0.8622315413468113,
 0.4389839923081096,0.7915644802426578,0.17570836186159622,0.686924115111227,
 0.657231882247019,0.469318414216342,0.5385639436759603,0.9543472729562745,
 0.32832714161065935,0.8399751712207687,0.5501994071353696,0.2918230463981488,
 0.15424074035357793,0.14020144811864022,0.8359736241234084,
 0.6997969178903348,0.922732479270474,0.21917683025696866,0.6066094965764347,
 0.47929750255401626,0.44979771734553553,0.0381714645288308,
 0.7277246941751547,0.7388799623654246,0.9367296094488458,0.400453915114662,
 0.9941238322154643,0.6121543090946042,0.3945865058510809,0.4994338187613976,
 0.17975317881546782,0.9237480245600556,0.18412652777412797,
 0.3940123710449427,0.4427910334091467,0.9464428295546796,0.8171629098108285,
 0.7300199723833862,0.5330110051267466,0.6758596264842551,0.5203691700188217,
 0.5619063976327197,0.4995471949138752,0.7950470600238181,0.6914961256096903,
 0.9793520977781331,0.8483188365593833,0.34828323167116504,0.5295530238287773,
 0.9678709317521845,0.12015206528243327,0.6316984851752788,
 0.45657305351189537,0.24025191396641166,0.6670996093610863,0.524392819585792,
 0.8929914354802763,0.3317033740073032,0.7335952291049339,0.8510284007753091,
 0.1527760276110114,0.23560686831198874,0.6957536770275856,
 0.32786131030207677,0.2710693666736784,0.9983642870637675,
 0.12912939909864263,0.3033072353632029,0.23616088116286948,
 0.6007344403937958,0.6921420370552439,0.5946973702047762,0.04142480968376805,
 0.4470898864686126,0.9613912357088521,0.85782526345953,0.8512093401890184,
 0.6322790622916574,0.4590722343844639,0.8524799861636718,0.28565967066498243,
 0.31849311806710323,0.9205631725320779,0.9321995505827613,
 0.13734876934847184,0.8752845434540673,0.9335490632801345,0.6660581706208023,
 0.5152504634757025,0.7102143188845292,0.38488312340552966,
 0.23413787175492662,0.06645588436589167,0.999697112307385,
 0.29236756278042175,0.5588839987607848,0.2715912173439553,
 0.45888046891252143,0.30738748085418865,0.5304036038242665,0.373753685152012,
 0.2359715279389495,0.3645713527066492,0.6110389377421104,0.49676354888541174,
 0.7061044562871405,0.5994162542795324,0.5034443629412418,
 0.014090923941502131,0.9827675419807046,0.3982709385829071,
 0.9370560538437397,0.5491765536471387,0.5897796061008427,0.7254856582783282,
 0.08451982912492895,0.6700767132653799,0.3355788085534519,0.7336533411978395,
 0.3949167133988247,0.6661192869979892,0.12532668752959508,0.1011790455633137,
 0.562331194239168,0.1884133931444576,0.13024388478649285,0.336253571446102,
 0.7168851855033456,0.4130494087461488,0.7069009948889516,0.6491634451739892,
 0.7707062940611614,0.06454954566326165,0.7703530316312065,
 0.45210285652418525,0.39372429548751686,0.36629355484595283,
 0.9025635196011292,0.31220749032535566,0.5613603983643513,0.8866186910360827,
 0.8978351598812508,0.08750002966000414,0.9465086632352427,0.6917507274394555,
 0.3929501124629361,0.932083342220689,0.8891997871468362,0.13146693201374438,
 0.20139714501663053,0.7692372635788113,0.7324706654190833,0.8158764681555761,
 0.10295212357441563,0.9914650144768582,0.32303441828906676,
 0.8734629359016888,0.5897231621172608,0.4893556600787625,
 0.043095860867221725,0.38430366012466677,0.05932442889593659,
 0.6325565176790762,0.5285381364265032,0.13672032242225396,0.4341906466031249,
 0.2954999060017728,0.8073156423164191,0.02690699237510219,
 0.010101317804808274,0.041030284317758015,0.2585323784495752,
 0.6118715541174664,0.33554292255567375,0.11911626052713031,
 0.6508883643989778,0.53333727620453,0.9299874888831735,0.761219992077721,
 0.5597957948851496,0.25561120264619375,0.798008413793243,0.2844112769439253,
 0.8163963286219602,0.6649501339622074,0.9252788408351923,0.29314693307212913,
 0.9790616964687388,0.2559349561215758,0.5456650439165449,0.11958481455518921,
 0.5099776098308668,0.8830526645971741,0.5375708252172077,0.8366335998463184,
 0.7215285854209986,0.6364476459976016,0.7445453879231132,0.24860515963051621,
 0.7957135533500679,0.8452933381525553,0.23741510387208997,0.7167539487334703,
 0.8416591082820158,0.833552032029548,0.04692112440658369,0.5470431757770797,
 0.07570462204806505,0.9987844806544351,0.5996758233197218,0.5293428405100703,
 0.5706263371240022,0.5470879273670773,0.9749158283565043,0.29966378328557275,
 0.31192463749629684,0.8067521823063329,0.5423445029721549,0.7602820883602089,
 0.5917774556456893,0.04957047657135727,0.0408326410592339,
 0.21765794582389508,0.32970662584093335,0.7075898621737988,
 0.02077775492535483,0.5149632380524253,0.9918572201222283,
 0.38090242781957984,0.4005055653507752,0.4408274837035734,
 0.49353148412114145,0.6318063071128228,0.6083745054843193,0.6449206986173269,
 0.36413029740057545,0.36622740261476405,0.32054331231811606,
 0.5906701351275419,0.39480893324430055,0.37728533403862585,
 0.4235342577062018,0.6880471612339587,0.5875238903359383,0.281469088742757,
 0.9285039800404699,0.09475392922190429,0.7602119305479738,0.7910142597909079,
 0.7442886474541984,0.5010073039428424,0.47942594426003826,0.4444535929834096,
 0.25999821811749435,0.04719168313505717,0.6348326458696242,
 0.3317674859094193,0.03034167323281023,0.7472610888306086,0.9266120590563791,
 0.012982397863290362,0.7199914912077185,0.37995129570137676,
 0.732543859557479,0.850036201559267,0.03689345954317358,0.6315223947344554,
 0.15158223850436225,0.11458756626800004,0.3642289574479348,
 0.0815114085090155,0.33631556228617343,0.6627180525530074,0.8299733781002133,
 0.24732757797261284,0.650540086761543,0.17787033845571254,
 0.00814730442182865,0.5717670837289055,0.13505604454306364,0.658467068688373,
 0.7865117336202636,0.8740191695711412,0.20029631016988847,0.9720952006050307,
 0.810587494117938,0.42439363802772,0.16730751495628604,0.6022909991171118,
 0.6771280616541289,0.15557119206840875,0.550704362740932,0.2650534952469191,
 0.8406866077150065,0.8051835760318459,0.5995595220307595,0.3672175553355672,
 0.017917708071749083,0.8801540669350081,0.7709018293788091,
 0.4472028906466141,0.6642396006368673,0.5405491280868148,0.15772306642760647,
 0.9832235774683488,0.4208294490866087,0.3841708403331603,0.8378238977983379,
 0.49127518120884717,0.46807613608321885,0.6964109570017669,
 0.9464502598512295,0.8354209845433094,0.915245444292265,0.05161200507522323,
 0.12380477445344251,0.490628474214878,0.49367397524164636,0.9776585237034948,
 0.7372510612371952,0.31463746704247164,0.2900979402520161,
 0.16759201381095024,0.2663579431333992,0.4523706931622111,0.2272819270889036,
 0.9901253119323293,0.40032590096949194,0.919932656753639,0.26601075346091085,
 0.2718148355416168,0.2660227495566998,0.028030907950208683,
 0.5412827194472807,0.2032374337215337,0.6936435788108436,0.4281847044348315,
 0.5566338104327313,0.5995861086565807,0.22465573195302269,0.1937990735738372,
 0.392438921654622,0.17327942719601108,0.6364162660709386,0.8447586261802812,
 0.17388344848329795,0.33284084425272953,0.45202262561674167,
 0.7787146351770339,0.8335675138814265,0.7628871903485898,0.9666266762888249,
 0.46206821590088354,0.7682773724158335,0.19854328124517706,
 0.9814313073910359,0.4298376450310153,0.5510447189523878,
 0.058052202918594475,0.3605746187052049,0.8501961132656092,
 0.6611007449491999,0.5960961364277914,0.05622228864295731,0.5956891888937683,
 0.44189837941692245,0.8069425802014203,0.28923684741804967,
 0.4116227682278163,0.6138239878104496,0.25572114171655913,
 0.03511017109405845,0.6026291106871405,0.4559686480785091,0.825130548344584,
 0.9586839617373868,0.2096273655713159,0.15245885829370165,0.8936621733050474,
 0.9362162887564804,0.5302411879742978,0.36579470919089707,0.450713897362366,
 0.28343872289276806,0.45202234368152394,0.14065776088455306,
 0.5854011238325605,0.8953290419695277,0.5157857331586624,0.4486610237612564,
 0.037408358049892176,0.20671375293153071,0.8108519031677335,
 0.5352652153012044,0.9531690761321207,0.4739937152671525,0.3499060774702273,
 0.026700531855529563,0.7372015776439541,0.6114310435731855,
 0.2294294617732554,0.07181111381412553,0.6370541601074492,0.8004389445701925,
 0.8632906398517893,0.6783868533169727,0.0019324140174865612,
 0.39716665054683353,0.18035965195887993,0.8903919530995784,
 0.24148285084381116,0.41953077206381273,0.31749807460862234,
 0.8568875437820164,0.6977816299582851,0.39117670965373164,
 0.12132999224131147,0.8913365849514824,0.37210982983361185,
 0.6975832872618194,0.007558227120249006,0.7039104309158305,
 0.9739829807868438,0.8019121235616876,0.14026840351688752,0.910657233317248,
 0.6372529159964946,0.7811203515940475,0.1504415178185592,0.433280275355166,
 0.1320618262686798,0.7256512315796104,0.0025935353424320073,
 0.7355973322666923,0.8843630965865118,0.34932898337369955,
 0.20323306008395203,0.07018493663238379,0.008135444381418777,
 0.5854642461174857,0.2504097396539907,0.15133512801628568,0.71883694514993,
 0.7389379695770193,0.4481071400237404,0.6156397648703648,0.4183582356365725,
 0.011418588660559559,0.6277207007926504,0.31293044965705663,
 0.07208825937609598,0.18063303624542548,0.7262609671316466,
 0.4996315344970903,0.6300505669612363,0.6696689323083393,0.11383759259753157,
 0.6626233764255103,0.7502781387780025,0.5628114543990886,0.695667668619661,
 0.4465136135614889,0.997395380945888,0.06340783929579852,0.4258244017413998,
 0.04933702302069898,0.971890646862175,0.08610841312941431,
 0.11794621985430787,0.4620321036400814,0.1125920726077061,
 0.25949702044313006,0.9999665319758553,0.6468530497963517,0.293643763313332,
 0.32805819989306895,0.8130715599613355,0.22643691133822708,
 0.24255949844904656,0.36946886218886243,0.3157364464599094,
 0.3143488750612575,0.7074286653670167,0.7440550220191047,0.4585201349474536,
 0.8071180231803536,0.8124065366240576,0.6636606608205271,0.47878821772398117,
 0.10549339500098365,0.38916018997402124,0.26559908907167196,
 0.20973471865268256,0.9774828760320098,0.7005857849864484,0.7831148434540599,
 0.1781646322048538,0.8520381794012579,0.8204340744055196,0.1316880945647495,
 0.017204063231710642,0.3081538597061333,0.43557592139543233,
 0.1915156662255878,0.7945431185740081,0.5529245481077758,0.5941232099123259,
 0.49922892041854916,0.8394187365703952,0.09367404338565,0.6014139409685829,
 0.38985509536333396,0.5720938501321525,0.781099693155838,0.6188325259252455,
 0.5804814751047738,0.11455904881383083,0.2906108721327355,0.4217606246027996,
 0.7474950913661776,0.47024747840281944,0.5761538060259821,
 0.19134089049229963,0.7376922227505568,0.40210453975079974,0.497641563033428,
 0.29486575903644174,0.5202704660819949,0.5723091598909291,0.6100823432169289,
 0.42888588276716433,0.6645995702044305,0.7333881945324887,0.582442055508123,
 0.3916607361784391,0.18806900894834877,0.691369276547311,0.03887809642763418,
 0.610592249767766,0.38376451889165253,0.6808260865603541,0.7476928362655435,
 0.6946999275834922,0.8378526803481815,0.13785705136204984,0.8181539169990543,
 0.7821238656492157,0.8354218972408545,0.7209255931027143,0.5095141546653059,
 0.11354304697442674,0.5515674727450978,0.15600983949335812,
 0.6515312473596193,0.10183221548925392,0.7996635596209954,
 0.39559147538844874,0.13853587886999508,0.5616960026127333,
 0.12015501608791457,0.7093437237263137,0.057299689677347776,
 0.1753196840106811,0.6877844496748141,0.2675224780831077,0.5728835914769962,
 0.933617096644527,0.4584130721571792,0.5075845709541112,0.6835986044047121,
 0.4435636512487613,0.41914677811347245,0.7947084336291872,0.5906318180125937,
 0.6304521790953379,0.27306856784230715,0.6791484260747189,
 0.22364785817859278,0.5635156732494354,0.14685694244841085,
 0.2285668731048881,0.509048351085273,0.8836334239669137,0.8145907951401977,
 0.6904315638524986,0.7916960988956538,0.8040541405497771,0.34494760726585094,
 0.980500086241855,0.8723502245552015,0.827329605856434,0.6815841904489196,
 0.6514632444636158,0.3206110971268161,0.9437687725317833,0.4631558680206658,
 0.709402776307547,0.013863909948620523,0.7661678382867992,0.8950319004022079,
 0.652874010919187,0.14958903306512372,0.5760605057549097,0.2496089656455811,
 0.9098839700975273,0.18670465177200146,0.6435917248004432,0.3338431899575077,
 0.043139611477608364,0.4135006654453377,0.7648811858182699,
 0.27497350833663925,0.38739054089117053,0.4257154738201898,
 0.8504873836911234,0.19456683447203005,0.3830760170705705,0.7165293440487266,
 0.3804123530410979,0.3898615881940829,0.700242125506381,0.3214073180394156,
 0.955739706330616,0.8588836718143169,0.09565431418032933,0.7474667054971492,
 0.396966634938674,0.9051592709520052,0.6010052323455315,0.4632928726808515,
 0.13971263668429446,0.2526696546156866,0.017854010542197263,
 0.3441655867221578,0.4134025677919586,0.44488948307198417,0.2065018249268682,
 0.6636606696782261,0.5946416703104012,0.16016630549229283,
 0.17577695067018717,0.6299557901136653,0.5842779862164328,0.3679133276250359,
 0.44110625258542324,0.8942088960131609,0.47121438568260854,
 0.21888413615504443,0.7712331582897433,0.4056421342591503,
 0.28767092116907955,0.09820380794069883,0.9667596458970218,
 0.3588625689499876,0.08769054100722984,0.8204952577227207,0.9904585435549912,
 0.7470749022256363,0.8489568520866877,0.6303149987391067,0.7486429694571475,
 0.947685856180825,0.021095511930419142,0.960845612739479,0.1655504059220565,
 0.43159457262050027,0.315156040879901,0.5928242740779035,0.18923295770538318,
 0.5527002451166411,0.550207616962667,0.9212999789725531,0.8243301792378367,
 0.5906400850706813,0.542435786477811,0.5058058914180352,0.9652674432279733,
 0.6352149300133948,0.6776617233125531,0.43531191434002026,
 0.10987577445601382,0.8896793540829845,0.35921728605260106,
 0.14778272592028796,0.14378581450953487,0.14786885975897857,
 0.06673392251554222,0.9562292620967879,0.8529980570894993,0.1929256040710916,
 0.08757169451010793,0.8443074002498541,0.424689321407122,0.7195295803002368,
 0.9134099549142669,0.7271494338485114,0.598739510398725,0.26106909436744286,
 0.39836553960362675,0.9520199929313413,0.5179725629523133,0.9944692780642453,
 0.6890855324501763,0.03449324850890123,0.6074531150433646,0.3631434611911386,
 0.0026415838057031182,0.07655228857153806,0.41465442991266155,
 0.31723248412780647,0.4139763892640419,0.9408272370717877,
 0.09277618729916015,0.39443120556907896,0.4553334493000798,
 0.6132634458902909,0.2784710008078546,0.2116088237739011,0.9006624612971443,
 0.9979950841466391,0.29775425387112986,0.6219995564418922,0.8840211277229061,
 0.945934526748669,0.36643910286074544,0.4935180379982087,0.10481656932164973,
 0.7845161887597688,0.23913466396310923,0.4951368550177576,0.9545854931536722,
 0.7972220573154377,0.17698791471292108,0.9714903210119195,0.7874710176560414,
 0.7169870084273572,0.053309639160026845,0.4372305654232327,
 0.19134551887999462,0.7263724212754374,0.689828922698403,0.4289242697297999,
 0.9118563635455517,0.43261791117035475,0.4849348497970907,0.9598231628080947,
 0.16048900622284623,0.7345556398209319,0.6558393569085044,0.4033133874890522,
 0.8828232477456817,0.9906795964028601,0.17115697657485984,0.6498581565259829,
 0.9062443788445458,0.5203865185307531,0.8327728124288493,0.573910117352427,
 0.730870407649276,0.6491794268854226,0.7362329722239367,0.891219822907503,
 0.5022153285179604,0.6251121542785014,0.23623606506138195,0.8749171033144801,
 0.2902786074465977,0.012149493812707912,0.8360041206830174,
 0.24986317297150462,0.26402535417198814,0.16429966098236215,
 0.0015861681911285563,0.40819205117341495,0.08629837400245188,
 0.10800169743483212,0.11569479227092683,0.12814393599714746,
 0.8653628812316345,0.7456495688765623,0.32972834365158676,0.4010675849158317,
 0.4669103409626991,0.5795285900770849,0.12074924296597833,0.409803595396354,
 0.718321219854903,0.6730914631182051,0.6162620920201263,0.6560579496190863,
 0.4454410814384997,0.5427927497385611,0.32615624045776204,
 0.48339602459199993,0.8384492465657267,0.9200137174397708,
 0.16295836744255132,0.5174194740671271,0.9881441376794994,
 0.24943623840113693,0.38389910119583204,0.8309470805218483,
 0.38082558217281415,0.032420109364998395,0.251247575634606,
 0.5760222245961407,0.3558910293689862,0.3940801494734475,0.9042364784390297,
 0.9743222765394536,0.5645876418773018,0.3204181561484841,0.4092998282923035,
 0.5104325399050366,0.6707894266790737,0.07711322340796611,0.8687391265636488,
 0.03446482963421382,0.16980333008882387,0.5952095950357932,
 0.7417272557974702,0.2382794011350824,0.44057805081564916,0.9918718745521556,
 0.7202059819985754,0.8700058465352446,0.4840987354923726,0.6397577072681288,
 0.7958480268381746,0.7121306121596688,0.9030891665032129,0.9233821897325217,
 0.09511666782049089,0.053120911814444116,0.015440866849860457,
 0.12193443814681026,0.45166739349456475,0.9344382816276967,
 0.7673868148174288,0.8613309602605412,0.5026628433006515,0.2869245525730263,
 0.5201743886154737,0.6586975298589521,0.4681624470834904,0.20185566256552268,
 0.30877630757688157,0.20978889974453352,0.2987898502789116,0.589892892610353,
 0.4605681402887316,0.4804682491425579,0.9169885133780322,0.7140609631854066,
 0.41781040588416607,0.9038374055135137,0.5987843205730983,
 0.20566624071549988,0.23351287790438313,0.49719073114929946,
 0.40269033828086787,0.7168456117999513,0.1021669686879717,0.6848661645459652,
 0.604906567798308,0.5761812413654787,0.32268761242602517,0.47077919013588454,
 0.917468187901467,0.6662255206361875,0.5896207175107864,0.8788662782274674,
 0.9179097477719552,0.016536721065018578,0.09392930696339596,
 0.3956274929784124,0.9950512122857,0.9999935508567569,0.7600657291480265,
 0.7625213554573071,0.059463586930774026,0.4415976335756985,
 0.7015317258666367,0.9312275213555565,0.932499649807995,0.6788491869317258,
 0.06808993567148958,0.28575165680820613,0.06320876602563541,
 0.8503153884469494,0.7876408443938365,0.4291515912369567,0.5518840642743928,
 0.20566334810419007,0.018776080261139638,0.21710266304443782,
 0.3177043892377345,0.7625320701861653,0.17294377081485912,0.8772763006953751,
 0.48497932428575186,0.9570500091103793,0.49122316181701686,
 0.39239392499640735,0.34293746518274515,0.9669343823047435,
 0.6220506671899669,0.06664416716742416,0.5941033803398291,0.2216743314533668,
 0.13051482575550688,0.8822540877404706,0.718213259929837,0.2661632850586739,
 0.5146330887863546,0.4502020449512808,0.9086534649083897,
 0.001972755491953637,0.09977217191537613,0.03389516119908309,
 0.7459307024381109,0.43445184542255655,0.8925795837444823,0.884780156587967,
 0.2146509684354201,0.7831289768377814,0.02266464505440202,
 0.06832525939358436,0.27602506180298547,0.5657029052019815,
 0.7769092623581827,0.08599775183878189,0.9642914840005988,0.3264369519555419,
 0.19903354054534128,0.8818614537584184,0.7638207552482823,
 0.00721518448359626,0.7788573650023551,0.37325367569463674,
 0.012633877428205653,0.1197673130232546,0.09676792546040391,
 0.6559383001807874,0.031369658372471965,0.7390188246969713,0.256035887503713,
 0.3037690733692281,0.5422222490621008,0.5715611354387923,0.9823235811860924,
 0.8718435105158604,0.2873478396007859,0.9671986344804699,0.772952272121263,
 0.856498449232656,0.7521910986469914,0.10091061835738502,0.3959415601919609,
 0.6142374843111345,0.3480437223577807,0.4387968848346022,0.1741496265171878,
 0.9813337713443961,0.3795039258824745,0.6327964877426911,0.9137545844968303,
 0.07537929709659785,0.05553929438034033,0.6077570326198626,
 0.10225628395272923,0.29440945610795966,0.4591660587509324,
 0.06792208830605906,0.5363485911760836,0.4944889438068747,0.2556953603819009,
 0.03565909304655879,0.0316428618185155,0.6960533308349915,0.5897206276007516,
 0.7495911696477806,0.6220214124737936,0.2722059261434159,0.1530379213659736,
 0.579684565604519,0.9937439943602419,0.4358882953237816,0.2972883802075862,
 0.46177149047578636,0.11117677636481527,0.9719110204032707,
 0.14052510216397263,0.7254456126321012,0.819831432566637,0.5784021842331132,
 0.08324578816038608,0.7634000384712989,0.1074697372671849,0.6270127471015646,
 0.6142612648868555,0.4535916806407817,0.7739628167566271,0.12911494235410992,
 0.4342253895489996,0.35718963331382625,0.12893502103715382,
 0.6044278659303431,0.5876289061183139,0.15806797278581775,0.9240979499610025,
 0.31901401241632776,0.7467873246630364,0.4326407574940485,
 0.32169730359394777,0.12147350709318105,0.738723421520246,
 0.46232504422040654,0.8649791003793836,0.7795215133833278,0.5306917820347448,
 0.15454670132707748,0.06493251440229209,0.17881838867189226,
 0.9073732752599366,0.7705125720977292,0.12885143443216351,0.5759597855752552,
 0.4173808981234681,0.6551440867554532,0.6705260401231069,0.5921508042965541,
 0.03974965022305044,0.08054818079150083,0.7929618981912371,
 0.5105412020628195,0.41519490345146337,0.46899952731386974,0.809902873399585,
 0.9909990068461284,0.9631391760404717,0.9550026750995619,0.6923100796499412,
 0.9317141693698576,0.7654929674278108,0.9075391864847545,
 0.045704889858464326,0.7916257770649778,0.5045511053481573,
 0.9621364171993355,0.0957345009976065,0.1290016357212722,0.415332708574112,
 0.09918680725492612,0.7654280029920958,0.3770924944349585,
 0.06471848780860201,0.23463580872149947,0.22339465815526194,
 0.08312808057781873,0.6504031962247051,0.3257478789565521,0.5986081269266351,
 0.7875191271555039,0.29940281223694,0.8991544692743771,0.6350451436765322,
 0.04492386866170539,0.22383637494020547,0.7314894044294827,
 0.6660149528766164,0.34770904427503824,0.6289072164917682,
 0.17323967322198996,0.07133753081459482,0.5060855234438943,
 0.20444820626497762,0.1607176874012537,0.2982135259210571,0.6877649198932919,
 0.29118720446781876,0.5503131660314635,0.1481196129208152,0.5300098950123875,
 0.2562600251351633,0.33046458796673683,0.657436033132005,0.7404428241614975,
 0.9530500220229219,0.8213386036091077,0.5340510199982201,0.8490486368018567,
 0.5463091797375188,0.4809837936661393,0.43116881108212635,0.7583805163514726,
 0.5488450288215663,0.5199094496862665,0.30297353473608624,0.7894409918195755,
 0.1901770527474751,0.8548241258296865,0.12981726402997562,0.3735674605763909,
 0.14576594416136957,0.45673045464225226,0.12723010053311956,
 0.5715563995366666,0.40411826641783477,0.15880820421594932,
 0.49056481313437894,0.43332906388656023,0.43400462430647146,
 0.7024931836816687,0.07010918443001535,0.7572780844185772,
 0.14904686281893986,0.43363134275951565,0.831860573579143,
 0.17403481322852998,0.0029604571149970393,0.06670841634463631,
 0.27937021375972415,0.204483696094645,0.07317383072105899,0.7163711487425566,
 0.9876755303140794,0.061760868986418305,0.05995911110254737,
 0.4471843892791817,0.6479495918735394,0.211459568609639,0.24140636988307462,
 0.8212040428802682,0.5306402686107223,3.6538451991630083e-4,
 0.5442550774252672,0.5779451344379765,0.19070392507693945,0.933421359374934,
 0.42133590983940217,0.4142195630226051,0.4523715146110133,0.8168470783529501,
 0.48238340406634195,0.8393162353334538,0.7402895290641143,0.335524966136846,
 0.7149650826422411,0.35588706230519807,0.26988084878632324,
 0.8743541481376528,0.9389191873747353,0.585239991163506,0.5096316527814158,
 0.16970317339280172,0.04971013858586115,0.5255066722996361,
 0.11611017028839898,0.25362959405104457,0.3003389415894445,
 0.05388921622651677,0.4458394717666443,0.8806821762887072,0.5947821997894095,
 0.2211256553975759,0.5196537985789567,0.9309578237502675,0.8620905637045388,
 0.4681002066038651,0.5380564508833052,0.08489737010590914,
 0.29770996758419155,0.5355332328479796,0.12634115647345967,
 0.16441635305588032,0.47293501170607644,0.4468028337857042,
 0.48681943072180667,0.378045774306241,0.7263904354792856,0.8460835430640796,
 0.21863432220219747,0.4841952235793039,0.0337805237241815,0.7765439278508377,
 0.8083107054951432,0.9911119248049334,0.8351909302373458,0.4689273755387473,
 0.4818690220813028,0.9067657407356354,0.7825161943212546,0.3120931949070399,
 0.00468752997471833,0.7449608332071391,0.8756348775921063,
 0.14521903169152084,0.2427448126017926,0.22899327510968837,
 0.21868335005605466,0.3611210197254139,0.382326653179608,0.825899936127367,
 0.2596744958876668,0.43212354477799364,0.16665870421417373,
 0.5988855844952176,0.18472084179522197,0.12321677661801467,
 0.2543843133525284,0.4333636822825274,0.6886291748109326,0.45282121604120906,
 0.12097695065789604,0.7153921666836154,0.6327486597588403,0.0475285618689949,
 0.9292408511104367,0.6947924742177252,0.2181452223533633,0.969162444782266,
 0.36082545119071285,0.5922111442014204,0.3283188574795851,0.8181045404951023,
 0.8601510774091886,0.35876798450890857,0.592243607592853,0.8506950387008876,
 0.8258895611144834,0.870805533846147,0.8305848609414749,0.6888471540743566,
 0.11092586062635057,0.6728531665516909,0.6763955627772373,0.8087922245119343,
 0.12559667519248374,0.9863938554797677,0.18718903576374024,
 0.3055084596541554,0.42635926965661997,0.5236538678473134,0.7292104821196399,
 0.21579040711515463,0.36001898059898396,0.7178289511361344,
 0.6586110547305899,0.2414836264093525,0.27214976308774497,0.5777644402236833,
 0.45284604666196115,0.3704644972246429,0.7001350077463353,
 0.40315692067648046,0.8247955188317379,0.42643613573903827,
 0.0699268450774001,0.445515129594383,0.334212030573685,0.45236379195128784,
 0.8228858118386506,0.5911109379206775,0.8934080469030408,0.5713809074617551,
 0.38644521135447396,0.5381147769001375,0.5567733477966522,0.3362259515603957,
 0.41048829968256306,0.29889354190179007,0.09517653830715622,
 0.4689672934664617,0.48873597817651937,0.5164713230533332,0.5626997793128212,
 0.4728686589872194,0.9922672072073568,0.09889115127436021,0.5620327803325083,
 0.969656080482346,0.6468775878627104,0.4583437103219137,0.7652145190805928,
 0.21894439347612282,0.24680959528117208,0.5032947997289862,
 0.6939211886165872,0.7149255930230629,0.862780896961619,0.3870270061870642,
 0.01926335468636209,0.8332659855359101,0.8214716807167641,
 0.14701280158269525,0.367741875604503,0.7439829072419105,0.7442985130042832,
 0.007188320685470928,0.12464226597056172,0.01840186635184926,
 0.8480232192504141,0.6212846174498975,0.6618518932760078,0.45863949623781,
 0.6084917756297401,0.02523558435999773,0.5481911745897388,0.5771347055917392,
 0.4262950846095266,0.5773077284417534,0.7780008878285034,0.39100523321667346,
 0.220382244653908,0.008960290657318293,0.20783465728190897,
 0.7108017316760855,0.28386269159721744,0.27447700358508287,
 0.6698477480169431,0.37017139542789634,0.8541709761834265,0.4488670105522894,
 0.49359942603499474,0.7447374357953906,0.5186327650643218,
 0.06766270519292017,0.9215446254755355,0.41046903270246893,
 0.36305032308296525,0.07243720568009526,0.4420308322650336,
 0.06225099819596491,0.8939682978844625,0.6197620446984173,0.5581690411054689,
 0.08267478164086672,0.6976446630826426,0.2842308571235773,
 0.08542464571089814,0.7728037028207219,0.3832264077449028,0.5465467728657213,
 0.6580434336641956,0.10974082624612014,0.15411989674671134,
 0.4149961894193419,0.9560633307841447,0.8011853632390485,0.2267216008796923,
 0.024725258519741278,0.2624491792385114,0.5621098073333378,
 0.8413254680649693,0.7789435898240815,0.6004412585881347,0.18228935370202182,
 0.8816649166823352,0.28430136725524413,0.3304115377526118,0.2179680029443527,
 0.7958752733060859,0.38321524087692693,0.3707489078503625,0.4407070584024795,
 0.1217287558142548,0.364296785155765,0.8050406263128155,0.7524442219154691,
 0.9382251849372049,0.5989822890200878,0.7801750950372064,0.7412316163245762,
 0.999749290489458,0.8350035275923388,0.7762299280525514,0.5123627892591784,
 0.9930529909921788,0.15438595432860058,0.6776833190749292,0.7840257128715721,
 0.8324310781346835,0.43115811408139015,0.8949090673824117,0.9448434866443226,
 0.631860173809122,0.46630264690788303,0.7748399577819574,0.17509024984426347,
 0.7828337700249185,0.7661903341149988,0.26269677744223197,
 0.014076484184191274,0.9399192255406515,0.26831464021812823,
 0.9909918786849078,0.9454374816885638,0.6740075718574957,0.557863906243694,
 0.4722311768843305,0.5639246352957743,0.6701368907401691,0.900220929157802,
 0.3659651984792902,0.8806972945664701,0.8797733895994501,0.4236478792094277,
 0.3343976837392213,0.40636040607965995,0.4108230383029907,0.946548939954638,
 0.03003425914291802,0.5810153401871195,0.6228371313570715,0.9146194252066616,
 0.3277020256612262,0.6747360931407907,0.7744648818069247,0.9189219859460225,
 0.31705812302584946,0.6907674380439836,0.4090068989132646,
 0.29378234781676094,0.875086989884746,0.9104284604293092,0.631249210995821,
 0.03090429849472065,0.11735261830821275,0.6147123772335765,
 0.3027661184015602,0.09712533827832892,0.3736839540399951,0.9286418206317755,
 0.013995798230094447,0.5487493421013246,0.8532958159779447,
 0.6879809714783964,0.19134749918135252,0.67513983487435,0.5234589591724019,
 0.9051973518828674,0.835902707013875,0.0599785970009703,0.10561813877540893,
 0.952986935780171,0.13918061152504724,0.782284357468323,0.9489982423597687,
 0.8186900375818393,0.0691592609833589,0.5233023947670891,0.21479804299881233,
 0.07340062586008334,0.9241455932892875,0.9288788885240736,
 0.40123014733611084,0.32075454854815966,0.3537777492666495,
 0.4964963611841382,0.9448160035481927,0.7791625710452662,0.20942951684469735,
 0.7061243280530074,0.528112483606618,0.37002383561557384,0.9614855754069278,
 0.5620538673408536,0.18637103292845159,0.16072672850465997,
 0.20555120979501906,0.08612800493808281,0.1752242802444316,
 0.9705419911978935,0.33731855077168726,0.8609335854502521,0.4429561155292402,
 0.5618377182201129,0.08500150550910524,0.8671375765204771,0.1117476213756583,
 0.7147554245602302,0.5956046211198143,0.47069783363935225,0.9343182390829805,
 0.07653744064337442,0.051697709566223216,0.570569412758907,
 0.8988926009268075,0.9155390912743234,0.9949513108516488,0.9060422504868815,
 0.04379070342017888,0.8295487769721867,0.12750749029841613,
 0.05977335826568775,0.16818530568229362,0.8775059855487851,
 0.5678517353764265,0.31220730796842466,0.6594162816531471,0.6992977725243831,
 0.10232057057117683,0.8307560554462678,0.20123686967385834,
 0.2540831135830899,0.7899257760014837,0.30400497133356263,
 0.35783347789805786,0.9557875220960154,0.9781390123355299,0.8440791717072692,
 0.010271440191138792,0.09630789725948308,0.11613637780423758,
 0.8820135661590245,0.15638263901418648,0.77997961907295,0.2353121084593497,
 0.7131100937135715,0.05841461327420827,0.4108855964039263,
 0.16580961240541003,0.15298082144610203,0.4178250627085238,
 0.7155089015857121,0.16873840772036874,0.5714890170525443,0.4442841996396589,
 0.8302213444326769,0.7428263569690206,0.6942630510894435,0.14587378283020658,
 0.9878611823949185,0.9460977909198515,0.7928072608069102,0.2641499389566073,
 0.4579843679061636,0.7976821409163278,0.9453712745047399,0.976668204143974,
 0.383045042856209,0.5652913995447423,0.053810571757558434,0.9740091692268182,
 0.37769280197137123,0.8109499443310384,0.8944369127120243,0.5059818173608087,
 0.3028834915859535,0.9564111060387731,0.28810211212550507,0.6361989692261089,
 0.9702386669225918,0.4059639003583606,0.3300248991692246,0.7361133599791355,
 0.42138574388144234,0.2597677057993407,0.8075753724103294,
 0.28226360682120766,0.4388411977169593,0.2521716577265072,0.1401280700748635,
 0.5801290036965149,0.7527206021316009,0.1894802302646783,0.35522981587228863,
 0.33243657079892086,0.43494614081651695,0.10002019771472315,
 0.9831461477421398,0.909871602609661,0.6428065144597904,0.2701906102942153,
 0.33139653628280497,0.624196975840814,0.21717143115786908,0.2289525668738437,
 0.15425017715809908,0.8058880115842404,0.3299858823340629,0.3145918633543876,
 0.019681145901800146,0.6135785117414504,0.6538235320381786,
 0.6895112729946027,0.30504411692315625,0.538462852184772,0.9651128363772778,
 0.8500875085109607,0.825505373917494,0.40948140499660246,0.1059478045919221,
 0.9257227521335195,0.7905528586635344,0.04392949340632368,0.9332204410115356,
 0.8579440582781632,0.5866809909873463,0.08501593068030777,
 0.012165042503555434,0.28499482326585235,0.5389263342405967,
 0.6909910971379116,0.4518189290422451,0.33417480555467083,0.7334901863849724,
 0.8987251653564963,0.13761349868088835,0.7284846670365834,0.8272338268564516,
 0.5839344790498906,0.2575812009622286,0.6893459931230723,0.13655886527055405,
 0.3118436386915857,0.49780942638006775,0.8762709790826568,
 0.20077899813597788,0.4015858462928116,0.43218040832626836,
 0.39844092524189667,0.6007280252806031,0.6305106626164849,
 0.0011385391827449798,0.29919001553338787,0.25624560018607745,
 0.23850980728979554,0.7905085565634239,0.9682924185422976,
 0.36163823601637746,0.5941781615628433,0.5173389802665957,0.3429510419826426,
 0.6402298608656368,0.840742764505136,0.2600536992079592,0.10237934880064259,
 0.6781285805013976,0.33207986384006993,0.9528693444902924,0.1787115290881276,
 0.07250272818426295,0.4313869991278162,0.8903851884449352,
 0.46889044179997785,0.12978083144675379,0.6656257794717035,
 0.5122301254228834,0.15987968282319953,0.8504242846661237,0.5736621853377017,
 0.34025570419485485,0.06840175173054808,0.9652351174390637,
 0.0033609459106056505,0.3144406527523842,0.39921429743345893,
 0.4730583491574454,0.32915417997845753,0.3644066748668089,0.7432520949004653,
 0.8835903126337681,0.13211920294005974,0.2990081435747709,
 0.24517924876975905,0.5019520745319401,0.07268147301688355,
 0.7322568734453812,0.7460531264395204,0.7165325203074893,0.09253523942579878,
 0.16365930945471774,0.38075857925278644,0.8610266923232297,
 0.48011439939610034,0.018420660221916685,0.11577640141376666,
 0.16288919496701293,0.27371869558351425,0.41494651318395326,
 0.9095075577270196,0.4528825801253581,0.7858951688929143,0.3594124183002827,
 0.13567378678123376,0.9662402679115417,0.7558696702219891,0.2856219266807615,
 0.058669930413356775,0.018542276700708515,0.86335249343474,
 0.6244454923520817,1.7975833078132109e-4,0.046167050023020906,
 0.6681733671413823,0.6604298683904587,0.6238931615486757,0.5196123825019654,
 0.7768408971492291,0.7564969504473077,0.6436275286833172,0.26969849296980886,
 0.4189021344312477,0.24022875925574572,0.8855352462634319,0.9015407292320811,
 0.33831083196079437,0.3888986969856816,0.9647909051510536,
 0.48509622282703735,0.17741199372947647,0.6436067765817931,
 0.020079605238706888,0.12742391851050971,0.3238739418192652,
 0.2817783320577545,0.27296427132987433,0.26142288815401504,
 0.7006354264548637,0.2212982353327504,0.673702696660902,0.1651925893536481,
 0.20704563910819807,0.27979297103804024,0.4182411609487836,
 0.34325885469538797,0.42762299020214956,0.19390570841526455,
 0.3623053613431205,0.21096769582708486,0.6803973117743226,0.7530492775203371,
 0.3028230792182419,0.8915766457108856,0.06970414591270879,0.738251460089636,
 0.11497480058549947,0.408078459562763,0.9342431998726105,0.8216056857886873,
 0.37934769065478136,0.532797168591991,0.41760118108161837,0.9633492632530771,
 0.45899527017867525,0.5826454250732809,0.9622415354973519,
 0.29849010215341687,0.9129723508743758,0.4640915150919098,0.1878914824974729,
 0.50931333828792,0.8009351169900741,0.041016661172111446,0.666798119901747,
 0.8683322663959306,0.5729994681850148,0.7005366115963562,0.8261294942733398,
 0.9233538918792538,0.20625241165908514,0.40111027198662375,
 0.051471472810054375,0.26289725997674906,0.04865267090873937,
 0.1224802287502913,0.17404427928924648,0.8760371277068646,
 0.16165143852722053,0.7530705709724103,0.0969724174862433,
 0.36686809932124576,0.3051608938394844,0.8710716455774467,0.7918705465952189,
 0.7782685001127093,0.2872279357152797,0.5183312187753225,0.1164381939029937,
 0.13455955319952184,0.1131056510552313,0.1273430743258388,0.5587592036416035,
 0.19848860621789666,0.8177027516742783,0.6525583807650566,0.5478876344963707,
 0.8621193805320116,0.8975204210466978,0.7933372159159542,0.16061242491263428,
 0.89855357136113,0.8883350924574492,0.41862772711010965,0.6133865070175069,
 0.5668283306358165,0.7898134884922653,0.4687314763939804,0.2770440198915123,
 0.6112777735091126,0.8800506663627207,0.4879633095828678,0.4841520940382984,
 0.8008523768900233,0.7828569204012474,0.9326615021745853,0.43930118202556356,
 0.9742399063682396,0.9283500464893053,0.3152777100263984,0.13204271530552258,
 0.7073410305895066,0.3248838477142819,0.5704812740339332,0.801887652506317,
 0.8441624769437861,0.6920348566998509,0.4009680870362152,0.789822435317706,
 0.8171392823182853,0.9840327378329646,0.7196460457573798,0.1563262933913402,
 0.1624502825978248,0.8421318061887292,0.6815940948751693,0.12216205324178608,
 0.05108586529943826,0.21419329158543066,0.49105613910012713,
 0.9796788174139186,0.6207425201272982,0.07465732397463554,
 0.029760542493835196,0.511267160036893,0.6022483859535639,
 0.13774081066391575,0.9315016384243712,0.08942034888643968,
 0.7321008678835182,0.7394663880100718,0.1351190411997819,0.7613420371553513,
 0.8932425103928319,0.2561323274508922,0.0876892274489195,0.3535187122213551,
 0.573390357302297,0.01990160061714319,0.21550128134443625,
 0.05308190090393994,0.08705138045898797,0.967822332824386,0.5668659914593803,
 0.5769949618687866,0.385345445771491,0.33082323313335027,0.987603068561825,
 0.7205217460450009,0.09982451532752323,0.7737842258900746,0.5335707280713029,
 0.9420987468113919,0.34517191571679573,4.3178571041102565e-4,
 0.15261190553075665,0.9417823617740265,0.41572066784212747,0.686481690773132,
 0.1800347748993154,0.984988505268036,0.24641544971195273,0.40875503886526654,
 0.9913860197602472,0.1955255089286273,0.045155399498251314,
 0.3073551987663734,0.6467346567819429,0.18116999301601822,0.648573283585941,
 0.31460292541371715,0.13714504147455064,0.09566625894023262,
 0.2983801885439229,0.8697077714086032,0.9858367359481591,0.730047052453894,
 0.49102070360898065,0.5640392051848413,0.1843055181394916,0.7745403762877164,
 0.5502568529937042,0.9328538403550678,0.22189799199498295,0.9515134723607007,
 0.8156254099245853,0.8681936600252674,0.6907573509301588,0.8377433792712713,
 0.815503863724025,0.39338552456226306,0.34914138260873917,0.9710374576865731,
 0.016839881243166488,0.39313976581330434,0.8605840166549961,
 0.20361608541427145,0.2894650287698928,0.9913910245082733,
 0.18072779388149152,0.3927912238020107,0.2668708147383674,
 0.27839222422113385,0.5027724070822643,0.3552474429308946,0.9955460263358971,
 0.1866930203712489,0.8326885601076229,0.7919698649530158,0.6094996555968754,
 0.6387334913908062,0.6283177712914874,0.24887674304398322,0.2474732841468461,
 0.26797271044055737,0.12191217142246846,0.24756531693500705,
 0.16310152133957734,0.41456042345903243,0.40179856112570467,
 0.8380368755574821,0.9858754955815725,0.9859355501657282,0.47534971669644055,
 0.13835567740666488,0.04409361369216025,0.5038139393681884,
 0.6127277014057463,0.7589778155552601,0.10415333938755367,0.6883445276046399,
 0.6809743538153454,0.932016215557109,0.6008220287074711,0.6984057018293528,
 0.21010659087591743,0.10580847190184772,0.483296741730493,0.8886243140446476,
 0.11492956364382778,0.15363982438805146,0.0981386296248341,
 0.9519514585672847,0.8284368322281478,0.16141915734546364,
 0.43142759399014197,0.6692551955775604,0.7044303465415624,0.7131151386420206,
 0.4309912569885662,0.05466075251788005,0.10145425619043658,
 0.2907503246914125,0.48480749655016087,0.6138489648634994,0.5283294184257988,
 0.15970536390248502,0.9190087179460249,0.6815999573661007,0.4660544644879824,
 0.6478712882784702,0.34273854493707445,0.6734402688674108,0.7286047361616558,
 0.111761428994263,0.8901503793885264,0.35809835769229725,0.9772982677733365,
 0.267818277860408,0.37689313328462504,0.013466604528281745,0.361732802146,
 0.7030840347218996,0.8743129229201042,0.7878506831095429,0.08993314611429248,
 0.0995100924002652,0.971274770712449,0.3210930752964938,0.28909524559294997,
 0.6655406453892485,0.46988567414680915,0.44025098699022625,
 0.17433676133241782,0.15786760822492873,0.2768038420998341,
 0.8782817322979872,0.4415517093424871,0.7895062443652517,0.4140091394587795,
 0.9404681812489072,0.20772589992514845,0.191512508855482,0.6496452734053658,
 0.1303470726723236,0.41506494854400133,0.8517866625550023,0.8544387502286583,
 0.8680495589487467,0.13467331027379947,0.5628771243402075,0.6674790544228076,
 0.08912081861342169,0.5712314250751099,0.724592859544194,
 0.0035454658305379416,0.5159846479903482,0.8397651711146238,
 0.5061799073867795,0.4409713820652529,0.0931925473658114,0.5417353366805238,
 0.3465328961925529,0.0185298834197839,0.27825986623849164,0.2588639106395051,
 0.12125155278412258,0.8024929573210852,0.34395737045952723,
 0.5695620797919316,0.5873789556635904,0.9996556815860468,
 0.010456561913211648,0.13200623875561168,0.20943626661288606,
 0.7879563497758196,0.049070125458758995,0.6741893767507348,
 0.5062899589986054,0.32141583662228557,0.4932300689164957,0.8243311195672179,
 0.23538399165715307,0.8115302115471401,0.10240295834357216,
 0.9409168721673503,0.15479101265868767,0.7503003168189945,
 0.41495940083007643,0.2754793854937919,0.7833549480606476,
 0.04541423465291938,0.7584306293321814,0.21678270691873858,
 0.23171083383320568,0.18643393210928472,0.2514917537654331,
 0.09570166199017283,0.7141240839021724,0.22101874292654922,
 0.3947761972592192,0.6536991308529396,0.5961822465806123,0.9608386612738051,
 0.6553934616729861,0.3960472065473247,0.39483973203511824,
 0.16709901318707232,0.43048965445279297,0.6510819968964227,0.539270283055337,
 0.43656036123064657,0.6633043917086755,0.678257843961105,0.1268142799044325,
 0.4770751824542979,0.8589282512814508,0.9900817889448285,0.3242157342317027,
 0.05487354901466124,0.38240996943202754,0.8677080491032413,
 0.7091609436599877,0.7266238512805688,0.8677184550746,0.1723616286886832,
 0.5911072242256563,0.050894974669554394,0.3462184522702839,0.60083983131139,
 0.08952536691342461,0.9141492312195463,0.41482174983927544,
 0.7147668208562031,0.7765498609110455,0.33342634676082006,
 0.23448831124733938,0.941598608778866,0.3714581757574261,0.3788095506357625,
 0.39122779856744816,0.46429547526094517,0.23742314034451506,
 0.08943720695343205,0.4677446260820308,0.3604386404834202,0.2767751070286396,
 0.9314535589654244,0.753333466079993,0.8260276165343994,0.13622070847701195,
 0.36797547798972685,0.8755740935261435,0.12548333791417376,
 0.8358277134927083,0.8420533444081784,0.6512807992362065,0.3890804260539892,
 0.8447821422159629,0.769516358089468,0.3711668973995992,0.020091377946190425,
 0.21968147803765137,0.8770214701758817,0.5257737649488861,0.5878496556203378,
 0.7781189835724612,0.07052655575115263,0.4260438448963041,
 0.32374132343149753,0.6564015824906072,0.6851949625179518,0.7933447047950221,
 0.5184250743525878,0.9929145526934497,0.27856070668578325,0.9770038549190462,
 0.6139949612207892,0.8263788570702195,0.38441969866575354,
 0.21385090561244524,0.3674843511446587,0.572244921102314,0.3323038562098185,
 0.8578948377609604,0.9573541023496329,0.11356624811204252,0.383257313866044,
 0.44099363084480636,0.9307528023954251,0.045481108351963506,
 0.4608356887303734,0.6098734792681682,0.7503248080587718,0.27473000582206863,
 0.42114732564658086,0.082545773045405,0.0560246322750686,0.7843881465916966,
 0.7221086845300498,0.5138961342240497,0.016019879887457122,
 0.19188313247538757,0.19857889030507803,0.8883007798035877,
 0.49103020493181826,0.9330306691996988,0.27966098734313705,
 0.6083842346596076,0.9098277353923999,0.12074047329249593,
 0.15740544598217232,0.03678487335528269,0.5346920265703642,
 0.4229118213307623,0.24944664886495127,0.85927781892873,0.5756723406386695,
 0.7103460424942496,0.847109406124132,0.9696662486918329,0.27616861534301385,
 0.13991413941306607,0.021011724293379053,0.1996601939729914,
 0.21375075068417715,0.7330499294127091,0.36798299070530516,
 0.8734333886693072,0.873185212915822,0.9456861993567759,0.7189227221576269,
 0.9531939755398819,0.2176961017884318,0.02168109106016103,0.8533857533517998,
 0.7230420499229907,0.9273977064133879,0.2663678712795786,0.06302005483100548,
 0.2866377559128731,0.894672744548789,0.508032160424015,0.8384538069675899,
 0.4946944781019772,0.6190802270771824,0.8180221644562303,0.11522707787388697,
 0.07758522700318748,0.14836248358203963,0.4057638944027697,
 0.4112998100537054,0.8465306534140987,0.4878658987830147,0.3313535588859854,
 0.606132876587181,0.8322963846174343,0.9159378864252059,0.27390752608908897,
 0.5205878980545915,0.3890724985460847,0.001000551335147426,
 0.5047308999218018,0.09055857991078553,0.7031244649112529,0.469561926699145,
 0.029346980119789867,0.37729788748668536,0.7357868866174695,
 0.6333916171924541,0.7283413971331297,0.6650688265104205,0.615532917477249,
 0.9157548145669396,0.5912641609126418,0.38376336828329083,
 0.005313869028339635,0.5229818992708121,0.22301414188743152,
 0.643930913023963,0.01647550471063719,0.5841709855956065,0.6759210396355676,
 0.36734860485222487,0.3094745973770171,0.8377334436172275,
 0.08913193832948452,0.8294924047509623,0.8766178823360792,0.5940588931195575,
 0.5684176078634288,0.723991968645347,0.4720092844282322,0.682204466942791,
 0.4986481841917352,0.020902464059144865,0.23425031211038805,
 0.543661457781295,0.5647483413337631,0.845584935881207,0.7022301039388139,
 0.01627772124400295,0.5100766374540258,0.07014889681041692,
 0.6048433963197571,0.8827376491912691,0.3768646872414003,0.8343105833168213,
 0.8726055628496322,0.5269716072561138,0.025193433462775583,
 0.3105117540546445,0.34115922338981475,0.6710067329258981,0.5367349010320351,
 0.3683177888926886,0.42142440792057423,0.9995735339904943,
 0.19959282215837404,0.9186301280412618,0.3215608943025101,0.6037978126465176,
 0.5892046267633564,0.22055560957241782,0.8492820914233707,
 0.10164717145250601,0.9694008788041324,0.676005049427648,0.9576760007044316,
 0.05614786368413738,0.7352619579403313,0.6575381387544005,0.546349107858317,
 0.11381273923543289,0.7765909057807238,0.021926473064295737,
 0.7478962603788035,0.2716696882195627,0.5430468402286012,0.08565804159128543,
 0.5321745569789739,0.6673981290884418,0.2696518266589618,0.4987945949533312,
 0.20625926820173557,0.008895528062334401,0.5651806377414076,
 0.1742439895911314,0.7198010238953811,0.5866873346517442,0.7377357172630198,
 0.9524899249214615,0.5323876983024395,0.5846191699101887,0.5627472350757781,
 0.43428203764191897,0.7925695400273095,0.9119987823813562,0.722738506484687,
 0.8906500844013475,0.8939924915890435,0.7917014318860004,0.5295446792489152,
 0.019382832125780847,0.34505707207532443,0.24722875922957088,
 0.24140267532183246,0.8858868353850339,0.23774773630822366,
 0.2314560900703122,0.3715071578577659,0.05502756049154123,0.1877876132734898,
 0.2659864603234766,0.7974588794108269,0.43076500978109133,0.5268378153844018,
 0.6316385208279846,0.35924082236655663,0.6617178839922158,0.7418333910003161,
 0.5350819782840495,0.3202549091563468,0.027857624538847214,
 0.36985727879017094,0.48627402019340626,0.9277403176983765,
 0.5304944773835021,0.2980143528293885,0.04869968004640279,
 0.039930880616352304,0.8195170440265042,0.2303704739997836,
 0.37014208730274945,0.37826838407948793,0.3042802220481263,
 0.8377057070044239,0.3952208873705242,0.8820355331845905,0.6975166386049665,
 0.6777177019428965,0.8810812018572691,0.04854817848452775,0.7684287515955623,
 0.029984050076715674,0.7306332016161087,0.12317508879296146,
 0.315023134814005,0.7184619397068773,0.4249101409311521,0.03030625935898812,
 0.21355300984402226,0.1126155835106667,0.7667086548747545,0.6270173832371688,
 0.9282823277345341,0.0038161261259580526,0.1462561389314252,
 0.4777965707665448,0.32981338495731927,0.956361174167983,0.47586564748576254,
 0.17151589593838656,0.7283024393668114,0.9220007333882332,
 0.20963728023739236,0.7832400835781715,0.40601868515880823,
 0.9599490349317872,0.5874472915214131,0.6086669627285948,0.3492972959152725,
 0.6730526428738622,0.4327747743599284,0.029977712462426642,
 0.3101147951321811,0.8387668070548318,0.05002204999140325,0.8754605883348854,
 0.910664837266205,0.1458692722455156,0.5427493701666574,0.2719294848069257,
 0.9207197428166428,0.13431822091471368,0.6498614424751068,0.8103528752630854,
 0.4861150404524722,0.19459841885316997,0.5101293782254274,0.6969306796517644,
 0.8012987177504126,0.5961442741984591,0.9752871182216677,0.1423712104227025,
 0.7042039955291836,0.47872765017455,0.3701564937278511,0.818340001643053,
 0.884485214110431,0.732079310210086,0.031204780244445285,0.3680623771288205,
 0.8106577594191071,0.226162324397708,0.10423924785635541,0.04540957388700717,
 0.967813638071218,0.47520655072003215,0.3453115273175591,0.2809139790604226,
 0.7082531898711295,0.6994739272632249,0.029849957421488993,
 0.4998848520769812,0.3884043406023058,0.40226299931553466,
 0.04092764698098428,0.4782701410247041,0.31586114786205965,
 0.9624223910177515,0.22046252084615947,0.7023237257057123,
 0.040513880590533846,0.5011560528837888,0.23912028656273332,
 0.43664855017897275,0.2718051877341444,0.2692010362571601,
 0.045470717388659376,0.13381425748385212,0.22636916834977994,
 0.3319670565887859,0.9821023901190876,0.10808767261484031,
 0.26435622652942814,0.43303233346662706,0.6323461210294629,
 0.15366970537648983,0.5165713094256781,0.09134990646337315,
 0.21487636635037566,0.30437460729211196,0.6729443923889012,
 0.45720552727554054,0.627331756072717,0.5174773983862209,0.28542358101396537,
 0.20209158159606433,0.18303906447455753,0.48950683436981735,
 0.520549666324694,0.775695938191114,0.5845232957771426,0.6867846478859362,
 0.13495580795428586,0.2948404094144277,0.412195741757989,0.288933453813894,
 0.004771321956175467,0.1104526871465008,0.6303188734627035,
 0.9109828870834532,0.6875937789873213,0.36097225750218875,0.3127322764260172,
 0.20541721241366262,0.45082947356832226,0.4907015519601118,
 0.9362290141690751,0.20633572650620746,0.20556047476908812,
 0.9945021067060846,0.07715089389221841,0.9665345892343741,0.5355248967409946,
 0.4853008932173122,0.2069321128212479,0.012572923308509454,
 0.22302676736143345,0.5437587207076934,0.5538933947506364,
 0.16782110489724567,0.4925207282177856,0.52454136289683,0.9021981558473392,
 0.4147225071436017,0.448143491791105,0.08309103651419991,0.16061928642712675,
 0.8573921609260395,0.44984938599071955,0.24769089922964438,0.811535254248631,
 0.4760292904137341,0.7003578912481147,0.6457952631254235,0.4215088191543538,
 0.7072627075839787,0.8584447640761906,0.17921268357155706,0.3376380710655261,
 0.26300411158531456,0.08117233872405272,0.5394155925516386,
 0.5048686803740174,0.07002262322038733,0.10161194005381957,
 0.3778189452750407,0.3751907950787967,0.2919728356363175,0.05010794103848015,
 0.820236528667975,0.29145467791255375,0.8699563184711788,0.7089607054776208,
 0.9857770537711344,0.25044753885960946,0.6319843022480977,
 0.11431668657361804,0.9862686770757827,0.4219194997331991,
 0.29565462891703465,0.01744084614596675,0.43864899497373677,
 0.627112397368777,0.1428699010306268,0.5577631857359284,0.3471111965667728,
 0.4370705318648882,0.7642237539222849,0.1229926574602751,0.9069359557324539,
 0.4339507364991728,0.8859952912511334,0.647616972152265,0.9309912531547555,
 0.29196372218164357,0.7021529990097299,0.7029186747208959,
 0.027801289073699276,0.792600452220974,0.674955312989254,0.8061158557575191,
 0.7451483814396701,0.5760550575770725,0.8664334309606428,0.8761929215848472,
 0.9414570185631355,0.4246869376800334,0.9464619810627236,0.9865504879083742,
 0.7776326524757834,0.32354197142835295,0.4767919149800057,
 0.23736779064815128,0.6615876106175899,0.6194817085647216,
 0.14173031254174207,0.9991010889061434,0.007539204907861086,
 0.11597136613638859,0.34628526532274306,0.8037885709956263,
 0.018421210467501115,0.4911906478393251,0.3638066534790765,
 0.9237467006465104,0.541224618165391,0.7727749405954559,0.5825626791568913,
 0.207992627475446,0.6395230612947386,0.07474770868951641,0.52498267473757,
 0.6895071372325716,0.9627474972633819,0.6121030054657438,0.42191135163230475,
 0.7134809998868353,0.5476966591186748,0.28389545519521,0.25910150495237305,
 0.17246524632964233,0.23912057703014644,0.10424039479288216,
 0.734385662412823,0.6669139788148916,0.9189947099776822,0.4345404369561868,
 0.6187758778151469,0.7378888761144723,0.4932864349924925,0.789506282852228,
 0.3640488559301328,0.46972691853967796,0.21322404031159148,0.306786038402185,
 0.5898643069005951,0.8971794797869699,0.4322526083840833,0.8224731387499653,
 0.09899615199128631,0.7055020903716882,0.0049676914348293355,
 0.08750185181810644,0.4068150570919471,0.2323047244581663,0.2868740229301653,
 0.9120626621920034,0.7028971520734405,0.7590738180859433,0.0657947408975208,
 0.9773403808145221,0.8786489065042993,0.9693668580579378,0.3087054569615948,
 0.47957048987677475,0.8379835078628431,0.7313156695630648,
 0.37090571256720795,0.8838921838971262,0.0928116192941646,
 0.03611919248563422,0.6536062724953238,0.21693226075699346,
 0.5015602374294256,0.9635371986079005,0.27125870369858796,0.5240558573436641,
 0.30418061895952997,0.37440976293485295,0.24454898612296017,
 0.7069917504259841,0.317985395369915,0.2042723008168752,0.27335511073307517,
 0.4699626797410553,0.7044913821560557,0.5842640629315733,0.4673134207107007,
 0.3638669364002509,0.770316849740066,0.7258495486995094,0.6467797338535762,
 0.15059773172394164,0.213543090932102,0.27053577531571005,0.7178630539577151,
 0.7227221002099341,0.5048829895204254,0.09553157348517716,0.6366848619972325,
 0.7473184130087831,0.8522579590585864,0.2084155044020537,0.9622561032035232,
 0.4742618432656026,0.449719159059932,0.09492776366432953,0.21057486452948615,
 0.026221607582712525,0.22431620638828453,0.0671483105646199,
 0.6709977389196343,0.0457651999403903,0.54106637527508,0.569729330514917,
 0.1948061643057718,0.30487621905490236,0.20942449763363147,
 0.2636526178130598,0.6679596599848192,0.815152767046386,0.9588402056253509,
 0.1014385501161843,0.4409078023040831,0.4547770758254285,0.3482295859539817,
 0.5562861376567769,0.07249702360857224,0.555053887692283,0.7615404308798713,
 0.07888798024978005,0.4729039274178376,0.04076445609618673,
 0.09854695439165506,0.594791571862303,0.9877473256407863,0.18954271719893367,
 0.4166344234685734,0.1316659572020874,0.5516380385410875,0.33374938534747056,
 0.4473976995638328,0.1986681254079936,0.3356122818512429,0.8032961152813315,
 0.8579232963552084,0.44409982176638063,0.8937939487841904,0.3567221345503414,
 0.6694416384683011,0.8232848471314911,0.8139237870581275,0.10132304818992366,
 0.6851814316316557,0.7566450828554386,0.6385493755890571,0.389339837794086,
 0.054211234085646876,0.305305782359386,0.39033829982786494,0.482074929478434,
 0.5799459381586891,0.20152399861603987,0.9086108341349256,0.5751155986328659,
 0.10805370628423505,0.7475802087306876,0.2348669298378122,0.988332182756504,
 0.5655770293254035,0.6861676604254441,0.901472789976921,0.00823516055825868,
 0.12359225461019452,0.033717054332955576,0.5411400841730708,
 0.6128181833399104,0.47335299949632104,0.5724494270394598,
 0.24766106703659818,0.3577923653896826,0.6088983113320884,
 0.12489880986800705,0.0701066368198261,0.801214322469041,
 0.0034224339958833783,0.6421843218012152,0.5280830342410666,
 0.24862855290821062,0.6018984803346369,0.7596424440717492,0.7083136314026454,
 0.5596490959225568,0.4864274069116391,0.9967347465672106,0.3242893514803432,
 0.4877669309415864,0.7762516231845679,0.5012309277955492,0.5567010203471205,
 0.39953595897331295,0.16463780931884286,0.6409855021184219,
 0.7430657892576977,0.42402195502087525,0.3610348853408789,0.6662067577208289,
 0.5528200370326529,0.7891595364697722,0.9451861699653571,0.4840374029811323,
 0.39358768638495145,0.1721356533551096,0.05458346708752593,
 0.5301341333054608,0.516149575073188,0.9337510862413932,0.9941084564781012,
 0.6003529139587862,0.7129413958741677,0.8441272142222117,0.874480396544983,
 0.05771459345689267,0.09864414497936569,0.7677494172641339,
 0.2928309113013432,0.5487036156283487,0.1711117609738141,0.10175447995414966,
 0.38357255459028705,0.8796680986971795,0.13886859743462399,0.464077859105964,
 0.37218744489709477,0.3433503301486859,0.9674839774064005,0.210545513922897,
 0.02793687303412551,0.07738735975216571,0.5126762495728152,
 0.2421350847568109,0.14800855367631538,0.33670924828644433,
 0.26136628297576514,0.24094682905460418,0.3924822742282551,0.774833268145708,
 0.8661486898011556,0.04200077310316552,0.8244643226161552,0.4917944506276176,
 0.2244670591101774,0.6709871565464783,0.8631988475237471,0.6427494358377145,
 0.3590468015778222,0.3019825809535086,0.2486529287042344,0.9603753570326862,
 0.15308297927702652,0.6380232818145051,0.7861274908586282,0.9135055903341937,
 0.4261000864429775,0.6929205079066062,0.009291198748913154,
 0.7100689540967771,0.21406800201810217,0.05951144480431769,
 0.5097709722048284,0.2187102389476414,0.6703696732323031,0.05763770102871968,
 0.7003461198089456,0.22280685135011058,0.6881985939051921,0.5087993473269787,
 0.9177668778917857,0.586017050051711,0.9844931770855212,0.5847255177879749,
 0.6521788703740413,0.6037742629872862,0.10746826726313663,0.9582235644824413,
 0.3582545876837311,0.9314460726642452,0.7649415413581555,0.41241916434630765,
 0.8477925785030718,0.20714544694416326,0.07534830224360234,
 0.29719601335471435,0.22097778088617215,0.7265750559130326,
 0.2029365014899739,0.4710297937717943,0.20849840324028202,0.5615103484119472,
 0.9395345551010238,0.7912109382011985,0.820414969291566,0.19190470835716156,
 0.6811723255667238,0.7919479600710706,0.05764728929140872,0.5588720350392423,
 0.10301244704747581,0.7498708396715607,0.667768067847278,0.18376028079369322,
 0.4366574131157682,0.684868047657141,0.5048806581413565,0.017317506863074983,
 0.9914686044712651,0.06941953682499147,0.547828444924892,0.9155649126072934,
 0.8588812348707967,0.18860016957705528,0.16284219515561182,
 0.4644592402172715,0.09979144556175579,0.29049507220362747,
 0.8969282731736878,0.4052874690942735,0.2647406812844495,0.6238703787534611,
 0.1327696371430579,0.7505843562558822,0.19723590228504406,0.6163617464217633,
 0.8281005027763614,0.20296538961844024,0.9875253499611704,0.8764675901729554,
 0.9005358651119638,0.5023559257156863,0.4952726136133001,0.6191396418491159,
 0.9399377732230004,0.8932849945967457,0.867137561269977,0.1427578437505812,
 0.9763765702229003,0.550711817444757,0.7210351226200179,0.44243877092212447,
 0.8586635241763373,0.181001649012013,0.9300145709444277,0.30197410288904136,
 0.45872373289792145,0.3889802978287863,0.7147601294075858,0.3133543926209508,
 0.4571606265591942,0.06929691042436503,0.9381826356634563,0.2496124616880795,
 0.9818249665977193,0.7023121684172956,0.16605848689650593,
 0.41679245032157153,0.0020608764709856287,0.19734153544646227,
 0.05949569082438988,0.044628712531840975,0.7801394578541341,
 0.17439118077507942,0.5170268073972585,0.6080889584065807,0.5363536616516581,
 0.7518790446358087,0.8965195243052202,0.1694673150484315,0.45722631335756925,
 0.7191689357433475,0.06530470442229352,0.929046911813498,0.36372151822918064,
 0.6292674817688774,0.6404993223563201,0.7691576090597163,0.04489464441997282,
 0.31082038163570225,0.13656253184476486,0.026337979471472095,
 0.9956651110697142,0.36546983444780956,0.5579514798737856,0.7361601586604946,
 0.8212419187903416,0.6943101891760979,0.2776254510231343,0.43156586821856013,
 0.9426018459825144,0.2733468237272495,0.9228545859807858,0.8713377709443809,
 0.11039812084640444,0.38674320577550814,0.02551289062872364,
 0.2698294805866924,0.5569824097078994,0.08419804316558288,0.7906315195687411,
 0.8365880587693709,0.49991259444082914,0.25850251128230695,
 0.3951140230999949,0.37031992685627324,0.8984004090881272,0.5769711210810593,
 0.9640463485117278,0.38723075051629274,0.03164053134596978,
 0.3163552684430565,0.13247355460410737,0.4211727087547408,
 0.03824218171970617,0.03857651383827765,0.4141769328386615,
 0.5832504162665096,0.6435163630056607,0.768665133192535,0.13554005152261883,
 0.935329255370709,0.2801264577736915,0.9177155223494987,0.20295888539807705,
 0.2248830301994189,0.12508682377947222,0.8501717968444534,0.7991005448852503,
 0.8992195121066526,0.5924364334157963,0.10689047292346066,0.9447857544911965,
 0.31500682780648026,0.0830350814108256,0.853022461604104,0.3073258821152234,
 0.25506296616749513,0.2252134567116203,0.21353725505680665,
 0.9834684694559982,0.6198448237714786,0.9192373968811562,0.09453006394070718,
 0.9540127679678198,0.11665547542493337,0.43329757271347613,
 0.09258619535003154,0.9631331514045695,0.3114478699819758,0.6430104545617503,
 0.5237631785128728,0.6925615231295171,0.011479204372174423,
 0.6678333858532148,0.5578456137555875,0.9513505823723843,0.37110504772350994,
 0.3740276422943152,0.34275529295973806,0.04794292088505092,
 0.3160696540650031,0.6216137833711731,0.2965011559631956,0.5187538782691193,
 0.6047612208143633,0.3338149600580613,2.0921093338122088e-4,
 0.4873914394590342,0.41681958976457834,0.039006880903677765,
 0.07239620742164499,0.16401057563896215,0.04469394430376361,
 0.04548718651052319,0.9660858033965504,0.5005621141993882,0.5898334316008617,
 0.2942849698225751,0.2855724753464477,0.3693851123435219,0.7847138139137403,
 0.4902818922917287,0.47596977657238104,0.4482809061430888,0.9092278925970759,
 0.9675649001820684,0.8001997255497842,0.8825920531218013,0.725120532355197,
 0.8926127412752544,0.47702436400596704,0.016386088148087063,
 0.915733723052996,0.07802362079644654,0.7626652239310818,0.4166626039702759,
 0.739367853821993,0.43526616468720203,0.453624912024282,0.6543597230664091,
 0.7388797943057233,0.5359322611238942,0.45306922379513126,0.848649729612393,
 0.4958759244338736,0.07459003484365445,0.7621027800313123,0.2762638915420128,
 0.6190011886412511,0.02525617405274161,0.2981312235033927,
 0.03970812299817439,0.8182866315069491,0.6979710257624844,
 0.027561695883314585,0.5285343576017368,0.6578872123314304,
 0.19910842443518484,0.3858112411758343,0.3160250013293714,
 0.16815103785905583,0.1653451690656459,0.2550214235533145,
 0.19828884841411443,0.07822088799911309,0.5172680304370896,
 0.3309904897464986,0.769426492341048,0.6518713700672336,0.5525138634623504,
 0.4331867050498258,0.3768830881191809,0.15645338972261502,0.3499949316856452,
 0.3700942995061274,0.2720560016594695,0.9922764777365113,0.0469272128082554,
 0.3637788199746138,0.4233956355660746,0.7039220568992197,0.5298787734163193,
 0.37933318395164695,0.6026451559311495,0.36571680277495666,
 0.9738970775244586,0.8047921296404983,0.44779674533654756,
 0.36716077690652726,0.476220584604958,0.1379354849673886,0.4688493196546557,
 0.3565773468124156,0.2463834599620609,0.8218373284271818,0.0873133171501349,
 0.7990152225302685,0.9617147448730798,0.4498482695266546,0.35498242083197695,
 0.009952369866997435,0.23131733047855352,0.49456370634943314,
 0.5012573793014876,0.7061237128049119,0.9814262098051423,0.8564824381688546,
 0.598485111208641,0.37677286271747823,0.2691454868026997,0.7318399746833043,
 0.5393990792658915,0.25501973341586637,0.7652249398915848,
 0.18741430856033936,0.12395406188477165,0.7612710332736349,
 0.21474794642210315,0.09853333148614296,0.05123853305379655,
 0.8717152786186483,0.16744812802937747,0.6694364493053824,
 0.11533215978369304,0.024717421578243526,0.4179409687503337,
 0.2836261377445929,0.47307442238989106,0.9441475795779575,0.7595317948036724,
 0.8896419210835205,0.486390688024117,0.79450562494419,0.5223129553375647,
 0.028806056392860047,0.09932152533004479,0.2556744439196642,
 0.9030386530254061,0.8447025961613639,0.2799633719596586,0.9465717555914914,
 0.7874160599665866,0.3629808102992934,0.29607323852255474,
 0.44902982367716326,0.96377747602171,0.5171386789762707,0.13972872973394035,
 0.8377532351763431,0.8167471777197188,0.421503779279614,0.8956572058380713,
 0.6297205598208133,0.608493456927066,0.8041961518564834,0.1186545991043817,
 0.7964666490613922,0.9128660968818731,0.9592914448653399,0.9593665227957173,
 0.5395932683122866,0.24725020157005828,0.2270637754301983,0.4205232693759029,
 0.32753394494597154,0.5740089619805977,0.6889589319563415,
 0.037240680848603436,0.03866032889614912,0.5889264888868291,
 0.6935757270665428,0.35522121293011466,0.9277769567157969,0.0320550452215268,
 0.8627949573228206,0.11305850129782025,0.94166476702981,0.02681914449364997,
 0.4447792288191377,0.046945379771733964,0.8814350779233069,
 0.07094309314945568,0.49286994662318195,0.4691716993800693,0.441913969831608,
 0.751512795375881,0.2343576961021584,0.08334297746954888,0.73042736908654,
 0.7234842622286036,0.7367061265940648,0.7873341361331146,0.6946626011973385,
 0.5258511182164853,0.5762174559667593,0.4533596053758989,0.8034896207190514,
 0.9976299617254193,0.5069192595897375,0.5344513101380887,0.752843389408676,
 0.1394486494485846,0.21129167151288974,0.36630164924829656,
 0.9132691446204224,0.9150214284746961,0.4774009875061441,0.29307224392816444,
 0.2305213532913163,0.5472930786981924,0.0869286971429154,0.3823694740342215,
 0.3258888202892791,0.6341069045078085,0.680411264588601,0.6053992740157298,
 0.9062268604217888,0.5485501939137118,0.8021522459803263,0.9354822413545337,
 0.2533782090648493,0.7432598786885819,0.6121361737386253,0.9437795860688089,
 0.7085968669505769,0.3650507013966138,0.21798522662296127,
 0.09802555346120911,0.3427551637819666,0.07597596665345807,0.76431810287284,
 0.39281000706010205,0.6673242101443163,0.5998327894198707,0.9014601283226067,
 0.23150835218010968,0.9532939657019319,0.5424161614238391,
 0.20902951099746314,0.14499897889138136,0.1977341907790625,
 0.41461313283799495,0.428092510824542,0.7637513918249227,0.48454746586946906,
 0.9286373072934728,0.26248939676064376,0.6923585372970141,0.7908515044955884,
 0.48260454524803675,0.3266064349879487,0.37706397688126514,
 0.9814534223789473,0.06764416378175586,0.4505100217115313,0.7467978795805073,
 0.18636487878295782,0.5170778925312478,0.7309240933191385,0.4133418608124876,
 0.025001168869340673,0.5856569171676789,0.21147264732326276,
 0.44199675593261867,0.4626301672397485,0.812508026698599,0.3544194180719712,
 0.2833074656445109,0.09526115289757997,0.32064754675011176,
 0.3433550239597145,0.4882058495687023,0.21436088098882,0.43976941478946563,
 0.8809478837479031,0.23405453812817933,0.8403271753189218,0.796684542237874,
 0.22584620666352873,0.1947772137881545,0.9892851023302012,0.8088649896938117,
 0.21303539946650418,0.1351958380000986,0.8013487059173372,0.6199885270039771,
 0.23632177862661818,0.8681551671176427,0.2491269087920811,
 0.040127842648942424,0.036639771375465036,0.3219265286805484,
 0.5423249129335432,0.8561856038235136,0.013682041344677986,
 0.9901197750767238,0.39975292454394784,0.6132546351213374,
 0.12715066307330636,0.6528106675188337,0.7881041062461831,0.5433719333111218,
 0.2888917194243389,0.14933332529011512,0.6723249053265835,0.8284294902729468,
 0.9216897811412502,0.33687044525188503,0.9330229580489204,0.5987762881771366,
 0.7959752258463861,0.008825535397100848,0.3134109798074808,0.991964497829391,
 0.019417724723273078,0.15261550904004495,0.7664552039668149,
 0.010105584888720043,0.35008509842042956,0.7063218503703677,
 0.12371657291012661,0.3823864840510779,0.01728782028040876,0.986940709991241,
 0.05290835423877982,0.5680301899161817,0.899539629917003,0.35076252641566286,
 0.7502742855012696,0.5392231121815692,0.4637493002072538,0.9602362635178788,
 0.8962220822777668,0.05398323568615837,0.3831068991905955,
 0.35641704432173127,0.12933987287009452,0.31593525368947617,
 0.9805280115579056,0.777750622177806,0.963787396477348,0.6151072965388316,
 0.16060051604990255,0.8759377678265611,0.20610406017053318,0.624952624835265,
 0.9978902787336607,0.2557393381616444,0.47793107399982404,0.4427275878029173,
 0.22712515167668945,0.506073280053244,0.46946745001792356,
 0.09008525386297839,0.8776906242748527,0.5318175281523302,0.5270953880030709,
 0.44774078056646427,0.1398977670791114,0.2569390575722924,0.4790293112805054,
 0.8232681214438208,0.7134372328064928,0.580393982170036,0.9061745674601972,
 0.32325937757051204,0.3478508812818244,0.8028282551671634,0.2657409314465716,
 0.9167058136593713,0.8676704907762474,0.7598347201791606,0.08029790402789072,
 0.2506553922340513,0.4277108452685524,0.5491885879634805,0.03138528202174029,
 0.3475125064713934,0.998566895492156,0.3723482924581467,0.575849954679351,
 0.11172913833908116,0.44016704173073196,0.04708498217153212,
 0.3788645827012157,0.5921139923219212,0.3928559585116188,0.8383200721879853,
 0.9942696803182584,0.1554333478956884,0.2977527131647797,0.8439556768054536,
 0.6513614371729947,0.854095355523567,0.6408485339203539,0.10018488340700937,
 0.8741275838762487,0.7299143408263208,0.9368156926245756,0.8899953133015842,
 0.7759540688088821,0.8413702281642227,0.8249774947697437,0.17157907610084344,
 0.9365324692472513,0.2002202225343761,0.4703458252503203,0.591515812400765,
 0.3744045144359901,0.34535215031826816,0.4809814502377283,0.2579375342794956,
 0.2159588724268201,0.4301004174384617,0.454109581022339,0.9226741155797048,
 0.029430012384220272,0.44299595630644006,0.7295533856440634,
 0.6569583632043197,0.1737295757952615,0.3346202718985546,0.8659692341190547,
 0.25164447488554553,0.14434261018297256,0.5698291641250969,
 0.6545034061288033,0.24820567226898926,0.6751919585739927,0.8596072180351185,
 0.8249435324329284,0.9561518644181197,0.5222761645446093,0.48696076388823717,
 0.5897879993683489,0.8575962198727787,0.19967866686227076,
 0.36597819443738133,0.7456847303275795,0.025857768013209492,
 0.7314304898852926,0.8698219281894312,0.8269894052404313,0.06531975010899349,
 0.11602880835441698,0.14949952803248245,0.35964713304827245,
 0.8367148610376864,0.37743014551405996,0.9569413379865593,0.8305254598200424,
 0.21099278315856596,0.9943317194214658,0.09511344772007224,
 0.6129564453542496,0.07361477498516844,0.8909564222400195,
 0.15264740008729116,0.3052552728408935,0.9117631862735598,0.5966381836913501,
 0.6993524561247184,0.3775994368203883,0.8485825581194005,0.9206359637705411,
 0.8908908283060174,0.477616553335789,0.2510416019410251,0.06068645671753792,
 0.2127850850657349,0.10562750905055418,0.38246693034833246,
 0.06219677716079086,0.4634312143833219,0.7004311125482031,
 0.34346648213765896,0.060329721541103276,0.21624735966312814,
 0.18997877243072425,0.30406511688983584,0.32533875664742806,
 0.5861818965250541,0.5006078226063962,0.7458643946821115,
 0.036626946736980415,0.049507777104890294,0.22492566008885817,
 0.16221391633090676,0.8379520187733205,0.4246696710575032,
 0.02263738105873292,0.1532467109712643,0.2790731323530121,0.4525889444501736,
 0.872880496031398,0.22434005725948403,0.5834204943783561,0.9269668001031993,
 0.5898353217669543,0.7601896649600955,0.5346153501328451,0.46005761610355567,
 0.03030549846758357,0.802469464957534,0.09912655164206197,0.4715039491761588,
 0.31663792043324324,0.8276146688709197,0.6349377821654706,0.9993859262717127,
 0.028883030420620015,0.42318980256875394,0.250437209170694,
 0.6671516160272003,0.3864807712496384,0.21153850189570622,
 0.06686101170434089,0.9589620648704473,0.7887049168455443,0.4856151205881154,
 0.317578473786279,0.891706605355659,0.8085357171775733,0.3221079737240614,
 0.2738988706319998,0.8355080967778858,0.5705075044432902,0.25990132459460447,
 0.45993288448652714,0.9426828674200189,0.9930476601686457,
 0.03356113706242314,0.566959474785532,0.6971615374771638,
 0.058721201112852284,0.8578074153592979,0.7017708775800442,0.85701109406907,
 0.9132774806835348,0.6162600772069402,0.4480232996244836,0.8192833339240699,
 0.47929043968766927,0.3024748435423237,0.7973902209965843,0.4965907711686818,
 0.4289757950265305,0.30849576399518486,0.25958123379219333,
 0.6655437332476096,0.7796172351633306,0.10547231958556391,
 0.46237500073461757,0.06823473290401716,0.4982603897659008,
 0.7216620901693305,0.32042812274997556,0.267488323046554,0.6631143369640937,
 0.34865598388146335,0.14375549373868401,0.6515496994703132,0.97579892005149,
 0.6923150750257685,0.5599992003130825,0.7409122278733928,0.271967820197756,
 0.5436992815083166,0.6184005701906882,0.39780695271980915,
 0.49047094248624956,0.9206631101571985,0.35997003430778873,
 0.5210250421539117,0.16946059978240724,0.3047479544897381,0.2312930733555555,
 0.29741410266204016,0.8111685678705237,0.05440513475409259,
 0.9364458872545831,0.25708000944811604,0.1706909625012295,0.4214197977146412,
 0.4106118727147362,0.9607144102878084,0.7861351864324388,0.9869402172206496,
 0.2293007185441186,0.9541661088669626,0.5097875771355468,0.6795035229027614,
 0.11275824903681575,0.3828359125174764,0.015395912578899051,
 0.9176000319640054,0.8456659113111858,0.1935882252548644,0.24929111793230685,
 0.5520030570447926,0.9537669146468748,0.06186275643845418,
 0.06556235001892985,0.5769404343889545,0.28096854417907824,
 0.2275445903537766,0.8092901924342729,0.12112812190974542,
 0.48680750176491905,0.28856387910799364,0.6302861815850787,
 0.17895853787627192,0.8197218383893532,0.4636527836775416,0.1746780367452595,
 0.6726502284167405,0.20597404555865872,0.23137175244396058,
 0.7197668547900999,0.9953776293092297,0.14079313830672546,
 0.13117229979895917,0.7161047511634775,0.17345034798685044,
 0.3245150644767807,0.8288648819124012,0.8056677148289841,0.27096974258481676,
 0.4545627471445115,0.8996271183363718,0.28362933281719815,0.409631294677423,
 0.2804532774725149,0.7275311431539518,0.2439033642928664,0.5894754410340188,
 0.03176682380195073,0.7730461700890745,0.20541630778279762,
 0.1514954198149453,0.7946434171408252,0.5962008796030662,0.22008625244715252,
 0.44969019782857955,0.9001176820405785,0.4314783322841951,0.7049703668696111,
 0.44993803591729953,0.41785525697475556,0.7780499901088493,
 0.7640396515594494,0.21613486176669905,0.020368915843965718,
 0.05723094004076279,0.6159696923488447,0.6444869418674836,
 0.43451356774515937,0.9002996453854694,0.3057899082179636,0.5015268175337516,
 0.46630377023717395,0.7977421902294078,0.4776944892074203,0.2221910498544285,
 0.5218742851405053,0.8864668770452032,0.5247215389985332,0.06885912608963851,
 0.6967676047347259,0.6783362684590963,0.8925768998281185,0.7920853831472358,
 0.6503449718083795,0.6458507244301985,0.5212509487714299,0.9021226564241984,
 0.4323450506236469,0.3548422018706098,0.7658326919157146,0.7999036783826098,
 0.6677061982755463,0.3447892754228066,0.7619112291000519,0.5629630422580288,
 0.07872568751985975,0.8009856806986143,0.7486358157574218,
 0.07970097644258789,0.3228729926866135,0.09116921294642277,
 0.21783252621154647,0.4700314564468594,0.30836257956873525,
 0.6964145490483372,0.2932759421824419,0.39919515185200294,
 0.01053505118822673,0.423703174588824,0.22101367554817997,0.991102111073528,
 0.2537132898521254,0.39292669168557326,0.8288702025019967,
 0.10872136121564413,0.40911128798107477,0.062074325991577806,
 0.14539173234527303,0.2995219819393956,0.41682499125203454,
 0.47308317475752404,0.9214958506816929,0.44975464869068815,
 0.9807032828605783,0.2686160411367333,0.7249948766640323,0.7945993519012783,
 0.21641681165790538,0.6911927624261833,0.3581780636946025,0.6493855982530513,
 0.5877182533730247,0.02460017181990981,0.2269862611580118,0.7756928112517099,
 0.7534955186649632,0.05642535675047011,0.5577372337651779,
 0.15800987398135458,0.8531203244073555,0.08332793301663699,
 0.1791411384667052,0.1031875603770932,0.6679477159644956,0.4007019136010911,
 0.09147679772047856,0.6602969020454741,0.7517202130192998,
 0.01887370468508842,0.5854443567477748,0.24747774487743446,
 0.3443978577772936,0.07361449409076426,0.8924755191369186,0.369810564728406,
 0.02246150243533862,0.47298366186954566,0.20321024753779215,
 0.6930959087177672,0.152653187207327,0.5024666460872095,0.004051468271498826,
 0.7754953007619411,0.09711458046755772,0.4707034991750225,
 0.38306853792068907,0.4245818885469347,0.5457498440423316,0.6583058603099249,
 0.020228581732650852,0.34649190690661225,0.7902752376166252,
 0.03247140020238859,0.22355250398130555,0.32737712654271145,
 0.8244719920209483,0.7719878404607874,0.15321398506534678,
 0.027225570922371323,0.9466868028799881,0.32339878906035224,
 0.36062352127999964,0.7180845081965647,0.6039960893355125,0.6747351823929095,
 0.10897544821521743,0.9970180083103216,0.18230313411558696,0.117989772990169,
 0.5014940605861615,0.7215913580542446,0.3202592118426998,0.5806087878557712,
 0.10957726368337495,0.7607791067333028,0.11907762884927486,
 0.0746892853123079,0.9768283730078627,0.23439396218547126,0.8651415674189705,
 0.7707207655828474,0.8244993099648658,0.315486876348547,0.26774903227688773,
 0.23140147534365807,0.019219507159149996,0.6757980928196626,
 0.8388155476692045,0.8109959696966418,0.2896243067073734,0.32574732016304087,
 0.5008552068713793,0.6385936871376905,0.8571151487698833,0.09392584075097954,
 0.6135159598399392,0.7523438942276167,0.7444458537542205,0.4716679659415378,
 0.07845873619169985,0.6158992928196326,0.5976759293896173,
 0.09282285274369884,0.4500563385116655,0.030537243614860543,
 0.6293418888521775,0.9852870407841008,0.9571878596945086,0.8646221815278643,
 0.8766743759832516,0.646545022755019,0.39646399002437327,0.20363939252849195,
 0.2780727325102611,0.6570044556323957,0.16185052795928478,0.9831075614918744,
 0.0040262450762375,0.6485661791722712,0.5817794828409806,0.9708361966859316,
 0.41639616358896836,0.9514742815129481,0.1641738388050199,0.7048742699161369,
 0.17432818067436517,0.9923367242992247,0.767641457116985,0.3839608478954083,
 0.07017044650667903,0.8747433729479146,0.2531918802964227,
 0.49536654904556476,0.1257737645305459,0.5270757594024112,0.9543647877979229,
 0.4568451885855199,0.3535053614202064,0.5899973999607071,0.4887728652027249,
 0.21722595745663142,0.7713808460910131,0.9662370141692914,
 0.15231378844352494,0.7039748481052345,0.7766725582868654,0.5399375977583749,
 0.9513626578111287,0.40055441761644284,0.04897697002223089,
 0.39625459583984024,0.5685649451523771,0.7893745269422724,0.3051586426842523,
 0.8045598932923476,0.8502037464423043,0.02763973598699776,0.8533744434845716,
 0.967045845191084,0.8004458524573513,0.3899479829410938,0.8011323243606386,
 0.004384623578695401,0.4959688121250323,0.16293080546904404,
 0.40099336261312435,0.16476907017438513,0.7346549966178679,
 0.3511701168629936,0.7527399787323749,0.755795984437219,0.31415787646860627,
 0.49301285410622275,0.9050744334541809,0.7901239791220701,0.8028444574514416,
 0.13663608700772478,0.6615590494623271,0.45415222942526967,
 0.4129342515605281,0.6428090310628668,0.3147971265140095,0.32617923128887916,
 0.15282840837846345,0.06748082065028782,0.9342035338554954,
 0.2730432014284827,0.4607349428738816,0.03843059942528704,0.7550342749595621,
 0.35650031758427125,0.49846602702797316,0.4425356230430527,
 0.3383878387877536,0.8662315020647608,0.6404012150773915,0.5973133702072435,
 0.6932269258963918,0.7393405865620595,0.8687249100530937,0.7377744196555398,
 0.5809154604481874,0.7700826924010652,0.9323448574097506,0.6570018974614247,
 0.6720053337463009,0.8816070406123091,0.4405459276105106,
 0.048838858066036384,0.1822145576083659,0.054778493293807795,
 0.2760689043879956,0.5724227742558651,0.9701087541115695,0.8007600649337465,
 0.9896242923312187,0.08411539879598318,0.5791898231863606,0.7618912458701396,
 0.9095466222062223,0.11616979630785562,0.8583315126325648,
 0.16129154497495213,0.5938445804175212,0.14056209528521169,
 0.8624705114707929,0.5079329162641334,0.4491702905134166,0.366094122624507,
 0.8860170489457297,0.6125216334375228,0.7341922651609177,0.11442276366998405,
 0.641242948456273,0.0659160946133317,0.34811029315964326,0.42572789160065416,
 0.10160563191550886,0.7184536356750028,0.22556201160306444,
 0.09671774029058289,0.1279161700040472,0.9272501448285676,0.423959184048627,
 0.8735176272800306,0.8444085645211313,0.7843170089029474,0.3308161102035231,
 0.46160332547520466,0.050487457106048295,0.9777307732046076,
 0.7105517535208893,0.35975386601812465,0.3753405833475725,0.7945458094921691,
 0.4132872727370488,0.10396160819612499,0.7146908293485332,
 0.36837873971855273,0.7769200521687514,0.09361688801078882,
 0.19204381175644747,0.03288785833303842,0.06407924788786445,
 0.9485833066804829,0.161713617337496,0.11567911610259152,0.7167962751161585,
 0.8112554397011333,0.043005164404328555,0.11952403527238797,
 0.18531498874034436,0.8587145483515757,0.2490999891113953,0.5324102315141138,
 0.16071873397739356,0.7075541546647353,0.06952696892712007,
 0.3224897517908675,0.9685564407449807,0.8153052974955336,0.38636517910234125,
 0.36744942669171443,0.2361706397100145,0.8391419506746226,0.2922724203780037,
 0.02989563921639149,0.7805623155959968,0.21326077410276445,
 0.38457700065117295,0.07215858622975513,0.8498973403862018,
 0.0042934124952193065,0.1215247405478268,0.2143141830213744,
 0.5992700384225705,0.4422666422951027,0.7030433813348635,0.48082644189440493,
 0.6206850278938343,0.012312901838276158,0.8225689962347713,
 0.12304477510562395,0.09836620406832397,0.18885684994306828,
 0.3674349028005811,0.1795197504674465,0.43082458709517724,
 0.25323356966773436,0.7844016738893363,0.835295013035429,0.06545478562778473,
 0.7672552884656134,0.08042253806945276,0.5497705455083529,0.8247142860889314,
 0.26618817315972254,0.7794583898257732,0.8792829462472759,0.862966315884971,
 0.5330834613274877,0.48197438729527375,0.8607183573217106,0.3674063237150431,
 0.9644102701808136,0.5623061459751872,0.05907468820402406,0.3409786425568294,
 0.05751833662681838,0.008796454680659704,0.5559995429266371,
 0.08564232337882238,0.7374356358555832,0.45328357439591016,
 0.8754266348834671,0.7391418716119745,0.7658632774808459,0.9274054439504571,
 0.594834957778189,0.7696967420428797,7.09207056836747e-4,0.9884981060860653,
 0.16941705361918125,0.7770925349191199,0.7007179404380126,0.4874470165833512,
 0.7515179960558146,0.6579718459290986,0.8902214472293054,0.47887205111733455,
 0.17219234209057221,0.15088698377890464,0.8037451820079314,
 0.6249948103241977,0.4336611211045679,0.21656845651700818,0.7906236975891051,
 0.08102604046954553,0.7837732258111998,0.5808774838552937,0.9962627822721795,
 0.513245522253493,0.29725341127360805,0.6607026589733158,0.4658468233998584,
 0.4793357253121078,0.42684446752828953,0.6545174425465197,0.8373122873234101,
 0.978524692177015,0.8729892269334916,0.7828965841988607,0.4832308828476193,
 0.4077610319031765,0.055751729323382127,0.15272678852731747,
 0.8043981283244772,0.5710113116382024,0.869449696631667,0.9706435031051952,
 0.11931415076450647,0.049313722667360116,0.20974450224958063,
 0.3668573309388493,0.34756858410174485,0.42324518017195767,
 0.31054089531254603,0.622534574134281,0.34055277567609865,0.2984043862708,
 0.6801298575384725,0.8014397750984839,0.9281550319528593,0.09192099810379883,
 0.993494791346344,0.18023746543596375,0.692914714031248,0.5761304060968371,
 0.022048262366104243,0.5747954740005629,0.5810768211973311,
 0.3738426006499609,0.25839069360313616,0.6664850861756175,0.5747092444610309,
 0.7119536942044489,0.9314128200841578,0.7807752055518974,0.24521305846122887,
 0.5616488818138167,0.2290640314202569,0.3371233211346486,0.26137477578463086,
 0.513013045930331,0.6519392740803067,0.3530498879193694,0.6865161330521529,
 0.9242292799944725,0.7848115553334123,0.27094683234452877,0.7472231162146462,
 0.5282653087567464,0.6047223854916405,0.06902138713333394,
 0.02860634707373766,0.08795355086143031,0.0783602235928963,
 0.1306201571201745,0.13548970619146283,0.39144973064743194,
 0.7926591674541428,0.9485007366034265,0.6821888623820269,0.22023209207695849,
 0.19032364389310308,0.24127227406414664,0.8971363408960845,
 0.3180331937769286,0.006661837638496371,0.04233834665908964,
 0.015457586670214685,0.8268304577210391,0.07650518759259262,
 0.5380594522799398,0.6249280211082405,0.9621724549487212,0.39014043984435465,
 0.3950152214198922,0.8462208447743271,0.3807281727312869,0.47413727592164057,
 0.4407028549402676,0.8803991797854217,0.5506632371806761,0.7714740309591407,
 0.2745923497196394,0.9569264861458401,0.5110622081096792,0.6969735273827046,
 0.9092801487072446,0.12424344923506092,0.040727264071566016,
 0.6835130016466697,0.13676057597021718,0.326374909416692,0.2877053780027867,
 0.8629238534910776,0.3490025836660551,0.8353259209902939,0.4088027048955467,
 0.9913126187346941,0.1850055259951382,0.5491351804978769,0.35053907746652413,
 0.37947120045869087,0.021770587640668637,0.70206977416998,
 0.34112824682663223,0.5862703469497766,0.5315455025226206,
 0.24593227718789257,0.968528424989664,0.007400451025099675,
 0.7189544683134101,0.4313045247220263,0.3964679123103514,0.4153991765559112,
 0.5822486059904327,0.7516902388446182,0.9237130009034302,0.29452319390021287,
 0.18721783910062206,0.9111661229341366,0.4041985517472765,0.6696140857193841,
 0.3108369299142495,0.015368105528548281,0.8547425043336077,
 0.20354569900300812,0.675276621661554,0.9890382885309029,0.5542255956322042,
 0.18690580523590594,0.18621414864756192,0.09724944637493727,
 0.26595531211143797,0.40056866775589306,0.5965894093107332,
 0.28528401667935843,0.9531115681858142,0.0625259826850928,0.9785592390463905,
 0.9575498759131355,0.8390847431340857,0.30752661257267444,0.660694634415544,
 0.23266293759377288,0.866565649944332,0.7307589479293494,0.03360592959013842,
 0.5209710546142559,0.6356808692633683,0.24879163075182098,
 0.17442059775695629,0.6096118681822411,0.17805418620199465,
 0.7857319327553196,0.20282553601609177,0.4763073878361891,
 0.26016755241227885,0.42787864453385405,0.9211887388867684,
 0.14641031323903142,0.6132164198269503,0.30945273732060796,
 0.8359400662538473,0.216800188930255,0.047694177673966,0.9136026696787395,
 0.0330814040723425,0.4749604910651426,0.965268428689182,0.6962104082807873,
 0.40894785125305944,0.06766399719488059,0.8964192278180593,
 0.10281715097902577,0.47785488655401154,0.8653389674635948,
 0.10440323151306852,0.4757299165935751,0.7231792039827128,0.8908538959164182,
 0.39548526612945856,0.22905942752173414,0.6328272720497716,
 0.5689618695284397,0.47108580476973105,0.10810532109818516,
 0.9867533719487742,0.47803386345814225,0.6489813456842545,
 0.24889106444500642,0.8799140370810075,0.612554818790474,0.37827436453292085,
 0.8655469626628524,0.5420904733751352,0.3030834869787822,0.9040104357861739,
 0.6032896116520021,0.2338754401531321,0.6419170662585475,0.6169925030415935,
 0.9011781320236485,0.48336312508977675,0.07402270367880392,
 0.7324951413598871,0.3974251375911563,0.569842664677562,0.03735882751407493,
 0.41192292953307885,0.4218022971905986,0.9949473555160543,0.8974297162135179,
 0.14925429570873772,0.6170071986468394,0.12972998001297553,
 0.6696649957548322,0.6280632211101812,0.1617723193402405,0.13366298512534736,
 0.5020712691820041,0.8449263609725405,0.0945961113830851,0.6985087283878985,
 0.9438609897497003,0.8431842443610202,0.5743973776229431,0.6383296004198324,
 0.6650264458951304,0.9920064744088601,0.32754389550499174,
 0.12079999109883155,0.9190504655371755,0.5497370375964197,0.5142541132859756,
 0.5640093764955278,0.31059065061913416,0.23400385351898012,
 0.30878015342512277,0.6107052496708967,0.1136192256129781,0.3573754513556846,
 0.9846565977440113,0.9936857658521298,0.7721089971858643,0.5246487943044752,
 0.1702061608964005,0.9946703995440008,0.4336494962007096,0.4939875371288852,
 0.07060579390677646,0.5800268322333894,0.09905325010774124,
 0.6910669014172843,0.27018635764458543,0.6130846994849374,
 0.31685816999754157,0.38021155316787003,0.43543047498864673,
 0.799578033393023,0.04363782377142733,0.6873765014286594,0.06834999445345358,
 0.2029892391194208,0.3178333990553691,0.16172635494585552,
 0.16162302416506502,0.11822872296114784,0.8745982398794552,
 0.37681933468964746,0.5302395032275905,0.36719933400499194,
 0.6902364109346563,0.9658413187938722,0.037372604834425505,
 0.2539756208238466,0.9481878603615251,0.3099112770612399,0.45568611468861464,
 0.11424455271450329,0.9498513468904891,0.2857113874082817,0.8068563654278644,
 0.6092086378842843,0.9600045092249939,0.7449747661518745,0.3989188479332799,
 0.11200422308550995,0.6388623524837795,0.9684885496375006,0.4035551335972113,
 0.06723387229985123,0.08979443654590291,0.9441261587613335,
 0.5706800609903236,0.6477249303554286,0.32411001653842764,0.5081784731782882,
 0.27007451194673215,0.5331087412133073,0.2836066316979029,0.7233346813519299,
 0.38970196136842694,0.03618520992216823,0.6011873091710778,
 0.9057328560224461,0.017205371410840442,0.9171194133186583,
 0.13289776254074326,0.5012359228817801,0.6762059444381892,0.6044603868755374,
 0.28526126793805406,0.6630513994281838,0.7087687705672795,
 0.13205057029015999,0.786779883705358,0.551456941442624,0.2925806692010853,
 0.2750100083943001,0.30254477976296756,0.9266514236590261,0.7587970247502619,
 0.3902459309523263,0.9697707760648392,0.988264000234536,0.6922101649580477,
 0.70945067192459,0.4664532429085859,0.5780222190891761,0.0892879922565335,
 0.08685120753235676,0.7905566288911782,0.31941117632682314,
 0.2856567717068841,0.9336259481629497,0.7464578392686467,0.1602399687957049,
 0.1207122762564331,0.013423559090651405,0.6169582595015057,
 0.3583551799913285,0.6900218270446167,0.7403619211504219,0.7320440298169854,
 0.9322055544345336,0.29789827186638707,0.4789298658897928,
 0.17554083299202272,0.7183205606844826,0.7676235351608849,
 0.05459222476910286,0.8888007149130652,0.6036381114855449,
 0.09808578645936128,0.11386832037013506,0.4804777153496279,
 0.5960649391286748,0.20259647809956793,0.2877245598967011,
 0.12059322486078128,0.5813675607337653,0.5451943032706557,0.2130369787932569,
 0.8370442425679185,0.40372804917902094,0.3816497403525073,0.7539028420673597,
 0.373792370722387,0.46186654574053876,0.42124799181315686,
 0.05555399815779494,0.6618202657822548,0.06583148900710212,
 0.07007529546945945,0.503226031856117,0.9684450999447175,0.8387203929333635,
 0.16483890184405436,0.5583815750089193,0.38268628974249264,
 0.8614223312197736,0.9021670867542679,0.8473581689914687,0.5069616072378664,
 0.9332730252161622,0.9083227851381164,0.8216681716573079,0.48059208139443566,
 0.06316235042623874,0.4954994300773248,0.02854453337490015,
 0.9348464985845123,0.14659656895500428,0.4870004539076952,0.9169747447738088,
 0.7642167721829947,0.9494992365206094,0.4521015583175726,0.43739370113151,
 0.5234045015123305,0.4891386008534162,0.9641492753531353,0.35998573327373184,
 0.5223692675973808,0.4469471895018853,0.14226164095278604,
 0.22759992878214452,0.01180863585975278,0.0408272120333375,
 0.9931343783131048,0.29073858615428716,0.8087232453734192,
 0.10449280722912213,0.9401760281081187,0.5108171455528372,
 0.29789702762395254,0.8955368782845458,0.26321193472027393,
 0.33350507333927404,0.343176508304621,0.8963737229753522,0.4758167370593285,
 0.04749888041454231,0.9639621449162012,0.8901532195785602,
 0.11458988407536819,0.8427774628418825,0.4501337759793067,
 0.17317356685333873,0.4424255001965144,0.963784479293398,0.2800224735788699,
 0.5847617623868452,0.6838878777399212,0.06583515398910522,0.3498735723111214,
 0.12599093953608853,0.8051188325032629,0.22144747438232604,0.964788888193781,
 0.30954990788678316,0.8098221467193955,0.6636834282413835,0.4398682038865197,
 0.5921888378144452,0.7298818643991665,0.6224312137846374,0.04367226501483101,
 0.6017397359986538,0.5775301427191138,0.3640974029135101,0.5240128703465156,
 0.4822529997538513,0.19243112071350765,0.6865764780021092,
 0.19085227927398551,0.5866814502837937,0.8307443638804907,
 0.21381775167033867,0.1797608798023359,0.8612111966040527,0.6559783978028647,
 0.12257368414233161,0.6087313803653522,0.63669045493641,0.07822882539124032,
 0.1073890785341749,0.45847766318401306,0.0382209086596802,
 0.23939148885087536,0.7617232614934911,0.24578554552243848,
 0.9245255252375834,0.2583676657808055,0.05313880981065178,0.7244600354799849,
 0.3155563898740851,0.8958030198878215,0.18167260642984095,
 0.33588386798221814,0.5091579837624852,0.15848309035169728,
 0.47864507632647135,0.9450735811939427,0.9362082087014195,
 0.27970306005697376,0.7796710000596075,0.6676762919921988,0.7473852429210309,
 0.004692142595574711,0.2764253909335379,0.5469681612704467,
 0.7276292271262017,0.6813440794460892,0.156849377375204,0.6919001490563162,
 0.05134787495010151,0.15777482384831565,0.5083323419819963,
 0.12682393533456915,0.9127393495371504,0.5990324117565071,0.4689772207019559,
 0.23949570112839347,0.7047506907879911,0.7735945409465549,0.3624724647144085,
 0.8190722698245008,0.8736777793961721,0.02729617124525119,
 0.21952590051819498,0.7958979387495223,0.20294946387163104,
 0.9715306021465112,0.6804927229531261,0.4950614574572645,0.3902719297057117,
 0.29706743797300583,0.5849606910992391,0.034403959270888684,
 0.07598792529350773,0.7451431503738678,0.7835988919603853,
 0.033869427615528025,0.9731274290435113,0.20953649592431267,
 0.967032532884121,0.8071377816410878,0.49005723684930635,0.7958358182219089,
 0.39635411809371845,0.0927420442247826,0.6580886226512143,0.7363159568185867,
 0.3399705906410567,0.027490550894098353,0.4458979330564641,
 0.7049546164655681,0.23083818760535246,0.24702604357901403,
 0.01081971351768607,0.6637183416285389,0.5254473840435837,0.6408626294302866,
 0.6314707549533505,0.346658286168136,0.24446147725047385,0.6547567900797715,
 0.44405911789957087,0.2963274441953827,0.07837964761643279,
 0.01191584940041357,0.9019888895715792,0.6052295311345651,0.2850602728039776,
 0.13704002778898505,0.34608701212121606,0.14272666870355966,
 0.5257124005046662,0.22232419301133777,0.5693134688562915,0.4967654516150024,
 0.7715548024519588,0.6299397364200668,0.1823143229507368,0.2880345853348244,
 0.26883325544851855,0.7931337460027473,0.33494001418722785,
 0.5352543296636506,0.001692248623008763,0.6297552766871405,
 0.9670753164732856,0.5840485389537974,0.4332789143296588,0.4486918270660848,
 0.41491403391420634,0.7629649199338694,0.5530879731415145,0.7802551656466384,
 0.13508185452364452,0.6584553635222814,0.8961997091877,0.0012097527108247963,
 0.4581300468094561,0.21810057491604895,0.2999207746417213,0.8736899100920588,
 0.30832660094580944,0.19953895671680633,0.727314651424097,0.6637571273731669,
 0.33416690052036957,0.9846188948292242,0.08175688786647373,
 0.27131708201391225,0.9113412425686318,0.06617089864099956,
 0.47147523084686893,0.153683124153005,0.7547043467358883,0.08297801912155434,
 0.14849583225655272,0.5728500754264281,0.1664278090033039,
 0.09371673002124692,0.08968437509333016,0.7265293166821181,
 0.20734313841290475,0.9426128696400914,0.2801611967387714,0.7268372826393422,
 0.5748646650166581,0.6347031905592668,0.9463020605987913,0.630374605528695,
 0.9765377455218104,0.8736329112929562,0.7637334257579504,0.26697929839483026,
 0.8056837058054024,0.8790187967179357,0.4241644554431674,0.42572660040134247,
 0.9625274178490099,0.03669617498378097,0.61040514222532,0.6092079848034966,
 0.959801371010001,0.5354771257240852,0.8061012787873459,0.9510554781709154,
 0.9074165824964625,0.08287395043494084,0.07665228915190003,
 0.14008880351019615,0.9801394344307036,0.9953436096446431,0.9229987419806241,
 0.6755534079310854,0.1239703333324883,0.062184586459059776,
 0.36960371246305335,0.9943256317195037,0.09188508062623246,
 0.9133929339995603,0.6639442756867734,0.33784477177546557,0.9819777408557219,
 0.3214240138473383,0.33909478925233816,0.367259329558028,0.9803437714396929,
 0.1586353335819033,0.7719673561560306,0.0352318190980464,0.8622128230668142,
 0.76252695237301,0.8770324593360025,0.06369315027214506,0.7707626617502301,
 0.015244230290704652,0.9591524455417771,0.7812067067888336,
 0.39411118903163056,0.8931764262567249,0.9968243376667236,
 0.38456134067344205,0.20514651854865884,0.9852538200624319,
 0.7578784494759694,0.4026144949447664,0.9679480805839995,0.2555459745679213,
 0.6968553693661941,0.6784152624706022,0.46232867279150835,
 0.25632747149733637,0.9473913741768594,0.8152436738555773,
 0.34433767732385934,0.582036124446004,0.15311680737039868,0.7990518016337882,
 0.007745815159701852,0.14212784470048534,0.41260760594237844,
 0.876828782139731,0.3878506443040234,0.7801092101280103,0.019998848736180852,
 0.8806725747051755,0.11498401263040248,0.3214005881359737,0.6828303655181958,
 0.09506606022766528,0.037285374180811126,0.16664745365931555,
 0.8683865995298118,0.555334237246406,0.8824399606090375,0.9132352969702159,
 0.9902922735619191,0.9402055938861672,0.5968661825705849,0.453865622849528,
 0.11159572688913533,0.4044351962062569,0.45883699760741103,
 0.051472895449037104,0.3186516430564311,0.44973250857011493,
 0.7762378163963716,0.6605346135050325,0.8381982608775593,0.22627129945581292,
 0.7997079838932508,0.7574817355482713,0.18463049165474033,0.6219856751805884,
 0.9833179968499759,0.906915009742765,0.3780642164508148,0.07104715725348254,
 0.30515111152460483,0.8410805802244933,0.9297582990513918,0.5224492445516369,
 0.06563430722330321,0.7371174298529171,0.08632831749399184,
 0.28788897369405886,0.5369936346502511,0.32510186055494394,
 0.2887314699408513,0.40505207819639666,0.43626160195441654,
 0.2729248541013588,0.6518594298593892,0.8176296955645055,
 0.010983748038030705,0.655384229174633,0.1814225446727027,0.3624845439795258,
 0.15336409578400045,0.9015611853001892,0.5139969036892251,0.9267798367854865,
 0.9936678234968768,0.5107420076283167,0.4719794672394513,0.7587345701611419,
 0.037814293069437066,0.5627024674345273,0.1312009375051304,
 0.6214251983408166,0.971799327325712,0.0953983701276766,0.38685336307080964,
 0.34999632813862,0.6935747803544707,0.8597765802022674,0.634972601470382,
 0.9139190389396121,0.032996413622103704,0.48964992287695297,
 0.7478633739878899,0.8417150398834947,0.06043134097728531,
 0.02501132117433813,0.13421909046939873,0.6127664176285039,0.267279930177192,
 0.0922875683535701,0.4098540389954417,0.629193852018223,0.6020502453796353,
 0.6826347588625085,0.41892654451163047,0.24500038597793883,
 0.4260227561085739,0.48525849869632987,0.7143870256169051,0.6848526086084753,
 0.22835764727194163,0.1247019263999295,0.20104926303798465,
 0.5511562104525306,0.3005215328496853,0.06714229533752969,
 0.44013032202958435,0.8277180242403835,0.4274823405633509,0.7401509929846238,
 0.7772496739809225,0.8927084803754217,0.9824161961760282,0.6548364876365832,
 0.07729637734056793,0.34747681437568767,0.44149710872561465,
 0.018263651533939718,0.2513347498959926,0.6605120626234009,
 0.7669955518981406,0.37490735934152064,0.10418732251589902,
 0.3397444653151558,0.5897385099854249,0.6627096283058082,0.27811464825906684,
 0.43327240843713555,0.32237910916114176,0.8821886279750154,
 0.9082631075708427,0.8668795300837244,0.7446569642643401,0.7850950842050449,
 0.1940262534390409,0.5498397704241815,0.3517688421755777,0.9034767183629209,
 0.20319495679984634,0.5921704116973159,0.37963892140217137,
 0.025455484143220075,0.6590836582816652,0.046474249286160396,
 0.26472523770967304,0.13306791006962637,0.8284687306203099,
 0.05406182493014933,0.39466526676616853,0.8886477318183816,
 0.8849184468142455,0.12748096569612033,0.4841497472815083,0.9854451992673803,
 0.6406652782323636,0.5943783265156869,0.6673059103975698,0.43837624333840464,
 0.1918921593737719,0.5268830503877301,0.6963838144279462,0.23442800294219468,
 0.3729834246786964,0.47191957699732456,0.5090180446180305,
 0.23605067790091516,0.9732074625264402,0.38819334492303037,
 0.07447436435706245,0.46332023630790686,0.6757682085929946,
 0.8488326835436972,0.5767811108424449,0.30983118715172864,0.8121234674821568,
 0.09071456565337854,0.30876497710359574,0.2642484365118323,
 0.6629123220356135,0.3826777928257362,0.7151749339464307,0.33246758647869956,
 0.772315678819175,0.2293081720978748,0.2975309818119144,0.2613766440236093,
 0.3242998860179266,0.85473436174698,0.29100808290907665,0.306780936602564,
 0.29676239392829085,0.9651584197235693,0.33855208988733076,
 0.8519867632572392,0.12248150604522257,0.2926797727958468,0.635854238373695,
 0.8234953106793711,0.8259390455768982,0.23630342609280441,
 0.48625738907269056,0.8418570003016972,0.6613609128119191,
 0.10767166526315508,0.5203923572202879,0.4498667490006121,0.6870719519802383,
 0.6154437921312713,0.6223023138544417,0.1395193787786868,0.635518513343862,
 0.1393618942590601,0.6815203786239993,0.3736620628984033,0.6342377290382253,
 0.7732974457035312,0.15514430192275674,0.5229936803627039,0.9727583873566417,
 0.3779185651670941,0.3735744848505359,0.25847037055429645,0.6034636861101418,
 0.5004765901899038,0.9543614416661791,0.6189747666897242,0.7619608798638742,
 0.4507653670434415,0.8226192730355006,0.6356336732882572,0.30464976189320236,
 0.6704395100022181,0.3348136959272131,0.8034374714925717,0.12553082638633095,
 0.548611113981702,0.1627039403725692,0.9245708749822281,0.30989535005670554,
 0.019658878664191914,0.7365973337032881,0.5994844833386738,0.993773105807326,
 0.12841748088834892,0.12428977587838919,0.9097872168679451,
 0.1292444048819016,0.1934922895859278,0.1310445546620379,0.3470818327680357,
 0.05399836039829409,0.5242578871038075,0.8292310769943119,
 0.05406288685991534,0.25644404365873463,0.33220508436050633,
 0.15599725030934064,0.9362930098502275,0.04618337340027123,
 0.6792066313250955,0.37388075173132473,0.22790173436189565,
 0.7013508409413878,0.19098107000765596,0.8955317815592861,
 0.06710566199088863,0.44365594362885474,0.5111970812714874,
 0.4540553270457186,0.03838010014306026,0.611240578947582,0.9968470209269499,
 0.11539029818084523,0.03773425550613796,0.6549888226249946,
 0.04541485196750328,0.016619010314102844,0.05356609602305329,
 0.597662523839761,0.5197334265103333,0.9095187387392799,0.507631925370393,
 0.12585434342723234,0.1409822538301868,0.5087889878341173,
 0.27574160062484665,0.5708526314345926,0.38209941009598336,
 0.9224402537556091,0.08679069525353711,0.503872895847276,0.0669107557048435,
 0.7588105984289533,0.6465804233551888,0.6005391498519725,0.18653447345143548,
 0.6219878933690534,0.42028632431474344,0.3298261122463184,0.4726211141296013,
 0.11325302156316996,0.2393061849410273,0.9766798985502076,0.3797316692838445,
 0.3357004721760355,0.021402764797118667,0.6254236171225502,
 0.9074041428727558,0.1683373212017314,0.37200395303120826,0.138381275949508,
 0.16979624061352872,0.5494887508474467,0.3244662394873029,0.7920882201315739,
 0.804121795374376,0.3929425323585344,0.6181214416453339,0.7231726573289389,
 0.9832661853675515,0.3316679979812024,0.9532350857721403,0.8787032413851403,
 0.6040789348799461,0.017264589050210555,0.3417687983298594,
 0.7399610206908993,0.4401724746565582,0.19602099187440214,0.5991018175427936,
 0.6433260385468381,0.4160068346154051,0.9805802744772129,0.18033650301794268,
 0.3299803522389645,0.1830207037519811,0.6774166192144818,0.19643383068510578,
 0.6010522244153702,0.5149698871631245,0.18360909668436665,
 0.23918912371424206,0.9752566378099435,0.7839763504021997,0.9787633028824887,
 0.04200294445319519,0.9679651270847462,0.7111985340579301,0.8207715434172719,
 0.7116332463774963,0.6924663175384204,0.23541150067599137,
 0.10741411894197195,0.6338604848761719,0.5728338267190272,0.7110780600903253,
 0.20891076430042232,0.662235301609213,0.7980119540456074,0.6939740763102479,
 0.7227027967482718,0.6392457161829322,0.2989789618561921,0.18070245723332534,
 0.31741114941232074,0.5461619007305295,0.23710141391735562,
 0.9557058102579222,0.0061254596210951195,0.9182731442664017,
 0.6031044928611076,0.08289927160498922,0.13945918276104863,0.471411387942826,
 0.8893175846958765,0.9272303507315597,0.40989595375059895,0.271790959480123,
 0.2642079596430096,0.5726121992360251,0.7913968982888956,0.9831333512570923,
 0.6351250261375977,0.5544409281774637,0.3254333174823134,0.7359914855992699,
 0.11610789800964383,0.15572083582463736,0.6397548021758758,
 0.8881578304561536,0.8835135341881453,0.08800823376070133,0.6556215105754184,
 0.17001064536499144,0.8695262356630254,0.10964672737036918,
 0.8914552792185068,0.30934430330429896,0.7207569794115394,0.9488398706289828,
 0.1522071907355147,0.8562759916408648,0.00301851926790353,0.6961739540100724,
 0.17369243654316824,0.4227551841616233,0.8918114229029607,
 0.14761119160154967,0.7168111905696614,0.9072102665627834,0.7726030578732035,
 0.03647462003061874,0.5251417096067965,0.219971438146787,
 0.048027892368211256,0.14520850734784718,0.9821873255035393,
 0.10706635742682136,0.8516462657609214,0.10950347196959453,
 0.1356159931809756,0.9229494704127852,0.106366699123568,0.18338999039631831,
 0.8203674994117177,0.6652340017025102,0.4749190408507452,0.7270185911653723,
 0.27497544267718943,0.6498538356523601,0.8551607479371143,0.8217066958739805,
 0.37450374293087774,0.13893566599930118,0.05138134746368839,
 0.5186950926410075,0.3952241195206697,0.7995001412663979,0.6697108108580055,
 0.331745847320119,0.31484414069192124,0.8114155327630419,0.19425088232831966,
 0.9852482551923164,0.574321786024702,0.4725426051512749,0.32109400352080497,
 0.4487880247233671,0.06526608924276356,0.5515141044564448,0.7093706473558429,
 0.7055227234250697,0.20688840911171869,0.7272460258280785,0.8065091190213681,
 0.38441092407171196,0.8100515664491065,0.1268901822262256,0.3767706653997396,
 0.3582560268723566,0.17268175488856863,0.4255520021337955,0.3498870797749881,
 0.22298021753111086,0.448252452484022,0.22962940787816777,
 0.47657223729696585,0.2611769672560015,0.3296786838118372,
 0.10814100443913621,0.8254120055128725,0.653873664307086,0.4267770212669669,
 0.28885894444784577,0.5457901565102152,0.032663429271551636,
 0.6531648499988789,0.8193021782142196,0.5485690396874836,0.13330053193444003,
 0.5014843179471408,0.286581786039772,0.8759545226503281,0.13711800254333473,
 0.47634870910407834,0.8856851932950718,0.19272148006662182,
 0.42401662817400243,0.31585785493857577,0.6360071113424661,
 0.8590654035485233,0.23917822018323098,0.9032028523671903,0.2743397507745471,
 0.36551454540572326,0.11229825933781501,0.5580854687166851,
 0.23834041268719064,0.6225073573786472,0.07165709244255403,
 0.09065267917161723,0.19425634027458638,0.9421736562151762,
 0.8133997150865866,0.8524310572245042,0.4874047826065274,0.5055340411451619,
 0.7456111348454817,0.7922624347581185,0.6295436955062165,0.4974593533767928,
 0.48979852400839496,0.8129896063543353,0.4130516577171337,0.4645487249027641,
 0.8701476022085743,0.152465692902073,0.8423051154181179,0.3706116262546356,
 0.7732707454762874,0.36899287657787716,0.6058691592347286,0.8987655775327795,
 0.6141120684267937,0.8691591298246757,0.5949581237158769,0.24198484332396653,
 0.7699374569241002,0.5115548215267997,0.10419174923496155,
 0.29387918298660176,0.37028986209634907,0.37489241304770815,
 0.5863786320439784,0.5808169160169457,0.6974206720735165,0.5851662236170254,
 0.8782543816909472,0.3948686111232358,0.660587122802763,0.3322151095120841,
 0.31222660882208486,0.38674601877974635,0.7997576824035231,0.829270731019135,
 0.8345541694327289,0.41708443252324834,0.7102178557660576,0.5696002827336851,
 0.3698710249381827,0.31574242064323044,0.8453100415256478,0.6220770552995619,
 0.34145959157211414,0.9963165617798279,0.03005117050285888,
 0.3389445964605575,0.03942815434036129,0.36203640954520777,0.804859847253438,
 0.9248013445603965,0.24801407587412116,0.2798125669060296,
 0.18497192418783237,0.934723974115258,0.8660712007154477,0.4440392507439265,
 0.6088075775541976,0.18789940909864755,0.03239307985123552,
 0.7104359723841077,0.3711918064292057,0.6341653106730321,0.3423488289314901,
 0.37138540488541705,0.2438483430285494,0.9796205054951691,0.1644523038597051,
 0.5034970627024856,0.7158477566818119,0.08412320553495178,
 0.039066897939729284,0.4153320012017927,0.4204203029543675,
 0.3938373201138393,0.7200561447400072,0.5804423440144467,0.46963555758217856,
 0.31467527225552494,0.2839849303913663,0.6892007347856951,0.975097254101086,
 0.6410122696099577,0.15220595171924312,0.9614054026785628,0.9967469956371471,
 0.708704037401054,0.2106252911054476,0.22725140381764186,
 0.023026736549786597,0.5676296481896524,0.7827406709379978,
 0.6037728936385711,0.36823869257617237,0.5032942355368635,0.6693447803078334,
 0.6370250404897521,0.8413048445335625,0.14748761916011,0.6679596878944358,
 0.9698444295239701,0.5635011289210452,0.5221137261854241,0.6565828958014304,
 0.9239935500597158,0.8840001124787418,0.23284161063422582,0.9386298763824067,
 0.6193326443386505,0.5941987461048318,0.13533091714828605,0.9024573998327031,
 0.6887802606550468,0.4554909403490297,0.25521025020598687,
 0.32792594891441473,0.661002543031976,0.21812785938659718,
 0.22880743926414915,0.9608866178165977,0.6173932164029118,
 0.22531548033206827,0.15508387449267613,0.15317353313370652,
 0.2649699566259681,0.7194432167499436,0.41928668364561783,0.5187941174928486,
 0.8206724643671549,0.7899103417572038,0.4833599194620182,0.5121849918358922,
 0.5579235720095748,0.8553709622140234,0.03373224590549517,
 0.19745315059338509,0.9551787352453817,0.15200337485994042,
 0.06301996805123544,0.8750216058561887,0.11196909334824223,
 0.7223113665064413,0.007898889291363287,0.3481632343899955,
 0.5495305664730011,0.04312705621232271,0.06580173361110608,
 0.7585989911617745,0.32121659757169274,0.9559333736714087,0.6054615066160264,
 0.7502228259925875,0.8146703377232158,0.14488263320493378,0.7774646836478224,
 0.1266511124093156,0.6828091014261919,0.5507105742208965,0.894071715688573,
 0.5878438571291872,0.035518378149984486,0.9984061673310664,
 0.22754256544214402,0.7659326967192974,0.14152689075899905,
 0.44103988398529803,0.44807704122594094,0.44606030558112986,
 0.8490826769384833,0.291818222673093,0.15737378936258017,
 0.0035427911711094495,0.002467000667036423,0.2489960173009026,
 0.6166926547379367,0.4600684263875401,0.8426240800212234,0.11165072506843954,
 0.7113400984084524,0.222840947687455,0.5121282531725596,0.9513473055402364,
 0.6805743951394081,0.7013320735388526,0.5546796542878546,
 0.058682620588215495,0.3616961964737916,0.40206185843584075,
 0.045651731004160645,0.7999342344885055,0.2980913520178028,
 0.40748007165770883,0.4676389713473946,0.021313381734986647,
 0.806884416725339,0.7268562283139787,0.056478807400076714,0.3089616078947166,
 0.22444825575614036,0.6424420006423851,0.5130855628501617,0.6853845196227528,
 0.3053361848342192,0.4041993455864974,0.6648781961148695,0.03256170260223079,
 0.1610626936191275,0.6968920489843295,0.5479526673358613,0.8665402894883449,
 0.45009882224266096,0.15223448696422381,0.13629768419682997,
 0.5007750334062903,0.8019176017702342,0.06190217353277905,0.6098283369783393,
 0.9923270455647347,0.19875929891231736,0.683533990518722,0.5062204450771028,
 0.465198291191263,0.09343677006592932,0.657245923544993,0.2995977790644987,
 0.27563715553833923,0.4710522517245681,0.4526653685233636,0.9429934360871861,
 0.3787537084668119,0.786560706275328,0.17913379389861117,0.44570347164733115,
 0.5612180762469396,0.8932215196695493,0.8664255437653833,0.4352999246650696,
 0.8109565922175682,0.6461114881218493,0.33021436616970157,0.9519644267970867,
 0.3639897884985831,0.8748121465587029,0.8673859371110478,0.16653575669972165,
 0.5263502299137174,0.44202617609683137,0.2360137038305372,0.9014846467284988,
 0.5673879910038926,0.9124484141095618,0.7212452774287784,0.1152402131796958,
 0.20923476189844314,0.6567977568000968,0.05091604020168394,
 0.19813537114727053,0.2588451137165979,0.20054222501216468,
 0.5362510763316791,0.2501997842819842,0.6540387869312181,0.3132343484040758,
 0.3507843378077784,0.7217557114670193,0.3773263000397127,0.7877595424981964,
 0.8745179179329047,0.4169307285477666,0.39466123083860816,0.5950878321617112,
 0.8970497707376831,0.041889496194806086,0.06081867676337205,
 0.4595915034558956,0.7520421116452363,0.8599320899182284,0.9345815590717157,
 0.5960948426004533,0.37746208655319813,0.0664058382469076,
 0.11093061379651381,0.3264041514960956,0.008411191384499639,
 0.22721657132444406,0.5554007191306841,0.16266890155153546,
 0.5885527553973892,0.05238746662064009,0.7084932539798959,0.8019159138371991,
 0.7700152799005695,0.6805479899174887,0.2873402351768517,0.12327213755818978,
 0.6218902930185142,0.1905308310053686,0.9454729824990558,0.5050920696093036,
 0.1727432205203473,0.11128666500799922,0.5716956392292489,0.3119149058661561,
 0.4607121882066275,0.5450732863981,0.4694981673936891,0.01624902500462999,
 0.3641241430164688,0.10901355652904021,0.8313234984211026,
 0.42195446885017585,0.6307454773499326,0.020063325342154492,
 0.7686459618117506,0.6990588701019766,0.6419952650607359,0.8154513348200486,
 0.9494097220528588,0.6900252687556466,0.6878979219854856,0.9367574737412703,
 0.6044040930627848,0.38297976236055153,0.03951361635394324,
 0.06133390690496732,0.41734765870187507,0.6154159900762721,
 0.3700197426219174,0.481433248014816,0.32278972357270774,0.9521185072047708,
 0.6126803360460323,0.9881521143655604,0.5397115735322182,0.7046016453403311,
 0.4445066701435435,0.8736810194176654,0.07770797262072882,
 0.16862343389073797,0.31514003040903427,0.22548143545720967,
 0.517571204097135,0.6078863114958789,0.367867057819137,0.8056138167205296,
 0.9782692750641656,0.026529837102144738,0.1789794342338351,
 0.3725272098279687,0.8034476836306371,0.3822342962205366,0.7219339261677475,
 0.008405917368201732,0.006545898679779949,0.7081364885537907,
 0.10007487917860658,0.8791522442314595,0.9842454892807617,0.5763273816550905,
 0.584737249166031,0.43162844820638613,0.8111146278027078,0.6140903660120774,
 0.22638184647819093,0.34919015577201873,0.6266419034814145,
 0.7640598438303287,0.310064328585435,0.6234676707406706,0.5559417001861753,
 0.11904584371013682,0.991541082620281,0.35570203087046914,0.2565240512067599,
 0.7142053384978729,0.009904015562930635,0.7178263707206446,
 0.7812337643829836,0.9412049533571184,0.32574179986241125,0.7194933708698951,
 0.27340799167811547,0.10952167411758496,0.6574592597736545,
 0.10809864587871487,0.46345545823408807,0.7585748063633779,
 0.9942627683314424,0.8023255733730595,0.14998487147261697,
 0.39602338463502873,0.1558410368941452,0.6096056803242706,0.8165318387202034,
 0.5434688774236285,0.580858049499482,0.5091288495894314,0.2986032468001868,
 0.9079014840013506,0.5342387055427322,0.08373784034876586,0.4215291399479687,
 0.5396027496474165,0.17559876448976985,0.13412981740740493,
 0.4926798887791426,0.5677529274286873,0.5920273671184589,0.31864375805553513,
 0.3802780119612207,0.24684735231183375,0.22473009375396014,
 0.5112471301217307,0.07449212368700131,0.9784044636521368,
 0.18241327093544835,0.8601039042431813,0.7016377104661554,
 0.23897780614971142,0.9959234279463485,0.10334455568233525,
 0.7994564435740646,0.39000455325503425,0.47996484311388943,
 0.1958001804222833,0.5741040662865303,0.0970719814214911,0.8826180409663726,
 0.8757917526582936,0.8807563474620169,0.8134088556094499,0.3997705828860335,
 0.3769474974234025,0.05652180307932475,0.9939626962424503,0.8669307656269165,
 0.9848045978733181,0.6052601368087229,0.387752430500538,0.3985674994682422,
 0.7782438929539455,0.056806436511932845,0.296672515509248,
 0.42564229322907665,0.7065786279332584,0.6324617664510431,0.8120123432613144,
 0.703719544869337,0.42738456306305783,0.032116874756481395,
 0.9093472187610068,0.44607416877493933,0.13456038135776394,
 0.19899857241313545,0.35867161895145805,0.23086255668546818,
 0.9534472293156981,0.2880076408229739,0.23867128210978217,
 0.22744782583854417,0.28243957966702904,0.8437917165578026,
 0.5567434122204126,0.8572632864109568,0.21054803937160826,
 0.16982056381834776,0.3214288057925926,0.8895972445835189,0.7592406004845231,
 0.6204528381886723,0.31177125224372404,0.6509861341536033,0.8313383713865433,
 0.7480437994819797,0.0054019203244956016,0.45570159066827176,
 0.5504673601223706,0.42202573255367615,0.6284819471005249,0.8979251930038854,
 0.3316479567220648,0.824153042234939,0.4723758838840676,0.1490301127649749,
 0.02295542980105869,0.3528185473509857,0.9171641937700403,0.5310922142756088,
 0.1384389807016395,0.6560672761658882,0.36741968031435346,0.4675380151400653,
 0.6734750252520709,0.32414477258843544,0.196078086001914,0.18505800756234336,
 0.12061283780606047,0.6692665706917438,0.29025403359337987,
 0.5403598275032532,0.7495741313522244,0.2706068808104123,0.47171234177791643,
 0.5733152178329542,0.8250292791376143,0.9718598919919085,
 0.009728828979455706,0.6823345960836379,0.5461560186324612,
 0.14277318208474132,0.46490563531923046,0.804447826183191,0.1658716649612636,
 0.08453319256715297,0.3972969718397664,0.4500765013752529,0.540654927653252,
 0.9983465949957869,0.8913070118502686,0.879918062667434,0.8900978336705274,
 0.5557531728881737,0.8718190159369839,0.4776988905177222,0.11035208620528203,
 0.2819162846198411,0.9434476763664091,0.3631308945687072,0.8653644786140423,
 0.6976969836613854,0.509185378772159,0.5810009968798426,0.8747544110116481,
 0.9386300445018301,0.23032089597340422,0.9743040629531894,0.1158458929679046,
 0.27191529558170924,0.9374138976247464,0.2251253706227031,
 0.47017344038641784,0.31124721499003505,0.24902952321871097,
 0.4626233129921873,0.8361948181784208,0.11658537725029483,
 0.01692629686356384,0.8031500793724731,0.2552429097433161,0.9811309657241787,
 0.4779527494646043,0.8865129201873393,0.1631362859999992,0.643097503064639,
 0.4872155725477565,0.10793062764122474,0.7124397248071601,
 0.14698231395417094,0.7128971058520943,0.45793219441161304,
 0.7835516440772985,0.5348666606489568,0.5622070162248144,0.06450743503562018,
 0.07083354741215497,0.6357995843916536,0.6406223745447405,0.9047354542580495,
 0.9476094339620403,0.38868799454011005,0.5700655175321678,0.9240351297779115,
 0.41363076675478216,0.8422576853246597,0.8147354777792597,0.3496400689846836,
 0.057816575980170204,0.9969471229351933,0.6292349354411878,
 0.03767968996970461,0.5518596319427996,0.20840734948807682,
 0.3979735451081121,0.8690853156990417,0.6869026016921489,0.895565442522479,
 0.557731890008856,0.21204632794550216,0.41085184847200784,0.6551106345190821,
 0.9874433557143558,0.14751356498929835,0.8931166297458379,
 0.47877864030441764,0.7013306320432295,0.39152972467102354,
 0.22406646613233172,0.1915019370840092,0.8140773239943022,0.9547718174386223,
 0.8484583314649481,0.7393934005579318,0.7263797029824582,0.9353539146848092,
 0.615866272136272,0.786625592361049,0.4727608985406143,0.008049970735604872,
 0.16479903400965057,0.21924756914147303,0.19089919819586343,
 0.9622045616841421,0.7796121812006905,0.6774521294304243,0.6190519359703482,
 0.9692833981883303,0.024215559845008117,0.5330144104071389,
 0.4662436978122294,0.4014159098113028,0.10487949977129019,
 0.43754371502574396,0.827170150550669,0.8046481011233482,0.6830529310093159,
 0.3930118494058803,0.7445472650480236,0.2231310672630782,0.8226723493985716,
 0.424231751906458,0.4148955661164848,0.38359166998257765,0.7804255063603199,
 0.3612267822779789,0.226237993911531,0.46274058057630374,0.6052833932939026,
 0.13321765571887978,0.8684845848138248,0.98622552572384,0.3580937640216041,
 0.79055806327018,0.031833123241100436,0.25394434269898514,0.3161204471819078,
 0.35900344423352404,0.32239321100696783,0.2055609879717527,
 0.16723470663854767,0.006530342993958049,0.4236422182919237,
 0.9040059022489744,0.6329600054103727,0.9909826971275757,0.09650941187884521,
 0.636399106359342,0.38456054175889853,0.48678090061964296,
 0.016028440670814326,0.6195868538392387,0.8266876626498895,
 0.7302817259465589,0.11071314881048755,0.6748092189609742,0.5422372945010382,
 0.2854262755296122,0.9756925589432185,0.9828787393259582,0.5538044989562617,
 0.26779089275752954,0.6094476426594035,0.42486748981328604,
 0.4287111561349487,0.6035071273354081,0.8745171008213247,0.2288617286681187,
 0.3822569558248158,0.7359736874537859,0.708903095200145,0.5293600027683181,
 0.13677187801316248,0.9634087632234732,0.5826364647747916,0.2424211299927289,
 0.465119100697308,0.28043829981175716,0.6061984734822383,0.1424984598841038,
 0.34982892399287624,0.9512703290362001,0.34720739197461314,
 0.8058123617068668,0.33562127323107205,0.7435188293887227,0.7833825196719598,
 0.7522145074095045,0.061570355464928395,0.26241007093350144,
 0.27160861445273166,0.2771272889530574,0.9891189551846662,0.671454569832469,
 0.4282385716268766,0.1310444608734075,0.5625359656047958,0.6748388154518787,
 0.4717219874690808,0.6908680508952089,0.10583286478992615,
 0.09831874790559214,0.12442477192749946,0.27242325482982,0.25387032482217375,
 0.4588370480204731,0.059578761921663226,0.8475033908362695,
 0.4641625735731869,0.9121289654033041,0.6888750889669288,0.9450292244451586,
 0.7017351482266115,0.2820231489889016,0.8773935525884878,0.5933244110745428,
 0.996384091672801,0.9211390750669332,0.22274875428503593,0.9395801020889927,
 0.3282356358179582,0.2522643693168891,0.6201285455551633,0.7272347045902056,
 0.40537979017107295,0.10802751481048278,0.3024411217541495,
 0.2497722928864362,0.3520385810438833,0.850113816495689,0.6650138400035657,
 0.722364926615783,0.6098515438736355,0.445639940777631,0.6001438622898171,
 0.6943362915868659,0.20434839641679559,0.05028927693589502,
 0.5886742578698055,0.969561431493487,0.7329773286967519,0.12742783936079705,
 0.27141381938550224,0.293958162318096,0.3902859993101627,0.090768851035963,
 0.28588443760286264,0.4493303655705265,0.5338043697811389,0.4849589346279247,
 0.6666740098802754,0.9825142310267014,0.5069613614303009,0.8117273430089567,
 0.18670352027742665,0.7016399232069714,0.01660730919586173,
 0.44736353764325165,0.9427068228475683,0.7243378610657947,
 0.02336186643373439,0.36509813208371433,0.2827921044666737,
 0.48062057073586617,0.9099190768469138,0.320625713715299,0.46328444810714275,
 0.25949321282010473,0.8004032664584124,0.6616002394872162,0.8356731568890446,
 0.32869824751056453,0.09208861551450265,0.9080688124219946,
 0.6643018607699133,0.309594075794312,0.7251562936612087,0.060427542311482285,
 0.019303590128842307,0.6428322106618098,0.31062770383559035,
 0.5400286648958417,0.10111660214010343,0.023046328372309355,
 0.2764878602932932,0.8497926167869554,0.046536018667796664,0.514662522147784,
 0.24678117658088994,0.03294736219474337,0.47164247524843816,
 0.5893690976145308,0.04599710536992918,0.7735514434266002,0.9244726081588104,
 0.01756153193954968,0.9830967776528535,0.7579274582460994,0.6676436205583208,
 0.08674582636663042,0.13549092959188902,0.7052905037630496,
 0.8115483378053696,0.5124213769987667,0.8625304419871465,0.9732078942543861,
 0.6088622342834008,0.20076495347783574,0.39939982397485907,
 0.40467587481338285,0.8370597620858771,0.2779054424307381,0.427547493120635,
 0.3070467471904442,0.20177876017575003,0.8954023682936101,
 0.07184046345259176,0.4082256353325717,0.02764963753573535,
 0.25513467131861867,0.8638596007846288,0.12830895630585093,
 0.2180090885254079,0.5036052413157862,0.8544450816888434,0.20983632239172678,
 0.28512736575629627,0.2929328692082559,0.4536650718026083,0.6573596332220553,
 0.6622089041467283,0.25240887606871854,0.08970453830572223,
 0.2529649797153538,0.9499131567375926,0.6999708481953674,0.8842689085288619,
 0.13893949511704295,0.722996796529955,0.18631411918208873,0.5112360179199943,
 0.48495838411272985,0.6493254472172928,0.9666608115362272,0.9474903630812798,
 0.25074989120513114,0.5387440423279828,0.1604199363852189,0.6204001560582436,
 0.420335430413711,0.6710700498510673,0.6055980524916155,0.44404624792793146,
 0.20779577588775844,0.4239083479851339,0.16959459786689812,
 0.9723138551212004,0.25694523534149216,0.313720127187437,0.22571961778658423,
 0.11146428931473495,0.04306834307030305,0.9288606744810954,
 0.37024829490997835,0.22764423968540637,0.8656381257142102,
 0.9005134165433076,0.7093849905082759,0.22431220592836387,
 0.40417112056952864,0.84527632052145,0.3989161322834106,0.012291048498149992,
 0.8243404031209239,0.3885364378526093,0.6854077954166101,0.7544079033634706,
 0.40218282601902255,0.08108663897885227,0.7968442592374556,
 0.39868992425610983,0.4043055352600916,0.1928452854751148,0.5376364150685742,
 0.07511206296587902,0.7811098996938384,0.5486086036924376,0.7297216874694135,
 0.08129784848645771,0.3772080464422841,0.6160180026470687,0.5198751422134634,
 0.9815204156341008,0.06592863598457188,0.5490314790130281,0.8235766020442954,
 0.45166503584595996,0.012484353681220806,0.6324738138162347,
 0.9353237236466869,0.45503906426497087,0.7096236304793374,0.6008383958449645,
 0.563023697300768,0.3912680416419154,0.1774727739798051,0.5154081686287421,
 0.8844040362057055,0.02722663374026224,0.6555361490633282,
 0.49107685464490214,0.58872479070893,0.9820221770200689,0.9524630008920052,
 0.8826978051791612,0.8714842973942547,0.25628696281422525,0.4502903543114386,
 0.33915334958735743,0.6420503964687346,0.2817321360404279,
 0.11545004501937528,0.7020330369990021,0.22617476694244443,
 0.6648199638522185,0.19211212982406844,0.7337607331163151,0.5936589726185727,
 0.4000488874182474,0.9730590495788651,0.21051020985980806,
 0.47403713555069615,0.061309873778911395,0.34352565429141246,
 0.05838955581072469,0.7963530590261659,0.5657624348373927,0.8030649725117713,
 0.7858317403658295,0.1445931686637475,0.4856166983522442,0.46867104397273884,
 0.2566259375735773,0.017036474742447516,0.1915999344344652,
 0.8665066631797842,0.6124791032290196,0.639929785247389,0.28266669830262836,
 0.907199905843673,0.5651979055341673,0.7390546302016117,0.1405697705351452,
 0.11133396326537148,0.6710920907679111,0.2834102473980127,0.8318471737152723,
 0.13639713699576816,0.5154990288294374,0.9724913580639643,0.6705874018750976,
 0.17165564633698505,0.10160885363576,0.2450432195798623,0.5939715681661933,
 0.07183521203250454,0.8020618831634706,0.15167009797692976,
 0.9586343278001538,0.42013789219024544,0.9448733028394035,
 0.022008615231517648,0.5207885708118134,0.39000361451463006,
 0.5309641920089729,0.8801452828862719,0.8400214398512477,0.22091890032041694,
 0.350581705799225,0.80129775034941,0.5267088223340919,0.626890794636018,
 0.055170063788953505,0.4572528840357043,0.812169027223799,0.1830617365815146,
 0.050186465115895085,0.5570111024990156,0.7135098655763664,
 0.9747443735109562,0.9130665831737268,0.34080324121561667,0.8143991760739193,
 0.9104600782277649,0.6869714414079695,0.8313933946484623,0.161012956117913,
 0.4537897975457099,0.08894688897688907,0.6982138475369521,
 0.12904189075907246,0.8066353375082546,0.1082387555285943,0.5731698168827628,
 0.3780636888293447,0.9359265356069404,0.5435137031468668,0.8946308060135693,
 0.9705498353745646,0.6270791480223409,0.9786483378065364,0.09850869274784835,
 0.363179562770939,0.7709659399506212,0.9565973249909281,0.2047806228064586,
 0.48946299725154685,0.8025288355555737,0.8907633992230328,0.6234470460718489,
 0.8285556906042163,0.7934255047931311,0.7619317427955259,0.2725142225061069,
 0.34868188238352493,0.4467937533636611,0.954175629734633,0.6894627372111143,
 0.4582209950465346,0.07736993942370451,0.5611854665524183,0.5049702934557505,
 0.42535917785725286,0.4779940569055019,0.44046747977667366,
 0.7035703570424875,0.19983292259150676,0.5941637625472223,
 0.41135107442957386,0.33669891808651353,0.6038447293406478,
 0.058301884643334745,0.012073381625149149,0.8520046554681922,
 0.007174359269514818,0.41600182876772307,0.6477788236399878,
 0.8805035112326512,0.5857891214451586,0.3140619977678496,0.08714276449958558,
 0.3115547857167601,0.7744435035574518,0.27474330223499854,0.7212572448017354,
 0.6978350530565068,0.816565312109948,0.42393317613649595,0.4225208146144982,
 0.3241937578671288,0.4617871490341561,0.839442509428985,0.31785234441950605,
 0.8278583629867547,0.5895092201662679,0.8640120612279532,0.3727394230440271,
 0.33466571530275024,0.054989570773712027,0.24260219181812825,
 0.6936537634047107,0.24553743106046866,0.9260960931508012,0.5694353758982852,
 0.943699857721466,0.7952930098300067,0.3280992600585094,0.9502817779489167,
 0.5596490146103172,0.23289967949251023,0.1769293419525505,
 0.41067285742295034,0.7794208035162912,0.5165069176569208,0.8497915124090747,
 0.02365607278828885,0.4971141979355541,0.08042772068940685,
 0.5974889039163167,0.09952759790029553,0.4617547058250857,0.5895152054833247,
 0.9098375465195692,0.5280867735937002,0.3221794595387649,0.4424170285928639,
 0.7687416040777635,0.5397928160335992,0.7481378269139466,0.25431748660751063,
 0.9242672212846422,0.8922249765438263,0.8400087850548655,0.1348470325476333,
 0.672592020339937,0.03213042556003254,0.29118669998539937,0.8817911797639113,
 0.6725434801809403,0.11903194196341593,0.4824577573437222,0.918906370361664,
 0.4714553969489286,0.4048276381063888,0.3982733033483553,0.7234437631719709,
 0.9296680272611213,0.8586312503203627,0.6071280434199762,0.504053019902097,
 0.6353885104183917,0.7505866665248357,0.352330298613167,0.013084742800977933,
 0.24547245883816016,0.7178614921751325,0.46911363125554384,
 0.8986473976653404,0.15635812765007928,0.9071791240458579,
 0.33078902784379327,0.7740846714571884,0.12500498748640598,
 0.5432270853187434,0.013232327500619645,0.6521978672430508,
 0.4379914654898256,0.8992138280256532,0.06977398011336178,0.6608963141861184,
 0.7063420340757712,0.42601581149676426,0.43963219273265175,
 0.5940744775040433,0.5735893497040624,0.008819590405002353,
 0.8007349602505026,0.034549473236768935,0.8926503966232546,
 0.35724840379181033,0.05377255137373893,0.13728765858422043,
 0.426236272811399,0.9181750323761363,0.25800875258909584,0.04620864220395082,
 0.15397349125744597,0.7185918607527694,0.38675744746843377,
 0.7925861312255447,0.27124534468936135,0.713453810546398,0.17630127290033326,
 0.5427096461735745,0.9508338547113095,0.8707526818048968,0.5617029883883465,
 0.8409811824996783,0.7375188144316678,0.018723579530499634,
 0.7698843217807186,0.8915095463661237,0.21411876651647843,0.524811780094135,
 0.5363789306728202,0.4721991124971152,0.9329580498136687,0.28461977464628296,
 0.14320694533349276,0.9700242238042934,0.6860704572407998,0.5490924058326592,
 0.699095188484085,0.5792465006019322,0.10919684896901272,0.3126281673999607,
 0.97906395825379,0.19481183784038503,0.28025737568309783,0.336483446859843,
 0.5525915499063683,0.6862780657463153,0.6925293602345632,6.90281623982214e-4,
 0.1166533686759802,0.9860224170523498,0.1940154959373529,0.37901633916170974,
 0.43621655817383176,0.002993682138777709,0.1425579324737215,
 0.995762580599585,0.5161157734196085,0.3564184031532025,0.24120090254319204,
 9.383955311059733e-5,0.993628353481935,0.12993604076955423,
 0.9093450731202652,0.9045302360906927,0.07244198265095891,
 0.29393937739612974,0.7495455668540321,0.2690467173790704,0.6842987652217021,
 0.80041026888477,0.921503117127014,0.3537961512977802,0.656373533712598,
 0.40876483228951355,0.6328681375358598,0.571840657458337,0.7670672237543948,
 0.5380751403375307,0.7677757793982728,0.9391788655996302,0.5742242295425144,
 0.4217841937727228,0.4975502017343867,0.06886836149048636,0.8222786137765359,
 0.9217158987010938,0.43189151355246,0.9300301376427375,0.18768922785412412,
 0.5177695205302801,0.25695882239188717,0.16056403315977363,
 0.4135165434582956,0.5191360743629349,0.7842905869263793,0.6701521305677048,
 0.02196021361517253,0.9116087231953085,0.2843382598161387,0.7551714202107825,
 0.16702972977482078,0.49658904429699047,0.5728236670840724,
 0.4434068747958412,0.9602133811664102,0.006010300637421806,
 0.7361911705828467,0.03767050156002505,0.9532633025932552,
 0.017950668949508142,0.5643726526822118,0.7442243738419685,
 0.49560633237962715,0.7058268889662668,0.1821103413928975,
 0.09192870975547063,0.7185861263364267,0.6325371036618759,0.2781688215490642,
 0.45851182454857287,0.508539048157714,0.3905910276269817,0.9236767055695728,
 0.4392194962238076,0.29412512310205496,0.15110301076395016,
 0.08061838810709598,0.2037416288478643,0.3455854783212763,0.5314851062759471,
 0.40966815254339894,0.033936793871009785,0.3238616904299838,
 0.5485014443557447,0.8527449304449329,0.35870006455899395,
 0.40255160938750634,0.9411315017134007,0.53783040070532,0.32824976972644015,
 0.19567117686779478,0.33738594966613566,0.8961151907291959,
 0.5672616247615934,0.7655510326392326,0.8834563847804526,0.9238659941556424,
 0.5123911105608494,0.6132864436050289,0.047725734598840996,
 0.23825812311890338,0.3687035430475841,0.4817156322674777,0.8387100170040087,
 0.11706350430079837,0.09499024224825814,0.9945950025192378,
 0.15333005607870487,0.01067158927469447,0.41386214527027687,
 0.25185258857612314,0.3592579200856779,0.25422343842018147,
 0.07882000923555355,0.33492212221742823,0.5837612345024945,0.08492683143682,
 0.13598001630950696,0.8679976510275957,0.33803051269714246,
 0.8818398631024656,0.01672205118347392,0.19795581943442386,
 0.6505408042208459,0.9430939156746306,0.2436120356097019,0.2997580267596046,
 0.09890465622172717,0.7944489741859075,0.8751635252582496,
 0.025858408539695565,0.5101621606022655,0.5385232538721847,
 0.7015360587885956,0.2740978713349832,0.3722903512371214,0.05857124456668705,
 0.23583813221682748,0.5805841403422789,0.9922483557406626,0.3020809522005936,
 0.5084891658912484,0.03155081805833215,0.5208082179638419,
 0.34363871770659615,0.5133081858408673,0.6544588611994689,0.7476533661507048,
 0.22166913682143685,0.8171006138244221,0.843682779671664,0.3578612549651321,
 0.28878232911657253,0.8622459917451737,0.807128080643432,
 0.018551270292824884,0.1232674363261016,0.6240577562519437,
 0.8635253435087937,0.9094075682617077,0.8579910427817801,0.8793674784295664,
 0.26374608109656883,0.014961778674113635,0.13618365381262965,
 0.7248040747969311,0.546268458262197,0.19821318585098402,0.6706794076854805,
 0.8066358976961653,0.12768802735499052,0.4495060067183514,
 0.16552781544118078,0.01712979819575733,0.01567093640626316,
 0.6432659442452755,0.7511338657415437,0.6626102346868358,0.3296788955935589,
 0.7606288906136064,0.04986080496066858,0.14704785391202357,
 0.37351387404403424,0.746757580129808,0.11921745071514644,
 0.43303175530038385,0.4887584787857553,0.9691397822895561,0.8552884471355865,
 0.14055658609848942,0.3876817651197365,0.32564836311697976,
 0.8244742826903645,0.993655900345127,0.7491418962087671,0.2989423367744518,
 0.5319951324655063,0.14295969712576262,0.4808995609566431,
 0.05262063841407694,0.15918023430694805,0.5321727317709137,
 0.23082990468369335,0.8690228640170145,0.3089378170081002,0.5456463199104975,
 0.10334978768061676,0.5717482316141294,0.15928789501963703,
 0.7723356967673977,0.016267138296409467,0.15717201325296704,
 0.19736867910879352,0.7999478949120288,0.9721451665249555,0.528404529280185,
 0.9286899596444121,0.3375817221889097,0.8648266122691839,0.3666258716777653,
 0.8031196762068649,0.4361605704191539,0.07445830401432896,
 0.15673967175642467,0.3260462875332928,0.6838997958671164,0.7404407502136355,
 0.5025982220157635,0.39230239063510275,0.9826133828074332,0.2722784247763341,
 0.5110296130261982,0.5237582766780129,0.8803292325158563,0.10273529562515948,
 0.2075752335717479,0.7243143665669224,0.30876410530648857,0.5512277555166317,
 0.2381376159928945,0.4969414989757648,0.9801939159371613,0.5398468037134444,
 0.11438468728556261,0.2568294583509725,0.05529915937866814,
 0.12326512253815913,0.5724601820247318,0.7301428609979375,
 0.08618654233994238,0.5574374098177621,0.10103041783588118,
 0.7993108175273016,0.17177601717390412,0.4713767878773961,0.98792257358166,
 0.6806258749305651,0.9002801506226255,0.5083843611446739,0.19245885165510046,
 0.45952851725072463,0.2056651110867601,0.5082120251472643,0.5333070458843481,
 0.9080905896110293,0.9927218154637908,0.4994606638040595,0.27139617860213405,
 0.48770513347941524,0.2829794626733043,0.5049050461384196,0.9184615731358508,
 0.8766555802716857,0.9925693977292946,0.5900821469341555,0.21867433271545755,
 0.7812543586084226,0.2715066114108622,0.4162658804498989,0.44976018187315614,
 0.08288476463999861,0.3986673794668233,0.5661036109518581,0.9567122798626722,
 0.5039898647219733,0.40107992090694733,0.49489194412595827,
 0.042695381411722466,0.13761214814998646,0.6255910624310707,
 0.41231455799957006,0.8665093498913102,0.22592715834355237,0.817871358570103,
 0.5280060515069409,0.11936531600063194,0.7438212894959674,
 0.48798867111199673,0.35978599117823684,0.548148901082765,0.6180961074223427,
 0.30433121822196674,0.61983594089718,0.6372497080861307,0.8359096744591896,
 0.8976869291861668,0.41168718535043414,0.060488103925291026,
 0.005081115464589114,0.4128549227442243,0.919221932656157,
 0.34610507943613544,0.9367425772923986,0.17492052689274262,
 0.5286333620831422,0.6162152747505818,0.6794313032827665,0.6500516553608783,
 0.15476454960320019,0.5668238607374673,0.40344179871131314,
 0.3372554384918425,0.19494824211511652,0.5978285379466077,0.979733360724989,
 0.48421710844951815,0.30773497537849337,0.36911369977847774,
 0.7794817420757336,0.5615232741229574,0.34103782750491884,
 0.24199569443053337,0.36442832780254875,0.6590197923350143,
 0.8864417720085073,0.5430070162704482,0.39908900864493724,0.9642444169808317,
 0.7920584234445602,0.30651895437829046,0.7019444493515771,
 0.29703669704338576,0.6668395715223797,0.4655601587842485,0.685789877348802,
 0.9060042144265023,0.5843577247263831,0.2144938729124456,0.44949195158525534,
 0.1837727851587203,0.3464589537918922,0.6770783563102781,0.7776556705313195,
 0.9002469267743243,0.45419707415383836,0.6345535984473895,0.6001015371002113,
 0.7006341841607232,0.8488785906975028,0.3984048695614566,0.4821174016398081,
 0.7211404365781757,0.5991401646593904,0.8254157383913716,0.4971211397608457,
 0.3811024325564011,0.18053252279715404,0.06486054957122644,
 0.32628814284790686,0.5006920680334206,0.3102501741944893,
 0.39182457340541776,0.25152544758641426,0.6995229594086844,
 0.7060106101102602,0.3616757357151338,0.5723208575558334,0.9775683019188862,
 0.5661474863778232,0.2587350891809772,0.7326895213536376,0.3233111869412092,
 0.3138391455991041,0.1210620128479647,0.7885250315027224,0.7178109782451269,
 0.7485598591047957,0.9181747408385891,0.6168546898612087,0.13859103060782363,
 0.800039441694216,0.03494861011003936,0.27386434956155936,0.8890031195705949,
 0.10158226072925003,0.961988663572146,0.5568824117521413,0.6591543079462315,
 0.6131226893738757,0.710557068402095,0.8744399557136586,0.502494548163254,
 0.07274324797180243,0.7747708517160274,0.592630240920008,0.3581029535762108,
 0.23260486176619088,0.4937684932143047,0.8458369901799221,0.5445479274288936,
 0.16226782484313484,0.3272512453340535,0.20728864837918015,
 0.7456948233653361,0.27392388434045445,0.02856066629376497,
 0.3634561897126525,0.4466237621642679,0.11220555958159062,0.3309722210866908,
 0.5601993041970392,0.9253507220215474,0.3484750262019519,0.38788409606611773,
 0.07101222827452691,0.5572507081976095,0.6762816370729582,0.6675382020679593,
 0.6189559803421881,0.506132012771817,0.6151430281302027,0.7009939198492683,
 0.5347714533125493,0.14931296151181794,0.797987292324115,0.23913880140396127,
 0.3118792889807287,0.4416885784059874,0.8036850379972014,0.281160795667885,
 0.11347620684253024,0.28298386984195767,0.3887361921720577,
 0.030959282662132104,0.1329890597971788,0.17775549675772773,
 0.6543358080258362,0.5354807341491932,0.7199973654532646,0.7461973064639429,
 0.5470653448630243,0.1748262085904978,0.5198235700255133,0.5256633739318474,
 0.7619390348901101,0.19379560259736517,0.9391193627535528,0.5925685992591494,
 0.8411090328054899,0.23696855585061427,0.18513427281274364,
 0.7029580500513324,0.08786162246341944,0.5112638457709107,0.2369938687829466,
 0.5591015527394626,0.3649410762715445,0.07108913717842047,0.1666037773736223,
 0.38735749199320235,0.5016937672250983,0.2251677600571843,
 0.48677294319206144,0.9516035035628279,0.9650707136672859,
 0.41952428929946284,0.6877005514814186,0.2867497617120719,0.4329594381170202,
 0.1422692209815446,0.38610738927040744,0.6842162076051932,0.6010059986791297,
 0.8195069868298932,0.554377222073253,0.901284044869429,0.5702532144109754,
 0.5658935052057502,0.45378703933102793,0.5331449087215747,0.2443226077443823,
 0.10064451336289737,0.7254935152874885,0.2852928122678673,0.560120905691692,
 0.9788315024675448,0.9251428510838958,0.5256198831242471,0.7149012311652883,
 0.1502073187205567,0.14553915955904428,0.8278524397581947,
 0.32303154534197975,0.8132686532649398,0.9177399217331498,0.7860198758466268,
 0.8199650190825332,0.4684608068709052,0.3362689905691668,0.8426198650015065,
 0.13453904933716743,0.2789510105041846,0.5212126045688985,0.1707824875181162,
 0.1804614247869537,0.25594318838482355,0.8603164550099145,
 0.17787756862973136,0.016162324640052517,0.6503831000503651,
 0.7275195260035533,0.7208486013016423,0.14192876639635776,0.3099854981201816,
 0.6933103676492245,0.06357981853077832,0.6302627386768636,
 0.11227497936334119,0.7116304574058974,0.6876798924854061,
 0.08884796827830754,0.9539131176179456,0.4834307366936673,0.5064453542223075,
 0.8032925777647707,0.6705580576521324,0.36793602319400764,0.5107249877817381,
 0.063404507096528,0.5263962618031028,0.9816254964733893,0.08773857758420489,
 0.9526294077531461,0.8921453496294377,0.2189234239197284,0.17231149833990933,
 0.7403034393890201,0.04598852382079255,0.969377099939943,0.48322342472425794,
 0.7388079614943135,0.8607356804611013,0.7408638248370665,0.9803638975017955,
 0.8282669085459821,0.792902093787678,0.18063381590459415,0.2780598306988191,
 0.049224433799771816,0.23321001462990054,0.6738887711940599,
 0.6226226227686323,0.3589974916867722,0.5559176381090857,0.6056663587938856,
 0.2424039818180873,0.9508117782411023,0.308745250779625,0.10591629137902137,
 0.29810844148729876,0.43891941932823764,0.40256826895608366,
 0.3550695961890943,0.06921689249534568,0.3845902828497211,0.1908102470644859,
 0.21917829282279178,0.7883843301239251,0.3488888830438077,
 0.20046537887611515,0.14119308353404902,0.31248534921344984,
 0.5352615355784024,0.20609698025948064,0.029706476665010695,
 0.04428345871209194,0.9790501448477855,0.12119071438396622,
 0.3717407999079231,0.4294547819373863,0.36603043152349,0.705817085305887,
 0.13357412338964592,0.15006438081659068,0.9309039095633439,
 0.2251914772694812,0.8687788292617246,0.056144643877748646,
 0.6275182387192617,0.5749815816045449,0.41574380784947795,
 0.19701765668926896,0.9712467961731037,0.25493819157350606,0.749115238985843,
 0.23616991775553675,0.14294804644302495,0.016427272212927613,
 0.9012052662846084,0.7924050966723284,0.7805318825680676,0.3544712008196278,
 0.48421960962547295,0.672534489572247,0.4123525832806615,0.6860638327099616,
 0.8788784498073402,0.5975815251349108,0.12826441396968014,0.7697820010937798,
 0.5496502684174418,0.4208277679992729,0.09513471345659952,
 0.13429698889789243,0.1496582005471645,0.4359971230624844,0.7575393762516703,
 0.8384491481166969,0.19798892773985277,0.7741134163419143,
 0.36072997269186324,0.5411303185252649,0.31387511242068467,0.63446485628506,
 0.3860914904488473,0.10209447125057969,0.2109124839764276,0.6438442724324963,
 0.6967718385502107,0.4299746886607656,0.2537097119038919,0.977709103956925,
 0.7771046915097469,0.3117566685619322,0.8692347741273743,0.12455865469434624,
 0.8603406886272829,0.26613066983993794,0.1247525548114653,
 0.47566724812406114,0.572875626990916,0.38204953340768355,
 0.13536613766614514,0.6868056336098881,0.6383907649134049,0.6410743228824192,
 0.787516719208623,0.27058910268771397,0.25656005518162817,
 0.012922352020204086,0.0011339816548856518,0.7894495809403892,
 0.9047143684936653,0.22813732359335126,0.7317330574126822,
 0.40738235289625946,0.7967697280680232,0.9824684173441027,0.9282579747733866,
 0.705640654821069,0.3682067556834796,0.9018749799163585,0.9944171513902544,
 0.7177340462332847,0.7083767753599929,0.41492862111720985,
 0.31662504641994405,0.9386213076010745,0.8235199171230277,0.396841148688461,
 0.527886119960047,0.5703361468510328,0.6155524969222556,0.1037209743259574,
 0.01438545699165339,0.7423450825806049,0.46112776296428326,
 0.8128731473660034,0.4371383578550263,0.6760258544789182,0.4465389534084053,
 0.9650599753533854,0.3428856669901029,0.7370169882936607,0.6339524836034338,
 0.06373855994104638,0.7701026617920239,0.0021404709132077926,
 0.9274564891676209,0.7418632764663814,0.0973487380456215,0.32972918164085685,
 0.5931285995332259,0.9377021177615349,0.9769921102278573,0.9459710610625116,
 0.9277644507864489,0.24552711533714577,0.03039930957163628,
 0.49635261939952136,0.9891994491495539,0.5268888755309138,
 0.31907424949954344,0.8808906143096927,0.9873362952681851,0.5660573467447638,
 0.74222448158484,0.1891269275304962,0.800153326540944,0.6883525442317152,
 0.18642653464518388,0.09300222411087722,0.7057415484235784,
 0.5340515933771878,0.11382947584087966,0.12751334182255758,
 0.4587587542438406,0.2061823909254108,0.4612839444811393,0.8923271997615345,
 0.9154809491460177,0.5857160383830622,0.21250757605884485,
 0.11094888911814571,0.5567601219762659,0.3323004248691862,0.6981493795601348,
 0.06083705034469988,0.7026326666642443,0.27044822220453446,
 0.5466910319245436,0.796040200884962,0.7746329520079287,0.9005553590730185,
 0.8708634121538184,0.9160148346598302,0.23330757838967098,0.5063416844596839,
 0.20514865742615207,0.20569534522695776,0.0940718280290509,0.803958336056686,
 0.6035465765807093,0.8913625379442464,0.4653551264785061,0.581002120527006,
 0.7626129779482499,0.7597146540299344,0.6498855440848805,0.5621052648548002,
 0.8563696644932701,0.4581371927028526,0.15684270267547784,0.8203370007644095,
 0.8352090189571446,0.11928975383906215,0.3875017754908132,0.8486733434200806,
 0.37537710980070815,0.2029071107974325,0.8643910080902897,
 0.02017279516728354,0.6770496689380672,0.46756502271722555,
 0.18786737136410736,0.4086059273005025,0.6017799313729257,0.7223760190624703,
 0.9856162102986473,0.2562142562097417,0.6441765021876716,0.8960506211416295,
 0.665095196497019,0.9475380598667807,0.937107298562933,0.10980833711819749,
 0.3077072122849839,0.377505078613921,0.07303255002261455,0.8852727432444808,
 0.5900594695262661,0.4674178401712922,0.617107191667701,0.3068933934650159,
 0.6716364630146912,0.863834363469746,0.28333011866277813,0.9038453622497493,
 0.3807627995894106,0.9461120437602447,0.7031614068656851,0.03097649135749181,
 0.21186127623218942,0.05302487446385029,0.45866021246879907,
 0.6194513643533622,0.3853345419561264,0.09230330604554504,
 0.17624806434446227,0.8395968939148194,0.6503550627078705,0.3111355116339616,
 0.36887420504427615,0.3040072246135562,0.555091902239476,0.4118931931981056,
 0.059879206459181655,0.9539610442589023,0.8877503612611206,
 0.36295540235406665,0.06707137663049556,0.5855137175664844,
 0.9460642971735798,0.9683943436157766,0.517484894558833,0.880019544027209,
 0.04236188477236258,0.9861370479990035,0.47762015589983176,
 0.34382159678713076,0.1721132747177544,0.2839464541808646,0.9339153197543822,
 0.45597949305812346,0.7220121914663468,0.5393624604365743,0.9009443490674991,
 0.027215817503656425,0.5587895754644437,0.4134162380409673,
 0.06679929145103702,0.237871952706334,0.6581418569630835,0.01935639873236894,
 0.6711918758948493,0.5970983003784656,0.9124630292835595,0.7951786630147537,
 0.5903948957923221,0.762698808438993,0.6449154942575108,0.3288370949420508,
 0.07432993853242453,0.7835517603520592,0.9112614144316421,0.6058010052985556,
 0.6367798026156992,0.458671048507976,0.27836456685484934,
 8.289677394257566e-4,0.9490971038149056,0.25332330097319844,
 0.21189348824773568,0.7072400385159511,0.7166829379954103,
 0.01054610319293292,0.5704566049702687,0.6206846938226875,0.7794459874933795,
 0.2626898924896818,0.43626561495077487,0.3941774397889075,0.7213714822671123,
 0.941708257957474,0.9539472630814759,0.46493357899060994,0.22707950002007238,
 0.34427660864264287,0.5817958863763015,0.39416796492328443,
 0.06280100063816074,0.2691823454795854,0.5473633545883906,0.7608579140271049,
 0.17469662730225943,0.5040297974482817,0.20886102741275703,
 0.6521680791539065,0.3615068713008027,0.3479114237386943,0.7904359222709815,
 0.5071985482290633,0.18951567809794656,0.516329604095126,0.7255765134531688,
 0.8518285807430025,0.9321805440626549,0.5691988447750191,0.7465901439906907,
 0.5524323324323173,0.8846558530157181,0.5840730083528497,0.7932578283103269,
 0.9708296929272193,0.5627069535709304,0.6043873375598035,0.805317941109096,
 0.6147591642101939,0.4382826596732473,0.3338654643605722,0.8550042507543006,
 0.8786447390648882,0.2763659915903316,0.7359675997093705,0.19199415954599763,
 0.9977283073841967,0.7947316395599658,0.48183649542487816,0.6587445919755426,
 0.7015435761396646,0.027368496812288035,0.08295847838623516,
 0.8328339852953294,0.06873106050265732,0.6612355637028973,
 0.10982019894164563,0.5999945962363302,0.15802012628537643,
 0.31340326146222064,0.32945290413980277,0.6653048038172188,
 0.12426639496925862,0.7000306553258882,0.8645853334830746,
 0.24319575823612283,0.7393121670805066,0.9444087053131184,0.6629218070849532,
 0.4083206197515624,0.7745049305081182,0.30545002626905615,0.526771955962773,
 0.9557162258586739,0.22211333121427013,0.32121826591452696,
 0.19771570243386571,0.15781565557445498,0.4876979301173381,
 0.3943596804585362,0.044066504732288014,0.4279424874611244,
 0.2445362687086493,0.013186374472022644,0.8211072319466115,
 0.9508815579161638,0.20433522707411456,0.34260446188577465,
 0.16024261257151828,0.9499659354700684,0.8611631621326299,
 0.15710731418453905,0.7571209278495759,0.5506845250172383,0.9314931925343488,
 0.7681673686333397,0.8784170298311377,0.8476403124628692,0.20504482979990957,
 0.10188646000745827,0.39563464947703564,0.043681365258459226,
 0.09766317553278736,0.621176696706593,0.39385721415133246,0.8888222939661929,
 0.8712412167379041,0.6806907935515513,0.6141063808095949,0.03598643538743018,
 0.5005433522093582,0.5381165437462203,0.6417283992175895,0.3978263906670094,
 0.139424546997734,0.4203329203219227,0.709776350297702,0.1286201130552207,
 0.44301228050439323,0.7556349035487714,0.6970763427456178,0.7008270221219741,
 0.2702193756269,0.9688367853201331,0.5825140906362956,0.5227423163142676,
 0.7788411681638477,0.9530464569118418,0.6144269361267838,0.7033205446708215,
 0.9738005044831132,0.20354077631612544,0.9934955920399011,0.5800859017972007,
 0.509381463120712,0.6566722658321941,0.3791071023753443,0.5394253891679816,
 0.019332340648803403,0.3078576025384969,0.8549227449483512,
 0.38454622707295893,0.4932824273963968,0.30299366356890767,
 0.7816868273840051,0.7370323201624526,0.5490376806267361,0.44830904083039214,
 0.5752673811473348,0.842245619064117,0.6908906243239443,0.038606014055687754,
 0.9599192090733468,0.33255882210201326,0.12907375861269132,
 0.11236182902698677,0.29144637785247174,0.4517721283027427,
 0.8045720395322027,0.9758836757771737,0.6643672349250138,0.4747692368312626,
 0.6498808681365249,0.5856098910370264,0.8798414873335111,0.3656826236900731,
 0.9404642827890141,0.7833220575595188,0.5371053994501436,0.7415802782004893,
 0.920345397929423,0.3381535134560262,0.30193118880890035,0.37539851699564464,
 0.9724187769074465,0.005588198003940503,0.23909650941133842,
 0.22963163156828015,0.1290972847286751,0.23951146814834023,
 0.7141627525821345,0.3050367127782291,0.23402161839830682,0.7817439521782311,
 0.8219651892358576,0.7650654406107994,0.5052562779100729,0.2804022279426702,
 0.31767740676370737,0.6914733807045639,0.11769725241471207,
 0.5863464073489328,0.9672043410967774,0.5210445952051231,0.6870186384625883,
 0.518235058924275,0.8928518853371625,0.8581814029711731,0.7376978351833047,
 0.9817056266635807,0.47988442777859697,0.4720415488211053,0.4122095827666422,
 0.7769496192534359,0.6954475653525863,0.28562183631397997,0.9870847898739247,
 0.8049392669561771,0.7697898703633732,0.7890352434238577,0.9701824760828128,
 0.21460017779981622,0.020082413697652957,0.32684999869833853,
 0.4334563006723552,0.45197741004401304,0.09243780532044832,0.52343974853965,
 0.8540206729164127,0.43432608260011185,0.24181559148105425,
 0.5633535119665494,0.211499788716339,0.41544513713416886,0.9467449663378115,
 0.7631568667973571,0.8711649018747293,0.2631421777137334,0.3513595333362233,
 0.7606766474388691,0.5272600849193202,0.8235196234601595,0.46579874479443806,
 0.993799691871486,0.4882929380120139,0.49609538544108467,0.7827987822618503,
 0.39112526786179214,0.6574038109980547,0.055194689404321484,
 0.5663753904218898,0.16712281549222885,0.7214247637360949,0.7592469937033217,
 0.37421205738457675,0.8267014210271934,0.5819774034339931,0.61471494012116,
 0.5972863853965242,0.044257810292370414,0.7714886301185135,
 0.6072263376301628,0.0010362172457184515,5.975853006965348e-4,
 0.21491948384492454,0.02672974608960388,0.2361105834740278,
 0.3124256118582541,0.6881741909592377,0.01717498399890549,
 0.08783663108800466,0.7529279344576392,0.31859121124841394,
 0.06321804722423285,0.5980231314844029,0.4473734186815661,0.4560410914061862,
 0.5256688487497243,0.5717742843408657,0.6554141218451421,0.41150604831687376,
 0.46794500214448354,0.4879773876517255,0.9555729406380532,
 0.05504309437225474,0.41577075831286403,0.9459156470985414,
 0.7138162072933213,0.6945545902635416,0.4256966200492791,0.7556428103957202,
 0.3595433915975854,0.11405568282946288,0.3551775022329564,0.9446285628869687,
 0.4401540933296414,0.24097013974388604,0.4491555726051836,0.3146737476047121,
 0.4906681284622809,0.46887275105808557,0.818823643914436,0.49030841062784025,
 0.19562821916020678,0.6169226420252417,0.7591401679729773,
 0.49389555365492455,0.24518846114918824,0.8335015064672341,0.943941683050247,
 0.9073444228645928,0.11541102256435365,0.7617934011854472,
 0.004603092081541527,0.7253190646881016,0.36783541605533987,
 0.7034869510878524,0.38360810551779323,0.10528059933258538,
 0.5536088019053828,0.3680030197406834,0.42255120705122784,
 0.21084264470354164,0.3011705552239615,0.9149699572695602,
 0.38475677342072423,0.5620725430079133,0.37810333653859607,
 0.3252826418479533,0.19009397563643227,0.20165766475734304,
 0.18984750013714324,0.20537358367381187,0.17209928663523888,
 0.9830718026401597,0.22034749817684096,0.16653320959128592,
 0.10915467344323115,0.36749716826706125,0.41005045294763387,
 0.6095360973156806,0.05738375437990917,0.01864410624557955,
 0.21386942332199066,0.08689016469495936,0.8044978051900297,
 0.39084966209448513,0.23284056922702068,0.05249142945054919,
 0.031814793166108024,0.9822756780988259,0.7767741968353915,
 0.11764060089302109,0.31604669558808185,0.05603988782947367,
 0.9247923333274759,0.002050655754995634,0.9144372768000009,
 0.6720165599494334,0.2832570558001618,0.22957158966685043,
 0.059474474480475714,0.40507681292781483,0.14482092389387935,
 0.23314246057537946,0.1275060191960038,0.13175953548169383,
 0.5988600824419054,0.21572379458235347,0.43314340732066225,
 0.7307203464922901,0.5315851105884223,0.045207229188856446,
 0.6108728993422944,0.5577966700073645,0.1744196583599127,0.49786437175679765,
 0.5457966912704263,0.5795985955872178,0.3455237981136309,0.8916842370591901,
 0.7985410755465304,0.23910534438880815,0.8059893187974991,
 0.49290968527686563,0.8231992503841488,0.757573232775575,0.877872593677282,
 0.8017003856745578,0.7538114006444042,0.9979303768914936,0.7900121167079508,
 0.6348639555749114,0.8780107120918981,0.9659436341250457,0.3694069974434,
 0.949269432829924,0.9205366722060722,0.04152345887741393,0.21367700517738186,
 0.4138622712895943,0.9842622242307117,0.08075276960415756,0.9743926182796642,
 0.9747376018485591,0.003210910573569037,0.8075499092369947,
 0.6377044591960765,0.3460381338716009,0.854443356403462,0.4588141246532793,
 0.7317247762976844,0.1371847137557134,0.89804752582405,0.6480056875395002,
 0.23740532452537522,0.6317547887618566,0.6161076487136294,
 0.07209655806269066,0.8202768687476313,0.22593498995163142,
 0.6149785925845499,0.7205762484194705,0.2437989336923918,0.26158661513952697,
 0.942670085984399,0.2917073073644233,0.9792844850772964,0.6601877257586036,
 0.9222735850759527,0.31565444421364464,0.6227693234151312,
 0.20179103744669746,0.9661122742969963,0.05183117390026104,
 0.8105487859297955,0.2182918011430054,0.9645854513249565,0.545695306135311,
 0.4069374719907056,0.7316575945862149,0.3607323296501269,0.2867599614726921,
 0.9259536098312007,0.19273697696564573,0.7755324568175755,0.145659915113048,
 0.8573843799059666,0.5059788929765088,0.91449130196379,0.2017828701217611,
 0.06227466627768252,0.05186300543327882,0.5407957559748755,
 0.6430781151255799,0.09563535813120849,0.8988853543966224,0.760168368305326,
 0.25603848223416403,0.6588803950489811,0.5106957437922461,0.2826507046314589,
 0.3423055121160309,0.7704465063930194,0.3530430332030937,0.6220680765987159,
 0.4511339164687366,0.583672255940931,0.22305858751132412,0.12195561077555384,
 0.18889230289775238,0.842052153628689,0.8956290617358594,0.14902768103770214,
 0.6869126025207729,0.11285285633310871,0.759598896990418,0.8703856694788739,
 0.044066265593826315,0.9537251871036312,0.687290182477006,
 0.12014061832938816,0.6067858640358426,0.13831662640122055,
 0.16381740632440212,0.6623668125218342,0.28851872407351564,
 0.1931010750002521,0.7134548536665009,0.08849673021364934,
 0.09261972023868204,0.6261874797638759,0.9396793792488417,0.2644762823154502,
 0.6709971055032871,0.7153164694326295,0.7833906550013683,
 0.009562707649359958,0.5716413558607318,0.33131411824577095,
 0.5386734723810885,0.38318726184930785,0.6559727596868845,
 0.37525856362648746,0.03237323529516489,0.5570862714797608,
 0.7545648913043869,0.15385441741889638,0.12088689665025343,
 0.7504421818153862,0.8083714674160255,0.8905561465065968,0.8000372518638511,
 0.5523084239593526,0.5463321766299992,0.8433159739611507,0.37428037756912536,
 0.6622328358852186,0.49583946587452576,0.859710833367646,0.26588652907565946,
 0.4177242805683996,0.14637781751019208,0.2467235795839562,0.6147670712629375,
 0.5411580500502012,0.8605675836711509,0.40886262844118126,
 0.037180013027535375,0.044461892727809205,0.37229420204462327,
 0.6586750641292527,0.4875187091756017,0.30481289578064286,
 0.32332642688671287,0.7325946054163984,0.7924201386676973,
 0.10929955725133256,0.7784057321291667,0.06596498526774097,
 0.7174903640158357,0.4030176589230856,0.9811131623403538,0.9232143941940164,
 0.019405058937535635,0.5304060222934418,0.20222972673708017,
 0.779271298292578,0.5949918875981284,0.492166771278542,0.8922410615929127,
 0.9011323599969883,0.13605349080725793,0.23694477271215253,
 0.9636756384371926,0.5573119830433967,0.3934234028485367,0.3128639557065114,
 0.93985581059618,0.12167326048205429,0.747507194278522,0.4312423051420592,
 0.7919702529521395,0.7071410096208531,0.7366286996007622,0.8785013644409728,
 0.6096882981481472,0.34209356720694295,0.5664238846901058,0.558161503266854,
 0.8486456506923317,0.3267249061327687,0.8085562478469104,0.9432533594191664,
 0.7376551086993767,0.7363142079262152,0.24200725460265238,0.7527631626659164,
 0.47234883005267836,0.8835534983569808,0.7260562073119208,0.8080354918899407,
 0.17078159420171113,0.1593693275470377,0.8387121167510121,0.275696186866521,
 0.13139895098553644,0.1904034911892697,0.43007140347407746,
 0.7657987685140027,0.1172433275643805,0.7235047717486937,0.7139528821272656,
 0.08978821935061343,0.3677247231704375,0.13660809891594983,
 0.3519627959250058,0.9005217243393001,0.4913842709449402,0.5141280189444366,
 0.27125801961923035,0.9109673410062866,0.6844791160252063,0.9058272810690181,
 0.6768493181990113,0.5598810286994558,0.9219311156285405,0.3727852076196788,
 0.4068164500978416,0.96685298891567,0.40905624951983643,0.7819563247258952,
 0.64112906859157,0.9636280841358363,0.5248108038213898,0.34305877855749844,
 0.6808733164310332,0.12992780582017704,0.4608836579376634,
 0.25927782204910077,0.8819550009151589,0.1264382769944078,0.1493183372032385,
 0.31230565387258624,0.9664692408819336,0.989043221166932,0.8652374001099264,
 0.544203984088234,0.8436425897132616,0.4899108177162672,0.513016367431172,
 0.6208911140649439,0.6474915893502364,0.12396205045066311,
 0.18673530539443262,0.2411756273920509,0.6782304604449872,0.8751379585447303,
 0.5867459956061698,0.26901790171391315,0.33036701545980596,
 0.28358683471200186,0.533103360319472,0.6305926257210589,0.5309797529739521,
 0.47319233384715487,0.7545289565282909,0.7637985877889409,0.6826399895097274,
 0.9035627762239924,0.17222892481262975,0.14095108825738334,
 0.1470106792830832,0.9421633913348567,0.6471936460037343,0.30348548554655375,
 0.16343564481420847,0.012360993241874896,0.2705101311748628,0.32948204435731,
 0.6852741670591892,0.6601108881716654,0.12022645548263211,0.3333681683764631,
 0.28338157802993913,0.8538125614267842,0.14937566899412358,
 0.9332107024346251,0.6861312096966272,0.05075541926036242,
 0.29502765616889004,0.9149681481322128,0.5209535872798414,
 0.25049094100979774,0.38651952132865286,0.5708446557994248,
 0.1121306546060521,0.6884647332221728,0.6626490172815009,0.10239939893702832,
 0.32271484629890823,0.426929380261757,0.09804928400188384,0.6161743499432935,
 0.09646595721939599,0.8888696224602386,0.49813631167548333,
 0.24073522691188065,0.6021427979014605,0.8606930408232862,0.4568405636387971,
 0.4453268194331601,0.8698704830467429,0.4062836646589414,0.3178548903061976,
 0.4048071704938535,0.23351277187536745,0.0821360816872867,0.624437714351593,
 0.48517124575455317,0.2472027904092846,0.8834624752989448,0.6491317084053304,
 0.5703191008120079,0.24690997861619468,0.8076163047956286,0.9526358227811553,
 0.07482152329818292,0.8335439918519739,0.7838396858006524,0.6931441532032135,
 0.9401284242045866,0.312281920233634,0.45984760234657407,0.9191229866237904,
 0.8657293358704976,0.9456839978151668,0.06913808675570321,0.805466040730118,
 0.32291952967384463,0.6224030557062014,0.8673483640820103,
 0.13412132845292302,0.13380499823317926,0.35253377487169146,0.27150417731828,
 0.8555258171047267,0.22906737164705204,0.23958245139896273,
 0.13551364685912004,0.7484307678364777,0.17548280263587002,
 0.7371705630329849,0.1417486703066858,0.7305564221226318,0.5939657293484575,
 0.7614008725655164,0.7364023905377524,0.3583679304073487,0.413408167563029,
 0.26043131431953526,0.9234481098423458,0.23979206251000462,
 0.6112829032539749,0.9723755054897204,0.5613667940401399,0.1492248260104022,
 0.9627627269697581,0.3666471180273202,0.8179025904946199,
 0.037164111235275965,0.9864997537671832,0.5919124811493037,
 0.3620115766312598,0.9050011196947647,0.326850131102261,0.9021471330177903,
 0.239110454021646,0.08612136612656973,0.59515816641318,0.006984553142696193,
 0.3214615086861832,0.5331559490995711,0.1551989722560141,0.3778683467849382,
 0.28807815805614334,0.6755099632294967,0.8916692341778771,0.9897398497625122,
 0.9104708153633934,0.016313433594330573,0.8208877985262377,0.635826081593957,
 0.2841230037308373,0.9855017827653619,0.005700440362714776,
 0.7749678432010344,0.8593105033092985,0.32565364909029426,0.6344008301738311,
 0.2704104158922517,0.6546248175219355,0.4132800592670124,0.32376900289515576,
 0.48409026717336023,0.5260864027072764,0.9338490665855506,0.1716762191157999,
 0.5415232272299342,0.3297771034248793,0.1272522259155997,0.7794938305242709,
 0.5972287565418393,0.32635471575392916,0.013819932956730008,
 0.02275884290888519,0.5200993506719144,0.13078791904206566,
 0.00948602888422534,0.9089859735932126,0.9006300342734915,
 0.05524449845476487,0.9685102729648996,0.6620030967771703,
 0.25310573332327924,0.8160924587776464,0.8267782741420957,0.5522204109009292,
 0.8704571559321563,0.16738329700019472,0.6072678542283666,0.6655518607227673,
 0.6226728181331981,0.3011129824751717,0.5842518797642853,0.9518148655717542,
 0.7757860116615166,0.5918572713293853,0.3178721546523896,0.31376224011306875,
 0.8856641484820714,0.3989921965769514,0.18705697020811152,
 0.09166576050161535,0.2676673378422125,0.15770616420848604,
 0.20058926358384455,0.9051440456322729,0.9378315086130021,0.8362271232302749,
 0.5268221139364682,0.539819886194163,0.14233571831282898,0.32247301078634527,
 0.04567165738933854,0.5071529030261551,0.681486216044502,0.03676558194281265,
 0.612613876975921,0.7115180157985173,0.3956466345227483,0.516824458405962,
 0.3134324056059161,0.8811417519090111,0.42743427340389095,0.3281994514818082,
 0.19372799889966097,0.3001579961104902,0.3745197015501398,0.4141007117802258,
 0.9551083589671157,0.1896020166403205,0.654653152855918,0.03236274409605333,
 0.020764429483530966,0.4542873275319549,0.7765280146637114,
 0.5127865368895145,0.561531274818141,0.9009135350437611,0.9383148643910081,
 0.3692391384906004,0.10770117931529288,0.433803487342845,0.7085829484881321,
 0.6027703359182746,0.8433625677092127,0.06121760670576759,0.4062967870936387,
 0.017293329168311944,0.7837799293452089,0.29135600996623956,
 0.8727390782557969,0.6662816058051959,0.35522159100996376,
 0.09894510846234228,0.7520792817603248,0.9125252293489379,0.9513654698291162,
 0.29074490107611006,0.9055159711941776,0.25627208832914294,
 0.46627228480635596,0.9314985730921035,0.8428892341081926,0.5105219154118623,
 0.8424805368317674,0.4069331919513215,0.7159702821920961,0.37454915837519254,
 0.30863480395305387,0.39715165597482915,0.7152732537322201,
 0.9019639004538841,0.5146802069213628,0.08538906599042106,0.7930448614091322,
 0.3866054418718956,0.6992432581625578,0.8913347089413096,0.9248477507053896,
 0.8562750643001905,0.6284873631412515,0.2673569698536885,0.17330882688026794,
 0.6074350348949035,0.22853023858309018,4.3377397379895655e-4,
 0.559239351151992,0.4059864351999082,0.9311742682441736,0.560185386949185,
 0.4429617059971931,0.7418497672220211,0.2137240233360166,0.8162768671759838,
 0.28384785801752643,0.7331465762499542,0.9092484548368996,
 0.12061976418878706,0.49018452871092744,0.9680796974399971,0.630684530048099,
 0.9749984134984606,0.14052908635295658,0.986845176114785,0.8346742724567984,
 0.8951157075093599,0.7774817884732352,0.7878352122772989,0.9831719366630857,
 0.2548001205169075,0.4722339211582922,0.6484976739885874,0.266460929188725,
 0.4129766325947083,0.7783564976574315,0.9679311295308347,0.6660459765574435,
 0.545010815744344,0.859704754555211,0.07415448422518878,0.5208581738061476,
 0.43325348464871116,0.46569792616528516,0.6920711801174582,
 0.9326809365892162,0.0714180271069027,0.9164100877047948,0.2280276099265055,
 0.7326260540270934,0.4180566048221911,0.45984429414454087,0.7029819517342821,
 0.7905145638768422,0.050269504012094846,0.8476946223545317,
 0.6961313078542705,0.7326591190438361,0.62064636117846,0.8258243905909055,
 0.8462668995040411,0.450916378407447,0.9943270711075789,0.5210230712194976,
 0.6491135059432849,0.866347618331369,0.3037696609059132,0.01880896273849253,
 0.9863021417009858,0.5399709697994491,0.9561729105985258,0.8925957330302512,
 0.8700265887726324,0.8167396997490025,0.9223877049352115,0.06798294726799348,
 0.208212035449578,0.1292249013478186,0.6327857612006257,0.7187510324657532,
 0.4239508981897362,0.5312621047689791,0.9963720403912284,0.9048867081887453,
 0.08464644358320761,0.11739298742285176,0.3237773194464607,
 0.7099940443091861,0.27786231897631897,0.7873440606482408,0.4295552686840357,
 0.7833364691915138,0.3550717554168429,0.0529644530284189,0.8275377278551286,
 0.6318954720878935,0.12326276163401417,0.01656458243028358,
 0.4088971281231025,0.53062491771635,0.020385307601983427,0.5795421763122968,
 0.68712613110033,0.5698088682203041,0.5635708245304395,0.9585562714796113,
 0.8618005043114376,0.15064605494299188,0.04865631322583297,
 0.6000417303061049,0.2768610827983693,0.18722684441953497,
 0.23616372898915527,0.22246024130048891,0.8782888981205306,
 0.5956791934415486,0.8583370386274061,0.1729031625131403,0.15798457482903683,
 0.5281321338653725,0.014106955635679097,0.608713576675186,0.044105726345119,
 0.360880045372152,0.47331726816064557,0.626763162782074,0.9033814552618545,
 0.8705688324027492,0.426820652758118,0.6175009216907918,0.17977931845748762,
 0.7915039170499166,0.8226941392314334,0.2740520147670904,0.4837075978282529,
 0.7927849284033424,0.12376730808726577,0.41245034269968506,0.568970488493563,
 0.2943525127507248,0.10609439501761142,0.8032741955677897,0.3207993458699495,
 0.6847480470192141,0.6179970429273686,0.8930837714450395,0.32664428476972707,
 0.9484651614545627,0.6102862341163773,0.09656500487215813,
 0.18826392607975473,0.3985541613477783,0.10102753651676655,
 0.5116966586207463,0.2717231778959439,0.5312287139950006,0.26658271460523253,
 0.6449695408846685,0.4438150216075154,0.10878198100570713,0.7211667401065575,
 0.42380322627990996,0.13293394469080633,0.06335758743162245,
 0.43441583784624016,0.23195237284447323,0.6607518569817262,
 0.038845617157174606,0.7112078205580834,0.6778377992843947,
 0.6774667405797664,0.4217912906308341,0.46111073025554394,0.6207938468534346,
 0.31668175868576043,0.604669311315646,0.6888051501959644,0.19863456030860827,
 0.5507795808806615,0.2692958850500592,0.6585663538529808,0.8431468293906222,
 0.46572800322591545,0.5388200819691827,0.26300231945237895,
 0.7057610832543436,0.738582931452864,0.22801085747812821,0.30814759851693985,
 0.7752961284345993,0.8970269284634399,0.19225868895761622,0.4039485070514698,
 0.786417278745211,0.8252939919590723,0.2282976968584839,0.7326342638247868,
 0.036005985433080645,0.36701219770046867,0.4799574254159118,
 0.16259415970656632,0.6113045642895161,0.8026051537009681,0.7862466576248419,
 0.9403665931290167,0.434221792862968,0.2775141560575144,0.2204816799813223,
 0.9835833457921035,0.7497780143394955,0.3457071425954369,0.8891442145383215,
 0.8791948675198349,0.4793542495787293,0.17913255220817081,0.5296398475784309,
 0.08346196151991736,0.707057913242171,0.0021787903149880217,
 0.6468648551074124,0.679249187311936,0.2731209871222354,0.5222164317592584,
 0.45963830349452184,0.7319309798011441,0.3817806211308221,0.3417349745938101,
 0.03634968965486485,0.4600788691591946,0.5086678461084561,0.9858518721021328,
 0.14029257892068647,0.41352308866536713,0.6471571647764283,0.955382194799737,
 0.3870482498707639,0.5362521885599707,0.9299118750483572,0.3904744244477516,
 0.20116965826058453,0.3492173788204487,0.2813163312778243,0.4901038809954883,
 0.9738264005858158,0.3710673843841139,0.27501186310730763,0.441841413805373,
 0.9761087634330584,0.5901160071064473,0.5674135787444509,0.2768814624601639,
 0.9258169040094123,0.5075852199586953,0.9752830711351176,
 0.013882043084273765,0.4702138551635481,0.3844457649231334,
 0.2899069151319644,0.2542508646131497,0.7286997165946794,0.7871693303229322,
 0.2904602295150883,0.358570259232297,0.30976258506635157,0.08054174134144498,
 0.04625037427371337,0.8710239176012461,0.5650085518125757,0.3605035140435847,
 0.390922229930624,0.896011721906146,0.5670215056340561,0.9622735060141732,
 0.22737190778663907,0.15621352257439725,0.3392165781785006,
 0.29245549461922193,0.3201650910780357,0.4749217564797721,0.8875840913149072,
 0.18954734993632505,0.34606899053736795,0.08521931674075944,
 0.5242161577981099,0.28383444735568875,0.3270499822727134,0.9762491246323899,
 0.4094192108284914,0.5383614329543626,0.9028669271156247,0.3657469420455396,
 0.010733225600348062,0.21978000346884874,0.8369113790008703,
 0.5486758694817924,0.5712148269711013,0.3068038038413061,0.5993620126763013,
 0.43324866522289107,0.6758077611301835,0.2686141663739747,
 0.11877320706236572,0.06888359107614372,0.8175801898608488,
 0.47134756728609917,0.33422053856298417,0.854616723752846,0.5151554860304786,
 0.854261949996425,0.31925218516847464,0.914733861474384,0.5637753334965407,
 0.714959157293468,0.5823957341004041,0.7047650380968028,0.7029051735968579,
 0.5264320484333406,0.5153186155510981,0.644914063971329,0.49716503357811936,
 0.23786424439723075,0.3782033837907379,0.9748957194134525,0.8653830351329161,
 0.257832257044297,0.8035765408342573,0.8219658463810269,0.028821823815621483,
 0.4849481403902307,0.901910562736774,0.5032753898537395,0.368191080857598,
 0.4851367098090569,0.875119103038951,0.8778075665933195,0.9457630696236767,
 0.48874366674076297,0.11075953579131881,0.5250431308577528,
 0.3335097211708592,0.4353414409537735,0.8302609363361053,0.6479035627129203,
 0.37221673020843793,0.9475744685206582,0.5804252774261762,0.6318089737910911,
 0.3592303891518044,0.5633510741442381,0.8666330721214975,0.07381872381279175,
 0.21601382978636918,0.9315470278057213,0.8902833734597883,
 0.06532368346658535,0.5824047907814407,0.7125202747581535,0.1899732992949964,
 0.4948035275120771,0.9260004510844696,0.03068867250336771,0.7143522994167714,
 0.9674335364363975,0.2565829353009721,0.3173777342485946,0.7305232665890895,
 0.5521804329334027,0.9648548401626091,0.37673313294097754,0.8866590904904471,
 0.9777747487934774,0.30763024190053667,0.4270980887100403,0.2417797257448796,
 0.1522906986686564,0.5118959634889375,0.04217010572780977,0.0705064867159404,
 0.8061797809825137,0.28501615080630616,0.8367345173410812,0.8730358980068645,
 0.6395253986670989,0.5383604556530533,0.8876644484218917,0.1852878726406616,
 0.09622119219221359,0.7189096523911002,0.11895190752804308,
 0.05466616803065161,0.9620587282027298,0.8685945270975866,0.7233996064092014,
 0.06371486490793177,0.3563459752055065,0.3764095172314592,0.8275348131602505,
 0.02910885726252077,0.23631978786853614,0.5065000858809642,
 0.6196281043030649,0.5923624637814947,0.12243049087460944,0.5093034468061438,
 0.16241173274508602,0.004385868421867278,0.5477333293865284,
 0.5641953884827595,0.7770305084139082,0.7481116222660561,0.7040321300497532,
 0.3155284310846861,0.8587789681791929,0.9700072418697958,0.16273741365423566,
 0.03570164473651294,0.5913858634050074,0.4530963972416706,
 0.24493656981795753,0.8192953258512878,0.8743578494386792,0.6761514432764981,
 0.46365430612142355,0.3131669498468054,0.9253224682434181,
 0.15298499702142065,0.13890669399716193,0.1150673766725504,
 0.7058955653897429,0.9709992342523801,0.3228219175828594,0.1519717564541606,
 0.6114772216268592,0.5288215104433264,0.5463225952346826,0.2279822703809371,
 0.7381408781062295,0.32423328041546473,0.845639742135271,0.16408188246855304,
 0.26769960882339583,0.23691283445773514,0.7390434317407104,
 0.5550371295580453,0.7089010129968862,0.1214132687626075,0.8746397557241227,
 0.4557670430147801,0.8644384915184976,0.10635433424418872,0.4492035218401229,
 0.6813559618056101,0.9554833771275818,0.6036910848358684,0.4019663855288689,
 0.5239504151675556,0.9273919438878642,0.050528350236964004,0.722462195210837,
 0.6773748709383816,0.43472345500836174,0.2221381964112359,0.7800013304725627,
 0.8742950064680741,0.8777092116427982,0.6403156510933361,0.14905920209455947,
 0.8523040344970569,0.27183748576862155,0.21058628743057906,0.616152249664354,
 0.8194941436242515,0.7184971469052396,0.25980790081141925,0.3896967608490034,
 0.07851947690242733,0.03031965397096703,0.5929873594309135,
 0.8760694365228968,0.6464639661766445,0.6467644439106387,0.928535217469666,
 0.030744823517777298,0.7548466638075321,0.25924610472862564,
 0.25456069497959777,0.5088941665236257,0.6519344823958613,0.8882050828243582,
 0.9377865879349984,0.35898265787081485,0.7944796215461524,0.4627681980518087,
 0.6748939904272306,0.9597952842844067,0.016601881092040083,
 0.5347583808452449,0.30631932710607523,0.30640142154066885,
 0.5674822750953884,0.6772612027089986,0.5132080582645768,0.5074543751240829,
 0.5413356715885367,0.07311038613005783,0.3086951559589536,
 0.06216422374939434,0.4942118358630342,0.923648369541489,0.6588689433662092,
 0.5241485615392023,0.5459159048823429,0.09723261356421053,
 0.18712912587681396,0.9674462989436541,0.1112864481130551,
 0.30579585589597824,0.28442219022516013,0.13403804612843762,
 0.838867969843707,0.9316343498167544,0.8141288529572931,0.5896406240778395,
 0.06445759781183691,0.025951969605990888,0.7312038242508109,
 0.027629535503527425,0.4714445329844623,0.34680784505477047,
 0.10385489459412556,0.9669669715167651,0.4757825182801585,
 0.34753632380282884,0.014482136398334844,0.47278343862151373,
 0.2684450705411542,0.0017144597785699878,0.9766255510422176,
 0.636427249208055,0.5111172063771883,0.9828394168501444,0.865725076842084,
 0.9814256112836272,0.8688225360395219,0.9829059150453143,0.8215985582461045,
 0.1102384196427405,0.8004801964085836,0.049338561599700625,
 0.8318980570383189,0.6423152405857415,0.7219066380693138,0.7698678320746756,
 0.672178705024246,0.5802806346540416,0.06858860312431103,0.4105187470476015,
 0.05801233946219875,0.6121363657404875,0.3519370923266515,0.3806598679901133,
 0.75106469652559,0.44050787563360017,0.8527147665328885,0.6706291449183528,
 0.7779535590440813,0.7336806744286619,0.44080619557400635,0.7121710078195214,
 0.6097843767120827,0.8817434563684381,0.6083448496913159,0.49646513818198634,
 0.41979449561667326,0.031208849392368787,0.605781422598612,
 0.17615023386759776,0.7222930442993757,0.5714314137577654,0.9855843406733662,
 0.7785987788376281,0.36026164113060966,0.5070459181789362,0.6192026464078921,
 0.49883345048271943,0.19495500807298694,0.7171941492822147,
 0.7963142951532732,0.08267756134872206,0.5627033444980152,0.846635590521712,
 0.13328564191978576,0.29469851395793345,0.9362946576201995,
 0.8692001596660771,0.430862932568851,0.438293602256685,0.04464856404642936,
 0.5491814179834277,0.6948122090012993,0.589941682812373,0.6338078940825514,
 0.5838882152122563,0.20872801915834927,0.4076545387597985,0.8578277383855446,
 0.11415854754359767,0.9080025883638374,0.8287173071781887,0.5248722914889374,
 0.29946732352466576,0.5435938851552472,0.9054185234812251,0.5313454813267247,
 0.7640668230090558,0.12803096317297313,0.7223262338094205,0.3187685010518615,
 0.2717658980573553,0.11619690666779714,0.10097762884564176,
 0.6014619770016969,0.7549705681218063,0.766037460949778,0.560628917269086,
 0.9548004810470803,0.18309814226812093,0.051698227673374175,
 0.35541337737373047,0.1526945517804572,0.9303039731449306,0.8893219016537536,
 0.16547759612653112,0.048798806290938246,0.647274159525563,
 0.19285187438752838,0.31460322393347917,0.3520858873424736,
 0.6142423055682239,0.7410260436299313,0.6886420416001982,0.3748379500771294,
 0.9938453120319732,0.482632266550725,0.17295723812665476,0.7650198411203365,
 0.371305152748552,0.008808434605214965,0.865199406315969,0.6211588848724845,
 0.3726344047929664,0.19568091919666197,0.5083727161133228,0.7791253269058162,
 0.5692109296561131,0.948672764912459,0.2696127463934245,0.8294720400105149,
 0.31827484514798454,0.10234798114026455,0.21871558777505506,
 0.38973936492734684,0.06206892574827638,0.9019429661253066,
 0.04931345854877056,0.9899287552340225,0.5136115884724967,0.9185043102581201,
 0.4991117567400841,0.2450283411749714,0.028514452857351102,
 0.5515452917539678,0.9392198157527888,0.1243917588371185,0.9492340865645363,
 0.8498503078351678,0.26808511682495995,0.9535086779035185,0.7441536850840278,
 0.9611182512976597,0.18527819978480675,0.828441297912766,0.4616793071250075,
 0.8709015702772003,0.8339554563034642,0.039257453151096966,
 0.5049744821983266,0.5905502816244458,0.8212940380264053,0.8473503613441797,
 0.9809826365082392,0.7007715297419144,0.08728543060053218,0.7164572383211387,
 0.36965417237665177,0.6718804255110842,0.6425610091253418,0.4600571640854484,
 0.2072385419402052,0.9201501463453562,0.33167458431918206,0.8906992308379135,
 0.4360723127297693,0.39747621605920447,0.22573720557822063,
 0.9099051953194881,0.5503848014503636,0.4139005144685537,0.03762816152812254,
 0.08900099341089596,0.20266044590780907,0.5893802317053737,
 0.9257166900960175,0.053791721274869086,0.22070968912564837,
 0.6433927313684451,0.6154552771755671,0.7639324629016557,0.16823635954772176,
 0.3744963733661355,0.8610271820720325,0.18052081862575753,
 0.08949714239827644,0.9298438127979471,0.28616870408792916,
 0.09777078983128673,0.304418059783645,0.7395644391615345,0.7807026780011863,
 0.5957688509594335,0.00398857847385381,0.5158674259766585,
 0.14755179388587547,0.8403970659925283,0.6338826159885553,
 0.08776863874747753,0.2332334202487033,0.4170017736731205,0.9331760963228758,
 0.38250176238278255,0.08117564559395285,0.9205286950985192,
 0.2346927488165771,0.2886107135846181,0.5562209021128914,0.8723296724116949,
 0.29156536634704544,0.5195022192823083,0.1789315881095188,0.8976462518190305,
 0.572498801549446,0.5911563618586774,0.9800364195755457,0.10707776830901017,
 0.5830514506267243,0.6111307352910651,0.39774184467314444,0.8454034624939365,
 0.45764948110599035,0.5720774959837319,0.3407262471403777,
 0.29205013830814286,0.10137052550486936,0.3277439633041078,
 0.49762858786825037,0.12074836101517805,0.8938303682765154,
 0.2783414640098888,0.3527982908226559,0.5332138185374482,0.6399846664935884,
 0.9984596257817754,0.7943690966102951,0.2879648597962192,0.05859591289959476,
 0.24110111782828203,0.8310382030488558,0.23051467359714783,
 0.24511132833415072,0.12357771939604556,0.2697751055311254,
 0.025947168464129367,0.6880499766621853,0.16890503678544755,
 0.07279878063466971,0.5969828981032801,0.5712562505770952,0.2519310367362805,
 0.6459071312864801,0.14615696477512485,0.6017434589493983,0.4188872479725123,
 0.05136260770482304,0.23634772371092727,0.10376567836030348,
 0.7132098604898081,0.03635280573875577,0.007733686513886506,
 0.07010638667523572,0.0534274763573781,0.6400722825203241,0.5075871851924619,
 0.673759081891811,0.18936728274594605,0.583252869321883,0.7543710373432352,
 0.7135949123223814,0.12547489523760225,0.8379036790202139,
 0.0017881508163172244,0.8291839622349274,0.08754327717336008,
 0.7311829762803885,0.9575307361141554,0.18676567196711658,
 0.08217060912916085,0.7865266220092938,0.037976960852228125,
 0.9016657265412937,0.4975812342833881,0.5310117613376055,0.590340840788166,
 0.3785092144484066,0.2894109912767956,0.9113807203952592,0.14313437139385532,
 0.4551251244451293,0.33179110987313465,0.4184649169667335,0.0761734715971033,
 0.5937238572889953,0.12504752928769503,0.30958000108686734,
 0.7562377363974503,0.26889635455022165,0.8167638302661859,0.664772205520632,
 0.7542001591069909,0.9604262618280215,0.31989011757460917,
 0.17219330323987092,0.4259059608235152,0.7161428629033575,
 0.45078307935598527,0.6642247613646715,0.3094999515828766,0.7680380697758271,
 0.4596337418570928,0.4498977749227243,0.711268810123369,0.05952589101153127,
 0.5499619027907516,0.6476545609137964,0.26231401988893754,
 0.36085649255343055,0.7637457505604383,0.25571881737233615,
 0.40241672500357606,0.7089121926680809,0.17117720343947274,
 0.6157264711385431,0.9499650664836685,0.0260033450435182,0.40765745135088016,
 0.9849841479104827,0.09406973843023003,0.9588737292582867,0.5302130430495886,
 0.7671506823509229,0.8052122349879567,0.25297668431556286,
 0.18860177395310807,0.7864052860909622,0.9987471704511167,0.6118680723096763,
 0.12721578247638277,0.23797363523469384,0.40116361574876847,
 0.1460670230498351,0.2797078417311306,0.47152333588673123,0.7707915695531007,
 0.07361777150196047,0.23039033727348335,0.5676747344377133,
 0.5038153511258224,0.11523026176484297,0.4584702030897445,
 0.43032619815994144,0.04871436036564647,0.34384280869266914,
 0.9819762420939594,0.5600309009709328,0.5975572217163613,0.8371344806024678,
 0.14200921474100958,0.09090672472395345,0.3326627580892485,
 0.06327516582005932,0.8089867957097397,0.39230956840039466,
 0.7332955845682377,0.15212356984049435,0.7616984393462984,0.5896400328091828,
 0.11512594240112317,0.29703396486726796,0.9172133672120181,
 0.03243930761753999,0.814505948725038,0.18573929933695765,0.6886222441145673,
 0.7887210052960688,0.8292833760042233,0.015664008258612028,0.949025222163319,
 0.2908804624587915,0.36494010374131614,0.3638592825371987,0.72988133634882,
 0.44037768189834403,0.08152680991495243,0.37940483298699024,
 0.42494381533169545,0.5517325748621917,0.33688593954717183,
 0.8124992203111818,0.2773307840454423,0.6163606593795993,0.7430532801948107,
 0.48766617547859115,0.37339406378262585,0.8664653144673371,
 0.29753861857364605,0.5606032627752268,0.6812163560051459,0.2326040575570093,
 0.10270330746911449,0.6647353470092061,0.9916396476389068,
 0.08832608442268042,0.2329507360560571,0.26679958250099267,
 0.7399695716015864,0.8729292407093515,0.862613555392511,0.6875639938385266,
 0.16424548943644768,0.5008581752743662,0.23154525405736337,
 0.11428563006228354,0.7900645616555534,0.30507954605946463,
 0.6307527758411952,0.6874890058227348,0.7815663687913446,0.4934780785813,
 0.12713133702880786,0.6161388459500197,0.5820630672276194,
 0.28079629836373454,0.24165581571987638,0.7743724345860783,
 0.4896589437035236,0.5525993132136481,0.34443334109774826,0.8864638458807571,
 0.2435783310010634,0.17612405977385315,0.5982550477113836,0.1928465698055517,
 0.280332466928808,0.019996293871167414,0.608445270616913,0.1044924363715467,
 0.08708358748061673,0.9988124325280925,0.352405709141405,0.50279478531017,
 0.5606167835003131,0.43441634313327515,0.2173196661251655,0.655497767903873,
 0.7669994918215285,0.5743085249421677,0.3432299882777039,0.306203565174231,
 0.6674345712790244,0.28501117496981176,0.8975310899117331,0.6397191201420102,
 0.7133265076963251,0.1837964238280272,0.0648865652036773,0.12116646237404083,
 0.6853070713175411,0.7800621951908644,0.8676001363396842,
 0.009546146693207747,0.1100362408315454,0.08084606864917776,
 0.4262181889056844,0.2700149461619955,0.011620741615161467,0.861989984517518,
 0.46605719207539775,0.7935965340251998,0.41221501124872995,
 0.5979978233889084,0.32801133848955555,0.6000019141476811,0.5095363944544298,
 0.5411477420477532,0.2570341012239512,0.08504444825679569,
 0.35980029738481245,0.39554124105390154,0.8083693256185429,
 0.043325227186355675,0.7807418305358421,0.24844312454769657,
 0.15979342612167902,0.09802390341703004,0.6387724461510094,
 0.6249827133241663,0.035727438441434556,0.918647364964524,0.7358282517338908,
 0.23177209391677145,0.020800277655247945,0.6393298197054416,
 0.7590543465636677,0.9118585149554361,0.1799706597595555,0.7269397636151869,
 0.9575434010322117,0.3399611713035213,0.9713191968826288,0.5839615983764336,
 0.4426589776805373,0.4655980435586964,0.021707775027892895,
 0.2762050433059482,0.5018931257818805,0.49438457696712845,0.3911212987027499,
 0.592813411671544,0.8398177076614034,0.12402182849610055,0.5599578961371648,
 0.910539320171859,0.698718269177751,0.7738752375924554,0.3976444425695139,
 0.14058172720592976,0.9370529621253766,0.36602479914761343,
 0.04321843223380872,0.5618257394908616,0.8526386574721244,
 0.17991236119341492,0.20553468877721504,0.2714465025149001,
 0.2957967856221061,0.9251862921050188,0.9955086511318976,0.4816280665723214,
 0.9334730708967932,0.3730043917061916,0.5840641667916473,0.8419135071826661,
 0.9526703719303611,0.4489678160750973,0.2018707227666836,0.12862341425340862,
 0.1316222323675248,0.5172841913916831,0.602439119108831,0.1667660850476954,
 0.4385534997634244,0.6331224647825265,0.047388954866636745,
 0.13861996232703422,0.5830853534807737,0.5440072594408266,
 0.08800741018704594,0.4552736273122724,0.011998073859681435,
 0.6671399044939565,0.16085487614547544,0.13876439686408704,
 0.8853678480918884,0.32407141471712997,0.010969929409816892,
 0.013846167640188645,0.19101499035977576,0.9462050439407831,
 0.7454388191016132,0.8801834758825212,0.32800482290511634,
 0.05558582218026753,0.12444449756897114,0.036588867555131466,
 0.7202186637955867,0.11070155216925104,0.5919992411506012,0.9671111219955683,
 0.042477489504368804,0.013395174111129893,0.8033813050442428,
 0.8812707000923078,0.022646499825526578,0.6368213749143411,
 0.5937968822907224,0.1270821439806843,0.7278358263351176,0.7144741913959692,
 0.07748868323206703,0.9105832916109032,0.7167637962990413,0.6864100082054225,
 0.4044228381140389,0.16232125535189368,0.7968134853899432,0.8705768282332835,
 0.7655664791284996,0.8688997427953383,0.7599474977631573,0.9678285564061306,
 0.5586541356386365,0.1554722144183861,0.13481604308336048,
 0.22254472716250018,0.5703525464196767,0.9409929487062098,0.5098498158982431,
 0.18319235018081326,0.7538295808260206,0.9972519031608349,0.5255963441326426,
 0.26097458309899946,0.47853349351571783,0.956151941492974,0.7946156531214833,
 0.7821350691736698,0.25738218084451936,0.9656575497429761,0.6677850015834604,
 0.08706537529069591,0.8716813115444917,0.47867563465887997,
 0.47983669644981564,0.19529320388530136,0.881360929143926,0.7582484304287023,
 0.25271827874287833,0.7188885402836523,0.5207675825726885,0.7035731430011852,
 0.2946119743505248,0.009713105041173287,0.6064929165373755,
 0.41418106403040233,0.026852834818030913,0.7556838165518339,
 0.9826309284780825,0.40706920534985613,0.552280666086577,0.3930451180268708,
 0.4780691582126497,0.5953754243508959,0.10200903809121165,
 0.01452530219085546,0.3067507026042855,0.5382480232389817,0.7412383893891381,
 0.12817310183319452,0.16399037011134365,0.20865803089599433,
 0.8803651938268342,0.5249236951524863,0.9275829667755235,0.19673309141020423,
 0.6270988068500398,0.531869845579664,0.47113737298769753,0.5425044285602866,
 0.3988022026800111,0.015136403209448424,0.8086454915882451,
 0.9432839906632514,0.010571571566904137,0.9339960013187759,
 0.39356419698464706,0.7255601738917338,0.4346856242028234,
 0.07726697072748123,0.1379735837496694,0.815282925329505,
 0.0026987186193904167,0.7975161146277324,0.3464523869820295,
 0.1516719155958387,0.6395401504515867,0.7760081318299613,
 0.056923026646586106,0.8311042407392913,0.9061077029678689,
 0.6762389270244815,0.4988283300747348,0.18722501924000423,0.9860582175361495,
 0.23152806113972269,0.8133818301700858,0.6892788927023179,0.9665555566601827,
 0.5651183054179276,0.643583965966414,0.44246891377909126,0.4661259464110309,
 0.12805996139909415,0.7231669954276808,0.35355469138578366,
 0.8637280998312648,0.5931317205742158,0.5304734411478016,0.4045914511162,
 0.4926008695105668,0.5791756860992333,0.8067571949582959,0.9077530575660968,
 0.6395152270417004,0.38060353975815087,0.16652759375836768,
 0.4194336160455565,0.346449983280518,0.870235489962315,0.5864128965473981,
 0.6558589389939558,0.4456950388108324,0.16417088242995792,
 0.47571282055109254,0.898901280631405,0.5152150624072225,0.806563694351234,
 0.9944024070698174,0.7901882458997823,0.4317716685964367,0.4457718397591468,
 0.30705894453982296,0.8071445597718296,0.07929031780658125,
 0.7550460493158995,0.4244607408883966,0.3641282486334091,0.3450873634138404,
 0.14337760721377513,0.9775684185763516,0.6235602461412528,
 0.33376580720000804,0.7942972668705885,0.1771572164516041,0.5084467930837944,
 0.6057102159918545,0.18087700955745534,0.9991630078673697,0.6261692660713593,
 0.08172159826762915,0.5988685433428086,0.5358721313583297,
 0.13308422420800103,0.3637481113757781,0.463324672860034,0.9031944987567609,
 0.7919396104796099,0.755210205914824,0.6004882156578828,0.9125843284140751,
 0.5344954094656039,0.27893037278477995,0.15776210174063854,
 0.9623498550054999,0.3609321365527375,0.051881259584126815,
 0.006768015686153039,0.5774677334820072,0.4355068102967682,
 0.1933511434053249,0.9969727604259588,0.5038246752673253,0.6230336483649295,
 0.2642004575113499,0.6691735191018361,0.46528730328839385,0.6699579084752858,
 0.9247829887511786,0.9257684136458092,0.7173022344225953,0.2766955498428304,
 0.8158281956701833,0.8544349117492346,0.8509548438969635,0.6824985467651559,
 0.23040186828863385,0.21481638296411398,0.08938549908933446,
 0.06600975505412299,0.7145941370972496,0.44924412424002336,
 0.5244380774987458,0.8403456479313166,0.3481356344238752,0.45736143978611543,
 0.9813334361556304,0.8847970025643468,0.5906048924155277,0.15032082556697324,
 0.5733648281773039,0.8259834262535736,0.8493476605202847,0.6999159091908781,
 0.49585579874149055,0.8802400261844293,0.5017653250964845,0.3932122127897695,
 0.0739813546883421,0.28052756230080345,0.4427502171498231,0.6728879921643628,
 0.3898163482387096,0.15100313590095937,0.24471945717433652,
 0.6006160300799331,0.1614723398737793,0.5069638256169874,0.7348653880998139,
 0.22186680115544977,0.6349615749629985,0.31668168402911834,
 0.27076188594366013,0.9885000350983293,0.7537828351567232,
 0.42308627487250106,0.7122030254585819,0.2361955070627868,
 0.13120047683405822,0.37728890641610024,0.9844763008593542,
 0.21302348506823898,0.23363799576742672,0.3254024791625201,
 0.9413599050932716,0.780855143207608,0.7414355031950203,0.7897771360217022,
 0.8271246431528299,0.0015354483546309305,0.6518609414921627,
 0.8095620768731554,0.8585727132057298,0.5707550702491806,0.9539603296252641,
 0.4935357893522946,0.8207204234276197,0.39193663349505836,
 0.11480753550363954,0.959123173212243,0.6710115974181871,0.15720951432521513,
 0.8159371066383644,0.8801662963297157,0.3754225918528642,0.9045623776317882,
 0.7137917519620985,0.19106661100318012,0.082473905316507,0.7173500249670314,
 0.3908793357211575,0.6053431105358918,0.7857205099868131,0.8629686509585345,
 0.47281297716421533,0.3561683916354087,0.6413189270416868,0.7772911034401468,
 0.2693034216193456,0.6167766591472644,0.5051078093795178,0.750460538475424,
 0.186399448833382,0.03944621903275869,0.30811594896001715,0.4487848129896468,
 0.7987875009762804,0.37629312633510204,0.1224871118142229,
 0.023145456756030858,0.6828755639487042,0.041867605154141785,
 0.25186515772091767,0.4519711386313324,0.5962573240382177,
 0.22986043754731744,0.9035281685183463,0.1757932411749804,0.689727717128415,
 0.95471013268006,0.369676360723035,0.3959048588327193,0.9490526119822719,
 0.5428201423870286,0.8728337363548385,0.9044058924369411,0.8529429433745426,
 0.07470064524454867,0.9177839188424627,0.6097679180457575,
 0.17063314083922543,0.435224274093994,0.9244781640484487,0.27044182639139736,
 0.6926176851080031,0.2560085061536369,0.1291566553968153,0.21182850966754563,
 0.463464360594672,0.4459311067659103,0.3843563499386945,0.8441818307960574,
 0.8900182517154821,0.4695531151397696,0.71174958037158,0.3513120907006524,
 0.8800439427693014,0.16282037295698493,0.752656211415633,0.1699732563092703,
 0.23574804979276232,0.4655072064495812,0.7248013234878652,0.8790081505622465,
 0.913888815520738,0.12404765434606979,0.8645316686838554,0.7054478261347823,
 0.4119747771647124,0.8272586812031124,0.7016425599982011,0.3850594167464263,
 0.1630790798909012,0.1088488915633069,0.389153163369504,0.7736386499642944,
 0.24982342026845983,0.16602910884086408,0.8363173629756504,
 0.8490171397656483,0.5430319764028241,0.4309837214485499,0.2231769742458427,
 0.024225713376352043,0.0054020868801401,0.10460471303617858,
 0.9089876075456582,0.3377443333447575,0.989349289823252,0.5504606189413191,
 0.05300179863355092,0.334270090917943,0.9481277026111548,0.6373338972080816,
 0.6135612340204055,0.11371125601657006,0.9491003003322167,0.6035660878282085,
 0.48094774736592805,0.14164740371476703,0.4298788982011894,
 0.4712314014435314,0.7506473838814975,0.8300012348389102,0.7683947550464779,
 0.9695788688921897,0.42400549860967207,0.3568409900174231,0.3109582588608196,
 0.7534753199233898,0.8789460273250108,0.5198986474975995,0.14298642894920055,
 0.9495535698693591,0.6805734861462298,0.5875037200735299,0.4153531886975156,
 0.725902147016572,0.4287758841430185,0.6280810378604834,0.4866058981680401,
 0.23910297143257042,0.0025758162964467646,0.4839445585251927,
 0.4783800812262461,0.8083962162862122,0.5924928714193072,0.17546546672085528,
 0.23580048284724464,0.30025837377295983,0.7794149756031823,
 0.24435812902549436,0.12769289561002606,0.32521722272407483,
 0.4689190580493562,0.8340604434967254,0.4109168229158352,
 0.021524604775125633,0.43268985331669885,0.41517464876679644,
 0.1091011934735363,0.22482058429494445,0.5708812844800882,0.4314620486332519,
 0.13239084534075862,0.351085614264809,0.08568278395824169,0.6960470637796784,
 0.5133361333508424,0.9452194849086459,0.6431807361187923,0.7592434241801279,
 0.8661900672383012,0.21061538399483892,0.8405087383742076,
 0.14015742225936967,0.8905149279205761,0.9174169370496174,0.7889140740262949,
 0.8652506134314709,0.2979483134668849,0.49209784739866513,0.7628856471257203,
 0.3268787161770922,0.7239172808363248,0.5176005797501222,0.7455116326183633,
 0.7082032865362065,0.6458803426780164,0.42442191478945634,0.7700474113801613,
 0.5336166114314614,0.3867271914461585,0.7883350069293299,0.1509320932014936,
 0.1616941012935158,0.14213860781151055,0.9063432822403328,0.5757602489451277,
 0.22759702776373325,0.6535630106891673,0.36771102892170127,
 0.46803782778007086,0.07488079920413915,0.7180159210124412,
 0.06661105085693053,0.28822833562198324,0.9169768329943608,
 0.33678143232168223,0.29549945224273,0.3510792586299767,0.8162078430901241,
 0.6050464471447419,0.868364516415711,0.6172078904894834,0.8302304843541025,
 0.4948861238023057,0.5424206472861431,0.36735249597507313,0.3261160832671447,
 0.061412953241802515,0.8023505677838464,0.42681496933220286,
 0.2770896242892701,0.539257907209742,0.7210323537004316,0.2980651163764332,
 0.5795740952276225,0.7304472237953231,0.3154991586015681,0.5023259247442117,
 0.18221428904164594,0.2696758816109429,0.7468454083551305,
 0.35991395805124426,0.5515523000471272,0.7998705393267409,
 0.04112758844529696,0.19951013653802796,0.6828347942100513,
 0.3421353517219332,0.5120027719223857,0.28618042488147,0.3540103685004192,
 0.1965910458026513,0.9400910500626154,0.6937836434014462,0.8582779916008009,
 0.24703052751507049,0.5600720369696461,0.8712913373870698,0.6624206044543461,
 0.9632332186365957,0.37341530148827407,0.9140107964397801,0.279565793775582,
 0.9763120770933782,0.9691601664939565,0.6648235949213206,0.7979138881902558,
 0.43437501800532996,0.3809165042332042,0.34925281456555746,
 0.8241302560336253,0.1759319824841793,0.138759064897233,0.07632872530960189,
 0.6467222988924202,0.8336831006055453,0.18228627805361508,0.5650009924188806,
 0.30697361522226596,0.42168587398487767,0.2641008432290497,
 0.7589158783940553,0.628705790643001,0.02730726468423539,
 9.326090714589785e-4,0.8834646176925397,0.0961928833811263,
 0.9827907900364824,0.06779150038990134,0.9444024411302606,0.5852984654236707,
 0.7160341719236489,0.7100486045265257,0.6511628937919878,0.36714023044525745,
 0.7806903789242154,0.18157125287136233,0.48618302455986684,
 0.7560251317381818,0.8437068633797773,0.7197901054194582,0.738362150305138,
 0.9323179281982448,0.13133094221647557,0.6262166001094842,
 0.11553410849071866,0.10979124571375454,0.5221015031304019,
 0.7140592909123008,0.6437752224341836,0.02625300659767338,0.6028342948895413,
 0.2954487077938761,0.9992381645200412,0.26440784515754656,
 0.27489944663446186,0.3550065702336901,0.7892182150973904,
 0.09094680403236566,0.6993329509300383,0.17696000908321896,
 0.5810430000377179,0.07610034287520484,0.22504357513017925,
 0.6400703238876783,0.8548252040943739,0.5903893366822324,0.39537795703476075,
 0.5257183748302945,0.9229496387293947,0.9858282936612589,0.4917673050975788,
 0.34577527786790363,0.9575077804569448,0.8216471859925818,0.6362048629716865,
 0.3886731189188388,0.4923755237004839,0.7766340621267334,0.05061299518064988,
 0.36355215131564445,0.08663259041923133,0.43870220321877307,
 0.4057283739684543,0.254383397561298,0.028301091959205604,0.2623929171015328,
 0.48415112862260734,0.27400821360454886,0.02612916181076097,
 0.24914977929404192,0.18167897319102366,0.4534703917094104,
 0.2888224448199681,0.3314633501063925,0.9147205626998274,0.6653815898998846,
 0.08905271467183562,0.7373027168910188,0.4613704718032131,0.2643629538051545,
 0.7285998282224271,0.27520129818100003,0.6095144733945722,
 0.29885269263626846,0.43805180023494705,0.5067358276492683,
 0.21060472316660006,0.5070048704993733,0.4202183540918568,0.8079609996749747,
 0.5819504477335309,0.10021249941108512,0.08343590778656784,0.118943099135986,
 0.19938390525002536,0.21683969948345083,0.05778040597143663,
 0.48617974773348727,0.8764002843540296,0.177334223225729,0.32031149619533117,
 0.8950626286609173,0.21181828226175137,0.6060576517433959,0.8882465054642711,
 0.07630168952768379,0.47022010089457167,0.6283549297417427,
 0.3655237088398253,0.2378535863512986,0.07734809378713325,
 0.22415054307908677,0.5565578019345496,0.9963210514098745,0.5136135370109782,
 0.6563762736222283,0.01023243242779781,0.9915184467898545,0.5006137378017202,
 0.6605822104584836,0.9193316403974439,0.1026516899920773,0.4602234108660015,
 0.2622113261922654,0.8711151127814676,0.5085153723259483,
 0.037017047564212646,0.24319960347692515,0.030937568507245494,
 0.9637096467852364,0.7920750430293335,0.6790403173647148,0.9733036087692093,
 0.2760981496453001,0.9225900788774384,0.6728228668127478,0.5745851364369796,
 0.6285569960923825,0.291037181766574,0.4260051717105836,0.935212457465273,
 0.27269775503199556,0.48019322731224057,0.9862893256366517,
 0.5992446021628117,0.6022757521794617,0.37028696060990607,
 0.36148747593899055,0.4784856204404535,0.5846317259083785,0.9177613296885627,
 0.08395503057936626,0.7369258622735345,0.6315034022851619,
 0.26014747626375734,0.7331220330569453,0.996112232204309,0.6371632520546655,
 0.800171325977881,0.7561224658280573,0.8099814973979126,0.3247219998984656,
 0.8997959809316918,0.5692250675628339,0.08713320829053717,0.9011123826621034,
 0.46714775507186,0.6416922268018025,0.42175674967987864,0.3272282779804818,
 0.012862263375761906,0.3255847776531342,0.29212654928452864,
 0.4027593616006113,0.4565993068708828,0.8924977178376254,0.6509499024385274,
 0.4935686772678032,0.4470858830842823,0.16904499820003305,0.7868481487979457,
 0.8572879996397993,0.6384381175534367,0.05219182991081195,0.7455034504550574,
 0.17009080060366177,0.28984467658576285,0.6198643233605043,
 0.7679359559308757,0.7134871036331504,0.7719293093080826,0.3423128266378389,
 0.9276884608029757,0.539339239453098,0.3465276795910479,0.9000313728367044,
 0.27462479762168635,0.6141963277730697,0.35051753670872,0.28702000025795027,
 0.8596451746140145,0.5337941248147334,0.5926689485818667,0.4459382101639663,
 0.6152219355810202,0.23079527759275797,0.5518305144651913,0.7402407973571071,
 0.6398916704041333,0.9447491992309729,0.9160892672225627,0.8227913455894103,
 0.6335961294847001,0.8513057490103129,0.7307221365998111,0.7422811284374932,
 0.2060431057305948,0.9267662273119717,0.6694699662228698,0.44033548565441427,
 0.4832014763990605,0.8351072502895158,0.04040007711449556,0.9626919253422376,
 0.007741396152729174,0.3312739612093798,0.2522020934262269,
 0.5639897976577184,0.858645249340408,0.17880615404263833,0.9097975368877216,
 0.27142368416637896,0.41663894761136633,0.4157377923648211,
 0.5438124283767874,0.931252062317875,0.31469960629447213,0.257162582285383,
 0.8118397070049275,0.9213449489211997,0.9282475360123902,0.40544308492922265,
 0.3876224341510218,0.24127357263320348,0.09104332240296187,
 0.5497921925667509,0.6845336261553192,0.8943004895863762,0.04698137425016169,
 0.2164638479899681,0.6696731588258534,0.30818294813161307,0.5185589154247534,
 0.6163045942872497,0.8832228023936479,0.45550010041135314,0.9598843897233038,
 0.12058823093039384,0.080321609540569,0.33855491469597343,0.4967334361286585,
 0.47202029573973014,0.034684063356422934,0.010162879302466177,
 0.3247327526214505,0.1637819860513342,0.3650667840529752,0.07474381142356268,
 0.8841884894120684,0.6292625833428971,0.8675331482972047,0.6458279007573057,
 0.11299686338733961,0.020236830104122805,0.5789906978820362,
 0.15950113235200902,0.5861529221435061,0.9821754460695666,0.8725024400846606,
 0.7419534184528268,0.8552708758251435,0.6722747548456084,0.2223939774099034,
 0.7068276348205176,0.41253622303377646,0.9163271194055662,
 0.10783320708887623,0.1845477013142779,0.4480229978097957,0.9250173399040081,
 0.8357978972228765,0.09773070232149739,0.28248935020715593,
 0.3575588343263062,0.08258278287031173,0.5748729704319576,
 0.43382552027425303,0.8955429225713759,0.11036998274378795,
 0.47574208402164575,0.5867882118060062,0.03905109611425428,
 0.22325603302582275,0.8394173322607499,0.5544120321732676,
 0.21372730279214391,0.27574428496373593,0.5629100128246198,
 0.47478208661440235,0.5680158974009906,0.07373203299976416,
 0.6521787265410907,0.9295570228323349,0.7525172593676235,0.49018115634942216,
 0.10487518007606333,0.7896810578440092,0.6932203046269584,
 0.35471142995715244,0.7133631295200566,0.4373037770341629,0.8557255191025573,
 0.05263143292991457,0.9150559675171697,0.6037338575570814,0.3708218425832879,
 0.41546619921086514,0.7652435527268647,0.36410063635876677,
 0.5645272794585341,0.2904977087671414,0.7767038033656517,0.39257705579945434,
 0.6340363271329531,0.9154368049544606,0.058782895769357135,
 0.23965070717293546,0.3807504312266492,0.8131062641441066,
 0.33537269664773217,0.7760894540576503,0.33464555168477006,
 0.6683890040014606,0.4287745451249185,0.3954694946863361,0.4232415753372136,
 0.7302255795925496,0.4583738639603664,0.5507249096112724,0.42519219509922235,
 0.08329344227070423,0.5976822132187265,0.5688810528427443,0.2258043066409916,
 0.8454141678208698,0.16502706444048465,0.7074524572726801,0.9276562651522147,
 0.24367665809642813,0.36188969512275626,0.33233152458610427,
 0.36865866774110156,0.7668342350723844,0.8447396357728724,0.3558286546621593,
 0.34134639841913117,0.996238466736178,0.2388669206253844,0.26638662417408776,
 0.2726297915075159,0.9876492489868972,0.07037126784688463,0.9735406651074546,
 0.03596981372992247,0.39261220377325023,0.5349634065977948,
 0.45085949437466155,0.8646656810216756,0.02655698256367256,
 0.4347793319749511,0.25397840144712747,0.7232020075586885,0.8498253947498808,
 0.20878868347192436,0.46395347487300487,0.9745969664945053,
 0.3123916058346332,0.7458640020895512,0.5367334332737163,0.7572221945474823,
 0.32530051011181427,0.1494785747527534,0.3787310633091113,0.7009556119767488,
 0.8446078892774495,0.6644517636493538,0.11435686279534107,
 0.37878060707054384,0.9156649003860367,0.29114835627956737,
 0.11869500362303798,0.646156875615338,0.6290146965906211,0.48748415000793877,
 0.7871898717596232,0.6169929743000674,0.1953941598152671,0.7564277832701708,
 0.043225761806601204,0.09071899291343843,0.26975580446483083,
 0.5921353716437838,0.21355678831555613,0.48578481437209886,0.305617472571291,
 0.9644724305954974,0.5354363331059783,0.8161735070024723,0.42999121581775723,
 0.5663108782870558,0.8545870450512022,0.05987386539643391,0.6720809812413433,
 0.7126304351071342,0.19901270620206624,0.12908763313657567,
 0.8306938779607432,0.9970107682027343,0.5875471933808991,0.992855143771673,
 0.4861235524303579,0.03739509122484275,0.8791958523166962,0.3976099324391743,
 0.3184639672877191,0.5407434909500041,0.38646118437230736,0.672151623291001,
 0.04414797538050841,0.3257585421713042,0.7240768877741467,0.3214667693077474,
 0.7671312074029181,0.34956884091689777,0.7371364542204831,0.4344044340814581,
 0.8621545672631387,0.372719311453083,0.3506522015103254,0.13510781525009552,
 0.6514962417766405,0.921335671524075,0.010815708700007476,
 0.20248741742125453,0.6098047929448296,0.1778280291270531,
 0.18027299659436313,0.2440690557427765,0.40464140198169496,
 0.7784608182425935,0.8119407983217555,0.6265120339467107,0.7312943995107907,
 0.8209443625893293,0.1540486429457174,0.39592819021882886,
 0.43409018143881317,0.6177728923308854,0.0259948308664133,0.9989760463997589,
 0.9214283252329444,0.13978668310971476,0.7052750309590703,0.4299570236381347,
 0.11964827627018049,0.4681616719172387,0.3739750421943254,0.884473002364396,
 0.9151389127802871,0.6775710358401084,0.13129314199835784,0.4210600889967464,
 0.9653336158408021,0.7282278790694031,0.38078105731843637,
 0.43100083589087856,0.9632445747878511,0.7087643046833142,0.9911439607270865,
 0.8463419078823516,0.11966373764659632,0.6227658473814035,0.5529929307568089,
 0.1523971890103878,0.29094546847531433,0.3929841103534869,0.544828728797789,
 0.522160498324761,0.24476945197138789,0.18508879631942565,
 0.31369953469467027,0.4680525019985149,0.02998102829489735,
 0.8869675983147662,0.9154358900571268,0.108507435930984,0.9456057958687855,
 0.4241968895548148,0.817814256030229,0.6679081204445138,0.884188380678252,
 0.2954041177156025,0.5779102926870666,0.986881423597057,0.6082516543989935,
 0.8714764563109119,0.2988153211319853,0.5945882880792278,0.47444989096241275,
 0.23850611212746897,0.34089335733258697,0.5408369263181484,
 0.4015382614835338,0.09840536165970426,0.7810155395083083,0.6318748135047938,
 0.39964572655728614,0.6177880954290107,0.9435313126213803,
 0.28887871367636864,0.7654098097443998,0.09869862460924328,
 0.17313214162085988,0.005311690781438783,0.8607893096927207,
 0.9107358397120374,0.4986634514356637,0.77003421712514,0.5841965856772557,
 0.6421578351488034,0.6973902723403318,0.23028871881711765,0.807257352094556,
 0.5597625306184041,0.18806730896595503,0.26698254579337355,
 0.2061309894245964,0.7233901021773044,0.5344889511189511,0.7949553517298665,
 0.9619566508460143,0.57977170746308,0.23345966013973674,0.3351470273801632,
 0.5907133107274384,0.8116391319193252,0.3992654141275319,0.9472993571388367,
 0.6035783586322138,0.8556866588919251,0.5957776617044184,0.01704930674520322,
 0.16136210612959712,0.28552947749727686,0.03245534923213811,
 0.7669102247830264,0.0069010573540240205,0.4480441793294395,
 0.06567428388146823,0.35123320923649026,0.6905042883616997,
 0.3990664918929875,0.5973487256227314,0.6882987858483116,0.8770334719873488,
 0.06835821159147737,0.47801261213646384,0.2200444008740956,
 0.5205408758959091,0.5907718033718364,0.8186121725077963,0.07520355694378211,
 0.9007872643904619,0.3619479099755061,0.8715458968817023,
 0.037038745101008796,0.37466547880085777,0.4244770228403406,
 0.38395464944611246,0.26158548076049437,0.5190417887673108,
 0.37533052056654836,0.3719880737790182,0.5838348522856197,0.6286890823632815,
 0.23833578028390157,0.35360081049918723,0.01863848080268038,
 0.6201568224741442,0.6010051051198851,0.7298687114950023,0.1478545387611362,
 0.9218986662421167,0.5789112545574332,0.17942252002999448,0.8020468712712265,
 0.7612536675802484,0.46836628033727545,0.1949714890463341,0.0862581905754084,
 0.7759419205834915,0.5848544318677873,0.07245948188611417,
 0.21767202670075458,0.5811587480459034,0.23484006719137085,
 0.6022203399833603,0.789496899466851,0.293860517091625,0.40139294912313717,
 0.39043644604446837,0.32730681596638334,0.7383871498130167,
 0.37977953275471554,0.6415306932621244,0.8481848018559482,
 0.05097194027466201,0.884059619194704,0.9812326296319714,0.8284478786311513,
 5.86402917364004e-4,0.9359661759216247,0.5295347827945981,0.8399657172070619,
 0.20670950355628592,0.5736950869712734,0.7599435387973832,
 0.08974840962374742,0.4721598521091558,0.6679257497555116,0.7158726373125773,
 0.6273131375935165,0.860622821426676,0.45427028485200327,0.8194562876480914,
 0.676303928862803,0.9379533483218023,0.10042204787503095,0.12531612278956517,
 0.3775832240877357,0.7932414929372348,0.4199268195668717,0.34338172934792344,
 0.3450433947036805,0.18394816030748318,8.920529690237267e-4,
 0.8770787171494636,0.9376850629726879,0.5551624934028797,0.13838783214797246,
 0.36051347584592086,0.13164200280617533,0.23500416778704192,
 0.402059526621127,0.18438103589471722,7.366151668489751e-4,
 0.4791057570349616,0.29260917860540825,0.27991284350851664,
 0.6332074030452086,0.5910193779881254,0.7008207965190616,0.8224365998092118,
 0.48451981549953116,0.8921583955564131,0.018040712010686133,
 0.9430174422488966,0.20425642997262394,0.5395565782520866,
 0.14614512866649798,0.9466241307960841,0.317478250965454,0.01132651679983232,
 0.7054472094576801,0.35778591962999085,0.8939585316069808,
 0.29714879298317165,0.011807292032898209,0.2823746337312796,
 0.6482138872711618,0.16069221873569162,0.30887888874980785,
 0.16824766494336096,0.7960666806810889,0.14758736430468777,
 0.3674817036623135,0.5622592113963241,0.8754307784607347,0.24105473961401414,
 0.024405871905628262,0.7525430793950403,0.1835983897460005,
 0.1802122128750756,0.2802014516592267,0.828624294842638,0.5784379911231751,
 0.28778908544434,0.8160617350813577,0.09096826333689045,0.7276086448447678,
 0.7595642240396736,0.28283593651501315,0.2837943380724366,0.7783962250947224,
 0.23017495553842682,0.9174160480075632,0.5442988961240998,0.5121194414870164,
 0.8025379168611072,0.004278917190947973,0.2692983955802468,
 0.3197396703078378,0.7598869824398001,0.5822413099668203,0.40521809512747897,
 0.28810240304172374,0.9433448258504258,0.15613723907342947,
 0.39754408568107513,0.30697135588579605,0.29577964990407113,
 0.02375064555590134,0.24015332041944892,0.8664863873013876,
 0.3008851584343595,0.05173584619149618,0.7881034360793757,0.6193382070526419,
 0.3051241158512116,0.6846047362227331,0.02797318983943986,
 0.17357545825026932,0.04680328060979233,0.7470446480693802,
 0.04844597697854347,0.5798766922227583,0.24953843380613439,
 0.9931099948194096,0.8570384848558881,0.30186441536280584,0.1425006041258482,
 0.12434435992149906,0.44501631386928653,0.31238537114589204,
 0.22794489883328073,0.6415870514969367,0.009769162857474867,
 0.1795106502557029,0.19159060933840522,0.3290336574967676,0.1573496470190855,
 0.16636936871242447,0.7037037058296421,0.040443047511039,0.786410544523813,
 0.34489990542964877,0.3520375065158895,0.4151150127741253,
 0.35984597806061913,0.4361433380280524,0.41913474602328205,
 0.3958697268276522,0.8420084099399897,0.1827956556738023,0.3613850261030439,
 0.248576226357901,0.0014213817384848282,0.1466772820135296,
 0.43536989413760785,0.701331680648007,0.9093115453630487,0.5519541227366161,
 0.28004756313953594,0.8999651676446556,0.911835138081792,0.7375521933912625,
 0.5186705175216387,0.9031790734405282,0.1563645453529494,0.09212085921273294,
 0.09533757817453348,0.4162554243991492,0.9169761223261835,0.2425534076784145,
 0.013892636555153937,0.6173593702686023,0.07676042762465662,
 0.8102591697693179,0.5341263187521761,0.5669771440203816,0.3961759250337038,
 0.33190087858742334,0.08123264592552759,0.7071818837622273,
 0.08525679602816905,0.4536890659168564,0.4963868309614783,0.3611874292336469,
 0.556633899545566,0.7431960526815988,0.4196876076466136,0.41316140685647373,
 0.6406628323812396,0.68845552305538,0.5718561825544755,0.061122054047104024,
 0.07427504071968938,0.8367795317505696,0.12336866432708271,
 0.2005832108883503,0.8336323017159478,0.5197863603678164,0.5472921119212555,
 0.43172239866248974,0.6689466845167803,0.848739767276036,0.8495293907152814,
 0.6827639638151055,0.8182887972267362,0.6613512052535973,0.6071593226653457,
 0.3087651795985965,0.8940445158815626,0.7053579534909744,
 0.030256407522919804,0.47078086588811086,0.6522408039810944,
 0.9546538618281021,0.16411237669238565,0.3682470010427089,0.960811899790507,
 0.27160134242426603,0.06271742329928509,0.8911459399364702,
 0.7590064311828654,0.11324120687292838,0.2408983669781637,0.6992598959515524,
 0.6668123166254685,0.7203617170670211,0.9264069478126912,0.8950500160305059,
 0.04576643485879062,0.6509966418959756,0.8132938658785587,0.7065126829642239,
 0.9282742661695131,0.09511586929633287,0.09951979161334124,
 0.3744218713899403,0.1882681389343619,0.6744509265662054,0.04976141262240141,
 0.5258742527017921,0.19325839544368395,0.7417770921366069,
 0.11589953215358728,0.693583268581221,0.30013579292183123,0.5709959832368541,
 0.5967352483479226,0.6007931163105638,0.9087272154581145,0.3600111520375383,
 0.22796640778321697,0.6037933801290845,0.2588825231441736,0.9960492778682691,
 0.3877650126052661,0.8397913625875548,0.06509506640939922,0.488806616106611,
 0.34049781911149746,0.7868042540843158,0.29186347684584857,
 0.22880198519183448,0.9287732513446466,0.30517555012248954,
 0.30093170744651165,0.3358253344314064,0.012621882588683242,
 0.12234503662579999,0.5556897224064362,0.918403035023486,0.7475690967023367,
 0.7859321897204202,0.08522019751610355,0.6013212868335615,
 0.31017579485308566,0.7737179344214483,0.3871787928743794,0.4508658611891493,
 0.42293589254769315,0.8302368076575323,0.015899819303010543,
 0.42189815520327223,0.5137495409913924,0.6943500991856892,0.6421731961711324,
 0.02465467053154502,0.06601007906874068,0.04892338162252219,
 0.1662948484956851,0.6111288757182758,0.7438239990059226,0.3791794674604313,
 0.6608741366403068,0.580741118379167,0.5925412265109523,0.16760386996481091,
 0.7248723762389162,0.22105081161406626,0.8460176456063025,0.5312336328983547,
 0.7855514361287952,0.7790841311160757,0.3530790614131365,0.7248447509963353,
 0.055264751844291915,0.0335325349575607,0.6530609178617731,0.916025450748009,
 0.24891782820722064,0.2574445684497074,0.6614563689242803,0.7144760407161002,
 0.052857314286130275,0.22885770657325932,0.5458967392054687,
 0.9251815497514055,0.5338543224468391,0.5576151574206744,0.9797961659603595,
 0.3362185493305774,0.9246363957322553,0.6133092490878823,0.7085899912037674,
 0.4587974972219553,0.3131785237031536,0.8965721096935073,0.3175212977700945,
 0.6106666207791599,0.12492699886705882,0.3185105887401083,0.3939080979597892,
 0.869698997759591,0.803656234374607,0.023537212656531303,0.5902061938415221,
 0.1742087635655334,0.2449017186756014,0.6893467026687247,0.7799216025791541,
 0.8504974766414555,0.3495048287450373,0.519036848351564,0.09716849760275814,
 0.18625274260365876,0.9926776308250739,0.9952151806547569,0.6879159345024269,
 0.36693194882053404,0.9262177586004458,0.5071380330265316,
 0.12759702800632955,0.8552140724509868,0.9237837591954783,0.9282110959007768,
 0.6011759300427324,0.15022069003101657,0.6172667195308092,0.7574329106975095,
 0.39688681323782704,0.6521053313937626,0.8828984621785572,0.6633577843598853,
 0.1569454560827962,0.5814132955250418,0.2116203385169949,0.8939396035711059,
 0.8793285389590902,0.9436735499132975,0.586496009487947,0.3157387838510124,
 0.47252616976220785,0.04071381203624136,0.20570022590007597,
 0.8963059750630101,0.42978394615984483,0.5620599346087174,0.8953855256868287,
 0.7393962547238493,0.22343394140886796,0.5990118109849863,
 0.014313377349703282,0.9518741277345053,0.35605177262146426,
 0.7803814140852189,0.6157155986120713,0.5266339936162231,0.4058567395590843,
 0.6148773045876744,0.5560559258701326,0.4672236539344886,
 0.018818822000183588,0.22006111390222838,0.2632819574628995,
 0.9609599821547969,0.08514460951544078,0.4178619886511288,
 0.21468840882148077,0.5566515466638349,0.47284829704412845,
 0.6756935706985799,0.30071205584480976,0.3122605049021093,
 0.46825031742517276,0.5880668442571075,0.2773407377435235,0.5858645179924126,
 0.2968268011181072,0.2958512729049304,0.33270593489834666,0.6070397189646686,
 0.9202184799838234,0.08221069534959613,0.19536925874962363,
 0.10646072311223942,0.7367159715042914,0.2962925424134605,0.6826129605017299,
 0.6366749248252248,0.3877177967118478,0.04668178635439102,0.7961276553236659,
 0.46577212465341766,0.597952935949844,0.15419583372052814,0.5193344334959079,
 0.7524811527907109,0.28839146313012587,0.604895142879013,0.47940157285957685,
 0.7114893892876524,0.5398308364816226,0.794167861526524,0.07136516649635194,
 0.871527472852962,0.6885614755293439,0.6529576576492501,0.2782670107814287,
 0.32082203257997555,0.32421361171480223,0.342503842113867,0.9780347313925091,
 0.6433541797008235,0.09645026210188423,0.9621466136163905,0.1703213209616341,
 0.26934818588264653,0.023318068648429557,0.07208433338251685,
 0.9634575113667567,0.839764129734645,0.2847122947026014,0.17697986423916734,
 0.3663728611326832,0.7983566394544105,0.9744244848200212,0.8598317741677903,
 0.43174064306369286,0.6436513140278702,0.292333802494937,0.1184495649474222,
 0.3816864603794734,0.6134575151014907,0.9040740545299749,0.6574417754644847,
 0.048205427348858465,0.8328405291035392,0.7141502115783473,0.637412974337589,
 0.9793446233209855,0.963913352293001,0.844179103018286,0.8249837296360812,
 0.22503901628235623,0.8579790455197464,0.1566486746836876,
 0.45954548468500866,0.18149758757310464,0.05411802564910129,
 0.21202739018095784,0.6519975285348119,0.35396193944740306,
 0.9842552338238846,0.15200569469870273,0.98882899775198,0.723206284795428,
 0.3985472797745162,0.37687061820031265,0.207290849969736,0.9791161132087731,
 0.2415425455278598,0.05989362545488408,0.11241595121573733,
 0.7258796716766642,0.33318687498594257,0.8961495713481116,
 0.05703246705065512,0.676683679844735,0.09605477451999078,0.4919896807289622,
 0.6617819054263933,0.9164484799083805,0.21489025568892817,0.5337656376271793,
 0.8040785107490703,0.06882172007322096,0.01717889635624892,
 0.9811753637424118,0.0965794544087587,0.9623337726052404,0.2062993369927011,
 0.8446430714142507,0.03782863704114914,0.9874433260375225,
 0.006471948928002402,0.8804839152423979,0.9639970195826771,
 0.9974440178621206,0.4191874369682842,0.19285982690682213,0.6869443417341846,
 0.6215053813266097,0.7707946113927464,0.7980302627737236,0.4787685790162063,
 0.4678703319504882,0.9870900845422207,0.012064169933715396,
 0.3341252785488562,0.6580652458092102,0.6859993424601758,0.08483677005270784,
 0.6803414856757984,0.6814036822405383,0.77973278629905,0.31997845667492064,
 0.9118239907758985,0.3737335694767049,0.568204672039331,0.8980904203906551,
 0.23089328452100222,0.08822481809670768,0.549371495570057,
 0.05864867168512078,0.28502725462880074,0.32113030647341356,
 0.26826145112519173,0.13721036469161596,0.5598725128631982,
 0.4651342353018345,0.39019557883060585,0.8082666866178612,
 0.41319279390997865,0.39798046343196525,0.9225798225469168,
 0.2368600481719847,0.7213383614824685,0.962199647226873,0.18899301054420903,
 0.8151813079114012,0.8359421232930868,0.2140514400730391,0.5601167555981136,
 0.6027535948697839,0.5581391935129401,0.7255229360877042,0.20675823614494615,
 0.44439043624239494,0.14686791138548316,0.29499395471043044,
 0.2905488472879698,0.1772815112839461,0.4576383884171149,0.4794801423393036,
 0.6268832427254334,0.007320318863712094,0.6208254482264901,
 0.9573476462643097,0.9653141411628068,0.5314674709759708,0.5673717029720868,
 0.9081168320815185,0.5518489996323938,0.3875605299809357,0.35303661205861436,
 0.9809066821697212,0.0736348396021167,0.13073406747016425,
 0.08482961908282072,0.8802511212711972,0.46685660978409727,
 0.08898401955211266,0.23250194031956017,0.12300645775513919,
 0.04602679344771787,0.9974733994518794,0.47715771029381226,
 0.13318780014958598,0.5724583441788935,0.19638047197813036,
 0.001584433231262894,0.9828194781943792,0.9583289221535923,
 0.2729236975302367,0.5223174864095829,0.43445353978301304,
 0.14877439953746807,0.5907967612686632,0.3687064543284194,0.5606659007697359,
 0.5761803641987173,0.9123934762388953,0.7743528491481185,0.7962394418921139,
 0.029168583587244612,0.20598004281055182,0.6321258942445396,
 0.03136542909111517,0.18526377126483062,0.29219999198669,0.3971597990013921,
 0.819602054815264,0.335467073940791,0.8524335788433769,0.5453217024359738,
 0.9301040900913973,0.6724545993273797,0.07831198776781945,
 0.05144073909110458,0.16678649689105562,0.9882426282680988,
 0.8571328049162247,0.26665729896745083,0.9570384486932716,0.4841055590695127,
 0.5690524935779931,0.8883138412351016,0.09512712111547728,
 0.10062733081886777,0.4004887349833983,0.31036790892747423,
 0.017355456233547728,0.5816975411997546,0.8708369862528706,
 0.29471393288610004,0.574458685777208,0.0593613067663582,0.4280379867114966,
 0.3818387807315219,0.3109927918066049,0.7351917354772176,0.7415298502960556,
 0.6255505040324719,0.37929987554959466,0.011194863137847899,
 0.32173106741644575,0.917380768549717,0.15093077067752436,0.9374517876123528,
 0.11180392144971829,0.04768967870235152,0.11990802212764118,
 0.4300937380050319,0.8098962365110925,0.2528841525339509,0.6355274401268727,
 0.5896131582193673,0.1641891538126986,0.2076752864919229,0.622374617478465,
 0.12279035792461057,0.8040766972518723,0.6646153330373739,
 0.031274183446407644,0.5011363128682922,0.9560956319615603,
 0.4883464445020419,0.6015754164290227,0.5583160971790079,0.12253875311354345,
 0.8467851989379095,0.8556234471200888,0.9803864901126896,0.4286071841557354,
 0.716767586280545,0.05768252817764552,0.7991725006497677,0.6772047393864887,
 0.08299497871904027,0.6659976085816283,0.16703474264999663,0.66220291660153,
 0.41207163485764264,0.655224092752157,0.25581676195842806,0.8672008156817971,
 0.6334849261846833,0.8714236423520159,0.34192778457649464,
 0.34399130860284843,0.014786328588716646,0.5698623849125588,
 0.4847782415653725,0.008835638089892517,0.09415051944221309,
 0.5263503040737215,0.20788370525427036,0.1475208771411608,
 0.23606930974928286,0.3851798222190814,0.9121504339300177,
 0.37739964345519017,0.06529827702143098,0.2225389115855806,
 0.3849859362618314,0.7910883139370769,0.6155048539614352,0.09288961425929632,
 0.7761263274966643,0.612511298494173,0.06555059229718008,0.5174608903178651,
 0.37841260897527285,0.5725198364276087,0.8426834753525064,0.8231264750289045,
 0.940938152003618,0.44778089996911774,0.7108727870015976,
 0.012948274338940413,0.58908555005007,0.29727474158206224,
 0.45367995139132666,0.4988898959293401,0.4125196929001165,0.7763334859815496,
 0.6752010648531357,0.7774787912942437,0.18106027853669915,
 0.052460015745924515,0.39604446629294277,0.8259059157191313,
 0.5788065379729508,0.18783658668656178,0.07151556074875376,
 0.3426847334224301,0.2311201521219609,0.19430709013934078,0.4876053971704404,
 0.12821475818868466,0.5221639527532197,0.5268352925719921,0.7983692583482056,
 0.44440767837971384,0.32735881994597116,0.38936924767606607,
 0.43603395518137766,0.20839779416074566,0.5236282493340821,
 0.35901448851932205,0.10268763808156833,0.9591103470887032,
 0.4678415397236092,0.8525795131358148,0.9800864165576585,0.7767562582034925,
 0.6925351467294082,0.8997041693943356,0.30549314777612724,
 0.041131438296243106,0.7544840877823927,0.060530839712827134,
 0.8203628361329843,0.2976962359339267,0.19039395967273243,0.8162514056684363,
 0.5066207632632822,0.42600376918448035,0.940262903833306,0.605327508541042,
 0.9695223117342316,0.6902588601642643,0.9921634954378495,0.27857447473749697,
 0.9161716581218382,0.1343733184189415,0.08239960245789413,0.4247553908132813,
 0.8083843538565948,0.6937482533083991,0.3040431052652496,0.43032568040855934,
 0.21658828226852112,0.04597509176341996,0.7056147428668559,0.920464277445488,
 0.03919938348328289,0.556192482796583,0.06193524240661474,0.1806772413035722,
 0.32011266788881265,0.10508975589760894,0.2545287304184193,
 0.7545223718049181,0.32629982437345584,0.04861446368623101,
 0.23049968152539047,0.6320068202366208,0.10239864344847449,
 0.2682085119829648,0.09967747519486403,0.8156853386526157,0.7038933915635054,
 0.9092225590917916,0.8018208884539887,0.7478253046486838,0.6423190726372532,
 0.7290990563716031,0.21082308542811,0.345658398298447,0.2418098577398401,
 0.6113503272634313,0.7848204486496329,0.04899685237200124,
 0.14367707806086094,0.7094760004752931,0.919888559981111,0.06863597808653499,
 0.4394684933166434,0.4550667533771383,0.38426890933647473,
 0.08942784647703073,0.9818696124742673,0.24368725245791512,0.550952284148682,
 0.04989654010837152,0.7189946717699535,0.577313495506498,0.45396112001574707,
 0.40468781851557634,0.6252064892460797,0.8412610314503606,
 0.32676130264718817,0.14409616817675563,0.8283071699933988,
 0.7470607918893334,0.2858462393700312,0.2470269032207334,0.7933140657634576,
 0.2472073873932844,0.2798119133352335,0.00808052560431749,0.7557816429366727,
 0.09474516831098445,0.33715607044664686,0.8621966961738347,
 0.8716004357839052,0.4971423316312845,0.37237100333641515,
 0.18945662137664943,0.728959963179376,0.5087063165368363,
 0.014506823483342424,0.5711513546922181,0.30292740643939,0.8130740133438582,
 0.6698576209248674,0.2904456562548454,0.7582973719760793,0.6696911962577093,
 0.06081487655224771,0.8812910851759058,0.25998759634418145,
 0.6847353151118827,0.6591580620171076,0.7357029642046687,0.7484391976363273,
 0.4669456317773989,0.7784863437439721,0.953000109711287,0.0721541933649642,
 0.7536998485702208,0.4291075450302887,0.8879013176753374,0.03070948048127342,
 0.07147272326707521,0.43202891625482254,0.3738152298498487,0.913493163188843,
 0.5275774178505587,0.04508964372399138,0.6618986451440102,
 0.38549014779647506,0.798115886069576,0.531011804563823,0.848293950660533,
 0.5511278291651168,0.6569602281928686,0.2801017402472179,0.12774280253807302,
 0.7290000157522913,0.9925186256885457,0.20143246671892379,0.7859438857544463,
 0.02593263366900278,0.0691967930989269,0.9971792548018776,0.9468981176626924,
 0.7597094261576177,0.3563830082732784,0.7316288132873481,
 0.005980662102386014,0.8339294137387424,0.760347546805405,0.8702759676613437,
 0.9075923583241683,0.17591642905650284,0.2904503767240788,
 0.11433628871801782,0.21700517640194938,0.8150914702354184,
 0.01922970979585792,0.5762480984276507,0.22479214302944484,
 0.4103703138606065,0.48264708390052424,0.9972567486277629,
 0.37455861055279893,0.9823629780064542,0.5195347625448571,0.8892061171695642,
 0.22106032929394637,0.39325184646119327,0.8831718538553,0.9238792090555714,
 0.28063910842632367,0.44375249936131156,0.3521532676631619,
 0.5403888257268306,0.9398640143005386,0.40987824432345143,
 0.28224205944452563,0.15447905723162425,0.22100124578726454,
 0.7864746616952393,0.07906807449296238,0.9839829856597376,
 0.39536516246383846,0.05062342169874823,0.9606065592939619,
 0.38406905898901833,0.3246109883164868,0.2007244305589445,0.5941818122840098,
 0.03653758414578123,0.4049381977695412,0.33871550962467256,
 0.6051825910581838,0.14945788468231547,0.08359276908221336,0.648155354059752,
 0.34153090442494716,0.3524744192728322,0.4696905359010002,
 0.32855690137594906,0.20760792450989074,0.41769573775965874,0.32462464512622,
 0.6087448191644691,0.6275124922256969,0.4822867595300676,0.747551788035768,
 0.05693495450288777,0.21741576446843214,0.7373656960272289,
 0.3129431733368173,0.38512537877587516,0.4481420017273946,0.6218399436298787,
 0.4287411958156472,0.3660557456632283,0.1642893932884062,0.5818980776741887,
 0.1788086087210654,0.17900057047902573,0.31320023766995797,
 0.6781607271631221,0.8991839419159602,0.11903643682047782,0.8042378411330008,
 0.8894024499327353,0.6136147145010703,0.25592041045540304,0.2845184911020624,
 0.410082836972814,0.1612807925921076,0.019005248338871183,0.9542384580144956,
 0.1006751592708911,6.456418831934929e-4,0.6016938678832753,
 0.8990990806624043,0.9589343618896562,0.7839586324994502,0.8235263908774679,
 0.10883268247144195,0.3347221941082904,0.2672178059068513,
 0.45233174991922154,0.4261891988824049,0.2084552966642358,
 0.0038926247228013855,0.5153159249452282,0.4653471227233511,
 0.9163975939614479,0.5637648382091841,0.247045286137633,0.4677327156946902,
 0.6690382522562134,0.4262757114590072,0.3909991444213111,0.7196852801476964,
 0.5121096233221172,0.7801684072009435,0.49622392016752903,0.0468097929211122,
 0.8259664008758884,0.19299787144797065,0.3291908131386275,0.9665031187044733,
 0.5822678468519061,0.4584068715961185,0.657807304631144,0.7505711538041504,
 0.3765357605609795,0.2948707972001431,0.23695850319397704,0.7847047485573605,
 0.4974026130020094,0.8766503072437343,0.3065444079178079,0.5036263463427266,
 0.1342347412076066,0.5617201117336214,0.8282490740831975,
 0.023243111431022978,0.7334156929384203,0.2572907519896298,
 0.45799729485557794,0.33662534404167843,0.7000101200650891,
 0.6116870794544146,0.4614494646375913,0.02842659602594455,0.6782914447039625,
 0.19771172860811614,0.7709879864923481,0.8266175502266564,0.6909177453134202,
 0.915288391222721,0.46025779736335726,0.03149324465143477,0.5325341389085763,
 0.3414529961480446,0.4323913949370496,0.5217130706469215,0.9021611497654208,
 0.5319908564502933,0.5620299531284527,0.8887464314475063,0.7266359724678476,
 0.27086659912299316,0.8911800370766848,0.3731838827704339,0.3592520093110033,
 0.9354442156043864,0.8638498897292912,0.9648733460039206,0.9565535817840706,
 0.5743986254148298,0.16754549243512895,0.19397159604098713,
 0.04161031630104073,0.7735170968881627,0.946995861865365,0.28495561973162187,
 0.7609407784171565,0.6071532015579624,0.6333980069130835,0.988186107167851,
 0.45707188777281127,0.9042374868069846,0.43305706689643697,
 0.34430438985377365,0.023721090831008818,0.5272859573252825,
 0.6020111486614944,0.29601745171457017,0.7400031652043209,0.9232609500106145,
 0.06044784503510703,0.9959379250229031,0.5529331816564252,0.9048247706251267,
 0.674522896799667,0.15926989112442236,0.5749515509523173,0.8961680568932101,
 0.445737713612412,0.6774984225361884,0.5570868694170651,0.6121961310945199,
 0.4501007498539972,0.20719627472279878,0.6164827773976534,0.1516625363237838,
 0.14504810939841672,0.18847593686084096,0.9821749479157472,
 0.48856460573937166,0.9341144591393064,0.4475133214110908,0.3170175026989748,
 0.3947302694802599,0.206292291565503,0.5083422728043903,0.5499403086243161,
 0.9512421907610416,0.4350582900377077,0.9695524695429167,0.6882249594867169,
 0.4996831233734552,0.9132096861216549,0.5655789171719179,0.21618581382195756,
 0.01773961132188351,0.9205169232879277,0.4058042832168002,0.9086081133227234,
 0.3180172503038926,0.4713089254904015,0.031338942525381386,
 0.39131121895585685,0.8873309190548945,0.2878113434864281,0.4950515732050049,
 0.9130146375881083,0.9211909906220535,0.30703436762864444,0.6163068874760005,
 0.10714758430613636,0.6188559214897511,0.06922585684310611,
 0.5825547133038379,0.7089740363050349,0.6480938479906226,0.6966208725481181,
 0.6957082217884915,0.7398009127063816,0.19046481249545844,0.655085905821752,
 0.8589323441639302,0.619750568398102,0.09409007476033238,0.19406125096952853,
 0.20657101900136066,0.5033380097557486,0.6824436122862968,0.5728117800111722,
 0.5599412200924003,0.4273681163979679,0.6139174648012554,0.22153191335334277,
 0.5289437958307767,0.7665291541174097,0.33225605146401116,0.3540979136025727,
 0.5319952277870547,0.8600380470221929,0.2969610392084826,0.6540199796479387,
 0.31708251302770796,0.16718374272741465,0.008175445667650827,
 0.6869771554308656,0.5542046561503384,0.1563576515255125,0.5821559457737221,
 0.2807906792957413,0.9569540595761482,0.2852432415848096,0.8856676458739335,
 0.972423398502061,0.8473312829113643,0.03245105825758199,0.06094631931816119,
 0.6463742226332151,0.2944420572064099,0.9860972777901998,0.6656022017807229,
 0.08830135215169133,0.9397052276942394,0.32255118645023406,
 0.13332091028678228,0.4857987821738817,0.14612368222056193,
 0.17308505885907888,0.4294471555167986,0.1357378873823979,0.6176931911665412,
 0.3422391664974028,0.1281779408718381,0.04162435588739133,0.8894854007616491,
 0.6488824493028618,0.8926038266803563,0.326886352485812,0.9786628909805855,
 0.11790453969465486,0.4822394522377297,0.9279762440314676,0.4543713441641787,
 0.7226572427137415,0.5789470336192397,0.5374796701438168,0.5764111091133588,
 0.9351395201966238,0.6848071219182238,0.13694618195195063,
 0.31687619213565554,0.27533294394979113,0.012808021160237493,
 0.5040444359425987,0.9168355819936247,0.017219250525687846,
 0.025571104675282674,0.1982966360629843,0.055697433182951395,
 0.650030469201106,0.23662296186190157,0.8114195275274396,0.9412920643097368,
 0.3625025816024843,0.37526980066682236,0.8089885134399523,0.9811500344141291,
 0.9667301383982584,0.22328639171172926,0.4024871933468318,0.5412592032209058,
 0.9588150899932917,0.16597058550522348,0.13464448670613327,
 0.47729881248411743,0.6551669079429452,0.6346730998642514,0.8960733931916071,
 0.5033029192640351,0.27803079091390714,0.7848480718328488,0.6493357292840944,
 0.7013820548533085,0.2641692203694107,0.25328191493391694,
 0.07636124204595884,0.44599979158897995,0.6498757092355254,
 0.09418761780552365,0.3276983196519342,0.9283269229012212,0.8628410969352068,
 0.13700673626606696,0.772564573072123,0.8092091350593373,0.103968721505583,
 0.2480105763061623,0.5999871003672997,0.703159534301654,0.8447214809736504,
 0.5299850984395726,0.3509707580399286,0.5421246212194102,0.7590168232925867,
 0.3905117958092442,0.7195179585430088,0.005061572483185728,
 0.057953001002304716,0.6646564427133864,0.4389862181075148,
 0.7625928349771901,0.8858439438008847,0.23872442330348798,
 0.17889617429703997,0.43678637540759324,0.12066457809815978,
 0.368176636994747,0.6533834327708239,0.6797189756708351,0.4982825549844879,
 0.06940490662947285,0.8440476048654442,0.06810999221887837,
 0.31166140881828186,0.019951880140772538,0.7387642657390365,
 0.15373064530353697,0.6139340429005297,0.9235863853343584,
 0.37246848453341264,0.24003833920076156,0.7404801281742667,
 0.6707883469889705,0.6611985265371414,0.21026943179516433,0.5672928504234811,
 0.027663880373110628,0.5128975705240849,0.012989039520342582,
 0.22121576168147616,0.7038737470961135,0.33026964938393755,
 0.4277446436963954,0.3567186652451457,0.6782554283017044,0.03613030144398466,
 0.6662046257191625,0.3808166260872351,0.5334253812768305,0.3991305743826905,
 0.30057394983975705,0.22785718246161268,0.9371625717799514,
 0.8318700978826488,0.1739740509435248,0.11042344226585932,
 0.14299346250471523,0.7696622590276101,0.9668262664585205,
 0.19858556359059143,0.4794502155264906,0.1946505504852477,0.3900680105184212,
 0.9283156932354338,0.2742065662661515,0.214659323632172,0.6593044928298795,
 0.396071529921155,0.7643650810031273,0.575440878877489,0.32985833676894105,
 0.12182603019472182,0.7073223333619035,0.5985318015111725,
 0.18032177928313198,0.06248787766384911,0.30192303583983415,
 0.6493063042025432,0.80020157698157,0.08717741912536492,0.7614978030530496,
 0.2730064530217231,0.7080811292078272,0.8494669769180582,0.5833277048751434,
 0.9962823335295877,0.07690137428573462,0.6097114438734609,
 0.034116416817050244,0.8838914073672178,0.6048766504787025,
 0.7865062324957823,0.4237226732042857,0.701981028100269,0.47321554501222274,
 0.9221350295646431,0.23306475150223516,0.3302814406042288,0.9664324224456188,
 0.7834995089246359,0.7980374776834755,0.23047774623950879,0.590811722207966,
 0.5617373996836614,0.755059635445041,0.5443348999751298,0.7520930198102372,
 0.10688361855677742,0.4495367342105535,0.43313916646317707,
 0.9783902563587585,0.7714275482176969,0.49641963217199503,0.5209016143179837,
 0.893715505943729,0.5309954436586729,0.8338331295732206,0.49700825389389625,
 0.01830097261859087,0.5872750853052046,0.36662272463898493,
 0.5539928868939076,0.4968900250061017,0.7593377810322971,0.31179737131989327,
 0.13260179261190763,0.2549038566553009,0.0483670787261673,0.5213808818981478,
 0.8041070914697117,0.04656692337228663,0.6046284104246954,0.0504462978102711,
 0.08221151018990658,0.2974139779304177,0.801348979882757,0.7553603022794572,
 0.49029038074712383,0.828942309705686,0.21720870637730116,0.1953838544619303,
 0.7630415421931974,0.5828281383132656,0.9053891033288497,0.6713040743025528,
 0.6784385945095145,0.5977191625856939,0.41002489113514395,
 0.13419580439408907,0.409925558466641,0.8517338641587148,0.8378919675249691,
 0.43928901266695786,0.6819294010136183,0.8282577075165535,0.352509834906088,
 0.7344192731280843,0.8481210683902317,0.4702556603384759,0.7678638666515312,
 0.8849116122452542,0.59879198572939,0.9908777322908676,0.396666010378533,
 0.15412879363973564,0.5454253024541772,0.5255619874589764,0.6802288599582524,
 0.29040216380920525,0.1405187770070483,0.20285777970883467,
 0.40730966433719284,0.20018773504097853,0.2225467751287339,
 0.22679282045020566,0.9373878079792146,0.4076918402738139,0.8328790518834542,
 0.6322113608107796,0.5193170531225029,0.7191025010356714,0.6342181286150277,
 0.8287686427186931,0.8622089777414385,0.09651800302690883,0.1059541890387562,
 0.3418644304663526,0.5931508411492539,0.2712569532110969,0.2946658250285402,
 0.912257829488901,0.2988446859288447,0.06604723492714637,0.9851216897384499,
 0.7277243625079338,0.7811329509745594,0.3269997706226391,
 0.015511967369806201,0.06382574084877468,0.7769954306844163,
 0.9562782859947138,0.807486279328601,0.2201018922974558,0.11998096349810472,
 0.20991754054663436,0.14632785392999603,0.870770073130641,0.8743644781509603,
 0.3199197903822282,0.22020361381127873,0.9781198404691469,
 0.41371299909582226,0.358958666521368,0.580142622255136,0.980732977260982,
 0.6826643699168514,0.3375031086711734,0.044001472340175596,
 0.3523201871051911,0.8526228853160663,0.7328757391942382,0.7034388214925674,
 0.31639207257531976,0.23351045144753768,0.4478730997805179,
 0.5597602489951677,0.5311412735028521,0.3824626847332435,
 0.013140262935067759,0.9747503897119645,0.6369658568589668,
 0.1774476691165724,0.7362099848575225,0.7863353517626223,0.3824249303018702,
 0.2649530172902991,0.371701637320232,0.1606292705213903,0.11205577538424605,
 0.539021705030187,0.42129041851277294,0.5262211981129014,0.27548331325666553,
 0.663222399477204,0.5985640424639103,0.35385675781539183,0.9553496085402116,
 0.7120618425815766,0.697265763076495,0.8432352743891122,0.384411786884543,
 0.4136024414657822,0.6830101055726538,0.9507758357460153,0.5991132045868778,
 0.022157079386604828,0.9995314964751411,0.7123233244585622,
 0.9674051179930597,0.8522150835996178,0.2888008011565206,0.40224153986523714,
 0.06149850346619812,0.07013529102750682,0.7864710789516813,
 0.8741022086543293,0.34485003461149966,0.1834978591311771,
 0.29969872289910804,0.015512501154501535,0.6957027921010884,
 0.20067228637251344,0.0915044401389924,0.7704403199571217,
 0.18949975882548165,0.3384168252696198,0.7836367661682875,0.9417901742336929,
 0.7587954150311963,0.9591632504785549,0.7027244518318347,0.2170463009902468,
 0.4133583159261641,0.12861917712888782,0.33818997681253105,
 0.08007429975865754,0.1811842928998426,0.5570448115548653,0.6654938552037984,
 0.4547941719683767,0.5073062956328788,0.1790803970201651,0.7900153787476077,
 0.912026515971292,0.4323281470690432,0.4265534970666227,0.5726772197973827,
 0.6680864760060553,0.6511151269356868,0.46875121449315016,
 0.45574227987003446,0.755109807125523,0.9079724686320162,0.07640668182601262,
 0.6748448242622622,0.2542602514085024,0.19819318618301285,0.8602042728279549,
 0.8702262089394925,0.41361296237538536,0.24594686509335728,
 0.9598709395620624,0.9955573764887867,0.5463619720537954,0.15190965030380743,
 0.5030709666352909,0.15694208932647313,0.8455984664401405,0.3879164390464722,
 0.8744982848097514,0.655554571791657,0.34499889002360384,0.7679467138515907,
 0.8172199926604199,0.14903782320433523,0.5270811342582653,0.5542709681794957,
 0.998085849831199,0.4386745113695909,0.8851359862950895,0.03245034091615995,
 0.6552327482237114,0.5721218309567786,0.6634048918422509,0.25576701865671736,
 0.9442786422473975,0.09690252240462816,0.5746141485206512,0.4516645846468582,
 0.7733753966326655,0.35294884157540984,0.5615717836998478,0.6463831463775529,
 0.9830159869527599,0.17456899864428854,0.02606577738449578,
 0.0144484164063452,0.39789873458937197,0.9451775386930301,0.274200671562987,
 0.8550932371138483,0.783829127881435,0.24374274422687792,0.8881494931525726,
 0.17370244013599057,0.8794303151479398,0.2316427290216454,0.6868145034728355,
 0.7877523434508441,0.3451847031886949,0.6259476102975968,0.42455399137640537,
 0.00553063314969926,0.2669064005012445,0.12235247991501408,
 0.5422187152838958,0.2345985073168333,0.811463536460696,0.9416325780273743,
 0.47294985635751075,0.5547537025218572,0.44008279866095523,
 0.9296520819619091,0.47186234034106467,0.6962410147867836,0.8596435657692917,
 0.18349362840626293,0.437690311576324,0.35818534737730967,0.7615047312532996,
 0.49380026445687764,0.7215949745104555,0.4007841711134438,0.9589549654732908,
 0.5374061466513175,0.9855259759536159,0.8943055303132397,
 0.018791823754009562,0.6446035725180307,0.4633201840290936,
 0.2815724568045995,0.05422461290301195,0.268195943145188,0.7899476244645331,
 0.568918036102772,0.5064471363585421,0.8573516503728595,0.648432942727187,
 0.1089506254226087,0.5187613789855127,0.49404245443664274,
 0.42604697668638836,0.3435322292767762,0.397166099612953,0.01815931389989578,
 0.6400758897111931,0.38255316977827636,0.5857517629578948,0.9222404784377466,
 0.2577823772846982,0.8800422393385174,0.4009250870424115,0.0385422356401014,
 0.3569806833112409,0.8191232387153223,0.9644718040282644,
 0.009544683833703438,0.27657547184991516,0.2896097183124958,
 0.7809857704221111,0.49224058261459247,0.8173866440165306,0.1979424492004498,
 0.2857832820699543,0.9956505808688908,0.8068159425753048,0.30899642242459413,
 0.3878506273486787,0.1228385670484392,0.9400305882348783,0.8757850634602656,
 0.9081588081603748,0.7622371824735132,0.7397379855841502,0.9149565745292106,
 0.03661573820574093,0.4988432566803396,0.7759642788550213,
 0.48400536213600676,0.8867448628363617,0.5843910727105818,
 0.09726276800679556,0.1409379873785629,0.019031108386811768,
 0.1964585153024656,0.004238795981202692,0.8292692742674252,
 0.3107602300357998,0.08460048947286536,0.7528978259660931,0.6476262007392001,
 0.024608962135460932,0.7360086406286178,0.3337753576146629,
 0.39274526296529244,0.3794646095122741,0.26147652292680734,
 0.12846414407486906,0.7668278479626617,0.36397433878800944,
 0.45600622959916914,0.3099019571198942,0.06070523863409916,
 0.9233518786467111,0.9237826709842112,0.0243294267207812,0.03917852773473873,
 0.9442815513872984,0.3801976392135199,0.629741217642937,0.39790508023068105,
 0.2241425273437624,0.0747674810537532,0.1866080843804505,0.755843667941361,
 0.7166354777353926,0.7390261205822062,0.4448593799405228,0.05908846050580041,
 0.20415492370432053,0.6373400697291337,0.5443251740373698,0.6818625465090584,
 0.6968588686939674,0.2999769318736196,0.6304042541487754,0.8723286119566152,
 0.12431169484433346,0.9947700737430913,0.563877200931354,0.9721653378983712,
 0.622603369474962,0.27608735756732994,0.3709635731777857,0.12365141069971375,
 0.6494151324170376,0.48396932220345956,0.9212471530194251,0.7068085960900481,
 0.7037474284289583,0.4946329938758659,0.15106414241576016,0.9757762248955133,
 0.19821264348057532,0.6259070988841321,0.9304785509692483,0.4423209893789072,
 0.13943681216359893,0.17431172545398466,0.10835404590743192,
 0.1549986732100298,0.854540759613055,0.18958709630670612,0.20448454291424434,
 0.8096994087099318,0.2633412796985254,0.06324120043201265,0.6432302309589806,
 0.009493066387271099,0.25097692162038454,0.6451901582808822,
 0.9829288495531147,0.8650419584660936,0.9621324898816703,0.7627296067654828,
 0.29687391758907555,0.4526117511257589,0.6468259838092414,
 0.41495821245247333,0.7869099899716083,0.3270114251324687,0.4202844522566118,
 0.5273548897033064,0.7041211989146884,0.8319587103953588,0.16384554420633002,
 0.2974064187798301,0.421651491598825,0.33968306024493855,0.06024181256664052,
 0.9950829029917244,0.48835056967859136,0.1444129868077818,0.5740721074274094,
 0.3188821391368295,0.06572289979845147,0.33653614222499084,0.502524608272954,
 0.5651338021778506,0.5652243905269729,0.98718806627925,0.20067658496074547,
 0.8788067070693972,0.9111033160728388,0.8484424999876179,0.7534969013540377,
 0.716023457121835,0.8041296487473331,0.12010217596787365,0.9313604049672839,
 0.24781545748942424,0.30880356704880874,0.8851823402564959,
 0.4703052671198782,0.15032834113993232,0.6324465299840075,
 0.04221177827543765,0.6307926885311991,0.9162427477789582,0.3352828369173988,
 0.49535290915333907,0.3359309988994913,0.893133841253517,0.6009138164203864,
 0.9942138765475348,0.42340755292899945,0.1837085072520528,0.8095975174218506,
 0.4340167414891216,0.3889211928236892,0.40221429134448716,0.614088229521224,
 0.04873317672093269,0.9980016510405557,0.9796165276604136,
 0.31821302773419313,0.5972584365840579,0.26890751826729054,
 0.21390228253730592,0.6476212909464987,0.8972629624342192,0.7730632121117137,
 0.2643818681872172,0.31228062282788027,0.4092298385850698,0.1600649241341825,
 0.497551876651538,0.23712642198521272,0.7638443103078338,0.34505407848454883,
 0.9149289800828724,0.30099725318300197,0.8905098908850921,0.7080023510705317,
 0.8025322377064568,0.6793323391778617,0.9519612024093588,0.2691150113119809,
 0.4457679461990183,0.8138503484053505,0.9809989831130035,0.5740737131895164,
 0.18866600028763614,0.05247123958988065,0.17609797970286267,
 0.38484240537949765,0.9833678208946709,0.22036844356084306,
 0.5126624716610261,0.907341778063351,0.6228037289555436,0.4862924885123373,
 0.6242386604167315,0.4093745980935952,0.4476167977230475,0.5486452301178049,
 0.8178718520326588,0.8927601113498059,0.5678537625463584,0.2642348772038192,
 0.2509461433280278,0.8669173002885351,0.6252751120457641,
 0.007731388148752627,0.06241846592174416,0.9530279568416706,
 0.3024788487692309,0.5952790181203389,0.9979755614260231,0.3975796272863926,
 0.48103690720191594,0.8568087523926877,0.3207715576833787,
 0.27134290171154796,0.5002831792553711,0.8333464241348137,
 0.33985552336421243,0.850986323007018,0.7548384447601579,0.4849436116191783,
 0.4717977549285064,0.7733080156637253,0.4258926922867765,0.26645151505739806,
 0.8815862832152845,0.12440031290111708,0.3758765726783153,0.5297932327024933,
 0.5972235711951834,0.992584335239891,0.6419508769912355,0.9047639416166957,
 0.8282555558393524,0.14171063169000297,0.08089536684648613,
 0.9903497202084923,0.530596799497571,0.1917830869480811,0.8624885999072021,
 0.9405518215101455,0.4278734764844324,0.7954878519917239,0.5345198465443272,
 0.9698976480344397,0.9474480607739749,0.9966837906353809,0.6842935212111693,
 0.8107870055792751,0.17938048542075669,0.40626973384096254,
 0.9341540580490998,0.040455123232235435,0.07661219857512303,
 0.09304017029390332,0.4356519554874253,0.3675832095248539,
 0.023347450493623523,0.958496502162987,0.07522325125032814,
 0.09642851465996294,0.2352233521499184,0.2456669647843992,
 0.15833918431107108,0.9880131770119105,0.8162946650502382,
 0.05621130772172078,0.8862199077644463,0.06253672423007228,
 0.6694732952342624,0.695555084642653,0.5031397846730765,0.25528913881834225,
 0.7159035111571969,0.7901873647001832,0.5983467965363743,0.8323706165346518,
 0.9925304751653761,0.2989604130452109,0.31209488900642557,0.6226140328703039,
 0.9523520074392366,0.649732535236243,0.9207601859663955,0.1719083045689267,
 0.23198788995075614,0.6430016211335579,0.6178260374980705,0.685348950088253,
 0.49016951190728575,0.6442374786210818,0.4560261387577965,0.9822360741664422,
 0.954928764637546,0.2187314584766754,0.2295330168546461,0.20910073132895446,
 0.5335823344685077,0.18204652885725592,0.7154084184601947,0.6154911841397994,
 0.5515736749328286,0.002944328636371618,0.4855913617311607,
 0.33333345568939876,0.5254451701650334,0.90236209645963,0.6852962229889151,
 0.8601682894066386,0.5161921383165404,0.7512891192625276,0.9445831205885877,
 0.10291998202099195,0.1529779197048875,0.7819191550030381,0.6571021899048004,
 0.2791362753834694,0.46877496105245775,0.5634778853881368,0.6184494675868698,
 0.16980114766520982,0.23888429241335218,0.39591681784813826,
 0.23061451644233666,0.5247731976754357,0.14731561151118777,0.6346120478673,
 0.27029721514816296,0.5655701554320407,0.7697607131991387,
 0.20356802123207718,0.28900856588534496,0.630099585069627,
 0.007528414363302205,0.6792275019482262,0.017950748883981515,
 0.28976200115451867,0.09354632060072976,0.3735595194301655,
 0.03850972549573839,0.3074095488001177,0.9973712881591088,0.6357361762126806,
 0.7811529085473894,0.7134894987193757,0.6120905682089266,0.23481659170267044,
 0.5006168789477803,0.3448472602914002,0.3040025065600517,0.4315263619373112,
 0.7381355019841316,0.8491699025064919,0.03929069851067912,0.5771757822590965,
 0.7656357169851941,0.6641455372743547,0.054250649310277343,
 0.48834120083367116,0.582715131442401,0.8196696104865059,0.48979066484098577,
 0.22994434186334523,0.97873050375015,0.5215844441062708,0.8597514753156115,
 0.1864447877497446,0.09727312335134852,0.8585442567621505,0.9663939033999696,
 0.4411216440704895,0.8625704260661281,0.6105582969114497,0.7563036357086634,
 0.20314219943669165,0.9743125918421665,0.2722665400924602,0.6915712673927704,
 0.4558851334160986,0.7681440801061675,0.21917997584040627,0.5234433078878087,
 0.22972168096007262,0.7629655225078599,0.5621488236513721,0.2592091043650181,
 0.5889755707947515,0.697494651137512,0.3680404623037087,0.5574635166065525,
 0.9445679828704154,0.7166157730997467,0.22493096514248523,0.7176087410703939,
 0.0561026823986257,0.445880453664522,0.9254914711377351,0.051681351336196224,
 0.11311459921124545,0.10897785345591426,0.7415289707800502,
 0.6167209531591301,0.7442091549300238,0.36973714632022525,0.8783187223591806,
 0.7200363898554469,0.9806894005001582,0.3862942825115524,0.8738963005474502,
 0.6577459827175558,0.21820096209845985,0.9456650222191081,
 0.42698928652026136,0.9935949608040369,0.5301561333851857,0.9757832332810565,
 0.9841530984140543,0.46153016649483003,0.568611018409523,0.8218882068344097,
 0.7336299831484246,0.39631525646251486,0.7178065592835897,0.8615957104647757,
 0.7462953929119946,0.804496445686583,0.19188041603133732,0.8633474063528634,
 0.4580245917668643,0.09004058615313393,0.7427308296224765,
 0.13827132952338694,0.12469021019227666,0.9625182107978781,
 0.009555163770161301,0.24835159208531166,0.9716986645098247,
 0.8375906580931689,0.21091009604441113,0.7091579078366912,0.6621536194100636,
 0.784079384201978,0.9344815478033658,0.7768943789827432,0.2786898926917635,
 0.12440632619884251,0.7320107386042939,0.9619029154864529,
 0.02218316005651999,0.29429234645433877,0.5271513490695603,
 0.5370899029949373,0.7564600814129416,0.4389966802401951,0.39332142113000423,
 0.09559992853285415,0.6370062601108335,0.9938812884470274,0.3197523378191308,
 0.3522674510118321,0.8097167108917693,0.7215530100174177,0.49572295972369174,
 0.19304334434720627,0.19035649934365217,0.5187819277902805,
 0.7506893504457688,0.7394743901715354,0.46308581239527735,0.1469388634421246,
 0.6919061732432018,0.467978001816429,0.8111510244739011,0.37848618349329444,
 0.050887396266595664,0.9054450772839591,0.0891621205633717,
 0.4362661336371676,0.8736169730191543,0.1157237473929541,0.810756492152753,
 0.35848536423429156,0.034971723276860045,0.8150635303775526,
 0.3090628860136867,0.15311263957296506,0.25944816013094407,
 0.6379481603852243,0.2693131938458231,0.5677618653888263,0.3620776774290022,
 0.1779553474710701,0.2253312581115896,0.24580273067730873,0.865568312798313,
 0.8311755111278105,0.20116669337580073,0.8591498533165322,0.5503914731498694,
 0.40414550995141063,0.7582235044967052,0.2754303079930912,
 0.17341409423561105,0.11950202641904384,0.07705620880873387,
 0.3966412420288896,0.4856414252813097,0.8396098009669193,0.12261380747213613,
 0.9043008108145153,0.9138732473002065,0.17972469886524578,0.4403165363189252,
 0.17621578998036558,0.483225592962754,0.4119824832041321,0.25537141777813566,
 0.29759046720116256,0.5288366614396425,0.5198041537160571,0.668311360376225,
 0.24869981612054082,0.9243554913528268,0.4380111865887826,0.571292048180054,
 0.3208344197282885,0.5430601372466682,0.11606177375797488,0.6811964445296159,
 0.05812817168725626,0.7188414223868375,0.23116672561378782,
 0.7513961978919055,0.5736358703602533,0.17052968183081996,
 0.28410259147199945,0.5541735376397082,0.5513554324861419,0.6705407211026002,
 0.0811850823607253,0.692142133071102,0.2574696474914292,0.6002433294749028,
 0.3154785407683557,0.03603993897293989,0.32964887463830095,0.879121592329415,
 0.37310682540720297,0.45664196973267,0.9155369945853277,0.7771824175254199,
 0.5084818909761625,0.5872015543365074,0.5957769977835677,0.11668204217431188,
 0.26507775283907997,0.013763045883743663,0.8342249631593581,
 0.5651337571851944,0.8715656749420755,0.1352173492796761,0.4473423075504803,
 0.9740944943810661,0.10638045270139385,0.21711725556322847,0.61189308068404,
 0.9213185711193367,0.09795203421703902,0.4574208490252206,
 0.09448284979828048,0.7735817555073593,0.15999166333732318,
 0.7562118569476189,0.33980066212972515,0.5084436409067662,
 0.016211898498938004,0.8803350407510098,0.9889046673443354,
 0.5877630397218689,0.879483422919134,0.08231522078538989,0.580112521269611,
 0.8913876707255398,1.2676280095380577e-4,0.6179504861877043,
 0.7895450506099482,0.8175590100289325,0.8020280453024102,0.3339734665369839,
 0.6733070419771399,0.5989801994357243,0.03283590400371117,0.3889826901569816,
 0.9111550464142774,0.27877161748282975,0.4438058424594292,
 0.29485563564933204,0.5831782529475696,0.5278434284498226,0.4409944551503304,
 0.6544972210105208,0.42462224014462036,0.34554367551268383,
 0.19033216887565496,0.7240794497992125,0.1770021719512891,0.5636317298305222,
 0.404448253228503,0.4930086192356746,0.877113785120605,0.27843479960776274,
 0.9735868832384305,0.14549859927374553,0.8550357779011333,
 0.36283572063854086,0.7040701786416403,0.17635585804027598,
 0.9147310143735908,0.3221436004084838,0.8989758697172007,0.8020198491777143,
 0.05888995012789966,0.6099347244660211,0.672445624331367,0.4805101733321695,
 0.7288787413220105,0.17713961359917796,0.3489291989113581,0.5560112535498243,
 0.49606596305404305,0.16062414340432896,0.6742000374869838,
 0.9305217292544576,0.7359024208365356,0.25157739705709514,
 0.48525672384831464,0.60843860557063,0.4303742628053344,0.6831617103799436,
 0.9955518275151688,0.6851238494615266,0.6501342302030176,0.950369068337356,
 0.1587430222769346,0.9964729168042364,0.24904184366379256,0.7837343415578351,
 0.7784478298049632,0.1038681329504958,0.7837439391604275,0.8978487600745714,
 0.07578955082294914,0.027162428265842142,0.07804745389797951,
 0.7105788039877505,0.30616476446360896,0.24476610722725112,
 0.4613866361610619,0.5215058411973504,0.41116255577463134,0.8465254600818002,
 0.45323362305568093,0.5299209882580563,0.6880691748363544,0.6976718793996635,
 0.7632831653596298,0.03208330800597181,0.44985632796971364,
 0.13797006846122062,0.5666093376204022,0.7119729498391301,0.5966373709954896,
 0.8602066454918964,0.5043050592601183,0.09844100633263764,0.7773949954038061,
 0.6882248968291883,0.4334040154953255,0.24298100095457265,
 0.10894503455187043,0.7165262333749721,0.01139378549613701,
 0.6692350519703566,0.05840539529357269,0.10771148683303056,
 0.33133378188205986,0.1549858982432415,0.06031180479186493,
 0.2249547248304432,0.38398671268001583,0.33931061188086886,
 0.0770745389142018,0.4958485168206257,0.19727934963604288,0.9242247728470478,
 0.24085408533237518,0.5435247813868195,0.7495627721328344,0.6216428275509945,
 0.6846399749006881,0.014539866779344601,0.8954636191663079,
 0.7484869764926354,0.37233233239345753,0.9201840437242451,
 0.11241284062689882,0.5981495502242312,0.6468329130217744,0.3557136906404853,
 0.6515007321928165,0.4037216947528116,0.9793196712810543,0.7758999752243881,
 0.47385899522245734,0.18162069063155184,0.09647446240042268,
 0.6854249341673732,0.25875589273412924,0.2317470485647226,0.7850287524412963,
 0.5998732111523098,0.8472631931064173,0.8813308110732389,0.9581658629436229,
 0.39210082345394226,0.4376574801221116,0.7328318169253062,0.8409901002698222,
 0.9597521971359262,0.30059257913769,0.8240940722948062,0.13261524780369238,
 0.7814865042448175,0.7737578328530176,0.20057855416732284,0.47031016737777,
 0.8277103558796031,0.6390242258635003,0.15297884499772707,
 0.13072984361518947,0.8353618443005432,0.9173795309502735,0.7532829073886237,
 0.7695705274997465,0.287735513419915,0.7452313815507692,0.5638954796055531,
 0.3666398049808933,0.09806733289599734,0.9191524168994901,0.7070558405094057,
 0.1829286950894825,0.37295881676866427,0.7347463137948264,0.747905601502956,
 0.4726970377330304,0.3018974960458207,0.16478287226637334,0.74133933073127,
 0.7252449076986125,0.747360121397115,0.06138757332225753,0.4955007527824261,
 0.6342476196485461,0.6413821578665558,0.13750695684491143,
 0.02060130754054945,0.6370881034678864,0.2304529932808358,0.9847509866401954,
 0.5360783875725275,0.583029945066485,0.762701405901732,0.8238099355435038,
 0.44729203870238965,0.2936275296150007,0.39861425994075805,
 0.36047334432965206,0.5824234029691234,0.32796574849809057,
 0.2870021038310192,0.993873941011455,0.3181337602257437,0.1634742559731186,
 0.9907882811265822,0.21486745157522835,0.10014049869023067,
 0.3568459565293106,0.16436165178160633,0.21618422855900443,
 0.29822578556250123,0.7498877003704214,0.9117421771314095,0.5387219295722552,
 0.08543737408203966,0.973616613317709,0.7328552637471709,0.12055555068649437,
 0.9887242559745898,0.45875038230975973,0.592877596164446,0.6580547455592323,
 0.47852988993193923,0.3365860543977327,0.3843579025395196,0.9011237242775448,
 0.6437035128104677,0.1868352805464667,0.9188123094502123,0.41991379097929526,
 0.5677825196973167,0.6299169499702774,0.7245445499499432,0.5111830853980411,
 0.34435937365383595,0.889933655045484,0.259716258240551,0.02491840339612239,
 0.2034239258564522,0.26897157662803095,0.17158374215755168,0.325736395562677,
 0.9073863552575278,0.7536634258642423,0.9444046435643916,0.666045393037415,
 0.7240689244051931,0.2577742828750278,0.3040091074601414,0.8422679643160973,
 0.8318951729114981,0.428526246362253,0.15410162080075795,0.9501029226441564,
 0.36206703475982416,0.44206049761772204,0.9352788338633722,
 0.6520894573766526,0.4723276771197523,0.15252308383521707,0.5156353321229552,
 0.3936804425415169,0.7317601875636897,0.9290197252701383,0.7004545773540204,
 0.6642844343080796,0.7539190227049557,0.2601346172257364,0.12162428842736461,
 0.7591085470293398,0.8003609021655644,0.7903160171124664,0.8941977963455258,
 0.5647352270672116,0.3966787409002871,0.8720600141748575,0.1521848995215599,
 0.21582559996164186,0.45175324839603226,0.40791663324490623,
 0.24072521249637224,0.8769269277542532,0.9147548138352308,
 0.32530411540878834,0.8047370645366938,0.3756013269148628,0.6430929939323554,
 0.8621221190217372,0.8928413523793153,0.8314342042354911,0.6617672323217348,
 0.9533472656351945,0.19345439554496302,0.9730383702554017,0.768404196342799,
 0.9332433952262633,0.7286129173486938,0.9632319987683197,0.8419202046970071,
 0.4348577051979181,0.5462513099281527,0.297682525262561,0.7724458976164872,
 0.853564416476712,0.34051240113489056,0.30329131650304364,0.7416500865668122,
 0.6272596904694238,0.12530006600191257,0.1414218980240054,0.926309752012543,
 0.42201861194207513,0.933512350842002,0.20268860946062117,0.809635592111629,
 0.9145644976732192,0.9431590029293846,0.7235212294202982,0.09810457500266379,
 0.28668320609871556,0.6230696343725222,0.4487695821530979,
 0.07371545762366516,0.5944167745992037,0.8731742120051184,
 0.44365697801736004,0.5029207435699491,0.12411941732722298,
 0.6892701072159226,0.2714743798577569,0.5049837757248195,0.69268352294708,
 0.1830708558237183,0.2029763538228525,0.20659384106506362,0.4038158817121561,
 0.7143855603635674,0.885845158093634,0.2634365825965771,0.1641366383726972,
 0.9123346491191834,0.17601209525645367,0.14671537646079336,
 0.6984272770657198,0.7375915454184215,0.5411879647196496,0.12043039987180282,
 0.2522958435923315,0.2841017602108411,0.8264658152256172,0.8556563971880439,
 0.6833429731733895,0.590060896993386,0.7624251372766316,0.03445887815649773,
 0.04308203178498149,0.0724355752119683,0.6553915098029565,0.4793080628241738,
 0.007080063424953376,0.6231159912513994,0.617997564532038,0.7241370939974761,
 0.8178371069867706,0.1541247849799703,0.776970207405908,0.9361840002961684,
 0.2994355634742022,0.07850040783794077,0.6510747199678754,0.9797835976590062,
 0.07336991573050533,0.36969889056523775,0.9289334108505465,
 0.31102057829347873,0.6045907287888527,0.09174478161572763,
 0.9723901541491002,0.42648517725204993,0.24290462763727394,
 0.022563345320230965,0.36563207232417694,0.36907114394460494,
 0.4556718501462552,0.2363820151956969,0.23850799984323767,0.6872064543489766,
 0.44283375452320295,0.676600313546439,0.4809653428295322,0.7461312935502588,
 0.36680312355690103,0.4621003620745332,0.6778839603582228,
 0.026742299755778687,0.5457564283698858,0.8530466348418261,
 0.9043779066257972,0.3224222645074438,0.5916242018259257,0.7718295193565777,
 0.6667432479159043,0.586173353623119,0.4465000122393761,0.8480042048499435,
 0.6624853673127747,0.24724148913947452,0.9152136298193138,
 0.029763364520040914,0.5953270393191886,0.9153656201390369,
 0.06119210276656162,0.4645035291293922,0.4633967119752108,0.6219695057770415,
 0.20975746477824675,0.425865571870509,0.9354432469750071,0.9694310118375924,
 0.7180221549208862,0.8736193020040013,0.6766729971330603,0.5911681794952722,
 0.22721913565381213,0.44342648719878286,0.6833468988309888,
 0.25831380806541093,0.24635459905391455,0.6425324184186615,
 0.5886965203066402,0.1749166852474977,0.6553338771343676,0.3357833218486539,
 0.9126165836126511,0.3358783316051649,0.7451259996158489,0.13531621604857114,
 0.8976165739720537,0.32326641241265797,0.8759311099200424,0.8486535900499774,
 0.5497095672702902,0.3131408903949944,0.2983153309496388,0.15312592223564392,
 0.09787698552446011,0.760449238539177,0.02996744097203441,0.9134662389248652,
 0.675449590024106,0.416456742901415,0.5515461362122712,0.1397486890739219,
 0.9569540160160473,0.5245895646227365,0.1358140508445409,0.549247212606703,
 0.9682611687678808,0.6240207909034274,0.3637334989063824,0.36709555157520857,
 0.35344597639428654,0.49521820912771475,0.10510309733525236,
 0.39267327730558654,0.037874694503608586,0.5645608817250595,
 0.3794182839289191,0.5547618149805461,0.8844031088663611,0.3924045973936916,
 0.5707703421117173,0.7470199015038469,0.9290058056541199,0.8321463150361539,
 0.40223643945406373,0.5144056593698303,0.02683745834576634,
 0.5828425469865179,0.1822392728642337,0.5475210658270749,
 0.042428146157537316,0.792044860151641,0.6397337996114155,0.8793574287717648,
 0.045467204365498115,0.053902947648196475,0.4463767819935759,
 0.47654128096798387,0.9075785347569896,0.9386796407088067,0.8324959479040189,
 0.499575261649269,0.3640677369195876,0.9079135594786889,0.8601171297663348,
 0.8643842403634442,0.11059882672053667,0.21920792130671585,
 0.7030755756393845,0.2808130700870928,0.20910261301147148,0.1762276043901787,
 0.07622921204743793,0.10042274949571439,0.3926253756704028,
 0.31274399269218345,0.42517116096243335,0.9600581150692878,
 0.7171837440617218,0.6425601684643176,0.026398734746296837,
 0.7303678473461391,0.33017762523691907,0.9025251889167962,0.3094410501196623,
 0.29790399483528995,0.3460880853857097,0.08917138695744986,
 0.9035804899381116,0.23467837627435584,0.49583252789010834,
 0.5282858747180008,0.0767942457559353,0.42701453079615226,0.9267605436530678,
 0.01926198907272192,0.075742759121466,0.7761828371844516,0.08193387051824064,
 0.19352879840568182,0.6895864698386941,0.09975786582250867,
 0.5244451121762651,0.5220557067655691,0.27877959688302234,
 0.19976870352709086,0.33916962417385166,0.22908810850769756,
 0.1270793469641025,0.7788683979569381,0.6068345899241432,0.6407396325017288,
 0.013270833112528857,0.6312675924361009,0.3866358175892348,
 0.3777278841514864,0.3155880651885954,0.6895883920126745,0.764379962238616,
 0.5021256052575808,0.03839912232318021,0.5932986239721314,
 0.37185264106054616,0.11512278050705604,0.6312089532465576,
 0.6273046476029235,0.46997428578337697,0.1417512812584104,
 0.48316816814007957,0.3366631622545049,0.035165329225273934,
 0.7387720004891397,0.3621539749738719,0.8060957302159434,0.06084099395455722,
 0.7915803729582064,0.5852001838771307,0.507897240746635,0.8769295644901498,
 0.17769548215801556,0.5766096797390706,0.9532279852181951,
 0.41834843375775255,0.47741149755008305,0.7540446530006971,
 0.8389472325141362,0.38839900159878027,0.7153673184589696,0.9389172628308646,
 0.40952960352731926,0.4654398488180327,0.5871223808622416,0.713536538576425,
 0.2861547040289002,0.44300240009095804,0.18019648818518097,
 0.2913675014686744,0.7786939593143369,0.4744850339148884,0.6606007414293589,
 0.1678788246604146,0.2891403652124239,0.02325112186085121,0.2786435858788334,
 0.2270987439100134,0.11210598898576685,0.46153425199486586,
 0.42601730455530185,0.08593114774188981,0.6615400555129172,
 0.5622672505153848,0.8782133693911144,0.8197265817301413,0.7420000186359617,
 0.877288018114376,0.014781002068576354,0.7011072331270545,
 0.37738371482610145,0.5807004504483038,0.4115101609817712,
 0.39041655666630204,0.6171959115099207,0.5426517376026715,0.8931242271860664,
 0.5480233541553853,0.9628431419660766,0.4342707939061725,0.6095225396350625,
 0.22104924645484858,0.5325133702438202,0.2689914949812011,0.6398609975004856,
 0.6959918647794281,0.4369831955147835,0.753771960772164,0.9970479031870269,
 0.13913975900946818,0.35271214682610597,0.7328012558496051,
 0.20550992584199623,0.30237010273551246,0.9247871791841608,0.743314320418223,
 0.7868479568794589,0.07244995260161369,0.0935489120335502,0.6898521156385329,
 0.4542876280508844,0.137735293085528,0.30592556849299546,0.15102223155316574,
 0.13591367392597142,0.7422802061561072,0.19880549987416962,0.780642540220528,
 0.6565607105836612,0.4813836989097894,0.46199260441196854,0.7244650209922624,
 0.03684400725403114,0.09657464903112256,0.9033910785331357,
 0.004598151268198736,0.3870792575406722,0.07546678404770779,
 0.9362432237902809,0.8082889833103157,0.46858946500944665,
 0.14586100036522898,0.9550766053342059,0.9646419188362556,
 0.09508802864830301,0.04169983153051815,0.8858873036755095,0.889435873504001,
 0.19687027525137846,0.6486962818101347,0.6927321373925656,0.6291178438625572,
 0.0763594244611453,0.6446610148621176,0.7304315556561263,0.5009862918133472,
 0.6112153488094579,0.9232360689060328,0.47409963776036,0.7681176535434227,
 0.5678787535006804,0.9619348228784836,0.09187189932261497,0.2669341067127129,
 0.6145348734004172,0.32308062596203435,0.8700331983193565,0.3335399924478063,
 0.24302698493039943,0.6121268762402113,0.12144726107824577,
 0.9489148157619292,0.40691404962320377,0.17219177705316357,
 0.1932667381124249,0.7545968021246175,0.4942143030750369,0.7701993929508536,
 0.3111171584686423,0.5178967820058887,0.38526070040559757,
 0.21368453558636014,0.5710039030690878,0.36824935353393595,
 0.06252213205143697,0.98087961460107,0.31471409299325837,0.6575121485426986,
 0.19388601148468843,0.052963783435190814,0.6970248865173878,
 0.8658641279812418,0.7986241400394917,0.5431477080586836,0.2676237885376006,
 0.2815856845839282,0.05830965366377949,0.7881426902923641,0.6320105376682326,
 0.22872183533261659,0.4984046832125696,0.5931987135316085,0.590178205181205,
 0.150172439207635,0.7278699239552411,0.4759239214667659,0.2297918467095854,
 0.06823510382985098,0.4725840358628075,0.4917106061144296,0.7246792891024408,
 0.13164675629892908,0.5018679621575095,0.5045836707817672,0.9167737582744097,
 0.6461748922078135,0.4697324185489008,0.5526992696716547,0.6813014782267112,
 0.39870607023914895,0.12659383010706005,0.33760955958452576,
 0.1996773127886038,0.9686841670315682,0.3171210991347947,0.0850433465492082,
 0.6247695436974934,0.3357186155391543,0.6679826291588659,0.8329655783830642,
 0.7996937113492872,0.9689699181421726,0.480302385764211,0.7109894425279533,
 0.8499934636206642,0.8249642916415416,0.6017386625443825,0.6338822660210842,
 0.833355550113178,0.9212254849566979,0.4850554838297394,0.768701506174065,
 0.24512899227460305,0.9550508389188912,0.7193876664956029,
 0.29009401888007824,0.5859874365072871,0.7417637017380515,
 0.08180213321996455,0.6154109202430872,0.5758758380929263,0.5578320780527877,
 0.2419306836912648,0.29219382298643826,0.9402433422607169,0.5562292161759971,
 0.038377970585353194,0.056744605474973175,0.5684972755707675,
 0.15360840126522757,0.45910465850598614,0.7181077592265193,
 0.36294190818353145,0.5897432403833442,0.014947809322252703,
 0.7775866026511123,0.2522519688362883,0.12138695242626774,
 0.05679360605680073,0.5853277195053714,0.6981105740844507,
 0.21424941145407983,0.49878570988425497,0.28984204080813136,
 0.9678896320574985,0.43962077604000094,0.6294250407023276,0.8925221890518886,
 0.4599160854869824,0.7337978570629328,0.7346097262623267,0.4709896702239571,
 0.45050821768011307,0.38591474214654065,0.4337599078883021,
 0.8396629096092153,0.945277726111389,0.5679905859933297,0.6910160221695839,
 0.5723869908386069,0.5953082221136474,0.11280605397748633,
 0.43475762898697023,0.5771910539626803,0.1088839945510186,0.9488451636444668,
 0.30113876890462954,0.7259120775906818,0.7786139474607602,
 0.23125513962075606,0.8662876895282442,0.8632125978812247,
 0.048625191829929526,0.571659538478912,0.4294948567840071,
 0.26079697928120305,0.8830589105535952,0.49632926569433056,
 0.9024136711358578,0.20092719412047655,0.2658120546612227,
 0.02676748100685533,0.9639891230912669,0.7965180100367293,0.7712230988745142,
 0.8658894478423943,0.5684117777465708,0.07735382064897012,0.9177512771302111,
 0.359537653771801,0.24172173036163702,0.6636040985847099,0.8209212579086729,
 0.3935051402501881,0.8104094476335773,0.31098158286965705,
 0.15851139333989694,0.40612679686047315,0.128196562389014,0.208056387021464,
 0.11664319763009257,0.2567298696498348,0.878790031718587,0.6629907756415018,
 0.8821401061496744,0.11792215513284288,0.24704166967307695,
 0.09045428112566867,0.6132826712420378,0.9654007958590042,
 0.28736346510944377,0.17579308428056595,0.6032724010908901,
 0.0754005385737907,0.09221174919831787,0.04210155179785613,
 0.2798654909185292,0.9489344693279956,0.3860924130138811,0.24514976215858886,
 0.6995859259451895,0.12620513101084496,0.4467003262915864,0.5056762701970299,
 0.5269006254910001,0.9577800229682963,0.5202899251033053,0.765346985641665,
 0.009051406271296969,0.31623466613700846,0.29579942077982513,
 0.9796858151116694,0.1600569230741612,0.8487126320594043,0.7074166720194415,
 0.8329193430291869,0.24601736124221818,0.8457149638197426,
 0.36050940058424585,0.23966877137202902,0.3840461984328628,
 0.5355042170128633,0.2672481209874713,0.47894327288210925,0.9928848333025234,
 0.5679542129679698,0.1504037208813782,0.8859718986430192,0.18083507869657967,
 0.6965359805213533,0.8743495330124746,0.08768264542788695,0.7820888401895477,
 0.5041526410617768,0.9716630790390401,0.7624778440456688,0.7905037866239425,
 0.9243873884103951,0.5532346596139074,0.8737914818824709,0.4051647175905424,
 0.9143782723146321,0.38299649371932454,0.17508046105471609,
 0.9036875207732146,0.7062222401512689,0.3117721975794837,0.8289995686311863,
 0.37810040461622174,0.863570903208971,0.8999717207293731,0.3933482352355553,
 0.5515671814308488,0.6351063993719255,0.14739908490134002,
 0.43574625752308915,0.8738677581231944,0.8481798689286835,0.8466184816165198,
 0.7218119089421294,0.6154086830825949,0.8548501667977461,0.9536928835357058,
 0.07453617942489443,0.4994746978611414,0.6409525994391239,
 0.14727677072834622,0.9859379168306992,0.6318782420546427,0.227715875785635,
 0.16737197804908055,0.781071178516428,0.9244928570687917,0.5462979257276314,
 0.9295977561326352,0.6363751615210044,0.41377788047229735,
 0.11115523053187881,0.3901413078159832,0.619786093052882,0.265560030015794,
 0.9271569774313773,0.7549253043068357,0.31992247300986665,0.8529273924857491,
 0.5916792453056496,0.022288782084777692,0.3024577456187645,
 0.44610867421967093,0.24449480209028474,0.1779282441055643,
 0.7415563687996052,0.05884988423483939,0.14070903591229744,
 0.8952441555140956,0.6740097446760183,0.11196765174348755,
 0.24616587966173964,0.47220779851227845,0.22659735991052044,
 0.7667319656840567,0.3882574943561159,0.5877915619589963,0.9410045990141273,
 0.5902826027724971,0.3845297256318738,0.8404032607533226,0.04638828041513954,
 0.9896354301276459,0.11582873814009631,0.7855679376844465,0.5497471302584527,
 0.7465755649341672,0.8094529759879174,0.3731839043477949,0.45883077978736375,
 0.09051723189180783,0.6828715068339231,0.407626723004879,0.3007739296041273,
 0.20263531987461203,0.9265590181355547,0.5539738830410537,
 0.05615874125832032,0.03975083741071239,0.44187198168531217,
 0.8038669013842634,0.06259178444452207,0.4757358573749233,
 0.15470581538833184,0.3626901176525592,0.924253571348384,
 0.048296016799412866,0.7878217598947332,0.04562781416082895,
 0.041963505903205744,0.0060765176809147725,0.8104767745355081,
 0.8279780884274328,0.6072753985183039,0.04329047180658174,0.6155759305246656,
 0.5123275932570276,1.0417557540676192e-4,0.8563095789935121,
 0.494269753302239,0.3344818263204814,0.3102977247497919,0.10944637917250599,
 0.13436728551280597,0.7326524482648554,0.7659240946292489,0.7670818281762128,
 0.5836539563621176,0.7741093224974362,0.41567471087673546,
 0.42473458401001807,0.45670324729652634,0.8285374091029818,
 0.08841381340044496,0.11587895847090046,0.16409927645841693,
 0.8316569675267795,0.2630519617516971,0.7295310263391266,0.33983217956511647,
 0.43964032185668744,0.9034294323925975,0.6246511598421619,
 0.005336584352555129,0.818040803451089,0.22393114156283023,
 0.23635158930150646,0.46137814377733166,0.42724190002822904,
 0.7847942819840892,0.08819718778523367,0.9335499646946668,0.8849340249871638,
 0.11009359518581452,0.37625043028904903,0.65663181302543,0.39737652823639913,
 0.34064978797110734,0.5437246987792997,0.6498296610482421,
 0.07108656069365271,0.22998043285041836,0.6805888951355893,
 0.5237303018170364,0.9276816771719368,0.6680713904653728,0.5755174348671109,
 0.32288037701057026,0.8927993918865785,0.5581031174706014,
 0.07370261414624657,0.30332930619465603,0.9950048724458761,
 0.0845307789759101,0.2433148347367926,0.17519305297279697,0.3383836312227221,
 0.4763842574438396,0.5559602474156835,0.42981151580111365,0.5505029523791815,
 0.9279206846292974,0.1133970875866318,0.14444424090087893,
 0.21868732289231685,0.7033946819214169,0.22981830745963772,
 0.3723917804186696,0.02384805012709168,0.6477859298961168,
 0.37720761593188523,0.5127653431624779,0.1003042621720297,0.8290232744899864,
 0.2838541345157648,0.6129359458033687,0.8154915480090202,0.2608837734082382,
 0.6582611284353143,0.4542228776397399,0.6442840254670902,0.04510390934453112,
 0.2258719568368336,0.9336243347948302,0.38679873770204676,0.6370076111866398,
 0.06878058100323903,0.029645857263896014,0.542612614081526,
 0.5224595179944841,0.5541004017266253,0.40656490878181184,
 0.09205990167573241,0.8873434730327207,0.6101298487003589,
 0.14144697398701112,0.6459313504461152,0.03271539087118036,
 0.28537761765011305,0.345466065072652,0.11288199913910701,
 0.007759296304338137,0.8213101909234329,0.4936585124854612,
 0.4346427741328188,0.9440764347217819,0.5942543335907403,0.9282301185828505,
 0.2867159035839737,0.7578340007791607,0.6396989313506781,0.9917932618672668,
 0.16474600893084212,0.08607298738735314,0.30550031469546846,
 0.2168496167310484,0.28512651371355713,0.663091060964265,0.41354496988823586,
 0.8593565978147245,0.9044967963696922,0.959366741961736,0.6494345157855629,
 0.23542211153822068,0.7009036794219178,0.8508057414353408,0.4422428886366069,
 0.28732578182358526,0.38331697027671063,0.3835636350040714,
 0.4418341542839357,0.9015303586607826,0.5718363031353507,0.566050311489071,
 0.6605624508421054,0.6799414033670476,0.20461275548143687,0.552641099324588,
 0.9577892905872908,0.8600829385764067,0.7673089633854665,0.39971893624840327,
 0.3817300020368175,0.7116164389077422,0.16672520985122752,0.8368035697507105,
 0.8606936386571461,0.01261280543424681,0.12915660625611292,0.371321547854468,
 0.14005231295057718,0.041349307075272956,0.6105891174972108,
 0.22974032618218476,0.6400826424241471,0.5319882075863126,0.5484952315192654,
 0.8553736621992655,0.12994446912598023,0.5973304557834177,0.7861655404330764,
 0.5808293574793986,0.005789234360904838,0.1458879722337011,
 0.09311719678807384,0.6540616010780536,0.17618444507092035,
 0.09605191052735429,0.9608890015170805,0.498917462264894,
 0.019688923496838928,0.1386262685390489,0.2980684002349404,
 0.1654462712373027,0.8974685069009208,0.8646807568778828,0.7395737352245441,
 0.9070862340865876,0.4597641943217816,0.8272338682723351,0.351347943599281,
 0.11964517257805796,0.24169143279251515,0.9774782982881554,
 0.27982171740332396,0.29745367659529864,0.7607822377183424,
 0.41279565323984124,0.23782392993512846,0.23748309572804938,
 0.048913937378587735,0.21565365604393816,0.04026524709171375,
 0.325456824700347,0.3296999716437099,0.28877641572354606,0.5399024666625932,
 0.12316172004696457,0.6663178146496332,0.825259682499384,0.18891386227372375,
 0.5898475880755043,0.3301488823467139,0.003135453447744463,
 0.9484085267044656,0.1703741963425378,0.802095857301995,0.7238220972169721,
 0.7478435347403439,0.3461042581384598,0.9549834884216121,0.7790707780913774,
 0.17668157878714497,0.6155095050397644,0.1304592062894031,0.5155474017829725,
 0.3520464597984124,0.5805590427486779,0.9348931840294077,0.9713436251784401,
 0.5540733875353635,0.6832487570619781,0.24088854415432592,0.4402475333948934,
 0.29738292771756003,0.3920480453912346,0.9781551524608297,0.3909405397698702,
 0.4719620528224757,0.8336370087657072,0.20517887793030432,
 0.43926124925792276,0.32599921450724834,0.6947014669455275,
 0.21138300217446804,0.6216240958703058,0.6441441836031747,
 0.35447936513669087,0.3860019166270572,0.3478153540507576,
 0.04704138378337552,0.08817022236220928,0.4902001759979978,
 0.9638045819651023,0.18719278035523013,0.16321376441446533,
 0.9647495847200689,0.7492339798100651,0.7029673987134369,0.5299905493420695,
 0.7873155399692733,0.17226208097845863,0.053516288838149295,
 0.4778925264241225,0.5301447156502557,0.20520992925490433,0.9843307077013704,
 0.1730105019989422,0.966675834987979,0.5364681769681083,0.7941736821010285,
 0.3065021625654394,0.2913172740146618,0.13042281100627118,0.7766526689844646,
 0.024191828889725486,0.9422894369052885,0.10723129880183047,
 0.09397086255805143,0.6071253846133094,0.1412901379715158,0.6878188194277598,
 0.2628468436190403,0.2967434444604877,0.27816482413856425,0.7776864695966801,
 0.42357125758295644,0.7842266445403357,0.31077918516693614,
 0.42046182189730474,0.42312326012302903,0.2490942007418998,
 0.7904189505543319,0.474603893470798,0.6789240292183732,0.1761629310968571,
 0.16768472185249161,0.3919256514824083,0.19283761282506828,0.682882392504482,
 0.42364218960511524,0.44199173177596185,0.6676898742851713,
 0.32137179667763793,0.13804272301452125,0.5718109100731027,
 0.34051617157102854,0.653018244933432,0.8812565755402648,0.9517470362098887,
 0.46918806697664284,0.4108093192432265,0.0912298181827138,0.6599896580123318,
 0.8840505269960247,0.9349664926118666,0.021293424145984896,
 0.36222309262523744,0.7840868125016217,0.7361671163416407,
 0.37957316059953927,0.8421943668788192,0.2655677559631886,0.6091661322281714,
 0.6743769689031278,0.4790984002915315,0.9265143056094827,0.25401749963476905,
 0.10201657301757106,0.8147347018930979,0.92944026807952,0.6280499550115128,
 0.895063641385715,0.36009583176337034,0.47364529689758883,0.1719551416173566,
 0.34087059447004275,0.1800682138733759,0.8580094978296694,0.2950071871977965,
 0.7569950945469293,0.27295971663406526,0.2897078562906872,
 0.11478790897132551,0.7752515748426572,0.4644054353962901,0.8338621526392924,
 0.5974658570419834,0.4075665296939559,0.744430867015983,0.5858590450962483,
 0.26037682884765156,0.5507682970190544,0.7039437702800746,
 0.46565991293363085,0.38021609729339545,0.937950682584253,0.2613859642857146,
 0.6529043538575625,0.041463372455607606,0.8695779125764705,
 0.2550374560134969,0.533102163063691,0.800846645719617,0.39466492632062855,
 0.2031901334204399,0.6045977674517622,0.5129768296383165,0.9801025982315085,
 0.18947068973098324,0.5137479309099675,0.06908635169374344,0.829073148293935,
 0.6609463016489876,0.7404405242591673,0.39307088082874764,
 0.22091978676679325,0.7102399579738967,0.8783533349143436,0.8169205386730969,
 0.04518446944385335,0.3462847552397841,0.9377214416325487,0.6292884705447988,
 0.07825253656329068,0.12101216030661499,0.4546837464354895,
 0.38624814886829595,0.8924889200250328,0.381910002399099,0.39582178646400423,
 0.9898785195862441,0.828709770977637,0.6239427700402788,0.81042446779363,
 0.7153812734112005,0.13783410117349015,0.25533079378968904,0.307218377796286,
 0.2849888560780438,0.0752162668409182,0.31757879999481775,0.8486654993840256,
 0.11575052118047857,0.5224567042438718,0.6422245052000153,0.5731657501105138,
 0.869304005525222,0.7826011396050991,0.17627912512254784,0.7625035812863158,
 0.11914916202345016,0.3300947043234025,0.17896055563969115,
 0.5184858062177158,0.8012366040760841,0.9761424402390396,0.32440604758010305,
 0.12228344356097354,0.9369857075495901,0.15632151507963565,0.906271109494355,
 0.08622046296776986,0.812014756810504,0.2167624980950298,0.5887842825051,
 0.69282023416224,0.8257042210145433,0.4839245187976535,0.24827658653093487,
 0.6553388003590743,0.044687077963551425,0.1581823773754365,
 0.2218801592535653,0.31060811528861065,0.8426499045919655,0.5850275997118727,
 0.21295156474092458,0.3155894871406195,0.2248377280483893,0.9122959930036578,
 0.2851579231915191,0.18467429269381275,0.13096207835933593,
 0.8990922855341698,0.9844686337112449,0.003912850750020236,
 0.8645669489981387,0.6877091010898613,0.8917656373317406,0.6092611939011705,
 0.8332086076162053,0.7058117284650656,0.3417022212660279,0.5615587709362522,
 0.2675276498221292,0.47564124914842376,0.7579148523703383,0.0830703573094077,
 0.8554093658144585,0.5162834179570593,0.13455104897539205,0.6094447959245044,
 0.10746673624783576,0.5944239195561405,0.8422807350631898,
 0.24210295159839978,0.39636974435922434,0.05519620115277579,
 0.647025729067557,0.16568286534919874,0.3437421136819897,0.5255511715330554,
 0.4950292297138986,0.21323988689238593,0.6386435542321498,
 0.12983417285160093,0.3163216360077258,0.8480050206015882,0.6191791696200308,
 0.5498143139845789,0.4985785632874644,0.12942025036079785,0.690363616703305,
 0.8393414566660827,0.8436517944158773,0.25115860806656987,0.9804630481460169,
 0.525465210996938,0.3522357234303697,0.46968075374629503,0.09484217255015215,
 0.6879725262169296,0.011879821758311926,0.9972093835799589,
 0.4635755232081673,0.3614516926547582,0.3608358984658646,0.16781020066171204,
 0.280864970579623,0.5836804450915919,0.9859327926676072,0.7805834942840411,
 0.5416959451799682,0.3881640640083953,0.7909436386650481,0.4412301491996734,
 0.4052452633491318,0.376150464855178,0.9347567451544267,0.2809454861070908,
 0.2489426924299113,0.9735247648456529,0.4332258793027509,0.5107404571251607,
 0.9719997276512078,0.0043462892595709235,0.43889556832943977,
 0.8560757367706023,0.4937887548113925,0.1153176497456252,0.2750399950322212,
 0.5463846306218775,0.26735866817580867,0.28694462345030836,
 0.5224121076828124,0.37732559333398663,0.2869522668121318,0.8903585225136231,
 0.12235081325153996,0.3599207598748797,0.808760670273626,0.1912946633481416,
 0.9296743893649432,0.16857406237832806,0.9898495441536032,0.8894525490322034,
 0.44336689986005495,0.07174818717475862,0.43616041165443886,
 0.11772826917442858,0.7474260043324227,0.5068419293546917,
 0.15609496033126957,0.30843647960379883,0.8068661615413177,
 0.9061788816793327,0.3370639080105591,0.6723451477649435,0.5453646323420172,
 0.37091168318753254,0.9753956323284149,0.0014452225049779166,
 0.021465479094653128,0.5069893850073147,0.18205605425286897,
 0.7424951821269373,0.7573655574085677,0.8484743406211045,0.9506456030637198,
 0.9479689020204474,0.7966754804570109,0.9003108563220182,0.4186246808788763,
 0.7471992861196096,0.6669070956796566,0.6696756152118397,0.9651122357096308,
 0.7012795919500319,0.25087350816144705,0.5398115994341556,
 0.028690354384719563,0.13902296051897178,0.8082140837476366,
 0.5878958445524309,0.3948075189806589,0.03829945449422034,0.5034699164502212,
 0.20881913254155315,0.15747286887121115,0.026578059087886574,
 0.4566471249845149,0.7628183888714236,0.2314488414273459,0.9109175919689755,
 0.4365309509405225,0.7687221827640106,0.7627511590626226,0.43563728495069176,
 0.9053299362422476,0.9392359770937574,0.32864246983593026,0.4341187812528481,
 0.39276283839648873,0.4807137792337963,0.5289505344084086,0.7271785654925454,
 0.6949579007872523,0.17166444662484082,0.8197598266878316,0.3122006552630793,
 0.7026004693812404,0.006799166525935929,0.43579018920296264,
 0.4820775589200228,0.15635671294815778,0.3966628152422924,
 0.09262545042852832,0.6033935841655897,0.6105901734685372,0.9188109866660356,
 0.5620487062926416,0.7051326159632876,0.458726661424817,0.3312346810310096,
 0.5641477094258133,0.6046759489112333,0.9342014830307631,0.44181381566756284,
 0.8100036744228176,0.6848895581614054,0.18592672876733185,0.7397476070124703,
 0.9431656211879622,0.4204936110877219,0.8457557108243409,0.41340569281933803,
 0.03948962681408652,0.8223451966593335,0.39178013283789614,
 0.5016844973968215,0.9685535506269418,0.8105556648189499,0.38858269808824175,
 0.035693604970429726,0.6474984742993295,0.8257082557304674,
 0.5655792575862674,0.6601721820638116,0.8929965942828418,0.14754073935696776,
 0.3577550581784742,0.40581543068134285,0.07111350578288,0.24276901610903578,
 0.8022513593994015,0.5374318221901783,0.8724325108302488,
 0.020316623748590446,0.6095585674977971,0.6460624578530357,
 0.6506883081022227,0.9344657829082177,0.2685398655682585,0.9284125478790105,
 0.7672095153639339,0.18199639069143314,0.20133712460685793,
 0.20069955130024442,0.583172716812756,0.9142333764233601,0.09589492930225918,
 0.9537665891892892,0.5729900499368998,0.35870871297447726,0.6157028643348504,
 0.6383424258368092,0.23755907549243704,0.9183402465697013,
 0.45289117071428997,0.23047664905867815,0.26372425767411667,
 0.34605874961510397,0.9685429350983082,0.9171881840711668,0.6916418767568344,
 0.07050140351393785,0.694691909245368,0.5879977470612072,0.9784348424660441,
 0.9604157951974541,0.3051747858415603,0.22427134358952872,0.2813110134013652,
 0.8557859155798218,0.41725335786860174,0.9256331813260031,
 0.04394851870142635,0.4627697348807296,0.4899549577856325,0.4882272687823057,
 0.13962736944276155,0.7618546213919497,0.4826157195618793,0.9510664221078176,
 0.44169384411744095,0.08213649143308288,0.8101111611287081,
 0.3729314484790225,0.19262576660769848,0.6627474912523059,0.5346162690637565,
 0.9652335564813886,0.511688147478997,0.45270315839942277,0.15409387082706905,
 0.9695581036552794,0.8521382811664303,0.35566908683650966,0.4811950117227827,
 0.919904447339734,0.5324564494991173,0.4014125428320051,0.7936686174746477,
 0.2510431791472696,0.7543537666876439,0.24509378461975107,0.7985927292855707,
 0.382271884944121,0.3740560454207027,0.8887779792225616,0.7981796987459517,
 0.5775905348390596,0.28224384052684537,0.7948706991782777,
 0.30484498471113775,0.3610729704005424,0.07075779139300464,
 0.038401085841399896,0.8685881029579663,0.808689884032127,0.5859355472428116,
 0.7317536978995701,0.5147873225741626,0.5135691235176709,0.5097009315789439,
 0.8702452236312211,0.9478465207293233,0.7272849667489535,0.941903077529852,
 0.7703288360594592,0.3994105518026203,0.07123383575608755,
 0.26275862402769123,0.8992695081791873,0.16795183808086767,
 0.7848517623671444,0.9746266347304792,0.08315464150930585,0.5186296896601734,
 0.6199406718881664,0.41303048299713063,0.6145357473773181,0.8869125178929654,
 0.7369476017435883,0.07042145590969084,0.3327482903340597,0.1336891661157823,
 0.013270417180729077,0.6707069076996927,0.5144233779126107,
 0.19449376415645636,0.4021100188480745,0.8561738871191977,
 0.18386031468109765,0.41569582364119506,0.3305999053737605,
 0.19610221831962593,0.40333429839183665,0.6813599725645737,
 0.16458194654765546,0.32072783445546105,0.9222632956377027,
 0.9342522337757254,0.751875560540209,0.17758136988747297,0.12981242171429186,
 0.6591264500273697,0.08034362325462019,0.5843717045544449,
 0.09596854168066049,0.9450259826705872,0.44809999752550256,
 0.038193979085706764,0.20805097537056538,0.933444590960747,
 0.8852326123494405,0.21265773954634692,0.03403074553110663,
 0.1944451698870998,0.9948762470022512,0.22746753521943663,
 0.18084016560097038,0.21998228111765517,0.6688141300690935,
 0.6527373022967173,0.14508905837502228,0.7642658967147056,0.5728745618549791,
 0.7521878539166003,0.936685248787249,0.944519603557229,0.5046566471373432,
 0.7801103560721143,0.12929461067760661,0.3835802630489784,0.205109191043361,
 0.8258895397054926,0.7149746577820211,0.03574775259803631,0.5696663784290719,
 8.425428548813896e-4,0.48541713802602904,0.4393412425883193,
 0.11781317820170512,0.5092179401118003,0.41711953818682734,
 0.7652454622531203,0.8335367372236343,0.7095269295237897,0.6349341377320308,
 0.7196690223211663,0.8036622972837743,0.47434787381483334,
 0.03131519411352879,0.045439614993908806,0.3755900769215824,
 0.2800815069550989,0.733012527699149,0.36145927758628726,0.06060035577889855,
 0.3189273496170142,0.8790838198256545,0.008777125661168284,
 0.22315451517145646,0.5515184887020336,0.7409125735327036,0.3106270454328376,
 0.9383130536480616,0.08108293364073016,0.575352054489508,0.9288392712123402,
 0.9582234716514366,0.860326840747896,0.687463880314771,0.8423787089733402,
 0.017219227818508265,0.2408531653205569,0.35972902962155306,
 0.2601923564263102,0.6846479912671469,0.2756592669299589,0.909186764968932,
 0.4702056437198835,0.6724135404759647,0.8099091199577124,0.1880465938195034,
 0.832177938818353,0.2557847097940328,0.22698068451867048,0.9622008405156255,
 0.7173668727765627,0.7538681732395306,0.7048733383289185,0.43564200423058386,
 0.44672069122063696,0.973070020780566,0.42130333124581554,
 0.47460653198300806,0.5077899552971107,0.39566563857540826,
 0.9877324224006682,0.5999256889387354,0.7632191939871621,0.7154222858780038,
 0.7703577139090663,0.959377794745927,0.918514534800349,0.029271300154019242,
 0.36232704267778326,0.016851324453000527,0.0992082776278651,
 0.8100034111835838,0.4418299948022062,0.9101981891172469,0.6033330151244104,
 0.04117876916526564,0.2056535865929795,0.2960336000822412,
 0.019553870146217234,0.31670159060697056,0.9025119195147278,
 0.013202334304465269,0.07249991947250711,0.11479665526528049,
 0.6813681216621883,0.9093122167053838,0.21921866581575522,0.950292683001615,
 0.8996582069485137,0.15796560905937584,0.8281125808998929,0.8993479912358727,
 0.9681131401142357,0.8743128833250434,0.5874779029707523,0.9113356844073817,
 0.6380594803115847,0.8952700620962372,0.9847248914009474,0.41611894885461165,
 0.905561107389569,0.17135271222871085,0.7176009549329102,0.9107903146708501,
 0.9326930547700926,0.5175992281443166,0.20817841862215214,0.5869703458530869,
 0.4792339746389789,0.386239243056855,0.4298686385544981,0.8154218637207873,
 0.31356824998130106,0.9929670290067528,0.4757512140894785,
 0.01470512139511182,0.005127388998618865,0.21546605549464315,
 0.403088484178176,0.037888458241838174,0.9679696916512605,0.6209954834547208,
 0.8682142376897694,0.22399224813914986,0.6630609286513715,0.6071279534971943,
 0.5715665776741068,0.4083376263063274,0.36858739504653903,0.8211604174330475,
 0.9399972334464601,0.43175300916700887,0.47147253036521575,
 0.7472827649325915,0.9828910982473534,0.26881841309587584,0.8365992590485457,
 0.6749958305735712,0.7615171173196916,0.10276137950355568,0.9408343332752699,
 0.9561608197876998,0.3051295853405964,0.3499630521503472,0.5247201225897064,
 0.6751950062414374,0.7286585156977087,0.7514249847704412,0.7328315650271847,
 0.6565047476084382,0.5350974036583666,0.883264181818388,0.4496648773521532,
 0.6894932572257406,0.5086195854135194,0.34968279218963105,0.7072560106414589,
 0.4649944884663445,0.24738783076417992,0.8692727706605917,
 0.09314818863707186,0.35370510312325976,0.4683884596682252,
 0.2381838322226636,0.6979950785356457,0.5989224159841966,0.6673662001904801,
 0.8231212198962479,0.704896345965699,0.5045700258479132,0.12157498855430582,
 0.8347879858331131,7.759866926628778e-4,0.0062264540093741605,
 0.9232204333711064,0.6999429210430655,0.8011238240135315,0.9852530432450396,
 0.5729094407945184,0.5389279855431741,0.3812812198353821,0.45141426673221685,
 0.2371737644038293,0.1233309971300125,0.02973182126191909,0.557458261498417,
 0.8194504313145846,0.3152652565306484,0.2590781672655401,0.07591802146415838,
 0.34897242170820353,0.09151751081599324,0.26404969442109216,
 0.641295393771156,0.05951120295149426,0.17496830809570285,
 0.29860901855070976,0.24972312797188467,0.9250861919464298,
 0.44612198170343265,0.3607776419187345,0.825246906264213,0.24109830935126253,
 0.2262194468451213,0.6150011047659337,0.9931388747731091,0.5868463564048484,
 0.24554888228530836,0.1754667768122461,0.4912351468637881,0.8195380535404071,
 0.6883169844729605,0.49167776531040897,0.3791512832206828,0.874081639288169,
 0.34931662940466957,0.927373991012632,0.8815485857894618,0.9956406143146271,
 0.8213176531841304,0.7370284704452995,0.41047088041085633,0.8396342243980042,
 0.6687607894388211,0.04878542567698829,0.8922071720890082,0.240764987680806,
 0.618799098357515,0.747920651968306,0.8896711398208824,0.14304554164287064,
 0.5561998211895524,0.9397398323260013,0.1193528216133165,0.5828228770724868,
 0.6062764237556566,0.7265073019540419,0.04822260456261718,0.6581468446401373,
 0.7800207936500688,0.9464307059785301,0.9419218826562271,0.4058624486200286,
 0.7472630469802288,0.1753935356791022,0.419032642718702,0.9286145031497999,
 0.13662204684889767,0.6737763228113034,0.8929286942860131,0.0770124256959216,
 0.2625415282688187,0.5764326874021681,0.14265651033614724,0.9923085973011617,
 0.3526360218423441,0.5436617189778005,0.7279800309963316,0.2358725129227659,
 0.5580559177943915,0.4237338715744693,0.7682996072126324,0.5148856836735048,
 0.47986064734957445,0.9318309653092187,0.16139744954355162,
 0.38721936195997375,0.5677363848185506,0.6735776836739024,0.608768992396425,
 0.30029557895846526,0.29663721767895335,0.00655240376126065,
 0.037401376845971,0.6980077773315104,0.23339415712302136,0.41262208811640677,
 0.6532996296176904,0.3940694199195127,0.6147710629079749,0.47466919980550315,
 0.7085153376160274,0.008078498578597437,0.669463169694251,0.5126437413763889,
 0.25367832225070586,0.9406560712794728,0.509667069564324,0.5393678200417177,
 0.0793530946322093,0.880144198194888,0.1928462667942017,0.08760220283789599,
 0.8825000809305223,0.8264084729862953,0.34724930962899647,0.9280280864729251,
 0.40186974383648777,0.2801958255530167,0.2513988081066655,0.9495061525584083,
 0.8171135072759326,0.8800785371687648,0.04801416201906239,0.5213697112166622,
 0.7158758401100597,0.24449883084644275,0.20022747449386724,
 0.7935988113556013,0.19190512303622598,0.7824680466948348,0.6425930553554052,
 0.9857464039365296,0.7779089798360636,0.704212378610468,0.7060586716076153,
 0.786509551614726,0.23590023394544624,0.08561237384864195,0.8493385137425946,
 0.8359547885360674,0.7745301249089745,0.44561717714275495,
 0.37792576764132924,0.19611594223262185,0.13664222182408914,
 0.9514552330611925,0.5995361640417426,0.6663184120613821,0.49518546427785204,
 0.6820328173131376,0.8831656665851745,0.5937049407618865,0.22754437277319373,
 0.7783508444269611,0.4922995935790241,0.8551499474410075,0.362815671553522,
 0.325332072338274,0.2596440315475501,0.29002662023114456,0.356835964013667,
 0.5294245624118182,0.42797642425767013,0.38238995450704505,
 0.1073734442341987,0.34562818742846657,0.03468994787393209,
 0.6584938875553477,0.7454412083917012,0.13415273943574502,0.7464499871939749,
 0.31881944444664767,0.28157002026323696,0.9446747578418984,
 0.3498465138446112,0.8650777615748558,0.3530196432333369,
 0.029750292758116248,0.8965356786159322,0.9808631854428544,
 0.13021830650201405,0.8793407609229912,0.514477313780013,0.21025143989594297,
 0.8997076929352769,0.5106465909728417,0.681761893485828,0.9080078158980024,
 0.29580766141883397,0.11660791944257753,0.9745086015952148,0.718489873910576,
 0.08925779694534741,0.5676591286610563,0.34828060081636425,
 0.07860633307567844,0.43453821343381915,0.16686561931221044,
 0.6718330441444952,0.2753914359752264,0.8994066854841863,0.05935040139300063,
 0.25707974459306204,0.8819677524860007,0.9456573010181122,
 0.06943831563751901,0.6146726052353082,0.25292590439989526,
 0.8125989522400787,0.4872990451462411,0.5670689318855802,0.47888824670200913,
 0.2525293488979371,0.31200044237387725,0.5914551977129997,0.7429340111369459,
 0.9377622521664073,0.8758020084316613,0.07086432600040138,
 0.17304771910070815,0.3815628855703086,0.510045070787425,0.9293941824156116,
 0.555598873743982,0.3429513423328564,0.302976887461212,0.9524431844949676,
 0.06806476296272257,0.7235695374798669,0.14429620966237122,
 0.4991246073844381,0.511512040315216,0.27897015979451156,0.7670520060995213,
 0.6014094739390008,0.04447646720134191,0.5512120010536986,0.3337358955237305,
 0.6322450316447565,0.6950901217108103,0.9899596838266588,0.3605135959268929,
 0.9892792898736527,0.33780968433132796,0.612697153115174,0.30014639429535794,
 0.07354811450590548,0.018322805045681068,0.08695882011696987,
 0.8495199259069925,0.22392693237663175,0.14650310468779426,
 0.7800059671887087,0.3518735839548487,0.5687935297338551,0.13273157615594977,
 0.1339349226599087,0.13213728976278938,0.6563493457013903,0.5303460079714782,
 0.8309608243703968,0.3449241475800615,0.21306627448330095,0.5490554697567619,
 0.625575568317768,0.160954482088985,0.4277319753045479,0.255411229730381,
 0.7724878575453038,0.8446193584970463,0.28027335662619324,
 0.37783548275630463,0.9740682908409323,0.6807941036411004,0.0921180261768324,
 0.921568310979536,0.29979114321389544,0.40037182203961597,0.1463889708058289,
 0.41117150507937483,0.6812590864344711,0.8262390707394176,0.2214434989172469,
 0.5288782189290093,0.2902567611688365,0.9639728477317727,0.8645836303047426,
 0.9166685571309022,0.3211627880353032,0.5546756967123437,0.08874023898179673,
 0.8390485923107118,0.6481944390251599,0.7943240242724657,0.7511863516134618,
 0.12372329294829032,0.9357958132468853,0.22186847039738944,
 0.8225069749181677,0.7648882174753647,0.1078648644691017,0.37220043201484,
 0.01928763560486757,0.7435984530268873,0.01365790195580241,
 0.7839186179034029,0.42123747975244163,0.8875859692295957,0.8171309808030764,
 0.33036466269426346,0.13543591910084274,0.5252545909929708,0.675676384287712,
 0.8021685305005317,0.9536307290366925,0.8833085432923485,0.3628478914919435,
 0.12516180361677864,0.5401586967571002,0.23265233984413447,
 0.3175950772932472,0.5177904525389793,0.7235437644655006,0.1383050203160321,
 0.5583097116250993,0.29983017385140287,0.7538910994662752,0.792858786551652,
 0.689770171885778,0.4307432869662725,0.6473438062485588,0.20586729820493488,
 0.696439399272422,0.639186527602405,0.3223107971637459,0.8937483082281763,
 0.3232482634896877,0.6649332281753565,0.42316075474632886,0.7031760455269029,
 0.7515098531099067,0.4111524517776681,0.21709728780250614,0.7334750567644784,
 0.7790338332855784,0.647328138440084,0.11502994479030693,0.5496040067734964,
 0.132099046571183,0.40025547400184835,0.6166335820217642,0.5462517353841584,
 0.38818850002269223,0.10956750579488383,0.7879081141503638,
 0.32018457617676666,0.03196967518196625,0.9751532659445243,
 0.9861274354040266,0.3388915143578286,0.2684250646910429,0.8449299448968676,
 0.838720602769179,0.6328883266073325,0.8709928739726265,0.14743753698841255,
 0.44907643720330914,0.38074548795049246,0.3091988557453631,
 0.3536015881826945,0.006560839932996387,0.6659985197330034,
 0.6890398656956053,0.6600998667604524,0.1220273904737702,0.8904071278642617,
 0.76371270929835,0.5748325308768681,0.6144370203327686,0.9703774659185354,
 0.9202996797589376,0.45418321909941817,0.8753685651995449,0.6553532406526024,
 0.15319371313952945,0.646935620157335,0.8844449305853536,
 0.004709324117066593,0.47036645007152356,0.6145254749507975,
 0.10621180535929531,0.5907121143783569,0.9636774953128674,0.9268408016764004,
 0.614542554838681,0.9654412262113283,0.6926180797833216,0.21192791992521276,
 0.538407604698493,0.2102397360897691,0.015070945671448732,
 0.04859418302084528,0.7233338200145792,0.9629317752876908,0.6122287305843629,
 0.8370890770748167,0.23122747940759425,0.6004168461084504,0.9473449074994411,
 0.4509952090959184,0.34565781151715114,0.39393713007539366,
 0.012897579402516346,0.2690944457092642,0.5625221150015873,
 0.9506847383234027,0.6828445044010008,0.18483270310135047,
 0.27823332960292624,0.404295022707851,0.3554109199189962,0.7152299027376676,
 0.5247093654246199,0.09293385968353995,0.014148292557704112,
 0.20559734959195375,0.5628174763294924,0.5647504157169927,0.817752432099956,
 0.367118308249892,0.4901301893082386,0.3255799072705413,0.1443984664671667,
 0.14465624217007855,0.5837433415956589,0.7231588376581237,
 0.16673366717908056,0.3573957723422092,0.6871849994542854,0.6271431250686691,
 0.028600218970110558,0.7016311505662481,0.6985118013912798,
 0.2933478949884212,0.9835449249192778,0.8375344526422331,0.26656441472672676,
 0.7603170828094041,0.09393235015282864,0.02167038227279605,
 0.8013388196610499,0.9462492650477035,0.04430759309968335,
 0.31972784043964564,0.2636586111248237,0.8261411906491968,0.5461202681175708,
 0.8859544503396009,0.03764533486939714,0.4186824188401408,0.1936127815048243,
 0.4355521713623707,0.4102239324739222,0.8913460060574218,0.7211420098835314,
 0.07156381747287732,0.2616980481712683,0.625522752878808,0.657991109859419,
 0.195394345285686,0.22860958754799077,0.5085711802575289,0.2792603999565355,
 0.14864321211723874,0.7391801966482696,0.8376850708195537,0.3862677010438601,
 0.6708619823961124,0.615861971984989,0.614711524789911,0.44508456842215716,
 0.06353994491128834,0.3238601126275713,0.5263303494427426,0.8688431828004617,
 0.9733764831767507,0.4426421568059892,0.4572812099507947,0.08104533632521238,
 0.5403952164762947,0.8849186506322159,0.40556348316938196,0.3051871508651365,
 0.9743064252158586,0.5282099098904439,0.20845691907062658,0.6787650884220371,
 0.682135726254819,0.1290545378587843,0.9452991351489893,0.921068128065755,
 0.36697928008711067,0.2864762032336926,0.759695481505782,0.2313296795640668,
 0.34489113498161084,0.8609955235962166,0.9863081442342634,
 0.06226033204965664,0.302598215026216,0.5174491914047762,0.5607268378393785,
 0.6687975697780946,0.1924291732954876,0.11506795052692298,0.6920709528412747,
 0.15540894903473146,0.39030192466306957,0.3227894024778283,
 0.9509304017238028,0.9545884099788386,0.8603505802010957,0.15796218036565302,
 0.5300048949736531,0.8183012075359816,0.4785162864663298,0.9285618741202681,
 0.8411506110309975,0.12972816887114014,0.9980428716239195,0.9139363315689908,
 0.8487185454263563,0.4229426867471411,0.3903078672931146,0.6858507355459239,
 0.41495813765082823,0.33485108485249493,0.7328601501732283,
 0.5800092751720539,0.31106179584850535,0.2060786693527834,0.2179409822352324,
 0.9020361092184331,0.9527959419896106,0.3036388290000004,0.05837847810085117,
 0.9104773665498248,0.15128744159512064,0.9893456308252518,0.8135167305608011,
 0.5662996603064042,0.07300430277568426,0.6875926998411455,0.6611671284276642,
 0.6002224224536606,0.48150242777044683,0.28604777346995847,
 0.8567639482973295,0.22413308669909549,0.4941382668821829,
 0.00388003961824257,0.6763185689851448,0.4531650342727258,0.8851076839666115,
 0.07737938598097227,0.20563237423073366,0.21663690569471084,
 0.8329800353206142,0.4455487510533397,0.9617948264618241,0.984913709532756,
 0.5902976694192759,0.1413164381173705,0.28951502518916006,0.6434379331042483,
 0.7909663768678465,0.46574897870880616,0.5043846798965419,
 0.25868420493532385,0.16608192618006912,0.595428994823594,0.3696289336660157,
 0.023447107357821406,0.32875268485478526,0.7467949066301042,
 0.8442151048505707,0.37897003218667025,0.9545650890505651,0.5509607843009469,
 0.8243594601187499,0.180111943845228,0.510709687274705,0.6283941985608519,
 0.7601789902254872,0.756931005606281,0.2122736973730759,0.9754961734455032,
 0.6893459060201429,0.2239655389548122,0.18184946461673035,0.5615839602968424,
 0.9294572100148699,0.06694713105221362,0.1574817259601975,0.167145723484719,
 0.40786443368099523,0.2660468233639779,0.4931470785525832,0.8069252811713481,
 0.04152221358757724,0.8416429458147167,0.7822952370505039,
 0.07567044089864738,0.8169312312165499,0.4826193246497009,0.5240322129526213,
 0.547937832370723,0.22277436562887842,0.17981998545775302,0.5995266987501094,
 0.6668028164646763,0.07847200521338171,0.2457333514992185,0.640036678119722,
 0.013512240919137941,0.9366345831526206,0.6387565268606703,
 0.005425680243848041,0.2237320389487899,0.5919799369687009,0.589352596235301,
 0.7657432518725077,0.17066318470414643,0.027499235862090576,
 0.4757194890866757,0.6624427677641365,0.6691755972153617,0.46441115625619345,
 0.3448708857544207,0.38059206533286294,0.9538686528401966,
 0.27916111694469614,0.5623923464632169,0.9778842599016279,0.5079670811605986,
 0.14364052651050607,0.7911327423012161,0.7187111912828209,
 0.34030603017995154,0.08748807912970835,0.03118198714487075,
 0.44899122457010765,0.20686302923385136,0.4148420825266208,
 0.8176536370719035,0.12469891019366508,0.8180327871411861,
 0.21110314511950046,0.3276312322030499,0.9217211209041245,
 0.22026133035391415,0.058774587028074476,0.07060126128782929,
 0.7893611935449903,0.21472432980857048,0.2189179520572423,
 0.31604594084953475,0.5564500506906055,0.8982220505570567,0.2895784016472204,
 0.9235420134798735,0.3594729599463726,0.8848784358882036,0.5440424253767947,
 0.10293291442737185,0.4654020339641852,0.7111002654597249,0.9077896048043612,
 0.07791932015057224,0.05473264767198416,0.15719764030845385,
 0.2477505622184849,0.6576278021882471,0.8502877236543793,0.2507527001703931,
 0.4555467745408861,0.4601137654539389,0.5004399530435557,0.505349658104723,
 0.8258823768121827,0.6546690345823747,0.7926404903294831,0.6790916019724061,
 0.7759032248648855,0.3165142284381841,0.10587192818987212,0.5302761800742902,
 0.6814828498346247,0.4086775195845469,0.013041327382359746,
 0.23324553156191663,0.046209045522400016,0.15552828573089406,
 0.912887305145243,0.9181503875279521,0.19500985413280758,0.36302335153947474,
 0.2012040199216928,0.9641045584853615,0.9140767118867755,0.5717494661127055,
 0.9146248907268987,0.7182700062827658,0.8203544465458887,0.648180759173055,
 0.5846832257679346,0.9519201098530166,0.6674305939336685,0.30208434678215124,
 0.31816150885443995,0.2094521807926386,0.18462336162002724,
 0.9032608741072511,0.9075744930188697,0.8573434064269955,0.30364133468597865,
 0.8515581723807656,0.3051424290537482,0.8815727939661308,0.7858455949870368,
 0.5496980960184408,0.07906376229950407,0.42626515458886804,
 0.6091333073204134,0.7470853535668737,0.7903413229939263,
 0.005443255087654086,0.5425511610900072,0.2169604212413314,
 0.8607733566621389,0.14828153023183854,0.6314576353953774,0.1037568911419422,
 0.2540863718258949,0.8553374295186655,0.14951629762930518,0.480366737326964,
 0.24771648626829978,0.4121079189351833,0.8188357888875162,
 0.23810896819591476,0.04739638803522739,0.4467406416852826,0.362157888037936,
 0.39191588870761007,0.872545902020058,0.6594515434373449,0.6488144837854084,
 0.18179797613703064,0.8178266767033631,0.7355212953957111,0.6121074428792049,
 0.8611633216284783,0.0054428002686724986,0.9808869858383797,
 0.6465374161577088,0.3363630281891773,0.18099984250477608,0.9336476987933198,
 0.11129590192354943,0.579606841652394,0.5345721575483571,0.48511200572955604,
 0.638482075153948,0.04754122590906629,0.21537675729123262,0.433040777472542,
 0.7260743838857746,0.9548840356726946,0.8828906008095987,0.8023316306746503,
 0.03839265714062445,0.887709096984224,0.6343776012211692,0.5790908981332665,
 0.16328090892389047,0.8768667450127821,0.34619286522859594,
 0.21619716500674813,0.8605638708977127,0.21882438770699686,0.598058728823313,
 0.6100161340157042,0.19601902212540423,0.16324670688386744,
 0.023439301141129465,0.20369211142231103,0.5906248018488782,
 0.23113462197959134,0.6202551767672649,0.48814521417196377,
 0.32447397089053565,0.6026727644899694,0.6402861130424704,
 0.20016536794884487,0.06613935293682205,0.735019162909753,0.7723754146565522,
 0.4938839338776362,0.5106185185587199,0.19009683566124735,
 0.23278192099842765,0.9748188959534653,0.8045804380255256,0.3504277430264596,
 0.6232224068292139,0.9724883356369367,0.9214286951081483,0.30941316938840313,
 0.7491875011593729,0.5315309238422367,0.5537267418087858,0.9884322818734456,
 0.5903743310104046,0.07766799985429085,0.2573297311186069,
 0.47724687824146894,0.6651175526374024,0.0054195659792326545,
 0.9206880198053622,0.5671852551520212,0.109396502059672,0.33057691701463787,
 0.425579332452259,0.4816542311456522,0.8871708064949921,0.2080535263494151,
 0.01816076525523136,0.878046216361085,0.8166100118249557,0.8315170836452199,
 0.07713260691966428,0.5457401524065608,0.48003322357713596,0.65469041337929,
 0.6280708962532049,0.7015323031057233,0.02622991445897438,0.658150046032041,
 0.10407570671454125,0.2088059204247047,0.414286976708517,0.5875911864513201,
 0.5042456813108936,0.1474880406533401,0.7855879254270298,0.28465859435921037,
 0.3095341901758373,0.8315453163735939,0.7070523143825314,0.7894570052148884,
 0.291368131979966,0.05323427089375288,0.44586606906835013,0.3839649632924128,
 0.3913780419731858,0.6572978454452246,0.6746829185161424,0.9396005724169648,
 0.6845365879590051,0.7931997029171144,0.5092083251422216,0.8764044725462496,
 0.04136581560059949,0.7633375005082771,0.007693206868594649,
 0.919651162341218,0.6994803537878475,0.16099104102064554,0.4054575595750639,
 0.3101022974969635,0.6138594502918984,0.027142422590470727,
 0.15701172480796421,0.807881523791265,0.5545353353340744,0.11029231123281702,
 0.5100625895909062,0.6998210740303524,0.5422251982667076,0.5330733233286697,
 0.662719792823432,0.1599347970584818,0.9626243861666256,0.8864049390921548,
 0.2884213358565866,0.8235648596288,0.8563064551236135,0.5050958639850973,
 0.5263165055040628,0.3625135614123448,0.5594437082341883,0.15127228886624267,
 0.4935409422925381,0.10532756483437045,0.193574868917749,0.42003041560602417,
 0.1404644050248668,0.12075716968658923,0.47167800879068555,
 0.8743281171262426,0.978238672244829,0.116008821015253,0.08445351721250183,
 0.7462173285267533,0.5383394515241813,0.27924975724105483,0.7538634548606592,
 0.03395414982728828,0.2842053692966472,0.4912599541530046,0.7617172939022522,
 0.25830011186621615,0.5206280194637485,0.20692164516943923,
 0.6391430843136922,0.4064263815514987,0.9997606528770616,0.4428587112207105,
 0.5766911923345116,0.7383879600506269,0.8941629840357563,0.4291717121618367,
 0.14665131476184468,0.07662785128834015,0.503267072989947,0.9067389877870664,
 0.3288189560079672,0.7307674006790413,0.15992318185294596,0.7936511109813771,
 0.3241072173099223,0.8794245324546749,0.5291837366758229,0.8883847592895122,
 0.5758734660986087,0.9861148680349959,0.33724458433074744,
 0.38966768832843324,0.22799011819920523,0.6625558433395996,
 0.6246673775709556,0.034331322275942266,0.7411803867463617,
 0.1923592390129364,0.4519679010127845,0.9828433635825005,0.9594349999021603,
 0.09202392956396976,0.8096794871462344,0.4324120255982967,
 0.22024102906012244,0.8782705717421335,0.42325297594973,0.18812512457781638,
 0.9244878895334512,0.8613900908854273,0.572952252884594,0.6249003994201457,
 0.6814905463659984,0.3149637942737822,0.12413337740473585,0.8209001022825275,
 0.7610126299008855,0.5343043342546606,0.8093550411006059,0.6465976636462857,
 0.9682408553539308,0.7338497006478036,0.9960686064362754,0.6718694603833845,
 0.983633115716512,0.028916432845313977,0.20560337175522692,
 0.7623932862886332,0.3737267000480504,0.7698479763247174,0.8342808433933894,
 0.5718613906980224,0.4757497313870931,0.20621991332430334,0.4809295198575442,
 0.8580069928858098,0.8356908349306446,0.4799782525370354,0.5242774013156,
 0.5806333415781124,0.09746414745121856,0.9086610729090581,
 0.07680415913739302,0.3613318298096173,0.9366609715679193,0.8691382972571329,
 0.8493829025749746,0.5002302772267353,0.2992877119451395,0.11409161569392967,
 0.5344025972134722,0.05412778796325346,0.062383080534324176,
 0.5049151800309348,0.14201792674277947,0.015658945485805043,
 0.11921655220757632,0.7447430212553252,0.7266679929635403,
 0.45290887847910344,0.943558479403992,0.29462311010569886,
 0.22017410640376067,0.6976255231039488,0.46978705564041523,
 0.7500100151133556,0.5872768981147141,0.7245903953489005,0.5418013975660804,
 0.802470621344771,0.29333876605560305,0.09281889659190712,0.9300751166991672,
 0.28935311948404474,0.6887055022803859,0.7453967708780903,0.875135571442927,
 0.16788894629049333,0.19724273028131334,0.9561378111034599,
 0.01444351432115698,0.04694910714944833,0.9178974664198796,
 0.24327797951468955,0.189929831340241,0.7726903879127732,0.39978815162875003,
 0.5639530876570128,0.15947108515397268,0.9743642367809507,0.9209096215194329,
 0.9878503503337084,0.2774186444568434,0.9923015550579555,0.8098833925132667,
 0.7364139013878139,0.4957574834366424,0.4048016713623743,0.952184688571988,
 0.7272171805502614,0.9095822916144957,0.8947463085246585,0.6321182797338295,
 0.1683197117462003,0.2552339958190284,0.2813247319484653,0.11551747107591126,
 0.8693191950494716,0.10433542988092148,0.8605414448157652,
 0.14630949688853023,0.7760988737714101,0.9458371157397114,
 0.05268057881972843,0.4175341974422918,0.9963218105523425,0.9217350152082971,
 0.46252028276969326,0.9087984358379451,0.45326770908477854,
 0.29940240707341126,0.37226453010499505,0.21444600645532041,
 0.30550071185611516,0.27818716856150716,0.7154073095419476,
 0.6153528508313048,0.6162365474808366,0.38689876642726073,
 0.04222618518948362,0.94561821162293,0.8491008415511253,0.8302207724588262,
 0.3098193460096115,0.09958635707146724,0.6553287745546825,
 0.06649261775982596,0.7672667049548401,0.07997263609792293,
 0.7391680322682213,0.7944478375651817,0.5376713479958267,0.3262982062199523,
 0.18924996099356783,0.936393335903563,0.9281668653089477,0.7194482199139544,
 0.651411459320729,0.33424682584398857,0.7075163623724167,0.46191686500871143,
 0.06877268902527,0.3133789094376289,0.5601858555239884,0.031925434558053856,
 0.462983804961141,0.6106794098222813,0.8050003612098511,0.40574213173659224,
 0.35150733692182035,0.9837021261168825,0.7871673716587808,0.6773570019828694,
 0.16607945006623992,0.45514881760334935,0.7329632954103608,
 0.5784511651968627,0.8766422367872049,0.1953426521189522,0.5747201237621439,
 0.17622993116703056,0.2801154969483618,0.7447778257625178,0.7598611989850975,
 0.8157463254307677,0.8030458082856785,0.53721177527016,0.6517153291959332,
 0.9872511332361323,0.3483860081189971,0.6620244636146072,0.7596526717838936,
 0.46493305299554455,0.8057482138350809,0.3123159029286192,
 0.41644692944442585,0.5030246955606559,0.6234232154084245,0.4408147976684015,
 0.9824878538052423,0.5980857331473313,0.4692886219240926,0.7910770725284263,
 0.11234428006138686,0.3220978463508305,0.345808011516647,0.6999716457094162,
 0.6261185028615135,0.7425655815541399,0.16054628131349513,0.8707182777167589,
 0.993460557960962,0.6021990056946285,0.3351512764219101,0.2871607672551706,
 0.07125112764475772,0.28899788496365475,0.3654572515407156,
 0.2854430167503572,0.4414622533689121,0.6647191469567264,0.32419241373325924,
 0.09112994600257585,0.15083988569572926,0.6931157277961592,
 0.8471772070547998,0.16697129292410873,0.5268700041053251,0.5113633053739134,
 0.1507000596368363,0.13119178730173897,0.002378238901691976,
 0.21787537692839265,0.9685780751480318,0.8010778927844047,0.2705720109074977,
 0.8992256444479514,0.47423162177756695,0.3012546486622799,
 0.09289791206479825,0.49549932004074104,0.7663344109151886,
 0.5063573811526241,0.246764064870857,0.7420125242654864,0.20857148552530014,
 0.1884182375215534,0.21777354904391932,0.8965956385573838,0.6743943941145124,
 0.5775600213583109,0.36007527786933724,0.3774104630732674,0.4482456789518906,
 0.35478380715853197,0.43344126407924,0.786955704807713,0.4952886395315278,
 0.8141086824897039,0.6065535673521707,0.5204582203344251,0.6234612986053185,
 0.35646641707106863,0.6912621250925601,0.18620325765299972,0.918141811679612,
 0.4959531260568354,0.364097500727639,0.2144437350948657,0.6646233062416819,
 0.7283533548207038,0.12395843863543543,0.4720372210062008,0.434674230432762,
 0.39972424928564343,0.03252734410065494,0.9062419287467323,
 0.40948563929283943,0.6571550087899767,0.9452966012606795,0.7185513249030561,
 0.9367823415553375,0.8939628196386954,0.1436824352859678,0.8769466377992587,
 0.6423417344407387,0.32894961118203403,0.4310664519679146,0.0195544835072734,
 0.1548344263025423,0.10199817484905815,0.5945669389683891,0.4339571291710881,
 0.7272312655850507,0.6726950576182412,0.21670932195123693,0.7252929007210059,
 0.043392936489030864,0.608687372369133,0.9261540658280621,0.7064365774250169,
 0.0673416232243802,0.28206074708081286,0.6314640221409811,0.2859926536938763,
 0.05775499871414669,0.9947655194946325,0.9954733267830831,0.8298075088878772,
 0.2003477901241102,0.1606613226512167,0.29623686529976756,0.5143261643390429,
 0.8676294646989525,0.7750633443894392,0.24852079706420493,0.627168103761932,
 0.8313594246942164,0.42608775265410015,0.5391212943620187,
 0.07678866022000352,0.8726398442619284,0.023131200291709453,
 0.6510353876730212,0.11779753005126992,0.16183240360719064,
 0.23225618698361106,0.9094861624227528,0.5882960186781503,0.6544955428510743,
 0.9255681770620052,0.5392056559663545,0.9193797076108912,0.6165836862705057,
 0.5579529402569733,0.8560295657886207,0.35414155494005206,0.7506688659712881,
 0.08036990095523122,0.08360514619372938,0.121099815616865,0.6333917957085475,
 0.0740860576376623,0.039722191295842646,0.2995015997049053,
 0.0387075847520304,0.4347770637436774,0.6035614687216844,0.06327447397179264,
 0.49738799908428444,0.9273646472949599,0.843347635435295,0.9605304620315311,
 0.772957043474399,0.9321734909951553,0.4307310804486918,0.2358234492667559,
 0.035501220813612955,0.314863489985416,0.7820798534554242,0.7069711175249321,
 0.8280437610513709,0.8872854704146106,0.7460332106929757,0.53877627088448,
 0.8976631797206073,0.6208626491319078,0.3760156265191936,0.7951941132012607,
 0.807304378268586,0.4193715065525194,0.9161332681861064,0.6034132665299149,
 0.5433703769178773,0.8146550959316208,0.24339204943048687,0.5144050349692916,
 0.017366741804390573,0.7191060753149473,0.8050542392483703,
 0.5848132313093022,0.6505361308354448,0.13185918217114467,0.7301916034947548,
 0.9490824150804125,0.31116333781965144,0.5309596069934976,0.0976793196170671,
 0.11206157967267527,0.032205848688227845,0.9862183141981887,
 0.9996467006886053,0.31881814626495075,0.15149385711883911,
 0.44259947674676536,0.752886999508477,0.3279413652211389,0.7951974054247586,
 0.7541328265669262,0.006469459689699031,0.9917174346469234,
 0.20440172708186644,0.6116794852337581,0.6657067785174656,0.3679594187027653,
 0.49653049579515274,0.2706863663974436,0.8840831878184406,0.5131403743409785,
 0.06276466660037217,0.539701759791658,0.21547351219524424,
 0.11738109822902276,0.7224129133376189,0.7128423937846835,0.6737235284807496,
 0.22701023598126358,0.33516989897238214,0.09593594973294384,
 0.9368591917697702,0.42528735879691626,0.024999524054914124,
 0.8337109137036361,0.48828335903837494,0.6504018287718788,0.8142220897224286,
 0.5756194702156985,0.37889841989969764,0.9950220465469664,0.4468088846330215,
 0.7684072774394927,0.6202271158748562,0.6096935186471958,0.6282464152845544,
 0.02340922065298623,0.9758155064215366,0.755617490825127,0.3669687684676909,
 0.71265562540677,0.25623306099464505,0.09550926866403864,0.092324370506905,
 0.7319482712066503,0.7701225511480527,0.5817695776399576,0.238446251042418,
 0.696286297542431,0.775739499407205,0.2850950852447841,0.18881143697827363,
 0.6458202325420538,0.5104929278613947,0.9928746824991485,0.5963470294821687,
 0.4593313288128056,0.18557122198309373,0.7525148179168345,0.5313829044995065,
 0.7453666601616076,0.9560307687585419,0.12976260950297935,
 0.14814569247702092,0.47362107676906073,0.18182560945369963,
 0.7905498784210714,0.9500179941449355,0.013296639191253457,
 0.7836236106682035,0.8537147237243615,0.06757337408480235,
 0.006280425289891256,0.6841766085238039,0.9488651719906647,0.869492556525164,
 0.6604884936599553,0.20669140292210542,0.593714204881829,0.03932543145051226,
 0.4248338550848427,0.9318571929405929,0.49545985557581895,
 0.053099124346374205,0.8912548240372717,0.15879081041523513,
 0.11893707828018829,0.3421239973221465,0.7885396596280108,0.9336653570500251,
 0.31815772402640063,0.99054462712147,0.07048425604752828,0.18767310947480764,
 0.9744429947505676,0.7615726220193093,0.2575616738466009,0.41733492059177413,
 0.10809580847162359,0.1818762611188206,0.39663085061934267,
 0.7658271680431854,0.44764634532621395,0.9343338953498038,0.840908114341533,
 0.41385639543698627,0.25232767863625005,0.2653106652563033,
 0.21390146796263854,0.46846155155575153,0.46239371350169456,
 0.23722013998370461,0.3493464355628587,0.7844631556868124,0.3518950714380915,
 0.3559210703377663,0.9308380323827282,0.3661324522462591,0.2746354955624729,
 0.35633011143350957,0.4115226288635526,0.04588659160851971,
 0.5224952760216244,0.1963475536480318,0.3618026443625123,0.7991807983726135,
 0.18744323825272324,0.4104435618506481,0.27344579451727447,
 0.7997746531267595,0.9868543704616449,0.21372191065164314,0.3685123636224782,
 0.5024960016207588,0.9819651745850166,0.3909729694646913,0.51029364230839,
 0.6233131972260644,0.14322860945859772,0.23974792362747366,
 0.5034921008436788,0.6367335086864971,0.034258248890665866,
 0.7072456170431476,0.38589977663409414,0.07767064684200298,
 0.2736868363375292,0.6780401882026987,0.1322067830760607,0.0686712219625375,
 0.9775708212388965,0.5175527677989622,0.14449225389336484,0.9875715579763463,
 0.5711601344796889,0.7062789064208774,0.96127541134481,0.030180466008383866,
 0.9075087637367462,0.2899835377949428,0.30444602014527433,
 0.14260046167094376,0.004394035073805758,0.7019003305607785,
 0.0897071747631597,0.7655029223779719,0.9921656647568535,0.17620037669288113,
 0.23510086674865593,0.07640650391350046,0.6460848674609372,
 0.12311249755131992,0.8896274658967412,0.2044741981364867,0.8762379981830205,
 0.39716490046875996,0.5497512396283778,0.29123118567711814,
 0.6956712231398471,0.35002039805248475,0.5246666166424537,0.9999874423623963,
 0.4532106814793446,0.07852515817505212,0.9094401256341641,
 0.34749181104484994,0.5573561164128342,0.19448673449413445,
 0.3933626473897025,0.7240900667037774,0.009414714367052879,
 0.44638651299118126,0.8734045888300221,0.36366322291440434,
 0.0665608076354085,0.2812195841938232,0.48536321702404217,
 0.41735421450790233,0.7058663060099135,0.39054965581478474,
 0.7726737685487037,0.04389891808310831,0.5753096181378533,
 0.07151652297016697,0.16465748582994433,0.7781337269209567,
 0.2680401706272214,0.7585691295667571,0.25510500987621976,0.5786232840078415,
 0.9720907842084943,0.9883430790746481,0.30536530730690215,0.6598679451072094,
 0.747892996550503,0.9269075575273888,0.7278482439733908,0.13912928670593616,
 0.2377379998297906,0.8890252911747789,0.36944334612516905,0.7910322908587023,
 0.02419520604209957,0.2272449423472529,0.27546501366916043,
 0.024114762985146476,0.4223568800996258,0.24121530137359537,
 0.4471601885487697,0.6388185108819436,0.286928818246608,0.11257782273449812,
 0.9455880221355131,0.8884884130386287,0.3980438699785137,0.9298858730526739,
 0.2277876619037038,0.6206328814490735,0.6391068512029263,0.7922438001172686,
 0.8732249812480117,0.5251528882744669,0.2331196725310417,0.4424189275696212,
 0.44386057331558926,0.12452102227425621,0.0645703945868179,
 0.9854582544863305,0.9426935591295227,0.036120570856157275,
 0.8578559273398342,0.302429938077984,0.44065129384196655,0.20691183970574345,
 0.48087054409842744,0.4952724833826434,0.2107405445094681,
 0.27590416300371223,0.6719212436861284,0.6139861808412974,
 0.21776021472275287,0.8050859279303392,0.14402649895823683,
 0.1757265454633209,0.21961590764904815,0.6731544549061763,0.444437307723099,
 0.27687212352648216,0.8934609381541141,0.0800466744001791,0.8611687098545071,
 0.7928136285827175,0.3103063107044369,0.17918391386182597,0.8847094179304755,
 0.14858789004000716,0.28837101193899517,0.45361350627917174,
 0.6446262832512393,0.5456020676650077,0.9470494442041503,0.8320281430601824,
 0.10148361237847203,0.1212766657236698,0.19282526144042533,
 0.8596584016068652,0.8356791007969954,0.18253158370551748,
 0.05825166801722237,0.06215428215325458,0.44115028670545875,
 0.9907060422108052,0.21394505884302506,0.09591868963305605,
 0.13177655652552744,0.9051476489786192,0.8371008500519687,
 0.46752767535908757,0.7618681647351145,0.6112401650184204,0.8155163029661537,
 0.24865135654796355,0.25667320419383755,0.852446001137359,0.6258202439105746,
 0.4906998060799467,0.9189841788573971,0.060883361045133144,
 0.22345577235070613,0.10765384426285352,0.04107573163775102,
 0.5358342418304354,0.5098410426998559,0.511716137809881,0.014675851844035415,
 0.9460353350868026,0.6911991719683371,0.029965583893422854,
 0.5809527122062081,0.9229430741616457,0.7943978668587491,0.651287403269787,
 0.6504142696130857,0.8726761840680659,0.8191125367651262,0.8503895374907326,
 0.3907198067951033,0.18452754638603608,0.22088941917566485,
 0.5793349295548967,0.41194860634813224,0.34297021719861054,
 0.07195581146081564,0.3796552785399546,0.28357893581843197,
 0.36021992847181894,0.3532282773899903,0.0447588070820214,
 0.20911694170825956,0.14353341821946686,0.007041702360722768,
 0.9673998791524061,0.6347650401077639,0.8595750995594988,0.6313253795124062,
 0.8224420797660825,0.8856057029860046,0.7030984720272833,0.14780140745353154,
 0.9622316511080321,0.9712394558004699,0.3851884411469704,0.6415342814014836,
 0.111987473744246,0.6192451307304958,0.14604784680493732,0.3249073430616216,
 0.923512624902568,0.4969920165463748,0.04630382724040072,0.2289205114876285,
 0.08507360770247285,0.22221146601479802,0.9837486024402542,0.545962675814843,
 0.9866397623077162,0.4652801001160779,0.3097468195322456,0.8767705752082129,
 0.9759036338824073,0.9163290596157961,0.6851581891400231,0.742178096352439,
 0.015414207001291924,0.03252964702510952,0.7494484599921861,
 0.561194415837725,0.2838965378261491,0.843765396121249,0.7890055447166211,
 0.9879079735575872,0.9852863608387017,0.7491841350577633,0.42674836304021413,
 0.9552664109635592,0.5733695228723155,0.27864592998072313,0.8831600768573904,
 0.9324697207405586,0.8858424493863949,0.7754913985940123,0.07221994770934892,
 0.4827310443975533,0.7280329350471572,0.5601503978306114,0.2560942346320144,
 0.15584561251163653,0.8864715511463329,0.4158380120125045,0.6713905897524194,
 0.40863816141854614,0.9457390055634826,0.10045211788410402,
 0.16223967723186627,0.3257442973566139,0.4960369415214452,0.7871420905494255,
 0.809759687431637,0.534542781081803,0.7034472534419564,0.8691476191159838,
 0.41207983841504836,0.8849998182255207,0.034445647144741076,
 0.4600681726206767,0.5013805540747468,0.05194599120792376,
 0.36068616120157404,0.5667008002373002,0.056613060913846325,
 0.16784852501500203,0.1286402083452578,0.8601315217264378,
 0.006699478992422181,0.29465959974845723,0.8311182921375823,
 0.8861698203225751,0.13954149726025555,0.7133715742477591,0.577984895787699,
 0.8983171369592584,0.7016409544968686,0.73535442993175,0.1951517812983401,
 0.8275975641866885,0.6892905921511264,0.9302604809594037,0.44795218490492483,
 0.9471856047132631,0.28070967785661527,0.5547940813982114,0.293829969922183,
 0.18774541541836065,0.21615489411343625,0.6065183327872746,
 0.19522470862943397,0.8903332837344043,0.3109454381620087,0.161454429099115,
 0.44627284717365034,0.817139854353239,0.8845516202840352,0.7527230227867009,
 0.034465893666211334,0.8615427095748549,0.07153697201427112,
 0.06832943207515418,0.929522280844433,0.4839304664224665,0.7090193458139102,
 0.0641459963232498,0.85166020657282,0.6816651515832465,0.10207967437089449,
 0.1800479538508849,0.5563562362142717,0.3300326114524683,0.2816006819344301,
 0.7263610063834272,0.12636852403521426,0.25423863701885674,0.970555992780608,
 0.8255274065566577,0.8120613385632737,0.06539033207333111,
 0.07477887654354276,0.7043181832819139,0.612670624510762,0.6382291392920914,
 0.41758121799797043,0.8438679761746707,0.3377535392542025,0.9151724624725082,
 0.7477182508072395,0.41509563417974116,0.6317599944148325,
 0.0019790550842669896,0.46976126943621355,0.7359282810703363,
 0.6348481439326761,0.23334316213953554,0.5407935538847981,0.6896224784358111,
 0.023960341753256453,0.8213886324143361,0.08489301775382385,
 0.15083318675652546,0.8837053689936617,0.29843768702830287,
 0.2472171481166483,0.9909604910021118,0.0327793654479942,
 0.029148883845891205,0.4790045627044497,0.8426393553352131,
 0.12194557686994245,0.049227263301747204,0.30202695041754113,
 0.12069093948533649,0.9417894084688987,0.6466593655880317,0.6613647518340974,
 0.35551139840291046,0.706000853083616,0.5620446834849283,0.42586147205739633,
 0.4664286501502366,0.5308926092608641,0.18745640018718768,0.963891236388456,
 0.1726422105187173,0.9730019042470835,0.005930085871421076,
 0.7960957233049711,0.5035390577793821,0.9844727377698367,0.5841149562781556,
 0.6131883905090344,0.1477193596882238,0.07665925103213123,
 0.05092894779245738,0.041571991334171665,0.01587833210980083,
 0.4608518710402746,0.5894814020066678,0.67541065854485,0.6869409945729893,
 0.7264335632345782,0.3967026897721787,0.99254059924105,0.3453536058416202,
 0.8462116485146928,0.1828065009316131,0.06646374548429912,0.1990381841524449,
 0.680410843530526,0.7438699540347982,0.3134609162010281,0.21881975534291853,
 0.28800828781062504,0.5774682740414717,0.17287813378544503,
 0.7013934070297327,0.16861052718970104,0.7394234628093832,0.3422482308806124,
 0.035568977045113526,0.14081757645681892,0.790199421971947,
 0.3384267756580338,0.7066174772138739,0.11090869986765983,
 0.17370548474104697,0.2791156145632083,0.8735871275440557,
 0.41925957216979637,0.42085552774682555,0.805439401305269,0.8470573227162354,
 0.499692951442653,0.173912571704249,0.017880492917910273,0.7375007635224694,
 0.13436768693880408,0.1234048602608433,0.9658833557568712,0.6067892915381581,
 0.6086236007608166,0.9996601385276144,0.3339489824161025,0.5003286341628,
 0.5601362674074299,0.6112607111362012,0.4564104578921785,0.00483022396720123,
 0.7983950578463407,0.5823390680219911,0.6929843232225273,0.9508154816684488,
 0.5641418855105224,0.13351172921112486,0.717273684211585,0.8205695137820569,
 0.28396888817545307,0.5824895118873147,0.8139030634242873,0.9707075320018376,
 0.5739516604211958,0.7619433862869304,0.4203090161856229,0.12812474085655323,
 0.14019552900641952,0.07182905809415396,0.5982780049844928,
 0.21879910804430236,0.9187520130521263,0.7381170159847746,0.6302148894923767,
 0.43772656013863465,0.836039903460079,0.6036650567967087,0.2333855409643979,
 0.6452773352257387,0.5648737279915624,0.41675647989222986,0.4722469979752949,
 0.008136902541958979,0.2150745583565561,0.6332248560809854,
 0.41931377657650903,0.743517261079035,0.14304780233840664,0.9995276770395042,
 0.7666778318997205,0.3318639532856117,0.22419340935922383,0.8161724079184685,
 0.27984743216617514,0.9833100169550549,0.8579896431782412,
 0.16491812479862444,0.16917443189372605,0.20106365318938646,
 0.44688511777587403,0.39856105195259994,0.19709904053194116,
 0.652305103205122,0.8031337488644954,0.9952378006053637,0.6892587705220721,
 0.5236877166901749,0.09274184108056593,0.8876703947886893,0.7233579372628253,
 0.1866057906723767,0.3685089378534122,0.20843781974562314,
 0.30291764243841135,0.9740427072177411,0.3633420715906449,0.5875337262333012,
 0.8922321486856108,0.1941477264299598,0.8106394959398661,0.9144693028725354,
 0.9282834025327251,0.7638661366695091,0.061006097240131174,
 0.5691972169318075,0.23761319586199148,0.3027837412977272,0.5684802259310141,
 0.16634552884664722,0.7704928933896164,0.5027553144196351,0.7580956448378264,
 0.24968921927843502,0.5034637060119627,0.25456939762050734,
 0.6605267208685648,0.09762536302091385,0.470058127797796,0.6696738043494894,
 0.06372398154630055,0.014427655985219778,0.887438265748147,0.754757979124864,
 0.6053172550759829,0.20228717476461522,0.6020093826173174,0.4314096051976454,
 0.9654837826470481,0.03667975950610314,0.9189273373978739,
 0.051690849181553866,0.6917070057489719,0.204386132596854,0.0173860686236903,
 0.721783359841308,0.6752055808771604,0.29284459181719924,0.9794551839797954,
 0.40357898598108877,0.2929612660405554,0.7142871089134458,
 0.17605447679348896,0.6130570178109658,0.502305473614034,0.9916620136085097,
 0.8661168040780403,0.9991708293582722,0.3864728528695798,0.9213319506604053,
 0.516572666386486,0.7442026391647661,0.23525193795093924,0.8204588300695007,
 0.020281308553390387,0.7515594496488354,0.48135934916807643,
 0.48435219886589453,0.809787264426271,0.21546750605205722,0.746570491046805,
 0.6841944405392566,0.92763087315089,0.6799588761783136,0.2051147952318465,
 0.35035449835298205,0.25135911676459277,0.5742816673490352,
 0.29221963758675173,0.306730302782119,0.589837888671187,0.7968418302783247,
 0.4497702277625816,0.5302348647362916,0.9085923811048886,0.27702707216177824,
 0.8878504707948167,0.054035221518513255,0.3360923281800763,
 0.8096126093418529,0.3186494277523024,0.9624965569711584,0.9200264196184568,
 0.3213657762917349,0.21313166593525534,0.16604330938970624,
 0.4741965926171169,0.38021580921461484,0.20644731164632435,
 0.4223094274477954,0.15798318208110396,0.5961704720172806,
 0.05527630916013493,0.5933930101371028,0.0880019873934339,
 0.04420973360465541,0.8020832193409795,0.18802758072093062,
 0.06827198475980778,0.04691866466577599,0.7062626174120378,
 0.09967923738521256,0.47868417657718354,0.49763312316706376,
 0.6171290328804127,0.8538890129085537,0.8793249238127258,0.5309928767119826,
 0.23356658648612827,0.3888394296337807,0.10253160999075195,
 0.4726130918154432,0.6508426894410989,0.5114115836692364,0.2729765496355312,
 0.6927441444873104,0.07123906783759737,0.8008723452231153,
 0.24062410377665577,0.051032679897753064,0.8981119892264294,
 0.2535930040029597,0.38700550870499373,0.7821470636844153,
 0.09464143867843644,0.3781243095326765,0.6041934026085595,0.5822817497602851,
 0.9092942015646411,0.10015228080872618,0.5309385414322942,0.9900545180812391,
 0.30667854954967777,0.2860804763955265,0.8839353138770127,0.9434684442404316,
 0.19770617135070778,0.12461020282471158,0.45725149924647923,
 0.5920444356095041,0.7874314136978647,0.3727256753289996,0.34357862640694925,
 0.916891921174652,0.3793658964143062,0.3685544600226961,0.06317468299384787,
 0.9908795530093324,0.918356187093005,0.2936323124749739,0.8150328957936285,
 0.30720646181082967,0.33236767262323885,0.20162804075701313,
 0.4969715397803395,0.08987735648451989,0.2786514796229904,0.5668740106745546,
 0.9879757644518795,0.013742425128864366,0.9112585877914587,
 0.3094756145741486,0.4656266607568085,0.49698746494160373,0.444479105558267,
 0.4462420514334485,0.023178141577272715,0.4275703119260933,
 0.3679885589211448,0.5450022179829647,0.6252576374480119,0.9995246814858126,
 0.5058438298472729,0.9023880738273188,0.7500243136081042,0.09524894572409615,
 0.21557422105903812,0.24299981889368227,0.3574606598632789,
 0.6375071620556065,0.8961798189608836,0.5576923916075144,0.13370119834982663,
 0.1724200839089639,0.1990534794464831,0.5833657780635567,0.9671932231635986,
 0.6954130541143118,0.4718611909087903,0.591297408585075,0.10838659068053802,
 0.2057493472339066,0.49581627521168214,0.8320603317608158,0.7605081718759422,
 0.07834655767789866,0.5243076584541179,0.038433585269904036,
 0.12426121389976774,0.0390891465374108,0.819559943760314,0.8706375388133659,
 0.33520621026034036,0.4782525357539862,0.6351285304658333,0.9599906072151423,
 0.7291836925487821,0.4783463393689521,0.8341353219583416,0.0767359971931093,
 0.160063531609563,0.45752890029208204,0.9081474725461709,0.10238530859165407,
 0.840848437828106,0.18323165230892347,0.39329061370702656,0.7106133265665043,
 0.22618335113588417,0.34190251239280456,0.6611644072066952,
 0.6959019743519095,0.036484867159961576,0.24476528138416354,
 0.9742924268380259,0.5912720230094316,0.6948921954399674,0.7450405394182147,
 0.6805448794830355,0.962166370786727,0.1791842958189679,0.01525400870232918,
 0.7706311114955128,0.8491230548309473,0.3644683519278089,0.5587621697707604,
 0.8579214007102854,0.9963806454457795,0.8751404401809286,0.41196007831768766,
 0.24472059840366878,0.8600790279930827,0.689555653117443,
 0.016875780570398513,0.023987638247883858,0.693295177519436,
 0.1816464366470243,0.5999222120735748,0.3342089241379992,0.39193463858052635,
 7.922516500143972e-4,0.6277034843439822,0.6340186777422104,
 0.9468272459668652,0.5220576093788034,0.9897497847153693,0.2732071150510583,
 0.44562734454673203,0.7341787337969571,0.5900052622910473,
 0.21211429268808257,0.1062002774023405,0.45903116740813077,
 0.8041598387286232,0.382824421751337,0.30595443644831644,0.626751824797421,
 0.7116012103483886,0.6166044379261246,0.7169918289888344,0.44445231063497415,
 0.4437958512431256,0.0056029553145686695,0.19199827960263216,
 0.5670426327251553,0.18932856670109777,0.8547935515411619,0.7736861438408651,
 0.7205734646586125,0.21469853617814327,0.22467457960062198,
 0.35677248641926884,0.6860269074115682,0.9035659389300957,
 0.22838009892945266,0.1497108882110929,0.8030592779282119,
 0.030958969478075082,0.690833516882302,0.3637982378891077,0.6569840747881964,
 0.5629917648727825,0.9407314259441405,0.45563595066236695,0.5211289685011239,
 0.3364553746515566,0.07457447654961169,0.5347231880250402,
 0.14662441675847204,0.7101004506056073,0.9952245119817197,0.6418100041138249,
 0.9752769518527457,0.33918741433230837,0.4271993220739311,
 0.24035221823155584,0.8701986056892045,0.989787943509345,0.1117623591678164,
 0.03460502056328918,0.5104520208296562,0.09108606928920504,
 0.010253061955259435,0.7445143752197639,0.9618626203296697,
 0.09926021241720417,0.5474775151205058,0.9289759237634965,0.3443687538309521,
 0.42967409311778304,0.7772768273975683,0.3660821445986133,0.3427628210886373,
 0.24017933602311292,0.94330047436089,0.27929729254802516,0.6531115362939858,
 0.9921282677685007,0.9673662366517356,0.021327361656657717,
 0.6439447618054138,0.3894826536142282,0.7404438520218604,0.5567777390184179,
 0.4627503107581098,0.8074982991565129,0.567570426679933,0.620643749881935,
 0.32442542483193626,0.8829703714184967,0.15499571505613063,
 0.39410806846244717,0.8276115553395588,0.45431962540128656,
 0.7141980062292868,0.8691608418011978,0.16044990309729779,0.6872855071983512,
 0.461089550887944,0.7518565300791543,0.38707377545079324,0.14357345617302364,
 0.9185549076955978,0.024345318884878897,0.8092415218553651,
 0.6054902013742264,0.8187065207034907,0.42459359731120394,0.6749224060785834,
 0.8294429825282832,0.820503496264097,0.8407911056534848,0.8897782002805917,
 0.9779519075771556,0.015621754406290655,0.788985603908065,
 0.10316341852619859,0.07162435032414927,0.6709409624037787,
 0.08163898583561113,0.18566281061450596,0.38171301693793447,
 0.11226792410108155,0.2235967862376489,0.3314808367327027,0.9199755689507784,
 0.3217664941288114,0.8081780422139491,0.7602763093119671,0.24667611854637728,
 0.42520998561195755,0.15715223490078734,0.20158188740217386,
 0.6000925189754028,0.07328782973673931,0.7784826216704425,0.9555368063827128,
 0.9312313145720235,0.2868704687412288,0.9384418750464196,0.17562699604329557,
 0.5893007274546359,0.3033331374536741,0.5198990054609437,0.44801308212700097,
 0.2285104827914619,0.7601892874362407,0.0026531379825917023,
 0.8165907195684212,0.28525915693770065,0.5398532115245679,
 0.03827703262414883,0.11891815313922349,0.4694715109953662,
 0.07746602229012645,0.08812502248674847,0.5648121062343711,
 0.5917604470490098,0.7464572275062797,0.5415032556369804,0.25142734920480647,
 0.6946153625559252,0.29774621336747853,0.4621713695003957,0.4717570215618371,
 0.16713136135449413,0.33696424479979603,0.7395079647865986,
 0.7076706971358404,0.3230980408447781,0.43294753838856637,
 0.05597862622824168,0.5679481409586565,0.6203786955183398,0.3918838662736316,
 0.754262333254594,0.5444641632156976,0.22033881978731107,0.8727971871360496,
 0.16761878892263127,0.9347773293271932,0.9228909041438634,0.4263517086066455,
 0.712809311501547,0.6935767440226499,0.3665837529476208,0.43217577773257254,
 0.8268674102272454,0.0948689661657739,0.47555311033454484,0.6182737378342755,
 0.6824842007571659,0.8959963467157391,0.6979123418258821,0.41812448225743803,
 0.14621617209630178,0.9587216355062647,0.09096616617153586,
 0.42890742371945756,0.9196305962819545,0.45411385509939595,0.762030159833066,
 0.9455010297066729,0.9548216290076237,0.675246719314702,0.4859399405382073,
 0.09059662832842674,0.877098341138378,0.7714707932846794,0.619175182272486,
 0.5826649096476137,0.8443985214104588,0.6638200767974584,0.3205465312276916,
 0.8333244670867233,0.38799809178146216,0.6955384360708614,0.844387097745053,
 0.2911127295447479,0.39891865416862027,0.14532123286698084,
 0.35837133771507024,0.10607514848270072,0.8289767220704087,
 0.5123949222412476,0.9128958999556517,0.28383826600011863,0.9038232433666153,
 0.3674105305268912,0.35201027619983705,0.48303200875538077,
 0.3193012675530964,0.8601154147480838,0.8963260694815419,0.20077658709164659,
 0.23913784833904217,0.7625670591402819,0.8427210519968253,
 0.08079874278142307,0.6695625555716843,0.23292688151617036,
 0.2905631451422571,0.7311523892804329,0.8228312620715106,0.15252713614829916,
 0.5938864780547531,0.531019438959774,0.2660044921454313,0.5801093248838544,
 0.8419058308520504,0.33051768329894515,0.5438811157158921,0.8366746769363765,
 0.6379591009466271,0.4127150086743303,0.057928013345033724,
 0.7659752739472934,0.7951887798837021,0.06026464074875215,0.3151377076605486,
 0.21620143620512233,0.6415172460919969,0.32233010730139056,
 0.8109440236690675,0.8644323180133573,0.4643154968779387,0.6549859701926366,
 0.8681598724466124,0.021943488768841446,0.588242811478482,
 0.049681683325454196,0.44096057534465494,0.7592952637656565,
 0.7943478349201614,0.8332930582899749,0.06318292077255683,0.6582785575555993,
 0.24367919207422128,0.4709333903806594,0.3885638077534628,0.6597706864747243,
 0.38352144225247586,0.393861921293863,0.03020484662017686,0.8358881575370225,
 0.7133749194553435,0.17222237880641578,0.48517186353972774,
 0.6273343721355296,0.09066642395978952,0.02477313545080906,
 0.5475801458376491,0.2968373305545384,0.8014101444229623,0.11313275313497806,
 0.6424491711575873,0.9951475156312403,0.18346244398571343,0.398321043236451,
 0.14070048242066047,0.1048902847357236,0.6469993286579927,
 0.48357558312994264,0.8738608199985348,0.08451093436102242,
 0.14649979514077172,0.5813666764270402,0.48915503289894746,
 0.1753183631980938,0.6638110276384004,0.583615460683797,0.22879830011771818,
 0.7353608842874919,0.13777762878044153,0.2110667846447274,
 0.42819939299846466,0.06738498155009409,0.0545678734093521,0.63688293744088,
 0.2294707508251832,0.2469000588662783,0.09969644302919634,0.0903737176614755,
 0.5985621184188101,0.9417720574940855,0.592751950557052,0.11763889790032578,
 0.1167293416356584,0.9091084341837594,0.5051830573666357,0.06992718010178023,
 0.6416677763204366,0.5957883397633874,0.6881796347703074,0.5732921090293324,
 0.16181960445912058,0.750729353388649,0.937991992724543,0.6329814639606323,
 0.11841952345158124,0.15896122090099052,0.18599655109662694,
 0.2662028170052896,0.801520994076924,0.07278106179081367,0.7068408313927271,
 0.056757220292974364,0.24708021077819964,0.47447566165659316,
 0.11154924625168294,0.6022763104297495,0.5109600277858298,0.6195521268157534,
 0.2886806439954809,0.8124045781217037,0.980938121002779,0.6438953577562991,
 0.8112311864693159,0.9365134627276119,0.11947599417674071,0.8200672578448345,
 0.8223096331083803,0.23928695294913838,0.7010789567806712,0.273335319870297,
 0.47172950450511053,0.520601787183755,0.6450729078210337,0.2820544944279034,
 0.9544836443115784,0.8870588639823316,0.40082665406382223,0.700869242904388,
 0.6852690039800757,0.02412730724321832,0.5617229184615774,
 0.007037813436983176,0.3926698881227302,0.4300431638928618,
 0.5856453083206281,0.8154391241301086,0.02206471746964367,
 0.24706221167999665,0.9052233174375868,0.28884705986597625,
 0.6025801062509881,0.8917239470774172,0.002655012156277725,
 0.5820763410276191,0.6250426676058821,0.2635268052933657,0.40209183012649297,
 0.07043292084315922,0.037399183401403246,0.44988138727619953,
 0.34317965645717385,0.8310448520105221,0.6616818288415486,
 0.11389842565244324,0.17999084525642162,0.27116888078209467,
 0.6676011250545819,0.06802028159479301,0.8986260114409511,
 0.39276550339475136,0.3085213618305075,0.4143165321090826,0.263252823677287,
 0.6959154920012132,0.7703218789676302,0.2753614966953515,0.6165154942553843,
 0.7589237479566502,0.5602972211779912,0.05638741017578741,0.6244541418258207,
 0.15920507293941333,0.16407234868445364,8.897272599699413e-4,
 0.11815752922618838,0.26173884020132887,0.09044697855139383,
 0.2914729851610105,0.9739275712647653,0.43125368652927365,0.2928724091158149,
 0.37123476709208514,0.7451665943981622,0.8109038331547856,
 0.47271615144531187,0.395273281365102,0.6406777231851666,0.23669426295996165,
 0.7304704251525291,0.7057879394717447,0.2661241027635197,0.703855349433899,
 0.6124506688983541,0.49744019812170037,0.5021994919579009,0.4323807153068612,
 0.7489748750701365,0.1651425008898877,0.31129445308128123,
 0.22637631730962626,0.2617538839205701,0.3312332660953681,0.7949574793726326,
 0.942053575312319,0.1468416832620667,0.6054578434470113,0.4624475490462072,
 0.8994174188502095,0.3768071286930229,0.5632962044689521,0.6731134764742541,
 0.7266039081005075,0.024927588272387058,0.9204175790527137,
 0.7767223326951516,0.5478065336164588,0.82584946481425,0.41058900871063986,
 0.6238923397723599,0.7604157110303835,0.20488824313627862,
 0.040625879514562424,0.5801751820266066,0.6395633281194061,
 0.35674095420116037,0.7926688045318565,0.8930419654601369,0.3560605123661841,
 0.7192018244061965,0.517342036350803,0.8815237887775977,0.040964189572746945,
 0.9600080948622632,0.9256346304114377,0.03229235577130296,0.10073204754733,
 0.9400716327826053,0.32841307361357386,0.30664310327243494,
 0.8681306346087687,0.3854530779159677,0.9521323931547507,0.6296494993175158,
 0.8685144792096002,0.2130704039165332,0.05769898570222076,0.490734293276202,
 0.5000103348035111,0.8355430307892253,0.6741428143865145,0.8933301358033132,
 0.4150454899962539,0.2466700728936292,0.7207753683620896,0.46389591741875813,
 0.49317467712582663,0.6408803401903693,0.7891120550731734,0.9535834071565548,
 0.6974136969346708,0.4055912637192418,0.5547613912723752,0.8182438765994751,
 0.12333342220028976,0.08951324389485282,0.9162306164419483,
 0.7782374438248788,0.9658180548875813,0.8112718761496808,0.8234222150523346,
 0.7555244918835803,0.3250694676701853,0.22053173846846874,0.8647873456974828,
 0.27737178094350756,0.027608306167096863,0.9028413390581899,
 0.8606500392260257,0.5640338775512659,0.010945963667193581,
 0.8143669047554659,0.6941342515724754,0.1767001390005457,0.7342150677292093,
 0.29373935296380227,0.6581612098485304,0.7671713673775011,0.4603765295619028,
 0.6615218849496233,0.07632367518313443,0.15771582209902923,
 0.3258706744790979,0.178433268682199,0.009591470328927931,
 0.16325835134941213,0.6241132749023289,0.46358926816666124,
 0.2010016260232843,0.965029718817559,0.8034588956579255,0.126176461689824,
 0.9744197113564956,0.6657966917806495,0.335702587152079,0.8028306443710714,
 0.6266772018730722,0.6510734517835707,0.21414966113945588,0.3335952353535274,
 0.8098296074613043,0.4648116244073285,0.7483521348226784,0.333236129038873,
 0.3320165167964749,0.7533545217629657,0.8786381808406998,0.6841903424381754,
 0.3278831813066737,0.06138072823410412,0.18965666210041687,
 0.19913979734127163,0.21921843381736095,0.7802689898713483,
 0.6340877668486008,0.23668053259667565,0.9153041696818038,
 0.22508077913438518,0.6251891958589239,0.25044164728605656,
 0.0072333113616760425,0.8956782265470775,0.891706808204452,
 0.7273295026303516,0.9253005147390574,0.3201408790007576,0.16053752269402888,
 0.9410213544062095,0.704885337452144,0.1189558303092988,0.7607160125089688,
 0.21517092419562556,0.5411615284024783,0.11700382518843899,
 0.35724137530953715,0.8444052305266637,0.08854077602822252,
 0.04241199689935515,0.9436543350348908,0.6954481948407562,
 0.12818091085748406,0.47585605656439534,0.11085467349062905,
 0.08816413653436772,0.9554753783909378,0.4858811047969035,0.9098814389743886,
 0.6067246041152023,0.7011916162622911,0.9533442665768667,0.33631865146476836,
 0.6980724116748087,0.21050672476596977,0.14399792713677062,
 0.2350187219230584,0.7023356069036902,0.010785454339404699,
 0.28109056173048164,0.09649342617301093,0.5201531847398504,
 0.31703605714264316,0.04556485440251512,0.9163729210698248,
 0.42218874788990357,0.5834803816365135,0.5065114797195307,
 0.18767993223089796,0.09148049461790553,0.45379383281240815,
 0.7626409025757147,0.8789744440774125,0.2932019086939336,0.8537619921137418,
 0.8706536314398674,0.41973434314242386,0.33464736229902026,
 0.07974867912665418,0.7653200907695796,0.7487890980816159,0.7563612315046945,
 0.6847548411551941,0.25194636751962607,0.7059681159328675,0.9741293248237811,
 0.0455604762587849,0.08825843364075425,0.6784298016210507,0.7830872657234218,
 0.3729145512827775,0.7839011823607643,0.43713108675036294,0.761512312413383,
 0.15912438636119364,0.8510965489872956,0.5166214948276785,0.702475613135407,
 0.9399755486707118,0.6553018485595324,0.8445202475261341,0.11519693114938445,
 0.05678835832594353,0.7027599817565902,0.47572700640342425,
 0.3897100961166905,0.9616960502790409,0.6239532736913296,0.0472901614816843,
 0.754934717653981,0.867951234433562,0.6882380502998169,0.09791396915059802,
 0.8584837229751363,0.9334280894487695,0.47491148776075387,0.5619861241632375,
 0.8629802860822322,0.12751936097886474,0.4439907412090002,0.9327258197080596,
 0.6467729227863644,0.32344308896282836,0.40764470358755545,
 0.7019595078408015,0.4957627376596483,0.2079492887952924,0.20905615870620453,
 0.8524457137414375,0.37991992764335203,0.9602462853256906,0.6089922152410027,
 0.4454424489513221,0.5981581061191137,0.3549691780894214,0.775644911638956,
 0.26217232785959355,0.12962859270398108,0.8925132545576281,
 0.25572763927740016,0.7637663891002833,0.2163758021565052,0.7182739048906074,
 0.09842541182529185,0.32625800436521624,0.8449320823833664,
 0.8706853698658934,0.35652524269002106,0.7310173854429178,0.4911421112161479,
 0.27011956192224995,0.8994675170196819,0.08231629042599298,
 0.4396926828154224,0.2116126498234978,0.873852902800331,0.8223868183481434,
 0.1605107137953572,0.6596594433770729,0.9212430973409714,0.7023344555431097,
 0.31524218101949963,0.9639489103607743,0.6404021898445236,0.5862143721563309,
 0.5617014072171832,0.1065976433556215,0.09658372505495216,0.5119256737386542,
 0.2931861088682952,0.8101747163535061,0.006990473243815631,
 0.22458174292379474,0.2053352814217052,0.43036068311062536,
 0.7387235478053542,0.14753148662504334,0.6037874882600697,0.498175015607662,
 0.12117403794335346,0.017486705738781638,0.10837662512913515,
 0.7194220448760128,0.7810358934146959,0.4896041122141124,0.16709157451011225,
 0.20632770498876152,0.6977716825356471,0.4667630655308871,
 0.30516116381585023,0.32968135664060716,0.7327123895845964,
 0.2518069737315477,0.35079054572684387,0.21912224400302638,
 0.8273517501137507,0.9617400137297928,0.8658907633331192,0.1400056649559176,
 0.36565389255346925,0.7640789960066839,0.471123424057068,0.6270103364959937,
 0.8690636184334067,0.35001131565503074,0.08329729025494359,
 0.6344977253342906,0.44556265167770626,0.2320003373695393,
 0.17378169759670403,0.22244337057053853,0.1985825253638328,
 0.6956729919390389,0.435459699585805,0.15265408278379378,0.42104530714557886,
 0.6151309774447273,0.28972497960180965,0.8583104232647764,0.1829231844102438,
 0.9315909530026427,0.4180141350153447,0.05347975191376986,
 0.06564278026404269,0.6563322262891218,0.4420397152436478,0.9081470420271889,
 0.2627628433196212,0.4407705143800985,0.6997329355855099,0.4685518595475554,
 0.09516612564778382,0.13838932044956442,0.867761595593842,0.8726345576327929,
 0.6814072886066755,0.36816224016113874,0.18344363264630736,
 0.20115213293804068,0.2721119166366379,0.777275396587958,0.14431861004937807,
 0.827972959615176,0.12614132069196016,0.6953007451270372,0.8638003516732189,
 0.28712185833659765,0.9616793228522414,0.8457140393251674,0.153310765256925,
 0.48601075415998685,0.9035091717207347,0.14030072245853298,
 0.0695392293181265,0.2251687342190809,0.4774736331413605,0.5206938620215568,
 0.09449067390933796,0.3395814413183583,0.29405154224808383,
 0.9921297745896398,0.1282583371374676,0.6772278014823679,0.178201032734011,
 0.1317882251646778,0.13532796743135034,0.6048191918231286,0.5708043424557165,
 0.7346100924371303,0.5151352881552497,0.6306035282481608,0.1289710922595081,
 0.9860841947359934,0.028702532189053387,0.6240982249065012,
 0.5020009773723195,0.8959655631143014,0.17319811206597735,
 0.46982300942648225,0.17141087437663005,0.8964004448483954,
 0.9313544491636966,0.02340761537373104,0.046913999083798785,
 0.0502979827097183,0.9018676281927106,0.8034950672147076,0.16964108950442602,
 0.9930596932008731,0.9742867705161853,0.17785094368910392,0.5121343950403641,
 0.8828137951113683,0.7308693760936914,0.8813277795896038,0.8671408105872022,
 0.1885972094825692,0.8401469170824334,0.5122061408326537,0.14465703803898167,
 0.5796781698887723,0.049260035103233846,0.7768026433850865,
 0.5284297528644811,0.9027595071808285,0.23381792126891865,
 8.338012092898861e-4,0.07169235952942621,0.8838279819723647,
 0.17056411311722097,0.39923551324555795,0.6662047815092466,
 0.45111775075481986,0.028736524406586472,0.6240926932772879,
 0.2663061944500026,0.9278920551907254,0.42724669946233473,
 0.32819768317510745,0.6736563523855976,0.9494132781717807,0.8738685617856035,
 0.9097282740657836,0.26699685432692044,0.06625732125340589,
 0.6197196692755974,0.9144141135798018,0.48351500517927337,
 0.49603142240175224,0.04651551438965873,0.97103328868684,0.6802950313591161,
 0.44558745258986376,0.7704220018211951,0.4818427379966277,0.9830397523144084,
 0.9566715900612036,0.9195727318611483,0.07401647219558938,0.9402934272604933,
 0.5316265954032053,0.9339255019932636,0.4927789153989486,0.9190738447433541,
 0.9345884249710372,0.6894309673525046,0.525036480934121,0.664046760973045,
 0.713968036004786,0.4925621931796972,0.6934933244011388,0.5560239457057222,
 0.17100555837832232,0.7155521330912794,0.8484689139919238,0.7554191465102063,
 0.3740965895144778,0.7028896492492623,0.06691387239420687,
 0.36893808529514094,0.8074234227686308,0.4969881566248501,0.9571565574312838,
 0.5577731130062212,0.4911463538161118,0.4724104991354868,
 0.028698083503087446,0.5467425664898656,0.5208768411217989,
 0.5481018892001992,0.7337037285855375,0.8718731943876906,0.22260700884227314,
 0.5806287377052122,0.3651352024700544,0.8770015948469932,0.5151850415258065,
 0.35683930159878297,0.8466916345699295,0.6922942951476084,0.596600549994931,
 0.23084578973873016,0.811674338841128,0.697722035653749,0.8878252462948221,
 0.36215376652081,0.519436433334381,0.7404428667433757,0.14987364143560233,
 0.802161832512314,0.8749993178218691,0.4827684142180969,0.060829512370934036,
 0.7137387039737806,0.3014472293102485,0.20827524480935944,0.3326388191975933,
 0.026419256987867512,0.7292706268163802,0.2668676470109492,
 0.8912466916771388,0.4203039124227762,0.6021043248737197,0.46900298399910223,
 0.6556900979224309,0.25298067290547843,0.2931901187795114,0.8715719943503543,
 0.2732879834977977,0.15222188678342252,0.02931736091131021,
 0.6029645636664293,0.6515012709521126,0.16336642484383312,
 0.14949912161720436,0.5524262067945085,0.002540619205719352,
 0.3825659749674286,0.9087191737024659,0.9482566672993732,0.03678925632058927,
 0.1869798857917525,0.4615313547561788,0.7079062990601964,0.7244750896955172,
 0.2415102644230238,0.5201305110930161,0.02290187061566895,0.384081269685549,
 0.8755069060810234,0.8183577392622324,0.5316021583441435,0.4802685350149809,
 0.7321582077793612,0.11404908388711676,0.7970062931072021,
 0.14302123377538578,0.6958157712489283,0.43367714457804407,
 0.16590401860572912,0.9896502706192745,0.0800173039615053,
 0.39807058198703404,0.5980184807291087,0.5136408658196893,0.8704942148853142,
 0.4551937386716983,0.8224065836912781,0.9841568511337934,0.5132596029151633,
 0.6204753113338409,0.39444497271585643,0.5181298185817811,0.7968598473275804,
 0.6600203724404866,0.9843494413283176,0.5228492152467266,0.12108669235555514,
 0.6392127128092936,0.9960297509574018,0.24232349433728662,
 0.11749522361375908,0.5839068216178138,0.49034753627381356,
 0.17965432028495154,0.41445319870686936,0.009625769776940962,
 0.7643812103273618,0.27210663316243056,0.7099763552676754,0.6612212251990831,
 0.34194980214555926,0.05766732185684653,0.6492597889226275,
 0.5124497322705113,0.6590794818945033,0.4459851138072144,0.24094417290483028,
 0.31880760445036094,0.0029722102008937767,0.981743918763217,
 0.9782152361607934,0.20137455840419438,0.5350048957961486,0.4659095273963342,
 0.9812907359691346,0.5355183837845167,0.9532240028459901,0.5259774212489361,
 0.6439455027114993,0.13836421266825916,0.8657860141724493,0.6383293018584896,
 0.27155642245081224,0.8498974582001276,0.30998197779085657,
 0.36539129101969203,7.982925348993808e-4,0.09084648538336071,
 0.6592647033455885,0.9975747300490262,0.27475922659081775,0.1882077795006395,
 0.277962227484156,0.7538088621637654,0.7244486543628131,0.6553448127952458,
 0.19341644412574044,0.7770101886156436,0.2130359056890181,
 0.032056154094456946,0.7719043362993785,0.8318312500105189,
 0.5962938098956568,0.12961556071944325,0.4228656417313632,
 0.09157392593648517,0.9015720855248474,0.5338678795462604,0.7571879506535198,
 0.010537347567213473,0.33567695617082594,0.8892782706940342,
 0.023132193093313824,0.5930481325657482,0.45944232688534126,
 0.7533743435544796,0.4658399558371955,0.584482639018515,0.8168789921248314,
 0.01815128344584649,0.873981097105242,0.5180881454306487,0.6126275739378978,
 0.32234594238570147,0.3003745107776651,0.9772211754829909,0.845559804388703,
 0.999705198862211,0.7579575261854031,0.9642364478551582,0.8154290452135715,
 0.20037085190022563,0.9465559271209337,0.9520686675096808,0.8948513931280515,
 0.8878985219986453,0.2407145343171373,0.3933125781725867,0.9803194000457636,
 0.057174826441176774,0.31048187229910706,0.8795559604619816,
 0.8622616941125092,0.4593762959704638,0.1506257042271485,0.2636488295633983,
 0.45003752524033547,0.5428493590835505,0.3578057842478293,
 0.32720861040599913,0.4096319129161191,0.628919238175268,0.0891425630953554,
 0.381147099080276,0.7925082533051032,0.22133342143648116,0.2686394334513993,
 0.40255212439118815,0.5447287111289427,0.013042628120863875,
 0.9790851603922434,0.48450994549462834,0.2101525432409792,0.6521203198048886,
 0.15338274203447055,0.7463629300139601,0.24482362505219812,
 0.04683553128113571,0.6358243386632129,0.004424001800579336,
 0.11548198662907416,0.22240524451350752,0.3588167820121416,
 0.1424298988295728,0.29712898344628613,0.6405390322758406,
 0.009223538451994195,0.05084228303962479,0.7404334821316061,
 0.052786376310064886,0.36619562679111217,0.00861528745489526,
 0.06150533743259168,0.9992075319029351,0.17915076278290243,
 0.30676905056031645,0.558448094193409,0.0813913199663604,0.10437017334602117,
 0.04002152746536902,0.5993967981702703,0.5992042421225373,0.6745382053586803,
 0.9560706144962827,0.7607211810699926,0.30547048323325376,0.9774181034380215,
 0.9270726732268855,0.8778342831885912,0.7704984353788793,0.4558819879025252,
 0.24703550813790998,0.20703994638513545,0.9438834318263838,
 0.22026345832617145,0.3317710964771622,0.9489785167580266,
 0.43375155842878255,0.6759617282012441,0.9871259831561231,
 0.008562466252854395,0.4136114779217651,0.13949222922711968,
 0.8915870214078178,0.11041926461004548,0.12351360179485027,0.740949460039774,
 0.6602833559281965,0.9751114319022097,0.08559292765013682,0.1375876416523878,
 0.5082618123107803,0.14935984081331466,0.4818032301452527,0.7798419984771212,
 0.4315939156508174,0.7847455363243876,0.752131054136525,0.6179149776641205,
 0.34287415008120736,0.08417294132737863,0.8120442444698271,
 0.9317755650352705,0.3529768963208124,0.3888942776634323,0.2956229950577449,
 0.19241848063887146,0.7817703910644731,0.685764401111169,0.8977110017221525,
 0.6053731649844837,0.4474396011975339,0.9299826243837684,0.448868186247946,
 0.019854988240183502,0.6110682891905052,0.021598972787818416,
 0.9705568462199051,0.5682952973692794,0.4969412651470253,0.3470183429140862,
 0.905219839870418,0.3897410959969967,0.9151008242093306,0.227888489095737,
 0.7011302379714894,0.8703168659894817,0.7151706086874817,0.9948874734944454,
 0.6032622440975299,0.25060658850804673,0.11889333056975859,
 0.6791542402695241,0.09973372663851465,0.08985979352994755,
 0.28475875642994974,0.8709064152358073,0.24127646700605165,
 0.12098059918471282,0.5779121745421562,0.4192285502453015,
 0.25627107336934984,0.8708946009677819,0.6721666664423662,0.4136604858012841,
 0.7076834397652717,0.03547910763146245,0.923712207389027,0.44928973610808176,
 0.08876390295062997,0.16968297900486706,0.7123351857087512,
 0.08748911409268756,0.8759719294800058,0.3863876318138664,0.6679990033047899,
 0.9854593806286547,0.5599276404000374,0.7003877795181185,0.7333667909500128,
 0.3771686857365314,0.9025611495623722,0.19665909126560333,
 0.03602731686715965,0.8916568455130471,0.48759564067469896,
 0.7470340873039645,0.23912803491933965,0.23212028598869616,
 0.8334244357272288,0.8155873093636027,0.15048538833097913,0.8785354970867498,
 0.7472935402758416,0.4783841903703113,0.39308472593048616,0.91327095354589,
 0.5904817410231429,0.5261166913462005,0.37950384279179206,0.5076907927376134,
 0.36066227502715287,0.8187240363624047,0.6275101117931869,0.5963680658859507,
 0.5691382699282821,0.42819675632807896,0.9769875226341639,
 0.07399836643623592,0.36052396195271574,0.7085740185602966,
 0.025966760833782243,0.38602938471403825,0.027260983836221664,
 0.34105000824817966,0.19973141041072762,0.16988865255882568,
 0.6848729367334974,0.7252187758241394,0.47188707034141386,0.2030964142710987,
 0.7109656137369024,0.10877383482692471,0.8827227704834949,0.5063220735264199,
 0.30640163275678867,0.46333191128319406,0.0768600159124363,
 0.02617254265543545,0.9868011806523481,0.38108682889753365,
 0.5208210397122975,0.6219379323755524,0.8305890710036077,0.761082957201885,
 0.09261117620773218,0.524254620684324,0.8380067097376149,0.1912641258563038,
 0.15588092238433737,0.541158131671024,0.39001286958266923,
 0.26038511329456393,0.6344608939428985,0.6064494140303419,0.6410395105808577,
 0.9563322557618906,0.05614175351793782,0.5947130951687226,
 0.19911678025226565,0.8760384761245162,0.4304197216275263,
 0.26118173921591914,0.39487007074900693,0.25544309844444046,
 0.28555406365898073,0.27150372294256486,0.3544758886569356,
 0.39246168389549496,0.7378200501266304,0.4112958060793219,
 0.42022301269337936,0.8159842904687116,0.15784961860270685,
 0.7413539933336664,0.4766214584522287,0.3650755283866046,0.4356791476693326,
 0.3343674620755023,0.971553550920673,0.4858195113436541,0.5742598711991467,
 0.10654153766805152,0.6863639621405853,0.7740575602084692,0.9170895571776069,
 0.6386208039309315,0.19001610645746347,0.31189330645127555,
 0.7099575190178851,0.11084327535599914,0.2882800066899176,0.4710449072142314,
 0.41130481591006596,0.7778842812038159,0.39500031333067154,
 0.7954798828272319,0.9129394202208595,0.6830260429917993,0.8703538751817079,
 0.4851048425031874,0.9498861543199871,0.8394665351307063,0.9462687583920005,
 0.8428833561215592,0.29662849944452774,0.4900070452089643,
 0.09680951578002461,0.7993136210889928,0.6976433068927002,0.3784398948235469,
 0.3821423859750046,0.445097667705846,0.5620369748999887,0.8630367495650024,
 0.34849099292381114,0.17508013903887598,0.8923954439217977,
 0.8989392366322694,0.8340323967368161,0.33866038398368326,0.4196837495383634,
 0.24963654303320582,0.31924290533578037,0.6170919653229249,
 0.3749297973003187,0.5856854993131249,0.44954349912738945,0.5840319687403479,
 0.5465532378997112,0.9944827628364014,0.2141671363823563,0.15126034316613457,
 0.12291345655911068,0.3041642469974025,0.9125800365226251,
 0.44979668098016234,0.15634004292498915,0.9148775949026337,
 0.8919149922420969,0.6504347416629886,0.8010549057488262,0.8781805792099096,
 0.9080900017565281,0.5902098851427283,0.7048750427303592,0.5615703477163245,
 0.9769955576838568,0.2283757603635369,0.8821605605346773,0.9623105969213012,
 0.7989803429805009,0.07045589362772287,0.06997254788196983,
 0.19393190213150968,0.1473456063209987,0.8289224801805093,0.767470303771173,
 0.5436868290652886,0.11808235529884625,0.4050677760432816,0.3151571098420556,
 0.6446664546538448,0.3868510923487668,0.5788756659406289,0.07151426820341511,
 0.3580607350992089,0.426736721618834,0.2665083748340773,0.7823379129823611,
 0.9159024429681959,0.9388586940761061,0.5441364579900858,0.3931896401299272,
 0.9530518517266107,0.7716780326622944,0.34250405764856173,
 0.19840400208511855,0.6002121044509404,0.7566991295687093,
 0.06865809999187267,0.639268631268828,0.8604739283880095,0.9685290143718893,
 0.3680420823384969,0.6115457718333716,0.12888868878125104,0.8427595844430423,
 0.5279208921602432,0.2562556005050808,0.9148049743972964,0.6569527514261362,
 0.3456491584554543,0.6485997502180036,0.9993909376213771,0.5051188983951062,
 0.1385848301637903,0.4270851983951388,0.912230913030621,0.48338104401700366,
 0.2725486486265818,0.6540399438037399,0.01369727108718044,0.9225099071764331,
 0.20351578771831003,0.6194722648143007,0.8405951336317369,0.2651447751734064,
 0.9501893596865127,0.05435653513514249,0.03271321364236046,
 0.15164825742736232,0.4476547658315504,0.0294012461095271,0.5380989873980191,
 0.36528111154261167,0.20362737926979224,0.4480671706556898,
 0.7960401688994992,0.4340491922798705,0.49356391961305557,0.9585822082784983,
 0.8566608156751543,0.9235545550180911,0.09258538212649015,0.9927458422425124,
 0.25222337066543943,0.6706088775220181,0.8398204102582286,0.8950995761573444,
 0.07092657717435846,0.025072515296327036,0.11765199676402416,
 0.4183472768090032,0.1894196821928089,0.49052143125947345,0.9734485188987482,
 0.7614856620562221,0.1905769755712361,0.6331087815209964,0.8141320885506511,
 0.3946293793911184,0.37572517315994436,0.25756748548942354,
 0.9581051174441129,0.4793214157670972,0.14053335032719838,0.944053920221346,
 0.18290281986184365,0.027195652656438085,0.48969418052295066,
 0.9417356224642847,0.2735576856037958,0.8583721774927369,0.09140974980534977,
 0.10075264761123826,0.6464259078780916,0.7559505571963115,0.8614565904878879,
 0.1299889396702345,0.6612297528042247,0.08402209307802966,0.9818490332227473,
 0.20121471793251466,0.231279695182386,0.642987842460972,0.7276652873679249,
 0.22224629705122312,0.5839801559715518,0.950963577757451,0.9188720683012707,
 0.450447011723915,0.022181221565216735,0.1293636705586514,0.3564640205282541,
 0.392180689575391,0.9812863818315501,0.7304134763840375,0.3368947487807551,
 0.5658393362324124,0.4055195589747068,0.3549541293312072,0.0904862108393627,
 0.24801163571998597,0.31973869249046494,0.867889953458401,0.2714279707189903,
 0.4210275320056408,0.06006197285556125,0.5806251202092136,
 0.12781265895615723,0.15658747244149762,0.8348312374847122,
 0.3115017844946327,0.9503598013801261,0.8465234336699078,0.12508009734067171,
 0.5579197322079174,0.28692835970699204,0.9069119706251138,0.5409235658329113,
 0.3840285963948401,0.3942840478986458,0.419765223839046,0.9674846779372435,
 0.4163738916736942,0.42850799633984105,0.9346731682115137,0.8393658958517669,
 0.48417291029301557,0.9915343727587389,0.5136044521431615,0.6260552288688875,
 0.06895625840245356,0.050690773036712544,0.5697381361062707,
 0.07455142553849903,0.44212142222829764,0.8463758371928793,
 0.9030545898883994,0.8692694311948662,0.3966230294866171,0.5945235206045205,
 0.3516443246283637,0.6715050887301479,0.044837586143584685,
 0.7429562329081849,0.27288946828627303,0.5701152192769953,
 0.31518593339738477,0.38262259466717197,0.5590722452267411,
 0.24833240984774774,0.7275793163722627,0.17005101753531926,
 0.4460221937222971,0.2640162673160622,0.06309736418297429,0.2333808187309787,
 0.04799758840179169,0.9769232503907536,0.7603730903080667,0.7832708395322944,
 0.6413010768286025,0.4849607040781412,0.5890242895388579,0.6516079062951305,
 0.8673029925776811,0.6903343991389281,0.40898290494974565,0.6924892798969667,
 0.3952638511607608,0.49268992852933735,0.6473570174921541,0.9974543561647115,
 0.8670452715659895,0.4701993823791413,0.31981102887573554,0.374363594261252,
 0.1603235621581458,0.3640846249628076,0.2615848951988413,0.7490223128237922,
 0.26563762678696146,0.6643538675332258,0.38267009613498215,
 0.9517231861367598,0.344117451769658,0.19430967601461446,0.735049235181616,
 0.1299655087280659,0.3910479609360148,0.9396697502469493,0.17769543601686488,
 0.4135728397482461,0.24230702308225127,0.8510625507516304,
 0.04612658905720168,0.5785504946820252,0.7585390787658249,0.9237906411861063,
 0.2042589789441892,0.21354194129067006,0.9993828788387654,0.7725754001890243,
 0.6225439633343768,0.04136375285989735,0.38903309103970296,
 0.2474825828359022,0.11174361111223596,0.5575849498587598,
 0.21886729311453568,0.09384666174388834,0.661217843672742,
 0.45897533170216065,0.5894085305981935,0.7356341400244024,
 0.16682305468587688,0.5255516774382922,0.19992545670059547,
 0.6862500897524476,0.18210075735493936,0.42851471872691593,
 0.43479408366694716,0.7967151459736638,0.6368753242933376,
 0.023675829699584816,0.35093838593164683,0.4489614681069102,
 0.20084636211457985,0.5994695847419446,0.9180226214707978,
 0.07493525708953297,0.8130801546286175,0.645193617368745,0.4716757954886379,
 0.02770266130385146,0.6269193017429625,0.4346998391313003,
 0.22437091605638126,0.08929030817917616,0.5871576055658881,
 0.8778107927595112,0.3418905767843108,0.7011151157701508,0.19424806592793564,
 0.9633719762128671,0.14070326851525206,0.7884925966006803,0.2023227080079414,
 0.7336492739485282,0.4548408773118582,0.9993622855262696,0.5976820664431469,
 0.9775580436997658,0.89538963226258,0.08387851726716944,0.49803128907826455,
 0.5138876444339551,0.16621648684147672,0.8918011137123222,
 0.24988254471604965,0.6399830508270054,0.6144189971236282,0.7508480764567724,
 0.40086154127694673,0.1837082560935165,0.6520300066280371,0.8010376393380079,
 0.7996661906253548,0.5267297273927465,0.4223868608679675,0.1782215735512378,
 0.8074919580690353,0.9902926738161116,0.6590847109306737,0.735692752012807,
 0.17364537274197334,0.3653978188604502,0.5825864181282762,
 0.24209579027065187,0.9602378940644744,0.09966189560620708,
 0.17022329521414103,0.12443381385449226,0.728818075752228,0.7479977001839635,
 0.790822537040728,0.8461302923569981,0.9171850853183365,0.5476728837624893,
 0.010843376292694384,0.9379873348984237,0.7333268305349723,
 0.7929522960072316,0.4063764187274417,0.8017034948703287,0.07877244592433286,
 0.84288155311696,0.07632669976479378,0.19060324508659765,0.891155527212464,
 0.5897821988401565,0.9068407274561633,0.7167005608958831,0.6476739164254591,
 0.7245632251505689,0.4814651469970095,0.6960661855354814,
 0.0013975973831614752,0.264313543664094,0.7124253646667914,
 0.06401114744706793,0.21538498624813474,0.3550354144297655,
 0.03650287898944948,0.6818142071260006,0.18762560374305892,
 0.8789991875564045,0.2629302177105922,0.4399006297032846,0.6540695810216388,
 0.5020925368198896,0.9780917902031332,0.12918076908114107,0.7628914104747915,
 0.42047721401413396,0.2613811031520257,0.11114646483684187,
 0.28495752226166815,0.7133242625273253,0.18421132398908102,
 0.35354980165664873,0.2554584142408898,0.9588911562370824,0.710401365425516,
 0.4684877473937905,0.45348268549117665,0.5446965951505178,0.7360750651958353,
 0.9934263855856963,0.22642782085581525,0.4384770274365487,
 0.35641899753729545,0.5314640693772091,0.06937711600440544,
 0.19934896632749144,0.05777262310492903,0.08475021012239825,
 0.3935969242928735,0.8740894634973242,0.6810559598604634,0.9969510691229713,
 0.31164346437314716,0.42516243210070404,0.3063822481448737,0.958060348919328,
 0.3029464334831776,0.6705700848905835,0.8892053161549137,0.5146694021576352,
 0.5481627619860442,0.6692600425808051,0.45103761647783414,0.9955026933992339,
 0.9797351190678356,0.37857317116939526,0.7676317466095051,0.6331320980135511,
 0.07725568421507978,0.9971328263582319,0.025525620207279,0.7984394360978132,
 0.8026418459580029,0.9005179280321176,0.5268234484651888,0.45316674089164277,
 0.9137854582428239,0.5593291564070267,0.4237424188534904,
 0.011130654298682097,0.9618104393711991,0.2224242887567276,
 0.5121289356696159,0.18212473508516813,0.22121675459638845,
 0.7076691117369954,0.6673918338463739,0.6668989789736793,0.9464530567460365,
 0.34801428221697317,0.9922082058528393,0.5309145951195374,0.2039079930235519,
 0.5062875217231289,0.16670091159192957,0.5694238738186999,0.5161784902393372,
 0.07248344544931395,0.1560433204118843,0.21878161257289885,
 0.7298479744267993,0.9556937126165135,0.2959846161129096,0.14277693890277443,
 0.2799522038234007,0.5958050150160006,0.3261204104241,0.3139550731059789,
 0.7307418458119825,0.9330722844620228,0.4152685795544512,
 0.0018916152931367503,0.39787197382183004,0.22129913740206264,
 0.13328625756852963,0.12296176439395323,0.11990798593968877,
 0.5876039205922771,0.23127138786416412,0.204518832271967,0.23188291207324985,
 0.611596765280697,0.1533959148543469,0.6226033668782427,0.23352710248416486,
 0.3950451416252395,0.5913102011865052,0.0899822735433139,0.09844131237746068,
 0.9346074881697342,0.31072006409266417,0.3766076643111478,
 0.44353002831647015,0.8788600110530632,0.8666916495258645,
 0.07345746971550715,0.49196611540657775,0.3421052344043377,
 0.6854259773959073,0.35593865574604167,0.2596986325210221,0.7608133511108893,
 0.8885557842587494,0.9839044319422063,0.1165913121035631,0.27979909206101206,
 0.06744116414148826,0.010161202833054883,0.6987735639474103,
 0.4049649797119427,0.4398158784872075,0.6554728945735014,0.3325924997442823,
 0.8832169865305496,0.1498509943793106,0.8594470253528889,0.8524774685832156,
 0.11071902803812772,0.6525221837880507,0.25528215708986346,
 0.44611175049619733,0.5227543696888644,0.46216561064601747,
 0.24439340095326523,0.6537143198175399,0.0853188734616881,
 0.41688896746461523,0.9704788858749478,0.9050875887698431,0.4886284371221188,
 0.8160492834499502,0.48358292647610857,0.4425076940993552,0.93918931135797,
 0.12857319499080389,0.03303079986758217,0.05918227944402232,
 0.05874979288645288,0.8882349513773087,0.02998733857048641,
 0.6322119876684531,0.6477171893525747,0.058809034316606,0.08224029740407823,
 0.03494523542979999,0.9049007765240977,0.04408405144544003,
 0.7070816381620006,0.5457801026378256,0.183062290862107,0.9997077291115001,
 0.47742391511768023,0.12586918478163123,0.35859519067172874,
 0.18271657025341215,0.8376871560177321,0.9475586927216855,0.295625840302288,
 0.129230862888849,0.8636327521822234,0.1682492959568196,0.26064109994459816,
 0.39646881839670456,0.6976628339447375,0.06983456515384245,
 0.8990274921352032,0.02883416324966359,0.6074414314584962,0.9365335471265375,
 0.7320325591372465,0.8826062851962808,0.311057763426568,0.023519565094196282,
 0.9802664813407856,0.781454911153861,0.8011465592773542,0.290732024175238,
 0.4494566830591401,0.14121057518849556,0.7355386551849921,0.789000177123571,
 0.5178342674624714,0.3739406261099313,0.05518139253343235,
 0.11917648128744296,0.4715681757325547,0.8370213777411696,0.8591758243992949,
 0.5202247328814344,0.38291802690683396,0.0923232861386033,0.847142209281227,
 0.650245878275768,0.8305969629707358,0.7283844662551776,0.23095133182934235,
 0.7630978597288317,0.3990759971910257,0.10424008512372551,0.9690427241089731,
 0.8739130186142462,0.46024558192160436,0.09817974422565001,
 0.04693752299290899,0.15641584143978426,0.034029280823278385,
 0.918131438936918,0.2997631326573271,0.39115532504399564,0.46586863382746224,
 0.5042658441080344,0.6771304729205162,0.5752993945442425,0.7819955884514226,
 0.2085127508584773,0.16626598419467065,0.5307325359014765,
 0.28382305392170293,0.23924007120147261,0.8971497134202321,
 0.7479979280971755,0.9542988565570034,0.6926426068454709,0.6524418875823658,
 3.318958717646314e-4,0.13899631375088972,0.9015543459179882,
 0.7464377869469324,0.016953757916325474,0.48735858337744753,
 0.24725671114134062,0.34227877409734453,0.9161079665702199,
 0.6687637497424755,0.39044635189547805,0.6112141235177372,0.7975293083689083,
 0.9580094445730005,0.16060362010444518,0.031178849687532817,
 0.2874085838385366,0.4644241094382645,0.4067814444049894,0.7186526836737243,
 0.4190986520345944,0.0012634317828107422,0.1528893570241654,
 0.9814498682846625,0.6331728092429333,0.24668923342965687,0.6886996340126286,
 0.038770133025948894,0.5944872710777369,0.837029714670905,
 0.23862558740947792,0.9821884171748477,0.3475854023406463,0.4345934805962992,
 0.2016120861897458,0.9807650020217319,0.5698115622653034,0.2525444944554257,
 0.6930597089125416,0.6821103978575559,0.6021262347470558,0.5120292641382539,
 0.9209654780335936,0.6906922573305825,0.7767394832767756,0.7878766737355298,
 0.43764350481304126,0.046256026229060665,0.9008831433370817,
 0.6645202755694709,0.4748737777823383,0.6648606842585685,0.10212249561860731,
 0.9945597911958216,0.0819310898837774,0.27020169443776476,0.7468995295619445,
 0.8751413048012592,0.21006594068732443,0.5502630504092633,0.8721033164871643,
 0.8695066145964357,0.5294987480629749,0.39296566430114055,0.9657366493493887,
 0.401167350622465,0.657938347134948,0.5942404231620997,0.5019229818324233,
 0.6098570864953073,0.028428200868619524,0.29560545985727915,
 0.6715624482456454,0.2632186305232801,0.7100955429736,0.6073389129030241,
 0.6925759434481499,0.37345829976013833,0.6762543046255551,
 0.04095202187963376,0.9189189372176432,0.0998113316035355,0.4519526433400395,
 0.9095214692704753,0.8458336791638366,0.6799064747890765,0.8768497438936355,
 0.3929155766889707,0.46422770696197735,0.8948614647035815,
 0.34243638386601627,0.2577741103735426,0.013457096877392827,
 0.432037021504054,0.6348882695501035,0.7659513998405881,0.5591779072628901,
 0.8487489875512353,0.8646477741219215,0.7608715842671926,0.4900252626815067,
 0.6814309669431524,0.18510454416834266,0.2838114062977387,0.9506930067903694,
 0.16838448142724005,0.4208913422494214,0.846735249513118,0.3234726191273868,
 0.6112883120129129,0.4165599513692063,0.2870613681356369,0.9418391024668809,
 0.6127821968431475,0.640459255966563,0.36819972079291574,0.7932187188421198,
 0.7857762178200157,0.09325722901091726,0.5478163384951932,0.9018079079413043,
 0.7140992608647461,0.16681178723572399,0.21579878611628178,
 0.2843997389447003,0.6591279670167656,0.24883729504971075,0.8516923853948419,
 0.47495333880813007,0.11929745023451932,0.021875352614165178,
 0.012622886780173315,0.37900563918643115,0.5606687752834134,
 0.1805631767149174,0.35605417633267267,0.7085499622605039,0.3121755506384797,
 0.2785566543816359,0.21578327557911048,0.12757252598671354,
 0.0556739457272688,0.5266012963042663,0.9886798880487777,0.5638219298082054,
 0.37464791163824473,0.03841936024371795,0.52405963578771,0.40397456343165916,
 0.0676064181630045,0.0530059529632636,0.7443628382207491,0.8029615967687583,
 0.46080342078253955,0.6679776262909302,0.045083537742453994,
 0.9273149325877967,0.4548956476297421,0.9843977080485223,
 0.021390791713363666,0.2298551840663744,0.5955667668525816,
 0.4345746175546149,0.5482933910606018,0.5751029676228135,0.6759799350191763,
 0.4024864717984473,0.3123687666203758,0.9463430304851628,0.7354585683981975,
 0.23921165414160273,0.11592159803926017,0.9237014214056679,
 0.5121050094615656,0.19555018112710443,0.4363922064344482,
 0.20584030892281846,0.722554397694634,0.7300090965789438,0.4371858616726777,
 0.36156184711287753,0.9540945525060694,0.42944745113325433,
 0.03243194467591348,0.2338127072661269,0.9902468386093127,
 0.06149984525938468,0.11122191364654455,0.18559815590727657,
 0.337396850877684,0.54361592400941,0.7788936485600453,0.6657244087131247,
 0.4381667661990233,0.3688145659097478,0.6118317111310444,0.943892053858797,
 0.7631564988015948,0.6002550641940195,0.03450141831258213,
 0.49370860044740483,0.8986418417177133,0.2161627740046944,0.5172790028746358,
 0.5754056685613506,0.08673294927642805,0.3183216190837086,0.8964064010942329,
 0.438620997979563,0.7292371955387709,0.7268405555486126,0.3283697550000505,
 0.6915879941752897,0.25548967227476926,0.6047550333571277,0.9671060551166133,
 0.49683598147028674,0.31821618520137274,0.3993083271954927,
 0.19298425218777338,0.26287776913700334,0.8219449099946223,
 0.5055785329355175,0.4528998298242737,0.15481233217210266,
 0.27338575910034857,0.5451286747769237,0.7079603660886227,
 0.04793045780871508,0.7079209771919639,0.9170147110878162,0.3712224523738299,
 0.23263471407182612,0.6016920620979599,0.886270730832012,0.5810424441389089,
 0.9673954718451986,0.26567334211217897,0.46505420818223975,
 0.39182726393801126,0.9034696924348279,0.7695489788086066,0.5463851783031323,
 0.053374820045244076,0.5262130748590839,0.14162934523152504,
 0.4574687532515661,0.4042772630601438,0.15122977335553633,0.7845693011923972,
 0.44959081763841147,0.006821536057566524,0.9708909048892564,
 0.8710436637611161,0.5798524014472757,0.20524903692506358,
 0.23959158312310092,0.6266115331615874,0.10107076718742469,
 0.030219133091054506,0.8912041445007715,0.8116436761099085,
 0.44215986391551987,0.18677305188112114,0.9452649597898062,
 0.40649179129198654,0.08178726385428148,0.28281387116749246,
 0.24483527507077196,0.09033330370987369,0.4140593773250254,
 0.6756533461391113,0.5903121199394832,0.942278575317036,0.5204546620349123,
 0.6934322493066746,0.3504256922320663,0.74610072323657,0.6788391714765342,
 0.2945820285063865,0.8525817975493611,0.9209302582554231,0.5769510039306184,
 0.32697363069353336,0.772528307777454,0.8128779018458689,0.481620333354823,
 0.060225272868650426,0.5299809790668789,0.45753036483395315,
 0.053202422613648825,0.1329802664715427,0.7477262577868742,
 0.7738351745138667,0.9819816321291555,0.09830396730688262,0.7689199856584832,
 0.6405009041694154,0.27482049521464536,0.7382014962102188,0.3053597061239227,
 0.3150718650726515,0.6294761529379718,0.8952641596045584,0.6406242343657239,
 0.5346089424542562,0.823924578053041,0.02659700247978536,0.4912760913600106,
 0.19460142236823208,0.09936142931701863,0.8214027327531751,
 0.40527227143338307,0.9136969800371266,0.4584415984877529,0.8280184448334306,
 0.7236472420781012,0.8397543527772701,0.4797480473368947,0.6356276759083069,
 0.18194364189360068,0.663971094684324,0.37148748692646205,
 0.10478607657340211,0.5450291767445776,0.42241335011359604,
 0.7683177759721715,0.5949204641439324,0.4821913814885985,0.539411949428769,
 0.929734590415927,0.02429768934801635,0.4238203840977597,0.41315863154690735,
 0.8981147712225954,0.18025388230036898,0.01191881852133525,
 0.20227174353588517,0.013871825577522534,0.1796552768638644,
 0.06023933514377977,0.7429284076910383,0.48283056872053376,0.833269460361359,
 0.034308707946201555,0.147949324947506,0.8200800744556057,0.8719573856459583,
 0.01946387314928444,0.6710400931390276,0.2332732450505437,0.904410999448881,
 0.7047028029775735,0.42855027196645434,0.6060655639871336,0.8436661783698622,
 0.6268232951892939,0.23308607161052108,0.6687962932383756,0.6754443939947085,
 0.8657943448927099,0.6935956530284021,0.41313878219081657,
 0.17354074482259052,0.27260796176305513,0.31240872830735733,
 0.7703863026900779,0.6714919936215296,0.7028461202191272,0.8992242839844842,
 0.8127868073023958,0.892251967108629,0.8238319765320252,0.2682294249623277,
 0.9139892093124449,0.1163870026151459,0.8688355731446818,0.8522355988781491,
 0.7219531947049342,0.31265519029726974,0.927059622522697,0.7756293970722884,
 0.4058696208959661,0.034651577297368785,0.3963836149663845,
 0.6042213064843461,0.7991118371273699,0.9110694054308234,0.8150543015655846,
 0.2487214726916145,0.8672849081178917,0.6389818734155522,0.24743478156471366,
 0.5594400888661741,0.7416281348692775,0.6333693017909299,0.7347755722011422,
 0.44630861408253075,0.012606022976249243,0.3748998531169825,
 0.9422624905262432,0.34232492559055183,0.6411768698516355,0.4435701815304093,
 0.2891457490287732,0.6025342937188927,0.47916023511257544,0.3238127187870061,
 0.9759881855674998,0.4788110640108384,0.7689131557871641,0.36914167145134,
 0.2823333196013099,0.775310058121791,0.06959914011295365,0.2461469924500156,
 0.31763166403426957,0.039868366270173095,0.9668139897484477,0.94653176742866,
 0.45293414664257603,0.5713784205141581,0.27915956333947334,
 0.1138359728367494,0.5384896724166921,0.9321668393471165,0.24037714382740072,
 0.5290038877010668,0.49532403500767075,0.9542296525838809,0.2184864894944144,
 0.8643094925642035,0.7471582038841955,0.7797056686909116,0.1673825045886037,
 0.010406511519405415,0.519900482640735,0.6114772613316567,0.2530964884496536,
 0.579683820638395,0.568578009056515,0.8167166781186695,0.3308729698290205,
 0.8586355550436539,0.6834008437669877,0.43166503323535077,
 0.06814209725078124,0.5272314829446488,0.8720119928863659,0.4420499151234021,
 0.23703477802161477,0.36884778873541946,0.5716941935707355,0.785826125042586,
 0.4607153179739003,0.8911485293931273,0.3968387913779309,0.3184528532265938,
 0.5186300025001614,0.8552113973640838,0.6832060958123918,0.18723541173970215,
 0.7236389818805328,0.18204439851662302,0.20087428541429642,
 0.0926785044356283,0.44937854529128607,0.6985078770706415,
 0.02273208135223248,0.19835614461608542,0.8630395453356892,
 0.3777460649482205,0.22426793831114566,0.4057394942237743,0.766316708879752,
 0.19588983425519246,0.4977004926028459,0.2872638739594602,0.8556359867583856,
 0.09271817236931257,0.29320315601917,0.4178484812258879,0.9288608381147024,
 0.5184093156200872,0.4443054899380059,0.925591500816245,0.8300865569069854,
 0.6150425967207397,0.19641750156492643,0.3519781039800831,0.5640586983341942,
 0.27530619228385844,0.9949372912960152,0.4623952615764785,0.807103267066826,
 0.9433319525568155,0.03813320965291034,0.004943581266687924,
 0.0684753471701709,0.8897146373336967,0.3883176066553131,
 0.005255724819659857,0.7983070623749444,0.3525373724776283,
 0.22169033118373516,0.8626203950002616,0.5711665830270352,
 0.19962889482416746,0.5360142197923724,0.5499565297363153,
 0.33541951107985213,0.5697837245711063,0.1467842335066213,0.5804721299574596,
 0.36867818311691447,0.830480024180194,0.024617528147120704,
 0.9218387439485193,0.6675715984461046,0.7364852094886247,0.7422021119790142,
 0.8188906589856171,0.38792144393330286,0.43435837347827944,
 0.8846175649278647,0.8614979734459381,0.2670587241889142,0.8645617623418737,
 0.2792096946121272,0.304730747750776,0.2936534633819844,0.3818131315577401,
 0.634660217414701,0.7050688353536736,0.8993834271025681,0.8003983128946763,
 0.9246498412545868,0.7834661721884948,0.9755469141962859,0.3255350864148392,
 0.17221914031311147,0.36469423816935187,0.9527547620587211,
 0.7630283500602377,0.1094671154941469,0.31522219545158947,
 0.40321211805031065,0.20171696638181236,0.493793486251619,0.3266859565830029,
 0.8852003349070325,0.6192171069270845,0.8910717826810809,0.3788766847257924,
 0.8636429156014501,0.9144028297878559,0.9675812748628724,0.9525192746530151,
 0.5607400693956249,0.14387399423082947,0.9401803692297547,0.98753960181297,
 0.3743996740048421,0.4216759751114306,0.3646621602622546,0.19699122423678317,
 0.793648456390327,0.5377870136006969,0.3071699077180394,0.8326996966676868,
 0.6350947485525319,0.27737180439821263,0.7253271592931284,0.9936672535749558,
 0.902994447013421,0.6357791092738104,0.22952475142806672,0.895287789085292,
 0.19330355551522604,0.5697681182853629,0.09551954453659806,
 0.9322712624529995,0.5548663380416426,0.7065775370422426,0.336765752602583,
 0.8489094232566481,0.4527702066133831,0.4176043146383227,0.19051619284180266,
 0.7370883420883345,0.16804370755639753,0.43147621729911545,
 0.7938443279967845,0.5873033858969485,0.46978422127658304,0.2822174042806055,
 0.4853249167585262,0.18051680638716638,0.3483419505450245,
 0.22649130700225728,0.976871563507643,0.7334633332665887,0.48954767932526067,
 0.8295230951131749,0.7434722226955118,0.9864779088123499,0.35011338531048675,
 0.0016462784929908159,0.2525952105639565,0.9707684651726781,
 0.7037758287886875,0.377311117304976,0.20668167573203677,0.4478002929452366,
 0.8804569325373007,0.14422475561252956,0.7947310083204253,
 0.20392556167064113,0.7660968789483051,0.039862312401776734,
 0.2304449117286539,0.011672981821930661,0.24467219610613344,0.14614828639486,
 0.9434352849496106,0.157522713427924,0.7788047245854586,0.1599972672947767,
 0.17057418583778605,0.1977614123475856,0.7185682270447823,0.732570083334568,
 0.08117938057088125,0.46190255162343896,0.012680916340116921,
 0.9819933100218449,0.9293061041475141,0.5370497781099751,0.3114911935757244,
 0.17645709619962568,0.7449053643638697,0.5031017880542284,0.3381276733664055,
 0.9203777230761202,0.2156277233673931,0.2223879618964899,0.8381757359548513,
 0.29262881552526243,0.0019068171675910017,0.8692123749343695,
 0.4491977132687853,0.814184609227699,0.42681738831665006,0.16988808715431325,
 0.16773340550282778,0.16435615084307564,0.9123920999008495,0.721921663122457,
 0.18419473770250605,0.45559394280779497,0.9644104757731282,
 0.5785322035324508,0.48774376599659586,0.4110932091033781,0.870216237330899,
 0.7148154533135828,0.22234780247758168,0.5837914092707251,
 0.18374688167489772,0.34555985959368163,0.9291095171270535,0.90218606197713,
 0.31684785767243107,0.3696411210017909,0.25081297758788124,
 0.4508292755140664,0.326946442599784,0.10466212205375014,0.3895141490931774,
 0.3504670815055144,0.4488598248018981,0.9992723461245053,0.2909444644066729,
 0.1230213341328239,0.9480255115919937,0.9497792819244099,0.2706305889285283,
 0.7821907842094665,0.10015499168007425,0.2962974346104139,0.571123241733948,
 0.8053269713893961,0.050012576319903634,0.996344382793084,0.6299492152582881,
 0.34627995777081466,0.2155382773595742,0.5851125888195687,0.9442793775155897,
 0.3136781896871388,0.978795653641866,0.23716455083515953,0.9688555070930596,
 0.9581665300555456,0.40943055076683765,0.20302470336318623,
 0.06527244531250864,0.03741438977361822,0.858096549548759,0.8690403851282282,
 0.8834822101767271,0.8984551860223069,0.19626407824962566,0.4059539944579581,
 0.5569912836384896,0.9388516590374483,0.28570345130243835,0.3120205791536197,
 0.6213070979832381,0.692438519903917,0.6130484881205269,0.6454122191838612,
 0.15882631301601902,0.5049385615289261,0.46801160908351636,
 0.9857619493297852,0.08677548511345179,0.23482556584976555,
 0.01001593392091471,0.27184096166554494,0.11490212558296031,
 0.4528451631647238,0.0926919052881825,0.5562283129092995,0.9815478245626456,
 0.8235152032611324,0.23840689277049332,0.9082388532914478,0.4614979380128318,
 0.055758806608613654,0.7771125784612739,0.3444267290845342,
 0.07053310810984414,0.1871102714501195,0.9666779068006135,0.9380674169140104,
 0.841200760933221,0.8017881637722859,0.9960155369351559,0.7682878933071161,
 0.9502445327153921,0.7018516419432598,0.7618853916950399,0.6697566074918051,
 0.8915346454132181,0.9987027647150893,0.032302010504297596,
 0.9178055169296923,0.5780350054433449,0.38400344762022187,
 0.15947981216907914,0.9488268617313756,0.8721256841788727,0.3035090969814176,
 0.6586742336527539,0.8702949231225885,0.45127064689626684,0.7730501608872504,
 0.5020595311656838,0.10746718955286194,0.04146391676143857,
 0.001896010810930715,0.2579199700172664,0.659333335811517,
 0.06105498624942918,0.4770646091189018,0.5917568770986518,0.8061838052010537,
 0.5418901293047895,0.027137366032975896,0.3442873480910984,0.920839697016451,
 0.5497097875763857,0.3357859696988361,0.81799046846167,0.173010466673869,
 0.15673603959542148,0.012564058248481058,0.1742537704165381,
 0.0856534200040846,0.5261683456691951,0.0185344764362374,0.5235140402346457,
 0.33713893343660484,0.1628037824556393,0.14076636593816005,
 0.1682341221333291,0.6548821231295746,0.37557630441982115,0.8260224745964961,
 0.3133591481370823,0.4981728964634018,0.8645447827446675,0.41773293875295836,
 0.6296340905917106,0.8941633697283975,0.3745260054451005,0.7801561779910614,
 0.06881141340774155,0.15758659845584055,0.8036948424903668,
 0.5681938703247946,0.4766473731873768,0.0030576841176163105,
 0.8339987189415133,0.9442112272499864,0.05258823934657608,0.1783786043443255,
 0.6365723296365261,0.07795710762206776,0.750611554836453,0.16072162726193773,
 0.26983295697366994,0.09559060018243537,0.8063141344916148,
 0.1839983268234373,0.3984200418342777,0.1186075120362009,0.05158876800207057,
 0.09681415440117425,0.7887502447267576,0.08108678949343617,
 0.8790226847099235,0.33723751380497935,0.9520749991213555,0.7745445961635791,
 0.6839805324715078,0.43966667894799083,0.3053914287763453,0.6840718306564952,
 0.16941044354586565,0.3330005675643847,0.4726329800766572,0.9628458920314911,
 0.25367746848877426,0.2881141473842088,0.43211703709130367,
 0.39046558228389994,0.2912783405853505,0.9009854075018354,
 0.34593619956358235,0.4815344514185925,0.18026118050823736,
 0.9000687658334634,0.7772013343253484,0.011226463160492473,0.732017747401381,
 0.16251294770557534,0.21500439571499486,0.26886958012823814,
 0.29076556601577774,0.809485337875106,0.61149763514817,0.5214570710126014,
 0.20266579114867933,0.19283614478446354,0.8858236385579012,
 0.7259620849584781,0.7542879614202579,0.7385897060375677,0.9008437314905149,
 0.18639290744488501,0.4095202443498053,0.6008505269271768,0.5548971808723682,
 0.9091267179904969,0.7895139694025968,0.2925844797322962,0.3268495411392526,
 0.13671723520524348,0.3717708337210488,0.09782645255399314,
 0.8055641461152797,0.908633259243584,0.0273498043219379,0.1749329581366258,
 0.7363044774793179,0.712879486236975,0.6137419697328277,0.09298002681643847,
 0.9796784995731779,0.13996780608481418,0.139293089380236,0.0846073063570878,
 0.33873143741310996,0.23013858840821066,0.7922313435931372,0.518774268737331,
 0.16416871653026277,0.05404450606236122,0.9911536430159349,
 0.11056425819528415,0.11884995158556289,0.10389664853163705,
 0.0874041004277254,0.053104420416129616,0.8734565961432835,
 0.6440194861014354,0.34392001631503244,0.9719178004836074,0.6336977696437218,
 0.6714400732738252,0.31935730697605713,0.10580688228308888,
 0.8181042837549168,0.001940183315347399,0.27396089483258235,
 0.05580076578094595,0.9682009454908561,0.4668264175434378,0.7977274539536157,
 0.13239372283797624,0.98734286290108,0.24727643733794247,0.07973124057005732,
 0.30321052441114216,0.7452314934166029,0.7093048094450316,
 0.02445430416783445,0.39236846633499745,0.6520097962247898,
 0.3566573581093677,0.2634157108600964,0.09905471305046243,0.3252607448914643,
 0.30215312745963707,0.8064920629927597,0.6275861086378622,0.3294209941711924,
 0.41718830785312666,0.6370038335999549,0.6147299910143958,0.8609893374998641,
 0.14658445543058962,0.5282239421375038,0.6649265150571536,
 0.29636022961932795,0.5172162167410288,0.5539848075741204,
 0.05968276652028037,0.19131148548161092,0.29790371125648285,
 0.36579557142510266,0.26129518569781285,0.8839911509099438,
 0.8992375916215287,0.751964537236766,0.17056900470750613,0.4822712621272186,
 0.8901469139783346,0.41757473660873023,0.5237836756172397,0.8457175207194012,
 0.6930584462274115,0.6878090624797234,0.1979871289611964,0.603636977730823,
 0.7281826309749992,0.8401411513753211,0.10616315568006385,0.9507266323688113,
 0.832397846995631,0.00363394664304173,0.899452716001054,0.14748236724828678,
 0.9890141908588032,0.7115952301934949,0.6957089876086292,0.9849808141384166,
 0.28185376447343835,0.5593334439844427,0.5791274149583033,0.5460190967019682,
 0.2193407278614069,0.28125757021394293,0.020197751920953633,
 0.47297052459502587,0.905193861845051,0.6172948708715289,0.04677722137536999,
 0.43987649786533756,0.14220664823460205,0.3364677107613384,
 0.7381631979022072,0.16953127371613252,0.5621097982822386,
 0.24605971115942382,0.11832363668512869,0.5624813815354278,
 0.4168118532142844,0.3454374145440191,0.5424513561857769,0.8105027879652873,
 0.5025563497731723,0.14058818155070618,0.4464609993445733,0.8494538524618409,
 0.916218469045518,0.8866766824366846,0.6313052403182371,0.15971126491344223,
 0.2133371365766935,0.3884452501894784,0.21482102756495436,0.2108628406053903,
 0.8912479609687546,0.5469656143666147,0.9303561154149887,0.630810247906251,
 0.01507049047944764,0.9790689987097503,0.6067174318014458,
 0.025458298325060857,0.09280528561906731,0.6603120181864385,
 0.1919906549328484,0.6480366489118712,0.15758670458278146,
 0.22429577254213418,0.5809353877810075,0.9437102614444433,0.8424398066773812,
 0.12435307788605587,0.32085363892661956,0.16848201006194496,
 0.019166744026593352,0.256821900072679,0.9949307598944531,0.9589721740169245,
 0.6139068945895717,0.5550601902493423,0.022349380167687616,
 0.04716718398089781,0.18737284600072712,0.8683760852251202,
 0.7121426827792693,0.9736696014237909,0.8074473051824038,0.6926276551360907,
 0.811537509888681,0.17773307589448573,0.5724778220019164,0.9285321799040911,
 0.059868889344845355,0.8714016169217175,0.12251555561602057,
 0.3809803069468827,0.9637449018946697,0.21894962408555607,0.7639186440169558,
 0.418699666769463,0.012142928540639475,0.5333834295443449,0.4538474378242935,
 0.43206397832043786,0.3557623161870632,0.23808621096344806,
 0.26026423968364487,0.37350109075727644,0.5415867702925543,
 0.6838172169733098,0.17817487954782596,0.20329767834586365,
 0.12783768792414663,0.6434708672751226,0.7269725615074449,0.7269553275113477,
 0.9287141276607017,0.734016863558156,0.3445392724604701,0.9719177420347912,
 0.7134841367127029,0.8183643603304357,0.263934511580014,0.1296850979516958,
 0.33245446903875175,0.8485095152843548,0.6464406033136426,0.1583005922443812,
 0.8321018586740316,0.5703463095521912,0.278970343947357,0.6851035880004844,
 0.02662842340992355,0.3315846759946989,0.5888977302053224,0.6378901933851967,
 0.17980079092648682,0.7858456263041099,0.8154607394069932,0.8667394226627041,
 0.8613998392522544,0.7877513731811528,0.7901196700770067,
 0.036647281524790465,0.917451308302971,0.4812077179144234,0.3885320548532104,
 0.523796486112845,0.2502243276664444,0.6882431592063314,0.4201463742204695,
 0.09793190054316081,0.930456558309497,0.8684595573242652,0.5288953806891872,
 0.9690693396169512,0.389710991559329,0.4689503980537384,0.8896888274603033,
 0.9986384230581977,0.3685837238488914,0.9086884421259981,0.5544678680141473,
 0.47014292487220644,0.2776384024217673,0.2936578681036205,0.2518174710423283,
 0.4416499967695151,0.7841959778091769,0.42682771981281364,
 0.26917646300876497,0.07264919394638303,0.7389270880209947,
 0.37914571161918953,0.04947827015153061,0.9971362124255218,
 0.35142198152774484,0.40199451329308844,0.6758932811372655,
 0.6015585013794943,0.8975854361792261,0.649400013747834,0.17569647065680205,
 0.021296993567054168,0.6124672340224582,0.6631505861495349,
 0.7648148242610394,0.5265146475765828,0.9679182173588098,0.6514849712793677,
 0.09893711190540644,0.3589289793091136,0.8688361264100692,0.3166133537447383,
 0.054043466319413924,0.3115306706350711,0.050458134984623926,
 0.7069318724100703,0.13182353119188328,0.9197028531887113,0.5319145504077192,
 0.42518919972270053,0.3078575259814843,0.9886120727517577,
 0.20682873762115594,0.8103536352274675,0.7937800194501153,
 0.09464446343905653,0.639894049132254,0.8399606524305789,0.38194123816402803,
 0.6079304670089174,0.7130736103108444,0.6283071012852633,0.5038477728113245,
 0.5062582256656273,0.330824879331174,0.8819674660051344,0.7859345505760732,
 0.33805309669258465,0.9360694290370116,0.4791205341735777,0.5623360481203157,
 0.5975400375341517,0.9991967900121923,0.3760044747985487,0.4352720048248917,
 0.5635886303662166,0.9815885471838106,0.04491570842387993,0.7768465125338098,
 0.5920146371004529,0.46710716851983414,0.795892069222879,0.24753239924152926,
 0.3317494743360623,0.40053468012681837,0.4094273604812524,0.8834352065580068,
 0.2281175272889937,0.06589598595029356,0.7043821894477545,
 0.09582807605308286,0.588665804364454,0.4288123343102547,0.44301030584851664,
 0.9585501739730854,0.8195380636693277,0.11790337923765348,0.8437243417339254,
 0.26682008415368963,0.8688130339353284,0.6504253541916625,0.886247822298673,
 0.06924376219109929,0.8774955884202577,0.13012896821083197,
 0.7057333449670556,0.8553936467268284,0.27805804289858516,
 0.43063696681518215,0.1877496097537893,0.02868355271966372,
 0.7759231679281017,0.694088160175147,0.8772039020147685,0.23232936173829755,
 0.9860929134548522,0.2372040855545714,0.9467728950807509,0.42161115789811143,
 0.9246954644375154,0.46383830745377963,0.3813355960133611,0.7991131612112508,
 0.43990307778157944,0.08889540729493006,0.07556354309853375,
 0.8608641179725811,0.6646781253271656,0.26282072490967434,0.7181245509449142,
 0.7387190027683082,0.04244606416083807,0.15436482874631718,
 0.2578936116261741,0.870990401820144,0.6227682576509629,0.6036521794341735,
 0.059117981134695685,0.0016499869030246472,0.011951302387026796,
 0.6786923508371573,0.051603298327562164,0.11782267188715667,
 0.8082416183373986,0.27107963660712153,0.28551600523494614,
 0.13718297765241516,0.3417805436160528,0.3738885358810673,
 0.31151017157787886,0.5715233476734692,0.8541651414365791,0.9763816517115438,
 0.6573855123862895,0.9506445803656611,0.2425156180409337,0.9684117900294916,
 0.46576606153321043,0.8089232722702092,0.2899237635991285,
 0.15032137676355384,0.07298611751615025,0.2317995787523658,
 0.7879134199169192,0.5747112479047155,0.6164494062487783,0.8076190910088299,
 0.9013111473966109,0.04399758707633483,0.9162429189696348,
 0.21685657241270861,0.3257761621952626,0.6899362501212396,0.6118239695708346,
 0.28759481833022904,0.9750421767595067,0.14805105433217391,
 0.9527282401945537,0.42857337844415144,0.9925538557549664,
 0.29194465314159546,0.07581142082838577,0.8178403463428292,
 0.04760279102817133,0.9805339744367756,0.9058862454668716,0.6931320222874067,
 0.1960936907356814,0.23590439064880275,0.2813090484085172,0.8068238815892558,
 0.9766533063116348,0.27482479286882355,0.8859676752464375,0.4411979506480268,
 0.9260821639282953,0.353993789679271,0.2861991976952989,0.6743552258302348,
 0.9777066802974778,0.9415023301485482,0.06707416432013968,0.3857795819725046,
 0.16821493669442877,0.6627228371967645,0.6190047811005237,0.6327021724035855,
 0.22868725607964802,0.041150639716549864,0.06004664294969764,
 0.5261646155585901,0.29694055125168894,0.3291005717207941,0.7666460244971322,
 0.24440078822230082,0.4950100251462062,0.05623055854118064,
 0.14258542714110212,0.7701287218904997,0.9392663379381528,0.9297116366983564,
 0.4293113418176907,0.956970868316664,0.846215179429969,0.8899620498468669,
 0.6887297527338863,0.565613591990185,0.7652296828740057,0.02387168205943746,
 0.3255876121138289,0.523920660436801,0.4545203071726214,0.40080559878445166,
 0.05060939170147105,0.8330114559495069,0.30606396850684614,
 0.2650299531968463,0.0684717655946131,0.45522855473899637,
 0.10220267143962614,0.2938011164036848,0.8111484172392102,
 0.42810441744294836,0.56413262422466,0.2372055605065898,0.45133693898420746,
 0.13153666492157767,0.5836596607808862,0.7703325757389987,0.5272244207132655,
 0.7406626379133712,0.08571958809040547,0.406138954465634,0.7683329091127229,
 0.10859289189672239,0.9749430876930154,0.23256507056047437,0.77012485332313,
 0.28473777440130754,0.16283448405916623,0.31358935331256355,
 0.850350638288257,0.8623860377447155,0.904859244365978,0.6485005770598724,
 0.21021115447494165,0.9454308462534421,0.7587715701785833,0.6644285045855409,
 0.3911145395013237,0.33153623767838303,0.9625498679676274,0.5181176006893804,
 0.5778571131409169,0.6609886006560307,0.2120391938400199,0.31361220221907815,
 0.48988760686073174,0.4626015259937327,0.19925888569208983,
 0.49307708252573157,0.8234391866085783,0.0016398141166912295,
 0.6631383327747017,0.20900181424568398,0.37922125242614013,
 0.2852287218750854,0.22616880955193053,0.8692449367466821,0.8221894260309758,
 0.9181044422682023,0.5057268675554787,0.4067965750654644,0.1675150156598526,
 0.6941503906473095,0.49853625660200973,0.021662255764225535,
 0.1725880493930666,0.5908678489020431,0.6094724495265442,0.8320551063097871,
 0.24918581765230297,0.49117322439144695,0.2516391420865901,
 0.10759139714685939,0.259237176696839,0.7231627622478133,0.29326272795553165,
 0.8729937907918996,0.3606183846880523,0.5126641644876917,0.13069645300088872,
 0.5665666113854879,0.17820888221494602,0.13249376545520253,
 0.7899075980954277,0.8756515177688228,0.6817062592390981,0.6458159953588887,
 0.7977546559638033,0.09171916292127891,0.9267249422627362,0.5266802367595605,
 0.48664825910250187,0.19924908884382098,0.18432497833161243,
 0.5957233048948454,0.95111720868107,0.2426936831508517,0.8630190593347552,
 0.11575480292089069,0.05491920410943962,0.8309793318252999,
 0.35615587006718363,0.9192040985030339,0.7130417355008649,0.291702045656694,
 0.22070224318069498,0.3805367633617259,0.08040754632896452,
 0.6404847642742392,0.5606471422692544,0.922293439553357,0.1564826091690934,
 0.014966536874874925,0.14881049937048807,0.20999514191805213,
 0.37698738806690657,0.7751046976102117,0.12206059735925656,
 0.6316279444501162,0.8506218275487742,0.5567244035125442,0.09792351637278518,
 0.7645152586325082,0.7708618866197949,0.520415014360381,0.3943058790707682,
 0.034740087352872484,0.6219182232058292,0.7609020330952454,
 0.0309078190329507,0.04861639168043508,0.8041267084824754,0.7940043759949605,
 0.7260132376954339,0.029956271206447393,0.6919804949665838,
 0.39165019585937433,0.9924128411386391,0.4501441187499785,0.6725038330134743,
 0.6464043730218157,0.6352962443144813,0.4721085332972865,0.3387942913614822,
 0.7148624344386432,0.11308545920628621,0.93355129123601,0.20212214649233995,
 0.6460894687263881,0.25225662936243853,0.7176448773576232,0.6532749814747488,
 0.3274091052439462,0.6972432553828899,0.11107891630779076,
 0.11538838650467376,0.8686747800233504,0.1678611021357207,
 0.21633215318691457,0.17220935602393506,0.4925712265712654,
 0.9309543379555285,0.21014216799270802,0.9521782966669704,
 0.005503237508879599,0.7661187961657333,0.5348144630424028,
 0.8232065880323592,0.40826226445931013,0.8310777722329563,0.8304607758035026,
 0.045361663543715736,0.8424691351591361,0.8383910628613972,
 0.31313688819772845,0.6664995976852036,0.697976536213121,0.0946957781378881,
 0.24496057668331483,0.3615050564006942,0.7666543554524222,0.6221721893098667,
 0.31602519061466494,0.6451066768459203,0.506933010110677,0.8653866933278012,
 0.3208388160297626,0.5982770122818765,0.39282342063071185,0.5806178928435024,
 0.23569844442811028,0.02193352889823119,0.22759473827196708,
 0.03292796823828459,0.6020408329739573,0.44303490846796556,
 0.3836930969921828,0.38145594408072336,0.2689709864803387,
 0.33998600509368326,0.5599250146831779,0.09108884460249889,
 0.21648272883797037,0.3527017942607861,0.3537204012106231,0.8994135592773198,
 0.07849541355235112,0.23908377429828576,0.6467698073549905,
 0.42639212198131893,0.4910938587726442,0.9432999016480389,
 0.35913058346058724,0.8848171945083778,0.6959462321100854,
 0.35216773446835736,0.7084852436033903,0.6010853603732844,
 0.03505792066820146,0.36913608272180065,0.7231942809376857,
 0.13183324606474234,0.8889876348421456,0.7395788537441872,
 0.23224971308900721,0.9037361742048194,0.6352579131978419,0.4139658031487973,
 0.21133647879976714,0.8913880685159458,0.8103162347026401,0.9655678880303131,
 0.9568220675255005,0.5841893555102614,0.8565525636187501,0.3429698900652879,
 0.7034005272112516,0.25023898336403916,0.5361560695726968,
 0.23234911954066528,0.6748282247890991,0.40222950618477726,
 0.8721900835565131,0.47256068262018547,0.6986209569058575,0.962983300219325,
 0.6775211134951377,0.6510767576857838,0.6156007592576518,0.16002786263184188,
 0.08673848419233698,0.41898123855602054,0.5963557583253507,
 0.9280763577674864,0.9954032270351838,0.7351136678819777,0.23680836335656996,
 0.2818946478521853,0.048091043908357056,0.07482215554556182,
 0.4640564344067747,0.496183755890526,0.10855649200943884,0.6483713721448674,
 0.9297262418080114,0.10368048434189059,0.5893978384781332,0.6166059766484157,
 0.8042505145874872,0.1367095508251841,0.09007904221940177,0.8125804830845742,
 0.13347502906165765,0.02498704613895275,0.9765193624142695,
 0.1281416382431685,0.24908726468051856,0.8803292046545322,0.5071294550659277,
 0.18066490942585967,0.7869180567691627,0.1256377100784679,
 0.09471057988113518,0.5881104064017375,0.4425319700346614,
 0.34941905634918824,0.9373515327773171,0.34461681683339596,
 0.3478502096420065,0.4469983037129758,0.8790954031525882,0.9288168857550332,
 0.6248181753873363,0.31661347028632236,0.6196677418871028,
 0.35099800611334353,0.9294048043309383,0.5314203264177323,0.2223793476038265,
 0.6774820926187237,0.9024158520925386,0.8557595211950557,0.3164176908788616,
 0.47690866599665904,0.9675796684269412,0.6547600029055276,
 0.14107181504080046,0.1250532148743032,0.9542392573132445,
 0.23936115118582535,0.45401846928304557,0.4672521719536009,
 0.8016147342360007,0.2483544721293427,0.6630288832193711,0.7792394564740657,
 0.21333016188809273,0.077152634360387,0.600926470477484,0.4406155017456749,
 0.23687469775351844,0.32673123866277287,0.4149235623101598,
 0.8112624532003007,0.30462462559715975,0.183289806161551,0.24045496985728676,
 0.24078747605756545,0.08901242641813212,0.36998562787305833,
 0.6867456917471065,0.4130505735699407,0.6055840282233472,0.6955793210671437,
 0.2348878279822344,0.7632822621295625,0.5200023763465426,0.11308284245648537,
 0.05893589976807512,0.5951759678309951,0.9317248847954259,0.4259200519695875,
 0.11279857993893672,0.9518483158092418,0.509953573564675,0.3989970325288582,
 0.4106349731828356,0.6136155612976754,0.36049310617316455,
 0.015799465412735914,0.6604315675608534,0.2013059808001556,
 0.16676369049997397,0.513416329083613,0.27419228057206047,
 0.20264241231431446,0.12536241932944314,0.32545457670879374,
 0.15374034230947442,0.21058184480913167,0.11945562888753036,
 0.5506256187329571,0.5470757264224564,0.8561431265906533,0.5822427712261657,
 0.6393442219413111,0.039550963947758255,0.04309693460105146,
 0.4447712220867175,0.4270468370992212,0.7338902447496487,0.9186877077253366,
 0.41853022869054324,0.8673595550205163,0.010160120491721658,
 0.5600290719269656,0.8261265944699296,0.9734177065010772,0.4058892770107351,
 0.3784836099787948,0.6799492523920722,0.9395968805814084,0.8154563440320183,
 0.6885196526004361,0.3695577517134181,0.25540936427578254,0.7383387453396202,
 0.8143860737873638,0.2401108514851933,0.3816334564898912,0.7818487927174481,
 0.5749453603274292,0.5372136687499147,0.9048092440880892,0.23619869862664256,
 0.26403652610001527,0.20306478275716544,0.3009532633075678,
 0.34848023194530475,0.4585846802592013,0.7264254862645942,0.4911782556248889,
 0.5805820459083542,0.5077464597831459,0.2406981133731232,0.13919345962871155,
 0.31147651352999817,0.36682848718179095,0.8786290820039697,
 0.4154766923102807,0.06536315499642298,0.904272893865351,0.03898531832102814,
 0.09841057723618873,0.6224978826143076,0.5339477662272649,
 0.44163906210996806,0.6799585694719777,0.12441289112041032,
 0.5058725824705813,0.4334445215610755,0.14883235796384697,
 0.06361284940242617,0.5369891803836564,0.2056974195826783,
 0.06385325765223782,0.7250946130422096,0.4304318783604091,0.8340424532546092,
 0.5595283834895781,0.7413616981232463,0.9629536301559334,0.6499106493030972,
 0.7202540639741061,0.5557170912043936,0.22493656671192552,0.758952989566324,
 0.6054415561576645,0.645829326697825,0.34185207558380515,0.6706623805165153,
 0.7589132357234146,0.7783252488362085,0.6536607826026007,0.6024827071475181,
 0.08831184804322945,0.5058704017715143,0.8241876464924993,0.6224697093359958,
 0.3446760129330072,0.050323832907738675,0.9469166519549388,
 0.3547296340167928,0.8298049971093178,0.8379637621375864,0.3520637555447762,
 0.9406281936050942,0.1387052765651905,0.3964092065098086,0.7173082806275412,
 0.5604276964690038,0.46419971938309057,0.5973869207435438,0.6216967413378257,
 0.7720098548941224,0.641788863334436,0.335962627214456,0.6591076787318654,
 0.26394145155719506,0.5418407463699024,0.02981345653679801,
 0.6006272982922997,0.1612019265332728,0.47546129106888146,
 0.05643416317405958,0.9103571470367331,0.6721384244957351,
 0.42344758318958664,0.5686560362730042,0.7850787280252207,0.7696692468691673,
 0.08743086067180639,0.7219050755384481,0.9634828806611305,0.135205372356058,
 0.8826869810408884,0.559420279254013,0.8022853358881522,0.13207211366990224,
 0.9352152493861415,0.033091205324234485,0.22958008897761495,
 0.44240481634818174,0.5407931078809785,0.6975369435942518,
 0.39917105616214066,0.060069075913603664,0.9334200879112663,
 0.8160872914589758,0.008191951620783566,0.3431314763734329,
 0.24207818368490597,0.02721238213240451,0.029809121241768488,
 0.9223635882905068,0.00474249610042099,0.1772038712816153,0.6848222701205228,
 0.2482589401585138,0.34654113645698914,0.9725322325334398,
 0.06106363015114358,0.6930717587414111,0.4764639729646224,0.9645761194119843,
 0.11909375450142401,0.5976934969460627,0.1931226435718174,
 0.04639161340008868,0.7308731220161973,0.39079636814843344,
 0.4319400398528631,0.4641218399790742,0.5399299149457302,0.7935655990914899,
 0.8512291089561078,0.5917628837200041,0.03820677624346791,0.794245498402965,
 0.5768333886256588,0.11502312463906927,0.23115599954136568,
 0.9994321209226131,0.2313049272651415,0.467239847094045,0.6947482601673434,
 0.9213266847283081,0.36021804434705185,0.3479327108356397,0.8918605611602546,
 0.7724439044876378,0.7458073395213602,0.9215546511205208,0.653869276580325,
 0.06458490673242556,0.15334535079394462,0.6606746079108826,
 0.8400801070107446,0.4688771958530511,0.9958032543375374,0.0511729558625319,
 0.7332575009598067,0.7806363055083122,0.5460365858573792,0.3094090868128292,
 0.07539691458803888,0.18829091033393464,0.05253292370456297,
 0.027191960807114612,0.8270389129122186,0.7604038072984152,
 0.9114304277723111,0.8558359675502935,0.48836795378103615,0.9440930016960309,
 0.34430805569187783,0.2573413072564441,0.03748946249257856,
 0.18887812099192136,0.833633500128713,0.4890666219228035,0.2632763130771436,
 0.477421331894047,0.4092107210753385,0.44289666933122374,0.8064929363714466,
 0.206629057071958,0.02732680637428475,0.9958820264867474,0.16775104020181297,
 0.5021407817056138,0.6715793006082199,0.25086720097086157,
 0.37688520915915247,0.5555220480456143,0.43904324608459044,
 0.8891222973373152,0.8284655645423492,0.5331457148445817,0.4670626015531393,
 0.7430152135816003,0.1323435641399897,0.3849761490699838,0.21152007381075888,
 0.9778458690450443,0.2094858377035571,0.2976287315457832,0.972765480473296,
 0.429221666964425,0.38965729899876367,0.22459726381116074,0.6886385424025843,
 0.9843127423178337,0.14554391239222841,0.6371542237409161,0.7343642282124886,
 0.3965993015797974,0.08362051083277633,0.5991524202533538,0.9350408075528046,
 0.6059157954345918,0.5785144953423063,0.8256538610371241,0.853712150934334,
 0.5218216857304115,0.4664030543702208,0.7803398602438345,0.4957514548671633,
 0.4940950205647632,0.6790063946466243,0.8817740908018019,0.9791096932263442,
 0.3874426447132421,0.6263410827509654,0.6473453378771695,0.9971789925747933,
 0.6823971863565277,0.8997898257879328,0.06049630223631153,0.9119786932970233,
 0.6692294713023783,0.26207421609404835,0.22956620798268046,
 0.11991502673456322,0.6126397167270781,0.9924044943780506,0.9399766637680373,
 0.2970388552487768,0.46212015624259817,0.9333996051550099,0.9375405398987415,
 0.21267537294065075,0.5629385993089534,0.3984963815786744,0.8104645874193219,
 0.2627295313793072,0.16447071366108446,0.20975484037450753,
 0.18930556202582238,0.3260931818838877,0.5110492309641103,0.4372587091836354,
 0.009348609827748322,0.9660524231286358,0.86484069292226,0.589390029159414,
 0.914861213730054,0.5071516420298334,0.49628493391632944,0.837260484829068,
 0.06713852695441669,0.12497730678565011,0.033328523723831105,
 0.45274925549427725,0.04661411774027335,0.9014510860111307,
 0.16918850119910012,0.918461267601503,0.060677351660548884,
 0.020773638404861794,0.8959017901337373,0.5606664030285751,
 0.3963151349096177,0.6485686464498258,0.9228929537265864,0.15175674058744248,
 0.3881540469071004,0.25881751759459126,0.4986887661992746,0.6962954674513819,
 0.8815787854207788,0.44450279109753776,0.24279869503559903,
 0.6416731901016849,0.5724168456969742,0.9464399279013254,0.8790673838969636,
 0.9232353365522655,0.2854732374541905,0.148802361845487,0.9976503315257522,
 0.842848549293148,0.9140887001298383,0.7410445825466586,0.6537036356638413,
 0.5291568383445099,0.46640414885868553,0.5015697134663007,0.6238138227856246,
 0.33696612685010197,0.9379235486828199,0.35679477916753466,
 0.3210623254766669,0.18803417156552182,0.29714391959199293,
 0.9954609010726931,0.6423591305940043,0.029536673215317055,
 0.05419476251933941,0.371486523108036,0.5250820627208104,0.6030312034876437,
 0.558387918924847,0.6376431573961461,0.3378443868300547,0.6527571919269938,
 0.28402158311850056,0.16094603298962495,0.9337862784670806,
 0.7706998771403888,0.765280488753731,0.9221110717355803,0.26545914665003467,
 0.8623150920725132,0.6131186456346878,0.8985015494504713,0.29211966934575484,
 0.8721686907826802,0.520705985057644,0.021173593789714507,
 0.26299047362634786,0.5628689403779146,0.3330330936096638,0.3015053498739767,
 0.960311328125003,0.1254361284862301,0.00920114316599896,
 0.028242853279169422,0.8776810259020161,0.7146762244079959,
 0.23590502349341191,0.556833459584935,0.7037340492120284,
 0.054076282935486386,0.6235409914264253,0.010536677111047665,
 0.5050754547992681,0.3646987181747484,0.8514704171038422,0.8732853850797397,
 0.9004008895753194,0.18588611880814376,0.14341428077356133,
 0.5328256650003306,0.3454183324737593,0.6674753697984511,0.5074003369589382,
 0.08542180813536437,0.07977016101903522,0.5800090532455974,
 0.40188790530925844,0.2682777854573728,0.9967077286041433,
 0.07872736945025882,0.5591762598879831,0.07590864228211891,
 0.9186689735989495,0.7933455639995344,0.10253172301258329,0.68330260257279,
 0.06833947677024732,0.5147904624605752,0.7416429343870012,
 0.08235583117905843,0.9835216136004795,0.7625730053864667,0.3060596195757921,
 0.42963429508021533,0.5834275278715663,0.9086211905392352,0.4845912928011713,
 0.35086839401386216,0.35097821741877044,0.506963083001293,0.805684248927809,
 0.007800555896247174,0.5823620236277499,0.6737204903259366,
 0.33210621171408494,0.024176570314842327,0.0869863315642535,
 0.20192428377715588,0.868086411235855,0.1739140018379126,0.4183907106150374,
 0.43601289743699856,0.7180303945012756,0.3647665129583334,
 0.23708206814672605,0.7402321581618431,0.839671781643193,0.3494998779078885,
 0.3465044472250213,0.21492997528143643,0.6059564779091625,0.3729134445202942,
 0.4279215217791056,0.03257785083756459,0.03968929458666082,
 0.6465391568656078,0.05400519473900334,0.10623630359160685,
 0.37388348948107164,0.9086638388760775,0.5101550510474218,
 0.19081355206799366,0.9212175379503089,0.5372067020721422,
 0.05464869806318273,0.8470815034880288,0.5935236288851545,
 0.45422566462775693,0.40773371911601963,0.570264482589663,0.7929942717977179,
 0.7663324165198373,0.994575645398013,0.30484006988476875,0.12640749706693555,
 0.685752204570603,0.3940617231364102,0.22471737827309968,0.2041549947959359,
 0.9125085861457919,0.9996880000942752,0.34436612792867616,0.6685021057427268,
 0.2556165806229511,0.543602145886373,0.1964758463309757,0.29814952758943336,
 0.20114006506585802,0.5751358071081546,0.24640701552911526,
 0.0693437331481046,0.932867825664579,0.3922828887771943,0.9273359876198775,
 0.2656175638121463,0.16438219866647197,0.6760605807494955,0.2854199828792999,
 0.8109342310674251,0.1720584569026522,0.03698308534797845,0.4058961044879297,
 0.5388633685441127,0.40212447041207333,0.5320918370340165,
 0.06857977488071532,0.3991671050332525,0.09073361406258995,
 0.2188554849822254,0.6948701657438257,0.5047174647490488,0.8676709987022573,
 0.5270026092070346,0.26231673144332934,0.10195898497478428,
 0.9840907632710448,0.4326307685137969,0.7675325696123091,0.2947241410979631,
 0.6961479981532512,0.2895476563220165,0.6848747217239538,0.29438825421475445,
 0.8669977784064817,0.2625266986297693,0.7869697118441044,0.806466326335986,
 0.6700335854997367,0.45797690138064584,0.09984515853233589,0.583893469804983,
 0.7899967500045563,0.5124964663482212,0.46934096606104114,0.9055008824457381,
 0.11964993638366683,0.01285009176418117,0.040357437256344886,
 0.08996645936348469,0.8274870173459634,0.7033669717039914,0.7416590366189093,
 0.5920070531958319,0.7566528537257481,0.8441443867113084,0.687441847137455,
 0.026126506066773136,0.12282783216959359,0.5093165258179453,
 0.26613509764301513,0.03944097643943145,0.5023629694662617,
 0.029233632416222566,0.7268547657083508,0.32643334093305043,
 0.941070143515116,0.4236165696460241,0.5194815591269893,0.25039866426658786,
 0.7082957080478177,0.8720657645970078,0.8674650001070903,
 0.043353705247057284,0.3658795392044264,0.4700255311204804,
 0.5885449431358669,0.35175491279712434,0.6293150955799106,
 0.04813096220275692,0.7630062372904042,0.4257659672365022,0.769838012699159,
 0.1603697250704481,0.8544087263933313,0.7928534103456758,0.25404775754154096,
 0.24062533409196174,0.8796879739759137,0.6596789820245929,
 0.02840091079675988,0.3648214243765665,0.4282700869110375,0.7194436652049523,
 0.07263360156119936,0.9779094350601808,0.7319960617376499,
 0.14184124880429871,0.04243591056174645,0.31765334872409623,
 0.9930025778447306,0.03391758876292095,0.9178323577962586,0.6168797185314188,
 0.651866051313124,0.9798574359486235,0.06747183296532677,0.1840905368167406,
 0.006964531285665543,0.5707102604042397,0.658852540858391,
 0.02282363898280093,0.36268084811597734,0.8147298532568454,
 0.1927182509709674,0.8166144663499126,0.6614713711088108,0.9613156737554789,
 0.1956086009339535,0.013055624710607505,0.9183125763044929,
 0.3651766475659941,0.48991635308386416,0.9084712935237524,
 0.38143316152016105,0.8120526991365826,0.08151951372036237,
 0.18058602123136258,0.3516182633313246,0.7994019463528376,0.9404617178575452,
 0.4187328813359611,0.9878413729939295,0.4703451238604326,0.9120978736831474,
 0.6747354835038735,0.7483136580593904,0.680810733063674,0.7780555022031233,
 0.4725282248304409,0.13204207805924706,0.7261811585013325,
 0.12516961723546893,0.9971546741363262,0.41845272769362885,
 0.6456451664532343,0.5512826084749214,0.9886688350033395,0.9554162099433405,
 0.7072912855821037,0.10488534855808407,0.6620390871564441,0.8508966410230744,
 0.7759012932908951,0.33661773000074835,0.6820475393139523,
 0.47774719620079464,0.06764752060727675,0.1383432596434978,
 0.5138537470270897,0.43963635497776843,0.8113909345357956,0.3247081332526818,
 0.020114776010618507,0.5828160568742232,0.20423940545024333,
 0.9635846674839972,0.22722709450244172,0.8346232923952133,0.8369652078064695,
 0.7527650641455346,0.8714297971697034,0.5898219151642997,0.38327444314203796,
 0.5316945911382737,0.9593008493930552,0.3475949610981637,0.7960049154463696,
 0.6378147526520578,0.4948637277479301,0.45704469259082126,
 0.25051415487081286,0.5965482819139808,0.31294231249653637,0.874864512109824,
 0.02926117932246175,0.7510093424388412,0.14197696216279876,
 0.3362068385340181,0.47832708562656445,0.04761216818946168,
 0.019763027191947335,0.6675201089230792,0.6306375622026311,
 0.9183554907905057,0.3024187868456799,0.07210163651069146,
 0.029941320027792817,0.9385757419061331,0.8879331755087025,
 0.12061862821947544,0.29765178900459943,0.15490020068694144,
 0.13902097145629122,0.11751599652275468,0.47871218882844224,
 0.6184283917520158,0.8863972341547768,0.7160251924808626,0.24744933139118486,
 0.7476625467234872,0.6545741435175769,0.5813466938225264,0.45150635957399465,
 0.2524799502391233,0.7871329637521154,0.5091066170774501,0.6200121156956938,
 0.5564356767199476,0.5026746678846872,0.7077528169166692,0.9169014902142887,
 0.36701995504274176,0.3325693538618181,0.9057495959875963,0.317902919055406,
 0.6788532521523413,0.12799798488817393,0.11184775770912658,
 0.7676162073156285,0.0811918404326819,0.9279566739349768,0.5964257525600987,
 0.2306763215337736,0.6012009125801695,0.9777440882412716,0.25898175069850615,
 0.4093854237552371,0.5877684190149264,0.8132807633610056,0.24398647180822952,
 0.9050040382256886,0.04398320195816208,0.05063654817918739,
 0.46245747554931527,0.5417565113565763,0.9357286939920947,0.6682766876757077,
 0.8427547466267041,0.15571998195335401,0.8483240957038496,
 0.46959954734396625,0.9557325448594293,0.8525946286048728,0.3076862311216382,
 0.43671934904368925,0.21958323542839164,0.05383635068666959,
 0.5086279466174473,0.9798849370907297,0.1989140782045694,0.44975866059448155,
 0.4998479181400076,0.980356905130895,0.5873883201671206,0.23158995079073985,
 0.7879016691598263,0.12893807748000885,0.7387740847896548,
 0.23892896966045085,0.07572765234577594,0.7259311786824312,
 0.2604363742939385,0.013359465668046688,0.24667302487189358,
 0.4778137718084259,0.5428936577392434,0.9079857717682804,0.09689964810240315,
 0.026993051007671354,0.8724402006281088,0.5299567088467928,0.395588215980207,
 0.6138875927588285,0.6299141694445755,0.9541483405080924,0.7304510490649179,
 0.22117325509782315,0.6601679244658705,0.46835791705412977,
 0.7058431997275829,0.7751478567528476,0.0690083512287204,0.8262030948473228,
 0.8240449861718773,0.03904914103459434,0.013702316946373583,
 0.8717085622550027,0.41911370453326113,0.817617946796336,0.523921915673413,
 0.6882864444487999,0.0055766011624988066,0.11533425054844282,
 0.5343091850528567,0.11004560487954862,0.7764551540424078,0.668471905003217,
 0.5950786776067437,0.631054488912885,0.2756298844178887,0.6932593636992295,
 0.509072486794548,0.7744747675610796,0.04189507879105936,
 0.028470021298375947,0.6424234596579765,0.08532960959382074,
 0.7978878317509372,0.1547752847411984,0.7524880535863386,0.9543975439221939,
 0.9689588069963716,0.5950902978104041,0.7590703905795471,0.9161739763086668,
 0.9057406604085879,0.1335924458321145,0.7958807137921392,0.47305371658916096,
 0.13334471137196013,0.2219774130127482,0.5748974408952359,0.8689285921685668,
 0.12379669200111942,0.9245871497603032,0.03161918167828792,
 0.8630345857196893,0.10592556779419082,0.795857088394026,0.1543734220441222,
 0.2094994909409037,0.26452028513117054,0.22639159361813987,
 0.9828099634856242,0.7790620218836041,0.7847586637246305,0.6484317098829929,
 0.8671565630683844,0.582892589074925,0.29088419289743417,0.01619441550203471,
 0.7537087348021893,0.44951709634501,0.4170311045165104,0.1689632563619814,
 0.1733412065522577,0.9820803624029208,0.07459720360992261,0.6324573045424292,
 0.2998051348410602,0.5118956534023015,0.4980197548776639,0.7243545360529219,
 0.7827602732456249,0.12766473427752978,0.6021125877484004,0.7847300088389498,
 0.988127249186165,0.13757326233864042,0.7347566583360035,0.7029363640379342,
 0.02085927927551756,0.025275632897169165,0.4021042743461327,
 0.558537805827695,0.02802885182098347,0.5799616608379902,0.9698813671047964,
 0.012108386219980338,0.18307702525545078,0.1391422021972839,
 0.36284433660989435,0.7712284584252257,0.01746803451852552,
 0.26507246376796156,0.8291854474328876,0.4819832109349209,
 0.20433829756472988,0.30137667234036014,0.1465732955173088,
 0.2659029204902831,0.06673429935165653,0.16199834940102065,
 0.6073669623142998,0.3304122481084164,0.9204866553529157,0.8032342522349638,
 0.2692607399335827,0.8462402596366149,0.3010510971033039,0.1093207574042071,
 0.570814388907768,0.564177978240641,0.5969773154718143,0.14852545781300464,
 0.3471641431956134,0.7375583039488718,0.4409847545757668,0.6146104161244247,
 0.5132588262347636,0.11746567900721316,0.4395305292521414,
 0.21244598284530014,0.3095028372995452,0.07716431216665343,
 0.09594018300287221,0.6510160602421479,0.325904359484005,0.33778734293184354,
 0.1580405177501244,0.3375578080455037,0.9230256850308186,0.895102858824803,
 0.9640893321692612,0.6579042182521426,0.6186123072505247,0.3820102634167125,
 0.4457394528562302,0.5975091651460341,0.7699395832103104,0.8481583056905111,
 0.21464017018688564,0.7377051513370041,0.5376301350352024,
 0.37839119812921895,0.42106759320015985,0.9512078728674668,
 0.29152327287947377,0.5593736605412108,0.6477424586733994,0.4611157718771923,
 0.6520590289441238,0.4786113836396477,0.6863710917919252,0.8654216861809317,
 0.44182080529033496,0.3399752938830717,0.13525028580977172,
 0.6846468200197919,0.8067391040062721,0.6251801673231291,0.9713239060590851,
 0.13436431312730013,0.5199236135205194,0.8186027048910017,
 0.25664755240130865,0.2384987432091078,0.46269525657862653,
 0.5501966996924142,0.4894804528437544,0.1280729774704461,0.8372327800179585,
 0.8927155184258803,0.47433036531776596,0.019880716896038875,
 0.6685507921833378,0.4278067039963237,0.5467841898787484,0.22282283430867034,
 0.8821587337018629,0.18826005393469902,0.7086199487726932,0.4384380338765621,
 0.6649047829840765,0.6284792301410084,0.5640248729451826,0.9224206789764514,
 0.3151610331126219,0.1841194379510256,0.9273939285726266,0.47140277713895884,
 0.7097189224402689,0.8591524623850871,0.22600767778560127,0.6850205676732616,
 0.8688158656202938,0.6905156475203724,0.944838545168349,0.6260974020801601,
 0.6952772509811018,0.32454548546588646,0.02684143156900909,0.84672666504969,
 0.2374232053865898,0.9213927003801817,0.6620229703969225,0.6995132738391822,
 0.2800532555703428,0.6936220085080745,0.1408496682071757,0.7387812568580316,
 0.18323360416034662,0.0807563137515287,0.2971021292234347,0.6499621692787148,
 0.09734145277123907,0.3062496742104872,0.4529727946636455,
 0.07057476588323641,0.4496448350962502,0.6180108580454049,0.882327080617747,
 0.7894228433360767,0.0215818514792534,0.9356051164560092,0.03968819745279151,
 0.6045203077585263,0.11533120791518836,0.2141336363400833,0.6649279652377703,
 0.40406177598996995,0.6069136227753837,0.9509288429850964,0.7664097004807328,
 0.7915922141078838,0.8020932393790132,0.4420758245197307,0.5188526128482088,
 0.832907381770728,0.02418422177659263,0.6508453920582026,0.7392365350287143,
 0.17510933621532987,0.6901250057864152,0.6056708816083028,0.2535316216568074,
 0.49126585987624993,0.05091900522945325,0.6790020429445071,0.925945062878385,
 0.7918019532266565,0.8597070645772988,0.3627544720435347,0.7564247369720084,
 0.2233133725843438,0.8911468763432877,0.7156934424300634,0.48300836638211453,
 0.4818762505715056,0.026643016813100884,0.5256668390820901,0.515800460480628,
 0.34345797230756425,0.24945301558178967,0.6360839339608942,
 0.25867469268921583,0.6416257253710237,0.5320227705275693,
 0.13135713244718594,0.3744193303078368,0.5592213526041528,
 0.32913962475349134,0.40196984334845354,0.2871926265567275,
 0.15572519232740056,0.15469101836219834,0.0923128848922743,
 0.46367653082412763,0.010952613411829848,0.15740644657316794,
 0.7084920783349322,0.9613354410197554,0.7525453503823036,
 0.051753803311304525,0.07214492523524707,0.7151757994515253,
 0.836707460268304,0.23094001533779296,0.2909844272185177,0.3207087462925011,
 0.09846730080503852,0.7411666584955499,0.24507644476128632,
 0.6955611145532559,0.2286253981341706,0.8649041561590255,0.07611540249717796,
 0.01877871512407492,0.33204959895557584,0.5562007236019112,
 0.2578476087629388,0.5384988684346664,0.8545892118601133,0.426856415849137,
 0.3486143552201896,0.18833021986352494,0.7993069736505825,0.5317085718706341,
 0.9213074225034532,0.6034484475919786,0.4073120524659416,0.6365122202893505,
 0.3622402034347385,0.12435516898775811,0.7295853254642728,0.538147560620115,
 0.7819401122422835,0.46159453675223006,0.17960466484831417,
 0.5696365701112537,0.31235129239686155,0.9520877417726726,
 0.05735019451632151,0.8558898631009091,0.10254157475451853,
 0.029523042621804385,0.5560893610680272,0.46209065389516724,
 0.6553361768221853,0.15806427006330448,0.009983375270421035,
 0.12757785017587875,0.1585707082173502,0.7926188495963993,
 0.10188293125673609,0.6511654245417234,0.6615066444673459,0.978703424594671,
 0.051179803819927705,0.6600872557229706,0.77648619103712,0.9832753526545777,
 0.031396228856744424,0.2552969958320953,0.9362217000134351,
 0.8713950751537733,0.6754293030157014,0.7065476177426125,0.6165459295880766,
 0.03979526835876768,0.013723379586674467,0.07409336487491225,
 0.8438286934435161,0.5096031142905688,0.44779305764026966,0.7016225980060626,
 0.17109026886520717,0.4819193166099579,0.3100244343380353,0.8372169464539607,
 0.26536018158570907,0.7151432957190054,0.10009758976036853,
 0.5235358006601597,0.9086811113896932,0.7424930507404875,0.41705480551631746,
 0.6918184632764492,0.2824660459216819,0.21076708717070936,0.5678713502833349,
 0.23089769753216638,0.7173707061054212,0.28843702891335077,
 0.22008401703876423,0.6361031556881792,0.4820880568046171,0.8845896861567989,
 0.06996712575935737,0.8429443277633678,0.18812501592231579,
 7.524719431062366e-4,0.8891058406139842,0.18004431145150734,
 0.26845140727904815,0.45734018634598705,0.26059619923909094,
 0.8327651952255758,0.7972044151671466,0.47109298927066956,
 0.13454697210893407,0.8844711379949683,0.7423200093769395,
 0.09072749549180958,0.86851498132691,0.8416082609519514,0.9197506761436249,
 0.1818828175778544,0.4386324755711264,0.09794359075849413,0.7927218646913577,
 0.21036645508048757,0.8759285301894078,0.4661378288963717,
 0.014324859347406016,0.46939883884928857,0.19138666281746575,
 0.35892647216033524,0.5042000333784921,0.5467404563029221,0.9267221586953683,
 0.9372588697780153,0.8386293673995849,0.5065499809977145,0.8357331436274598,
 0.30219984758829954,0.8162250610740271,0.3049058511492919,
 0.40147847875726617,0.7829996250951814,0.5794450555377395,0.685688759063151,
 0.6327414261595644,0.4861228474760514,0.024509699667558005,
 0.9057798604764248,0.11178681787446254,0.04453079251522685,0.518226184385265,
 0.3121318318262579,0.8876539616987688,0.8030451841581906,0.6042912849241197,
 0.9195574898664453,0.7402471155151814,0.7906408416413703,
 0.030904334732948402,0.8576394169064097,0.8816875638939116,
 0.6166102319921243,0.5380330409969216,0.5913949215407919,0.4629041813353547,
 0.0013509876582035174,0.046722455900929494,0.45957664331271875,
 0.11758408936744313,0.4418122004165045,0.8903507185125914,0.9758724929154103,
 0.6687843748405566,0.05519320927041005,0.4619975889212842,0.9894888082764346,
 0.6769552659297484,0.4136847479002306,0.7942203636266627,0.5776547948095039,
 0.9427825775944262,0.17768544570775002,0.5993154149219426,0.4824697958035198,
 0.8690827532987921,0.5364056956252424,0.5418099829324793,0.743737064667253,
 0.5504404074385619,0.3898287838484228,0.5449806087811724,0.0642474412579348,
 0.06513554961281409,0.9725632018000201,0.10100732945242519,0.859739491678358,
 0.7103841310164567,0.39745761565310334,0.8214164810539644,0.9128710773197792,
 0.1990607496876079,0.00843214784234425,0.16640608487370157,
 0.9018139209726594,0.874667784770808,0.05054314913242153,
 0.012235966825052125,0.4030299072531037,0.24378697515059589,
 0.013256705362368226,0.9923008294967142,0.4864270830204328,
 0.6222421763090921,0.21876199273148078,0.6047795785408353,
 0.21401909104422223,0.943919949843039,0.7599652620300705,0.8470400512108859,
 0.8167463967241666,0.12979066178069099,0.5897511662479747,0.6530490886385742,
 0.18851849551938793,0.8944478185033077,0.35603265091852476,
 0.35246669625441096,0.3669553328481654,0.1449596700098721,
 0.026976365229602006,0.8707402964245278,0.15401971664198721,
 0.9962593983457139,0.7560446729132121,0.670332079855702,0.8122405507261184,
 0.6742151433482482,0.009604183490975582,0.3185239011412292,
 0.7255286698080403,0.6298044086126793,0.7732348564025024,0.48186490336115817,
 0.8117822231571517,0.7182844445900323,0.1832256381731634,0.02555414777225984,
 0.4545337706135729,0.4518965761294762,0.6136731484534591,0.3833613100739467,
 0.22589500789692019,0.23061590715487723,0.7167315264632952,
 0.025699485422127077,0.9270194803762889,0.45830362672122227,
 0.054342771648810295,0.8786212186738721,0.951728881378241,
 0.12349862861275684,0.72314499698426,0.18084991112860205,0.6188388548813498,
 0.06207435034649933,0.7514238020328844,0.008119224135794023,
 0.7120470328185825,0.3375287879420068,0.5916402321445854,0.9743614852911825,
 0.3878884903015336,0.9040584449200406,0.5182032865790616,0.27263746061275274,
 0.6896504653267732,0.5907105135756252,0.6296130687784874,0.18607267715519316,
 0.5962200255528041,0.3855007464845617,0.7165066094850805,0.7292261784384635,
 0.7824416846945769,0.3405938990517562,0.2199218753834491,0.8084328703272866,
 0.3010338086305563,0.8964415686010729,0.8640688498524546,0.27750552821654384,
 0.08005644442269633,0.022276426016306194,0.16577196007746575,
 0.3333122079887023,0.49237514237019,0.940134810890578,0.10412554581537459,
 0.11984180330117455,0.8183755623704712,0.44246448665842664,
 0.40317397136024746,0.429795778841746,0.006362568511189659,
 0.6895133264688147,0.5989392947094204,0.8014047272811423,0.830381417281624,
 0.6642319580228944,0.40223300944019647,0.3651396157626098,
 0.26855300219864275,0.4286928545686002,0.7160133807567743,
 0.32265364144128805,0.5396025430547706,0.6435637406461546,0.1968116181669659,
 0.33991327266423543,0.4965712489380917,0.46309017906263716,
 0.15470106565472663,0.6187077433890801,0.38558593352170567,
 0.38858225582086514,0.8428051384885682,0.8994449560293016,0.514296645614901,
 0.6394245690269,0.6926604558604966,0.20832020162767972,0.7653989894106694,
 0.16286165903054484,0.8895557230412006,0.8310610196567979,0.7594011466058321,
 0.1780610047484923,0.6951515824857598,0.7245002102727387,0.698699345194954,
 0.35593468317780586,0.5085726023151801,0.3749916746306301,
 0.06401368838370314,0.6384369592203715,0.9739340687753674,0.9504037054365546,
 0.8747894598428541,0.519285455714807,0.37209198187924686,0.7427236894471798,
 0.008233041722745194,0.8650552184987231,0.8562732701299496,
 0.9767332969794127,0.3279992340116843,0.9690156974965984,0.3863604884871805,
 0.9502740024361495,0.7525136794411771,0.6409170518894763,0.07596871893162704,
 0.3701536355425664,0.9261625672817061,0.8779233713462082,0.7811031772830221,
 0.10228456339413572,0.3725458968543828,0.46669706918802023,0.81578781887889,
 0.49905252252421406,0.3504655312797965,0.07427278663989223,
 0.32710014238891916,0.7410430264041239,0.4268164220944306,0.7068822056381561,
 0.5515474328875002,0.6641568049044362,0.6454407297175909,0.09226088198113191,
 0.7284800536146565,0.314348695495416,0.0046362100178514165,
 0.5072489273143374,0.7492977859100344,0.693732025924876,0.021771703772261675,
 0.6653700835855593,0.3056280193272446,0.9042566377309029,0.9318053615897807,
 0.06073121584993879,0.39742925165310894,0.28118933883683095,
 0.8031468444047694,0.6423456959062666,0.95099356211955,0.9381221246688305,
 0.49232307879231085,0.17447720244313059,0.965302195475784,
 0.15826637142368294,0.5213065575796393,0.655583704626787,0.14962737562478,
 0.026487410088617103,0.9395006433096978,0.18670451874941785,
 0.6600285700675729,0.6671099162734762,0.33609032065538236,0.8449423436986033,
 0.0021046937237026997,0.4181522602157156,0.5736004529448836,
 0.7394660481320803,0.19626090763862625,0.88631609092772,0.2688817459368503,
 0.9969027190708943,0.6527441766178409,0.947129102261283,0.7199837707547625,
 0.1734715249885168,0.65452003364621,0.19845395359900886,0.34421746194593017,
 0.9589095904637837,0.24473910598057214,0.19009719680926396,
 0.07208898830442778,0.2566767614223623,0.5168058637188243,0.3513087765785844,
 0.11921962697665367,0.19850911762202128,0.7738089846161548,
 0.07869106171740214,0.325584027005976,0.1560157115073304,0.7359671487479025,
 0.3836093684710322,0.5997144625985835,0.7870959488726041,0.9363707144906853,
 0.3003435743286521,0.6908057793919666,0.6547874425026011,0.3662035963314012,
 0.5743994967714161,0.37758392686966324,0.18075098409947266,
 0.9404252071585653,0.9488130171337235,0.5860686962428475,0.1847871451765012,
 0.16261253859366542,0.15258521418753046,0.6187223631302914,
 0.9322336390061231,0.10217124136550049,0.2744652452258014,
 0.18419997857313974,0.0991115884434538,0.4678069559806708,0.5745255560599154,
 0.713218766591639,0.1168863537072492,0.6662971289104112,0.9683156292194103,
 0.475509616237451,0.23363963021820877,0.8404824836888113,0.5341783704972011,
 0.5993248462550629,0.14925120515902124,0.5011998008784373,0.7131874725296712,
 0.23137488794989847,0.1717771694104172,0.6705449459981053,
 0.057756981433327303,0.339137789763695,0.9348554939203756,0.6700343785161784,
 0.5873557097308412,0.7473477856058872,0.8293982848839248,0.5901480897280995,
 0.7040562556145601,0.5665520521504092,0.7044868122165782,0.2957195536348335,
 0.9750738372174882,0.23427112134192907,0.9240834145442307,0.7377311145463713,
 0.559280511471196,0.5403035662886155,0.9621256472321892,0.03993454181014455,
 0.6967020086627985,0.8509350613935144,0.9962632438677482,0.25933756336072,
 0.4816262305213842,0.2574115343091026,0.08814279120958302,0.6622947570529647,
 0.463775016874736,0.6294993321310166,0.7596663243910353,0.6420528967053896,
 0.40362952799022445,0.44012376998367153,0.49368793391172994,
 0.42360510968067777,0.8858199383981276,0.2096817162391561,0.2915295348882605,
 0.2729306469413805,0.29818213262180593,0.018322591618449735,
 0.9185598612741289,0.32665685670731914,0.050364961028744104,
 0.24660432810946276,0.45317212575902144,0.6886633350273306,
 0.1452161177924709,0.6293431262259316,0.6104481692454562,0.12798503763731206,
 0.17621773137960017,0.6584498547165336,0.8438145277980751,0.5803246898167317,
 0.8130974236550471,0.5379170455254328,0.39563885526065123,
 0.22791597098144012,0.7950378606183464,0.7747469728455147,
 0.29574921388842745,0.0584454271476571,0.7668240976662033,0.5489342278809358,
 0.221867056281289,0.10104850505266627,0.020978770249996836,
 0.8069952618919455,0.5379884510998288,0.3597494216963669,0.7720570158591555,
 0.03640690316329398,0.5918789976628804,0.6892664467121539,0.2767134971606593,
 0.9886576041613976,0.14093118353210587,0.7773895809176745,0.7824693639203029,
 0.34214365289478066,0.21924493576709725,0.8580873931838976,
 0.1455949926528537,0.6355233423615627,0.452089154148163,0.3043599320354037,
 0.022006233659410568,0.870487358944863,0.8310986271642337,0.7691067088205961,
 0.4716427235044671,0.5191271523279624,0.915780014575196,0.9238481217641035,
 0.07541764339749202,0.0039161653356312165,0.9187201300285532,
 0.17142298064430272,0.9680791100991003,0.6845784643518575,0.5089437210645439,
 0.8113865565153607,0.28978253997607717,0.41399561886308567,
 0.5696804068469674,0.48610834176650286,0.6365647715020223,0.9059890492258811,
 0.7765724951536982,0.9602420981318264,0.9240887619156076,0.6249559529209169,
 0.8061286573767535,0.9710229501320394,0.436095391170748,0.14851027723120946,
 0.7505760734346316,0.3622636712084971,0.14986158268239136,0.8356158199421058,
 0.8130445224957117,0.8252546491964834,0.3023350538551207,0.7284109368643649,
 0.19928908571520387,0.1512481085683166,0.8100618870838628,0.9246401473556735,
 0.964476048618058,0.09400754097692343,0.7739658853791941,0.7468765098007316,
 0.5849127326295716,0.03080870401889957,0.32890369187817736,
 0.5726055187835953,0.9714817631461411,0.5158366659306404,0.3484350003610651,
 0.6903537618845962,0.4070887853758731,0.19813853269407378,0.7980866573156017,
 0.005285648666932263,0.8314626240746598,0.8818957238933227,
 0.8378256254127087,0.5298901393796396,0.27537924716070106,0.637010006025623,
 0.7345512085225003,0.6146367155906898,0.446050754489269,0.742825284314003,
 0.7869331656286588,0.483002286323994,0.7255987565902254,0.5788702187904815,
 0.6396364061261905,0.7753127533555986,0.786245239365731,0.7402274093759266,
 0.14445262325558228,0.6257715048188904,0.8816261736648754,0.9674120155851033,
 0.7819919118093631,0.8691090477660861,0.2912961879211631,0.45489108285686,
 0.7365823098358085,0.22813215484360383,0.6525748455394224,0.3340212142650335,
 0.1466592299926558,0.7242138127967735,0.5277765866120753,0.4338963643314747,
 0.7728826866969205,0.42369762226141783,0.16704537137873543,
 0.8236274325991002,0.07346215964974134,0.9302977794829649,
 0.45988122484575034,0.9214136670551158,0.39000267013352596,
 0.2671688407804811,0.15905246468835466,0.23559542388943788,
 0.9455749935516193,0.5786201649397843,0.7991730347944086,0.080484186461377,
 0.6393462981100156,0.04574337379415416,0.22649632443722767,
 0.5894645150926421,0.38474314994421643,0.4233620719036316,
 0.18046036465244275,0.5373754225369108,0.4403222235879489,0.5217035520421595,
 0.6648328268367043,0.4017031996401328,0.5358709213938783,0.07318826903820874,
 0.49394180147446454,0.4329267370109168,0.4101390303145084,
 0.007670620007668161,0.24222929332379206,0.7317667127469824,
 0.09243580730130374,0.06254803753122506,0.8401809392877414,0.659575565012816,
 0.1937268161007164,0.7440640402970319,0.15186897608998007,0.6366723310041922,
 0.2166270373201209,0.6826987019311814,0.9678024829644828,0.5146944164313698,
 0.6555748890830748,0.5221941926157078,0.1352713762112413,0.8420357120559576,
 0.6212568777280362,0.7769954067320242,0.11194908331526854,0.6794701559671923,
 0.7064684085166885,0.8071790612267243,0.8892376562218466,0.3210531298887036,
 0.8952895601279159,0.4812368565033104,0.8929990571788288,0.05592133649341757,
 0.5453804499280319,0.8930176389331297,0.6032254662305977,0.9834129633691661,
 0.5840348104166749,0.20888150164331198,0.356494205037993,0.9304475852594216,
 0.6604534198499,0.6354157016346828,0.9450062022828284,0.9550349733776669,
 0.973866751233956,0.8408756203960421,0.4529316502387387,0.6425683908587652,
 0.6716319439874393,0.8211793963655587,0.7315712868844528,0.8564648096045882,
 0.6676790482448266,0.6594862008328893,0.4403079404758943,0.8029069677592762,
 0.840278756697008,0.9061670125355359,0.5244187070951365,0.608689978697467,
 0.48510780518523877,0.4046601142385644,0.004285361306214597,
 0.41825703584318397,0.2106394663221559,0.47675090320544666,
 0.5353378422558871,0.2423784297158127,0.7625760713614559,0.2463086116751032,
 0.9246537460073481,0.6822520902676568,0.5245829000300279,0.17768107761044472,
 0.9126791837576131,0.6373754025460613,0.5778245925794631,0.6706508920736239,
 0.9740966159478428,0.1778427525031533,0.3810451621837798,0.22491449978583045,
 0.44998490642191014,0.6618025113895014,0.3980598271002125,0.4678966595531001,
 0.4638718621284301,0.5020892362090643,0.45967394581162024,
 0.05646643270760643,0.4101165319767045,0.8544326079309952,0.8666437219013943,
 0.32156038323514324,0.4532324207462153,0.5771256101346791,0.7006886720220906,
 0.3183065214953753,0.5147015455419143,0.7151730018791054,0.15982574617367784,
 0.40255454284382497,0.3777225727795005,0.31235167601726876,
 0.029173054700475376,0.3373136330051786,0.04154587650266195,
 0.4212704616016437,0.405808882412959,0.5137026875798906,0.6795537947848432,
 0.7757827917050202,0.7725566366718872,0.2152496226317404,0.9171753717762587,
 0.6908933286966497,0.6582392997132565,0.6845364842721379,0.15367922970089265,
 0.13985504098274248,0.8201956511110958,0.1482967428787949,
 0.21496201308343854,0.5090988969488222,0.4291005851864639,
 0.15574689008342624,0.13899456918637432,0.2919005805571667,
 0.39650536657148294,0.9139829282507715,0.5828242620318645,0.7309822337510093,
 0.8007125775978451,0.546109399932616,0.1605883294765983,0.32022744740712084,
 0.037796322803362536,0.1350608835390935,0.20594795356802598,
 0.7804473823019916,0.8595881906186696,0.8972127435428303,0.4995954208755641,
 0.10762440238185111,0.5614493302772546,0.24687874051009517,
 0.7975133149628076,0.7493257515952458,0.8331721737232366,0.23969492813979343,
 0.18279613636296121,0.06959450963613067,0.236409292312304,0.6387689828260035,
 0.10473944157066639,0.27281157092240516,0.11043519984461614,
 0.15991285885061512,0.6265381053184307,0.49042078168749237,
 0.6726856465760876,0.22509664336782786,0.87622577535451,0.7385672071117295,
 0.906849248731339,0.9923281141816193,0.02159632925671151,0.7937111232584941,
 0.4729590104157877,0.886997241995381,0.20682005477298848,0.4246040885379233,
 0.0621285141848692,0.2237825371550658,0.7358369915711839,0.7904438567291183,
 0.6669911424633721,0.1465968896393275,0.28352499549297083,0.6389407841839029,
 0.9044963229590017,0.056543097617236415,0.4297226437606564,
 0.5580325452551539,0.2753342730841575,0.23403341023175017,0.8525120616290983,
 0.9922280455689887,0.19609406616731495,0.5466369964301956,0.5744179140011179,
 0.12018502922159746,0.6613357824944154,0.1891782927608786,
 0.11306406385591261,0.9104769160772769,0.3246339985566298,
 0.10637970925951445,0.4106889089717386,0.6371356657078272,0.6143031102909153,
 0.0680479449199084,0.11980451972705664,0.535390637718075,0.46763690210895825,
 0.03244173535836725,0.026797513186352928,0.03138629773299384,
 0.8347417046948274,0.47153997641342016,0.288211469663485,0.9761709810630257,
 0.5035744461041205,0.8554087890500094,0.6604818500782053,0.4070968227320345,
 0.9824585838397637,0.2969535651585228,0.5092264433090568,0.18721122056134898,
 0.5362680413907285,0.6403360741344992,0.8847144765904307,0.13513087292407344,
 0.28630615351212096,0.8830293757945639,0.2594900383744613,0.878701521035317,
 0.33770107709604646,0.9662537271757614,0.6210448659158577,0.9805377595659389,
 0.1831938246310345,0.408269289584275,0.6560934531390379,0.7123009736744021,
 0.6020323053588655,0.9465755022428476,0.7251385564605161,0.21024390247224656,
 0.37602036624860913,0.3655891558509453,0.9711300579060723,0.5355417810104156,
 0.25100311810707054,0.28637871694223693,0.991785622599604,0.7600650536480434,
 0.857398981187431,0.4779394772733896,0.4267940513064481,0.2796278872306699,
 0.13964498320688234,0.43044743206037994,0.713378057317256,0.5903390441027022,
 0.9520528794568255,0.5696001161521825,0.6970532907560942,0.02324592585821572,
 0.5004126310337799,0.6777600721506184,0.3167082901146516,0.20771104413872232,
 0.24509629020584078,0.3182476240283939,0.7991973215913258,
 0.03673386276560908,0.6977699583795894,0.14194439932941694,0.078115882703236,
 0.33015977404482133,0.30137639388910686,0.8962584726259135,
 0.11460941464123187,0.7713791693756524,0.05386146061559549,
 0.3834060826602975,0.9369534570261798,0.9841501562001986,0.6623496477535716,
 0.487795793147762,0.2521061293261502,0.8577342503845157,0.7420052917873396,
 0.2724110626823405,0.8517529021442108,0.7892643319172025,0.6558040732836753,
 0.7985450554236232,0.29570287930885675,0.7642491811254364,
 0.42470895137597564,0.17506226580710138,0.7956254546615088,
 0.7894272432126848,0.5966956609173557,0.4294514375226537,0.5429778990376471,
 0.0458865882641033,0.7117428692559565,0.4627377128803678,0.7230699066276306,
 0.10409466462888395,0.8692327172635853,0.7515606665760863,
 0.48911012838309464,0.9532239270367504,0.33186366403363965,
 0.8176842092757208,0.29927124203704913,0.46591058771075344,
 0.4537570261550101,0.7717885906072379,0.4994389955266898,0.4996511981987516,
 0.19005081837858684,0.5431381977989073,0.9278436691232493,0.2834141542343964,
 0.7736734132019671,0.07433414133512106,0.0024081356400256215,
 0.9768850990463451,0.5733242322280416,0.05326566697328139,0.994511576206586,
 0.004640284971578534,0.4226035590598948,0.24794590475093192,
 0.0064688120996875265,0.09930585518313217,0.5151448389766622,
 0.9064163009574037,0.004180276885758172,0.5291654093943382,
 0.24887234862830954,0.5641460848339563,0.4716090047585699,0.1813844808742866,
 0.3102671252372914,0.7697309580337083,0.7744967421642311,
 0.044059129696664034,0.018765897241222573,0.025953505772391305,
 0.6635713577177099,0.38676468179659496,0.23027837789543193,
 0.4904281063670439,0.4463244341474333,0.5603176383519757,0.330534189413879,
 0.5177594981505758,0.07703853420638285,0.9114796099879412,
 0.22288222854741924,0.6799448006095422,0.438585948932779,0.9141903825445233,
 0.17605538500018203,0.20280631690850592,0.8739775204086239,
 0.2152053327586979,0.7146272602610694,0.827217006154776,0.8858771663866051,
 0.6313298612519306,0.5296010131573076,0.8726719631138542,0.15865938604487306,
 0.27117752525511096,0.6164410136370271,0.9887273699567767,0.2708553692092408,
 0.3496969094585767,0.6591228177487206,0.4008414859775741,0.944590090741479,
 0.5675469236784421,0.5944973787314245,0.279157357438541,0.40162612678584453,
 0.9310660799608468,0.9060231737260611,0.9305694510151049,0.618857380701831,
 0.974707878997034,0.7530552990576008,0.26064086936496855,0.3118995318954386,
 0.35355222205032977,0.38624132878591955,0.2316161537462479,
 0.11083906826404799,0.1566619601140099,0.24559779277204186,
 0.1510362283598884,0.8121437854389083,0.6563180537401736,0.6746612743337166,
 0.5792018610540031,0.8225782707630475,0.08126338969708247,
 0.20120606087073234,0.9884300848198837,0.8871294157705658,
 0.07566065544820144,0.36120616084636314,0.1443919919892518,
 0.9136800508267902,0.3980405958872434,0.7805328007961087,0.9376735944055619,
 0.36104655982534894,0.38547818309006354,0.30877060994540184,
 0.4696414507844263,0.7995157435657074,0.4102657882295352,0.04474469206081999,
 0.0731047770098171,0.1681960239174004,0.4060196641435445,0.28332141668317046,
 0.7468912250884592,0.12854078294980964,0.3735005655116287,0.5741335690760845,
 0.47457204730031954,0.42207113597715573,0.41928315720097065,
 0.2110254260303921,0.618559255596857,0.8075550325790927,0.9479601211153441,
 0.2846084201708132,0.0030986273209673287,0.11966554392907525,
 0.1837703213940266,0.7951661165478869,0.35768161454930036,
 0.12590182595699018,0.39280672261787974,0.1771042911955495,0.676966185359829,
 0.6291318293274542,0.924044194280872,0.03545062071558247,0.08901239120333837,
 0.21281804785787728,0.8432802575183179,0.3050686650838521,0.8242047630013426,
 0.8682520077073052,0.3377565129947613,0.25643988603925993,0.7149523772743698,
 0.9423796911397233,0.035792887861300304,0.6806975977138409,0.928461212324816,
 0.8052720433738889,0.4966921446556146,0.16137672444965023,
 0.007147569550993715,0.6106458541492276,0.8363764227648298,
 0.31443546848135817,0.043131530795931794,0.6598602144015316,
 0.9671149611237644,0.3031953536667138,0.9309846851180072,0.3125840376556023,
 0.7701442681081491,0.7937607124523456,0.2055558135740112,0.4637043986011684,
 0.023151607964087106,0.06158588029290468,0.6591613795303998,
 0.8838930494203665,0.05004805771247811,0.8360208001162505,0.9211534867812485,
 0.5980146224700289,0.7320400872494263,0.8126011062854613,0.25601916876182873,
 0.8428624766806915,0.8648340328105917,0.8317112725329171,0.7538977476509059,
 0.40331938622077135,0.4529597624088626,0.22665357849794632,
 0.5192131380421141,0.22502228957889248,0.9092402605511941,
 0.15492481852827722,0.9447615733690821,0.8169602622882439,
 0.10211734872363953,0.5223976134776895,0.6088774334961382,
 0.24962040645302985,0.6098456152904176,0.40840167952508444,
 0.5018577038891074,0.9434498181183315,0.40349507392663186,0.1441496039693283,
 0.8878660974600421,0.06203740649449341,0.05202969458033935,
 0.7098981318555279,0.13804876038495362,0.13349222005096073,
 0.7063002285622002,0.025571277683828497,0.068468549746726,
 0.07522670937863218,0.27663645470927045,0.2278769102443643,0.067336617293283,
 0.31961837403899884,0.7160252925710519,0.5853447062657435,0.8800524392843637,
 0.14864587846141397,0.6810659467354802,0.5026568355798723,0.5507822422889934,
 0.5492526785515952,0.7344742073319155,0.84966845331535,0.16727379528836095,
 0.296077277225845,0.37360698499463774,0.1775457951423799,0.883657251570623,
 0.0875117205461049,0.187044155750397,0.7398713653315715,0.5844467976375144,
 0.6628895796462954,0.4551621891270471,0.18419685043462286,
 0.13850557538449138,0.7391726201200449,0.9451952172783039,0.5815661450466485,
 0.01977485639966492,0.8118410218645966,0.4856933825314653,0.7453611333294333,
 0.6831565982539916,0.04272753173883359,0.39696809043726,0.7718840193882782,
 0.37643410152544454,0.020874730520145213,0.6536982674965215,
 0.8276310241703135,0.24237994179162115,0.09931793377306719,
 0.4832896442155742,0.7279010699499625,0.9248747078601858,0.12105843645421588,
 0.4963558753090348,0.9601861467506868,0.9134527691680236,0.17860780148124145,
 0.19360648709467276,0.13894630984260825,0.6941228537461739,
 0.2428984901605814,0.6902772697426531,0.5519107004260371,
 0.005506418064583962,0.8085772137290381,0.3097319874096073,
 0.5825926119496807,0.6298496308768713,0.19611996497129824,0.7052387271365356,
 0.23303740441934784,0.32406994020498836,0.8833831953401101,
 0.2869381410508878,0.40764368864187894,0.2943599765319501,0.9019863426182914,
 0.8671230268966852,0.3405007873695196,0.7364199784806346,0.5446993929992398,
 0.570404134407984,0.5158021257937425,0.3724740669265625,0.16453612391327166,
 0.14199970013935637,0.7329381303160868,0.112773176332717,0.12763529050243272,
 0.4499849785483725,0.9673342479425928,0.8839676849193472,0.24606606935468744,
 0.6155688234077058,0.9886103776284445,0.17546959015692765,0.9487892866883902,
 0.1095539650794296,0.7201627052968491,0.4186944474154034,0.44844057468846577,
 0.4273098137274506,0.14877155078772142,0.8278077485132156,0.8901365652872276,
 0.14224607102787878,0.555654056493415,0.41059765629300893,
 0.24843518870924375,0.5094907194376729,0.08026771379536712,
 0.36188265089954774,0.36659110042192977,0.6379444068412514,
 0.9869559515605943,0.16056016540413176,0.06248394149907077,
 0.24051977000265912,0.05233954956468079,0.26095955882593236,
 0.2352337491146771,0.3068296302468535,0.36862845021250557,
 0.46900393907824134,0.825599082702934,0.13644733338943815,0.7551061418959137,
 0.3025828681955548,0.5891186867354073,0.1940968379389042,0.17450605431087962,
 0.5971807815814663,0.18954341259561736,0.8625470243030829,
 0.08371371837833597,0.2767500030583254,0.4896608176750359,
 0.10165499676340695,1.749322037939427e-4,0.8345105261057517,
 0.37954175298015924,0.303815628922422,0.4336142931323401,0.326769814233809,
 0.6066584668709607,0.31607764104125424,0.8802283527877043,0.5626725124524594,
 0.580425037422924,0.36727372217290344,0.7123021644301408,0.7452037172225534,
 0.7655961434483622,0.7756209668949776,0.9566562487911638,0.4860628226214454,
 0.23448248911775493,0.4057598054787307,0.5146812960172334,0.9567963591723487,
 0.7188848289854853,0.41661654600198816,0.4223712450239647,
 0.0016297216036197382,0.42106224329242825,0.8305675814929906,
 0.10079561988362307,0.1699639677835626,0.4307140609949782,
 0.13332294401889222,0.45472425123824267,0.8680236205267763,
 0.3118362854289529,0.16283638821210578,0.6104662416920923,0.8161665441849564,
 0.888320834491262,0.922479309725543,0.5123786803520828,0.1517258519151169,
 0.40909002878847056,0.6993047623571629,0.7792967702014164,0.6174876296820143,
 0.21002660281039143,0.07254533031416,0.30628214108060714,0.8868604386018519,
 0.927494182943269,0.40439655426786947,0.905241847164636,0.7255380109052089,
 0.5856827144539062,0.6949826678200413,0.33388700922707537,
 0.30902545840164397,0.7198314928395416,0.5738063890608662,0.8824305017736906,
 0.39826319987144054,0.22522325533479715,0.7977067148790187,
 0.14357452729916997,0.817661293889986,0.6071294045904615,0.5839571585634074,
 0.3912870476594169,0.47765743599171595,0.9253057427659462,
 0.13277776180500145,0.8717192937061543,0.6417338790778994,
 0.12500096179676445,0.8023818662173854,0.3847118296690484,
 0.16619285281685547,0.19667537062106166,0.3136524118389553,
 0.2977061369420334,0.2654958863779683,0.9314290948252146,0.7207501778102291,
 0.4584711674937829,0.08892405705415574,0.27480493462029254,
 0.8544188233115118,0.6170171170823858,0.9702933566145684,0.3542415190916607,
 0.5148087959006059,0.41260806365690295,0.8059010264454126,0.944957086480508,
 0.7681397743142566,0.7970002448519952,0.5637145000914149,0.7809262049545818,
 0.896761663415934,0.6747937182309061,0.3930913219767429,0.5102446797061773,
 0.05630209340653136,0.7557159019739881,0.3746955478699101,0.9514493156153209,
 0.8605933454076378,0.8272339457893905,0.11910537842845603,
 0.45224444104262884,0.11882139939201708,0.19289573011237682,
 0.8677046201120859,0.4602624015892096,0.7788196036269218,0.1600421584730234,
 0.25877821165418435,0.29979249488637416,0.623473374284836,0.6931779568708855,
 0.8970898759073068,0.6343494004994563,0.6893281793353345,0.5050767648279836,
 0.014748846880171973,0.5267224786738809,0.8398852333120059,
 0.3405995624706988,0.08640012858537116,0.5990706533229226,
 0.12332816481585274,0.048458414467261735,0.44598206950020514,
 0.439633550415356,0.5545127786602472,0.8052471673743233,0.3612107246590607,
 0.46662009095262624,0.20559224016454303,0.7174535415124185,
 0.2854356585353801,0.224584485310104,0.8494846380906987,0.36195113937328094,
 0.4727183185919541,0.32272108779389663,0.6446793049434707,
 0.009991840796639617,0.8345576326407146,0.8615627985953576,
 0.2210803184031913,0.8717517391693339,0.7327953378376372,0.7965831189127273,
 0.8411940234214064,0.752293948761089,0.6805103276676647,0.4489383704013783,
 0.09083732019032165,0.32590061370954637,0.15030632864856275,
 0.2547626029262532,0.15738749117143236,0.025356656447944492,
 0.48126494446413903,0.31764709531027124,0.8599619927450297,
 0.7723367830577349,0.6798660853935898,0.3606669572222938,0.42446358506035464,
 0.46963184028288263,0.765376190805112,0.6484316858668092,0.6320547254026808,
 0.9465802499181539,0.4010174053808493,0.36145819529090195,0.5845558453500238,
 0.3405055702861307,0.8383117363904591,0.4828970616322129,0.3852237219740139,
 0.05683570531163107,0.7428804922448259,0.238826861781259,0.7558970238197071,
 0.632187829494265,0.23275908299804227,0.9178553341913661,0.9932243892392545,
 0.10487795284398105,0.42850388315972876,0.4793332051108633,
 0.9944530296494097,0.12077027384924377,0.12558426297341063,
 0.006446975210031347,0.012553454123862773,0.8125430414115357,
 0.6742574783005972,0.4023492738690123,0.575675221607998,0.29162587182190647,
 0.8415662923678362,0.38658502122754057,0.8321519199960609,
 0.05953997913767772,0.4684225524977538,0.14913946231580777,
 0.18319664530686497,0.3146967893617787,0.17438691812428242,
 0.7376405350504341,0.7326537420180524,0.44176738629565915,0.7541566339151852,
 0.38993984399903714,0.7776787413082548,0.306741757153946,0.9719964701279372,
 0.8809590271374239,0.06298083389090003,0.3661216541883704,0.4569455491509743,
 0.26380560009175746,0.8286891589523725,0.13123262959058923,
 0.8313374383090767,0.7380946897301299,0.4220655943744718,0.4927333886832913,
 0.031488408467888473,0.4431264603791689,0.7111531199809176,
 0.12028973014954292,0.818799614388478,0.7907706452151853,0.232067919777719,
 0.6907892515599526,0.9378505198877057,0.6695449254873825,0.19444473161855647,
 0.44162729022491054,0.27331597616798553,0.7286056604106169,
 0.7758995649082431,0.3182531499080202,0.1590170442926021,0.8620297726218525,
 0.6649586621350929,0.54483045371158,0.5751006363695879,0.4224902901230648,
 0.5846181032591816,0.6385831814599937,0.23842642303005113,0.8678120900522303,
 0.7980640301887029,0.19630008223685247,0.30580474138847946,
 0.6103298386726558,0.21580663801206434,0.6821242729904087,
 0.42822145334732875,0.9568956951767372,0.8486099488106436,0.5228627990075975,
 0.2550037026641927,0.41756597412968677,0.3564686123926516,0.4166551969716247,
 0.21277447937540994,0.5104678481697665,0.4749278726775189,0.3798173307958589,
 0.7147655355307346,0.09326877198528827,0.06966801824255209,
 0.5646952649235293,0.269032144775477,0.45328523859225067,0.22289163447518157,
 0.7909543524265931,0.11195782665095522,0.6085479053470013,
 0.08325850277882196,0.8807141641833249,0.9139530206348662,0.5237513279374538,
 0.811361210709075,0.9245973716393147,0.92136508506262,0.12217905955837782,
 0.2445131796733051,0.13061978246197048,0.03657068963737331,
 0.5544694341508135,0.7967008566934104,0.43510471940854634,
 0.20385818169462988,0.8871244356680619,0.4767759361892455,
 0.32412047836378055,0.3450319295889098,0.7795839412478798,0.4274779744356825,
 0.635630208744141,0.7074383670509706,0.6750485836981608,0.34080998754665137,
 0.8128720643102374,0.12318900570198232,0.0027853803506886488,
 0.20321515873426288,0.5016330837567737,0.5654898943075758,0.9028638700599302,
 0.5545065016957508,0.3355878659176714,0.2822885263791486,0.5964788050772394,
 0.30128208139204693,0.6182396571776746,0.620930933995866,0.3193679224002418,
 0.5944597731050871,0.6324396553484896,0.5160892039735088,0.947900035007873,
 0.5385681325522984,0.27204606559514355,0.5428267238575748,
 0.09367905070687177,0.6053014349447973,0.008630468523481705,
 0.6348988678991144,0.546523903462834,0.20671849629552153,0.3762707858148391,
 0.8265254044736324,0.25231614948486203,0.4164614469687995,0.3246588509151356,
 0.9738689047682871,0.42191363711633034,0.28589142590459093,
 0.12204546003246575,0.9320079424675686,0.14211602653102495,
 0.12753448199048856,0.5482870454072652,0.12563157744956854,
 0.4882827656902309,0.33165166951842096,0.08720273680397272,
 0.4785028188473638,0.4438847097715748,0.14514473814759776,0.6388003736204075,
 0.48881427306322645,0.555776965841272,0.1956271042845663,0.852371602911702,
 0.7491601325624802,0.8802679134167801,0.6471931886302135,0.7284423851524489,
 0.037707912671767096,0.015090921846657834,0.24339508268090926,
 0.8911494918838496,0.15835269191312284,0.4982206766339665,0.6146240295497054,
 0.21562089480668223,0.11632313657698878,0.5747444019516263,
 0.24850855064015576,0.14836112480958707,0.8481027383258701,
 0.7744555199545664,0.1686040745807339,0.22829765665462842,0.5359141654695614,
 0.8098943001671434,0.27169409949391743,0.955149697468263,0.03933057490589209,
 0.9142337069101748,0.19020843991374725,0.2760227951776555,0.7041790482460851,
 0.900120924150544,0.7965406949529124,0.25180186988091713,0.887226899137271,
 0.18145251037184296,0.14377768056985352,0.3312950239893878,
 0.6753470508118408,0.43906230196663953,0.9261506925873999,
 0.25931139652286683,0.21316729559834968,0.9101424722603816,
 0.03913193768964318,0.3842112991929164,0.3404738266584292,0.6092541720116259,
 0.9221294994538859,0.2249008548391369,0.71889506390985,0.4261218195829899,
 0.0928613754458516,0.3895552678931513,0.17603829955506733,0.1366549096626718,
 0.40775856039440816,0.25757366682814453,0.8721163158492036,
 0.06221215553042403,0.9525101994577408,0.6017485277350217,0.5815269232644686,
 0.24268796384571567,0.09752655374663366,0.9408459438578031,
 0.7082299998886472,0.1026677516262694,0.637520867488071,0.5486163378969153,
 0.13751088190995098,0.04066213382401529,0.1020352534118989,
 0.7738722216462907,0.9365244067144866,0.35991603941669226,0.6371494604951543,
 0.6626904880881372,0.9615674214274845,0.7960244612468075,0.9068519840387472,
 0.008188060211726023,0.18238786302988574,0.11760389599614363,
 0.8711432373041692,0.5671055661215023,0.9934714998956201,0.8113831425587382,
 0.5689834319251688,0.4246616547761416,0.25723436254064347,
 0.14196001286853743,0.5337731054821103,0.9147608652701354,
 0.07664555108649773,0.5627154283279587,0.9843743915058847,0.5610789628486949,
 0.303584116252267,0.24690249552455457,0.9102970617517715,0.7997171110384952,
 0.17677520837674976,0.9355869230225792,0.21752833971214458,0.421340801677907,
 0.9852412659528758,0.9715402535875899,0.3999323887915317,0.4700747819911879,
 0.31839994751400624,0.5062441508033102,0.6918263335215988,0.3842870985030359,
 0.9488303014171149,0.7998492305343599,0.41088933028012786,0.0433411824593285,
 0.23882183485056696,0.7821028656485285,0.9591595111251852,0.789613283778318,
 0.6218848026826542,0.2867306772164735,0.4406058335583434,0.555984978245697,
 0.003263309695397343,0.44663387696359624,0.8983605165189321,
 0.7137077691599008,0.16263501098441402,0.42656279004245223,
 0.4491262906237079,0.6977768632846625,0.8229752104951326,0.4044216412607724,
 0.47061249562111673,0.2774605916497862,0.9071968530511891,0.5630624917110683,
 0.6205079179430748,0.7336193280594128,0.1772809184988008,0.1127036550153937,
 0.46362125313154967,0.6400840888024113,0.37802040793733793,
 0.5931928500768249,0.47368837169460065,0.9319115282587463,
 0.32281202511847884,0.3533131811698389,0.5918752674405334,
 0.29263577162106436,0.8638717787847701,0.7299959255414292,
 0.10068913307418781,0.6605897616137297,0.6158683106599515,0.9730078313325264,
 0.47441609989662425,0.753141014601369,0.4426029222441532,0.7241839650750406,
 0.29682298093799186,0.20765613725272847,0.2702656409789018,
 0.3634243887303773,0.5981236967068322,0.16659414634897796,0.639799197660169,
 0.18616403252091707,0.9189354770423592,0.5131232979643956,
 0.15054280450233526,0.007123605732512361,0.14204168809971462,
 0.3681017614436237,0.16866506762121725,0.01657398390042586,0.500001933736685,
 0.43122493061916234,0.7281596938831837,0.4387325436217394,
 0.25985991581274126,0.03942848210419114,0.4135102588836237,
 0.9494156967930614,0.7458309400918193,0.07736375401883411,
 0.018984186652169832,0.9334506973962488,0.5649385073202985,
 0.7343781943302752,0.6092394275141254,0.4567182684675595,0.3842997093633205,
 0.5660494726941544,0.18393142397308726,0.7114153515906488,0.8896052821404382,
 0.9801343397735269,0.9733116353628977,0.42028677523701485,0.8279873209629456,
 0.9880390871654612,0.7335048560475582,0.5800698449369901,0.40346229523088484,
 0.9430421170509907,0.21792843501592107,0.5067501846757314,
 0.03648136898637666,0.261911961117975,0.48451546250701594,
 0.01499451888540726,0.12929366491530825,0.5924772392709301,
 0.24938619092297243,0.1386570764156595,0.45607326987770025,
 0.9886389100500252,0.2265985619463906,0.33849370125453815,0.4956359515660824,
 0.14781795524920627,0.02171340092311702,0.9082249119160969,
 0.7473013377795356,0.5102181447755817,0.10935035614505528,0.583996010095515,
 0.6507265387702793,0.01334423125920503,0.030909953634977172,
 0.23276186111870123,0.3075724447778665,0.1732539605546598,0.7502760630368692,
 0.5722214133324454,0.353736758273951,0.46653865036547326,0.7830032058476503,
 0.31899636456846325,0.7837995555872898,0.5511373596504063,0.2478953100738781,
 0.16244018276626504,0.9620684516135386,0.41180606994829094,
 0.1260342005695334,0.24631152204990137,0.05897856956373093,
 0.3367007079716513,0.5468273398380151,0.2906696480717097,0.31097066366194637,
 0.40215685574035487,0.8404433492493456,0.7338384820999033,0.291318259047581,
 0.031566840429797405,0.903715315627455,0.062064676682417,0.7590692432971184,
 0.1390467765359562,0.4049067880965137,0.8890046468018975,0.17796925791161677,
 0.7554686871622428,0.053333151084556274,0.21469045451870428,
 0.9755249085415857,0.13087606092193682,0.6562599609662684,
 0.44036785082048324,0.6688877935169103,0.9441268712677094,
 0.029987697196326102,0.5789471483073235,0.3150702655973079,
 0.6674447172604246,0.07411340128923738,0.937439324777679,0.5454893357717503,
 0.5416375088659984,0.7854398063645991,0.9550032757941112,0.7766717630849598,
 0.62604459835443,0.412746402655666,0.427916837698479,0.5268931073892955,
 0.3160866005026435,0.7558735330729153,0.13545950955581543,0.9898849374417938,
 0.41047062768849063,0.5786625958452045,0.5545234282156115,0.3853207833621993,
 0.5131643346331779,0.3959479113856841,0.9677920061985871,0.39654254472594497,
 0.8126247764674209,0.6724713414623193,0.3067698405540549,0.9940684777780718,
 0.5677154619797545,0.6071363741583753,0.9874520379567874,0.10100428504097114,
 0.37627390760453816,0.6818833327753858,0.707447979101978,0.0701623728947558,
 0.7562303263046777,0.5100605084091059,0.23468831900577625,
 0.44000195054321534,0.8245087529488028,0.40558063030279623,
 0.6521385212090131,0.6633805965707215,0.7481283730296208,0.9588790512029951,
 0.6512146675031129,0.19774869792332128,0.953480003349865,0.7461936040612577,
 0.05712322205734244,0.019959549264151222,0.28063832987370674,
 0.46582120851127695,0.4320335295182264,0.5761552406306016,0.9845890051987944,
 0.7521575788711222,0.7458564344028109,0.3634698921436551,
 0.030415483274625288,0.17771538798677755,0.4710220265824908,
 0.41124790569282144,0.3962500975096871,0.6930881045730974,0.1559141095416242,
 0.6268349895956581,0.32456117493849324,0.46469106643216707,
 0.009516798327011244,0.6567099615898411,0.43146323530659925,
 0.027201659113354704,0.764760237490612,0.7436205507110727,0.9269491578237037,
 0.8243637032694671,0.48114042768981036,0.7602152228539711,0.7211366416440107,
 0.7566169340253868,0.07569705956231354,0.03262877322847646,
 0.30289799212538526,0.5499174397656138,0.07729435785423588,
 0.29444245846276795,0.43632890121694956,0.044022890987974184,
 0.7728969980107118,0.25082868966890415,0.35647675394389355,
 0.9247737840375226,0.33192272656225474,0.42583394259429974,
 0.4157670618017695,0.744006940911787,0.8563168522116902,0.6159247040148378,
 0.2682928756633194,0.34972213641609684,0.6803742433602274,0.4774750953090071,
 0.6850346521587134,0.7583262101467978,0.4920356636695118,0.9687952579386873,
 0.5599697949270481,0.29135119273983956,0.8259032917939149,0.7796106972899102,
 0.15725291700163746,0.2566475534176059,0.24144293610009704,
 0.37837786176364196,0.9223215508854576,0.6060655312509293,0.9616158191062135,
 0.3858079961396047,0.06454873823804097,0.08483479344229339,0.859787247026248,
 0.03527936894740957,0.28667420684267986,0.34742478464864845,
 0.2867988673062638,0.37572811569342857,0.0623221549667754,0.5531100248180676,
 0.5636788277331172,0.5684564604566216,0.8442757556882272,0.9000443247421503,
 0.06444851059261236,0.48557394448739144,0.5764683566531148,
 0.9899929037066182,0.4106384474816972,0.08005643464932832,0.9468182161862861,
 0.2394891229879743,0.3872536137677217,0.8528322683613428,0.2693942567665233,
 0.8958462356973341,0.19486833940209725,0.1945090990180962,
 0.11035093467510926,0.08828648050771837,0.853722664971383,0.2607991766600849,
 0.6879209412402618,0.24509311948459644,0.8117891243408549,0.8825236805346318,
 0.6805570669999298,0.6015094874634604,0.3420892231459517,0.165461739281498,
 0.29440954653803153,0.4664861557913812,0.8450761318987143,0.4441743077659689,
 0.9622641832042217,0.08403146994525024,0.6396290467736063,0.3166510531697565,
 0.01723836508083898,0.03615732642816394,0.32556603715327626,
 0.8455690840829215,0.1035472758253575,0.18316911320270224,0.4734222923511294,
 0.5637622233183225,0.4378052304759332,0.2668734932672914,0.4434533415494153,
 0.9283719851791202,0.7933358210765612,0.7271606631194185,0.3205968249693937,
 0.8070935895012672,0.9323664408515709,0.7656097058465319,0.8281318980626841,
 0.9598232056314822,0.2301264718135989,0.26804272034510035,0.9641005028393012,
 0.7876856175107505,0.10028569304378143,0.7132466627088383,0.7879894001252431,
 0.4930423537183346,0.5656946666151189,0.3977120142623205,0.4093970631991648,
 0.06506383949121641,0.16168296962440076,0.3116176067968597,
 0.3266739041405071,0.9685234222571799,0.10038742710998405,0.6416672940645896,
 0.8196804875656972,0.2649815341137791,0.34760165897673045,0.0976190726057945,
 0.6886927471980913,0.3905799189363104,0.5434750221352957,0.9494974126952611,
 0.09216160092331149,0.048896142447095836,0.5883912091008845,
 0.6677713546264605,0.7231360176260345,0.8179389436505513,0.16371874452629576,
 0.1827250737332431,0.38666646851566633,0.2252187854877976,0.3155712506347579,
 0.7702240498846711,0.22667921092153387,0.7912109686763197,
 0.18654315471962157,0.2603474974735466,0.1360588507802748,0.5184429768478498,
 0.5796984301799133,0.5315349579186216,0.11530952907932668,0.5615900038316624,
 0.4998969023027354,0.2342245880865803,0.4515162139516504,0.3162043976706671,
 0.8475073706911315,0.5116916717669117,0.5748191156979324,0.770830668729649,
 0.2629205520516592,0.7026553604180109,0.5637449294079124,0.9200533716768634,
 0.7278481977165621,0.9143710565587237,0.7955780562303709,0.5472886116709348,
 0.4879879245409503,0.38167245838485964,0.45843228058469465,
 0.4493461852491667,0.5156041472897852,0.24054208085805062,0.3656686648919113,
 0.8205390199289287,0.935774567534965,0.19167186945177628,0.7470303422483476,
 0.3013543541363546,0.7848244647230357,0.5911100981610612,0.8206547748676196,
 0.6204924659548579,0.7375733554248713,0.14510128684654422,0.4197116846178063,
 0.3291966200377787,0.849050215985657,0.9187571179099769,0.29887763562353264,
 0.9558600894869748,0.09372774802280304,0.8920236256841343,0.5780546851690492,
 0.7554418230229547,0.649203231293145,0.3170775077833058,0.7902942435467365,
 0.3535306304320428,0.5618072721852944,0.6048661232711285,0.09434164745936047,
 0.09344732022485192,0.8514607053269793,0.8735355054022226,
 0.18395322245404766,0.5896251548124882,0.6190593118092601,0.1505321589715385,
 0.6921463550700846,0.7554815909334841,0.4508142611407524,0.9872166858077678,
 0.43527730865166236,0.11969204809197276,0.14739900467760936,
 0.03913145180845712,0.45313335046582215,0.7016508963475104,0.20947948126048,
 0.9645425254258606,0.6573840220225946,0.8064795165845636,0.15071455868342265,
 0.9739170067852434,0.6491160395043759,0.35001143887114705,
 0.22925477772133518,0.39980499867776587,0.4989890545012303,
 0.42353901558590745,0.6414631295680416,0.04049063220321292,
 0.2727329786197066,0.607855001711201,0.13476811247053222,0.12118586212800764,
 0.5745134554509841,0.5347924926969838,0.8614908265733606,0.9722937661087225,
 0.9350057253797401,0.0948736508132092,0.21702142191723406,0.6053481808886851,
 0.5659311149897671,0.4692527376522312,0.018181953891079794,
 0.5736046708430067,0.36884871670974595,0.6872236882162478,0.462336290427521,
 0.619706876092038,0.15438089179723513,0.7735968840123687,0.19616402368191554,
 0.27196850428308383,0.9804031416143859,0.8994127869222028,
 0.10852746977582994,0.3077313459926534,0.2195046325897696,0.9973606717386552,
 0.616129702043437,0.7849257692805875,0.1873185021515298,0.18795858265977405,
 0.1195711534278141,0.1959686554590524,0.8615230235181268,0.33943591716378485,
 0.5151173038243451,0.7336160040427775,0.21610289318284626,0.5409035786051021,
 0.7482779621885535,0.3385757554960922,0.17956754278657594,0.5493985001683349,
 0.08400782840685905,0.8563576551718626,0.0032346694325319225,
 0.560970290000701,0.9064154405749216,0.773960170600897,0.2990235374470154,
 0.8043548936588715,0.38256177785972134,0.6596395777996086,0.6498784147206274,
 0.5474940872256944,0.8267597084575123,0.5984089760640796,0.11511892069402352,
 0.9424063364866982,0.7352395131482007,0.6566514141924888,0.8252170997849447,
 0.07441534123147286,0.11870809376328217,0.6936300663569914,0.947076684067862,
 0.7573674210043012,0.4270797457426345,0.3354111090439288,0.2072635784854806,
 0.2895138661848169,0.9687187160107597,0.9767358797191149,0.8715610197293919,
 0.6194846152880983,0.8063252916932523,0.8030519751716554,0.23065152459236127,
 0.8985341309382319,0.29477910433491583,0.7753043837605568,0.9861537721707111,
 0.34175365871176944,0.46041452072564115,0.3304428863035368,
 0.04508429506370937,0.23237782207312363,0.11328243494610868,
 0.8130346391821169,0.786218724545012,0.389372370763194,0.5449227468188569,
 0.17802707083662828,0.5847134275018775,0.6292781627559721,0.6254610644298922,
 0.1559966184771444,0.13064152578621768,0.0045975746430810815,
 0.07402208696661883,0.39364906935504906,0.013727829906720057,
 0.3922487272291528,0.0728350945217241,0.4684176223776544,0.1465126784767179,
 0.9371794817205774,0.16662830692240282,0.5946309126488457,0.3807761111239194,
 0.20978708222644582,0.5675056360164628,0.9837245839148181,0.561425703587112,
 0.04412942568659528,0.5221224262399471,0.013685752180052857,
 0.38723115557140897,0.1358429240709662,0.6148533600623723,
 0.42638520984986306,0.6999928866634736,0.27708962924028613,
 0.5566334312414054,0.21558020676312095,0.665855331167677,0.28075728014919044,
 0.07204015777881845,0.6455388107090149,0.033811425586654065,
 0.7873423297906922,0.03821495725321955,0.8950717746242938,0.7834637488794673,
 0.8227953897734831,0.9835376149060705,0.7199962137837008,0.14696397364608338,
 0.23722434037032636,0.5509930744946467,0.7448209213099782,0.5724289779632301,
 0.42308237306322083,0.5814281153798825,0.4388614930467085,0.9945604979198367,
 0.9284632666995046,0.2752197624443906,0.48497348240884675,
 0.17396802117822419,0.5329864874652426,0.302999299491865,0.9315824781150719,
 0.23985653785836525,0.4802774969356648,0.13808279400235846,
 0.028589963613024905,0.8559414086420049,0.94478883673034,0.14973550404111358,
 0.16682378239352746,0.28039796146464924,0.7739754819923776,
 0.9060745910925284,0.46415317967458714,0.30316495393988385,
 0.23822593472202724,0.2323856828277343,0.4445448694480756,
 0.07915584308083878,0.34221833469618956,0.36092029006401094,
 0.056532468124216084,0.2959668087863475,0.48076072503412126,
 0.520107850936911,0.31987495823304424,0.6612255951599902,0.7970188141914636,
 0.09595024824634413,0.7872230046552251,0.5685475667353257,0.6899295309793297,
 0.8618730952776543,0.47829902639769095,0.5677894923647929,
 0.36065852267028364,0.6519464888020418,0.6096750432590798,0.8366555750945017,
 0.6739345555227447,0.20395584580225123,0.844468876387171,0.0980811313103247,
 0.7867678769795329,0.39040625930666417,0.5339614319809987,0.8199455839426725,
 0.01784685021830157,0.9156174690201643,0.8364003280194909,0.6850076822990905,
 0.8997979370097948,0.6964451580338025,0.404119138250842,0.9619967904830565,
 0.24118627603148923,0.3740208065535885,0.7691688011484765,0.3217654247784585,
 0.05048636889554725,0.5134331221633108,0.29403302466212144,
 0.07170748959612527,0.15595752156704612,0.8894517981938794,
 0.8918333910720162,0.7332292089709245,0.21590376204475448,0.8130117432780384,
 0.20941857993494994,0.20867853321397145,0.6872123899327914,
 0.6224282009482537,0.3504495522632143,0.9764282158658195,0.8504347457852197,
 0.006560883146795815,0.7610482763875771,0.12154577396075594,
 0.9845652257764962,0.7781935386074772,0.10701908731515009,0.8438064814949752,
 0.4331849939001202,0.5723552927482759,0.8378660392384587,0.45702699935204394,
 0.5547258230526353,0.91774145088124,0.24142581877515656,0.9817956018506078,
 0.9435607183118172,0.5666659191578544,0.17718350932715432,0.4237263819592001,
 0.2820822960375651,0.13779075610837976,8.023445743942093e-4,
 0.09693208475584336,0.7351474584567637,0.6561047611752374,
 0.13850675722694206,0.17643850018307117,0.69801824491202,0.02862908292280475,
 0.1909154829649392,0.2690664447821842,0.4388343984647002,0.11548556670407306,
 0.41351021163823,0.8894532943914977,0.2632809341083848,0.7666821628549673,
 0.40261626722916144,0.635630667560489,0.29307484371880144,0.6061379497396198,
 0.5279399080181231,0.15201777216012735,0.14435847427337023,
 0.5822962093435274,0.06533083262966033,0.019785744168989816,
 0.7581411067264407,0.29495660125568113,0.9314679432992929,0.4651794364266162,
 0.47973184750892683,0.474327925028585,0.6548019354234639,0.5719836854655699,
 0.4896404249197107,0.543727288050205,0.8161158600446576,0.21352341184749613,
 0.22113272399214456,0.6345169277555857,0.03549037519126996,
 0.012139281903675414,0.6494198264110634,0.6628053251965063,
 0.8208965909637391,0.9035495923438209,0.5610596702629173,0.9671101016641935,
 0.24717807879845666,0.2418104618601008,0.20210963981007224,
 0.13920345923920152,0.5283412422339435,0.766293284790322,0.8171901738894373,
 0.9678836722776597,0.04559287840664761,0.3967350115801046,0.1697770378457878,
 0.6726047074906486,0.34733869358610725,0.20143657044660124,
 0.6653546076797519,0.1344346959487328,0.670549194296067,0.5273355622724852,
 0.7830035923466424,0.16746291849009298,0.7917676442230301,0.9137422496749017,
 0.8803776756736542,0.16973410656132715,0.37390305824633496,
 0.7284423961407572,0.3282971113371811,0.1405650046855469,0.5077542143893132,
 0.10253446934552002,0.20036674323036596,0.6136224690643943,
 0.9967586929997454,0.18571837335694585,0.9061375795671485,
 0.48460468456099015,0.5705202225475675,0.7433323474173392,
 0.050652913490730866,0.2539473906865457,0.24894273912420772,
 0.9793723576632578,0.00938176756930853,0.7845285356679195,0.9267466033685617,
 0.9472241598548873,0.863769693364056,0.3722613027578884,0.1851660819181561,
 0.8885049440986857,0.9236725047492076,0.16953499561896757,
 0.07617570365043425,0.2374921527948859,0.2792411602850762,0.3350856304236287,
 0.8454532116922402,0.3037679926620882,0.16025081972778032,0.9405165256314756,
 0.8057812653311645,0.6400737164604514,0.12253042022169092,0.2684926937504444,
 0.19140487754443192,0.4538119033981606,0.5059651370799242,
 0.08224857880931158,0.11256826409273546,0.45179294529206027,
 0.11274773953561001,0.49268160838148223,0.09857504509164317,
 0.6029739681157491,0.35750185601458173,0.8372986091640897,
 0.26375822987986464,0.8740320721694984,0.3750567338705485,
 0.24788600521668325,0.4507499790051078,0.40091570982657476,
 0.6018514748186328,0.6778616102052893,0.09599795917151488,0.3430255810340296,
 0.3478334114660563,0.10679518159726664,0.024485521381846342,
 0.9659730665191765,0.9063853402843633,0.03262152786218431,0.3696631058790627,
 0.2948023649453595,0.8853871193160734,0.12562737569313986,0.8156885368002692,
 0.08350190107984945,0.9377306734458051,0.828996124808684,0.2512593800604811,
 0.1487529391616903,0.7670331641610029,0.7572551424267912,
 0.030624954885560385,0.9402950213673843,0.14872753867530064,
 0.4633765041413209,0.2367992213420178,0.5625332776219901,0.275036928502708,
 0.5430734018530398,0.4243789245671328,0.6402443606892876,0.628181176675016,
 0.8417441012298743,0.4362013698343201,0.6568894711518496,0.9403960250023304,
 0.14939717811672804,0.7780150681295647,0.3179892622472792,
 0.21456858030851222,0.14567628313320247,0.4875751722609022,
 0.6558413284891373,0.8855182669203209,0.27711155868023696,0.3787435709354986,
 0.6459900929429965,0.38762953643669107,0.4377296645624392,0.5274898122083821,
 0.5794048549434064,0.41020583050670634,0.47167179463133324,
 0.3467015158369593,0.3450778286045503,0.31996449993371334,0.5296154671933925,
 0.03805514849380598,0.9958576951266622,0.1551787039077107,
 0.19122214533304238,0.9788538241484339,0.439408167660329,0.8254431456084153,
 0.6599099842815375,0.8535325330708649,0.032689701334234766,
 0.8592976310416455,0.020950054012356945,0.5196811137954156,
 0.4506025892937632,0.8136073983143807,0.3296323648416677,0.23781474796655988,
 0.5628322795086392,0.5924936987946435,0.6772443368732843,0.950454725290008,
 0.6418799401420968,0.3051752110536218,0.6031110091088768,0.9923761153157822,
 0.959047643313364,0.7834266731998116,0.06807580542081815,0.8045552665753042,
 0.054765826176557786,0.570711602684417,0.8526515072726827,
 0.35662295288152446,0.9323175971957481,0.45064822654765446,
 0.27458508456846387,0.8362490273637635,0.9085801202109249,
 0.32139723162222966,0.26900828261201126,0.45286972854167096,
 0.35486925524141444,0.9226519609253361,0.24484054361113006,
 0.5217036984853812,0.9754551088094553,0.8422193352242974,0.8025689415385543,
 0.8167270524300932,0.8369886906046319,0.6079675329506292,0.8236816751149154,
 0.4196541619449563,0.08366657547286027,0.18455381735911325,
 0.43632870523418665,0.12214125595211867,0.3954393423701713,
 0.7484868545831191,0.6670406109168796,0.5403125351025859,0.7478942746778133,
 0.21742084398514527,0.25944782563371793,0.8821568554302969,
 0.4430160937125387,0.748970480381341,0.33131284798625593,0.3890299665432818,
 0.089373980570014,0.2882959807119396,0.237933718336779,0.27475815034272144,
 0.6342465637066709,0.5555352754331712,0.5196786848190719,0.1065639183754945,
 0.5369791775898708,0.5583030669240633,0.5594613328016987,0.5229341415535791,
 0.361952022234607,0.30604776329839123,0.6027128193631701,0.35394111908510106,
 0.7000125636956667,0.5216086915136093,0.5208948650695067,0.7656286850992592,
 0.9497877214802477,0.3431211291832892,0.4128503500977033,0.5418911046416248,
 0.5836614744058004,0.8074129808953647,0.5154488844400107,0.39751702523690424,
 0.945994415115934,0.48447822784371575,0.4945434153907935,0.6097887906957072,
 0.8898949278694476,0.6038856033236015,0.16917486903165924,0.3690024052903156,
 0.01086583237484906,0.7845231894445565,0.11350817853667594,
 0.22505334799792587,0.10309282911229045,0.06488836816011911,
 0.7176650000678227,0.2501452761901133,0.8460574775992833,0.4145943788001939,
 0.3520480004266998,0.007263882607179473,0.3477996630516563,
 0.10785414745853794,0.5545635457182629,0.8877440123844328,0.8868462311074149,
 0.08127387265716601,0.4462400200916894,0.055577893714502946,
 0.3337871031789633,0.7839827559577284,0.5472660291218387,0.8666981698699942,
 0.0036636566502044055,0.2442132636994936,0.5230950407978929,
 0.9508310875489893,0.06247530123470002,0.12489748864806116,
 0.8595613883226334,0.13764978642573722,0.20514344190819866,0.711882963419043,
 0.9684559776135785,0.7839079823268353,0.9948851975050769,0.2076570823699192,
 0.17478134535573397,0.45964903092021325,0.6290650595298778,
 0.7775263204776751,0.18264230306061546,0.7642664981884796,
 0.06885121859148469,0.8922715500475887,0.5426707458245004,0.4971319437490387,
 0.06574540890713443,0.014947759348009737,0.7267245831248671,
 0.33848199088189523,0.6392384120654901,0.9756181226903983,0.1967093450639601,
 0.6179008463282805,0.8653728890395683,0.6743376411387713,0.32449695817665636,
 0.7443274524263129,0.26837064707135216,0.6542934562580093,0.9950781031146987,
 0.8336974674415869,0.7247202021201335,0.10002526466352879,0.9312476027347539,
 0.5701454740209622,0.16008007599270146,0.9534829346057992,0.8757664675559851,
 0.1886018100957607,0.9975089574922765,0.7219147638834482,0.7884233242318617,
 0.4161705301300206,0.668388452379793,0.5824320793458408,0.23182626594377942,
 0.5344440855963558,0.82360951495937,0.7781015681328747,0.47359009971973887,
 0.021832446222543878,0.417589805744646,0.27014474599310234,
 0.6776691636843484,0.8049122883393823,0.7902697534761348,0.09649817121793125,
 0.07236358381884689,0.4388304447792932,0.24966905633610426,
 0.5834832832835369,0.37168295916132266,0.3558726553281317,0.330205334301366,
 0.9706636012039573,0.8626666341523713,0.3028233336473596,0.8441807131268988,
 0.0010906884656197713,0.22900801008914895,0.28514323231809746,
 0.8622027309840595,0.698772950659011,0.17016794164855265,0.6384931440484302,
 0.1777876178488379,0.7770663044899001,0.5510463506699156,0.2887929604547309,
 0.006651894428309824,0.9176648058898332,0.6870344302493122,
 0.7604491773668551,0.49098260886664136,0.24248306709576695,
 0.13149342957976717,0.22505948940036435,0.24872387513643424,
 0.3346905601929513,0.7290598641364493,0.3446900044858875,0.9909256403798049,
 0.4419072934940287,0.09903488477206968,0.9299436830196609,0.4887536399880831,
 0.7876417831148864,0.9513644755588078,0.45877538288087627,0.2954018099314004,
 5.961194507764223e-4,0.1889086972667393,0.7020900206420565,
 0.18631816213490582,0.23423511862604662,0.09918314897855207,
 0.43800759574413906,0.12730322662988158,0.5882321607394313,
 0.7067911982898953,0.38259711005554076,0.5895264021091278,0.3676096717541585,
 0.9745114762043249,0.4699659491536984,0.500448766302036,0.7071244923537463,
 0.01702291736548145,0.156662798174092,0.5591642780261239,0.11599990691621509,
 0.6557383707485306,0.7406697615245311,0.2296135993585291,0.7736686715215935,
 0.4390506046740098,0.5702473969888548,0.613505246903189,0.9752235999614901,
 0.8225400406915148,0.8871647099094723,0.9645603205534636,0.545886704879195,
 0.7833611886541336,0.5034514033327406,5.200512022240922e-6,
 0.9609428556445333,0.0068212019512570166,0.8129137926642944,
 0.8155574762582825,0.2482875559212756,0.17552533154811523,
 0.48476195287187285,0.5542973987852172,0.604313905357847,0.7744140712410521,
 0.850165357792463,0.8999390898049441,0.33459942805332044,0.8071792329732576,
 0.2758925521312271,0.2980933584316795,0.12765503680458945,
 0.26833883427155647,0.25933073325476164,0.04931139731081946,
 0.13910412634730918,0.17530656454870952,0.6099373243907613,
 0.6495806417206368,0.5780279715892538,0.23571911198363082,0.6477950263870146,
 0.6629119595166466,0.4649819392069111,0.8129192882021482,0.7733790310154829,
 0.948062821466309,0.7569660074269495,0.14849600674232932,0.826002819764634,
 0.5652706675427819,0.14954882245093737,0.07842332808505748,
 0.7372671247677985,0.6670399231178854,0.506370520047668,0.984664444355004,
 0.9732602946418557,0.20464140655901952,0.38338328430379276,
 0.5637333812319363,0.8383267610439116,0.47407207617618763,0.2475238701118122,
 0.6136875634321444,0.20529423518406642,0.6202750579522865,
 0.11856896300068409,0.81764788253847,0.760647450621243,0.6931391882430595,
 0.12871948712315373,0.7499203944191342,0.34102692894959885,
 0.6245848088334204,0.08826544655846058,0.9908210923423196,0.8500663020668842,
 0.2571840503027707,0.17754898614495362,0.541435835937021,0.39433560239735643,
 0.1309502024800121,0.16110767048225139,0.09456582918690304,
 0.7343162386299132,0.028726194949605377,0.6030929968033231,
 0.7528601762977841,0.14996202661320135,0.974762383994498,0.8942646549982546,
 0.7797165893718797,0.6204476433037464,0.6660698129916988,0.5772050557273366,
 0.3338774268711019,0.3477897563709853,0.9178629752831177,0.6948186147001679,
 0.8785294551986265,0.1442179668500615,0.024120505574388718,0.524828308007004,
 0.2216398527361303,0.43165611724894126,0.4222687539692026,0.9804516375858934,
 0.1057134781289294,0.9510363952881867,0.28773094782513564,0.2629592339902489,
 0.6722702252127037,0.43826054371689216,0.07524515468551762,
 0.6083849680259028,0.2904259850751828,0.34247050329442497,
 0.010282236118028143,0.8546030846302972,0.33421051854643324,
 0.25463167758862204,0.41706051529689514,0.8269197126649369,
 0.06159010537620335,0.6924892080084236,0.10000506164780665,
 0.6043565393943066,0.5326115481114316,0.42768321598454273,
 0.09121244878949497,0.2045426391235584,0.643622681881317,0.07973558237093092,
 0.10832327181363632,0.15874545917368854,0.08870641747912766,
 0.8849634848356659,0.5542880927310581,0.6396512921760592,0.11499135287977891,
 0.8486328177132559,0.5849665593273599,0.03979836108764234,
 0.11920567245709668,0.3040074436385959,0.8126794022429542,0.4307975669668953,
 0.2506874623667401,0.2593147030063554,0.6715935564883236,0.13802430837310142,
 0.361825913840947,0.4072705060992907,0.8394672267587229,0.44955031339309504,
 0.2103871898236379,0.029878757225581642,0.23233832775996965,
 0.9012142211440926,0.5797572813265548,0.15620430449214096,0.1528258254226479,
 0.30421277889357967,0.7090672618477464,0.5331951852826851,0.3130686161109997,
 0.7849043481062364,0.2194634093875737,0.7224068059190182,0.06276474347325378,
 0.06916784071140858,0.04595535631433367,0.33325501669352,0.13363966545381933,
 0.832482130522216,0.15669603373410623,0.5784194049692845,0.22863093106907795,
 0.08164771522529879,0.5697946313171302,0.3408636490099345,0.7086731906741665,
 0.19918470273572086,0.3076375469213527,0.7462821219757226,0.950893424849764,
 0.08561643628736881,0.7205176497233037,0.5435675595726018,0.8983208392564802,
 0.35461271065664635,0.18867516670204265,0.5439066456789603,
 0.3501417145511102,0.9404745663946903,0.44411174473042325,
 0.016227544897878032,0.05054875612748444,0.7177873210718754,
 0.49603775653538995,0.43007511417813604,0.5564088523109092,
 0.2069546410228913,0.39841264441329427,0.13687677634667605,
 0.5723982635996995,0.43918150459677596,0.947630516618615,0.577361597550255,
 0.39570789992633504,0.6354092443820553,0.8556542636063145,0.4140233492930585,
 0.09763764199410285,0.36201197784041705,0.4775634747553196,
 0.2407709338710745,0.9299726649660508,0.7325867572722577,
 0.031021366473535927,0.14535219049119597,0.06671319360446071,
 0.30621590497427076,0.3914790558042487,0.99197997525756,0.5344879966842058,
 0.9621093825605597,0.7854300502549401,0.060728845961364075,
 0.6494610397774904,0.8507406007990691,0.6730215356315874,0.13346978528654418,
 0.113389735217167,0.49992889110464656,0.41723070444974175,0.8030028957184941,
 0.8354264281457794,0.5936241346778101,0.3778641475818951,0.06394747960726788,
 0.42070062172248024,0.9837271568457969,0.25160107942459153,
 0.9547200044646651,0.9045038076684627,0.48691584882192207,
 0.40240680942191265,0.3248335788350466,0.6333828954914102,0.5729814604234212,
 0.02773285986521845,0.11083348638770674,0.5479150160458798,
 0.41072620137069205,0.6239119911795754,0.03137703066834607,
 0.4299986940705074,0.8539935183960051,0.8809489858462627,0.6796000617022111,
 0.20224331967189013,0.8432579914230858,0.3675244854462517,0.6525841850572426,
 0.2570380637050288,0.9114867984695338,0.6009283852525472,0.13054688638222856,
 0.1847735574018714,0.09408060866387413,0.520281179600504,0.792101652982288,
 0.2608232796473875,0.9443706153767195,0.3117289156394194,0.29077120855665506,
 0.3002574330107275,0.39218849376245035,0.3578988449264979,0.9177659868027029,
 0.09702985354539795,0.30776304403037213,0.24175181876446072,
 0.07471769584082044,0.9154888568320341,0.49221551839475985,
 5.850664254667404e-4,0.181265912128298,0.4950346892610765,0.4415256685394503,
 0.7093781146077319,0.30202127165884696,0.6239423556222398,0.8125504104653787,
 0.9833127038569585,0.9957769084737365,0.16186091734984953,0.7017842699970182,
 0.962774918473507,0.2736438867066173,0.5879173229530491,0.7357392407957167,
 0.39362277729042516,0.12656554663195207,0.6047435808348378,0.618299156274154,
 0.5044726979844821,0.662083578768748,0.11724377791894602,0.39313298164292776,
 0.31032276133033443,0.3744523709896591,0.7011119990831,0.16789773441802014,
 0.29526431290876654,0.4176643055232291,0.2462982552059816,
 0.33930808064237195,0.5724934518120443,0.2977122107178264,0.9068810756201733,
 0.2178367424783989,0.13102482710235575,0.6572070576013389,
 0.37541031502124766,0.6533674305092224,0.5545829985140543,0.117721197628193,
 0.5734759133005363,0.4991558792484111,0.4347165566860073,0.5456645838288905,
 0.5716479460255531,0.9376018574068283,0.36720739756389875,0.2662206504416137,
 0.03951337440716285,0.5488853791253903,0.1009576020989188,0.4488621341377411,
 0.41278164960496755,0.9494873057514618,0.36606532975184913,
 0.5120109912095557,0.08968984575530325,0.33535377342457695,
 0.38887116522138454,0.41872371090074334,0.12413740499337911,
 0.0028166830490695682,0.08112180813754355,0.7361197373237803,
 0.04549719070881497,0.018862994072774586,0.31416123200469204,
 0.6263668540980877,0.5068460783092579,0.8636517101750737,0.7271254406021916,
 0.6730226730049731,0.7765530362903817,0.8049038544621867,0.3241225562734633,
 0.8691297610325235,0.4862051560166899,0.602016272761503,0.9584803328679294,
 0.0740154855010502,0.09590428298244458,0.7738458570781397,
 0.40959041614313374,0.3514328866341725,0.47843729799979084,
 0.7420211172460455,0.3495354509154439,0.5583617569458156,0.9150999424863174,
 0.6090528625638584,0.7152595702485891,0.4574079805795239,0.9432662536726445,
 0.530622315404009,0.03878534787792687,0.763301741836397,0.19798412855677883,
 0.334112803455329,0.9688917805140995,0.40731061033420224,0.7186347620568712,
 0.7436647282786801,0.14835129100354172,0.6584161605877944,0.4021015617376644,
 0.42891261861381447,0.4056921563647762,0.0568717630643053,0.7564479856530664,
 0.5511460803506163,0.7666726063479871,0.3395754657391231,0.8227151105776755,
 0.5257066258236698,0.43320593493221793,0.6953921657856842,0.879816212669933,
 0.5545207678974264,0.8764402881533112,0.33672595703143704,0.9880129662088875,
 0.34560404804758493,0.0528987250933356,0.2344988181147839,0.6601662595729733,
 0.5276536313478434,0.8969046334066408,0.6454479631264841,0.27315251965789267,
 0.4407144584520958,0.6166958152256921,0.07519044781557016,
 0.45192596605314383,0.22888120518342248,0.1505978481950654,
 0.32927424138821204,0.1937249300795486,0.08629526037139212,
 0.2835141203994159,0.9679323227270963,0.3271596348718616,0.6435899924236198,
 0.5304364180448229,0.5753772730961393,0.9452852213269215,0.33758389239470177,
 0.48430571100069353,0.1830460198697509,0.05068884428308662,
 0.11695488110410279,0.5343917283035023,0.3991264438806613,0.5190096987272094,
 0.8200556741367291,0.2535042502872531,0.5159523258367379,0.20439176998674669,
 0.3285195977637887,0.15457299974983363,0.03501694055226978,
 0.9298913559890549,0.6315482951674251,0.02731867516282982,0.3097927603244658,
 0.6419977872375819,0.9851224227567266,0.889763545200354,0.022985362635296824,
 0.6241896401303156,0.6469132815235918,0.6072872055502857,0.4926832921385095,
 0.4086148953214854,0.5975875171701691,0.02566207281996047,
 0.40945510025797693,0.8037838206889377,0.11822111771520105,
 0.15395221036093987,0.21366114481120546,0.8527369326313324,
 0.3521289510811213,0.12214750710594191,0.524761730901126,0.833396923637438,
 0.6098024859573484,0.3348256372613405,0.8392053317227395,0.762261602109253,
 0.3279274379064161,0.48213302027282123,0.8731142293261027,0.9451209355947119,
 0.8573385522562575,0.0911690954516905,0.989171057217011,0.2931280438997682,
 0.20678028017839278,0.4181482716963745,0.6777590807764398,0.4778985688534876,
 0.9815741949034421,0.2953830014488439,0.528994702001224,0.5041680375539442,
 0.2911207412850547,0.5269070806278426,0.3358378941232094,0.9080669366795903,
 0.417323121520234,0.42853710462273176,0.2064207272916032,0.47472998547517464,
 0.6012591548236486,0.17879472445072053,0.136966470468179,0.7535937755534574,
 0.10887866120709488,0.6168348497670013,0.9640535374317418,0.2895487675519568,
 0.3945130741802151,0.0823165882790502,0.5241831043604517,0.8545331729904045,
 0.18355183865239544,0.811294956324765,0.7824726646531215,0.4409573607644416,
 0.05533421067629296,0.2713168443234468,0.6117656806408024,0.4294631464837183,
 0.625212545469865,0.8259213453977314,0.425516641400425,0.26514341205016945,
 0.1412209275578079,0.8228212962688893,0.11321708723010293,
 0.08410621762145365,0.27876323131289893,0.6780858536004031,0.847594381143486,
 0.02667229660721948,0.9845390947843945,0.9035784153036569,
 0.13471174740429703,0.5059576269149879,0.8247786703918457,
 0.08029819270988214,0.7097186054340252,0.8537114342580105,0.7283684211755974,
 0.49182312261158023,0.9929596495602777,0.2875478762146735,0.9306004236724477,
 0.2914935080274711,0.3859535471511405,0.7687503690193419,0.7762482616944587,
 0.5619974307874657,0.5529706509843993,0.7894691056852432,0.12413514033574347,
 0.1252649906528881,0.3290254490296902,0.18808246877730272,0.3487266612374561,
 0.1886509491259375,0.6814026702677203,0.5140506466933493,0.5979425902787591,
 0.24411525963090897,0.33222031625823845,0.1229474578473796,
 0.7685252911441487,0.9100472474773055,0.5694547116464489,0.7754064080248401,
 0.01127012632634894,0.12633459548613024,0.5293425825290636,
 0.5996855829948579,0.6638333244097052,0.17883603328358955,
 0.03624337305314651,0.1286740361513542,0.5744738777895165,
 0.29570412196353013,0.37459290328519323,0.6842421027272253,
 0.3564271901384234,0.3178024520635431,0.7491987762983656,0.5569756686800817,
 0.5153655762406992,0.061383879130722296,0.5672343098393104,
 0.21931016343311915,0.9582868976064688,0.26372080337715276,
 0.035765946532361315,0.29375422899435855,0.5018747577105279,
 0.9175800453795466,0.22991264708368575,0.14121952179833297,
 0.024596967171470197,0.700284078655753,0.09223699734552537,
 0.3384622600548801,0.05164587764769435,0.03824279222752547,
 0.6906089499168624,0.09520522658539832,0.38753107353244676,
 0.4419880771943492,0.5866591902431354,0.8167112389376213,0.9713473521509142,
 0.8217037997612604,0.8981209544100796,0.4375127054607043,0.07222960476583073,
 0.6700175863376607,0.3495068711822925,0.8664240576807,0.7816937332232582,
 0.13531787482002167,0.8640988391344844,0.0011003308976355042,
 0.36104839204664907,0.3768029615355485,0.46403830099877774,
 0.9411237188628052,0.26280081029637337,0.1428663647993903,
 0.10529209838796161,0.0275181313930819,0.1412206165143659,0.9925139753887717,
 0.6713552521785676,0.9554675049234718,0.8626775905689935,0.5136687174158633,
 0.703023915885379,0.07837926234555248,0.46078143712027175,0.9287301652740027,
 0.49197672224215117,0.8218678237621347,0.8522038102836513,
 0.009057725653979998,0.12552201566490218,0.8872611746546547,
 0.06473831604007751,0.9880497784522184,0.36140046038950946,
 0.1641002607121892,0.7779619715601813,0.9974710049408717,0.47222636636026993,
 0.7967414365918998,0.06757588643262458,0.8647758870334752,0.80881557041743,
 0.3693837793570648,0.6928862296849798,0.3541893230351474,0.7513953367909434,
 0.8735464892317922,0.3180704228307747,0.6654879961172875,
 0.003959094353270687,0.5310427306575775,0.8619115560959958,
 0.6867860119029863,0.9051630435429686,0.4000888776382473,0.9930305807381884,
 0.9448622564990523,0.4911722786054149,0.7935236269236592,0.6079269877214596,
 0.08477109723567244,0.0431740684447125,0.6010694113360161,
 0.026481413103478513,0.1361607108887375,0.8026360048822649,
 0.12930691819467144,0.8925075789839831,0.9737189857214386,
 0.06359418271356887,0.03214857717649311,0.018955259276744796,
 0.12463712993758758,0.9675228276474495,0.21281484957953611,
 0.41775254987333366,0.6690368483114689,0.2833485708893655,0.8301549065091633,
 0.89003296589064,0.32922000130672957,0.7848605426702123,0.5419074788153158,
 0.5274087472336862,0.7642235554351863,0.46807673223347745,0.3311978183493084,
 0.6136203976590495,0.12440954334383025,0.3805874594920089,
 0.09499863693382393,0.9636308612248282,0.1604733618931018,0.6353527465070288,
 0.16939680889033826,0.12282059203413542,0.35725736045502066,
 0.13223387483795468,0.6458382279836907,0.036204812589942015,
 0.5785077885868863,0.3132426621524099,0.26125919320713686,
 0.15600979673075832,0.9368027533442613,0.35997339769204506,
 0.37754778566627534,0.8962579245673133,0.7666261407158761,0.5969432081089913,
 0.6545877270780478,0.5672869255897911,0.605603009170919,0.6459225748334345,
 0.4385351687577905,0.007793489935523867,0.8451832323372391,
 0.3633223831831768,0.38798567166091547,0.7293093729320591,
 0.30956252767330805,0.6484289663723738,0.8501304120903236,
 0.17463324206149222,0.02089678424845609,0.7659543675318236,
 0.7401116891070538,0.5594747527560661,0.13041672497707701,0.969942177073949,
 0.7125349016132502,0.9310655800584218,0.9961608186315808,0.5583619674868248,
 0.2750094841352436,0.19673105310564387,0.7368918412991787,0.9806701193457935,
 0.9874471009882193,0.29048259726702585,0.5085861989813352,0.7039562683524583,
 0.23307556498797732,0.8800484368999734,0.12919996284259538,
 0.8721048127681587,0.6902565000251848,0.7366336124414572,0.617198083707173,
 0.8660663766929844,0.304736652982749,0.1446617269820467,0.8780960072127076,
 0.9822002280631014,0.8344865080344795,0.5262385515831283,0.5621259090281585,
 0.5716530979681321,0.6418430673786204,0.8694379057799215,0.4681934863146897,
 0.8638756235811514,0.9837400087143938,0.04607217548307141,0.897768275597642,
 0.48778658131944586,0.3431427342755866,0.5946708398540363,0.7748073341191198,
 0.044850016756486366,0.14609191964338297,0.0760702200501715,
 0.6947965334377271,0.06461841989805095,0.3108156367836796,0.7490996437564335,
 0.22265977322894648,0.36555872628261643,0.4852475272052369,
 0.0760214783079386,0.07145931079994428,0.5438942658479816,0.9843550031344039,
 0.582959412015185,0.9474755497421153,0.23398596830500573,0.29823239314419137,
 0.8707159815820902,0.32592320448621637,0.7722475424742768,
 0.15583192272050306,0.15470665788038263,0.6463766151120278,
 0.5681429032929288,0.12201437844948682,0.5767079326248133,0.602019469086478,
 0.5327085605796472,0.7252671285455523,0.2619382138433657,0.7522327510011391,
 0.646611277955873,0.8773685000681571,0.21495652923673358,0.30141384044706676,
 0.7745217755891927,0.8323483253728086,0.2988844083391399,0.04921141243939775,
 0.49901757190545193,0.7101051277635109,0.5310104449868541,0.7441653018759234,
 0.7516157161833149,0.1451289734847272,0.559649634986815,0.2610959891093706,
 0.160619171984985,0.20772097156547387,0.8453252526811758,0.6115083701020358,
 0.1619775443532745,0.4322163412629805,0.3420875032404731,0.498122437097666,
 0.35219128303090697,0.9145664570821694,0.14799213606448813,
 0.8141279111963425,0.0654430965430628,0.6550545423254467,0.7055634961899558,
 0.3109102514711215,0.1376582345838826,0.5008835866798227,0.5359719051138643,
 0.32418152053570726,0.9648461872236713,0.28472614364195137,
 0.22127972063733747,0.19957074073306647,0.6925395604206438,
 0.6857118750447752,0.2242875181812155,0.10544641320905479,0.1116356800214563,
 0.20691905138104283,0.257096161567575,0.8217967674457121,0.47291780918677573,
 0.8111877396342277,0.5744420149500665,0.39226430621973274,0.9257608417309593,
 0.16263119709341312,0.2869436493548938,0.959634057520216,0.9210992836044717,
 0.7034742426323519,0.39694224287181457,0.11530395964312301,
 0.6740722720007867,0.2174767267737281,0.2354131409819754,
 0.036407719860355625,0.09361070095892188,0.024938657249377827,
 0.1544092305894138,0.14340187984406394,0.9202659657522854,
 0.43531357629735823,0.1977668932895672,0.14601511053250815,
 0.4452825381038491,0.23679206831084432,0.7487172207144578,
 0.39011292182701884,0.618016800725219,0.20865663663683887,0.4056258382366795,
 0.47645583710592,0.4881126675250276,0.4719703454100497,0.5118126599209091,
 0.37414202601921187,0.3925597373032693,0.13324266432150178,
 0.3933840818512577,0.2082676812744868,0.48171400680945,0.8719568074706779,
 0.37317514019134,0.34801512507314714,0.5251123946957038,0.8836490812803622,
 0.059584857022311755,0.3411002962321631,0.21886756799223817,
 0.16971164868362898,0.10629836102913115,0.6169546370097004,
 0.30911466075101013,0.5324490900970064,0.6399109911198378,0.8278089669198893,
 0.3506715117785797,0.8291708003982732,0.09276136712489302,0.8423474915505226,
 0.8880476694966456,0.9131712485641703,0.4375206484816381,0.0469781769386719,
 0.6881290503759132,0.22962329310055385,0.06362187937024477,
 0.20759474115336118,0.930263735731241,0.724774406563504,0.9514498898611726,
 0.22586350535917366,0.4199541653650368,0.49354899508102545,
 0.6228311644188694,0.6956519826816873,0.9935172218407482,0.6541074668634248,
 0.436964764616358,0.27456668896302605,0.9493873665330091,0.7680921811253189,
 0.2590923179192066,0.6473756732709989,0.9408156194434992,0.31479480494079126,
 0.6956702680539744,0.17424108272218053,0.38630058257649846,
 0.18248064965868838,0.35078726352445444,0.948500619917777,0.6957690267350867,
 0.2985707928000707,0.7249573522983483,0.2939557184755687,0.3596301274565399,
 0.7753444711007682,0.9504163471802812,0.9589283866532394,0.7605587302628843,
 0.7590827926743795,0.68135369916899,0.46097302821873876,0.3520568227465376,
 0.22639173487556097,0.6642701992551456,0.04672424612826809,
 0.005084452530415784,0.6607228609918431,0.21816107296055698,
 0.4642328353726788,0.7654113143524073,0.24629108788187248,
 0.10774992617814005,0.1251119917819632,0.9271534783298468,0.504613989304249,
 0.03812899051734353,0.3457887248421616,0.2897037246888944,0.7617389181602852,
 0.44037788277993006,0.3075653760385516,0.5320661789220553,0.2519084637525759,
 0.24665392896902094,0.9511047059361771,0.7589750461030198,0.6230476988405749,
 0.03991874534138773,0.4184312473452283,0.8203777275735077,0.4327616131628944,
 0.8950913752055679,0.9262174757712875,0.04632009840426221,
 0.22941714455376516,0.24688906967232283,0.8900566872098097,
 0.4361449366769815,0.6857174522750644,0.37955280145681025,0.7256388372203155,
 0.4615836420765228,0.33031873069223383,0.763200925153595,0.9786589591112921,
 0.9646937563142931,0.9118653242631932,0.4496151597783242,0.7253262482178882,
 0.22447967342155306,0.23825647041109577,0.5432676543352039,
 0.7612395025503987,0.7206670358689261,0.18219718046900668,
 0.22144110082927182,0.12097594287855107,0.09277502684766392,
 0.3680437399317805,0.848810331961971,0.7786656874864978,0.15133738262269403,
 0.6912581598052011,0.025151384490994744,0.8882668999458491,
 0.8678578720900003,0.33525312215243663,0.6367101438981425,0.2871036610642508,
 0.36421372492606174,0.06745790027186704,0.5751326639884192,
 0.09139954483989277,0.30960786255829786,0.8325808029435301,
 0.9764554192228803,0.8443010934454992,0.1522655000935238,0.1394233174967685,
 0.5575100144902062,0.30817290944547215,0.9634844645431572,0.5002548989193603,
 0.5615202624861475,0.7212537087354496,0.5590249762408497,0.14259845767152812,
 0.9298018353606541,0.8028799851651996,0.6269461424960535,0.9727817823284244,
 0.5486308845640333,0.8199679795890777,0.12912136610674108,0.8547023402218782,
 0.7046103988001946,0.04266556113656239,0.4146786621769326,0.1178412914878626,
 0.02675243797787774,0.11743772083913262,0.535997380506267,0.9482821258864149,
 0.451543161026257,0.8998547313831685,0.3762105901799154,0.18747291015513423,
 0.2761754741771685,0.9984592354380175,0.6922228884137037,0.5986735243732075,
 0.679419289795859,0.7528920279766099,0.021658720643374818,0.7815822185107915,
 0.9298066671227897,0.04168055048885144,0.6939535280004167,0.9652777070883501,
 0.7131852503245613,0.13141999605863341,0.23135644401248778,
 0.10749499811069896,0.42135141149483435,0.6447609907412145,
 0.5538380505269058,0.8773904799552461,0.5003177159049095,0.3111260800632264,
 0.08769157362061897,0.2718214150336573,0.2688988438021145,
 0.26415937639858555,0.4281085155359209,0.12349984123296565,
 0.23423693947069846,0.7662237250643362,0.8026650786016014,0.2986395181362066,
 0.49483911218466425,0.7178766052615625,0.22847596572305695,
 0.6388237356558355,0.7209193427960425,0.3131533128868016,0.38234335247800033,
 0.516587177432478,0.11260718801496228,0.08935994385118295,0.67618118470925,
 0.6735525786966167,0.9441338074677216,0.9210674556665972,0.5581920136720165,
 0.5105840700796257,0.11139543198077229,0.15976111373888846,
 0.2727924472787733,0.1766497911672369,0.05309661157951073,0.712694027090115,
 0.7155176120023841,0.6994078446286028,0.618739125679115,0.9000090259361745,
 0.9554400649570082,0.9198781962926983,0.40482459094676826,
 0.21585682577883336,0.5601431031767141,0.7740105863104783,
 0.39047081534440664,0.95252986545748,0.5535873461474347,0.471550506332873,
 0.22822720244937567,0.5122908137774603,0.7618159237009084,
 0.41911504163790936,0.8544928310151163,0.44160440936234635,
 0.5411423741345353,0.3698617327613579,0.5238743894476967,0.5745897211174517,
 0.14759629923386508,0.8944207185677836,0.0011440332459465852,
 0.012680667460201134,0.5794874512362282,0.32996483962266787,
 0.4647846882174881,0.9996629813867796,0.5299355107277576,0.6674912938421249,
 0.47293299213599616,0.26819358137766736,0.3869994058250764,0.522636183752403,
 0.6965330884850455,0.2824514377722498,0.5384393820647059,0.5837232221558984,
 0.7595721028135825,0.6196803089286305,0.5385745873223433,0.4289548261972591,
 0.9762115495010769,0.05367241240902576,0.14217382622357932,
 0.6622913884594489,0.4468060723688507,0.9890576439316526,0.6459886566530543,
 0.03286055707573865,0.819321357894217,0.292697493933612,0.6715871056035511,
 0.5603526803031459,0.3862808230000314,0.33693680838269635,0.7740989475735978,
 0.7352016617996945,0.3791742050391044,0.35629164182524076,
 0.16302802645715597,0.42386403549638596,0.8729558525010346,
 0.4894616467073585,0.5064097737810869,0.22535552436143902,
 0.07816227045694946,0.16930574623786987,0.5498079223660866,
 0.6593983776709517,0.88448140974293,0.9820043638934665,0.27485076173035716,
 0.18943204267596292,0.773572392438558,0.9848160135533199,0.3990533563405607,
 0.2807796283284806,0.27821818932895737,0.379049657343974,0.788081686472967,
 0.3077495321449576,0.6006624222885298,0.014695158272978737,
 0.044046839563431694,0.5733783348471889,0.3702095658315343,
 0.025012901180682445,0.10873058767014321,0.48802391473709283,
 0.29330869984258945,0.8123235810804619,0.8117000500739331,0.5134687698662845,
 0.33847351736276987,0.758626240156457,0.17551826096572476,
 0.24478228305897776,0.07778570121117534,0.7536827022481901,
 0.4473579401268599,0.9656232531897292,0.08004656527870202,
 0.47613508254731296,0.6714046686264946,0.6394526952346313,0.9437504663623205,
 0.1050351681457754,0.5352965507374513,0.6907216996894862,
 0.015699610904163386,0.8939199332177127,0.4167119505204069,
 0.7993738775107294,0.50892151142389,0.7256744857941189,0.8873676163467248,
 0.7194055974522002,0.43033633195353926,0.19919700878856283,
 0.9402207789237679,0.14160282127463064,0.1000802308845391,0.3348287355363049,
 0.10954342367690184,0.05378867461428871,0.8588168719249989,
 0.9492776062458732,0.035277416118113636,0.7696659121537038,
 0.9555899886696035,0.8954812117085613,0.3481072230258585,0.04015076225807712,
 0.6466192803340776,0.14304227515878065,0.8072971568794579,0.5866085243996552,
 0.8604241701970445,0.28677238487397816,0.09123714767503355,
 0.24975684447010538,0.5303211709533238,0.3648028977249721,0.8756507879137543,
 0.46006816330978284,0.03324474348994966,0.015404556083750909,
 0.07697206998417028,0.764144907640162,0.948491310288746,0.3963748784987927,
 0.9494147164891998,0.6895979274359615,0.9797520983163661,0.9288610823493562,
 0.010241499827485345,0.40682169118163614,0.2640769571126452,
 0.3887229370749633,0.1874859896214549,0.554680294365949,0.3069561332114237,
 0.04996840667631963,0.3901455610590062,0.5258126083812905,0.8793361710122272,
 0.5784083997575075,0.3773048171428832,0.8071889058421535,0.8681339367755646,
 0.7923083781047001,0.7110449957092502,0.3381378501109654,0.4388522118282493,
 0.495347854161876,0.9476175024113462,0.1611343863605097,0.3928918990799505,
 0.27826714786674056,0.5079464852484145,0.18298451622541312,
 0.09019457063661185,0.6047730512273466,0.7213769480171428,0.058119242017317,
 0.7228264510449801,0.2309428496079019,0.45773048285724593,0.0994490466636283,
 0.8728429238980855,0.04463514683866976,0.49442572145653774,
 0.39125264798223036,0.8372850080825209,0.4900777526557589,0.5475308248121076,
 0.2928487896326507,0.9190521063337478,0.5874304058431369,0.472412452749837,
 0.9661035947310181,0.10333716500526247,0.6149909636159574,0.6470271342158775,
 0.9976058392354187,0.2632452125025415,0.9027811572056399,0.31398787182579446,
 0.4349773957033829,0.5170337682162277,0.6599369123511097,0.04428933662287615,
 0.6643021641579155,0.3787221907955062,0.03258381496072027,0.5660342811124967,
 0.38981407307895954,0.20099115684176727,0.9184917971700188,
 0.25025170879957037,0.1828100819084244,0.9068671870894628,0.531603063444014,
 0.22262618211262353,0.7069218379301792,0.7151672987015005,0.4896832628031832,
 0.5984450019626697,0.166458821389734,0.01577735107889211,0.7178670655951005,
 0.03762388802658867,0.44893047591545476,0.45120250157522734,
 0.43463006801154214,0.37009190273320713,0.6493146364373502,
 0.27649780424500325,0.17399303872125138,0.7987922217402367,
 0.044644672071771474,0.8162558067196308,0.8454735056916763,0.558079920191141,
 0.6405750381194774,0.9240752659921824,0.8333635483553291,0.5382953699306514,
 0.8343087488941904,0.29034856853800595,0.6205369839738352,0.0767187803672309,
 0.5712007244407507,0.18654953565627297,0.009296825890506288,
 0.7779474415741487,0.33576131973841095,0.8722496136164325,0.7235921035958434,
 0.3317423723566352,0.901012691885642,0.8537043556722655,0.9712128059752946,
 0.7288149340810229,0.44316739636858005,0.13450434839273706,
 0.5429433828341002,0.12068123891364801,0.9829524941825617,0.6972320028169811,
 0.6140335512122812,0.33186624176026036,0.3111051246316592,0.2960978483624105,
 0.13986605303729382,0.25900773126350085,0.9581495786787676,
 0.17911109283357718,0.9660676912175648,0.20835533419595098,
 0.8519751017106328,0.011583792311302687,0.19957953567563802,
 0.6255798870000477,0.17490281625436832,0.882733784507154,0.7148401518785384,
 0.9628276825680333,0.7364146366025937,0.5996729838085745,0.9509286211907375,
 0.14412773084865516,0.1328925617962612,0.8529048653888738,0.9608830015832768,
 0.2001306443318821,0.900558596960477,0.04049268557538266,0.34906326076172833,
 0.0770720621258445,0.8624063834460881,0.876955938644146,0.7452473156641706,
 0.7629403706839278,0.3923096796005354,0.39711321344592787,
 0.33579134091870366,0.0070163560558365035,0.26870978321637284,
 0.4650703559469772,0.01377882124712193,0.19921894993824807,
 0.45228018990802876,0.058777305356426335,0.3485453205733431,
 0.03953438893866601,0.5691426662119677,0.6504164976175939,0.3593381953782271,
 0.7756433010940449,0.32526046989872326,0.8345821445205001,
 0.20598499399474157,0.09917014816250336,0.6329532338905025,
 0.6476315393868287,0.7513186845232784,0.8064224667475965,0.8239243816163704,
 0.34289601249501267,0.6477497736588292,0.6301584550138069,
 0.18622411139468498,0.7827221237371951,0.20342984615538018,
 0.6134087902180492,0.19190617588128767,0.9941307898634425,0.3172820300912058,
 0.22100914864938914,0.4658394256900664,0.8103431008069423,
 0.42241245226694335,0.5632575554752053,0.4980022015661474,0.8610251355125593,
 0.004255650493336094,0.56083983380344,0.3125429393131073,0.23709900814775609,
 0.9721685463285007,0.4163614741040206,0.8458526432422883,0.619433134601061,
 0.1687610878713699,0.23453840684448435,0.09854735794037506,
 0.6943721380358329,0.7499995858211644,0.5571588856813482,0.5837307609598505,
 0.8034488525114369,0.724986659931582,0.7847862733316002,0.8448712002492803,
 0.9489143895901948,0.6408423161723632,0.4030839044970076,0.5170545270032225,
 0.9196558176188803,0.6363149293739864,0.15788062031140648,0.9069479732711361,
 0.06840527373020866,0.27867274131005104,0.8617305522150209,
 0.4890907575403799,0.4242266377290501,0.6647028654371103,0.8497123050241102,
 0.8159123139411119,0.7991905959244248,0.7549507166457182,0.8408471422063909,
 0.2136986179370185,0.19286767669014326,0.553080201634836,0.02032766674354114,
 0.300384973736296,0.6787525971424315,0.18235293041221068,0.0520160561007067,
 0.8078636016396323,0.7716914907945551,0.8119047941933183,0.6018800854134336,
 0.6015573348661666,0.7126792137795019,0.12346430351558946,0.6809025187291161,
 0.06745218138546372,0.1281760687236717,0.6325633504587592,0.4720893519928464,
 0.6164996616697608,0.6614144559112898,0.021799229575650836,
 0.31853198715233266,0.9800595379400996,0.4625385568970064,0.9855185929234682,
 0.006269679976472653,0.9083922662134326,0.7286328184200334,
 0.20204891731762042,0.41018924449939187,0.45153010752829326,
 0.657571490048394,0.10138363206834988,0.6960629474555068,0.9645784738093428,
 0.19285424495428416,0.197457809973821,0.6352632548499375,0.34166032307588223,
 0.4935828311610241,0.7308023632252255,0.787119547005941,
 0.0038437798175453564,0.5186387518275228,0.5108959484000914,
 0.8697964716651556,0.6585782766601846,0.9460828146015485,0.23060436131644801,
 0.25392943268059165,0.8822030470139158,0.6130084800940442,0.8161365799280965,
 0.656569407135232,0.5046239518365133,0.5297281017686457,0.7393255697503472,
 0.618041592332825,0.6747848096431356,0.4543618688769655,0.7516562340907945,
 0.8101334864877163,0.8785874301523927,0.7639557272521207,0.9795391522363267,
 0.13603941394372387,0.45462792043742906,0.07487315113956905,
 0.7851822053414337,0.3325450373320228,0.22622507539418535,
 0.29849917344880283,0.1368037614148565,0.8198076691517491,
 0.10907800185231797,0.8458049176291639,0.9448099959797129,
 0.11906364226796207,0.40596474473161615,0.3501795824579017,
 0.3691623533437782,0.7611326952897298,0.14462407541028988,0.4763132984847256,
 0.7906047346060386,0.40751920822022747,0.3646842218378423,0.3922596530949931,
 0.1415866917790093,0.3318363842736194,0.503436078550346,0.4834521869037105,
 0.5976121398637189,0.4778058937730052,0.3821776946397675,0.3739299495099848,
 0.964568208116894,0.25105582798865955,0.7943870021047463,0.30353869473694295,
 0.5665722495316459,0.9781044371462213,0.8196933242513004,0.9523191689703416,
 0.5772564716687304,0.9137113387503532,0.988975050577701,0.7681770649252773,
 0.7295189047915912,0.15382382341255696,0.8793433942820905,0.9186587609673259,
 0.3818414869185335,0.15405569975462152,0.5292855479157349,0.9561269585110893,
 0.28348578282748305,0.7502140024009768,0.6096990141626828,0.8114151769697476,
 0.4528132442413243,0.47164453156769426,0.7264489942601035,0.9238127062017379,
 0.9906470589474986,0.906342864370133,0.29125170449218096,0.16105740782683364,
 0.004445441496434777,0.8431362320247516,0.32174430276254096,
 0.45730243273252635,0.8138550591886157,0.25696824540396546,
 0.6543398650767919,0.5254785510354945,0.2717977941315498,0.08268189746976395,
 0.8848729539824008,0.4551092029310161,0.7063781688456754,0.36627711670711094,
 0.20023885382869444,0.28637026316777403,0.7876678180521867,0.850647605516725,
 0.8544676321896976,0.3304713575480297,0.8486801045472883,0.3962475890094824,
 0.6304393437826075,0.9591839774946056,0.5252476451230095,0.31375479046024435,
 0.46603864333431233,0.2811155995848371,0.7641659767281266,0.3550653569957609,
 0.44709392507318435,0.4646366166610435,0.6423505366371631,
 0.24128595659764118,0.7330380441809823,0.23466332196840278,
 0.3813438003985651,0.38291453568982503,0.3194248322781921,
 0.13060478553717947,0.6074893569022589,0.4277393927003308,0.7055839862095447,
 0.44361920147477574,0.6616719116248456,0.551124334647174,0.4742124332784209,
 0.7555360837395388,0.700618650643468,0.5694128992066346,0.625223952350518,
 0.6434979395874881,0.6505955293585604,0.711088007095093,0.14989538786880308,
 0.9327473513208788,0.26991909457571506,0.763418134725617,0.9815506208724544,
 0.9516126044093856,0.8096647112619523,0.02428099193427935,
 0.07685004493716385,0.7055933405221102,0.6355077684052837,
 0.059762303770668046,0.38798788633232617,0.8068146762777735,
 0.9835718198999163,0.263416663255712,0.8967357404167134,0.5581097209546474,
 0.5971971758429899,0.7002270483726293,0.3571333991009914,0.6895984294507688,
 0.3586967917490582,0.247017685416429,0.7882164189252078,0.30889098472510357,
 0.03177350358035658,0.938497359313391,0.30850019868105927,0.5937014301681182,
 0.2700134278733405,0.96607194738377,0.8182631024573566,0.11141547515265826,
 0.8908629393318728,0.0661407882021724,0.7088053765941383,0.5427899385250627,
 0.6733804415853744,0.644490123898242,0.6302227758218506,0.3733991267662987,
 0.5817809601647919,0.8632588038202759,0.0585475356211711,
 0.013557978073824195,0.5339140144384845,0.1113696864649023,
 0.07150852798874596,0.45711441403862996,0.42165969123863833,
 0.9914003964287985,0.20481052362812668,0.9049086856145515,
 0.13030123814207295,0.8926775039911184,0.5739570653126078,0.7158257931111713,
 0.5990406327301996,0.022706077861230578,0.3409057994310267,
 0.9507336457985766,0.18972072165300224,0.0015140594646670946,
 0.11694524824563979,0.4617755039613187,0.528472718909298,0.05939686784440967,
 0.42724317722787897,0.6759366623093295,0.861794990067924,0.12399366076150276,
 0.5681428321120502,0.6606523358944014,0.6230151385340367,0.6194188674536178,
 0.19015066445850493,0.3757706512093333,0.8005866268634343,0.7582460018328567,
 0.9450603757138432,0.7057328882254577,0.21226996744669568,0.7392977738955628,
 0.6426163427729344,0.7429436225063665,0.36371345242586517,0.9265421180092918,
 0.2794860763988214,0.767211128960271,0.3261875054157378,0.14468111862816269,
 0.04307881512540768,0.11446970568959003,0.7145142192453254,
 0.7507233579486626,0.5885765980196612,0.13422442334065188,
 0.04524277205222693,0.5580929975562672,0.8564724892175709,0.4874038947966983,
 0.6694315210939147,0.829350103528398,0.19181561286489512,0.39645024340035884,
 0.2255083396487857,0.2469223080586036,0.2183795211167593,0.36810402557784805,
 0.9030916521392517,0.1735200884330852,0.4268794396883975,0.24558747412401893,
 0.18788743030130517,0.9658277602144008,0.9174998117619289,0.6944909154733535,
 0.9742996687777182,0.6233263599891439,0.23438145763554452,
 0.20278497167253584,0.8359508263623332,0.7173335091440921,
 0.39900309978218895,0.4849527284748001,0.9170671005561644,0.2803333305157767,
 0.895540200121619,0.25780681497311286,0.9270901547696528,0.8598700124673947,
 0.8241504706880025,0.2868752796398746,0.875197421090947,0.8132244115872307,
 0.4791690369223629,0.8019172356675354,0.035277557371541146,
 0.8735545670706509,0.7312571222295261,0.39796912955113384,
 0.31350553889880994,0.41288929846175537,0.33261436964673863,
 0.1884587213746125,0.2903969142695244,0.4087513293894959,0.16396477264093767,
 0.8497203296177075,0.4151327230765103,0.9163154803504687,0.9144482507553502,
 0.5729792855677751,0.21480173786082646,0.010125701422509903,
 0.6560623804207777,0.8790474110084611,0.4103612834018011,0.6785966319204537,
 0.684346506148421,0.8716441426250298,0.2130127541727207,0.7317688045585182,
 0.6580604223059944,0.15755812347011344,0.7032027543144016,0.5136355206365749,
 0.4551617989411435,0.2991081611255081,0.35776057504389214,0.8539529178870231,
 0.7458777491321129,0.6322207775424659,0.5277180402148389,
 0.044567934296695144,0.09119093179556104,0.3954001041861337,
 0.8318597159743668,0.8166269987521256,0.9104268340041104,0.919379225579015,
 0.7209504191759252,0.4277058027208026,0.7467477513682131,0.6300677576161932,
 0.9788235385938531,0.5417470443132028,0.009397510075986992,
 0.8184504091945827,0.38280509402948015,0.8297264291418247,
 0.05541016946462296,0.07748401668864324,0.04747907441776755,
 0.7035308122959566,0.24547471674030985,0.6284783829624775,0.3766349051488588,
 0.2910374134464292,0.16450247802914097,0.3886251709112364,0.5006820838367031,
 0.19109457204894464,0.8399781069532797,0.39734050996767145,
 0.45149335114555655,0.19783308931428245,0.6605935007710746,
 0.04992811735328573,0.5222346428249192,0.816497543339011,0.7330293246854738,
 0.6280477305268333,0.7244306415333143,0.20881744835328075,0.6218078236291096,
 0.7465999625225489,0.35648595683839424,0.7333365929439937,
 0.019902850063174715,0.7209746658207496,0.2883739630835437,
 0.9745260934928663,0.9908584889240991,0.041597545009008474,
 0.07684871754396794,0.899903621945926,0.9369579135481187,0.03577684290087646,
 0.36759484959167177,0.5809912899261277,0.5344364097088625,0.9002020255903151,
 0.12358037661787258,0.4608818594140761,0.06472004045682933,
 0.9897771170273676,0.782148967910695,0.12395866135891442,0.8990295899422492,
 0.4886052194777971,0.4452425590942217,0.47173525726067733,0.5372189048317482,
 0.7543188249566639,0.21231647964349631,0.8956783041891724,0.5876465713483368,
 0.8231610845584566,0.8344411543211809,0.8973103886828768,0.4613858211159412,
 0.5777855305071977,0.7144280851188699,0.9464921919744048,0.5261500595553166,
 0.15400779714082002,0.11102273596777956,0.4729909107317636,
 0.9721498008998434,0.876725198690627,0.13084181351946433,0.6539359395781781,
 0.6486104612051984,0.6830641229421449,0.6572105065842335,0.5001759186802543,
 0.6725071668360472,0.7575431674764495,0.9173381808600014,0.7205160980758772,
 0.9643055969911616,0.5891328074515619,0.2700561807246299,0.8097330635032346,
 0.9033082368260882,0.6849650371656159,0.2806814863014673,0.8225852234024784,
 0.511538777762913,0.41508023409591344,0.037958832006178445,
 0.09567528152818272,0.9418900049528244,0.6332189893247336,
 0.27213311992994327,0.3711129710145624,0.10614647018153645,
 0.9790184081805032,0.7396818636492485,0.0594479927316347,0.9325131311694479,
 0.46355546920072366,0.6891553434705248,0.5153120667969993,0.2413487625809102,
 0.2761726949077947,0.7195237665141769,0.9437587217622126,0.1795098277079935,
 0.7458796962796973,0.1467730945981206,0.3735470569632461,0.7145016006484608,
 0.5830618534282154,0.18696088821355739,0.4713236396842848,0.522584157410245,
 0.1429920748443685,0.21123637304259257,0.2411117810018777,0.8249649531419806,
 0.7156547755482607,0.43585998275834625,0.5354237170977305,0.8109855817661554,
 0.16703932864309035,0.22422151107853072,0.7847148518580004,
 0.23792012077269387,0.8143497199551765,0.5885433380747896,0.4327351159081374,
 0.78597467508788,0.24173149465745158,0.6979619061188577,0.8450658410046613,
 0.19427051404708173,0.0017780651878684939,0.058467489468648104,
 0.22431683763780108,0.14399125050149786,0.6994891887714574,
 0.8563246425037773,0.6161496890779302,0.6332586954203872,0.5006315946858658,
 0.05353736877959947,0.8927804370034991,0.4217628209446108,0.433094075124396,
 0.258766370597178,0.10660358296566763,0.8899726757872143,0.7019491714439328,
 0.14435703337020578,0.09644762011908325,0.09834451674556055,
 0.36705389893814155,0.4421851219235281,0.6480866307532593,0.613501962053367,
 0.1478333526544654,0.6247789195836578,0.02693643061711737,0.9868006610455586,
 0.3773096096435994,0.9089343209188503,0.4227024328841974,0.517559109566283,
 0.3034904725306966,0.7052832918103253,0.7877961500156423,0.9926408692769507,
 0.596550817588135,0.2787204521190235,0.5262114553148577,0.0340819246256201,
 0.758266157131353,0.32608836551277387,0.16577005236834208,0.5774444076125752,
 0.8170522109677549,0.07669298982638661,0.1682496978311696,0.6604318022676877,
 0.6132372897441453,0.6725119624459537,0.2487885562562745,0.07094249061848701,
 0.9037833373933212,0.015711000706919553,0.5461851589928801,
 0.9850327671333531,0.1214722975595306,0.4685571614348123,0.6573584960043943,
 0.8555492416366517,0.4971472933802267,0.6355704771161285,0.7178051276557276,
 0.2382127945208703,0.4699389167453175,0.7616563281750024,0.5677617401014202,
 0.3982818122664119,0.8196052307548085,0.9264827931661992,0.4327968790930006,
 0.5201425114669808,0.9255998717416989,0.2797672096963013,0.7766731973451929,
 0.026118806287157015,0.26846407753824086,0.47424603187892167,
 0.5505192693455792,0.8021973758961276,0.8278244552177698,0.7071311092543741,
 0.5441004175822545,0.21944688818709135,0.7756167216329153,0.7188405075967347,
 0.48633881415257363,0.07789334934569281,0.08453521499525252,
 0.5586604122398205,0.6572238205800625,0.1693551753244089,0.08692162156695216,
 0.5443565417929578,0.2427817260868136,0.47596446627787026,
 0.24482306468190884,0.6871464704088395,0.19949934112846712,
 0.6803300905695893,0.3426864959342051,0.07114308793070856,0.9868444572428481,
 0.22914931979910835,0.3438815075148429,0.8227941802609912,
 0.48203416250733533,0.9148003693842665,0.5913171955027444,0.783517671898065,
 0.8002416652639492,0.8933336170261295,0.11727196340595858,
 0.21843722504610452,0.2512585793768104,0.16908055759415697,
 0.29527738263069137,0.600435485400695,0.2605151963561755,0.41801531647241874,
 0.024732758160898483,0.7849374356502722,0.4538783151820722,
 0.4822252039058219,0.36140140437444734,0.5362927138114262,
 0.23404834311002265,0.9633031232881266,0.636909273434688,0.18351789437192367,
 0.09694113862533438,0.7699512112713048,0.1413347061655047,0.6566150820635295,
 0.4605490667281673,0.6289913397156379,0.38955404262910964,0.6644907972872919,
 0.5722369842228378,0.8053040805874331,0.9497954659878272,0.8377933261323178,
 0.5776000438548045,0.42742190362166366,0.6872423175531878,0.7891476540169567,
 0.5837250897163755,0.7790551030057143,0.026063610569192375,
 0.4102409990424516,0.5400423166655448,0.3033296703829479,0.21876831513684714,
 0.8738567777335426,0.7429547476475091,0.8179830998551074,0.7769628546719461,
 0.05498132387057042,0.8477670033124732,0.08687958077328728,
 0.03536547598425932,0.8588107808972527,0.26933108232917835,
 0.9447024588105452,0.5617285344451737,0.10956526487219642,0.884995633979853,
 0.16430466634269836,0.3669677988160486,0.7870505119262071,0.4414177189605515,
 0.335559140200164,0.9344130078629304,0.004105379045062918,0.2585450268088144,
 0.8434817893720259,0.7851542962216721,0.6440657117534487,0.45016705519812705,
 0.7843180332140803,0.566632561085085,0.6608972239466174,0.8062473977453185,
 0.879819737299936,0.7843338217563771,0.2136447194986667,0.02127812493106218,
 0.7833399645890308,0.1253347427969156,0.14011829544265442,0.5476583781635658,
 0.3458231025454802,0.7483597989467826,0.5341757899360924,0.08601970947431892,
 0.05726716435039314,0.6063532171214897,0.24952806623715118,
 0.12823358200374402,0.536049020386523,0.8873961993203764,0.28632162234560765,
 0.7615441952014579,0.4589959473203581,2.621584588069492e-5,
 0.5646478097023544,0.2654942990977359,0.16956663764776858,
 0.36813127112244537,0.9280106657828995,0.05934671803988345,
 0.1388110427227357,0.12655002229305334,0.6868423497867115,
 0.38847929244067136,0.7660149638273731,0.44319820106824614,
 0.4523422328215312,0.5767944125886785,0.30754872685712087,0.8131646655283126,
 0.6892028896106434,0.1249540823717672,0.4820309492951944,0.9703163407976315,
 0.3578534021432206,0.4612209498260629,0.04263627916584167,
 0.41764743777853963,0.16844570749992915,0.257149026277501,0.6960995935305915,
 0.22880696506033704,0.5301193105737351,0.533574624750001,0.14988067537773564,
 0.02184156059175446,0.45594953443031705,0.845622748146659,0.986713037902628,
 0.8303174509357505,0.06706822568570303,0.2593111007108899,
 0.05691256478437845,0.527069872002426,0.19368624464866735,0.8562880613097489,
 0.8431085919799902,0.2530974097349612,0.9758465014748368,0.7484833966530389,
 0.6829764286115036,0.5849031764232869,0.967984390215644,0.5408601163237647,
 0.6762199042572095,0.1602004843421463,0.30600658906125877,0.9460680325034968,
 0.45852706140384014,0.841353939669909,0.4738731962257178,0.42973248597141966,
 0.8141543501954374,0.2936243254425428,0.8509799762487288,0.04300048633864406,
 0.9601100664809037,0.2293979014998918,0.1385486118728112,0.3003388804947471,
 0.2559029570274822,0.4727167280801301,0.08810470335939935,0.2329427307302978,
 0.17548157057230152,0.7170567921025304,0.2074468094862969,0.6646443688045027,
 0.9929769553290447,0.7673650492227634,0.15310982930941475,
 0.11131595114733239,0.7518203173958525,0.9469614084741571,0.9692979035335376,
 0.23555800699717566,0.766152171571465,0.17906472687182162,0.392371923475358,
 0.1699685461730619,0.5339127789072926,0.8739450216219967,0.5068170019254861,
 0.6417761583604362,0.8127021944175339,0.3440434657912561,0.1164042189465242,
 0.9946610342992318,0.9025173045486761,0.2008832812219632,0.36283446116486673,
 0.7234530857929825,0.11755986564673682,0.357294217013729,0.5655309284881853,
 0.413960520739697,0.1601344016133417,0.5212812942907449,0.1347458716810649,
 0.4413447409838879,0.4096948036401794,0.26537221543113765,0.916490365501237,
 0.6223927449338627,0.9304863256239139,0.15248009468770107,0.7722266088817712,
 0.6783720158637906,0.26944710562454044,0.6636287996944187,
 0.03180601344868583,0.4606639079333644,0.7255491489852022,0.9892847399079525,
 0.2489668056732498,0.28000573836387077,0.9137908424962511,0.8170039291083208,
 0.8257470884905656,0.21348812159060782,0.9059883925118087,0.9849997219860159,
 0.5204344902983598,0.13652297920762513,0.8644403120288359,0.6778299769460928,
 0.04584536546363194,0.8025188778272556,0.563890323003088,0.20675716501969232,
 0.5817455446057208,0.5141852533932627,0.2699653994477814,0.45153314541602607,
 0.1307757330613375,0.653212343135445,0.34836109421010686,0.8192402974534885,
 0.47426408149318766,0.4939133722745589,0.25251663285985504,
 0.4323738607672267,0.5088022558045175,0.26114394083633397,0.9109273382417216,
 0.5754756218807452,0.6310343507935148,0.19017310811225907,
 0.010300182340712838,0.08778102835870583,0.6111423861612796,
 0.4533491679653149,0.9134418654183507,0.8885036142170086,0.8313248608341465,
 0.1341170643848445,0.997073599904843,0.24889531273983723,0.8627494180097173,
 0.4606809219486424,0.6854706539573219,0.6270604449730954,0.2014590494188454,
 0.7313710496415181,0.3953700002839651,0.3940298808546654,0.16712605398685565,
 0.009628637115816852,0.4434337895061927,0.4932900590514876,
 0.7567439086629371,0.3471983296288417,0.11674773061743027,0.695370700106455,
 0.4370452156751974,0.7936850483056981,0.32910432199824535,
 0.09030221795782123,0.4590354377967273,0.7539492393139039,
 0.34070608171472805,0.9332233005659283,0.22037505588471118,
 0.8868092258253935,0.0967589970210746,0.041950089102326005,
 0.6789397902816001,0.33918716875529564,0.40709550019186114,
 0.6195991488150112,0.3666036391595554,0.04894485203740917,0.5716573970741259,
 0.3252655465343738,0.6222421774615888,0.8715242621367816,0.5173530275815973,
 0.014352056859120332,0.647121448567883,0.6481243695513634,0.6118137983917574,
 0.709917430228472,0.3363459047831129,0.018954422372743363,0.9746647126250264,
 0.4408753459380942,0.7953421099768212,0.006401077335087546,
 0.4915737811107732,0.38115805883670006,0.6715252396337195,0.8472505703112223,
 0.7423073131583223,0.15884823967910566,0.7048597279006334,0.4696457663299578,
 0.7950430949396141,0.5590109998144358,0.5507586625542837,0.25490799066899295,
 0.16964950023715364,0.58325043500527,0.26384109919711274,0.07182452986418786,
 0.04215278499193942,0.9156963160349654,0.18848042220840822,
 0.40232177572978367,0.9870663592733329,0.15613160739297627,
 0.8016310902473696,0.6355527321258453,0.27203628695137316,0.1720947711257237,
 0.5334072941528549,0.6585364036511351,0.5637307021118291,0.951152233696526,
 0.9012691915143468,0.35224254614406014,0.9729669094574921,
 0.23846514778405437,0.9044574277979696,0.6382361640047165,0.6593439997968638,
 0.9572405107004252,0.31369878084677705,0.33197023005423765,
 0.6661624368949197,0.7785082626598504,0.8913177753891859,
 0.009231741210167788,0.08350283289975102,0.3279709476191496,
 0.6135571971880942,0.0019937453207138356,0.8414594955047335,
 0.7674957514083788,0.5655603455021168,0.13151719077362733,0.1557458641519044,
 0.9991711776312066,0.6471109509484336,0.2443988583769131,0.755989130735207,
 0.19100851181006362,0.35644223938238306,0.3239516408217209,
 0.4720868976571252,0.8207192369647051,0.9910696466836506,0.4891139798515157,
 0.560186173270564,0.7999223431395007,0.6472088597882864,0.8276816764351567,
 0.9400752662016412,0.8017379886797763,0.005066514178109993,
 0.4857645281862346,0.49499088590884643,0.4160278500064646,0.2179623355392124,
 0.26843679422234334,0.7648114941780936,0.4006846539243484,0.7483338245840908,
 0.1774833905037826,0.01068070499745466,0.19107480733369187,
 0.2087831933052653,0.3846002162253379,0.9980282623672498,0.37584093563742305,
 0.80248216319232,0.4156470609992452,0.24305534634002746,0.5995962596863235,
 0.7216184576090616,0.172011244218089,0.7104228838208125,0.43453006530923244,
 0.979993866930858,0.8980088095548444,0.5734703716173105,0.3063324854353189,
 0.4817421682561127,0.5839937174601986,0.7899427834975152,0.7164671192382649,
 0.9256991401338885,0.7051776847127234,0.9455896254951631,0.2830206956270065,
 0.08819848780820527,0.7093570172847815,0.2184335156205902,
 0.19030654083438714,0.38205359938403105,0.3745394239729398,
 0.09099352040246955,0.8192279673372502,0.9644813576240511,0.9684402199316248,
 0.348583545330334,0.2739862002695841,0.06313773509274057,0.9629740938024358,
 0.632040724417421,0.6749812881676214,0.9530671082609177,0.09508421650968679,
 0.1109888781607955,0.2570943209156612,0.8199913832667316,0.1917739696630889,
 0.469256110764639,0.2767004308615275,0.7159381611638242,0.628727907426696,
 0.70435545152301,0.8374613757786555,0.14806170292378695,0.6570666862846666,
 0.7430197331048332,0.14469587379448567,0.7544423032891269,0.6784073529901702,
 0.3604055582531722,0.15359392622384427,0.49753434136593966,
 0.9036800841780148,0.3260740748062465,0.10909073217379628,0.4607027234203924,
 0.8005802270728363,0.8956696487822836,0.3162007052426231,
 0.014006680735646482,0.32329064847976496,0.8347577070544988,
 0.8696800083141556,0.5478344900650609,0.6891842579139658,0.266411587358644,
 0.3478640628982451,0.5618331851012023,0.3257496939086998,0.5757084481741548,
 0.8719296668547579,0.7146471193247597,0.275406073789616,0.38299449155996657,
 0.31925234962830784,0.3243063508496963,0.3450731019260481,
 0.18874267108337772,0.8081373907673797,0.3055412425510513,0.7525951269032459,
 0.9110355432678456,0.7494954904516672,0.05900426308050866,0.7655369122320185,
 0.3509160360587996,0.21930321898178928,0.22091314704081366,
 0.9798132138701505,0.06069675377830652,0.2582073644208206,0.609748747603614,
 0.05713759372149729,0.26870594915652046,0.2504202671883172,
 0.35331144075548604,0.11534314504366994,0.7640006734191396,
 0.26033889003140165,0.03651045559356614,0.5297914914408509,0.52974513957561,
 0.5600226899069806,0.8327497494130613,0.8670120314190614,0.5548974385367242,
 0.9794354975700587,0.9027520095519246,0.011387051036490625,
 0.1462143174796604,0.1987527874555729,0.9595242453745051,0.877838253059981,
 0.7128376255580495,0.27296576555853047,0.6714396462054961,
 0.19047704661750708,0.21168301879746954,0.7883879648799251,
 0.12018528695982345,0.8565834825487119,0.5348830557955366,0.4358028634320076,
 0.9636079406981082,0.32202091978006786,0.83490891500665,0.7283351506952216,
 0.8012919015784928,0.9143288872741666,0.43442727583487073,0.9356481696781432,
 0.4733024074976251,0.9505928809330133,0.35365787111064306,0.6360639747334554,
 0.7564393699277162,0.44792471199313266,0.7616160085600348,0.5846838038394171,
 0.9724397192081519,0.7795333457535046,0.5729955588391134,0.5486191688820101,
 0.39242319412558035,0.34690303024279623,0.40939802871608866,
 0.14901702232008762,0.96879381536424,0.09118984705888955,0.8413475659214125,
 0.44592517726797254,0.6441231773209557,0.14098216865449698,
 0.5607938246547124,0.6677070175388123,0.8199381096139385,0.3649872239495734,
 0.17939846977604845,0.6102515874113464,0.9973320388576756,0.6218101895928861,
 0.029788589596200454,0.1664341258433979,0.27809177012187525,
 0.5870072556040211,0.9962042204470678,0.43305458819046183,0.4891782528964639,
 0.22461494412890648,0.22113719490025563,0.9195720674109084,
 0.5411085179625948,0.5634803769507857,0.5005010665061057,0.0781910254324879,
 0.09709402195184191,0.28887833236242044,0.7344410215488512,
 0.9094700675946225,0.5952517020638128,0.6723402476672775,0.121510781666202,
 0.6157032834038985,0.3617647197588214,0.03140546044037962,
 0.07014234828301269,0.6766413045339816,0.5335693053064734,0.1951642146460678,
 0.9521242969184982,0.36802923572271773,0.4279082197280384,
 0.49889299241937735,0.5962073949298123,0.32614138973200224,
 0.6439015114239223,0.4396974170998481,0.11743133978745035,
 0.16471150714249683,0.9420366671511989,0.6823696126325333,
 0.09650684671341281,0.9526090763293882,0.5026348921684979,0.9537748683036683,
 0.749704017439882,0.447691213471036,0.1482136675360416,0.8254099469800077,
 0.5209186645631504,0.9727758818430294,0.15397259059269075,0.2111913296305371,
 0.52740141345833,0.7207226064283554,0.8869895726044101,0.919083688253949,
 0.07096362001715217,0.7342638301264748,0.0332483885186976,
 0.15719237100337125,0.2980188817044268,0.040402679269605724,
 0.937786278610869,0.38695725149438354,0.3677150236288823,0.10804423304401911,
 0.6138775488432606,0.5390222851553477,0.18920903946260303,
 0.41027026288876467,0.8906099388236296,0.2474263226137604,
 0.23965446516374556,0.2838787175926587,0.3809363039145266,
 0.43137200628185224,0.8071436529000086,0.8837299867268541,
 0.24555196248268374,0.6177491996310887,0.8781566186591748,
 0.45694550343030516,0.12883176471611668,0.40037988820811055,
 0.4386939121887947,0.11860117565999673,0.6818597430164176,0.9692850035764763,
 0.7326219392863145,0.8438799800073316,0.3806534626562682,
 0.013499643712766352,0.33736204665432,0.4068989395397402,0.8747917022875134,
 0.8819504952011823,0.21861764729073674,0.40024647615294096,
 0.9939495923191183,0.5873972353803893,0.09836767805274926,0.4830028699683372,
 0.2611831448760922,0.3151812598802912,0.7422723442334797,0.4970533012597831,
 0.5027043177141124,0.2682880676046444,0.8128153591111741,0.4275600140012169,
 0.646090198739298,0.4194424522011718,0.45595569521932733,0.25209524330188915,
 0.32349573564140777,0.5891099681970507,0.5657454149505343,0.767694552402688,
 0.8095812534975311,0.6283323201269325,0.004718435838036772,
 0.3438708775373286,0.8806864162330192,0.729460303628327,0.31337732307730193,
 0.20548240923051908,0.23577868759684384,0.1951888954914709,
 0.47108490467709874,0.7386443044070257,0.13685454499643646,
 0.9188254286200248,0.604218253639928,0.3795403362664789,0.9370188286598298,
 0.9432219468844567,0.15052126103382746,0.9647446517700552,0.8755384896300508,
 0.7247004863351034,0.31299983049265323,0.1964121730416145,0.6420958717740574,
 0.38357089300921643,0.9979916362705837,0.7529705535816192,0.6942053308529533,
 0.47318370345957583,0.43238514957968943,0.7951131766034215,
 0.02849729941008572,0.47260342520323884,0.9694486711362366,
 0.6496868206484466,0.3003191937092219,0.34348054294733055,0.9455003715467414,
 0.9527368821709838,0.07790598147088501,0.24546098289104812,
 0.1561690169069978,0.6169050279433083,0.6474701519315922,0.4903633324473452,
 0.0074233141081766485,0.16999344100689018,0.666305647493761,
 0.24183490608548197,0.7450832355151331,0.9803244576068808,0.5694362541889534,
 0.45073743997892957,0.2667088740932424,0.13192475201417309,
 0.45129221382718077,0.16706538038971708,0.5799036021965764,
 0.5898504890855458,0.25559071916332543,0.9546390229235275,
 0.12741047125528548,0.8303655793177533,0.4000304672888578,
 0.012085639329494913,0.4759388949073362,0.3400278797476459,
 0.7097529702964963,0.9787508287980212,0.612767849714755,0.4776562093527488,
 0.49241864655152146,0.6361592523851712,0.797141822860204,0.7117252577635197,
 0.2296012460819541,0.9318451280982873,0.6504113272264995,0.3857641053537991,
 0.7078412658483811,0.4426199144532186,0.7090552933381823,0.8168687802921186,
 0.8931961369814383,0.786958355776451,0.9456799185544491,0.45813775763439524,
 0.21069019329425298,0.6823028745822329,0.8009126469787377,0.649582322950025,
 0.7362766845973961,0.4830678176950205,0.007264121469286344,
 0.18713854315744738,0.3691596674793358,0.43768260194966424,
 0.8264093239803854,0.6766535082959703,0.8938316301993654,
 0.019287912964451093,0.2506597240109171,0.6624248401384968,
 0.40369855195645665,0.7060569832359082,0.6143748762542467,0.4932372137844844,
 0.11149834185195795,0.321333196431155,0.22231776920723056,0.2882834762420715,
 0.30722102402579343,0.03882244774121646,0.8500316008038171,
 0.7891326581842179,0.7741893081648097,0.41598687072764606,0.7458186790195847,
 0.05489550343695215,0.030476672205897226,0.6502973190787521,
 0.9286900981921398,0.28414882305120037,0.5887622307275324,0.8760834954784831,
 0.5384004277561918,0.9980483724282443,0.830891573517922,0.6020113002914288,
 0.9350828430756882,0.9475705556818076,0.4703492643960022,0.60092880460707,
 0.18907672913584583,0.8076663377666754,0.05114880537587718,
 0.20808121038969407,0.35005124251285347,0.8277258949862298,
 0.3257462596005589,0.7505119213884933,0.1821735556728452,0.9841394609856308,
 0.3738944939624791,0.6016361144693838,0.45520435917253677,0.9162230047043697,
 0.15023589020432038,0.061492938716900314,0.09113398312325893,
 0.6395846262824986,0.9359558458418282,0.24967086849521336,0.9206452600040409,
 0.6019935456112524,0.3834526580690466,0.5489305615281376,0.5098488180743939,
 0.83796416311798,0.24794953319007051,0.44470467499652233,0.7885084905566654,
 0.2562254079544187,0.8141239106721951,0.20575254055181857,
 0.17234839295390325,0.23061306067982745,0.01804454424940094,
 0.29657052104642667,0.824302829002809,0.24748938504563767,
 0.37784039923488544,0.9233021071300114,0.1009702322627184,0.338282047384154,
 0.8293260971892015,0.2753269677590817,0.151383188153989,0.33041971885977506,
 0.9535149692374443,0.35624514690706,0.4111373949429048,0.8760395142251058,
 0.2628587550443291,0.17425741669823047,0.6568387509019482,
 0.43839964432106204,0.057028564437514695,0.8303968427871081,
 0.6467194843323799,0.6658476547126766,0.04429901615599796,
 0.13630657907652843,0.6967813748439229,0.9428534960084851,
 0.46038755888092475,0.24969697051869022,0.6906900691734481,
 0.7275535704820618,0.06931221515694619,0.9877063859171403,0.2571903830163953,
 0.8782948537935695,0.5251621519471334,0.9977089470517827,0.08097840873883277,
 0.5252169903591731,0.47134636093426596,0.18271555192943156,
 0.2127690619200039,0.4528309356290008,0.5255603941146549,0.8578582157588315,
 0.606313966413563,0.6771594861547858,0.12351409795628587,0.31999731714377677,
 0.14321393455339304,0.5044019177274388,0.1582654756145343,
 0.38054520318971163,0.4234745223470179,0.42773021654094623,
 0.4680898432979881,0.03224655671650245,0.9756455648255599,0.1447481054112807,
 0.9583201442487082,0.6635227968380408,0.26704460074953096,0.136406176042553,
 0.948824413816898,0.8539906916571286,0.5678456737245046,0.6359008102154107,
 0.7487312298094029,0.42833983471432535,0.3984332226363455,
 0.12913018422509825,0.8404107079789348,0.4290342459483638,0.5107058020593589,
 0.31707259942722255,0.992560194312083,0.028349757095213923,
 0.6302180084267535,0.4903069806307979,0.2567978436918632,0.4776856851465202,
 0.4083654206552364,0.06889806843766844,0.13441764330479866,0.514236160186973,
 0.19126970301425295,0.37951717932850504,0.13800256144605094,
 0.9375367222325227,0.6955855194603527,0.9484881673121329,0.7477275792475032,
 0.9560198453400923,0.37142915382054453,0.9938765027814805,0.5949136513490391,
 0.008767630307395913,0.26455934062304276,0.4823679692192038,
 0.6904342962291867,0.5110876149843298,0.46240835375383627,0.924053073831293,
 0.9769423325539014,0.24048230928450787,0.5727353730110603,0.5062039136094953,
 0.8824185078442518,0.3850708649662937,0.7341693553874468,0.19151221007389596,
 0.6054737203084563,0.6168461270931042,0.6284538269160926,0.20514931383085244,
 0.08521205577164215,0.7247064277064873,0.7765157049226857,
 0.43596478802877114,0.6040311795203834,0.20822796374999486,
 0.14977327315206546,0.49443244991134716,0.7083047926773445,
 0.9404703800655664,0.6688457413019544,0.944106464528201,0.39726489878357396,
 0.2974729188063254,0.06437388421200319,0.9746928487932768,0.7328647863795795,
 0.201537087329424,0.5779228817301935,0.7301912090723269,0.5365586643473469,
 0.24595641277405045,0.19673224666533584,0.4843715223980323,
 0.19148263464755233,0.029044120435736898,0.9737713705541413,
 0.6829127866411544,0.2517609268515101,0.2783742342799297,0.4079193961840776,
 0.5207260451672837,0.826291420967639,0.35714171501196723,0.6505631805634254,
 0.42882407270011436,0.5356556119875239,0.9058857566947318,0.195595708745106,
 0.11761635315093333,0.4636450895280171,0.19352174595898775,0.864535818366526,
 0.5864285900893916,0.7504887530292227,0.14228820089675986,0.8134606214102711,
 0.7833815527353561,0.2886036921084497,0.47532131033934444,0.5115387799238451,
 0.7418677640947114,0.3661862100494737,0.8413072620094675,
 0.0021897680022601174,0.5985319071358844,0.5177915296555856,
 0.5967255356100307,0.05290463203846141,0.06866172515569768,0.417547268219326,
 0.05382896497662759,0.5709265379816159,0.8082126094563261,0.876642468557491,
 0.9088110480280623,0.814666095560487,0.7867419541749695,0.34575582101124025,
 0.863224060432973,0.8634587531906914,0.4278947509446397,0.4089901902824178,
 0.0033503623315244146,0.2041381915562528,0.2678727876871255,
 0.7856945986300871,0.9679996292335296,0.19030655441396338,0.6700065259219015,
 0.9353038908033076,0.5984852754657787,0.46731158746902846,0.3918646388194176,
 0.9398619979179889,0.5984150217426432,0.31259130051405604,0.9612180160168124,
 0.2306439545506015,0.6880934068066786,0.8467229233446552,0.22918213270168053,
 0.40966500619017077,0.6675070168709629,0.788933054837844,0.23532255751050268,
 0.23733087290000643,0.048411937324124854,0.7975291522475468,
 0.05192143289107154,0.7291413368153236,0.8036099121573351,
 0.23140493543636342,0.6887362206711455,0.8544130732218262,0.9056497110145403,
 0.2105904180535626,0.9879985826069067,0.1151657807489409,0.09917277149603954,
 0.713794939632916,0.6330660145219543,0.9660542278667448,0.5552297100278849,
 0.4708942818470807,0.5283611889442958,0.8424970871247022,0.21117979070038595,
 0.39614364936307633,0.9519035520248085,0.565165295712577,0.3736585380417059,
 0.19399476103759528,0.3879725194443644,0.5171529106927613,0.7695479319487976,
 0.22785973970603224,0.20088475526727168,0.14719066574748485,
 0.9737294014780697,0.4470691630491668,0.4892704469782103,0.48431693274195087,
 0.5660765808263934,0.7491049008263908,0.15569945142293484,
 0.09913750315406211,0.162210766972522,0.14331407594725876,
 0.25994271895536203,0.921420560545019,0.7169024492367403,0.3493669597584079,
 0.023814182166106646,0.6946910296597247,0.3082341649814003,
 0.21607673148770856,0.4028558169379861,0.2730978554783988,0.8095934014453415,
 0.641196244474896,0.45431546080147767,0.1784259264171837,0.05168515731683032,
 0.07668368986300089,0.48906572136101023,0.006256676713090159,
 0.016447077126966136,0.11215335271528504,0.6348839934446746,
 0.6779027340666052,0.47016484137178904,0.6854247588734621,0.8733922208528696,
 0.8993164484581284,0.13451567157680888,0.965162661747432,0.20301882619856393,
 0.07061172616835887,0.34038802943693514,0.7794773148933059,
 0.7679671131715397,0.5262416827443689,0.41658582249928533,
 0.38873794073301127,0.4473760946203479,0.5311319889901303,0.464959643197833,
 0.4159263030663646,0.6027523782194797,0.9106150604270302,0.40787380684826646,
 0.9084875692115019,0.45974480058862854,0.05736820679474919,
 0.5300101849461801,0.849283669128675,0.5541146564759583,0.40935574994816337,
 0.46928430415995825,0.3337339722575473,0.6929258013787762,0.7622587251442062,
 0.5818119575081948,0.6111611055976899,0.22247763143255816,0.6192059169697812,
 0.4332148207048425,0.6014990740010229,0.7854378994353861,0.6328073408982158,
 0.712396511398849,0.5233964404596533,0.480564928628165,0.41445477505502415,
 0.7311264497946356,0.9458398466890976,0.8389722704692639,0.13739898655206217,
 0.10859811291934318,0.6291681018135896,0.7266336955841383,
 0.08341099523679385,0.8284246800094075,0.3849520326405437,
 0.07112409155727906,0.686304650428027,0.6837387006938521,0.9517997102377784,
 0.497189900095275,0.25477881887610965,0.5940387558524948,0.07391297824920162,
 0.4914159428549927,0.8745991323081154,0.1319183677009016,0.8114265188874094,
 0.6538383377859722,0.6549619175345672,0.24183858747365394,
 0.23655942618788162,0.1976852104738267,0.7485794560796052,
 0.33898448453653396,0.820307407178871,0.32510622102619113,
 0.01225065122467095,0.8343424506316602,0.642024838053902,0.36393782503553374,
 0.6988484923466247,0.6219622405655378,0.7658620678265345,0.7410908779901966,
 0.3144962096404238,0.48704143777098696,0.7670519714431324,0.7701329822032261,
 0.14549874451046307,0.4625026287789308,0.61620519927692,0.5964584796853647,
 0.7684574063278804,0.04490881300568339,0.1423086732585913,0.8045826212131608,
 0.13219961479033893,0.0979695815549535,0.36627356867245453,
 0.3249228167576512,0.9024507332232697,0.3890474658240236,0.811627680760903,
 0.14743484648532168,0.43897821595182573,0.06824741483505536,
 0.3726633484415419,0.9896947423375769,0.7887809973619064,0.6821540228511728,
 0.7550255314954445,0.4068425600390101,0.913092733160724,0.6020727633697075,
 0.01502016474094725,0.850173721275228,0.5747254663571613,0.24628311990561425,
 0.5699391686763557,0.10899439591167204,0.9864618916057237,0.9691387660583305,
 0.8611151698449695,0.9919208679967564,0.9744333491361801,0.06410893434386122,
 0.882078352311952,0.4554084106443904,0.1586070088071283,0.387275149067412,
 0.39018890159620745,0.837324941059703,0.8745255225781028,0.04702221416198016,
 0.2270427325869504,0.580672799889208,0.07455622521623972,0.6651361088890316,
 0.4862608108598919,0.8435131128609724,0.6105105745893229,0.9572378471533823,
 0.22619586910627199,0.3780899707533206,0.7334098369889197,0.820866553025724,
 0.4144210761844902,0.9755124875474981,0.2768272066152866,0.6970529549471994,
 0.9839953793823116,0.36759678179450084,0.9029818410047129,
 0.28764297742633804,0.7070727197194536,0.2118571902171288,0.8724430437348287,
 0.9022077993016769,0.9815690882983827,0.8397959273172793,0.11378224201667297,
 0.7950005320181972,0.06325675213559767,0.9832569732623133,
 0.27048250738323576,0.6906614392691509,0.6987219645119305,
 0.07137934504782673,0.15179925011848971,0.4402273501993559,
 0.1421445703411166,0.6016528943352939,0.3880705704492028,0.7484371231098775,
 0.5262160335710315,0.8675027763902555,0.30321378310405045,0.9598119969506692,
 0.982279430069528,0.1684452135177873,0.9468925983986917,0.22183047413523127,
 0.24217904633625342,0.8935793751409472,0.09997039248694772,
 0.33206470375853425,0.4736936860185552,0.11465587274107591,
 0.07591410399821341,0.09984163408788282,0.755228544938374,0.7762050081612737,
 0.2560837576362811,0.8728078754016616,0.971293250033622,0.1718097645750689,
 0.5294154649944297,0.5545130347997348,0.3251318427753723,
 0.050939423821324104,0.5044401535454167,0.31342119050024886,
 0.7803317895019324,0.056425929093302574,0.2129636103512076,
 0.6258406660035086,0.6211737049307723,0.4070825409983625,0.5522035718925739,
 0.49344922285196136,0.20451051620897043,0.4072777792239094,
 0.7998655206077296,0.6300281002648908,0.1396223242844591,0.6833961930683732,
 0.040023288465511064,0.011263490239523555,0.4939875384428474,
 0.029070253890299158,0.5678610154450573,0.2875712788775511,
 0.27545693399341675,0.2301441438759002,0.5570876329877946,0.9014212153329052,
 0.7577239506806018,0.2318071023647147,0.17478214880468823,0.521211788652506,
 0.7476087907308449,0.24971714085656083,0.6393613881789416,0.8146398293785122,
 0.9249875249697365,0.5878844220207118,0.5929414718572729,0.5791335773105131,
 0.6558923870570801,0.965798443698529,0.4325948679326431,0.4491238852116828,
 0.7091689875095605,0.629437318425065,0.8579901210691155,0.1913387134944412,
 0.3805135171831797,0.021424749900509577,0.17647360940334433,
 0.6809033673395131,0.4995337440820411,0.9320503057673198,0.7678168017526296,
 0.7099439966619211,0.06769324541110444,0.3924623110067291,0.6618006382074971,
 0.5556881308015826,0.6902498861834697,0.1085731085221191,0.8437851385300136,
 0.7309528486496903,0.9202377561443453,0.055465431485226646,
 0.004756902373142857,0.7755926606402099,0.9241999088631893,
 0.6237499793751002,0.0761750867614911,0.9820836295602725,0.9724419691153294,
 0.5357224995251774,0.9553148305604111,0.42019193202565486,
 0.09170867544581296,0.7622511631626789,0.516173231606268,0.28399236452866905,
 0.9241217450047776,0.018489333837630184,0.3070209035646029,
 0.8351891550275168,0.2765249511482982,0.4566707968869643,0.5408371751327458,
 0.4668535359739808,0.8643312515603969,0.6165622974983136,0.8034994254714328,
 0.4419344401424826,0.10541552094524698,0.7092461813234818,0.5143093606708297,
 0.2649496356145188,0.01776888655911779,0.21119970951546185,
 0.026787435760090128,0.09478988476443995,0.8053360946451674,
 0.26577788043299977,0.7008256946879006,0.7507567273056599,0.2804978833442149,
 0.590081463066774,0.0131743358969445,0.16917116616810368,0.16376459977498803,
 0.6368393917295826,0.04839319259297059,0.9323252411996686,
 0.04465640327639486,0.01557981125302732,0.7805037117229823,
 0.8680074476192863,0.36983804056527414,0.6518752275863174,
 0.16264139337701433,0.597890488397619,0.08907303033999625,0.8188821774578513,
 0.5719411830806665,0.22526632687492987,0.3351907197570596,
 0.008889904053894337,0.9294864024805133,0.1526155624440676,
 0.24231692208184763,0.9142599092787737,0.010838723317116772,
 0.21421074326722744,0.3024647307910293,0.30069466629070285,
 0.5444201212502203,0.04029018269723261,0.6192787727098464,0.7569022714454476,
 0.37334828209667825,0.16389553670092538,0.6076855523391633,
 0.9532081701642426,0.6450211440209448,0.06289780994597527,0.5146174048802926,
 0.2090083594564459,0.09858528697330593,0.5195322601422168,0.7363476492548012,
 0.6182495634059477,0.3751519779077066,0.18158243189338452,0.5605644977068532,
 0.5932001413451632,0.7494941016660795,0.4962245139360436,0.39011768321354745,
 0.6321079709990682,0.4928816601250242,0.6242825156831832,0.29383577710079,
 0.46198477625347045,0.34653360752203666,0.7078421951069664,
 0.19441930963886145,0.7532146943089314,0.41939888099925016,
 0.025780867032756838,0.5520211780037183,0.026005411025651237,
 0.9994647578002496,0.4555780095574392,0.9644296709400811,0.96383096159964,
 0.3869497837561775,0.8797363794618505,0.3642543135009585,0.6662561088178716,
 0.07785368121474878,0.40535818534644497,0.1637460012055868,
 0.3737457704446461,0.8008799498372835,0.6948124921313013,0.03192508746272171,
 0.9969319137249378,0.5384764899479897,0.3932830680411604,0.40384840115506915,
 0.9774719422967276,0.5819379963071422,0.41842249637728846,0.603057286873592,
 0.4258318978682788,0.6035232507902504,0.5274236154667824,0.11066563267603868,
 0.5670236023404888,0.9507713158005238,0.357892968173266,0.8653575745263894,
 0.7885202587034952,0.9897562932015367,0.6208119522027076,0.1376382669113866,
 0.5322295898941676,0.3936727005147358,0.5419242011308756,0.07192058237648391,
 0.5340524428314188,0.6741455614798688,0.006560298244542184,
 0.23326390608216785,0.17943811748099248,0.10634090884390113,
 0.7905606306352251,0.24806946722911194,0.11976749125768515,
 0.2645399477446285,0.5329208430107599,0.749271331224419,0.16619292035146427,
 0.9487818353415018,0.9787583815810156,0.3199334317101765,0.11745902945247078,
 0.6874652495262729,0.6538834249910375,0.8556402504047019,0.09348205123530473,
 0.6561567983357214,0.509702819635881,0.6303877015340202,0.04559392951445651,
 0.23095864026059776,0.7085987349310165,0.6956043021064929,0.2737211809157343,
 0.4607899619057201,0.8240746552469475,0.517225519682252,0.6341303259341081,
 0.23496564087911476,0.780273135744035,0.9771671229815105,0.10206140409979292,
 0.580697614900598,0.6845918274888894,0.4647031537731191,0.1476740707986921,
 0.9009518956060484,0.8366497938112892,0.8889147255568485,0.7262176155326143,
 0.5567992481697147,0.6986450561745728,0.31913299209235924,
 0.06732336491607926,0.19453899737518054,0.0808840130852948,
 0.14641911257787044,0.60177639034188,0.6115201546161635,0.6133868743610538,
 0.181134476222518,0.42085651172489924,0.7031474305308474,0.12354253400813175,
 0.0923006379850776,0.6616239396057946,0.4926753445803451,0.26200921670660504,
 0.19115652506693914,0.12864372693729342,0.5346133010139684,
 0.08694132432691604,0.08043079293821287,0.8015051760209628,
 0.2227326702433189,0.7899036048312094,0.718164219684034,0.9039997961413093,
 0.9143503867588154,0.9214576579920281,0.8600632003851927,0.9720497299805183,
 0.8008745562562405,0.35175774122182757,0.733529423308034,0.07656962141157264,
 0.6321406077088917,0.7944699426193806,0.4800041235594519,0.24191769127036378,
 0.7449287710513983,0.88598109532948,0.14072563644995573,0.03802853715997123,
 0.8437896785122165,0.08973900834675619,0.25473073761304477,
 0.7555220647855754,0.9113985109845408,0.39629897913800327,
 0.32571191235439556,0.65297499867707,0.17230570633011721,0.7293575228621902,
 0.5938289775813832,0.7310189226472105,0.4581786804825564,0.6748337794412708,
 0.2616396318670622,0.24680750209547553,0.3297281276797104,0.9847798386787098,
 0.706132113516341,0.32729991386802526,0.7837722781234918,0.8665964221164955,
 0.364257387531005,0.9437873586941161,0.7116836338963685,0.006793697264648291,
 0.09000322794677496,0.6341695209783731,0.14628740785861738,
 0.4988150467355892,0.4937331125230404,0.7745855279196591,0.8303251843902032,
 0.8527727975188764,0.7748559524837594,0.4051473048639871,0.94874316486331,
 0.7719418584257532,0.24789848893889388,0.8711004070364353,0.3870468073070723,
 0.6377062115316059,0.4309804916697859,0.8320933660650677,0.4305063896591398,
 0.88934444274273,0.3192379317790053,0.4850070997666973,0.10364051511413819,
 0.848003841526394,0.28507906977698216,0.8344664775841206,
 0.0012536671772865882,0.34041790691418006,0.14341455546745652,
 0.05772294851567483,0.9143893065921437,0.935559735683024,0.09567519550618298,
 0.5066410676334426,0.013352108516883399,0.37490303744432885,
 0.16830118303658181,0.7757296298931573,0.8095924361044516,
 0.22540679544621367,0.6320451704453877,0.871173238585562,0.13607138151424858,
 0.16230913861786933,0.17831048348364176,0.4695320676195478,
 0.8334801056822724,0.9132056625679906,0.3179143653605949,0.37117569939780604,
 0.9837810779930255,0.5297833729717546,0.4576239057494591,0.08886977289051157,
 0.38437345154261127,0.8456906039823815,0.7324164979203266,0.4369954291361955,
 0.7118723156492509,0.9794844195232885,0.2720189935594902,0.8524559556049766,
 0.4314603242307583,0.257319274206262,0.49753215801548845,0.8701355363102843,
 0.10871449718858583,0.2637900272305601,0.0843907716398743,0.6678802992083472,
 0.7638199378864976,0.6443293238551053,0.3976064806057241,0.35059867338444906,
 0.26259104559402724,0.4692245005788548,0.6437127824216213,0.7284920591772487,
 0.4387427350300275,0.45261351336422184,0.15466221577828487,
 0.24710576679441543,0.29233113769283736,0.09946886629608453,
 0.4914050869040688,0.05683999130343542,0.32427533563363276,
 0.6412892760507132,0.6314827861385952,0.2935874297633668,0.6186478864514638,
 0.665852022166504,0.3568254013540876,0.844674573683227,0.897450462939896,
 0.26352837592708367,0.23262854508152575,0.30930448286110956,
 0.5897478183928753,0.1971955666236488,0.035684198914688214,
 0.3161036788009757,0.9698677282891595,0.49869176986795016,0.8454557791775787,
 0.4182165395709806,0.340718765472938,0.7559511780867417,0.16263383597063297,
 0.3279717508518146,0.46196980424551415,0.7404669976468766,0.8541573151452613,
 0.23835140047395942,0.3561157830963265,0.8731029212888379,0.6165947355453366,
 0.9568321152389258,0.7883567405456651,0.37046130545475286,0.8551387464472935,
 0.8673259258847565,0.47801421906749897,0.6743561133597421,0.1678315722850039,
 0.4666035169823336,0.8136497375186151,0.5208032234025516,0.6020234115492406,
 0.2912889736120322,0.05203043301181087,0.11378217948503433,0.984325738426709,
 0.9163526590715736,0.9456122316511075,0.9677721265121648,0.8679488382241872,
 0.20898425592788183,0.9235151796329689,0.3000167010706212,
 0.09371801138682523,0.46096895496830403,0.15932540397190242,
 0.7994660934361192,0.8549964877916576,0.5104414498387836,0.40726468816984673,
 0.052049081402596564,0.3394419232718764,0.051567325743792636,
 0.0233931464008017,0.12792911606837676,0.9880740411103673,0.3783755469055209,
 0.1263150595316921,0.1604824440023862,0.7535116191554291,0.22177925937186538,
 0.5707845335430746,0.6868465433345518,0.5105541233439195,0.3904441545601761,
 0.45859572381106317,0.10769979144432473,0.1371849958628244,
 0.9079323440633454,0.18627355115664657,0.5571103100150568,
 0.43361988961730924,0.02036028391663458,0.29973485676061884,
 0.31888070415371383,0.3845443173698879,0.26651121457526794,
 0.5349674078591361,0.6203189276180441,0.28145116110723756,
 0.15877821123023983,0.04650010966749529,0.8024537189859836,
 0.8808280191929969,0.04912296150642592,0.7348447893641707,
 0.0012213322781954306,0.6137829654262129,0.2347817698609902,
 0.1832694671511349,0.10448528617600328,0.06276384683795122,
 0.2823344997868745,0.2501872603193165,0.41202967686692404,0.4831393740671044,
 0.08148911032933881,0.7905731731392291,0.3922176962584889,0.9296909910588478,
 0.3733629933118421,0.7356588023448754,0.39133382051353993,0.9993197914287267,
 0.5866023304937144,0.13441153262001448,0.826352602498198,0.18471144261642658,
 0.674141410927252,0.4141768134507974,0.20888716690445475,0.4136762266640954,
 0.23896255619042184,0.07184743247440573,0.12866768639649584,
 0.061289644809865096,0.37040910186793685,0.07797560829648753,
 0.21726421161963283,0.19298723616342572,0.2323381235445321,
 0.3442407202310933,0.7953260954728464,0.684954213760276,0.6830012865661952,
 0.6167066988529923,0.9165052350913544,0.5941978384088692,0.2983208479711439,
 0.5141510322068216,0.3595154801386461,0.22793187443827567,0.6442591754774463,
 0.031824807481346085,0.17770768925399905,0.07596793235167631,
 0.6025096443091743,0.9155592406369454,0.8392920155464154,0.20829492275837458,
 0.3254110106087791,0.818965916164861,0.09809824534625489,0.7570197898630825,
 0.32749009412742347,0.0842798796314399,0.723642770767609,0.8395556637431079,
 0.5381203240063533,0.019699702576886913,0.5338518931202472,
 0.4450902879423754,0.16546665011293826,0.036141233329622935,
 0.8125806205721798,0.4876203705352964,0.15962413298953582,
 0.43488323093619385,0.740888030724675,0.4713837524430331,0.03337838034176899,
 0.8003614110219935,0.9120252585603742,0.8638353203692952,0.4857308182037784,
 0.8982903884866565,0.49060330470106195,0.44949974035103113,
 0.8806784894131864,0.538081109574544,0.23906654282953166,0.5666262406519902,
 0.4283339060402822,0.6828330975022113,0.3378718406094998,0.5110108037637119,
 0.41785440397209417,0.7299138521649107,0.6930669007352094,
 0.46327645006606344,0.7062476397563737,0.6294653627909698,
 0.007297400089805173,0.08364926092124658,0.4701956137987041,
 0.8218192756573357,0.20518903418186252,0.5673463557783487,0.4072709868605614,
 0.04539424736202402,0.7806655008163299,0.7086680698448276,0.9601519901661986,
 0.09809247243700314,0.9615605703442546,0.20047074328650316,
 0.6166183412767672,0.5859917249688319,0.6162253810259497,0.31175704590324294,
 0.7435970672613499,0.028249066230622955,0.003298134372574646,
 0.7294510158657915,0.0767882747711669,0.030796375920040697,
 0.47345787948279516,0.23514257087900337,0.05711573183762508,
 0.9434210065097596,0.456512020862186,0.631097727986419,0.20314742006067854,
 0.46254192027930874,0.6271853793863578,0.502509111198232,0.3386453743897514,
 0.025511094644182775,0.47975818286437943,0.17746235230238738,
 0.9378458116846765,0.8410257238764124,0.5682238405485486,0.9636503747568929,
 0.5505237445593589,0.5048086373751915,0.44439526205748026,0.6210605998318486,
 0.15398645754068363,0.8401599668913684,0.3925797091028562,
 0.43906425800422266,0.4405407452103204,0.5447799690301239,0.3726873214703381,
 0.3298402043520031,0.6770751210904904,0.42213476109331105,0.4522221291726758,
 0.6146543076887575,0.4605668038058106,0.6172693605930711,0.8562627949047174,
 0.6145524833432966,0.788510900611387,0.4126582604609559,0.5927001271554091,
 0.8044973101250205,0.8352568532700885,0.7301989388797664,0.33890703724244287,
 0.7650683308707459,0.6445206058120944,0.09972363787165961,
 0.22571769248365348,0.22150024041055405,0.12103566133685251,
 0.8070365100736354,0.19810059996581109,0.3767568833527498,
 0.24378482728196482,0.04920746048927538,0.19046317239168953,
 0.018804662584785636,0.5197352916955336,0.23045828957402836,
 0.26959589485740365,0.3735727376125241,0.7855886985296079,0.5332142420191021,
 0.20668617956193636,0.8421927946718589,0.19981452053502446,
 0.4284257316652732,0.06164812861032265,0.5393772449712397,
 0.008579189760051675,0.2407068802643776,0.05212658159723749,
 0.2703504792156083,0.11130982462368211,0.35321635754785097,
 0.15057843026273765,0.7009122991822003,0.8359530197193606,0.937871891772163,
 0.08284184406715278,0.5151340442273957,0.5502324618473575,
 0.07374257551932129,0.18160099926774753,0.929426835574913,0.3808254550162262,
 0.8060420478809565,0.8406114174919954,0.32102474814095383,
 0.37188696109915276,0.3527130288618031,0.6994582773885994,0.5529141065290808,
 0.7402434146398225,0.5547361183504634,0.8450602210158946,0.7937618507358254,
 0.8574856562850397,0.41827329762088317,0.11277088059237783,
 0.10350119041395378,0.15822054963003263,0.5520274687821827,
 0.7255241368772474,0.5021140908494878,0.6026564331738996,0.5629372257280153,
 0.4121295246066866,0.702356667899714,0.9828945989076501,0.1026401373323872,
 0.3645472805430501,0.2500809955995871,0.551206636575406,0.46317551993233985,
 0.49763777885125526,0.9158887377788052,0.6886738193413362,0.8497447088728686,
 0.9537383348754312,0.5254326959571813,0.8465555306110248,0.7539129703224269,
 0.0617411717636126,0.7577559060674406,0.37989677563458313,
 0.16149208213889954,0.2887910149504752,0.3791759613967498,0.3908745152084879,
 0.9765506344646434,0.10260049206889166,0.16163805048614666,0.411166089814033,
 0.8453454318184861,0.017562105883379586,0.05494418851409155,
 0.20459388615641305,0.12261923524674256,0.3913698045782724,
 0.03656777876844841,0.04104749687302944,0.6316092042731288,
 0.1914249186095347,0.36712669110277873,0.7548655839614339,0.294687353408154,
 0.8697191324982007,0.3233913948435385,0.25418538522715395,
 0.39051228523590154,0.0977333069754035,0.9060480993370951,0.4410265262642328,
 0.5065122739460689,0.04251022239984048,0.42604498984006156,
 0.9661658298957092,0.23587375511486974,0.12838754681066344,
 0.02953588204580826,0.9888634528556155,9.134311197210865e-4,
 0.43688841265967016,0.7944133234337343,0.20934525952245964,
 0.6238979208045251,0.6200381190413329,0.6378592993724598,0.30020735690490985,
 0.7204882169379158,0.6038274779391528,0.9148208154064875,0.15988197513612246,
 0.7770899873325577,0.28503060503931976,0.004606037293892973,
 0.8927781861593207,0.009873380374719076,0.6425540084253332,
 0.5823799346823232,0.0028738082484848215,0.6716749035311577,
 0.09168643040289082,0.864198231655982,0.20608057663543522,0.6840701032951655,
 0.20902345785610998,0.30203273749185733,0.9689607522201955,
 0.09045831295737994,0.7356838168651954,0.11447205617138223,
 0.20572088094711594,0.04996719973454444,0.009145423005307185,
 0.3695076291994207,0.21227699428296676,0.7996543771815967,0.5677719143982539,
 0.6204569391801013,0.7786448203126735,0.1046420703955252,0.9787015129004291,
 0.0033686005446339484,0.07276233669997156,0.562579459224976,
 0.878164179132509,0.6863657290087708,0.8721956603748064,0.12222302642575056,
 0.33644913385348685,0.7060458201946112,0.9069687045177348,
 0.27183872856148106,0.2913655971025074,0.5452620494161743,0.8603416195899742,
 0.4894330643659949,0.6085536403267819,0.36988817853883926,0.675499542529319,
 0.7497229852492012,0.8329207506056479,0.5530519444668759,0.9531954692894957,
 0.3047716836404246,0.620995819896144,0.24098850845432207,0.49060010553537337,
 0.9743652852184158,0.6373404272001354,0.29641307144715856,0.5792634864995729,
 0.05010263416899796,0.7982151265767714,0.2707719553149114,0.3747575257918936,
 0.09540918156232636,0.058993997842385126,0.7788396952716047,
 0.9360405368958693,0.7599170134912278,0.9591302189026445,0.73408997349538,
 0.9272006120599304,0.8137414032491205,0.38307965042763736,0.2023688906003286,
 0.5060337935531302,0.3641289891650371,0.5222682658525202,0.893154309191367,
 0.689249592190676,0.9931525200853406,0.4207175087791798,0.8785296509086201,
 0.051095983109384124,0.3070440475724563,0.6067076738564701,
 0.04869767397519886,0.14523230853351266,0.17378689056513197,
 0.1948574480274401,0.3405184966015007,0.5106683882487699,0.4231424302901772,
 0.16620336591284768,0.4773859831818329,0.4320426184097722,0.5639044802368048,
 0.25147425898628895,0.011605488963473687,0.7273244613342089,
 0.4104891706656797,0.4970995991127687,0.2827132174986968,0.33029817837069053,
 0.6632641038180045,0.4873706121354333,0.7806253190432361,0.9560713497460966,
 0.7223270598249064,0.6550823689729848,0.17287096492690723,0.7312932420512043,
 0.7162647273977779,0.8556522934677904,0.5621534101872767,0.46353366720675926,
 0.4197522429870558,0.40491680537594965,0.607720504966706,0.6868046283273671,
 0.26992246154951094,0.42013546216537156,0.6893967360066597,
 0.9016563293744679,0.04880639118866592,0.08182798203650454,
 0.9946182564767071,0.8655734575599595,0.19482843291010088,0.7434324244462998,
 0.20013522272605144,0.4485045226056368,0.2402013193728878,
 0.21056370994741447,0.5334487545024597,0.5310975235219244,0.9092242493552464,
 0.3248176382740532,0.8664531792260249,0.5640650404342337,0.7324109799501726,
 0.25500042523108224,0.8661757769235084,0.04839881554445169,
 0.9998039351577787,0.9440408282693817,0.6471876905308933,0.1411563469230731,
 0.1581739884043588,0.6970897708423606,0.7259409865907974,0.9206536810659597,
 0.6612580711739295,0.07058800341524973,0.09007493830866331,
 0.5658762621290445,0.7507705342730876,0.05817312597684654,
 0.32229240865309305,0.4613981822170894,0.5436455506261181,0.5859792870911351,
 0.47428533855024757,0.33027531573646607,0.44130707666250435,
 0.857414200916311,0.3056934052500492,0.9189126303274244,0.06915425134935316,
 0.3210647774049419,0.92257036250333,0.052314662011461066,0.5426669922436567,
 0.03910697273286878,0.6247226895252562,0.6633675076136265,
 0.09129288925257562,0.880901209208034,0.8187156773770737,0.5567526830809975,
 0.6258681701267529,0.228595720125934,0.7725581574523187,0.14761656909038479,
 0.2636452962883976,0.18879373112192532,0.6250785143962565,0.4679423880552652,
 0.807728837821598,0.9815875618312646,0.09542908811132422,0.22899199182736785,
 0.8770446191883876,0.49351351933548016,0.22004970859392725,
 0.5530113054672166,0.5333057572481424,0.6135917107357686,0.42586854866080726,
 0.08009843280149809,0.9341171270657226,0.33526012858397647,
 0.022411556083137385,0.2977971274068475,0.5279064878374132,
 0.3466223121845866,0.09508813612779332,0.8058700859619945,0.7577178694850528,
 0.43043852990833575,0.9027967942043862,0.40490498288571675,
 0.5956261884920947,0.9335650115639916,0.5161624415693693,0.8488841122092505,
 0.2595416714787908,0.8620136116733383,0.3972348902765298,0.8241994222314991,
 0.9140255514593871,0.9821353639177482,0.729675838489545,0.6228513855082726,
 0.07793950223137558,0.37794579012542306,0.7564338669106869,
 0.6130773193271657,0.43369733607596306,0.5943236007173823,
 0.048362764895835686,0.7598730478938633,0.8550899582717008,
 0.03228699626621534,0.8250337796099042,0.20978297575979177,
 0.11353907439669642,0.04511165041663601,0.3150854561069707,
 0.10194053174030993,0.04665837144662055,0.1548833114794168,
 0.6330231857111733,0.003590978719565041,0.3582560448807868,
 0.4706882700596354,0.5569474965531199,0.40113657287005866,0.7606724917711245,
 0.7327469591475394,0.7679474476074531,0.5038935008719145,0.7767424928734548,
 0.41166474977647505,0.2494904380643982,0.8482374433817764,0.3744345969492924,
 0.4805927642060185,0.3246643248499361,0.7063514580137006,0.894476465051874,
 0.8314405642819214,0.6375253285227387,0.36607834281686125,
 0.07707988562440049,0.03517705081762734,0.3742101295408786,
 0.2740027901653357,0.2577416657914243,0.9669571906790028,0.10454817007907491,
 0.7729730829804088,0.04268423629776208,0.652017085857938,0.5538060336858824,
 0.45478019899009847,0.8653739280907292,0.125481544513788,0.4231087661750951,
 0.8293178123709155,0.6225753706884715,0.03243361887307061,0.8245089779142544,
 0.7203964167092707,0.585327564946269,0.7003466291946675,0.7583977694649536,
 0.7720407024700732,0.31617252775613314,0.9450392221005003,0.5045896434082666,
 0.1267473000503042,0.25264945508909475,0.3003755523244346,0.468468210351872,
 0.284011273653071,0.7190838252328929,0.02784514662149984,0.9291176914675563,
 0.9022992887485574,0.9182482705062267,0.37836292063729515,0.9552087619333113,
 0.6867557730167667,0.777008873913921,0.8432425453987754,0.6390649914473249,
 0.1761851591631034,0.24999804506572343,0.5217835935658133,0.8781884627303713,
 0.7138483813953869,0.13119180246167783,0.8737534937087923,0.8367264007281108,
 0.4497282852230148,0.05885842227217131,0.19653680664883477,
 0.3906829991916363,0.866743085239025,0.3365729492172638,0.9651013568651023,
 0.6807072199928067,0.8191620785084551,0.07493120012465893,0.3445985598037511,
 0.09924291750936454,0.3430479622476419,0.1710823910083621,0.5081184854116796,
 0.5899531329532832,0.1461623874731851,0.9211828460772233,0.8344176382833969,
 0.6655862439986685,0.7095996789286169,0.6343009097442355,0.7543703278542443,
 0.9348971666495574,0.8051898669849157,0.8549103503360725,0.9086299120838022,
 0.4960076293852679,0.5068351738746676,0.1802225712009482,0.30471066792262724,
 0.24582384922745826,0.12385795707906755,0.6615811097068552,
 0.1748317996547618,0.30945792285915674,0.5467301753303814,
 0.32244303523520523,0.802382314115773,0.3683635976509875,
 0.025666374096906752,0.2880483270488179,0.3259990280375977,
 0.34598630888393234,0.22381120721755332,0.06266612849686526,
 0.7731715540198396,0.6306627953613739,0.42105913220016344,0.5452078667723846,
 0.2789369834601503,0.2840241093401934,0.38493505836524344,0.8330418953009031,
 0.13545249103905632,0.43181916454136005,0.27941936893125896,
 0.6278205171187106,0.801872209019968,0.3226343547088679,0.9303016409059092,
 0.7993496759261483,0.16219438307198952,0.41575385496480777,
 0.3210905367398391,0.5897676517161319,0.029847169631249226,
 0.1284079271277514,0.004350805442526973,0.20988587284629723,
 0.7265012568377576,0.5870035410679764,0.15399451001901499,0.5262328686953165,
 0.6825305295075669,0.9773080309979105,0.6245919977002959,0.26883505904855287,
 0.8121290960491623,0.3976479676691036,0.7600272117407907,0.33366181346474166,
 0.5441049974315412,0.7864038889664074,0.878480943310847,0.09247695810083612,
 0.7852640382273878,0.8906858109655884,0.43719484678932763,
 0.01135060895091089,0.20338696806056156,0.9826270323453682,
 0.8474845779526686,0.3905898316297036,0.3319709662545136,0.7823253983928211,
 0.6210013636318107,0.6467689855554686,0.939179966635831,0.678341902029248,
 0.3572827128336925,0.838924902934349,0.6864157700261037,0.8223651337053058,
 0.7669861954370045,0.6568289137168674,0.12066501519844952,0.6317617977839244,
 0.822804687587001,0.5383930097667241,0.6489150354336061,0.023571982208317577,
 0.647831374426518,0.21080852146751494,0.8320226559952147,0.44756647659987714,
 0.29241754054663605,0.8283621478114774,0.3720439710827055,0.6751078222747625,
 0.21564614419423112,0.6721492255133805,0.7047955844691987,0.7511361166838486,
 0.49614345326065257,0.6883878487462557,0.6213058292854441,0.6010882270751011,
 0.8324452957311445,0.08172382894512964,0.9241365250728195,0.5490547089135333,
 0.8807197155900067,0.8926822213531466,0.7029525485808186,0.05204485232641498,
 0.6069462445076863,0.9732915802184899,0.10182349141930902,
 0.04128332039148508,0.5069749504603935,0.8548287227455403,0.4551804396433785,
 0.3262853620765471,0.08438167209251812,0.9582609007216266,
 0.46862543320084304,0.39511922713541536,0.8481207939115971,
 0.3410206316000195,0.5572372405271637,0.8538109866725857,0.4651084395271635,
 0.6099491584438744,0.9550169326169323,0.9515088830940918,0.677877036875263,
 0.20802082973465175,0.07738506771305476,0.06682419894257874,
 0.15973386630093211,0.27502709187572505,0.33482214786106934,
 0.28020288014992834,0.5414689599640704,0.929025243173538,0.07308883695567592,
 0.6707732981774034,0.8804347865898764,0.667530630357001,0.2660463429495017,
 0.4402342196372906,0.3964386438307709,0.25449614025877765,0.8710763142471769,
 0.2716956533514421,0.9240760731352472,0.731278572954785,
 0.0026830855683357413,0.9485672542784145,0.7376911150716321,
 0.265764537002138,0.19774774764947134,0.4336020890629304,0.9681751585104081,
 0.4394783825884926,0.5232024956468253,0.1178758861825705,0.32722330582993353,
 0.4712704857901707,0.3817532089152851,0.047459802164781295,
 0.45575710365715727,0.8772723580942794,0.3436375504259158,
 0.03370605093427015,0.931219213316509,0.7123786364826344,0.14251488055966188,
 0.7289210045635365,0.7953163732334328,0.34966440444370395,0.6897328870293548,
 0.34494946162646734,0.7744360039907121,0.5652877710661725,
 0.021466511233223184,0.8601257481853773,0.4221329722929472,
 0.9133825077654947,0.6032251093304309,0.5957568338295889,0.2915374340585608,
 0.7116569840089217,0.4884597731076231,0.5091156729468571,0.9345893058311244,
 0.25848180765033757,0.9225830589406117,0.30534447684519206,0.296386330058215,
 0.4790887781268238,0.6575155430969133,0.22349970304008726,
 0.16292751444354014,0.11740593826339007,0.40104081396821323,
 0.9194350777326501,0.4912411346899562,0.05982229328602129,
 0.20517098611321705,0.7391813073295176,0.8308770766270223,
 0.17602695579174477,0.8665333368103253,0.8421485136565223,0.6719388051963934,
 0.15410933501224122,0.73698513127501,0.011007183285686839,0.378876978859905,
 0.4490832925118353,0.7355087315388628,0.6315014967131676,0.7104898824002476,
 0.4540176487927433,0.6407551419521795,0.6066142570162872,0.7577213660633972,
 0.4980061118469218,0.03401987111802829,0.5428008290520556,0.586924598521677,
 0.15676391072853224,0.23670986830011775,0.3778051924705379,
 0.8367617478452185,0.08254190947694562,0.8619882404982199,
 0.17877135394933574,0.28334717460320114,0.28998772823229335,
 0.13885641044547736,0.3853187334046524,0.32319811687560007,
 0.42727380082265876,0.3644249468903037,0.3491862048883575,
 0.30741086428987063,0.04156551654819529,0.6885031714801861,
 0.2470766385309846,0.20300707878186885,0.19631657041599526,
 0.5532088289718946,0.6190152009587023,0.9173344471886583,0.3485067460662514,
 0.40143427744072013,0.23756811151978197,0.06487886905448415,
 0.3104222527945846,0.6695871822597906,0.7233572349180041,0.4518266387103159,
 0.0164436225377943,0.7149140400926186,0.6987727683556922,0.5569203054412064,
 0.5193756223629071,0.8472160514195605,0.1097834494176515,0.5122031267212521,
 0.8544349517993501,0.7106451334118209,0.8457522794635225,0.5332134652279696,
 0.5677057275351443,0.9275644152451544,0.20739896764863575,0.6141305862085207,
 0.7216626284125783,0.723804262360262,0.5072740546610595,0.2859971298860534,
 0.8202193151276895,0.9251859571234111,0.6919521361720463,0.800673069547571,
 0.8072742058052255,0.9381690002976363,0.9137919654215205,0.3625674010537595,
 0.984356109785515,0.5811173380695552,0.6724073401798972,0.820365074382857,
 0.9005822323125879,0.4610649704923585,0.7510059162864323,0.7375904044679668,
 0.43742679946665597,0.9000900784988457,0.6035653141976397,
 0.21119318027580913,0.9917957383322632,0.6394387213189122,0.3687587572796588,
 0.7096658845351089,0.916859705637694,0.15578401606713088,0.5241873437797545,
 0.8685290578710402,0.760228112868746,0.15079339449827223,0.9223011060020953,
 0.0225913668806057,0.8574578306179617,0.9081150736825192,0.47307537867767,
 0.32003917163027407,0.031049564086719994,0.02887807899392436,
 0.7543125830346864,0.34156023375204636,0.8008651566455438,
 0.09963063166193609,0.13475832489284856,0.5792652857365189,
 0.11105224317355089,0.3218050619316901,0.6573045305934879,0.9752910165319014,
 0.5748266370706049,0.7633435686664168,0.836696281050035,0.47925983121789606,
 0.3216386334245278,0.5709033345490149,0.2165518894198163,0.9532740118570997,
 0.8165362498356779,0.26493399976186427,0.6366449641393899,0.9710383215609437,
 0.24565113037827002,0.4932461534774901,0.9610710500788348,
 0.28213132265971674,0.2874888090992528,0.6187231060648518,
 0.18281439554375878,0.5431006236129711,0.06467899667128485,
 0.6658581935828783,0.7113718328071382,0.12617299529968262,0.213320511928056,
 0.6523359918037124,0.18034069090429372,0.5626448412909911,0.7262074205664169,
 0.8582128057067128,0.5387574375092417,0.7409396540547927,0.5477454426752317,
 0.07509968758132624,0.868703923329653,0.46775426526938235,
 0.10120286393913258,0.6126905287911668,0.6435037823223659,
 0.16393649904822338,0.061720271391615844,0.6316335927188766,
 0.9107217806205485,0.250337900676035,0.5616553278242811,0.8841687347224967,
 0.1585091370947752,0.6050474687118351,0.8572483711483735,0.5303389346154725,
 0.43184290387796,0.18558006537617333,0.6961638500026578,0.3059930976973736,
 0.9929800818551306,0.8739739960705273,0.1640880883246958,0.9057784673294017,
 0.5546366867159033,0.5484363534621735,0.37579119059363,0.012684819204885955,
 0.4965549782636365,0.9710476353947954,0.33020015245864687,0.916204736846921,
 0.0018179706241395754,0.5937032764436072,0.9750705540600182,
 0.47883371836358823,0.7162338009300239,0.7135800886333089,0.967904187609749,
 0.9708637576197632,0.3563254958171993,0.22525860322513147,
 0.05893643991413822,0.8001173965510592,0.5554014930343635,0.9920104962438694,
 0.8075027891086584,0.6676010204312063,0.3821081959981055,0.0698647748012724,
 0.0018360239928725264,0.9407873947406813,0.2733198582977602,
 0.8168489778503975,0.16547022856339755,0.9294997914989632,0.3493040319716493,
 0.6572432096088341,0.5836046862936612,0.5847919619599331,0.7909177831088179,
 0.27237835105844277,0.24596610339557112,0.8240117201249771,
 0.3990215876191321,0.4379150772777495,0.7419255608605293,0.21415418087177263,
 0.04463392883072759,0.7437093205290578,0.8493710826193731,0.7001392226302668,
 0.48820720008655194,0.8218622066661627,0.5139732368508675,0.9777011597270772,
 0.08171854275687052,0.9084364539530713,0.37935643881525716,
 0.15593214205853367,0.15997666798029453,0.0763581972584565,
 0.31045583705003255,0.1382657615300129,0.20562662295437695,
 0.33398015867747166,0.7143215787958802,0.5509960317883763,
 0.40096815476258785,0.9108141138478475,0.8873649657985951,
 0.11050386627085351,0.5199906951356379,0.04640291945679498,
 0.3788942610743158,0.6711431935261417,0.5117424670763235,0.0874424758006116,
 0.773973841363992,0.044337389442596575,0.7696367204745133,0.3917346664998048,
 0.6669401490420294,0.8084161289349143,0.7527190234159667,0.10213104708672649,
 0.42562570855640236,0.24967169058139993,0.03917976424762082,
 0.2573178232538874,0.11777391712723728,0.957190670682813,0.6925796731467199,
 0.28869439644808437,0.8746850803087027,0.13406080162926914,0.393709011817281,
 0.6435340013398614,0.5324778068049167,0.3285801627205932,0.8316153906728329,
 0.15265072645790012,0.838483486207946,0.1873565520087469,0.3660550865365465,
 0.13803597658493527,0.9858225059027425,0.7656747582136908,0.2451745398716949,
 0.8441564295143212,0.37957171038507753,0.4754988827541452,
 0.39670809902327087,0.04086413808586553,0.45612910979459464,
 0.4659887159173348,0.44401541548184587,0.6473215453726096,
 0.03752624091319123,0.9100319813573835,0.003766800208390153,
 0.9029829873973374,0.0830515947578071,0.019120669131891366,
 0.23963101819390975,0.9053888159708166,0.3179175348600095,0.2172692831903631,
 0.02548612731411204,0.9848146105508383,0.761234720965684,0.7498488940779141,
 0.051511818525789566,0.8078330676983706,0.8539127282054854,
 0.22716122581947473,0.7656183427266432,0.10628670130582007,
 0.4884294140423987,0.006809900931959656,0.5212595674932393,
 0.6495143571228013,0.6269236687323212,0.619708541784187,0.907510791728728,
 0.21554938844403237,0.17096799445542843,0.3420781942191945,
 0.00891578937367321,0.9687569541887472,0.45812864156766664,
 0.8420414045055332,0.7736062857114763,0.5404191514805218,0.6374168066717651,
 0.08216668616169798,0.8020641991552898,0.599216012871482,0.08033578052421597,
 0.07136638020824093,0.5864361536991413,0.6508576278869131,0.7567448852730093,
 0.4559985532712544,0.2877979652836288,0.19308852425130818,0.5497370906091296,
 0.8057419089477658,0.6028294695399069,0.18007448885952382,0.8695194373118333,
 0.8420147240970255,0.21771354142889,0.4532542223790632,0.735608420359283,
 0.19281983687253745,0.12276580061425257,0.3633646893041085,
 0.34646165147490937,0.23669368542223435,0.7559477925792402,
 0.13326631258743094,0.2601176137770135,0.1179634535435331,0.8368827086483634,
 0.33668311427194797,0.45339923429411,0.08522823656510048,0.26886454808286675,
 0.047918949300837954,0.47799875673505243,0.5889834290524809,
 0.616215316079296,0.32385065040276606,0.6826425800563436,0.8256149320010908,
 0.4991201096473762,0.17190771492267165,0.195590262033437,0.8427696313696555,
 0.6140924068160593,0.6646008468348206,0.1346263376318897,
 0.014706323837151203,0.02428439531154225,0.5508563116362694,
 0.6847808122417378,0.35324028941194496,0.1936060112812633,
 0.18383369490186174,0.4943776947944045,0.40910031468875174,
 0.36369373287698703,0.699204446481913,0.7151486365881905,0.4135170139675517,
 0.19287153064265317,0.6753334762219756,0.7195931122058179,0.8948566216257219,
 0.13138797904690047,0.8508558467245232,0.5579536309779652,0.7350953621767405,
 0.5956713754131772,0.6881297816964448,0.1689538670364893,0.3667794157672145,
 0.0724041614055948,0.03737098252140314,0.4973754823938016,0.587434478091616,
 0.9450197528825101,0.948639225133419,0.10341504399171142,0.4883873515153718,
 0.841515160323463,0.9900706666845455,0.8058977857509002,0.5118816085642971,
 0.9742925247774106,0.002941023459361425,0.8960324104287765,
 0.3809524562366924,0.8747530823917883,0.5259318387993547,0.34685250334068274,
 0.15332281347336452,0.8385377373756876,0.6632752063269025,0.3613226295106736,
 0.7443835779795696,0.03777723734178051,0.5171865504855439,
 0.34506741085351667,0.8864077242604638,0.03237041735777835,
 0.6133109552293398,0.8586838527756648,0.34814525133183727,0.829809870819162,
 0.5423503786517566,0.7403862879318766,0.44974184630038405,
 0.46202301516465616,0.8526159824614736,0.3784452615768721,
 0.33718564593381806,0.06615180201751847,0.8622175411400159,
 0.47312789029915103,0.12308908219127412,0.14302103381713172,0.82844649960868,
 0.3330265154380134,0.007587033047571801,0.6464641893727173,
 0.8663781261736161,0.6034949303841141,0.6129606546616069,0.17268876820845813,
 0.42025098655494597,0.03014640854505901,0.5969715302740802,
 0.47443261494852085,0.03490324322246996,0.8511460491086797,
 0.2494062096452393,0.9849347360747474,0.09535538744296834,0.2725850087749988,
 0.5810913609937504,0.9459432480422477,0.21156645508309335,0.3785778796989866,
 0.9822572354570218,0.18072380924040898,0.8182168855832602,0.8285772201338956,
 0.23912132176394574,0.8571942238980382,0.32740691865228133,
 0.21762121870380424,0.7703408875818006,0.6833720178794811,0.5692383907255484,
 0.7263585506644654,0.1323577441296767,0.883482858192858,0.6785066633392429,
 0.6214756657482147,0.7674547049764217,0.4111291135774714,0.5736807274753025,
 0.06205843469199268,0.6751190778132742,9.456835956899168e-4,
 0.526979280756473,0.07089660227033912,0.5553683187446147,0.8737562893492608,
 0.6921857640093869,0.2168024176257456,0.07680115481416394,0.3613624496572414,
 0.15286987607261804,0.7916581721299655,0.6839257814269585,0.2848308967447465,
 0.14769068203606894,0.4552504562494092,0.17893167187454906,
 0.6141306511996695,0.536169123635336,0.14049908494508911,0.9508835691583,
 0.7213507928917918,0.8219967820459925,0.2892763785694008,0.13364240100341607,
 0.11267412001065824,0.37800694104252064,0.7131703343269595,
 0.9701395123330921,0.436965046029526,0.8438069918881961,0.3597472540536024,
 0.8167385439010841,0.7991537829099129,0.1566084631731942,0.5803490158176238,
 0.9515818623488721,0.3194854866495871,0.4615049754266487,0.31757726223718274,
 0.5753190494832726,0.8808380472055736,0.2530999532945688,0.3361587541961848,
 0.6747871861118648,0.8332075955191488,0.018599094198148758,
 0.1603856353934845,0.6713660719014731,0.546328602739602,0.5792738871866948,
 0.7910906722273532,0.038607940707524935,0.2865954020830034,
 0.13650919107213988,0.7052128446201232,0.5156149868035707,0.5039702168617258,
 0.12106388899147069,0.19051269155847583,0.689981241958796,0.991736163889187,
 0.2938731235957138,0.9835383416364453,0.6073681214382092,0.34567949593412295,
 0.06331854445452001,0.36441787157780203,0.6461630411683714,
 0.8065485576212187,0.6043160547556699,0.9832668442482412,0.937397000047101,
 0.7314460254106164,0.5482016021714753,0.5050485997963668,0.3152520277565909,
 0.6256720505400207,0.9808776263766044,0.7365016217030238,0.7163843999456441,
 0.6308610771653311,0.6442854117326472,0.08782710762840118,0.6096865154300265,
 0.12307066552190471,0.028739904416232975,0.02676205433581258,
 0.09036638850102519,0.5456381972226443,0.7576814760204444,0.8588036379116032,
 0.25176104343769434,0.03440007721102245,0.5320175235881274,
 0.20038290134153647,0.40524320513134227,0.3228773393793557,
 0.3852835469839536,0.6282556522797823,0.07797406899383097,0.636128422814958,
 0.9929394706511515,0.6001111502072588,0.1618406825466061,0.6636042518326317,
 0.08956547548890914,0.9088481646316975,0.9895186479637589,
 0.14595448284384227,0.06610874027086433,0.6776295829634055,0.553150667265072,
 0.48773041934796635,0.7377798889183761,0.15301325068778548,
 0.05361292491326641,0.6955317422225896,0.8061551763820743,0.8432294991336786,
 0.07764427057093859,0.12006205751764687,0.6367920300610657,
 0.5692281011562872,0.6001747463629661,0.6823240651525863,0.8190848137432392,
 0.24289756287765984,0.4828350380544304,0.48872593182480006,
 0.8353993585584258,0.9422458528642459,0.9983883428034762,0.31129953664598564,
 0.4772164206149915,0.423899725508038,0.5346128766998178,0.7588682818215486,
 0.9412375413610201,0.6860655084172198,0.5787096654049028,0.7393377720881075,
 0.10135091540715602,0.06702285108136685,0.09967966771962333,
 0.45929930298155697,0.45140693813199406,0.20318416316273613,
 0.8547106784306205,0.28848980503924637,0.349483096897935,0.4622663760043506,
 0.049954825727434926,0.11768986915649138,0.0580511456123699,
 0.5781970253315422,0.7834087406686306,0.06930195773903802,
 0.19238449105408884,0.25697232538571857,0.48740860766415084,
 0.5313497550903648,0.2715672754410684,0.8999814012798955,0.3847848422352811,
 0.31475112397481353,0.05589545590327272,0.6394699532767846,
 0.01183292915644829,0.7987414084434621,0.6468923322324395,0.6757594469740511,
 0.32869803315419,0.7376803517633104,0.8944496095717515,0.5921994709173427,
 0.3499634615498668,0.8887660416522892,0.0972152027983132,0.8012099282153222,
 0.28802324057886153,0.22141935924376333,0.1771081045990165,
 0.2951922283483144,0.028119520086637206,0.5752151872137905,
 0.8995282049622875,0.9944364028789409,0.8208352104455059,0.08386046937770586,
 0.3445927892136057,0.5749986428011647,0.922528258110061,0.5454666727448573,
 0.40966457693373104,0.46736876385706616,0.8027371785401423,
 0.21771704872101105,0.8779656748777988,0.2559040362681302,
 0.14686907718453057,0.9687331890201942,0.6972901742574614,0.7406148438770971,
 0.4194238802801923,0.8640800489824634,0.5067094479036494,0.31411361344961564,
 0.3739541097131396,0.7227486926248984,0.37738795660168356,
 0.17077299125962664,0.3454780185666376,0.581449633369263,0.6252341562651474,
 0.7112522296924946,0.601790317050766,0.735405045096482,0.8650234147094297,
 0.880020066968645,0.9133621165656787,0.3467332768239181,0.22946683115738542,
 0.03315583530045707,0.9171829324163386,0.21970772496428326,
 0.5340749438240863,0.09057388100746733,0.6887098985626646,0.4707984663227355,
 0.924375862511586,0.5329516218623646,0.47307840758297803,0.8691888331627645,
 0.9953150576326066,0.37801824833183395,0.8304354854199428,0.9864565959205254,
 0.20491883859679638,0.47253785598889164,0.43392636617842495,
 0.5958477703848333,0.9610036543656277,0.8164554860680859,0.4879595117925054,
 0.4253357820690824,0.8355754396290375,0.007846618513063008,
 0.6604148448912168,0.7901901716491445,0.22088444893327708,0.8380610955153545,
 0.933198350836712,0.09302633119940318,0.3262871539134097,0.6588122780897737,
 0.8062579446021668,0.47401822569957086,0.15601680915425875,
 0.44930409619292455,0.39074352175115556,0.5835141317006115,
 0.5353081500466663,0.7711797709390611,0.9229676095188946,0.27360269942877435,
 0.012325803912652256,0.35129949336654975,0.3449248936790348,
 0.10107479658546248,0.761470600312064,0.04844107489623206,0.8187259567338533,
 0.2967721260461672,0.6931292903283932,0.3843155502855017,0.9031627525495872,
 0.04316468651749583,0.97566794540434,0.17049252048279417,0.2078690948224653,
 0.6972674756197456,0.6396122702070266,0.2152889076524327,0.6910567801324108,
 0.3422257103939793,0.5327641654316435,0.9411007483895821,0.6746049461479373,
 0.3970849012883685,0.9439436736925783,0.4836614638136192,0.3687707926672372,
 0.3615998758615956,0.8215235404068308,0.6940748226983549,0.06372535635913157,
 0.09384905419414313,0.21608984284037458,0.019960625153760403,
 0.31312981802749995,0.8376917096612159,0.7307010866069973,0.8659352914664672,
 0.11409423266099383,0.6201150885635635,0.8733528621195844,0.8948137247312069,
 0.9916933147093556,0.5452651897134815,0.7014041607601331,0.13234917810124158,
 0.7394729911386596,0.22242962677884337,0.7954862577207296,
 0.14333025198309968,0.37181200513907076,0.17225974686836065,
 0.15281904629939314,0.10163477058614756,0.012019370886943825,
 0.6776319565137544,0.8976307909247068,0.906647098124622,0.7461411289196497,
 0.8042510777693561,0.4382972953727198,0.26745580963882887,0.9329563203273579,
 0.745876387479511,0.21732988928225305,0.5578150287605423,0.4399377778966427,
 0.9915805246933067,0.6859094812512225,0.685913597678136,0.20026627657337626,
 0.9168914628808432,0.037163791447088035,0.5084820365298236,0.422383733761021,
 0.3714426441789098,0.27102264401256726,0.9348718201791315,
 0.15900903888512852,0.3486823639410428,0.04550957218469398,
 0.39917966525176884,0.8608619266855231,0.8083911417685856,0.5663991137338832,
 0.42607740769773506,0.43736921647040605,0.7748608383779637,
 0.9213880292788427,0.941449384880489,0.2346477126675055,0.04657046992795488,
 0.8923185447622891,0.4462617240623389,0.29823125920851323,
 0.013373169047286138,0.3027522059334402,0.11203375441273089,
 0.5207279059958678,0.8052978553765396,0.4089503289701981,0.4491027515730359,
 0.6415858826265032,0.8494476494624427,0.4532229077050166,0.6204873215921641,
 0.19908374137364715,0.8109509377165446,0.13282772157789358,
 0.4249840802651528,0.7263737599693632,0.4599288800978276,0.9123733366308171,
 0.3645617535140032,0.04595654883882172,0.46018924374495285,
 0.7000474201826095,0.5549749282122687,0.5140758310778046,0.33617253898259214,
 0.549195298332683,0.2894985738228929,0.2886812154841423,0.2896376026839017,
 0.9220024243501386,0.8319273795820638,0.21468821795970783,
 0.11626939160648475,0.410714035140062,0.074619278127532,0.7075270340234309,
 0.15904303774388673,0.16364640655273066,0.13470745868518774,
 0.396379586965376,0.31465851674468914,0.10315919685582153,
 0.11653836529429373,0.4820042336887044,0.2885293973664379,0.9849459921188487,
 0.24714184441474274,0.6008598790782601,0.8713318539309779,0.0822213709380436,
 0.4732191562263721,0.7678740772392069,0.9392768461304737,0.34474218009918256,
 0.6594033251382159,0.15646948854640474,0.35871936672827176,0.390662697425869,
 0.3775634833613025,0.11629143198977598,0.578914476707895,0.8492583490861637,
 0.0824498680756548,0.5515406953213826,0.9578343347451223,0.18453227469710498,
 0.35174756938944063,0.9868306899000064,0.8064071075737406,0.9267203972846738,
 0.8126098228590358,0.6613188910320615,0.3378958530812375,0.11748891787346105,
 0.8658329937589191,0.8018059502252464,0.582318953192853,0.7709133440194893,
 0.8660918344628964,0.5459100612834017,0.5253121136541795,0.02888003236584402,
 0.887443149582764,0.007020953850689127,0.9921694728632461,0.5949967351542746,
 0.5980415141311877,0.13156157169073557,0.7116038918950345,0.5586096818684797,
 0.20871803244806308,0.19378224397717858,0.6898435146438121,
 0.6659127383273769,0.22620031129596962,0.2283556802388368,0.6498645822710701,
 0.37840728530705126,0.022453303905794697,0.2577880199032915,
 0.6212631151147461,0.4545041765242682,0.8925210739142151,0.6446915642939448,
 0.47860707000606106,0.05434641138088658,0.88991185762877,0.7151142537045146,
 0.8074943467479612,0.5355932984148336,0.09473534911021197,0.1740357977002337,
 0.9149352240108233,0.7918050315647768,0.3643546828433162,0.30158085352505726,
 0.4207754449197474,0.8139780614759085,0.79876885773386,0.469386594471357,
 0.6829599068974729,0.12826117881105814,0.5819013948620756,0.7955196188629455,
 0.17315669986110915,0.341830857236028,0.39334808804668353,
 0.15678985445781402,0.9789730969542653,0.5936247122637459,0.504274336129941,
 0.15771765766438073,0.5602906950557618,0.5230700949181646,
 0.23215615667486433,0.4501518413243759,0.6784680860180753,0.9404031906578095,
 0.446177819593761,0.11759343955455892,0.5917725136510106,0.72494691231772,
 0.030464148855314677,0.20432133556971788,0.9001643493097606,
 0.11951441163509868,0.767554313036865,0.004183958097330076,
 0.06870037639859783,0.9542308116783151,0.9257423824342614,
 0.49640403386386034,0.9526962200997743,0.4323374987332578,0.8299311756945843,
 0.15997614346116384,0.09465919890793972,0.353916546788164,0.763944489044274,
 0.7152379851259506,0.24351895422439707,0.07512230578929102,
 0.5003030507157923,0.16096260895982917,0.6299217631150836,
 0.05663534502450718,0.643825003778951,0.214965890353655,0.809755015317706,
 0.128153261060324,0.38992403585177926,0.5249219581544733,0.8507070672205819,
 0.8046151646879389,0.08621770450131161,0.45915538489046326,
 0.32185305679031073,0.715441887849543,0.843752079339783,0.509706433917526,
 0.9690127560742909,0.9266036663210545,0.7286432694940767,0.139484127357704,
 0.3386589600748927,0.22931633258670214,0.9904518184515515,
 0.14947430294985375,0.5155726562013228,0.4154094923012218,0.369198500140712,
 0.9185669904926592,0.5150266231762497,0.876621415779693,0.06978236895696721,
 0.7121675431920417,0.4718046875135885,0.47786891095446504,
 0.28402444149935846,0.5497771409100667,0.5885311214290452,0.8884374367901995,
 0.9664186073896964,0.8548604215589082,0.0695946268003993,0.42978358948517437,
 0.9629580127067296,0.8433186059118536,0.7083754309235939,0.7521594790054467,
 0.3379221172271155,0.1397178562808753,0.24082812658182107,0.6065075616184203,
 0.9265024347175109,0.04611759218808731,0.3926506198952403,0.7663756128821345,
 0.18004593711299277,0.16169603382907682,0.8262295324132595,
 0.38972920658693855,0.2590125734267056,0.11159415001241924,
 0.6490855738817748,0.4554403216873605,0.5759594098109845,0.6654753193730005,
 0.8291484475030478,0.6548359463360909,0.21870662931550333,0.1286264612700727,
 0.16666068987633909,0.3822075351997173,0.8924288562122683,0.2716094964309419,
 0.8239642621555884,0.6065190857973699,0.4054144656018668,0.5733560507378986,
 0.09910669295296848,0.9506502433537329,0.9514768614646457,
 0.025486461233905633,0.04681162242491321,0.5199351200163104,
 0.4551672594382087,0.29360947672307547,0.6098018205071946,0.7159471554495198,
 0.3038994612529581,0.16748947039413142,0.3457948322991089,0.7680279138876436,
 0.5387828279628357,0.08579298344328756,0.17715361543515318,
 0.36529935299189353,0.0034025457752381305,0.5785072413351635,
 0.3099222330115512,0.4739739135505092,0.8881962072620725,0.6382456173899684,
 0.2951176983173358,0.7057716246921247,0.8096879241955752,0.8825535180502427,
 0.6977327387785746,0.8628670028243212,0.5408914529629607,0.20929645607944536,
 0.438274533645006,0.3011852455862182,0.1372422457727791,0.6378800258430691,
 0.11509897155346827,0.047213243965316676,0.6016791547237724,
 0.0716662249672737,0.44909273398391203,0.07620762743877663,
 0.8172659570577272,0.9566497098953114,0.34105754898959817,0.6836283152729857,
 0.13822929820535002,0.7337276973940671,0.4904453640172538,
 0.13669701133631706,0.7974535838780933,0.7380448082430967,0.9486559418168539,
 0.25181625804281493,0.7394965054238716,0.9628393983129448,0.6981647194927797,
 0.03206431690887335,0.6669904540414938,0.9342262389535859,0.607435506928278,
 0.18747345177275576,0.9973317069126202,0.31697279757046515,
 0.4444325010466931,0.5590608518298713,0.40763904839651044,
 0.15544218109013774,0.0781122799439109,0.2855467923375814,0.8538184618894038,
 0.8630252199920488,0.7428184483878839,0.16028724464624355,0.0446145009860246,
 0.354012415062657,0.37988998801071927,0.1880863339768194,0.6889424264903423,
 0.35653074322908107,0.7114491932377953,0.23388602742967424,
 0.9744558030876816,0.39235796760476327,0.45944286462582395,
 0.2806630530353247,0.8517710483394751,0.5319771407008892,0.43993765018320974,
 0.27958554664306534,0.5577693344115366,0.20204498861494524,0.725606694647296,
 0.32883693564156635,0.12948211655737807,0.6698948978381167,0.7453859550526,
 0.6945355449272723,0.3746553034619351,0.015368183096453603,
 0.8368126840132558,0.14449009046602046,0.9644751762470536,0.345758107690872,
 0.4735595835212357,0.13391398800016618,0.5237024619456656,0.982118012670629,
 0.5215997282158548,0.9038548596954358,0.9476192399824037,0.8748253587252016,
 0.16242783408938966,0.29546952646094016,0.8501358904967367,
 0.2126374605452036,0.2659076772306783,0.14233888656025173,0.5961480778259494,
 0.3833423737873334,0.12715389849531678,0.06322269118460333,
 0.39291922792184863,0.8310533376543412,0.7650105387187953,0.9635307878669651,
 0.9062920955287982,0.18470749629790273,0.4006301088629558,0.7116078470352453,
 0.6365647367826862,0.26925550780437013,0.1643123946757149,
 0.25823096492262376,0.9373005184314227,0.7724854795898022,0.4366229630638583,
 0.8740765634138714,0.8065916317617411,0.17055932969021326,0.8833305965182632,
 0.7338022094850423,0.7249849428374908,0.8865018958513435,0.7429445262531389,
 0.9624143405728142,0.9082252320450934,0.036093464153694965,
 0.9653651181325156,0.3250915432296382,0.16401278693548715,0.6041079805292507,
 0.8940780766484536,0.6913908793211843,0.9893635939177536,0.554105573520434,
 0.6462831490815204,0.8947963107179863,0.827249624280597,0.2799256310792809,
 0.3580558281732703,0.4223107149271915,0.8245498936304703,0.9982010315737616,
 0.9288231076823107,0.40794634714080147,0.1322002239791753,0.7728107627834047,
 0.5959396692553279,0.5372947220003148,0.8180991817638993,0.8523728889094717,
 0.08414305613359108,0.23798185690003126,0.06600038439959488,
 0.9262615703798945,0.3483698970404505,0.22502897821763135,0.2346468111429626,
 0.6719082918894363,0.334509420794154,0.5908698083921798,0.9438749745756769,
 0.7037075610871719,0.34517880205335427,0.3296857152096496,0.579657232693209,
 0.01098014117384527,0.011634015179611934,0.8203546241666357,
 0.16116430888677746,0.5267781276929917,0.33555657087953494,
 0.7106865487760262,0.3081332457484526,0.9663755025786607,0.7519419299441849,
 0.6871606136613759,0.4238622691048314,0.30155411997805404,0.3503803859084421,
 0.9097840367357846,0.549855916014012,0.8497653766706332,0.30398444457562046,
 0.6582656282101258,0.8594242841987317,0.796887142537009,0.4638988620860861,
 0.10404958619861826,0.2994886928916808,0.6756547670097098,
 0.23054797149465478,0.568769274663565,0.9549477743641921,0.47143318346861185,
 0.2488662896268239,0.5243813437132329,0.3934623366459433,0.4241233259977656,
 0.6918710364372096,0.11426657392609552,0.3417408370550771,0.6618343452881398,
 0.3801538787258457,0.34429824649099794,0.8535947216596224,0.7119893145932013,
 0.7690562958656499,0.6731991451811529,0.8126897422211448,0.1902214793115573,
 0.9672553467445913,0.750031969211851,0.24303807020040902,0.14926169381944177,
 0.45895011996069035,0.7818793576634013,0.8855186328172953,
 0.019630568780851232,0.6691549777784364,0.503840677687768,
 0.08522238621889755,0.27836550471602317,0.32174926622018707,
 0.3195299082891562,0.24209084908618483,0.4009778747248277,0.1919940987232337,
 0.6308171347971454,0.20461788311364382,0.6122201923705781,0.9891418777197964,
 0.45549733532886627,0.5194965212655851,0.053360846219604685,
 0.9225101869189588,0.6393076007751396,0.5186300138157502,0.2980474884651052,
 0.9209316978442945,0.5968185143944412,0.7072387659817421,0.695060238669687,
 0.9759938732641307,0.3376309647573734,0.8386738491911575,0.9687656162288031,
 0.7049419249113733,0.93506080920888,0.19492839607537182,0.7821740584583105,
 0.05074958304158772,0.7414169486422285,0.8400322475407923,0.9830981559921541,
 0.5641985545281085,0.9071725200429553,0.3214163075485489,0.5622399786969349,
 0.2757212473211518,0.7937612075004794,0.4909645182812369,0.7611743346807567,
 0.9216524459599669,0.48054695658659496,0.6850009336864601,0.8618400523326355,
 0.6249585765456966,0.4614202693773388,0.2941481745285217,0.00909251292837665,
 0.4357967762874805,0.3087353233171716,0.20394969560784015,0.2589195016272251,
 0.7677175977736347,0.7331889415007466,0.5129057915401836,0.6702041815351832,
 0.22933973398158758,0.3904958718044018,0.923708039088684,0.3457582175075159,
 0.24527424352983473,0.5144797297197952,0.9805107048200556,0.9624895797728963,
 0.06602034201717999,0.9774082891947629,0.15490055576683048,0.489170568090734,
 0.8562815189420188,0.8449733300901282,0.8037370653518155,
 0.058097386568889076,0.01043834558219725,0.30956792528360466,
 0.7760079505636865,0.3595774629890517,0.7836814595713806,0.5865787980451533,
 0.28540035906450156,0.5520870464839581,0.4845075688889833,
 0.19052707242954436,0.030609912553266172,0.8209734556676692,
 0.6705212841820317,0.7471650566362263,0.4022309248784739,0.6633952597716185,
 0.20697302594647515,0.6468856427423801,0.9710266865713866,0.9126149728992945,
 0.8620815563010898,0.7238065196317484,0.46367002621438347,0.5737671738333874,
 0.7811872325547684,0.27028797026654605,0.15065578819608294,
 0.9660486596926974,0.9341417245072479,0.42672350864451436,
 0.14853967607027507,0.7869238050418289,0.9163327017714921,
 0.005113847937131988,0.10926668936496697,0.4188298283338394,
 0.9407210417176909,0.4052480287003094,0.5074916895809946,0.7554524703282838,
 0.05457208425239102,0.132307163661193,0.5059193380335053,0.982603714710998,
 0.9401323252147257,0.7589561217569278,0.5683854878482585,0.5523600413952487,
 0.5595341843178403,0.1319855794215089,0.8589613980467935,0.2274895989368143,
 0.15507399782283704,0.2888303117772104,0.7878944622779318,
 0.07462351964666036,0.5110573324508294,0.6494417168188416,0.4555972804250189,
 0.8445417043903723,0.47668963345112636,0.9074676971875362,
 0.06334589426253068,0.5708242088186651,0.9386875236772729,0.3366313149867154,
 0.9777703773322177,0.41387863699515903,0.09586086469113231,
 0.9554667891332327,0.8392390086296228,0.29193106519074097,0.6540239190528085,
 0.18848787144999324,0.15285448639033206,0.11462486428682483,
 0.4485227958194331,0.6418627691994969,0.7069576933960554,0.3981535371509015,
 0.6580192132139746,0.9957884763996154,0.850414272660639,0.40070599187189393,
 0.6103915989248334,0.1631527274302338,0.030318787404190473,
 0.6001632031823038,0.03455251127123238,0.4122054234008633,
 0.24737702748663581,0.6026444319160784,0.7380483966185807,0.8189444531766027,
 0.5897295623998317,0.9506775836671583,0.7193109851387477,0.690267793952297,
 0.5174606307149251,0.7877289348480477,0.4166182870863956,0.22419318806135868,
 0.6883788761635014,0.5254868603698363,0.38874243655771046,
 0.08164593535777176,0.2686134352990279,0.35861297331269204,
 0.22086156741890095,0.6848686369439464,0.8547464138006816,0.3742848012405483,
 0.38213028847399233,0.37317164596997743,0.7646992358795203,
 0.7452620511263448,0.10901057048159901,0.23649428897056912,
 0.7082500499972457,0.15401900546609992,0.3178991838208012,
 0.20487724253614215,0.32512730032996595,0.6210381689811714,
 0.5115199614423809,0.6274460630388082,0.9870507616345614,0.5344107473333302,
 0.127435240964613,0.5270249487882669,0.7394050497942884,0.19965819939160423,
 0.3884472844400789,0.5196165423527556,0.8974381763989702,0.7884589634046245,
 0.025722153176036988,0.980011755626895,0.28185577024159025,
 0.9953597630761617,0.959592697528024,0.9690301161913674,0.6907586144283091,
 0.03628463498822221,0.438812756280635,0.7110519426913142,0.7263504987827449,
 0.9856532826743619,0.5613020074758556,0.3872588495259013,0.41010220017639665,
 0.9085627107923011,0.7347886579743448,0.3918922127074179,0.9080611748615596,
 0.10483734323443605,0.17236483180754258,0.8136514294385928,
 0.6229539134373719,0.64920711489261,0.3470847237770682,0.8646067332651478,
 0.8405684034253902,0.2565129690500021,0.8325314270622387,0.8876113078186231,
 0.24043549620923166,0.8833082143593566,0.4126748151579879,0.710467776956739,
 0.5780261462467619,0.7393817680067887,0.6386782883107501,0.0812778391390333,
 0.8544804630486517,0.927541726186821,0.795013083167618,0.04889362046464685,
 0.8127660225706959,0.713294530899266,0.393388091488717,0.4433486817739073,
 0.5080235623866552,0.1814952885942167,0.7083756618716067,0.1759561496533084,
 0.13961282491270632,0.8305036904431823,0.6323416144161009,0.803524196639601,
 0.3248166514637123,0.43969946744916244,0.7274862516690103,
 0.30684068332139314,0.6223048336222657,0.015736626506688167,
 0.23767846607595744,0.5705657150882778,0.8784857880432807,0.8941509443245346,
 0.9550803512025741,0.7620925897704259,0.36399105981391766,0.2942252803688157,
 0.1234322698897925,0.3653842364480585,0.4066619601630469,0.41956031247650527,
 0.19892953355487797,0.7686664296879955,0.5127260749178466,
 0.19068488260891814,0.6121264638763088,0.2819320345951649,
 0.19517074192966066,0.38337269207053004,0.386820019915483,0.7338377194732171,
 0.24943363165441623,0.06772677684831796,0.40986342769381123,
 0.9135784679280426,0.8889135803254825,0.3989096154145596,0.1492739581674084,
 0.13880393774635058,0.638891458356899,0.22448936951842058,0.3726315796092532,
 0.37198258214534274,0.23615707355734583,0.3183973626561645,
 0.5191404201556703,0.7656225706395126,0.028573971076941818,
 0.044827355344925324,0.1837202738543744,0.6062540184691438,
 0.9250205036267284,0.8328340301934117,0.021775917032911152,
 0.9161834360748411,0.4144426688000564,0.6333257519434201,0.9169843840985321,
 0.7715532888874975,0.09801169948709176,0.7419096708307059,0.6857045331554401,
 0.4358821196325128,0.8005698653345251,0.9831703656740994,0.5042703808776059,
 0.9477593366497907,0.871750080583968,0.8008083580556784,0.8703185997444309,
 0.8899382420540826,0.3995361990607007,0.9459390593144326,0.7324480931562001,
 0.8451007571414181,0.22515654919547157,0.3947910877789912,0.4578113787187271,
 0.4520302595881117,0.7703422191385294,0.8778949451081166,0.2876653861712707,
 0.8041246317441012,0.162259454103444,0.7698826380019448,0.8241949561388654,
 0.6440355478175184,0.09101558935155463,0.14688357365897287,
 0.2524600995551468,0.36874933723435777,0.2519752351385157,0.9279312136282435,
 0.013267578391271573,0.16461510839311821,0.44023818505503165,
 0.388035015586092,0.027553223277864802,0.6757981974962063,
 0.15153419760096987,0.526054543226232,0.5039104418600155,0.9353563386184905,
 0.7270948864401502,0.2132763731068401,0.6834152646302556,0.9768128106580699,
 0.12816730542085042,0.4794674152738039,0.9064260147824087,0.7719289509685829,
 0.039270430632716846,0.3454810526501356,0.36993734979384585,
 0.6041394059006201,0.9623155963568829,0.6733758491401828,0.6900573015031052,
 0.1215964456194818,0.39906504883890315,0.33877839007032917,
 0.5697757212942718,0.7060914997096011,0.08175620572863485,
 0.05853406588441257,0.4973947727605684,0.21204648135983362,
 0.47549190866598057,0.9884819646775324,0.03081823072425205,
 0.6054575357601908,0.6930583168181066,0.7239144002825217,0.2869431235259472,
 0.11722141529858843,0.24393049010333157,0.5585561101050504,
 0.29637299313961196,0.46438756709866347,0.19549424889071787,
 0.7013425102930161,0.5741099327777287,0.6435044431048462,0.01391325332852722,
 0.7364705110849551,0.24086802087140247,0.365517243670746,0.07300378544422581,
 0.27605772382300986,0.6868095283235992,0.7616853889693473,
 0.37578829584819196,0.530066693618914,0.8336612837010711,0.47733428730698213,
 0.5125487725927508,0.18964242884236537,0.3053468654977667,0.53976511435935,
 0.6562480457522883,0.3597086515601461,0.9413113832850619,0.22413128048552167,
 0.39742794208510257,0.03417200406443954,0.9111145403513813,
 0.7451796456779834,0.8930886713853242,0.27200578257896,0.80181971971329,
 0.09522375280108353,0.9904407797986896,0.7430274489000098,
 0.061550399988517634,0.3473315123789047,0.7670631693735103,0.209518213784205,
 0.8823728501790936,0.6909942761972463,0.7900663337844057,0.35496489217800675,
 0.9953703833016079,0.7794504848225354,0.9745833457619452,0.6290154280262754,
 0.11953077990706606,0.08025959131981508,0.21349475788201377,
 0.22082915679023818,0.12647757354169076,0.19633237772550816,
 0.6374765022782719,0.9575456992727737,0.32286693044692116,0.4637604399443399,
 0.8674297060367983,0.8448481559291171,0.28052388701769315,0.7666075976631664,
 0.8010440584064893,0.42080622779441035,0.7220951735423151,
 0.33966404762315683,0.8454326958924583,0.34158526800442157,
 0.6055405949936088,0.49689459494453225,0.04373938436344427,
 0.5951421142183586,0.04278167724063742,0.9486891026957125,
 0.21116380197609197,0.1886573470895282,0.02230494421729201,
 0.9126967468942624,0.2585326678143838,0.3325599184048684,0.14534300421785717,
 0.5483400101465842,0.13113920656005484,0.29453533819406486,
 0.3267061237247575,0.8417768128146029,0.34761803136735525,0.9323816815558702,
 0.5683922246649897,0.14896712467264206,0.6840198019474919,0.2087346467311233,
 0.9641769000667778,0.7107688691197397,0.9548328855256086,0.6078860344276267,
 0.5451657930274524,0.6412820174969217,0.75938388015254,0.7756993456968789,
 0.9014245637794173,0.18530504234076906,0.26261710469599475,
 0.6551972015624211,0.5105652584204953,0.6098681444678191,0.7122455721793134,
 0.21368255735592256,0.6815087154725247,0.7653533951677709,0.8179398194592428,
 0.4199927300887779,0.4159330817143301,0.13693862616531904,0.7346837292816716,
 0.1528479687301697,0.2821106244012722,0.5699155749051582,0.8335197467176939,
 0.8278081186057322,0.42308920701743724,0.35633988114773074,
 0.7182318836145759,0.0030707405873852123,0.8167683055865386,
 0.7270464371599656,0.3356884065311594,0.5712446186182822,0.639189509258517,
 0.8032313603274133,0.028690235042310053,0.307320234115285,
 0.47810381287101433,0.2254135859290608,0.6634483589324557,0.8048636183745851,
 0.7321852538464733,0.01934730250071448,0.711742475939679,0.7855253121429806,
 0.04581778261578229,0.7064956743228157,0.2942006490047102,
 0.42934105187473515,0.28560765366250107,0.6719861561693203,
 0.9264904920362769,0.007011970292412184,0.8782589071914281,
 0.6072866353500485,0.4309272623586531,0.7485799728366069,0.5177689281904911,
 0.42966568368050795,0.8578424290522538,0.9058397358892963,
 0.15110946131248237,0.2157714837543665,0.16167271494171698,
 0.6437335706203446,0.7298164831650971,0.7819877736180794,0.7788505613575871,
 0.10475310805625482,0.5695649919735712,0.14359146807941026,
 0.5528883733640134,0.08181388732740758,0.6376603006798611,
 0.11910991341743149,0.6941938976671588,0.5580415198246468,0.5477290904197223,
 0.369516205053811,0.6273946905201617,0.728867980240135,0.2787114545129674,
 0.14095631236953898,0.6273014222983182,0.8346805733975402,0.9806345932069657,
 0.45126473498447317,0.8928709204175814,0.14960167244936295,
 0.8082045636998927,0.8910917921751056,0.0871090747842791,0.4272429161740904,
 0.18928363698522332,0.03121842328694191,0.7685268130185688,
 0.5152929557484058,0.7268257893806204,0.47198462690359744,0.2028683514142604,
 0.09357630704614739,0.7471166159512487,0.20280038397900857,0.604693141368348,
 0.03146541523418489,0.883286349879363,0.973944261896098,0.07349844005236272,
 0.8741768265773525,0.23242715158291416,0.16214059302881834,
 0.8297231909511663,0.954222008393973,0.666848212558186,0.4078448853499773,
 0.5198083089093377,0.9923993257746091,0.9653568069840255,0.15320678637645457,
 0.026417162922434012,0.8341705044556168,0.1785757416998761,
 0.8946837868834603,0.36585428233931405,0.005886304919253793,
 0.36128237624936455,0.5755201394682753,0.6713155107527182,0.5375325553668902,
 0.40316408107125357,0.7841530130674387,0.41674959154026636,
 0.12973020828944815,0.46601771655453694,0.9219604548813753,
 0.6813145673139054,0.1888523357705344,0.28760051462314395,0.3487076987226708,
 0.8953449186935221,0.20711061138089426,0.6023013187412168,0.5426555654588545,
 0.8167558855984669,0.47038875780166267,0.9640974729203995,0.5204188677806968,
 0.9765735156143657,0.018531457611401403,0.7774235158711904,
 0.34931389700977933,0.05454575431850184,0.17764057584392368,
 0.9983744353726793,0.6193748295528836,0.623343239827615,0.014713248900885767,
 0.3275004829301632,0.9062526207106255,0.4057552417129727,0.04587119979663645,
 0.21792018875466224,0.6775845603723499,0.3422347235204868,0.844693072819976,
 0.08063830353619494,0.5772135880735633,0.27692119372570134,
 0.07220659775316707,0.5094465833353039,0.22366794486353503,
 0.06016399274034534,0.5365270937915047,0.6465874530375119,
 0.06392817855614441,0.725072641396391,0.3870567151306671,0.8834952742846074,
 0.05238721397852719,0.5175685793359722,0.25281067065914553,
 0.5950779098649344,0.42435659621486665,0.4166377514014743,0.5144344114746273,
 0.45168201435481325,0.7044505285551763,0.9548790417707897,
 0.39882620928844925,0.676938733226292,0.8288468436404164,0.43334272307741195,
 0.48975408928611863,0.05836970461031432,0.17760817394009054,
 0.6419172851630544,0.6104152920234345,0.9575616485330809,0.33229057572191356,
 0.7925036693290598,0.060177220614898275,0.04778106393624548,
 0.6601164267666516,0.46798677655337073,0.02136391385475156,
 0.8704317556232455,0.30883980892557417,0.6023239416972453,0.6863158885229108,
 0.03280615437879286,0.2047599091418303,0.5178506604370099,0.3281229624095372,
 0.3055579042645268,0.601421244282705,0.48172028064117633,0.22270840889745813,
 0.9635438929142277,0.7549628197039082,0.7932252680162829,0.5742461777840968,
 0.3928787546688919,0.6350747510338122,0.9419861978546111,0.817166824336933,
 0.3346438580186567,0.10100398040110825,0.2981016274684509,0.1100923892206721,
 0.9238957197111142,0.6967925876797875,0.3553075736556339,0.39804853949575913,
 0.6966161492520926,0.4906584101947299,0.8312476646946515,0.8752088472407499,
 0.5455111959567398,0.35002293638472826,0.5327256503647326,
 0.14011684427372084,0.36500772016465843,0.18149911510304506,
 0.48570157510805934,0.47896688030546486,0.21345096826889165,
 0.1326938840461055,0.9137205137025951,0.9227665634846509,0.19290919804979367,
 0.248632790379555,0.12742886220391658,0.3951973904300561,0.8733496389892215,
 0.12106607685355053,0.9604002107111055,0.5777340315651055,
 0.32805256931579585,0.2397595433296884,0.9531303510703766,0.5971864034235304,
 0.2732548942653823,0.5412007865235684,0.18979412893506864,0.891445883837181,
 0.9784180179004354,0.8105651694887969,0.7082192691015633,0.17993816172914534,
 0.23282615208773727,0.917522035840355,0.011178083631126379,
 0.6187844585325674,0.00763713987662229,0.6335065463741436,
 0.36561852224614644,0.4258012043909932,0.6532879802871736,
 0.06371676572470941,0.713066454186096,0.06410589448037318,0.6672518487688792,
 0.2572916224676668,0.6690650417017505,0.3190196040633795,0.67071689631143,
 0.7135549877232228,0.2917912464168524,0.6562233498154822,0.6581399914524881,
 0.7425735193358713,0.3690453230400993,0.6257115202698001,0.9256643251987983,
 0.23090183705095635,0.6532027537673967,0.603222034132707,0.05002503716748807,
 0.4430197440938636,0.4643175432187122,0.7889224607559358,0.9107531286069751,
 0.5642809825661197,0.961485846378728,0.5513178365816237,0.4463685145598357,
 0.5111430313886101,0.09213138866151038,0.019008321464425326,
 0.7168254559329105,0.6837836062197424,0.6029669343935558,0.15427230026227545,
 0.8083420645610611,0.34713756516059446,0.5161783395910209,0.6000367612091175,
 0.21523101749034146,0.06878557968696997,0.042161416486445225,
 0.7768257233491804,0.13629619025719886,0.5942972297744576,0.6078050162102773,
 0.8804240473630703,0.4936865956352591,0.6489575406765034,0.9462800839975867,
 0.6352381595306371,0.16336154919956802,0.1087917522200571,0.7687301280596671,
 0.42758590027814325,0.29280795574892793,0.446944537220086,0.4161462808702632,
 0.2694247343608649,0.4562910050107303,0.8106186209062897,0.777490707765214,
 0.7111626603791398,0.1046424267783631,0.58610277304465,0.46526283052525663,
 0.14259909110047198,0.9917676714458981,0.12453127227197835,
 0.1251566228665757,0.5897061742238597,0.5717796491931866,0.6182764895916769,
 0.6204805384464079,0.6393359884522821,0.2256136139961905,0.7112672323241315,
 0.04226495626245519,0.773475204122717,0.23594650643128678,0.6129987039663372,
 0.9787953264239632,0.18584848949598198,0.8062669722383408,0.8405801622544018,
 0.30639721674198295,0.6403801003217972,0.8131868352389238,0.4322093010451915,
 0.6649164185039206,0.28068490666240375,0.8416938431228037,0.7206437504090746,
 0.7816830009443814,0.2848303925670157,0.14678902010497796,
 0.13439133010035775,0.2995477369973998,0.35774448455162877,
 0.6641690376757495,0.29049434617554537,0.001762744128703364,
 0.302455550543953,0.05577479281028219,0.33140114267051746,
 0.44557723411605465,0.49986864521632035,0.6662803319711227,
 0.7900849018770437,0.7089348473908311,0.44422169843325954,
 0.10840076850122549,0.43743804700442135,0.8861683140671773,
 0.2044921404705372,0.9386007845342735,0.5428634855681753,0.7616647688333034,
 0.5045687139908598,0.47359467888442197,0.6757117058011654,
 0.04919320077754874,0.8164599843649523,0.8444609611025113,
 0.19262192284518265,0.38084891359681783,0.207489634691708,0.4743936667715608,
 0.1523549668077384,0.11656877107660224,0.34659667799980054,
 0.9009007080822758,0.7399796165621962,0.5843184786687171,0.4131261876497079,
 0.4696304141517119,0.9192161591686302,0.07993380120281035,0.6966030702693933,
 0.2407792008820241,0.12098095961955802,0.5377269948064556,
 0.04995064642530744,0.12102541763747432,0.2534911987346332,
 0.003438874470270359,0.18652578295152245,0.7654263551857196,
 0.3636571830405392,0.15182281698244438,0.6906992055484459,0.3183449616826528,
 0.9583900833037374,0.863569220767183,0.06540615231473246,0.23688671205312684,
 0.3252784360058635,0.7169309457229442,0.7734397946679136,0.9639616773201899,
 0.1040542278861909,0.7374850376334058,0.41008155179839123,0.8894517887417219,
 0.6475487060452111,0.9365973640188989,0.536297671175625,0.9550841573511863,
 0.7307932186946796,0.4283479061040887,0.8149435184481997,0.4604723279152133,
 0.7439744663929292,0.22072694614300026,0.826781111554586,0.18148519947646713,
 0.19906550001194123,0.05238593100518929,0.8756352159831549,
 0.7557052361381436,0.5066573652029354,0.5516687602526088,0.9566955504685327,
 0.27917888114310585,0.12127712020825487,0.8674179375628107,
 0.6070543233993347,0.8721486282127053,0.743539974363029,0.6952241833788106,
 0.49530627040718744,0.6184752768019695,0.5203605918819918,
 0.25371292357184205,0.9348184360063858,0.5554920242571788,0.7732844575191173,
 0.6050171859470346,0.3436487440518937,0.225997419792357,0.9844644041143857,
 0.14789763047811466,0.3697602052909742,0.0628685444746434,0.6634511134776329,
 0.664580596829632,0.9880322091843041,0.15867856602722663,0.8287345505934126,
 0.21470947183225175,0.02373472408344235,0.7819933411143056,
 0.9788724457841317,0.8340678233680595,0.05938798049700822,0.8350462015380423,
 0.5075593664762561,0.19153791453687474,0.34353337373043336,
 0.7482807562488439,0.08567666712963562,0.657540392646188,0.11851802838088132,
 0.5021306881590001,0.4337643814523182,0.5686309492098729,0.8686242496275467,
 0.3439584473964655,0.7133331442366271,0.5553650246486229,0.8069717116982029,
 0.8938896236338085,0.7286524237180552,0.4950418001835776,0.6880706720370089,
 0.5622832339156567,0.10106804626629429,0.3071780695287313,
 0.15866144160908702,0.9637755257880355,0.4073649294763373,0.6146247595532033,
 0.6291370786260679,0.2686301903527145,0.35895896492334933,0.7895348975050003,
 0.21043926617527575,0.3210980010511778,0.024115799616995726,
 0.6589848943788305,0.4770688452627436,0.6713754744438035,0.02289207881890365,
 0.038807323671155136,0.483594074837007,0.6100933361004556,0.7264887665522443,
 0.46904579152295334,0.5128257415577473,0.26653493691456875,
 0.9901931266298197,0.446310285034895,0.1252394506627713,0.32590877130894735,
 0.3841481241258269,0.36081488053910793,0.05731274907170403,
 0.31815591670675825,0.025193436369809197,0.4227880222678134,
 0.25560025679490017,0.026861496393808082,0.5425860444626031,
 0.9850208478867627,0.7712419534307959,0.8674826733114478,0.4242390942208508,
 0.42418620819369146,0.6699415560080278,0.2102800816947734,0.7956903894845652,
 0.6061145645796836,0.5800951415390713,0.7274845268203235,0.7672164883788698,
 0.26282846499905865,0.30591665280155034,0.7767020152260707,
 0.06485345956242583,0.5964471557086102,0.5919216115281829,
 0.041382036103194575,0.5884178525719997,0.034744514805586935,
 0.912689157434464,0.9799450850948164,0.44160141067936076,0.329993667490858,
 0.9708042677608133,0.624121954894219,0.3777061796723088,0.7816271889558823,
 0.37011700214859733,0.520375481365336,0.636852285105924,0.8681643162328577,
 0.2872378980179836,0.6209966834730822,0.424497434703931,0.5875892750886728,
 0.4525473395836418,0.405274645920076,0.2655557141324998,0.40729837686853987,
 0.10720087566708614,0.3432994524185452,0.8914969588362148,
 0.12852027491068863,0.6308893013251103,0.8708825513259963,
 0.05959464106984036,0.06521095528177479,0.5185916977714683,
 0.9141676877574981,0.7095355784604251,0.6866432204337976,0.2697215235479665,
 0.5694815255970356,0.5264789492542973,0.10092600624731729,
 0.06929679980784098,0.14820195594536978,0.5302286371945942,
 0.46675194459819114,0.9486281670490719,0.7994723857860948,
 0.33323841417649647,0.8420627381685752,0.4340110808154779,
 0.01325177380886955,0.6334501362927519,0.642239407468238,0.3219954420983767,
 0.5199257584175011,0.1662904199635179,0.5302986952079006,0.8091898881140717,
 0.6296047656968801,0.9458539072577752,0.2413515813008803,0.8007358689670598,
 0.26276301235488686,0.4636165777171435,0.5402651348908312,0.4681431385162035,
 0.08298368569477088,0.896222520083594,0.7340425528593254,0.9816282493636351,
 0.2520121295430505,0.2185402978801798,0.37544520968325856,0.6855150631470093,
 0.5872282517572714,0.5734847046536556,0.18195717423496804,0.3927628872320277,
 0.7319546730173634,0.6234351380768796,0.09655415830463343,0.8784012554081296,
 0.48011745909659465,0.5678030025700567,0.4305685851865777,0.4382358516268954,
 0.67085247681257,0.21878038892751606,0.21658472042837806,0.17831933472746353,
 0.4226163882591998,0.9635893820389442,0.3965864382777726,0.02117538817305953,
 0.6488385313154724,0.7629825535199883,0.3262097653658017,0.01283614065483285,
 0.46827032482205877,0.5043772987272778,0.22372981189182295,
 0.6499587528632509,0.29001730561624706,0.530540649588426,0.3964441625875391,
 0.9110713998555565,0.43005489634348903,0.628384601376078,0.72873341806109,
 0.17755887067730747,0.12925819216434298,0.5906637158705823,
 0.9301031708755364,0.26529650051693565,0.15114241039531062,
 0.3550102294988151,0.3689475211992185,0.508136032366477,0.7636833547867126,
 0.96304624195458,0.5163674269037497,0.18843848781576145,0.6860976109729457,
 0.8001393328952843,0.8668316717264618,0.3365514812690654,0.6296058601641352,
 0.28777137455326474,0.2241315279064341,0.818631460433649,0.644593885990076,
 0.4001736566653048,0.9283931412526214,0.09698912831780193,
 0.39749411718392413,0.09556294964781153,0.0711255061265299,
 0.07518483760744155,0.45913272908695824,0.30138746051795495,
 0.964600460798193,0.7582978787220181,0.9889901979600012,0.3509133904164442,
 0.6250095059557189,0.8791326572504653,0.5225748138484357,0.06836750416089254,
 0.29526467197538975,0.5171946677185915,0.6018018461417926,0.5951363256152745,
 0.8815961529315114,0.18978721558799183,0.49442907301649863,
 0.2577967973076628,0.17366262597217796,0.7922187247446107,0.7528203711583761,
 0.7433317160391846,0.7109252163245455,0.7723220477622232,0.09775909631898538,
 0.3964141612950566,0.5693444587566278,0.11933491835977827,0.4515614196645755,
 0.7748896221316625,0.11485129226149593,0.5348837060008972,
 0.04303305693244963,0.3401180531658473,0.5604101343431487,0.9940312938528761,
 0.04382532762261382,0.06706581065022055,0.14744258751216488,
 0.4697110157416211,0.9646508976402739,0.24144101819851982,0.5888905044543239,
 0.2332172196107808,0.5450522086502221,0.9619269641509328,0.9975275454172929,
 0.3702229964925885,0.12501222842826665,0.549486607901011,0.49315365739896044,
 0.6533637719091906,0.07299567524141481,0.8518688275208166,0.2507267167852325,
 0.9168099556950228,0.1702808766336844,0.4808364010146118,0.29418788507737315,
 0.3554504420743816,0.2261308774972297,0.16063442724139865,0.8519692290454212,
 0.8243664096267448,0.22289886061297504,0.447788857123784,0.04855022568785616,
 0.1808831587945704,0.0930317595851422,0.5873815274865631,0.3665323878499471,
 0.05972388879405388,0.29475213995324556,0.2542813742199316,
 0.6555494167721124,0.48524061103585203,0.49675076665774287,0.599529873146561,
 0.8121625997002129,0.9673802284727568,0.7398010807015888,0.4982927106723293,
 0.8677385032519365,0.7889269433052049,0.24976762238626404,0.7502641472903697,
 0.9281408389938886,0.6209145386575219,0.3321196214531327,0.8601115405154028,
 0.9031675681477319,0.8354275710021378,0.5877351514994877,0.12966171689640726,
 0.12178736243358079,0.8302083874407064,0.4276116562868092,0.9407651303284885,
 0.42135812290598584,0.8697251708003789,0.1399183804353653,0.6774831043012421,
 0.1538951730499254,0.8060561050911855,0.14770776559569443,
 0.23998147514921986,0.8847374941030185,0.3147449706525268,0.9486763230560465,
 0.19863767787503217,0.582463813264714,0.5872112691176378,0.5831214889329044,
 0.6214354740285787,0.6860256452696816,0.303226790319878,0.15274613591744035,
 0.21612902551630608,0.7313197358724077,0.95817970829907,0.45496726601219994,
 0.9505214844205514,0.1659824558169577,0.8481612611497695,0.9420881427811414,
 0.7051988026512769,0.8785367476235818,0.8705516033750753,0.893861631409962,
 0.6734403387585024,0.9067081840847241,0.49005031548558153,
 0.37329132581252733,0.6959173407405885,0.6811940257782013,0.8482695230928784,
 0.054125606253928815,0.5872133627064036,0.5928193029132574,
 0.3945074891158087,0.26031637349053427,0.6927026582696373,
 0.30607468970918383,0.440866343163395,0.7538182794479275,0.15771518134775775,
 0.40424560283070976,0.5672945377701941,0.724710727979405,0.3125127141444555,
 0.13004784767286848,0.004464500871453447,0.6515222000476313,
 0.9199732409669128,0.4745692586654148,0.5160331976925112,0.03965711161318064,
 0.7024364889650623,0.9568956814110399,0.019685800543955967,
 0.8765246504815195,0.4964942455385989,0.8688404554833171,0.6567004335304393,
 0.049387237958592234,0.9588181166723672,0.42371060103986113,
 0.41151610443624087,0.7856538664208778,0.3170399793643137,0.7221316395466694,
 0.14514181689227434,0.2756170182737583,0.6736081079997707,0.7075291140149451,
 0.9019682998769286,0.16570733009223737,0.7185569612039031,0.9563776318964765,
 0.0032752302797508515,0.22770938213882352,0.3417340320458476,
 0.5257481992114171,0.22880935732667107,0.3414535401724941,0.7078810700699152,
 0.6326056637128756,0.41826076172306736,0.5782894634898753,
 0.02543646016489154,0.5181911441981615,0.14427857554529966,
 0.33181631797327604,0.8314692169501058,0.9094114054974203,0.737544403389203,
 0.2273667209762067,0.6662313507659302,0.9933441791507697,0.3910558798826772,
 0.43053608460830395,0.05080306846782234,0.3718801791188796,
 0.43885466274912244,0.006483930469027066,0.21290859765212677,
 0.003639323306185105,0.09228425698816711,0.9143194331272804,
 0.5011381367222631,0.05331794910497445,0.9251150063610084,0.4405383163365372,
 0.7749272355292973,0.6971966406138509,0.24911494400645784,0.8607540602370803,
 0.46118989733432214,0.13367622131056422,0.19357996776868747,
 0.5870688784662736,0.6161346838210278,0.9480916015568749,0.5214636757526311,
 0.6450921762209502,0.362177128370742,0.6671603620942936,0.9919437853872768,
 0.7093595271567823,0.09073465176545148,0.1807803980567938,0.2843791745341848,
 0.3900357729599703,0.05040603034783597,0.6894215448910831,
 0.40142753163490685,0.8377973081121343,0.9599865861718677,
 0.04709025101994013,0.7945380765919415,0.9776679476625715,0.4386962338196343,
 0.32794408575874134,0.13571477043783442,0.07874923587724458,
 0.29582411477175075,0.5681578268155566,0.43361367914489857,
 0.20408715526079813,0.13305568775534793,0.3974617755228107,
 0.5430091570395844,0.5799236717940232,0.10363468416113486,0.8077878052690425,
 0.37989086684557083,0.4527920125011309,0.385531185355823,0.5258416255782933,
 0.9085391540989325,0.1251470806507039,0.2929760082490187,
 0.008116833656338085,0.512198793445609,0.37036255736738044,
 0.3348685660155043,0.6676587995343916,0.7009057102423212,0.6853990362826782,
 0.6731271454870541,0.5749850355137034,0.703070303730965,0.2550962155116211,
 0.26017190974454796,0.556820134353079,0.3420685224526384,0.8013959674564664,
 0.1945304082976691,0.9574046573803962,0.33353885892836954,0.5228716153111996,
 0.6567049049106531,0.44311576400796393,0.47784589747745254,
 0.9871982987883371,0.8769343156777245,0.5604049798188859,0.5178387347130247,
 0.7346708059493733,0.6379493208783245,0.9820091496333556,0.9681290552619283,
 0.6774225580655546,0.3605738066584002,0.18759204218567593,0.5013244108598024,
 0.6049990200560658,0.019788444810054373,0.29533151429267257,
 0.14364832910139247,0.3647659417424556,0.9170627311292163,0.3139772447169338,
 0.7184740009370132,0.6224389910911817,0.7964144704564879,0.663660969926157,
 0.041302529109485064,0.3955126683776288,0.2606553954065831,
 0.7857114990497296,0.6993676286487819,0.23981860793559684,0.8662903352209278,
 0.5517156859210369,0.7249628504157642,0.3009430679685152,0.3826701022310215,
 0.13216696029484165,0.5340983958359924,0.5997731030999243,0.3324843326278094,
 0.5565166992554439,0.7696641632839891,0.39106359346862707,0.5101555486466668,
 0.6587723186586623,0.9479644013718548,0.34250235788418615,0.6215949948018605,
 0.8721285278133578,0.9118247270941284,0.7786693153272088,0.1250365374500194,
 0.7867275761584327,0.5865165071323685,0.6391293540596157,0.6631201888513409,
 0.6134650103626619,0.2127204724926005,0.26897577239809056,
 0.22781861585029317,0.08300126721090317,0.7639332720209058,0.325928968445248,
 0.07400930890203616,0.28093898973008,0.9053783609749692,0.21610988491948613,
 0.6830744443982759,0.7798528761220443,0.2518415987484641,0.8639179122583922,
 0.6806186346456156,0.2681404383257785,0.6161742663434462,0.9846313501421794,
 0.7753807488938286,0.7174094898704465,0.47596188805349504,0.9502590178322259,
 0.9424936668936453,0.13060320665998248,0.7211364602901695,0.8116345616411987,
 0.11036343171402185,0.8939951080612809,0.01931274680880435,
 0.2824671386381553,0.4327613329363583,0.46062869676965545,0.7733103665749397,
 0.8329069471742444,0.03451207278489987,0.05270614081358871,
 0.9823582062946953,0.6263503355005869,0.9651112666879453,0.8926623364897969,
 0.0656506547150092,0.6256682360492097,0.6101909016012446,0.8779874755499777,
 0.9090754323530308,0.7496889236480128,0.4467419205989427,0.3232332504644213,
 0.4402921931065855,0.6516799969996443,0.7024036324540905,0.668724244912316,
 0.2087669458678285,0.048772149312138136,0.12841000880973896,
 0.41891195223549127,0.29679348797740224,0.017489389982757286,
 0.8273699756930162,0.21462361545473096,0.15403879737393156,
 0.8345064278949185,0.8198602977870602,0.6036130616761012,0.5561581988915488,
 0.8705260294928676,0.8142642791475414,0.9410111469148643,0.5049044058949321,
 0.15391598629820524,0.421735236832428,0.5258790042847261,0.0630290482835395,
 0.9203608587362276,0.053239218450300774,0.533458366216525,0.7843440194543337,
 0.2937016949841591,0.4817466550118795,0.6425382844647439,0.5980393113513021,
 0.6603273514504151,0.1131995886388587,0.03857987676489061,
 0.27359625325200754,0.6912454062165129,0.644206776096277,0.5286036346057219,
 0.6187959762265375,0.807887584243227,0.6453209129208164,0.5169448560355708,
 0.553642127149194,0.1203727298379148,0.9669091582029985,0.7657302844400335,
 0.9141781434344026,0.8281557755451581,0.8894915031356708,0.896742029804401,
 0.4363196070679485,0.8741436828800332,0.7171100815835314,0.09250834624240034,
 0.6795184937527441,0.3699682358608706,0.13558565563294833,0.369647221412557,
 0.9049131741490752,0.6843546776251145,0.41766363228906866,0.6578843674872231,
 0.6212403874324928,0.8759415901556823,0.8152251426665877,0.9483486917798019,
 0.47360998304510504,0.05349384476177177,0.30550050275974794,
 0.003250837022994002,0.393288035138061,0.1349374357140477,
 0.14691135784014886,0.4527065418528464,0.026839885776549055,
 0.1141654688811492,0.6381543677547346,0.8335905963851324,0.9183874795533834,
 0.9826450573207426,0.11013063162070424,0.5033444016400197,0.684362949246388,
 0.2797348809941037,0.9164027600310145,0.7734862753449046,
 0.036437376089602624,0.2894216864464976,0.674530796560802,0.7258342907999332,
 0.9110006733202564,0.9307774281067106,0.6934906135683535,0.14399108199705035,
 0.5746443850017101,0.6489885990253419,0.5125189211171839,0.6111191141612113,
 0.2974035652199081,0.42102681851255785,0.6094610180281064,
 0.40339524422085826,0.712146178961741,0.7467095895335251,0.708105718318057,
 0.9870264445189865,0.2598053115969755,0.7196392345382252,0.6041441075807714,
 0.2793259078635345,0.9883308588871648,0.263616857037968,0.7450157030859269,
 0.20413868766908216,0.4143686422836351,0.3251539261811124,0.6125910945213177,
 0.9364276683224811,0.1522436010755973,0.1373508134802921,0.37675430850426517,
 0.7354467181620338,0.6564910568203334,0.4399945881342404,0.40924249749228647,
 0.20669536347563988,0.12969500887746843,0.3540362372778115,
 0.16052895136854384,0.6199027782164956,0.3832924752198633,0.4174039949397442,
 0.19611008138035668,0.21563474084648704,0.7931061972350182,
 0.5499303344836588,0.4694315261881662,0.9756108685059258,0.48754199356264394,
 0.4768969801171191,0.02422167793060248,0.6762286575041564,0.748717411320206,
 0.4122915937914361,0.30077871772536524,0.15819142592712343,
 0.9583959147428329,0.9022456735290045,0.0918901862802286,0.25863689336429774,
 0.45945416870899203,0.0886104560541412,0.6576962824161778,0.6028393915174982,
 0.207790258461377,0.3601345649820801,0.3309983780208742,0.021723227928418964,
 0.8936365609172267,0.9733477725820722,0.9598391219211223,0.5588872015544045,
 0.5877874125456785,0.7870469803648741,0.8514553204277904,0.10708243481140622,
 0.18950882451577833,0.3946605584065902,0.9494456329163011,0.9058430685711376,
 0.41202438458545254,0.4283517346745017,0.1873960660614964,
 0.005795085376901765,0.6636388850644142,0.5831975522555884,
 0.7276504241759777,0.9118675927035479,0.5041109414349234,0.39290176950050093,
 0.6749479924240548,0.9703009818430978,0.39257626099425025,0.9557078020325325,
 0.794501240179637,0.4641082987171672,0.48016234924438317,0.6970109907176786,
 0.7772930071284305,0.1894354655921512,0.7120354309755624,0.4770949198353861,
 0.6854360644863196,0.19705776264293384,0.12628650362453464,
 0.8522800336449035,0.5899098966806838,0.08431890583707247,0.9617157649232934,
 0.15716057827115526,0.5910000885216089,0.8311789436412095,0.412746309909247,
 0.8812144206576233,0.027923395403192663,0.20554809579481492,
 0.9503336444240665,0.34926125642696604,0.8669219185110573,
 0.05519582282783997,0.4653665021763055,0.6177683622147605,
 0.07643295647450876,0.9568908972212642,0.6424559765921631,
 0.04361661328788524,0.4281643927257478,0.3571891628519791,0.8518122544961493,
 0.3316189067499703,0.8553109929085064,0.58238969845001,0.5670113706264076,
 0.973288484664961,0.1841900068308837,0.5173183630714621,0.32798758392369787,
 0.2668497762825949,0.4059503680670853,0.7446011334834285,0.1123478021574309,
 0.7249004296997664,0.5628812694554217,0.9028622718607537,0.2628915372723092,
 0.15305891894258594,0.5909281004506654,0.2806500203096144,0.4600249832671294,
 0.7127479894441106,0.8548174475605106,0.6231615806208792,0.36051337810401374,
 0.9487261011133139,0.13755394389335496,0.35717169256822534,0.879728501070274,
 0.28311062126097875,0.10707895371402487,0.12162259037113998,
 0.7935292746621792,0.44417123064757613,0.702201976089291,0.11872962237894558,
 0.28723105595825416,0.4772294809127915,0.8981584778708979,0.9110147905568238,
 0.08846888453616586,0.46025116221497486,0.2984319761663836,0.363658237174903,
 0.6557369192971954,0.6436783295318416,0.9102632428232686,0.6579030205141,
 0.44037735490470187,0.9986520373983384,0.5876260558942106,0.8631061993089927,
 0.9197182590073859,0.4549540948530614,0.31349224931905706,0.8754607719951019,
 0.3407948981075516,0.6393181677794366,0.30018482185094497,
 0.18797748896744404,0.5923380031758451,0.5443577922929522,0.4978100185147176,
 0.9999132198288322,0.2200181253627046,0.12876987444658816,0.7371697497271752,
 0.6951886496051862,0.5751081158100337,0.44832230897419145,0.5879088909488634,
 0.4948064697736062,0.5968953776364633,0.5007456286854187,0.4453360589531543,
 0.516659803178725,0.8379968135520829,0.8368543324839425,0.13043988007111118,
 0.03706510885578296,0.9745047264253287,0.6577675203839977,0.7162103817250668,
 0.8623055397589398,0.7605533222350138,0.18548680538873852,0.7735516181558998,
 0.08674004248223488,0.27849193471374256,0.8891847553781386,
 0.9634325350004681,0.09882939312227312,0.048704776404071004,
 0.40021719543868395,0.5585013612026083,0.1690295812309477,0.2756462448203634,
 0.7898745691546974,0.4049647725517658,0.0849065984587195,0.2420409655054817,
 0.3220184949806081,0.2525378648745735,0.17199183778853888,
 1.2760143338386154e-4,0.04571808208236883,0.43837264581360713,
 0.5873469726063572,0.56898873478659,0.1040993912167103,0.735463232929953,
 0.4963819513992729,0.32325944648547866,0.17468245604454968,
 0.9195799391171311,0.4676094116960897,0.3991530106267176,0.5231349881023453,
 0.31229238484285116,0.9294013974923997,0.7186104304432006,0.6805950172681495,
 0.3495022533934242,0.30151810273198154,0.7445639847237316,0.7129568658462715,
 0.4585760350719904,0.10503853757265369,0.8002554965452019,
 0.14010434348274114,0.2798144598762311,0.9956637071224699,
 0.12680642572236367,0.4895948199700352,0.9953433094561096,
 0.34674899797696646,0.061763750816796414,0.868605048158197,
 0.5607186432520823,0.8716018068298256,0.8254878877944557,
 0.004641447809364485,0.13384624575911586,0.992879555200285,
 0.7435592085699185,0.4562307646506719,0.5897789976327581,0.9572003569459677,
 0.43533623798406307,0.16939946436730768,0.426263890182736,0.7510273344889062,
 0.07947549885821259,0.49111462765756353,0.5961695565235092,
 0.08942051202085965,0.5564087632333077,0.8907412448671764,0.7597063861696136,
 0.753991512546279,0.6147484585820387,0.45784464092530586,0.7547872965322538,
 0.5499727616939679,0.8882577969928601,0.3695522663118087,0.5870877479043717,
 0.5141144987060567,0.7608032990557527,0.664219609564215,0.7923255654549358,
 0.5943645958282513,0.9725464549557029,0.7579262430332367,0.9623939124446685,
 0.3914417402293633,0.9326906219789692,0.9719385746262393,
 0.025756640526421126,0.860577497793835,0.9171237651201027,0.3928046440791654,
 0.14717176948909538,0.42863145656079127,0.1485937040318055,
 0.16936354650470575,0.2609521952128605,0.4086255337029059,0.6621379378163523,
 0.8777331893725004,0.5214831613048088,0.4317771779316265,0.5015881393496058,
 0.8992353360610792,0.033750931831981634,0.28039882801241767,
 0.883221629252649,0.5910475187220273,0.8841456557823757,0.7092217105755785,
 0.6235053601504577,0.6496575355142306,0.8802090124413773,0.9280889436977372,
 0.47924728283863494,0.9929988714705991,0.6853033492932986,
 0.30845933675017945,0.6986337896277481,0.5015369105632592,0.7157487632156034,
 0.5458213589491773,0.8627352780863218,0.0660848041974671,0.6765713969476547,
 0.002608457388696306,0.621091625502262,0.3076740049635279,
 0.27196199339129623,0.5039062376314571,0.6308143526441327,0.5273759944375557,
 0.840379200123391,0.8729318471096349,0.7512093740176565,0.11165667842112648,
 0.26948435415225436,0.6425908074988959,0.8829127003260889,0.8756012170459021,
 0.21741140492151345,0.1798882296773543,0.03941768689638425,
 0.6500172437461142,0.5653401503558833,0.6258516424785362,0.4585069965756049,
 0.25383353926489727,0.4032885009273498,0.70610541539671,0.5019495016349914,
 0.6070155228617322,0.8580427345481,0.007990644382828815,0.9638530741666825,
 0.035213744387111,0.8553318899799307,0.06804917335897054,0.7251930548252845,
 0.7356484585056302,0.6779313636057238,0.741307287633182,0.8169632630068833,
 0.8910225961783506,0.7725588138706059,0.9445250958761757,0.33261117134084695,
 0.40529101916719745,0.9512854010367618,0.2817527587583992,
 0.023101705837866193,0.7996979417024084,0.684326313383483,0.7482446186324966,
 0.28015982520988447,0.9226368581169822,0.5873216003709238,0.3361444303731762,
 0.20649431740444413,0.885071464831109,0.3216834463875796,0.9438136009434502,
 0.6943035054956097,0.8634942784893322,0.02619433621506062,0.6265535633053564,
 0.325672434377937,0.17593847070058877,0.699283651018944,
 0.0037672666342619676,0.9997245904711284,0.20419077796843477,
 0.051394591716179217,0.2101850334039097,0.5229897440548047,
 0.5510150379138785,0.7918448918199482,0.5093879328134736,0.3138813326578793,
 0.9188808414011891,0.1264555051552969,0.9717273547333451,0.6070505906729957,
 0.7892300394111187,0.3579576599152896,0.7428056612498531,0.16270650590211888,
 0.9541912275873464,0.37352475465787827,0.24702436139449724,
 0.8710810123203011,0.882697613578874,0.5604207354704854,0.891783495401828,
 0.6211729462242791,0.299257452730917,0.6803883255403199,0.4693649675477668,
 0.6422312104932008,0.01934593417509456,0.45603510434050587,
 0.7618576566543127,0.8221486689652961,0.9879651143238086,0.5729915109659827,
 0.17731269987293208,0.8714740322398257,0.6182344626469617,0.9216280125315477,
 0.6514579673406588,0.6530843028039034,0.7223671977423501,0.40939581347552256,
 0.1665724737215546,0.06527664992771798,0.3556981405028047,0.5892521719555343,
 0.9239526758526359,0.4229979732473298,0.99329486525815,0.8261666078247574,
 0.5518711720251869,0.49787382150978354,0.9801013174796088,0.29137412920522,
 0.8139715074730669,0.08501917511993407,0.7355354037379052,0.6822167652825119,
 0.6263610121333817,0.8821642706369698,0.34581631136823554,0.9221304152260901,
 0.6208218908641755,0.6338098481061244,0.08893385796490094,
 0.13341827940443274,0.7227407406590137,0.4830776444174304,0.2937379501597063,
 0.4628099857057538,0.4781814359970985,0.955638529442543,0.04360514955552908,
 0.7211495175869654,0.04079364246738315,0.30755160993833985,
 0.12187640802289212,0.8609479932474411,0.5679351223318259,0.9340214980457955,
 0.6583925217518545,0.5005374143396453,0.4933990358455127,0.5582610291143117,
 0.9067515474444461,0.2842202595119099,0.5532384159590605,0.758257846081568,
 0.32459876754532835,0.5666326231594977,0.25211679910006135,
 0.6059115263549893,0.11704534011551637,0.8325321613972202,0.9333300844713216,
 0.29288013652318423,0.6255689846020831,0.29381407166011797,
 0.25222090464029157,0.2826033882689011,0.04758590148406561,
 0.019523949085506365,0.6610827351143818,0.4453638386597072,
 0.02975046550665572,0.16371234439181093,0.6265528968504688,
 0.46569544002543395,0.02003796320250628,0.4370286373749698,
 0.8965434341840776,0.27945764329740885,0.7946598231082889,0.6506628445340279,
 0.04878093652840865,0.3776869079895424,0.6473802595335983,0.7044045447559698,
 0.26381015418568343,0.9637630951600176,0.7724495989030757,
 0.48874439976653006,0.00424600455763835,0.11427145778442382,
 0.21839330401015145,0.2903519336721996,0.9683425807761195,0.8870903284864601,
 0.8095169152090493,0.41479465157520456,0.6557143918823565,
 0.30046164163360245,0.46954084565663523,0.03198669166126211,
 0.11986253129994895,0.6804087881743692,0.11632520761647414,
 0.8224591850100987,0.060160992465971286,0.5684824349528532,
 0.0141951598627581,0.6062108317145233,0.2863588981408848,0.8760814374240827,
 0.5076302934959899,0.060316431815789495,0.2482671595426157,
 0.2001540960235917,0.31113240734027714,0.8668865731722363,
 0.32883822913102656,0.6306401936281486,0.590232280380649,0.803805805409451,
 0.3828927320349459,0.2253060611661455,0.5818997579837776,0.9416320925892326,
 0.5075417487497398,0.2670371377542303,0.12966757143203778,
 0.36234549275091144,0.03185030579085946,0.40616631427470873,
 0.9167111228126646,0.7808253373517986,0.4612571072786722,0.6391776519245091,
 0.06860573094432731,0.66614949715191,0.08942033065197286,0.3766209133494023,
 0.0629530453442837,0.9176941796994523,0.45959081297646687,0.7074943686835593,
 0.9407617466402282,0.10154419490538025,0.6519149784158736,0.2941516716015412,
 0.21284343131724703,0.7377745212186367,0.8268760848759421,0.413458405583391,
 0.5759542648327249,0.6522807813387583,0.47871711503303394,0.3986182268761139,
 0.9071177990792485,0.521192257513757,0.4278348576020504,0.18318476973675568,
 0.6183097153140911,0.22887536457693136,0.5002156973865266,
 0.12468840721030383,0.25284349324231503,0.03788628325275023,
 0.471681585617031,0.7904066513392419,0.777539041457126,0.22797802330136385,
 0.9782698028184639,0.04155488729109691,0.6686287888721462,
 0.23341981224121489,0.45229529749502273,0.12353604471923707,
 0.22078522104829512,0.4694697175720107,0.15377911584194282,
 0.5866829420091101,0.5827633674714219,0.7470880196253761,0.5214068616439065,
 0.06930161076919505,0.1179444376993547,0.6435027644135419,0.5658755836635866,
 0.5748238870786568,0.16928452155523988,0.8504291864296425,0.7019126337027591,
 0.4305776436289166,0.7209847979741094,0.5249899480584593,0.8409920785183488,
 0.9007146636735591,0.5702001662179003,0.702741648609625,0.026234981694381432,
 0.30402400079861347,0.9046829773676156,0.7910156180938498,
 0.43863664801172364,0.9856694075536514,0.008498214846562657,
 0.2885409249264843,0.9347335335319391,0.6387706098215057,0.3643283671960569,
 0.657515561028312,0.2571513344061407,0.6589166917993057,0.9121456522722835,
 0.35653683401816993,0.7873742926857994,0.16933081182315213,
 0.8842927455749274,0.7847201048677074,0.2422233410879039,0.6310178432442721,
 0.6087625394916645,0.47469732945980025,0.9914357507778659,0.9633086419904309,
 0.10515125228453437,0.5118089703892019,0.8279517678305718,0.7952650357136443,
 0.5533650618193735,0.15041198340955408,0.8625539719933866,
 0.07004266614527033,0.10187030436545885,0.284435427663715,0.7716322180871401,
 0.7681205721194085,0.8765839736457609,0.24395528341953387,0.4489841972315297,
 0.8353274916333608,0.8842194512240382,0.6466202381293962,0.23057662350857122,
 0.7505271101609184,0.962742637054429,0.6923020846470718,0.2811836493207299,
 0.5250409907614522,0.42974674463502294,0.9573703560044619,0.8078241437308159,
 0.17183156788600995,0.631906938386303,0.7572198988255865,0.7817890796673317,
 0.6719936392090763,0.9117976155540145,0.15287223629008695,0.5903821569934308,
 0.08748643309283066,0.6016698493691577,0.3560455443340673,
 0.41035325882859186,0.8333432247250457,0.9250325651408601,0.2429220097012572,
 0.8341927997243932,0.5293186806917325,0.6544970201030317,0.6918247643577702,
 0.6353329077577035,0.632456830892245,0.5888437492608395,0.21033412308497001,
 0.9345404238163557,0.6716905951985772,0.3937211571393261,0.07283048174503526,
 0.9483384941088839,0.5704627194275695,0.6540796020648011,0.6712398367847618,
 0.08223248529631322,0.5658734510461129,0.24182751625156917,
 0.2616948599433381,0.3999156944140442,0.048036125412281194,0.799239822052275,
 0.5197824483084131,0.5646760856870258,0.36003692897115724,0.5130199329426788,
 0.07814209875125344,0.8481818460200091,0.6673848464509644,0.9086587048876369,
 0.44522141981839625,0.044841594452752265,0.32344141900165646,
 0.25837130287307375,0.8022019145676647,0.7180170142609806,
 0.17386635970373954,0.03555772490635134,0.19798724537898815,
 0.02761571029838017,0.9998938345694132,0.3086827452108102,0.7205909806433269,
 0.6460824431146355,0.25201179664951334,0.8898180270148606,
 0.27384932544000384,0.2190610438799505,0.8885151854991861,0.2496602834070314,
 0.26250074528553824,0.34138919772589926,0.26119794352324865,
 0.4585462969111115,0.09041890127620267,0.9045252920923783,
 0.07345124304771011,0.6229372303670759,0.7049151384313613,0.8415391915286847,
 0.5427889598548308,0.6243705999049763,0.7301495762424036,0.5454529483967702,
 0.1069005378388369,0.7922477008045508,0.09480526559525415,
 0.07662421639638195,0.8172281676112457,0.9237241512840316,0.9448221095231412,
 0.9017089162511513,0.8129389980134369,0.8804843314273738,0.4191376886446785,
 0.7193491539195693,0.7591066326201077,0.9255640698944695,0.35582183891153596,
 0.089639603388869,0.4930788161036155,0.28051967802036026,0.04728276960063149,
 0.3810361181088725,0.5907653526725076,0.13323129037505732,0.1680326787708234,
 0.36129132212827286,0.0713877745656546,0.35835861536269753,
 0.5878422601893881,0.7554656353731293,0.661938857571135,0.42898034947159447,
 0.8510072041581389,0.060401078414220266,0.13111389459844347,
 0.3789258183504046,0.760014849294803,0.591423944718876,0.11873613383782977,
 0.14747479450687595,0.3806604777969108,0.39356044670057355,
 0.12376182221302623,0.3534056239411759,0.06875038391654498,
 0.6909957659740008,0.24599157183723042,0.7846382014622488,0.6296301870892397,
 0.19108103275886723,0.217188549344566,0.5922969825050427,0.28903950979617243,
 0.2758054636765771,0.841075583715195,0.6746569092834382,0.7154716973934656,
 0.5345340226026944,0.04266374848646126,0.15245376393291898,
 0.7724708766785486,0.07470795181124612,0.5928394683983366,0.7854938296141519,
 0.3182655333200217,0.5262333042965154,0.21066947567039596,
 0.26218449529244126,0.8787566977745298,0.6905122627958133,0.6798090058224763,
 0.02887039677728631,0.9532898606326483,0.7048201469582216,0.7808460731813941,
 0.3055457850922929,0.2785004261219801,0.5388781497742066,0.8419162520027187,
 0.618735168287751,0.84325261021876,0.2419570972554016,0.637352693104789,
 0.23202182082092593,0.34654716102090144,0.8438181785175352,
 0.2689470937885048,0.38519818481190704,0.38798426972265754,
 0.5341551071849007,0.47310290296604507,0.35208459117476754,
 0.2983122729943293,0.41634553422465104,0.8509986864077165,0.9359117075850567,
 0.23925971772986598,0.7436835438076146,0.51838435748442,0.6029480907972503,
 0.8122145417467938,0.35925804276620077,0.33020534924299694,
 0.6747174295785232,0.37082546055958865,0.6503280654385972,0.2448114266573881,
 0.14181509762485356,0.022921831657094605,0.15844943245725374,
 0.5215711332572058,0.050632820998767314,0.328013162037051,
 0.35284823018414146,0.29520895986870554,0.778955283157815,0.2338464291433513,
 0.6897850274632674,0.6808872355223635,0.8080638705785886,0.2392723250532448,
 0.956430937360981,0.03933502817748724,0.47306667714198725,0.4908501697520805,
 0.952084061746382,0.7934880024109032,0.9741512680649878,0.22728864725201492,
 0.7805862057454424,0.949536251318413,0.02412029013118966,0.0780352655704144,
 0.17944380527352166,0.5486635075717072,0.6680322553288203,0.5839323612232477,
 0.8131765355059599,0.8735319119133076,0.21249285586561917,0.7947095969342168,
 0.4981892924753567,0.8463677511179004,0.002003157846172221,
 0.5999996303178291,0.5742254473044799,0.19211030013395503,0.9228652776572289,
 0.7443709849348392,0.704593886178742,0.8418003444150784,0.8430034885506021,
 0.7629459499816893,0.8286993217669832,0.484986261518069,0.2804528695541527,
 0.545208758727083,0.514378582614211,0.7762918633045773,0.7242969246522664,
 0.15401412696018424,0.8588217249655614,0.42950745642193044,
 0.6721482711759075,0.14317626738400813,0.8901046298805897,0.6256891915242748,
 0.8631290632789349,0.7210444431146872,0.6030473280865546,0.7862514307962265,
 0.25221931959862853,0.27127089074010646,0.6729963476121581,
 0.6605882975068043,0.47917065821319316,0.9785057626594609,0.5218294686943099,
 0.03787930367715353,0.16142113020341942,0.9938467198535778,
 0.6869034237002258,0.5653652643961266,0.28759072359322935,0.958131938348382,
 0.9990493109665836,0.7007811397688106,0.3537072807327617,0.7973117617917028,
 0.8426030834869396,0.7455546941835993,0.46385700224342785,0.3157818817087995,
 0.9378158199509329,0.03242894233136551,0.648824470177418,0.748778693292045,
 0.9596085928555961,0.660148670562053,0.3452571778077529,0.4116382788308861,
 0.754770879054959,0.1998674055156916,0.8780034119169591,0.3622974624720461,
 0.4197289647005912,0.34257719146428356,0.9511361889866407,
 0.16148325650742523,0.32470567184102306,0.5564488652989747,
 0.7621947164480949,0.06645430362635119,0.4516457189369275,0.813268910804,
 0.41769060896162713,0.456414732955401,0.140959453881066,0.2986558286400415,
 0.08446211470937182,0.32193042323693544,0.9900741653664681,0.925837851573848,
 0.8518711671673852,0.9018354416775831,0.6878341905326233,0.8090458806897557,
 0.14047167324542498,0.09810569196457941,0.4280722147169086,
 0.03553928634320025,0.8165397713847493,0.7933885886317393,0.4289054849758237,
 0.5991890830842475,0.7498417613427143,0.4991676194288355,0.4299762108172519,
 0.7233081326851277,0.332307346801063,0.8574960196664749,0.35025310915131025,
 0.685554162151439,0.3545990822369255,0.951497198853878,0.6925383670155056,
 0.6484822453983575,0.7304557959490627,0.872521608565969,0.9817836553923134,
 0.41286040044266314,0.31319323722605585,0.14992770578486103,
 0.12986589379845137,0.11903346348341692,0.6381234782134437,
 0.6023228095704393,0.6770571051856261,0.19241399923967506,0.5222463255928921,
 0.6201383466662629,0.6955583187252578,0.18429936558059268,0.6855496666628864,
 0.25872085140289725,0.9574094569394082,0.8310856036985876,0.622004948073119,
 0.6437970076821432,0.5493145241163095,0.11965089752512625,0.7932645206753829,
 0.39329272509821744,0.27236552386222357,0.29888163188087224,
 0.7535197587034258,0.5733199357060041,0.5159592219390752,0.07393496591113535,
 0.7351438875707845,0.7882897705887113,0.5298511420376576,0.8331600673659579,
 0.1241312381598535,0.8968820831056639,0.32962366950044886,
 0.013726806528941538,0.4043192170061254,0.011639688680202265,
 0.9132040274609446,0.020176266242609375,0.9655177189477085,
 0.5171335922763434,0.9515634966252143,0.5012983197319167,0.784996202540912,
 0.38503748404007176,0.5888034151440215,0.47956197092597563,0.974683516572317,
 0.9005469256744248,0.4638210260238995,0.49554619408536527,0.906852673708451,
 0.23672312174693544,0.7333321382402542,0.6494574022168585,
 0.027430186357213504,0.9682303913763055,0.7216217900972781,
 0.7436509443283331,0.3353584269690463,0.5734314609576767,
 0.020862937269788717,0.5136422465072237,0.08442715159477854,
 0.12006318496710344,0.3958819580567424,0.21872786666628163,
 0.7724681088823325,0.7181301775626912,0.239387771956021,0.5518209178772571,
 0.9949202119168132,0.3258621168762227,0.5370978800226519,0.868772241901922,
 0.8746098595199956,0.623232078924786,0.21594922931619986,
 0.0027351833569960737,0.5960208739660208,0.4535114825519826,
 0.37778735778629513,0.38734818967573736,0.6688662977959728,
 0.011371909498524069,0.4088690910603284,0.1900337578247464,
 0.15499437300399344,0.308515066888292,0.684046745377344,0.44229098855057936,
 0.3450593312279191,0.570127792566374,0.7103615676357289,0.8882863421527487,
 0.31868342827217166,0.838250208347116,0.5464844068248559,0.8766742115523247,
 0.704431711561055,0.5638213525951806,0.9588840881691224,0.5302654545639522,
 0.7608176220813307,0.9670498701955612,0.31888863627071007,0.8092682276105391,
 0.750276644173154,0.3572350145616301,0.3941651266834617,0.8911929197134074,
 0.903121667097958,0.30452357656219375,0.3626250173627321,0.6883732749782581,
 0.6878026176706761,0.5863326382060436,0.2269807686056775,0.3626951143496824,
 0.005495124246609251,0.010027864522516694,0.6206448741951347,
 0.3673414969669161,0.6896397568943323,0.9502267468914823,0.7038508660735123,
 0.5920192490536166,0.8842431013948109,0.9415125034461479,0.5351762983863072,
 0.6800940415845907,0.3227557428135238,0.5929917268771562,0.5310706237852433,
 0.8511731082484787,0.32077796972491346,0.62588280737135,0.9827965858584604,
 0.8011735393900187,0.2637200820628648,0.3490705956330863,0.563552380506832,
 0.5850296012357946,0.5606049821853017,0.8552611166031068,0.7286279673273596,
 0.9881069262439828,0.7671452190327317,0.6017907431328369,0.5109080179498853,
 0.2615857147829125,0.2624335877515479,0.6231973895484009,0.7178925596835214,
 0.06680048235215585,0.06790802501362814,0.9912638867829574,0.479981212981464,
 0.6115257958457216,0.40613227735832813,0.5229181545777686,
 0.38794737724039274,0.9070146096731476,0.6682371082353347,0.876911190591229,
 0.8355712724784452,0.1352590645249513,0.814350233237981,0.7457394265038626,
 0.5475126023451163,0.408356152227211,0.7146403565159967,0.7526151077782364,
 0.320533993971963,0.7548218578114264,0.49515110224419756,0.6090515377799108,
 0.6695678774427027,0.24608393377007232,0.6264445264273973,0.1939984559955389,
 0.16213084045690784,0.23887225250534916,0.8366369212337397,0.661719457935788,
 0.8320747501372576,0.43126802870273173,0.13546754257996751,0.444181936099761,
 0.6201868703791593,0.4061557332123612,0.9893339576956559,0.312865910888531,
 0.2079289944410242,0.4143913439760115,0.409316707797034,0.3529453201231555,
 0.8499487539358992,0.5657865951762564,0.8628912778766852,0.34802117935354093,
 0.2487406494469283,0.3024666890246237,0.12170696874105902,0.7510405388265505,
 0.05974159075463348,0.6391291399267257,0.5690556374338753,0.7539342955811716,
 0.7812327221160407,0.6817582664502035,0.5063650764704989,0.16160175016248768,
 0.8972050791547509,0.32324672999731385,0.042193486455583074,
 0.37647394529190226,0.46085044229513594,0.02480324284172175,
 0.9907336401637041,0.3612270517667784,0.054316110860607614,
 0.7409651068090307,0.3523560878648905,0.21246658819461117,0.400287970242325,
 0.014408188393887689,0.5706467965019105,0.840561528077723,0.960434098748192,
 0.44693497130923043,0.5127500507943464,0.2514164046083063,0.728814893723975,
 0.31324396394700094,0.2333535172237302,0.7577556692313375,0.8243291576624627,
 0.9170130315059761,0.9483962817046563,0.020149299330390003,
 0.8999308913548195,0.8693633968839255,0.4922722058985175,0.1124810915773714,
 0.9548350268369485,0.5991202930125586,0.6823859538454112,0.9260621084847428,
 0.7413898809577388,0.3044163487730841,0.5297720970506246,0.8056458685340061,
 0.678330258734246,0.2091611728042927,0.9203834878727924,0.8240687534034221,
 0.7618927490919261,0.5476508519870495,0.9144764213168962,0.6220524048598326,
 0.9372016527981878,0.30501544167673367,0.8855981871454215,
 0.011268803586551357,0.49428398571347754,0.5492909020797585,
 0.4055217112327987,0.8543673552984946,0.03628521758678993,0.5948499724699484,
 0.9639647241570735,0.39365853519376537,0.4904287335802786,
 0.031894538185293086,0.566016827732834,0.9063336872903788,0.7138618668287415,
 0.8705127308131531,0.3445937891597175,0.7508923817905822,0.7074928595538718,
 0.8037541502997554,0.33631566541778446,0.48528154026557624,
 0.6158910859624452,0.686646109080153,0.007214731213448067,0.5952962235390462,
 0.8240294171591085,0.8788761664034451,0.4425981974159291,0.24191051335608882,
 0.4777924227955841,0.2369913560010536,0.47806762300488026,
 5.496993322477994e-4,0.37333680860429796,0.8525201505700064,
 0.24304677823903842,0.408869708901314,0.29364677104254444,0.8550341537448303,
 0.4997117458094372,0.2866684355193465,0.31317549885765583,
 0.34518635065672143,0.6022335959251941,0.5866904257648927,0.9403287915128732,
 0.03574396259616264,0.3181896533873132,0.9133430201330078,
 0.31503705920012426,0.32719037841619636,0.022220770854961214,
 0.5436493589887308,0.747323789127526,0.7994865164261044,0.22204456009295337,
 0.5393141120796416,0.6956973841535641,0.9138960548369106,0.36735627612880695,
 0.3880214525258322,0.23585125642772098,0.08290578740338028,
 0.03411143957598439,0.8430977913868545,0.7679983655551191,
 0.10285395284856558,0.785402043170483,0.9164608777657219,0.6673761837478084,
 0.38510910385354324,0.363730443316128,0.4398071151829719,0.2731892589048849,
 0.36166494292749274,0.5884674603575795,0.2932598060290357,0.9099800941746684,
 0.29519804343171796,0.986870134046322,0.9577743197959082,0.8005474463289528,
 0.32730979404785754,0.4739656144094183,0.49109101606997385,
 0.2085995002757982,0.7533387168702816,0.47585394084987165,0.6613183339207034,
 0.5883519076184057,0.9571426393926876,0.9639980716629065,0.43966087830385614,
 0.36491833597387613,0.16809590396207763,0.13064099270499108,
 0.19426641111073772,0.9913211205532908,0.9457597381265349,
 0.04978922950624454,0.8227120072005927,0.653223262418021,0.206922440701528,
 0.3729788481558832,0.2040917320707536,0.3944879652273553,0.4445436492469561,
 0.6847217964988452,0.06087448200930845,0.884546162534169,
 0.017327885506228657,0.24462864256609707,0.8200288419022694,
 0.9561907345255796,0.8322150795318096,0.5697803476392765,0.3294651468706379,
 0.5589874481741249,0.06362416792747805,0.3751408343399082,0.9647189557038373,
 0.5296574157563642,0.3097834691322302,0.9560391977220436,0.4272870022335393,
 0.8313078919585417,0.29182435484230584,0.6690885247869693,0.8774870225946733,
 0.5457414301789528,0.3038908795827052,0.08604800515698707,0.3415069186105496,
 0.6598730822942364,0.5072869463614234,0.5490399802455183,0.6960197327731324,
 0.9787912939075412,0.27965814696708535,0.6504475754960543,
 0.30830903505166385,0.5175831994200977,0.5456628885305037,0.3539493045457772,
 0.3202814190191162,0.9780047716237303,0.6498964794452258,0.5986050427499217,
 0.16601193560767658,0.9037331395485544,0.3917561350882942,
 0.12180113569769113,0.7357615569522852,0.16925212890090946,
 0.05379809148068293,0.4161602239828084,0.6931911776991326,0.6375643509763238,
 0.4430573336400243,0.07729548822074828,0.9541037788165765,0.9123000787876179,
 0.3403623054198003,0.5073620529604231,0.056562572025215196,
 0.19297923014611318,0.6718426404459142,0.8387578974663862,
 0.24164077059989186,0.48279205790306046,0.39217172633163266,
 0.6776069919800081,0.5265523158942155,0.22198307515902438,
 0.40730948004517387,0.7346496302695414,0.5599698856247433,0.9107468128851376,
 0.9352785783134674,0.2592782723832001,0.5662955803247057,0.4678323495976038,
 0.9852783626892663,0.589821102769474,0.8413063455478621,0.11122707080335115,
 0.4387810926173846,0.5217823611362623,0.8399389000212273,0.6280810700953582,
 0.7935299101300329,0.9442113360003893,0.998546795388017,0.691276996282832,
 0.7846366546545154,0.6563116926592076,0.6345034621650711,0.3626889850340066,
 0.883258647634784,0.9064331555182346,0.6070756850899671,0.14647843315220432,
 0.03679410355721924,0.7316772835357171,0.789387624575402,0.08651524081053541,
 0.7263467219522438,0.0886193163987622,0.8762591591701651,0.2618907126728922,
 0.5320252029740477,0.915281760245447,0.8514070496372103,0.028603422485307894,
 0.7899314837016167,0.8403875552444233,0.23648844198930963,0.9227026008714677,
 0.7881719297394114,0.6442355253094946,0.7651867023051366,0.40771005889381506,
 0.7758745175339461,0.2643524693829714,0.743930633498423,0.3363366202167519,
 0.9320185890947899,0.3085856100430491,0.318854307373049,0.08637908105324899,
 0.8255894868532534,0.9383634972224179,0.40694550003170193,
 0.15976718146181312,0.11507718596803762,0.0821676539301065,
 0.06489162416847549,0.5431509782862363,0.32915815238192714,
 0.42924437625758116,0.03880997611533876,0.32173825229049335,
 0.22630567800489554,0.19346992780240369,0.4312734568418767,
 0.8390382026032204,0.926588404147307,0.3389416441787435,0.4421378221033684,
 0.2678402124021326,0.43964196443472137,0.217708368405654,0.432554282293825,
 0.8871801864263648,0.3825159697966616,0.889032845682129,0.867315772102975,
 0.20183611085264386,0.06210174041779559,0.6601865902659487,
 0.2621623244556689,0.6066975887397046,0.884065423409677,0.15585566990696265,
 0.9376387339222372,0.9940693838000947,0.8275682360216431,
 0.055935379682383046,0.8795207937230276,0.12252183090419433,
 0.14093770823871377,0.44871955621927073,0.1380825673392252,
 0.6483359490617436,0.019583774307409874,0.2582661044266217,
 0.31867266438747965,0.9657156349697833,0.9769991648867304,0.1803563394748403,
 0.3349769165927273,0.7488721235860638,0.30943248057468065,0.4400209068411356,
 0.042156083166438485,0.7967347067771718,0.7439075931573453,
 0.09028244712406741,0.18542264844632383,0.20828533187920195,
 0.2572623668660289,0.1848511339836748,0.26177703772220706,0.5145153816039413,
 0.9944709046400041,0.28823878427277416,0.6293077798564363,0.7563809552967045,
 0.6703102086845647,0.7269325754686831,0.8768754029134554,0.18858659868066774,
 0.22587061489626692,0.525767789984951,0.8690247167808899,0.4029901651416632,
 0.21896732407786934,0.8923681491672865,0.36293085709371997,0.788365039964264,
 0.27515342793005304,0.6913980164512492,0.6927561012901103,0.6318307954754635,
 0.8834140085649801,0.20518868850459415,0.05023243974472691,
 0.7456296563777455,0.37839037167109024,0.6851023527950983,0.6887343843669042,
 0.38206060004859,0.5212369548998601,0.05230668893069734,0.5857562443728329,
 0.6567790845254371,0.20869558796429366,0.7447518199847929,0.2021946803768181,
 0.5801477888957974,0.6160333180863401,0.9372143988319223,0.5235555285757836,
 0.9346752994897342,0.008603515913724658,0.6659415296652665,
 0.6129018972570904,0.7737029672334963,0.6151254137203945,0.28054225011921363,
 0.5635249465982516,0.6098372642548969,0.5425342175942072,0.30279074436262887,
 0.5883367767616807,0.5869486483094862,0.36064406092910417,0.9546795426985354,
 0.26456064212860797,0.370009895903044,0.8409025777267635,0.19825804211157305,
 0.7587450083537033,0.2881248868421822,0.9605621648443912,0.4276500585555463,
 0.10767717104153351,0.3019597357003996,0.13092909225380323,
 0.07313886545555293,0.6466274551648374,0.786833369765287,0.9660161769677347,
 0.10472538758026817,0.1721989362924733,0.14159350393471448,
 0.6916385655310449,0.38659123442362575,0.8819321708698656,
 0.14909047014202925,0.8891708398023868,0.6836294887149471,
 0.07941983553751553,0.5569552153980845,0.4098117031290006,
 0.43181371964546755,0.8883638361591164,0.8284625850337973,0.5570014147434695,
 0.5884651113508913,0.19792187082542279,0.1875051666545513,0.9475447245404163,
 0.3947050885366288,0.41663808343289965,0.2196035132997325,
 0.25067621000263285,0.5663150225344212,0.5037501403489563,0.8598998304664052,
 0.1743897099900027,0.18369521199601235,0.43799581667214627,0.926765420924267,
 0.4650250901170989,0.8189344124366095,0.911548706249706,0.6413350666663906,
 0.47389668091416237,0.6921865074651152,0.18135548475819163,
 0.5373881346160498,0.25362305635245597,0.6210334502853438,0.354750621669519,
 0.3943102243815484,0.686477992413836,0.36614815462983685,0.927454755121191,
 0.15354213436575137,0.07605965738873566,0.6158232381211949,
 0.8680837543732172,0.8105061311258068,0.09815207099869616,
 0.39899183676121763,0.37065604243672157,0.6127376193634376,
 0.8277519285692021,0.18960543938055463,0.21390678409191077,
 0.4658139045501364,0.09152873641318293,0.4214954121140717,0.4494485640589134,
 0.5232560231598405,0.5654514532188022,0.6944638259900551,0.24463435382285392,
 0.6164498552293732,0.13332394554499538,0.4891810318699914,0.51211896081595,
 0.6620870596849637,0.8791495888715414,0.1760645774981655,0.8983736684494295,
 0.8651308412814078,0.5095948991767867,0.9792778556808052,0.6718502887339505,
 0.6984228133872721,0.5105244148225645,0.9838111569132338,0.5683674595746685,
 0.2413663940547114,0.3643421207415394,0.9751427443873287,0.6038295133194748,
 0.21638643115193368,0.9561109340227772,0.7800455998139604,
 0.020063525868313947,0.8055719513328113,0.6474624539041969,0.589392508250474,
 0.9290834727110127,0.16815877501086962,0.48135269492423205,
 0.5302189780316198,0.6414054149562498,0.7613151954679714,0.4104377414211421,
 0.2865340540041683,0.4596643024777216,0.6178363738700023,0.3232651079202916,
 0.7853662148405703,0.7248615558055489,0.8471989543372743,0.5426304320787987,
 0.1287665079325786,0.46551670644730003,0.39830610307997383,
 0.6799933338735143,0.20390325703050916,0.6574194313874933,0.7654888768194882,
 0.10616277910603611,0.07592674265380528,0.09453287802342392,
 0.3485649298255059,0.538460886738018,0.008895802592734037,0.9040621374567489,
 0.23411634472002574,0.2963522493479739,0.4407312707460107,0.9915291248786461,
 0.42593419387911413,0.926790541790254,0.3145911753364752,0.909240531356525,
 0.32709591471539223,0.5126012547132565,0.8432919841314988,0.6860538524560319,
 0.3495492517787444,0.09662864840435215,0.6332387852692704,0.457541102717713,
 0.19709219538736134,0.46122532588706866,0.5141961431900834,
 0.13296369663737662,0.34099476330301526,0.009895758638787111,
 0.8386488745625663,0.6441088382375759,0.444988084679223,0.7630212128997017,
 0.3715034711352604,0.6356869935746596,0.3305831159387027,0.10095425331136487,
 0.6087434288669006,0.15222744884996597,0.9852831778770835,0.4464027931244653,
 0.5028609357852745,0.6648925530205497,0.9974580100275894,0.8659442937512559,
 0.20573777647469016,0.04046295076912698,0.5767123543879173,
 0.7834018795894699,0.21576884861404177,0.905216932943153,0.3610101164139792,
 0.2547691337002732,0.7327438699486796,0.843397284401315,0.24406568812671003,
 0.682875778045293,0.10409445404931505,0.7260689042852921,0.22186266689596001,
 0.9261498458842209,0.19105559193929644,0.6830495196821275,
 0.46996303654732585,0.4116655405905316,0.6772828886746821,0.9491159939312641,
 0.9977054725966219,0.6781248570921968,0.8576120864562325,0.4433711294054815,
 0.33125250462604894,0.43677466700529743,0.677114722711015,
 0.26938343136024967,0.6019153482900054,0.6384323843441209,0.397607607041631,
 0.6472656512424017,0.8244608570962473,0.4622664116978842,0.8462720294613842,
 0.6355885416531263,0.25709264892870076,0.88926177155049,0.21445874684367539,
 0.8329395140641475,0.131659018183094,0.43062636671455223,0.6582094716781408,
 0.5881974596158557,0.812010867322635,0.19718774454365062,0.9731985399116341,
 0.4224749263943215,0.6725048744881703,0.5917838566914339,0.9713647742283162,
 0.7432212527446455,0.637830838403539,0.42578202807219323,
 0.035092154918862195,0.7766083281692785,0.17258068295588358,
 0.057576720515555846,0.4347430028705931,0.671669514546757,0.6079321868501566,
 0.5362699296989374,0.6254993628420495,0.3539881193420258,0.349488548597753,
 0.032390355732995535,0.3845745390111084,0.378930483159138,0.6442107050716124,
 0.8868661308035153,0.8080272414651791,0.30801514740711766,0.7050561922019418,
 0.12727244463535137,0.7235514979083698,0.36325888443261023,
 0.13052900936213097,0.7441437163911271,0.08661615101787501,
 0.3311187814425768,0.9791228794987326,0.5497205979656269,0.26744959476254204,
 0.03545502204361728,0.5534768610398465,0.6356563313475432,0.5611361836634768,
 0.08337955695167576,0.5768314042201341,0.7918555456011552,
 0.001549813014814938,0.49868339086283553,0.23919041629695692,
 0.632182724791895,0.5093547232411059,0.9349842301332213,0.6488660546886138,
 0.6489004342196941,0.6378569608548015,0.07180828376838544,0.2300617238528706,
 0.8202544248142029,0.6281384581337417,0.5719478684641113,0.15800447093554226,
 0.6052009744359874,0.33374574729906636,0.3786734786705538,0.6854376489438788,
 0.32106629722140756,0.4009317504411939,0.6495425304248403,0.9036455434518368,
 0.9964865642365672,0.8580294526948278,0.9500418394316931,0.7177496579145115,
 0.1891263375272736,0.8423574146431274,0.479709242764755,0.09695092913019021,
 0.4055663422752497,0.04240129108325741,0.9222061554276606,
 0.004122264093581962,0.9516662394789357,0.7941214935036985,
 0.8645071186090046,0.10382603583541217,0.8162873785190244,
 0.47925736143603825,0.0848564265810865,0.07734325407219389,
 0.9553143415744569,0.5916517960389271,0.30981599011368866,
 0.43419838251182985,0.130855899280082,0.4377372473650416,0.6434589275830348,
 0.21062555094460855,0.6493559100075805,0.14412150009805746,
 0.5500710839750879,0.6370426894655071,0.395335637142489,0.03283186721283049,
 0.5366650719530295,0.5368863329619697,0.5257675036551539,0.680630578688474,
 0.0803117549786242,0.7356729864632696,0.36133830682262824,0.4371673876536152,
 0.44930879026180304,0.9093402427527988,0.14018011687687237,
 0.3477683156704794,0.8535145528000093,0.015952846393664122,
 0.36330965520258485,0.0035630111755545757,0.2150075582445602,
 0.4100429188958178,0.1975327023667658,0.6816420654532399,0.30275777996337816,
 0.4903959434324212,0.11096488603715293,0.574887091596424,0.9539781501501982,
 0.5592198207190929,0.4841924610032061,0.7091970055198805,0.2648903380745756,
 0.05161195709019961,0.9833939304327687,0.7383744794121866,
 0.33030959850941266,0.4582740114417301,0.3751823272187994,0.9024956211839655,
 0.38415961054165326,0.32098653497117147,0.687360864641832,
 0.08375844682281774,0.26547906315749814,0.9150560241670231,
 0.35954089170985215,0.45993738539760187,0.1455635413569586,
 7.150651339822289e-4,0.8984202988594875,0.9592749427016584,
 0.33345586033202357,0.9194735382036743,0.7821727218812922,0.4733617654520311,
 0.0730933987053789,0.6548623719670842,0.2942976876078076,0.2691684844763934,
 0.032753113828213665,0.5628697464107336,0.7095124904087814,
 0.07560444550255863,0.8088663164439962,0.8100158443114197,0.7290602900689644,
 0.7968124505631198,0.22090481485904667,0.28103809272590596,
 0.4642358412973602,0.3221267402829424,0.28954919083183306,0.8877465341834828,
 0.8494975906726656,0.5422159912150115,0.10639445029074435,0.2701147285345913,
 0.8162504861508595,0.29574713428478505,0.9500388336805278,
 0.17870990955331856,0.045460446725603454,0.8110981752431379,
 0.09623194691901094,0.9345457368339867,0.11275154414328825,
 0.6499473775560692,0.8184288223519667,0.5197008251883629,0.27299550819772744,
 0.8032511674261813,0.6550062683401761,0.8703560036092619,0.55773554922219,
 0.789111300937975,0.8401108918877991,0.434418033183953,0.3646429562301292,
 0.3573360887026582,0.530008959355261,0.5903802831692276,0.5623306762717032,
 0.22346740736839488,0.33297053992794867,0.3957207332450471,
 0.9654268735675658,0.9977654845247023,0.8338728669584917,0.8841593374324791,
 0.3298205595029413,0.6059244624706861,0.5170800916388824,0.491341337914428,
 0.332829173479972,0.9739616223611279,0.8347898195792584,0.7672739809626913,
 0.05310770639167106,0.5701111796182313,0.3690314565896269,0.7491773735046448,
 0.9071316797123576,0.8434451221200081,0.39651051181819874,0.8619552476265577,
 0.9532358409126722,0.21614012488478151,0.03769816121986391,
 0.4474125944965873,0.013746472832902201,0.9081199758969569,0.962095365960327,
 0.5498036613990579,0.8525463208217626,0.9432986493425805,0.9524038921775887,
 0.04981461437453738,0.25889241247953554,0.8743100903040789,
 0.059578023685224246,0.5703430924357599,0.3853829822973165,
 0.5603631294352455,0.3730746431132529,0.06359952971020988,0.8109600428567375,
 0.5213274467087127,0.5448499315783422,0.17012731074679932,
 0.07175501789603478,0.3846492912686178,0.48005306414222126,
 0.24717081076220604,0.5087982602623369,0.976134815701163,0.481081877457598,
 0.06997654898268313,0.6003787144503905,0.3019667126643938,0.5709110945254338,
 0.3007856019327697,0.9454140554242365,0.46548771607235395,0.7318282297627584,
 0.45296658294615977,0.38335447286162716,0.9864665490298165,
 0.08153394702737549,0.3318528706583638,0.3867271884395759,
 0.13223540244145682,0.14842150066881576,0.9188942611954264,
 0.3264469424807859,0.9048334626472018,0.2868068840163839,0.4726679295251902,
 0.5757224222010696,0.9103574896866341,0.5293273330508789,0.8283067909442053,
 0.7255396312553749,0.08099447169255414,0.9364386495557633,0.8193926110556076,
 0.005389992038149161,0.7910037279801929,0.15742578156607223,
 0.20549424546122708,0.7454644153682688,0.9041578517601637,0.8964919475548343,
 0.5289155195689339,0.6482597619782765,0.13673078591272003,0.8081568599763559,
 0.8381058029311131,0.05796208783817791,0.8861309219043249,0.5713367344262493,
 0.05899478131940239,0.3205940692883066,0.5056970922583315,0.8433198375823469,
 0.36732307131327013,0.6516372623048928,0.0315927022712833,0.8877827250741775,
 0.43541418249656116,0.563274819335992,0.9624266404245361,0.23006416170431598,
 0.6465422452025058,0.3037029405982252,0.8962924158154788,
 0.031127710213388782,0.20678469064156024,0.8092286959948568,
 0.9683080910072945,0.647810159100702,0.328849393674835,0.2932223247769794,
 0.06608278991597016,0.3954092517719562,0.8921950337231233,
 0.27003380865892535,0.5109132858418765,0.33957358194846343,
 0.7834000677291197,0.5937376281046709,0.8181500570553739,0.3533933108745738,
 0.7756799914572573,0.24019348215630854,0.5698753861855748,
 0.05063885215776298,0.3295359306001493,0.035553212458975136,
 0.5388130585410614,0.6557409646656931,0.3560560734321163,0.23808498696491665,
 0.8462117835448745,0.843030241613679,0.23862795509230073,0.15237562416980344,
 0.2939922142970177,0.6462658011706417,0.882879470111447,0.7108479647775383,
 0.7386008646754587,0.4034665068719172,0.9170131408255076,0.7539442852463605,
 0.3928763976482267,0.6937979996516864,0.10914220802130181,
 0.29113604915713154,0.893779464348057,0.2347563914397447,0.6027361219343693,
 0.8030580477313999,0.2870969814462828,0.19128903738314462,0.7920270091114654,
 0.24277967750852114,0.22261933449616256,0.6356095922588798,
 0.4695777786040949,0.784079029597187,0.26887558186356475,0.6342360850518564,
 0.033644405486440765,0.19219392282495884,0.784893039528338,
 0.1790176478953699,0.8680172302891519,0.4514259291338777,0.4445433373945272,
 0.5011693381990403,0.34912086353484706,0.9017836729389132,0.3373638683566169,
 0.08186482532444306,0.7815964427342486,0.013912655239658234,
 0.3042549764811142,0.5166023839703734,0.39297780916655034,
 0.20196434473492564,0.04471375413300793,0.05925714186789066,
 0.3128852083260404,0.4518951252276514,0.5531379306351267,0.6446257284531159,
 0.2812981929391488,0.4138440692373929,0.596834397737331,0.6392940899170653,
 0.8980871020936362,0.7211038061859996,0.7676945327333424,0.8882932006096502,
 0.09481809197199065,0.042076819883164585,0.749660304256208,
 0.15246475919773728,0.9353490343552878,0.9976781758917139,0.6392576498032819,
 0.7524267660648905,0.08188736518972461,0.8421893842572431,0.5858023434029482,
 0.6711126316844185,0.971991970666391,0.7028565979759226,0.22696521218207988,
 0.08554794889708717,0.8470762411114459,0.5699452183647313,0.6306998632554596,
 0.31838014071197906,0.004622740844303808,0.8297526030835708,
 0.8052501376290301,0.23055189599877002,0.2718993142751924,0.1648312189779907,
 0.24397171918899685,0.08087967549848762,0.06600594897239564,
 0.7127386685986894,0.06408730539027852,0.8172313519567646,0.849646949352369,
 0.7290183563088588,0.7024317412595946,0.7379347846450265,0.21951854017894146,
 0.07351187775349499,0.3159443293987162,0.3603348362532369,
 0.27034538480148795,0.8509102096243221,0.7032325584222403,
 0.39332606735732956,0.592951460047904,0.6817758748505157,0.80611723961549,
 0.4456226069689223,0.40485098760621685,0.5615563387535955,
 0.005071297979992728,0.6933892963985478,0.142537415683765,
 0.36055280118520805,0.9348192057304391,0.2356225676866146,0.9485810201213766,
 0.602112472456365,0.2840000201534273,0.5227027407476077,0.5320923060656205,
 0.516047645692441,0.1599343145214075,0.2237452131762543,0.5721787367193667,
 0.6626551724221743,0.9609575758270409,0.41141352160712363,0.5089654269801049,
 0.4593846238807301,0.1571054663877517,0.8611812830728331,0.15879515635177077,
 0.18467098346605337,0.3859032736062966,0.052698580642112525,
 0.8614022849543916,0.3372700865677448,0.1411478297511337,0.10649214392107265,
 0.410870202600804,0.9193373532916198,0.8066509170038343,0.5027610978416549,
 0.6857636547829357,0.531026503821406,0.5728685929773861,0.22976606446842418,
 0.6418657605353,0.8440503262454564,0.8305025477110325,0.6746169226258673,
 0.6780092477024326,0.6905576716434007,0.27196655610253073,0.3286737289756544,
 0.5603234024546307,0.90835000914824,0.29177981983255574,0.352717363820197,
 0.9453198886187454,0.16731023230135533,0.7960946859497829,0.3244551107692686,
 0.7404757231010421,0.2543048528240006,0.5319945971463934,0.15014463709469972,
 0.0628750495749818,0.6401851098743739,0.5185928294773904,0.3283341421549081,
 0.6761331415879394,0.37190686647903215,0.9183332823997895,0.366865358029587,
 0.847843735412469,0.6723888793361373,0.6468670794155692,0.6076041269658408,
 0.1840370628097786,0.09698602263898692,0.14626729428893215,0.538923389561819,
 0.47916547057457515,0.7430361167985065,0.0829197470541102,0.9395964984422214,
 0.354861713925423,0.470747835433303,0.44302392866755613,0.4062663635799809,
 0.3375515645001024,0.3166856430308914,0.08014384706946398,0.4806195228529607,
 0.06060388910489556,0.21934411544808996,0.34213186527461226,
 0.07804354612602604,0.10714643935548063,0.4472162604386437,
 0.4671464508849318,0.7191724042002898,0.8451396565020213,0.28264113655389045,
 0.1904404707526638,0.5167280673786702,0.09627661231005535,0.4774373543525048,
 0.988178022745458,0.042436573486354545,0.9997430037570577,
 0.47201311916249156,0.7279033077996836,0.12651481345184878,
 0.6816578961418177,0.18684285112937082,0.30787600753715216,
 0.07832810189526995,0.3313074564688603,0.7327755845883049,0.5716001455760753,
 0.39703615862011643,0.4914903383859186,0.09140697342933435,
 0.7557917896892663,0.22345477088441246,0.7020805482773821,0.5941661549861565,
 0.2628471697336805,0.4143597889418664,0.056600779447455896,
 0.5242232601090594,0.5786779519967153,0.6543911952236114,0.30985600517982215,
 0.006890930179972443,0.536806544682998,0.09027032341257424,
 0.11998474678443083,0.5082628425689384,0.78439221329874,0.517476998578824,
 0.3604722003430385,0.7387458942986508,0.06967770153843067,
 0.16404026440732178,0.6110207960085836,0.5005989163688438,
 0.19449199459915167,0.7999552858387701,0.49658137220292375,
 0.5771732691560092,0.8107363347468575,0.6529884251705049,0.8597193538627032,
 0.8835169746021233,0.6004172512096718,0.8072998663057861,0.23841575044207253,
 0.2615693683922399,0.4225015478926023,0.6149171110009694,0.2434905468060199,
 0.6268497605559811,0.5612690799373592,0.06540480545809968,0.7147776083940445,
 0.6405082155665689,0.2090735786903013,0.4972043083294734,0.6820723736229529,
 0.9432505398385383,0.8766320498461091,0.4440260234105923,0.5706989864953091,
 0.25939510142884525,0.4125563951567718,0.2372730407588144,0.6203462872528472,
 0.6789157295905716,0.916619782295974,0.8620291337743278,0.8252936851289895,
 0.4674516464575116,0.5470527434550635,0.8589417819394132,0.24147120143680756,
 0.7628698472864629,0.07471513196863888,0.6346627017746337,0.9463208514369298,
 0.4060885619543333,0.41805799323472126,0.35518517158178553,
 0.28695140745164005,0.44697741390521295,0.37754234041350065,
 0.7264767566761954,0.7928095360123109,0.5313888640561297,0.2791134765355574,
 0.9781453595791187,0.9027816205626753,0.060291848993935804,
 0.3302724921613043,0.9879636758004828,0.7267757820641142,0.03812756219294111,
 0.3946358390875365,0.9450945656225811,0.5874386266600893,0.5904335560506269,
 0.23363690376942536,0.064091563250253,0.42663321807858823,0.9169252534964426,
 0.95288241292279,0.3898775641508947,0.6207433740876095,0.6726199648906704,
 0.22829049390420486,0.2020326625451737,0.8862437497681431,0.4863153475711457,
 0.11191590714200994,0.9702992852075952,0.20449876137217715,
 0.10578638115769878,0.81304798960347,0.26406801584452055,0.7201061607125117,
 0.9776600421940213,0.8876126503246566,0.6416015032269383,0.7458312199334705,
 0.9157394961329531,0.39873374669433936,0.18343035872057978,
 0.9674520736465351,0.4919414304956854,0.4454342206608125,0.34716475989132967,
 0.9670952267729019,0.10421150713172977,0.3218244952082745,0.1920840695002356,
 0.3478316139989832,0.4944284263672962,0.39822197542039106,0.3837986916755399,
 0.6491646779691167,0.710760941978138,0.9351837257784259,0.7763110076500039,
 0.3539514188899433,0.5999040854920661,0.4799460877559999,0.9065942072906138,
 0.5281375857027462,0.462892753568629,0.6332370658632901,0.47102188574118475,
 0.9086473912004924,0.6524976351574197,0.7482627698932186,0.4926764078601359,
 0.32486774040756894,0.759092323073899,0.9523799129713907,0.42472552494456317,
 0.6145489620134522,0.9948885295210194,0.9846634303160005,0.25563125208928683,
 0.3282338904832175,0.014626785134688847,0.878190479441116,
 0.04301923989463774,0.32255262857290423,0.7314571209312559,
 0.34484176697662816,0.3168345585525022,0.943502186589432,
 0.0010114893357867638,0.0800861179438469,0.5758292315308383,
 0.8062037918078325,0.8975845077890547,0.20505955797235242,0.9476787487609744,
 0.9415966998660728,0.22520290437696655,0.9373926491584303,0.7817518108417434,
 0.01401396702097324,0.3800243427685819,0.7016494382771805,0.9135381293226572,
 0.7074828591904518,0.16526493199816839,0.8730690157528807,0.5563176263403379,
 0.8624369580749405,0.9059996800734379,0.7555322271224897,0.14565810518137223,
 0.37597497865796825,0.7643522543716185,0.27703526664466505,
 0.9566833943893015,0.025954635343213073,0.47770717631507953,
 0.2057860536410414,0.3965434661803191,0.26945079166951424,0.6352999175480696,
 0.2032607553662028,0.33292169266511296,0.32246855891415993,
 0.9467996037112179,0.04267522164223847,0.9563194617968418,0.9672450231508946,
 0.4661479876947102,0.19084522933975623,0.5356041944645965,0.6504653839174089,
 0.4803085566646357,0.9368078144888404,0.41208299202007215,0.5950661617627331,
 0.2056377327106398,0.5984877631973334,0.9945456690837841,0.3515844672118782,
 0.45078137057795675,0.5194620436426428,0.2418868400624643,
 0.04090425135976461,0.04810125341364613,0.23728448376347944,
 0.9594098983832806,0.24263677655867688,0.5610833924873602,0.6249506381429382,
 0.21112841719446118,0.08008874037382085,0.4043853506755394,
 0.08074873217042589,0.1380884373579634,0.8600108026223715,0.1318278223963012,
 0.04705115571137064,0.4305816551425111,0.1794411368110932,
 0.38172111385184726,0.5331963572744022,0.5620667179262251,
 0.02376559993026761,0.28890598385085653,0.36783796735945573,
 0.2598617140671279,0.08127722680673544,0.1273471325361385,0.9613976571259465,
 0.025203014472085727,0.5322948768873768,0.890971862577226,0.2028868905654162,
 0.6938838027768062,0.65833384048543,0.10078369247932029,0.4589254830608167,
 0.7441676220506038,0.3845542425015739,0.9419209606453316,0.7500263235424146,
 0.18926535665689936,0.6391453304974747,0.8480702634400458,0.6884149844476263,
 0.07124617988341486,0.2811999571696261,0.8028083191526607,0.3093880275275611,
 0.28278830556291834,0.24973148461525252,0.9283311448944583,
 0.03658516813248114,0.6705156773735479,0.6096457382423838,
 0.39592768870859407,0.7264134354084225,0.402596324057515,0.2786997267017617,
 0.49267117972008656,0.6680247836062266,0.5982711338412356,
 0.15088475486868658,0.930705474926995,0.6104233559796353,0.9020182952477483,
 0.8465580606524611,0.5443581429885196,0.32260251992190747,
 0.27628207377137415,0.8573375795631286,0.16368343259196194,
 0.7682854038121838,0.036208412940744616,0.966120722633411,0.5910809834492029,
 0.657614762781474,0.9120062950859149,0.9532392098938876,0.18885936529981184,
 0.1569832406583751,0.7157059686406816,0.18391313186641467,0.8390152801174939,
 0.7028295015907283,0.11273955663053759,0.12485112858534442,
 0.27702238740034413,0.5951254319175069,0.36866847094079014,
 0.8809556626242863,0.08560029352509868,0.39381698400699094,
 0.0715181593080435,0.7723708447593494,0.669932014856569,0.5036115591387641,
 0.8647952650616592,0.3691095577198602,0.03757538135130778,
 0.43461188379023286,0.4995974794213115,0.7960212878659403,0.7762146944249674,
 0.890633520395368,0.39474330902959076,0.42601173552898897,0.9866889628029984,
 0.7457578829583377,0.5616035046764551,0.15448159219801938,0.8730557197427661,
 0.8728111285118154,0.1897950394626755,0.4617688283849548,0.1691425025965042,
 0.6258381223365115,0.03276046555570611,0.12076832218858136,
 0.5777029142602845,0.5923553778403742,0.1299890609494102,0.22406196445990334,
 0.8402352118847503,0.7014006298955571,0.6341639227722269,0.8215230301425531,
 0.5147888056256894,0.7108411419088289,0.21538583800924882,0.2506120865822157,
 0.6033994599172304,0.8882444097807327,0.14073248479785383,
 0.22220774460677495,0.0015711317624043097,0.3074607304240148,
 0.9044440050316604,0.43879773779622655,0.35007050078258095,
 0.24627156373668624,0.6488416163483366,0.9404001739333903,
 0.21085784130210639,0.4707671535128821,0.6941229312691699,0.3612850139504935,
 0.09528386359117091,0.2659761153253586,0.08674601927180314,
 0.2287849574667965,0.8211311489813031,0.481525396715478,0.3360529493692801,
 0.09337461187493212,0.7615407485733453,0.7167600789794005,
 0.32614517609436977,0.6410636074906524,0.7473248009056601,0.7416781477506247,
 0.2812690058276137,0.9613704398510131,0.17555052067501986,
 0.29925446261977096,0.3687676118761105,0.37273126501646625,
 0.16332687507719623,0.6687698526116389,0.3585316060061021,0.710755978033762,
 0.7632993729291803,0.3729108829988691,0.5971354275980171,0.24663545125656672,
 0.9195513550906591,0.6282873409947994,0.8635875803573585,0.6970020703297973,
 0.6789218120746754,0.20807725747703365,0.5654664076618537,
 0.042264810277973575,0.6586389939978289,0.38446862925149217,
 0.8806734062064719,0.4640593216652378,0.8154015419914994,0.7154291938100689,
 0.5464248281570924,0.8805371592181421,0.16096179004434508,0.8594054033851615,
 0.7123527126569693,0.7336880255360583,0.8272377661720007,0.9221334754552949,
 0.5848546418790278,0.2196640356973778,0.22658307114489828,0.679471044873764,
 0.06166955664597784,0.07060545499425519,0.5955028116768613,
 0.3569057559168539,0.8308803637170467,0.4805717046249741,0.04500759422876366,
 0.2821852750178373,0.07723367564313999,0.0864584528166128,
 0.0023370286923924954,0.9716156959794503,0.12969555337915162,
 0.30827912695313275,0.463969571933203,0.651883687121122,0.40470772981764913,
 0.9342522767025008,0.30718326997870915,0.5516779835829578,
 0.20097600019266515,0.7955694291832079,0.6257992994939526,0.1386343106589809,
 0.539333092447917,0.18172245167883927,0.5684611201936904,0.5451043965664469,
 0.27371665511433174,0.9786576561645046,0.9565398460390508,0.2815310826754893,
 0.0571292920713653,0.7527330800418474,0.014934541997650763,
 0.35326639668896775,0.33236117589751346,0.6174959157194471,
 0.9915276034840911,0.7239063296075984,0.6464002651191015,0.05120068435830272,
 0.6539689085248347,0.47644802050338175,0.4395035635891986,0.2182713095722011,
 0.16466465013726173,0.18845608259338875,0.5818717581938612,
 0.4047452525802203,0.27606119657586436,0.4189114729110117,
 0.18759609709748837,0.14537278278370336,0.13278122335854092,
 0.7339550326748625,0.6467869737937582,0.2998700084785981,0.2906210524011239,
 0.8480764219361274,0.8018811695766828,0.9787899506869521,0.5323256546550097,
 0.3617971031829442,0.30299376768743513,0.1772629058699512,0.7329703909190742,
 0.33168014039728577,0.8184092496144454,0.28500359300273137,
 0.6445377997017003,0.28500808808915734,0.807445278048226,0.3834640904897402,
 0.30237929614583603,0.7860717956900712,0.4886180660286532,0.6931398389674772,
 0.7361301125433719,0.20053117580411106,0.9037826233254157,0.612429225031225,
 0.9051479658637513,0.11339615593056518,0.18351082628200288,
 0.6743292106414114,0.9565279531758537,0.005333401083332134,
 0.8014445741255125,0.3414869489160923,0.40944081036122726,0.1870750511566136,
 0.39135171516391176,0.7024317760995948,0.8795725170172741,
 0.48500815513805007,0.4602154911561438,0.7743196272060169,
 0.21922383528827527,0.5700446748185454,0.0013700543193363446,
 0.5103795067567075,0.05024468203695065,0.3112693849606153,
 0.18275699680096313,0.5891070863874188,0.6863046276058034,0.4587445496869076,
 0.43260846891873483,0.20876316748947443,0.10244830471361221,
 0.32898354821245457,0.9439932041196203,0.406102638350312,0.5295657752898678,
 0.744407697165486,0.5046763850526054,0.11599887532452602,0.7959624037315131,
 0.4235377160049292,0.6581523271056323,0.6289860102047365,0.9828277632984214,
 0.22305816270677603,0.650813322916727,0.9406605959309133,0.613261512906598,
 0.7840840226361316,0.7796023207472593,0.32320492999099837,0.9019802725704518,
 0.6076860297929888,0.20129189670459036,0.7205860263357416,0.1903293883178766,
 0.5818490594408573,0.32337585461769436,0.670345682676859,0.6007513679666426,
 0.6015867010493458,0.8664758848390294,0.9660706318006331,0.6136607950390026,
 0.23880579708229133,0.25347273670928383,0.3162981693609551,
 0.039005948464558515,0.2556792838697203,0.48297739807379614,
 0.9223419906706853,0.30498617981279763,0.47823348673189636,
 0.9305874863938608,0.30847184829361296,0.49252727362619897,
 0.9489052829397283,0.14554951922808057,0.24217671962449805,
 0.4076258128995346,0.04732773766282694,0.4029458234321588,
 0.33958934559117226,0.841156612378402,0.0830953866845241,0.36249185024985164,
 0.7129931798479621,0.1183825330910454,0.5184546150710277,0.07679510016515068,
 0.7027139858404888,0.5160957426140431,0.23526753153652424,
 0.39465131491829586,0.4944936081382145,0.3434862814941456,
 0.023886927996441987,0.7445185628793269,0.896625891180304,0.373050771068425,
 0.7110451950751041,0.7870253466683731,0.04413045050279152,0.7283857579170256,
 0.9537954236946006,0.49343769454707553,0.15811336947501287,
 0.6760811917934453,0.538106174451132,0.9385808210651618,0.10470410644496686,
 0.6960842383421524,0.5051071001700168,0.476246948894554,0.6492020525576555,
 0.1376373021351176,0.09050161893508157,0.8923640319461674,0.8195383268171794,
 0.6603058433967203,0.07539756760924265,0.8645709490208462,0.5494786868230254,
 0.6659636807602843,0.02581127183969434,0.00704912215268283,
 0.24416313647343868,0.8986455317387358,0.40759721774770263,0.561833030992702,
 0.18661982895012763,0.6548615774766329,0.2533939517328401,
 0.37203505389131997,0.348858809193365,0.22413544955839237,0.2982381044929089,
 0.536126650989312,0.74151929132043,0.4960457841037602,0.15719742532120318,
 0.4646293387662085,0.39348915141788654,0.1511566316142512,0.9920121066847953,
 0.45879497453395157,0.16614344086766786,0.7993023070267146,
 0.7030908733061795,0.9012420532230332,0.8506332752666803,0.1468629402164121,
 0.4932291185658315,0.6268001485786908,0.5523119512051794,0.38546505009433074,
 0.916307533516531,0.6493716288913222,0.9170358262855426,0.46965479549501865,
 0.6236322216067598,0.5404943119629971,0.1819622362369293,0.7415575213843789,
 0.30292238553910744,0.6043483254528265,0.8742754593209806,0.4162873524417985,
 0.23155889588821066,0.20051171438523152,0.6369641913824517,
 0.8302507136311484,0.7237684498997877,0.2767194129597619,0.7726130844558191,
 0.005379685496047171,0.7601607087820746,0.1452992181103726,
 0.7996957884503275,0.4377912720642674,0.5780260557181917,0.08275787969823845,
 0.29285427018146903,0.47679250280349716,0.7780233039774078,
 0.9520983537757803,0.8046187059506641,0.3292813045884284,0.821384878581647,
 0.7667824906273518,0.8996951820300856,0.7032648878449965,0.8494870340930323,
 0.677417336481031,0.5382046761221989,0.007640837218178764,0.9238940274170478,
 0.16165461231121403,0.7615609326827311,0.5385321717264324,0.7821799854043686,
 0.6761302481048594,0.4999737283911593,0.9370200878213336,0.49792437572515424,
 0.7713584311957571,0.3191493548700569,0.8453392893715878,0.8053566070134991,
 0.9447143219800171,0.12554953457024998,0.8389456700911375,0.4872874925925458,
 0.323025639537032,0.4028043834994791,0.9804396416391069,0.9492124839556201,
 0.4119255899583196,0.1777330989771725,0.36078695090149415,0.2665434370639289,
 0.19324595829309088,0.5205618088217824,0.7953951805897099,0.9997336287373032,
 0.8939590999223364,0.41873744079051345,0.28501665082282257,
 0.6480274883364414,0.9451846538611283,0.7740476420416563,0.6894718023961827,
 0.09170202758962265,0.4073380007288515,0.871858736390693,0.7735221537167891,
 0.10870444736735174,0.12389651281560554,0.14916726678157644,
 0.599755150783566,0.22712059944328034,0.22149121670065264,
 0.18996927929813945,0.3090488029511994,0.5484485042064478,0.9590526610441988,
 0.23801915866091483,0.7766605548148321,0.04615403272179164,
 0.8047082166771892,0.2866053495872183,0.07523282977994428,0.6602947658997321,
 0.8305865338973716,0.5528286165635768,0.28350053043516,0.5967703103812789,
 0.2096484877403264,0.37082388103911545,0.4912506233585917,
 0.08560790234912075,0.5230984504540218,0.23651338788915488,
 0.3903717457128667,0.9509494982140373,0.9313432777378616,0.9552049419043136,
 0.6050491781304275,0.605993353282572,0.408079215329068,0.9011565273213705,
 0.35270577191563957,0.6062813055173331,0.21160197293880756,
 0.33906394382996996,0.19893599748002067,0.9479179177228048,
 0.3363895471949173,0.25329308254849625,0.880757095516666,0.5665176173715388,
 0.3660841144342989,0.8001111567720147,0.4973662763519806,0.46653111880588105,
 0.7694560458643335,0.04370383208117601,0.7715268259999525,0.9240901570559703,
 0.6625847718759227,0.262363620999548,0.09843580638537353,0.7276228356618698,
 0.8280368105368501,0.15955779618946442,0.8795434374119769,
 0.010557184653055551,0.7614706220407624,0.6218519806347144,
 0.9665696695244278,0.23454093266996223,0.6930515050631576,
 0.23487795829376057,0.9841954842119374,0.14088124488071063,
 0.22856019029725438,0.5845498969256524,0.7747918078983409,0.5113978437592613,
 0.7859482103129474,0.844932925003347,0.13646069209697587,
 0.033523602906087024,0.18092229809210303,0.8217216479621038,
 0.4132865117217327,0.24531191891814175,0.4582468996609229,0.3730731429906867,
 0.3299356188602659,0.30987890094150705,0.42935174377001784,
 0.28931601196737633,0.7000292052217106,0.031585700251152926,
 0.24652998468645504,0.3470779885699554,0.9304536160612845,
 0.43381976867106076,0.5385222296161555,0.5694623272134384,
 0.38283866731398486,0.13457601839609046,0.2821856628924946,
 0.3571324928248487,0.657461690613737,0.4710038968049677,0.24509603326881868,
 0.5181472536212859,0.789593808074685,0.6791547837376524,0.4516710373853394,
 0.908912962760118,0.4809155924211411,0.6046321914109548,0.12505626660515778,
 0.8614002244067788,0.5348769267475788,0.26478593567942366,
 0.11307164117863289,0.8071892692099006,0.5226213516742104,0.8743369608371047,
 0.2955676730210862,0.1307265749397114,0.8494968494010583,0.9422501412008846,
 0.6877056890732985,0.5135776716872419,0.9525781798542776,0.9651858745978669,
 0.024681763115956046,0.32226542315320517,0.3698309323379547,
 0.8668043073910182,0.5454404815637048,0.3204653318857298,0.36264656182758004,
 0.9931592828569161,0.11851937305791804,0.8831153533411215,0.5529503100653507,
 0.03386199796829026,0.8642746521759566,0.48872077509286294,
 0.8456723255042994,0.3586605807772968,0.6253530627175669,0.7109732693454327,
 0.9146693289812312,0.33886575912086603,0.9842058903063131,
 0.07995828151240547,0.9387493817839744,0.8521715624212796,
 0.33049794342384187,0.2498376573774448,0.7433784221924626,
 0.27874964982126604,0.8816337909002385,0.4111336388094806,0.639802438890733,
 0.2665588226201112,0.9555405370043349,0.820501643301179,0.8512551585878401,
 0.6095250061446187,0.6532833972142956,0.8215632857168932,0.7138117076432061,
 0.828405479424768,0.5761160637441268,0.31562444507151854,0.4161621209320272,
 0.944887885083458,0.9163759079371666,0.5844942901883253,0.5271070147559576,
 0.22301575861598988,0.21177650653031443,0.4871263874203525,0.930158142732435,
 0.5347470405920794,0.4790240636943378,0.3209521833203669,0.572014754104554,
 0.2853773587734081,0.3915786991610213,0.9901527483983482,0.20061882960824517,
 0.47440817080826125,0.7452463763226791,0.1427516244293534,
 0.047325636865512055,0.9257912171391018,0.23746858993477193,
 0.03690417880480401,0.5065773537616329,0.38784608859644587,
 0.7618634463918912,0.37285832630154714,0.4139934386758136,0.2641214696994454,
 0.9460856320441876,0.011986697624633003,0.9058082541048199,
 0.7704307989631206,0.9922479542417063,0.647636198505738,0.6791330067356842,
 0.03290004900527976,0.5127308246252966,0.3421161850983636,0.5497895577376269,
 0.7066183135646096,0.5567793463709522,0.3111731782432159,0.6447746467825711,
 0.27930223345688265,0.4685952550878768,0.14839822336495379,
 0.5052649616963759,0.06260582258511826,0.4272098555009807,0.8966208977920958,
 0.24103997049749393,0.929744646904808,0.9240013573619754,0.8407972966892387,
 0.6116751915256379,0.702935758926722,0.7488380917411184,0.5322354770511425,
 0.1961604298957904,0.1335225466272899,0.3198865936944837,0.2181564483209758,
 0.010652718211874634,0.8569787710188678,0.5009601652298984,
 0.09380684187934796,0.20760988621340093,0.8981366103140308,
 0.9326118161962498,0.5859417991653068,0.6834970004222929,
 0.060451231265015504,0.11549814220793198,0.641626774396741,
 0.8899221641077592,0.6214417207535073,0.4514880373757102,0.8336505224636331,
 0.4530990387230439,0.3818877131469012,0.8327265173165463,0.3718893425156733,
 0.8218091971511738,0.3125144624254066,0.8562050332644466,0.30161726951449996,
 0.6913687772058862,0.5227259068126704,0.23743701643061454,
 0.024615159772103823,0.018764578213857463,0.45527298741392686,
 0.2334166777159049,0.12454063497257573,0.40776483227582905,
 0.39438363235633556,0.8385414290248941,0.05246379988148453,
 0.3982897567678192,0.4113115389825941,0.15647087965924122,0.8900545336737913,
 0.9419037417139098,0.784457818601394,0.3995253044954371,0.7728449781720852,
 0.437209993799874,0.35832893989708214,0.9428734544308491,0.8244517200269035,
 0.7976179349699682,0.5355630060279519,0.19604720547855092,
 0.28223622923381075,0.6607097666161303,0.42004895203075376,
 0.1423355640338665,0.9507204049331777,0.6021651056006141,0.6253851559693996,
 0.9195169005651777,0.5243191723160063,0.5054794471994546,0.6331050573458642,
 0.65707000680606,0.21422742387582727,0.5944587427849825,0.9036779027695001,
 0.8007868914181158,0.30434960949046497,0.04592280084342515,
 0.40116585274399597,0.988630609860503,0.09900871880349316,0.5631268535695653,
 0.40619867969869294,0.5611770532603746,0.5804592552005863,0.8913072859860154,
 0.878839404688033,0.4557681539822004,0.3486123367830317,0.845781667213056,
 0.3473637724701134,0.6255957381101515,0.5634494388101206,0.3771739458701293,
 0.8359850495840142,0.05572762831532385,0.21348515511443056,
 0.08194274342572982,0.14202726458781822,0.33350462132403547,
 0.5578685018195794,0.05089841661584371,0.8819055090531287,0.6618947033270922,
 0.48024356381707334,0.47357515172911513,0.1422828450522522,0.522443568876303,
 0.4122747602955993,0.8308915394212864,0.2491645362710191,0.4703766990280164,
 0.977053473735775,0.07643128446070002,0.0033044580362034948,
 0.7925148839881176,0.015517338575999462,0.6392371192368501,
 0.23084042575316288,0.8160528120981873,0.4603493809784561,0.8791695664711299,
 0.3821796120249532,0.8762759940520546,0.016033449060919414,
 0.6568686497193453,0.3907743650841934,0.7321613038447068,0.6807635866059984,
 0.8728122837719196,0.32146064352327786,0.4653031752353366,
 0.11420699310486981,0.5730263058021168,0.39912099531466005,
 0.2399240562994227,0.22483092934694304,0.3114198812042931,0.9218021365569834,
 0.2660564162208374,0.7191539054484312,0.5155541629867565,
 0.009047922726945101,0.7981182905189952,0.4430209808469061,
 0.7081051411089836,0.2548895992962158,0.7259525023154347,0.64027704300369,
 0.8034166239599397,0.7173528510156781,0.48727433829694966,0.8344686873513737,
 0.23795891488442922,0.4489460442846137,0.3346563832670817,0.3346175078895762,
 0.7086519979844942,0.02406723955950596,0.23168082087138697,
 0.33011481996727177,0.3322198042027178,0.8572169723300568,
 0.011314310366393299,0.7286685963018407,0.7299130416877274,
 0.24744111667153978,0.17271397385558285,0.7948784000956766,
 0.24287670915179416,0.4856149136109802,0.5485696292005938,0.3218352007604329,
 0.8170134521754948,0.3160992414843645,0.2660374389849315,0.2771221153929968,
 0.7085801543716683,0.8427741855671613,0.911978126174603,0.5917966427461271,
 0.26687197488755054,0.2990321490788692,0.8221008515108609,0.8182115871489766,
 0.060092815372244424,0.8935163313699394,0.8235113842853983,
 0.1088561309409255,0.7152755278343852,0.24837322258216277,
 0.47472163312220994,0.04284002034982326,0.280786571834321,
 0.013706990758673987,0.5502740332497089,0.629721604367282,0.9391407233351199,
 0.3600121713752549,0.21693074673215207,0.46247704309899507,
 0.3807400238134405,0.45769402727447406,0.05060967086522805,
 0.5079157624426515,0.6216814634671991,0.423459931837184,0.3871220853043478,
 0.41124872282213387,0.7711278407193506,0.4429312037038452,0.5225124256958276,
 0.9939337962938802,0.21361677013368863,0.4926315815446133,0.3276192192960278,
 0.8237208184096677,0.7425778499589024,0.6595668252447222,0.9989601349539611,
 0.5408356808950853,0.6431388909310778,0.6056158841851801,0.11150679061833269,
 0.1344258373031989,0.799398836058932,0.024571951322666985,
 0.45614450684269414,0.48353592208173624,0.6013152424048402,
 0.7130323741250534,0.8850154416406741,0.32230720493948606,0.6106731338109139,
 0.9581829115968383,0.6470022173909196,0.17301587240364025,0.846102708235591,
 0.2933115507744055,0.707694743419323,0.9187399660832435,0.49215361217690246,
 0.39255526397705043,0.9226768781537269,0.1519437410967749,0.9815259587467898,
 0.18079720544741518,0.44756251594684304,0.2529811486978899,
 0.4620300584496194,0.9874607876941839,0.925654137700263,0.7705664091868503,
 0.8583842952756696,0.9330798039744692,0.45643804201876337,
 0.42581730463827616,0.6622973239038504,0.8158174790250757,0.595164531039492,
 0.1551362571826802,0.9787803815366789,0.8577755274940874,0.5513982757884778,
 0.8155053510574252,0.26447192282230225,0.7553808443966844,
 0.06780755356150392,0.7933133054123829,0.5831620261124498,0.3391026939420255,
 0.006337562126494389,0.7007266112901818,0.10289945942123546,
 0.6310321274726133,0.043218675554207664,0.3264260635872023,
 0.5208617511644635,0.34647717628866215,0.37720691944550133,
 0.22381703825911614,0.12616854761993268,0.6876215035595815,
 0.4091365421743457,0.8713099396588737,0.04142531560328799,0.6086918750632011,
 0.04158818511651774,0.07605038525435537,0.36615147401607806,
 0.5524676016347647,0.7390041681562541,0.6432865288717805,0.8446183043178326,
 0.8951049145683895,0.6364705227594077,0.39826214721592756,
 0.11598273291136996,0.7877513042628548,0.5202051686151579,
 0.023818153018865806,0.874947617328406,0.6927841788177853,0.9227677665258155,
 0.9401453862410062,0.7341994266195856,0.9170216059947734,0.10670043610335167,
 0.11198418222438633,0.7483432741549721,0.371755720341165,0.8111537294578868,
 0.2526239397483667,0.07791821084864003,0.7209509629986569,0.1610290118261728,
 0.9434819937913071,0.33114129592490293,0.9678106552700855,0.1345988996277473,
 0.5190729350732984,0.33863921252943596,0.5676833906436793,
 0.12236268270243977,0.39995447898059144,0.23249227479470458,
 0.5942954514563864,0.7895459079525436,0.7204627642639052,0.13265718728352394,
 0.12084829453785029,0.4207751553164851,0.44570173142711456,
 0.8709471168160217,0.05670261928008835,0.7627678747158292,
 0.25718081355716893,0.31431321771358733,0.3925591849327411,
 0.19808317603984826,0.9236217900173809,0.14927512760942718,
 0.8214488378600555,0.36845769229778136,0.5323438096651024,0.7646073149036459,
 0.5269242082004595,0.18432736541691497,0.9030693783914677,
 0.12695942765434254,0.2955124722792468,0.23816952823875515,
 0.2860476918705055,0.17644539499368506,0.843769924316054,0.760282576685994,
 0.08292398448006966,0.636513227031751,0.8405121089465094,0.12290020116675582,
 0.9769437792549069,0.06852786387342236,0.07381984578911538,
 0.2910433241760194,0.7305886999682745,0.2585899556680745,0.957372138236409,
 0.9759068272808826,0.8630116783233173,0.6658860022267162,0.26006993597709915,
 0.7508627747742813,0.8582233737442627,0.7888120052447642,0.558835740697092,
 0.845013379737875,0.7760920289843053,0.8765694369229498,0.1696330396105773,
 0.7084420110107925,0.7974056309770378,0.8506535668552605,0.507808384131376,
 0.6301083360636903,0.11111338035160989,0.8877447532734661,0.8125462020120415,
 0.2075404618128207,0.8348565778604543,0.28815496035843713,0.8205989596019146,
 0.0433949268866034,0.01667854454393325,0.49100678635532036,
 0.22176705919088402,0.3474242368829962,0.4118141530203506,0.3799162656027333,
 0.441390679814877,0.6226373465203305,0.9439193026296205,0.09815443292012327,
 0.9993086351187797,0.6088136008707489,0.682916786098066,0.054803079550882705,
 0.9338197565831821,0.6324743243596516,0.42336742525871374,
 0.49240766740899666,0.9442738001985822,0.021365870923342634,
 0.022247505019672875,0.07332370752014128,0.8981138075845407,
 0.679115552672712,0.8265258939432463,0.9861893731979096,0.33561025978121284,
 0.38723951319921435,0.2838700480763603,0.8844324735739798,0.8276868556502481,
 0.5853166175076024,0.18141381443831062,0.6875589113652836,0.8384034003784347,
 0.05976668069860891,0.33908442527252824,0.38435893421029754,
 0.4445947048523454,0.9268486123814139,0.048135713556638704,0.52901555294369,
 0.33925306674213895,0.9171663709422193,0.5190381497641599,0.9328145193332816,
 0.41998927976155975,0.2917038976825561,0.6116227529908888,0.6473064546073578,
 0.5421021644219025,0.7241945845231447,0.3489797945284485,0.9158804722689271,
 0.8065615371602679,0.9672715141537688,0.2553750008860668,0.11508943299227314,
 0.9464608942741912,0.4114159461110559,0.8274179006982091,0.14820325024096326,
 0.623360936507918,0.810393535358572,0.312079712098257,0.8370196242809582,
 0.91307528537705,0.528972942403556,0.5892831452390063,0.3584120814027507,
 0.07209054901066003,0.49242955428591617,0.49378841180841593,
 0.566044890908666,0.9168481877908026,0.14234486310289518,0.7503353685930556,
 0.2853716872192503,0.050545884390323925,0.7897852289371416,
 0.6924375084809907,0.29408998884800297,0.8018825958656842,0.8212765397322127,
 0.03541281661852769,0.9944380571228144,0.6689961091955351,
 0.45905682916670454,0.10875246671464478,0.883569543124424,0.9101437424251829,
 0.1906589975086872,0.2649565968177554,0.23278881901589066,0.9685705287031393,
 0.13869866379271734,0.8517393160484893,0.9108077954482656,0.3065704783403884,
 0.6208538097900785,0.06484428134625309,0.03454002896216857,
 0.34095086141364084,0.7588166829903045,0.19050602034322572,
 0.7012422579490882,0.37811824223135604,0.9375464025939634,0.9349595148819723,
 0.8310904987553407,0.06619843389161939,0.9822039855817076,0.5329226423657494,
 0.9817769477354781,0.22470447734319565,0.19846388575772567,
 0.11785729153518298,0.9166504793060789,0.5439516892783863,0.8752617267527036,
 0.05845444723310067,0.5264687624698593,0.07959147295114355,
 0.8579622052272134,0.3314059071007509,0.7476054681729506,0.31116696625719187,
 0.5408764198901359,0.052759563819556377,0.643454925635363,
 0.14230477239867456,0.4307852180600157,0.666229983219468,0.22401175648927563,
 0.06052178347948389,0.8273548441972876,0.40486904429273685,
 0.34795017378450377,0.06687431636522945,0.5748155597449508,
 0.7297837071960898,0.9879585566048843,0.022159621720911682,
 0.06327151982448165,0.3731851796618041,0.9043510501807318,0.9808847152168072,
 0.09631105576713417,0.5969949224155477,0.7927041576561173,0.4496128633578518,
 0.366218024453951,0.7942332991466473,0.43916059186071643,0.9799567131812048,
 0.19151392517186117,0.4760474032532238,0.39576076010309247,
 0.9126273623607409,0.996991275523863,0.3079048503538805,0.1000048566073608,
 0.36212908571465385,0.14700794427373332,0.25945579235444727,
 0.6327893920858082,0.09611218938817767,0.7969687943340763,0.7189693832478751,
 0.013259416682734448,0.022164711586244112,0.3121626239272626,
 0.5815715096009005,0.6144019961464752,0.8231124865795869,0.4240057791548164,
 0.45832861250595935,0.6505932836677899,0.20150778369157907,
 0.6314352613269669,0.20586616112362333,0.4313771966310752,0.6790811647175772,
 0.9809744816463942,0.10814741297761776,0.27021281634580185,
 0.6039269014039824,0.03461009534579329,0.25844804800255106,
 0.3563007120515478,0.8224275808996242,0.5836312353564994,0.24495620320837452,
 0.12320903618989931,0.8831626982738909,0.7255260516773302,0.613123511040171,
 0.4195505766227414,0.7813786373318484,0.11240369726316601,0.830901093872968,
 0.05057076210497735,0.6157818777986366,0.5657084137320747,
 0.17433238018877106,0.34134151497647003,0.00937870661853557,
 0.25169139058146106,0.9318014580115769,0.1950232002119594,
 0.17157637413120153,0.4374651912522096,0.48990371299858526,
 0.9842447699712572,0.5586531341356753,0.5316343906286339,0.7541099834487052,
 0.16349132776257014,0.19140225543368505,0.9608760562353379,
 0.47149311145119654,0.9270181295142433,0.8368234555628927,0.859517173831225,
 0.9472457624071396,0.47579532922219614,0.5803418780067098,
 0.010111508095155819,0.9274979019356299,0.21418563533146784,
 0.5215680730249097,0.0020457980025407885,0.9728872337093203,
 0.13451993369812032,0.031359677287297605,0.20908846063202324,
 0.8456812930324282,0.5674924414174867,0.19578277678948308,0.771528557360109,
 0.6112775717491764,0.16702644349792883,0.3649024775132794,0.4477404835912153,
 0.2464113340774059,0.11629009473048235,0.41445990944527134,
 0.48814431947090453,0.7379475647381686,0.8494736786918732,0.5291167712537653,
 0.8361199666543346,0.0619589518965612,0.9306644986888581,0.23115276911634774,
 0.9563477099724399,0.28750624805111724,0.5986268694516018,0.7327282502871403,
 0.4451582223902505,0.7166792088939866,0.6908310364325122,0.4401256500409749,
 0.3545501590737695,0.5004746076329584,0.48513221441939813,
 0.19990745859118897,0.030843807289825276,0.5721621861206879,
 0.9169721675521482,0.7317397063170216,0.7184490976738619,0.8756850978268955,
 0.12982867365989614,0.18046391774948323,0.358452353686492,
 0.19511302296279087,0.2884698556444213,0.5810708905366024,
 0.05559978695522427,0.8837915328129414,0.022555465290301613,
 0.5088203176041209,0.3792452330610776,0.28740809515582466,0.3323978774046117,
 0.497905354246597,0.02671303330723762,0.5592769041382557,0.742168170748958,
 0.0654153630548211,0.04217264552373767,0.15638441813156345,
 0.12932168623995377,0.15744539342832553,0.5912733004325287,
 0.9386529011827633,0.5913819018614613,0.8296778529266919,0.6936539111551825,
 0.8889048006481576,0.7551007153070718,0.9352397271864844,0.1980034322007117,
 0.12791452463713648,0.03799263433997235,0.7718807638872772,
 0.4391045280250355,0.5737851743965734,0.31311052867748146,0.9782271527126627,
 0.615789968088168,0.20028314317060514,0.9749092751762689,0.8078919399055752,
 0.16496271200631552,0.47868645069901306,0.5938611345412971,
 0.9395137423621268,0.23790351040435442,0.08638528553499647,
 0.09123036198614864,0.6199071784113314,0.4622280279118992,
 0.03074283310092607,0.7264951387262857,0.1098160248371921,
 0.33009900715822094,0.15346650789248217,0.6699717557849301,
 0.07606811066123564,0.9255051035940881,0.307175319777778,0.8116776849294635,
 0.4159954627265028,0.9865190176232705,0.07124643984106194,
 0.18894436353313504,0.11314483675284814,0.5637041234773162,
 0.48751595697597205,0.30965001153244776,0.663097443893967,0.6615222551116609,
 0.7888319221733203,0.7306661964621947,0.08065050939322915,
 0.15857010216468137,0.94361630040238,0.23678670729360263,0.03107498998020264,
 0.602810410889742,0.7564198946494759,0.5286905182443472,0.45829930998663637,
 0.9986711831867454,0.20703488643918422,0.8769574613727351,0.2157310057690044,
 0.4426782686355265,0.9177825197499524,0.3553200228600155,0.6714898917748928,
 0.2503862365695473,0.7188150377254799,0.9437514021005731,0.2934433949297264,
 0.6425604590948405,0.6569023161572547,0.22221895633389777,0.5785388253993361,
 0.4146413156014638,0.7636042062779609,0.11907657638368718,
 0.40063652662430405,0.5412610075453874,0.2088972098455154,
 0.12446185991928094,0.6855123249389046,0.2361568279213926,0.4768556672474701,
 0.6947550349737418,0.07699849073060128,0.30968448542981786,
 0.7250506457346783,0.5480652648403959,0.04966692504665038,0.2146716378406115,
 0.17734686756935303,0.07623756349691313,0.8055750378439693,
 0.5314996721908855,0.29992465329901263,0.7596686896323336,0.8020439307213055,
 0.9284207943486917,0.45562920728312295,0.8744672441435563,0.933950819229267,
 0.17588614371687716,0.06747618737960903,0.20435364043160364,
 0.42247339169866216,0.40482802977737753,0.42335911674225823,
 0.9667607352352569,0.9361351424604258,0.49380480340551813,0.6901263524697533,
 0.7555976832795711,0.4727415468133842,0.009146394328621188,0.674818620391426,
 0.5903837624689615,0.11998773238301497,0.18143066151014287,
 0.011630024902613112,0.04367184640141242,0.8630830273035368,
 0.09591743755980764,0.8904580295423511,0.34161074037058603,
 0.5324004318909099,0.889188252734618,0.60705140442283,0.5334585151911089,
 0.7517778807610193,0.2756630986461286,0.6801539601046379,0.2869443497296309,
 0.9118069159376546,0.5408670790155371,0.5026194033108117,0.49387787047561127,
 0.5409612553180565,0.46178598388442094,0.3954577457290661,0.4600469340601223,
 0.8238665518056127,0.3999570248413289,0.5148931178363164,0.7600610768696254,
 0.5397671044874016,0.6108242440805692,0.9381812532026207,0.8065712448458052,
 0.16897480028342993,0.35187495467232643,0.03252287792403552,
 0.6760976867050331,0.9092073778572464,0.027236218477867613,
 0.36780800776557376,0.2721155279929721,0.6001218241141097,
 0.19093174015025205,0.7100544445192873,0.731683953940455,
 2.769832294158281e-4,0.5021010355298389,0.4718491899326125,
 0.7378661145532778,0.44294450519697126,0.9881418539759189,
 0.25883410285765274,0.9056348408410435,0.5365363402513127,
 0.36725078148198786,0.35792193903157377,0.46209673233830084,
 0.2595132356292632,0.024585173674500194,0.15231834503537456,
 0.03920689118835052,0.2989444086537726,0.5128279924080983,
 0.16511779168670138,0.36987308688787035,0.8083250442712862,
 0.4420069332425184,0.3557010001780627,0.6213316424324966,0.3597659241239509,
 0.25163142488278667,0.6274883652756498,0.5278813830586887,0.9707199663143565,
 0.787947986872621,0.6247424494657837,0.9586720051693324,0.45501174492538077,
 0.5612480491126663,0.2316663278724801,0.5439052917543121,
 0.004487897077264491,0.9070435765199963,0.11739107572048679,
 0.37386953097098086,0.24122383065203934,0.4512823504199347,
 0.45499761771598846,0.25565360895665656,0.5326834212121405,
 0.0385433253098868,0.42285117392799876,0.5382366810014472,0.9527983110853877,
 0.2700582541084119,0.2989439311453348,0.505820158682204,0.22385787021512504,
 0.0622442798126559,0.34072663461312425,0.26171809621720366,0.631480153487905,
 0.8073957311659832,0.25938534825388126,0.0335178023106506,0.3158056298287115,
 0.34520616671581716,0.5027923124013087,0.7608935375786903,0.9215044239581832,
 0.7719058779251602,0.8486600015192151,0.811757495494817,0.8623935355299415,
 0.22102161350689897,0.8740232980035758,0.3645196441078993,0.2586121547706991,
 0.3123403305654323,0.018494415848838153,0.7879182396458968,
 0.2902389592480479,0.15308841414039365,0.9390258831511981,0.9901962316350746,
 0.3632478324822306,0.4080045241456165,0.39807576847042936,0.5009370893664982,
 0.45858352710932637,0.13827534770709127,0.025593226874985886,
 0.21692650831538407,0.2791285712197139,0.7825848538797586,
 0.18037101234603636,0.15429480509820737,0.16858227471125375,
 0.7366309216416425,0.5703070114203245,0.9279976189019038,0.748158954305152,
 0.3408261258536638,0.5178068917443535,0.8952129738948775,0.660970903837656,
 0.47064748089367536,0.5614504637245876,0.264267771244002,0.09730968388273764,
 0.25466957472638196,0.46443815588466053,0.7976308613876957,
 0.17954421411576493,0.9100367705094093,0.041728415409768904,
 0.7938710052958533,0.09332223921551308,0.8585229534974642,0.6448767288009054,
 0.42898077315696603,0.004096617862239027,0.7644599308461789,
 0.7446432221060978,0.32583213862637017,0.7588279476147672,0.5151603715935481,
 0.5799397596173728,0.5825205096221463,0.6141242413728558,0.7503678671348467,
 0.1830605757046312,0.014573450558061696,0.02531497400514149,
 0.489758039182542,0.8506129017551635,0.03703828783740293,
 0.007387813720047154,0.5057642959541393,0.5974487245604619,
 0.6801684000388585,0.5607946743614112,0.4632135971783522,0.329917375139533,
 0.38764195384943156,0.7360868406090311,0.0719619893626906,0.6995881416352945,
 0.8031806785288704,0.17184247148381682,0.941091088257914,0.298881871104254,
 0.02454090913116136,0.2792872880159385,0.1349200951858056,
 0.17854759522246233,0.023147643402241402,0.8242594036393575,
 0.430378247059816,0.42110089722522037,0.3342708990222082,0.8935180209361104,
 0.02786216037079503,0.14277044960163687,0.8096854658991497,
 0.9284550457076812,0.8840724254847881,0.9912814045405769,
 0.033589131900736646,0.38418342241424597,0.4692190140409772,
 0.07442290144065522,0.2617967899284208,0.1391042372699448,
 0.046547122012103426,0.8302002027412306,0.7391509166594421,
 0.13937513602659013,0.8480243342964797,0.12123044450236853,
 0.22373725556060453,0.6417850986821807,0.7420321362733293,
 0.08082849747004639,0.40772112847108577,0.6565074252243386,
 0.41750499752069403,0.6567259490389625,0.5540860818348682,0.543392688091768,
 0.8504616625141087,0.9125439899943927,0.15021850070818044,0.8414197845276115,
 0.10763027470475472,0.6123789612314462,0.6911741815132646,
 0.15262153146726887,0.6291460324695749,0.6039540042235861,
 0.07489628495996181,0.21433751606564788,0.923142794955956,0.5676925943472786,
 0.011862627918281299,0.761869092911074,0.03533696429286359,
 0.48233558458934467,0.6744208589479382,0.7524431980146002,0.8390150174478294,
 0.27535347276903,0.10846971345009493,0.10899354270471018,0.5146286895516026,
 0.10383823462806452,0.6179038103105242,0.23881994655394123,
 0.02111586060913595,0.05257519134188904,0.2277576235726858,
 0.6203221487127183,0.46036313299273846,0.18215594011042113,
 0.33296637945802776,0.32925515262788885,0.6932339243329617,
 0.3302493877311584,0.6623951695909727,0.25608463719797614,0.9191825160222614,
 0.5927229832520562,0.06599474987677256,0.9820184423676431,0.2777353483569631,
 0.3942439571609371,0.1786597559540022,0.815180834439866,0.5979327956818998,
 0.576208402364587,0.06166499972096284,0.7574993060939097,0.35576528857266143,
 0.44555363144878146,0.1405274760474382,0.7150982550866842,
 0.026405347203334983,0.5678571512575336,0.02419145101676723,
 0.6804335866556982,0.12037936504845814,0.23748003566450593,
 0.48856340254472397,0.19741214416989028,0.6130565743885068,0.134976108573171,
 0.7029732791274643,0.7024124584910006,0.7599530532619417,0.3812159295249199,
 0.9251918824103664,0.7715012036210916,0.5084658994089695,0.8082846981422588,
 0.24458383198247224,0.12315386687849861,0.7243311011402699,
 0.7510925126414134,0.06683551512805253,0.6617478636929711,0.9695763861924448,
 0.9387590890133237,0.23991793975146547,0.03596839477494895,
 0.057546354714297476,0.6261087220431623,0.091042233115333,0.617231325315124,
 0.27104483126556267,0.9163897385004913,0.12814805343688107,
 0.24485217276125504,0.39129481163732704,0.6455091001040208,
 0.35635288152287437,0.25840331169379993,0.3276968493308493,
 0.20486441299229308,0.023030656090273327,0.1889006994845056,
 0.630978429033709,0.6285110811802527,0.880404600060851,0.3157397337880907,
 0.8844154452859835,0.021229219513035913,0.011458511627280954,
 0.3956385462771763,0.42877967358454994,0.9060667753525463,0.8665200349003346,
 0.9551602617480923,0.9771134256133149,0.31383584772529893,0.5507333565899806,
 0.4872908176570787,0.7087979370019711,0.21413433419693095,
 0.12607019619926685,0.43952774141954765,0.5378305062383402,
 0.7003251458250731,0.9381188703060277,0.9868891050198151,0.6710603962090056,
 0.9102277387468605,0.18271222283828747,0.3378510795675914,
 0.20486624193712422,0.9061927276340948,0.27384030758686584,0.511903393747961,
 0.07317419482600274,0.5004321710380506,0.6759959084277044,0.6729431135610705,
 0.09408025760480809,0.6438826180020183,0.7331417688749292,0.198382755803594,
 0.36776078046661076,0.45139661121378716,0.8418836870970698,
 0.6415632435070632,0.7403363769838844,0.58079733597714,0.5578030830964364,
 0.7797253971307989,0.7125951623053268,0.6999732570880397,0.8222223099982067,
 0.31065059923139204,0.8538484461279583,0.958131225033332,0.7243575759187212,
 0.8225231507650174,0.14773009208932186,0.7662353826337552,0.8564603090160939,
 0.167511868853362,0.1957831904948566,0.7332098963758609,0.5211875066774021,
 0.6917408076413479,0.7050376091403396,0.34201345107139003,
 0.37420716982273516,0.13016496689914248,0.7809169306848533,
 0.5470723799380695,0.01637023554443051,0.22098868529537752,
 0.1315583445814299,0.7112524713956737,0.9283589990357836,0.7857609392813462,
 0.8000027015996133,0.3363065806783154,0.1201568972029674,0.865336913149648,
 0.8540000575721439,0.7807519763908377,0.544809836593218,0.19867606259578152,
 0.016382115695844912,0.8962913046332586,0.3287108473338869,
 0.8509266146618534,0.5946191158868339,0.7960804543806566,0.4593546893936289,
 0.18089671073619218,0.4794694583939846,0.7582144740185937,
 0.35817306174525776,0.24884614414513317,0.45143170406039257,
 0.18896694943452463,0.3095054542001566,0.6143937832751182,0.3177487813315264,
 0.1423708803078153,0.28624820939853324,0.6332215493443183,
 0.030080748562805626,0.6211237931842934,0.6627792443775187,
 0.8861712043856526,0.5680811019861485,0.5266958934627842,
 0.033784143703688985,0.5725282988135699,0.2495767432443432,
 0.6938468311705714,0.49813346262390157,0.42514650218787353,
 0.5663721446642501,0.9640776660802126,0.9083257770732214,0.6130096806787658,
 0.11837293955788286,0.4358692165670144,0.1090190668136537,
 0.33674901494972354,0.6627675185101726,0.7532299102775302,
 0.29327833499808653,0.6767641731662123,0.6000598290273328,0.8219060516954916,
 0.30179402636092867,0.12952180430513816,0.410164204838835,
 0.31977387387744816,0.12084194274513427,0.4337627151634117,
 0.0537042887093272,0.5835023339664022,0.43090738145947205,
 0.42710432443852064,0.9691180193165829,0.25406814874258854,
 0.48724404031631996,0.37328670656008756,0.8352987606464647,
 0.3967009704504315,0.07165884937701894,0.4024627632360138,0.4085263217492533,
 0.08879197503761238,3.910750973756816e-4,0.5784910507749818,
 0.33148106866637506,0.30151391965788243,0.5799148068927109,
 0.18401533757619049,0.8406048429759165,0.5613755991311222,0.9391914847147115,
 0.9220955874356631,0.7912333991254378,0.0342632954635359,0.6908971480274275,
 0.43470722718275234,0.5624759182988947,0.40480744422482584,
 0.05503910622589836,0.5863697826436388,0.6047472526075847,
 0.10666107980039086,0.567850185796966,0.08088445024633029,0.661259422970963,
 0.8800102622296856,0.7988292772353263,0.9462713693900114,0.17558036153148326,
 0.2682581392922564,0.15978113009742478,0.3198092610134956,0.7077558031504364,
 0.5524895215290402,0.9523371062168725,0.5020586696119345,0.6905864968322275,
 0.9287158528948317,0.0097032446880283,0.41652893159771187,0.3912380972021312,
 0.9780733190274531,0.6120600381950824,0.5047530893303488,0.04416283529742615,
 0.8842340163288585,0.9072223519134028,0.6638206921026538,0.4024149346431005,
 0.2088816058798888,0.0580265696524912,0.5949750689808094,0.3932648820541169,
 0.5462945577730033,0.28134096619032867,0.29618070211603187,
 0.6994000314799445,0.76983223641907,0.06647656064074803,0.716673694439436,
 0.17198589638514716,0.6525933655784594,0.5511479924924011,0.8010588620204993,
 0.8039949441529938,0.05125334122717218,0.8331158552671923,0.2787750691927131,
 0.5834231284002048,0.9938077997571868,0.2091112711475085,0.9998620740580081,
 0.7019694714083052,0.06775189091670253,0.15450824310114986,
 0.14413637251944134,0.12102093312208073,0.587612606359519,0.3381881967226946,
 0.5236961557591511,0.35504226040270503,0.892956610835097,0.5356905868175357,
 0.4615284245981509,0.8481043496160642,0.15302284449604486,0.5668223827842177,
 0.3224239269278725,0.9597628873838792,0.3582485321225912,0.4378699421334227,
 0.6304472915501823,0.05355733658442363,0.62607833022237,0.5533267332759114,
 0.4771470702976903,0.7341224924411909,0.8484566898788088,0.7511776642124725,
 0.4352980169589873,0.7481929114104054,0.8444056414894516,0.5987849337845935,
 0.9824122050144745,0.3407553724983622,0.5555461397002674,0.03053812070879014,
 0.3887669471444666,0.8562013485411235,0.8291575101489989,
 0.003868883047517202,0.44924505184188446,0.5039241306726187,
 0.9976456463815616,0.061949244595494224,0.2345730590902566,
 0.6771737254815302,0.13509376688580355,0.6565674061815537,0.2628625212492476,
 0.1733205810196219,0.8707231498588519,0.8704096159166758,0.9034993583044059,
 0.5081138271887448,0.18998184565608767,0.7304886447792603,0.7519354754196276,
 0.344481842319595,0.1609242652893288,0.2056024554559568,0.165932044927385,
 0.1680066326889167,0.8001760944023228,0.07901719820163045,0.6150388462854232,
 0.08846014487436771,0.6014216250890003,0.29074551488800093,
 0.2766647613271813,0.15019642086433227,0.23353841549464183,
 0.4080687175594483,0.43124698220334645,0.15978361154441412,
 0.8622254951773289,0.7757410733865524,0.016990526061133826,
 0.9849078978536974,0.941608454957567,0.023064087130447164,0.7591671155305287,
 0.4074229111506824,0.7957759994790496,0.010017695651093672,0.991601409228136,
 0.37157396261104136,0.7786512684957508,0.18814669217857405,
 0.24527753471755043,0.40749825447088106,0.559318935779858,
 0.41394723014235935,0.5780634273164102,0.4537751597455696,
 0.07498077833995664,0.7391218492789837,0.4827717473883233,0.8724297029533423,
 0.275462238370185,0.6756106444037485,0.105260383061195,0.4921890897560579,
 0.644188455453973,0.37064231239644096,0.7621292454768498,0.2947932647679833,
 0.04972595128963586,0.5037943319961432,0.8844269697513891,
 0.28330466644353736,0.4014710066328433,0.08166963442455488,
 0.7622960488000574,0.9672599177550124,0.5834519475986841,0.5069075740605096,
 0.9779911666384044,0.7366308581802937,0.744115729182694,0.67460231805998,
 0.9922010585047597,0.9691727002845415,0.050817148479511465,
 0.3956908870155499,0.776850698732133,0.29819344043260154,0.9439328380384434,
 0.010989741913008055,0.8857109767678002,0.46828517013343196,
 0.18254473804791116,0.9411990300915664,0.25953976515683763,
 0.9982917005518137,0.9979862722098557,0.07506855601439011,0.9134745607128189,
 0.6617506595982822,0.10786093068405578,0.12827446182204816,
 0.9380028077171056,0.18969350757684533,0.9265771046101546,
 0.01922692932416048,0.12307134378487972,0.13636943698006765,
 0.49011254187102526,0.7477431132301642,0.0687125298921838,0.7494444634939462,
 0.47743548910811273,0.7724675418292363,0.22170754259584124,0.554023317155433,
 0.15060330351624507,0.30007008986215733,0.7537770144656234,
 0.6131843214272686,0.43591358166467886,0.5100678616624911,0.8840633480102323,
 0.39457113228120444,0.8749937998909132,0.7529702984530303,0.2284537540166306,
 0.13534802645964228,0.7045940748655994,0.8610895388927251,0.7185944464241798,
 0.22299749014099535,0.8673885278432903,0.42169673449188305,
 0.7993389872755542,0.49015748427500383,0.7787386740301523,0.9500979883355807,
 0.46194057981924797,0.039293645854672765,0.9227610967567026,
 0.4058051549288043,0.34169844833738616,0.7370433238284424,0.5337458604368174,
 0.20659980106094167,0.4837750062136904,0.650701551877799,
 0.034870278091398976,0.8223299070532616,0.31755488252065,0.03687938500842325,
 0.8050860038919616,0.26090737498430994,0.9752008297108254,0.9913280223121447,
 0.339249609648533,0.10805792592259578,0.8460974357355722,0.7579098736710068,
 0.13444935813880676,0.5195446994430492,0.8153265166746209,
 0.07686845475700166,0.881090253272642,0.5873150634296899,0.28561761042929457,
 0.6847135450541589,0.7306254138858496,0.7776489025568829,0.07860733359126992,
 0.8194771217709919,0.5952937131571182,0.35387493824342653,0.6246323870341479,
 0.2467363465136283,0.735854864060631,0.38097243825090443,0.758367056363062,
 0.20649455224428048,0.6746348036784114,0.07383075281934248,
 0.5310690651260037,0.36006443362948815,0.8849068543386645,0.4877117235554893,
 0.17052209589750855,0.3351203487636646,0.1908447200847364,
 0.41844466214967535,0.03113791759307638,0.5628896230705998,
 0.5844633257481857,0.5369068640589045,0.735804868153008,0.9335423955125411,
 0.25070116034542744,0.8356136649480127,0.7520236813554753,0.8474985946160709,
 0.9626022249704652,0.9071559366148277,0.8274171489148162,0.13910857259262333,
 0.7804445473961558,0.9164383410708689,0.7101685632981136,0.7794122909636575,
 0.9719979973988324,0.3823283763978015,0.5231241282057536,0.20121498187483278,
 0.3406643530922915,0.9771409740174192,0.23211675982216895,0.6665729191788918,
 0.1772894176702322,0.2772091610914722,0.2320530960490278,0.07709797064792712,
 0.09987698608986162,0.3301686364359986,0.5887028849848398,
 0.047236676296483204,0.9157211829243536,0.38791005795881806,
 0.9857170131721317,0.5664853613750856,0.24524472547567344,0.8396478198255847,
 0.1496003723758187,0.21817259103060438,0.17785469042634505,
 0.7964972889622647,0.9588711025184377,0.3862768824366424,0.338079770161553,
 0.8369522496375482,0.023918150498360102,0.1426538676861958,
 0.5861744170988715,0.47536513234802036,0.7656363650128124,0.4504497824683821,
 0.6654309983336618,0.31515802273276405,0.6398788451407771,
 0.13316262045979976,0.019191473279075666,0.8773131459499884,
 0.7129132939252755,0.273569826431422,0.9697492672263714,0.6295495927459407,
 0.03312481219431518,0.005878873758973335,0.016713907725206312,
 0.6390046799902066,0.5056613262299496,0.8187323790235028,0.42491362867378857,
 0.6107823928112041,0.25631255054918634,0.8846722235800459,0.5492307012309893,
 0.824009022482348,0.6134507289480202,0.6017190070764572,0.6811693895637855,
 0.15699895923493234,0.8597903290542891,0.3449823322472052,
 0.12246098883499046,0.9303198514177262,0.5301539467120032,0.7135324768195829,
 0.8423051243228058,0.2234189646782605,0.11347505870672014,0.9420663106257471,
 0.9008304105862986,0.9742962491943675,0.9377270852946481,0.5045303375380294,
 0.45131028680776863,0.5618430329906932,0.2909378301688358,0.8721200560497839,
 0.1068838096746988,0.7642958917772348,0.5910375924401533,0.5837965092392188,
 0.42109331617067447,0.7369545778843352,0.12098868910084448,
 0.11818708334986716,0.6026346053576186,0.19361224719391734,
 0.2787018543817319,0.03757277164950501,0.29017937311948017,
 0.6084827358408296,0.38852893047285153,0.6324635728738676,0.8598830054260662,
 0.5746092530020421,0.8504748267015716,0.17825709536227774,0.9850678207039094,
 0.41963759115608934,0.4321673968502727,0.5723582278545813,0.7647405609468166,
 0.4245135588935036,0.8051759807238582,0.7100373986605018,0.3231215397567504,
 0.043212189815517,0.3033785402204162,0.9710391721908835,0.012413960008575975,
 0.5373316682668667,0.27555110854414455,0.4464862615449696,0.9782332590835926,
 0.368641724721277,0.5459973263582506,0.8732595073761106,0.12425010490102117,
 0.4576702265089594,0.32389064544788226,0.746535067431734,0.7686005376279309,
 0.8388462474914284,0.9504378985287926,0.4379199134977465,0.6348852200604389,
 0.522978783831889,0.7451726599722122,0.7037857203636876,0.7521048651260974,
 0.8083934009571636,0.8864921700094425,0.9233057539563925,0.17627101810656842,
 0.2896209601928923,0.005243720235280858,0.1423122055357665,
 0.5225690050196596,0.49148029694153617,0.8749354295817191,0.7062618299442166,
 0.3578806264392964,0.9537793265383757,0.7562796426265881,0.22810239105455954,
 0.9610544893880827,0.6031845041945386,0.6337066435531938,0.367610983028131,
 0.954221968695609,0.23427847918164701,0.6959109245596529,0.6902422820478988,
 0.0892106746048641,0.11120411577683831,0.2802451043407723,0.567288041028231,
 0.6562633634489573,0.3764476540843277,0.25166895642847276,
 0.015524807021337295,0.8794282582522667,0.2708681190843043,
 0.47443525467704506,0.3563724155626874,0.3033663566493606,
 0.14771644582551768,0.28392720473754995,0.26929888477359765,
 0.6707674462704399,0.709216629114213,0.9698835883382536,0.31196622093043913,
 0.5626478690673375,0.17327999763511615,0.3001474032423689,0.819612371751269,
 0.6927990789802017,0.34591453442096376,0.7797340293277711,0.8637021161549931,
 0.9095682137155346,0.9427288156512805,0.6281277252378025,0.4113188505981785,
 0.7049790501114617,0.1932245026577597,0.8540359743678816,0.9909389782279941,
 0.26337882642172494,0.43374533135832527,0.5065922870580898,
 0.010276462731389935,0.9662131389049073,0.8142457398720399,
 0.17693829606395106,0.12894442378368876,0.7376279786777212,
 0.13145364021866313,0.3361666079279352,0.9553118896298802,0.3454911055418235,
 0.6731672721620588,0.050485858823214214,0.23950329141416116,
 0.7127250857324577,0.4172701374177017,0.7827178198061454,0.43971060854676025,
 0.9320335707683793,0.34733400494306466,0.7468510939216977,
 0.18257404028742164,0.4103695821543719,0.36570651124185116,
 0.8993333041789767,0.20316788638092997,0.15651574396676182,
 0.6618194821465987,0.5507349971833374,0.5460448002378809,0.42290411269123585,
 0.6482629487227349,0.8473220554962742,0.3272026574927507,0.2180531889388473,
 0.6528626090248081,0.8906333783000145,0.20934382251728545,0.1018988091576607,
 0.961983396480468,0.44351759501274124,0.27289379498526933,0.9826473171022687,
 0.7743521424147273,0.2790115974703291,0.28777395266164796,0.6144408626831225,
 0.5125465884401417,0.9550224896053441,0.5028731822685233,
 0.005463512487080879,0.5443889626216466,0.9756779001471304,
 0.14472329855501465,0.578589220382071,0.6203772865315994,0.41211057197726697,
 0.07312304860929109,0.14800714306795815,0.9082070585800993,
 0.26175232406884197,0.9223497426468996,0.42240191068505817,
 0.1685965992867533,0.8218815341845668,0.92029026405758,0.3880690117090616,
 0.07866165949765347,0.48255134072014016,0.4194121489528848,0.120900556862509,
 0.18220204574265597,0.6260569095561568,0.02618922489030573,
 0.12275114996525771,0.4515511384251205,0.46536396678802294,
 0.42718689758816375,0.5134187741837875,0.04773163387345514,
 0.6457425916104237,0.0890936685231487,0.37177874423012136,0.1973911044189537,
 0.28091684543173434,0.3903138680181474,0.8066509755847366,0.7632259767300165,
 0.8615569706392648,0.2795787600382299,0.9151083676476657,0.7031005518641603,
 0.3888716417232586,0.4801094418641464,0.5688507925024663,0.8450227749051034,
 0.47406433928954317,0.18373378301787713,0.1113940707760116,
 0.01359796901864918,0.84078790752716,0.9788421751112364,0.6702789926852334,
 0.024690818855260144,0.6879812105797435,0.7605266640462016,
 0.7439586635129316,0.448765172561301,0.4290378571484784,0.0838587319194043,
 0.3226688644945894,0.09038293706445277,0.01330466992983348,
 0.28206635955721093,0.00664978980267783,0.3169633254178903,
 0.2474845390324233,0.8153656092964318,0.5181027496669198,0.16225536378214478,
 0.011475888048935734,0.6523992810019176,0.08902477694620903,
 0.9974615689330697,0.27439494046395774,0.8880305324718824,0.8140312547462749,
 0.6823182872085299,0.11268323072391728,0.2712962428654061,0.8985534031332579,
 0.6872672370790288,0.09677970573753503,0.04083420313904729,
 0.6131441500139226,0.602207368864285,0.496232421608657,0.03437888956550861,
 0.06811574417207245,0.4471604703819193,0.19650396370614676,
 0.2988162729852657,0.9192146828242093,0.5955454646138997,0.3935491151198225,
 0.7091019679554901,0.5744975338950191,0.3897575506946249,0.7450547438573016,
 0.41593527619474935,0.05440074843184162,0.6753073861311076,
 0.6115717922412733,0.9770283677997784,0.030297424230709025,
 0.3325834705844326,0.28247495757598917,0.6518093646849318,0.9323153518402308,
 0.3649627499133412,0.7437372292549154,0.7029776470237784,
 0.004529897941108918,0.8839810796662648,0.3825823619230524,
 0.9428990810948902,0.1251558982421811,0.02577992803169915,0.2854439217345457,
 0.9256960093693428,0.6720740443327591,0.19910740912670177,0.6658169329750407,
 0.3407893371495515,0.06907649825565232,0.6267903957582341,0.2908111098839712,
 0.36692255604664237,0.8225303743958728,0.5514049682040747,0.4827638643953106,
 0.40908016843514383,0.16325268120283332,0.37398163103878845,
 0.7884539785397687,0.595787262935575,0.08790737821549843,0.43548644305137096,
 0.8352538449113804,0.4976590892114422,0.6452435999978404,0.8729214228679939,
 0.9064215480332567,0.6183657520445874,0.31307141307694897,0.7662841987171983,
 0.41578399748361383,0.5510887131426916,0.7012599980618278,0.5628145178519374,
 0.5244276441976465,0.8545957956607164,0.9056861729347829,0.7348614934708761,
 0.3747654924610113,0.269771268632475,0.6038679571341775,0.8721573107476812,
 0.566815826290673,0.10504458672401229,0.10721307100447419,0.3516197320218616,
 0.909251054038319,0.27710571316429844,0.007889533685990324,0.710139741498332,
 0.6308903298811955,0.10457976683196857,0.5520396260957365,0.6256910802122799,
 0.6545373368970702,0.03206304080946265,0.7799055475696691,
 0.048901869529376696,0.7473382790044676,0.25372863110938626,
 0.33819963356904004,0.9534687741255238,0.035101044775008416,
 0.5086406782440845,0.03888943994514149,0.562081735839794,0.6222529266615441,
 0.9680049353963968,0.9100971963010666,0.24218172424316642,
 0.08749678291736163,0.8288080732692963,0.021679652792249482,
 0.5289540189114684,0.9856607144475966,0.6547723724670476,0.40289637339675943,
 0.1531655550867601,0.7695104336008166,0.46140104482239097,0.8042059776531327,
 0.4631650164176774,0.5128714589509829,0.03674474532029426,0.7911180147716743,
 0.8051377712523438,0.7859372827474945,0.4836742775842472,0.4811405551954431,
 0.07951056616834107,0.2785095460266447,0.47077950594039697,
 0.8795572918099331,0.2874634511748514,0.5962550343488552,0.06892433018016897,
 0.5665586877682892,0.4140019922502163,0.8773449905318456,0.21627544843066504,
 0.6740732930273943,0.8403957382186189,0.030940402285693613,
 0.8987288459870704,0.31040127030571363,0.23988844818241262,
 0.07674683339558752,0.031133590918311915,0.047781694064736424,
 0.8854738840269737,0.9875766243103195,0.8051993745125288,0.08007288330947615,
 0.8986496860402448,0.5286078683790424,0.9997883077015004,0.1523554743177189,
 0.12657404036359243,0.11323643135464934,0.05934267863953446,
 0.3788816788994712,0.6642506958971793,0.11781324016996875,
 0.49047518129142353,0.36038583334954044,0.5627556206223217,
 0.23867385475673486,0.6429160531418023,0.7366202777981754,0.3930573841160624,
 0.15596915083615126,0.0948960810567211,0.5215879061188983,0.8417432110759775,
 0.5036290908304788,0.6260328928996032,0.5718171526379481,0.3833277374071318,
 0.6625725638030873,0.24891037675849415,0.9380101374563483,0.9071182021251233,
 0.49579481583453333,0.3520275971145217,0.3650277001620945,0.9111317530392,
 0.05994027113011602,0.7242157021498606,0.8515864505582,0.7020396963934199,
 0.7331322332875813,0.7041393073776429,0.7691725924138488,0.3878772455337629,
 0.051637625905369244,0.3424885019271251,0.6611212548547221,
 0.34232326847152994,0.6702734546486608,0.6728018254421468,
 0.29018135608878604,0.2204155293976391,0.8916590809461115,0.5902983939441734,
 0.287807558538401,0.7826504208090866,0.026519979996006415,0.6378128424589328,
 0.7007289413776978,0.3812754848194797,0.1721219327665664,0.4190513171845657,
 0.8444942466354951,0.487272138461188,0.8514234657998018,0.658266154131314,
 0.6669871393760229,0.8341059942627025,0.9623049987820803,0.822074568921245,
 0.4209853053035666,0.6315964347517,0.24475542806502137,0.7449253632538226,
 0.665405960390918,0.674818350150028,0.6960955655605005,0.6364740129012607,
 0.5353062662704027,0.132482274701488,0.46935648914411976,0.4946912796844495,
 0.5795270547059451,0.46132896326302353,0.23352599580852507,0.914494307726193,
 0.7002269077999022,0.7468515215408448,0.5401945563518142,0.8539619587865308,
 0.3261741977271275,0.7387398579366417,0.508610177055281,0.2690696865572463,
 0.6667686072334156,0.7132290784677322,0.8470929736491226,0.06303021773774686,
 0.9309108048756981,0.2575381427061004,0.06761903510671163,0.5435456510076956,
 0.6868788707273423,0.48324111953728666,0.32221554553796994,
 0.42462959939904565,0.46693283582892475,0.17314604109261,0.7840444310140208,
 0.8878358013654345,0.3463697535031216,0.3493788671613761,0.8188116761372917,
 0.7751283300447954,0.502638436769629,0.31980536729570763,0.1414487411885883,
 0.7942730842655162,0.7848179863105227,0.28258280168227534,
 0.12703208917206898,0.3856930961389202,0.328733285223182,0.40059530200673577,
 0.9057369618420519,0.8654188010564754,0.6897025081242663,0.27958726263903366,
 0.08390496174690465,0.14124108058077622,0.8810296713134642,
 0.5004060720587656,0.46693725650857365,0.012095738726728866,
 0.30681806486034047,0.2470288599567686,0.1928933783690997,0.8656516317814301,
 0.17361394316088075,0.06778588622629111,0.7231711932099371,
 0.29924797719460017,0.8486459871454863,0.6114343554095795,
 0.13853894674563594,0.22689109062633273,0.21619595812908265,
 0.24121283563202955,0.16125403375139435,0.0018809691943194595,
 0.5159513391618858,0.41174456521166247,0.9639753804469147,0.6816310853123262,
 0.10610010791473456,0.7074394102320087,0.1303889674402512,0.9287285770643774,
 0.1039786684201447,0.3575545846454705,0.5733889221271257,0.21338369752459696,
 0.6059801218274959,0.30569907805445395,0.9062264752503261,0.4773972223021221,
 0.3144238698694055,0.19568343898637375,0.22222759070690945,
 0.31404469986658745,0.688801842400837,0.8483514169684195,0.6429610029880543,
 0.6959471684330952,0.9866008855209221,0.9249009501530169,0.23555663596150223,
 0.7799092294722181,0.26382837689479033,0.6774698404923754,
 0.09393736910088346,0.825950556281802,0.5896834702688438,0.6742616505312732,
 0.28459509232423197,0.20225561809268233,0.9633139049581634,
 0.7959666479145946,0.2914309388337686,0.3651963670272934,0.816804430603976,
 0.318336244706118,0.9521565774127817,0.786355996391419,0.9087179348651692,
 0.03472244136450109,0.8109266178893562,0.8616747519985847,0.3643522734254244,
 0.5919281010050454,0.028173426309246352,0.6470108985113338,
 0.41994878270497604,0.47123594387501755,0.288898128120725,0.9364102969325366,
 0.5203083138967787,0.31317828787504576,0.014997016310843936,
 0.5159770088757878,0.6790091353441543,0.16577404021031483,0.6437120333899522,
 0.4287646202045181,0.8244215907672681,0.4663626650881366,0.02252966552675928,
 0.3310306169545907,0.04771442245376012,0.6889538892723507,
 0.16956930925125957,0.6603460855476813,0.6325516617144242,0.8863528279484798,
 0.08147752665636032,0.961326661699535,0.7253903242499943,0.16744967476432837,
 0.6266440673427939,0.31622324381063405,0.5627199299291048,
 0.38215059071928825,0.5353156937002534,0.4681896339540452,0.4375235329119548,
 0.9332128406510576,0.2770744702847212,0.8995915502783558,0.8996537906178111,
 0.7819370264189289,0.2318706712763312,0.7579555778264931,0.71830714532614,
 0.16711028136361916,0.8615982128803988,0.5514504963779703,0.5550329453689807,
 0.7472082009427389,0.20494082680943726,0.3542856522750095,0.9589611311133495,
 0.3088112997820044,0.01124701857740873,0.32414995494024323,0.873403383846803,
 0.6030734220492258,0.06156107740487671,0.9325686935302734,0.8854893438472747,
 0.07689088165674629,0.9989178282247313,0.9926830689378535,0.4996737159393443,
 0.013823108938717743,0.8417935252530319,0.37254437376929783,
 0.1128458986082842,0.6245744674875707,0.8884927762097323,0.4109210488731475,
 0.1580496132846284,0.7122107245670404,0.7786052443437703,0.22584739904611761,
 0.5965285825943226,0.40997555019371135,0.3928205155164709,0.9805112928303141,
 0.46657211273655774,0.3808114510623749,0.6708581641661153,0.8355791828272746,
 0.7086503439738767,0.4921020555083364,0.1334998852490893,0.7503611878600015,
 0.6236288834673813,0.9041817710975677,0.2086025430354279,0.8301682091147905,
 0.7440632556129193,0.3154178230021364,0.928785870330024,0.1993796186230894,
 0.6050540288676777,0.28119741315417557,0.23285308458651544,
 0.09781383195815818,0.4142838484005471,0.10720279618185402,
 0.6817141901903864,0.4611131776931032,0.19217821815618052,
 0.09330105319571425,0.879182155392978,0.2456951738496601,0.6201952142506431,
 0.8678265982883125,0.017952680467519633,0.1850269663207046,
 0.9482612490157255,0.5452439200369241,0.9396055339445722,
 0.019128709292476342,0.3147343558100557,0.0016751478861950497,
 0.8778032272617527,0.6197870846999022,0.12702920225056702,0.4920700706208998,
 0.7826113417859921,0.41164052790391925,0.9222415315678385,0.5379102292061138,
 0.3305078530954222,0.06915007521841043,0.9047184853785567,
 0.21157845898123828,0.28599095557048926,0.9346734210354641,0.531334730667112,
 0.20846349740049563,0.43566845779459307,0.2810807905172936,
 0.3337681547325294,0.3601670747556419,0.6812800951811244,0.2860772217247546,
 0.48317572393681285,0.2427836224683203,0.7087589570887363,
 0.47551338457660175,0.3048524105623497,0.8895324851226154,0.2943786436256627,
 0.09413073256415327,0.4942525166374584,0.9702755664966004,0.6294480623506711,
 0.48091728099741016,0.5861242818443756,0.35604483609354953,
 0.1011273935261604,0.6582232978305695,0.12283301453876483,
 0.09736283666471224,0.4567685770010973,0.4107935997246762,0.8403211197990208,
 0.791633262141204,0.42166496064316406,0.367144463412611,0.20749914390776325,
 0.006460610479761986,0.7994301837442249,0.888742596044505,0.5229109402337805,
 0.5857487510955653,0.17906854987541065,0.0740186288837118,
 0.20395146733096714,0.7773518226710451,0.8894749428809393,
 0.28234167153052325,0.5723045061636989,0.5099838742363016,0.6457240326437665,
 0.5784675279363471,0.9355357868821093,0.17458257965838253,0.7592689573266566,
 0.5203309254154149,0.3296952054486497,0.18531389062230186,0.6348327722055475,
 0.4192665758612266,0.9135336778710058,0.038665738960795504,
 0.14340289061093026,0.3846815637743819,0.9697610138125942,0.5026336171445257,
 0.34416282513101903,0.2504007157089768,0.42738565598949796,
 0.3925643543430414,0.17976567463562554,0.02757181505912809,
 0.20261353076819621,0.860734021217855,0.885721491655864,0.08991240027943181,
 0.2156632142474434,0.9519597692263149,0.6612881376903833,0.4763070459517529,
 0.6656392477131214,0.8236909301227504,0.6278179152089401,0.12824303248823432,
 0.8147368282711229,0.2081922327233876,0.0854772203571148,0.8809343883273827,
 0.41737914013485056,0.28560120487936247,0.7314190844747813,
 0.8930146344201093,0.45959372177275704,0.37794813470194133,
 0.9519201379224134,0.842106118738592,0.1325634381634817,0.21496095166368934,
 0.28413475491429185,0.9877301562355502,0.16012965578003668,
 0.41727833943936765,0.2942785266674536,0.4450642512170241,0.8638985832433418,
 0.43737504787550885,0.568286254405853,0.9105080428958566,0.46518181463372754,
 0.5374178381355881,0.7222558027669361,0.8174695083854574,0.6859714364818222,
 0.2828500949891197,0.9241683160200341,0.8286515126362305,0.1123022895643534,
 0.5540637708799471,0.7731059530218416,0.837479663667189,0.6061737789861645,
 0.18034516692980174,0.15144486236412358,0.663933038985824,0.7201546172952582,
 0.8220201995119543,0.8994200661438181,0.24350149956609135,0.6386554270738655,
 0.5765167811128906,0.3317847127234579,0.49887547015491895,0.3690969280028783,
 0.4047405160844507,0.8267279877169149,0.5795325170571972,0.6118484547094208,
 0.3139853246945309,0.19684442821596582,0.07208551319399126,
 0.058313972564480454,0.963956198070937,0.5207228848563116,0.7406451994036813,
 0.7381935350573422,0.7626944045336904,0.8834375068727524,0.7694687630580667,
 0.07042783367565975,0.04572463274011196,0.3620554362791135,
 0.8227993710507708,0.956307170317529,0.8125422503969832,0.7408549262003814,
 0.1713606861153686,0.4990493765445059,0.4505826375972527,0.5279693481060044,
 0.2317145989838898,0.19859242522903164,0.5692062641577612,0.8279164119422138,
 0.5604308225891227,0.26841986148254127,0.11127611350075117,
 0.17169468515215358,0.14681171936786108,0.9339615050367386,
 0.23768154909931283,0.9474426700390257,0.09974353927611068,
 0.8080064553805254,0.3348405716083751,0.9043894131446811,0.13853639472637191,
 0.25342754597101025,0.8998652451689536,0.9553356709160588,0.6868654319541238,
 0.9886956491568809,0.49026330484966973,0.19240546106224832,
 0.2038270653707609,0.2472949246678796,0.39328502249707087,
 0.09258834008985228,0.8866827832697616,0.26747503728172495,
 0.18318243365554554,0.4449902182026745,0.13094169199085948,
 0.008276774605253223,0.3959871351392521,0.046637794166924706,
 0.7416970487779425,0.02533913433543622,0.3609073723726075,0.6192226920684041,
 0.9824573428512573,0.4693981181988536,0.34887693735142467,0.6776552439519122,
 0.7602282568545377,0.9898164688902937,0.20316650178470086,
 0.02857808217387514,0.7748588003979662,0.45702647952439546,
 0.5643348172120304,0.9995045460326641,0.06333565836854815,
 0.31634605165009555,0.49958489807762685,0.23885906682329294,
 0.8941468260203544,0.3102651703311272,0.087953365819612,0.9614576258431636,
 0.23255839069739914,0.98462544000134,0.9451421809575057,0.8375462369161413,
 0.23808500946928424,0.5465156949865733,0.05453130745547041,
 0.3333360816098706,0.9877618192750243,0.7203582602714713,0.933466168255934,
 0.43116482184431915,0.6527451089152253,0.2586278358533429,0.9997970312987002,
 0.522343452884522,0.6239612605885873,0.06969503966598334,
 0.011047734063064474,0.10681498583050153,0.004472009508561192,
 0.4536449469452434,0.2725478944809837,0.3311153331255593,0.2272186060496596,
 0.5681986343083315,0.35124875533079614,0.3014187970996238,
 0.10849819892801082,0.9914795803953106,0.5247891192588177,0.7570531859804287,
 0.10279582244634078,0.06991034669526752,0.27040984477543095,
 0.3611669445634824,0.8358587516756518,0.2963553155498301,0.2704000592535265,
 0.9689464903213517,0.85536896060018,0.6677136169774391,0.2911749007600971,
 0.2043573937748535,0.18348401962797445,0.30102589837625515,
 0.9897304419788334,0.17783974476086517,0.32845756022263706,
 0.31437381293104727,0.7526178167636693,0.21920335545904868,
 0.30210682525185895,0.04252917822353042,0.5537199792508108,
 0.7278524901253918,0.4086223365037256,0.39705269126424525,
 0.40852169791881365,0.9121868427279822,0.019438902433435712,
 0.39836584708229783,0.7261297571682935,0.0546063556716152,0.6789350202703228,
 0.2140229102406549,0.4299028540523866,0.08414638733664337,
 0.05873288548729594,0.29366678035606386,0.8979319014929505,
 0.8956256751755086,0.3998773514745284,0.9656578173697794,0.8651292913633953,
 0.6569686399493628,0.29491053595294403,0.05535062142932312,
 0.09509448802151632,0.011960703370086034,0.6889567411415919,
 0.44693870392925705,0.32383990875678537,7.344928618782109e-4,
 0.4184737162963498,0.5939015537242358,0.3918983670456986,0.4673921595710574,
 0.22239889693735582,0.971218603821132,0.279094666535477,0.01726955353332349,
 0.4641484648915325,0.052477463918708556,0.5909296532317626,
 0.45456237747844397,0.7980163920014325,0.2902701492051223,0.8585436410089619,
 0.6847062615264348,0.976091266955503,0.4641448191768841,0.3659160683717926,
 0.1783235068295469,0.2671207942015902,0.076370678333613,0.45410048827416727,
 0.9380480081376819,0.27982620816186676,0.1092070209106607,0.524781755040136,
 0.6220906874303871,0.31885814521231126,0.5952093461505636,0.9562080973854106,
 0.5941167833682034,0.17846669613855126,0.49232055740867675,
 0.3121740591464498,0.40028412236251243,0.8095516910464337,0.9238141147622445,
 0.34706072691161227,0.8757018462216487,0.08633197204526655,
 0.057079157156779914,0.43299094621031464,0.31652321314659704,
 0.40820596113986474,0.5899758125447594,0.123348185156967,0.6733099500145722,
 0.034186493592462064,0.35924592493894003,0.8756635891194514,
 0.19086861425532775,0.6355526823236891,0.30835798685985194,
 0.043088550783713764,0.28124700385574974,0.5519878574900411,
 0.6294007158533094,0.534814873090067,0.8425977050503711,0.4476037751736278,
 0.3060503204226547,0.2139736524563034,0.1588895486279447,0.20074446358199616,
 0.6121958169020087,0.5745536177596055,0.08111973437368547,
 0.30952253590972423,0.527318099728454,0.772804741847239,0.062060268097435256,
 0.17576585036487025,0.5948847244409304,0.4903461408535279,0.6061416251070321,
 0.6111986013452876,0.2956590849081375,0.2563487508863539,0.8595812782777732,
 0.34704574368330754,0.0791218015387889,0.8228495249477633,0.5181863085505832,
 0.45538276618413576,0.6047826928036399,0.9415457529964973,0.684678035834465,
 0.6010481279529438,0.8671270200632727,0.3773024177380371,0.6074245966223609,
 0.9059539001269165,0.6842436849865984,0.8389312507839886,0.21064026401257885,
 0.5737887591158829,0.38318686619814,0.04997707033409382,0.39977400218699044,
 0.18957896361064785,0.6136470132639984,0.11741653381319717,
 0.46126724836021293,0.8834166274969277,0.1758841960850812,
 0.35635416644856055,0.7155308711663849,0.16488882400139115,
 0.12090690805479354,0.6260564967105673,0.8535909925129987,0.682614571281639,
 0.09938678180103078,0.17119851580803724,0.37715908388826824,
 0.5292158694953973,0.11239213538643322,0.9777852161294601,0.7297225003761304,
 0.10793742842876786,0.2568696390483982,0.22622613871181774,
 0.06882796894021326,0.026127938104679505,0.6019972359069015,
 0.9393183538569972,0.8023473774336883,0.6335018229185396,0.10209969755450632,
 0.4598407563838982,0.09815185001047444,0.10532629869708221,
 0.9868209418211678,0.08482275929491978,0.7853515634534344,0.5301093918893385,
 0.9800367469667248,0.1343921862534977,0.3729038067179504,0.06703842883589983,
 0.771372357353919,0.31213324860805425,0.29826805721673766,0.754281414298717,
 0.2697441305665621,0.03829403513484664,0.9240907666125301,0.7056344181215397,
 0.3866071018788919,0.23718565577303607,0.43788831783655713,
 0.09182578548998499,0.9239503370801514,0.22172416870698908,
 0.3424400426775487,0.45503017911634225,0.24754036288364112,0.954400972509339,
 0.7672637513219176,0.5944454602347213,0.02537921772518137,
 0.31580610126162145,0.0271457756930793,0.8424067337281118,0.7153830873751791,
 0.40921504719494106,0.24657131628673123,0.21400109946333445,
 0.29691554814039134,0.04189629610390688,0.24286967716371421,
 0.6217689757864703,0.33649391725320543,0.2967330685470789,0.7812546122742386,
 0.6796294536994035,0.05003920617607105,0.7359597112502267,
 0.46378576373096125,0.2581475055637289,0.7816064830018067,0.9865699293777059,
 0.8970273397055987,0.8175593772423274,0.7762848717346553,0.6495222209949726,
 0.001991781221604816,0.18091721510051373,0.28655775729911903,
 0.5953526459053485,0.0983946883389959,0.777237804689089,0.3314232670167909,
 0.5261274394952581,0.13277842769457815,0.648292744504662,0.4107837360401425,
 0.21934757099618363,0.6578699390321023,0.8889057883406415,
 0.21977216984797565,0.4410563322994959,0.06637851345561263,
 0.4227055440678922,0.5058752557863442,0.6598490334483913,0.18929164579096058,
 0.6994973968759042,0.7160119109987936,0.3569762722681604,0.8554060189974172,
 0.0966502135725199,0.5740492616872579,0.8635400959581065,0.5555742303493416,
 0.9451112093544837,0.8329781008159841,0.4617475908970188,0.7980996868612635,
 0.5493143360552016,0.22088442621065074,0.08353851561906589,
 0.11945411634550918,0.7839598767738297,0.20845101980109615,
 0.22311227642489528,0.5377454124406995,0.8332747349638912,0.8049044444634192,
 0.7907527452807319,0.558224326549672,0.6480446189206233,0.8407705376079723,
 0.5314870102544716,0.766604550503063,0.0182003540297061,0.7396735096266533,
 0.4206952945343201,0.7380825381479221,0.3809234007484471,
 0.021284154344660244,0.9312634909053352,0.17851873935802298,
 0.5597234279138895,0.2544008149991045,0.12826523396923462,0.6050907783323968,
 0.4829869585697363,0.6053565533916712,0.48513213580162917,0.5577174190691275,
 0.2437691189837441,0.6091776650130214,0.5723707533681597,0.6725902292184612,
 0.7560486713108034,0.32408557851694086,0.9491915510405454,0.5658031148174801,
 0.8823907708719607,0.4876834904705023,0.6461509595317002,0.3227013944267546,
 0.8985424761268342,0.0258692826657968,0.21776583795743487,
 0.39353084852106246,0.3121864028056548,0.23253927664144047,
 0.6722434422792403,0.3862240257170696,0.2147359668831489,0.22819269856575553,
 0.3859967093829365,0.9483165859728786,0.9617348541673408,0.5222495999926428,
 0.3696697972674673,0.0731866712201541,0.017298580261198904,
 0.32257655408943653,0.5053192023196952,0.6655325505235165,
 0.15813916191311395,0.22390956925722716,0.6051583653405509,
 0.18682453092464302,0.8577740828675215,0.1417702704871162,0.7349310805333262,
 0.6818047341983325,0.15572313218894518,0.8517537479885405,0.6535331797622527,
 0.7202815163418791,0.3210916219952945,0.5844972596847482,0.29477182969724924,
 0.812944705313109,0.4926028893129988,0.8839183278989251,0.8939651033766325,
 0.62950855254786,0.7609564765776666,0.20272709071756945,0.45131936948812357,
 0.6624826443983786,0.017026634903186677,0.2535338927624331,
 0.5847416843597395,0.9179699137638194,0.8288785098549656,0.4694812594145633,
 0.9918522638665345,0.17589494179409815,0.8818599272794414,
 0.11195183262756447,0.38962768667591396,0.36624817578599456,
 0.05118871563588845,0.5236986142938546,0.0808561827817107,0.9981420505503501,
 0.13266583064709658,0.058533551515835835,0.5126384031459277,
 0.13637123252790617,0.5130764946536959,0.7746016702431177,0.7291040996974146,
 0.4347403910620845,0.24941718998050233,0.6647074319934314,
 0.04429876783181774,0.669145810952874,0.8191822327293289,0.2557098656981217,
 0.8994992292221209,0.6347739869649944,0.3958791638708897,0.6035320257363779,
 0.5338496903180516,0.047014235317685116,0.2060436931364038,
 0.1425312624553422,0.9115894414776702,0.03017457978397109,
 0.08063486903389327,0.377277318695436,0.11358633345580149,0.8699584619507283,
 0.040020210988632954,0.1865585835385285,0.873834846207971,0.4189623458584002,
 0.6704305989760329,0.5329254906843459,0.12276354428403824,0.3592304391279558,
 0.8298125396535806,0.011683307055348702,0.9193845861335113,
 0.9365346712333036,0.9706762981791485,0.734471932543992,0.8682907480238363,
 0.5765648134848088,0.038387105695457335,0.4450313973864807,
 0.48152810725261275,0.10423386754406594,0.20731649275977515,
 0.9308916255633376,0.5887517454120371,0.09100010862541486,0.2885297615990934,
 0.15853717653488997,0.3640757457493514,0.09475456675293048,0.980485369715011,
 0.9347609368564248,0.5875657399291703,0.2210280855612834,0.8118175624516599,
 0.0871825895800763,0.4113936821737245,0.2787623235180169,0.04911573084461618,
 0.9269469733775634,0.5910870573206221,0.19978783118837007,
 0.13064352117226563,0.8745416759735989,0.4785389485788101,0.6749357892964875,
 0.16347218063524926,0.6028511570460012,0.23644202513268742,0.572971963023817,
 0.7020856651852522,0.15563730247042484,0.22675008887476278,
 0.9522485208303162,0.005686343809118144,0.06441610181393143,
 0.34111857932519485,0.3051335350284655,0.5403345744760255,0.7702080877167117,
 0.12332986306704152,0.27382535173207956,0.16530512886376303,
 0.7611334060687657,0.9911498000686678,0.17205796558596187,0.825153959372756,
 0.9969372180283743,0.48454948027149347,0.7675055854121973,0.2368707844988286,
 0.8501539213413787,0.9268157289645673,0.44513890731903427,0.6326509373275456,
 0.3529684289825864,0.6543575356649185,0.39849484627945975,0.9966332984622186,
 0.27333765631715956,0.7753111550239106,0.16124048560215798,
 0.7826992069312129,0.3285136646421791,0.44782498755445144,
 0.15392148148573492,0.6734968370165689,0.717844882726263,0.19270465603864306,
 0.3381583754445665,0.6137270649954327,0.9239046681060293,0.6052544308690576,
 0.6034282510729686,0.027767686491382282,0.9287027021453141,
 0.14882743561638667,0.9873512505103985,0.2071488888301457,0.7607561226032842,
 0.4462790920378865,0.7874472069788035,0.797037688746066,0.634072002098861,
 0.057719164452229466,0.0485830107506009,0.9425745141568843,
 0.4832284250198633,0.8287615265464511,0.6379572010917701,0.5084800326336364,
 0.33836342902878935,0.558219949604705,0.03202435675627591,0.98415259213464,
 0.9314393084708179,0.30118706413438434,0.05307858125563991,
 0.3293054509518224,0.7700941771316844,0.7249316531007602,0.20376842296431996,
 0.39060095665165984,0.7165333825905188,0.9243444109306097,0.856733473120254,
 0.7209161179519452,0.4695942802398131,0.22961088595968104,0.8948498656848113,
 0.7549288930391311,0.26636913496959547,0.7869630339775824,0.7748545020857874,
 0.8700799356955012,0.8803447342465403,0.6100874055434057,0.400816298828919,
 0.8893643058286699,0.9229125295038887,0.9719676391034485,0.9632444085967933,
 0.3869060649841669,0.028787163819135486,0.419406233219226,0.8506077049639836,
 0.7456446532801303,0.45615761660226073,0.32380485367602896,
 0.7078476853822179,0.838553465115895,0.26855566541497167,0.4647034893162786,
 0.43059766307025904,0.10249664117810997,0.39423883097806867,
 0.40913695422434204,0.4735622759812117,0.28404628854547875,
 0.6968742107238387,0.42007950856039744,0.974522733246864,0.8782937425146925,
 0.003854378263921765,0.731039477942467,0.9683367888448979,0.9341755357260472,
 0.7104127048799387,0.5172106142655065,0.5604544337498102,0.9393367546518822,
 0.0650255843033104,0.7806946061450937,0.004933339720643359,
 0.5148579757516429,0.7266018025942548,0.7498202929309142,0.3029918590041888,
 0.28492613409112044,0.7292737084274601,0.5552943543226369,0.9203080553092865,
 0.7404087321825588,0.5102717399897457,0.43841092951408256,0.8109601329399775,
 0.7860168673597603,0.7369683769510956,0.08482969330083945,
 0.045511069078180144,0.5901399868949702,0.054039107988058444,
 0.9589491045444146,0.658849521248164,0.8221733172995556,0.33019286294003547,
 0.8352691468482685,0.45508593375336215,0.10685750804109129,
 0.43261134742087703,0.20348310961876692,0.7623317441933354,
 0.5098781550407093,0.5300162644124663,0.7940428293847378,0.7238759788017166,
 0.7785170004731783,0.9473473896980485,0.998656626113062,0.08624661965435076,
 0.7357423029955223,0.3557696294006719,0.734524923713845,0.9121054934571109,
 0.15620422410618962,0.38077156983901417,0.4582778781539636,
 0.6218129459068223,0.7559344850904443,0.5873865087112764,0.9208840843112647,
 0.7351874335631831,0.15664688000993676,0.9044430632715026,
 0.10439120748568054,0.17310344274925593,0.47342478615115535,
 0.5495473763104826,0.6907765477410623,0.5651287804627101,0.739846255849824,
 0.7175881601659644,0.022850900737678592,0.05016862275249556,
 0.4194556084944976,0.16829385880468573,0.19305911964608735,
 0.45256988517479757,0.6632563567133976,0.7497472564468867,
 0.47558229108176053,0.6923467599896941,0.6793937643455641,0.7075962765901394,
 0.01568313307306468,0.05847177597750386,0.3815277090517645,
 0.7506736060132556,0.5367532455047438,0.7477191680620255,0.8040242723191291,
 0.3107859806308526,0.29841145269034264,0.3652791405885861,
 0.25435975818870893,0.48464903322494723,0.2455309963859219,
 0.6308344214574111,0.34192295850663157,0.39452965005343965,
 0.46092244143023386,0.9366860642945037,0.8654926122921104,0.2867474668338925,
 0.1865711437422357,0.22918260518557987,0.1847821433130542,0.9941603669509751,
 0.7257328611109369,0.5006621352500051,0.8373880741649844,0.31283338516032366,
 0.8543133929159586,0.6572260421335673,0.6146240727335717,0.2893501257283718,
 0.025414996966083958,0.6140782850729419,0.6098102079176119,
 0.9833400426339785,0.21050588304803675,0.7748840570819078,0.6272429114912133,
 0.9598363881699279,0.1309598560207088,0.345912318786333,0.036210718784995155,
 0.5706012046865149,0.1529966942818286,0.7806460226932053,0.7307408165741985,
 0.26057539337843805,0.2658143189910591,0.15663941066423537,
 0.7396273507769856,0.017864441992693103,0.4254912565523179,
 0.04858117512578852,0.8792864364294904,0.5585034690536648,
 0.10362401472444649,0.29747503395959607,0.10564116665967538,
 0.07894303939709257,0.3709159964472346,0.11175604543604745,
 0.8058739707995558,0.3255042639189687,0.04072819613288037,
 0.13201825724317273,0.6467799373503257,0.03372065449246897,
 0.2306102154290519,0.6230206544036401,0.1803687315855006,0.07235009647206014,
 0.480523152392517,0.4527229439830962,0.5239915037229419,0.9423857938955567,
 0.8820409081655276,0.2551147964237306,0.8856490557856491,
 0.013368552818415202,0.8534613648146014,0.7498661558332632,
 0.40786191157910523,0.6414677380164572,0.9777425513789327,0.8344006994002778,
 0.5359719268836655,0.21163417919477112,0.4989774105539099,0.259505784088769,
 0.31573537038955635,0.759097344540904,0.7619249581376979,0.424019496354049,
 0.20221920027224982,0.06435195082223388,0.9699124149861268,
 0.26138092669320123,0.870530639074928,0.25019617596396626,0.5598442870040382,
 0.963696033674357,0.9050514840452661,0.2512524719821939,0.7382049149576361,
 0.10488631826283512,0.3312549382290455,0.9072609887052463,0.6723080504837675,
 0.7008913426495718,0.1776758030094312,0.04232668635517156,0.6068801036781715,
 0.029360021637990674,0.6035732765086297,0.44024694068703785,
 0.7627919549303063,0.8756110758821347,0.9526439208741484,0.6120861510970936,
 0.010276999777886986,0.33044760198799406,0.9965856770405178,
 0.9575373000116187,0.09957036645265194,0.4496788632791746,0.9444373132538095,
 0.25301348774911503,0.389300659776058,0.24133299513686068,0.3326677921729825,
 0.6191770506947923,0.05233930572783663,0.6229559373734632,
 0.43641264446604067,0.9561552014117433,0.3829852512089761,
 0.0011590029769247279,0.6979050599661978,0.8083420564011612,
 0.38197952769919463,0.09915570902772008,0.6787748337730843,
 0.1447941583187221,0.9789532722009414,0.36507737588850975,0.3822114548478729,
 0.628708985163445,0.6610267503546131,0.3238635259958895,0.5033836347601442,
 0.23607794952825867,0.4128675774275742,0.3317376168301468,
 0.21516488024603764,0.7315471482643777,0.8012169698316143,0.7680329863158967,
 0.5529427701176827,0.3014832280768506,0.8411085008945188,0.7940277088638005,
 0.5877496932329047,0.639477973626282,0.7520570886665735,0.9532373487602648,
 0.2953579754334579,0.9878676321240277,0.6532683376705378,0.4244892568934252,
 0.2968963836234353,0.7772316892666554,0.22109453462069162,0.501551513574634,
 0.8926333901772202,0.6082769827444426,0.08051377782687175,
 0.006664986533992612,0.17424650450350554,0.11237666651682388,
 0.8400047418984522,0.14580900103159267,0.6045411627607331,0.9671331426535882,
 0.9145270739567772,0.5277300432798412,0.7060549065818585,0.8303230838633237,
 0.6494215632654049,0.7906654392605841,0.5235984672660651,0.7949361141307265,
 0.5355708914215397,0.47188385229738194,0.15572880013763402,
 0.6558741994325086,0.15956606315441801,0.28165299154950285,
 0.14240234962958742,0.8054814801342026,0.9466999943275441,0.0898865855652119,
 0.8582824621827548,0.9922878705952513,0.3831731873964329,0.5786601708556527,
 0.9499782305310533,0.9040291963728296,0.5244431203415632,0.28479929222908085,
 0.02947833627496499,0.8739083017594598,0.5617731658972859,0.3214935634668742,
 0.4757926700976184,0.3469935323753156,0.4985386072144721,0.3402998384032855,
 0.5012300312421722,0.7042240963686617,0.6721550062820185,0.7473701791393834,
 0.751246900149369,0.5374835704124302,0.7229201759475101,0.3622605904347207,
 0.8848597658638689,0.6047954694096149,0.3831338951805101,0.02139726381825824,
 0.4838741028557513,0.37887145718330706,0.7710330230392822,
 0.049278264240373515,0.6567564109677424,0.6137895282899376,
 0.6015631126384419,0.4905483526704101,0.7012079105041101,0.6880676518670862,
 0.6759778884586181,0.7655087671407089,0.7732281978621116,0.8718434646634223,
 0.35347739306030523,0.19358203232072435,0.1556057452547278,0.859979232490839,
 0.4938495713361033,0.7029490087619111,0.3384662559349394,0.3893790531432634,
 0.16119725298876564,0.5416838123000788,0.7650167243369754,0.7212751523465663,
 0.32908712177112875,0.18802579860505186,0.3058477187453871,
 0.31670707296300205,0.4899732275892136,0.518198360772044,0.29907300157680394,
 0.6341751150884525,0.11157265490613733,0.8808131696394655,0.5444967673252992,
 0.6357629248102938,0.16689313871282385,0.31891923712278925,0.545179269702905,
 0.10684874151986479,0.2817817713332954,0.41100860121291716,
 0.16774042147420354,0.8753881637728042,0.327582833143664,0.08348905070643087,
 0.5667415304505378,0.7608529066220213,0.04818216891167926,
 0.45960024560786394,0.20710993925790533,0.9644177734943691,
 0.5368994289111955,0.615179914063911,0.7783347217832827,0.3066759842472686,
 0.41795149661763054,0.6623023168554427,0.6060005108381058,0.2750755930460803,
 0.7055968057305804,0.7287754815534806,0.7487331488348442,0.21297640904275306,
 0.6585687067415298,0.9989075893874186,0.795008900663947,0.32939646489502217,
 0.2571805785468866,0.2768266089821486,0.2225341588653691,0.5419092134155852,
 0.03763969072064344,0.8917450863767431,0.8939228871503226,0.5040217567652522,
 0.6876411417058803,0.7089937987659509,0.0703981127130957,0.7770540854973168,
 0.12894765481811243,0.08669665773141955,0.16222136985076063,
 0.6741111018751098,0.7454363446060824,0.5764150417568059,
 0.002077064348129931,0.5728140405526299,0.3407955334928938,0.511382005089199,
 0.34207858148749626,0.9691659919651517,0.7988224588411555,0.290011307770043,
 0.4028674532475339,0.47203802413403595,0.2221515452855255,
 0.14751781248469298,0.42363273653641464,0.3612854221587183,
 0.019059926031115615,0.7600063103914826,0.6309702419804305,
 0.9816630112488903,0.3255452696128637,0.5617776935257837,0.10995798760763709,
 0.01724643052022312,0.9356561378189686,0.14820193216549937,0.878683982244574,
 0.8281960202869678,0.5171103599144422,0.015993849194875787,
 0.2452049265165408,0.3102324716559832,0.1940964133637082,0.3746488269976396,
 0.49659631618565325,0.6104663201320017,0.26571783962109596,
 0.34661713969903507,0.49137921152216857,0.09420839321816998,
 0.8702017597808149,0.2441329116899723,0.8051094610489246,0.3579834892522349,
 0.39041230586569153,0.19222729437971642,0.5097442645396529,
 0.2834513725545722,0.3097979749493216,0.8556058166178564,0.3887562370562978,
 0.3602441450949361,0.047823092822410596,0.5075073192949116,
 0.1538796009385227,0.3362227025681518,0.9727018183302001,0.7129587719368997,
 0.5825661635594457,0.8131701657814954,0.9882945994155936,0.3189727029386712,
 0.9917991952168282,0.15705950170335725,0.2433772960344226,0.8696575297392666,
 0.5869987798486946,0.5464399010696582,0.4222690314416999,0.3659941901021321,
 0.059624425958856864,0.004641434145700618,0.14673094877964776,
 0.0582176993059732,0.3871798492774162,0.5219359323777164,0.41553700634095236,
 0.3186854902870163,0.48856192054792136,0.7333416090494573,
 0.12012442073578444,0.029929701456303692,0.2366526134476421,
 0.18774738422349835,0.022017363503790133,0.8680733122010709,
 0.02664873606178486,0.6005055965056671,0.7029533813019677,
 0.36342707105015437,0.10380128566969782,0.8204165575576786,
 0.6985483916699455,0.7208768523519601,0.7355760946002656,0.9080037646991163,
 0.3720729519266276,0.4162332654247063,0.19746778818324817,0.7014447255437929,
 0.732110031704091,0.4561049321591484,0.8425419374509693,0.6213124048945909,
 0.6816031344870631,0.04082997182003856,0.9499736205106686,
 0.35204968447598284,0.5572736849362815,0.6196895000047701,
 0.26906606402235766,0.9874801841040002,0.7107021019189733,0.5047588254553961,
 0.5192033337076734,0.002199385534716436,0.48781926564165823,
 0.9572169966570406,0.03147349870109861,0.4476084173639595,0.8090900406322038,
 0.05646247894379741,0.7381208477629622,0.6754452643024056,0.505542116516244,
 0.43901260932938935,0.42339055321010144,0.11136336217491882,
 0.4627398283532773,0.43000044434523466,0.11770044960219506,
 0.15431451781305605,0.06665529358485578,0.2208030544796279,
 0.8399612797092538,0.6740507972880133,0.19736354160889302,0.1773032379338867,
 0.08681617607077441,0.4748917681849686,0.520733705819253,0.3960654200962874,
 0.38894099721605047,0.6629661685251754,0.23994206857848555,
 0.0023239672948273116,0.10792288397353444,0.21580520055083852,
 0.3121068881611718,0.01259608733148787,0.13820511038395444,
 0.1935996725752459,0.881286951177372,0.18705654684087047,0.5236378291230576,
 0.5576098147955739,0.27738297696158487,0.532048142206082,0.10225856321742377,
 0.14170888441246432,0.026599652473088153,0.273955897277958,0.638387876227973,
 0.825018704928214,0.6639625493724978,0.6723947014618701,5.793184176863253e-4,
 0.3655430959766157,0.25648501799313683,0.9146953298876008,0.920821870168144,
 0.4621393824119562,0.9470310797817414,0.5994300584501315,0.6387317946331942,
 0.21336198178014987,0.728813332151629,0.033451536560078154,
 0.9260951711382633,0.4327944357522344,0.5474962090259055,0.28059841149164755,
 0.6318426653785488,0.2418410525516247,0.8741808310987568,0.40705294006182924,
 0.6811005127682129,0.0386744281521052,0.6372988110487684,0.8450875615035665,
 0.7213189378005168,0.7408156845900404,0.5878327657115564,0.17586449569388385,
 0.8792635399278563,0.2671023095253968,0.928177241359344,0.9155269733651519,
 0.699979280565685,0.5694001533558588,0.07454940602778803,
 0.040613442066652516,0.1001349592247105,0.6045133409143683,
 0.3248262924336519,0.655091219383223,0.7820191597020053,0.7644087146024604,
 0.09863489612394805,0.6015430056915011,0.16555684753995337,
 0.22759910102874148,0.3049038401648497,0.365259043270461,0.4842968918795917,
 0.43737050346764694,0.6280620718502856,0.5045046497412453,0.7104274677090748,
 0.9231804566700446,0.7630264050340189,0.5297773919053117,0.19740843423045173,
 0.4704835519448304,0.7779975899018989,0.4454082189243763,0.5468152959753704,
 0.9045750121334306,0.8244762597221041,0.011231712168198849,
 0.6876721033984907,0.4499138993641648,0.43551332110514607,0.6422588469173974,
 0.36740408452747475,0.19062555015840954,0.7059154077153706,
 0.24153115096775835,0.019238143109555184,0.14009624974309887,
 0.2760520919575451,0.4067427076628257,0.981196213633368,0.6202676918560852,
 0.31438781866154275,0.039426753891163346,0.1440677708015342,
 0.053731120078623906,0.5212241067238285,0.21970769884280184,
 0.3491031295255921,0.6909421222175934,0.17219780131767048,0.6098052143952044,
 0.9715356193703414,0.6945801476658909,0.026125207178476106,
 0.8023231060518353,0.5314164899735283,0.33062632909519407,0.3103143439519025,
 0.3143432174922156,0.4903201523371846,0.24821868511553424,0.6054584604902753,
 0.6459709446141741,0.382969143313281,0.33317757284525984,0.16444538453928748,
 0.6041820778599396,0.6739893416752722,0.37716315720080573,
 0.29325748474856284,0.720942360828156,0.08414881946578368,0.2579742187119889,
 0.4888654861364258,0.07288216023510341,0.7570313065305364,0.7511938018232064,
 0.37616916143949186,0.911939285455017,0.35398768799555624,0.6089429224289986,
 0.8891302422779697,0.010079062802596628,0.30429471323465507,
 0.952832997634688,0.695806351599342,0.5520073953877263,0.08164013164805062,
 0.15329396126678774,0.30620934347515516,0.3397469727786634,
 0.9761244363560445,0.9362213218818852,0.8598801180812148,0.7010672030360743,
 0.544001465671475,0.3851686759633647,0.26098193475501186,0.3512746451319606,
 0.7365162707611236,0.48889657241844586,0.8364541091416586,
 0.026449945914554296,0.005366675250560338,0.47552740593229614,
 0.29114503969711625,0.6060322891016403,0.08256012485116326,
 0.9746967290287838,0.38263962313043276,0.903831978120559,0.35592715417295995,
 0.886602511515664,0.796527895604426,0.47616784957565816,0.24179417168043138,
 0.37347548388639307,0.38711537206212376,0.7857474592485211,
 0.34818002885930244,0.22712949048551467,0.739959502026169,
 0.44788425085844774,0.7495281463184585,0.5866571560957798,0.5443341126655981,
 0.7696703115048,0.6672805348248997,0.9638219613237218,0.6368831531974622,
 0.5625398475248695,0.16824571146237355,0.7004543171847922,0.2073077110059982,
 0.18375727078062265,0.9161777330927898,0.5250161861486397,
 0.044950555218339905,0.8898261712974765,0.9083042245241966,
 0.18377335269680228,0.016292001976179415,0.49996932177925413,
 0.5547176806856176,0.06588768851717486,0.969188509948448,0.8028749653595975,
 0.21967381770358163,0.5537886270224144,0.7496698998132905,0.7777041894966222,
 0.05278981114362502,0.28623130296731625,0.7458607418670764,0.675534480992503,
 0.4245031362052345,0.12805246830709893,0.8842376032639516,0.6505174398972773,
 0.021898100775696205,0.3370428278595865,0.9426906369498106,
 0.9278472573145091,0.5904962497739958,0.7714210695727919,0.8498679555424438,
 0.05067761635779655,0.9259516206890978,0.9074411626466816,0.9457717492604564,
 0.4763779051807765,0.6783002193540949,0.5960435623529196,0.8829374304315074,
 0.3220475874687374,0.8774746857761987,0.5814558416203425,0.5376444293532139,
 0.9309381535730096,0.968481537174527,0.9935191801348645,0.6395312332642029,
 0.8284173647158602,0.23292557431682914,0.47600078852293437,
 0.7600261445995063,0.25107753240951713,0.1077275498427045,0.6402959932000907,
 0.5526810289671826,0.7399138969818719,0.5763724708981378,0.8088891079443176,
 0.23823573143255883,0.7970030354717165,0.397170111237942,0.7943319068960547,
 0.34996326818959045,0.959704984408048,0.08042527346101636,0.9229344912850811,
 0.38586081467549327,0.5116964514982848,0.5352785707408666,
 0.21534712439581327,0.7037088833565005,0.9089935472925613,0.5944480926096859,
 0.9193241737132766,0.9695109835432605,0.5777467934100283,0.02481539732474991,
 0.8134409630469083,0.2763835706048371,0.17152582808826355,0.7388159141666095,
 0.4059782363476595,0.303874429878937,0.7651607224867658,0.7279828953753266,
 0.4189997011300495,0.5227373108380202,0.14827998875215032,0.9668815451086602,
 0.7123618485559767,0.6529230405902728,0.6496445087084238,0.8569290402251574,
 0.8248487673894697,0.5530186690012172,0.06267927974472776,0.8994380143588303,
 0.02582419138359504,0.927224291619635,0.1130774223635338,0.05158833701514043,
 0.8363379581830922,0.9341449094211364,0.04139549636759743,
 0.18006038967485416,0.7156037371829159,0.2805719381357711,
 0.41625262088051196,0.5906556858278618,7.43083347291007e-4,
 0.6643953986134532,0.3627450096832161,0.6921504957814207,0.924694547024976,
 0.6608607469513976,0.8139735213888601,0.9852966889426064,0.06875557249072572,
 0.3018615152490478,0.7980724558410878,0.4409151820344399,0.7502874829636885,
 0.6338506315856464,0.270579174753389,0.04607175336136371,0.649090005933811,
 0.5847612818483199,0.7595231831291722,0.9245459645570682,0.8278667411044068,
 0.9671232513144223,0.36314435128767375,0.03309235168916058,
 0.5845167095784601,0.9727946186221287,0.38336557854855824,0.8111625114931222,
 0.9335141968763097,0.009556813886034132,0.5232229694647887,
 0.7972792414649263,0.007219805813564295,0.7568183979377682,
 0.39979166317921655,0.7554888143660493,0.9090756744531271,0.6859144012258813,
 0.5364736464368324,0.3748777248364641,0.6075789423786168,0.5139412690877597,
 0.49098954158343466,0.9741045756849934,0.4324081766686205,
 0.11073163947261433,0.5192634591088983,0.5553828402885366,0.9585216913396818,
 0.5506157545874997,0.08849584061732119,0.2778420188718248,0.9385388879416328,
 0.9897183733395761,0.8136965614597074,0.726999704324543,0.07273422845442401,
 0.16144680921621402,0.9219350205968979,0.7278277934910261,
 0.05782877235716399,0.11437227619965518,0.6844470649661261,0.505004953755494,
 0.7972863077099126,0.6652560906222212,0.15783490989547666,
 0.31384320509689445,0.42842358077712406,0.013218027571854862,
 0.9713984841087937,0.576780423519327,0.9233732230570859,0.09285008427694497,
 0.5692026619901156,0.45537058842099765,0.16584697900193834,
 0.5890059540026703,0.6846038516282341,0.06272447548492754,
 0.06126325220145967,0.5371672009176576,0.9729649049865337,0.2908339060996983,
 0.5603724214846671,0.29921751053958046,0.906801051183678,0.30376075092229804,
 0.7738259255278478,0.854255139170908,0.045831836307898666,
 0.35087103663514996,0.9017682482960906,0.3481456488378819,
 0.09302043538289806,0.677187827143594,0.6684106812695132,0.6361037726124902,
 0.22643043134273666,0.6973860827390549,0.5297172194034856,
 0.18194785809096603,0.4936496581272525,0.276855976631148,0.7882531367037818,
 0.2581384195081856,0.05246171266268074,0.13634558153801857,
 0.35153439326353286,0.9219026150740329,0.600983423548356,0.5805679743265745,
 0.9058248649918595,0.21503503610736718,0.3961777503266285,0.5506581605064488,
 0.7529293408438157,0.4530902899698268,0.4892937090206593,0.4130077735640951,
 0.583484993943245,0.6856347762881367,0.11147694217606541,0.08116969198779111,
 0.13508068717471255,0.8037250912191105,0.011657334329926439,
 0.8246356850702822,0.6458647036287563,0.5401050289962632,0.2774236331084653,
 0.8056933709363239,0.32882837188798686,0.1590514356807483,0.9844151894673683,
 0.6029532533485136,0.11110269702718667,0.6692775537925761,0.953069547718343,
 0.4313587518327492,0.747860856861612,0.3408928987853298,0.2271522361263426,
 0.9426582658211597,0.3576128982274315,0.9075181506137882,0.42983284435668967,
 0.40467714127308096,0.13415219889691332,0.8753769349170517,0.632372113139189,
 0.8284797430265145,0.0063582926434619935,0.8229687950600321,
 0.1380556505408076,0.6855630083955767,0.6842711227370959,0.23942499496887304,
 0.3259503009012905,0.6499150503700314,0.10417255450397656,0.7403439718027255,
 0.21617673171947682,0.8701787423009344,0.1418200371697066,
 0.10241895909284904,0.6361772853052399,0.5669764998480735,0.5060318902945263,
 0.17267077475572423,0.09158858814350546,0.616015097766551,
 0.11988073869786742,0.9168150333524057,0.03609449707480106,
 0.23487666150930908,0.04190968972741205,0.9860362952670441,
 0.9408563345797201,0.4746932305104906,0.7222859919692541,0.44304877192658654,
 0.48981660273451477,0.952235748214812,0.45698310003273657,
 0.15354575430796547,0.16890915637043813,0.3654074343481176,
 0.012563432326332169,0.85951726581418,0.7791706759656015,0.6634746540178019,
 0.9124520713220009,0.20841006799073514,0.3008509784138089,
 0.21822055841560994,0.8968205906380534,0.5605746990888596,0.9244449674323094,
 0.3724091262641933,0.19868358161020516,0.9934775929229638,0.6097916690509162,
 0.3027620411987595,0.7881779880847389,0.03010474824801257,
 0.34835194031917815,0.0190651638569449,0.9753312714283439,0.6522710328512378,
 0.052532041126851725,0.28931420743859027,0.657085776873747,
 0.7268196116094163,0.6211033097172141,0.6604451949667908,0.5243143550578848,
 0.0942084712148914,0.3522668026596818,0.2085332787350368,0.2654410161122416,
 0.4897619741576974,0.6704335390735559,0.42978772967699497,0.9992092630728471,
 0.02908787631790921,0.8639340778248596,0.642206132420674,0.37633309816394467,
 0.46782650065164666,0.4354339879271605,0.9940485086231453,0.9867298894739709,
 0.03941195794139429,0.46064272361430836,0.42806327030410773,
 0.4760399398969658,0.6868311449594787,0.012288333028088405,
 0.22313044825695372,0.8722990873440638,0.16844047909408122,
 0.027584618807837358,0.33386304534847067,0.9013034391896315,
 0.2898846608607465,0.4998293769787796,0.8397730244688536,0.9812230379787429,
 0.517799324343971,0.6139962723821583,0.793344694846764,0.7691005289864323,
 0.8508849681929064,0.12776604242827982,0.9999386880435422,0.6396681358592816,
 0.16517230166945607,0.35399008987428326,0.5321228321146028,
 0.2949043497468682,0.7526110309361137,0.02929010077177796,0.8830925903859671,
 0.421773317026235,0.4498158191879402,0.8025967961955842,0.2633960000045259,
 0.9745942698392698,0.42053776126350195,0.9577361684185247,0.9066894385355924,
 0.29704548612714643,0.28899711582375776,0.21045648902811298,
 0.4296797207091958,0.8489630239592175,0.778343169852751,0.826598502506862,
 0.7536702365705938,0.0011530033551363505,0.37269282677965787,
 0.618093191475007,0.7547494598637444,0.8736880352574953,0.4826505295437139,
 0.6045422552687046,0.21111366371774598,0.7860113170586573,
 0.28469742496600137,0.5030159576131898,0.08005114180230977,
 0.15836545878795805,0.9294619534687644,0.49422855362822415,
 0.5510817583248544,0.5977520583486111,0.2132059061421574,0.9340990236331742,
 0.09319065942312821,0.7600701161208089,0.3137599825025247,0.9131275131874139,
 0.1731716296812309,0.7940636534263613,0.9216358323362008,0.9064664946702745,
 0.8600350137247444,0.7495886470810195,0.2292933371570719,0.5919787663103908,
 0.11661395737114133,0.38659413040253654,0.4913790889163695,
 0.35799216949221924,0.9750019549076282,0.6761554767540676,0.6071342941790693,
 0.04734872601574369,0.45438096285949736,0.15820507962492536,
 0.12248715258521237,0.1257968342202037,0.8061809636755577,0.3807749647438893,
 0.5292128354223327,0.8843168578273282,0.8062574127747834,0.3253623232315467,
 0.41444184844533927,0.5051444301255978,0.9524281577610416,0.1476710971794416,
 0.7916574930534674,0.1345415417091328,0.3004546419362596,0.7247781082686985,
 0.5523772486469329,0.7836843468649514,0.19114924411668055,0.9011616928715038,
 0.0800639494537243,0.8529898634619325,0.6612727208506627,0.982678785958289,
 0.3331705175673765,0.641327458821448,0.5764725827465808,0.7852988109933741,
 0.057822257131958765,0.7346839400103468,0.46358674131824296,
 0.9333044606339687,0.44236206726372784,0.4356314687731031,0.8597278787576172,
 0.9824434189487525,0.2661571465532633,0.4026573501009406,
 0.016611222316096685,0.05286518723153222,0.7124868057184038,
 0.7913271236718394,0.34625218424698334,0.1855221450781892,0.78681293079071,
 0.349654979674181,0.11987869158852527,0.9910944846145213,0.03417322389363797,
 0.404990347548565,0.03718825209512944,0.1142256499697264,0.6898067618743753,
 0.15748324865517704,0.9614005476410361,0.26759497792709064,
 0.05030736143922465,0.43987282003846007,0.6899106831694896,
 0.6774675480418205,0.3270525959420525,0.4926318399543981,0.9630532961946566,
 0.7295472931659184,0.10792030969320698,0.7911176539728998,0.4634068188629228,
 0.4416574659022652,0.734501437551873,0.546765274120778,0.3107954297158788,
 0.7046332460590591,0.6832642955860007,0.9945660195403727,0.1652219549171302,
 0.5954494027300354,0.8122345098032943,0.03445371957830068,0.6781884756974954,
 0.3021583395337857,0.6650720697300609,0.6375008044916173,0.2590386214271839,
 0.3966616995464082,0.8976785730508225,0.7259161457498302,0.7929725139092353,
 0.358253556315695,0.33342035182662333,0.9786615012736327,0.7179797726859312,
 0.8601662507193979,0.03313568313201387,0.7045627664947139,
 0.19536485214368793,0.4966874165706392,0.9917641313919952,0.5697646669736713,
 0.11998553596760109,0.6740480570743583,0.25846244020941456,
 0.2723641923173906,0.40522162037043274,0.19022817762281752,
 0.7628199282847348,0.3444885002433864,0.754521072278922,0.5831123597451303,
 0.6817547141351501,0.8228984467060453,0.9133731513685979,0.464038893810643,
 0.06722000660714511,0.5691754455913413,0.21016928082516328,
 0.10352018989475231,0.07934011799907792,0.3245465252194406,
 0.6326346729360599,0.47234443995243214,0.017149551736332147,
 0.6943681640703394,0.9992637821720871,0.03282636727081045,0.7439367730052537,
 0.9586287390595688,0.21536073290887914,0.9522234342426288,0.908294114120036,
 0.14558849341062774,0.6749585773567747,0.7581291426693397,0.7145412019523659,
 0.8848061992578542,0.4018260962370275,0.7957752378603176,0.7690401132205994,
 0.6068674295064854,0.05468672791820972,0.09115411024511755,
 0.6737375326039712,0.004896710103939883,0.7382922781540348,
 0.9847940498385506,0.9008381726602194,0.6655530844765547,0.5037715431516259,
 0.9230031913555051,0.24714771788678358,0.9470857800213909,
 0.24490397383154805,0.0946929627579941,0.27902736074970125,
 0.8839478263412813,0.7394722391046348,0.10233120428891684,0.1536010756165841,
 0.2855119566612895,0.3570815013372278,0.28752038288786674,
 0.11495529233038171,0.8343865786931745,0.5904892599780391,0.2974844819006607,
 0.21098013146343775,0.5724455250467093,0.3667474292234002,0.7693112548978811,
 0.6809353805598782,0.3709466722612702,0.2730011136764363,0.7510872780207763,
 0.3238574093568036,0.47781479151014417,0.006013608569566675,
 0.08192598860433131,0.8081285286958809,0.9948748500522947,0.833540786041697,
 0.00958860731843203,0.22310588003813936,0.7052793200493618,
 0.13540939223978687,0.7748291432369235,0.29455609085942647,0.638374941760194,
 0.291388806112462,0.9383218097165447,0.21681581836052743,0.7371872402033447,
 0.43964367431381735,0.6165370747995899,0.3134750513914325,0.8128975462062834,
 0.2401581177824672,0.365221871499955,0.14334584491012214,0.29676751976518323,
 0.8895007656952676,0.8659676968515566,0.6187450319794937,0.6019456413462392,
 0.5107917713411447,0.5346642148353651,0.582943717106447,0.4986310972634468,
 0.7285308616697099,0.46099585540193866,0.6823872287273794,0.9321671782031931,
 0.05589489746992249,0.7970756430612651,0.08107297469672647,
 0.8930993730779961,0.16022974248340327,0.9814962661754237,0.4802927548727264,
 0.587991933232958,0.37804631696432167,0.824106963466444,0.6161126027393813,
 0.8662026137690603,0.8522066874062275,0.44782055534809984,0.9115011490010747,
 0.41796204632524747,0.032139722971228135,0.04947473670606395,
 0.5253703347717875,0.40051280961011526,0.28592124101936645,
 0.8350067331131504,0.8641007711794565,0.9433293757373736,0.6837415388153506,
 0.48613092352539133,0.17745580145698958,0.5944072997322722,
 0.12591113960934386,0.5220399893300695,0.09019633505118874,
 0.5776025489582715,0.134998259167586,0.4048761163865837,0.9665544479324941,
 0.8289258603656304,0.7852648984822758,0.09265487604439271,
 0.06822220635837528,0.9740826534375813,0.6895020665969679,0.6623693214898921,
 0.4903008241261332,0.4803009941013564,0.5261321936944897,0.26446143728468974,
 0.8201374759188592,0.05108541273404865,0.18143901581136312,
 0.2324360174315364,0.8721802806728303,0.7392878863080731,0.6046021519832572,
 0.708193878999509,0.23022011982946067,0.30989980572419706,0.04402095407355,
 0.7711868571928303,0.6155922813789385,0.11809091104335934,0.6979429696754987,
 0.4702998806685128,0.45614836879216936,0.7026003832987283,
 0.08007645479483783,0.5956568309340754,0.2817576144830366,
 0.38909855451677033,0.6372090627430753,0.05553588214189409,
 0.5574876654231593,0.07685052913798773,0.36846828174292856,
 0.5474943658507356,0.9297801797490676,0.2681306341710845,0.4251916028701712,
 0.9907208865305537,0.21980255959464778,0.534083773999309,0.42208961148058566,
 0.26671576879510983,0.5230670073509673,0.3431668297811614,0.9754670798703861,
 0.6152185536037903,0.9936676648645478,0.5252610561252217,0.2877131943165364,
 0.16382710524187827,0.34961883978620045,0.17550027672707225,
 0.7672039516395771,0.6994510247055816,0.3364260791855348,0.6567391929877974,
 0.5128756749124807,0.478323609505813,0.03921802152662868,0.8359888428383988,
 0.746073988984433,0.9834005176450947,0.2404619144221729,0.8357213824864711,
 0.9922151153751637,0.8282658706650282,0.8202094266638712,0.9718539673725977,
 0.5134601166815269,0.2170964731435996,0.7385283401419185,0.5883094513433573,
 0.1328680015766236,0.493132262159088,0.920127790329773,0.3929615527423941,
 0.5751628103848537,0.6066508134818729,0.4541936321536232,0.31027466871310116,
 0.5990654463447387,0.806738775427472,0.4176291267791994,0.794855773269233,
 0.7487983648933806,0.48550899689446103,0.2988284551352115,
 0.03395082877607081,0.40427564503859836,0.9787544747539195,
 0.8733708355623598,0.7288722659745013,0.20834955857153914,
 0.02592437889827426,0.3808366362087324,0.27627818914109337,
 0.20068141643159643,0.4617419106545243,0.1227819201561644,0.8244359410329934,
 0.8888669291871246,0.32651458454631,0.173022495046111,0.8190187044535202,
 0.21472317013821196,0.6451994437727222,0.247592199557776,0.6354224525470574,
 0.8014242432621276,0.8679625528661183,0.852213874878805,0.9018429803628857,
 0.730402949878382,0.9730035432356705,0.043837749854382935,0.3798160487174562,
 0.46121083327725243,0.4937824768109905,0.8533429146537327,0.5345237414475822,
 0.7274160507206402,0.75360568664843,0.07817017514636693,0.5133478165327968,
 0.17765580075609133,0.3984832213150773,0.9790054328721505,0.539634470886396,
 0.6240534108746723,0.05624863884602438,0.7215980998247448,0.8325252887791853,
 0.16630277648230185,0.6779931043199998,0.4704867178938057,0.9006073258977977,
 0.5462312968230728,0.1480680674054735,0.10491521272069504,0.4691254123434512,
 0.03843406929905102,0.6735148402079146,0.2070942685210383,0.3914735604077202,
 0.3886448227615924,2.772688733458928e-4,0.76214204773516,
 0.050518918626303644,0.8709282013001391,0.47447204180591895,
 0.15726462750603865,0.3468314569580192,0.5449440934207678,
 0.04177161508693106,0.07093130979565254,0.009259932519254832,
 0.02051956242901054,0.6063674990322454,0.4157169113232507,0.3873850914842628,
 0.5610103069865998,0.15893047373820102,0.6934301626173116,0.7105331451597464,
 0.27506240655356384,0.23698192261731943,0.9605445333406921,
 0.0992010916734739,0.95120140999354,0.05477589397195881,0.588062863518433,
 0.42870187801479087,0.39366685557488057,0.5746014373513019,
 0.8027610003407942,0.45419214050655476,0.8890366579466378,
 0.016737953209511214,0.7735467071037759,0.8916614585689818,
 0.6754614672157865,0.5333845350514168,0.5257330022574483,0.9290623967930199,
 0.9955589057422671,0.37693789966190594,0.8063556770936636,0.7386975810273287,
 0.9235525081636105,0.42876430575075597,0.314485676272954,0.45704002533840293,
 0.8775763217308961,0.7124352649964938,0.5130897909417363,0.11473618649288442,
 0.6153682320121395,0.6077427374643914,0.7586982190425919,0.1789115329112585,
 0.45815229373854105,0.018713017533397158,0.5298722203605917,
 0.9752897044713976,0.15318158604009358,0.3633572254547969,0.06122086568842,
 0.8099015429189959,0.5596318709909613,0.16324325772380632,0.4593345885325779,
 0.6442854665401425,0.30545515934448686,0.8117410160566423,0.3960013950605048,
 0.5544363043274476,0.08332139436225894,0.40696456469191467,
 0.9704821999386598,0.559236423227012,0.2462344180372058,0.6013876274671236,
 0.06126677068823061,0.5887579076793884,0.7196502925683451,
 0.38151916059109914,0.3987861492689935,0.8222014851862222,0.8986881734484929,
 0.28256917388089875,0.5382429656711751,0.7259989795110604,
 0.08707306667296189,0.044696888352499964,0.6942178831626176,
 0.7447651269432887,0.7281609624404713,0.5461529344711135,0.2156697232073177,
 0.4473424270604336,0.9973757786651535,0.25284248731067527,0.3382825512090669,
 0.1799386354642144,0.7667196343348164,0.754343030313264,0.9493882668738504,
 0.3161208678816404,0.22029773392831142,0.8557783001091945,0.169304517454808,
 0.8848613355573449,0.13397724945692024,0.44100045190384896,
 0.5346543525529727,0.5852941026834082,0.9401377093666756,0.2608703737687119,
 0.6660937090646095,0.7294711480712672,0.8439527396183157,0.18266554009932645,
 0.20866705548302467,0.9514335015791187,0.06891924480830025,
 0.8052705550097846,0.9719663538687077,0.3520940646607005,0.26491146720043723,
 0.5926185421923609,0.6851703757229088,0.8486778217089495,0.13443515488529623,
 0.86622516938783,0.0923376907155582,0.6332660694221751,0.4875922569803637,
 0.9590842152785903,0.4300573038886526,0.7043540240123761,0.6577151834054337,
 0.5017058188845699,0.45007287987673794,0.018078350255997466,
 0.789481624926205,0.16337375119516961,0.6994371164914798,0.8873597726462349,
 0.06298795972928517,0.49031178698585953,0.47375074012029283,
 0.8323994416744307,0.14854331708409751,0.8635514047851394,0.2804732919109896,
 0.6546804061160583,0.9457841294877407,0.30147618652209285,0.7185975609226354,
 0.9349209678873439,0.6889117019554826,0.38710550527359444,
 0.26523244372157395,0.9114822287333084,0.4473575640170341,0.5451368992048216,
 0.8632116935453623,0.07677550693816682,0.6818326619045991,0.619094684823819,
 0.39363062566574714,0.7270703254268446,0.07333772134990757,
 0.6424761966886378,0.41489902743888607,0.9495421786123811,0.8852334712866753,
 0.757423001675718,0.47545621451748143,0.495078725560506,0.8000917296137549,
 0.8390159039088341,0.802798430905248,0.07059135200274691,0.914079201949599,
 0.41651313318745264,0.4260136488764208,0.2588029995786707,
 0.31438208725608574,0.7364202091056888,0.7115947372716007,0.961355858330736,
 0.012103040415663235,0.553940278385064,0.46396044364608124,
 0.042204255053577944,0.28860286590975315,0.6583348999025644,
 0.787464344588579,0.40592901961734373,0.9153587568260823,0.35222196058450905,
 0.4831110107415969,0.3673313409704798,0.5224121102071005,0.6520275644605878,
 0.18275437026312424,0.28064622802183603,0.9814841959809427,
 0.8785372380753785,0.29800913901849835,0.39024871981061093,
 0.9565942165134363,0.4589263099933337,0.2877318728540388,0.4970208950030949,
 0.47986546606579616,0.3801492161959018,0.49248730918869177,
 0.9550067211070546,0.31017311882634313,0.36472106796409176,
 0.21856803411578807,0.7147616098377009,0.5085273834918209,
 0.41703757915595385,0.037760943095031085,0.8989976701527279,
 0.560557900309102,0.3315363129612041,0.39528880724768756,0.5008664297888993,
 0.31981649950191604,0.6637945002206503,0.9424904595075525,0.6042399914754251,
 0.8028786676556182,0.36620871134945865,0.9136371472267477,0.7562428371668113,
 0.2278493846622609,0.5801021769392334,0.2899419558684685,0.5323417165585118,
 0.4502054668715444,0.03980175117936713,0.27618384451166555,
 0.21833986235097824,0.8718121620806706,0.5285507284109072,0.9717318904546329,
 0.7131299357706629,0.22815078121691967,0.5259062930809848,0.6900602176476733,
 0.3885134039479373,0.07307745019883627,0.14509295098049946,
 0.0854382245297508,0.13170432157913947,0.09467707406296422,
 0.7985024335262035,0.6195612277197045,0.7208344502347506,0.3466639190594535,
 0.5344658799459089,0.2923325810339392,0.0884353961329607,0.8023518836549633,
 0.16768723719482892,0.3077098065017886,0.044008905375328866,
 0.998879956462655,0.5668205447207139,0.4921859369674749,0.7073827324559883,
 0.08169836717260959,0.4760327001592273,0.45811303111317125,
 0.7652378689372472,0.3456041417760891,0.8128584494439095,0.588775131179263,
 0.5280684338540139,0.7827188336051514,0.914368959557867,0.8389212035309841,
 0.8658431352061535,0.8061666070544423,0.8922699303820444,0.14657530905456828,
 0.27022780199742436,0.49569140099318976,0.5100081172232098,
 0.5798144100464545,0.69113095111088,0.42694839878537194,0.7276992553600986,
 0.1409300303155513,0.13776504514468302,0.4114444790155034,0.298270012675499,
 0.591142703131752,0.7252237809629398,0.4927385105669353,0.9084572685709984,
 0.36481388642223755,0.49931053294182415,0.34732456676663626,
 0.027914989135812918,0.5444248385368671,0.19009125459340592,
 0.018717935469643288,0.9149376411782937,0.4436303829803435,
 0.40012193221478176,0.9646312688199461,0.6787687489573574,0.3331246675585351,
 0.24758283182555074,0.8123302709961553,0.9607483057472525,
 0.25471774420259874,0.7319794608767757,0.572276198418201,0.8087669514869116,
 0.9656350603656528,0.3122530626504124,0.10838312763798086,0.79483096292416,
 0.8042686760668717,0.0408188966906502,0.33998827203860116,0.5822477882072453,
 0.20938890259207,0.8134046120470101,0.9796369407576385,0.705730777462828,
 0.4892469034465843,0.5823029598069179,0.816301040103861,0.5592671035684065,
 0.5909789891120424,0.5911731925899447,0.9552348782353179,0.7896317935002861,
 0.4482007105715373,0.7769522446242806,0.4873241387543684,0.23096549110605769,
 0.1973875798471757,0.9046377255960453,0.2202075858204937,0.8017390111643069,
 0.6174880430902047,0.6225285823870615,0.6736443456006928,0.9677884278130322,
 0.16310446430710757,0.49674160856991667,0.694142609377196,0.1278915328500897,
 0.009924460768820698,0.9181687203095817,0.11131767417943306,
 0.17942421947154852,0.6218353529037057,0.14221072614308672,
 0.36709252446240925,0.21503242515193488,0.8522084835848822,
 0.9083345380160373,0.8388736783037635,0.0851040168049585,0.8246201950557943,
 0.08534904980427194,0.8564425625721142,0.8117410961321322,
 0.17024128221527146,0.10196078328458169,0.6791363501607751,
 0.4467848830893816,0.5846903382590098,0.6256350629707117,0.6175315901074346,
 0.7519068908679545,0.6353966350630625,0.9589307441307967,0.5953401582733909,
 0.6866900000861794,0.5141540534745874,0.9827131096934446,0.8879793360569419,
 0.15935371013778754,0.848924393618826,0.08020633793143772,0.6783902523956136,
 0.020254176420010372,0.3711149605517753,0.4478305504101827,
 0.37851641053090457,0.6338345538164142,0.23531173738713895,
 0.8251483699553538,0.7321302060995621,0.972601335732963,0.9275296779982873,
 0.813903424420231,0.4699320255762339,0.3120017991416112,0.7449044340341154,
 0.3960090039010543,0.14498121869471414,0.7380574896211104,0.9129389529883012,
 0.18490200236945875,0.22359194390280535,0.2746512384339004,
 0.7472557707000965,0.0593708449112037,0.40942810141481334,0.7573393264953915,
 0.2550410088389671,0.2358393346054941,0.503703985345874,0.33708314009116647,
 0.836993992015176,0.13075353768709375,0.20950047645892023,0.967502619115738,
 0.2872986799853511,0.6097638824678905,0.031036923230544522,
 0.5776371374539659,0.7320062881582373,0.407798006268575,0.02305636435429892,
 0.9855549859928705,0.24397519425698921,0.7671560981342699,0.8599812250333554,
 0.3764975271960477,0.23079909440570512,0.9256418573919097,
 0.15998174776275387,0.9073198905570621,0.5874155798762688,0.8869941706194777,
 0.3349152284100724,0.08484539506890043,0.9818918943605613,0.4148325075609698,
 0.956100362179064,0.22795326567756125,0.7641527155571815,0.20227203059918852,
 0.9346786448854226,0.22540105111307596,0.7042158658525034,
 0.23965988558150408,0.201647621014989,0.8949662460053719,0.27357194117448547,
 0.05076633331342806,0.49940319142491707,0.20306302039629565,
 0.9202905094444982,0.9644860769406023,0.4918715012541178,0.4743140559014781,
 0.28499718793482887,0.4013137746224855,0.6705693227647306,0.4009849143158566,
 0.1758791498464024,0.302453826796095,0.10701372757359218,0.9593092055994292,
 0.31287348258277525,0.8514075246844615,0.9797047904035845,0.765587964161654,
 0.5259067501040909,0.244308161300815,0.05885488011237994,0.2908979355742387,
 0.7838011356467802,0.0398535498162057,0.6474388724492413,0.7053000711606007,
 0.48978378512501286,0.789521325346941,0.21760826606928052,0.8286145998296887,
 0.4408832756179901,0.3607388546573669,0.2914060703747202,0.39466698962719815,
 0.2883832281832436,0.2694428533803481,0.9177871898563655,0.5353801776257577,
 0.4105739845558578,0.030108498570849074,0.5680867953864219,
 0.24456712334625674,0.6000475766856738,0.5996044321881061,0.8305513013488908,
 0.7523078707228759,0.16882660540122596,0.7382347673536417,0.4129944703177949,
 0.41971668923654426,0.8862275351563371,0.6745732843435994,0.788535001501987,
 0.28226829914419493,0.5220275011302998,0.7976043699173765,0.8585238813120247,
 0.6384104985230472,0.4541515479260394,0.9894855225202293,0.19932316958116636,
 0.4571874586429773,0.08051519138818308,0.49240814721488935,
 0.01402974562333792,0.4108055633243377,0.12029295006412055,
 0.6947652579985474,0.25556886257555267,0.3809349259278042,
 0.45485106433465694,0.5040699829568783,0.007277278440877999,
 0.6135708665228657,0.7824082358412497,0.04137324751580285,0.5081329753522609,
 0.15820048007704102,0.38457279929758215,0.7737733928160978,
 0.2007578443899869,0.02335771409512255,0.6921025525829719,0.9730862522166972,
 0.44313762901150433,0.8156099235690731,0.06998363281491116,
 0.20732278731988818,0.6601400801988263,0.23248596953600098,
 0.20007976526771087,0.8683297204061922,0.8000917617309364,
 0.16613927116748384,0.11466192783413032,0.0635479413065192,
 0.5530944976997624,0.025426106704732,0.8051925075086664,0.09226460334949871,
 0.534116768026597,0.7465156810168899,0.7452071517486129,0.21803639138734832,
 0.7120701253808582,0.9407978567002498,0.9416813715528287,0.6807178286434477,
 0.6707138395046918,0.36454367542740906,0.8919739407640751,
 0.18707723448910962,0.9444061441635607,0.05183948058699617,0.426996616641903,
 0.22648875755911413,0.9552209480674902,0.43873317702438197,
 0.10512176950006735,0.07684560427775644,0.8107933144703761,
 0.5070362521276233,0.380021247675875,0.6339687755779957,0.5042186589166245,
 0.1513649414726551,0.14087554824018123,0.6683416995856604,0.929777350240951,
 0.9418186128390381,0.7672514461305203,0.3137935890984992,0.18325557877596843,
 0.25777880651305607,0.6767585150867318,0.8175836733734073,
 0.13148729297634165,0.45313974289475134,0.8030837892044831,0.762181977842283,
 0.5592864680611753,0.4428635055024308,0.9422799998741039,0.03279498342670706,
 0.7640348154085429,0.03635579375137121,0.550998151039328,0.065660954583725,
 0.4056173374329196,0.9597965687173855,0.7259745392600951,0.7307093054778048,
 0.9688789480852463,0.37623558627243725,0.046490575306393334,
 0.10029269730640789,0.034698976351850264,0.20699583940475064,
 0.09288771948741903,0.7306021588930464,0.1908413230789362,0.3289567057389313,
 0.9518984085421359,0.3181328824121171,0.728443909929816,0.22899876629646443,
 0.0074177687087506605,0.4200071860117144,0.9961976678241532,
 0.7023496082121616,0.9302397427897079,0.0016038959434377942,
 0.5656871265674027,0.25323058348990957,0.8381921225218822,0.9836728723656699,
 0.2914679323347419,0.36830547550134907,0.10144805478291119,
 0.06917286729894756,0.4984878846293289,0.9901506775058515,
 0.20487595848046258,0.6732866105543976,0.9419228520713931,0.7099428925306259,
 0.4904080498540506,0.2232343931359464,0.29534716507987735,0.9167673579599992,
 0.5047547851707942,0.8734484966013724,0.293064075179013,0.8652353189920768,
 0.10402496349243451,0.878106455856661,0.9583774297328643,0.6352995159959481,
 0.5613028867662161,0.09724067843827522,0.08273889852708594,
 0.7024897430401192,0.2316554717301953,0.8551020673237879,0.8235817881166188,
 0.7017462826574717,0.5370565208312469,0.9530456676071772,0.4118609507536378,
 0.7093716545809926,0.44257582363604764,0.45524571229098665,
 0.4862745351804797,0.9543923659626733,0.23966109295570026,
 0.38384510777912373,0.46576111485281757,0.4638003752669515,
 0.47196789332258626,0.7909577388964733,0.6482019195044432,0.8348062850137123,
 0.09100726775190071,0.9055892004109068,0.1721073375331602,
 0.13411383193080728,0.573868752589615,0.5458297048555583,
 0.054109661225471006,0.7530578633097533,0.8410852743825737,
 0.1679578468888836,0.7496311701532516,0.116454105996454,0.4467077237033982,
 0.1209483498842876,0.9941383838284108,0.13902135131702686,0.9396668111220677,
 0.016669034299302443,0.04100082410975103,0.99970560293229,0.7928236621119822,
 0.56919123364709,0.3487910951790557,0.6674706990912495,0.7278456770501458,
 0.5773369824234126,0.7360255077152105,0.1475129213201064,0.6456998178304674,
 0.6935952188287611,0.583130044418442,0.06555859790795127,0.44187239616720964,
 0.253080639284748,0.21255780418611026,0.9513183210308331,0.8957710871948449,
 0.06430524281848671,0.22515927806699443,0.3439775793271378,
 0.4365263560140794,0.5232432921176043,0.24927679931560853,0.4253496327337415,
 0.692210168972378,0.5434046695871331,0.195961091168129,0.46386529518984354,
 0.04004445424586123,0.3616586060581122,0.2311890418037319,0.9688201894274138,
 0.8907948881936371,0.051449310334321985,0.006202677904329379,
 0.42750149951520533,0.09118990463136989,0.1401784859402314,
 0.43175065870420004,0.8425753508697845,0.5850263875901542,0.4485778594925057,
 0.696325392900055,0.10809548861538099,0.33170716657391774,0.8136135492436718,
 0.45117178781995093,0.7772917610953527,0.8916411436769603,0.6304897385258562,
 0.8269352547800767,0.5078285482853239,0.3065859321323756,0.4569861509579364,
 0.5854953252726098,0.2276600410563363,0.024198705302235446,
 0.5937410760976667,0.8626255565753796,0.7780041251278147,0.16383533545424145,
 0.29946497756529267,0.7731798913284825,0.1022835686464253,0.6850867367604097,
 0.8708873623263232,0.8100284481752035,0.2343774125652085,0.6120521350954042,
 0.16373068513728806,0.029821794869750118,0.5873571704018851,
 0.5691222745856385,0.7161213433184083,0.6517224704525126,0.47264254252083926,
 0.7593600531204586,0.21442564339541736,0.5066697042627157,0.4399012716764854,
 0.8977644146186885,0.9421950896470009,0.9837888642880053,0.33396611198387505,
 0.012110900617897213,0.7851986422858855,0.7957590069596195,
 0.9190262311086267,0.19162177660547242,0.5435717266402192,0.7210853145577651,
 0.6766576967969175,0.5401377779269542,0.06943865898224544,0.272172059155642,
 0.7692944093835451,0.01280711703356896,0.4116266406370437,0.1530013824094293,
 0.8532674275898704,0.9057265462239048,0.7483207976511749,0.08513579963166129,
 0.9016065163198569,0.5005199019138441,0.8734188289131374,0.7196990943648301,
 0.10954271307998098,0.2507815837144336,0.59575639773701,0.36978967256585116,
 0.32474809487893375,0.8991361624038696,0.649130133576348,0.6124625977981772,
 0.33752294054702825,0.8041093974119572,0.5875520636143137,
 0.24071293873179067,0.08179312621388135,0.48015163546840467,
 0.6996964591469983,0.18231749810697762,0.5092675895864232,0.6306235869371044,
 0.05423570657003085,0.011350180408160493,0.0016250991244030066,
 0.9944149395380998,0.6042859545671891,0.7695260400684567,0.8665668382845337,
 0.5052602150775206,0.4476584024687196,0.6120069040002791,0.7310492265918611,
 0.3215769861373433,0.36441431438009253,0.21894608714470143,0.67985373491934,
 0.9217075169917024,0.8480524484775028,0.9248036229460319,0.611688145656967,
 0.22241845747142852,0.5121731787255906,0.17475410024946558,
 0.4669485511314799,0.04443273318910512,0.6759936490538934,0.9555932869820727,
 0.35461879044388533,0.6094286673683339,0.38905567293985166,
 0.6728692105175764,0.6919368394890508,0.2840558015944932,0.9270253787336303,
 0.9502690450733033,0.17582350376258427,0.9627617257290283,
 0.21271931420116097,0.10818703659579487,0.7828939740187348,
 0.5836494948491902,0.9853123641314403,0.18140539744128426,
 0.24603848827701524,0.5930805440417818,0.6763290262940513,0.2728600757410421,
 0.6614581702773149,0.3904127482707248,0.39180119415254755,0.6056761983213973,
 0.2647345223674634,0.13438625572471807,0.9941985899085897,0.5811820680376353,
 0.9372092999760351,0.0069318964023222795,0.5906943884632313,
 0.36475136873381264,0.37660912924656664,0.022345652950032213,
 0.9206103027127246,0.9356063703292148,0.42983242004706845,
 0.17521454258323066,0.9591237943495762,0.8516385424844952,0.7757394290248726,
 0.020923055682959935,0.25259801321081865,0.4664201128715163,
 0.02437063512160642,0.8362093354149643,0.6430010449926155,0.6405144980330311,
 0.2991134934228754,0.9890439221414787,0.9394700482921609,0.3973160643895446,
 0.9506699125324048,0.8716902321766,0.14914579586296217,0.366785271704112,
 0.613282408223452,0.9701651597196904,0.9261622488868652,0.8612994685324772,
 0.45751534604200605,0.7421198087197107,0.6706815483861682,
 0.23862088912851698,0.5853539378919482,0.40807259231406123,
 0.7595719284031525,0.7364735650074749,0.06894970849623627,
 0.09505985427234886,0.7837620449988368,0.31148353522664407,
 0.17527224718829104,0.7033464491261785,0.39700612954680436,
 0.030671400950446626,0.3360072001098421,0.8393303787827397,
 0.2873760706243308,0.7056665299984654,0.8826459983821676,0.9858643405880027,
 0.8229148598413973,0.8876184461560976,0.4677085567172604,0.2679871322269882,
 0.06234331424781481,0.7309881702725679,0.6350516558629169,0.990519570481797,
 0.7595518559645638,0.4370895242808215,0.38888883320072865,
 0.026331928712748143,0.8053252507408435,0.1591329821406514,
 0.053133478202257445,0.5558569470923413,0.9982855594207135,
 0.5888550976196382,0.6348833147926874,0.5305121632508192,0.17624302826000549,
 0.5088786753513944,0.8422814083511445,0.665190843566698,0.8377690565718814,
 0.961907159107519,0.3086714655604532,0.6798023574905363,0.545950748896956,
 0.05284053552792878,0.5731966524898635,0.5845283456647954,0.8454223138412249,
 0.10274009888671543,5.825397123000453e-4,0.2208103664975638,
 0.463401458068026,0.003988620042479907,0.9344802401780626,0.7493323204955792,
 0.5501433996796935,0.9975147182495907,0.9287047972145495,0.49202023366437153,
 0.9969467599225831,0.7772583519051885,0.12894425631679485,0.8472661093706843,
 0.6461073009596608,0.9791194280703132,0.5751098981312563,0.8897182894181079,
 0.9129970873717965,0.6942653823302153,0.16295923740907958,0.9774898192690192,
 0.016076053425123082,0.40341390717577874,0.896376475224067,
 0.2270263648914994,0.9969863201775956,0.6739631421755949,0.8946037815016373,
 0.8048473934694566,0.8918172960875858,0.5298212829649154,0.05683951275151378,
 0.1353401147849579,0.02671067360488144,0.4422877109727773,0.5371968530285443,
 0.9957770969254706,0.6350091320312844,0.901648539929679,0.2180290836707809,
 0.007009028579758159,0.3197613318893672,0.30129117852310794,
 0.6817330144014099,0.0639016627836061,0.7869358417963968,0.9183306572461172,
 0.5268696417864688,0.699317470676366,0.8451283940108684,0.1009527014702516,
 0.365920477393742,0.29071990345478227,0.024158055287217373,
 0.7076905014690649,0.720445210211938,0.6558907608717882,0.17533750704199424,
 0.5845735760831586,0.3563102088167458,0.7585862768967239,0.44619818913632914,
 0.014096071805901444,0.806329159039117,0.3169085022542494,
 0.15525519773323793,0.41217484405576776,0.5428416006949619,
 0.21383257541411937,0.9737233978306521,0.577684525252614,0.6053182098328457,
 0.2281909325850272,0.7284773699623115,0.9306559247977733,0.2014636856550691,
 0.39517649882637873,0.1545297894229175,0.890723055352424,0.9785849887938154,
 0.31193739810867505,0.9214945638685206,0.2997373322621768,0.916021627461536,
 0.5023549690447219,0.5659236008452618,0.5651888790071118,0.6588708388644996,
 0.18522595549952903,0.4174426832677425,0.20112680701229202,
 0.6862214540187965,0.006796996569295777,0.28542273178630606,
 0.7771655011839934,0.7533866525791263,0.15383322666533095,
 0.19590023067522233,0.2847712045082359,0.023627291607645096,
 0.24708661232292206,0.12332549172282092,0.5886435474225779,
 0.7288894617118934,0.9856240611572458,0.44695168615286485,0.953522757833275,
 0.5943067812471255,0.6245046239550189,0.5714663794423469,0.07788793563215501,
 0.6661939060242597,0.9341418973334014,0.9042205005764486,0.6095089841757251,
 0.6165449339062875,0.9255145814870716,0.8768142802457247,0.27436986816782394,
 0.8155358757503572,0.3380286617269007,0.01733740300934772,0.6076282630015082,
 0.7683823239803846,0.5566906753170506,0.12635654190424062,0.5206093988219335,
 0.9482493176544442,0.14707562407458785,0.4410598205025775,
 0.12623649073389542,0.09901542069518832,0.3309558259737885,
 0.6333547162717481,0.30449541383536105,0.43025627154057033,
 0.27208566569007653,0.9497392087071048,0.2529823383098255,0.7363856192145946,
 0.8065534126437758,0.9352504415454098,0.5957782538971327,0.608268901241862,
 0.36700373596905533,0.3819199072366063,0.4412895218998547,0.8295117193142759,
 0.7174511459915172,0.6768086476993214,0.7916317023491017,0.4800518437792559,
 0.5701900180066041,0.34336281724733886,0.33555940443153764,
 0.6461593233910086,0.8971153359781804,0.8940145436610605,0.9585444888311994,
 0.8364972935119466,0.05558392862652939,0.720452452450083,0.05616897693459766,
 0.9748631955497117,0.04724705701915266,0.31650240733128043,
 0.8378928659083764,0.4799065530257285,0.7861185776299021,0.21436911607440912,
 0.028826715248625567,0.7769679645043684,0.018688499478437204,
 0.38317631693900034,0.2765816691586296,0.9935377073936753,0.5301247511244894,
 0.6974738669541074,0.30712632316350164,0.04010811867202391,
 0.8116439266613533,0.6910353608767084,0.6700164313840442,0.3205949574545326,
 0.6415454074903965,0.27366938330137236,0.06268313581898821,
 0.5739171076770053,0.21507250318867321,0.35845644983560543,
 0.5291396561085127,0.6812973589216335,0.33267202882319735,0.8929696244093692,
 0.45517987917397507,0.6684249278866298,0.6480007415677923,
 0.02949476975930232,0.6723081308916199,0.9438554337137222,0.2993506035234349,
 0.1410180392790803,0.7537252121742402,0.9375777483469424,0.6458226644113877,
 0.0049459613298079574,0.9828469776791544,0.8392615507249587,
 0.8656834924587902,0.06304045627658605,0.01374008666016191,
 0.3562088425160349,0.8674538839445538,0.9543666229887076,0.9478987282761921,
 0.7282087856811488,0.038959162610273435,0.36228081334617024,
 0.636476947966239,0.19505808881830977,0.2995035598843806,0.9402078514182978,
 0.3565309056576055,0.5428810787581202,0.530905811394641,0.18362463348222235,
 0.9959599073729272,0.8094509179834934,0.15030374010914693,0.5911234747543032,
 0.1849880300247404,0.8178049737095473,0.8483716713528535,0.7222295014265724,
 0.938203928697041,0.7702498882956467,0.32267014116964254,0.7649236179217782,
 0.7671105216578133,0.005678379000710887,0.45988533777889207,
 0.1732063484695474,0.863609016158489,0.11153315668548636,0.5275838944269671,
 0.5373395640693414,0.5079149526775227,0.6590182159035439,0.8303662126035802,
 0.19757605474225992,0.4098892478214212,0.25943958282191115,0.739160510331982,
 0.07850520988293652,0.6420963391862403,0.2998465273856379,
 0.030409842911164442,0.08030695476113181,0.049319399628933724,
 0.4382368099203533,0.6584605055946611,0.8340841440605139,0.16845594738977876,
 0.3300716912134848,0.9814491586770953,0.15472836514686805,0.766019794239065,
 0.5142918412915582,0.1040631785655689,0.8587490585820893,0.9019559755433476,
 0.7327041476644226,0.37266804406241893,0.5830691082000079,
 0.41732929574524313,0.1913379288746726,0.5796067896756387,
 0.11265181114587741,0.7531311190950338,0.3008449709683929,0.9731435196375179,
 0.01995106692515114,0.08821016070675958,0.9233610306891812,
 0.7472688019239717,0.9561741636347179,0.27338983307895315,0.2758631808651303,
 0.05963401768592247,0.5193193759054595,0.6634769201050659,0.8557033497746495,
 0.07317477403163453,0.9421169275512348,0.022618574710862593,
 0.33503716468616984,0.38532377095033254,0.6965863069618378,
 0.3406519375996421,0.1396567179053132,0.15895179626611655,0.8092499294577593,
 0.17344557947846295,0.7150997143602885,0.9917078158930761,
 0.24498197805828814,0.061866721868511165,0.16800367035554975,
 0.7730126704220726,0.49807111222089373,0.7687360000156651,0.6345699732542682,
 0.751551019530297,0.1240540667122364,0.2280717714524363,0.23492781625441272,
 0.34159334175184586,0.1041532821334481,0.9790319857231915,
 0.03286963181517144,0.18100751269296278,0.2854314352420614,
 0.6454308918785662,0.18786313489250017,0.8255108416386284,0.8679520521557641,
 0.08290593530616908,0.7282266551178593,0.4863590195639025,0.6049118806649757,
 0.05403911078380619,0.1899058717088311,0.10224515200133277,
 0.5907402483834494,0.09205768536184589,0.5382047116277455,
 0.21001600963791744,0.07268033741666302,0.7065282547684654,0.571193306326577,
 0.9619285527138945,0.821170815904979,0.24728422302170905,0.6843740661853271,
 0.08242285579147168,0.8536858347920215,0.6055078368480824,0.8691693015057397,
 0.6695288453525886,0.21157621018052586,0.23031381523461025,
 0.7184095321318538,0.6479080105941358,0.404558997149268,0.9398632491576812,
 0.781978164221803,0.606499665581864,0.5548362470570681,0.48016811909138535,
 0.9731527669116853,0.5720930652953297,0.4192154285604476,0.6438365951929639,
 0.7433559969124116,0.7374045361408076,0.7925664028283106,0.4491649942613003,
 0.4049620389655276,0.21359370855755544,0.8945805941831615,0.7651512428541407,
 0.7696747197940755,0.8341045207185325,0.11501995887428229,
 0.41084295547473737,0.9804608296794196,0.7723863353490544,0.4560104584123189,
 0.4787913325883607,0.25683908093431074,0.45807534409468564,
 0.4463768026292977,0.3669059011442668,0.36118784457913844,
 0.49981217429606084,0.9208851854073631,0.9563192002515158,0.8902202720597292,
 0.4623372616652164,0.6934502904450823,0.8991212272443796,0.27364027658565804,
 0.6875861257603983,0.7454926303292688,0.35892202276144003,0.5667208265668405,
 0.740490190469343,0.3815816897565486,0.03108852644348792,0.2470153286735456,
 0.4004735688497082,0.24378950218990036,0.2893971585651124,0.3923463743918335,
 0.23375507606834667,0.97629966949359,0.9462928344628637,0.6856908482817026,
 0.4872089820197252,0.8925552794497148,0.6202849612439537,0.4575962002104955,
 0.31851491224494943,0.7125373912725259,0.8237089099168023,0.6267796297775472,
 0.19264186836763908,0.13989648778368158,0.7139988586175012,
 0.8402918020420183,0.39542436492968214,0.21966365047977765,0.991027281054679,
 0.7022016994278744,0.008870989752482461,0.9498969700061476,
 0.4941681036736605,0.17429844018843044,0.16789443746001387,
 0.8273431045035815,0.3288149039118564,0.10433124504730384,0.8201617314002456,
 0.36749003940932157,0.43781587471381733,0.35423479953803527,
 0.23826184095621994,0.835677933062111,0.7769829922939552,0.6834118838869123,
 0.08574266222638072,0.19628475877909057,0.07879041890718685,
 0.2593603867492086,0.7584342744235516,0.2890885870239175,0.07623461055370373,
 0.07899564664701997,1.3130986583953508e-5,0.9720035183538864,
 0.06522372675858534,0.7512965083712277,0.047409917140779356,
 0.5103648125167688,0.9674562575258423,0.8242702613683538,0.5208297737943577,
 0.7817986203553277,0.9182400443130854,0.14949531925395465,0.1523189875321077,
 0.16133304419650418,0.5017381750582958,0.2093041347291219,0.6015364489320519,
 0.056898715807420785,0.13134786944025223,0.8768842085112607,
 0.8766798841187692,0.15518341153882864,0.9831316214674283,
 0.10817473701870006,0.3262783334802928,0.04612355799813417,
 0.7474251523349356,0.896886581862082,0.18897175956135903,0.14377620165793503,
 0.04838065918721579,0.2825179773039497,0.14499961570339692,
 0.9711193486266309,0.9645528968129311,0.523319329312676,0.495676789864555,
 0.9450617545315785,0.7962592113484802,0.335879170794648,0.2315561969346307,
 0.9145861415996952,0.2038346956365682,0.3769509401701754,0.3398520247261827,
 0.8370835126936788,0.37926593342309023,0.7668889431245379,
 0.21988927402921954,0.7636332886062813,0.03147186782663269,
 0.2164725886932768,0.7590561953557394,0.5690530716051911,0.3210718816717649,
 0.5682388706483881,0.01057941005306895,0.8659905733707997,0.2588186896733612,
 0.6328494873251107,0.43798018261060045,0.7394508429643775,
 0.14587507883711293,0.1450421150296073,0.8473410783429776,0.9476472872068943,
 0.5575846835630035,0.6157031737930665,0.6019059283983452,0.09235840012320151,
 0.3712993046026245,0.6976633358763406,0.03769658300730461,
 0.45977761242018944,0.1756751899800113,0.7160899382206968,0.9110004609464906,
 0.4888857734441274,0.2513900224296587,0.7013756566898675,0.18875398550406264,
 0.8798699968452286,0.061215689358154246,0.15714164197694336,
 0.9480624790292325,0.17860041712163643,0.41476942588430865,
 0.1747790465368153,0.07471252430598463,0.05227468183358108,
 0.6319393467991955,0.4511844513341501,0.17674963266828858,
 0.34948866453622407,0.9930053927139818,0.20304979589261918,
 0.4215679304745573,0.3888453919071265,0.6421849346579703,0.9146953464177663,
 0.7400615055237146,0.15161029665171244,0.6042381139630268,0.5882275388354733,
 0.3825723060748407,0.535546623332344,0.2042858094716844,0.0603433836838595,
 0.46095157444683665,0.4573797754290049,0.29768234267461235,
 0.5514835125927661,0.054818285035978964,0.7768310642628683,
 0.05339075792816883,0.4241584343164172,0.04517993227715578,
 0.9547219994425743,0.551621991155701,0.9056983072230566,0.4259942569660483,
 0.4305197632511868,0.7649832719980182,0.6189819466292726,0.3010310648207444,
 0.08042134577092785,0.6024464570106506,0.22438578998445713,
 0.6234880468508781,0.5396364796758149,0.3054418708613269,0.06589050819389464,
 0.5713155838185031,0.9647143723191866,0.18174056250966797,0.9645214288443816,
 0.485108012051914,0.8167554681456353,0.3092180609294911,0.5553411953415431,
 0.4317224321144423,0.06858805375627586,0.3609863636584121,0.7633242870640637,
 0.8998554564532282,0.5175777690363241,0.6974722717997669,0.5691322194181901,
 0.5812111289617288,0.1621866364323995,0.8405658379168319,0.6165200345430089,
 0.13588686677297124,0.8053928412218541,0.3020223779038117,0.7560332832104102,
 0.5356235393453641,0.19481282979945969,0.028575574920604474,
 0.8533060312899179,0.37772537328638245,0.31882831256031297,
 0.5475706003095209,0.591601045161135,0.31743726815598794,0.23388326637344758,
 0.5959991879459584,0.6879536389116644,0.3670091344105968,0.07804238441307187,
 0.4440546557793652,0.05423105950777973,0.7129536215944691,
 0.15182994939610361,0.298653073920114,0.6810729014048054,0.13340085910410482,
 0.703418504538818,0.27550908848558864,0.5995299031015922,0.5365778853784942,
 0.275890534863362,0.7071397786364286,0.6562183189719045,0.47883687293824706,
 0.9678876990599103,0.015518480442471283,0.8348442318803277,
 0.14943659874820903,0.5341309481586941,0.535881503335709,0.7786809771725697,
 0.32999612475450935,0.5349813091734219,0.058810055276789064,
 0.6747448909328599,0.2854179624094352,0.5672629617925884,0.9940259810767132,
 0.3624572587289815,0.8985931154862934,0.9486870987277778,0.8401341183524007,
 0.647037572728504,0.627714061472826,0.6594164555517271,0.5587938955496153,
 0.28280127043728776,0.05430937619641385,0.9186458181493179,
 0.6122537842659693,0.49887220540709876,0.7142543640460061,0.8081013293531629,
 0.0542893716316567,0.22740266959502942,0.04974638101227569,
 0.30701292560083515,0.8823527905641486,0.45297852492262025,
 0.7028382897567965,0.23019003233487,0.38265736059720934,0.579248609320958,
 0.06625031249831137,0.6325587706411967,0.6288061641917557,0.262768634610263,
 0.2634575717803187,0.2698472951027364,0.8913740693686163,0.680502532485356,
 0.9333313708629087,0.23951187434942023,0.7285978607625171,
 0.10465333879911232,0.14986267282517485,0.1799193688877847,
 0.3414902498223231,0.9647094928892546,0.2890928821718115,0.1364688601420876,
 0.939036802115042,0.3491805551500853,0.26306051744279757,0.6429821865671471,
 0.8682891794182909,0.13891081662503257,0.8423491507364622,
 0.23475287101136988,0.1822126968512986,0.6894472594853623,0.8190231593427406,
 0.06268560737308038,0.10715418553198475,0.8980629085137393,
 0.44204022962474165,0.7734200233880522,0.9662581350286751,
 0.26737215229519606,0.5760726070672201,0.31324558209515263,
 0.11017453369628583,0.09129183348231984,0.5742024869935962,
 0.7265104135902158,0.8720741859853228,0.22971292947906874,
 0.49310000576199275,0.8537965796372854,0.4922587420252963,0.5649185082183906,
 0.33294463552919284,0.4879193819171749,0.31270093309062186,
 0.6607235681656962,0.4769135742901105,0.22390155667074263,0.9616343231658901,
 0.3105948908974543,0.5047521298020712,0.4111893023052914,0.43860726063559075,
 0.15924746317566418,0.21667408191084137,0.9187653817600642,
 0.5432623208624043,0.9253122617711992,0.77791155163834,0.7583561103845544,
 0.22994998924090992,0.6861197544838435,0.889509202601738,0.02110636314446468,
 0.511113291374064,0.6582863484469322,0.9771746732920865,0.8599532918105042,
 0.5521077741242988,0.5605265171804615,0.5371107443299434,0.8258641508730232,
 0.2941936021523853,0.39457889167200344,0.3810908330424777,0.4367343666265249,
 0.3353824801251286,0.23136518848328735,0.04721833525031993,
 0.05963336012915821,0.9352819273367521,0.930533822938535,0.9231969519292392,
 0.4949967422288627,0.9202834502066697,0.5959157002940548,0.7778873778456189,
 0.558062885159269,0.25901081007544746,0.0460855863651497,0.975224303057274,
 0.35477230389872083,0.09454320291492013,0.9253617685313419,
 0.5179780886949382,0.5426055509364717,0.21518857965932126,
 0.15573243170852746,0.8660097062060087,0.26553472579858317,
 0.4090022511385667,0.401462545549529,0.5680513733523215,0.9020084267919342,
 0.010404301959566875,0.1449729243201544,0.7930801565850809,
 0.1744468789078748,0.04235203386133479,0.5937362289676067,0.3853017087951901,
 0.5033659047473822,0.9887662572645143,0.9097417888747446,0.22043289104356678,
 0.7369860628053699,0.4267240968702277,0.8012021643732605,0.8929480411533056,
 0.6025066633627362,0.7549844795867584,0.7834813984399858,0.6291362541558908,
 0.0960280380835038,0.42313817683007127,0.9468376489681694,0.7707639855641919,
 0.8435360082534171,0.514285268880811,0.028464912723152325,0.6495174863621331,
 0.32739787162649014,0.44023007882210996,0.7564073569006939,0.629501823705648,
 0.8553701549532368,0.42699618508502224,0.35423042154076834,
 7.705212185957766e-4,0.9258501277584082,0.8880040965657645,
 0.11470155376139524,0.7917751858340252,0.290044451628936,0.7738177952791753,
 0.4683344075241609,0.19987252934792876,0.6049828992414422,0.7118704342234087,
 0.7378523529614901,0.2811064535106945,0.4581946026496776,0.2242932574406049,
 0.9490968389831245,0.6866339278702732,0.75458966597247,0.06893247698655758,
 0.87895994259322,0.8574411027983506,0.31911159164773917,0.9934811556782676,
 0.6101965292800077,0.2091250426823137,0.7763244615744498,
 0.024781867891578435,0.5169480670936523,0.17578357523661714,
 0.9618209055809607,0.6222833143898185,0.4448938939886078,0.1497717747110099,
 0.7055887366772228,0.3103086701414294,0.8774504851437942,0.3193202332768542,
 0.9369364026056768,0.9816229681628647,0.9582572314596487,0.7483380579250594,
 0.2410874996353809,0.5651035735210468,0.6806080188086704,0.11781736678492327,
 0.4886234656555425,0.5380118077156887,0.061573305863069305,
 0.3176190605666952,0.9677217815402236,0.17963860756699424,0.8535555385418733,
 0.8380529129435058,0.687165722610036,0.4205981439691914,0.675985203164417,
 0.3962717555510864,0.33025598281070845,0.9434062361343171,
 0.45629049277486733,0.47626528215520536,0.5512804642984462,
 0.14782951539952127,0.5970413034295066,0.3541465475882991,0.5170148000550661,
 0.2730995452835199,0.8314528728205385,0.4271310859560743,0.1967590255699374,
 0.784041166583184,0.7820983950972676,0.7897964184297034,0.18045792087327195,
 0.9884309912439782,0.3424588463473802,0.40928276975815014,
 0.13241668870031775,0.09550854010979859,0.5020342571914196,
 0.9843714910622724,0.8148336419865321,0.26392047986076794,0.6590270856909792,
 0.7288973578032107,0.8559073712749496,0.3366749405035414,0.22910036601386252,
 0.63103153858111,0.396115296268689,0.9066787883673798,0.9076777865582235,
 0.41376952739321493,0.39662998788402637,0.2294563476986482,
 0.5176358108651387,0.6313194539172563,0.015220810496340498,
 0.8075164767314756,0.03069446746313098,0.8676260303743848,0.7247966259482501,
 0.6221996612611299,0.44361503820655224,0.2877362347973461,0.9494137999127064,
 0.7511024335255573,0.21347047606932712,0.39359814692908,0.31076744429997405,
 0.647634811561578,0.7019153874500282,0.09063147143452,0.7355070999568989,
 0.17453474245598788,0.28629936233912445,0.3772428986776315,
 0.2794644686081804,0.5435353054466441,0.9000749111202522,0.5004898068922345,
 0.07055905661250828,0.38714434810886933,0.06767879020882173,
 0.7456534715445211,0.3505835945925734,0.5138009439279766,0.6240445642492587,
 0.8921623595817051,0.2497285549177004,0.9210411003136241,0.7838471550559547,
 0.9799903331860331,0.3027168923427005,0.15245118847134154,0.5156027961816555,
 0.4412072152461666,0.7669593719221537,0.8705809078967157,0.45976551737308125,
 0.9918351733085166,0.13125947723939257,0.9053935400566011,0.2438682363057203,
 0.3447861170507176,0.7943595229441728,0.9022940075387667,0.6646263139753705,
 0.49463537728145623,0.26889626337132955,0.1289329760978375,
 0.8765504346219952,0.8409222024097798,0.30374785176681884,0.7352648552140405,
 0.4796391426907749,0.9748783331437992,0.007478887723051941,
 0.2385820799060442,0.15017707834323457,0.6481783395353119,0.720995197227093,
 0.438672133084103,0.05594681213527297,0.3259346753805815,0.8061845497205159,
 0.1688274521421541,0.8881809348288051,0.9930254284132175,0.09650028764319785,
 0.16568394620650306,0.8026201539903324,0.19245045494765456,
 0.2259987300690911,0.5871676404731081,0.49593174366518955,0.8370226611763611,
 0.9616406379826671,0.41990466447008545,0.8897473245144639,
 0.27848394978904256,0.10459166423502075,0.6182811419188741,
 0.49124700540406774,0.3871177510568923,0.4931798538389568,0.326540699754588,
 0.34125199096685477,0.042563068276939786,0.4287963921587178,
 0.3823073647473709,0.16826798656587894,0.36353214827189784,
 0.1398500906661312,5.151562862276693e-4,0.8806875403366681,
 0.3261297060908237,0.9706370559793435,0.5849458242403841,0.5088523199866641,
 0.8995560331273579,0.2702405433609273,0.9943446425048713,0.37395470423769794,
 0.9207329250173659,0.5177519240551282,0.9051626187963868,0.7280849699465579,
 0.04363378998560774,0.8013344021130231,0.5819189816777712,0.6230993793876947,
 0.5753170960303804,0.010583659230544384,0.092004144787871,
 0.014041578003104016,0.6785830188002834,0.5032190997205589,
 0.7025607928297646,0.27534425422119013,0.6732619401308809,0.7605232428831015,
 0.42252082494754273,0.42254729475560115,0.6841254712080705,
 0.8810148389847969,0.4770511873052524,0.7381566595654601,0.4818128904704134,
 0.44539971918432764,0.6486681950347075,0.8359433553009347,
 0.18624880356431928,0.2883896892094061,0.7313735489093217,0.1830755809925788,
 0.5422839323343619,0.6741099990291182,0.5326155270647412,0.150857606260402,
 0.695644677198497,0.4342948534140777,0.6788312320576361,0.3798582798082273,
 0.02728669390049765,0.6262753785779357,0.6011618582982229,0.4727815054478013,
 0.11183907903502821,0.986541474204378,0.8146724413129138,0.9413749390784055,
 0.758619439898934,0.7256983715464528,0.227521649739862,0.5823247974838012,
 0.32761492286758587,0.7340144302831916,0.16222600719570412,
 0.9070144708203969,0.6157196723276606,0.5332799027412546,0.5508407633896839,
 0.5702068939687108,0.3196743610994932,0.1731334084685554,0.07734202875940965,
 0.0763117577665472,0.5648696397720725,0.15189637523446775,0.3466505638992148,
 0.3959596572068057,0.44147070686321854,0.2515744197690174,0.1970908311468177,
 0.3232230334985483,0.8556886101940276,0.8261422701739222,0.14738669228539314,
 0.9078537841814732,0.0886715600441983,0.6624983164982071,0.5419463442281406,
 0.9375386091991781,0.5611873776647036,0.5228714777981489,0.5954532177296279,
 0.6443695451141083,0.08188042753034885,0.8424085279056913,0.631655459095315,
 0.46954366777617607,0.7271983951483034,0.621962498972307,0.4117701983080376,
 0.15748365970591383,0.06563987654603287,0.2816486180370208,
 0.27261096405824814,0.6727753580528273,0.4089842552094488,0.6464536285089155,
 0.8521694681791699,0.5751842880469269,0.251777994338128,0.9659860946655938,
 0.4037362675344315,0.06838642355410784,0.06087481168994935,
 0.8236017632883709,0.5930677421799622,0.5480916678423307,0.4010717395323886,
 0.8903711588115103,0.8104794381903169,0.5516591105350814,0.343429789667394,
 0.30555423841971474,0.9096003351206781,0.09055118468877865,
 0.6963325506907478,0.7384750650747353,0.6617422132079653,0.296908164076663,
 0.2731390982346277,0.9722280709522195,0.8693040233536019,0.1815776022245903,
 0.8744795990196513,0.6074178035784668,0.017814186551938227,
 0.6075231558792903,0.5729264646780655,0.7986855653981813,0.6491058097626214,
 0.5355913957691206,0.6382394370305577,0.44501957882035936,0.2204205589992485,
 0.07761423780244237,0.2605352962537427,0.9071461745410997,0.1907300553648985,
 0.20567161752100893,0.8581244188653437,0.346046310119302,0.654636956531208,
 0.4108811535816941,0.953827058982434,0.36262794819253275,0.6091198383615601,
 0.3298115754961646,0.4543767243964866,0.046717712022471436,
 0.019707563000799233,0.8900860214352116,0.031064479235542675,
 0.4989382774766997,0.288898221062896,0.24030674707213684,0.7624275675860019,
 0.8201006029413076,0.8918878098262663,0.1912781315896791,0.19336072720432107,
 0.23072324361161223,0.5286732397042474,0.16946658084104027,0.282293327510749,
 0.7400819707679396,0.3502568458624715,0.5972771646365397,0.535780902573619,
 0.005800286802366217,0.43210535947871775,0.3977643131126515,
 0.2862481280374509,0.39866683633935185,0.9923661663110989,0.6164489849800345,
 0.08516695143263886,0.8694508202708968,0.9162696640609377,
 0.24273337159704345,0.4254042110282691,0.7987252352351207,0.7552983687889745,
 0.039855745570645085,0.6737327596647341,0.712644942963736,0.9900877071356705,
 0.0680680589037016,0.9979717110423177,0.38809130772900324,0.8053411625542295,
 0.119148308107772,0.6709643378094496,0.4625690595286923,0.1699208714533943,
 0.8802333784771756,0.3205768553414853,0.5337732577613251,0.8354949224490357,
 0.1875961272629365,0.620113688335261,0.028325683091062537,0.6724449006758698,
 0.2852476272609512,0.9762632379168336,0.010957208438107213,
 0.9850281622027597,0.4507591443575264,0.09755299864053102,0.3230046267711213,
 0.2518150427270962,0.7911847574756008,0.4708688070990148,0.43187581918849893,
 0.8570016860454144,0.008932036861966197,0.567696867773983,0.8377744286460855,
 0.20523765781074765,0.4536876236196278,0.4397502692898342,0.9675902838463702,
 0.9797356177658973,0.6113385185247321,0.550098761165191,0.20158199370418983,
 0.7462901484396257,0.027145252203195303,0.10095117839036738,
 0.6702850230616482,0.6224201173856163,0.9683871666039221,0.33816669833049373,
 0.19515656971994721,0.48356319912569923,0.01837990159684999,
 0.18212136114049438,0.3758705825585511,0.2969414250179039,0.4272231905394195,
 0.2637354567406184,0.314019028781092,0.12820677655378332,0.8085312900020611,
 0.8084998470317739,0.15422605366996034,0.006084414589991427,
 0.5502740136218066,0.9939612348289981,0.7794427511948561,0.4809704865624198,
 0.4285285879261005,0.5351370084740393,0.879302118014438,0.9274185443391071,
 0.4730368128681328,0.5323144478237429,0.13947275330205133,0.5873544353104625,
 0.16811299122653933,0.007306297129200834,0.9139694760009756,
 0.6601015335439163,0.5681160044968525,0.8831396878423199,0.593735005730931,
 0.5116128012447229,0.5066668176204008,0.1621277432482221,0.7349371339877986,
 0.3289774900460285,0.6569733623638627,0.9964146248359601,0.6632906627690203,
 0.7698386842421384,0.6562667668832283,0.43498187674578914,0.7298039297753957,
 0.5623275469071433,0.09628291403159195,0.8756274205459256,0.4132924304080817,
 0.9584876204398883,0.06250191937644545,0.00461674907702736,
 0.4310228347877929,0.8388575190219487,0.016671545440778313,
 0.9047449399261472,0.5334214644407733,0.9950718800925626,0.41407599826477826,
 0.46694030687827504,0.2560143106517284,0.7477295537797489,0.9118822985462837,
 0.7016965410923077,0.5257378481976923,0.07877873501292776,0.9286056397375185,
 0.5732908816384049,0.29368967959559633,0.5048829169459739,0.7053639465506009,
 0.4526470572232837,0.28585694885763613,0.2908175472966732,0.9110479587011899,
 0.5103838067629107,0.22919343037961692,0.04271157636426626,
 0.7500722207919122,0.8558335438501244,0.7525243241377199,0.37195155998587137,
 0.15922660424838653,0.6174441777639312,0.3894595281438783,0.574994409794139,
 0.3604199843220436,0.31199721568036143,0.7979936844226379,0.9247597986507259,
 0.7123248012232377,0.7444914501904641,0.9490071878937354,0.0440842331499639,
 0.5042208489197205,0.8573485807687563,0.2587510574553965,0.7733786224118376,
 0.49410001541933557,0.7284751730045066,0.23785081668581143,
 0.9201541015047106,0.6024263460051529,0.3201769783572017,0.7888668867426173,
 0.5748637098891766,0.44105886198418986,0.32855211894465763,
 0.022300737871670306,0.11662566969771093,0.8218289961829202,
 0.6121007586456655,0.685958954911524,0.1680310613275996,0.13693877517308528,
 0.1879813275078922,0.0836066553322895,0.1913839545773296,0.5470723530737244,
 0.3686519290579774,0.1892286187644001,0.7801520470102273,0.0974069854935462,
 0.5765603319415253,0.7826134845550087,0.4109741887073727,0.8862762943460084,
 0.31857477870974404,0.9858755365283468,0.22896522356879967,
 0.5317647420031509,0.7526701367424179,0.642528943759603,0.5442227535108899,
 0.5891367806104946,0.03179776296108938,0.4056965645933619,0.4441743656886623,
 0.8440247217987908,0.5386727687159086,0.6058474287755196,
 0.021419277250288915,0.43691338217150844,0.30366018158148833,
 0.4882838661463639,0.592119423446866,0.8909980340709072,0.9051417767256725,
 0.8649397670076776,0.36223611207007644,0.28262514112164405,
 0.7813628789529503,0.30764117556917736,0.9895562906219401,0.3255040459643709,
 0.751717558671367,0.09315239713896073,0.38976971423252227,
 0.46961660406511796,0.02311658478779055,0.6462857015617877,
 0.9408960741022275,0.46535262920940146,0.4956953401023315,
 0.15711482518422815,0.9690816930077726,0.8834145134445957,0.6601102260292944,
 0.6237062691373605,0.5216098798022988,0.17034068519939383,
 0.13858079473845164,0.17533630207097017,0.672132237918418,0.2410389409932423,
 0.79289202808059,0.9538886560171947,0.5567942722131798,0.600114647162281,
 0.41760207324992826,0.13670080971477916,0.12907502576574614,
 0.49390920645184344,0.2088571948457042,0.1052420399641435,
 0.07846670436840752,0.9961224568742632,0.28262983890553384,
 0.2785331823710492,0.2899261373916865,0.5932876662281561,0.5289255356958115,
 0.4558798623023418,0.6925640778116463,0.998873401162759,0.2950476588392581,
 0.09670902109701218,0.0492448283621012,0.4734256552812772,0.5106391069513754,
 0.7532828089853871,0.36057289207313403,0.7100107773388398,0.6588425018785633,
 0.20882561783879694,0.020966124087154814,0.3810923711374915,
 0.27326870505482626,0.36304202611193614,0.22456981056434178,
 0.5722453173534278,0.030831312156540003,0.2803554597283904,
 0.29759447329179123,0.0882628591641913,0.8725619332502567,0.4486964056458511,
 0.16378911097486193,0.7339749830239359,0.9233435904655725,0.3250690699226786,
 0.07326937343663908,0.9907044693476844,2.1218462229632973e-4,
 0.9418615890084063,0.09167948647540425,0.5088998059245186,
 0.36786775241093606,0.9310858249099808,0.5771934059287825,0.5320346858465912,
 0.9965501928161985,0.727513146709638,0.11378435955277943,0.07294003067175814,
 0.44547737608256754,0.5601197196664682,0.7870424609455824,0.8232996398581242,
 0.2999615424137586,0.19714583973056143,0.6946082261695601,
 0.24830127818744885,0.9111436401231389,0.34136105046820486,
 0.9405743885347634,0.6729143611547235,0.736704569379548,0.18413221066746566,
 0.15661950037288241,0.32954484222125846,0.5987374928260351,
 0.2197979746404296,0.24645314770543314,0.7305235724589404,0.7807010623840964,
 0.0937724956763557,0.19564412784596008,0.04458105246995281,
 0.8786896870561529,0.9247791686022767,0.42743155682434153,
 0.48217904042284887,0.9757761696201508,0.6581090686257804,0.7407628799663256,
 0.8305682350298663,0.402604981689787,0.49091226110990593,0.3060141942822846,
 0.9727808548042152,0.7729755768972106,0.3069826815710197,0.1960600064843726,
 0.8457471895707946,0.290105530560522,0.9017839140463794,0.8354529953574701,
 0.8144644580998758,0.09431625304355151,0.1770498496608508,
 0.19512423092143527,0.07703523065337681,0.39123983708888777,
 0.20523537617943877,0.619698796412241,0.033185305434358714,
 0.7110210414870923,0.608644949977068,0.2640705149985768,0.04172658055812217,
 0.4532556119371064,0.7224155801689087,0.8509537668361676,0.9733562342969628,
 0.983832044814809,0.6200140240859997,0.8431667178445463,0.8238755982521654,
 0.031142851210478883,0.4196471346887152,0.8827309719296834,
 0.9410286217974948,0.738223954164537,0.7339887193023191,0.7004811556428441,
 0.13449771165993618,0.9713372857974562,0.2814736629370226,
 0.46438822507535593,0.9851726198841193,0.49527671595832445,
 0.38504266177134716,0.4711773442573204,0.15015672651185674,
 0.47949648921279453,0.04254970305563033,0.4767425796176342,
 0.010736631306824784,0.4421374641472944,0.6429577749154596,
 0.9373431063499026,0.2523740216746161,0.30130590317229344,
 0.07150646943152472,0.5704057688005354,0.014855781244868083,
 0.7654712436954508,0.05514107915168598,0.7596783038427255,0.7354344794334051,
 0.16064696890347718,0.5138069041893132,0.1913117238516835,0.264892040273463,
 0.2773315944835463,0.20864930771134083,0.7214960686643195,0.4078391056679409,
 0.7966314029315877,0.09266990550510001,0.15386306427440422,
 0.6156648885403375,0.7478685957308867,0.5106940690472922,0.7748890148162615,
 0.1011794082098445,0.136854317055212,0.9545777665387218,0.4816657753370165,
 0.5187325924577971,0.9352236827665763,0.29796288914308333,0.5202162478453204,
 0.7858841958872032,0.02246890443872518,0.7378323745376053,0.8972350801250923,
 0.6882676228975533,0.6031675081790239,0.267076129064458,0.6612691912395225,
 0.17329612113066328,0.8225087052792359,0.8296226212596556,0.8369805725804804,
 0.3725026882448922,0.5060201115545172,0.9365361027469691,0.6107962525138984,
 0.1397758407722134,0.5437299912008158,0.9727028258937557,0.9999136338441329,
 0.6911936257678029,0.30101314203907426,0.921108231246232,0.25138890654738444,
 0.17558075866687517,0.5893498848217275,0.15457562205222564,
 0.24024825053206567,0.9599261793329477,0.6778693569570342,
 0.07629003471087281,0.035664552841946584,0.8071704860469452,
 0.4147387873005718,0.2436261219022795,0.04783593468313718,
 0.45146396605459393,0.04350152546468977,0.9701518001863747,
 0.02985752718478618,0.10810360827571885,0.341635809553989,
 0.07623751567466763,0.8045577805521444,0.9332178275074166,
 0.04700573674867592,0.9513268682872055,0.898907688445103,0.8429068565467788,
 0.6144087504972832,0.4639574106923151,0.41093833627579834,0.9133894612245779,
 0.534983434663126,0.6948099138634298,0.9787395563229951,0.9691798761788293,
 0.7743964579543192,0.3420576307962989,0.14723027792745413,0.7558058799380182,
 0.4771855280914664,0.12722724724506185,0.5042250908779331,0.6666107460433645,
 0.09053906369389952,0.6274867843874532,0.32379399164805267,
 0.3793754752218964,0.4308629875009433,0.024224437089747086,
 0.05754986055757738,0.35894191455237245,0.697342638871038,0.9832628190376302,
 0.9376291142664341,0.8171666669265518,0.9367891810481686,0.5308498253311459,
 0.9420928877146235,0.570323036602792,0.030232983173364647,
 0.09466254906702565,0.6577486948140007,0.3928819100870016,0.7310175673113887,
 0.4639554135505285,0.1417202422382834,0.35064346339832186,0.3154960637678116,
 0.9098680477005735,0.3036000538472957,0.05816676663006293,0.9286576096750011,
 0.7406418159504035,0.41292826474203037,0.9991756155207097,0.2320322462247859,
 0.7231355207246662,0.5127836680021305,0.23063590336741624,0.5248436496440411,
 0.32634122492725237,0.30434394012501276,0.06330330262511952,
 0.10510725349972694,0.23765667115981248,0.7107043144978327,
 0.3448913392052164,0.2832677403200128,0.649515497409429,0.9132076677587524,
 0.6955130474450358,0.29603292750773613,0.05030215792392356,
 0.14444090987647373,0.6351946314309879,0.9955315175328049,0.6736921706016801,
 0.6486762460981841,0.8420497349652366,0.4272838970716536,0.20297378287782109,
 0.7853650493865095,0.7489369839702998,0.8965332831652154,0.42456308282136035,
 0.08927116428832771,0.2193464058747676,0.884122169180616,0.6273357172715732,
 0.5997561871704701,0.3760570818384825,0.7441600467762226,0.9685837231129999,
 0.22321218983668956,0.2606756602193492,0.8507805255422402,0.5805729245194435,
 0.8913954433095169,0.4965629329592809,0.12075977750173716,
 0.07591890304833238,0.04844093506201541,0.2721138719820524,
 0.1615766793098019,0.10409872293887401,0.14471263268947643,
 0.1585746952534549,0.8060692710779499,0.412785577327202,0.639202169117257,
 0.16502180413194056,0.8531235883894506,0.055292328315372674,
 0.959272342122718,0.06698823816224919,0.22774477517501868,
 0.41089607406762285,0.9791287093051503,0.8362548269531713,0.9822747870590389,
 0.09794935331510812,0.32563510002212825,0.36610519123762164,
 0.18807881881999355,0.20127699751925732,0.32449387539335894,
 0.8065335937962953,0.12869794927760358,0.7584958523142392,0.7853839472843817,
 0.14459311859412538,0.11926234146287673,0.32314762029652744,
 0.7469199903656185,0.24551017997487945,0.06385568894252724,
 0.6182506438101564,0.8932091353275865,0.6171861688109648,0.8927091104033272,
 0.2378882114819203,0.6461078552977572,0.056478660081016874,0.734623975398177,
 0.0055769041328429525,0.1282322129795207,0.9489058448404226,
 0.587934206434422,0.8580586671828752,0.6182184516381278,0.17732520525049555,
 0.7123295733853537,0.34164645901956325,0.04232255079258085,0.601192347305215,
 0.38395047379652714,0.6927998495892386,0.07056689307468955,
 0.4822134766364905,0.022814865067383505,0.6082679510805392,
 0.9992407457889869,0.8584021786672034,0.011214911945168882,
 0.6789975132446595,0.8090403824110551,0.18505091078321156,0.6449396073050837,
 0.8180170894462554,0.9703239935052526,0.6534530281671012,0.6528962184940651,
 0.7826965997266567,0.2200247331348255,0.011092483351623605,
 0.11802333779448482,0.49402813287426994,0.4043691400930387,
 0.6603881012540325,0.7085310533940836,0.08525640777470844,
 0.35627123570065033,0.750604009001377,0.6823656236828256,0.5746558643678592,
 0.6279978424500949,0.2946081141090915,0.46740281338709044,0.5768722618599236,
 0.3783172480643835,0.6805147365165575,0.031602972230594384,
 0.42817406483667486,0.13785095041961848,0.19869769353580724,
 0.6439363913574406,0.9039270124981089,0.8077236522674431,0.9307787268574658,
 0.5884044887971381,0.39121701151583377,0.3745371730418684,
 0.24690080423459926,0.17650411549332645,0.7333087428830144,
 0.7643928860353925,0.9256563163139676,0.9899024302826295,0.8070961762761312,
 0.26491469611892204,0.7541299576795322,0.24502045149299034,
 0.7783258796164695,0.7645192033896936,0.8463882117374517,0.25216245587224473,
 0.8442397566478206,0.48772235140100795,0.20038554068594505,
 0.31679349259764744,0.6297076887551175,0.9294749684222496,
 0.17980270629279682,0.8683890950517242,0.5368979961971068,0.0313404747618643,
 0.37275439441838376,0.32616727782055654,0.8608369851378773,
 0.8270054950044886,0.2446681622850565,0.07969241116596315,0.5449447344205988,
 0.7896600906329917,0.24047868433811814,0.25744980028666564,
 0.41690048411485237,0.45596088008931623,0.11637367321642922,
 0.5748914132489857,0.11822813552496969,0.04494863462328458,
 0.7862133669711793,0.10049311507793857,0.5067916555237645,0.3880967163546357,
 0.13879047780075038,0.2772960769580337,0.15799015350854395,
 0.4017309893393213,0.3773215727865764,0.9846371758743226,0.3202398446927156,
 0.3182304340394376,0.15955652721698688,0.37933352501355744,
 0.1795885844558509,0.0647856989948481,0.6312635531482866,0.9438189284305845,
 0.6905049544838695,0.20094261340127528,0.2409810329049451,0.983490423404589,
 0.4313617474555599,0.3279483547551404,0.7760801017321278,0.2006573275517004,
 0.2341557024314448,0.9785929208396864,0.30513750456523103,0.7754507439893339,
 0.4541657270290371,0.10613270731477642,0.8305926609094993,
 0.016762844496976625,0.903890400571015,0.9433999767198757,0.8989038692936341,
 0.03721981935119967,0.02619881528637913,0.2564118683288231,0.515592093463769,
 0.6301465125946375,0.8579701182049508,0.7188093754370714,0.5980913163499983,
 0.8927915570023925,0.5370828861510205,0.5767591785917295,0.14796344429405217,
 0.4697312848892471,0.041258137359656444,0.10496113828823717,
 0.01451869366842884,0.8392961593376382,0.2123788251346509,0.7380880257787238,
 0.03870957076840498,0.9106730546371644,0.16688496536569386,
 0.5079567898702873,0.7723440206975335,0.10109403835490038,
 0.09837121778594349,0.9764634035177522,0.9814258334155266,
 0.09478067907658361,0.15171732068071686,0.1894830668408879,
 0.00789848244663005,0.6566715473122717,0.07940087952287811,
 0.13250377536943447,0.3629254384537003,0.4202110679424702,
 0.41626919087953707,0.16335586813904712,0.3616333470403589,
 0.18284028618970605,0.4726647593817057,0.8696944825185611,
 0.42957412352453705,0.4684272421201061,0.996111553539865,0.999145507225651,
 0.6939690068698987,0.2900193840935935,0.4514269065384545,0.8782224139695851,
 0.8034785258407878,0.4567405561054797,0.7060351256485226,0.1993737600729114,
 0.9491005420222789,0.8540600048080035,0.9783393486981229,0.40923815423410803,
 0.9890299981639205,0.27231742824230265,0.9030751987398768,0.6077994979423126,
 0.23760295911537788,0.04336218229065858,0.5311987094509539,
 0.7817226373571984,0.6528719810414508,0.23886495822356701,0.257071789278449,
 0.038442184219454,0.5370079243336913,0.3332988873059898,0.7030475363390504,
 0.16988770337687586,0.6617405642896447,0.5351392256602128,0.6946699511046456,
 0.9486286586711287,0.13839596235231844,0.9455750634283331,0.7250096710278527,
 0.8489577712400618,0.43473344300703476,0.7541690117082709,0.7210650352078173,
 0.8179714330485865,0.9630822102345321,0.7708679700793828,0.14603531742705822,
 0.5518184485527639,0.7620186870650958,0.23186034646864384,0.49070285550418,
 0.05773199769528836,0.7981214443492436,0.9549010489980062,0.581197244029221,
 0.6663300830827219,0.48854138921637147,0.22838578742816606,
 0.23645604718754754,0.468489434880749,0.9560057027823664,0.7099696217433565,
 0.1684066960610936,0.6549028975786394,0.7439885723566529,0.04076600834722366,
 0.0753311933936397,0.47271420651853757,0.34893071390360764,
 0.7200311958223647,0.5572568392330003,0.30198101852598724,0.7973879546386545,
 0.595168628022993,0.4046331468574287,0.36030589682978875,0.6949507617732014,
 0.37962071413874277,0.6801572304668397,0.17147523918913554,
 0.14781242399346373,0.6532216584600835,0.6467835457029262,0.4939108817093706,
 0.7477539940911566,0.7448187909987231,0.23908308588933347,0.5644738183662729,
 0.7765128251817323,0.7017977615053334,0.6232306788628946,0.4316941077141636,
 0.2407976534203753,0.40737072643610606,0.6491367220380913,0.8768342752980209,
 0.31508454227445226,0.9929953453972404,0.11462983942877325,
 0.10110236145487383,0.7195242072771493,0.03190141814351477,
 0.7805000753718607,0.36024421292491293,0.7528760734298933,0.2699973947191615,
 0.22106861955714752,0.3566462253285243,0.13202001641861927,
 0.33064107446281576,0.26688091082559096,0.30111315163802244,
 0.6294039512384391,0.9125541452908116,0.821510952824753,0.2982465705962002,
 0.6638576854195853,0.4838176738423423,0.5971797016862608,0.8849021890697358,
 0.6318596660761041,0.08185862929218257,0.46872852007314847,
 0.9531153750403841,0.8915619939185755,0.7560013890836355,0.7235275751627246,
 0.973600572435817,0.7024901779143856,0.23393248169108694,0.919012322566704,
 0.9951658683071257,0.8071610964359136,0.42613792462705735,
 0.17017795066742536,0.006813282885604122,0.5185916047174888,
 0.36446436627754764,0.5578902547914842,0.4138186692146195,
 0.49073694009145064,0.7576098826172682,0.6296087455487624,0.4010486681126928,
 0.34658990316767235,0.5038784408554947,0.8873847376852401,0.8120767759462022,
 0.7946924750183412,0.7913227306951834,0.9076671583725728,0.33680332485330355,
 0.17213373376572916,0.01779515825984379,0.4495352432467723,0.79184442858974,
 0.4952808886280431,0.33629732399704504,0.7911149839722446,
 0.16246513876448432,0.6122111093695528,0.7708234902047115,0.367581458397783,
 0.5743066305797142,0.721590644056343,0.47952147960861435,0.994140693466644,
 0.42341559928128025,0.3067936258603574,0.519777977261594,0.2963885946606476,
 0.38881213878812004,0.06643661611768159,0.48536387185739027,
 0.7341053848519307,0.7273302427196472,0.4678589427422556,0.7742780092164857,
 0.8175655787833936,0.09341183316596569,0.2659131103679354,0.6395983604253729,
 0.07402454696093119,0.07158817099631931,0.2867263148882083,
 0.0981801244024263,0.4531298691300104,0.22464483186478845,0.5055554673693774,
 0.4746044729972715,0.1052231595035108,0.12733818338802205,0.8443660184997873,
 0.8261051332298249,0.8183608250179164,0.563818986773951,0.4070263403959269,
 0.44178439677255554,0.4235170647184787,0.559772411525168,0.884166712585566,
 0.07033383552041561,0.46356586992445303,0.3455878619146032,
 0.9969288831905596,0.6871743648155716,0.12042082686595212,0.8929076671477464,
 0.8624915712639526,0.7124419093417677,0.32932414132560683,0.7859056827107487,
 0.3743263889584939,0.692741622451966,0.35718655212271866,0.5456816524275078,
 0.4774938777912693,0.6760235065894232,0.43566781010454625,
 0.11407044787300724,0.3570416353007222,0.5746143379665517,0.8660003257093228,
 0.3026861468064028,0.29734052384353316,0.12286257616157936,
 0.3785953710377463,0.7968589549650703,0.4389362001143441,0.390275978157139,
 0.19597224175135808,0.8248313454492026,0.8888779841093749,0.341196057427659,
 0.6972743060988623,0.47566614595281975,0.5976689134084786,0.5385234815066281,
 0.1714185892140988,0.7906540876482718,0.9805943153676389,0.6857718965575033,
 0.027512085728698676,0.5259817644775048,0.624339247880964,0.2174961428522445,
 0.7952100024065704,0.1579442322227239,0.32221079186004387,
 0.22316347857288343,0.6242646574664154,0.8462744805733601,
 0.20008675138705156,0.044993317368759,0.03624917979933662,0.6537339862593651,
 0.03551088028181126,0.651528484844094,0.9883676130242964,0.18659939519127267,
 0.34403135751179514,0.8801882221010933,0.3873251669455914,0.9640336659133624,
 0.6524140497446711,0.5223368004701865,0.864763466923557,0.4423334285131203,
 0.245497042671027,0.9082610258406125,0.04027387567672647,0.4133043496920854,
 0.22922463551181183,0.3583349340230777,0.23682118382797368,
 0.3415338538168615,0.9478921653879404,0.64368788640543,0.7760699407599796,
 0.5736025028084625,0.4288392676840812,0.19939373225799795,0.397200801675633,
 0.4862237930218345,0.9867118984913952,0.852041867806054,0.2637059209372824,
 0.033988774676693034,0.9352123759271791,0.7565594993045,0.2884137729862245,
 0.011217413881348559,0.9612032251065596,0.2356961540599678,
 0.08556825959573067,0.6878840122393138,0.80641711316965,0.11144569112862313,
 0.6546964560655988,0.6972714427432714,0.6109301188177768,0.11548115798168679,
 0.39984080953427215,0.6439613176086473,0.9623553434860587,
 0.36379442862180567,0.5047266604047345,0.4318191383236929,0.1187330975924431,
 0.49791162297148017,0.6431880281799931,0.43529580759409525,
 0.7984752331313059,0.09073297804814628,0.5006537776538702,0.880433086091047,
 0.6670445688600157,0.4119028734723298,0.04113004582150992,0.8949368431363133,
 0.2625739438198632,0.3575822372672126,0.7807115191266659,0.49155354921447825,
 0.3711615924941327,0.9071992930704139,0.2482303585240353,0.9662395420843723,
 0.6577509403002624,0.5999254951234728,0.5837283791267168,0.9649132878958658,
 0.39726467492238715,0.526152125037776,0.9895237606626409,0.5112465825625814,
 0.21540158770059015,0.3336735940599608,0.8125966622591939,
 0.28162267218807213,0.7818696769453822,0.19057821905668193,
 0.5095653325693754,0.807977533951798,0.2436646691355554,0.46356416341458606,
 0.364118213984829,0.04243388258397751,0.3644910911905086,0.17927243699665163,
 0.45850334666258763,0.8936250941708217,0.872893183517712,0.20041616247694016,
 0.11954312286146451,0.7280565447945421,0.5779532509345837,0.5458711017882344,
 0.3116588478039207,0.10571968266404086,0.926863696748039,0.8137714980636963,
 0.13474992232521066,0.3807368299694752,0.5576165836301495,
 0.00942959597083981,0.14064785573324712,0.7199814826435071,
 0.21422037478897338,0.756926117451679,0.2220968893878006,0.6769019079133413,
 0.11882301794652617,0.8125060341921312,0.9703270410240786,
 0.31099898739811094,0.04193969162240441,0.672996571182275,
 0.22540593927472197,0.7031778978065293,0.17781778804152104,
 0.1234055173502282,0.33328771333263785,0.14177167947575042,
 0.15672242163399186,0.7943004485290827,0.6648555334981808,0.7922232798833662,
 0.31341537111182105,0.06131813216974202,0.9821353021822172,
 0.23230729249022242,0.7417175777382276,0.8166473577005652,0.8757637261561876,
 0.07217797004111293,0.558172346038833,0.4133242213048489,0.00232492350642044,
 0.6631379177048985,0.2621735400197085,0.38213753553065666,0.8746543947644034,
 0.5043688254769184,0.9018276318593889,0.45001020583548124,0.7823194238369499,
 0.48521417907481745,0.2650767835518745,0.4302146695928619,
 0.08764711022952776,0.7023757188151768,0.5022706750389436,
 0.22531431862759677,0.25959255083874333,0.27693205085532946,
 0.7695016387683625,0.5562667404536278,0.5578323717362134,0.41292643713634447,
 0.11705529203753695,0.07593362528810355,0.21687183797868093,
 0.3995463616531547,0.44303642057423653,0.44781547271119837,
 0.9640148653230435,0.8036672407747508,0.5194155162459639,0.7759878154707549,
 0.8257951998899008,0.810898581240122,0.3590235957908371,0.38451833007543024,
 0.5514171107699051,0.984363082955187,0.9621067767659857,0.9105641606333935,
 0.916287567005111,0.35151418343715857,0.40253233082678885,
 0.08917723480478312,0.6102290046020226,0.16805053318098584,
 0.9819620344353692,0.1640464420963732,0.0028886015881821603,
 0.8702176502430083,0.21730013943664406,0.5210565602601339,0.8436576375315723,
 0.9927480209910802,0.13191815439341292,0.47617453843041757,
 0.9842634907999263,0.74683102589331,0.8122973860484308,0.40056753670784173,
 0.636518052182379,0.4505988689447028,0.13884369739456548,0.13746771510239209,
 0.042174750833686825,0.8362001961464239,0.13009558880052463,
 0.4756619087304276,0.4776111181622882,0.33593179237033066,
 0.11102600083143666,0.7324316829106837,0.8391733304137915,0.8096414050676839,
 0.3565180733817892,0.2342080365853667,0.3695775665177644,0.8830360897273369,
 0.04459812548619135,0.8503702997529183,0.21054618302400596,
 0.25203666103047473,0.14118760686220655,0.4576842301587053,
 0.17364251725593582,0.233756249357305,0.7279663335707562,
 0.030955610037009196,0.7962567189556693,0.06332499336600139,
 0.20031793456455338,0.08737174596411956,0.265549457205678,0.9600780687661622,
 0.026029822067948682,0.1508823627977326,0.20934040711246438,
 0.09348890795873688,0.09157605281899273,0.037608180398081004,
 0.23322783989088358,0.8488285353647091,0.3494015427947412,0.6436947008963154,
 0.2045390673625701,0.7632502897931442,0.9890661056581964,0.14106740324760736,
 0.4875092603434875,0.4057963488663827,0.2906451237563218,0.7108232731411553,
 0.28350643151676014,0.9861655877244562,0.013456792064862944,
 0.8007552716783977,0.548946702595634,0.07656096104550325,0.33080359743628807,
 0.8705626758359069,0.1700236786950995,0.40393520171978425,0.6593626423657194,
 0.7538751853341256,0.5731077143338861,0.026498146860905236,
 0.7820544244185366,0.30172821745348555,0.09847479000924053,
 0.25824873904485046,0.8325959319387708,0.17964439497910334,
 0.7602021993803991,0.27037076399493076,0.41336668776333163,
 0.9046918424524815,0.5055903414112523,0.3102223285723269,0.3675271135038295,
 0.9011442325269418,0.498386678770508,0.5563294330774416,0.7193209550439784,
 0.030722402232978285,0.3937379751635456,0.61259537805514,0.9915446118223602,
 0.5089519130679063,0.004686726397872376,0.43402081712868623,
 0.6264288278628264,0.2660148410217541,0.6129260464015982,0.34716252863608676,
 0.4234984522000883,0.6816467149799205,0.9985487516777122,0.9901090912458731,
 0.8068804424005527,0.3923974045519809,0.2546320311982644,0.6422213069564014,
 0.007105879954338201,0.627762815581532,0.08556042987385737,
 0.9280760175630814,0.06289164312074536,0.7432945351887452,
 0.43284750239681435,0.2608713186473244,0.980638154833926,0.2778751864172625,
 0.49076993664516,0.7845998488203678,0.0021698273566889137,
 0.17737461320926406,0.9254545194985161,0.31136764513724124,
 0.44523199187173346,0.11562052617204865,0.4495655324700991,
 0.9737868287273427,0.3648410595099607,0.1317229361201777,0.64645681871572,
 0.22176096698422743,0.7220262571741056,0.5159433220055821,0.7249199140242507,
 0.5827010271936237,0.44320040254791004,0.9295532709708035,0.5330824566232526,
 0.5661616549292133,0.23685277444075192,0.990827311038541,0.05152850747433446,
 0.09286051305299581,0.8622198283305408,0.10460004129404754,
 0.3049597341281135,0.75214583261262,0.36616029478948375,0.0991157050179633,
 0.4265780381240768,0.3370963031200547,0.9759590838038311,0.05106919037101154,
 0.6089398148018494,0.9398090283654043,0.8417446964493609,0.9308744996420844,
 0.7547215084323076,0.5550442501902925,0.4000041563595069,0.3956330437005471,
 0.7262830644698468,0.17970634246457062,0.819574456995965,0.9596786022853704,
 0.36460428564658764,0.4074270314490739,0.030004721974523862,
 0.9195859685389723,0.5745766758749384,0.9496435342972539,0.03705890926258748,
 0.7852791082228703,0.0272027658449685,0.1077197695143235,0.8785668749364955,
 0.7739345347418719,0.5129306256462377,0.8107412653933398,0.9153853530616002,
 0.23618506217476065,0.3677949014574382,0.7510227300301684,0.2980539725921171,
 0.15474026338360036,0.340062053824561,0.10639611228300916,0.8853124553084695,
 0.7354153242042827,0.06312585312841368,0.11943699822132547,
 0.21159474392548006,0.04419724583885587,0.6078129063876203,
 0.056978938827119574,0.9791772636225996,0.49333800583744436,
 0.8142683001019109,0.32907561816615416,0.8045345514791329,0.9415298828730622,
 0.7566702491322694,0.2405592003604088,0.16428405318687644,
 0.25610628653476775,0.953141460562552,0.7369591866742428,0.1699292367773303,
 0.5500738144860812,0.34452643348965795,0.6759711879136963,
 0.33542930157942386,0.49738789313173914,0.6093644855183545,
 0.7588508934177433,0.40539202908966465,0.6876335536297047,0.553001944959668,
 0.6442756212396,0.11548642475058934,0.549134712840952,0.897791669320376,
 0.9550288377072417,0.6653207359357524,0.52157032853627,0.96710718165941,
 0.688592267985065,0.038348326035355096,0.5267576019469831,0.2031541536372664,
 0.7838313838325116,0.7284513707990824,0.0898819171419144,0.49153859111311216,
 0.49625237162389657,0.15947383321797526,0.9377372578146098,
 0.03093523272519627,0.38025457085409076,0.01531780866694854,
 0.8774538369105765,0.5789883271600725,0.8271576096154295,0.7670966522282658,
 0.10043259939840044,0.8955136467968138,0.830745230162086,0.05286058001078664,
 0.2851328694712153,0.26186942241599387,0.18230306609069846,
 0.38671960241520165,0.9404903426786443,0.6985102472687452,
 0.28775180046013715,0.8850719688064667,0.8788898033957729,
 0.23543377887407424,0.8918835468485113,0.9206344837645714,0.7091148683364985,
 0.4574426212598567,0.9581440713161803,0.2460439247181888,0.8961486820543066,
 0.7801097606246993,0.8576136579275996,0.384616394631906,0.24925452738761067,
 0.7550048117976141,0.9825306666941902,0.6017290559424078,0.26356640772623097,
 0.5899448803393006,0.30641043140710034,0.7425846380340048,0.2020884256136083,
 0.4653859389797753,0.26956972651634914,0.5642213469331325,
 0.26634937464084407,0.06780642229581357,0.3112794518822868,
 0.5976469799250319,0.37295902667253567,0.8289335962698857,
 0.024801795682205308,0.9326930984576066,0.8515448494076443,
 0.2645042295309816,0.10886864158708598,0.2515265446722301,0.6256296771364147,
 0.05836085623532339,0.6659308930162192,0.9413402272152167,
 0.25294342185524477,0.29936794832595326,0.38911542631503937,
 0.9389651677050879,0.8672878821325244,0.2916812159060689,0.7009577621868398,
 0.8550876463848611,0.5462127941882998,0.5518628260005058,0.7491453930551661,
 0.3137472491559701,0.918408606756136,0.0848907409050087,0.35500360634506234,
 0.3721681247925469,0.5251597067840681,0.5564245381121303,0.14032802303851033,
 0.858895666027321,0.8051038449985473,0.07926837032276068,0.890688776643321,
 0.4443443847944657,0.583163788251894,0.07481216459800033,0.6979371255364651,
 0.6675321668598437,0.8212396751171899,0.3951741189326672,0.5553046602854211,
 0.3143676030472521,0.3871703306057702,0.025549283491776875,
 0.7556129567497556,0.9669154160521949,0.15906083157758188,0.2969018220953883,
 0.5109204416595876,0.7417757519070309,0.45579642186532254,0.8592956423842141,
 0.582799265252057,0.8497296186034832,0.23917717188166188,0.194399646612622,
 0.7014401616560383,0.023899652250038206,0.16453733495126566,
 0.4407550678521387,0.23108170890833146,0.03385907372829289,
 0.004352393600956628,0.2880160467100856,0.17867687696718282,
 0.43409519714777567,0.9040205861436548,0.6454823277275115,0.5034006815071915,
 0.4332042689070611,0.7233262394481362,0.6732133493705176,0.8806762204659246,
 0.6525931788491882,0.8943987861911769,0.8927606592572864,0.5994814417848555,
 0.33238693786986584,0.20239265619284186,0.701197883031864,0.7635982429564423,
 0.8988785067724894,0.47164379297445624,0.3274995776401963,
 0.20766325137310448,0.06899936360618486,0.0457448896660837,0.335525206677201,
 0.3129042868870794,0.6243781906476987,0.9751005724256839,0.40388847661912686,
 0.019245956998482527,0.3192843798030487,0.9708511705845045,
 0.41784323668492196,0.5018110886415749,0.10450701339176793,
 0.9794127051938384,0.3144858931367751,0.3961311487934803,0.01861538264403828,
 0.17476029767672285,0.38579188325915625,0.30983174298040783,
 0.366308599829236,0.6593831288277999,0.5899971059204097,0.4816429575849488,
 0.6751535743334313,0.7873494931418542,0.33055295800413576,0.5054362566188046,
 0.043770589255689885,0.17228409923177268,0.5190453589545557,
 0.5065802635721909,0.019856892116772218,0.4709942473046904,
 0.22233832221107175,0.7681849119084276,0.2969014772504348,0.8233926439547261,
 0.36903035055177236,0.07038216071068082,0.8189054858071798,
 0.4135759561965239,0.810949513949849,0.9992254749896158,0.3782363973616669,
 0.4702469235045996,0.6842693910090679,0.09021532496783002,0.4400799783997992,
 0.3738668810305359,0.8918441489385116,0.5573561752403217,0.8637152407270955,
 0.29411711912432115,0.8241563951162144,0.2766631260148573,0.9742896345326094,
 0.3310648793873652,0.4224742311903089,0.10437766581921726,0.2741194582403411,
 0.4639407632250664,0.3784973423933471,0.317721625288537,0.230827581507969,
 0.7017909413069481,0.8551632198449507,0.5380664218880056,0.4282447992147288,
 0.5927551830923053,0.6620063585005456,0.01234329414200952,
 0.20372535592345453,0.21096942170964172,0.03400806774947995,
 0.05669468709876391,0.44874107369108407,0.24705500286920845,
 0.8809920785186869,0.6421682855184779,0.006968138938301061,
 0.9653406655180646,0.6150936821099693,0.6317107554739817,0.655022383756732,
 0.11353941874844631,0.5477603488787209,0.04638460160912283,
 0.7189691305151835,0.7461756080425584,0.6545657552414543,0.922228472291348,
 0.1418907968973988,0.9832535011623513,0.3804765935318293,0.5034302990698735,
 0.1591477841290263,0.04426341941282175,0.2645596926060301,0.8299214541476859,
 0.7005002115684971,0.01400161057481264,0.4188272907370638,0.5203605316872092,
 0.19461855500589476,0.8165738712808414,0.8272695246541084,0.3704435768552413,
 0.9515991577588947,0.6548270638370892,0.879514885131763,0.5375948895300315,
 0.1296432242214971,0.5834221616253769,0.07544305844083443,0.9353760850948427,
 0.18001298277248157,0.9989921647812454,0.5014851976193561,
 0.30052242228446024,0.7902789010724491,0.7034427992259067,
 0.039725548600114235,0.9335433936907319,0.34298669858307995,
 0.9348917267141537,0.6977946809668271,0.027514127669650668,
 0.35270050060208835,0.16918547015295837,0.03786515464426554,
 0.5733167205479633,0.004459270252642256,0.6103917271342276,
 0.23315531568399206,0.8990307602122976,0.5129535197583591,0.9858008952671229,
 0.3388213835263659,0.4442678678164579,0.07570794777635625,0.9890740651071168,
 0.1173715887547142,0.5807693127714585,0.7507232903586345,0.4851105078193946,
 0.6358936400834981,0.5193248247233887,0.9038168249959858,0.18721524887389185,
 0.36961224685623695,0.7379979159915713,0.8932495757725333,0.9843474753970998,
 0.4941470822158478,0.06745556577682854,0.4974941952273233,0.8525946513025204,
 0.9169458943906863,0.24238110930376844,0.6931867521860626,0.7217188223970687,
 0.6949664876970643,0.3613282321536667,0.5120412448970431,0.05415796428238706,
 0.05802719803375578,0.85319610102622,0.5205960256684335,0.15038580647855726,
 0.3559082869184067,0.23230296420668894,0.11871923789249783,
 0.20761722781671266,0.5454259568345303,0.2950524765351351,0.3650171340662831,
 0.27869028005157537,0.6577499333429989,0.7251436746236103,0.9996917830824321,
 0.7093375727340103,0.46153823948338824,0.8872549753509129,0.3461660091608999,
 0.13950579427469345,0.6659618255001518,0.8421984417545585,0.6817074580194434,
 0.8822686615229549,0.9268014846507782,0.5296328648235455,0.05323440791641365,
 0.9288933763637383,0.27505337969418453,0.851706465647916,0.6279066121168643,
 0.4412204116146987,0.12472914342658703,0.6139170201919469,0.5309236141352472,
 0.6946496284704979,0.29620621918594814,0.463650508279864,0.11582436418438324,
 0.713362895555105,0.6174384687408873,0.943788834268043,0.6237019848312043,
 0.10144087028581505,0.45649675086383845,0.6982085327622768,
 0.38789223922252736,0.1703106102549381,0.756264637576624,0.4292344555847861,
 0.14641808260115474,0.07936882471575302,0.9914333268989725,0.625606617676056,
 0.2556006466582982,0.6263394637479822,0.5967193340434536,0.9789141782294403,
 0.7830427475200432,0.10857544594286561,0.04206183350710213,
 0.8347544204812815,0.5528261225596559,0.9844160795297823,0.25990527056745205,
 0.7507544807023667,0.25503991945263127,0.9102906156322723,0.4537984579025568,
 0.8202815046787887,0.1558867062598337,0.8169801984065789,0.8746867377996466,
 0.5166501570105115,0.7680603333522423,0.094283437051774,0.5336330525118461,
 0.8528442326397342,0.36928319128911813,0.029570632795606677,
 0.2588721029982669,0.807894222495088,0.6785913486930588,0.08867328845972888,
 0.10197235207415212,0.55076973805699,0.15094948187487078,0.8370769305019741,
 0.1567805638161599,0.553391077762724,0.7996881614661755,0.8446678154238856,
 0.19947239261883443,0.15587892457824315,0.8096650041567292,
 0.6002681092233872,0.1798294445010612,0.4077547478961674,
 0.048949714153544766,0.8362427442440578,0.5582702276313956,
 0.7995416191114566,0.24060027071223178,0.400966922134128,0.24524284193143642,
 0.24748489689042397,0.5707653616153708,0.90496007199249,0.9220206724349421,
 0.7465186868576226,0.13797157481113032,0.8741355408134779,0.8299680088874273,
 0.7031896120721871,0.5279871724166219,0.8686548044574375,0.8760075912006298,
 0.7284232733466377,0.6911096195165694,0.20634201393607432,0.0663922799048694,
 0.4304011075806816,0.21042430577025217,0.41061771474531594,
 0.22201873712365006,0.5206362595474866,0.6711703402185014,0.9499111642245547,
 0.4317012789304927,0.826432193968838,0.07306487422674557,0.3896526956413263,
 0.5521030036372061,0.33217092057409014,0.884844049759153,0.6894092142954601,
 0.7826697518865663,0.17388327613972532,0.4283116219066826,0.5825296566511781,
 0.30276284880792037,0.0937993320283399,0.10018101123499357,
 0.08884254337227593,0.9611472143880765,0.8704752153822722,0.583263875628811,
 0.24217726362216363,0.13874357063739906,0.46662792648925855,
 0.7288825242917685,0.48728840608205093,0.7586317202609695,0.6929398335409076,
 0.4023826917018609,0.10467314684997231,0.7089346550994843,0.7639702386225029,
 0.020502840983077975,0.44235774990149523,0.9385853310048039,
 0.9912507983607585,0.5216820283834246,0.014702231042505898,0.804324728691563,
 0.26165677590382563,0.19385682319391107,0.9445225906992435,0.284785320015453,
 0.1538712183907447,0.7733862717005283,0.19455422450229176,
 0.25675376300350183,0.7315981480945246,0.06492215670638024,
 0.05413999911992062,0.891063634066026,0.3288343544373734,0.8030173746142991,
 0.67890081222526,0.7111894495587834,0.37946271621595473,0.01776058340727915,
 0.43668812623194886,0.15409630044097633,0.5282138932752656,
 0.22477826773231302,0.04930268129063009,0.44974070757726925,
 0.06712513950966736,0.6696243453347572,0.7419468685636899,0.5664809521223516,
 0.11786472865518993,0.3025619868623254,0.42770284761322297,
 0.9238932371203132,0.7842595525338184,0.26127107717795195,0.9849247089196348,
 0.06301106159675951,0.763485627465855,0.542330923746754,0.07962299427329844,
 0.274624201018699,0.25199151533471564,0.8463361188801063,0.8730594778965965,
 0.4714982098902145,0.8056468353049024,0.7136121689649667,0.3498797410765919,
 0.46314439055142487,0.1611479076219351,0.4359927186776399,0.6248495560787923,
 0.9689976364286114,0.8805092248003876,0.3486728738260947,0.5657569295927745,
 0.8326916824803635,0.6229937268645557,0.9152067221666944,0.10549888781853178,
 0.9541793855324624,0.22263329934416143,0.24784769936822248,
 0.12784853238618932,0.6475758406632702,0.18631711730692502,
 0.04099116762954169,0.2751533389046834,0.043006820760040476,
 0.7475936031339585,0.5634950957963355,0.42944434049388414,0.6438560783170475,
 0.11650801920886877,0.557788604271736,0.4035712644568037,0.4570827597541801,
 0.9074672271909571,0.6545516414147425,0.7715182206436287,0.8432909635500121,
 0.6931731848576931,0.4265389207102013,0.15244129152497077,0.8381194330608952,
 0.22508420722669586,0.6343077258681397,0.9476327860794052,0.3805174256894851,
 0.5405596888499485,0.7437997460971468,0.7018769512691692,0.5522112863494958,
 0.49129138894265023,0.18585004143765782,0.18035607990418545,
 0.4239962663524792,0.7696408255489281,0.4261280245797865,0.7270653645401578,
 0.5392798061391678,0.9935536332113389,0.8534220230135279,0.32272663519529043,
 0.4555664714035753,0.982158516015532,0.6811134558486998,0.67564885930519,
 0.3414641291402156,0.04198292565178474,0.14510720305711378,
 0.42148560022655635,0.23129607800912133,0.2198063018529759,
 0.6306981828539057,0.7817631178035305,0.08353486762516604,0.6765006848188013,
 0.21590975015523417,0.6557858520610905,0.8264225045470842,0.5594204217078336,
 0.892911324706523,0.018184166469666252,0.06070798720669168,
 0.23458362679709777,0.6962747090894088,0.031878358362113945,
 0.1550536200360556,0.47713841975402393,0.5438081803216563,
 0.21491099042077733,0.5301418798378773,0.5332086861649415,0.7422820704179098,
 0.8034701450755587,0.11646106654610933,0.3283379330240913,0.2737796642300835,
 0.8228705928286537,0.3993225008934109,0.5154203671513687,0.5129329048027302,
 0.5919639736660381,0.6629110150333262,0.7315188755256918,
 0.0031911607234649253,0.9222346058718418,0.16575943326250475,
 0.5568344478257511,0.37840769400221774,0.8233830967816443,
 0.04850164922004807,0.4979082241035325,0.29958871419583755,
 0.4459673663532169,0.0544759456952173,0.6991368290375944,0.3714569341331232,
 0.36180037415460964,0.1948557209613202,0.18328704310822375,
 0.3711919619526539,0.8927372281581648,0.7807002938903641,0.7255956898650577,
 0.3462894678289826,0.47711408729720073,0.3411415918269872,0.4448227843943856,
 0.8192306724228032,0.8720592285696792,0.5150798776975072,0.982495056987291,
 0.3878032319285697,0.1750767899072655,0.9172908387004738,0.7298807838116643,
 0.06824309268537687,0.8537140896278554,0.9674166416698476,0.6419317542835924,
 0.3541282415839767,0.5276222572764673,0.8651945448433481,0.16615086269885937,
 0.6970485501729528,0.08859469257320496,0.325381156161775,0.16762940808412197,
 0.6309121939734131,0.7464326206857452,0.4956576097243144,0.11412077906695128,
 0.3018912431140739,0.7053281577350564,0.35803588735574643,0.6288930164681561,
 0.01369603942779729,0.4172675369828587,0.837604182570455,0.41798072848612844,
 0.3580934232160543,0.7510217927210304,0.01126624188535752,
 0.03671838171097597,0.17587076589168582,0.9009825559217757,
 0.6974651059855691,0.15308384539829567,0.0028310489477009204,
 0.19818850941211774,0.7028032636628996,0.3828228097380897,
 0.18944174043880824,0.4915314559000613,0.42113065648944226,
 0.6471452669101794,0.7241586471990992,0.2703404310387627,0.17886845770357507,
 0.3741432549529017,0.293577599688824,0.1996000252063589,0.6759728867362497,
 0.022287546833608785,0.8216394952020964,0.0554396415600823,
 0.2924093012854839,0.45344938044605854,0.4123163334055522,0.995723924697703,
 0.1698887954494488,0.48479334469667923,0.7742369374494424,
 0.43613440874530696,0.6676026366596778,0.9177575005166473,0.723806689722516,
 0.9425499938653867,0.8344106878276447,0.8778869562179483,0.89079633312128,
 0.8107106587766191,0.9107155629435844,0.12967263925463024,0.6433561457962405,
 0.571833571542258,0.9550455632728977,0.7386230856685779,0.5074081248035822,
 0.32436029463684535,0.38906339709148363,0.2661673537678142,
 0.6110923376967128,0.8628538424921849,0.3698674140177016,0.477219193096833,
 0.6464426844594809,0.25569086483605097,0.8874114576615388,0.37971209335447,
 0.7733179284325108,0.8846465225322946,0.6844210285259946,0.9880220134862843,
 0.9727162866640904,0.8754859629101663,0.2549644775129175,0.5036413907414454,
 0.29463047866407377,0.2117653839595749,0.69793770326741,0.9655790432011692,
 0.0450824250228441,0.17765444175929568,0.4093325450181119,0.6463490993228265,
 0.11184518436400448,0.4975069182738763,0.664593088888222,0.8841964229405369,
 0.22952908045182363,0.08241152465178747,0.7680978861697608,0.729144901524629,
 0.928980112083126,0.6743646264266107,0.2970792329920333,0.401272220466217,
 0.1432884079564254,0.9995591587088775,0.5260114110886913,0.6874033625685627,
 0.05349611045959335,0.609207196500087,0.5011513194873103,0.18011770550853845,
 0.2571050175612698,0.026667654820581643,0.4709097561673874,
 0.6075028077816462,0.2892394758827397,0.6088955227670036,0.4971948907495769,
 0.7696213556091057,0.1463400565675631,0.33817404038720134,
 0.006038199482265472,0.8808284735131097,0.7045605508417631,
 0.9784310100278895,0.384789336372545,0.8678907460765726,0.6120831648497579,
 0.8158497483875351,0.233598953128343,0.10934765825711246,0.12215092552452844,
 0.5422535799636895,0.12601593972348035,0.34518043668260157,
 0.9892657376089518,0.9602651227046528,0.4308567154249672,
 0.060333182257720996,0.7382406315369431,0.8044358580184261,
 0.4234510019456109,0.44973541070109857,0.4617764044699624,
 0.04792156710967954,0.5608819312120762,0.8830587495717421,0.4690502309358354,
 0.8192086979081207,0.7358293265340509,0.8476242074136368,0.1423336117206152,
 0.4934815995197368,0.8730961954945232,0.24976430363623636,
 0.48306163062028684,0.289651714698264,0.002828714835285129,
 0.35744187683208883,0.01760900363518081,0.22077793570716697,
 0.6359094861650894,0.6698771985776534,0.2255549083109456,0.09595464906753515,
 0.15825053504111675,0.5014091544212496,0.5947162882571079,0.7870420216825815,
 0.16395216053123507,0.11054606477424112,0.5883270327805082,
 0.2718086228514348,0.099897004713834,0.10765445198870616,0.9308145347739661,
 0.9832429177874361,0.0876665339647329,0.6931358887820778,0.09893763716489579,
 0.5064256012547096,0.6790603333973738,0.7265081959405143,0.9834290115119928,
 0.5733946829385403,0.025628973330413474,0.216334181581049,0.7212641951720565,
 0.11360275224992744,0.6940917339748984,0.1473714517331155,0.7178402192708859,
 0.4269647537038255,0.8837882848029803,0.8068296508958753,0.18721738892939288,
 0.5824292100816403,0.9338839896673183,0.6750899797306807,0.19124952213287405,
 0.7638369635961247,0.46030475824484607,0.4970293778282041,
 0.005377014763927157,0.40709954319572117,0.4183203985258396,
 0.8142228704699214,0.7850506951332215,0.846028180182567,0.4107758675831218,
 0.9037722059579943,0.358455712303025,0.12725566667213173,0.5332435703246855,
 0.9539276743997271,0.6282328509562211,0.4836329509191969,0.5923558715532159,
 0.3448900671267243,0.24085408056126567,0.36106475798839277,
 0.36677204046441836,0.570230387424205,0.7751937999246055,0.6091366088421607,
 0.5386371298886156,0.5401763637381823,0.5199873355360974,0.4025774289193996,
 0.9802246615507388,0.3747509977614337,0.5710558547789375,0.25381534321165644,
 0.40139955141382966,0.6795941744144899,0.20389989710574996,
 0.9036580204989959,0.31334780516180305,0.754436954753225,0.7264927831928879,
 0.5438235660394747,0.7007559605125773,0.0853295084153427,0.6472440199749997,
 0.38702568141567406,0.08301771798111213,0.8813353168829764,
 0.4597766350113014,0.4197462338978132,0.9689286416170289,0.6308681326126498,
 0.83967536797677,0.8491515847109169,0.07936665678227772,0.34941291361715987,
 0.8534692594947759,0.5614016443214478,0.9029518631245557,0.17296198430093124,
 0.9489794874396926,0.3888019843313709,0.9178485563227633,0.502512526769859,
 0.2632098935514151,0.27753877668075844,0.928968709239593,0.06964694632957447,
 0.49007573296655815,0.36764809946664767,0.6678889325597731,
 0.27682015919333447,0.9176174092949239,0.2273361341212764,
 0.02074956641961867,0.29769633329685985,0.044515188840979714,
 0.34347324543919666,0.7348625272176172,0.6248475544570627,0.9979918265076113,
 0.33727953938831967,0.28924084908494296,0.8959149719551804,
 0.5353383565725263,0.6766713196080757,0.3467093398908886,0.21461590996232527,
 0.6670036026917914,0.3108505726225529,0.6114441339886654,0.33627976012589444,
 0.7637638219089071,0.04192712131947851,0.6115977144284636,
 0.38429260512796104,0.13384789354757687,0.18412400687115293,
 0.3788750886346137,0.6536728089927296,0.5261964696867275,0.4443407485316273,
 0.01509844345380218,0.33697651428090447,0.21713633413983757,
 0.30417391556966356,0.48792494522120544,0.015771019036146083,
 0.4663960219497978,0.8331108303276233,0.7010752730181213,0.49124796683614735,
 0.7640068703816351,0.6891677492023301,0.9358901340841274,0.8964724625239386,
 0.9049287010694618,0.7125345753535027,0.2134166268102542,0.49891800175523204,
 0.5235233743306994,0.2821564817402762,0.9414612684341306,0.578027010972111,
 0.14414836149382038,0.12265121361801423,0.28084069196839034,
 0.7020927496669607,0.6334721904128375,0.09380539058246806,0.3968086017271433,
 0.6080743461538269,0.1263804273576835,0.09422083987609009,0.7548628516297995,
 0.19683652831576615,0.18100405055943847,0.8889310308353343,
 0.5132573355439572,0.25537176084408464,0.052420549512906534,
 0.6505850997178315,0.5570387828483757,0.9721031163055911,0.9975458253407705,
 0.8815732510909317,0.80176558501349,0.3359757888639543,0.9404263064816042,
 0.6146964076762295,0.2718467707411545,0.9846146984379939,0.08129527310260165,
 0.7991013955575166,0.4233228891941033,0.13918195442577175,
 0.09787764509887809,0.40082940140102763,0.6047123537961719,
 0.43691111179550424,0.3357392710309469,0.5456224355286462,0.2086271162839205,
 0.7880908405971165,0.8117155920755262,0.9735329946883438,0.0807989459186298,
 0.9424186210220686,0.24123147075500828,0.7470135958384737,
 0.37021224239274175,0.5275640054477445,0.1235350896973183,0.8299592277493,
 0.3934900111869183,0.8434861804869772,0.4497301624952639,0.7615295643278344,
 0.2532125607614626,0.14609910502773926,0.03886393336013283,
 0.8261327880424554,0.011647942281501233,0.7837870326947466,
 0.3508387351482114,0.9157978714797084,0.2976380659047323,0.31853580947141413,
 0.8558215883290676,0.06324116550770886,0.3190066504864547,0.8577068021898007,
 0.9653204079419222,0.6472061554189795,0.4009212097567978,0.7844798607248366,
 0.13390260789207775,0.03349170715042438,0.9321119855156967,
 0.23691475733842593,0.3456463951121733,0.31638575185950646,
 0.14269065733685116,0.908082172088624,0.3975227494700123,0.4197009250853876,
 0.8247379801684939,0.9469268894137961,0.5009114501876114,0.9572517621081361,
 0.6446763537604933,0.44899479282631993,0.5951188427131818,0.9943805221102422,
 0.007118312216025657,0.2293409294890435,0.0907176224600772,
 0.8519018478536934,0.31956229670950054,0.3581587532682945,0.5653412124441034,
 0.24363607908864438,0.18704101347580637,0.43319123499727363,
 0.7124107375222011,0.8768941286523436,0.03495513656457172,
 0.09005866517975347,0.4209631108998756,0.1408857204184254,0.4195596669621897,
 0.41496144777044275,0.7521809063418339,0.13469627686936625,
 0.7657213039670956,0.5991282508420969,0.5286641083733287,0.4296673185173947,
 0.4077672175105961,0.2363253413446299,0.7573470752817197,0.6073906685995637,
 0.25933857870692667,0.5316131159027985,0.2712420956339403,
 0.47366956822063633,0.8957714660024544,0.11829965535297782,
 0.4564250114031191,0.15719344103325827,0.02729665190095143,
 0.6735145932641798,0.6534346755904397,0.033880354949056946,
 0.5742308435138813,0.598970445804945,0.1708865332618914,0.241493836984225,
 0.522903054774323,0.5063972555807287,0.9327633853936312,0.9195198096488024,
 0.6299322721946847,0.29060523122497817,0.5917871412655263,
 0.25645591412721036,0.20778538147937864,0.4356093888320465,
 0.21494442178169615,0.9037886931693522,0.09814079638608697,
 0.2023349318459382,0.20689771737704554,0.39704759626510977,
 0.32961980755736897,0.45312216395626437,0.022548769196572782,
 0.5736974036529655,0.751964612003862,0.5137315951543417,0.5223918755357866,
 0.6291845189897581,0.38338224746778704,0.34013257072745173,
 0.8088531583091325,0.5368071348014332,0.4967791944677714,0.2974215907419794,
 0.8148775930927815,0.04075640320517615,0.32197592603035896,
 0.38903390730603493,0.08430358041127095,0.9699545146754053,
 0.3181322501726124,0.30213898856493315,0.03483330789864547,
 0.4124162914820514,0.8970377816229081,0.5962817232309019,0.7465538865483738,
 0.6901533374359015,0.023126452102550266,0.5044658675859928,
 0.9045095298344702,0.8465702533171513,0.15342234200204197,0.8627788616696178,
 0.04953587832163853,0.7483064344657204,0.9591830804526379,0.8337451030567652,
 0.9415923716827163,0.713076282679576,0.952835830390166,0.21220496192277616,
 0.4288773757582053,0.8865097191864928,0.5693042551712477,0.4767533916632999,
 0.8401687151205856,0.20632256508134283,0.03624613963183909,
 0.5544848654879713,0.6530672384789138,0.989737474594653,0.2019144231750143,
 0.39538068619241873,0.5482950984306935,0.23762871377271366,
 0.09019618223812231,0.5135640956376608,0.9966947820362178,0.6609743500919356,
 0.8012335376529077,0.8077866611391595,0.7058029300994524,
 0.027123016199003702,0.5190364643246328,0.49291139775581805,
 0.9859163676308129,0.5565766457210228,0.7761229594070029,0.4776553611302594,
 0.5109829246391282,0.6224178375831765,0.7667575426833892,0.03681078036720331,
 0.3170378158826954,0.5130656760934257,0.39515502137918845,0.3537865177212478,
 0.17350451709614367,0.5895301507319195,0.24764546899612727,
 0.06781112334652495,0.6730848043047624,0.4343715117902719,
 0.18125850539774924,0.7839922271278913,0.8197540901758704,0.9214051782495225,
 0.45154852605472207,0.8823087170402072,0.17916481148777696,
 0.10629089020204963,0.9735745953523176,0.12061622876705713,
 0.004302352801701659,0.32198572313483576,0.44000241557933917,
 0.15946757419222002,0.4944470686305362,0.41217059902249276,
 0.8082043354803787,0.8941579687422688,0.13525226200714546,
 0.12650009795096062,0.0014288500386619996,0.11578326403235861,
 0.36860293594680926,0.20704011162486602,0.372848031269267,0.4833285660165676,
 0.8267995544745022,0.1293264059847099,0.6366760978945736,0.16215233795500272,
 0.679004321957187,0.7383612253089266,0.46287737266532136,0.7894303562195804,
 0.1975734112240466,0.789440353679523,0.9372288551501796,0.4978773992126161,
 0.23880683197583308,0.7587200519570257,0.6313953006253666,0.3611291037127693,
 0.8449911672670787,0.49581361760540854,0.5781478784047147,
 0.38117054094032765,0.08174528478934218,0.08796653393093123,
 0.23565917188865693,0.2697680847874886,0.9406025966244143,
 0.28933696951185495,0.9689830680047847,0.5776403917147738,
 0.17318665844102255,0.8417305545261325,0.14066220035360366,
 0.39588120629798995,0.810623187691277,0.6084101499885407,0.43177806099198746,
 0.7035673588414251,0.19135986576097574,0.9556820557813995,0.3722469760708872,
 0.5131881533018897,0.19123200241720517,0.9465424256215484,0.8120402428542327,
 0.20619071498238417,0.5931349536334376,0.22092592225152208,
 0.8393436205091451,0.5513941536678717,0.3192260008958486,0.7682291141492595,
 0.18219762809825713,0.12536219197306075,0.622674872499616,
 0.26049256025415535,0.6628879511130563,0.7175691843030734,0.5101457502524295,
 0.1445210055586328,0.37859964122363676,0.7158260478348633,
 0.24083958158086438,0.7585962757927773,0.830906772654215,0.26244225758170825,
 0.47708053621636504,0.8790460115348183,0.4403220280583783,0.5700382960645058,
 0.755642195052156,0.9971212556213903,0.8976398416147985,0.22833645033557004,
 0.14645956709057228,0.03933329857928092,0.9123083499544113,
 0.4728236696969832,0.7012986403993564,0.5720940663451298,0.32141359012215676,
 0.9241739270147323,0.038768298517464994,0.9195236917418087,0.569125896090305,
 0.9809355109078036,0.3143529069396267,0.8681588430546567,0.7945036610445843,
 0.3251673109800264,0.7484221021569297,0.14881726694386976,0.1262351070723522,
 0.1794952214497202,0.5991528197744405,0.4760924090300386,0.03262680905196336,
 0.1925857058041025,0.051866073826280945,0.9788923898766536,
 0.6268155066052513,0.3880355324860587,0.5747352265890062,0.9326965013149896,
 0.961945884062808,0.19823029906924672,0.8904423457860517,0.8642879599395035,
 0.6084579031780447,0.44919667662240925,0.5081303314002074,
 0.36832696449814906,0.3076269901045172,0.486310215246772,0.9140541579968934,
 0.5902201709099435,0.8792813826804252,0.06761696072701429,0.9678979124490136,
 0.507456980041094,0.5231467647214928,0.9093908343406583,0.3011850963127598,
 0.6455306686710923,0.28662650499226094,0.37937008294710595,
 0.7161623822375224,0.7162200010087975,0.7744536940955956,0.14811558761498966,
 0.3903643458747653,0.8044168688914163,0.8501005381657025,0.29021331868835654,
 0.925198688866677,0.7856216159303893,0.09861124013028788,0.6690778046487521,
 0.09187503310196965,0.6676794447012913,0.22044052958797522,
 0.7133313668848807,0.6188763760935915,0.9547281619720812,0.8920543950471286,
 0.36161235252850155,0.5516113731112235,0.5649826072022875,
 0.17637205297035696,0.7023980797572855,0.4781200267610052,
 0.37130923935820914,0.3045141780147954,0.5559278302346947,
 0.15387618096889322,0.9376349388348761,0.5774467670313783,0.5329161379078216,
 0.13645823717984618,0.5015330377278013,0.6014071214524914,0.8354133505712313,
 0.3780785115632814,0.08828898607367663,0.7862601393493689,0.5795887133640929,
 0.48618624306553615,0.09257830412252688,0.7992450615034344,
 0.5757395854230477,0.8719590329294995,0.32161304250019196,
 0.10310859048850651,0.3916966052720934,0.6786946052758641,0.66486478683794,
 0.9107344953465628,0.5272706316604084,0.31772342414447863,
 0.02658387779029958,0.5882350163297987,0.19409452764348134,
 0.8862816210805651,0.027699907032646354,0.0704091150730094,
 0.1688487799575784,0.17456493915928428,0.5535137576260533,0.863639389739836,
 0.5661826341494737,0.6702005285827918,0.8748496168399521,0.6138469954226567,
 0.11885770097276938,0.9563634163265787,0.33147473331566524,
 0.16342150218474372,0.6436515961452958,0.3809766267912812,0.8866642363921421,
 0.4952185423918417,0.10182434694297282,0.4184987936800979,
 0.46198365195768987,0.6732093891903095,0.8433393933206036,
 0.18395911713201996,0.9267473306560162,0.37354051658099685,
 0.6005126627487615,0.9060050114214969,0.2465451387145141,0.3976981445855705,
 0.7951374856089152,0.9438956930651585,0.9486304807688898,0.12492410157127687,
 0.6274041869956103,0.20997982666028825,0.15867743960734304,
 0.25688939830908275,0.821748016282454,0.6232094129315128,
 0.043450153660061996,0.2698279823308253,0.9622040982726789,
 0.2553129193856969,0.3290412454075555,0.258519870787126,0.6771573317719912,
 0.5677882099893559,0.9691060086313626,0.6224856790874578,
 3.485347990745957e-4,0.7602560346739211,0.3030056517529546,
 0.9522960719319399,0.9879579474957166,0.987653032984202,0.11027913067822392,
 0.33237014370374107,0.39495637346136037,0.22959668937878275,
 0.044486162992800504,0.975999813845247,0.15131383708055524,
 0.42532332046462107,0.03491305940150258,0.32409438321994033,
 0.3505854784289173,0.21906424153268356,0.7618584056336389,
 0.03042251280611208,0.7370506920035949,0.6391916045579132,
 0.48679487917575626,0.860298372815881,0.17826071292091195,
 0.17907243961495622,0.8520677954842306,0.7045170506087566,0.5597163842080936,
 0.0152240096171532,0.5522755254136351,0.29803701542925076,
 0.35948394139039663,0.20648239809827917,0.19555733350151683,
 0.630926506170598,0.6345000347423663,0.7889001124075614,0.014754009949395352,
 0.784490948860042,0.7032290139920678,0.3400461058552936,0.9569717179725138,
 0.6105020774072875,0.8971815262870544,0.19493787414221453,0.4486761813189788,
 0.1963831359106346,0.08813780732704879,0.9046567837147248,
 0.16634732490131854,0.7272370685437209,0.3004484852558158,0.5529398404453192,
 0.59691763145634,0.6105126385524358,0.4077210219730679,0.7233479859999898,
 0.9167166370080946,0.11487823694279853,0.9482300201601,0.4760829448770698,
 0.7968689506314011,0.0046262035289980075,0.7064759558455275,
 0.8273626564999688,0.009973750178518204,0.04167297122937397,
 0.9123032928421828,0.03213038168318916,0.6051738278080744,
 0.15043934296182215,0.7322850467334125,0.7890273361624364,
 0.40897080181160705,0.33100127989272465,0.8513354262111696,
 0.5221074122616515,0.13603454691534578,0.37645917425073205,0.918849867782807,
 0.3782738453826173,0.7232896211539294,0.4205813541611131,0.04724149102638553,
 0.3743008043186107,0.6386090622202841,0.9270296407729006,0.9550533639003354,
 0.6427862568197731,0.13328418466923253,0.43765497528662123,
 0.9834368581563077,0.2118430656111101,0.46968086029338607,0.3288947914935578,
 0.5673802733592362,0.7165059180774912,0.3314086213271463,0.3867841595635667,
 0.6585751120861472,0.6684148546356032,0.547011061500921,0.2943477234521582,
 0.206138442344585,0.47724170213066186,0.6703643825611425,0.8507430264805445,
 0.41613503172051913,0.6422684064032348,0.5748102630829225,0.8432585034432908,
 0.3424009926228875,0.7407981338662871,0.0814401413521334,0.06628004522175834,
 0.9200739943355909,0.11630191555678127,0.9359395480123742,0.8627617484309467,
 0.7559097218975431,0.6665092109196769,0.4964286200557375,0.8416357267915998,
 0.7594122951791292,0.3263288480481674,0.206515027948882,0.4359266749573254,
 0.3015454314059449,0.1925981906398333,0.5898707845549775,0.465607824646168,
 0.8571866625050563,0.4118086974465395,0.517754696781573,0.6701478628122796,
 0.6907522770242066,0.05407107362100971,0.9653758330366844,
 0.45766990393287277,0.8205840000207977,0.7838380332498506,
 0.08601740364669508,0.7597507210811599,0.19394074887845025,
 0.9209055188775627,0.1375149690921872,0.04949800603967569,
 0.18654552148268932,0.20431595865680996,0.3551273896027235,0.552725105511238,
 0.8245219593334544,0.3168124617103889,0.4789019863423498,0.5160091912398679,
 0.5199444242505199,0.5621350984696125,0.7010219356014892,0.5809595446613074,
 0.5081998828092861,0.27228668919361865,0.16258807512642393,
 0.6871494310305266,0.8206285834052233,0.31958529120836787,
 0.23468311732706004,0.5544459556099589,0.6193755749277864,0.6439931183053341,
 0.3537111523130416,0.09700139807133945,0.977974617149101,0.707105184879515,
 0.4271392666759195,0.5732644231574804,0.8475669319812988,0.37950504943248387,
 0.08861282792534975,0.022811212016640936,0.8155146319773331,
 0.1083644818037407,0.6600560025064547,0.9424769639411822,0.4020180509617449,
 0.6069529186157508,0.9991724674483594,0.6943599333881453,0.47416388884219596,
 0.8093260697168384,0.6652580269675948,0.24478110015758492,
 0.12882027678899155,0.11119188635961519,0.564504157648455,0.9164318113757864,
 0.22833590463261544,0.7676520664487652,0.5383141387515824,0.9510819525906004,
 0.005136274990997336,0.8496105553621236,0.6644155057208518,
 0.5660850941917561,0.9142480551945736,0.42386366298897915,0.2023421459093534,
 0.9366833945666826,0.10389803366202743,0.25394069428270494,
 0.5141086922352783,0.6820927950156245,0.9288742668858644,0.4465097367036205,
 0.1710333146874835,0.4511170052387037,0.34337818039117873,0.61441756093467,
 0.08482208726145068,0.2108922601458384,0.9361344627981899,
 0.11449912481482949,0.5814734631117172,0.7023495713597327,0.5208891839716221,
 0.977296583179255,0.44841603011414055,0.08101270062352373,0.8380564171693107,
 0.24716746802277523,0.4765059366386387,0.19739123919200174,0.697151124795435,
 0.7085333935782894,0.32960783495955615,0.7683474885203945,0.8137524793560383,
 0.2856493002264313,0.5196952006473995,0.9646604252080151,0.3895938883537924,
 0.6364550536905192,0.1685813183089836,0.9108595330298912,0.8390330419584684,
 0.2679642424464346,0.25137444255416963,0.5791276869548148,0.9139858576787655,
 0.15563597843857668,0.6678427105870495,0.6511549326540393,0.7867918309692632,
 0.463400400627343,0.9343711761690499,0.6974915455348156,0.2105456007428358,
 0.5681868523969688,0.46897118740753496,0.3748176959949263,0.8453968461754373,
 0.25665329746078713,0.36071050431568485,0.954230212231853,0.936968112384233,
 0.3569025482234296,0.4053713313570092,0.9193987440838219,0.8375630410509937,
 0.2243961333404979,0.27482270495930516,0.8071735313664122,
 0.45132516546798784,0.11389231648009313,0.5095148233496609,
 0.7213706085070428,0.5085498563650493,0.32644996422537953,0.5718999795204301,
 0.5153355429538011,0.4769105607650511,0.649318011387119,0.6833494923566692,
 0.4999794899722878,0.716142676718282,0.8032705878125572,0.54438338758896,
 0.6761641498027376,0.920383595562016,0.7384447969060491,0.4672466214049632,
 0.12707905289798105,0.13940876139401748,0.8089067796255658,
 0.029866956191570426,0.7481590581838075,0.14037401610305766,
 0.39797293928748234,0.32738302885811177,0.8393521039744477,
 0.9620586137920674,0.9356902865402332,0.9381607186087219,0.5629569690428669,
 0.17180935259829377,0.267625650453304,0.6870104949227315,0.9399144937996053,
 0.4803670489249926,0.8769967223196096,0.8826663168833025,0.7578163824274923,
 0.4076407471888708,0.04094168484980554,0.28808166826334614,
 0.017114129787379362,0.8679615993654552,0.7451730903971764,
 0.4567197987222421,0.26636457555998905,0.4435790496430977,
 0.47318564773712546,0.9137035636866048,0.47426567013364895,
 0.9153173879413684,9.596187004432544e-4,0.33938986097451074,
 0.8610531237255792,0.9140002580896655,0.4798659082641614,0.7525163002320854,
 0.1075680628939365,0.8155706553986565,0.5537210593225288,0.8043415596335141,
 0.4968880541539138,0.26186011641112883,0.04563057423072281,
 0.6990271037457078,0.13189983003794448,0.8236787520233725,
 0.10378381860757468,0.6113567775519924,0.30404035283158426,
 0.9480776826012304,0.7944172025119993,0.9807393707728427,0.4351978045330357,
 0.9121880226285193,0.5739983081596234,0.7211031059455084,0.4931259672023127,
 0.6654482964800708,0.8087156925789305,0.03954121005882638,
 0.18769650779502656,0.9769350971627523,0.5759118066279726,0.4659187944149793,
 0.22290738384820163,0.614400462402657,0.07857342307451265,0.3304986764391483,
 0.5577682747076935,0.5167534098694192,0.9313134111274364,0.5375668406066672,
 0.853036599401046,0.8276607114671002,0.991138642431648,0.48128599619387125,
 0.9450996116487137,0.5391415723376396,0.185924663812153,0.8917668237012935,
 0.5925637052706073,0.7231342802154952,0.20194912593487058,0.9865935027018271,
 0.5321605407896058,0.6732015920080392,0.7286635388215152,0.3580352810534443,
 0.3924544619348168,0.8889395881563842,0.7057564886783362,0.4030819999796995,
 0.34102604195330943,0.390266797637279,0.6258924216914443,0.6098528012123806,
 0.8759142210975026,0.5200171788534451,0.33884859997337313,0.5219533662850371,
 0.4432785351989925,0.9784884904230168,0.5894082302833541,0.19341427449626514,
 0.08627048287276773,0.18368031909627658,0.293178847128576,
 0.04051669008041037,0.5377133606352429,0.33286196862428497,
 0.4780550975215083,0.3684415950275589,0.34358335690172825,0.1853452554768621,
 0.08097120944410319,0.9099764008454487,0.5591618211869896,
 0.023508424233896874,0.6183951749655172,0.5242803623307208,
 0.8290398805479255,0.4882730328634143,0.11488814174548778,0.6044108074059439,
 0.8430444555896836,0.9128852367491737,0.33324896282353755,0.682079768164874,
 0.0070122852710684525,0.9073143200032604,0.8289699716000876,
 0.393177291094683,0.7872822759237,0.42636700391968396,0.5103891776640089,
 0.9873311660213753,0.6278257513150264,0.9913123861931917,0.5156010156576879,
 0.4199776108625938,0.44980611142481075,0.1737239954997427,0.7910649803274786,
 0.16307246404096754,0.29753459701645735,0.002022372709575171,
 0.3872370562892872,0.8285255371727164,0.8917937596347135,
 0.049226679368147286,0.05832362852069717,0.9109765189883734,
 0.10694518383125362,0.5403192716625371,0.9785329322410088,0.9461487612703593,
 0.48419992225504915,0.14660844520316496,0.1769711071029635,
 0.4661164034571843,0.9013079268020593,0.346462928325157,0.40362112010108253,
 0.27497446660066394,0.15510533455361974,0.6503103557032819,
 0.9075958652343401,0.5753250737620704,0.9617160466769409,0.09470713989272495,
 0.6398109575644102,0.007051114084219101,0.6715644954552544,
 0.6276597047042265,0.6457312703470947,0.07063397577564978,0.2700133360700192,
 0.978312906918536,0.27369783857377705,0.6572726037646558,0.14112980632742722,
 0.0032274217525593007,0.3124069407429455,0.46353663267500145,
 0.40595895165407736,0.3160461742708386,0.7529318936363149,
 0.34986164515999196,0.3649756280044065,0.9397673234898472,0.1947782636068468,
 0.305866506537092,0.4632829749850733,0.8685857843882087,0.20896363986857847,
 0.686420535771008,0.4209306972193232,0.5896665512386193,0.23526549249958395,
 0.2159452709624491,0.8369762454916168,0.28052470725987844,0.3268586523968836,
 0.561809036773738,0.371857041640284,0.7142495780321002,0.7666581975811322,
 0.04865150008415575,0.1267470651151329,0.024306717231243558,
 0.1915108144759543,0.9013159343794469,0.8639018653862539,0.33811740640302446,
 0.9035281318539646,0.45106518345957225,0.4477322788432344,0.6700084536487525,
 0.5900601814350875,0.07664164429503417,0.5598139660595151,
 0.021070522549504034,0.3481793209221381,0.49992482466387533,
 0.6036243964320478,0.07632868759794253,0.06643080001803636,
 0.12292561042584105,0.17026054951798308,0.7861721015016732,
 0.3095509142552282,0.6271618812924039,0.21099701691376416,
 0.05132787503184022,0.8809524494800347,0.20328622451928402,
 0.4310516335085419,0.522303614999006,0.47962489056754043,0.6037919137537489,
 0.43610133027330034,0.02327010293189713,0.6723978005315631,
 0.6872983570132714,0.6305227928601459,0.7953005024763995,0.8188758183058715,
 0.42030411315934,0.9486878609026319,0.07990251472355314,0.931091162197604,
 0.6325379417735498,0.5241021430543116,0.46269843098638697,
 0.02442484932188005,0.4417646764841826,0.32898640992799033,
 0.5129976605070116,0.27645030431807194,0.4192862145621006,
 0.24371174820994324,0.3385723452741426,0.4658201861775002,
 0.24981324719869902,0.7161101630710599,0.3495548955053662,0.5838679727501518,
 0.7197402686451234,0.24649293362395208,0.4509868953636893,0.467834205226797,
 0.39366394233443813,0.012984477669223726,0.4234496264163291,
 0.30801339294284147,0.18825399091846284,0.48538862637121405,
 0.22023272969340824,0.9657035572704016,0.28915197286087857,
 0.039852967806531914,0.25917180656132466,0.06036545853735387,
 0.23441264589730137,0.22508443593295635,0.21865934210837723,
 0.4846083434684342,0.8976802497539258,0.10453906243052535,
 0.18218136229099913,0.507302896505067,0.030038526500972718,
 0.5327049311909988,0.5864529459774146,0.338713568818785,0.9827859693236441,
 0.20023050624042216,0.23792075227703247,0.10962076695038914,
 0.8316665685851716,0.04056367503325764,0.3409306472535808,0.3094232096826095,
 0.8509183129774051,0.9134184190586158,0.8883756844392713,0.3663649420714984,
 0.19851292161265055,0.920044967173363,0.4732275134976446,0.2051041414890522,
 0.9275736087856495,0.8760900224798626,0.6139548639244032,0.640904291519104,
 0.23340481656127354,0.4300203881838288,0.4611060788126147,0.5978414161974648,
 0.2997394996287056,0.09033129837542797,0.887982857300367,0.8905077694765415,
 0.5371825790102456,0.5637754979889511,0.3457174081292165,
 0.014418184233885523,0.773932138228926,0.49190425776195423,
 0.15359196089181715,0.619737046011875,0.3926024423785973,0.9004758195394584,
 0.8374015279558967,0.9783084068875166,0.9295472745893466,0.2831710161871287,
 0.8667872242661129,0.5095370807631352,0.6018108584036918,0.45242087895447813,
 0.1434564602132411,0.5137648513442934,0.9397392894541324,0.10908313578093076,
 0.21217450851583397,0.26757948593110015,0.6745368060011411,
 0.016749210564288908,0.8646934232564317,0.7848311045749303,
 0.8359872332885337,0.8533468172878543,0.07452247665062339,0.8225208469556918,
 0.08142980272427858,0.985585370820475,0.19832686818564693,
 0.10221206928148185,0.8226133632343785,0.7053972934844465,0.809778977071927,
 0.7324089241551475,0.9987207068505245,0.7472781630085403,0.06436900596853534,
 0.2703412977062274,0.6258403219248675,0.5612419207110466,0.7172864481704317,
 0.23012628083795006,0.07750253809551089,0.3970565862397366,0.092794034194145,
 0.3315350984376898,0.7415150024089738,0.9303710233299283,0.8668364184563899,
 0.49735875114662675,0.19400822601883805,0.6595206104155329,
 0.8807126246191221,0.9082671390538155,0.298437533434902,0.4361051736153537,
 0.6564024612775021,0.5507820375443295,0.7750990025567416,0.595811293573715,
 0.961149892025253,0.5958901606593042,0.23610646182643646,0.49834531814775873,
 0.6840565611866076,0.6650814354710555,0.4771330869463919,0.6645746607013744,
 0.5895144114774581,0.01495788304346013,0.3454045989051092,
 0.10039073034207324,0.6838459016261931,0.8132517087515806,
 0.18552166755273136,0.8295123501687517,0.47584775819447767,
 0.7054677155834315,0.12515988369208775,0.11994623886682021,
 0.42438780361156025,0.1939336038334477,0.08731866694775015,
 0.001999761637378805,0.3466712642139934,0.38313972628558757,
 0.6115163947478461,0.33089127834588583,0.45694348575206245,
 0.6737745016171988,0.31771835298047557,0.017707637256535458,
 0.8354706202356834,0.2684423521447814,0.0607738094836594,0.722274540021812,
 0.3126286723672752,0.5280962738640528,0.6320323806913021,0.2977308082168104,
 0.8948968305642597,0.2693895961795596,0.8064648415100734,0.8870394050330057,
 0.143185223582861,0.33641137304802926,0.3577698808955948,0.18238144698431413,
 0.17318167684221697,0.4510523212455303,0.24928074616844154,
 0.47811963718948347,0.39734785768935943,0.24027433071039694,
 0.3768655658322053,0.35073860915572885,0.6022159994510258,0.1920997919147326,
 0.5371073117543279,0.1835559797366828,0.4762469290437955,0.2921373435164213,
 0.671786030182729,0.9863297117615787,0.38621938951586166,0.8807332558325293,
 0.16216442013698562,0.6264342217892291,0.372175149755839,0.4727659498899133,
 0.10214938345406943,0.030791998023113276,0.29878404278643744,
 0.5527099919277875,0.007646124387230557,0.22686094258185785,
 0.3237708820666656,0.266705090570884,0.8022045391421292,0.7496803586217188,
 0.6893637262826648,0.14172993136112033,0.032513170644072664,
 0.22599275532933594,8.414177557491964e-4,0.28998531507250325,
 0.09174729652457347,0.2933991815330077,0.6704494900920208,0.401746938224234,
 0.09165815683789502,0.20935760438561146,0.7336749928117751,
 0.13762746462565878,0.050238240622341945,0.1420135602971916,
 0.6456417941277705,0.8450570036890046,0.2573798354246274,0.3708803024031011,
 0.18354586210198642,0.6525535114874521,0.6611503483264417,0.8394102705773483,
 0.35923157769021974,0.6903962048704656,0.3859835459902431,0.5975984398888093,
 0.6051858418832002,0.572984573791337,0.0490508419416803,0.47408667463769893,
 0.7946613592767272,0.5325539051631389,0.19017740380457993,0.6057478976685238,
 0.8480983229915561,0.7482212973501063,0.6211252346180316,0.8571117385764164,
 0.2732941112483881,0.730033603706986,0.15639417878301887,0.13113179619146575,
 0.9985213041764088,0.1277229106694049,0.3213278015306391,0.4129723007210264,
 0.26449372429621354,0.6813451107563252,0.41482562696171643,0.54146801330848,
 0.4497967951803473,0.19920776603321722,0.314323000745916,0.8086605528506412,
 0.11964745163157997,0.5360110791506649,0.7276312602158631,0.2229303038590621,
 0.2745644069545583,0.7930675673062282,0.1993833764409233,0.8181514621113004,
 0.8594861738043138,0.9237831428817952,0.2670722250791814,0.6788773089489177,
 0.5687582690351485,0.7487995705047914,0.11327977597396188,0.4664176619482282,
 0.27479128978921874,0.2472692898917921,0.7289105563721631,
 0.15246497467331277,0.8618983740656568,0.5615981278432008,0.6097276342906641,
 0.626320496315691,0.9877026314108264,0.6591468585280049,0.17126330953350988,
 0.9358836829830819,0.5584174336892974,0.8715663952387668,0.08876345457422452,
 0.28449698047167005,0.3042121609060343,0.13908081294440544,
 0.4202797760780643,0.3789103994597045,0.2249583733444731,0.48315142292048097,
 0.21495950312626788,0.18413587886421479,0.40413132327596035,
 0.25235380931668094,0.9958893237160257,0.7226860295532658,0.8948869974967145,
 0.7729979194723289,0.8921598358031446,0.45159457431267935,0.7812928869165996,
 0.8627106316742456,0.9567922465861883,0.5859215037227593,0.6648192016675033,
 0.2976850017057736,0.8611539527678819,0.43725094434882494,0.7323034382215315,
 0.584148942731111,0.02013642293352491,0.8470968788291829,0.14885943099098842,
 0.30437486057892527,0.2696950291001501,0.12209931687377029,
 0.7253129508522717,0.08675632475551476,0.9898451406297827,0.2929684815154361,
 0.9820892087152102,0.43746363916769604,0.7097478003853915,0.7038086361848378,
 0.7258710014594458,0.6227903879676151,0.5476727194158071,0.7567796478939405,
 0.9589823415017815,0.6540841470564497,0.4405364400908304,0.01834135680360205,
 0.7835158169217552,0.26789638910734737,0.8907801284774003,0.5903746264657925,
 0.659187949054776,0.990150193914431,0.9505108983662155,0.6925668214926316,
 0.06109688494593324,0.3788499580214585,0.489469771733521,0.6196468843556251,
 0.4487703032293452,0.17158850783154245,0.3334054568648793,
 0.26030750777968836,0.3196032887440141,0.2586824982833168,0.9546300037024482,
 0.33684110538943646,0.05752941227687569,0.3482367782146122,
 0.8835364371115629,0.028964177582055717,0.7242396271464693,
 0.6893845362554393,0.4406125279827027,0.423752928485133,0.5290929339446369,
 0.7354443405356774,0.3863748395598119,0.7953527017583947,0.8345477217119562,
 0.1432928888042556,0.44923358682888626,0.6837307255725479,0.0734816608362091,
 0.9278708030805438,0.19956312684997568,0.9735736450904523,
 0.15573820455065845,0.25740851340020665,0.13721144859243173,
 0.8525549002821681,0.45870201871344707,0.1813489674865395,0.3141138532965434,
 0.8804408407320208,0.1146476692675833,0.2590276888764097,0.4909765711694388,
 0.6764359818024792,0.19713520584424726,0.3536162210934132,0.7350232904898143,
 0.421885995158954,0.22162813972057693,0.23555928313137842,0.7956502348984924,
 0.8768786691003551,0.02772666040138483,0.04365834209710151,
 0.6343656580980616,0.9049864650625472,0.3368266771427535,0.2609723758048109,
 0.9877076981272686,0.5776630505656244,0.058982206303843765,
 0.6110304965275478,0.010563411564908476,0.4258552529642178,
 0.6400761162163063,0.42069771764137087,0.03448065776278075,
 0.33164676647668734,0.9680075296310219,0.6653471661316027,0.2577924459880283,
 0.8701718212946372,0.5358986437636831,0.14706073273114995,0.5174694596570614,
 0.624250985681709,0.8610930307980061,0.7568336110571839,0.4999021009885948,
 0.5885281494218457,0.011540875349546176,0.9159777629907434,
 0.3430538463618922,0.5538882873845352,0.10444998769668057,0.6268118716128688,
 0.5260469679120084,0.8441265964478991,0.5648606661883672,0.8777562839922679,
 0.13025128205079572,0.33897974571336653,0.9731291341800735,
 0.9330976712198759,0.947397395604742,0.9739614245315573,0.3315965544168322,
 0.5084482370959786,0.7867545661554918,0.7560879083568389,0.7602051542286675,
 0.1756095528488021,0.15811354644205677,0.6319557930816807,0.590977576820148,
 0.28584525676688144,0.36118117708100284,0.7477262800831501,
 0.0500360073633086,0.6818071562841009,0.35407501039287503,
 0.06778850652517687,0.4059515314631762,0.3145148217824457,0.9392361415653039,
 0.4485438041002947,0.497427376108288,0.16140424687185362,0.8494804756721032,
 0.04999413918596207,0.46010766461046726,0.14868673030131774,
 0.8010297933974524,0.40418881171338383,0.3699947695711784,0.3949892578875783,
 0.5510462341169355,0.3562988413359658,0.07300614596148614,0.8534783750098479,
 0.22228347225444178,0.11412994336742144,0.29998728070372715,
 0.27977761909269705,0.5128224213762788,0.580331869563199,0.41642089388114023,
 0.28813660082792014,0.17648306333399144,0.7224309660501215,
 0.4583197077393377,0.024230925150121485,0.8768431111023423,
 0.19337825807441233,0.19654515895493918,0.14891104450070802,
 0.46333877891021646,0.2928492407504356,0.49012006433079924,
 0.9268266894471207,0.9919403493925033,0.9801041677850387,0.9870202654379219,
 0.824998130408146,0.9957421166498595,0.7290278593188713,0.9624405843848493,
 0.21835882900551573,0.046722855811535124,0.7210036674511361,
 0.4016699275734563,0.4202447305885739,0.7307137829563355,0.9790525609623085,
 0.3934301933112755,0.34002899290262945,0.9536927371831334,
 0.06350837091956973,0.1717025087965076,0.7011988535372443,0.6506048976843872,
 0.8858920672366979,0.15350806435556474,0.085967816136997,0.16990378065806477,
 0.9398477129207665,0.7588340980352225,0.42669347045572603,0.5375185271270799,
 0.754719395073259,0.4802043143721624,0.6696094512471364,0.39523505900105493,
 0.8913573978620555,0.7627035328831838,0.8894222886802011,0.7516573623212854,
 0.9608213951955505,0.17214713510575796,0.5647566098029774,
 0.06688493754927105,0.7925278488105131,0.5773670139652606,0.5078261468534164,
 0.47022691286515195,0.3373302304736261,0.14658802076766264,0.917639794349643,
 0.00617686695476849,0.8273430821737886,0.43305894206769824,
 0.27814060315483335,0.5665260224223839,0.09057998937297762,
 0.3693917892067894,0.3724839301723517,0.44040107567391884,0.4479306725258181,
 0.3023382910645205,0.361528806731205,0.852934570669203,0.6993084685974483,
 0.21251887286989068,0.1388771052607759,0.3918400173375898,
 0.18659200238732432,0.6256199683717508,0.7195068852217598,0.6385849414204183,
 0.24032649267348494,0.3024533653433569,0.568958129883063,0.9501789339759905,
 0.7251711675046244,0.051490565935664145,0.4920923555210047,
 0.6505935362664736,0.6559692808522172,0.9416746728903397,0.8578001111251619,
 0.545778826580352,0.4737959364228883,0.4140851321724869,0.7690118772419633,
 0.9188777526157572,0.9507501007498933,0.900505418844376,0.7338316615004369,
 0.16717601887128886,0.6196953548668913,0.9862017413120517,0.8505826865847841,
 0.8267450939589905,0.08019007536684608,0.1630233826775741,
 0.43100289913169854,0.39378886248745815,0.9796092108665257,
 0.05746015846965258,0.4881217192495402,0.30992908330277047,
 0.08149331865217102,0.5080975719245939,0.866350491999081,0.12730902353169715,
 0.047309386964947575,0.05358750457618855,0.7618402884553847,
 0.6058242943739773,0.4070714008539049,0.14064204376542655,
 0.015808108162372325,0.729471405955779,0.29209779109982326,
 0.17823509233203239,0.29257541676417353,0.7677535604898393,
 0.08236722660670137,0.5018855576226817,0.5172430637744251,
 0.20616900337619593,0.3704428995406386,0.7413688437865491,0.8488448594523689,
 0.23125994687521545,0.01995349379698308,0.17532216409127854,
 0.3142126003112532,0.4711270693020656,0.8337057222597473,0.2142504919366488,
 0.07024956929081116,0.8443921046393607,0.7288786768477834,0.5034777751457081,
 0.11886137381481299,0.22509243174013394,0.9619840970894609,0.762701595445695,
 0.9581148950454625,0.30892488397067974,0.24920525922857062,0.402688974674424,
 0.6126166888444672,0.7693254910816685,0.6703805530387921,0.6380357500859007,
 0.7595656241875322,0.8571191264885567,0.21037571562343504,
 0.13919198970168445,0.5467836924364038,0.06475867881330633,
 0.35014107716622767,0.7623494249100757,0.2414871078332783,0.1803430445934825,
 0.5436401806377797,0.7728098532707522,0.6320114919154185,0.38332916599060274,
 0.4643069090541202,0.7547125061510971,0.02154830129028351,0.7270617884438041,
 0.4727254087705254,0.7408547375374316,0.02831585761508193,
 0.11817648778461987,0.4749394208262684,0.44100584633629514,
 0.38552097258307294,0.5382796760947417,0.49990539680651,0.07415465647779829,
 0.03472640020731266,0.8457655866012511,0.27787469452059044,0.704111152827226,
 0.3744146912379881,0.32644916004067603,0.3077898730834898,
 0.13880888129971647,0.5225519803904337,0.5550655629303797,0.4331941725497429,
 0.2199643722249034,0.54225515075256,0.8115279773197419,0.38194474232099496,
 0.2513062499346288,0.2185270878757689,0.09338849709270403,0.9395098076829719,
 0.8305183816524704,0.9828034464040949,0.03868899396198566,0.9604342978507947,
 0.7850569096829564,0.08555359672138818,0.12014807188841115,
 0.19432802830662066,0.8069672742324127,0.805701748095971,0.372445982306744,
 0.2720077282087122,0.9204399478439498,0.4725447592932911,0.14345821799800118,
 0.06315492692373215,0.5277761188164314,0.9987580032728384,0.9285782529759831,
 0.4486829181391716,0.8104667470593854,0.7866684253958801,0.9131583600293134,
 0.6231400978702492,0.2682679670154886,0.9094544892651889,0.2273630167275037,
 0.3785032368588792,0.7321151004167332,0.2893729510847429,0.963953320305058,
 0.0020684264136954233,0.3265146639937664,0.7643310347638098,
 0.43320593312625333,0.9573579395624324,0.9456092929227645,0.4042314231258005,
 0.38393957899183806,0.5715010256419041,0.07676827643300532,
 0.9463981265972073,0.41989206160443293,0.01244309679495359,
 0.21557702117602018,0.6983651956785989,0.5084083857631594,0.9362007680863811,
 0.47933331453577466,0.020749014912083252,0.3942355405062309,
 0.48424546398279844,0.9572739415132474,0.6073070199393965,0.846699088032771,
 0.7544619437691668,0.09587591722231237,0.6741453918207182,
 0.16107822494792878,0.8682682371457776,0.935044705223893,0.17242815555445068,
 0.5825752768957518,0.45511593752818336,0.22571554265619498,
 0.004562122738737573,0.5011962542266507,0.709080930253008,0.5452865481729539,
 0.5363913989230534,0.8193074407037694,0.1487999900158894,0.294374319902991,
 0.13765779694068214,0.9085746928859653,0.7179209989240217,0.6264540014711795,
 0.15234702407104272,0.7155785037596981,0.5556113835368546,0.8426769869188399,
 0.9830406772755244,0.30651752267051857,0.6264590251826215,0.7527187009601811,
 0.8007815502893126,0.760974807282959,0.2626640009452065,0.13741192930373325,
 0.3425529094360724,0.45148251522437943,0.9327803617445737,0.7585724454434802,
 0.013607217981898678,0.6176110880888938,0.9663532935191314,
 0.4085101005719536,0.48549250801428634,0.500193284873686,0.8067123809005317,
 0.22238784161428282,0.009310723521750663,0.33365835239273356,
 0.636555750874128,0.7263698612613291,0.6607456096491308,0.273976004471456,
 0.8018040620020519,0.4413876324250585,0.7068296907455293,0.4703543168456743,
 0.10633144360309454,0.21741221898886176,0.1618679935895062,
 0.8877708775793164,0.3500705839699654,0.7770783719215821,0.04420661533141368,
 0.09384916269879928,0.31742402279301685,0.8232858603347126,
 0.9641325980291061,0.974749574387327,0.47361741771073784,0.20263420037997482,
 0.60696710838127,0.6443967713468761,0.3305306589527103,0.6504590094058194,
 0.4046500326622966,0.2375361698503169,0.7013834411021878,0.7661204659407325,
 0.8872955591807885,0.09529675184601327,0.14549019472420166,
 0.5463222347836781,0.16693518738457125,0.6762026134688792,
 0.17999461128352012,0.21666158724376594,0.027929196566915282,
 0.11490228147089376,0.5331437291994634,0.0772952111131544,0.8587732555093263,
 0.5424445295452014,0.8744838088925604,0.49762211997339767,0.5332585287132162,
 0.4545111673608585,0.29438255639822675,0.47109807838041196,
 0.3113637793587338,0.13876394060830977,0.40690204729191126,
 0.3417196948681963,0.009725435181471509,0.976144059008415,
 0.24034727022364644,0.051607805833533416,0.5026797668027524,
 0.5567294682551123,0.9038349255528046,0.3152423759624857,0.4194974528852915,
 0.3762630096232972,0.1405079629206113,0.44045682917412143,0.7097300895310623,
 0.016417320655427114,0.04366150406620195,0.5519230234075466,
 0.8385636060968636,0.8684247082218555,0.5040459509206145,0.37526294082087164,
 0.4898565782151191,0.10954972123939832,0.8861042217938041,0.8228023031792152,
 0.44754924015600095,0.5299681312402504,0.06530171081074043,
 0.7638036392255612,0.5705621118109419,0.4130571402005494,0.8070293002664563,
 0.7372858680533384,0.8645516009067383,0.40575036456318214,
 0.08080904150695856,0.5088977860638728,0.8879177731256824,0.9625740434649921,
 0.4497412856094799,0.053540410606867206,0.049916859062902,
 0.07305039301928062,0.01549675122122407,0.4319281837253569,
 0.7380217638085995,0.28076721472113453,0.7339639226862985,0.5980293089019513,
 0.05797647167775177,0.8645655537747532,0.8929855303890254,0.6919469617721595,
 0.7416635499637467,0.6935475906869835,0.2696417629566268,0.41612099265000546,
 0.5974861565798874,0.46792952687053435,0.008711626239008452,
 0.06372736901917553,0.12785126319148143,0.443751266215462,
 0.15347712216900655,0.29458487821637314,0.9039655719464734,
 0.26638223440255315,0.26280675821024,0.3064733136633402,0.6838372152512349,
 0.7350862554360225,0.3637869472624996,0.8033844969225387,0.632315080888238,
 0.6633698175135913,0.9452721242645126,0.19208809310143626,0.059166864113671,
 0.4276064268098848,0.5572754193936954,0.06583968420231479,0.6626895716867897,
 0.20448840547641478,0.9086394530264957,0.7750339093821752,0.8501299286921574,
 0.5736848728884919,0.9130988743937555,0.43244619844505594,0.6266604415060086,
 0.094206897275805,0.55314488941136,0.10796500663149933,0.7559086867924537,
 0.28071683615696896,0.9148511501964683,0.0521443819174624,
 0.24764937506414275,0.3989138031435361,0.356522731608354,0.6284575701233479,
 0.1752435336148166,0.8557573038069798,0.8054142674634678,0.11264789113965512,
 0.7050100259163179,0.7618053804424999,0.901377806578742,0.8764449881184433,
 0.8970809970708702,0.9874950429455749,0.5511381951449833,0.8495536229405588,
 0.9028109968695517,0.20174451602416543,0.5124784663632505,0.175385057268179,
 0.994688968741241,0.4245876668673686,0.7376687645120579,0.3336859828116494,
 0.0743197682586878,0.24048178501556272,0.06232127634358442,
 0.8313304546918435,0.5106839677306924,0.6739504652802498,0.08902362564791588,
 0.3204495477942957,0.5836319361536291,0.40779859048146905,0.5385634422195327,
 0.21770305257333988,0.38714041225538054,0.9157112446169718,
 0.3702795731833519,0.570443194656813,0.0914089145143715,0.9689403962989432,
 0.799292264062862,0.4302785759491137,0.1432777823232676,0.9731211053021654,
 0.18428474206321166,0.49428160239879726,0.38203945583278565,
 0.06138654933468535,0.9888147812874635,0.19658180861461927,
 0.9750914656546519,0.43256133668995367,0.5861941696377441,
 0.11867855243816339,0.8477434731280312,0.007361132215917721,
 0.3477070769913502,0.005856738211341139,0.9026588222144976,
 0.1883493517352478,0.3407725594254414,0.3249602408652581,0.10432764508129644,
 0.39775796930484475,0.9211404580930401,0.8866270738309899,0.1325515143504682,
 0.7407943093453436,0.6450642432446464,0.5205818690634652,0.9128178873561833,
 0.5536465904047236,0.9567567208597139,0.5570541828627718,0.12743138941152055,
 0.6999456052413136,0.8010121330305485,0.1734402134075076,0.44776255027639245,
 0.19427219339034996,0.5326421671693707,0.2088835416166125,0.9791226380100522,
 0.7435311342844466,0.3533777392954115,0.7981079541151106,0.5642848741821044,
 0.6342613547119895,0.7809911193078986,0.29468632526611827,0.5235243069488685,
 0.826370982210839,0.6863061116982696,0.21868573591686546,
 0.0012386280517311388,0.6375640826075604,0.9893379516560177,
 0.4936924748654731,0.8831357688427278,0.5820174550139181,0.4098517775806012,
 0.6143007221236335,0.624613945816269,0.46423598407697386,0.6487605796813238,
 0.1304364279531406,0.396926187211328,0.8001078373624986,0.6529107576924169,
 0.5554888623588483,0.07247480085326719,0.3962112183219276,0.1323313673927473,
 0.165681633314689,0.9564434596363834,0.07366198851665762,0.11508275661619605,
 0.7083803526248306,0.423360188652676,0.09626207472322823,0.5849552683333016,
 0.6771874229721616,0.08041970296604528,0.567316942507711,0.739898797849983,
 0.8767066838246504,0.9692314388198346,0.07649192580187125,0.9033630947139356,
 0.6507327882469502,0.9018103948837045,0.5742396684850676,0.09161655213821973,
 0.15543529114724652,0.1783411082194445,0.9575130086854057,0.2938227948927987,
 0.7709462438046538,0.7589251076418722,0.7272772449701177,0.972652411574672,
 0.30561732653174767,0.5028663402100749,0.6946407666600871,
 0.19335615313893317,0.5147897478382267,0.25456494535388974,
 0.22311370320764068,0.6390075193980764,0.8427609886663507,
 0.17743997410956613,0.827099561979927,0.3863371853435347,0.9698572500111314,
 0.19023379876973057,0.35226594956737534,0.20759657410901167,
 0.4578901879609395,0.5091675890249,0.3122480301962156,0.9859011437272512,
 0.45830564174971755,0.5486809712357851,0.536396326275593,0.7065584100122569,
 0.9777423380249394,0.0745844991801422,0.07023280373932472,
 0.36987618352935225,0.2653257414319494,0.012309829907518166,
 0.631975162125814,0.11270884200841835,0.9769384283511755,0.47782232511944234,
 0.7908285493788326,0.8927039585638877,0.3402636980029521,0.4688792779587164,
 0.5018996681728681,0.327988647074553,0.5502286860170057,0.38487708482117633,
 0.48212861697036136,0.34705445123103484,0.08351629984332876,
 0.5178067315402434,0.9378800194434243,0.28392708528521104,0.7961506247433341,
 0.3903144876339658,0.6884661002627795,0.7442815210133772,0.23382373338247264,
 0.2106582795178732,0.5516019882111244,0.23759076344230667,
 0.19377694763442754,0.24364408260066972,0.942426600107481,0.6618207082865545,
 0.4317693872168623,0.29345574503885097,0.5509438989000571,0.4377687006208293,
 0.4606349544772226,0.740059984572925,0.9228708696770963,0.21724323243840893,
 0.45028305883147157,0.10449210518151997,0.09535508995958408,
 0.33171766602733754,0.01440705654868546,0.49508390627655463,
 0.2541118999732329,0.8839008508212113,0.3616851613493841,0.3220425486007663,
 0.8447244649105555,0.5725865718267382,0.15674943965656774,
 0.49863834865423096,0.2425911690128547,0.5863000935383661,0.7186717335409107,
 0.3446951253701912,0.1949013491509023,0.39769674101500063,0.2697819771911045,
 0.6869969561770514,0.4566194547090636,0.19196040564007955,0.5836216979283478,
 0.20201338432424654,0.03401807235708654,0.14536979991917665,
 0.4457083419581517,0.7121399573549302,0.7050091542244288,0.42356780474533073,
 0.5555420205135102,0.3242538924461392,0.25363818815635064,0.9173126381661956,
 0.58172761321345,0.5911478947845797,0.4049568813956983,0.09030729663054504,
 0.20398808433963878,0.07360038079777864,0.9319009180240921,
 0.8141434981183822,0.04832917910811685,0.8839995352864454,0.5555847088886064,
 0.6390568663007605,0.09764877095825708,0.5283067483755294,0.6252768977455215,
 0.3630993376483521,0.33364769761550994,0.8658331324692928,0.7051488161898934,
 0.4285459255175704,0.7123232576962241,0.9858307275721346,0.5001571311167194,
 0.6363016862914013,0.29580937690924936,0.48410977780618747,
 0.3731671740261294,0.8720777846235319,0.010857035002775683,
 0.9403911962828352,0.49000167621981383,0.44174586885906675,
 0.2476910499007262,0.31031413334636904,0.5821964960953072,0.7418604733659864,
 0.7925961177189284,0.8953595269653255,0.13488110733422476,
 0.015410282429341282,0.4528348043885907,0.33812591377642465,
 0.3305791863342281,0.13899609845264893,0.25491690934421385,
 0.6141827020723675,0.731093226025257,0.37884583553490503,0.5085384954185517,
 0.27422294728295027,0.2090934435974875,0.53462371808417,0.16247649956512056,
 0.9538000835016139,0.5769041105940086,0.5692953668611882,0.5707595896792516,
 0.6439207679696817,0.43626723130781453,0.39521857072282773,
 0.33165821613022184,0.5553563857306831,0.7080205586775065,0.6431684414458435,
 0.8172923450875391,0.7830146235377528,0.04776300915839904,0.6299279748400335,
 0.7874681250450236,0.35749022177850864,0.799813984641178,0.3978818472410739,
 0.65866466830412,0.7658060611473735,0.10758700538653931,0.7186188069134714,
 0.9821448184977741,0.4363457582760607,0.40391950448545133,0.6840883651328561,
 0.11109441040132173,0.45335161725883477,0.43934924831248956,
 0.3665415231022484,0.5956549575030378,0.6735852002776092,0.7057396118711402,
 0.9084196367696167,0.9280409759212986,0.5763892069533442,0.570435040263278,
 0.07897821939904381,0.16822250825980223,0.28506599010763656,
 0.436886265218013,0.03027709220662722,0.6109039697354482,0.10302835947722988,
 0.6570683618308479,0.696615661454609,0.7802576708708824,0.07559816738313951,
 0.4931496414777422,0.1080313889251231,0.35982573629876047,
 0.26218751277484365,0.22169736602258083,0.016400116671901666,
 0.8265721237450729,0.12629655379077154,0.5631745521586549,
 0.41259967688008603,0.5526380388991697,0.028988109791957983,
 0.3958914711285023,0.8590128098869071,0.007974349983508322,0.611032304059165,
 0.5959636167148097,0.7637797633021175,0.8746837565166831,0.18778701394970243,
 0.9657093213494634,0.7090138263924468,0.7306731158653452,0.9573188951205462,
 0.1872725795616924,0.5952414653201906,0.41620178689930765,
 0.06197018992381198,0.12864979274723132,0.2740888236051039,
 0.9773114240607048,0.18606256255770193,0.055788050075087914,
 0.6404443147178815,0.06360548542688826,0.9792712180830061,
 0.0010188053002658792,0.6234408280041637,0.4012188647820103,
 0.008966232977091093,0.8761906696597564,0.5547397682866709,
 0.12777984322135172,0.19135055643354126,0.8428698550520741,0.771316312500064,
 0.7281715958867774,0.7566225770030669,0.8681402592010803,0.09066581812242003,
 0.9887543804754251,0.8724180873475454,0.5863861772977574,0.6321612558814054,
 0.7418779958537871,0.3395139799211727,0.5815747687380761,0.09669736722179223,
 0.08204896590264021,0.32583036539032073,0.425371101568932,0.5448695370582135,
 0.2464289854416568,0.23754902270790879,0.6099857103131091,
 0.40564400523758815,0.169571982116398,0.8034091850880187,0.4061304198600244,
 0.34493581901760584,0.8834096113522502,0.2810374235027664,
 0.19033596269598452,0.5205517824752063,0.5504727828101094,0.620424872530619,
 0.28350717046318896,0.8010989610785794,0.6489376594078747,0.2965894139009202,
 0.7674068559601712,0.2609175478888991,0.007610312702882482,
 0.31014631159147543,0.492755631365015,0.7511912825552236,0.6270850283113653,
 0.3686021236540913,0.4030918631368047,0.2912143991219608,0.24837957119816856,
 0.788625597794716,0.7633045557957835,0.625399963847057,0.4036420172084444,
 0.9712488138745732,0.08967622825208643,0.8908713234654776,0.6465595066201463,
 0.7633428188445336,0.1959448026036319,0.4647898077710414,
 0.036368040755328734,0.22477905543941323,0.908222097823981,
 0.2569844085489751,0.7050781257342225,0.7806694403932335,0.7099423410976611,
 0.7161974377505939,0.19380372452804573,0.3354560888071191,0.6428777649824774,
 0.2751037841384478,0.41215266153814134,0.011073587030824728,
 0.37203132191411337,0.8934978258509636,0.23576565668811478,
 0.9887463737900305,0.6271151089392645,0.7689417842416021,0.8997017920173863,
 0.8138671925544744,0.4435029196520791,0.39153795795899926,0.3465564348926723,
 0.890664080984914,0.7910881798057393,0.40291170899424245,0.9025296787114647,
 0.9313648929175384,0.2229192463460481,0.7065382193681962,0.2845192813691195,
 0.26986436857311324,0.4080382963267053,0.5165278747116245,0.2822039450238032,
 0.8458886140675478,0.9276471114516683,0.8392865827447915,0.6716469856754681,
 0.2374484320123461,0.6404576155657045,0.3323922853903254,0.5967982529622075,
 0.45190853386325835,0.6505157074734782,0.8535217695768843,
 0.23744943307172306,0.13343539644537228,0.8395289670982111,
 0.30653029741396354,0.4821132042170496,0.7557479470810938,
 0.020630927094403728,0.8484976644017639,0.5941572216098946,
 0.7769928031141555,0.47821091401396787,0.8164936738838624,0.5141695086584385,
 0.6800123813823744,0.1086404860249367,0.4219455418675171,0.7375429058624084,
 0.13065165831082637,0.5000216803118408,0.9721174920564761,0.8732641940175168,
 0.9984973258341998,0.02333045984549631,0.4032436357708138,
 0.05361880028532218,0.8806123955975786,0.05614194166314124,
 0.022956184413231773,0.38007355037963975,0.14373347884089582,
 0.6285994347863562,0.48635736821867515,0.7716338623267407,0.8352908610883856,
 0.07185538657234203,0.4615578271013867,0.19804476333286514,
 0.8507502693355251,0.6491471266840969,0.032974043683824394,
 0.32787285433491853,0.4942372930244595,0.4645396974803536,0.7173336934109212,
 0.6346976569773999,0.6766902349128396,0.572561853900312,0.44347987477417206,
 0.8267297731887883,0.3381375196559916,0.30658666314662486,0.5646111104800926,
 0.8873765621387676,0.3507835018872637,0.6583662612864601,0.9044891344388081,
 0.6768542555641047,0.06345331511356256,0.31526691970750864,
 0.8074264881452472,0.6452825376939052,0.1757210396313239,
 0.022139450158961016,0.23004563919317011,0.4987196593176765,
 0.7575464647383352,0.3564483553143861,0.7556557171034974,0.1290466172316902,
 0.050422723566674654,0.9579012341471438,0.5556440991985112,
 0.6873802133081263,0.8696527547945402,0.11777283566988217,0.5552236419796457,
 0.9658793965158311,0.8133263673143275,0.28435493043384696,
 0.11254377395314541,0.9809139692275816,0.4781303834743468,0.2929956383882367,
 0.23361695932713156,0.8253522943313535,0.42433653287657336,
 0.6579950068881911,0.9687782920509678,0.6576922612152667,0.5622477665738256,
 0.7122345363271307,0.6659388381109427,0.7457759866575969,0.9775528447342082,
 0.5188217116591639,0.33787258253760777,0.458859597265486,0.8753084047307418,
 0.9973372447322952,0.043776614650658296,0.5786454329928519,
 0.33184349917716915,0.5167133218114,0.521334235311947,0.29608171596370736,
 0.8757996918222776,0.6958856231199323,0.6403124323271698,0.09079960764733974,
 0.20869400790201198,0.5241248060218273,0.6941132991383833,0.3682868138401325,
 0.7980913984238139,0.9413426834834118,0.9421162519364418,0.46307409130682453,
 0.8491612834937305,0.11708182411763413,0.2343638099695864,0.5491854531844224,
 0.3278124405585585,0.6677627808982658,0.40996721992053364,0.6542945191794586,
 0.9290074303147531,0.6014978007637113,0.9016067876032221,0.7055520234468479,
 0.43435165620093785,0.6150714517488629,0.8114174554725048,0.8877454733058139,
 0.6840631605282745,0.34201114254136067,0.044702166105409114,
 0.23902988762425958,0.17553109390082755,0.8384767523113269,
 0.7010021204210056,0.17388683479293987,0.274850248732591,0.2029719715264211,
 0.8267253496497645,0.020265067579276264,0.8998702396714897,
 0.09690467284950055,0.3759530795255581,0.4599662235469657,
 0.46050886496790433,0.21636080819969405,0.45330310007591135,
 0.14492792791161435,0.7729995225244572,0.11953143911230346,0.772845305072889,
 0.4266486875477902,0.029605264819815735,0.6772737311326816,
 0.06952857699214054,0.2410728030233591,0.4026769150764862,0.5267399508158865,
 0.8418278774862721,0.08205959032058396,0.5682391210101484,0.4438535126744886,
 0.5942525082389852,0.9308999522526318,0.8513204788928554,0.02579717972186768,
 0.057232461176653815,0.15803202129669658,0.5179676656346259,
 0.17579295038518794,0.8709010598640938,0.8783732971928541,0.38790891234434,
 0.9472454712601985,0.2193634506029557,0.8232485990063695,0.7415056145767606,
 0.25681743999627704,0.2703308053229345,0.8540309331251703,
 0.27884470795012806,0.8678239663391506,0.2492725649518014,
 0.15944595230834402,0.5804831781640034,0.6601810354986707,
 0.023804976767158093,0.5176696566590123,0.7214059780844364,
 0.3047678274032205,0.2601217755290527,0.5939393712982896,0.18243819591323518,
 0.3170938408006593,0.9352469567070464,0.935062655693216,0.41694278545655317,
 0.8304154386732209,0.8711689371337026,0.6396934448258307,0.5172256366708029,
 0.4988097987223249,0.8723150159599237,0.5434053788903346,0.3250266407416965,
 0.3777394044482977,0.4801995747255592,0.8843722930145557,0.9464802785894247,
 0.6909525646155736,0.1079331788598914,0.9431172884205409,
 0.057565570513189446,0.8423708950296898,0.9790384243192912,
 0.02810014865529853,0.08157965928146338,0.4136542749407188,
 0.22427526270861275,0.5234445515431324,0.19362259662228143,0.624694186317827,
 0.2408441917360663,0.8804392700575396,0.9923882132811996,0.6614981422082596,
 0.09577625115612054,0.049576790508134794,0.1557289042930572,
 0.814258141858632,0.37671370027390405,0.6472328256319193,0.3578866059793162,
 0.8338824234946357,0.23001458710057887,0.298804683773802,0.409941053713066,
 0.21077610534237,0.7162837620920349,0.1943496492489869,0.9621228548976111,
 0.42452090805717646,0.6890425740901905,0.13240804985746546,
 0.30563591936970746,0.6030755270845154,0.7256770095622428,
 0.49388165368534687,0.804349112375403,0.5469886006847102,0.17010373388358346,
 0.08187287625656814,0.37055762227310407,0.29482975031889247,
 0.5642946789965554,0.2666161212605279,0.23702233029323783,0.8977463282966718,
 0.46845633893672867,0.37317902938210823,0.36938721319218815,
 0.7506474626206304,0.9214739224014276,0.1395066205177491,0.0747207203573812,
 0.6837780871718011,0.7013915745124257,0.6147512028320421,0.5231998054736243,
 0.5504178611629833,0.8622339077672984,0.6506244099364861,0.8504408103669847,
 0.17244299267413588,0.5548267520090464,0.7369362114036498,0.3937933896715875,
 0.4514922328917097,0.4479751461369441,0.4193180173971225,0.45388942254009723,
 0.835306438332747,0.928794815912826,0.47591346224592024,0.6869282971927332,
 0.3507720824704853,0.15462874389880665,0.930184309815492,0.0906955697872961,
 0.9062195403831511,0.47639750909711454,0.5614119537979586,0.3661208253135938,
 0.9809513589285771,0.6955973937424097,3.815619987412866e-6,
 0.7397910973998475,0.8234615731948346,0.6944163051428899,0.26722219081970355,
 0.4724084968054294,0.2240171822644753,0.00794714999004642,
 0.03838583989474631,0.15001241551098476,0.7927339531575879,
 0.3155273233463669,0.8412508531355686,0.12026527482573424,0.4409052105134561,
 0.3393150499551859,0.2474035488204419,0.20760268276176996,0.6122368162624376,
 0.45727329072219736,0.6650979163156012,0.8816482919510354,
 0.36195322960733356,0.2585709350066264,0.8092681929083044,0.8532798051308915,
 0.3290873883156624,0.7579330003248161,0.20583986628856543,0.5870777753225033,
 0.8791526867914086,0.7768014582339476,0.03209644203743123,0.8444736645052955,
 0.1869517590457409,0.5295570602228599,0.6447167433479851,0.34272252456425645,
 0.5701515973455114,0.2640495358686339,0.8467499909876326,0.30549090626507325,
 0.3923072792288578,0.8935111404750455,0.0267080990826567,0.12478247061591519,
 0.9144362352700589,0.04384407566918791,0.4141189278034262,
 0.27210424028013236,0.5519913147918482,0.6677850119831431,0.8378296294208271,
 0.20061239360586858,0.8678003755240097,0.850378285022613,0.15800605725998618,
 0.45505914679855575,0.6781040109350648,0.23411919728887975,
 0.2722663272902299,0.03368532803249502,0.4529490267766575,
 0.43792559756404315,0.9404478898286123,0.8504565360654028,0.7773923874598774,
 0.22448706776586724,0.5425113196741523,0.6403941269850648,0.8542948875600349,
 0.42176455858913653,0.20295975533980104,0.768631116292068,0.2782428665544202,
 0.17506414671622583,0.7214087246689556,0.536372609485539,0.4244285517036215,
 0.854035248701671,0.08771283814068209,0.3468478788026914,0.5456008668236728,
 0.2416343689048729,0.8245987710109868,0.5257729649264159,0.568612757475079,
 0.5498434428157905,0.22102788959826847,0.11558889253635818,0.325631929004109,
 0.7065975618555995,0.858819086246144,0.7123418098271548,0.033463838859471906,
 0.5206821207883443,0.14481995117405755,0.5778121469413372,0.7482137002312239,
 0.3934013211175019,0.7997218681528268,0.503224786553744,0.9848171544305675,
 0.07711742387836928,0.9792124024072075,0.6307705990468219,0.2740976616791069,
 0.4446664853911412,0.2702514430084886,0.5941961504557909,0.21300166932211728,
 0.9053628248202771,0.91884756554289,0.9401450402662664,0.7338810662380834,
 0.7378170672231236,0.29444560422979693,0.23635613715201087,
 0.5055556041453559,0.6299345423260165,0.983653321027874,0.7896768717889942,
 0.3459097408906826,0.24074173782297137,0.07563495959817546,
 0.2798637656277465,0.3951530264623464,0.9743642382123999,0.1624138843235462,
 0.7116049863744135,0.5872952018757015,0.4330773792966741,0.2390505024263152,
 0.6861822035910027,0.7690493499593413,0.49106304718776383,
 0.39610975467940235,0.08575273660138194,0.08877977564545525,
 0.6735205779967849,0.4280672973209757,0.1591344508569601,0.322161797167262,
 0.3343801115835272,0.03395049461444333,0.5745033071512614,0.9045203607071264,
 0.24034384708160794,0.33495728535806746,0.35155946492078594,
 0.9014450668224273,0.09103244336415195,0.43352825923638183,
 0.49930897227022986,0.025158890776394327,0.8613015759418352,
 0.08357002269098845,0.9198217975141764,0.7321535668996092,0.6833321725520006,
 0.862921602383044,0.5232847409142715,0.9201634946849738,0.24983667007363541,
 0.09610161221377123,0.5851168566982461,0.019412864612241454,
 0.1434812601148504,0.9081478380079009,0.23238086936466495,0.7796701274499606,
 0.34830523699983895,0.5572340248101509,0.6185043028042247,0.7554089801926128,
 0.9354397026919112,0.22779629537820079,0.8607935882332807,0.8086487489905235,
 0.3945609369259305,0.09971921745910073,0.5584080479096433,
 0.055886178722388324,0.6139338575636573,0.9576499496136518,
 0.05827379727362936,0.7516168831705843,0.8032864328296898,0.544963847361125,
 0.023343514982997582,0.12005971754319056,0.379559745652729,
 0.7527835391727816,0.7746884318736834,0.6167507627141324,0.12842089540631674,
 0.9625440093138125,0.6652117284175698,0.5986840773132099,0.2383990642645889,
 0.21605210060935864,0.5491817316854969,0.5148156451734223,0.0764375338496659,
 0.5524929091194091,0.14274038590224702,0.17636161195289934,
 0.5945810220320267,0.9792826591400974,0.625772536738811,0.8805044709763046,
 0.7935931209544799,0.9610483968159551,0.4589441415012322,0.6788979706915678,
 0.39615314251419564,0.23117712226063059,0.9158942561286014,
 0.9433886343672148,0.8215694602263846,0.8212282386676603,0.8461494445751158,
 0.4736109395125875,0.8922074421848473,0.7215349013538791,0.6160353341402428,
 0.0663696629494347,0.4171522101101892,0.033168263641095264,
 0.20791936219948526,0.012087018918351289,0.7945418273621464,
 0.9872790397001815,0.8681244579576578,0.10552403445409597,
 0.03209775611540466,0.2270276505781339,0.5069769039103047,0.978680071857756,
 0.959607179654422,0.4914480742275634,0.7857939209769,0.5754701888870121,
 0.8055056569364941,0.9442149488710023,0.3515271776067833,0.3239808571894137,
 0.034953471782472434,0.20306803706346277,0.7883608291715067,
 0.5947127650965002,0.662646818975834,0.12071163604860013,0.7379816821038478,
 0.9803206100848822,0.6653081233886502,0.39562911225779385,
 0.020013073855678742,0.8659625002762018,0.9869709000741056,
 0.4739012798520691,0.10698629914121371,0.1290022068531147,0.5647589452686853,
 0.17734408831970694,0.3357997688877812,0.9808725066110539,0.9141097623793221,
 0.523405118872898,0.853500205834015,0.360116617782146,0.9238069620677837,
 0.5140823340056438,0.0509219252456935,0.7464020535626893,0.8835596508820082,
 0.7111575083898696,0.7161117924233791,0.28165768467131025,0.2993746447060728,
 0.21965201888553354,0.3628077293459874,0.11094831757210044,
 0.6548599961013181,0.3091778071976947,0.3437268964995256,0.6392366293869594,
 0.9972318083554278,0.6794299167984553,0.9435558563087484,0.6248142636964534,
 0.22797355926911134,0.8801540729282048,0.004985362891708256,
 0.03577163320401733,0.7791742018977831,0.6220497309079672,
 0.49194869199895275,0.8954080332561554,0.009616349135410429,
 0.11294558277040623,0.42161629243598187,0.7840092790318075,
 0.9292128960378956,0.14256519234810672,0.6646452278092627,
 0.40228161022057407,0.40180964930899743,0.0547641298403162,
 0.3788139817430015,0.4176042056362975,0.6747155334764513,0.3151767808730781,
 0.5681237931476114,0.02556004371231635,0.618702486404157,0.38105627526669794,
 0.6807359003541612,0.0721372803546797,0.054980079274567606,
 0.4548371742917128,0.7021125978457223,0.7729770465197027,0.22822703449040194,
 0.424326757510589,0.9488271048180678,0.6639196797865028,0.42503059928702003,
 0.8072767949223619,0.7976001771063772,0.9666281992498639,0.39805644511883875,
 0.6498574055755205,0.7717902628863788,0.41204208598529934,0.9115515109413344,
 0.7814271736580193,0.619295423187067,0.2135876004281826,0.6276887009226176,
 0.27853304229330766,0.7661091529117184,0.8913586698105307,0.5240893744354107,
 0.7013789087915118,0.004841684037061533,0.06886012349810233,
 0.7504353262903526,0.9915088808617155,0.04743178295309547,0.9386361459825934,
 0.5498305191206856,0.6141685513916771,0.6544151029992606,0.41109974341537914,
 0.5249636753211312,0.9150961713762112,0.497865575197188,0.3072962694120064,
 0.7980949525015384,0.6977156137705585,0.27010424709218117,0.7392928860211962,
 0.016704897527572538,0.3348892232989623,0.9775358617654768,
 0.2833155657072366,0.439460315459845,0.8017243820455552,0.5760621352004591,
 0.7999345595717082,0.10648784905173647,0.6224753189618271,0.2660381568871113,
 0.516967604456664,0.9039730792090047,0.8635367890243679,0.09060798258266756,
 0.8270603288640095,0.8430446849612435,0.8607968447842946,0.9133464661811092,
 0.06284595172638541,0.4053296717445565,0.9431868441435093,
 0.07125450933571154,0.640185194075294,0.023401826789107738,
 0.09632975797525278,0.7680205209510096,0.5537232956611363,
 0.29435141715639523,0.43641872361834033,0.12033502317863409,
 0.701710647160809,0.28158167163227177,0.17048314417079258,
 0.04002366505786226,0.940087034007654,0.2471420652867049,0.6725102159624193,
 0.8208479900927824,0.026090611974014855,0.610512117883788,
 0.17887478059218753,0.4580358338302022,0.2266313881656924,0.7770696825300352,
 0.7008094030237617,0.020810421495844555,0.012026436429780296,
 0.7664114515473265,0.44343995746459064,0.698757416913394,
 0.012832720320463231,0.9682461996895336,0.06447177822862582,0.8248833961625,
 0.8968589647870192,0.8347017067823506,0.9896568914619058,0.3207552347235014,
 0.9003131635631647,0.9034748159686825,0.8973057117658599,0.990208768677385,
 0.7662377051451988,0.14085912911466303,0.7145126930112844,0.140973118322401,
 0.6227539509245833,0.22742902777009466,0.7102027712945216,0.3660329832075552,
 0.8283194168132193,0.7926763337331144,0.9879617949551085,0.7148464798894294,
 0.21371824524255745,0.619687904989671,0.3735302379718066,0.6797584722545255,
 0.8065422821220583,0.5144533188615233,0.1535872764988051,0.3594490935297854,
 0.9593932442252111,0.20706657703577835,0.18615358051335684,0.360692477959079,
 0.6022207362266637,0.07333440466397978,0.3935272286636776,0.9583462556570841,
 0.3640285298771724,0.07146308121019862,0.6424275529975896,0.9366586635374766,
 0.10505130385244588,0.6811301488788312,0.2207382108941962,0.7712593617682055,
 0.3654836553513102,0.278070141679552,0.40617158466134673,0.3187175426533193,
 0.45565180499538294,0.6986458883506765,0.4015863072575241,
 0.09404191254243122,0.958538247286814,0.5462928361831096,0.24211149336347182,
 0.2570913970798939,0.16847876470793532,0.7054990044025555,0.6025322389993855,
 0.7542416764536796,0.9782254132576721,0.1487668148484259,0.4542467357378176,
 0.05652986606256971,0.5452536809743203,0.5824482988173045,0.9167711771218268,
 0.3235422171100747,0.4030270847090346,0.5737772317151452,0.25253542569805765,
 0.06775185086497226,0.25531756152475216,0.19036002722091372,
 0.10307321602192276,0.7345365979403928,0.4594385342840296,0.490508242572571,
 0.7019131105853346,0.25702446955000124,0.7829269595199182,
 0.14672662540245174,0.7569609355361628,0.6378031634229225,0.5404609474759466,
 0.3231710397416673,0.3143136617870149,0.5580065312036325,0.5007008273749006,
 0.4527835101512301,0.692385978297724,0.4754906242836454,0.1895804780898288,
 0.10529079449254841,0.7363225777206601,0.7954282047918001,0.7320553018896053,
 0.8422918440891478,0.3986709451495508,0.35604134290019496,
 0.32617631315204454,0.03404758853490586,0.6537345606253013,
 0.8645886953650741,0.6493556627566928,0.7736909162603826,0.26901834373642863,
 0.6088473028779535,0.5511235138518608,0.6500752270682824,0.8575421479915398,
 0.6461802497357956,0.4728937568561883,0.48298796761532414,0.6368264333129244,
 0.02524689984499251,0.5636949647708128,0.3495371578245483,
 0.06789524158977733,0.7332927900798666,0.23645471654773775,
 0.2848505100411307,0.9919832544880235,0.6952261224261674,0.5044735074684591,
 0.3234355219793563,0.6454889383599345,0.631429323067577,0.9814433012965921,
 0.892581795863501,0.7448854132187667,0.6330545424934737,0.7804683675672265,
 0.7640135380261573,0.218183284987577,0.1025089592496039,0.13223088292302454,
 0.7172444795627138,0.5838019603441885,0.1245496480360867,0.8553075406477223,
 0.4047168395979742,0.4088424400597608,0.47494699646339356,0.0470388014641927,
 0.49083300439398725,0.893727551607203,0.9168240678677265,0.6020718134453005,
 0.537428163845655,0.08685407807466627,0.130444943714205,0.7456644011110881,
 0.405668615606124,0.6012663107791325,0.016165062863441326,
 0.09122875394275765,0.14605895550815462,0.028440909741262788,
 0.7161043846528644,0.12777592297747797,0.13753212030591955,
 0.7692617856371504,0.6420439618595903,0.24201681999061186,0.465582199422238,
 0.34755349683958736,0.4876377004476735,0.4368251181729109,0.4997772125432206,
 0.7355331806854972,0.9225964443900022,0.8426253995511814,0.329772031983789,
 0.10921052563956601,0.7525279859904102,0.3233457238326275,0.5855073575531884,
 0.06401584951554673,0.24742050586710684,0.2605552380197166,
 0.6906273195572693,0.903817058524319,0.32321062654241184,
 0.046241025945509606,0.5712738232281842,0.009345333189896499,
 0.7160516157921583,0.40546144086371294,0.9953636905041652,0.9436630402685485,
 0.3690157471522547,0.6996397233760429,0.07714617942251578,
 0.43425572724613004,0.44055145920292005,0.23828424965978967,
 0.981072544388004,0.6230899300915491,0.7525767489566262,0.16731991492175124,
 0.5456271270426166,0.7552734056957202,0.04934739519414533,0.9123709026292021,
 0.548317960470023,0.9337152912396953,0.9238186780103782,0.08796753855310813,
 0.3623468689356373,0.7634748168206797,0.7538503979079478,0.7343568479560207,
 0.5096296279188336,0.45601676040563976,0.9917256399270032,0.4754238856533791,
 0.112265120682986,0.9617580304899193,0.1270911950331668,0.9192442626320168,
 0.2783807282285533,0.6962022922021096,0.3217168317650442,0.5766344684473939,
 0.9790918638907218,0.5673760725814968,0.5481205853765343,0.21663206485064102,
 0.11875807164287722,0.6550175739270776,0.6193790843634794,
 0.45674592662498137,0.14115448365938577,0.8746957511001178,
 0.9101858036690371,0.6176840972095893,0.49153202568214427,0.5283119724126238,
 0.6376926091949624,0.17780251301480698,0.32558467693817705,
 0.7661917795688529,0.6825505172986304,0.8192638116307549,0.2592794374865146,
 0.5984872250418923,0.26922346613937487,0.04892776957042533,
 0.9155011304760368,0.17863888135227768,0.4418908319642374,
 0.32120906427304163,0.8848935425437401,0.8255244854046178,
 0.12133742075237353,0.8512994306155971,0.22253936748275382,
 0.10133031704761986,0.9622818322259491,0.7838942719383043,
 0.21582200276102326,0.7338677650545068,0.8041920392687485,
 0.10049374558992585,0.4950370117645335,0.9871256093702638,0.9189918638249455,
 0.027649748378851036,0.06665637444535688,0.6706977941823383,
 0.005438613770234024,0.12421546982495524,0.8067624050787726,
 0.24688604197582498,0.792402223488853,0.6041482013384225,0.373570783585051,
 0.36994878860471914,0.16472550938375397,0.25389880484308325,
 0.32795426480985257,0.3890994261650448,0.7370540986004805,0.3072322148619411,
 0.9639985315860112,0.13486006199553746,0.677264075581137,0.6068167030293612,
 0.17467506518181086,0.019192158022558692,0.034827026238718584,
 0.12426470190001249,0.5151050033586042,0.20278677630724384,
 0.9010471861238017,0.7134901213425063,0.49361498526960335,
 0.30734034981322056,0.9100310595050218,0.9707849274126045,0.7145324917069493,
 0.3710157608872485,0.4106805053229621,0.8637082153429173,0.5340790428366935,
 0.3307227473727482,0.21170469829983074,0.44582737724343424,
 0.8956580758867597,0.013360819084384468,0.27239219401437476,
 0.8540259442700296,0.9489999087950404,0.14898686616547552,0.469584670330061,
 0.1719998858570272,0.6176840437583477,0.11790917364167153,0.8503503678878638,
 0.22875344647376594,0.11023894316911353,0.4016859560794237,
 0.9054074323902094,0.3742440308919559,0.13272963329091447,0.7220274809482021,
 0.03049080612601862,0.3763149699795154,0.8892729192646232,0.7597025617560078,
 0.7154133507193194,0.03628404556065368,0.9090423568319831,
 0.34281732636256823,0.9005119367105262,0.8654644658450868,0.8915282344889288,
 0.5365846598210875,0.9206171942990923,0.1601286540029404,0.9328971762354382,
 0.07929566429898716,0.27948188569887433,0.05591469893516965,
 0.46332824218076407,0.40882188630829563,0.7494317736384516,
 0.5787979617717964,0.8856908742665603,0.24877430154194002,
 0.10802323324113594,0.012698502850345017,0.8895563139220432,
 0.4718214871719475,0.9059356296089575,0.8971307169761815,0.9110257479284796,
 0.4254183525869111,0.22156734920674626,0.3714828947436495,0.7787518287332618,
 0.024940758388860673,0.7028484444014951,0.11609055131727974,
 0.530320150847917,0.9428823304608629,0.5298236517898732,0.6030093462306892,
 0.33824460541042767,0.4420206838098518,0.674407589784068,0.5330478250767191,
 0.5137998993430548,0.9057452970760633,0.7305079963758951,0.3982895090458709,
 0.9801963500470894,0.4952507490149177,0.9334479289026221,0.21324288722761897,
 0.6221189356206007,0.11750574418818471,0.24239809067031814,
 0.03992918897835118,0.9408899454964293,0.6220522812033619,0.601556103700398,
 0.5342501586983256,0.7863268921103038,0.4307891490910766,0.9673356109154186,
 0.4980826992255415,0.7678869221204757,0.10899907339939874,0.5751017837726834,
 0.3467236276767104,0.04037540987386157,0.46765680672484355,
 0.5297456720799909,0.9525576938891837,0.34197662021925523,
 0.26101647911519565,0.9729482572314008,0.49894967129889434,
 0.7805051139817789,0.3358703040408526,0.6560049954700391,0.35352222264646027,
 0.1983166734186551,0.7421108404856522,0.1784537776538908,0.2779366484829586,
 0.23439977796403977,0.8667183031524588,0.3648285720868403,0.939700486985323,
 0.6409864146608701,0.6740006645311367,0.8216570120462344,0.4699638060177763,
 0.132361100396849,0.11532924849414583,0.848516352413935,0.8517469630875599,
 0.14565242304606718,0.3781896501282027,0.37053453692711447,
 0.5252370255968473,0.6822774737683158,0.8809919785444497,0.3862878753068274,
 0.23638472010573608,0.6072209692653481,0.03181409811734248,0.300860651659538,
 0.9365843119867893,0.08775512321893075,0.43068964397682297,
 0.14117844948305158,0.22784674165240992,0.3200309533778335,
 0.6630405879254516,0.1288166074704732,0.4712360741925027,0.863413823336341,
 0.39751355770200836,0.8392083275442734,0.5742051016693944,0.3931911331533837,
 0.4895267038784563,0.9730499033653868,0.23793705037895552,0.522995415132299,
 0.849632595539679,0.8586991780380475,0.1235430442698815,0.8642998438524856,
 0.19158203077759262,0.8510933865086315,0.3088721800784685,0.5800590907614722,
 0.3172043202972007,0.2456665244829519,0.8721910994346452,0.5813388411593522,
 0.6115660387716522,0.19367625748938988,0.6459545874678068,0.1452332540195489,
 0.9819650367225204,0.045539592400620554,0.050946750043008304,
 0.3886244180771127,0.4665583797952715,0.6442992540699325,0.9683745987168112,
 0.27536633782973785,0.2860195058171704,0.5952641624849989,
 0.43859791578145957,0.5599815417834535,0.9050385479512013,0.5986313366661392,
 0.021589479690412583,0.6197843281882868,0.4020224097505589,
 0.6194643691681667,0.8862924335344518,0.63449716592888,0.4281098721978326,
 0.677207536538854,0.5872810606248939,0.5177346778203271,0.5414295237553796,
 0.08913304814915479,0.33511650513863644,0.569078407215075,0.272798891395154,
 0.7577355620721894,0.8932977179856506,0.2192786340768471,0.4735886636750023,
 0.27020267756733674,0.7962269782939418,0.37968422222122433,
 0.9662389249717075,0.5242562181451744,0.18093641820467699,0.8949756286280443,
 0.8883019330305553,0.5736901888803193,0.9767286567725002,0.5024701258753121,
 0.7840243591860214,0.03545605123450568,0.8578438333440893,0.7425342906094342,
 0.8534580943396572,0.7658520529735897,0.09758105770759284,0.9606084993803736,
 0.7166410416279135,0.5909411210756916,0.6410131775492732,0.5913339237456721,
 0.9817309663569818,0.8648150901600065,0.40766278578218196,0.5864687864406702,
 0.8087637162454372,0.892654821905392,0.2168077204862161,0.7081497567769728,
 0.9376147477777623,0.7924683664140367,0.9328521038269652,0.6449967444520806,
 0.5331845614276867,0.6556072627643645,0.046400676499488536,
 0.4463242353964647,0.08374507836502598,0.4386849360709185,0.3767366001885991,
 0.9933235603213135,0.3309281845213071,0.472817266841002,0.19237085303537083,
 0.2900983838296316,0.10735796546505572,0.06945088427511348,
 0.7136627632967121,0.7088535802618331,0.17239568812465522,0.502246504954514,
 0.7811599872789715,0.9547829403205232,0.7793653914491568,0.29812277550489863,
 0.22194872196671367,0.8925839724997313,0.8404313486751999,0.6164571617187964,
 0.2585758706304164,0.22665326989762324,0.4632417284016669,0.4369886616661729,
 0.9394456509300196,0.165410115917355,0.5026753351314995,0.32822659123978326,
 0.6521364588951389,0.5114161587866046,0.9791757308315284,0.28384604263432334,
 0.7478200722406388,0.7445962644691345,0.627736102101528,0.17125954803631782,
 0.6215208118044819,0.1795466196761576,0.9129349635220189,0.43327675534498244,
 0.32737670472883185,0.4660980341136245,0.1366031829736114,0.8207479601550958,
 0.7627974920871955,0.21083692606625937,0.1371567427081677,0.7879901350493601,
 0.9620141244747282,0.47750831596716126,0.3131715279490599,0.9216085235646226,
 0.9567056128153157,0.7592096384812501,0.48071482998610615,0.6465277384844427,
 0.7602783908486892,0.45306673965801636,0.19715135093294278,
 0.04117196232561704,0.2156817698771175,0.612325335680255,0.6349759778651184,
 0.4783005520392636,0.27172239826508837,0.3564807252305484,0.5777353376816526,
 0.7070821344880984,0.3193165397813872,0.5814512414475118,0.5621753125078965,
 0.6310994341263029,0.10430004420994521,0.9536913918962857,0.7511988063701225,
 0.42136724656063995,0.48289399559976065,0.12665701099694693,
 0.1668362034535007,0.0070579291360003404,0.4577524667801778,
 0.2616838162807297,0.5413962781791767,0.3475707148118595,0.869637364488525,
 0.6101553870841668,0.8453363158499546,0.3691907587959389,0.19936072757626044,
 0.7124451621164853,0.482916381782336,0.21395726197683007,0.8464979710346364,
 0.8961792335317951,0.04150705843531377,0.45655732191858167,
 0.4026646480265895,0.36302997915422985,0.6276090902239096,0.5342398181925023,
 0.6057000106196806,0.5414288066682191,0.20543888314851078,0.4026596523194367,
 0.24991261698955036,0.601144482531053,0.7107580865403755,0.8393919105991371,
 0.27080084871338805,0.6414488044037499,0.12180821802520991,
 0.24235707822029595,0.8002998839982377,0.9751834879316373,0.9308931835662313,
 0.636103402172463,0.4601829979228056,0.14006647383696424,0.6288033919256439,
 0.05649033799388625,0.10243114590414604,0.26496296416289056,
 0.8052214740998478,0.07084304988950652,0.49687438103782366,
 3.339016791777949e-4,0.4274999137377554,0.06952985679643497,
 2.8327438736708288e-5,0.9977039526175784,0.7570887113198453,
 0.641940916568181,0.5656459693407866,0.5610017234686224,0.3185389088143036,
 0.976275591917714,0.5718234554894888,0.7517183713808908,0.62156784729068,
 0.5620645166190046,0.5017479054725901,0.2761622114684965,0.9445577459675087,
 0.21427462503826367,0.3614814291039692,0.9291133518750772,
 0.06386030745520299,0.9158037899241233,0.19977588677243685,
 0.7475040231838889,0.9018015917915028,0.35141151911877133,0.443208395678509,
 0.07368223707632193,0.1648861815082262,0.992553396125649,0.6425986423031628,
 0.22872177182157494,0.04361399814885947,0.27993207908474815,
 0.6309360757032594,0.8899078944651986,0.49739190011339895,0.6239292184610664,
 0.4510118428598622,0.9805299227083811,0.7823107920111425,0.7642487950106305,
 0.11899087098001171,0.13677353989142138,0.7369703241516401,
 0.3232344201719962,0.6625361844913648,0.9019762296726331,0.28330958562602915,
 0.6532915462306291,0.8172023509439313,0.8876331670557617,0.8802912808170267,
 0.23037475027330334,0.5779232130897614,0.9749660964169099,0.7102258288018388,
 0.23276796511119235,0.9742759010421383,0.4658382342276135,
 0.28277445985282235,0.3694692972182865,0.2868793727222959,
 5.415657039808952e-5,0.2864348341787435,0.2982114488311197,
 0.15256935304228625,0.0317019246092074,0.2967028115283725,0.7664426967374538,
 0.6956111072461413,0.8299578912642337,0.44077939828716506,0.8098248724370481,
 0.6692504998953799,0.9279317966422306,0.03449219411103632,
 0.19073693811402492,0.539215576658168,0.7020318893548106,0.3244834112097793,
 0.6722966093018411,0.8273150281353766,0.8380112371647865,0.5507124678490043,
 0.13094583827619755,0.6327357067848172,0.21594634232651955,
 0.9792268790255817,0.1889657221134401,0.6773567193168137,0.05028202797115722,
 0.6908389112585068,0.7218515251861124,0.022278599143263245,
 0.5085721346284457,0.02230621936845978,0.04638731514439831,
 0.4633923502175422,0.45264564580921574,0.5358693375770103,0.6543216954724708,
 0.06729456593265093,0.5420406664694609,0.1950907725523101,0.2785015978090434,
 0.9491291490218333,0.7819707673181068,0.07800377777251977,0.23704632308757,
 0.5546220735465445,0.3843429714252231,0.6856300909135676,0.32102042931268326,
 0.5322207987081206,0.05646141099230939,0.6349952583526748,0.4419018879033698,
 0.9409939627988522,0.9188373385959663,0.385147112585999,0.3910712954607408,
 0.533928668354805,0.24089692683660202,0.8152486305054022,0.5823047227914293,
 0.3357303594038784,0.6789773298827092,0.0513422647571975,0.20842829812374541,
 0.6963874605814913,0.1546710599638933,0.7044717498158274,0.6908256270721583,
 0.039605080649963265,0.7415546002721354,0.5089766168337937,
 0.31099475547496525,0.33399449599786757,0.3276405952208803,
 0.5567848023960443,0.5826598236846977,0.4553296790408772,0.6781268695124346,
 0.39161470109248153,0.4800162550906458,0.47605927758133215,0.736226537942041,
 0.6549108759939626,0.811127677456443,0.5123406248759066,0.5112466568595191,
 0.24185346315496314,0.13846376907915492,0.8389911856061736,
 0.09041510376976192,0.6386887191132365,0.3089399573919511,
 0.34221119438393466,0.243718678770936,0.5973028235022645,0.7869363173650774,
 0.1633720697861426,0.7148747243494734,0.9117875396298117,0.4171415208235938,
 0.8521269296434454,0.0405331825970926,0.191839112377576,0.5755278738189459,
 0.9582470500337122,0.4933715573712708,0.8363870573315371,0.15021114311174,
 0.4997504637698552,0.9350338696651999,0.6925228112897194,0.31568748818801484,
 0.4992655072216333,0.10106592167113804,0.08379446584104788,
 0.3892457239329965,0.5744680572289527,0.7939420959701741,0.7674343105703091,
 0.6688890748429133,0.8473155357400047,0.249115460818925,0.524437405503407,
 0.704161530578672,0.3674298345524366,0.8747676988807038,0.1211281801275701,
 0.3988430473151303,0.770971208291862,0.7632338207763776,0.3433036036548536,
 0.4272749990124608,0.9963146811355859,0.549350173237873,0.9571557934497579,
 0.7007670255384474,0.10135213020968548,0.0575036580968985,0.3584113344167945,
 0.9758519850077988,0.6577288726456865,0.7595101380894611,0.3976741119019399,
 0.1602426525647418,0.6721787562349102,0.1490064789278902,0.9177675133422865,
 0.7999307328233725,0.08271655615967066,0.09012049322437499,
 0.37618321966866153,0.7992467494245387,0.03981614179268411,
 0.12344857498319772,0.4086324154202151,0.2026619194216559,
 0.26652046401503304,0.33497920932695324,0.9439572349473153,
 0.46615265095984826,0.9878019365978782,0.6930617012933297,0.6034988919983518,
 0.699613863618149,0.4773663623148515,0.14259419455319156,0.913570194468647,
 0.7661085940264469,0.4226190001184812,0.32773113488695615,
 0.08072640681437693,0.7999978190321089,0.2962140649441861,0.7522935307101521,
 0.7210935233475644,0.1853104099595433,0.46179266235864846,0.7705997231198364,
 0.833542831550866,0.7488364369146764,0.6726074471236755,0.27955506074514114,
 0.9789962724769145,0.140857653246924,0.608386517343384,0.5866825097387711,
 0.7330227351203843,0.7714738250824905,0.5295755986415245,0.7748457150213973,
 0.8111750542951007,0.004506102821279168,0.6519109550718092,
 0.3947505555571469,0.5684456138401542,0.7612909143204745,0.11403233271104718,
 0.0707908215700278,0.9695101065424928,0.4018365997103559,0.8057088177927259,
 0.14209163298388972,0.7711843383020289,0.03509937600107016,
 0.9768336466504334,0.9450851195197327,0.9196348884402273,0.49701942362617635,
 0.3200162189928437,0.13539326142963248,0.5055428411090691,0.9676303778672793,
 0.14867768012002214,0.6744159235058573,0.23371894211748068,
 0.6995531714365152,0.20225512515752764,0.1875009773571591,0.6492645156019694,
 0.16576460378777746,0.93160685225977,0.2351845333004784,0.9465769893374061,
 0.7232304353330409,0.42785830151273596,0.7426217713904799,0.5405437538870059,
 0.27737396738463116,0.3434499840464279,0.047337464418277686,
 0.10069651452000494,0.8700677263832911,0.5540150195159569,0.7803304142929023,
 0.48438447551225927,0.06698971804983178,0.15689291500826053,
 0.3232009917287937,0.19307769859394552,0.17775575173407498,
 0.3237614337954715,0.6956346320392008,0.7089041665891336,0.6151650240426524,
 0.588781706693311,0.6700725916319098,0.43291261945967363,0.42012359573873737,
 0.23989455796090764,0.037806101900312084,0.7022506860124562,
 0.5044828388003249,0.8910683986405403,0.26279289539738726,0.7099195712101039,
 0.45902526216110406,0.15826224256579202,0.21807022720709235,
 0.6847021565806546,0.4805562804431839,0.6070660123331656,0.809976361501735,
 0.22215053270242913,0.33004018887174913,0.6827264432253136,0.586202874499308,
 0.47835022180479025,0.2002515491055662,0.8863582576950163,0.8657560256867654,
 0.4088681676463509,0.23666338905345408,0.06962350599250311,
 0.3739406130815828,0.5215812458991929,0.07175068259885653,0.267841319533227,
 0.6327934286224359,0.9006020043811591,0.2502208147941535,0.6425443058757444,
 0.19323569066302593,0.8172923618921046,0.40889826610578894,0.678023150788244,
 0.5140832877081601,0.7029383166716624,0.8241684329329482,0.5139855529179275,
 0.12094067930252883,0.7145429921907025,0.7224992193419726,0.3153091474384412,
 0.2628849959039353,0.9351973339599078,0.7563470371418756,0.04649983052641482,
 0.013143991781733533,0.8360045709184912,0.6442547316566049,
 0.5745164339125293,0.1365499032365607,0.27100540191891775,0.5109368464396876,
 0.745775834618326,0.3297808095687236,0.3280405041465787,0.2884814913723245,
 0.7475681659220346,0.8240433658083213,0.78482087270862,0.13191358813635556,
 0.4153861537293886,0.032575776102815146,0.9352518841781798,
 0.07340821140059484,0.030021086001435937,0.005660564735695095,
 0.3465440849931679,0.863910599173049,0.3799240693005633,0.021676813086331737,
 0.2940772255012114,0.6524226185520245,0.6236903769662773,0.5826341608211385,
 0.9522227124589997,0.08969374257198626,0.7566366033148388,0.2654265234138332,
 0.1271170862653367,0.18284411319705807,0.9705626503199472,0.2639764522202589,
 0.49198603375897576,0.4981724470719908,0.19253982529471092,
 0.7462924747931107,0.9470831383616547,0.12345604045155101,0.8316912254572159,
 0.6461079664183002,0.06061740949717087,0.25737513078367347,
 0.06563982561084214,0.9128742608858083,0.38877953614185157,
 0.6742120945704563,0.6978822835666322,0.804132175840382,0.6456615648203939,
 0.19287911848823436,0.6118333190063029,0.216691201420126,0.2596163142654131,
 0.4866137318494176,0.6014568926875765,0.13429019170791912,0.9044027633422216,
 0.016842147844024935,0.5039228974567812,0.5394582054526113,
 0.28848661421420685,0.36355345198370337,0.14651454443578926,
 0.4053768046174673,0.42438366372395275,0.38319644895703187,
 0.0793710632366531,0.8750144359558925,0.08553370289420725,0.0335273140227349,
 0.8507675379047247,0.4799948423199395,0.8636133343301963,0.8657670954688571,
 0.29201704450122934,0.5440701729757089,0.11696190497130476,
 0.7121504598804033,0.7175890509551911,0.9772570069599558,0.1345196680327575,
 0.0793620513084613,0.49710137729461423,0.5385551179859885,0.8760795647206487,
 0.34342089909614537,0.8033607691419817,0.023019224478289524,
 0.09594084639480327,0.17331431248405627,0.5940552680466032,
 0.8043709228589826,0.5862913927034604,0.9812221262475062,0.42309306038956773,
 0.23023145251741017,0.30175331381825,0.5026036022616602,0.8177476154567367,
 0.9653753341766457,0.6743925488074034,0.7735141591505992,0.8910288378593751,
 0.46173547043019847,0.389697014922042,0.5284428654273212,0.9994371287853784,
 0.3270932855117693,0.525431085047576,0.4219618747628423,0.5808270105023022,
 0.5512835912894297,0.38093887510299007,0.632629236589001,0.7194302179803715,
 0.20607527396833014,0.22978860618528907,0.47465663112685674,
 0.4350990290522623,0.5960188601227152,0.47748079881242045,
 0.22102042498862218,0.70474577460691,0.6557414926605193,0.05240098655020431,
 0.6706324959028928,0.25680595595585354,0.44256787006948706,
 0.5642303229502281,0.7133959518724594,0.43546066869870204,0.44547018495725,
 0.8502592216778379,0.2281815168764567,0.2951595724320648,0.5182131915733708,
 0.7248080497865609,0.17504184391515099,0.9383151617017872,0.3795362769334317,
 0.10382067091427671,0.5360507846844741,0.4379119676433738,
 0.18115195222301106,0.23604177514310187,0.3537690640808957,
 0.044772718928681954,0.38591523949357565,0.9004539137016238,
 0.05144858071513336,0.44322797391557467,0.44968679307636783,
 0.532447613909057,0.064829379973387,0.988888279951192,0.21845709200335162,
 0.6346382600291646,0.9240468236320157,0.5417919145397866,0.6854975474640219,
 0.6955178829562817,0.9189185264282724,0.6454809366222447,0.655806452612489,
 0.08310969790245548,0.7660663956599348,0.7557013229618504,
 0.28499366511404345,0.43957731978409387,0.8020631000914227,
 0.25898555114425914,0.7120846626213833,0.4224535783520288,0.3396031643966265,
 0.4343443546190564,0.8509793798660259,0.6184696916026357,0.17013188631847054,
 0.8096731231660952,0.042241056731562754,0.6340763056796377,
 0.6547154266194041,0.4412391792895608,0.5016473612709076,0.21923962182778822,
 0.017524526187324585,0.40021057613403954,0.9523513094100267,
 0.33341820837036895,0.4311935389396664,0.23380855995781458,
 0.9701608621048743,0.14186812804559246,0.18606542495475753,0.818070565965732,
 0.7969301790107576,0.659299641630191,0.54356186845803,0.31781190123900704,
 0.16772846764346505,0.7759686440385686,0.21547658146094373,
 0.13412110973659874,0.34461721911134113,0.011407349840048764,
 0.025104909150538313,0.22160682030311973,0.7784494479000785,
 0.12160758041130804,0.6008672654253969,0.42804965918919347,
 0.6946372391617807,0.9904838100025524,0.039584180947889736,
 0.4134259007609956,0.39396947601130605,0.862471197510687,0.364115183575648,
 0.1589275675971027,0.19449878097087425,0.4599894730551388,0.1885449437863359,
 0.8485861092710597,0.5760821116405119,0.8878828794631992,0.8104607128462571,
 0.8389541768872054,0.790845484896529,0.10491837864967979,0.7710393867002093,
 0.6282671419670962,0.4458250517686565,0.6355325365421776,0.948309602714412,
 0.7622404635812511,0.044243635349481836,0.2197299979883962,
 0.014368740604944641,0.06988052535719214,0.01338048680904591,
 0.8181445039063069,0.910234583298147,0.4125961236870509,0.9640527238321742,
 0.4821136679415301,0.6438922908752372,0.45857512664043376,
 0.11259649739420807,0.1658716653395771,0.37890757203948744,
 0.2169686512802469,0.28482406100934643,0.24230752828172508,
 0.7377606541239343,0.3556119768709378,0.8864332383773608,0.857529969825881,
 0.70080111989157,0.46231928858769,0.7537115969787397,0.49804550223806066,
 0.3891276588412236,0.5446626411044853,0.23870930026369375,0.5270709602855539,
 0.7990267788441168,0.587488353574984,0.3175693634405524,0.5617356513782161,
 0.6152554226320358,0.8971901401391577,0.18734349978935882,
 0.23751574602269798,0.7773386794536521,0.12836917204217058,
 0.8326935535116177,0.8183256160165969,0.7334604410294381,0.23825141098602542,
 0.570895935240286,0.191517331154469,0.31604675351261025,0.568214353063362,
 0.2998678202489663,0.2120355602422883,0.03370243133129969,0.7506984561236751,
 0.8951563044750662,0.8497611799733025,9.645849238083537e-4,0.522576877682495,
 0.5061197730012535,0.3779979793540371,0.2443392658130531,0.9781763554405095,
 0.6399142135455008,0.020621412171504705,0.3063595984524583,
 0.5362010532107514,0.9856403496362948,0.07513304224518103,0.5663202030340166,
 0.6615763820224955,0.05321909792402624,0.67479816129455,0.1904602081130521,
 0.13921145996630846,0.6893500925345152,0.7676324564713902,
 0.11018383591734304,0.6615082298213576,0.9079751955771362,
 0.03754024573226422,0.4958546007365905,0.2032409232880994,
 0.35027391429057597,0.6067843912628469,0.7667586874020252,
 0.017790736038818977,0.6593678351403321,0.5080148094363325,
 0.8930295027436678,0.8370026092997883,0.5637613098601008,0.7728080040461379,
 0.6675756418850608,0.38716485788378885,0.5908882811408454,
 0.29147061684287856,0.06365879543816388,0.9521474791575151,
 0.8486774327281954,0.4975214401242394,0.06353650313725723,0.5976851144460726,
 0.2521648501675908,0.30042638015695,0.4610902146369449,0.35426642107728457,
 0.3193846071030646,0.45763525751072365,0.9556007119519278,
 0.09830035794741254,0.7716669981097598,0.19780730249705292,0.618024119787951,
 0.3287995172697554,0.6019357522770685,0.23977734669367978,
 0.019953098984836992,0.7831534151444607,0.5822404909844133,
 0.4327898015234607,0.7468669780275545,0.3140797470605329,0.1369331761995911,
 0.22642004771661206,0.8333288715928159,0.018788102083436842,
 0.8916145821736329,0.9234577207765536,0.6190527525377877,0.08501062495435319,
 0.13680744194904093,0.13576524521109912,0.75448351698202,0.8219879728092403,
 0.4020977812210266,0.7987701602898062,0.034177199370055145,
 0.46670249617394743,0.9530176117669209,0.8960029156092408,0.6064188470306555,
 0.6592731869181997,0.8485256216017404,0.8185084235549249,0.9755823276879239,
 0.12734292165640004,0.6084511595182226,0.27582489039595004,
 0.2220629741653728,0.5681547241977679,0.667003776517787,0.24499438693380027,
 0.7334670547309865,0.14767877397515794,0.9406311551186051,0.2695719343016234,
 0.5243622175764472,0.6684070656083939,0.2056220064939377,0.6201070983240283,
 0.19888498415820643,0.9039514977580663,0.5269929602027776,
 0.42349858690244746,0.6993708508120557,0.890927816037329,0.3212524425373898,
 0.9858313082888595,0.7977992289414415,0.3924143071935583,0.08809748720729271,
 0.6867020420760966,0.40632124658067614,0.022471326350986587,
 0.04058253651785926,0.6954992809201858,0.279924355876251,0.1897891140627188,
 0.5213147439040309,0.7106121384863495,0.4605068094649387,0.96405060367676,
 0.1666318338814715,0.9321296776540545,0.8596639930937094,0.23700049565202808,
 0.17823757978397625,0.593021001201178,0.44626681095840404,0.4783569651877728,
 0.35379814438740176,0.1639540061071284,0.2216374098436853,0.5004985332731012,
 0.8522429804271806,0.8431287796328317,0.43884956052491575,0.7507981243833097,
 0.9809442554341714,0.545411596173371,0.15228005161244917,0.2678662227809978,
 0.27524069315263977,0.655606235615307,0.8323880845097259,
 0.055238056771420574,0.580393218014036,0.27953366324141915,
 0.37884520827143453,0.20137604927720631,0.128800136107048,
 0.17999171189108776,0.5222835539787547,0.5148682707669173,0.3202466678044994,
 0.32514651125169225,0.0600598535219381,0.3102704977840566,
 0.048874801899533415,0.7020398144550672,0.9212135361580434,
 0.34663897600370863,0.8010404188108348,0.6302719972782601,0.8353566798529518,
 0.7367938393117272,0.041688737845057,0.24387780883268095,0.45077594175648616,
 0.3989667939458812,0.19243909814987104,0.13912018908056445,
 0.5988570717233813,0.14459602333953936,0.822388956581899,0.45056020516822815,
 0.9278948279256315,0.1873536332802721,0.4150706617941219,0.03428034011910841,
 0.6555816744634808,0.8393569364771287,0.44848235953958737,0.3139168712440994,
 0.6043843491766312,0.14574145588641274,0.3594059385701347,0.7485908603995324,
 0.11492030053599167,0.8365232562468184,0.2527709087482599,
 0.41642498366359226,0.0713556769853475,0.8284619437236918,0.5164383623284651,
 0.4617979022405603,0.617876003227316,0.10107059342344327,0.11558593470891121,
 0.3849520490553753,0.8685831869641601,0.6733178928159803,0.6476237009974666,
 0.0316378315111866,0.6683282275054053,0.31750216426112243,
 0.10996182608510552,0.3750887383551461,0.8406667041636103,0.3717339842013583,
 0.3906991290263344,0.5122803934519049,0.035975516583120815,
 0.8769388431007297,0.9859527610119627,0.7180115866703836,0.23971104563176826,
 0.27564475726047233,0.8695211039115585,0.3298044125593266,0.5565963115198571,
 0.026160507917075915,0.8927120116249773,0.3391474856912966,
 0.8376555677179544,0.6775707124723557,0.024329859832602496,
 0.08391092013573143,0.9662451087294635,0.7564210404816397,
 0.060547671679398785,0.5250411711779217,0.7712613577055474,
 0.4895114397716974,0.8756716474783461,0.7414355120964772,0.9827051788042994,
 0.012358188922462987,0.5395356023281657,0.20739125070809605,
 0.6104973742888715,0.5721211432376307,0.0433473186883464,0.6592490038153758,
 0.5547512751063641,0.9103351096124589,0.36046160151797746,0.9519246992494462,
 0.20430653870976934,0.6562712374266197,0.4824977679105441,0.6044356984508854,
 0.9247323230035585,0.61055160217561,0.849848510763746,0.4098103894331562,
 0.5024393848430635,0.22843581176793115,0.17146872650408518,
 0.6973716493875208,0.7906301071202315,0.0693177788431969,0.35694301189401056,
 0.6562975995846938,0.3626955899238897,0.1450851637375805,0.408010616496709,
 0.8014428430752137,0.21395990923218222,0.14918913994973748,
 0.5696228771534126,0.012222073804152966,0.06686346138427324,
 0.9407275437238903,0.9809678004621546,0.1906651100229596,0.24370622688103227,
 0.6040850746216739,0.21776101981140905,0.5289116888627737,
 0.31802243921171225,0.2651048651678751,0.5024443965773365,
 0.31004294115438347,0.10899524089325752,0.6855255883420424,
 0.045159024611683085,0.05447239802804371,0.43464998732647486,
 0.8392414203556294,0.4975476067991377,0.02275263678229733,0.6181312975083404,
 0.7166735549670712,0.5357259134393709,0.3126605853150607,0.9562067696979359,
 0.7461289102824054,0.37722351030205037,0.7059494997928951,0.5434304394391438,
 0.9352803858049342,0.8821557105766514,0.05800224042005364,0.6613438051953663,
 0.4914894645061775,0.5208423568712863,0.10277364965146008,0.3009978112732433,
 0.11045317900079288,0.3776564750504977,0.49968634049717875,
 0.38035182598680883,0.9209818668304743,0.28517933318761657,
 0.9015259326820028,0.09615209653248724,0.83685160202411,0.5127505706559804,
 0.8047768301607618,0.28566216390322563,0.7791561370848639,
 0.024045899797654258,0.8001713759030263,0.061316563846387506,
 0.7266280896985279,0.34635162286856624,0.33123689091285624,
 0.7828213239947059,0.6561487491374827,0.1187143142274385,0.9657646908001228,
 0.9458793278659755,0.4716230598645722,0.91508140505314,0.7506056352018465,
 0.4793762722227479,0.4081513878379611,0.2572417303721053,0.00777617262271435,
 0.8375858085365555,0.32807885935460124,0.8041325961470207,0.9844194154812579,
 0.5968049614315685,0.2601128423837371,0.4500024502802378,0.5435423866073545,
 0.11880067420082563,0.3812206476087986,0.00712023767436154,
 0.2584275157776743,0.6902512881760751,0.5393230126777265,0.3255058090457331,
 0.7458105530398377,0.6120223632722512,0.6907380195230417,0.6220000180405667,
 0.4670032667538816,0.8110308691882945,0.6073940604122148,0.22915082104536888,
 0.2924747814519494,0.4624759484149119,0.9973116681773002,0.5417818968364168,
 0.13245424674560347,0.9361058545345088,0.6035999447296899,
 0.08961008788668168,0.10479941033500784,0.241694800638663,0.3340078713136052,
 0.13872470479472665,0.6966485297849722,0.4636776146322452,
 0.45744271658333635,0.27173527204584447,0.9687719946510966,
 0.4305540835614401,0.3771880622267877,0.3369781819344788,0.49074823529069145,
 0.7570241779665269,0.5519112084472636,0.41381639531416015,0.6104167989585891,
 0.272108151486366,0.7359361784917177,0.02775809273363994,0.7613306568056312,
 0.33282467002385197,0.7014714537457271,0.8118229250282835,
 0.04769671968589628,0.27378208892775924,0.6440263489590737,
 0.9571141471434454,0.9632292476919584,0.5114970558698488,0.09923040770635427,
 0.4317908508047349,0.10940299581292778,0.8831040606136301,
 0.36450897251500414,0.8740237925486305,0.3450219892941956,0.4945372209588026,
 0.9785006610270365,0.029619005552592226,0.8137712635245563,
 0.9849927658664382,0.945930148238316,0.03580390365902486,0.1587950505595348,
 0.546969162226793,0.2075496900547995,0.6669472576292217,0.08797011403936683,
 0.49409843141560006,0.428946985812375,0.06304465557564021,0.640125316554643,
 0.37939257319653596,0.6235145994212794,0.2846150053533012,
 0.22389346803490606,0.2996961514924974,0.6823708759693815,0.71157045430597,
 0.37940484825104903,0.9481205765726938,0.9007941197599078,0.5494872405429305,
 0.8759801665891505,0.5519834025705075,0.8227788488601692,0.5743250354584041,
 0.19320103329797678,0.9401830603202322,0.7965162419174203,0.7489660536325448,
 0.1462251675008589,0.5446670668240078,0.9807258810433741,0.49255515242632886,
 0.2870090693623676,0.5539897721124548,0.2731828590573546,0.6532210278913294,
 0.036919746269607145,0.3644797008218589,0.3920528320600638,
 0.9141956764980984,0.6784040448111005,0.2795755117911811,0.49443733661305533,
 0.4253495878057456,0.08944429340657611,0.7631323360047306,0.7426553402796624,
 0.950997161475776,0.41086113854910244,0.7975670917016402,0.9227500165074627,
 0.1371125837041871,0.4583474295758694,0.06807931068609674,0.6246988726609721,
 0.2031891228318593,0.7257769171160035,0.4801861308571841,0.8700039514002296,
 0.8690709343450977,0.7086634963555712,0.32220406172510185,0.8082896829320863,
 0.3409170115378535,0.9876439792035456,0.6443937634959316,0.1374344186275439,
 0.3408179507564466,0.007813932180106065,0.389993873159285,
 0.12678026859449176,0.9082841124617909,0.10816762327920582,
 0.2302850361647535,0.894164158132636,0.7610179501787406,0.6876087890525981,
 0.4128408842719,0.4306339818392243,0.21231616699175404,0.26322921332769644,
 0.622173686938224,0.7500571848560665,0.15418000722493996,0.7288436450751173,
 0.8963412472246062,0.09910448614661105,0.3578461954429537,
 0.13538289476801024,0.826700508485568,0.791228553299099,0.548036415873201,
 0.5436515998503372,0.44880804050167034,0.5167415429919693,0.8373830393459637,
 0.9458716987215794,0.9242145447278273,0.5813989066445582,0.43061920382407226,
 0.8401318270594541,0.6350911528713999,0.23904137788943436,0.4044549592655665,
 0.3424365705978456,0.7360001465481724,0.7094570550217197,0.23346473622164754,
 0.03264278653477737,0.5235856614083197,0.3690908787124251,0.8760378279470493,
 0.779400732338764,0.750188973644687,0.34067525830000656,0.9335560607935383,
 0.6393920625710274,0.2235886118790542,0.7095511435260115,0.09969406573901929,
 0.2531139195191058,0.11926684901610562,0.1718440559922817,0.2459955080785432,
 0.04124291568394689,0.9292692157570527,0.34926081039399404,
 0.40365919680169704,0.06617805934792198,3.6754924675841405e-4,
 0.4689612493558717,0.3282045992223859,0.8471048546172244,0.9354810332639276,
 0.3113090355730377,0.7063630843986383,0.829057300671836,0.425660302540103,
 0.5008295938737287,0.8378973371916124,0.7571208343675782,0.14353916028473235,
 0.15798778873706865,0.9278160766197813,0.5056940647542869,0.6766713951047116,
 0.5186242437366767,0.7775098946147936,0.7337633309134763,0.5318464253554148,
 0.2979254568647537,0.19282991974539287,0.07248949760278911,
 0.16776486895890275,0.561200506570024,0.5572082049033593,0.2128553188692177,
 0.5876826094991511,0.6887948633381892,0.11893221110947994,0.2393594264179384,
 0.43848833505943263,0.9495597321192044,0.32899321817768856,
 0.6254187961959213,0.9053168205913806,0.21700625670808515,0.5140483131284186,
 0.7069483110288473,0.7074521409599474,0.21957547984828985,0.5371823966143106,
 0.9339835138001896,0.042560750837805306,0.5212664362944689,
 0.5826555347209782,0.557171476251384,0.32129885703632843,0.12055682300685455,
 0.8401716137055741,0.027039855446196137,0.3979311544774855,
 0.42514101743975763,0.22558217020352034,0.09894073355083166,
 0.07838185970903766,0.7724472680478125,0.13100237175774632,
 0.24265734364404545,0.2638507924859138,0.45731159404687327,
 0.7478017384301839,0.29442609665616093,0.12340293101742139,
 0.6874949666739597,0.8518649555685415,0.6077998312241086,0.16812018263504913,
 0.8043551282693334,0.2571355093619879,0.14458975573691057,0.8385635823397337,
 0.8235895845054394,0.3907674935348625,0.09791346221546526,0.9646493593511745,
 0.48185293252643857,0.8210879773231186,0.882966195690274,0.2857545381494755,
 0.05285842823184905,0.948495912977157,0.33533803686380603,
 0.37354379950024386,0.035235507371959685,0.23317058023011983,
 0.0229979795657419,0.1603184717469951,0.7096841456301098,0.1721244492676277,
 0.9149844806412019,0.39790868786479183,0.6015538656967055,0.5014589488981872,
 0.25723379330466667,0.1985027357939766,0.9843022630794214,0.3641815222358824,
 0.7182292615419394,0.9081363283998375,0.6988149230252361,0.6352226979262615,
 0.26432150122059084,0.7220255734715718,0.7949663970022345,0.9310931218651639,
 0.21422091778917185,0.010791337739749829,0.16349786578909198,
 0.9007652025896804,0.3731828358470958,0.5185113694987181,0.42887351802267604,
 0.1192505675381671,0.26786990832698,0.8898884740349027,0.20500611427278026,
 0.8794605486464894,0.5661176429932009,0.6601886271648651,0.868894911825215,
 0.8881452846720885,0.10618124115105632,0.139081826597673,0.16298536853767698,
 0.41827234724451456,0.1859290589125402,0.07545641379711487,
 0.2977662629339295,0.9470687596865012,0.1622511827018588,0.5185926491917903,
 0.08942471814451725,0.3189427074110184,0.31768891386251696,
 0.7956441737370605,0.14608743961561832,0.7198387905121348,
 0.31424049274662325,0.38731004071924147,0.6852750556268942,
 0.48455888833887284,0.5814679342465952,0.6800175624161235,0.8314336852129389,
 0.7256944607788259,0.7189188413152803,0.33335409669142424,0.2661156711893977,
 0.8402594124095454,0.39527843310520616,0.6296070368379674,0.7784715790798661,
 0.3819036739632422,0.247717170368832,0.6762323271230947,0.689706441788321,
 0.3612378089104664,0.055726983445136824,0.11164215129603772,
 0.6258496057701128,0.1850220705932858,0.6212543120773718,0.3382042430186355,
 0.6175761131047494,0.4651018238324015,0.3715732752814266,0.47581055663874483,
 0.8319962785516783,0.6716882297828132,0.3769569905247536,0.02926729430061281,
 0.6825413015316211,0.721837816940354,0.8129634280994608,0.36711739635199614,
 0.1148376313421392,0.37630089769187425,0.448661392038591,0.7680778042111281,
 0.679545796781554,0.4947286423396666,0.7643511784070512,0.9539784990285595,
 0.2779746968657222,0.7909412746766491,0.4197724851614222,0.09658479388128627,
 0.6528967580017679,0.5509803407433524,0.9756582009720992,0.38543310505868145,
 0.45612237784740584,0.6912310175844762,0.5145258914670128,
 0.38948725594237665,0.749196352099364,0.13365449243716798,0.8333722931611328,
 0.6428886842213698,0.17917298022724193,0.4841967499315504,0.3484557602965369,
 0.3924602886515973,0.8436432740848141,0.31835356761695943,0.7654203289477882,
 0.7299973085119615,0.3755485334430937,0.7279933956793684,0.8414797005234534,
 0.8313797771074152,0.317801940869642,0.9411421005224092,0.19721392400101434,
 0.46788462484789617,0.15806809488983453,0.5256053345631879,
 0.7364458074291738,0.40765628791548936,0.4495054868591275,0.5581849486071964,
 0.8979217201550863,0.321593692838146,0.11004250902402213,0.966225227077741,
 0.26587891010579834,0.3655631623535571,0.07862273821394572,
 0.39069881620704927,0.864989089446941,0.04909727971624067,0.5237518964845336,
 0.50348299279739,0.27459403471145427,0.5856889933654644,0.11291389195507762,
 0.2327806032015941,0.6879938783525349,0.0028493151102040315,
 0.5272766808514288,0.14456765986189601,0.7517450183511779,
 0.059968606699107974,0.7051207663965942,0.5386957694454013,
 0.08407720457307921,0.7830821011134276,0.9200441283080202,
 0.44202573939965806,0.398279472990139,0.9398425342737511,0.3856745800168927,
 0.6613070627758884,0.36593647142538377,0.49971044008645016,
 0.9532616572634474,0.8435885656823047,0.3833810892485795,0.24746513168980222,
 0.13366023051922682,0.06288461910114085,0.9604476649191642,
 0.6159731458826475,0.31800120102151896,0.34681563738594945,
 0.20747583625902632,0.5431184154350414,0.2716164252190293,
 0.17569998880029836,0.45754069371466155,0.25124519992809524,
 0.9172673058363326,0.2684840216677714,0.824404781248736,0.24368749273832402,
 0.441649791965967,0.9402203426804717,0.7772179311902294,0.8666036254801476,
 0.14796541834381705,0.5400539404650674,0.8571553734152371,
 0.049973452248507355,0.6991341998209356,0.318489663220063,0.9099338352342596,
 0.5679539442014745,0.5356558621264793,0.8394110662041588,0.4569690052870331,
 0.8164554418931067,0.5670967104363973,0.10926036656999505,0.840357671095264,
 0.04929473160030684,0.7024974582740882,0.5854696627085625,
 0.25658529648024286,0.2426965672305823,0.7337575801412982,0.2405666192250976,
 0.5326495605109116,0.23119137355997976,0.40784519623663995,
 0.6722375144314452,0.39219041775302876,0.653297433552066,0.186671414538928,
 0.047486188601366774,0.6493761010989739,0.4381959625783336,
 0.48900671987730326,0.21029080366264796,0.7224152284530769,
 0.8799193344300948,0.42417826232839617,0.44113050374476426,
 0.9893519753802386,0.21768143835303744,0.758739476580131,0.6955179277051776,
 0.1596157233820551,0.08950984891235159,0.5617114106794011,
 0.09736595640280887,0.9749699565004291,0.446939820883138,0.38650569499023013,
 0.41619348456262284,0.22927585245546322,0.3947707671515339,
 0.11624545597624458,0.7047650124882526,0.6540982210387625,0.8442488046588446,
 0.6599336797023636,0.7908780357548741,0.6119689591591233,0.44008230077603283,
 0.2868975650467035,0.33465198790604944,0.5283325316655217,
 0.09091738905842872,0.053596106311394,0.7592675713564981,0.36843393392432766,
 0.3750917128883313,0.1617249296044777,0.45446971271799896,0.8920499052925549,
 0.8724131766010583,0.6893467195828927,0.17254678764451636,0.7433092376511503,
 0.8659180040122104,0.21792554547391552,0.438608745510331,0.8139919117968697,
 0.992973947691042,0.22726141586275195,0.24488954655795048,0.5464956656963197,
 0.29784973776694146,0.29984731316177216,0.9916458875369772,
 0.3317679587145017,0.19154109316733114,0.8146502564637588,
 0.46770608036495465,0.7036341120945107,0.525434672421736,0.31807090525945136,
 0.9326526043133606,0.8690361878942525,0.280227531558787,0.9996057968882939,
 0.3838154450985285,0.0695956501953382,0.295286656797801,0.15374720267634356,
 0.19368989562865435,0.8568406960883597,0.1802718380344298,
 0.42963868950883566,0.5934279353204678,0.47911876532756903,
 0.48528375706176785,0.92628027021888,0.7457341528309767,0.09333856855242995,
 0.38990319168910825,0.6785832989345147,0.9520714998955605,0.3164016851917941,
 0.79850423093185,0.4428445546831685,0.11832781656365166,0.47342654167683795,
 0.774002244833531,0.5022413120270209,0.46574248773937577,0.9184873768613175,
 0.3666091507100533,0.6771449323593618,0.7562011751958947,0.8288100846057537,
 0.18646202476032192,0.39811429152889355,0.04763666408245115,
 0.9955362264586716,0.4254172725658607,0.038123650519763075,
 0.5740685284926301,0.845880904332751,0.059316357479539006,
 0.48979896330711836,0.6151137995157535,0.16562263426762103,
 0.5633334662382872,0.29736455198082723,0.06113350751388147,
 0.03987512136960858,0.7158205313339719,0.27892716445673016,
 0.031420295459772984,0.1879873795788204,0.8871566608630072,
 0.8574388077437853,0.8155725943949975,0.6164806810271583,0.8890166914292273,
 0.587883844810238,0.958014403074816,0.22622686117583124,0.7884207526675535,
 0.20077735273018094,0.9827359076828857,0.1721592722585239,
 0.20773109222962505,0.82183690561642,0.6193237975100967,0.6128682338679531,
 0.4874724137301766,0.9262460576929736,0.9518322124109376,0.17829742107667068,
 0.6584342007535461,0.21971955869282123,0.944081116664434,0.6175243034094566,
 0.42157492777686534,0.95231685036165,0.8316717405857357,0.5929700327789259,
 0.5202096179146344,0.9844167992681299,0.5884931577289306,0.3199166702195544,
 0.985062963585909,0.5509877261672416,0.6921839466355131,0.00933686644781695,
 0.10085637176990003,0.7201118792310591,0.5360963791300041,0.6355657051637255,
 0.08048121001370667,0.24780763708786602,0.1808949721790697,
 0.15706798996239224,0.9639097493743066,0.5790122384514268,0.6359960397611945,
 0.7812600451488692,0.7919947466713521,0.7641198615497853,0.6905976826039674,
 0.8812052405395325,0.11919215270283856,0.872581373658044,0.24850641556376962,
 0.5859492947073637,0.0642348675869916,0.9723024255951868,0.5856817893181336,
 0.06649547854593962,0.11714373456798743,0.5140509245274932,
 0.16868669615601517,0.23227739127114455,0.5849489977675706,0.7624626182523,
 0.32064965401967416,0.19062896779865757,0.8913437460727893,
 0.6317530536159283,0.8694336347159327,0.666209203448954,0.9515450842745038,
 0.5249146424173752,0.43685847932034283,0.30530824827079717,
 0.21875628536376068,0.4420975568171813,0.2326932467025078,
 0.22541158086702973,0.2784327750791147,0.15108107617267952,
 0.7512924144735091,0.6147757399292769,0.3461569829119966,0.41248516692194626,
 0.7776855283380028,0.35263054935869287,0.7052448210533994,0.7962823317147992,
 0.13002382123207767,0.4943087998332645,0.7323370029583139,0.8683642846518584,
 0.7910704990592997,0.572690398664412,0.32875339131004055,0.17781243780607392,
 0.08009606446562767,0.7410822818254609,0.616498608433343,0.40414513008301967,
 0.5451237989482565,0.07871094702050963,0.5333441742891717,
 0.32243414328623576,0.8641051387168046,0.9460416468065299,
 0.14046143740994843,0.766290830277321,0.09934692346934915,0.814097639776395,
 0.773703696096794,0.9456823962083745,0.1060924923579939,0.3303696625478302,
 0.4678203989190882,0.4627368210645967,0.4352435991614818,0.3129064136046248,
 0.16713506771644215,0.8020463213693543,0.6741383307976584,0.583307144951898,
 0.7060884003486009,0.4940240020795581,0.3980673026720124,0.06754703279723129,
 0.757427769174087,0.30401974280257094,0.6596454577327538,
 0.009068421926982762,0.40505849288283546,0.6929150031821837,
 0.828855036599192,0.8101298046691299,0.7603809368739713,0.3663666484433127,
 0.7042538468091244,0.8453719052484439,0.05926565629797376,
 0.20162583215681895,0.8043863070989158,0.16677820639320373,
 0.5049522403673496,0.2639872985896241,0.2165564216061806,0.9059955880572813,
 0.6204692347329883,0.5409167354039035,0.6101707584821259,0.45217624827402747,
 0.40170921706947293,0.7659417012836834,0.7979905288287764,
 0.29933574709815736,0.4594355378256746,0.9566355886059258,
 0.13308829406257217,0.9521232386790524,0.15790365819701857,
 0.6521674093677791,0.6904326868682559,0.2538762323747529,0.8519677496153207,
 0.8230804717857378,0.4635412714999855,0.07867216827820811,0.5771995867335756,
 0.3456814636976431,0.9787740725265408,0.053996215621097354,
 0.8641897810468393,0.44157024997700167,0.7650616761687961,0.2002602819084185,
 0.23348047441669662,0.733277442048136,0.6013733900768045,0.6643652655469983,
 0.33689827007623985,0.047329933595688045,0.10238388999653614,
 0.9082221181209877,0.6609227821467881,0.30259109275920715,
 0.22274678992937635,0.3444710653624826,0.8834461805847765,0.6193738817464471,
 0.25773793882611207,0.5868645056214681,0.5398570797931684,
 0.13351069269279003,0.6815067766361069,0.7951294070925731,
 0.23408734560234545,0.40776369303398086,0.016886049449957974,
 0.271589970230977,0.788365969730819,0.7133568812361919,0.6276523195561612,
 0.020289789922116497,0.7041815701297128,0.6705612891556613,
 0.6918946614589847,0.8963620400382363,0.43016680149099074,0.6983416783594115,
 0.62484948587754,0.9398024635123079,0.3667190362682107,0.044744561677198424,
 0.5741895509559442,0.8248066967653652,0.8036034131273212,0.04839478706166789,
 0.8446338682179899,0.6820316939639433,0.5392478766120211,0.13215412900626056,
 0.48946881577235746,0.6843237997862953,0.6476241829625888,0.5478524397209891,
 0.03484696466158377,0.3542225581449292,0.2741259818231845,
 0.44794736709591343,0.8659284425000114,0.049802861306039836,
 0.6169961289445849,0.03427295900383209,0.7614292371231626,0.7300585571167627,
 0.4939242310963444,0.44021424093916395,0.48019678163819224,0.626914209902406,
 0.35160840380050895,0.46493527115068933,0.38416680545054105,
 0.007356783533556066,0.706041941190571,0.13966840882080855,
 0.2848414731897233,0.9721686448446577,0.771600079278403,0.4697065042879416,
 0.6661743223776698,0.9295748169385353,0.21835716633780788,0.6378421920700812,
 0.4982977237755687,0.813125121651393,0.07131710026453264,0.5018987533694756,
 0.303724114289305,0.10919777418040788,0.9896602826228231,0.15434974356021813,
 0.5165033182927535,0.7674106158055398,0.8752925061544139,0.5563156114576676,
 0.3061262492202723,0.9891364111720288,0.501355949114113,0.3661880792432042,
 0.9632117602552773,0.47993786925312454,0.9607864759493724,0.8266159796061405,
 0.7996148002870859,0.06374496290566856,0.6617376161427001,0.6101367947014762,
 0.7133543747263771,0.6677399339174434,0.5501162857944063,0.8606635564093095,
 0.8628960608680151,0.9223363086049977,0.3890418341623435,0.8719353814817487,
 0.9685486159998737,0.0790816123808169,0.03887228805807352,
 0.40477155056127256,0.15759675841774967,0.46717915367422247,
 0.7981852869170663,0.35779833584493237,0.4531629940049353,0.4376252519228835,
 0.1103315708995436,0.18881502833960018,0.006863009595306835,
 0.15772804398194307,0.5823785020395015,0.9789311954978559,
 0.17132117314527917,0.551746573392928,0.7910450105807117,0.4609569638277341,
 0.23731996520875243,0.6857083028357372,0.10037939287094422,
 0.4706189211099293,0.3682298226494357,0.6177150983560941,0.8719403267181871,
 0.9970747339373917,0.27995430451712966,0.38199350660709286,
 0.07911645391250177,0.5648608878993975,0.13513792448604156,
 0.6111425869494092,0.45617908151368414,0.6875245049805679,0.9880712499305996,
 0.7977944686864772,0.4149001106173533,0.5876523993997207,0.45971608912067863,
 0.8073762841332718,0.2448186937704656,0.4200874130846457,0.7204000204299039,
 0.6675091412722839,0.8137384373064677,0.9708571960023217,0.18276327319423702,
 0.20328268151231077,0.1435780208460843,0.04758315855046025,
 0.4161089137822064,0.8802287224693137,0.44599951188384446,0.4715716742364946,
 0.8526112258069652,0.0794449455016959,0.9790285008749465,0.12997467790072448,
 0.1201331007907186,0.337345080557214,0.915453829024068,0.8575870126340696,
 0.02058419058470662,0.6661971643002184,0.15932600458452073,
 0.2998858782103071,0.02604860211208193,0.2052409167947452,0.580611463297236,
 0.45772684935578467,0.8540903439641205,0.15289484663786934,
 0.8987324553775846,0.6367075252452807,0.2108505744579886,0.7319099203871136,
 0.7081566451718954,0.7450296492611878,0.026654429223211817,
 0.06644709019486206,0.5113368842503325,0.2074334334862633,0.6900939005124321,
 0.7981715937270235,0.7758298792693763,0.2818138818581435,0.9303163746736508,
 0.3248721090238449,0.2529172589869768,0.5464976710983674,0.5992456377555085,
 0.7111444860410872,0.5927398906958047,0.006205333510908351,
 0.6597216730350645,0.9995572673116824,0.13331791502725554,0.7621105948899628,
 0.5338599237771444,0.01832773968106305,0.8816685549070096,
 0.40696786167257715,0.3872163020913626,0.6014050797447901,
 0.04499451237278462,0.02914990767851533,0.8877625214081473,
 0.3395507245480488,0.6605275074373833,0.6867422667271168,0.17623772248885894,
 0.6424930090711568,0.11241275644706084,0.5822488299832977,0.9309131582482436,
 0.7311865952459875,0.2825977896762393,0.9429355480163563,0.1334579449174197,
 0.6672443024046038,0.31993117053453046,0.9024298187669536,0.8368393330013028,
 0.1398711219197809,0.5318328802124156,0.7047130271027816,0.8688531533089485,
 0.978155242134672,0.3651235344007403,0.9250030187755269,0.867913749937385,
 0.4264767132742533,0.0985490658279976,0.8832545417353515,0.7127632296315564,
 0.3258980199746073,0.008417997368269092,0.21726583883742268,
 0.7705768917798409,0.18824072101689993,0.8628293329049641,0.6224512935597714,
 0.21667118162910137,0.5670825889275445,0.6098980274545849,0.4496654848692766,
 0.1318105777906493,0.7702319184398569,0.0939423837266804,0.7966144264928205,
 0.8284881326795422,0.9157555135069148,0.8469892157653536,0.24872754477535508,
 0.36156775643121075,0.39324389762411693,0.2966247509155253,
 0.4257237113114192,0.8315426793118225,0.011741885344924374,
 0.18186549957766496,0.49036195724927134,0.07073534048732022,
 0.6949228608773965,0.17560035335717405,0.17872892006407626,
 0.3673743031094949,0.42809634312167544,0.588180872986543,0.1898709596662208,
 0.4777316853280299,0.9769715841448816,0.13345774527757714,
 0.18904749838695856,0.6261715679585139,0.2409930511763626,0.5529689608082052,
 0.20004029494950404,0.6432286469723563,0.3073178626797062,0.6912307104547329,
 0.7617724765551821,0.9413826469085252,0.07510010719375049,0.8120093906725157,
 0.40055841877469545,0.9737361016304134,0.25226019936780375,
 0.6763143797109432,0.3123449950331558,0.8710383914881571,0.7198089557868923,
 0.13389277882117134,0.03458659393040564,0.5830719918044117,
 0.8228352137546738,0.44162356327877417,0.15056696019782712,
 0.9103826820847081,0.5656599124203985,0.10485048508283212,0.8997468499771578,
 0.3212735748541108,0.8892082022521232,0.059952525736737816,
 0.7777242007556782,0.8146498941289498,0.8116914067969748,0.06955620108073235,
 0.8778524228918069,0.5141695590240696,0.5452228470636002,0.9746315693787592,
 0.5380908263089192,0.8411757957177444,0.05103575289813067,
 0.21569380959956808,0.15878476150670062,0.3313084491787214,
 0.3698648409637091,0.6653162957255206,0.14403521761401672,0.9751705294895022,
 0.6772258000331987,0.13229212577011407,0.6709952060236222,0.7428693217597503,
 0.19295487756509744,0.39395105673718034,0.8085755910815035,
 0.30877672529529965,0.8006266802905796,0.33727643634018845,
 0.3829079620237168,0.6556849905837413,0.38250210912378546,
 0.005650732496347399,0.7624918842710404,0.9980978536798827,0.081550866915449,
 0.9670898796389966,0.966584147212325,0.9830870785123292,0.4413787755710442,
 0.5542344657966924,0.7655338880322984,0.5522771726922957,0.1878115833211711,
 0.20873034831341586,0.9259435529005222,0.6942155496521025,0.970686038260145,
 0.3261544820496578,0.3073048076847892,0.9064612120105772,0.3221460808021722,
 0.8907970002868281,0.6581407792505535,0.5538704903250069,0.5239932662915042,
 0.3548531474368368,0.45480750119860536,0.6853316483693703,0.7233127540375603,
 0.7991646878376789,0.1834332519202434,0.1814824445003198,0.750662181874075,
 0.3293131453155078,0.3953065402573557,0.8059978171447714,0.7049667940011166,
 0.4770706276248958,0.44563095097828787,0.17833903555364672,
 0.22537730159464842,0.7960313835351256,0.8844418233721434,
 0.06082315590267462,0.8415309659569482,0.04459292292582062,
 0.6344221866493951,0.5041013560355283,0.47481473276792874,
 0.22409241989379025,0.06714393605689428,0.40574431839436165,
 0.4388101375735035,0.021282924220239474,0.0838395205498983,
 0.12306284767370446,0.9614434803262482,0.11405135308566539,
 0.7922650602723675,0.9946827316634022,0.7071896926811623,0.270729788809133,
 0.1192254992912557,0.7286910520984939,0.41057845057530995,
 0.47695623214725025,0.20538890942480525,0.5667101309215936,
 0.34312647018393105,0.6198482475352378,0.43130502859760755,
 0.5102925925080528,0.7962666676476305,0.8539011836843196,0.8962633838215689,
 0.5596618645906868,0.39587531876664517,0.37731125482079375,
 0.5581158084311831,0.719169388335501,0.5764471477669202,0.45904795097668283,
 0.6222244237722372,0.14137387419655956,0.3200763011071809,
 0.11476340421576114,0.23016767810632321,0.021400258165283237,
 0.2875664577302989,0.22200301331038885,0.3654339167120201,
 0.17645766693775478,0.24829116363698833,0.9359715030029213,
 0.8006363344179006,0.6003732667222251,0.9847401136253261,0.8580596061793669,
 0.03639622053952907,0.8600253166558178,0.7300863698691288,0.771289994463141,
 0.15477940666287016,0.5415406294123487,0.8822719583428162,
 0.23915623748510906,0.3760347187636198,0.8966338235421736,
 0.17007886129012006,0.3292022184486454,0.5161067634533998,0.8792226812622834,
 0.022730926499288873,0.09814771583547488,0.9631048092095111,
 0.44019298441570653,0.30351220853264604,0.7158213234538686,
 0.7154480937095569,0.534578235111209,0.6264435012591184,0.2887275444079197,
 0.07745878294835329,0.4092881205087783,0.9086881901418844,
 0.19930701392318362,0.5889252838392393,0.5224471621386191,0.8764154270104862,
 0.642075935891789,0.8323225213261767,0.08534723146460466,0.04280877689720386,
 0.9393474025880311,0.13517654743549368,0.8982450933778531,0.6259695737351962,
 0.32613509547984143,0.8749669117143603,0.767177438052968,0.2802022289409388,
 0.8826670695755113,0.8719296098012356,0.3030079214590742,0.7820038586739031,
 0.47638879139918533,8.605508720771482e-5,0.9710766185240487,
 0.07632447047404467,0.7201023385933366,0.7498391661103847,0.9688198909594096,
 0.7514903206453405,0.04641642687846481,0.8929119838926305,
 0.16709948801622498,0.2731369765687075,0.5374646513464109,
 0.30900195407679276,0.909491235235278,0.635825171315086,0.9717531060538309,
 0.7315562683447306,0.7299829299108893,0.39118358210234694,
 0.38762970892437865,0.629201837377823,0.5489840809331432,0.24128136943971168,
 0.9563844417690904,0.02257620962149498,0.4680077127538831,0.7510333783525699,
 0.23785724033511446,0.6252731107200605,0.8579004041642927,0.5838310642100002,
 0.7018269112440282,0.03699501153345608,0.3351161774822352,0.6277274710906171,
 0.9745559498578242,0.5857711443143407,0.3860615451999855,0.9290462832456533,
 0.5826933091485279,0.0014082026662656233,0.18773375453565533,
 0.4992922756235628,0.7654526034471676,0.12596978627781052,0.622095704166141,
 0.6492325758132581,0.1330653914576767,0.9011598885180166,0.08509042833447789,
 0.8770535070224205,0.47351977994919636,0.7046987510139271,
 0.31726253558972595,0.3941264946470442,0.9393751297805087,0.7358538697297246,
 0.5187771108228731,0.4880680803946946,0.5948054049930059,0.35853950782925603,
 0.11075902776342894,0.058247128464897324,0.43553582520584755,
 0.0029103485052981437,0.1285296569201333,0.9658032054691963,
 0.9458246735759019,0.7430879657698486,0.6847212337162494,0.6944465480411572,
 0.8643340651239159,0.00868481499018603,0.7204152832144096,0.8762427637434801,
 0.9831929763652894,0.308621994066564,0.7311124336516035,0.11113243433777586,
 0.60506802835811,0.6939606498838623,0.5031346045080507,0.49484190276691586,
 0.12285576790720087,0.5809863473763895,0.716635351293918,0.9217417879939825,
 0.7871314068887431,0.6002803403103772,0.13693478712561014,
 0.07362605747109519,0.7837830521269777,0.3365428300969804,
 0.22918955061512214,0.39389826416177953,0.5284221144630601,
 0.7695918675839717,0.624150040075405,0.25819244855164525,0.8891020203173939,
 0.35666149452109175,0.5772859809258851,0.4725128076663848,0.7068225702569422,
 0.6151973177808225,0.6351165527572478,0.7418345758534078,0.44591020997596287,
 0.5914367164720049,0.4772901962483639,0.9709208026415783,0.6734069923865206,
 0.17728375818908992,0.6907915853092359,0.7931201957397337,
 0.18625788334011317,0.7284815084691497,0.003712882023462094,
 0.24326319932483131,0.39163829199684796,0.8190083054238732,
 0.6827986360988141,0.44312838710353586,0.34111079747520623,
 0.8510387948972458,0.7483882106804176,0.65255821125578,0.3686550207943101,
 0.41945164092377385,0.8898510014894916,0.5391444038135149,0.6503691684629052,
 0.3370519616771488,0.37251269618011207,0.11479322793466096,
 0.09766455636017035,0.8463794255804168,0.3879749225686504,
 0.06744995016122812,0.09885275714125386,0.8852355282068651,
 0.18785196250596203,0.6454329206839861,0.23734789601825146,
 0.21986348446350512,0.9660189998227326,0.42352586046310225,0.635019861924554,
 0.6313969577861738,0.5186590880150193,0.7367181879237451,0.3815063823942765,
 0.805265791972484,0.3054985073814406,0.051708450267421435,0.9367604783475205,
 0.9393910303941214,0.7303386532039423,0.5616092979195959,0.7204641388598196,
 0.13312154148233313,0.8555108483647174,0.7675368373765401,
 0.36492522744516387,0.9116301862047289,0.9712058235370922,0.4416018159541011,
 0.2337098961523868,0.1312737534168722,0.3543440660822279,0.4665582358215894,
 0.9121734721347865,0.2418171993156586,0.13441254319713392,
 0.23632703037374891,0.5489650826741884,0.32842592615350563,
 0.5030258089115992,0.28813985237780204,0.6701239823497608,0.3275430391569514,
 0.7411519661881666,0.29039918206295456,0.16844158746732207,0.927688491773653,
 0.22463362523228314,0.16035353706344646,0.41194880273067835,
 0.847185225144023,0.909643170785976,0.28678157767480883,0.8840749563430845,
 0.09291499107947732,0.06294950550472667,0.8316917023246119,0.530208181145161,
 0.08710561951321072,0.5723897813844103,0.41558025045811287,
 0.4275373887935743,0.8497392319912479,0.13916540566915026,
 0.15015610675784152,0.46025483956033275,0.8958212728673123,
 0.23828209927019428,0.31392735840830577,0.12813491383482467,
 0.6255299295392355,0.7730958177358671,0.26283396507826917,0.41764004335186,
 0.756237990058064,0.8742218250351491,0.35389856859339197,0.15195249900108576,
 0.1230283476966414,0.5089417766115301,0.17539866716553842,0.8641638145814621,
 0.1388684604939845,0.046159507408184974,0.29717480290535325,
 0.7688152623189933,0.44352923296074453,0.7467464170357948,0.9559733529238127,
 0.3409841380254448,0.9531747048882053,0.7407331859993618,0.4614056015147572,
 0.686840465565806,0.43235992099218523,0.8276510880011206,0.6231587100869758,
 0.038549714317016814,0.22211646044704592,0.965346766160721,
 0.8756192802565285,0.8211199137025128,0.9281347883480431,0.6023951370771436,
 0.45015907531676413,0.30322754838653665,0.6913616309199497,0.831836815794657,
 0.1028859374205352,0.7256125959447154,0.8228411478306942,0.8045734576758902,
 0.4007561873853064,0.9407113652774782,0.3283529045459739,
 0.058795926518399155,0.8028319921181862,0.086427822074436,0.5458450427606488,
 0.4423953239816073,0.9741735044035987,0.5909494989517302,0.12279745821384502,
 0.07652514769557273,0.26463567864867343,0.7637892398765008,
 0.5425930667392589,0.72500302503487,0.37029896369105564,0.7870242932800218,
 0.7806262502094965,0.7435429763698688,0.6272087085332085,0.2935728910629388,
 0.31228064073050665,0.4032292107674236,0.5054382154909037,0.7146462972850266,
 0.24264257463876915,0.5984211896582797,0.8245947345731616,0.6936711907207034,
 0.647164963466216,0.703622955126044,0.13426935894354697,0.6753237131401495,
 0.2637561594824879,0.8024616105610944,0.2525301150418481,0.4689118036670191,
 0.6906389569914768,0.6309788642973707,0.748347061803614,0.22612344558202724,
 0.275416085234384,0.03690701094249782,0.26568044210959785,0.2846561096302025,
 0.926649296629688,0.5382062317096891,0.11660717023106804,0.866130538417292,
 0.9379771579133249,0.7217713372652399,0.7219604799421792,0.1567005401882866,
 0.2721567228017334,0.23718990122932593,0.4800798013797547,0.5479126713789222,
 0.14078632019214032,0.0903214006885662,0.38784013877508416,
 0.3204147101807219,0.14192662807720002,0.8733049235938216,0.4108853544595056,
 0.4951238416320205,0.6315005826854856,0.3307520168046053,0.7344800854360295,
 0.4797279791138521,0.12044197887177177,0.8816481876146831,
 0.23946742697640033,0.3156101715367825,0.5820555088846459,0.8795856825923735,
 0.8555318373842516,0.12197061430778988,0.689479373140315,0.06662722361388396,
 0.07769269681645952,0.11160636632441101,0.5622256209589388,
 0.9758278418203375,0.6477580261157017,0.7021907778871433,0.7702335244783549,
 0.44826379509773895,0.6190206693385014,0.4935482102755768,0.4532871636776272,
 0.37803671092641933,0.20566924759285088,0.5519394467407329,
 0.27703661955814607,0.224997860436582,0.00802347868228892,0.500376514552261,
 0.7094747552027749,0.3181639305179125,0.818611856280301,0.626753367927434,
 0.8057690700296639,0.9956511275331357,0.18225190761869858,0.3412426899778026,
 0.38920964967930405,0.1760716184277874,0.31953504726511883,
 0.4171744461625194,0.3052026753334174,0.37748788669690203,
 0.11227135654062936,0.061826154033986946,0.9128635089878608,
 0.6987590800112353,0.5029344009600641,0.5489017914915847,0.7485354483300521,
 0.43133814213393395,0.5251909712315241,0.41465299834159286,
 0.15439935504907298,0.9822290340239931,0.684880221371132,0.8665588652109439,
 0.4379744760839278,0.3461468458996133,0.5918872753941912,0.1298526478499945,
 0.48208974307894625,0.4888321699604211,0.10664105681505853,0.458144517686891,
 0.1750312330947703,0.8509645400004677,0.8825849455494991,0.6320823567999311,
 0.6634952140550898,0.6368583366657522,0.4818976405370413,0.01876338898005203,
 0.23117657395296687,0.5794540398687158,0.13093371346081928,
 0.9410778275741367,0.082050072236164,0.331992859803769,0.7179714546204635,
 0.07402225228392356,0.7388118921537689,0.9595298317594271,0.8931737482662228,
 0.5976290333669331,0.21431684223401037,0.08695148989477608,
 0.3508524526119283,0.26300631380552164,0.7336432052760461,0.7309652403990758,
 0.2933741538262271,0.18870724908490288,0.6321185981752226,0.6881809603960365,
 0.876724607841455,0.07333229608865222,0.6706039705025119,0.9903587284830042,
 0.07077865510577752,0.3134891830816251,0.5079618011059048,
 0.31384384210076255,0.4287113083067786,0.3087338260388207,0.8710119233683418,
 0.5323703135929795,0.517821887862893,0.06154666668040021,0.35452242133167333,
 0.8380922068033547,0.8586605492069689,0.6827492204773908,0.21030806440051975,
 0.11803179237540795,0.5286711902947365,0.8581581988368714,0.6327323753469195,
 0.031352144429207085,0.23169727551858554,0.810614593163775,
 0.29518281009519476,0.10029778018231295,0.3484153074668198,
 0.8496160868162442,0.28171787969107653,0.35248424461149075,
 0.6375238028964424,0.10900828850476885,0.19912510466574518,0.829163635544766,
 0.7324023409175009,0.6658865555612125,0.8712358953102904,0.9410291165983191,
 0.24177909215217652,0.9086916124081088,0.11528306297610824,
 0.5382616515315749,0.23843343396760286,0.664999462749626,0.2360217670590441,
 0.01208321096703835,0.8780746393198736,0.6718664129485488,0.3320140647225084,
 0.22812929640263846,0.9771270371887153,0.5316690336667466,0.742083818230931,
 0.6476917520127619,0.6652677006271265,0.8550735643788334,0.30001387495832854,
 0.06668989385233348,0.7001430317059386,0.05366349839576079,
 0.6662000763094357,0.6562780931464287,0.891351834895186,0.27144280568635315,
 0.3820882562119755,0.6142068969617507,0.36007707791482246,0.8536730461656644,
 0.11854885975789353,0.36057438448338897,0.603542777283804,0.5044824446926159,
 0.7072262831095341,0.6176726224273059,0.982553876585184,0.9053973968432377,
 0.7710307606957116,0.3600773854018925,0.5778893826838021,0.6305089155498914,
 0.7208561387065189,0.6342015366019439,0.48986304658082147,0.7854581982966435,
 0.3955020495090954,0.4483180166774523,0.5361879023671846,0.7043614348442002,
 0.9795167694537346,0.9319665165835671,0.6132032939840536,0.662451218133158,
 0.694300655505663,0.25827707220559293,0.20412350266194035,0.3393825544819965,
 0.5889058173495043,0.5062384628015344,0.3917477610612412,0.20377884434101057,
 0.35358832610203317,0.2412019241276443,0.9730275272888315,
 0.23922761481790156,0.7895412320313718,0.9499167704304914,0.968185696358855,
 0.31868810149112536,0.1859324794381435,0.18580563401753392,0.254649322567003,
 0.7833962993091665,0.9501506106102239,0.9148536601953368,0.14358344355782493,
 0.7533166359427612,0.9466824013636961,0.7743095661560158,0.19466120086671468,
 0.9489208570227387,0.6005669140945319,0.9196020339771447,0.6869435434747942,
 0.30197212668797546,0.4447023518985149,0.11369054307818316,
 0.2903409569958155,0.6828739336889013,0.20121547914335958,0.4597657984596286,
 0.24332180449133667,0.9512897892471452,0.7267458395554893,0.4891641718478419,
 0.20148844363521368,0.7964869192548409,0.7919951954465104,0.9496175411403207,
 0.8409464342746112,0.95622008109474,0.6213913108463194,0.6949409235688329,
 0.9869924907552254,0.7279456306029296,0.08121393649096542,0.1478844203267432,
 0.1434990809408614,0.2536992957033728,0.4615737849310496,0.8311760483941534,
 0.027689879702967102,0.3057677331846318,0.11994207135918034,0.6329008546864,
 0.524704753641175,0.1745304367157201,0.9342578106665465,0.3103256933722265,
 0.0033957241963072926,0.9515540660940166,0.8191630998992304,
 0.2713602479681727,0.012501789045210598,0.4226204761219148,
 0.2666481393208211,0.21366334129154685,0.9359272264369588,0.8044887945368929,
 0.5006752896098734,0.9432617411206551,0.07743133783702849,
 0.18829473032318123,0.6024988361858333,0.4689766424738069,0.7431279499835626,
 0.09516204697199315,0.957926616072272,0.40289907387902446,0.8163649721439005,
 0.004551556398932677,0.5127082580926826,0.35883602758520317,
 0.4814799992602501,0.267738212262845,0.5087846873975137,0.8435316881869096,
 0.5563762017866457,0.8225429357109377,0.4704197575425322,0.6988156232178429,
 0.396046372664825,0.13893976959084586,0.486032337274079,0.5034955855924694,
 0.48886032602399854,0.673791149802203,0.9259360298142261,0.5425276587923139,
 0.21271290313734703,0.3094168015502561,0.2652569653063722,0.9015363092821292,
 0.5036206025771522,0.6347670171514143,0.45125276336730347,0.8881442884094937,
 0.33196151385178074,0.35168185519792616,0.7715953019397763,
 0.24391578370213385,0.6404790679812117,0.9579090277726339,0.7173133497386495,
 0.5941273573633998,0.34987225289534707,0.5811880531589994,0.9276333877332675,
 0.6515762497527773,0.49209270995751286,0.4677710002244706,0.2703746697194723,
 0.06319163044544351,0.9111903443955356,0.01025433575488599,
 0.5165151426467846,0.541899422617404,0.4153750476374799,0.5978883218836666,
 0.6137412129424479,0.16618140675440696,0.15071903511013374,
 0.7755972236388039,0.16781395549840716,0.317301000647494,0.11692117026570326,
 0.479668268771859,0.9376387260238307,0.7287570759129628,0.46614899934770015,
 0.8017971376485966,0.5973885983768911,0.5465352782006776,0.22449598105911006,
 0.3474953763613021,0.11051677611082178,0.5477963991651684,0.7144386251859488,
 0.4304069785541401,0.24496804505155734,0.6271332979775589,
 0.23762015024804717,0.7939056573161973,0.41234524548437945,
 0.01882425770649354,0.8620265031368263,0.647174485249224,0.5584150417304735,
 0.8126021863998923,0.45494612065061824,0.5134676234127151,
 0.07291225148067615,0.15873466344282905,0.9898122976442918,
 0.9815902979980797,0.8093717063161913,0.4162950927664876,0.29521376671499977,
 0.19470400468734295,0.7548292555566404,0.7288855682813893,0.6992705709453607,
 0.8269788746370338,0.2731824745104836,0.06636837256495665,0.6269115920404683,
 0.867659562865919,0.3893573734609336,0.6759183793455417,0.7309108743037276,
 0.2763543050747421,0.1902310063429168,0.26365842522226945,
 0.24965401505107065,0.5807077806938921,0.7741951794093387,0.6016338375210575,
 0.6502714121109078,0.3774606665614214,0.8733839820946733,0.07249791791718474,
 0.9609427754285292,0.4407376127720716,0.21655366299553147,0.2882442596800723,
 0.19895463290518767,0.9976975368544732,0.23013405512829133,
 0.10659940126641287,0.7485570992588954,0.7435228446030391,
 0.27473733241611176,0.1499128824545376,0.48804418067898603,
 0.18821870156026654,0.9169100518717106,0.9838807556745058,0.3436170434921899,
 0.9065774010148244,0.047941119872865534,0.28150530074524693,
 0.8344591443776594,0.7176301703462143,0.6657814784955881,0.7210163819700619,
 0.26105236089291406,0.1034582810917214,0.9862986996304406,0.9315845176905071,
 0.9887595907506745,0.7103740559370684,0.39289376444000723,0.2179427740186619,
 0.29915026878988216,0.9109238599440673,0.5040353244884126,0.6254449553573301,
 0.955460668224931,0.028585311856435758,0.5071400626703844,0.9855478574761428,
 0.6277490735693855,0.29863702479010223,0.5396183307777662,0.8649212531907338,
 0.42879632256112654,0.22788744094468827,0.4790766177329694,
 0.23326854090777283,0.42251992623451373,0.5506845156306346,
 0.16781245609479667,0.4204977815508868,0.5704762564685446,0.4359937228657206,
 0.008804959162866388,0.03883399730424408,0.14532318153787505,
 0.08427640775808865,0.4586715571386589,0.6455028081139402,
 0.48420365714005753,0.37550860470501213,0.7190388329601708,0.53542290392407,
 0.6919741060503302,0.22924369991633475,0.6913823847878023,0.8367072126163201,
 0.8541913607137639,0.5456001525552414,0.6540904956711628,0.8404880932721897,
 0.8593463529156092,0.8165388548659359,0.9934121638219865,0.3236825732023574,
 0.2740438553786819,0.5580584597144805,0.24444785351688747,0.613887704770804,
 0.48431924604080034,0.7903882191988351,0.020190617079994366,
 0.694395571662667,0.3318265018089095,0.7897768263544661,0.9986682337464833,
 0.47762231836430213,0.3801930815421438,0.3773966056703274,0.6789766992440897,
 0.13513646062375673,0.9734891607134458,0.7335264798420138,
 0.07737987682798741,0.4538251952090353,0.004065217803446375,0.26869166394696,
 0.3061260400875262,0.5452216092704106,0.30139956823030534,0.9978950673896051,
 0.8177757049748083,0.7390478739388133,0.7802054972110437,
 0.018802525832841144,0.06388446418805205,0.4666095137049001,
 0.054389161876506886,0.8913450937433264,0.5845323272913387,
 0.05063302516068802,0.09191254988328956,0.7294044713757885,
 0.7543129320561858,0.6229414645520716,0.011831605218906471,0.415780241530174,
 0.6009964329946091,0.17073506633737678,0.5614338577271816,
 0.42820792472125546,0.5454292946757948,0.28334225984506667,
 0.9462022291888897,0.726574185593694,0.6618984008875317,0.4416771065698021,
 0.08664087443099033,0.8978471025597309,0.34837356767094474,
 0.29528527599918675,0.01831311407772629,0.7254231142898475,
 0.38593820680841573,0.9386970772567036,0.18274078085181544,
 0.4518659523027887,0.7838720574777551,0.21178568661672115,0.5989761959734209,
 0.07020989069074823,0.9899470504082616,0.5025944059961699,
 0.18987520414498937,0.8512927327836584,0.8364815305598848,
 0.28157829796680756,0.40721878650961696,0.685531118329903,
 0.05784316709037984,0.4991563733304166,0.21668913164931292,
 0.5708660838578652,0.9557731206978246,0.44071491323450207,
 0.02638809016174859,0.1742140714584658,0.04219109014058875,
 0.7772281753871328,0.5862557175496648,0.6027739205329353,0.34750723865813904,
 0.2637121188224403,0.6139215446144572,0.5036628186239791,0.9659173608854129,
 0.1983571022797821,0.03803374511235458,0.4948415342338509,0.2130705715527451,
 0.09749060014939592,0.6390216922087746,0.8679473151382646,0.2020881127502443,
 0.5804747559482386,0.5501431902693877,0.9156916935482894,
 0.014337976048720114,0.20777432712279598,0.49170738630910704,
 0.39193343622869903,0.028716752531660394,0.8620690162736596,
 0.6234095524708552,0.7838606306320963,0.6870914031280085,0.7670200910909872,
 0.5326617817591086,0.7516034215973014,0.8478759471081705,0.2947680623353095,
 0.5462200762045898,0.7667438680262084,0.5786232091497612,0.6331753546903658,
 0.8631101573224363,0.0073535460683255005,0.9737643460516949,
 0.47622546179659486,0.7943481892561355,0.7503688639563526,
 0.20258813186005142,0.7183483256390593,0.20579973526474127,
 0.8409291099021472,0.7797872590564698,0.2052770199780395,0.36457202531337796,
 0.6717097592583705,0.2577389625226021,0.7658411482982629,0.1749668776100589,
 0.49409898450284195,0.3712252461566766,0.06473602322760896,
 0.38289372862751503,0.2321322466555047,0.047140580566612966,
 0.23944294914218778,0.8011180107217687,0.07960939883983231,
 0.3027622109432141,0.20642599659264083,0.06115403929851304,
 0.5669345645427848,0.5982561284480922,0.589061055722456,0.2900969837633913,
 0.9291555953965801,0.8848130480386701,0.06531055412837783,0.9233346678118708,
 0.8347548084089811,0.10360879746794682,0.2980584348468517,0.7347184793630602,
 0.45649141002359683,0.5086955864033177,0.33237065754641226,
 0.6735169657980601,0.7201448596764687,0.3771144486769348,0.3775830148398551,
 0.6266869502232951,0.8854376420267258,0.2396790664102686,0.4891862965575565,
 0.41464761401517625,0.996040169903326,0.4129825943026638,0.30461297469923854,
 0.765302764037939,0.5377514445585035,0.5090145613262533,0.5519218935845296,
 0.5629600620105888,0.9100911039520619,0.6764900592836922,0.7763428741817857,
 0.8124224742294133,0.27020432405151396,0.05991506159384907,
 0.9643419289181508,0.03946411885323475,0.8458412502255352,0.890729132241011,
 0.7537511089359549,0.7721858978171698,0.8813816497886242,0.8230895653432316,
 0.11129117317201698,0.13093089224178722,0.10151996524732698,
 0.8372688679602625,0.5256321657813453,0.3880538266143444,0.9026054275896952,
 0.749454734361579,0.8604353970055145,0.0216900022204366,0.5731381920387527,
 0.44873696080279835,0.5452465823624819,0.014724675926718,0.997513872206677,
 0.8365481329510436,0.23978362585000745,0.6583079091929451,
 0.14053332582141287,0.6930169431602126,0.2037431112579806,0.261325983073899,
 0.44517388396974666,0.5280895686394956,0.34464853116546834,
 0.5061135658258086,0.47960657553755415,0.6642851321632867,
 0.45441256109413075,0.10349310787364108,0.7141584459098709,
 0.5478485619674678,0.9570599642822679,0.06147027220901413,0.3162327504554493,
 0.07028401635529447,0.6293447825549932,0.09438426168780434,
 0.6226360015408327,0.2644784158273866,0.13453898751636473,
 0.27519599658277283,0.6517329735951446,0.7102526518675059,0.5141789519086961,
 0.7057821688107236,0.6020668933681321,0.8914264888779929,0.7747639030642949,
 0.38030253135268777,0.7277797982204719,0.3181761683808664,
 0.006731799252581272,0.5423639387149521,0.359475117652019,0.6452414233780143,
 0.06615448894173148,0.4784893099816512,0.008792082692223313,
 0.1107776736089997,0.01028782410238327,0.8409767074959811,0.4350738834894676,
 0.5884750406381892,0.7093744120152544,0.3463984469326937,0.4588296124780432,
 0.4677297348168211,0.39969535902487296,0.8131296311829992,
 0.050511697499251995,0.05258815888897306,0.3294761746249375,
 0.9307160383508286,0.15332319632497593,0.819299230440101,0.23205126191678138,
 0.8015345531769436,0.3726479487658084,0.9646462444135518,0.5812834254843431,
 0.631753260914886,0.20802015498680504,0.24570496212143578,0.8200037452876906,
 0.5596654053749668,0.763520128131008,0.9832930734760601,0.027494950341112823,
 0.6591998124462553,0.8935369547947647,0.040502376827207365,
 0.5027174108041861,0.8865760960683982,0.12110559638094087,0.5224669996411814,
 0.6583896260893205,0.04021357937521819,0.3320992982002061,0.5373245462444788,
 0.22613127545281875,0.1827071913355388,0.9997362701027126,0.6533781330133522,
 0.6562728836502523,0.8610735935705685,0.9579483479045708,0.9618630952654185,
 0.5960412521818504,0.32375633543701676,0.26126250295597475,
 0.9960220576406478,0.7779521595425569,0.522370222513698,0.24131347455769558,
 0.7075529078285518,0.17333982352813782,0.9926338794601718,0.771492374978288,
 0.8193904422037483,0.815164830538013,0.4273269804769896,0.7091226161211961,
 0.6132280604078711,0.037681731597171586,0.023475696702434545,
 0.6420335381844379,0.9323404469469614,0.26997864014279804,0.7695776373937306,
 0.34680584528075264,0.31118607299644807,0.20222063313178018,
 0.8457084846933912,0.3491479373371933,0.7077812257552233,0.46090110223931213,
 0.7360498259454598,0.14530523759357106,0.48732373497417714,
 0.06694545563025212,0.25816015304492057,0.5888839474809271,
 0.48944263879356487,0.14277544582248602,0.45815200744958906,
 0.2643493420426831,0.02799524341524917,0.024474282807804504,
 0.8158584739099832,0.0565631546832317,0.11926322488071228,
 0.14909727593796218,0.045990893772412944,0.1604051491824967,
 0.3824218687896208,0.7999664182033905,0.15729687517502722,0.9498556425875666,
 0.886718752209975,0.11358514991278157,0.10677508148546677,0.8717240130767626,
 0.6501540872462422,0.019682560968893204,0.2912366046581134,
 0.033451150881325886,0.9747156688562835,0.9019939861520907,
 0.7037098441429722,0.8637376178630392,0.3652070197268944,0.611405045785616,
 0.06673292353137672,0.6739345889740678,0.31014787052865755,
 0.40287284668469026,0.3656522287001688,0.24918260897336308,
 0.013195056328885402,0.6366212517811574,0.47872641150925466,
 0.48352564387631614,0.1483241181527034,0.9460064677915314,0.8010866442746766,
 0.5844616786138417,0.21096721892329595,0.6992813075997811,0.698159713115357,
 0.9775111079257681,0.28487412384788935,0.7419118480743583,0.9741587009189072,
 0.5102327578522324,0.6288629574186867,0.4490559349649824,
 0.0031171220228738417,0.9525636046903734,0.5866710274221035,
 0.25433687190148213,0.146068035551569,0.20386241988592368,
 0.19328104560741932,0.8020186737415063,0.2548823815237401,
 0.07928361990616128,0.14773282745632432,0.48630896340710017,
 0.3639536297304862,0.3531471740537031,0.2618332078195008,0.5221605610983843,
 0.28147242387002247,0.35528469852684397,0.9160625222223389,
 0.022571687989317746,0.9745080651681763,0.6443923958523035,
 0.5118864174282823,0.6810294326235773,0.18779197035786155,
 0.42038123284175877,0.9941208172924537,0.5455052222080994,0.8710540635684485,
 0.6424895607361074,0.32692088687098875,0.09270146254568234,
 0.11749330024446891,0.014926057682630889,0.4308963640392186,
 0.5549137405136233,0.02371747312106831,0.36293605798633455,0.919282294031554,
 0.054389907624508904,0.6038403821967613,0.3332304304468876,
 0.47991455177776254,0.22127360648424843,0.557709415264186,0.2073335458686708,
 8.268050197843824e-4,0.39380800169123187,0.562570996664845,
 0.21813342513271572,0.4721593279681048,0.3666501502520949,0.1195895496276177,
 0.8810680183612878,0.9070089520375063,0.7924851725431032,0.8003179593995522,
 0.780210995886988,0.7546877291413377,0.9514398833929306,0.8987569941291486,
 0.22828280817691338,0.12232081749931323,0.7602062381835168,
 0.30212887576431824,0.6084760036058351,0.27965501195358844,
 0.6975772485698735,0.5456029317903479,0.3192679382832696,0.19280779589821895,
 0.7244662108197873,0.9072563488345549,0.07073808424870964,
 0.19506735869285163,0.7769926551374223,0.2221179950203238,0.7316470074197227,
 0.5270224413707569,0.09323555783136439,0.7358615643846772,0.5791588315690579,
 0.47972527842002455,0.2410280778906697,0.02367988799600007,
 0.5956511855708895,0.6825680408245138,0.42614668765098707,
 0.31263371792896666,0.9410832490066939,0.05311669228527216,
 0.10134433615015725,0.6907841659093309,0.05987852416315953,
 0.7085589317174316,0.5785757704246898,0.00677282606823304,0.6671610356745877,
 0.7101685030852367,0.6522785212298874,0.2659846303783203,0.7053707734022225,
 0.4628983494856239,0.6673228287763169,0.8906881110123321,0.5148997190658418,
 0.4372954908580442,0.7200370853436777,0.30116933815945335,0.5000890087642952,
 0.5783071156970242,0.05161960829767298,0.45067002694589053,
 0.014918801976851914,0.9503696972294459,0.7484955451211365,
 0.7052184341933481,0.4878301022447946,0.6000162236137612,0.02919529188782788,
 0.07883168542374153,0.9780511543452577,0.26922296673442503,
 0.8718141185818471,0.5234808963147761,0.5832120620920496,0.44760386520167483,
 0.3198575505604462,0.4646252848819302,0.5761770757906004,0.132047758264767,
 0.8169016333706254,0.6913205645286582,0.543759113708488,0.8746217674235913,
 0.516656729817871,0.8325930966693789,0.9499740546828233,0.33241789421932855,
 0.20840344830633994,0.8583666094659634,0.2648207441656364,0.592963367076606,
 0.7574030459676214,0.9256595700032405,0.610993361115322,0.06620174092093245,
 0.3095981347603546,0.5153893826297229,0.8032807768931016,0.35438976520462917,
 0.8435571117319474,0.5701815410161486,0.564367254008456,0.722296893158244,
 0.04764159326104411,0.5884551749099955,0.8625818631594814,0.4851187404905968,
 0.5017129402508677,0.8553629331902619,0.7618942314726783,0.22661686405260395,
 0.55435515319115,0.6661340077269676,0.11013124180245848,0.16205658048426463,
 0.5981397688748976,0.029378138541273247,0.37304419570920067,
 0.8324132872789738,0.9422130297329507,0.8521578416610983,0.7269854840264961,
 0.5129945241433969,0.7204254245233044,0.18006277548705119,
 0.40148761283397294,0.9202711877131322,0.6506351772504991,
 0.46956048555989827,0.5558299713229876,0.5780939709146594,0.2104603922560988,
 0.844213308561999,0.4478690621378364,0.7015969764693489,0.2914022370354714,
 0.9000416430045896,0.1800752379195475,0.12402838325383558,0.6134300814685002,
 0.1412046552549363,0.7581779099219662,0.6401263792207437,0.17054164344935208,
 0.6362411913368936,0.3643485597396564,0.1430881815023174,0.8729745205011823,
 0.29418757980743415,0.4999298133729221,0.40058040075859014,
 0.5990028455246958,0.22944268078803154,0.3092033905226854,0.667602227456619,
 0.2930153256459951,0.7236139191250828,0.8944615842301284,0.45831135330061834,
 0.8512620088134598,0.9354139663717028,0.9880371271472354,0.15241372307858225,
 0.0988426221518307,0.45427920827395907,0.49709975124319217,
 0.5717725990358508,0.5629202243642498,0.8117095560436719,0.18020878805003782,
 0.7373214723896389,0.8991876646655973,0.8893670053361087,0.44569165639600994,
 0.6300352441225705,0.2214176265413097,0.7980724888710791,0.7622549774699102,
 0.021502608073135754,0.829440164390624,0.5714405722691037,0.9648856347489797,
 0.9364852686639682,0.24496744415782545,0.2857149433351954,0.7539883589920436,
 0.8583876620331874,0.9237707143627831,0.8598301554248177,0.47904281692982253,
 0.9176498992644508,0.40965006298834905,0.8904894465379634,
 0.18299173183612993,0.07033844841233516,0.006571260860243955,
 0.009729254292486167,0.09019482889953323,0.9331128699657629,
 0.871375728135595,0.8171130198230492,0.5949314424368246,0.43122081502627063,
 0.25378214341435634,0.37403845318812057,0.35330282482940834,
 0.1278854699983193,0.6379801584607931,0.62556929227782,0.7813331929376482,
 0.5344600138992688,0.15489299397066358,0.18022406442056094,
 0.05408781788447081,0.8329459522537005,0.9182052321785432,0.5058478934097499,
 0.5675165637896842,0.33989665770076083,0.44891235179007527,0.591982167254321,
 0.05871323218144464,0.1426568388142524,0.7885517397094507,0.3388010636649702,
 0.21156902306760872,0.963900804625553,0.8576230354619915,0.06801993816887708,
 0.07909727287407242,0.6935152918110405,0.8351125769226897,0.4087887640139033,
 0.239622156834439,0.08286761336030413,0.9334197768666975,0.7127231587890991,
 0.016672511740541074,0.8335574598965336,0.7564319495799321,0.8772654121423,
 0.5965453919341159,0.5145241018103197,0.42536439797404246,0.5147266832657732,
 0.6497179313731558,0.3671728259969841,0.8694923895323226,0.21014789441100457,
 0.3915381318965261,0.13048911798144192,0.05500262626442143,
 0.3074492742489062,0.708945990506413,0.179309628491096,0.7436675324624821,
 0.8428995210636969,0.513609741851489,0.6740393867474535,0.9511139678870313,
 0.6787696020282736,0.42032017618071693,0.62841554125193,0.365460881625139,
 0.20916505617867576,0.11941180910408722,0.5695587007082736,
 0.022029118074718412,0.7678705737663757,0.8992487569733496,
 0.3464166551887925,0.3497432785029968,0.5200080508826003,0.9292614234216442,
 0.06708832501344497,0.793808896056845,0.7895748067842177,0.8085312491266514,
 0.3058147721347366,0.48790333115224827,0.7456378706143382,0.1344695887418358,
 0.32989083711745415,0.5490142982481981,0.6719207279621051,0.636448171225068,
 0.47567129033263367,0.06134714818193354,0.7684015020105138,
 0.46395970561065947,0.5229803166383671,0.4035109304583405,0.4896779782180555,
 0.1107650527088041,0.009212746850101872,0.1030059343887958,0.213705794785489,
 0.16322981504365852,0.9571806109577167,0.8960330183213366,0.9291579666178866,
 0.8255843862413856,0.1786521674335957,0.061267487967174405,
 0.006687674202649818,0.3235381730610253,0.29659455136112545,
 0.24690113524840096,0.2849771718941264,0.14819077685249094,
 0.32080799677445426,0.02844824524974854,0.7299964565528347,0.654264658358357,
 0.5350463050166681,0.4209129140093656,0.23906137902323565,0.6925947805104201,
 0.6559981908352489,0.9810865808694716,0.755273686085679,0.5058177785969047,
 0.9249259576418998,0.6337182832207053,0.4416272266461443,0.08540501288950408,
 0.8251010201431186,0.3014584276244858,0.6688605919746333,0.5965067744405821,
 0.5291549535286378,0.9608618858714969,0.6484023177418419,0.47784562821443,
 0.7804184906088326,0.4899605161592081,0.7799462119451468,0.34391140680454413,
 0.7346212879684915,0.7636489702247136,0.8128186558296564,0.0365384805932798,
 0.7278464859842727,0.2780656440690942,0.23823905334230855,0.5693454587994563,
 0.02986846474883187,0.5977652211339105,0.19740296680666658,
 0.6847044104981819,0.7044854214487005,0.9523413355235771,0.9715735798478619,
 0.9424568151125207,0.7491732986061197,0.8640439862645947,0.8183138360607828,
 0.5773481147662444,0.1733737144448303,0.7268428020829039,0.7954398642074063,
 0.35591036637084983,0.0709608761900844,0.17415231318805258,
 0.35670013498674946,0.7356968904051006,0.30448827608912266,
 0.6014012432047715,0.7781546607743746,0.8057783009292795,0.40464957132425095,
 0.7045237925398389,0.7838300484589215,0.13234360989599636,0.9351538050246884,
 0.8634990340549634,0.6357123885171262,0.6125518189157296,0.8089813536324854,
 0.9160539985143993,0.7999047931330517,0.8096857180544552,0.7952332475980725,
 0.8758889132366446,0.7197754322538404,0.11007399049708755,0.5079905648634341,
 0.25526255278410737,0.8701094414762932,0.03172046812640661,
 0.3125829669090364,0.046080262685068485,0.896051054914008,0.9351953176701201,
 0.14460427136110954,0.13681971883535426,0.9955284343604088,
 0.9211546040849421,0.18517267807807325,0.8435482379919834,0.7727908622893908,
 0.6424896098237151,0.06388016271363306,0.6890547104517335,0.6055722927608332,
 0.1508983025631535,0.8444451995325036,0.8512785336001649,0.7043521499047313,
 0.9143604742119433,0.4548775337267892,0.3219632993467777,0.8055833494765988,
 0.4406015941639534,0.9314055622138552,0.25259610521301823,
 0.22081894963281334,0.9395682060701602,0.07149575734468683,
 0.9709422006330506,0.8242809959188212,0.5902064787005739,0.7347879586177897,
 0.6919466689594662,0.986840122940359,0.2490791150366155,0.36521952631566,
 0.9558167436457985,0.6058028211307755,0.8999516215606039,0.9250119726602677,
 0.11609971028091304,0.9644051476568702,0.8558502785873325,0.3963546945013511,
 0.7654119527763718,0.7290626199119753,0.5884273252406756,0.8243859448090592,
 0.45751184408057466,0.1251128123417441,0.8780122611183124,
 0.08861386188617215,0.5524377715048394,0.22883156104798874,
 0.18080609340215703,0.006573452806345359,0.5498810525458974,
 0.4210755661111023,0.18154375784460486,0.35010665391727014,
 0.9443621869242714,0.962981567600222,0.126899016626295,0.5878891083259847,
 0.34962431833410523,0.14277229446343886,0.454252267538822,0.6780184363923606,
 0.35558179675113066,0.7719489077438801,0.849713421852848,0.6866799110918336,
 0.2218945253248451,0.8046956372705858,0.8508048624985979,0.7549189124605394,
 0.9193184055972253,0.8053165083045807,0.3049845308203638,0.8168808025761195,
 0.6267510223723177,0.8590489670248531,0.021383662537698966,
 0.4840335553829177,0.47274620180379234,0.9194099339831563,0.6514573178439242,
 0.3152021042199131,0.9037196908357856,0.6509435621109243,0.01508344585683874,
 0.7119324972230208,0.2236575841134001,0.32537752555790456,0.9306994007471204,
 0.2850442565541955,0.46874541160915695,0.34067033080751075,
 0.7459089660700904,0.034070113167693816,0.12350068158770433,
 0.9968669769676299,0.877026038752667,0.5130956968380989,0.9707055357548688,
 0.17047100186040365,0.5102644926180347,0.9500473452214704,0.8658990219671843,
 0.4713416443609877,0.8654266455167741,0.5630100069666124,0.4382341867265217,
 0.6183691363578779,0.1740862804146136,0.9097931549312179,0.25539480761580236,
 0.4480998609287802,0.7439075226016973,0.006568778449278834,
 0.7571113386824927,0.8439892389691389,0.2858883425208033,0.4091109188668678,
 0.5031937605359152,0.8730666578207232,0.9873326945874876,0.3381789725350177,
 0.7111197325498679,0.7380168761723782,0.770401276468442,0.6522390352412509,
 0.9703793355559369,0.40486252920948596,0.8116887168472758,
 0.49460668356753623,0.12814096458543933,0.3609512930803803,
 0.41434362049826123,0.11759657533917534,0.4909921648190648,
 0.5036213333561858,0.044103829528091665,0.1448837219089022,
 0.19999638124178243,0.891779890646433,0.5111198388592575,0.9035459684113218,
 0.23279895080256818,0.4253919236147544,0.4932913721507448,0.9410669570748368,
 0.12265139224449917,0.6242733574234867,0.9831453639088288,0.8603863335569834,
 0.6835888729546546,0.8231458437661172,0.6937969508338195,0.3784917525011169,
 0.260004530820946,0.39229500887856716,0.13121316702717556,0.6295242975177839,
 0.409898649346791,0.8626018280113403,0.2504263562345883,0.8712399041882799,
 0.9820329186415987,0.9515687290176055,0.1925462062460691,0.11579826088951573,
 0.5794617670179006,0.8652548220732013,0.7390156566439674,0.9587356258591423,
 0.9174228783830073,0.7306674238517339,0.09190580549263228,0.8205239739598326,
 0.4869846674358391,0.10034748014597661,0.8909913969429045,
 0.15087514172463345,0.3848786136015353,0.0689147853109654,0.192544318665817,
 0.44918965827924284,0.019305499263701398,0.38938437396207126,
 0.7761396295105576,0.9614362921188535,0.9999824778092437,0.814422220367197,
 0.907435703558262,0.6954024036371661,0.29866817202233553,0.54314278599132,
 0.38201927207220876,0.7459870828194348,0.5885201831195652,
 0.48400387486239804,0.2609445224415965,0.3541439469742874,0.5709199875550923,
 0.549914168216916,0.6317521922815299,0.14332358814261192,0.21485252401039034,
 0.4319013868596856,0.02034562119457206,0.5372790122080726,
 0.32634040711906065,0.13309282566593605,0.9413545505935944,
 0.9501797808718713,0.734384463201286,0.29013627574395173,0.3367445024304808,
 0.3687512029047536,0.060194325293013606,0.3049223981279784,
 0.7835388500466278,0.4912279686770562,0.6905539339607475,0.00872496313708615,
 0.5179663645545619,0.524250519118712,0.8940788232863857,0.26397636094081944,
 0.3223535487361391,0.9986380279336591,0.9287458452800214,0.29801985677778375,
 0.3805555900993808,0.5552983952247057,0.37745162265371524,0.3521722852829169,
 0.3861322929580965,0.3305116216418006,0.5675271321010623,0.801433767575839,
 0.6897549489778044,0.7920920113110784,0.8516838424782806,0.5294877945021634,
 0.2137450302319751,0.6146953610998066,0.8954169252305098,0.33933404797915323,
 0.056742822953344096,0.5824750667072859,0.30729791596875833,
 0.6059446499891175,0.15303178486298397,0.13985387959073503,
 0.5864012056560872,0.03736983964100671,0.6073785939897709,0.7959703290438852,
 0.19759015732418095,0.889149758268554,0.562451550450407,0.32853230705531955,
 0.3207406960983923,0.4936254290157579,0.33624298956864884,0.8671185271366554,
 0.609896734752113,0.3333771002825249,0.28240466900759575,0.24388645957794575,
 0.9567417694442593,0.3553474686778384,0.6958055653330747,0.2381601005132329,
 0.3836665652374146,0.7093017939335312,0.625622335356311,0.018017240148217484,
 0.3374241048328285,0.8998434092574779,0.6594859865748149,0.09061706506929834,
 0.9929477030297322,0.47719599143934144,0.26722050635165473,
 0.9926105510085586,0.027028457270828032,0.3378411327781178,
 0.4132146349385649,0.19666372148588862,0.5315619805487384,0.9007097330123432,
 0.020553224725839092,0.691513381854606,0.07610901072433152,
 0.8445700888395555,0.05766325660027671,0.08448029239511845,
 0.11783018510946319,0.1560182842319281,0.15865836043156367,
 0.6101455316129647,0.2872627746426385,4.596428179610701e-4,
 0.5069723112425613,0.09893296467860557,0.6017334830874065,0.2700136140929337,
 0.11359397091732903,0.052122405842407016,0.2875195949611088,
 0.3486271162738541,0.10813264855628557,0.6304006737751346,0.6004352474072262,
 0.7201850107858943,0.5864809811449938,0.7231156628971751,0.801802957335249,
 0.8835465727592839,0.7160815987433399,0.5676850597429015,0.24244569527603987,
 0.8975717132505368,0.6254634129076561,0.23635634304060527,
 0.05326565243407233,0.5162189271274514,0.5986335533831237,0.667169814652775,
 0.4488208223330039,0.962594045670946,0.22887712966888074,0.622695151034111,
 0.6948363564514644,0.13961137272105528,0.39081777703218357,
 0.5346884345870178,0.5904183925629609,0.8550569053174087,0.5454877981640067,
 0.768310639528064,0.7924619542700987,0.930971814508568,0.8773927068912433,
 0.481416641593961,0.6787682962705336,0.8052053850966859,0.8588629399760876,
 0.7114016740804862,0.15395888930157087,0.9050495445420839,0.4370437054444324,
 0.3363273087730958,0.01248851650058247,0.20850732995368104,
 0.17158214322852672,0.5941095287036084,0.4347845046893023,0.590343793443506,
 0.21677460145486116,0.8596526761704744,0.9611663449584514,0.5557215031346121,
 0.427748329955295,0.2579769689828192,0.5843026478881176,0.7748481228401373,
 0.4420465121018867,0.8502350960741195,0.9807598810271828,0.8167011136395412,
 0.6360513623438191,0.22526048087013695,0.23764170007850027,
 0.4677083176133604,0.27808079550078824,0.1762863860154661,0.7977510058523196,
 0.5493521535584182,0.7566999681654337,0.04557791154458635,
 0.002489119894523717,0.8126902463108849,0.765843061278373,0.7527587099845968,
 0.80962055868197,0.1803980238714915,0.12768581765800568,0.34510959709232236,
 0.25841582711213684,0.7448765424327779,0.9997987722913586,0.4304239842608961,
 0.35328651682566303,0.6518635186748812,0.22751313700449938,
 0.3349844156574908,0.830201022623968,0.18671529619152416,0.813513396040386,
 0.3154037972171625,0.5967725919781924,0.3864107264359734,0.1711109582243917,
 0.3343581502640549,0.5308654558452541,0.42151987159325877,0.0780898818029826,
 0.20362214473202522,0.4981321308121375,0.5766482839093329,
 0.007985736468425153,0.21207750703834105,0.48599800549821337,
 0.5936839778578413,0.9083338502142263,0.6283537305998597,0.45142668074441017,
 0.4367511567728186,0.1028945439890645,0.7430606297446919,0.5797532014633516,
 0.951877753872206,0.5654093759502073,0.8107340828623191,0.22217085104084733,
 0.276416711006656,0.544072330397261,0.9909302470944158,0.3196400445381179,
 0.2156978952351315,0.04120477111853815,0.3409951098503229,0.3690966293576513,
 0.6476989620975807,0.5371113477357935,0.09219720358284611,0.524361668707074,
 0.8840206199255457,0.8932737252078754,0.749677339628734,0.31482618232838366,
 0.23477825014023512,0.6278519464897608,0.8497263822593119,
 0.38473467450578325,0.6893003063399281,0.08744607821699935,
 0.9576910696184886,0.3740538456857443,0.9979606002444855,0.4254278352977847,
 0.28356240695466206,0.44762057851129977,0.43049214667321034,
 0.7910238275128477,0.915037092783718,0.8505811555170374,0.6274256640518784,
 0.014869714483803365,0.8217084097818522,0.9841540830205633,
 0.2796719510815575,0.8951203399313715,0.9467845273826359,0.4860520434438689,
 0.07578741109256681,0.2815217962313683,0.3128824085176821,0.8057457684787948,
 0.1581468251850069,0.14573698629065845,0.5935829010114343,0.6874543495854146,
 0.8007368961154472,0.9586389807519211,0.47380880219417354,0.5523291946167492,
 0.3348302715399718,0.4351068566846211,0.4547712572223469,0.2781025732057497,
 0.04732489798561801,0.4544302507626169,0.2877946992154339,0.1497136133073408,
 0.2844954471569189,0.815926686122905,0.6808403125988186,0.5554884938811118,
 0.9312956738029553,0.9282167305066589,0.1877009606263188,0.35561196696528774,
 0.20970701086193766,0.14012709578262073,0.5310409258966382,
 0.8805143423308383,0.7764788064327203,0.016173047006105223,
 0.34008628976291455,0.896710329707725,0.9176724208176409,0.7807168171098803,
 0.09559316232600457,0.3008847749769189,0.6072302328676075,
 0.20030225435793525,0.9346459545359291,0.16922282119727838,
 0.5269980786597519,0.8495088299913484,0.866935584570377,0.7263719886267032,
 0.9226310925159344,0.6782138836093212,0.7695655022764393,0.9795958193884439,
 0.8365531441676842,0.9778666466196174,0.6823056614328866,0.13268339388531303,
 0.4252428454127889,0.77289971644566,0.6252011720434395,0.199137734880391,
 0.7117101511526203,0.8845995854739703,0.599788575625985,0.2865989199201432,
 0.8149728131835486,0.11332007000836519,0.7281831296080477,0.3960152387556878,
 0.34238568351608356,0.8206947384712422,0.891258408503626,0.4124853016329253,
 0.603201381708008,0.5054508352585019,0.15601028745112144,0.25366928617651796,
 0.4698755809890355,0.9254565785315733,0.5789141133955233,0.8902178330603724,
 0.20596474404527632,0.7679589529106097,0.4646430248597493,0.8954875817775191,
 0.7475835747088199,0.24499767742180667,0.2817912488749773,0.8482488373709799,
 0.171675126106277,0.6510610948666397,0.39139464249071454,0.17628302022125153,
 0.015777633864788476,0.5684139100677804,0.5725798841030043,
 0.3980308602617678,0.7678394581263505,0.6446545420352094,0.26274759883621335,
 0.5868830128080781,0.22205838053769245,0.22473390883802558,
 0.06621551201288478,0.8044893951653471,0.7989707862041577,0.6533098556438163,
 0.30954455296714056,0.6662918908285377,0.9354101026933002,
 0.07361736167129618,0.7107855482964753,0.8395357243123731,0.6024460508341846,
 0.45611373798663135,0.8263884771428185,0.4148887599614187,0.3485359304777683,
 0.0640295443055574,0.3294341241957188,0.3873651735876904,0.4435838175707606,
 0.12183399503282077,0.9047690131635191,0.9806875941849812,0.6270799267638466,
 0.7698635049564299,0.07480229676665795,0.0071387210637734455,
 0.8934135239465018,0.5031621263084545,0.3765574721908683,0.29258343598355396,
 0.4791675887402599,0.33839974416649,0.050866880299022976,0.22083253281253779,
 0.8412589976011016,0.5034959302198271,0.05375592916164973,0.5865898579908487,
 4.923370748293188e-4,0.7791987054894218,0.6070059758582869,0.799350472242361,
 0.1328222153799139,0.3886403754026613,0.8699382731934748,0.7160251113149767,
 0.8345619128414142,0.8160991750980039,0.14256851355543043,
 0.13518423262392554,0.09176215723862713,0.5689280068607963,
 0.7974709805255283,0.9519758659801,0.25178563035197044,0.4673715409869055,
 0.4524288251325278,0.863183146385897,0.17623890548795584,0.5672098122830491,
 0.29813232003114754,0.40014144705955834,0.2610097639960066,
 0.9930503882442691,0.9510042364811937,0.6746617627310688,0.6196448659157314,
 0.7221816390201702,0.5666956181363536,0.05407535258204199,0.8874477566903898,
 0.8859905970985471,0.7619344437135892,0.8280609789578562,0.05095631905869791,
 0.9694793791448982,0.9364958837734904,0.42110593714098044,0.8635092283602422,
 0.7809493239588399,0.5643531372528457,0.9085426302359556,0.5997861740157088,
 0.7831827912823051,0.28402612090774615,0.9830077429760133,0.8370722507396584,
 0.6105960119723983,0.5329222814562868,0.7759248964125431,0.4206776928405591,
 0.8482698482674738,0.48024618897557525,0.41113383038599915,
 0.6517801223788587,0.5019349530257812,0.5160791560730991,0.5941617663878895,
 0.43732567047978477,0.04934977198563684,0.2989646861350229,
 0.12532310955440384,0.6623245794110222,0.9642871322437108,0.748427426583966,
 0.8798633327448413,0.3687918133232966,0.746763201913869,0.02054873378670441,
 0.04138401839593098,0.4890212563096409,0.6973708643451683,0.8606431217346067,
 0.2862018870186782,0.4819473894415546,0.7339325692764295,0.8869436755260482,
 0.6288033123351384,0.8168790819571514,0.8843027081786894,0.7405600220948982,
 0.5922126431831318,0.6180636871026199,0.343104110775192,0.13532140097305323,
 0.7182051644885445,0.01008266644534539,0.6543268776333531,0.5810166720794264,
 0.8410570559843966,0.6037421158111297,0.5522102650350209,
 0.013823380894447235,0.032269332658733885,0.021541242018865203,
 0.9178907695675315,0.9861767382155653,0.3902046173975142,0.5801301596067686,
 0.7057390749753918,0.4213231266027613,0.36253466354278374,0.7923302525906571,
 0.8422022851466792,0.14269234055125224,0.6474498106140566,
 0.10912958543753193,0.9284177275673104,0.1816724953313934,0.4064640851129626,
 0.043376334749109846,0.26054198687477714,0.7862823270076786,
 0.324859513024887,0.23328821212794304,0.4872294501969293,0.8382179296378323,
 0.9246512911820942,0.7558683420261206,0.4470955140624754,0.28549281241757074,
 0.25575218134788924,0.9251671689630847,0.20281172904322164,
 0.9812566164985298,0.010653325040587469,0.40964308093849366,
 0.2458998099982984,0.37955072569954185,0.8450552670194952,0.714902238813732,
 0.06547237896180136,0.5251877588647956,0.11275830355974681,0.346514024675474,
 0.2282647057421079,0.9020667822966774,0.8051070620455548,0.21790062779153474,
 0.864415886159894,0.44343459972798305,0.6656324558405904,0.31661114741869834,
 0.2944848015354131,0.3176753308891611,0.2711848198281005,0.3120589666070699,
 0.7521578266860618,0.524180771352967,0.18579989285208354,0.9461199205603663,
 0.77354752795857,0.29447633113171745,0.5510839945993047,0.007913337949864041,
 0.4554365096588052,0.563048925522716,0.024780836276393714,
 0.38534540647997007,0.6030107433772026,0.6464736079745385,0.8729699546340028,
 0.6241247474999532,0.9030751022139376,0.35287279936425997,0.5398372799925891,
 0.054351444179699726,0.8468894423479674,0.15589151163207937,
 0.5874715614349367,0.06234694454417511,0.02562449768004771,
 0.15098623196899996,0.005126791259097541,0.9957930262533968,
 0.7614333911180279,0.9331999510716942,0.20126434081225308,0.9601199270321151,
 0.5453527629903311,0.8811239143195341,0.5166132175327594,0.25837921116564466,
 0.5601364027302004,0.9967491543566708,0.4606083856550496,0.839915827804232,
 0.08921372849257292,0.6731833465934941,0.8988561350345239,0.5244128831936906,
 0.8397517347713258,0.6546559750419103,0.9988925913243414,0.880309660325127,
 0.8383605719758769,0.6186931393258005,0.1386469673038735,0.9584462759166346,
 0.02974968351087559,0.6646535659061187,0.09136666243595593,
 0.5294782312001757,0.878305940498159,0.03557974190408708,0.1688330072731924,
 0.22035955457588063,0.10743351850644522,0.5569418590401918,
 0.5673801907369694,0.6455646580275696,0.4831146098848629,0.17641853768080928,
 0.7357051279552924,0.7621262950193762,0.35565736471970044,0.6329885087388346,
 0.218388882216351,0.30756267601653375,0.11020689532318917,0.895350635992347,
 0.4133397959888001,0.6110223588013145,0.874877589160238,0.8120955128769582,
 0.9337554618391789,0.9468243980288985,0.5154339933358089,0.6922277892557104,
 0.4407925078484306,0.2997653885831667,0.648319352499844,0.27515240994830836,
 0.29572143496111525,0.4207129456495271,0.2426192860678904,0.742365626140394,
 0.5743695600381655,0.5525386118907207,0.8111446139620524,0.07496961813796821,
 0.46004146243999733,0.7082828428111111,0.907417554463209,0.4997713945345018,
 0.2916770145064912,0.9582334872660891,0.09676839884997568,
 0.015649502252009895,0.21801889281830178,0.6013036103972697,
 0.24834795530168352,0.5405234177610697,0.2547924449675817,0.2197148645739342,
 0.7169565190995821,0.19454745260615847,0.785703477734984,
 0.015766223770165344,0.1890820846133745,0.4099551932912102,
 0.6174016870993311,0.9889694534147998,0.7603941151599546,0.1840373223877394,
 0.2433466118768779,0.5899025475852958,0.6853369601154966,0.8779257627127093,
 0.6669585966636498,0.5775890329944287,0.9005402815260504,
 0.034094762013688884,0.6913847443760655,0.6180726686240404,
 0.1703292410160835,0.06020786968895986,0.7072020588833193,0.4033918203280431,
 0.12699445427862743,0.46470811284397084,0.8736603661185529,
 0.6527295329799201,0.1611539521114087,0.8915863303790391,0.2914987232342805,
 0.39973391357401733,0.6533224120628739,0.44433340561685275,
 0.28149836554156327,0.5259470065708575,0.8456899145640602,0.7672240262283209,
 0.8078319536496871,0.5792169086714768,0.5747617305380414,0.4484030241334034,
 0.14496245831211296,0.5869224285799072,0.6685377847862047,0.9592985530192317,
 0.8214415800408181,0.6997935280619522,0.8431650534065659,0.9130945789748557,
 0.3404663111579267,0.7802063951929117,0.6768877658713226,0.6494258037071415,
 0.7137651259469839,0.9801256367644631,0.7481984373746989,0.7543133089498485,
 0.06554894004902934,0.5970947967549305,0.8512198951264878,0.9329519705043742,
 0.08024805538589519,0.13076685147500078,0.10339525377621472,
 0.9328008611019205,0.42767407059530393,0.8716174677995437,0.5938544089148277,
 0.11599355530206668,0.8378344390913034,0.49107088131924037,
 0.6960569229338069,0.3544223650572331,0.05482222480266252,0.3647701776074316,
 0.32688096928661126,0.16963951474019967,0.7255569710380084,
 0.9670071740144645,0.8755796038924116,0.616470135866617,0.6247882699608484,
 0.5311062589938478,0.7547831671306979,0.652308386144219,0.9387617950297236,
 0.9575558499863576,0.4582965740270839,0.9721322373939203,0.9279080350005053,
 0.4534232151612213,0.03846887697662593,0.811394966967753,0.9294408592548933,
 0.4023550207464296,0.2616487699964728,0.7264625417332765,0.6939640964516143,
 0.4645644729294125,0.005268579167318599,0.9854793251073556,
 0.5880481572463727,0.7952854911763315,0.8509057858614885,0.9168144860493369,
 0.30033116287590667,0.0190506188635583,0.08956690352978003,
 0.9442986723565774,0.15416793341179724,0.6122100179369451,0.0732171505079422,
 0.8807516850690307,0.2141926491010796,0.4754447812406617,0.9351534678509943,
 0.5181726927514074,0.7901866887493574,0.7645150418007773,0.14116145762121857,
 0.27312193016805086,0.24759535108096797,0.5638721299407695,
 0.7418412006530941,0.03196745265129397,0.958187905200664,0.9663999010170631,
 0.29924089177491764,0.8168290182144271,0.252720311373953,0.26532636890261185,
 0.6445967625260076,0.5756212411740438,0.35580906624875197,
 0.043903583965189674,0.5646850435329297,0.15665442413091624,
 0.1878513206451652,0.19729644603329033,0.9378351999350756,0.5525323506184247,
 0.5417501493245158,0.06123662095720728,0.458964528270829,0.7324733609867992,
 0.44420862249931425,0.04376933447285836,0.9831535930193794,
 0.7934732135191886,0.6026842087135454,0.12541916659598162,
 0.23151766463342416,0.04928861713023658,0.18442450156695422,
 0.11726427204657919,0.29686872844724,0.7934495053087889,0.9935761262936934,
 0.6296437444869922,0.9848909349199507,0.13904190033282848,0.5267778021262316,
 0.7440823674497874,0.30727698261757774,0.5295276828360423,0.4142782744021001,
 0.20889542118865934,0.9060689208689439,0.1010918265756876,
 0.48777143774909926,0.11899867054405944,0.24257008863676577,
 0.7205258989483291,0.07671517789091076,0.33293534582649464,
 0.8203392672183434,0.45386876526998343,0.47768378618996554,
 0.8758707450204503,0.7372530615399856,0.5727092771646793,0.6930788741915213,
 0.6891320532439807,0.06691776043269015,0.10118850793649181,
 0.09645437778601795,0.8466519284858212,0.8140065125435604,0.9119830060056118,
 0.24751060341424602,0.11199094463872439,0.7495961922183176,
 0.16788537770573564,0.1140709774751596,0.02405461543153231,
 0.7944814761492545,0.18883153186974055,0.0653746757318262,0.8746198123119009,
 0.5254095720713009,0.44302487832091064,0.17942359658822316,
 0.7198126168592253,0.6479295567876968,0.5300011331831629,0.3507969853060605,
 0.4390917038801474,0.7291941447887201,0.004342871109241253,
 0.44252116794905383,0.789355368189243,0.4282167014297188,0.4056855362555828,
 0.36608935046216917,0.7785911587316401,0.15520528951223223,
 0.001051266716161403,0.6913796397766157,0.6011451289205839,
 0.8766090827367042,0.6246784282371854,0.5702353984706561,0.3391948114968588,
 0.0273505262531758,0.25251048895563044,0.891389669439223,0.8203620473179506,
 0.8149583030060429,0.7676491759281152,0.08069423465504011,0.6694158453863593,
 0.5677046946742914,0.9879704583592114,0.689414358458027,0.20132418589358114,
 0.30294173626599263,0.8289663511085547,0.11821776263020245,
 0.10860452007661203,0.403916600182891,0.1973437170687935,0.7666941461892616,
 0.2012102586759429,0.0864677209347805,0.984979668481981,0.2125808094107453,
 0.6403263018479521,0.5363729427938746,0.038795324375028195,
 0.1330166020306367,0.2819979461662657,0.41333179224879146,
 0.49933550184093334,0.20942281744171642,0.7708112081356068,
 0.14510329307476555,0.03313478596989228,0.36551448992605606,
 0.13885112404615696,0.9264668025992853,0.8688717704904058,
 0.20532130196509302,0.9527743005369494,0.10834504020113322,
 0.7543068021905833,0.7251798784791028,0.971266239018072,0.8315230011101858,
 0.6167496880666765,0.20191959474076204,0.6994988320300539,0.4307994113082013,
 0.14001426489303304,0.32936961866055625,0.5593519173759135,
 0.4243994082431083,0.15249795255705156,0.9406861667449602,0.3265704558824549,
 0.4954490195883322,0.7766944251464682,0.521145766473776,0.2186443025575734,
 0.2475321898078534,0.6532542625885618,0.8913317795291166,0.11405496223302525,
 0.07848374081342424,0.43133153354547704,0.8178636993575784,
 0.3866747712270616,0.8286841242916804,0.947048592607518,0.6928787606956054,
 0.3612297432999241,0.5204740074531088,0.13715067950965565,0.5983906943375227,
 0.5812084303808579,0.4456286103429655,0.2639807699584752,0.6863180053880324,
 0.7043145584227553,0.7883129073243671,0.8587025708114854,0.27582229298510663,
 0.5328445318483641,0.4760361142450551,0.44468960699776405,0.3967456147477233,
 0.8007128865607214,0.5606352779038324,0.8933329513606582,0.7483098033406012,
 0.20215414466645043,0.2895903287299505,0.3179071190823781,0.7499359464165912,
 0.9500232839430294,0.8397976094123354,0.4192241293046661,0.38726314495586955,
 0.16092400760785286,0.8020871064251044,0.05051773610565069,
 0.13724983918064781,0.7934917224159339,0.22860290293677987,
 0.8375866142223876,0.47577463960609057,0.21484906951230032,
 0.09513973123675656,0.5902129855621038,0.9360393009395656,
 0.03826610529027319,0.8990059420526503,0.9358570861125245,0.6731104824349445,
 0.17135430005920504,0.3630942339716303,0.2957439990187918,0.3118546313844919,
 0.020698555522667195,0.9313033148455376,0.8610090262374598,
 0.4887821269477287,0.6323908242286389,0.3965512400116833,0.7916061601853839,
 0.34865109726449806,0.8938384198838137,0.05314881293083129,
 0.23805375577563093,0.7137052148484879,0.7661561265020496,0.746279337754976,
 0.142002410546901,0.3892921162940357,0.640799012545194,0.09832716931641017,
 0.6467541826802377,0.05172459150515962,0.5538698770978175,
 0.47934462117072174,0.7584114981807265,0.563459890027053,0.3443913823911523,
 0.5193264798735807,0.43049091757671254,0.7685353832685349,0.5774533360089663,
 0.6226380444782901,0.2439362098763499,0.029864959054318185,
 0.027517035680881263,0.9034967113476652,0.9624621085274068,
 0.9452343207384143,0.9706410433343176,0.0968221047422061,0.5030189515268664,
 0.4009436653168319,0.23657288246331476,0.9649153104537019,
 0.43048216469412426,0.977386027020971,0.35029842305646,0.9223128582295651,
 0.24272462298862607,0.6060036921626746,0.5358025689542495,0.600597905135144,
 0.030186009653112578,0.29838728223605526,0.24583477428167966,
 0.38844080681424387,0.25642746920954174,0.489014128891348,0.809218722272105,
 0.39742161236882845,0.6086085472574728,0.795561619537172,0.5185053257932605,
 0.7507653988306469,0.49892945816662193,0.5756330492063805,0.5023963110663227,
 0.2671882446490599,0.7953153154416985,0.12859256539528652,
 0.40061144877493127,0.2665400912637388,0.8568223492068497,0.5293857914806802,
 0.21405153179128866,0.5824369691761966,0.11016633331378511,
 0.5250441513659639,0.10291242485250052,0.5770826277418791,0.4813268234398098,
 0.792366363508307,0.932269546293617,0.6889851533785002,0.11220863899443367,
 0.7130911813674072,0.4393711555624801,0.8644515144319663,0.8826975797020399,
 0.5446879956239874,0.7813238040058658,0.07372184817234684,0.8066671167209649,
 0.25217498681741746,0.05178961597621057,0.595746747789613,0.924999795313004,
 0.3152447602040056,0.7874603554057881,0.6541123198992398,0.44553112597643874,
 0.7690048430584769,0.2924036341021372,0.26602142256739403,
 0.02748788465468177,0.47434058446142513,0.35533816908396443,
 0.8712312593906563,0.6251593569621997,0.48723577115142425,0.9630977219292669,
 0.931391467987985,0.066649773072055,0.0864732100606912,0.599882198264459,
 0.461908312568593,0.625002334914369,0.7704953922978736,0.7581328205917369,
 0.08537882922018669,0.6260313174562692,0.36417505596426447,0.910363217571345,
 0.6884032706945415,0.5422016152775399,0.7111987741077335,0.2591939653404979,
 0.3050571314225846,0.7839109886097038,0.19534093340945957,0.4595522752406058,
 0.7256671950665797,0.16107739865742365,0.8110351738533312,0.9039518238850663,
 0.3957874845700904,0.421437468683853,0.6301316342470229,0.6008257922559505,
 0.45189220657274765,0.937506612804415,0.3329169240379203,0.6556725137481002,
 0.26843870910764256,0.16741208540488295,0.5021620317428014,
 0.16296590957594648,0.5860752929026981,0.8492896134426738,0.7255762379474628,
 0.7003655276613094,0.33265758020111125,0.9565929946558089,
 0.09818483304149783,0.6676325169156216,0.8250651871255219,0.7522552647613464,
 0.3356467852511821,0.9532724064183572,0.14929547548628241,0.6257267863420476,
 0.5655359494757688,0.048527547237772195,0.6107627520382014,
 0.7000560273627948,0.5205571652509408,0.05390634103703518,
 0.12812510602963334,0.6983032279131006,0.8950244682618202,0.6508309791380339,
 0.23739988152810299,0.8864802960483715,0.1228739572021662,0.5265561729573257,
 0.39575524851792376,0.4240957201124538,0.665199796236029,0.9627556536076415,
 0.8734327594022845,0.3929426917105977,0.36111078995343926,0.7899605796698448,
 0.7339030745153861,0.8384636916641669,0.8079418571942805,0.7305072984648767,
 0.6100162499206331,0.3067255041060333,0.7401847316520638,0.6112167527944767,
 0.4512496275391378,0.07766267417628447,0.48669077832786256,
 0.46671889680482703,0.5188003523823578,0.1115805441473684,0.7682706642964165,
 0.9908753310110717,0.4222783612244513,0.017904048683499485,0.378418536344987,
 0.13769766076837353,0.10300102239287146,0.2041498366431349,
 0.33898490320169783,0.17450934095084203,0.2638319470429846,
 0.056032570293661155,0.7109910992718189,0.504271881642865,0.5120438366506609,
 0.12298759625165712,0.8780620196942944,0.6339208292241584,0.9630669088019326,
 0.0795781028449899,0.25677743686807575,0.15425744554663556,
 0.6688269526317607,0.8380372735610312,0.07761415305067554,
 0.06531857136133778,0.1607320896461918,0.22835869566644296,
 0.2924610308301131,0.4653612432895158,0.29923695275063245,0.1724371068971915,
 0.8490417480828979,0.04935812021224972,0.21997284655416482,
 0.40365581227363556,0.5690222562704771,0.3631349013357027,0.6379769811460414,
 0.46843697781917726,0.3224617114483954,0.3002793617303513,
 0.032574655059702584,0.0735710699438441,0.3380187992197151,
 0.42656161673482484,0.45010877330374677,0.15555051986654544,
 0.3540066670545915,0.14118313991270703,0.3752278624191365,0.5441094392511012,
 0.9931428795324575,0.8043486410838999,0.802823257227938,0.41540939409540434,
 0.7164257578531156,0.9789089179561138,0.3576408837953555,0.611394176393248,
 0.3463652520944783,0.29245596411930164,0.9437265063009036,
 0.42199639390837584,0.5229610088846249,0.39708946512320753,
 0.08389619382078861,0.9426944051923762,0.8159817700467428,0.9296636461574914,
 0.15482114978212314,0.28828801119272696,0.6419289660630543,
 0.2295987901858776,0.8601162264587294,0.011624095576328597,0.586864173395347,
 0.7230335133403296,0.13579051658872654,0.6756290072020237,0.6891150808153844,
 0.8634768704918065,0.33657358020950523,0.8118583281132098,0.2858092525561654,
 0.42400919497817136,0.6796443153155745,0.17086857736239147,
 0.8961138244934369,0.1641717326441232,0.9415441042171365,
 0.035487498855984456,0.9127181582160808,0.7247437077860905,
 0.1498314823164747,0.6402383059048959,0.021251088284642505,
 0.6651309462028353,0.8652391783170535,0.8744289601835065,0.14082056462740133,
 0.6121239147868962,0.7664320479775847,0.647398791042928,0.6748190783619722,
 0.6470167100668048,0.33170651279698116,0.3253673760982121,0.3151740523909643,
 0.5819057975533701,0.1367666419748198,0.7785571737059167,0.7494298676376452,
 0.0769999223608242,0.752014304882094,0.0633608056971604,0.446302285484371,
 0.6665248208706525,0.05534143475833664,0.07901056622934699,
 0.9512542984218295,0.5134599067420723,0.2446377704568682,0.12429650456234453,
 0.2634721037028054,0.5324995040471858,0.07722105466198848,
 0.48537560590784423,0.6781075602811809,0.07668014494887954,
 0.8641557821931602,0.6036111310513548,0.7734457919914312,0.6878994414594588,
 0.31034049881669556,0.06337967515527809,0.3109138351983609,
 0.5608615695087019,0.9664252633565416,0.313134370242513,0.8893513938535448,
 0.10791810621095832,0.333269924581745,0.14045405981014913,0.7589757419420068,
 0.4370573125867292,0.8932864335297418,0.25290670810632954,0.69844200302681,
 0.035715769596530667,0.37737893655257404,0.13504526128377747,
 0.5247966746014338,0.26216613682573453,0.8808572734776045,0.8882841875283991,
 0.7335936613733717,0.31871735858993955,0.8704916196940226,0.9502152907256866,
 0.4647682027719835,0.6885886330708353,0.7624200532398177,0.9613450459491227,
 0.895917816778329,0.6365718984875248,0.34786339607016004,0.9714764051895851,
 0.8682285286244289,0.6401680426952521,0.9022866821038735,0.7855749162678736,
 0.6346770733167313,0.25824300188324856,0.3229461028135707,
 0.36396986994308345,0.08689047888767609,0.5517430853790949,
 0.44870262079532286,0.33203048901028476,0.18946329250720662,
 0.5732091228009526,0.33392509245711777,0.39746800316748143,
 0.2873007052399481,0.9743749007642208,0.34026270255448376,
 0.23854937421961686,0.8879682415395364,0.8184308574376606,0.6576592268444325,
 0.48031621131112023,0.2213288853826968,0.8961759892930594,0.622394960298098,
 0.5722646674844432,0.762156424605271,0.40012525810209465,
 0.025740852144592985,0.19816576869137648,0.10886354599143122,
 0.3364383061438657,0.443232480778322,0.2992201831747522,2.810525399954411e-4,
 0.7085196155583893,0.5286851476542889,0.9915788635550856,0.8121318424539045,
 0.788712784886629,0.37313460825901934,0.4470390607045861,0.9431781898736041,
 0.3325883449303668,0.5263669267999261,0.7847065691588355,0.7002654329462605,
 0.4640139990176223,0.9963535377014868,0.9793436479446104,0.6768492156688732,
 0.9622006413162733,0.03040748236966606,0.06493059898516895,
 0.9916695485602423,0.06362212389163335,0.3760266352502126,0.705182631956268,
 0.5878667204745373,0.49327123918798366,0.39430580860225906,
 0.5095510298069069,0.6061673606171806,0.2227158232833808,0.12612815543908673,
 0.20998428945055658,0.82969719676069,0.9160154980986647,0.9688553345122678,
 0.9676912215604401,0.4578366417560801,0.02150572752926061,0.8529505189359077,
 0.38823354345795247,0.26533708364975783,0.3759933325803536,
 0.04291653159833986,0.19280236885347002,0.1787580557413725,0.504868571508304,
 0.7855664037012586,0.3862866985089549,0.0776572234412668,0.7928878181386093,
 0.29527703554076146,0.9500097535239929,0.3397742202490539,0.8917034089017962,
 0.18089086064192827,0.4015435278442818,0.6180922873704826,
 0.18741936314407293,0.1498220360469419,0.8409040749060215,0.7363925281007534,
 0.7863856990276314,0.4661465929270767,0.5996719496693304,0.7697973576687952,
 0.05229577247107353,0.45613805977333877,0.1773523867888842,
 0.2556583413705986,0.44096659887132594,0.7728699328486358,0.9504444036644697,
 0.04642245367131004,0.7155597688220329,0.8906819854123476,
 0.47010975801041865,0.11807824928851174,0.27629813995030106,
 0.9247012162256122,0.34668536515257586,0.028197507254100085,
 0.6882847168939449,0.5814716954395607,0.6956077320630366,0.6479936718017157,
 0.5823223641657408,0.8141024489697665,0.7424044637205064,0.12691393806118345,
 0.6740670643007284,0.9787150440083747,0.9240821847409698,0.5733030228909518,
 0.7201193386601433,0.7553660421969951,0.2276778898071784,0.26756685582926953,
 0.7006629757536068,0.7277968225319624,0.2973563930039125,0.06653130936862306,
 0.5293412333556031,0.8420071571797825,0.1366057523219456,0.9619916645654327,
 0.6197381574869906,0.8586049472258949,0.43623440479099296,0.5115558859111449,
 0.13046227286207746,0.6910855794954618,0.45542206544548614,
 0.3704568708156474,0.4059934896174321,0.7739227283717637,0.6438050753718523,
 0.3676963566899949,0.5797682930330595,0.5179846490102673,0.8615512113687349,
 0.5756903842023822,0.3464641936368339,0.16081468818975342,0.7218860307246552,
 0.742133289116135,0.2855675158157336,0.7825609343084273,0.605543339688231,
 0.8968031973938527,0.9739253357062704,0.11197193191221166,0.6233554919035798,
 0.2141620550812483,0.8391081972645618,0.6275415613113329,0.7434567382667644,
 0.31797828979151144,0.9481256261261758,0.22555719084568182,0.879585004137968,
 0.047327365286275125,0.6819423113871633,0.812174840810767,0.3050793871977495,
 0.8341780089027415,0.46911283825534267,0.04795266868969017,
 0.1731084590381019,0.07796967093832585,0.5439119200594452,0.7870628718110648,
 0.10826169797693774,0.8222495979940846,0.6568035591222854,0.9109599031263764,
 0.5494056559374491,0.7588503257305491,0.7034702967531247,0.19058559744579429,
 0.486194498964731,0.8633954596465567,0.7628348857972116,0.9970816072109819,
 0.3177005369614535,0.4990788399402425,0.9746427507631881,0.23591329280987883,
 0.9998527958215301,0.9873557181075702,0.6041296897962805,0.18720599597957643,
 0.44252185671441857,0.7166026211617369,0.47898819946682314,
 0.4903692029161689,0.9744847285349557,0.9333791270787306,0.2780376458793017,
 0.8288856272436979,0.9994598355504204,0.5264789420368697,0.6019376550476468,
 0.6601640415839445,0.08583758049285661,0.06343999926402744,
 0.45837647603441767,0.10852133359923277,0.6665078717480839,
 0.8448674724627453,0.9827296160889807,0.8875301328436124,0.6650773803493529,
 0.5447945501463478,0.5442257407949715,0.6176518512614089,0.2897003318739454,
 0.7240925615875354,0.7632604926055161,0.5179689248566242,0.8015232768714681,
 0.8525624683502806,0.5577929823135027,0.8931227741251379,
 0.024493490362299974,0.6453692699416294,0.2870212041533353,0.750147791004484,
 0.09146741310175666,0.8714193624325365,0.36660782952770765,
 0.37710358856071347,0.5481043274367209,0.7393572828278305,0.9315409704364553,
 0.12885989757656036,0.9570870804019886,0.10332315082713761,
 0.8177489751325436,0.9222824016997304,0.9428922214795553,0.600963000955976,
 0.47943543920942644,0.38088640519431816,0.9043021251473613,
 0.6607354062053666,0.24395783329177823,0.7403678651710844,0.427942124711967,
 0.24017251307927345,0.3925700224644372,0.22041804800795906,
 0.20479427816048168,0.47631748623235515,0.374183681427365,
 0.22458252273859458,0.1920119807795082,0.6039566966077887,
 0.0013764066907714367,0.9942314362462623,0.7927447933021952,
 0.6491612140680983,0.19411499337668647,0.5208166882188154,0.48304875141349,
 0.8842007920308474,0.5655779760866346,0.8301410791308089,0.3813744752527801,
 0.9258738615969131,0.13122374853402408,0.401523875284978,0.9550176727353387,
 0.6751679181081163,0.9419737251239112,0.5353809048063134,0.5604215262907917,
 0.5899317140302517,0.1969426939976332,0.23488368867979215,0.5612743146249323,
 0.16862709882479088,0.41458751224663604,0.7416624628378884,0.71164088892303,
 0.03989162751260378,0.46845008526549936,0.46633686697067267,
 0.4725271350895863,0.47266845065189544,0.07332961523727588,
 0.2085077450456284,0.18455102036870252,0.4310957275436037,0.3792702968942028,
 0.7463312083580806,0.8145908308489397,0.9417654590897422,0.4749218479810694,
 0.4940305266757258,0.19856921534969296,0.6422965422386404,0.2840589195564931,
 0.845840097954772,0.5257895657500703,0.2720060245065341,0.42935357178234557,
 0.311466231459691,0.20167896414131825,0.29456593300571077,0.7489858956700588,
 0.48855144091055724,0.6373469404002576,0.8471162952820527,0.6193648854359676,
 0.445237696788354,0.27517779715443935,0.8559411497822194,0.09416542595924537,
 0.04803591498207449,0.24318015508234014,0.5408906807681915,
 0.9089387991573695,0.69844696307431,0.13419572985689265,0.6444628611582274,
 0.21223352036957732,0.980335048292983,0.08277896667032336,
 0.43636560331491947,0.19942088108538525,0.9228666845502508,
 0.8539817626493532,0.9670439292865227,0.5695919082045515,0.07580173061682194,
 0.18044816362576732,0.8906818660956062,0.7616417092237616,
 0.11668735712509426,0.6024506408089677,0.4888369253616308,0.4619799245089957,
 0.21912446648311812,0.8227492597320866,0.48071075359611337,
 0.07838052897797942,0.5058518301235557,0.013065133554653996,
 0.3177211731483598,0.7808013300531923,0.7379099245167074,0.7982427894428273,
 0.44390048747830313,0.4284740492558107,0.8986806816250408,0.9897847544809609,
 0.5249879582866732,0.1057726388388005,0.6988457869693007,0.386597341646153,
 0.9004960985347839,0.6130312191615054,0.7302296899637087,0.2054277720329264,
 0.6379279811645376,0.30335714492768395,0.36156334858643857,
 0.9237901970249172,0.7306605156471784,0.6800431042172757,0.508503207001581,
 0.5464882807510475,0.386591902155366,0.14590232848928175,0.31178208555314957,
 0.7112240933724924,0.10412516704874164,0.7293862636148882,
 0.036009317169132804,0.20827542442877667,0.5969669242536623,
 0.07898489812479825,0.20313831342560285,0.617919274145849,0.8516229735193083,
 0.4520018058102545,0.23112587802650952,0.7187595294715704,0.961211934173503,
 0.7126060933851494,0.6952100927971776,0.39495113730995546,
 0.15409538009576096,0.6206332764761291,0.28508142271164183,
 0.8008556590020701,0.1202629052352957,0.8713121045315936,0.07878337198009289,
 0.6418217322953343,0.4536971331188364,0.44990022109388805,0.1169256860686344,
 0.5068730678036706,0.30207985707033425,0.05292766149371997,
 0.5189940598457157,0.5997470632211912,0.949002841522473,0.018382452259643856,
 0.7640792760502138,0.9569296588940345,0.6224204851505737,0.3897322768667316,
 0.5841377817030126,0.23397913040624452,0.5531152286707696,
 0.46449360471059464,0.08809506357244656,0.6335429098203961,
 0.9599671588438153,0.9585230500469661,0.7211289913674122,0.9020489604873031,
 0.7713431450849194,0.6023516955193324,0.11657031791228534,0.459813767785357,
 0.47231581426324754,0.1430650707518757,0.5329553493334865,
 0.10761319013496862,0.7892258094954645,0.9659730350439243,0.8986694087292323,
 0.674546061961614,0.5749070039116754,0.5868911610145328,0.9912270477623788,
 0.7969407289769513,0.014801328253442048,0.015820945158665722,
 0.8340592214408955,0.843242899699184,0.5894214604172584,0.628324680621975,
 0.7251162860403253,0.42603546572042506,0.46356458549921764,
 0.010460523848567993,0.716581574864834,0.06983559911827486,
 0.8254174070631586,0.09677744351456119,0.37751921255993603,
 0.6188592317116728,0.6122485764745,0.7114276817700144,0.30751781377525733,
 0.08464544698772658,0.5325080465362886,0.4634848786776614,0.9451650738074614,
 0.7859500785930783,0.7001266053026745,0.35051574287205567,0.5375926189855588,
 0.25353195623292346,0.4009370232305316,0.9430319952736032,0.430426345550593,
 0.38017420772193244,0.925100560952907,0.07426495743410455,
 0.045448660875847424,0.45089252791318046,0.030139844485422396,
 0.7983287506742459,0.18166388223723262,0.6782936050767562,
 0.035963528953583346,0.04863483893536147,0.12758149415739006,
 0.6078502485356964,0.1995022652789472,0.31810756840551746,0.5858980205754849,
 0.12508506775301298,0.5233313046372811,0.6407723239207724,0.1956428231215167,
 0.24748574610992258,0.8648044746980081,0.3024025124217149,
 0.28984631549001516,0.7954011455768939,0.12120612772677075,
 0.4509489617444983,0.17925941437542048,0.5269310661558941,
 0.04987301808443123,0.35246093834630166,0.048067285403930904,
 0.6634759121822317,0.36341142665115345,0.4407989384467882,0.5618942537783763,
 0.42714142852252834,0.23091912005336612,0.8011886113365885,
 0.7478887113210255,0.8636501275450191,0.6319403568891073,0.3396743528409344,
 0.8150715298884649,0.8383535469871592,0.5644771262400222,0.43374343797278136,
 0.23752020405124463,0.6342559686095041,0.8218378354798962,
 0.26590191841742483,0.3177129707768701,0.8945254713065179,0.662338487609842,
 0.803143755884843,0.36855200215363404,0.9662986786547731,0.9268806188982179,
 0.39196252695662437,0.5127510966537215,0.6008551926913412,0.7914924923390897,
 0.888824380971312,0.8034332848856115,0.9435521215780259,0.9127611110057632,
 0.8522631590889923,0.13599106812102624,0.8317869215012661,0.5785145909331316,
 0.969615840995506,0.15238920242741094,0.4656805943918524,0.789742502399039,
 0.48135634405122973,0.7209902519866235,0.2689372308169836,0.7675918285475931,
 0.5221774562980165,0.6123785586373468,0.9300194685368839,0.5995868901886289,
 0.7097575336539559,0.540700318000415,0.8062971364028354,0.6217554716265039,
 0.7208620105596318,0.21266922777622832,0.2101361071973188,0.8894405519822185,
 0.7050439484077985,0.5244162840111007,0.9831037478632604,0.9060881228804846,
 0.9866516212850738,0.059315750047009286,0.9859301082868503,
 0.48962344682992776,0.08130448830449899,0.2538009941883609,
 0.45407890496966574,0.6280177406035391,0.2366652908263024,0.2645007670043755,
 0.06658948753163241,0.505298944134416,0.9411541959609766,0.15677082742201343,
 0.9294381067645252,0.20529950105392547,0.43527090644847843,
 0.8169367035666977,0.08940647089920584,0.75336734155517,0.2689161288511934,
 0.019029248898418394,0.07559193289872779,0.640225738023843,
 0.33946143023700714,0.8473662100504279,0.25247271048843967,
 0.8766304116357786,0.18868938570560423,0.17300233576257618,
 0.005062258910121131,0.3538283808656131,0.13577689617108812,
 0.9260350279857267,0.47209555544095516,0.8775067375051203,0.5915148569218365,
 0.03774501810964037,0.19708476532212893,0.20745452864870018,
 0.5687780007738892,0.6674556505511187,0.4261202340811072,0.7774440970268918,
 0.5719894056449574,0.8299483213373925,0.05607147097456566,0.7211732957267932,
 0.21316906871950703,0.2659198890476777,0.42954114839964697,
 0.25887555644801774,0.22969506011473784,0.12335912753431177,
 0.20515629604805674,0.9012512496171499,0.13421805793040265,
 0.7691983550581061,0.4459530032974188,0.9428024892175019,0.39219910114108114,
 0.2804662218018663,0.31391579695844474,0.995121285418461,0.49940857397516525,
 0.5640211722555183,0.5924720551201955,0.522420047665242,0.42067013248710916,
 0.2798459443444117,0.9130047487358081,0.5830927638229434,0.5464891508555464,
 0.6581657042824098,0.8879786907665358,0.7418576827403323,0.469326656514252,
 0.0045290062388254615,0.40895140785510553,0.6208446821276377,
 0.6980942221046148,0.1140431761506171,0.05940279750650146,
 0.46243937201793006,0.9234813093700277,0.11311354648827565,
 0.6813451453152863,0.02322528644803279,0.8176557313140643,0.772093052598514,
 0.7814828452435484,0.9102800299758612,0.0883067113721151,0.5118635065737401,
 0.5772606981163125,0.4943370165902753,0.21731980794493144,
 0.05718568040688632,0.7334971063764131,0.11356820005378754,
 0.20899598057159718,0.03084367631800533,0.5536028779894099,
 0.5168851429183697,0.13150695310419402,0.8211967633263593,0.9729063268727058,
 0.34188456473895956,0.5193595016141799,0.14832278081662253,0.533618008538969,
 0.46143652517505873,0.23930826393073434,0.5279350213029371,
 0.5830329775321353,0.40217281977963193,0.2166196785364709,
 0.23446410805348217,0.006929771073563407,0.6388208525442909,
 0.4749894381322075,0.6671188132677952,0.7326898140869609,0.20447949977187796,
 0.717857350359019,0.8630277423446228,0.5860376239085676,0.028861553440232668,
 0.5360607177632073,0.4670446005179162,0.16050261020607992,0.7163485942184482,
 0.57727642394594,0.9895663126222799,0.685519118990085,0.45991649065702145,
 0.27229188651630454,0.4588583146885389,0.396291185694,0.5818302869635135,
 0.3947968016012944,0.34090288806990476,0.3429661309989892,0.7682285804430001,
 0.10963194699956891,0.8252268430060266,0.39516190763665915,
 0.22078338705031286,0.6857932782097664,0.14328147350133047,
 0.6115975134733037,0.4448723740122762,0.49608810338878095,0.2167371622131793,
 0.9096925056710393,0.10018740830736639,0.4109146762726109,0.7374393210298906,
 0.7632798797568072,0.04240653050813348,0.2009646335367865,0.818011811749418,
 0.6315012217453704,0.9842612817688369,0.32186791493842704,
 0.20779113629764467,0.46646500625666165,0.056000840157156895,
 0.9282740364598305,0.6647921841652027,0.5773668450897856,0.468027287219287,
 0.6604561069651048,0.024762420404033425,0.9886375977296087,
 0.6145672773390023,0.3885883724763939,0.6656853292531668,0.9424634712055513,
 0.3988664418814871,0.7653492726757939,0.025576132138681595,0.530883528741209,
 0.07691592212886111,0.282797425522312,0.794912638750993,0.9044984255024973,
 0.4447862048774175,0.020683049406392096,0.033722625246077054,
 0.8565790167215188,0.12939208943005043,0.9606384102399261,0.6136251352331136,
 0.700471545025231,0.4000803639031475,0.06452728909853545,0.5252054063742413,
 0.3100630066738985,0.15421658692712337,0.6843128206562351,0.8917110925390961,
 0.5007104060697034,0.8047927334072829,0.8354080809422834,
 0.0011776260935772997,0.1770410683593897,0.5635402064752835,
 0.5976531813482395,0.8537421236892211,0.6461391370955087,0.7121444822811096,
 0.1168213581724753,0.04479744836496635,0.1880448854216492,0.0764171507696102,
 0.7674593356238008,0.12780302736920157,0.5013654279291728,0.9095556630984691,
 0.8279150984585472,0.7608500364632165,0.6579433786737421,0.20314999835497505,
 0.8802238638862449,0.08327700252755477,0.3748391337131354,0.5331784460737061,
 0.7567097181868196,0.7098871546773694,0.7221008794960491,0.6024060394926036,
 0.7746652441142616,0.1065304233893164,0.67796081804089,0.4954336193148706,
 0.5694219369438092,0.6964157035128388,0.42176376657253134,0.9229633461360218,
 0.6846512945577515,0.27160232332345524,0.1626526848525175,0.7653608023824248,
 0.5955314628952021,0.18515317689327504,0.9221562627480115,0.5827833665225948,
 0.2502548582436085,0.5251868283903424,0.7856634007117775,0.7556283564539398,
 0.2618322001459725,0.3007283482066806,0.2961035673923328,0.20114541709472067,
 0.3021100996592432,0.861487634427582,0.18268891550230282,0.8278007076640038,
 0.6908084378086115,0.7064465751193395,0.048643590880454,0.03401703867590744,
 0.5840882700120331,0.7659550668710654,0.855986931410931,0.8786469028582182,
 0.16605481391114596,0.9361061242348101,0.3496523955488986,0.545685205682287,
 0.8416591346246541,0.7513593254547145,0.05432408777424391,0.7606657865720702,
 0.6753740561709811,0.3921251492508777,0.18919982997957796,0.9201264877208802,
 0.9875617842236377,0.2700475833260332,0.45205617393255804,
 0.07362004759302843,0.7453424619879374,0.11481984427367375,
 0.7794194567110299,0.16491286232915425,0.7991339560450456,0.5104588929425428,
 0.059762058789293415,0.03393641609643505,0.2291552575771072,
 0.039334448969277735,0.5917709377121451,0.19758573528605017,
 0.19998639732782408,0.492206048096847,0.31866250148027264,0.6917699056953581,
 0.020006712413897176,0.897617765977245,0.931146722465871,0.17354897759076926,
 0.8458821035804224,0.6085789234212183,0.9365259608072141,0.9190288943204822,
 0.01097896436378265,0.31532772935523934,0.48363066973791413,
 0.7232907097243971,0.30354154990859306,0.7540617479976062,
 0.40205502094143497,0.39328165643667035,0.27798828595179237,
 0.6787581347137048,0.42115559212450027,0.8157808905329645,
 0.07151066074848333,0.24370675418105414,0.5095320018763342,
 0.08696863371522923,0.8708513061658167,0.7925158161407202,0.6131884409416557,
 0.18449870709099325,0.17626901167660092,0.9737393854173356,
 0.6141188287350802,7.870373071534598e-4,0.2724375922251969,
 0.04152536223508929,0.8554205273151682,0.2610511335371859,0.3813119174812891,
 0.8122810251260337,0.26309976883715036,0.99307528423412,0.22800690032749804,
 0.8131845202165157,0.8913894491616866,0.5573182288911758,0.33849398455830815,
 0.10454506641271077,0.31741838996241223,0.7656763659784208,
 0.7983910589330129,0.28697187749879316,0.60235137212788,0.015441731094527622,
 0.6807685923691008,0.3712328806336408,0.18111057385593654,0.4736566192139531,
 0.362477542619428,0.35024775447284473,0.9200539729135426,
 0.043562666743505285,0.1937930921464257,0.5115595394840488,
 0.22665058218064127,0.37238473810542205,0.49893824520059504,
 0.9353415556154843,0.2091097734624603,0.11982983042636108,0.9625546674998544,
 0.48561398884834905,0.13098600361394275,0.7130230841972106,
 0.3081642683193492,0.024555919957635952,0.7146441979231166,
 0.9907028245295671,0.5313994771589766,0.7664301731948882,0.8375939473819753,
 0.9224055999857961,0.9935896523562593,0.2728951235721,0.32458258173546983,
 0.11697930259859801,0.7004172331171193,0.6211144352227784,
 0.20598879940634318,0.08259916164420433,0.5822366962672766,
 0.6188403582387161,0.18192030980282936,0.95152526916659,0.7928437724444993,
 0.8648357258429747,0.8103024401494046,0.6950351054335333,0.3135022124735789,
 0.541253886191837,0.3700409082976278,0.7811249953889161,0.9205221204875349,
 0.1512236447135472,0.8468489820081821,0.596786787497078,0.7206373561503364,
 0.5399295849675947,0.9935884095178225,0.2000594076340526,0.8039055197498949,
 0.9012863285826747,0.82183566078316,0.5950262018947222,0.024257838514798236,
 0.07603536202776906,0.10990147787720561,0.140224374249161,0.5993778575050127,
 0.28306058537996737,0.8545073757820427,0.8232191886989922,0.548955508770655,
 0.6519482274742806,0.8726954999812503,0.06168841154277338,
 0.42563876997234906,0.582445006766539,0.4724591165774903,0.601142625964692,
 0.8877444253415497,0.15012792210082204,0.13497201212926868,
 0.7264310380352461,0.9662263562601443,0.5195196698162154,0.35486999589909474,
 0.052695680986493576,0.801679213511233,0.631176190490834,0.15058175807712537,
 0.8566112843250522,0.5910192329362111,0.7204578155734653,0.7124893376805844,
 0.12021454798318199,0.13676840434577409,0.9051122912883905,
 0.0626549354997239,0.9311799649882577,0.6599140948809812,0.25575127430486966,
 0.2514382148930684,0.8012788825435787,0.9239056314013139,0.9078785048381479,
 0.2612740899572914,0.7789871724960996,0.9545208321794018,0.7726374042871204,
 0.789676175794399,0.4002068556425725,0.7054882366691849,0.992704931597389,
 0.6688577231690885,0.9675639460244513,0.04565286872491947,0.0435039838829685,
 0.24773291041433243,0.84703430624819,0.5332960707955463,0.7012791731132944,
 0.45824343309171733,0.4896763460478083,0.8862942346611066,0.6920390071843363,
 0.8790903077895017,0.6751918439654192,0.09186543336521424,0.9122428120445916,
 0.8987534043488338,0.9132687062269407,0.8736775952989919,0.130919927395583,
 0.13107552223561958,0.5680143378347513,0.16923041500530023,
 0.35067116283371336,0.4105578276927163,0.5908635026900501,
 0.16217999992079501,0.9381630760702537,0.07825280463945128,
 0.5444014007647829,0.911086929438038,0.4843590969247823,0.10814570051926076,
 0.7444431969649202,0.6274053228303249,0.9347525821138548,0.704636001984893,
 0.7988824555075008,0.5815651037936537,0.5107238658550578,0.35285476232522284,
 0.010463229935421037,0.30903139527068224,0.5923146450183113,
 0.04657985972817347,0.15335441684953066,0.8918122467222798,
 0.22397530935889942,0.5004598521296368,0.7223869682119743,0.6852762895784364,
 0.8183146993854522,0.7792744001860031,0.9811868913813806,0.6213269938217115,
 0.6712333758525125,0.7118390585418053,0.2682341254339212,0.06276369141863325,
 0.008575335841622622,0.8551884486930307,0.6742850045775383,
 0.41140041990531484,0.5124315733093145,0.6670876667448631,0.5891093957271417,
 0.4576863219576761,0.0312860112506983,0.3452497411744091,0.31606572909263564,
 0.8728956986055352,0.8599508928054976,0.698876204918474,0.3857501260993411,
 0.7323294354644653,0.30314452912378964,0.40162539975885225,
 0.5568051906856026,0.8263542631344777,0.1917462372613976,0.5347357912199419,
 0.24094463315016912,0.7123968570108373,0.8752282800515545,0.5915819959430608,
 0.3751466275662775,0.8627604726369447,0.009051259617601115,
 0.3856035929109005,0.1364085171213173,0.4818941411155997,0.2870105219398853,
 0.6471622879336412,0.21596617370239068,0.6104032951932868,0.9518185278954995,
 0.4716859157208383,0.4824583817669972,0.7997453211721197,0.10117293806634375,
 0.9506289313964288,0.3758423646979975,0.7124018945603765,0.7953372973942667,
 0.05455816882487485,0.6665857627940915,0.6337614227239707,
 0.47913456231119345,0.5409778270988254,0.7205514654416739,0.5004255971478319,
 0.47904298490134356,0.39645198010445604,0.26717532101649666,
 0.29781344670857846,0.7080569433632407,0.6842170895313252,0.131319439802686,
 0.18686001229839477,0.9616702184413868,0.985805813836787,0.9671017818047781,
 0.5220497322586748,0.3128857638382353,0.07024511392259969,0.7210436896039798,
 0.9639565297940788,0.6591948875431747,0.8634145577178916,0.02677234109913007,
 0.21877001560283504,0.6668510050896111,0.5108951652986199,0.9104597805574555,
 0.497870589670519,0.384871249760674,0.8104455965358589,0.09841024347896066,
 0.7553234958934908,0.4099583636432138,0.8561249859808364,
 0.009910616751159895,0.4665370570762982,0.5246853842674702,
 0.9291422164199791,0.709070141613613,0.6689153789102876,0.18643243744666405,
 0.10430176477237385,0.6538048344278257,0.1700103007419892,
 0.28294835879645175,0.3136004916883446,0.9137167817004833,0.7127724293597979,
 0.26934468094659336,0.7728113743979007,0.8490158390170535,0.5370558822075792,
 0.5504283154019657,0.23336144592074548,0.2476434104148223,0.1371651006738348,
 0.13404300265072233,0.27598032077002843,0.4871458486347837,
 0.08270231623508106,0.4525865308091689,0.8866415363158882,
 0.14613079193809286,0.7820556383867523,0.23950369446903952,
 0.9062540697294278,0.5628636589034648,0.4364337916006702,0.37876961522849584,
 0.5772899760210618,0.9840138065441719,0.8082414834084625,0.28700017982921666,
 0.8757821801388321,0.7661925570107602,0.21858627195122704,
 0.14519800341825073,0.05183846334399689,0.5083514479659923,
 0.12295568925299494,0.3990457159176476,0.3520163634133302,0.3961176081282509,
 0.6902315933434884,0.2137281748981542,0.4541408974095572,0.49964789272007715,
 0.6594280696604176,0.1164753727490323,0.5367022430526639,0.5907717415714588,
 0.33605558765470367,0.26713186012851264,0.6358398204681721,
 0.28788332678294914,0.19948801926052484,0.6845005188882127,
 0.21783328348589204,0.4135492198900068,0.7056711323738458,0.5322610569565177,
 0.9116760353577357,0.731658800746809,0.45738471550001336,0.9713097276403708,
 0.2565808046769372,0.7262416369570873,0.2531238528543329,0.610109374201634,
 0.1288461757287459,0.5779156460552757,0.7949243458776682,0.6469259884972185,
 0.9759766933296187,0.9066693942600076,0.9747904258219029,0.10477990265543069,
 0.8560748755861732,0.43350589336455647,0.7868363949218679,
 0.18047273328215097,0.3728149078017359,0.7987833617490184,0.1368026406072057,
 0.5558058085194792,0.5572405993250784,0.628379069536741,0.881099185476109,
 0.08752289789189549,0.30198567475294225,0.7527817343809553,
 0.9030995385090504,0.39400981230765497,0.7704057763193151,0.4054761302734923,
 0.14322267106708741,0.3155034192920513,0.43640644937962036,
 0.0101899110667375,0.29804378149534894,0.3782958020187607,0.66507326564468,
 0.07534468661078919,0.9096885349284416,0.6123911848345678,
 0.16671095757467158,0.53411866291982,0.012871012092329437,0.654697123712255,
 0.6468321555505913,0.3255576142251271,0.3911922593492463,0.46939290757056484,
 0.5657548927665828,0.9652022303781213,0.2325808816566013,0.8859935503111199,
 0.6321458792133655,0.5389499995472256,0.7103655017168728,0.7885610828113593,
 0.7467928612896392,0.9004545476159955,0.48413551189668935,0.7537366236449513,
 0.5309399317180273,0.0013634232896294618,0.33294433304141136,
 0.22101037469714468,0.12921354585232814,0.9499750630787124,
 0.7206500188926216,0.6002834451144686,0.949372681406022,0.8267388796089006,
 0.47527703547318667,0.6032750952369924,0.7655239496279068,0.181312185287265,
 0.7335952474246047,0.544382120789865,0.6956642927555649,0.8676215751914799,
 0.6888010817895632,0.2813084524666317,0.3311908367044445,0.7688616296212677,
 0.39146729183960605,0.8475028848039247,0.06812916903050725,0.753920285439188,
 0.25693770761802637,0.06217079874990228,0.6413906728380787,
 0.6691132631293627,0.06414542184162031,0.4382651398281249,
 0.46991972479868216,0.5264519439181883,0.8202529829382644,
 0.36105915145117606,0.5404749474213313,0.15137469235306444,
 0.7566704584141367,0.9410832206382702,0.7575583884937074,0.19578042711727028,
 0.9502037405635485,0.9772049559824142,0.33774215803904395,0.8707382363472795,
 0.07385721850331461,0.651131175810716,0.3286328986377335,0.9067516576455645,
 0.9882180445088801,0.01576071459111883,0.9782443737380282,0.9016849936000897,
 0.6599785600039675,0.10299374941663275,0.0032321115172930792,
 0.9181008305836528,0.05728827961153116,0.758038415960746,0.15241241228780744,
 0.08077857333476335,0.7670822473979763,0.22135239361397185,
 0.4175521430066562,0.3311611566298591,0.3285840190006759,0.9741830561373773,
 0.17578663152145524,0.9370383761135076,0.46267373086507835,
 0.36675838450701925,0.22431549674701567,0.37261832826230146,
 0.20478227882345457,0.8783130657942222,0.0011551476977352815,
 0.13161453752921226,0.3697101231716243,0.10793731215567559,
 0.1264139057937791,0.33809332516410906,0.3499103908779214,
 0.39571516051075584,0.21764145754303377,0.5522514657784784,0.212966041835853,
 0.8090533849043802,0.4813242186153964,0.271630866864479,0.18747450691045908,
 0.1984206933824878,0.7017135310964662,0.4578702322890592,0.6625543074203013,
 0.19308015096202036,0.2721220828823273,0.828414325023832,0.8840826672548651,
 0.31538064927238574,0.6242428110142626,0.69085754190072,0.24115049468934457,
 0.9826439974259095,0.8809884275501643,0.6043023460352928,
 0.017925536592830027,0.4864217247034154,0.8996702510757857,
 0.8219823476908179,0.7748602602693087,0.07381025108033634,
 0.11180348972659337,0.003734890514175526,0.6225138549127384,
 0.3953198731280215,0.535505897468013,0.6083152298991384,0.400010562900887,
 0.5207155206981512,0.22927214122784834,0.7028035261030837,0.474955924811607,
 0.8985340295937932,0.28501249911272564,0.5357018670627764,0.9264460711562824,
 0.752162777337209,0.3927165264457829,0.8877060947744457,0.7691732400408198,
 0.37766704395533424,0.6469501422054357,0.4810206668544357,0.498957351088985,
 0.2507179677859098,0.20035258658841715,0.5588045664196435,0.486274262138085,
 0.48915191527546453,0.7848317590529986,0.4812328055503511,0.7086605588944705,
 0.787123389341846,0.907654679499524,0.42659416304006226,0.11457836776013475,
 0.758296313488049,0.6811399009236764,0.5099433598623381,0.6287874409454625,
 0.04394498954943504,0.8762157630419166,0.8019914583222384,0.990910359499518,
 0.06944735026989646,0.5757135057725185,0.21213970941161875,0.845347286881952,
 0.9766832639371871,0.3610929975524526,0.9461711784545593,0.5549344402730143,
 0.6969036982276221,0.29562062494240904,0.5992524119116975,
 0.19788282624752118,0.4935671457726295,0.8697346193026386,0.4738268600678026,
 0.18882940207595045,0.42897061133065517,0.06772231364060011,
 0.2298986947129451,0.9330436637466951,0.23755790984357394,0.6722317064651664,
 0.8049067062555709,0.4602566871761662,0.12815505016270845,0.9851709143173376,
 0.07866417501997902,0.8727940765111022,0.17008908892945884,
 0.015080318350361277,0.23010002550193187,0.003968468076487763,
 0.1350950726242911,0.5780810650265373,0.38021888945550764,0.6898140224820091,
 0.8805161502084097,0.07093529437828416,0.6666766926325614,0.6521104892015364,
 0.17362313172670563,0.5947247184792314,0.9997373358745892,0.9378249749892495,
 0.1357188792513685,0.666100564326058,0.035926389551971916,0.7036682639244582,
 0.995592764229377,0.4330572484235836,0.5612733204916113,0.48476136388713,
 0.4718898852166441,0.6375554032842796,0.2599020901234833,0.14417641759692035,
 0.4572105681146855,0.7559652981862474,0.9149664710916693,0.9568532277805,
 0.5970482317151228,0.14155739074205087,0.21756278757516523,
 0.0858441923542016,0.22053288391417525,0.061604551876609626,
 0.9249371340482241,0.48538036217624825,0.5213098946868001,
 0.33626391122035637,0.5555704083346005,0.6385932379227315,0.2866237690056034,
 0.28662915764048114,0.4862462017987006,0.9414751799580383,0.6872614266409435,
 0.3755886734591851,0.3176211395465405,0.5103411409406727,0.6599420672436703,
 0.06575232615723037,0.158501020794787,0.3096594092871553,0.12475965208107564,
 0.960419622459089,0.7502619688499055,0.5619257822910003,0.26045177550315257,
 0.13672660822623584,0.05712781475260853,0.7840696584530978,
 0.9863710403833641,0.9032969501310469,0.8571198076022138,0.39151051038454954,
 0.9121099666967755,0.4961630900466447,0.9246828791864957,0.7645693001447338,
 0.5202903565959365,0.26593340725994274,0.6582140616543449,0.8627547031883452,
 0.2847897231490212,0.0651949641469004,0.24618476621661056,0.313167187811373,
 0.8533270138328373,0.8885445613207027,0.6563805144525532,0.07905564267339893,
 0.01706625584613708,0.6524165794259956,0.7976667020976721,0.9141089262602891,
 0.3712067410753749,0.5451937199214172,0.25808440072829797,
 0.09750450274710998,0.4424331964273138,0.9890204960139942,0.5333906260536293,
 0.18899014083571186,0.9883793305149434,0.038789561505927384,
 0.4441953068752099,0.022067563851707384,0.39763576715712434,
 0.14660849087490324,0.2260515869265476,0.4734150358050989,
 0.18695956104328804,0.24483094413903095,0.33686344491388065,
 0.22560022099292043,0.1337161875137013,0.8537232557202598,0.7931608177836509,
 0.9228046154135061,0.741364904134194,0.9208140372606771,0.6974935537085133,
 0.609862971138034,0.686576602102704,0.9134055899735075,0.3986418197411089,
 0.7728212261403271,0.95499380903212,0.07872860083459043,0.37697081571088586,
 0.2528031583190907,0.9836403564286701,0.21610606493383167,
 0.43908590993152363,0.8583178342198355,0.45342337133263444,
 0.4255474939484445,0.24287188725936215,0.534522424541523,0.4329282319455776,
 0.7724815525589743,0.1307257002710639,0.7625596189575528,0.09448846294252733,
 0.1087252217793031,0.33328051375181666,0.32283667181748277,
 0.5005913976688456,0.21117855217267445,0.9555776353504638,
 0.04537419840512891,0.11095972855065062,0.7523148364954092,
 0.6942959720481757,0.489973970587848,0.9744958561744459,0.1612621461786039,
 0.577176127487311,0.17770767526456954,0.29541613345523277,0.6234050273376909,
 0.1858613800319291,0.010184299481533876,0.2270707638279218,
 0.14303712678724545,0.02363226697221621,0.4640443329901076,
 0.8070411427457023,0.7695020362722671,0.6593415375772748,0.3020534090251482,
 0.6378535591558511,0.9008261434102707,0.8195611992666748,0.6853197580182951,
 0.6707581834444429,0.1766386558646177,0.1484855510259403,0.5743994140393783,
 0.5702944250159179,0.7217282974400271,0.18243210092563367,0.8253311182415524,
 0.28604848498437474,0.5385309157596984,0.5423397520479385,0.8117533434196352,
 0.30574994826441837,3.216762493013503e-4,0.5579274679267711,
 0.5455879254154751,0.3474018963298481,0.04130584986555319,0.4515357889293873,
 0.6661569029020035,0.47815692319281045,0.6443748204327201,0.2881948214739143,
 0.14286595333546392,0.6036602919797011,0.7767274636043957,
 0.05075275601160045,0.11962529256833121,0.2679463093848756,
 0.8001463541937331,0.6317870045551707,0.7042528412311188,0.6852227565959618,
 0.8661663038840511,0.6003957239102669,0.19321700820965193,
 0.25345901811826665,0.15567374314458593,0.6115127684972755,
 0.7879986052446166,0.22509183868779725,0.40158820986317745,
 0.03213486435574364,0.7973218336689103,0.3245429569083249,0.8567239279866147,
 0.11307125293322873,0.5797279207065187,0.4893521180776783,0.4570672104609632,
 0.999305948842225,0.9402002130718644,0.3811177765755459,0.2753605581052432,
 0.19710938922550314,0.750496796298929,0.21364295099047637,0.3636018485172796,
 0.86402833452126,0.5553892944032374,0.37617054490818513,0.6906691510480797,
 0.45802084029345436,0.8022481073904103,0.2495526429604833,0.5433184935222265,
 0.5987663109616389,0.21921922232359703,0.6409718462816018,
 0.22898887975250515,0.6879587133609669,0.3389580259828928,0.9173818217678358,
 0.49446922447826713,0.07526610907713727,0.45784733575924186,
 0.9479746082331783,0.14388770373662296,0.37808377010466754,
 0.033845149306095945,0.7929960721390895,0.6241488769280459,
 0.7311072832166879,0.5733636215686106,0.9418579125830225,0.437954466157797,
 0.4951423500347736,0.24606520302169743,0.3430979323393527,
 0.45388823649251653,0.4446958936257024,0.6322326473521407,
 0.46543697104023396,0.26355044919644777,0.9298006190413113,
 0.6131445865874502,0.6925279754683942,0.19466558623502317,
 0.36472408866654715,0.4000583631718909,0.76164853657009,0.37921299884708815,
 0.8939086968149057,0.9146906972800227,0.8137047814001288,0.2934409884549186,
 0.4147423135869506,0.07258385735596928,0.21598992869127398,
 0.44365229784189464,0.08462256380526045,0.3510596268334567,
 0.9823653368693507,0.2830597120327628,0.9070740441202422,0.9712871363256309,
 0.9537394941140322,0.8838427017497328,0.8962038877916192,0.812958246655349,
 0.9791607279810379,0.8701062602764641,0.5883645823452156,0.1963119145864609,
 0.3976510429078379,0.9050476882727896,0.5874429990233373,0.9056359537322795,
 0.045613875249295166,0.2849569411150622,0.360818436401632,0.3443761646442779,
 0.8285453120527975,0.1513269499167298,0.8176631646851801,0.06620325814396077,
 0.12707457150897583,0.9123383642405251,0.8399651509257179,0.4533694284850418,
 0.18198181328547203,0.07143853730151717,0.48348318674245294,
 0.9328502671770211,0.3348544056355841,0.9893205443489148,0.0452323352015811,
 0.8214283360787004,0.8928306229038071,0.027463113868155253,
 0.3782533543544724,0.09509505021248643,0.19561076523025012,
 0.03158105351812279,0.4752496466180771,0.7056844149470809,0.9448141610295334,
 0.8048850783587733,0.25377724727784345,0.6279979684329534,0.2965561586204214,
 0.7848063405061361,0.4104096209191531,0.7247800206519772,0.3755380627675666,
 0.7219744683209602,0.4405535621797846,0.599420877614481,0.8940394534399828,
 0.7377831801170535,0.8288119992570985,0.7806731213670148,0.501187152521903,
 0.08724125907766367,0.510375083181745,0.5991588492379639,0.7224572811025667,
 0.879052382288163,0.9104116240222707,0.6945767719711619,0.745589309687424,
 0.3486567359087005,0.3872319345503503,0.04825154727212988,0.6128090379444343,
 0.706550882905959,0.7602306619651203,0.28574185122457785,0.8113198367734351,
 0.5700947253891997,0.8877224555214945,0.9699103087563768,0.5825467499582524,
 0.5632357800413814,0.7231753795511434,0.6311331486074854,0.5493886283614366,
 0.3520394696841572,0.7981143555170809,0.1881783817957452,0.01902885491107198,
 0.5932072741382043,0.7965742280494567,0.46598748930159495,0.9579074131468897,
 0.9702434400214051,0.9982693478673638,0.5116929762060929,0.20106022732350792,
 0.14016433441846265,0.20947819957871394,0.2782489753804648,
 0.4248063840317712,0.5659336323464365,0.7659107901672259,0.917916602351649,
 0.18983616783514623,0.6935995607269088,0.010539082039234926,
 0.6159087005991367,0.535500261575272,0.7600945181462173,0.9117052110370903,
 0.18944232257322813,0.9021264051270429,0.057600140761384755,
 0.9846568345636644,0.15643598487916455,0.4100740119127495,0.8139931811787838,
 0.797781607267869,0.0806351633376321,0.3970477556831369,0.9200485056365639,
 0.9961379258796457,0.5074365549264663,0.023485800858043304,0.457186099311199,
 0.6266671077522108,0.3510931742466179,0.41849672379110814,0.6406211923577628,
 0.8861292371218423,0.9865223253906219,0.1933546412453936,0.10273315555564833,
 0.672808455006455,0.6922652184450367,0.2739013600338559,0.7948597347130744,
 0.3141775510351974,0.3567050249022351,0.7222675373404274,0.8925336461236352,
 0.9792851389262073,0.908728607037523,0.9081641875705854,0.15605020278132908,
 0.7538063501057235,0.15544387625074585,0.010790778788248079,
 0.1617529147150749,0.911100605479691,0.939108408560027,0.04610119010205593,
 0.33976104887247094,0.5191218863240161,0.2567880680904969,
 0.13514374001225427,0.022147553931573638,0.1152956628074131,
 0.8060046596898122,0.09752856536468357,0.3279164227962216,0.6744124177917505,
 0.5386907631078972,0.7628323714581833,0.28205209147457877,0.8982027561712761,
 0.4849397691262324,0.8963745379738903,0.6365341927834541,0.16726691234546553,
 0.7116039698643378,0.18979228200084242,0.6929415545350026,0.7978257641227934,
 0.41407914252604117,0.771879227062251,0.4074441070941899,
 0.010820603732308975,0.5026810131011823,0.7742292631863043,
 0.7553975500771106,0.959062008515837,0.8441053301121625,0.36749528854960756,
 0.5479188258595991,0.07903848086989751,0.1944145797825091,0.9737527911842944,
 0.8199051054971175,0.42928472938016793,0.042490871903231864,
 0.8028785427733518,0.26729834168882893,0.7063016601256166,
 0.41181021931628115,0.343259445840675,0.8485914244929729,0.37439880209829046,
 0.7033917634683459,0.010365476536086105,0.8837399675828098,
 0.7966978437257229,0.9453181574983894,0.4726065789167073,
 0.047360038096580404,0.3569823289992514,0.9672402161183843,
 0.41059304604527325,0.10240235218033178,0.07816065938237127,
 0.12393481703780806,0.4728278305241591,0.8594663034323712,
 0.03173912835582671,0.03692535030723432,0.31375110027935427,
 0.5596282764074454,0.658580286133276,0.717319416472308,0.35109341870728306,
 0.29190804705127626,0.46777209601728154,0.335142684559263,
 0.38009920049586154,0.1326587264326563,0.45390433151465936,
 0.014571603080145157,0.6657319653845036,0.7638005740231736,
 0.9627798838932949,0.08942523378251754,0.5786176856977178,0.5120834163888036,
 0.01632866345739392,0.7431990412950176,0.18506642630105719,
 0.6483045574322731,0.47662745123030614,0.2655262803822045,
 0.14481187470673196,0.9599841589028826,0.8392590198126388,0.5405886347229406,
 0.3031700252461097,0.053264105065825684,0.49929378597813134,
 0.7492055366994754,0.530740350578041,0.5777593074755495,0.7745771362434577,
 0.9464084983570872,0.8946250999778808,0.6856013616724236,0.9189786806150372,
 0.7343401607750797,0.6155823943446184,0.8971660631826532,0.6739142904789045,
 0.24722030605198453,0.3726962312998888,0.8708596112435043,
 0.19418143360863482,0.08402057494195248,0.24433426373804812,
 0.5058872509339047,0.83272668191869,0.5345940335224875,0.4677805327153517,
 0.06928194955751676,0.2244983875567128,0.6410705360059822,0.2717938945651187,
 0.10241330304581864,0.601258674069785,0.13731733261397405,0.8500268248956433,
 0.02546433399969772,0.06694287534351517,0.9261398567413383,
 0.3885692144651256,0.47846662141199614,0.71545045966575,0.45176153579508405,
 0.11929416494330491,0.15882611549275172,0.1492081305009696,
 0.9887862012667168,0.1965401498758843,0.31310720520576585,0.5660044462184413,
 0.788264031664407,0.36779543288283123,0.9288629036216786,0.2802610184952644,
 0.34557981366328583,0.7988630745802838,0.5080318765539982,0.5329040862123952,
 0.43424505073387987,0.9700912930288987,0.13858005740819546,0.791461039693067,
 0.47763362385216546,0.6442722673146334,0.07984555324224851,
 0.056143842639886254,0.36933108046568475,0.8924733343653629,
 0.6257923732720182,0.8688136052838912,0.14806856521842726,0.1251962388855018,
 0.3323866084545919,0.8885901039132655,0.1929049682954117,0.3986019551059311,
 0.8388179139876614,0.7749056559740297,0.0012142970477908577,
 0.4776856660516535,0.1444995509153778,0.7671897893156078,0.26600780262878976,
 0.8071689800275607,0.5051462088778096,0.24790250360027655,0.3135441833672331,
 0.6493525680822412,0.44676642846682,0.1341731587531979,0.020440916102979267,
 0.814911089292222,0.8528362294239742,0.9178388537467262,0.5014749766999897,
 0.32007587961880657,0.6791503824486652,0.5525529787116464,0.7395532323646862,
 0.6453601870505488,0.5134010361892627,0.6512160216331406,0.6924967191937266,
 0.8215686301431482,0.45891124525413074,0.7447757980776961,
 0.20015288533062447,0.02197478631045413,0.021334480948914214,
 0.912949921633432,0.6677570240682288,0.7400100772009025,0.028047047124916347,
 0.029328086986869284,0.8942884004894627,0.8570399341300327,
 0.34617562131494684,0.9494714273846319,0.4470923754834376,0.8297356273779274,
 0.8344899632963064,0.8685010059004652,0.5894662271351861,0.6957745181360143,
 0.7314789836397513,0.26305954689568334,0.8493994727018324,
 0.46308799578114435,0.5454749734246174,0.5519605962129495,0.4253571802392033,
 0.4579969833580968,0.32278888447933185,0.4565841195028888,0.787509224737907,
 0.26732193687798966,0.4475445379318307,0.8474618218395275,
 0.20756592217710823,0.630339062099134,0.9278318081169523,0.2825429474839424,
 0.42116819307524056,0.8918872537071162,0.5422369928063301,
 0.18629871587470082,0.23912387710962324,0.0789435918386251,
 0.41972360897116046,0.4530424504808922,0.587130547744046,0.8330633638036492,
 0.9015118873677304,0.693156032505919,0.22316940029486898,0.5629635706806659,
 0.8445778045149543,0.19733018394863122,0.9962589984608297,
 0.46292567839347265,0.3732700003271533,0.6796334341084742,0.8730093024085303,
 0.8525979802147379,0.5390301087458655,0.3187295783575723,0.8505918639017347,
 0.12716041714893056,0.06604955761451803,0.3338306606526462,
 0.25175754318325416,0.9433361244924364,0.24286636154256103,
 0.6654103720524883,0.3141320845505854,0.5088772864264974,0.39933929381155175,
 0.14309141842858286,0.0040936555681634346,0.14827864358238652,
 0.49363636313846393,0.4967833832331291,0.5545893197038407,0.8005640053250168,
 0.1027320883493521,0.3755643943698561,0.890122163778894,0.30542806222748475,
 0.9793470356018231,0.29959611928705765,0.5811645275748865,
 0.39885468327092666,0.2771454559468536,0.8298018379981659,0.66359407772585,
 0.03563864081074297,0.8553526689032712,0.482332297717739,0.4957914153512408,
 0.7487433169341426,0.6389038100091975,0.6757504623940727,0.8690607545893019,
 0.7466584790717578,0.4713175010149697,0.5706006153384465,0.2006791392667664,
 0.8352043312494378,0.40256648306421494,0.17638276301987654,
 0.06489523506155315,0.7686151163843707,0.907736303479379,0.2635060796066261,
 0.21395521071634505,0.9855618234534501,0.3988115955951932,
 0.18183925170494375,0.5677833433756285,0.30499726593690013,
 0.45323404216058394,0.4968929816740567,0.5016196605005914,0.532336224735932,
 0.1819875588793861,0.37618516895524534,0.7519426993278955,0.3379477833898208,
 0.978901316596024,0.9888039983905214,0.7303634802831359,0.3507446670451676,
 0.7515791327259246,0.7927840322744014,0.025049960150651085,0.520625041171951,
 0.18327018146904273,0.945701435294701,0.7907638427092052,0.505492599503165,
 0.4482474833898543,0.3765446332352188,0.8721348641993898,0.4485525353569746,
 0.37872668647880625,0.3401168602964666,0.32604770903418423,
 0.28895591654632735,0.3361072607119949,0.6852268505961578,
 0.14824278153147807,0.39008988026155056,0.29814415522136606,
 0.2252407838936904,0.5975051110177195,0.15760549682626002,0.9400638337943259,
 0.6719181240015828,0.13523041276502912,0.038432666434532914,
 0.7313599955291159,0.1736273627683127,0.44988825211060257,0.2310493857320941,
 0.0015196034757534171,0.23230317626385788,0.9383262540528794,
 0.1376188706842968,0.41547912268091414,0.7143267831771916,0.5570275932907262,
 0.09169500639873895,0.07659171892334515,0.006893271030821602,
 0.9947267335474848,0.0696540329231361,0.4713266381189828,0.10947826671498728,
 0.0820833956503162,0.6245509235249463,0.7669315719429041,0.4581049534125049,
 0.19633392805056382,0.7746909653362768,0.24923346726115125,
 0.2355315366499362,0.6529566358985641,0.7409527127262499,0.6867276313363171,
 0.35410435722790845,0.8384728736854111,0.48371675567603956,
 0.8920008088045255,0.5977072746320371,0.18694226067530062,
 0.0034119618242514083,0.30413472410538234,0.21792884294296377,
 0.8041754091690906,0.6657820368931154,0.8559984467311842,0.7981929641300394,
 0.6982167318324999,0.7643126398690029,0.4979002758158698,0.5538381676793638,
 0.5829326863377001,0.7054928498470485,0.23235526338607104,
 0.29391489012077476,0.7878091462987232,0.21798376143948917,
 0.7203111374030984,0.46207589851145303,0.7405558799329653,0.5958635708327163,
 0.24820640101007618,0.38698000508512653,0.7922799504292326,
 0.5850267923280363,0.9498698062018947,0.3707429366168312,0.22748270580428342,
 0.17035454314015652,0.789157740359807,0.695495882691,0.5968734061251776,
 0.17125570440612758,0.1463644716511434,0.2716875842208203,0.6336621064845677,
 0.41749121008879864,0.6805595366188804,0.26127048838509637,
 0.8373150236408706,0.10011851383316828,0.10488550139477071,
 0.2752146218753566,0.9381663581310284,0.0581113421561672,0.44473066084825785,
 0.7004148786034582,0.5495183186868746,0.29067347693730794,0.6239047618974096,
 0.3084829995540268,0.4398700385331379,0.8114174809344344,0.7904523448445144,
 0.9600432721305212,0.5341083129797352,0.42995467163959766,0.5039727403854158,
 0.16348048985936314,0.06046526908294381,0.2488448721400094,
 0.3061813065533514,0.21644906818029952,0.19229342333338106,
 0.1733694399564204,0.6017305246121892,0.6539327397931688,0.4327610981306551,
 0.2178454800559919,0.2198945517925096,0.3376904595761846,0.8263378479190042,
 0.9369926291156853,0.3376196108767284,0.0797429803375107,0.44295904154440113,
 0.03505748642694484,0.38011794735707105,0.24659814870837027,
 0.9682199884473129,0.13480316610983956,0.4795049992521364,
 0.17125869327221177,0.2534112883617581,0.9009498461754454,
 0.48692578720571544,0.2864145197523882,0.823409726278214,
 0.007728049490611122,0.5726724358688289,0.04371296296076188,
 0.22155454416395393,0.9754981100677564,0.6004980907932158,0.9409016914764252,
 0.0067887299279576485,0.5998772395821492,0.8572138112411842,
 0.20510659611833648,0.13955870610915566,0.05195876285859846,
 0.12249832310049924,0.1356027165427861,0.018441751863807476,
 0.7398279949512836,0.2163650645395243,0.3137183899744691,0.7026178694944507,
 0.13223247997557042,0.03197119992810271,0.6798552756119257,
 0.13987315110705256,0.5180442612671866,0.006613314380887858,
 0.6655514783852492,0.6326404762951053,0.33646300509601534,0.7776525229119778,
 0.566012647893641,0.7631324568137275,0.5948071426292123,0.9480916703309018,
 0.7269564718221765,0.406776762127667,0.1538319392659936,0.5155248999303406,
 0.38400828757472905,0.7873353375375919,0.34752545633767806,
 0.19527732606766524,0.7573512959843232,0.4174322784057062,0.4636387508256967,
 0.7229423236823859,0.08195121217636803,0.41025530053604486,
 0.4651586382029096,0.8295000535666945,0.9880481566104842,0.40892277704638313,
 0.2674990260677311,0.9101238304369397,0.3119052747748883,0.09341321615918319,
 0.10637507150489212,0.12560020675398842,0.3508895486515944,
 0.08213792483062743,0.38911373963844853,0.06453446052025724,
 0.5157944302200645,0.5477088265933061,0.004799748362841583,
 0.43091525974866884,0.8230692807112971,0.4733590286078099,
 0.38844230761673526,0.3663390134130575,0.3004741112261594,0.9376440620281936,
 0.5793319784373407,0.034713723454747436,0.30680190419261655,
 0.2582508790654223,0.7552251607014434,0.059153210233390596,
 0.5270135338690373,0.22281269956080474,0.04457301346483389,0.728031197806336,
 0.7833143045838034,0.3576113283649973,0.7007148253470842,0.37464547850627183,
 0.7968645131215947,0.22763688578589814,0.1497419086055114,
 0.03504926945424802,0.3290963314448955,0.9805433209964357,0.7112182206845955,
 0.6479955284634692,0.11987998149278078,0.5542694905513622,0.5505347524635347,
 0.032896334580206354,0.9941172748303738,0.17317863508037612,
 0.02120957118558542,0.23884830906113153,0.26290173561040364,
 0.9096747213030627,0.014169693873787192,0.2777312914933894,
 0.3325836678217814,0.7430284740940449,0.09842396793748032,
 0.06221842633274266,0.9481014347663453,0.624542607858208,
 0.030853882532953292,0.4106490872509426,0.36399836242558814,
 0.6977079339603794,0.4341528581598926,0.2627670837895838,0.43028906761457586,
 0.35625901739232935,0.9850401218976753,0.5854576932002693,0.5660091772958191,
 0.9495978863254931,0.7040022832587675,0.42508197893869715,
 0.18970033209551906,0.8490322198690448,0.034486031638026216,
 0.6754068134810434,0.8749923400071488,0.5747995445188658,0.7322433940736502,
 0.17853093905223838,0.6084126559354881,0.8415002896750629,0.8650197023334376,
 0.7148002769928614,0.13314123788277676,0.8060159974335184,0.2468155958324587,
 0.8541922922012033,0.5413623146340537,0.4823305866260146,0.8463466473885989,
 0.2403983010966252,0.7869256449623013,0.5940242014358693,
 0.058233691055326386,0.4305107496108582,0.5504314065921351,
 0.1062570305424132,0.5773168951646716,0.45421978571847954,0.9380239723485805,
 0.19325546047967423,0.07457039249656083,0.7291862121988643,
 0.03068124436048647,0.8553963052753132,0.4202285509319308,
 0.39830020544844924,0.6199942722450545,0.580353116452931,0.18301239396915092,
 0.30431587578885166,0.24700810202408852,0.6108856900194565,
 0.5477650640531041,0.8697772418106282,0.9002916924373022,0.18941437362057856,
 0.5561590335508744,0.9406689301583828,0.7310922898863415,0.397011722592959,
 0.7205935353194588,0.9964519498969937,0.24756075443923387,0.7783421937303681,
 0.0014081374434750105,0.113571279220122,0.5919633643243402,
 0.04181823881716085,0.17047301393136916,0.9143003984860157,
 0.5526665521875158,0.6556263956809879,0.4357860007383787,0.49421932569753213,
 0.6227764967036729,0.552652317114306,0.6600315738577074,0.17935067447473774,
 0.04029975635261518,0.7362963903953196,0.48642030011399806,
 0.4146303659468503,0.46557451641902337,0.46341951722526187,
 0.09844006039148046,0.995808449731399,0.7769473391419567,0.06697595249536659,
 0.7819370867944778,0.9737536949267732,0.3258777167421696,0.6938143604975142,
 0.2680133025247665,0.8984681062129436,0.4632969603906095,0.4167327580438276,
 0.5358462576571432,0.4356848204255247,0.67313636183793,0.7810933755719653,
 0.7355405092236544,0.40402256226325584,0.8395693279347024,
 0.04459172794955846,0.245105881933115,0.16952611589114386,0.5840542641062992,
 0.5909219818095188,0.3772502863426843,0.14247818030823312,
 0.40747239616284214,0.37701277413028167,0.7928890137897493,0.469608225649791,
 0.49901319795747423,0.9406248384964082,0.4753528595246501,0.7747388562793661,
 0.9876649584450219,0.8474415048844998,0.038658514460724325,0.71940725581752,
 0.26463905084557526,0.24985115869060415,0.3872373441416802,0.618922507482996,
 0.9974350355364512,0.6788166282306908,0.5484088930898352,0.29082756985396674,
 0.41066939850537565,0.3784835876701629,0.7366811906818399,
 0.07472749535181156,0.6662211831586777,0.8804145692011919,
 0.07236848499032866,0.12556775820608812,0.0974914790632343,
 0.5235333930601717,0.9776857355838617,0.6896106816067921,0.6050061354580092,
 0.12556058585583507,0.957134700452214,0.10465077648906007,0.9954322683972151,
 0.9620743832289446,0.31275023334689867,0.3698738533457465,0.4391458794079711,
 0.6362369097015728,0.12149292737047102,0.05438954509897065,
 0.3817453257599407,0.13679383742705253,0.12243463132035615,
 0.14097133051690147,0.8654042451295256,0.7369601290618631,
 1.554511280331683e-4,0.8778044962282713,0.006166241456039612,
 0.2967556241552809,0.34135264647971086,0.06487276524140695,
 0.023477275609394055,0.0061250611073577765,0.13090637988634046,
 0.825267988028866,0.720906275461594,0.2987758884457715,0.26337804958761213,
 0.19082945971544862,0.5717486798959827,0.6624651994101798,0.0899289126597066,
 0.22324410118751337,0.32044443151597024,0.25161669483341764,
 0.12656249371492867,0.6195957620069793,0.8219795996968515,0.9555804092281158,
 0.5252527531288603,0.028378860021342822,0.2129966439567138,
 0.6040577284155509,0.6060370064706833,0.14686493740699103,0.9436530864176623,
 0.5219781017429921,0.05495292794493434,0.28187550091769464,
 0.014560120965751655,0.1129227163224007,0.758822159934619,
 0.49434704281867914,0.20277056669963356,0.9489891087523122,
 0.16972616695239917,0.009217008354692413,0.5246485934304559,
 0.7049662340745477,0.9508516958834781,0.2322766957958331,0.4217318606179725,
 0.3448206708872906,0.2705045101072939,0.4376957397117557,0.5513788971500035,
 0.673233556366669,0.7279381708723818,0.055172428030461296,
 0.023550490937954915,0.9299998975728093,0.7735223013546567,
 0.05310661544629536,0.16387864367558103,0.19313820973458273,
 0.6218693807616457,0.9773294224978386,0.9102622687718738,0.4061631264698674,
 0.015673723034925313,0.6965364402575169,0.9116694128335766,
 0.5466815793773724,0.7224183420028898,0.7930773914875536,0.19620947960234592,
 0.6708841878578528,0.3373472999159376,0.4850876946313779,0.6806322665072578,
 0.7461839919121119,0.7821852048588523,0.4323747564923338,0.602325969538598,
 0.07881055376117085,0.35438533983689746,0.4406216784943511,
 0.5956599997243295,0.9582611306384916,0.1773866398380981,0.7870868762381167,
 0.08598468966807782,0.8504556030892684,0.5668817311443726,
 0.47491901712683593,0.8513638594347086,0.9588151013147367,
 0.34326217640536805,0.22158820442107918,0.14726200577269077,
 0.15731742843543273,0.5614725236687854,0.5274244496356288,0.6983458316206589,
 0.9842676698689454,0.32541066286960274,0.39979720331707924,
 0.022173425178961037,0.25154504197214855,0.7204773820946839,
 0.44848516778861536,0.2762803370091991,0.7810114638133003,0.612272544424853,
 0.10806905489828034,0.19446064557211218,0.8875468941297041,
 0.2891334513189576,0.1869406622102865,0.2255427875298086,0.41930440153316717,
 0.8420844571413695,0.6398520744495845,0.09282215112085013,
 0.21517942220861352,0.2834482363245677,0.5363306330194607,
 0.13005529440944652,0.7212527098422253,0.6644371023489603,
 0.10057682909373034,0.8631677718985981,0.9256149979287401,
 0.006541222722438178,0.06507197951735211,0.3156640203134392,
 0.12679551842865033,0.9597117387300599,0.0831588633004382,
 0.31297595605599393,0.34786200188399563,0.8031929789532467,0.227299250484858,
 0.576741154189589,0.07739367435858446,0.14848146788868366,0.3449767453916889,
 0.23638389048798158,0.05475397447076813,0.7117578592309501,
 0.19909649795964324,0.3829478570320549,0.16391125402280793,
 0.4027858011638925,0.8541825598858223,0.6521389783573449,0.854580290055425,
 0.3553811222155968,0.12906467387401233,0.43240750076493195,
 0.6865587904416052,0.806880840049769,0.7467320782044948,0.5941324600653042,
 0.11787763102065618,0.23587148283350157,0.35135015236716693,
 0.24414852037037693,0.5468992783686425,0.8828679707916123,0.8199079267823491,
 0.9059795778923234,0.2412882611824556,0.6536188724395212,0.9832945146926093,
 0.932859352083644,0.8076027089913457,0.1032416160946873,0.6497242185308816,
 0.9078142125297625,0.8857573379824332,0.3254710423754218,0.4505738794803219,
 0.26662994578835464,0.1371536629212563,0.18353581759400284,
 0.20166624297496438,0.999384472862271,0.7368881235851061,0.1295235163333378,
 0.27640966820156,0.5538595559882421,0.7454278742204664,0.7285368896547284,
 0.6191273770153206,0.07228687599171901,0.04773267925001401,
 0.2431109068444075,0.36023876685962397,0.0930465598242397,
 0.22881397783199497,0.6338012521306411,0.08710620711501471,0.889577895786987,
 0.7826926224693868,0.002592620859762196,0.9812832502006245,
 0.9679470662800342,0.14406860342035122,0.6509401571793662,0.1999967463105734,
 0.7455577693980983,0.8033715698008835,0.8498455358027298,0.5958234473752061,
 0.6243708458723906,0.028105661559371997,0.8547474393420428,
 0.3441403534736841,0.5176011986636313,0.4066607806281173,0.9682712848750142,
 0.779078818799711,0.9443594120632184,0.22449584296990654,0.9510229286377802,
 0.5997782875067561,0.3762479894809293,0.46046510871738133,
 0.49380017940073284,0.18850784130154097,0.38942975840379745,
 0.6788081458724788,0.6078884369710001,0.27932583754565876,0.8925144359689499,
 0.6230943500662346,0.9759724425495933,0.23735632937630935,0.6726323989663557,
 0.8780476019334899,0.6843224211850102,0.10595768129021266,0.8995572606989998,
 0.43441467614763607,0.3170455778465442,0.17965898052706952,
 0.8865439794508094,0.6029588476122847,0.45885212824063704,
 0.05378264450928116,0.7425820962347603,0.7121228146446805,0.519570960825302,
 0.18649321391770168,0.9322782713228036,0.6119218149142992,
 0.08314689261248698,0.3693444318847062,0.6467048924770591,0.763291078459679,
 0.6513711004989117,0.4741269942622782,0.5536796249880814,0.7814977905400116,
 0.27084619640911134,0.8681550228533366,0.23709671297697332,
 0.8299961989559893,0.29655910355013115,0.5068954295477592,0.5795572435062474,
 0.951562616609256,0.07190590523412,0.40335108436369616,0.925783552181001,
 0.4671868608497852,0.2953247528513705,0.26128040859173374,0.6121782366315855,
 0.5932370143260458,0.9023993261272693,0.737693996174634,0.08334673569174689,
 0.7234441511059174,0.8007324937622441,0.506568954335971,0.3067724466760742,
 0.3685151341423645,0.7599570792311532,0.1498204615073384,0.03694623361968141,
 0.4509212358934873,0.6758183179811894,0.8142328443020497,0.08035895152093364,
 0.7702913310432229,0.026880262012522693,0.8445466286026618,
 0.5307040908828066,0.1945465211731051,0.501323904973913,0.31875737879019916,
 0.36601529611120265,0.8774775084082698,0.9765931739742056,0.7407487698475727,
 0.25891708245048994,0.286826538213678,0.6589530375152736,0.06187801822515837,
 0.13676028095342507,0.7933738767435974,0.19868084763484262,
 0.043338424244219076,0.29851370105272934,0.46539595865855177,
 0.14790252179467567,0.9729154735856504,0.668817817156803,0.2194907209575382,
 0.4496726195047027,0.06749122258569695,0.5649582064951744,0.6553040208359632,
 0.9000888707414257,0.9698345782096933,0.1241237817532217,0.19639696835397458,
 0.5938342585085626,0.7998888969475909,0.5156949144307073,0.22738637142469176,
 0.6830026221456067,0.15326799096993649,0.292084969879046,0.9799032878019164,
 0.11211955291266029,0.7016330946184033,0.1299275262559303,0.4066011823921901,
 0.21278188143259602,0.3252143082343898,0.441435601884694,0.7832757080326125,
 0.11695459474321834,0.9178543523090213,0.6937692646507477,
 0.18445509310546215,0.7766673357602808,0.38231808416383917,
 0.5373393101516366,0.8435173952283518,0.05909915005956334,0.1307187197848767,
 0.5131456386714702,0.3709907594130315,0.05780495563400123,
 0.32475275583414787,0.07347845620475957,0.8934078522473001,
 0.05010481577246284,0.6413557139176982,0.11717938890463042,
 0.2874644904964304,0.91482496627597,0.7025447411919388,0.8957221106574486,
 0.2981344012180116,0.5802082463980615,0.8282972627578851,0.38025602808104897,
 0.7271183514095844,0.29592920946932244,0.7855828791891415,
 0.22114813997035676,0.15018556399649863,0.22014598800294705,
 0.46267283541034954,0.9107527857929817,0.8522871753284396,0.5190655881373627,
 0.7046621635694312,0.09878802875860859,0.046612364363297965,
 0.7996083204569138,0.6230898172340311,0.5542772197025807,0.4401236756702087,
 0.8897425197660909,0.4372084116865942,0.6169508221534657,0.6448764320257938,
 0.13473388632320893,0.6534901516537296,0.603728407924655,0.6817211712969965,
 0.8525124564658463,0.7820623378296045,0.3295692607540396,0.4933879538705507,
 0.5796714263591408,0.002243653746995977,0.22072117453278206,
 0.07381369618177336,0.18224250569614386,0.24123962121853004,
 0.056947924483130574,0.7119865666506893,0.13398028201378887,
 0.9739675811263716,0.7261040776613201,0.017844142782538652,
 0.13777403302989055,0.2491882899864003,0.8057767080926672,
 0.38431802590062025,0.3928766930813955,0.8633450661503621,
 0.48002362740974724,0.1253057295324106,0.26987715518846933,
 0.20642351547824245,0.7320021539252223,0.8161104751251234,
 0.21711253536171826,0.3085579360274484,0.4442336851622488,0.2818039558159515,
 0.8296309278837174,0.09971451247929264,0.5147421276668075,
 0.061544492797006356,0.01767438302128066,0.7209992495301967,
 0.15570684382971933,0.43012501227232514,0.5140044902927644,
 0.3453745556968235,0.8672761987528848,0.3828371119442171,0.5943995484829404,
 0.9211188571971216,0.4558585801356312,0.864944867904344,0.3936148539744404,
 0.22624484848186788,0.8269058011475292,0.5819368953813017,
 0.26137513307286653,0.482494538780033,0.9320444013417012,0.07304544676012126,
 0.5196790314493833,0.7382442367384764,0.35652124578359357,0.4380099328644984,
 0.31334095685953234,0.9051542442846134,0.9632865978265398,
 0.30555076092248823,0.23753482159853978,0.14389903449073604,
 0.49344676768463713,0.5312957009396662,0.1941653649823325,
 0.37877583010272786,0.2223076438765763,0.29546041066815776,
 0.35085162747005505,0.5969397969455126,0.003734027421416597,
 0.6191978802199013,0.3652460302072018,0.9951737187259925,0.6851493508370938,
 0.13987462057626,0.44734751834792674,0.0014272817949366745,
 0.09881976204798026,0.10975460924738611,0.8600559469721112,
 0.32149315235355236,0.17151371533649096,0.13166156233289716,
 0.9960432349457276,0.051823085518638456,0.8493544454967145,
 0.12206120276267052,0.4660229106592193,0.2146356894209276,
 0.47271066544318696,0.12407245400767697,0.38239271349099524,
 0.34367833576320106,0.7704031236641455,0.8846589035326846,0.5088977061116883,
 0.21440029285290962,0.05751549037570869,0.07940678545559177,
 0.4099268573768067,0.3390038387404378,0.06613567436236678,0.5699722224608585,
 0.7387498167864296,0.14263674673178817,0.7386655721204574,0.1121914647529535,
 0.1300849615295352,0.8637873480162278,0.9585699204228313,0.9046971844201636,
 0.4858214583286625,0.4127420491895075,0.49795882349710563,0.3420229921609379,
 0.8648976845538279,0.28619291750195175,0.2217467057853938,0.2458288841995574,
 0.9839050575759765,0.8239426147987161,0.758370116227743,0.44183948948392593,
 0.09360029947352044,0.17892669912154013,0.35187909014200325,
 0.28546281849232447,0.007693505680663182,0.10926933440751652,
 0.40279390291234174,0.7547052448280782,0.08996705817509665,
 0.9538912632126649,0.7519195481190122,0.5488803998171198,0.8526830544612473,
 0.7160929515154423,0.5678537488542454,0.12928694086210912,0.542824546814722,
 0.9744415212304398,0.39619326334629834,0.19454253188335457,
 0.9417832133026565,0.24868620238228922,0.8337433330566564,
 0.03158429053413736,0.4859617165441834,0.9998703242012901,0.8686394972807693,
 0.9792267529977764,0.1110866011941466,0.4921001023247318,0.10530687084634904,
 0.33591917086049417,0.9898790666339159,0.0699995530614379,0.5279380502693399,
 0.24571647526047424,0.845985055568255,0.3731150485921306,0.09483304119898439,
 0.9718703119813494,0.7246557770543972,0.4751934388603941,0.08183412421314429,
 0.02814094466076489,0.2090636554929598,0.9288642286869205,0.859146373191733,
 0.4607626223400074,0.2012546400935048,0.188232164053039,0.30670508530158536,
 0.8447148123248068,0.4512096387319242,0.5426517728351706,0.449381342115567,
 0.317198205951879,0.3865552345959331,0.3376272801498832,0.6370855930096581,
 0.5762741635649655,0.818648891066473,0.6577319897714956,0.3278900043394396,
 0.4144358067170606,0.9152421871216022,0.9361374473010352,0.8862901607723678,
 0.06366737468255812,0.9397456994881495,0.5345889574505536,0.5575514489074759,
 0.8537043069513802,0.639021501885027,0.820241169587727,0.5011228545996882,
 0.013681337569034224,0.18769516403122133,0.8921590925156602,
 0.7039170840046493,0.9503486500830648,0.14530513173721704,0.9246049240095896,
 0.9008822898313741,0.28653650248512696,0.7647434177017092,
 0.05117647517171742,0.3207471088057239,0.989683309956405,0.2249790507987912,
 0.8203425308426568,0.5583926116741529,0.49186644907463295,
 0.24640575042512913,0.6226383559878673,0.9537697662804138,0.3888565436186624,
 0.6009645623852974,0.6654389960479159,0.11340826015329264,0.5440851224174515,
 0.29230723422696037,0.5019414644139324,0.7145778841303594,
 0.03528032338846743,0.46194862160999295,0.22831262711916633,
 0.4118464266827321,0.5804187032140448,0.13079727343232084,0.7331137110282078,
 0.635815698033419,0.5656785540479484,0.1820038393671788,0.3330127506481181,
 0.8710705264810785,0.28559014669726024,0.33634194153196484,
 0.6062307051317647,0.8999307145942963,0.515769542146184,0.9796752951462862,
 0.06792253836459206,0.33354704787061207,0.637704296246811,0.8531999734802571,
 0.003641040260617845,0.3171047888870483,0.31230416360137214,
 0.47250965473333717,0.38481333235644666,0.1040073165194173,
 0.8438581131493201,0.5380110151689188,0.0018461866866380516,
 0.8363849775014068,0.9877653068668754,0.7019792469041007,0.682891036787531,
 0.2646811782649806,0.4258183276148494,0.9738875470072369,0.1171499272699833,
 0.34066453458135126,0.4196548469497139,0.9297139785226631,0.3605649353108774,
 0.4456389248090602,0.7470793650821347,0.4768699601360258,0.3850759862030714,
 0.9884853229771802,0.8694491319433594,0.6239626207048943,0.7569963519852673,
 0.038395818869278564,0.1919404874422631,0.37642052608516574,
 0.8422890037990309,0.5425441472004796,0.0815993212818984,0.8081621335052656,
 0.15194409354963168,0.36813700748752165,0.2830500439331278,
 0.6182647743340693,0.9582597649758471,0.9633278116760831,0.3235280426399326,
 0.31168495800427165,0.04970180061211349,0.848663788693308,0.6282380465005595,
 0.9533015518220574,0.5242960446428614,0.8375952769000747,0.9168615599512833,
 0.17176194244118048,0.07240807760661394,0.4218339089982227,
 0.3153601768756116,0.6144576186664061,0.06265686369960877,0.3571754642722814,
 0.6562988423378697,0.5169528394807658,0.933560977427625,0.7257518609374182,
 0.7629121291556928,0.5730758375490288,0.8866034518537227,0.396002859587116,
 0.2820433075892055,0.7731680480526035,0.9201763457456833,0.669450298727628,
 0.643200875333064,0.8297286844016857,0.27651651538029587,0.08432382660333904,
 0.24785455156677005,0.10055036551763918,0.6637597911938421,
 0.2565906476877018,0.35386161110396475,0.6239677816753207,0.7193084543258113,
 0.24658588027992856,0.017521579745577442,0.8786913465614237,
 0.6303056520607013,0.5950091643198652,0.5196763336894659,0.11444062473500927,
 0.4725444457721174,0.1466954191333849,0.3800487385104815,
 0.047000948052902114,0.7744168743019493,0.6122239226179852,
 0.3209754833485663,0.4897454738946919,0.4002643104181923,0.40157558612911814,
 0.24776117506677886,0.9316141632689354,0.9624097641207997,0.5651857848143746,
 0.6299511571789093,0.5872830430145207,0.6475041684472421,0.6221291187651913,
 0.355205386061065,0.9101149659969878,0.22214900037448204,0.18639259097557592,
 0.773578622593858,0.5272320400387414,0.22706505060297622,0.26618125538883586,
 0.9056999162312687,0.8828495506060519,0.028753597561102517,
 0.25863191816518205,0.15746291010875124,0.3831008415557867,
 0.04047256863224802,0.5195822248826292,0.532140014750461,0.9059680157262886,
 0.5124918340827456,0.3096347299309885,0.07680271800871286,0.6980623723980289,
 0.10488598928193349,0.6654649217736517,0.3074198790157434,0.2887011234643274,
 0.12117406454454649,0.6595690264025724,0.5461367417328775,0.3808150184623248,
 0.7673883660517321,0.5006530542541917,0.7646560047527355,0.14155010636380194,
 0.9383346796856601,0.16585269898432475,0.3368262353059952,0.9337423115121567,
 0.6448501900784832,0.24490379551353536,0.6346540858224651,
 0.27854565988826885,5.7954722417097315e-5,0.8256309007501722,
 0.7861618829180761,0.33696819910831055,0.41465364736743027,
 0.13382828577111838,0.714417319344134,0.5261046855620151,0.6868596124535692,
 0.28841645538446925,0.7220711463477416,0.6725419989017617,0.4136651964817536,
 0.09448850268745779,0.9642114860080362,0.7208208381738128,
 0.22652371983241737,0.3764950497436065,0.6950072669904251,
 0.28247428727732893,0.21933612434020167,0.3903574715181035,
 0.7363552926204288,0.33458339655805047,0.44026041716475806,
 0.9462775808103545,0.11944071639998433,0.3112725910973554,0.5309532514658513,
 0.9674716085745722,0.0678560191919736,0.5885711746223246,0.19849896014204704,
 0.21021528858010885,0.755519411919705,0.3295161255916177,0.1225408312345595,
 0.0987613613345577,0.16210528411271408,0.24556092985403044,
 0.4103146695137795,0.7578958701672329,0.6870020714311725,
 0.010257032315365455,0.40346577641297465,0.6126415552175842,
 0.29389460784061905,0.7346742537799096,0.6626700222627424,0.8963731963779656,
 0.07036866049502799,0.35100430896272417,0.2860824783945476,
 0.8324763835160351,0.22072223631311771,0.7751251413109479,0.9630726194506233,
 0.7156685564251254,0.513762885072231,0.8953875744097926,0.4101199538349468,
 0.625777835133855,0.3546166013577837,0.8855222758787773,0.08070789889511387,
 0.20960492887137328,0.3008913676082712,0.15669753079211868,
 0.7422792141916604,0.8554285410009668,0.42258011173692656,0.8155782880627646,
 0.49476370612955944,0.17365512978907016,0.2591695038989431,
 0.15115262342868885,0.14744719227709302,0.7152333476360946,
 0.6906559531114486,0.23909263969691286,0.43205490161647286,
 0.23146769859417615,0.7710079336414575,0.49987474289283307,
 0.12681830279088802,0.19516040611642382,0.21167459888981743,
 0.08431202253214898,0.6173658781843199,0.416241007700069,0.657221692195681,
 0.6184331283427332,0.7129115379609936,0.29993860298160935,0.7746783300762841,
 0.4145515967317589,0.8278563613349761,0.9265418523928135,0.5322503337605877,
 0.35505051354261696,0.8831453189287551,0.18881034569817146,
 0.42224180385583043,0.05971769397092341,0.945322382798149,
 0.15781986892582356,0.11540450704022054,0.5741310279226233,
 0.0807756825198459,0.3422416931127431,0.8122036011792385,0.14375271218039531,
 0.3981169966040301,0.7762639441472369,0.7819953758425693,0.2909497120080987,
 0.964468479683369,0.41666761678865627,0.5118517702792158,0.8388486583291076,
 0.42858185175036034,0.8787965613974669,0.14194309046903308,
 0.5391062487697322,0.5086740319962101,0.44369512532332733,0.5729577277988094,
 0.4864376436965734,0.0999445406408519,0.40333336177862233,
 0.23679180328348248,0.694684567339037,0.1964855862646122,0.8750435339402973,
 0.45487804979843005,0.43429861439966344,0.9027970844428136,
 0.5392838489864555,0.20915443160255343,0.8152952688685131,0.5739280699850404,
 0.4783516515201619,0.3182764289920945,0.020176233553504064,
 0.6107245870395416,0.5656810951451969,0.22486215184017966,0.3401694069552168,
 0.8302748805973642,0.021554249431101846,0.5340535055755941,0.774199150287019,
 0.39573079560752045,0.16113703843453953,0.34522611071774245,
 0.47006370822945165,0.23858290909322777,0.6549203400780524,
 0.7996110935963585,0.8955530283955192,0.23056881416777442,
 0.13421157982840226,0.8879236468894931,0.5969007259956243,0.4442263965323707,
 0.12394284576861159,0.8844192597399421,0.7292544804700593,0.5016618228863647,
 0.4379604503333756,0.587407666919292,0.16145250023766422,0.9001400436020401,
 0.5434852059041522,0.5395457062988294,0.5045930665144303,0.6260255674260815,
 0.8895479090000444,0.17426319621894715,0.6205212259306263,0.7519782779604093,
 0.5454452350137269,0.658688370890958,0.7652841126847856,0.8384202633496269,
 0.3145031683758215,0.39297606111396133,0.9573866514660714,
 0.24053736400029768,0.9449924553194152,0.9075172150032177,0.9600472217807883,
 0.7542906836333994,0.79407223751733,0.5853451443521186,0.1083736739772958,
 0.2246939467725414,0.5560778188385306,0.08052058546697871,0.6716955075756881,
 0.6087606675141548,0.6229995742746601,0.42709972606608315,0.8613548469258744,
 0.5473377727735499,0.15860204813761103,0.6988211720394467,0.914157786138243,
 0.7056684142212669,0.007381526384800607,0.7202313951137249,
 0.7102672166935317,0.3655909884714985,0.5593201582115797,0.9354738027409017,
 0.179522656647779,0.3568069696516173,0.31074187094727823,
 0.015844666118080997,0.5213568356248716,0.8339297579528218,
 0.8638609790684736,0.4627915931880221,0.7955255599162789,0.9072816270734007,
 0.37205655535182836,0.6162778673940215,0.5250607560313505,0.4592920199214011,
 0.12507512954850597,0.5262191312153455,0.4032295452572551,
 0.38743031388844806,0.8233505806841847,0.03164713122865537,
 0.012719761172501176,0.6854673221829992,0.9480673047721285,0.722117003257669,
 0.8611054900744739,0.7292080748850408,0.3285144844344511,0.5183279593192686,
 0.9802881342430602,0.6325568514961317,0.47580543906083705,0.2742807943149098,
 0.38874809226803697,0.4959613074647983,0.34148351692189327,
 0.7395417075484515,0.2657589082861529,0.4444219320242798,0.9602333438923589,
 0.3813653247040393,0.4895469750586473,0.7536556264727406,0.9396174146550711,
 0.9165804900226783,0.5366819566478914,0.10623073911337433,0.732432829836528,
 0.9778391871001791,0.22126454360966685,0.9707781475995131,
 0.005457055096988794,0.49932721647357825,0.899419847745326,
 0.06260190681313793,0.7388883194194555,0.9514421572620687,0.3577456628996366,
 0.6611138672190456,0.6080350984705784,0.09460114065803371,
 0.04678466328898545,0.6493583068644525,0.4130510201228592,0.7727351120981042,
 0.0353878819202994,0.23111354903913417,0.5068502620175865,
 0.28839974869853036,0.8002179194495393,0.24678636182883384,
 0.10178302523300109,0.5863109324491844,0.47920261039254297,
 0.6987828531600887,0.2811288693164302,0.2664500793369339,0.2483026058288037,
 0.7693702064607958,0.19978592191760847,0.13153085962232958,
 0.1776623989078765,0.2142078429023656,0.15319906874801226,0.9229886001033466,
 0.39141127940532283,0.30260001020857374,0.9955428870826191,
 0.6136282950471337,0.7163973691241922,0.25949023176674146,
 0.10775355843611312,0.14755236078591838,0.8877160326581429,
 0.057717546807907194,0.10438286469217772,0.13170802943918503,
 0.9374817338916281,0.6062556202291751,0.05090298338046195,
 0.20987305297305892,0.3129913968450251,0.4472507979695023,
 0.18899035920095442,0.8898021616593805,0.8906143803807103,0.7290709228865162,
 0.4046254180666137,0.8692286763457471,0.19860109729363962,
 0.029139765455260647,0.7111881784031311,0.870954843611402,0.5155575088349535,
 0.8179261470797712,0.20843912697661726,0.5099597808931394,0.6820953058505795,
 0.07418732979270515,0.34305191760584297,0.9475443550877323,
 0.4018288069143543,0.5515870294727276,0.8034393426193376,0.03925765052067609,
 0.9428596046630222,0.710101707180319,0.4650998991219927,0.3819330487741087,
 0.004168863696367819,0.647910424072061,0.522209861398151,0.9858981834222438,
 0.4505071838357284,0.7352491982770933,0.8859702232885661,0.7263649523607019,
 0.8412482204031733,0.3639183848826565,0.790770845180427,0.042606454756265566,
 0.09005603155227271,0.2116635997442715,0.8717572349468555,
 0.43567041266748896,0.4469682063799518,0.2522262844652138,
 0.44664890306313987,0.7105023836336543,0.8958626653467768,
 0.16361713551768797,0.6759949904505402,0.372278051176802,0.501226417655839,
 0.4394013033780074,0.5875470906173215,0.05475692507552177,
 0.09911768509366004,0.8369092595449323,0.9121420862777512,
 0.08912116646084345,0.5082204395608415,0.31568095542736985,
 0.9372114067189751,0.39777216119744996,0.5841998921979469,0.9068267871569584,
 0.5742109836172337,0.017856358593911548,0.2656466736342752,
 0.6063288010647141,0.39631709095631185,0.3094551574272911,
 0.036959603981364664,0.41058267639924784,0.061232001406643755,
 0.30681681683296347,0.5834714891331257,0.5349896021306205,0.6750029407407868,
 0.39066568881694375,0.44647760681014814,0.6040529179581047,
 0.16334585969565651,0.2159166803962057,0.6484240180325122,0.2978339222301075,
 0.5877025172956962,0.6666419858927385,0.8694420003073965,0.8444367782068206,
 0.3980335921219753,0.034923369095032686,0.5043355267216096,
 0.7873115982766307,0.5783787617268723,0.5345639032520838,0.6319849511148184,
 0.9323320844219182,0.9986033009696058,0.15346696323867537,0.5952762738008593,
 0.7588845875146594,0.38150898372132036,0.043620250197300536,
 0.9696023684912447,0.13520291144255225,0.016802849884165028,
 0.5375049074945307,0.5136169804449364,0.8243876861142233,0.6628305016096481,
 0.4737489823101987,0.7472727030793411,0.9603609226349,0.7415901195002029,
 0.45011855053397476,0.9036112696938857,0.07442524882707835,
 0.9698690984648831,0.767118575840999,0.9664785162402778,0.15599908145378738,
 0.8314104500999124,0.3975813228741858,0.6677727230931276,0.5873500220908778,
 0.8572972224402361,0.04429230037676213,0.8385704939412032,0.322792097966877,
 0.3941678362459219,0.9089740997955773,0.9663573447429962,0.6302586967279346,
 0.3500815869438825,0.12536231772800466,0.06871701658984819,
 0.5030047537342456,0.4255125593608561,0.4924683495001152,0.5914691738988236,
 0.7794494512639564,0.027100189622609627,0.3175900390048042,
 0.28620034129396976,0.34905993382257783,0.4512229533236175,
 0.3807364902003192,0.5790647275925175,0.42898263885320065,
 0.49539222588776155,0.09907981904854146,0.22945207664310108,
 0.8761721529145214,0.5663865001954015,0.9214631162132926,0.1047090281720976,
 0.804291260711646,0.3836680504330878,0.5070146451445614,0.4994581076878053,
 0.6226910368341648,0.1878351463549598,0.8549115278878792,0.561058178657804,
 0.5329016608337387,0.34830471709422584,0.4889965859077572,0.9381358064139346,
 0.35106179491747347,0.7112361072110819,0.3308148617715606,0.3998127599707296,
 0.013759412887474731,0.6198159948870665,0.2675880125855845,
 0.48600977600478135,0.826001668941442,0.4457488826696592,0.7882364459314588,
 0.534324048428976,0.13403022057296077,0.5917765103605552,0.7863121077281615,
 0.2963294983371608,0.15737397053095192,0.6300805153563506,0.5303567504390149,
 0.6156208754002271,0.9680152400283828,0.6227909125231178,
 0.047923129222386174,0.8742349359744186,0.8350213282840443,0.563000187407021,
 0.3947462801486181,0.35352437607751475,0.4284995340498785,0.3905449612624272,
 0.018309726488801914,0.7289235892225905,0.27601711023383024,
 0.7144736292927197,0.1824987129073864,0.8972739797276427,0.35043322813930966,
 0.13384133811737664,0.4025034917383773,0.7797329731677183,
 0.18474552920365772,0.25618887087476916,0.3547325381549322,
 0.2796850365385447,0.3584998618623787,0.03704133369156448,0.37358205523455,
 0.4679792269173303,0.26009865446024993,0.5125706888624958,
 1.6544197346113343e-4,0.7631280542030952,0.9385763651743106,
 0.7444972976928523,0.2529348632325228,0.39508031417446343,0.4333461598408539,
 0.6207247103027183,0.8480689157932209,0.5416859132725765,0.7141480020669605,
 0.16157105700134733,0.5762651402181033,0.791351985083326,0.9879541520398917,
 0.7198560313588247,0.16467812262623693,0.35316544130962346,
 0.9424254179526526,0.6296173137389505,0.8298174469242796,0.04865883747579858,
 0.29574867224319135,0.6367790314632927,0.6192782024738479,0.1336054041300948,
 0.6642682209324631,0.9430151820870268,0.41602313698848115,0.8112050065683747,
 0.1663416036424865,0.05670430175667507,0.2229901617854878,0.5198048703528755,
 0.7151283503999141,0.01826799107865207,0.6466555866730526,
 0.13991556458862764,0.22102535896748898,0.8635720188383041,
 0.4387797076406186,0.15190324237203467,0.22749784224142766,
 0.9900663244307637,0.06905074873372385,0.9596067743788033,0.8509837090843302,
 0.14020432196768928,0.4143226087139903,0.3845322964480645,0.9010252735003864,
 0.9882433504222337,0.6185827703267861,0.8595461083765538,0.9040028135344659,
 0.2223025337831026,0.8454134365692959,0.18351512548692872,0.9588120962255124,
 0.3264734933606863,0.40383047088742563,0.496720495636073,0.06961425066332172,
 0.20863783761463217,0.42055766081448775,0.2971388744758894,
 0.5663402421843199,0.9371067882243026,0.49206177033044396,0.8400212710108061,
 0.0603146374986524,0.2172773655479695,0.7445911308935338,0.11985094462575585,
 0.8652567312787058,0.2264170985109819,0.38956095762856313,
 0.00204896769774654,0.7828268932831579,0.47582391339768493,
 0.9572736433001375,0.0673252762038421,0.17414074491565923,
 0.10673741230794387,0.3190407589748935,0.26136886570288553,
 0.9100272881596027,0.1850538893278062,0.5718757433270922,0.5088467930385892,
 0.7445961821733308,0.2348874029255542,0.8360738585437624,0.10465611181332546,
 0.09153292216916065,0.8269002069189381,0.9566145634310693,0.7222494245271529,
 0.08151313152284345,0.6020224915396214,0.5233798720993973,0.3821392161400874,
 0.18307568637586535,0.40205564576813757,0.5472641065798196,
 0.47193963020541774,0.22211123946656286,0.7238142961268893,
 0.5875339597611486,0.0026529288308292376,0.6255516230816184,
 0.8694893875261658,0.8949721629925564,0.3617145404260119,0.37296306102214694,
 0.5761279954900407,0.19403451848390108,0.35394777185446047,
 0.6960298055101971,0.5904759453633331,0.45987879921502084,
 0.47170399221269455,0.507491325938774,0.3647006184204621,0.11270489721246402,
 0.9822902833177573,0.9773557459119487,0.254399811473523,0.5017740050901808,
 0.6969092231599698,0.6118032221008892,0.9504427721797524,0.5871767634530451,
 0.8862364152797704,0.08333447046062048,0.8069741138716329,0.5189846948245198,
 0.6834518021466068,0.7308638868744541,0.1774240526247418,0.4893945591265425,
 0.9388152112548203,0.7646860964839403,0.540532193427115,0.5916189490653406,
 0.8497679809006324,0.5532886359175265,0.33277963596036153,0.5849385202503938,
 0.9694650410141474,0.0075533292515571615,0.05733039457821376,
 0.7455400731849804,0.14427829810705783,0.15556161876082708,
 0.32799356996161577,0.2851365385392399,0.36138218663215327,
 0.11389165781411392,0.599603557013285,0.6788764622810262,0.6929182381305137,
 0.18063249847481821,0.5077028882676506,0.4934918223141612,0.4611790884621749,
 0.9051805323767375,0.7666019014090438,0.7393364655860558,0.7898243484322456,
 0.39475308924940344,0.04933458312007999,0.6028166808029481,
 0.8002180984304743,0.25110535317129756,0.8347244250691661,
 0.09282599496120814,0.7924069862974936,0.754347256793087,0.9869336930314839,
 0.12827781991292475,0.0012759868714919964,0.33414134218619096,
 0.3433426199560431,0.23356176820825414,0.16087530786120197,
 0.6768533047564889,0.37818214813002415,0.3822750662720109,0.5818755187075626,
 0.4407289355719999,0.008361635543740809,0.5935256131291918,
 0.4168522700446068,0.8051226437400887,0.055082626141680135,
 0.6992871079742811,0.9997749621430625,0.653859416709976,0.20910908054509025,
 0.4483866389335728,0.17817527398725297,0.32620777499913456,
 0.0015206702633721036,0.18616494242304282,0.29541115899900083,
 0.1419074581826597,0.16774358119591626,0.3018609018777729,
 0.09763716041012194,0.506811561646874,0.09500252880255311,0.9877654035684464,
 0.43066552108571443,0.13325130588801914,0.8285652653229915,
 0.5033323419004311,0.8082642103552686,0.4735306247138624,0.3107672373705651,
 0.9933551693774053,0.020068795968908093,0.5016814138272249,
 0.9393053621764827,0.2614084982335869,0.3682034064512889,
 0.004070319309022241,0.18565093468386684,0.031138350033017126,
 0.3815326246783317,0.2621663432636415,0.14908508919852625,0.6297001073651506,
 0.015079247723068123,0.282974460014506,0.6105849802058536,0.9208355427828884,
 0.34935091703765364,0.03702394204963888,0.2759114024386531,
 0.9001474221408728,0.30108844076261054,0.030309153038923675,
 0.8077818732050274,0.40206076986272843,0.30425531681477924,
 0.3292361622757122,0.43578598332725016,0.6286482639616606,0.771240136976808,
 0.15211129758771036,0.882713643313749,0.25783255318666454,0.7215596844423174,
 0.8093777829534727,0.9187550711444258,0.877199788509755,0.7287617656517732,
 0.9301365394857692,0.8637363941531212,0.47335820159514197,
 0.16402008051663985,0.9358114161891745,0.18030960354503567,
 0.15144252685043746,0.7110476920186588,0.20734657911086873,
 0.6802155712573517,0.7885904596959531,0.8387146835172299,0.6022321186540015,
 0.1648781162626971,0.09001379577884394,0.8481988356502246,0.4775258471883541,
 0.9835121750445205,0.008834236052270183,0.47581199971557164,
 0.45920063287673507,0.13377011319085064,0.7264480202334681,
 0.3418602283263621,0.5878270556324092,0.6639663568815208,0.9231666061907293,
 0.5694386394626108,0.902736525829967,0.901078805289446,0.5933995856667154,
 0.5721473291339514,0.17855153487399966,0.38429042100449384,
 0.3016433656335701,0.517311609259637,0.22730930873831812,0.6724308016264735,
 0.9736916999203629,0.7603414091919738,0.4734169414214987,0.1472308661988463,
 0.7928704702845139,0.4791084107147635,0.19091489249015248,0.8930249050320228,
 0.6240658192088082,0.21082673689792464,0.06001105465049317,0.262607944292385,
 0.039300781992709255,0.914712489167089,0.7667611769863978,0.7031120852510899,
 0.785937873927535,0.2528793283709495,0.8991946867768168,0.8563578976801485,
 0.5793976142381217,0.00800423132176531,0.5512470243837594,0.6789687686773707,
 0.1735850567465973,0.7033147496494959,0.5300294267247988,0.4839186359991494,
 0.9421610817060216,0.6894082674004116,0.2553813451174304,0.8734354694880402,
 0.05423975290702532,0.9981757547990594,0.2385918733199065,
 0.042619433703305454,0.20306992234005172,0.8513682716754244,
 0.7438163886816991,0.8159712800615766,0.19231273668923154,
 0.21925211127857558,0.8735792582166071,0.2784244616840511,
 0.07965297487521217,0.03334145506502373,0.4657087588760035,
 0.04406776253767919,0.13570355144890067,0.05417152242097334,
 0.36228455015565963,0.09330405391447094,0.07476826914529167,
 0.9771930924664338,0.4099097331242262,0.6061619614864335,0.9469185156002174,
 0.00837017175865229,0.10428990090740942,0.4079198365927347,
 0.26851064093003163,0.8028189124987672,0.3819274979713767,
 0.06883987389845292,0.24311222992532788,0.8675386076695851,
 0.3016123754735355,0.34945552947828007,0.28209444924446414,
 0.7864182925665192,0.2740791694949909,0.1188840167765356,0.43237666547487763,
 0.6393435902667813,0.5596002685141487,0.6079625682407008,0.43413279243147507,
 0.9935671272543055,0.4926836160186847,0.47696399594695216,
 0.47028210218100586,0.7951298525494614,0.33912879302304133,
 0.9306798361767518,0.466582353954082,0.5934704689313369,0.9622581954242784,
 0.9424837482495786,0.3834359604039076,0.7280076965364973,0.8690361176817185,
 0.8666245759639286,0.6232035476231099,0.17434156147675572,0.1684314306410596,
 0.7750784271951139,0.5123925552259074,0.10509985126042953,
 0.14414602211343563,0.7298846433580753,0.7245362198434723,0.1272790108970574,
 0.9112024233427036,0.7375256602467946,0.22622285644936202,0.9940716746129082,
 0.2853890891856702,0.0811875493463694,0.3510935625929048,0.2185721000327333,
 0.32777740520973053,0.12214639167349117,0.7086074468749934,
 0.7887066538083686,0.24958125589394153,0.9420092209663383,0.412155197016253,
 0.6553081471917607,0.919184599032126,0.3557113937932901,0.040480350333955584,
 0.18764871360518642,0.9419014456885739,0.8682475216581258,
 0.39145699464999795,0.1977914183816657,0.41726603903870696,0.480975516400433,
 0.8996059022043479,0.18079289807357057,0.7186562294819065,
 0.22239681751854845,0.5232006771715846,0.559250388182671,0.8904404296659784,
 0.44938344368529504,0.40853467134131527,0.8069730079573748,
 0.022511994090501974,0.4737132816414861,0.8373127929250265,
 0.7209104163632115,0.40977990088150285,0.8624499379128362,0.6216094798406326,
 0.22627387098953988,0.8565600329348918,0.1291708795864277,0.5300427131825535,
 0.8765318442754992,0.06237012127290864,0.6901614764147794,
 0.42721641544005395,0.20852695601336468,0.1637856522110197,
 0.6251875942425911,0.12171774461871976,0.8412438877584386,0.676435261803568,
 0.45265099159483335,0.699590929726976,0.5584920755074011,0.7602122916085253,
 0.9857663006154403,0.5725957465410823,0.2832502980064817,0.6638846020047741,
 0.0088281529025942,0.18455796691133097,0.8331816660628159,0.8930026433739925,
 0.9750582910900483,0.18824832086713505,0.16204918841408178,
 0.7915347307126807,0.5713643273627799,0.23181145921938118,0.6627299520497275,
 0.690948715072856,0.45025765902961234,0.2381585151849308,0.38661997479277377,
 0.7599187804846095,0.9276322464080149,0.35795143600932466,0.9697434460359726,
 0.39026301791338036,0.5476402783647424,0.32829349577311007,
 0.5500093931722254,0.15233763230606734,0.2953526562844778,
 0.07668044374079941,0.16120544666528502,0.1466669170884931,
 0.5833191982593091,0.4316517973047741,0.35203051007517916,
 0.36707560203666445,0.9891095785974717,0.23655397846999227,
 0.8249953788296046,0.40995723623371816,0.03081641569486271,0.496140692875385,
 0.6869474284820853,0.9184552383104007,0.5291070928451546,0.31123098653296966,
 0.30579412371989456,0.581412254457526,0.12654491190185013,
 0.22169933701774414,0.5875284782705597,0.4846042045794421,0.3499353074700853,
 0.28028040679284616,0.7424991167175651,0.6913214893215285,
 0.47630293896867015,0.44448948078936024,0.17962946605707641,
 0.18063711095133672,0.853483918491208,0.01781928751563555,0.6134069011051212,
 0.09295073285768984,0.15200552707456416,0.3683976390647965,
 0.19904944911685374,0.06027993748668614,0.28473396747787316,
 0.6846717132590018,0.9177915194415011,0.9844328421359587,0.4333292738714527,
 0.7861257400920156,0.3556975518689157,0.3832512079079,0.9032154191221623,
 0.18962300462069093,0.22140811114387193,0.3233741140507024,
 0.25041411340844233,0.6198748277187586,0.688858200196034,0.07985932029880394,
 0.3018049593449438,0.7129767929662894,0.7810582862690293,
 0.055304236175362176,0.5552986197952832,0.35039953751016806,
 0.12750683560721432,0.2075490368997066,0.8391666354652627,0.8874559112399619,
 0.8198594635250888,0.23798565359643686,0.7501099250490604,0.7928396821228602,
 0.4985273919818132,0.8301955324599478,0.28137833371361265,0.919194493306492,
 0.4106417494714628,0.6150520460161368,0.4904712997123668,0.2399285249797416,
 0.42023144088902575,0.4311831135829096,0.2188605634951677,0.6380504029828981,
 0.7424446561011119,0.37984088409793393,0.33897741749913746,
 0.8076298081417962,0.29302777954658665,0.8299428737272665,
 0.021227745050835467,0.0649143058225371,0.10744235971629945,
 0.9693576971442435,0.6000473799963892,0.12902293844993518,
 0.39764636030757616,0.5780590123324998,0.49621342192877393,
 0.6358828600350023,0.8151888314519946,0.38960104039338017,0.4449974437316204,
 0.8629185667638044,0.20065888526630593,0.37195362451097425,
 0.2590468087972627,0.9143663515345042,0.10873017889084335,0.9725202223546889,
 0.44102914852823205,0.8674525249000103,0.2556200902039294,0.829014029346787,
 0.1895637178017645,0.3808757780897939,0.5554707624539029,0.5794180065197744,
 0.9987221615667663,0.4831992067253016,0.9480789516107011,0.34957662387119726,
 0.6303351383529414,0.8990330566671929,0.1373951319477028,0.5561749084096288,
 0.6266356875011141,0.8382005965078521,0.9381028589768907,0.17700827582959033,
 0.19799425900564283,0.41785809923815864,0.8214058706061259,
 0.8403687268247908,0.9387181852411365,0.5175550598712765,0.16133387905799645,
 0.19964437193726392,0.025031113628077972,0.7112630364256449,
 0.6793266327495784,0.001135465277302239,0.1281627153731416,
 0.8473637089385491,0.25583673148199626,0.8367937833081421,0.8701785717527306,
 0.4177188412247158,0.6288958962703557,0.05909346230430301,
 0.27470112598299234,0.5886232969407171,0.0806370688177872,0.758854864401691,
 0.48687770450845325,0.583305630091364,0.4755520999014947,0.12455264837420266,
 0.27191851029233616,0.5689436893653094,0.2686738743311339,0.3927182106841044,
 0.7101426015222878,0.776331080557655,0.27077762651011417,0.8415027896639169,
 0.1863761845423596,0.528041146570649,0.12680848712695303,0.665184258884102,
 0.4345028494599321,0.7865581287318398,0.7896295830803324,0.6255059439277436,
 0.08868461427576546,0.0980377572523804,0.9088394158717377,
 0.02545500653813393,0.38356346563674104,0.8387674992803532,
 0.9853686082516105,0.7092406552604824,0.012027837906603,0.48780858194917065,
 0.664667486067937,0.006222846487449751,0.697037032737504,0.4851686067056522,
 0.41362678645206596,0.5833891993659183,0.12448431230227941,
 0.5103980176072944,0.5174610740798492,0.9324258882599407,0.3311918085814627,
 0.11627009396277788,0.5849503252677875,0.5726288528447627,
 0.40058796106555017,0.2116568111089826,0.7190662659252918,
 0.22777660515846287,0.326450470986993,0.006428769079010466,
 0.20400919929796968,0.6731155647022872,0.20245322158604406,
 0.5278329050004039,0.19916838827738825,0.02354770591867017,
 0.20184411465453023,0.9508669581802355,0.5907047568060899,0.4318492763453097,
 0.6441629813365671,0.7613268904278009,0.05679476723466792,
 0.016335297161980455,0.4774924197859435,0.21247758781120307,
 0.8158657382269047,0.11716072574796277,0.6211325540508494,
 0.20661417397528747,0.30393092195166793,0.5308052216643315,
 0.5777434413080309,0.7288165790358011,0.5870645144875603,0.9244284748192723,
 0.44328514049915124,0.680645171509508,0.2841581997014955,0.7571070715966621,
 0.032349800290926156,0.8554417740664,0.9580230309620781,0.34729463934169924,
 0.8945596307888845,0.8748027878280308,0.7772852314520035,0.3615007770640689,
 0.11656265153568512,0.2359204413373439,0.05082189235153256,
 0.7126484618433032,0.6078900229963867,0.5872814262788667,0.735845217047338,
 0.9827403123809331,0.8901553130298024,0.15873698916450418,
 0.14504507094985308,0.6529359490544581,0.12408781381514133,
 0.5022435445512057,0.8817584516133233,0.6403535863014991,0.1990337888000122,
 0.9333352863509243,0.792040594130073,0.15250764023163432,0.18810695249009224,
 0.11470926101495371,0.39771347445492156,0.8106465869247901,
 0.9458781748722436,0.29106438105510024,0.8823565502229342,
 0.27600343266466476,0.15003124742078666,0.9307888435757883,
 0.8230852671688362,0.8980116540841035,0.5903209390995781,0.49898969235397017,
 0.47309283059923124,0.9064538123722561,0.09152391797102066,
 0.7325562875171882,0.2651558050183346,0.4564853267509651,0.4278737195282917,
 0.8202532020100575,0.4159246887696456,0.35633420013184003,0.3153688643471605,
 0.5686005760097153,0.07241723972344138,0.6000775732955045,
 0.13746095832143312,0.405578289818901,0.8285859813753734,0.6623188738181381,
 0.2233852452038282,0.19850067720395737,0.7020234463826792,0.2590074221086265,
 0.029894870310036237,0.8894071965277268,0.13924685810456772,
 0.619428750678871,0.7461835203823024,0.4862209414485773,0.6771042897589924,
 0.87096887503045,0.9497314945703261,0.8073884555436633,0.7032315120545243,
 0.3414364840506081,0.1143358300991072,0.24638519967782846,0.9126570963203072,
 0.3617328204263863,0.5185539461938059,0.3648728420651065,0.3778377373899955,
 0.3565241763809567,0.10776305072330761,0.7418981999416151,
 0.20529577409437505,0.844918277293465,0.5215976933287747,0.984306558758143,
 0.792818626681688,0.43965155684406376,0.7333437851078117,0.5657472964389562,
 0.3894025413816782,0.8065217192438827,0.6557776961252193,0.7236986829932334,
 0.396408998968776,0.9813288387412038,0.6282840676692356,0.7634674323289311,
 0.9589480078926045,0.06486801896406102,0.4362240790967449,0.662104093515421,
 0.3330566755042885,0.9400471892686326,0.2617355767897088,0.5183815442564624,
 0.8423152967225349,0.658205712436348,0.9501208740935881,0.1908741577616666,
 0.9429416115990643,0.8659611799340368,0.3929885329463314,0.07009699221482735,
 0.6217334234377969,0.5514516257681666,0.1836741014266834,0.17625745707660667,
 0.7876438329378479,0.31823078390401405,0.17929472387933199,
 0.013571319046897234,0.17293775418933055,0.9191514359643528,
 0.8787981314885619,0.3422691627988498,0.5450052371308558,0.5514017268730409,
 0.7712718098230102,0.6952316442906241,0.2310656189797088,0.6181882731393868,
 0.3208416837538425,0.13317688790710203,0.7892426481467093,0.6117678539838682,
 0.6869211531435112,0.8023003349452258,0.14010387676554348,0.5061428334087752,
 0.32328735113116747,0.27261491537385607,0.13889861785678292,
 0.9666199827653094,0.596705498125628,0.06843766756303826,0.5280014909724093,
 0.32895029905234463,0.2963871790110736,0.189922961166901,0.16012962726005486,
 0.37614216991910976,0.6750581261985446,0.41853687919204374,
 0.6801446432278437,0.10423554026863613,0.5681352548863356,0.5253639359580895,
 0.3334086918025564,0.1592995851579735,0.992455167411675,0.9868855090157155,
 0.355275643576928,0.022103560213881934,0.5300076311381435,0.2879033838692222,
 0.1680162889215131,0.8093075392481051,0.5867125569889974,0.35529980363121516,
 0.6768996146101811,0.14588112853580437,0.5612743176815217,
 0.02101198072509758,0.8298135912661163,0.9223952345398477,
 0.33997642795130223,0.1769312994961283,0.10877137544132287,
 0.4614496458882509,0.7215432593293681,0.16026384217890088,0.9667571579027584,
 0.8526314763510885,0.16196915537246792,0.22993037535626504,
 0.5143778842348323,0.8440655315192365,0.8101876240960431,0.9653357775374529,
 0.7123487253390925,0.8279177535164556,0.14202589361920792,
 0.16469973457147014,0.022183099385091554,0.7339874491151245,
 0.8102312454102211,0.24757873315752532,0.4812133539415355,0.932448121489741,
 0.5399095272923359,0.7325666444697402,0.6652604691170215,0.4157469888580191,
 0.6229042882399397,0.8104152719399497,0.6713547816505102,0.4835872115551165,
 0.4224859015681831,0.0012349221420081546,0.8348943756411815,
 0.09708215173135293,0.44601250192027175,0.19953617104384136,
 0.5510455399209293,0.7041476197702892,0.5623232645994709,0.5933467055154865,
 0.749285349473763,0.8708860401970114,0.9301983241471985,0.29643002531965146,
 0.09235265983925889,0.27919736802839934,0.7297090767642005,
 0.36363801779314553,0.40033566663707953,0.019426388210711854,
 0.9966553513393979,0.359356410007066,0.5605743074100751,0.8917186333962205,
 0.9587781891320868,0.982725424769895,0.41131344479766574,0.9525907916244943,
 0.9063434614156644,0.7286130474795189,0.1731259978896288,0.353412798523311,
 0.27534582764868576,0.6212426279506408,0.9491743226349209,
 0.35992352758766244,0.4212173625984761,0.8967978307032856,0.9601866884492213,
 0.10393204860451277,0.01926908505566105,0.5070580236006931,
 0.8483004984467627,0.3499114047381263,0.9730427976749565,0.8915202120144824,
 0.8798886986143861,0.31444010736937045,0.377608183212814,0.2399654153341786,
 0.5060891572609892,0.5430206602805566,0.5115114143540316,0.7127726444024579,
 0.4996236932653624,0.007218825006779728,0.25210311101742566,
 0.6766456964186348,0.7611853672961124,0.674586180088175,0.9505768458651184,
 0.5285930390764797,0.3365144783697559,0.7753144943442506,0.05975529810373892,
 0.7821659864317626,0.6222417733274205,0.9924024119116985,0.6981729184625316,
 0.5228680770343124,0.9598221695702222,0.713528556682125,0.906133708619335,
 0.24276034474955488,0.17447160220672775,0.9772897679112175,
 0.4024513861987281,0.1631588451053213,0.7593185331254482,0.21018591775956375,
 0.19055991794403693,0.39984110388588123,0.666496055833425,0.3216422159350124,
 0.5446453033250144,0.4071239313450693,0.33001053186466434,
 0.10943051390033431,0.5204599043796257,0.3792347964529037,0.1662670826944005,
 0.4395037010499736,0.3644574991064097,0.2749662960308369,0.2638525825289333,
 0.13276786113073913,0.9465247569038197,0.1870001746760277,
 0.06077187546975704,0.5858004275640769,0.46413694846842013,
 0.32899877645236986,0.6316956666558103,0.03332778512046797,
 0.43515390058237635,0.8353499322196815,0.24510910224685611,
 0.9660014485014506,0.6048354849794397,0.6138437253356966,0.23749029300426427,
 0.3268785593574989,0.9889056731482259,0.2499016449311131,0.25288119465965464,
 0.7513818518512372,0.6087504157575139,0.8530498477383054,0.7563855130567542,
 0.15063221651626668,0.1417844735033451,0.8911089243657673,
 0.38925989159954133,0.3378313150337783,0.17448850258716997,
 0.6222440417884778,0.7900728958749763,0.9428662948055658,0.3776110830887007,
 0.11941964165956276,0.5812145282673852,0.24993883222048718,
 0.6580621156308102,0.21053549709709007,0.6903716546634896,
 0.26575089040377986,0.6201465396213636,0.6844678632409842,0.3501541512912997,
 0.5835629141518885,0.7192824892952981,0.4621399480576801,0.01717143760677886,
 0.26539766361980555,0.08996544186269784,0.4045352286844286,0.283415620823785,
 0.8678516990058444,0.31810846999147946,0.1538984489485652,0.6939114209658134,
 0.8031193974674293,0.8207762794206609,0.8177920194121686,0.6061609336489541,
 0.16139606293887931,0.31939436288876577,0.015242287108474417,
 0.6803013926122825,0.9291957823870194,0.06319911959276259,0.5814281649784144,
 0.23104623574184724,0.8052430689075203,0.8233357783707435,0.6746414879468297,
 0.747766280765612,0.3775863692344149,0.27738085906052623,0.13462257199920924,
 0.5676124383766599,0.5439915324358227,0.3798692918509514,
 0.017613187789534135,0.01874902051336036,0.49583609938411666,
 0.8294192267798919,0.5631609406667034,0.6723617629438033,0.38563196027241275,
 0.8263453041643328,0.47338170838784643,0.9097888586552204,
 0.10569387529762742,0.40995350831781274,0.7061439847983905,
 0.5962978754567094,0.8242933953914405,0.3188509728676576,0.3122803858670745,
 0.9804629805573412,0.27479457345281577,0.4313110705954182,
 0.16741573189021675,0.5545271004491754,0.09785183813596798,
 0.48606840386013817,0.10194182881034375,0.5747667555101592,
 0.2295679058098724,0.6149079138268047,0.40121677255828336,0.5834575144368013,
 0.9135317621199496,0.9321738767137867,0.08188089558972478,0.6794308457597869,
 0.7323600593980154,0.7526890091483089,0.43835920338222545,0.6419328136974984,
 0.4616393764635244,0.9515714888671095,0.16925030149344666,0.1544460022187938,
 0.8925601413484074,0.5691554894536583,0.6637893855659026,0.7138825176491658,
 0.980568283261245,0.43723408788206,0.16129612373815272,0.059797703359472765,
 0.5171334996081104,0.0980804010227806,0.6138106649920957,0.9590491757214394,
 0.2644698863543059,0.10219442482171315,0.7352744715176699,
 0.16756258496648146,0.04745844515277586,0.7978606523178721,
 0.5195294362380872,0.6700082403753146,0.36124842217688813,0.4168510135341169,
 0.8357183445297186,0.1295533498695538,0.6314065477283258,0.9412894631103967,
 0.9993165154779818,0.23724476900207947,0.3723086905381159,
 0.22027114349226162,0.2588364208436912,0.5684153962240072,0.4314941590520708,
 0.04175328264231093,0.720393418730404,0.6370107566401484,0.9780619006920577,
 0.878136580985951,0.6264269271542623,0.7932737942537965,0.8302240648312696,
 0.14907728620214888,0.7285989676803666,0.39614545275025526,
 0.9780343686398164,0.25633529930959553,0.6867879289091917,
 0.17941203086320334,0.7870156224411861,0.8249672020638701,
 0.40756968344941513,0.6415773290291079,0.2837158315140722,0.8326995175319033,
 0.15165290969309952,0.07931360016528477,0.4597393424819515,
 0.11409483673408771,0.9102808999971872,0.18654884222844148,
 0.003099614950220264,0.2500945447078027,0.27779642674556626,
 0.026981815065995618,0.7749874668165672,0.8095039205081987,
 0.12917039163699517,9.78383015582951e-4,0.5530647865227802,
 0.3293818773317525,0.012322443027856234,0.15677881037389185,
 0.7549515264777786,0.5082302421820888,0.19327288677431065,0.4536595107236123,
 0.330653455946861,0.2660631094622671,0.32934527671056646,0.3527637593432349,
 0.21657056845256517,0.31330536768617645,0.792177118982571,
 0.28883326609687576,0.92006311678587,0.8846146305295732,0.7007030446213748,
 0.4501739990326348,0.2651135438680945,0.006523147426183207,
 0.13403049877994566,0.8479498815616018,0.5448545287206459,0.916755713895452,
 0.7055313363452488,0.24591035089814361,0.9268345372417206,0.7438053477682623,
 0.7988364913751741,0.8870973151716857,0.13404651151601632,0.4917077389193849,
 0.5035792491448273,0.9858271270535783,0.4242924584704776,0.7185607976338488,
 0.07102197244347264,0.4088076935517967,0.8606488808574103,0.8556950180958484,
 0.7355074002887362,0.2685097666244841,0.5760505305755024,0.4215043184296442,
 0.9077827851355897,0.24367792418558087,0.13667293177618955,
 0.6146837374045478,0.6181139099985862,0.3990679693810675,0.8519351816606261,
 0.08988051342320924,0.8973277068051355,0.36663708044603327,
 0.4759506282637943,0.8085467533956995,0.8720158190822616,0.5215498590969226,
 0.8701836935741236,0.6196064900759827,0.869620540156914,0.09947656534932348,
 0.1369520430242247,0.05440001200125533,0.0711553292955549,0.9426342546178804,
 0.9341044549292782,0.021054637647048513,0.8856966911848431,
 0.3887551035533984,0.40318954158204245,0.6752325525069405,0.714914916102433,
 0.8643165701046416,0.6225898605400699,0.25447509407229374,0.8395900783718955,
 0.15546700843977068,0.045497540711304296,0.5495109967670389,
 0.940948174316499,0.4189370749751453,0.4891476445419106,0.10757025511995555,
 0.19790808225933143,0.9901821029466027,0.3854289058101006,0.5598906068744964,
 0.7353077035899039,0.30261206527863793,0.0643751742924984,
 0.10809366945956689,0.27223086880053404,0.7297738499586053,
 0.5506461791680524,0.22205703853914505,0.5193329235330226,0.3305063827024427,
 0.31193860109043714,0.25718690921764953,0.07333713256600927,
 0.7146275514694579,0.25730304655060043,0.626830683343529,0.07484018078523391,
 0.9329919608541843,0.18964950800256275,0.06282050763880176,
 0.3798559434253551,0.8842112550536283,0.6698077768958668,0.2621785728888021,
 0.31038579364264596,0.762587312903286,0.49484392315449666,
 0.16161857881262542,0.9691570378651827,0.6292845118754814,
 0.15206560917636724,0.60344347761827,0.8614074120502938,0.5195215909575324,
 0.7352832799819273,0.9389273510370967,0.15953308544624,0.3673564183398911,
 0.26543028772392807,0.6949571818584708,0.6000248773385733,0.6894812154699497,
 0.4156473987238203,0.4958719921956133,0.8967281352173865,0.7065464158856217,
 0.7934487334167126,0.5833558667810386,0.5955340534572955,0.34022964054020854,
 0.9092798620842394,0.9635852388119697,0.8192329060241241,0.8953540063985136,
 0.7976049119948856,0.6323464754512885,0.620287858170563,0.7099981821409818,
 0.6487010713505666,0.3371218042695163,0.4591929983747667,
 0.031809833829237744,0.5030785696179221,0.7873652759823938,
 0.9677013024464786,0.18264586888673429,0.4621110063251087,0.3338354969797902,
 0.5974471244280942,0.9164218774069404,0.992203917971094,0.816825895134386,
 0.6527826618371951,0.7136374784358127,0.7129389940937368,0.5046953287003124,
 0.5727086023180092,0.4302921891235597,0.13606889132255784,0.9834863102261451,
 0.04016642640611967,0.00504030292979174,0.4416364742269264,
 0.8447267870090664,0.6040480345537032,0.47103663263025175,
 0.08275351059291525,0.4824932520946732,0.7543644635551326,0.9485114091143988,
 0.49258824200203577,0.5192867758097657,0.6102656853600812,
 0.28439735482289386,0.7135673554859082,0.635010813686559,0.47622707614550686,
 0.9600667311754132,0.744953100393173,0.29993039234428154,0.322136950049854,
 0.32118473875910947,0.11652654627642178,0.28740649128042683,
 0.5174332664190943,0.7405067995248956,0.2743640013195059,0.9571067681132518,
 0.839341466845988,0.7410565650895213,0.23748122173129538,0.6134983972831746,
 0.6341543848005542,0.7956478487031678,0.08716481298744994,
 0.43121881058714506,0.46645078810289364,0.6432009386768114,
 0.13191497697371812,0.3600474665548462,0.21210638617007693,
 0.9710320047153898,0.9424747115904726,0.5882729783916539,0.626154821459226,
 0.12077508892658961,0.12258823833646548,0.35608657455960213,
 0.29929003268333987,0.8244199943012337,0.6987543028923374,
 0.31448590183645075,0.5635009545988334,0.718337280798449,0.9751826264694736,
 0.36394517286237993,0.8056800806077117,0.6559452605312939,
 0.43384740264641386,0.4291586655117492,0.7359249393036558,
 0.44491767058415976,0.8346035331039667,0.4156317914833503,0.6412041505095957,
 0.04647175263212622,0.857766794399873,0.24537649733170264,
 0.30728326287218355,0.14421630696511722,0.24503494248223623,
 0.4387027483919351,0.69705893826774,0.2090940541166485,0.962472019536153,
 0.7309441668340199,0.008153268440534545,0.07074103435863932,
 0.9866386718042495,0.2756161770365536,0.7386154960978645,0.9207776268236183,
 0.5886464764823969,0.9882740254828001,0.8800621374400215,0.17445315628598523,
 0.23719312068424503,0.04882563008387519,0.4964965727603281,
 0.5611212599008285,0.7698288770904667,0.8245365502491029,0.7265434283343217,
 0.6752290440796281,0.6907183528441121,0.8065593749208242,0.5631009156792609,
 0.38048727636129154,0.11747244674753266,0.044980071572102265,
 0.23267312087005598,0.06198482953838269,0.3781106433952369,
 0.18740885365802773,0.48983991178427944,0.40371730986099286,
 0.1749573865689522,0.03552051469766282,0.6214465784181142,
 0.26024669188275384,0.9860515903128815,0.09955411547175297,
 0.2653418912365244,0.44490851383341934,0.7065407498688852,
 0.04734454338829064,0.9104429200874429,0.25680376353719225,
 0.5871655626682702,0.8347217588559469,0.4918169983418317,0.30222404928342184,
 0.5961826868224284,0.3665000368540524,0.6093689845204,0.08749530637580771,
 0.9857576833857309,0.11340235220057693,0.3303390942174388,
 0.31354056187833734,0.2748075651514287,0.8243562742557434,
 0.054692750807851964,0.5138363806697995,0.4454110755234293,
 0.7763839311996912,0.09669641026715459,0.9379782000897793,0.2220868304285587,
 0.9233533784819142,0.9624591309506,0.6325017567641985,0.38533395766269696,
 0.22131811598244477,0.4584717651054081,0.02308054634399359,
 0.15966470870084604,0.5111399821192455,0.731069067445185,0.3162955437506465,
 0.9283813046204435,0.5542525262587799,0.7597000041243751,0.38938876752749774,
 0.8640475076889634,0.03363860339849811,0.34634658118121686,
 0.1752579002146586,0.2853119097768133,0.061019748821705355,
 0.9092211138399068,0.8812237337734229,0.0708387034816963,0.9888078131931618,
 0.7068081048960637,0.41982186728151616,0.8300136703427671,0.5540602949329008,
 0.2670629891213713,0.17458880411885103,0.18382495285915867,
 0.39732975229364564,0.7763308271461266,0.831203746370419,0.8629635568925846,
 0.6927949359282622,0.07424259380571852,0.24178122536908442,
 0.3520405509672522,0.7428086638794874,0.30220541649731136,
 0.42178705891156143,0.7221803362525093,0.23041339830237262,
 0.8550905500123427,0.44908160155579036,0.010633689222308562,
 0.9914149752625937,0.6598517785513648,0.807864160175638,0.8430713631431259,
 0.9341803387345465,0.5201702619884933,0.2995467367020047,
 0.023326765105609493,0.4669608696564822,0.056529273486690235,
 0.5492512909339168,0.7719588751198242,0.26827009734724006,0.8010325289545983,
 0.16551684248760234,0.8175969463456764,0.6640085276528661,0.9220336676056384,
 0.7528229457570377,0.18415603287090676,0.02074477926141538,
 0.3763235223619812,0.5760248035848524,0.8604194405686167,0.05451536758234199,
 0.0889490116121312,0.4879095358149299,0.4776819438922424,0.3049287204202821,
 0.5774148690112031,0.8205674279031439,0.4479522429877787,0.6081458142249785,
 0.20201944084331414,0.04274387889183251,0.5269910461320187,
 0.6248576917686005,0.5119125865994524,0.34828535079124245,
 0.40893148571791216,0.9419552634068805,0.5484949515806512,0.1117422702286196,
 0.05797058274337619,0.8269905707497921,0.11128271841001913,
 0.6515079726269521,0.4650128902252859,0.019208062625939104,
 0.8116519730317038,0.2017797815180281,0.33005861272790105,
 0.07798934668579982,0.3262315471599905,0.45257953631225656,
 0.7296490792521317,0.36875320768831266,0.5085557812521987,0.6106427749623422,
 0.3584787102517949,0.37290523465089576,0.9223593188045114,0.6913576711610416,
 0.33222792918177635,0.7378217652045117,0.6479531509078887,0.9784959295663725,
 0.10102972889281159,0.8137405463872208,0.46491127305006363,
 0.7135158033509185,0.7517446595188687,0.7262634953187377,0.6832859395723231,
 0.3082757140346285,0.4990164269714362,0.15873891361409487,0.552243796371242,
 0.36196655249971654,0.03558007414241737,0.883346960507287,0.6601973915291786,
 0.9171431812361707,0.3817272755272729,0.7914337524719568,0.480112416102246,
 0.6026292896855323,0.7497708257149209,0.2072413000052088,0.2300257717965969,
 0.7135857053717668,0.8492663850587459,0.6811906175923961,0.8139625082501458,
 0.35736390062832357,0.5622785897731544,0.28308614422623835,
 0.19935946295212303,0.2563303489183043,0.9681368693198223,
 0.27160660298849937,0.5426573828717153,0.5970674805534747,
 0.15457265092542827,0.15334179024832095,0.06895622427424053,
 0.7423651414453709,0.5738468649404453,0.6163138405581272,0.9845456070079457,
 0.9430526769245989,0.2186745954493997,0.9808545121811009,0.9978003905906205,
 0.5307968450709101,0.6773952479008406,0.2762044826281983,0.37744150381886254,
 0.46013089068496116,0.1672293688456714,0.46723885864502823,
 0.5278436713698385,0.33538449749782395,0.3655029272891791,0.1624932629595227,
 0.857335790372609,0.10473486502578977,0.40848477903054536,0.8687793712587937,
 0.3488583211787405,0.24729050473317882,0.250298694879981,0.9997276271525058,
 0.5140770555550335,0.20364333331395623,0.18395209713199856,
 0.5818999971482401,0.14905067342239753,0.5204653493431514,
 0.11520981691946952,0.6081655463910668,0.6131931762679443,
 0.017419204237824837,0.7528376961061152,0.10410950298972899,
 0.37502306689967224,0.4729483867435951,0.8261495888971604,0.8180595862044489,
 0.3039867670883236,0.7660427182545602,0.6466607429368306,0.08184961819875936,
 0.6578958801195833,0.6588234268002362,0.9144650917421563,0.8086148827443254,
 0.6125275294157786,0.15525038253250745,0.6371642096417609,0.6874550217795697,
 0.11283400970731727,0.8708027721885035,0.664043654022963,0.886606694408014,
 0.46485375419149444,0.5740500199428764,0.6199272318644924,0.5595140249049079,
 0.8807189227836765,0.9696465110265576,0.37648963599228924,0.6466410554035629,
 0.2367763495861206,0.7339114931138119,0.6403931232819715,0.7953314879007287,
 0.8002720894437764,0.1733819551465876,0.8450620014536182,0.9619784039592852,
 0.08308212288105676,0.9851205222746082,0.596428355098226,0.47873436984861917,
 0.3910567155018597,0.2301600110182015,0.5621442436051663,0.9279828211717329,
 0.33911472346078375,0.0782928821425285,0.18936195542948786,
 0.5130070312704826,0.06865967905086856,0.9150855325606703,0.5322286528714771,
 0.32731110412476216,0.10946351161636036,0.8796967110554117,0.436921263575066,
 0.038118138644192556,0.06057314617310583,0.5832600293222405,
 0.5604301173194776,0.19362286188574507,0.3184189625782847,0.7242556240162696,
 0.8201753936680363,0.7400858919779865,0.02933248190152793,0.9489795583838346,
 0.9215299724613986,0.25336369412202864,0.1433504228359408,0.7728348594803316,
 0.27142435334980664,0.4025410389484685,0.7579782898299708,0.6713680042616882,
 0.3111375844542671,0.8656905145817566,0.846815941342506,0.01641709385168233,
 0.048117933337594065,0.8328686600871684,0.09592243412163903,
 0.30096458095754075,0.1159906587453895,0.12752974947302742,
 0.6993656194859832,0.7407055350093498,0.5565387505037546,0.6262128520846741,
 0.3146851897458458,0.8129816880627624,0.396324609478367,0.7966345208038509,
 0.377433173387386,0.3891684004783602,0.7178879440960614,0.3280122808930904,
 0.48977171453415225,0.9748816290619432,0.8773290531881288,0.5181992804332375,
 0.6026326872571146,0.9695712773218912,0.9776008973840249,0.06459799498947072,
 0.18085563632234458,0.8212870632446828,0.0886250237937316,
 0.017268864365691194,0.509798492500051,0.22599425589821898,
 0.5523182172600883,0.30264050285220767,0.4567588271593337,0.8925198987670836,
 0.20538930070556827,0.7053941827152156,0.5989502759233309,
 0.28182085407731206,0.18943220738728528,0.83763457327716,0.8900094340941664,
 0.7517357979715298,0.1594578414149388,0.48165026326543814,0.6171752613385167,
 0.4216638261232135,0.6638160304481374,0.3603126129969233,0.4872449318632913,
 0.09046851373777509,0.5268791825462602,0.20093333264517654,
 0.14525586313276495,0.5753477622346843,0.7046564227629466,0.4517862867119906,
 0.16339750683506948,0.16096657390524238,0.3534607763556523,
 0.9820800397402893,0.16122328667091856,0.5988773609318396,0.5005968522774208,
 0.3114433250147438,0.11238615266846097,0.24861207388074702,
 0.8379822811255496,0.1947323433808198,0.4153886923973963,0.5841409439255465,
 0.4288667571967525,0.7076126019073051,0.7019563858302863,0.2753230621231826,
 0.8441676046350086,0.40773044670197056,0.47091250887319236,
 0.19644663765457004,0.40428793396161766,0.16645453601890003,
 0.9888586317339138,0.3201349784810217,0.9772559171840854,0.2686588025069522,
 0.566626433709771,0.7800940883240322,0.8078616669210229,0.973658607468121,
 0.8340341738634882,0.9130986557285762,0.007641204666506196,0.588058930813979,
 0.6864803133554973,0.2882746419973393,0.6081292642205929,0.8400285909710292,
 0.09231404523412756,0.4629531201893722,0.01064276143014875,
 0.6329791803635314,0.2706663115790101,0.5163451617939235,
 0.013987298830373862,0.9740076074755394,0.03818743632526378,
 0.6256788753417146,0.012995647849123104,0.8867623596712833,
 0.2560222198843993,0.4808829438566028,0.7248986760192625,0.05243830453842613,
 0.6793485761385647,0.8635641909510465,0.29402197198261415,0.9196622809630547,
 0.5135944189340225,0.6579970723312362,0.4881449867040587,0.9661211178547457,
 0.5497996572639662,3.4191352222345817e-4,0.9879813320742976,
 0.5411845061488738,0.4199078489286441,0.1961856080967307,0.20844065617422547,
 0.15176101822482657,0.5143612431803816,0.20599300093839312,0.974687011610948,
 0.5098988236457653,0.6264132395438187,0.6450247351227458,0.9797029823437834,
 0.6553714897057464,0.43675086858164014,0.16519175330297742,
 0.45126604682662186,0.41667025610018216,0.2585038638613316,
 0.2387128987722117,0.9853797834426923,0.5681458016366661,0.5899938649205233,
 0.8007644240159922,0.9178498446772165,0.38915966501416077,
 0.08839691872549627,0.5042615411116911,0.3570195679306818,0.3149678536146935,
 0.3182016623403764,0.8855299844299773,0.9942631452142822,0.2776792582639702,
 0.02960648695247292,0.4552870048224067,0.1527753737529859,
 0.06878167179915784,0.9820243558357408,0.16890734074338076,
 0.7133272465744899,0.7409383794612475,0.4072138774940256,0.9026478182392775,
 0.3681426275150227,0.2521468832201922,0.6295981895278641,0.7025495820714974,
 0.9716602278711048,0.5920122466039381,0.324701386574197,0.11195542772183953,
 0.6199827918891503,0.5351510861653268,0.46709371109977704,0.4780917788017853,
 0.763701034226138,0.22528894453289383,0.8966496990133846,0.5745132502492941,
 0.522344444956291,0.5097233354569396,0.1620175869606597,0.8442627626297247,
 0.1742199131712111,0.02369564116745615,0.6310120570435047,
 0.13388022586775505,0.5680711697495814,0.4965568701295029,0.6655188580854478,
 0.13121332887346826,0.8102675095692379,0.3969714106109926,
 0.08001820120321013,0.6661184174101181,0.8385440288801143,0.7661242955950593,
 0.4970028347582409,0.3640771764837474,0.8745611235717756,0.10637513807601862,
 0.8986092333311344,0.972600385564104,0.3444037612570443,0.20856213377903832,
 0.44005960876144534,0.6009421314601747,0.7387920045901082,0.512637759060605,
 0.9909399684330324,0.8037329474266598,0.7874401490570636,0.8960073827293729,
 0.995478968782046,0.30020831703258066,0.8621279688807002,0.7898017443834346,
 0.93454906028755,0.2683692550885237,0.20830171990380086,0.7480084789888326,
 0.5040489547900268,0.2176707345230372,0.0704163722918767,0.9407164506600312,
 0.40553008067923857,0.37018944038519386,0.5234455132224736,
 0.3042376167154026,0.2585157829209269,0.4367334757731154,0.3624298901270868,
 0.43544266827993583,0.21332978805769798,0.3963276918371398,
 0.18613115170141115,0.10513239773602845,0.7074034777219346,
 0.9961093682863525,0.6026667384784795,0.8575549862906797,0.1528950158733039,
 0.9745850562501336,0.1683438140469723,0.6118576145143994,0.18820456048683176,
 0.6897063768453,0.31855719922094394,0.8697122183528562,0.03874028293633636,
 0.6447712743502512,0.7695061900416661,0.9564517486125421,0.6427227245976583,
 0.5388922537558759,0.7151086412059959,0.6045226175657057,0.7569667458410708,
 0.26719326544402244,0.4404485769462698,0.729644140978839,0.32012333277708593,
 0.781050652831644,0.2777558165967444,0.06470245018003118,0.7504593918430951,
 0.6667132684949432,0.5009950882393535,0.035546052614768286,
 0.8333982751396983,0.4765330758370727,0.302584237456056,0.9139641517992283,
 0.485447618366461,0.286170256443382,0.29091832882927415,0.4568110760156153,
 0.8805259470955962,0.5543674080028068,0.5843330305744153,0.9480940553993737,
 0.811535922787376,0.8673255727064453,0.8390900957498959,0.7425271817933526,
 0.927330046766381,0.12759541847038647,0.6328535564240028,0.4485388440149466,
 0.9674799711026335,0.1499250597880366,0.42944280368839516,0.5202687447660099,
 0.2303971981985874,0.3960298593680742,0.6144063029030296,0.7608746899704194,
 0.22407884267671552,0.02979974019484999,0.9168941644358507,
 0.40641317214953365,0.9765499020367677,0.04538437285386232,
 0.4397389243491522,0.8973286136125164,0.43840221733265994,0.1776348127693126,
 0.16214359464624328,0.8235561383674024,0.78254332642825,0.4671269627018675,
 0.16282004233225822,0.660750767177092,0.20003501083336883,0.6038158419972267,
 0.8264332008583388,0.8891495581716626,0.03734091475910328,0.6794361546127231,
 0.11596366713071482,0.620323908709423,0.48014106837499393,0.532913859822009,
 0.9486322981272661,0.20624513613044937,0.13062395125992243,0.716756784090842,
 0.7968554282272898,0.14579968759177098,0.6374095572974788,0.7433590584400629,
 0.9285316971404588,0.3919333892949648,0.6188134428405164,0.4195452470302127,
 0.007505936650296086,0.7315215626164697,0.8194923562948605,
 0.23512149454875197,0.3842886864336864,0.9738872636582322,
 0.23492411020695803,0.9310842335491618,0.2571961568605372,0.7306522150028131,
 0.5843984381939245,0.22596768739685746,0.05719421728622265,
 0.9141931527465796,0.4207218375790438,0.25735989748432386,
 0.38240366052879127,0.13081681157839475,0.15716840625825623,
 0.06369850455890835,0.6744601495501121,0.05151718299284325,
 0.47085352303018235,0.3624746901954683,0.5464036032685329,0.6703228034477815,
 0.014639780883212272,0.38472161725108067,0.6636323219186079,
 0.5289972956632125,0.9622031639488705,0.07766986457554836,0.2430772961248282,
 0.04528854513887837,0.12003690854041793,0.4332918758903871,
 0.5179721540204025,0.8846162495871673,0.9366851142694308,0.29029819763520026,
 0.3037098271783497,0.5649901028177489,0.3895114846300125,0.48169749589884625,
 0.33325298541930193,0.04540783209913535,0.25149251039082376,
 0.1210867909205563,0.7113017017770993,0.2850053353552957,0.7784642836375967,
 0.9204682950967733,0.24460225821912607,0.4347782575615926,0.7884431669031572,
 0.4600624822484629,0.12621716761843949,0.9449962324731286,
 0.03280154013645309,0.41411461460419907,0.5551338682979503,
 0.3266261487389729,0.31125698076696995,0.05292714963153977,
 0.27399404884514833,0.2817309517238993,0.8596642157552888,
 0.12204913488796021,0.9784682179804519,0.21160731241501385,
 0.09470706331795853,0.8481328231651311,0.20687044087776796,
 0.16214774936176146,0.14451349273389713,0.004554393635531118,
 0.22247495545592877,0.28077482438211443,0.8481964783773509,
 0.1733631471088628,0.5850801068693209,0.7137908249358172,0.61432639754363,
 0.7539619038675645,0.044767685968058024,0.48463275690776486,
 0.1944479558356188,0.8957212226683688,0.05157165145555398,0.7944035643669654,
 0.06397834239489075,0.9793598986482452,0.5410112560309325,0.2288755982534636,
 0.7411592225411123,0.8283267013452331,0.022677775788727872,
 0.9486776657107874,0.08148552189480995,0.8406109971336356,
 0.10976123001516236,0.6473878473443715,0.9798699169843899,0.3531061222933347,
 0.3245495281927546,0.7194162491645111,0.969523510822531,0.18074725234191835,
 0.8636687645042098,0.40649638097361107,0.6250775899164547,
 0.21485156603644362,0.43543020872236504,0.6395888853773823,
 0.4039458312220212,0.884057795604085,0.1397128161015575,0.05643135140475741,
 0.4032386888282725,0.5723425953696822,0.5925694377071612,0.04213063460021915,
 0.25939002824630675,0.3319712652473018,0.07347932198141649,
 0.5937882915280521,0.061100263502772756,0.09424692033454796,
 0.3064171270270024,0.6206073556234653,0.6992242705018565,0.49208241461393676,
 0.6441576487072992,0.21932948174300781,0.3505452488945847,0.8871040962156089,
 0.6300992824499876,0.8212572484856061,0.6760008641020964,0.48588092412909856,
 0.01552560672774883,0.6109913939747077,0.10702095321925675,
 0.19995271372306445,0.7489578768119162,0.7259961189739956,0.9520342837128516,
 0.010296042854483378,0.9704889610969697,0.911689216128769,0.640500299015675,
 0.6086768604034662,0.8534698431451994,0.3606032830803233,0.6859504216636709,
 0.11930401690051917,0.798215524873389,0.9374746485203694,
 0.032598673611474616,0.08520876272153388,0.2814440075519712,
 0.3817179010697097,0.7990425045995929,0.7754533202121187,0.9809081503941948,
 0.10567514631043462,0.6974238965162443,0.21851695832896156,0.680161567167823,
 0.271135245480958,0.9774738446271085,0.13326503448478955,0.7961930795021253,
 0.8716089669306,0.7195179936838599,0.7544155079945818,0.24617092635956717,
 0.043818243745750385,0.98873505261484,0.5979529944008208,
 0.014584575258028587,0.6914692817324463,0.430714440291466,0.5859079402400442,
 0.1622960169113168,0.005471584541929975,0.8024187547201751,
 0.21808590517763637,0.4468897018565775,0.5514278651504626,0.9900490772858369,
 0.6207938914083426,0.9537242518368086,0.2071543492644069,0.5143896098324725,
 0.5202683287030853,0.9029125238291885,0.0014580301939109974,
 0.6810465945527269,0.15123279979122906,0.2703955174480288,
 0.19908012949322396,0.6127876297512991,0.09747344002820535,
 0.6150098803793189,0.2003654514155654,0.8539501731528129,0.5525789182673702,
 0.06467788406862107,0.3052449297931499,0.08908359908532937,
 0.004557499989731539,0.7789119204249098,0.45004550744666805,
 0.21575927141747253,0.3930090510337272,0.6443935588871507,0.5002771687804255,
 0.8179956715584563,0.6966729623474505,0.489566875154798,0.9784267611750299,
 0.24258072846323664,0.5071994407704816,0.3896095042200691,0.3213937840420167,
 0.609631663233141,0.23280310366556667,0.9686358370928533,0.6554970033265083,
 0.4234966515424218,0.9299221199403825,0.3676761431585748,0.15565271303422745,
 0.9116486706489303,0.7166662944890819,0.2612727257697468,0.6069898548553309,
 0.03718390606743993,0.8081485320066795,0.6432037536808493,
 0.08692728206463374,0.40146504203036126,0.45609771523849874,
 0.004607903245376965,0.9088197303420749,0.2983808810350457,
 0.9076280472213207,0.16983091800481487,0.9066645310302328,0.9528257040264012,
 0.9712147895691092,0.9246011917000536,0.8557427988590043,0.9228768858520857,
 0.4364110612638037,0.9203842497187826,0.3251351481978164,0.5433204472197627,
 0.18857695220541226,0.5049379622542914,0.4149953746767745,
 0.10229015963186283,0.7341010794060311,0.6429309889395816,0.6060422084463325,
 0.28490457762930244,0.6059441246692804,0.25987786928661605,
 0.09280378181935867,0.8660205737698526,0.7254118139470616,
 0.056023565808688836,0.3359694245374326,0.6878081878470793,
 0.6566407809869099,0.5895314950417794,0.44212614180156473,0.7248687327995692,
 0.6351369238518394,0.5367457094319626,0.7713635181855991,0.17219456059609872,
 0.9519738945382571,0.6748170684193837,0.808889199267474,0.8264752885509871,
 0.5597438819655056,0.6452919218167859,0.1625137928048731,0.3979537908938061,
 0.15560014201380357,0.4155901346377201,0.4426715881899381,0.5250813714680183,
 0.15301414208318864,0.5750982314650135,0.0035324127571826347,
 0.4094278672162035,0.9518403975606846,0.28767203780290174,
 0.15773688525650864,0.40792913282615395,0.7135941541840335,
 0.09066762745388779,0.11274937552862196,0.749024638318424,0.6402895130464393,
 0.02165956034759786,0.41935813457026294,0.053317779356536255,
 0.7276718492073474,0.8566231372014607,0.017134600325624638,
 0.1249001590736224,0.6931290780885305,0.20709553315632845,
 0.47102424941144916,0.8011588290710668,0.3528246276825808,0.2724615386809863,
 0.46491835142149185,0.010718120863461134,0.5407580434910854,
 0.08684079957134494,0.6912055292880243,0.296036286909484,0.26565192501006374,
 0.46103572441334795,0.6361845662635509,0.024467163447010942,
 0.24529926964714677,0.47297207924055207,0.09150137328747543,
 0.7200278505385933,0.02072458650818587,0.11461902906006838,
 0.8759290616654064,0.7011331089274011,0.7756802057607958,0.27599397543415394,
 0.023543483878827764,0.42875830900711087,0.9266739716166869,
 0.45094321980222807,0.3050392554796345,0.34358255446415553,
 0.5980839990400924,0.7135684593773985,0.7848692908179504,0.9918771029584824,
 0.8385917483466381,0.22976425934971856,0.7129861890613165,0.1528964339188471,
 0.7747584145766198,0.47555466991957407,0.33654487833677427,
 0.9646886272190828,0.7137430468407662,0.06225520411469998,0.862316943097684,
 0.7802477908115395,0.13948905091589425,0.5603078880997187,
 0.07903108412497739,0.8789551621947201,0.3889007998141901,0.6553424108675167,
 0.265871624165931,0.3180920958210246,0.4295406510489619,0.91843198032242,
 0.3650041021721647,0.0999771145361229,0.6895193210634372,0.22567185029661052,
 0.1759051615382301,0.8392133965759001,0.5461204797883843,0.09375026876362624,
 0.010219478603381016,0.566412560971924,0.32167771377915544,
 0.1585862333659801,0.8729389394268013,0.10130101252281021,
 0.41379838774907785,0.3792748959388852,0.43418770381258176,
 0.4603785700864156,0.35621959565155736,0.8585964469468605,0.8394054947645505,
 0.7154940792544435,0.6119719972769576,0.5043999336381232,0.25761377298718935,
 0.2448887999859679,0.6407485691605113,0.548597596186685,0.06327753455438634,
 0.2607098057910421,0.6195906859182951,0.34953564285238947,
 0.09724406175731892,0.5207085750225605,0.9091568430538391,0.1541132043541248,
 0.9334463953425544,0.74734823948129,0.6176178431444201,0.30949288927746677,
 0.14805709675246037,0.8217417119997014,0.4318991666359181,0.3790879225547594,
 0.10969017573665252,0.657206653236147,0.955347864290042,0.5673020926052557,
 0.21252387292210095,0.029719405963264833,0.4423753302992126,
 0.04257048260679741,0.40087763745974625,0.05118652201565288,
 0.3492354445070226,0.4075874410229775,0.4881236252761767,0.10856085483736566,
 0.612849709308008,0.41937703316329134,0.5460235403400778,0.22249223910145777,
 0.813875829711554,0.657146470875505,0.06487504918897835,0.10736983947087297,
 0.8893713866500915,0.9398569723626391,0.10948058169178077,0.7411660538453028,
 0.17779781938598227,0.024054587775089398,0.3269029200426129,
 0.5019820377938089,0.3650545753828697,0.045178498819235635,
 0.9001992958809842,0.5436393854799563,0.5469150675200662,0.1336861215029348,
 0.5604982186990213,0.018440337215754776,0.29175411474178725,
 0.7989454085265794,0.42577955736699924,0.8466110130839187,0.8903347159401355,
 0.5745866133932112,0.5650776074604523,0.25281832267255133,0.9511254540709599,
 0.7568245095487485,0.9880772466604533,0.6329304562213416,0.2607813856394876,
 0.047452921643405976,0.09609330570560992,0.1164129352037151,
 0.16375383858082526,0.38022447931658976,0.4597796804571437,
 0.5563052887983551,0.04941002058261912,0.49467719936943877,
 0.5205451904401794,0.002624856874315573,0.6639619802474704,
 0.4516850792786731,0.8575284567091156,0.9504048259130504,0.21180337434194474,
 0.7922574155250236,0.9630741945311563,0.6307906474116767,0.5758823965287247,
 0.2973946472021347,0.5878676478067096,0.8287449845718287,0.612622736870863,
 0.27637519458914184,0.9767835772595447,0.5745659324797596,0.6610503416399793,
 0.997401588675336,0.06724176304953466,0.5124708435659785,0.42983422492840406,
 0.3730686472686211,0.311623861524718,0.7051972665245606,0.9239938819130139,
 0.286712033035845,0.5685298856138491,0.7995485300185426,0.3962027899199444,
 0.7643745774912215,0.9161212376290618,0.48008232988701416,0.464201205147326,
 0.3702823049297109,0.5183346428033172,0.3208945495311888,0.9528866018422085,
 0.6581518441837386,0.18965450431568165,0.5227894835359598,0.4438334693382505,
 0.2348501964567421,0.6536231043942565,0.4006477363301384,0.9502521682868681,
 0.3203925679615671,0.8045932041593241,0.6396243073664395,0.6408693284132018,
 0.9381878430260793,0.6542931074555789,0.2733280370394724,0.4754339372699501,
 0.9637893825587794,0.40339901600449735,0.6744213258740149,
 0.23609155864657538,0.4287045210215439,0.20005768726354167,
 0.11943715793721699,0.23609930347479136,0.9586130957708798,
 0.8825814264920673,0.9569717916426146,0.77731094135964,0.33011054480320257,
 0.10528561908819845,0.45646408174205755,0.6901351187572543,
 0.8619849274228334,0.25260606865586555,0.059427441186969965,
 0.8816919744786517,0.030081371314933048,0.6827723574852986,0.355462257729407,
 0.4206792300796106,0.7322314609945142,0.9035797143554632,
 0.014011119278419915,0.7537850744952092,0.90727436571888,0.298498963671591,
 0.6561653122521418,0.16204832520698087,0.6673274657741354,
 0.12311623113852621,0.9770407532395257,0.08768349784091045,0.547141849039104,
 0.729403798391524,0.18324829431646072,0.33172169801270723,
 0.35570534689835553,0.6503642717674387,0.2742895174423611,
 0.10129091888814346,0.9993323973161914,0.4704650338517412,0.7729854946375078,
 0.43061081849464544,0.8558710601877563,0.4955285919849831,0.8226060624975643,
 0.7099000284809069,0.13914029235899217,0.6138140854237591,0.6119584603460501,
 0.2954610721944212,0.5590308850760694,0.7348975339821032,0.18999813739287186,
 0.11892943803294487,0.39148265396633297,0.06893770657939424,
 0.08773522053953531,0.5502767925021815,0.28032957240083,0.6299850116119263,
 0.329638826549556,0.29097384392383496,0.9243928651848549,0.17854260205961148,
 0.9632858511758037,0.125110337186918,0.2704284062557163,0.026235271320010467,
 0.366131307038714,0.09332047171459601,0.7102287916287474,0.445782656479345,
 0.6568197633648154,0.7163243141040668,0.5972837764363278,0.6571342459870455,
 0.526922703305556,0.8780292815680429,0.5904334900025618,0.6058737390619857,
 0.14610459733882164,0.3744481635242053,0.4117819955562373,0.7320652901092978,
 0.7978578829262744,0.2890317587187199,0.6914719888290749,0.18435918281751673,
 0.5451396303384063,0.3359434916560867,0.558468870345849,0.18446437206281807,
 0.6580310062889672,0.6811439125955361,0.02182819042605466,0.6145850085906661,
 0.8856148071909467,0.024119244083280145,0.43515409252840276,
 0.30044503052427673,0.25682311449212214,0.7656081846348343,
 0.3515663166946347,0.558572806948465,0.056280249699150575,0.1081642826130893,
 0.32555680699644685,0.1839935311004871,0.050087350300215716,
 0.7657049989617398,0.22030839595722118,0.2941889813689871,
 0.07455394839164964,0.11929492207326098,0.7495166869383088,
 0.4915472792754869,0.7057502930796947,0.13949639078737064,
 0.33807433567438183,0.8553447641191826,0.6608598493237814,0.9370061867139132,
 0.1356531310129102,0.17338683098276753,0.18220104712775853,
 0.01799313990668394,0.16664602549738827,0.8920867811116897,
 0.24707759913105032,0.8058021954444148,0.6783206446791851,0.5153582925116255,
 0.029566722444452354,0.30223051954940183,0.6841117653978843,
 0.1731573998737601,0.3135859455489912,0.3747862623888648,0.47113741462434156,
 0.32721820048854544,0.04926099544718632,0.5879619497995795,
 0.29138520577558236,0.0017546279804558917,0.9977295655128983,
 0.7610497787839656,0.017147123908132444,0.3660223781516535,
 0.35496183759692657,0.1551101793047106,0.25039171446747943,
 0.9887550280662508,0.6466963263770749,0.33250535045545637,
 0.05557188477231745,0.6899664160444163,0.835131437242299,0.6654784802264139,
 0.43057269931294107,0.9064747528819371,0.3807422832655514,
 0.36087537641066414,0.24882196075567176,0.6189314986438043,
 0.6805181036437935,0.5590223850162248,0.22039433029540056,
 0.019592881370861237,0.24714000111119772,0.5454895561054368,
 0.35184763190184143,0.3457824305538095,0.8260963702029549,0.5147517332746245,
 0.018875300776603487,0.13804115115391025,0.04168690002750708,
 0.12193812084229416,0.21479018318885978,0.4312830570635362,
 0.006861881821511706,0.9430508779829837,0.280434991267402,
 0.33453829320981554,0.019178383749523742,0.417005564646309,
 0.5340352383209916,0.5038704473114319,0.6018590386487601,
 0.039108574892944725,7.714859318869172e-4,0.19153255993794993,
 0.3964201208142455,0.7366455217563973,0.5252171032241499,0.9025178611295955,
 0.5664121346620704,0.026572479300973617,0.8440125945810165,
 0.4790006955406283,0.1534123115775866,0.13641258205921247,0.9777365315889985,
 0.6418887053587279,0.8506042242440832,0.7245092342703876,0.8493337396845315,
 0.6501210323017289,0.2869264015762629,0.9844664399440411,0.6668802251880762,
 0.4709041524752433,0.2114954118490916,0.44818452689496946,
 0.26178060845532536,0.4304547194709415,0.5852818760825282,
 0.19069682385484443,0.8333167880661816,0.8281515594515448,
 0.43723040782434475,0.26673962249932837,0.7785136345784726,
 0.08805639401068555,0.41930434337304934,0.4974789477796191,
 0.022241535402412804,0.07290122741267524,0.2821902790146491,
 0.023953693044714952,0.44012174226502265,0.39270994415402405,
 0.5076260143179135,0.04684714355210795,0.2756111752128273,
 0.49932356797591637,0.2959473636880474,0.9002655017258415,0.5747525886971959,
 0.9914744485273088,0.5615323683854425,0.7433208108449821,0.5192742530841541,
 0.6876008329433918,0.5864061239675196,0.6054205082820417,0.15305491634276835,
 0.053898668824733265,0.7904320363213335,0.9310583595894947,
 0.8204121630103989,0.5106187160437831,0.10569268099950757,0.9542040817596601,
 0.4407241441863141,0.27354378548351366,0.956057438823539,0.8141878572144845,
 0.4947592673574175,0.28857426925561547,0.5684015087969183,0.5862707595201675,
 0.4616471815859825,0.2025341202893326,0.028398261923892387,0.856903992768697,
 0.958728480062642,0.29790907725015525,0.5201441432332841,0.1488606887481303,
 0.36397708383814575,0.6852891827610476,0.10548222733318591,
 0.7258056184840307,0.8407132305363745,0.7008325485400863,0.5484178493403585,
 0.7729599385324617,0.6977837361422188,0.27244791882103236,0.7291690671987442,
 0.9053340593061038,0.6607848982493655,0.7707873615381285,0.4066003100739678,
 0.5819754374440839,0.2290079176120312,0.4076924299030944,0.18798118822079046,
 0.40408807368206956,0.09510454260424805,0.8602159115056602,
 0.25044725141222013,0.1286057029006188,0.6799414691292958,0.6475078800908123,
 0.7171508521344974,0.4510904631243984,0.3889785411401997,0.10188592995616319,
 0.23152150744060096,0.5391337718957787,0.007558679182175787,
 0.30269263652412604,0.06634351636453462,0.6759770124233523,
 0.8820754404733746,0.6971021879607152,0.1784298187043759,0.19014870051485422,
 0.45542788220079666,0.5095790137614262,0.7530332313277837,0.9114328382670346,
 0.7014073862388455,0.8129020565389264,0.15705170893588827,
 0.09665730924673155,0.3196996202489645,0.055529192776531344,
 0.7935704694128288,0.07492331632660931,0.4221562620243938,0.6508723993547049,
 0.8826642453013724,0.6916811741407762,0.26105815583107483,0.8626278875088801,
 0.21316428347519945,0.7847707517866438,0.7418696612202871,
 0.45713894460020243,0.7085334407478909,0.7427842397401145,
 0.06464605086173114,0.9707448649455847,0.8086170205688945,0.6684984913982118,
 0.6407150118038288,0.46795953658668366,0.04864754990919806,
 0.8924026161124878,0.8245032490577392,0.9209446679946686,0.1805865039323512,
 0.6850117876477755,0.1484676432076888,0.5118774181554071,0.4144964541764501,
 0.5134053220767363,0.6347139206791335,0.7302931659132463,0.45023362683637846,
 0.9729418400585894,0.6950931140097136,0.31289236014489774,0.8027806073385325,
 0.04591687720746507,0.9567477404716889,0.40119199734698285,
 0.9995898532222762,0.9853257654426752,0.3283456270013013,0.7093058765164115,
 0.2643414406299043,0.7345171817562641,0.43941493459353,0.36874463115370315,
 0.409932402218656,0.9690677965111268,0.6178478186926742,0.4353484535732256,
 0.759227119602473,0.6156676736850978,0.7360911043005814,0.11935843036782601,
 0.9417489409931468,0.25910572840484813,0.19559531886201664,
 0.5933854382068091,0.04940803288625806,0.36574819439145956,
 0.9759654790169535,0.37427812461060994,0.9251974651147264,
 0.17892859554162388,0.14329331105808107,0.8829827693287136,
 0.38080469005769957,0.24347247061419208,0.004687593050204164,
 0.9221837946808102,0.5335848819528988,0.9273953117263477,0.8144188361263398,
 0.5577067423209794,0.8145139396293625,0.12107440692244309,0.1720999241673402,
 0.42128787197153394,0.6285697533029267,0.1402822855503717,0.5347054326341736,
 0.2967863667912667,0.5172280519059244,0.5651911166140401,0.37869777850714137,
 0.08383382165911701,0.7920926822795384,0.2839991190070602,0.420144336722887,
 0.7140462343517203,0.021021188882359376,0.6172672473679028,
 0.9814517757460373,0.5387748052486208,0.15203441336008094,0.7175975531337359,
 0.220280494719565,0.4151988507472385,0.20265206122766743,0.13308784352052683,
 0.01674475512052276,0.5191637549819207,0.2693671400499239,0.7348468128345628,
 0.24354034186164986,0.43527432426029244,0.9492132176039743,
 0.8226996143818379,0.6539202506847656,0.3722494354230802,0.6837271906657431,
 0.6046184272426661,0.060794601911222546,0.05963143230569057,
 0.945509289115227,0.8391318884426003,0.7610357464127658,0.7579577197154379,
 0.8985634086714043,0.27235736679902633,0.2598814054467281,0.5626740084239069,
 0.26166601937543565,0.7334394666504542,0.37650338607224165,
 0.16000610728760944,0.47774192562545825,0.06027654054909504,
 0.5239111683631463,0.42939802619072986,0.7576201459880604,
 0.07921415193447423,0.03300637259073391,0.5514743534506094,
 0.21675415866232595,0.4736368299736795,0.5440466805735987,0.4225740917212101,
 0.38369927180034047,0.9484348970588641,0.4979998779047857,
 0.03749499931604494,0.061121432036319034,0.884564341756919,
 0.8176256338075012,0.31397373028796227,0.07592097699778211,
 0.6772845707820271,0.06106786353013749,0.22467470963865444,0.916741614074658,
 0.7405038954321999,0.3624898281962283,0.4377016186708862,0.9839695735539116,
 0.40896966436010274,0.2799100338795151,0.6432838708680221,0.7469740338519343,
 0.35594082607888267,0.10249725310776825,0.7130420253521258,
 0.20395702794507553,0.15685033729400588,0.9395967971533384,
 0.20776989582505567,0.6140366870805393,0.8672229008993902,
 0.24209748153538213,0.05171976462269201,0.9691993492934297,
 0.9759686895194074,0.9872214990180395,0.07122983997103782,
 0.42084569867155386,0.7452863613532887,0.3071785290249398,0.7639675068266208,
 0.8871685416281374,0.5233092630727005,0.7238440650602229,0.0447228033217173,
 0.23331153835031904,0.6019571088205331,0.46409977282173487,
 0.9820161150730471,0.2572093327599845,0.43087786667782535,0.5028474475400272,
 0.8377761881073813,0.7113244023891773,0.881616972239848,0.7730830765308941,
 0.10821733519440313,0.2147969121097162,0.35884601078487777,
 0.22417383244559608,0.16481120432820273,0.8295172261775767,
 0.5405154297905531,0.5003870362071933,0.3634278046430971,0.5002740211181365,
 0.5304117204540225,0.6473851068531256,0.3389147453444913,0.9605341806584822,
 0.3998870702927413,0.18831166509474484,0.5359668151936249,
 0.46437501168440454,0.05709638423629748,0.6180104953431189,0.761886201344021,
 0.7780138271465388,0.7067089042126722,0.4277352546131218,0.6387099756141535,
 0.6395266668090352,0.7345482710631316,0.6177663862519615,0.1065869187701507,
 0.5483029416086529,0.9483428391916398,0.8788874954451764,0.28022507528980123,
 0.791971228134447,0.6292841956034673,0.06225642695457023,0.9442948139184918,
 0.1654188183237042,0.489382449167695,0.8025634450096745,0.22240560226479045,
 0.573050916103481,0.17100807374379878,0.049751366349835124,0.350611957673001,
 0.29041332164082634,0.4671530759833872,0.2764714257816937,
 0.09349397274179871,0.13076323441162985,0.15889713237308933,
 0.8796820796872581,0.9378426371440657,0.7392393579389238,
 0.030459980066246617,0.8382011673582653,0.23596170321196563,
 0.17844914853647376,0.40739807904646863,0.07365880519657031,
 0.8842191255183326,0.6945895799162201,0.45321499980335167,0.8548271093049976,
 0.29659611591761836,0.23231962762134795,0.6100866368973386,
 0.6551257166507407,0.8727402514513679,0.0015875630853733336,
 0.11319775740155413,0.32518686530965146,0.6069103459018806,
 0.5573282380022885,0.6599052681724595,0.5162405232275312,0.9905495597225354,
 0.9224043697705111,0.37130711702166286,0.6309789000253844,0.5282970819575646,
 0.5569850049431987,0.3222570897391296,0.25453650626854274,
 0.21671491665625653,0.26505008901539073,0.3259715485599409,
 0.33538613214955193,0.6397822026624008,0.5615681695742778,0.4496980834390829,
 0.169219082972366,0.3990238807115374,0.13709110375996825,0.2684986351285641,
 0.23541927235803883,0.3130855055890974,0.891036925230208,0.6317751273738575,
 0.6432197165161067,0.07498125768987507,0.2555907412714662,
 0.17291833931255685,0.6410243709358434,0.4428274268755731,
 0.32634683321241875,0.9845543347132127,0.16016565387684134,
 0.49220558234593803,0.6663697869464226,0.683818364434006,0.8437029320866125,
 0.9895213374005933,0.7879422025309485,0.5631408999724312,0.09662214049447382,
 0.40733658669300377,0.26596768200310494,0.25351580482645897,
 0.8965318107392298,0.3530919805263355,0.4499364027056223,0.679786295045586,
 0.27160603145396645,0.2682465894118893,0.21150789114611646,
 0.6321074410355172,0.3140064079867202,0.7485537582655534,0.15653576410465397,
 0.7760638218760181,0.5982274395822864,0.12433522408133102,0.9209352789178261,
 0.9084800642209979,0.08061786184469633,0.9931009599714643,
 0.08090898345046593,0.5682957321327783,0.5023129471266997,
 0.46337538344829543,0.7495040646593673,0.5994736791129364,0.9507885644862435,
 0.11834228284998594,0.6674702546443042,0.24252455747550627,
 0.8460061016500302,0.5337177220559228,0.6585912580966865,0.8727358547571724,
 0.19056048850131035,0.9626205733703314,0.7960615037125068,0.8002945460862593,
 0.5835683520753552,0.3730982125153628,0.284547268587796,0.7854259682567917,
 0.9425062815813142,0.47574141309935203,0.6985738600129326,
 0.11675433941168545,0.2170358903342957,0.70642575837162,0.7813285329048042,
 0.7925033604880749,0.7791988788843077,0.913717491996789,0.25285466487227515,
 0.188507300336902,0.2541067590813859,0.36839792539746974,0.29606634615853644,
 8.615730644445652e-5,0.7873414769314575,0.9172349891431025,
 0.9675634157131524,0.47938678450863736,0.4794986493343426,
 0.07284590673081592,0.36893654614827953,0.8317480403783353,
 0.4907509459600973,0.1582134641401689,0.15495993933092878,0.7810111442262779,
 0.3645757409677117,0.36103297346960417,0.35932097367900095,
 0.15962604167750027,0.9175112306891386,0.910612071728777,0.35008888171361763,
 0.9867541781148801,0.5448869184995228,0.4148955964566329,0.49528795797058656,
 0.18399645281558474,0.7200662408627628,0.6417489388523732,0.8145594732860943,
 0.529026166952825,0.5459504870811761,0.3056271614607966,0.03559002141651302,
 0.36807174968343903,0.20674342107135224,0.8927099999664417,
 0.3751265571629764,0.5520463962144819,0.9062761565495902,0.48519624856036914,
 0.6041823879798591,0.5200101514441742,0.5063365918339187,0.9664400510728112,
 0.9600728422885949,0.8473556711387413,0.34093126691915165,0.9503441646062503,
 0.9823883009358513,0.8662955481233304,0.7410470131873782,0.20029789956949018,
 0.15580763457905877,0.030354563241793753,0.8976386657769925,
 0.05813487195975775,0.7875734533795787,0.7197756366153288,
 0.25620637787008504,0.8578842417815414,0.44008355190656523,
 0.9177113725920449,0.9460556742242595,0.7318337625502489,0.1807990617377443,
 0.8716511351170574,0.536242998140398,0.16459880855300313,0.1236905048836131,
 0.023079318213726663,0.6195987548352515,0.8563731500377725,
 0.07833211087575753,0.9170358324085354,0.6355447709450609,0.3666663924860719,
 0.9433358347567279,0.12695518157255126,0.13280608406287453,
 0.08160472129785834,0.1163190499163328,0.3362887400528476,
 0.11933528542857852,0.6608408089328279,0.37008225268250305,
 0.3931701903750858,0.7304024591317829,0.4349248097751123,
 0.036784918140302714,0.7593898685843444,0.1991235798806077,
 0.5168171891875043,0.24238821158355728,0.646752955764955,0.6986254013012153,
 0.9548780999356267,0.08568893788669352,0.8971160217970826,0.3029577650423685,
 0.49349226350859865,0.6757185401026973,0.22977609864837234,0.385653190917601,
 0.24671163648866978,0.4476535642220746,0.4288309355896799,
 0.32524885649222535,0.9339759770667763,0.3912433100391801,0.5260275940802429,
 0.11961339460480647,0.36808599681088794,0.41847675514358995,
 0.49691679615333173,0.5489355313694905,0.3745049746446678,0.7926445156163724,
 0.7581685844362848,0.9043326719603962,0.493410533861233,0.22291328437932612,
 0.7648425945901753,0.698234144912442,0.9038627292282375,0.03116581517396455,
 0.07837170115110381,0.18661680101816736,0.43264361167651144,
 0.42981134397684573,0.5817910684566459,0.5071367698141521,0.4012782895620043,
 0.6289858781011672,0.9618145840200534,0.7607348699421839,0.827892842988233,
 0.966094033331542,0.6791249978240866,0.022368221700140856,0.7472729598202377,
 0.28564031537761003,0.6532849110805844,0.7036057676918963,0.8829967156809043,
 0.5829504236637334,0.7093722906000913,0.6798854012614161,0.0704010519934315,
 0.5695714200240061,0.4296800765044714,0.7881420499558047,0.22325944691810518,
 0.951147085525317,0.7502310568607661,0.7370958374076493,0.8371057889895914,
 0.41819550460636057,0.029475761860280225,0.5677583285258527,
 0.43065114549938577,0.5825026216890583,0.6156145479551127,0.829117433345432,
 0.17223721007645953,0.6212430506902052,0.8427945716846332,
 0.45750770750841696,0.4531662782368149,0.9420473399244406,0.9999214969763417,
 0.9794071073022106,0.6239993332821009,0.4804325953298546,0.6883913836362796,
 0.5654973373983869,0.7859140666946389,0.3312611972830215,0.7368510761237814,
 0.7925156841411755,0.8773250892358152,0.9187067901103442,0.46855548238095435,
 0.1533218790108839,0.32497516770767765,0.6906761401794397,0.53956304575312,
 0.7676612865307177,0.6715304375083774,0.3066958753559307,0.661231203714564,
 0.2628198688615022,0.8521313053685639,0.6721348911080163,0.23511464304859386,
 0.0903454103250465,0.011513082276487174,0.9865320807428563,
 0.1739213879257503,0.8732114940456398,0.7430855279164783,0.2225235642060992,
 0.8125647859176658,0.7601011560288448,0.14546940902959649,0.9870989816023795,
 0.8592087084174239,0.10819796088314304,0.8517948104672587,0.3563926827953021,
 0.6616197775625678,0.3626968706392091,0.14225835329253267,
 0.020320011395059012,0.42771407852357246,0.7183491772367178,
 0.7303280323001882,0.3722760641164379,0.07288454011356893,
 0.47856196133208984,0.3282300821720492,0.05755987702194798,
 0.8907143089889952,0.9006632872694927,0.18173445986749115,0.1262596177396419,
 0.3595023881649614,0.9373846027379481,0.9043343645864835,0.33316022935021405,
 0.7260776610100687,0.7094824783549881,0.4333515135226077,0.9464093242969812,
 0.7233925706070452,0.9004036615352701,0.1189556356063386,0.3855309686196785,
 0.29026014208630313,0.9668307205547689,0.7865102876758637,0.678983041059852,
 0.3398517155261317,0.9193197315992641,0.7350684006592207,
 0.053388025186353216,0.37982774294205623,0.738800178223802,
 0.5914174055622767,0.6229427973180199,0.5487655183546417,0.567060169105212,
 0.6758327524808404,0.8653144787914362,0.263795821912406,0.8621577606899496,
 0.9068986039063588,0.7503691807138804,0.3790778659783647,0.6394880737855775,
 0.873444164197883,0.5716087236886397,0.7394956299630379,0.03063767667168371,
 0.5381205350015084,0.8181132795473905,0.305647391968875,0.8648537462318784,
 0.06853217035688408,0.5102461620441436,0.2070812853116153,0.3528197107511063,
 0.3056349766435338,0.27848730546239386,0.09726256364211794,
 0.18002970856036404,0.7919931000932874,0.9083022555951343,0.732949157922253,
 0.33379792544037357,0.8281607427964021,0.07411442135843316,
 0.5818595423411241,0.7660289088820699,0.7071811124515729,0.8266266596744084,
 0.6351979685069794,0.5944283568012911,0.8111460807337588,0.16987940472107943,
 0.5008397482522652,0.05502901686531958,0.1260196138268026,
 0.14466872211142145,0.9778432305744033,0.8063401007568423,0.4447304666200931,
 0.9003542133488587,0.8095584173606063,0.9893913487308986,0.8934960173500684,
 0.9868543141544108,0.14614351400793024,0.06348180986784291,
 0.21331738676076273,0.4508386157962536,0.5378845349343709,0.8829393693033509,
 5.901115417475955e-4,0.21934667006809205,0.30367841402477425,
 0.605787371352745,0.9781173495926136,0.47125924487991655,0.8266797346940102,
 0.885857510906062,0.661651571638718,0.7094801680886762,0.8098251133049414,
 0.17812721376468177,0.7644927703108428,0.21315968355513892,
 0.23318908473595668,0.2745931476026362,0.03386230676633584,
 0.3136124578428957,0.9129250121746268,0.2684798029464851,0.09944073591002478,
 0.5832807414178008,0.593864416673197,0.6047129163444025,0.42437234454638995,
 0.7442507832975945,0.656577351639745,0.4590798381811929,0.861568963864394,
 0.7020725463881088,0.6656861870455573,0.6513069209642091,0.10009657970762686,
 0.8347976233492624,0.21430235920599228,0.34261402691088894,
 0.6766036061472098,0.8881549068004948,0.16633225344341462,0.8880404271936014,
 0.9933892454341962,0.9644656392313882,0.40988815061351425,0.5191161692988656,
 0.814763164949924,0.3260664976549511,0.2754974568328088,0.41150340434716937,
 0.2981254227410082,0.1771513764724194,0.11512597746841768,
 0.11436630785394608,0.9165442305058772,0.8226608136386653,0.97125925109993,
 0.03557051818938828,0.3243040530302297,0.994640502726186,0.1008310224226785,
 0.006437154837641623,0.8537019993508295,0.43262187577324107,
 0.5936234560137811,0.31724094294167227,0.19030697137375407,
 0.4057938617116891,0.9264839396773149,0.024392553209124146,
 0.8168394920818435,0.10626437754865958,0.2539539800678112,0.6824640711582057,
 0.3092953546616476,0.3743429633133216,0.7451921757680685,0.7100530289670064,
 0.5066267056594346,0.6506600983795279,0.6672061660888393,0.07423850075856953,
 0.71815776478156,0.6598840265773065,0.08455300650883857,0.4951008747161161,
 0.2942631665771196,0.6455171685064771,0.5211604569151783,0.9214025504071677,
 0.22766808245555237,0.3185721499155567,0.7991164135545034,0.0848196284952929,
 0.2943552449855068,0.7857917602652177,0.7565565385775035,
 0.014100459635812412,0.5301782243798143,0.7788615426126697,
 0.6304964082750113,0.5540209007809489,0.628240324491091,0.24177324790115928,
 0.3306790228239411,0.6794069351356875,0.1997086838426292,0.38659546580552107,
 0.6629949257507801,0.138586671553259,0.225190299511984,0.9096179183825288,
 0.4395797890809935,0.14304220804030965,0.7283103002577181,0.791632204711924,
 0.5610830619427798,0.525936762626511,0.7957915026986344,0.6807074096982707,
 0.5275247535772315,0.1697390425396127,0.4932483211292634,0.02391485859015441,
 0.23235878993271863,0.192291811107011,0.9224535559594649,0.31866009786227534,
 0.18997796339893125,0.5089621283943427,0.38724135761461287,
 0.3551624607349587,0.2996728783983863,0.5089136133332981,0.2721895590036003,
 0.2873676470747122,0.8328560496631626,0.14643374416853316,0.7685186854335202,
 0.8126501568579801,0.9579414226816648,0.5199405411826649,0.5961007600249904,
 0.4802403229384392,0.8173189055173206,0.49184080271002917,
 0.09448573714265818,0.3446992061187525,0.763388868194035,0.29634245617599075,
 0.7744502526738285,0.778807279791846,0.0751105569800875,0.3475916308014253,
 0.3088236545061046,0.5587600291747156,0.4260236057319349,0.2630057317934513,
 0.755941776145182,0.1268965836303062,0.7856345845014594,0.739852056536518,
 0.7432937913094074,0.3041965962012232,0.09474190787616932,
 0.055638718369448226,0.5205153560303692,0.9087185157610467,
 0.18450725515854705,0.2292222235674518,0.6347411407456935,
 0.35686054157567826,0.9026418722203075,0.2897048977744987,0.796157306825411,
 0.471792674064648,0.5959007328205491,0.3528692274349263,0.606730416754967,
 0.7529014196018673,0.30113345800381586,0.8954207218898412,0.2626904262546145,
 0.37272419932123946,0.22328427091010128,0.17781201944401137,
 0.28288418645030466,0.3618376824945865,0.6405275413830274,0.8905833679557954,
 0.19356410515477995,0.2420742742365991,0.3579350149399474,
 0.16774359972281128,0.42062373169816614,0.56175393586715,0.7687997886728789,
 0.37533038521155326,0.047848865950467,0.30238649759354674,0.6519751318856595,
 0.9638739150342435,0.8594581908204626,0.23671924900075714,
 0.06669814654048645,0.7006719627060629,0.08015708349126927,
 0.4120756898506156,0.9810701443488028,0.36085003872038623,0.6087105300148563,
 0.15385924838030096,0.7577081446619002,0.46823815810097225,0.619438998687465,
 0.7367179666349726,0.6086299217831979,0.9475365580204316,0.6301224598805559,
 0.6737277257074002,0.9240559905667161,0.19525776013127694,0.4164662610941575,
 0.2738341177283854,0.1756632704619745,0.9688849958632457,0.8565195236500202,
 0.8318123881979971,0.00827223230678864,0.30167002059288983,
 0.9770457658024202,0.10852017623023114,0.04726615777256349,
 0.4519392208463072,0.3227432833395877,0.45646421749785004,
 0.27165273104889986,0.8125896398315092,0.20842893315233324,
 0.5179177621097663,0.6762571245329874,0.6195702572535766,0.7530571887061838,
 0.3753156539681992,0.2670669771633887,0.07691654151055383,0.2148793690327092,
 0.9612128676320426,0.2737698721126951,0.47946961436333857,0.7543868217718451,
 0.3341464086125384,0.4983246319204926,0.6421909098864921,0.9204459459072121,
 0.5120883585354998,0.12849933783336775,0.3982776679504415,0.870885953191927,
 0.4251581156685996,0.4383420160436804,0.07307310847507786,0.6177828849522049,
 0.37757832907382227,0.6082357441869886,0.7413287967961021,
 0.007327294396088591,0.36356899469052784,0.42976647474235974,
 0.43201128733493754,0.44217563665814763,0.34880424788251374,
 0.978639774567153,0.7289176287041298,0.20963958192490462,0.484657609688686,
 0.5134441597002968,0.2740625707136708,0.3426764296344784,0.9361831184218646,
 0.46187460114172263,0.570725905215244,0.787864323044285,0.6043521593907236,
 0.6784401956684573,0.30174553884156385,0.2627107665270373,0.8863402291523012,
 0.4818459127540309,0.965405194663956,0.679288118006564,0.05937682542253564,
 0.883910539714794,0.1146776027300308,0.8177709282055139,0.6619891601955393,
 0.26322211589033884,0.09656802672175524,0.5575207077060625,
 0.9868044635755735,0.760428031739047,0.8012071251226792,0.8756320670981526,
 0.9846116576477528,0.4194807251899382,0.42774681670447023,
 0.48721351056113316,0.027525279843605976,0.6872241953506981,
 0.6549167884470458,0.23288847006899094,0.28875600228751175,
 0.8917597559447347,0.47483251946825267,0.21606223767820476,
 0.5841451550038433,0.5524613407964305,0.5276718566193215,0.7363337404245219,
 0.27421142819211897,0.20268289955570473,0.871405684264763,0.5201991261224794,
 0.9056233772828909,0.9963107329372531,0.9870880625706685,0.381211417162339,
 0.27362965389136373,0.76978422470633,0.7612450814976158,0.9176590994277561,
 0.39682815972030316,0.13587732632505944,0.1157252812855698,
 0.14473320779423493,0.16688943269110357,0.6472500630909449,
 0.27212366191981774,0.22131317161535302,0.2255382964526409,
 0.21627379661797708,0.1103401634309189,0.0711414692580069,0.1492937480857207,
 0.21566876764351228,0.9452022021570403,0.9498369937255559,0.7915485105242842,
 0.8479279375085625,0.12528563676354731,0.33256823168432703,
 0.008336278274492348,0.9034217532050423,0.5589764858961658,
 0.6893128449618867,0.5861247514512133,0.805109783871823,0.5837576911304454,
 0.632093626425565,0.2685555411020456,0.62439393342972,0.7431193810704284,
 0.894048180270634,0.7649237523190264,0.11129718423697987,0.15523315922118897,
 0.9047066380473062,0.8617470758420778,0.32888096198408057,
 0.27455929984214644,0.5925591016380173,0.29220032415549735,
 0.9950894191892905,0.6136700543879942,0.4720402217831723,0.314195076422522,
 0.5931566078919253,0.031322776946963615,0.9951162059628152,
 0.8359191928771438,0.566376888942485,0.702030169460799,0.7835873397988812,
 0.7395989735926372,0.8395210952761566,0.6224245130644772,0.17668113619655657,
 0.7029921697363,0.7674487457922776,0.5070067131925698,0.475019609943618,
 0.10413717924308996,0.768853081923548,0.45390135335715265,0.6264771643574543,
 0.327367551474653,0.8477559966060992,0.9621329253717922,0.2330880162858141,
 0.9886762878556113,0.006171398705265729,0.3215225589481028,
 0.0871416950372067,0.4450456691872403,0.35429740453178316,
 0.28212173082487135,0.3514918254983862,0.6599334245442681,0.3682236247362637,
 0.030011500023888416,0.7112076516511875,0.017582504726209258,
 0.7025630677870249,0.4621287942256713,0.6563870999222752,0.921110023051954,
 0.5328102883927274,0.22505181355654036,0.5778866905777272,0.2726183010932389,
 0.3809763292954038,0.6896329334558097,0.8497232559441308,0.4348949857498152,
 0.9505284249615346,0.5332522453389305,0.18235611127558826,
 0.49489036471190895,0.711402396925794,0.19394092339766467,0.5862875276732082,
 0.38219013903465815,0.7561681056918756,0.6816869230131973,0.5979187631974974,
 0.8777154419538192,0.8865809304302181,0.19071495857674325,0.6060125410833859,
 0.5102345956152783,0.9947341570168475,0.05341775010306127,0.9647727231105976,
 0.8950905759320859,0.5151890880384873,0.6919558934227716,0.29400986230700354,
 0.0250690565930205,0.7244023328832431,0.3460095850979087,0.9903505312919875,
 0.43016527866558685,0.7070652315109216,0.5375293199450942,0.5086626330395363,
 0.7109618035311202,0.21899063418373277,0.23260042894606991,
 0.5785165534477482,0.7329928981496974,0.5200061638307726,0.30498338050301854,
 0.015937437376866037,0.8142797652325843,0.6438734180463588,
 0.4451229113105867,0.8806387162045912,0.35289785660663453,
 0.37328854890686713,0.5964598426412289,0.394356862236971,0.9436514539903343,
 0.14511820586676116,0.25774413490122394,0.17103321508763836,
 0.6865251528068286,0.5330633069309356,0.8689154478825691,
 0.013229117004066593,0.6089247052236092,0.45753679843052275,
 0.615382778738706,0.6247456990392651,0.1182458224495806,0.4297687430266631,
 0.608506403217934,0.8105665370035405,0.11012398784803856,0.4752496342508794,
 0.5836598069347558,0.9015141527320875,0.16455617016083868,
 0.22238981386378398,0.5875479521422231,0.03575934861139818,
 0.6310326127796959,0.15286857612736282,0.7853754872714951,
 0.07873143419970274,0.5532057813978197,0.6507751217119216,0.5470569013460042,
 0.6346556575914317,0.9529866644878218,0.07647418307586296,0.0415394044085563,
 0.1903720044438214,0.6207206635821034,0.2611646911249219,0.9741259782565849,
 0.8111646011352558,0.99100552792943,0.872736351702372,0.07982134735933921,
 0.028472853286304,0.8333754008655221,0.5591565127527782,0.8912321315643896,
 0.6968932967221104,0.7793655193335708,0.8077365246029813,0.9698206364284063,
 0.6278941022825748,0.26503276339098014,0.6604078831638,0.011574402715082699,
 0.2159560986101694,0.44121956194433976,0.1426493472409196,
 0.29099389840185463,0.10436902047673735,0.6008720599170105,
 0.8899522227245197,0.5573255478729885,0.22646345841906568,
 0.38184157062226276,0.287581639522567,0.22183204638763054,0.5525268649064876,
 0.9601508747819132,0.67432353486198,0.47753229523974483,0.9146305234924913,
 0.8864988270984964,0.10171987977172647,0.7131484674981414,0.0920990818917371,
 0.7636116183730128,0.8458820695182964,0.3122282364801897,0.5998998222521613,
 0.5516785681643838,0.6864920195648329,0.07120645636305467,0.9549891380777888,
 0.5872006964850175,0.9392924765489546,0.260081349964948,0.5491343659966504,
 0.426981056851096,0.07780722399797257,0.2982778697333758,
 0.038195202165646536,0.9127238581074237,0.8719302850387408,
 0.3192364543203303,0.2743294083967265,0.7109459441652062,0.5332797970066357,
 0.7476183321171203,0.10012422195940984,0.09748701330178622,
 0.6122707156754864,0.45074578532855714,0.2911043126951335,0.6329964904933196,
 0.404531520916099,0.7070519154977365,0.026906101724214482,0.5864451538702713,
 0.8629573741252048,0.6264625752382939,0.8125000689885264,0.6997838942462267,
 0.38062227878356436,0.6545854403219256,0.9971862320637306,0.5804293513501771,
 0.3061564466536769,0.40501449634800213,0.9053334029441962,0.8867341579866802,
 0.018433083722435573,0.9609685070230675,0.30773610496929416,
 0.7419626816128687,0.3031517140914508,0.6231833936432358,0.1827361561502524,
 0.4145180751466844,0.7485115850545678,0.7352186757240723,0.7287193728346094,
 0.6386226710453604,0.18908990999432285,0.6431567429585549,0.751181011740385,
 0.8290320606932838,0.303677134134245,0.5848952301097791,0.5674395494361182,
 0.30886878574146037,0.9004672762568794,0.37358640384930764,
 0.036645174412915815,0.895593847595754,0.22313338772388758,
 0.24872639650522155,0.12149888227933736,0.046906373143170255,
 0.7945256283335191,0.5733981116265263,0.6474906103864144,0.5868682314857344,
 0.9877798579741119,0.04208338362945119,0.40736978653986256,
 0.9541619542662766,0.5629029425469262,0.2634306585987771,0.5387517780853982,
 0.34916075711435035,0.9062638599411198,0.10915535455421432,
 0.017967441376879112,0.3756316727401604,0.7215727605050081,
 0.5848192644860525,0.7687137519131582,0.2796787812454907,0.34796512248926426,
 0.5456329270311713,0.23336328698512732,0.30709309540016827,
 0.4454322569440985,0.458461089576931,0.3711001164208729,0.9056145901087858,
 0.6631310117850959,0.027662596570418474,0.7294520244999487,
 0.7693008041178963,0.4679702906265998,0.9716562150415159,0.03590657184160184,
 0.9640572610535969,0.30325571772851456,0.6410905846113953,0.1348619245635636,
 0.5093812196392241,0.30225806729545346,0.7760931073386758,0.7174514603694354,
 0.5596334566436578,0.8970755188579767,0.21361131836910596,
 0.22830835506310854,0.559909950846964,0.23135778361342552,0.631762863380187,
 0.43292195702627456,0.41843946536094756,0.47437575747378435,
 0.911818413951378,0.5896450179053239,0.17177766862421273,0.5693812322516283,
 0.1820916441654551,0.5131075555792106,0.26469827124534806,0.568196039508828,
 0.3238023945700441,0.26329035851732085,0.3144486419989766,0.2447342873051107,
 0.1620252803328478,0.1428838189429068,0.7952145246715854,0.28953936173671313,
 0.607744428214281,0.05511270624050124,0.5177509898457645,
 0.004612103028363124,0.8358235274996099,0.18414166062893655,
 0.4271758992977145,0.16052752868484443,0.8982221203799717,
 0.03936159316635768,0.5081913413304391,0.2588572998648587,0.7460446075003203,
 0.8159173287108225,0.9115250350400101,0.34536883254210926,0.6945240063033103,
 0.9698708295662951,0.8273158092538427,0.9655502178923339,0.4064517772536149,
 0.21209461990336798,0.4634713217288544,0.26878121087705276,
 0.4595221402215741,0.5237715874199704,0.7866016913788101,0.61169741980224,
 0.9889664289037798,0.8152372503125663,0.9972367838657786,0.9884594320671849,
 0.023959759586417606,0.5245022827348972,0.35673616590592294,
 0.37124375326969006,0.21714533796739732,0.7678041975834233,
 0.26176750070969823,0.2660640180005993,0.8268636110555267,0.9406047479143468,
 0.5545569189929496,0.8189860274096499,0.9789421952617156,0.8270335721229661,
 0.6782389566936089,0.3324648737401338,0.16983282762942575,0.8835727401358406,
 0.9592682601381931,0.39611002805530515,0.3180163747528999,0.1739257295163097,
 0.3487131163470193,0.9840086014397953,0.3466810392996944,0.8674069345610685,
 0.5547414431750074,0.28416911123648647,0.2826553489265633,0.7863303988829075,
 0.6098418561446972,0.9957121346505199,0.5692694784126132,0.6104755413646984,
 0.3802630108889594,0.05171539007921522,0.36948541947877567,
 0.013171554460854162,0.8712128916154633,0.15312698980602213,
 0.905420034272662,0.9280097094575976,0.05203656589199812,0.5545036420197245,
 0.9543374165493365,0.9006938552541747,0.5267126029481753,0.33718857523066925,
 0.8879905725647079,0.3941918404966296,0.33444606046068426,
 0.06522543486660237,0.5019348378105457,0.09434004432706478,
 0.2410353261104914,0.17066542383308836,0.9355077593795117,0.6288617219355453,
 0.3237569309905983,0.037989618596040886,0.8247881417088966,
 0.04931556368801937,0.8700434684203593,0.5100755164429884,0.5768579036984878,
 0.7658927743591768,0.9508191558185782,0.6946203329290367,0.5532196543485947,
 0.6574814606410355,0.06127276609534171,0.43087996248790406,
 0.47490760554124445,0.17702100681645527,0.6874215203757721,
 0.6609591559616455,0.32794041613548996,0.654266011712107,0.5058451541215205,
 0.6532789852311812,0.06928314951223724,0.6439737356201536,0.8758269653997705,
 0.24175413472741691,0.6324549979214686,0.655812040399933,0.03348753849601871,
 0.41013748994700316,0.21889757337639715,0.20781443145317124,
 0.6570801978695082,0.9585012291523324,0.057830040357419454,
 0.30735638715472846,0.04216627756887115,0.6447931553431077,
 0.4109775927998005,0.8976821947451455,0.30768711550527594,
 0.015980237904072325,0.5931440682847903,0.6357920084419352,
 0.4311054617053357,0.34795803693438465,0.7548151719522962,0.9623534696031147,
 0.5750802684027769,0.6407683769846069,0.5657183609500044,0.9528680335270617,
 0.23452216547544147,0.20104046582229373,0.5989467908483436,
 0.8179833552910372,0.5225786009959071,0.9722988766609728,0.3883953927994027,
 0.45823846079270714,0.11341505505557603,0.8447582699783456,0.32313814259622,
 0.25737024547083276,0.7731784524022487,0.7839294256808881,0.329165484819819,
 0.6413955170041289,0.12150810984397753,0.1028627951523633,
 0.31313957937254544,0.4494207003694356,0.08836733150838649,
 0.8422224900931823,0.7755505362611612,0.7459032456900573,0.45397324494848945,
 0.37205553080565057,0.5832805922690997,0.14104029784510164,
 0.1334499605214421,0.8780705784340376,0.5952799270397944,0.25180404391471134,
 0.8905871061423495,0.3818789888827546,0.13198195335493212,0.7439815433365067,
 0.17910649006558488,0.8393576162354108,0.9729144710943944,0.7112163998448112,
 0.45763989623568246,0.02884379157618966,0.44972494218000403,
 0.2977230593477429,0.8777932625637409,0.534381463898761,0.020602722198749146,
 0.15927268854341592,0.09018317036847834,0.1429145836290261,
 0.13083561061841076,0.7916991855792448,0.4575856662986345,
 0.44472457597430404,0.5623057406076887,0.8090064476019354,
 0.36818095146465957,0.32206853406883207,0.7088860014722586,
 0.059578313290736706,0.8300727364285225,0.3749753250654989,
 0.5744803922167026,0.6460798959592883,0.4308490469546771,0.1143924535029941,
 0.2907290598151202,0.10193879623787272,0.12337226349491193,
 0.23073175012637903,0.07550254098083242,0.14392341113152107,
 0.003926737881452036,0.78354764195863,0.5317469740159879,0.361679887363514,
 0.8901963386522914,0.9412412132058795,0.5083944393404843,0.5377724464013663,
 0.8601668031500361,0.019447275401544006,0.708834750337934,0.6593153606045565,
 0.9341214911481348,0.4433978822286446,0.22406191910702578,
 0.39155301544268617,0.774272106503631,0.06570889318577289,
 0.33437464013625295,0.6517228842581098,0.221278710070003,0.48730729339039436,
 0.12251937438384242,0.8564296661781712,0.09780754620417653,
 0.5707723666156386,0.3703765231123507,0.5281963177555236,0.26859604638323265,
 0.7938306913979857,0.6282698071503934,0.5765922623215508,0.8845789997866185,
 0.46893923199426757,0.62930106296934,0.535607687844004,0.7595287984165408,
 0.05641801847274064,0.5493881503931475,0.8721964768864884,0.7450405431964945,
 0.34990712060034634,0.9920944124816702,0.6557232604834496,
 0.21940683153165486,0.10135862894991932,0.7977656938682215,
 0.3540128260041686,0.29995210321786325,0.5607393359167797,0.1476782529135634,
 0.7350562862471736,0.07186616775798293,0.111142242356421,0.6608053558822762,
 0.697569100125899,0.5084633864796428,0.945351229970341,0.8408855436047862,
 0.5755372195868196,0.37227888823745714,0.665996594734923,0.39854026138507925,
 0.5487700661286977,0.17922344762989406,0.7774351147642137,0.6879190874600485,
 0.17849608991430077,0.5979237884177065,0.8106346686436492,0.3647455574622227,
 0.2592195222696657,0.8266905363459103,0.3182492729802491,0.9710312913247221,
 0.8068631142674201,0.6723305856208566,0.7364266984385397,0.7720719909586036,
 0.5265342279147425,0.6704444694977904,0.14008838202369944,0.2663334748316617,
 0.18232052818990918,0.19209467743377973,0.3113522686871445,
 0.6114676318747029,0.7812721593079582,0.7657990697615749,0.7018849044711104,
 0.041149699372932336,0.7594213794147051,0.3478963062759999,
 0.4811011877098449,0.060614132859660774,0.18378675898388797,
 0.2489622999431289,0.4807736818697549,0.9156655308567224,0.14719220081939088,
 0.538644740547672,0.8527700026590266,0.5656835059752838,0.7681261155732401,
 0.7898970304629751,0.3284293534498911,0.1092626288420061,0.6528612858081795,
 0.9397643967095217,0.06829320692916274,0.9846446526647745,0.758598091902278,
 0.30983085120235576,0.5319720049300968,0.3809484736658476,0.9267286234470222,
 0.03984068612688707,0.5836181131233404,0.477591970910901,0.4394991373171535,
 0.28927291760603047,0.5303627773819487,0.41995296067500043,
 0.9892864606560559,0.31342364257893385,0.7218556593919204,
 0.11297327713355332,0.7381736394283832,0.6624061613240096,0.7152441801944884,
 0.9487952360264253,0.054757533018774396,0.5170596645034831,
 0.8778498014189938,0.4583878403201075,0.9944805559365295,0.11002988621964205,
 0.9343113029292006,0.645294652488788,0.2756992823306057,0.5791335696641142,
 0.034762344941286605,0.05503062626720534,0.6013910085321886,
 0.5790228538651209,0.27700186586725406,0.4846357860881385,0.8819706206328012,
 0.4151641678691176,0.1143500254115064,0.10706786063733909,0.8677193468284441,
 0.3027676851467592,0.6397139874836992,0.42237246445819954,0.1500470937829146,
 0.6502231244006625,0.4040376426124417,0.31261941369529034,0.7453445467760278,
 0.4146495390720333,0.7140827031086224,0.2995300496052353,0.4292729166544029,
 0.582043550506802,0.2833994920491225,0.3519707361728711,0.5318180580851704,
 0.9113196782370757,0.5276724748253253,0.6988444343999172,0.18200514235590837,
 0.6190113095546046,0.22575388620092418,0.061427084564895096,
 0.7883658702470094,0.7616924766679785,0.4879354967253612,0.8370862998857302,
 0.7934925103608286,0.06696600935788755,0.9487914784166668,0.9199891331293137,
 0.9535744555014517,0.9506457496155107,0.8447723662769423,0.4161071260895698,
 0.2650243382388393,0.25688215815985505,0.8720538931290268,
 0.06349949681513678,0.6643271460081457,0.329715625647226,0.40720780570654613,
 0.6101362366839235,0.17849405698172027,0.09243078916765723,
 0.8923896044579812,0.01257688652272515,0.16920585047174963,
 0.5411663155446221,0.39395516913681206,0.009257256746751819,
 0.7874374801739414,0.8164354551613563,0.7045654187248732,0.11306666540930588,
 0.4606087817784974,0.6512010779743648,0.0405170848877604,0.7733975622345343,
 0.7030505411919423,0.717991038159853,0.20248244317651942,0.5407228314662637,
 0.8279611882377661,0.13741421380493135,0.07575817192281242,0.832070157110761,
 0.12875125024062095,0.24312909700402552,0.8834760766629403,
 0.9927237748695221,0.5986488912075794,0.7608850941031383,0.28855901326733413,
 0.7251260977090532,0.29940179002692136,0.3493238293103512,0.7070622727352572,
 0.8152289849206513,0.2944678303024826,0.6863120821677295,0.6904671848538078,
 0.16418319934324244,0.31769234199832663,0.06262427685595928,
 0.8222792399925949,0.010893192761153747,0.3602905227020159,0.694935245221665,
 0.11640714644169903,0.7120183491683356,0.5533976312614406,0.6637743831721589,
 0.873004409475937,0.15103269384500773,0.485668136026862,0.35356992562391776,
 0.6875650553119829,0.8100947155103922,0.7140058464491158,0.9455982172092017,
 0.4134541568706551,0.054217771944931936,0.42998360095863897,
 0.900144272810481,0.7026957094219138,0.4048261025374533,0.1757227421309544,
 0.030231616461382615,0.22659788370452572,0.6424957603499606,
 0.9992913110848372,0.5010493657328688,0.6173279667056527,0.25859748897275103,
 0.39051310173640685,0.6854748587912607,0.6085149307125923,
 0.42374832432269416,0.36627859909038596,0.8514550837781178,
 0.6348165087902805,0.029008716531689682,0.37878738185527916,
 0.9887241929051115,0.4855553059156922,0.19871507333165905,0.9318134671314462,
 0.8643949902886426,0.9126073176867386,0.29999834373218115,
 0.46484282543095135,0.062434304339560764,0.9783085644142112,
 0.6184686152105326,0.08494551698817632,0.9368614566640668,
 0.28536785391291286,0.2785361326385766,0.25692102081399226,
 0.18267816945849047,0.24865105477752347,0.2592089805751371,
 0.6457427895700694,0.76333952953049,0.08207028086767398,0.8515929789741203,
 0.8719144535233988,0.05559460752317913,0.5199840670902851,0.9485063240623859,
 0.27346983951358583,0.3457344750754501,0.9712981794968669,0.3204802202251491,
 0.502336985067604,0.5189460290221539,0.40174128883900617,0.3173261563993345,
 0.5777116658256278,0.10061916998337772,0.7752730531751988,0.3551329301448759,
 0.686524615690544,0.5643034874417248,0.21359337511707766,0.947366454829356,
 0.1331557423870825,0.017008008930975205,0.00564203767132565,
 0.5724570633023396,0.2755129862444411,0.9969777109394219,0.31368344608409715,
 0.23631336624513644,0.356786022144159,0.9085501612706848,0.21501453672605808,
 0.8502214087115142,0.6175946591741869,0.8414653849029892,0.5443923481364137,
 0.9329982378886305,0.18148090825219232,0.9227618997693607,0.9327935074845826,
 0.40256362706455884,0.7883499639348612,0.3009550687860503,0.2628255652813869,
 0.8980210792711425,0.01392031124961357,0.5431573061780263,
 0.012162122948297882,0.13575679302325594,0.21019120904365884,
 0.7258852313753412,0.5531609900856864,0.2818404916684695,0.9879230323017582,
 0.4785295559478814,0.8105217476196328,0.29509677032486137,0.6899893093329525,
 0.9226524349367599,0.759138468737054,0.5199808512275663,0.03832889080300239,
 0.7809581283254636,0.6112870257316328,0.5984371117081501,0.34916178344322,
 0.11459532339764067,0.15114748562559355,0.40382046352154555,
 0.6742219960136494,0.10735732768864126,0.3799343431593458,
 0.37437795600583723,0.07683414813523637,0.4076204753276558,
 0.8456310110151436,0.4227388822987551,0.8251236060201855,0.02982909162563796,
 0.6906972932549519,0.24837566175221704,0.1406774563161508,0.7662976275388211,
 0.7533346311406609,0.6365888868007393,0.9824717620683086,0.8046226984523986,
 0.9310562803023973,0.9130533046988133,0.021571643026376486,
 0.2783557114590137,0.15572199999392777,0.4674326448388413,0.8579708905816393,
 0.8567756671390507,0.1153254734555017,0.9449156629980426,
 0.024219373107499886,0.7437234917825082,0.7078289444123437,
 0.2579150010453036,0.4050363028330052,0.22136522508916479,0.6758017541799064,
 0.9935041404927167,0.7018392018334791,0.6562248163853797,0.4882680638859075,
 0.40961412353762694,0.6926730132215878,0.9666981161107474,0.5967304491259685,
 0.6320912195227627,0.12395850559758648,0.11544139537685427,
 0.7489787851286028,0.1882954972380395,0.9320943040362998,0.5025765856197549,
 0.3837354980819606,0.35576071095468165,0.6831465831215514,
 0.14259425710540186,0.4048376229914801,0.6403220831589019,0.9903663356948819,
 0.8175430754247381,0.2337071069759944,0.8126610294821137,0.5841804058428594,
 0.5198595921352785,0.2065164395198229,0.3528639913928673,0.9930376235675415,
 0.05346624053069127,0.17486905407962505,0.9155715227414839,
 0.12123442389496375,0.622967357181603,0.019896934539946454,
 0.9999683063467061,0.6066233800456806,0.37162024155533135,0.9259126365227383,
 0.8490556309240664,0.18357401752273406,0.8262991070402925,
 0.05626531294811976,0.23250582573953693,0.3408815418748423,
 0.7857811648261396,0.5344776648070785,0.4205506425306925,0.39869532746808367,
 0.41742042549212566,0.9281423241616519,0.9170435647634694,0.813421345955352,
 0.008786707650629721,0.22172472188374648,0.23778623088322604,
 0.7564738122066611,0.5361665669838535,0.9356266924269508,0.9635045167397603,
 0.8624323341507543,0.6063564900652604,0.836754830324447,0.8466158598292299,
 0.8942415517360273,0.832345977973308,0.879770481489552,0.17835829148818794,
 0.09420619552493581,0.762146073022475,0.7709622561580134,0.6551742330768824,
 0.7723686495408164,0.7039124327445953,0.682591929621794,0.23019024727595816,
 0.9210404011742284,0.2088953868499278,0.4348505797008373,0.7451453371037466,
 0.21488300764977564,0.5836270887020061,0.5560139343231498,
 0.48842942909918574,0.2023084762068994,0.9111976985546953,
 0.24065980106788654,0.2439402210619548,0.43459039625324425,
 0.5960611438979043,0.42169059648301366,0.7218644502236101,0.6289043324367577,
 0.9973680452340906,0.6028587141640701,0.9379373890552318,0.8009062125743652,
 0.8881368086228242,0.7045813339566059,0.9392614101082173,0.3376630906143203,
 0.2705002510693699,0.6222882841641146,0.5701160050532122,0.8613512530705038,
 0.2608454517969514,0.5568543920674827,0.6123031414957326,0.858559193813157,
 0.805689272275391,0.08250651925500196,0.5920923983777686,
 0.024154437106789883,0.22263810741864865,0.22305517831779564,
 0.9158478012064581,0.16448558596560547,0.1890831806334261,0.7698586732580047,
 0.43247326367343375,0.7608059587605586,0.03477318350000569,
 0.6866126796852459,0.39152328738295217,0.39911123774620627,
 0.38001773095194313,0.058415252617340596,0.20811595746975975,
 0.5908424079298998,0.3253065794467046,0.9119068271013409,0.7542412763462192,
 0.39863312553791985,0.17891550300206283,0.19953213254176194,
 0.03242006012270027,0.45127044459712506,0.13422904449659412,
 0.2056516358563073,0.6285814742296502,0.8845316002019414,0.7955611769691335,
 0.2036889854573528,0.12245719343113382,0.4218022678093676,0.7415088723804345,
 0.9893105998261275,0.26424834378385187,0.04598538767288629,
 0.005001476940753702,0.948872688453312,0.44003268108358895,
 0.5153400169560495,0.7263462947836715,0.3431169945594562,0.11846245787737542,
 0.05099933214060015,0.7264834116324458,0.8732967817461895,0.6065271221128149,
 0.10215362926912996,0.08528180645792638,0.9816925006241488,
 0.8232222774083032,0.2526194998635526,0.8582690294726065,0.3584647095139569,
 0.6595413956154029,0.289968012581757,0.7638511236793081,0.8033054919916006,
 0.47262769639345226,0.09887172593041615,0.7367784028749529,
 0.5564491559046061,0.5298856282294354,0.9002916733840622,0.9346800512924544,
 0.33848035846962243,0.7318927998513017,0.2134908507607226,0.7623027996696946,
 0.6000056296939872,0.2997926724941147,0.5655196877620372,0.00988869586342389,
 0.1301623703748025,0.5541084661939488,0.23378510213022463,0.4809267341679033,
 0.850534548505648,0.3316508102149881,0.7475499471040535,0.1534653437624699,
 0.572104870037637,0.6961061499552864,0.2886177886148016,0.4776769230945128,
 0.5141825409320481,0.5257420779756131,0.34093808803908676,
 0.41289144538577394,0.2750292009437978,0.11572452665694488,
 0.07189382858974591,0.05571953213686143,0.8405183227447411,
 0.31321718206417215,0.8848443432067479,0.6247028142938297,0.4721628883089575,
 0.659512620158274,0.8149357538291214,0.5247343950912715,0.45286411450112407,
 0.6156662972090191,0.28467539431858246,0.36208076909095244,
 0.9373658338977477,0.9264738789513106,0.131077321453274,0.7102500382863456,
 0.6704826056598232,0.7583921604150398,0.2672012440424575,0.4923074451907984,
 0.6241443616319985,0.6045614611530101,0.3145816878902664,0.21561692334528282,
 0.5728041772466996,0.9849636108660057,0.5016368257510326,0.5153470149100816,
 0.7012420357902545,0.20648277358557032,0.2926236017873447,
 0.18684540158568397,0.2999450280463051,0.47604661147737803,
 0.2163036273490413,0.38064726452795006,0.697603009075699,0.09133488319101879,
 0.8517840168330434,0.3047639231463476,0.3993012511021803,0.7148250321834679,
 0.8931552205448469,0.47340641197546374,0.5024887661765066,
 0.28703164812749793,0.9995298558088739,0.05709163314447996,
 0.3578888348328447,0.05000258826189741,0.22164764003721216,
 0.8956043021331427,0.7272650097056512,0.23771291557034224,
 0.45822277057243466,0.11667877905828006,0.007594382810233946,
 0.5556509797080597,0.8768532798696892,0.6032599833239669,0.2208068482373906,
 0.9957863784374352,0.616104154235099,0.6033266732152316,0.7829127521511776,
 0.30141718477403945,0.45209156351818847,0.8577215940013341,0.653036502706566,
 0.13046318617838426,0.2638357049414165,0.5736204587296436,
 0.31282724903479164,0.08249975195664216,0.3663799114860786,
 0.7758601908702566,0.2644600598110759,0.8091252780156952,0.6984867756319726,
 0.14790187667276267,0.8764758991029155,0.41336555426509125,
 0.6480753840056047,0.7548784114543176,0.6639309472587556,0.9432845560955966,
 0.3988148184668736,0.5355870537848852,0.3238874834671127,0.9115009915524659,
 0.09625986530721342,0.9525172188304429,0.03250067457454919,0.097984192809358,
 0.5177712076482875,0.22511334512280012,0.2800044638666047,0.7232485020279267,
 0.140590291509459,0.2266129958638099,0.3933209307306288,0.7061826030996017,
 0.5110748478807573,0.9520075677637122,0.6917404244440221,
 0.011354736620766026,0.887052114360303,0.031343764896062165,
 0.7623005077391928,0.38818903863667265,0.6344918301802143,
 0.03237806618538985,0.47436526553169067,0.04591270931334113,
 0.469392828998594,0.22300888376204842,0.7019585318540281,0.18451769242310745,
 0.07037350061837022,0.6236825060662079,0.9900453422841262,0.247841164009728,
 0.8598522233186993,0.8579157437022367,0.4656137176987184,0.7970259455299146,
 0.3703495155559192,0.5689734535264912,0.2566882587505729,0.5868521027521831,
 0.016348996691876283,0.3887143333384204,0.7256825983599748,
 0.03779791622495121,0.9294354000182687,0.5145115466991049,0.9178146926276884,
 0.2281112575204034,0.021626429569116534,0.1000125668914229,
 0.7770053032085364,0.26430135706469327,0.40281806273771703,
 0.4909427760718672,0.6438728663602107,0.6275634305727373,0.28793111940033933,
 0.4206746247004052,0.1313091661980862,0.4888671526836925,0.13185132906327846,
 0.020476425399698295,0.2595885908604385,0.5296162243775285,
 0.25142266559073545,0.341441075051239,0.6385373083904534,0.10294180362184913,
 0.9264959356548947,0.6154860626869423,0.44614831123239573,0.7152272724940181,
 0.12009044952689407,0.4760596567301849,0.30192972443616994,
 0.27519462285164165,0.259244589241757,0.968257837472316,0.8283138709552852,
 0.24498638925710226,0.022090378918034892,0.7695921205466278,
 0.7618567625964966,0.6181374592699842,0.8597932067431714,0.8047552831815666,
 0.8955391702449571,0.9480315335404574,0.014474583922089446,
 0.8798434713154756,0.6875654837944772,0.34680540846922514,0.7426018920562762,
 0.10012838408318925,0.8853667907839068,0.0891746387038761,0.7712430331825728,
 0.11555398649273396,0.9101121545024881,0.3934252592611466,0.9039746565939576,
 0.3395355657377065,0.4122694736462097,0.39457102496563623,0.540043643514384,
 0.5587756534508327,0.29479003716511265,0.7764360241702952,
 0.07915482441701582,0.4680435459366875,0.7885318524580333,0.3261026364354186,
 0.8045075752337929,0.3818906448331263,0.80639301374365,0.5029822031679715,
 0.41038892119141057,0.6787146698364344,0.9452733688849781,
 0.09781856882406625,0.4386154680795338,0.9972401306092707,0.9088724626116003,
 0.547523157342139,0.4649580400914044,0.20606407310487285,0.1305680728107641,
 0.5484898305468452,0.6861327829794132,0.9115135004248176,0.7644481255701083,
 0.8699275135146594,0.7110333261341516,0.871240286707543,0.9583965438651671,
 0.5564230789966401,0.359820384162417,0.5520390038900542,0.7610584391668862,
 0.8371112003680969,0.07409311793261208,0.2033816859856662,
 0.022352516569044467,0.44341102426176227,0.20070925947093388,
 0.5285168319079429,0.8196470110612433,0.2555983417269786,0.7800017179560964,
 0.016695052468373883,0.2022119148695649,0.660087887464062,0.4940893477360989,
 0.9285502619044801,0.7058117111549653,0.27644365464488296,0.877024926802398,
 0.25715484388416854,0.4484597388378314,0.6385266207958137,0.4658249126701971,
 0.8421756035878067,0.9499296154928087,0.4134368996091691,0.4387048134587107,
 0.049202868000247646,0.16355258395798777,0.5730404841693604,
 0.5991063731008005,0.9966886761703141,0.2042419584776175,0.37237925611007305,
 0.40508740229697426,0.9630949255094852,0.6802835170361113,0.927135726200741,
 0.11293747515227781,0.5908297205005437,0.49380390769530735,
 0.11023161171690521,0.20016380644849985,0.8910352167668948,0.817895838160993,
 0.3178650775578029,0.33851769008266686,0.10916553002822538,
 0.4801474409210965,0.5015409787233877,0.45662934009454825,0.9817145376415356,
 0.7844786896488022,0.29845468793227714,0.615478290838475,0.4404599784310401,
 0.25805682467814295,0.37083406671762253,0.017340111537730407,
 0.6111528059853318,0.6114331224990522,0.48793628941561773,
 0.43342429805867333,0.2508648493457446,0.10704837192523431,
 0.0879484012915609,0.4050492029545819,0.5255668335992328,0.7391714838342353,
 0.4148776834384432,0.1627116626159043,0.7307832648586726,0.10281073639666272,
 0.5620509396382696,0.6262406874315649,0.8714375058947504,
 0.048311738715894115,0.37994381417765233,0.7945658986984244,
 0.017838393501873107,0.22071868812674,0.8433662639421637,0.2725852599841837,
 0.035126581721390826,0.8173199181639466,0.9383962505802581,
 0.22784399748343875,0.3831964431460084,0.815555194280867,0.6755393226707602,
 0.10085892998969193,0.016979062184540816,0.667417506595291,
 0.47183227877715894,0.4907760570463139,0.3708339315686846,0.9020233742604843,
 0.9887003261336932,0.10886339803155365,0.8128467412207342,0.6170266354076195,
 0.7940446974211978,0.8853583526572782,0.4855099590602734,0.029364584887832,
 0.6932396272598498,0.7120692156857997,0.9404712379804817,0.4028536950568308,
 0.6714661648436109,0.8753395235457553,0.3571621139226163,0.18379563964267154,
 0.9195734152831473,0.9595560720858872,0.3144115091213232,
 0.008953829070684538,0.6791710159297106,0.11278407860236184,
 0.8678257100703608,0.5345134790648912,0.8369885905452186,0.6640523821025248,
 0.07591961181018536,0.7101221847943687,0.7550306760414505,0.633374280131526,
 0.5516059309139951,0.6702326346684214,0.903466542645128,0.04172550658444485,
 0.9253435922737245,0.9124283864928249,0.2813835831776911,0.9945792499990358,
 0.6207973949983238,0.06818791647074962,0.976403485104992,0.771684743826608,
 0.9426326010402101,0.6610619982649965,0.40493976348843774,0.7289804468360951,
 0.7624389417373606,0.5917852708924592,0.6342365897073039,0.6077731291278524,
 0.15806103629257462,0.08509628833682803,0.2768764256486994,0.636398231664659,
 0.9119909721167699,0.942905105684557,0.34588108487391844,0.25012902651443847,
 0.3562562415970304,0.39599209011866165,0.8947079141496036,0.9412263954690228,
 0.6563971430683567,0.08938296855184313,0.8216033103708441,
 0.25502241694694794,0.7624459410620811,0.5372132245054266,0.8330525366171884,
 0.4810560994193088,0.04338214021989695,0.4022537785531727,
 0.10382915899178968,0.43620029776294356,0.5788131317092632,
 0.00703187102428815,0.24053475732174723,0.15607011867668497,
 0.6643375118022313,0.3950195357084114,0.21607108837816047,
 0.26060942508402096,0.3835063494624964,0.9163169624384245,0.6501735470756362,
 0.5778162139727507,0.351256577846078,0.338732930608705,0.6073099595281808,
 0.6013296984280009,0.4806810765129229,0.3654767665158789,0.5135817015263021,
 0.1834291918940001,0.2351881353817158,0.06502613896891485,0.6132068735458184,
 0.5987863037562341,0.43234352162202283,0.7556377081644481,0.8405988237483245,
 0.5789611086026808,0.9132206097529715,0.2753070950730391,0.3828722732723113,
 0.6736388428568999,0.5349855955717417,0.5561411472172602,0.8683914439512664,
 0.8244662872708131,0.1476410154330734,0.7707290610816299,0.504832727919265,
 0.7730898740236487,0.4126199703660114,0.9682750496694991,0.3967969995695072,
 0.15301408487287071,0.8155706549546562,0.601760641530442,0.5690016814143739,
 0.7092676323954008,0.5755209433000316,0.10342686901762421,0.1505199162384483,
 0.1349857011025043,0.980413409359382,0.7323687643636385,0.3508050657805295,
 0.3382579109067565,0.7756804263052866,0.2788655968235454,0.02325769659107635,
 0.4342536238500183,0.922370275308506,0.11521817932403022,0.8129035557360794,
 0.23922252973363434,0.6785555694374207,0.18530339689033126,
 0.8246877705781123,0.9586625944226468,0.7420403488718825,0.525540526127268,
 0.97140891656831,0.18898190552167005,0.5820233185237635,0.8938696384218948,
 0.7344737922329262,0.9871115154178434,0.6032508299638818,0.8659996382443126,
 0.8021473828242743,0.10638355130161659,0.7231438784705887,0.3028988408561677,
 0.09008701128022689,0.545181360049859,0.8940284049102976,0.9741509353310868,
 0.6930959215873459,0.7896002438427714,0.7776752758585486,0.5817307076170669,
 0.7799516611226126,0.21071338550553265,0.803094329549544,0.27420032151431917,
 0.6402007681202512,0.7177172937113382,0.9523765495491147,0.43598242512154506,
 0.9900281187046202,0.9950693996277603,0.7868275784127207,0.714592600739981,
 0.5924544710789129,0.41832667036483506,0.35109767296023064,
 0.8909596808089122,0.5643447075877083,0.02353231240466125,0.6668924845637446,
 0.8238931309396011,0.7642933160652347,0.4533291835986293,0.4845830896995462,
 0.8990770161001242,0.04456755535153767,0.979647473300212,0.438767631693597,
 0.7755495480022176,0.6636561067105998,0.3006129496167055,0.03066125612822168,
 0.3616735937249508,0.5208219090095751,0.8799956880444747,0.9994754382361867,
 0.2499056626805629,0.5965004417555075,0.5133131886828312,0.19657184472724443,
 0.07465255742487975,0.9852034744053546,0.17683371101918965,0.834767442920602,
 0.6170893158041146,0.31472580926185045,0.543909757604101,0.8759215596438288,
 0.3072617513212611,0.8473425223700807,0.6211078405984705,0.7400199207330505,
 0.4149274000732225,0.33692684259732986,0.20129646933708023,
 0.5099297334982199,0.6280583137450059,0.2714741170116206,0.823164042964299,
 0.7756156218948898,0.13326242751914763,0.07128774613208533,
 0.4396389162390364,0.22272507512487538,0.45092214344784,0.07354835763202305,
 0.9627321896580526,0.5755743272266206,0.6471372871480142,0.5494199368972854,
 0.285250436842301,0.0637386763306449,0.538927313816543,0.9212618306128292,
 0.10278884162547797,0.8973397697872134,0.17281738881562148,
 0.9666617583729105,0.4754903146042774,0.2629303667917886,0.6589047317500291,
 0.0598182453716023,0.38597458369475324,0.8268852612667754,0.7722621846223863,
 0.3202175373393743,0.7299062857564521,0.7549573366107388,0.6709044369821744,
 0.3745286074648322,0.18843999837493997,0.5953753274546218,0.4662402000182825,
 0.8614285296334578,0.49397130651099863,0.7147040910371593,
 0.06769497292872884,0.7144847800868226,0.3565771146011236,0.5361573019963559,
 0.616083142266622,0.5843479484585474,0.8894505444566148,0.6014669429997761,
 0.5476689019228577,0.5240569415045221,0.31962866564685255,
 0.06828608523597812,0.002174642300837082,0.17693720771123145,
 0.5565265113516288,0.4647576625983062,0.01933803911559273,0.9071484486223601,
 0.9093943173089389,0.7829011323695731,0.2441867764340675,0.15059145305107746,
 0.15287676289568286,0.21698077173065589,0.2780456396921963,
 0.9339441962729964,0.8539960236762691,0.10090788928229699,0.4163863888444158,
 0.7896060275862129,0.13817087250957394,0.7281317182846542,0.3654624084161786,
 0.1412343010032936,0.14027627527668596,0.10109358069830277,
 0.12449522523492995,0.6360781062655688,0.55783387029398,0.7137349169558178,
 0.6656487982421101,0.22633277570090404,0.9260751377354834,
 0.14210774629876022,0.008239828880866162,0.0821113052407424,
 0.6283560059950662,0.06106706493986258,0.7227711731189926,0.5853943747835191,
 0.15320348551378415,0.6102537030930162,0.15050367078497295,
 0.9326123296522395,0.5093238813470096,0.5202329107673697,0.9031901831719297,
 0.014946200190780079,0.7797982649016902,0.22994048895275476,
 0.4842534357281981,0.1224249500684601,0.8555221292924852,0.5222386975403432,
 0.8899239871203349,0.33624383630908405,0.8310346340068938,0.371847409402687,
 0.4898635311413122,0.6807881289881287,0.39558619471587175,
 0.11665975762146163,0.6433340826652695,0.17231306859512752,
 0.05189710918908985,0.633355311872365,0.0479828234550268,0.7072611949514154,
 0.5484393610162707,0.6984958129045679,0.29589181384119645,
 0.36269038620191707,0.9033492989665647,0.8231497173929595,0.4181877378301745,
 0.1643686152490258,0.20923678444953897,0.4542725623792261,0.3801682274987499,
 0.05618128544882994,0.9108020080828124,0.7427497353675439,0.6035241776318361,
 0.298374557708629,0.2781415569228243,0.5738511454979182,0.9446234089259979,
 0.9674596054551013,0.7604651422622999,0.9390980716412755,0.7925487870556152,
 0.5576326075184067,0.813120159390559,0.6272031958102311,0.821530261958935,
 0.45545736577496887,0.15361677040461974,0.049172569638145425,
 0.6605275068075944,0.9955717584927768,0.26143486322943477,0.4493771028330964,
 0.9002037677688105,0.3339054568394655,0.02221493520490092,
 0.039903936537980966,0.6490319124826552,0.8229439900727205,
 0.2237116081004915,0.1713032239970249,0.05268379239865251,
 0.15181789123079747,0.466762015924131,0.05598954398886069,0.2487747891180404,
 0.18744606605305114,0.7562010638470759,0.7961062235783865,
 0.14574395540437968,0.33791940212279736,0.4892918392120812,
 0.011272079699129134,0.26281966895958364,0.5656281892255951,
 0.2991152952269438,0.8312458226407021,0.299658219672853,0.54604307651273,
 0.5206894635457713,0.6390999294882312,0.35585468745174453,0.531984199135535,
 0.19490071234453044,0.0446360152097377,0.0273804244351048,
 0.04438383941201507,0.9452019675255616,0.5225432436179626,0.804016059510941,
 0.48730846968266617,0.8551665072836951,0.6545117206291075,0.6488462121377344,
 0.7973512935742315,0.38944773660689735,0.7167289671480261,0.384006435944361,
 0.18181171938850127,0.3826866147725313,0.6084702256164614,
 0.023250561552900062,0.8230325111499142,0.07127589969471049,
 0.7037405993445669,0.44653908683764687,0.3221933815463336,0.2648011372823783,
 0.734340301340829,0.2134982673162622,0.9087802966789689,0.17167196024447595,
 0.8044603957278045,0.13059313064769484,0.9925306849569342,0.6521301728126454,
 0.8638438330606405,0.9644221715101593,0.08178379920528211,0.7318723956786422,
 0.12078894484526537,0.30979946599801267,0.9612337306106946,
 0.5350826065939491,0.27608321784967493,0.7372191287415266,0.5465978924499424,
 0.488851972672789,0.6899047657527665,0.6999340341520248,0.1745751630102761,
 0.9462607191657602,0.42660520854469497,0.3913559383044872,
 0.30132861352659823,0.7804253899634931,0.838847242689892,0.610282784697724,
 0.11302606483218314,0.12373799516466422,0.45587347826208957,
 0.5882950825806429,0.41869263112052846,0.03859398079489085,
 0.3927340188271893,0.7339696426736865,0.9767817035097892,
 6.960346167282916e-4,0.8868876031418608,0.3776340096497204,
 0.3215106362325062,0.20636792368139767,0.5122813265142355,0.5921790230048731,
 0.36550973104838047,0.09473379795536596,0.5640962965999424,
 0.22100077108713956,0.1405235714691767,0.4400646701170373,0.6299151051602532,
 0.9469120982517186,0.4972244457156856,0.6918038940087096,
 0.006051089109848751,0.5768817349879022,0.812872731379994,0.5789683330600547,
 0.23645417117871204,0.3591510754263709,0.7710799373618714,
 0.07851961726943968,0.17555952830886346,0.013779047931098853,
 0.5728296330766284,0.19985601547199328,0.60534638425526,0.35712301761433507,
 0.4657408756893914,0.698562674246579,0.8082292743572146,0.39771051008314395,
 0.7019225651867879,0.7432048629086198,0.0410903032576293,0.22496039789261457,
 0.3869830069682699,0.6890350442208236,0.8266568981485459,0.08006865303409882,
 0.9233802040786689,0.3390213613450782,0.9636497210944095,0.45700151859455107,
 0.9702874278959879,0.44868287905528137,0.8253086358170856,
 0.31004639191181116,0.8933834655617529,0.8656624226865877,0.4844531227057052,
 0.20998939045546516,0.6946801073760547,0.38930512756560265,
 0.8463541643817362,0.9771893984417647,0.44639132984499275,
 0.12062123470970831,0.9711957189108089,0.4472416795260238,0.6216334115981905,
 0.1812363703346218,0.06334716401233764,0.6089483932727658,
 0.28355731574118903,0.6677513993666188,0.7999390169329333,
 0.48793295964718086,0.7081863606134084,0.6291248749790019,
 0.05939341712959101,0.5613558919602545,0.5162966488155082,
 0.31075852289263217,0.11511390887482265,0.10666347480725236,
 0.7676026340555225,0.9872420341481747,0.27430692344609653,0.8631442852233121,
 0.7195309442575732,0.22229405856315765,0.1417854074159104,
 0.49798362055670675,0.3735162291588865,0.5424300540820381,0.5331946754027699,
 0.36998383878943775,0.25446446962994496,0.6370492382345541,
 0.7690465152578015,0.9924467904260252,0.444456328561749,0.5115088891493784,
 0.08531703726142059,0.6746436807225878,0.37642806140096596,
 0.8792970463189644,0.9251263318916091,0.2679412474195334,0.6395284952408384,
 0.8298526212111625,0.7301756404566094,0.8240835376590439,0.3742722424108781,
 0.7337149682065894,0.8992962840817558,0.5052697073092823,0.34146492669195316,
 0.503713982363822,0.027856967395467058,0.19487928254458942,
 0.9503955556511455,0.07174575178868081,0.6141673389858298,0.7490446259575629,
 0.4420917437771843,0.6610864753137293,0.8681704118799257,0.7181772860266493,
 0.31321583962044897,0.6438801778610789,0.48106713119919076,
 0.49025303415326404,0.737484705657922,0.02490209536571797,
 0.17331145642621704,0.06293516672822053,0.1985366050004962,
 0.6209906886806008,0.41974567221039705,0.49831506354143074,
 0.44318091577629204,0.49569538870932117,0.695429045265386,
 0.09530080173745137,0.8939258579183691,0.5818541282078463,
 0.34154043002385004,0.5897282702598218,0.3358980452448077,0.883890832158589,
 0.8342013423457919,0.9834481870800957,0.23759746018723416,0.5355146692859741,
 0.7420519838691633,0.9520182688697972,0.30133304249599524,0.8576762986068396,
 0.1062677946367423,0.8466757180317827,0.5989558766317115,0.6878811660894463,
 0.21677829758461875,0.9777451256002657,0.6551784873419911,
 0.035639048946829455,0.5294694397047088,0.9792449842687887,
 0.2948007585776019,0.18388133215016234,0.8006849179776243,0.7053303808776978,
 0.42068041331091943,0.9882507844378878,0.1447630742247129,
 0.37460884246984427,0.6994130628115252,0.711990372813162,0.979595970417017,
 0.040527302069442994,0.15303960840005837,0.07125284960368516,
 0.32120522242299254,0.1964759395900274,0.3422663497726419,0.5234654021461607,
 0.035875658766776786,0.07285999254071729,0.5774537189907923,
 0.7878988217462769,0.7220691713501094,0.5425420535532417,0.3273071621179917,
 0.20754146870209733,0.32495282802124503,0.08852217780433247,
 0.23599377891816842,0.7193614128620343,0.4994677685088166,
 0.10221795044252568,0.9332685275968489,0.5534797878529921,0.9896419281969439,
 0.8234695459275195,0.13861401397682216,0.6047663139234091,0.7710912483451527,
 0.09074410091001694,0.3051416652796284,0.19594510335194715,
 0.12075002609635943,0.036790183979228175,0.23281301520907127,
 0.6561995546795091,0.18565836457639462,0.25447211968925876,
 0.21687514664703278,0.2392831971043945,0.07902887126504521,
 0.3705744505750286,0.38409009240713965,0.41528124422473056,
 0.3453284794299223,0.7528525733597375,0.7760786444355239,0.21774279265154328,
 0.407062737261109,0.42834122930933494,0.017529242976551007,
 0.4331142196437484,0.7957945539682993,0.22638798863434018,
 0.00620979817320555,0.14175856622098837,0.19730077386676637,
 0.6087344188257867,0.9594522162602763,0.14165645746198763,0.8184373920020104,
 0.5547499132624427,0.7135631893570206,0.9079564606189031,0.8500143780035644,
 0.9729826839056595,0.931298324658554,0.8084364994721859,0.9920845297233105,
 0.399590292515022,0.3002514329393313,0.3780625921909452,0.5177679145781795,
 0.013667255189129479,0.8687903143538724,0.7855689834167283,
 0.9337757197727806,0.584711031391856,0.3756760775069983,0.6626774659343373,
 0.24366442748526052,0.9201560731061409,0.24941426465501382,
 0.8995579536077623,0.43814823450688745,0.7555156274307401,
 0.29878276551870864,0.9798433278369917,0.02659304653719774,
 0.7117463784072904,0.10007031998650162,0.20602371075570058,
 0.8610247003434829,0.10947389793813933,0.07379766005881061,
 0.6270483201386519,0.30131897787114825,0.41558377848421224,
 0.36361186047705296,0.4962426966341509,0.4517027921493837,0.7736462649752618,
 0.6098721075504241,0.968268237264203,0.12581865545005388,0.9271968044918927,
 0.7672564044535101,0.6180725239115664,0.11855231064349403,0.5782749600213944,
 0.4059358067239752,0.49461716210525764,0.262852758581957,0.7431692846751362,
 0.7230256824779882,0.4434799222697341,0.31919463732437336,0.8986105282084662,
 0.8924194154605803,0.9155402128035801,0.37972534850424633,0.6515023879721432,
 0.16832070943543964,0.9059561949656526,0.7443377910956391,
 0.26174120711649274,0.6737975666523672,0.5937122548437418,
 0.40177553226466256,0.12153541222578157,0.1952547668819523,
 0.4996365999357326,0.2231412222487117,0.6648808901373497,0.47366408282636785,
 0.9583432633723918,0.8793254921505996,0.9902571873294722,0.8195405705669253,
 0.1688545767283146,0.15773007350787782,0.17763131911846908,
 0.0494201203836232,0.10256584294846993,0.29144038584278953,
 0.0394091901122694,0.149786252329426,0.0872239147872903,0.016981540027805386,
 0.18475911434004577,0.8848605054348956,0.723769665831663,
 0.009649291177431518,0.7586252621386494,0.3039121976966649,
 0.8494091874595299,0.524909918394077,0.15360930060474032,0.15274824679938492,
 0.0031848339330968756,0.15852456909361345,0.6616333633391903,
 0.8278019931952422,0.5149452180220185,0.3781106701758752,0.7109730380857878,
 0.094109986505726,0.1437606737444308,0.7159829408362329,0.8080154101912103,
 0.9478722989798046,0.5601868918294092,0.455876611213775,0.20600190983114175,
 0.9675102842250142,0.11169741651101883,0.7078563125541238,
 0.06390574788092518,0.9631669650424264,0.4990536066108391,0.5383602380588208,
 0.6248861351879554,0.9289802505575889,0.9591348595331282,0.28722445708632727,
 0.9985528445230148,0.5035982572025477,0.41969549095023595,0.72077456366636,
 0.8595881675789211,0.5367330300349781,0.4932847803497875,0.23089831152772744,
 0.2143488733897959,0.3540030430870066,0.7365360161523322,0.6413826600872906,
 0.5551440351895075,0.5128638010293913,0.04721049163670299,0.1998422650501428,
 0.8509137621975978,0.06927239485813752,0.14048438839145416,0.674695249677348,
 0.9756129115011434,0.2875643997605558,0.17059514185601055,0.7916495249647757,
 0.40718681197477125,0.6501717044449707,0.8731513904280581,0.4630375191387548,
 0.08381729768901991,0.07886802685724414,0.25185436647014103,
 0.6328329574093623,0.20343223334691263,0.921678674260489,0.2494228228132399,
 0.4026822722687382,0.07426386881983038,0.45982158463272893,
 0.13349924694420867,0.22118558099340868,0.8103893325186358,
 0.5489676335936193,0.3299530111418719,0.6769884971711612,0.8529736913207313,
 0.21245855697152516,0.9771782138439433,0.884422852518067,0.41756326291053447,
 0.7659100585809211,0.8386167011694704,0.8250316202350441,0.4870161689054019,
 0.09525036063611181,0.20451261569335943,0.9089863708372652,
 0.0906416135586513,0.6142435393875647,0.34197054693785434,0.347824332809881,
 0.4491426713073613,0.6135893537457353,0.45845170682491665,0.5480023793635833,
 0.8068365430912355,0.699011440464922,0.4594283838923059,0.6227039163578634,
 0.5776589968878206,0.176697045032153,0.3796779021676848,0.21931220203082924,
 0.2685950719756841,0.06426083088178869,0.8432615803222537,
 0.013218870196116894,0.8260753842629528,0.5774691016561545,
 0.46431074715450704,0.22003076898167873,0.8367967849837421,
 0.6578769057257493,0.044798962739180714,0.09329211108459401,
 0.32079043147643826,0.5205104252095359,0.6072623683311494,
 0.48383617459262906,0.5987300755748284,0.46523798036983377,
 0.2987331044864576,0.5895544509617049,0.8534238717302363,0.8312337441222469,
 0.9937475369677035,0.29584264175269104,0.4584685586878533,0.4531957191441073,
 0.11883243087980522,0.8790263704148791,0.9355893880027996,0.7352938420813362,
 0.04598528737565655,0.6279001571344968,0.5978381809962762,0.5211402139953141,
 0.3008902414893754,0.07062013341516105,0.8864992777939349,0.7466220048259359,
 0.10628579525715398,0.8642718769790873,0.8515913642392631,0.5343852655054999,
 0.5149664365982868,0.9309340014138864,0.2872254117414923,0.6217264027552087,
 0.9436800071889364,0.9807306843055955,0.03433382732143264,
 0.39557277050457573,0.7492933673966825,0.010690081144719366,
 0.41845246131058156,0.6599631571496358,0.027584221711375312,
 0.7062549188001795,0.26452934567794806,0.1535154757944217,
 0.027587988534528374,0.42716263043640124,0.04851442435153397,
 0.7003649397801424,0.4507775163727442,0.6633108674264232,0.6577776639648112,
 0.43427382801514036,0.5995701587535357,0.08258233066875975,
 0.5684739136169463,0.8352621199386259,0.2388199617546658,0.02084741113690769,
 0.7941304733667971,0.9564913407652366,0.6842564991611141,0.24604649790995636,
 0.7468334284831827,0.03591813183354775,0.06137477659771595,
 0.7437551531074026,0.9513139726104528,0.7563674201244059,0.7603773559867892,
 0.5992066386076086,0.46673555438398096,0.5031579537950195,0.9141298701185535,
 0.19203656115365086,0.8031105913202471,0.7611030499823356,0.4458438203876407,
 0.027592103975605298,0.08945328861409496,0.24538644680753297,
 0.11872248888709747,0.9343964115576621,0.19929567605220755,
 0.49042624418629355,0.5507050615712517,0.9895330059967886,
 0.45671792453264004,0.14191598538850725,0.7212697560962413,
 0.5119960544663672,0.31652849125514837,0.6204899264183101,0.9647838978816039,
 0.3598404760334861,0.5681106578885022,0.31815340815840765,0.5812948437778103,
 0.908740949917273,0.8952729313687573,0.9716052769869086,0.5060184019230647,
 0.20341607793002958,0.6238907626958329,0.04599500277010482,
 0.7579995921963273,0.026228116306994842,0.4894819056560785,
 0.7055500130387615,0.6694734621537882,0.05308074366762938,
 0.21646977461645023,0.13021262267802158,0.5240147931829743,
 0.44415838873374514,0.36204478482267677,0.3280487508742779,
 0.05377109765937893,0.7292931209052648,0.6768443414742191,0.2493137195666455,
 0.015090649938584666,0.8595749762339192,0.9823495834415724,
 0.22591530064684773,0.9486557212935409,0.11825366510748403,
 0.8443379841438561,0.9759262788875628,0.4076743915421217,0.47094003328759515,
 0.4156198496186464,0.08068808571348962,0.5526273635070218,0.8446280955508011,
 0.6031082929938335,0.8786815490255887,0.15101815021720377,0.4014460327356163,
 0.9052644699457123,0.9241757876705127,0.22650276233805344,0.6285769985826729,
 0.39239010875537594,0.0034987972313130777,0.11785370674697648,
 0.541915925545826,0.42955646169028516,0.7337727905534759,0.5336599227965706,
 0.5533145878060431,0.720930631432051,0.9401589516831707,0.4364541383111883,
 0.7220183587068992,0.12627626859561902,0.5235913911200142,
 0.49830818139366406,0.4693376681284439,0.8489102437188196,0.6855206620752902,
 0.8281737843472825,0.7306934966310872,0.4840082831457203,0.19693728761010332,
 0.6421080765616716,0.8852125938547896,0.8785579783616345,0.07969084433173235,
 0.41867087498585964,0.34471205696272067,0.20046413094769167,
 0.9341034416859522,0.013635538749829812,0.18286895295456318,
 0.17889628028720395,0.0036548111491034874,0.5745227604202332,
 0.5171354773880946,0.9511567934420917,0.8639331637111473,0.15166808506124507,
 0.3960897559503882,0.08640173795229411,0.32586375187680483,
 0.3423446456894992,0.35093043762971987,0.9814968925933661,0.543846089669936,
 0.5382922483790248,0.9396620972699004,0.5165888151215281,0.9510250712093224,
 0.09861694633488838,0.4614730815189656,0.30287322089376856,0.66171605169364,
 0.4055333429556933,0.35304716010934756,0.38347684506957336,
 0.9672828245432402,0.7533176692362481,0.9936949168327921,
 0.0015175805823537392,0.3814138830081837,0.8073205153884668,
 0.1896373789989998,0.3940420522696638,0.9838816051981057,0.44707492560155504,
 0.9149788582746383,0.8761808714085396,0.87919851856058,0.3635825896934052,
 0.8286857547056987,0.4994605978640557,0.7294452552991216,0.6974523018643816,
 0.8777369682621607,0.4677716231776552,0.3921715497954624,0.834446078819886,
 0.13105009580130178,0.5330285642050536,0.26579151381636756,0.091804953493556,
 0.9369083008170758,0.4403326460930258,0.6943476945547525,0.2652626417020404,
 0.0891965597959602,0.20048668077194032,0.8823383552794186,0.6914456969457895,
 0.031079017847362334,0.42391512312867863,0.216352442155746,
 0.23436490281837485,0.7177409731681942,0.3197262148387089,
 0.13565410645019904,0.41348611759446274,0.4573683849123662,
 0.17181413783384125,0.31435326652438145,0.16228760025422861,
 0.2029886108553962,0.7337859850877982,0.42423624609416644,0.9204581226447175,
 0.5090155823940496,0.7246567822595431,0.4024354342639691,0.8035255667075188,
 0.9950976007020493,0.07478742800901206,0.29917612122289183,
 0.9024504565083497,0.04256838272381924,0.41559922687667883,
 0.9686900888516992,0.8377455937754336,0.6423447267009428,0.9907289521904163,
 0.4979715843996866,0.358529035864368,0.40358095967254437,0.5810269745001122,
 0.9608777670297927,0.9191396863866881,0.2594126775917046,0.7405664308615976,
 0.10878546289652058,0.4461624147291213,0.4856563576401989,0.9236073934534098,
 0.5199900429029896,0.3832971919612179,0.4590546030645761,0.2408325704361628,
 0.37475243436174455,0.5896099352341693,0.17209190091484838,
 0.7090619376362196,0.22318002268933423,0.6286804101852921,0.9115662159346982,
 0.5148338533093999,0.8601135750501753,0.16613395902843264,0.7124120616281232,
 0.8893296269276203,0.24525508724242484,0.31642638362911824,
 0.4128769437366202,0.7561491627736483,0.9840097355940792,0.6983657106736662,
 0.8353999580786967,0.5611054845518086,0.21824783817053084,0.853265107351306,
 0.511364765186911,0.008202267680980313,0.2552535625366287,
 0.003279450019465391,0.0795236516785971,0.42642216624585494,
 0.850943912791372,0.13846997203034728,0.19882120511161927,0.9793649761648427,
 0.17185006341189502,0.6325277961854857,0.7951276326456281,
 0.29495760540025573,0.9891873191807767,0.8210743617264715,0.8104631944623847,
 0.5161003958761623,0.16764985876704264,0.7167222810703154,
 0.34644579054198477,0.8695750599919585,0.5676338406220605,0.561656892116951,
 0.5912730635922907,0.26591923060118194,0.9393387543737589,
 0.19207740854475341,0.04181401199651458,0.6209092324885671,
 0.24886331373917414,0.3771030020497689,0.37869637752185425,
 0.5209817253069161,0.4043597069618169,0.39716575312064795,0.4255765651335196,
 0.49652127798069723,0.2814360378664664,0.9229920540731414,0.8865701474258026,
 0.1300624933645198,0.839125688176061,0.13230571440047012,0.18147381375656646,
 0.24038831042912034,0.41563060978949873,0.8003312796566954,0.668630980462578,
 0.4820116435311299,0.988972600213732,0.23412040216136687,0.2125987815684266,
 0.2563588067944841,0.11223220892878571,0.5224969772483996,
 0.21035773808143676,0.18832032232686247,0.8825341424662846,
 0.7111198932167226,0.44041134055744624,0.654958078882588,0.6300053654298042,
 0.2315280610115905,0.3700729807214965,0.13304174724522722,0.6782961118132048,
 0.7101109213011838,0.32917367341453163,0.5609053506662858,
 0.05441215480598682,0.47703805438722735,0.4052490637508428,
 0.2166238794794354,0.655791239786764,0.9353745837438646,0.003628470744226675,
 0.14586543794006857,0.4985491855760721,0.6588061122191485,0.7880006508405419,
 0.5824511392234315,0.22461228825501034,0.1709658286132616,0.6533819640719354,
 0.841831314083002,0.22333379114773244,0.39128958692514415,
 0.11434140491634193,0.3416159711061172,0.7820611679802981,0.5629355993509835,
 0.23272851798824934,0.35812728218290324,0.10282060566085338,
 0.3036329764165714,0.1871994064668009,0.07473672103439943,
 0.38705815577112446,0.7366443051284915,0.5058505937730043,
 0.07752991173184032,0.6415296171852325,0.6884772722924026,0.3041594226423917,
 0.6287702176507316,0.41026244015531776,0.3402825878348893,
 0.15354986902667378,0.19967830128077135,0.20350917498966214,
 0.2402185542369476,0.21651537161185075,0.24383464514286857,
 0.5046692173765958,0.5325897741686311,0.4691920904726259,0.7686210294146084,
 0.21965674690803727,0.27837219757921783,0.5999154069957816,
 0.5107547091345613,0.06966920006958821,0.4969950336596244,0.3160039075734139,
 0.4076037634948376,0.35973895236026376,0.7696511299891833,0.5318688389086225,
 0.999757473499771,0.9410280106789634,0.9792381952316881,0.17048117505963212,
 0.5073608102806825,0.9998974438792897,0.5043883381553363,0.07420780230613044,
 0.6375356307220797,0.6548393211023728,0.7919108182446551,0.7688541841717965,
 0.8087540603503807,0.9198438344741887,0.6990332779086187,0.6210592612248255,
 0.9624712379423194,0.3668017585580724,0.9468083293933373,0.7039433197380139,
 0.8214782832079633,0.873156746037828,0.27819072543327117,0.44864442823857476,
 0.22681561623657276,0.41116838695404123,0.4131313399561518,
 0.8867697784729973,0.9311456944565754,0.7716460249667797,0.7790941384763072,
 0.6641733321115706,0.3141500972517449,0.3325463301013891,0.3064147918482991,
 0.5566335080336254,0.2809537645145709,0.7885584236331286,0.66847491466343,
 0.12955410055667083,0.5355688814734361,0.7097221047482877,0.2787314932412889,
 0.563868379805792,0.6960821266481286,0.7209062952168139,0.6961457647578569,
 0.5867311457853714,0.27467940020321036,0.898420386741969,0.9506193961435275,
 0.9751571733597123,0.5208348923826552,0.5159907898206738,0.1758806330790239,
 0.7824955138210253,0.47737306225636855,0.8989612545788631,
 0.07549757903700305,0.06928705272959612,0.9887022649425942,
 0.061292342246943465,0.7590755439866698,0.47402809475060137,
 0.587327485482334,0.7301088418170483,0.06843939589749615,0.8664545540285705,
 0.8972295637663088,0.42147839829632894,0.2598068874845678,0.7117476646759704,
 0.8507604199939895,0.7429060274241023,0.6724634818956742,
 0.056178103052170636,0.02018482060141613,0.9799949541696863,
 0.5685097825722791,0.12317222056330857,0.021088016410858113,
 0.22640945929982947,0.9996679889906598,0.583271534645261,0.4042612265426918,
 0.9107917592798054,0.7417539406094463,0.4545845342114968,0.8251807752596233,
 0.4906143245471246,0.9930983173572728,0.7796829381097038,0.19453539327954794,
 0.2096726349742999,0.7828505667623142,0.1741888775550947,0.4493865423385617,
 0.4665498358399399,0.3602227269392274,0.993911292009335,0.7695373784803623,
 0.9927471389954879,0.9955462397152206,0.13759635374124568,
 0.10720101526764436,0.11410765943719481,0.24389343053835988,
 0.6710633465126561,0.8767612548490751,0.12117507634201352,0.5080612709616372,
 0.7362297031906877,0.27554432406374774,0.35307914490059833,
 0.8164132866743167,0.09684816454114753,0.5140465564895825,
 0.04048275012333302,0.6346762319866754,0.28605519825693193,
 0.6584980421264074,0.609490177820348,0.15818666865207087,0.9937411541258985,
 0.032850602109818494,0.9599104528409359,0.8623795968975415,
 0.7793906182884758,0.9571654176819998,0.051726686282769885,
 0.5168944634119698,0.0649703112177098,0.9576123338104167,0.6838095587864957,
 0.660534098582968,0.10989782503643386,0.8839371696833237,0.9629164382469309,
 0.4427270807397439,0.23284997715841538,0.0649611527738434,0.6824127840950133,
 0.12843358034327113,0.7563279034058034,0.8384566043347458,0.965274191933827,
 0.3214827373975021,0.8206558435697231,0.40213972604255144,
 0.09072906029043826,0.4687786640252962,0.6322362945493358,0.8485017509974118,
 0.7566332917215842,0.895949369281586,0.9064385548845568,0.6010985842834868,
 0.9549226608801276,0.47307700188331114,0.7822496191451583,0.5593232575846457,
 0.1897645064875444,0.7830400164511765,8.910741662158106e-4,
 0.5684522130457156,0.49297864679623515,0.12829590233646726,
 0.6523448027210459,0.5688677298775403,0.17784150694662637,0.9215558789017072,
 0.9038752751688742,0.27060351702049523,0.9090140956875108,0.9484751566836866,
 0.8691103210425292,0.10886527383445788,0.11536186703874352,
 0.8059792987399421,0.6053710631201733,0.04862301211156095,0.6035449765459245,
 0.3132322733341626,0.29102495739733814,0.9834831587958198,
 0.33383704085767174,0.44623001757809977,0.3687456814064676,
 0.0896534415480571,0.1053926394776924,0.44253348855564545,0.8488784016100029,
 0.9482724870439897,0.1653785778717204,0.6959284357070908,0.23771436762125442,
 0.5057468138071061,0.9571087201354368,0.21149715846482087,
 0.44296293406585696,0.5647537562197512,0.2561975397927547,
 0.024481650152449408,0.11315264663488778,0.6106520277055719,0.68804744648638,
 0.5853818349945803,0.5544234343639319,0.9997556778610058,0.3025057265248925,
 0.6385405853523135,0.685975033199909,0.25654246840632555,0.9831208064026131,
 0.7019158854705085,0.7456513011285771,0.9641079094079149,0.959364040501944,
 0.7848340404175815,0.006011812927107174,0.5207360883338011,
 0.022894425888092984,0.01933227816156169,0.5992344211097571,
 0.10251600903473879,0.11605879806676622,0.7944171464621474,
 0.6621609712424458,0.10513036876608395,0.8245300922667657,0.5593136504904714,
 0.895183636409187,0.4967922136850824,0.1434911368324021,0.8913793838935747,
 0.5328619866663048,0.593004942099884,0.8136643978073521,0.6682287964378475,
 0.830910631771322,0.7547109919711098,0.48919327770913223,0.4267189905091734,
 0.3749540174851641,0.42841320361351487,0.23115174889965973,
 0.2140377036436628,0.03765885754271492,0.7457127163073277,0.418783010785295,
 0.7784709705940904,0.866287269415963,0.3341345046627887,0.446194312644002,
 0.017940693418285636,0.447580200737819,0.11581051444141033,
 0.37978693512025086,0.6822110519216089,0.18816579186954763,
 0.9077691792676288,0.5752234661335932,0.9195943812969477,0.18046503612715048,
 0.7446931899488055,0.7068768602810859,0.36888250393122934,0.9030854422828428,
 0.8891195075967766,0.03963523741740271,0.44922566901722183,
 0.07242392122720176,0.24773049775871503,0.4201647976678037,
 0.05163346629486709,0.06408372067660639,0.2758019003666412,
 0.6448666901173759,0.7430078827670512,0.12292933193030842,
 0.42462164922711454,0.2227849541399325,0.059303083819492786,
 0.22956293382833248,0.5184706279214683,0.20082136506122428,
 0.0873722021468063,0.08210209446968997,0.9255689232080104,0.860764360994991,
 0.13746270466826105,0.5905174143278493,0.7162181562393243,0.5870477693291231,
 0.5000382550817057,0.8200157830316283,0.6706997742736156,0.07484330565961694,
 0.6912617939097191,0.21122889529391553,0.04239158831859746,
 0.7553842739744487,0.18578885713215332,0.641340719798188,0.333003814790377,
 0.5755577977082482,0.6540008735057453,0.20788355475353903,
 0.16528401101369927,0.6235727318433889,0.24687711295573944,
 0.4502155158453014,0.8710875065089119,0.9748653459148656,0.9912335194604986,
 0.007460258443116596,0.8850160668683349,0.4858828662640656,
 0.2848227334293728,0.4419695588154361,0.4100046869736502,0.3673771752607061,
 0.6772627203067738,0.21237779409892443,0.026135559630332428,
 0.7529160333050312,0.0765600595793332,0.42776119996922557,
 0.26146019264637976,0.1258376635241215,0.09049736262848285,
 0.7133309772902451,0.011740751970574115,0.3820880774895674,
 0.19274106099656896,0.6880747673619324,0.9704316208621465,0.4199546012562819,
 0.11497907050028933,0.07278644523704991,0.9937883942609858,
 0.5186875509232939,0.0791705329240151,0.37458916234082185,
 0.20917981157092347,0.3549432039060979,0.8462329897776212,
 0.020881966435466914,0.9422387600349903,0.6610539905628183,
 0.7488481585282958,0.27979993199780295,0.5713395725505893,0.9476701118940998,
 0.18788496487203254,0.37361600941996065,0.6630697847071116,
 0.6210719883961386,0.49490851538470704,0.8321409948563505,
 0.06037408512149545,0.7165128170097735,0.12413900400882982,
 0.7304113085222146,0.8500058134737107,0.8024438582301378,0.24513517047266042,
 0.007627343030846556,0.3988648275832404,0.5594206655613716,
 0.5558909205370698,0.5432642235628983,0.3199319503970176,0.46668870825814734,
 0.048030585329034814,0.541256023954862,0.5818881545779347,
 0.30711222557351026,0.5176684783950806,0.6871834614413649,0.5108772435175273,
 0.5264871393020124,0.6317111981189653,0.17361692495590264,0.716943150177669,
 0.8247184333724993,0.9611558602378119,0.2219842444558663,0.37276835246931905,
 0.26000415077877936,0.5820564316938122,0.21712516277686555,0.309599574899054,
 0.3345557463543848,0.36845790670901846,0.5482816837470871,0.3121652138742509,
 0.7152835790133708,0.6982032315447486,0.5473838279063705,0.6198475803288325,
 0.8364696293618115,0.13953666302354162,0.7378082367774992,0.7827581749428616,
 0.19867614342678785,0.1702011412613097,0.6683550042573627,0.6738771731647606,
 0.6514063723387891,0.41317484937614335,0.08969349536228388,
 0.7861072674183012,0.4433818427083568,0.011988096628317235,
 0.13967831231882177,0.37310724913299087,0.05833786111818884,
 0.8500345178200479,0.7411670204409877,0.04903190325805962,0.5707115447690808,
 0.15731601894494474,0.16249505710210466,0.5564201120300217,0.788762809593314,
 0.7468458051526394,0.04534278953603088,0.04633535710456782,
 0.18854114763660057,0.5504265593751091,0.578532915987689,0.39208573085371556,
 0.592404136936758,0.06546326647975642,0.46436873018303604,0.6255314974194405,
 0.018621952380223683,0.7059072613597298,0.6126377807555853,
 0.009160027955079753,0.8451745230247649,0.15603937092678555,
 0.8065242661381672,0.4668547040292671,0.7319774935392435,0.41995876739251636,
 0.4506213882782071,0.419312163551005,0.6260516975115389,0.3311813165355586,
 0.819879073354493,0.0979762969286051,0.0034305666361855724,
 0.04454016214560075,0.5224535021794385,0.17149229666229027,
 0.9091640008463543,0.3623261891668301,0.9463710865113601,0.22573501642389293,
 0.9917606334588618,0.765768901051883,0.8694962811581479,0.7925838446442429,
 0.0882650561388536,0.006127599086717783,0.3032718668269484,
 0.7981488456553977,0.28056519074098896,0.15834110424206393,
 0.8752549830664442,0.33267067499977776,0.13343981602250732,
 0.42221173828814973,0.2273101831901374,0.4093464908275195,0.2769780615456784,
 0.14294647334494892,0.8361324259567672,0.4548624271672391,0.3828710986780698,
 0.9133751988621303,0.7687402850254528,0.5021470552717744,0.9387854477852422,
 0.5424224310958586,0.7152691449634926,0.0448681896514026,0.5181022675214058,
 0.6778757154711884,0.8500564200610926,0.37918355050532715,
 0.34278152692241215,0.35786518618453733,0.4252455752530987,
 0.05471556869640071,0.4055831054996645,0.34571441703447037,
 0.018113664289221232,0.4675091977256288,0.021844719965205983,
 0.049998336965225265,0.5672585722286921,0.9247926517435705,0.994787095594671,
 0.7199526003199015,0.03924789224466818,0.44420660962408454,
 0.3186378349023832,0.9086499202534742,0.17699353862385048,0.1641967849472461,
 0.742653430575288,0.09927952712801491,0.6070324408381504,0.1751926009744511,
 0.42908254643868304,0.835428572310387,0.25212902637200085,0.9319604280470553,
 0.3815721755185435,0.21944410106065892,0.19504758662898136,
 0.8331013813103084,0.8238389871622909,0.41992222537923984,0.8173174686869877,
 0.8761821448734746,0.814114060676776,0.2656990017307196,0.2821142754435051,
 0.29125545653105755,0.05202792689953306,0.8068467031155333,
 0.5137676920637452,0.6459413870455613,0.30386206621534195,
 0.011622775737917146,0.7601124982391405,0.26439442205761443,
 0.5206252633227516,0.9251910707886863,0.6846998983032808,0.6409445266506841,
 0.2683939171676559,0.9854835274211605,0.4199876458014623,0.08182240072349056,
 0.44375064170158296,0.15011630309839308,0.10600437423606557,
 0.8412865565577223,0.14816603560151032,0.8407445653328391,0.8929714305582647,
 0.5001902821530254,0.6784700935445194,0.39743725940587193,
 0.18523810348017034,0.854104757725894,0.2209106210008569,0.10592462264457359,
 0.2355668586357309,0.8240455020713323,0.9606242472138371,0.25545786689963246,
 0.6531450691139564,0.3484268833937413,0.8694716206219217,0.5304278062457836,
 0.29414545280815263,0.11903921218409219,0.46608273330016825,
 0.6741695944647961,0.7004510944722785,0.8191000702053217,0.3451497148733583,
 0.5166101097859197,0.22108460544173125,0.48834053418524104,
 0.4429642943362192,0.1913027229020935,0.0742698240290689,
 0.006679016130012094,0.9525664675276403,0.38672147941740054,
 0.5754243175834008,0.619777893973283,0.436268758888655,0.9756108546232267,
 0.6224908362141213,0.7532606528601872,0.8473204077854398,0.16602466558988993,
 0.0031695423011399715,0.36317370444187824,0.40937499384475684,
 0.8336479157669349,0.03137580237316051,0.3637167523557089,0.8404466512212517,
 0.3492325395795213,0.33125674923862125,0.4649459543815989,0.862251996707734,
 0.7042100246246277,0.3665946954409074,0.4767370574452868,0.08035027646356074,
 0.8899831649327222,0.9087242818686535,0.47763280334578284,0.6693484341212224,
 0.8753735283188897,0.6464276616483406,0.7687497267659321,0.5456696332308635,
 0.6276347581943775,0.45449487051954407,0.7541614387473251,
 0.06844733457463159,0.12077959797883242,0.2943686184918598,
 0.6008569881386088,0.398224400689819,0.5583056006029405,0.9233753594762383,
 0.24674251036607953,0.5065324890002993,0.23005646300143545,
 0.6545221066732059,0.5923306162007471,0.020678375540309157,
 0.34248894990992196,0.5832973020211558,0.1380545661972299,0.5330085366996874,
 0.3864214409056076,0.4385990605166702,0.909163741022555,0.48609273804376174,
 0.02345560512280498,0.3071326492536922,0.4733255928695195,0.5853068094081302,
 0.7527056574008504,0.32255372578584074,0.7850281239334874,0.6795920714977906,
 0.4352900342813957,0.9672739168501057,0.14637551833335882,
 0.48019053139061363,0.8965992480571008,0.8897050906064856,
 0.43778626312043367,0.39529483335788007,0.6942965400117521,
 0.12371659275545854,0.47265101902054396,0.8045012120530255,
 0.7653714784186029,0.2579276543289496,0.059103356693673725,
 0.11858562874647538,0.5031272906221169,0.9155221746091858,0.6037496343935974,
 0.8400334243795514,0.2947342565637875,0.5023765681487912,0.8726252338751453,
 0.8974868994505137,0.3212418878973391,0.3927240449290228,0.18743251469514066,
 0.2964654668081457,0.8493912003538819,0.6678207373592059,0.3833791039177018,
 0.8528216183505087,0.6808471704522685,0.3590834241855916,0.3966779793648141,
 0.6619654804830924,0.3661866791724273,0.009465925255170426,
 0.1232962261973487,0.6375275197952601,0.6533549934565257,0.5258177188213001,
 0.43352941849036586,0.8623521485851979,0.16117310238980287,0.941628445540722,
 0.42656214699324724,0.4610045913944807,0.8713725176297126,0.3022041104682992,
 0.033064693835475834,0.010599224396142692,0.9224203681202477,
 0.23670816110920534,0.5672941091750469,0.1616591977655415,
 0.43365011988092816,0.5708036435990488,0.738781175382262,0.8003284408029813,
 0.013559403215877897,0.07752447060157441,0.8536833151900477,
 0.06820441342489647,0.9195781893433503,0.17682138366899014,
 0.8578416481713773,0.17129611274353884,0.8893439143738165,0.7946215319114113,
 0.7154621696792232,0.7582074247067102,0.645170643175391,0.2662573428189875,
 0.7792972367540943,0.5472256278139377,0.891074144812633,0.47348934606864757,
 0.7236963589794875,0.7263750982117538,0.2312603465185591,0.41785333922824874,
 0.28455381057914064,0.2252311716258616,0.009720147564582726,
 0.7693758466537844,0.6156629712813362,0.7498234233891979,0.39890633215539983,
 0.5355095645198521,0.2244114613251582,0.6697954274386939,0.4405480023182988,
 0.5491419322386774,0.20806731020364178,0.35658189505523497,
 0.6667915655992621,0.28212490531006684,0.656228540709672,0.12155877806496695,
 0.2585758432138666,0.9244992374871107,0.7868959936245609,0.8455090150362524,
 0.9529475133794163,0.27814920482033345,0.5384110183568134,
 0.0016773474352125284,0.3310531129569618,0.6991573522879055,
 0.20883938981215722,0.2795663613003928,0.12344916801765471,
 0.42423712393897917,0.49567124653182526,0.3567638609985867,
 0.13341772339303348,0.4005528863844705,0.3056388581174999,0.1546464635998841,
 0.562552479682376,0.27683246964680286,0.5873137470326091,0.3155238703636635,
 0.7438817975012929,0.9142173293549819,0.898881144785186,0.6469409766636142,
 0.699230226593137,0.4712222355851712,0.6358619847412448,0.26351127119577655,
 0.16451638389211376,0.12020502224378116,0.5173118759556354,
 0.8356239516781215,0.9194717093858091,0.6973649831571146,0.5926436653815578,
 0.23929428733767077,0.5205377179269476,0.8895984238309492,
 0.25282642886252893,0.8523631742786152,0.447092473542967,0.6985706561264066,
 0.2203724107350592,0.6586378093612058,0.9928457352569451,0.20596823138064058,
 0.845595567410145,0.479669243702077,0.7205356891267046,0.33913296779299285,
 0.9307429044504961,0.015646146339373668,0.08766648210982175,
 0.7227069812690052,0.39226915722597244,0.31057609395025,0.38895958524210295,
 0.48447781861835026,0.6152746435311673,0.3156720901618707,0.2970722141703369,
 0.5118193666594303,0.9070659716882191,0.2771513859890681,0.03284599231246421,
 0.0938048187443884,0.19946068791923577,0.4456211670073149,
 0.04278751950963111,0.6283502276221147,0.43668490265828175,0.341345249201369,
 0.11543211945254839,0.11508668955775658,0.032958482716436754,
 0.058949352220916684,0.5465367842116615,0.23142357517044787,
 0.027083397626276096,0.9009563795602796,0.8603189476352144,
 0.4561979676904624,0.13336404909733157,0.09851336473314953,
 0.1766890444512701,0.5413781941913783,0.25575370854666934,
 0.07282178455829724,0.49089159764929635,0.5239220107645013,
 0.9510395341599316,0.35651538545585637,0.7565093585398145,0.145480574387156,
 0.07178490340125254,0.5334275913439843,0.6353106476659819,0.5561316025101529,
 0.10949485870401798,0.6823342732080909,0.07078765068797266,
 0.9569070657054142,0.4047425310907812,0.4758715357470453,0.4401859033743951,
 0.2519060911390383,0.5012225732444363,0.35065617619903755,0.8678267807992506,
 0.4292048852580266,0.14362518004210656,0.04884919467831372,
 0.5171092297757715,0.14582082028766152,0.7327530200933159,
 0.05687946155504642,0.8463324572961133,0.2809445502360396,0.2048038611481764,
 0.0350914111325078,0.5140050116787018,0.3730300138086292,0.8083965701373924,
 0.04279375562045218,0.8987186440036669,0.5741885702175274,
 0.44884254969799164,0.8941934385524403,0.00794733915863044,
 0.015282666981825344,0.4862966497052382,0.624849234444333,0.3848667547192983,
 0.09160616515972675,0.37373425691824247,0.810363655696752,0.9606355500954984,
 0.5374926047836395,0.7148116353200004,0.6623305483229405,0.8498651456116392,
 0.45346533316824067,0.8826446974278238,0.8865088636760942,0.523653776648972,
 0.7148789377348405,0.5350057580513943,0.19153195380406718,0.2626358254168665,
 0.4210171585455572,0.020933421009023845,0.3971197708415328,
 0.2985171070464089,0.8131340066669213,0.04017523769394049,0.9181767214900826,
 0.7258006691298247,0.830912796969071,0.6939911756030916,0.28068280741276463,
 0.40510944284463335,0.2801228112000512,0.7249696275867619,0.1975817290308194,
 0.050165151528841156,0.7572288374659126,0.22688640421578155,
 0.06656057640844537,0.9177271396766751,0.044507297764462495,
 0.5565104358613596,0.4938484456792751,0.1780958789641417,0.9142093950487646,
 0.9020460054445821,0.889645654551718,0.22337731810516626,0.799024739629256,
 0.8767269397304522,0.2373444834369246,0.6050065790444485,0.6075757587369256,
 0.6096788123178439,0.6329090895779845,0.6565712340228781,0.8283362036590491,
 0.7431937843749585,0.32121034839119333,0.8266488233840562,0.8199145205605547,
 0.21393084024975595,0.3663561275201357,0.6685729474656491,0.7465778529109028,
 0.8215411625999204,0.9040049848535567,0.8281214662420519,0.35284433576158614,
 0.5314343150886629,0.3497816065078362,0.7052599677188638,0.9091253676446093,
 0.24457778902143235,0.6493150207188673,0.1703400644202835,0.9160995983338622,
 0.37524257942186845,0.322585562874097,0.27846496103264384,0.6004438294339158,
 0.3788099323060605,0.019646870927068738,0.0347703572793302,0.271409158292671,
 0.30860795142890096,0.39670534915084077,0.9140444204295597,0.412941483973142,
 0.1974761199885794,0.9901308121337558,0.1286167068829135,0.701650871162421,
 0.1507738556227436,0.4102476778619918,0.6069590003964201,0.4556568572318974,
 0.11118956997377827,0.285811844805712,0.03792158682322855,0.4806867953229764,
 0.6158407841300305,0.6564687719236155,0.36917244623312984,0.6365072561007659,
 0.6742777854222141,0.15915097170482095,0.9272082479418883,0.7051631657543207,
 0.3098030757332666,0.14793217136654635,0.42950236867191083,
 0.7633352499993463,0.09840960105224839,0.6480455137495564,
 0.44390447852942816,0.4675840681673652,0.44151861502892664,
 0.32166960470400596,0.628749574073654,0.9691495619674737,0.709975555821121,
 0.07942629381531019,0.08385902999264172,0.2039653322590228,
 0.5494486396438222,0.18922819421886838,0.7055577306972282,0.8141747407619268,
 0.2578949440960876,0.6053023268493201,0.12047582711556648,0.3379875661121978,
 0.41165680876933797,0.17103106788653422,0.8791150301839429,
 0.5210566404565469,0.5449622602712013,0.5051646897992831,0.13546100490311785,
 0.6697751808853135,0.027267703168276314,0.16665629742596655,
 0.8575165533592176,0.8720637305896998,0.21229068511990667,0.7892429228711079,
 0.7093374658875551,0.061329275838704844,0.8891270225511481,
 0.5918431748067664,0.7400022963119406,0.4916050535350971,0.065866965263637,
 0.8066847621510288,0.27515049057724117,0.36337261051333125,
 0.15252167668583594,0.22260498868152012,0.051610426851021796,
 0.35167329850775475,0.2036279224902442,0.05999732861157048,
 0.027503852827334363,0.5548791745885358,0.9732150274159759,
 0.7754220083618723,0.1360256922557278,0.3763244202186682,0.07460615698548623,
 0.30233619582390836,0.8957143208777059,0.9527901174360026,0.5813120734171398,
 0.5867521157264797,0.926410424992345,0.3021886014549986,0.31039564761201,
 0.49461193325813846,0.270971563241881,0.8425096413806116,
 0.046217991187689544,0.8364219858904466,0.7313153375700133,
 0.9382119873873718,0.7809384308585646,0.12421667502211542,
 0.20371877112627446,0.7695235415836248,0.9155316458942011,0.9241525826898148,
 0.6676162277581659,0.9315558823900063,0.7176809575183682,0.3735653970356765,
 0.0845159615861576,0.30877806622525006,0.758966106418991,0.3332888797515088,
 0.2174503793681759,0.30109277013236957,0.6693802302940032,0.4591411339580247,
 0.5974952149971813,0.1837708544053137,0.33833934119969633,0.3431767132009992,
 0.18231379583030372,0.45131648937948887,0.6347210100863894,
 0.6879627473022736,0.3789648887750803,0.9576111545095087,0.7710772254746214,
 0.1398443661684422,0.4024289465396159,0.8903085669382038,0.3701695079781223,
 0.07762168329312802,0.16248620866611796,0.6291326457301871,
 0.03349422080748099,0.8413266904978832,0.43898588589624477,
 0.4196972929601942,0.9533672769315931,0.6975030408761991,0.7150870855285174,
 0.9722381867593011,0.3501202963862379,0.3712587718267678,0.9701362601523549,
 0.7055769104575382,0.2749150026307037,0.16716626747514218,0.2273959425427564,
 0.004980480557635669,0.2166260263475447,0.7954594014801829,
 0.7775841962636982,0.7215596829848256,0.3977466051880707,0.7437831443992654,
 0.1013089216726375,0.9881770742180812,0.006254559299875173,
 0.9264880282517619,0.06202626473041584,0.20513378948108607,
 0.44671825587554503,0.024261476660745407,0.9373972331010935,
 0.772176698044872,0.9866842200090153,0.8582675270873831,0.06380387008160704,
 0.2583140310257076,0.7736114740214964,0.026826234695285778,
 0.7155124975580952,0.8218527401079684,0.5301396933350491,0.5267662502429694,
 0.745364478021807,0.45188760366174696,0.18533817057311475,
 0.024681191849199258,0.5106339621148218,0.28154830865664826,
 0.7160392016851393,0.6649891259085486,0.8138564587374968,0.6647730379514076,
 0.6466639845345072,0.31880528889123094,0.14592250211784275,
 0.13829114118491637,0.2468022647010789,0.3724672883550435,
 0.03821262651340063,0.3127118093893433,0.26853897749344724,0.531756412613943,
 0.4423671075372424,0.32213395792491506,0.916718108297676,0.03904464053335921,
 0.27101957595850934,0.168507504564812,0.9610477772652822,0.963848546712003,
 0.08051337941465697,0.8567421592375402,0.09922946867072313,
 0.24234384997304093,0.4089706887277308,0.6461987148434324,
 0.19667324786907248,0.6380630313306986,0.830426041226133,0.960318013017651,
 0.722572918862731,0.21624642599993182,0.16015060484741495,0.3004301815492141,
 0.2628818456291936,0.680907107563971,0.6773429233386431,0.732440848098306,
 0.5014349885465192,0.7879722759621213,0.8105120458909758,0.4156644817400209,
 0.2338130495355294,0.6403546531513066,0.361392683041583,0.8785518470173478,
 0.33221861579947765,0.15399530506059123,0.1189422929283922,
 0.36761504090440433,0.18452577645147306,0.4628845217221885,
 0.5579098026816138,0.6258478199548158,0.23093869080855567,0.6443568471835134,
 0.7954578139754493,0.2458874419319479,0.7876107806962661,0.41156071172014974,
 0.8315147726027473,0.5327511736338063,0.5105821823235006,0.41615245808592316,
 0.29392806490001333,0.7281934089991319,0.0558263207849391,0.3822077169113617,
 0.8507970188715994,0.20162543159380675,0.520521911951999,0.30387025651697674,
 0.5560820397120544,0.5639640122991071,0.5664451266571248,0.7381421710379105,
 0.025275225772472765,0.0929384143604598,0.4556319693715184,
 0.4125007572546744,0.31599367258852296,0.0538421034072567,0.0803092581545618,
 0.3703276231988588,0.4040868542462711,0.5609107952848988,0.03532762516547905,
 0.18407228393422226,0.799876703412308,0.5627496231909819,0.8302571142964964,
 0.8613654380943389,0.9521823884391158,0.2595974427873212,0.1289649220345408,
 0.6235078297701753,0.9618128133450154,0.7180635030945974,0.5721059177179186,
 0.7308897606917568,0.17262195597297902,0.6677703843693572,0.6893116361684822,
 0.7322650919692143,0.9867730504451837,0.1557223687068271,0.1733673494908513,
 0.6515953715729216,0.5672475594401846,0.733271233930505,0.5332624040401431,
 0.3422784275553411,0.32585419183454833,0.886552681914289,0.40732239204349985,
 0.072689928547778,0.23833032353841765,0.6298143217434955,0.07412903661519099,
 0.23409619374438329,0.11646742337644866,0.4415252176750658,
 0.8395308691715359,0.46150408408907806,0.6217481577399624,0.5683052014588285,
 0.3534393135803957,0.1431790342928173,0.007519290296449377,
 0.8067714875604826,0.0068099141352333525,0.46591117460960796,
 0.5472349727388495,0.03325493383695022,0.6244036251238103,0.1700291379696084,
 0.5793937465244803,0.1525160170833869,0.9705179341486677,0.08043746795332751,
 0.2870710940097093,0.7072429345686755,0.4149708560828623,0.9295482376618244,
 0.8086021842015891,0.7507289879979698,0.5511663894763448,0.28713381777992253,
 0.11297045748953616,0.9576417181862048,0.9151907524363145,0.912547106719294,
 0.22850755749784823,0.6369511003739072,0.33482238095888217,
 0.40739319527318085,0.8579006535273667,0.4246619053099884,0.4428911326917241,
 0.4281328583551106,0.7340048516862907,0.128066156318277,0.7097207873596746,
 0.11473618492869808,0.8843285675230445,0.14101445909333388,
 0.5328109232222998,0.16455395689122942,0.43410268319111434,
 0.6076906278833851,0.6712082597408867,0.6489490258305305,0.13482500499191363,
 0.04237366922540664,0.5139414752737949,0.6553519871315027,0.5516095344703176,
 0.9342664555368569,0.7356038976198799,0.841524341053227,0.2306949067555336,
 0.6906547153347217,0.6089351877183331,0.2950707717967763,0.8107437514031463,
 0.5910054410692712,0.07969331637256039,0.2359414439973957,
 0.03174823979741104,0.7930170189057243,0.4314582525774505,0.3361953789666873,
 0.30052160859240273,0.4632342286174299,0.27888111301816754,
 0.049228129330358916,0.7200358115764889,0.2318684168320636,
 0.5907058801376688,0.18908667446483274,0.5171449067577699,0.4057680100579466,
 0.3856007499234574,0.4906439747317872,0.5600013241688364,0.5201839645941604,
 0.8507575200678571,0.1487882451661423,0.9684657712404772,0.25578791824590597,
 0.6337520764917889,0.8226775150128682,0.772550168753793,0.1331435524620378,
 0.5013715463180426,0.06305106905399394,0.5622198219945779,0.5454832236686927,
 0.4139413588661164,0.8748375488806521,0.2174735484147391,0.7979293486784387,
 0.540351666338349,0.8883746515563927,0.16022246830517262,0.8199966543852255,
 0.873162822812938,0.9706370296399008,0.38806987925487924,0.5335838065738485,
 0.20570099048162083,0.08801302498494623,0.33476984125058895,
 0.9641564123210925,0.5985901954483899,0.2841904671842833,0.4687648939535184,
 0.8272281746484189,0.12146962470843159,0.10585968487942599,
 0.41174569680418094,0.009572061058893966,0.8644457151414182,
 0.12602853051806595,0.5265227471361653,0.9976816280161163,0.9976937930854903,
 0.27483225234865594,0.44758767025655766,0.8214747209235411,
 0.8486581115337087,0.9138628324477228,0.8351017753748395,0.27672932059061905,
 0.200176500832686,0.2627751723948214,0.08163618522342841,0.6050992537718949,
 0.5069095641208111,0.22894384592375372,0.07515861533945423,
 0.013654116557102136,0.33506780180829976,0.5316527164758337,
 0.13615375423222364,0.7601989441585345,0.14266136151827968,
 0.3931172189092601,0.5953634168061406,0.3571757079185117,0.2806085814832162,
 0.9070186472860678,0.38232812472250943,0.21594808900205464,
 0.39627970047033134,0.7555425981614481,0.7858877740091579,0.9444619291018608,
 0.033933461146545385,0.9228528453739839,0.2907412511416375,
 0.8501723718574911,0.9218138780497049,0.8906520483111229,0.07780643511840979,
 0.028192409115538886,0.6426265566295694,0.24264052933651525,
 0.250356314729014,0.05853498862878537,0.6243612030642706,0.7915751128779188,
 0.25929150491664044,0.19068254312208965,0.9293942556178015,
 0.09703680579168494,0.4469239281359516,0.30472842294026636,
 0.9097094710481121,0.7455964808502108,0.5151000346712985,0.4888265454118058,
 0.28475558243300114,0.6483741336938388,0.6582003306080789,
 0.10544324570314323,0.032474373935104994,0.423431226606108,
 0.6078961789591362,0.28618395473952196,0.5651066374383814,
 0.013542407728612704,0.3364152321353734,0.20547469891277903,
 0.7985366947916333,0.3366918304087205,0.2908651738039878,0.48217840714548466,
 0.27257746098987623,0.44954120952806054,0.6604693576687025,
 0.4265637150758126,0.5976496300838527,0.9257263280642422,
 0.029329276435433926,0.9963083681854157,0.21769683384637573,
 0.7415523246544976,0.709048004130051,0.6738726893791754,0.625106664927652,
 0.419830537254834,0.8441339018989038,0.612451869502374,0.4493632838452235,
 0.4130406695358426,0.6440359247854689,0.9430692249942566,0.5398362580284595,
 0.1892019255004742,0.35603711995972354,0.6913384979198219,
 0.03678741832695431,0.15446145233605124,5.939711298629247e-4,
 0.7106340623172221,0.10521220480312043,0.24291333059359288,
 0.9624110368734989,0.16115845969946085,0.8470328529716843,0.4532395271960492,
 0.7326422406283224,0.6816141354393755,0.6890818109046857,0.3825650339875126,
 0.08792533435130179,0.4858798629405843,0.41231363553228406,
 0.20969870303894544,0.12251966572442674,0.8141585742140304,
 0.45832892647358137,0.6640393214788016,0.6564110529738214,0.3215304744051187,
 0.06397546341235394,0.4325645782767715,0.08664026239058875,
 0.9805628419244851,0.9555358119032594,0.8934875848003352,0.3157266689073708,
 0.300497138419038,0.4694256586578476,0.6426061891335196,0.17693523833203673,
 0.36392006770601504,0.5764270466253507,0.39249503619989934,
 0.4542767114372521,0.693397662539742,0.5998901739644472,0.14605575115223368,
 0.4563253970133655,0.7553279515970042,0.7093076932453306,0.857411839841125,
 0.4747394711956051,0.605305959270375,0.21417569370652434,0.03777387428792789,
 0.012913168702737066,0.34570434029829056,0.28689234195265456,
 0.02438537207073499,0.7924995996387012,0.9934906004456334,
 0.17649920294617616,0.39470263155358953,0.38564343596397,0.06687669615620595,
 0.23452216340797227,0.9944668586769779,0.44688976578326234,
 0.11958972017534719,0.5661170826427265,0.33349313115111334,
 0.3049990348844488,0.5416152833584269,0.4330302509637145,0.9589345834362539,
 0.16982377409670824,0.9783973488245117,0.5113034562862186,0.6602887988843357,
 0.1447193018470765,0.7159213931084157,0.49854245938355124,0.4339443306265277,
 0.22874646441141355,0.4901959198930961,0.5217073095674424,0.9403230047424168,
 0.16900533773031456,0.964760372991411,0.5935379454701694,0.18950033115663478,
 0.1783961083003146,0.31725533242823245,0.17769536760522187,
 0.2588838200646242,0.862765758050235,0.6395972192562418,0.7163831026137012,
 0.9645976356897189,0.6440557445075622,0.5311485897269697,0.3485372644389868,
 0.2664989835034087,0.07775211468856136,0.4989688201128438,0.8799825428172666,
 0.4686474840108038,0.15412205505070786,0.06651891145621824,
 0.1010304483652883,0.31826025985024864,0.9860097437492864,
 0.17320244370103377,0.7326797025368511,0.6753225538251636,0.6064184900622674,
 0.12467830482008202,0.7585939018556872,0.7428182954572147,0.4156941955416086,
 0.20667573858822863,0.6725596788617612,0.22120309941200833,
 0.9540048422063477,0.510152856470506,0.029809676692348686,
 0.22225383246699149,0.09751639440489845,0.32881733482435005,
 0.6817528062624316,0.45188339905616415,0.21590882135635525,
 0.13798715924192162,0.34331382555816137,0.5725950947278997,
 0.43498398932161986,0.9858004987286717,0.4656216134007014,
 0.04975086689736663,0.2374084167731194,0.8606624770763543,0.6564643554602031,
 0.9978389269315487,0.9274297271524775,0.9146015708077839,0.5660586219030133,
 0.27192626045973156,0.43131425715010663,0.6864236675444477,
 0.17447861539591003,0.7649076634926781,0.9291014579838046,
 0.48113652282486075,0.6950691061511742,0.49191271053176,0.3622547206029605,
 0.8965313627566525,0.8413136436682697,0.2255571127475805,0.7510793837873122,
 0.8094441288490536,0.5048004195939081,0.9856055639777491,
 0.025009732102564408,0.11848951447510847,0.5647627360346333,
 0.08355687833408565,0.5951841006685553,0.02326586548363141,
 0.055404540785469525,0.4672094533543186,0.7975653641967805,
 0.2843172546101711,0.9897323048444459,0.2767347609328219,0.6159306393951427,
 0.3896651928856314,0.42372151155680604,0.22389219868218668,
 0.27341740576276463,0.7269844474574875,0.08149091387451879,
 0.8162118634343469,0.20654219945092422,0.9667543109654099,0.9007130559191631,
 0.44274285856351425,0.6199699024795859,0.47534645114123575,
 0.3427799241431947,0.15097730133451115,0.6711300603783242,0.7412334904632547,
 0.4207031946092574,0.3341243978324947,0.31908154402544076,0.6290044178328547,
 0.5686670641915669,0.5800580438978349,0.6845079589954567,0.3031022661539695,
 0.8498660152571601,0.48594358832190465,0.46746613969783446,
 0.6234267331740584,0.8027195777126757,0.651921846755779,0.06897011611030601,
 0.4607341014646851,0.6945030849349785,0.5382246608061603,0.9457867417757072,
 0.34480016452799245,0.16183864067769127,0.11949139739920689,
 0.7844983048290042,0.1254391047452671,0.6668967851418679,0.25644676594962346,
 0.8325708552796318,0.5704871714198116,0.03484958838473862,
 0.18133779421055418,0.9558521139825954,0.967300767015998,0.869574998505446,
 0.06467322051455959,0.04926955882148998,0.948755554589349,0.6961357414232201,
 0.3696432669294377,0.05395456646952379,0.4580772049224213,0.8709952294152793,
 0.49049786648258653,0.5009974582687947,0.3520053138106213,0.4950420722107135,
 0.46245417428483726,0.23256596244845584,0.16292115419784614,
 0.759923951578036,0.6291980467302679,0.9426555804032095,0.5216383361633629,
 0.8637542707545249,0.9389669088889431,0.9001096075839116,0.7021910783050174,
 0.056176553384703065,0.22717378848445202,0.009040264357122352,
 0.5306300383963445,0.8739378415145298,0.8241787175655135,0.27346973147649567,
 0.9145683095374054,0.5736969250711914,0.7773595306319722,0.36477893164884334,
 0.06021772528884839,0.15010081523096297,0.91130087270813,0.616283552323647,
 0.6400042996514731,0.15386736358783715,0.22275725597276064,
 0.7997868750743918,0.5176269952292302,0.6942822617652716,0.16469370289774554,
 0.768320922326126,0.8917731998684337,0.7343692669220154,0.29988820151820184,
 0.19379800092938126,0.46738799318326896,0.4560795048224595,0.866279787119464,
 0.8475030479138147,0.9019955803046256,0.093003242887015,0.32508425276985997,
 0.4700540144902975,0.5034798780491079,0.9521193593836029,0.6997314833442445,
 0.634684947351674,0.43337354647936865,0.4035664191045051,0.553499976675319,
 0.18599414304769257,0.1623486848643424,0.4525409155363658,
 0.24754060369203656,0.6262834750721736,0.6329486793209257,
 0.42372447493428433,0.3597512760565673,0.49929252751306075,
 0.26541268966548826,0.366389122182881,0.6636274458313195,0.5598093020364951,
 0.4486468689728378,0.09266649685548811,0.9480762416686586,0.50078736975088,
 0.11203054376495891,0.8532850104715218,0.3725999886251562,0.7648844501744599,
 0.5484986332407283,0.36891428650779,0.6298023273338443,0.7912617684035956,
 0.44347700449928396,0.42630886511642163,0.5452298733099409,
 0.8492229352886069,0.4538898505315807,0.35598628680183064,
 0.0016566089186869082,0.25223324617867826,0.19334852758536192,
 0.4420507920673219,0.08531397863073986,0.41051117411551297,
 0.29915707345337905,0.6033345632257381,0.41848460642149077,
 0.2910985030945462,0.17070200265948,0.4715384896484107,0.010686162675179744,
 0.09512514253773352,0.7992225853500933,0.9554541798527786,0.7459477481156805,
 0.6136039246304222,0.18542364111723897,0.5745815527169802,0.9326460508661618,
 0.2486883815361074,0.06424422903097005,0.8565238769089975,0.4271495289103153,
 0.606185328106457,0.6247157827549942,0.5624664069454618,0.028948892433992568,
 0.4892697240880164,0.6130774297957281,0.83474662787055,0.06672184867593223,
 0.3612426592423018,0.9443598993262916,0.013527871945727243,
 0.05446759369867338,0.1860543298677213,0.2369438378544817,
 0.11108415552559692,0.3503045000080649,0.2945117888160693,
 0.39069864879192806,0.11344139556974653,0.5766439072711435,
 0.5715629728347233,0.1585234906654739,0.8655689955855634,0.03595653988453096,
 0.09030184618215908,0.5196516429281784,0.4047117474343088,
 0.051525293169426445,0.774289050874687,0.7186625358867155,
 0.36017085827668816,0.700168307976655,0.22695773565977406,
 0.32834211381861533,0.565083270052403,0.6152659117681656,0.701342522396216,
 0.03274915475597251,0.07797538859236186,0.19371301001638574,
 0.8853631244398961,0.9895489690559053,0.765327589987458,0.043414380890305915,
 0.09112876185935548,0.6497024532063473,0.48596871841059974,
 0.5598612465749289,0.6957866984768036,0.5262740887925592,0.07219247602980972,
 0.20842572521904923,0.45847033701150364,0.3790459198732694,
 0.5739788542708613,0.6720279437001428,0.9063212161485423,0.9826773563734875,
 0.6938839362942884,0.42211964810879554,0.05485360661518757,
 0.019215853518792847,0.34975089840409235,0.8276730562144148,
 0.21172969290591714,0.9273955166808396,0.05769084972815708,
 0.8494042219636282,0.31307890633557267,0.7158018505257756,0.6138111699272585,
 0.9520857156725577,0.5794810338204309,0.3919554995641843,0.5443566005898555,
 0.10086773743806221,0.8647165824069566,0.39734782272343605,
 0.11975561600972173,0.8994172816327253,0.42343364465126765,
 0.1760030910947885,0.20442818177695932,0.9945901944500342,0.859308360656292,
 0.16113591443080255,0.9242389026100971,0.6378176917097185,0.5188808228254241,
 0.7399361605914074,0.5732923532278218,0.4735126238235636,0.7196619097107393,
 0.23105091270663014,0.8264911350481419,0.5553953270512009,0.3404636851015349,
 0.3817950327133265,0.3204679380772031,0.3692455426134731,0.994371058696939,
 0.8777648325045765,0.9252983246424559,0.32265284307763886,0.3402083379205826,
 0.47282433147575187,0.8704741186624424,0.33939983327761514,
 0.5061811424586922,0.0812102405875762,0.7460401034308974,0.3573321297167883,
 0.35230589366430176,0.7825528026308984,0.2602853216697085,0.6220023452270417,
 0.5129812729501103,0.8889143559426411,0.21931596808464404,
 0.37818710259995436,0.8936836596008968,0.12100227916701234,
 0.9890623311039403,0.6236752257146684,0.0876260167988856,0.9213884797278634,
 0.7645592653794223,0.7658173318344649,0.251111344832025,0.8085567709476791,
 0.906190732622428,0.19994988344316544,0.16874578891202519,0.8219211994126875,
 0.15570262208168495,0.928941685498451,0.9442226438511822,0.33783967078378874,
 0.0011199473637487634,0.15045418665140764,0.24412650789827617,
 0.8859837985809789,0.15821020242225847,0.2203022510186552,0.8302895032668229,
 0.4864773546026949,0.686053836731662,0.15041075130784654,0.5286989458784966,
 0.044191518773669225,0.33390040429073786,0.44345018303921857,
 0.29519650070074777,0.8735474050512778,0.87149844231043,0.7165690984557491,
 0.13019593880407276,0.8186692991682006,0.7392222628634337,0.8931216345042663,
 0.3212038112842698,0.730701851892508,0.7883393213256652,0.9810532002180457,
 0.9153643574311546,0.09199637022388885,0.8664143216828775,
 0.07892622241212455,0.1626850109709651,0.4847296531162163,
 0.05680605899801505,0.668141265460178,0.8991006488808749,0.29331612777394,
 0.4080162023188816,0.5312574543352055,0.1812111506461599,0.05644763544127407,
 0.6170796248142993,0.4390856534213765,0.49311695295069746,0.814009323359647,
 0.23015835799490947,0.7950941768256548,0.20487298183878655,
 0.20485590764790484,0.9025497608402738,0.7797578613074015,0.6348688444331124,
 0.6682885416211397,0.39738035975541974,0.377079382042658,0.7794325527585627,
 0.8658648585305443,0.9493199483632769,0.7638813604534452,0.38206849493295747,
 0.5236471319665116,0.3115406655015788,0.44430055402968827,0.6391852294869012,
 0.317625561694024,0.17245248680968606,0.20455757603253255,0.6169818904360489,
 0.4308563549754574,0.45685487194441976,0.15546112799661937,
 0.3547214246553232,0.5508235875644765,0.7355311620322175,0.9675581035587973,
 0.5028620908187545,0.8146190240245321,0.6673449369855471,0.09021564354418843,
 0.9602516471184699,0.2635711761071109,0.9452272304972527,0.94251009687072,
 0.108280110058666,0.1355425474847406,0.12981616251173322,0.08417241714895884,
 0.6854653333907417,0.45299610796208845,0.7336451337550137,
 0.25912820408435144,0.779095646776644,0.8921573376770771,0.7496226087311201,
 0.023358274604515894,0.5187009315335234,0.5329294657117183,
 0.17226891473143802,0.8205051629914438,0.9647112239379279,0.9017896148672657,
 0.3960779928503534,0.2813538816480603,0.8572686601291128,0.3444283063178575,
 0.6118413435429166,0.09974154613522357,0.11083821203321387,
 0.4080638724941228,0.44245786809307197,0.6847259585535039,
 0.29543928662655583,0.31838800101577003,0.6775676301590193,
 0.08332835432689167,0.8789774764691439,0.27058121875350616,
 0.8454869543902946,0.0999940444036781,0.9433345561865465,0.14474959582984492,
 0.11620804553914699,0.4071363317358868,0.6620887038249642,0.9441614812709221,
 0.13330141141281804,0.8370226727695389,0.051223186104390916,
 0.019706708493235303,0.8487587743954247,0.942091156738059,0.8639758474277045,
 0.3371737656403152,0.06437096424928346,0.09127765831834134,
 0.6581159629705965,0.03927887790338691,0.6498717912556096,
 0.014881319035005225,0.010195616489802495,0.5407608428447811,
 0.05066907406753507,0.7329306185929053,0.989716250614718,0.8140620725054557,
 0.9983777670177065,0.2852886294603536,0.29788121962992165,0.6635110911711488,
 0.6933398047175026,0.5168069348589518,0.05892167154507244,
 0.11105079060479506,0.8866188724629885,0.7883464142975615,0.4457924590590352,
 0.8923660671539403,0.5249063961274744,0.2005875407730997,0.2925984832799764,
 0.39285521901387854,0.8358174021965008,0.06475529615862463,
 0.8148142185555428,0.4042854302879829,0.727840982205723,0.3167202825743832,
 0.2379895886979666,0.7770544197308759,0.20979043650524232,
 0.10812984027764083,0.4090476452977665,0.5870739444614611,0.9640820420250419,
 0.61401577669907,0.41022614602454854,0.6459469150009681,0.8214486304424485,
 0.10500696631947215,0.4433491643987647,0.22555962623383297,
 0.46388106651879224,0.34374652067246425,0.47550720396806134,
 0.11224148466540285,0.7126269111556116,0.12442058885939966,
 0.8683055563843481,0.4937003765992719,0.7320382592328923,0.4215863121990798,
 0.7620938795699068,0.7240881855648518,0.803278109883225,0.9126755498527962,
 0.9731996428690631,0.828772967287645,0.3484079027756206,0.33781448039058837,
 0.3721437069969642,0.7700790252168979,0.7846056274799954,0.45770683066108586,
 0.6071285427371705,0.10738171126360352,0.9136892736263563,0.6643473432929076,
 0.8462449044505329,0.8161640600655429,0.12067366279832592,0.6919200295392023,
 0.9574284724379741,0.3163583601040928,0.31525259570651,0.555466487687726,
 0.459019896237039,0.29102245237329893,0.7841422678639544,0.28014273509943277,
 0.6016029121733909,0.07522548411172891,0.6559178131035751,
 0.19951883103071688,0.6084757156442053,0.09803698911689485,
 0.46437509953073486,0.6612951395679659,0.6542843785924032,0.3162421339725011,
 0.7907552896913113,0.6923355252084872,0.38908707636347906,0.7201190140349462,
 0.3083578106774135,0.2700877671131635,0.10037943353559131,
 0.40976765449090913,0.5424012874017844,0.5311997741297378,0.9010308423112817,
 0.18858355694371032,0.44342679077082636,0.1702370231064344,0.57249928255328,
 0.05231911927702004,0.7969333553512407,0.7919983058998521,
 0.20325939687025163,0.22085481617705205,0.028283258124424004,
 0.7561696820153209,0.8687426362209621,0.08205617149841304,0.1682352277329381,
 0.8383847939149347,0.17609871946484856,0.8861133156066227,
 0.43234551851527536,0.5920257397254999,0.18765633222589506,
 0.7583870732410573,0.9047730128242041,0.20084293478504034,
 0.49554513089567753,0.9665719882445903,0.3714381270726147,
 0.16687198612714782,0.729953919659599,0.5075965306220954,0.23030604742435712,
 0.8137296813376884,0.24950753817755755,0.3541934876280519,0.5783065700078085,
 0.8406559806331861,0.4736881626656696,0.7716275811866846,0.7101692349402715,
 0.3860734190367492,0.6161965174396008,0.5787469489488211,0.5876969551542893,
 0.6179749622712145,0.22398209895806476,0.12366177965576441,
 0.09556065903149602,0.05588696717566932,0.9034882276484312,
 0.5680985344593793,0.3994788796083941,0.32013291437844626,
 0.28292361300686486,0.9383947507056587,0.783750533417693,0.8923792309867056,
 0.49972990253479177,0.49213036343341077,0.43548706989505537,
 0.9111763383714219,0.91353599155832,0.10571953840775716,0.7501367881672658,
 0.8178841989760044,0.3265431528743783,0.7499018972828244,0.5784796245903201,
 0.24652934227454204,0.9593419893229659,0.5325338410688539,0.8344649373435462,
 0.11750994501094669,0.2560283273290138,0.3821698552869215,0.89041929390413,
 0.08191449684944285,0.8004912002652482,0.12777168919217508,
 0.12776745716891158,0.8660662170692424,0.1646389250272683,0.9324074808381718,
 0.8582707308187149,0.487364549812211,0.7554251744540937,0.2801772125438253,
 0.9768918901199833,0.2719033578439134,0.5532244616609528,0.1838751005550855,
 0.0473864173063433,0.726755100846533,0.7091557588459038,0.722462500333058,
 0.7607194183945697,0.6980196105286082,0.21441192183465574,0.9975792905558262,
 0.03561433442068496,0.702371468894118,0.6518157508905105,0.16402528210445932,
 0.29406459732596457,0.8152448502676343,0.4076473063187487,0.5336636678077782,
 0.5051595451610285,0.21643627894383644,0.4440395640129935,0.6640855299173016,
 0.4164697840114642,0.8381945729648126,0.5126062778392293,0.8960401565848841,
 0.8748540309809317,0.01045859615948952,0.9890502462694427,0.8018494038315497,
 0.7410168905812773,0.9447620295914367,0.6737334034926734,0.6918035044786793,
 0.062400642329190625,0.654787401318268,0.544493812096521,0.690765255120699,
 0.5724148520492301,0.04465877542746699,0.9424643007926775,
 0.36760976199950823,0.9351265482699984,0.5684939497754532,
 0.34161402414680575,0.9393774272251266,0.46802178676409845,
 0.6716896486274231,0.4617028844261988,0.3361448617459756,0.9081383158155054,
 0.8404850592642508,0.5009656059631888,0.3945182679705762,0.8691911952464348,
 0.24715135160544854,0.88938298299314,0.6563406534747523,0.16421834031246563,
 0.12586078107071785,0.6803084368467058,0.6067466492631415,0.1821325632357279,
 0.774153473701056,0.6147640492413632,0.6963728795566644,0.2507283481646516,
 0.8917777778832819,0.6449836962347874,0.4759554046856743,0.8486777194318962,
 0.4374556680678262,0.27132041193999323,0.8948960828534398,0.1606870737806687,
 0.016583352928350337,0.13921822741178302,0.48682695800947096,
 0.11592681200392863,0.7597177923541024,0.8427418416177439,0.5600284062290416,
 0.47874834878888084,0.14882216264403136,0.5661060446690646,
 0.8155729473124748,0.7750114681822264,0.33370905612977686,0.9878470428591806,
 0.060823125405365386,0.25355635725232095,0.005074646489919288,
 0.4707985453660851,0.10246094456595722,0.2441747588116825,0.6266351566832055,
 0.9830998074518877,0.3987067621082868,0.24071488589393364,0.910636581857421,
 0.8836375762926545,0.27286940620288525,0.3032094151110737,0.339100370839163,
 0.5139947360492482,0.340811881135342,0.07630538629730943,0.5705338797873802,
 0.9109797477490187,0.05985816748125117,0.7813042796412001,0.6987650577623534,
 0.27943737309190064,0.7270642342357305,0.22173883612812462,
 0.5217745074495768,0.9955799718908819,0.49844419100256643,
 0.33448816107702095,0.844173250809926,0.5709714822864898,0.2950760619094407,
 0.20574801070622772,0.8019898830598299,0.8152848680623401,0.7972365231792555,
 0.23536746890746818,0.03546673693786806,0.8295274247866035,
 0.36042564652759557,0.583103059737468,0.6463343264893815,0.25406845929131106,
 0.2849822719842112,0.8199622902964403,0.34552230140748597,0.617596445746111,
 0.9934072860776485,0.2215699562277228,0.9015136663987389,0.761171055597717,
 0.48540621949478935,0.6163738029393584,0.31725467094028603,
 0.3780878165537105,0.7749761279101478,0.2918047104377488,
 0.030783234697976614,0.08054906650656635,0.3509573857232514,
 0.02725537786005605,0.24864519917704708,0.9799543290122064,
 0.10721078017116237,0.3423274604819595,0.030330794347278256,
 0.2120544633926722,0.20074104848258434,0.011966002244219842,
 0.9340169407479719,0.02047390527120163,0.16618035495021566,
 0.3580118247022709,0.12471204033845729,0.9467801511931951,
 0.18508521037076364,0.42797364351030864,0.21662387914195513,
 0.24839156038342758,0.39333709878818834,0.5831443804281126,
 0.6774853254659816,0.9788563978129573,0.6829741113369052,0.1886166445290861,
 0.5876389876567754,0.964955159421331,0.8853298485840913,0.05955095915356379,
 0.1360976367761223,0.8251797638365916,0.4152718303508055,0.373346512432563,
 0.697613515283537,0.6014583128491093,0.9960810742224409,0.014641727882489963,
 0.4008728670573153,0.8408060122294342,0.6740903975009571,0.10212293705974707,
 0.7389531638570715,0.567749361562644,0.4515815738747496,0.5447033130226808,
 0.4353548541916312,0.5307559264473929,0.846881549964739,0.4381573431332566,
 0.9923632933182989,0.22918607477355613,0.7775127275995218,0.5048357439113333,
 0.29415653678577003,0.05181390390058127,0.7980952757020966,
 0.41849855733611063,0.2387049527796118,0.7748503432945011,0.6045951451186444,
 0.44216823100160063,0.9624425350772194,0.44668352369413045,
 0.25936188313306596,0.8022912451597747,0.26800726331823577,
 0.7094175072478045,0.7606327372663565,0.6477326703083027,0.5780531505771257,
 0.30931014771514,0.016304247106691316,0.5862737422204546,0.3538038102016504,
 0.1676804263307311,0.4407672413341619,0.7736622442048726,0.8913722684943198,
 0.249546981650138,0.06461746509700539,0.9153999055465702,0.9235262451582489,
 0.30461689660864777,0.9034826200430386,0.28355204791161903,
 0.14641383658573237,0.4824608312298392,0.09919077053090297,
 0.1825906555535961,0.42466919632522404,0.46253910644343155,
 0.8701857774013521,0.8261971458500111,0.8019373218528746,0.8613444000787303,
 0.7686267432841104,0.2104444455302682,0.9553218290229326,0.08236795013121789,
 0.952856017149235,0.2699821551168413,0.6944587999438976,0.29018028344065805,
 0.617707562089073,0.8734672670370718,0.5481274379616301,0.7861170062562098,
 0.07602957986418302,0.23797117666339562,0.7182813953599412,
 0.10327800764840145,0.5679666110346648,0.8897993570694568,0.0563115343744619,
 0.2779252017234436,0.6852147147717237,0.3249644870067724,
 0.023271421508649803,0.8170784273934382,0.5092090807020193,
 0.8448032480463046,0.40140432842241824,0.9420716267862893,0.9703093116580113,
 0.633830447401878,0.9362202895278877,0.8735071883505614,0.17852115964494586,
 0.5409018508870171,0.2995234727999324,0.23493836399704338,0.9548918440556097,
 0.11466628324278272,0.4528692605819651,0.5284840249937157,
 1.2910564093238097e-4,0.12315539221373306,0.3830083964097878,
 0.14007632950289806,0.9967850604291328,0.9623563373117282,
 0.02122695194468538,0.0325821570732332,0.2575660793991269,0.5044382566343667,
 0.5479764683076871,0.970831864187345,0.10699359352246252,0.8143183913665192,
 0.6745720483209867,0.4418881316851305,0.28976211436728794,
 0.17847649980017533,0.495923645938117,0.955460749294314,0.4070320698182548,
 0.14719865683878797,0.6450572684170247,0.9931554652377143,0.648833881544296,
 0.5895461116144498,0.30012653969210235,0.07602796699214875,
 0.5115665476973041,0.6479924253243535,0.34227280760042367,
 0.11484360300722041,0.7783664803552637,0.3824239380009853,0.5613785083795769,
 0.886208560639776,0.9092808401841513,0.5384058471145132,0.40029839493370456,
 0.17740552222879935,0.07956446243310189,0.023998244447066996,
 0.6572195358476219,0.14942925214170177,0.6080917256333974,0.7086428085171544,
 0.7986805746439564,0.4231190318005106,0.7228023825587756,0.3285658248779305,
 0.10435792298307234,0.0296963708668031,0.2540689264487709,0.5716558618417441,
 0.9051307404156832,0.11800626869665065,0.7571968987844568,0.7130690999853935,
 0.7275613243679216,0.5000110756468009,0.8106259985765312,0.9648398256734545,
 0.498546920343542,0.8134051301693758,0.1294445731887045,0.010796108750287026,
 0.033006942553382324,0.3955308637910263,0.8179894229292827,0.983035888048064,
 0.17321609366678326,0.1300766604445872,0.5152616531947045,0.7310521396071954,
 0.2403682638900606,0.013977975081016614,0.04456009540789929,
 0.48377378343316113,0.02741369294874152,0.7335012314155671,
 0.11449255373160094,0.6408410758358806,0.49677238830647963,
 0.27178684002761355,0.9465527056862668,0.36983146339903017,
 0.8244157712537824,0.6813431212601467,0.10694866662478886,0.7559277042102068,
 0.12518467926038135,0.9992643942543226,0.43909491973534354,
 0.2022471223759329,0.2176536484870889,0.6309792944214562,0.675218541467501,
 0.8294868036646744,0.7710423594395823,0.08820198336899743,0.3976019553261202,
 0.9165818636297327,0.5623552072251722,0.5482171988424315,0.9359668128364824,
 0.6910781699912287,0.023307759238927694,0.2782666867995397,0.943339637376001,
 0.38283710003097526,0.713857284359762,0.03841049564792176,0.9099779460461491,
 0.15390321025281062,0.04302789436965904,0.5281777944417381,
 0.22923382984009333,0.006950486482234175,0.7635753434167625,
 0.5804761727445026,0.08951287439581823,0.17895693703039006,
 0.3397283057251983,0.5108409386567034,0.7127344884216241,0.5870453301692461,
 0.25726037826102677,0.02870587600765473,0.17855886612047966,
 0.2296590374206079,0.8835670481101068,0.5888714678244433,0.1695126643818805,
 0.30995854746761964,0.7399850813200873,0.9288554909110498,
 0.16190957273495066,0.02288745458790864,0.30335190885706764,
 0.7198350501283461,0.6329633036562943,0.697315127501872,0.7820949384057578,
 0.1846525914800674,0.032451802688040154,0.25231229215482953,
 0.14184248082913098,0.9645777519250587,0.738580530489998,0.7479054849563838,
 0.5801403911280676,0.03320145825947518,0.9596465273041429,0.98036845408268,
 0.004829209402290924,0.8366803288928405,0.6473894064542522,
 0.8658404283759422,0.25128443260088584,0.8178651281655659,0.7493591723874924,
 0.8105274890004361,0.21815188775774785,0.9318282557808908,0.9656369406258298,
 0.42862762032227186,0.23534885754267698,0.25052971953462544,
 0.33074293747643135,0.8545005723386532,0.23622708768566358,
 0.9757789166451976,0.11401916024346681,0.2880563984546368,
 0.23624159831024782,0.2247438107228913,0.5126031416682516,
 0.00468216394689791,0.5621256591376329,0.7763005587804749,0.6310959290471849,
 0.43453704304485097,0.9627559796537063,0.8656981132031539,0.5936484423823623,
 0.2680781937994621,0.7949173578082004,0.23726451344071264,0.9892674100018041,
 0.5897148340082695,0.5198827101591066,0.8060403620583616,0.4128514373083052,
 0.7887015830970461,0.5985901579421448,0.22659982005770707,0.7638854387426584,
 0.0586701354560899,0.8996668394312286,0.7993809152814356,0.9916460975341879,
 0.4556330062531341,0.8274866344626239,0.28414642942828605,0.5958910865121391,
 0.12408884656593078,0.18922633416644652,0.5261658639835627,0.268304400526572,
 0.813102691188877,0.10170597205811349,0.4031759948638818,0.954621287705732,
 0.029048037253457393,0.4044332325071319,0.32329556116876645,
 0.5433008930723423,0.38547866839050815,0.06100531388185759,0.410163401934866,
 0.6547588349872934,0.23930273801801583,0.10247647618102962,
 0.13316371285187845,0.8840375756000867,0.5135041401082757,0.3972193981687534,
 0.5567733467731778,0.5950712456448852,0.19092268235350818,0.9018215024339566,
 0.306383820836428,0.2303714684572581,0.7453491970286483,0.6009898451316181,
 0.9491984492468435,0.7257994458375361,0.1454097068099126,0.7719909904736589,
 0.7655079025268374,0.5751137790068224,0.5285615813308712,0.9781975687266281,
 0.39019031221978917,0.7603617921317881,0.3916193444202426,0.8616417560657734,
 0.4852860626179636,0.454551024634803,0.08034886350314685,0.07238494755856573,
 0.990690316397612,0.6625341858046268,0.45997736986205273,0.663624142911449,
 0.30135226152211514,0.9171203182702405,0.9415472584881556,
 0.12251751658897692,0.9692725341192454,0.3370542633857565,0.9739955845480743,
 0.6280642839469243,0.8153377298090668,0.7552520611120873,0.6526168519802986,
 0.32219092647383374,0.21498232572036047,0.5657940883354771,
 0.24736467057424705,0.12406267282277073,0.3939171492085186,0.514447543330184,
 0.4589455014265149,0.20760158620460223,0.7667450664280875,0.8270327751570132,
 0.8718773475081985,0.032408549404835285,0.07928184804611971,
 0.04340741214911081,0.9415748191589463,0.5083867293896209,0.863622664695122,
 0.025928575058289383,0.7099742552684009,0.1562436665454311,
 0.6648586045889215,0.34777558163980105,0.28289712938956235,
 0.41258563813456406,0.9372956759057902,0.6181122222333041,0.8242672450816807,
 0.289215888260438,0.8329369543808886,0.8378860607713121,0.10319312071693532,
 0.8639432514447791,0.5164178929424471,0.40488546641396883,
 0.00167969533404988,0.11717532364518568,0.389637964371904,0.8879770090616442,
 0.725052166287102,0.32471817710269546,0.6299752708544076,0.3620304976322176,
 0.6312038000533282,0.16002301806508112,0.7076474603805046,0.9912783403389532,
 0.9230527686045814,0.686428554409054,0.7562246440526504,0.3208467275746514,
 0.7036023821272747,0.7957948929351555,0.9903374116824585,0.5847384912501072,
 0.14904578480066355,0.6768206859140795,0.8292165399809447,
 0.10081891354585026,0.968831944801981,0.8319741151308736,0.718325240443216,
 0.21903597726374902,0.7588937905423043,0.7308277557528685,0.5110655809358924,
 0.9115875125906201,0.7483769066649273,0.2796867495536932,0.09799452842679413,
 0.01083831858391815,0.030222815284591364,0.8512837261442172,
 0.768993010722071,0.683750211087171,0.2730494831967848,0.7704355718817263,
 0.815451301495901,0.9372650829180426,0.757374834139409,0.08037401290089574,
 0.7937620501093635,0.4587831522493868,0.12910690100455124,0.035585116480776,
 0.27920581474241957,0.06525001419203347,0.3004759781031219,
 0.9817834374740895,0.4039925201699305,0.26398356025300307,
 0.28954057474877115,0.7470305121652697,0.553570515334241,0.5204033245086397,
 0.5517529608136329,0.23585832719675737,0.2270927971544161,0.6062955831331914,
 0.8737089011596544,0.8221172437698554,0.5869103982489927,0.299089601211907,
 0.2172230847140506,0.2517739704483719,0.4432191503541938,0.3952684075138856,
 0.5170774421560894,0.722980362360394,0.7578308649467721,0.2528197253480894,
 0.8177898183600435,0.752042114343937,0.7059886298438489,0.932424396264436,
 0.019924743738601314,0.22654760521693285,0.3689600232604868,
 0.669134631845438,0.28057053017009204,0.9339312789726907,0.5591061227506086,
 0.7406018348140727,0.5803166954216873,0.7453436954108903,
 0.010315468620564872,0.650924229467544,0.41374067759347855,
 0.41509724407396076,0.4725600674591749,0.7025982141188432,0.9365374352520345,
 0.9899582808276484,0.2120185858778626,0.5437942651816645,0.5011638095315563,
 0.37861698791236176,0.29875958436333994,0.7850793101199891,
 0.9280150592520722,0.5115985829337689,0.9251471450920289,0.7602945734683859,
 0.7524367185116447,0.8116337044203561,0.2721928083540328,0.6657891545800003,
 0.05264655598645207,0.9285740222951062,0.8846731383943094,0.9770228022631561,
 0.8800731673141461,0.8123316384589071,0.8145104267415214,0.6085723992480422,
 0.7920667854238463,0.1455330347688344,0.011623272360706594,
 0.02395187548840627,0.01312638805870614,0.9419123629390297,0.312483322537745,
 0.3494111908647496,0.8223167372413139,0.11954336153791734,
 0.30962092994933277,0.7732143931149261,0.58884494794237,0.2176096043746012,
 0.13628506081453273,0.9360801092267248,0.23368702760462956,
 0.9673061681574078,0.9254328901244182,0.9598637392515156,0.891304737092939,
 0.20225787014859087,0.5464514487564844,0.758243910469745,0.09134441091279055,
 0.7971324926211926,0.5648895891636029,0.18863986607299565,0.9237870411665129,
 0.9080430796636684,0.6909339113627894,0.2795519103760602,0.17957452906989535,
 0.38400404769946106,0.6249714735557421,0.591333987287488,0.8040159237678508,
 0.90524768286511,0.1785641103932406,0.19806831026784566,0.4841130312336305,
 0.5698464866355876,0.4734518211091965,0.2659373977004842,0.7415543155851797,
 0.09307065428367534,0.18285312656106634,0.5299766228536908,
 0.007113082409486049,0.07926056156438632,0.025039637830276518,
 0.8045315955810945,0.6422024480556549,0.03785358559407481,0.731235154846573,
 0.5687569848380031,0.931241806107409,0.5866135540277189,0.40835165833793974,
 0.6316809854627853,0.26584001997489914,0.6478608021999346,0.3651314912185044,
 0.8679939993119382,0.1289219861022024,0.49408950186526157,0.15172383453645,
 0.5525419326963722,0.8712746144029418,0.704879771608967,0.06419271121502534,
 0.6231722324005622,0.523402338520868,0.5178250241273883,0.05211570817545952,
 0.9298737019770633,0.3501409663180778,0.6027351277353785,0.4146044061202281,
 0.23818920131347276,0.01361165007210674,0.6035111312910376,
 0.47984231143735145,0.05190437022144345,0.6684152856929301,
 0.3499209119374962,0.48936447759333745,0.08205750967624548,
 0.5164891230638312,0.9935927361156521,0.348155123130501,0.036113420845228505,
 0.49443355017300716,0.35157584887068694,0.5048998388359156,
 0.7598275186627351,0.01957432654096447,0.7296057258307788,0.5478263298301886,
 0.6441503332812063,0.017894653533854443,0.7217163705436067,
 0.8081582997788677,0.08063199767378204,0.9411918349223668,0.564978623151283,
 0.9228938370248593,0.9238546885393653,0.9129555549132397,0.17922947162993952,
 0.6300804723545729,0.0720716860636813,0.4785195780872399,0.837006769498755,
 0.37660755174891625,0.5230340848155481,0.8658303764196564,
 0.30765578247675407,0.6889580592271367,0.5227950166808676,0.6258610990247757,
 0.7204723977158018,0.5836973737297401,0.0796150474723557,0.9254274950057211,
 0.9889319105297599,0.8805153567419308,0.05162301601149144,0.8950465857687332,
 0.303544013131989,0.6529230743300147,0.6759480402103505,0.5987724762443372,
 0.9395985542036229,0.92131341869843,0.6381851377259147,0.36572237984404854,
 0.32324811050584856,0.43205760989286024,0.27123293237179624,
 0.7271129300235982,0.8047843159363537,0.011353105617020631,
 0.9854092601765793,0.08038669139582788,0.5340577939792734,0.587248137861668,
 0.8616112873978714,0.05472403869882947,0.630134186579122,0.5969404576104645,
 0.44087165439607967,0.3145374591081298,0.9637858304915103,0.9333592908589585,
 0.4335909993254812,0.1688564180183294,0.18443379634007429,0.4502130422499814,
 0.8219603298052728,0.32291293297826096,0.8705849311553516,0.7777902598738666,
 0.5489228620257798,0.7788878819101114,0.8312386188259685,0.34631146630022647,
 0.27378439936252663,0.1991724053102485,0.42286175242969537,
 0.6578910990322918,0.19666157461483125,0.33929571514322665,
 0.7233816342327772,0.9809516922724103,0.37092715517886676,
 0.37437223413526977,0.0527286419809011,0.9415135365621135,0.3463300130243595,
 0.3951492964500516,0.8792361719215644,0.9323037800939084,0.7901729704820998,
 0.8867322296469222,0.39589174838670205,0.17413077481327754,
 0.19856650206840099,0.13903294475543881,0.62592002072896,0.2626663855610295,
 0.4131823348475159,0.5126642712021142,0.8302204955133009,0.3351140610638037,
 0.30785962802845823,0.4558735641079257,0.8165283694861962,
 0.10630234123612137,0.44137483260782284,0.9059043403406156,
 0.1833557762555924,0.602791229425882,0.4455879349028915,0.8747174497513744,
 0.46576391059428124,0.8494218557449853,0.7165784928877493,0.7255466044610426,
 0.4036130005831209,0.01558491466886569,0.5286224633722235,0.7780651122349697,
 0.36816310874190494,0.47065534565781497,0.8792465381315174,
 0.3910911107737385,0.09412134109230019,0.18498074951226717,
 0.9839201438458746,0.830489614702048,0.19143885586951492,0.9418294148980787,
 0.40894781736447505,0.916279498246967,0.4543185916392102,0.7700630114236323,
 0.33734863557204586,0.0899334637262611,0.2911545373969622,0.3650928979166872,
 0.1337277003041013,0.22213039661796408,0.6663780975070739,0.5614635580171231,
 0.017259920614842494,0.2468285369167844,0.23759157847448953,
 0.5802394238315556,0.6163062252775345,0.03941369715083998,0.7930163484166372,
 0.22637164485237327,0.23336563636485275,0.1516363326038792,
 0.4481704940474527,0.7981787904576045,0.24093500099175358,
 0.14742441365813364,0.2439368373156735,0.05396713198530201,
 0.6206294343884265,0.7548308914689763,0.06071981403169624,0.6133652915847185,
 0.4623378082623797,0.7977685455211687,0.6267364927948328,0.5938648627887788,
 0.16680509783619335,0.567796218570762,0.13909989671826917,0.5358026994718779,
 0.6108047460752539,0.6557858835664195,0.3170022880871852,0.49233401480816563,
 0.8906814496179205,0.7699231159551236,0.37063656843009507,
 0.28004074734668505,0.9712208152622798,0.9747297931941297,0.9323144144965346,
 0.6301854876686567,0.6311502149799901,0.3113280311488459,0.09264678479710031,
 0.018520435771246624,0.26282123881051156,0.727321813324541,0.731525583621411,
 0.17683794729604607,0.4528492251220767,0.5549187148376451,
 0.31186254756952625,0.24310123180763887,0.7039917756157505,
 0.6731874808258478,0.9006385626328022,0.999704818426242,0.5189072084428559,
 0.6263247903810233,0.13101991791330703,0.32112125950822035,
 0.18983123566174842,0.6412546627493357,0.913747950793467,0.8901679608396661,
 0.8449968445575418,0.033583610761033555,0.7800697340832297,
 0.6679010641602258,0.8520374437495692,0.15126977103267714,0.2811824422807153,
 0.5005979687210028,0.20507049619117024,0.13368855154895898,
 0.6074891369769612,0.39822754496356605,0.32448152998351854,
 0.09993516596406482,0.019719266602583052,0.34987187087038174,
 0.20564779554603407,0.8335352885497439,0.1378149211208406,0.5853575801471174,
 0.38708112688637286,0.5319202392633597,0.7442830431406634,0.7330443861410734,
 0.9084436521588388,0.7900558249522386,0.6790732865424876,0.45788756772392913,
 0.5006290694247779,0.6238498307572952,0.42997726539922176,
 0.44324471354104233,0.5272336344492139,0.3193719709129289,0.9588106196418309,
 0.9067309970135728,0.7933554071799352,0.003783406654229493,
 0.9890017128171916,0.7837743092339847,0.07273849600734483,0.7154119183137874,
 0.33447550582658825,0.6135266265853653,0.9633091268649647,0.5037327550682585,
 0.31294953479222154,0.3375276355649982,0.5512412339178997,0.8472430072919849,
 0.611922169882945,0.9603054459742791,0.03776503567082601,0.9539118245829313,
 0.8463008023016018,0.7670297610890412,0.473594428429553,0.2487412410417965,
 0.13939419560158617,0.7665336208792545,0.7091441118358803,
 0.18572563240157125,0.8830934449919567,0.40903218626359594,
 0.3276012751913444,0.9659822350129359,0.6003088918392462,0.17459374285703033,
 0.5975331410001107,0.5483577594975197,0.9817717965692804,0.6977030803617902,
 0.34567808812027023,0.996321892940445,0.47974633773035247,0.4206732914172383,
 0.6420596503820135,0.9571819473099988,0.12339293428174047,0.3035142955307124,
 0.5138263626279748,0.4910761187897156,0.9609533701077809,0.5191085435952048,
 0.4582113887429957,0.30765332724994043,0.7308364273389992,0.9610429284527426,
 0.3236058970253315,0.8547918036258888,0.7190208508242827,0.5748593247261472,
 0.8381634501389994,0.894797015051945,0.7045107356413429,0.10857283932616801,
 0.48694207192314864,0.5359768651183013,0.42093333515954345,
 0.8126763140598211,0.815157374831346,0.4612393378132882,0.9861102300593112,
 0.5659501596403682,0.596164111064801,0.9800785163545316,0.9421977779430204,
 0.10907670793997515,0.19010309398090097,0.30955708016622374,
 0.43404999795365196,0.2204368346095199,0.7916946331591304,
 0.29562127362281443,0.02297988504043047,0.7593774910374476,
 0.7947896484619225,0.5568728133447947,0.6310007504356663,0.4168042884913117,
 0.4874642854046478,0.19870269763236637,0.05022195273458652,
 0.44292930693378074,0.1407390681162789,0.5372149041189122,0.6726200656571858,
 0.8603615792457265,0.5366934863576427,0.06451274287505404,
 0.34836337879810453,0.605445320540313,0.333288861966324,0.22401226651685957,
 0.3097744091399448,0.4788460565183805,0.545124215236504,0.2456367083778751,
 0.07540127346955305,0.4653984853918316,0.7500597960377948,0.8000931669062827,
 0.6785301516357508,0.8345538206107224,0.27604572963360785,
 0.16483076026055854,0.007716383352636025,0.6307572508089759,
 0.9672376723410737,0.8526547083007888,0.05334786469808439,0.144485359165882,
 0.5101252542083096,0.5240659935951818,0.6714247113645901,0.3624315853677694,
 0.8200636330248128,0.2366713275104152,0.4071308024665544,0.6416232881188922,
 0.20601923295857083,0.07724006359653968,0.07836293848159681,
 0.010385137703455838,0.2457099320669155,0.4791606794173946,
 0.07335332582339393,0.7722786474241606,0.7882508932697374,
 0.36446339822065554,0.10209315871661284,0.7353509634368018,
 0.6610689426610975,0.03950103505121494,0.7676890941165229,0.7855427457297991,
 0.7478528032876498,0.2994958822981757,0.45669802497681644,0.2819913555788165,
 0.28824598265756196,0.43241892005234195,0.11549106890334093,
 0.20930107976245438,0.8888631223834342,0.4823837505399652,
 0.18596351209542883,0.3347535114849065,0.5237734189431414,0.0582779784772024,
 0.9708769149667381,0.8376838326691477,0.6467018316488045,0.3778682356190276,
 0.20663891513247035,0.20982459957432043,0.6552632033175492,0.984672634608836,
 0.4608384457047423,0.7079071181412436,0.8513041997058057,0.18110724462212968,
 0.6966842360843741,0.7350046694576601,0.607088022588244,0.8718604401196025,
 0.19065215517206846,0.36705953824022497,0.669557215062196,0.922321463590936,
 0.07572830294243738,0.8964271831351157,0.5173701354272885,0.5272651941507722,
 0.13071597524020884,0.25281986998112205,0.12240812456076788,
 0.00446389036902628,0.2739149749668691,0.5737045998226432,0.4363812742467891,
 0.2470701854138102,0.04744323134911088,5.179304658343709e-4,
 0.6618923119566701,0.27934519994481966,0.006176874691369516,
 0.5802351369134423,0.2034027043633163,0.6503288014283874,0.7061003723734254,
 0.7245528377886773,0.26809355080054387,0.9206136893248504,0.69412101151845,
 0.7565350028835747,0.1432174126302007,0.9882646647515633,0.6429528462318267,
 0.42063909324932314,0.7094199979975733,0.008920398144048658,0.36167684858981,
 0.2453667727823725,0.4133834129385576,0.06363813951102659,0.5579588482264584,
 0.5704087487841011,0.2386334298484697,0.38225009706809454,
 0.17696505176679578,0.4916419903646827,0.03993729407241631,
 0.9021496353017697,0.41084563986029266,0.08736251668312556,
 0.3949313125353996,0.6737001542230848,0.13441282366796126,0.8593192749061189,
 0.7187377452683188,0.7161973697064311,0.3329650795998258,0.19188114237480025,
 0.5950757492070194,0.37355507448477177,0.29367022622622785,
 0.49664803611605457,0.8076378053396792,0.46344109500997166,
 0.7189453437511103,0.2238113198271522,0.35356383761289,0.10769327655193295,
 0.43962355293616184,0.754069862957797,0.028960553559298607,
 0.6996067567377136,0.3668721014784383,0.2097531133952032,0.9716941287744447,
 0.6195324862682112,0.832537314973444,0.3738587273097591,0.803797052620716,
 0.9715420439675974,0.8589882567947476,0.49236476932551987,0.5058871737697961,
 0.02304863807768498,0.3684752016129913,0.5365399927219566,0.6212885065898779,
 0.6054292388263348,0.6522228952900667,0.8094108318041003,0.290072832641855,
 0.16347964457343778,0.44002169699443605,0.06617705210388813,
 0.6899096027761332,0.4358860343461557,0.5853741692754963,0.3986380336782329,
 0.5663855537630147,0.7700607126154659,0.5185893863460951,0.7319627024472922,
 0.04368717328274263,2.3598804256497186e-4,0.8613458726475648,
 0.22447914096241361,0.11827535238933973,0.7823910883992473,0.774125927656954,
 0.22056350249744805,0.7612908146319017,0.9719495953916275,0.7203339114963894,
 0.2441427581582889,0.9308990971872546,0.004747312975132867,
 0.08726335047912992,0.28835413097229556,0.08568958442012187,
 0.8645434279319805,0.4112893132090085,0.8111091424547321,0.6766954287771996,
 0.26221590567206854,0.2850520784124707,0.09369580043107029,
 0.9950292307576935,0.36257816703708257,0.9898552941011104,0.821161852386869,
 0.6346742580569615,0.311569167854687,0.24162647428270811,0.20974326197688176,
 0.5581397912003437,0.13460916483671326,0.01527673435369703,
 0.6453388266992289,0.6394440011871372,0.8470851414520739,0.08868307642924911,
 0.22326620248101503,0.5747984136240508,0.41829361073081806,
 0.49937850846840504,0.587250437724652,0.049927261422975655,
 0.015196262712647624,0.8426114322762324,0.23351399265304895,
 0.44471764829314664,0.8201212931720541,0.537805526705415,0.9714509859241222,
 0.4699356077495973,0.22395088682325093,0.182119650734495,0.16930181453595272,
 0.5212908880582804,0.9995357870141237,0.23785148553813973,
 0.17143254900570715,0.34825322036379247,0.6477872594234428,
 0.5487919544992863,0.02520814006185801,0.4781179633685574,0.3454787806055539,
 8.588261826992794e-4,0.24376401360847422,0.5128766656395448,
 0.33280495325901294,0.32904546487025077,0.08957739636160944,
 0.22756122838900172,6.231358628530081e-4,0.3718756345197973,
 0.35465855879954555,0.5573816981669065,0.653794976669649,0.45798243200851096,
 0.4808082522058291,0.20357800633295753,0.35256851331406047,
 0.2904901191616718,0.44600019956656844,0.819939676195618,0.06046660295331463,
 0.9812016573055312,0.7100824316024641,0.08427395390537695,0.2582524681196561,
 0.38820714679515433,0.8358635001755516,0.17156813090570333,
 0.6174324976087394,0.3453779291886103,0.46546194442928424,0.6251009301746301,
 0.8474244678288576,0.7246558621770396,0.4182240338593811,0.878589666769119,
 0.7625150602749888,0.11294103168967795,0.8213019924203793,0.9475797849258494,
 0.38219950515915346,0.8593441437166922,0.6614961098694956,0.5727909347778164,
 0.08873451570234836,0.1883198883764352,0.005537100961746533,
 0.7955145542548848,0.39290335834805157,0.5288643868354287,0.819486755574013,
 0.18345409132892254,0.040406932305217524,0.6063789623107353,
 0.6074724019305032,0.32495331371152136,0.42693439328522587,
 0.6196861899229245,0.8513311914872377,0.46856521031590537,0.0798823579921657,
 0.5861042200146801,0.7761773571933087,0.7650721274199347,0.901947341415565,
 0.4780317748801566,0.9052444272376128,0.32397276272494346,0.7330826572516619,
 0.035250355163321756,0.9148820526614804,0.8788060335916106,
 0.36229724363476623,0.2647327825518253,0.5075799286239161,0.77151107286504,
 0.42042167988855594,0.46464533559293697,0.831409889546005,
 0.45019324356557366,0.9988133084987287,0.1664463855116043,0.2853428594562726,
 0.11028789288821716,0.4324093005975561,0.03914720244053804,
 0.5229349047342513,0.3651484985627651,0.2058149553854216,0.27990056350311365,
 0.19866686731005945,0.3134797403120071,0.14202252694124706,
 0.9317815119235263,0.5995105539728125,0.022636787304656902,
 0.7053780455129529,0.6987655142101757,0.3129291908817623,0.45183657677794176,
 0.3710134494741669,0.2546717926920492,0.8712267863454679,0.23080682908777006,
 0.17425135723380214,0.35851143807032293,0.8443957326046233,
 0.6865373483487711,0.7538419504947175,0.9442334793150339,0.9665189009902981,
 0.8344984746703394,0.4364095059643396,0.5107788509697808,0.55392147187959,
 0.5234319508904595,0.5427785424894835,0.10207998219891012,0.9373008034791731,
 0.2796337341489499,0.807886752244771,0.6058075076249266,0.8906793961295824,
 0.81922056306751,0.8914149429802759,0.839400375101522,0.40835092958745145,
 0.3493359558080885,0.4370004001169031,0.431637960377129,0.2639300518841392,
 0.6070466713269282,0.23914848617333861,0.004704036962979052,
 0.5131444737173545,0.03446416606071934,0.8051456252914532,0.4279450177076254,
 0.252434161140231,0.12877812909921238,0.8050557720092646,0.17959360175274686,
 0.43714493153100786,0.24699049242837268,0.8046923868752692,
 0.6232670166051935,0.6499491864764937,0.4712673433818576,0.9230752146126975,
 0.45426772619269395,0.2865557587080608,0.5144903747175433,0.1603810581909496,
 0.6653219234432841,0.7784613729315693,0.09930407827938526,
 0.36780448768501206,0.5072632180399009,0.4761296926192625,
 0.43231104889878613,0.8458970613020641,0.04570188130835162,
 0.6453979560790899,0.026186691938415607,0.028885548300756292,
 0.4121510105218339,0.4382760365883528,0.8761158233560379,0.5177786038242677,
 0.16738369313126578,0.26673239412618677,0.2641982668897779,
 0.9604589024581985,0.2151218966747097,0.7742939958084517,0.8341203205189366,
 0.9026394318305475,0.9579995760688157,0.4861923256823608,0.2309832676242224,
 0.35389810033565516,0.25759524513150345,0.694547720880483,0.5270867948515553,
 0.8931215471679127,0.9655231005068813,0.9298683343916785,0.12613956032799145,
 0.48118323600854807,0.904408175169688,0.4788838505645886,0.8076970521267567,
 0.7684784166565074,0.7075682270721101,0.06385031414947495,0.8974431632110704,
 0.8726724686594873,0.3751256650045687,0.35968777280379205,
 0.03487402114104288,0.501287708610211,0.036377367520376014,
 0.9838198279747381,0.28934476934432474,0.1895103187138073,
 0.23894980596079762,0.7722236114385735,0.9767490828183909,
 0.04257457904227979,0.09972675213473592,0.6105849847838749,
 0.37295395400658615,0.6420364338808371,0.8007444535308968,0.7600873045454578,
 0.09291639677490615,0.18243200145151284,0.02086844208458516,
 0.8022055896850311,0.8156880824436168,0.04607857998516385,0.7052947640925524,
 0.5411035987237454,0.8950553857958458,0.8208436521757155,0.1370223980194708,
 0.9344846824335137,0.721813889522532,0.5408472840539809,0.9247265836055981,
 0.9079492822098534,0.5937679246660161,0.01279645707432353,
 0.07742236337524355,0.26710241285599656,0.005826514470206656,
 0.5761846771582794,0.5662451785374777,0.4990119914277188,0.9627818284900049,
 0.8000553630562226,0.755328858021008,0.5875700965080701,0.6219784582561885,
 0.7696124536887075,0.22481716632320614,0.027324520935723795,
 0.9483712356041788,0.4420335750501112,0.639600974991793,0.13800909528711824,
 0.3991907815708552,0.918912723347778,0.6215348997761028,0.7688464484605865,
 0.6896733726836265,0.45181533561326603,0.4817237418914564,0.7573017834410494,
 0.2107350840599953,0.6256086132939441,0.5711383798889049,0.9515325567645114,
 0.3217278254674061,0.017642473034950923,0.8005688949895136,
 0.4997782496575397,0.6563445244050237,0.730006355626843,0.46222098135926626,
 0.5164869544391738,0.3916456926245585,0.5523951458678116,
 0.005138531207249164,0.21402666310422425,0.030114254577257804,
 0.6605843971083355,0.5574958802092403,0.03072504533906162,0.664974100429357,
 0.43379447483890066,0.808713633498861,0.9205801274234903,0.1165027910071228,
 0.7971733082958022,0.12460662853916515,0.49578823027488017,
 0.6023004200184969,0.22902471453594775,0.3145985412457275,
 0.19586410234834584,0.526332502349086,0.6116557436928522,0.6644192336281004,
 0.6379364921473364,0.9541482934903952,0.8208836526714781,0.4605261699512875,
 0.4237271547929369,0.15850357360440592,0.6390132844164482,
 0.17725610886119525,0.23297809258517665,0.8690187820180607,
 0.0653088984084691,0.5679861649933418,0.02235417052166222,
 0.47686255133026956,0.07873251765928369,0.3073623442100768,
 0.6804870788361351,0.8459157686768071,0.518886539961762,0.9439762808103459,
 0.6265565869659978,0.5326281057140767,0.001692284932771293,
 0.1873226148174496,0.39254443550149487,0.48016037171906956,
 0.2210969351558809,0.6389740139457621,0.5284694234824018,0.4823076108284252,
 0.8093275518115827,0.6983758475926476,0.20894170566999637,
 0.34636215216775423,0.8921570534512828,0.5470947998584271,0.6002142660887919,
 0.12044814815634375,0.9971724666690114,0.05238704741405009,
 0.3278284361866901,0.4815455736315831,0.08499962525024829,
 0.09580661062282148,0.9873195550206547,0.7766129404415874,
 0.0076804894873513385,0.5001828687359703,0.2706156197415972,
 0.025175915873094312,0.023191295294954672,0.12454127077690769,
 0.7059195160514465,0.7973516637418205,0.5179472544977866,0.1368438732528715,
 0.9848501430610475,0.8117297720317679,0.5552189082155274,0.8405484104031871,
 0.7541708780851927,0.01692181768913026,0.595596322709832,0.7005348952047493,
 0.32726797770294547,0.7322518928057747,0.24518669995994335,
 0.3116065995730375,0.8773137958905614,0.39861864505844513,0.5492195263216366,
 0.24441471460825837,0.40995112394164845,0.15487998462691088,
 0.029405971757364258,0.16936160641904785,0.2444406588472734,
 0.7759267319395661,0.37503590687214516,0.7089911593330155,0.6782003821489957,
 0.029556100140785357,0.817618633449191,0.008804560530686567,
 0.7186819931519759,0.49722842655827315,0.8306901889317128,
 0.04479648895861155,0.546324224252275,0.08574521734087748,0.6128315360903132,
 0.8438061463899424,0.05016163439739196,0.23907359588214971,
 0.16034767111096115,0.7784523336949697,0.27742530624224626,
 0.031088760280623795,0.3079449012894324,0.8732007907810093,
 0.4392131780703734,0.987298111464476,0.8365280611264172,0.15702538199962757,
 0.45622231239876454,0.24622066084811134,0.26527869860498177,
 0.6953659158126326,0.8313555498154863,0.957639375882994,0.011313495853800815,
 0.5243285266183955,0.4639409070179109,0.5618570489474026,0.3672838225419679,
 0.06646720047076071,0.3039594459637809,0.8384438723051314,0.725867427211938,
 0.11540953160326173,0.6520999276132449,0.0032829398086694495,
 0.3944205455305855,0.46685731739396197,0.3212186901211588,0.6856872328092085,
 0.31543936559199837,0.841005219061927,0.22788740871040103,
 0.16684248900457654,0.6595776907000847,0.8888088921651818,0.7657606908275241,
 0.32359947131973654,0.036806049110615024,0.36469115812021047,
 0.9432500577276104,0.6718776679972625,0.526050196966557,0.2623795354563542,
 0.04104376544991073,0.30507155781921835,0.9024296473109205,
 0.5716435195515245,0.6906981864562027,0.27208252758151974,0.8967869709207938,
 0.048265991980985024,0.4087662820379053,0.14417120663049898,
 0.9154597364666401,0.3073445626490696,0.47143249534817155,0.7365365098075592,
 0.4496812452026855,0.7100988091620032,0.806080162116039,0.18525645970604654,
 0.8989493634809392,0.29371469207883727,0.1456933069754922,
 0.002068699890601766,0.9097400727044596,0.3671146682685823,
 0.10575299121842263,0.8818036181743903,0.38279755890016154,
 0.05763385382529573,0.5850777822490749,0.8919820122950676,
 0.48401599269466666,0.9151993195879962,0.635757567976154,0.6864457052370609,
 0.518027607483158,0.6713860285174182,0.6061531208889283,0.2765916165780924,
 0.5360406638944747,0.4152057415536412,0.8071088147431728,0.36040682223510057,
 0.48223907309695757,0.8717885015499849,0.4294903105557467,0.540877124169493,
 0.7448838537277591,0.5287387760899886,0.2423574253973363,
 0.010389404639914113,0.3185612582381253,0.45120390425966284,
 0.6714249626543107,0.33228114353473004,0.3336355017293451,0.3941427049989038,
 0.02886957738804452,0.9318721635117394,0.43980671920238157,
 0.6671397612996871,0.43350384540340725,0.9882907042210082,0.2135861127002695,
 0.25541317818876463,0.7961484415001994,0.7954395991331218,0.2802033527863531,
 0.5914553864029364,0.24334298769603036,0.36357927795364564,
 0.1585737705153264,0.5175540673585598,0.021573929473538533,
 0.5211092821447998,0.7385309820791097,0.577516947364988,0.7596302383209171,
 0.7693935885981703,0.5814899145245026,0.7907354007292564,0.6329104034847086,
 0.7951891345456059,0.3544547569693415,0.8031850229737695,0.5719286553315815,
 0.07306957837331962,0.6652008217907706,0.6563906174846991,
 0.05709979872934545,0.3877958265989273,0.2601295418758045,0.609837788614274,
 0.989824388480548,0.5849424402899196,0.3244112454572846,0.23778817919649542,
 0.24831030121323483,0.9194499502663596,0.23247237991003644,
 0.009006510487562736,0.8993090273853099,0.7295786798820558,
 0.1682134756416679,0.839019111684046,0.28553756291820986,0.28598674533332813,
 0.5468840316940983,0.8359825349117478,0.7283523267187318,0.01474743141116075,
 0.12838945440024674,0.9208532383505685,0.0010309562536066075,
 0.46767368243213214,0.29306576346313273,0.5161110044911061,
 0.26620487374654445,0.7406855600610389,0.7474912872234732,0.5978038894723527,
 0.9584377350263964,0.402968283681694,0.8375586420755252,0.4494141011249695,
 0.13992474103952135,0.3401273103532405,0.5307078270416538,0.6573817279291407,
 0.49791848539847017,0.819127209960633,0.4208865368558147,0.04539093092560731,
 0.645054872971636,0.7305788647665902,0.0460771826515759,0.4831251755827066,
 0.2678849607372261,0.9667803668802736,0.7796827362273033,0.4746230535685282,
 0.22561416031402348,0.9725042761628424,0.3810147756941995,0.8704068202145678,
 0.9092533807771984,0.33607150960098464,0.48628923210003183,
 0.09544091228715623,0.3080205166743123,0.029645320068111825,
 0.4363108190920705,0.4172429621983882,0.9361903972936246,0.9969002705521802,
 0.8219561642232435,0.11314448058415605,0.015224660696701475,
 0.6717176256955394,0.04984449835804139,0.9428908540652697,0.2407313224283083,
 0.40026039033267324,0.6017434657554823,0.5684064991964879,
 0.19523525385714302,0.7163341021986044,0.7771710348579084,
 0.22963021741948197,0.8283320054950092,0.49362618512922896,
 0.10477924174513153,0.17340054202427624,0.4925199001495485,
 0.5316289665591141,0.01856691154719281,0.09714391923232979,
 0.5318730662314246,0.7126896090706565,0.6210331863517138,0.1168038431164018,
 0.47346389777068143,0.07725835073144238,0.4993967652251371,
 0.8131182766398791,0.8526737816097907,0.23888308448782303,
 0.44284304120259554,0.5646256608221188,0.8734629737965862,0.7693310288236833,
 0.43774997044128794,0.4806237240001001,0.31929567156408034,
 0.06211647754121752,0.18375090939622873,0.3533434559818771,
 0.3309437587986418,0.3175318125773514,0.6392205465912874,0.8343951758379394,
 0.15895224006551778,8.852862884501933e-4,0.2852554227508367,
 0.041845374913148325,0.3086365397423949,0.476710336477008,0.6953361264487679,
 0.6483307437773781,0.23364916438555583,0.9734697840406901,0.627376258817699,
 0.6217405093202566,0.8626342540256038,0.12014189408701237,0.5955239901969952,
 0.22429989858777366,0.5722851078152751,0.37688962992672226,0.46760781564444,
 0.674134074554148,0.2577447440189369,0.13148669786682787,0.09402271623653358,
 0.2707346633468398,0.8497668979169402,0.4604321304082307,0.09598760353871083,
 0.2888240422430103,0.7910302231390259,0.5918266631451016,0.4444362565543909,
 0.16829811906427117,0.04579833051734117,0.06958980719231755,
 0.335278416307784,0.9170583511783912,0.3411807430199817,0.052370673286770586,
 0.9458775851574129,0.8357461306980288,0.689579402975302,0.5991536663127931,
 0.8517618760516074,0.6078706157501571,0.6690353447248382,0.5409599982571078,
 0.11506003044746449,0.8014391749063791,0.10736936370082972,
 0.30515486090457566,0.4154060322753481,0.8362440214120193,0.6150240517599302,
 0.21247875837007146,0.9458524232578894,0.8423359529506842,
 0.18080435633979142,0.707087705117871,0.47534346362175306,0.6196330370289446,
 0.0318526515637273,0.2818046867923578,0.28360985226004387,0.803525996302494,
 0.014669175459576134,0.4361105885263322,0.8030424022431137,
 0.5548828294781453,0.5794023719417601,0.48808681460304126,0.5669665813742799,
 0.007666329290947016,0.48383544380411114,0.7835584634384738,
 0.9547955451838097,0.05475547431764238,0.3833020443044881,0.6546328173656538,
 0.5130353493673406,0.1088486024213402,0.42761606675887487,0.6027133325217668,
 0.23389733045023042,0.34149238315077124,0.8069573173615219,
 0.3528749306197738,0.3264796476959161,0.8300220235807947,0.2805712084762646,
 0.6203941174516391,0.15574717659043436,0.9748906336507378,
 0.10997461745654313,0.029852838459917486,0.2348519096574243,
 0.6656272965840009,0.848415562820384,0.587395707808988,0.6211119155967535,
 0.8742473650130995,0.35409764339348115,0.5941054925033631,0.6906380234197229,
 0.2836655799069391,0.22319895521003974,0.7398133230593814,0.8384737147782512,
 0.36719050841743917,0.7100616058183629,0.8686174486142391,
 0.37847699275048263,0.8884560738816077,0.313739914690624,0.9701524005256519,
 0.9382568335011328,0.5540899916524459,0.1982183496152805,0.29368753719640295,
 0.4246274420310755,0.7079819154699409,0.39602928402207616,0.8382425178841468,
 0.29358374674289744,0.8467983656552149,0.10291417961774418,
 0.8183553805393288,0.21494295777295425,0.9130912228773732,
 0.22284463747402683,0.3652967511586831,0.18490587223995547,
 0.8803756400767192,0.18244696780431435,0.782758315492356,0.4436998505109382,
 0.734898993161275,0.07004291348709268,0.6629683209957479,0.3504288228991468,
 0.890329107208432,0.12665451668830086,0.4859495727375063,0.12740880966006884,
 0.2584751175511295,0.4578697614551007,0.1253959273451446,0.7373853300881739,
 0.11034073757036811,0.0350104215473952,0.3866189632146234,0.8101737058882875,
 0.39765737539737445,0.0891373503564642,0.13623103957214477,
 0.7938959077638301,0.4017488929623684,0.9419561329827615,0.9517891536094583,
 0.6583250975507599,0.05301985779179652,0.6199905906443584,0.5209061508521502,
 0.17595212192913934,0.6326161476503074,0.09964770158834213,
 0.6280889576109958,0.749256959607452,0.03174928584823933,0.9056136014577486,
 0.09444268617083684,0.5060481803136222,0.5271977942535218,0.9944138548916526,
 0.44703514682829193,0.358998977249137,0.5933919858845852,0.22362780310840336,
 0.8024590540168127,0.20232501022744986,0.7173328642121167,0.8600532059751458,
 0.4121866553051441,0.7231219708417832,0.01218553791607524,0.2961646563652338,
 0.22818006619564546,0.3278235835215825,0.4340903718386814,0.267930850519012,
 0.5636477614871037,0.5846663776285705,0.5388301704956592,0.9452071584775041,
 0.972680723753678,0.1576259360690634,0.7635360702594564,0.5010557290945064,
 0.38198234552577615,0.9380265913351331,0.8771486601247307,0.7946744885123336,
 0.10057187950525448,0.3173308161409849,0.41298096525457706,
 0.10942104138519237,0.12586294871776949,0.8127177102333969,
 0.00280495076741627,0.35031276264359146,0.9688692749100069,
 0.7383684785566722,0.059467437680876145,0.7622782066241661,
 0.25364392792873447,0.7059729200007325,0.3815778679406555,0.6012945663233671,
 0.8904134449408924,0.33552907151239897,0.6728982551189951,
 0.05624712771185014,0.3405361662002765,0.5731354552268376,0.772927930748438,
 0.3139601077430505,0.3123299044981802,0.46594084065940033,0.4614733112635311,
 0.2018239223834868,0.5153404115791113,0.5052629846221948,0.8573660582825577,
 0.7672867729247839,0.3986692153460132,0.8252010673875662,0.971874968905535,
 0.10279609485888297,0.5127981406440228,0.08894740794034295,
 0.34184446891837117,0.2316044390474673,0.3087977185224825,0.5396532002980446,
 0.6806979576844638,0.5989563410365238,0.2803608534185783,0.25832803772143464,
 0.01820772651023539,0.6437792340139037,0.3915438484873387,0.6311267069464315,
 0.4749859713064878,0.6714928688535083,0.4202662600268282,0.9134286467571416,
 0.9126301264687426,0.7771937063320978,0.07853419676397422,0.6919276279051029,
 0.12987633874170235,0.010716713056319604,0.7826160442972697,
 0.905168803990711,0.667407770859509,0.07520778996268884,0.4493813265760628,
 0.5017452819080296,0.9271210537632217,0.18385744049494002,
 0.08531761999323928,0.8847790511599745,0.737512636717997,0.5026344599846821,
 0.18227893294581032,0.1290533299915546,0.11385466103159958,
 0.7475257471739005,0.5024560032290601,0.8542169314320491,0.7487794093893235,
 0.18460916315072284,0.566450551701028,0.4741845612502138,0.4707615723024322,
 0.9355628829573573,0.3553340017940505,0.10400723868431505,0.5208400528875468,
 0.3539444487201522,0.8276116722124518,0.24318068876969667,0.8634381101786406,
 0.04978841456614114,0.12521709864097186,0.8972106023172407,
 0.04152329522748155,0.0297446679832426,0.0941494128927608,0.9988859299818261,
 0.09125664858111604,0.16256844085499222,0.7981138207440793,
 0.6136873727416516,0.7704758621029498,0.21545792164215138,
 0.38474488789758154,0.6785663794447141,0.5039835674489785,0.5128513857661717,
 0.7960608866004221,0.6391250640792746,0.8068877154524196,0.07673851701829892,
 0.6818832835165437,0.02885626478460246,0.20140692835215213,
 0.8827670648095225,0.3784735386110841,0.8135685494978864,0.7093450489358007,
 0.7679785538626848,0.45462427923221393,0.024663866583324756,
 0.6441571570922039,0.30868299026642565,0.6055934545351195,
 0.13942708894416567,0.4102839292697835,0.6462866818362014,0.9369233491779012,
 0.08597645633212836,0.9786973698998045,0.04981825388837813,
 0.6341779292545201,0.92791544980657,0.4297704018504873,0.4787061348173848,
 0.7475434210121208,0.8214836981031877,0.06561023209901296,0.6497341436822877,
 0.22388233269374813,0.3230810526148359,0.34767528118622615,
 0.7772058852359434,0.37282164194534917,0.633323422158495,0.5120054743976895,
 0.3185475781375944,0.4901420689423045,0.4601368803085506,0.13086298529075813,
 0.14447408464488443,0.5683543713834568,0.797979365357036,0.28985881979177996,
 0.20152798528949323,0.47539652196061133,0.8107158289525355,
 0.16449806037548076,0.37362489539241106,0.6404305790489865,
 0.07478285105878146,0.5871998863858522,0.6782932851094547,0.93239184240196,
 0.8088620944520928,0.08674911157075604,0.36194190245475355,
 0.4658163519281052,0.908437969364411,0.3552973337207713,0.8351005334345474,
 0.7338486969201325,0.7365909158522777,0.9083962933300856,0.09612535033857494,
 0.3028284002625281,0.2193463006225871,0.6893671691900743,0.8419533878723178,
 0.9455365340058618,0.08854093526195128,0.965996561214833,0.05617786577850237,
 0.11330354426329337,0.10516437890600883,0.660349081133792,0.8175988183256369,
 0.4985078875050958,0.3335062685464547,0.1176737623192422,0.5370329744930064,
 0.7798513011613313,0.8155009885539682,0.5301334143059914,0.4468101042380378,
 0.44088073646198533,0.4786896824055866,0.2678412183946144,0.5534409853669845,
 0.854797319151209,0.17746842801554596,0.014735696078287619,
 0.05537247803459211,0.032219784102599514,0.36443239374197867,
 0.5473573835431724,0.4946170864757986,0.20683934507203072,0.8894027282747781,
 0.9038926488839181,0.8713746716066985,0.2783169371018528,0.7246623699927255,
 0.9103023815790152,0.17994288902061362,0.6525034526908654,0.6369604215934734,
 0.9751969027515978,0.06296668102870395,0.3772966808774708,0.3057389052029852,
 0.8171048500012382,0.6987766659631404,0.150873699063795,0.468039360164037,
 0.25566000440634173,0.4806680767032555,0.04506811258594223,
 0.2823597504150164,0.9445852779116538,0.015092241533725126,
 0.2645055623492504,0.3880165722052954,0.12340294829144993,0.7518172039961687,
 0.6849651480343398,0.7629904947712387,0.6146115964901365,0.8151454725803023,
 0.6756392046779234,0.9661227315938574,0.09712553497815035,0.7146046596816349,
 0.7035234037238582,0.7077763120856803,0.29328708042447516,
 0.32881023399268794,0.2789912013874086,0.9396434727378854,0.5913069806497758,
 0.9068536994938288,0.3876558472230085,0.8293056419307714,0.25210150646282714,
 0.7945541461618228,0.6173986967349084,0.5895291452161292,
 0.015738484685484444,0.07933166098720501,0.276135537330208,
 0.4248981917934489,0.5663175504147716,0.49940151202390803,0.9853522845816358,
 0.18370218332772636,0.22041322900200633,0.33340410813012344,
 0.209184189279631,0.6459184096025123,0.3914199002985085,0.14700087550218988,
 0.5820803250854403,0.935965971076479,0.3703562555500879,0.38488044298470125,
 0.9159072121878952,0.9050248966939451,0.9875470447036548,0.4449216553575789,
 0.2703332430986738,0.5432078804767918,0.6096184362997288,0.11345629922658107,
 0.981610505925828,0.23597944943201699,0.9776312028844264,0.919055243786381,
 0.08484224435935028,0.9302448511301084,0.47114984643728086,
 0.9020897341110298,0.5833403701104934,0.8866183980966192,0.33224032494241473,
 0.8777842501391553,0.08332442296215681,0.3786584348739075,0.6999547187497038,
 0.5410403853983543,0.7484077347087784,0.24120954986178655,
 0.06869812891291871,0.8082041588103043,0.9819950001952614,0.6541924134800667,
 0.6953087984901494,0.6618811714040108,0.7079338856633289,0.6063958289678513,
 0.3478515614574855,0.7325833689624807,0.02277929034076398,0.3854342066788048,
 0.35174491194783186,0.7723281295934216,0.6881125879075195,0.3997452129512059,
 0.2034102969055428,0.8471197961497708,0.6773999111480503,0.312048039608388,
 0.6343605001496132,0.900864852571474,0.43395866025532026,0.09990870059392909,
 0.13681399396358562,0.815547205071669,0.2627306781269312,0.9286369945020125,
 0.1916425433252752,0.8241193413387728,0.32654563770976075,0.9991652725150503,
 0.2359338161026927,0.9590948824838035,0.8594047344795291,
 0.011984625515569336,0.6063705015276707,0.3304089376282603,
 0.9807234313081302,0.02176577234285748,0.2253534582580231,0.3268987876860455,
 0.7783909079381174,0.5988211015767935,0.14839486629011978,0.9368124207376584,
 0.0359623264366008,0.0854679270323151,0.10160378443470486,0.2857746969491931,
 0.48939328427853557,0.6945966779573601,0.9981151266263468,0.7205139987555961,
 0.4808998628007921,0.14990273111295194,0.7349652814492922,0.8492170641116829,
 0.5655980790422535,0.12502134529288655,0.32978377421063254,
 0.8132323599777623,0.012858672195204646,0.23785920328103394,
 0.5082879579788322,0.9529733483638806,0.8702392048002412,0.5181528830269662,
 0.6836628667769398,0.6299883474093697,0.61820636365436,0.4926988099288332,
 0.6368533588285059,0.1852837066625478,0.8996198034146126,0.10285527934296135,
 0.33054706513516063,0.45599117150645196,0.6107115973052198,
 0.3868988081922363,0.5461147839642562,0.9462430268860009,0.05235982660710514,
 0.44288765335677316,0.05112446280966043,0.30245321844980433,
 0.5424186546443227,0.4248043742351113,0.7700968678280036,0.31088205815054437,
 0.4276556253307171,0.04146600285785451,0.3581602982115286,0.343928509329142,
 0.36369006109087876,0.08884829982022469,0.9855460335940865,
 0.026110273253417082,0.37824802711877414,0.4892492219973765,
 0.5137195871580998,0.7174106256151453,0.2217287513900421,0.7822115191868857,
 0.7844640842045523,0.17733046407859843,0.8590622548800261,0.9517092000994137,
 0.3244139089650073,0.8849001871926454,0.9001801935069637,0.4194441792716731,
 0.8703806117678083,0.7714990660420037,0.6368118085239134,0.1760943536493581,
 0.7604863072820569,0.02720334916915179,0.6744195396406503,0.2040064796001856,
 0.12668896955303988,0.06364853234570655,0.4601490515748229,
 0.5038224115402675,0.879224727343622,0.10476174247596992,0.6644209784508729,
 0.7567270040386803,0.9640620703856583,0.13988649490687854,0.805741117798381,
 0.45650380020506465,0.4526429478975674,0.7827022418670134,0.9774060118741136,
 0.3046853204995439,0.32885440654834863,0.07410224137901444,
 0.18813798075438704,0.24423907985295013,0.4193450016638478,
 0.8446190400655417,0.24697992088968712,0.4053364315172744,
 0.03673007810172635,0.4149434980697929,0.8599715554811295,0.8243084091682562,
 0.547660449069022,0.03976818813109173,0.107325816809141,0.8363622932481937,
 0.1306461876056202,0.18016568000285216,0.7455559733559898,
 0.19474594901469988,0.8713898981218374,0.07399948718065075,
 0.021433951007270746,0.49193267321609246,0.6469071773746276,
 0.16791770440171272,0.9062875947972958,0.8909476190992989,0.1570782855984273,
 0.03586345930880763,0.542136722067779,0.16328268075808972,
 0.49602667627656327,0.8776805101757569,0.8071024330679522,
 0.031426139633400796,0.7372866671674112,0.11464041093017441,
 0.0463913768032983,0.934840747684737,0.4848862743477991,0.6290202660619579,
 0.6560039156871103,0.024462514186213258,0.35050443087179883,
 0.6032666512234617,0.27352503050653987,0.674601143083071,0.8480025418894734,
 0.12707494379050255,0.10151917976694491,0.4731082199248049,
 0.8544261776556558,0.455273445805366,0.4722874389909073,0.7524308237126534,
 0.40343589165934135,0.9822392555848293,0.22001131655133876,
 0.026707433443913686,0.5382830672896831,0.6813108148058631,
 0.6006996558138478,0.19234751750007506,0.16427435894701314,
 0.1671296524441327,0.2927645463474844,0.19808016742512513,0.9454874986379588,
 0.8117143505937832,0.21895008526589999,0.4203758093128528,
 0.18205302008394375,0.6704767756071286,0.021365289912979146,
 0.0521365879098592,0.3347788286929563,0.10077715134317278,0.5729108165567101,
 0.20987225645939045,0.11830348851865846,0.7311663643385834,
 0.1149157098931668,0.2843093223510127,0.6597879395503332,0.5978922929555304,
 0.7331315466260673,0.7013973631419176,0.12014755461731197,0.7082506765562574,
 0.223936379354031,0.9252413341634075,0.8033598138894202,0.11970600333030401,
 0.10149675928195734,0.7450641270134213,0.2694640997316492,0.5799872988879234,
 0.69560779813591,0.03078673253950248,0.3020830632604694,0.503903391153541,
 0.243150598206341,0.3742098896093988,0.415709840075715,0.5150027124633724,
 0.6024596022237438,0.5155084754847883,0.311083890747084,0.8405564306689849,
 0.7288645531358484,0.7518217812826995,0.4827168816651135,0.8783209562953397,
 0.767656547951127,0.8346147633862113,0.3021786313772391,0.7347454982192184,
 0.33775705163381353,0.15653889159783518,0.5189536428003908,
 0.8641368505607486,0.2587947340442909,0.3258103729018571,0.38470109277020303,
 0.08524056857808238,0.6862436750471388,0.6206115045048409,0.6102789866224579,
 0.5434905260026279,0.7359355987311021,0.8713290042529298,0.596105851843768,
 0.26745329495150494,0.891588072817364,0.9475991751398357,0.6206703037021566,
 0.37621840940644113,0.8921568037027203,0.49169037627974477,
 0.25366262575756215,0.17865525530125104,0.06574453130212599,
 0.8556628475981852,0.12407312470793963,0.8712111239491626,0.9733882024019376,
 0.40778763457897804,0.4410102111732823,0.8672580800148705,
 0.22402437028769429,0.9949893112105644,0.7331212282833095,0.2084025909255639,
 0.8011720195561228,0.1228260690390156,0.13296885265727665,
 0.26594364560012984,0.40088616083910655,0.25648080190693734,
 0.3180925556260783,0.2813396621444586,0.4970500454340979,0.21468079180230237,
 0.009839212509468731,0.6532597570726487,0.11675009159516292,
 0.4407398267915196,0.6248998544769292,0.6132713125078406,0.45727492235614187,
 0.7597751829080184,0.6967339937580348,0.02748472653590417,0.3776288361172385,
 0.5571910619075463,0.6425556948133774,0.6777266264939213,0.598268640720647,
 0.9575709403998668,0.27029489165754883,0.1819840687580564,
 0.44436903079491497,0.14692543129544877,0.3635553250574447,
 0.31972934074619697,0.8110939672284952,0.5504893420264705,0.9286305423995332,
 0.8439990357428151,0.16278516102584384,0.6653370751628904,
 0.05556421887147345,0.4729287826878219,0.40248667289473705,
 0.8333865281134514,0.663056033677693,0.525082886102556,0.7883495984233096,
 0.9279248926343723,0.5979091169535838,0.94332665626168,0.8164364535979678,
 0.08341009629744278,0.3696206363636173,0.6510863352065911,0.6205513286225189,
 0.35430925380934186,0.6847338248779978,0.5631908418436332,0.6026306736934547,
 0.537438260340243,0.10033227196880679,0.09527341231366282,0.6020527448991896,
 0.27071662049244927,0.40163526593303633,0.39190001044661305,
 0.8346632653432355,0.05234713019220538,0.6532359584536374,
 0.13182928148950868,0.24298302727241594,0.10191441672749457,
 0.03060948411829345,0.3154963243311222,0.40824886318841547,0.433136057058256,
 0.746764808980986,0.1103038209248941,0.1166395346785497,0.3401918177640022,
 0.4730462960005575,0.9023302617699133,0.4993286434792793,0.6148537268599366,
 0.27620545843181943,0.3035254437774103,0.9768684015088424,0.7377362105089539,
 0.22775395884909333,0.46100426282371987,0.34277082760244915,
 0.9058231118252504,0.03256287422530035,0.6398314259396379,
 0.11764344940903149,0.6227817014310219,0.9513717997106742,0.5805358320143454,
 0.033290779558476036,0.6238294676365334,0.6454933294765497,
 0.9493628053226432,0.2957317052204559,0.4098097437999785,0.6874814940717296,
 0.7394734289097059,0.0024414611188559032,0.7627106038908882,
 0.6713757370121328,0.19372540650421888,0.022242260091849264,
 0.4527938030377102,0.84344338372415,0.606329113528076,0.9584073807948399,
 0.15512306039267576,0.23346739548766582,0.6244622324752884,
 0.9822272301485048,0.5244423272634517,0.6548664917708426,0.275131302317861,
 0.41776565453303605,0.23000948233457652,0.03864865676555418,
 0.1730538388549644,0.39918709372355055,0.37243563852987194,
 0.4393409247492437,0.7928580551781136,0.25513805453532934,0.1078903292842186,
 0.9602795993639519,0.6535937684889362,0.23500744745807045,0.4372638615444121,
 0.04664743903035862,0.9646525267091076,0.2196808881361384,0.7147091077112179,
 0.0047310082001318365,0.23359970296106025,0.9101108084815822,
 0.9613763426736667,0.38550185730335307,0.621802778041223,0.9556086877666851,
 0.4279333798727909,0.006480108462637468,0.07249031447043008,
 0.0035379068674394887,0.769586377572573,0.37211346308205917,
 0.7571709966778067,0.731283434202962,0.3997817928096793,0.5895001320951335,
 0.7709957335138071,0.18470336786063657,0.11356919726670212,
 0.6762589864208618,0.6944626024290702,0.5924816617602572,0.3171359801727499,
 0.41197711637306966,0.5515401664512125,0.857568336458322,0.7061598826614139,
 0.9635466787364511,0.6569447797428097,0.4290520124572942,0.8745594985158013,
 0.06188855611632427,0.8079701046572836,0.4559892373995039,0.9800879249194973,
 0.7352403845123375,0.18294469008501013,0.5722361715053464,0.7943954905334829,
 0.8522625767161337,0.48109236224937146,0.2040132644568068,0.450630543239651,
 0.04298973181447496,0.002859235107570357,0.976646587753166,
 0.7533925448108664,0.19604044910336893,0.29528624125656355,
 0.15211792248903688,0.45608962553452226,0.5804990020968082,
 0.5947860077920927,0.0011232113063656124,0.14387656568011153,
 0.9108475378034492,0.4313047533754415,0.5303997984240851,0.6504482229866438,
 0.8029285855349466,0.2646967516768902,0.7613605916099271,0.23748415106401488,
 0.4172279435607449,0.25798236226215565,0.732832372634462,0.12453385544953877,
 0.7792716651704272,0.3563011603211663,0.8971838296759144,0.9358691778859716,
 0.5960894624026879,0.6579005546050394,0.9841177378312936,0.8043615127241376,
 0.470181223181533,0.042012167746394935,0.9960218795467529,0.3561766996631358,
 0.6913353377879109,0.5739710395983102,0.6464853838809191,0.3657913344824171,
 0.8734464499593263,0.6532851539565031,0.48526861066486227,0.5707928635341467,
 0.5119921589906684,0.12563436921624593,0.41705188768087265,
 0.4913997390014788,0.20649455826634155,0.6223040604835876,0.651589815216597,
 0.9962352553696946,0.9137806225043307,0.08753033573982694,0.4786194103084451,
 0.7743138927113624,0.31218088566226665,0.7035369000318568,0.5449205837648033,
 0.25085068803571375,0.4516493033852793,0.729874086681023,0.09639195218090968,
 0.7616784340449015,0.913343780445962,0.29638155949570544,0.2562031576697349,
 0.8868786238681793,0.7507466603721682,0.38406631687580606,0.5916403228741608,
 0.8918158199726918,0.4718341570355804,0.005140184046579566,
 0.4386228586037483,0.7428014956083228,0.518571753377048,0.7654826600890046,
 0.2546023424134696,0.5645812984159777,0.10900943346175462,0.7539030854497533,
 0.08497814556173333,0.44953204865908725,0.5258095107672417,
 0.5554831928258702,0.3972284027473191,0.020164223008972026,
 0.8078173174964752,0.21195247029201114,0.049644086959101275,
 0.5193261586007467,0.17060967435210717,0.29703655966533626,
 0.7010488423322994,0.19948455764152184,0.32648927292907914,
 0.49031973430078235,0.47945417849082195,0.4282920624316432,0.644939664283657,
 0.23303540420422042,0.5568070307808223,0.7463744173152675,0.8380749920620112,
 0.9717836153815937,0.9303508205298445,0.564106918500091,0.858638698300596,
 0.7647080726737503,0.6578261238329811,0.2751184067218293,0.09885818453405104,
 0.825945521677971,0.6645668932930016,0.2292038498041249,0.9299483947032295,
 0.7154239874688935,0.21082377378375672,0.5149822319173111,0.6667705248113754,
 0.33676127207607676,0.5319971164049873,0.4616229295243852,0.9014890021758768,
 0.18554019571584157,0.16833526429893775,0.8531275137794403,
 0.41830352854289654,0.9734940070395863,0.6483869703013968,0.728028664455666,
 0.03627923488829865,0.7260924669614772,0.7807526833699977,0.564170257867277,
 0.7873299373672766,0.5168132871198652,0.86911253749302,0.9784160757871002,
 0.9997096599792986,0.27947524593751794,0.733583806475668,0.2539175291220218,
 0.09648665111052657,0.2282695842382313,0.494817081518655,0.30460671103590486,
 0.7168004057756759,0.5534503639331828,0.5633874917579104,0.20907844980104862,
 0.0733400132451838,0.9018045136618534,0.5971891619220959,0.30038933075725605,
 0.8765425410711289,0.9366702891744408,0.7833059553966689,0.4690195533719176,
 0.22898791098876625,0.79524235090531,0.035307366619471536,0.591223989596344,
 0.00969432835269346,0.195743613905214,0.6621048317342709,0.4033209535600546,
 0.7112120469806614,0.06445697935243433,0.5476158337771811,0.9579542907994536,
 0.3218601104827218,0.8607030030140596,0.002190652845532104,
 0.12160837557339954,0.455796202129218,0.7780131000597048,0.6164884366476374,
 0.47175799105584626,0.6792733641289271,0.5175779092441523,0.4644759436406848,
 0.43641578803997594,0.28728912910426085,0.2338639400621605,
 0.4996624199394307,0.37512861936466413,0.9818376628881015,0.8375050587888989,
 0.08448540115908698,0.5818375461212085,0.7293263874607997,
 0.18375282162519668,0.4556890359526108,0.4611558959473355,
 0.06096400830695614,0.059588428016615236,0.5143417371974326,
 0.304784162812646,0.839800513959908,0.6624904377130458,0.20690325480921934,
 0.5074065921909171,0.5063843639381732,0.29939625767945677,0.73889601285382,
 0.7549952202741134,0.27675318630128476,0.22642540918971166,
 0.03405996889596041,0.5163980379157852,0.7836592909193886,
 0.24163484459979712,0.10522283726534576,0.33254565686316606,
 0.322143387974005,0.6823348589616199,0.9827851669894763,0.4481749806568327,
 0.2884077045897159,0.08444912056428089,0.4969294257505805,0.6197934224185118,
 0.7911005495269468,0.9951595977977782,0.7234788981990112,0.1259525604696733,
 0.8589796162594683,0.9764390059718132,0.037819557905948153,
 0.8501047498501162,0.48803387929823383,0.4174910911454234,0.2662444008864935,
 0.1342868120368721,0.43063951136066125,0.6669063661396414,
 0.39953675477159856,0.6838408723859959,0.5785990891607121,0.641616971016403,
 0.8347289734520845,0.4010440182645204,0.4323594991966787,0.08828470237048003,
 0.808760256510001,0.5081297049504617,0.28905998095754337,0.6432198850675637,
 0.4456248635547404,0.1758690412267493,0.7196254805961305,0.979622120378311,
 0.34411006855079085,0.1861940460317767,0.8304719091774321,0.2975861796040944,
 0.22561298930166895,0.3286399799839851,0.43191841353010463,0.677562300697543,
 0.7126084744593965,0.5478435812931581,0.2117755287210129,
 0.053336787434628885,0.9634807890962124,0.901645597037462,
 0.28499680519935844,0.5411062691713562,0.680840755861775,
 0.025201016011110577,0.8806311036082096,0.9014254194689568,0.208950435247529,
 0.2524003444478846,0.9157583601910192,0.18177986376787214,0.7847562116444627,
 0.8518715504780486,0.06803171735291191,0.5149089687399192,0.8971417629553535,
 0.7548188885307289,0.8861730532840439,0.8407701320969178,0.7548021538339892,
 0.3754295713631731,0.129426552275289,0.4050040350598607,0.5845108557588004,
 0.4711694696916289,0.7582233996248976,0.2243746141204167,0.16462158863427034,
 0.280644273032425,0.23562031697814378,0.18030741466653877,0.5933097319407875,
 0.5683802687775145,0.7686839479026079,0.9610324037557437,0.30195945236291,
 0.8158021464319896,0.7184438980831431,0.4894156157176439,0.055119281566821,
 0.031223637515657154,0.008181938063740235,0.7676607942859032,
 0.030777231350873513,0.7045129805915462,0.2963912940487382,
 0.9845550810403819,0.18308767530052095,0.8689980149336821,0.2721059629705205,
 0.16248953913575648,0.1882596470091339,0.3204329567032378,0.9512133830227857,
 0.9972801455328265,0.6927173727983739,0.07371170688185524,0.7736286404694794,
 0.004056067686895259,0.30793420608878574,0.5691300124864435,
 0.6911626642536507,0.168153117932911,0.6685105410542643,0.29071929603897084,
 0.7557961879726189,0.48753077416388724,0.38764813719978786,
 0.08455903385376307,0.8064845287123034,0.029566176287328383,
 0.7167394800986006,0.7727726179691675,0.22464117336690437,
 0.20022066444048459,0.36636311372748365,0.4131894246283653,
 0.2783934210009269,0.6969966851614823,0.26302419982002656,0.266490309773173,
 0.49584053009221984,0.659272146873666,0.5542642986947451,0.2367595576577438,
 0.0844276205384451,0.6238157000783208,0.08887612507855047,
 0.20429261373081908,0.0016804245423336361,0.22194688879394597,
 0.7755197038558095,0.8346810966188225,0.30375389953282106,0.9525765316510029,
 0.7533006534072708,0.36834829054436435,0.19987603375720486,
 0.21821333093484296,0.2796968121411808,0.0023471613008320125,
 0.5859137394047333,0.8111590139606771,0.5036262185571723,0.14364824300913248,
 0.7362942747470502,0.12969208339760985,0.11238089318765998,
 0.9034600368583936,0.8817748425650035,0.3558001499618817,0.4199215986899044,
 0.3914508983542482,0.44242699078229386,0.11440380660661742,
 0.2519588594011074,0.011404938368327788,0.6966916974847205,
 0.9006690437595986,0.692947586261852,0.9033751993556924,0.4983255421757864,
 0.9183929899166081,0.3175622332516085,0.9831748129569806,0.11301584119540797,
 0.933616250576114,0.24340810379370037,0.016340037660854567,0.729276808261836,
 0.015799296006828167,0.14230188490066875,0.338782667632184,
 0.4118358125643463,0.47010733264083493,0.34319882066616103,
 0.03448019972878247,0.8872072119311312,0.32004331374461203,
 0.8249920880189986,0.7394338028022277,0.9479527871371044,0.1572904101779744,
 0.061473836673380355,0.992144393624214,0.4492084037361438,0.7612293616008392,
 0.24343424331553232,0.5024463603252347,0.6280686617120866,
 0.30702629167011997,0.40383080444110797,0.235362286744772,0.6571311814574923,
 0.7630640163292991,0.46480144343135077,0.06856052325653372,
 0.46053707659039966,0.02036240235891773,0.9587404706969408,
 0.6332816890814815,0.2889623157061567,0.21774446821928062,
 0.44858126001377074,0.5642977218757099,0.4077182812488518,
 0.20343304991970157,0.24041632609566754,0.9382470094006045,
 0.19578278159228135,0.5167949315562913,0.5705311126953578,
 0.15339885302103284,0.7037069355765262,0.019422682157727467,
 0.2202092280124861,0.7156972853746048,0.9407206258420785,0.8045593828301962,
 0.14353331123920998,0.2001291906688607,0.9728115870740379,
 0.30190907291900015,0.7750454223577677,0.06139571840442004,
 0.3995702567136976,0.33759305644410886,0.3622318541400539,0.3158720424417223,
 0.3989583143402954,0.7446534577451556,0.21549160176872595,0.9132660065059732,
 0.48612235684021454,0.14306183247459314,0.9445730310893138,
 0.7308176609986842,0.5442292027420305,0.6501485990950744,0.06859588856386978,
 0.2048162611117238,0.690573157633476,0.9293824118408408,0.03614428649610846,
 0.6654511336334301,0.785025859028001,0.9324054885585368,0.3699076386680047,
 0.3556086815959768,0.09213990908901426,0.1510911198587147,0.5832996789513732,
 0.9734224811787168,0.37507558889641257,0.8349156971640398,0.7186486146455207,
 0.16621613029896853,0.7477531274658904,0.36450215626694094,
 0.7785439628300704,0.22548366577903567,0.38023564705903046,
 0.36470903988435344,0.18527822360295165,0.20381989082985918,
 0.0881559927423986,0.4656844396335562,0.6746063675921358,0.14090327178825457,
 0.23302338161276515,0.599102950601008,0.9493103857726585,0.617879945309078,
 0.12183092924649674,0.7304447800148901,0.5379826391615469,
 0.09794495186719132,0.2481883715514177,0.05401575463965558,
 0.8050608565998523,0.725745136917257,0.7033144620286539,0.3402271652440767,
 0.357733965469227,0.5294409982074448,0.7935698801500339,0.4963273839039558,
 0.793913050018641,0.6958078325294272,0.2870821522393765,0.9349170058541084,
 0.0027193688572466623,0.2819306793384718,0.3200580218230257,
 0.05430682038833179,0.03096417342817137,0.8298261531801623,
 0.6999753014118029,0.7249245132138569,0.6565017949809188,0.29416783179427874,
 0.7158724704567031,0.45421138112099146,0.21154266966537483,
 0.6809149364256802,0.997561782638329,0.2189675292340013,0.8589868113494815,
 0.2479759694707362,0.25220153761746034,0.7021383222306878,
 0.40102310090971893,0.6910761188283415,0.5537212830627846,
 0.22041904533940682,0.17759023989136213,0.12570285625375943,
 0.1956232917205286,0.49284498964850354,0.0994767234254762,
 0.20130739057976632,0.5530391757007775,0.2750530728455751,
 0.06147661522513448,0.7573936671503644,0.5485471201871381,0.2731290193636823,
 0.735711213991062,0.8711733546714671,0.5031695185885224,0.9527058259757848,
 0.47375968275875036,0.8329337188786794,0.611576490995512,0.634258341455472,
 0.9311491164368411,0.8751794573740269,0.5716418842805188,0.4556942730986374,
 0.5993542707679351,0.7702311569720185,0.4542483417101274,0.5402580286848397,
 0.3066411048376809,0.879692568450373,0.9520564167001004,0.27859136897233727,
 0.21091460569073595,0.6320697282868316,0.46827960234959776,
 0.08500732204639683,0.5417025740490674,0.5978549293005192,
 0.05919774552887336,0.421322836585467,0.9922287305913894,0.3526342670355702,
 0.4277622594282492,0.08271202662789379,0.37904939547281236,
 0.03507466403223258,0.6718631989874163,0.28346364974917493,
 0.12462735375465994,0.2832183377655476,0.127753086358799,0.27036770212084427,
 0.4608648629254123,0.19293936688121027,0.5394543460609513,0.1149410416930956,
 0.9333971050060393,0.9416006359681612,0.20281598795214673,0.5165605629320873,
 0.6425566898314299,0.8024099349808947,0.0794789443641557,0.5078967735582569,
 0.6769387593694893,0.45220514729257677,0.2527888931083576,
 0.10552951404080946,0.43557907452752265,0.7134568539230493,
 0.1029358878487665,0.07417957528635422,0.6159783218161516,0.2749295837415038,
 0.758321716639261,0.9211152688034219,0.594757025590729,0.7641726690556359,
 0.4462560739721103,0.11671994147007103,0.47861732873131946,
 0.06740811531074642,0.44635411195673125,0.5549228840776351,
 0.5566868744070936,0.37827857123756237,0.5476637356036703,
 0.46667433077060394,0.06400912790701907,0.5515231151840321,
 0.44685564605921324,0.348411054602059,0.03251920099070471,0.855010537365812,
 0.27453939882656864,0.058705549671525525,0.4283884581786974,
 0.9324192672307461,0.03349453241888367,0.5957004968776867,0.6891097032334459,
 0.9964176432112869,0.6474603075450545,0.7214040691816679,0.5007224942779402,
 0.4375040424105763,0.7991945886027605,0.40388271573920287,0.9866843741853022,
 0.5714757798881984,0.9499620856748553,0.47782511037997866,0.9582475535564106,
 0.16858256089188517,0.889148864762269,0.9000831427726587,0.3693158617950163,
 0.12411259307387834,0.27128098347866847,0.8519706824647704,
 0.32648630947985613,0.05587366455962761,0.9243071030122287,
 0.9897125622702877,0.3283431390505098,0.02993636507534403,
 0.27592930698746043,0.11640206520712326,0.7861593781240616,
 0.8870412540158641,0.4733456236654695,0.2702074078186344,0.840379976221491,
 0.029891099753852957,0.23357410181110405,0.6652070396089533,
 0.737112835413461,0.8038862872594487,0.1431699011498403,0.11884071386527062,
 0.831910128594826,0.5262472164383007,0.156925134292906,0.7535295293611872,
 0.7875289237684191,0.8073934146112927,0.21620270471571157,0.8934610932383898,
 0.5679558014088182,0.12123136737343576,0.32545520043722664,
 0.8826957619044783,0.3936638958100247,0.9940774724076972,0.6051626799468596,
 0.6388309994982531,0.9434332510797899,0.18975647874524504,0.7334261265374288,
 0.05202841876538422,0.3913710537790521,0.01362590044334222,
 0.8961915563144334,0.05947663973477135,0.49817346266724927,
 0.9498792126216885,0.3534604874175574,0.8235882928760231,0.6824933156108103,
 0.6497319248223109,0.2824060479024907,0.8273245148375704,0.49104331990232786,
 0.17845421898617475,0.6614211374620493,0.14285375499922082,
 0.6858467440319047,0.08263017837454656,0.2863022114462521,0.860262999185916,
 0.1606276001376612,0.42726607544082496,0.39497959380937275,
 0.7538409422063056,0.41719584508566065,0.9432073542352662,
 0.11654268863407691,0.6864410497689677,0.04919573893287987,
 0.19669538678230514,0.22538847778154392,0.7240489309404186,
 0.9324984966055653,0.5958573407250616,0.2822995885680414,0.5193909060841018,
 0.8256870625606259,0.715664622132911,0.5929930569105777,0.7406934301109854,
 0.1851189072856958,0.45979887760893523,0.29572057374396643,
 0.02093097217013684,0.226584876223486,0.533266126372255,0.015793297615727897,
 0.4590790841101977,0.4773813031806804,0.3054077991427172,0.6151068199818018,
 0.6181999961490554,0.3869350078601761,0.10742112109629476,
 0.041226406423863216,0.731200860250577,0.8100320883892713,
 0.14704068352831035,0.5785061869840578,0.566994438204327,0.1704794260610707,
 0.9754681202208438,0.40157311619918135,0.8208582539368418,0.5188072828949855,
 0.869768759895924,0.20027766787206147,0.6778883137220322,0.070023453066126,
 0.9065738249927265,0.6956753177787689,0.04872172693176857,0.3121859662485629,
 0.07661112768822564,0.1285325908262288,0.5852590992083508,0.7338343286027096,
 0.05763005587786463,0.12362591721465621,0.16248405941181976,
 0.3042030634032248,0.9114969131557525,0.9565403683154131,0.5564723395889848,
 0.7353295643966518,0.8274335376549188,0.6954065612539756,0.18817975624891303,
 0.2319168977644327,0.8126312421168929,0.8779564033853915,0.10887479590933735,
 0.8724049204541201,0.227645638073668,0.5522480725730078,0.6039368631441616,
 0.30076175281573325,0.425115453987552,0.16479197797163714,0.9282981624730945,
 0.04341909398261967,0.43687807800598777,0.8501869653784969,
 0.6729921722340042,0.6097605498906274,0.5394056648946601,0.08701066827224901,
 0.773722954903577,0.1610389078208866,0.6780603980911802,0.494084596286284,
 0.6770109225564084,0.4376728791940018,0.09954580275351332,
 0.32922866537336226,0.5925559888191367,0.7084003303955887,0.5943220894878376,
 0.8078185566527959,0.26581024881679527,0.2878250130416451,0.7335314153927501,
 0.6495644365318833,0.05363062576549571,0.2038455801075627,0.8799067102757352,
 0.2779452030803681,0.993410845083236,0.9422664813414172,0.6467829438985069,
 0.335954615031556,0.8657298560597179,0.8062080178753978,0.5884001041956893,
 0.8811849847005098,0.4947118534203374,0.025912138311316002,0.781011023347084,
 0.11281499675567974,0.33909578369745685,0.7761534266369212,
 0.7339274594210371,0.9692679518409135,0.2435056179816928,0.32720314650680205,
 0.2619324048839897,0.8072844745268338,0.5155513744702154,0.18470250367278163,
 0.8543831745690387,0.17671302847764458,0.22166607912714564,
 0.19988860302646438,0.2734190950062696,0.7671679635009283,0.7415545299564635,
 0.5811524534709528,0.29260318277892994,0.6508405514814862,0.3425695447594129,
 0.29061363662993545,0.5037618447076344,0.38247631221012135,
 0.4751510042645466,0.9645991563758705,0.06262994581192538,0.8846869312121938,
 0.24337089218747054,0.2617714647574988,0.685842492425319,0.162359259803466,
 0.35566450646885106,0.5575773143521169,0.4354923113361058,0.4387104274549738,
 0.4449341094260455,0.7721855833082163,0.04621459353365176,0.5664754983072673,
 0.5218825500414965,0.03673768367194097,0.5969180691104755,
 0.40815036134371274,0.6325922473104375,0.0905811919668118,0.6522349462785949,
 0.27237572591979364,0.8997892310746849,0.7655454983957938,
 0.06218038885068522,0.3226660078088679,0.3723955743135451,0.6941754146583416,
 0.8021753776211156,0.8212540204500388,0.8895029862569159,0.7167854923722842,
 0.4511657440738557,0.2925219784538815,0.21326677914273684,
 0.18553330239042376,0.5005849339478224,0.6152868232281813,
 0.13572057792020542,0.11852614606878775,0.48604803158591403,
 0.4255640002563861,0.9292075023226242,0.34957296088803547,0.4253921390265516,
 0.4866995857715305,0.013390226592626453,0.28523560610742693,
 0.851263890026285,0.626708069192724,0.4882914692947474,0.8951214825829381,
 0.1549620238847339,0.46612004461686596,0.5380099890045281,
 0.01580998930239752,0.5545116616146931,0.4908920293058552,0.7187916316593734,
 0.007834953935672129,0.8894016583373988,0.4153577991475731,
 0.5228137075021136,0.9007499862795687,0.6640266294647214,0.6289512496265294,
 0.8230598990318561,0.7173089379251782,0.07690430386076208,0.9613463817991497,
 0.21088223143084084,0.8094888375004176,0.5094183852289327,0.5629306154539092,
 0.9674407353361245,0.1421324091348234,0.370247784067673,0.28200991671970754,
 0.12675959834051997,0.6073579682181812,0.9731568990985032,0.9945666031269198,
 0.14638756281782972,0.9722033716257303,0.38507392301619103,
 0.7286815825296302,0.5542663671795213,0.6745063064466484,0.9884437962654535,
 0.42579139144352407,0.028669956870640112,0.9244207526027085,
 0.8285162177872782,0.8128107032143592,0.9455883422424172,0.46104316467736994,
 0.43252774395011295,0.2039767391504218,0.49891719434582427,
 0.9227148870803416,0.10450651864645188,0.9411885100829053,
 0.11513225103578051,0.794130748131246,0.9363365840373502,0.1808176716563178,
 0.8498990685202654,0.3557540453947068,0.6483930387184742,0.5880094294868078,
 0.764592389462338,0.19482335000756523,0.7747594075029213,0.7916094624527988,
 0.8697329220801158,0.38019838614395907,0.4555597509884586,0.9591140226800267,
 0.14416911275314248,0.30335272054613194,0.7046345483068794,
 0.4090199516538736,0.05436673936268521,0.6969169925993915,0.5088659334762502,
 0.5224589873814999,0.4672899091051118,0.5415713906364137,0.2799355050805744,
 0.9379289862006067,0.41076265007629387,0.9183496753429106,
 0.025523484781743377,0.063559211636668,0.8942160615962699,0.1715151506437118,
 0.6402487258482412,0.9440644975131809,0.5467846935786705,0.5200928301803198,
 0.701264791666764,0.6481250522176538,0.5006668456897232,0.15409848437051588,
 0.46140436592949197,0.9197022317910428,0.22950423001845877,
 0.2502665683551105,0.5650578169584759,0.08450387236134305,0.6928435855882189,
 0.7066605779037383,0.7443387605209509,0.30224043081187724,0.926517480577875,
 0.4066290769697355,0.5469325643660936,0.41067281304503633,0.7239767968625498,
 0.02770394760728012,0.22236900130925252,0.9545685018989573,
 0.27508503034126097,0.5558714829856557,0.35416182747471847,
 0.7102193398336409,0.468623149588812,0.8697817871365896,0.7860301227957275,
 0.26636433801210524,0.8827949599503893,0.5640757744589506,
 0.08972592227786758,0.5905717081154959,0.5419405259767824,0.7849657516204773,
 0.4832791309981972,0.24986416009422996,0.03022401954902887,
 0.2227138325296938,0.7123924899744749,0.5208249641850178,0.7457054128681633,
 0.9702602850174956,0.32558418607907635,0.6071654907777945,
 0.013733544375468143,0.5901279645234809,0.11426046994848704,
 0.8341753571670874,0.36925760419571974,0.19495795514286385,
 0.07594272175276684,0.12249522250759315,0.6328399341451922,
 0.6970802616450263,0.7541255075977419,0.8376271840827064,0.25879117934320883,
 0.3754823958216339,0.14150261123925234,0.32128766988471014,0.543518984038162,
 0.38441415508822563,0.5446082429302093,0.8869122946670802,0.5202905875336851,
 0.6566308315468876,0.37937667342215275,0.11303430971785033,
 0.8627691097105111,0.7664507879722885,0.85946778683983,0.9657200457274223,
 2.7368253336268467e-4,0.2052130800057239,0.05311599195841499,
 0.6990919119391107,0.7551959210754969,0.8384903079047851,0.8879640310270097,
 0.5382712377783067,0.19269369284886917,0.23997045004552753,
 0.4965046062169767,0.6372211552453227,0.33709760321822746,
 0.010691905806806568,0.14745707787444606,0.7753391386043584,
 0.7671866113601435,0.5490598613267221,0.6459531281260744,0.8235136990285569,
 0.2301637736113875,0.7578318300094296,0.5869755144017833,0.722225285121237,
 0.4501778684607538,0.0807915615413054,0.057955922196638254,
 0.15001237641475962,0.6440573813674589,0.7918567903115514,0.5331766668145614,
 0.37497333136276,0.3738158870171817,0.44230696684356385,0.942250557369369,
 0.21377826665728916,0.938124312885504,0.2011146512515094,0.2802843940000228,
 0.8088922853140313,0.4942275563141032,0.9919912028518636,0.9890401167411209,
 0.1423163611574948,0.6223805245384756,0.39171008922676676,
 0.041091064053308535,0.6153461425770544,0.08410642276544145,
 0.6818287338320492,0.3369982156440656,0.27199823932224776,0.7722955336343401,
 0.9614082796203811,0.13453237998767031,0.10340565571276228,
 0.9306926829824764,0.8750593557761132,2.260947954576542e-4,
 0.9956337707973576,0.2080659662717158,0.09735060528886486,0.5688260067693015,
 0.6018421021896189,0.8512454052384508,0.48712184462995034,0.2901731775987103,
 0.843932894525825,0.09537651293393878,0.7191215730783032,0.1152808408714916,
 0.21161901496290136,0.9997344709383206,0.5317131468825288,
 0.16651226765818627,0.5525890626384778,0.05274666064818534,
 0.1444056623636737,0.5709455485956463,0.15475869235790307,0.2691378280542944,
 0.14264359978726715,0.872075401738365,0.1058025000193985,0.01804896473045048,
 0.19750619346436948,0.0899990307479207,0.6032108841201553,0.6764835021333734,
 0.07990831134220644,0.4169248428733019,0.9876274441999962,
 0.08688976822880257,0.2977325378764023,0.0357200236195071,
 0.22393545129323866,0.956548134946036,0.9563549303589896,0.3072342356792239,
 0.8972442162034365,0.7141328068009474,0.009811505048809122,
 0.9265886606468622,0.6943523040703778,0.5218494540869552,0.5962174205427532,
 0.9745176255598982,0.7655937713176375,0.3559107463675496,0.4044287408247992,
 0.028681939768168463,0.6845318875433345,0.14998491645221934,
 0.5548673415143148,0.7241217747041702,0.21321498874775113,
 0.41181592314125903,0.6577269629904237,0.8637966715147302,
 0.25194876439960545,0.5364589708625721,0.8752273933945024,0.2629083008684766,
 0.5485604366071338,0.530534988453438,0.23035573240800256,0.7806831303580316,
 0.7649209485014109,0.13190425174559184,0.508575995002633,
 0.022331039179839474,0.321759623806243,0.9395831199002159,
 0.06296209166117128,0.947074496598007,0.15223828583294252,
 0.08249842655808237,0.00863220157682254,0.9489470173878221,
 0.4020495240188222,0.7753524773569279,0.8483917955258002,0.2788837944720708,
 0.0991339557366806,0.4657157580145683,0.9683686045935732,0.8122155783934332,
 0.7510033663709462,0.17393062075155352,0.35612120603507513,
 0.11597313291700806,0.924797212439523,0.5383641641613042,0.32466742972238005,
 0.8065588841665974,0.8106289842533726,0.2084178835228917,0.46320618586670137,
 0.861248933991814,0.8786796184454738,0.41194844554137366,0.7710376229539718,
 0.2624231695095681,0.5188118011946063,0.9785181888685976,0.728376474514073,
 0.8364281617185401,0.3180106065108832,0.1822205975752389,0.7273947682576265,
 0.6126248574796412,0.30542585095043007,0.5265549061366445,
 0.17747408056154101,0.6116777916722327,0.9869512886015485,0.9003690042275589,
 0.7855495652449004,0.05499951573984163,0.7194498961948123,0.8062945053650171,
 0.9601875224079293,0.46741157030355585,0.06796846443702564,
 0.42106813474321103,0.5861826206122602,0.5469803801364052,0.7449337174095063,
 0.15615130295604085,0.2945843275422705,0.9844470661077396,0.1529887254420077,
 0.0842343144984179,0.9822816224867348,0.6466421849825431,
 0.055622116808187516,0.939411286592826,0.7787813884145282,0.6841968081831782,
 0.7759539708547878,0.6732046618344857,0.5921063954727801,0.32227301284226184,
 0.5021480814577353,0.219368661100525,0.7794448779377945,0.39209149597773174,
 0.07881387979280108,0.9540389879890313,0.8723973812398039,0.4815963009608384,
 0.21442495795916217,0.6099077673412038,0.06058759970374206,
 0.19769505320182457,0.7194870834449819,0.6723176235394838,0.4410502408879541,
 0.08729890346467162,0.3697793214624553,0.3464563180165223,
 0.42768639562003175,0.33785855848940716,0.9722558764169393,
 0.40017811968071537,0.3959185050355327,0.11480281246963553,
 0.7740192521947767,0.1731514747769476,0.7151882386103177,0.6094927998538122,
 0.9799770560112557,0.8205244061384921,0.057485569614345366,
 0.8203942057564753,0.5098509391927174,0.28191729523335995,0.4805797821645965,
 0.3073975153285593,0.041224432969353186,0.22658788224863402,
 0.6133766137181264,0.7913774818061171,0.674533612970105,0.14495964683059248,
 0.039881670005755265,0.2647107157411567,0.1961221250579004,
 0.14764964236443723,0.5512267488855463,0.3126049236509344,0.6918362015591516,
 0.9479478473452388,0.9983265026728616,0.9768302548060365,0.8751034920455926,
 0.613505366148809,0.44608012596531355,0.6243853448010851,0.5206594686934501,
 0.24843181088916688,0.6368218324942648,0.7727399267454386,
 0.03452248617457476,0.4211235934192702,0.3966074750773585,0.2486622879691428,
 0.6678573912718114,0.3305927515882078,0.956600404781842,0.042982503475109346,
 0.9064217833549904,0.5430759748758759,0.8379810713575072,0.12386124840259949,
 0.3513640909436683,0.00767606801189058,0.7794909200720933,0.7415524631306928,
 0.7901456959226908,0.2334859986234955,0.2887073671406316,0.11048120442119547,
 0.0077285417016372815,0.1647493758994556,0.4240270871469849,
 0.37424973096758674,0.7331329323521473,0.9855566263753319,0.632134402215587,
 0.22489212997605035,0.38399147622261776,0.6427946812564241,
 0.3468805761983489,0.7810183973853821,0.4215143752558824,0.28696606239203937,
 0.9505847519049999,0.1518413012711014,0.48057373194449615,
 0.21359075393395455,0.07810504736538404,0.8962771322583625,0.827845205013267,
 0.27909005389200503,0.28550327501209305,0.014214279079273195,
 0.2506888661340756,0.5604022081444031,0.27304615089295725,0.2065370233749424,
 0.8397019702188883,0.24491767609189286,0.04881304953027377,
 0.39072226481427663,0.6268011122307637,0.365740868635665,0.9129145153580396,
 0.7899504654247024,0.07156051882214653,0.833936876348706,0.8033365657517859,
 0.3046523328291094,0.4503003030744004,0.16871162612961133,0.1516711968235369,
 0.21647260883470887,0.36379229079982944,0.9238264141540476,
 0.029944538091646655,0.7731014799764452,0.2095159420821895,
 0.2640320361656433,0.06591218599496906,0.24206764659766478,
 0.9800613494972191,0.026799213607966332,0.34618151104006556,
 0.29088283011330895,0.3267278641574505,0.6382701169190292,0.6345821415239311,
 0.02602949845170288,0.7951872680769102,0.08511778588382402,
 0.7012506242163589,0.27715949623069236,0.7585994223671634,0.4820432634744751,
 0.10931558032460165,0.6374326243996439,0.2782841971369544,0.8783491243184596,
 0.8392375897359442,0.24262902857277935,0.8798597270958282,0.8255750037007943,
 0.9991101764273957,0.3551941172031011,0.5645930529874776,0.7919132023084626,
 0.51764101778493,0.3677153107653771,0.9261810520335202,0.20617551670420597,
 0.8544001342738783,0.6311588570532592,0.3026318474246089,0.8960512337072325,
 0.6231040294084569,0.5953441381231437,0.6746864335555519,0.8655756669988985,
 0.060985917387885724,0.03445317662156211,0.6795359780207362,
 0.441497319006948,0.6191201686436245,0.8783934445600801,0.8928511065517803,
 0.3118615656253263,0.3473427451386937,0.5901475974202246,0.32398224242716334,
 0.8908829442146006,0.2742182275426943,0.7786872758584084,0.563325875769553,
 0.552952626355014,0.6964810653437608,0.6223760332621966,0.9958688286418836,
 0.7923969889601483,0.7277592869933177,0.34104456151446905,
 0.20348807284901704,0.3658015225364192,0.39124504337017296,
 0.8193166705178743,0.41596027400347324,0.4835556162883542,0.809014820554763,
 0.540793304453456,0.5818034526560288,0.8171221592357574,0.15189975514912268,
 0.5551846776516618,0.6609709946219033,0.22273676120376995,0.7650567827278517,
 0.9711955349455632,0.9125717901364525,0.9373176709563511,0.17416493748774142,
 0.48292834090433534,0.8853592603171564,0.7363312325526863,0.3693142102322816,
 0.022345392791747076,0.24576469386653144,0.11912401269585016,
 0.05421114347650735,0.8624686210219144,0.6757317843369014,0.8178436556370323,
 0.4401089370220257,0.7361997430946251,0.718937096224534,0.20256667780234217,
 0.5546175582136864,0.3312330470769773,0.49182903082240115,0.5074530978095448,
 0.9657767574870892,0.6408506028193103,0.7412351639089055,0.5663228791534789,
 0.6763320195275083,0.7263952847892691,0.24715786707547993,
 0.19516267887461192,0.5464594253106475,0.6289860685937908,0.6183442007221062,
 0.15294809824896827,0.27693781105456317,0.031837595851690326,
 0.3126350749023713,0.0066878923764356335,0.5373080530182477,
 0.6101828100307483,0.2869574032946205,0.6570518362474105,0.4378460458730825,
 0.5442209082570055,0.38025600391085235,0.5582809560775812,0.9612871882979558,
 0.6469470561718366,0.555041686575114,0.6186075370650533,0.8320872824882615,
 0.5675178444453354,0.9670476653828223,0.3716824809592526,0.5909202530192557,
 0.30082985081217006,0.04061814957129217,0.5254077198641837,
 0.18904432621269307,0.3220075312211652,0.24572371617334166,
 0.5722420252629036,0.053641075861212584,0.998400306274671,
 0.09882033685484004,0.7761685082647564,0.798487346345476,0.514281309777746,
 0.6540588966871229,0.5540590734731493,0.2678748667152172,0.9264836727596809,
 0.16907987633759758,0.4535633445982211,0.05295601645230219,
 0.02029202515778572,0.5774015783411349,0.28137895926043244,0.652470823494836,
 0.12671375519289918,0.6513914194938465,0.13447856140066217,
 0.33128412551901176,0.5915673827028072,0.7888898032837743,0.8922659788439917,
 0.03782289759903856,0.05958882736401916,0.6632847732771827,
 0.055159230596306674,0.9656379143954608,0.35418887278794176,
 0.4529279898578684,0.257211392781844,0.029712775709694883,
 0.34852721036572043,0.09039318248852224,0.90014373845171,0.6015894229924864,
 0.7544184961308976,0.7474249953856941,0.9817479672098871,0.9383718534772576,
 0.04745313523442929,0.3915097672488661,0.3201020366352314,
 0.44672867436725194,0.13016080335392743,0.886308481630641,0.5392783765376373,
 0.4621936281870589,0.9070062045695477,0.15545181014956255,
 0.07438581181829307,0.3937865527359544,0.22648869788697712,
 0.5811769375697597,0.9746156380577237,0.34845736096863567,0.659004778110369,
 0.997988144533304,0.5916466987655805,0.7201771950091023,0.19540859767843333,
 0.6215629481587809,0.12071163778231819,0.4561686826632718,0.374980556654972,
 0.12700613047772924,0.48234513927231415,0.7138684073626046,0.650094895170299,
 0.30371658023934134,0.5087959260545795,0.6152278190644015,
 0.12692639113577475,0.2537693510301403,0.4132685786462713,0.6610583238361819,
 0.49711017341698915,0.7350097391689979,0.05657049290727445,
 0.8581309640035973,0.7442077950909838,0.8393248021750925,0.7544227908782905,
 0.1577550122222532,0.9024497879814977,0.13096406268232763,
 0.05229163629092892,0.9398097760630258,0.13830071700661906,
 0.7520319412184205,0.2326968093421542,0.47836784319261094,0.0642226890392058,
 0.073375038492975,0.23861179677747046,0.11755836365975558,
 0.07733249268233755,0.4546501848800767,0.048025116510095445,
 0.4631519758375551,0.5978055592238313,0.6165870308591938,0.9023452697152873,
 0.6477628032230116,0.11556083094460745,0.38985566904064695,
 0.9633421480865163,0.046768773045986145,0.7880746942559919,
 0.43058638055505716,0.7987349003443829,0.8631284415436518,0.6958504376372852,
 0.8057837577490626,0.4131017110349757,0.8393278180371276,0.487532342595852,
 0.1267384242235301,0.386229418750932,0.5827188526494592,0.033279557033369156,
 0.7692784719338279,0.5568866472853115,0.40026340914117753,0.3755933592265265,
 0.30871691430161463,0.07641512506207304,0.056829843050782314,
 0.2860133324484331,0.11599662630315266,0.9843224855102216,0.6012932160834173,
 0.8501878790180414,0.5744840672500646,0.21215893975094535,0.694571128533556,
 0.2350728242203033,0.2573817635820581,0.5019565190042182,0.8796439382301253,
 0.9945008846581475,0.2916097253127017,0.048680451992262674,
 0.22798379462134222,0.3682734267882791,0.0035105047515324728,
 0.719019482884732,0.029652407919563584,0.6972764392153152,0.9897619895591497,
 0.20845861053745307,0.22442572371466996,0.750536822233682,0.5211153207857505,
 0.5595082801511874,0.782421215216249,0.2307216305741806,0.17370971294330095,
 0.6406097790361652,0.17188991691904165,0.8791896551270465,0.9859445469775641,
 0.9566628739164681,0.02567595001644607,0.882128896490814,0.0523538509619611,
 0.8450497936897574,0.7536141070820569,0.6193086691373691,0.4922888468330985,
 0.644467260591681,0.00921855863334109,0.9517208644626983,0.8548068460733934,
 0.004095771851108876,0.8069925436516746,0.40643461358256583,
 0.8009578524049421,0.7103289193161784,0.8750929772068066,
 0.049723723577010714,0.412770361409999,0.33192254719883496,
 0.8885747880026591,0.8103744471153554,0.553677382771921,0.24397172036902182,
 0.957185186505784,0.6231590426260619,0.29067375220387115,0.7460499811938631,
 0.7192385068002054,0.8321275284325148,0.3854416893231022,0.5644569652642816,
 0.6438224607928041,0.6360375990251155,0.5567889595876541,0.682340961047446,
 0.8890689011219428,0.9569632159622992,0.1813086193216823,0.35969005138710697,
 0.7005409539219949,0.10854503842971353,0.5579974580214664,
 0.45629076458668993,0.30775421333106856,0.2222139165276239,0.381881081933781,
 0.739095803220287,0.8109151296335736,0.39619299429033283,0.7709743083769841,
 0.5538257265792387,0.5107113824688887,0.6230841054824742,0.7092221682084735,
 0.3263626616561309,0.45262318720577743,0.936325634325579,0.07780930352803561,
 0.625579624581603,0.7333841975944843,0.12837489568078886,0.2521079600834425,
 0.9488691861096235,0.3643856490672458,0.7645889492721659,0.34255269188929127,
 0.2191684215702855,0.7580815443977775,0.24684134902665755,0.4674444931197391,
 0.06701623267079704,0.9229655240918396,0.8540420097846586,0.5331820772833472,
 0.9006265656607997,0.07247127370819983,0.05208906658979462,
 0.6131133290233133,0.6442872486825544,0.6085352318072874,0.0650184519519269,
 0.9838041114641,0.03536137030503428,0.2905234880040376,0.6801398928040354,
 0.39190587911363084,0.5741514947855243,0.05729026752610178,
 0.2762410506375922,0.18179332146763894,0.6507292732084247,0.9290117839418931,
 0.089699018899007,0.08143881364717631,0.08076428817746961,0.5555221308922922,
 0.35992063814209874,0.6948526432738261,0.5666662807242097,
 0.03560358779259576,0.01451994010486568,0.8214892866886729,
 0.36548259754050705,0.5441503920342875,0.9108631759953298,0.6780698283349136,
 0.5545844964442159,0.07663032255543156,0.23857695300835813,
 0.29759264587920975,0.2633078993415908,0.18273487122590348,
 0.17002566590909318,0.2690310075232064,0.8678063804179538,0.6238820645160964,
 0.09120282205430685,0.8216516139305023,0.37317884206315843,
 0.3723970338540098,0.19556670082675143,0.9764889167724485,
 0.09782959497233468,0.30617311374778355,0.15358815161338712,
 0.9255328109753339,0.42863341386220566,0.3946310491525622,
 0.01764913479441388,0.3142127256975611,0.23887761914394867,0.320222116380773,
 0.3198497519023489,0.7508872933943713,0.6014386134527234,0.8508138192846201,
 0.0069407282627574896,0.1181024603194395,0.5263098748877241,
 0.7384301383803473,0.5766495884985678,0.2816993574569038,0.9616612090050963,
 0.3653423626229829,0.4284896404352164,0.4462607893889121,0.1039529620783699,
 0.46870513560856253,0.7540415700292149,0.23484856419525402,
 0.23182555832246643,0.8627396214905458,0.569613425323469,0.8161220253875826,
 0.5054556121280602,0.8500638320536139,0.4975415663841549,0.5426272931081244,
 0.8156595186262797,0.6215370947580166,0.48489655553292255,0.5649114729403203,
 0.2168210597140764,0.8730098499561684,0.18177195399112733,0.8348772691858093,
 0.19515883431549697,0.47362366311689186,0.2737684648948657,
 0.20662609714366054,0.8750168586527047,0.20318556988376169,
 0.11981886996791591,0.9953788502004931,0.2408350056820967,0.9084446733071132,
 0.2600500291308946,0.5903072054841922,0.0023846750914477965,
 0.5135974519780143,0.8318794687122623,0.5063450891168576,0.6435112154716021,
 0.05503738245534351,0.20236939911059182,0.6779642609787063,
 0.8645636224137104,0.6086404835868245,0.6355425228146142,0.8366761574188034,
 0.8864040986276138,0.16677333743400657,0.7745511264409972,0.6076154375977905,
 0.8170518290217819,0.6242935244855159,0.3705661475695159,
 0.019888344329630825,0.48002768827773934,0.9183789313886412,
 0.2692012068852405,0.6184336352026675,0.5039675043545618,0.8231686981453438,
 0.14250235836691716,0.7068668400980216,0.011753087236047621,
 0.7259362718742458,0.5132077400561975,0.9881468382564365,0.6342930444857304,
 0.3010817793895716,0.6156833961919858,0.8222281041895277,0.07069808999484117,
 0.7704399858197241,0.6927151548372972,0.9240354676811477,0.1519637008804231,
 0.6070916516911408,0.9113138460929394,0.6759002924653393,0.8913441786844762,
 0.8589936914581603,0.06495065549004542,0.6900342726966385,0.8108935370352238,
 0.19859789019413454,0.8971087329521241,0.007780897802480591,
 0.7088222872403397,0.8132960759133013,0.5869464959873396,0.2155564950151203,
 0.419958352671964,0.243268053394032,0.9008581197996605,0.4665542644614322,
 0.7051612003583227,0.6858915938169607,0.2525651636171329,0.9357813834664819,
 0.018928434949861517,0.23816681843572485,0.7267804315675997,
 0.9110679166287866,0.9098858244285198,0.9536403703846378,0.8352063320888897,
 0.7292216437625567,0.046945712810376516,0.32914999464766614,
 0.7338655988805017,0.10302748969364572,0.9461462710895843,
 0.18572771919496112,0.6921059577487243,0.45887203958262124,0.895827461590368,
 0.45993407322482793,0.6785580555232712,0.40184773783861194,
 0.7948571421264511,0.3588339605232316,0.8527009233163705,0.24232376739834127,
 0.044744327730517064,0.6721910991251738,0.9388092887951933,
 0.36371089428965664,0.7153374106541942,0.40317119360225373,
 0.18349701574043842,0.6617544371407902,0.625117730951472,
 0.0017809640781314418,0.9763908665469514,0.037554713075221846,
 0.2938829391199248,0.7365274832076022,0.9138394744468676,0.4283608553601458,
 0.7065860811380026,0.18701129892448343,0.09718065838177403,
 0.33457006568722825,0.36080948429379545,0.6298619720169141,
 0.38082874149632273,0.04596111242256895,0.5634315011409062,
 0.28497420005301255,0.26760467012015476,0.4600951253727944,
 0.9271172602679818,0.7468780097677228,0.958977757815442,0.6075278966995226,
 0.39420063584870824,0.8701854436778611,0.6353767421390207,
 0.13567949957642844,0.503473519005408,0.4193282003462466,0.6022989300392873,
 0.9260491085743042,0.8984000910194955,0.7973628560959024,0.30323862720519035,
 0.8830293508900291,0.014729318418178927,0.41631121279064676,
 0.6536831259629813,0.7884216265256694,0.6277972258317934,
 0.051273551378565596,0.3791693121087635,0.7465352206088103,
 0.3614473514455089,0.9048303729801677,0.11253582726944311,0.3088977162036983,
 0.9352353495921248,0.48626765536664707,0.6540250968686573,
 0.20840807080550872,0.1822247563345536,0.1264669316247563,0.2540453610694383,
 0.24590850473160197,0.34233111319889353,0.3865302537842845,
 0.19933280137609133,0.8645317364637422,0.9993833502048284,0.4226381038262854,
 0.7060596226667024,0.9828567542756995,0.46958662005147345,
 0.42588583560256776,0.01388122603277353,0.5991295520035219,
 0.6308717005982546,0.0754554624474949,0.8144186667117306,0.29678143896774245,
 0.07384518124360362,0.5430688452884536,0.8578653488731749,
 0.13376843085957235,0.9204787976561272,0.14625935257011347,
 0.3944501596742799,0.503240067264387,0.32182278080229554,0.05184874892304725,
 0.6872215325921057,0.8062758693184465,0.6845644398618288,0.6424406056212165,
 0.3587035810634295,0.082343270609046,0.8027176710917454,0.346679821317589,
 0.9274120236441256,0.5318577733807378,0.9534202885609295,0.9202583625069338,
 0.8916420061360564,0.7218017085855485,0.1274426090614892,0.3567552379757961,
 0.5155843691598447,0.03957633365876556,0.03520280442407531,
 0.3289050867861877,0.7239947782891658,0.47471394524676924,0.4492523276221122,
 0.13753021038210345,0.7408051465721919,0.4128302863851747,0.4384801709509534,
 0.258975294932813,0.13097280582900828,0.579245988804144,0.7587423853936076,
 0.37815375340871826,0.316785115991458,0.7169229168392718,0.6340081673890582,
 0.24662322140501303,0.6512569190760527,0.3983410809888851,0.5635671862727535,
 0.4633113757123952,0.6891437925138678,0.24315128602603975,
 0.37079796323717484,0.5141907485897568,0.5437796773691655,0.4677965782296789,
 0.33637620951022995,0.0948427936509314,0.5523804145523669,0.9048698333173856,
 0.44439949718070093,0.21470625782838704,0.9767353820494503,
 0.04740146424155478,0.09685538889528555,0.38539637370020685,
 0.6975639691040127,0.7877858738999554,0.46519719717890917,
 0.22160294001549286,0.28915035953749213,0.4867362650179837,
 0.06288270439391042,0.7957348020337295,0.6357267734656993,0.9927231595553752,
 0.7983904886263558,0.9250325005515428,0.32067246342645106,
 0.40608432244568515,0.47172564302940034,0.40306647584142763,
 0.7559256601127167,0.5170534005675554,0.3015382711697492,0.5914070703214626,
 0.9839553633489793,0.001563946266848415,0.9672469962200894,
 0.8174088742374331,0.45965306390706806,0.9064384577675147,0.9483451062757312,
 0.8635775172100314,0.7441980835263101,0.7841608270512466,0.4105424039456248,
 0.5268932774671389,0.30174367917240597,0.06620789457207155,
 0.9386007688268935,0.013283389691322478,0.5674935191261979,
 0.3861545352066158,0.7221141598603791,0.7815585635271446,0.2531778826637697,
 0.9875253598239484,0.4871328085113906,0.9426252822422801,0.9130164013463004,
 0.8886846575790969,0.03494092937518234,0.6485498977311317,
 0.09978721322809614,0.1969699440368109,0.7446403390901127,0.6955059493527337,
 0.06646609112005908,0.9782510237789486,0.41476246736484024,
 0.08396686494818362,0.49676219089891804,0.6454694871625671,
 0.08295999856502811,0.8636656387147024,0.7107048991988651,0.7997796010338373,
 0.7072103248989808,0.2984280183340452,0.7183366385216686,0.47609147711848543,
 0.9643188561114995,0.27856748462413694,0.23480613092728675,
 0.11040474197038264,0.9241194784978191,0.5487595094670834,
 0.20085663530244502,0.8379479467369206,0.10732973631047849,0.747529050017994,
 0.2879719701213056,0.9160712080685114,0.9792638489202703,
 0.0076378184660557125,0.43877897267945476,0.08657265375853496,
 0.2900282433003448,0.8545240663229987,0.1130651770017298,0.2458272749947289,
 0.9011000667709169,0.7409295750813519,0.8316226718043954,0.597053061564309,
 0.2486715232932205,0.848473775545647,0.6239442051699444,0.5419157912142656,
 0.6593073714804757,0.5131155258116138,0.8719125500837709,0.2886049599157846,
 0.24627377351896163,0.031688113144598185,0.5506216698315382,
 0.39871828969772194,0.402875141899643,0.02121732570382573,0.9477061948380081,
 0.7662237756759339,0.9756982941248422,0.7087602297452642,0.245867798424953,
 0.4803399651541508,0.9759214301280025,0.4068322506564219,0.7217627987319031,
 0.13006717173150018,0.9186851185146202,0.5530342685422981,0.2232196964749008,
 0.9980975099493126,0.3992388459436962,0.1832262627735991,0.24896130987253495,
 0.8462733801076774,0.4356150179464404,0.020577010899598958,
 0.9855002028367759,0.32987103252371075,0.674903611450137,0.3200766104362537,
 0.6588690912292112,0.09857749087602163,0.4765375952299862,0.0412371908833048,
 0.3681146823978615,0.6027846580349392,0.4314640239340467,0.5859452350141937,
 0.4519387474640416,0.07609838331861873,0.8561289467950568,0.604439297825569,
 0.09430686808981847,0.5206858524360851,0.5801842085536203,0.5619784272757187,
 0.9837181648231463,0.882194332507759,0.7278848549841959,0.4012178306465394,
 0.6605364207727322,0.5440570585312031,0.38380921322945083,
 0.38035599819031884,0.8656264215941447,0.7800278279796115,0.3879477991270749,
 0.15810781422340414,0.6631397281295437,0.8232475662791153,
 0.16934759144363942,0.585008736756338,0.291384848683263,0.5915810903045897,
 0.6384260137843275,0.19605936902014953,0.8040065703542091,0.5814125963454643,
 0.1547176423484874,0.019087415815930986,0.07546159651387219,
 0.5595845565438886,0.23405016938491785,0.7237790584463875,
 0.31722196615377396,0.13603178783908665,0.11298182433872805,
 0.3548440708186862,0.28213843596459365,0.1511947130191007,
 0.20901496617277504,0.2540623442629618,0.9181015820619994,
 0.14947830883832947,0.02279534957651641,0.37804695829656754,
 0.5852843361372381,0.3289518432190972,0.9686346383625142,
 0.055969919705113336,0.14302759427513467,0.10980818918496671,
 0.6423516560165007,0.03526966751009253,0.7351738294376962,
 0.12054111708043824,0.5665956121758522,0.6539857453937463,0.5009094546397962,
 0.7975827850386714,0.9792263718744709,0.5297993757008127,0.36181635232036713,
 0.7277801693543793,0.8556116274594887,0.06173313414037351,0.8053214009717641,
 0.2745951278811647,0.42584727129463984,0.511412155328188,0.09920456584792792,
 0.429066699259007,0.6027287961865496,0.05057950560763358,0.15670830275519643,
 0.13126805073739212,0.21253465840868369,0.6623446916702013,
 0.49742341408418567,0.12143003085175619,0.48468475774592545,
 0.7422042710677765,0.03290692712101473,0.8855498291274734,0.5403691837301681,
 0.8837634967685475,0.6813638121879306,0.6829042674179135,0.7017680687823139,
 0.11075232223174925,0.16122943932560807,0.011922215275111392,
 0.441017824776221,0.45177206865150743,0.6063200080522853,0.04970121180595244,
 0.9383377683382156,0.5808492463528452,0.38096537079058257,0.8720395030966193,
 0.05353493864415115,0.5214838495493537,0.11101296571847918,
 0.0025354384877995795,0.30713556734083425,0.7471444227456625,
 0.12025069890360662,0.039537069094853106,0.3398935616275359,
 0.8499957861489146,0.7635666096953617,0.7929226300881753,0.40965783877667095,
 0.31017596100210465,0.10059892373589163,0.8153365486825058,
 0.2918753996418061,0.9439588620588021,0.7671651992983111,0.7532911400901677,
 0.23226856981733845,0.9505981953010245,0.9006650054612332,0.2916966332098374,
 0.22075372057044274,0.47900451506620656,0.816143415625516,0.3879764839733568,
 0.4529454080229629,0.47264015331576115,0.45388145493540544,
 0.2080811387024264,0.7882936853017802,0.5767892177597084,0.46966522328564864,
 0.6752476009386219,0.8008675801097535,0.6698465267870615,0.7682318282563131,
 0.38199835487580025,0.43969721901994063,0.10452756769537319,
 0.8750113726061348,0.4740961342280803,0.016571433657544876,
 0.0057035530246197785,0.4424652764463727,0.5396715875733489,
 0.4888062259408029,0.8775093653586801,0.8634535587585709,0.5272531934442091,
 0.3258596229133738,0.11996170109901227,0.7109879780675801,0.6991892658202473,
 0.25168910900579156,0.0240385253390224,0.6788185007829954,
 0.33434743892575847,0.8256723091704847,0.8880129246441821,
 0.23837870570269415,0.5536799743455396,0.1725014696877849,0.4254620708391855,
 0.9479346062716262,0.31915713641655286,0.2164993843714632,
 0.39494093735517966,0.024057769496298897,0.2508922798232436,
 0.5082256402083503,0.6503056250100837,0.38254675530600957,
 0.39266485241000404,0.7366558060586849,0.30005242062451176,
 0.10537028867380538,0.6018776720444168,0.7276439186898089,0.6900949329110382,
 0.3119177290776586,0.0813648346692557,0.22670432853051026,0.7449970363558341,
 0.62085159192456,0.6092699933507759,0.45310741152725065,0.16039372497048698,
 0.9337685227531921,0.11744179385735942,0.3719061820669979,0.524087465671387,
 0.9778931502402802,0.16467895704114788,0.2156150755123304,0.2339996397561631,
 0.43437148768271716,0.6648911643543977,0.5150959143949124,0.5308440815366853,
 0.31798218431651426,0.6010597250398756,0.05778063028811986,
 0.5097685030033486,0.1482654974217673,0.047143240360811745,
 0.7208228378081599,0.09610102531576958,0.525338596678124,0.2984463012711309,
 0.35645192695084216,0.0024797617730678567,0.1735978371208785,
 0.026887298303661322,0.6874114045421944,0.0880761875003585,
 0.9281023717346842,0.36911087443617485,0.9384715156038682,0.906497382202825,
 0.9535816350489678,0.8615871217543852,0.46192758327557437,0.4301008228628861,
 0.4768807396732654,0.11743240267820032,0.5270582536102115,0.4348378931914084,
 0.3230935958889132,0.4556450519411901,0.9920633269825855,0.25900758071149454,
 0.636942529566223,0.7692472125266238,0.32979664184104673,0.9194330334903185,
 0.2528562790802089,0.8393301043012783,0.2188170109202231,0.6140251113445668,
 0.8328289844464498,0.9865363762739595,0.35564423085162034,0.3591910026020999,
 0.9337777013126287,0.577116942729055,0.5725560423993249,0.13483318625116458,
 0.7634487765078054,0.33295338459062396,0.9517009887813451,
 0.41554896919397477,0.8340279250052256,0.33575808962659226,
 0.8421674065107867,0.0875825356725437,0.7442148780478612,0.9792009274601602,
 0.9669729703525772,0.13304910900150135,0.9371636634557027,0.5477335346008037,
 0.2840307785959152,0.26396294134218723,0.7667250380854151,0.2884311528114414,
 0.011831965459428218,0.5365361742969007,0.5022204596333105,
 0.12178179537853817,0.01859124773642362,0.009825394785791475,
 0.9465367927957756,0.5949802038521,0.8292386743125277,0.5164156960383377,
 0.158679932005251,0.0880222249881688,0.9217949557445895,0.13923270569167867,
 0.36110100144607915,0.33808290177700473,0.3439572391383273,
 0.024253638454142967,0.2795550104480903,0.8230438586690543,
 0.44782169268430216,0.8483998236178782,0.3190761971625562,
 0.37082343389088923,0.8179618227933514,0.967237351119577,0.8353883088313582,
 0.8335840631501252,0.5035615911513864,0.39240921551137664,
 0.02682162547820388,0.17222342194234785,0.29053038434518785,
 0.9532317269277929,0.3659238841500596,0.4599372694210574,0.7742803081239366,
 0.805698848946216,0.006504590374526331,0.5089612936722915,0.436294678566718,
 0.3150378893833936,0.8477800915713262,0.6294535518001867,0.5898677862671939,
 0.8697609784559756,0.49587991177901203,0.0664824783639012,0.3200997813702593,
 0.7513697008369349,0.8815891405088044,0.20000955169317913,0.7722903379809142,
 0.19347818889254043,0.561788750458511,0.36551605925519637,0.5552213479755672,
 0.6330693935522582,0.7261572092150869,0.8421983584136397,0.1593627976757792,
 0.34433638915626663,0.9065648505856503,0.09516981232847865,
 0.1096520025474278,0.3902535335453443,0.5278671509090621,0.831346557803522,
 0.7721950673391998,0.524175147588668,0.6101965885184213,0.4644902754954341,
 0.7013711333937623,0.2494822597174967,0.4415598231379282,0.6420789708325252,
 0.9440959295961365,0.5288164416748136,0.6016841089092155,0.6387276004883778,
 0.8924871638700534,0.35593119134756157,0.9861187398607585,0.6991651159743393,
 0.7349860742278082,0.3647571349601363,0.562399824365958,0.5202604861146869,
 0.7173119482450022,0.2507589646335051,0.5431428451048087,0.1578600843716481,
 0.44739184880747385,0.512167464298893,0.6809743284622197,0.17630342688691503,
 0.16228397360141078,0.6978027284045,0.28878618963497626,0.06237517420425753,
 0.0858563078679968,0.27292401090632246,0.6192137630271147,0.4456396342644169,
 0.6321001336800732,0.02440478374263566,0.2610450333852341,0.5021963100659479,
 0.48014364852682856,0.5257941752650874,0.10860615371990823,
 0.49420223442618183,0.5629777598872818,0.6602259614706582,0.5239968584942147,
 0.3663988189927446,0.7857372150076896,0.7437645690079595,0.07028176649124351,
 0.999914940704411,0.630187705528,0.536937022056226,0.6961645832118951,
 0.780172797921954,0.6459177653941773,0.5859491865878146,0.5797778625157075,
 0.5009210401663005,0.7733845566997556,0.9589782610137753,0.620381009718227,
 0.16824933105239614,0.8740588669544017,0.9626142627185342,0.1718893050107837,
 0.6385184591372354,0.6425853561909041,0.8037789053885619,0.3607263150690234,
 0.3847981661733093,0.5062582561168376,0.9480620513241141,0.14900447623809876,
 0.30448291221819546,0.1977516695754976,0.9279560979264614,0.3399942505320963,
 0.6145965525766961,0.00864514171414632,0.9583140208620331,0.7665175460842772,
 0.3199860561307649,0.31000114065507856,0.7493184738993922,0.5735176943780725,
 0.32208684632278894,0.4825563443407975,0.7334983588398873,0.9640010695331606,
 0.26257771907894567,0.5908904275498327,0.3779466577991356,0.5711140057704214,
 0.06843652554930024,0.942598042096402,0.8539380078906448,0.0743500979573487,
 0.3871058589328531,0.23522554710737964,0.28744173648969207,
 0.03324278963675387,0.1805279042843193,0.32579859167935743,
 0.2252919173717005,0.047597731651647934,0.4321457315622884,
 0.13905770822186692,0.354521864253433,0.3106216004049078,
 0.018101001556070417,0.4879013325714885,0.34401188942993943,
 0.6806394522194575,0.09135443164627233,0.5328128779408667,0.1120851939979941,
 0.32413044987359796,0.6128542746739662,0.872320524909163,0.42283997257233763,
 0.6014082671455698,0.7932443218518008,0.6400326935105544,0.3301447216353145,
 0.24967200073631013,0.5452880167418185,0.3028881477281856,0.9129097379622421,
 0.3078848468668032,0.4237840737639543,0.331945369959486,0.6472641738892136,
 0.536259004014146,0.44526860317550687,0.7815105654983598,0.65024488898151,
 0.6771869626455761,0.9672208535516205,0.9597105263758123,0.4488625259360124,
 0.4299920283549499,0.48718703263370433,0.464765427066983,0.5280057350631078,
 0.4968035702125049,0.7557115515780162,6.246542520849108e-4,
 0.6451461709733018,0.6867374978725378,0.6444552048451324,0.05885662689696358,
 0.6037231924950525,0.4628658307267569,0.23769657714288084,0.2586391211261595,
 0.9150736579876367,0.04234450809830559,0.5531373355731418,0.7401642138549821,
 0.2164617659099739,0.1375120474352245,0.9950269134479605,0.32669061988225157,
 0.7268063366790757,0.8464174043299264,0.6053950068807993,0.21113046572095817,
 0.8188381883562511,0.0424458987087345,0.08871558097093135,0.5522031576011189,
 0.13667581148913777,0.9027745951214005,0.8921289011146776,
 0.25814622786634156,0.025971959299130654,0.5300362106026969,
 0.8209783904199947,0.07874182866186419,0.755765011189764,0.3519478273182224,
 0.6298472257952613,0.731042299521226,0.39927136453053336,0.6803816215370848,
 0.6412686414862869,0.3104390201309898,0.24956562866662846,0.527976232211627,
 0.3598058964230637,0.8154748523087217,0.7665035555203409,0.39947280610121627,
 0.36648831233733026,0.4038799834552783,0.8677983318603785,
 0.19032294022920915,0.9868884086922004,0.5472532956346926,0.5222803872773687,
 0.7170926311612499,0.4841009055121558,0.48700818152702796,0.6683598660146814,
 0.5071781180637908,0.34962829604082213,0.0017841392031626713,
 0.47587783973120645,0.7205074090580929,0.6511337106789588,
 0.31904228620851904,0.9880279082984904,0.6278192555675308,0.6467398042539179,
 0.6170017236724427,0.27802827203965097,0.6621512192633979,0.8729838032508063,
 0.4775947026209675,0.6131150208499094,0.23316461706747915,0.4689871632090974,
 0.8651892093007335,0.47268908151504874,0.6044574577474023,
 0.06942354976800758,0.44019085613986686,0.2614034849905582,
 0.9196945663330773,0.18513182225302272,0.7852169732778935,0.8886858453552978,
 0.32568143479890777,0.6965117164561042,0.07965177556487013,
 0.8088034294172838,0.5043527239771104,0.338288977760856,0.35147422236544135,
 0.680338388965063,0.3828752057907582,0.6806260287695757,0.5824922202772429,
 0.6303090541462578,0.11485318689429314,0.2944243987101145,0.3542219007106878,
 0.8077126228468757,0.9867238210415511,0.9281440992358987,0.21376626966803824,
 0.9386410102206576,0.15038095613816338,0.6303470663747668,0.8501501133398183,
 0.21738394257998062,0.38485369898299104,0.5955676729638577,
 0.23451834227013035,0.702723967360652,0.45301151797384787,0.5683428734861257,
 0.5900890570402443,0.7593653813862621,0.30095782150866734,
 0.09056833667813691,0.6659833016899157,0.11815838445345772,
 0.03762172441361655,0.004733352994445883,0.3229616405508997,
 0.1937379713194416,0.03878071151626117,0.5783136143845046,0.5173790175536972,
 0.18670263427787448,0.2853022203391692,0.4948081229437513,0.1582626475209915,
 0.8536389202294347,0.9406253959881149,0.35137012111029065,0.340347914028721,
 0.479023985392675,0.03690693885979146,0.6249438420875774,0.916259776783038,
 0.29086345485932164,0.7353686670985158,0.3914359886240063,
 0.48869509415970647,0.5134935178937317,0.26126370248432573,
 0.45003067864798485,0.4989488107838257,0.2199051381118422,
 0.37523377880815256,0.916369915531123,0.5439453854937346,0.5155766560751827,
 0.3516760570582047,0.9044528124379586,0.14596501663831596,0.9089639054977366,
 0.21385073280867384,0.6307733279377341,0.36094255324569413,
 0.9899831589183533,0.09493685283729336,0.8183038333388917,0.2825722029949558,
 0.14611653188146645,0.5392461136437311,0.11565883905023155,
 0.7030005244414381,0.8065914545618248,0.6166839577683176,0.26091355884553147,
 0.44836912133726825,0.5264621240975346,0.5037974965408527,
 0.27100780741911046,0.8293893899701996,0.7663050249075987,
 0.07988916761764375,0.7818215709933641,0.1547916780641414,0.856365277868415,
 0.30686187335179727,0.7667867613496891,0.7780054320055035,0.616091964381691,
 0.3117464434268228,0.26283240698153076,0.5171583343367149,0.4342440662881253,
 0.640382053055266,0.2534157983226566,0.12719029362084633,
 0.026822252134277358,0.946887773965811,0.31825052694942424,
 0.037823367103742345,0.4107973240486038,0.9009926629138437,
 0.8680559013637332,0.7230014424986958,0.056900424635647706,
 0.4595120739998372,0.8643984273778894,0.39724838901640247,0.5548687076049651,
 0.7499256675922913,0.3354352700877601,0.8327580192430969,
 0.035591880376948826,0.36700120359541333,0.9010954503062651,
 0.2928571940315051,0.41413012482766187,0.08814232752944007,
 0.01473048411863398,0.539620069387907,0.16941113060963886,0.281752413315957,
 0.3758847737032913,0.5632869028416144,0.9152406743426973,0.9892999959700949,
 0.7758817459903873,0.9964247581379141,0.5404515221853756,0.6957977779284116,
 0.6471510480822034,0.6017723625051162,0.7427826638081272,
 0.0045737423934196775,0.747956437103621,0.9143184860646851,
 0.6830415456670967,0.6524631548666815,0.598654374727737,0.768488359148157,
 0.8322193693683637,0.5813569135152545,0.7674908850415151,0.9319195680853551,
 0.03482126003394859,0.12945885179908867,0.6689812530979319,
 0.46879481969940684,0.33351733451703536,0.17192512726194575,
 0.6356794064014883,0.4998605636301129,0.5622503424548536,0.26762045141513846,
 0.2672708073423612,0.17579573735972942,0.24782358411232364,
 0.2531123406927994,0.3643827110046479,0.5015213011962623,0.5347741965761721,
 0.1652471202641328,0.7715478914840425,0.4080549084390477,0.7507051076294138,
 0.9493595430233769,0.44833893463439245,0.3777910881130425,0.9463498461474098,
 0.42198005503350444,0.9115239974994442,0.5286823612085935,0.5272871634036789,
 0.24886978984565,0.5538235738268944,0.9320598144461935,0.12641263758893817,
 0.2672593980740938,0.5180399636537139,0.3564942219460425,0.4734678048334493,
 0.7368977659787955,0.4539654885890413,0.42444990535280436,0.9755194632788546,
 0.5506282943779398,0.15237995463885123,0.5083934383877646,
 0.00885550512945521,0.12745641342305858,0.6028891035431706,0.894983338348307,
 0.45668986720312443,0.3486243059869274,0.9084458619730822,
 0.16186672816315006,0.7427590814563819,0.8935059722221017,0.8485443024268075,
 0.2765816268128167,0.432693127514469,0.3054491678111433,0.23282265594696583,
 0.28979820521155175,0.04988607720568883,0.13085364528206678,
 0.9444065264662803,0.0900759483004836,0.48684677650380126,0.8528123424014427,
 0.7552404572561933,0.9563276257505956,0.8930786284862731,0.667681767635675,
 0.15469717195129884,0.6400856312926158,0.16845901655148143,
 0.5979987088762485,0.2671278403725251,0.2901167596096391,0.9893597163020196,
 0.09607383653936818,0.6169446387551932,0.7401928583407053,0.7941410434742061,
 0.6642058634015819,0.9952140190577972,0.1039211428170087,0.7203374272900595,
 0.5739178849879631,0.7479453930506538,0.35988206645454734,0.3238344644693236,
 0.5072739108283906,0.700362223481286,0.765579074025239,0.48808337016621584,
 0.0686712693402316,0.1873223045330371,0.4145929723341464,0.27152448871094426,
 0.8448347210222631,0.84942504819993,0.7334177286474127,0.3846577023857506,
 0.17374623437512626,0.10627472506735436,0.3204143823861293,
 0.0032474883696359536,0.22671400059604252,0.42133338636028306,
 0.678028466846563,0.6567308797284384,0.4344169831991552,0.7212649178006115,
 0.8336140876038481,0.769780759981243,0.12181888303879596,0.9463096041930329,
 0.3173261193155862,0.23244481817645557,0.24446084261980494,
 0.6169708761624858,0.31260655351109023,0.025503611544399396,
 0.6383715446315248,0.6728427939101741,0.8252721110560076,0.45106638945281374,
 0.22606896824436007,0.5088402365541138,0.33006194239859976,
 0.08962392894660032,0.968886190771644,0.6933782267688019,0.07780646640033972,
 0.5844358688908087,0.2685689991222975,0.17943555075175033,
 0.43326414459905604,0.9545372378676473,0.4820561970567211,0.1362108093798,
 0.1615609604191327,0.9378415480739657,0.8571431073460355,0.18907417353502665,
 0.46768296084015293,0.0047025416462276315,0.36825571602826446,
 0.8040774942403748,0.5142638742716625,0.8853478838413954,0.9715432847005947,
 0.4509562579420693,0.5895406694628565,0.16779918683242856,
 0.13988046998128212,0.08110455398399052,0.11472322881972674,
 0.9349331041614883,0.6238180856769855,0.8500415180843542,0.8657835026673145,
 0.14277999527916563,0.24647279785166054,0.7481731005317669,
 0.7050587445447937,0.19350776835733152,0.3960649684022429,
 0.14410122872252573,0.474932982816058,0.1173622721297356,0.9197206070794571,
 0.47307875598391025,0.8026443338037109,0.49772400763564817,
 0.3249033700700271,0.8619936732509247,0.07480864961442457,
 0.05030275478670376,0.9368642095198211,0.2954696521224357,0.6604256284562227,
 0.943826843123716,0.2872554751007079,0.6043588692173381,0.5306814630620166,
 0.7412274582021621,0.30888111509297467,0.8983352176786786,0.2441004545952914,
 0.8683457358440386,0.75577286041981,0.574263019186912,0.7897489525012489,
 0.38093895133318656,0.726888860128029,0.8471891988528951,0.5739909869142943,
 0.6262939720001118,0.28933403060158014,0.9484701910737837,0.3542571734540505,
 0.284209737464579,0.492427081537669,0.592927175773223,0.9434392859759788,
 0.6062208575035734,0.8265250568956126,0.4768010148717161,0.5759935725874137,
 0.6171178531550724,0.1753176650658701,0.9202400228556986,0.20813539715525597,
 0.725588434325233,0.22682738617741727,0.44440596502745644,0.8757580112360465,
 0.36741291093375916,0.2496149666568015,0.2875656389856398,
 0.10107262408849593,0.7033968804214475,0.4535082177974804,
 0.39505771215322016,0.2802212115824325,0.3848763267863876,0.8880551703436119,
 0.6390749980937968,0.8262422960501272,0.2609922872884749,0.6564650067653462,
 0.6681670513138225,0.6474105918573099,0.3280580678662588,0.21437441977356886,
 0.8025248697478481,0.3463437963696365,0.5319472249740894,0.42566427130521123,
 0.07198605790312163,0.09583581400495078,0.9843189746787285,
 0.7511123850103221,0.4195931136865121,0.007052729732633156,
 0.8021479385228029,0.18020876491439297,0.39687870088527955,
 0.19771752991428904,0.04548964570824343,0.9542824646787968,
 0.8274112505767165,0.8073845301273121,0.8487710225064392,0.7394479854891007,
 0.9571506574519983,0.24592167372453,0.5349153511348119,0.33056252228427807,
 0.8118607205886669,0.1872108246479215,0.7940809798366041,0.837494535839158,
 0.9397620540246264,0.31499958276131745,0.3476444044981408,0.818349693793794,
 0.35512290561024507,0.6066089450666693,0.023321070245680087,
 0.8045128490790635,0.5750590300252112,0.3095424188020115,0.7948526770384685,
 0.7711101529331505,0.5407562911365803,0.24761056729835318,0.9161729106562395,
 0.8001798975925308,0.0017471127282424614,0.36203967691151373,
 0.35162755556563374,0.34333746946814836,0.6445346793145227,
 0.8838738912201642,0.2329977076508425,0.8697836432667528,0.9523349417747086,
 0.5788672340846561,0.09832173036607983,0.7459134822380898,0.8998240991674267,
 0.8044211877478474,0.3025993212998307,0.5265320957643478,0.36418811522888717,
 0.1267411898351294,0.8467707911185715,0.2575797998704351,0.16874821876352242,
 0.4905779994559025,0.5846834725747578,0.7681255173588121,0.741205916619332,
 0.6559810688752266,0.49387053652064317,0.4548215740997221,
 0.12552891823996015,0.9769288559160226,0.9831603958215336,0.7086432282698878,
 0.6905264681781296,0.8200857199421027,0.873490111594932,0.12554241864745275,
 0.694766374091134,0.7139015694084719,0.4300868237456673,0.35155509313794053,
 0.39767385998367133,0.2064690757197304,0.6807263841075804,
 0.011004777577338087,0.02692839383545864,0.5000068355756008,
 0.6912390658203765,0.41276779557346577,0.8775118223166034,0.6105240475598732,
 0.3593379677978643,0.8449423204707373,0.41074867723487307,0.5786556626235906,
 0.49052495540060437,0.18102928427485032,0.6274833536767307,
 0.4509605553049443,0.8422801697303084,0.600778203892237,0.46019109500426714,
 0.76622129768706,0.7396858288582759,0.4675172092927019,0.9995277504153729,
 0.8197587618420422,0.41956820681174734,0.9231665025434439,0.7651677754621847,
 7.166236914835089e-4,0.7508896801012978,0.792787109441223,
 0.21781724163496397,0.18716784987174984,0.020915238356127408,
 0.5917169192436874,0.6782224128987133,0.8939449523643924,0.2369967906960262,
 0.030941359010451586,0.8921266315839574,0.7467858428918855,
 0.6614302588311972,0.53088855039158,0.9908023402350452,0.5895316692543223,
 0.9793915687818632,0.8770027397140121,0.7038712998461584,0.19292284345627508,
 0.9817301734285794,0.7548556259775272,0.44328253499966597,0.532712493958132,
 0.46785947926065086,0.8108808969327933,0.9085080992509507,
 0.023628111573421817,0.7735776570491375,0.3470284530382456,
 0.9799516217032413,0.9993815603097438,0.48686226829044676,0.4247105385747231,
 0.3118988092145175,0.05748075507501138,0.911860330317102,
 0.003242570785026766,0.6747167864751922,0.37190270714963425,
 0.8764405801966769,0.5788967641879477,0.3600115349471694,0.20968814214243914,
 0.3976932055335791,0.7653619397329676,0.793212976910468,0.6828413725377437,
 0.2687142652677994,0.8342961999945249,0.8609109360691589,0.5970482913708192,
 0.8198154605594452,0.6540423221602554,0.4168097565033694,0.5346445285833918,
 0.23100626126662438,0.3173477179411527,0.8341514245522943,
 0.0028669894135084917,0.9027061768651945,0.01034402855193306,
 0.5526031675357376,0.6978399990470844,0.7150809562635793,0.8830644339285392,
 0.8130463888011842,0.016567635710847406,0.1432517942154743,
 0.10600565670470374,0.565826090784299,0.005078504068530942,
 0.4964976147293656,0.574577593453875,0.5434973186295615,0.16631161060953115,
 0.111378303352487,0.7313065617599981,0.5036838527219198,0.11170842665263458,
 0.27521696265113305,0.02680836421219468,0.9824916747971024,
 0.6095300856026425,0.34115768737876606,0.6524363853933177,0.8164348077785459,
 0.7148697950300356,0.734308628731988,0.6434979601875236,0.4124344651863733,
 0.5995066631496645,0.322447597566789,0.5392513889514956,0.299292662438726,
 0.4537711153906465,0.9869827766723848,0.8074558340613841,0.48692282384651375,
 0.5190937167480261,0.3428849353509511,0.9476645967691927,0.15947652030685855,
 0.2079348572950943,0.1416888755766308,0.33687738513216425,0.5124595146499787,
 0.3475445779274208,0.8707351065360642,0.249838993087232,0.8738038174653583,
 0.3254927825474295,0.6857087084316053,0.61461966072977,0.21667003884442915,
 0.8356258093849914,0.36156470863267887,0.20771035023239848,
 0.22174937563944574,0.43582089653030565,0.8851970950026079,
 0.33074112651532306,0.2527678299540802,0.20194399571146215,
 0.8154274272831089,0.2121960137170964,0.19413634885492415,
 0.07746607231490876,0.9525943401779311,0.09426281732957953,
 0.9493838383936414,0.21313266704587364,0.28573123756783336,
 0.3433891195688511,0.3720351176978911,0.654563881619318,0.23306827462180513,
 0.9111937941634302,0.6780295044372403,0.5187305930429564,0.2744735429912357,
 0.5668120947798194,0.15861227913003306,0.04401414520372271,
 0.8522437885234786,0.28629091979086196,0.6654045361913759,0.7583900010459065,
 0.004228649503980053,0.6654960221883572,0.801057305602596,0.6023063759123982,
 0.513035673496554,0.7647260979918955,0.05477748060033871,0.26596863364609624,
 0.793648960674104,0.7097617304662046,0.7573473920777174,0.5531720324511146,
 0.2817489965905733,0.2592708470556545,0.6673034889901336,0.12016664870357818,
 0.9973592109135756,0.5792153291487907,0.0698150025284544,
 0.006849265904470547,0.5538817360344085,0.7572412357174003,
 0.3960214099816046,0.26188086222750817,0.3815835472675596,0.7127150123087481,
 0.0013547993281058934,0.9939445422447388,0.9299027553555066,
 0.5430756992458049,0.6707151648121714,0.6764629810178633,0.42792093699096057,
 0.050782753675285064,0.7296189340280035,0.5486013820087197,
 0.023495165532643525,0.07676113688159825,0.1825535442464269,
 0.8735484206406114,0.408484715965704,0.10595341314467244,0.30506353226600214,
 0.8027963042994438,0.6013470607817738,0.14638137707636878,0.6615311434848904,
 0.4294085815983807,0.46613512436123095,0.08022896858582806,
 0.5512929359979685,0.5752870768654066,0.9029825249125264,0.9276239210831817,
 0.6880035130070681,0.8771582244363371,0.309225322951146,0.47897259896386823,
 0.532923455756718,0.8822372530016418,0.9650360589790477,0.8768195957465698,
 0.5536921983130749,0.04009357062154484,0.3018677713534629,
 0.23257768076926144,0.2768848447253983,0.962107069394658,0.6190382266379542,
 0.12189094928134947,0.14138472365633936,0.8042611530352719,
 0.2787848703160357,0.6668292595337242,0.25096714666142317,0.1490372551662814,
 0.6930664056781939,0.7607821789431185,0.30030267656504683,0.9041546070976615,
 0.07489322755672068,0.2823039884634171,0.1784919323614329,
 0.28164397886906023,0.528357686844931,0.7821873472568099,0.2504245260827066,
 0.729260045595798,0.04779429395403911,0.0015979740949507715,
 0.2632153794916715,0.1939144025519588,0.258873054488579,0.6521966031798926,
 0.6367669701291144,0.9281782921046432,0.08106585533821031,
 0.06865760923222097,0.3097730748666321,0.5713302278082965,0.8986791194647351,
 0.9188617212423466,0.4691201905845144,0.20902310444574446,0.2857417598114921,
 0.5171200817446474,0.4728081503668574,0.19415819505879306,0.7707001203844448,
 0.1876191459464418,0.6069483768387367,0.3659894558645954,0.7003356723200562,
 0.8982097427497461,0.22688129135301782,0.051212064414652536,
 0.6617786537891482,0.17483112562399816,0.5642671269423535,0.9020470788737345,
 0.2645480474980002,0.6342583495312328,0.45833802054900374,
 0.47952970116121474,0.3688060245040773,0.5515718706858375,0.9011863492548956,
 0.9414259387428034,0.1549681857606724,0.8422005173004401,0.9010068336489161,
 0.0225428102412627,0.1817911505300639,0.7017008922677149,0.49345192438265306,
 0.40841990240597803,0.8526027778473786,0.06685553262042043,
 0.032491218172464476,0.7885010311888021,0.8924366565533869,
 0.7737932396836874,0.1382860733010962,0.8430675708666838,0.4462804065236976,
 0.6666971150723263,0.9204866797840034,0.4568493004086187,0.26279413104531524,
 0.031352679015071994,0.6795921999861834,0.8273396188853188,
 0.8169265153026303,0.04592979259702057,0.891175172906158,0.6229894591327441,
 0.2080257044021322,0.327244149393518,0.3405486141381906,0.7129252456491991,
 0.7219998777294366,0.6270139775144332,0.17445405386748358,0.110617959909867,
 0.08086823679233535,0.6882675647330145,0.6150306999769328,
 0.07868238440280817,0.5393034256611274,0.39646332214764113,
 0.6643259416502366,0.8546674895518648,0.6610357024746318,0.6959155382332511,
 0.6043878955943339,0.3269790129760137,0.17545443344262523,
 0.34428897917993273,0.17239344266588752,0.9248524319156216,
 0.8960042043299736,0.8872811891363893,0.10287603336557849,0.5210142714263639,
 0.5183203204313429,0.4921998483545116,0.2689497342888578,0.6645221629609104,
 0.33312152197671263,0.8636837941226272,0.47783752776628063,
 0.9442404029423076,0.05104432699710226,0.29532688521191197,
 0.4187142535726017,0.37126616404967827,0.06757509601910128,
 0.1157261784843121,0.9254974545722199,0.5780098934706335,0.26515279185135,
 0.5166666304479883,0.7120816663368554,0.7501868955860322,0.30258992334808377,
 0.12210789819232326,0.6377604573684076,0.18252010781061556,
 0.31276795171188443,0.04715950115771095,0.7971481101868648,
 0.34213727551406103,0.1353813893116318,0.41380503687582826,
 0.7390047080688491,0.02497481148241365,0.6221719425773826,
 0.14115359958216878,0.9448631073658768,0.09885701368807864,
 0.3366408391557544,0.817144529219706,0.7406003075404073,0.7259783054596388,
 0.5031677329417967,0.18364716330720254,0.15654445036314824,
 0.8661747513392827,0.7629537254883486,0.2785415989352731,0.7024238743048632,
 0.6268043738330955,0.10588128907884409,0.5422810203926232,0.4118052952697707,
 0.6137020650574216,0.005669977744277133,0.12398724677802198,
 0.06263690747997208,0.3307059039947092,0.032280400773529894,
 0.4789996354516629,0.8226977239662995,0.3883755053381682,0.3322468645822897,
 0.0431896734757129,0.5362846012442267,0.7561554142516796,0.09298607655413382,
 0.5852081285104965,0.17398952583177518,0.7729944515880354,0.6569201344386191,
 0.4137539351853269,0.8874591178600023,0.8232393715043604,0.02734889858309697,
 0.1706903235758761,0.3417795664808203,0.49853335130637866,0.697166086151386,
 0.26650560866935324,0.8088898667133522,0.8375406829920793,0.7283205580048352,
 0.7171706895591732,0.4525935690373395,0.7606029212064834,0.9746715984604383,
 0.7804265601067779,0.3897892435621436,0.09375043652453241,0.7987799065250039,
 0.36055464062997444,0.5465052710247109,0.06934571617935381,
 0.3598562861620327,0.2650589586928771,0.9699137086366116,0.81755519960258,
 0.752416611704815,0.1238919940533939,0.1138929985442142,0.4487200477760872,
 0.8037307672457821,0.010654858077607021,0.6083758046356373,
 0.004569149148324514,0.03825701434071138,0.5925197421307385,
 0.6898094771129363,0.03484282915872561,0.7425297504407531,0.8985634205754238,
 0.5425844910816644,0.37413431546684095,0.6971506170853233,0.9815301096371828,
 0.662308671239857,0.28403345561651094,0.6969450184654591,0.9440658810677445,
 0.14759692071240194,0.0312997232069141,0.38515072180860765,
 0.15715340757322827,0.49681159306620126,0.6240704948963849,
 0.15469210451174653,0.14580832354915185,0.3757074776086373,
 0.7250706372117811,0.45128756181809226,0.8650662367924629,0.5959692749091727,
 0.5540901059480509,0.8957745742865194,0.18344955619029502,
 0.44471225180558804,0.409672028131022,0.621901079968383,0.2730979295691812,
 0.3889140877005245,0.4010193248328563,0.12167246883916194,0.5784736399462262,
 0.26852929664670455,0.939784508694306,0.7543454283609183,0.7152195293200535,
 0.9190156476980745,0.16303791103450427,0.8011304514111405,0.9179828484470753,
 0.47512032611243893,0.5374079276962178,0.7261968294607504,0.9792620948515843,
 0.39510365446497764,0.5563265504221541,0.594027791983398,0.85724788941501,
 0.8215081448864243,0.6356108574703082,0.3030620011356351,0.42491046147052103,
 0.9736011561669004,0.22181446688362738,0.47264597204618397,0.72123987660485,
 0.567174113774429,0.7298256061631423,0.4316121529893293,0.07984008312615698,
 0.15815537260095403,0.897768397725152,0.8109037197101904,
 0.026832351734301985,0.9177241563426562,0.37586281624167717,
 0.49804798336477774,0.015915480706015117,0.8072383067173056,
 0.9283382215601675,0.06531943899040227,0.6817920577490457,0.4738846758621188,
 0.24849587428919506,0.6014756755697224,0.4440158422053837,
 0.0011736824719583705,0.16371195984853115,0.7470448957343402,
 0.1832008482857479,0.8244497764492892,0.8512618840891201,0.7905138433514358,
 0.041593721023058694,0.3238230395302323,0.48297580888666525,
 0.7599322060990332,0.2772520076543633,0.6702613625558331,0.07439483756524723,
 0.7812131388629105,0.19664791085729627,0.33178276495424075,
 0.4823923027069541,0.9225946798248936,0.28723938690716755,0.9695567353806953,
 0.07798923558686732,0.18892444988604706,0.7056143031049631,0.490515742593825,
 0.7249975790401956,0.10307342004380671,0.5475024024292777,0.9874465480243858,
 0.7495135093473836,0.7068270381014741,0.553795880772698,0.44186182298972143,
 0.8346745550579365,0.8809492787127224,0.06696809333328413,0.9558110568850752,
 0.9881993854070665,0.6488745241163341,0.38887496629187046,0.5409567810860462,
 0.06657101808323818,0.10189156460011972,0.9774891684925986,0.989882591309878,
 0.6424589388653414,0.07270430145366202,0.6630319583229551,0.7449319262111872,
 0.4688885094232249,0.5319714073008592,0.09040326193956538,
 0.26983594988339643,0.15160597582809088,0.13138630060960643,
 0.952110999360142,0.41951582231542306,0.6277066641980803,0.7969644986460299,
 0.15966752900865078,0.15625337373199566,0.4868348816880651,
 0.025894769475589552,0.4749734810804985,0.08903596093181398,
 0.34470310803661286,0.9458074186832379,0.8158555869479881,0.5947721067471563,
 0.297448993134127,0.1560209469030065,0.08266524234775996,0.57892472379864,
 0.15926421970994564,0.5295726754724598,0.8583359314433446,0.4559320645857944,
 0.12564964476969187,0.11414705164387828,0.18164410150342025,
 0.36135773477543087,0.7013939566625351,0.537376743480034,0.8870081015139321,
 0.6130480237200367,0.4601230943955954,0.9881892147629598,0.19819452136788918,
 0.8416214055192899,0.618079848345822,0.44334332224127415,0.5062206600535081,
 0.5563768629630639,0.5582767355273771,0.8485940978978558,0.7504965681524502,
 0.3752847116689131,0.43506529944103056,0.5042131954412447,
 0.19975354672220036,0.9762465566072259,0.24262648246346896,
 0.0884725810853384,0.8073073708756481,0.10919992878484264,0.6908661078565693,
 0.0803376436536376,0.021622143454780862,0.7272689060495667,
 0.5874554988422721,0.9561529304736169,0.5106421037645702,0.3232058795676662,
 0.1001538194978675,0.2481263649884582,0.2315023121343025,0.15814995139815313,
 0.1483402459968064,0.3688444990901464,0.05941859352446066,0.5438833666222813,
 0.6405084837528258,0.6280255267887142,0.8550552581851321,0.8671534728330674,
 0.6883325646375476,0.47337475447767585,0.5767270958279977,
 0.06885126648089313,0.17925705982443774,0.3931571343662529,0.87324106259015,
 0.1556770445448593,0.3058806283528298,0.9940279734771579,0.8116336858098566,
 0.38900517529430656,0.754569985069238,0.9093989504285727,0.4561292160778929,
 0.10688328561319804,0.6574686148756389,0.14451471225516088,
 0.8591465079808467,0.9107019751999593,0.12256571835497232,0.6489585930991129,
 0.2225102820442143,0.47651745769490494,0.7608064649586206,
 0.16019864580966214,0.08231250636572496,0.1758922512910439,
 0.5524360205228953,0.15132543797818243,0.817269838732164,0.6431417965082025,
 0.3658654964897925,0.8123079905056063,0.8535503959129831,0.807628679524972,
 0.8764158112686673,0.5905711002356845,0.6625152067052503,0.21484556326172233,
 0.7074619250609153,0.2637468307357308,0.3616273984731887,0.07949090395601466,
 0.5541462212223869,0.6633859134255772,0.23267334827202446,0.756085283016934,
 0.4429569054762478,0.9258046752526141,0.5456099495762862,0.9785941031757934,
 0.8154130660919181,0.16569235130017823,0.24848713241477272,
 0.6760658285611398,0.2590078969820394,0.5376678735285498,0.4652474830205403,
 0.42182440654416475,0.28775041962013137,0.3883596664889719,
 0.004274722925093899,0.16912609428379555,0.5179142912578422,
 0.5518137796716003,0.5885023775805579,0.9123862114527186,0.4612114639418512,
 0.07084017202124904,0.1312134314746114,0.5549191202212383,0.1290347597430136,
 0.723276805411145,0.9909087351998362,0.36682661285649354,0.1541773121447647,
 0.1614327957936953,0.7450331603744903,0.6034550377727257,0.2935242687504951,
 0.6458043254888183,0.7941872449188424,0.03020020452040062,0.4664433105830891,
 0.44573128739717216,0.7505771656360589,0.5419835872211781,0.5912075466575651,
 0.938076022304427,0.27004602235978503,0.06071867461429559,
 0.12021232215755517,0.44168932095529045,0.2845659509746963,
 0.7335472350801512,0.5401818105860412,0.11970187624699402,0.7929352880407542,
 0.06879601138388636,0.5261500458326708,0.8854969988162452,0.8261315473364169,
 0.1780758328763523,0.18067758333809025,0.953193149950724,0.15678259984561937,
 0.032974195162992714,0.29755923813714036,0.8581100293065994,
 0.1614516132347098,0.4040556273669027,0.5135029786120728,0.41117064427107397,
 0.17681594616314,0.49263269907202145,0.7572856492519764,0.13941540964388877,
 0.15061555579161934,0.1876545896314753,0.6616670692664846,0.6378095608712259,
 0.13627171132859361,0.3619281123352496,0.9675465777890787,0.7910300371946882,
 0.07876697445849412,0.9246652504797163,0.461766084716579,0.9325946155706493,
 0.3585639833149178,0.4121122262338309,0.9987534083561606,0.07133472835192545,
 0.95240782065103,0.8967030558605723,0.0790127132878824,0.11726743943308393,
 0.5902160752167691,0.1806540844026584,0.38844051250872824,
 0.46019787729358397,0.878230209686958,0.34133837999747185,0.7637716551934108,
 0.9138573442697029,0.0790353241264079,0.690854093310921,0.5924810837360073,
 0.5714818163625962,0.1599549317911122,0.16311894452264775,0.6905774712373596,
 0.08880702816201358,0.4154301394246035,0.40861371155390147,
 0.7392220345010195,0.46835346686724755,0.8118930845561744,0.1459229327513859,
 0.14238694712672184,0.7889652847022717,0.29778265818600624,
 0.3879135523430519,0.2686789108406642,0.6885447045094987,0.48052192155668727,
 0.1813488809515652,0.7039580890415987,0.8088377223175816,0.3415255749171153,
 0.5792890150917687,0.13171136604628841,0.4464537676952145,0.6997012072852826,
 0.29725707909405763,0.46755242796280083,0.25856267390252086,
 0.2841195219837416,0.005345399059048295,0.5482530290494838,
 0.01062472524278224,0.18618474703401455,0.9557310297946631,
 0.0674905664500336,0.8108995365039805,0.5919939875269491,0.8899901892127111,
 0.7520743426513218,0.16614284294089066,0.7287516473029005,
 0.18868027937143217,0.13064627163765186,0.15764873654719136,
 0.5892259267843003,0.45596442192588615,0.292120946052542,0.1734997720504734,
 0.18491292328674225,0.5694519498792957,0.7398929493128621,0.5260832367246446,
 0.04127486042158024,0.3282219978508234,0.773962248935213,0.9776313026074362,
 0.6244574769949037,0.9849479195685912,0.38635907914459544,
 0.33381195859138857,0.48660084269565074,0.4124759208622213,
 0.028118007397757605,0.2794073236885747,0.08808883862599015,
 0.41048482620905835,0.49701182373379416,0.45884223108088684,
 0.5226526171326856,0.35074887399257015,0.6073894559038984,0.054907462110941,
 0.4002237426630567,0.8173347671614426,0.07483307544718809,0.4231728041500876,
 0.23584569672544964,0.47548837833628377,0.8514928162075862,
 0.17605113767374658,0.8218922752848931,0.4428020864660913,0.4793059828373588,
 0.14821691657254288,0.38151725854407714,0.6411530974031223,
 0.3234355851200611,0.4437938569164612,0.5049021164051217,0.07351144071346905,
 0.5446931366519211,0.3533044336066796,0.24006396730105684,0.6879955550545358,
 0.1470182415881518,0.09160819880504056,0.463410911408328,0.3094644820225009,
 0.5232889724392553,0.2030167589662213,0.20288936587133544,
 0.06437133047429311,0.9050415219273694,0.49288044433574973,
 0.6125273482660228,0.11005967966833441,0.8270993347200194,
 0.08979570816848537,0.0044939178976010075,0.9899937561729568,
 0.03261697328006696,0.8657002201142575,0.4669502412954736,0.5574503158914578,
 0.37045735872847096,0.11415913939014988,0.9613268069601726,
 0.6385130053725412,0.9695540827004505,0.5379390959988258,0.2749191820751362,
 0.6775933794640174,0.8111353039006115,0.7296013828171677,0.3050069993220972,
 0.6265317233172566,0.24856048364150807,0.5533450505076354,0.5057373152518202,
 0.6382156855291496,0.01741703280072604,0.7155505720980948,0.5264039292639568,
 0.4813095968264488,0.08700255699749204,0.41314441078656283,
 0.25209489149757003,0.0016605849279851537,0.4142967161951583,
 0.8524497885517489,0.800255317538409,0.6045960837054253,0.46526185169595535,
 0.8994376461808731,0.90291326608159,0.9720041970827925,0.3705767093577442,
 0.7770675620578077,0.7200457846559336,0.30183715475839845,
 0.02403071220050057,0.14782022574880227,0.22502146257124034,
 0.49531781071165326,0.16143464245814731,0.5951339427272069,
 0.21057253929107733,0.08956976639624226,0.07886480138884822,
 0.756263396412808,0.3778868496664083,0.7143378178396711,0.2812856320436373,
 0.2559527789337728,0.5494488091514392,0.017836019835880013,
 0.6723179766542986,0.8381822608286102,0.8568847556653737,0.6231940783303894,
 0.9272428278132621,0.8014980427892021,0.2895077288569071,0.08264587565582426,
 0.2129798813833257,0.1557891403062054,0.6388133399211023,0.9412841327218266,
 0.20471447495372042,0.6093159102893739,0.9471359835999255,0.6881490996722012,
 0.9593323233241667,0.37964724381026904,0.39659288029114004,
 0.7673101306558995,0.4566897805535113,0.10848963429735403,0.9342127043936074,
 0.5484034425037805,0.11298900752142893,0.7088544817967569,0.4455448635992204,
 0.10254776418190614,0.26151470873050964,0.1870014741108339,
 0.21104466359375162,0.49613525075868103,0.5606744473608045,
 0.33647480160290744,0.7354308405999936,0.6424926133489791,
 0.17781998267199106,0.7102264739701767,0.41589395660418627,
 0.7059070514637422,0.10921418306635311,0.8388592720813468,0.5307335738732596,
 0.8848229521186278,0.9457138065351198,0.718270749095529,0.11555438190017486,
 0.3529721678939082,0.046463164511692634,0.14264565837437893,
 0.14227357104829474,0.31219931463298767,0.7791040713754972,
 0.34394952551777336,0.28505023131440776,0.58723966728203,0.687428878474365,
 0.6880056200170754,0.647835176357686,0.4749725598649287,0.6164328815739851,
 0.02469903525768946,0.2788938146888047,0.7507140897962205,
 0.14856719399045237,0.13829900535365036,0.30143309197669477,
 0.887691370037838,0.40287901426849837,0.77122836511526,0.8615596094771953,
 0.2943044945382711,0.7852104722944963,0.10934385431004923,
 0.009133551709193943,0.7057366087368551,0.45921490821962585,
 0.22529683306215775,0.6507393015827581,0.851075039415398,0.5260350595867616,
 0.6341917008142686,0.4637736451153479,0.9033027806871746,0.3601910826036294,
 0.7670513515497634,0.7488005485425022,0.3665341907650238,0.7830763812089868,
 0.5951912235035253,0.41865696300242405,0.5686852512751454,
 0.25983544747177345,0.4344387848942912,0.14492649012672,0.9982069510439147,
 0.6929092048550434,0.939348576192372,0.48556844187516635,0.9230532611581971,
 0.1038580316050971,0.38409808219501673,0.2036225454802214,
 0.04437947919966323,0.053966651242028485,0.08853075759434093,
 0.12808177611132476,0.28808313033979627,0.7111228366869288,
 0.7953781092861877,0.4143844801276114,0.4359761772340818,0.3524889774459773,
 0.7141221323265536,0.2619870460681535,0.6155814043930924,0.4335561423327129,
 0.8644012637885095,0.35613510634327283,0.39632964936743953,
 0.24099992651097069,0.22976971527341528,0.4857816375524946,0.888338502445575,
 0.8334604262735527,0.22076241753294967,0.6566747165399256,
 0.0025500238236069173,0.6884293442368543,0.2756874639439987,
 0.38291116821902427,0.18600036424592625,0.8897884682109244,
 0.3560971228034293,0.1968089564936033,0.09078878752855257,
 0.09886192340456113,0.45409443169187746,0.7236788104463187,
 0.32503477216239995,0.7802355309842995,0.19702611045562934,
 0.5087753719744219,0.48222805100558985,0.25698399032343844,
 0.26321631341248675,0.3375594416625315,0.28088952610777296,
 0.22641966657773038,0.6077055792006865,0.8899069869012008,
 0.14052031408951893,0.8181892097900635,0.7935560553739219,0.9333798567796832,
 0.8432170576491439,0.24544835422822342,0.04612853113259585,
 0.7232723279867325,0.3896544004861998,0.6619957364824293,0.5336048945110572,
 0.6321328504177923,0.04448048880428579,0.8899997244255674,
 0.07246318499210658,0.9850212111749803,0.19843476324527487,
 0.8239616269764309,0.6439697896213317,0.5856435422152262,0.32013426392119737,
 0.9688951818087345,0.02732955196975495,0.6576816490733072,0.5022419416313708,
 0.3697138747755657,0.4256373103530444,0.7685624607005401,0.6531467794906122,
 0.10594251345129957,0.45772688488946556,0.8922654674337704,
 0.9239748976982868,0.0843760607816676,0.7417785610943495,0.8548435654938947,
 0.8837986006418949,0.9926205673827203,0.11247319576139214,0.819655241628266,
 0.6327353422496568,0.15331883058984863,0.6976321306625854,0.7025924216686545,
 0.03672309882242564,0.9504639945397748,0.5534735195130762,0.3768714069878225,
 0.2055997698360762,0.7045612654541595,0.3225491938360875,0.3279970684809743,
 0.06155372414496252,0.26661952105627373,0.46860466443305104,
 0.03291105188278176,0.4788457388218148,0.6285078146226192,0.7926595373393291,
 0.887120286428153,0.6505189654436765,0.919958891532971,0.7484786031467539,
 0.7683105942384629,0.9601929653643213,0.0021152432489651707,
 0.2586118127861139,0.7088562818687797,0.28798449322789443,0.6959999614448881,
 0.29899043174778095,0.16591138441519926,0.8474534536145035,
 0.7216501421745561,0.9407289263418069,0.5347983841351733,0.7339513310375433,
 0.009341545129690432,0.5862369000945581,0.4885891456669187,
 0.9358081363172894,0.1131944181365172,0.8552550139143474,0.22085178510844605,
 0.18461863962365355,0.7108376268514608,0.12136150854362882,
 0.37194122721680034,0.6351030723970118,0.05937014461667489,
 0.6751449752143883,0.5379838755245594,0.9265128001198256,0.584257146628733,
 0.6675826470767608,0.513331846716237,0.8256095640582358,0.7749677885741647,
 0.3343179132468823,0.6303880973672406,0.012273072947236452,
 0.08609747351088393,0.9216146500343344,0.9862968202118426,0.5408740284144051,
 0.595056296126537,0.8900595470086387,0.9747367533001539,0.5388081611582335,
 0.33737510042933405,0.3324121979901423,0.7954817464540767,0.6990606573558026,
 0.38021307248744884,0.12961121021105138,0.9897182993983937,
 0.25771236411611653,0.8406974660030635,0.38171509025104244,
 0.13165714558389485,0.8500435286393854,0.9904291127798311,0.8397955627685784,
 0.06657907110655004,0.5922641562118479,0.15900451482212408,
 0.9155333096600231,0.9851077017852314,0.6142973782743789,0.530962653712789,
 0.2669647092425884,0.5594945309647008,0.17237398549047056,
 0.023289101264080014,0.8441217008072346,0.794252967829282,0.1152341271040439,
 0.7183682625982591,0.16960874237811618,0.8671122432281297,0.8151671619012378,
 0.5563090712163169,0.9756382844623012,0.7808572478834185,0.17655264652066527,
 0.997278381446081,0.2595723712923276,0.8424539162174357,0.12939039454163104,
 0.7967655920608291,0.39653818730433354,0.4200613749347486,0.6688261870257742,
 0.4895078695550873,0.9877112120485417,0.8164940250670348,0.44445782688300894,
 0.5207511360738428,0.13356428642598872,0.22411796130063877,0.104382139263413,
 0.84832320490602,0.22005600975497053,0.6631566138371668,0.5907789957395069,
 0.08031365883532837,0.6947710864377473,0.16645377244055903,
 0.4621967110569678,0.44214987467882694,0.16948163848683573,
 0.7773393478516838,0.5433639554985767,0.0014904712283090316,
 0.31289810271413954,0.4979487388011752,0.09406038717959875,
 0.7198133670701745,0.6245186247659543,0.4605833076191942,0.647174949882878,
 0.41780803891296026,0.9905545416291197,0.5793510352882909,0.8534582531334981,
 0.7952602806284965,0.963906082249431,0.6521387483458444,0.542401489262744,
 0.024893750000644643,0.48455003733421287,0.9242948342869244,
 0.5599925217980054,0.8901361577251711,0.07512035997262156,
 0.42465605794110706,0.0351064920882006,0.9777895178860231,
 0.06605292457155931,0.3711248148986619,0.6414090449930031,0.7109450156624981,
 0.5146998778863819,0.45916957324150265,0.678188375606897,0.3732144796726613,
 0.7844315629181199,0.14124481330816419,0.2129175912417436,0.7580020565029253,
 0.9545169237502154,0.7096652539157404,0.35171106948484276,0.5701538895815085,
 0.4082076505104759,0.5357382031163168,0.3747293462001682,0.03913443111958226,
 0.8324567879338373,0.8468539918984143,0.896803298659278,0.8373152048568676,
 0.2419627023738219,0.43131778634566853,0.6512142151149664,0.6323026606484488,
 0.2774937904607825,0.9526193844269831,0.6777175416776025,0.9768277326181263,
 0.18061491392311124,0.5290180446947116,0.4620438280696457,0.7579039919506124,
 0.2729307266582236,0.9027733504707622,0.29116532602767387,
 0.37108469262895916,0.5151787265509031,0.8728134087882563,0.5818907159504465,
 0.31897499867943313,0.8402787978139493,0.5109939078425797,0.4083342146463408,
 0.5333310996957307,0.35999038812679063,0.1282138313944109,0.5005004005760759,
 0.968675693968966,0.6676031425205156,0.7604006332707102,0.07135733216967677,
 0.7062140303620649,0.3302865418514268,0.47959579205770697,0.8951857811067134,
 0.40286207195243073,0.13033784660620318,0.9582282035735074,0.604219998116953,
 0.6973444446227225,0.339707144328957,0.022764981706320908,0.70429059359969,
 0.6203273853418805,0.6021337245503944,0.12975717662707742,0.8023071523464012,
 0.227701292414214,0.5980376354710872,0.9518126746407324,0.2691149714722536,
 0.7061648486903747,0.3023956888314734,0.46972444243694766,0.9906223172920374,
 0.8719175487022561,0.5148817351990277,0.8585392779327332,0.9979917264421317,
 0.6103061747229052,0.1111357368521605,0.686419144742185,0.016153905523120682,
 0.8824760723720648,0.06216285943724731,0.6840694801783846,0.8676876373533987,
 0.43973788507322686,0.8148888168991438,0.9945448929158703,
 0.49197592601452844,0.6346403869246009,0.3505190181996879,
 0.23727337307918506,0.574572231347978,0.8395584590006083,0.8305569455810624,
 0.15271495858116746,0.24199731294598814,0.7722866409020386,
 0.8220324298692341,0.8957118044478617,0.8965318297416277,0.3262417808958882,
 0.8907458133557798,0.3955695570168496,0.623547320237396,0.9755577362175187,
 0.3267211334609259,0.4688612657844775,0.4750444576686481,0.3278984185068339,
 0.9292189981571389,0.5017048649324838,0.22389478984567612,
 0.21959078389846176,0.9707674205963807,0.5931497543918082,0.4257361747121652,
 0.44204557633742114,0.7727728154009139,0.4375527689976634,
 0.23544949437917706,0.20426397234646032,0.2136450108309046,
 0.19783889829136092,0.9059077701887503,9.170912964721056e-4,
 0.9647314796404427,0.21225823129837018,0.8848665456988137,0.175552406196414,
 0.9709155424224931,0.9710614297479176,0.8834056122053185,0.2454458844318007,
 0.012292927667022835,0.18582995233668365,0.5199168908670653,
 0.3533779155778962,0.06968960364679222,0.9175148684196841,
 0.49053761723561273,0.12195066544630762,0.6466635762506846,
 0.8100896844990311,0.49914790469510384,0.7791008583488369,0.9181662406351719,
 0.5756508238541165,0.011823212720520893,0.6814602023684005,
 0.7063779319847937,0.9373383775987028,0.533119312094051,0.340531522599709,
 0.051690783356264536,0.7911200387244612,0.18663314483716353,
 0.7392608832060669,0.3595117267887913,0.2381366874582782,
 0.011198054446326267,0.5798124612162082,0.4906537105547665,
 0.5677974907655117,0.47051696367743867,0.44027951917365415,
 0.9970719449345774,0.30392311338035016,0.6261104733238139,0.6300590479783169,
 0.16550779778469027,0.8737528584289099,0.49264605651190063,
 0.9218151021575572,0.45018311245160003,0.7491058539519609,
 0.48453124281256044,0.5536347807406213,0.18857160773059356,
 0.7425008245241439,0.9979854584804393,0.7308191705352194,0.9861129129441665,
 0.6890667315477335,0.9219410428782921,0.6186813457236046,0.6447641919946997,
 0.6027405962093692,0.16705567070313387,0.12261143822945042,
 0.5830513799299628,0.3018919424790423,0.3015550263928548,0.9969391251364974,
 0.39105466006867284,0.9468101148461614,0.7260567996722459,
 0.49145658113260837,0.674429104111869,0.9170412169529445,0.45918525299272217,
 0.8627123216829437,0.4140964225171885,0.8781706934533164,0.728718963212808,
 0.23086913680015042,0.2798792643673518,0.875055261254593,0.4270289467165823,
 0.9477744541271002,0.4480215780996031,0.5694747513103225,0.8695533769438205,
 0.6688087639571516,0.578243505345358,0.3640032057189363,0.35474221852619736,
 0.1187118814008471,0.6258999826253236,0.3873319741407323,0.22895909208544207,
 0.6670250757114214,0.44796605557084557,0.9745209378475772,0.9330779627693564,
 0.5334371251513357,0.25378673896386894,0.9995457052102902,
 0.14090576880471117,0.07475047932566581,0.08372482082411681,
 0.26778881659473397,0.7821920566857836,0.8894126152183259,0.2990489578596903,
 0.21435781131110454,0.9538293243139623,0.7389755523916514,0.5762096187129214,
 0.034015311960556294,0.8413439794254799,0.2245257477396566,
 0.3043803250632423,0.7980747419783996,0.9221469944942686,0.6899859271860395,
 0.9841482140840832,0.06858655952836823,0.23378981638196827,
 0.8753811062890704,0.4604232647848141,0.07325478108235406,0.2000045887156804,
 0.2803968294750496,0.8539305747528024,0.2267407442537317,
 0.054836121390778314,0.0054045817983844335,0.0024208292526176622,
 0.36011549751462213,0.8303637378511894,0.8843294494360568,0.2948902195266353,
 0.6944323172869963,0.9232860048341051,0.9262158045275896,0.2506267356473998,
 0.8942054694141321,0.8647154460913713,0.6659817898886388,0.075077988477677,
 0.6418666500754833,0.5705541210377111,0.04773717077363082,0.6100544539136802,
 0.7584510753809237,0.4583438334193277,0.26046121223117746,
 0.32819150496184035,0.14313819543945105,0.29975934076756305,
 0.1403763633448828,0.5404789522641424,0.40167956046406017,0.6249970807129948,
 0.5337869591516061,0.1456172666873916,0.6276422937427264,0.21946740343172577,
 0.9162241314988743,0.31730947036230295,0.6381620354087791,0.1608897001020827,
 0.15887999859859936,0.18930485079770398,0.152329989797803,0.9473592057762932,
 0.009852159589933684,0.39031900601721803,0.2534547846020213,
 0.9680538529280793,0.8581442760534106,0.05684121127745112,
 0.07484031020250237,0.7915274742875819,0.9180704051439623,
 0.40251198593865234,0.09496436570948608,0.6595551274281231,
 0.29786006837267065,0.6746563803805953,0.36556648071584574,
 0.5940657275530774,0.3758798544261456,0.64604737753721,0.36057943388476954,
 0.5666916221583567,0.4753138230826873,0.21206763051017108,0.7171036360178362,
 0.8223580019872448,0.9845444317215049,0.6267202066291814,0.6004090255062452,
 0.3278411213147949,0.03575530871587762,0.5535685670600279,0.7658468785094529,
 0.321481312872959,0.6177666780024431,0.4738178307212194,0.27806447202551243,
 0.2859627339294166,0.25340370061631934,0.4867642249577302,0.6396991508607417,
 0.49803285581343726,0.4007093465686852,0.8849504466965569,
 0.15285855716772478,0.9263226960903508,0.24921943839595173,
 0.35748365143482974,0.9849405304341015,0.7588601707254714,0.6032528465496195,
 0.8356486936364625,0.8253650965799796,0.10885465392951121,0.943715738821613,
 0.7685293232998477,0.12100778730644901,0.7084407984142239,0.9345080154142533,
 0.33755752030870434,0.30949848145709813,0.34819631641826154,
 0.930550269558063,0.5386924134272095,0.4546557156421437,0.39552490590668554,
 0.3606403762939334,0.3974696796785572,0.8117000346231376,0.7067075089634196,
 0.3771067684274675,0.9236649065014114,0.737184393589344,0.21568490720510602,
 0.18837305705215945,0.6871170783085158,0.745638656449425,0.9151469950812613,
 0.1267316289500996,0.873710869725934,0.003936563391295889,
 0.004235370631602953,0.38096796874158323,0.33420161184747044,
 0.8556898271977583,0.2139067348661794,0.45936974195854163,
 0.38308976148488905,0.8394809694735574,0.3717167626586989,
 0.08452657706852351,0.042629454809960055,0.02148128096166868,
 0.4920999676829466,0.7852151765124005,0.7969347081381748,0.4794360530164352,
 0.3798757516154463,0.9687666436650672,0.4533090204102139,0.2516128635149061,
 0.3316505050880134,0.8952022911013796,0.9236899506715435,0.833767978987316,
 0.33086999217041724,0.12272019539389567,0.9689337837409935,
 0.6790344517472442,0.814641495261256,0.8270429226019088,0.8836582257503705,
 0.0016587845489854924,0.9005962326132069,0.6578061066274863,
 0.6591866158247068,0.15565616162692175,0.20358003145905246,
 0.33200472694436034,0.01883824494327324,0.26127971108792547,
 0.5591485295778937,0.12796521329341437,0.06578408612548525,
 0.8993292342034264,0.22289895136278792,0.6045730429589002,
 0.031575087458765916,0.535407191168605,0.5690235372862752,
 0.17600113725305078,0.9354528562996356,0.009677617777261172,
 0.39908505633433833,0.320195796690509,0.3439747165816657,0.31651553367337026,
 0.6913084332604401,0.7105639219826967,0.5226843574607059,
 0.0047822489583900785,0.8009855084327089,0.8644818021264843,
 0.7396512829682457,0.8690334036290082,0.504596377607052,0.5496304402310195,
 0.45329343417259094,0.2792539646115413,0.39824938971772283,
 0.17733663139547406,0.34674822636931624,0.33418777052328985,
 0.6437385604030957,0.7189230691636826,0.9508691113659874,0.8799346319862976,
 0.9078275095609273,0.12556757600759372,0.22056256766081772,
 0.32459906239099434,0.3114131131463923,0.9274669726159588,0.665728501766454,
 0.1553493142656246,0.8974946003937624,0.8030167016524398,0.6843658660453935,
 0.8064042863689445,0.6147313025177568,0.6504347228577844,0.15755977997803283,
 0.19555599062286777,0.9739012945366907,0.4365063461655053,
 0.38544049174981176,0.32753259289025183,0.027931264246400245,
 0.48978739978928787,0.9898974142476116,0.8139353544884619,0.9784929991297645,
 0.8817115894663679,0.5017907643304136,0.7657144550703702,0.9226886829411586,
 0.05292927238219414,0.5917680330383159,0.5571368752170714,0.8293453830995796,
 0.6038874759992424,0.5541139635300376,0.9339648701461671,0.26851766308776504,
 0.4429686485233566,0.2766180811455249,0.6750671890039125,0.9571728713180014,
 0.6931918242520727,0.6218623780194138,0.9876853891978566,
 0.0027247697396970416,0.03186631059886591,0.2993268531015776,
 0.6000347846558574,0.8785238266750937,0.6693140360512866,0.2829718281223492,
 0.2702250454416988,0.865962869251221,0.3141159560768614,0.9309968745930048,
 0.27548829683857734,0.6105552149511501,0.027244877744860174,
 0.07486020470812016,0.17474763782306002,0.015871930428420278,
 0.4070147127907693,0.944686711904094,0.47415880482282646,0.333494371432526,
 0.3875988673673799,0.613294657734404,0.2077629242666169,0.25267275087414554,
 0.9089074573118772,0.5886607277059922,0.8196140701804386,0.4368667792977138,
 0.49737930747092296,0.5563553754012529,0.23948529802259455,
 0.9707404843889117,0.9350247885564401,0.38370266836750444,0.9414607700079045,
 0.8161497990461323,0.7245145881276263,0.8621456660932391,0.3973433617094878,
 0.15593450895868288,0.7594998730690125,0.14473242680550957,
 0.8935137888997173,0.3018829090205436,0.3091056479194887,0.17379063771296788,
 0.1850162940375839,0.8867063957209816,0.05453461707368734,
 0.32676958287370395,0.7082005250395194,0.9048415925934338,0.4577042445295657,
 0.013323685583831013,0.35285929787124914,0.1728573541290036,
 0.32251180123007894,0.9322337337392512,0.3535479738970647,0.5515281162497356,
 0.7651548884358653,0.39844298336611583,0.24213829877658044,
 0.1269821832565654,0.2511462134744079,0.32923395476967354,0.1022662565543141,
 0.41914170037972087,0.4051928452998861,0.1606931503004938,0.4460261204049263,
 0.2564484668757474,0.16340141843087974,0.6060374307330604,0.6895433629779765,
 0.46331306830956787,0.5066928536033228,0.21719481960308484,
 0.6730197391282982,0.37567343379784834,0.8082065569368279,0.4153939565185413,
 0.10360886637621125,0.37075155691909534,0.26754518953828377,
 0.3242188217127121,0.9305107116090631,0.4460641341770686,0.9284390335451735,
 0.7649292866786642,0.3366945499127092,0.9865140965688859,0.9982697847825572,
 0.45631783678718885,0.05886270523614423,0.8939717792935088,
 0.05547591300923527,0.32941105421878947,0.7157692338572615,
 0.25257564426690293,0.9064326827853869,0.6693948574832387,
 0.04107179766431579,0.16921637675070633,0.812479320243678,0.5443704755071118,
 0.6653021858684462,0.43533750238820246,0.9852033961036939,
 0.11624393347033679,0.6650696552899898,0.12693004231024707,
 0.06727085992501414,0.9825983346231542,0.24886337290792881,
 0.4338621367512998,0.4430948821697209,0.8767809186744108,0.7971179132735127,
 0.8336567732831506,0.6906631191068229,0.9643137654743676,0.1433787091664379,
 0.8856640495873674,0.7341495875074723,0.6919255308446861,0.5354645190855056,
 0.7718524124798385,0.34249884728894653,0.10973580673185612,
 0.04349359732574609,0.0010795435524941777,0.4691104839467104,
 0.7709892012306376,0.7165796917930747,0.9764591266531881,0.9239184133420204,
 0.8917260504361931,0.52057008013993,0.11517515527997546,0.7443568964184528,
 0.6350519493432469,0.4783360620533055,0.5878417628553583,0.5079609157991846,
 0.9342170598112368,0.6054747658588079,0.07759146463411537,
 0.03738296660641394,0.5289594141738233,0.9310534687937202,0.7107064996593894,
 0.022289099167137283,0.8178290346762844,0.37885649116017295,
 0.6904875912472871,0.5326907730097136,0.47225759123462985,0.281308931567695,
 0.030713139255647004,0.6449216768720165,0.12129134579416567,
 0.34280251735477796,0.8661468712988026,0.32450029898119404,
 0.6330325763755593,0.7762879337164096,0.7192293998717282,0.6733166834763095,
 0.6235592775114078,0.6319974022672616,0.141147670607769,0.3640756648508525,
 0.26019350764741744,0.43196796507673907,0.6609855415692357,
 0.4690908688420984,0.9297525789004093,0.9079275889363897,0.23025371927605587,
 0.7368974409700703,0.8235784004105032,0.4487442964580993,0.3731397667765761,
 0.7659638101099728,0.02735360303846024,0.19728119753940232,
 0.7534718166461625,0.2503691706855594,0.7009813524233136,0.7825531001966805,
 0.2212213111903114,0.1773959275557454,0.13188087874289267,0.6913975279460463,
 0.6346374405394534,0.38305165946231656,0.4729189243340751,0.9644833348505193,
 0.46139604895450215,0.1328436031756367,0.43667540854654563,
 0.6323267511953805,0.260106799497573,0.2500428248333071,0.7681368835379826,
 0.13881792499745993,0.2956054934974215,0.4571261082650646,
 0.22261995170238702,0.8942233957743532,0.16290549172940472,
 0.4543896669027766,0.7009163089787191,0.6827025614750762,0.08828370570420219,
 0.5594171790463872,0.725420832846662,0.29574224895314316,0.3266827676845988,
 0.11958691770607843,0.1480754199687776,0.29222957712259834,
 0.5432355591367372,0.4920797403180305,0.7765022847393028,0.8779646210410939,
 0.09762689024596949,0.06800318418575968,0.1943220314666424,
 0.052338430079818865,0.8298902713056557,0.6292952395183739,
 0.4038251742747204,0.23980569716040034,0.9228874039342038,0.8669508597269701,
 0.35236879875619853,0.589958452335474,0.15663379244693576,0.895609461929561,
 0.2024783876434274,0.8882940064695566,0.20494196830275857,
 0.17515520762906822,0.4200662078991968,0.29462638027150323,
 0.6813442855127912,0.16825245882335782,0.06817560197267858,
 0.5730678232213233,0.12075979908600887,0.0679602102134419,0.4436470136142081,
 0.3146175132426954,0.09874398415913999,0.9871590650629447,0.3685444882983495,
 0.9301921979143682,0.8653843351277603,0.4808360332505197,0.6878078979726207,
 0.5254005157491035,0.3077269344823268,0.5156118733090608,0.10271380992957924,
 0.5267048079388172,0.9074527410121855,0.8966173234550697,0.7388915852662696,
 0.7226549134241906,0.2307855076706432,0.3984890214496922,0.49092217641385716,
 0.9187178325165655,0.3901880202385255,0.538035422160013,0.3577694942702734,
 0.7524272380498596,0.34288692400284737,0.05716382774149509,
 0.2558050014963431,0.3859975035042489,0.09448766603287284,
 0.19149633604950522,0.4219705904628793,0.5066401704473679,0.9319578823252522,
 0.24594798271287044,0.9627092297658227,0.6244933937987152,0.6463695519857962,
 0.5915616048469617,0.9677060892239422,0.387657398918017,0.49710311279526154,
 0.3340318931228874,0.1363270035547638,0.6466271753934901,0.22812916976943276,
 0.9204427363087619,0.13194573340608784,0.8855294201505103,0.9509248488921939,
 0.4914380180041522,0.6011385800346382,0.7754276198655651,0.6345091974248904,
 0.6541868104646679,0.10160011736298669,0.7058059417309855,0.691917459372759,
 0.5440642700271409,0.791813905644701,0.6143180732302388,0.45513407661323513,
 0.22551936517699267,0.40097705084213064,0.4035640097778175,
 0.4806577785706929,0.8935215196197814,0.6911225388996544,0.9867987087163959,
 0.6083334104058981,0.4205961430952314,0.5993231146811319,0.9301624334432228,
 0.1537361180828839,0.7301679140542768,0.2506488622500389,0.5658246802728575,
 0.9488141738767941,0.7970503806768179,0.7822429654694124,0.5098653978328993,
 0.016125529960418494,0.5129096906094206,0.3536062569048274,
 0.5792425559133398,0.9236987419370599,0.3974623421752024,0.07932929561026336,
 0.06700093211471758,0.4694331300259418,0.45012412846577865,
 0.043176077380552735,0.45179549216314885,0.24726982133983466,
 0.19114368043204444,0.019301853233827715,0.33963487546886684,
 0.7857805390737831,0.782612975274574,0.24881572268025431,0.8052001130057734,
 0.5054353748100991,0.29703403873685963,0.5328078205591892,0.8881691480574317,
 0.35078253519455693,0.19127790722301707,0.4855150486252956,
 0.47763425806075666,0.6599056982208864,0.5848405557299877,
 0.33835165588685623,0.9555005314305298,0.7474025952115133,
 0.37449870249356865,0.6387488964587782,0.023872227250376632,
 0.47615008535902326,0.33240951538476415,0.14974257160413829,
 0.6619471433650899,0.6832103551659096,0.4023299700294847,0.943935742264864,
 0.08521624020493557,0.7513750997937739,0.18685907759324738,
 0.6825956833629905,0.20340921634909392,0.7653617171189095,0.4774558380204037,
 0.4981274411050719,0.08117099144015805,0.9050434469023427,0.6114042083488038,
 0.08495155091050333,0.4895788519408756,0.6170177288262991,0.5431399016495315,
 0.3827300935182163,0.9410825900689022,0.46503935507586114,0.9866506418205995,
 0.8371891223289788,0.4927099496047227,0.07473278745254863,0.8479305601518847,
 0.2825416005895681,0.8611397353068817,0.9163436906575932,0.6670137609704188,
 0.36283514381975945,0.02477946997815106,0.12815233313733332,
 0.8390665959033499,0.930344248924149,0.841079036562988,0.8515304870936541,
 0.5475108645657684,0.14468200840713452,0.6593586512577609,0.9266694764427875,
 0.6588855147427943,0.4145717962898,0.46916132184774895,0.5675182219992904,
 0.43729288193003957,0.46901059778146326,0.6498954369666178,
 0.16629106071843747,0.19817626405658095,0.9181230095857921,
 0.3520443817311786,0.2566108929783887,0.3384267190367318,0.9284066201275636,
 0.8401202177734278,0.8913923690266574,0.7489859655578033,0.6036223297571005,
 0.4194632577260332,0.266300274188356,0.239308661083381,0.8312125041209165,
 0.7852478435974256,0.3565334245597669,0.21989592388137813,
 0.08644190366930438,0.17666928909751234,0.4167045811620309,
 0.3177960842639157,0.7238185000096014,0.13613345216945216,
 0.07157079881924766,0.9882229035864747,0.9577408284220323,
 0.11173368318207022,0.8754630559688805,0.31347005703702735,
 0.11795185939095054,0.026356825204311485,0.4487111285723375,
 0.6600007266827601,0.3854971712928362,0.6496356035657671,0.27311731503349035,
 0.43740643203701757,0.8113815625860611,0.9662557959480251,0.9645635556407288,
 0.4249573655051234,0.6803216395047743,0.5319409464981014,0.8155920055576785,
 0.305422482082361,0.32929743226434516,0.3654272622050653,0.08757193154935938,
 0.9819185563370415,0.2425106796634522,0.5852392015680605,0.576983292156739,
 0.35503931763611063,0.8366070034882137,0.6419887174929514,
 0.31091763244007264,0.4495950142044728,0.1525164103999651,0.4911588463838007,
 0.13575655538703957,0.6690550847768899,0.8032566019291638,0.1143716611020078,
 0.1750316170780275,0.3221773479223137,0.46816290745358424,
 0.029811868284179255,0.9597701365384662,0.28802812046058524,
 0.30551249993162777,0.7599267068681271,0.5462384576620909,
 0.09331836981435082,0.48843582093161875,0.9978849010177323,
 0.7791108798615196,0.6673659190229648,0.1497080279666344,0.9672258206259916,
 0.9106406078471685,0.13709188270876915,0.46582643071963037,
 0.9946888035373544,0.6946003009820578,0.7640044219960399,0.0516934012555027,
 0.2857884798238318,0.947556235474107,0.1069354838305081,0.38462049615816785,
 0.6132962516455353,0.8812653710339398,0.49921891650790173,0.3541678410702235,
 0.6811976697711685,0.3914085784548128,0.6971700219932851,0.7773168864331885,
 0.244978380176743,0.18388773268616776,0.5468758344432649,0.7707953617272881,
 0.5547156000043719,0.2855525084187802,0.1646942503322808,0.555329586869465,
 0.6779571735027815,0.18996611175163824,0.33084994306387094,
 0.24560326409282318,0.03729236986051754,0.29346731850785923,
 0.3105947619407078,0.3105591731120141,0.6873151535004773,0.7783734888418569,
 0.4298575284066676,0.9807407407350061,0.3838275565968099,0.30788740523996916,
 0.4517834755605641,0.26257590040575773,0.6820423158554434,0.6681564918184251,
 0.5812839468762468,0.894184894065285,0.662335067500878,0.23800503743138934,
 0.35732944764520125,0.6403417720338672,0.39538369390706396,
 0.05980309339430545,0.9592840722272296,0.8839352846011783,
 0.47541224964827855,0.05783193334592596,0.3694078075564078,
 0.06788502635641125,0.7528277830619231,0.30819666403786694,
 0.8485784791118436,0.34311272963455686,0.3321054668642338,
 0.08694211511833627,0.8586671684928823,0.6622003318542815,0.5989957487135832,
 0.942097409264689,0.08162550144273095,0.7196537606314234,0.11963628476432386,
 0.08186414432582223,0.8893975133764663,0.4522697028491681,0.8951749122789209,
 0.7534689708163304,0.11017953935007174,0.8280106073276995,
 0.37194183047138674,0.9641411096135091,0.40816858575577986,
 0.7624334183281323,0.8500990277056925,0.46772326846702017,
 0.23928097768340317,0.5719705710753527,0.8946130616116619,0.8048871855904033,
 0.8125567571641734,0.47485144123161216,0.15710643128103885,
 0.43490417992531794,0.989366360502266,0.3506448540525243,0.09776099203116428,
 0.2947613377160295,0.7825057559694181,0.7893782776298666,0.14121226006493537,
 0.24185537186070538,0.244386994058857,0.33237483847560334,
 0.01783250184801166,0.8991633905400211,0.8353224666166441,0.874802646119117,
 0.770302467879426,0.12407125358408355,0.22872887373998652,
 0.37587015344956987,0.35485721509874735,0.6089868233732665,
 0.3773896900972671,0.8364151095434563,0.9712418797795416,0.16676293302725642,
 0.5642833061880116,0.6854405489797898,0.03682088693688823,0.4721688947999635,
 0.04050655325548813,0.19606822611712826,0.996054642772881,0.6607186881152662,
 0.36897504140269377,0.530015381000049,0.7371826142648862,0.4688680982885388,
 0.021749941341638213,0.7406964661886385,0.48711533918867855,
 0.9385413693744887,0.7874903743630564,0.6118893134611765,0.11431619374459046,
 0.13248785849863054,0.7958706272348801,0.5717293000315965,0.4251422142797917,
 0.044793860120268536,0.3820346488221249,0.22787538517118877,
 0.20122865640491205,0.02841852946796175,0.8285251689991053,
 0.35165770968274424,0.7680579972646655,0.6316343767103132,0.6253101617594677,
 0.2756749714719018,0.6488822988581426,0.5907957531056358,0.8642495945281945,
 0.1337138986394114,0.7256670713122494,0.11042083626970123,0.498161073771757,
 0.06277375018649944,0.38726316397161387,0.7398872860808351,
 0.4383558768164687,0.7953691648567635,0.8514131829130909,0.9981522323615415,
 0.7039343710199307,0.29482218696567675,0.04456619309139298,
 0.044731467510089495,0.8346311984386764,0.8656191677939241,
 0.5245313281218285,0.17894931871888076,0.384199954954776,0.0720838238094288,
 0.8481583514397096,0.30740783819505557,0.0023924487818076656,
 0.29681068367979924,0.17307102446790057,0.24392324877822114,
 0.39784405896117125,0.3991585616629827,0.6499902700839307,
 0.011383998993779398,0.3410423433702844,7.492780094680018e-4,
 0.6257546562987909,0.7826929089260332,0.9721678932880772,0.02520197463934204,
 0.12959193201813324,0.7723516419404101,0.8413675940505358,0.9359487044354485,
 0.13056269005444765,0.9439990725990883,0.7473567301512641,
 0.23518548519652782,0.031509062855772485,0.7032749732334852,
 0.8010637576952633,0.028717798096814295,0.0995279691285873,
 0.7818539153659972,0.186171029409549,0.8711178231476463,0.13872843114940858,
 0.3636542723442392,0.651005217853235,0.8870261343169509,0.7677538158018992,
 0.8078835715395123,0.4181633705230843,0.32021066580713353,
 0.40012669038245907,0.45472815769229324,0.5943669950426608,
 0.6886857446486614,0.7163875068399325,0.7732404315036727,0.9090116676234401,
 0.9899574010540373,0.046240358963450245,0.0030174107145984896,
 0.6397779169199164,0.8501356655707125,0.7473425580012516,0.941518030708338,
 0.7333450189932794,0.4293380491145433,0.13541292461062826,0.8070103008170872,
 0.7515656476022037,0.19766130924112235,0.031861555593728164,
 0.09185151043802342,0.065581224535459,0.844038792639672,0.28959980972741495,
 0.6942695546804799,0.7380944162119233,0.09589155890716317,0.5269697884065127,
 0.6003356463113427,0.7347272079014682,0.9383564933933968,0.333949205904552,
 0.25579651482304044,0.1631306211394975,0.8882631740617039,
 0.09323508963415283,0.4535653611082595,0.9182495097071659,0.2512494667491444,
 0.05470698644587868,0.7273240439082067,0.7119545305978192,0.873198696571688,
 0.1848027629576201,0.07370907996667209,0.17825238750216776,
 0.7291370202060976,0.23981303638571638,0.9949817959949073,
 0.21315746356024956,0.3176741511247594,0.700568452753103,0.05646896295763104,
 0.5100633667661385,0.26181504099160613,0.4764560480767728,
 0.36754368496619905,0.639187265661873,0.6456248492209403,0.8790863008584612,
 0.6767460867871744,0.46066205774862046,0.20515084029349873,
 0.9481093371386429,0.7139414476251329,0.23798806996498234,0.2126711766753815,
 0.32117620701227323,0.9021692877213549,0.054264649355869055,
 0.2348129196421962,0.21271017601344422,0.6761492720777417,0.9837698098088613,
 0.06725149917358342,0.23747065241136978,0.9096825561355589,
 0.8820258976282922,0.338742344300485,0.7733298933219078,0.31895797517956737,
 0.33935881031858184,0.2860554138819881,0.5521454125805163,0.6217540398155396,
 0.9893768190723297,0.33473371214926406,0.05713578244485096,
 0.13874439897299795,0.6904111256881821,0.1185150740998504,0.6656741430013084,
 0.2374909473177672,0.19044746691274128,0.5959998540557349,
 0.24937242327775977,0.7930155380901487,0.6246482208138646,0.7962317280513302,
 0.655715374614446,0.8815454535509915,0.9004716941863442,0.4518614323366712,
 0.5152957393728642,0.30074442143708535,0.24327043615764,0.9427226199110678,
 0.3340641264502464,0.8768070334135801,0.37183311292904264,
 0.19243087711505535,0.9549312381857344,0.6850640766604009,
 0.36704671386219445,0.19736535188842574,0.280772662050086,
 0.034945801682516486,0.3194376109604937,0.9277668286090541,0.413413656299064,
 0.9054679726028528,0.5374042574733332,0.7994940395203042,0.33156267281594554,
 0.4486838005111913,0.9591205939424841,0.8912415467722641,0.7651200990849519,
 0.4543013727669897,0.4282997624604229,0.3775764782872503,0.9193362660115331,
 0.7763123392272956,0.4567369337406675,0.7391963354035089,0.4872693631008236,
 0.6979681369445276,0.776789704220137,0.23957918242859488,0.4668458819167689,
 0.40479386435078,0.05959338033286843,0.07228734098755918,0.37274981327379686,
 0.8976639679204792,0.5852077711859154,0.43539097715928854,0.1419361226813023,
 0.9464951323702198,0.4531980072257742,0.9197746716885383,0.4728235071766691,
 0.6307278825635837,0.684926470997759,0.21495234497391402,0.8022617488766577,
 0.07678950080944313,0.3677119788067227,0.8624129938960782,0.3757027271988176,
 0.01558123972664327,0.2627330327640356,0.9128239111152097,0.4161998469400733,
 0.6949859523096386,0.07688075161222785,0.04028178383798575,
 0.7221781604531681,0.7977094507595849,0.8917163367799379,0.39212011643625555,
 0.04517673390840504,0.7539238591548562,0.9515592664415045,0.8805020391326223,
 0.12490536107372174,0.3552696800724613,0.49970425311515465,
 0.5363782055179844,0.5764090315618788,0.6322658239755787,0.6814968772820947,
 0.4601110835804151,0.13948000079049072,0.686625414571981,0.14140838593995708,
 0.4768529631725992,0.9042210484472434,0.8676941689169952,0.7079229125011791,
 0.1791015066375412,0.04171921733956807,0.2677136330767066,
 0.43279011381468047,0.7937388221869371,0.05357802989497018,
 0.03619480748227444,0.2192112887779153,0.8518418140776093,0.4834701908328184,
 0.7874539332783491,0.7798690852293647,0.14787094418158475,0.6500394157812693,
 0.2538325946810607,0.7905883525233334,0.5187954778692668,0.798132493884068,
 0.6090610109394868,0.18103086120059853,0.49576536390540826,
 0.8832819024379412,0.06108748184374302,0.1732403403508137,0.2525702575673797,
 0.8356691081607233,0.7587565451906081,0.0012192820407032068,
 0.4860870595934873,0.9617777119084286,0.22095855008932874,0.5537399526933751,
 0.3221053232241583,0.8125302849132473,0.4147618688876523,0.7292884480420252,
 0.29051183901015065,0.4132053508111938,0.6209772876627639,0.8081334257071379,
 0.9742551893967444,0.4950707078317115,0.2106887039243729,0.6665390985150202,
 0.6371565489184484,0.31544163103275535,0.5233270334569595,
 0.11039124901040775,0.26946892950199186,0.2298361406067586,
 0.7386152089643759,0.2215642254638276,0.04186421559465525,0.8627466054451423,
 0.18469518615717684,0.228153245749258,0.5987254070416368,0.27406429908956753,
 0.5976104566836247,0.27012603624161635,0.7968306162116482,0.5992768050318507,
 0.036908839220102685,0.35119379771207226,0.9688404108966792,
 0.7674758539710074,0.7730247674659309,0.5966962072838453,0.40597691933309565,
 0.7476822694888798,0.3331586656154448,0.6475760402988888,0.7189420865729148,
 0.45120981578203245,0.6695780964079402,0.9863582627679729,0.6924877953794506,
 0.15556347896830536,0.014814270767273707,0.4693359588371109,
 0.7882519681227989,0.7437914097007654,0.4044157607638713,0.5745864744541778,
 0.47122063936326897,0.15959393434555202,0.9745587111310096,
 0.014654317384285687,0.5690996439356243,0.7294544618685042,
 0.9906515498559701,0.17385191832266544,0.023234885754679357,
 0.05851041482218189,0.6209093544450834,0.6670157901881258,
 0.47349404735271305,0.20410342533763903,0.19118559111645694,
 0.7595564116280809,0.8549936446714793,0.3907697193796209,0.9498996928734994,
 0.23348084171865535,0.9972773140439126,0.44915283051003174,
 0.26752221315443836,0.4153767218796073,0.2775954424883442,0.961708863153925,
 0.6685102889456818,0.7073733753355235,0.559517217700467,0.9910428501685093,
 0.8987879298624107,0.4859831770887848,0.1940092495442055,0.7985352192202243,
 0.646937166183826,0.20395626058578298,0.5541234061646032,0.634484382907964,
 0.40312154640038955,0.4446461591052522,0.44192222274920545,
 0.3431435284735783,0.6832096954987072,0.4286150305861113,0.9043453467924808,
 0.4998695426122207,0.6596281746102995,0.7362937287694811,0.5822865736428788,
 0.6850709740045127,0.7528073109842874,0.8716841889759066,0.5628090902440166,
 0.6217029360843811,0.16472975861770855,0.6982542039528626,
 0.08649947879332454,0.3101178882453649,0.2105544516468253,0.6629694234590493,
 0.2684569918974353,0.6067946817789742,0.5517821765518913,0.7862516315497585,
 0.799032278412832,0.7767856555285808,0.18714891305455794,0.36924197440775763,
 0.9800240707736079,0.6076389894342413,0.6852660626062553,0.8986646903228761,
 0.1635851800737066,0.7096928974050977,0.7875212158763549,0.3417752812997019,
 0.8765052585262906,0.8474467468190391,0.6578251480287679,0.6344534750612438,
 0.6971752626364356,0.2564097099592687,0.49788014733335384,0.5398448301315035,
 0.6457743979868614,0.783024580307157,0.8375408907850883,0.33532103500258437,
 0.1441273396267766,0.727949172583269,0.8820308078824531,0.1841819448410582,
 0.6897065080149853,0.9578062257917939,0.47585730450679,0.5655498149336619,
 0.8240108290142565,0.11320676192280299,0.8673450136338718,0.9739935001472818,
 0.7554004907735817,0.6075946837725928,0.26895291929440024,0.5188042941586416,
 0.27910292136190584,0.43574644090285586,0.7879246197872055,
 0.28802616354773314,0.33303896224264695,0.6319561107859935,
 0.09863988413605629,0.06725530811906222,0.41924727347681745,
 0.5817034802871142,0.692250738940663,0.9055314862747715,0.7482668446341717,
 0.7270801973850238,0.5059961317724393,0.24309658288666813,0.9510792552545421,
 0.18256062103888726,0.7264330929272572,0.856337565945737,0.8086912247529262,
 0.25510135544666224,0.6142647045258537,0.19455448237479056,
 0.7575522887889137,0.21779682036416603,0.28305915990527863,
 0.7785930217278668,0.7558151782429958,0.3633703409678861,0.21441350102100443,
 0.7259383690381849,0.5692757363819905,0.4833357831852275,
 0.012688391791017661,0.721257773735364,0.21627002266928808,
 0.8374301548358459,0.5643176744352183,0.7543497566649261,0.9570312468364675,
 0.547478242640569,0.8732485404704067,0.15304869590047943,0.15424309570103267,
 0.9979242181013948,0.04612630785630323,0.0015501462925172627,
 0.27737797969008005,0.9033987394610504,0.6873867487768786,
 0.34231682736361946,0.11050576775315935,0.8773759726798298,
 0.8540357492764811,0.26562908372708893,0.0998364490418917,
 0.13268977563418283,0.5859588808741371,0.1441518227924491,
 0.24301919373000758,0.572430738273609,0.546481043159494,0.7995504805817599,
 0.4550455589260408,0.6415260887039261,0.15617372476198066,0.3784782311817728,
 0.5795715374563948,0.9888349437009052,0.5359766509510048,0.5493422548107354,
 0.617347618760848,0.27697972482414834,0.22786976438414874,0.5031499821651919,
 0.20362263385926105,0.08590018407501465,0.24777146195984512,
 0.013357922864500882,0.3266982567170764,0.5946038791496637,0.716159266131079,
 0.8711026280608156,0.07771609559262382,0.8503873227544161,0.8991643032655002,
 0.16059840213579635,0.43332682598762196,0.5292963647872259,
 0.06562870654696085,0.5502564605507742,0.31455823372548153,
 0.004945078046589613,0.979658128794008,0.41596445812910465,0.456576023376987,
 0.33447234831786643,0.46911668837895726,0.8506289081731027,
 0.9055849779914209,0.1240526952734623,0.08273980986199514,0.7725195681294869,
 0.7477261978334396,0.8027710421851502,0.8977704876955683,0.4711414700746448,
 0.9475541486573673,0.05727953763545901,0.304926125743326,0.6164258533652653,
 0.22924195018802818,0.3406373757067114,0.01792296436747387,
 0.07578792612707275,0.7449296697240582,0.5901993811898707,0.8175163214994129,
 0.7869277645501208,0.06593809260489936,0.7108997317481006,0.645959287447553,
 0.18922701283932375,0.056673852997333296,0.4444058206630901,
 0.3711727462251211,0.08993918231433717,0.9449632176709544,0.3731782417021451,
 0.6858544225702898,0.24801250513351658,0.22810000649400108,
 0.9412639314561956,0.9904400444806739,0.32155413496887497,
 0.36944933203642516,0.3642524397568295,0.38300253611867097,
 0.22587554791499798,0.06939301867255498,0.8960162140507625,
 0.8371155017389367,0.5751518460503704,0.1374749175172305,0.7302893438535503,
 0.9400941385104936,0.2838510103419536,0.05942166611640998,0.6692122659961293,
 0.2911903036115313,0.4277717808012347,0.8420859377838756,0.9093858069890032,
 0.4375991662406983,0.8508505566902038,0.8797296134812994,0.6492175387192471,
 0.3102222469665019,0.4564713818524556,0.10393422829803012,0.6024379171731704,
 0.49770803159277655,0.2722744465320779,0.4097799993563134,
 0.24248469579276088,0.5393427968343314,0.9450007198946008,
 0.18110215043022238,0.11828155145791919,0.4184594691673722,
 0.04601680291402366,0.03992297231830266,0.8307911635683288,0.669275124323657,
 0.7946409728315371,0.6945377284391936,0.31205041434351455,
 0.30357521739985305,0.21165242202026957,0.45831733554586995,
 0.7158099525556472,0.6945379237468536,0.08152566967324526,
 0.40857521851265377,0.4814294142404719,0.4773097895044377,0.9174714333045595,
 0.4411492072849217,0.23635315673177903,0.060665508380491784,
 0.5998228476250262,0.12802214532688938,0.16232513062622855,
 0.3208699999643805,0.950726731176683,0.764692392601627,0.42300263127326143,
 0.5531293805442584,0.8615439068800927,0.5380430843119142,0.533068163841596,
 0.8988272456809725,0.21271624303528203,0.6593651683244045,0.7648948284692834,
 0.3997455967557302,0.08874483730365346,0.6844986079739804,0.8934435681515073,
 0.8683034948954061,0.6271052723910296,0.10843711853981453,0.956055492291319,
 0.16869923376123896,0.6157097898108899,0.5841151335491168,0.7503067399242213,
 0.786824400283427,0.5693139057259575,0.22615912340055855,0.40043959325354805,
 0.1668062152586386,0.9368082259073063,0.7705737396426579,0.34236367050999583,
 0.3301258672342371,0.17876808307184255,0.07279258191847271,0.69005770135906,
 0.4354485216120393,0.02866086897519704,0.4415143078330189,0.840339740873752,
 0.5593219033024671,0.6791007204538325,0.34020973914811736,0.6768970826059956,
 0.06242977506277436,0.15259312898592747,0.6896780975265463,
 0.03531229136353342,0.949318904138224,0.6202816421756778,0.3147639364820488,
 0.5416690244683101,0.0102599206810714,0.03241675931492061,0.4015753179979865,
 0.13091750805897762,0.5985150128814797,0.13927271218772275,
 0.4932388515908531,0.5353218563899576,0.11859196936130023,0.9344765447956127,
 0.6430747142907224,0.19277596322651835,0.22685928222102492,
 0.6189276090763384,0.6201315160778685,0.33359022958924,0.5990323799766744,
 0.027983572123640665,0.7624429518222768,0.7330543232951016,
 0.18681047588119903,0.41531452279266146,0.6988918938556437,
 0.5526383550283485,0.39226576949982406,0.2383835804759662,
 0.27183827880458045,0.16537605906464736,0.011374667856607679,
 0.6941739649276155,0.8110084702894119,0.20186864241579972,
 0.040775501964712246,0.9840480198659607,0.5060959626969029,
 0.7983211163804704,0.3954635124750757,0.9189349337854854,0.7530585098040053,
 0.12869513527039345,0.07047468730847373,0.9007011654055717,
 0.22346056817108673,0.9212880525097993,0.1383298362294173,0.7182064868135302,
 0.9460657229867598,0.7006407268256143,0.6513140155527798,0.5730798082272341,
 0.14056904410655313,0.1285995664850882,0.08477028496008698,
 0.5455008280105784,0.8255583068409792,0.4624427459939995,0.7476012109079688,
 0.2883375893202632,0.5497321878964512,0.46422971737892316,0.7909139829178077,
 0.19779506179438455,0.0476231518577247,0.34190672784527343,
 0.1364535721033553,0.8111783299916139,0.44748718302642687,0.9609482499727763,
 0.2380324621557971,0.3815727019742564,0.6952511152322347,0.3373527201311012,
 0.40851377569876346,0.6218263265233472,0.025894806209775023,
 0.4854592157459169,0.039750364862113474,0.7921711222528128,
 0.9770951649400476,0.24565752106567917,0.6272228913311961,0.6818981165552681,
 0.6512680704285725,0.08704620734551427,0.8801381421585208,
 0.12270498026776033,0.5343091841235982,0.6661779111201186,0.6681267617084159,
 0.24316213271839948,0.019651411470919533,0.5500685344080358,
 0.9859932034312164,0.6042582676303003,0.4827370835658624,0.5575881023323365,
 0.5314205511501411,0.11327763715016059,0.11948828271882461,
 0.9223170157811439,0.9506934089253484,0.9605442746443822,0.32237731961333127,
 0.6278566682042992,0.2630223988804078,0.31634589250123923,0.3277159069159016,
 0.9197184570556439,0.5955027219649698,0.6196220474767334,0.5897156227923214,
 0.023297295741058432,0.9716357436208656,0.31852367742988885,
 0.2910233368357271,0.19499162126186909,0.6160719794095709,
 0.14411482696320843,0.3713567245654382,0.35920630891318794,
 0.5026488852432155,0.056925045033176414,0.23183505462154697,
 0.3423334238816921,0.5921937063947945,0.38824770870165526,0.39483943548345,
 0.8078302979316361,0.5779131178656213,0.8707663149953211,0.8627377467614012,
 0.49150150824639516,0.749773447876839,0.8833404977196345,0.2150248023412099,
 0.34180741510734425,0.9390137839409383,0.5626561736167958,0.4427810259724487,
 0.6363234184297362,0.31858974144594043,0.30803964300875897,
 0.1539693838614461,0.0414448442263563,0.889729224441679,0.7164049691764315,
 0.7572694245557677,0.21512232586359226,0.4524553936662369,0.7347360373345897,
 0.8006912490089837,0.5435179552065752,0.37075462033024276,0.9995681882859402,
 0.38297088828157133,0.34487981755347696,0.2364303279788803,
 0.3052721821227944,0.509906645034374,0.20175631083440826,0.5696460115718509,
 0.47701528351484024,0.9332743840634253,0.5254807034783476,0.5014562288784029,
 0.1713402532206072,0.6453217994207556,0.7654952336523629,0.7634717619785159,
 0.06942081033794656,0.2636872431245947,0.008263718339057702,
 0.2605619309152162,0.8326997132001389,0.12631717627603756,
 0.02570741935476295,0.34633112600087324,0.06470525964277041,
 0.8214958442638458,0.9418961381970367,0.09902988423816395,0.9233765980586356,
 0.8403590713181764,0.6531273452966798,0.5816736107374659,0.5741788011834905,
 0.5243554500112372,0.34760111198372445,0.2525045856722259,0.6650991831904678,
 0.3475790755874597,0.04277598626231127,0.6832858949099103,0.8117340533999957,
 0.2595132036320936,0.4269682453691024,0.3539524892470114,0.3610681391249102,
 0.742066072354408,0.9463735304563097,0.5442528528220205,0.5728417419402101,
 0.6635787062744314,0.8295060734511303,0.5719472448592258,0.04434054155926048,
 0.4689950870805484,0.12629976697089307,0.4358030164324118,0.6372548572822846,
 0.06496536696934452,0.0879139176433017,0.7259460429612026,0.3836654555906922,
 0.5954669912885219,0.45951070076508915,0.352483389110375,0.2972141184443816,
 0.1872268307008884,0.012557040995586743,0.37993663947584866,
 0.47595099311227695,0.40704358332374213,0.47059725666148333,
 0.7645545102254063,0.11622216880891134,0.2926229988482615,0.7264954570008143,
 0.7613481295168214,0.8838800964507956,0.37499009683836737,
 0.40030190375921126,0.9530867518630549,0.31258682595899256,
 0.23638938931120668,0.6210452388599581,0.5088672687065159,0.9010739564386108,
 0.21723504676299188,0.5847674279792434,0.5537851203200784,0.0848653528488924,
 0.40815958254466655,0.6051178291117515,0.34887367182675977,
 0.9218911902993522,0.3926434399598008,0.974333462699226,0.14262795653630467,
 0.4964152174983947,0.457600348183149,0.6183153861184096,0.8431249071358363,
 0.44558260214949197,0.6322474085092887,0.8080002162876978,0.6900913939023032,
 0.5874017101569677,0.9072255687780142,0.3136432210393416,0.5908609724936261,
 0.2886071710001513,0.25741681550033846,0.7559804557601775,0.5397405207585466,
 0.17440330415431826,0.711452384884786,0.70104648569867,0.38345670721616054,
 0.35624442481059493,0.3881555579782787,0.15303586679441095,0.578126806629633,
 0.012631619250929371,0.9900648944568456,0.5924277157846438,
 0.6636102876350838,0.7324910139608403,0.2635838445279338,0.14653739290128898,
 0.40409869240978036,0.03242253694281394,0.7382510581626535,
 0.5289913631028509,0.7833167622985674,0.1198171899309246,0.465465651797924,
 0.4292772111165709,0.7531311150972919,0.9408681252461019,0.4795192612058513,
 0.6545712412588791,0.5012584347401258,0.2579334274704934,0.15709498004201472,
 0.25815899891403793,0.09235688425570032,0.5192163206457635,
 0.9193161932639088,0.8354194016172924,0.9757356906702253,0.7982652035779086,
 0.7349227741750469,0.046938593906094006,0.24088725848238335,
 0.6014934326360618,0.9114779554790119,0.11020555709161228,0.9258021252947247,
 0.08759602639394093,0.7910299172294358,0.7201140331343135,0.6275257857737438,
 0.2546769268911142,0.019307350459648154,0.31220452362830686,
 0.6763207044299983,0.12335074780242761,0.620316291078945,0.41239765349429647,
 0.8419990981425084,0.6067342552216808,0.5157308335819655,0.7498961097195678,
 0.14578219927349423,0.9852765450167854,0.8255527778892281,0.7044948968687087,
 0.49497069188680565,0.6075963377561862,0.8576901403533538,0.4480504800892512,
 0.6335675947680919,0.41083170417815573,0.5203107935756097,0.8947306319158719,
 0.30153698245639626,0.07540211641928596,0.9615159525548641,
 0.7414134988888753,0.7878360519093024,0.03795472062783489,0.3307601841889837,
 0.8554900381019039,0.41663027110206474,0.5441289409324539,0.5042448893101953,
 0.5512149830793801,0.32205364325147023,0.7642232947977317,0.6041785871573249,
 0.5768166441060575,0.7642988159621885,0.3451995563712553,0.9642079527104215,
 0.30989400460928906,0.3626839718459792,0.9617543066915982,0.7759968430166715,
 0.8886892052276101,0.34086428597509755,0.7524683175500916,0.1527425360368142,
 0.2644762943379513,0.098690318432472,0.022700492397513594,0.5328290007256486,
 0.9647687211496258,0.9963114258363678,0.03333048993824517,
 0.12991194233162118,0.6054798943712958,0.3724561520913089,0.8696788535249758,
 0.3342893097739691,0.1927989802573814,0.7272291743627098,0.12463261430219097,
 0.933325059814235,0.7766381714744678,0.7105908570608277,0.06124732260650456,
 0.5711332972094253,0.14854298688478695,0.9724745294583156,
 0.004304354994328907,0.08332948000786355,0.8707206466331043,
 0.8548855226456334,0.3721590066068977,0.16013163417991327,
 0.22032961797025208,0.6644520121221806,0.2758933376965301,
 0.16503570464560358,0.3062247654174026,0.5248245335497899,0.8107089865565669,
 0.5661914489734596,0.8651343325518943,0.01941805428684351,0.7709836602642084,
 0.43747762935194534,0.84814595051456,0.786299962848125,0.9898770769787711,
 0.21735102356459313,0.046743304545959874,0.2767338583727206,
 0.7637575607539353,0.8719315012051554,0.9422158042213553,0.8902982019817804,
 0.41559906161293003,0.5611053354682521,0.6074665407355585,0.3664321275530652,
 0.10429456470017784,0.8154692406034052,0.8453907563570953,0.7501653838903745,
 0.46853107579439146,0.8172884432641074,0.12051966085104282,
 0.9295805153776716,0.17517613325988102,0.6568631633872634,0.4750434608692725,
 0.4776659756681495,0.6262885582639477,0.31871588075652446,
 0.15752433839156632,0.05918381013017404,0.9327377562624046,0.758538160753262,
 0.39719113438546794,0.4637423310103469,0.9346913103427903,0.8420866625360907,
 0.909213189654074,0.6761624843108397,0.8475087691273848,0.8073574866796098,
 0.6688837847813129,0.8780649090661292,0.23754526878434945,0.8439028340189707,
 0.55225304054136,0.9598867432041177,0.5031348111790713,0.47355028442431935,
 0.9152218974328364,0.06531533617298746,0.8339612548071125,0.5454793358571934,
 0.8990773890079916,0.01978059131256893,0.9526245237345309,0.5605352554301428,
 0.7283616420373918,0.5345128071503008,0.3786203740172214,0.9243661086118026,
 0.006808352584131483,0.2642298168061503,0.39758056860975166,
 0.170575961934048,0.2342077095775904,0.49600933961861604,0.3483166850590623,
 0.35769512724808583,0.16925860689377015,0.46846114913077086,
 0.45399663573918925,0.35802642714873034,0.1551930785350749,
 0.6351596889039185,0.4474859597101184,0.4621967581055918,0.8166885904558349,
 0.6187490783042771,0.9336733988251432,0.8741667172027417,0.3795190704591309,
 0.041452185553283005,0.3082760470487055,0.2873178335041454,
 0.8705525662078349,0.03281878476178646,0.5013226670113557,
 0.04075052783312927,0.17681957246966828,0.300808646728026,
 0.03591050152505004,0.284683571267935,0.9690577882337486,0.20227450029159733,
 0.4710867330232761,0.8821646637126009,0.42395337416215395,
 0.39922344449658653,0.6315307421731994,0.0529179696197688,0.5322016327447507,
 0.3876094546463146,0.8074597016667148,0.8340846342016914,0.3080722330937997,
 0.17725207890797046,0.006574418344057031,0.28982533918313347,
 0.23390318097246854,0.5932987106137937,0.585949898613896,0.7235072002998942,
 0.8712918468479568,0.05610332487271452,0.4523160933015027,
 0.17718909496226565,0.6713010857250851,0.6008680660751599,0.7512774625067785,
 0.7736793299335183,0.4728198275794838,0.9014536354650585,0.63928127421673,
 0.9108192137534963,0.2895844128839773,0.22991294078575564,0.2272385628382103,
 0.38038509839885215,0.18701467932809623,0.33674490801180434,
 0.9891180667446259,0.4195759659218613,0.11509333391934518,0.9360330645325212,
 0.05957895801181434,0.9358724231598807,0.4393238090464092,
 0.23809461680679922,0.4476457466765046,0.11253061620064442,
 0.3628908223843478,0.8822189149211221,0.2692851527899549,0.3683031728937032,
 0.7587025252033018,0.3915319218876674,0.6502726311461232,0.15942457518302844,
 0.3545172922616254,0.5791180103204763,0.5733016651510634,0.7576459068611768,
 0.7967430006408495,0.8281681874209637,0.021848652476620156,
 0.0010009561542613232,0.6524673282052866,0.4029910711928968,
 0.6471667549224766,0.8732567787271361,0.26843551871311955,0.7343072971977437,
 0.4527071491227289,0.472095368719624,0.4596152234042489,0.35203298928873317,
 0.7968946793169077,0.6278921102466147,0.6205934763816302,0.6953274456507144,
 0.6602859164256751,0.964812100786841,0.5294135296110726,0.27661366499244755,
 0.28856318828786265,0.7776320585792369,0.855146685962584,0.39999731493543433,
 0.9677451481084728,0.024819924272562632,0.6620775917218601,0.671110267241028,
 0.341536104398632,0.14232925793399964,0.1826183835882449,0.22458444799298716,
 0.22804313678344612,0.5999635551488632,0.7221130191804579,0.691432483340226,
 0.8797319620885438,0.5496374173495018,0.6170652927313425,0.811321428509792,
 0.34103153146463105,0.9259439963821752,0.9703690154543122,0.6441243742451003,
 0.9292525209858589,0.8577018386227026,0.1312955170940413,0.6775226640855543,
 0.7220030453536826,0.8131572338844213,0.4661215961960231,0.22286500993222902,
 0.21591653707966774,0.07088660390692014,0.1354924216470259,
 0.8311763465917298,0.25169282284275163,0.35843870074999895,
 0.4069745293712037,0.5924859745392981,0.8409131888038267,0.5702365780684765,
 0.8206973995751747,0.6349151365975071,0.8964411817250737,0.8614387798061126,
 0.5655111147927939,0.36926567750161887,0.4829145098632148,0.6364086841955982,
 0.9100074465545577,0.6925685686978647,0.07497144907901254,
 0.46794900358546987,0.8140262108124577,0.08310376528741703,
 0.7787709057245074,0.2725715069012522,0.48665736408207605,
 0.45882930985759374,0.5260611219505172,0.37313204050848936,
 0.31714887545888093,0.25549677531811554,0.8437342174187101,
 0.9016761024246479,0.9742916245248878,0.917130730969762,0.149989731807591,
 0.2621844458373066,0.8529119249035217,0.30729038125859387,0.2691328114127881,
 0.7406379086435594,0.8681732189639093,0.19941203551811815,0.9186946572722221,
 0.5588750361065823,0.39085563304312,0.7238346658039421,0.30975077321675326,
 0.14872593730399408,0.6168570533657238,0.8960251012294178,0.4529093342946693,
 0.8370635104539398,0.37614179758748956,0.11966490018587705,
 0.05838647311956768,0.43284305182890104,0.649137243104029,0.8561157927909017,
 0.001381337165278218,0.6309057262623169,0.2419940437983673,0.586726117855101,
 0.9393615357571593,0.7205699835208426,0.8812373547640757,0.7182172106143501,
 0.1831670810819912,0.9597297833516673,0.378340794263509,0.34564937273134333,
 0.9205993136728695,0.712031481594263,0.7248730225444162,0.2704425117510061,
 0.11015207041613828,0.04643718735831637,0.4283702806231888,
 0.47680525012160335,0.2588993447225152,0.7531740602704512,0.2705744906289661,
 0.5676101412908326,0.7328191307191139,0.5765835635550265,0.11532341894797615,
 0.8669704339338237,0.42059853117996004,0.9571259520438711,0.908331868201346,
 0.9697682622461842,0.09233645328287865,0.12654712182795913,
 0.7698554164033198,0.37306656097811186,0.39570975519372664,
 0.3592836268577988,0.729238593295464,0.35825939308078236,0.7617064016627668,
 0.5951884164123679,0.7160676396481097,0.9338794082749856,0.2742079255635784,
 0.8633661384473759,0.40920706270929774,0.7208622986047251,
 0.33604910303095836,0.3322995210730778,0.8277001522605356,0.5743943493673987,
 0.9232923366840956,0.6357678897436871,0.5322460487690538,0.33582184429485984,
 0.49467127394130705,0.027019525222939378,0.9654525246039322,
 0.9761477503249159,0.9146980690081454,0.8101695461949062,0.4897486708743558,
 0.37409179085961053,0.11854128000253583,0.4334714454532922,
 0.4186997044617202,0.35488469041885407,0.2883314117019625,
 0.17656320510589485,0.18853854272888326,0.2982079415220472,
 0.8148854943176693,0.5975716622579099,0.8889618763647853,
 0.049606118207348615,0.8995885018038825,0.8313837097204404,
 0.21956593040359906,0.6399113225197034,0.6555316894933109,
 0.31113441178781875,0.6835481528977119,0.6240037253702722,0.5399430356573516,
 0.10892119883893314,0.8734846082902498,0.08153182240395862,
 0.5604774047096209,0.1139812406340085,0.2439347305407481,0.6753325163969154,
 0.6749624225578018,0.24594790348570217,0.8031333113356189,
 0.0038130365045538905,0.0889030627773264,0.1980261483211858,
 0.07720573403427877,0.13574599783386265,0.99915163295534,0.20433440009403603,
 0.8915188675345596,0.2208713346001815,0.5845223876382316,0.20493976763352206,
 0.24560619278250329,0.35486196100251843,0.8035246468541513,
 0.031841235692615544,0.5875098360199713,0.7805947323997396,
 0.6593850831662921,0.6019439492372446,0.8840063704436767,0.7270775959754576,
 0.32968092468905297,0.830391503882746,0.9367089115965154,0.2009011907337881,
 0.9346249956172036,0.28715325566192296,0.5540386231132666,
 0.23079484853859233,0.22460787431853335,0.6360639401599814,
 0.3924866777730035,0.05632986789395733,0.44163268355866303,
 0.7402390043562324,0.5229429220320734,0.825035605496308,0.8311689532367954,
 0.7990168421243615,0.09366819371100077,0.09977776396341376,
 0.8785955421292616,0.14576654900184594,0.8846777583221049,0.1602636788491577,
 0.22024158014440753,0.7387871862631583,0.3584037400877145,
 0.42757898437547615,0.26513541329561374,0.10513382065546262,
 0.2440992157049634,0.24572753956601967,0.15443037329040954,
 0.33008385514795924,0.886335138992933,0.06540281001327453,
 0.07034888535862727,0.9608859674859662,0.20147042507383128,
 0.31681115330959486,0.39238611117931943,0.639503756043022,0.7063147972447036,
 0.7065149179579325,0.450382887981992,0.19109084505758467,0.17376985037987513,
 0.13736632982919572,0.7182395585553343,0.5554650883343832,0.3448298471069762,
 0.6546512053714125,0.07397945165243591,0.1629347877569559,0.5278214560288912,
 0.6147113450826871,0.1297510182918138,0.9706500619380348,0.29283398348707856,
 0.7211682465609365,0.606971356882423,0.8071500780836951,0.9630757456121395,
 0.31526523587720734,0.2586231812309506,0.3949226416877103,
 0.19414486525963603,0.9182873502999394,0.7117222110995224,
 0.13198250894953312,0.5225531400294146,0.265702120125889,0.6646575483016148,
 0.748271012970223,0.5871603426593636,0.08058382381535134,0.09013031512208747,
 0.8812162645368607,0.13329818519766257,0.9804222537134831,0.8316051210124962,
 0.37682688523228647,0.6086098562586154,0.6711522135836385,0.5271760788308014,
 0.5550267231813559,0.9952330464149539,0.6613407158023863,0.12810321940908764,
 0.722066594401361,0.4584437662291737,0.2546269153159749,0.25048686968682654,
 0.594394660977801,0.0665417724006453,0.4201417532527233,0.6582025882109569,
 0.9509424090447336,0.7672755320208824,0.029669983000586875,
 0.7106407805519217,0.4886540547810573,0.34642693543221925,
 0.36525496375373057,0.502956123342765,0.3233300959596279,0.3664591652229179,
 0.6745226423206843,0.9527920763829171,0.1740766557711625,0.5239491576544352,
 0.2558261855574421,0.5343435449350763,0.5605747818764477,0.07767089729857224,
 0.7627693842317937,0.0023665646143751218,0.21060314635834332,
 0.49457952177242614,0.4062473669526159,0.7172458829725885,0.2886270377427622,
 0.06808675680268139,0.44237955310958255,0.6531518042944265,
 0.7238793822787912,0.4624991200604638,0.675498621864957,0.22740692572504484,
 0.5821869929296262,0.6626124402723668,0.587043039879127,0.9489684258708188,
 0.3811520707965905,0.5391494106438063,0.9686227092044373,0.1671710913429183,
 0.6450185442741868,0.4720201921086211,0.2537420819851136,0.4467048365567722,
 0.5654993885509144,0.8903734710549018,0.8568750188318653,0.9205995125956213,
 0.2299679913712489,0.844282178572458,0.7553199935443028,0.5667941121243211,
 0.4147681871494793,0.3881115338675116,0.16645469473329366,0.4126330716444764,
 0.29549718309267414,0.39154796647080714,0.7875795426586492,
 0.49280516141881126,0.896567429845064,0.2398598316949656,0.7835778399869207,
 0.6572254003122262,0.510628296079726,0.22738951871726298,0.8951035808142723,
 0.7861008515361128,0.8204384765122309,0.4203062020314804,0.30981042973952655,
 0.3431659632701536,0.29801771765613305,0.7134071964709419,0.138940315174933,
 0.008365718564380953,0.07140419786721863,0.1694545940202966,
 0.15842490973682166,0.6144567005920335,0.8229300545494503,
 0.03933234975129274,0.7429904962498186,0.33709193529856496,
 0.3114298087639066,0.15694502857558656,0.46684299731597734,
 0.28769439521097206,0.8140805056962976,0.6950463072945412,
 0.08852735868747919,0.12492827959991026,0.7707206439590536,
 0.7529732419126212,0.398390513509456,0.6441089254781575,0.6011332190837093,
 0.8883654690878824,0.7687274041836321,0.8856727533731953,0.16871762893987075,
 0.591369683529811,0.6388610608795577,0.14631397400627688,0.10120972481160795,
 0.5760241696468112,0.9934098071123783,0.28697117186494503,
 0.22113706135296396,0.31050365239392275,0.16962535142338264,
 0.9197971959654101,0.4744628527246022,0.9627813183930881,0.3530532046297725,
 0.31260643080110906,0.47885138657134185,0.08086467664379648,
 0.09421606191389909,0.7763122778723419,0.849023651207059,0.9751275191323837,
 0.9878203041651035,0.10126035459650207,0.4997855229445636,0.7446634284566208,
 0.9970697961205424,0.949994383547107,0.05719330070207673,0.5203684692633276,
 0.8394177647627797,0.6997791289265342,0.2025577832653067,0.05986817988667692,
 0.43105670270595675,0.25096066477373,0.6299064227734298,0.9987453789722016,
 0.8574684761952617,0.35620014137934586,0.02269882661395828,
 0.34850314639822577,0.6877286072810764,0.9201478283105539,
 0.09838039628678164,0.5398679140266592,0.9855112376486681,0.6005996765624668,
 0.30764201193930996,0.3981719437643355,0.5148348664074575,0.7128850590879869,
 0.7222089745313314,0.7162807276702821,0.7308328403856978,0.5411181769094155,
 0.17577049670756595,0.18141345440851409,0.41508078597607134,
 0.08681850299203009,0.3437331398592336,0.4938479564445931,0.3544614416817391,
 0.4443729343639343,0.9246518908883469,0.3290730142613445,0.5112730420531273,
 0.3351557086725129,0.7823325358510445,0.043454029964541396,
 0.9470599829756416,0.2981139507472954,0.7957838788153636,0.7257451260014466,
 0.4814548723580534,0.5004755224007242,0.280313891121147,0.9060336044774037,
 0.8684914659414389,0.49478274366470376,0.5358034934474054,0.7817780554290628,
 0.39575757419162183,0.3854021458614981,0.5645636072230554,
 0.17542328498725956,0.4371306869333329,0.61814739661658,0.8976699485692977,
 0.9997821763553533,0.3683998144183923,0.6183547946110147,0.3677914747219986,
 0.47090938640283253,0.5707365959177076,0.8293436594191856,
 0.22660636396824285,0.9891995335341323,0.6426708570230565,0.7854793964912752,
 0.4244938273207114,0.6175696199527083,0.7527577590819069,0.8099369990415707,
 0.7124278900403186,0.34543848015545986,0.8188478851037384,0.6233199954010166,
 0.8888878910861316,0.2600927010578742,0.38444227204805903,0.8702881838082088,
 0.15614575249919171,0.601862834141998,0.49789980443554194,0.584646319321947,
 0.6346434747394719,0.020130777300899738,0.40848396298289247,
 0.07450734088406041,0.4511858021134565,0.27421026454966124,
 0.5578370191083658,0.887712007130258,0.1210509646564698,0.8390058512689336,
 0.09889991369265771,0.9926400370099224,0.8914618884912854,0.6058388250408075,
 0.27545829519270193,0.6020598964094293,0.4372948914009036,0.7396350719471276,
 0.6826102438365641,0.6356086476287742,0.4746984091855114,0.12345559060858968,
 0.4643207353145874,0.478057293001525,0.3662632872114029,0.05785547903540833,
 0.03631795408039806,0.5197178723674257,0.3184019981169489,0.3114828770108017,
 0.1777932250801082,0.6723225277481585,0.1491535104252416,0.5077131483851006,
 0.01384416502616126,0.6104526853719571,0.21391122917193683,0.341444203449894,
 0.5140918038358078,0.6764240566782799,0.7694698111366319,0.43559957528025117,
 0.8268721625681091,0.9137974962275318,0.30063796603105686,0.8570368248597725,
 0.5470908683201603,0.9590194342130466,0.5777169264977225,0.6656665636271577,
 0.5375016508341847,0.9447218470589888,0.974383785984952,0.615158628938814,
 0.15318244459697483,0.21624797982498334,0.8279997422714251,
 0.8831638555653204,0.21341388681580975,0.07127546962909359,0.158602034077737,
 0.6543377674605538,0.7411411313540636,0.9430635875559498,0.9255100476609349,
 0.541223751237903,0.1723122153253177,0.8192162994201423,0.6729314906168491,
 0.380900319659347,0.34787694646385503,0.6741515576537314,0.36081693884763766,
 0.8430757282620156,0.9262862216094001,0.04412026397107505,0.0945676710154294,
 0.8962608274330849,0.49601526457225387,0.5127891889251361,
 0.06725765958115071,0.5846517286394093,0.38099167881388896,
 0.7228096733381749,0.8925602126077415,0.6367074801655503,0.9938451409203773,
 0.4710366482388634,0.611848965222347,0.4246223020139065,0.5248513467395463,
 0.5188686552127146,0.7938511175847257,0.022716942026081988,
 0.31293440505830494,0.14217377024469524,0.9002197898040246,
 0.36663529344653856,0.10315350668867052,0.18922657894152606,
 0.8170740692203056,0.03272916387882385,0.6868335320088826,0.7704574136316555,
 0.5829277187675287,0.6401162212530023,0.9512425628491351,0.9963695644953647,
 0.31410775154164217,0.96433392624238,0.11175812424408693,0.7626892812399726,
 0.2454047509694799,0.9033529462889143,0.8690721269499155,0.5974366768994004,
 0.29935088336729765,0.6609700124146326,0.43432676996003694,
 0.13667678060357513,0.625005338852606,0.6174235252201477,0.9985703459163227,
 0.8165849082109772,0.5423266911082547,0.9825112878493147,0.8837213153242012,
 0.5179761342035856,0.8103823983066087,0.6745923262104904,0.18668878632961516,
 0.23365680090004892,0.10071583339138956,0.7116414939707614,
 0.8084401790440572,0.30168250538704233,0.7836567422529137,0.5641867004562001,
 0.5953444112407813,0.18179082448160244,0.15357692183064453,
 0.8412225061217109,0.5535911650107121,0.3847734519813232,0.7641287772080115,
 0.6185440016209206,0.3135152338562216,0.8201995001093543,0.5888124124487519,
 0.8046414016238447,0.9689268214929638,0.9905705600744051,0.2776721059972449,
 0.05844504499594705,0.7331413299272578,0.9982753476247385,
 0.38100724590181956,0.4863115258758628,0.8459903475955484,0.5155400717908836,
 0.14206364668273475,0.04128474534401827,0.46264639892408066,
 0.1653508243682661,0.27133905628128674,0.684557749644637,0.31218057026511536,
 0.5592705294530361,0.9997571333696172,0.8900486023890439,0.7508379339903184,
 0.27982774109038233,0.7281610561309327,0.1459463217678243,0.8662158611430115,
 0.8694309541484192,0.559762759955656,0.5413484921786703,0.6035846500031159,
 0.5557627018431407,0.8772598444157327,0.9628376492028969,0.8140727892074542,
 0.15515872770439043,0.3475737901126297,0.6053551892712722,0.6757331360738525,
 0.7567321803065504,0.7562812925335991,0.08296234529663293,0.8310102538468684,
 0.5444042685578449,0.08692554035520983,0.45725052574202885,
 0.6968195539735836,0.6475877889256636,0.24725935875021743,0.9140873612629421,
 0.34312728768383716,0.22091238752867537,0.49992504269785754,
 0.2517604169466434,3.0175070567939333e-4,0.3624219721675581,
 0.6117736267040688,0.26442118382835367,0.08589282773068263,
 0.5996958038203452,0.03133409503366136,0.23967754151147636,
 0.2935951653518313,0.8571447127270713,0.18552908740629792,0.6286783769503329,
 0.02883590056986507,0.9235010555917278,0.8530724975361782,0.4540425062326914,
 0.3141264291378042,0.8528582543405473,0.9453647103083946,0.2548969776292578,
 0.8622483369549208,0.11585477050467796,0.04588116813259824,
 0.3275826170884961,0.780651329387042,0.5377946106502403,0.5215975899831288,
 0.787585949138689,0.9443714966176204,0.9460462768516543,0.3780895819971304,
 0.5025751674882519,0.41256932181400074,0.3198872474158324,
 0.23787011193882224,0.8274303901096383,0.6921864690727637,
 0.09579174081845088,0.7589378688673165,0.17318993011717965,
 0.7839130085076591,0.7858277345020825,0.6016625741826939,0.7173285097519854,
 0.3573506546151971,0.372979333420841,0.6105245109768929,0.8052712431544636,
 0.4068035347020602,0.44383515610568525,0.3926897913785973,0.5281241201018895,
 0.17049952077675357,0.18196127523221273,0.6377435998482198,
 0.8304045252129832,0.4291767045463266,0.9431755385214444,0.45372856271546613,
 0.30396687290538704,0.4793611397437215,0.8298809150993022,0.1474825350451956,
 0.6371525848036365,0.10162050688190138,0.979684637898027,0.9732618485144116,
 0.5500976326753385,0.22467054904529538,0.750984251899564,0.05447299230801972,
 0.30557761279452467,0.415245230700659,0.7200346522514715,0.1468570714938746,
 0.46566409716126067,0.5247010385226749,0.9216920063743186,0.2697739207900921,
 0.44969372106698946,0.39979267435189125,0.5226281691057733,
 0.4032664915586123,0.5586112583079599,0.9952736765824146,
 0.050238445889611905,0.5652500020309278,0.7498522111202449,
 0.004798875604504649,0.9329653380230778,0.5084533730060551,0.299063845038519,
 0.06394907998898625,0.09239583974454724,0.8194841046759139,
 0.39149633116599714,0.2966513609061391,0.05968099207399691,
 0.7094675152976896,0.38121063137721944,0.34843406858418446,
 0.35626510478687523,0.8733464321587402,0.5876893943850797,0.853422543490259,
 0.020151845274640534,0.7673666056934302,0.64043764686288,0.3882950585412641,
 0.8652623061521573,0.39507554717430593,0.25578568496508125,0.660143248099401,
 0.9722976514519136,0.5996883533298489,0.08075963811593523,0.8227951682254874,
 0.5426252899227284,0.03704687996308986,0.06492870234313686,
 0.032259094755782924,0.5291672527257625,0.7960095496893553,
 0.039139313435053724,0.2770794396212485,0.5045408208296053,
 0.9353024233255908,0.28390809800251193,0.4454445409103919,
 0.34813198113259924,0.022825786732559528,0.16309789390711016,
 0.6007265747304436,0.977706816097224,0.975370778748662,0.5418937461227128,
 0.20408346021079993,0.428577010604156,0.4999871353138029,0.15826578355547993,
 0.15845859769423387,0.4400617018643528,0.3116057645048417,
 0.18844911426744415,0.6505132951387624,0.09265492241978934,
 0.6019862918242144,0.2856044862077096,0.6109466183670677,0.58212323376097,
 0.6559158667962003,0.9580270099701246,0.4452363344185657,0.6976273581515845,
 0.9730896987852831,0.10185642388324978,0.3243954692037787,0.8370233567653528,
 0.09142131115403285,0.30418126107336063,0.7235928637312882,
 0.5004123517287521,0.5674990896549411,0.01919468953514869,0.6039810060605616,
 0.34816786171473113,0.8361136484914725,0.844927697691171,0.3823383990554794,
 0.9986705334509609,0.25579891625338735,0.5947578546070632,0.6908393749247365,
 0.2525717256590809,0.18157792127979,0.07241007018997647,0.06674688326334188,
 0.5349253630261318,0.6885374758243136,0.6410484261709408,0.25502599276203297,
 0.9943509563255684,0.33879149507646766,0.04894083287622508,
 0.5507991800898001,0.7822754122100437,0.6905365269273573,0.8749015161657117,
 0.2665530346373234,0.1958173095426866,0.9220193762321082,0.03591075121832388,
 0.24536731092755726,0.26498136417639606,0.21317735636463198,
 0.30635862145846415,0.28839285079634847,0.8061991502277975,
 0.6376196816959682,0.5926408129873946,0.9519871596100594,0.6325714129771675,
 0.33030223092078637,0.5255489946970073,0.7012417836185794,
 0.08415314922435568,0.9626633684719855,0.612646915406612,0.41182595161721014,
 0.7227712260210907,0.9737580352223599,0.49231308617158076,
 0.12253128560308324,0.3728778541863117,0.7021209950468549,0.6082285924523656,
 0.5684119472823402,0.1182711332266686,0.5032067368346825,0.9079696959704707,
 0.16314597580996493,0.638590838348708,0.05548272223523765,
 0.03437309371934327,0.4894728519246676,0.4066482080233662,0.2700342089850146,
 0.18368358438479893,0.6966795777869084,0.615273583235547,
 0.0038050756399339303,0.7750068020364875,0.182916085055423,
 0.25307208572121975,0.48098456261842226,0.38605995285808037,
 0.07126878066153419,0.24248972748435327,0.23938584675656116,
 0.22376750576079163,0.21747867802409826,0.5541574242502845,
 0.2154406117336528,0.7010989319857973,0.55674607311175,0.04668235567126622,
 0.7397567103664944,0.8376263256580727,0.3055060023829088,0.7176238738905034,
 0.16932881368282415,0.41767566946833523,0.7392976381020975,
 0.5847282578429843,0.6855078468485121,0.6385428861203221,0.7877708151050231,
 0.38616278468303966,0.0326438293788569,0.9137925611361524,0.7956298069089304,
 0.17391850427080624,0.815855602627196,0.19596181054424866,
 0.17924750356838026,0.6016328140725508,0.8744592463486474,
 0.09511297903055249,0.9925616250279318,0.8939442742107275,0.9924017250855997,
 0.27308949295903806,0.40046831041774644,0.5682311872292578,
 0.49217252436780656,0.4364697864760725,0.42635043173732357,
 0.39305130293276525,0.12898665971127432,0.8180121001926631,
 0.7967099071451154,0.9711214033164638,0.5138651394444295,0.21684043798741037,
 0.09463885876138411,0.7830995526237822,0.9541457132807749,
 0.16685558522825628,0.9280598631540348,0.8833798682229064,0.2806464559931916,
 0.7141873680636811,0.611046356822291,0.10487244969509579,0.25383871642554634,
 0.5787737765545623,0.2281821940489719,0.19319244560608717,0.1878107323160474,
 0.46301734809247785,0.975087179026255,0.5941574828375151,0.7073352338219028,
 0.3978930591746659,0.7602044858015933,0.31109160372755296,0.9054916933619468,
 0.169928348885668,0.4630006562546436,0.6906966138138921,0.45488002262848093,
 0.6280226765803415,0.8313230959460126,0.14212298154740355,0.5062411546574256,
 0.3697507033985191,0.15711134033819696,0.023368181678871114,
 0.5014628261485004,0.3822915214165392,0.14703971914433356,0.9150576021256999,
 0.6165702182794892,0.36671280619452495,0.21199412356737013,
 0.29785908382009907,0.26095249088539774,0.2788935487953059,
 0.20603022707517415,0.7204349005535572,0.11502917300882076,
 0.9479082220288618,0.3922023178739461,0.5014245385710001,0.9682559964043339,
 0.8570839302096026,0.8278487324118099,0.42868283610457003,
 0.07261559217690305,0.21411527998746405,0.6703479742360834,
 0.6297772974879788,0.9176037837805977,0.29814731771568637,
 0.09645140959608445,0.9949602860250338,0.48996490865375975,0.789242300231893,
 0.1657150226325842,0.1478467722581741,0.37389974551349714,0.7108051809676473,
 0.6262347007796703,0.76023996668545,0.5432136550482592,0.5078880117809638,
 0.09243213826169316,0.5836789035313137,0.017740110080497362,
 0.4151693259778364,0.5231131211667365,0.18287688001415414,0.3627024781960291,
 0.42101566383280076,0.8080389583164079,0.9381500734248819,
 0.23100138824684058,0.9236600052734437,0.22467618833110814,
 0.6151090304974264,0.7804677894849287,0.8822483935335173,0.3630006464470117,
 0.38274042506018824,0.16768788014406988,0.3976085839734489,
 0.4787153735891556,0.4228962960917173,0.5744995597000104,0.16971803399913243,
 0.9463951431327551,0.5969214375001813,0.5698244517288151,0.3781570679581211,
 0.9397074482393782,0.40059276835308033,0.22200459382212323,
 0.7795380061124119,0.06830629001650423,0.4039592358398705,
 0.056515547026303015,0.7046410296855258,0.1456231260433114,
 0.23020264387103961,0.951830794918051,0.03815381751484426,
 0.16524693520089695,0.17718522296078087,0.23763036774576218,
 0.2362945231786865,0.24179962805048238,0.5291007718351536,0.5180076195309352,
 0.8247511428075901,0.48573925658464157,0.716195425258862,0.4982395840861964,
 0.8905844464576446,0.9917193388784944,0.6613380615859114,0.2524694903828636,
 0.0383378712906306,0.25762330894965924,0.028270900444338443,
 0.4702882706080792,0.2673777151289176,0.19910788976308058,0.3899906714585234,
 0.2285240885469173,0.3428724808337694,0.26955226732737536,0.6190980068618341,
 0.9873533696321957,0.4496380291340678,0.5546601383718022,0.32352603029805505,
 0.3170236631918415,0.4361702254261952,0.492004917037426,0.10508446466247379,
 0.8655892013830351,0.8346984809286513,0.7155256995159568,0.20811705990265972,
 0.43142358733443176,0.25205542783957846,0.9333230902455879,
 0.29165737205957276,0.8112734687424593,0.10390136968111907,
 0.5743904588415288,0.5825226678447937,0.09000029209893301,0.9950905436623187,
 0.9481584061772083,0.758707149987926,0.39853656626739997,0.5914067485660965,
 0.15618811544858513,0.029649137783290258,0.6178654205565944,
 0.9414568515765177,0.049360437956720404,0.10764588374650674,
 0.999031971671598,0.5278337809971224,0.3271017501165099,0.5185963871741766,
 0.13547137092169692,0.558576326463024,0.10945966786927663,
 0.06921862908155829,0.21196842612254474,0.5718671580543565,
 0.2775052120595528,0.2727270583599568,0.6775898187943513,0.09965890202617256,
 0.023576115438209366,0.9270961638553233,0.9049236529997201,
 0.2815220642930749,0.9126619261784938,0.2980235953599617,0.1822227534279801,
 0.3073077674628939,0.18411811251939048,0.03478550967070215,
 0.12512049943067505,0.04240201323741444,0.23488879421301356,
 0.2474219596697006,0.26771268274021287,0.7954305179887675,0.6527997679818754,
 0.22185686499610546,0.686770669407607,0.04594608485891938,0.6872410463950729,
 0.02447737754276691,0.13753805586919732,0.9905253938282325,
 0.12997020436070428,0.44766242646335364,0.5979211608969763,
 0.4431479008814351,0.19104330908220046,0.38293820083600494,
 0.6077275362867312,0.4728240894693162,0.28428502182839377,
 0.25355917030147357,0.07915698827151818,0.5395153874059087,
 0.17248699664141298,0.36732931451046935,0.6984179317838223,
 0.2164501602477611,0.7261404856092987,0.9042766342211074,0.8136747418227813,
 0.45142971249472974,0.3782888532594453,0.6120264205165243,
 0.16206085074001741,0.9793193386903193,0.42819714371723727,
 0.048194071038935915,0.5310621064887364,0.7040840977126318,
 0.8922864188121286,0.1963108384090445,0.35527226462491335,0.4802792222391662,
 0.808720328990762,0.309497744307633,0.8783781819451136,0.9784644654412104,
 0.842807024006476,0.9496456711110035,0.5259357179505505,0.8842725870429381,
 0.9952383526642375,0.03325707931623689,0.42226228275593247,
 0.0324568442740103,0.0958434332380993,0.16541821091072106,0.6469384990748023,
 0.217567612381813,0.3331221693681017,0.7629785892062244,0.9908142628487051,
 0.05666528068411425,0.08847055374874269,0.44407860305711555,
 0.5995249834234434,0.48197155577335027,0.21800638047148313,
 0.5092322309769732,0.1662554612404148,0.8713795844976973,0.5354317360841568,
 0.7320464941850946,0.70461034289025,0.9479537643640451,0.2577063885220433,
 0.9596659098457416,0.29375969997152285,0.7772803176856365,0.9705767528099237,
 0.19247784858063344,0.5093031828371288,0.22462713808584234,
 0.9801609902484003,0.6211798759322835,0.9991443477042197,0.513348159573545,
 0.8858863558046197,0.7097318564077871,0.9811800151483627,0.8270555202834271,
 0.7977750684925258,0.7864326583271815,0.9767457948852729,0.4470910193540508,
 0.6607342901907067,0.10013781167621216,0.875872798398355,0.4283836068295436,
 0.22862831388887073,0.3199990279985382,0.9712659175819349,
 0.23017323511083554,0.851641893098544,0.22022199341684878,0.4015405200207812,
 0.3238027905880604,0.4902618364273079,0.8099385776566588,0.3993842153612387,
 0.5226375055290727,0.5040993038698245,0.6645900798113944,0.9296874540983893,
 0.9077822682941494,0.06343715336802469,0.8184325730725561,0.6357792795859938,
 0.4142158766631697,0.027427964324283405,0.5901328165833284,
 0.7664326268284942,0.16339044850096607,0.47157425846591294,
 0.6390013705799712,0.1545070050158085,0.0685373842067244,0.2017533145828183,
 0.8613074233885522,0.37752083224639865,0.7015636908158626,0.5830070834249945,
 0.6769413171698877,0.6814127104598183,0.29601812647453674,0.9896645702961339,
 0.5479696678024748,0.3447193787875499,0.760193827724531,0.4787625641752902,
 0.44461093005326113,0.013158353004807055,0.5151924876281014,
 0.4785307980370366,0.8439190511101564,0.708734156596562,0.5728594869252323,
 0.5249482338286521,0.22120700944677063,0.704899062647427,0.3898585988156986,
 0.19442882856280064,0.8725651873858399,0.6852707395819845,
 0.33233486101787557,0.49824060129030345,0.28271108397138156,
 0.4347613761486544,0.3449946552327281,0.5578942165717662,0.811642618437661,
 0.17809439480781386,0.24398068976187315,0.4047711114403655,0.45440629837735,
 0.413871659731033,0.5185354319802603,0.8931239356345498,
 0.0070237367082707225,0.23897233131359807,0.12336251153307431,
 0.6986380848723929,0.12822840991622064,0.17583136986511683,
 0.42875964150527235,0.36652681299481127,0.3491970620273682,
 0.3799877944148051,0.9530017642161854,0.49448839936450384,
 0.14819063186053305,0.759013341545939,0.4455936983842723,0.6738029232170442,
 0.8949506615205044,0.9929731240839403,0.6089286380625584,0.07844335996234197,
 0.29808302701645795,0.8479125089084172,0.3652409365317718,0.3681068619943467,
 0.888095148449316,0.2665599027079749,0.19242364265484957,0.13308487688496806,
 0.2989289180979966,0.05341835642724302,0.6020537188752645,0.9075634658172897,
 0.6824347035227206,0.7176936218516612,0.30271991541372856,
 0.21525926775163173,0.521664868321154,0.33559663883778057,0.3616449013851716,
 0.4097451882011186,0.4627514647176293,0.38958996715042216,
 0.32242985270797253,0.48025428619486576,0.30865344211407886,
 0.24967771714453169,0.14964785620896404,0.11653154398763788,
 0.7553015573961511,0.01703001410750904,0.3301244138079965,0.5225115465709943,
 0.031051158646729693,0.7014997343025744,0.4474181396190936,
 0.4304560195473067,0.5684040880311249,0.32702164195261485,0.8477287929516067,
 0.11007109507828117,0.35408855985815046,0.002967041086394895,
 0.5897950645623421,0.9946682642995807,0.31779635743030654,0.9088559199867686,
 0.30718328186400945,0.3420778642822643,0.8787312626591193,0.6618826484516187,
 0.4624950405403504,0.7897405935494927,0.5131917032469465,0.8487774156887143,
 0.9377051738551085,0.9257445811037543,0.5869870994331274,
 0.055496471407155856,0.16127857831132109,0.4667576321106497,
 0.04220291130860376,0.8301558421019049,0.4947017289580412,
 0.04645579858271898,0.5821366694188655,0.3603549786349133,0.3695812464864454,
 0.7910331332393882,0.519156257322356,0.5660156004483781,0.1506600983910057,
 0.3239236025470775,0.5158764939242064,0.2458961186624825,0.4547489857944593,
 0.844302873914221,0.11363320545385802,0.40234139771296284,0.4519517009359677,
 0.21846036868045726,0.10315857582030008,0.18070150165564092,
 0.6136568903434904,0.29934169363725016,0.22474208655200945,
 0.42918848435477697,0.7685218928899591,0.1416959381348083,0.7056708442448586,
 0.15552103101965042,0.35191970004967765,0.3527767865196223,
 0.5394413414368737,0.5467083465428765,0.6751497666538632,0.4718347268979247,
 0.8045592660316623,0.944761652021415,0.9306239586715106,0.4893719120550638,
 0.3641309370779142,0.21738923743000127,0.5795811311108138,
 0.29369961291828206,0.39889663763298444,0.47963043757228807,
 0.2523436453309593,0.8047971346117055,0.8079266613055924,0.2098927740413563,
 0.5550947981237058,0.16233112727795596,0.13176835399573295,
 0.8975935588764157,0.3775213153720519,0.7121373586891157,0.9595533978021807,
 0.7892058539015383,0.6780711954296049,0.5127776489483354,0.07821844052151439,
 0.008737390361805208,0.26157717420283044,0.7163419981396268,
 0.8627338280399022,0.9069852093645266,0.27016740240273873,0.9116807195602163,
 0.21026690380988855,0.6852142060519845,0.5444990866827689,0.6986627440261421,
 0.45887088789882546,0.563923588737129,0.260587152886971,0.7551435373475072,
 0.1672655755339485,0.7716893045978448,0.3981967812904015,0.6683915420413667,
 0.2779091303335015,0.20660969767414816,0.272217780593322,0.23043051762911781,
 0.572841799484345,0.4153856564441145,0.18942239957628948,0.6282952798765483,
 0.3757419124697565,0.5100147402761447,0.5531706652227975,0.2329918272304713,
 0.2893910122096556,0.7209883848961111,0.1083288846158349,0.44290734897219786,
 0.859881333244728,0.005181932624818697,0.9562352888186263,0.5590910064996044,
 0.09126403687565698,0.16681929303199627,0.49332114492297685,
 0.48567516648189435,0.09547207546179315,0.6540410461915872,
 0.005488587822051527,0.9416839816520481,0.7655652450966942,
 0.38531286863641756,0.7884970564642195,0.7247962752053431,0.8678849796331295,
 0.009814753780021834,0.27781174385400165,0.5576572364751015,
 0.14241892227186015,0.4916302041141074,0.5071672987734008,0.918771963086388,
 0.35272071397404026,0.8139967572809996,0.6148658849507305,0.7231138022514554,
 0.9477887770901283,0.3849282869884242,0.06488421646786291,0.2701486483065324,
 0.6066101067271144,0.6304701586216823,0.13510277343492028,
 0.05587544811913936,0.8117493717107577,0.6778778467041461,0.964284859583204,
 0.963753940339342,0.47342445337913963,0.5629179128080093,0.5005490356926776,
 0.8855276693399506,0.362074575042051,0.21100319421891534,0.29965727702100586,
 0.6397539535694076,0.753474431873252,0.5662123905906206,0.2734419037654222,
 0.8040729482515704,0.4063715466118254,0.350141460326837,0.3714066598200856,
 0.5824474274669657,0.2802538260607875,0.7551199736165006,0.22884171688930233,
 0.9289810366703971,0.8451120588122935,0.5534119333302358,0.16713144469267238,
 0.6286672966865339,0.1506488260052753,0.584386912136368,0.30455158500798407,
 0.2980665644490905,0.7868476126475992,0.5577434202377717,0.5075556212701673,
 0.7211818605925228,0.963829249274835,0.6317360786820134,0.47744646654170375,
 0.916481987305236,0.7655521118415554,0.42562512922488693,0.07124152730070044,
 0.366587404877723,0.5170803145730403,0.5367684531519623,0.10094749890995469,
 0.1618825348779681,0.6054385864526977,0.6157137903531149,0.5170745182898753,
 0.46177816765389246,0.9203231442569163,0.5390214503196402,0.531540584779911,
 0.5833313982030216,0.8123622712011206,0.16660323980678615,
 0.01689056216401985,0.6332014804798245,0.7671657396979263,0.7844567039405814,
 0.5239968220282067,0.840433226774707,0.9380751473402286,0.6885942743131919,
 0.7534249487783878,0.9956761480810421,0.16599023208882957,
 0.09245853303029161,0.37566770139102856,0.3164911462450273,
 0.6317874002567646,0.13341659260888894,0.9383827007516411,0.8713440846830899,
 0.7082251601650584,0.08830436243164497,0.6540869896957857,0.9803543752927051,
 0.8242764750123687,0.8119391548481105,0.2944972787754059,0.38050653288244884,
 0.9037353963543899,0.020599982813516826,0.559312841225924,0.5872934899566598,
 0.05920099623274133,0.5398931114659055,0.48135371580401143,0.676551734713672,
 0.7990116681896527,0.028770070409202875,0.6648945918893268,
 0.37426075860292196,0.1852209108772177,0.2814591182487357,0.8384577296093421,
 0.028340498995463093,0.8394929552900354,0.7323909346382204,
 0.043769902108006575,0.25528087930843724,0.5985578405103673,
 0.7061942010845909,0.57911861548704,0.07976835635706048,0.5550498907476396,
 0.3108638938560815,0.794394281009204,0.6762232039903202,0.28712446881483744,
 0.5687721770293332,0.5661897242488297,0.6628133168417967,0.2894638847233447,
 0.6897613710437503,0.30432994769942134,0.004861504777886383,
 0.02654639220754451,0.3887310350463792,0.5768431925945154,0.6455125535176226,
 0.2237257113700959,0.7656238912436604,0.3950611878004263,0.9050351645210958,
 0.7126714344198708,0.16913747724155215,0.6342763803295474,0.9881262358537295,
 0.17304302588423803,0.2987788201402247,0.8663004253265808,0.9164187300360368,
 0.8640312337085483,0.5232194785355088,0.9544891955033219,0.7012915455162168,
 0.3548796613791467,0.2571381673156864,0.5006369480753824,0.193786290548712,
 0.10345098943898046,0.8396139106453101,0.5799938855619516,
 0.27935607023119247,0.6090999267417124,0.6938249012449624,
 0.27209685288082397,0.7460988930402424,0.08460495392315348,
 0.5305675784674412,0.20898330471999838,0.27656570550073933,
 0.6814965486456356,0.13266189262629213,0.905874392377474,0.8695609033662705,
 0.0034158496829421736,0.4263760433876769,0.29594642320307707,
 0.919125473870027,0.10151017867576961,0.573107068042359,0.216276465558207,
 0.8969489345469343,0.20396851693843132,0.21697519426169032,
 0.19597442526917042,0.3980214894352483,0.6170157413205591,0.9201662710448401,
 0.38478397324243996,0.15428677528221502,0.37182921508978417,
 0.4774962103380308,0.8529778762895761,0.853882965730077,0.562834576172657,
 0.37801993388226274,0.6407677425932672,0.972403477919636,0.1840394301649605,
 0.04890542335654657,0.6939422284139791,0.5712324612196304,0.8030586455140205,
 0.007953715233579683,0.8756526711531925,0.2708872636696823,
 0.2916041796050445,0.5507161423765643,0.7903958956285331,0.4869802060715104,
 0.36555809773644565,0.4805319812292278,0.19046745248478714,
 0.08614003591808927,0.686301896847983,0.4264486969072997,0.8647708552800754,
 0.031447986725230526,0.504336026371163,0.01782169764268604,
 0.14597595536983266,0.6243955160327256,0.17392414035980208,
 0.6891802465420568,0.4409241396478787,0.340613314521514,0.25854415862136015,
 0.05222718763500789,0.2778070916498141,0.4433982553205501,0.4523875582905238,
 0.17575198867590647,0.3937949426572267,0.814722002480253,0.7334190131381864,
 0.6821152606652526,0.14734924988454345,0.37687418671473627,
 0.9012180180979525,0.6275080111226364,0.22024365020864178,0.9475830642494776,
 0.41023889714123285,0.34890980772369384,0.19194632442382698,
 0.16888567968407575,0.25466190032013336,0.46932869888116535,
 0.14892854233639385,0.17702923837926643,0.2126680060816586,
 0.4990568678057341,0.021742718493464985,0.9389107303645682,
 0.2743871901420136,0.8963147282143747,0.98437391203839,0.8604277517643529,
 0.31422639552929976,0.8251512632223148,0.9135952183631785,
 0.12601350315930804,0.2353212994883761,0.009874312863792811,
 0.5646391218330479,0.25148434230945227,0.3884345114303367,0.884370468609189,
 0.7475649137622198,0.7446390481771628,0.385000361356565,0.6525063599955353,
 0.33570834723227405,0.1293250727665438,0.3515608439073389,0.5287953503907131,
 0.40698759860863487,0.6343748415887768,0.1906273898455264,0.5887556861898636,
 0.7840956149025121,0.24592521916783538,0.428174490525041,0.8148339403391603,
 0.9298423612957942,0.671942643743931,0.09236783535312942,0.6531207349745927,
 0.43723314915432543,0.7680168959310454,0.6692803727846859,
 0.49115983326343315,0.05554616698969528,0.5832283419092128,
 0.6480046401121294,0.7162942445765247,0.26911632941267616,
 0.03345940451103946,0.1578718892065598,0.6174343388234418,0.6862942994506087,
 0.9149312868009373,0.27963124905463577,0.8392694581545053,0.867340490348053,
 0.04140614644833984,0.1633343271620521,0.4116400927553374,0.5616946915642013,
 0.6787858244427937,0.05384933348863219,0.14438408197139085,
 0.4564151330923222,0.1279578290780259,0.21355588040671902,0.7408171125400584,
 0.01929214476540586,0.2843218174832203,0.8427199975103326,
 0.28419417576203376,0.00999979360072456,0.18384032165019137,
 0.11202383331183996,0.72635754308027,0.7273714312908636,0.4597236893293277,
 0.4145609827141383,0.4870809913861842,0.0077908729880874494,
 0.5537695004989153,0.3394689030982768,0.16896356145716507,
 0.48229429965582393,0.7537047589302319,0.8970995843757579,
 0.38443981726538057,0.14650240421992533,0.7083064656227906,
 0.5184882859162214,0.7718382930157268,0.6399129374576282,0.4578422136112673,
 0.04809236250885218,0.34721051248870205,0.14642285532280974,
 0.7892039326459062,0.847705231585375,0.8320890556928917,0.33337105534826716,
 0.8258943476105043,0.436256342684331,0.21350751561872516,0.5633336337187435,
 0.3400090613267326,0.14891090662629902,0.5462225886884047,0.5399941212624932,
 0.2757032410894237,0.7146094969669887,0.6916166559049343,0.6424488245288914,
 0.8434548205810604,0.03997223054065446,0.2960245267862942,0.8970733020226269,
 0.9111514062667284,0.7172483460724121,0.22051148681588195,0.0214194315458337,
 0.28298210648316957,0.06288410664063449,0.6768155201789232,0.359685208282681,
 0.25949448205217907,0.18571500371232974,0.09542663049567102,
 0.17011314582493764,0.300594449151077,0.6250637866566578,0.38719441861876347,
 0.8679313823142069,0.07147188425709261,0.2849887655826051,0.5486823354035365,
 0.6113290786039967,0.9463619211727406,0.059355037215361506,
 0.46969656302911644,0.0941610922389502,0.4603413605749874,0.304609506670535,
 0.41314326422455927,0.7986628678086667,0.3354316269003299,0.5221540164928632,
 0.48976425512693345,0.4310382893369944,0.6613055490289721,0.8230139145970303,
 0.8331053641214815,0.9701715834668273,0.378036740717707,0.16643212967169552,
 0.5059457614862115,0.8175202965447279,0.1678853957708074,0.22258242922604055,
 0.43568926241737227,0.4421727363682547,0.44126466280215815,
 0.4105718883224202,0.9786311568902691,0.6788858984177356,0.18468124512809536,
 0.3341107924392889,0.2550266884783706,0.6955976916110257,0.44310459651856104,
 0.15112188831944195,0.949930680667135,0.8816709339304155,0.20552653682926492,
 0.1600095010859891,0.8102550845142291,0.6456162809953706,0.46630290040767497,
 0.7828813623119114,0.43245202734201305,0.8113337851103005,0.5063412140665573,
 0.31550152848944957,0.7552212521423394,0.6707233656252675,0.6382741289784052,
 0.9168482486042043,0.01330725680600986,0.7181301861194336,0.6150645444480738,
 0.07000673190277307,0.2542710370426946,0.8963494722955395,
 0.18524471647989904,0.38264680314408017,0.061051281673488056,
 0.374725118294181,0.9545755614147826,0.7499477073687082,0.8983986090368854,
 0.6064426904024204,0.4354099546890532,0.6892490826426964,0.8086053235084891,
 0.98641269596318,0.7983412355535482,0.6097618788616213,0.88419370410735,
 0.7791576349073813,0.2234081094840923,0.22695588941523015,
 0.37695964862017184,0.2891461401363795,0.9553057344657607,0.9311062668459232,
 0.14751170104315825,0.2574886678214018,0.05013638408966048,
 0.019682941632613193,0.028327915660584235,0.05930417880895744,
 0.4398162577500212,0.956032651193167,0.30440709743267635,0.41992981108621663,
 0.8256568368822736,0.04693641935619175,0.6627678778570892,0.7484520388717428,
 0.416791502516773,0.6983684754719691,0.404888885501982,0.6897312254043262,
 0.5579162161142223,0.7380714528569069,0.09591075728414,0.3409038638437768,
 0.8500649428593026,0.28468503032205283,0.09601115834663487,
 0.2927270407947862,0.22986194198416476,0.31948089418817094,
 0.8895912613615822,0.05286170407563906,0.6132460545436533,0.652548955555382,
 0.9053810577489602,0.21916268937097905,0.692088503555993,0.38952641656252895,
 0.4167180136565617,0.31775467126376955,0.7382741919139804,0.9180812237740239,
 0.35994451135063354,0.5151169602630513,0.09127825961800484,
 0.13859434792310588,0.9649719717844123,0.9083343577938278,0.8968287878077366,
 0.630370647885795,0.5959238723066755,0.6093422463292314,0.8286555074509758,
 0.45414279349572106,0.7140804149994142,0.31418219703540695,0.787391307989459,
 0.10349841825239159,0.7899354684354224,0.39336022950507155,
 0.09200589504478085,0.1484425441847257,0.08116909204889511,
 0.5739265020840792,0.8889410713724046,0.025144246477193044,
 0.6304996132535244,0.3113233221281031,0.1772307707314088,0.2235875595349055,
 0.7149025353910224,0.440440537872554,0.21941553023997873,0.8090814253046383,
 0.2299203998462629,0.3756765322516601,0.2674934206358913,0.2222307219756403,
 0.3475510181255582,0.9987840954987236,0.9470210484319156,0.39287258088032107,
 0.6529111844053297,0.8273929952935749,0.5077660991539527,0.7709504051230126,
 0.7602176615610587,0.7797875905476209,0.7513061031881594,0.9787529775234389,
 0.45588914770262867,0.6344518175930295,0.5830626380133972,
 0.43205707950544636,0.08984283107550772,0.19017167950713598,
 0.928404995066883,0.07303966400436968,0.5132109945256851,0.6091511738357559,
 0.4593199054614564,0.5073843035849301,0.9034322881181821,0.757841375235797,
 0.9012251505522704,0.8546105172229038,0.5632054051576132,0.4638083888629404,
 0.287982794890586,0.8789210051209069,0.41261530712846306,
 0.061215626704073545,0.15010361729842447,0.9074393514282741,
 0.12474679156132396,0.2535875860694755,0.963674750065574,0.6332799414068154,
 0.8272298572293781,0.04450057359639781,0.6972735651654215,0.2627202735144224,
 0.8916398525717794,0.27742963871915216,0.20859216703166006,
 0.9393036685700255,0.584776554414882,0.892069910534146,0.045290209843071416,
 0.5730393300422709,0.9278525992012087,0.3448865671812532,0.9273335619900952,
 0.9097087689744473,0.40838500459016913,0.1533760602637152,0.8027543714216023,
 0.8309834182386345,0.6529756315618739,0.15160706127063417,
 0.03463845761801598,0.20772396059645093,0.2168697332338987,
 0.9661648265676056,0.4345135467123791,0.04324202085639717,0.7847185922792974,
 0.5353741582608336,0.3734653723965664,0.507318090829395,0.0428359776318763,
 0.4602528501896975,0.8932920151257442,0.3641047090823307,0.7017764242006086,
 0.4463240486915605,0.1925474453352819,0.3176547074689996,0.13578389520902845,
 0.24914996390816502,0.4205558893179985,0.4504466855296825,0.785835332313078,
 0.336636861975839,0.5097554497651597,0.1857989079778376,0.6928979124832184,
 0.7643739249574399,0.5081109234042074,0.6484847425105055,0.07122515451871247,
 0.9525129132161247,0.6962622324353461,0.9570303062565625,0.6920986990726877,
 0.10120510126561366,0.07855657053063536,0.27044076990951216,
 0.755204194596723,0.8482138909125135,0.8625267236807895,0.2742638714067944,
 0.13802332306972886,0.44489173101394797,0.7474255868760623,
 0.12018521147325079,0.44511467056191567,0.21254836911331076,
 0.7298149573098671,0.3149622051488775,0.9356364647371846,0.392979145487795,
 0.8310914159571655,0.5383002031816373,0.8776705378986623,0.2522072090135371,
 0.013073050830842314,0.29070725727392555,0.27913503126005657,
 0.8168972322503874,0.7438826317746803,0.6253461992920961,0.713999297776448,
 0.8517942003870628,0.2878845362813771,0.24241253501018956,0.9818055345047316,
 0.0423091885578043,0.8740334770190418,0.9083724393753387,0.27907823851992664,
 0.5846800860546557,0.692507453570426,0.29425768602319413,0.571682440778653,
 0.7613476896892202,0.3568563177925622,0.9807513653342814,0.8812388241358033,
 0.5474269689831467,0.611703839643457,0.06274000849718409,0.6194797330073314,
 0.6724553736568353,0.37651222893149794,0.6180736360735879,0.9624740416171464,
 0.7467301317376375,0.1048333252479754,0.9348287405104796,0.9355875516334424,
 0.03299078359828411,0.34754503407009807,0.09271171819046453,
 0.45789626884211376,0.327002622858533,0.17296773770586138,0.5252198856308827,
 0.8833476288550741,0.5933953629892121,0.5735552373355849,0.3595094727834688,
 0.8112567629060137,0.14766762268616518,0.9318353064138452,0.6251125288620757,
 0.8114446069865433,0.428152515303329,0.4902598136399533,0.3997463255425341,
 0.40035623842618984,0.2312144038567543,0.4281178766110665,
 0.41038481069614896,0.936114600725888,0.901019920581663,0.5143095941331037,
 0.80945122440657,0.62847298851257,0.9869601566084774,0.6172268698799668,
 0.6692763042436903,0.30092445857057193,0.042462558945023265,
 0.3531450509314993,0.19617859954123884,0.2459321035574047,
 0.36494960802247045,0.7838905036904011,0.5840651239346107,0.9459562829136703,
 0.09242580850337367,0.9721858035630981,0.877903753905604,0.9794501968979685,
 0.5808790922428834,0.20450744804651422,0.06309934634361902,
 0.3202376613744292,0.6154362142362033,0.17273990552957208,
 0.17270781917062106,0.902799758964123,0.2810267514022269,0.7960434513503127,
 0.6741782011704658,0.39649797031504663,0.8483767745663426,0.7432238410070402,
 0.23694251845808068,0.45624932980268473,0.5555721753149373,
 0.05229081029733096,0.5470293206005419,0.8422798987532005,
 0.19640647747500395,0.21504966768613187,0.7552544870590405,
 0.002749352304439956,0.9943841579564677,0.049766021757754575,
 0.17337762598317685,0.8107156899479485,0.6296184615070983,
 0.29275190967178455,0.3758636274164666,0.941446840705087,0.948986413749846,
 0.8460247411833176,0.17893629788850163,0.017439645788968505,
 0.2571303261730633,0.6396909631248985,0.16862067076809528,0.522130238427924,
 0.9012104753210887,0.4766262445838759,0.8873602752205386,0.6947541681323793,
 0.4001674111267519,0.9002986957176433,0.09906894561782753,
 0.19560810938189832,0.09564548406427142,0.8300763677336556,
 0.08951421561130335,0.89893708416069,0.9001979634063357,0.3887423689443261,
 0.5720349332341547,0.7392299435850407,0.6682140982162169,0.429310785866294,
 0.3571765752649011,0.061781478213572716,0.44955040494604726,
 0.10780081678781106,0.6563444100865238,0.9449524269986483,
 0.04315670994094223,0.0839464596409194,0.566189104234963,0.4867029052190115,
 0.3003395862160546,0.377037291510685,0.017497148858162115,0.7843124538754583,
 0.27785962774538453,0.8185379688939731,0.19101104360202759,
 0.38850533297361955,0.9635019458754345,0.15339124460876163,
 0.4720386440315846,0.6523643861158294,0.4161248575915337,0.9621282813368773,
 0.6436380861225065,0.5839520230544619,0.3824476500246966,0.898056284096285,
 0.7447545640529631,0.1286795839530046,0.3537765970354054,0.6215610274754416,
 0.11320674049324053,0.03736196499864641,0.04701144754635489,
 0.5003556414021386,0.39672098811473044,0.58675108259848,0.028922247251994104,
 0.24115978579840391,0.19441270602599403,0.49461143561398757,
 0.2859208168864036,0.9174732630361592,0.9699075787775684,0.8944110267056655,
 0.9058478427929282,0.7730434097802132,0.3439471537578397,0.47362880071293123,
 0.4454802022771075,0.728767162557083,0.845903853194941,0.3150368482927439,
 0.4318415277732024,0.19964155560500685,0.7075101676811493,0.2340359961631937,
 0.41068813942638316,0.7694116082216306,0.9854241290255716,
 0.25734522901925283,0.0460729472083532,0.30131805424107005,
 0.9228235335285567,0.16323220212949507,0.04914885678625336,
 0.8599079963368131,0.8380096678914435,0.0427992724234455,0.4724667725141495,
 0.7291672589493389,0.13461072979853816,0.11278669188592005,
 0.08973276946343967,0.5743639320105559,0.5400260345312791,0.9721211505621437,
 0.9904916245363519,0.03350690175214588,0.9019514361439228,0.300795275725828,
 0.8099208120626731,0.1836022235235606,0.10055639842703923,0.579481498330305,
 0.5944648410614333,0.40951644274497756,0.12026162838650212,
 0.3594400311960172,0.20623411601869424,0.01119252341875221,
 0.7553880528522673,0.3790147395418767,0.8540028943323658,0.02055103002117764,
 0.8004610244254252,0.2774465391168446,0.3978919202251846,0.8035782984376555,
 0.01622320389561005,0.5973019356321607,0.36693808389784044,
 0.09104860607193321,0.8741319116174693,0.4341368906584888,0.3096998727479492,
 0.10438077705238846,0.5019883539126868,0.6096596642198249,
 0.20101919021532677,0.4680502813300512,0.6730025067168269,0.4399205998316773,
 0.807246660836817,0.49861670235171607,0.16272460161762747,
 0.23184655986074432,0.594708060076793,0.022519188911859,0.3703119911255147,
 0.641017138347485,0.4225103461644105,0.4117887377818662,0.4095455273582026,
 0.8046608146657873,0.49316946114415283,0.47044217171921043,
 0.8475888515349894,0.16588632018687843,0.22886412981014337,
 0.6860782897942794,0.36891746434087247,0.6017840873834623,
 0.10533385588450472,0.2826940921807284,0.4793303585801383,0.4164150036253712,
 0.7532982407978716,0.9310709244954228,0.09443702474556903,0.177514275997257,
 0.2429469396203957,0.5587838290267869,0.8526814551023392,0.5587376361844786,
 0.5206218836047672,0.31766378086104774,0.4929950512893605,0.8661342413234534,
 0.7318634511691784,0.9135316762303678,0.5425934797134402,0.8122242224946482,
 0.3393594638410008,0.842432931701569,0.9479763977909045,0.11073079604267744,
 0.5084193762738969,0.7867648404194004,0.9704669435440209,0.8351318806100506,
 0.7550372859900377,0.5484629263839786,0.3209804250562137,0.5830181485501158,
 0.5167309216247112,0.6177080035993584,0.1678639943209248,0.43834880316210867,
 0.9958413181602929,0.2780341367006889,0.5371983966001621,0.7579062548144097,
 0.8381536945579621,0.006818679769934666,0.39326292228313353,
 0.7583236933062381,0.272414116907182,0.9741317727580991,0.41673537899387014,
 0.9649577587226662,0.07689358650809397,0.0027349640410819775,
 0.6751310782736383,0.27426305861844147,0.7771522510876436,
 0.11858199004573822,0.6256468372677599,0.10077828993613036,
 0.7153020360683504,0.07642213627922656,0.5901218080708566,0.4025105644432003,
 0.7281029899210245,0.7121170740633643,0.9135097791794561,0.23513692057656188,
 0.05427865823959532,0.2238963999615391,0.6902738671422121,0.4999833830558591,
 0.1332831001746443,0.5051884339403399,0.17510945652717724,
 0.30929367270916974,0.8438602620662848,0.016996414046296815,
 0.1070196878568388,0.4266206652754214,0.2028205589929971,0.6821864362954546,
 0.6306785118836568,0.4614279642063237,0.003561616435814541,
 0.9426732032644821,0.7384564897020317,0.9501854306878679,0.7639758884919725,
 0.5050803978447331,0.9359710246150749,0.6766662720218589,0.7483929538271012,
 0.3577942669135342,0.9315211557750717,0.96813566342504,0.12746177072228693,
 0.022359108552143514,0.3636737505995208,0.5095752944787949,0.882840094527773,
 0.018439546945900553,0.6324808039813312,0.16974506922984223,
 0.32623564836887775,0.8683028552744096,0.4334476163679589,0.8189321068692481,
 0.007917729147617347,0.6871044523072636,0.5931964966942647,
 0.9891800286877943,0.9843768611003708,0.06153886777612494,
 0.18335508827360192,0.47100254709513834,0.4900250151414748,
 0.5316360096239166,0.22090647234637895,0.45488383728905524,
 0.25920141089998094,0.007252066249995859,0.13208323979174708,
 0.27531636520165526,0.5371433281925849,0.7587856170400161,0.8931846514377048,
 0.6973216559922336,0.48829323985437423,0.41045283238361396,
 0.6734415359827941,0.7481099689749419,0.5866981354717856,0.10465059514841735,
 0.3234168061311171,0.32079672760614475,0.14259513548900482,
 0.17666348185387248,0.2002088917051027,0.19804739952953243,
 0.3507318150808213,0.0628996150225416,0.6249187851294471,0.02925037260603147,
 0.9377692363231243,0.8579705399399506,0.9979399908456639,0.9621574210424999,
 0.22161009926630748,0.509772154012846,0.21186493771387438,0.4791478629924636,
 0.006346850328510456,0.39020398590930616,0.5624265042193781,
 0.10643692850479214,0.13789812443895932,0.45487804489601746,
 0.6243927520658538,0.6782602467011313,0.9066363740719073,0.2725516413061153,
 0.441462771243456,0.38783031288617287,0.8517083337070591,0.5472087028225503,
 0.17857529112315096,0.49183764338343494,0.33619489338874375,
 0.7709287762801824,0.3418324828522826,0.9872053130839956,0.8987183251630219,
 0.21290410488502698,0.6001962226935529,0.47316605417426016,
 0.5867288251325005,0.18546367480546977,0.8565777878812152,
 0.022971782408836128,0.6885612484175085,0.9168855561064624,
 0.7311777441143228,0.6491730098073183,0.020347836950835974,
 0.5623406245269337,0.050071596554450926,0.29432595569041964,
 0.5364155734523627,0.18330387825127126,0.4023677043127102,
 0.22970394326820276,0.2859670827838843,0.24598113448837833,
 0.11708826071955769,0.8965103384764624,0.09650153717643639,
 0.4765842416068621,0.7088465066226709,0.5262055088145458,0.5961469139052854,
 0.4767387618140483,0.5974215910171787,0.05478216707939476,
 0.16605749771145362,0.16421618252819092,0.6786951681633131,
 0.19693331848056816,0.1683095318853316,0.833419718482632,0.9167389802751912,
 0.25175461417120204,0.883439496324374,0.7555834511115227,0.0968210068202695,
 0.9134653620345157,0.5873676159065229,0.11360980232994156,0.3923718479340229,
 0.805733552022658,0.42991422259588197,0.791724915523308,0.26066316149836954,
 0.575045453705381,0.09859703558872157,0.34921739545679353,0.8518385455824626,
 0.8127799272493647,0.3884292344400311,0.3047995690627001,0.1269645362956131,
 0.5152289753358259,0.09718342971817995,0.5068709962205548,0.8917918810673862,
 0.08824236015107478,0.5093514338007111,0.7809242370312588,0.9511023653774234,
 0.4494475995171592,0.23988852506255498,0.6715108270782797,
 0.38577034977514935,0.31809051355133033,0.5787886795649118,
 0.6300950257529777,0.3464944176538435,0.7388681609812184,0.33153915311422355,
 0.44956002359165637,0.5811134467861035,0.6333064368954436,
 0.15261175030115748,0.2536386804334444,0.9442087238559492,0.5951449448284659,
 0.9435821694946502,0.5192613204611926,0.605071405179876,0.8899901539215647,
 0.5481987403898688,0.2244830222098806,0.7907812401996845,0.6713932539080569,
 0.3255949457764029,0.2914416942015292,0.913723204177436,0.9196050161345921,
 0.3035071001899877,0.33842056344737836,0.23890290799103797,
 0.1530454578827397,0.6869127154985041,0.703654460795139,0.533051112217609,
 0.09382657318419863,0.3898412757573566,0.43426688028491,0.1281956433576399,
 0.8416300454579857,0.8369659198431711,0.6615094118549112,0.4049057891923451,
 0.679306341453459,0.8080413737837103,0.8607366874962876,0.12585651652308227,
 0.19548179671736554,0.6243730824477769,0.0014425488723334157,
 0.46829479515468986,0.0607698802959874,0.4720376460336093,0.9895361472364016,
 0.6817427580486358,0.0796541951960088,0.48134414432410044,0.3515164378394743,
 0.5389432152697384,0.24195566449750738,0.6965285743201385,0.3409491896597454,
 0.9832937368548191,0.5970690370667392,0.27284093580617896,0.7339471164065438,
 0.02222148601766849,0.0997918166787174,0.5645921734444967,0.9049558037345465,
 0.3985769279171971,0.43791635861050204,0.35558402020173074,
 0.2578493511767952,0.1751676795476823,0.7452933963806357,
 0.021592281914720735,0.8165032747528139,0.9359176447500173,
 0.9969080399087064,0.7098742091815242,0.9602025607506565,0.5568537416732465,
 0.869838584403306,0.24650975860033875,0.35655780858303,0.6141610360404444,
 0.7525705859564387,0.1495641682593707,0.5012138437594115,0.47549710957593583,
 0.5962155784905274,0.48451115458569416,0.35358095391779765,
 0.7754264896536576,0.9702110925822369,0.3361073336060262,0.3867267407764585,
 0.012283100247104883,0.019240724895195682,0.5904498487625561,
 0.43783446024283146,0.05601781293120944,0.11917626559594374,
 0.8415956102990797,0.024471739107162604,0.18189935284168612,
 0.06119280000345273,0.6639673473646548,0.18317328653448417,
 0.6873991593806312,0.3649490800082351,0.031997418970970304,
 0.0826448534161226,0.5504956164674516,0.9687716165479806,0.20663243468970993,
 0.9716923251384072,0.3196928314061489,0.6916260923556984,0.7903496306319269,
 0.9534020861357925,0.6767846914763066,0.9785293581225953,0.08923951646527506,
 0.806752993134865,0.6295206313975265,0.5936402798730284,0.24972772769868878,
 0.296041241253009,0.6646685494695916,0.9612582019281696,0.11389706037323,
 0.2811481033065576,0.6812671018096593,0.40019676886428335,0.786781899299964,
 0.39815589852244304,0.0896804451762655,0.47396356219128855,
 0.4122316583325446,0.28758666189239845,0.7912791451984795,
 0.023219482446935724,0.41997726440722294,0.5207557548809023,
 0.7212453075349348,0.7762080018655828,0.5953046320459338,
 0.010764794302387859,0.12711744014649085,0.530933903095671,
 0.48887046425455083,0.8722050977893149,0.9421848277640872,0.8390669385969578,
 0.858356031522523,0.8100812586978302,0.7925862143957236,0.5654926000914732,
 0.11606962112271602,0.31844983149527883,0.9222725672761624,
 0.24322731074863901,0.2292775579674703,0.017825722598841298,
 0.30573299261770603,0.10650876560798628,0.06479092971168432,
 0.5454572986456616,0.6611412739262582,0.289021839733996,0.16594781527065328,
 0.10165898583428912,0.9095597734487751,0.3585768484466273,0.5070878602554177,
 0.9896933942332355,0.14081006685113073,0.5900275422354033,0.5472143053531777,
 0.2911425407253587,0.6408405681071808,0.36836799827232936,0.669741469449926,
 0.2488978846691301,0.703191753426871,0.7076733132175609,0.8473104073565546,
 0.41795718057223863,0.21700118213518804,0.2490532690756584,0.523878966517434,
 0.29158043254054467,0.10737307021326503,0.5404274848794458,
 0.8795651039396887,0.47704723464790544,0.9225695605556808,0.1546130807320072,
 0.5934340722048418,0.17269480151499084,0.8118086670407497,
 0.041231206684958055,0.9624169205787027,0.8343119079004645,
 0.02824325130731553,0.7236108594143933,0.1601058468941906,0.880401423857989,
 0.862669606188583,0.0618875476300389,0.6416071458803975,0.06604688740778497,
 0.6947070714180684,0.44230225168896686,0.3639295533586032,0.5776466224563869,
 0.6559421210838793,0.395936424024368,0.17578900846069978,0.8166258045166501,
 0.6288624401903803,0.22205679970099346,0.8489884219248272,0.558540584479016,
 0.7993371148397784,0.505283284430609,0.06687638973379295,0.07709663848738812,
 0.39509404373657653,0.9908547869423135,0.9326620427036412,0.9161948900779762,
 0.7171212803819647,0.27229249335376926,0.5219525810737873,
 0.18251858496154327,0.6645900074076477,0.5060245559466074,0.1862463058147774,
 0.7638276672113871,0.04509334275883414,0.23319348906243487,
 0.8738490399166534,0.7318243453550397,0.7078992887905082,0.1979656446371807,
 0.7050855825623151,0.05698052721945113,0.741517829039025,0.3141729971021193,
 0.4821416599494358,0.3562142951629229,0.29473418507031024,0.8227376531171441,
 0.27516464140881014,0.8830121597511198,0.9001808162479819,
 0.46128450354100914,0.8105065978506709,0.0921616123010911,0.5271332858106684,
 0.09826965395915133,0.902809914298699,0.34860430355772465,
 0.24237619248426145,0.6234116908621632,0.009114229338125113,
 0.7000066130778714,0.29981611027296007,0.6696667271753396,0.7631626212210978,
 0.5128113260383581,0.7660351878325294,0.8640920200921849,0.11449503744459821,
 0.44445895177988204,0.7609302034263804,0.9217704082013386,
 0.41713928807992673,0.998430828621474,0.6820040972691043,0.33260440964981286,
 0.6707441095743694,0.5204947538296478,0.7742029833701372,0.8568936070165544,
 0.2752044398780814,0.9032208174130318,0.11379186931791507,
 0.20902207534742479,0.5809883283320145,0.8858152182781627,0.3556569617351951,
 0.7736437976514862,0.7622445170409257,0.9268241633164743,0.97158063739987,
 0.6011320163262088,0.972489214850173,0.1566280645825917,0.21059824057889598,
 0.6707594346291808,0.8601875284016868,0.07877264929368932,0.546925092786913,
 0.844568462205592,0.27346530073396047,0.925154746540848,0.591395493543212,
 0.27614407247636386,0.723510845862116,0.5141961225057822,0.02218507026905081,
 0.06847350733551172,0.3640711579884435,0.7744922327322946,0.2469719324730858,
 0.9732499114267464,0.24176400043404644,0.6303067196892009,0.7021452030989186,
 0.012922428219197424,0.6328898480900623,0.963771345654403,
 0.24535634382568494,0.964991427298673,0.5984258369430865,0.6165510365165139,
 0.05673465124671506,0.12507299363535185,0.7658018120214523,
 0.49909777458584725,0.011324488792815113,0.5422223249444047,
 0.1558126083956246,0.4336476696578613,0.6763544911526566,0.43626378013052014,
 0.615035929866459,0.9778196649387414,0.3740906791167743,0.33998238991834673,
 0.24401140757483808,0.8945793528785486,0.3230319467324734,
 0.27205477780267184,0.8036114230927608,0.2313591880442134,
 0.14496016155847102,0.7214413084538105,0.0038963659045203602,
 0.6547000167793039,0.4439615162606494,0.24252505191393114,0.5354219933872862,
 0.6741643842561482,0.15952950525651222,0.7457209915783455,0.3872091965408413,
 0.4049207002486275,0.5067555798087604,0.7154381045859309,0.7533943866763535,
 0.6451568343502166,0.04925997231928392,0.6523698332530716,0.7009082088044676,
 0.043287194700344145,0.7720166143041266,0.08204681511948697,
 0.8767598772968865,0.8262398858464635,0.4112299888376041,0.9201870101104708,
 0.5473037113950278,0.3594617190642424,0.8221997870635854,0.9679071402828613,
 0.5597268691854229,0.3725896706420335,0.8187985417426343,0.09152241086344437,
 0.48192913580672414,0.9251212775199417,0.8504890990203099,0.4308175996836743,
 0.27469737637201097,0.06617691917830082,0.3229483488218089,
 0.5946837950144539,0.9480381963208209,0.6967373978837446,0.8560955010386391,
 0.39048167019108504,0.8307371168958807,0.39839200427582133,
 0.7277888393324686,0.4631440798840192,0.5868339960443776,0.959168055623003,
 0.3806793787099215,0.19110337007235934,0.9228498966622838,0.55492397715449,
 0.8767873791837061,0.0087547335276984,0.45469504066744715,0.762564808470799,
 0.5698428227764296,0.9957077073194229,0.5065339101442468,0.5821832742275251,
 0.6959189815363324,0.46753734886180864,0.24153420462351272,
 0.5547400498417129,0.19424920235031395,0.3559124431857301,
 0.01622825618505508,0.8296434778764544,0.2320337411200264,
 0.12473393940644162,0.2700045769783137,0.2769285046776043,0.0962565307666653,
 0.7992724308153842,0.34379549764393313,0.7484976038761908,0.1587657159418756,
 0.458345154839808,0.7756994888522727,0.6350538041922232,0.3960375480762375,
 0.040240914646734294,0.08579104513758573,0.2165342647638675,0.48420592122481,
 0.5268491459802042,0.7574399507190092,0.5224276338416483,0.5896056566368724,
 0.05840504366836419,0.35060033130902946,0.7160648365925497,
 0.23556271889211355,0.6875713159937681,0.028519881055979135,
 0.026846603939939095,0.26413223663708696,0.23136914006116216,
 0.2534283815554266,0.37163167026154764,0.3042395062683766,0.9493674784611755,
 0.12499081997897665,0.8171648606852062,0.75658840121085,0.3657891250491926,
 0.5155816450923677,0.2328264226294967,0.4315519122227618,0.9031373997457713,
 0.5124104221055691,0.47551166271451206,0.08796052264099896,
 0.7610275603087833,0.40165680368793777,0.8935148318365758,0.9829479473451724,
 0.8530397307005501,0.8810632039169644,0.7872581886241599,0.7350023009675541,
 0.5796692657063367,0.2902794000094848,0.1411698832919459,0.6182347350887922,
 0.6617254086326785,0.7809575760176939,0.7950722968587142,0.7597036932829625,
 0.45706613207929114,0.20255925599636226,0.7823433680897436,
 0.9215276804873194,0.9238051259899985,0.5133356596905271,0.9282701840120351,
 0.23417027979980798,0.014837062730536865,0.8225031742313229,
 0.3337417786986908,0.17266736268778804,0.17608352496903001,
 0.13092540990484813,0.5809706969436282,0.2020861742469381,
 0.01402434454926893,0.32552517844072726,0.9025475817402392,
 0.2811301596546999,0.04210007613256983,0.022415506918903816,
 0.7201734331235438,0.5332879971522604,0.7925179559181253,0.7293467801796717,
 0.3451297687902726,0.930278427503656,0.2001777456749928,0.04754712631375846,
 0.504263914472095,0.2616127377543208,0.10838093724045506,0.12778158786634186,
 0.36590962701365304,0.8148188362137974,0.035120033103125836,
 0.28817603159546246,0.33023967200999926,0.30618403883093026,
 0.1533463036522884,0.07202317644320033,0.6896719812431248,0.2005724614426535,
 0.46132616671685445,0.5658992489115373,0.6864736889120225,
 0.07717990734392488,0.4780013290970416,0.9539449003589036,
 0.026245175135432897,0.9694864603907896,0.5138198898432143,
 0.5698803222544522,0.8483819198035152,0.5180870096214893,0.5521952137814821,
 0.8980188163610381,0.960069639131843,0.4792879434523092,0.9376330474115698,
 0.36048354566498564,0.4219349827183818,0.606744753926206,0.8187881803679589,
 0.5249877800129049,0.4933062511444175,0.6878797737687633,0.09438804658055644,
 0.67132683795931,0.6096949857977186,0.44388884733213074,0.9215702014136081,
 0.8674701683527666,0.43141340146680407,0.4908715376840812,
 0.04855546416234524,0.9100367860179233,0.6545419353570754,
 0.46610939870980905,0.8723809358533843,0.23330287351542123,
 0.16150232346162874,0.2797095588884403,0.5687055294247186,
 0.11770883700612411,0.7235280941506056,0.8155492606119747,
 0.22974223455287557,0.4383392693334497,0.7600224197403416,0.6836435998403246,
 0.2623500842451735,0.2971753905496962,0.22799621949569915,
 0.18181970356340005,9.809299983706277e-4,0.24888595484667042,
 0.7675944334986027,0.3788452709841569,0.03775990325670253,0.4248010206901426,
 0.3512962123373664,0.5443950984141642,0.9048116730251177,0.06497701337391504,
 0.15585815066148634,0.751787091612624,0.16410904101842538,
 0.23046770289529273,0.8198033086016798,0.22615724738057308,
 0.39490577656476056,0.038688402682180545,0.6232155025085793,
 0.09203376325187729,0.48635462022816356,0.19145316498393394,
 0.8721457315042337,0.05305356288330265,0.7412365654424651,0.8451486281679115,
 0.36595717602055067,0.9611779186421475,0.36135554569574113,
 0.006350651438861266,0.19789519620439955,0.056018376849999996,
 0.45680783039364337,0.6424542099174381,0.5694404439167216,0.6651218043260565,
 0.04634819857126815,0.41843314278602906,0.09798742076165634,
 0.8851951844807522,0.5254528474985319,0.01829852246039332,0.5429911042830087,
 0.9038386969627198,0.6109758994504149,0.2721036829095833,0.9510913077522662,
 0.05013444666052458,0.6602623215125014,0.2838519577239347,
 0.39277937589449463,0.7840767179505798,0.7668442791798792,0.5232540884613872,
 0.21665690673384042,0.4325796895877553,0.16404591707284355,
 0.09588453926761809,0.32429335551033245,0.42227017504539244,
 0.5492212607035897,0.06067438674890402,0.9246432000154141,
 0.29735987255035745,0.9250684523475846,0.574507692555148,0.7474746523569487,
 0.6372192614566033,0.15606609190907073,0.5565593088624816,0.3419842301096806,
 0.09018243208152665,0.4145332448160136,0.14664008777363902,
 0.14119960327862513,0.35004941535208756,0.4575270051092597,
 0.025848794994670854,0.36963663155671267,0.3087383077384954,
 0.6035056789634781,0.7001555431278559,0.20001580817148845,0.3409239783508524,
 0.49931786492211283,0.6964207926500268,0.35963496248515503,
 0.7333267281307012,0.6985000279877899,0.33729853900328954,
 0.06386493464984488,0.59098463605787,0.4195209610082974,0.6444980529389581,
 0.12477341226513528,0.5717549869431688,0.12518983200281641,
 0.5176285602951534,0.08806647354247144,0.8332169114584198,
 0.20660764698455136,0.652122107810748,0.04734651244669519,0.7232632676645182,
 0.89775607821076,0.37560237751627723,0.8573708665106655,0.7909501847381899,
 0.11570965246470422,0.15399556954775595,0.5631548741024048,
 0.8886395978428493,0.03674270114298406,0.31457594021806634,0.863768583008685,
 0.2934847264474535,0.5029019345917577,0.7065863376018732,0.6369942526628851,
 0.27276405955856153,0.565256180358369,0.15481190541882184,
 0.015638744461971776,0.568472667511329,0.7193107749959816,
 0.12750611389870803,0.7997268804558911,0.2738295758337306,0.8946587599132102,
 0.25586428540655726,0.9193647538432377,0.06090670605538473,
 0.8986217251738415,0.1966181624595097,0.35403958991255025,
 0.28624325241536575,0.7573486470271792,0.3497354823594059,
 0.44697830167183406,0.9311533280407707,0.9391005569169528,
 0.05895254255925686,0.8423554201885171,0.9554947877335368,0.9384052389114838,
 0.4505825010406542,0.6694033034077557,0.04182480665972177,0.6273464249650057,
 0.8477688423319938,0.13209716341721078,0.9169814772534507,0.9822017154291687,
 0.829771619730657,0.8850387765704989,0.047232094617921994,
 0.08136342554506948,0.8787945754315025,0.24329948535178614,
 0.0331642514058027,0.5355693260109811,0.7705069220564278,0.06881447749405101,
 0.6129869849068907,0.2958290100073423,0.35757739449285464,0.6548126558941215,
 0.7711885290437326,0.9993283343239597,0.7240060217444767,0.6019383670865968,
 0.7245205126585893,0.6949149384676788,0.7692853775127287,0.41152345420513026,
 0.37976401225134493,0.104135411187804,0.16655528093340277,0.7725552724873639,
 0.12486864887731564,0.19541841517306402,0.8543424539834235,
 0.3545028229501933,0.048086797510204504,0.4116636394512675,
 0.4464487591022367,0.9436696220969408,0.830276319325147,0.27938005847486425,
 0.09166987820125228,0.8151516561570388,0.030531796356151997,
 0.17752845840050224,0.14115461277175345,0.007751641567094625,
 0.41420183178630243,0.8252902059367454,0.0382348315553096,0.1326584642595402,
 0.8581238212631671,0.17703284643629935,0.46753898819880624,0.728008098732298,
 0.8898114940255549,0.6497606267142737,0.33797929330076826,
 0.06914625647901818,0.07825134778153442,0.13702138738967573,
 0.42823264238404735,0.7029060299395189,0.6029747930590293,
 0.43104822602174375,0.752796456195675,0.8832081553768294,0.8795152418853464,
 0.04728271658479544,0.71775805100362,0.15828810026412932,0.6375190500331378,
 0.5840305611155429,0.6911116820057269,0.9291299407830075,0.7215768756983127,
 0.9300063974548718,0.3085553182169891,0.3432864270708421,
 0.048064907223566045,0.17008388121720397,0.8229761725886181,
 0.2230947841077775,0.4618298232607069,0.22550775645240584,0.8688305709622222,
 0.8382318504798665,0.7458224602203708,0.6643989316169928,0.3699053663938022,
 0.5203776298048026,0.39761607892587514,0.18211121623380555,
 0.5302629636235343,0.19132382112843305,0.6031240511244526,0.3813992910310524,
 0.03976444534560042,0.06932777844133287,0.45782146584812844,
 0.03353621985497646,0.9809352193681684,0.9276664684894964,
 0.03614837801208737,0.46063150338122694,0.6997213917365112,
 0.3204790530396966,0.8623433508654711,0.4415198602774335,0.6884149743077079,
 0.8267126572665634,0.9195438048362944,0.03484487282480331,0.9816181798056595,
 0.610550366193117,0.13030438710009673,0.7869776560295444,0.8241343296145186,
 0.77068241967026,8.174219298003882e-4,0.21294925902032835,
 0.28524595217277426,0.18267050343874114,0.5382367121342083,
 0.8731239225460181,0.9096431170121959,0.44728561112378873,
 0.49044424750989424,0.28629667241261314,0.3425265221121686,
 0.2740894573534991,0.8468512660730458,0.5350678907414763,0.8370317077193453,
 0.20043679122429503,0.045243462320185945,0.0286217796125261,
 0.9987328509564903,0.8459559501534977,0.001746356674020233,
 0.18037583801625867,0.9091205369459114,0.025021108983134255,
 0.25505747721889815,0.9530401485973575,0.06024008859463981,
 0.4863611741411549,0.9214782385879053,0.683746144762617,0.5582106077427905,
 0.7026950729791599,0.2930463887555338,0.32979326848688395,0.9194614859971159,
 0.9782600593698001,0.6297212111437694,0.11505274521931308,
 0.27290583315791006,0.48978822551557644,0.49264268296487046,
 0.560848166052656,0.9516069272903183,0.6122586902298206,0.1039280298310804,
 0.8023262710782858,0.7976376287792831,0.49179891799805897,0.6277231803323539,
 0.3063153980502653,0.06043785028052906,0.7926101381051416,0.8584333476101509,
 0.5742977967635335,0.8521998968680117,0.5363749513322938,0.08248695530383576,
 0.9559005413853188,0.5098109408642958,0.4191346059897836,0.3398727440327105,
 0.4154791617164145,0.844136321198297,0.46198100343729,0.41321556447469054,
 0.07682179436525682,0.7062456421527136,0.6540137602127423,
 0.35339460839901216,0.9080530803844375,0.7921544897525438,0.7276103627700241,
 0.9621952165569857,0.7643478635492821,0.37746836630414704,0.3209542957911551,
 0.9452220449984786,0.3417052187890258,0.31102653120047974,0.8787259070302725,
 0.7472578973526817,0.47783348132893533,0.7087685609452117,0.9151854119342011,
 0.4741244241445397,0.07351747544070719,0.6660557131899418,0.7712853592853538,
 0.04540522537875091,0.661778451109754,0.07574759769494077,
 0.04488412686694998,0.8587972502946711,0.66342077450161,0.613823455313676,
 0.5484000167007139,0.09187496238519433,0.7547128068752755,
 0.47069689623162503,0.10423603146976168,0.44125990687645245,
 0.5506857685948512,0.6950329351674329,0.47039741599992724,
 0.03226919400232853,0.6607792837283458,0.8006205098215365,0.8175936020772757,
 0.5388873323626785,0.1522926473414522,0.9139939301654829,0.3647737553491599,
 0.8193377465863502,0.6375124216105004,0.6477576372021744,0.4320916880306519,
 0.05057852368653293,0.01470413824348027,0.013916403522458931,
 0.41055593839267435,0.8045413922558234,0.13523314462312708,
 0.7895567552978815,0.6956977845820518,0.3280610095946007,0.5665948622187614,
 0.08149685554307717,0.5435955699116738,0.7015333259973108,0.9910250902412507,
 0.2052838815195922,0.7879649462831212,0.9216901109804808,0.8752634206304168,
 0.02868364650834565,0.4728434246695622,0.665732342877238,0.5523217255413037,
 0.5641823790299602,0.08764468616191201,0.29804996441239595,
 0.02425685746397166,0.9301731657494307,0.799991327318233,0.41943293856805086,
 0.03340698084606741,0.9986302203004896,0.925920418609591,0.22508138779101428,
 0.9970151568831687,0.3599660387732917,0.580297412473624,0.22399142341362555,
 0.8431741978349402,0.8359738796374674,0.6411848344035375,0.3306400494123036,
 0.009068724640217796,0.20263702334619382,0.46813840387998473,
 0.8287923999498717,0.47291788038315374,0.2365555492445487,0.5135383501553298,
 0.6637226100374581,0.6463744795536773,0.7916221672235143,0.9448770821025373,
 0.09532774906523989,0.1151027199230158,0.7126155025670763,
 0.13865421431723457,0.7514753128648837,0.42304577008070143,
 0.3699126986771353,0.9690478964712103,0.6102294267947739,0.5857401911783103,
 0.6171134655484549,0.5288501033492927,0.44359442322971765,0.6633630082219792,
 0.06952356692170591,0.3794809184288055,0.22815903819348138,
 0.24210022721409996,0.0415107023552872,0.9594646339309618,
 0.20937868415784633,0.9746158703068627,0.7884550307016639,
 0.25712143184991887,0.1657534625065027,0.15991750588948395,
 0.8160690128818994,0.6209935608664572,0.614959748414947,0.28176519789934074,
 0.0033852187897409625,0.07173112001892235,0.06353308551937698,
 0.958073738603889,0.7080042333695188,0.0397481062506756,0.16974815427467782,
 0.22819879461521309,0.5278384439400859,0.23942450206841748,
 0.19587590145093348,0.22737974174817888,0.3518022709266194,
 0.3651473675581486,0.7678249465845122,0.908294512845713,0.20886828777260857,
 0.4264101899037346,0.20199198724464895,0.23189146020938312,
 0.7814954478285197,0.9033095714867381,0.5943951954902118,0.16029601575236385,
 0.34892318955497803,0.4013359198101053,0.01819357009522138,
 0.045739867498346576,0.39114429331958844,0.2167431919499596,
 0.2888111805911532,0.36334445855196706,0.19563312010729728,
 0.14405147795118678,0.8927807756070648,0.625131554295852,0.8064909418227698,
 0.8437826227374626,0.15563675622044826,0.8222411450360876,
 0.15401456699554872,0.43647462208495946,0.713195102764818,0.7211227881411324,
 0.5296184657126743,0.06454754211908531,0.7176947024469369,0.8812567960393809,
 0.9419159949009911,0.6804042444011137,0.15315477286281576,0.8151552006344015,
 0.741163986884616,0.9630341747778002,0.8856771101102575,0.900567863030614,
 0.7895241880231246,0.653791520412814,0.0025902368148558175,
 0.45875667833681344,0.06393867976332657,0.2487118194368345,
 0.0880340640541627,0.811835866170896,0.9896262171262833,0.7680793333300073,
 0.6611288815567821,0.6989796603196871,0.9934289127578932,0.3569800241319211,
 0.8398813076417005,0.5507575953561212,0.35825920021555224,0.6799174055620152,
 0.36094489484313474,0.6090428939564193,0.1832970341139133,0.7155767342616813,
 0.8031565851323917,0.07020305609160749,0.2827165641338134,
 0.07626692686867698,0.3836353218194095,0.6296261647397561,0.4653322699025657,
 0.12975687340241493,0.8922819921352725,0.08662709974894067,
 0.5779261632156094,0.5251395271010365,0.2303479776299826,0.8146387465431949,
 0.8701847460698195,0.48156185880265323,0.9430440327863877,
 0.43163526181945155,0.31267020270463475,0.21517050772874402,
 0.1917824945132064,0.19280288354187802,0.425104140220791,0.601379582789429,
 0.40912007578267984,0.25944009738224183,0.5899167629878246,
 0.8669516321986608,0.6876176588789198,0.04229468589465457,
 0.22900720236930716,0.9534602339238596,0.5792311150895367,0.8241173615788044,
 0.047126865630582215,0.5653106755912698,0.06618142299413488,
 0.2693596680080199,0.25373116585485866,0.6224473140415598,
 0.24617044447027714,0.16012031209136657,0.7847123098323795,
 0.05688706774193808,0.5492801363060006,0.9610176396079756,0.483698832037184,
 0.28930511204772724,0.44361513964444577,0.7346187954005147,
 0.24147576204559762,0.3582163858954549,0.6581595552714342,0.456123957809329,
 0.1526669520022882,0.507948693028569,0.4724518546784102,0.7307481787863852,
 0.07561205784717295,0.7238555660907146,0.5669155498242328,
 0.42507969384394384,0.15059791873051231,0.95919170152825,0.7813058159801618,
 0.22616569468329062,0.8870814077877155,0.5303456143479105,
 0.049322270180963135,0.22796550944204352,0.6402758706890637,
 0.6401161121570517,0.3358907624055173,0.6576729918289755,0.0502010332771583,
 0.37354123171441833,0.7017651547788395,0.8485076861224694,
 0.30863167216685183,0.5959313540569072,0.11186039126232628,
 0.24153820688581829,0.8799862310718205,0.7026034355196199,0.4460879253339962,
 0.6929130480756375,0.43261785887665816,0.5650394717395599,0.687065600858605,
 0.3371903305581003,0.2645975228504872,0.9867966452170229,0.3433176705001335,
 0.948825531555597,0.9944133321054633,0.5261143746852541,0.1528572181628196,
 0.4692149769934352,0.49539937719213467,0.14562202505321997,
 0.4725941071236306,0.5649585691557817,0.9739807256827576,0.5278026985783906,
 0.45103260878215834,0.17437818560428398,0.6269471064497749,
 0.7664304974897523,0.3118812551153073,0.707845868010157,0.5364024857341446,
 0.5174026908901089,0.16037885934647045,0.2952163240179866,0.6908365445262296,
 0.8280671572537576,0.41450649217134083,0.4237998779241522,0.8289542348223464,
 0.8789508673285967,0.35409770475500435,0.5058803519547129,0.5232658891146861,
 0.049877430488831864,0.9652838533678463,0.5797636669696434,
 0.3404132449320598,0.29509609434669515,0.802852854485892,0.07684450149316702,
 0.8225374361038873,0.5136726480426953,0.27352985721074663,0.7665294175690746,
 0.19496186497719603,0.7309844908120475,0.6536048555898434,0.3902966388051309,
 0.9742026891619655,0.7903030226938672,0.425386206167777,0.35185530332721404,
 0.5742243890427418,0.5114571662850826,0.16617747081182177,
 0.43488631026035207,0.3130969996458266,0.7519525642534112,0.6088621031614428,
 0.4612965997392299,0.44409225677695086,0.8447964314165108,0.0117411634523783,
 0.6039524405087808,0.6511354183045649,0.8552179848426862,0.9959994765838772,
 0.7220724424249705,0.5832407424499636,0.3526732178896087,0.9827007079420031,
 0.406407371197682,0.2987947449200795,0.3441521890934409,0.2384403573414342,
 0.8523572541588468,0.10506298107487977,0.905290568807472,0.19540208148689597,
 0.6395386441668892,0.7544164923492118,0.20365779341772605,0.3397838081921093,
 0.06273051691308273,0.330329014769152,0.6763460632103241,0.8133315987779213,
 0.6883683652905056,0.9204041678797453,0.912956799583301,0.9086287659271517,
 0.19366203357356893,0.1170889142838828,0.5279184313895767,0.7155400110857533,
 0.421046352610283,0.2647722138157862,0.8129768723119043,0.6703698345095357,
 0.8971758313894462,0.08508346855297622,0.7414182613731234,0.803960050220879,
 0.08383202018627334,0.7295360663419918,0.70778113015676,0.3274948021056012,
 0.7133311540973006,0.1849329476645979,0.07504611306114306,0.131490405680874,
 0.38567372910919795,0.8212535776616908,0.20510518698765978,
 0.5273439011117738,0.27930875785442255,0.5148479997396145,0.5875963427928966,
 0.3737284623091006,0.7285828881511687,0.4102395564935679,0.7830545700137471,
 0.6034403529101136,0.9750021239590928,0.5009936211656288,0.7574875682180382,
 0.661903386595326,0.3396240369769242,0.4199224606872083,0.3406089693675771,
 0.48974961543714335,0.8728187686430695,0.9519514076047283,0.5169114281838183,
 0.14413922979301608,0.7688982189397467,0.3805545453020588,0.1778833963658223,
 0.8570572630886224,0.35468896910194436,0.31551396182464475,
 0.4737655411686086,0.3175344089727177,0.5043284182377601,0.5188527163035795,
 0.03238086005092444,0.11935766029009864,0.19760031732618522,
 0.12611839151423077,0.25334232861218287,0.78440821983376,0.4872453213812614,
 0.9760936775908758,0.4293160593913785,0.49058114080623394,0.6858722347757081,
 0.7914392768553787,0.2790434738800647,0.06756434072011586,0.2732367917172638,
 0.5975367648066099,0.20604394750889632,0.913017237332212,0.7539138699344016,
 0.6454354391744336,0.7399457685419375,0.29954746047948067,0.918990341648904,
 0.3164218324151855,0.7198410467506415,0.6925573420000969,0.8871290621882058,
 0.24715473024198342,0.6238977067338715,0.08430219550252338,
 0.6983684108666558,0.5864428201294087,0.5840406254363213,0.471321629246396,
 0.6434093849151135,0.6224984156111771,0.9847752870880676,0.9021383477198303,
 0.5242185939082961,0.10463805315454289,0.8073291646997163,0.9129219056731593,
 0.7382545286741027,0.5111019893698215,0.4662695308019704,0.11591576347419552,
 0.6525491555926369,0.42499146324984916,0.08898698217656331,
 0.7661387825886743,0.5927676204013395,0.18593898777801587,0.6361710451284762,
 0.0703746120872416,0.3894942022590874,0.7404662921326695,0.9955196368792619,
 0.634911187034696,0.01032753825109034,0.869886279855761,0.5090906275926327,
 0.5842697107827592,0.8787945094510486,0.6729403800532172,0.9606869055301526,
 0.4687328379064829,0.663216063007289,0.6534323241838105,0.2472653800507708,
 0.4030213422914952,0.45007318963570786,0.5839834605295762,0.9597296606700655,
 0.4730538830189942,0.37867715889302045,0.03305389416178295,
 0.9294270170090566,0.48501988552221165,0.9420868330396119,
 0.16101729159806322,0.5877012096217085,0.08730045205913728,
 0.2355158984693504,0.9378911462062884,0.028327333535846266,
 0.8367846594965436,0.6187515553779407,0.5221839207642383,0.8224257563763528,
 0.737800500158959,0.7723568552255841,0.06406074040303666,0.8412776307156948,
 0.2516738928984258,0.15099986735725923,0.41217846338709907,0.668363731970677,
 0.8583673520007868,0.8950469238135808,0.2814123985318685,0.00106623502669434,
 0.7002409109721872,0.2960547533666774,0.7924293065102191,0.7958266605077405,
 0.9043590959474705,0.1343826524819104,0.5122516223536119,0.18674129688326158,
 0.24306072678061919,0.28789270377324205,0.9315118250691515,
 0.6542210496873759,0.8581278059905484,0.6359316947965891,0.7607586995838622,
 0.2925728780101319,0.5063910914655051,0.7625333999718797,0.9095880328158636,
 0.4517412167516547,0.6747169426337306,0.21449039594773245,
 0.04278196122679845,0.6886270249150785,0.7298675657826401,
 0.10626723869974375,0.14138695013819746,0.9244135380079036,
 0.3648201326380178,0.37226747612378164,0.7287438767687177,0.4017157355100911,
 0.6134080577603833,0.3231730544550072,0.3895996436122714,0.637120226552786,
 0.8395199501206922,0.5436463468530275,0.29274335872625756,0.5292617135328094,
 0.07834944294910451,0.49166053557385747,0.21167685225952215,
 0.6983783522784774,0.7341991732578587,0.6015489186219471,0.9665684606636815,
 0.8082966263320229,0.9759050287515738,0.9850047750798996,0.3379294336062463,
 0.13645172032174013,0.5559620331405709,0.44841772772150024,
 0.9255635661542533,0.7720726206234352,0.2814031300450148,0.7736034098814469,
 0.7707429783334171,0.15761460920219295,0.6704786115455086,
 0.23394645912733003,0.9329072409182328,0.2972842071227245,
 0.31184935892739307,0.7443480089818745,0.9679864993212357,
 0.35947624200997397,0.3173444100891366,0.9368956949791722,0.7824195444107241,
 0.12531090356895302,0.11518048044543894,0.9175671342790771,
 0.5850146245243808,0.7318721526493676,0.7803804233407529,0.9366386151319062,
 0.8864314307056059,0.9977990539060231,0.2490032395011691,0.33442585032246164,
 0.9463044074827732,0.5491120240778632,0.13129530484105467,0.7437069398690185,
 0.884111559645838,0.34834224867302455,0.7511401580314291,0.39905200649528105,
 0.7592001982322591,0.696821250561208,0.8815653886749673,0.8772594168943297,
 0.19478951733229932,0.07112383062056271,0.9819003143567259,
 0.21171266685630719,0.6505052739962291,0.14779589643528634,
 0.5401024642261544,0.8442937889172639,0.12493917782119635,0.7189268905929184,
 0.7519558612753798,0.7717721670705935,0.4679142070834813,0.5893533434306514,
 0.7859725106371818,0.6486374300919323,0.09145901245615806,
 0.10486781515902255,0.1765616691689238,0.6101126092524151,
 0.20794958983527423,0.010676162403841838,0.7688901044116009,
 0.3398745573753573,0.19333916492497705,0.34566585141193606,
 0.3105494750296751,0.6152630420980802,0.4037421238090997,0.6474700889570806,
 0.5398331581463057,0.2840876331226925,0.3019950226198497,0.6775281968552056,
 0.06772587036361799,0.15936670631655692,0.7274623614375002,0.505936820109978,
 0.24063197999911656,0.8625766604497962,0.08092026025206067,
 0.6971276081737129,0.3366810873739078,0.22756803814204274,0.4813592104610336,
 0.21018876092678784,0.6963192330197442,0.46295782072034664,
 0.06975930858620194,0.04854043395227903,0.7659687735241444,
 0.9148003719440635,0.33407682470008093,0.6153992180588833,0.6257381887552823,
 0.12065836705466138,0.3013326807719098,0.4556674809026007,
 0.01641651176817127,0.8611620677347496,0.3048909144330457,0.9740183871134052,
 0.8095405080500852,0.23323430971955117,0.3401665275453888,0.2985418003938294,
 0.2962187172348677,0.06682309395073838,0.3673845853838408,0.4047824207123114,
 0.03016115011389997,0.9196687846595251,0.2519406691226709,0.9422789500171392,
 0.9730243187402176,0.3365481143135691,0.7644830843660267,0.1410953199823185,
 0.7381261878918076,0.9512013444545856,0.47824586129621927,0.7708771717838446,
 0.25536627853016247,0.1802738191608363,0.3950778836176416,0.6493280844063278,
 0.2980721265535299,0.6899680478200839,0.5963218181063268,0.9802921628004537,
 0.6781212153158181,0.31205216008778824,0.4633891441723901,0.7930425484001978,
 0.11218872762344245,0.7246934157675147,0.29888130663251355,
 0.2753011408314271,0.3566370237282346,0.8700347251838924,0.5080859485802934,
 0.4052887652248519,0.04015628774179869,0.14295040598505926,0.50683092732693,
 0.7593778855672542,0.14064461321830102,0.7190964400854759,0.6637081178083937,
 0.6273767871058367,0.288562578040197,0.9950600903724672,0.32922649283135175,
 0.33154587804526303,0.5787024458170669,0.7007133135476102,0.7394269135890348,
 0.9687825969665969,0.451887689831574,0.8401043067674644,0.4139203018869452,
 0.7958369134482127,0.3319470474553824,0.5829918900972498,0.08629278119211181,
 0.3412097583312612,0.4386189648287856,0.0025554762062038883,
 0.335214323644599,0.32178333517106716,0.12889273755754893,0.3406349532534563,
 0.5154242242129208,0.39326819157997106,0.4824117909263247,0.9204960517438604,
 0.553900891541356,0.06420585937283552,0.6091405561630844,0.8169770984760955,
 0.8553655824219003,0.06951365860109604,0.24981392000039682,
 0.5768339568582477,0.5531036663253218,0.15596758698964475,
 0.22246810069089473,0.7596784491401342,0.8049492758500416,0.8631230752932811,
 0.24517302031473864,0.5554195131276867,0.9163481740370033,0.4795548601538552,
 0.11086833169653598,0.4048670721739749,0.8308852206059627,
 0.26691737510317803,0.9311681164091682,0.569617034396272,0.32820289791704527,
 0.8930331986891725,0.465841275180651,0.4553936122791187,0.03837166878586529,
 0.1553447648096442,0.5520715192698169,0.49468607794464026,0.6613398802255426,
 0.9194220303364232,0.4748930677772776,0.2919974443920762,0.9874539233079621,
 0.5279021177068381,0.43541984859351335,0.8622102557566717,0.5332232883675982,
 0.5868253991960437,0.827000755903745,0.18376688773091898,0.6854237627403243,
 0.2501008868182737,0.08978202042409322,0.5122596407728904,0.5503738783051836,
 0.11521258862248818,0.4875235869020281,0.7084903011492253,0.4545254346927059,
 0.47662686223881634,0.1066594589218045,0.9995510267153319,0.5532853458905101,
 0.5476372820818647,0.1810411477050775,0.5047864905536196,0.1557645173131028,
 0.8610480890691082,0.8924274278190092,0.849336733161508,0.3917327495621986,
 0.9936057826775755,0.23232740841668265,0.5954229498330817,0.5198176137172752,
 0.9369543369529623,0.88452575599258,0.15972303674689636,0.6505688705769731,
 0.29426255145556257,0.41836749204922086,0.9877087199615884,
 0.8416001055730714,0.2877483909232489,0.8177769930847782,0.4552226310190606,
 0.34957477209438836,0.5215734451324766,0.9832896114593968,
 0.44808115639995505,0.9526763203265834,0.7732729031056169,0.1900934269056508,
 0.6847346604217236,0.728420025944547,0.37733665344425704,0.675863362664452,
 0.7229446030796336,0.35440260573299565,0.4659275954326879,0.4699299307163516,
 0.919133423930107,0.36819252981877837,0.5592547950757621,0.17873038337722313,
 0.2688909675268437,0.6540259615097362,0.08901323401666483,0.9648135323389224,
 0.06493096478577987,0.18456139225149404,0.7643231556635767,
 0.008761825734676099,0.6426130717428375,0.9322971072291546,
 0.7184162276101305,0.7187181214443177,0.4186469275418223,0.4801641419640007,
 0.020345378498130895,0.9549868488795751,0.17688163202671303,
 0.756081915538639,0.05632745160536612,0.46823621402659166,
 0.06956878946417389,0.3097300863507362,0.11662432046468496,
 0.2775914447315124,0.3111823036645479,0.6725733069982515,0.8860428614632234,
 0.04110868706507653,0.5379402867901092,0.3377073424274589,0.8469861728288264,
 0.350183533248305,0.8307379303313427,0.9030223612421053,0.9207177367568933,
 0.01353715561647606,0.3166671233609124,0.40292423985211423,
 0.8666421895844378,0.5168290414163833,0.7648548654295663,0.20433404867458171,
 0.5443724792986449,0.702189619374788,0.7358208160649005,0.5901437411524364,
 0.371508334053127,0.15118143483632873,0.08026605818358346,0.7891759751360341,
 0.4904460635242329,0.3669498049341604,0.055394158687738004,
 0.8012080688535834,0.2916991906795633,0.3918258263075991,0.9057405559800622,
 0.9184059551804543,0.7412430672703902,0.7855670860444146,0.9490103609187468,
 0.9110634295887812,0.5964884052713256,0.31531282592231924,0.8990248689964763,
 0.5542988037006427,0.18941270651437692,0.8127781235021934,
 0.30409862742090865,0.2204049829959167,0.09284805808170271,
 0.45723318757518305,0.17534222901856522,0.7088994425479086,
 0.8981844695825557,0.6698267448947213,0.9992552952759672,0.35554767517416186,
 0.5438341260873218,0.6183931149316861,0.5636165188239746,0.8329702199999267,
 0.7184115250459587,0.061477857886586995,0.9942749015722558,
 0.15583196635015528,0.7841946936018347,0.9830442959734272,
 0.09532665432263798,0.5430844543996217,0.10561531060863694,
 0.7960347699505563,0.35132556643660207,0.17381636441171722,
 0.9244491680256919,0.5900238601271257,0.03940111695795401,0.5767840606249571,
 0.7991094651369486,0.4474686725520235,0.4122187486804205,0.9927680824272994,
 0.7772311695142897,0.7223587243688057,0.995098780666894,0.574160234841831,
 0.2363144066927536,0.8247696141823531,0.39394013827164465,
 0.47929335017813035,0.029946074469582973,0.9327775478052379,
 0.30416384864301416,0.009434183967990561,0.8474908872122083,
 0.5280515842325351,0.5640916549235073,0.30347063757332815,0.6294197800652825,
 0.08594540499162007,0.5407179442894532,0.7296942030331396,0.8918553543353014,
 0.8158003694307978,0.41620164705625584,0.7317565299795032,
 0.15096940481913024,0.6836123148738141,0.9924674544553338,0.6470450636566917,
 0.3224899574469213,0.8991814684581798,0.7989123783554752,0.47593967673166193,
 0.8496669068227082,0.7872151923542166,0.2788869688614567,0.905076743537429,
 0.6201727087425138,0.9872310376415273,0.8466294498172297,0.18642408014308431,
 0.9516367548726988,0.40045251223790457,0.42881147384566765,
 0.7432092386273473,0.7672369925025122,0.5001396457379317,0.02459886428980207,
 0.5807136394380443,0.419627558988359,0.009840236838460914,
 0.32070631732644106,0.7160136124952743,0.2611689733306487,0.8734688025026265,
 0.8129331775540859,0.7275963648915399,0.098759491998886,0.8135632018952426,
 0.21142605650486657,0.251930087444728,0.44085131768355124,0.8221093434730321,
 0.6573343213976508,0.7144386548243696,0.034352678578367546,
 0.9842864647000165,0.5956599437134027,0.11366206129314649,0.3341862217334345,
 0.9772161138408475,0.7740505122564939,0.05075180759822162,0.6559718433361981,
 0.6221005196016769,0.5195460078501654,0.3736673751674042,0.8658101678716189,
 0.34720932833426665,0.8158371298581667,0.15787338567361486,
 0.12184166843079192,0.38954033304970537,0.5653399074383465,
 0.36309287784709554,0.2615502388210572,0.1903041014330713,
 0.41459178548642606,0.9745921087426112,0.7599927948780899,0.5523424783943814,
 0.5978761131069471,0.8909077304653259,0.7227442995039712,
 0.014181779542567496,0.18385712864207937,0.07683233706877068,
 0.8223874865572665,0.7545856848985113,0.9423633626361921,0.5997610929778943,
 0.1733310375881052,0.45569793626816835,0.0944782514913215,0.2921063975342212,
 0.6688002037500422,0.4287493267317737,0.6738564363718924,0.6676744819513034,
 0.05513846408445444,0.5368929188626264,0.7163172324639949,0.1857694840574674,
 0.015700555431689245,0.598832248047973,0.8976293015359043,0.271940048514529,
 0.3663450965420163,0.4273741660717335,0.64384136903478,0.6037260156044543,
 0.34329650598527517,0.6800423133775351,0.45286671695478775,
 0.2880293677345821,0.6831268822009076,0.8770778096675248,0.9549870582836111,
 0.21076175343418513,0.52966484046379,0.11221939930244373,0.8609927934867548,
 0.8732041274941372,0.8030740775374104,0.934636198295754,0.5966910141527654,
 0.5289978494590198,0.6861789147888677,0.8499816115357002,0.34868550507722207,
 0.4475712462038097,0.41549755429329105,0.4877856190476566,0.7005612053472995,
 0.13767168476922154,0.9147663107630071,0.5847591535003841,
 0.011502311448584823,0.49940820177495315,0.5798840943930587,
 0.7942950360886983,0.21035279284369146,0.8321204292466603,0.4937214139940045,
 0.37011168135066586,0.9999770195078135,0.472676150002427,0.25744211558185404,
 0.6888458916021211,0.5357452500529436,0.33768515817710054,0.6555088170912531,
 0.35518669603570585,0.928071905489179,0.42836832983679274,0.5717160799375816,
 0.4690528622443261,0.9370432474491467,0.7246270000831065,0.38373190088356823,
 0.015697994444733077,0.07711753895160245,0.11083907405401994,
 0.04091206799774538,0.48033690960066555,0.07548790868299027,
 0.7565497603692669,0.4565061785658022,0.6308305038955154,0.57790794023169,
 0.8803209859543195,0.029077884632519257,0.3431335624240973,
 0.4956433152465912,0.14322296090618203,0.7129790818541428,0.9086179968621971,
 0.21102509868922803,0.5941087531041638,0.46372379290393984,0.937007839222596,
 0.7831564430287284,0.8626484687472272,0.20121023219244394,0.7475928361457731,
 0.7904948071273683,0.4043704891754961,0.9447276696027919,0.8680506956027401,
 0.613109603435118,0.008267369867025698,0.6996058247119471,0.9788179476459926,
 0.6946678329091096,0.8080925402450456,0.2398781650236843,0.5647344198438338,
 0.47942532707941565,0.25669745089525664,0.5830996878474375,
 0.8076996741202112,0.026362489717757787,0.20241949745545007,
 0.3458476360804241,0.5564907295903239,0.8222200051705381,0.11835189569172044,
 0.2354857917312665,0.38937015945718545,0.9154002448548721,0.0836254246885304,
 0.6632910890009447,0.9714867015027214,0.26831623175964436,
 0.49239027744255515,0.8683609808400485,0.8130462867075606,0.5183103717292203,
 0.8730078565311437,0.06142314125348869,0.9783411158962975,0.7520203219930687,
 0.05928837864767622,0.3239744784720521,0.44885079887163204,
 0.9872212016171723,0.4785036118119612,0.04436678787040638,0.9457518791233107,
 0.6262081710001289,0.35973023363997236,0.8583051393078798,0.5591453841996967,
 0.37262145453473516,0.004075529143242607,0.39297789426238716,
 0.9427189339740002,0.643921584037384,0.06962118457049493,0.6109438435525396,
 0.9863743987160248,0.09131683410927427,0.24879375392125436,
 0.15450388555307448,0.6589883139582396,0.3139234474369462,
 0.46389430156721234,0.7593200460401176,0.18807647927366444,
 0.15296224448853346,0.7640867684554595,0.15591418758145714,
 0.5928229044743613,0.5287912626579526,0.6033989071823093,0.6875072652261272,
 0.6175571224351333,0.7392073144673158,0.06053169647515699,0.8383778982999276,
 0.9940315988545478,0.8141273121031856,0.6337217886155732,
 0.038450109775370556,0.3838986533518871,0.26830818769291853,
 0.9894855801599234,0.9511066437451619,0.5052148493260664,0.48694606342193203,
 0.6408188554970764,0.938376690193907,0.26822015831061574,0.8077133701446233,
 0.9405201673298867,0.4168516808990905,0.5090303657036526,0.4198406680974329,
 0.16555914934767346,0.9841959692060537,0.3177406964581456,
 0.32179464885574116,0.17817241728287225,0.05726945340770606,
 0.19931222388509762,0.790636353891917,0.7449899983906231,0.6818762552500787,
 0.8650952971030476,0.2853650152732917,0.9393260477710935,0.4634902915607947,
 0.20781492249091793,0.8548311219804283,0.850807323227572,0.9129640352158351,
 0.9817884300383858,0.10183882490882079,0.9555124429647639,0.9565102541164695,
 0.48083916504073976,0.3326833220216502,0.8567334118785185,
 0.47203325197470114,0.6543253933880175,0.6151475210897868,0.2831550533783548,
 0.30857097681977685,0.900549025157966,0.9965891644908873,0.36236240309540324,
 0.7540216704328013,0.02891953530018876,0.9787659928852963,
 0.030135012246142012,0.3423033263846723,0.2824503409803043,
 0.8987208073220702,0.2643513064190556,0.9193209582505685,0.4004583849049801,
 0.029960127918086776,0.972444380620528,0.6542361642374644,
 0.009226288207508482,0.11428512923301315,0.7843910980002883,
 0.13286588846623182,0.6562710629939957,0.5443416766866053,0.3027254392365417,
 0.21987259356947264,0.30155328147283567,0.5185451991417439,
 0.44073693837801786,0.6410003073299795,0.9063435423159824,0.8693353230968635,
 0.43175039630409995,0.48139269463955026,0.5097868468454513,
 0.9950123920695065,0.49039854463844923,0.7752498792246213,
 0.30041582850312776,0.7096408497625958,0.9776818581282172,0.4466240093603361,
 0.8988314116119512,0.5403640875455425,0.3124102660019814,0.8633393690543462,
 0.25503734183018034,0.3889269136201634,0.4521037099843538,0.379665589695793,
 0.7181370997946468,0.6015387422978097,0.7116523453460919,0.37060392892146066,
 0.32826377182366373,0.31112096521842436,0.5346990296512736,
 0.1699938741564554,0.25543500986805123,0.5899833770611926,0.4551057803809684,
 0.9392732442849158,0.09443491870920429,0.1602858939819123,0.8212694370450164,
 0.987565489218793,0.3631470328967731,0.5359276750602661,0.3434960156784378,
 0.6565038629844064,0.7305694848912189,0.6642804418537686,0.5830115029955296,
 0.2597344072774729,0.018252963475288486,0.877040084162528,
 0.07766685065856016,0.0710801198128791,0.30812324363508725,
 0.6896716103965496,0.4332265773997749,0.6373052029290617,0.6755676847563613,
 0.047527084043199386,0.5397642764251701,0.17550270521007694,
 0.9541631970395205,0.7873997861194403,0.9973990480940014,0.9591393405248366,
 0.6166194411991583,0.7373741681218055,0.34254441368896027,0.8963788600483136,
 0.11034585152630305,0.8164366496901799,0.9910555774546755,0.5622342437054021,
 0.06351399203216346,0.6782364278340662,0.3712918278951217,0.7076743589409836,
 0.319728106927607,0.2771477927852223,0.6350324125542643,0.17890165670705593,
 0.43631215940988766,0.6316450067073588,0.09873117704629775,
 0.7995263753156995,0.4896601745358088,0.24830660416187977,0.6884832176233043,
 0.6946234434664943,0.05635539449164828,0.5445305208298582,0.5276244770004915,
 0.4407645347109025,0.07230271115690723,0.6218098361189182,
 0.08003226684700482,0.9886310008677639,0.9165878514157367,0.6001334763996672,
 0.18645748260918404,0.6229530018075595,0.8681761364344573,
 0.07996164267646744,0.9234044875222133,0.904939784060318,0.8944899555580348,
 0.7042246493423743,0.6195434742060743,0.6518397879323935,0.6704982748209585,
 0.7255528283244769,0.4393717574007966,0.2601614129749218,0.22115294224288184,
 0.22482600788555684,0.27433692288253475,0.012990528743488494,
 0.09408241168151088,0.456661855602156,0.713370478450539,0.6141447582086792,
 0.9038582926618481,0.31753960249724456,0.8658099436763027,0.7691791298355837,
 0.46286257783940754,0.04554617100485581,0.5518523123555372,
 0.6363792567252917,0.45313904413268935,0.7204671428282525,0.2716753484889666,
 0.00990641247339763,0.8243410718311164,0.9430052996941736,0.3573391341925669,
 0.29830105333931356,0.436914252884834,0.39797297163222844,0.3817634712178096,
 0.40503535143078206,0.9497825872806186,0.9525677258213543,
 0.19897416173175098,0.752877425806079,0.004289624394801184,
 0.4957515308895779,0.05265659078057083,0.10471272114774699,
 0.8520220283297496,0.19628670607822762,0.7443674054898255,0.3998843009825358,
 0.8539842971341769,0.8786354247986945,0.3421307855479474,0.4794211658910832,
 0.10690904160764614,0.1774924933047981,0.11797337883633563,
 0.7998699832646503,0.5404704804958158,0.4305652050409179,0.6556851337859841,
 0.9504705161577682,0.15649126845076322,0.8002830230278163,0.3269182508296966,
 0.8383429069376388,0.4318866015731898,0.41207667169693785,0.5584624187091782,
 0.9547698979974776,0.9524143739582394,0.1727480388204028,0.9794511263454446,
 0.6240382242969926,0.07529157524949459,0.5488523369467774,0.5571173290904202,
 0.0279922118121807,0.19474291029829582,0.21776442956482633,
 0.4761545954029166,0.01617298045969351,0.8798161395771461,0.3724965808145324,
 0.995117451062513,0.21951718535600284,0.9347111639867013,0.16494033104459982,
 0.07498639173513344,0.6821422517779122,0.1614744643029793,0.7799788857864705,
 0.5610839321021525,0.06615779853799186,0.015208507355487955,
 0.7708239520869267,0.06739584205711013,0.5481324859779356,
 0.10549027571116643,0.9821174495488442,0.5679539272896855,0.7165316682992918,
 0.5923977620944909,0.42231088103296255,0.46235257825172615,
 0.8264220847855275,0.948487856123629,0.4412411163866965,0.008725028303717353,
 0.48941209176924083,0.10020074539373824,0.21586491080140202,
 0.14890272970399976,0.15547502886999776,0.219096041245588,0.9999433762245685,
 0.7121729033427765,0.27970959926786965,0.03393557788283219,
 0.8589580260881389,0.3919747488257289,0.3404857324923872,0.8461594152208503,
 0.8001646557227001,0.8786310209218811,0.6097684740202173,0.9534973866353919,
 0.6299101757862448,0.8887660978093681,0.08964697766471896,0.7036900847007953,
 0.07510418764456872,0.6732601319254575,0.5566281548009716,0.8126640612241399,
 0.04984415379858065,0.25858471072535916,0.930655616026608,
 0.21337553268629994,0.9147443809686651,0.0777625117296491,0.745554638793073,
 0.5775826947326518,0.2118046091895086,0.4760068741594402,0.25164480037886094,
 0.14381783586094354,0.0310641733208189,0.9744781119582171,
 0.24965755504177345,0.5101170396229308,0.21639859478871637,
 0.14890716029688378,0.5015986943508391,0.2308948691022895,
 0.05710617469817514,0.3615250444234972,0.10637415232286274,
 0.6623622641383943,0.19894477806930866,0.6380625115455136,0.6057414457182216,
 0.7811082810921012,0.5646235755830488,0.11021981348931398,0.265717293317032,
 0.6057819221475124,0.8614326464125491,0.684754435912933,0.3347930537992958,
 0.12092322821582857,0.6164211703101691,0.6122068046509704,
 0.14764206654642342,0.9939005410809312,0.12124918971396503,
 0.16572382683309628,0.5385607813145361,0.20088766211097286,
 0.9777902115154751,0.44289568268575763,0.554001907743395,0.6507971180948696,
 0.16219600505944987,0.045148973844754026,0.25107163944513233,
 0.5720916039321949,0.04962741992246744,0.1745319882485009,0.6555916949558068,
 0.2500280942852233,0.22706036097468385,0.277855524201872,0.10729779025513586,
 0.1270745038838077,0.2713262742173831,0.04051799613968654,
 0.06227976880544894,0.7412158099924836,0.6044115777472086,
 0.01246020640107548,0.22951441504684267,0.3629773656214139,
 0.3802230696425952,0.7287195350642542,0.24697643478033426,
 0.02276452346446889,0.8617635683405439,0.54226684693888,0.673171501653066,
 0.2717958092440883,0.27016715951602255,0.20820102621659475,
 0.2461096682262558,0.8369345816839144,0.257711982870958,0.34413271741189977,
 0.06040953137169458,0.0766798899898058,0.8237746270826529,
 0.11185067126900561,0.35262338802614224,0.7730253785995534,0.768964581500545,
 0.5100597694258115,0.23915503641538693,0.5592216599291239,0.7620369428432767,
 0.5926102282199012,0.8441215515905653,0.8039828590079405,0.5899266145248997,
 0.734116234508684,0.46479620625295603,0.740807913540083,0.062468981991276706,
 0.5624298628282213,0.4342874681182112,0.9209983068036646,0.7522966288235675,
 0.2654210357918485,0.6589906073437016,0.11751259122967317,
 0.11564025413514312,0.926915135056882,0.007573119412299767,
 0.020619478495226273,0.6735266243721341,0.6006596651089856,
 0.23140426184179475,0.8128320230864514,0.18833871885392062,
 0.4487407191063856,0.36104083822799427,0.8352776899768138,0.6988608745333285,
 0.17429325755764002,0.6075637419429215,0.806068709428198,0.7983142106839115,
 0.3218392693507308,0.3382593256304163,0.9701221016520889,
 0.019694104349436414,0.9531381421450013,0.3542808097185888,
 0.9521879993474782,0.5562757703257418,0.624895251013334,0.33747915512711213,
 0.2923371297790063,0.00370670493753944,0.47761953514397903,0.595142102434695,
 0.5011581058608009,0.21072272955345817,0.031110842264343797,
 0.936106940446488,0.9740432121299069,0.9829886896371284,0.37618484565840227,
 0.15081515776015553,0.26424677939311303,0.3879137392876916,
 0.39152972470649394,0.7378052492952163,0.5089827665052783,0.7848361526204002,
 0.18121338992755698,0.9683120938258772,0.8944447026475869,0.8556669745589748,
 0.5273862386642432,0.4438304440107954,0.19878162415937772,0.8817727668474115,
 0.8156295129730041,0.7592445144546764,0.430490507327175,0.3117400832958137,
 0.034660597779148206,0.5402221830298113,0.42735362410515887,
 0.9174745547620021,0.1896470814925063,0.9893408565948522,0.678608498588487,
 0.8710447858189498,0.8678627195945301,0.056941410383855584,
 0.4513093737199132,0.5627655702891967,0.10786630207997916,0.6661703143841124,
 0.013058516146634558,0.22017608825296764,0.0727950893368774,
 0.8890377764298759,0.4489679564098362,0.060257160287688905,
 0.10901955921031359,0.06785057572042297,0.7402495544246197,
 0.06127338288189732,0.08432286500205766,0.748413766997646,0.26933777715283,
 0.160524594871563,0.052901393963843746,0.3486092267467006,0.9831915735285474,
 0.10206316461416365,0.9539342288336666,0.9429023369326199,0.2357387141859817,
 0.15729883260137834,0.03280642619122065,0.8582179565499377,
 0.9527809131960908,0.3981411016614391,0.09809228012937266,0.5412259462768056,
 0.4209063714814204,0.6855549809886547,0.5629148537652837,0.5378132624040494,
 0.6597419875699464,0.45957367494887014,0.4848464470556664,
 0.13374269657426363,0.6444104343124638,0.5795832018123319,0.3047467870548457,
 0.5769977899308046,0.5523496926843654,0.9828613153832391,0.7242048190266963,
 0.37074206248981767,0.1956028461966286,0.3564309600630503,
 0.45382980698949216,0.6181509550094204,0.7126473995961646,
 0.18818337146657138,0.8389173673846623,0.3932991105396484,0.9062756319334799,
 0.38177214192249675,0.43785136002452263,0.8349350670589415,
 0.5253057782585329,0.205777317918162,0.5964753380106513,0.8270542119861535,
 0.9530878585262211,0.982710634454012,0.6709703072180455,0.8625542911710995,
 0.24177045832244926,0.2217282062291458,0.46097314337043105,
 0.05118430416881836,0.874836097934798,0.029629351440557672,
 0.22932424962778009,0.9142611595202926,0.1906474204404226,
 0.16644398304025199,0.9655221847980456,0.19251526212801073,
 0.29038809421636114,0.4650978740625843,0.3160781000182624,0.2935793090005263,
 0.9799447118830755,0.34614560362662816,0.7830690453268744,0.5321155044743726,
 0.7764643949861249,0.1595635160233545,0.8886852331576125,0.7153039533706389,
 0.936326285344728,0.38150643720853605,0.5220945438835026,0.836447920026803,
 0.2873292409022802,0.4369343958220774,0.0965831840444088,0.92537270739724,
 0.6263379333606321,0.9700424932353271,0.48840434593205617,0.2421366239655779,
 0.6410790693828198,0.41294179279624843,0.3191053697271867,0.8178735827691189,
 0.9420424504325864,0.819983598734622,0.4040625453069908,0.5738977185706796,
 0.921113132742244,0.12080930519284927,0.9030784128861244,0.6469099595706331,
 0.2630309538775024,0.14316111178208046,0.8428294778033557,0.9088733529356569,
 0.7244555569126689,0.47076062344352276,0.4240473790048178,
 0.25147888325870615,0.10950404610844433,0.974114216653608,0.5337364522203212,
 0.4925861602005801,0.44810793185938,0.08806863843843171,0.8627609147038624,
 0.14148691103497946,0.5731840532218597,0.4532971688401558,0.9369213288862901,
 0.571222563997155,0.9794371604981879,0.31824336161970457,0.7547475622478537,
 0.8959595671046808,0.1535048416264746,0.16177991855228835,0.1746234024503922,
 0.20557525107620023,0.12812041754766523,0.8500647998694462,
 0.33555539182010785,0.3194663516564975,0.7968777596177309,0.8637395203780094,
 0.6781574332711712,0.9149591260824373,0.6042958086432995,0.9482693599384202,
 0.6861534398527022,0.3102476922805705,0.6887419935399227,0.8747378119520975,
 0.6067637512987053,0.7809764200808708,0.33156184431365787,0.2257898030835378,
 0.8306882878383195,0.4318268402008003,0.6600002686341293,0.16984709402961717,
 0.5237488087105746,0.9730129149537768,0.5251707233389086,0.12344551134871373,
 0.36797435127731737,0.18687454906710566,0.5196853126535379,
 0.1215495136068594,0.8343359721255312,0.5722454727842405,0.4391020967767645,
 0.3402238192874184,0.5111986885533237,0.5663875677152128,0.6019973928541594,
 0.09852962962370038,0.4591761640321178,0.6782100565436207,0.5494612105828912,
 0.3165039326505542,0.9869945118876032,0.9701630779748539,0.7610653677179174,
 0.9097014748571645,0.6567846275225386,0.7429772705387238,0.9411701440938854,
 0.6999537404953589,0.397077356714447,0.24442056727229433,
 0.030222107662556952,0.4716925777539964,0.08891668115242235,
 0.3749845306806351,0.11852984583132964,0.1963668198105194,0.8135470294293382,
 0.8030432397443968,0.416672549477313,0.9827989906837729,0.5684224057709412,
 0.07569814446933165,0.6749345311084747,0.6261935600365182,0.7249602527627715,
 0.2552646229076526,0.34666190949948295,0.21311731653566146,0.515132271914573,
 0.6456220208694569,0.5741739735442897,0.7818301555436858,0.21475817123071006,
 0.626781996518593,0.5188838776894849,0.20088260320545714,0.797030403482536,
 0.6256430178432401,0.733156509016177,0.23836520273800266,0.40563488675443726,
 0.18856614671069039,0.04045030409323802,0.6643566628617491,0.399476881648662,
 0.9745583232370574,0.957163664353757,0.062362398809253095,0.5239434905478526,
 0.06981036822484044,0.5297447005634466,0.878567391020236,0.13559494618049805,
 0.8572095061311591,0.3158750559481719,0.1841473392114522,0.723056313940927,
 0.1327335062587509,0.2607419459704913,0.20746003883221398,0.300864839363713,
 0.9880256653898624,0.6216126112840963,0.9272463718424849,0.5316019379827445,
 0.5705223936608916,0.19370807838612392,0.7009948874993428,0.8353787045095254,
 0.893087470491643,0.30499875670489496,0.6230335445318107,0.9996424899491733,
 0.5800023893051085,0.7601386867664452,0.8467595737834599,0.5441410404494929,
 0.28759579936032387,0.037628010329267214,0.4237926627406381,
 0.48825161940585315,0.12006401226720642,0.3064991174929842,
 0.33199908766038055,0.5657336561564761,0.9249853796284346,
 0.19901645121728417,0.22798125152970639,0.09748154480562554,
 0.0260535048433983,0.06984724457446945,0.01148778023887409,
 0.7690105590111908,0.2849501739881948,0.9432035724187885,0.6151026327458043,
 0.49758786010406275,0.7061983941210419,0.6321879112046271,0.5798405850466601,
 0.17287187875573595,0.5779031849015239,0.19778155742103976,
 0.017018811343897045,0.6631821755433452,0.6258019572483123,
 0.08348111819330228,0.33739786671483385,0.4617321187188075,
 0.40345779561332973,0.7447983036862585,0.431102446387037,0.9108899378331817,
 0.21471898048548876,0.7094957651472603,0.9810972840159614,
 0.41875229441262696,0.518324610232258,0.24594803850083058,0.1152399201647556,
 0.7911957943248538,0.8745135772981174,0.15446141572907102,0.4418814554205175,
 0.3730166822590214,0.9361568782108612,0.013147334493130436,
 0.01283377637457328,0.8973965010285265,0.2676340141411392,0.1738377614094958,
 0.33446229981612274,0.1645559834792223,0.26707897989089,0.39640367140095667,
 0.4202832761800904,0.8869521885175425,0.27841312325028456,0.7581058544119603,
 0.5665634136288874,0.35023368347368444,0.1874808586996901,0.7463982765783541,
 0.10424862324068318,0.8926035823894727,0.8454676532562707,
 0.40143261890869875,0.005007861494737087,0.45613066316677897,
 0.06316718697663615,0.7456368804831459,0.4725352127721689,
 0.11586999785410101,5.108697098422654e-4,0.5503221068005394,
 0.08101719216157244,0.5142148112981879,0.7317716887251886,
 0.42274896859127764,0.7859410581676168,0.9163275616301009,0.6614200746719489,
 0.24032351556814635,0.014674437830996423,0.08980506131782029,
 0.4364535302629411,0.09481981677763118,0.4975574535506684,0.5352628613173565,
 0.9650099529543018,0.8399523280163191,0.8494241179716189,0.5211895405093947,
 0.7228222243608086,0.6663314270453756,0.22903408761428168,0.2516857843362609,
 0.12314829092344226,0.2674314396440539,0.9348466391367868,0.5357945758019366,
 0.5524478384205282,0.7502285872238001,0.03003516136151463,0.8403994439155275,
 0.0384688846543183,0.2034518946059115,0.869236546407934,0.7859464359994724,
 0.7910976004889484,0.13142848526756268,0.8089671870186945,0.7210981020637715,
 0.9877106948938019,0.7757337375587067,0.5701108309913345,0.47382840429485773,
 0.8983225888866361,0.5291059253456403,0.3502546820156983,0.36958023626338443,
 0.8873266133008157,0.24080028614898363,0.6896339585384056,0.4273591341754769,
 0.23638778890514722,0.18119575821810818,0.34802959520440235,
 0.12197509452003719,0.13612484161774252,0.8894141768329935,
 0.011324496211312485,0.14025892581718646,0.7502947917898002,
 0.49611701685927767,0.12234252336423612,0.16200414097842242,
 0.9883873600168095,0.7162724493696856,0.7915225788441722,0.8325797895405883,
 0.08695291978508513,0.17316433641062623,0.3453373742374827,
 0.8766931250691234,0.9846604373861305,0.04399764764339387,0.6048382635912226,
 0.443215274219364,0.5088655552197092,0.1143927708132706,0.4959022425964843,
 0.9175688810820664,0.5878619048738936,0.7505258328715897,0.49659464136879317,
 0.6713415576975178,0.6038103250387323,0.8448244309708175,0.6720877548455602,
 0.40132108998953175,0.790438689723041,0.08106658000846123,0.5509333462173026,
 0.6382169142155807,0.2632564917315716,0.42953404567039666,
 0.22246765356914588,0.8751933274387825,0.3784006835196907,0.6397879785018519,
 0.08526031597699224,0.0010639128061726844,0.23881275926319834,
 0.6411010243344103,0.9917305144851374,0.6608549073992389,0.9594396665522983,
 0.9020777108751793,0.9216125586073463,0.05832895944532812,0.9196108185420204,
 0.26260280370790257,0.39456493970453066,0.4380740509005119,
 0.5060752309376054,0.48581219242060014,0.5746685530960991,0.7218714915101535,
 0.6795736641219903,0.5925711776376712,0.19798675646846897,
 0.40820186612451925,0.2306902066398081,0.2654902330950276,
 0.057852088184440764,0.7462487977145711,0.3231200655184958,
 0.28277086824742115,0.947760595661103,0.35272931900275717,0.6228802406412205,
 0.9919431926329805,0.5689864838355796,0.7361784317043181,0.8636298838140896,
 0.0782888516661483,0.4999335949604429,0.29391546853234185,0.7017283110268197,
 0.5142596750262592,0.8495062247964593,0.6106425876523498,0.4647642992463885,
 0.6543194926644009,0.16139112582974746,0.5119888129287771,0.064258198561013,
 0.9732308420221525,0.6966373809775962,0.2716624744924345,0.2750913892559015,
 0.4217402869628031,0.02918238108155724,0.1677421518147687,
 0.013024256199615447,0.762107539211172,0.6698276365513361,0.2536435364354871,
 0.28591233197589594,0.1711581392908087,0.10751416005180503,
 0.5163366029952934,0.9593852836063088,0.35823021850270664,
 0.22568391373501273,0.8847706911244394,0.33147390521828135,
 0.0338116993085541,0.5547339287533146,0.6181095590530158,0.9854336125389465,
 0.35070839962628,0.9093457953769479,0.08178649356893175,0.5121583772922791,
 0.6862721090404386,0.015528513002400102,0.505000608770476,
 0.13935709321647471,0.04563501987687424,0.36770550699380067,
 0.6664010459518142,0.6868501942370883,0.14672928275846975,0.986935883530232,
 0.5700785306646959,0.9070141189695509,0.6758578363775665,0.6883880210043792,
 0.4412163788565111,0.19968393237225757,0.8734531095814279,0.7456715134369769,
 0.801617679029598,0.8682405309010413,0.020176744635947985,0.6483102325483495,
 0.5561831105392628,0.25592677120239904,0.911574053689503,0.5661630361010401,
 0.4196342146343299,0.2980390533322206,0.4378694518840084,0.540926279965877,
 0.4430034116174537,0.9972865272133257,0.8680758164620742,0.1344030997689355,
 0.4606502501767562,0.6853195773608837,0.26818314195311577,0.3046521573864861,
 0.7264418261877783,0.2710729903650656,0.31025287592370066,
 0.23458289441594105,0.8079774426383264,0.3188274673812277,0.6494944554540362,
 0.7563428099794028,0.3509202097151518,0.5431954172004801,0.4915100365428525,
 0.49363186129446834,0.262388191990218,0.06986778402838167,0.1993767959244942,
 0.08962348590220137,0.7792019510521803,0.9475307869145656,0.631134697611621,
 0.809904476276562,0.10323066824100502,0.09064823808345712,0.6091180973114011,
 0.9344618031647599,0.11901362246262537,0.9877419539291808,0.828751831320793,
 0.35418671689314796,0.6816653370616257,0.5115782584127215,0.1851674804790998,
 0.45124963659449324,0.17961771361360424,0.19510048473016217,
 0.6183107371440598,0.4042392860744105,0.0748689806323406,0.11478286081173827,
 0.357071067729249,0.6065962243206732,0.4440388057783018,0.4798400481897662,
 0.25153350469823743,0.2625627235433371,0.6993480885901271,
 0.36051212980367553,0.3233804377418802,0.9751530612168652,
 0.07069081915989361,0.09015149794290545,0.8476718544318969,
 0.005051371708654218,0.6688006268390083,0.3554948506318265,
 0.9531777130185941,0.07380286779325329,0.6567887391742113,0.2507026334110909,
 0.44113067276485207,0.09852356055776978,0.6522639409400852,
 0.8593316495162197,0.24625543035648134,0.3387116579523295,
 0.10349749764739513,0.599747618799978,0.7654462676443378,0.8961474450384181,
 0.254453505697105,0.38394076713764047,0.5590194488563,0.5225873670205017,
 0.16856876988266023,0.5783134806473545,0.9362309745490537,0.9071980612641752,
 0.3708729024637416,0.5629205673553571,0.9888656115526709,0.8282401518656017,
 0.11509049377926661,0.5995018444618652,0.8886134037414473,0.7286480572600608,
 0.2866724821298131,0.98696781246317,0.9883331902876885,0.9318462044163242,
 0.2019643348091642,0.769443582208321,0.9637664879629617,0.07931218548845331,
 0.09040186735104683,0.5289161634600907,0.7155972605421628,0.6874869238232176,
 0.27246067491882653,0.29383676954753457,0.8951538720107908,
 0.8051459965054792,0.9673771788951706,0.9707525491248074,0.5787878594755385,
 0.5086480747842989,0.13546557862197028,0.7798682130566484,0.1780130582008549,
 0.34507443242668057,0.5151354069838683,0.9957655010762526,0.8176404082184879,
 0.09341170244736952,0.6651628154602934,0.65095416798207,0.5487362343660392,
 0.9824623242964818,0.7834640849512688,0.3324151735789542,0.4618765345072219,
 0.08002454883087962,0.15715964707545527,0.3439314463247155,
 0.10088117719056033,0.9889816043392152,0.6439711654480662,
 0.10087555171931184,0.3763476241026512,0.969299254350423,0.6474595939595179,
 0.6857625975001724,0.8545970515913801,0.4323402170362236,0.8636002150535411,
 0.456109080486349,0.8162368146647249,0.2617287939329751,0.7665737565843437,
 0.7357187307060609,0.36468539551136836,0.7970305877313565,
 0.14688088806487964,0.4002461020273379,0.39294596258539827,
 0.3372690778066969,0.9457148721318659,0.15250234105860616,0.3465587214088167,
 0.8386503365157288,0.18526565866700218,0.10127253382085943,
 0.29241023778146846,0.8907566573993073,0.05481316672269554,
 0.9591192451371652,0.8619778879190738,0.2456623487563604,0.6009253691511043,
 0.43427815664451186,0.05190438833095501,0.21876877096965708,
 0.8999319851713694,0.21697806096693506,0.6262852127687764,
 0.14820033157051804,0.6111397179974152,0.06344909078081362,0.85250579981061,
 0.2078637694011216,0.5987754527553317,0.034887573749786016,
 0.4136433199396984,0.6878155784876253,0.9219880245219613,0.7400833098884734,
 0.09953662513560357,0.8788311483069025,0.8271941256498279,0.8451735223713115,
 0.44512852207118614,0.22379982964199285,0.9857878666218578,
 0.20890801253599456,0.946457336242968,0.8309728263738453,0.7652788923106666,
 0.9196521260759903,0.8353616617387932,0.592960802170967,0.11204014446203636,
 0.20676668911890317,0.511424900952578,0.9905424568563272,0.1887813151672013,
 0.8029406702807951,0.43882290169757776,0.5238058544128925,0.4199395150970793,
 0.5120360608693713,0.1421705548324771,0.29796395284049015,
 0.46706806009171464,0.38500501642158214,0.47574106119589266,
 0.5555570816027228,0.7887250604237739,0.4756511316792734,0.03707295380420761,
 0.9240911884596622,0.1048592200999734,0.8380144324435086,0.368544109177144,
 0.6535857506067807,0.49046287967778623,0.9925573518775712,
 0.23876033680557163,0.02950632029770439,0.12235296425682651,
 0.7024875173102968,0.34591993706740387,0.5831594950622497,0.4535282994587063,
 0.9508062026418029,0.09818650047132171,0.05491867682190554,0.751881471396163,
 0.8126049369323254,0.8987656274282122,0.13387715440433046,
 0.22412397783624594,0.6930638056967523,0.3052728887842052,
 0.09604025833443319,0.45082027963500204,0.47004362377980435,
 0.06196145689319543,0.6831442275746489,0.6551843496531714,0.5187158782856718,
 0.7439766311212279,0.8798484589280914,0.39335756376086517,0.8437990567177285,
 0.23799661742210687,0.5209679910557058,0.25398864722041836,
 0.9141772366600642,0.6499625949986335,0.23116404101133403,0.7929491250498871,
 0.7495482984214905,0.3860147464987447,0.9638508009467159,0.10458112468032643,
 0.18436031403890807,0.40194929572067006,0.6990503300899168,
 0.8796312458388998,0.29792895009758824,0.9182198175621765,0.8662677407833682,
 0.7332704510126504,0.8333605605055466,0.3331585036607987,0.9435751135743862,
 0.8852750188321801,0.2964678809792507,0.5413922035121174,0.5527122085247329,
 0.8739661410554738,0.07710798714669109,0.5237908227484285,
 0.006046584597218407,0.7988056164701091,0.1427639712825446,
 0.09637324237326106,0.3066933632541392,0.16513587691155562,
 0.7654035779750169,0.18843384435485622,0.166972066207836,0.7074032848740063,
 0.5046341035480707,0.2899484167781846,0.1579181499522715,0.21744482043102642,
 0.42910159137693016,0.9928487371879066,0.20111444036511528,
 0.8971719794399652,0.6675746882923617,0.9423828391188647,0.577250430283712,
 0.04537899902330533,0.5332537413255896,0.15997994586943454,
 0.43549197689647523,0.4586987575600068,0.6265928179033708,
 0.022659976555515438,0.8632406699967528,0.3395513409552857,
 0.47865879540423084,0.3755363733548508,0.8013124686669849,0.9392125474035059,
 0.7160160029402451,0.0994908324406717,0.5286605498027633,
 0.025948880531765695,0.30918710738391153,0.4657116802458393,
 0.3902487081746582,0.3833839323907626,0.971097502393227,0.5038408979358181,
 0.9614369718791951,0.8247179569819872,0.1946059534485044,0.43607670407959276,
 0.33074200882075466,0.5068205391398402,0.7637176633681835,0.4116130756797305,
 0.9283138758428551,0.26124046392369726,0.20052008251487463,
 0.7253760106683306,0.49278202472376365,0.8705460033049374,0.5555640625494027,
 0.686121983910385,0.4249369798113385,0.14128951842734028,0.367744111110046,
 0.7734977535235745,0.11967242929407806,0.5494177239492544,0.4998305581405552,
 0.15644289909102316,0.04025293214765224,0.8089983309063197,0.260257019555329,
 0.735475888460189,0.9492663540701238,0.7409339010544865,0.31479054796790684,
 0.9964074360125161,0.5880097857891126,0.5819970671009728,0.14978281273577632,
 0.4207982707794311,0.05327322381082311,0.06173513437514,0.25467367105925687,
 0.8800586577422586,0.5616175953942923,0.529049890654342,0.9036232989316798,
 0.6575673026560607,0.8433144810676054,0.014225737525453419,
 0.6614637671643868,0.20130214297730287,0.4289970957179309,0.3752786256772439,
 0.07599427913096823,0.4805481294312375,0.47966344725075294,
 0.30399217951095003,0.5053849378559403,0.7762153259795467,0.4169324831566157,
 0.580286437585815,0.267892239455651,0.8406494543209578,0.008965555731203834,
 0.7217616450046233,0.21044540398296907,0.24591690710295244,
 0.1275524195610671,0.007500844705813803,0.5675960359397306,
 0.34359123826506877,0.25222513053557927,0.29346677784122677,
 0.247195902164406,0.3553100890485109,0.7156025820185491,0.07441917211961824,
 0.5463391057361073,0.04218307317419645,0.48734513803829504,
 0.23385460647468626,0.5185183368138522,0.562150341666482,0.6066605654386175,
 0.6336451245278304,0.8402908951380662,0.949019573871176,0.2959890276180789,
 0.14412259808340155,0.3942680307745863,0.6310029528086102,0.7001677422175346,
 0.3788926174014008,0.5178485167062964,0.9062941095112609,0.9349228025097136,
 0.6974926892305515,0.5992349034369153,0.3170930927464999,0.5017813352303506,
 0.12033935382712113,0.7053986204976979,0.8625192900066811,0.4447595169404561,
 0.6725579419348724,0.22609058350055555,0.7814161927611466,0.854168348087973,
 0.34694227178398074,0.6048288884340529,0.6434676844785041,0.5511223714760153,
 0.5158096032174975,0.6760705404394975,0.5676388810986123,0.8340999393810695,
 0.07495097164447706,0.9681323680835928,0.3538218371350541,0.6187007205829889,
 0.30717181706431795,0.004700889814324927,0.18041378966226163,
 0.7234677111263309,0.318952601486881,0.5395485865770642,0.3908203227571728,
 0.21789736833882845,0.6171202884874875,0.3255029898591033,0.3304710054168616,
 0.192375919554979,0.960024899918893,0.5858201747396687,0.6469875487294989,
 0.2005701106763471,0.3193946132476718,0.495340263049447,0.8645312629009683,
 0.2769536674214035,0.39615330154379713,0.9093029574944538,0.3866280050595893,
 0.08540023279862996,0.22910105018682914,0.9511594640836983,
 0.3033977546905855,0.9287214326345472,0.462356525374059,0.6955536710998929,
 0.2764712878620811,0.8089220061884452,0.5913142114852784,0.9767726038281249,
 0.4335520540085118,0.6965544871914411,0.723097361496384,0.4261387669316712,
 0.2003264057597992,0.7860116464464825,0.919652586970976,0.7155209153002495,
 0.6889181121526703,0.71631599477589,0.017821589915741676,0.255972929586864,
 0.49655332645899186,0.9937050713685388,0.9538430322904032,0.6374981373085608,
 0.31732486274681304,0.06901446951886392,0.7592896231795562,
 0.8671584461135389,0.6565051186604426,0.3954891063555809,0.40161546627316014,
 0.874302037720021,0.5647512318604202,0.6505920833498401,0.5035354184327039,
 0.5444128995118509,0.09969321564727107,0.9646839692874436,0.5666654556840639,
 0.39535285281677246,0.5619223079969428,0.24093325735293425,
 0.6044950039949887,0.15274880992671402,0.7953046737656985,0.287159833455255,
 0.9416801094034946,0.23634725229004738,0.3113204632930302,0.8063831434107056,
 0.02961276577632388,0.8052787013244751,0.46439706020149074,
 0.0849243304316536,0.21106984584877264,0.14106795982374576,
 0.46480544820532843,0.9001107623215104,0.817278960224735,0.1975601653457726,
 0.45534888885184666,0.9211957188372386,0.16342173391469228,
 0.038038695445296256,0.5425789062939063,0.36978117139325073,
 0.5662383415958567,0.6863859569274531,0.9532494554348753,0.8479357766312786,
 0.8167388101454445,0.27716810428597505,0.12607709093642538,
 0.8689637001692164,0.610505807868503,0.3435878190072369,0.032608845336718706,
 0.047163876093342716,0.016807156274086243,0.08396235535902108,
 0.47977870932128575,0.012304440978162079,0.5516104349359456,
 0.6255960968150475,0.19187923499077242,0.7260962673992684,0.8419670622118242,
 0.06893623393120307,0.7575729313918892,0.9140592820525103,
 0.19760130953684285,0.763988062169161,0.9135187939136501,0.1014793299986525,
 0.26272975234755724,0.672402559110292,0.4807528624356281,0.7332511071905351,
 0.9489046755545251,0.656753936801413,0.7700085105595309,0.7156619439933624,
 0.3993936204684877,0.3488527816946061,0.8668167836856605,0.9834177038669045,
 0.08259814227351292,0.35177378335916376,0.8811443625445172,
 0.8385194837043967,0.16453740849711485,0.5462978987194793,0.1480168036862738,
 0.6124837496422173,0.17790062499515658,0.7989582046988012,
 0.09445852269933963,0.7427200969841911,0.9796725354376163,0.8084029545442047,
 0.31404489132886926,0.942662597079983,0.3205912130540566,0.6735460582643534,
 0.8259947295455766,0.2844733092446279,0.7087604318998932,0.4692017833919355,
 0.9595670097147391,0.3349289504412847,0.8872019646823257,0.875903340359912,
 0.9014359257733644,0.8162997214986637,0.7041163670602729,0.7841820512480902,
 0.7843216589645895,0.39702627223463216,0.9592747456458749,0.5917466764426843,
 0.08130593985038348,0.9592235112979527,0.7533884681056974,
 0.22530365363562033,0.07766596517008961,0.3490148434891409,
 0.6760877262163043,0.31480430516605196,0.762190266001187,0.39423113883930627,
 0.19681254404889337,0.6598470863687143,0.05696284877911706,
 0.8856914044399583,0.48922695630440705,0.023515436392786304,
 0.9343987762225007,0.25776589997073796,0.4826018157435301,0.9421627145751751,
 0.37877217762047544,0.5054056276130802,0.6522423324460085,0.7551906239290924,
 0.39692527403131594,0.0627859581941782,0.7845226981100076,0.8370934279780349,
 0.27970196110003853,0.6879128772585594,0.3437971580681106,
 0.24001685478947665,0.8481534579411554,0.33586674216456824,
 0.8936924915222889,0.801878767896282,0.5996392797511204,0.7706680356862752,
 0.43523856166885755,0.40545008340519817,0.29415020806089953,
 0.8371424072797323,0.3624255539740815,0.16487298984977306,0.4516946802807903,
 0.8411413049917801,0.5732521136887471,0.9383564319400961,0.11928498043304248,
 0.23519485057848732,0.8654260680316614,0.3922032288601327,0.3673072033468602,
 0.780075564828631,0.8270310632860709,0.08790196597480437,0.6714944167447033,
 0.518986673428945,0.7083697387815814,0.04635701842500928,0.8279263582331562,
 0.746536315209169,0.6822223520526942,0.997673110884531,0.6188182657800324,
 0.5092384392401592,0.12836530900429655,0.03722125184864633,
 0.9176695574978768,0.5003995114800494,0.5520633881641922,0.26427768057682455,
 0.17783505402432098,0.38702704739542426,0.2548819809740904,
 0.6676579799129038,0.3666294663655567,0.817768454662309,0.25980768596073045,
 0.9980913062157395,0.4990223176510362,0.02524796640508109,0.8569865861062367,
 0.22929345123859413,0.7870509760577072,0.05929177934426466,
 0.21458914771051563,0.7494105423179204,0.40638291385410386,0.756556102772104,
 0.6711819203988428,0.9134175462917957,0.773154588023867,0.5905710526059341,
 0.13626837795769142,0.04766010807484766,0.8154345410215045,
 0.9050868779240463,0.2379850099855232,0.21636952498442064,
 0.18261066290419392,0.525701468589603,0.2987419649240308,0.27112482816294836,
 0.7432416490421948,0.11404233011286624,0.7237738217589598,0.7830977479329057,
 0.746976221528765,0.12726199733787324,0.23143238110271414,0.836514837077101,
 0.21921451649292079,0.8024466989836584,0.2725295120050094,
 0.22761828727117983,0.445675822465046,0.01505026022091771,
 0.23384619185893563,0.19328399365643711,0.6446345253673831,
 0.9840651024335051,0.9365975712395535,0.722462949553201,0.8602563800320105,
 0.867121347853351,0.7962361496840134,0.7193026337916483,0.7426794469286874,
 0.3905943903211774,0.2880254629548129,0.038496167468126674,
 0.6550503314214677,0.7244344382645451,0.032328651331708924,
 0.40011967680704774,0.7881947897271405,0.7487667340474783,
 0.16994617947675483,0.2691760578985012,0.778593080423537,0.5914118591110432,
 0.5005194273865738,0.01586544169760773,0.0524801371918151,
 0.12699118079497995,0.19147721955199992,0.6356809867171808,
 0.8688079646593538,0.49263798569659656,0.3797203069665792,0.9658373004575025,
 0.8731670232116503,0.7570637081243661,0.13494341826866185,0.2029440458387899,
 0.7585925168605504,0.6085384998622337,0.9644431573617475,0.7007692709106897,
 0.6879894261201407,0.13696555903339624,0.8300724629505828,
 0.46861566844281644,0.17585238924453106,0.060604669286249746,
 0.45037120986707924,0.1719596376802024,0.6476033891712568,0.4969521073582097,
 0.8177685615552156,0.394040597153578,0.16102153981873402,0.4250495250160121,
 0.1271410506442192,0.9106159041324781,0.8515326541781971,0.787937753225284,
 0.882947756794605,0.6114520413572601,0.9082409582648797,0.004067200293451201,
 0.05030621613457287,0.4508795413028732,0.6800220805903981,0.7674406093361786,
 0.3057478582469516,0.15144133980027885,0.3868364264401436,0.2844518466586373,
 0.3184454804461947,0.3304624234016674,0.42232161786353917,0.4913198393991096,
 0.6648354731084936,0.6544181370513447,0.1393242011000746,0.7999419219719812,
 0.9118919521729264,0.6562991719857725,0.7704795040717601,0.4262009069857229,
 0.38659600184676157,0.8389989870714452,0.9134227515045744,
 0.12469604787851851,0.23645832212667994,0.33249760992993604,
 0.17092366278586124,0.15851533937004714,0.17418537516797006,
 0.09503250437379773,0.6145908707430423,0.7520281488331201,0.858597219443795,
 0.7794439340168531,0.29006707586972547,0.8028341785475605,
 0.006721449088188058,0.3019355075825032,0.03479997161199522,
 0.6043031052422658,0.6659089944529814,0.4778089264403822,0.589276235935312,
 0.5319570897144613,0.4454821333072101,0.9859328287946281,0.2674088787131538,
 0.8246906337564387,0.7280799534455853,0.1065852933113115,0.7874859355968971,
 0.268867925476336,0.2070734736255163,0.9112846323179289,0.09507588183424176,
 0.5468312544126712,0.3320260712985079,0.023628793669425052,0.216316122455678,
 0.5602157423553401,0.6806809900502161,0.9676625521096759,0.21266476505647325,
 0.6087115066293425,0.5376068487037172,0.7422242557841846,0.4600318519375458,
 0.23039531443263517,0.909638753889194,0.1307126882385118,0.2268834548380757,
 0.4502544915450619,0.7592248036745598,0.7636149303519412,0.9907900474272044,
 0.5128523705793653,0.4429453044648918,0.7500833615568254,0.15414446548395466,
 0.5272093236495798,0.3181743249452067,0.0794314811494422,0.18302731662642346,
 0.19768822260860097,0.3423602840161656,0.47921991575049994,
 0.18294321445515926,0.981894894636674,0.14815873003101798,
 0.35369346294586046,0.12445782841614061,0.7581358220498201,
 0.9439393956769877,0.5912339360394672,0.2455460283324895,0.7159129459422975,
 0.03457783705851192,0.5151728827180584,0.45757251847110225,
 0.3844800673482687,0.9892728696208695,0.5430401618716368,0.6084463934118834,
 0.3144468302147365,0.4361416900142854,0.8787097100111231,0.7737395073181705,
 0.7215053021679705,0.12935773259424665,0.034102032025364104,
 0.34115686423911784,0.13562113796769348,0.8604473044872826,
 0.25516830236790333,0.35277149955455955,0.8241349650516115,
 0.5879968951250308,0.7376787602825936,0.6886599843006637,0.33959919995989174,
 0.20775236173279277,0.6569958018834504,0.9329299536257181,0.6933990085103938,
 0.7463466670769698,0.4108233364988457,0.22441852266181328,0.9409589350499625,
 0.48105281542445233,0.5555035941933774,0.5908822099670101,
 0.32485908939113484,0.7964826598881674,0.11585886177217608,
 0.9808327911738642,0.8684058586489464,0.6765910217579839,0.8896216296639237,
 0.4221646540082934,0.32429682448546726,0.4401324304388501,
 0.10468871393797174,0.768207683373475,0.8848517287760342,0.3211862509807293,
 0.5202263033695869,0.9308469971271367,0.3251134840588774,0.8530814056873105,
 0.2178750130161311,0.5977805461465385,0.5174615866346916,0.9044122417477987,
 0.6548644266158251,0.5704690163183505,0.03177161066209322,0.3243721783644309,
 0.38424849670947536,0.6710400507580077,0.3810829252749067,
 0.07635820889595524,0.9589437931035822,0.22114401553134233,
 0.5517999682280081,0.7798287192978614,0.10967667435518491,0.5528791298214683,
 0.3032367572422018,0.059757538774728114,0.5489945853323727,
 0.8898666733222657,0.3870356765728171,0.07260012160091789,0.9074929639272268,
 0.9379694598405403,0.29418329196850956,0.24420140154522474,
 0.4985197009947042,0.8457576607983633,0.6237269784363979,0.19439881773847945,
 0.2588649816371623,0.321901532093402,0.3501037883986118,0.3258423442960532,
 0.7098871314076581,0.3223681533030248,0.2957606921729391,0.7093652333874024,
 0.7962057581505863,0.2213775519547052,0.5602128709797065,0.2146486459400625,
 0.02734236741951246,0.1815254852031325,0.8233504508274464,
 0.20576619237006033,0.19917689565490693,0.6251827842357635,
 0.8524872747190716,0.7678727506197661,0.7163832271870366,0.7484106511586487,
 0.811307780033888,0.3779778948039899,0.01800877885047414,0.32627093210529656,
 0.8413280200184544,0.5469518224633877,0.8021360426570293,0.6397492957620157,
 0.41241469341057435,0.7705307404564519,0.10468525729270095,
 0.42994213029901784,0.5068497346711134,0.4376180266258889,0.7273765609506445,
 0.5897088593326406,0.5305760749054949,0.06911696247085397,0.5674184388663998,
 0.20747719370514595,0.6819789402247863,0.48527540529668767,
 0.13343963041191242,0.14712976998853478,0.7565147834886351,
 0.13174085583987039,0.07967872410272081,0.10509831336723308,
 0.2397940382658834,0.8355603963409846,0.045092194680003694,
 0.5734114096401806,0.1498895820687175,0.6064449315048356,0.8082716628646517,
 0.5223215340469957,0.27124866848315266,0.01920139023034584,
 0.8841385329451223,0.39552694495210106,0.9388479660448534,
 0.011227934042818632,0.07889707375409549,0.15960509993248861,
 0.1916308302273071,0.3509544119933945,0.052532752450462294,
 0.6625632743359162,0.24788859847327271,0.006515221281540695,
 0.9083818637379262,0.016517675243780783,0.9609551125779967,
 0.2261744471989252,0.3100841987178594,0.8809743815846983,0.8205146875311365,
 0.6060851417266182,0.5682492262276657,0.6582259567968793,0.8765894494997761,
 0.5396174596628425,0.6342818427593229,0.36083599376648345,0.8312424950511205,
 0.6212558557257977,0.2294356049474987,0.0381132884248232,
 0.013333569377086074,0.5825696905037415,0.2757187469568404,
 0.6708524077720796,0.17949219420098006,0.13378155011300275,
 0.9458930204932706,0.20712876974232997,0.22843941408268909,
 6.071985208009334e-4,0.7284111378681567,0.1326019921901358,
 0.09842218188129237,0.8133210692336669,0.9970882287695516,0.9400639917263494,
 0.522202067250832,0.0429438417026895,0.9973181581577959,0.3935515285154043,
 0.575707873237887,0.16747220020304976,0.4141856732230057,0.06693323397120543,
 0.8921787145885838,0.1633265880949588,0.13736582403940156,
 0.19608163300651715,0.6722581561189991,0.045589338796588774,
 0.9415875943400741,0.04188361125478657,0.8898049047288583,0.8848907971198303,
 0.7090933848780743,0.7024376541696052,0.8105958976934562,0.5872453255471842,
 0.545332317040468,0.8522840493016768,0.8605572818924437,0.5177361959104981,
 0.824705154838199,0.22508445570505997,0.3339636293728028,0.850068733713199,
 0.6546886953605433,0.32380054277217296,0.7573654416684114,
 0.18090295461871353,0.9226038868204384,0.0581810695941446,0.8727202953792033,
 0.3069425282562682,0.500980171037678,0.507899119653126,0.16305373874922913,
 0.3931447289432406,0.32661718835810993,0.9849796638252433,0.6530440016821976,
 0.6361981099962956,0.4547876126635404,0.044733730153632134,
 0.7933503308595558,0.44164610757593403,0.505193011012832,0.14353747398681505,
 0.1451936909125533,0.6517180108961584,0.296162732499747,0.18192554509396852,
 0.10020173094320539,0.6481485100324375,0.1753412601297446,0.893607664540408,
 0.7773479072078143,0.439621580441126,0.8517887240446707,0.11059423122708023,
 0.2928296915076761,0.24892446557218073,0.49852822520816464,0.34560989413862,
 0.3425337873512082,0.15972368357585087,0.6876525539526165,
 0.01807713052351889,0.5618858078314457,0.3003278393239005,
 0.27172174316505404,0.164154317345246,0.2136188943987659,0.7199776918405396,
 0.5782617383748236,0.4092888042016438,0.5674401981870877,
 0.0050296995801069055,0.0017025691093943696,0.9694609510550253,
 0.2735053730148914,0.06801620949993326,0.19350890002455656,
 0.0387638245075792,0.2555668877848605,0.7437052459660496,0.14773237082133595,
 0.12116329752118948,0.03875097647580583,0.8825080089761725,
 0.4537907422616202,0.12086063828886118,0.9202120841588911,
 0.19443405369342082,0.3735133769864196,0.9717211531330769,0.4259991462887581,
 0.7340625453659961,0.3153681899632512,0.9227902346206157,0.6362952204717542,
 0.009399564081263967,0.8138478220909728,0.3640112027516508,
 0.9474634757912851,0.18175259326265658,0.33410845765137676,
 0.26391930711797895,0.49818596391469494,0.44112572655906535,
 0.7048619869968342,0.9486570495049641,0.5901659703940825,0.6966246634782006,
 0.7395202086256886,0.2966179286010465,0.7031501436719629,0.11692119697776204,
 0.643679202775817,0.5583935233117602,0.6023354491203055,0.9819252894583244,
 0.3314975813969949,0.8600754211998267,0.8348089663072267,0.763519163139184,
 0.05126158853028484,0.5832844627101855,0.20184300770621866,
 0.38856002833536996,0.32706313305546464,0.23266938049969765,
 0.8627609278166357,0.9460976363801566,0.9827006520837559,0.5451059786847308,
 0.4410626128126782,0.7309881620563612,0.9735335551018076,0.96339910587415,
 0.1823573109071287,0.7295208925885,0.7921598028953477,0.6041423061895234,
 0.21103507975199975,0.8873580846009794,0.8298891648994878,
 0.28289646218780906,0.14586688753282417,0.917120298510728,0.5509317951773004,
 0.14248258824472526,0.3523367638582803,0.9525445232364183,0.9121879978719833,
 0.03893479235142239,0.26496258186654864,0.2810183025214399,
 0.2349759569148111,0.2895098884480116,0.12132484210734251,0.9944583821969034,
 0.8740663130226881,0.966977786250632,0.5055081052749368,0.44710887741135874,
 0.4981793001702922,0.07611811125857904,0.6320843530748823,0.6319527871479946,
 0.11955936066247219,0.21157212738070907,0.53014859778933,0.9143264312408637,
 0.05981254267024727,0.9105610961783269,0.9779089752507202,0.1520438089772539,
 0.3647337673830984,0.6020002208677441,0.6368766955112768,0.870402052481516,
 0.9505290899571198,0.5279928210328274,0.548821942758008,0.6658135383012866,
 0.9133377886485663,0.318808815358966,0.3076782114425005,0.31215708102235606,
 0.6617099141803992,0.5801430221163363,0.27743495272563923,0.4769333218837214,
 0.9690373480772384,0.3535136721508173,0.4628165912064899,0.5503896351766984,
 0.004267115579124847,0.6666939765320739,0.694431908518464,0.5771886283563371,
 0.5172133783193724,0.19525205250812294,0.1144829223707825,0.3511297682333473,
 0.6118779606535585,0.6588477549451444,0.8925122229209185,0.7013891705652069,
 0.9972631760584115,0.46184548435135886,0.22348890137977184,
 0.0073352764016835215,0.07341236423171271,0.12841971821220988,
 0.3192318406362703,0.3646387857860307,0.09243045016038498,0.9275984929286903,
 0.780867471500055,0.1027725747137217,0.2035552242219978,0.19705834597187444,
 0.7894964572343834,0.04896379985503452,0.6887228001339679,
 0.08290205786632321,0.12474368637386257,0.3470829829649311,
 0.43351318983969445,0.010293285467406488,0.631012009915663,
 0.6327385049838682,0.3387470773710082,0.8728925696786413,0.26160279768753036,
 0.0683212584376629,0.06806628210958121,0.13386190457530311,
 0.7965467285549481,0.7220172979660521,0.40383735329190884,
 0.061775932174866655,0.7747892527626715,0.14883667963764857,
 0.03950023313716222,0.8960805529269792,0.5221683254584117,0.5677648353182198,
 0.19868432147203008,0.9210387176702685,0.3970763624038366,
 0.23597303048450435,0.14982867459145677,0.5813107164488092,
 0.8187360088295809,0.3671164147485304,0.30008648568082863,0.695906262694008,
 0.8339055608513702,0.8837553542526418,0.22299262452414537,0.8815305580963929,
 0.5289459043491117,0.40652813845131097,0.6631932605391907,0.2687944725969378,
 0.28894847160057036,0.7770339371446922,0.6859745442242898,
 0.35098176001921877,0.1331571279049476,0.42053982096791354,
 0.20282865202082034,0.6739931900952898,0.33894181917836086,
 0.9645687242031106,0.680198511136093,0.3311316182063474,0.32206568774813826,
 0.31177896089328105,0.8571406502597911,0.1561824959375434,0.8709843680685236,
 0.5920462799635499,0.7565970527458935,0.8521070603062801,0.4978739186523341,
 0.31617482810086883,0.2144673320268652,0.7858077676168643,
 0.26233673243261146,0.004759525745647641,0.027521030445457617,
 0.8684696636345964,0.9837050955374296,0.4634571935409587,0.9086736933398746,
 0.5814534080205535,0.4441099065973826,0.40886175930958024,0.8548862075370339,
 0.7133911976843307,0.9559666162801614,0.42158555811642173,0.9685827180327812,
 0.9033666847574683,0.3464448095282512,0.745040590639813,0.7607031443784742,
 0.6104895304109014,0.8224749443528344,0.6295076381598916,0.6376569214057968,
 0.5936149334264713,0.01723282172571272,0.2572522816274314,
 0.014016879044178698,0.47396570589965303,0.5091523442844417,
 0.36316058048031286,0.705023444919124,0.8393419836111989,0.8244055608470955,
 0.6972040215093399,0.8502931180840118,0.6835435458118914,0.6306960754695816,
 0.15451743036809695,0.6003574436528041,0.3396058704872812,0.6749643512985334,
 0.9499771083649359,0.1490706415655476,0.03194421398785674,0.2830529827333871,
 0.7838416148735792,0.09222137152557908,0.7518132322290895,0.9295255544772808,
 0.013646071046655384,0.48767300817185166,0.5585053190480885,
 0.8063014867984413,0.9399705165641856,0.5236554712131585,0.5973332331155379,
 0.8288271490969282,0.914279239439395,0.7447030302224985,0.7276512577360984,
 0.4657411197796564,0.7366199889798231,0.7832843274337952,0.8734864214004691,
 0.08139748567201166,0.6511557739058066,0.09335065292475142,
 0.9862496496385167,0.584458322269153,0.31118406922119757,0.5550970730207467,
 0.026088734008514858,0.08765855919603105,0.5558260921151136,
 0.0720329086683309,0.04660473768107054,0.6991028063482773,0.6265188444806159,
 0.7960884342215363,0.042298553985557485,0.8017708608031582,
 0.8761187192762101,0.5251529790836587,0.5733291159644521,0.9786966854459689,
 0.028848587335358777,0.9842274637950994,0.6185579563058997,
 0.1404522841639141,0.5513413072607769,0.4149843050803662,0.3151310890853064,
 0.6388673035301954,0.17137284103291983,0.45889767024682115,
 0.6493579320613109,0.6505934010358094,0.06963555507156538,
 0.42570279508813635,0.36171443895848754,0.608930732902992,0.790375371177034,
 0.685968209583573,0.11505926949253353,0.17904111120504995,0.7603812811159483,
 0.5584355330511266,0.24292360906868393,0.3393184268971068,0.8521112322593698,
 0.15231985533495374,0.2886404483880378,0.8379380089256039,
 0.38820482623579233,0.1572777849834629,0.8245161386667741,
 0.43618303316117735,0.5313042469311453,0.673191541411354,0.25014342297896863,
 0.621429449312036,0.8344779946246882,0.29926126432269884,0.7685518671024519,
 0.376690215625884,0.18743993696848404,0.7346625996939472,0.5409015846234131,
 0.4065818693002462,0.21109853365155096,0.20971216228643286,
 0.13073834915869353,0.4844613865655002,0.537475069171491,0.6366871289953834,
 0.6588234191002196,0.9413716927664706,0.2612945427570289,0.5167790504951978,
 0.8988441980762096,0.9572302101620854,0.9028968679014217,
 0.020601285920608392,0.42380187229291577,0.9324448972696305,
 0.7817709118947243,0.10055703484253409,0.08465114900282456,
 0.8353706689635915,0.24903388066406873,0.36313160170546044,
 0.6547647763515049,0.9806317660598961,0.3438912792290253,0.8330080941468725,
 0.8050609309968361,0.4087271511530898,0.8635846528810569,0.20810624048504223,
 0.40100689284478175,0.9860915365271001,0.5910621980945109,0.8363499665762771,
 0.7912636644866881,0.44287700307038924,0.2981537344966425,0.6260718805463229,
 0.11223415807700843,0.6481368563474685,0.28966087032133625,0.554822313686606,
 0.1687135727126381,0.5188989060795948,0.3914534370295316,0.9434649874127904,
 0.7071334431304527,0.19222426465979536,0.17934427885898307,0.389877510541902,
 0.9814944059681018,0.4428841710210636,0.8465348913885806,0.4275043944271236,
 0.683895906571007,0.22566513136357336,0.7258945833087703,0.05734682676802705,
 0.5948887213156606,0.7726373134616704,0.34180196040748456,0.4715837052665741,
 0.3115362927089651,0.9688745429058399,0.7076681055944269,0.402722871880683,
 0.6150060424657335,0.7917469267894864,0.4771191893061051,0.34755868803969203,
 0.2698936044953355,0.4589845524855982,0.037851338216509434,
 0.45788702840176587,0.3287553512281802,0.9930153950124262,
 0.49990857634210006,0.27376132471763515,0.9473343036139871,
 0.6926659894014746,0.7194367682480983,0.7179654540642452,0.06035301121382297,
 0.10899559426755134,0.6643587642024422,0.9736369282417296,
 0.10908647914202518,0.8089846319908723,0.037875240871483085,
 0.03464361667744975,0.7686584168252497,0.7998788156068829,0.2422093289840478,
 0.25893084976390524,0.7856728672904073,0.3087145181357924,0.3370868973012152,
 0.2842206827300011,0.0023029304302920606,0.5079581410795274,
 0.4059063328561089,2.6835307672778974e-4,0.33010495881550617,
 0.17714555306691848,0.8351962235746181,0.07171464443604769,
 0.6740199523946709,0.03013657781789203,0.48841982336708334,
 0.45222829818826427,0.3277436380278379,0.09103645232103919,
 0.4535338940810607,0.6784992868922806,0.45620513895213033,
 0.04955702818391505,0.31041329854798805,0.1918607546486194,
 0.3940398846594747,0.4857270138779679,0.045297813203956094,
 0.37676825534680225,0.9550617133958282,0.19370957966930913,
 0.42526178652623525,0.27101883025349605,0.79868169221111,0.04460300255745919,
 0.034017863568547635,0.3532324779997539,0.30846327864488665,
 0.21605284009909909,0.7163661428501874,0.6940496398443772,0.6904934824838269,
 0.8727521314106759,0.7800573641507318,0.8930280900373784,0.4235039138689395,
 0.3020046161450598,0.8267258258941056,0.14078749240041444,0.616982090864029,
 0.022806725786104765,0.25705360325682225,0.5223767459853165,
 0.27713360060564096,0.4447068951926042,0.4078708419253022,0.8557917957054499,
 0.612787894969008,0.5267362543456457,0.538147432784898,0.967863971636869,
 0.7605965712536553,0.9898130174513668,0.9225867258642327,0.3902906372967301,
 0.36297099124028964,0.7781158484638943,0.7158558852017344,0.8371032272292841,
 0.5016695834574636,0.6285880837753578,0.7241983325687236,0.23151247428260424,
 0.5290649014439625,0.5631509893943678,0.48050086257458735,0.5766668816678141,
 0.42218395000699993,0.8710970926996623,0.11990624888294432,
 0.5063928371893361,0.9207422088135713,0.6212573969307116,0.6134757257211019,
 0.8600643704404974,0.716296149061702,0.25062606726546466,0.6044876328689861,
 0.0574699575615627,0.40529511030142373,0.4391800771686043,0.7347482758995388,
 0.07082096952234129,0.04642843429139554,0.8442574600431516,
 0.5311225553140225,0.1466516726540773,0.40730390482157963,0.3594199586487963,
 0.6705031173833603,0.9097656874168246,0.5339095134620824,0.9078700655637102,
 0.1522200441172471,0.1123817894205199,0.515277573752197,0.8573810558566225,
 0.3230157343725597,0.7502039541092529,0.8672027465008084,0.3642031832322211,
 0.045943125862447975,0.5192705763417101,0.4360529159305737,
 0.6549179778562185,0.09367081729821125,0.6678068143832077,0.5395932541469914,
 0.29623381778615043,0.009168616331853063,0.793538274295591,
 0.06828067808301697,0.5832461554637055,0.8685545435414629,
 0.30825797367547125,0.39258506809100446,0.465846340449671,0.6060678960119332,
 0.31150877098095087,0.9806055509521513,0.053308681002301905,
 0.007766760401352046,0.052427649952935695,0.2638391512561361,
 0.4792936082541558,0.02763729148949068,0.7890548638611392,0.4034667719771309,
 0.7997298074314249,0.5559470210129522,0.2665954770845309,0.4611008892341334,
 0.8567788938841303,0.392063026564588,0.10295670714544336,0.41258869110334573,
 0.5571046326312439,0.5480669684313031,0.6003747243387907,0.7355719819277644,
 0.4001410940530392,0.6629181094543606,0.47178367405034627,0.1036783683726783,
 0.0716687153110307,0.7556485667888875,0.014934335421479261,
 0.20447395249872757,0.15561500209170342,0.07892977231287668,
 0.6253686215853703,0.5290337246008994,0.33331957586936867,
 0.29260459905454217,0.5810985726555101,0.27059979495882536,
 0.5590600496704211,0.7113801068945051,0.33970018585467343,0.5924920771192157,
 0.3313693913200675,0.3198761814883655,0.8360729717688323,0.8972693377835963,
 0.37604458161171606,0.7126010441263619,0.7739217395441079,
 0.017508335189974744,0.335298964172993,0.6367518767745057,0.6604081637581993,
 0.2566256191879681,0.17493200861566638,0.4259719374066794,
 0.24341620333829772,0.26973771906251365,0.2509171812720221,
 0.9139201163422769,0.6892788980051425,0.9150668424432372,0.05895400119607297,
 0.14351207090859242,0.5076440311619052,0.5439657174172955,0.2074549093460356,
 0.9734419915604388,0.021266525658741653,0.8792479297677092,
 0.060534061739271894,0.8847393966229278,0.30980482740868787,
 0.03028716935072151,0.7624666089764024,0.27131050752869534,
 0.09512888040314538,0.6581580572003991,0.8872787815533625,0.8260580330044783,
 0.7611783677295386,0.8427424372571324,0.6963253753744046,0.3307052309311872,
 0.8357210158559708,0.9189759945155874,0.6383531756103129,0.04451617492386517,
 0.6161997258277986,0.9694161821108346,0.8959081327901964,0.927706156364971,
 0.6135736161284582,0.4604077920778037,0.6701848144056118,0.8665528081461915,
 0.7962122868430337,0.17081446437761372,0.892613446127983,0.20346367659571163,
 0.5979333049023433,0.6081558703205975,0.6099654346701694,0.9074427533562716,
 0.029735997139267223,0.06635437345208817,0.37979888653561433,
 0.3905537103219008,0.4196955346204504,0.4603621875695666,0.20979529377183703,
 0.84521876404382,0.40213343793415546,0.07027544619452897,0.8363884771276792,
 0.575618403559406,0.7531354895031511,0.04937895952038007,0.34368724190905264,
 0.48027042323249425,0.5475345196153667,0.34397989115083716,
 0.20599257617407885,0.174059318495845,0.7901266490579905,0.5081696071934492,
 0.14810844773431975,0.04787001355293563,0.26423503465813525,
 0.17283394164502153,0.1766355815501227,0.1739049438260184,0.8822666197240193,
 0.5268447675482375,0.09221057795609722,0.5521901279334243,0.9327080477232885,
 0.13323844498081538,0.048466654162677925,0.6228491550894595,
 0.7122972533612072,0.9190727821167207,0.7707204059322571,0.8842728270225142,
 0.7540647523452173,0.9854554090699954,0.03620571904028691,0.9501957065519715,
 0.08647674344224976,0.4640219471086444,0.9440710003566544,0.8638332406435955,
 0.6686160264252288,0.2663400312307562,0.3833244071013613,0.5126797440917712,
 0.5196959186618174,0.6795366856073153,0.1555441527452105,0.5773290303043023,
 0.38039376702490457,0.4669058725954205,0.12068945950114429,
 0.10175503959027832,0.498115861532779,0.6981389165747105,0.22992303652433865,
 0.19656896935891477,0.8878262883354145,0.09247848990819385,
 0.7743133880666068,0.3838020690988768,0.6364192668005039,0.731263341222854,
 0.5678113005218798,0.7895035305440381,0.8353745939767971,0.8222654539877736,
 0.7773715506598716,0.49415121627093817,0.2506850563035785,
 0.021513562497780914,0.48763101665454034,0.026639336763603305,
 0.41379990069755423,0.8422022497653342,0.7200513253052829,
 0.43449725758768887,0.8424119680043136,0.4267071553989329,0.7027381993697309,
 0.3874138839172918,0.5039223858093537,0.6130375125535324,0.12465480198530265,
 0.13095801277764685,0.0932687838925339,0.4417109727600732,0.3967922472901766,
 0.09990259970061788,0.808688881740403,0.36784717989818516,0.7603108468855528,
 0.7255448536500907,0.17646206373088646,0.7628225109756303,0.410369832887554,
 0.07934823913705813,0.1988355725931743,0.4047932348857437,
 0.48104409090220346,0.4194899102597195,0.6910376316538336,0.1550249173285677,
 0.09095324231428825,0.8613264076857579,0.114436558556688,0.3237904201442783,
 0.7914453544728481,0.43899330449285845,0.4163606262733248,0.9826429425238886,
 0.5844145414411523,0.10864896072704233,0.16768995457769587,
 0.7669414526883636,0.9465060809350657,0.40968095889365175,0.6356874933746914,
 0.9472062512368989,0.6029487929355933,0.6372774030353908,0.07738851710287709,
 0.3640464596324705,0.6670054589897306,0.8656088298319466,0.3360189890821994,
 0.27073186925021375,0.8993294602566746,0.7797301343145312,
 0.49611688043702873,0.6032317089984045,0.9488098456717391,
 0.22393559756895898,0.15000483624009509,0.5556387185455347,
 0.15571869306398856,0.36742805844760107,0.9656229776642143,
 0.8057404541684373,0.8251971069122904,0.46297991886782996,0.7391355562688001,
 0.9660462438525601,0.46227667410590245,0.729222274980282,0.48651161043067437,
 0.40920846058368765,0.6166423302984819,0.6588963921076386,0.7047390373589565,
 0.6096426785242239,0.10780660606681614,0.7937474580475478,0.5874695281934821,
 0.9060104860214468,0.7674060964218181,0.3141256637793127,0.49745260452880236,
 0.02237663793354694,0.45033282024630017,0.6709443266246604,
 0.23311672090940405,0.3615862439336478,0.5766042672597499,0.5400259651190726,
 0.6532095953184336,0.43627017206801544,0.07137074750033967,
 0.6911820042874118,0.4204691283446903,0.8366617580125512,0.8446905578700244,
 0.7825946579913287,0.14559202978344155,0.8709802177502898,0.5650921186824848,
 0.7829243657749152,0.6781213563335059,0.9800284819250408,0.16070578638106325,
 0.7471558853192771,0.4601160785304108,0.10429807561759263,
 0.020557348317484347,0.33234070349826583,0.3140297058826851,
 0.7298928527848804,0.38105875811989454,0.44112398727208035,
 0.9340517853394709,0.862002531479282,0.056243013943949705,0.6138884853790207,
 0.3998107807094,0.14222115420777526,0.7416115661191784,0.9095769938369617,
 0.589375814765892,0.550824191837475,0.6520932473528677,0.8258025826662705,
 0.8740286625687286,0.11261945097139991,0.8576166706686674,
 0.22042425136541377,0.12124355746728999,0.8574609079803882,
 0.5112252335535246,0.3230858543443017,0.31821701836553584,0.6888506220624007,
 0.44549296731472154,0.10346162664541525,0.9448115017201445,
 0.6909658956014145,0.3801001306869789,0.2812887810409954,0.4610884817320653,
 0.8326712422364679,0.7037092522448626,0.2353518214093797,0.4705616858582826,
 0.11093937906296725,0.5276749570920676,0.9982650850687842,0.8305018709132082,
 0.6593115242009288,0.24189332934533636,0.16925956370333317,
 0.9815655149276401,0.39368738592312624,0.6019910282486354,
 0.21832923809894256,0.35850846300339023,0.47932937483479043,
 0.6905047381940594,0.8532780781538686,0.9507023950563283,0.5640224138864419,
 0.8375362814956494,0.024875925909708885,0.8748446273104248,
 0.5204366692474851,0.8538090631785024,0.6263544669401172,0.20322461508995615,
 0.6177538531817319,0.3330609140895403,0.2769334924827027,0.30966738697671914,
 0.7693610858527686,0.3395634515201339,0.2059470794521695,0.6345825880647342,
 0.03332618399864762,0.6320648961779111,0.03125766355779103,
 0.49579278593561926,0.34411261446018326,0.13890135588814967,
 0.26164974584100287,0.28944638505833087,0.3310158855159848,
 0.5325128053906912,0.002052655943168502,0.8714792934791553,
 0.4139939599911564,0.17472337179392283,0.9435327867732163,0.3439335423458132,
 0.7394299334191029,0.40489671324693366,0.5718072273406806,
 0.35114321620084543,0.9234449438955976,0.1454856030755738,
 0.40970858018135736,0.32308683900867086,0.24160087078879033,
 0.11120700017210217,0.4079041927076904,0.14971261502847755,0.746502626318875,
 0.1670397605271945,0.07342787963482078,0.8703782380807902,0.4623770486803208,
 0.5714153249571041,0.8987930451290568,0.07125096587002322,0.9166592230218097,
 0.8317713650594016,0.6705885433270858,0.9124589526902835,0.07047004357202802,
 0.846594679631116,0.6837760190290955,0.43334541695425377,0.47161747257967523,
 0.8378146574617834,0.8640230718839439,0.5316388552302352,
 0.061359487371845156,0.018004837869943557,0.5078346448304025,
 0.4048780544339118,0.7256765805016977,0.848800762237493,0.201715792087933,
 0.2922045790070016,0.4477989385751584,0.6617900632631357,0.20693452231941856,
 0.594479356390716,0.02913048645336047,0.11451411092344521,
 0.40655068102763536,0.5568000091710212,0.6055052921772043,
 0.29972305585030246,0.01228861341440668,0.05497465421001746,
 0.8023679186669888,0.25954406914606254,0.6388276675404522,0.7495986624094644,
 0.5350697888836784,0.2876565724985697,0.9164441620910077,0.8549308400767773,
 0.6320886478781528,0.12105583720908197,0.41256188006940775,
 0.2947563007164663,0.25560267429731887,0.5534109826721657,0.4927296824015954,
 0.3651977479736118,0.7480702353073179,0.26439495470623564,0.3850036872280468,
 0.2567229392028417,0.6009219981654601,0.6268346541767132,0.5856509388220936,
 0.14176267759920236,0.7563356059656504,0.4929299766103745,0.828001885314724,
 0.3767234201272339,0.16228179414282518,0.5894202508670745,0.857058388782202,
 0.5899907649832297,0.828701326981773,0.5222311401792924,0.31360505024773655,
 0.7690709959333021,0.9472458285500219,0.6680141422265891,0.3817784944276946,
 0.6971457712894292,0.9261240354907644,0.31486040240480095,
 0.10310893325043469,0.33140742077746344,0.20948659362218314,
 0.6355290884790336,0.7256762639146251,0.9529184825350421,0.7863775026602589,
 0.8476413808640184,0.46164217788747175,0.786112275034153,0.9696763756681512,
 0.3067010727530236,0.2611255606494829,0.5934230894124614,0.6169323910313228,
 0.2571282279384739,0.05304365366269881,0.928734404052072,0.8545869503098861,
 0.019580391205592185,0.7174605705871198,0.4573246430218739,
 0.5914113757911147,0.1674427702070881,2.987174733504183e-4,
 0.8822607601175274,0.003676152636173713,0.9926288855139922,
 0.7546327636238903,0.06865076180453455,0.2966248993469963,
 0.45570116765160296,0.08027998148217563,0.6778773595739225,
 0.28989696582495417,0.8850384256719706,0.09157882038253407,
 0.6543452083583425,0.4186698424513625,0.805620870670803,0.6363462377746294,
 0.003948835317107879,0.9196358339885418,0.8860110364410925,
 0.4717245974444687,0.787357128214032,0.12731524111211145,0.3800756632111585,
 0.48624917986390814,0.6291033512090757,0.1291627596095677,0.901801455365768,
 0.10919525581702161,0.4683248188959389,0.2514678907726504,
 0.16090538868643944,0.8580960372920596,0.14759694640645882,
 0.37845179976423404,0.3890073243240487,0.0171824262068887,
 0.49053095560496074,0.47586298730183685,0.9922557165247724,
 0.7624295165725641,0.058804994393206744,0.9193265049355616,
 0.05106663204796391,0.678288650324212,0.26681275102724156,0.8918104318032545,
 0.302538011566976,0.8538546925473874,0.15419017129689305,0.21979844073832255,
 0.308799243680127,0.22120099339426058,0.9298853358433646,0.32453472388056426,
 0.6916130926932169,0.6790896451581314,0.9883482845246495,0.7367282101043591,
 0.30360537858414216,0.3889050029398523,0.01689431032569877,0.575572638492045,
 0.053782020050354595,0.33444403398969214,0.8767134633199887,
 0.9509526135925281,0.9857031570888728,0.025799507976617564,
 0.020905804555920615,0.009643827058670063,0.861947613581918,
 0.3895119714371874,0.8911889800551533,0.9071661666073048,0.7141377980857322,
 0.9864108021793834,0.5867729164604937,0.16365683535829612,
 0.22322255216444897,0.4089212925167034,0.6806714263111405,0.8440299606628211,
 0.49220543152781737,0.6034616068593867,0.7165434861568809,0.8853559029218652,
 0.5902967709084163,0.7940932318690138,0.45140323696039153,0.9104926112553823,
 0.19699493326817785,0.7797705898461921,0.4146458680479268,0.7202049812798553,
 0.4572404889592513,0.10461836144373238,0.17759137887986154,
 0.05689912896749849,0.8540236358234162,0.2932970090956609,0.3801829287092553,
 0.3446458075166975,0.3359021043776992,0.6241628825683636,0.536957280007504,
 0.8101797366591176,0.11706682018663628,0.3850570058528682,0.7710922845539372,
 0.5003274392448299,0.823474565325863,0.2467909561310585,0.205878885809064,
 0.7742357049470581,0.8097993023540272,0.23538756777781522,0.5306829595040822,
 0.8154735610402897,0.44556911950797684,0.989798372080603,0.8464385382560968,
 0.9147309040504871,0.042691246276506556,0.747978385213522,
 0.11410986092283648,0.4664263011059766,0.8925506268742961,
 0.46508292769227466,0.7284083037894007,0.49527582258401937,
 0.4961584584951694,0.26411097745545864,0.6465747601707932,0.9311816318909646,
 0.17862302822966603,0.9975134124260888,0.8850432904005998,
 0.047052085823648504,0.6081831462507679,0.4168071247385715,
 0.9008257283379513,0.7251095900611468,0.7318695292277893,0.7207462394439896,
 0.4779000952648438,0.7790106248562099,0.4356825760980555,0.18751321033270663,
 0.27259589716325217,0.19870872351903313,0.2964907230648667,
 0.20960900732357168,0.44012390399895396,0.045597270068736706,
 0.8092881205797896,0.6805746120791879,0.700050231185645,0.2122869728599044,
 0.466476183479274,0.08680650025127801,0.16082066964594943,0.9680074570986472,
 0.015468782718284557,0.8068639120097972,0.43986267981756755,
 0.4249927126567967,0.12062458264782594,0.17034835261671866,
 0.49119547568112676,0.5491022181254217,0.8812606952007881,0.9317043808373869,
 0.13392736843511688,0.37805626777904544,0.4079200446002599,
 0.43675265537194485,0.09898847235423958,0.4078656025343549,
 0.07529155469832438,0.589733993593665,0.17562441983138877,0.3636157595961438,
 0.780075324138117,0.021207108502894978,0.1525078050780122,
 0.17701507056932553,0.2103877922848556,0.14899651380030454,
 0.9982234044881143,0.10871704246376912,0.2792183420866565,0.3274662962734157,
 0.17503918962691445,0.44405256070060195,0.0726699004600917,
 0.9654879701906757,0.37429046896075546,0.42199766106504943,
 0.28119967902672405,0.7518223482534114,0.19948534392891348,
 0.8116876945448264,9.326588372361444e-4,0.3697095132214788,
 0.9192368107825976,0.41397274077271273,0.5002484740780729,0.5619693609483921,
 0.06822851665005947,0.434882524474156,0.07461205329080312,0.1357167037163648,
 0.8211003030795822,0.3920875405901303,0.5011525065150576,0.3321076941952785,
 0.1787289269417558,0.3556939218761185,0.00793909308331886,0.2423805909090323,
 0.6169593954318568,0.3703601949625499,0.4955489361945504,0.8471002002363822,
 0.004215069676858807,0.6788271403366982,0.3594323857642413,
 0.8105649931932263,0.8871998403116104,0.659834873527688,0.736791535146384,
 0.9981848294446091,0.5538271815198762,0.12748088459394635,0.9436865456580231,
 0.5977641226139963,0.18871459679244862,0.33328475316558304,
 0.32578476702516523,0.6866270341093961,0.21134625504581772,
 0.9708170718212911,0.6948884725057802,0.6323070307933061,0.07901491215979017,
 0.3180784980844946,0.5108263899621748,0.28636223349140666,0.6106515304656595,
 0.169352539307553,0.03090068755262987,0.9251174277427195,0.2662675998411842,
 0.6390661534331378,0.18507239265552777,0.2841998425172342,
 0.25559560284045424,0.5131449708187573,0.18768022941455265,
 0.9860540259344531,0.5706977078554112,0.4448368870197239,0.650779651857414,
 0.4167732671244221,0.3566810349022749,0.6851988221638413,0.4218081191640798,
 0.47486490494388345,0.3174832928237469,0.21061033119312667,
 0.8842581716083929,0.6445920213068164,0.20632847730458703,
 0.18413057771116592,0.33785927858309606,0.5647483516930721,
 0.055436077458372646,0.20464533326502088,0.4626402973448922,
 0.02215712939640624,0.3974022748609307,0.25986920658703105,
 0.22275970713122817,0.712984587784499,0.03421473946004594,0.4578924868902271,
 0.957315019476388,0.485135643045846,0.1588466043904917,0.7174333996764855,
 0.1301046952056154,0.8445942621045923,0.9034577358347957,0.5533973068358251,
 0.07923073176366213,0.5851859572749879,0.6854998680493756,
 0.30951965110556634,0.04486038867532571,0.9611345482314728,
 0.06928157121801337,0.34177369968244276,0.305238505456338,0.5190123871440481,
 0.4315950354397652,0.4502165917082599,0.9351904934361949,0.5967292298977498,
 0.025667586786771057,0.015350013249353922,0.3474696867135334,
 0.04895276808113003,0.8259264436440468,0.3164135845203443,0.9174712259326938,
 0.008272928470284935,0.04877163236539206,0.31254522073877555,
 0.019917013256379734,0.45659600460043825,0.2863405415526864,
 0.7107676277834967,0.47958000340732254,0.8284115096485539,
 0.11835309896022417,0.3053034407552869,0.8918329307093519,0.6768719535869546,
 0.629885340026575,0.9043039315137572,0.15222381678809405,0.17213437625674666,
 0.6406636532928062,0.660727232260485,0.7675349756540153,0.49379456361104135,
 0.3744284924765835,0.9956822571618564,0.2534773608469607,0.8620441681443268,
 0.24669194161455898,0.32974512260498645,0.5601922469813571,
 0.6555338413295674,0.4108888970405914,0.615528219503003,0.47194602622564497,
 0.5441529187010382,0.7859921774361049,0.16196541965464717,0.9404589333401938,
 0.9417599383390698,0.6441184581579693,0.01206000234688609,0.5015929265085096,
 0.026858280279359437,0.718788704346582,0.6789893913084541,0.8017955245936246,
 0.9351335892692536,0.5110580184005776,0.4704383735807066,0.2166535205317689,
 0.6001885087839004,0.06845502565688588,0.9857254755476267,0.318624969111008,
 0.03717367638535507,0.041318686432370466,0.34775968274170344,
 0.5989882875486979,0.9593425046701349,0.057710779101029175,
 0.6239065425151951,0.8843079977048718,0.13377599200297685,0.5318312156658185,
 0.4504826271849718,0.4679312875758007,0.21864234971566432,
 0.44857940680416664,0.4800180630585742,0.3102474314490743,0.3392359876085399,
 0.4125467858462668,0.048647196444426055,0.44998453494593027,
 0.14445461238842638,0.8892192313079839,0.7096131074396427,
 0.47576966513928187,0.21393657660305432,0.02961843076183157,
 0.07604962639877999,0.06371006574161586,0.9590072355495682,
 0.4580178497504491,0.24377534620803998,0.17184675067716082,
 0.6574020228781865,0.7513544167883393,0.6249245818383036,0.7176436031838589,
 0.6767313763307061,0.11793405806647572,0.04968326809268209,
 0.2262808683415808,0.41648324842840256,0.7183771601779745,0.7542650284616661,
 0.590779116910128,0.45348941249604946,0.2825397668031354,0.16096256773541218,
 0.9219282756257987,0.9976795996820759,0.2553346183622459,0.5068200117796976,
 0.9554725400732037,0.5439640505359132,0.24287674535074444,0.527566805182237,
 0.2733468509792403,0.9860098806246178,0.07855862772249222,0.6308965294104426,
 0.5997280177736136,0.9883828147238863,0.9324568250355061,0.16487646454500582,
 0.8087270719678862,0.041703478029716634,0.9712071003285118,
 0.6421553275955795,0.49610372946792736,0.3461471932952914,
 0.16257015997383306,0.8955049506708177,0.2908191933271794,0.5235610682154355,
 0.8363690853156155,0.9904317657317999,0.6384653048907759,0.5342689181549218,
 0.23197374992656927,0.4954632436126277,0.6753440696462742,0.8632541283500168,
 0.9373351691567572,0.4458717520281211,0.5073692492874871,0.20651002059597812,
 0.602413823353978,0.5232008012048238,0.24350582551803623,0.881521314779628,
 0.783003076000123,0.17020148728887197,0.9823855708836817,0.6790343198949115,
 0.7650543747822258,0.2572485764674125,0.7982735870213549,0.21796549401061405,
 0.6526572697894352,0.014314045692786381,0.7706346910909989,
 0.7232213291618484,0.12385585285547818,0.953655615798566,0.1766589431249559,
 0.9061518090171834,0.33962248795671846,0.5234081388320385,0.8634102846920179,
 0.5344074399677191,0.36237594231971526,0.6072587844540127,0.476477734244815,
 0.755725062181752,0.12199681577203836,0.09893392483366403,0.5452601459067992,
 0.9028658164667088,0.13171721230681366,0.7043007150614533,0.5911942256838534,
 0.9762851228360756,0.3931566488193563,0.9211532881207302,0.4026564280989554,
 0.9397205802286702,0.7779432484222047,0.9280473975443789,0.8930775028774947,
 0.1260322676488398,0.5761221813515666,0.9121524123881861,0.5027808582895605,
 0.8317970088353988,0.8561979431262742,0.7795077703782946,0.3004468121352256,
 0.6072362606438576,0.4866976594331053,0.014707297186122847,
 0.9029996437450698,0.37426900107185757,0.5034262308852194,0.3194238569785388,
 0.8048248166802663,0.5260721738802613,0.008382041498633863,
 0.9006919103436043,0.5077952731131465,0.6034372640649907,0.28625556698668797,
 0.7976151699365057,0.0856925982811344,0.5555456279300132,0.8858085004422797,
 0.21092833605815797,0.09617596475960133,0.2100706822965659,0.50650987469929,
 0.5473537837625471,0.6303864463003469,0.5621159934770924,0.8499433755186986,
 0.5774502848722569,0.2799847935308861,0.22231098687152584,0.5762444146846585,
 0.043760784852935775,0.07936519839018152,0.5037678120001605,
 0.08233423697157949,0.8059835059620152,0.9669430539602999,0.3945935296172298,
 0.0783965007475963,0.21785959079310757,0.23292361159120611,
 0.4560446499329752,0.14840950237992,0.3969668606148963,0.36481723445497705,
 0.6861282179231983,0.38032624130870785,0.973848935719387,0.8268547622423073,
 0.42021829135404465,0.9050594567606608,0.4955861838897353,0.294778956155072,
 0.402628055568067,0.3624107056272703,0.3756840925659598,0.5376787389275883,
 0.586199526427472,0.32116616476714255,0.8860467612643963,0.7749311201270271,
 0.6156334859616427,0.22128048871881978,0.8315529479984021,0.5880056145711575,
 0.24338754128353712,0.3986265064608314,0.5139665965022137,0.9782400664872266,
 0.4111943981256364,0.4572253193862552,0.2225830440910651,0.7208965517720068,
 0.39146377227123585,0.9767378130919393,0.4373454051908583,
 0.15920871388534885,0.6250337384555585,0.8250778785494166,
 0.43957039735259584,0.935430414176122,0.7681642585181153,0.05905758404764083,
 0.3194173657597774,0.21478428136726357,0.005933809262607892,
 0.2854651933533904,0.4366038821777488,0.9537786108692738,0.7377936571545476,
 0.03499060622127792,0.8166373558818563,0.3036879636785422,
 0.06830265765980092,0.4373890769873625,0.01450133159840472,
 0.17832168732048082,0.617734860231888,0.013596119925655237,
 0.8879491332019744,0.5867703313889959,0.781406171773041,0.5273772621843101,
 0.8255962867138362,0.31014016613838313,0.13855923632772615,
 0.6580931767893843,0.6486853536468469,0.03570724612120979,
 0.08570428246012451,0.7930930143265952,0.23537211928973345,
 0.07403346166684766,0.8633784833387315,0.4808330040705069,0.395021057729949,
 0.7176303831296075,0.5757497439962808,0.7253333965737734,0.5810704033573364,
 0.7775020489304386,0.3151533181488386,0.6026988064660754,0.5780517813972461,
 0.14749243535564305,0.881199560447316,0.9520797692297032,0.8748861466823898,
 0.13850481162895045,0.8115060624169484,0.06729505064259045,
 0.28415386201038284,0.3187191338761671,0.6562130903620852,0.4757668160786087,
 0.7720812573427676,0.5210591479170915,0.6309457629686748,0.25533197187718815,
 0.754093777332806,0.7122771099716063,0.8400277214892914,0.9446314274219569,
 0.9763907543447499,0.7297078521098399,0.5495862722233382,0.547330102505259,
 0.7263158576313482,0.22550759109618723,0.229186709429944,0.38751401474321734,
 0.5352998804682784,0.25882405655722596,0.73606908150368,0.9434580138112765,
 0.02783139211042307,0.5894545424816077,3.0601620779235894e-4,
 0.6313085489814716,0.6359710536932298,0.8667223929001059,0.5326287419105117,
 0.5196490222185592,0.1573638553835418,0.0814224084960059,0.24996924309837432,
 0.06318332013471273,0.1337750028338025,0.8028592000198868,0.5438889007274729,
 0.11860853426320894,0.9591584135001158,0.26714930991909946,
 0.3409305917354877,0.9979015358017769,0.05102466666173289,0.4748611040093589,
 0.06693467722640167,0.04279606205797104,0.9592819327728441,
 0.5618124555000749,0.1219932797211365,0.9318457980529992,0.9093568844534311,
 0.8983515484245893,0.5115793361506061,0.44480922408713364,0.515772342823963,
 0.25609278873951,0.7047847321643615,0.9616643687149599,0.1539613081626362,
 0.36992616447031124,0.6672528736482898,0.8205909675159966,0.7067041843949484,
 0.5727073034882579,0.19692108370462413,0.6217693331935017,
 0.06188223281373362,0.11284665001386518,0.9400803680134018,
 0.5075533703680336,0.7059678358305859,0.488739679095166,0.7451795285264275,
 0.4695212476383521,0.9529767304662616,0.14566151556178486,0.6030054780462618,
 0.5023671745395127,0.05026467616121,0.13657588732350034,0.2211064170341278,
 0.6890170879904152,0.8026218984785386,0.2964196755780957,0.7122556091511061,
 0.17280512303306206,0.41760295771474965,0.7213566914049998,
 0.4800792163422505,0.2004286110428951,0.5827730189902246,0.4023944229045602,
 0.766135402163928,0.33113788105647657,0.4976729442634794,0.6770779792515846,
 0.8686829149366805,0.19276418349379842,0.37158145185324576,
 0.3833187627649698,0.9377565675358043,0.41577138565246685,
 0.20713648465687895,0.5799276009531864,0.9521984833947661,0.2998263886440552,
 0.0524862984508665,0.6531920941766332,0.4549021697616178,0.8403887351286419,
 0.6317357123335329,0.1791451060158359,0.24333920748554627,0.7202344424221422,
 0.5170687812141102,0.37732832820961126,0.1573477359307207,0.2605639891659096,
 0.08190724518955939,0.07105402291145668,0.8002572826925034,
 0.31439353054317176,0.5224767710601976,0.3545919679946621,
 0.0077901479930198025,0.40548906843776056,0.702859358516552,
 0.12591538328541563,0.8351146766915535,0.40385204918365636,
 0.7106500269036347,0.6731067519854397,0.17924601100314275,0.2803359664394406,
 1.9161232677655793e-4,0.5196730117494066,0.6343522591787938,
 0.6526559588200803,0.05697224815611501,0.5264660263125425,
 0.22651869757785792,0.3060006662668304,0.07170486956619426,
 0.25114509335545854,0.6427859176968928,0.8925023768116828,0.6626395223466844,
 0.9192991167136152,0.043218432158074305,0.43811398999012763,
 0.14969117735776305,0.3485259569803273,0.18460559039596447,
 0.042166381599232916,0.2783659397784267,0.4846721807239547,
 0.5677689298241594,0.6338582567933678,0.3155171247405343,0.9175918304226736,
 0.6539573990501187,0.1809106785757102,0.8401598356727543,0.5654511854830087,
 0.5912355250782764,0.4378299756419166,0.5194018401470359,0.7872083198240896,
 0.8055056212426701,0.3553316927342709,0.35301149518982433,
 0.33255454090914327,0.42627876061857084,0.33176788039467175,
 0.7096210141008437,0.5539380221926641,0.9711708969334686,
 0.021339053072131353,0.12180152974185454,0.3117560302707436,
 0.003867045814397896,0.26689182634094233,0.8243605754870362,
 0.28872929202565467,0.19274082765126277,0.262438652670274,0.7556897852022121,
 0.6808936814710576,0.9981281826019612,0.6299552445704609,0.4581214907658593,
 0.48898165739109345,0.2883865415538067,0.5811121256464215,0.3858554717966669,
 0.4706189036176577,0.9946729955268292,0.6284277605279068,0.9072681075993237,
 0.11626209969162338,0.19194609344207236,0.23670179587716345,
 0.7872926873114335,0.26482537724273203,0.6770374909283331,0.2225593076753729,
 0.8909483139853582,0.3078761561349689,0.4429235244019004,0.3445809429901219,
 0.9325640665957269,0.752565501546948,0.9790492010479204,0.6199027646917143,
 0.27663339142561405,0.09058036659052093,0.23765817433208958,
 0.642741521718074,0.2582713817021741,0.8632735761334526,0.8482844336750258,
 0.23162533422827902,0.5994417169743061,0.5219454026265565,0.4154309325526002,
 0.6469942101830923,0.8730605215264521,0.6375372986978609,0.2998887825548945,
 0.07219578000037641,0.9522786114710371,0.2582282755597256,
 0.007904804899114493,0.8242479433034132,0.7194707681768469,
 0.7797706173329855,0.3867765281794464,0.6229472277699849,0.609696011870905,
 0.561201769636861,0.12086427735554117,0.041505922689302865,
 0.7201447841620947,0.26524706595503966,0.31202514098934786,
 0.7932297073313702,0.595522153195109,0.3141269920716163,0.7008256383688136,
 0.28843961519171135,0.6274045170761692,0.35077530084478314,
 0.6033607145357889,0.2522203755240665,0.5787967496998474,0.7001592879749696,
 0.07970477780807017,0.6463248525251242,0.5717374717141203,0.6290570790666852,
 0.1362572307419334,0.6134457223653363,0.4651053627864925,0.04461824870580666,
 0.9240467312651288,0.38026782808519644,0.25950766095393174,
 0.38018894539492787,0.017630883287788945,0.9292909790176541,
 0.798428532557793,0.2604037543326584,0.803355479278386,0.3321794998817167,
 0.32118120462584254,0.7761998297610089,0.7431462723168039,
 0.12677472619238106,0.1731494453486062,0.03548632098345528,
 0.6158384863815702,0.7770095998932369,0.6680907305200843,0.5441191275473116,
 0.7539456491335427,0.48525633689902914,0.47420683353851156,
 0.5012369188042981,0.6504997941703992,0.19924436054560735,
 0.23852820700568267,0.4325358348182785,0.8547075137694653,0.8826605862410549,
 0.850614562667645,0.5947329074947196,0.28572894998857934,0.44775614076099624,
 0.24035116480813812,0.22280903040432765,0.3951702453244039,
 0.0610708668722767,0.9097745324142651,0.9043526942553597,0.9204046557383793,
 0.49198797071288036,0.36835940158895697,0.14629554870794625,
 0.23682058250296323,0.8255593536159852,0.6267236021580316,
 0.22960506432590566,0.9052756942967345,0.9251620261388698,
 0.12777362013915095,0.9752703368013126,0.3668082432288584,0.8420484110396753,
 0.07908608193002886,0.9136952683394111,0.42707530968565033,
 0.4666264792063478,0.07590185403405147,0.49483510479186466,
 0.9885645374933293,0.7364319476402972,0.07659979961578522,0.7675472829187987,
 0.883114722767461,0.33363771173888834,0.06593212795221748,0.4508971491968462,
 0.34550119367922794,0.7047987930105931,0.3203658930341948,0.7465498456561472,
 0.7473596075177277,0.5472217145986927,0.3688257238771565,
 0.058583966476787874,0.3667136830568698,0.5529896433175455,
 0.02909472909716837,0.45780353712573096,0.6454651908894372,
 0.6648106319452052,0.885534188145456,0.37121334956191787,0.17711513744068286,
 0.018468042907499926,0.5849075347500282,0.9516195455249379,0.886423868013375,
 0.9905387503635283,0.7543214338991645,0.8950961784861672,0.5531798594531172,
 0.223701948491157,0.9443052607327861,0.3378012309189584,0.5812401884920451,
 0.5042616400404243,0.3719547730814724,0.771525710949272,0.9812886039994113,
 0.7554811476962027,0.3646040549220133,0.6594304699710355,0.8721870725854609,
 0.6629067811010698,0.5651165409942953,0.028285986012840603,
 0.9000498914134497,0.1290421657517341,0.06760386794485618,0.8098336847026073,
 0.5369830818943219,0.9106450230057521,0.020667349953195302,
 0.19904210647760867,0.4849031301827985,0.21066280691872408,
 0.8634037553097953,0.4656689676275889,0.11707917807508594,
 0.23750331125570112,0.6094206644332801,0.07889108018903024,
 0.4554421779481048,0.40628369796960373,0.6458400452423534,
 0.29330529371441105,0.9222775190147533,0.39484055553286823,
 0.2345964700713178,0.28537887446619237,0.5908072952216123,0.538331085866567,
 0.8938986715130931,0.6752684548085442,0.6589836982338478,0.7651512517656072,
 0.94610537369688,0.24738103879946782,0.5180017452932125,0.437892702803489,
 0.6299585941881374,0.8581862782817676,0.3385294141579255,
 0.010455306054667823,0.7579686230607393,0.411871924725265,0.8921187054507086,
 0.12583563681689602,0.19831633550997851,0.5209025048423498,
 0.9343251246561416,0.3133704862313025,0.7293029858834662,0.8616888392032973,
 0.20314652384766718,0.8876308647030592,0.8153123435016654,
 0.41769228060264585,0.40661539474826014,0.5977408413729918,
 0.8188596551270187,0.9279123753910814,0.39078887222114567,0.7128280135151601,
 0.5989344066794418,0.8980047925192332,0.32662160547265484,0.73306189940339,
 0.32278967156257565,0.49024565400353615,0.08399379577617705,
 0.06529366538023973,0.7020542958104735,0.13310031629620445,
 0.32531279335996877,0.6504951477997131,0.0364708338522759,0.3811257408747768,
 0.8130325585956586,0.8305940405774411,0.6400990154797682,0.8763452446041144,
 0.5516621368619544,0.6004967803265725,0.7995651916527029,0.6443819173365976,
 0.897220264640835,0.1301748725122709,0.2708269633596194,0.9912365345070991,
 0.8482090818565587,0.832231238458271,0.5555129903683138,0.04835332897666744,
 0.9910662740760916,0.023848330446257227,0.770842068155938,0.3060853446460494,
 0.6139022493402311,0.724567119195839,0.4492197127132461,0.9104328991326198,
 0.6693969564174898,0.14192367569103692,0.1990158522843949,0.5747210073219482,
 0.5781716264261384,0.9236503208393371,0.3401946675833337,0.5385231192440977,
 0.5618868903823556,0.9366850902482469,0.40081090885871296,0.8834878871475573,
 0.5382976914853307,0.4457527181821277,0.624520523776829,0.44689497488171837,
 0.9794702118944562,0.34485360159715417,0.5376089866947307,0.9477951186288389,
 0.32121488985250957,0.8639082113160419,0.47189087042056344,
 0.4657351985490329,0.6229947294611475,0.10915602454892281,0.4434372837683702,
 0.16523993498906342,0.5982956540866151,0.11595546988386485,0.197088372302463,
 0.31223294612889163,0.07770361587258956,0.534756430998885,
 0.05278411848800546,0.08906307797590085,0.6872687418270238,
 0.5852682003889363,0.9459747518658193,0.743291809614503,0.7570769713835329,
 0.35678330434041006,0.4293991232037462,0.7750305025851595,0.5759854372914208,
 0.06312970116099204,0.8230813523190872,0.49324724681150545,
 0.6735373125762854,0.11339682026846265,0.6145605390477916,0.370892851013888,
 0.12590370951508145,0.6783811908636701,0.6914389725981691,0.461551644356037,
 0.48048579472275965,0.4137557508254588,0.21792739898087643,
 0.4724791029546639,0.1646450118573538,0.8770580951925651,0.06810880584107104,
 0.8451952923555709,0.0028005405158182395,0.6694593854668337,
 0.19891792213212744,0.05512292890516801,0.6790698717135798,
 0.8991937298904865,0.26733734703971734,0.33008035787813417,
 0.6851898614050598,0.8349817947681502,0.21903168111671456,0.5982623274957054,
 0.9008357686345776,0.7142018191905286,0.562161398010044,0.19845480505883983,
 0.47367903403019007,0.25472264774677866,0.6997630388703253,0.9036879984079,
 0.024074226240075824,0.35956449572552773,0.44471423391455844,
 0.2173857805334558,0.4627727710021454,0.8510561066572943,0.07667172519446586,
 0.05958451812028631,0.18986247233902376,0.1567796924159619,
 0.8328805763905072,0.5328598706227847,0.6339802620999306,0.08771482403639053,
 0.9691628082615901,0.1296731992355945,0.8310077422262349,0.9419267601903322,
 0.1397080371047792,0.9009620725584178,0.7619899262770699,0.17182276246887263,
 0.1624443162908965,0.36864810365087386,0.7691564244944293,0.8333271169625658,
 0.5788465877195098,0.25456411049812744,0.674298540770906,0.45984663437628326,
 0.6516873515178407,0.9346021745772825,0.8448591027905653,0.943303210061141,
 0.38488555314731854,0.5684963821817307,0.8715038418880384,0.8897032450789887,
 0.004541053541125661,0.08177130165742108,0.08909208936133495,
 0.2352606465073276,0.9503122735030596,0.8235249527575098,0.07832531819141775,
 0.132735480541189,0.09275410088195091,0.33027437063194354,0.410119787135644,
 0.9562895449088942,0.1817223416577637,0.7061243584480132,
 0.015321670625468853,0.933299162388397,0.3535107644458798,0.4754444715181523,
 0.6312590400210183,0.2850224597734372,0.24524122995090314,0.8630213935539615,
 0.8532688216044678,0.8975121802099842,0.528816615631782,0.12861143229099392,
 0.43990641339847336,0.8050487981580062,0.12923124448951173,
 0.20240568864150044,0.2631674442780483,0.38458672718737286,
 0.38724602425754906,0.8249605089536849,0.5243221900851445,
 0.019562038450002484,0.6380529339344182,0.9939017462589386,
 0.01700398404031256,0.008674738547723049,0.10122952235132066,
 0.17950967058091394,0.12568346444755585,0.21993507122819256,
 0.9506455209765035,0.941751028139307,0.6798131819650173,0.36971480975172977,
 0.33702720060746816,0.9471497225936487,0.4729562981474228,0.3669910302818632,
 0.5191331849355213,0.41539196240329157,0.9987046083213563,
 0.34696893124711736,0.05289601615610873,0.2330905833848792,
 0.3350836058985365,0.1089480920809649,0.47357002586169383,0.1626629192201562,
 0.5959764998803698,0.4985788354335313,0.6879216090749476,0.512926401957129,
 0.645114302124141,0.371779368693678,0.30512240535864044,0.10124541042946378,
 0.61314103896568,0.42299596069028234,0.906999013989763,0.31891945237072883,
 0.0254876255002201,0.8248173263318598,0.02453173126190822,0.8704083723774775,
 0.7371599413446133,0.21020937601165668,0.7797451302725443,0.379318404132439,
 0.9534579790097654,0.6980926903277125,0.03639879648980815,0.8724887793811252,
 0.9409226456811811,0.5551764724392017,0.20168907336262765,0.9336403103187506,
 0.7827017615997551,0.31631217455337723,0.521615481211209,0.35984551265592835,
 0.19384803540442064,0.04370526373055661,0.4905913789627895,
 0.006812648077129335,0.143511255796395,0.8072045048263484,
 0.02978104453481456,0.07084926832729288,0.06828531291140283,
 0.6320113440410117,0.5543413970898968,0.08690630195692739,
 0.08509762140858335,0.21084520020710473,0.046203797962566906,
 0.44527558876141127,0.24771322303304788,0.25277556392681666,
 0.7652087312449778,0.5773122745558424,0.38562361478484397,0.2020712438647151,
 0.8593808277637649,0.18649577410997997,0.19996636693547476,
 0.19018796309764163,0.4782307736167163,0.39586860634831444,
 0.9748672830974208,0.6019159830591577,0.6674580049989337,0.8104430165468302,
 0.33950231833467703,0.6435014470440442,0.621638299613659,0.7890731828707199,
 0.7813399611179365,0.8279948498487548,0.8251791614160917,0.3821019866617703,
 0.6609940533908907,0.4107181817568939,0.2958233511553686,0.24058342733334115,
 0.25658294891899214,0.37572191342552497,0.9114644680987732,
 0.8869686765800411,0.9451432717105133,0.007147815544988445,
 0.14478214951403834,0.9361862063934125,0.5696988342045514,0.7098592543631341,
 0.6350770748774408,0.43605634451709896,0.33800254609076297,
 0.7674610517814618,0.44534890249438,0.7426207070487324,0.9708981037262958,
 0.06049111766481863,0.4868220556633527,0.680203975389808,0.8070903279986351,
 0.8124298443878946,0.8262593754172814,0.9259911098773317,0.471592645379049,
 0.961369313174686,0.2288885615223747,0.6208849048037055,0.8767690772717577,
 0.534155052734688,0.12503977798645682,0.7176374037920199,0.49498926934095655,
 0.2890185863196144,0.6009461608087912,0.3765706008422114,0.8104780799701921,
 0.2859111751096258,0.3471433289626531,0.2659504998159634,0.2836277354527077,
 0.3888706098582234,0.7530467504588233,0.9645273414707284,0.5007956402383253,
 0.29547190950752256,0.5093042198869184,0.7158750610776045,0.3653278459987148,
 0.7954226776011554,0.7098628318969463,0.7823108919954027,0.0897628553868175,
 0.9105433311902581,0.45848938759942637,0.8225517824612352,0.8362470629722245,
 0.8979623489206474,0.9399474510038146,0.8701052181287948,0.7445749404744191,
 0.9657425361925855,0.40953324531075197,0.41000593204200586,
 0.8064897886201278,0.09838557034709328,0.5586142346373665,
 0.46214001479941313,0.48701000430370456,0.28826725295922184,
 0.9084621011841939,0.19209583779001183,0.7826078464844015,0.9486628187590458,
 0.12185283016493942,0.34468021873643795,0.7528005017391517,
 0.7633625866085824,0.6845540003886387,0.9868413410758202,0.8760512649315644,
 0.1715289662632159,0.9878798666805892,0.69264259412031,0.6899925961599763,
 0.6092173817459967,0.06709611985600328,0.4802243737582914,0.5368428714306916,
 0.28100582857669276,0.7695845067276287,0.7690343249887299,0.8710357926344915,
 0.24014960418213926,0.3420303790137076,0.014800805025673491,
 0.31503713688200374,0.41714768035833316,0.38239126854082306,
 0.6525744068024613,0.7259746715194866,0.6348261668117666,0.633056271855312,
 0.08866365109016261,0.24185593924452764,0.6459659660921013,
 0.6168937927452325,0.7926378085300072,0.2107516132808065,0.19552434287947906,
 0.34442410554941505,0.5042919258599611,0.8371457492373919,
 0.27848688812582933,0.934963074423319,0.2941246664526088,0.7594084128034644,
 0.5555327529056242,0.6492632956506872,0.19672950541994072,
 0.12862969171247085,0.7340318710726046,0.9617998064816703,
 0.027044337906494076,0.41192733462450515,0.6290294372178673,
 0.23507605544029053,0.8912970344252118,0.952171479004204,
 0.056987595453153905,0.18394064907941288,0.18016549099196988,
 0.04491464468451334,0.32804737635299064,0.5402204446045955,0.650075385294954,
 0.07216681332656327,0.9903105406394612,0.008747902228102067,
 0.4681175411916785,0.24511218301557525,0.15401695718804875,
 0.3963247419317122,0.433123302048288,0.2502600138369525,0.7838284055396921,
 0.900423277595181,0.9475951905588627,0.7942810241684349,0.9063927673869492,
 0.7120693416243292,0.5344168406858134,0.4126072819578692,0.6310057759479212,
 0.9485753974589345,0.15675731406864413,0.7504798030738198,0.8469690248239267,
 0.049557320352353385,0.10576253869066377,0.3910824284030192,
 0.7632435339913007,0.28635392481028865,0.05927163572786176,
 0.45989207132814636,0.34971064012997977,0.9483734773883272,
 0.9604408058550359,0.9734976466711118,0.8601986323880495,0.6355483599512485,
 0.5422924247298382,0.29101524702269876,0.9160068705269049,0.991552890997226,
 0.017042497312593774,0.63393650006927,0.4809273524627362,0.7362094546812576,
 0.8580179255843748,0.06752187431516654,0.8127900005719833,0.5633971627614378,
 0.4987995074059056,0.769666930395341,0.9021570742603132,0.05820249641649833,
 0.13475025414697872,0.11171438906939557,0.40232462567462035,
 0.03286313950674735,0.28012260667586775,0.8753569276178766,
 0.22083103677858507,0.7698343985480599,0.7437311032740568,0.7315863920346991,
 0.6637479962899259,0.9297295957504454,0.2805871030669558,0.7458165999056813,
 0.5085112717830292,0.8223929928275655,0.08624779634196889,
 0.048167532981645445,0.598998648566474,0.7543324661811703,0.3385280761777567,
 0.6263919143808837,0.6018937448662688,0.5292059997290142,0.734318423847389,
 0.30607247118945946,0.5070154518484514,0.9426094656609804,0.8387942933623883,
 0.8296192501653898,0.8885990597792163,0.9530850491399717,0.4394645415757438,
 0.14235312286572477,0.3988823182473932,0.951957653623153,0.561256275015032,
 0.18011237304633,0.12178216292838784,0.31007904706829736,
 3.4253071448053785e-4,0.053224812793827736,0.6960408705511544,
 0.8343074839372752,0.8554283367841647,0.8151530226306648,0.8182138652244805,
 0.2600348003435362,0.22252321903225325,0.924893041850433,0.30112946739231017,
 0.44280782680854247,0.05554322207562079,0.6394351617705171,
 0.18327326364590746,0.5411791511364213,0.22097503910647942,
 0.7774653630912453,0.4843386353351875,0.5424416182481512,0.7153668132277617,
 0.9382404159341452,0.9302844955202072,0.531389601378715,0.4666719302170875,
 0.5323542386418458,0.6887224369178444,0.9289054336894023,0.0623657743404985,
 0.8122138222764198,0.28873813715719054,0.08442766409316826,
 0.9898952365092322,0.7082230270289716,0.8769766881806659,0.12459360527281849,
 0.8758338004001928,0.41543970331890656,0.5681283945075355,0.2053590492174815,
 0.40026275047819626,0.5153822328566784,0.6704035142329557,0.3885410232659908,
 0.517488260709861,0.2819066519060809,0.80707058372281,0.3474170911227018,
 0.6387325542545159,0.6088316394001142,0.4116404518524208,0.7989407290982413,
 0.6749723612196785,0.9412684923585147,0.18526078281654235,
 0.03905824618922327,0.4516416573187656,0.33033306766500425,
 0.5950921446338564,0.2786369889720697,0.018101367671169544,
 0.7672999277341039,0.06122404384110858,0.8437202296865525,
 0.09935756827196596,0.4402372087621481,0.5222918656249302,
 0.08485046591790679,0.13071908256672737,0.3145577656385208,
 0.18735739893516423,0.3370421562074438,0.059693656698710074,
 0.6562428458075504,0.37660470645393096,0.5273600488651083,
 0.47721830898475215,0.327572133451166,0.21193947960640425,0.4750518625183817,
 0.1457752934080242,0.03557010231883628,0.47486854612220053,
 0.09592649713501311,0.6628196747390797,0.18279542689341344,
 0.0873243959158051,0.4354151785557411,0.4931518854409118,0.07803178714631975,
 0.39497933861920687,0.0984823354372607,0.692949190592145,0.9941912061446364,
 0.5425780849666586,0.7873835067825629,0.4624350242996691,0.2621215868479748,
 0.5352576240686076,0.6411405176577947,0.23914134712469015,0.9811063157936604,
 0.4789689979769024,0.9622631046569855,0.5551186697167617,0.3798801490943301,
 0.6122194294599328,0.7429244893464094,0.9449740022823927,0.5137669067910643,
 0.8012018783017609,0.2856200507625044,0.2301959494983089,0.8704096471830889,
 0.40585887278658683,0.9527982989922552,0.07018083958380616,0.932113641077838,
 0.2566173552232084,0.8846930730178565,0.323534602798197,0.8568636753327948,
 0.5739857229089821,0.7040201879383635,0.16270398079846327,
 0.01095461993694491,0.06570818479953733,0.7035674763615851,
 0.5061682519886964,0.07050378577977068,0.5708072820869307,0.8067304407672925,
 0.592944812591296,0.20279485921400697,0.8793217692324561,0.3745004242852724,
 0.5076121929513114,0.08626170365689034,0.8405553952974558,0.8240779635088372,
 0.7896602356527965,0.3338403051874004,0.734845925666383,0.08472555623053535,
 0.4540073863916775,0.2320601072458226,0.6731240226103545,0.17787360806856167,
 0.36447318839605214,0.9905244779831847,0.03375427832844369,
 0.15239397087863937,0.5420649064365326,0.00897934592385341,
 0.3606685517396876,0.176427969039736,0.4734690164437233,0.13084902540767507,
 0.2779136464691405,0.1903377819872688,0.8594936536452906,0.1725438893570238,
 0.21117268565711445,0.9958885414355376,0.7886479253608757,
 0.45660203886992423,0.789922153581396,0.7917450553407342,0.3904773707417991,
 0.35075248283859617,0.2754577284918631,0.001978729572011062,
 0.3646781111536165,0.02939136661004893,0.2532647595622399,0.4156976656223421,
 0.904456652689587,0.7169614166051714,0.7347545430242077,0.595047761745525,
 0.6900690251720093,0.3656617777150212,0.8775933965826519,0.8897326949364752,
 0.8729174496972627,0.8007163702263325,0.14926229682398062,0.3578235990708638,
 0.3774917671757205,0.18788638214606734,0.1464839259573455,
 0.32358238999402256,0.36354524202514205,0.2111606035874204,
 0.5996129613771773,0.1903191983595761,0.15354542964539164,
 0.30857897540112167,0.9053319059142497,0.2616478702436422,0.9200389257410975,
 0.8310700741759822,0.38484550166037523,0.006534845499010178,
 0.997371261052868,0.23875092772234807,0.34598024548457906,0.6243997757605846,
 0.891612014393718,0.04434385893120285,0.7143913257878535,0.9001778703431044,
 0.40331013725309706,0.05604976307070619,0.26019338844767026,
 0.45805233025518766,0.06145888668582644,0.32949145902536225,
 0.42974096923227645,0.20678044109096938,0.31060060272867374,
 0.4536078781594859,0.3586695396865861,0.32048571932666337,0.6508328614016378,
 0.07665900791288538,0.628993477832264,0.8811090104770732,0.543956044328624,
 0.14512526591995545,0.7469006085879792,0.6486468369935304,0.8194636360416881,
 0.9117283310762854,0.7725249066256437,0.6040265553041246,0.46876902544086707,
 0.3620021679166422,0.7087537602908132,0.6601847374528582,0.2360077322752694,
 0.12469741714756655,0.9071916373739263,0.4666998326942795,0.7723181370408778,
 0.9391500340028689,0.525642710093759,0.995596672716357,0.8374821840213951,
 0.35675284401495855,0.6319590844861643,0.17246989444483762,
 0.11288641663151966,0.8448305727470374,0.2020340461005279,
 0.18890053136004914,0.73665680433681,0.8759754681714584,0.045701542207008394,
 0.024140716224112957,0.40555760661290075,0.2200759231499787,
 0.5723654786104756,0.787036531999503,0.7126361195808839,0.5511255858646537,
 0.13878673531621777,0.6730936360808084,0.3522890596136836,0.7935796433461896,
 0.07266116978395032,0.03754817186657511,0.1166403550284143,
 0.4477773159796614,0.9832317904022311,0.26133078722272696,0.1368981290902228,
 0.6603172771934889,0.471283382813133,0.8784300138270084,0.6681337644631178,
 0.45048114184600774,0.029115905688792187,0.2570073262203171,
 0.9828472585759409,0.20645101757068396,0.7389911906664274,0.6893026566150472,
 0.07186567330486493,0.10198339220519703,0.5319385447595042,
 0.9045731472205624,0.7565263815430117,0.04791946926966062,
 0.37419003414908447,0.5393858745472511,0.851049788815014,0.5736520622487017,
 0.4032195119810331,0.4406451013297077,0.9212550554319656,0.1464026904932798,
 0.09976392593578598,0.551109913880786,0.7575667369401397,0.4291307098896616,
 0.573629137048044,0.5145839633250793,0.8827475054659185,0.44991612840077,
 0.06920997622246405,0.044917249726468045,0.18180354742276317,
 0.6854122626569586,0.8777878546512015,0.9612917841733283,0.7039911624593098,
 0.5227173166081324,0.21020827168252532,0.26469221225235073,
 0.6420564355228714,0.6872270590339802,0.19143706644499403,0.9896787273433238,
 0.6333116931500337,0.5935929319653341,0.7105426111638816,0.34535604408450604,
 0.4088161763776068,0.6669545157394835,0.8357892441525147,0.11540230121029638,
 0.7069058955830904,0.4165139803290614,0.6448538478148033,0.4930008780850319,
 0.3806683772703128,0.1897285695204941,0.029839748514871478,
 0.5496367869327901,0.24523899382471948,0.6721665791149662,0.9663333671935669,
 0.6330751960238622,0.9310428211551735,0.6502926856468835,0.3343934060168575,
 0.03754173998570853,0.08025515113485715,0.8410427214543992,0.567647271134726,
 0.016953578657655233,0.8982265866796906,0.8930620818553596,
 0.7365857665993034,0.5081756323284654,0.39248338330002974,0.1429527968986658,
 0.24288147252410497,0.7521661673585984,0.2919080527603558,0.7524762514421713,
 0.8803940341212688,0.021402988347760887,0.9122112429568259,
 0.20143831985891447,0.3552982808082533,0.8810613141605237,0.534268517399729,
 0.24535297520625798,0.5679342820688554,0.7765690847661996,
 0.10546666083788625,0.876773562712802,0.9659789705745548,0.6689912364139218,
 0.682940752633605,0.46358981246808895,0.8354961242631938,0.5205998155030207,
 0.5357721613250926,0.0572061582148089,0.22853237206629418,0.9637405014004726,
 0.13686880726639317,0.7642632257155688,0.9673333108043433,0.9620302637157874,
 0.29003575798144277,0.7382118725095197,0.32222134962277726,
 0.9912228056479954,0.2710094001444058,0.9825253204803868,0.9324927860505623,
 0.31810081130164203,0.5130196917081835,0.5424345153406545,0.5941034065661828,
 0.05543832863654652,0.2929313754287606,0.6651292068389856,0.8009578444848577,
 0.19549775303958128,0.09662138431386602,0.8881449655333081,
 0.5913477241280971,0.7790965773866644,0.5477913183746999,0.465576110188875,
 0.4540781130065875,0.23229925273476915,0.13509940612875038,
 0.26169825931970225,0.02408563713838019,0.9364518092607645,
 0.01272814776531439,0.42488024143104974,0.543711039475335,0.2627888669312535,
 0.06033779816654672,0.2090992422453113,0.5521375581035782,
 0.03165750043584603,0.7371961656853526,0.6143083012602724,0.3669855501800027,
 0.6409803075357035,0.44036067740671836,0.9063141038520656,
 0.04190450542047508,0.0786406838092496,0.9864266115859236,
 0.22359481718305274,0.5002016794179822,0.7968401996533386,0.6685342826974346,
 0.9802672290908162,0.6775666024113439,0.3844921406485111,0.4525800785117008,
 0.29131133446068513,0.10340384168747163,0.450173363286379,0.9243047784275445,
 0.3103266716468912,0.1980297749361264,0.993395665591862,0.560242584137184,
 0.8922427556655648,0.3391683835341298,0.8144091474212214,0.36932822585073144,
 0.6732820676311954,0.07798454782978159,0.3181455030842143,0.5636945632278861,
 0.59387067384178,0.10665698128337708,0.41205547852765956,0.9185290922359725,
 0.6656730725343745,0.27958524255993045,0.9494149483258548,0.4726948245085587,
 0.9753692844706581,0.9817115168795962,0.24272953113228135,
 0.08089638673760302,0.3980538801154794,0.25009262659173725,
 0.08579930340698316,0.6791291670334588,0.61557149745906,0.5000271934026258,
 0.8716385029680961,0.876059938414361,0.36612113605991936,
 0.052663032439681956,0.20233524946377057,0.304422469083526,
 0.7721479241616271,0.1533841597865626,0.5170592370478491,0.12276819237695746,
 0.5340550574672575,0.11213792789624422,0.9618768430991462,0.6134778269131463,
 0.1511848127812745,0.7008204659850251,0.0024801526714957056,
 0.00385243022820847,0.5184334837052622,0.9980661935025369,0.6912411775174633,
 0.49104266528915963,0.5768522228975105,0.44146456386500943,
 0.5186253267209874,0.9881512966422178,0.14147738810495103,
 0.08362508053619178,0.7565792554592286,0.2640618481239835,0.7877600788211866,
 0.8187862562731654,0.8658576341690188,0.17873522594036084,
 0.040369136368026526,0.8386553944270699,0.4435221428387457,
 0.5396447218330386,0.4592202851616326,0.2256727229773825,0.4907995976944761,
 0.38493446353516403,0.30939780213707946,0.5049520251795135,
 0.5587552661178103,0.6368542646004196,0.21313364294194692,0.36065720854421,
 0.675893410735739,0.3403114224296503,0.5936346494909035,0.9934507480802062,
 0.8655726458810987,0.4936095767420071,0.4721027760069749,0.7979775444444025,
 0.3698688483738146,0.004468793108794666,0.8865037430628624,
 0.11249060379610332,0.7211428829814813,0.7103191423693413,0.6564023968859204,
 0.9170535598507163,0.7298913658319852,0.6607099019089362,0.2742303956520723,
 0.47228741857439605,0.10556658238505201,0.48762279139470277,
 0.7858790721425908,0.7299249330320945,0.10991039723912555,0.6899368686291985,
 0.9274228721108092,0.8201951012748194,0.0770832516560157,0.3710100500734961,
 0.6563430251245045,0.22715409209979764,0.3360763230126943,
 0.23654460836058222,0.9431751020828338,0.2086667995429583,0.0632078499886134,
 0.09912887779356772,0.5580633109241845,0.6220887854983289,
 0.38677600293141756,0.9965290021125841,0.3043524155415006,0.8304540012199235,
 0.6124179563103455,0.3704248588628375,0.2563666974127049,0.831814889450076,
 0.19977909104599856,0.709615621970645,0.0032728263422557857,
 0.10256209869047939,0.2575768709778934,0.5211290502446125,
 0.019427647203418386,0.5590579302536777,0.8056540461010777,
 0.9044396037225295,0.7732044784767581,0.02536488304055995,0.2168591800138211,
 0.6951167955710852,0.907799798874861,0.407604310539312,0.4493379329291297,
 0.09060308145969587,0.4945780825115248,0.006252083220880222,
 0.09596470350952158,0.9213057188788479,0.3532728908530457,0.6103307289230612,
 0.5746251987653429,0.8542973414280861,0.28556373300861815,
 0.23168220334300915,0.8687513580643994,0.7160686444710271,0.7449260125035136,
 0.3450686310185562,0.6532441896092858,0.14082009259869022,0.8513592999100392,
 0.67143425900678,0.07771750175685133,0.1040710851014941,0.23572950246444158,
 0.5250508661648693,0.6736964190657079,0.4336447116689407,0.2832656780036177,
 0.21805546839702628,0.7350947296775443,0.5586344230271565,
 0.22555359599507707,0.08872317925559903,0.4421422673051487,
 0.4905025318386853,0.1928144115789685,0.06428370068410816,0.4657067484176608,
 0.14150326511920852,0.37606713861819674,0.7867164682792032,
 0.09884507275938259,0.3101509775339828,0.7560648101479354,
 0.0015594242730319063,0.9208720873862889,0.8554458615403289,
 0.5078526408416306,0.307381568357589,0.41796518320195053,0.25460597287162645,
 0.03923752797276048,0.3984854467512364,0.38951265957726877,
 0.9664799410591584,0.6028263197068965,0.1480414858307737,0.6134108994135826,
 0.679299260741562,0.8121722873786279,0.19638287879952399,0.5133120363390069,
 0.35986316976358523,0.5662166991972525,0.24826513976331144,
 0.20635909398345575,0.7704383398291006,0.6825736111053442,0.4177249483640795,
 0.18966238651491307,0.7159584387562989,0.7682930050194283,0.492359850131626,
 0.4081465712061202,0.3486247544811032,0.1821673070283757,0.7745802683883797,
 0.5713466067592593,0.3641206583983042,0.5737113470448353,0.47948492019296995,
 0.8273887028670429,0.7719275405047027,0.5776090811471137,0.15909821220848597,
 0.7443080635306654,0.6877093825074159,0.045431844998442994,
 0.19828229853672108,0.597177073928987,0.4973363156366686,0.44293845529473863,
 0.6278030810793229,0.052231723327974766,0.10298199360296945,
 0.9500657195288387,0.44835259038621633,0.591123036175007,0.9041062193264491,
 0.5102754033982895,0.16336256813367278,0.9246027634575389,0.7626704987739606,
 0.06957866323885409,0.370310551983815,0.8395177352188608,0.11202380371517628,
 0.7213000681955908,0.07102815254236594,0.20411409324397123,
 0.22806140928938046,0.3634934035723669,0.6615471691977363,0.4883046861834188,
 0.5689764615375891,0.8412588695538193,0.30468207317854623,0.902340385157657,
 0.36046663823345826,0.4619729069851085,0.1958088789665916,
 0.12883137923296673,0.6876836788002827,0.9600540601209482,0.9741130216576502,
 0.09269440698814968,0.3203952161967214,0.22788269338409062,
 0.8324830125105884,0.47263166973244786,0.28434767010376516,0.799634612303952,
 0.8694913392126278,0.6155411050890869,0.7503617313969094,0.5532796955766791,
 0.13639232725225525,0.07843640077670577,0.5036844786742856,
 0.5186575601194768,0.8147235511462588,0.27704971121750976,0.85569980971978,
 0.3133924518496136,0.007361218686657289,0.311950873162795,
 0.30757720343578965,0.3667022972019063,0.22646180537403193,
 0.6416427265206092,0.21328428612922656,0.7874724983694817,0.8580348390683304,
 0.026122570096262177,0.914891470410062,0.7063524564708278,0.904707079111383,
 0.9868274625986689,0.2532662292980825,0.23980527107167826,0.9439218540621336,
 0.2980520608788908,0.5782135385337441,0.22949824555021991,0.5291168787766743,
 0.7837791813965821,0.5881194869199875,0.5436964242487726,0.21261211550514492,
 0.40518340217211357,0.6391041465208142,0.695598037927881,0.225639847046063,
 0.8737943559301814,0.8529493636239499,0.7875715958586074,0.4370773137138273,
 0.8738025245209249,0.05962243308367976,0.5105177234155419,0.5898670549685179,
 0.8104883884018613,0.5860636086905922,0.11125914931906644,
 0.40761042320158825,0.35849552828498077,0.19310651134176138,
 0.16607834146960443,0.021987985242898245,0.7832263085368747,
 0.551271009267074,0.3304681885408024,0.7541853466897601,0.6960035376651597,
 0.7636162455850115,0.9388170755017532,0.9778757451982627,0.2287396262438608,
 0.5815782960797932,0.18543320754574488,0.2586723151800332,0.7248565028654733,
 0.25285418917335467,0.8502155547286695,0.20526433062657534,
 0.6952790819337119,0.8776028987708039,0.4220762569893809,0.6798141272073397,
 0.8341162008797995,0.6587731407022354,0.16436379265983925,0.7893866461600152,
 0.7907889833482197,0.01862396134282651,0.7519564066414998,
 0.04216658873458856,0.6725463304180537,0.1708564657787286,
 0.36008854705548177,0.5497613981035792,0.46100293700832917,
 0.9080953637774689,0.8519005678039461,0.3674536096202131,0.9115473034485982,
 0.7254529548870705,0.46719928662699806,0.6098977584440316,0.0710611747121962,
 0.33191261647791725,0.05792555459736337,0.8844490340158117,
 0.11196211233017939,0.39571414105649527,0.07968684328540265,
 0.6317572405828948,0.5499679527946756,0.07416002508891573,
 0.35865053940685954,0.4971998412463078,0.6983813492582054,0.6625684069456573,
 0.929944929042033,0.6141069115278851,0.30226699908579957,0.575836873809375,
 0.1664585075921019,0.3586292031521948,0.6928300229184269,0.13906263888567838,
 0.197060244708675,0.007960906735153817,0.15393900721899645,
 0.22050797879225958,0.19694566347428688,0.28501736004474665,
 0.9935190316677376,0.5268916341168807,0.011736914100315632,
 0.5988729797567017,0.12332143682794339,0.6154187509208282,0.8379954987028839,
 0.13392415312428618,0.9392285679529638,0.975349785195868,0.719242457601627,
 0.02781086246124309,0.9132600505677508,0.46370421162154773,
 0.9642311991099095,0.06660171743150534,0.35723774263780184,
 0.10178400051138836,0.6059495967454402,0.9587531283083238,0.7901222903843732,
 0.3896420840417931,0.575366277373608,0.7510437861577477,0.7671035270434164,
 0.3534809634438151,0.2325252286915066,0.8739823740060143,0.1643438765493368,
 0.21682990366441468,0.6451986186514777,0.5436598847979522,0.6666080851584809,
 0.22526351966150027,0.9307692491913482,0.6727642296697793,
 0.10040748236130004,0.06932014222443594,0.1985818278686834,
 0.6073693092791257,0.5621620355315121,0.7708240227370262,0.07075617214521968,
 0.4421805383391678,0.6977931338055496,0.3808514729267939,0.6396085754560703,
 0.6925019544977691,0.4020402053973605,0.6934130895634818,0.24203080540525224,
 0.8750568787841836,0.00567448815633842,0.34366053277081854,
 0.16524513538489538,0.3002069361272859,0.950783936682545,0.534048024516585,
 0.7607710031087577,0.5592589371032289,0.13626749305593222,
 0.27610793032608094,0.90929951113231,0.8276690151097095,0.8187003282353901,
 0.9356884505284008,0.3005999029288279,0.5905986979545155,0.7461848364890813,
 0.9249916708740977,0.6358220468211015,0.6780457658560505,0.7666281883042915,
 0.4382591428391588,0.9046839555741168,0.5970910832405064,0.4099005368981766,
 0.8786554479382044,0.1714621823698278,0.16979501833748722,0.2945154295563086,
 0.3110352527077287,0.9023134519021314,0.9726401665181381,0.28286868931591524,
 0.3112749869153114,0.9075221729940085,0.596406223625289,0.1870278597042695,
 0.016535827457958563,0.9623179373372585,0.1881531324237451,
 0.0314164616041861,0.7825805147898472,0.06632777102898668,0.8061028880778276,
 0.051039472855110346,0.802682887712678,0.8593163366780248,0.8023698379720988,
 0.1841105327776743,0.6543511348121323,0.29370238828544104,0.9576738591952085,
 0.6046699463980959,0.027313212649334062,0.9406926291777212,
 0.07705198511999423,0.9255013805295058,0.8563019199322266,
 0.17752658340520022,0.8585229816695895,0.19580401349067778,0.153921782825875,
 0.3180665331815875,0.30249270503641357,0.02071512671499176,
 0.7050793504750216,0.8218421854619091,0.3372064734251252,0.3636868489961823,
 0.3912233612262136,0.6726889746075168,0.29971587778223496,0.1667562713702968,
 5.365562238330845e-4,0.5015969214970873,0.3923296425415028,
 0.6079973351592902,0.798969042442493,0.48883841226880864,0.5896773394305015,
 0.9633682931942091,0.6567610125448764,0.5988466262643126,0.5367252390998115,
 0.03997583917347369,0.13984065579483163,0.8041994885276333,
 0.08632122519773089,0.6269942480577164,0.8552412159562256,0.7545898215435993,
 0.6433303583627298,0.45438959808431867,0.5220284085196422,0.3742391278613385,
 0.11884631469158258,0.4334695459506134,0.2414228701290152,0.9216997641880704,
 0.18473806537844573,0.9199723130075692,0.014269616960108089,
 0.9368946309732016,0.5876821832738378,0.20563809414931844,
 0.33445893672307525,0.8473291040542027,0.3686916242746856,0.6310862650196145,
 0.5661682668952499,0.27921329276545903,0.7504758283051739,0.6204556025251119,
 0.33280212840680556,0.29439351434103755,0.02629463152457967,
 0.9619331209962726,0.27016916528160295,0.7823895916117299,0.581659751439867,
 0.2836629143258479,0.24991118564745007,0.5100847295746882,
 0.14744179455488626,0.6950593506768079,0.03509862299075017,
 0.6572973111916248,0.3789359262320612,0.47128363350717717,
 0.18241643268835694,0.7676847644967798,0.10202286711163921,
 0.9411336180716812,0.3756420943678034,0.7706521263339776,0.8636547951915946,
 0.4712536412251098,0.24618431694604992,0.15493167378985773,
 0.2366426196588397,0.03529577704696407,0.1071221576539515,0.9863400681912731,
 0.6745443023818837,0.04306882120296529,0.420717904928007,0.8527437922997679,
 0.3946732949308571,0.43678747140348984,0.3680685420249512,
 0.16179372694150695,0.11659332224397079,0.5910292765764305,0.480746805342143,
 0.673223874952073,0.43825621928429137,0.9154287844452951,0.11216032333720438,
 0.6065857037407817,0.4654989278436219,0.8857385609364612,0.2768290924744162,
 0.5082202357170461,0.4207650068760489,0.8333767530317191,0.26074377234669377,
 0.4613450333743717,0.800664174712761,0.4335368279564805,0.1420048610287329,
 0.564531935528325,0.4991003614742118,0.826147881051679,0.20003350193060865,
 0.7259311861345552,0.6424547064541429,0.15906388029699248,0.7393256483026502,
 0.8933251582521198,0.3811716040100165,0.2210328607677048,0.6988717589851099,
 0.38432858448787377,0.9971169011701897,0.0982946649761659,0.6476312146682447,
 0.60392167699411,0.08620407802054098,0.3382692962937117,0.7193643386422183,
 0.953086454786119,0.7701650788402693,0.036643599318016085,0.5753114322270779,
 0.5377187330781191,0.9112101497815689,0.6842564749883241,0.09157912216790076,
 0.755842429536093,0.5259068194373334,0.7205419252193563,0.40021337304544946,
 0.4466268505411005,0.28290349509493695,0.7111009624222313,0.6547816912489597,
 0.07748127679111072,0.3163418753673308,0.6738218667271259,
 0.38014048371330855,0.9993388314930644,0.5481650418299115,0.856073776462904,
 0.4752798055438897,0.03661848386607802,0.7201038767093723,
 0.43274010846261457,0.7499834199016487,0.1641877122683001,
 0.13819016802355377,0.4875536406883785,0.3121593825595763,
 0.48723234119943215,0.3152390368197928,0.1611374441025536,
 0.06526212836119893,0.7116376991582442,0.2867591621324893,0.9780766817371092,
 0.8622721873809208,0.4376061484970042,0.2727172820685043,0.723702234701186,
 0.3978474297525557,0.3230918453588928,0.4930650607083378,0.25352670920558995,
 0.499644373904393,0.4096885684581658,0.016808984990435816,
 0.018522975540278153,0.1670352144579389,0.8192665808764394,
 0.5814307996407113,0.01411686972644921,0.7023232746351602,0.8667814038345575,
 0.21980338927457244,0.931023389932996,0.10413625400230131,0.5640200735041363,
 0.25793528376595365,0.20178812146149228,0.6482979404204707,
 0.9786010989338304,0.5952612722117845,0.05933549024462437,0.1602292401780383,
 0.4241942166682571,0.31949412492756724,0.0641100388136755,0.7387216883525756,
 0.8297331197181014,0.26106368048648687,0.10979825815061028,
 0.6513015603228924,0.40363209896635466,0.9417833714243404,0.930125368475122,
 0.6363869056773442,0.8361693473757024,0.3544497073143801,
 0.029361126849423935,0.17504848249510163,0.7371200649957397,
 0.23577722293553083,0.29544534411494217,0.5304653087121002,
 0.1564806242606449,0.1900648373025019,0.09720677575970493,0.5064468460019047,
 0.6215711500353174,0.8273563885946277,0.017327099399648938,
 0.09779066936242697,0.34765802203521623,0.6103665118915299,
 0.9421520317293204,0.7427364741896694,0.640493601500075,0.017852259138484983,
 0.3056196779677567,0.5979808355491087,0.6388760673366597,0.34649172210950907,
 0.8035139164763553,0.37169057635078984,0.26038313230942767,
 0.6460426026568363,0.3865042105936243,0.7665993641803567,0.12164931825080283,
 0.3941413702712069,0.9408325649222691,0.8906881080527089,0.7712887519636433,
 0.4008703412242297,0.8175327281768556,0.9818720319947047,0.8840852176089276,
 0.7270117208589864,0.7621722035387715,0.8724848681973604,0.536301130262041,
 0.4586563444830849,0.6860682400952541,0.5565286277884105,0.6512432236143725,
 0.39971729941416545,0.5159347368005027,0.8588076460687566,
 0.07910736600138912,0.4833435142693504,0.3859281692928792,0.6114780145766233,
 0.1138036711210193,0.4085642862344313,0.0908410137807113,0.544728696633981,
 0.5324718540343889,0.41050952964705,0.13126361056174873,0.44003863097810425,
 0.049944406193208946,0.5200694299361429,0.2983648166359981,
 0.4680871661576227,0.6162185995873339,0.0020800048000678695,
 0.9381934132579169,0.11537745796974941,0.5360411020540584,
 0.33646448577361865,0.5781768175366472,0.18675077111521998,
 0.9944177990238158,0.014290063712974366,0.7115007967333153,
 0.6219329652788654,0.16387395110799585,0.7796522012577383,0.942804231894368,
 0.2431589531188113,0.13167235275502598,0.41430711349520577,
 0.14643686053653304,0.9196746174751824,0.05382909983428463,
 0.36903582054865147,0.10358680132797726,0.9566114777526411,
 0.045816362318203074,0.9857459127728051,0.1657487686355117,
 0.7345294402796991,0.3293825144871446,0.5500781809308571,
 0.004318559471536365,0.2887327721655847,0.87084657708916,0.04740641447996974,
 0.11201218816820757,0.8342155772376142,0.5192194783637101,0.17670397782921,
 0.6722486329231425,0.4091500719206068,0.2087036610877676,0.8047689506108984,
 0.7291307460690727,0.9352849443326379,0.9177095545499563,0.0816174755809157,
 0.6735057197064286,0.1918666425160519,0.24369230211814807,
 0.45027032374654086,0.6914033309677099,0.8739144927774579,0.4140811060334906,
 0.7252352389786536,0.5757583091778727,0.09215745731654412,
 0.16131601053598732,0.9048435538756383,0.3032434117402778,0.9490286231938116,
 0.7980109208503909,0.9888559879836305,0.605879195549688,0.4909620460607046,
 0.2050883400432184,0.7908058582765823,0.12209060432262642,0.2690340648260534,
 0.31958346205360166,0.35300742725053147,0.19352877719295525,
 0.6755903213039584,0.30854037705111215,0.3773237865004636,0.8156796623976389,
 0.5143454521523179,0.04130770628808289,0.6832651984034541,0.3434843707764742,
 0.23586351979907527,0.4652467066313337,0.03387628924382491,
 0.014562340316796707,0.6078754108510113,0.9418571784672958,
 0.6536599311725911,0.011224308389766957,0.6119367314129044,
 0.022716845165004496,0.14893963658156573,0.0898833042047259,
 0.6739321538504734,0.977259717121719,0.20110339919524767,0.7027340335716854,
 0.8707473243019211,0.2663919503818494,0.9001979779755067,0.4131057886710705,
 0.8540947178837139,0.9696869975215411,0.29993559567227923,0.7204677200495779,
 0.6038607219008152,0.38567709721625865,0.8848657143029658,
 0.20524675756274624,0.5824296903172982,0.617666781698964,0.07142630990369214,
 0.09424093847137216,0.9725321057956313,0.03356079348538277,0.649272181714834,
 0.23708715261414404,0.22775880142587746,0.29365539690288855,
 0.27551281971249353,0.7872527982922876,0.4866392514099186,
 0.17003584445343178,0.5869497368710311,0.1921285641499113,0.5122650394368805,
 0.39323889659721745,0.141192056896131,0.1087905409934744,0.15294926826728839,
 0.7903148324540779,0.354395847944734,0.957800058816162,0.6430994079175772,
 0.006710605528448932,0.19790341104455633,0.18134983723099263,
 0.1996257360441147,0.20394832753614434,0.6739041419612136,0.8184721375914967,
 0.5863062087075765,0.13275870074629348,0.4611666514401027,
 0.23138275559564392,0.5218722265525739,0.47813218112601463,
 0.35563385765937194,0.9333912857743406,0.6473207670734106,0.5195976037637134,
 0.2750010663653898,0.6227170065251895,0.24671538223604927,
 0.43869079393778054,0.36335437920279634,0.3940797120635029,
 0.07716815239519115,0.6308015392712698,0.8013809894514691,0.9588875672578067,
 0.3860954864152558,0.5584567248378001,0.2019093568588597,0.06359693702027258,
 0.47305565564316043,0.33394786111580366,0.5012385644207586,
 0.9459858323453352,0.6376848352467699,0.7508369217631218,0.8228566424541472,
 0.791911766205891,0.08060164273903303,0.17381919718881433,0.9139610117911675,
 0.23721795642092447,0.17946414985444648,0.09149253033535676,
 0.839815968789912,0.6121676847073032,0.26026105271179156,0.3790614708216644,
 0.8043257541141905,0.9486325030035493,0.6847623261510472,0.8976201571412735,
 0.8687370257621528,0.019140543128604248,0.3730939689353826,
 0.8969967198979107,0.6722497089505838,0.9801354104953021,0.5099536935970155,
 0.2313200052634855,0.5448617353751833,0.15451863690315015,0.1848195576634074,
 0.23746339586951304,0.2568107880539583,0.4775699410536377,
 0.44995137695656573,0.3979221351349419,0.22992054574602638,
 0.14939497964074555,0.8872585557225217,0.15951540578703094,
 0.26378968271637104,0.7671025894356116,0.5531300121975931,
 0.19124645934346174,0.8826763655093359,0.3279111014707099,0.2237681755901616,
 0.10201726145405021,0.9465450877724071,0.15316448841521024,
 0.6859801544999387,0.3099026151206712,0.3182664558828716,0.8541127407568435,
 0.2899150176139158,0.507604362967546,0.12369621565777511,0.13050387962146792,
 0.528416137387937,0.6732622754146758,0.9072334915800138,0.8585043759026352,
 0.418337464732066,0.4867503051936696,0.5846766355535782,0.9845999389390964,
 0.544102717313616,0.6934869649058123,0.5739873930646889,0.19402249687965079,
 0.2428460412012926,0.8569325386345285,0.46416090553356115,0.7435459059586795,
 0.5270895198948659,0.49069779458126184,0.6152627649879554,0.8283099833623842,
 0.24214597909604652,0.4121813368582137,0.451475434465025,0.9637190038622026,
 0.516917974902389,0.5216792997408162,0.9758830234149006,0.2267737459282818,
 0.9538823856185712,0.33421624737882005,0.7922025239855632,0.6786529650631583,
 0.9799254164185008,0.17068586245887896,0.23977991554580835,0.404912020329537,
 0.866147016669008,0.354066992857887,0.20408392165032097,0.6374426235877817,
 0.1334689138764743,0.22199101514568853,0.015530643775917397,
 0.04586413928310318,0.36202499946834654,0.4003423811104029,
 0.11689396241361871,0.17196137758321872,0.5733747659020323,
 0.3338102031607594,0.31533867176034314,0.8213309386897105,
 0.02209711191556163,0.017993670756661584,0.5439623377215349,
 0.0791217226419858,0.10916725410918493,0.381274373667446,0.969172719137106,
 0.7437743453521957,0.6794941585845806,0.1352240307305257,0.4538583898395352,
 0.646153113229265,0.742567017617246,0.3673063659979249,0.8060167352223311,
 0.9341721575988774,0.22763540952604555,0.38586738745202076,
 0.9318666614848481,0.14019616453614625,0.25257786348683264,
 0.13559598123887362,0.6433226995423095,0.8345398451493153,
 0.026521633637973663,0.1481012994255715,0.8462819549782561,
 0.41799731241113447,0.5056712407705689,0.8998344487033978,0.317142261069448,
 0.18587429844037728,0.6970938141937779,0.6357637521515954,
 0.37044520036909545,0.7039664624872047,0.06393364779383137,
 0.5025662166846968,0.25700233634240244,0.12530858548649726,
 0.6256649805198727,0.6227569204346868,0.45327787542682796,0.5131995775417145,
 0.32886522855376643,0.9785149316160013,0.12154073395221776,
 0.43026365385861176,0.08597350546159133,0.46641114637356695,
 0.5113242954100771,0.35034481676258633,0.4860585495373494,
 0.25597627295357916,0.6622364894997342,0.765170354456184,0.9050836475275332,
 0.41594467242405564,0.232009589297415,0.17684198894683034,0.3870617071850522,
 0.22914194739993954,0.17284513816981562,0.9003765020964003,0.294940644377093,
 0.19652527238075068,0.5791744739018435,0.825507718031457,0.2095112387556738,
 0.3007223245032913,0.20186954587203199,0.12951263679596325,
 0.9387973141965612,0.49887797676778867,0.49177940374718776,
 0.039793451754791875,0.48406661958445796,0.41122013431165305,
 0.6863227823873836,0.16842079574862945,0.8485690897157007,
 0.02930479030311195,0.7643668080891249,0.862606739872585,
 0.052150262478682374,0.5481196132368856,0.11289176271294699,
 0.3466373299264185,0.2529033767921177,0.45221336748714047,0.4217207447688356,
 0.748608950990437,0.24231638605327643,0.08925470971524985,0.5318939139306388,
 0.1826811343098882,0.7935216867412218,0.5915407882948097,0.7724498480726122,
 0.5954598978362108,0.7104824830019143,0.34942682306780004,0.9436821253140324,
 0.2560506800408311,0.021315002986881715,0.9159584453328196,
 0.6299613363543383,0.30154280791011656,0.3326738751561674,0.3328482178743676,
 0.42378403629760397,0.4199016580007019,0.8432957505738761,0.8971308577337316,
 0.939234268894354,0.9999808788624537,0.9587010378469375,0.09411221064219444,
 0.032505138392534216,0.7928160919635852,0.9583751951438545,
 0.3015062876095067,0.40447310374174295,0.4398779623535868,0.525632732891364,
 0.6580730543099519,0.036407855206147643,0.21724395726687973,
 0.9690405123385268,0.3431431192592974,0.4718268702026207,0.8172776466253528,
 0.25759612894389594,0.7971333803746917,0.6350271829177632,0.7801227047248966,
 0.9121406339670215,0.07146365991179127,0.8676699890275569,0.3319636327231129,
 0.5220414550964904,0.8914334069487294,0.6639781504187416,0.5030020474387249,
 0.10828994831906813,0.7042120964926222,0.36849352658398227,
 0.9235069295671102,0.334304664771057,0.5256808213784809,0.5469048881099314,
 0.001118815418238217,0.3957681361226525,0.9351712199079442,
 0.22635170800073479,0.907529490637451,0.02651413755449472,0.8793482375356493,
 0.8538770263808823,0.8543910068698602,0.7807070870119047,0.8840423652357202,
 0.4033735295187134,0.9162799639338239,0.3835025174838782,0.730993827911521,
 0.36733019728457783,0.9312074863208427,0.0054360123082616285,
 0.17238895500824414,0.08723802943863401,0.1772223575824402,
 0.9890347610818434,0.6169852605135308,0.7085542253406256,0.09937428248088898,
 0.922690199976856,0.5421796970937686,0.04036180950547397,0.26206525488179255,
 0.6737411242697046,0.4195054781141171,0.2790720387573222,0.853573109026363,
 0.4857764032929559,0.9636789689255563,0.3633043612120682,0.29355831230117024,
 0.516052066831183,0.3122197164557984,0.03545936399992278,0.7426003438931373,
 0.5202298586048826,0.5558751232883462,0.4080258488099032,0.584672423244958,
 0.8133050251823479,0.953639927519109,0.912074426093158,0.539419663585834,
 0.2444799919393429,0.9146324729973202,0.7429103468772957,0.7409514909049879,
 0.9681004921384607,0.8492628807483233,0.27836135637535897,0.1037541421232302,
 0.3233877252296149,0.010930980011998659,0.4172413599044479,0.208803031748502,
 0.6281376590788993,0.05942908025904903,0.44533020753131536,
 0.9135430301999842,0.28050554024681007,0.14592882722300737,
 0.3376333870095647,0.5784148452377429,0.7894203675015449,0.15380975538780062,
 0.33936453557607327,0.20858164923120293,0.35919860410139237,
 0.2826151406681596,0.8103078437306745,0.7185732677349083,0.7018928669704819,
 0.3507318590903028,0.7247938272156803,0.2483552588112956,0.21752712641799732,
 0.5014253444914184,0.3887732801144226,0.5369122632017318,0.1049430267385365,
 0.47425364379152657,0.23507675053596655,0.20949336649328965,
 0.8863027335216288,0.6073050795644759,0.4501190620894445,0.5488626594849731,
 0.21838049103872903,0.09572965024166213,0.3505150131060496,0.986466743087387,
 0.6938445700113758,0.9944699510564561,0.23465045775656657,
 0.39081524519526645,0.8761677113125171,0.02614621913767423,
 0.19425494980994706,0.40989678881256264,0.1738863459456944,
 0.1327570782729276,0.23732532778683058,0.6062753150749891,0.1314472933183508,
 0.039329302354485396,0.05044812816594946,0.9758965362039557,
 0.30052668351847267,0.5740490314799661,0.2509447614361129,
 0.032068466945110896,0.7750567161694173,0.9322090756349097,
 0.16219678257173353,0.8498466907836004,0.7627012896815214,
 0.34684312016062746,0.5568724484657172,0.44138040182962524,
 0.8570671204062482,0.07124126195610392,0.723790059481843,0.547035162502866,
 0.20425116807132038,0.7987175709290327,0.008484544710749553,0.30235714881161,
 0.8459258470220313,0.013105580939882788,0.49802476963565745,
 0.6458637996312352,0.08002832446113972,0.1303684043642206,0.4500248589022007,
 0.8821276168419643,0.5018604915554834,0.3628625763960809,0.8601262544809299,
 0.10213310994725733,0.3840495554932708,0.7291095696324141,0.3345955005119158,
 0.6458166321330202,0.3171081726313145,0.07125896084986272,0.7421053112899939,
 0.05115039737597504,0.6909173793810237,0.9486219631917739,0.8900425862519061,
 0.261165166423344,0.26686607580211763,0.2614989476830739,
 0.026630156774178126,0.25266214056131253,0.09427140379012788,
 0.9438244572887107,0.4518569575336351,0.003906310568898341,
 0.04096590593938643,0.3845570640382633,0.002273724648445552,
 0.5523229009673042,0.6219834342487207,0.45894512041025615,0.6178532641554605,
 0.7134280012014975,0.3402729773819446,0.07149366135244783,
 0.27853629857636375,0.020888168709785493,0.8854534196955173,
 0.7339111358020666,0.8554568110794947,0.34633141276189483,
 0.24554024650336204,0.9421032778204165,0.24147979340805836,
 0.06229382032191777,0.7323594444927324,0.047305475621616244,
 0.9353564475922467,0.7004352630891433,0.010439534665376504,
 0.7372729135487225,0.7419595887899164,0.9728600608204427,0.01665180053393145,
 0.030230664165970977,0.276269264753197,0.37286401028808025,
 0.31190768585496054,0.3887265558825884,0.354354329727123,0.9031333771579032,
 0.5825007095144992,0.5496942202277846,0.2688694098465695,0.6877702728748681,
 0.6449286034563877,0.8904614776892631,0.6515677985507752,0.39711446089489877,
 0.7596827794931398,0.7295465246692642,0.5430089536487931,0.9797609898415205,
 0.9588252484238408,0.03166873081039723,0.03463711791461099,
 0.3319422660817275,0.675347287284784,0.29048250044573154,0.2690840360570159,
 0.6925410059062452,0.713708763704191,0.2001274901268595,0.052682431160157916,
 0.032611156101800653,0.6513925797533756,0.43676992075480114,
 0.6337307141934513,0.5626466027803324,0.619596746304858,0.7607532340245,
 0.7634476445550368,0.11451537731632833,0.5742045064468063,
 0.08560136211636815,0.9183000646581237,0.7347708182985442,
 0.03868692821971831,0.6548115457467539,0.5419913407550234,0.4287267825303258,
 0.3684837386194757,0.16970623843030397,0.14402455107994128,
 0.5089271044892629,0.6376045235225867,0.7999421399030223,0.8037897173931805,
 0.6465795285959551,0.42191175555616955,0.10080328145099904,
 0.12833126809427353,0.52451629341712,0.7915953853641524,0.21801106912790735,
 0.49767962426010637,0.7205809972119276,0.2622377643465207,0.8865134938194068,
 0.7857367573086347,0.967203140880279,0.3263285485467273,0.9917643810228123,
 0.47879349226517887,0.45426663737204764,0.9103781778192169,
 0.028919132962965666,0.7596912323637734,0.2972810652200262,
 0.0010425179965117826,0.27535326817484973,0.9661252829970796,
 0.17889838446315565,0.8309802972876869,0.011079693267439628,
 0.8137595388154956,0.923196894730607,0.9626980428445192,0.19705404213506317,
 0.18431537543786236,0.8918524178453697,0.8523726894956478,0.9762034124718513,
 0.715258535070533,0.7062924137374513,0.8922233939798639,0.11149538448947349,
 0.19985546254139885,0.29269488195566995,0.38257510586320587,
 0.4173525808259765,0.6241525654260135,0.4435360310635068,0.9165242637912028,
 0.5828340156542947,0.2599942530196536,0.8756452601863229,0.15658763187643054,
 0.6678644305823945,0.5624705078990192,0.16274561668937315,0.9458285583684349,
 0.08764099992723706,0.4531056431509015,0.41719921847666774,
 0.7811620868098015,0.02145601831278887,0.1774306953628786,
 0.10003095343792823,0.8162434422231659,0.27229444199574093,
 0.4444432353194001,0.29094545352159173,0.13643465750261397,
 0.9571477293430667,0.8652045262143168,0.37053724799393584,
 0.49579589200884266,0.25820990578353464,0.8731528175878412,
 0.2022134407109043,0.47674232023964225,0.7583748222649229,0.847962374085243,
 0.04039714259813021,0.502749496718128,0.4757173297096373,0.82414115448293,
 0.9108074570889211,0.6346353582776885,0.5263226125566591,0.2928665129012612,
 0.6977953826277055,0.6780071352853622,0.31922206725844626,0.5963574716697793,
 0.8273299564220343,0.22819394979619523,0.12447952478214508,
 0.28473424666070823,0.9712966928984017,0.29676000773194255,
 0.8106985082424825,0.9605230263640655,0.3096127226685177,0.16177037939010996,
 0.5623066301932184,0.2859486001673648,0.9613480995867791,0.7638701899120683,
 0.2530193231663036,0.6228691320309732,0.5861984558964565,0.34607023786828883,
 0.020468643818825405,0.7952342312502072,0.19231440772046537,
 0.5541459755954765,0.02773821636242868,0.22591573096028306,
 0.5081936662669487,0.033439718969410004,0.746299991823631,0.3179002868739179,
 0.31048326092095757,0.29454324380230457,0.2994448884498214,
 0.19205689328767894,0.1324422313603707,0.29758862475892145,
 0.07883599575264644,0.30058167035751127,0.7654405199837532,
 0.3409355578037867,0.04630256848442782,0.07407347671403719,
 0.5388762989421236,0.03395006183040217,0.5527028203186706,0.5493427106965869,
 0.6421781929529589,0.08494551407618878,0.3117638758491711,
 0.020204955483664633,0.20995708617876185,0.44233361619011635,
 0.9277265617871464,0.8841478440335817,0.8261541943754896,0.3286063211720698,
 0.7267799593543853,0.6149522453872892,0.3616213346667052,0.2894165804304025,
 0.6447548262518357,0.7994968107086838,0.5891763959017379,0.49740914669370007,
 0.22713078900052242,0.49806835243923386,0.8195961971187381,
 0.6523563258542496,0.9812508209355335,0.8158509381179355,0.6031499095038504,
 0.338126869267692,0.10622225307402766,0.14932303205031738,0.909665628812057,
 0.9172000518881303,0.9594575341511264,0.3764553748885784,0.7460006927927032,
 0.6108165405340242,0.7741501307238204,0.749080582173389,0.2616397963860846,
 0.1663500274727926,0.8091536438710103,0.1406978462717049,0.092880707631536,
 0.5593991381290535,0.3793263419802929,0.8640442547301151,0.16680168011426866,
 0.5702833765538396,0.33729194243319593,0.8810412953222576,
 0.07426492194145873,0.7578150605597801,0.7258395649284233,0.4999628845870464,
 0.35142456422752655,0.23875893329381148,0.46709609512607686,
 0.35493949948290426,0.4833079967404842,0.7458631327434887,0.2536644076836455,
 0.7723744209594312,0.4759572594776651,0.3107196866000117,0.6107485202494389,
 0.8892450030791684,0.8595356944835368,0.19917659582573688,
 0.33076262868504447,0.7167854145886303,0.4444711742339994,0.5928112106607182,
 0.575606323497544,0.206832231683441,0.7311234100741436,0.6907125282728231,
 0.006456442505678783,0.539123186413643,0.8065786114987417,0.5185356447651279,
 0.041465816041304726,0.5080083032181291,0.05338603439424805,
 0.7162466349089361,0.9740444916714767,0.36511941912557444,0.9734228299931394,
 0.46245310479112056,0.5842338284682598,0.21340551595572044,
 0.7691599636981546,0.6228489066440462,0.973599650471614,0.33545282653174013,
 0.8965488427812539,0.022982082904684953,0.9062173714760515,
 0.0994774869225532,0.8103885156333291,0.7498543312721279,0.6095489344761608,
 0.1394971442735874,0.5988167755121785,0.20889971849499012,0.1910050502404106,
 0.642205394396366,0.6934559980029997,0.09487259946914539,0.2018612887475334,
 0.4714718740464907,0.026819921041491557,0.5305530312851405,0.427945613075277,
 0.6517635568161267,0.7095807786442606,0.5463660295740154,0.5068729880989273,
 0.725147705987006,0.9196093390860061,0.2714264194735675,0.053074380475784944,
 0.21663546690284508,0.7921671021937529,0.24866280710599076,
 0.36483397408424656,0.9276405472630608,0.09202414296418304,
 0.8129413066411451,0.93821260974374,0.6045779867661858,0.48412791694073454,
 0.13593041889195834,0.6994852463100464,0.13502440254155146,
 0.7870075842752369,0.6279505481513025,0.8154423911833744,0.8465992858478552,
 0.061218012233935304,0.07230386552169088,0.4846827798158909,
 0.27941773669742687,0.04459414028315023,0.40006429823857603,
 0.4199807714655832,0.48968682265984365,0.0679151493884359,0.7989122373162614,
 0.07749270035188993,0.12801654308751753,0.6159008463731156,
 0.9805691318803238,0.14035731205702295,0.8864563716565818,0.9586735756861331,
 0.5937951732278968,0.6319094147741924,0.9377835772432501,0.6418494213473738,
 0.7156317396099904,0.9236478979843783,0.617766463738086,0.6494079913787665,
 0.4105987160909118,0.3340365500001292,0.5971137165801154,0.852684725835749,
 0.9038361786577307,0.5510130966115215,0.239713524615922,0.8515342032980044,
 0.5858556056022846,0.9576293724786028,0.6161180538582794,0.27026778463386214,
 0.557618139945369,0.5395319166548883,0.5099016479085325,0.44808299698399967,
 0.4730547327214485,0.5876446768123377,0.14361992827509618,
 0.07273480057956372,0.061423825375303776,0.45328861119008423,
 0.3854110144382461,0.7661022130445615,0.4840428104775706,0.9978672092173153,
 0.6922436584124946,0.915847444486516,0.359859593424662,0.18540409050326978,
 0.839635181647412,0.07043964536615688,0.4623314190546307,0.7647519802681093,
 0.9895341893988856,0.08267362529976907,0.7820810352632649,0.7084880867141737,
 0.16513621785546362,0.9049972401911739,0.28772253345643617,
 0.7744411045541008,0.2593420233400152,0.4852383217891101,0.8514293000375708,
 0.11711146240216153,0.20112147952978698,0.9577891446165782,
 0.7591685738598128,0.15725320504035945,0.45391193943532626,
 0.2482253061958476,0.8862507163005029,0.7856256879958825,0.30141733409670524,
 0.7754528988116937,0.957851265147726,0.21934601648366037,0.13404693346934782,
 0.3991678419648865,0.9467839258010767,0.336792106216518,0.3743499083140782,
 0.5645303369949289,0.7930807989567674,0.7846931133193367,0.762573364033433,
 0.011814924438034224,0.4607420392700523,0.05435243435119619,
 0.18569003723939914,0.04610811238171608,0.005328529235731794,
 0.3837661164642472,0.0016380200222161356,0.6640655983081112,
 0.0870178857622197,0.9987789930815726,0.6700218432307268,0.8385057990964797,
 0.7585399504265278,0.7033468064075963,0.22105871935750498,0.8049903884778326,
 0.8984458149837672,0.2854489297179553,0.644671416327932,0.5318760226884066,
 0.9370212358857667,0.35452498401505395,0.7361803234013455,
 0.43312165650172885,0.6629483912184277,0.9689877265506935,0.7932165857715311,
 0.32774528875521014,0.7406967093488647,0.06377944167735006,
 0.6748857500215671,0.8270580316360512,0.877451993360066,0.7724111759880372,
 0.27789808535268623,0.21957973266765374,0.5326777539828336,
 0.8984952887635944,0.49476579646058727,0.8565200805009704,
 0.13899607980160866,0.18167218290089227,0.8445971500426127,
 0.03613120428379801,0.3050583477308917,0.9528561884032071,0.559566136685031,
 0.5228116575476196,0.22154528000190288,0.4808204330576764,
 0.10232567087658351,0.6678321248838862,0.14548984862985936,
 0.28760897234260496,0.9809165168289598,0.2700056392406096,0.5300679375152344,
 0.11422534418795327,0.4493351384352857,0.2579315503607985,0.8293394602432874,
 0.2746471209367828,0.07654301868427771,0.6872530950517222,0.9801338598273733,
 0.4933291033433088,0.08536895041700554,0.13418110892565638,
 0.14373492116826847,0.560182748315792,0.8229249728043754,0.9738269549958009,
 0.43621110747154535,0.43231066676148444,0.9973018781488353,
 0.8570089137881267,0.0894251623712704,0.2146600647076088,0.7427613924739482,
 0.61874963891989,0.41131190748188207,0.32394581446796666,0.38346083479139537,
 0.9959394422591966,0.7709135197791329,0.24754493151723156,0.9486352472025835,
 0.9682239129273233,0.17255074060306275,0.7420859409279683,
 0.19843741741972898,0.9676037864033284,0.5045326429816294,0.4880813370736773,
 0.007534280844511687,0.9090672968847544,0.6888273836603092,
 0.6489183954489105,0.5899977919202539,0.3761914683354891,0.04057011066362992,
 0.7897524519854365,0.3095841396375565,0.38660758202701906,0.6846055504947817,
 0.43962859015671407,0.7684626903288635,0.9154138859682442,0.2607135537795743,
 0.10971495840762302,0.8166784576666376,0.9535085151536378,0.5083957475430644,
 0.03569704721911571,0.3696390955492551,0.6765426605783706,
 0.22314584506870616,0.6997553891019728,0.5376608610114314,0.990493027520561,
 0.6749192569432144,0.8404148633322209,0.30336082677067433,0.6949999425739964,
 0.05788984764591065,0.34350639551914364,0.8897164124913673,
 0.6135191271055291,0.686649069825271,0.3962214729636402,0.6418720436630383,
 0.5204911933038917,0.24021897047811958,0.6820366847595606,0.2801171480141449,
 0.8679300207570259,0.35872150787724366,0.9407501445143492,0.934441941692211,
 0.18411985417759347,0.2803481476178664,0.014119536969234958,
 0.6484544201895148,0.3304781568515325,0.7916541784078144,0.6518141019544483,
 0.24128070017804504,0.30665730295552207,0.61056149015236,0.46895669294445397,
 0.9583098959897224,0.7621570767298187,0.7630049852678347,0.7458767599386571,
 0.8607099737336745,0.9510021021634061,0.6442690496466436,0.645488765701087,
 0.5234489349475056,0.5720511786113578,0.7208931194431999,0.6984432878305981,
 0.5043080303008881,0.7813473343214842,0.5761026224517186,0.7637738672019987,
 0.10211077207273567,0.4897450633393299,0.06131768574799201,
 0.8808314873037828,0.045256095166432586,0.36116026988123506,
 0.23203943344517097,0.20553512883258163,0.2200822766008348,0.383476068139497,
 0.7587344565414914,0.24607191328284173,0.6569222127763424,0.4809868807453309,
 0.9403226399838016,0.6938733019868795,0.18813072479099568,0.7247713577699216,
 0.38358885513256014,0.1491224132577691,0.8390160435244419,0.5973458066835805,
 0.5929613351884584,0.9171342617910104,0.9753127532339785,0.4412218969638203,
 0.2695294059492507,0.5885857001857753,0.08452987310796478,0.6316538496547106,
 0.4774457280146659,0.12214095498733313,0.8862023035235049,0.4754260790946816,
 0.07276950643956415,0.39563459094357567,0.6266356773896382,
 0.28888767051146014,0.26058933985121224,0.9390514928622219,
 0.04755492697927888,0.011970162830909237,0.7916187789261329,
 0.8921397933917458,0.5975376119931519,0.5181435718507995,0.9263301592623457,
 0.11706364859610474,0.7676370656539477,0.43339225149614014,
 0.6599695507884789,0.4232821343378914,0.5829340509204612,0.64259766812527,
 0.13257590780409212,0.9416263028486009,0.9396635661376825,0.2665209532664796,
 0.258312967378849,0.5070178388995621,0.5395235555060501,0.4210637592786277,
 0.8576130750963569,0.4344395295500475,0.21321318572033898,0.9764319110704461,
 0.2624989824197438,0.3265978303993936,0.40179052044618613,0.7434206984723897,
 0.9369035288269235,0.15995276223731636,0.38560225116162794,
 0.44053904325230786,0.04807279136653353,0.1959010928111301,
 0.059867798485657264,0.14906661395999643,0.20372494459774493,
 0.782422169891784,0.5600427284707947,0.2311119465646596,0.5280318930804054,
 0.08414805619149457,0.14220785388597923,0.5275585788266083,
 0.5523345250365168,0.6400691430468186,0.6711196589416216,0.09641437370423322,
 0.6383840327346522,0.9929381299572253,0.1270939408036258,0.734448579216032,
 0.1399727427004278,0.4060759936483974,0.0861305085180365,0.2289168971160197,
 0.682458368279721,0.4915459050743718,0.7061972112856003,0.4911080718799704,
 0.7676371021178955,0.549696070211753,0.93343639846786,0.7209936913368651,
 0.09484550940730962,0.5773758604253234,0.1303858719821963,0.4689005510338904,
 0.7922929069355014,0.2642449547485044,0.00967699881471995,
 0.42173051054027655,0.9015341771005296,0.7267205556506043,0.6084714629496699,
 0.7577173261643713,0.959879131878743,0.3290520985172628,0.39559077973846446,
 0.8599444348112589,0.0034036474092414837,0.8377044006181605,
 0.4310118516328506,0.8731373227676009,0.4654312998439135,
 0.007803582774529194,0.6234271014057743,0.38374465934981217,
 0.12764752122919132,0.9724738878224248,0.14990013036810612,
 0.08787288470010934,0.09736649087668736,0.9282515680993575,
 0.20204452002256545,0.04152916083371205,0.2752733720771121,
 0.5667068359852131,0.6201224909551725,0.36097800679820435,0.0909842654880102,
 0.2656016442950444,0.4241030148538254,0.8535013493711486,0.8726483839294599,
 0.6509498315397745,0.39640246533976653,0.16478842592763787,
 0.02396289434778931,0.48956537711389836,0.4505440999646728,
 0.7999449032032179,0.30608754714050257,0.7221289495327985,0.8265018468197659,
 0.7246490507201568,0.32964854804751076,0.3907923075400527,0.3854925740766384,
 0.6087173882548264,0.005499630728372562,0.40741148382353476,
 0.2756237551009769,0.7635940627614615,0.583714122807318,0.01594966665681752,
 0.11123201401815797,0.009542286488944174,0.5599809316268848,
 0.5790094611136885,0.13884666435884685,0.46786706799585065,
 0.7211964874900639,0.20522621094282723,0.9256501378094792,
 0.48004896845003686,0.6328356100898697,0.08466343305813062,
 0.9790958678216461,0.06282350621469324,0.5373051652477415,
 0.24771656619403226,0.10511880520155037,0.6851171866664789,0.472302254483296,
 0.845309143883033,0.3220966580842419,0.5616827665333557,0.5378282158167179,
 0.714641090428305,0.4087440739152828,0.6342438365195243,0.8609196759016312,
 0.4669737466985743,0.2091365023597993,0.439664621132527,0.6654013385300286,
 0.47928703294695585,0.3214133883510455,0.20612990123644304,
 0.21727246059366168,0.2546822248283743,0.14641630987505583,
 0.7676767338955987,0.06410121110719302,0.6906217215945978,0.6109175933676528,
 0.153266365667022,0.8145801800973165,0.9554310499708089,0.8364935359366213,
 0.863200792256076,0.4352918505216756,0.18828488934424947,0.3202855012498417,
 0.30666751027893857,0.19849980662982958,0.03974715223255598,
 0.7668529182112173,0.3200205925900266,0.2137048460789912,
 0.056901940720334054,0.3926071295033966,0.7715586679728381,
 0.3296972623486788,0.3164919882337527,0.6164244311293959,0.16167117768351125,
 0.12213613627954689,0.5429639441661768,0.6443231982673222,
 0.19743119035711632,0.5039053226735569,0.479079222965341,0.5010738150641315,
 0.4181497944732733,0.7967096305428552,0.40702996965172333,
 0.36018241160477404,0.9228731705735629,0.06966142250109308,
 0.9532900800594274,0.4347950196754423,0.29220500358796897,
 0.35399707044337936,0.7586871029454122,0.342576225238458,0.3998900756592656,
 0.441376162861461,0.01085007045125097,0.5197893837817268,0.990896214660827,
 0.21584042551811533,0.943102375341611,0.6047544994641119,0.13454338272126443,
 0.18467656348689365,0.5373102865728138,0.2928347975468214,0.7748244469894991,
 0.17286749151718594,0.07515908659592152,0.7775071250291089,
 0.6562546162310622,0.8077781357674554,0.5500275629064528,0.8332145173892527,
 0.8717277344836198,0.6445468766400012,0.46570136858551003,0.5367529628199471,
 0.5236255135233482,0.9023331478279266,0.4569240676475408,0.19148406333057577,
 0.028639631098085205,0.4245758406799557,0.6533897323458668,
 0.46460736286581494,0.9731935110648001,0.5577062693489498,0.763640015602951,
 0.3281741260490547,0.8416402107134163,0.4270076403584322,0.3387508925626245,
 0.6566071745904224,0.07733425354908818,0.5569538932504575,
 0.41224440141175167,0.6735713088824952,0.15676823664886974,
 0.09014010012052365,0.4098936466084101,0.3742097010587956,0.8353958106966003,
 0.6116086713803707,0.39572266945445667,0.6105679862071326,0.6255175102377931,
 0.5756465008957613,0.5131827949793285,0.719802596696849,0.7901973582890935,
 0.07270226998177631,0.2981044127661109,0.5151982990312528,0.9435550560948038,
 0.41186115962969805,0.6222421138183594,0.17343712837372616,
 0.8516197356016291,0.9416265150761484,0.9575435963788765,0.8627246788093912,
 0.7622646049531766,0.4479093854498223,0.9335970187338878,0.6784708866706128,
 0.18494775016201148,0.06727166573887766,0.1889849830825775,
 0.8204453306988897,0.40102223533563464,0.8201877011965999,0.7497860616299541,
 0.43132838617412117,0.7108399413390705,0.7657853701586613,0.1090346371220473,
 0.5871515829640203,0.9177650940774041,0.5262766102675067,0.9875731994455934,
 0.37516512951047665,0.7437614840530755,0.8103183407926322,
 0.08307152415377772,0.19348990234401597,0.9050622868614023,
 0.10450983377734002,0.18446134980366413,0.2319703768481859,
 0.5525508624186362,0.7128857445609351,0.37727642330613353,0.4136464422915034,
 0.6688042975269177,0.04461815955518578,0.09773136303559304,
 0.4128781099675385,0.9887299348948896,0.22154677037545345,
 0.40618183627097104,0.9765515918659585,0.8153355619976604,
 0.21950301131628258,0.4426286508059142,0.04601271189103173,
 0.028482877910065874,0.7674518881297564,0.3387967774647369,
 0.6890711467439169,0.4760422246111937,0.5769344443894124,0.3840304881748118,
 0.7141147718904685,0.6941795946003778,0.3203845455955612,0.48050158775709817,
 0.12457299759153961,0.6903048848436094,0.528580094858918,0.8644617189057954,
 0.1738720672034988,0.9733165437998504,0.9582209065376205,0.5995324473367322,
 0.3873131433419752,0.28495529959038646,0.9677163484825079,0.4853310644330613,
 0.9434559510176904,0.47447025648510954,0.44215743175016275,
 0.13273308828871544,0.20710480998777414,0.14899754396559772,
 0.9392030503457853,0.7310090962804008,0.3252691958363547,0.3729969102646046,
 0.22181105303959048,0.864769953102763,0.36579852418774206,
 0.011507394578450847,0.9455281966248069,0.06720761536090858,
 0.6204627263220557,0.9505969376886494,0.05944560301644575,0.6989174487767325,
 0.4369876331586243,0.6847962366677437,0.26755379141075364,0.1629616542338398,
 0.07154879130021286,0.3591137231082857,0.3544943310817944,0.124127179034904,
 0.2475176248623281,0.967632705493766,0.7700160999957986,0.5682259474933617,
 0.1823398515065856,0.9848933057303149,0.06512757895758359,
 0.43291469082520617,0.23424608341307518,0.7386995293362425,
 0.6820356080944897,0.28234572597038676,0.23146705293660363,
 0.7721525616208141,0.6645434011642469,0.8337459219233204,0.530325162313361,
 0.38578832530123364,0.335803695579433,0.686582505725471,0.4019869323371569,
 0.6117296891422364,0.4430205332602256,0.6288483669523849,0.49436644558613063,
 0.29166880642082693,0.41608338865384276,0.752888033559708,0.8824047059498699,
 0.10232659113471909,0.7129104510360539,0.3319256181702901,0.7357273001835215,
 0.7185827827423029,0.8940799065432388,0.5420954694795617,0.8360624536678072,
 0.5125208017167519,0.49015582967506555,0.12839204388705794,
 0.16427263453749064,0.5217060598729281,0.4179052370488081,
 0.01686605009957043,0.22026200818480735,0.4879122315329376,
 0.2980867555301542,0.9225262819635217,0.10020348763984455,
 0.10425694199147917,0.3674618618481794,0.8072896029375315,0.3905506454514478,
 0.019008051268631032,0.6628907955599629,0.9622239062597243,
 0.9268123766019694,0.614978791174302,0.48418789451178046,0.31368205330938403,
 0.6497316692606738,0.9101632427318744,0.9128727584506203,0.0712994914969507,
 0.14816467559825253,0.059128152943019874,0.08874390892747341,
 0.929973400383556,0.7675682910155053,0.35880574506903506,0.37084567503584653,
 0.9043293797295692,0.6371081115430222,0.7333687947257372,0.7329096161659266,
 0.3185723287526455,0.7077924798779341,0.3171134530968768,0.7797279508262339,
 0.6908957622998416,0.22362154809270718,0.6154858779932416,0.6645672921772332,
 0.9741510207342112,0.8001481561677934,0.25139758307141136,0.4694257480997054,
 0.13843288756593142,0.9702541678096981,0.19815659695848686,
 0.9477133985250713,0.6187302385612231,0.2806403556534559,0.5658657517622275,
 0.4481952619650288,0.56151450948894,0.6516358554157435,0.4439799440091745,
 0.6775348804275361,0.3685634955907968,0.5884784513933704,0.4715114046661919,
 0.4079983324707174,0.9581509273483609,0.8160454121482575,0.15540452360650892,
 0.15626370295532532,0.42336336622252957,0.5504119869995411,
 0.7904214093855555,0.2602059210538522,0.40233896970756433,0.6705355759621562,
 0.6484826347417862,0.07973989161172579,0.6990639580733926,0.8936258376350924,
 0.6203700760327446,0.4801150565372039,0.9173263018454791,0.155872400769959,
 0.5018816536056713,0.35407934486794634,0.7077695366996626,0.6521741696053521,
 0.9690077202272555,0.8523849810761037,0.4559145452979757,
 0.058121091413210646,0.42747330567410746,0.37111158552727874,
 0.30000205185853046,0.5813465086495772,0.8005601010215092,0.7489955062699968,
 0.2998408905515044,0.6760677466980987,0.9194839120501127,
 0.026496797090750035,0.4321593233597598,0.3794309440205219,
 0.35245395913737787,0.7756755763766264,0.6723472102676558,0.8270591814254122,
 0.2922568176316863,0.6702292177129292,0.3460356383388875,0.9227681790960683,
 0.623580448485434,0.9359032202400572,0.3769866398700926,0.6293293366877902,
 0.250942415408998,0.2092453960768318,0.40215307055289906,0.9753199556099968,
 0.34942182415514433,0.19544569290459357,0.8755260900992904,
 0.7145293583683169,0.8304456486394101,0.04523788933723616,0.7725124272657968,
 0.32379725420569483,0.16377497778537198,0.14154476273263217,
 0.4921025498852464,0.2208706938406474,0.40629292739255796,0.6405395482253524,
 0.05934840998744351,0.8714039460552179,0.31228395199161463,
 0.5185094843278777,0.1818393366190324,0.0498451374498714,
 0.043652853185734974,0.7012238209923978,0.12056097489637874,
 0.4780409727648859,0.16919939769079884,0.2014607584849737,0.7517474574285252,
 0.4705596884531774,0.7090390533284255,0.5005577867222157,0.8745048663707324,
 0.09359745676575082,0.6565107768117335,0.8067631530832421,0.2555892226731197,
 0.6117732448285975,0.13350639036518974,0.9600728669131985,0.9115338991011718,
 0.5799031248468866,0.2180429687099671,0.09712921317216117,0.8515888319569974,
 0.11885000274892321,0.7532464408148027,0.6785321425203324,
 0.32335310809743556,0.922871559804912,0.2311085214082701,0.7388437368508098,
 0.45291911491290193,0.9485092489645527,0.41321987513069514,0.179082839396292,
 0.06461480051344026,0.9075786101748423,0.5960092169480483,0.3530635059746119,
 0.7686888704189988,0.010123152018887471,0.40191948889074636,
 0.8722158582829165,0.6125201009817408,0.2215639465706254,0.25925967481049517,
 0.14985218540027,0.8420349250570998,0.11340962334311133,0.9297441151092387,
 0.14392266344248583,0.2617231294675375,0.5581774809244767,0.5449331925427214,
 0.9602709081566038,0.08565457317282987,0.06407461912645185,
 0.7804381353207092,0.7154863854616411,0.08031614624724348,
 0.16470471802534836,0.41312621200095834,0.6598932582031456,
 0.5447758810849078,0.11951635417506634,0.14209735303450988,
 0.7742675400562822,0.30272935326949246,0.8317682680797306,0.8404152364860032,
 0.5969285045584376,0.11349234340624603,0.2264763923671108,
 0.14656428496207385,0.47552497293057605,0.7769907894367544,
 0.38235768883340515,0.7965130108120904,0.858992369463465,0.05616496100083368,
 0.43252713420424094,0.6834919407874143,0.962972495747859,0.5301269520151393,
 0.5251970473248235,0.3771915552912204,0.4992387972943507,0.31673023053411364,
 0.22362515290417306,0.21963458716557283,0.3954076429975487,
 0.11994277642742346,7.41787124519e-4,0.47606707420590366,0.882059709457499,
 0.32451867532003775,0.5383877852425466,0.473073147919741,0.5784777369876064,
 0.04899532477860935,0.6359344094073333,0.5178933537391861,0.9218170334522169,
 0.6658666048232119,0.3932104369135876,0.8618158805369751,0.856916713692746,
 0.5355599593476034,0.2978757110540069,0.19766864640502302,0.8676727335535175,
 0.2734417080203888,0.7501030952998432,0.9833350406916274,0.7870260823605697,
 0.4855985845968215,0.3488159263611532,0.9961970158247195,0.47350369455979435,
 0.02494246472160966,0.29617168547360107,0.4668820439870843,
 0.2540384173495619,0.7964861432544235,0.641458472750914,0.32250255268686445,
 0.955871162950858,0.9639500316803932,0.0530537769170224,0.5272782029368837,
 0.621104469327271,0.9084823840038216,0.7106147340236821,0.4749474309868764,
 0.8453140489242984,0.5159534716772372,0.33600904954594235,0.9479985797023642,
 0.6621728345993693,0.03803122239702872,0.7264208699173922,
 0.20543225682395183,0.41836271180841167,0.8211265926666532,
 0.4644824599211853,0.5258750416482668,0.2763080065003162,0.30357463894716197,
 0.43507362130792937,0.20407825144267666,0.8339908450468598,
 0.7632362918886287,0.820808383864552,0.809336684624687,0.7259576272216258,
 0.2918643817619899,0.44779085485149017,0.7776500386138838,0.9146284303778279,
 0.5360119831968866,0.15951909826424904,0.5243479458122738,
 0.13810717689787233,0.46009259596417884,0.8947789902090755,
 0.6162362448910556,0.9757678233782824,0.668637933759156,0.6867961229275034,
 0.22059876346077634,0.9066422907425002,0.6347140986805742,0.5542640116366218,
 0.5721488064483552,0.9103426997178008,0.08320298535242676,0.4240697736383826,
 0.3965656258119783,0.77178419029835,0.3029023567625164,0.3013739574513886,
 0.3244572119915514,0.6363747472069488,0.03652843348050838,
 0.022451448734370105,0.920746021067135,0.6308268246490419,
 0.35393793520841044,0.7740248554941854,0.5708969694279282,0.9655704165265491,
 0.987087669697084,0.3824978269526764,0.40469624254161307,0.3305831359525143,
 0.45705864165960586,0.4681450473536958,0.37057237950480304,
 0.2596502914170147,0.6791007700293719,0.9365134445650943,0.37893896411711203,
 0.9570020580475014,0.6005645991523244,0.43984332919063596,0.7818366754343997,
 0.9036781708076435,0.8765190311404385,0.2913356215925509,0.6277007461373855,
 0.12001460113557338,0.03879418387410283,0.14492636380754254,
 0.5722659966559317,0.5491109703895203,0.9498583885048896,0.9293189475786039,
 0.10004753856709037,0.11346559356000474,0.04844884726654197,
 0.5814199454446214,0.9946708823502718,0.15958878726148795,
 0.47620191640558274,0.358435247941016,0.24213598228079336,
 0.0017496407902969358,0.9531105745814843,0.3713839034023261,
 0.5332017138675931,0.6759372818899091,0.6715168423250383,0.4731497458260273,
 0.6011410245185903,0.32024697802528657,0.1995464794390589,0.911159035377325,
 0.398520177477264,0.6659786216939709,0.5425144651883578,0.3101290340829539,
 0.91274198810461,0.8427577966473384,0.4573696900439155,0.17536599637965122,
 0.8433336951996896,0.6347263237011701,0.7398586020937428,0.3220832731140867,
 0.8486701138856498,0.6958348160546448,0.1503480469839249,0.21393658013948758,
 0.733343877306496,0.7578946156052399,0.08741195093914544,0.34685704222446434,
 0.01974202269651404,0.5108908965705522,0.04971468966890691,
 0.7066828955331669,0.5457119390170552,0.03022483680354726,
 0.30697240435016837,0.7138317519843574,0.5843120661873783,0.4319711707349678,
 0.6404200137748562,0.3429130718327631,0.9482553228949228,0.4805448060006223,
 0.3303861124605191,0.38041593147687636,0.24343152176825966,
 0.6805343291462569,0.7402162245628553,0.4710868840526613,0.12319360159736725,
 0.6219818450392123,0.030740129474740252,0.6282672629094406,
 0.8199518570536318,0.5760365557750595,0.010901622356664142,
 0.7656283404379463,0.8920365069282058,0.29466123579937875,0.6041796353927634,
 0.4139567207505258,0.5037339337648293,0.42006939950075717,0.9817996221898503,
 0.06404546777875719,0.0042887982314705475,0.8697927188809261,
 0.5434604085981948,0.6391744566381864,0.9698021902277422,0.4591070771896968,
 0.582666160343413,0.3743700471154693,0.27341027147767905,0.6699119986688746,
 0.230978703288772,0.704345999065999,0.038574927599947184,0.33125032237911023,
 0.14702841289134516,0.5179317933589205,0.5576741378377479,
 0.14105601196999862,0.547758933030654,0.6947402844242319,0.728857333123493,
 0.9136985651776631,0.09081736701776855,0.3394760129987454,0.9152029229401458,
 0.10653916530125296,0.08169717012148414,0.28355744082258083,
 0.9487740853177332,0.5119272057364341,0.40238033413040675,0.5710720774081807,
 0.5687648255033343,0.8459492632438875,0.41799326864284037,
 0.08582504354965059,0.317597197170427,0.9252420453870717,0.2734153584689355,
 0.7790128243690277,0.5268573018228447,0.011933733857023343,
 0.26795432395546703,0.9681603978701505,0.7899646994637437,0.8332089126473372,
 0.49307297210474155,0.9495332419428263,0.9931710244142274,0.9457236804872993,
 0.8717960839282003,0.5013712336819454,0.4379662368337943,0.37004254326720465,
 0.39825101271783847,0.5925297462445521,0.9104655823222245,
 0.15173916583014957,0.01849391619547902,0.235171992690826,0.7014923171452153,
 0.4312029319055447,0.24921403462446667,0.5447621396333371,
 0.46651239850330206,0.3564614613892487,0.581909781668106,0.03285200751897499,
 0.009219649922138973,0.9232399650752092,0.5699339677595705,
 0.14128317238874866,0.20074503388142384,0.9003757070906815,0.454075490690128,
 0.5621826602967752,0.2683067231211085,0.13674856305248284,0.4958468256287798,
 0.6810245911775649,0.8799559487207664,0.577905201266803,0.9640263219486895,
 0.38187944022869524,0.15823138941595916,0.006722776638080852,
 0.5904752016742598,0.625648515664393,0.05895522323054003,0.9585929653464141,
 0.38009140986300993,0.7850228014770243,0.0881625597260216,0.4214615294661871,
 0.3001582772798702,0.44416187137633745,0.6879899500005177,
 0.09111225279518897,0.11459408780896663,0.34250743740005507,
 0.04214437365855639,0.7584230768288361,0.7536513021100377,0.6503346898782313,
 0.2016276439514606,0.4288192554916439,0.7437731362191182,0.24828824012612138,
 0.6037019922448941,0.3754247493763303,7.938832261950024e-4,
 0.6206642306489502,0.19082673482504342,0.8432523953803942,
 0.25440637483184503,0.9708241953734701,0.9762524363166624,0.4968564669853306,
 0.1925638167420627,0.45291137602386844,0.2004634171687807,
 0.30389827698135685,0.49740823977849546,0.41861988914940707,
 0.5632647205444264,0.6912313893965378,0.07986507653988928,
 0.27506578839036067,0.5653002960587096,0.45840470869925576,
 0.7012168910725083,0.508689317315047,0.10394786156412461,0.8968322333406272,
 0.41923902173098027,0.4973564498421068,0.2892994420777283,
 0.11429466167821634,0.8435630009499577,0.9711262524351818,0.5525781585268756,
 0.3300923341162536,0.6682683697517602,0.5444695401848151,0.02630156719832799,
 0.23042162954223988,0.01596126343926596,0.3277546459922167,
 0.35059074189180206,0.5286746727262778,0.8899277810107357,
 0.21123957237926894,0.27441927620618234,0.4537715244327788,
 0.7603935954574634,0.860586857150522,0.576902632578501,0.621271104764002,
 0.20016926351276998,0.03191306115246373,0.4979749142128027,
 0.6430035714340218,0.33853995909959655,0.1636479915505361,
 0.031689881804375375,0.28657872406547846,0.10800160335101527,
 0.6556498005583674,0.7265478396927517,0.3104166716430057,0.6868200470714297,
 0.37297290480160183,0.706650851531006,0.3020912352378463,0.7688934916329294,
 0.9932590909846332,0.573128779903124,0.38174799701656226,0.7494732915174317,
 0.7034693212429082,0.4964699405121923,0.15272972286310915,
 0.21839991978742723,0.25352929401540003,0.5286234978892056,
 0.42067594288687715,0.6422494410912127,0.5484916853510619,0.7110079017340082,
 0.17781317661624618,0.41166313460341186,0.2984094480969769,
 0.7094843742702749,0.6816848886477571,0.28498502393272807,0.5416582655170734,
 0.7698111988068077,0.774170589265568,0.2792414227939628,0.19252122109822511,
 0.22735981432400443,0.9113422829935448,0.5816185594626172,0.6163338510646318,
 0.7531140370094049,0.20665669867404268,0.9797013139180691,0.7074780902248017,
 0.17513757808373553,0.6615788678028071,0.9499014195495534,
 0.07507082441006707,0.756165233198686,0.7060179872117034,0.7481793150323304,
 0.4398410622266319,0.2994953745261383,0.7935350092485328,0.7063121308505071,
 0.7649040787579895,0.6558854373868229,0.8673516939672254,0.4640183098084022,
 0.20054254713614073,0.05860312106082566,0.6710641959520196,
 0.2862766745067502,0.6934324182269689,0.102403291441437,0.817380153801017,
 0.7535148537127587,0.9161102679511993,0.911361064281815,0.9590147478943167,
 0.9958530332513462,0.2130706662035805,0.6045483256702807,0.7901093812651113,
 0.2769155604959296,0.7182550538527812,0.04638212898502847,0.9360150323542299,
 0.9701511264461999,0.028980516964556724,0.5814253870187995,
 0.8532166198249106,0.0809737245311426,0.7757199948835487,
 0.049700120018345784,0.08098718576850816,0.1206094783747792,
 0.8951857716085544,0.8477537710317221,0.34633090679105905,0.3320922152492022,
 0.24887448775957444,0.22005667327212108,0.3797201281745425,
 0.5036662656995754,0.22193607913355218,0.23380908459739813,
 0.5694180648816806,0.5352878056442449,0.46413801869802696,
 0.09660540625321812,0.691880778200461,0.13490413785873911,0.9957256221783215,
 0.7985143157432696,0.5205984366067369,0.2658839747453542,0.765028360785164,
 0.15727754799416027,0.06728501458609659,0.1618762540848292,
 0.06843326749199807,0.120453223390665,0.9560646607641815,0.4374015607733067,
 0.6594426344752066,0.9454930734810206,0.10071385092354257,0.754490192081229,
 0.5488474838695695,0.09144472991593755,0.5088446524900171,0.5258163112622423,
 0.6403522838518819,0.7001287372401911,0.6874563512347154,0.43514844470341796,
 0.40891442465833017,0.011139863386475324,0.1649105647724175,
 0.34786656204342936,0.10260786104446351,0.7347960467629127,
 0.9036626649565345,0.6760217671924564,0.031128479871100345,
 0.5171503020566695,0.9531175770881469,0.019878119209966405,
 0.7802028679240731,0.544273716847556,0.49197807252955794,0.1549404699862329,
 0.5767772041080622,0.505943257552406,0.6963312677408087,0.43518488190884863,
 0.4586283685106677,0.5153994490312588,0.938109028016899,0.37242011363450833,
 0.4202304963565915,0.8495762718058448,0.519632337598004,0.878134352132381,
 0.10173281436490378,0.2755557111940041,0.8805648568467795,0.4181159817402741,
 0.5708032880869214,0.95794814604649,0.40382905700936356,0.10504393239498777,
 0.4673852037840148,0.31143874282844863,0.9591720084734687,
 0.09193103603337405,0.37414953964393194,0.8702023912479125,
 0.7488564892473285,0.9239072971707487,0.5035355905230415,0.3245218352154955,
 0.6802077317360021,0.14626078302488954,0.9642192340610334,0.6100483943127027,
 0.008743502943244219,0.6227873101646261,0.318310779893067,0.812804703190591,
 0.5821616695467889,0.37327811403763445,0.03881049623187838,
 0.9338157263994099,0.34011749404301017,0.4860624338124808,
 0.05387008413572114,0.5947508774366969,0.5024359572466448,0.5588177474534267,
 0.12125211489537702,0.04469048836722456,0.6574026643562324,
 0.9365848117145954,0.9059407677991579,0.5650577093717644,0.14759399468652923,
 0.8420151155161624,0.11502767958018845,0.6698308053951509,0.9474036303576387,
 0.8428605616165431,0.5247052894773959,0.0423808145981982,0.1803972169967758,
 0.22989929768939488,0.7957675712678676,0.22755026076820328,
 0.2404893822385159,0.7209355194375939,0.2533410282871892,0.3373349970349402,
 0.662758851196871,0.9242661600826004,0.3375771915242006,0.8405479585712256,
 0.3624163082350814,0.48541204607300215,0.11138108206223307,
 0.35869687436675946,0.730646018882273,0.04571614103802102,0.446160937064242,
 0.517703971424572,0.3555520095127285,0.2032759150558927,0.2374563724149792,
 0.35019367828357284,0.31374486159385373,0.5330148107279953,
 0.43161849694292054,0.589107360127979,0.17460803215853438,
 0.08806971136631048,0.7957964675806483,0.467285845027682,0.6415293790831168,
 0.10834559481421135,0.3871960884122021,0.6399271430166463,0.4889510001206757,
 0.018896500837769947,0.6574004761383603,0.15551505513401176,
 0.4443300654956832,0.727404326669773,0.451480610217707,0.3319362551350993,
 0.5865138508582505,0.9220940218312735,0.4489094135174846,0.0393019196405181,
 0.9035761330571934,4.573305626121016e-5,0.11167858517919971,
 0.5561361410120363,0.3293058431937076,0.7020306831039341,0.9539806745618522,
 0.041682852487041844,0.13202375651896248,0.06336347987803259,
 0.4781854353321032,0.33961155617859073,0.6007727353048439,
 0.14059078975051575,0.5454102980376186,0.5184884187655696,
 0.07969357516255782,0.3948861354747043,0.4668531459836669,0.8241990181423895,
 0.6341941943845547,0.4784815530444714,0.8276208659606027,0.9110048056331046,
 0.6646511087499057,0.6789061712288296,0.5982106150775216,0.5364564514728752,
 0.935631695394072,0.651114233639677,0.11293254251098228,0.09577719458744183,
 0.7717213502817717,0.6095021189438301,0.3136897116824404,0.2229324491138961,
 0.5663709720627512,0.2627389909508201,0.6544765428746971,0.7857470949941442,
 0.5471204840129027,0.4364151711388288,0.24166278191422874,0.8286419998299777,
 0.577047031736747,0.8800757117429613,0.43933726593615374,0.9968689945573685,
 0.8075360125169867,0.15290398738598687,0.8784361621993148,
 0.33104070847644185,0.7221970455352232,0.3895592184974075,0.7068288548783217,
 0.3428592614948389,0.8638248355141694,0.8908931885459892,0.9520246059962767,
 0.4754065149967608,0.16306726816493922,0.2814828591518811,
 0.061475311549768596,0.6071834730161132,0.3785381735221237,
 0.042669213366956726,0.6803250434870135,0.9467747101771067,
 0.015199636051498877,0.2841102332167136,0.1953487517023793,0.877931773873936,
 0.6855886691681243,0.20593547778772237,0.20519259697909464,
 0.8388858003186246,0.40979829470283535,0.45902707480330773,
 0.5421682110616804,0.6764359301690811,0.9885109700944049,0.9221514622594191,
 0.8804990661744259,0.3611734237767369,0.25465792620411143,
 0.29370828087666656,0.41279004409080466,0.17604319850944394,
 0.7021739413745899,0.4536745439175903,0.24029168178767946,0.9216698377592689,
 0.7485252072433587,0.11395947963616504,0.4891473726831359,0.5843599369607235,
 0.9908427534902411,0.35843790179979307,0.6072888933377603,0.6440549072822139,
 0.350721589694729,0.003015133278414517,0.2541779977734886,
 0.46329792598699626,0.1328887949308356,0.7861491657794667,0.7697895045517096,
 0.5518464761848026,0.18134756470588653,0.5138991443567564,0.4708559793800873,
 0.7334124556125765,0.8903850669002006,0.4916523048049183,0.19706108892413865,
 0.5787635239868005,0.6970105661275565,0.25597289712572324,0.6339057525753693,
 0.41874357202502055,0.11030895903229798,0.03860710630076425,
 0.36340982952050005,0.30636169988406114,0.006620455731591313,
 0.6906758891310258,0.5467956700012173,0.1623998055604382,0.613681576022533,
 0.23251335089543979,0.9118394984089722,0.1766475323429314,0.4686656998731976,
 0.17421889227452303,0.3744328377549506,0.017054375386391207,
 0.22870844218839326,0.4054653036319259,0.5760855566056401,0.8827677580937003,
 0.01870694859158284,0.528981330160458,0.3593662760016869,0.2110427983051032,
 0.37001559088554925,0.6182950126847415,0.032466877868609245,
 0.20623838418098062,0.9577939346904599,0.8777343107958089,
 0.37767786213434207,0.09917058004493251,0.26539139394013267,
 0.047075543916621054,0.4122316340585156,0.1587598993447198,
 0.3329064647169584,0.9920838481806672,0.5493769521723932,0.3977535503906252,
 0.8274908258364606,0.9738102872803406,0.6475989490674064,0.69651010606457,
 0.3484408093658249,0.7297785852158575,0.47710269189775845,
 0.24831107017889753,0.003653011935195183,0.1780476212466936,
 0.7832930720959073,0.2815103706694767,0.9749664826950069,0.26644335051637036,
 0.8430936439660284,0.964570208070154,0.3254050586060784,0.5690331339913071,
 0.4196645581226175,0.5392973104357405,0.8826120685401212,0.5395484337798732,
 0.5909495876973264,0.8020733053504965,0.49239020089772156,
 0.08554301957054522,0.5795164259035618,0.36021025914167315,
 0.47559891345637373,0.6360088979821199,0.47101809049438237,
 0.5784418579674674,0.4778885507359265,0.3490270463866718,0.8522471010789086,
 0.24333722967311677,0.6930553127323144,0.053144661398764614,
 0.40630862505362975,0.9508487491699188,0.621577112124465,
 0.025556027974188544,0.3077346306030614,0.6097478284386929,
 0.48303480196247883,0.601504421286923,0.9261518988097078,0.4738638627238235,
 0.04607192442215269,0.8703885238371909,0.05010372348723857,
 0.21831699660544668,0.08729399250910885,0.8602560375195563,
 0.24219306402881957,0.8514435147514104,0.18645861460107338,
 0.18758124474335824,0.04703328323331779,0.7966237663623079,
 0.15490345308832631,0.18833283795543587,0.9307168756673642,
 0.2239804627778571,0.5327184296406822,0.6509759981693769,0.23101399367459052,
 0.2643357692502265,0.04753357626056032,0.9136389335743793,
 0.37702142017196816,0.9974154302747891,0.7841860825260084,0.8220092865601295,
 0.5880823503796417,0.7587844492613649,0.12844142106765621,
 0.28636011717820553,0.010751564997614982,0.7181291707941949,
 0.5131058861334494,0.47272223157989135,0.37466126514376263,
 0.6112259277530356,0.9440367174763622,0.6687696857302784,
 0.020415593568538504,0.4096463397122817,0.9697395630358019,
 0.033558545362908365,0.6610378450784739,0.7709499608656726,
 0.2723810955053173,0.12196831139904207,0.04666732626997949,
 0.4707032766327922,0.4987577282553536,0.4727049309779766,0.43624572869302947,
 0.010486302039291462,0.6139843314499046,0.27539404729189876,
 0.5908752167236601,0.3328728101464039,0.04260990713579249,
 0.31025461327634984,0.9712832609659797,0.825788866317924,0.20225520993919655,
 0.617108221771783,0.5148421015123861,0.7142016881180375,0.6005382930336411,
 0.6370776772635653,0.21018634903444577,0.9480060194990417,
 0.015397488979187357,0.9771895020712441,0.07598095957858741,
 0.664839580526606,0.6313828391565185,0.593245243601313,0.2670895837768723,
 0.6058751974989323,0.5423614089494837,0.894011837884304,0.7706121901216736,
 0.3381061422591406,0.939192652597644,0.016791551227400725,0.7060887565733851,
 0.13860405552710975,0.08250329043906512,0.6236237524728553,
 0.6933151928310485,0.44877634444957115,0.020512631190528374,
 0.1690405482744488,0.30229280696923333,0.36891194412475226,
 0.45473132720306364,0.6527243463103269,0.010185857394360953,
 0.010138388125672138,0.30020389022607563,0.6030344286799952,
 0.18239612801513372,0.33743305862848905,0.1204673777063695,
 0.32751906083751436,0.9948651524993668,0.35709069132722193,
 0.3044334081049135,0.6811587647448404,0.10541652825754821,0.3213040066175803,
 0.21678392507402933,0.5380836064166233,0.9387225927594005,0.5663480021289413,
 0.7280095867159625,0.21173249374811276,0.9062573567491572,0.37020795715486,
 0.8423494819418252,0.09753701915639756,0.5568286301877772,0.4436055306055082,
 0.8532352079781486,0.6015943290096673,0.157666619885645,0.40064389483857277,
 0.5224094995575245,0.37418808775081824,0.27931067332369763,
 0.8514740154747901,0.4236284176180236,0.13899172542079086,0.2900500694646687,
 0.9060770002896302,0.5781017515104486,0.17926633535947567,
 0.11058102765076405,0.8796241656780853,0.7910089664907141,0.8781196722163465,
 0.4999032944805294,0.15368363391056528,0.38378278016214096,
 0.6275292977654063,0.9102254558193081,0.4659425914141515,
 0.010997011207372953,0.9671220377885839,0.6923133206812473,
 0.30030270852370355,0.7883526073408845,0.11861541001681142,
 0.30944064629459844,0.9917706156527181,0.5575517404720868,0.5535738391683304,
 0.08141029649155429,0.34445769410925253,0.6268443029174791,
 0.09695829194984062,0.3824093986837268,0.68294592902133,0.9322808508411397,
 0.38374288202122797,0.8522018019479184,0.40383658270144807,0.896612183071121,
 0.3294043002283109,0.4575202718197039,0.7818498894731888,0.5930278805715343,
 0.6087137218159345,0.09282089196389609,0.0678755416018042,0.4901646821724346,
 0.7930952314700807,0.07666240920135003,0.3177100363260379,0.3175758609984628,
 0.6242399973885071,0.3937721652991629,0.1259265707167022,0.5703518267380131,
 0.5160598585570162,0.5188773251001332,0.3561367406289234,0.21215357197460705,
 0.4029921252616273,0.9392194503889395,0.10716776664183236,0.9499846840727167,
 0.8909994551493593,0.8501012772741419,0.695870202648908,0.961916375364475,
 0.41245017840262543,0.33781134892312337,0.6397085125100832,
 0.3479601843008788,0.6334139799422164,0.03959056987811904,
 0.003007192871021358,0.7363025249336852,0.020916631544784514,
 0.5529427984103138,0.6738539043793008,0.9671666426501001,0.572155055301419,
 0.8335207016540168,0.6162159026360429,0.8573733042299166,0.49121074631170425,
 0.10991902179085467,0.4627276014788537,0.38671881948512865,
 0.9546483665648586,0.5421058679861165,0.21330493541334994,
 0.12867963546242356,0.7269825250557459,0.6986319164595435,0.6214460248760341,
 0.6902854649560547,0.2083838848097871,0.12007714407031755,
 0.006342909994399237,0.7209472273457422,0.11310097617694381,
 0.4596014961449365,0.8531710408760769,0.4625557057554015,0.42224685822599173,
 0.6861939325217465,0.7767284869305949,0.9850266637454487,0.06732851077220514,
 0.6651266185404111,0.12621950532421045,0.24399638344394836,
 0.9391776887155155,0.4318012962002056,0.9365565763373831,0.44107352794954435,
 0.6136120118963321,0.051170580470726645,0.42326773878126545,
 0.3197091545025512,0.8353430709159162,0.37237816827593195,0.7373946351768509,
 0.3910834186289769,0.2878036608173766,0.9217872099357601,0.24095126982329151,
 0.7643961533986142,0.44408350990820944,0.47278053358310146,0.633380480816698,
 0.600139441333897,0.9154061443083917,0.016964748169876254,
 0.17465380646252004,0.1139168396076975,0.8848459989512323,0.7037410807753737,
 0.19062253589497058,0.5613456314661368,0.2484553785765935,
 0.42156158920061093,0.5332010776827957,0.03381165106705997,0.697606118868292,
 0.3930131423851656,0.847125197024812,0.30387181222185045,0.617313985933874,
 0.7276765892174097,0.13975343235240867,0.021329394885444053,
 0.8426230112099471,0.9942857197047174,0.7857212469250869,0.44900308999304406,
 0.32824374456068994,0.004655367011789147,0.43104208419446277,
 0.888386068653767,0.3209938530596014,0.8829276861887841,0.2778183964130435,
 0.7649813680262467,0.4921807202534404,0.36304253666141006,0.8554614113547743,
 0.49613660388402814,0.28454885893778115,0.011329139738547322,
 0.15850845281942716,0.5570084901258935,0.16202684536782874,
 0.06510256480575727,0.15715404291654422,0.23417134904218562,
 0.9438635694323182,0.6877815448178357,0.43677634152291445,0.4749662692310904,
 0.24564489011473623,0.9119667456407301,0.3326099912090129,0.914589253304593,
 0.09085110016963083,0.29689341380300616,0.16030672184026185,
 0.37572219362942905,0.3706502901615455,0.5429065060293048,0.6692206477948526,
 0.22412431075330164,0.8869061475526526,0.9873835271921614,0.2366557796347707,
 0.7170976727355339,0.5064619122494199,0.04984874076389467,
 0.16494340285436337,0.32235423135541463,0.10471168870037462,
 0.9928161379818493,0.6130248459083173,0.209018432948053,0.38311039445726147,
 0.7069779110054252,0.6786585778648555,0.9504239177390128,0.5109801857233487,
 0.25972230432644605,0.2930224410353457,0.38212265398929823,
 0.25790516050965195,0.8587515241664471,0.5217554496091014,0.5153678104058244,
 0.6434125715841331,0.8769594101468662,0.5936117280178486,0.14211425271929845,
 0.35850122355200875,0.873104049261512,0.7731461586253345,0.11458880430122775,
 0.3360635982572464,0.35731643399026125,0.20956520514681676,
 0.8215553565764988,0.17886889020316188,0.7698625151089052,0.6912013390601236,
 0.822079158055025,0.2004011598315647,0.35471417634819524,0.6047363123871257,
 0.6072406618913453,0.8460674893781199,0.6893296722631508,0.5944018998168326,
 0.11817736983437166,0.01225860745038898,0.3334556259076119,
 0.20276673711854276,0.07559822923438564,0.8891213034680883,
 0.22761568286065303,0.4101146168927685,0.2251702989533293,0.6123160625481858,
 0.1048665705036379,0.03513281825618053,0.42426406652492543,
 0.9996380700548728,0.04144672282756601,0.7155702632277062,0.748717427032956,
 0.237044197792334,0.16049958758442884,0.3543295480411991,0.36363267110681974,
 0.6833011335854157,0.9365360417977262,0.07810624284731216,
 0.14629307880132447,0.3441473617023839,0.6722262972449624,0.6595465679529817,
 0.03432991840885069,0.973226297530282,0.07895321820253964,
 0.03154336643328226,0.625024338470268,0.99764376659313,0.5216992449947184,
 0.6339058145821223,0.7826120014583139,0.8042238921349796,0.7008257878353761,
 0.6379905926078924,0.40131360211613276,0.7110259192866897,0.6909732758417447,
 0.6500993194843053,0.5389471781947969,0.46466016546804645,
 0.19385601038253675,0.8899719127582827,0.6353033895098946,0.3452712163764804,
 0.803578958501836,0.42817696246485015,0.6100454157531398,0.7021153860853575,
 0.27051687287984727,0.9411598521247923,0.7178064673732503,0.2649151303583872,
 0.6589884987668633,0.5965048489014525,0.28550675058044206,0.6988271077728511,
 0.4644433888577928,0.8534037999402533,0.17546180862920957,0.2787145172186055,
 0.12744825009818128,0.3076754324674503,0.38207392385440386,
 0.6223410846044559,0.33585942536978775,0.5385179380584351,0.7839171863347444,
 0.45805388328736574,0.21863545297190312,0.7754364736542559,
 0.24738950638376334,0.5049125040837108,0.4631182700740023,
 0.24448867460691248,0.9045074468305477,0.6034352030085722,
 0.06945296285208247,0.4740046439589024,0.5183097708935719,0.0690601023892784,
 0.17442915326426933,0.49134655557496465,0.19695624004665235,
 0.301889888090787,0.09521848308340952,0.012700266661600401,
 0.20951645858664203,0.05142558800065733,0.17742247141314438,
 0.7426274472295398,0.5804288510947105,0.5850187982855273,0.8382193728450021,
 0.25911109256109055,0.1286359650110015,0.3682028745268746,0.894250482382828,
 0.8727774440658504,0.6448149177296347,0.907968784977914,0.7413232503756263,
 0.9120685878974611,0.5569488514857758,0.26634282428299483,0.835834398779594,
 0.36231571183535594,0.16786847801524774,0.8703254256371198,
 0.07671372836402812,0.05944606862671409,0.9518720087224192,
 0.1302403781391186,0.8059652680238657,0.7646095612366268,0.253143751273064,
 0.9261981998571243,0.9100925984758472,0.08914377021296893,0.800489113602772,
 0.15480654734044297,0.002740231483591593,0.09409485680994878,
 0.1208007812932983,0.8853010209874526,0.9333067315304322,0.2772227862384412,
 0.69894158907185,0.8086378581237633,0.6477338975654526,0.9704906027786669,
 0.417880209274883,0.4087083249940805,0.37663055780358246,0.5648047737645921,
 0.28523344782400084,0.7435434477734962,0.670991093004445,0.7287959362336488,
 0.8846314949341936,0.13979458895207486,0.06995037214370259,
 0.8794754012770766,0.2728810594376141,0.9834121782757113,0.5691626293800063,
 0.6132981761904132,0.11482345257072168,0.5781720713555455,
 0.08500739436042337,0.861820177114908,0.9192948806545567,0.9661902263730775,
 0.30291854162351506,0.6212604980911739,0.9784972078916576,0.9995564435627181,
 0.8263784864170184,0.9161380817122354,0.17016402085321902,0.5136947319069425,
 0.7102042104101186,0.42404841251785375,0.5004793398749317,
 0.48990992043198023,0.6106462627512179,0.9517951214153484,
 0.05881397994544213,0.023200192334197678,0.7397075361187212,
 0.2511671784503401,0.8017838452303978,0.5173293500936815,0.04449819609706396,
 0.31584937478415265,0.23824669019195432,0.16971569450664858,
 0.014201508894527426,0.7724393786683929,0.8230072198779237,
 0.9514090244609165,0.20838852766466465,0.36831298042967076,
 0.05100022084592004,0.7164498915168732,0.6213896148443,0.4542815679173766,
 0.696141072636672,0.5062332733002021,0.4364802350057252,0.03526594935197602,
 0.8349225600760398,0.31927118201094484,0.400583533368826,0.9385705366742185,
 0.34331758789691025,0.9983096853520274,0.6261584345457419,0.6404760163543943,
 0.476257300107804,0.3430825546453592,0.9702779519189799,0.13449406983058265,
 0.4329743604599373,0.8019158978674655,0.3975809237451531,0.7440755292080328,
 0.9802095959708186,0.26351180050386314,0.21815280599981768,
 0.6051866806451561,0.18593811789016834,0.9066593468566871,0.5733504310463358,
 0.5185174416846432,0.05871292340765555,0.274685832146077,0.2996913313445405,
 0.3579837398675606,0.7468319151679355,0.6085357860711789,0.20762602230447025,
 0.6354858956305632,0.7672915079005342,0.01443809034149579,0.6621706208517819,
 0.39296208249464104,0.05055245278336984,0.0020534046297834285,
 0.42768886621819835,0.7807757497341009,0.6544164439317227,0.5038878798745328,
 0.31680560967939475,0.7265606566838159,0.22659492840565232,
 0.6116783724799172,0.4184734603056973,0.9125916253296283,0.988628646957337,
 0.2372136286067663,0.31157055526230704,0.8607857340213023,0.593389908220136,
 0.16498647762542118,0.35134154372015447,0.5533968403289127,
 0.7752298666216593,0.6325224446157122,0.9077402131178393,0.6190323787232839,
 0.010568479261754304,0.8507841120062759,0.39356813494196463,
 0.7702948887097484,0.4975865424879591,0.02022091912392321,
 0.06888489099163686,0.5574746454927829,0.7656967225801322,0.6805209634434395,
 0.482553218996553,0.5082637528115431,0.2397663304272002,0.2589412511911354,
 0.9874411438658115,0.6480165513728899,0.5412575894614413,
 0.053671689795960154,0.7014876901009217,0.8271907611089901,
 0.4055407882484914,0.4780808470142267,0.7039608045831235,0.7925330567657484,
 0.3470829545873827,0.8316009116745446,0.13316301077697812,0.7645958691113345,
 0.037220100040903814,0.26947063148778205,0.35547203271326633,
 0.4937476817197881,0.023791676185019206,0.8521009730488482,
 0.9534563632679248,0.09759257956308642,0.5993793755538298,0.8679075914526576,
 0.37807239043815555,0.13343505313718773,0.36597265822901526,
 0.25658906926178204,0.35909105300624855,0.9921207266433612,
 0.7631378154368748,0.26715865939349703,0.06148961000978792,
 0.9171686864697817,0.9586973224443032,0.9330479148404149,0.36107637698143713,
 0.6138598220204656,0.1949582866723567,0.22599906067507913,
 0.39661646405215434,0.39676855014255163,0.46033739847389965,
 0.689596526093103,0.4533227462015015,0.1705826372574456,0.2009741436108743,
 0.6529749219002353,0.2167846070977717,0.5072043083270106,0.9077301740658092,
 0.4469974639338211,0.26323682289944594,0.8136727909821952,0.8136753155576139,
 0.8453530831972178,0.7652457724596472,0.5086053876641163,0.9675878159225826,
 0.9793848328930309,0.9112104253315194,0.20323913984021946,0.6007786093058596,
 0.14122986560352846,0.5000409990312396,0.32339442692923825,
 0.11850678762410727,0.06807270163902523,0.568188800103733,
 0.49857022441513577,0.7220658532193553,0.880127805660777,0.8343737473402799,
 0.9481418824942748,0.41861180987846525,0.748375183442349,0.17894261523847588,
 0.8201955896907493,0.9867043130052239,0.9054496010009875,0.8263617523293616,
 0.668672014939293,0.014159308555017125,0.8907232518043894,0.5726035763974435,
 0.6023710133225877,0.3161824479247457,0.5646756160217142,0.45344856954756885,
 0.6225239338360536,0.3111573463669166,0.8470939363636826,0.8740762847012605,
 0.9946355874596466,0.7519849684956529,0.24043145158723567,0.9759336546231397,
 0.6424913808446169,0.7300522771217041,0.5120077780372639,0.5339489590091268,
 0.22442435043774966,0.8903674109334331,0.24161399845685327,
 0.6849710911204535,0.07218707169580452,0.016256007043963128,
 0.4479814322097255,0.07136929486255639,0.03156302990127169,
 0.6504862874821764,0.7029431044614863,0.008758812658241077,
 0.018852544190221865,0.30033629856308,0.18802282533087877,
 0.24881031957562993,0.818446114679812,0.9938365665797032,0.37942932019764986,
 0.9893879514003504,0.8415710720457031,0.7577358490539342,0.5633828150698308,
 0.3023407038427397,0.7585571947232042,0.7945296705320188,0.48907247635775597,
 0.4676524199900902,0.4519609232040501,0.7124087140884772,0.08162497975044969,
 0.5826401718571708,0.6807523424386007,0.45696592715280304,0.3562696148786758,
 0.0977724845615493,0.044377348895124014,0.5885465214809043,
 0.3514090375642861,0.09039858179680027,0.19219229775127034,
 0.5137091335149625,0.2512334627948475,0.5770693868143789,0.48233121391285816,
 0.04276466144748925,0.023789341354805038,0.9107604927382577,
 0.2101632870093153,0.588496510600389,0.4091881533547811,0.33441780849942715,
 0.5764666311211563,0.8748142819241191,0.5365447218539657,0.829023502373186,
 0.96639289218646,0.2614864519782387,0.7993140876599408,0.0019958208996292415,
 0.6340598417066806,0.8381705663644258,0.3398074171863459,0.5934806218855307,
 0.9158892251123358,0.676081259106387,0.29842342011977097,
 0.053374586083406395,0.685961110036407,0.25592567435065017,
 0.2897898000363568,0.5812243010713345,0.10663415153949307,
 0.45074652535462545,0.3166634505117286,0.09540266441306555,
 0.7054531966907065,0.7143785046453228,0.4356758569754128,0.37277339986970925,
 0.40632843099850036,0.48921541140992275,0.3989532892789135,
 0.3608721457558596,0.4829758015083888,0.4164436837812262,0.5981507477371323,
 0.751904987558341,0.2377712306496076,0.7701607127359517,0.7117052666433362,
 0.6340704453249533,0.6613998110259952,0.12769724953101946,0.4178516075568638,
 0.0274726881795746,0.4577023635085776,0.3531808563142059,0.39927744028760337,
 0.6745411229815418,0.45704232833072644,0.8659829966570304,
 0.36964843481181764,0.24663020614117626,0.636768042158133,0.6420549989111957,
 0.8015401852962114,0.025259865433156614,0.04613008394828333,
 0.227593745515405,0.9441381208956839,0.7086708875966874,0.033177320882795125,
 0.579964173497842,0.8650495295527126,0.44066047034092504,0.06948374851857675,
 0.5761028427128584,0.8053372309478272,0.33434996467065636,0.9958429796929846,
 0.8134306260581,0.13468437627938823,0.7242638790778039,0.3613723632168684,
 0.3726157322818151,0.15757482118316124,0.1945367914419308,0.6764986760097673,
 0.11032154818908302,0.7406401800942554,0.8197677167945724,
 0.10055627527973965,0.49330238433468465,0.490003272410644,
 0.15702983426905248,0.8255365128549276,0.6717075354120315,
 0.38010614743977844,0.5519766799605723,0.6278742626430233,0.3962252102632482,
 0.008217897196921231,0.46028031906990796,0.8555538664822279,
 0.7632873491751649,0.6467548758360109,0.8917833397568328,0.6597566068984111,
 0.6161332661050684,0.08978915404315901,0.6925226993661978,
 0.16875310006861688,0.3319256889050254,0.48409766252061415,0.830619401679148,
 0.8418414024585867,0.0065650871726157245,0.7736056477423331,
 0.02809655687199053,0.9828428758562121,0.6828279703603234,0.8640258057222151,
 0.7449726332203495,0.6599889949509169,0.6252894287147951,0.11168340532103205,
 0.8499899181474753,0.29991890444514857,0.47128141577064864,
 0.5542916434556937,0.7674920513801986,0.03605862817328631,0.9394501332070727,
 0.2077771825188649,0.9863184577112706,0.21028008863886083,0.5545646640924125,
 0.8990352338359239,0.08693651796930302,0.02571614022513924,
 0.28403913370958933,0.2169573241483851,0.8662228869527178,0.6078824296477847,
 0.27647520892575694,0.33783210219248416,0.08476042197923128,
 0.7008495080963293,0.24553457643854526,0.4352111299536542,
 0.47652246379155205,0.806966551409772,0.8081343834929774,0.36157179748557466,
 0.9818471898105737,0.7284464108569179,0.806305233927117,0.7728595532225376,
 0.7363806154888174,0.7288220981094479,0.02491826016758303,0.3359162244206402,
 0.03261944682412221,0.7280874492566376,0.18379145965552512,
 0.4553654495365965,0.39769833640458296,0.65307602855779,0.9456229716843778,
 0.19110348639293961,0.999100966691627,0.6711938120526146,
 0.013127634864186555,0.45523019190952063,0.232301409352353,
 0.8364515671378229,0.9270713265885749,0.8123874502848567,0.7179305183120093,
 0.36979574074309773,0.5801551742327956,0.09594463586985735,0.407463435920448,
 0.20001551975012377,0.6603945576350895,0.8290059945718441,
 0.11542199359461092,0.5461651052991127,0.5861917417983997,
 0.28502025782280227,0.0636032845672625,0.5912246925068584,
 0.19522429907643857,0.48583072677493844,0.23531545675352117,
 0.8285374403821029,0.19287760166407342,0.29864546984431395,
 0.22795495344366712,0.10988192013569453,0.6239045623590682,
 0.8489074314395991,0.20012276844220112,0.0887137144755592,0.3555866453595069,
 0.7902496167343771,0.735993925160181,0.9171747596062284,0.8898823536871842,
 0.4723285850369209,0.3640509860804999,0.5158047456024849,0.588511519502862,
 0.8283210513511198,0.6351266054139353,0.013983799162984756,
 0.14029436054135125,0.6834737056723986,0.7715638914025275,0.9731194386069276,
 0.12017534417325526,0.18043807479624696,0.8526099146450259,
 0.4340203477213508,0.7364909388185897,0.6005974158062984,0.7977531505379476,
 0.04390326498310326,0.006697890910986604,0.7176864561827395,
 0.677166187579929,0.21092440779425592,0.5113261481812486,0.7122356834854671,
 0.6865939579996757,0.8431724162353059,0.8289061078266204,0.21488326683390702,
 0.7486193136190052,0.9191816042458169,0.5714437110173621,0.8233429267151903,
 0.5581025140168162,0.5301607991884975,0.584609991731722,0.8051841022701902,
 0.7630194093325107,0.5692682457334355,0.41347887426601304,
 0.03771795636328379,0.12169395043809295,0.6521004437070994,
 0.9815636083584471,0.05844263818282114,0.4139437277926661,0.9057159627685135,
 0.1923144618064896,0.6553595970531365,0.6004561567253136,0.8939126152048655,
 0.8178709023395925,0.6651869788002278,0.9697619301916041,0.9019523723064453,
 0.11640103261093715,0.03592135669582208,0.7685324249647146,
 0.24836785955280294,0.9437370339586338,0.18734138029285863,
 0.8758379846998547,0.7148041990929724,0.6443831212844583,0.3219907738438369,
 0.3357548491775938,0.3129706127205032,0.6030325490711361,0.4605026076507981,
 0.45029535263606213,0.7156952623823304,0.6409967921791426,0.6068940431684384,
 0.28271812082628933,0.1766361629277703,0.4331536384346314,
 0.36742184268871647,0.8154386161920493,0.5445175496216987,
 0.026624624311277878,0.4230345815491746,0.06351548267186058,
 0.223864785087582,0.9935475914272459,0.4798729401012962,0.614158484180544,
 0.22553262093687632,0.4515091263806317,0.4364847537964567,
 0.13015009013375745,0.9480871475959723,0.20305483918795697,
 0.35381939715486665,0.09759577788916096,0.5782350990960853,
 0.7422825629218609,0.11331915889649302,0.9099750781193907,0.4429746245208189,
 0.5711024715219766,0.9528238143933585,0.3874707614674143,0.25917569424953457,
 0.0772872557717309,0.7747761015771981,0.2412026565341343,0.7706931546427732,
 0.9400158923473212,0.717641810224302,0.33997673776035486,0.3387488117451709,
 0.345654781295059,0.8279889933414365,0.5594143567136354,0.5390745321072147,
 0.5433711221416472,0.48532439049336606,0.8231756674220588,0.9597674976944336,
 0.9398360762580537,0.029785189586124194,0.6181502875398688,
 0.19721896511295645,0.5060235927138775,0.32639609212831866,
 0.18478171414381728,0.9591024608949488,0.30270223460616374,
 0.8490809381258593,0.9047040020486679,0.1354644372103312,0.38207977714393526,
 0.3470410054596986,0.06927084283384899,0.5578311266565458,
 0.14466437767670848,0.8522228497568097,0.02561613095631443,
 0.9357844305776196,0.6238428581328733,0.9831145071153968,0.7649362621974642,
 0.4880253573849399,0.24429050387609508,0.6043670240081447,0.2847631256926674,
 0.6698600119040347,0.028477260989979603,0.06467252939359791,
 0.05448642373089074,0.8977379201251726,0.004765594676468199,
 0.511674432302728,0.47614985844392765,0.9555299472049267,0.13928056220893648,
 0.6800620845246228,0.4653766041229127,0.8062382394115464,0.2689423199282387,
 0.7637709489024179,0.6613611288529223,0.06160451094329522,0.9866426228866594,
 0.004327275414407805,0.5722397860443249,0.3710516099043695,
 0.8859527277191066,0.289866533710045,0.18491490703796853,0.1226164745763052,
 0.2057289098911096,0.9522048483128424,0.40507635732642255,0.8344560542834661,
 0.5774719776064696,0.26106877781449667,0.9994673762201837,
 0.11079779939209222,0.8968217826218401,0.3135768439253934,0.6509692722135397,
 0.18303737690354815,0.9698916014012531,0.8629343286663067,0.1557233592863857,
 0.5884962779833607,0.6014812065674333,0.23211046378474842,0.8792342141079916,
 0.48704465252543083,0.4734049941141132,0.5064934768160355,0.9853428767124203,
 0.17175145559315874,0.08361050179731822,0.5515060048804373,
 0.043921837697384336,0.37500068276042553,0.4181036133170981,
 0.6706322277173249,0.34591225153245786,0.67633751881259,0.37143671620980645,
 0.11561217732988027,0.8176323329608953,0.6189592095423357,
 0.04602089813540533,0.8035901470486103,0.24864427190163474,
 0.09101415946954772,0.7432056994703196,0.8831334978914702,0.3804491032667364,
 0.7950414939751025,0.5158797470498896,0.2763835792224507,0.6410103973688834,
 0.7162042121739451,0.6242963216869468,0.9360597160193023,0.06140505056361456,
 0.6122536304798812,0.6083701101627002,0.4107055456376688,0.36857593995866744,
 0.3072308341737051,0.28498045872064737,0.24092638117898213,
 0.3939680138360717,0.8630382213629489,0.25476154161758635,0.7503105377526087,
 0.44203327675177395,0.8889866009316845,0.2547662301004472,
 0.23487903372071617,0.9994605532784329,0.7169794177824982,0.4479844298492963,
 0.6328924939882814,0.6122193624800872,0.5025573776552074,0.7464980443058917,
 0.7833514048717791,0.5971155011540845,0.9410905021744627,0.30140995591438935,
 0.033084614014871394,0.8124823047184434,0.3659940855329539,
 0.3851101239705501,0.2019049226613201,0.9099084829342788,0.36798057508196635,
 0.465822216345692,0.9805475000901026,0.12043520822069231,0.00410110808636166,
 0.46727017565833373,0.9013672953637771,0.6874602571365519,0.7262705188027634,
 0.4902193604847721,0.11552054478814322,0.9745297235588541,0.7814080417389273,
 0.9352218312098649,0.9936678627921305,0.8604918164508651,0.8341851980662381,
 0.48798734718071446,0.812995616196243,0.8629085385292794,0.1686049276295537,
 0.746059434612373,0.8299604265613766,0.3331267740651518,0.6908243022110842,
 0.8102207250104535,0.6465286277005828,0.9850177040180693,0.47806593963974575,
 0.9932197537567856,0.9965363824855445,0.22102203594050462,0.9525458379064042,
 0.03044002018453762,0.10942903253472169,0.27371252110979105,
 0.17371005543951168,0.6018375116877384,0.900751669099905,0.612032691598794,
 0.4188222143878324,0.34383427864381944,0.007097660682124807,
 0.8873594307578055,0.5703061901446065,0.0672913244396195,
 0.029220915289085547,0.11433475167271734,0.32365784547657706,
 0.7753576872237429,0.7421221975492887,0.8018865462213765,0.38618896276326664,
 0.14847978074371104,0.5917771836356797,0.5844913831339016,0.2113735065499297,
 0.16389962421836968,0.11757401888468844,0.009210577632419836,
 0.5401926179371547,0.34629064989608715,0.8680201892333755,0.7663010518490092,
 0.9190237680788428,0.6381086007568956,0.8830778562324701,0.1937798811321585,
 0.2788279193938622,0.3358792424371322,0.9598522375080987,0.9561193867338361,
 0.9498752878857528,0.21479130096184595,0.6293133489998836,0.6898716720539524,
 0.27843564275416344,0.8097875688176758,0.44608249979579584,
 0.025518726653431467,0.7264152028942954,0.12214415722961613,
 0.31628762794841536,0.347226446863133,0.6524978263723403,0.7250516745472335,
 0.09543829888395239,0.7282647971224434,0.6125647854803536,0.7862773321151628,
 0.9525822577172363,0.13985806921653055,0.19577889431843942,
 0.8743924940510818,0.7470202720690041,0.5623825557007729,0.6820654001202421,
 0.09051346595676046,0.2967274674403745,0.0527060774698479,0.6985688788063888,
 0.3830923249552006,0.8255310379067416,0.5158397897658149,0.9163472745585247,
 0.4681486194306389,0.4996764357004003,0.5353362997271616,0.7049782668006798,
 0.425970732292262,0.6479483925116258,0.7366331683267773,0.49841666292957343,
 0.15774569198076627,0.80680494394414,0.14885040154710028,0.541724772738493,
 0.4152078559393225,0.16536604161942314,0.45635233869851,0.494915562070669,
 0.8108125592676504,0.8772820011923999,0.33775401850710907,0.8398084216158965,
 0.41542340737313976,0.30110022886415844,0.25728355118959834,
 0.2616152112350568,0.05904485444376795,0.783547043727756,0.8965019230658973,
 0.25885934501953745,0.5761169278330522,0.5874397247553066,0.673152184320029,
 0.05057459506101902,0.6302496573040305,0.894004697086836,0.4553642728061187,
 0.8050802691480591,0.46235306392634734,0.7491922271902721,0.7898141218269735,
 0.9164250914098883,0.6032182540380737,0.11266005619204755,
 0.02956148268530856,0.32690342475427014,0.43444975471311686,
 0.07621264224068725,0.4247448070396118,0.36943925245969866,
 0.9587401989437293,0.6501022039809728,0.7107106083354388,0.6503046801938743,
 0.41109033296893616,0.14837316977854043,0.7936986841669985,
 0.7605378429426823,0.13890475799198676,0.8087706998383088,0.7409885095578469,
 0.9299346639226187,0.5894754010018081,0.7465243044350717,0.182511273212139,
 0.3467450816469294,0.12798285890454986,0.2148449621116082,0.2945416679690799,
 0.6891285837611305,0.00994373889964717,0.2863425157761793,0.97313359568242,
 0.5517868418372625,0.6529637537866114,0.7340580760445198,0.7172694969805975,
 0.4173394914480977,0.28205585942391365,0.9490949054590023,
 0.27399353259490133,0.8968675867659219,0.615386370130391,0.30884607413378107,
 0.743910307687942,0.648784802915814,0.49292911361074787,0.7664053785565219,
 0.6997275332913577,0.39896089831621195,0.801451780247713,0.27913797515656213,
 0.9738133194976936,0.04235354187943241,0.14916210741655966,
 0.5089903982070427,0.4646389823298144,0.5886753471047321,0.5317780656688681,
 0.4985844342476585,0.5552940470085747,0.6741516593658721,0.7331848770342799,
 0.1408139530203345,0.4025974807874032,0.6064704490913829,0.590909448503148,
 0.3272660585958729,0.1436103025893165,0.23220212189376244,0.5807449235569223,
 0.057526752146700444,0.47172581569303385,0.6981495006844023,
 0.0583159344901808,0.27479145384108616,0.7310269966484241,0.6687783489147363,
 0.8099737241459379,0.8013278013482708,0.8415361157725489,0.4390292617568756,
 0.29799347836261825,0.8381299783559115,0.38873933191525456,
 0.7036886410794727,0.6077102056193051,0.9041150478495108,0.30307690914008145,
 0.6638340001713463,0.10947437871570187,0.3219334835455189,0.6247069123945483,
 0.5799957225580855,0.17773646312199776,0.4780170349428261,0.4581590962631844,
 0.024964540187725204,0.7329583831879772,0.922835137255255,0.5194283238069602,
 0.6151221265992867,0.9505631646237263,0.30712767578571776,0.7029371921570421,
 0.0613044336365729,0.9005860350458791,0.04948501320851206,0.7983455771091579,
 0.9890616016551887,0.16836105239225807,0.13522119956844758,
 0.7623308813791004,0.48930830103281486,0.5992603497871077,0.7254346382386304,
 0.055575056185735994,0.3380112647879977,0.8143769336277283,0.671649273443857,
 0.215457969520799,0.006651485953309222,0.914972963479248,0.37286738410561626,
 0.4043370923231828,0.5197877149117103,0.464359761225128,0.31028175670034885,
 0.3948806702456249,0.8933456638064978,0.2381621055621327,0.6694773602978215,
 0.7507293416272944,0.009144275395198909,0.48297416368553636,
 0.6129699722266032,0.9315637811913021,0.16131872078197185,
 0.007334421409439962,0.23514978395780473,0.7808060248836319,
 0.8516938436209561,0.9046567034831339,0.5796941026688113,0.914342124585374,
 0.29510950267962566,0.07677987020735821,0.5565911469878745,0.271154240856213,
 0.987323007080114,0.715300197125166,0.590907977879331,0.6743368476112953,
 0.4502408388680217,0.8781088315178284,0.5357026968490916,0.05146513698104305,
 0.6844873106047574,0.5258664142311977,0.5794709216853005,0.9127176054224952,
 0.33353455021809153,0.6800883766947895,0.32054034414108923,
 0.7449617585525382,0.040017222133589714,0.8553882603085795,
 0.1027029879032284,0.9142627203648653,0.3570865080150777,0.6344666388770642,
 0.7765709973435869,0.2536122398932803,0.20414340376085816,
 0.07639768878920983,0.7178662682756729,0.706209129927731,0.4440499851147961,
 0.2530106990914046,0.7170129369260199,0.07145124733027564,0.6200994732689175,
 0.7118889512756633,0.5942173108791804,0.9133387890567284,0.582945594198324,
 0.5376064756199376,0.6839123249283646,0.5869285148016191,0.7992214862725258,
 0.48834130726128544,0.9300807623927283,0.08250673727586733,0.427753510501776,
 0.5423709599623845,0.5886351519005681,0.2912632049760504,0.5795677030529996,
 0.0279058898030895,0.46465945729539204,0.45046551339131735,
 0.010469393725341058,0.06206941377144126,0.6243483856396472,
 0.5045964217333928,0.5277972045371481,0.6752595031458689,0.6705691296994942,
 0.4434071158896148,0.7339478480046383,0.3534149618243392,0.15576454364970893,
 0.8310585619490362,0.5799977132600845,0.6985430087550268,0.11990955076786924,
 0.39670400201859524,0.15671773742389783,0.09576551063724081,
 0.29026549178310634,0.41222048024949576,0.7505185645454819,
 0.9877177595612112,0.1610153673252125,0.7680321144231023,0.7758618835003689,
 0.5531630178508079,0.8002620816519745,0.7782950415269932,0.2282229956396874,
 0.731700031767994,0.25900778119472667,0.8221767041429506,0.827954513656231,
 0.7200886991971106,0.3603721259559147,0.5036929885046857,
 0.055619463619901444,0.5799058155803544,0.4789800610719961,
 0.7603237264933717,0.08920420761997305,0.5575221779498432,
 0.10049820827669376,0.9071107707812068,0.8605547935409139,0.5187943487570348,
 0.9671653658536977,0.8918136545849454,0.5919458680359592,0.500832843338903,
 0.08591611903638252,0.127326492080164,0.26729125957459665,0.3990795214639282,
 0.264531965636703,0.30791150174882687,0.7457830002004425,0.06188876271052801,
 0.08975676921008413,0.8288230161031883,0.3881825098913402,
 0.08630234401012138,0.7183908069148974,0.41451099864542695,
 0.5177540712213394,0.5182228858591236,0.1099736869933513,0.8792424457480399,
 0.6461599623577475,0.08240407886243983,0.5769442201916215,
 0.07871278082045319,0.5959222138920217,0.042579458099654355,
 0.5264320229101386,0.665434054623358,0.6584902048398122,0.1118517610013674,
 0.7587683062653554,0.3253892176647255,0.8640996671489418,0.7896615061902268,
 0.830480340037948,0.7478907262957443,0.9866631037948626,0.7290306632620872,
 0.08341437137539043,0.825568821174149,0.47660253809985564,0.6481352680202365,
 0.4888667564241901,0.9443920785158378,0.27797693205625085,
 0.17422038586648425,0.17156521148421833,0.8480414014461533,
 0.4118407699375263,0.5735988712206339,0.3505648809516665,0.12643619153191687,
 0.9695717027244163,0.9597974023557417,0.40208153726669416,0.7086500184567004,
 0.7806101073705665,0.3906001468842303,0.08780377926217575,
 0.005461997055624002,0.08396152077744357,0.460999650076111,
 0.21858872962017473,0.5158592627735731,0.20088250531642327,
 0.7004826328656695,0.8454994517741581,0.46190968317744885,0.6417872650190516,
 0.29963640633539446,0.7599107848704003,0.835031410616571,0.14340796127687927,
 0.38974870969782216,0.058210154278655835,0.9213956131234984,
 0.6129917357868963,0.25773097368223696,0.9334982700933784,0.6191695272426477,
 0.7567760294647344,0.4150262956513059,0.47320489631230445,
 0.36828834748852257,0.48246272847510185,0.7025170762471232,
 0.4028118898019952,0.8671973221665181,0.3016130483288574,0.6274736332989069,
 0.8590678915675919,0.0468705167501301,0.27719083750282436,0.5283077575832191,
 0.07247685633699119,0.5097730900170209,0.4906779433441584,0.8959229060899505,
 0.6507101347846896,0.6020679547614287,0.8534959618416125,0.49429368859448086,
 0.13749374341332798,0.028857168536130295,0.35585295614023615,
 0.5548701425128746,0.22828147799171994,0.6274535279250327,
 0.38891125308934826,0.9023545279132539,0.637288894197814,0.3957074805215821,
 0.06511607738396274,0.8551503801089498,0.09892670107072343,
 0.23198058245854847,0.4733660767373302,0.054515850818304123,
 0.42644255964563416,0.9658666477151073,0.13571354459704854,
 0.21549798775446705,0.8153989029642414,0.7595136653360557,
 0.31453197003170963,0.036539683309434134,0.8065730140768328,
 0.3220641529258851,0.45020017119034483,0.2758004365066383,0.6067055585360814,
 0.10916045653224482,0.026964272075020612,0.02830598598455103,
 0.545065232688352,0.4862914179771728,0.9103395221131688,0.8000302061938942,
 0.1279494898195852,0.13462850747167288,0.7478620002329833,0.2741884168066401,
 0.4102685722202425,0.11408992194950651,0.4398797965103639,0.9826198966586677,
 0.26003603053568713,0.9719899099407371,0.1891120969048613,0.5875109081171838,
 0.21727421248637857,0.5215321941386852,0.41976674447692497,0.668117864933036,
 0.9651021265935535,0.7296575701748376,0.7952124765878088,0.7068969145140775,
 0.46738111573699825,0.46086787507081306,0.7950794262538763,
 0.3372516127905054,0.6439733409728479,0.7762602520667572,0.04115660924167197,
 0.11101891745967052,0.2234147561677864,0.39962030525278425,
 0.7328057496469086,0.7746284486673214,0.5669582472244352,0.7040359388794611,
 0.44295842932173124,0.8676843021412153,0.48762603843837815,
 0.4999388879903416,0.6421205528393037,0.22630450629604681,
 0.18605596013279868,0.5550419872301844,0.5881507373328863,0.5420153297163679,
 0.7914135746866685,0.3262943227303753,0.9347241783377347,0.35742091281828114,
 0.2330594012683893,0.30278170703750606,0.28311125126291026,
 0.8602648174330005,0.031888356663702266,0.8352432980172505,
 0.6876305024150029,0.2072919009311689,0.7259069941854656,0.9514490979313851,
 0.4426712772601419,0.27555012608966567,0.12250276213561517,
 0.34744372907944654,0.07557169167210631,0.7532132195310042,
 0.17067037225697024,0.01151733198150895,0.1900281494531263,
 0.9068140793014771,0.3471408986005513,0.21027084880058244,0.8783483838459564,
 0.8491787680131424,0.17430168123458611,0.9434666416130862,0.7538923954983876,
 0.2045689487555088,0.6270503203539918,0.3409963333154553,0.4250614525057804,
 0.30361696528314097,0.889952121037123,0.30628250517655486,0.9235948963894044,
 0.6835267666721384,0.21648707168744563,0.06918681672217908,
 0.5745094699745078,0.6327417351461817,0.6299103350013111,0.04831014384307453,
 0.6346223347011467,0.6531109908016329,0.20605784514008085,0.7450508240303865,
 0.8527659542934154,0.8194297466892214,0.5170658283517927,0.9345812972680891,
 0.043177021226619594,0.7107283382552108,0.28483853824823957,
 0.3435408087795071,0.055752490773909424,0.2999743041102456,
 0.4378796998365271,0.8174078349017592,0.6186412629647664,0.2206192670735625,
 0.9534506268541966,0.8829830093605657,0.31890094049841056,0.4892598897198881,
 0.0960395986513497,0.6191751327257092,0.8916479707538564,
 0.034259712703138434,0.5014591124738473,0.5580810626857435,
 0.04190797930391299,0.6131936372795093,0.45176837072485354,
 0.20891879384502343,0.7075556073373145,0.9463904153999348,0.6847332820008759,
 0.9677515256879419,0.41613928236069553,0.18251747919439754,
 0.31587572596679603,0.5756693232826116,0.8033764262116581,
 0.44068217017934175,0.6790686956510782,0.7342941094440703,
 0.012362579747831726,0.2424583115124911,0.6347228403796099,
 0.4275935460864283,0.7397489697847875,0.05841751545435625,
 0.054784691872680424,0.807705923763725,0.4247075764620606,
 0.19392120911590982,0.3621524179157798,0.8611683932862065,
 0.35313995014457933,0.37259778081264605,0.8993426856150795,
 0.34597912623103555,0.3530684892486281,0.3199365746755559,
 0.18314063653274493,0.017825815885253227,0.26117914295638345,
 0.7774863104456475,0.5466652903721837,0.1409440795967517,0.623294966854779,
 0.10847519049963839,0.363183128870202,0.5532632878803216,0.5040125354618192,
 0.5896711433426408,0.666373217398837,0.21751920414362358,0.1379000566015628,
 0.8259769434810209,0.0401303942402389,0.8778254843561845,0.8618075529439607,
 0.33826501334892267,0.4185341462657476,0.10149672212938943,
 0.7766991589895027,0.9096929565268302,0.23983822946304278,
 0.36209444759871945,0.03933430072589261,0.08745915832739892,
 0.34154545490937194,0.042203087001644324,0.3293775628976915,
 0.7104564096657039,0.3087964612650941,0.795348340207504,0.46749873034519496,
 0.5847280405810504,0.9420149681499607,0.541308000678077,0.6500070068131687,
 0.04960591038035367,0.44315404010087545,0.4233423282188089,
 0.16791142723281738,0.4898254340859265,0.43929680595490816,
 0.9845931470424298,0.7827933507900001,0.33781081343645414,0.8300052797210438,
 0.6759297568346025,0.30999650742008733,0.13432278344211712,
 0.4737286586107077,0.0030010109933755746,0.15143042542570595,
 0.817783351863962,0.00898228414154012,0.259935511001291,0.9882302816716945,
 0.027494429116536434,0.28958825662925425,0.324378861951125,
 0.4849624615032424,0.6488338120100688,0.9357422414188192,0.13547030759453838,
 0.3347855543142667,0.5197291085256944,0.4505323181410781,0.943632438533542,
 0.5483574009731101,0.2989843833570939,0.11757520635657392,
 0.03773214295417138,0.4185018752749208,0.2736784547602994,0.6285097493724668,
 0.4389180701008314,0.3236549365096114,0.8912161739218158,0.8139038902305574,
 0.46810640412467075,0.7031500805206788,0.02249901045917102,
 0.32024007460600423,0.3592092273947819,0.6142567924133628,0.9608443211941887,
 0.425989532023885,0.3296491274996982,0.04581489074566436,0.1495153165747306,
 0.6166943201799908,0.7959635430144828,0.5078281357082159,0.5779772920702151,
 0.7027764583209225,0.9797433952763331,0.3193013027820585,0.6550750327911022,
 0.21885398834753766,0.826114846982266,0.9995049899118061,0.32805818037791745,
 0.6760841125741135,0.8769290963774231,0.6118594062573866,0.9068724995166645,
 0.7950368584927099,0.15410941233035225,0.648506164432067,0.3235723572875219,
 0.8827097611892234,0.4033505873637001,0.21559500495806094,0.9105338910253745,
 0.8760945029840584,0.9146041997512298,0.7809041869464552,0.6909084171853453,
 0.7147021838596924,0.8853630579686593,0.4775565339467239,0.5215568320897638,
 0.07396925705587865,0.3671205414620776,0.9257731933660769,0.9662171129859088,
 0.25285785104144565,0.8320295220687932,0.1482307449886443,0.2400100680643117,
 0.8914090990183619,0.4582065219102517,0.6687631182410141,0.3783787698578779,
 0.909678550089412,0.07402705305442858,0.39560084133958284,0.2403769777244158,
 0.7417282202538636,0.974444711281451,0.9063197254436017,0.43184665847040704,
 0.5098198360240003,0.8056075518060555,0.9449584681242341,0.4395799454803978,
 0.26191619818523515,0.6008726043578022,0.3068940848674122,0.0780599365552066,
 0.08269361055320423,0.09533540687063091,0.3240091747737154,
 0.47794803807437825,0.5521802320960072,0.11120127049064044,
 0.19549369780355297,0.21888356554828525,0.33031436758333643,
 0.20012822954020593,0.6164210094957272,0.5821870857206933,9.7893285736117e-4,
 0.7868986520389795,0.271210414792843,0.9394525838082285,0.814648779379508,
 0.2543820700456286,0.34747191048688597,0.07553265170082946,
 0.4472681199574302,0.7889128809053831,0.7699355189511066,0.6471009192428745,
 0.9750257826863709,0.017098847630087177,0.20518479826262115,
 0.14415459067140812,0.4604946570821751,0.12115212671847186,
 0.8314404255973181,0.3027347802824868,0.13899718305326036,0.8301222524524481,
 0.30835597784316626,0.1845810193722781,0.85662464253417,0.5083869594984007,
 0.48462031591779264,0.4589794374371591,0.026655713288503136,
 0.3579954098808069,0.5300305049688165,0.2463665478143493,0.29187364470929,
 0.4991256129793762,0.8907887338629463,0.12516142783139395,0.9077292497838211,
 0.420206486774112,0.13472340821342077,0.9037296232887815,0.45480005966267434,
 0.1603252330266839,0.8510510967067472,0.3870118075305976,0.7732078521110449,
 0.47074614659568403,0.4799541334447627,0.08813120031878374,
 0.19659905691032886,0.46165802012588053,0.4989402300005593,
 0.9518659164933866,0.022030210485302293,0.7215114368633477,
 0.5604645005480884,0.07866000875526613,0.996133577667525,0.06080909091225428,
 0.6825386514270043,0.6298431320020041,0.04624982918820442,0.1635758604384394,
 0.18616049481748487,0.5149361178925193,0.7402136570128236,0.9273064452939762,
 0.5088994975888207,0.6142457467080941,0.017243132232165226,
 0.8093040006891898,0.693489553705231,0.5166621298952266,0.8316044751291648,
 0.11575635183076483,0.4965124612552566,0.47545883046223103,
 0.5120238239105845,0.7763030728443912,0.0350757180502258,0.02059213760379952,
 0.5067613256758884,0.7159173763814509,0.17130406464081382,0.3123195568987023,
 0.09032476722667393,0.9750813674240766,0.7546787097646349,0.1790485078972608,
 0.9602699263537882,0.327246898329841,0.6552263259199756,0.03878044476818743,
 0.9885292163930647,0.2078348217289323,0.9389062174428289,0.8409260355392046,
 0.6503070825291112,0.4591292561821366,0.4739592494520837,0.9968426296045398,
 0.038169548649911955,0.7514054791105016,0.5933413464701516,
 0.6963223016252412,0.3635758813385229,0.4112898164067921,0.989777746364484,
 0.18659753319479078,0.024882990003255778,0.9633601569431166,
 0.37508918463306373,0.9942643978341482,5.662335727500967e-4,
 0.4688353217648732,0.527134416065191,0.5559344562211906,0.0329001007342985,
 0.22704645322920403,0.3807694321747306,0.46058497099344886,
 0.06772744513659612,0.2774152824355596,0.37564023484676,0.6566794684051017,
 0.9594696929233244,0.8036221331580937,0.2204355893492066,0.8815641488654122,
 0.14493393633242557,0.7552162078287888,0.0385034273508037,0.8165748719308143,
 0.10160302906843843,0.6841901832234685,0.7907310569145538,
 0.48095487155168726,0.34250285484407594,0.8428900269025542,
 0.0748381141329002,0.22928325442844633,0.5099538928882381,
 0.43678874109716825,0.0770791725427703,0.5768661373431131,0.9025960431975866,
 0.15551478246412132,0.8094160732595445,0.8919628419064071,
 0.04706327941034549,0.41938477111048966,0.679303782166829,0.7981006300448507,
 0.20482129383193803,0.9965311008155315,0.24493291630044223,
 0.11064684651965873,0.5294411866332452,0.1056381102536128,0.499076371546345,
 0.5800510532296488,0.8260107345506097,0.5680249522840021,0.5437496817096572,
 0.3878749679790021,0.24385414448474618,0.14891996355192527,
 0.18494088864566272,0.38097647309533766,0.9204265477123895,
 0.6018233181555599,0.9198079318447329,0.17277783194469643,
 0.07887026129159125,0.9428397412336955,0.41102595010753673,
 0.42853574470473343,0.01926543654639301,0.8866098649763446,
 0.8496188515369986,0.055008489324681276,0.9980839765726389,
 0.9635786808788129,0.10001359958999168,0.12056100740734155,
 0.5984226245838482,0.53293308993843,0.599194333384704,0.15785538526738763,
 0.16690423988452907,0.06123470185978053,0.8155655441239217,
 0.5875883604109245,0.8037251208666873,0.36451757241358196,0.5616190494100356,
 0.8562872992496326,0.13656737143791664,0.5015014115557691,
 0.27405437872936067,0.6332442903924665,0.657794494523139,0.6351973856260456,
 0.5073662301979728,0.5696138744608268,0.05972929616205902,
 0.14638329872244094,0.17883091004466634,0.646865552198486,
 0.05306381501191859,0.2106237433513629,0.14654971537068695,
 0.42403190885291264,0.690311994381075,0.5600989837330802,0.7672484671841154,
 0.03216379155595328,0.4054292850609307,0.25774064429767307,
 0.9473843255083175,0.866199286636001,0.186581051120014,0.47501047872388213,
 0.7690649044258826,0.12587535550538176,0.5614721942608562,
 0.20864220123747523,0.3872733734818946,0.20318753613693719,
 0.4611637777639871,0.7521183328850218,0.5138358189463442,0.7660003978141368,
 0.5610534571062735,0.6487126413017823,0.07723259729814425,0.2918436194392834,
 0.2061673833433978,0.7101942643628281,0.22287266785093984,0.8231017947133237,
 0.307802372524438,0.6801532009722081,0.9458636878850198,0.4378555190539686,
 0.8371738252509698,0.5603771578449583,0.9606366086863604,0.8986532879958559,
 0.32643412271713745,0.6543387797427189,0.09919508023451518,0.434702630167229,
 0.22496279554393794,0.46880656731616244,0.7813996560892176,
 0.5355684372831822,0.8822240359528867,0.17660316577483248,0.5518969499095193,
 0.739674410615265,0.13061970437366777,0.06823184708193886,0.1719570586675343,
 0.24870559236924805,0.043843314890505414,0.8678407248329294,
 0.09774523063166507,0.5617444240616711,0.5520657796473233,
 0.39830273166150654,0.2994776454863124,0.16651860392470985,
 0.7051283848188891,0.2602130746394825,0.48973061971888543,0.8422620746915837,
 0.4329667004481106,0.19296657322086308,0.09868502693131909,
 0.9383949280362691,0.008732160553187374,0.1409549135463254,
 0.9040706667538668,0.16289585995149614,0.6978513458984079,
 0.16638116586020912,0.17325183210742678,0.6773485803283257,
 0.22204882440015272,0.19092550216075688,0.9717494566716746,0.820895995686451,
 0.056779627696417645,0.07684343306511177,0.33291957907397296,
 0.0954936088173125,0.1607717949584404,0.33065311893410176,
 0.18918663828219473,0.22604231527605867,0.3186423530426489,
 0.5656427873716765,0.8794614797688163,0.13658587543934309,0.6686679588223355,
 0.35398191496431364,0.7082548539680211,0.5474819543490601,0.2640898696084544,
 0.7158558806010358,0.12430704831514527,0.653948669525115,0.10079233121089326,
 0.45146936973620455,0.9528480057688845,0.6170755831052139,0.1753501547401647,
 0.7671514517250542,0.4026707249996575,0.6783039604669677,0.5317203462333177,
 0.2718460644113001,0.42526316345529036,0.005878417155081861,
 0.2898591520276096,0.2582915238162157,0.5094562290269359,0.5387247619831043,
 0.792967670971263,0.40421800868589886,0.9098613384655714,0.8352294340725139,
 0.1724585820714638,0.5477376172061094,0.739728465953348,0.8338000513794102,
 0.9468248998151355,0.08505117769668213,0.5490971488518481,0.9605392191790278,
 0.32912505042080864,0.3776305429144361,0.14753685092897229,
 0.6597903877966068,0.26302069514565585,0.29728249745056623,
 0.39025550083118765,0.133058796145898,0.34705249128948235,
 0.46391982423186906,0.6266693141837593,0.6537330714514914,
 0.09851030931190552,0.28065323323644675,0.8433759631515769,
 0.2169029888573608,0.32387501230304705,0.43533371498055096,
 0.6029245661686873,0.22745463105134978,0.1895678240936085,0.8400214542346399,
 0.17507227641064227,0.9463714357891465,0.8642880863358248,
 0.19665898388636838,0.9450288090057319,0.09143145933872054,
 0.4652270844053952,0.6345693786622871,0.19207360933655448,
 0.21774825378161344,0.3722278236543698,0.20625743061657464,
 0.4064998328548952,0.022136404869863013,0.6361178797747846,
 0.5447660857121163,0.9759511703316397,0.6775021391917349,0.4419916737367492,
 0.3228935767682567,0.4409694457501122,0.368893546174033,0.30906061650263694,
 0.6472756575632377,0.04065947355333288,0.8797185401365409,0.7817806918119432,
 0.6093909010418471,0.18403324898206397,0.7195292372396003,
 0.16783314849621578,0.06952530171931193,0.2667509804587076,
 0.8629605725931946,0.587090001833265,0.1843313329190409,0.9330476555560823,
 0.7982576685032783,0.1904636634453145,0.5610373129105432,0.7714407756370094,
 0.3207535507185506,0.9723099304607097,0.00912985253088161,
 0.08144287745373069,0.8038272695095537,0.032220324021345004,
 0.7513832159335511,0.5963679627702528,0.8281652977967027,0.5640498728424614,
 0.03205415744540663,0.2846101923291857,0.6333302058879553,0.6997906406211918,
 0.6135528958359598,0.09207975642671595,0.9070824399450409,
 0.08145204273716355,0.07379595900482527,0.12032137198477377,
 0.2792116948398936,0.799002757907005,0.9864850079810633,0.5386422053043897,
 0.8547679985754215,0.24829990483964082,0.16671615776104864,
 0.6457500140925723,0.6281886951956266,0.8576841000164261,0.7710777418756366,
 0.7893300938960566,0.8851089386993345,0.8643989280422177,0.29021467818277524,
 0.8267102685092825,0.7714324424259458,0.5961315692830544,0.9759536925269832,
 0.959575479052561,0.20672223096221654,0.5171591279691102,0.8187392205227265,
 0.24107698236206554,0.3868424220325498,0.5383035281826078,
 0.43047909316123234,0.14837309314171776,0.3369320242168371,
 0.8589311721853935,0.8516930999618028,0.3073282065277336,0.6497812019908077,
 0.07196330258845818,0.37405839255275886,0.7764465469533057,
 0.4165479715863972,0.7882572309037105,0.6965605530393731,
 0.024750501083280474,0.7533364898129548,0.11070522156602036,
 0.9102397925276376,0.7346412634215526,0.2297911836337001,0.34207517925494657,
 0.8708874005448268,0.2880565085611948,0.35951153879355136,0.6547806878094975,
 0.08024490733508305,0.7123710036901828,0.19915404505293555,
 0.45991250324236244,0.41031315713605054,0.48982005201292744,
 0.06958668076941099,0.3801204088518757,0.182635097507706,0.12741007716676467,
 0.9919380388057629,0.6734718028444787,0.9418150599748836,0.1945158095681948,
 0.7772787402051358,0.013813307965777621,0.9998185501793961,
 0.5121960044950237,0.7015098955201524,0.7079791088076517,0.840377882163443,
 0.10490711139584485,0.2623415602340181,0.3329214439382201,0.7112979330563028,
 0.9528312707184103,0.8166146353525779,0.8398962540403238,0.5361181509046846,
 0.632294349689155,0.2771455699355103,0.982643077891572,0.5264150650039208,
 0.17017833688118356,0.8959390491028484,0.7546680437322526,0.7446747712615882,
 0.87457381036683,0.8486811881802455,0.7366619629956621,0.992666706266111,
 0.8366895059431964,0.4882412832635613,0.05714755077146805,0.4873009201749847,
 0.5889944302970698,0.5901372064481052,0.1875373345390261,0.10879919429566731,
 0.8348570697062521,0.054133657676648905,0.09221816917264236,
 0.15219268935199315,0.05729555137032516,0.8489164232818927,
 0.26463399515051655,0.18917171369771346,0.2979816833975729,
 0.1510030101684695,0.4342232678482094,0.22591996285799654,0.6923952209576113,
 0.2504547723901317,0.22357079535390423,0.417662738627292,0.40718613313464747,
 0.09154882390326091,0.4834598996232028,0.30255334433793124,
 0.14704007293811716,0.019491635197253387,0.8833779523112498,
 0.05934004821388439,0.4545048906737219,0.6129210649729142,
 0.24031045003669105,0.9146562745585619,0.15897881089190768,
 0.5689594439852332,0.35587049178796193,0.32139115448044075,
 0.2490813730179513,0.39645365315904046,0.41500386292325586,
 0.3290529961050216,0.022404840398781878,0.26259577012581714,
 0.048284804954911054,0.7032897793032908,0.11556323581612948,
 0.8987517303774101,0.83332130216684,0.012940372692411772,0.8826666876597179,
 0.9154362730923935,0.1220744487529477,0.7394536282918674,0.6050423270510737,
 0.4520690793758976,0.10082219128997383,0.6247968793158454,0.5737369275206812,
 0.07608048705889814,0.7757165817025813,0.8372815066977771,0.5990641859727889,
 0.01488322257115493,0.33438501311720537,0.08782458340884347,
 0.23554710706469384,0.016082829898323325,0.9814863935560356,
 0.8634623078154899,0.7288109506980166,0.8826757573446761,0.39289859857299536,
 0.5687884123040661,0.27260968517860507,0.6131448146939014,0.5411789781006129,
 0.565227744084759,0.39577662226271404,0.6526852768621111,0.557843532498622,
 0.45364243910440416,0.5300949759260973,0.2262963553488505,0.9930882838903696,
 0.42779962368819235,0.5217004856426362,0.8998421297420086,0.8935965742970915,
 3.2284923292746903e-4,0.9433965384345805,0.6955899759478694,
 0.4921321106590637,0.7971992554101025,0.29143960674830394,0.751790784490406,
 0.8353622340449712,0.2601178827940789,0.32915423329915305,0.7089728035082032,
 0.6962393196328246,0.12436695518890262,0.8885935799234076,
 0.01348859042119499,0.4969989302446215,0.6169005431462325,0.9755826752963537,
 0.8181277017381561,0.33867153267940386,0.9331033724312932,0.7523169130273393,
 0.7436570120594207,0.6309090451203193,0.046780614207429405,0.520879917555059,
 0.06668762844542553,0.8934347536392925,0.11109352125709337,
 0.5813594728728277,0.11585636364861873,0.4609711297891457,0.0377868094328484,
 0.48581377590980945,0.13101621240051387,0.8078767682234436,
 0.7260693738049471,0.8361818446455588,0.21740775481220098,0.8208582096995696,
 0.8527187642519617,0.34395840292630264,0.4102468679553922,0.9966668803968548,
 0.6878639962708969,0.4121899176701791,0.82375437479453,0.6972087603365023,
 0.0703372010718879,0.6497003987614597,0.3412590230584118,0.3803603699059045,
 0.13201636474519507,0.4552021183317604,0.6904772651987524,0.92501778374187,
 0.022555672433510643,0.8534537083029876,0.2433645684620508,
 0.09669523229708166,0.3315103551935886,0.4585736146524668,0.8326946837948388,
 0.40930068220713633,0.6274616956521845,0.1258238033645377,
 0.003293426447939618,0.1597124215418526,0.1438872280151322,
 0.7326472386298581,0.9021152634259443,0.712366562289311,0.9478245218123652,
 0.38808493254099086,0.516331370560515,0.3135086575543624,0.18946428071299115,
 0.6137624243694371,0.7275595730856069,0.5329360759238589,0.8442322674077997,
 0.9644518154306303,0.3662965614416631,0.24223397242659972,
 0.34637309985961773,0.7019424132436147,0.21682800841056615,
 0.5142537869039235,0.32298116468952265,0.39673274860137997,
 0.45867549187597745,0.010039482600226357,0.3891767454842885,
 0.5483994164199224,0.20169326069267457,0.7364679626564692,0.7388325229482023,
 0.4753881548438028,0.5163905785578752,0.8020540673347951,0.8505159746980704,
 0.8640125726421356,0.3356111100853716,0.5969868699772066,0.6006290348674234,
 0.8576052926419018,0.9109749528227807,0.43014922098973896,0.529111884610335,
 0.7688737706220871,0.3426998608218792,0.8923676033882473,0.7409397101619721,
 0.6532629327357496,0.192391632300848,0.9457074202768939,0.3104148040580609,
 0.7721404894057168,0.7514148413272094,0.851875966663361,0.9752969771216262,
 0.5425225706012676,0.6334730290479496,0.8750675784826457,0.38390318048174155,
 0.49225925723549846,0.5752974650902597,0.30721019761826607,
 0.01639152360354368,0.847065590902161,0.17825124658622682,0.7056542705609642,
 0.9203408486728458,0.09762834941287468,0.5284533238702848,0.951842456067996,
 0.7487708387870994,0.6429753056743244,0.17181264810595054,0.3272803790383382,
 0.29130118608191136,0.06778481351221088,0.6294694347886746,
 0.44194350208054123,0.3100582165262392,0.9498345238394648,0.453077942206767,
 0.7052275252601652,0.15773901335484397,0.32562555062299336,0.989901057629216,
 0.12223934084269106,0.6620588574253643,0.014816591541220725,
 0.17665385495519903,0.31793638425476445,0.7380803046906143,
 0.45994766991644087,0.05608073480762077,0.48466219058297366,
 0.4009851667650961,0.3839953907276037,0.8837014526593036,0.766264694388398,
 0.11949909819313553,0.9847142212658985,0.8552117193052878,
 0.0016429521185716833,0.6535335692339963,0.8082245088186267,
 0.3171232077569959,0.9324769625641799,0.5940199882732726,0.8032948864665954,
 0.6320777019611395,0.48334070373519744,0.5650164976624876,
 0.24668357796454377,0.7432186688278047,0.20212582004987933,
 0.18666427878643288,0.9195918532550559,0.5787333827221298,0.5881210204182054,
 0.6844458230657128,0.04676322543224898,0.5749543527296808,
 0.04675352629283436,0.17113499180146086,0.6052832785729936,0.800302948380334,
 0.4846518882467906,0.42488183688150005,0.19765796216763187,
 0.23913321164678936,0.26063671301076285,0.7681556298169397,
 0.7981121408638647,0.30694451626642794,0.6326633396258055,
 0.16741926469566315,0.07847526145196215,0.8158673222394418,
 0.30250955923066347,0.5662261305268039,0.8235334657459065,
 0.41042482267913727,0.9864860389963037,0.42930766170808576,
 0.7993221613029302,0.18911205087678695,0.08186360360412748,
 0.35089874300150903,0.47347591126743716,0.8097316733646555,
 0.08336138322991893,0.19372386587561774,0.8676693247381834,
 0.5471339489329705,0.46537981805965234,0.03340561488012561,
 0.37070250762606183,0.9953169668284926,0.2676990910646344,0.8017774768344381,
 0.637691656683347,0.4200061809289405,0.2614635202799883,0.08453391709824665,
 0.7261077783025813,0.2540694112012678,0.953767739288948,0.6543255152421462,
 0.49272679799398367,0.3850226377037471,0.986608199458501,0.5166182564594051,
 0.9729728969200927,0.9483279050266502,0.3734792844875783,0.4386439146902361,
 0.831309628483553,0.34069216428670024,0.15657453827756074,0.2976202072926535,
 0.6752452428870225,0.7893270548918028,0.5597405752835445,0.5737407098166676,
 0.6961467785546512,0.5691852458933987,0.013629402095095222,
 0.9640582004046915,0.7546618429982733,0.3088509089992818,0.2262723425505705,
 0.23736801049358303,0.9869634128708191,0.9361823993063548,0.1869961984902221,
 0.03668612062880361,0.2612419304965178,0.15164246090379685,
 0.7745635329466549,0.03670085068009721,0.49196951698019464,
 0.2804844927659147,0.15791568243057907,0.17526772094615273,
 0.7896257602397793,0.8419498104749302,0.7176814815927574,0.6732311888134993,
 0.07454489830389055,0.9080973375318401,0.7774542141823171,
 0.11526276418594295,0.5049241566555713,0.38227323568698923,
 0.8068827590925851,0.7744408196438061,0.6242176034531327,0.5681677387088652,
 0.057161609341280784,0.998728393186298,0.3545196118335804,0.879634396690827,
 0.02842594733891557,0.5097627369560797,0.400761939256415,0.2106187569002116,
 0.8228053919893449,0.49988895387755927,0.6706249867052837,0.2518711452138316,
 0.5061600059797318,0.7694681888189012,0.6263065164214588,0.3037013841100533,
 0.8835759643432807,0.13436993693840327,0.14467406415206485,
 0.7214260901086331,0.5270967821825098,0.568403740200269,0.4571044100416364,
 0.3880094567700122,0.37723905841370653,0.1774875732057879,0.9758383452312641,
 0.689297846471369,0.9695628557484883,0.9798567502574637,0.6344827352047633,
 0.9640329276075037,0.44582566703982174,0.8059787894244357,0.5113647969058804,
 0.33206810662888475,0.15040334919884935,0.31354165154760205,
 0.6096385423716092,0.06514910157181775,0.729409463827205,0.9279852900017577,
 0.8646155179019416,0.07771747229562376,0.8260886345214868,0.2914187255204206,
 0.7590806026072432,0.6653817960039071,0.24136325917957613,
 0.15932914347258653,0.1395098313375166,0.9892296009171284,0.8735271671548402,
 0.30701753914869423,0.8680598110261138,0.18360421247411107,
 0.6871910230905066,0.8198235750866144,0.0557639273113395,0.17389462230374098,
 0.9069430901389197,0.45578181361231684,0.7408310497940708,0.56968587259591,
 0.35192102465221087,0.7557273654368959,0.3888103734251609,0.975135741925632,
 0.9465103251834825,0.7501156319786916,0.09421952064509354,
 0.38809364721092543,0.3492070250287541,0.014495969343589543,
 0.17346029075700853,0.6476836699530373,0.3596923643134339,0.5368777895430965,
 0.07811338218575647,0.5917486350249311,0.6470164818206828,
 0.18897826903110226,0.09173369458056069,0.08075519963622313,
 0.7278229081130249,0.8752251270040777,0.7208847816303601,0.05503464406655356,
 0.018262411852787652,0.024586147875995445,0.07669431505256208,
 0.8306702722624432,0.9396700808276434,0.39953300626336985,
 0.007262902439390206,0.17558550827746267,0.8749132568098237,
 0.2333012123029422,0.7003428530598543,0.42691455260075073,0.799628315989698,
 0.10559212584604383,0.26772171613819506,0.7531621959820065,
 0.6893229248080308,0.12000003617703137,0.71774396012403,0.010190152974883793,
 0.875898967567673,0.26493886087823826,0.08337536375689192,0.4307803517882761,
 0.6113783953797668,0.030930222294719756,0.4663880421371642,
 0.3784105056973265,0.058604329289587365,0.7626059987108641,
 0.5664980656293908,0.7533512555815094,0.6245251972487377,0.49906146173950294,
 0.08237443504003694,0.513227324275985,0.023523892970211513,
 0.03029998239022491,0.31591389375499257,0.567167159352864,0.1825428389644087,
 0.8517663264844034,0.5252388198328519,0.5650076908953408,0.14248312016760134,
 0.3423999660675878,0.8786378168280177,0.831863753962403,0.5536189903041738,
 0.4058526036795793,0.11367601271809269,0.8313030726072933,0.5188373380841952,
 0.7612824841172003,0.7269502966748512,0.45201426541562073,
 0.10446246524063796,0.25974717059576535,0.14862227244171733,
 0.665569198515816,0.07135421620944649,0.6944779207867445,0.32159342885590414,
 0.09682411815012004,0.2850189827448626,0.7001007591339676,0.0651803313932845,
 0.1656895557276219,0.7531270101916052,0.9641158030491218,0.7342628614053729,
 0.5885177053780747,0.5536653151777072,0.8956053620438268,0.1683130054437203,
 0.2201385644637852,0.622890540932662,0.6941094628159457,0.3771163818028993,
 0.16329445054200786,0.21388667429246766,0.7450717251033198,
 0.7135732202891014,0.12048642145415922,0.6276098430195345,0.7454164380956307,
 0.8146760200109835,0.288552767962415,0.9072307988079226,0.8946365420406616,
 0.7180648906282666,0.769170128644507,0.6614300660577355,0.2782521329374973,
 0.7195990428106295,0.8224302657632315,0.8982544826608975,0.5742695618586371,
 0.17379785935599723,0.7905297885460595,0.9529392353276581,
 0.37397163577507153,0.34993973078138585,0.9522544338953467,0.574913893278817,
 0.06799979719722404,0.6060861598107651,0.9293620657950653,
 0.12450418227524951,0.879509227325323,0.5128246839974189,0.5004131623556399,
 0.927118048972204,0.8156264241556472,0.5012078407011977,0.4135016755399785,
 0.8307189439876997,0.9962667052742398,0.08908533485514869,
 0.16485955600277158,0.749060108293337,0.4641847381793839,0.3956544411689218,
 0.31303768168218826,0.45985237873518436,0.40743461812232185,
 0.33729213997888596,0.44792924372472864,0.1818727903236791,
 0.06890042750303949,0.583646872201322,0.5084216568988076,0.6343127149890017,
 0.9063403829015734,0.16828686303235485,0.615521594347287,0.3530033714027857,
 0.6866247865084569,0.5874389416559331,0.11207597750090414,
 0.24203852072774457,0.8563668686263448,0.5781045785157654,0.700728770958136,
 0.8455488190463863,0.7376625683893946,0.499246456450816,0.43580029630840444,
 0.2205235880772859,0.06036963227659853,0.14200188151280602,
 0.008086805266384056,0.07629558560246197,0.34707382464516723,
 0.5782828003348606,0.04527150203313601,0.7249396762078087,0.5120215410438811,
 0.4150163382952684,0.918875925296836,0.796975460839703,0.3859959332136771,
 0.40287794826602197,0.0457703297823393,0.4678651983696791,0.5727879166018983,
 0.6792966041986088,0.7470059469900994,0.052725345877697394,
 0.8912254839335821,0.46434224480727915,0.1799868133798902,0.7249295136386799,
 0.7164162119858075,0.7513132465258385,0.7177305656057779,0.12278704954740494,
 0.723526836327005,0.8234714617190386,0.13779529013156555,
 0.005450178282965901,0.4973197375682741,0.7967374743159951,
 0.2033940526833906,0.22182634217167563,0.20615284746005347,
 0.6951593481362194,0.1211033434038542,0.9085571225318906,0.472903035860584,
 0.5011344540504516,0.9419197120042088,0.4525301468231331,0.19293919788773994,
 0.7872451243513624,0.9603136630816992,0.504116174736475,0.5228906443759586,
 0.45474700932440193,0.39394275819035374,0.2108473146154255,
 0.9334205211429367,0.1992071436988907,0.7422564939635747,0.7624782040974581,
 0.8221011714956343,0.8187196156565557,0.8678616910846886,0.35071649032802765,
 0.24268657305734032,0.3399870531455238,0.6239158567945885,0.4792823436813809,
 0.8957190229817115,0.7765008458977043,0.26740167052697283,
 0.31856240605148656,0.09155745750514799,0.9557957437627661,
 0.5189365050647174,0.7716457821020862,0.6385748033032784,0.3211347739320647,
 0.02331995104686113,0.12540079973934726,0.894515673789518,0.6838004208024909,
 0.16324832570242265,0.3587834284319764,0.26246995571737086,
 0.9686424167220918,0.5797820084089188,0.3457693011735783,0.06629750960354952,
 0.9896047314077548,0.017532532923541067,0.33690351854504197,
 0.08191053632538592,0.3891968884040836,0.14896198966084873,
 0.46145651395441456,0.4528986317137018,0.4144808275521643,0.6726932771818248,
 0.3588209910029827,0.21851639748598706,0.9314534885245707,0.8333506278543951,
 0.5212695308376275,0.6267748942012615,0.8629357675551075,0.9686279233313773,
 0.4421788421807714,0.46534913922744103,0.7265017424191771,0.5347001064478246,
 0.7363306252017591,0.9622557078210622,0.5242619158452824,0.22664308856874127,
 0.11225266702195336,0.7103249348621451,0.5120594127937462,
 0.41190537406769434,0.5596000242320932,0.4188261082129058,0.4267830187410352,
 0.03648427937576926,0.6817145247903715,0.3495225468689619,0.8062937978180114,
 0.15436175864452473,0.4923866314012155,0.631796900694516,0.08472278976105996,
 0.7711092370141568,0.42195981486339185,0.0800526982386669,0.2799850176870047,
 0.4287028532793967,0.9707437473467424,0.8493295830118224,0.8735539609970211,
 0.5092455958824167,0.13554304557518737,0.6557349015911228,
 0.030013837195653648,0.4852200132412571,0.9618277911130185,
 0.6826612957645916,0.7012902339341021,0.3643510669337948,0.369917952946187,
 0.12472145209461627,0.3815079450340191,0.3523244999504551,0.6514112286987498,
 0.21954234952148366,0.6163317261442097,0.8826665023631639,
 0.07689455438275539,0.8791459960140587,0.03842131640001789,
 0.2114629063965855,0.004930964835861862,0.39928622717903395,
 0.7576207151205036,0.9757507489831255,0.02149980919772998,0.9206407002527889,
 0.9228087404113096,0.3900746224140833,0.3192343366497167,0.23228747872396438,
 0.18399430384935067,0.25275465340661885,0.04434416123986884,
 0.034493058905590845,0.5512744584088016,0.14687172662525938,
 0.5022857200045433,0.47789221331936793,0.9368512900589002,0.515207902273571,
 0.8609728533905513,0.42484837711950574,0.1589990157036002,0.5794672037234265,
 0.4407355748253614,0.02147257761003385,0.3062693134128093,0.4763624582958731,
 0.07199453225707841,0.6590331155175072,0.3984282192691543,
 0.17725517876562802,0.6768746055310754,0.6058203595315876,
 0.009066225750957058,0.19144155442344524,0.9666458714462169,
 0.19011754171183548,0.01937465844783315,0.13834175500807933,
 0.9952035578077827,0.17742131168087372,0.9998978427426489,0.7660936302238787,
 0.8672783714994824,0.6992534786429027,0.8575407510966757,0.16166423638179395,
 0.7116226320660052,0.16782255578064909,0.2929334600931498,0.1829476346062585,
 0.9899110471981574,0.8536221856434167,0.247919653311803,0.865717313156253,
 0.5366244386739213,0.5209511858602051,0.9630733871517291,0.8205263493469948,
 0.12346154139340215,0.08368012741991326,0.7074555546716338,
 0.48071215096348596,0.6459086838576376,0.9561562642953432,
 0.36434593852627783,0.4797795635477158,0.5094932972120619,0.5539979731254907,
 0.15450566429358648,0.651258430171237,0.9155162835387958,0.2618926892515713,
 0.24665844140483495,0.9116171546344473,0.2014073709973967,0.9001726723938815,
 0.031400773167037266,0.179139809318843,0.8063124878683001,0.8908466021439694,
 0.15206607772790315,0.5147321405711552,0.08797862247648158,
 0.3500492019003689,0.4084068782513739,0.7888682975685408,0.3866466749777492,
 0.8278763745833704,0.46863844793487996,0.7265679162752688,0.8441370493840357,
 0.4465296395850321,0.2710944086895949,0.691328768684595,0.6358458218218792,
 0.39165383941384313,0.33931343766585287,0.8760852881974444,
 0.39075196330764406,0.3304920525626561,0.159718770467425,0.18522652540972873,
 0.9763852648416383,0.7453186666955028,0.29316232844620527,0.8184240649068769,
 0.7152863430735086,0.383149662105307,0.766697173048294,0.6884409417382389,
 0.717430405958743,0.9781014852640476,0.6418283604887158,0.859852696090083,
 0.04056384344410624,0.8032723799390086,0.748614357163055,0.6069257916189544,
 0.8392997096976876,0.6051669684234012,0.39273118986892064,0.6056544431733001,
 0.07063442213469118,0.24141106269557788,0.9352052012116909,
 0.30571167849745073,0.02048080743509262,0.07934282004862603,
 0.6646612850718245,0.16344035171979587,0.3882420158723968,
 0.046803678644554036,0.4952141268010567,0.585922782943903,0.8683279005758969,
 0.7084418837890736,0.32171442781255366,0.8365614020861593,0.5387916457536549,
 0.15162621092098716,0.36235640246734235,0.30713621390449597,
 0.19101304263590768,0.057615233983048464,0.03869292457258078,
 0.27843320055513576,0.9473443608737184,0.6684053226464713,
 0.49393439750594914,0.3009128059261513,0.38516950896547686,
 0.5215267901971452,0.3284149156566447,0.5532037532363885,0.31479568719934425,
 0.5834892096713332,0.1413695562910715,0.24770422032781048,0.9229636781468776,
 0.4429983756094318,0.26290795781804377,0.5146107637730176,
 0.31686368722278957,0.4027910308131295,0.5951378847889648,0.3520183792283281,
 0.5476470344494045,0.46559900022272327,0.22858336717097005,
 0.2873224106445297,0.9715215148529477,0.3816813576678697,0.1445596158395096,
 0.7034729515481207,0.9120074966593205,0.29906024766503947,
 0.03344199602539688,0.2324980158455593,0.7267619369556212,0.7088313757879158,
 0.6932632823208035,0.5786662091376282,0.5930685476592229,0.8930901614039873,
 0.486606680810437,0.47996889039624746,0.281002723131207,0.9995516015791677,
 0.15744757606502358,0.6973545889850492,0.4414929365951553,
 0.25626173289807475,0.14312318480500574,0.6764570697286668,
 0.9023128712676061,0.4096033905712766,0.5087415386905987,0.8061295960409723,
 0.33567508365497034,0.07108674862116215,0.45485478937042256,
 0.13610196249129292,0.7282610244543315,0.24860150070345344,
 0.08985518227374578,0.5393493261447352,0.9504661916365418,0.7982692740739286,
 0.2504920384680941,0.8687418930550903,0.5414779468196665,0.612023880525519,
 0.6499462272080595,0.733728264339991,0.05269343031834528,0.4213036956637354,
 0.5076776052316717,0.7423224607019924,0.021516115978195027,
 0.5139516403947934,0.2202392926740947,0.8093249579515861,0.4270522187674963,
 0.12385214898255825,0.5392606333798621,0.557001766121367,0.2989283115383541,
 0.9576331460399156,0.46773049387603194,0.34075020384751586,
 0.6685951765575294,0.15110289622510187,0.9894186864982673,0.8741211402646712,
 0.11152978201093189,0.9017012509336821,0.5228904288602815,0.9567470307260096,
 0.772216779802002,0.8573502724914789,0.8767184767553239,0.1583141398498168,
 0.5117322280688992,0.06410210166920094,0.41152858527534375,
 0.9089363746014034,0.743608593250672,0.34397799493583725,0.6174680201389043,
 0.8711773500212353,0.9247414400905529,0.1946676331537085,0.24493957781229003,
 0.4923066800202216,0.988375455720853,0.004744969354358308,0.5328472483545235,
 0.0016668906419282647,0.6537710354836193,0.46955051117717517,
 0.30913679445344755,0.29042050245719353,0.3047102304082763,
 0.11577648688055808,0.7410280613879978,0.63349955357847,0.9614568211188047,
 0.456678996569994,0.9266389075249316,0.3581946081169112,0.35348624959968666,
 0.23019053066231931,0.9850224733671212,0.19838718711657077,
 0.8632074383607359,0.7863450792103543,0.9979702154749859,0.4102621049622037,
 0.7831792145779094,0.9359167720961631,0.765748259346207,0.6334979520430358,
 0.19945890878194694,0.7475996118520998,0.5037126866165493,0.8972395847372002,
 0.33225829949302876,0.7205237774811164,0.11911893841501253,
 0.8578076581578573,0.164979295247482,0.26743600554649705,0.27260795738744825,
 0.22002721535600878,0.2806308787288535,0.22833063275627685,0.750615942668102,
 0.7858229591452274,0.5897846832173391,0.3802399055691337,0.24672676827581264,
 0.04552903961096866,0.15759515968275495,0.7334065141957713,
 0.07887249786151618,0.6280307809159562,0.3536839857775127,0.2821084335025409,
 0.5827314865570935,0.033829751896507054,0.6687113574362458,
 0.7155564288235005,0.6789156028827085,0.6612781916380052,0.6802902508331948,
 0.6335790311247107,0.10444823457895192,0.19885823068283015,
 0.8394761611275261,0.9258486627068783,0.10355376378754344,
 0.40560110945828376,0.7713576174217724,0.21771656552380025,
 0.25845150104647374,0.16162357592060528,0.4595935225743225,
 0.9282340182745014,0.7699267923082624,0.3448916520590357,0.01735649939317485,
 0.6821964280625528,0.14680852282213996,0.37564140320724904,
 0.3079298051681969,0.7490773070711191,0.5743260648381778,0.4940110279919343,
 0.7993599947494906,0.9634775429073523,0.28318918416334105,0.6740213949514396,
 0.9349960979037999,0.6145234890388904,0.5241744854995527,0.08871383084089535,
 0.9311356896600539,0.5111039010882907,0.45922252357099347,0.9338976442006641,
 0.612307704092108,0.6165952018127503,0.4581957812966827,0.7876674338394702,
 0.43481279722502064,0.5231470005945927,0.5008439175746923,0.6573980387129778,
 0.6974742784977849,0.23284715863182803,0.9825123040921921,0.553067131986425,
 0.9452352293189382,0.988721499016744,0.5998894158517207,0.9564420766065209,
 0.01257388942014348,0.7181650578613367,0.07609283055541449,
 0.7407830278520214,0.18431381522892776,0.6158481471713,0.3984175911920975,
 0.609600235549459,0.989582766230009,0.9064437221855579,0.6095735559074889,
 0.6171147084893812,0.4961943012471429,0.014965409943120012,
 0.009235315457225224,0.3516276639528444,0.0030367336575992665,
 0.1899395690107224,0.9764840469049508,0.20421876566643504,
 0.24360934071536045,0.07109830192965705,0.46230270952249375,
 0.9878287645887741,0.2503330049352841,0.49863315008756337,0.8810979545321305,
 0.9670474629905842,0.9092152835874739,0.3814032026899292,0.23997792241222626,
 0.3740442616531463,0.8346931798125796,0.7714533326737699,0.6997343905202974,
 0.5186630593538331,0.8131616597017409,0.8431679331778794,0.07860483261386442,
 0.5429193701007277,0.6723611481645466,0.6356917393909021,0.6696073779587514,
 0.7267986347966615,0.023040901355434462,0.3334344345371554,
 0.20275150769240413,0.2858441169762962,0.3028536404647696,0.5792648023806957,
 0.6480305325757062,0.31814338898579475,0.4221075095886465,
 0.010205444121811458,0.15345674288270372,0.5072735027918032,
 0.2997909983177679,0.9401051257410608,0.1454390454303145,0.2794247448358711,
 0.39931600109446264,0.15866422514400913,0.7511824420832518,
 0.8408708844268092,0.5875787198685068,0.5183458175640216,0.709575554185486,
 0.12148296309313145,0.26120030744172595,0.7282929719782077,
 0.2156055835315055,0.05036752168746872,0.339377364897273,0.01760716410157559,
 0.22127946531586673,0.8989535837898099,0.6377039509278923,0.3019789167046547,
 0.8708096881251667,0.8052984604021697,0.7949399975775601,0.20168582371815447,
 0.4366917350354992,0.07114270413018309,0.4743736237078179,0.5168996772753468,
 0.9749166988073487,0.3216159080361872,0.394209395648746,0.5342838730172346,
 0.9006265993859321,0.7682769973098718,0.7368262437915896,0.16103700934641707,
 0.8522539304444019,0.6646099537679978,0.12464512516175774,0.6386268092644355,
 0.1941396855662788,0.9107871296907921,0.5689522863328556,0.26315643055547744,
 0.878169636529504,0.1092451564382656,0.33928244726288204,0.22713916846769222,
 0.5861341882935078,0.47434149874671394,0.7933512823704502,0.659892546006284,
 0.2323239886754418,0.9079390684851497,0.3956771639474098,0.26325484175626657,
 0.973097612292738,0.6989671004008288,0.0029019725970620414,
 0.5800494771267828,0.09730798680247088,0.9055232772858792,0.8535083555267956,
 0.8791564943102483,0.11600519433538292,0.32716209172551647,
 0.7528592093649462,0.5732666317077304,0.17953811486754612,
 0.38446456935207374,0.28695716698187,0.6426417283204543,0.1505275378818417,
 0.6799759241000288,0.20187691204709535,0.3619872484565648,0.4309903316981629,
 0.12113026457530118,0.0659438744239742,0.5120781252690665,0.5624675436599951,
 0.5337316647318661,0.8179206982015385,0.16528203473715408,
 0.20866185910583412,0.0127571408134326,0.14526780460299638,
 0.13467949534030677,0.5923555834287784,0.4040509281605458,0.8930389026033356,
 0.8529754666796567,0.8208895418243012,0.135063675245351,0.814424186815856,
 0.987315376195671,0.3961360002375347,0.6414520637558375,0.9113940301700721,
 0.6004060555484244,0.8342593965414957,0.7872331229564029,0.14088370856578758,
 0.48670274927803014,0.46037208493276793,0.1513497891271638,
 0.7809619689935746,0.8860020025773374,0.9571314635157088,0.3306575212927073,
 0.8196910645224272,0.45779867771001725,0.6726216996638533,0.8489005234914833,
 0.47117281551926493,0.3548568208337338,0.5163659966543019,0.837372773177999,
 0.16400407497984082,0.40753276726622467,0.8058804258414758,
 0.7336791103633501,0.939955861090645,0.8352221512042763,0.4853154186345878,
 0.2713304945731324,0.26473807420978335,0.38283553308078866,
 0.03745022200906434,0.7618779424396441,0.5657239142878945,
 0.06463460293176482,0.7516589691320351,0.7766654201313641,0.5949364403688413,
 0.4417639300069013,0.74793794628544,0.6455811308520647,0.37824739411757635,
 0.9986135135830754,0.4317038611402668,0.06580944461027116,0.6372497157721239,
 0.7108214828998247,0.1546417500197571,0.688529930881627,0.2338468085571862,
 0.13991854176774177,0.1957354163580427,0.7744483760665462,
 0.48061794201946784,0.46068715994195786,0.4033477809916246,
 0.41257604644934054,0.8811233733704428,0.4198402404229472,
 0.42902735673171755,0.019231348008998395,0.31724560133299606,
 0.19506345478842535,0.5714693786826639,0.8963340352629047,0.957457051134854,
 0.11197754682348582,0.909727222321442,0.0715490980813912,0.09164255397970211,
 0.3061905632165911,0.3375382801509854,0.4333390615968489,0.49112884648849653,
 0.4176883125792521,0.0922459233090328,0.6601861375224605,0.5957274000633066,
 0.13054748890543277,0.7460266941660642,0.7517971697522546,
 0.0024022128038250212,0.1685064356109376,0.06865340840636736,
 0.525649534084042,0.7177746149335772,0.08637240863569828,
 0.006543942506839429,0.8511330133541553,0.6234026444442046,
 0.9253429102540774,0.3119502519535112,0.06517664841668769,
 0.44952923216937024,0.7460499624206537,0.08511537751906584,
 0.36774934742438836,0.7933696297669497,0.48138274389050184,
 0.6122607038841342,0.435620619235525,0.7999622333116708,0.44850799615782755,
 0.014194153155664524,0.6210637815294597,0.6906673800640233,
 0.4282603877365594,0.2833437188882346,0.7506819279973956,0.3555480206564595,
 0.47834223525303865,0.3703412146295346,0.6008207059515577,
 0.03441866695440776,0.12110909934972436,0.3073715434132168,
 0.8008838669961724,0.6517027337910773,0.8049813877707799,0.585391091223566,
 0.14554883830470766,0.184323248522146,0.7091674263044072,0.7742501264658455,
 0.054578054774786944,0.6147003518584631,0.6303847620353354,
 0.1507877109786565,0.6471195667820406,0.8119542875126852,0.9281592144111381,
 0.5577343057596961,0.10042657007683853,0.1767180819544607,
 0.26887813815784545,0.9618813571024307,0.06918744142937516,
 0.7982149903093552,0.4446375852848521,0.12354121611973101,
 0.19211623868554328,0.06635489723344823,0.976988657053682,0.9681852277661565,
 0.8864341335556706,0.5052013780233728,0.814901076222612,0.8937028726271699,
 0.20691715892970464,0.7723440538027367,0.06786126093236389,
 0.3938086932458503,0.7355489782066962,0.650434493149317,0.18179329673207922,
 0.7795310475580354,0.3228377999669255,0.17335887318266885,0.8392512681459354,
 0.2361631326915118,0.5038898230600227,0.7453296631018079,0.38151433792426725,
 0.8263985871330054,0.48006019772649855,0.13924452617464933,
 0.7835409479963618,0.12222346406630513,0.4267948239617614,0.6573572224694881,
 0.1896074889867273,0.11768158581408006,0.4712513803977384,0.8716230792895985,
 0.11423133392653817,0.28153525207895413,0.12390641829140159,
 0.6518400193783693,0.9232764352076089,0.835102844383708,0.0952825702891621,
 0.5204128099995909,0.1619276659808171,0.19149627816492,0.23817835174683466,
 0.49401137886329216,0.4415078028222882,0.9270850970682492,0.9551300314961712,
 0.30711997437797733,0.14692969828038638,0.3853555078796036,0.563265800534978,
 0.9261764826087283,0.24426209191157233,0.2620395309896071,0.6655400635884742,
 0.36600398036589776,0.6170359377987722,0.22839789105585628,
 0.8140592581418203,0.08820958778989063,0.24944911547723525,
 0.48870096334952073,0.10306156070647654,0.8195129949145027,
 4.3335119201204364e-4,0.8522713904875084,0.5394907802372784,
 0.4737585742645094,0.44071304781269716,0.3442489253291279,
 0.27915415671968147,0.5894768898181558,0.991505305750812,0.4207724525901474,
 0.8312331860124945,0.6696313158095762,0.3381304721797851,0.8032126398512137,
 0.4056124880554114,0.08356996227486213,0.2753291331369221,0.5134294512857164,
 0.5253206503571696,0.7291255000427572,0.1548606403281162,0.3697483538822589,
 0.373174621860341,0.6455532610425447,0.05571020511164171,0.315487214303942,
 0.6400150014702758,0.062440779123222656,0.01636803033364498,
 0.2474860874748087,0.5522765737932761,0.5156668012903054,0.8590412642228517,
 0.9885213689360156,0.6796840940029405,0.39032862755253717,0.9003378609367874,
 0.7565368842304646,0.3160140124226075,0.9965531717480545,0.4242822124041805,
 0.7511698514516648,0.9405015894742884,0.8437879121370455,0.6672493493372839,
 0.5214815852452861,0.9004057942947853,0.2928689566784143,0.17943148829008881,
 0.37875163168897674,0.9398247322445782,0.6308778675779119,
 0.19938052923593497,0.04828459146647335,0.9410428853279845,
 0.7609013880308904,0.8858340681144606,0.9423807518211397,0.9395721504376527,
 0.42118886065134187,0.8820040190132095,0.41463418948030795,
 0.5831623181259394,0.9917212567309197,0.029763095559155794,
 0.45720619941458496,0.4603522479882751,0.8715191005479502,0.9481659922530251,
 0.5146110828938489,0.07522482340305181,0.84873099197435,0.0204746421998242,
 0.9526793201064608,0.5125660987001014,0.9412669798564587,0.8770308628798699,
 0.5591164695577563,0.9794450418085849,0.7747348726696635,0.6738994050215839,
 0.41556536373660113,0.6613317011828144,0.737638142911639,0.6227513601556518,
 0.9762739809319074,0.16453311225709288,0.9623818848196003,
 0.09783530355324443,0.01212239305520546,0.49789341677825016,
 0.7677005932224619,0.873176944582877,0.9056074440766478,0.5970275081188736,
 0.9446232172186149,0.43828027408994397,0.5303811959188036,0.2415171799088207,
 0.402686876767944,0.6570232989955485,0.9996303932262605,0.9714938105656767,
 0.07147268031558296,0.14336342916451794,0.8723624475499505,
 0.24663288004705253,0.37821981875720323,0.4053724744485294,
 0.24246704738284697,0.5017760550852266,0.5398194622779353,
 0.09951813108980223,0.019438559183761916,0.38300197377404754,
 0.05059452477525439,0.5119614448615051,0.16747810030547094,
 0.7147302373283397,0.02049175386482971,0.08398129090374318,
 0.06502266272637303,0.45868143573398357,0.6201536392325917,
 0.025149629239317628,0.7228806957547586,0.33002797541965057,
 0.3063025795310984,0.12981633291800387,0.7961556841676634,0.9158703537058155,
 0.5087685077106229,0.07660651605303137,0.8054715438799508,0.7359809856435916,
 0.8590960357892174,0.3960771821507231,0.6723979777526681,0.04389834516745739,
 0.37508883905301316,0.5989037189165608,0.9079772365284507,
 0.23454069670434086,0.7885484269700298,0.8144931292808449,0.548811391281683,
 0.8043867244339301,0.5994226867272417,0.9062421842338065,0.5122752418663447,
 0.5161752424285111,0.06297467214182328,0.7274600036994749,
 0.17599838115130628,0.6733203163731435,0.049687802391110925,
 0.11525276204106305,0.7709625851295969,0.28911976507674864,
 0.07557156591970904,0.6409098333065386,0.2559177701745212,
 0.28387535423449684,0.5964166662488256,0.563730694124515,0.05457480107018886,
 0.17203973402924766,0.7351961610666617,0.6386021704540178,0.8643470084256163,
 0.6277834656714534,0.34184124337613464,0.12484068167961604,
 0.3063677513093789,0.9580577647849267,0.6803209569837453,0.0518322072626064,
 0.4182272741328251,0.7501731338581928,0.9279015257267833,
 0.011673864237174203,0.20760003403189264,0.30781475352519416,
 0.13549980630429703,0.7082295677103083,0.7924334289718699,
 0.20118763077745427,0.7470360269347494,0.8288655170721791,
 0.41510598675800026,0.3711305725809664,0.011604331700509563,
 0.8770864195581506,0.5767631238519098,0.8062995013925923,0.3410854145504155,
 0.8959141805263907,0.9535336873331255,0.04403928698814652,0.812408680272096,
 0.027170845718067382,0.5267503684375693,0.9165880905080314,
 0.20864656425213823,0.29677369575256773,0.8473340613203779,
 0.3906011515727701,0.19340707692107328,0.8763038495219061,0.4880427646549782,
 0.12773902724357988,0.40894944340143136,0.0409025575559433,
 0.5639487702158521,0.9071693549209715,0.5546154831140239,0.952435934635797,
 0.547794994274685,0.7261458620908736,0.37858361376844896,0.7492515558336086,
 0.9585037145363545,0.6413166281433205,0.8072819178781203,0.16849046856649186,
 0.7655769003125524,0.5702501822575274,0.3276382403826311,0.5354861232199213,
 0.44573170878332036,0.8644454676599617,0.45883177841804634,
 0.33710619183666735,0.05933803995028408,0.28251157861741005,
 0.5962725509038708,0.8427911828212492,0.4307868851921872,0.6706499342657126,
 0.615369814974316,0.12264790264608161,0.8628310505080027,0.334579614454968,
 0.03509844166177911,0.9790858122220102,0.8934968260977169,
 0.28153070170064154,0.9206234431033592,0.9568912492528401,
 0.021577881905932572,0.37125198046353425,0.022046908228309237,
 0.8815322007183405,0.8239012096634616,0.31695811030027665,0.6570897268136427,
 0.9065351615158246,0.8594388789625985,0.5360002544398026,0.8000791851292269,
 0.22751628394165024,0.5159502102795646,0.44559894879212625,
 0.8250507711627086,0.0621961295187663,0.5556063472605568,0.5640987462766838,
 0.8301408931728118,0.784222396374685,0.3217740801675397,0.07516013264450438,
 0.5850851454544488,0.5655252450899195,0.7371633705650048,0.3661437390758625,
 0.18599407390975842,0.26347595191373774,0.7174232987287409,
 0.11363218361632033,0.17128877943641163,0.2681001124240082,
 0.03661382990975903,0.49546937847985717,0.4254649445125481,
 0.6179468380813818,0.07929589805159243,0.862155457380306,0.8676500032915748,
 0.24030343770499119,0.24129412294200925,0.38879151381689725,
 0.7489137621440624,0.46199290947783267,0.6497929594472935,0.7117657746112784,
 0.48345785862558965,0.7403801063420274,0.7949016315970284,0.6170673880964004,
 0.9823846758559549,0.7790065506115581,0.1064524928764069,0.9289359421519987,
 0.7862104829629704,0.571054660440548,0.4649540970013697,0.7621195327165348,
 0.3878704812122459,0.8995055522713848,0.2857286061911528,0.6673185883126659,
 0.6226112222683933,0.969857187177971,0.9819985426584307,0.12382220736692773,
 0.43471146853821385,0.03916411492302185,0.6637536091063125,
 0.05607429157529009,0.5842376242994477,0.2693798655733538,
 0.26033639590613133,0.7158035996431513,0.958530285142943,0.24253934402422495,
 0.17931741833009418,0.6929782926156504,0.5203566337395343,
 0.48757854277827684,0.14131798313737842,0.3116055075168417,
 0.7796672733528812,0.37785815354802377,0.09503205926712488,
 0.7440780034008564,0.5136516269842158,0.7262144145341524,
 0.046431198093069725,0.5001290236051492,0.9326860860417976,
 0.18770001707911832,0.12016975493770199,0.5649877743336232,
 0.10477089571919262,0.9802469081997707,0.393354919367673,0.47903765711688606,
 0.9516041040455904,0.33403676302223784,0.3199113095878032,
 0.23043700880675566,0.1313601198327491,0.2519581309243688,0.2632595981688821,
 0.7450473591594695,0.7214153485181979,0.9003439735989887,0.32815580258211585,
 0.7876427354033492,0.04958264333627738,0.6426768907616087,0.405188157069012,
 0.09178766309521957,0.7468627878913523,0.9602437514896219,0.8141297235025632,
 0.8623577887922272,0.0877129876475754,0.369056762351455,0.7069526200513232,
 0.4901147693068334,0.03069415882949278,0.12446421276203656,
 0.031958144906547914,0.8616117375913221,0.3696283743076001,
 0.7523544427475836,0.030138743506392562,0.18645488663311716,
 0.1811997820804222,0.3145899300797661,0.9627376356371287,0.8927143841384684,
 0.40421099901615576,0.8696059856453203,0.5552322627028768,0.1484133534947918,
 0.36886347322069135,0.9302548426114163,0.6429008968301594,0.4452428502846014,
 0.44539858295199697,0.06524538610285491,0.476468425320449,0.270603183335829,
 0.09320933863675263,0.8871311428277538,0.7069581241857469,0.7550521099730336,
 0.4342981240834545,0.29494865495622646,0.08040571943792629,
 0.8862676456844301,0.2727585474829898,0.5330721310540124,0.2055868748178149,
 0.995596408686046,0.5539669199657709,0.21615767847605383,0.37835428354989453,
 0.18901157255039913,0.05509904375040031,0.048662981343737255,
 0.36528992261998805,0.6011763886137924,0.4293602986859041,
 0.47046623932815446,0.7082201427328115,0.7066852081881991,0.5577816980217416,
 0.45752944576929166,0.4951381225458604,0.5912587745911636,0.459263337400172,
 0.24946209625014426,0.5914767161200184,0.8703214789164682,
 0.39425349888825345,0.8400178840131225,0.07181692535670425,
 0.9419847065581579,0.8036493579879251,0.1588216844788244,0.9685617267915747,
 0.5103874835234272,0.630984778148376,0.5614275143097462,0.30231157440619405,
 0.6778487771665652,0.18269873098360678,0.786720102792019,0.8506785627680293,
 0.18873350950205414,0.9541955995407218,0.6200222579922005,0.2286214716564039,
 0.1991587830311229,0.5005705514964665,0.7371420583956748,0.17777032083623523,
 0.3556352587166929,0.8060738891319644,0.9801498017340684,0.9580061657258906,
 0.29819288168584335,0.013689222285446001,0.6368614969513265,
 0.13094549347560747,0.844869773347343,0.16311435222208548,0.7681692267366792,
 0.6982127572102388,0.5698316121634289,0.5941441536689382,0.14838691625622236,
 0.8946150458284697,0.13465990609864142,0.05170021873160746,
 0.2260796106454408,0.3302813586167974,0.5935715036765469,0.41140552861598556,
 0.8252771539543506,0.07727789365530169,0.8624616210372675,0.164453415748405,
 0.2587654080918178,0.10849407666079647,0.918155552805839,0.09174846406429193,
 0.566627756226808,0.9457774255292775,0.5123450509063968,0.2812734183179407,
 0.5521507551928674,0.595214431523908,0.2583708242660804,0.422436243636366,
 0.7108954770448407,0.995583727922961,0.14248094521324728,0.6118806877644866,
 0.49648971496532956,0.1454476538493994,0.3311186273688662,0.4575028509280269,
 0.8272738956423928,0.2672259534290631,0.16037610871715602,
 0.49604828565071735,0.41469324649768635,0.15588696856489792,
 0.9830559297247109,0.9342655285865944,0.73390282906495,0.1559032979274343,
 0.26403144029147674,0.5006114785366582,0.9636984574130525,
 0.09373266945724623,0.9710769357509916,0.3625076846558668,
 0.016048945229300005,0.9134532172551223,0.15204882801496367,
 0.2608755855562145,0.9789567807530463,0.396232687400639,0.4653150561385174,
 0.8291821321425688,0.24623350486331086,0.1419674090393892,0.1664944207650485,
 0.9870357275129501,0.08258034775395662,0.7782330859646858,0.6827767670922559,
 0.8245514654582808,0.7474745215215701,0.2489667059684051,0.1042386759812628,
 0.5522604014669306,0.49155958265824895,0.5964832981838131,0.9458451164496664,
 0.07668367550457267,0.06478168401154094,0.06428342138074239,
 0.9824789948839961,0.9862358009207377,0.28552519063351367,0.9586828154153895,
 0.12548221161239526,0.14013337232083323,0.6715331167103941,
 0.7822439593983553,0.951336152462186,0.04879241975453663,0.69209296700883,
 0.9375358891776886,0.19402002306925625,0.2949396164249227,0.5541741869583758,
 0.8304369540676393,0.8480583117496662,0.6135836892224821,0.6439888518624031,
 0.06294281606640606,0.5047232987799724,0.18838985935831964,
 0.8951643386257417,0.44897091546381807,0.1482551061414381,0.4808489549245226,
 0.8969367554318294,0.7052875380905715,0.7017003273325678,0.9698647017159441,
 0.8128872084550205,0.8441612155237428,0.2952710371488272,0.02492449653494999,
 0.6520353390550239,0.5678860705315117,0.0014094546487284365,
 0.29681160291181863,0.3562635095841076,0.7456051145407896,0.6642158672512701,
 0.7608967947789984,0.4232755559548309,0.9611420039206542,0.8932161777383995,
 0.08203789030246911,0.6331470522582145,0.7201634943065706,0.7513274831670624,
 0.986670810650277,0.2252132890855787,0.26934322081105555,0.6444769518590326,
 0.2252121682804512,0.8860795892005119,0.3322597750135634,0.07454279156262444,
 0.6585072147287514,0.1556567380676055,0.6355566465107797,0.17946457787906978,
 0.7566271500718542,0.2093876447594828,0.7373246079145301,0.5466360625595744,
 0.6571372080636458,0.07017984737701977,0.084320139471182,0.31512872301153594,
 0.11108740007846807,0.635097241238507,0.38343834849830827,
 0.45964902253642403,0.7348318302340738,0.11890165516639162,
 0.11058981572812421,0.40969258353332005,0.11668783626502977,
 0.3884195607594504,0.9668127094905523,0.27179524348116724,0.7775573378302877,
 0.6556629449062823,0.2162450697558327,0.19631602352009225,0.7823930812523272,
 0.7246145268575763,0.526450317917932,0.31972585802865106,0.5287991306949877,
 0.7929264191251724,0.003709191292113223,0.012426065616787274,
 0.9842251727338593,0.7380672297899564,0.35904877089805476,
 0.03323383930619994,0.6594231203611697,0.40765952594392896,
 0.6800114214196948,0.14840278235414905,0.7656598951148549,0.682390178788305,
 0.5906591324139298,0.16513866774707886,0.5467350183637502,0.7742016679339392,
 0.9623602351226874,0.40225842253973987,0.4088055779842179,
 0.12024409964712379,0.7876419452237642,0.5804047940354891,0.9149716120046584,
 0.4319642723437074,0.00720777641699244,0.5607967730411464,
 0.27802277634916994,0.15762299239541278,0.7025818298630224,
 0.8890618380468305,0.25248306334156245,0.5518828606916556,0.1551253181746014,
 0.4876540245851938,0.03670229848788542,0.1844944864753697,0.8694100448655827,
 0.7428511980894266,0.4415387372660311,0.6747371846957064,0.7065895229058801,
 0.04657969008578111,0.2643780665820229,0.3974004783928715,0.8113725972855025,
 0.36534377330402523,0.2663884669259958,0.7866493034768114,0.537437810035497,
 0.9699430637251424,0.9382035654385749,0.12539094939498063,0.7696979889619773,
 0.5524263205748263,0.7358291257268263,0.2738910175578104,0.49135874111677325,
 0.5427227268777002,0.2003581735401606,0.5383430562594806,0.2248078555328239,
 0.16891604807822458,0.893367077140518,0.38367418126319675,0.6179605589971874,
 0.8578930068158978,0.1205082525017559,0.23720681692034984,0.8133865844651366,
 0.041172211545565274,0.7144651374199863,0.7811196767432693,
 0.8125372754952641,0.5461531919283199,0.8681059083341425,0.8326477034410671,
 0.36524741859738763,0.5088900423702032,0.8381065417203191,
 0.14938869957524548,0.630648867350524,0.31339273858434036,0.6884354252258045,
 0.2715360045656591,0.7634930580944388,0.588034599082772,0.8458025491391478,
 0.5939083405379085,0.9717677354620022,0.6376990834301398,0.5035635944411199,
 0.0877800512228506,0.7582619498392033,0.9520278244392869,
 0.022435039783319732,0.8764993338720034,0.8045634351248999,
 0.4481712606659145,0.7098433284710177,0.9361472076744309,0.2648252375822022,
 0.48472456995866975,0.6132648019290241,0.12514042820966753,
 0.01688664872407608,0.886214885915916,0.7158189702309131,0.09763295252154391,
 0.34491464044005693,0.9515429262868104,0.14037710996951058,
 0.43625873523654546,0.05651312628828331,0.08274123302770187,
 0.34410612492858184,0.7324742018275778,0.11547801164974669,
 0.25860089615965587,0.4631755442665495,0.025727295993170918,
 0.4762554924332968,0.5813885133524467,0.1405735192223707,0.07756350735049866,
 0.35585982883101364,0.2731845565886629,0.6349161540223777,
 0.11444207813966967,0.27103293467053124,0.011172123572164572,
 0.5617951022246409,0.5318728752473896,0.27565872926741175,0.6419047076683551,
 0.36358478410367834,0.02850831455808578,0.29477090542440143,
 0.4676448602915757,0.5271405367943126,0.17019580445843918,0.3355336153198313,
 0.8887918189888464,0.7347577929481671,0.4864776108928667,0.2225559228492574,
 0.6994500614967843,0.2663336653098779,0.1297101157940218,0.09745721559900822,
 0.5003061088074398,0.309951111579702,0.9618534290653094,0.7085145791256808,
 0.18319520704033088,0.03162336523471887,0.4958684228156486,
 0.7341561028672673,0.39138297849424486,0.9878188446251607,
 0.08876724794835755,0.4676963045748628,0.4245715571365819,0.3189006475367754,
 0.22784157330788546,0.7109553037625513,0.26187892339929164,
 0.9905355765378142,0.27841733357489873,0.11761191582247277,
 0.26540333228935076,0.23626917119305169,0.8855572212468824,
 0.6152216777352412,0.17096071130273238,0.7144341155631813,
 0.042700769912548564,0.5507619902134024,0.8179048992408151,
 0.7486608964219574,0.6047371280070147,0.7084526724943018,0.6715242179940938,
 0.024181537800346042,0.9968716683369186,0.23350304329418115,
 0.642076928079077,0.19567438244780178,0.18808337501653483,0.4373582239709358,
 0.1939431939074956,0.20366113655477402,0.7633590193911328,0.524454440898804,
 0.38580011409862114,0.4141863650884724,0.27741305606218347,
 0.05941224332386352,0.5334544267637603,0.30101587109338757,
 0.6203638143128829,0.2501311740840496,0.18006339347967182,0.8150081491515245,
 0.15013727542143063,0.9989390568639047,0.4181144520942406,
 0.41136693538195934,0.6599278896517117,0.5119072700457856,0.4098383827240605,
 0.7232926595374293,0.19928962438831543,0.05386627907946062,
 0.3578872061212215,0.6423532101513135,0.6943649613448235,0.9901644073716676,
 0.14787691813113268,0.9591021710886536,0.9322500788951789,0.4188652769688588,
 0.19201761094104974,0.30608671671097,0.21479257769973403,0.9326010041740886,
 0.01011962892023699,0.6140618791532437,0.5739573694761263,0.5909912866153701,
 0.955839715637071,0.01926798939959562,0.46075554548087794,0.7622997500069582,
 0.42604723850368487,0.7855859369490741,0.9413046442832292,
 0.40819972187572773,0.7513799555905104,0.7820672015581117,0.6255256741625052,
 0.4540980856348448,0.3782843977041591,0.8849463533222335,0.4771178053587348,
 0.9615997428056633,0.03131474307746762,0.6254669803360253,0.4147966176969131,
 0.4140799868058451,0.2713565024112893,0.9560427719605896,0.569062603291179,
 0.02556007816742467,0.399566028277774,0.983800054283193,0.2679789497792231,
 0.11547041143560766,0.39339578768012773,0.1777814358621228,
 0.19143978021687014,0.8228018027887076,0.5773306111132742,0.930822992296617,
 0.036484619393315376,0.44457515677501,0.876996098770596,0.6123861487266564,
 0.8252464741610849,0.333665240222107,0.046488482724417524,
 0.22555917616380616,0.6377917855032985,0.5633878244149926,0.7728876834596866,
 0.661365553179362,0.77111514924192,0.686954670902259,0.7390103967639803,
 0.4159414882264252,0.9744362606322249,0.45687440830147097,0.6661900584415283,
 0.7262240653660724,0.4640062687394223,0.9569748948367188,0.14971884066767205,
 0.07261342937508475,0.8802564424626756,0.26018005441383163,
 0.19928487339363588,0.8651810107581578,0.5783287147601263,0.7676634587844967,
 0.17827112066988782,0.3363495379568684,0.16781350854821353,
 0.18532792964522404,0.7804354788855942,0.5545451366570633,
 0.19345398220971421,0.7419963039830987,0.5173987418868684,
 0.49604621141298344,0.34224620061011757,0.22366619852184222,
 0.8676173742595098,0.2560266028510012,0.8994782297172754,0.451823867374475,
 0.6630502510423673,0.6642312466114068,0.10623373833035898,0.9512273484519733,
 0.6096024652742019,0.1114137315960606,0.5142730251916255,0.7427813489080536,
 0.3131136429773995,0.11057773702219564,0.4965400352057924,
 0.028118344651594906,0.4123446624818097,0.21579291892496966,
 0.7743981502622941,0.07648407007600333,0.15182534273487047,
 0.5956852551833864,0.5087939102006662,0.36966367766510744,0.3854076831840837,
 0.9268930682628219,0.6076912129989047,0.5718245424181845,0.4052210789383024,
 0.7344502627290193,0.10312344360845938,0.2550126412965442,0.1659012964527019,
 0.8422632815516239,0.5997251015315151,0.6882373579678688,0.9064261430360901,
 0.39742380939570066,0.718624538058985,0.7935603043932689,0.683890969906626,
 0.2542252371179443,0.6210472123880001,0.800428861566251,0.7587070078011438,
 0.8696277508334342,0.9941342388567687,0.6320981916469742,
 0.009055976003889299,0.9583751599583823,0.4921046016425279,
 0.8319633155318528,0.8883769613517319,0.2261376223910342,0.7188323206740566,
 0.34333239632572876,0.8364567514406724,0.748285408964152,0.20578451627401384,
 0.3381384279851486,0.8226772655272068,0.514191941662997,0.45292883414154683,
 0.9874294595920976,0.8492088219344834,0.40951988934931927,0.570144715010622,
 0.6829412714331533,0.5390364924289255,0.9310491740620981,0.9459548767780013,
 0.8039930573545977,0.3218289537495408,0.9647572003771662,0.9036979790160362,
 0.15149915513569556,0.5613385379954049,0.5353735733286387,0.9045403482206987,
 0.21086841832843217,0.5285948490207641,0.4098197639551139,0.9220142519066227,
 0.6633574207361717,0.795402239329619,0.39760087984711157,0.7178473355189818,
 0.6770832892757842,0.1561950759665418,0.9023473175408961,0.4841220585745254,
 0.11572830098848264,0.4011310684850854,0.9516650329642808,0.8110273355594838,
 0.731878830300119,0.4510153553062739,0.7735531745121011,0.27029941921451484,
 0.2609041554029897,0.6927139678168464,0.8345253118673635,0.49211992054117093,
 0.8472266265372168,0.5379071224100205,0.2670980133005041,0.22583839934047956,
 0.07068462897775585,0.7872601659692886,0.7406498052514895,0.8703661999758613,
 0.7966397149330269,0.6356254502183322,0.20049212427679342,
 0.45551314844092083,0.4627587391864132,0.8275402474177,0.7413244903747339,
 0.35690001706814833,0.9795328605240947,0.8288805126880451,
 0.47127401815746905,0.7328755268947188,0.003615104247512546,
 0.42711336053549676,0.951264645059243,0.7210996675932404,0.09529880100747967,
 0.10544496676533399,0.34177212433709925,0.4616672413581142,
 0.04215719215067493,0.47360036403434436,0.38959344911984417,
 0.619943098305666,0.6594543027482261,0.148109811449599,0.8300128418636158,
 0.6578750890872919,0.7845271158377032,0.0048797489277919714,
 0.19643855261636034,0.18632958729593374,0.6297069223142536,
 0.4441164831238342,0.6141363836793733,0.8837290043402646,0.30847497935452706,
 0.3153872581978078,0.1634784148566647,0.6964163044329841,0.9591953152379122,
 0.5121934934916738,0.5729975996123388,0.45897970210648353,0.8650420183396019,
 0.4396892887517234,0.43326386744042467,0.9469923000766551,
 0.16717785001475793,0.3241953908319999,0.6189067405888705,0.8901122759059447,
 0.18928866585092374,0.6703174782159924,0.7182757960367939,0.8486322513204386,
 0.8489417725638715,0.09188643904268079,0.7895600925510533,
 0.12450903002706193,0.6832646938381551,0.38535728711813266,
 0.12701223599082356,0.3295442380769864,0.15806846413367337,
 0.6898723634725151,0.29823801088425084,0.11927110001104557,
 0.09911034999433599,0.19985918100213296,0.8484914182599776,
 0.44382748285278595,0.38483831806018176,0.4297401301982763,
 0.0890271637762382,0.5068001052652842,0.3001218425862854,0.13268565186053172,
 0.5132208042024364,0.7712330838587759,0.6996265031815675,0.9980352288196186,
 0.9292119101249731,0.8450322757911685,0.7682375061157366,0.05848611866972697,
 0.8241131043205355,0.5223201386070604,0.7451962936150895,0.2451056147318592,
 0.5360653798218329,0.7003800676011852,0.21972669392847222,0.6103670739931626,
 0.1642385052920462,0.9485645054966633,0.7168426818185336,0.8113073495293521,
 0.6532814948529029,0.39259706165340325,0.8611836459055238,
 0.32357059938643895,0.5468193052030056,0.29149000573084216,0.511347521362126,
 0.5862546105126157,0.6571904008185209,0.8383921221220157,0.8388594168145034,
 0.5886670847668625,0.33375452063979194,0.20891728838055823,
 0.9417530999657799,0.9981891879656388,0.5533820845526799,0.6527088267970522,
 0.10551608916307298,0.4190166162261235,0.2666937580590514,
 0.11560776993035238,0.35158524402264235,0.4634159352843925,
 0.9057467674725305,0.005101565009684728,0.773114375059894,0.4371584039475408,
 0.7222407972964602,0.11838628491125591,0.9626744439221067,
 0.40534778620391354,0.7692644419345658,0.8951836043280205,0.6719704798536335,
 0.39368809736731425,0.3377079593961908,0.5502498500601167,0.510970075508584,
 0.4935694018285557,0.21833540667516604,0.7277681981593954,0.8283949690376154,
 0.861824211693282,0.1433731075030915,0.6932128708255031,0.3209225854531018,
 0.9372704624037403,0.68771595553921,0.2370525457024184,0.7282754138280209,
 0.23693954327690148,0.8146296269422203,0.2173945085461918,0.2649511345396811,
 0.38323804917051774,0.764075937882645,0.2653296862030906,0.7344271356267722,
 0.8861207794555661,0.015529994061263208,0.9644914111888332,
 0.3212214069674447,0.1541434277999223,0.11084650880328406,0.874047081668876,
 0.9487691299475575,0.8912825575754915,0.260896780607619,0.2683460776799602,
 0.7694247132185863,0.5161870507693684,0.10660305367522771,0.205197142970582,
 0.8248928172255408,0.6867446599710513,0.14749669164022483,
 0.23843978123082366,0.6647688083542542,0.10622232147629251,
 0.5631481858722891,0.7316835625529767,0.6458405215773014,0.2740144188509319,
 0.2579465711950957,0.7572004343155226,0.6084343224429812,0.27582580299573134,
 0.5320094508391142,0.17048860587825632,0.22209145062367408,
 0.16641286444789405,0.9372147934500499,0.9555371128219547,
 0.05530638876142091,0.5608510035657133,0.9425505085793662,
 0.23365777311288038,0.9799557145386685,0.3529628540288442,
 0.44574450068862403,0.46210409703288613,0.44670964645596967,
 0.7468810218872933,0.7529343230561472,0.2861665825079831,0.6296389540993946,
 0.6154561821020751,0.39999265930216,0.21850972483234254,0.8822315096097387,
 0.2442837016800119,0.4804636850263466,0.47238018541532345,0.5880910220399915,
 0.15309629979498385,0.3669402086132879,0.1684198280934769,
 0.12885506817768422,0.6835680621935273,0.48639919233676976,
 0.43250387653728506,0.6751013776754238,0.45221371446109393,
 0.5372901922056852,0.5246659253776821,0.12529838909406554,0.3662382141147005,
 0.733571687211617,0.04434374126882157,0.4652537673850736,0.3046450036344157,
 0.2855818253868544,0.4183038258366013,0.9875672757813927,0.09685776825953929,
 0.9402986250178786,0.7057443759248886,0.4352962619079169,0.8725886672931478,
 0.7333613510007444,0.31328106985460524,0.9808527059222947,0.8390495896553813,
 0.9212610428583812,0.5838921794912506,0.1235986558697213,0.5232303490054889,
 0.2668492694915038,0.7327005688539269,0.44877772351087264,0.7148294474784102,
 0.16999790535564419,0.939228400461021,0.45743191308629305,
 0.05413672869979769,0.9554715121984453,0.21473159420796006,
 0.8789937512721551,0.2787419192521928,0.8327270976139339,0.9010332725429259,
 0.28058605493702593,0.43149566039407117,0.4509017942575946,
 0.7078399670130462,0.29946871836329625,0.30494887388226244,
 0.08737587423597593,0.8156279456102523,0.13960570075018974,0.71547501783871,
 0.7113499798359433,0.719409672212618,0.30686830370033813,0.14344829778163193,
 0.2411738083419308,0.02116663285215792,0.08270744996733415,
 0.12480546069677567,0.018118760212071705,0.1475247637367859,
 0.8497904008239986,0.07212007623051231,0.02003513449918426,
 0.6034896509993722,0.808213712063476,0.7596393917151835,0.09356508956431564,
 0.3990677556535177,0.40938752410297363,0.009983457597432022,
 0.6448558414453174,0.1421213832610735,0.2687305937895824,0.7581907069308423,
 0.5874855089517055,0.400007547420268,0.9686602407101936,0.9352567229111123,
 0.2617487216045623,0.2141046749787282,0.08934837153696373,
 0.24649286283161465,0.6602463027308255,0.22500784744557845,
 0.8914235133625801,0.7764341976772243,0.7863489386679933,0.3733268277163494,
 0.29872775067106616,0.6618849634181743,0.3405449332651498,0.5183445598081132,
 0.6038975411763092,0.05860591251809988,0.9382724859240222,0.2557751422501796,
 0.8949319971384518,0.7579483156705572,0.9966469282960109,0.10338472285806011,
 0.7220283882935483,0.6608491921193791,0.4518133928284179,0.19238834837884689,
 0.8125234239477517,0.12556674734064865,0.12079293325768814,
 0.5932048444442097,0.8430245512762534,0.8235257346370788,0.6770537432424675,
 0.6173449993311186,0.9471380220711204,0.8626353437915806,0.5080246044578702,
 0.060568180478261846,0.8346780600036278,0.29294742963979814,
 0.5578976844589861,0.26927726640881045,0.4937095428016268,
 0.35498097755336766,0.8010846760816074,0.2569420820132088,0.6585000666802285,
 0.2913111928007057,0.81932668017042,0.8146879561638617,0.5640447674384219,
 0.34122111017018075,0.5789721812865717,0.48197233208808155,0.775440887026503,
 0.8652826116758935,0.6181559281493489,0.32715552321873986,0.7229455826274751,
 0.23590838834038075,0.09425499824865136,0.7024085986081565,
 0.6615159820650354,0.47259511933610654,0.8776819807400681,0.413252626426337,
 0.16191676193790516,0.46635415195206664,0.09175382768712248,
 0.10304632744815723,0.898018058109134,0.10752632632663583,0.9556005270029114,
 0.7987878222550252,0.3877592430472979,0.4046735246931181,0.2541263963427862,
 0.30326795340341894,0.2954796807578268,0.34275007041609573,
 0.6343905452420284,0.8255380945167807,0.8046651987542124,0.8186677852604762,
 0.1665324756304104,0.1644336301885575,0.1050474857124728,0.16949584723446365,
 0.6667529772485449,0.10276272708075396,0.8779766838526453,0.8360890622158871,
 0.3256095242275624,0.6204869993847382,0.039926944198499625,
 0.42274904648010336,0.42524708182083415,0.878329156738217,0.2756478233881605,
 0.1306516254945358,0.041463681321232126,0.8506483569506833,
 0.09834111366361287,0.8265473885747865,0.8658108033140938,0.5379500599513506,
 0.4316400076702491,0.1795024710443479,0.9686875567453757,0.636579875599461,
 0.14207011060965602,0.49627307926760045,0.14750718608224578,
 0.32713806455072925,0.5525185197771851,0.6463705288325031,0.6761199167983254,
 0.35018371356977673,0.9592020670216505,0.8109535032128786,0.8027197160366405,
 0.7410446532623103,0.45846516803237614,0.35587794639244463,
 0.8273163364058014,0.4549118218291932,0.18999376412182967,0.7167403963648621,
 0.6471318500006413,0.517871253749078,0.08645257150829655,0.20443108188868275,
 0.8546961842797479,0.005480896963815929,0.4107225499740941,
 0.8656920851066623,0.3799998109784255,0.10376456378632382,0.8094909583154212,
 0.16216514919362268,0.3222410104068276,0.7599555150258013,0.3375865253744632,
 0.7808449447614747,0.41501685194990623,0.2967426266417573,0.8080121922438873,
 0.7298216797240402,0.3802799227304051,0.9777792802536933,0.03319412641417585,
 0.9078944396483031,0.06235601065283203,0.8220172809732801,0.7538527988206061,
 0.996199790687034,0.4828411613942942,0.44352341779750903,0.8720989582499842,
 0.09074163265065194,0.43133436266946246,0.8431161668446348,
 0.8356826737179727,0.7170323456695729,0.761057823682902,0.3123664957893323,
 0.7285118354428207,0.8002349781084561,0.6241335794812538,0.6030945504877504,
 0.8911254113885794,0.22191087502432616,0.45773124605902904,
 0.9735935717373252,0.3418286528954273,0.9997631003429642,0.3351073223145349,
 0.856228841031144,0.20483578873314345,0.6419956925428157,0.4577110753349479,
 0.4032653230034289,0.37487280808957557,0.04927870746174645,
 0.8974774513821437,0.48193202540303715,0.09951821894711521,
 0.8812999812671181,0.8262073582684807,0.11261961538282816,0.7288446970651019,
 0.9127028596432362,0.1425171003394229,0.6382490620218879,0.7208304481038692,
 0.7060579897018006,0.9669436044882956,0.7113000971420769,0.08099418758898858,
 0.10485671303740651,0.9271713096687968,0.1827627382760053,0.837295919649665,
 0.2970617322227227,0.17930664069440305,0.8649120999036858,
 0.07007435795806194,0.5848231072612642,0.8237854130323756,
 0.22993113854618297,0.37198086788172136,0.8565746214904962,
 0.22010619555273014,0.5810153360330943,0.7932906454065053,0.4044953528352214,
 0.6203958355380629,0.32698440487108116,0.8607700106479327,
 0.24764775144926854,0.374968810548573,0.35157000563374785,
 0.21048695403984818,0.4363552032848371,0.6751009397655527,
 0.03529599774310954,0.6936522131101532,0.3465417702180821,0.2559424481352822,
 0.36846246054907295,0.6076855253081884,0.43538739240872637,
 0.9268888548642529,0.5878289435258105,0.26798166090698794,0.6422140364561862,
 0.7493843177779307,0.029155744905077774,0.32825775188590234,
 0.46123161215277175,0.6730850689885043,0.6184281631485665,
 0.009868930443715174,0.7123476878786691,0.8432075628642997,
 0.06058650924101838,0.17462896574898357,0.35508244146080814,
 0.039305361116020565,0.3346603139469638,0.2957120338657604,
 0.3254209443648267,0.5635541945579912,0.7482784077210998,0.20498461036307902,
 0.7607537517117805,0.970205642504558,0.6138909891965953,0.23539788466625666,
 0.9177003836102083,0.7422893712867837,0.5023232495067639,0.15708659453656504,
 0.36652588069749603,0.1921753312307105,0.9865773015583871,0.9859491582993832,
 0.07871200535237566,0.2885213179530438,0.21932276264761463,
 0.8423425778014199,0.2574597289870538,0.28570889385089926,0.7265364223618085,
 0.0806859528768833,0.27801735932624916,0.6126942640146338,0.4094220946171817,
 0.15964602436652975,0.32437021058242343,0.7631666418935829,
 0.021589780900369338,0.5660418164050411,0.025230651134456994,
 0.253138963112734,0.38706723940657584,0.05798870098944453,
 0.16169317011825146,0.911390440128344,0.8502986267824282,0.5774810188872648,
 0.2708023260944957,0.6053383458522692,0.15819506706954867,
 0.39218054628541066,0.9630707967457292,0.439277446910724,0.7070864040776667,
 0.6181915030494028,0.3740193182623003,0.7653938961869483,0.1953868150380127,
 0.22851331018613819,0.9259405362429389,0.2559504981931062,0.5379878044699704,
 0.8389184112569547,0.9689549116564625,0.7447535881142471,0.3917480316310439,
 0.04369614317897941,0.3084192777511997,0.21834788495362223,
 0.8403152187503891,0.6786605998647735,0.672644702967504,0.5643010792469328,
 0.9161386446926615,0.1101543630842503,0.5349707096143365,0.14719055582276686,
 0.7967467777073458,0.4673850782591442,0.49180017312472046,0.5921638471951951,
 0.29792093289791843,0.8277725642504742,0.9622388862486442,0.7163699956162861,
 0.7728185133179292,0.9022988349409429,0.6216586286959838,0.3165124214201087,
 0.42297929194757555,0.5421498084372952,0.1790367504594974,
 0.27557073792230546,0.8717141241951174,0.30165111826184143,
 0.23447713816796212,0.8706994202779715,0.9897292068582803,
 0.022189133292315866,0.48680669004174426,0.2531256784072494,
 0.08603421857103011,0.3636933271404059,0.10786684211917485,
 0.7340310969942931,0.13829778213660993,0.3526623859905631,0.665897578356207,
 0.0029594196138263973,0.5343263268542093,0.847324568228993,
 0.8828389848531014,0.9793010741001756,0.40863731944185466,0.2795243170810211,
 0.45768625095846927,0.7523785925533908,0.8968144136095905,0.8845533142712823,
 0.44330235907748483,0.06300634196437904,0.0017356802517102254,
 0.40311290954708945,0.7350473054959388,0.26511725280032883,
 0.06742606255833672,0.6182217542083625,0.3514700098640975,0.6371851759849712,
 0.7999649449003854,0.3364146814510609,0.435967178991578,0.4516341247576806,
 0.5484239017128819,0.3687752230308927,0.5798453564712176,0.3412643584548086,
 0.16423444865468062,0.6921310675466229,0.9399225822450027,
 0.19609165827310204,0.8803853449426863,0.030390209333162055,
 0.13032228249695932,0.8006923016320766,0.967087171015215,0.45959780795856353,
 0.1437582565263681,0.5519674423317743,0.3100987296520753,0.6460140680387935,
 0.7483881341692542,0.3918567873958031,0.5816643859780019,0.9155036604883467,
 0.3074093671155885,0.5005163994759696,0.4657848373874789,0.35854260083721623,
 0.45497083223529156,0.792405737044933,0.8788082603680599,0.9855724361988466,
 0.13290952482956142,0.0951740870783685,0.5962340889177178,0.6124636762962864,
 0.9629627989266117,0.32543825546259453,0.6167260917387893,0.3598974026125339,
 0.5118413724509869,0.11368197195634167,0.3523473115457165,0.9970361952119108,
 0.7001284269974396,0.38522080914642187,0.7716628967683302,0.5594499201890091,
 0.8677636220171732,0.22536902427216798,0.2632894602141538,0.6230433359914269,
 0.7959216736873208,0.6469684105542068,0.9260637362884002,0.49376411357568917,
 0.5418573355015842,0.4072742857258319,0.9133085442244587,0.7120342163380883,
 0.06757598368817619,0.3482016102953778,0.5048439339382874,0.6933566204706981,
 0.9731686480366554,0.30615826630810583,0.25663749554485116,
 0.6273273809208872,0.12441980713983958,0.7122084391608808,0.59645538083934,
 0.9356338995825746,0.3657927112560484,0.8044884201474766,0.2859578763743358,
 0.05312710582998603,0.8659566191145198,0.1637927901326648,
 0.21632885985325967,0.9254265794036249,0.7892238738270577,0.7779537252027414,
 0.6684254870723064,0.20711433784847666,0.19638632193498384,
 0.23865178097527195,0.9110072112665826,0.4476711362695769,0.4261671310434906,
 0.35317492713959076,0.5442297593056677,0.4349039005760509,0.5542501694076601,
 0.47159730848733983,0.37731300257918376,0.18600458482390503,
 0.9198979504262383,0.39566552157310064,0.6929439057522385,0.1508369779686195,
 0.4890125424482151,0.44689350324085575,0.9017349741271294,0.7654347558501032,
 0.3162669104405892,0.17613483415755105,0.6542640462725741,
 0.17897382732811673,0.43293323679094764,0.13550161473728872,
 0.41707627739118724,0.3293467473635515,0.05158549113942912,
 0.24179357438301707,0.38296248882072315,0.11126834033539779,
 0.750081837749735,0.7347749264117375,0.12869941613287406,0.44211700245226704,
 0.9554685283862657,0.1250160790154159,0.8112901853358573,0.6126950471514009,
 0.33620018531162954,0.10024734074268038,0.8858337649787503,
 0.6521929262251995,0.741505850431852,0.36652650336507897,0.7125097092378159,
 0.2800403819716252,0.17589160390787273,0.7969972665675411,0.7670187185135442,
 0.7888799975081943,0.4043495186417486,0.07281416858842082,0.4048288973765126,
 0.29491029674143976,0.4674771446726964,0.2198058351318397,0.5638943898817893,
 0.6348065949436055,0.3359560329827357,0.43568554911583046,0.1398047003865559,
 0.17298675945096664,0.6890345191915497,0.5270796283396991,0.6779735617997781,
 0.22175126079259966,0.6314400494085732,0.7115817200068072,
 0.18535385124622084,0.4818658275465051,0.2981042528767591,
 0.013318381516599365,0.4497407186220912,0.4660484479394601,
 0.5279074843035508,0.3352707912460531,0.0971086730067986,0.9353400047310876,
 0.13788972839719082,0.8415248616322398,0.10849792374317557,
 0.14754878680189898,0.6007009942686297,0.11549526236918117,
 0.3814377217481154,0.06452330286527042,0.7965624981196866,0.7366808795039809,
 0.5765554575016245,0.877063168778718,0.3645240149796942,0.9122328388316859,
 0.5672828927585818,0.9299250496507007,0.6913111486707728,0.5753702053746131,
 0.7224036867727485,0.7932928744375283,0.6166608191400494,0.11649250942351541,
 0.23224806068240766,0.046918993766736916,0.0823241395305695,
 0.8578751550745284,0.6906874867311276,0.4368899970199861,0.6564086399902274,
 0.7396284169249687,0.5849113343406279,0.07993263778788995,0.654594166160677,
 0.03451633037815427,0.7194912100870587,0.8506442508554188,0.3012990542223152,
 0.3916391368509686,0.6969687710429715,0.34929601161592505,0.6143366333430124,
 0.5748755568657985,0.4306095775577533,0.5203334747403356,0.850727949542954,
 0.9196733822860542,0.9155445013294918,0.6027316902471835,0.8980314397561858,
 0.2986506213519945,0.15198589016485164,0.31873511306713165,
 0.6966795394097666,0.4484933716778595,0.22903459572168228,0.4643819900663949,
 0.11282898069503933,0.5438095608328287,0.03736728692384905,
 0.3736604546255431,0.06551502272429377,0.03999149403999036,
 0.044238947138789175,0.07562323816023264,0.7368992538959026,
 0.42294997023873415,0.6550939998867368,0.877615021588007,0.5534918130168427,
 0.9649858703411291,0.4712183836127217,0.39246439473348294,0.8553835843798043,
 0.07651584675333045,0.5338883634202607,0.5000748821795068,0.8198077709748101,
 0.964393071718293,0.8754174833499272,0.6831815913746908,0.7771908922417444,
 0.44498246722430634,0.5819037382649246,0.49936864469653985,
 0.9808582081097914,0.8204554453182958,0.8656045111908414,0.6129806824820935,
 0.984384176856341,0.7877140329353918,0.16094785056572358,0.1973462268325008,
 0.3027425082020977,0.2252497842888631,0.8396038825018028,0.6063151212273954,
 0.8324224313440516,0.6176311772311535,0.7704777561932469,0.6205331530206553,
 0.4350077851160902,0.39474069593771244,0.8685405327606057,
 0.17631748771332267,0.993925129293975,0.2590167906044234,0.40216693639692835,
 0.26452590263736975,0.40511086734007096,0.3991899558069796,
 0.031150379980136367,0.36256808155705034,0.6409776207793666,
 0.7917900819923644,0.6390715631114379,0.839745685596108,0.7126816255806532,
 0.7905347348626743,0.271381353989565,0.8593132222465235,0.07255900681196936,
 0.1749716178825621,0.40672985358479297,0.8483505447378257,0.6646997420033113,
 0.5530266887413882,0.5940428878525119,0.6442499209409592,0.2647193637855304,
 0.31055000695955215,0.4871076688184205,0.6404153972570815,0.951292802959549,
 0.030425918104836414,0.19919818658738342,0.8552178981952161,
 0.14382816983069924,0.5208854615858891,0.4087240890428212,0.6120978286359934,
 0.8510411450520262,0.13429163660796684,0.6325886957690237,
 0.25940560578785876,0.7130849071224077,0.8636369409206792,0.8764612853269773,
 0.39787286821218815,0.9221885476797164,0.44001311174500846,
 0.31033373518726515,0.4217099521873199,0.841239324474159,0.9407327807612789,
 0.8321319581101506,0.880021921340591,0.49209091884907796,0.1132568418827471,
 0.311581659347439,0.4754425612675911,0.4343159133259822,0.37544114141177354,
 0.28344981463664143,0.1389606853066785,0.7728710635855347,0.5473499966786618,
 0.2256871527162957,0.7392667909747869,0.627615170519262,0.3871136610797884,
 0.7484012004091982,0.8619795316616298,0.40668480533566587,
 0.21143686936155892,0.5803541257045997,0.28852712132564895,
 0.15178260729311632,0.297572713460317,0.2555473287230837,0.7902170596353872,
 0.16471541975664483,0.17560693069854172,0.4594248959583255,
 0.09999909148209485,0.5809934286270664,0.637855105093667,0.16439971027344635,
 0.6455859052543423,0.9880081363501214,0.3159439117761822,0.5928852193967378,
 0.552771653482674,0.5866051838406323,0.6975812177175188,0.46298993371801733,
 0.23934445825435657,0.5433772827699282,0.5578917723623601,
 0.29350439798699035,0.7099057603657977,0.5731834482769498,0.9722966984073446,
 0.21522567380944146,0.887693071689418,0.06691627204667672,0.8618144371330386,
 0.734271627397805,0.4302818295605846,0.8940581271657558,0.9603826668749945,
 0.3554267410235594,0.7740467389563307,0.6651759287739805,0.13501532376837644,
 0.447057152092345,0.36246796181506824,0.012116241324061305,
 0.30016873447103865,0.9879697983323906,0.19019999413631794,
 0.9576404326174233,0.8864171668173044,0.9565834616487775,0.3300189220942391,
 0.8215523468198802,0.5278146703035149,0.11746168202166019,0.4873785861836054,
 0.627252059894772,0.4556539808097396,0.9307587903166171,0.0342065417215941,
 0.8162513640842212,0.1386111995616317,0.18681242993808245,0.9444880987860125,
 0.575004942624383,0.777877635084867,0.6454585152010407,0.9417172057020124,
 0.05366554197408613,0.630277486246598,0.9471128687118197,0.18985103991218544,
 0.4881764751946783,0.9257730940482434,0.20080025907843402,
 0.47744018726163473,0.24642691438746966,0.8365385007224623,
 0.10906190087995826,0.8769486490578408,0.24518633313553262,
 0.37520768695138407,0.7982475927622269,0.671147173539611,0.4797464311028733,
 0.8833038648630366,0.17092123217056354,0.7549837750241493,0.1993015234952049,
 0.015646096767757522,0.698267033258522,0.07700789155602039,
 0.44950341700431895,0.37115567770957203,0.9262344649464965,
 0.9488470070192653,0.25801035719050847,0.9960354159145179,0.8039835861330157,
 0.6339879766759426,0.7796269116307787,0.8885625558574397,0.0795505518814934,
 0.5047214527252017,0.7224999072018496,0.1106868230725162,0.3151492070467714,
 0.5097599549831651,0.02575138219474149,0.6353827198340775,0.4008670048474138,
 0.03299658822988294,0.035088894204623355,0.44352207052565407,
 0.8894493015813691,0.6150635416100335,0.3248764199574925,0.26806919563180354,
 0.6788927978904677,0.16964639470934628,0.36376023740032537,
 0.8992681218226094,0.8442757225113994,0.7461274857992894,0.24719933779083636,
 0.11388601657765096,0.048752685346767244,0.8931739733976537,
 0.9048471998804071,0.13485431926746694,0.23283136937591542,
 0.2610083342402033,0.5097869591115812,0.9550530185702147,0.26189487793832134,
 0.5766277942117872,0.9926194007201886,0.4194627513113727,0.5661915212766605,
 0.1612301199800491,0.6573321490263577,0.8101115599772589,0.3544165536430627,
 0.9154451872363831,0.3176104101565195,0.0015354613266489459,
 0.7301046481827917,0.39335512937277495,0.5306988327233579,0.3211681692376558,
 0.899443514551302,0.15344161356342223,0.8032312891899924,0.5261664095562721,
 0.9482964090606908,0.8138087861120508,0.3495161492217622,0.4305404995007617,
 0.8716327079174192,0.8145324834954268,0.4287484674270228,0.7907772139753079,
 0.0030912013011360484,0.640939369519928,0.5009468257245625,
 0.39309466251487557,0.8243779748788002,0.4076209812499484,0.7043581310637029,
 0.7419098299911017,0.7627213972145109,0.5575120681926248,0.24562187049578577,
 0.32947234851415574,0.5943513610996769,0.1723581737319121,
 0.36541605776880404,0.38089996425048156,0.20695236938951667,
 0.7573602557868385,0.43232125218733153,0.24462888146098138,
 0.2941212375206216,0.25171719140636795,0.8918558007822662,0.5091753656272785,
 0.8061129240034242,0.15375026080337528,0.18347606682912032,
 0.12614935240988945,0.5672745958779059,0.23298126159974253,
 0.5348536386720678,0.2067548202753906,0.964987511311389,0.6041399534991481,
 0.07400717838966087,0.285351903502747,0.5430505519475372,0.29626675197498453,
 0.29865011599420366,0.2080904932179607,0.6956517298569022,
 0.30612796656854924,0.6449422899750317,0.530947029051071,0.3914722873479073,
 0.40691765991060735,0.7517937576902077,0.651194387696768,0.7059117198486554,
 0.6482391430756984,0.6725258313132618,0.46162656701426874,0.8296237839292766,
 0.7554893582257628,0.7798537555531422,0.9437001196586329,0.9751519933168524,
 0.9991394859171543,0.28437241687257697,0.24470858826356434,
 0.35098295543130986,0.45103207089063524,0.5456866317749941,
 0.5537868044826062,0.29868831462003664,0.38046250582465735,
 0.6813754895392482,0.6672126444205022,0.9225575077803098,0.1375256329405412,
 0.3618012555027149,0.8706042135589313,0.575406784164024,0.496023504918655,
 0.5489938853991941,0.5329182132894664,0.03299939502775551,0.5821202662066016,
 0.41069173068908993,0.27174377883232903,0.5526242976167665,
 0.5636959492309124,0.15502493954939756,0.9545894267322111,
 0.14841911914764028,0.8126962855584006,0.7804983238026573,
 0.21714771990882964,0.5909946214130809,0.07574446401004042,
 0.5996215150509675,0.6777913396187152,0.07480416255356326,0.9072927490642465,
 0.6493150099661819,0.7980718724682536,0.22409725010280102,0.5245868976101815,
 0.5193935430944078,0.14566956599310343,0.1697692232987963,0.6958447428826287,
 0.3302587577421475,0.6741368539417231,0.5221860921746818,0.3106141946401261,
 0.8980464583826956,0.8460520226261116,0.6337890671395062,0.8917673413582491,
 0.8547490187458586,0.17616901584611022,0.6743196455521089,
 0.09852027337028169,0.632503451742745,0.6217805394762721,0.2273969619283377,
 0.9963656988709884,0.4126488665451383,0.02113867007815784,0.8585700502049591,
 0.9315187096482199,0.7722338229289507,0.9701113941374234,0.5696043229628738,
 0.663420077697792,0.21637907391824585,0.9453635575518,0.5812059468863369,
 0.23502830067565972,0.8916159371698428,0.9747010916505554,0.1912204704207805,
 0.4966768026555597,0.11557624710961967,0.3704587596045469,0.5005183743212539,
 0.12799671556796044,0.16032011837975846,0.886600151543649,
 0.45033980019383646,0.39007830805406685,0.25989919652460547,
 0.9889872303301819,0.48004369204962427,0.09034276863363322,
 0.16872717541586646,0.4467337981226671,0.9848353682679881,0.2710622418344064,
 0.23182284228571348,0.8449821295195175,0.26165727599729793,
 0.9208106465209911,0.5734143073175846,0.8847820540853573,0.5140977891116056,
 0.06507205458531529,0.6444431222597894,0.3450486005485979,
 0.05689499987433344,0.7341043434663143,0.0546908195616177,0.3509612746208125,
 0.6551314340159877,0.7059517820415873,0.7070993977862726,0.23010016926771604,
 0.6622870197902474,0.40979021929283055,0.18897089496940045,
 0.5783826040824939,0.5610901509312759,0.0031881479358925047,
 0.011594818333198642,0.08656308302932347,0.8317561137800555,
 0.21187717644170356,0.02758960066783711,0.33124233157422156,
 0.6525350308511078,0.9378536653135305,0.32574397813470835,0.9932262380223366,
 0.7065630126610744,0.6758057473747296,0.25155554673130887,0.7704240362409575,
 0.8664411267390572,0.015265249655311841,0.4961290951542989,
 0.39233520276700473,0.23870297227171566,0.39952994891444105,
 0.6709430610710353,0.3080854966086237,0.11957072332144514,0.8444302869913971,
 0.49174141209583166,0.1394823434567266,0.1369009556748515,0.5974499419679848,
 0.9089132841447293,0.05811459354921977,0.6239040365967421,
 0.33477409443465256,0.47896901014808846,0.39553097405727555,
 0.013460201225254154,0.1147850191702472,0.09200217847695025,
 0.5790386280779793,0.07909292997856532,0.8794094060571078,
 0.10018212240127311,0.51597893904263,0.41054354111283897,0.11373435314529401,
 0.6826488726961844,0.5979698769555217,0.1874742996451222,0.5490140619585369,
 0.567651058819832,0.892042804313091,0.8947331569527036,0.7499012871565628,
 0.9478868481754581,0.543448954475577,0.39158598439233294,0.8346581896896708,
 0.2458599527828501,0.3497143212445162,0.06762994262295474,0.5357187971586523,
 0.33281339536552235,0.4505730266871312,0.9989739498354043,0.6585881708185054,
 0.14265327156081697,0.2130340052857186,0.9634555748920338,
 0.34650040395919557,0.33907887190359387,0.45973929216685305,
 0.18377424542935283,0.1214956973178224,0.5468348159578635,0.9981875673747315,
 0.23684202740636573,0.05865608355357543,0.1552061154034603,
 0.006532969584663761,0.6598699350844374,0.9241423846525579,
 0.07689868113860154,0.6471782620968042,0.1530096178427337,
 0.11388017282390861,0.9019389961774759,0.6711629093617859,0.7550813442141275,
 0.5599419333572974,0.23064198569562033,0.8783209220190753,0.5538859377161411,
 0.2626907745995093,0.5759319097028391,0.797437001622365,0.767025736674221,
 0.9266208187632156,0.5244384498505961,0.6337261294505784,
 0.011082235113489158,0.7634927766826869,0.05554290986944399,
 0.5074530163887789,0.0176913870935822,0.7875399421754861,0.15835856005201487,
 0.29905381599873704,0.6769167374451891,0.8739083233919492,0.838477233471075,
 0.07975680765953996,0.5133445411529127,0.23499028279158363,
 0.7547579771133086,0.9118492598088276,0.7773249009561889,0.44003681761338365,
 0.7615477800725654,0.9386886424082609,0.6353625070040311,0.6361053132406287,
 0.5444985463359008,0.23899489852897005,0.3327446946721253,
 0.07557080989735565,0.9504200241681471,0.26700234174050264,
 0.6501446726535716,0.5586486634376724,0.3263001509413921,0.952825122176774,
 0.8986821728572297,0.2580492042359951,0.23941045833352748,
 0.23865451500534796,0.2798846799080448,0.11295252893800667,
 0.11261944904293897,0.34589756908392666,0.7416542572214013,
 0.6834077505081766,0.18452200728481338,0.6225261717205349,0.351123204671258,
 0.8191923217776969,0.9985495221648738,0.6063616585151436,0.8492672526612051,
 0.3071108045567772,0.8919622235978226,0.42146018623207127,0.6574470500286834,
 0.3041416312787979,0.8462706073880844,0.9920857857162582,0.3200183367052313,
 0.27075213337090986,0.06608462427648076,0.31977442942301126,
 0.8548385130964895,0.08681585808591774,0.7372868006897773,
 0.04347809885413556,0.3006719286943542,0.5372828988509898,0.8242258989960023,
 0.5308158363767942,0.17894698660298103,0.47808878144734657,
 0.5784685230298798,0.9509881401803131,0.852820430206456,0.22607073615553108,
 0.8474273147479136,0.7074654484017733,0.9060189166854715,0.48046280572939193,
 0.3973305466650149,0.7269657460643889,0.9637517365790613,0.38495283218949794,
 0.06470608150747859,0.06030452066026326,0.6926762868307648,
 0.8831077404574575,0.005305160581982826,0.7459561199237518,
 0.7124868862437131,0.1811723905091276,0.5634727586768494,0.49575660133429234,
 0.16031737840001115,0.08181428123159684,0.6712225166004588,0.227288742058936,
 0.34103802427753094,0.28571560019459297,0.23452653492256137,
 0.29451290367828953,0.5104795994608474,0.22348791021973746,
 0.30333687588085345,0.8260155128662166,0.8937648204954879,
 0.22303784443024732,0.25179630447831547,0.490386703840454,
 0.20290548020644483,0.44590067702559255,0.03510380532607793,
 0.8582546353329761,0.8163650789881278,0.5723012715572211,0.5345266416802981,
 0.3342664277908002,0.6671057756904484,0.07887561864781345,0.8201004585908203,
 0.2883236873896333,0.2749227418206731,0.9629568427704547,0.8825166687449513,
 0.6529457364775146,0.39708507204590127,0.658204162226008,0.9668305604546397,
 0.6878439897944651,0.5654296716522164,0.3428751674392134,0.5441003529504193,
 0.07376252682403983,0.498387414494332,0.07177403588163445,
 0.25443412359992323,0.006344940330860527,0.49272759921943876,
 0.6922139392918032,0.9126188488291913,0.31898185156997383,0.8596366970991677,
 0.05915522184479527,0.49594851360924175,0.3764328728127172,
 0.22239936529212778,0.6532955343932243,0.1440928187189503,0.7484339971202749,
 0.3742473040038129,0.8181185896879354,0.9540154415821303,0.7901110312900484,
 0.7155086802755013,0.2716661085397659,0.7355587338305749,0.6300996546395703,
 0.7118746043519976,0.7056007940698809,0.3452180791149865,0.985239923518983,
 0.050398229477133016,0.04424763534003373,0.3892788923776277,
 0.13513574836564934,0.9641548921266584,0.8373160417671496,0.3077652800466347,
 0.05941799115603663,0.5688722522041014,0.22516583642895827,
 0.1347422192519715,0.6961364732775576,0.37523936501875366,
 0.06083767262815187,0.19893025754748128,0.3443108822015568,0.645268850455183,
 0.8370752237860376,0.8240390682539157,0.8600208535010183,0.9232782647262222,
 0.3920093852070915,0.718841211263433,0.14916123406551007,0.6321481823477874,
 0.7661141642030157,0.06964820965474805,0.4465446779965234,0.6756888620485852,
 0.41523174508440397,0.16723090111912575,0.7013281530368201,
 0.32050586079356935,0.9045628217144347,0.606339123052229,0.45811927265926056,
 0.6121143805468634,0.2933622514263977,0.7552065240456384,0.4345652600449579,
 0.2516634443587271,0.859681772481587,0.8044583080751602,0.5509952765600398,
 0.7915634250523185,0.3777633340858646,0.7563439085373123,0.2124285918879142,
 0.8504073894100915,0.3290892216396447,0.09406135380287772,0.6422284030993781,
 0.7505442783145013,0.7783631639884923,0.6884748277294614,0.652670838739111,
 0.6611965879812804,0.8645459495079776,0.9935870620774114,0.9534949164667836,
 0.6832915079095787,0.13711184650250297,0.5355508582406113,
 5.56337277450969e-4,0.6643984786974182,0.026850928943931818,
 0.8516908661510388,0.03768547995312488,0.6414394330343381,0.6154020989835974,
 0.9765107269984838,0.9850594322040739,0.7199557748390129,0.1720135568965494,
 0.28067381047766937,0.43134480485982707,0.4307144982939304,
 0.6249904474967964,0.250898818028757,0.865760012304696,0.3723214390279177,
 0.9943783082700549,0.8335809214358807,0.7279191687285052,0.18211749101559227,
 0.6676689750088031,0.7479119274083522,0.400644408740096,0.21352310037133937,
 0.05277862428681579,0.15536927694964076,0.6226472973308378,
 0.3847060473208026,0.3793836679123883,0.552079214450127,0.8061092184918417,
 0.6643832241851179,0.2654685980752133,0.08669913110795124,0.2401918503266366,
 0.2058498934176708,0.005474217730535358,0.8476697765228982,
 0.18435272050532048,0.74488960757458,0.9422975052582294,0.5599663117212076,
 0.43089657626720435,0.8941891355025414,0.03347520125751857,
 0.3893232762970227,0.44140315900330795,0.3842797098021643,0.0948494347494786,
 0.6393698911942315,0.7855221868327258,0.8969459540700118,0.4135210461404041,
 0.520512233260414,0.9959312722874398,0.1165906327928492,0.9456489521184885,
 0.3851349043375615,0.10725171253871846,0.3422602296888738,
 0.13698561486683591,0.2140782551904764,0.4435484539766307,0.3944277224737983,
 0.40816085814078595,0.842174144634724,0.31299213461825814,0.5840802769061697,
 0.36899536052506554,0.027344242425657583,0.2006894649897648,
 0.7047274236330291,0.02990960390917996,0.8953474688572601,0.5939272221928371,
 0.3049882959732627,0.19323684735949187,0.5237113148915091,0.9474539754468,
 0.5210798622152745,0.25548912249782807,0.5499108277930681,
 0.36222928812727184,0.9398737144575068,0.7074384275778941,0.5843357429609041,
 0.3026930783097681,0.5305532334553995,0.6698156759086878,0.6044435609181057,
 0.12300542128364766,0.18202543369489432,0.530272666932114,0.1728408649940939,
 0.46889622285448507,0.2516082339871357,0.42923303401517166,0.658394637089419,
 0.0853265283594482,0.2119942942237626,0.694110740162849,0.5352884280647513,
 0.4623540485182366,0.04884791685875689,0.1630728391651437,0.55204269672704,
 0.5847327443218793,0.9760293032648805,0.5714913032319129,0.6434857648551392,
 0.19553898965178207,0.6284193872709621,0.17169745928294522,
 0.7741835921895165,0.8704472017689494,0.2843523597450217,0.8915080649123764,
 0.707556285602981,0.14815383118169623,0.8438538000672191,0.217669786573359,
 0.2971596656093699,0.9628112234664906,0.5384741084962386,0.9304177521297204,
 0.8359377126160775,0.0233321049670695,0.6887355912118817,0.6558525426367087,
 0.6745171647610599,0.9123319758116305,0.7726759299304544,0.6883881251410049,
 0.39474226456672323,0.28649737729045766,0.2760029510694716,
 0.5839131587280809,0.08816696153105763,0.3190110451033985,0.5295497849576585,
 0.8437739192807905,0.427472907110927,0.8232991114575251,0.6262457978742115,
 0.9608249910693213,0.751565243363463,0.7249163899176525,0.2067788460497838,
 0.3656128842856571,0.6016177225542997,0.7821193021133429,0.02367647803771533,
 0.9272952124442422,0.6374353594215367,0.04366351751976316,0.9651316007448272,
 0.3017654342697046,0.5126172048383083,0.5773467160227927,0.7875343025197324,
 0.5058976158153835,0.9702769549499619,0.5735955095260146,0.5448881045956103,
 0.7045597781250111,0.9573396101945936,0.6376731438591072,0.7702429749510926,
 0.5062348852593213,0.2788266270803045,0.8774089168596896,0.9294109586010039,
 0.9437868933938152,0.9148863592892319,0.3803855163957648,0.7459370909557553,
 0.13346607125262044,0.15032989853613177,0.17599565447793808,
 0.6976180751868439,0.9586808521927345,0.763785042553978,0.8956854706975739,
 0.259547845303878,0.5035132873350541,0.4200618764550309,0.8675467079892689,
 0.928113689521543,0.8157913291929734,0.7181885628099073,0.9022392054644999,
 0.13277541403708493,0.5634100626378568,0.6733950350798988,0.8249581980392844,
 0.1444586862372199,0.8708655886197654,0.980581938305089,0.79023575116387,
 0.48339584291325277,0.005044873050530474,0.09486991070340134,
 0.8924432979805192,0.16371569614970516,0.4734584269579787,0.8940667044514327,
 0.09670408348853876,0.03422418392229165,0.2579157487006223,
 0.08955545350115479,0.3559257397346425,0.7053193797153674,0.8434112985889946,
 0.223283911302388,0.12893582300929252,0.15904468360603585,
 0.12430606127379717,0.9305075977419142,0.9417215973393596,
 0.04697938699203141,0.7684636537936891,0.1560400476087218,0.7386192466972844,
 0.6536650658827363,0.9335305409358944,0.42992978642411994,0.22616635777408,
 0.2660386068147923,0.4556855019615511,0.46518828219180175,0.7282964672406813,
 0.3768427059423123,0.3433740011692412,0.0670591218265042,0.8501058708558313,
 0.04382380182831913,0.35192414329758803,0.03653539878946854,
 0.3348280496275571,0.26244960460605904,0.88985453461237,0.5633321192229729,
 0.5271414579680842,0.5286197513133924,0.13800309756021278,0.9177991188881214,
 0.2775095610449776,0.14619877627740485,0.9804267525164221,
 0.07145026426573253,0.1644101026447844,0.9678106582308773,0.8434172993981739,
 0.13603655352129385,0.6342444499567108,0.49965136211253514,
 0.5137556752324054,0.9005343728306363,0.5936902535059186,0.835352924585241,
 0.6390254830216385,0.2863018716505765,0.868096447764244,0.06395200410217139,
 0.7031651962190163,0.2524653111998939,0.4187514956270596,0.13906034025972858,
 0.3165031816400493,0.04395202810506904,0.051839211578487476,
 0.17973720253819292,0.9861456377702941,0.923665998269628,0.7586526485256225,
 0.5873531158177907,0.7015248029568554,0.20938120247080372,0.2598169779924363,
 0.9943873307133856,0.32864385707918087,0.7944210155181635,
 0.23740984407833288,0.15215007004932346,0.09338256659522715,
 0.7706411353139065,0.6435529550652829,0.2031990652951885,0.3898453179067928,
 0.8875085030586491,0.5288698857660928,0.5940806177946585,0.32814062385088694,
 0.8448135394670503,0.7519363883672741,0.6046493058565924,0.9366319399001325,
 0.9337142889368384,0.1671932196263337,0.9852675078927542,0.13696770033754235,
 0.08280085931956793,0.30029003996082293,0.8115423699525844,
 0.05355567929141947,0.03261620355593209,0.3345789851278702,
 0.9620110948221653,0.002198810903494275,0.36416806061216267,
 0.14177182275161027,0.6584413407186078,0.13086634707234013,
 0.7179369834869825,0.996620896778746,0.6322663984091447,0.6497522919729273,
 0.564000064217635,0.09536522584652762,0.7573126215681739,0.8881364972484028,
 0.7569059495575593,0.3250907423969688,0.09508161657095704,0.5024101583693519,
 0.9794298514004169,0.8896413643398983,0.598342426644354,0.20431236215391624,
 0.3025760781629555,0.07499893380826583,0.32861817133858495,
 0.8518530471461497,0.03496012404130333,0.7488050807465948,0.6321791921986345,
 0.252348720931429,0.9687601873186082,0.04295698178292784,0.24796966390713893,
 0.6868779967041765,0.15515400957291692,0.22355935752735723,
 0.5126396740892262,0.08143365289296622,0.3374592519108648,0.2323262832261963,
 0.11325852767940325,0.8552662956716238,0.9713336744510431,0.9192479917168535,
 0.05378969809855516,0.4804103669526254,0.3975326960376069,
 0.21806861708732572,0.39494209015937787,0.7856615916947168,
 0.19133030619825953,0.08227845805521117,0.5640640047563976,
 0.9587272099522772,0.7008811162527259,0.11028303045140297,0.3003336944525674,
 0.919235619886738,0.5460995109999707,0.5534179147740247,0.6297656318019,
 0.48560923354274965,0.2931799677870941,0.3698975195270475,0.9756553641956215,
 0.14292955039378974,0.3138127567545528,0.446962236836723,0.9066335482947109,
 0.05212477943968219,0.7861938155852916,0.026578726346277004,
 0.3788058952611032,0.2614375019388592,0.7265811824505827,0.5139751745590195,
 0.0038898140265035686,0.7521990547244197,0.9827549934736506,
 0.9251949000284033,0.7329663355244014,0.4250703971522106,0.9614375468493651,
 0.8634217607502592,0.2304275378787306,0.34877706358461424,0.7727359749810817,
 0.50179144013145,0.6299503910383861,0.07717420390502139,0.11491426058127774,
 0.8387809584025624,0.2475066486417773,0.7776696547555223,0.2765785124423783,
 0.9693487816717723,0.6091237157463655,0.6934207058869483,0.37496500632720076,
 0.8398384968472578,0.9894330331505456,0.9793270359606973,0.9400632920949996,
 0.41656286487644634,0.17479583313980318,0.9153846877732343,0.945191536912843,
 0.2249346184795894,0.7882069252263504,0.9351636097175228,0.5536107543707142,
 0.10198433436648435,0.9610577584034936,0.5961678309576116,
 0.43306855584481707,0.7872770317992167,0.41820677309448184,
 0.5586182163276452,0.16437447957395346,0.4517571578565891,0.5941097548511997,
 0.45214434451937635,0.4098241543432999,0.13834016311618003,
 0.3527830985879288,0.2232824164349807,0.8330508267191299,0.7634199360163805,
 0.8641849589271027,0.19826339436329166,0.4107613743927847,0.733091338298056,
 0.5806728976091653,0.9600575044792423,0.6527624566586647,
 0.025779218626073463,0.7384679365940562,0.5390560901773174,
 0.1412333520552348,0.5511490968998451,0.789521591850351,0.5491408084130637,
 0.1883463835320922,0.5824640132936847,0.6934560473187029,0.3180015629026318,
 0.3064314286437788,0.32377884635605725,0.6108083750217609,0.9633803431718102,
 0.3469264715805729,0.012790650748482069,0.2800123982579594,
 0.5708885521034867,0.31208934448377734,0.6381183867890375,
 0.14350335872439945,0.1426369975993118,0.7591164378690444,
 0.21366728609775998,0.25342181707665334,0.46633093730338593,
 0.36153013806088563,0.7565493763882858,0.37775757710499247,0.727107100560828,
 0.7533385648744159,0.9700129634086018,0.4004532421010537,0.04919951270685119,
 0.464575072319653,0.7652223843034094,0.530513455251334,0.45332713549484704,
 0.5913253845672308,0.7018862208260519,0.1478642896630309,0.19938582518751324,
 0.14579631990997477,0.11538899995754603,0.11584916230445574,
 0.1807659249095932,0.8242983136175052,0.2900776793996366,0.6475431906102412,
 0.5498109267783096,0.7319243534853459,0.4046981403349581,0.31910875318106324,
 0.8989219782380482,0.9446066493860268,0.7118470553776283,0.08051813744972114,
 0.1602432586759932,0.17031196803696036,0.08155145576857425,
 0.5572917713270246,0.6697184757053763,0.05288491731080647,
 0.24917684603039258,0.40145836473838714,0.3549419131432129,0.679268949683894,
 0.43960093950384727,0.19524337474704412,0.8806995340544292,
 0.47905148309744694,0.2013420165453247,0.17177296991755964,
 0.5935978761320129,0.5989223073752488,0.967892530969745,0.3128138430912232,
 0.21723711848947713,0.7298031389264203,0.439580189003249,0.7893384681118097,
 0.07222455794646765,0.8868111306419532,0.3748973817379181,
 0.005452241697064042,0.3127361340485889,0.5643214253246193,
 0.24168764123074882,0.3094979367642896,0.33039090324005316,0.722886884943786,
 0.2874307150132217,0.20362064421568649,0.357209800242157,0.5375342140830405,
 0.31106699033898444,0.3356166978592868,0.5894317279640063,
 0.12554256510442352,0.1826342082034409,0.8951665893665826,0.2741229211982543,
 0.04776192203858087,0.954167177589262,0.891784383880175,0.5082281935230488,
 0.6349822804091986,0.04795435462037545,0.9678500348481169,0.4522860662340903,
 0.3003310990733684,0.7855407261454522,0.24443788935094846,0.5769917510759122,
 0.04285976965740257,0.7951050981204366,0.056569992467299435,
 0.279217605263933,0.4191965496864031,0.607977222526103,0.7138095587013082,
 0.6496002324861819,0.07903046634308153,0.8097575421356689,
 0.45226373506924533,0.8905734360277535,0.42751164664018504,0.651816357706716,
 0.9847995118042443,0.31638585782113093,0.3463553010403695,0.7281998476664141,
 0.9619419214416909,0.16676842281076787,0.7312622323560629,
 0.29062442195030425,0.77343476192404,0.18613580999619783,0.8876956937544012,
 0.353228108134542,0.6969603206060329,0.13624067184685573,0.7985592301190468,
 0.7349074489866552,0.12690872070900172,0.3852075953139479,
 0.0011126925371192087,0.7201832905910092,0.5141779311632415,
 0.4784023107254012,0.7845281822963568,0.02804461724488827,0.3615969501357843,
 0.4349929175823045,0.37345082185035916,0.2600621749793389,0.834611491764991,
 0.7071129728074036,0.21297130820385513,0.37608838512932097,
 0.07166593879764405,0.27759122567005123,0.753000593127692,0.5599909995914464,
 0.5640466546730556,0.4141916921059474,0.6582052258452129,0.5828306839473708,
 0.8329782601301481,0.5956375229472481,0.30889853704413706,0.9040645987510235,
 0.46424134508710424,0.365725519894626,0.7981676753743228,0.04878031061454591,
 0.20534162162062675,0.7472120466059198,0.8966818970581498,0.6783149841545988,
 0.5542570018943965,0.563555252141388,0.8026387449950275,0.4517532993887191,
 0.5872115911428698,0.9386883417985201,0.2656727762729769,0.34671440777639406,
 0.6414781984424138,0.9901786159141547,0.6585015825338446,0.34537909236085773,
 0.5893368505251824,0.10093926140056764,0.550185228783296,0.841082321768303,
 0.8260772725631064,0.8979747184069606,0.6843453608048421,0.724728591481647,
 0.31139228239095684,0.9930430717679329,0.25456283155065984,
 0.4088207737831854,0.8758273934098251,0.5627226727386878,0.17832346208871197,
 0.269019159519049,0.49995842547760805,0.7674183439857711,0.1974846659118159,
 0.5221479988311121,0.027669147413222683,0.22413181076045208,
 0.40239718161299654,0.7192282831135375,0.3246066027035627,0.8845633680633387,
 0.7489687865284188,0.25682705582626786,0.8093677094971317,
 0.37257020376533323,0.3135536375056346,0.9186209366796135,0.6336521214940682,
 0.37647646180962313,0.6906579205507645,0.974272465153964,0.6571557400645702,
 0.8386163033805711,0.03935673875137513,0.4641607849262447,0.7153263270865335,
 0.6765472929078168,0.19734736427551336,0.7253981183078239,
 0.06169525874562776,0.8612455107507578,0.9826720462833642,0.7823023655490079,
 0.19829094199400554,0.11036651557314714,0.14814448370277356,
 0.0665565453839333,0.480797579339156,0.6135504089881226,0.8756332771268667,
 0.05759479685440583,0.050938411122644744,0.059550756866143306,
 0.9573911490338554,0.27340532464282075,0.4355148039705782,0.6570834318920357,
 0.37900969290000397,0.7357295847582876,0.8290188123334193,0.0179775563987713,
 0.487537882305187,0.018620396384146365,0.7154186714102938,
 0.17633771180758595,0.3886327668220768,0.15395888349988773,
 0.7709679418539391,0.08840717017231858,0.3237617621269392,0.6176542909011835,
 0.33402492743708945,0.456557381104167,0.81472574506695,0.647659332739797,
 0.9215450881873367,0.36442222518302814,0.9010830681538267,0.9921125245114978,
 0.19860056103398105,0.7165801495110391,0.6687589005454595,
 0.26705498922328363,0.7632465838176797,0.33857440040410514,
 0.6748913118494461,0.7465045454042655,0.3198289945010869,0.4087554077813166,
 0.34101397814496126,0.44663447637476283,0.7745177774095781,
 0.7621275095255374,0.3764575672786665,0.7518058459307049,0.16810548733105812,
 0.9499355974211635,0.9722801947695657,0.4760860115966221,0.07692540261563396,
 0.19434797713594698,0.24005232994244607,0.22737781073857355,
 0.20136012935331837,0.8030346030549389,0.31830284252649554,
 0.9649883728261847,0.23847932600831,0.798771203811235,0.7210170125076462,
 0.6324016905655906,0.1821741511505055,0.5894689910464856,0.4410229959399541,
 0.1079637530055273,0.5333509865199243,0.8384007389607234,0.9337281658968896,
 0.5445143722319865,0.1523610871421539,0.34478223104390227,
 0.38956649394416554,0.3118558162471645,0.263323122895051,0.4663360423309444,
 0.40693275168007026,0.19242900898984783,0.8582210665342634,
 0.2966145627537705,0.602733889258309,0.33968072229489865,0.5479603943912367,
 0.7704237380319041,0.6638742761875812,0.0776722231369239,0.3105229279267193,
 0.6676043219754841,0.07106962536706618,0.46986436154331124,
 0.3871974071973924,0.12622241507854948,0.8513119923582705,
 0.23745700424043703,0.7521974215366617,0.008744512103235569,
 0.9206849775787864,0.6856788972847072,0.07862405428958319,0.3614932949226991,
 0.7224704406127711,0.3018468161437766,0.748510025865407,0.8279805047123179,
 0.21750379535554898,0.9615330553110145,0.513657483402037,0.3332577343585511,
 0.20507857738742685,0.1177930389962214,0.7744406971074856,0.519251411201243,
 0.5585983411408239,0.27625292094611875,0.6735611042975579,0.9726049700948254,
 0.2289681531686838,0.5445363072028453,0.6988082941890571,0.20429953280271407,
 0.6645369289543602,0.655610831729206,0.838773795733369,0.19796178244372853,
 0.07477476776373104,0.30909938849941554,0.5195900305058325,
 0.07792518292479345,0.3875092685526812,0.2148267969840978,
 0.28490775848191985,0.1870506951388924,0.21257744480097862,
 0.24148952370123433,0.10131497924701094,0.05859394267040341,
 0.141226328209757,0.15237834305579,0.27536653230811525,0.44031685229524853,
 0.18648381653350465,0.15323639246852094,0.29733199185821235,
 0.5380942727427711,0.44717167277267944,0.14157542245708654,
 0.0661953014796457,0.6407546696466587,0.10219993677930506,
 0.058811809429976436,0.09426160869415712,0.20182875811911172,
 0.676089367681207,0.32943578470148227,0.9902773681594006,0.17325778403782266,
 0.4553391160299173,0.6698668095063279,0.4406935134182738,0.33310075446251763,
 0.5542285093777259,0.18853567810579863,0.053428809561474,0.32441377879664757,
 0.8425259488890743,0.3397815777953095,0.6392444158571174,0.5782116427469873,
 0.472381347798433,0.3746305593230449,0.5793411321695576,0.7843524248966423,
 0.8389089100787384,0.5520518815791833,0.42687810079704414,0.651590841592041,
 0.17114230753307857,0.212366042888401,0.28748931480969686,0.9152036582608418,
 0.7411406190097836,0.6353508884552022,0.3985122295746757,0.3659157858556048,
 0.2150606246815946,0.04497179203137669,0.2711876870970751,0.7265577802940745,
 0.7796071041919816,0.9230077242791793,0.86931921804583,0.34538037996229376,
 0.22018808609341833,0.9832904633906115,0.8554789666025612,0.3779194626980821,
 0.86212161878356,0.31680286330049656,0.36846394707453867,0.7483742252884356,
 0.7992940840683298,0.8448109798679508,0.5820284929659211,0.9975876269243101,
 0.033255185750514005,0.14515245292679968,0.3854693252478749,
 0.1554156989489096,0.3982217993076679,0.736315778557523,0.28435331240872974,
 0.6528615125519482,0.0629741501001515,0.7163953472340733,0.6569202239167311,
 0.8221312377263453,0.49599241759044066,0.5793107206914337,0.3824058216970613,
 0.9503004077550611,0.14332914281397213,0.50574921682353,0.25992132724059314,
 0.815227613006413,0.5267421141715373,0.01595201880346464,0.5984838629482592,
 0.6523777259215425,0.4229178129392984,0.37792701106636317,0.5662623420849725,
 0.8950466566293082,0.02563849971803922,0.1083733278006026,
 0.17653864301333455,0.09492272349182618,0.5514329628570148,
 0.19004110216092984,0.12129255421054708,0.7281581092548158,
 0.12457271650080903,0.8414329316109002,0.9223629722189927,0.2109120868537857,
 0.708289035872681,0.983764648877624,0.2861396005564839,0.5590143291679083,
 0.3206818991258594,0.7669786379198118,0.43444726225479313,0.668114696969609,
 0.8936838888334958,0.11740453135017836,0.864016454299286,0.5407954009386089,
 0.6352249947812462,0.9603638824293378,0.8334360508085187,0.5882719365929447,
 0.14775989038439596,0.9716425672339394,0.9500437774907261,0.908450265507946,
 0.22433911655510586,0.6775405701713377,0.6597880052479191,0.4915153129032701,
 0.9443539666608616,0.18497472489267053,0.25768334492360545,
 0.3112916651599723,0.2156626649436746,0.7476577703644234,0.7468254568025608,
 0.998872025590786,0.09303939166503816,0.1542199782667849,0.5354346394466727,
 0.6655563478484321,0.020204413206710736,0.5851041856429388,
 0.45956580243932177,0.8874931076826806,0.26061271013249643,
 0.034347868913399915,0.993389269674154,0.05228921508843465,
 0.48139479536165375,0.30905774185446677,0.4248700303639168,
 0.07566631530319734,0.4097007628890601,0.2915464902270124,0.594151925219756,
 0.502328516404095,0.43682673078403644,0.9805773211721734,0.5781041672150016,
 0.8725515902071026,0.857881944413544,0.9009444521610436,0.26802161100446964,
 0.8726039234556084,0.7477628195480704,0.8604010410455919,0.22622578470000676,
 0.6341885457980586,0.18715419749732454,0.7693915870820114,0.911108480396966,
 0.18949602991871406,0.6577302289858312,0.2906543723446571,
 0.27557936993013166,0.8431149383655687,0.35107707001930466,
 0.7243407439580242,0.49388981609284155,0.5736463589177196,0.93188015748071,
 0.11010240264754656,0.041799955130330746,0.23949305781176877,
 0.6492813526234085,0.3111477083247891,0.627125547698075,0.3813493003314363,
 0.5923525878373166,0.007692276233266515,0.8718186832968026,
 0.19151255210973495,0.8838478952647911,0.12440633549819347,
 0.5522011534447775,0.2969371141646966,0.284168471705912,0.3295242607889929,
 0.741266292546789,0.6548821542394101,0.48364581965737097,0.6526984695862232,
 0.11509872142414779,0.3730540679297365,0.867991511140097,0.7590576929541862,
 0.5176286849469873,0.07489360087501029,0.7128197739123358,0.1287917269625859,
 0.9738804704439659,0.29901673602194867,0.0632199750020851,0.7995871728517993,
 0.5285851863622678,0.885304965710514,0.9484725577709711,0.929621157501306,
 0.9860230094241547,0.707929921471419,0.8667318863216826,0.3533126708724801,
 0.7147887009705234,0.1344612602554559,0.18982554676023966,
 0.30160212352343074,0.2678529378224077,0.3379316408921932,0.6654828504546441,
 0.267778557214133,0.07754612216417833,0.26670697164490453,
 0.45706430738496395,0.6907485408147788,0.1517602656261704,
 0.17407246126437648,0.2698678739092324,0.38166422437701175,
 0.1492198783652151,0.7928281001585931,0.7243616658276557,0.3696787572179211,
 0.38230018044582703,0.7605300939396796,0.8243479759947475,0.8674410495980761,
 0.2607508228554569,0.28751362882388365,0.10288805087997932,
 0.5612042358375577,0.5605147601292428,0.638515453848508,0.16427381403699082,
 0.7350922038639659,0.3598292990252544,0.06805894389751765,0.7433275994452146,
 0.24466923967848608,0.2672179694202582,0.36766743892601983,
 0.038876390045699316,0.8461302819372726,0.6344920475752384,
 0.6030060314135535,0.24709884716530683,0.14413103543409034,
 0.7772745041515378,0.44719967327165366,0.7864369034859716,0.6412982749868752,
 0.10226648684624995,0.1718964251792272,0.054256911628327686,
 0.1506105424247587,0.21338802390716272,0.8640626891595051,0.3003108343011204,
 0.12103574672938244,0.5979504892452031,0.07996137951629179,
 0.36505403732487707,0.468450780618595,0.14792084658160898,0.7540925300029594,
 0.06300896118066279,0.5093125114072009,0.9784506340846041,0.731933766893202,
 0.751956909710377,0.06073059347404264,0.9573800637546102,0.34054422481314706,
 0.9298011902416383,0.6508903686055371,0.5725616000750052,0.12425114413157845,
 0.17965505747562693,0.5883976325059314,0.9151785348191169,0.8684834805826318,
 0.7415480100476244,0.3162658201180958,0.26632380022412494,0.3609763431763918,
 0.8231103768004394,0.6213278710808029,0.6200989603702498,0.7390015251585539,
 0.5663989092346511,0.485498962103577,0.9574149452989645,0.36865197143980954,
 0.6767436550904592,0.9479050511975565,0.6721597356062282,0.3153167938343244,
 0.2384518972806694,0.665406624030026,0.7697320769592492,0.5527286393548606,
 0.42110267995973427,0.38376484884107775,0.5637618562877307,
 0.7113015536808084,0.39312544147424056,0.6378438166724106,0.5469068089099568,
 0.9552446406371065,0.5037919464616096,0.24860610106309944,0.3175700381838674,
 0.9676737611755852,0.9352331399660088,0.6320389151627357,0.28554626816342654,
 0.257316990650442,0.3913450649054667,0.033976957290538845,0.2801530297286873,
 0.2969543722475274,0.9929170001071012,0.9888749404367327,0.9587143861500744,
 0.39262873648555363,0.7301700414409034,0.8818230589510354,0.7085690911450515,
 0.010373422241632735,0.5030570601685876,0.5434867308652445,
 0.8900552470846245,0.8528373090448063,0.5397514898520734,0.38395312508023083,
 0.22392128153096102,0.790047736436316,0.07776760902476465,0.4603107539920309,
 0.8311502451476903,0.493743421794424,0.9471772668736246,0.9146959090065676,
 0.41342870870297743,0.11536424936894185,0.5664600205017095,
 0.7509279735920453,0.49725000673567465,0.3028792356047665,0.6352477092374819,
 0.4032390760595226,0.17371752880702973,0.6006560748960849,0.9790874182551924,
 0.5942256502738498,0.9426489238501402,0.2986829956998356,0.6094798428609861,
 0.3694541469724568,0.5920310747010379,0.9519161456317428,0.45003568720691656,
 0.8657890405990641,0.3626355987411737,0.21063510819336118,0.6633470480946287,
 0.9370709577797897,0.14592083517213494,0.3733552438613994,
 0.061333952288001536,0.7729925098561783,0.6312325820636052,
 0.13824282127814091,0.7197042080405845,1.0205138270480418e-4,
 0.8637291829323129,0.6813242995604932,0.05345322322242707,0.0488226198062246,
 0.2973808475967681,0.04151956869026685,0.8861979350590679,0.3553627413009378,
 0.7253813146913767,0.7367925489261871,0.7819290732927862,0.5677160668380848,
 0.09547496323400928,0.772913284333178,0.7494711937772285,
 0.052966427324201315,0.7977463588888491,0.5810475634280011,
 0.2668357743279085,0.65507992719254,0.3768215234982729,0.8459265996662422,
 0.2340442523411298,0.48113007998079915,0.1009263849380736,
 0.29652466275935685,0.07378218536300707,0.27519606020430476,
 0.5289050053578225,0.18920599982285902,0.3463993446395205,0.4021776737575281,
 0.5458239393266575,0.15750938701904615,0.35020845700142333,
 0.1541247333007979,0.17299897924387453,0.8534022456058992,
 0.18067473482705876,0.4980016927798132,0.6896915225218219,
 0.18149257538537755,0.8180473857293634,0.9917198504804599,0.6263152168166775,
 0.4500281982046159,0.7387174350745236,0.7953494532361464,0.7967162075863321,
 0.9365739805436012,0.5888941197260145,0.7917276354483473,0.13818392027996773,
 0.08282146535580648,0.5846093396378638,0.6875990814414666,0.5066217684724473,
 0.7930186908089105,0.45372331818662726,0.15600267939867163,
 0.15272762756551983,0.9425184853961827,0.3137728710467812,0.8901632404772921,
 0.24158406770348773,0.2663370973261493,0.46807793360775785,
 0.6997576881347094,0.36216705382488945,0.23244465650592339,
 0.24726566948693218,0.3669822104395668,0.5102830619035398,0.2808876139163562,
 0.7389723625804879,0.923374364331707,0.958956487423246,0.3037372460167982,
 0.39840041192672004,0.4620424919076076,0.12972510253915936,
 0.9324904136472907,0.8073529551861454,0.2973452286239002,0.7307627777883698,
 0.9722103509561826,0.7674279065458445,0.9275010204864107,0.9578870139787898,
 0.9556790914695162,0.9112974064338397,0.28012761040191925,0.8468773188840852,
 0.8850171377719348,0.4730524554182971,0.09449900569623271,0.528063689609217,
 0.39376536494123204,0.7568896035954092,0.40829310255263085,
 0.5812226716682054,0.5529327875484806,0.8335126044676129,0.30646457526461757,
 0.1472302027915573,0.7072801290238198,0.8342588397545633,0.40894136322664354,
 0.6948694650076379,0.8654736164895035,0.8825431163059149,0.2180116732921341,
 0.7297539701108698,0.5122130618425396,0.12189406078918275,0.851118146234889,
 0.021241928627566575,0.4440764607481148,0.9962478163165039,
 0.10078070650548032,0.5751255128952998,0.8865158456084297,
 0.17519504846238498,0.7035964548888378,0.6341189549715522,0.54116620994573,
 0.2874842921081242,0.5494662257583219,0.2880756324161473,0.2193416447950164,
 0.033929424819339804,0.3403695113182915,0.9439783504221462,
 0.9970161014456733,0.498045373180922,0.96767559418563,0.2906310347376626,
 0.6901121370661859,0.9140461204371934,0.0397758778860704,0.20573845884142183,
 0.5903353344228273,0.06488500487232862,0.1416631598763739,
 0.11220915906996187,0.897862088291428,0.32881668296378297,
 0.26656465106771654,0.45586067083678883,0.11769802505061333,
 0.20960002058539173,0.19111706698566,0.694786120586256,0.416842006200551,
 0.2317090947283642,0.1804808125008811,0.38216719966458834,
 0.10716189583896862,0.7591985176550404,0.8417008569253627,0.3453974706496621,
 0.05437889179812294,0.9910949714740206,0.012157423186694993,
 0.11280316700171122,0.621076077823613,0.5044375966940734,0.9115936711365512,
 0.24310030863505183,0.24075635084884095,0.6167504885111053,
 0.5807203575072614,0.8829677721089245,0.9159805051861304,0.6495911449762811,
 0.08875860252391043,0.424366747725305,0.9150578323305546,0.29421414642324306,
 0.894221659926862,0.013302164530282767,0.9547431356429457,
 0.13853633582298785,0.8061812249722228,0.38449796513739565,
 0.16776378154062055,0.23070505740591063,0.005045313961145714,
 0.969926028034555,0.6007860480206244,0.017654624569596544,0.8168404194567189,
 0.6604966108466361,0.7571708727325618,0.40072346287308813,0.8692570108805141,
 0.3163649290511481,0.6055409189538706,0.9671396896994964,0.8358585711972448,
 0.8180462638946054,0.728481425373311,0.7315624474401937,0.4535367596053691,
 0.21704231480147762,0.16196812466471677,0.09151168636596552,
 0.820049838793614,0.5725617424309748,0.3246832578345463,0.760438248617815,
 0.45926468035834433,0.3350061585997055,0.793258303620406,0.8122072891958478,
 0.7029732984494854,0.1695911635679358,0.30401176522731665,
 0.05264112031913826,0.7466157809455076,0.8241725203352614,0.9571549674543053,
 0.15736236451577335,0.8139448636763668,0.7182167247636184,0.5459059308506669,
 0.5810399612636491,0.9199808340477931,0.27767723072576844,
 0.27019463281937006,0.6485298272218425,0.9757916953502812,0.5684805746166904,
 0.838279889274912,0.8675013431194541,0.6021192895133984,0.2579948412428055,
 0.8441543419093019,0.30660181564948086,0.5857678839825929,0.8417448300791496,
 0.3676132138611885,0.45719815718229717,0.408742370622343,0.6411061179091079,
 0.42989292792640244,0.9995193219395351,0.993926070090629,0.04423572371202522,
 0.9248443356671562,0.18681559310016504,0.4991891718337119,0.6772608175453582,
 0.7060614168491913,0.07832025208598359,0.4759649340954226,0.2086835088193142,
 0.19524956058406018,0.24142364840066144,0.4745718162979875,
 0.40927414254078665,0.5008951992701487,0.9380882660983411,0.7734496658130692,
 0.019070958138383665,0.8512041656417094,0.9252258042889226,
 0.006080757550511184,0.6127293807491327,0.6564244052243126,
 0.5687979895351788,0.1176364447632835,0.3182756861094317,0.16305540409274066,
 0.5754878300280106,0.656263533958506,0.17925953022350205,0.41784675688335915,
 0.3545155360437533,0.702261980464911,0.6398535404842551,0.39379184653878296,
 0.2189252442361146,0.44322363002198895,0.9676648660871744,
 0.34735283289003704,0.7707492012721645,0.40125531290060656,
 0.9580364748647294,0.24306271069610508,0.8453411737702871,0.4681602667498955,
 0.9782026375070074,0.7243296212911249,0.04879610472917728,0.8144409154674969,
 0.12244459785447592,0.8125727721907265,0.9734751960671978,0.6588373706396228,
 0.18489484969243963,0.6080327647705017,0.5988089235586267,
 0.10170317708840315,0.33194790667954455,0.46633454272896846,
 0.29354747159039085,0.5446531873028201,0.5644228855369389,0.4533572973074095,
 0.9313566498400367,0.5062291734461454,0.3037344061964574,0.6351566954034072,
 0.7211368180015728,0.4979182983876921,0.9524484612075375,0.9180889129664975,
 0.41501124381402443,0.2702721958096387,0.584430772809466,0.48205797149401075,
 0.28578991156998435,0.22637361765706943,0.28523571556454763,
 0.20764505627706265,0.8830594833622369,0.9131135888557299,0.5646160690308543,
 0.8763473146665994,0.7157786586741426,0.9059939955418826,0.2462678799325767,
 0.41448584066270866,0.7496240307741597,0.9411609937545796,0.960501641832575,
 0.28415321131415316,0.27653897789286797,0.6991498510195249,
 0.08992056873287435,0.1552581905728787,0.18661829419533738,
 0.7226112757097685,0.3175356240169428,0.9745980320868471,0.05614869182506954,
 0.7466483707626068,0.6400693787831241,0.2189721543556884,0.9349997061070429,
 0.07923274181867257,0.3538157335559976,0.3688602106861171,0.3583441366355625,
 0.3428681133761402,0.19465276219209482,0.3465971441805916,
 0.37865719000230635,0.6384408354681195,0.9071969299121453,0.571143766590205,
 0.3468671884784583,0.7757726339625821,0.7897930809350799,0.8146223004785408,
 0.5727874895553016,0.6548692266268755,0.1042496738974853,0.9949168759811844,
 0.047875521865339166,0.6523556851795992,0.545086921275677,
 0.15707201153095718,0.9416198162005357,0.2413641085208521,
 0.11901618985802098,0.9551151248844607,0.7214188033337245,
 0.25844419975553923,0.3457997721187658,0.6559210344668517,0.6796850742900664,
 0.6890127451076999,0.43731322106499826,0.07184043936327145,
 0.7216681480084662,0.08638182746265388,0.09908667014007821,
 0.15803908433723934,0.9646541343158567,0.49021743540003004,
 0.20188606251680707,0.47946826840070844,0.9428248426385556,
 0.5481095120196721,0.6792224008321407,0.7787995667601332,0.9531379911068971,
 0.25515431753889684,0.724467220976768,0.8431436435868298,0.8942030216792808,
 0.7155187537002157,0.3278404221356125,0.09147350630511297,0.835618754963311,
 0.15870645507999126,0.047231660622285454,0.555397351839643,
 0.9405504572451622,0.5682695817311577,0.405834028592468,0.10730917376880922,
 0.6583168905757997,0.3839815719404034,0.2515042239684445,0.7425284346438631,
 0.5219309159258445,0.6289058430385841,0.3779602358893339,0.44338730699602713,
 0.28266871597287924,0.8994897863884601,0.9686431844626027,0.3217482810388257,
 0.10045452519440479,0.5144806825442012,0.46997073335048867,
 0.5279684046290412,0.24995866707391157,0.743185637382039,0.9882917750139691,
 0.622741301276437,0.9502589995441862,0.24462628440899614,0.5971953626781431,
 0.8849903660336765,0.2919136676672097,0.4624534780357541,0.20609722287324106,
 0.7703281948114057,0.7656873621311752,0.5433862080394137,0.32174423435373056,
 0.18424775648524472,0.47354571004663537,0.03130030082439639,
 0.29417518836578416,0.2644676688749956,0.7432281792622691,0.6742023126378986,
 0.4608967933874464,0.9538164457268183,0.9138383144069558,0.08608110635872046,
 0.8460733998806533,0.4279407845886989,0.8946460741729146,0.9850201131867802,
 0.46727776638388885,0.7336708798595711,0.5006905684486109,
 0.47716803267047914,0.2955547455433374,0.29591114180484346,
 0.08345114555658961,0.21267117510319722,0.532169370259653,0.4371946949212141,
 0.2892322450886422,0.8477349618597096,0.28232259730595344,
 0.16545959062109883,0.5289403574160519,0.35025131826083566,
 0.2446901460572224,0.4445446252947727,0.9477096418815334,0.6347653817914085,
 0.3165166413620172,0.8806064008947919,0.1635905833440403,0.7002820454211324,
 0.42405357244829556,0.7659931050622854,0.6560401810413292,
 0.06951836946147849,0.8904261142466714,0.5990518612659257,0.9218005821253825,
 0.7423794636249784,0.5948208061666533,0.0542001433132524,0.1802262281634719,
 0.8442396371278146,0.07312448368131774,0.47076505700260274,
 0.13610761465934573,0.22910712783365728,0.0526059937076343,
 0.8644110268023645,0.5521714275894648,0.9942399435556045,0.30535568550866443,
 0.21392602501165714,0.749530112577311,0.9883538843067157,0.8052970854069774,
 0.8109569164226285,0.6494158433316507,0.7334750170038469,0.43169089211834444,
 0.5643488885921135,0.5541180822924782,0.9434257164848048,0.4860803829533529,
 0.43589108834908563,0.10537879565310049,0.8855395468774185,
 0.8463248905690975,0.9860171057466248,0.12917653135356089,0.9631352718465692,
 0.22106407197326006,0.27079946974157987,0.06994036823547334,
 0.7215942349466555,0.6510447254987252,0.47896000435514074,
 0.34807327081907014,0.23969350792195476,0.9646714755227951,
 0.5966402024471463,0.04444545723472437,0.6027230780465734,0.9167168714606359,
 0.24510927609430022,0.8424549461083507,0.7569414053734134,0.8440676769731507,
 0.18543108146516807,0.6301364309980261,0.024888007531457812,
 0.4974681579218063,0.09921023381939609,0.062092825055330314,
 0.1304063527913728,0.5826189363375388,0.6036735008330575,0.12062207704687988,
 0.15806425063765928,0.907477792479356,0.8570556846716262,0.8593642360218425,
 0.5300559944132385,0.863825207990956,0.5966112096332842,0.644033745820566,
 0.9316091045755525,0.7583675816145224,0.47609252200483876,
 0.19414351496076543,0.7636236334354758,0.3442049976965006,
 0.08083184374951469,0.8907279726326455,0.9858615978419943,0.9284670741693983,
 0.6322588064868782,0.9692860497905279,0.06087918042173324,0.2757224725366402,
 0.15928481954048535,0.40688761269885143,0.8448982775766577,
 0.15137613797958793,0.29116143954038765,0.4917049461542493,
 0.21229722195626832,0.2630220528111291,0.04907406040089335,
 0.3843128584759685,0.7505914767031586,0.6705438996888367,
 0.008537311668722625,0.40981656166288327,0.4055694332016485,
 0.28931514144890635,0.8434067144819193,0.38146409136409953,
 0.6056400628487019,0.4686563367234654,0.247151328688061,0.7309917338758815,
 0.8697544653304894,0.4726469240364005,0.32325118412172327,
 0.026987130749788202,0.03276433415701874,0.40200176771275775,
 0.5914799289516326,0.943499973419871,0.8786314200441598,0.6294290822498962,
 0.7164868971126984,0.475778380333431,0.04907382264553639,0.21584190065588793,
 0.35417496185487407,0.08374958306874747,0.6901016222626157,
 0.6267874090657817,0.26833225659931603,0.7703858729313101,0.7126722101765537,
 0.43673784330755916,0.4798366848864367,0.5609933064071733,0.9570664476490455,
 0.40803443451616017,0.08573584830065939,0.9164697623407166,
 0.08943005297493545,0.43843420241562614,0.7316765001350045,
 0.9689899476223027,0.7226225714901848,0.03586508523844034,0.3988381328293745,
 0.27058868180874096,0.45553945833872456,0.2903325087329326,
 0.8769256608439658,0.6176758374286859,0.16159582735804112,0.9372356773636284,
 0.5965184080141781,0.5613741742936598,0.6665570041477077,0.8117144630657984,
 0.401874742532066,0.7222389404847249,0.4166473043917577,0.07731780778016173,
 0.08588414734289673,0.2466496055643278,0.6785421970335846,0.6317419259579338,
 0.5299908779745568,0.7790223842349907,0.5822337617341415,0.8099164769866223,
 0.4796011917056573,0.04363348760659802,0.8570268956283363,0.7056419151967663,
 0.8952403113459336,0.980781527974289,0.20272593712689224,
 0.011999708497694828,0.36267930525767555,0.17447792199825862,
 0.16937417586088344,0.8919051237213768,0.26572892227248524,
 0.7308275355410604,0.4283354385214051,0.843088244950672,0.5831702362126646,
 0.6245964055100319,0.839603977097349,0.5939488348874593,0.887707564527334,
 0.7911979851514727,0.9870583760564386,0.7450604056576359,0.32818361443633237,
 0.2736591376928055,0.9667437784490223,0.20167879761934626,0.6293349760807023,
 0.4699042788811205,0.509455108536765,0.3369172664256681,0.33211419645621276,
 0.7261461288760264,0.3441726441112938,0.414380083452888,0.058165449821901194,
 0.17008326839078103,0.4298389547178265,0.8636781030271332,
 0.07501658969932623,0.43525692665058124,0.20207013154452425,
 0.514061209480483,0.2896881364736993,0.5706465305473194,0.21439624254344314,
 0.6095059529289166,0.9861628338466486,0.7125735118243146,0.9104432866027381,
 0.6140669045772884,0.9633587029613628,0.5553415170481792,0.7077650226214683,
 0.36257064342122414,0.15551136141901978,0.021294932061654892,
 0.6031819936818829,0.337476167293433,0.8339296333222151,0.7897367562847325,
 0.18271869260680007,0.9050087332291193,0.7362434006492073,0.1527919449371472,
 0.29475320314027786,0.3043856708385527,0.8074964380821034,
 0.27724868366213873,0.0442855411404417,0.014863503308847448,
 0.7946348848565844,0.8285468402357886,0.023833467569855493,
 0.47921044737272656,0.6434923064192104,0.7365513745448385,0.9849911009143596,
 0.2840461071191326,0.042111488351750936,0.17526897522430918,
 0.801003532720179,0.6726870223644,0.9041635556863726,0.5131932828894823,
 0.5573107379564333,0.8475679039373544,0.5743666779327472,0.7826690805306782,
 0.14804410880688923,0.7323187756320878,0.3297254004631591,0.9436517016153864,
 0.35811007194095346,0.038407687170776805,0.11003299034888514,
 0.3655978713974959,0.7696512172408322,0.6262812237716028,0.5639515308791812,
 0.1589456973040274,0.4889983052040593,0.515036631733947,0.2335328437300067,
 0.26946421775984575,0.12923483731231822,0.631664357102706,
 0.33459510153833516,0.27945320267993146,0.3526547101103893,
 0.0986186376325997,0.2338945504689447,0.2486998380587805,0.7343514092392771,
 0.8074923120008357,0.9234439732779881,0.23476424844815003,
 0.046194358397794844,0.5213746699528577,0.09114993796786541,
 0.03536826229903112,0.292687679065038,0.5625954328103239,0.5605533708161919,
 0.2252958540935952,0.39787697528344934,0.14426940731213445,
 0.5085821406123794,0.032330751992396656,0.4596252315461705,
 0.37552826790916805,0.7902925949105553,0.46232834158650915,
 0.2533876784323138,0.3707432468682951,0.07198583359451671,
 0.40077885669335445,0.6386409023625699,0.013572502305979173,
 0.39131570886623346,0.010380683595787588,0.9587420567002611,
 0.890862838970498,0.3331826945782471,0.6420523589612627,0.6528431095891817,
 0.5053709460772006,0.9902060818681284,0.2378434592928953,0.07251515848956835,
 0.010447481068973086,0.8455942698016187,0.29044329012166836,
 0.6508718229781373,0.16015901669095234,0.08800743313718618,
 0.36952030709757666,0.09068065668775416,0.19189548400541168,
 0.032020895882594824,0.2342843321741057,0.7066981230880424,
 0.43787783476917974,0.7083810463566474,0.8054369608589015,0.13271136493683,
 0.8093471005975654,0.6971942844502692,0.5354270360409897,0.12972975408840992,
 0.5568935730144738,0.890071726234399,0.9856222615233423,0.4502619337958401,
 0.6038578979176745,0.18822828576276185,0.09932015346131717,
 0.4868124571741772,0.6320153927127014,0.23201857769748246,0.7263683484906623,
 0.17752145536343233,0.37850845570148184,0.14019797563486558,
 0.6595187565267682,0.5491164069133697,0.992918979999656,0.22696873660208938,
 0.5987219632925785,0.4529335026209429,0.7899473338640027,0.48461324191009836,
 0.953441064184111,0.2703862810313341,0.5262088197689476,0.8179970240561801,
 0.662093693685164,0.27429861949317436,0.3694648463215693,0.5726302883823857,
 0.3254903015251899,0.6885338857912883,0.8477806811597873,0.27303088304280587,
 0.8917463911801249,0.9940146057834467,0.6900648034044545,0.7275431453469472,
 0.7019087793202495,0.07975661834929548,0.29833768768360824,
 0.6155222709059015,0.3805705583476343,0.9230828711397827,0.4103984070086778,
 0.8913403353846918,0.5234662054182274,0.408610034820877,0.5954042573132653,
 0.7909742070521721,0.054661209107730424,0.8088446240620595,
 0.8228918991427512,0.40671544478078103,0.12803170921261797,
 0.8219055346680012,0.77755251453589,0.596714977374261,0.6665378896810351,
 0.507920061716605,0.8318049952509773,0.5585615224289951,0.3285307840642252,
 0.4364891740278375,0.5966170344381995,0.3110049103251351,0.8905542293028965,
 0.394543761287939,0.43689733639193873,0.5553351150062285,0.5543570730713235,
 0.24098834355470344,0.7042722597397068,0.040630996790871454,
 0.3071703580597209,0.7438725089752356,0.6690030930673678,0.7912873488088266,
 0.03831330432846014,0.8544675053974651,0.5126523096611362,
 0.23247868750996759,0.6103887757433148,0.12912405216765066,
 0.6200107595297325,0.5942661277784996,0.8852841766102455,0.127124002210403,
 0.9634610506605397,0.685102350607348,0.29504449189606796,0.6504144563291965,
 0.9255864389466102,0.4467728573285029,0.7897267318641124,0.9031431190473562,
 0.9443296827223209,0.8093036794307988,0.7665564969171091,0.7003275094401918,
 0.661272910358891,0.9586175582860624,0.10687452684133836,0.24311979853374,
 0.06675037503759063,0.2758555897509749,0.7981450972535498,
 0.17471906591468822,0.2604799396838634,0.08729312397359978,
 0.5975588114085132,0.4024924933512686,0.6163749767137998,0.5347625549069501,
 0.5549581733620421,0.2582567900351931,0.6673273128006232,0.22252510527732672,
 0.32534466062172473,0.9963962600607077,0.6356036306711726,0.6023031476094446,
 0.7699140835302836,0.46124850035668186,0.07358529430963423,
 0.8999151621623116,0.5264416678479601,0.6856940833342202,0.1405166853900759,
 0.048115345546312915,0.6395231893376648,0.5763198780969121,
 0.13942292838463088,0.594314469367186,0.3668694547326592,0.5954204274529533,
 0.7321161553364602,0.9526002189744097,0.19998991491535867,
 0.31849499515009616,0.9829075277470242,0.2830246986329723,0.7459801815440473,
 0.527326261396849,0.560450897576499,0.8671455248432501,0.46351330443532324,
 0.2019330371665785,0.8420754508171323,0.6364430478841994,0.9773892370333828,
 0.6269243586972797,0.6892499291881474,0.9024006248494569,0.992584468373019,
 0.2832503228312542,0.9859605133797811,0.761996456914956,0.7786422071941502,
 0.6000480708988339,0.7820578910109324,0.5818533130166352,0.20693631710385063,
 0.5567883452029043,0.9033827168792098,0.6186539281521467,0.2867482401592405,
 0.8926158618288383,0.4830197640487194,0.1737136857201247,0.8878583333369576,
 0.47677743543899664,0.3894109374909709,0.23507326698840825,
 0.21123566716821363,0.05216982174912832,0.035834119807137466,
 0.24780366491822248,0.7652646911876972,0.8727912822544015,0.9412176417902064,
 0.20087042690390988,0.9649025867033302,0.35831900853151977,
 0.2774090469937158,0.32179296329047447,0.13245511035344149,
 0.41274561774074825,0.507026200731176,0.7460043016540672,0.5679617271906605,
 0.3448036982691409,0.08273884483890626,0.5687351886657905,0.2746781972581561,
 0.693649855800029,0.9644073149198504,0.27290093170925944,0.8317941415264559,
 0.13896090765818625,0.004123208027160041,0.7028484895616361,
 0.7942263885995171,0.4484709845654137,0.43748758660438236,
 0.16142595717268404,0.2526670204084297,0.6198330414615011,
 0.01754801936492445,0.020722030232209754,0.8204801824443189,
 0.08616378844569894,0.6519954670741126,0.6984289747171722,
 0.16678582760228045,0.709410301706167,0.844692311601431,0.07832014309458157,
 0.9132627599951899,0.06516360832332357,0.47974155579389954,
 0.49223355643773237,0.6815372767615135,0.5385999727601853,0.9359873588932304,
 0.7429763174995737,0.037272805545242926,0.5396943912635009,
 0.3536579317224755,0.8075394921993718,0.7100828005916519,0.8679135865192763,
 0.3477634674755866,0.051124396952029305,0.753914924806347,0.4437988526095399,
 0.8409140426135691,0.587613397880839,0.3856152026497448,0.4389230899822213,
 0.46271255430480196,0.6161326821592642,0.07900641196600122,
 0.27549162491519863,0.3607211547710697,0.8255052468764859,0.7748897050441416,
 0.3321464343187257,0.33871056182068093,0.09498809086665705,0.976384017049236,
 0.9165047530667978,0.28457021418644535,0.9553606762493677,0.8415330746796833,
 0.13279318153857855,0.7078045542294615,0.14748572808783833,
 0.2624497022920852,0.22285067993254892,0.3543362387976793,0.9260925580649503,
 0.46113274810428395,0.3616948020583206,0.004289013879039905,
 0.6098486279273089,0.9156450115595484,0.17751832036983628,
 0.30934706510844856,0.40042620239558935,0.9793161316156028,
 0.15656351279930614,0.39685539937037995,0.699220643056465,
 0.20120848888707443,0.527722002066389,0.2395449365797051,0.6690939937896654,
 0.9031946649884061,0.9949801110027768,0.8098369943116082,0.45153647550207643,
 0.15344510215353757,0.5918467824871784,0.07126942868326724,0.735920267237846,
 0.14033578318171014,0.2570851405308714,0.7934271517728555,0.9215710678532545,
 0.7940374308273092,0.6158213533442295,0.8076302371313098,0.5501659086304207,
 0.9165900166556575,0.5540938796781103,0.987625611262129,0.5477759733406775,
 0.9685240820251007,0.4111711599821133,0.2871602984220246,
 0.006436644467954555,0.3657700574089846,0.05476635511976358,
 0.9681001867080616,0.3922963680875807,0.7423943497552967,0.6751944368923591,
 0.7206642466424035,0.8748386281467642,0.48071017185367504,0.8468574257244473,
 0.8670019862741564,0.18357779149937825,0.34936186026923943,
 0.9509810429168282,0.34820026515132874,0.5244267951907642,0.9259293951356277,
 0.3416528853656442,0.5565566320112939,0.1789202283414172,0.640654046706842,
 0.9098928231274486,0.10331007256268931,0.05035981332598305,
 0.4386327754200511,0.5347494283287038,0.1911101476932261,0.8192108022637794,
 0.9954699484032137,0.18993357782173947,0.3844245125474144,0.6292314225251266,
 0.3725970108877791,0.040765399804974733,0.013417429848969764,
 0.07462338964458659,0.8803575022076757,0.6670353408963793,0.6783307819852147,
 0.38181300965730813,0.2342247812010968,0.3123588923868079,
 0.05412703693922083,0.4755130835258674,0.3536275963200962,0.2870233997372288,
 0.21585513686534297,0.8475231308153601,0.4737047986316919,
 0.42566429527618366,0.0712899054372127,0.7183814285071068,0.7824983391803715,
 0.45025309354383314,0.6839366292390644,0.5373024684940372,0.9793456000523947,
 0.6050143122463375,0.2780293876840989,0.35598022871501966,0.8348546329196348,
 0.48754831797121334,0.7132937238185485,0.9185453060244244,
 0.38845320999044985,0.16151248486645764,0.04718428461564994,
 0.799631082723085,0.6750124783734193,0.23789741195878378,0.14095297580568755,
 0.8471006190322505,0.7481771097402834,0.0062366863959860686,
 0.5072289662856947,0.7703107130751063,0.1817384763021016,0.15549440945975734,
 0.5690017267062508,0.4674794818301604,0.5876957001316895,0.8887106160425233,
 0.25752143841285713,0.5065969881077178,0.7906407897212062,
 0.06672423805361616,0.8659600849612595,0.3872605639713975,0.8944867825413592,
 0.32166121113126445,0.7498260708866455,0.288307411719146,
 0.019718495490524512,0.11014491853051067,0.11461032194792686,
 0.5171126254367182,0.6106755525041023,0.7516274542496092,0.177047522954071,
 0.3380114555365207,0.8362205504891391,0.36559280301039887,
 0.21728815661225165,0.021885592077325988,0.03006512074792278,
 0.47890510179150625,0.7456000150395383,0.5061645655295847,0.6058502791778516,
 0.34211445274761354,0.8913857335499871,0.0804236869040873,0.1884278490587219,
 0.43470815104341176,0.963483326630398,0.9327277859888952,0.1635649269468311,
 0.5008719402173039,0.0790197618418016,0.3437455628540209,0.14827330223219293,
 0.6462910031982172,0.9159107920289747,0.20337198542405,0.4400904670626442,
 0.9501933540617451,0.332488892465709,0.33531997727308516,0.9540397307140932,
 0.15209607377111356,0.6806441895726105,0.04423948043566739,
 0.7544747626009698,0.7797556164994648,0.808755592117207,0.5441312816069799,
 0.8899830743157496,0.6400281007391855,0.34699997326861176,0.2947748907280826,
 0.6831462788549463,0.48082894901106643,0.9305764329528791,0.7184362368514564,
 0.7939254046573544,0.76876014267616,0.20547875793224157,0.30335152589281067,
 0.7542171442001367,0.7910208005475192,0.9222554359908861,0.2428541706919407,
 0.08376373108367441,0.13704037646009037,0.7232169369742277,
 0.39672058555895495,0.7985729360233323,0.29349509459072587,
 0.6965162904294385,0.7815952057373523,0.08704387323625495,0.7181263181113646,
 0.059778364874855106,0.4286987434498556,0.24082165646764464,
 0.42668373860222075,0.06487329395501262,0.10081223250799598,
 0.9610208821511145,0.009975829633135258,0.5881933156355619,
 0.1432803486978711,0.6099904510356956,0.666115445297445,0.06172779523715366,
 0.34497468669250075,0.19377159085783546,0.8007732004124009,
 0.7327604693037476,0.09776398894442895,0.963871893587044,0.24584718788247417,
 0.27198264454309107,0.8326272293477911,0.7144375880833027,0.978324556459565,
 0.646782747495776,0.8866138378587414,0.377218536341395,0.8886735246948374,
 0.6896821483792309,0.4673326896584147,0.10241133035934924,0.2347044075678021,
 0.7658976161939492,0.25217678459880155,0.5930344708738097,0.4824993408155005,
 0.1714686326353263,0.6426277918716743,0.08971289546244132,0.8506109105889564,
 0.9767410914328738,0.2853206703371697,0.40789840042441716,
 0.44374992000789315,0.15027514538052011,0.7528946300906493,
 0.8076198291591776,0.933776967055995,0.3786051364756242,0.8658168727649092,
 0.4915291272279664,0.752002289457618,0.3524916891380213,0.5560437908189876,
 0.7463866094835084,0.8736820103059014,0.6716900823698403,0.2263975961739586,
 0.5448851575964562,0.8733942016924919,0.30736395797523464,
 0.03280474158410607,0.5396727601131451,0.4231327796271933,0.2570995442875321,
 0.5240246178129103,0.2459253436464477,0.8692077495712389,0.5347016504308431,
 0.13894999239520023,0.2571055931440983,0.8140565116310753,0.5725664800570053,
 0.9501114782577557,0.17071283295657136,0.7993021170201311,0.2531505812869017,
 0.8756863306203672,0.05713767271921755,0.6383316916868934,0.0639619387300413,
 0.4482191416023644,0.2194755086827983,0.4960424308256097,0.5071692551513348,
 0.4379551431325681,0.2350810225224108,0.11851712244054213,0.667229511190119,
 0.00402398736430587,0.8068862150239755,0.5720552211418021,
 0.35658453886204144,0.22041566582232297,0.6901343501664128,
 0.38256217189213837,0.9882827395493303,0.5512000462172908,
 0.20529566916765984,0.47155751751260944,0.4685258371856653,
 0.8958978168168521,0.6655363834832914,0.934682690184085,0.31447653251177154,
 0.6206242489680049,0.4315968222965806,0.5210462551717032,0.6645298959045536,
 0.9663463510647263,0.20787089371475143,0.03980570494962832,
 0.34815437888367107,0.3096030728832154,0.8227444141616999,0.8553569017720991,
 0.9658156107064175,0.950697215789102,0.009187725788212031,
 0.26477669733874665,0.3133318671409656,0.796430466012635,0.256583892566254,
 0.0962139072959991,0.8017585716266045,0.8518993515636926,0.1346134561483706,
 0.8637096532438564,0.5285244847477639,0.30426732321755523,
 0.37476745744998574,0.386689736494776,0.1620953865235688,0.6821888470933307,
 0.11151433823272605,0.6592114095047741,0.016571389302490913,
 0.6925573226305699,0.7947705894084383,0.7326405589062662,0.5761151782133553,
 0.16499749256853136,0.1506774895340064,0.8399206120760967,0.4074057913430774,
 0.49204325326666454,0.34130243802937477,0.4750347494685836,
 0.9851751454706882,0.2954070398689814,0.7676514473824294,0.08872055412285529,
 0.2502672163862232,0.36056027064243124,0.14335411937482245,
 0.29716160965832117,0.3058159688433856,0.6958289871121816,0.4941388937290121,
 0.2670142633272218,0.7985790444856516,0.4929568599386719,0.4931100856796419,
 0.33706126102230705,0.6790288083461402,0.835579319672327,0.12142185624055069,
 0.3959859429233473,0.5639710653319175,0.4082904730214503,0.9840752489699783,
 0.005582301451500404,0.6690846059116851,0.08369169784820418,
 0.6994903433156109,0.408476609601536,0.5543810159324419,0.6445301359708319,
 0.787459408775998,0.8282987570495247,0.5534909896714508,0.3890626602017082,
 0.11148773216065244,0.9065688457367264,0.46958188804607015,0.689525341156432,
 0.5245628399021275,0.10420634312198518,0.1457117912117144,0.3199398846738659,
 0.6320614233867463,0.04206960635476986,0.958144441014343,0.2354941068671793,
 0.7126578842721136,0.4981234368248886,0.37134687115500986,0.286283234849135,
 0.466093390355353,0.4985206685138921,0.9513387338752052,0.7163485440570357,
 0.1941991084366702,0.8129375679864522,0.30349671151857127,0.1872770775548478,
 0.839423116479345,0.7897722812300736,0.9840647696784419,0.38613996255467287,
 0.6907889870680953,0.4100624004840312,0.9715902617959628,0.8714156032588203,
 0.527618077416634,0.49754205029692533,0.463258718682678,0.9973988809571421,
 0.7047104754463704,0.4650059771602042,0.1970186433539327,0.5637896535685184,
 0.8125826537316112,0.9578336825548361,0.5370660642677078,0.4020624695419378,
 0.30830009107717793,0.94240361951492,0.8947060710662926,0.5198048798943431,
 0.9478321666415709,0.013571940227507184,0.24078182801828207,
 0.03501288228020838,0.962623170958334,0.9611789208163628,0.9043251314485292,
 0.8635854753413393,0.13839310786239944,0.8543247891984722,0.9532525029908313,
 0.9359347643108173,0.33881844372514014,0.28812749646837854,
 0.3868815949246657,0.26740393158246756,0.17643663549196564,
 0.5297305710144812,0.6698438993240959,0.15708395519370277,0.6550747894964228,
 0.34762055084426646,0.3027467946983087,0.12306569072236795,
 0.4753033248310724,0.9003936406495765,0.4316839745862646,0.4939572273282219,
 0.03726797538550852,0.8198073984039376,0.7967025526455007,0.2079185405504932,
 0.38376400245585074,0.8063422542836985,0.5500318149137114,
 0.40252642877523803,0.37570504441593566,0.48001849104801153,
 0.36135366875854247,0.6565608343795166,0.7593696145608315,0.5166401112136628,
 0.8918151118245738,0.9615729653543572,0.1616909678367663,
 0.030918122648162738,0.4915656127083404,0.17197556092342814,
 0.06614683814713274,0.5385550274599397,0.36707771442827075,
 0.1927253031327606,0.740565428779104,0.9905449957635813,0.5842673167042531,
 0.8100528370150938,0.31666215348414106,0.686994787946476,0.6340519785380087,
 0.9156756853506302,0.6797328366345942,0.37441987355529827,
 0.03627351407294377,0.9853104346155518,0.7421372919821239,0.7407960526035043,
 0.6328687025865011,0.46111301018291584,0.966116228787181,0.4602129597703989,
 0.29129719268594434,0.0728371853100066,0.2939793763280707,0.5726547472684,
 0.09723935779490012,0.5874527518738795,0.5514555584987348,
 0.13655620110032096,0.050003017234244385,0.48042564578861324,
 0.9867009277914381,0.6039675882275911,0.7986637407821454,0.19254727079580758,
 0.2533889621822464,0.5639910094669298,0.5155415119756687,0.03608819319051204,
 0.15746867356977923,0.28662093198927086,0.6485167630561164,
 0.7360393502135271,0.49426406790396193,0.885285100096814,0.6494388355618971,
 0.17799139051273027,0.38825568061829474,0.2540170752186819,
 0.5369528604469431,0.3263231081748005,0.47062767315251575,0.9597439221375011,
 0.608245759781348,0.8415391076631381,0.18240786680717802,0.9818510948016963,
 0.6768500397001912,0.49105405392044654,0.7690326284806126,0.6764349908663732,
 0.8178524185575554,0.13372997421004373,0.6508365577021364,0.1585838115449143,
 0.15836096835888636,0.13670045439331424,0.44576066463988073,
 0.8254287082925797,0.721493885381335,0.20518889242945637,0.5737956901174605,
 0.45833672628684163,0.3531833277513794,0.0846820602948658,0.81426140601991,
 0.2578251353244646,0.8490358175485545,0.5459534063649675,0.9735246147592704,
 0.49714913003596994,0.6435018195181831,0.3649830736742732,0.2377766258712738,
 0.8129664534534242,0.9011302880808603,0.524976445036746,0.7194645123948961,
 0.25933254020803775,0.4356255290570302,1.4343687247497794e-4,
 0.4570033029952669,0.8755486810361939,0.69406304012799,0.46913935715710753,
 0.658683081742513,0.12421020390041437,0.7041151817797505,0.2991810928421311,
 0.5170263815822035,0.8413409111223366,0.3676062954328979,0.12605049211519914,
 0.49608508777488924,0.058318124176536035,0.3960327635109333,
 0.2346083320289447,0.8007576025923898,0.0628831776791069,0.7144673367293921,
 0.9722540680062308,0.46289409704692863,0.7943427976887423,0.808701716437767,
 0.6672640908265322,0.6401361606111002,0.06999008686145769,0.3056554475165886,
 0.5199893398726657,0.7091378747916337,0.30427571647121243,
 0.28082184846564584,0.2725428413647869,0.009050871951660766,
 0.2013132239035348,0.5155403381704042,0.4233759352456994,0.27355228097933404,
 0.9665302325439808,0.8358556869314953,0.5873315136357162,0.6026457429862735,
 0.6868529490601948,0.6266700185166153,0.9754446190487661,0.9844779122068551,
 0.37822273773740833,0.36374199221501136,0.051154933619163456,
 0.8948062288526624,0.20794356524799396,0.46229917963753164,
 0.8403768629876182,0.49881158975225726,0.6563435761273467,
 0.44993984197419146,0.22725347816393182,0.19085317352712905,
 0.6292479038197277,0.18437375061728978,0.11854461260307347,
 0.28329040905475567,0.41847899751285544,0.6067587242958473,
 0.9167243395512413,0.7406532033822903,0.9534289543563701,0.520447823001374,
 6.2533059495673e-4,0.3640667936154669,0.8806814377171405,0.08963358214455952,
 0.9256973842323557,0.693848673448641,0.21436290751529374,0.8819444082083867,
 0.6998961984800914,0.2538915895618845,0.08598426617886235,0.3578084430870615,
 0.5343171243665445,0.6818286878678782,0.8433227909218377,0.7133578487174639,
 0.32065252894611285,0.11507496328682121,0.15382981651212058,
 0.40120465987151066,0.7656486698291369,0.49861069150017,0.06227299097041372,
 0.665847982372583,0.9639107109817224,0.4637566742812944,0.4332486662726832,
 0.6116150008483818,0.29874346631007964,0.6456046177819691,0.7059304170882021,
 0.600955198775158,0.4285289604882123,0.48846625206295924,0.04692714734018155,
 0.3998293975779422,0.4440257911146803,0.48685387183776885,0.1419906527182765,
 0.07437767096695602,0.3350659441012611,0.08690926547538347,
 0.7727213742822117,0.9286074551370644,0.6402655407605561,0.1598031277849291,
 0.6335904373049923,0.3835238350261597,0.03278094338936999,
 0.39861754091489554,0.6954095262095811,0.2654633989499706,
 0.45974169142915655,0.42555041919137815,0.7969420626310049,
 0.36493892189362065,0.020743786779366724,0.5142779464802139,
 0.1064435459554468,0.25491656880877056,0.21099402198957729,
 0.30400653867593785,0.2383616061845819,0.21506969886237226,
 0.2032822307594866,0.5826995426592984,0.5258898063764198,0.9966835727414559,
 0.0394505499757275,0.009234407219456453,0.9726497838333381,
 0.18921843200067923,0.3585601463485657,0.23094644884186533,
 0.2801970040983922,0.6777871466097622,0.3394799489166853,0.3631907360619834,
 0.7081500079636213,0.2746855155863821,0.33017712857464876,0.3849810902968611,
 0.9266184843563656,0.5494472564420566,0.6067911421927855,0.3616499739214426,
 0.6296683247632634,0.6642083168187415,0.958570725617101,0.07598060809775642,
 0.036355004215392506,0.4091492440985045,0.805397389923637,0.6696710254692494,
 0.4145752344740117,0.09105394497230879,0.5146947217538991,0.7328317370739847,
 0.052159627806781295,0.07996060346970446,0.3268757499620878,
 0.9115510609612689,0.707727827910973,0.18114749818384168,0.9925576018803173,
 0.9249553939452849,0.3274713609722485,0.2661455902609303,0.8258834139354734,
 0.506315802387038,0.8610053530251545,0.617307348561284,0.2190656203159287,
 0.15079829161580682,0.11574421823072778,0.025411122620717097,
 0.5007855355414128,0.7508671874836156,0.4643241164860633,0.47138740246692556,
 0.9053809819200191,0.9738660563797455,0.8565786828421826,0.27370939138997596,
 0.04564069976228469,0.9397970171832102,0.7581553271303647,0.5890909365821511,
 0.5437340202719806,0.8791552511705557,0.7537121713879482,0.3504245595396933,
 0.8445046420474835,0.644258686378206,0.3354772822788208,0.57469490163827,
 0.04477211933570202,0.38697834493783434,0.5161799677607775,
 0.32886011804927373,0.27783195130259986,0.5223134694681025,
 0.018241500466273708,0.27274711415207553,0.08697223947998134,
 0.4675595986981085,0.1372855993988611,0.023717512113894168,
 0.3204401814013783,0.2853533909883832,0.6279663522756135,0.5040973703146697,
 0.04603853827798898,0.6722840764677247,0.5169638023351736,0.8461316142029149,
 0.44622637255030817,0.7795538548923151,0.4559447440098543,0.8102566352774938,
 0.7189670642793192,0.24957985634531032,0.6282220820795789,0.9821272304348072,
 0.42452046688399747,0.5801168813616937,0.8447508708670372,0.7930309426483337,
 0.23649334334476146,0.8115492321836923,0.724372185675215,0.5480790411021459,
 0.6342572883182989,0.31209720989779566,0.08491795478263275,
 0.8138543896843666,0.42166459034563564,0.7146526718131846,
 0.09133861397772003,0.7521343500798137,0.9482290636230885,0.2461018917191442,
 0.5921349844878261,0.37279418827896227,0.08017537337656733,
 0.31194220533531014,0.4664779000034017,0.22232532795073767,
 0.19782429245654676,0.19350024475885275,0.6606418505632132,
 0.0496904101783987,0.1265541995627384,0.570722128873385,0.5007267570443407,
 0.5559660173299663,0.7215028602253644,0.5515776706889585,0.6010877164212072,
 0.3030172191198689,0.20217330192354632,0.49998406079545554,
 0.5437031925863722,0.25777872976731075,0.14164675982290742,
 0.18231593202399043,0.6660151912884016,0.8591387868083367,0.8122434172438006,
 0.9824966892103512,0.7050036615490852,0.2389159539540383,0.05916849548890757,
 0.5566475249106582,0.1617383139285654,0.6168414310217543,0.17352305834270687,
 0.4361508051538848,0.657307539319222,0.47295799343851586,0.1897331973421792,
 0.8892931323646517,0.20608765538872764,0.6340465241321076,
 0.028574500279325288,0.2607080347030105,0.2775317058075679,
 0.05349303279911122,0.3734366381226307,0.927287727892947,0.5838730374887908,
 0.7990437483171823,0.027088947543447706,0.8044145469396461,
 0.5548675783514972,0.7526318385391005,0.12402043886128267,0.1279764903321019,
 0.7204734413791678,0.5301134114622017,0.8765204235386787,0.2838207118449829,
 0.17096858347148625,0.18427228446920685,0.44701444918214606,
 0.19094000926603605,0.13530404596202494,0.2180565949962,0.7844517460668728,
 0.19686534000758482,0.017093792947444175,0.5557075268841635,
 0.8404381984185646,0.13237182663694935,0.7708800532460868,0.8904134382153849,
 0.5342631671701851,0.46738930411936574,0.1494977475235726,0.2839023789919187,
 0.6145503088318568,0.1106045468026543,0.15693724195566272,0.7836385259125643,
 0.8272618563239489,0.8190190730736515,0.11518022131658145,0.2987849421657415,
 0.16977659293076264,0.7247796596298083,0.6963964147806818,0.928350218381046,
 0.06755211932894889,0.7898274269029147,0.7724779091850862,0.8643005394128838,
 0.2867900037367431,0.7131390831237232,0.9196244518265774,0.4924331977993087,
 0.19940680837103053,0.979988994165454,0.2819329380514921,0.2792291633059527,
 0.9272652854384713,0.7776638441875202,0.6395500526328896,0.931881597981074,
 0.24431798364976398,0.3577419174641143,0.42139940902450834,
 0.09733274093290856,0.18839749463866384,0.5979577950741044,
 0.43303750106430416,0.6687726503624466,0.8422350979385088,0.9954345155093801,
 0.2826759189937921,0.11199526195507281,0.017695620989456762,
 0.18520325617538191,0.857065966546851,0.6946798065457517,0.41337518412261676,
 0.028153325274950514,0.48097627927063924,0.5690354551481506,
 0.4800320493816238,0.04207945926562906,0.9869507962011423,0.732321570920644,
 0.8039077504242715,0.5035122201664145,0.9900377799662741,0.3913221327207539,
 0.5658145435330664,0.9964134764235363,0.5765975814499376,0.08374002376457257,
 0.25740319962753444,0.5867777318654402,0.5535560031521648,0.6594322859636126,
 0.10440678836502615,0.35506549501170714,0.9466428656521573,
 0.0011230514167476224,0.1085236486067086,0.36241159268274026,
 0.41345011912890783,0.43945377371455163,0.35073703044809024,
 0.5367687492925584,0.23640860358450366,0.6457788991441281,0.1822375809913468,
 0.6877043543221731,0.102610492348985,0.07897146874888339,0.847864901260903,
 0.8751592623260205,0.7713348572597635,0.6127416199924756,0.8918656232111618,
 0.5407575607937724,0.43190327917274407,0.5122779768854456,0.6348872105138107,
 0.24732490155797104,0.4609526022899577,0.13413772950589864,
 0.0735082510854308,0.229936653973416,0.3653055791749197,0.1515856904641616,
 0.6879029407763985,0.4280293200916243,0.4786316489563607,0.40128312165444,
 0.6154526117699074,0.15145247402334783,0.558412195939251,0.396709847323802,
 0.8523317393754329,0.5208990684906081,0.8310634328158253,0.6891823673032522,
 0.482704740376126,0.5820675516689463,0.932814867314637,0.8587355684047089,
 0.7572672297696903,0.2650044086422356,0.015775109554298394,
 0.5884219073419302,0.35754236968109554,0.7567987703167753,0.9952945511889356,
 0.19774744998791927,0.7783212707497412,0.2713895251219277,0.7620044789189663,
 0.17623647195059555,0.9875245253718559,0.5046403595265151,0.5436561659187955,
 0.29669764034359036,0.21922841011443828,0.7591956551826893,
 0.9738219511233235,0.29833851413975765,0.1739538714884018,0.287643651832696,
 0.38990092111956454,0.14491182764452726,0.6964225128378219,
 0.7552721659848176,0.06913350845905586,0.49287348203101256,
 0.8098165631320444,0.7437100444226132,0.21299597504771384,
 0.16966757968926094,0.1575360705224741,0.2403779849690516,
 0.34458990101078846,0.21804390681795516,0.14115881695077992,
 0.21649842617779502,0.9499461715999911,0.9032392020774876,0.9171778160570896,
 0.17926670466145578,0.4195012882505952,0.9808511365279025,
 0.34785907508803404,0.06696179411969405,0.8961661197636508,
 0.07232154648124722,0.5977845521521197,0.5327511194835158,0.9754131924409951,
 0.11463979587073592,0.4447505257299371,0.6120258574716368,0.511273272469345,
 0.18399983042916024,0.5429517103450188,0.5220367964980285,0.7918672505611374,
 0.41553068095496937,0.49995846238447217,0.24577154946688817,
 0.4757521854546004,0.24668646036809927,0.22454186566160317,0.520557245918352,
 0.9282948569076627,0.5413203383791925,0.5243076084859024,0.2932984138804886,
 0.39455640571405115,0.3867380262172855,0.0017413024951060674,
 0.8797052503383178,0.8946182637580303,0.35896687319101495,0.4012816272267634,
 0.13041203523495726,9.559404592067722e-4,0.24404750942515252,
 0.8502923218193693,0.5889057521163662,0.14088155874432584,
 0.03730894193011469,0.7934918655998233,0.2619081138931182,0.9542400555461075,
 0.46507199878621275,0.7374346301116842,0.6860462455251549,0.3288285594650119,
 0.8348186025901123,0.2615550807314754,0.5003964578997577,0.46627987768300005,
 0.06693088277972048,0.783343675771699,0.04864680726246384,0.8496421506665028,
 0.3084721936406798,0.17584062353152285,0.6575017563426095,0.6859082566374971,
 0.39970884147228325,0.5055274521502247,0.16830679672527493,
 0.22707133873428642,0.10953325385415835,0.9307685059495769,
 0.6561519847840647,0.1702644566931253,0.1612725706533228,0.4762020382114125,
 0.7227727391601044,0.8981674861360437,0.4049879120896005,0.393389143919308,
 0.9205404198571897,0.2929171390075064,0.6391016360943444,0.5975329427268398,
 0.6633335499555503,0.917794005777716,0.822744460996155,0.9465988320090181,
 0.152500887426128,0.2687138840968387,0.38207636764486685,0.6096476781968995,
 0.54344050490667,0.6846179230746264,0.2800507750919481,0.6404276043782294,
 0.42356901905806876,0.2663236205429348,0.8928032487710693,0.4355431707365115,
 0.9475232465899198,0.8644722178488162,0.8034895779918725,0.34766573380372834,
 0.9473337583617151,0.4625701880222046,0.7986184711510531,0.6804269693303533,
 0.7278518564233399,0.3051708056961452,0.40602952253688784,0.3318874029782558,
 0.05232630838999164,0.26361516959865683,0.7270633291826656,
 0.19755743988130314,0.37045689316298713,0.8172914655994685,
 0.08653233785409742,0.8386437193123516,0.7184645450657776,0.5614205605537866,
 0.23416776113074234,0.7620263432107135,0.4382462456162064,
 0.18039043784913378,0.3804521197040015,0.6351574358476417,
 0.29394975398282486,0.1119919404252041,0.2289595277748695,0.8872556892326482,
 0.8430202482262046,0.06777539166915758,0.31218919416605195,
 0.21964821077710128,0.33025728760654416,0.9202866496672997,
 0.9407283147693972,0.4987125364395353,0.45095904273197207,0.7598080286929496,
 0.7731162325347515,0.9321421961018078,0.030097264686231306,
 0.6644758426288855,0.8360991387268587,0.7927753991450754,0.47059662089197696,
 0.29317540097612027,0.5804312005594904,0.12107814546523499,
 0.5571860408986654,0.6560039500302073,0.747863528282658,0.14402555405738904,
 0.06924643222286231,0.26079879037063325,0.32476892710735217,
 0.6031563039190498,0.4852451873485143,0.09200392029040372,0.9480627803965369,
 0.27739055750609953,0.14724086132082548,0.833819705759158,0.7167363016273921,
 0.2019160086699504,0.9773735546817476,0.14666756427325778,0.6584192152894873,
 0.05965633065992426,0.7515351930086547,0.45450978839412914,
 0.9119871681148844,0.2654983204252932,0.217240790745302,0.6915805120584367,
 0.6588154435101007,0.8329380843727765,0.8689302300823591,0.5361680535196118,
 0.7352522194758945,0.41204993609363216,0.08777598622503646,
 0.19975568446263747,0.8783350230764939,0.005714847497413911,
 0.6807434361889005,0.5688932232407172,0.914080969390861,0.5545928700984969,
 0.5422028426475962,0.7973931300450735,0.5985774405391371,0.9778098360587077,
 0.04447114344114933,0.3535916360715652,0.007582377659804518,
 0.34522636232101145,0.5606779251021211,0.7453359391164877,
 0.18285900446087588,0.5761461810180587,0.30342384403157907,
 0.7924579747576005,0.7312902641906738,0.43088714026588426,
 0.05171947432354507,0.05382451212304795,0.1533896613815272,
 0.08621850329933678,0.9498003381952405,0.4056862932943197,0.6445934512916819,
 0.010362908884234079,0.7622475235948855,0.556548930642966,0.7182292655372285,
 0.8794456489280655,0.7376486144738122,0.32759425454681346,0.5306081811043697,
 0.7353887742882139,0.4912338566028983,0.019675700709987698,
 0.7154141133225738,0.8957595063571053,0.8675952613059975,0.8644620290772231,
 0.08345272973726114,0.9690184481462268,0.010102935351382958,
 0.3883296370488081,0.18337951885802972,0.2658587626218454,
 0.10839659921964029,0.09480766625785864,0.20361372028851987,
 0.27018721914983823,0.25646710862429445,0.6118747253159472,
 0.9978671834419809,0.8772603108690429,0.6385840322842445,0.344531679592077,
 0.13036020105241397,0.8017836815191459,0.34929284351880374,0.893922825203127,
 0.01915890237879969,0.36168041787819105,0.9318181271109823,
 0.5561235846055608,0.8768635937058279,0.6982062925333796,0.7464054218047278,
 0.9475327245035724,0.14946165666446742,0.6269359539903682,
 0.05795098072546756,0.23677859063812445,0.9144150556597483,
 0.40459249455059165,0.34011302509084373,0.06954165492415032,
 0.37143033238409495,0.03575971459314431,0.06315663937835791,
 0.42226164036585223,0.9284058931633207,0.1299444649053294,0.9581255632259013,
 0.5382167687230904,0.11617199245122611,0.24424136855222456,
 0.16109894462152574,0.3375447189500119,0.1516136818483771,0.4168981595400366,
 0.38366812675096507,0.5800777090328102,0.3003543511036868,0.9695199798424207,
 0.3915991466058587,0.1490314957578882,0.1542086711657492,0.9676680150299385,
 0.7623049130015171,4.721412185669793e-4,0.2498747504328377,
 0.6692576811078879,0.7308024624967143,0.9811029069214424,0.1134413150554292,
 0.26267421508156064,0.7517748163502348,0.7108060225337944,0.3277409395829153,
 0.6907275027124608,0.19675242416812944,0.20934745972850033,
 0.2735816882028519,0.7705806654396017,0.44946435962082654,0.2862197645439767,
 0.10996730947515276,0.49216170290189865,0.44668845976243543,
 0.7060254825619051,0.20913620031893387,0.5439073155450419,0.4755413002247678,
 0.5632835199920351,0.9494356875501898,0.07357853339176246,
 0.015361065786410166,0.5362139696014234,0.6749283370642507,
 0.8554559570439312,0.9331151208545813,0.973336188620066,5.332711547536562e-4,
 0.957882382902513,0.4706765737374491,0.932784177571759,0.5793084866744829,
 0.9843196491069898,0.1170329795933931,0.9918393637897874,0.30319002277554297,
 0.5049850486115873,0.7718178423693204,0.38333156483972797,0.7149534968067992,
 0.36441780681103875,0.6334765743050164,0.2247638602362707,
 0.14105756522862645,0.36676926173607105,0.03158715335875273,
 0.10028815249549039,0.2929434040318726,0.4793500842674301,
 0.004554990195769659,0.24898046331920642,0.5878546667708722,
 0.2963431221760965,0.8307938093924961,0.021912269159390263,0.963411720488724,
 0.9345455429569612,0.48561676092219275,0.4158894645012038,0.2518365965683327,
 0.04450852137106742,0.6228214115821623,0.5598556497648424,0.8172823778458506,
 0.4103248708270961,0.49348492674354694,0.1862215094142221,0.6031404150853699,
 0.03770241698735077,0.9091399189375741,0.2490046390763303,
 0.25996492910531843,0.87155731994421,0.6947912746684233,0.3423195556498093,
 0.28471947751444315,0.6620130264792061,0.46429743676174984,
 0.10971289600238499,0.9247495262507043,0.5983658092233076,0.7115931241452789,
 0.47063970825165613,0.5273252690020701,0.24032880077348107,
 0.21616906664201296,0.10267880025812814,0.46496703611494283,
 0.5305022422154752,0.254103481329375,0.23965636006948265,0.7336066159636003,
 0.8406706640379665,0.44688943429963335,0.621963764105055,0.4732862065355603,
 0.8075638192715053,0.1202064736832944,0.9326452837030135,0.5318355218843305,
 0.36337025828115144,0.9272856182546859,0.1573596003727188,0.8268605557812923,
 0.833862948375367,0.3100738969055611,0.5737183893256176,0.0203042478849621,
 0.8432499642941192,0.9919455762793696,0.12929481886741334,
 0.002812445456899404,0.9112657298087302,0.5473981139468549,
 0.8243545869944032,0.6763137101063275,0.01473607328399451,0.8852894639425294,
 0.07651109534091383,0.11705200622247813,0.40079687196681124,
 0.9045376087841838,0.47400991821311234,0.5833026947898045,0.5916181686261928,
 0.44130969280279086,0.7768966962083379,0.9529919637505586,0.6888747298647814,
 0.6745988836401707,0.975032980660835,0.17632068362593978,0.3468367518311327,
 0.9314082561576924,0.9974154110773472,0.6683274520123982,0.5202587009558275,
 0.33279101526234534,0.3981743906947214,0.36820767540383714,0.673488388838119,
 0.5213397803712916,0.6742766895801657,0.6051550088035915,0.7455416827513655,
 0.29165617420099954,0.6353839099958349,0.14437689657643926,
 0.2089999602023932,0.5395555210894618,0.453304427358204,0.4379075516532339,
 0.9406857368044709,0.366662465202121,0.516380009484042,0.6694596680534661,
 0.10971271109503733,0.8245504242441366,0.4784524494219553,
 0.15548804920761705,0.19198809266077754,0.19450204857552256,
 0.3615120128889534,0.5940039224930959,0.19160215240475498,0.5367926395725846,
 0.4505145451368415,0.02452715805159389,0.562305051713298,
 0.022784029122518423,0.1364263870550695,0.41699246428407233,
 0.30531549700009375,0.96958425802413,0.28914243804751905,0.062837819866436,
 0.6874538874620029,0.5009017563546139,0.602553377144238,0.7597375055797158,
 0.6524033001325542,0.2602351680444853,0.6102614587942395,
 0.026136261940937877,0.07755459917102592,0.9886537992494115,
 0.47903437143996164,0.381512947217133,0.03669849396884439,0.9774187486092695,
 0.9160009507367683,0.7094725416891627,0.6257692974965303,0.7273480031736801,
 0.9275264991237516,0.30330761276641127,0.5374090350452669,0.3444065650576451,
 0.8840930963333891,0.5954323122674573,0.3621078958510432,0.6153107750676361,
 0.7542627392433877,0.5474331673735021,0.691079796598534,0.736645156751844,
 0.32563764423618236,0.15534923331214745,0.21196288979287914,
 0.07628717134746099,0.3795244638666515,0.8401688049210331,0.74023696986657,
 0.4640415249108112,0.1837355290429069,0.5789008009552844,0.7089618605737903,
 0.5748496345736775,0.7735845676462906,0.10271493834621714,0.9103528356831502,
 0.9496972677813601,0.692771698969109,0.20065387078511976,0.9375031949720593,
 0.6651365026965904,0.4258398570545895,0.5163616203609179,0.2504899293681838,
 0.27296753106635996,0.8320663357957453,0.5198842175508736,0.4588103878298575,
 0.8553655265713783,0.3780778841931062,0.6424475097286041,0.5254031612027625,
 0.6290711929420686,0.792446952135868,0.914351814485678,0.526906677197995,
 0.14743497363546154,0.8492838604180823,0.35760630607976285,
 0.09625353391860048,0.44764063774186136,0.9772003166062428,
 0.6281012793789501,0.9950888443627673,0.8503234491220825,0.9447602187797415,
 0.18925667616955089,0.23370283669106617,0.7621039553329931,
 0.7624928357653765,0.5023011627391302,0.6387043358404044,0.03136104114759153,
 0.3509497129086081,0.3045193730404687,0.9211629842319781,0.6781177443696569,
 0.40384774424205494,0.9460841223726598,0.10375486287600033,0.940919964468841,
 0.4624019228670444,0.3268586684282828,0.41449066531484213,0.6542021104186522,
 0.6347623077595865,0.21491104471035938,0.28314699494950424,
 0.6376816646733832,0.16445181646534712,0.8080874287244787,0.3667131935055441,
 0.6460921001547051,0.8739405991926836,0.2755170863682922,0.7019339194449226,
 0.06831710717329098,0.35224978385946093,0.9344690515541306,
 0.29930846370336395,0.9874665609604762,0.12171294066769056,0.144836721885086,
 0.0349848335077767,0.2627694781507244,0.313721488105142,0.1479407724152697,
 0.5359755972979513,0.229628813485923,0.9686786137895854,0.24435219252094842,
 0.4940768529782127,0.7477971448792995,0.38842858099364475,0.7607649348499216,
 0.4262896231118042,0.9079001141456119,0.931750519648843,0.9186373608116658,
 0.2517438830277444,0.049218382244231695,0.6392960484996077,
 0.5478814844099757,0.9586982245567643,0.25138940596414994,0.9058194406648326,
 0.16452703096342125,0.17204959107711648,0.06507394333686922,
 0.999763881044147,0.8240009004685855,0.41678073477856126,0.16857846480976513,
 0.084227597304799,0.9709085361117966,0.6150240377567703,0.25453895037683905,
 0.5421847487533599,0.8156293829728852,0.9663843323444646,0.5202883444215279,
 0.6381373990681665,0.7898098456474626,0.8380633140307445,0.28158934010311176,
 0.1853962789847372,0.5690503529356243,0.6633414200133242,0.49535302171582796,
 0.30712813509114356,0.9418175784383539,0.059196016962108855,
 0.28829645488353306,0.3717280922709685,0.5426941906415995,
 0.12879242550052616,0.37368260626857097,0.20166819907578304,
 0.9804145880382367,0.08706218409631428,0.5144027609164312,0.5751727607170103,
 0.5580203501422003,0.7165754258127115,0.9144414664452261,0.8970205006049616,
 0.7847384593384339,0.2572260727658511,0.7131048773461098,0.7705537712803423,
 0.8131618230892734,0.045485510841946186,0.6005867281161915,
 0.29015007027992523,0.42742229974695345,0.8450539111901801,
 0.011483948494820373,0.060195456099389255,0.19592561324722835,
 0.3312342326986232,0.8901348522618595,0.10385829686483783,0.2145895817681316,
 0.2893673459875822,0.2745955872067253,0.5568850240754099,0.10782478879059465,
 0.36246287454646564,0.12062995301688162,0.5855628156108365,
 0.6340407712337913,0.31832969327688154,0.7075163871965141,0.2735584795904542,
 0.44651789617452053,0.06863739946877567,0.36448868026335257,
 0.3982508984526242,0.6310114897053928,0.49520180163378547,0.4949585412414774,
 0.8139584517263813,0.8014515187110792,0.1797816745168701,0.21894535670899928,
 0.9794139248145475,0.995622885316125,0.7871655466157891,0.2656660631913855,
 0.2529305556096295,0.8713728251238843,0.8049369768930834,0.07575267397183605,
 0.09778726165391638,0.4800444262494604,0.7471910415452342,0.5607576947098771,
 0.8417455651018395,0.023289944863294054,0.013273649871234805,
 0.10252829950900211,0.278147836739463,0.31477643734824323,0.4647603709204331,
 0.6738095237388085,0.5053138233464851,0.19635697266393815,0.6635349972940431,
 0.11817864142924006,0.1493780784388803,0.5404391027489779,
 0.35228202391070673,0.9131060690531054,0.880407391348106,0.08165096008370298,
 0.0928725879329877,0.6268686029433898,0.14641639947888152,
 0.26470417154196024,0.9299956684416212,0.3301950718547203,0.797966167427613,
 0.3579218090683939,0.18901801827539044,0.7974470071749679,0.712520669188003,
 0.6342775843902658,0.6865589144667693,0.11460347051486375,0.5412765709446304,
 0.6417574653004786,0.3294464851024691,0.4565161332743499,0.7695121143476237,
 0.5455521187356159,0.6083798976844953,0.6444178964606274,0.16676606311373554,
 0.5308186870551588,0.9745187786356321,0.6926518108339638,0.5307945332256288,
 0.9232566866899105,0.6277500836252898,0.776552493855605,0.9168570628320695,
 0.41977998913822645,0.6710705477391655,0.7889885077328931,0.7417565211243409,
 0.5654317255453728,0.46784717232250084,0.3862785515409807,0.7400432977915079,
 0.9454113423356046,0.46825429879619973,0.8859469614944878,
 0.16435230453560412,0.4606838234677767,0.21317483747511845,
 0.9315307327924274,0.931229362953587,0.3482411294648361,0.8951974152767268,
 0.07015452595091853,0.16561268343016655,0.8118713015977058,
 0.6954132932600219,0.8158752994757248,0.9987333165904546,0.8750203393936247,
 0.9259309780039254,0.358367904654316,0.8033926964493225,0.919867233268241,
 0.4890581467375499,0.9470969913926565,0.9022693822915474,0.7632593400811835,
 0.8680930993017449,0.7696639460050465,0.6608771319496164,0.48261314384488263,
 0.8052100751451805,0.21148673475472946,0.74583613764692,0.04314369026160192,
 0.9187846802509737,0.564567828557154,0.8229465740067151,0.8124908779342271,
 0.5622770333824335,0.8326280404580014,0.5747402177124741,0.6582966991289125,
 0.8273744412437801,0.05827356774265946,0.09867967079788764,
 0.2735886832855422,0.8364309399059344,0.3109261319123521,0.1725616191468986,
 0.8044066844994122,0.8656863307745032,0.46745038860372734,
 0.13273860339258114,0.44104161017528243,0.44805994239322766,
 0.13037622011393868,0.9796372493509266,0.08730267033018646,0.877314589506289,
 0.09373689585138845,0.3813817431087235,0.9812440209897967,0.4468003855568202,
 0.1583782207819292,0.028466514586395042,0.29696506070713014,
 0.7021914655626261,0.3605192686714237,0.3138898212368342,0.05251973686389089,
 0.5084834761135992,0.48383911391485035,0.7655940755497843,0.6470242559666177,
 0.5663254156126758,0.35157352584995905,0.42699918403422,0.21757734335250334,
 0.02139146948218995,0.07563806885734925,0.3923714275144322,
 0.9929630191376954,0.10379816163746947,0.5744633735352942,0.5885241056123666,
 0.9661498866435436,0.882665632791025,0.7233226132576328,0.5524010231336831,
 0.23227860102248787,0.08766286222808062,0.017759072705599066,
 0.30179006757060967,0.37708307940836316,0.19609491017417902,
 0.07951270374627806,0.04359158171471378,0.7717088728839759,
 0.7369475690040321,0.042721781745152176,0.31922883267994395,
 0.9295694027530417,0.40257853928756593,0.4669241440850965,
 0.24561507292897167,0.7988078855984866,0.5366995112981875,0.9467492194797521,
 0.9377479242083567,0.059076764143030425,0.09066418244707242,
 0.6443910127718393,0.29447277461582777,0.07499889874384591,
 0.30053068908975444,0.5716775900703942,0.5233453318452007,
 0.46352247663221635,0.6813446391292453,0.15256325734969856,
 0.26921465706406356,0.9847107668666183,0.9911546821494801,0.9674435834920406,
 0.930850117959129,0.5570054485947552,0.27933863156304384,0.6809221252313907,
 0.26912549765845994,0.4715435499894062,0.7226540260184899,0.2917317153481731,
 0.8374708607846998,0.8067887258993701,0.23447447295903123,0.9806460893047995,
 0.9541605318084201,0.4975391806199947,0.37640786072436305,0.6099770059048314,
 0.6680130190810414,0.8062929454379736,0.29411648480331254,0.9657942940083901,
 0.16785021355849805,0.9376069087115556,0.9006557578787654,0.2055238040036952,
 0.606515886413994,0.6127917507447173,0.5325269580409127,0.7951837642013505,
 0.08784485325700608,0.4415472747548913,0.28875370601743555,
 0.16223757401981742,0.862309430183714,0.23439526373796138,0.3189742720294172,
 0.4014909585794343,0.14970570371597713,0.057540658383636756,
 0.6665821269601223,0.8260162277632965,0.5850429307444881,0.6848559841291851,
 0.9375754595432466,0.514145618796248,0.7682515938101544,0.23887413285305314,
 0.06055259052132056,0.6920145678263319,0.2565218446930132,0.7926497113175984,
 0.09829412077090294,0.598556535080132,0.2577237969136663,0.9611054540549533,
 0.031826266302084294,0.6950685644745057,0.6166669776345636,
 0.17995915787345473,0.3995960708925401,0.04520218426325684,0.618006143089221,
 0.3817493785790401,0.8554748905066637,0.7657315096017752,0.4329587457751134,
 0.5842726801137916,0.22280022986840853,0.7138204602882927,
 0.06384515304834693,0.4268873951539931,0.4075584201745347,
 0.41873051171302533,0.13857471852643066,0.7716088678080075,
 0.7635295080960613,0.1906207423049895,0.8831617099983896,0.07762296989484962,
 0.3833955639180381,0.9200651866582084,0.38582365100280613,0.3884127488033683,
 0.5690636964009306,0.727968023649392,0.28751917163056073,
 0.007134826162033647,0.22571445120390388,0.9496003681296741,
 0.6155482268220911,0.44294789724783745,0.09313556967521008,
 0.4316903418313025,0.1679174640420794,0.2633564212891746,0.6439376000926361,
 0.024266651716210674,0.7942963249019352,0.991844630220062,0.9770045647401786,
 0.6495374886396923,0.2399381220956831,0.2149289248292331,0.13286824747960135,
 0.03505500039943754,0.5608317799615598,0.1906929735897226,0.4573563633432236,
 0.999299400574878,0.9894929920703689,0.9065133575107261,0.8394202710038997,
 0.5824039461491785,0.7389497646496758,0.7835818532071284,0.1888182423162078,
 0.7650474807838844,0.9817266216701822,0.7323434166310508,0.19810468961862726,
 0.5535107525701612,0.3894760057194715,0.7077161237145275,0.6291584573032373,
 0.33458232004352995,0.6665489644203691,0.6305969825345878,0.742638270154563,
 0.7665952741047133,0.008455027190358844,0.7952974179205684,
 0.40175797782945566,0.2844445097813384,0.15188969272679897,
 0.1922850612606679,0.8501721727939117,0.5471423301618363,0.7521978730413261,
 0.25154832291358464,0.3004311729156254,0.41169114268848117,
 0.8963962661239109,0.29151566899355663,0.5794201322478981,0.7864915031733574,
 0.8921182092049549,0.3289488759572936,0.14193973016325412,0.9570986302688295,
 0.32953583148899723,0.1661154148902022,0.14864468991558044,
 0.8419831873878239,0.8625567820155401,0.9960393026989668,0.06582474123169113,
 0.1871172000218002,0.4347493366611015,0.8261644614328871,0.7172172603750241,
 0.931820586428051,0.007740252899047451,0.24067449778845862,
 0.5601045440644271,0.8011697057646326,0.2508878041936138,0.6379165249520263,
 0.20336237915045163,0.20310395330200015,0.9187197703391576,
 0.7585150765301375,0.230026057320885,0.35684333058773055,0.8845108422310771,
 0.7840199752765619,0.145848307302745,0.2983449345753727,0.9778127959523927,
 0.31829328652711564,0.15071660613273052,0.6410477831867127,
 0.7856640708881579,0.17799475689792732,0.7313520351024647,
 0.46106732212121615,0.32470981241307695,0.5716997580063801,
 0.47825938864346884,0.7285618011882613,0.890347417024691,0.02162548953633414,
 0.43325860162781726,0.05532067901619464,0.6390509551074719,
 0.11222806273659724,0.7035711224833955,0.289094718852243,0.7560357652835341,
 0.19343652334894024,0.3308278371898583,0.26277299800649145,
 0.30820873097075097,0.6795352817927401,0.07153314102109476,
 0.5634651530018557,0.34741176702397625,0.9181515453777546,
 0.08237355649465883,0.8879017101918072,0.4330779863598335,0.7119669625626618,
 0.312493110381283,0.8636464049075135,0.9570551758318271,0.05171938822124533,
 0.6987430751911534,0.5632190328650012,0.15260479176214548,0.2770204896247408,
 0.428851727976157,0.7419777067783884,0.766526300379624,0.7671198395929527,
 0.29313473746427454,0.24423495936865847,0.9404629702286924,
 0.34001147107351104,0.20059604687404342,0.4667786177441887,
 0.2421445899712249,0.40000552370119313,0.5577853495885235,
 0.24518426224006018,0.024784678026553708,0.15050294721910806,
 0.4136718591256193,0.7150148103366414,0.7227290376902626,0.17412294031811326,
 0.607616642428123,0.9700645458492314,0.884806906121228,0.692967756313088,
 0.4847437426690495,0.363067243805527,0.8680825084996753,0.13028239138498943,
 0.4383630509638181,0.9093882075138491,0.9636390572842048,0.3143934170645738,
 0.12859681618991037,0.2609749964567558,0.5393566734503927,0.658099337733289,
 0.542463229315827,0.38768860614161915,0.059098748116511834,
 0.18891214797345235,0.4042999178700635,0.6525587006513622,
 0.31906936077715153,0.9533118374314005,0.4745667256742627,0.3863828797068666,
 0.19476751767915002,0.32479771625050013,0.393778085580039,0.9543526860830092,
 0.6140167214573257,0.45025926452574627,0.4414393384592903,0.3452197612316752,
 0.9431483888154506,0.9479408699572976,0.885987055286657,0.4086639565774598,
 0.006370373002111518,0.6027140453636464,0.3319367706159526,
 0.9252423929887009,0.0044297172039662325,0.41094034914341093,
 0.005750113150064462,0.890952197873475,0.048366903691186236,
 0.27886053555412604,0.5083955881909898,0.11993990233203722,
 0.7181930759165153,0.41181294708895577,0.06375130117458727,
 0.9401295754863797,0.36027434908521727,0.3669779812362569,0.7493221610428981,
 0.1482108962238048,0.3842066957674133,0.7427120865940394,0.40931266345278106,
 0.25015328959495453,0.16164972001260947,0.313119995215428,
 0.014580257026189258,0.06787029207357786,0.30723767511077427,
 0.37921550118310954,0.34203400440673315,0.24582603198341502,
 0.44285999841209567,0.9934116520017408,0.252249263993349,0.5656714707215149,
 0.1659808737245667,0.6884911489646999,0.9239331369956086,0.2939699455768109,
 0.359435444395374,0.11529323546521109,0.004623187231998949,
 0.25043837385137757,0.029319283645333538,0.9904925316434465,
 0.3164394033862006,0.29953696720636347,0.37836393548588865,
 0.3944452426514776,0.34549286961862347,0.3035546092953517,
 0.31668017053481945,0.36097026042785163,0.8902319669193669,
 0.5717201894033609,0.21107729215231186,0.5908197598770323,0.5342888683250151,
 0.620531725251391,0.21826281194648312,0.9306796045189688,0.7175991182668517,
 0.19196317105788996,0.9207773839658719,0.17761582136279597,0.389009813863497,
 0.7459438263902024,0.9680490649196388,0.08586303005598417,
 0.39654421780073223,0.14176731794966013,0.05701659135372794,
 0.7685725123207485,0.7972275042784999,0.25559688419939164,
 0.17050601021698197,0.8626872434873017,0.42675696925747775,
 0.7827300463716773,0.24088223107013507,0.033455089688547046,
 0.26672461387246993,0.8023723149135701,0.9203722176259815,0.3114256109508965,
 0.42950036694168636,0.8987367945346713,0.6786651731979333,0.8073795594155949,
 0.19220987370207,0.2613903985854268,0.6850725094526061,0.6164908274214529,
 0.9465500216126801,0.6211920977442119,0.3505669822878549,0.6720214812854617,
 0.04200018760281099,0.09177973164620878,0.2205606750372937,
 0.05252565542149179,0.5341140469734785,0.6659330643302916,0.8593333348203567,
 0.38359308782095924,0.5651851963516896,0.6715637663095272,0.7772998871285559,
 0.8904922620156838,0.40261985747303786,0.031017438643470396,
 0.37417483526850903,0.07275643965607959,0.4742024678398986,
 0.6660420145267851,0.4872505637707989,0.7743565365780226,0.07739817571745344,
 0.04064687751400231,0.35025111445212886,0.3314621709817521,
 0.29269750917636916,0.8546022135123992,0.8860252942553208,
 0.27543813587848376,0.5565642586224736,0.9239638969920919,
 0.45285536749445776,0.7257230888037803,0.0811611922827763,
 0.15297679693049726,0.415081357920571,0.6840344567262874,0.46124647534630203,
 0.16838420874082272,0.5215420978560165,0.23242425319951998,
 0.36506616556015614,0.7380035589726707,0.3620043658492017,
 0.22694229061525828,0.25823825700052183,0.4355580751280369,
 0.2852963744773893,0.36477443350678174,0.478648709833443,0.2764137657178266,
 0.7886928360822513,0.6649801952873086,0.6235863295962272,0.9385958468252086,
 0.06499113868861595,0.7546298459070722,0.0367474847765068,0.904181490979025,
 0.1474126793546109,0.5410802700371872,0.8729636179939517,0.8565205332944593,
 0.4553705405120584,0.9513440012446245,0.3936850043142417,0.37364395507591375,
 0.5198881779309165,0.9978007979593788,0.7531669263382983,0.859903589637843,
 0.9327565971364143,0.3131018793943521,0.04365155166032719,0.7909381737435565,
 0.16063517191413124,0.8583662058300818,0.6135456059680592,0.4240022066042596,
 0.5900892463558689,0.4037683905012055,0.2914811388832337,0.13038605263283898,
 0.3821776105510537,0.15258211739291694,0.8407741794589074,0.2585646130581918,
 0.6467459899101086,0.02259461168349819,0.8954458288186854,
 0.061187743014495255,0.6378636391599064,0.11301969232426778,
 0.0527587933354533,0.6722829240055866,0.3181063980278759,0.24176008542006266,
 0.6931384123703134,0.5822237759123476,0.007305926324979728,
 0.22756043027038653,0.038465106417532646,0.26331560339119386,
 0.591232411972744,0.2024029522820363,0.9006238376944291,0.3372409740621952,
 0.8403034769430834,0.7828552403578862,0.8286926106035881,0.11542486444045374,
 0.9483784665943739,0.160475079542169,0.4694564538635482,0.7784000709632217,
 0.3773348951377453,0.33701259469181677,0.8219019514180625,0.8130980379626863,
 0.9388081184728491,0.9394412862185165,0.6558015479617051,0.08298947081952801,
 0.5247557149876948,0.8538897836911115,0.2646619763900754,0.7951163590162422,
 0.5138425740157829,0.2885904272192894,0.6756922307329996,0.9078420516652068,
 0.598077486856601,0.8475855173041629,0.07361108969336505,0.02056328234919713,
 0.8877683582481397,0.23194053602668174,0.45513172188517936,
 0.6703635687641044,0.5597563160019617,0.3507613559432634,0.24644498183978047,
 0.9200375112346906,0.43947570696869487,0.5586528693014039,
 0.26667762201987677,0.9147291207178841,0.2403365072740964,0.7355016661225742,
 0.9914402102755054,0.0032433448976567636,0.5820663096713132,
 0.21991248380452533,0.7171490768474607,0.5102172204227106,0.2361992877716329,
 0.3656577940911173,0.36127586789515354,0.6491623391292438,0.9961307706857246,
 0.206525313283765,0.6636766112959824,0.06281882429597319,0.5511180364246366,
 0.6305484719453625,0.6753393324825989,0.14857736671313893,0.3401630426903547,
 0.44696245911925403,0.041223089054687634,0.6017731464063771,
 0.40730532920226215,0.020098833391898774,0.89299209504236,
 0.07368765257160093,0.348318470946438,0.1362825898658271,6.01843813380798e-4,
 0.4315222927059279,0.8330819411503674,0.6387948279860949,0.42427280584494187,
 0.464780852710995,0.15156924591279108,0.4376207141282987,0.6587080578058062,
 0.6617260371487528,0.8872278022320284,0.693256482316074,0.7460348570257783,
 0.8243435026109266,0.058900746135764126,0.6275677421280724,0.742147852608098,
 0.2586078862416118,0.81603353218616,0.0151141768022921,0.09893641419469257,
 0.4503581293611568,0.9241358341163556,0.7040131051929266,0.46981843134683243,
 0.936536018607791,0.17288829398051686,0.211820970851885,0.9443485976226811,
 0.9034472206337479,0.9182250126945068,0.40191692682766256,0.90585612603055,
 0.7234843046406261,0.8339404789262882,0.3263871805738702,0.7782684960225458,
 0.21648817223620642,0.8866013799834532,0.868010369157539,0.4691797376865161,
 0.5275931192569134,0.8072016448415191,0.7113304672246312,0.8147233017156398,
 0.9635478120794537,0.4064409814749844,0.6449732247805721,0.37838575188208423,
 0.7378115811059847,0.4147257489182681,0.12033669960365867,0.8096847623744585,
 0.8843035548918167,0.7826604210732253,0.20866163611641397,0.8794153543767814,
 0.5510849082966157,0.6023389922600478,0.3859013358360106,0.29183627109259824,
 0.6829813796043659,0.9325734654947944,0.6268479915646447,0.4912129487342428,
 0.6004360500205876,0.002671185099466178,0.6191665707574993,
 0.11859995012537006,0.7832694738234232,0.9048439239915,0.009927381292972215,
 0.9383740654548769,0.31679726971549604,0.5198852524448284,
 0.01316694385323125,0.211432547473958,0.952027538739652,0.8195682248830288,
 0.1996018598992364,0.28878650238311776,0.06333425885965405,
 0.3184313576424064,0.19616404322141923,0.44057624629373826,
 0.22217933527949418,0.7075117586951483,0.401753658549422,0.3791262165125726,
 0.2111169664749678,0.08493685719339872,0.07321259897927845,
 0.7298224193943604,0.46747946020497166,0.6590803044564904,
 0.39379387835838886,0.2890717437906124,0.4139219408920797,
 0.23182960852361045,0.08069243413728466,0.3162278577310539,
 0.15404388649367595,0.23989431998545563,0.4632980630497723,
 0.5618768183295635,0.465317394156907,0.928263977476076,0.8168158300164954,
 0.11857422502608539,0.2778232726506905,0.35316196622575036,
 0.2208171445141699,0.06146085091589837,0.7942678627162205,0.6228143608324865,
 0.7516688055231094,0.5015087376812364,0.6392087493492888,0.7548249372617176,
 0.13034703922378432,0.6054508919963457,0.9890250376261391,
 0.016475804358227486,0.40372834147988257,0.30450632609528605,
 0.1590048145984826,0.8621382228327705,0.16467524477677697,0.614865768900185,
 0.025440680880876587,0.22838060749131817,0.8498120275328355,
 0.39914047978767153,0.5953119592855776,0.010986194786639314,
 0.7099108158568549,0.93794672470884,0.03693665193086715,0.5755346692938745,
 0.8258847546122602,0.8420293565908274,0.6714846916222846,0.1367787673235541,
 0.42662731372086293,0.021369071491141645,0.9080242880058281,
 0.08508751547461446,0.6334361647900845,0.7512955636965853,0.9504474448147737,
 0.5367396330171637,0.43745197324603224,0.9898585457641234,0.9080061508709406,
 0.2158194106126512,0.9973289821898148,0.8749621197901903,0.5415204383912621,
 0.19961517794952333,0.7276230839528403,0.33612213643657185,
 0.010887688406109097,0.24126639814121464,0.6694781410221913,
 0.9836605948891468,0.03289617598117334,0.35302639035894934,
 0.11336719708663445,0.7403766100747943,0.5753259783619807,0.9584257286656634,
 0.4717533902278307,0.9488814939424643,0.28469031883164775,0.7582625517924595,
 0.5004668404397942,0.5352286581101289,0.7018137852564289,0.09826558116423734,
 0.42924465091472674,0.5811822603292085,0.9411951321094105,0.6782201190311287,
 0.6539759474439697,0.43526575612944685,0.2291753235121745,0.8000339982620214,
 0.3926125245387885,0.7685497798352836,0.2496138085556192,0.6683928135006927,
 0.09958988259698587,0.5594230760355328,0.10944789790816711,
 0.5766520625837982,0.28532640968069234,0.11610469094789022,
 0.3091873934449688,0.9616150457993486,0.34824627379986584,0.8849094470274527,
 0.1136712140853251,0.521508501666452,0.2988746263344829,0.9534570163124187,
 0.20549064637563919,0.18122609516330768,0.06382673340123102,
 0.4183513368938133,0.0020998369537764683,0.22771434431222926,
 0.887950346895659,0.3946470855929817,0.578803342388652,0.025344742726331138,
 0.5282607380435954,0.026998525268616458,0.4980508322688386,0.86923429062304,
 0.5642165113856221,0.602808804721582,0.17266092786962983,0.7833292494939823,
 0.35988156792319015,0.5372346390949055,0.6671912200087118,0.7611250663737334,
 0.5727986459836542,0.10288821515870805,0.5029881074437279,
 0.37464903192488963,0.31015659898135184,0.4242967421451753,
 0.49250333269187574,0.7749765298669042,0.12467345715265798,
 0.5027653862384517,0.5747345320893642,0.15175165991213657,
 0.26812445010480457,0.21981535637850225,0.5400490561759678,
 0.5512596773301505,0.58772467956961,0.7073417436501258,0.8453197275914203,
 0.1340211513855908,0.8109726660273736,0.9947635290705591,
 0.023969083501006128,0.5951016582042296,0.47344864967717537,
 0.24505836653563373,0.7549353396689977,0.6106313412684833,0.332815903132879,
 0.402286849660457,0.8182631833984316,0.5564287860977473,0.32523687427607284,
 0.010961431239952302,0.042745172742301496,0.5723510627383425,
 0.2824983280928717,0.2967331150068949,0.8948289348038178,0.11316317303245926,
 0.08146719658871016,0.017738551207225672,0.7680913827847297,
 0.38530131493678377,0.2409718907095304,0.314356948706215,0.35930936233900124,
 0.7965173257395805,0.8512120010235998,0.9138122486400512,0.6713761951191721,
 0.500880193659166,0.6827883095898727,0.5505718600987338,0.8005572688376271,
 0.8193096774156589,0.9248195654318635,0.5824246929829513,0.7786749299592353,
 0.699732785663765,0.6915978559347167,0.998702214801335,0.6248767674891212,
 0.20166012890442142,0.6858750591169555,0.2882482408784497,0.6652612288654265,
 0.8477362391511295,0.30862214996568293,0.5287379390683357,
 0.18670459708265874,0.5673010558379751,0.33837708843776815,
 0.2167981988183244,0.015043695316847794,0.3473226393511233,0.909658410324586,
 0.1642389841901305,0.33418807940573514,0.7488851649684604,0.1483366717366048,
 0.8683312100594195,0.3293388326343215,0.2671316184648641,0.41753636544455286,
 0.6130814288514939,0.5257160441119503,0.6064195548190414,0.7897518694656123,
 0.5588164890588896,0.10836792579490317,0.5349467324214265,0.7521000650277426,
 0.23945845718603742,0.26271518462236154,0.4682153982632393,
 0.07941120570232685,0.4088984085548506,0.6086511661981627,0.9078995029004266,
 0.11020217661945153,0.033558935711783144,0.46044014698178326,
 0.31411174282639687,0.304897375265323,0.05757858837371754,
 0.21689945535215238,0.9022173000546317,0.874550874149429,0.9088937119078772,
 0.5539515140673271,0.49852120911860387,0.563019451107654,0.6577049005511441,
 0.9983709571474996,0.7407417891502178,0.2980583727103816,0.4305180848849942,
 0.32918837702014014,0.02996881387561634,0.5563367076389183,
 0.9567850984489763,0.6685064817016603,0.33130851668804007,0.534996240421653,
 0.013467317368138554,0.6113978123513905,0.12536592027868743,
 0.17148532839193853,0.5095407938051095,0.1270006014901499,0.8260560210202776,
 0.9738402356298127,0.956923104535798,0.8078446006572059,0.7971825066475086,
 0.38327419888715364,0.5820811774325343,0.3501777172082644,0.7191208795180718,
 0.7445161323502902,0.25046325818505877,0.27769912530149155,
 0.08209663816537771,0.3239801755732724,0.974249511191644,0.5935744156766273,
 0.7119648750644678,0.16999594094732828,0.7000327224233989,0.462846143984057,
 0.06369243655087131,0.7546406974187966,0.6279825904184058,0.7221414903313266,
 0.39792826302915596,0.3024841052335905,0.9424452217116115,0.737015764303438,
 0.040199142057102843,0.6862955547063865,0.7454784533410086,
 0.4416777743229957,0.1602554005749881,0.01553361414340515,0.1558974175043989,
 0.6832369180436927,0.5078310996080676,0.4436668594342892,0.503710227705002,
 0.09562354928596806,0.9158171691313134,0.13608637258053724,
 0.8799251336302182,0.4376279343809387,0.7683545952909604,0.5985059390835312,
 0.1539461147171205,0.12204158351571626,0.6428262948024046,
 0.25704250084275715,0.4321140328018641,0.06376250666783057,
 0.7652441310661255,0.20359780585373555,0.2762849995627966,0.4339584806600162,
 0.4915030119845737,0.7091889298344606,0.2617482300849523,0.6464964432661441,
 0.5678609405501931,0.4658795511870707,0.7962759389723696,0.8137048338156104,
 0.6457162410541393,0.05854459760980435,0.41010020610000775,
 0.6161874444332966,0.6001572841252282,0.878469962083924,0.5916756234161297,
 0.8128658447799407,0.011432156592240061,0.28741739916420106,
 0.49330646171720627,0.525431845330939,0.6159457285921385,0.09984601437433338,
 0.6958436853044658,0.9270776089862174,0.7747792273541891,0.12732831206347883,
 0.1895375227082664,0.8467459515755449,0.11896764426335116,0.2459315023814087,
 0.7277890949169599,0.1517532570524346,0.415365313031259,0.05895032269391132,
 0.297236852634158,0.6750640237581398,0.004008223692420931,0.7613674607885919,
 0.9228229196385037,0.314934313749463,0.7947435361017703,0.6643092171716926,
 0.3829233294728125,0.7389331800511948,0.6380898992279566,
 0.017835868067819005,0.3966415109231185,0.7344681035253684,
 0.0824271598329468,0.222421443984401,0.8934598602432905,0.9177542734983823,
 0.9291054318984254,0.34390957928868804,0.466500999133707,
 0.012328977461722679,0.7133259749480205,0.06846837836468578,
 0.7148205237563008,0.8577978400910279,0.2159141227541429,0.3337417604730595,
 0.6030306691233123,0.9508407687373925,0.5004109292898098,0.2844232195431552,
 0.988132350920358,0.982148325891785,0.481390023211379,0.042745094514739734,
 0.01385512115840759,0.7331254536244821,0.012291020743389391,
 0.5177956386899654,0.3746905717616671,0.6036407499761636,0.15590039269409695,
 0.779004670621674,0.8670925288246594,0.11534783907737112,
 0.014445864058360791,0.3861523835761854,0.9649934223100024,
 0.3943652306743831,0.6399181256813681,0.37054346293683194,
 0.12340171551971013,0.4624410051763007,0.448837867224932,0.20489868035921233,
 0.1614940659324483,0.5733435459439888,0.6350650721685975,0.947760252351412,
 0.9996116272897743,0.7149797115296285,0.8133291417576565,0.9034137710358928,
 0.20924834512450818,0.0657762623067959,0.9835628227021812,0.0674103670515731,
 0.40516274945933994,0.5030473730310652,0.19346801773791056,
 0.06630741110979876,0.03276375573036816,0.7377955804899898,0.820537917094545,
 0.12032353027530118,0.9509682110265849,0.41800720722546025,
 0.8177236883556296,0.14998215505489754,0.711446767136741,0.5043623427416734,
 0.6620939180086686,0.050019185355923534,0.25387574707101734,
 0.09495777079593515,0.5303314702995683,0.15620895236910215,0.637105137676019,
 0.07764279159917564,0.9308102152886901,0.7238462818522494,0.9976111734731177,
 0.7643010354055316,0.25353545282676115,0.06159931528015161,
 0.9754489199685654,0.24729232919403388,0.3477360124939779,0.1690571758122742,
 0.36030978877479325,0.5710587873365225,0.31686590413308235,
 0.4806769846242782,0.5172379932298881,0.18765105620174494,0.9679005281304176,
 0.7608853878317717,0.9702905689177246,0.9772884889342318,0.8673546046014642,
 0.6171878958816579,0.8274548743586461,0.47321927121900687,0.9609607453861448,
 0.5060879885178993,0.8563784794008178,0.6149078055799235,0.7062281930704251,
 0.5010180841110968,0.801806613967356,0.06029901449794939,0.4199577002467959,
 0.8333232075760547,0.6827947067346096,0.31075975358255226,
 0.45271569650314425,0.1900612286120804,0.12097228621397893,
 0.037939489843211494,0.4275004735205451,0.48152805247135444,
 0.715136895078597,0.11527312010638668,0.6295803800194071,0.9888509285832214,
 0.8750463302802979,0.6915700067175464,0.13659235042582307,0.2660303009440518,
 0.5530373144777085,0.16186502172454165,0.6254692726532184,0.3619178296082598,
 0.37006123799179147,0.2842638222625702,0.5652142907416643,0.7235116269613011,
 0.3555743003241909,0.21452628183522404,0.5730821917154669,
 0.10085442212740214,0.2272583750563123,0.40386075553610845,
 0.41150677830844473,0.38621451571703513,0.18906318758854423,
 0.7503251325788227,0.27889677024934956,0.2711590109505919,0.3744075288878951,
 0.9948523030764489,0.515754692799283,0.1985648512430519,0.11114253818112163,
 0.3518697473982363,0.9467227903508816,0.16514964747920047,0.9139544690338997,
 0.8511483088276166,0.29403902041332297,0.6747314813553605,0.5740965232081795,
 0.9231501605767051,0.8574725481404563,0.834975832144617,0.42500494087012386,
 0.9780845079663614,0.7839579577195892,0.9287997236751909,0.02778855211201514,
 0.21999534539410137,0.7729406459187618,0.8397179251651842,0.8760654705028085,
 0.6479595463244326,0.17985698591307941,0.40528876634538424,0.909058334831761,
 0.3306558683785663,0.9401017102259357,0.8906215704840061,0.3239482332614383,
 0.46584880921142124,0.2638643053194769,0.5034957011006033,0.8814175188347732,
 0.23982081069432803,0.9793036785316511,0.16142581273268042,
 0.19454393525847347,0.3241987002436625,0.7585578815435718,
 0.31618800276409786,0.4517495329664092,0.11633073598519983,
 0.48668452092753234,0.6995121739207345,0.7126706366705963,
 0.33838736117723445,0.4717222113160623,0.03398855737760709,
 0.18785864637627336,0.41318107019477235,0.026010219198611906,
 0.9768555613918105,0.19586159659624658,0.1552334588072598,0.9019952871282635,
 0.7378356240894834,0.42027745291570273,0.350140368453168,0.766503016726412,
 0.441071019856231,0.08707859231582671,0.03961573789317174,0.9058022461237032,
 0.5867042251670566,0.006245162309999608,0.48690585559421495,
 0.07808310892397374,0.5804926348626005,0.4354664908229372,0.7131015442150964,
 0.9072317802487585,0.6463591744350261,0.7078702793575429,0.8654352445087914,
 0.113484483408128,0.7674960001693191,0.8708174798536313,0.6725181500396293,
 0.23260443950041465,0.07571313058117923,0.34349687787284544,
 0.353949000271747,0.36130538101689147,0.06368134395698377,
 0.07826126189175509,0.41920376624211364,0.9042140829954669,
 0.30026129596638274,0.9924290204871437,0.16804518580921224,
 0.39676965967398403,0.23978494468472544,0.9170558682705641,
 0.9512345229451772,0.038188916410013096,0.5476081819104293,
 0.26291368224602085,0.02242034956066985,0.865452689453921,
 0.40905620481983285,0.930126251623124,0.8935435502248998,0.21565861220084603,
 0.16205434156866305,0.5745127239664494,0.24285963953874756,
 0.8354337914111446,0.8990800123486404,0.8492111287681057,0.1353392200029342,
 0.3742634919196982,0.09698452886424047,0.5372815025266285,0.3325544457030306,
 0.7185374193542851,0.31605337186363336,0.9713230622152148,
 0.21941738601740512,0.3913447720706731,0.2399130812758209,0.6271106995429464,
 0.7315816694252633,0.20772986290375006,0.0495772250522446,0.4387360062739273,
 0.6853730117827778,0.4312586118700996,0.22366060410111277,
 0.23304996432631764,0.3119125542309821,0.48690296284529944,0.847110891298283,
 0.7051416679105347,0.18778317827572488,0.29448243532540064,
 0.2846075154600075,0.6952449236994767,0.9549871073930604,0.6570709293389524,
 0.8934076420470745,0.5210876138665081,0.7709071435341981,0.5033896329314912,
 0.9624702087489614,0.6585881272417677,0.1352062085039677,0.31786218579342906,
 0.899076596864438,0.6910203855684676,0.6432798822008435,0.8970896894421158,
 0.39976181136270583,0.33713922689862574,0.6283682931127883,
 0.18243162312900696,0.8097368722640426,0.643593933110819,0.7077343257486618,
 0.977465264963949,0.7962539133914666,0.5002302947850693,0.2155346766619859,
 0.25078132611951187,0.30827178663859667,0.10355032698994182,
 0.826976483503368,0.5566123843348838,0.7003013797653141,0.6381609932632233,
 0.9158247539669963,0.8394830464596078,0.3805359353850555,0.2766392521190165,
 0.18628491952142812,0.5059134665717152,0.274811765999328,0.10501468480438614,
 0.9346066856320975,0.43794415068045356,0.4711037995443047,0.3849358419912099,
 0.2412119778593751,0.5762614785243844,0.7458968819921921,0.6736592290703669,
 0.32457266490105474,0.5630965056980419,0.6264292820084366,0.5207555661633042,
 0.11331046901986763,0.5526440178811645,0.8268236327968732,0.508844288282507,
 0.6241587138076968,0.9659809545057175,0.49666828935976637,0.8037510508879429,
 0.6000964486690111,0.427558396360751,0.3391644293795961,0.9786191531528121,
 0.44353070803841055,0.057271626627599215,0.5674830034017685,
 0.032417644600124085,0.5348467307462227,0.07333515683229386,
 0.19808666384559548,0.7235975338839783,0.5461678296223244,0.6191935826209645,
 0.7665480980717615,0.533308335351643,0.44907887969724525,0.42005715765908114,
 0.09462665820242255,0.7479039102561563,0.12822515393208456,
 0.14681236663590758,0.49337727086701666,0.24330638780639435,
 0.9449283025677966,0.1343815232098864,0.09661403723333561,0.6433576139359118,
 0.3207306217786877,0.1207021021739495,0.45094347124841416,0.347896571793212,
 0.6735369853056334,0.5396245580116573,0.7081141671738674,0.9491203972234665,
 0.35222215379588817,0.29433544125900357,0.42176568968383166,
 0.8799939997969075,0.041856664009854994,0.7049642002292735,
 0.24240396608125736,0.7640991428648102,0.1542786666964765,0.6406622622863196,
 0.684572063374721,0.6958311326777379,0.969977226526731,0.8034465102200347,
 0.03539712578350329,0.480059611620523,0.5409589847573275,0.970600671058943,
 0.9323638503126953,0.2803331241242485,0.8284794331578348,0.25977755274068715,
 0.3892765212736832,0.7147512410672374,0.08867161964671555,
 0.21220228558600995,0.7127033642419592,0.9121024227317791,
 0.031045430628173998,0.2337421937240598,0.4122616098110715,
 0.6664643590569558,0.22310852045504548,0.40391419639772796,
 0.6164684260089143,0.5810740241421415,0.19534907581340455,
 0.37416661643907245,0.4616735715114163,0.3058811332555824,0.944502590577928,
 0.6850080371160997,0.17145613122955317,0.7977252383510322,
 0.44692410972767893,0.8220035622819452,0.03677459803959193,
 0.8804166009927303,0.47006347531156345,0.23606980620498108,
 0.6620468909882956,0.7966006164197909,0.3230004673558703,0.7271685564588951,
 0.14334178409765863,0.09394790531620822,0.4717655250519135,
 0.5090216277170873,0.20738655597642697,0.025612446979033754,
 0.2656561021879229,0.042477985737212176,0.1865132000034343,
 0.01587674398091099,0.5897096131345456,0.6438593125416469,
 0.17899315425788587,0.6404309523148796,0.12040877270084682,
 0.8233311756782377,0.7430766107716503,0.37248877258198454,
 0.30613178020183507,0.6782890235489678,0.6491002831737308,0.408451062342288,
 0.7734708799755214,0.6234553427986382,0.9934493163367561,0.4321110949816718,
 0.08946768697527241,0.9032587918893592,0.12575382691669446,
 0.6072022959854267,0.661292263467383,0.16799921538017537,0.6671518059182282,
 0.7057390187044292,0.4585762231145867,0.7099680672203114,0.17724284874934515,
 0.5506967609869335,0.7988630271905857,0.6925388687231477,0.6369717073476812,
 0.5271928124798778,0.22600023418091175,0.37080631885571713,
 0.4409148747342081,0.44601076550372554,0.8555899178471105,0.1871023986418323,
 0.7966909860361999,0.11628603137539184,0.9454082010989685,
 0.03133706558255789,0.4589645514410998,0.2980581578970073,
 0.09166255121366818,0.8572643686061394,0.9780660228650281,
 0.43840843676116203,0.5389638167073759,0.9358070385752733,
 0.31708887916949235,0.15812216884426866,0.3090065729525735,
 0.7581364948404609,0.21699265169668192,0.990675627746361,0.42757733115165475,
 0.004289770894585931,0.3774055790980171,0.1894662290438609,
 0.0986968958995228,0.5358813845688287,0.6403765655226364,0.6423714112535701,
 0.8584831289723965,0.9087562256369827,0.5144467783382847,0.49751182973998187,
 0.14814066407975834,0.3573713315655589,0.6645825008184607,0.6635409862912826,
 0.7680684193418292,0.7857436455131587,0.8051238966790087,0.9936395796247867,
 0.7835386115968201,0.3826361834026383,0.9242112619455559,0.1922816229462182,
 0.31276071040537623,0.3511625998794402,0.6082628611867164,
 0.29520175039552954,0.7554605580155528,0.286076999310783,0.5310070914591609,
 0.19039877666643212,0.610549515462505,0.4899633632617153,0.9615972152320077,
 0.6515232719533629,0.48968563810542187,0.10566078429337189,
 0.15153610356376546,0.340111013677163,0.9565117214223715,0.23869293335118524,
 0.08599368694479337,0.8132661775399929,0.7872405242187374,0.8890087842183896,
 0.8430654900741443,0.22252425665806186,0.5087083095495857,0.9311862460672119,
 0.35504663223862487,0.4355442836010692,0.5372817512909903,
 0.053983517980364604,0.3748397843388681,0.8099498374093594,
 0.12709866850227802,0.20481300416671977,0.11974994845200904,
 0.39802412591466874,0.26754845326710575,0.1276736812959285,
 0.8995815312176696,0.6326223673937074,0.6099095823369552,0.9581208150611896,
 0.1441354429021433,0.7306933007312054,0.13173332069562316,
 0.13325502940188805,0.3470113079893732,0.770354477749788,0.9781698495668613,
 0.003280578129553069,0.4262806606207624,0.24996409315466772,
 0.6446335803053276,0.3711920590367175,0.922293829903643,0.40439401014382537,
 0.2811360006350261,0.2651807199916888,0.6432137757916473,0.47147571291048584,
 0.7111546470050161,0.22796621061000855,0.8998896916411301,0.8734094787899331,
 0.3212195165582721,0.28015567537783226,0.19752781486017967,
 0.3951070448305015,0.7489716783794563,0.571393317052497,0.9353167749431025,
 0.612289849852116,0.8562685347292028,0.022038216929932353,0.8236133811642312,
 0.7127173950990654,0.8158504026366522,0.3368820857491174,0.7394532748539047,
 0.7156175196091981,0.20449719874374617,0.9561464433341959,
 0.48577157533478343,0.13671151646376267,0.4257332248411587,
 0.09818933440203748,0.32410337098628994,0.7091254702326081,
 0.16200412881784565,0.7751815482316671,0.6793609127222557,
 0.48988428436846065,0.1752958208858738,0.8122474986601133,0.308966349183187,
 0.29960384819704156,0.2466381410505144,0.16653798346808624,
 0.03408246933755266,0.21221132106778284,0.16104357911534095,
 0.691408773422413,0.4360350400706263,0.016399299148684654,
 0.38173229915190876,0.3027477168748892,0.9886038797810649,
 0.00568586997576781,0.5307256685397707,0.9127217556614705,
 0.15601858783786682,0.32055860522166313,0.7223555460751678,
 0.5576904569718963,0.6292526580599452,0.12285406258096099,
 0.10291683632201643,0.49370067303279086,0.6646725349895583,
 0.8997418992532937,0.8882088929426276,0.7280987811666889,0.5525807507885987,
 0.8042433872393734,0.2297333051681114,0.2197923922611006,0.07791396743134982,
 0.5405340021858388,0.2262736294209795,0.2448885445075316,0.4420955454309339,
 0.23661188223841412,0.9968395646865182,0.9149889880667317,0.9584781470895467,
 0.2858357671470052,0.22208158747771145,0.5131629895599626,0.8675216855649295,
 0.5417529864832236,0.9260291121962843,0.6462881641769329,0.5343583394003906,
 0.491229540320126,0.486244235307006,0.902311014186264,0.3925527212335921,
 0.860207809531182,0.5161531856711805,0.23048180353842918,0.6635245646310065,
 0.6813986097473556,0.8223079494846413,0.12681856780063328,
 0.32163916666054304,0.006546820504292117,0.0821013486034915,
 0.6784660181945845,0.31208144833334805,0.4495066413077383,0.3791575639347795,
 0.2302005350562003,0.8540335586731606,0.019465599475774975,
 0.07234440443570389,0.06319224408147273,0.9179733470758887,
 0.43396524340555276,0.9430753506802019,0.6172665040942901,0.8402832854254338,
 0.7448228924457462,0.47749496821550075,0.5706545643671206,0.5329052325156595,
 0.25691605821342856,0.5363444853120712,0.3457725632340004,
 0.10914582938587225,0.2538021758230826,0.6692439750052465,0.4218768994667035,
 0.7919410909107447,0.40727563373903086,0.08641100633246646,
 0.05500326710353043,0.05210863230383578,0.9070961226270935,
 0.11642254309927458,0.23838139320699347,0.9539667499508309,
 0.25431232667185233,0.2038888661871593,0.2800910500636342,
 0.06848440051212024,0.44044851916767414,0.42643321337330387,
 0.36799745381901205,0.4844994451710002,0.6088987879049844,
 0.32066408070751407,0.8658479554065813,0.2666668802506328,0.279501576383977,
 0.9387960262064559,0.09534321793120015,0.5188770795041185,0.6041186717225355,
 0.8726056899872665,0.8928732410855331,0.5403013656738064,0.9759615873157559,
 0.9137136050882052,0.09449644659684953,0.08798189167339887,
 0.4679325213915989,0.4563584408234398,0.832510183093764,0.1609822821915443,
 0.2947847139717087,0.6837206318796857,0.6686218241746182,0.8828343051294985,
 0.47197406331795533,0.10360679781011772,0.3719948946107686,0.854854318869316,
 0.28903421714361666,0.8935028784422548,0.7173694075551851,0.2853462424568156,
 0.8565479596277845,0.2739714114713684,0.6340076730062199,0.13120293484244616,
 0.4945671687078975,0.969273289348561,0.24030205380419611,0.5311275784098886,
 0.6249364508451124,0.4256094044964238,0.38453222560738154,
 0.26196172751328706,0.8489984883838129,0.8397829277637902,
 0.17593852017461598,0.21333228339616717,0.8892735380256321,
 0.24983029625827058,0.20447378420444418,0.3723413172588048,
 0.5658835940919491,0.3813534647986905,0.3284158346260687,0.5890620544243017,
 0.03314268168582146,0.3352627050049706,0.6033183751535269,
 0.15889711949831986,0.7288603767585441,0.6882708404283816,0.8072378435068293,
 0.3929149782480834,0.0671786224426234,0.6657952970545075,0.8802611927456832,
 0.17259399280130616,0.9720271474703569,0.4034415484741156,
 0.09242530811204042,0.6836300878104951,0.2481340611820846,0.6011885408388282,
 0.4508900312073889,0.22635958427118186,0.30985837739170874,
 0.02015113336514096,0.8015146279987698,0.8483396989028686,
 0.17035550528473908,0.020000072536615288,0.7844366495326238,
 0.3113197628177825,0.3185751120745174,0.6550704530953225,0.7732584462815669,
 0.9588025125321289,0.02006972200256585,0.8577015535606958,0.8325649724658915,
 0.3236356538493018,0.5715044120586503,0.7180428499029802,0.43372974385445284,
 0.24064189738242459,0.6099479702399844,0.4129220430413335,0.9668850219508989,
 0.8362560599386696,0.4940057442325477,0.6768340405936393,0.14807202298900113,
 0.6603703364094415,0.44279241910438916,0.6050211541812416,
 0.47151097673894815,0.15771970946206104,0.7242403630493977,
 0.7676912335970655,0.47186151369864693,0.8964693385290763,0.6013840496530874,
 0.8325333631156273,0.3662084715265296,0.8764425935431632,0.18015670985987742,
 0.2748168455502533,0.4557405736409429,0.4006322542480679,
 0.007255070322149226,0.020932059870586106,0.010474623218822088,
 0.2084590364828114,0.8238513255079754,0.865094126277497,0.2947568270603711,
 0.20731455611400806,0.31702396314674075,0.37926535341853396,
 0.6327200583210312,0.46569302239229815,0.20967341684602847,
 0.02488930844209225,0.6442207914034914,0.5330922754720235,0.5382368282119819,
 0.2417392273893757,0.9927585018754207,0.052494996473785194,
 0.31254816658183604,0.5646558451162198,0.03340912332000057,
 0.7976106953161147,0.05896432242639127,0.9575925734162243,
 0.28048448707167295,0.30602450359858935,0.028919275898518615,
 0.6677915685038815,0.1221826785014043,0.7356656944824617,0.9764286346258224,
 0.4223993477704847,0.4858115896382532,0.942377685631991,0.3533015319538221,
 0.9109541415579233,0.8523667719656537,0.3448171573392448,0.5864240553412748,
 0.6613003154625583,0.00934625885393836,0.2306073882330032,
 0.22505057924355742,0.10108498838336932,0.5891337002982395,
 0.7688021315916662,0.0374569802993312,0.9136666610732246,0.9221382290491168,
 0.48083013510999595,0.018240327696813607,0.8290215166256847,
 0.18892433374067408,0.45017642503516464,0.20444753828048312,
 0.31969046362370257,0.3237666111300006,0.09305795297760955,
 0.3302854546447267,0.8614382129863194,0.057486714687853935,
 0.4339224353353669,0.9083291627413896,0.3847652816517443,0.813185827141651,
 0.9567727269422662,0.8245165043391559,0.385322823989499,0.06913342895310703,
 0.41493966509644975,0.30166793268625336,0.5940210528829325,
 0.5342566994866143,0.14768760399384817,0.8249530527857775,
 0.06637824945699777,0.769121232038056,0.698163321248521,0.4693246139484596,
 0.27456211906244077,0.8652575930920943,0.6682854840483866,0.73116951817268,
 0.6285088886072734,0.39753297715180747,0.30183854476293437,
 0.9912916315165617,0.18463886805380447,0.26725812119079806,
 0.18273278893507494,0.20586941997833363,0.8423210008233297,
 0.9276025125930222,0.2666685096157764,0.08221106895803054,0.9225384548817609,
 0.052722024706139514,0.9459724287149484,0.9900492178819642,
 0.4766241171444153,0.33137265141946814,0.5185123148402602,
 0.07485990780730023,0.1936970289686073,0.20322419524724378,
 0.9012774074335619,0.8173494966828009,0.6082024250049077,0.5539637084107736,
 0.25054746825534935,0.13181195290347314,0.8171469334565846,
 0.8494579345328612,0.22497758099363074,0.7013470683941058,0.284632702026517,
 0.23909483434477663,0.35663632667356837,0.4192164774052455,
 0.3534350982857767,0.49213269822129735,0.33146428806292605,
 0.7070797111187442,0.23113557637514348,0.8755243600475067,0.6404557282059247,
 0.41103416337942233,0.11071955639483899,0.840106495287974,0.8274031216006793,
 0.8499048038501988,0.1631587347714708,0.059313136240360165,
 0.25301721795136434,0.42292188729634805,0.06784200660578188,
 0.6903269533541985,0.8823684126683472,0.2646343834016448,
 0.020483470040862595,0.14689639637909602,0.11569651981824214,
 0.5188031925450107,0.149463548090498,0.9392120824241582,0.3457134106014069,
 0.1784119515319681,0.06798085790855624,0.7078094548539006,0.6749016703179169,
 0.17846973593178284,0.10753845000806261,0.9753613599245669,
 0.9248956854632199,0.5738366733392108,0.41598673906729333,
 0.14478443752061598,0.35232937067444325,0.601132871458761,0.5155552095457101,
 0.08978083545113558,0.08712166916775799,0.9173221578372642,0.748771353351929,
 0.5507542683874771,0.47428214316178907,0.9565347879907591,0.8094751494818654,
 0.795218383367593,0.9632720178742447,0.4167722335105605,0.5399076483231813,
 0.059650386083046736,0.5657905187190287,0.3569307294650339,
 0.15675699122656928,0.7827218902814723,0.8039282168892458,0.5041357097404231,
 0.9447633317326377,0.6076945385430239,0.6203489884254368,0.7126549152502646,
 0.9435602900141169,0.7196886365814716,0.6786934694900384,0.37327268710860706,
 0.40046747733272425,0.15662777911378212,0.4686400428800436,
 0.4852155008563124,0.9850839566607184,0.29671125950938426,0.5563821392857603,
 0.2936748248952452,0.9929112378799679,0.15977083355095745,0.877185338708713,
 0.8281106289161939,0.07945281117234615,0.6905335423237968,0.4024754350579569,
 0.6899543660877683,0.044946797644672354,0.7964240457796299,
 0.47975963778877795,0.8172248254231214,0.7483512955786735,0.4446860875364903,
 0.41335506734703653,0.17653955397183885,0.5902666257595587,
 0.5569165685925858,0.1815011429436839,0.420226133735096,0.8471249655712907,
 0.5389140329623413,0.8456000931356644,0.8700669324558645,0.18664160121147366,
 0.37403969247766045,0.15758827930236696,0.6057843431704233,
 0.6602918193652698,0.1661563963507081,0.5344411026827841,0.759486400989039,
 0.5360815526200645,0.2653767955265539,0.8722597649324872,0.24551222774804016,
 0.08818255440537082,0.47768537846010184,0.48216985178574945,
 0.566556489249315,0.2658798141783171,0.6225362020226584,0.6737171614349446,
 0.6633413358829331,0.7651905831100074,0.36886621310910117,0.662908710967783,
 0.9830110112177818,0.942828554651254,0.8453527148865188,0.37526544790184757,
 0.7250922576554766,0.7700691990440798,0.9636703553226683,0.9904158785562321,
 0.2665131032028337,0.6815670869473002,0.33856947432742124,0.6217608717484226,
 0.41017808301060465,0.08008970335426713,0.8319437025780269,0.811658978418582,
 0.8823834016588206,0.07826857153822886,0.2736271812247226,
 0.042187477733497514,0.7798178088642201,0.06413660329242898,
 0.3510161550132307,0.4787342072128765,0.18800369592181476,0.8626694178036912,
 0.048884382289438255,0.506518702030477,0.5194758162098256,0.5673310726899351,
 0.792543849449751,0.11566474198581811,0.39865642549096625,0.7084996692121114,
 0.46591035896369815,0.28329030890516194,0.8994590069598846,
 0.5094669855108724,0.8719466599909288,0.22774530532807102,
 0.034123514351800166,0.33348752423033756,0.9873902473908571,
 0.1734502981667424,0.5444440013939379,0.06842117520284985,0.5749905759488961,
 0.4676400980613794,0.3220747714869717,0.6983808250739354,0.2109949267652459,
 0.008144493706544687,0.8467156072499425,0.6969623272127953,
 0.8039944136328623,0.16555198647834546,0.41481568153128046,
 0.24286764158091612,0.6008698906893876,0.32774388075991345,
 0.8487665386028944,0.49722447459882113,0.6855492968741789,0.3830467032057733,
 0.18625252739005904,0.9336564405232977,0.8141823136118741,0.374848742682213,
 0.985656411986428,0.6592890278204877,0.9076528741723241,0.48476765737716043,
 0.3441637625405439,0.4949760867707401,0.11467003863315428,0.4183577173674481,
 0.35365262470981274,0.15534447289802333,0.4146342853506948,
 0.49579105566866966,0.036326481119709264,0.608950958541811,
 0.28517363605742096,0.45669875612028366,0.27125842797381505,
 0.5796722954200212,0.2962878152595797,0.5744903384633725,0.695702285838844,
 0.44727324900363874,0.7894869850938167,0.46266421440639216,
 0.8037638356812529,0.847689133417789,0.2316592467961306,0.824719659821263,
 0.4702506695871195,0.6262156626348393,0.6930587227141648,0.9974277583144079,
 0.34177268272271955,0.5777767793082705,0.002254670203749365,
 0.3943359993489661,0.7514600223499317,0.11946615119175552,0.4820052067496675,
 0.48524629013008713,0.21388315044252293,0.755626445364196,0.7176364811348133,
 0.994178016651657,0.12285590054304774,0.295873272876752,0.009497986444261586,
 0.16399169486364906,0.33086359545792277,0.8795072955829534,
 0.13274278374878756,0.3503158694345203,0.8701910095070916,0.9142166251191509,
 0.40579880540486835,0.31891746503783935,0.518680929909885,
 0.00422328867186994,0.21381763014485777,0.134390494477993,0.3805231016720313,
 0.2202015473393032,0.40793479117859976,0.031054747839796448,
 0.2360734120956134,0.8821297968865487,0.12781418304173175,0.8106355976454889,
 0.3736654766094949,0.9640611307489425,0.6821161346218346,0.9253175841311171,
 0.6442799097379631,0.92499619209739,0.24325888018171127,0.9308441105445022,
 0.09132766743395893,0.23795137103226616,0.19909716072133354,
 0.28040039453917476,0.6972534714316106,0.37061829920870837,
 0.06124866706642673,0.7819857767945666,0.5570601297204983,0.0802880052941618,
 0.09274109188546575,0.0304537510078684,0.5778680152255518,0.406448281676148,
 0.1709948455965009,0.9886566536524611,0.4758242919012843,0.7778013838610505,
 0.3815612233969775,0.535545525914852,0.783600876993177,0.7142856644780321,
 0.02899538776235777,0.950146626076287,0.03872004060721568,
 0.32928151333874467,0.32465379537608596,0.00408106351433879,
 0.9253428262779663,0.07835200136018339,0.6885646418673617,
 0.037116320311159856,0.6887107733616507,0.07024884494436068,
 0.8004796094200027,0.5705616492169427,0.2770659495901274,0.6437710338379968,
 0.7277498657881958,0.6039910163170141,0.36769129434431247,
 0.15422255289185982,0.2097976474315163,0.6248898683257335,0.5731060557204148,
 0.333129991796267,0.7330884024851436,0.061553017597161586,
 0.020266221279248375,0.7196885179475092,0.769250607156374,
 0.32707940247792944,0.34663915886341923,0.4103060045808238,
 0.44854686069355165,0.015840441885820566,0.4244325756483148,
 0.298240337365856,0.331009325262994,0.6223681168738976,0.8280056943030126,
 0.7357155475976636,0.4341501411185059,0.8968814663964254,0.17046505380131738,
 0.6036962454775728,0.6413040190070469,0.9314979537111228,0.620529758477339,
 0.7620886321624878,0.7929217129082688,0.2324207025141385,0.6131887610689875,
 0.17991288007135364,0.583059783896824,0.7331973238948135,0.6567254740942697,
 0.7968598327791541,0.7640713339120634,0.7176070382144536,0.8592208966100159,
 0.23525489926499465,0.14502076778303008,0.6861636623043809,
 0.7480440632687353,0.6759405851273516,0.9775446694121789,0.17635284050864142,
 0.6438712009191015,0.5286308143608834,0.5431456253956821,0.09234223349836679,
 0.18977975385429235,0.07898962532226916,0.06771835337793675,
 0.6801945083616778,0.01750135017907395,0.6544751815004372,0.8163501504771484,
 0.8102945940569481,0.4865940615784623,0.10766575429392078,
 0.45260182395682347,0.6617089633037919,0.18840665892754083,
 0.1296048832289095,0.2817082948046882,0.006691922214648982,
 0.7286271965355622,0.7488119634668325,0.8113274880239381,0.7595068450869427,
 0.37510199497559016,0.14752609560498897,0.3432040125672635,
 0.27313154244203375,0.8885046931047489,0.9306724287898097,
 0.06749880415416687,0.16457215608121856,0.08916316235891064,
 0.4430699804384406,0.19278647605492982,0.3150949194730961,
 0.05532378964738671,0.2762619367646847,0.9400388013911789,0.9128397578898845,
 0.47259501933974046,0.955891481405206,0.012554291474763501,
 0.1574130748768202,0.5313784416056042,0.20083290506423745,0.6627751438196947,
 0.21254201233849934,0.31650594843759183,0.8336576184112927,0.725611652884428,
 0.2180434032637757,0.9433502326427353,0.466101640375977,0.8452969719581166,
 0.6066663405771096,0.7383710843228827,0.8494261651533631,0.06650209987117084,
 0.9927587927071037,0.12960010709147307,0.016375629473536657,
 0.1350587381742171,0.21239722780610004,0.9384856339027701,0.6399596643876431,
 0.20611049389200053,0.2757534425323015,0.6222092669503388,
 0.45290393900319836,0.22338132135998945,0.18947088857183159,
 0.09833085181728307,0.23734061463861778,0.03176965323700276,
 0.1701253116458905,0.7146707497366637,0.33088343446301305,0.2020919233139319,
 0.8664262726565142,0.702207367698296,0.0657736312485313,0.7012601141918586,
 0.30868501206347265,0.18659179932685366,0.7070019753637633,
 0.4333400931095168,0.8480438365818759,0.732443613161324,0.15201577940193056,
 0.8134569408957069,0.24477017901343812,0.5158147877393477,0.876101575404216,
 0.13820579263871224,0.06359720999059204,0.21702724690513553,
 0.3867501914701683,0.21383039006541793,0.36818883168556493,
 0.7353235424390691,0.9880986284765401,0.022385845919110747,
 0.6295602298075703,0.6663782231883254,0.5703530060532652,0.7654636990542429,
 0.0534097679898039,0.004114258695809392,0.04592043899937248,
 0.028535722501991545,0.2416972306965005,0.883483082653851,
 0.36736917908122113,0.9351274653512646,0.2657972853627455,
 0.22429421577435038,0.20980164473444707,0.8631740758740861,
 0.23297890712935831,0.07795933835271096,0.4271116471328801,
 0.5337247755131433,0.05530871835547557,0.5342944708796828,0.9027721384681106,
 0.4613093637997161,0.014735575630967612,0.4713642975070079,
 0.2732067661560733,0.5304950381117199,0.941074741288282,0.4687721074644704,
 0.28745377908336966,0.3787239943095254,0.7276953576529207,
 0.37707394788104076,0.5335764033522137,0.05759499961771708,
 0.7481187552708974,0.23924233613225288,0.770747723331177,0.9325812161825368,
 0.9120958221290938,0.22610528494425086,0.9100808539805934,
 0.07528213403212158,0.6722714696623919,0.8923195591752321,
 0.37964944463866523,0.27700614701933823,0.09740726533875654,
 0.5355912619289671,0.24383495093840213,0.42946886504746506,
 0.9835627522675412,0.3390394854463563,0.47961102013682133,
 0.03262640269103223,0.5173057814744454,0.8199118796817993,
 0.21970753466766213,0.003995482818132445,0.5380134457376481,
 0.24877106967457363,0.6748122464598325,0.21350963118930733,
 0.19125007742383127,0.5883634816785661,0.43813169144354625,
 0.8698413543360477,0.8354402914546375,0.35365486089446785,
 0.029026593579465243,0.6655831491480566,0.7437444268054904,0.721753544110378,
 0.4667066045771715,0.2821134277276809,0.5047143311605071,0.7747612634827897,
 0.8930192391813802,0.7635318972966516,0.5611686308068776,0.02476068733976411,
 0.11976669978189158,0.21404999815814907,0.41526673019293114,
 0.9015940741267463,0.7630332254588544,0.7310386819303121,
 0.010742714753060878,0.461819848658109,0.8683843639401128,
 0.35073320915146355,0.13723620388756452,0.6077843670841256,
 0.3676899535848439,0.3077943114466638,0.5115559718130266,0.3089174836343066,
 0.15596906387132659,0.6240273606668344,0.5573622750788765,0.5005899825702601,
 0.692944251743737,0.6448908153918282,0.18040868277602262,0.7871558360828717,
 0.4066314134749932,0.79672354205067,0.18885404320760701,0.055729628849241486,
 0.40771798838662343,0.5893471100515777,0.8427516298633324,0.2649166559376349,
 0.1550872297844783,0.5651221636103936,0.975264447382768,0.3140599220776005,
 0.1474716793977625,0.033533382816623325,0.9541662370034321,
 0.7738283989299257,0.8950997596091437,0.9367265540047142,0.6178299315562665,
 0.7442052392030691,0.5970030400777103,0.46808138963926094,0.5699605583299073,
 0.9232821103778062,0.3814971657464733,0.6893632474377122,0.9702721626528759,
 0.14164216113328,0.9902854743421474,0.2290813452593683,0.5630994254569898,
 0.39280791661596115,0.6022885842181946,0.3267919236773261,0.557435160247413,
 0.8703841854039015,0.7917564790013935,0.5687114392855865,0.6543026395969372,
 0.9399816635858016,0.010461159763210715,0.008495564398557698,
 0.02725886882487083,0.6831512495630517,0.5408854721979579,0.8673261660574891,
 0.9630322308670394,0.10019315705641818,0.7696250283127339,0.6637188601016809,
 0.04191994274570987,0.1269933138154653,0.8615943764707271,
 0.34986874221193365,0.6672106619680759,0.9236953087851668,0.9321887476301445,
 0.43379986723763564,0.1013850612452013,0.42920622126877384,
 0.7586936289542611,0.6169461850011752,0.35897972252264265,0.3646917162547888,
 0.3652372873192813,0.04371642571905332,0.6311318543806761,0.3994395148858788,
 0.8015631808466697,0.35341274057193905,0.4017693326437226,0.0235426529550139,
 0.7874747134587013,0.7091205793033445,0.7466886066728514,0.2847813270432237,
 0.7582139781213314,0.01751579369422418,0.3841698459732177,0.8060640338529538,
 0.11646934794682595,0.05570036128096012,0.6407824128293897,
 0.8477882994887002,0.15966913449344755,0.2258284014569818,0.4649058541653103,
 0.7594413084490647,0.20212906179100443,0.2126412708588109,
 0.04603719417381069,0.6840630945794764,0.8185712568053907,0.3677347413465787,
 0.10531934578798752,0.3255571939465869,0.4134683113417461,0.9612604946121552,
 0.4697991791213818,0.9374547166964216,0.874491873118379,0.797820978338296,
 0.651394192951905,0.9613637045726029,0.7062013214353009,0.3978440242479365,
 0.7197165449131706,0.7207723643215262,0.10568066051582226,0.8770463021548623,
 0.9538686214165197,0.029307132638767053,0.3235011220455182,
 0.6703808353561069,0.8725672236682033,0.3851975394140277,0.4331729296679889,
 0.41523200782786707,0.38419363597436174,0.21817657377416766,
 0.026000471695816985,0.9053731888654567,0.7147964304128348,
 0.2398396283881271,0.02097395037719363,0.8479114210806953,
 0.48574151980346625,0.23548656462486983,0.4533452291468014,
 0.36027280523681504,0.02950859587372001,0.024668822209504837,
 0.5683706680379581,0.05305447574813327,0.901834773848865,0.293908997082138,
 0.09217105126277647,0.5304209796563312,0.5206920917254948,
 0.48184639565080256,0.4066851226840492,0.8849343709237532,
 0.08793694436398358,0.6463979084138345,0.21815939234769288,
 0.1565772393745839,0.36618716720651,0.14780366501682585,0.7822590791935807,
 0.2764913037846076,0.8485762012923352,0.9915187545775797,0.829238046455758,
 0.6355867703882316,0.4105571709515784,0.7979835078638203,0.1036797763355215,
 0.061023817580088124,0.04578228829377262,0.4870989541521845,
 0.3434156141532444,0.503857451361763,0.9228997059364306,0.5864241174200054,
 0.7163278444905397,0.17582123247079662,0.6703555875664166,
 0.36483095829558976,0.7827799845572188,0.9981799181204997,0.4012483086540122,
 0.3000594349061474,0.7751147491179471,0.11769192736732226,
 0.30265101867948674,0.3457954782544955,0.004221510882024604,
 0.9255054122407372,0.8307639406922476,0.737055844666307,0.3490704858469639,
 0.8560531509783106,0.14033436190360382,0.19261367130934004,0.783094712540092,
 0.8059239021419806,0.008299149499656822,0.48126116090186466,
 0.8558563006734337,0.15641022526316672,0.400910436221783,0.2932744702118637,
 0.5491022376012847,0.28871930234962384,0.72082192336729,0.9395974554851233,
 0.8502459874511422,0.1482223012019558,0.7431162230397217,0.3579344863467745,
 0.264643091389408,0.939913634786686,0.7710852239970046,0.6367161250768868,
 0.48040001276490507,0.15237664683537,0.5722493833124808,0.5177505240413525,
 0.8990724423013594,0.9970894920795963,0.050639072600656254,
 0.7360454379943203,0.4968881802002786,0.5926608578681806,0.5209016473706055,
 0.29851952231218726,0.8838251000931203,0.315955579406874,0.6086898363949826,
 0.03277412872869068,0.3193835086950749,0.10503288248515208,
 0.09718873179122978,0.6084948725353198,0.6811791746176371,0.7169923456879719,
 0.9615022825343384,0.40535681360594655,0.5764486499741313,
 0.18437729352456755,0.9405407552571777,0.307671676812448,0.4772142347151499,
 0.644441796955779,0.4141872941315363,0.43560375710195465,0.20886588489951163,
 0.6038281498088757,0.29021180752559295,0.009165858181737763,
 0.09296050212081841,0.9743594119894747,0.36861920467185294,
 0.16259123224574512,0.967616104384362,0.08125195969985965,0.5093944187255931,
 0.13436304956743672,0.5104815193679382,0.9289011054217672,
 0.12126981009390458,0.35954208255558295,0.36988993350620325,
 0.11729746464966262,0.26258642336494864,0.8748891506387081,
 0.05708631439512979,0.21460219038795447,0.5365499377673919,
 0.6961910171141446,0.6393516435941256,0.49704487616688964,0.4641997491715505,
 0.09871771286287678,0.15078739684303866,0.8290698246673,0.7529273265476427,
 0.9057910473269986,0.5255244584579076,0.23604638420759627,
 0.06396727546500314,0.7479159785230991,0.7892202251185791,0.8908439875037465,
 0.38222871985253937,0.9162320618637878,0.8686227940643697,0.5137576012593087,
 0.7373322139028158,0.23468301452352403,0.7459009998157533,0.5881383255624072,
 0.07304143872233781,0.3174068913476442,0.6564034413942609,
 0.20039379860176565,0.504635281917046,0.020521598365619842,
 0.7487539452367212,0.4177767514595835,0.9808953882810457,0.504617276056955,
 0.3617807288806827,0.08880687901562334,0.6730028196679312,0.5534245374897473,
 0.8665916682206619,0.7221008604432744,0.6913284208864258,0.8578797391195316,
 0.5721996679771665,0.7592416413346467,0.21394197198480702,0.7177249223613303,
 0.03837428287131339,0.30588526320443654,0.7706051339443556,
 0.6492579751476125,0.9385120707686673,0.6511607598524388,0.6489958776610548,
 0.6993694476407732,0.6524294206421718,0.7411958416577217,0.4729295232109031,
 0.0010129287678721788,0.8115416589805062,0.7366139192403831,
 0.6224257775374169,0.912129499926014,0.4866743045398889,0.4525052447505308,
 0.3436188864121503,0.14073052671601782,0.5301448760546842,0.0743899976247463,
 0.5264221643875481,0.41728433651998265,0.6663430793680079,
 0.08028815372422293,0.0343503841865912,0.1336391609989419,
 0.48309589471876735,0.9845317713548055,0.4018262602134025,0.6520182355540737,
 0.6502623423811186,0.6093929291859544,0.4900697886029588,0.7174289259750367,
 0.9260204701283576,0.7756872815975236,0.6911060080583472,0.6393476282255234,
 0.6636586651938366,0.2644802553708968,0.3160623841447967,0.31807872202368725,
 0.2475647782880145,0.41872652317089376,0.8355284933148214,0.4486422601088217,
 0.6213476163668474,0.691694066545764,0.93309230144691,0.16190635736616132,
 0.6948443625736757,0.8416198697804232,0.7171071884201677,0.606557070310125,
 0.908719366589975,0.7420357416792488,0.800902138256449,0.3201318663879925,
 0.11478326593326937,0.3113408989559955,0.28923126482502637,0.639728019295063,
 8.38576382584888e-4,0.7454385022908964,0.12040904704858335,
 0.43818506031271076,0.5742783087988867,0.3302456629311046,
 0.05148029732180237,0.6747144771095862,0.04449137343662135,
 0.5862468517666919,0.7593532337969834,0.08868743521249267,0.5926207169506784,
 0.3557912439011467,0.6087265171006546,0.6526407684795548,0.2871791197891891,
 0.7017078444579894,0.7521309526314607,0.9016872208480337,0.3542602356726411,
 0.7263221872179951,0.9441226336616813,0.0683597466895105,0.48069343625377814,
 0.4484627050532821,0.49396576094061806,0.3225657412968357,0.6191723003746781,
 0.6043256758436036,0.614765945896717,0.6019075832535736,0.42948072862281095,
 0.38895723476586597,0.3744522046716632,0.26413790796763803,
 0.20208154959156022,0.5660041748836281,0.9571558174434734,0.4560550789155816,
 0.6097693036145659,0.4084200907838771,0.9785849116671002,0.2094932891592216,
 0.533728271603415,0.4250591352580271,0.8873705183514405,0.4053225132709395,
 0.18504492777418668,0.47161934372710335,0.6808795584065662,
 0.6217297620083149,0.9326884443037132,0.9500037136399683,0.7023122144693776,
 0.24255752790499308,0.7001063046534919,0.3923283546263483,0.9455774077220132,
 0.6783458279618391,0.8147575919124402,0.35565306079925574,
 0.11965473564955786,0.16789063565474993,0.0023663987481030713,
 0.04152989238395388,0.5841981865521403,0.9696460490663997,0.43904764222071,
 0.5293653257616466,0.3462948627611325,0.34972009404749205,
 0.23278710144466241,0.4041598427180081,0.0031429960332209905,
 0.9582526068455839,0.4024170769440668,0.004956186631223658,0.874662461139748,
 0.4075628622278812,0.6252546877384672,0.8077083401901838,0.2524363857327504,
 0.40194616218636714,0.5002963968978862,0.43211639801112856,
 0.8621802684011298,0.7555184630158738,0.5110125996157419,0.7820491251050852,
 0.6176695173413936,0.5246584542938216,0.0067969134717700275,
 0.7395618674224674,0.5839934851004556,0.5679791906205292,0.4501802138038683,
 0.838135640829604,0.4808460745884642,0.6841216646613475,0.06995856363828812,
 0.4146476376491276,0.1918725518305634,0.7581782318587844,0.5228265268698548,
 0.8756858899280998,0.9259032833006822,0.217189900765004,0.13329099731224825,
 0.3801819564101304,0.7529317463107649,0.1560945659990539,0.7255746255991536,
 0.5331428978944874,0.36534704244812355,0.14694830470070064,
 0.9404580799350758,0.6798189633302757,0.03607399101107722,
 0.00972562594837445,0.8233120371152736,0.5780168560377659,
 0.24825895339355808,0.7675296922624579,0.732111326569388,0.7930663040646208,
 0.6635497337368472,0.6185389843399722,0.5626738416124865,0.3617034843873411,
 0.35257377902606246,0.38922678645926334,0.33654842964727816,
 0.17125867835043695,0.11132724995551457,0.666176617641279,0.6183413991836405,
 0.5736298067348711,0.9820914560479704,0.5419634753083864,0.6880205047146747,
 0.34810763388318067,0.994834924763043,0.22722407466114658,0.7210460698048688,
 0.04358219323490575,0.6163455475170547,0.7153200801684788,
 0.09384230560496065,0.639364908719362,0.6725158137424069,0.5781370900358086,
 0.3475094218881405,0.3089511542451535,0.6711045449609813,0.8794642057938532,
 0.5390848760258763,0.3359252071049652,0.2245619385392318,0.14862593071969388,
 0.6340240004350631,0.5166300059643971,0.672742489812253,0.08927851448264978,
 0.9711669545881548,0.7119610913314481,0.8703442354283034,0.31451722511319524,
 0.8264682794353824,0.15795028539502043,0.1919882760229047,0.827739214103167,
 0.0540337136884943,0.9451998666528132,0.6782262050653359,0.27821194923746084,
 0.8601085119040576,0.34687358045757544,0.5429779688727638,
 0.41068580213161066,0.7447039628183405,0.5034620128205789,0.7301790369877245,
 0.7100233747127381,0.3129015187441998,0.573015674517825,0.5630621573124952,
 0.6253256502009371,0.8745733263674693,0.8419108284377682,0.5387080947325021,
 0.3336878137667838,0.35709253552177356,0.05151013418187056,
 0.053812577388452465,0.8979332633361389,0.35117425974990923,
 0.986835946542974,0.4879654113247728,0.8893652825788501,0.6236788160003859,
 0.5791928728909074,0.6715106947290145,0.015367567817286432,
 0.6126151141408105,0.8950143004019554,0.8443747342688568,0.28381267917216346,
 0.7357491173551794,0.45243327284473067,0.4190165698488131,
 0.18929616113692915,0.5096722006837352,0.9686086909587631,0.9113151999215042,
 0.6746425900717496,0.11847854814706205,0.4335407858418314,0.3121750789648031,
 0.6796900824379595,0.6084625552344773,0.22338824142468527,
 0.21039045220870167,0.673825496441164,0.13010244391389436,
 0.22937476016215685,0.7410214674358288,0.43189829781856626,
 0.8967357143084115,0.24551355707614242,0.021803422629240954,
 0.3490380220602862,0.7016961093455707,0.47175726798744955,0.5657179080659551,
 0.3186171351289929,0.9560568024954873,0.546981391686136,0.17275946631474304,
 0.45765949974564213,0.2772177036448845,0.6360856555270722,0.2507634915541428,
 0.6481152897735435,0.27075166201341805,0.2561038424483594,
 0.005688359268130494,0.4745107402349965,0.871451121795578,0.9470399712369911,
 0.5945543033466785,0.3494726665341702,0.16859135338903486,0.6806805597985918,
 0.46857886955490213,0.2830322861855329,0.32907629115656856,
 0.3001028397525092,0.27326615508111285,0.1736796913893316,0.5167602287207317,
 0.5049498600282245,0.1043965027524878,0.7500501517859391,0.38148859656316325,
 0.3758671829608262,0.1906650388553187,0.5708813616181575,0.6337402743078042,
 0.659372223269426,0.601334143146057,0.3114658197211818,0.7481062424781109,
 0.7049379681420396,0.291437691006164,0.3873784844862197,0.07820587356270625,
 0.47804183201173167,0.1358065996502762,0.9681191025384542,0.7170027856495242,
 0.14988440634131206,0.3612476779707322,0.10620455891376812,
 0.4224278170454854,0.31562109407536465,0.9900879013312966,0.604636501717452,
 0.5443794586212465,0.9146365850329707,0.291595468573255,0.33779806357754794,
 0.6260573463261286,0.47060287067731066,0.3716696551799289,0.7637521102821615,
 0.31020044175675554,0.7260459586515817,0.8451758991397332,
 0.12535832973276184,0.3630929571489565,0.6032571333818431,0.7923441407768855,
 0.5937339680745644,0.4486845730833504,0.7095117082730683,0.5340977121427812,
 0.2789397262327944,0.26734883871419923,0.4964912968472104,
 0.22947164556555355,0.03584326951276562,0.9047542564288901,
 0.5759127044592436,0.3921461966539417,0.22474400400793904,0.7800365109061107,
 0.29359949507684313,0.25988808360405213,0.6996307093156302,
 0.05394165591938582,0.6853867701105697,0.4152151423731506,
 0.005676960962674005,0.958068725193965,0.7604216192526759,
 0.08973566637924257,0.10854304352017186,0.36682743477589885,
 0.9802159778268377,0.8097861224301101,0.20077371112739073,0.7468364262332803,
 0.30731816966452674,0.5346281325481715,0.4723439191469968,0.9390867677746461,
 0.7378246020623428,0.009239647899760839,0.6358021426504704,
 0.14334630627973388,0.2691929157415587,0.2335860719191536,0.9057433499130952,
 0.2309739643212425,0.25999958597130857,0.04910588547308048,
 0.19799813268352406,0.7987241585997641,0.38588856816530304,
 0.9564257912806946,0.4444373785509097,0.5754748840891484,0.6253790594303057,
 0.7132269967038405,0.2383038989487365,0.4685847295610276,0.8303869169604648,
 0.9402356122595117,0.8882079221018314,0.9047845727563988,0.43488034225216965,
 0.1251913972689278,0.18058578395238356,0.5677993813622336,
 0.02621164731862169,0.4944777959298382,0.20324165178918563,
 0.3527260987889137,0.3545240254712251,0.10857998544447911,
 0.20508086148340388,0.812987868176664,0.19626561726148362,0.372958139469704,
 0.19427230000264906,0.5014547159619811,0.11513886023519027,
 0.8872186786809393,0.546601730199111,0.0950734802069435,0.3924695325020716,
 0.3621656260274073,0.8234040352987929,0.15314042439059228,0.8140021433774336,
 0.49835947982095297,0.021571201465805823,0.5137280198298279,
 0.8813180134783929,0.43171540195813896,0.18025073076217113,
 0.43160661944485323,0.3283601498295733,0.35179182700816636,
 0.15228502612843609,0.13155987548359604,0.127842744849336,0.9505698117233294,
 0.7226457748812982,0.10882560467116609,0.5358625636471752,
 0.004232758866722475,0.7138655085362641,0.4563612626871789,
 0.8745057336825472,0.5749173615915332,0.07536992986972191,0.5967132449147661,
 0.3690739014947513,0.5610145763044598,0.311071726271575,0.3648272008755403,
 0.6188677995193003,0.471432396375391,0.8787474522093276,0.6195977602670607,
 0.66212666097976,0.5761762018322559,0.02308048333504975,0.4597038416210122,
 0.831281394977741,0.5987718180173299,0.2677150045034291,0.6284436994757947,
 0.8176540678253997,0.9920566950894553,0.41961907606748194,0.9829669104315701,
 0.06590798322422464,0.6705753264709448,0.20991035216165976,
 0.3022716610151518,0.886155750888552,0.7122413512252452,0.1817798293069064,
 0.5952855177404351,0.6910133428675039,0.8452104067422296,0.9849415113239562,
 0.8495792569027197,0.6671549039339698,0.8048557295169481,0.09242082019894926,
 0.2824113792214975,0.8333425809064886,0.897952687533428,0.20969513014488594,
 0.8102461720108186,0.7626824080302113,0.27651398088382295,0.3917175714302472,
 0.6680402165865985,0.7913605070378238,0.4550541880338915,0.9032858790748441,
 0.9849620911447886,0.5518713468494703,0.3324241876831764,0.09866144728726955,
 0.3667314703930359,0.5087332614647518,0.921608812978729,0.11687471395462457,
 0.5620025281490798,0.8724431553185676,0.931475319276996,0.9570923446488157,
 0.3782296339750514,0.8816729837360388,0.9888014943086543,0.9271904322243828,
 0.09584648412456798,0.047919944988117025,0.2624171895002646,
 0.292628501398161,0.34066307474172997,0.6794114483433809,0.9163255315361031,
 0.4517681435794195,0.3375583107606882,0.14550511928009502,
 0.17126033351467462,0.9520678747074534,0.6930341019642715,0.7158944640024792,
 0.3389350066828972,0.5447155924081999,0.4408627975263192,0.667742473626734,
 0.9877593144355383,0.11476527613457965,0.340195501509733,0.7221993784341149,
 0.217503938834942,0.550289415793224,0.9653688310530013,0.2826138439729048,
 0.8513468421191365,0.48055493920284886,0.7469135012268906,0.3844071946630454,
 0.011838781424011735,0.8624917142449648,0.384291071890015,0.4255433162287917,
 0.40008786380333294,0.8496448251545878,0.048105568882297,0.03209983297000529,
 0.34391370531548904,0.7069658812682234,0.3352635814110483,
 0.16552985563788092,0.161180218607738,0.3076538472556324,0.48914077214138085,
 0.5185631217163965,0.8477292150741529,0.27700146322722885,0.6666290929847972,
 0.9780018699382165,0.05515222474738346,0.8648131608681173,0.491987157211458,
 0.051363639682170725,0.4347302806179252,0.7016239183281342,
 0.7672739440191362,0.302803625998048,0.8128499416385239,0.8613972139079982,
 0.797197757011465,0.06897692514421028,0.921083947750032,0.6895651327377468,
 0.863166039109279,0.3070176272255358,0.8455574408575204,0.22092576263720842,
 0.4515120586082274,0.16088499829809932,0.669047089776382,0.48797624072597545,
 0.6307903955732939,0.8967105688859721,0.5069089574398092,0.9762344754330015,
 0.42965217351347307,0.6765984351466717,0.7469640457422932,0.9838574850936426,
 0.9369565803567271,0.97302072041531,0.9678349782344133,0.532775664219995,
 0.3063649063681797,0.8938356833747918,0.31296828950586675,0.2942750700160093,
 0.23760839414528234,0.64050634911183,0.5568133508934779,0.4298638350063845,
 0.5666190862493525,0.0925066755813505,0.05451267751671862,0.9391148345444501,
 0.22265674421829496,0.9156114521113082,0.2536036775411781,0.8012215106803886,
 0.4533472742654494,0.19338590123433974,0.9428833811760625,
 0.003401707337021165,0.6571957234076719,0.8772094804530928,
 0.7598527658227393,0.4296842540123096,0.015513199489991414,
 0.9302767977983051,0.19674686643380224,0.15873433868989295,
 0.5378247968751886,0.3928773868629183,0.40609885364032317,0.4385850791609045,
 0.31473197719627355,0.9218040480099502,0.9447201460638448,0.3506697824697075,
 0.32673572766055836,0.8193388083921821,0.235302458091042,0.11516400567613971,
 0.3814902042280619,0.9614293846808345,0.6442686632636794,0.4459688163657789,
 0.08645082261514692,0.12895681072680554,0.522200341505198,
 0.15181507802830807,0.42678280205389796,0.4802250250454335,
 0.9734819311013311,0.33747512746675257,0.4514450709071396,0.7843239974726998,
 0.880577393225859,0.5266679618201666,0.29973398847944144,0.5805339164284141,
 0.15103806026490518,0.0011330488760306068,0.7550331908886332,
 0.4169510736082671,0.5570583775437925,0.5355553181333187,0.5136004324444107,
 0.17101089975529626,0.38070394142489294,0.1918455584384975,
 0.6776105303936863,0.07618269497434726,0.18576316741099852,
 0.33571673127228896,0.2324885090454989,0.8343310896797617,0.2640191497436577,
 0.1367801288923739,0.3668495319396703,0.578467392988647,0.9768612375941144,
 0.9126345552683947,0.6348311514716589,0.4003810701658417,0.9547337847853545,
 0.26800115081647635,0.12130164155047041,0.3799603696184787,0.698717658627536,
 0.08187815783385677,0.8820913255251459,0.9482583567987056,0.8559370494256094,
 0.7368395516308268,0.6430206058418361,0.6761982504449082,0.9748923471376109,
 0.2821860620611262,0.6752483061864735,0.02209594268425019,0.8087452745505036,
 0.6966565169159068,0.5831544390325585,0.9008801306796868,0.9107957433617779,
 0.9366850721409554,0.549641580436898,0.5534659335297585,0.5892268531523969,
 0.3818806144578287,0.09791509537289333,0.36007093379729227,
 0.9259720190378955,0.442192696106374,0.0993935744608212,0.9941471425884671,
 0.4202521667572655,0.18442097027632798,0.7595861871760308,0.5866663363113152,
 0.6242455900268079,0.31500140587959613,0.3799889355869992,0.7764505935683456,
 0.8097236441026995,0.5678423080823458,0.20745738890690535,
 0.08456297412839842,0.8714164114691847,0.4455902558042565,
 0.47297783184973474,0.9321305836931754,0.037326213734699465,
 0.15463737770051778,0.7869935848997206,0.1791883418005601,0.54073346388433,
 0.3583923492710728,0.014117870798026622,0.3385891472846745,
 0.7923696315311047,0.41473289721220763,0.5232926367714176,
 0.059931532543310295,0.7481482110849299,0.5640074092336794,
 0.8302865561336392,0.6188882626284489,0.21461653082542098,0.2592974283840953,
 0.956930014846889,0.642617768497734,0.5642703388313859,0.10362910883780785,
 0.012581652518817044,0.5214828826612872,0.6515664920662259,
 0.4174742905128055,0.6443161904844009,0.9976238362594866,0.562041064013534,
 0.9801544459000225,0.12259586474221063,0.6569987391899537,0.615859317377735,
 0.15010827543686833,0.7358742141042008,0.5684298361164379,
 0.34638072171835954,0.878727032932662,0.12624912783191955,0.5983735426672449,
 0.5882957914082182,0.2117445562423339,0.5603197266214305,0.44536703515431586,
 0.6818783747696627,0.9736984486224832,0.2400310336627256,0.08453898665825066,
 0.24226018012597406,0.9069951244221789,0.09566915440673851,
 0.2535589285813399,0.31337068950653235,0.889653723280094,0.3342256167203367,
 0.653636375693014,0.551119040135473,0.8679137956993924,0.9530068559901576,
 0.9434721547972651,0.4191630172178258,0.43056696507817493,0.9491181641456806,
 0.582650330718978,0.6704424168442151,0.8963537913231697,0.34162723776292403,
 0.0787495759547534,0.9335455665984315,0.4872109656253092,0.46994604619661595,
 0.3502293344800874,0.12996044946590102,0.343454790106795,0.23153471976727324,
 0.6817363582403814,0.41816022221568616,0.8371046078982289,0.9786461185784041,
 0.7751583914389091,0.30402508629068037,0.4150220278712975,
 0.18252414053923038,0.3665780905362187,0.1628085096510663,0.2335307059697309,
 0.8613592771236255,0.3033685269278612,0.2783780423272617,0.8960351818387939,
 0.800768743695822,0.9242477678310336,0.5692089091112548,0.40574737500976854,
 0.03749250803446269,0.8715851010039746,0.8395904707406028,0.8502859022058636,
 0.8432996388096307,0.5625109907384978,0.4208196872514328,0.8939303692429139,
 0.9411276646099743,0.1767512627099056,0.5513744889875758,0.48144195748252905,
 0.3060796987049571,0.49058725964351857,0.45495924758092476,
 0.7386887659685907,0.49982338985631214,0.5101182973455548,0.6985691434019772,
 0.4985753166213618,0.4631432315217354,0.0540122985396605,0.22748465188501732,
 0.995709578562825,0.12354464902647455,0.4263081787344368,0.2771302147910305,
 0.2454177037094023,0.8392999849198528,0.6720618115108712,0.02001796949461665,
 0.2528177914317543,0.2678366340681849,0.8860001550829034,0.3851829749098463,
 0.340532114596218,0.8327574367734074,0.10077112320407988,0.7874453145052266,
 0.09072418186874653,0.506537068988913,0.4116915300313597,0.2914561379821764,
 0.3877936734809623,0.7003447574951651,0.2556569943571717,0.5244136159119613,
 0.3859817825164873,0.322457380215188,0.6239780341601914,0.8123722667291859,
 0.11938222602062187,0.35401101665458734,0.38382195056430857,
 0.13383662165309218,0.6949387294037869,0.27965266024547175,
 0.24666722746180436,0.8456112284683239,0.19269125863441072,
 0.9406550598181553,0.3029493486154926,0.2749669320100796,0.8556178768145314,
 0.5719657402763447,0.4079916593578874,0.8588331856156305,0.5918575223158039,
 0.8669095222651538,0.6128647686642216,0.7593822276170058,0.7465208857198816,
 0.11612618645170691,0.9414105507413034,0.5484574882525566,0.5183425765807418,
 0.20085647043157806,0.2680091637695303,0.5420990671619019,0.916853504596842,
 0.10696873568990262,0.41588493818907346,0.47556810063427624,
 0.10239857664753615,0.20539443009898195,0.1555212479605157,
 0.4308356980511907,0.7992226934376769,0.32047965698606506,0.722090493714522,
 0.6210502867093728,0.42636999112857643,0.28571905231275374,
 0.6478506674179585,0.7343939438999224,0.27570013844731855,0.1518874765105649,
 0.5516029707033401,0.9481650345374619,0.9056882494461068,0.47758314504975885,
 0.5054392295806543,0.7982258853304308,0.22900798455455074,0.6703267669560495,
 0.17911859501162675,0.6844179919790901,0.9832970072589566,0.9318388409367149,
 0.12495727653223732,0.06877783020428785,0.7821119590769096,
 0.9906502999946337,0.2360937623253878,0.22994748556208744,
 0.035254354269795174,0.5371635212718577,0.6445150162977709,
 0.17892273561484306,0.9719278886578138,0.2856776068825223,
 0.06369858023903674,0.8933533849043815,0.3295044580006605,
 0.11909080171167474,0.4104293776864507,0.5574660134237314,0.3751143013881305,
 0.18081880859974753,0.6711291729808844,0.07842859970912697,
 0.8184505349233364,0.8008844431427702,0.8887701314428551,0.727226172985458,
 0.9470819222276357,0.5144406412074516,0.8107789649624964,0.42077284167720286,
 0.18019798489917194,0.439531700854356,0.7301637268442818,0.08522386057488873,
 0.9959793715252346,0.646065034636361,0.495327300356727,0.7569442644699992,
 0.31753155871020555,0.2600818762818031,0.1283296264294499,
 0.15758811115075533,0.7972283473974852,0.9421543678089416,0.9690633173424821,
 0.7534070463930375,0.798343528115654,0.6391948865040646,0.3693542134061961,
 0.23711039693169444,0.02095054480838865,0.6488324238098495,
 0.36178988389571043,0.6046097784169161,0.47917877827196964,
 0.9071264661189766,0.3236087031035114,0.6843930297554256,0.7824360481896728,
 0.5285235246865467,0.27901161520717976,0.4429810246746577,0.6166937118651282,
 0.5205005371862851,0.5497951369696406,0.21391153323593137,0.8704721313677822,
 0.31412623512124793,0.06700865859241556,0.7170308040131305,0.541545537433015,
 0.1391361008713502,0.1401801631942624,0.12118140793058452,
 0.18607085119561295,0.31831421422078554,0.9818830423652487,
 0.8240049895864896,0.7028336038062708,0.7734717402170141,0.27849986435741136,
 0.9045029238443841,0.6783182465706723,0.1178026661431778,0.7215431438237442,
 0.16182412643180455,0.2318719185166963,0.2704768461675502,0.3187310491243712,
 0.37541592881929153,0.29173789842162534,0.03760060881882288,
 0.8473193005309682,0.9752042764343604,0.42048844381884076,
 0.02647606008310255,0.5031120462553211,0.16178496124063446,
 0.9395340947874831,0.11443081095973351,0.4435239577084755,0.44331059969246,
 0.16967347780625353,0.4707126806648392,0.5902008296726138,
 0.02430578816158191,0.24724808225549,0.49269363020120804,0.3614919480948031,
 0.9201038662064149,0.916632639474428,0.25112114957753107,
 0.060864103093653976,0.45252639533460826,0.6497888443810104,
 0.6542906074284626,0.31304409426168134,0.6992835050237454,0.5925562594708539,
 0.74527763795901,0.793567726508684,0.8934555796711017,0.6595510552452098,
 0.5298956197171769,0.6717053956617761,0.5285303063279936,0.09738370670520291,
 0.814348698309893,0.41069220974422826,0.8883536300162402,0.3575055584222111,
 0.9221802005530119,0.822325400620375,0.32183402406592443,0.8223300299486308,
 0.6180039188232311,0.022169195596177138,0.8860542296359812,
 0.5487130780023506,0.21510193045165127,0.6056141293673737,0.8433855701148756,
 0.5656948841122621,0.8192684297296382,0.9508431811647148,0.3210706630328244,
 0.6113954908623388,0.5252252830920475,0.44156923802097037,
 0.12067454352743612,0.17034833299225716,0.10684713406230317,
 0.0692999729984215,0.7729333091035511,0.9130225396733839,0.5196250844414275,
 0.4112331226772734,0.7562892948507968,0.7383362510879768,0.5819302588297538,
 0.9053843505355624,0.3279654742441096,0.09651506608452831,
 0.24933203343611088,0.8505585309322273,0.2882633085304521,0.3285298111358499,
 0.4327709147114237,0.5727400895481088,0.17028674172969172,
 0.20371091610613434,0.010156688949803261,0.7005635672912953,
 0.7891349627038788,0.7250226503314623,0.85707463237563,0.7023197960985635,
 0.9845721445812434,0.8965611393604112,0.5296627188853658,0.23487013824977143,
 0.4750755034877634,0.552877502717934,0.6531167537171472,0.5422328436926974,
 0.6996176131244898,0.571610703760441,0.35517482944294543,0.645233979559658,
 0.6432419021460387,0.4683693387344098,0.9656241505752066,0.11079683089265102,
 0.8662304824426775,0.3254694090256922,0.4314272831533067,0.39216376207396664,
 0.19129598006043513,0.9096465292414136,0.9074293436822564,0.8673798332215605,
 0.8624964783656178,0.455960688323673,0.32423507011977315,0.8563495858041883,
 0.5664664690785214,0.5774888451076109,0.018500900391605346,
 0.21829556396149674,0.42224794580819425,0.3319517606360898,
 0.6080478993087878,0.9163155146115285,0.3770309126131677,0.35946979650116795,
 0.3581481349013792,0.5313600782325103,0.6651829151860087,0.24228919611942557,
 0.05591854579809785,0.6991208999675924,0.4529340885454043,
 0.40068936318080395,0.6358432736088067,0.6561909108812682,0.6559185491611931,
 0.0283809171011713,0.8218054804120707,0.24248602023356636,
 0.03368851819239438,0.15325131316139695,0.7089089780684709,
 0.6424752795439512,0.4122186011039479,0.5005465486138219,0.1688566442964271,
 0.5906422508510649,0.655537349599379,0.2680005180082763,0.3429431320645733,
 0.2461297598346397,0.6118541307062165,0.26516871860728974,
 0.20890628011612977,0.45796320166874194,0.08027676961626062,
 0.6659366593863265,0.36053317582506383,0.5778158231333482,
 0.48359182206695583,0.6769800671482556,0.6366057020016547,
 0.12834617615225286,0.25507975446938547,0.5838885719803925,
 0.10434022481535066,0.24423613637768948,0.14186185019227493,
 0.3368158200148551,0.11396050473104002,0.27782782327628863,
 0.5345492076758181,0.15765592922736182,0.30370631548410787,
 0.1429695713216913,0.8379850871302298,0.09393481294658623,
 0.27379699489121756,0.28296891252787926,0.9085941327716217,
 0.6597288317316998,0.08509846953100375,0.16781208436010586,
 0.5389660693634413,0.12600138288930796,0.059159618652039114,
 0.21553812114877735,0.7847706867832175,0.29478833155245154,
 0.9277078274303349,0.43153835992730616,0.7103656020914052,
 0.04061191925719443,0.6438302084264036,0.4083571687116728,0.9493975297947528,
 0.9654018165095848,0.1924612528052102,0.5734729398364169,0.03703014704481278,
 0.5817297916770748,0.6276062442517274,0.053879996347453885,
 0.5919487211650845,0.9382155053531339,0.8290585551989428,0.4411234374734728,
 0.20272726027061927,0.8584479769697138,0.1347018920345856,
 0.020428706990249457,0.39808220428054486,0.5350799223967577,
 0.27011742183942034,0.8744273401502154,0.39647326593507826,
 0.7930765242723119,0.47502778003091894,0.001605811171077054,
 0.44831951407391835,0.0853818569532423,0.7420731797998563,
 0.47717371208736314,0.6478129323435711,0.0778011676834941,0.5908379020586296,
 0.2578576129956318,0.5329217587863716,0.8423847802878245,
 0.027427394514252024,0.8730609558472732,0.957886305191691,
 0.44192471900119257,0.21406977947536499,0.09963521459061297,
 0.565136408777666,0.120593042675186,0.6172199510179524,0.6254591251678985,
 0.3244166939139346,0.045818578829608936,0.6679105774740568,
 0.49074070275401227,0.36087130026178205,0.31735728978561983,
 0.2678428154665984,0.48766641635450014,0.20804926495309028,0.506358646508892,
 0.4975553570609317,0.715801884610985,0.5778202661173609,0.19938328386470594,
 0.9251253790268136,0.9206144706067277,0.8010518512906734,0.7868430305912365,
 0.19010539863951315,0.710878014717986,0.4326013635963709,0.30430906146526515,
 0.766722881498898,0.8013108889835653,0.42107223054983856,0.7173832876065853,
 0.18529672784145734,0.4395391566097502,0.43049789802307437,
 0.07026064249762631,0.027279290167007253,0.5070008392207167,
 0.6574750228473998,0.764895600007927,0.9361476027765516,0.05027616270508184,
 0.7234192100064352,0.27929697246931795,0.4809488821603759,0.8142692300928872,
 0.2164223007423547,0.9026917713455422,0.928222909388759,0.6436372336214455,
 0.011474438838839807,0.2870632174982086,0.1956516827894217,0.794513779070258,
 0.18572715441476906,0.8881975892778919,0.6594915038336497,
 0.36664540004750745,0.5488412676467894,0.8720089665015653,0.383034754744518,
 0.9494288307074052,0.8132486573648148,0.8633155975847261,0.9724466865750341,
 0.19366984337531812,0.46501659970505804,0.11787071428404028,
 0.3340202296078725,0.5564869956952729,0.6432322371552986,0.7160279611526522,
 0.22624633980537356,0.7217062990599605,0.8030549111351162,0.8919499592882186,
 0.8830708850363955,0.378973101787314,0.6737767772393387,0.5230961286891546,
 0.2518722605056125,0.4330505049256853,0.9057999909137112,0.06598968698309116,
 0.9171489313312614,0.4288376148853945,0.011485204657165515,
 0.6787331308728629,0.1917476882466802,0.05852973749110668,0.3149411065117792,
 0.42009987826297446,0.8365854247237205,0.021583340933623796,
 0.44009127765546063,0.8673121388620385,0.3935972060580213,0.6892252508526165,
 0.014505950388955857,0.0755161899677339,0.9594332930700004,0.973914400213117,
 0.17153015752652068,0.6407862482716086,0.1711759613205197,0.7405608295448827,
 0.6226943886333342,0.8391122691519627,0.9366504914846017,0.21964185972147032,
 0.32147079087271546,0.08047666760101801,0.686032772597659,
 0.06686617677630913,0.13325451766844754,0.7083758840481001,
 0.6440244232470623,0.15997885181648397,0.2866497099919961,0.9821398211626711,
 0.08010657809796329,0.0627175890258197,0.19629690518804022,
 0.8290711479914723,0.44353753074953184,0.21546751192241342,
 0.7184609341521828,0.1996851584416348,0.09114172797968612,0.9834040688206231,
 0.8376827610009341,0.6036444995448892,0.009849623890823223,0.586268055738843,
 0.15264219947768276,0.28290471932640193,0.8656220503943823,0.897561233632892,
 0.31540637822490347,0.26187631799025424,0.2086876813235483,0.336561803352185,
 0.31661545102996114,0.47467993653318363,0.550587347968088,0.9688919849802186,
 0.3776840908740712,0.4930056598111692,0.5391429628902272,0.4363328760497427,
 0.3013311428901877,0.7901020733429159,0.51859234791485,0.6386500543034216,
 0.8283410521197039,0.373057750595137,0.9216199078980818,0.19513247225011854,
 0.07338444508389763,0.634233617396687,0.09274444648681412,
 0.35704236365823694,0.8505777291739227,0.03665899325191391,
 0.5250008473114041,0.8821738065629682,0.9825732528794752,0.31709213585422846,
 0.6109406183699242,0.004558080749221816,0.8637886904694938,
 0.7158824256751919,0.044205032317271464,0.5027337657540538,
 0.8714842811251529,0.3602771920558686,0.8817097431285669,0.9013412218497789,
 0.6698684965117124,0.8362165667273408,0.9904776878724412,0.26199253977999604,
 0.1446104668507917,0.8213727338366994,0.48375347294898097,0.7526587106636317,
 0.275264997649047,0.8100910853331437,0.4176894331400335,0.0801444498284195,
 0.6583841604556033,0.48193809971114165,0.5024724590176723,0.9466309731059884,
 0.9284040924321194,0.8106921162445525,0.7121849432798898,0.9516816754587553,
 0.6182868115201656,0.2363668800345935,0.4688084134444894,0.15441447649557116,
 0.6160776546173449,0.788215693626353,0.7394437159842965,0.7013358629808527,
 0.19265817001446095,0.9432226443885161,0.705099532907622,0.6329231396963755,
 0.6306880738726264,0.3661333323507563,0.4495764409086813,0.08594740898138231,
 0.21116006111044217,0.40095517781381085,0.9968610042894299,
 0.8752657078635321,0.5744715792914936,0.9073642192268592,0.4976910751790188,
 0.6132365805940027,0.513941489524281,0.1033613225041432,0.2367694283596714,
 0.5720492994861742,0.8280140890129425,0.7685077166211014,0.2580712510361818,
 0.3533146195943283,0.5809265536509072,0.32688063042384496,0.5256233929481187,
 0.44499027034773164,0.8181822036393938,0.029168865273953015,
 0.686247740537734,0.12163022851569161,0.7595378649640799,0.47725523880071274,
 0.016893965568261615,0.9359898225150507,0.04956913983086142,
 0.11733020831030405,0.9902694228030181,0.533480696556165,0.4005920029284369,
 0.048309552667259115,0.27015746220675385,0.0853746915148148,
 0.010079870957365178,0.5672378629379102,0.5879782918511886,
 0.6204456813095005,0.5622548708085202,0.9868590213919061,0.9438849152461606,
 0.2983122652908978,0.18956877065292155,0.3666467353230368,0.8699929482981255,
 0.12642271823834972,0.2848723567553413,0.35854380743985537,
 0.18378905465551854,0.2851716115012456,0.5900951209649312,0.9487231898559912,
 0.9417914786172525,0.1798322722653637,0.05498789587227515,0.1732968904776241,
 0.9559478248868316,0.876021543660278,0.27184032052818297,0.06293020892040313,
 0.26374575553976565,0.9958999272744976,0.628136744476736,0.12408173519062793,
 0.7971358162016391,0.9945430381159303,0.5802489661186117,0.939619849313783,
 0.8489218747327162,0.6619143984393727,0.2896531408349434,
 0.061976743421670144,0.5330386542430294,0.1208428982965456,
 0.6336940367924221,0.09636502166112415,0.045273889530699885,
 0.5451002417529995,0.07691061846198999,0.87479263589672,0.16389047322850858,
 0.026145949301883964,0.38326942004789744,0.7366794744139873,
 0.19497833913541218,0.7818942596194435,0.4324919602589702,0.6295358997651676,
 0.3720994777689013,0.06678543292890515,0.21904506655564582,
 0.5557257589377996,0.03371134669792697,0.27893858244030767,
 0.41292229575839146,0.24859195077870688,0.674650514908443,0.7756216348291646,
 0.2566857501817288,0.888182482658481,0.3573924000024521,0.46236699979364937,
 0.029225168852827332,0.5718575525542583,0.6231819765642795,
 0.4701633262334248,0.12834438204609777,0.4315699539285953,0.5761712278027149,
 0.911264972753854,0.9184310136637559,0.008037314342250745,
 0.18890620908653732,0.49686008566571316,0.47309221290249504,
 0.8078162592710938,0.5963167428336719,0.40233181872657575,0.5392051355380173,
 0.2504621016528835,0.36534714802376533,0.5353823476385786,0.3836176136810371,
 0.22554546863487357,0.7903544261695121,0.5599336515638627,0.4051368286204037,
 0.549089374886273,0.24200021372232683,0.6734508701254324,0.46167457875051277,
 0.021559453390229355,0.717131918908142,0.2403462078522539,
 0.20897069100631527,0.7369975754269047,0.2906455061682278,
 0.46722085620882925,0.7778974702295572,0.07079694840573192,
 0.1409829761586574,0.19258037231228042,0.5131547400647035,0.5295580656616605,
 0.25554171303919326,0.05267276697135037,0.386698794549202,
 0.020813203731184338,0.4569565472343824,0.29202349688620666,
 0.7362661837484366,0.8679453987202029,0.5280847383324976,0.4864833069400406,
 0.9008288606311086,0.943693025612913,0.5274522835500999,0.6350368203119986,
 0.3135896317852618,0.2900859415397259,0.14863904930421468,
 0.12507574831169166,0.5023243949305247,0.40158351896296507,
 0.6695173942925053,0.19214113592399884,0.8581973668669062,
 0.17231416618243023,0.7829483834681992,0.165978316692297,0.6389159444867722,
 0.5475418207963014,0.23248163250735998,0.3162296368561882,0.1490048279100309,
 0.4195088850015021,0.31940753216268714,0.5997489542550937,0.4487816835667291,
 0.5954478584931046,0.968581363663512,0.1128494065856055,0.3629461582859963,
 0.710763742238993,0.16206975348761776,0.6270689736238065,0.4331233355751858,
 0.896963774406455,0.2978349946997506,0.3732497375403889,0.28604312832162204,
 0.05233385008757785,0.5348104316993026,0.17345366526896622,
 0.5462579656617448,0.6588864305213495,0.7315835459102757,0.86980714953454,
 0.06093224352494031,0.7312082386938634,0.2708661197377684,
 0.34006320358640874,0.06362488148645395,0.5968125164390684,
 0.06832198231903541,0.4187287022611704,0.12572742027934936,
 0.014886196826936482,0.16333934312396492,0.5991336779560876,
 0.9569253035360615,0.39827435659405064,0.02641658245726808,
 0.8818263994692527,0.8070875000537651,0.04792059667971382,0.4695827349297883,
 0.35570249305695534,0.848778321168511,0.23756251987597277,0.7724185419540115,
 0.6114382688742347,0.4718726032367274,0.13624925667118237,0.3914871256977098,
 0.9050916442546915,0.1573317289765117,0.9496426655296007,0.6803697318712372,
 0.5812948797345578,0.670857435725491,0.4067699547744864,0.45402706061626597,
 0.9203910067485701,0.7286498968158627,0.2580809896034306,0.8521304987380877,
 0.5437792850751418,0.8023752031077754,0.5308932691957002,0.8323713992707744,
 0.9670534356450303,0.7810492838394225,0.310638517691209,0.4735147867707298,
 0.5165401200787825,0.35061299000372215,0.37283045540288773,
 0.09176088452691056,0.8401778647062432,0.7726164992891675,
 0.17395738371069647,0.8920372513098416,0.18619547916634693,0.260003104682359,
 0.3866622407142105,0.20631818821458603,0.11554706697192607,
 0.28091940440050955,0.9770426123979841,0.1346206705945724,0.8274961960523551,
 0.6236328560368534,0.6409517189681009,0.07133295885262025,
 0.38221140299500156,0.6713791064146009,0.9448462851562356,0.0877126865103649,
 0.9466119673950216,0.5742682540102413,0.10782553457757027,0.5853746279797707,
 0.802545536291801,0.6169396332771979,0.39671712824236405,0.7418002773267411,
 0.9460798084838722,0.8023350640454081,0.8527404794722718,0.8193756193854782,
 0.9287815148896548,0.65835395953652,0.3612358268070418,0.46969780800384264,
 0.04702774831551537,0.09259920135909483,0.9284779260403012,
 0.5041763275977633,0.0053085521589566476,0.4464470626548981,
 0.9585521855339275,0.40922040333585785,0.07823061587319324,
 0.10240802500403912,0.5893635041221115,0.8723118363735923,0.9783255628430767,
 0.35885165433092125,0.7009116455176116,0.4283497775898152,0.9220708830722117,
 0.08131959819959578,0.07627781920950594,0.990219601866513,0.7823222996958605,
 0.8238530905311533,0.24235916246689326,0.3102221606573812,0.9349461176567062,
 0.8205473571551888,0.9555836910823753,0.28230408379226046,
 0.07568932631972203,0.19956859317335773,0.02071315582270572,
 0.9553878337305499,0.38340970149447673,0.21388882107241403,
 0.6737858271313446,0.41769889916572345,0.7109069811484291,0.2160002648115339,
 0.45287841679274854,0.6387789541112142,0.0850834525381533,0.5768818298815186,
 0.41107127690314305,0.4385244646336871,0.11529096469907818,
 0.7632791668728935,0.2803364965862205,0.4784353510462025,0.40069261018119784,
 0.8179928531351227,0.5148118047585727,0.11089859089069454,0.5844933802628544,
 0.2736970923412958,0.4529070406485366,0.967900432712179,0.41582320021012753,
 0.9282545771211508,0.6759410723971636,0.8724715512380932,0.7232483224470572,
 0.19476679166304445,0.2403200024762533,0.26712068579450343,
 0.7062535857543528,0.5226150293255506,0.9237089259321711,0.6480844701472709,
 0.391289500049648,0.11209579171184347,0.03756277710249434,0.725961640424613,
 0.8019216394539163,0.22932553723510662,0.9674695631002997,0.9076181428570543,
 0.9958122055636179,0.16534363203523694,0.45461192773246784,
 0.9432967753196072,0.38299316344659107,0.373162172923881,0.23187597777055147,
 0.7730055692211623,0.3986630754548638,0.916878103680747,0.7840852502527962,
 0.5680802081650513,0.7650415710108291,0.6082620057653086,0.8221022824972909,
 0.48410898980988004,0.9925105182741001,0.41544320437467863,
 0.41764514910491657,0.5649757619233517,0.22592918510740634,
 0.8923198461581563,0.2071201511678158,0.4794614047247068,0.12368508917461629,
 0.1277190369316159,0.8934528636480993,0.24234125187221178,0.8954506819570116,
 0.8519431029549963,0.994840903296662,0.8231828212407223,0.529522028941279,
 0.798022862160389,0.2951985468764824,0.1702211728809122,0.23913824072486844,
 0.3436759321210383,0.5191380453148602,0.8133013969632452,0.13745645820659624,
 0.5256572671953291,0.5688690409833359,0.714227218576843,0.4287667490982128,
 0.8882684401925321,0.7107033756279098,0.2805920141624173,
 0.019723690123028215,0.7571529845329492,0.5414873582389943,
 0.9776501448914302,0.3778520610963252,0.259886363987584,0.5610550000028663,
 0.0511964749813939,0.4445175372439938,0.6401057185784957,0.3192615263007258,
 0.45103572508430245,0.3802494931713789,0.920615224703659,0.4507432922079425,
 0.5030452172090379,0.4416343263474851,0.40409745194173474,0.4086857219861043,
 0.46635563696657056,0.9991698231132528,0.8962165153737899,
 0.17723269761122706,0.7043464878849405,0.04798381721661438,
 0.9814651980241889,0.5046170061926136,0.06982884592863303,0.7025930618340263,
 0.9948690031850116,0.4238761511546911,0.8777228764574113,0.4273836590924759,
 0.648893994844237,0.5243479062459445,0.6054609367524989,0.6710885788738898,
 0.015411758316973057,0.38920540577607043,0.8786105652341005,
 0.7183514059464268,0.9232889775640281,0.5978462233561539,0.6046656395086337,
 0.5355318638011448,0.13716307057316357,0.01963167698278878,
 0.3423488478643023,0.6361502093509652,0.18197531112773246,0.4687712947092588,
 0.6918541299003427,0.14702295274029165,0.4477427541275524,0.8455238161594668,
 0.8229672235311044,0.6141463864223524,0.6995962480698488,0.776889902782105,
 0.24183290325098383,0.6805920051902759,0.5466414879544012,0.7825799911712179,
 0.43277948502834773,0.5688477356724108,0.9250866308640039,0.8561964232086652,
 0.8581806350702573,0.13485693765738271,0.1563492212368015,0.8475515900259624,
 0.4041302662132137,0.8045849650856017,0.21572815718984584,
 0.22897911353063005,0.3242190779022047,0.5875592087687374,
 0.056222087818135646,0.34385795538399666,0.25702843015945165,
 0.8089002848267411,0.096819199984566,0.49628586081966464,0.16356177496877577,
 0.34882832656692764,0.3800912078819212,0.9029980127800065,
 0.03668539126316139,0.5456030596515716,0.5189473933840657,0.8270278774378325,
 0.3738937962921369,0.4622973431242535,0.26902808700192027,0.7811387616549736,
 0.33550375059277404,0.7883450695076434,0.6991685564177528,0.9515377468627323,
 0.6564072309574174,0.6437398708127788,0.01914911426395205,0.8414642328260981,
 0.9284399513832178,0.9482330912658088,0.29516485586429686,0.7823656676362967,
 0.3080188192878215,0.930812368924389,0.5551825777989283,0.9408478926597869,
 0.692794856214701,0.11837672429703461,0.3819236874785772,0.3264733402471697,
 0.9771233170690482,0.16963330701250023,0.7805074975302946,0.3708868814549616,
 0.1394066133526014,0.11422467808102266,0.09842268361529594,
 0.4206860365773838,0.10855335302014335,0.5771227619137872,
 0.11519705151245752,0.5845548657988331,0.48125567075016473,
 0.49565418363083824,0.5261216132315633,0.8890314288525464,0.6257284940392155,
 0.22425079758797362,0.9225243073838946,0.7497223443283638,0.4818175930088937,
 0.47172153212571977,0.38395976064773574,0.5295560295171181,
 0.3760914610484778,0.23657146750426183,0.8366328210564199,
 0.30577879998663715,0.2600929524308607,0.6756006994604561,0.5595682846995549,
 0.5047616250290774,0.20735498258462925,0.060863805428934215,
 0.5199324785749595,0.5038468501915838,0.6558439864226834,0.28064518321632503,
 0.726908716888341,0.3389835222904226,0.7408995890757256,0.9214032728952936,
 0.6030668243996469,0.33951522401548806,0.028540611209755595,
 0.865627989991235,0.4875112650393528,0.9134918309871194,0.6193094260994827,
 0.8507300620434368,0.20888306367144938,0.5558287916815203,
 0.36529433172721826,0.14114379868053528,0.8335323166018693,
 0.38380693679554034,0.4340831432646023,0.47860930845586,0.6718653711238699,
 0.19909751019359023,0.067149541815682,0.5247443835932276,0.5323173183551917,
 0.7708429046123436,0.03148435311062503,0.12311585548916049,0.821202688395715,
 0.32092626996049833,0.9584377509378352,0.48424759595602485,
 0.8586885137336913,0.2425994735037883,0.10258743462300579,
 0.17432054203343672,0.279578260779625,0.4325107160000846,0.8966487420586055,
 0.8516229331211786,0.008020382938607518,0.2383428587033668,
 0.7711124902994367,0.8390588423987632,0.538748486953669,0.4547137776045268,
 0.9915086654378379,0.09189217517572978,0.8969921385018479,0.2666617482413196,
 0.8125121297038094,0.7868419676874262,0.4139123676602626,0.374983881960821,
 0.4172097250042326,0.9883186130451871,0.9108428289809088,0.8456946161606794,
 0.20881545681069036,0.5768051878270096,0.6020298500954515,
 0.26472494741683816,0.7512821457669002,0.7086506493810103,
 0.008630709481587062,0.36249492316666954,0.47080981185996396,
 0.6958015081039811,0.12114382691130143,0.005321134120887194,
 0.7737026827139638,0.6950645710604052,0.735779175808909,0.30875066522884465,
 0.011352418311284529,0.23870649645820918,0.526912733508881,
 0.9135173295404209,0.22929516719835275,0.16620680568025958,0.693903213026797,
 0.30225871580276786,0.5517805066167046,0.7675760078373708,
 0.23178230587711812,0.27852452758867097,0.44269769372586554,
 0.6853952991485004,0.60677106909908,0.7178401736003108,0.7619176295671785,
 0.04374497109334785,0.4170995359129611,0.6161542733402362,0.7379832389438664,
 0.9627513446148023,0.3835494289270259,0.7901387367894849,0.43887397436021724,
 0.9639773860072112,0.013118284856347917,0.38229636189159844,
 0.5838736267110215,0.42734426600612585,0.6255913946101688,0.9603955742841426,
 0.8045797844195126,0.16679025339009868,0.09745995456816092,
 0.015074841776889225,0.18262435132062138,0.2177582352187113,
 0.1614925544226382,0.1214967626105603,0.8655539346614942,0.7191892362478589,
 0.38872700050062836,0.44839639391148767,0.02892002999393095,
 0.5791312402089028,0.8698680350486199,0.4613070158947028,0.6340004730148671,
 0.7633245772260678,0.7631906789257986,0.24556592381965292,0.990785201881279,
 0.03468725855042443,0.632267297710405,0.06846201257883744,0.9297828715204716,
 0.472543001017662,0.5249412426826185,0.9609965935131867,0.652837166347156,
 0.07442630306635845,0.7514566851897562,0.005459494487590089,
 0.9565443674547304,0.9360777212899346,0.19799001740171185,0.5502613343022648,
 0.798779118827474,0.5902121866754158,0.4475943921148877,0.7312157537190082,
 0.3399729669371462,0.37948087279206033,0.08454566474211256,
 0.1723040976311686,0.3683515919878866,0.29464209136939246,0.1032185325824283,
 0.8687561609503077,0.22944451717929837,0.4229493085189866,0.3949827500780514,
 0.33639438000884825,0.902809634505477,0.07554474484679763,0.4953145960355142,
 0.8780391686948311,0.8129589783953766,0.39139430691826016,0.7465130436990729,
 0.8824367444903655,0.9666873029678198,0.6041109024260369,0.25125763355449715,
 0.25874726301608986,0.38055451965197185,0.22286589954819136,
 0.8727490323687784,0.5122406067450693,0.3270673470094342,0.46193644459229555,
 0.5739673657888004,0.8049070474084808,0.809365230420326,0.8682101234185214,
 0.9064020105144895,0.0642231046666758,0.8107077642557866,0.5149499249979412,
 0.25558029674963145,0.1472028523663268,0.29389091139631873,
 0.06885563951438822,0.4835905317885192,0.1995258570088847,0.3818120395591118,
 0.7460739062685819,0.5589667320859998,0.5979746164011078,0.1870984443776692,
 0.7979750022540879,0.8715767436960123,0.3397646280938641,0.3543126719646502,
 0.42295276268957893,0.06952485237083861,0.5248465237370435,
 0.7104682226728993,0.39718189154641936,0.28583130338263407,
 0.4948372903708921,0.5925486694405272,0.9456512763652066,0.6420095429930828,
 0.414120394585463,0.180195823362868,0.8183897046159038,0.8102693928236517,
 0.6664647348887005,0.12322043080150036,0.34182207050204816,
 0.5272519698903705,0.9688053374742104,0.2558529262812226,0.9249635208553054,
 0.0327174508298832,0.4593169034843211,0.5490298549398962,0.5566527223693984,
 0.20526313318872103,0.1667720554755221,0.2658130871365594,0.7745676244638267,
 0.40303046351931415,0.7354644136748235,0.05204784231145687,
 0.8369779390377945,0.20835513806494388,0.9312167140122289,0.8199218221634925,
 0.23870218983181912,0.9520436839349334,0.9054149588724603,0.5466446096502985,
 0.9736338647317664,0.8185824076129654,0.34222361353634634,0.2918429241432652,
 0.28200172978491655,0.8005966508487796,0.9639995087905384,0.6412011364576907,
 0.46152900902026317,0.8417953216482408,0.6722557599463636,0.7325534205370833,
 0.34754463315186646,0.6190306254435185,0.9786930980262166,0.1451453052677395,
 0.9174194011388274,0.224833249850115,0.4600608215994092,0.8550555739220602,
 0.02516573605754946,0.4602691258088605,0.24642826226226133,
 0.3727116668895707,0.5537596880779664,0.11533162485190873,0.8394353959095803,
 0.9723748502934235,0.6824168579524369,0.304168332841395,0.43662990324085493,
 0.05104565452066068,0.5852330163698917,0.7411786144958835,
 0.28088486666789614,0.41703788069091363,0.11314410260579955,
 0.5235601921891258,0.7051432747440042,0.47367270621231483,
 0.20991525433374325,0.1661986513735878,0.7739344598624829,0.71843790291086,
 0.2769281014217232,0.9618551999740224,0.7469302564945695,0.0756659780151634,
 0.5796049684284001,0.9782642500889204,0.9919423775965048,0.6192952217402069,
 0.5875260592371908,0.028159399853870837,0.17433778184932747,
 0.6263249928926067,0.9256591224961378,0.7143883564311043,0.8989950555436097,
 0.0687379304072595,0.858438378903394,0.12591920126047396,0.07717610748531356,
 0.11309855190141249,0.5103250387480968,0.20806681197289012,
 0.5559014457809643,0.6054478697225459,0.33168816289207326,0.8305319019830193,
 0.20700440577559032,0.7258934848334795,0.6413157679250255,
 0.16631332768623974,0.7009916898223413,0.3607141333954823,0.2527237534007687,
 0.6361377652750436,0.8668306272098173,0.043691447524412674,
 0.6572764221500059,0.8248539007247401,0.07138432375009451,
 0.10478778172421765,0.7419372744957152,0.6539992781244447,0.4006476754232152,
 0.056974871005229466,0.8227994469452522,0.8933495255260664,
 0.04769525345411896,0.14582683452159684,0.9649816563316146,
 0.13314390186703007,0.23262367617074142,0.40073121740378437,
 0.22015656624200686,0.03999920020914227,0.2921369046687129,
 0.4691976626676254,0.6758307607974636,0.8997659888692754,0.6615459370406973,
 0.7890506370964917,0.047650120264194795,0.8250592312903439,
 0.36452460509317464,0.05134971974837743,0.3162526589840845,
 0.5078751734988532,0.4681906398268848,0.5724339118055138,0.8748544606102319,
 0.600205864701813,0.04591886866534789,0.746832648882083,0.8273941172666952,
 0.32246106710686107,0.21764226132206765,0.5734098175243462,
 0.7313320476461874,0.9089638925692294,0.08878718241830064,0.6121619432479715,
 0.5680442270370979,0.27124854493444417,0.8929297789626266,0.9112586897946333,
 0.03409517328568046,0.1973656971789487,0.6068053057958597,
 0.24438839458519213,0.5788658082310792,0.3880941890939609,0.2794090033683603,
 0.7057295609153433,0.053812258920523215,0.8288899996227309,0.784610081431985,
 0.5361793948469389,0.9998988102370949,0.31699466242299834,0.9158782265357177,
 0.22515252771656669,0.7254675451884478,0.4387921675236991,0.9118013386359195,
 0.14880030286918478,0.07202211576529982,0.7803639268699307,
 0.09661375374116254,0.30355435803678454,0.4929047740627658,
 0.3345861164473547,0.8247026206038865,0.8069556103617406,0.36528000023394336,
 0.9086699624058567,0.45986824649582714,0.32138631577920773,
 0.9618795833399596,0.6203820871892892,0.11014767162754846,0.785424091683758,
 0.7447078769736039,0.43548449077964224,0.016825363428783047,
 0.09326541966712265,0.1034484983020385,0.13178136120702488,
 0.8624593552469054,0.6301673095539113,0.38904565377841305,0.6921667450945248,
 0.919578424109592,0.18856291723773044,0.28490579042994213,0.2665855602849433,
 0.196628540122746,0.824810445544489,0.5815434302757116,0.4119711971822677,
 0.790727433273117,0.220994438417081,0.5171876123752581,0.7289002913716329,
 0.48898692418479317,0.45843199511580635,0.7308833803952258,
 0.05505165769047826,0.855270450538559,0.20159964274698905,0.9358898122165835,
 0.9180046220852308,0.5785886783127686,0.9304083709649684,0.07381854029104118,
 0.46809120824259576,0.8221002419171923,0.7127160661554746,0.7268784186848497,
 0.8886518459459692,0.14689198142401028,0.4383426055353249,0.9778071201605215,
 0.40632884733944796,0.9189906612215678,0.005830920661894057,
 0.5348993049926482,0.7384986518079338,0.26111816867284854,0.7786416605064868,
 0.38986576735859,0.23738857360295385,0.4084267320627212,0.4812241639949796,
 0.4263499328208915,0.7117713671202327,0.989501010059215,0.672004357343932,
 0.588347788511202,0.18856257981763236,0.92159570271675,0.9953404697177223,
 0.7866733449149288,0.3091615274686357,0.1668416650641027,0.3677497590881027,
 0.5991022987765217,0.8020574180102864,0.16960437352023872,
 0.30141710084349915,0.048715062588551516,0.4611755983215129,
 0.3633461826805967,0.7762133625487304,0.8234368262561862,0.7408626610590834,
 0.8051227433245365,0.9861246841349831,0.4812735975725767,0.4121107136451214,
 0.434477954616306,0.475971599680438,0.6550980230998718,0.9573603829276561,
 0.5508882438692623,0.2620236826235778,0.29555567170742025,0.3463559414253198,
 0.5916186649491267,0.33900356813586263,0.357396454720426,0.14763400905943747,
 0.3403300683364491,0.1380862163417611,0.8455179519256574,0.09283587156981299,
 0.38228993447567294,0.8617029540185256,0.7611673815954891,0.1675828122827433,
 0.31682823323114695,0.47963346501664184,0.8503482920972925,
 0.8524880224523161,0.33284791408904246,0.7501455369326189,0.842223107990046,
 0.11671634825647814,0.7736140692155352,0.16480459936932357,0.031905323635398,
 0.5921857566391868,0.2514494071017812,0.5059782214178968,0.45411216119045794,
 0.8258441045630492,0.763922760128745,0.8576303689279293,0.4037377124704232,
 0.8875906136723338,0.5746239300119175,0.8919141533158573,0.954127853169163,
 0.504297368218183,0.7171175979962513,0.9055349060355057,0.41912403128003617,
 0.108888496058267,0.6863503409956401,0.61363390069704,0.8297279944248249,
 0.5138698700699821,0.774914083978692,0.4147517012777755,0.02336913475647051,
 0.5132233827452524,0.21463664094527424,0.7226639455463085,0.5345989752285867,
 0.6199962594382904,0.007508667357044918,0.5496985747195103,0.530049407972034,
 0.513538967420477,0.6935584288433944,0.5599313256389322,0.7196766870623224,
 0.7063879096086051,0.726265084019965,0.6707675180840201,0.1656978934768426,
 0.9547037790000161,0.7244185646885709,0.1177348004978157,0.6017839317759374,
 0.8060446685083712,0.5352520592092905,0.8703645978280787,0.8573379203528855,
 0.7917644446960677,0.08885909797892078,0.7771269868509024,0.5814096974476276,
 0.49257249966921157,0.30594873400669964,0.2325295087026279,0.374872352964586,
 0.779578634250502,0.340741691910768,0.6716800487089587,0.08004830148336972,
 0.7818145458380074,0.02579755913755799,0.8513795161472224,0.3238477117173333,
 0.3513972894248991,0.1462774161179079,0.3927875252797317,0.16858632371512627,
 0.8297216901483285,0.27141119468085384,0.5384087554639116,
 0.08713510306074423,0.6152524829889893,0.446986908403156,0.5983910293740891,
 0.43572955416059755,0.043766344682513836,0.8082911212569064,
 0.10887125399685005,0.2344884425799606,0.7712576718551792,0.8091586272121046,
 0.9086019131979097,0.6105513996774871,0.14115555087215426,0.7743822226370067,
 0.28839244894799165,0.22967081516486254,0.36692775788775145,
 0.5980858680850393,0.38281467088129584,0.2357445276778969,
 0.46556744733322486,0.44626748742215905,0.9615280299636104,
 0.029374326758561886,0.30939355251430023,0.9698044424940342,
 0.6064275826004922,0.24856627294961664,0.04886054226506764,0.858825869846968,
 0.7115333507481365,0.7102839412049675,0.455675636194534,0.25126407288884056,
 0.06843433852542369,0.6644582665748886,0.7101498835046418,
 0.47337077220604984,0.3065168335058476,0.6372720262970614,
 0.12190632223133124,0.13525865928735226,0.8342559228967543,
 0.2821547751496222,0.32226846769174666,0.7647356776526752,0.9164457584912479,
 0.2024891633056456,0.06820703712030418,0.3509245281003729,0.3375013010129704,
 0.19061107237372132,0.9369720066368612,0.670633258255946,0.6422781018487453,
 0.7888625415744612,0.7928051854498863,0.29562449297138627,0.7920723960918266,
 0.6820966935072694,0.6115060832380971,0.14612894247603325,0.7846553421392766,
 0.8049450384617504,0.6656963155467767,0.3138963517409904,0.4917969893369861,
 0.7978529965352348,0.6339800160681777,0.007807348235207234,
 0.6778479046350409,0.07693151721657643,0.9445156552989739,0.3492115355788665,
 0.13049536812327822,0.6954656269831205,0.37139763473673093,
 0.6445505751572748,0.8605995259052738,0.4132664786952944,0.30367135645481724,
 0.4571536607301804,0.21884334246062664,0.5344669202704968,0.5958831385132841,
 0.7756771086728316,0.11694896199340199,0.9385855755864695,
 0.15400244519686868,0.9027617237612909,0.622485235438269,0.5946218880295259,
 0.21748760112526855,0.5910518112225781,0.30194954591214884,
 0.09841439507704142,0.9651723834455993,0.7736078619079735,0.6006824604614145,
 0.14686245780439067,0.7345768624374251,0.17136361268041478,
 0.18934640915265566,0.4576234440936048,0.5399997191314406,0.6933530999506238,
 0.5168119650182859,0.9458422359051567,0.07803507923002484,0.3597273683120036,
 0.8587161953011765,0.976308900983226,0.24457199903282234,0.3747798059909815,
 0.6001379478080526,0.27259208568867743,0.6101600854664946,0.3101850431488824,
 0.5984936573798971,0.7728667257087461,0.802783700349029,0.5677804715286819,
 0.14671428606715897,0.1600005044757068,0.22264342749524668,
 0.19513955103048986,0.9338101058311252,0.4052109745458896,0.612058060808699,
 0.7094454425104896,0.724604684355686,0.590418261614542,0.824577123434609,
 0.9835204495696623,0.6875899532272927,0.18775796640098308,0.9519346141256544,
 0.7446891542455972,0.11615076748136399,0.9018896273204696,
 0.33692279555449556,0.3568443438380339,0.5638024688791934,
 0.42509279142098344,0.4555244894327072,0.29227636812674473,0.752594061139535,
 0.9591198175131002,0.15987845904773645,0.4341883538779764,0.9459112954754787,
 0.5635717591301563,0.4994649702316796,0.4393697625915798,0.892225748146299,
 0.2909843193845745,0.648769595490617,0.28209650818446874,0.2955740029464955,
 0.8160095190346921,0.5574068834378628,0.7923760653706029,0.5659159363315268,
 0.8790062076178091,0.38153268734514467,0.5966407513964936,0.5527669617481918,
 0.841537816262454,0.7628627425085923,0.6110099131896459,0.9442546880451422,
 0.7842573599623652,0.8672432903758556,0.004212288412696941,
 0.19639172061500965,0.6621454448296398,0.043467497823186596,
 0.7426605663719914,0.9758865840399131,0.11489288651486018,
 0.40323923984040055,0.057769560254847474,0.6459687435804515,
 0.024426780228412714,0.5893133448971059,0.9124361216080978,
 0.07536874708516916,0.4210338765520687,0.32226267825735366,
 0.22652784502892698,0.2784951110316245,0.007814708846247465,
 0.8390984452437853,0.802111201405707,0.16093986696434792,0.08981057604327092,
 0.6225374623168982,0.7414306397006782,0.17549320894887332,0.3259145392563163,
 0.8644095703918236,0.5218675095007222,0.8913007918188603,0.512237456351752,
 0.07346667146641184,0.3480324611852833,0.7522316476289899,
 0.09793969438841899,0.5272351886842185,0.1359795233676636,
 0.18227987162225245,0.9570135863973901,0.4627988680737579,0.5387701113851776,
 0.03836871969328004,0.019604782875778493,0.8131964124527616,
 0.2230459243551589,0.7671032776365506,0.5308735611076099,0.31897023968460736,
 0.5034328134397723,0.18110679463760815,0.3514024326104861,0.8491997141004033,
 0.933628983856135,0.3218265530568044,0.680818648691893,0.36595813608249483,
 0.35527874766681933,0.45948826898473016,0.20804852919075156,
 0.7012913097049569,0.5592550216839572,0.6348890843648666,0.9012396220939816,
 0.13810971682432716,0.06802174867943267,0.10340671803346435,
 0.09650630645094216,0.9148158337827176,0.02584677668694757,
 0.6714918498483458,0.11887082217321276,0.4248140574546927,
 0.04257558967943431,0.40538321606287053,0.5087843073465458,
 0.06694863578023902,0.47966177651474595,0.7426734410189948,
 0.3231006551524117,0.5960524121615008,0.3269850098323027,0.9050313551842936,
 0.21700036737518702,0.784320829526328,0.210093013688064,0.6601583934071714,
 0.38555587699013993,0.3943457316937279,0.8170934714202692,0.3838279570432719,
 0.24606481669976676,0.8157096645451068,0.5244244379943365,0.902624612344661,
 0.03423013674150244,0.3956093142983933,0.9117337067462319,
 0.07604674384710541,0.8533957653836163,0.6154419122233326,
 0.24943899526892233,0.22276250808473874,0.18576854916972985,
 0.2998432852742161,0.4003694937280359,0.3456074408197606,0.5122661876361808,
 0.492495070396053,0.28497204994992154,0.45995468703798226,
 0.47480593354136713,0.6954602577493176,0.7390894648825267,0.2524631529256024,
 0.8525497192473941,0.4887084489555932,0.10791890139579075,0.7590209424267969,
 0.2626264816916355,0.8548152323168138,0.4269916761111535,0.34052443561058177,
 0.6333608830389679,0.6232283027099833,0.5148292001127068,0.15115946245072553,
 0.24413001883089203,0.10737810377095247,0.6145864824941548,
 0.28087990540155217,0.39568608507879455,0.6647525756812669,
 0.7847200749373628,0.736264025601686,0.14540795647030813,0.5245525935038835,
 0.7606538144968289,0.3735538317297037,0.22670828253380992,0.725572553316865,
 0.23859371918212324,0.7351685333897638,0.2740546264866275,
 0.028125070056083223,0.9509992253588714,0.12374954373359459,
 0.1258183977148719,0.8551455612174291,0.01830915498626684,0.9985764141031006,
 0.5425183089071702,0.90920804602111,0.9506703257362638,0.18548597413471313,
 0.6375933892613267,0.870651984123304,0.789494570395867,0.19113408717180336,
 0.645962750415995,0.8011869932578956,0.28955356652577213,0.5777478620452299,
 0.3812505269403229,0.49464451978151003,0.08798357376681432,
 0.5568275539151959,0.28106892224074065,0.8870057778504277,0.7611802822290199,
 0.30039032114328224,0.1431235750979457,0.20227300979385388,
 0.23630282858507046,0.8558145553899192,0.15864729502118913,
 0.1276623893998058,0.13395070080645122,0.1923117193828361,0.1606888938246459,
 0.8807074749085936,0.7827375998792163,0.1467599584605922,0.6286059610158771,
 0.5084648719567756,0.3701952379832679,0.47059736908326577,0.938724632384554,
 0.6511139257155788,0.04525245452123261,0.2361418249657431,
 0.49552068407536776,0.9399527331666031,0.07122785350317518,
 0.6384758310212033,0.07163018349052908,0.21840625644056555,
 0.1484165527305451,0.09632378128355956,0.3227819219617156,0.5010008122711309,
 0.5044358646049548,0.02202095252313374,0.0462222511500352,0.8036446287297241,
 0.5872100735178535,0.7821202261031954,0.027350225720032317,
 0.0515881197281397,0.22512955927477585,0.5959455441180079,
 0.28349154729112946,0.017179904592684925,0.9711775240719577,
 0.8039691048842057,0.5511018721549493,0.48122141276594466,0.7119781289094644,
 0.13251726063718727,0.8294202746668544,0.5298952440601897,
 0.04750734079289132,0.4793977907882697,0.7049072597151008,0.924906550846946,
 0.3865168726375646,0.088458407595423,0.5001827264800127,0.6164851357829924,
 0.4105281171998253,0.47384916509012565,0.7271386387382313,0.8900226577651965,
 0.5741776608001278,0.822257803755792,0.20555384442214708,0.3494043469566448,
 0.9471573081861707,0.05666085633508822,0.2442195597712029,
 0.25970955459294987,0.6906767599882905,0.8860128554217402,
 0.056910514584940564,0.6132530151882992,0.9013839422201155,
 0.18256642284566094,0.46403982672511734,0.9042544646423503,
 0.07420328585728342,0.443667350729542,0.12311214189340824,0.3382793463042916,
 0.9857333364645026,0.03256046391572742,0.8019746251046352,
 0.07011544814250636,0.7418823804404742,0.03251791741699794,
 0.8895314034771965,0.7215089664266806,0.9956994071000742,0.4834570126677432,
 0.7270608910572554,0.7171393503704329,0.5525957258025931,0.3592151755489221,
 0.4311813928396847,0.9189036348252919,0.3378495966378563,0.29663526926040407,
 0.916467759239622,0.17506815477945015,0.14839812664803742,0.7841737490158184,
 0.7032988006548054,0.7290635776298546,0.9217647090407634,
 0.039081853520404164,0.39319788051062265,0.32904946629829657,
 0.8610261159532452,0.0014436303254008465,0.5916869374823694,
 0.6723286596797192,0.7243463849367408,0.5961425467626873,0.02509163942285797,
 0.3071836301093135,0.2897070511190821,0.7294548580673956,0.6706286094515601,
 0.6026645578954546,0.5207098070649191,0.2981274624297714,0.1050397290262185,
 0.8014746292396995,0.5867135995475385,0.0426197034111615,0.7429814756541987,
 0.9002515672435026,0.06446798144881105,0.48107905271081464,
 0.6238640685739942,0.8297541897369108,0.43475480365958774,
 0.013890160995067014,0.5016011730790493,0.6274268380380141,
 0.28000680143081613,0.6514858587572147,0.6692351981646892,0.4003615452471476,
 0.9087557933308987,0.07859258816717785,0.5693886793819873,0.7422011015029482,
 0.5175094639744963,0.0976066428196879,0.17918208398411306,
 0.41845640732584977,0.4092348058176466,0.8842365815353626,
 0.28809988832147926,0.8218175569934225,0.23006027286083197,
 0.6459893581729158,0.9585218281182086,0.6714868681441835,0.19576016122907436,
 0.7108730052425742,0.5400727269064902,0.584415277648074,0.34207150704209077,
 0.732429199899468,0.5533197258847621,0.13715568129136446,
 0.005264355383243102,0.17153037820916872,0.7898896262731241,
 0.5449357132911052,0.03670963167941654,0.6209538769466154,
 0.04518296913700193,0.6945894699273915,0.5076318828052256,
 0.21839905246665803,0.625947713952324,0.92917472053713,0.2685558384062178,
 0.7824744335104172,0.16490918883960548,0.471530483296112,0.5371585392012709,
 0.30295884791430516,0.6069693039199815,0.12506737792812184,
 0.8687564639336808,0.06222059537181679,0.4671674427643988,
 0.22126101407158894,0.4336956456292219,0.08863142000277813,
 0.8387932912217357,0.6868959768150743,0.6826917599707171,0.8751122441378492,
 0.02524236556016035,0.7167723287986886,0.30179348461192124,
 0.10557917695519603,0.7247130321526527,0.512311526867755,0.6070741502733947,
 0.718835083017526,0.9181882482770365,0.7546848768707659,0.4043355532584004,
 0.8267051617214819,0.03496063457715026,0.49604088818975944,0.24757788875982,
 0.987401524462299,0.02423420618760619,0.12183701654990298,0.8345475035741056,
 0.20148726626787783,0.19818566522262304,0.2461584053096968,
 0.5776492977871943,0.4456137175979218,0.3784667249828474,0.08267798927071934,
 0.007533550509804421,0.6246706435199164,0.33991796832124765,
 0.7802513488277054,0.6913843750755599,0.817862755644397,0.7126279511283714,
 0.2333491349571234,0.21226052209098512,0.77283839842959,0.21058060083933394,
 0.6837739668513154,0.9467502702191819,0.22786807580124813,
 0.30424426250116476,0.7017542010594325,0.28111775007361495,
 0.8799843954988524,0.40692234687052653,0.8736134876963306,0.9933237284226716,
 0.18430839404627086,0.6380250807397958,0.7554135170224181,
 0.30841895140905984,0.04521528254165119,0.02716055599502254,
 0.21299519167755454,0.1647789183591727,0.20906769518306467,
 0.07010103118034405,0.22014538677200013,0.7255142984657583,
 0.23733056516574746,0.22918181708974106,0.09567560495769811,
 0.3313699494185195,0.19625026380363986,0.9262348593810639,
 0.27430946627897945,0.8414816188213955,0.21283615276393697,
 0.5406370974325313,0.7429769574160816,0.40107680828261627,0.9035084199514933,
 0.5592695450772167,0.26919449345040714,0.17670895301782286,
 0.2443178293319006,0.15565611583311678,0.18483049522470685,
 0.6398791148166415,0.9502233981784288,0.5517703939731698,0.8846921612881661,
 0.7339742667313004,0.48692752255591576,0.05711799321018318,
 0.6121454650501577,0.9675149223513422,0.2736859345638569,0.5260430712019029,
 0.6466713618187598,0.8325616355523403,0.8318975118941755,0.11494990447914644,
 0.07246169633999633,0.7157762064686835,0.020366977060275016,
 0.2796910130800516,0.6223056095936045,0.826674234680127,0.42095650597690004,
 0.8174687450300685,0.4898150550522491,0.5823705404953822,0.5626000530133142,
 0.13992251842866454,0.968186005803195,0.5455178016902819,0.5279464805813326,
 0.38869799562807494,0.047195666262906166,0.42606187753302016,
 0.7757486340187661,0.17745315867116473,0.371373210587128,0.8724532953448303,
 0.5525164241245512,0.4422057406875213,0.8833094556259743,0.39865817514308444,
 0.9453591702653008,0.49967713073151465,0.7861536298508781,
 0.32059645838558615,0.06814759452081154,0.028902411094670732,
 0.9143434981414952,0.3552352721923858,0.07099364998196234,
 0.21953065920720083,0.9711948093165412,0.5042535372442098,
 0.18964279346465962,0.4238486569445159,0.7923701602384945,0.1540337409635918,
 0.5201050196619684,0.9904855542500277,0.37425749634349437,0.6296857521866907,
 0.862786422717704,0.4162997516715551,0.11123717901841867,0.17237571276125452,
 0.37691326212192233,0.5600506431700988,0.8777926445270542,
 0.43855425574803186,0.8123180830654049,0.6699201520200228,
 0.25608442100644435,0.18638709236970707,0.6136257806695775,
 0.09517646773070854,0.28000038148700024,0.6606557928133499,
 0.5582513602911142,0.9455180030229499,0.8966432476906139,0.6265050961053584,
 0.2944534239178488,0.33554737176328875,0.10276091573456636,
 0.9924753639531846,0.5990115841037075,0.11344554175167088,0.445916828184335,
 0.9382468375890184,0.8058456276531029,0.04110517363990207,0.8836867733835947,
 0.014862352231385989,0.9927089540018974,0.7909451352350293,
 0.08601416572088116,0.6581303953977646,0.6350192195483673,0.4041188363890936,
 0.8949569916484574,0.4510514571533577,0.4543522610995192,0.6764660656939069,
 0.384060835561846,0.3057913828686526,0.7183668371585643,0.7566998360248026,
 0.3217834354023352,0.19715556801156886,0.7509512116154251,0.754853374222455,
 0.7467777498209003,0.531133332383764,0.4714139177517026,0.7220804448173256,
 0.4889700505603045,0.49201142665490827,0.36593609666391835,
 0.6362404055627151,0.2708038925370756,0.041698043093895465,
 0.5189613259742216,0.48479896918872933,0.7650756709168626,0.9848539035780518,
 0.8345802857677513,0.0848147420937635,0.18018751576559888,0.6723309831533473,
 0.6336132385088689,0.5690366343494975,0.02006601671700292,0.4641847611851786,
 0.7491237150454725,0.10711785876107471,0.14896180021912941,
 0.06524336576118073,0.6757446913560948,0.8619837628593604,
 0.04219042990730326,0.025740243412499808,0.13735311282872764,
 0.7581860617642746,0.6296925105670381,0.8573373741031298,0.5957114824769857,
 0.6287485427125261,0.7088092857470754,0.6721023938089999,0.8965798177360862,
 0.6083828833432181,0.7302001949068389,0.30042406942533995,0.0897622210111958,
 0.48606679746622006,0.30067758670605893,0.8693210955563299,
 0.7863170960921742,0.454387434092671,0.22249785348401363,0.2394884468062516,
 0.7798869229372999,0.9542806732465976,0.8161520991299833,0.9491007730992433,
 0.6196610903312868,0.1483021702992101,0.7429577468144143,0.5949811760459096,
 0.8793165354996362,0.750715700246292,0.2593160652020702,0.8268151449831244,
 0.4721729513832944,0.8422802297057623,0.30234796201731784,0.7785277842992606,
 0.06166125056085048,0.6084116764944282,0.4301438451166646,0.7336629758879454,
 0.5200691502711714,0.5590194478002691,0.08355082374379974,0.5721021771611481,
 0.12646183217560325,0.7260454683979575,0.06702505715116114,
 0.6243553779153097,0.06496220176292322,0.2922592373976032,0.8665726307819979,
 0.7104334915840863,0.3280638483546232,0.30516002589329927,0.762139692187002,
 3.4108059760062925e-4,0.9047910065872691,0.13668794128494433,
 0.6329171882360786,0.634571278936332,0.683865585459358,0.5993850680246194,
 0.0728994453483276,0.7408982436112668,0.8283269904272126,0.37277463713414805,
 0.12490793054003335,0.2758267793640323,0.8791797275823287,0.9929717841088742,
 0.14031455357543865,0.5068508832643288,0.26387426396409663,
 0.08877901588289794,0.6035633337176395,0.2950627864940446,
 0.10932638343255918,0.7792296507114573,0.9505029948769085,
 0.03938019274566751,0.8053632809831958,0.7198466696459286,0.6368864271273982,
 0.7340755347763896,0.4037132609373557,0.7721320373612471,
 0.012702652671534254,0.6619024345107496,0.865602353455758,0.9601763029974817,
 0.5492216761416355,0.16935222301540676,0.026595605762308017,
 0.8262963651901819,0.9354780872734285,0.4135063059286609,0.5332907181527715,
 0.4015955201647776,0.42025164833239526,0.868878011859045,0.5255803113754806,
 0.9521188939192717,0.004253350179764248,0.5840633068609765,
 0.5322959758612111,0.45170071882292806,0.2706516901018752,
 0.03216118752753805,0.06398136361160123,0.7718068469042149,
 0.06797795256225359,0.6436142264812307,0.10482609054822323,
 0.5926347295167254,0.09086350220608996,0.2608467966887569,0.6968128667956465,
 0.41302775161575966,0.5664891998776638,0.3911526842939377,0.8007710276376813,
 0.4147793441933023,0.5869502924253979,0.2766941670362366,0.9520011725249236,
 0.043426993012211756,0.1479955842122186,0.9615085895001089,0.59554717077211,
 0.08760984595021337,0.17345712617613318,0.5081302454879156,
 0.6875536674344125,0.21926008468189262,0.42576918001136277,
 0.6269093702381257,0.1437459781334618,0.3511124011955977,0.7933701159108063,
 0.04258018720472556,0.42978394081559856,0.6246013729303813,
 0.9711049953759181,0.9754123064383918,0.2553579149191274,0.9499440992356374,
 0.7666450015207046,0.30578585499297595,0.6652997521133718,0.2734083302293149,
 0.5537288314309166,0.06936378654774678,0.4917900398710461,
 0.17204840416373146,0.8224223899086244,0.317295175272845,0.3058252407726042,
 0.6887613269543201,0.8119097451768641,0.8745451738692308,0.05298232269529968,
 0.12261638278280285,0.036368145979636446,0.31433396560173177,
 0.43122734828120723,0.49485291852440916,0.051762246838588766,
 0.6402741943220136,0.7342746839539576,0.3825419647007552,0.7928862169351222,
 0.15198147636136905,0.8965083701314224,0.8591857448922309,0.1843409829692888,
 0.521463437202124,0.3111498314616241,0.3302556951179647,0.3641556401741517,
 0.10405398920987163,0.21700035627694936,0.6193635714452387,
 0.8471808519628291,0.11032338402480202,0.40375496336279415,
 0.2207419407901674,0.8562387941601665,0.5373090452789085,
 0.017800128275693106,0.08701874062031556,0.6926025072228159,
 0.6580845381884274,0.8857593430202381,0.6363622973972224,0.931392160974576,
 0.3520462751737342,0.836469465129717,0.3459956638046947,0.8707523697024536,
 0.6838654464335353,0.8395823428913523,0.9268246327551027,
 0.007313370522347706,0.8814991844701171,0.014669563207066139,
 0.6031317331993324,0.27392035038225526,0.22683105020899252,
 0.24561421290378993,0.25459437795464046,0.5001509176929486,
 0.7291983745012894,0.056133566617986186,0.7329948129376556,
 0.24062863905367693,0.33592751974226165,0.6362884822331663,
 0.7227595273455872,0.9589919739988119,0.8408541451191243,0.9776604267544384,
 0.648320244973976,0.8213234125394733,0.929160041516795,0.9139912571500484,
 0.6351510908219224,0.6140817889960338,0.23879419054849627,0.5863863157604415,
 0.6216666475607174,0.34346917796248055,0.3856546924676093,0.6303544701112664,
 0.30170565273777905,0.6479810270595993,0.3747278563476093,
 0.10508098208545424,0.0383803970454607,0.34368589480154266,
 0.6045750514727918,0.5082683486930322,0.7646409661959496,0.8012195777104606,
 0.12237323734502992,0.9447631357956496,0.5333079509221036,
 0.13281533532583012,0.5544446791888734,0.9235986246072926,0.8241990943051931,
 0.04748839432211538,0.052093980337103996,0.7277717220553129,
 0.3183753277609932,0.2278230638102372,0.15899042910729655,0.5479640999676685,
 0.37083844892053475,0.16283981775284273,0.6222284651784932,
 0.02192079401382252,0.14827957027706729,0.7402659684899711,
 0.2799662478815673,0.1407538221361615,0.4547136965357734,0.6446240209613403,
 0.0632144192340901,0.09037968415376674,0.05091014568201879,0.285859876361797,
 0.3266813995063431,0.6554130962555572,0.2623808519582035,0.7522770025863016,
 0.6537082507438169,0.7574476954173494,0.7734836531674615,0.36861532146950415,
 0.7191744609750086,0.034708697901880914,0.4162801875084532,
 0.5167789338650259,0.4354230947926374,0.14992256637849588,0.4748294586983828,
 0.6358180793065057,0.9343460339874097,0.5944514470827513,0.25977587997404905,
 0.0306754925169912,0.5482713110623626,0.18872863159253872,0.5997063965709325,
 0.6692742110949341,0.21000753732773425,0.4071074185879838,0.905659659136301,
 0.09888074433051608,0.9223121880571344,0.9789437440865922,0.8467643494246099,
 0.07656161197730449,0.5251460771850093,0.21356969537262893,
 0.1740218154126857,0.4102266687923448,0.36166025966020743,0.9696177683331475,
 0.8216150346274931,0.6986700087008858,0.647982583368449,0.4332122611613375,
 0.371508443221097,0.9535645095160195,0.4094493090796405,0.31335145648758245,
 0.8057363689709981,0.7259970660602023,0.312088783274049,0.2198437359570944,
 0.9674146897516163,0.3453730524088523,0.2445003519485337,0.11628649252168144,
 0.6119203361194144,0.1991658861136284,0.9961670284768465,0.3206680494887326,
 0.8307541408173518,0.959230552237335,0.9915265750352978,0.4262616670355397,
 0.8004989604482807,0.5662104970633237,0.5528258535213851,0.08166470643834911,
 0.5458828501297078,0.8364390741799995,0.6009513146392824,0.6863735825921643,
 0.2388078863218459,0.8863175425528929,0.922040085618301,0.8941048478352587,
 0.7109221964172106,0.10092297081649093,0.9933334900247303,0.7588597612842636,
 0.2507762693461667,0.7141404959310235,0.3133790950316514,0.7867289055927001,
 0.17253682163552964,0.04672551431189376,0.34908079007800374,
 0.5507492643983178,0.9767280224263674,0.4734076901879726,0.06988540232105334,
 0.055013870419100885,0.5813895004010108,0.7804644235439999,
 0.036717570973014535,0.17802568302568456,0.919608820036827,
 0.9881696759440163,0.5884566377471107,0.41396456066076226,0.6382444772610828,
 0.5114293715508997,0.4441122342744056,0.8105039640862715,0.3788166066778137,
 0.24837689128211238,0.009383603904874693,0.05272166540640877,
 0.35751374261842617,0.3725667750642714,0.3105117176733406,
 0.048580782831067926,0.20404355348345327,0.6156145982491115,
 0.416561529114833,0.5638095659206851,0.47788876205219233,0.7515043727249188,
 0.9360926839364804,0.6292463832670133,0.8340496748400279,0.8850098601345933,
 0.1610000068028964,0.47596019338607176,0.7772380198635248,
 0.24769556592511832,0.039094565813021176,0.19949880173377965,
 0.8040592353731552,0.8354598820080581,0.37626737922498255,0.4845432497862623,
 0.114474907988391,0.5333675481928714,0.19911495030836013,0.5278780996568159,
 0.44359696185955766,0.014398372984653629,0.7890636928652047,
 0.01138127859777427,0.32070791510656205,0.09916122507917224,
 0.8797989962048453,0.7070920511580341,0.9391342817204542,0.8851166895094452,
 0.6825957826451999,0.6169243969898538,0.2630865373844742,0.08557377350352391,
 0.10913815844567987,0.2257018481851375,0.7677537489280074,0.8486152830797477,
 0.3500113053602839,0.39375832860205795,0.7515034440287639,0.8262118065733784,
 0.03984208638500708,0.3811213550981236,0.38894890345472655,
 0.36586213578751947,0.9463250163265341,0.6743868634974534,0.6169253902775887,
 0.8401147436547812,0.49753026268977396,0.19703123743997952,
 0.40662818098737186,0.5633294498984555,0.12571788434765163,
 0.8487222342076719,0.9062760235782769,0.9194348102272004,0.7288392921889386,
 0.7518143187462598,0.48041579635307474,0.04932260866423421,
 0.7762521825889096,0.7567730505733755,0.29304156313644714,
 0.11010900939118218,0.979114146728032,0.7417604243297693,0.9039855815141722,
 0.5993110675849732,0.07853328151108896,0.28206707053301183,
 0.17622786805198987,0.6429639728468829,0.7427796737067167,0.8421106582565,
 0.24530303525083907,0.6617828134222425,0.8756185987022632,0.7733818200254431,
 0.9244442380764799,0.26380609539703737,0.7020755803366964,
 0.48392047690939854,0.5289439417677725,0.2743108060494125,0.4834474899654324,
 0.5544102978242517,0.7139058551087207,0.8604353152962245,0.45483404786989745,
 0.8807427356585242,0.5200110861132643,0.4801749014359402,0.4304774778067073,
 0.884574509743154,0.8420173415613676,0.7109851241431439,0.040930817259383745,
 0.40724110588160445,0.9626917706093394,0.6704903439830434,
 0.14962693701116891,0.029833264676127902,0.5948496508729157,
 0.9026947893515025,0.5874586688365842,0.8966141164122089,0.13707717357796423,
 0.6814407231103136,0.056029599394124996,0.18359313969872282,
 0.9242874597103058,0.8466322049160816,0.9800948110087009,0.11776849708441395,
 0.23042679039780378,0.8080532326376502,0.7648612832614113,
 0.01652588926795795,0.9141060187510601,0.016017964679226226,
 0.48563886859575134,0.7344435450851813,0.027683967958948452,
 0.34689065266109587,0.01895810690266231,0.25826895219961354,
 0.45944058355886674,0.8160186689595081,0.3394503192895125,
 0.19147781601158953,0.5826535087242345,0.8907347904527572,
 0.42498059742310246,0.291096141039202,0.5506327187634711,0.9557538942132611,
 0.28646572761962663,0.49639561376752905,0.7344753654381737,
 0.34679106624228484,0.18441294317291756,0.5298793548284481,
 0.14501799544914518,0.20335871283517104,0.30255741882788245,0.87168234746275,
 0.37407249081357974,0.9365965832038619,0.6781442190562624,0.8795833299510326,
 0.7845967813585267,0.6477363749038635,0.47334598214061663,0.9283319525312873,
 0.480805927545257,0.818052782448102,0.02923256466331836,0.6230747737691559,
 0.8650359884923442,0.9653715420128605,0.7208581867181862,0.8519390689535267,
 0.09281540660263876,0.20936322221479642,0.7253910142048801,
 0.6428563942449297,0.6910026391824604,0.6868863132510249,0.6522481023590359,
 0.3487451205550348,0.12562514299957894,0.7622225095770564,
 0.35826324057311376,0.7762353659306066,0.4548360371728585,
 0.07829021773521005,0.18615956454738525,0.12045559670191475,
 0.5060286185789555,0.08112384638152115,0.20358257720342587,
 0.30872576301746135,0.7041021536239327,0.3253022013938732,
 0.48734029905881215,0.3273714151934165,0.27750674143025955,
 0.35332911937865774,0.10273975379130174,0.7776185056503903,
 0.34390775498189385,0.682051619615748,0.543317143892073,4.938983451308054e-4,
 0.5770745091907715,0.5123016698051476,0.11965830049482562,0.51461499328806,
 0.08891445730539471,0.15693878499714375,0.681331172861895,
 0.35776017673589344,0.42525923495177165,0.331767422941732,
 0.25232901376370664,0.07353307736554238,0.14967081135030602,
 0.6796001337352551,0.5572883670732937,0.3197281889514072,0.4261606792009607,
 0.7325001180994436,0.45903370588293824,0.5084012477742605,0.8253472454758267,
 0.8042045280618155,0.15521486357487424,0.021899458956691098,
 0.5528379886267113,0.5537846765755149,0.9188856358662343,0.5258375588371017,
 0.42061006521827815,0.6761608229526344,0.7182613267849902,
 0.49982209786077336,0.9709766247867384,0.7688065465907543,0.2764056122052555,
 0.09397381702455199,0.7938384502969644,0.1547795011509826,
 0.012813041537277137,0.557981523353781,0.6542470913845949,0.3063395917177728,
 0.654151005351016,0.8344209107867602,0.3428188781237429,0.11400037964124454,
 0.48386024394951055,0.9220771203371004,0.8688096561322148,0.5819866370068327,
 0.7839304677819904,0.9997502525267623,0.6267518180104665,0.4809613899981049,
 0.1484909800387939,0.05129160419498324,0.4981714903681671,
 0.34407240989369026,0.3284822362845041,0.990368871289503,0.8386751520349384,
 0.9873580675315313,0.22893786324013543,0.3161411202425316,0.1632961809598239,
 0.6063710016261028,0.33343147836910947,0.0433081261437418,
 0.34477633935633656,0.1877812175767709,0.6358916010829762,
 0.12200523841579336,0.7622696478298814,0.04057109892501931,
 0.1350741239262082,0.7540250472251395,0.05574621153335746,
 0.035188984136318546,0.13046601129820912,0.8137440824234905,
 0.3878102884758019,0.8286275728204168,0.9915606414941549,0.1025088862923692,
 0.9768579313246849,0.25622827465787723,0.7989731707938478,0.8485708644775944,
 0.2093599834206592,0.9083031892435242,0.4634104441743263,0.6162364756701109,
 0.02528905479714394,0.4552041094981849,0.7541059776194312,0.7930899774464423,
 0.9535694506404363,0.9305066568041527,0.15558641991276978,0.5373734305807945,
 0.7930124785002578,0.43118056422720086,0.6480443168134501,
 0.22356024246319228,0.2090344941322846,0.8635953212132533,0.1820211906539927,
 0.6628015766038858,0.3194254709585842,0.6224503974696489,0.5637735879561077,
 0.2707461469414889,0.6399086639367623,0.8546366411727385,0.617719784422995,
 0.6566679819540079,0.3126885411510154,0.12967106269102147,0.8457348809688989,
 0.3368932697551237,0.32796968856358977,0.07698867270335275,
 0.3707554860983322,0.3540681627992972,0.41306708120107505,0.5274175012994353,
 0.23452452313474148,0.1537657255937378,0.183919368279038,0.6564372885009602,
 0.9420942529090895,0.7512897027797689,0.2899778921011138,0.4646809788993498,
 0.3163061559129563,0.14300415472445616,0.11682772709188638,
 0.9374002073899743,0.09667619319769605,0.454529240030358,0.4135800057087268,
 0.9699489703732198,0.6715935106656411,0.5611362451698787,0.630257227269036,
 0.2741469969754059,0.4514729376383614,0.2211918776056594,0.9381429976038743,
 0.01651441607993054,0.16922862070547995,0.9871895904009667,
 0.2263580656595151,0.8912087656805538,0.22826176203264026,
 0.44700914086727905,0.15417564883644708,0.26877177678451725,
 0.8286445853716922,0.7095144823633853,0.09450249020220403,0.4000011061280512,
 0.4342091332648873,0.545599467414718,0.2789606593846148,0.8083680497410541,
 0.2688087332406751,0.007287394222305843,0.6940925909865605,
 0.9790091611149977,0.38042027770608955,0.6669856671461212,0.805033743340148,
 0.2808032427086208,0.7625791001550147,0.6861539407601016,0.7471863722412262,
 0.05419198645204082,0.14762218980916508,0.033347508054134,0.9971273905296795,
 0.909605476435638,0.4479394645273256,0.6786607074695683,0.9725649598340907,
 0.8060125312289005,0.6459115451450643,0.36397821524061147,0.5130428286399666,
 0.5538966137013257,0.302619576986752,0.19204599807767475,0.8790783081520922,
 0.3600285975874161,0.38268457139889733,0.05251153701178224,
 0.5483659266051616,0.3649802475819087,0.7157111652523289,0.6778337067002056,
 0.3270827100302345,0.12643587195361983,0.8336174079989482,0.5132024671587002,
 0.03751410270420663,0.6548787108832987,0.154496867894223,0.8966961824182783,
 0.6473145774076182,0.8882558513418949,0.6173282425288631,0.5379870610710834,
 0.281498672772279,0.37496287189674904,0.12067672655481099,0.5678236754667818,
 0.9294393331600193,0.3876648195965874,0.4724283033918266,0.05081039132043896,
 0.19970884175920667,0.20598633007912037,0.2701111959454964,
 0.2624167224150378,0.7585285354011478,0.012197721607143608,
 0.28322079729055527,0.7194278028629447,0.061729715135496965,
 0.7324672861353546,0.7094310228849342,0.36376282895345036,
 0.09162440922773651,0.46630616880653486,0.8936359064271708,
 0.6371395578523075,0.5252290044346387,0.08457167669492471,0.7136601573295909,
 0.839169489641603,0.5135523740519433,0.6165998482860655,0.038358746490243956,
 0.011013592436922681,0.11542938144309522,0.5950600765623043,
 0.771296285036383,0.7605210266703886,0.9923514147918018,0.6727368495208791,
 0.8120530495967564,0.6314054964247579,0.9310530715970615,0.4194607556281106,
 0.366623047631346,0.05259725598079856,0.308882041175099,0.7951384276790615,
 0.579897752901536,0.9627211802378816,0.8767511147814593,0.5354801111836474,
 0.042416868988805145,0.5316090794551902,0.7777411437200306,0.628531268857318,
 0.2996276492869988,0.25005479573592937,0.12093102305175707,
 0.016594530362230353,0.9416311987457644,0.7657116163674259,
 0.9628037275894216,0.4533612314979848,0.37560775626391074,
 0.23828708589970848,0.09813954548361858,0.9884008399661005,
 0.16678259895650727,0.263409338518229,0.6668476752330501,0.2531615516333867,
 0.5134700589659799,0.29481265842517845,0.4971172035426914,0.3212333537591916,
 0.30342638968541635,0.19889906751823183,0.7080990177687683,
 0.7359721820395652,0.24931307108932754,0.2946144675604724,0.9197696245705914,
 0.023965986245019777,0.39326283817047714,0.8022852555321448,
 0.8074080779201421,0.9293360779429991,0.00885345515159286,
 0.049507376991381746,0.443487090828409,0.43240309618222017,
 0.11109568269867887,0.9051030900649217,0.5322309109354877,
 0.42076376136074434,0.6693555697194952,0.35474102352529224,
 0.11369812568715743,0.22036309852178215,0.01674476487110743,
 0.6106658502541066,0.4005873226209524,0.44866032390529065,
 0.43713204611147605,0.7496358571001405,0.3410278878921378,
 0.07471885055690974,0.04772140545960191,0.7528440011520544,
 0.8380059963748527,0.48633800678472605,0.5880067524990796,0.5758912887142916,
 0.7385013194812514,0.16077985932804795,0.2627659836300278,0.1909022706236576,
 0.20224248975044135,0.8674772983691008,0.9250819269510232,
 0.27721048313569974,0.8260839701041544,0.22007085431176476,
 0.5137524719273254,0.017744208808168516,0.4414623831643335,
 0.9892732813009215,0.15669523892584336,0.8419957323386865,0.8186109879197845,
 0.971375442438191,0.32213309030453463,0.6905949643169249,0.8987527365537394,
 0.2246222259376357,0.5608926163562665,0.5703585265976623,0.9780095705425266,
 0.5965429988601804,0.2585126178424578,0.6337621591107058,0.42067531881088893,
 0.7846535917343084,0.947143501864915,0.8751381903829798,0.6196650390371008,
 0.43134088927378533,0.591918698338155,0.22921192442625804,0.6048626733523882,
 0.7648948064003401,0.27481870672982356,0.002670159466035993,
 0.960184178120707,0.5310502999568074,0.9779451235439011,0.30044286072969983,
 0.0855714000893284,0.36796694044932554,0.27422533941762794,
 0.5596186516573498,0.9939826072896908,0.5031819457528167,
 0.024835325138309217,0.583870704227066,0.4354420855718302,0.468069021573128,
 0.14855878860400928,0.7175571520550941,0.9619559215296665,0.7766612257183149,
 0.9493458907756421,0.5934848640850098,0.8841540584864565,0.6036894489830394,
 0.11974232182340672,0.3063375684125855,0.9192234226668645,
 0.41050424029136967,0.49729782046179827,0.717706708071302,0.5413628829405069,
 0.22851473815750278,0.2300700585924681,0.04284262683658979,
 0.7383145541342446,0.19742386654925126,0.5126735701510635,0.2233781247209039,
 0.46773577808207845,0.3656637755156379,0.8082765333192279,0.9720684945769441,
 0.476297200442355,0.8784184194984256,0.7380599540702282,0.847222783373585,
 0.12195114309776911,0.9219282898340472,0.450150102358041,0.720815557533078,
 0.49689006053849816,0.8040062818524015,0.9645359421529671,
 0.24276236000114404,0.9484771771339982,0.9260350650710915,0.5514052871556809,
 0.9609549715832292,0.05332548819701144,0.3071665199138024,0.8759315233736221,
 0.6815358804302056,0.15693451955724536,0.484366195080732,0.06691366316818714,
 0.6887262997220468,0.24088703279356982,0.967523122796014,0.4346293624607498,
 0.24197256117468202,0.45825031380863934,0.347745327289106,0.3822841099212738,
 0.22658829874703612,0.8509403283674181,0.3311788644031882,0.7832173987080544,
 0.07418391583895068,0.6178089212166613,0.7794506163192152,0.8281237253733623,
 0.9702264259283344,0.16424660143435854,0.9665159121322557,0.9447520319652374,
 0.6763369755851384,0.25792843941813004,0.102356803692218,0.15493892710300694,
 0.45159474650527043,0.6193976849641663,0.346898770978162,0.4448126757850782,
 0.9015536429612372,0.842732498368006,0.13463154980915526,0.4078469214493776,
 0.638836851227195,0.7045664133433803,0.43596737383522133,0.2834232269788681,
 0.5657342344349936,0.5774350569846403,0.6826261191777423,0.6732081362579615,
 0.7248941066809952,0.5724059490434483,0.8637551132234976,0.2887947276507197,
 0.3568290582603042,0.07225528832868677,0.21297240729337907,
 0.8380091022843049,0.4055770233316428,0.9983169037992652,0.7850113615966744,
 0.24412895510256494,0.2663539377027371,0.08267456663217665,
 0.3469426203631326,0.1513089206749333,0.5536011340731074,0.12341422779779154,
 0.25688937662464373,0.9590806197883319,0.7872505590020629,0.9724056093710803,
 0.33545156180178193,0.7386191330328893,0.40898835545809753,
 0.6656236819123676,0.23925218619830668,0.38969321112232636,
 0.3406146887739412,0.7013431591790467,0.05156234020139483,0.9495850976598426,
 0.7794293915371602,0.9029919968822764,0.8826789898781278,0.27973391534051073,
 0.9887104969362475,0.5026615038196687,0.6542804955766207,0.453523206569564,
 0.6980946384210275,0.3948817293415555,0.8714228838158168,0.3433079428815864,
 0.533147719327823,0.5956495721457505,0.07023802309960403,0.3636807099765512,
 0.4145221392773162,0.14471436372222535,0.35163418890298737,
 0.49976122496695596,0.30877897517403363,0.6174417177072162,
 0.7299810055307633,0.4460577831043534,0.2088731496204831,0.427660674002005,
 0.2540235812248012,0.8008695023828407,0.2712443536386412,0.09673051115826592,
 0.8297988254608062,0.9960209709370131,0.7721657743848341,0.22781155369552364,
 0.6491193948667148,0.2309978882306436,0.33781393007550686,0.9262464398656045,
 0.42835729804349576,0.8656228631364435,0.9373565664315688,0.6548687993981279,
 0.7793920115399774,0.33593545396132596,0.707108760575939,
 0.029768693253755085,0.9243770392219564,0.11706999305044252,
 0.668040114710649,0.9440542911497518,0.6841320575180172,0.6773474384367587,
 0.17224069936738495,0.1818289268243919,0.09399218545943577,
 0.3986644804943972,0.7899523227711484,0.7998588238238535,0.44131890770181825,
 0.9951633173473841,0.7284970135187021,0.10268304009025209,
 0.09809394235397395,0.9132348588585025,0.8634942236654904,0.9773358400190809,
 0.2696693785323032,0.421036062971062,0.6129812300496028,0.9012317333250862,
 0.6731226991801511,0.9933161084742509,0.6668896335539559,0.10661183142658048,
 0.9452328245232288,0.41210775018081314,0.7436707003707943,
 0.17731258828099317,0.38707721848943655,0.41524177727951983,
 0.26204875255194726,0.3302785106440518,0.24931194749952634,
 0.2357363774394141,0.5519164282597834,0.5357949942346041,0.7555558715777307,
 0.9338821341659218,0.9505359160818754,0.8081694697487544,0.4410081018434623,
 0.057528640515312035,0.21566234641028204,0.5357751456910061,
 0.012172183617657262,0.019919237084301433,0.6983974203150192,
 0.9867863121307661,0.9344895904641243,0.4593407605800274,0.1881228038473226,
 0.6735164746780258,0.3427464967857553,0.20974795365870647,0.3566838212817799,
 0.8950404205413801,0.8617612283671969,0.058882778589899654,
 0.7462565772743576,0.18947451032941587,0.15514285256136806,
 0.12065986903235848,0.4600527404140522,0.7981284139209417,0.5477646168293961,
 0.7029994265483375,0.8566217639178474,0.7862584379520468,0.5987636938457089,
 0.1489001138651448,0.5910597050026316,0.3247295442671879,0.726195482456763,
 0.13356372259542104,0.7033306702174781,0.5184737404820537,0.9353187711471751,
 0.09677131166670416,0.7383178116525493,0.3422567944731889,
 0.33759626407169063,0.8682856592305359,0.1534649513823596,0.7504985100551017,
 0.8134997295061136,0.709659022920969,0.5032755070108584,0.6882948795083127,
 0.4892027112741759,0.754804817373673,0.5823921216535006,0.6873379981978206,
 0.4948633428023147,0.804136282071699,0.3056845848167691,0.8466615261769891,
 0.40688341377544257,0.10596792383619358,0.5027941588028388,
 0.6858836261641968,0.2400277743223368,0.6888757529123641,0.5382890463487636,
 0.18372571358965195,0.6951251928983779,0.42682809316887205,
 0.5312468013109339,0.8847172585990192,0.4131546964484183,0.458956521338729,
 0.9637523606840143,0.7414286052800868,0.22677856632056248,0.8028680201426307,
 0.8529680562567237,0.9344462447845804,0.6922468421672943,0.13675085915638618,
 0.17203759084038872,0.14970613730800897,0.7319945104564368,
 0.8543608616186074,0.8616538887583858,0.4902972260228158,0.3133481386848447,
 0.4947369610360013,0.9313435410721312,0.13100068052095004,0.8256703419105423,
 0.058119772727975305,0.5691160836501744,0.11918629871216424,
 0.11699159483429034,0.017258283462163626,0.799147661586836,
 0.9443361973907177,0.9496588644387355,0.6887499614295269,0.7213100265594447,
 0.5990252242317732,0.2807865833840463,0.1758338721501349,0.7197031863102176,
 0.3968252031162236,0.20142192689420746,0.9105939103653963,0.1835090995710318,
 0.8677183671637002,0.892971823941044,0.3452926744459611,0.8832388511265935,
 0.5457059020187455,0.9646022001733359,0.3857002319799421,0.6250026345117615,
 0.1671856581053276,0.4005578879192887,0.8127143716899609,0.6873590252550291,
 0.0040895613806261055,0.17415483360621298,0.9263890317230385,
 0.7784671186525038,0.7523244696082744,0.5358353412391338,0.4094376785234298,
 0.37064028333558696,0.8527309029680433,0.4890223455761321,0.5036530103000154,
 0.06187446132046304,0.5320844166362824,0.6823036720647129,0.6833136781290368,
 0.6969840341119361,0.12531678925438994,0.033647109048870916,
 0.13138415707234052,0.2331948311723685,0.061456889293569295,
 0.03556093578020125,0.6560150362694536,0.7702404624839138,0.9834286850240902,
 0.20539727309880562,0.9166467994439639,0.8835296030608664,0.4901519325589263,
 0.1067729375538582,0.48015124898337636,0.7111112968210086,0.8416966981900771,
 0.36819131528710203,0.5691547277016198,0.8244297497201873,0.3486075026320592,
 0.5569953924863171,0.9308513545519602,0.47462390092080453,0.5623015582983677,
 0.3359646955361758,0.4426655438486927,0.5202892749760798,0.7779628475453765,
 0.895666781076885,0.48316504267564,0.4662041638622706,0.6250133979494543,
 0.5072660084450549,0.5800068618067769,0.2519755558327613,0.7523531240164647,
 0.5631985842347286,0.36496532360061507,0.2471538844976603,0.5605444771112041,
 0.8438572128752296,0.08534760522376816,0.9653339159301298,
 0.07293437462124952,0.20971935180001977,0.12167990643984239,
 0.4938411353355152,0.7040234980680954,0.1256002756983562,0.30090150276226846,
 0.10814299244171,0.014491049449216642,0.6036912610057039,0.1976769340437099,
 0.2551201051940736,0.7242012863735042,0.41961175899367376,0.960511461667878,
 0.007487112304292887,0.6269262153258888,0.26987340680196703,
 0.4430715683094084,0.1559080226091336,0.5835843343689887,0.8992195184996885,
 0.7795097256669562,0.6064150103555198,0.5454214198160744,0.5831951527645348,
 0.5351155538084609,0.1092673124974618,0.7663178015404222,0.5178444880481744,
 0.1408487578037566,0.6411039379828178,0.04625652496032029,0.7552307051648427,
 0.2051047573535567,0.8472967154462177,0.735547645833207,0.9422063215641393,
 0.018614680341317746,0.19914058648916577,0.19244497268582628,
 0.28661361780356764,0.063687072637033,0.16875491492253036,0.8697388337939745,
 0.7671946249198555,0.6358771736474085,0.6356000607079855,0.20488179552979746,
 0.17563510959994955,0.3805499754437506,0.32878407143427724,
 0.7355491026091332,0.7511525866175034,0.2299927411931466,0.9202034926248969,
 0.32758355939596706,0.0675228757186378,0.1795309635309107,
 0.059200300895848934,0.5814687010603371,0.7465571401943714,
 0.11995954132660935,0.03042386906242167,0.5925896948269473,
 0.7600390184574746,0.12027035191422353,0.6489627534787941,0.8853233381818216,
 0.44655127227937097,0.8470658696340831,0.4274549750579656,
 0.058572901396392374,0.3596948043201106,0.6973828063217535,
 0.26809288073458093,0.49637544951358437,0.3850849060733571,
 0.5554029546945259,0.3028213617320117,0.8910480641011521,0.2219357102993419,
 0.5641343527646162,0.22486152346543875,0.20071931973809376,0.978509477205367,
 0.20765300771999073,0.4444312613738024,0.4072817254011103,0.3600565538802446,
 4.651934305996974e-4,0.4593779909521071,0.45171801707919246,
 0.40255434518704347,0.9721368330835358,0.020832892428582106,
 0.062471942146477,0.29799991665866243,0.3092555259923335,0.4055563055147029,
 0.8162391477190671,0.18569667096711995,0.673477353005811,0.18359672861926335,
 0.16655100951091284,0.528294317635252,0.6875494989273054,0.9798676110889096,
 0.1804841984703851,0.6344186086121925,0.35522890060210777,0.8471898036273972,
 0.8074674248342226,0.3423461153839622,0.32482284936065653,0.6298629172880834,
 0.778725679502221,0.936207609617937,0.9638582656343722,0.6963985135997035,
 0.4969928693808324,0.2126250337710761,0.6165800409251778,0.40533609494736744,
 0.6893002274549025,0.2819065358969397,0.5843546922880679,0.35743284683485443,
 0.6493469902324981,0.026183736824676673,0.006953779168568763,
 0.845219410110013,0.47098776723088387,0.45402925415806195,0.2086357541685866,
 0.9052165367260928,0.6286751779477415,0.8496257842189578,0.7121042525170891,
 0.8615577862668868,0.8485619479748064,0.1100026057716168,0.7659983027954747,
 0.7701097088909951,0.666470034571733,0.5033239533980526,0.6280261934506882,
 0.6143625535059134,0.5855225619164895,0.4219303521378569,0.4668462563533945,
 0.061526194396135936,0.5674910435579896,0.13288061202295653,
 0.9872651929414903,0.2958207298677167,0.44214027044204063,0.4917717122141686,
 0.8940429979130015,0.3166669906669971,0.5704564286585113,0.7819511725906194,
 0.8738582826911762,0.10285950409483013,0.7308325048830634,
 0.33922107380552735,0.8792224206741025,0.771055265285732,0.6631918772892974,
 0.19789400865998363,0.9596046057032559,0.10440015983378703,
 0.5935770082625187,0.3347351529702719,0.9902699730130002,0.8593592345656805,
 0.9821749215196673,0.9409151076235105,0.6247167827099118,0.6574269774678101,
 0.7238781384295575,0.14697570047561326,0.7329504103004336,0.7202311956386029,
 0.7695598512809756,0.25889494321170703,0.3106978668939273,0.7647208390954627,
 0.5482677729692771,0.2511127602769323,0.8610997450048699,0.40052328036129936,
 0.6700962653679025,0.7633726587333265,0.183126468044713,0.5588308769672135,
 0.7248243297662645,0.8493364400276938,0.1989015129852031,0.5505806229240783,
 0.6955014810798613,0.9298398646332936,0.43110790361325246,0.6171480090332625,
 0.438577792146296,0.09296280055816031,0.18949667540596327,0.5135227544877748,
 0.4544275386767378,0.4140906569594376,0.5758108250443199,0.1535194390197241,
 0.9073994604389223,0.39335680983292665,0.26253632542900296,
 0.5702475795178653,0.09935075051859965,0.3917733166402637,
 0.22155774123227023,0.12551826231290364,0.9722431755309295,
 0.18688759098149732,0.7700662627824563,0.4035428837086503,0.6499892952161995,
 0.6873313239068128,0.08807314501581309,0.44720108006792036,0.660958179340128,
 0.4775671419779127,0.8702469948036654,0.27921923912852986,0.329967431703339,
 0.06347619215520717,0.7184757919445123,0.05246954008489102,
 0.7708586151711476,0.7250867746542832,0.5334294389456746,0.8317189647851405,
 0.07770338791144316,0.8323070460783277,0.570784381156485,
 0.031107758993103585,0.9577117605617514,0.6654943728066434,
 0.05211026605400726,0.5259598622711953,0.4315043834187382,0.3000750959488122,
 0.6170359142713402,0.3128418501704404,0.2826739825453196,
 0.012948907724944836,0.11105484834166246,0.45361201306079024,
 0.48167875806767646,0.44491738548047066,0.5785728391887099,
 0.08381701345006665,0.7935530365798722,0.22079518759196992,
 0.6486868824237706,0.24266880322982431,0.5685127084157119,
 0.09263895525568522,0.5741343550930674,0.14510096078404533,
 0.7409481167881753,0.9556100184536127,0.9660182845036241,0.11375897841395377,
 0.5960776319445529,0.674516110726967,0.41367755109114357,0.48907597213862997,
 0.3646150014128712,0.5179972917739227,0.6420303741294721,0.6353955662443764,
 0.49368915068947383,0.025702746959329836,0.822284954223892,
 0.10470230861737684,0.3671560711928328,0.540594275173609,0.8251002363008952,
 0.9847204676962832,0.16671956565949364,0.8301868015090511,0.8265912472220044,
 0.5979849853747772,0.8541011964161684,0.6962298556470472,0.43311871453710393,
 0.9734246653588532,0.3889647050816911,0.9607604370561561,0.891090044008218,
 0.4947617939589447,0.2912349680622889,0.3076210096164457,0.7725905651562088,
 0.44090800771065186,0.8763694236425433,0.529178986697483,0.6044562784548069,
 0.8498309066024575,0.09755767687927841,0.14614212515356162,
 0.0945456483848256,0.9968395579989358,0.5063263932232559,0.958856053578285,
 0.546663006512143,0.14141155022621588,0.15842454032408348,0.628101487613402,
 0.93541994767164,0.8891247832933735,0.23409455005795832,0.6334353374409388,
 0.8992609438013742,0.388438452423122,0.36955881662840706,0.9301010101616224,
 0.011982414593432966,0.24503539946594066,0.7268302298814034,
 0.17724643184117195,0.2171490613468665,0.4518536555317557,
 0.49672941051257413,0.05122779313625736,0.4005102352387918,
 0.9554033305939339,0.5085899727577398,0.2763703457095892,0.9588080120162243,
 0.5645781873372563,0.7610699305723665,0.11632257774798083,0.3974084178014168,
 0.1927413503382709,0.7227556359982836,0.717204695155204,0.6651084506509328,
 0.8496808725804084,0.7957246426812702,0.9124715349137573,0.3113517915091234,
 0.40551329262415237,0.8499275937222616,0.3438517000739707,0.5437442209762334,
 0.20156930469362622,0.5838510094666381,0.6038210062948867,0.3284338457173769,
 0.49752405068549277,0.9011426660527306,0.006034276510882619,
 0.8393863536888971,0.8335637935498906,0.3109032711720269,0.7543040287681757,
 0.330566417403494,0.009460843276385522,0.36849795163291665,0.631665181233831,
 0.7287482174573807,0.020988296070459267,0.6138434223186088,
 0.6467068562202696,0.47518598756744745,0.009849783793412725,
 0.5436750536783456,0.11438286311500179,0.9919666533424052,0.861419112551599,
 0.6890766331022473,0.12787824213584187,0.9195422458330059,0.5235874597612749,
 0.8643630458422696,0.43200646172935675,0.024820383761056775,
 0.45604222465875455,0.4355323631678235,0.15613771139248322,
 0.7619014280296668,0.7747955936262924,0.0879717216193503,0.965653016040221,
 0.7342355629176105,0.24995461959592047,6.859681630522196e-4,
 0.8857898019626778,0.8699416591031841,0.41201353736354585,
 0.23136697666164485,0.2155761201709938,0.5993294286026706,
 0.022446617256900003,0.7081758953736078,0.6630536081673862,
 0.41122195956764895,0.581831679030269,0.6642809705749699,0.5045571524463986,
 0.9729317108461796,0.5749215207230286,0.2749649276217775,0.5693724947822145,
 0.12122850036731292,0.9854891005840863,0.7445718791602549,
 0.13062419430562977,0.6839122730543121,0.20036729607974169,
 0.39506763397574474,0.056255202935197146,0.5905506730321134,
 0.909333954170346,0.20633117071815554,0.742192895524557,0.17264438179905228,
 0.6889767643039163,0.1519647647897504,0.6908508610591241,0.3686011584790746,
 0.9042865923443069,0.6500062206199833,0.4812998766287572,0.41300178947448696,
 0.02394413780533311,0.8956206824434132,0.7747175388282451,0.8301312649516389,
 0.9655261594039165,0.08122945992963349,0.6985294651965087,0.853013207374628,
 0.006626041448581033,0.6457261712648452,0.8472615205214213,0.571125052445144,
 0.6220848972423634,0.5902121402534458,0.6167291741892774,0.5003005316456437,
 0.06992076694953986,0.6511173679464245,0.6460986811670948,0.3361526719522767,
 0.6126827453870675,0.03008743245205281,0.4071365456103354,0.81456548391717,
 0.48069579183036415,0.9912014817049328,0.666850165969759,0.8199599711884508,
 0.3491671246795356,0.7161331457101148,0.9673696288914658,0.3599732067640844,
 0.7638070916217081,0.030611204575500928,0.6821158201661756,
 0.44360399493078084,0.36125798111581797,0.6801903176852566,
 0.45956938649814405,0.6459589571908289,0.3606734856965401,0.9444768985669045,
 0.9796254474754168,0.8557917680393494,0.7815249242745818,0.997472593369676,
 0.7478024498083055,0.4642213641885855,0.5559031705317934,0.7001171824465396,
 0.3763605335013732,0.7495173927658729,0.2785364968364318,0.738173940366404,
 0.021290276475136816,0.25380711388882915,0.351833948293105,
 0.8404978453089089,0.6248994621488947,0.7504836636995244,0.48117039439990505,
 0.8701470633606094,0.3678717142050447,0.6929400856659644,0.6809341042956776,
 0.6441950609665017,0.5486721703119,0.08457563808650392,0.5893817734872164,
 0.6231454272641077,0.2206677502373473,0.1510449118099889,0.9347636791356423,
 0.7346624674171075,0.5455974059500632,0.09896895761787339,0.8036088585282489,
 0.9744798165578442,0.8247091312310427,0.14097826816761683,0.900795899381837,
 0.8300692021612015,0.12426370646957152,0.4754609666694327,0.8412484185279652,
 0.45340466560220394,0.480504175141911,0.09942252661184126,0.64025963285376,
 0.6286587675535179,0.13503801208452404,0.7308170166656933,0.8953097573695499,
 0.8235336799746831,0.7666874153181105,0.2578812023493793,0.6286319417718018,
 0.8183043308237898,0.07310838848553947,0.9561387661135747,0.2484508588661628,
 0.2218538665474844,0.09120068179416319,0.8686045119388032,0.8384281828497447,
 0.517584690718674,0.911859731499001,0.45984901900214625,0.7023130430922216,
 0.9272683527054645,0.5853730962642357,0.7712839661636751,0.4085936776678537,
 0.6486449300254267,0.66171299842073,0.7855972137766817,0.3646162069286307,
 0.6686317021452923,0.6556384845856692,0.9683587790886131,0.6859482722482095,
 0.35708909169962855,0.8528072137277084,0.46838939798942647,
 0.34237715602205354,0.25704867199996595,0.4417081781915766,
 0.8793656376145599,0.21979224078072868,0.21634304588933495,
 0.05485610041072042,0.6577421223985894,0.11200796465378848,
 0.23698069841003078,0.6973368347758515,0.621635964145856,0.5246622471959748,
 0.33008521781086186,0.04431296108904381,0.6860154993118301,0.740697366624661,
 0.7898846867471547,0.995988244597901,0.417778212193202,0.25269629155662965,
 0.3121648993194993,0.5491730252223054,0.4230581887002618,0.7863415300256625,
 0.09514761370060931,0.8264080275206366,0.15287872798374025,
 0.8331563899320694,0.27364585358317983,0.8288888765735374,
 0.07925443259763743,0.5240251375084665,0.029838200616873145,
 0.34774885662896415,0.4830857503793883,0.8868710178019424,0.5807144384113989,
 0.6283724988755548,0.7791571153462551,0.356954716226201,0.1793403131167648,
 0.8336711894964903,0.1806790422327197,0.8833116320471224,0.7981684267901515,
 0.57947239560767,0.16829831073272639,0.41749651903237117,0.6220985907660161,
 0.6782646031744688,0.6169312021039256,0.6752323218942913,0.39636752371117234,
 0.7129329991426623,0.10126713067580428,0.07790862772047225,
 0.06046381429548198,0.7062634802184548,0.9497280641980419,0.9048848794674833,
 0.6923658541888094,0.177949285462926,0.2682727409517094,0.1617079498511097,
 0.5799044324463427,0.4819222399821299,0.21732029267342223,0.1954063740701969,
 0.3721694763201797,0.28166822373531797,0.852681651754792,
 0.044431798293418145,0.8736723081393554,0.7482589935937634,
 0.8539139113498168,0.18230370413398922,0.4800567883026171,0.4892275645277693,
 0.4292598561656509,0.7832342068999848,0.6185291735877889,0.6618453988011481,
 0.2906965011029421,0.8360081694050642,0.6449169529494233,0.458794442569058,
 0.6299125841003276,0.293860056943259,0.7813627481115962,0.7696228169955199,
 0.544223695247462,0.39176677922351966,0.9944329388013593,0.564337730040296,
 0.9670760937055389,0.9450016474915025,0.2036059290644867,0.743509379566569,
 0.08114221800614718,0.5930623342317055,0.8574974982606608,
 0.10156424026678113,0.0258446358443577,0.4620559176693204,
 0.015044780224329979,0.8922243981310821,0.9112564658827369,
 0.6864958040576551,0.9713418296762804,0.4046245176747245,0.6374611667464616,
 0.0046849894043135,0.3131835551132819,0.37372467402292686,0.7427263284998824,
 0.48915682720968656,0.9628037908255398,0.8214263931737025,0.2527604790943315,
 0.16568827637657402,0.3023783512174456,0.5158867355746025,0.5800689668663207,
 0.9657927717363762,0.9056134017051305,0.6201651970976381,0.32330929494213567,
 0.37105763699239624,0.4974982782271068,0.6037089518304145,
 0.06524104155744304,0.4909125862431508,0.3946794067431745,0.3469561295793794,
 0.6299576297126408,0.8630822745691757,0.12976381147663796,0.6725120462319238,
 0.0558173551631459,0.9953110312853672,0.2686480209721882,0.8160843012393989,
 0.27269221338973704,0.8985106410207316,0.6123478932338307,
 0.09087366020904686,0.4356072554251532,0.9683541045267533,0.3759920598997981,
 0.6504206452115485,0.8668091779585158,0.5068044666427143,0.13774247348738378,
 0.005409910342961011,0.8426886129793482,0.8425804737875013,
 0.5325401780563516,0.24891257932079436,0.34716744830956026,
 0.9237603049674631,0.5722366694450346,0.6649012953260109,0.8480315715799963,
 0.8762425698655476,0.48026854646083206,0.07277126178032212,
 0.5965575978214626,0.3651812743348808,0.23013925280794145,
 0.21225842138542894,0.3773598256808741,0.20422955612742344,
 0.8406303816356102,0.8994157758629254,0.7633305501898,0.6407798950617501,
 0.2777795241141443,0.8298987961645383,0.2758540721539775,0.4392023821456964,
 5.28772453873172e-4,0.001604415024217154,0.21009693145122188,
 0.30161565462173334,0.2799164418526331,0.31068765802930354,
 0.0036398059094100432,0.6812419469718827,0.1350316576005277,
 0.1163050060698213,0.24163707262786693,0.9107782665855666,0.9068924387244294,
 0.35018507082715333,0.8190507979520864,0.7671076435699373,0.8133265171034213,
 0.061949934895197956,0.06337726362470186,0.8053946192763929,
 0.6057568033049512,0.11053915963982197,0.2319456022508829,
 0.08436478311043316,0.021080125022216456,0.8860272392536506,
 0.7653432209456097,0.537089714881137,0.902157850023282,0.2614532556566106,
 0.794809767056993,0.7336454593316123,0.3649356280327315,0.9871214230052153,
 0.05682108167668942,0.3243432931540209,0.7522849577596724,
 0.15854757384258034,0.9536809918974807,0.29020388572415357,0.222142177563172,
 0.9378132787056201,0.3942313497401879,0.45095393503517567,0.7186143587275405,
 0.8789541302768039,0.5766242672166716,0.42964407993144704,0.6905861640508045,
 0.3994058334586049,0.865652445284753,0.763524015109027,0.0997382661519155,
 0.7631581970722388,0.2288945161845688,0.7915638632805043,0.6629572164261118,
 0.042057195027169425,0.3889827848165167,0.4196733959654363,
 0.7020738754034402,0.5001052801409942,0.31748803465465425,
 0.42153856805525813,0.5842447506193521,0.8770233947955249,
 0.20878470924788317,0.04889014819611981,0.46540739732970404,
 0.19722540757939588,0.011922302064129364,0.6332540876915583,
 0.5148755148905069,0.3545479492702457,0.7840468119396624,0.6825104936432046,
 0.32564758237152314,0.645938853887946,0.43440612453900074,0.294862981590464,
 0.8027785106130284,0.1726235446147858,0.8010510324975562,0.33149705080496894,
 0.2691015348007214,0.8374605934882524,0.4151774669976359,0.39198137671716315,
 0.2697636282190179,0.14792227929703727,0.1454820478108927,
 0.014908874770156788,0.9224105447874997,0.2739037572983304,
 0.10945769927908089,0.06399248269440738,0.7284397718043136,
 0.3338570991935854,0.6802472143282403,0.36442542734045935,0.757690214840188,
 0.27006179920435247,0.9414201422792178,0.36267128096161805,
 0.8499863219299332,0.16940973230420253,0.589641090107994,0.3573457514395205,
 0.30576640794277465,0.18723187900538685,0.32729769729687885,
 0.3212194001583141,0.018231822411406196,0.9429052234738757,
 0.0749028389446158,0.3434120418073733,0.009682672986299545,
 0.6757771833798888,0.37204985211028696,0.7011350802444631,0.4114626415300435,
 0.232545224216392,0.14968051934999393,0.546308613411132,0.5854408560091109,
 0.6724221054913917,0.8146012593987374,0.991088462682343,0.20951109135195767,
 0.4341600061557088,0.22237290502030327,0.6501128943151201,0.4551413930242043,
 0.2967563969542657,0.2613837042727085,0.9335346403693316,0.5477156112008811,
 0.8785197422823663,0.023811159178031782,0.1290367583270956,
 0.9183883912978404,0.029537139379770205,0.2503744250602319,
 0.7027659857474222,0.9649139105870799,0.5110171051455233,0.38851021697869115,
 0.11494725098932679,0.8735958666069518,0.8583389849585543,
 0.47444448619126334,0.7552485841737574,0.00734108758835339,
 0.12900767908176314,0.40247622064640487,0.9079142566057893,
 0.8523972823328618,0.9112589773907459,0.7780384329391431,0.5142810181214772,
 0.3154117711565263,0.12850555445612022,0.645849672579703,0.6617477402080021,
 0.5777196794259026,0.1310054755655331,0.22282621172843875,0.2878413426287323,
 0.2642537569429526,0.24087964842667242,0.6034759751585043,0.7328146772996961,
 0.03291782633186613,0.9571782521823532,0.2196278024337459,0.8895297738673129,
 0.46075743081248555,0.10576648078699813,0.8948598663496589,
 0.9794434324674286,0.2103319809000227,0.8950578160186907,0.934222668992181,
 0.2050377619307907,0.2730071450259074,0.04461512271033463,0.7834403830548204,
 0.39952910257562446,0.9865697379451766,0.8951477498026091,0.8508882743912787,
 0.9312119740398821,0.35231661938654657,0.7360578506715998,0.573698493110939,
 0.33577021114223815,0.08038884043185246,0.9355169282479239,
 0.3388242416883769,0.4935630156043951,0.9767361663014924,0.9789857933529547,
 0.20807843290518868,0.39756398889544975,0.2553203243274248,
 0.6858512785768576,0.6099839240755641,0.24063382638805442,0.6790784674285637,
 0.02699146178588352,0.6442583993970584,0.7930558811562086,
 0.036422725941391065,0.8013799479479494,0.9958099060828671,
 0.22096439341458562,0.2742891128836473,0.39231218632535547,
 0.046778928636306616,0.30321104587075376,0.33908964992033663,
 0.10773401762775281,0.6930852233660525,0.8089886238541448,0.2920019998361876,
 0.5491767462001127,0.49339294399932054,0.5233691982246611,
 0.08354473191813372,0.8367775491490018,0.15216548089247584,
 0.06637910560543525,0.2534559204965553,0.8834362306159875,
 0.08164896969713165,0.10281840686151122,0.5360465112590003,
 0.7417987089931964,0.5279490475907411,0.8275190220428076,0.6420374441533757,
 0.8999892881897823,0.07581157902711166,0.5541126582295097,0.787758356687614,
 0.25257196713752617,0.9069232098743413,0.3946893695661976,0.8761627137615526,
 0.631494612346556,0.47838850896958984,0.9514096103179143,0.9925304006837413,
 0.8185200947572886,0.5085972144227373,0.8329898884476256,0.09413789687573748,
 0.9638062763253455,0.7730018003648014,0.41685708404004596,0.9032614608962766,
 0.10212867196684261,0.9427360652160891,0.12343239251233307,0.97374300986211,
 0.14277475704975395,0.9551757174770633,0.5662692038407459,0.7621108768537166,
 0.5278317937269718,0.0318487480071179,0.17445596188114243,0.2173056424602855,
 0.3238160672917172,0.060829291638530125,0.3997931642084701,
 0.8467882134894421,0.14507487044399403,0.18135181871767037,
 0.18646686354175868,0.16244433976018335,0.9438053414505443,
 0.19930156147105105,0.38322786068315806,0.20172725835000027,
 0.3829703035350387,0.19063705298503597,0.31334399373603705,
 0.5017073559068034,0.3857640302592157,0.034114081991663214,
 0.9473273327035869,0.02607748656740272,0.08095124853815006,0.849833152511417,
 0.22470166227927402,0.9399199307830065,0.39526201981711817,
 0.3280892380036564,0.05266455182554297,0.21506514517863828,
 0.9090702007867725,0.43225388364298045,0.14729934693265823,
 0.4368111317647694,0.26860583149174055,0.8696112529680645,0.5265632790309531,
 0.13137571392723646,0.4218537040157121,0.4387401468547718,0.2964265691586788,
 0.6395237326568424,0.5695876313316139,0.38454473306472203,
 0.042924088364700186,0.9295844852528605,0.6899928032903923,
 0.6098705139171952,0.26837317175148956,0.5293335130752381,0.2273615710393576,
 0.9106572270644719,0.40017392525798,0.7019612691611334,0.37717755626439864,
 0.8710476300210096,0.7671271404536354,0.8381413633145295,0.2196041383920997,
 0.056914641087115525,0.12050304835431846,0.6064797860544944,
 0.6600951930782231,0.29739229394134925,0.3081885673882938,0.4114058253500793,
 0.5592562693066228,0.31189802650475285,0.8575287098972799,0.4001763355329385,
 0.4604217056966112,0.3609702907208713,0.8522847664167448,0.4879526786034205,
 0.8538829146015045,0.9800248315353032,0.9705575899232527,0.3522218651731467,
 0.5444569812209711,0.0994773868102955,0.8447556174935371,0.20977251979045464,
 0.9853736034192251,0.2541492082132264,0.3916653570522718,0.8795810879494281,
 0.7604204975178104,0.27028648504882236,0.30466479748036934,
 0.9151160030676426,0.5263656733786533,0.21975534346795655,
 0.0025405474850035414,0.20027726682184832,0.2915072026900729,
 0.4167072279981099,0.21059989050020667,0.642493630691936,0.7959307880196953,
 0.9533627875980124,0.5787001834118242,0.9781268377327856,0.2476771889995416,
 0.03387875211064173,0.6154986233572415,0.8681290720200272,0.5989150404273579,
 0.16221596211361777,0.2815729514734804,0.6544397657834754,0.8167768229955003,
 0.18658784624770441,0.11308010673451407,0.496411686658345,0.4424216445441328,
 0.06706574666243292,0.8679405705285458,0.30002341769418583,
 0.0648224982519211,0.37999888816592564,0.8615617691705634,0.5825968289324335,
 6.559808434873915e-4,0.3374513421136621,0.436601281090208,
 0.40333969709600015,0.16404793150244057,0.906334460574781,0.5752588708472288,
 0.7725444913033834,0.9646524865412204,0.49724092445106804,0.8110431857353503,
 0.7421213411311909,0.7041184495090158,0.10522169966011807,0.8603082467263763,
 0.3122196911381008,0.5796986102259879,0.47901416195676316,0.1295973819826426,
 0.341501003471156,0.5438394511632084,0.04055376745995698,0.9611081174729342,
 0.5280537740601597,0.43817833874789003,0.7514633771301249,
 0.21578709276529506,0.9722968227268658,0.08294071919081891,
 0.5966717252425072,0.28948566715235435,0.5017450414117559,
 0.35343936857967584,0.813474835291723,0.6793491363983684,0.700038371357677,
 0.0032431716451376547,0.1795036998589321,0.2770329448846144,
 0.15203421689698704,0.42749644175988255,0.07004560624680434,
 0.7075559927939017,0.3185967003600638,0.2353054816906851,0.4357907888720808,
 0.42721664297770856,0.39844076571107434,0.7559083888946753,
 0.9427341235045159,0.0538714228777295,0.4565327775541229,0.9037566539445618,
 0.7336678803573398,0.023965217834233665,0.035892092147760346,
 0.5965764598357152,0.46108722523075074,0.36713785543274857,
 0.1375414544432727,0.896884742448223,0.6928419181851423,0.683813282022777,
 0.24034021552681772,0.6235056777257697,0.5618577618960224,0.3877677505350392,
 0.4928793725416917,0.8695879329544611,0.5532959391936212,
 0.010553036127507243,0.47348680782002706,0.11496981468831935,
 0.46943743718300657,0.30252608072081955,0.504765811604564,0.5840953098606774,
 0.013231976891937003,0.9108624784462805,0.0898970203135856,
 0.09278890212384538,0.8585798444624556,0.8156291512948025,0.8757714235691019,
 0.8767307802710537,0.5268061776141915,0.7985664115799261,0.8688476706102403,
 0.4646275125285465,0.23675845388240968,0.4176852660239715,0.6778400188386078,
 0.10915194015031238,0.6136245048614409,0.5289517122486845,0.9524296234385208,
 0.3763872408510468,0.547554279908756,0.9406687082080186,0.20656330043733573,
 0.02970562866892823,0.6096380914190808,0.7719353465310698,0.3244025421915633,
 0.2990366771103222,0.5951217612479526,0.5941297486861182,0.9941175616610108,
 0.9700654590846214,0.14522485910679783,0.8143885735083017,0.5144998835048336,
 0.4328708824359575,0.28523580263460413,0.15494606611659245,
 0.37849358210248285,0.9193847428734666,0.40876290760662715,
 0.9663354890779019,0.45525327861597287,0.4698730055600504,
 0.25209130266284063,0.5193216422802001,0.560914482842715,0.09563600950261608,
 0.1892838182632689,0.9208150483783286,0.21810370003295898,0.9772258593563767,
 0.4439595948438654,0.05105591958026534,0.9328606186471461,0.8707945189952243,
 0.9500595301293931,0.04645709678621446,0.6800905026487427,0.661117515380527,
 0.6890103520105801,0.7831818452231353,0.13925175401580403,
 0.11954118096889754,0.5778688582700545,0.29241636735818965,
 0.5747644759278473,0.73145176130555,0.8796724434447716,0.38077767843987564,
 0.9002298713145028,0.24957299801706068,0.08260943581427138,
 0.18266399381157183,0.9174323989605054,0.8745291727985268,
 0.42285415131099946,0.15698298411682476,0.8533990434400867,
 0.29712569465816985,0.2828968319213092,0.45166383270030663,
 0.8224444582767498,0.4476467449387468,0.49185187112112694,0.5634319192416732,
 0.9440860040218122,0.9875954317954905,0.21867383518195083,0.9146556951714168,
 0.8457881783521749,0.9667884552324941,0.15048614354541823,0.9898395163352441,
 0.2952335955305099,0.7916338156201669,0.7100512758986107,0.16470279579892322,
 0.9010561258167215,0.7880257710312001,0.15532974464181437,
 0.44975964189979867,0.423211568067957,0.572966843671243,0.6786001781171593,
 0.034330686804590727,0.011583901169781141,0.362017161830853,
 0.08136661978468662,0.11936215945577788,0.8822359807074686,
 0.30939391233218827,0.38180930740526364,0.01787757723193706,
 0.6614381455982369,0.9303049127252042,0.3709525920184701,0.30668044424948127,
 0.3257079298890877,0.039199810226235976,0.4267944887330166,
 0.9725224254935222,0.2889554397518561,0.11712738793614541,0.6777394059342265,
 0.6041797756004111,0.5241909353339076,0.9435855930544252,0.14201588965345158,
 0.05497792751542818,0.13271288173854923,0.4582113518206672,
 0.42439868435006034,0.06621603685030397,0.5215633611184916,0.90370907270877,
 0.3382613231828291,0.9660255844148753,0.9178933821527676,0.4486643854809529,
 0.5981771029112862,0.2565585303843797,0.06658018681291333,
 0.23416639557748598,0.26750878068187833,0.9970787929673306,
 0.7633817538981359,0.7963881384768527,0.7618265644898164,0.5628776658523582,
 0.8331857152733536,0.8004614276920005,0.12005593014914673,0.3372390324057081,
 0.25236795896013153,0.959199425661663,0.8544576199967842,0.7449098677822937,
 0.662448593592802,0.8627928505161625,0.9262338489163096,0.07101726227783434,
 0.06823940561911346,0.5406182871471968,0.5399009457646411,
 0.41952214513497266,0.44029878177449344,0.029020125519688578,
 0.6936096204540423,0.038663671753554674,0.9738304388438271,
 0.3685527855647267,0.8782989153454002,0.32186781424001376,0.6890949084743844,
 0.5710202482445416,0.9369767808871678,0.10555524788776782,0.7386426419367117,
 0.9582091083554014,0.47474314798892125,0.9144177715565722,0.9040825435688795,
 0.690381196583574,0.05298999478858091,0.20117211144850922,0.1837740720868738,
 0.6938981732344126,0.37375961048285744,0.18710203749787557,
 0.7165224198803191,0.46020191116928244,0.6244482516685669,0.653825489430743,
 0.5627229682979747,0.007821410992782662,0.11664367893053396,
 0.0032234616679414074,0.6050970200474208,0.521213512102398,
 0.9403393918526874,0.6771780992556965,0.27988449050721254,
 0.20828376863978304,0.9985872705915423,0.4930933412726737,
 0.14948428611915032,0.15225142114578516,0.802870437859651,0.7477406074000805,
 0.5667948450631759,0.7285737982439581,0.7072771077659712,0.9145423131424643,
 0.22892316744584307,0.154955379841079,0.002093942942957172,
 0.7045466127157677,0.3015655994390387,0.9199129706129598,0.8442031065387592,
 0.16908100742672305,0.21541995468129138,0.6264232593240715,
 0.03686422765799924,0.032781893928228056,0.6762258204171706,
 0.45338449882909415,0.6494934786846376,0.22360338635750576,
 0.054565281726849935,0.02255778841076861,0.6555564172000306,
 0.01809533640358474,0.2809043782429058,0.7174109604593826,0.3620407747916988,
 0.48351493839355975,0.28362322178216426,0.6967804108001385,
 0.9888979174325843,0.4927278641460431,0.9696631660302427,0.5092060877598871,
 0.843621589427833,0.7163206259929426,0.8751032628361416,0.7819316218485476,
 0.4965097302899223,0.2802406545439543,0.739651423478842,0.7351711114445099,
 0.6082067180718967,0.1417747559170578,0.5543649504570463,0.5859948857306141,
 0.8232296173976227,0.6056762436576741,0.48932808180066745,
 0.01913219223838536,0.28158795234493506,0.9203187630197345,
 0.20508317086098204,0.191992195687685,0.39593250045930817,0.6654645555631928,
 0.6177773605072218,0.4590031333321709,0.6642141294904284,0.27478275750169734,
 0.33343197398194446,0.006885788978425067,0.4093233279625863,
 0.8976043569539232,0.20630466203367848,0.49527630337414275,
 0.5085938014866005,0.008508122488191439,0.10900605682794351,
 0.5685504654980296,0.7889098348800564,0.33641411123176823,0.7730141131993178,
 0.07613026362319919,0.4622434468291995,0.982076773312046,0.14846537605296783,
 0.6579784005858579,0.8309580898861825,0.4107730633451465,0.4177730774690521,
 0.21647812616733608,0.8384006030868166,0.33239317846889516,
 0.41052846458022907,0.4193407710560857,0.4350753486842519,
 0.46436080212438335,0.3085286525609422,0.7022456381379737,0.6549145920615743,
 0.22545622765412443,0.13695111168309848,0.47891366512795563,
 0.2706180892432085,0.3671032089035011,0.3717505593079764,0.17441373867082188,
 0.845867289272875,0.5357450200439515,0.9951003474843051,0.29044916508717555,
 0.45073877549301267,0.2611063410570059,0.7220363705143684,
 0.02847003755854527,0.17808549878398605,0.3576377471117945,
 0.6294461508956172,0.49771668163825133,0.00483675241830972,
 0.7157231963788683,0.7159919902331008,0.7830032455186667,0.6139045082167474,
 0.03440774082781606,0.37335772096316244,0.43543858336380803,
 0.11715575483797591,0.6292137305589264,0.06170828170522891,
 0.40279682773659264,0.4832207519594316,0.5242997289628818,0.7082645694113615,
 0.5670931322475427,0.3465134531163315,0.020172188376493594,
 0.5627589903344501,0.6374392540374334,0.03889216140580898,
 0.20336047472651086,0.10396751430269924,0.01724933481625035,
 0.9412011814936915,0.6189428011809656,0.2491367785246943,0.1631849382707138,
 0.32019222813927195,0.6041191536272107,0.84855854664591,0.7012980716081955,
 0.2746653191206344,0.3519270613677129,0.708753073513882,0.7102895467627359,
 0.39568956358223983,0.04241527377738796,0.8462538380216584,0.254606308183713,
 0.7812427580840315,0.7911779921193163,0.3774871129229923,0.0611328589984268,
 0.10637677882971752,0.2471672774147753,0.3863920045555945,0.7956367664500766,
 0.10038821185898839,0.49131250380520897,0.9339361843625501,
 0.03006117106888595,0.8861277349569111,0.027157123411571726,
 0.5399552223833929,0.3600779822035529,0.95743099637089,0.7096686659363813,
 0.027371228521384627,0.17845769612545292,0.9517302811988728,
 0.5740944691072054,0.9547868218368779,0.8096860332947097,0.8364618694708493,
 0.9955690544583672,0.9916717569408445,0.883507903630086,0.628932333280625,
 0.6258175942385131,0.04971378575379881,0.21661225537806428,
 0.8224677332911259,0.4948231541653396,0.28792366056642715,0.4442331155325182,
 0.4889425132225349,0.19448885838432628,0.1496475701045169,0.804715368864903,
 0.6506991199814209,0.27509516587380844,0.1953421919785049,0.4136351161840508,
 0.3816682985897958,0.9639050984024132,0.7047953254479906,0.18558033839066745,
 0.019827596919519963,0.41352241708898263,0.8463066154827952,
 0.14337092037159882,0.39301276734837953,0.599164995908241,0.5822908055970683,
 0.02516336312311418,0.8377760679573408,0.6018248893697957,0.9591522707797919,
 0.09421613700425924,0.9772627901428932,0.7948517225634968,0.7602254021808976,
 0.9275038885320713,0.827212789757274,0.5760951671000832,0.31986975212781776,
 0.34205641040986634,0.703369487568825,0.09471806440927999,0.7145595021380023,
 0.3033311416962309,0.4570631650200787,0.07406619649732915,0.6619340325883236,
 0.6384393077870649,0.9405232977923977,0.10839755272090934,0.6787809528104808,
 0.27317430013715027,0.21756433236050765,0.5041632705177708,
 0.5152308552777991,0.6384474834991182,0.7250760346262107,0.7607754609822113,
 0.18808472936717102,0.9186829992031217,0.97344688209128,0.886147575614397,
 0.1826614143223647,0.8092033928531087,0.7333334188978184,0.09017632823009047,
 0.054458708958929236,0.04444789743569011,0.16193175957976713,
 0.48151921776271,0.43598176086232254,0.7783788987673816,0.05634045264276655,
 0.8715830661231305,0.017722972108242474,0.05741946981594914,
 0.25331904530328575,0.15547990687684243,0.13040962880721296,
 0.2142463929269658,0.19308253865935066,0.352533190425678,0.11891999193103953,
 0.1036004536432682,0.7629890156998589,0.19907250487202977,0.854541847337975,
 0.883695991141121,0.1315864685006115,0.60242557327562,0.4338061491374333,
 0.806145532304918,0.7623676928148064,0.3812965558918391,0.8968303578571634,
 0.4705282638173558,0.33022723605482374,0.03488051653848534,
 0.4256064689249991,0.523749020986372,0.5164924031301421,0.9148027942312111,
 0.6142253839143988,0.5371342966356119,0.3600063128616371,0.856930808923397,
 0.2574859318067724,0.7378337875865586,0.6266247636151914,0.03974883547099928,
 0.5961909008298454,0.7127262652578451,0.39234622318173673,0.9260370849849424,
 0.7870618706433233,0.4281043565901146,0.1702611432883342,0.8223339336132801,
 0.9030454587607731,0.6013603309216478,0.8012204234922525,0.7931376966866422,
 0.006751304838694461,0.614779158799696,0.19885232531312036,0.839378579259035,
 0.7029410668115014,0.47628923546965507,0.15224681859674538,
 0.8951549238925524,0.5400214688684998,0.5596623393908962,0.9472600466878272,
 0.4627621642458494,0.7567432867092566,0.9618722062187365,0.20613459915668775,
 0.3679610819624296,0.5314258374858264,0.6051372462684443,0.35152885914985266,
 0.8496174424696026,0.16460225581509447,0.605801986887246,0.8362511933027252,
 0.07593133553763232,0.2901555520853156,0.21354051243583472,
 0.7665743574918005,0.8054613794560562,0.8865490168691031,0.668833587722905,
 0.41270245759971036,0.05107288151831946,0.11460033688289306,
 0.4312961375866904,0.3951815117747448,0.11152590656040329,0.5625848727526349,
 0.9693862584599492,0.5644062763110893,0.9676321396300176,0.28144505832612987,
 0.9629773995747414,0.0860517539103327,0.8671850227324353,0.16646184539281195,
 0.6256157298403752,0.6254377213170919,0.5128622423124984,0.3329728674043698,
 0.41482687083251923,0.8231799253546598,0.23054806789046267,
 0.28506045396953983,0.34643508801561784,0.3354869542052714,
 0.011759632295677003,0.6980243330201236,0.8842032894203968,
 0.22520704144597337,0.16908817614455196,0.373133444905429,
 0.41339145830755275,0.1441717752934344,0.94414412688289,0.3191596198265242,
 0.4038170112471253,0.5371856234992397,0.6365149951420485,0.07487510506522621,
 0.41409367797327834,0.563012244282784,0.5516901185201103,0.19607625759547476,
 0.14182543892004573,0.1900554930214069,0.4618511294569404,0.7860808747686621,
 0.32302398619926753,0.44393506559704043,0.3885701966633124,
 0.42794580143225114,0.7620879067158508,0.14121741740306504,
 0.9282863116467162,0.3128480671543591,0.6419246567328173,0.795504309293003,
 0.13451266936230621,0.5889215458938342,0.9727239687084247,0.6984721311863762,
 0.7688337821758353,0.6685844877624704,0.7978643565772924,0.8772214399130938,
 0.22189215798187334,0.9382581572295245,0.4085317317221254,0.270090425424109,
 0.5386407676981604,0.4372140602001623,0.2504826344363509,0.3139053531056418,
 0.26735822976516155,0.02509476352359108,0.7829541152528692,
 0.49331168749376975,0.10379533445185696,0.4772665150275418,
 0.6092659178390379,0.6001173320312663,0.41456485253137343,0.8265494568092043,
 0.6953899599267882,0.5476567099662085,0.6986369068817136,0.533624022957077,
 0.5300057892204162,0.7943194593675754,0.24013764486550326,
 0.19094471125969126,0.20389883674066933,0.6524576776875214,
 0.14343957462719215,0.24727538631304802,0.20849524379952922,
 0.4072211058446259,0.7659171910901824,0.649868673471354,0.8223936539795911,
 0.1294305381236831,0.9503888961235339,0.5417203912338755,0.6154401061637179,
 0.8919450728455189,0.9667790794354632,0.09139234090252057,0.1033616702209551,
 0.23209961646729527,0.035512568041603854,0.06961981941275786,
 0.6921778248887758,0.07112708577105098,0.508898768044021,0.6779753414334275,
 0.3870560031245087,0.2854392447237294,0.15510862863934205,0.6998363019533385,
 0.5065253481090461,0.9370654291030706,0.5172942973524263,0.5400777770156459,
 0.07740682875091853,0.6298601046624762,0.28456414141993724,
 0.9773312911598544,0.599554073241133,0.6763883361493791,0.6650317475303514,
 0.6428390547130446,0.2599238701684371,0.6420153450569286,0.5483822656328178,
 0.5841285372345626,0.34609999102690003,0.355987364435581,0.9730571829603244,
 0.6223404580352799,0.603954070202282,0.21685159536887388,0.05529862942242991,
 0.9761049424039112,0.010020588856221924,0.5541620766123224,
 0.21039460119534903,0.9097978955179564,0.7576657601746537,
 0.32422833012729857,0.6677192675173647,0.7950812680389674,0.8470550062266429,
 0.07887051078909102,0.8315004830849536,0.7502668401170001,0.4734121602660536,
 0.6383735922633322,0.8866070202509704,0.9537997844561886,0.6614644293497762,
 0.7688443684202596,0.4154375636988755,0.8116226339586775,0.2189590649432659,
 0.7484469815344877,0.022437502673504972,0.4774248027595177,
 0.8962763931534611,0.24441831373133993,0.9198664594610257,0.1213996249862136,
 0.26529772600161006,0.7041234677450291,0.7793766544917469,
 0.05442033393524848,0.9765717861353375,0.4762899391526353,0.6644526601909954,
 0.5145216307961774,0.471547707237418,0.11394750824261346,0.05727162352097315,
 0.42143882396962806,0.8812610289833361,0.15185500319232115,
 0.09911728450537216,0.7389896396638767,0.9601451303565984,
 0.17843684171544538,0.5581316460684952,0.7998762081724475,0.2750365095180598,
 0.8845287413708528,0.6562200277801493,0.8400905655889226,0.5590476456098259,
 0.6323173690984816,0.2365447921259327,0.5779308130629758,0.25619778035335083,
 0.82846481219164,0.6794353475930704,0.8635860538726893,0.9937604035789643,
 0.6655269618721671,0.2868409189439749,0.22722085264035008,0.8129578021391218,
 0.8830630001741455,0.3605313016739523,0.19833888825017298,0.8880561903179925,
 0.518717773746532,0.811117525219221,0.31993868694003824,0.06515888668405745,
 0.8307511995145179,0.906519730831228,0.3951363733377412,0.7352323143373206,
 0.4627631290141515,0.7075577262652636,0.045212019933026926,
 0.10399549991349033,0.018105438725315892,0.596524262736049,
 0.7946410096766264,0.5950485245666369,0.44168968491132965,0.3314082993960189,
 0.41663541260927206,0.9197513420926013,0.8573289935870662,
 0.08477205275437583,0.042882320204975954,0.34821027223108536,
 0.5312561195636138,0.7200010490145,0.02199656915513437,0.8546189824489101,
 0.05711386340208757,0.954970707714087,0.5033257601109504,0.15261826936397804,
 0.3960370307000265,0.6645057297892254,0.495616774430919,0.3471416378594895,
 0.8416597140789368,0.12859727286910771,0.8085598568431102,0.2557330849860858,
 0.9980958771750088,0.31287694341314176,0.15661069193795252,
 0.8743591148811762,0.4501535781771763,0.12229914362722016,0.3677337277512098,
 0.8156435590832722,0.9400535026695364,0.3459072673312442,0.3496399386626803,
 0.6903071885645908,0.3479336755909912,0.7733893369710417,0.11386233044321015,
 0.902440107864465,0.10953906142625303,0.9065460996403262,0.24239496839420416,
 0.5432692561120617,0.3848694811797708,0.19194352747964616,0.4292567458839337,
 0.7196758023178083,0.11833518383362407,0.0891773043702958,
 0.36239091212423324,0.8640639293848347,0.1267213264632706,0.8120647440932198,
 0.6387865093430639,0.9091113553450676,0.765680524873106,0.16860564114240661,
 0.26427334549343484,0.8436580099426996,0.2888668840899039,
 0.41494869562232506,0.6517143760336972,0.4533877782896243,0.4344437566048206,
 0.7859355630407314,0.2232903473989285,0.7564334191472265,0.650911707058439,
 0.8783250340474461,0.43555287345512683,0.11683049824645797,
 0.45693444470135625,0.8664518318243735,0.15198688426504225,
 0.010206632365025059,0.5909336035408795,0.6292558000628937,
 0.15366509090720792,0.2328872920686451,0.987709357696945,0.7709520480198341,
 0.8375949957862445,0.530474474281282,0.3602413600061476,0.7013896355912341,
 0.13905340011582712,0.6968355410632683,0.35684842895047164,
 0.45063742462674605,0.07724995754553465,0.15252135739253625,
 0.5809635196176803,0.6441850629124823,0.4434947796998674,0.08092130225629524,
 0.7199344445835916,0.35222682200297584,0.7287459612350408,
 0.42975183993149635,0.7458201086687876,0.45469082665258775,
 0.6412116108252425,0.7893750220095295,0.6594728572611279,0.6596659295517373,
 0.6562865309332914,0.8176243840217978,0.09472790278599696,0.3009804073347344,
 0.8334243816033357,0.28282989271633396,0.4460475090681114,0.9367065159429094,
 0.46845270522791127,0.46146412646356283,0.3840601207707255,
 0.5729067613046924,0.18399537316097825,0.6689016747070262,0.8049739324516022,
 0.06289755333734026,0.8474112336845465,0.5122128094440871,0.395901907550371,
 0.3163619921943961,0.570467468469389,0.28637873626218424,0.896132029028636,
 0.3570264012872737,0.6087998459258597,0.2286973943091154,0.07036197018009993,
 0.08605472761171629,0.46695590525512976,0.9971007605744454,
 0.026097505606756055,0.3064141464066583,0.5605422335833158,0.443618835166255,
 0.2280792285120371,0.2256529406902651,0.20490013992811695,
 0.46202882479140495,0.3142019098423121,0.5167203370679482,0.9646559551228431,
 0.03789372641182265,0.8740944523150234,0.9135127080286944,
 0.26044577872613905,0.30725062463572916,0.17625854028540477,
 0.19357770018117537,0.04627462613965605,0.6780555710803671,
 0.5487746934301644,0.33227347551392306,0.09445618241137022,
 0.5048179301422772,0.9868565001309473,0.7537348631260079,0.4756188769245787,
 0.7596754397043972,0.11425637156763147,0.061955926203388634,
 0.153314566501489,0.04595075689440242,0.16951389938100048,0.6766257184909655,
 0.9112400120752422,0.9972514946976613,0.9780529306418595,0.7307935704620281,
 0.4049765645900639,0.7900276795219473,0.5550974406897362,0.33364429142114793,
 0.17202681690969368,0.9109649731595989,0.14788332951607197,
 0.6765544747735013,0.1259652990392881,0.24071801158849226,
 0.14860216698483053,0.48287748649160145,0.1938154966549327,
 0.5520158679710291,0.6748434595920552,0.29244910111385725,0.936641341647471,
 0.08286641412502749,0.04975619834531475,0.5701938002492851,
 0.6158769315244346,0.5625792572371726,0.1257939322383752,0.04931626494707386,
 0.9596013813431219,0.3811944919187381,0.5122267644093305,0.3119994093397189,
 0.4436706069094808,0.6461673633285625,0.27380983109888646,0.7567541286969375,
 0.1874412939362492,0.8866484918178239,0.9309324350684728,0.1609801588446801,
 0.6032698098702785,0.8289616724948711,0.232234611036934,0.8881323499016979,
 0.8439808932034164,0.383709189753597,0.9793160605770713,0.8901803897725229,
 0.400044675167171,0.6562900171101043,0.58796776832061,0.317931809472918,
 0.5378925841577474,0.3340453637422285,0.18635643137450486,0.6202517941361523,
 0.0627460561163784,0.8013132629087433,0.24946616498559637,0.6413746275443816,
 0.5536456349722176,0.08443399731758527,0.0036890975205974286,
 0.7583991889432516,0.09650726032065349,0.6868530918573655,
 0.10653405654168302,0.04159887892265279,0.039155313118597435,
 0.9286992766207042,0.872490274414426,0.04122422538101134,0.18833273181512344,
 0.40044156534444386,0.1677288035810206,0.9212418867237692,0.2873233717778235,
 0.38731364654245193,0.9726367169853558,0.41643585883800016,
 0.05691188304559047,0.8737802831991255,0.6806537638554991,0.698806285190695,
 0.3333194169185394,0.9035381292907,0.9163213100930856,0.46839273077940013,
 0.8414719666144533,0.10853079997558068,0.8198213840740771,
 0.11330460065261849,0.20279742154126057,0.48368020204305984,
 0.30519507809001123,0.432986305181987,0.011734485145227347,
 0.6197042576561123,0.4184270846962619,0.014636677151587785,
 0.30738652820216394,0.9331895472719322,0.8577907760296766,0.0760978774068819,
 0.12107022976083726,0.4023702990013207,0.8031693752291414,0.8590618171677401,
 0.7426395570647438,0.30077400745348515,0.3917543241001169,0.3613359040592481,
 0.31320252263886983,0.7793023075413359,0.9460116582764612,0.5249016299612325,
 0.826958440545691,0.3651801082514893,0.5040934257956052,0.30442074308113043,
 0.4751057739542007,0.0853398572562617,0.05717080799525476,0.3127372035081314,
 0.30481613383700445,0.8545717274874223,0.1951478700590401,0.2652478998423493,
 0.6225554655861387,0.25555674976412934,0.8645419585415921,
 0.47350089814460317,0.8000008557883164,0.15060592797306338,
 0.6728593770239764,0.31823355260741093,0.7521217645065198,0.9397223193583627,
 0.9071362194283805,0.8363652940250785,0.25330651243180324,0.5858018802406246,
 0.09625127867881744,0.549009711065256,0.1798976314246551,0.07299195026568295,
 0.70306240354083,0.9465743449709219,0.05204160533842983,0.5677245112675754,
 0.09195463837080553,0.9213121062745508,0.8938401456650972,0.6762270039496308,
 0.9411765324954983,0.8108638409131406,0.9798637321793378,0.7002299330549164,
 0.3453287056761144,0.1865601638503157,0.0703093210867003,0.9277808975855025,
 0.3137813861708987,0.24540896596044925,0.43653707070201897,0.902996187746385,
 0.9608559090607737,0.6382301150686428,0.4470571717639119,0.1887947120766481,
 0.5943035563814534,0.8953807525584141,0.9486668487171538,0.4786595392051666,
 0.47784826124248414,0.5871605035480812,0.09721978370718543,
 0.16299067205348794,0.627598102588469,0.20712614513798278,0.7383598818551863,
 0.034866020282533605,0.4288311623062031,0.19480201313241208,
 0.8929597384833492,0.53882272956993,0.7867868323540631,0.9828368865715336,
 0.22140485435671153,0.35278595363486753,0.2028543132121884,
 0.16524503663108192,0.138584794571876,0.05331125885260668,0.6816338489518832,
 0.6894316394711043,0.22251602365483958,0.7771726223168901,0.2527323789232775,
 0.9543594990011314,0.5422794427334792,0.3805190972514766,0.25218688102361475,
 0.28665325767208505,0.57437678277457,0.18942724888959372,0.7668259222598242,
 0.052527010250685735,0.45457599084939193,0.6602028545436599,
 0.9468570112600129,0.138361797844179,0.12430302378398417,0.4011801622908594,
 0.5815205001465833,0.18014824389789863,0.4552883281240019,0.9187180805055891,
 0.941414834542289,0.7584755696891072,0.886849825374594,0.4875240750178066,
 0.05837748774028151,0.6907419250271044,0.27900894320660874,
 0.36392112984064084,0.2714663216134634,0.7954917937669848,
 0.35276132856545406,0.648553090145402,0.043781340978036765,0.74202540803019,
 0.6977739353716781,0.4757887542903507,0.9014303332590201,0.7261276100644836,
 0.104764483678352,0.03618999271983625,0.8325878554679946,0.7434976167743048,
 0.9209757474742537,0.25117666527506544,0.2470344287866587,0.3764054667708332,
 0.3212843366918281,0.598579166176226,0.2847652641501076,0.37513977080517336,
 0.8468151076041537,0.9154935804975517,0.24226429655206883,0.5710934952044956,
 0.16863141315807784,0.9921427721952233,0.3971956612598251,0.2872140314234993,
 0.44636777544446793,0.4981247622174493,0.6866677780683593,0.92555651874382,
 0.9428171749462866,0.0797934911398247,0.4430690696228601,0.23830901179737507,
 0.7825366499645082,0.5087780320605785,0.4990568809995528,
 0.008420866514866288,0.9628811869533538,0.9055328227008721,
 0.40745928025473876,0.7340880067952731,0.2109153529567801,
 0.20143767446512117,0.3622772911372504,0.39109735954481895,
 0.2575352915777769,0.44099417237370486,0.9726646996544522,0.9544894418694901,
 0.8400278711953798,0.2345047619005708,0.571388105208341,0.6972645579276532,
 0.3007229246530966,0.2584510554261168,0.6123106935876514,0.8799854603279471,
 0.5273749494803636,0.7420586816378717,0.9688984997574246,0.5692255667037819,
 0.30330905288153875,0.7991202227372505,0.256676868432393,0.481483207251469,
 0.8780365662611944,0.29477520243689515,0.5605418443363526,0.4911244667132404,
 0.6463357755628583,0.23552198528784662,0.6123913484156132,0.4849920741435828,
 0.8780457204538767,0.1498441749029431,0.7188351269421918,0.4205628614187199,
 0.9414834420090721,0.6789487257553745,0.4052028927953113,0.16445934634671944,
 0.6059080196730451,0.7750432087550823,0.3853531165216172,0.8145576934120191,
 0.5340724894945648,0.019775098130094038,0.6312540353797801,
 0.9990264716022819,0.9297863101072275,0.19253135534654664,
 0.35438754168560227,0.9136362759278965,0.3742338245227105,0.6396705914927596,
 0.08311218230070805,0.44349178587805027,0.6412155109992216,
 0.1639564814212635,0.937906822877103,0.8911324062593562,0.36687699741217805,
 0.8728710987260804,0.19728648453166575,0.9457572645458694,0.5308940480382789,
 0.27351994918708433,0.19522829108123707,0.48828686411518407,
 0.14865465006014178,0.6404433510585819,0.621108990102555,0.6444703083983886,
 0.6664301591929602,0.9055705840740996,0.005180524234024819,
 0.4809687431078493,0.052478876262426755,0.7114933924130029,
 0.9483675899934298,0.5160781165730784,0.017505946085146662,0.345886133544289,
 0.6638648875604176,0.26997358715114994,0.5908238741778169,
 0.46253365488816556,0.6104379122436613,0.9169912106401902,0.2120037343579102,
 0.4787941474136289,0.845852420280515,0.0513834754800212,0.7767167281243329,
 0.5281460436402227,0.975852325694499,0.09711585519803267,0.43639949538297107,
 0.42897437356471946,0.0029152887346637346,0.003462056974212313,
 0.649768250813013,0.6869222171089111,0.7788041293375415,0.5025654531007705,
 0.5135109073717747,0.6930404469206128,0.033428956965058765,
 0.6110836400777343,0.603707103585338,0.8837801368694896,0.2548441917284304,
 0.7723587587101669,0.4010587069606275,0.8446121177424557,0.547243725513995,
 0.4576831007678377,0.6615258511165499,0.6803676259484417,0.9355732639443548,
 0.8669509131915489,0.04648636351608748,0.30541926297450117,
 0.4850860285779768,0.9463992976642537,0.533812746191047,0.03410267150628987,
 0.3928403471543792,0.7221463271252877,0.8906536174120252,0.0639631237589392,
 0.9084500692253813,0.9337240324088526,0.11303161859304889,0.5411088018000472,
 0.4507384243296618,0.38443313057978634,0.4392454649284935,
 0.20099167048227096,0.9924163132801173,0.4578686743652456,0.8370573770407989,
 0.8791518888273674,0.19716268163372386,0.1578889463108064,
 0.07258954818834418,0.14917327489731347,0.7012461107028314,
 0.7003093543190925,0.3935574300426009,0.3855243977250262,0.38818892320812537,
 0.9971489806020974,0.758264652382072,0.589193011420351,0.7179703987020007,
 0.13850808809696868,0.9516240344045612,0.042609959358699157,
 0.7461591231809003,0.6787663987080016,0.9077570340486384,0.6551728621738726,
 0.166214254370187,0.17359675701167143,0.022903207764038203,
 0.47698158828282033,0.28074266137183235,0.7694741179074913,
 0.8447791087222571,0.1576629367246345,0.16199752700293035,0.5822108012058606,
 0.4424845029016492,0.9004720391469564,0.40767642766304213,0.5617764091818147,
 0.15980407948078235,0.0074395260284861875,0.0865115010012002,
 0.41429452262160404,0.5438486672031253,0.7810328327663296,0.4295664331353336,
 0.7297402999966448,0.3471163093009526,0.32164663631210655,
 0.19264987547827694,0.13388854354444735,0.019257517341343955,
 0.5361858061322511,0.03508242429072994,0.8340074199284343,0.9096111213156646,
 0.3187150277399161,0.6324458873520242,0.2941602807174021,0.9536127472041303,
 0.15369175305771954,0.33165333544954745,0.21724412861573095,
 0.2079770686009077,0.8553169770927906,0.353336813842427,0.7235837643776898,
 0.8284650266169764,0.9418741412559036,0.39255962056966476,0.2665043838586195,
 0.25103674383934693,0.9590750495340621,0.9595916471357288,
 0.36052315119912215,0.5170643751440792,0.3782496759653614,0.9726636839090943,
 0.2666968294209655,0.8958267520577328,0.001271301026781746,
 0.04594669811108454,0.9894728953797852,0.29648118580348237,
 0.7953783342655941,0.03437958203255831,0.5165154027642709,
 0.08681204788783703,0.986252942517801,0.8750645014851967,0.2811674146534272,
 0.2847955761432709,0.7906378003078436,0.8972062781896051,0.27645156150372063,
 0.9780728684152669,0.904219536277633,0.24518797542388782,0.727114416941746,
 0.3277039776363231,0.0570036807485933,0.9248188737871288,
 0.032477115601155626,0.34995411309588975,0.7896882003675078,
 0.6478885660874643,0.03742443957397057,0.056418039017474,0.07813972119647283,
 0.8927249663671433,0.44644293905177,0.8722537713609445,0.9125309637318012,
 0.9867107415101853,0.5914188087275745,0.5698142128998059,0.9996660190307756,
 0.02223544377457931,0.34647374669480624,0.6490526282666728,0.82451259002313,
 0.18331572861234136,0.2510304974767953,0.45537187194612483,
 0.1638745151165849,0.6661591731287497,0.6193501066161515,0.9279182073913896,
 0.29071139863743667,0.408197491400633,0.17264878174703235,0.7890846880865003,
 0.9314900861654108,0.7868825912855619,0.6632757090379059,0.7380395768330295,
 0.8263201880854254,0.19182006851840927,0.5583209886796359,
 0.45273799731190045,0.7257398782707023,0.34672861532607113,0.393490822128156,
 0.7055399474447616,0.5368623139205657,0.18618789461696916,
 0.0029742308689054697,0.6438013192584023,0.5544829987100973,
 0.5340797124117904,0.0019179660383313868,0.7696383192175476,
 0.2719534227514737,0.45994729015636093,0.2217986549690707,
 0.30650902471940555,0.5187429927160169,0.18734917036513388,
 0.09883795527981687,0.5032353007481912,0.9535028404284568,0.4119498551572225,
 0.8401676286192267,0.18460412694164174,0.805848626125535,0.626593356426034,
 0.17372118001868364,0.19784712656534797,0.02592107157675616,
 0.8581405146358048,0.9225428736076025,0.555457921930647,0.9521488646341203,
 0.8866172036315882,0.18251544901867622,0.6131504839785351,
 0.43534293501270804,0.4535018457246489,0.30030359214097824,
 0.9344350774602828,0.8351963297930051,0.5732937173559869,0.5931102779926614,
 0.6717104453499633,0.1701950784899221,0.7883331646939582,0.5550641729049136,
 0.5242417580973956,0.74199200952919,0.4328604249790198,0.048143928576433126,
 0.4790295560638491,0.5937612848240766,0.31202463484980225,0.8221302073369015,
 0.7734505743139984,0.29916924967752534,0.34430322193983043,0.550917267018873,
 0.759251253785329,0.15276329156114943,0.2679107553384398,0.7014744525158476,
 0.5048819189246059,0.6308763454132161,0.7451975631888891,0.2348226729106908,
 0.20022997795527475,0.7200894490856263,0.19173670785665975,
 0.054002370623484564,0.2264892170106202,0.9374773311746935,
 0.47917892752180413,0.32547826696122173,0.7701483217623089,
 0.9480717002961924,0.8799606114220072,0.07676819847692629,
 0.42309546188981995,0.17053304897812582,0.22179160963454392,
 0.7149911500475402,0.38174788791291425,0.29592861600984166,
 0.12310576512363858,0.985983723439023,0.4105173592448277,0.505743845402027,
 0.5432892964910696,0.5650919501969114,0.7830947830401012,0.45268067420027136,
 0.2856743954849217,0.43814894149237227,0.0694138946322227,
 0.44420362503414734,0.5352604740938769,0.6591916707331525,0.7298474365443577,
 0.810871497616122,0.8123242063247386,0.10204064350881892,
 0.046642719396988164,0.8424405077288621,0.5373902371492261,0.462813585304665,
 0.0845806228429068,0.5737609875953852,0.9437154379961145,0.9854076102326084,
 0.7594199580194563,0.6880845236723302,0.8505775876990758,0.23604541708634297,
 0.6024305990980653,0.8835938791152695,0.44184240644925554,0.7601714813733971,
 0.47974430055967054,0.7780986822627615,0.3599111810071769,0.7583838599857791,
 0.08055337416651787,0.03604800578298384,0.6449160141916631,
 0.2904102468528458,0.8406233143352694,0.15987390589869765,
 0.36070723056801524,0.10881318507714977,0.45255230488970555,
 0.32628886785635347,0.9465617592612976,0.5490731792700259,0.2634653679102592,
 0.8292691019946811,0.9551081247595122,0.19266675601560868,0.343097504790708,
 0.6533969833794777,0.9431752964516287,0.4802078799353211,0.2782461502872735,
 0.7173882957115213,0.582859847423473,0.21487930268943034,0.7085671282953687,
 0.11223935324257028,0.8810072805044942,0.9521276511135417,0.547676389194597,
 0.08957318295606487,0.14017444826394043,0.9068553123341078,
 0.34579757264473565,0.5622431350287117,0.6735230304335774,0.9413483499496321,
 0.07221405168852768,0.8016115669126325,0.6310818505151691,0.2797109418214285,
 0.14680818360583314,0.36269347859303624,0.2890644941071392,
 0.5093106107197357,0.7920747794687123,0.46064662726426975,0.8026868251459605,
 0.13500801572773136,0.690632557462753,0.9884612358329506,0.12194041202179451,
 0.16315878625301994,0.03200432717321233,0.14851021148115973,
 0.7207496193873273,0.7766616362778715,0.6810851128524369,0.30453565174782726,
 0.3765987536638883,0.9957760902836852,0.3434448784465799,0.6662549116628189,
 0.3113071402247085,0.6207764959654103,0.8833814603155103,0.7621223053131118,
 0.7955076330494201,0.29546019049651473,0.5460686806898742,0.3609088146453806,
 0.7683966356653151,0.23345560516081687,0.7550234204065185,0.6031394900940412,
 0.27443212236383585,0.3091287882613787,0.3569232233108234,0.3567492981794582,
 0.9518096009189958,0.6483350012494048,0.7731694391236639,0.542929093447729,
 0.3527056997853726,0.5007260615705549,0.3813449456436727,0.9263260530233577,
 0.45164413733257924,0.5707079727103409,0.510805172713288,0.9836694090174701,
 0.27599224439355696,0.24266006016302588,0.7233946191716463,
 0.7060447584022481,0.7288242621999027,0.3713991497148752,0.15886336025421222,
 0.05293283833023943,0.6066868554206475,0.633831120543502,0.228630134885493,
 0.013379923919516368,0.9421405480485944,0.7205845489231828,
 0.6385241508099971,0.49090040498804277,0.48840312008194575,
 0.6375218327525565,0.4590558631527967,0.7713455525662298,0.11678894937818685,
 0.2478702518231084,0.9833344878993108,0.16569578424099285,0.2520423056397043,
 0.35071824042406474,0.7128152799513336,0.9875989565547798,0.3959411636688477,
 0.920644362607119,0.6529462547861146,0.7292321253980316,0.9758636220108419,
 0.21133261394272762,0.7139792759107103,0.4478407844500194,0.6836218190703933,
 0.48777351120951873,0.14904470643431822,0.5938108436629549,
 0.9546054982001211,0.24837225225866488,0.6202820359863751,0.5961001958389981,
 0.47501382328282926,0.32486930833169414,0.6800800415647252,
 0.04012210583830722,0.3635093413054503,0.2591244320976621,0.8903487642552838,
 0.17195193866866965,0.7360693437403578,0.8850981347657786,0.8574118323294285,
 0.4886218865929052,0.032977461753207704,0.10418485220647555,
 0.4853017906333569,0.8987455295966232,0.7085873560389182,0.7831004172014335,
 0.5809073317143629,0.953758601993811,0.7576494564181575,0.37702870945728806,
 0.5017542943746831,0.9496820802142273,0.06930673321680714,
 0.008339495147713594,0.009677149372061766,0.3365080202764541,
 0.37675238961738,0.08375500031390615,0.7903089931363351,0.5295444044606497,
 0.17016272999023985,0.08817103267313986,0.511759363644492,0.500669499471895,
 0.6267976273301022,0.7166979516348465,0.44590405371318864,0.5817013869201099,
 0.49439419336763535,0.5691328561542629,0.13329095432709925,0.618062252373259,
 0.09846151213583687,0.09453564673384862,0.7567161988341085,0.602746162913464,
 0.9247181146877235,0.3301706074985252,0.508443242957616,0.4742590040627527,
 0.12542709491954018,0.6733437665073354,0.9128074740867778,0.9856169605921209,
 0.5314510649315339,0.7975850448346264,0.1198687622821113,0.10999556120649145,
 0.3507055616298297,0.5585111846874572,0.3531422261206669,0.9458817670479182,
 0.23619015561456524,0.5077980051997549,0.34597387925331125,
 0.7700034902978959,0.9752534055815237,0.9887960840578963,0.0916177052396342,
 0.35049872177720776,0.06659547065704252,0.08443376902533772,
 0.3548938970036084,0.46035323585404275,0.607833820923973,0.9645159168521369,
 0.01766239101446543,0.4698391587902301,0.9437716093349348,0.7711950400974252,
 0.856954828212136,0.3753637473791728,0.551467262904928,0.6896191620079317,
 0.9937047703415371,0.023966129601668995,0.8288645562463732,
 0.2371638980975237,0.31477923922963247,0.6418613384394249,0.6004785081387015,
 0.7620030987044258,0.8071365233825364,0.7143784643236427,0.9968928846884341,
 0.177789506210425,0.6155577448188098,0.08266759864485218,0.4396620247669485,
 0.8158061712180807,0.5851642967569863,0.921434327774525,0.7633434662880039,
 0.054301250619142705,0.3370335715633205,0.24770427120197724,
 0.8057333903382089,0.43373030276177027,0.3974769959670952,0.884607624097782,
 0.5974361813213341,0.6930980560967728,0.07710549013147694,0.2961428342391762,
 0.741317674470344,0.9689273626682114,0.06835143349987949,0.9039193960255694,
 0.6975951389020694,0.08186533531764073,0.8498609770003752,0.6242064796969315,
 0.6488998509218551,0.9140165608634995,0.6114448551371097,0.5057997518334324,
 0.5149438311020818,0.812295055207211,0.8683258111178822,0.5821800005464162,
 0.5335219242365451,0.5572063775271582,0.1030430164887497,0.45511408612804716,
 0.36071291055831745,0.612043259540148,0.7344793928401322,0.7022372885822563,
 0.9744108142670315,0.6091978657294229,0.0876977748679384,0.5149602414026817,
 0.8373276090854087,0.9294201417000825,0.9127593156984878,0.5752203375731759,
 0.6872739088155123,0.19071305515678627,0.8623096897388197,0.8439714423493203,
 0.9151323044224011,0.759949045104055,0.784952891319756,0.6007120356908555,
 0.4874392572093291,0.6057654114650417,0.8197772947628233,0.29961852520858323,
 0.7943005384357054,0.38343071756287705,0.6012467916042605,0.38330092518251,
 0.09995883525142191,0.5363544291601929,0.02239506084264209,
 0.3658333111330839,0.03663795885438004,0.00739498145425288,
 0.7609360471012798,0.7551472041006922,0.8888331798141424,0.9782927560011034,
 0.7490155098910306,0.974093384287098,0.3746256691759019,0.4197304272675859,
 0.14036909622280402,0.31065143677380136,0.2771570542373867,
 0.9073458906875307,0.46870207403395137,0.46791959148202156,
 0.7522154718615855,0.028519220032448223,0.6835632590824482,
 0.8440536361944428,0.46404003527732807,0.40455714375151974,
 0.6750413729008304,0.30352899661546606,0.8661919942656444,
 0.31007954147249017,0.748585556664362,0.7085254542411236,0.36810997072904983,
 0.6466614160477324,0.4102238882177479,0.01693391489848306,0.7471438216049178,
 0.5801042997760345,0.597188640315444,0.49761603033212953,0.4225769832600833,
 0.1623601491446488,0.7764793329408104,0.5883052230094964,0.641396405144557,
 0.8353953660385752,0.5661709868806257,0.32313975701464703,0.6770821735534225,
 0.28517995840153765,0.3790596465942959,0.3636863134706233,0.7726451005274442,
 0.7703796511912879,0.5000059975187464,0.811976075600834,0.8092804982682341,
 0.971295839409551,0.8684133141453568,0.322164727973801,0.21192116970983033,
 0.6857757743231322,0.6436290098114066,0.45448108779259855,0.0868588274564297,
 0.8361351504449577,0.7688660480667668,0.10939700188798895,0.895562488154868,
 0.6671133498474098,0.7059258514325384,0.21026567926901274,0.4200724019812728,
 0.325000631308784,0.27397391200743904,0.5734946923379668,0.8983597481791958,
 0.7821429596370729,0.10145887470786508,0.16263071008750518,
 0.7986715078439909,0.5125705667523269,0.2937075427816611,0.8890520761219448,
 0.004676460692059514,0.19211262073211144,0.9055444105106571,
 0.4735260146974679,0.7726331216510729,0.13027277479355026,0.462281784934337,
 0.3355208298057839,0.11205045628452359,0.13431471131708184,
 0.9903737232158105,0.6099722409793759,0.20202758747945804,
 0.35826664505053285,0.39854659790184377,0.9386141445939157,
 0.8693596989332648,0.6838110006857155,0.2139345950355297,0.4670799289217077,
 0.11558370391139972,0.3577863850387568,0.38700531987525477,
 0.7325416374190277,0.3487455727683554,0.10843292348519573,0.9268079176527562,
 0.971251848001521,0.15474163699627286,0.9506609986143651,0.19190764510514247,
 0.8238417076317301,0.8007469227101864,0.184530061478954,0.6712337327823166,
 0.9590311805176502,0.9261709340958524,0.05327436524886264,
 0.25024616606175676,0.18810539479854793,0.3609717150184978,0.395810461754051,
 0.1765051024598434,0.10950526415064854,0.4290552213530163,0.5912310034056629,
 0.693118425217963,0.0125843317937685,0.9355186034661438,0.8777627954436279,
 0.5830267974796584,0.5893688913754823,0.8771241727912626,0.3582809979219371,
 0.4419627956993858,0.8734752273516498,0.7328856907292401,0.06740502857390984,
 0.7526319174866807,0.3499043896044576,0.6238300875379615,0.8801584818071855,
 0.4575706008171152,0.3018736898895177,0.27245248921897636,0.7436933173493034,
 0.9251814773885413,0.42520161174745297,0.4526251899128009,0.9762030647377304,
 0.4416468313344103,0.507127822940053,0.08131905443993315,0.14707562806974506,
 0.2935823533839259,0.08852751828447936,0.854814026391321,0.6831437290013448,
 0.00889098966127233,0.46853231825062325,0.24674186065319625,
 0.5405508919146593,0.837448745620507,0.851792568547778,0.8244386706310235,
 0.2151066874443056,0.2232364592048085,0.05002918967542158,0.5516614020712342,
 0.8002786253387634,0.8383696816851657,0.992438813370494,0.5862688519982341,
 0.7304833256452752,0.5227494599782674,0.5790559066067095,0.2644663739882559,
 0.9293166849503521,0.9128385966457615,0.9419267664851345,0.9401446348780834,
 0.1881281835316786,0.12389616490042943,0.8217386977933285,0.7329984510474588,
 0.09833136916103824,0.2960628307703874,0.21471490422198125,
 0.5468477159824227,0.18252804878877993,0.6646288676547006,0.3177932591914565,
 0.6187579470398822,0.5149834294170142,0.8246519465479317,0.8566753854090208,
 0.5823274151251455,0.710114824276159,0.8172307713330813,0.773942566379359,
 0.4609771823184823,0.1053114265891617,0.12548291745869755,
 0.11496848688359806,0.6074392940802777,0.9521217697773109,0.7819181402654128,
 0.6077831144729493,0.585448781024952,0.8110119833904902,0.8563998390123615,
 0.6782968393450379,0.7325067881145696,0.45152861017494306,0.4269614066850591,
 0.8912530192321098,0.010290177453409766,0.6449480312841036,
 0.9901278136452236,0.9849471428110276,0.8876229578087772,0.5573103866250816,
 0.6954218122730893,0.9703259662676355,0.951415932995071,0.9008402931282812,
 0.13899868030456086,0.8522268287811525,0.04481027264615933,
 0.25983775046865554,0.9029003342439936,0.4291049700735925,0.7912633476147459,
 0.4361741288245323,0.27513522650426203,0.6675731802794888,0.8734425324100163,
 0.9468934731152291,0.7473491388588437,0.3582991777397131,0.7322286755889734,
 0.33554972641371095,0.23550987812069735,0.7928413191280774,
 0.3544395466998832,0.2150000029362379,0.19523627786577868,0.378686899386915,
 0.5082495560352602,0.23742946645838692,0.23090775543709685,0.528471793042937,
 0.14676997619923793,0.050569832019140226,0.4020910841905935,
 0.21917261983881753,0.809514784911385,0.2149521168150057,0.18792847604610152,
 0.19401147300816213,0.693204897867312,0.2229019110874707,0.3307907451672626,
 0.8669003769115654,0.986350795099002,3.5021720856709404e-4,
 0.02788166991722263,0.2213538227170645,0.4195647489471098,
 0.19381081979150006,0.7815953112300114,0.45422141226480794,
 0.6026188814292756,0.3303977643958842,0.5157742208589164,0.43279355386101537,
 0.06279420870404506,0.9293006456798375,0.6809103190744358,0.8179465646312551,
 0.8560355111623679,0.7435352902464876,0.7796989340071458,0.40452687435431856,
 0.7194475634845318,0.8917850950660189,0.9704526088880384,0.7370332310746139,
 0.29329701949229836,0.33227402809810347,0.07448414890288158,
 0.32950436738792366,0.031095865542527634,0.944384420762506,
 0.8212785227739068,0.6191741923851773,0.840075511672899,0.8274948355132745,
 0.6542903652314253,0.18050935736643448,0.2555356635851499,
 0.18525566445581698,0.3159030372570164,0.10939766810174478,
 0.4958797907132544,0.2718320651071038,0.8553295492573285,0.49829751276090206,
 0.5561075016749301,0.2763478557839669,0.6581874946494788,0.8566104113726289,
 0.538243109629915,0.601825708998837,0.03908293089377324,0.05097803467775497,
 0.43444053453175957,0.24557292323426438,0.6394971808238327,
 0.0158628350216361,0.7307166856739489,0.3206713803083584,0.7863153234519593,
 0.045207092064230014,0.33224713107937764,0.07578666685513591,
 0.3018333953364494,0.23339396148652425,0.9372925078116044,0.9419561462917236,
 0.3716186122712668,0.2121492329846414,0.9016947444215384,0.775077446266808,
 0.3024271733485384,0.48162970144658945,0.3339671918749192,0.2658094434452809,
 0.42759169922695384,0.2920183250602927,0.5752658180945693,0.8479956643233921,
 0.7439091556053404,0.9215530850357886,0.2834674802979843,0.4437770090828549,
 0.040897715584229166,0.9751407817408376,0.9333856914410775,
 0.2325420582291453,0.8727764070014707,0.7387570618791445,0.2297805946321715,
 0.0449716151338041,0.7405082662688628,0.408776284492822,0.8178942145923358,
 0.7125997408218362,0.5201467924685271,0.7309710910579145,0.7071133049414029,
 0.3012763459428378,0.9587631560700532,0.8649307060607648,
 0.005866444267327409,0.8956956795239808,0.061361611597601895,
 0.1850453347128398,0.5072095513640156,0.4183884234077526,0.17670404951315166,
 0.9121715404079124,0.6205916643756056,0.9218591965142906,0.8427783880046897,
 0.03692020237337168,0.7634848776748028,0.7713601150902015,0.9852302131249812,
 0.5702911363705522,0.26722515492527255,0.30830282314786717,
 0.6633810853244312,0.6523656335741447,0.6533040979591055,0.37393769575686886,
 0.5218746820465172,0.8306940341343441,0.9842830407839913,0.17658180050453887,
 0.031221455776561657,0.6270884923657004,0.22371851379751861,
 0.3150532621708034,0.7694478119536765,0.00917822878587482,
 0.07509979641899156,0.5206060696721039,0.417645195063589,
 0.003965073235229544,0.8908661335607143,0.8666896507715748,
 0.44540949076876957,0.9942827700939099,0.018304962286908144,
 0.4769847750737879,0.2607870211739086,0.5941465400887935,0.5320564940041871,
 0.8040315751280289,0.7181379347508798,0.6907805877651112,0.15307513102504344,
 0.03376117196632267,0.11820246015214442,0.8305734510348431,
 0.9477681425547357,0.9414175261169186,0.05435545528330643,0.8723834369777574,
 0.45613479743453966,0.10113785711335943,0.2408322091640983,
 0.6722280607143883,0.41139112617064755,0.8631939522386999,
 0.03943566194323633,0.4824101385141859,0.5527448896839912,0.1788145816215705,
 0.5122783106785751,0.35116115552889493,0.5246348368878437,0.945721721125732,
 0.8838240422986288,0.5345051372265567,0.5400358584286858,0.25781869646302513,
 0.014535818458414562,0.9297067719818066,0.9741827562097307,
 0.8754000284192534,0.19800264404327972,0.7516732804901751,0.3273120969725678,
 0.48089382870534025,0.05375237647917219,0.36534349874371164,
 0.26393463001780604,0.2391832234650294,0.16128969164457452,
 0.5626644937937286,0.08244029757958815,0.47289511592146527,
 0.01939321174848807,0.9913080783520547,0.8574335339648766,
 0.39823312090127594,0.7076107383800136,0.6871313352037013,0.7519651530002536,
 0.9117121974488461,0.1278089367679045,0.10279391394598891,0.5167709311943814,
 0.7467310279502447,0.5227597365463459,0.5269292393760148,0.193960802711022,
 0.42918159827326896,0.41355605262691364,0.5740185701880726,
 0.9657800843391678,0.1682358686757226,0.9073469734952272,0.38752221844390455,
 0.431280192099076,0.25213883388987157,0.22520032843707494,
 0.22063618863401835,0.17154346270592635,0.07168563343426393,
 0.9297632256655143,0.664193844771645,0.3639754515469471,0.3102073904173156,
 0.1705996337824769,0.14286961494781492,0.9887534928856652,0.3503795591128789,
 0.16737867069222057,0.07768197683489875,0.4942494687646617,
 0.8146516044456873,0.7778119254565466,0.8576150754667504,0.7354392491936838,
 0.5701491333306631,0.5582093200754904,0.3835380767582057,0.11312675712086973,
 0.7169945008430814,0.5110948191301035,0.5784622621037449,0.07793210563250419,
 0.284608406935442,0.4575389028385749,0.7751261342876702,0.067663704966756,
 0.2978636497672149,0.29774378168307014,0.03877280799889393,
 0.46850824541098335,0.3714249178589908,0.016850094883656297,
 0.8941632425338568,0.021845176856247206,0.06470090613201662,
 0.8627873724953335,0.2762422756263482,0.0802696684816172,0.9045316550875881,
 0.6900810504488818,0.8025722305426555,0.22842818960408495,0.3183339445913873,
 0.5050577562657357,0.12870171240819095,0.7827920834438823,0.6296982768166302,
 0.8338167806595974,0.6457653810362652,0.7881454633783858,
 0.020993709493308144,0.44209966777954024,0.43520681545713347,
 0.4591966560542572,0.5533818970896867,0.5507495813671719,0.5997503207454074,
 0.5708995548759054,0.2685972341363354,0.7316101350385095,0.0787414964046127,
 0.516112616887703,0.9030906733783557,0.6782274573816413,0.09765969175714695,
 0.7610544945454201,0.9842342420967733,0.4793317918876222,0.4854203070182579,
 0.09764573779787478,0.7471097213371259,0.01095149659159611,
 0.40879936735526023,0.2570991697682108,0.7330197380570588,0.994398850948418,
 0.3289785164447774,0.3635337644437483,0.1688306671930454,0.45871862351602155,
 0.8062510143806508,0.15748621033109966,0.9516227656220342,
 0.14743799249069933,0.28767890622517567,0.3316935370541343,
 0.3524575695489426,0.10304406761091311,0.3542664477259787,
 0.33033609400111896,0.8801165834936612,0.250880336369384,0.0782495602936153,
 0.03749718790638812,0.1613518244582477,0.49577751489177124,
 0.7717863500323334,0.7260440905757239,0.7217503349541083,0.7776093715503609,
 0.04959420747270993,0.9304243241585133,0.5807087915750183,
 0.09418404550995985,0.9636164252756041,0.9489731396978408,0.6315225543801549,
 0.7402719084239076,0.3576888014852131,0.2951888669231755,0.82296774389689,
 0.9828609659999947,0.4326494954075,0.2853025879276061,0.16844865740548087,
 0.4697961220180904,0.3980634216997433,0.38469596137925344,0.4369754512984071,
 0.7270111014515628,0.9516891591611604,0.13156306215363955,0.9159818037193646,
 0.3785485834854243,0.5674921448701469,0.6498495833033763,0.3182106309477297,
 0.7337374868333838,0.42919348561981796,0.2600356420521597,
 0.37086307108196337,0.5691686071178277,0.6578090808260938,0.5187860096795329,
 0.1252571391491537,0.7772548149308075,0.24503829344182781,0.5063654573365732,
 0.07893974004853022,0.7495062855564444,0.5002887548867995,0.9718966288947216,
 0.529101930153635,0.13240041059185648,0.5508077455674699,0.11236121366125196,
 0.06935846842227356,0.39984307033074207,0.33514396209367814,
 0.3232358906682601,0.7524955707554273,0.41010139908915855,0.7172552992819399,
 0.5942375808988203,0.7689006743665856,0.4554233896231238,0.3921621839553471,
 0.7439349565315687,0.017213945030365996,0.13066732262407388,
 0.8204986673978935,0.2927748407238928,0.8101942077882908,0.271773343151847,
 0.2846532832430744,0.9538566025150139,0.6963505688302295,0.9877835771412976,
 0.12310772808414638,0.7571686081079462,0.9555483679423417,
 0.14847341379783618,0.6588633872124856,0.3003874790821818,0.9463516580942328,
 0.42967483510957327,0.5333956484804266,0.5303835747661763,0.841015653581061,
 0.5933899466241421,0.15592768200089036,0.5834798876559105,0.6481727385221389,
 0.33289760498318755,0.2587025799955722,0.5692296722833939,0.6083734560283648,
 0.1755696806856466,0.9092213355498816,0.4755579326076649,0.4347726204502518,
 0.20008380690681216,0.09633270636179214,0.8942080095361288,
 0.9007392845826308,0.6641484446004694,0.7171574492633567,0.2962077904862126,
 0.40057630943817,0.04648458147249068,0.5484619780726746,0.640834808696167,
 0.49696352234070096,0.47862786440259875,0.23819673224920612,
 0.8851489915877008,0.32437875762770374,0.8732551968823228,0.8721757660368988,
 0.7764846569064958,0.5519700581941221,0.5982565638279891,0.09421254741796248,
 0.7349200210521357,0.2181962202291582,0.7554476172200695,0.2586053199661178,
 0.7106345157037214,0.5799171232698923,0.9250685819544364,0.32386325232291535,
 0.44095625738068145,0.6626321508178075,0.705385782073756,0.5196381364301502,
 0.6625434346713134,0.3315646917157501,0.19622396610732684,0.3656686452711687,
 0.4603596475725926,0.8728935283744323,0.5593012721359102,0.5488087262455108,
 0.8595939536060865,0.5789897898343779,0.3850707761385931,0.9256382832993931,
 0.488035891063221,0.8879336524116576,0.18313431758043797,0.32484956236572704,
 0.26988698966251423,0.56853408617938,0.3455194398931629,0.07331336702802049,
 0.4321720387721667,0.6429577178424228,0.23614967560872013,0.6623126149627384,
 0.9979794308075117,0.9430113382078426,0.859604730963409,0.8000373648184909,
 0.6905809882038935,0.009419618510544736,0.592667523664418,0.9015680511001885,
 0.7277744038869365,0.04886300224927498,0.2982894220634994,0.3410589241370958,
 0.3499885148062234,0.2062076223459901,0.8744547320750579,0.2468961309544564,
 0.31512301973466395,0.515059454649786,0.5883332592822269,0.6226139585561051,
 0.19525257840245835,0.7208890659073894,0.3000822503348137,
 0.22988998413147288,0.22474094084809904,0.5113851474700214,
 0.8087774658099848,0.9721473506343488,0.13666817729100766,0.9811725854239576,
 0.8841773958184979,0.14304242806391843,0.7672190236213898,0.6181419044384754,
 0.8756824041742436,0.21405906528225704,0.6209312082224616,0.6375564801606401,
 0.4792726030769964,0.9201136108149456,0.47669675925805866,0.6665016751478522,
 0.8176722248904569,0.41364734966388406,0.10737589472195286,
 0.8827274921609597,0.08315190997464394,0.871910800504849,0.33634602705932415,
 0.6699704714988255,0.3584366100659343,0.24539975731889063,
 0.40081664868376743,0.8680857672261132,0.8872009756979004,0.3347175155690798,
 0.13346882357901657,0.5955829028806824,0.04659309132618206,
 0.21024699754196918,0.6874913065218675,0.34669013437990914,
 0.035123962555623445,0.9296011134811439,0.8221703847364755,
 0.3210677469901253,0.9873812903146224,0.11386083200657637,0.898980084598259,
 0.08132212213455303,0.47346230492852326,0.8270513085089598,
 0.7713236711629374,0.1279692049272746,0.3170419835414473,0.5137860674467533,
 0.8982766270835711,0.6922756932466768,0.07034397032768991,0.5082011899152813,
 0.02837598847029832,0.6147502506011578,0.062196030464595986,
 0.3428436433233747,0.896116993935334,0.981279295723453,0.7690438417328236,
 0.656948182139888,0.7845002703526245,0.22877731817713476,0.19853853357671325,
 0.8270399034698538,0.6176749914233362,0.5253007049641498,0.24246177793592416,
 0.4491192376608868,0.2718482088812242,0.2841457011065498,0.13361040701997073,
 0.3736747481805478,0.04526687945087615,0.9464507174613543,0.5717313525250904,
 0.40031740933531745,0.5892651307026155,0.2472407153267273,0.655938808952883,
 0.43641116547997694,0.6108240304435768,0.5940869790384347,
 0.058717779187618224,0.396704193274487,0.34293041295257654,
 0.8737357230103635,0.8937963877117603,0.36154919726695167,
 0.44526186169011706,0.1723431595047754,0.5783632168206982,0.721441197464633,
 0.7577816705989091,0.9242745992440183,0.3034642541428718,0.44105408079619024,
 0.09080649148535258,0.2759680312370354,0.6303091941333783,
 0.43664913964008123,0.13946266085977577,0.8086797050958296,0.971332191145805,
 0.30898057898025855,0.4657416622767462,0.07596060591948639,
 0.5262024626401632,0.4177128883859236,0.1394133878351721,0.5711366103600454,
 0.6719290551419613,0.2970849878852059,0.6570875372955228,0.4175697414663123,
 0.7185242435571484,0.18596613090294956,0.09897575437926198,
 0.02832538559268527,0.02544926879784637,0.3747811752368275,
 0.8671457673142771,0.029357398118003974,0.05578299238944595,
 0.9715510825379038,0.964202728597443,0.2545781104030649,0.9971598676556304,
 0.10746253117608384,0.8090336122227049,0.8896722595248904,0.5106801484948433,
 0.8305615915936648,0.5116218866823735,0.349164112859392,0.286178458077902,
 0.9846334903575998,0.5862445774824302,0.24184448862966124,0.5373248202165474,
 0.3794508939905077,0.14286105803453142,0.06367796146341265,
 0.9073288843214798,0.3477587519200558,0.27844337537509267,
 0.45596337886441674,0.32095875729302525,0.8540735893043296,0.578534149711987,
 0.29082312256483456,0.22915988622450212,0.38555258782784807,
 0.39186366476772083,0.9802047576429593,0.6838947506229405,
 0.009850597121649596,0.9301536715214965,0.49303157367703465,
 0.4352950567850855,0.06094837462261471,0.41787828648015823,
 0.16463646908530272,0.12632524182710914,0.5173120671853645,
 0.21587461882576686,0.8023646616530549,0.41435546167106285,
 0.49798505444268515,0.07250683194344543,0.9893699742634453,
 0.49613478948770384,0.31460824245869223,0.5552205236243211,
 0.9458449416721111,0.6700394148120432,0.9326303702209716,0.7542323842484465,
 0.4247516926495709,0.8448265594758684,0.37380027844029895,0.732392266505235,
 0.7754686027826163,0.8355183696915408,0.5845369007413431,0.7093425893374556,
 0.5507126660832152,0.496663963912087,0.44215152900283405,0.5605394231158746,
 0.6239006527203361,0.30784732099581846,0.2800589095142456,0.5522414189663057,
 0.804072329691047,0.17657992561383362,0.5111355641890082,0.8127078783264952,
 0.8867387855220699,0.7381408794075393,0.7188433949490575,0.3323310777264178,
 0.38785891786624904,0.027081685772963437,0.3196940770437303,
 0.9722888913473744,0.6475181448680573,0.9930395096422636,0.1535971517714666,
 0.15164198742211732,0.2695808090415044,0.9340189737030303,
 0.08046173244217258,0.5777713125877645,0.3258501129321484,
 0.04626048430831031,0.5428932944786419,0.5928348849232462,
 0.41008050668129514,0.7418682671510314,0.8253652181938201,0.6976606613405343,
 0.5071029035184542,0.8444370325002616,0.7513462026704106,0.8615446851941312,
 0.8117851419384419,0.4267369932945705,0.4364853488464854,0.8254852416888997,
 0.7418579851288685,0.7833149891011076,0.044767150213883666,
 0.16100850737010408,0.5193812396498926,0.6973135278264654,0.1486694736459736,
 0.985180743871409,0.3798700709516728,0.5738079050351017,0.38112976609811955,
 0.8206104876776577,0.015484215944744584,0.5607400044617418,
 0.8049381848518857,0.9727368225567857,0.07306090607227966,0.5595559738754226,
 0.2161512103622092,0.25602502147959716,0.07333250668918434,
 0.6095025968766007,0.7577259717004813,0.6990639394867753,0.5715795038010306,
 0.2770613472364625,0.47155152690790303,0.8331184330222298,0.6623996303869634,
 0.38043678192389796,0.7587464837946335,0.712519721283669,0.3973447045899652,
 0.1404040788897266,0.6535167984837953,0.3267871447014751,0.8519226346137163,
 0.11129144336763497,0.6785502176317108,0.06194690582446538,
 0.17712461165014082,0.25764693880156186,0.05122599791858762,
 0.2764400489490415,0.5226683995011719,0.7532965620189657,0.5285979348903185,
 0.9971838623949485,0.3487052213368863,0.5564192042189688,0.7383816495583235,
 0.6282472287008646,0.7493245557863187,0.08648751800409216,
 0.07672650234412959,0.2553625229315012,0.0027904765981094926,
 0.20548858786819413,0.4895589063929199,0.01598295876300193,
 0.6080636438107712,0.27045618804409555,0.03538601518742979,
 0.034322479250040616,0.05672732206903819,0.7227008253822264,
 0.30549830957442814,0.7108673858050599,0.5728284868969137,0.980414525791888,
 0.3757788048718088,0.5603444823501287,0.9135727994779854,0.9279183606508935,
 0.893560261491613,0.31789679150689465,0.6191941978372302,0.3733598326492822,
 0.16889095344280247,0.5130094603952157,0.03741171053764514,
 0.3525951352436847,0.5981357424767243,0.34528606557941965,0.9006362621448144,
 0.8621321715252087,0.10823601602015187,0.6348430810945656,
 0.15570620263490376,0.27273308412602526,0.5890917132972998,
 0.2981209694082084,0.9843966528086096,0.16388374581996035,0.8464728436919425,
 0.38670413666511194,0.6204477472535077,0.2934800990125135,0.2513369438186981,
 0.47891402469764743,0.08945182129489537,0.7713852018144879,
 0.5902801924666616,0.5088701277964196,0.4966459524351291,0.8151102693360796,
 0.9853129110759272,0.1904825034470805,0.6225647778808903,0.6735955049624915,
 0.8611467459841093,0.9905377130395204,0.8908156049256427,0.7073841675300012,
 0.888855017497739,0.6118392476186124,0.8177988135735473,0.5997377775284807,
 0.5843056558073665,0.0902778984504492,0.46530858823862653,0.5120789276460902,
 0.6441051758750588,0.25581972465329283,0.6747937054573893,0.5414910600706441,
 0.9583378492701924,0.8272276139683026,0.9799516462071922,0.742229106139636,
 0.9693402572755409,0.6178287437056785,0.5014492621270702,0.21834906016741784,
 0.8869195497948278,0.5547069070456933,0.9116460336080836,0.8045736255848328,
 0.94549620646323,0.8593358356708009,0.4555000356892921,0.0992562495443009,
 0.4179848168305552,0.6214931650002807,0.169479629865531,0.3735678336739654,
 0.3113460447868057,0.7522131112148511,0.5904283472933988,0.2890543954072561,
 0.0975353136729028,0.7648052089853816,0.04495878406318543,0.8669253158373177,
 0.6639778861752791,0.770020006737508,0.13524484455504504,0.926206591904178,
 0.8736246171755787,0.3058435786656606,0.23163337838165787,0.6472048133681265,
 0.5789663009641735,0.19978295473279073,0.516191349046591,0.5143000045187094,
 0.6005269667533945,0.08343598532161245,0.25468889456121946,
 0.8570409209207238,0.7139831477512978,0.7433979598414648,0.5831021332251191,
 0.8786367849027644,0.921855336916755,0.8215577530550474,0.057848375273074515,
 0.5600992381493762,0.8074093816734091,0.4020479183393797,0.9567743690301385,
 0.12388620520565863,0.6952127104471668,0.5166203202135196,0.594290379047019,
 0.22754489669386313,0.9751673709363439,0.8375390480288156,0.9626940362099119,
 0.7686094386666401,0.07041683124435705,0.040038885081833064,
 0.0712290797592876,0.5418408396473958,0.6607221445885368,0.08153675558073892,
 0.907895275610155,0.9910384550769835,0.2063134611200652,0.6889717551527871,
 0.2351258717686081,0.3513428543981377,0.5117985661281744,0.877691003705245,
 0.2945816595609624,0.30770270859435334,0.49789592595233445,
 0.023360768609880345,0.26274088151256536,0.4125292387476083,
 0.9075901159704587,0.001459131759612875,0.5207728877427505,
 0.8211181614004869,0.1942057813132203,0.5161415641039688,0.18150943832283495,
 0.9893600741325759,0.9468582027676724,0.7267487304726404,0.5081425386289328,
 0.8372270762992309,0.9285516565355592,0.6183571734313885,0.11416541828017779,
 0.8491958700924583,0.8639439303964304,0.4559810741030744,0.3826535621574969,
 0.42154144645734515,0.6376216573222191,0.5419280644577583,0.1997368456494666,
 0.4241380199961474,0.9516780410522161,0.5353189912248286,0.06424535791998764,
 0.19921362703389667,0.21409697039798492,0.1840841550616349,
 0.7402419495618826,0.2563173786767765,0.9418780837961496,0.1274375440647334,
 0.3700057749975316,0.66505688124638,0.6924426491384135,0.019113332934418747,
 0.20156442642481842,0.2863906335961196,0.42591065620314905,
 0.9146782605089006,0.17480477606117073,0.8047335034135971,0.6382286549133251,
 0.26762282337230703,0.7684739549443069,0.9156194744425475,
 0.11759334569769486,0.2833215926934394,0.9323505225733276,
 0.12543097327003383,0.20281158579332448,0.5647222897429529,
 0.23559487437669846,0.747222097185951,0.6362099707887335,0.5588690203296272,
 0.3831589032344762,0.5615580126788786,0.017939820245621463,
 0.1783455964001167,0.3002567048589755,0.951028176545001,0.5011698031846249,
 0.5948190812207418,0.20532335127868262,0.6533348925018594,0.5624453059950086,
 0.38851008862220526,0.6182471831800179,0.2085971549275838,0.9157201231450418,
 0.918451609279708,0.9883758489320551,0.4553345037317128,0.04573868569035311,
 0.6128809058854485,0.6776265980428177,0.44234239271757614,0.6963234337692432,
 0.253516148218599,0.69286978683877,0.4584914672776975,0.8495487483454836,
 0.9142013250738822,0.2094081012849105,0.8253099583445815,0.32826602272016425,
 0.3977667770234312,0.563895602230526,0.48316252858498154,0.639136856074251,
 0.7995266900335384,0.015296469104826294,0.5996174808731912,
 0.9449555071817127,0.3046646845128703,0.6065740235998366,0.2038094537372045,
 0.7298149624705034,0.6723074270161497,0.9203849497049017,0.4528396625731941,
 0.40136684690318647,0.3765438529643075,0.8322283681590936,0.3513404252453417,
 0.3705298599588599,0.8362183726451385,0.4883695655083985,0.4065952985652247,
 0.0437698795503868,0.7166556972321798,0.6657785603533373,0.12688361103575707,
 0.7827986579309086,0.06408060015630646,0.8529508020917433,0.893040546927211,
 0.7451039832497498,0.9050653971958279,0.6346440023414177,
 0.0039325589239225245,0.07108973383794859,0.12434627885745309,
 0.7073810846160262,0.5313613736242107,0.4471388479950087,0.16504727966784516,
 0.8336696605452745,0.8204862142779488,0.39746574628434395,
 0.006410544139971219,0.35670003040369513,0.1523468184748198,
 0.7228878805168762,0.772082508217361,0.6441334048113507,0.011593129019918491,
 0.12730790641286605,0.10877292316920462,0.4522609428989356,
 0.9124747851126334,0.3230196154848668,0.7417369177745565,0.20472602923146754,
 0.7283714742398352,0.7577743893821324,0.4053676540931842,0.5780261994313758,
 0.8697249018081219,0.9724241952490927,0.2736545148762055,0.7696794821065728,
 0.6040410325380249,0.3821400711426458,0.1891008515943844,0.02570890336702858,
 0.7615274867103158,0.3907743442684143,0.3925424840563697,0.6859819158571699,
 0.5665746172275856,0.6269642491696276,0.5460759060412705,0.0308397237007334,
 0.7145315044554151,0.7944159746357685,0.8460234676857159,0.4264922358575858,
 0.8406582269461701,0.9378887595615889,0.5214935958524506,0.45472969511909656,
 0.391560118163623,0.9115684582940426,0.48159153035119895,0.20783076394216304,
 0.8161144475201921,0.684221149614817,0.5139827028605753,0.08001198956602462,
 0.9089118706859334,0.1461328956607002,0.18000848373865574,0.7754157868497166,
 0.11904544774442694,0.4429980825117894,0.9688148458482017,
 0.41562498269364445,0.4962297783649341,0.88872585717727,0.9185153847372413,
 0.35345497278955773,0.4410655036853075,0.23830623252673022,
 0.5373039033297335,0.3496894326206126,0.657495544192549,0.6648038816975453,
 0.7681044058886458,0.5804408245116164,0.8000791721626317,0.8152388863367102,
 0.2081515604110753,0.5556619307751993,0.9811035869888273,0.6103418267667745,
 0.15695280802659872,0.8611703777433743,0.020830965759222497,
 0.6168798382493619,0.14340074683299409,0.1444368101923985,0.7788948392998525,
 0.6579409922009,0.9649390610660191,0.07837576912523248,0.05355424958096411,
 0.3560227990308179,0.4778249635224108,0.8447213331889046,0.9084078284337193,
 0.6040495129833535,0.040443236673507355,0.09516795598964123,
 0.9133982172779556,0.05603546805090487,0.9994211726087954,0.6939992330759563,
 0.39374068618381475,0.03316394084245733,0.6048629025817092,0.857746005945398,
 0.006148067147911718,0.354037136499728,0.6250126382523653,0.4393570723211484,
 0.8349956049598614,0.50096068926019,0.48212881063743096,0.04190416943408348,
 0.11966345073143281,0.2946010045698628,0.25884507776212384,
 0.3980929457198489,0.624443967789948,0.09147420938671003,0.30540449103167655,
 0.8624222721378529,0.3614298255857781,0.0736103504548532,0.6957891019363286,
 0.676951769642429,0.22376909061001793,0.6873201044749536,0.8860917795954774,
 0.6607873557146059,0.5174805749289288,0.849278540649291,0.6912390305876557,
 0.41503308638500336,0.022332487576515958,0.4071200676670754,
 0.21460213637536807,0.48697884364221167,0.4100863840520955,
 0.12344882093037934,0.30996849649800673,0.0022923687492844014,
 0.6754092170272126,0.22134340256669072,0.519225420428058,0.3123320319061843,
 0.12165038346726831,0.83276099749978,0.3914351124784793,0.4426172301965332,
 0.3394942709019093,0.09415020050194367,0.4549655942317802,0.4893748051818315,
 0.4870804149548307,0.8536812998087299,0.6465684724177627,0.607781745407745,
 0.9919708044915105,0.7886069201647634,0.6983854440172133,0.9473730613987845,
 0.6107250121557811,0.5830503032423937,0.6840101824664919,0.09530529805954269,
 0.34633602520502826,0.1415145282780117,0.3899604779211445,0.2459008878496477,
 0.15918927881695655,0.6800102039769238,0.09542481960922544,
 0.22975787203415865,0.17114935804456444,0.04102356400662921,
 0.7834577016546949,0.15859665197914985,0.21435295406447508,0.837350893936895,
 0.05762627948730026,0.16942795873093064,0.604004832405739,0.9958818031380484,
 0.6636730055464972,0.6020303641223937,0.6838785285711928,0.4345634711634655,
 0.7819248878218782,0.2438193317033417,0.8298936297427044,0.23202145627202053,
 0.3818555543504669,0.0777032851609154,0.4957097561821002,0.17315120029444941,
 0.1755550096946057,0.43613876025903864,0.9321216669136184,0.5626206348595559,
 0.5331718076631072,0.7867818642901174,0.07383670401104225,
 0.09344832454032503,0.6073844068181237,0.5615063087822242,
 0.10077185683753553,0.8920056093986046,0.48147359552103264,
 0.05588489422590559,0.9406596924031309,0.6535571956695259,0.7725050828875085,
 0.772622466808612,0.21579060041728038,0.595762369966294,0.9399898977431042,
 0.639892377735793,0.7857681358306927,0.9736609788849444,0.8328049553771303,
 0.09920739830086756,0.38436117995494656,0.26228524443912116,
 0.487600452362247,0.2106308830174669,0.1504913596878389,0.17297613612680107,
 0.4538051521013303,0.6512162030039832,0.7984879042801328,0.14129967660660037,
 0.8075936850032606,0.364653898606509,0.6566291625968652,0.530768204030794,
 0.9589612372872826,0.4832854913933925,0.7150429456882723,0.20543539215718065,
 0.4473779779836481,0.14231767117195793,0.1067603467183702,0.7714131585221773,
 0.955834851084196,0.09640297835557643,0.6903377080862114,0.02822578285896482,
 0.9375521588058382,0.03994184667912781,0.5935980496806745,0.9440580263942424,
 0.9813994118860829,0.2609768315483456,0.8089197677197238,0.8623732449610377,
 0.1744896899434012,0.707330261691876,0.1031183253631428,0.48922845985051466,
 0.21557861315763638,0.16331166266427044,0.8223641543776945,
 0.5629523813282276,0.8960471234674406,0.8899778722077409,0.1710230791868237,
 0.46247545862515294,0.5929458618599714,0.44263716954729126,
 0.7926655432915548,0.8200440318753262,0.5062187864558961,0.7793176628360242,
 0.33142730411679766,0.9509247455715637,0.8910105201964509,0.7716646827987907,
 0.03418765361602516,0.7151545498102476,0.3509426985641504,0.8543286638327181,
 0.6696903811001087,0.2547997750717157,0.5477488269489545,0.5236642940320679,
 0.38269328408275527,0.41094526504216666,0.4502455583086803,
 0.9712581151191012,0.05179286865649513,0.33734896562649985,
 0.5264232523685737,0.36425890901049907,0.0034279552374331113,
 0.09231381496226398,0.5849964264833301,0.9636831759697954,
 0.18121514665722482,0.09310536623902155,0.9112464290553236,
 0.8410162645783169,0.44698146289370366,0.4058620090859075,
 0.09500923102285741,0.5210064411944871,0.9818914101907413,
 0.25678849754577737,0.31642399896300044,0.30643055071587,0.8145874351210048,
 0.5475913156039255,0.9697245914584495,0.2854758206265253,0.06940652684336157,
 0.8344431465097536,0.24611352879896742,0.3448060560277014,0.3244747798104549,
 0.2047824534497822,0.3665173482626193,0.6342207614444942,0.40006912989957055,
 0.9684900545821785,0.3368867436864703,0.7650333966253829,0.6835625080221652,
 0.8516995798240599,0.9932480752271015,0.9913979659725595,0.5824574302538799,
 0.5051228925229015,0.9668594438200989,0.35258721796084413,
 0.27924556888734975,0.8044411623921198,0.7181752183168768,0.5481583012021033,
 0.05687659298686365,0.3010508944834218,0.7300779398890419,0.3442274443686424,
 0.305115007454557,0.5038898674129944,0.5007255590011457,0.5029026495615803,
 0.8639161003203836,0.0075737331284084775,0.781879249216732,
 0.6623327457570825,0.9130256216517886,0.3012356633618032,0.03525285466059336,
 0.6933017869867246,0.25098335722276865,0.04809912862460164,
 0.5806207277016777,0.719710627842727,0.18520099811765234,0.11744141129416663,
 0.6248865600206558,0.16289684646909552,0.4092757995205042,0.4502534253963668,
 0.24074489178762049,0.7714546751208337,0.9413550171804654,
 0.01460482479295988,0.5555713224931774,0.582333222365503,0.6798233728907936,
 0.39763107912604,0.6015616931704704,0.36410166818855894,0.01858169190643366,
 0.6447502791769008,0.0526926683611868,0.24542227683031903,0.312318993704666,
 0.7438055658947901,0.8148469826964145,0.9492683286959438,0.9242537435138812,
 0.7225699190342814,0.9746811219131395,0.5440766497342473,0.6667399037374703,
 0.9649291106025442,0.10887173569158859,0.6294961818935947,0.3153101498862745,
 0.22465481520269281,0.4345857936234726,0.03738445489975595,
 0.43185104095578375,0.14065885756705676,0.4877475692627148,
 0.056038204887548515,0.624910926300579,0.6467388558475845,
 0.08224490297821807,0.5698279939480029,0.3107026434346978,
 0.37295531494157597,0.8670392443971392,0.8073538381426749,
 0.09526339688705632,0.6277291684923422,0.23049658206165946,
 0.1352221646180669,0.5253054358876205,0.6221915096104366,0.6284712328627273,
 0.448507934831218,0.6887000973317785,0.9929397730716565,0.4626874275189159,
 0.6761061463848649,0.058149976517911095,0.46974468115021084,
 0.3918880123052131,0.7787758154778309,0.931250919764436,0.6244535738716067,
 0.07337286421760958,0.6931022998467414,0.6672873694288102,
 0.34213855073525856,0.04183823625870786,0.04983636939933245,
 0.7872536701295623,0.43808359585949297,0.31976352140865927,
 0.5202099957167874,0.2630530309295709,0.5694332484625718,0.6917777923437802,
 0.7398986267402035,0.4413500126176966,0.49001212433237484,
 0.012439457292631628,0.35728073950063133,0.5599550638038736,
 0.1800249169456951,0.6580126163759895,0.4708427001151211,0.08561235340822515,
 0.06278104411222818,0.10323874448091219,0.17269679249323355,
 0.6573650858201192,0.7161793723759675,0.516165634276446,0.7616559523136474,
 0.6024788857216012,0.4826506088220984,0.7143998395526691,0.5057170794188681,
 0.6071924092959284,0.17513693604561098,0.9522433234708058,0.9310859057975053,
 0.4484252848739687,0.8333522512002873,0.04587492210601041,
 0.41114210619893854,0.719341794520471,0.41580138248787835,0.8556265448043373,
 0.3061265122239243,0.9330538034054331,0.8358258859542471,0.4095163671039006,
 0.32310646342407623,0.9478072213380526,0.844646750316174,0.8127479031581756,
 0.6288227146910392,0.4623845662665266,0.3208727078806486,0.3159348552093242,
 0.35278199844683655,0.4340245608208394,0.6636052857832012,0.3060406599756268,
 0.328935964691607,0.19716879166264456,0.42677083677552075,0.3878642623253227,
 0.9534483200819239,0.8707154079079258,0.33589363208436307,0.99107434788888,
 0.5637971315222827,0.020398965103731515,0.516384330119158,
 0.27187965168943273,0.1644466882427077,0.34660128997623485,
 0.10887552687457325,0.33601147008029364,0.3402486384536758,0.446383459895059,
 0.420543165438498,0.7759793595943572,0.5812089692601,0.2467327394657408,
 0.16892249582144925,0.8393594182119172,0.2194658819709554,0.4512983325777222,
 0.9257928619263112,0.9453996610856485,0.8602612967332782,0.89532879208172,
 0.7419976589924931,0.9069370652790271,0.6506565178038174,0.20083893277296094,
 0.10706644099362916,0.5852379383135086,0.35948149804182816,
 0.31846429164527745,0.9085367781405965,0.059977003588657984,
 0.6942188777377176,0.46877079820778644,0.5117336999832557,0.9067410080162406,
 0.4511434157171794,0.5262682858072607,0.9032920325019679,0.7243978338827148,
 0.6713386997871236,0.8389234596195078,0.5494260171004889,0.8762534559853172,
 0.15971144936797332,0.6549959243570985,0.3285401321992688,0.6247626728462202,
 0.06043559486179784,0.011928197527137807,0.24955399946231138,
 0.36744565431006493,0.6441612021749907,0.3902357560694043,
 0.014258934079065422,0.46998579642655325,0.893256304060885,
 0.5425409896149908,0.1383951261379628,0.15860422558594345,0.2803030840755244,
 0.6276870170299736,0.8365727374284618,0.8892574747531772,0.5757775276498284,
 0.1287089080274948,0.24921672766118508,0.3022568682390584,0.8105658981554428,
 0.40658970268794015,0.28791133903708366,0.8859630344205203,
 0.5158404400136595,0.8898469707658581,0.44606615204743494,0.7536929939535071,
 0.7883015075518949,0.9672120749306551,0.5434719629934308,0.7395967395425869,
 0.6647585901982582,0.47645549235962115,0.8866911455062143,0.678422936938006,
 0.17679340526169263,0.4976108998381401,0.38893087544134497,
 0.8813762104805436,0.9870740378949509,0.9691940766503625,0.7433084395244594,
 0.812309917787067,0.9869828702825852,0.7184526087996391,0.6170695043265332,
 0.7859913049902776,0.30381154702224566,0.49278215819679205,
 0.4018144939982733,0.5272714462554103,0.7672065637212242,0.5271633074161615,
 0.7220651655425265,0.7548397730054618,0.5739846293113627,0.9523713591941,
 0.5986770585025056,0.8730195493620424,0.017981911476712464,
 0.5015997135263831,0.07847010353980777,0.27874635964827355,
 0.5179283133788865,0.6302998540935151,0.05196078822244787,
 0.26091312489171536,0.9946127419792041,0.3861225940280095,0.5555426737002757,
 0.7260120905376805,0.09615915713051681,0.9609611115520698,0.6326792084772344,
 0.8523379788252121,0.06754947138891143,0.5737256323397597,0.6531825532915961,
 0.5306642278177568,0.1026606508957777,0.5142772896632523,0.8778295045198975,
 0.633596500985932,0.8229435978437236,0.548265302769133,0.18400700981365437,
 0.9407715537642871,0.03671966610998767,0.6372024273631115,0.9555634746843403,
 0.6775814216097337,0.14191695935576643,0.504799791774123,0.5955173885795574,
 0.12024416645160951,0.33696442553377626,0.43338547581049147,
 0.07149724389963752,0.07733387094599975,0.029525221687515035,
 0.0702056333729959,0.30770706883457843,0.8130752227105218,0.4031823956123681,
 0.7182481915323491,0.09657885078797968,0.598252821871837,0.11004000178220841,
 0.00555322143433179,0.16026435595579203,0.20763204493638165,
 0.2037527366376788,0.8323509445224468,0.46312015105236914,0.9258902581140916,
 0.664112939449776,0.8986302643857371,0.15239536742513016,0.1201776493229435,
 0.5049369884742053,0.6778306533913667,0.8108358123807899,0.830266785501008,
 0.8967458760061614,0.5890286627840478,0.48152006054011376,0.7757621986096982,
 0.45207440738033333,0.010278174219745528,0.11112904433328363,
 0.8931240377801696,0.6390607252433581,0.9584062482140745,0.04276604491538605,
 0.6378264945143672,0.20538502572234185,0.7348013641838478,0.6538810430641269,
 0.8711128535133509,0.5551259895044366,0.5859162323060921,0.438417340717117,
 0.5995344177034929,0.45788032738475415,0.9182677964497971,0.3890247467761304,
 0.5866668562156326,0.6558129985470399,0.5696792994858747,0.3222750946755589,
 0.5439634678696753,0.5706929881459385,0.5487840212291296,0.2519397193694939,
 0.39375625433860006,0.09774456799974007,0.1340938590630939,
 0.7361282507167625,0.5765811602455035,0.268749720212231,0.5226311384030793,
 0.31099333592267453,0.6429780857308084,0.12883196982684697,
 0.050171088485239435,0.5700551951549875,0.33134764016376317,
 0.2426264110523807,0.08602215617133202,0.7792564631054709,0.7681073999462293,
 0.8253639977932333,0.1392764839133247,0.8753644629305664,
 0.050690141360023144,0.38824273432290324,0.497501384804347,
 0.27913347891741214,0.6722046687043608,0.03649770988064105,
 0.45989268664966565,0.4922723247461154,0.5140283106639351,0.7675796377659275,
 0.4875537075357157,0.6469045476792505,0.03326824925265204,
 0.08009958272571349,0.8936399799021036,0.7572012912422678,0.4782409544145595,
 0.5051422083804608,0.24475399558423072,0.3784259020753915,
 0.08551584113700439,0.15249950050600825,0.5402883333060627,
 0.9317282664256855,0.13399781033640745,0.9183724635162753,0.8928604768874303,
 0.4432500078174215,0.5237214454280299,0.9693990058002109,0.15493428529650732,
 0.026742705873446737,0.3290259005180274,0.39824723857130284,
 0.9402933706004671,0.2495817002664077,0.03700190174461948,0.3708768628164393,
 0.4099718884960053,0.8296234700430287,0.5151190820374275,0.9336300209496402,
 0.16848845746558394,0.3117114465896561,0.8530034523762093,0.2335250372865897,
 0.8218621752101052,0.618280868884042,0.3732210564890952,0.8755173132316628,
 0.1450168051101528,0.9503123442811334,0.733125751698096,0.9172989966258708,
 0.5815436507678388,0.6911633807105008,0.4488715211796501,0.4386349676416612,
 0.5429709295492421,0.22667802975640916,0.15905437724877136,
 0.9483099834178741,0.3719680161665614,0.516353528499962,0.15367058241070497,
 0.6598813394540226,0.9142992098244206,0.7107109506429986,0.3775233380835964,
 0.7625876796848882,0.22932303292917233,0.35232274165633803,
 0.39867028303243157,0.9257206994637717,0.47793537242247974,
 0.3752512237693012,0.8807978576195106,0.9710982481424079,0.7338246341166519,
 0.010369368402095347,0.585659814396323,0.1664128802321233,0.9408481967130399,
 0.16348545164671135,0.4821232580323067,0.4499648094399461,
 0.48230624559752866,0.078319842734216,0.1004395948854434,0.5048696940382227,
 0.8515551300543347,0.7078332631522305,0.9950195026317435,0.9113274047108566,
 0.6767999397395064,0.5976364559086927,0.30022725010474693,
 0.35045590263238713,0.6197265708221722,0.2291178254617856,0.9698136996233236,
 0.9943994409810147,0.8896333910114472,0.30556914365039767,0.6506740915628878,
 0.006882095833089408,0.6125628947011308,0.7538634498515691,
 0.23898513902394425,0.6986288783436831,0.07437912265007585,
 0.09436994718668978,0.9800722184532379,0.8498628395759512,0.6445895140637783,
 0.7518067393506283,0.7116402865788566,0.6513366257835681,0.40782941898370517,
 0.8076004899018857,0.2972777395928252,0.7856177402113559,0.1107713701382107,
 0.9449954573037075,0.3287073188621822,0.2915427448408864,0.9643573515115345,
 0.7663820467869342,0.5407045432292968,0.7747208356222536,0.19018198019282673,
 0.3192964394249257,0.9226463003503333,0.08297748100924118,0.5893954484202778,
 0.6452689169618273,0.24332665702368417,0.7469769992338245,0.9022710583922857,
 0.5129707806340803,0.12038861878564822,0.44377259175817074,
 0.7647240212642831,0.6487458598133684,0.2624870907519318,0.4935386040915918,
 0.13867591060585038,0.3067084277171881,0.3157204211944531,0.8675054015196866,
 0.7117886370017011,0.7319463932163509,0.7223664091545703,0.41816543882931134,
 0.09027104926275242,0.9261108677870095,0.30645129773632385,
 0.05427142337021229,0.7073888137660269,0.6116504616970098,0.2753135868185076,
 0.9476556154227441,0.8658942972543856,0.4849660172787792,0.5209183739901493,
 0.9610539973835268,0.5497129985652015,0.9334065673986086,0.5637016248279038,
 0.07358734311182136,0.6781258060165102,0.2405706390217165,
 0.26699911542402044,0.10288452069027254,0.18506335509228078,
 0.5580528539005869,0.09408975130124442,0.3693054176636965,0.164622827999507,
 0.032989343513057046,0.33076853321931987,0.5320476087655494,
 0.1106565818753672,0.1609836322625332,0.46219109003002723,
 0.49491787948015054,0.8812830276222869,0.6504123448295961,0.2157191775875047,
 0.443220441305462,0.05099416634852738,0.4812950129687604,0.37884722139785876,
 0.2130067622693984,0.06831106731947156,0.7871029789724076,0.527936269506074,
 0.2864325340536187,0.2366214304174813,0.10362492318738425,0.5076362433332684,
 0.792845619091861,0.5348267564117305,0.15153962411172772,0.4275719375832173,
 0.2011067534353781,0.2458785996632088,0.5434770020980539,0.6037845970588022,
 0.799922330341884,0.08487400272084222,0.34028122414200723,0.6582061795212328,
 0.755710616700428,0.1645622891963643,0.8754202179794813,0.040572924202029625,
 0.7861964703402622,0.8455941503881956,0.9383264316505088,0.45499824281171464,
 0.41123903306611076,0.8634679118630129,0.14599432883761576,
 0.37628292776069816,0.25923294533144414,0.1406522454362793,0.525879655797922,
 0.8611023231337734,0.2037824638993172,0.23498891807249123,0.7989747454827234,
 0.6317748034014513,0.21703861763861476,0.5681998469534469,0.6161857407141271,
 0.47907907170392405,0.31645023390427285,0.8469958905272562,
 0.7707324708119867,0.6864575897715309,0.1529285449373997,0.4319138173607143,
 0.09324451558626912,0.6648095521929696,0.750897454562022,0.7802944179985124,
 0.5993636494113482,0.5404785368517826,0.7244720828384619,0.8197473859257849,
 0.8497577871501933,0.17034582510840912,0.9015390107888431,
 0.018052735211701365,0.5933286166375735,0.8269244047051749,
 0.15301703297152247,0.6436020032602597,0.2655464406730068,0.6080814329820998,
 0.8985860888802863,0.710311021602461,0.564472127831552,0.7086572908010172,
 0.8772099352009127,0.7232402307404151,0.5491606000306966,0.4496695071669967,
 0.29957512451598156,0.13230700227872283,0.609448670825208,0.8433961657951511,
 0.7774140810430875,0.8434175263035896,0.9021258225096329,0.8833521484465535,
 0.520002866176277,0.25880011063873376,0.4271160733392374,0.063430865613607,
 0.21749966908499563,0.5482620930703794,0.042917268097458106,
 0.9221191340075066,0.3598520383441669,0.7179725903498164,0.6047182148396659,
 0.6202411107041484,0.8095058409395872,0.0667434827528075,0.06366485884985762,
 0.4218478938344187,0.22680113822909742,0.33222573207441075,
 0.18768644385622468,0.7808145614669345,0.15566421754171555,
 0.04448867739402451,0.4762177478243894,0.39056441466375613,
 0.44458263887147087,0.6430503803647876,0.12343761179174584,
 0.3225531330130955,0.016893052413028742,0.10431387490105937,
 0.3609606384847407,0.31142554217573726,0.2727072586540046,0.8131422815306648,
 0.988874252037583,0.05293927749855232,0.3266873466086022,0.11049916114280944,
 0.19516188499320009,0.1610918391207612,0.1542318636896115,0.2316462050783551,
 0.08610485593264228,0.29119050151069326,0.08649117217270708,
 0.02761772774665905,0.23932456100298594,0.29447212322537064,
 0.3308967691425575,0.868154081492174,0.453391623092537,0.29407131845581536,
 0.617573154896958,0.6459644159357129,0.5449547628611139,0.26655490283878935,
 0.61844260201859,0.849115184020502,0.7705659246020354,0.6600667577537417,
 0.5066618879179606,0.6567950746038534,0.18725201145434656,0.5794718666219724,
 0.762894078459524,0.008533654326639106,0.6366807324200069,0.5194146527202005,
 0.5572390443636633,0.750127455074607,0.5676332027740718,0.1589332863414249,
 0.7670491107298867,0.8752659746220567,0.20080595765312503,0.7532183094955657,
 0.943626150429533,0.8825043070346533,0.29487870536906535,0.08022648357497375,
 0.21682120747919087,0.9627301139738318,0.7039553867141499,0.2746561937845977,
 0.2796107851456986,0.37048811217327815,0.4416727996529385,0.4575125740100341,
 0.5508800992235934,0.29200422607417964,0.11145260336183094,
 0.3350743539910336,0.9119762512839691,0.40673901140162205,0.6571565289274435,
 7.621481864427126e-4,0.6892879900172395,0.4376024879069903,
 0.4671975590931132,0.4235336254035308,0.004670854305020122,
 0.09625178556015312,0.7199257125629266,0.9594025380714702,
 0.21346459958047526,0.7697822369697749,0.2792742674695171,
 0.38497635044422074,0.32039983658896376,0.21039802228775162,
 0.517059726272563,0.6279270394350702,0.386092934482021,0.12415546993184978,
 0.33134183434030173,0.4396402736729067,0.5933283024252359,
 0.39412559266640323,0.8183436981667153,0.750753632480845,0.6811432846566177,
 0.02971758889430065,0.9028652186285254,0.14873697275236175,
 0.08654271007432768,0.2531948725571155,0.9403762813199544,
 0.04249409201728693,0.13620660937361606,0.8320584186788087,
 0.5881360921897434,0.20520677179548952,0.26904469842285217,
 0.8675882484142791,0.8322432282539585,0.8671218766378376,0.25046421712015754,
 0.4101581135360728,0.3583266253397861,0.2635635458845017,0.2671845677519985,
 0.15827002311771576,0.40162254586537793,0.5977568073917752,
 0.5235237345765659,0.700462108899403,0.8713318363151468,0.8237330288460278,
 0.805441118768772,0.16781510663544208,0.9401480235575418,0.36198560030145277,
 0.9593994723086909,0.755732537783957,0.9062711012965906,0.6908594384316925,
 0.5307867975473193,0.9583677812150574,0.6667152326997569,0.7085234795805637,
 0.6090895015060203,0.5900262319328812,0.10634774552726711,0.0819171289427203,
 0.7053449178143385,0.4773237368726102,0.9731514617418459,0.23793301328997696,
 0.35931679043748177,0.2885573899434838,0.2950078418020452,0.2102733767032673,
 0.3040648650058486,0.6543881170753726,0.3294939544181579,0.1617781031316009,
 0.8494872412980806,0.8218213617946439,0.5786272512311201,0.6244818703942376,
 0.45499565061241254,0.6773427496781285,0.6500511935217033,
 0.04151806170535299,0.8007647913291286,0.32704771739983696,
 0.1484295728575683,0.47926794671705264,0.9280098097151382,0.1548554130613039,
 0.6710769784451921,0.7530513540008194,0.9462794586934432,0.5989527954440902,
 0.49057744951291793,0.8569890365900069,0.8248045827691135,0.7659322178248843,
 0.3901046421068187,0.3338867790922708,0.033257189595350733,
 0.09475260419621079,0.5353351257595201,0.1216964434811143,
 0.10128527269879661,0.3515463696122988,0.8064066450915665,0.8362361009526154,
 0.6099895664529252,0.24074561726548505,0.6529301689149891,0.9555516738016263,
 0.9632451223228123,0.6651923714106444,0.8343124195486663,0.14658721023233057,
 0.3104569275627852,0.24504545461709726,0.6657901264113322,0.7581878447716047,
 0.5647374638773719,0.3178133874418969,0.4143020823762529,0.8151543817767765,
 0.38510922180827256,0.44153279340566765,0.2862625041003347,
 0.12078300299330647,0.26775421883194583,0.6823188895697384,
 0.7290861146906263,0.24749240856202215,0.8778241997296478,0.922347266532784,
 0.8477312217248624,0.26298149065674037,0.13969845892987065,
 0.49752843524810864,0.6678465405180419,0.6126822055250855,0.9466049950986758,
 0.9889932275874602,0.07636154047521493,0.2788214616367638,0.4167466655179797,
 0.09891166497852688,0.6210164412190331,0.2582683182300255,0.562177297670129,
 0.9975718796710535,0.17130645594505656,0.09552875922342219,
 0.4461374732123258,0.08060883753462855,0.1424020184279009,0.2570424496653816,
 0.5831654128051775,0.7081940243443041,0.19991721029185627,0.3020795174723041,
 0.8985298256125762,0.022520771528703687,0.5440940969256277,
 0.6447860912107259,0.35936511178559827,0.8439873967765912,0.4151187693909941,
 0.8930174579792562,0.4840358833632882,0.14038951380862508,
 0.23358953346714617,0.39575540459345815,0.48021427343428347,
 0.9704646647387266,0.7459244398334846,0.880786146433257,0.2823172287935807,
 0.9319687948362794,0.9518398625160018,0.20690367925183395,0.1009116127581472,
 0.37777305788130877,0.600179493654234,0.711093033089741,0.1675819967644241,
 0.3623921371162976,0.5998730960014612,0.5703773855254131,0.8752514656941776,
 0.12520755919720283,0.2974168559417282,0.5080420940997851,
 0.11589951083057703,0.8187001500126972,0.898476068762297,0.6812597078981395,
 0.2904636364213572,0.8001045125678523,0.23867273002602907,0.9282455577930226,
 0.34777594509016163,0.6292726861718912,0.003915194577883474,
 0.081194612554785,0.007027520392252851,0.6851373728268659,0.7581823832937267,
 0.07059675556998268,0.8637359574098853,0.9956875026173891,0.3674499067686037,
 0.43289787867216567,0.21411872887546868,0.13165241913483527,
 0.6620111473952646,0.46577007123302083,0.12529748490404202,
 0.8867276536678623,0.8178036376171035,0.22265084536121915,
 0.06202179490948578,0.9470348315027799,0.24040039677207803,
 0.8481468914039287,0.254926933287294,0.01585561415996528,0.5930543199006075,
 0.4401374322106534,0.6390401346897076,0.8285818467715431,0.25007857509309384,
 0.29072373081223857,0.25204489305237665,0.36061842664250177,
 0.1467784614805513,0.67385397491819,0.2360537173748486,0.30856120975577017,
 0.09699588170255591,0.04315977834702589,0.8564454509259004,
 0.14741934250548383,0.4022804593547218,0.6786949228231267,0.4901359804071379,
 0.14392465361212747,0.4390266488025446,0.7201174476442173,0.6805957670059601,
 0.15484154819283846,0.966151746509063,0.433740650013292,0.3544952980372741,
 0.22778371633557026,0.22223610883806455,0.5599175332020973,
 0.7495773814876624,0.1653292979684422,0.673674227271868,0.11799706668831345,
 0.1358792790674178,0.2010311244765377,0.733129308520464,0.71309969623527,
 0.09709204424106122,0.5504352050311974,0.22788520594474926,
 0.04873735093147902,0.6333506090734845,0.5141792045423459,0.2860998930601004,
 0.07456603731605171,0.24303128459753787,0.8706751201848649,
 0.6901334624778365,0.5191516244474359,0.18569239590660813,0.5717555408292683,
 0.9384818806372738,0.8821360165885784,0.0015843406374206337,
 0.7638410158043395,0.04454767259221215,0.34103120399460585,
 0.03061370430492616,0.3609949711464391,0.6257156441781973,0.3484536024689384,
 0.5515993527932532,0.6857841155468757,0.26939758959052895,0.8870710068769901,
 0.42138969709871266,0.5425554460885822,0.393100208261724,0.6442824640134388,
 0.33176633507408193,5.855996676732422e-4,0.8408201172368479,
 0.30582581741269266,0.19340815523746968,0.1619140449308104,
 0.13569908425809307,0.07339192290650931,0.06896043026925092,
 0.32882687879652195,0.3755027401363923,0.9070871418340392,0.9773382961047581,
 0.8613650811629823,0.9852926662496582,0.46237523931601565,0.712538792786039,
 0.4778002894656874,0.6142745538136423,0.8058604161228969,0.03286256547126054,
 0.19255973648395974,0.1802972360422893,0.6231857770564245,0.2876946849147598,
 0.2201223881794454,0.8642016808559867,0.12501882072935988,0.6792095329172991,
 0.9034618361782413,0.8645227260496633,0.31648728792532477,0.9192530508713617,
 0.9604049025675346,0.05827955496438564,0.3155293099823493,
 0.13457968234967121,0.4277539771715727,0.4597801023149177,0.7161238482161737,
 0.13352513824145174,0.38016118360940443,0.6641623805890359,
 0.15789544557466662,0.700846905188271,0.624853562823134,0.13790831487590283,
 0.8201135375507408,0.8413213870241649,0.0787079196933137,0.21573121693129227,
 0.36800176847233257,0.6237180054459939,0.18834940335359773,
 0.0259291284339922,0.30782366831248775,0.5370404555513539,0.8774910693549013,
 0.12629859301758917,0.15598512299069756,0.034056736611552174,
 0.3582590351960915,0.5676074160906432,0.44582789752316576,0.9262158218132173,
 0.789771315956424,0.6937829328317673,0.9581257736663201,0.8062350800798601,
 0.003272802832977151,0.554722392449904,0.4658822384137833,0.8495817474779715,
 0.28740758873042693,0.261975994096627,0.9967435014731806,0.06544409174485588,
 0.9597199012005456,0.8950799565880857,0.046771933583853675,
 0.13803892840506682,0.815130199397772,0.43553112496400403,
 0.12844394759548927,0.3709903594602222,0.43068029619666404,
 0.4081272947980594,0.26281416685118675,0.5407945298871001,0.6753658128391069,
 0.14863221159526452,0.5110094738731612,0.23819710236710157,
 0.46337521588372865,0.7585128691154117,0.14487528250629622,
 0.5484819420075375,0.46696259107232885,0.7552015607265246,
 0.031036358788522844,0.7632805197632048,0.22733295091470995,
 0.6448966346184938,0.45002462368026763,0.44149214077813737,
 0.14237809465493934,0.4986912921057486,0.6603130742763603,0.5321426216201037,
 0.1463121159353904,0.8125773163725077,0.9967951018556032,0.12260257095011562,
 0.5809405102100358,0.832788573179634,0.5641567006209481,0.10922873947361578,
 0.1359520264177777,0.5735727948086164,0.4497189201582934,0.35145946700618746,
 0.23419515811594627,0.7349465939291896,0.16398580952153197,
 0.6195862489220639,0.9936277444140198,0.06710367167621711,
 0.06270336431786883,0.6342548136932837,0.06580628660923393,0.995241550855777,
 0.6076250074089847,0.45604796907913936,0.8803002555780012,
 0.20122729945944906,0.34890148172149194,0.2762204904144051,
 0.7832348293585474,0.9633388739736864,0.6116733448896652,0.7961978618854225,
 0.6468977417063607,0.7921949699717622,0.11425229278213211,0.1412208374427799,
 0.9505457398731106,0.5446810153071896,0.16374173819521176,0.2701816289407668,
 0.37657011751874336,0.8490834749265815,0.9452806707154415,
 0.07611061031638133,0.9196554968785853,0.33496113642273173,
 0.8893380050314591,0.27841465805720367,0.5279935568500675,
 0.33770800363232967,0.5212481881136308,0.778561130368068,0.1215308282782821,
 0.20796411364191048,0.6607000705911594,0.30454177220038137,
 0.3565468446687786,0.9707904012210304,0.5463774854056763,0.8559999145476104,
 0.40012338558341576,0.8504997420144158,0.17656937277311757,
 0.36248318243740374,0.31437599740035205,0.034702826839891565,
 0.5146839061293619,0.7297980526000083,0.24845926119992467,0.9244798567994504,
 0.44496382866106454,0.6461600852120322,0.7802051467917648,
 0.18944650845160316,0.9389437597164952,0.8406902070791911,0.7224496986863802,
 0.8442017742262801,0.6724782686357486,0.3694542064979235,0.9263987729373433,
 0.9684269983218919,0.03385329187908448,0.0169787922218666,0.2271213355875078,
 0.5814505070911938,0.07254790935677302,0.08698361921972986,
 0.44246924030012325,0.7840796030883107,0.011169738385001904,
 0.801287584401431,0.9350600897420092,0.6974155974954092,0.7235642686832836,
 0.9412041307930616,0.9297508842663051,0.7746659157922783,0.5306200734333384,
 0.5933054316046074,0.4354362184579884,0.5782015681852876,0.5043940474036113,
 0.4428450639209838,0.6700895462727644,0.7634178904920925,0.5242710252747294,
 0.9555091513826617,0.16507720193105102,0.44622902753335847,
 0.055380994182123944,0.16401241418501455,0.7478217889999764,
 0.05400463563789104,0.4576822540833283,0.40880564081761317,0.555882601995193,
 0.614263070334921,0.2267856165671407,0.4035924470522715,0.6498289078269802,
 0.8462356508812604,0.9433576781523736,0.8606564157548279,0.734556416150616,
 0.7852109363837552,0.030015529622803405,0.8654361635728418,
 0.38275317691714283,0.9810715693697286,0.5444597794717809,
 0.41063036244164874,0.6353875728960067,0.0018904512175029753,
 0.24002065330878597,0.24912070454525037,0.5196663838068156,
 0.8096598371843057,0.7800601769434562,0.357245388090473,0.4269886095158709,
 0.5098075858428519,0.8171662920697097,0.6405694252724415,0.3033523175346047,
 0.2959101769766458,0.1436947241848745,0.5114730188931504,0.6601097931638311,
 0.8739171215548135,0.32163085600053665,0.19142611920770225,
 0.16208114547323138,0.21825268415863064,0.8746689458904879,
 0.13780957395563986,0.5235567999070785,0.3720111750135564,
 0.045017612893919146,0.4208523762592774,0.4314666336432238,
 0.34331620683280684,0.549374867912818,0.14444077876466843,
 0.26275407778044046,0.45644636994130794,0.20451372552880986,
 0.4761166502320764,0.3180655399775062,0.5721295670425743,0.779542032382019,
 0.6593632008835173,0.7152494535075085,0.0038822964727641684,
 0.4465192762895991,0.9489073200661529,0.42586009523662516,0.8915578220930009,
 0.9306862267617454,0.693812757436665,0.9555021177152113,0.7268967928558773,
 0.8646501785267208,0.137504426011259,2.839212418221937e-4,0.7566334645180275,
 0.07005288779740138,0.6722648492975076,0.5774997900589482,0.8320000457846848,
 0.6537698893361293,0.4363142902197926,0.2501072405242253,0.6287623195200357,
 0.02656932611432694,0.42669691730507886,0.7019595552582711,0.743080304379875,
 0.4520842505080468,0.05809769434291101,0.3341857309310111,
 0.22664405399857268,0.9373257424435135,0.8656351255780146,0.6662088013850745,
 0.08006184044103248,0.8065925081753597,0.5332328456312263,
 0.057755861768106764,0.040387258092501876,0.2261902178129438,
 0.01735728923495372,0.22809104969352767,0.5529846863555108,
 0.35928855054767705,0.04876911664783268,0.47069767210277147,
 0.019248063759969902,0.36750973913449814,0.9001786009090889,
 0.05390447394909925,0.5392635033547025,0.1950618774160746,0.7454490234733654,
 0.9606720866235043,0.7599142972241001,0.5000060400211601,0.48539329552125965,
 0.8631664165755738,0.5967292625789702,0.34850188631841916,0.7547121192982702,
 0.823116029196444,0.9003373347321337,0.44459627999171103,0.8550322963200035,
 0.1255739063381317,0.33247392484055616,0.36737549534830005,
 0.04243196721761999,0.12277007700471942,0.10792887930510597,
 0.9697082140580435,0.11940126061340794,0.9647457999671466,
 0.02759399618873204,0.465699993506297,0.9459519329061935,0.5501794632915001,
 0.5674837425440942,0.655150059937544,0.49723574806875936,0.20407257621164276,
 0.0019569179593905695,0.46572515717407825,0.4187332097178146,
 0.5878637411118821,0.2168031120013194,0.49854061089066937,0.9162933742808422,
 0.8240420354353661,0.8620436619128882,0.5319873476029708,0.24198631829505202,
 0.9129685855118935,0.28347663059880135,0.9465450736897764,
 0.13105187452233125,0.915839481553693,0.5926453602943352,0.23766282387253035,
 0.35965410511344476,0.7041873127587561,0.11498538727437435,0.821057772495992,
 0.9639284384481125,0.01829650467141053,0.886225984873367,0.24826976859699013,
 0.476202040272776,0.6676024333687738,0.8541690726139332,0.8619748299835654,
 0.6074625139980423,0.9091560112515794,0.5226225814472114,0.5313720965535336,
 0.7884662002802173,0.7562746463188366,0.3675934340422181,0.8183934864138537,
 0.514827904393054,0.5037172365078088,0.7929293878041249,0.2875256203876084,
 0.47176135849662115,0.39857190640128626,0.1212962401939982,0.741480228608819,
 0.7252372945313048,0.5927001472870624,0.012091009245626338,
 0.48209243139294455,0.7699631259176039,0.24553912210340445,
 0.6115214302481279,0.09210468222404722,0.3055806431551372,0.2605533387937641,
 0.4102324672855743,0.8701003021254337,0.8462361245046218,0.6042935421459702,
 0.08230245695881755,0.2055201275585914,0.33938598204196735,
 0.2513583736113798,0.8929110814896615,0.1947007358716295,0.621655337224277,
 0.0667403507633203,0.792774469684907,0.15850564725337957,0.2146439372973622,
 0.43465986268888024,0.04944192060705754,0.639532763666249,0.7318721287457622,
 0.011644166617092888,0.9688972443459932,0.9395646189512679,
 0.8097128236779189,0.2716346857316163,0.3597037684932012,0.15988505379881623,
 0.1592479659900885,0.6586258419822689,0.12750030660812584,0.6845109356928543,
 0.16387280226955814,0.4410483286939645,0.46650462530801695,
 0.8287380287510409,0.37854535639664677,0.6410378040576853,
 0.35367416341217905,0.42183464360803113,0.9979783912019659,
 0.9390586292176645,0.37291034805964085,0.027007134799315713,
 0.6247401783287932,0.6296315111288856,0.09208985209024467,
 0.45311757164318534,0.6706989815812797,0.9162662492725805,0.876580029093437,
 0.6362267545634277,0.5010867166835496,0.9524320888707501,0.7094840748386038,
 0.8276929781574939,0.01963147414854216,0.001479881133850447,
 0.8558407453787338,0.8396355988293258,0.7829344167113425,0.40712494565709356,
 0.861735967518987,0.8177895817411162,0.26396356994540016,0.8780801375365213,
 0.05603614619834629,0.3515418920927045,0.4593811123606064,0.550970520889836,
 0.17191427027493,0.7340876922534199,0.7152894656995411,0.9351320303025191,
 0.3172476936467763,0.07041837901180703,0.698457734658595,0.07885616590811895,
 0.8884192638453954,0.36799004962347504,0.09927005490657481,
 0.04628860753559261,0.4814212081267857,0.47198589072554165,
 0.6967029978935354,0.7841266101626077,0.816816653732771,0.24065431729941134,
 0.37920260301659503,0.020679785477058288,0.3884559396849381,
 0.05624476004763579,0.564623577606803,0.6651837843024515,0.18402550286880393,
 0.5583225361551518,0.19229509345392204,0.42956122555870857,
 0.9502913042387295,0.7163017221161931,0.26894773538173755,0.8327689894336078,
 0.12362490250019187,0.9805308764992011,0.6056558452764609,0.8575812442581189,
 0.3811627180982273,0.9587215122410667,0.5956270419357541,0.7029387998311276,
 0.434782757993212,0.9623491443381668,0.9077383571755946,0.7293826232629445,
 0.8892002647969252,0.063547911947055,0.45798088195362296,0.5512252293615212,
 0.17138137658031116,0.9871409996997389,0.4578785716973508,
 7.230515893711376e-4,0.8706016007009785,0.37637059367331527,
 0.7148828114012337,0.5658870869756871,0.36361228176589366,0.90296938681312,
 0.827654912451209,0.10294815581021732,0.04644571807563502,0.5655937201983942,
 0.6432881075578339,0.4180075015051209,0.46121603252824395,0.357252927344594,
 0.3232885088923191,0.6304576729712581,0.9218913277324123,0.3686230307391327,
 0.3353196933455639,0.18845266622682222,0.12380264490379778,
 0.1548667672043158,0.4836816317984324,0.934257931700161,0.4375399789851271,
 0.10831144227850442,0.2090925816013698,0.27637000212905516,
 0.3882913333457312,0.0882046418634459,0.6527978864262538,0.7053625661910901,
 0.410062476841624,0.025552785140050882,0.9597166436221692,0.811783691745236,
 0.7528624721712495,0.7403518842522575,0.10888239563027657,0.7014792756639703,
 0.3490289091786244,0.1405301069245971,0.5682290239666298,0.38162790896829035,
 0.10899855717400098,0.7349113619268417,0.6926242787669299,0.3546830381520091,
 0.9746349667424609,0.6747856820168564,0.6567685234050241,0.4684147123019681,
 0.4210566928672206,0.18962593668496242,0.24542605912747484,
 0.023874138249602384,0.6336649236804484,0.8985523339661511,
 0.046378291821713824,0.7147895171876855,0.09095378379480434,
 0.8154906946567929,0.5536160004062932,0.1426907449612861,0.31808150783927847,
 0.16441797662005642,0.7273603507754686,0.43246139667543493,
 0.7271046831106166,0.5721724716493751,0.010000913621862284,
 0.18184158137000117,0.44731521827665455,0.5225918741221324,0.676635003448896,
 0.4305147394703116,0.5307976405447251,0.6280322928348537,0.8010848557818098,
 0.28472849619359875,0.5944055719930461,0.04426261635989748,
 0.8287364542402493,0.32099946680083946,0.1865049209458487,0.7454963689667681,
 0.4944703190909788,0.18095311168464323,0.5113609203926976,0.685311984861067,
 0.4910059927884711,0.2673653584618382,0.2671040997416536,0.14473810822471145,
 0.5698240940698767,0.11067781719114078,0.9275939788002248,
 0.22690986623188303,0.1778068973223067,0.8304222672784491,
 0.36000408921082694,0.6130893764804001,0.5240324447967155,
 0.030336568276838727,0.14185963932345735,0.4026653576733842,
 0.510026557409853,0.4224428637739822,0.41120053375559973,0.7772722828305263,
 0.2715474249766574,0.45998015565895345,0.09464051797245843,
 0.11068985746100701,0.31381663273263105,0.03203098674203664,
 0.990996238542917,0.37820962035284134,0.22297578246487093,0.6189499249193202,
 0.3232155664222275,0.10041965821164844,0.03390212803470838,
 0.21399415651717268,0.4229797764120101,0.32163817073614775,
 0.43356622208772644,0.053149895126884505,0.4439976188757463,
 0.9922327549956005,0.8535516403847931,0.246788148648038,0.6886484259232667,
 0.542270187007836,0.34591593849149904,0.7149781636044994,
 0.024377445754095284,0.18782039677125795,0.30628355978013067,
 0.8856578562467649,0.16793538163284416,0.05835654882901753,
 0.8826877178901558,0.2881424757616722,0.7843660484734771,0.17251002302646468,
 0.25149664640971003,0.20257180162381438,0.14171354839941608,
 0.6538460651755403,0.033684537063909614,0.24560726904541752,
 0.6760971324767645,0.5460312550975269,0.2349270601187009,0.43853199927358566,
 0.6536437059931903,0.4095600397136512,0.9344409642114885,0.25599023041308644,
 0.5699847305984274,0.07482962418720962,0.1641034014533692,
 0.13430050162400164,0.2158938772130452,0.47454262990195983,
 0.20507928972156053,0.818383299807559,0.5716500778288806,0.7913461578942725,
 0.44712642523502366,0.8855957034997919,0.907190788103652,0.33130215013784836,
 0.32759733280847414,0.09993690375696362,0.45816827781786407,
 0.10413878352074957,0.8224288365409049,0.7903042507039082,0.6254942893851033,
 0.6045612733971463,0.29581626957096185,0.9909387050300239,
 0.06265193015124004,0.5476388304623181,0.22350005804486361,
 0.6121777466514726,0.30776872554554036,0.838225179768997,0.295483541172339,
 0.25547416245499655,0.549095782387585,0.9021427324490444,0.6697132725760728,
 0.7362894198858358,0.20324326261617653,0.6987140446310076,0.4685325272500622,
 0.1873026982838022,0.1926482360083095,0.7576301768091123,0.3785357468952144,
 0.4154092230303812,0.23351550956959755,0.26081025741175157,
 0.9172159019224133,0.0790263412642781,0.13574726025663175,0.5344707533434194,
 0.027165876545613643,0.39245562645031773,0.6430976882401185,
 0.049027914006456785,0.5433702520371915,0.2713221329911494,
 0.1474385276204775,0.6395330327409057,0.1639560346850989,0.42852077100617736,
 0.07689542814039263,0.6066940057922366,0.626680944587543,0.17859024212549723,
 0.7061807359789039,0.7740467207142682,0.7668775103584792,0.9552694879440962,
 0.8519454742045026,0.6239686466238595,0.6616548143906066,0.35243090801734034,
 0.4450953801910418,0.3765348804935209,0.38851695376572704,0.6480223221272003,
 0.9953204990021669,0.2513154054297789,0.8584400602657903,0.5878630812702433,
 0.4603369346062961,0.40784187820283946,0.14386694719836213,0.838753608576674,
 0.7556014711463369,0.12700744698922772,0.35863341115275216,
 0.15360854742499963,0.7589085031234796,0.9234797770447638,0.3709949065486928,
 0.3542738520174882,0.4906671502234137,0.1919816472357181,0.08634356808026478,
 0.4815293164457044,0.7314240043601912,0.04036634319003185,
 0.29777317934057734,0.24115446743024116,0.22364628842816647,
 0.054370642228192545,0.9170768009673333,0.22217800763158502,
 0.141589452550834,0.19559265488838673,0.9377895852989331,0.9685062327949977,
 0.1592836656601916,0.8723628771461251,0.6427116018474112,
 0.009890400171168268,0.7258910246949741,0.04356406659620593,
 0.30558008305150797,0.6800096552118293,0.7227336270309234,0.5044362745007701,
 0.16693929528828932,0.6552624124220423,0.8155271768354262,
 0.031074713861183656,0.6097691005635832,0.34969691186439245,
 0.7022429596258403,0.9885035279678961,0.2935389920589849,0.5974179010135637,
 0.7758415005567882,0.5828253159588931,0.2509428735443433,0.625412900684671,
 0.7142833258630924,0.7555001839612105,0.510894882628194,0.25924845229943894,
 0.12087717655417507,0.9653467341681337,0.5159791610882385,
 0.05760080482264207,0.7816577294958652,0.5970376249610219,0.5621928321025764,
 0.3691981016292829,0.7262000980549894,0.07442335702920777,
 0.22425208299755506,0.5759005081682238,0.9538450105218609,
 0.028776193892250612,0.41371156513381036,0.7433308381938285,
 0.8515009849510193,0.38524902110386994,0.04956052704174463,
 0.8414651812553473,0.2360705045132221,0.9140304697085898,0.8883838919311556,
 0.9862138293289443,0.6642296897929089,0.672839171261071,0.6287734016029896,
 0.2621249706033971,0.4326206453128223,0.46003646433470935,0.4389530912116023,
 0.7935834497148451,0.3799840305159857,0.8163341898140977,0.9618154106157758,
 0.722839196546465,0.518222433348927,0.9771550552561297,0.3971667616158929,
 0.1816177836152777,0.6326718728024012,0.4895077922942308,0.9769545768967812,
 0.9375556164634996,0.114899772192956,0.2015794294381359,0.2764973614543338,
 0.007562412981420574,0.8284241937207277,0.08176973146680955,
 0.48812286917078906,0.7277303649760559,0.5349179080787178,0.4387314175374899,
 0.6023226270208177,0.003300232847402995,0.09644779509896972,
 0.9292379728138938,0.09175079195757407,0.9164601205410808,0.8348964135370773,
 0.8768293516944934,0.36755999797641536,0.19186639659881388,
 0.5671226085434378,0.38748933086642445,0.5627619621518711,
 0.027849181675750123,0.2999955788173071,0.987175635461026,0.7462935199961944,
 0.5343671000496709,0.7382491402538297,0.933469554527297,0.16550605112318728,
 0.6385685847164541,0.5129492526075421,0.8016022032442438,0.626513133126103,
 0.5856601999635536,0.346443197257122,0.01166729061807903,0.3149348423895513,
 0.6647197801546055,0.18646837997545151,0.06654205554952453,0.613006778216004,
 0.732956120181933,0.1507063997191832,0.6681547672900938,0.9924906602589679,
 0.17301010497847413,0.2513606903400143,0.32311063976124677,
 0.5011952382561304,0.6707680168787963,0.06946200163399019,0.7928424988663543,
 0.6311264068338531,0.5994127800181411,0.852608981295634,0.08001161296205028,
 0.5677240543801374,0.8118800797436673,0.5038232063168683,0.7611673910673966,
 0.8611043021910836,0.2767946847494561,0.42460810369063573,0.7416348876411072,
 0.054692802594830336,0.09689035356821307,0.5710834553926895,
 0.7617053406279451,0.6678964362284282,0.8770550372092086,0.32620272861275346,
 0.7865351778907755,0.06800980612357921,0.053370159504040005,
 0.6709863804809704,0.7018325879701867,0.3448955780101495,0.7258741884953968,
 0.8926299689453909,0.11624254367917775,0.462531426669726,0.8817578453480632,
 0.3482574837937876,0.09758720330908233,0.404779903934826,0.3708346103844671,
 0.8027688153708061,0.5702540210633827,0.4104437811093805,0.07498762243326962,
 0.253969450106455,0.10061600179448893,0.7801503352562319,0.17760896601826093,
 0.6366392722460039,0.06426777885846313,0.5476684096827109,0.184279010580787,
 0.30391043396590844,0.7357855396142946,0.4216849415918177,
 0.060230176145299685,0.9219409301432515,0.4500516112324986,
 0.004031331559312679,0.7197741261530856,0.22180609508197502,
 0.8711818339327196,0.13113474841579964,0.6340124686845576,
 0.36073214723542946,0.005348430767276424,0.3974179705155314,
 0.206430221052897,0.9021782019469948,0.5253717660021389,0.726884410764311,
 0.44629188448727874,0.12124467166876174,0.7110240424070158,
 0.007038298138118737,0.012277795069769892,0.2159635574393225,
 0.6241142646659047,0.01952233488002586,0.21019258673554453,
 0.8395805083188155,0.12371467162322458,0.8669581485266393,
 0.10751704104845983,0.272901439022568,0.5866452325007858,0.2596666508752197,
 0.43768904248575524,0.2709940900060881,0.33249304831456095,0.463226819987162,
 0.21678266477543895,0.8661994899097923,0.2928857334588383,0.7504415466262987,
 0.5058639338427602,0.9367414907064242,0.5559964680550906,0.3381720866241018,
 0.9932603888975337,0.5910927659116175,0.4837543374926928,0.8563391954360071,
 0.16727212704993522,0.18146505985052297,0.8527455617308053,
 0.8614076255096765,0.9710371688157611,0.7233196748975992,0.8266034550514574,
 0.29861072547521705,0.24431100648581283,0.4376687583698231,
 0.7583223945765876,0.6250189552280174,0.9790399396713673,0.8813877398033976,
 0.6657543815618224,0.9519821156644346,0.05648786903019043,0.759053493098305,
 0.046300368445609474,0.13462619817239863,8.013892552731772e-4,
 0.3685014166541961,0.6597278550615121,0.2811888802605703,0.7306643819632809,
 0.8603081178047123,0.16358569035749726,0.9420469082333821,
 0.40776317325335965,0.35867994082589183,0.8827130182249179,
 0.9593429901786672,0.9418098720869962,0.14939488530734124,0.3869653876056117,
 0.8394931436943894,0.505096129917103,0.6987240291939739,0.800082822679439,
 0.2449231635745348,0.32307529962887327,0.11274334052683954,0.741088140424003,
 0.6863278540067964,0.29481894262744457,0.6136935687622961,0.662886790867851,
 0.6557797958922065,0.06225834128466223,0.9440079089539931,0.4929266619650481,
 0.28714495289880737,0.0028940591071836552,0.25462164897538675,
 0.6361173162695718,0.8734241330666076,0.8133293655965032,0.9149880747813426,
 0.27910530083504115,0.5650835910676584,0.9474270944779765,0.8985588241027694,
 0.5103925865815571,0.5738525297537831,0.6568943190714709,0.48190131309858997,
 0.52444826287191,0.21368053546733112,0.6402142539374079,0.30851402716191545,
 0.45035087860207657,0.4898185549462264,0.36404048260155963,
 0.17010480718213616,0.7957143126617288,0.7431065849618128,
 0.06780427889209584,0.574890572707954,0.4900703650377093,0.3641510254484148,
 0.5629281078848143,0.1526104542837624,0.8767603255563943,0.06659435757225851,
 0.34235377222414487,0.557653171334534,0.0791285468495957,0.40898666752177126,
 0.24784472535971924,0.8782557513170558,0.7478168496466077,
 0.20268321696184677,0.1861275546833837,0.6579874597171387,0.9836873503928645,
 0.4900321305522627,0.09885308690125971,0.35456127793098013,
 0.7127515021433894,0.33211489734741795,0.9676619632373104,0.6776418445317549,
 0.2976441501683176,0.3665327140106379,0.8117421573599579,0.11622524885679764,
 0.7907081355262213,0.02377350205521178,0.33298017346536457,
 0.11766639945515533,0.998285640463488,0.46768088889499915,0.6679655783570199,
 0.7820090939587736,0.2595424636882875,0.7263244992124283,0.31975565596008393,
 0.3474580786772625,0.5244931473240477,0.051712651861820946,
 0.6291911476145627,0.21122261687038868,0.6970867159936703,
 3.978188256229531e-4,0.41041482939155094,0.02964182634017054,
 0.3441903580544573,0.7116801892147198,0.4033987217860546,0.5012784712940815,
 0.9582980925305385,0.07717680335154409,0.4133532617003841,0.2993357726450735,
 0.1145595688413168,0.5572078808993391,0.14582101604855113,0.4583221491381815,
 0.4593386571958211,0.3641260210916588,0.5146549333322771,0.29105702216682205,
 0.2734426527530798,0.8001800348056284,0.08259058263403762,0.5310211631161573,
 0.8627284304423287,0.41450106844898715,0.4905274459834261,0.4252328532351882,
 0.2994780922666257,0.2097065035339628,0.2605925265595914,0.05648875699248068,
 0.6743621364371445,0.8122451546666445,0.8507201203250447,0.30622940782383334,
 0.7101469840800164,0.06822092924539058,0.925878153262462,0.6059018383768926,
 0.44780995394388656,0.7414242733378573,0.4998506717894373,0.5519180152472354,
 0.574179089657655,0.18069489078155876,0.819013011286368,0.4457453770259433,
 0.8668633392311643,0.648088296309322,0.3588735766418274,0.7951037075438446,
 0.49176980194267805,0.9546987871152908,0.41749091857452203,
 0.2081663280113092,0.3551009402439319,0.0804883914078387,0.19151890337320476,
 0.02668209372193575,0.5912624516663191,0.6284487210363952,
 0.36046716707166004,0.556490147591982,0.011087898027017928,
 0.2953143468748385,0.4573750114598111,0.28455188001303666,0.4278876801574364,
 0.8164262709362977,0.9146087237158825,0.06738782316962155,0.6195269864685085,
 0.4980759469374034,0.7278009028167591,0.7021873263465911,0.9060278931041539,
 0.8543363663051656,0.9733712832324778,0.18489767876567798,0.603071696627813,
 0.06907243373096905,0.7525975904604011,0.90142636854657,0.9892993862749551,
 0.31074397666164233,0.9363618671832923,0.31472166622989417,0.978123396282411,
 0.4667739383289574,0.8793825678191171,0.4880833968339643,0.21658910427523181,
 0.4140140526124767,0.8998504208305772,0.43457459844749846,0.3122613422684093,
 0.3407098774770454,0.8325765995763633,0.6142367221305485,0.4531260472213532,
 0.2803792349627401,0.8343728923892403,0.9511235854665056,0.8850449507409833,
 0.11166113558201018,0.44558274218310256,0.9723089411596118,
 0.7394880588324149,0.07412703372077145,0.510255227404369,0.5197524112450381,
 0.4442946944933319,0.21581715604894192,0.9685428385223471,0.7287499131431908,
 0.32186037336100015,0.642623365176005,0.49507363949170835,0.6750858554792587,
 0.7469197377779873,0.3617806624453649,0.1821343432525676,0.6191848762845763,
 0.9262678393649445,0.21132988449601697,0.6643314697734845,0.8870867698594954,
 0.06816033386811104,0.460718043684772,0.7224029265682581,0.9813468844119811,
 0.7979379836340579,0.9717049280896732,0.3439483959722135,0.344500855333016,
 0.23438836097628757,0.2880864418165454,0.06601628032317397,
 0.5502976096537027,0.6234285669661032,0.23922104840524627,0.9023143653871354,
 0.41107024587176244,0.15822273111935492,0.6745631979107533,
 0.8073252859888382,0.978419005615986,0.12385043310001365,0.363835724556939,
 0.6869661971557725,0.6396270272524256,0.5174535515927016,0.06457369834900906,
 0.7582904492317082,0.9563058747705439,0.689531269162767,0.5633797379639169,
 0.8548582016672415,0.14263470316625126,0.6351595422041267,
 0.02322154442489799,0.780887759124405,0.06454205267115232,
 0.44420246197892177,0.7196058684899939,0.9869841758170058,
 0.10877352886340608,0.42850888593400827,0.4451646677157618,
 0.7498698555765717,0.2995881534231398,0.644484068479906,0.1664639913221898,
 0.4605304002247915,0.4178338665265219,0.035185767282918,0.44131036440732063,
 0.7838827471717988,0.6355202721643336,0.5229277861476449,0.4991315876179905,
 0.711336799582012,0.6331025222679844,0.9618445373533566,0.14371695418703312,
 0.45258672135567746,0.36173434339969524,0.6729655510993217,
 0.42821463168363016,0.06804661225198527,0.21330834365815643,
 0.6100221153213924,0.7129186440838746,0.7230889013097707,0.12956655325521305,
 0.8040308353812622,0.16566209351538086,0.8968539562002827,
 0.08660834287474661,0.1212427441797097,0.42498431213380927,
 0.008442949958773927,0.2739397297444657,0.1427660717342233,
 0.20689458787829185,0.8787467265692968,0.7852636930536584,0.9976470463560461,
 0.5679461243196432,0.7463451103474272,0.7035668507127875,0.40210330433893815,
 0.505556863296104,0.6342150703382072,0.8300847592522086,0.9384570801913088,
 0.7991541916640659,0.45777928933514433,0.2697097792223947,0.2627173179947704,
 0.4247483787915033,0.23879334849615297,0.12629142227629342,
 0.4816656395976677,0.9710452860762594,0.3831951038885105,0.8220624118921895,
 0.4018069745805124,0.4804434202687934,0.5099626185522143,0.20621613969190922,
 0.6385649895690129,0.9432999689928568,0.8775112951928571,0.08722627477743328,
 0.8035463741516196,0.8082149392995734,0.10262633936657728,
 0.12929434825932806,0.759017865370975,0.3652301361687412,0.9932002382744468,
 0.35984029542875595,0.4766656646218346,0.9598328426429361,0.1359493895573931,
 0.9333037613406096,0.4605941979789634,0.9170957970212237,0.9610204939968051,
 0.6850444214462137,0.8528912693903568,0.9004644133953594,0.29545059832487874,
 0.185181675135487,0.039608915833572955,0.36444918163107365,
 0.28271678227006447,0.6496613250351342,0.050234019376711125,
 0.9783369054620863,0.338030152514599,0.576547513116387,0.6406348887786816,
 0.9222827939493453,0.15250511491438223,0.17705109807211017,
 0.5033868234146215,0.28820605964732826,0.3840443226888465,
 0.09511567288561007,0.03160505067341801,0.290206869452312,
 0.44414852909253644,0.5494748790214018,0.5890409688642658,0.6898603292378302,
 0.27881532416256705,0.13684249321502828,0.035491392625879636,
 0.023782985258075318,0.44515020141573414,0.5609278933688624,
 0.26243379773443143,0.8737260850921734,0.022035529932681097,
 0.3718454188864272,0.12887695529571408,0.6528408961704897,0.4424807316897883,
 0.5960204109764615,0.8025851117223383,0.7368072933769876,0.7020564741442924,
 0.27302601549316063,0.20042776799294793,0.2117390113662787,
 0.5314678906840796,0.6358987465144916,0.6928088496388766,0.02795648590263433,
 0.07989006878688965,0.4929072441571667,0.8605971151187538,0.8365179537651709,
 0.15688040212672116,0.7279008865584492,0.5155624657616547,0.7581298932835976,
 0.27930610941077216,0.33629692693753155,0.6275446020928174,
 0.7277933814903857,0.198657413754178,0.9194894420376506,0.05082240738378507,
 0.14966483378459938,0.022956809683041368,0.7111314878848607,
 0.4799650481198726,0.0988198251520418,0.12897715259213416,
 0.47458394346832955,0.34798091348335736,0.9292864604761122,0.210277717015106,
 0.5011755923052832,0.18481663427870454,0.36651021265285766,
 0.4368278837846593,0.4061809123827894,0.4941330054584768,0.6880332417295483,
 0.009818064123080616,0.870645562285032,0.8554242233086427,0.377415664331619,
 0.7052864714989849,0.3764809750853395,0.8672157298643409,0.7536686215094263,
 0.07457033260806678,0.9818793975912519,0.15388612606973218,
 0.9623640568947129,0.7674562598500566,0.28606801356809275,
 0.13871850790999185,0.12102675773544957,0.4666107226408539,
 0.4730835718713725,0.9062250424905515,0.6578008160897807,0.5242846620228995,
 0.34417558284040883,0.8512003418105186,0.2890508496005775,
 0.39378230033114314,0.953010265352064,0.359331283815003,0.3519611062691669,
 0.21791950341525046,0.8106947862667052,0.020773184374748954,
 0.8218374324756081,0.2383487101521985,0.2555154223249799,0.304800838011691,
 0.6580740102216779,0.461706203999828,0.5816630670413289,0.6979184481362787,
 0.2966528926514109,0.39088395939621634,0.07583960853477045,
 0.29061592409169457,0.7132532280237135,0.5122741170034162,0.9106471242105623,
 0.05337484124910408,0.13064889172406036,0.1028703252931562,
 0.9199212188022303,0.0889799575229564,0.3105995938266014,0.23031274499642362,
 0.30725491079044887,0.21100708614928543,0.3645917992220923,
 0.2455026007891502,0.1018054137535166,0.039176342853539325,
 0.20697092140921858,0.4976919652377141,0.6669727785994413,0.1583759482465159,
 0.5184390764255229,0.021430569725274573,0.41185827559515287,
 0.5007460766737887,0.816336599174328,0.6870039615139047,0.06332337995063853,
 0.08370035506331663,0.07802685749824612,0.4653646759685569,0.986637479055359,
 0.14999737814808833,0.23084619692256525,0.48961722223215387,
 0.8364002366310481,0.6154212776126486,0.7126095858154138,
 0.019120186200690115,0.3680778321274736,0.5704710454494728,
 0.9713719338945206,0.7336779479634377,0.3903172331257614,0.5193197190661067,
 0.6033685868750935,0.44904797962580756,0.7012397011898911,0.6371946487281525,
 0.6229788706318247,0.3686704099219619,0.8748151583364357,0.06597973727289042,
 0.4804741240771845,0.9314752061216403,0.5639036754065488,0.6364155974177593,
 0.9712502630667101,0.04654611268229447,0.21282087000129735,
 0.5749356756380963,0.3743176175737659,0.14968795197689444,0.4005535193463694,
 0.24993833678098865,0.007224322978268249,0.19247371970147642,
 0.3519360772412471,0.35998607071698707,0.7720314324917511,0.5158612935769804,
 0.7516633157716663,0.7334645449429751,0.874069191293788,0.8837318436643644,
 0.30258351973896014,0.24321196676432077,0.28353220694822234,
 0.24795710693322148,0.1357759659258222,0.9179071456434894,0.2886620906288303,
 0.09201825133291008,0.7263753287718777,0.4198197677234724,0.6497618279018708,
 0.7763535458210098,0.6231707092674875,0.22398793812303353,0.1612766560071488,
 0.28609719945578704,0.4623881466110539,0.23105079817371643,
 0.18268006229714084,0.041492494749365116,0.07708460163685782,
 0.20603068906041433,0.23009396003691573,0.854053971128102,0.6928107596215902,
 0.7312696880638636,0.8946301387431479,0.682429911246427,0.6345250659991998,
 0.547969322124169,0.4626471366916126,0.7753305055328824,0.22659418232016015,
 0.8379700845385958,0.26851675774691464,0.9849458382753963,
 0.08565256649019093,0.7611046029320976,0.8638176992674684,
 0.10587841098610673,0.8067271221780375,0.8436354537099332,0.8880357426972373,
 0.39081554368949734,0.9139095977942003,0.8471564155870844,0.9730896388552303,
 0.9616321024353398,0.9632989550824702,0.3265770358195249,0.30276462124436887,
 0.8392725907997776,0.00373920123266247,0.5000840971812724,0.0952261936841654,
 0.5806973873270177,0.6472270637297295,0.9997496002348238,0.05074420258450463,
 0.6433771942496707,0.5565182005799739,0.78529027785874,0.8062799339867771,
 0.409304144239789,0.15213903857361777,0.7045119817397618,0.36074889530210363,
 0.9804820111230362,0.5999590152489714,0.07155644997875543,
 0.24052515303883049,0.6149017191390754,0.16758268707128365,
 0.07562451375682333,0.8121751874081701,0.6486833851740744,
 0.15192463994968264,0.4273298197409263,0.6204364227955205,0.6749645285450573,
 0.41984961084891803,0.4747531229521149,0.09056252090009898,
 0.5281251419309937,0.2248508114660539,0.564044228094853,0.06349865486974682,
 0.4525587459603192,0.774430537618541,0.3409131015658974,0.02062573384001154,
 0.049643508773814626,0.21907376113750165,0.5450965790122929,
 0.09682464202998475,0.1697123285375769,0.3503864340334566,0.3765889634160251,
 0.8375926398548038,0.8705756621501646,0.1529327844002648,0.11997075026048931,
 0.7934115806909012,0.9695511446295151,0.23587999689386718,0.5662124855048388,
 0.763638860329891,0.7502465770185466,0.4755239102386959,0.4809595740144754,
 0.2912670385551712,0.3437708117235643,0.8940653815373116,0.5311568125065284,
 0.34288819530676484,0.6602775866421458,0.8027739884186569,0.7863484577647916,
 0.027656823677929432,0.5213969509799096,0.16288364464426275,
 0.46378638614447143,0.28486812536837736,0.9919105821826651,
 0.3201951145209836,0.23507172039392532,0.4738108938794482,
 0.16731228715226876,0.1707558996588826,0.43571197552399954,
 0.19462945683375066,0.8891048346316442,0.1472393308384593,0.8812422516991933,
 0.918952491381327,0.138839448768875,0.03670737136229674,0.4549691851914114,
 0.2892253134255782,0.27667591597623,0.3043483687200885,0.09641239760665121,
 0.5807172408522603,0.12584574428354023,0.7938863813763093,
 0.23965590353309119,0.9412591143881381,0.4371606377823398,0.3413080938807602,
 0.5990724934214875,0.6469370179420623,0.886929753011807,0.3133295218807459,
 0.7115614701563094,0.24172598377426846,0.42411434390541036,
 0.3083037387765003,0.8768595136753746,0.31989437797337816,
 0.05730171822253638,0.3319349907964285,0.4396182009517219,0.8371004066341353,
 0.26790940742725433,0.3930966128799157,0.955813218995988,0.3914231087123472,
 0.9693502759657114,0.8224456406992503,0.9523248290057365,0.5826692098818166,
 0.9166366842416436,0.3847340490532837,0.5861261290376247,0.43478536552085156,
 0.3748966039186499,0.08509961857879855,0.3007366437544241,
 0.46522310144898515,0.4815846586816095,0.9201820689191064,0.8062803813698324,
 0.44158110594266475,0.8579138308280722,0.08240353775350895,
 0.44272155062591045,0.7691888693819081,0.25521579532165506,
 0.4647766938534251,0.500992787462826,0.13112703810606408,0.3946189150756292,
 0.312300846779756,0.2271875643473228,0.667602712373807,0.10542816109561626,
 0.025345375809406212,0.4482560126804632,0.07872747754487175,
 0.7536535794062572,0.9809588869178744,0.7535790718740171,
 0.022686587282422566,0.17314589666461666,0.0308633069694344,
 0.9120078266975077,0.27795511496766645,0.7395584690145778,
 0.12146238692818179,0.8628871465354453,0.10084036187632717,
 0.17674778131139246,0.6105754919335801,0.36038040416201056,
 0.16391538540867834,0.23603479302091734,0.14626294541240725,
 0.49313156827299287,0.4056493481340665,0.6703730297161696,0.9373529884331431,
 0.34639519735722735,0.5852610649550863,0.4640804541814715,0.2347745266394431,
 0.22857786810597192,0.7723119532115347,0.052001542156419944,
 0.2755992584771202,0.5586484547398722,0.8828689841913621,0.2942691619329175,
 0.4851410590526696,0.28546779417184853,0.9453885225909565,0.6265835063788642,
 0.19397120405276558,0.697994363625889,0.076841954658041,0.39676915216332576,
 0.9345046070908118,0.6925892977769275,0.06476408559008895,0.5625834572975646,
 0.6781526762118902,0.3822537763229058,0.570840384698813,0.8972513766980061,
 0.3688376266361264,0.9257365564267576,0.6432063740013991,0.792817425674994,
 0.629806836680711,0.6703072621632087,0.02088301024272543,0.3706914473479004,
 0.048711972186223806,0.8357440898881828,0.580604149246127,0.0353895076670504,
 0.3428013560676021,0.20177755666160702,0.32846783551159997,
 0.6417433393940858,0.055060960742038256,0.35579271768802445,
 0.10299523257162702,0.6249456958639238,0.7680050022110699,
 0.37150566288828546,9.957043934768484e-4,0.4643707461595983,
 0.6182672316059599,0.9228515798040193,0.6912009471043284,0.4080190489943214,
 0.7204114474251315,0.5464737786763532,0.6383008368845635,0.22936797370543516,
 0.2510481691672771,0.29236179978661014,0.3331594612300073,0.84907544527022,
 0.8408678864526948,0.3207852976428246,0.25481302330195454,
 0.32905372175633385,0.012057072857563633,0.8211708408593876,
 0.771476244941236,0.4039746142432795,0.7239451609618134,0.8676328773313248,
 0.8825420181595267,0.7561630651268195,0.5890846653732796,0.9870976710640078,
 0.2307780043215597,0.7362458891784868,0.02101833787600571,0.5031121659283839,
 0.854436210751922,0.8258107714660904,0.3809457244720701,0.661147130715153,
 0.14679012863411545,0.9880611917121381,0.4993697515850094,
 0.11299948704436713,0.46640375936292955,0.5635166368462445,
 0.3279885240503635,0.29306467192364327,0.33093877516199144,
 0.28903841906452754,0.5723108404849738,0.008532973683191725,0.9236514136438,
 0.06829286213556063,0.7991517522896407,0.8408306746964515,0.8068990505871699,
 0.1357088063351538,0.9457542474123402,0.12976971366496515,
 0.30342587070005356,0.18317860529159402,0.013429996869937844,
 0.4085045462030046,0.455284611950459,0.06479861623469696,0.19752929594742907,
 0.8632227349413601,0.9110800407954691,0.16653870487881373,0.661771323931315,
 0.6702522855916269,0.6911062727601713,0.05934862520696549,0.5612712401440902,
 0.6058435246418356,0.6846621514003801,0.7085677165360539,0.33739973482707375,
 0.10291620866627749,0.19800114499711297,0.9613726415255683,
 0.8626296772272546,0.9147182025561346,0.550983258453007,0.2414934570226568,
 0.9463887527834526,0.11107550460291782,0.9098401096406651,0.5506263985992278,
 0.6879023515991514,0.15014093225911584,0.7267350513958366,
 0.35891646120517584,0.8102882320164381,0.44318576650493524,
 0.8146486205560294,0.8006058369823957,0.3306107772697163,0.15271595466052235,
 0.3445424555017327,0.9945046058359349,0.7221447169350849,0.4796743971946408,
 0.6283656100637799,0.9550908929193526,0.023953924596029186,
 0.9814500894243506,0.6156486658592953,0.883295768770471,0.2194289881358098,
 0.9752748651361842,0.7756554002946905,0.6258727022692577,
 0.040594548113899065,0.14776053663788125,0.024678775793154473,
 0.5326573613387688,0.8236667173510429,0.1632581584685,0.5796018257365997,
 0.7640228101872065,0.6776413716016231,0.7294591135799261,0.9283586788562875,
 0.26231778279635865,0.9054677753690623,0.4715124308852793,0.4631093821487815,
 0.4485724500715025,0.46437133029524336,0.31350892209305514,
 0.9410862839165473,0.9501184642899778,0.07248608851167226,0.7243347916372074,
 0.6549642265648461,0.27395306470515723,0.13049502848001993,
 0.8408543482406089,0.7519306934344464,0.9700496117331198,0.2095182616848552,
 0.770390174371763,0.051833353988453346,0.8745922729098176,0.9097387037425455,
 0.13587232397853455,0.5621172106491059,0.2041256964656193,0.5692734796931553,
 0.625826781146704,0.8033757849373344,0.6328736653040444,0.26103241947395894,
 0.8523355295540457,0.6673095584582651,0.5007379256820016,0.8638886672514449,
 0.8880836438506416,0.9590670212599391,0.4130298368203431,0.18142040084460354,
 0.650490320866249,0.6221979268900768,0.9920886801000067,0.3559842274318854,
 0.7240178758863879,0.2281871786702213,0.9191167009708917,0.3716269919706886,
 0.6095855497692985,0.13613922805785783,0.2535469484549119,0.5613693193229087,
 0.6268880196716228,0.917802977960203,0.6272990437594874,0.6837845591636528,
 0.7883734607153485,0.765119195635608,0.8966718377885861,0.4755689751233655,
 0.16377995898968745,0.20929016751030338,0.6436728485814817,
 0.6907127914816373,0.89579828961117,0.13032925316580934,0.8972905598085932,
 0.7205212625458867,0.15841440031506648,0.0014092970295758178,
 0.35852301775627426,0.8429903968652396,0.9271299298044449,
 0.25997517596010955,0.5825430194613199,0.4364488790523502,0.7319378120086262,
 0.5473222195422836,0.7551406666335689,0.9645241334903256,0.7826320441384849,
 0.08442162323911451,0.3793695347097872,0.9221324608449403,
 0.11585428878448689,0.11191757792365098,0.9667803997429404,
 0.8355589052565277,0.5322475411817792,0.21708379359037921,0.7961089484161799,
 0.990039541835477,0.4386041466568915,0.40165796460432535,0.3524822398158206,
 0.508881783838382,0.04695107519814112,0.7961059957892528,0.7217903395816153,
 0.4546733912941111,0.6653133248037303,0.4086532562301618,0.6688114759272668,
 0.9618412401465021,0.9991367290130457,0.4763076058255563,0.995196764632474,
 0.2011200278976224,0.3340945018714524,0.20622117847538768,
 0.06202321589319715,0.8321514719580341,0.3273625381839175,
 0.29684583277039933,0.5418515854864541,0.33389200740424096,
 0.2208379594180664,0.6562101776534524,0.8593494967280537,
 0.0050978896464662915,0.46808060256445616,0.11334432501227587,
 0.6624275509154247,0.5114634364554496,0.29762259572398586,
 0.10851010336478384,0.5907096649913992,0.6669995213293404,0.8974392977168375,
 0.48608836652955456,0.7795911881002026,0.7716508883780779,0.8878676662682542,
 0.6531747352242303,0.5567734232807954,0.31463282038124085,0.6550136233388966,
 0.6624506280067998,0.9075121198569358,0.8803867106023451,0.4615886766748316,
 0.8666947314006611,0.7328357876400209,0.5841802614771613,0.7562686826152298,
 0.6090066616183026,0.6121592023502779,0.8781565786288332,0.15127177157761595,
 0.06299604824797056,0.46490825222818966,0.4699059908194083,
 0.020626497849886904,0.7906810856864711,0.7392794953573277,
 0.8458294901543059,0.10746533538920511,0.6531952683510868,0.3778873045975084,
 0.5077881934074219,0.9286818205651026,0.2991378008372586,0.08927482939357534,
 0.39838531706754987,0.793370952079882,0.32571781758158513,
 0.07829466019812725,0.5474552650863881,0.4708113601015459,
 0.06485382841737508,0.25977084397083405,0.16752460426565996,
 0.7296712589744763,0.26646496068784253,0.31691085961192056,
 0.6639301904021339,0.8756565544965049,0.40316385630335494,0.8762876136144326,
 0.39455763860643844,0.20011852274699438,0.884195689883578,
 0.004754221431450767,0.35302505109689286,0.2354818496232739,
 0.585074058135213,0.5039405371447565,0.38416025429113876,0.7322391951372692,
 0.3509988002536514,0.9429843055120131,0.7691748274660715,0.7861854095150617,
 0.10518482687739072,0.8092514903524209,0.18633080830141768,
 0.25736051256689985,0.3013679629651459,0.9288696284329292,
 0.05679655235133585,0.4663099625733297,0.30638767924034627,0.819620483306732,
 0.5623179275206801,0.7601287744614779,0.8825379071705435,0.40483833067496344,
 0.9237382154328524,0.33891764541731295,0.7979872193938112,0.3862719190518429,
 0.11569439269739545,0.9122914059489747,0.6037294076427155,0.5983345107244157,
 0.669376243956215,0.5640332369313624,7.640777115054398e-4,0.375752794768749,
 0.7718672476531164,0.9479735730162366,0.2821073082927663,0.9830842023510452,
 0.17037666631668547,0.3327390194294071,0.6094305773100032,0.3070834284965611,
 0.14098619226936127,0.20435798672682326,0.704139966032213,0.9623217583640362,
 0.8357931807618083,0.9532006848285598,0.8067903857916109,0.25900110587408376,
 0.4367793493993881,0.22534361455903829,0.9749555668181535,
 0.11905626807191794,0.7857091126905764,0.438310289395743,0.8222895621002491,
 0.14979338009431564,0.4023338090787998,0.7890686072571537,0.2857728756628247,
 0.2716596736838769,0.8840205872921205,0.4405700708397826,0.9275150620740712,
 0.18372642592831645,0.20639034946681167,0.03849635849188615,
 0.04438444488372251,0.09889624344659409,0.9083355520656298,
 0.6826168641450365,0.15661951949592712,0.316503637569014,0.07230496277339549,
 0.3601128290308159,0.9917825169554164,0.7153392925698415,0.8177127017111988,
 0.3970869975950878,0.9476623139261813,0.1274074089888374,0.7166592683997861,
 0.8127999695386888,0.39906292211205074,0.2854843073161847,
 0.06679392806981777,0.6261034535275862,0.08372247635312768,
 0.9130085378569524,0.51758014491232,0.5526494970724954,0.6576732112941363,
 0.825935263145683,0.8665649280488428,0.232498657891676,0.10794294571349561,
 0.7508078443687375,0.5238201439405962,0.04877183596974688,0.3034136411464109,
 0.8316436011891621,0.48178209141318884,0.14770787045412315,
 0.4567942412123075,0.64789066691392,0.1917981576995751,0.9310538146999016,
 0.7380985179512353,0.46128776194979837,0.30493340568464,0.15638070300141416,
 0.505205610283248,0.3774005749167918,0.7847761775432414,0.5847522553216293,
 0.6788960368871615,0.23244568752685157,0.6691539724888467,
 0.17031249588457464,0.21722563975190368,0.054766442141088945,
 0.17952314396209745,0.6994946608141773,0.7938668365671505,0.7153417975611069,
 0.65634756585471,0.44532460358372494,0.4070547747228058,0.4978745322431518,
 0.29266174690509683,0.8520243174381972,0.3077243293674069,0.9384710142933793,
 0.9900659168742026,0.8385869358076881,0.4720457427906015,0.6274066455516749,
 0.952681262358335,0.7319973871456277,0.4333714427605776,0.019065096259160974,
 0.7491251856146501,0.7716588482149855,0.7274263797302779,0.16430518606237965,
 0.408094180335124,0.0997875344665089,0.023953665110167166,
 0.21829797076882584,0.3311588433829806,0.6278095669732523,0.8952687409158026,
 0.03270346021804116,0.1306342484685078,0.145952006474526,0.4285395033289999,
 0.5059943204118591,0.675685637373702,0.02839589004795362,0.642676824062477,
 0.7029856177124532,0.5159410853484364,0.7536906636809539,0.7828240238125538,
 0.1964138037192693,0.6834714585907717,0.38262630150868204,0.5778434543242107,
 0.5896816497126433,0.7894628486155046,0.38185100667525207,0.9128542490395136,
 0.16014900048245972,0.08022543517962455,0.2561857894444336,
 0.44290101839695417,0.7783961985844015,0.11454701962043046,
 0.03115581350419383,0.23370332509741665,0.5770054671016024,
 0.8461937623372614,0.5418347144610864,0.17993745777858372,
 0.39979873105299335,0.5520386407886099,0.7249945745536605,0.8568098771375338,
 0.5234982266064836,0.08683354122580633,0.16685041864191763,
 0.9179264104260639,0.8599437182202423,0.2738165303346134,0.2263172703264007,
 0.18136516497512745,0.9488582096929958,0.5541663697028382,0.8044643896301384,
 0.2410545955623281,0.9721904211869252,0.03202516661814159,0.9836930185732837,
 0.4037280348947151,0.8724411369877423,0.2626534176366293,0.7261503486603479,
 0.23482193988811784,0.8728708332860469,0.08159754048012657,
 0.05776952375109212,0.20610397665639546,0.7117633469237605,
 0.7760618515392941,0.4674963220037107,0.7966545247612266,0.09792247474845961,
 0.6324277477792986,0.6168452891904503,0.7992105620944004,0.8840513477164075,
 0.38844942209284916,0.7150318672209258,0.3133596878304089,0.8009524396734133,
 0.0704410285474002,0.13594529439684744,0.7754958076035927,0.5119767546295447,
 0.5193750693566909,0.08833622793920126,0.9219320321245771,0.9851059536740341,
 0.843714531145967,0.8498861820937256,0.6389411272642218,0.4700231065259235,
 0.8848325882221131,0.8301424867559234,0.8605225234556424,0.8070037239771194,
 0.8433267491440375,0.7641798779766994,0.5734491812679138,0.3478397918736372,
 0.8028996382653282,0.6073937634506731,0.5676320990325093,0.9134080218341549,
 0.19886917738645615,0.0655969634966983,0.6868904547674768,0.2876033876344123,
 0.9059124703277454,0.6317381940137854,0.7057850954765277,0.5928657640304962,
 0.39411209758620636,0.7797610964281372,0.5430352575423707,0.6455048091212451,
 0.3257407370449237,0.127789955739526,0.6117746044757989,0.8984665353438441,
 0.2352143181470845,0.2136669397849108,0.4175068343110686,0.41955566919612797,
 0.28994129710127337,0.6565106147330513,0.8636311372443434,
 0.24897337416417198,0.9062726100999167,0.6641407671606113,0.5060909365737903,
 0.40551929925455177,0.5230795579503883,0.5733023979569666,0.6419339765936362,
 0.27785386286505986,0.9157196979887846,0.1908469393566935,0.5703087227238718,
 0.22359208095599625,0.8573374822791863,0.4288347136086713,0.8948519935742629,
 0.10871061316588249,0.0757535578710391,0.9254774633569199,
 0.07840020059504993,0.6121622272602223,0.8935613204734063,0.9180204479171037,
 0.8312646417084194,0.05612392542952638,0.041846855969818075,
 0.4815009943667564,0.6737856977284906,0.49334363340885734,
 0.23366345980677838,0.5115953567686748,0.509309576628532,0.6127154043053535,
 0.9663825285831463,0.3173672365396829,0.1828699462611586,0.9893459680394655,
 0.648558586637101,0.8860833910797719,0.9037053379664824,0.680282797953365,
 0.9743053692889533,0.4004102356053836,0.051257428818486606,
 0.2840571427780083,0.09197131156460325,0.086869986130783,0.28767316053987324,
 0.8106071291149586,0.09119277547265747,0.3974508364270233,0.3387688535805825,
 0.015506960363568578,0.3945147820154582,0.6624888290415328,
 0.3469926082628997,0.8915274758278723,0.32198361824712585,0.7008888202826979,
 0.14862451011395617,0.5880800324527523,0.21915750142228474,
 0.9261034114512631,0.4804728858281434,0.7754072431093035,0.7075341638782253,
 0.9960100084939826,0.8572618646856031,0.5259751462962517,0.6890768397681846,
 0.058370582081172295,0.42053676766773906,0.2540812196279,0.720635857926349,
 0.9028433741819634,0.6322317073007035,0.06834643051229172,
 0.49611268832786304,0.40254673561968535,0.0693612946866905,
 0.5693548934363115,0.06577154531095308,0.9277982324656395,0.4781986469686115,
 0.8888726826952572,0.2361316961919132,0.9803658351140188,0.4788834016311948,
 0.9219684491787106,0.08871602379153243,0.6943419029586966,0.8314341184342137,
 0.397466242810528,0.07771312382021267,0.8983194298598425,0.49221423975661405,
 0.03176879131183563,0.2383141323860709,0.30664599950797955,
 0.24219181534694412,0.1480934201382017,0.46519015665269026,
 0.8055453285174619,0.3697358897510453,0.8973155894386531,0.16747506852911054,
 0.9095290838317428,0.6612728861397303,0.08300098424765356,
 0.31997051057253123,0.08389929668146157,0.9287542439120483,0.272057099013729,
 0.8697135618531591,0.08672077854726667,0.8450161702661381,0.3759013676899424,
 0.25257076895902464,0.3996782472404945,0.3595397497362347,0.3514583757691939,
 0.3391088028932535,0.8741102911237324,0.26715671854080925,0.6549524504416716,
 0.33133352315377806,0.032478904142166365,0.9283284424638256,
 0.24423645154286355,0.907502586821623,0.0025057602274698443,
 0.3755008556204651,0.04936236349916956,0.07930981873705212,
 0.6632559479234463,0.9641607516818707,0.6999931101740227,0.8241555152042936,
 0.5723960813100506,0.33219011270475474,0.6720480599282866,0.7525278929677597,
 0.5966028169771193,0.5163381070368184,0.1768977477326873,0.8281440917047921,
 0.1909018001850009,0.21205766653929048,0.7442043529349753,0.9062354426218371,
 0.14858093716093212,0.36831002527139234,0.6381847371028516,
 0.17482572407715502,0.5197923642788705,0.16876844413029102,
 0.7154664045757063,0.6287170751909645,0.7289285229427289,0.6156943330247167,
 0.8197928977891985,0.657291113302509,0.4761133282982536,0.6854708126692785,
 0.3501924500313256,0.8967492150541095,0.7170019014077617,0.7341130718583306,
 0.19799914350017667,0.08784732953798002,0.7018037771257803,
 0.7667738363056981,0.21851021780013935,0.008411284892651305,
 0.9641002413074611,0.13204191747932537,0.6052831178286414,
 0.39674696413299104,0.7678806653438326,0.013000500322806308,
 0.0601829059479303,0.4166570658620157,0.01897574593028184,
 0.26659346314952903,0.8346831471639998,0.08524931887816012,
 0.23590577072441143,0.29771889916225414,0.05441475550590935,
 0.9310557141939138,0.15890598443423487,0.021740484313914177,
 0.4568354449033454,0.7987907058542756,0.2471458176849627,0.23602339592073585,
 0.37595110101461426,0.5450121091150498,0.685544570269132,0.5860290976115494,
 0.6717901736229537,0.8540018549744585,0.13823093156107635,0.6146185819829347,
 0.8465243972652496,0.28597755746172404,0.6343311180199817,0.4129863888073435,
 0.9838466431552092,0.23882275534612452,0.37936681507718784,
 0.9819609423014977,0.2452647486003322,0.7417001718483613,0.9515445181309525,
 0.7383151482955828,0.2848258235160679,0.9749685892811425,0.5209181320285009,
 0.17528348031585328,0.8688089529911615,0.3268935484502561,0.9818142555482553,
 0.4841759692581754,0.37424136187305546,0.7053687016583371,
 0.18597948936456776,0.46068689003581587,0.9151450234657982,
 0.6507684712642973,0.18719043697640703,0.5382209717392683,0.8300931073298574,
 0.9710144958033357,0.7614714809148865,0.6728023406995713,0.3651327513219319,
 0.7475513317590641,0.7381387041186538,0.03276598865711311,0.4840885188307894,
 0.21450114963531086,0.5284413967427904,0.6890056353117481,0.7236581670994578,
 0.34992514733015234,0.410364945496265,0.5755389138115997,0.7334523126893336,
 0.7588230969069827,0.6996829701833406,0.8547132015657255,0.7980045892351232,
 0.7754278388616342,0.2650856161431311,0.8134327888985012,0.38490609558223354,
 0.25507662544384435,0.1180222640717229,0.9565462372931824,0.8890853688952195,
 0.3271173533212687,0.5290521485135918,0.6935205350461577,0.07725524324224597,
 0.33345265305516636,0.8889060620695087,0.39185002076003694,
 0.19924131855684268,0.530845903243584,0.0795705029711854,0.9148119348997372,
 0.9316187136875406,0.7557748970033655,0.6279102999681158,0.04689590158037549,
 0.8387664289399235,0.9491534343895537,0.7103457042709722,0.6383111847799829,
 0.8507582981700984,0.5290048917167623,0.5110533293461608,0.04648546510422413,
 0.8966643031255999,0.810670386947634,0.4287469355598622,0.3078144154737309,
 0.7616988943026745,0.7600910863292248,0.7351462834332413,0.431512967213178,
 0.811134053885121,0.6013539034764708,0.535296321316459,0.7634787611305984,
 0.4263035315417637,0.6224435956975907,0.738034398533191,0.5799152354312062,
 0.5668144364290983,0.2932297690123432,0.5273473256300244,0.5693893353528071,
 0.5932062709734568,0.5119036648425458,0.15821845749221064,0.7758547868175304,
 0.7349946889511123,0.7138629118266963,0.3074404819046329,0.7338858156904631,
 0.91401464496559,0.24326000471014897,0.5567571781609169,0.6309022702053868,
 0.6398207566552597,0.7281829039220693,0.4751282259175358,0.3721918159685953,
 0.6092462016395488,0.269659152406891,0.47084924707081066,0.5659176640537567,
 0.7752501317107974,0.5613587412351934,0.806477499111508,0.4169123633713975,
 0.5477235841637625,0.8133425447977667,0.3530624410855209,0.6369547803899801,
 0.22413906957906982,0.719449757235605,0.3561893898836801,0.36324507718398014,
 0.8089183931156261,0.8602523343520065,0.6233710283796882,0.2241643152548929,
 0.8895643374145996,0.30455373668116514,0.2782519191098497,0.6739834134192028,
 0.6662298062952935,0.23985504123066803,0.12570098930324058,
 0.9753739754794177,0.17838950359613137,0.1973915670051264,0.8788273530115978,
 0.31409880902607523,0.8681170535443212,0.36109047960037877,
 0.8136301135527719,0.6297850968621699,0.6438890489218941,0.7868397651390591,
 0.08349665151478691,0.10983015982113431,0.6659261023095102,0.959866031483951,
 0.5611676608187872,0.8143957968219973,0.6518412208931554,0.40917723606722856,
 0.9300520068347196,0.6167347757506866,0.6719216991221609,0.47869168464419865,
 0.007598961805211579,0.3930006946272754,0.19890046182504761,
 0.9429235186064242,0.11342755384228975,0.4393846240918693,
 0.25108347197293623,0.8586583586027786,0.17499832462451725,
 0.5706248343960203,0.25382155545278295,0.5890253220398557,
 0.013119499946482316,0.9897020996938148,0.020500536337784037,
 0.8076672198794554,0.48181941746491364,0.3134482522031836,0.6114665586827109,
 0.6938425136441126,0.40896642537913874,0.07189299128490623,0.262861575059119,
 0.45092694919927057,0.2918592419228282,0.5791820656678022,0.218098524989096,
 0.30601063988455635,0.9061207607039258,0.6638552182595334,0.1685729180706521,
 0.9489635600284004,0.6028638019268805,0.0859752374372248,0.9789000492799927,
 0.8475135362862765,0.25724272277158533,0.3819212960960927,0.7662569827126219,
 0.6909137909472498,0.09506933641579762,0.8408561562190484,0.720398471433215,
 0.9350852174550603,0.4374072892464195,0.8650526937134311,0.5193648809841563,
 0.962128037824101,0.6032865607969965,0.6727356442124056,0.9930718577514641,
 0.495336641193312,0.21510303038736145,0.12764115280952604,0.7182235603425839,
 0.14130973499398203,0.6886531030043106,0.45231797631863657,
 0.4354115338126371,0.8243278122717538,0.6094518311432381,0.8197573715989723,
 0.26735719470199215,0.45495893863629133,0.26471039927565543,
 0.6199895387958427,0.14206731209419765,0.2907569062008274,0.8676917475564634,
 0.26519354144363017,0.35553865463187007,0.40985409488321844,
 0.9007185088275925,0.7906094997917243,0.13969754302674486,0.6725866640825819,
 0.85889060137424,0.40265047576116264,0.07655962727323384,0.09327212347495129,
 0.2589014675071283,0.7712805098430571,0.5133508347511372,0.6065925785718246,
 0.5482038176625367,0.6830058132819243,0.7035482577887731,0.5825147180251161,
 0.4698117503184991,0.807792720290091,0.04388216996848693,0.23590270622356801,
 0.4572097390769104,0.3122109563316502,0.6760371404836979,0.19267696242499455,
 0.5967069098947317,0.4382100863590267,0.5155841328216159,0.6125502449592551,
 0.4143126046404929,0.2879108925377897,0.4964154983872199,
 0.061430118951405155,0.0400624056992791,0.03826041441099903,
 0.7780271941495984,0.6734503494035757,0.1642741459029613,0.3773552802008183,
 0.11815665433711176,0.3395606685239586,0.7117108749259988,
 0.013311524194497104,0.28331042252999006,0.11995825671871307,
 0.8945526723503007,0.7456749724702012,0.21539010224414856,0.3928071180767485,
 0.3638179085067078,0.5931355206333082,0.024431408245757735,
 0.7692576187089003,0.06493741903294226,0.3907859805406826,0.7845504016221219,
 0.5060311540077982,0.6186179925118619,0.28760983132485596,
 0.47975646869079813,0.1524375682360103,0.19133868591157377,
 0.6577502737107588,0.9727584621337274,0.4100113837411583,0.5417843796675554,
 0.044039531441729896,0.20103860607548962,0.14338382307463637,
 0.09497692746031994,0.11059867158550485,0.40195352192037626,
 0.45365765276909287,0.30781831741087073,0.4598753039123894,
 0.026528733110602554,0.18108496205426172,0.42961721486570215,
 0.2979554992084352,0.8824251010294966,0.19187593277573045,0.2523974018868974,
 0.34186341226326034,0.44223077687755497,0.1675421146899163,
 0.9359876063301834,0.7826866664770411,0.8893921013897714,0.8181793019813274,
 0.3445660384521301,0.2428480555953546,0.10490057786871043,
 0.19968638710302433,0.3970140613856099,0.10946784372282581,
 0.35136068999801195,0.3511760433660607,0.6984745132420203,0.2777193353576405,
 0.6259010260541253,0.20426790562380848,0.572415241591409,0.11864862858530811,
 0.09791143264945612,0.134597042705064,0.6111934041970345,0.23276313598787612,
 0.8914966516186703,0.7421477097745973,0.379054856025427,0.637567293485418,
 0.6465539499679305,0.7754306630181309,0.0634814979315047,0.07926116210077194,
 0.30066310794930173,0.7090730284504312,0.48530270022582,0.7185260440401475,
 0.10834579171942138,0.16111082128884924,0.051905961799900924,
 0.4148444990545652,0.023894066039379047,0.11293331781508809,
 0.8094624602710773,0.2813094304382494,0.2808943104848354,0.40184368585761954,
 0.3807034119836634,0.48812549379431003,0.5618111138529558,0.3724756010976089,
 0.4048302756196326,0.1656136018432346,0.9732698231080773,0.23456475177252833,
 0.8758440661454466,0.2749911445793066,0.03537194039838809,0.4337207510054031,
 0.8941880639777104,0.4323509014706729,0.31966888388749937,
 0.44968342426036867,0.9018144263581207,0.4859120704071458,0.5150077230656173,
 0.04983700030764404,0.7974716886177202,0.2995704016888563,0.8940532594902711,
 0.20098850095182808,0.943405134142757,0.1017606758339108,0.9102740782018638,
 0.9154895642414125,0.9663648921680789,0.600755926772958,0.11684531600348869,
 0.14616870167437424,0.08893072313596029,0.04964935750417665,
 0.9362444649991712,0.17103545518089402,0.9070879018908471,0.1649264572742769,
 0.7040127698983789,0.4790862240987819,0.9782219496841512,0.06965973549364912,
 0.39100916150862886,0.7899351761047124,0.9878055162032624,0.3909466855591499,
 0.9058307172523699,0.5164513717441551,0.6351298753483129,0.824863972550264,
 0.0035012601373839436,0.3196817234229995,0.8722681166560965,
 0.06446624106589116,0.01341417484054197,0.7068693050899715,
 0.3506431198544765,0.0060688386645004755,0.9490560766401752,
 0.9221862983078521,0.9915060883854414,0.1062731536848136,0.22688397937135096,
 0.0469135359274272,0.30144449923309535,0.3096154032241145,0.5735505344625378,
 0.6489039023977722,0.32803197570164955,0.39823469314967563,
 0.48430469162847223,0.27099154769556666,0.21660872506815876,
 0.6088813246995735,0.3891343505451813,0.5799546948469917,0.12837233589239327,
 0.002948625178701203,0.936205526683429,0.49507191268906414,
 0.6566785210726281,0.19474017921992592,0.25646786901292595,
 0.5394427343781522,0.36868455390607335,0.9330827321124135,0.4762137626948273,
 0.8550355754647755,0.2246368319727733,0.45979917005453685,0.4943233567209768,
 0.8064353770634227,0.6409162713890352,0.4308472307893173,0.8179272480249609,
 0.5700831837272069,0.5230488485784429,0.31556511356365613,0.0314938580360451,
 0.7218144660972672,0.8679075586392454,0.27105229340491965,0.9777611319997296,
 0.9120875884458922,0.9430434653775503,0.5048002707732281,0.7642437561982787,
 0.7427670383407159,0.9983397587567755,0.26212574251433973,0.5304224854193466,
 0.39231400173897435,0.9059637842279799,0.8133065563604155,
 0.061705618893549974,0.9215109110157571,0.9695871660238521,
 0.10879781983692971,0.5368731303383623,0.8411962241772137,
 0.05806707846667458,0.048872338297282436,0.03389275069917441,
 0.5551770330214375,0.4057867268208494,0.3830025823944765,0.16940105874629463,
 0.8614277727955031,0.2802620632263181,0.31933622602153555,0.6196149652872194,
 0.07534351010070461,0.7865403277433619,0.415835394356064,0.5938734129958344,
 0.6109061156630303,0.1873058517456393,0.6732173143924945,0.10023364819246638,
 0.2151170752853664,0.6911476395993938,0.9035932996264602,0.4012845511076887,
 0.13462024896556524,0.003973705922439286,0.6541718253160664,
 0.48058751586655013,0.6318334181713409,0.6851206609958925,
 0.14834108561145154,0.9044479719378773,0.5771876566237417,
 0.05941495658768292,0.425022245423463,0.3363533459206439,0.32108926949002337,
 0.16714980769650478,0.18361045047966462,0.41039440866815335,
 0.7914782925827194,0.0035568952245847996,0.5209572717148143,
 0.43087455278880094,0.43680706689655724,0.7830092061144381,
 0.7747947660529105,0.8526089260200638,0.2930164568073713,0.4170302135220345,
 0.6335624941180255,0.4683015868048559,0.18568382330548128,0.6777488105230284,
 0.31959927662832954,0.9998735245056598,0.7800707761269753,0.4700354960773232,
 0.7622831624785145,0.42098727708022177,0.7122505282797917,0.9641274820127048,
 0.4080717962851179,0.7747755600642321,0.997187208559552,0.6951993316833568,
 0.11985583138772249,0.8147106299789156,0.2558090801994094,
 0.27089515136470343,0.7935635520669824,0.7378035761394834,0.4396817786651258,
 0.0293727139679808,0.740462635937625,0.655647797940586,0.10482762532422119,
 0.2697449109978711,0.9672179687899343,0.6450256165313474,0.7732757027202093,
 0.6216992645463904,0.1844637685115038,0.5639988526964899,0.11564088228760339,
 0.2589267739010451,0.8439626957090031,0.6368915281024269,0.6965122823606211,
 0.8963945431052992,0.7081970524807332,0.9497301373326589,0.16367441621433265,
 0.3124183823201242,0.09023680570859904,0.0885829798866209,0.8872826469294152,
 0.4252276990441576,0.5137188736306917,0.7073295593174846,0.356330681438887,
 0.8538492701722866,0.8263021947998792,0.8084130883022449,0.42417202835045953,
 0.6651222117119577,0.0722769397765004,0.9144011193903636,0.2062993930865119,
 0.6587521514171641,0.36859124421072553,0.7179766633828185,0.45321591119018,
 0.2573098102498549,0.35036492721864665,0.808758938552358,0.7786387406635977,
 0.9192469926814887,0.9035080728865481,0.2692682332095222,0.8027969658747635,
 0.8906665333641596,0.04357284798707095,0.17128271589493016,
 0.8726390560414358,0.9952888392000874,0.54181804787991,0.1818206065080703,
 0.3934955394285561,0.655670064771925,0.9761399550110483,0.9010619777247953,
 0.4196392705104546,0.009903564564026524,0.7984922469660439,
 0.7700963241491167,0.8149226319628088,0.9513223859560962,0.5971115900485161,
 0.2954098775737063,0.34399703528424497,0.04120777045323776,
 0.8225145351201733,0.9006691509972505,0.5158777643308256,0.3243004376311506,
 0.9019477407772256,0.6089479226009809,0.37129025429062223,0.833560666262425,
 0.02383282039560819,0.8258044084847775,0.21135275658076458,
 0.8715728576029523,0.9324227006064187,0.27230851351879226,0.9615494877880874,
 0.056529092641042666,0.24172246721917612,0.4145085929484692,
 0.8085602638293115,0.5448494777775966,0.19964673081987094,
 0.37877749636552926,0.5201945526420069,0.07768908981194123,
 0.33233194752558415,0.6452015963835763,0.26414522043194844,
 0.2058201063252515,0.6537712727812033,0.9292767745820341,0.10599207987252757,
 0.6835875164407791,0.5090537740769746,0.15393759028741316,
 0.17411429681026314,0.5552465596287206,0.9024387422557214,0.9568634308881846,
 0.9881920199918095,0.9007729073352072,0.8552134120140663,0.9089085690067159,
 0.3366317887960165,0.15287383858419445,0.21767799726223047,
 0.36315682131651617,0.5912600936382336,0.9469761732134403,0.2873390396516944,
 0.2388578721587742,0.5627871088267525,0.9277927942909006,0.4553368542066272,
 0.7457746957388551,0.26762027307061975,0.4287522585074496,
 0.048451743587422214,0.20603855592036047,0.03447366619549552,
 0.5080249260257914,0.5034902567252925,0.7830683276839117,0.48761562475440934,
 0.0289295497683818,0.7578158356482704,0.7281362395573406,
 0.020307540771484667,0.5694749995804329,0.34397319944030624,
 0.6255580960404332,0.4928397800728411,0.6864162563559212,
 0.030987777666984462,0.7813567217250385,0.1183940054259619,
 0.9714522962775438,0.3046647549044821,0.9504888844806995,0.17667980668197047,
 0.061771742327596524,0.1784389847173452,0.8819078167695019,
 0.7133863378358746,0.08077849848217378,0.5121757142489346,
 0.02325710103575418,0.5308925900333741,0.001274308036801064,
 0.22858569934212236,0.9657140034223993,0.12480797323976589,
 0.8950242392238755,0.7859644808757157,0.08189217191442077,0.7644308234357449,
 0.8984755304131856,0.7441993288707622,0.4054527956610121,0.17758247520044967,
 0.2537418172443968,0.3816189082747403,0.06742033351522314,0.1535649848883226,
 0.09810656203393575,0.005177708825321314,0.9410113008553148,
 0.7814322584379556,0.6572813725299027,0.24659958679674054,0.5875881845023815,
 0.9488900523750556,0.14889447154288682,0.0996460227558097,
 0.23380128690525737,0.17826400946225907,0.7987587692855935,
 0.6239926806572237,0.22646727028858504,0.5333225465744809,0.9005702925447754,
 0.47975041084899994,0.18782525159779384,0.4161015410853869,
 0.15502858035273315,0.02213244908025558,0.8565400401762132,0.501367221166379,
 0.5963860550643341,0.2640463380312088,0.6084617141729705,0.7890806028591001,
 0.7840518540215876,0.8299900388873265,0.2614422727011576,0.804260917153745,
 0.059570469700772066,0.8058290154974604,0.7589556299785435,
 0.6227979610652243,0.29163231471293616,0.7936545627206408,0.2653646897645485,
 0.35799394698602516,0.880120956830182,0.683109378149359,0.7005014329640855,
 0.7060988599885479,0.19878921015192752,0.5481349799170423,0.881946359303208,
 0.6703615529958921,0.0620072714203852,0.20443920790686743,0.8651225876610272,
 0.9996142222567087,0.6264141587900979,0.5304116998759935,0.9782732629142257,
 0.3229237147910714,0.9173233755959055,0.9072892284160756,0.8649789273061328,
 0.19521778269976042,0.7061829710530544,0.567802897635356,0.04513925293653087,
 0.6338420958184706,0.33743295716603283,0.5036743340443227,0.4932095230084613,
 0.9021688428218999,0.283458734413217,0.3024281554956275,0.04556550073816068,
 0.5486132777165923,0.2953361681168849,0.14037018987411498,0.9423911974603805,
 0.11457347925203842,0.9869077364287261,0.5410541793763127,0.6976465972021105,
 0.5817791646004146,0.31138841664238437,0.5904794567419729,0.9873724540139672,
 0.02452854498325363,0.643849934414151,0.7628213691808137,0.23828712031615462,
 0.794368166575708,0.3848231474967757,0.032606744136536525,
 0.20146884573218282,0.7069122659438869,0.7268178396185027,0.6708115876719185,
 0.5801384986336551,0.732964960557527,0.7748835157964659,0.9413007461833097,
 0.1792257814856777,0.0652661799502463,0.3969520333584905,0.49413878648126563,
 0.03740430874429779,0.11587580767600658,0.8421299663778934,
 0.4374389425733367,0.5702616449773681,0.6717248175892945,0.41413976873537983,
 0.6885702778522698,0.02257998852819243,0.5017301567444977,
 0.32494922376835667,0.6974194828732893,0.3921812094011573,
 0.38885486777612643,0.7846701703864865,0.691782290000202,0.1551179901551879,
 0.3955961756770503,0.4870734987483113,0.7359719377220396,0.6351502402217573,
 0.6924492092036807,0.44627183834023454,0.4626094658236455,0.672233778319225,
 0.3438116892635583,0.936950471443979,0.6932453122954871,0.12281755277474637,
 0.4637147774124718,0.5478039673201417,0.6888375276879874,0.995620418670161,
 0.55442519302278,0.8306084686915264,0.21377827186563014,0.9123003258642639,
 0.6963595750616948,0.7550546087690861,0.6079823689748026,0.7938329382918925,
 0.09426264008380958,0.5883119030275477,0.12051269625040606,
 0.22021229600789238,0.07349664784649035,0.7055005432300882,
 0.08290740059631396,0.24829861862431735,0.3955228631034725,
 0.1785278406490035,0.5362036327784849,0.6408978650554359,0.6949528200538732,
 0.6106950789233841,0.2793176477057443,0.132438156102209,0.9193023343087023,
 0.4981181508692145,0.7536873235346401,0.5276021897115459,0.9989483192697425,
 0.010385364639827976,0.8248993565326631,0.26502418013085716,
 0.35163081234210014,0.9034948380102583,0.03933734663581956,
 0.9327628599295149,0.28592090651108415,0.45290779324793584,
 0.3859258725225764,0.4692370953495474,5.991447790533178e-4,
 0.32165842998160077,0.983548352657681,0.7676294445406128,0.2873797557292095,
 0.12460619086032176,0.4486347422370093,0.8472334283646545,
 0.07553366775652848,0.4316386535924299,0.3193450766838929,0.2498725428312557,
 0.00878559247702504,0.1461488974433398,0.32103606396130413,
 0.35245437876012375,0.2903746014379641,0.36849638292690623,
 0.5491099082034203,0.1285394256434098,0.672323132446474,0.2766691228689938,
 0.5984122699076807,0.759073671426286,0.15840109091321353,0.49901154458788466,
 0.696291546396022,0.6360137817309472,0.3304269718250299,0.1803545472818744,
 0.5364862558165892,0.7003534045168993,0.343651610580216,0.3277759450865516,
 0.42083534140316337,0.7738033599782894,0.47347600311867677,
 0.12326900730481749,0.09332746709879891,0.0021517718396517616,
 0.22209295207836377,0.11567712570065036,0.8328762384217634,0.304247695664309,
 0.19104902829585702,0.6683586738738446,0.9651870050074398,0.3125714313960466,
 0.34319720482590765,0.24350358210269707,0.14073720419875446,
 0.22277594439295811,0.4081597739497471,0.4976762137317806,
 0.32461871010833043,0.9728380927391963,0.2664967289792395,
 0.07863256518341688,0.4350406987426103,0.11820390325089192,
 0.9052643992219774,0.5570743434590791,0.2479232600589577,0.1773944443419989,
 0.9644855926952758,0.24220321291114733,0.8703115313476771,0.2720396825864897,
 0.253728786368569,0.05185697955463364,0.6662943760214951,0.7680054081330013,
 0.37585793067578144,0.6978644262381151,0.20265782097861718,
 0.8363827092158868,0.783128080853062,0.6214583768612839,0.5016566502529056,
 0.4562005064300346,0.5854814951030529,0.35574323222716453,0.8713523015061571,
 0.5246546671676243,0.8907146393029169,0.9188334528458576,0.4424227209883397,
 0.06738075185348824,0.5252083763012597,0.46313504349646495,
 0.9937895855598758,0.12695458951502014,0.1472735495035522,0.9630474452513273,
 0.6052905306420192,0.8722795706621751,0.910768359142789,0.6025420960414091,
 0.6381791129808422,0.6181759599038701,0.11543090299381265,0.280551132126148,
 0.05447542443853326,0.5314724755102921,0.20853116918222958,
 0.8044566037796584,0.14850371696626663,0.5550778706938913,0.8319946068430131,
 0.22070506378194044,0.5238454003695422,0.014424464963360695,
 0.659021301097686,0.43162939347220763,0.13788988402082392,0.7473762464943641,
 0.08091792383146734,0.6024723986764023,0.3003672717248048,0.3175266214054596,
 0.19755282757838,0.9458840985653576,0.22967191987995583,0.03740721610477771,
 0.638791317959072,0.6137763940296663,0.9430279707983953,0.42682098800689905,
 0.3234427396147711,0.7960503618555865,0.8533496510924544,0.5657736837580639,
 0.5960733496358899,0.8952062823515442,0.1779999073575832,0.6968286011436168,
 0.3512921939461794,0.5992454477668915,0.05548049084732509,0.9548177552830228,
 0.9905824409553713,0.6586745329709176,0.6552043067580021,0.564282345788336,
 0.9162419273680015,0.7105304794188585,0.3965888148839918,0.7134877943631351,
 0.9128490706314534,0.25609076746239035,0.21858627210063186,
 0.11311451693006147,0.45029851649123276,0.6218579540382042,0.916292235335999,
 0.6087290004970284,0.0138988276987847,0.12313910554466678,
 0.14931626242120521,0.6767294298728779,0.6819648151515731,
 0.47205882973464963,0.6093049537253811,0.8455337530173126,0.3538246435544371,
 0.8241388599988927,0.7922874493551821,0.11337722359512314,0.7504838171153131,
 0.18968631262849822,0.722087693264198,0.9095189742210837,0.17072900943887348,
 0.9328095595338696,0.2594911272616045,0.2873891283405048,0.1761541281531136,
 0.9819127142122624,0.9266750266603361,0.36988774894150966,
 0.29613881848642654,0.6495070580628833,0.2350758907053747,0.5938531856911142,
 0.19246869346398676,0.5091076293406337,0.4759553896464468,
 0.15702462858112698,0.16642211518319694,0.30198029263185544,
 0.6923615084874762,0.833355393637594,0.905391833298143,0.7051637254368591,
 0.7413449201205832,0.7190561879872874,0.9998626643394068,0.7053068781960018,
 0.09873959890711648,0.8036042250219053,0.27042209787727167,
 0.26729351301988324,0.01160703800739149,0.5780623419047453,
 0.8176845242215391,0.974642185027998,0.2613346007186179,0.5589787007445027,
 0.26727102735052066,0.25267180524067334,0.15237897898943575,
 0.9580852423969525,0.5310639744354735,0.8199601900508198,0.3959611481439309,
 0.7564714933537597,0.008826231125703576,0.621948460607248,0.7762578805888162,
 0.4469911555522996,0.1101167838070829,0.09091283062031674,0.2531819020875875,
 0.4943739153405635,0.9605906821952764,0.4397575997447444,0.7378031405655093,
 0.48163880470972753,0.4890221598293951,0.27686831786662025,
 0.12208833629064042,0.18176564014369234,0.5145335473671778,0.378977068364742,
 0.17774045397928961,0.6879994292335482,0.10944946328092964,0.961989291136565,
 0.4818806347578104,0.6618867211542045,0.5487417642578627,0.24402989715104728,
 0.4655017065207351,0.3993247778853428,0.8650858918014868,0.7482676278661122,
 0.22041714154919945,0.5606685194963099,0.4172069414500863,0.7220626897267097,
 0.09126276592572358,0.24914454199335112,0.21144493586935353,
 0.36445252569489917,0.5860725906361888,0.8505028271614785,0.2073085719136547,
 0.6440536129674183,0.22162657595340562,0.05063983164041452,
 0.3319344982344602,0.8632739290910161,0.07283060823626719,0.8180230762128764,
 0.5756125858611962,0.4743152056866988,0.1675042221127342,0.8885716470450765,
 0.6845175281043216,0.1606662820877116,0.07399801687494623,0.7705000161909616,
 0.17405550578505846,0.364327207410612,0.7050409153086281,0.5294454545557847,
 0.4784248231220466,0.28282986689856504,0.025559718001502363,
 0.6982885573948975,0.7175796995119899,0.7747180578976564,0.3060749065832059,
 0.8746406749137877,0.8478519790788646,0.2551472178071418,0.5937516687479167,
 0.6425404600124821,0.5108896758932991,0.6940187259552292,0.5458747852932981,
 0.5730674792076335,0.38125606641652676,0.13151350891938562,
 0.6138343149975262,0.28648281923712204,0.6204296895708745,0.6869248400638048,
 0.8609445809725214,0.8235793405431313,0.9462771400332922,0.1476647240414488,
 0.3885343836521974,0.8001981362920969,0.6699035599370351,0.335403973664929,
 0.6248906607652573,0.0027598098118123415,0.39800935305463325,
 0.09090310595135032,0.9783903498257769,0.859951960276817,0.9160395038802546,
 0.7820017067114153,0.25584821988891004,0.5667926372896757,
 0.31773795492199663,0.5971915317128862,0.49611870272232905,
 0.5819272205920997,0.35496928106818604,0.4963013357166792,
 0.20191105201251058,0.9826005652834003,0.16087358167865473,
 0.5993407129314797,0.38655493067494673,0.9931965495618453,0.9487287614074086,
 0.6475992455992444,0.40131295233943287,0.4025251334239365,0.6616093770730127,
 0.6120131552686643,0.9853962196615766,0.8328822108429479,0.24070437066071126,
 0.7504431352044033,0.34543562261530913,0.05422807933547669,
 0.8795356626715617,0.14003811546552958,0.6321367240522271,
 0.39999812938888324,0.5270783396746826,0.562865807838639,0.905492392918908,
 0.7049252577320203,0.7594920859207023,0.011996617367289297,
 0.9904229429731356,0.9408147538083044,0.13021160150749045,0.371855225968899,
 0.6001125621083606,0.04305188517047087,0.989325460127514,0.4698812580736449,
 0.20137120156213473,0.45785821982373065,0.49612072348894387,
 0.9014172008454835,0.481846844501606,0.41509288243220344,0.7967561844598972,
 0.7610862974784321,0.33729901585799804,0.6815808835471376,0.9287293913160917,
 0.43080570801559437,0.28497964929687913,0.3506570826685148,
 0.32924247323202327,0.13209552949429082,0.37564259419154866,
 0.37064422273587305,0.1463614983627598,0.6972249745402753,0.9188921231592981,
 0.3167044328747747,0.832298415605236,0.03128767545180122,0.9643064979563513,
 0.5281738453136957,0.8925242852430103,0.8301220111840708,0.7245565420127096,
 0.43773323735669967,0.7959370382692641,0.08159843396426714,
 0.46783800217671045,0.4185407109682604,0.21273314941318544,
 0.6745922953969944,0.4769894053149738,0.13600965347618832,0.7393103862945957,
 0.44122997296779876,0.6661191780512352,0.045574213800289254,
 0.3450890944753977,0.36902026372543517,0.6003909505227928,
 0.22610592408798014,0.10407391372959274,0.8532223468995984,
 0.6212845900504017,0.5239301028975234,0.5166021513472607,0.6290086891714323,
 0.8045668831987306,0.6275713672647829,0.307621121162727,0.3878946316442734,
 0.20274078489963987,0.5634289935962621,0.6058666098606208,0.5079654268042777,
 0.7349046730226408,0.701886736568173,0.35212130698122135,0.1814489305871111,
 0.3381852852843068,0.30634176462119145,0.5228564416571299,
 0.18529481536656145,0.9940995063852086,0.47194098184833366,
 0.7649993089366042,0.27589374867250027,0.8775416947134745,0.4040375767237545,
 0.20351032891100362,0.3153095648340629,0.5680406938791973,0.8890318075958507,
 0.07882875233461606,0.6020257358302677,0.29021920701236836,
 0.8253579292194215,0.22829760492315587,0.12161266891822486,
 0.9954763858270744,0.13614586680212026,0.5691268649035318,
 0.40991318755423445,0.30297952463151545,0.5696410056116696,
 0.5840312951712726,0.44140857283510737,0.2426710877471301,0.7207132590629616,
 0.6986576567329355,0.30426258395892325,0.5345352599991878,0.8693943494549489,
 0.13244710222123635,0.05292975321016591,0.6015341942967993,
 0.5231516170068359,0.6400124710372477,0.9922079615447291,0.17799151533562108,
 0.22732909993936834,0.6320152079449177,0.38585166299284346,
 0.3131347975390135,0.1351779035689582,0.877746519826998,0.11361905712335874,
 0.5280394334536394,0.861983532457002,0.016316570907184746,0.5004732246697798,
 0.7635473756382404,0.7876867149199193,0.050506789386594586,
 0.41100881465884764,0.5264088625291402,0.8538239921285413,
 0.24722524705797488,0.1556252111124634,0.8441544453512722,0.8013352371368833,
 0.06613791895103438,0.1180573190077645,0.39231389568957775,
 0.37454158477849486,0.10882267527445433,0.44044912060605057,
 0.07376175799661944,0.6239078997784453,0.26005437833114553,0.423074580640571,
 0.15218618305728449,0.3015754892836686,0.39520088069620274,
 0.5070382396318003,0.38035793415966046,0.36230460197863956,0.244558755817458,
 0.4072513582513062,0.0015012623777390655,0.6957597428888981,
 0.4493944240918263,0.8037391436266886,0.15334697933839925,
 0.0021567394732613376,0.8007139680895425,0.05367560535183136,
 0.4553211113854224,0.6840198120593086,0.8777051161867179,0.6489335231921106,
 0.49268676832290315,0.3447844137153898,0.7009595049303862,0.507658127855226,
 0.2763824125186861,0.9259458811742313,0.2563192647139412,0.07959901447869178,
 0.1444429626687601,0.9188138572896601,0.24475242138360098,
 0.44307318751876557,0.7795015640648634,0.6074639080663078,0.5724532038532475,
 0.8742698178520492,0.19561340776064284,0.8183481423554414,0.2601375319824547,
 0.7614810974364661,0.5112156961649442,0.8791595478826284,0.5217736762863863,
 0.9733002833297543,0.6964424938353894,0.8407402716768834,0.9659247151696455,
 0.84184575229355,0.43950541052594916,0.06051002881402501,0.19572764201714,
 0.7883628034857209,0.9007262199922372,0.18566393288424887,0.7795311427477304,
 0.29055769352152805,0.050181951950931114,0.5861858648144666,
 0.37752336889010507,0.8300451154043923,0.6163473346476712,0.737270272296336,
 0.28943010714741746,0.9816315251911948,0.4714739362816409,
 0.007259396211606095,0.2765949164946324,0.8901431225820498,
 0.9171315517339652,0.6820304067046579,0.9489885611174277,0.12658879026536096,
 0.30582761859118046,0.1592902775823969,0.5655206328133411,0.7973129994372951,
 0.5895380066100371,0.3105225479360404,0.9896547270142426,0.5849269682329492,
 0.793979483392893,0.0865379867829923,0.6119892291552665,0.40705289652603394,
 0.5044937589483602,0.19642347612381605,0.9677489733440131,0.9651923710466267,
 0.6763173447550165,0.7110686626306788,0.43003413694284953,0.1579647392961261,
 0.48971474422880434,0.954009582557791,0.6237003591323653,0.7190121379435721,
 0.48907523410373566,0.8192412484631993,0.8592186793694104,
 0.18887936456568577,0.4369601436126941,0.6102152900397961,
 0.36437827754761776,0.25933346964526005,0.26717244523792083,
 0.8142395436860279,0.826063136535097,0.8823774586128142,0.5985057957089149,
 0.8332334952829248,0.04927316766995726,0.12866482396598067,
 0.8553321775987477,0.5376732947488411,0.6881774440783117,0.13563865741423342,
 0.6636208090516219,0.1712106371069103,0.14512579923013935,0.6061473846511841,
 0.8694455776919582,0.1647350036485028,0.7162547772596001,0.8448944777777034,
 0.6626090665874826,0.9175392919471967,0.02745133184618387,
 0.008610785211878857,0.37404135091205004,0.7908631637133107,
 0.5656661149287745,0.8938740680200931,0.10091906077694857,0.9614325222169134,
 0.7138926680670321,0.37002264097438164,0.6617042863341688,
 0.09442009339541346,0.6943305893121031,0.36923668383075336,
 0.6699161333639071,0.7292978817451592,0.9500027848419722,0.2607363056100709,
 0.7813972744823684,0.06344304101331,0.8525126184452533,0.4350654651478101,
 0.5635328661983305,0.4942799911438197,0.7370737976360959,0.16160697827299397,
 0.6764585748790264,0.6062213053597371,0.4975785027329793,0.9530420393800644,
 0.9142978167486018,0.6298694588148962,0.4059093658807663,0.3493325680766536,
 0.8537409524888429,0.6780925620491769,0.8915908925289481,0.799738326080712,
 0.20793237340354853,0.7502329423048717,0.2363198643073392,
 0.06543595931573565,0.2511149702789608,0.8017129063474733,0.172493481114668,
 0.9887821850734471,0.5178669657308562,0.043152381667258544,
 0.20142056504112527,0.3286012127271051,0.23638868005141433,
 0.7182618120004135,0.289475128604168,0.5458300862935895,0.12902597759615875,
 0.8990464624867744,0.20859788088676234,0.3964854694748239,
 0.31999009286770663,0.7773068222053041,0.9274328518902111,0.8917872661013015,
 0.49042662815928484,0.47447039923908074,0.7746142903756645,
 0.6281325485200923,0.3553571738459278,0.48757574061480213,0.9121927838600667,
 0.44735658301946246,0.587681235676235,0.7194804156464512,0.4286301111496784,
 0.7523901706185963,0.5830691293640403,0.45466342047178554,0.6500396187624592,
 0.5278793493524535,0.634770114546546,0.501154361490792,0.19249485996668259,
 0.2330876631237896,0.3141235429175785,0.08038735543575604,
 0.028095422920857316,0.9672810708478714,0.3395868940521254,
 0.8998893520527388,0.7843051234901516,0.1718818778551131,0.17347051659468393,
 0.4976006898662616,0.9997005850786618,0.8892297399678059,
 0.017799538994755504,0.15513362199023328,0.8845768553829954,
 0.9327565773326894,0.027924113758507563,0.6742425823156107,0.507247449760881,
 0.5508655663936073,0.8710849686812153,0.547167006320546,0.18910602968630252,
 0.016486354417810434,0.18742257872934864,0.6235371393408873,
 0.8025850194033423,0.9210081493002407,0.2727800571150375,0.9685091274982673,
 0.051356981227842824,0.6687295320550641,0.09596399376846743,
 0.9433979455774656,0.8159485661134386,0.2873976207611796,0.14402600747654937,
 0.36446375401803843,0.21271811825751297,0.9401794953846718,
 0.6068486897877241,0.2592491784302392,0.19283364864943986,0.850781258750796,
 0.24754406719813105,0.1277408004360262,0.6876672902552418,0.4870160992281991,
 0.01272963536470173,0.3034349415742076,0.062077253170434,0.6352859379077549,
 0.5745854249089801,0.4585129662879366,0.89369201873447,0.19011488997594062,
 0.3182384301630585,0.5295234526247387,0.48915524048848624,0.7497449585427698,
 0.858204985997544,0.7068939530801934,0.3593681964306584,0.28450840440183767,
 0.6613710045049163,0.2553404634982479,0.25431269701728776,0.5005960378755256,
 0.9319553339239353,0.1833678727071033,0.6959253561917993,0.37853434537981046,
 0.19771354042211842,0.9233999498748745,0.0029198473830281957,
 0.7575358853098046,0.6341132029529615,0.11526299370413473,0.5838392069904002,
 0.8493622976812623,0.5464293683114229,0.616620931609017,0.33932907520249167,
 0.02174028359103064,0.14476677070636879,0.022847279198663806,
 0.19138559891163354,0.10176206554347589,0.7265519790427402,
 0.46808347297047925,0.1435505758525466,0.3509567716545574,0.5333514833272133,
 0.9242167820985455,0.6900503353842741,0.8214360235335751,0.45670379447484544,
 0.8999573571619495,0.7894548146677915,0.6593300303031964,0.8872526421102612,
 0.7326412641308869,0.8301955247983933,0.3488682057965745,0.09493978011903637,
 0.33660857034133007,0.6803618613185061,0.9600107958305135,0.6127787204209737,
 0.20469981629036937,0.11331762735201756,0.1873094128576911,
 0.7923097684500522,0.034499563501270814,0.1488411107238028,0.872199081421793,
 0.7871992396056601,0.5525287217380679,0.2821663157660831,0.8183017775297394,
 0.6843791125210844,0.5993291864695338,0.8727950503366745,0.6540599646527178,
 0.7651739144870309,0.2428352743674025,0.6090571906359065,0.9570285828650343,
 0.3673831390290744,0.38073644743645396,0.7322858520770544,
 0.19071594780234835,0.9940984063730901,0.19144442466669898,
 0.8371611780681857,0.5422771131368306,0.7039462063510726,0.02469302001235829,
 0.6651758862870346,0.27548481647460954,0.9755752846311843,
 0.21724522818348369,0.4516215904336077,0.8174557714117875,0.3936027424150441,
 0.10742939840348775,0.17655787795957645,0.9967224310593324,
 0.47604717379641737,0.725259149462433,0.9616163570446058,0.7873357118305625,
 0.7483033584658919,0.5131711164375905,0.943313439522705,0.33162990437346374,
 0.837242860798033,0.9425326152165282,0.6158110846715769,0.08420037448619555,
 0.25457310303030023,0.7943896615184788,0.666715695573782,0.5840790658913749,
 0.7034201874346505,0.8211172903347462,0.5080329589049,0.2514176432808043,
 0.3425052506352546,0.6031102576436325,0.5557061645543387,0.9887888410356825,
 0.6299142097060689,0.341662200216428,0.32726038987980866,0.6960433599509097,
 0.889848646272201,0.7196319034999865,0.2283601999553072,0.3000627853984885,
 0.7230288904943991,0.8502541573940161,0.7720555847341569,0.20546153254306254,
 0.16238286778202704,0.588613103013742,0.12100574093566863,
 0.49751876622891555,0.42319707247849336,0.47634960635512136,
 0.6149198788168262,0.6171605832672161,0.15604606198935156,0.8420706372573569,
 0.03148400136845764,0.39157857764557236,0.3304103567252801,
 0.04776300364223218,0.4719060783419652,0.9235089218542611,
 0.42671679669621865,0.8194860517399931,0.24933494201869288,
 0.2315554898990272,0.6064418902970531,0.721370508381806,0.03744976297797342,
 0.48343362953676494,0.020021947696020503,0.27210414964902785,
 0.8849449344777442,0.7786983188349811,0.6044905427924767,0.6128261632805508,
 0.871456361113371,0.9491859134315306,0.006648551407660297,0.6288277202708304,
 0.7789403177937798,0.1555677802189357,0.6159688942555442,0.15196253512220537,
 0.8735326565502582,0.5965838431064071,0.5548260419457482,0.7559045692020102,
 0.42429671006576164,0.8580531963199186,0.7491006021833515,
 0.19379452207798054,0.26711522842224056,0.5697661368034843,
 0.29429826738573694,0.9992773620834006,0.30518041431824494,0.394469914404711,
 0.3411676447407578,0.8931209000742075,0.12936537687173955,
 0.14388422565555614,0.8473715176153858,0.7489561929103253,0.7302445229002309,
 0.5946721869640962,0.8929942116890187,0.6841328537148109,0.5547344089595473,
 0.1835019563300414,0.01626109550028365,0.8764263920071753,0.7540203873021329,
 0.20405982266257539,0.47135881966813564,0.12490018890938537,
 0.7402127392964741,0.7129835408007936,0.4730737910260391,0.40045181182365464,
 0.7140649727046632,0.031815385191979684,0.387808451815032,0.5988617329130818,
 0.8453901223751691,0.35548008786191776,0.31905544661085383,
 0.08146660192993149,0.7012761094694467,0.07694608857357599,
 0.4186904766379851,0.5085812937832663,0.3041796803672594,0.1444991997832905,
 0.9955769102465202,0.21778400819155697,0.21354161547164774,
 0.6421442658013448,0.6440293996609929,0.7306401761006159,0.872007630711908,
 0.03399594144513418,0.9131838988242158,0.36685496735539735,
 0.27262718493918503,0.5368634684736118,0.633153055400465,0.352090804316321,
 0.3447461120449702,0.6300821141259406,0.7250962068589087,0.722950014379334,
 0.5939777444984989,0.401474422982123,0.9650742243594659,0.4234788667577051,
 0.1645459626166621,0.27531426015553784,0.2648758243794652,0.8925904007909922,
 0.9512985953436695,0.4461523692896119,0.6852095066516282,0.348798406147958,
 0.8095107545323694,0.26256564067834176,0.6657218713128997,0.1654008061018668,
 0.5271854232557288,0.4201129581177111,0.6821024371503945,0.5528446052985441,
 0.5401119003851867,0.15966320290183178,0.3885156352467126,0.8422622864280215,
 0.426350771117351,0.8060298007830349,0.3789323100666452,0.9606471057055191,
 0.03495691557203162,0.6485671175340645,0.9138858006994999,0.8416275409795125,
 0.3750840956552778,0.5061809698305848,0.43313944177118113,0.5151530203054602,
 0.032845302809082155,0.4825723513165098,0.4256720444523917,
 0.05646980848867389,0.6362753818074108,0.7317613297158974,
 0.08085496416951321,0.6681347813750739,0.0377548791174227,
 0.15157295717091168,0.7691114807868373,0.5412808842205075,0.5568167431059506,
 0.7383360372512034,0.1756480283840567,0.7432932840065729,0.9109888565584967,
 0.1463856065306096,0.5498921997920604,0.6660704376043662,0.3484038408995497,
 0.1286778110403598,0.8039007365436535,0.328851233129855,0.290011614111538,
 0.5744348186496618,0.23508251293596916,0.614584061910756,0.9527126206586195,
 0.7177579648557367,0.9980448850076835,0.6121233991619919,0.8652948185031158,
 0.4917678455168454,0.6350223940040474,0.950567548018723,0.21472696340540487,
 0.5789313123481594,0.5187267799758486,0.14171555699724148,0.8723851759364231,
 0.32268488736009526,0.02840861000160433,0.14068976237652986,
 0.7155425962040745,0.8759805224334525,0.713131842475454,0.03451100448242772,
 0.07296582967906351,0.7104761813669296,0.804615609686165,0.4811182806866775,
 0.05541574834554186,0.23169922612990346,0.4355857010655282,0.652387264592513,
 0.6713947245436176,0.8081233390224748,0.1264636924406437,0.9242860351315021,
 0.8303321457023396,0.7700361960928289,0.6290741028896294,0.4187615443422298,
 0.7700034370710775,0.4052444473974264,0.7740453258210358,0.9928706919996486,
 0.1554533453985778,0.9079828447660033,0.964272778679684,0.34238941524148514,
 0.8601958102883386,0.9099735512446111,0.11281115313052881,
 0.34159153157028865,0.309527044388607,0.31907608548393007,0.3507355755455187,
 0.9991827292325417,0.37860505451599036,0.3683482999942781,0.6489464026766057,
 0.16277002853069467,0.34642230113228023,0.988308303774805,
 0.02471632726677253,0.6058787063853789,0.3115691774695848,0.8058689579267867,
 0.32288264785748255,0.5770861087766884,0.97479873058867,0.010786120252428644,
 0.48208484596485024,0.3306428126518356,0.2762841549375685,0.4303579957772856,
 0.10115802233831184,0.6492741883578566,0.12388037779033623,
 0.13195096899977343,0.05852505208734338,0.7787571635284429,
 0.025194974028144745,0.3052920179984585,0.26889679894412677,
 0.7072745530699596,0.41565887083943953,0.23853682426717104,
 0.7152532544568119,0.42768650820117493,0.045328273127736796,
 0.6899715807543546,0.22469774516238772,0.7840715313084051,0.554226233313205,
 0.9462541754793073,0.6541485616010898,0.7272747121597465,0.5892432330842557,
 0.5939378572007586,0.14820903109315375,0.9149101455351625,0.8096634331124359,
 0.9507902423153363,0.4415961890717671,0.17999752136012492,
 0.22860358029352013,0.8266317477477018,0.19154043797725795,
 0.09605828861380405,0.11444371484763327,0.5079354100799334,
 0.23675039118311214,0.32255040333549834,0.8849245614887499,
 0.06518808437547796,0.9371013168791488,0.7417724042751113,
 0.26780413035607553,0.7455671774556144,0.5132408108337753,
 0.14681145089898073,0.5498922969458806,0.21302669240857885,
 0.6079292896289894,0.2888796595343318,0.9277184369278436,0.38307833440364236,
 0.7642713565832653,0.07693817117936375,0.20991463054824655,
 0.6000534758168881,0.6204562895868612,0.5372396986592871,0.2838014594320697,
 0.8385582022008509,0.6267923167275534,0.7289952283727927,0.7159239329454663,
 0.07562847768469927,0.6072696589095872,0.8351280342386482,
 0.35020594059529697,0.10096253364763408,0.8456450044485767,
 0.05575783823986802,0.08173485930625579,0.1636211492455688,
 0.42557015106492235,0.2577945380875608,0.7704756233385475,0.15384951041577,
 0.42330465060250844,0.47246003990002516,0.27871519615560214,
 0.8067862994071832,0.45666926899890137,0.7964729063045114,0.7669399001444255,
 0.2558934760969739,0.9496233667018585,0.39324252155308936,0.2859949785761525,
 0.38523510575781994,0.05612316051875932,0.7126039383841757,
 0.5046922848982345,0.9371554428168825,0.8071718764579221,0.9306539459380444,
 0.18459885096306206,0.8306155061014958,0.6254848841533254,0.4241684403026379,
 0.1923391295794029,0.18784381192584687,0.2515444402905396,0.708044639126356,
 0.2416980794716127,0.2430755712097431,0.2324680065695075,0.5596868695489001,
 0.23181041638947697,0.7950213716155842,0.8469310239062908,0.8037469510496839,
 0.5496890397782218,0.22052135772810255,0.5302005727816493,0.9242734081830698,
 0.16539876841312195,0.65058523053953,0.5935088107257325,0.045113680709103265,
 0.15861396299936004,0.29144232453700414,0.9629557697014722,
 0.09118755456750216,0.4562082101664965,0.6321643294893488,0.6878345804179732,
 0.8261511889276214,0.7198979592010937,0.17221245335411728,0.4578135303879737,
 0.9281733193268817,0.9760276030809748,0.7662829058717757,0.9076857229121682,
 0.684239425718267,0.4621826832097038,0.5211897645327809,0.3001397838377028,
 0.08070621459761762,0.741851461718344,0.08172352279649975,
 0.08887888288311308,0.6253269704117472,0.07018828485156314,
 0.7886777608864116,0.6570365769391733,0.6726477277438387,0.20372720832457492,
 0.7826904604120593,0.5589769148255632,0.1683343107728782,0.1984212083422654,
 0.22830657832418688,0.46341964940112834,0.48944670585713,0.4158461488016265,
 0.3255439083579198,0.8048446821466394,0.26865136128681866,0.5914034642806676,
 0.30348816431086545,0.7396197571648062,0.3363951122654757,0.6325484628577938,
 0.478810705368658,0.17970744636564528,0.13363267785314625,
 0.31576273541019007,0.028114661144602926,0.2402132621259988,
 0.30139381025686296,0.7843018769521058,0.8220496265557657,0.6412488443464359,
 0.3501466757055882,0.8392619487813646,0.4379961567245655,0.2931554323012556,
 0.07723069054883325,0.5428535123196837,0.9524209149008234,0.031485958831725,
 0.7479294092784008,0.0927527002168429,0.5680534173103043,0.2388199387996035,
 0.4962327318028995,0.7287244249257382,0.7446038310463841,
 0.059346438406243096,0.39965774310670166,0.34864172347587596,
 0.22651243535844556,0.9256095562674708,0.8925337396134183,0.5401284853780839,
 0.32348621854665316,0.9818221679981781,0.5918892283591896,
 0.07849431141956387,0.522009562670646,0.9049274673198195,0.9620435855652443,
 0.47710326432629235,0.8271527954252089,0.27432462765856824,0.642062007330538,
 0.006462937003072988,0.8039457273233271,0.12145145088624576,
 0.5354895424910856,0.16182505617481868,0.09217033105505623,
 0.4252641821558353,0.6672325938773228,0.3551708352271241,0.3287345103763394,
 0.685206097967093,0.020053774476364694,0.5044472977911825,0.7780218606945233,
 0.9646362611145738,0.21746915127999922,0.6120911072622215,
 0.015832304807010922,0.9787126758161337,0.15934250508503733,
 0.9679982726440831,0.3462640807172428,0.6632950331820803,0.17105240483340123,
 0.4606994224213825,0.7178520635593805,0.7722173347996649,0.297469481224809,
 0.6839711707791121,0.8427855624690247,0.3659164743423412,0.6430220438489616,
 0.41112740000672154,0.011271161259552853,0.6862825425094171,
 0.13226214782703782,0.9750227466291682,0.34181776335568437,
 0.6985276163459533,0.920838708035929,0.5939547707827161,0.2918495517394042,
 0.3354815876771381,0.8269661789646549,0.6666696658452403,0.7962555744248175,
 0.12511218463516982,0.8804796325468157,0.14694757173245354,
 0.4228478175091873,0.031121093990848725,0.010266410478427135,
 0.6145171885559845,0.9725682808099887,0.9977346376736215,0.21638566936480208,
 0.5472585358941918,0.5859709172398531,0.084310419372716,0.5966122793597347,
 0.07852779238006324,0.06337510494292153,0.7578759015440315,
 0.1271013223163927,0.8989347557783215,0.9378964270032479,0.10558686309694243,
 0.5622910509304335,0.14380258248453182,0.32284452185864043,
 0.05569695342931913,0.22674933462008906,0.4281420810467843,
 0.7665901228379752,0.3412092116995622,0.5471944929889271,0.08046697119454715,
 0.44104883176180587,0.3537991854738165,0.10447485697671766,
 0.34125459983471695,0.18116617735694296,0.564944172379136,0.1490100665973596,
 0.7492536028792123,0.615994466865997,0.07313430420459666,0.8923521098607822,
 0.08814784679093446,0.022616844082833487,0.2367399290091644,
 0.20883660974451002,0.21060860916285884,0.6709386346380578,
 0.4780758055128017,0.6615993464207947,0.027680799461055727,0.347917849680195,
 0.2792228241255278,0.05476447358885006,0.4968466293200018,
 0.38114761642524786,0.3996677539815455,0.9212673714623951,0.7040155845235523,
 0.9509125340116077,0.41152457536912246,0.6450536371536079,0.2853091881198302,
 0.9732316920395779,0.2756614127743765,0.35215401260826273,0.1004479921325272,
 0.9859431466990406,0.524407931351917,0.2770805869391272,0.1734508958864307,
 0.37592336634200874,0.1891874764571414,0.9151017467650191,0.6517377683752091,
 0.02271215446988062,0.5585487374877668,0.9618896970898053,
 0.48382342474875306,0.004891336861489348,0.9304608943463044,
 0.6335646674213392,0.6385270280390677,0.7326125557150753,0.5833896308941363,
 0.8045475446110315,0.20191405506931148,0.9806993204834008,0.3444962582233899,
 0.9443667823063892,0.9433981191781143,0.14699981161783515,
 0.38936283607099986,0.5157532072548991,0.1740196243955071,
 0.11114063565620436,0.7907298162292118,0.3027116248305697,0.9488617960387217,
 0.6086931815357718,0.6570696788437451,0.6476754151472061,0.36497272516174917,
 0.5693180267478808,0.34615124712743583,0.6405948226775309,0.5809665394893107,
 0.3496325172300515,0.5287631611670621,0.14362094146373883,0.6000403865396369,
 0.19537021229942875,0.9120252923915231,0.2824325837386171,
 0.07965930478821648,0.7985033786612041,0.2914475139486461,0.4302977143550877,
 0.8710019525420796,0.5943389675201587,0.993371551284607,0.6689845761578298,
 0.09401392027865985,0.31384789549786973,0.47577958793034525,
 0.7946668731610477,0.29979042708927106,0.5684140557068071,0.941932810925222,
 0.5764356574345368,0.9970381963205,0.8604313858384228,0.7733555693584189,
 0.9282734164445033,0.4613096421838543,0.5531433706898169,0.20975444192145765,
 0.9836355343912766,0.39219279107004845,0.3558689242930081,0.6689182966626578,
 0.8915080651546115,0.9760950957931287,0.3300592543253046,0.9374634909417694,
 0.8070442675334539,0.6592594261898583,0.011015249966158436,
 0.2964201604534382,0.961071984003057,0.44811227326108105,0.5304811267500724,
 0.40351389412849314,0.04693179240099232,0.8131131996999851,
 0.3677765796281858,0.44579430822687305,0.04011157552022049,
 0.7688670534698576,0.1547666455711454,0.7315245340704123,0.5967320028501284,
 0.611224619343064,0.4726826951836407,0.4399950807634022,0.9649901564088912,
 0.09155446971609915,0.8393220252015006,0.967295137774166,0.4675469701414672,
 0.21555929544703367,0.5405544883350135,0.003953171289676316,
 0.8169790547611092,0.906044077614134,0.6603744660358467,0.478954376202662,
 0.47628530008738135,0.4591862167326588,0.126924103944832,0.722030805358076,
 0.5945953907947598,0.2387634195014402,0.5383733903678996,0.4836483177127332,
 0.9169334637423621,0.8664699497740578,0.7351286870536966,0.09052368763275231,
 0.35924739472450495,0.46840674449053665,0.09122731707357534,
 0.5493614198028848,0.29585339869269267,0.06755418736751051,
 0.4307137395604894,0.7677165180705194,0.5736493615554249,0.48697374356450773,
 0.6644909792946547,0.5062641227326713,0.015080826286872795,
 0.48788069340417806,0.39504338331544053,0.576718174152297,0.9280141573404608,
 0.09680585585020818,0.738113900512633,0.547527803038943,0.1057636693941324,
 0.24216368432842061,0.27640088552053277,0.7378160160836142,
 0.4742179748973152,0.3721589528430923,0.3275713541781352,0.917865482711516,
 0.996739845189027,0.5125227015013842,0.0455879230473446,0.299312421618961,
 0.03404091674842624,0.3225482939069886,0.5961228041335326,0.5466704683006595,
 0.015579306766860701,0.5151256439961805,0.6368547772622797,
 0.5513864979497416,0.12881161436887067,0.6335087205416045,0.8893235785104957,
 0.7519394169928976,0.9736727983243517,0.5999686456773367,0.8136737380357366,
 0.42014853962539445,0.37783357413731367,0.5335777007594286,
 0.3009461228257926,0.0320936302454089,0.28119495344084633,0.6974477339295649,
 0.3448646872431278,0.7586072677249225,0.7834234147284004,0.32045757221776927,
 0.4743061552474347,0.7161979474312943,0.4312534042672944,0.45809665689628654,
 0.23310014795989242,0.5064220028182548,0.7738578348802124,0.823260642014091,
 0.0067736307563990295,0.5224638584553695,0.4807732686731363,
 0.4810659262672917,0.6037461373235358,0.8706918025244297,0.13925024022678312,
 0.17016768434137497,0.8900475966193369,0.2573595672962543,0.715122645634859,
 0.06454195939622498,0.5868673652210143,0.335495643732762,0.44274755100381047,
 0.7586994657221701,0.45849095002261775,0.8303417422162678,
 0.06026656781188233,0.11730421910529887,0.6523733655686771,
 0.19768907753770448,0.9105846336426293,0.8948850159599623,
 0.007997602056215691,0.36825879247804916,0.9424918662418863,
 0.961103195422636,0.5014409839438857,0.47986164550932,0.01579358456807256,
 0.11878547701523579,0.8492886694191552,0.0854396581983845,0.9728405593497189,
 0.18694770815176331,0.38900601929272083,0.12303983299804389,
 0.9967000471162389,0.6321334548907798,0.8976086372167863,0.5123299175240512,
 0.1212511094275831,0.42186576729862113,0.018369189749079196,
 0.15328511845502168,0.1603226542849442,0.09243692516918667,
 0.3102899365816114,0.8341447553708647,0.870066310534262,0.7881763936820966,
 0.9684202697193149,0.09161660364484936,0.8123478843084503,
 0.12982986878962566,0.6746193988859961,0.5194271025545752,0.7339931930450402,
 0.9944540128662097,0.4889774184827582,0.22735342215421073,0.9485186279378904,
 0.4113879723025815,0.5796765763753707,0.1929897253703753,0.06494505735970046,
 0.21818688223503735,0.506134705332732,0.9667023636587265,0.13469332706187265,
 0.43579552117090015,0.6601344298504097,0.46150231871857206,
 0.3281709971818525,0.009604648522280712,0.7961066988633814,
 0.024692996220142316,0.0019692164083719987,0.18245118473426147,
 0.8908080719231064,0.49193756909495656,0.5588377008872003,
 0.11857170601990819,0.0012145413269066818,0.8449165656418315,
 0.2120323121250608,0.1649699752532413,0.40600120928062666,
 0.28496442454396087,0.7479356160599435,0.9922720828096749,
 0.14648047854182578,0.9194920005798265,0.9825170726715329,
 0.15841223117603342,0.7934743830401153,0.04936795781454717,
 0.1557093400225601,0.27297637916365847,0.15790117803268067,
 0.5281290599179318,0.5599716880848425,0.966088525038972,0.5336286209805491,
 0.8582482840649344,0.6012678438327548,0.33289678995640815,0.4464139301389851,
 0.03787219605886982,0.7074044316875145,0.3651071146497463,0.5928748823363164,
 0.7412086493752481,0.1553178345500852,0.3274036270340088,0.3396640581687649,
 0.301524750310009,0.4961227971377712,0.18270292576156055,0.06860309164376399,
 0.9151077490122742,0.2744183562231398,0.5529045179087488,0.20904173187075703,
 0.7072609950574176,0.847986641388382,0.6846104207605408,0.1274285197700069,
 0.7497573799034034,0.9864452014823815,0.16275991153307112,0.9476855090657273,
 0.6301179394998724,0.5224067498328452,0.24030564932316012,0.7846119460557935,
 0.08871537116133554,0.3633049778479698,0.23748993802772855,
 0.7949634681519407,0.5990588726261543,0.14041265926807134,0.715530639378126,
 0.38383587123990237,0.16647213464372534,0.6093582820180935,
 0.015977127519871015,0.7605979638090853,0.09818921201979136,
 0.4905966626750725,0.6556405205627659,0.19567898063985978,0.9793244571534994,
 0.23648894494422257,0.02005143737364623,0.9207613903758944,
 0.12152775447576825,0.44388440504823623,0.22745743093190363,
 0.08885847580206241,0.30025763757140456,0.9268706004863063,
 0.26154827858829643,0.22958770724793776,0.006899738539747902,
 0.16458079857508023,0.7279479271706848,0.28772419847054664,
 0.3637647913430867,0.18930827778886783,0.17716248417035152,
 0.026720200706126684,0.4739276358535228,0.7990393039465319,
 0.8523500869299347,0.9028568154516515,0.425320246294009,0.3336433618350301,
 0.5657375907974407,0.16170042439491605,0.8565282178744026,0.305240434489106,
 0.48453754219416245,0.33879964395223705,0.017058257999445603,
 0.9084932635518956,0.9452521828092676,0.8836091083038402,0.8923242409376068,
 0.2645079926951458,0.6007697881814693,0.1859088396871269,0.7976909187730513,
 0.23588934831161712,0.03659069309426233,0.10922684595143284,
 0.7325713621925388,2.1969833964208263e-4,0.30155415299793176,
 0.7071435079059716,0.4896600984326499,0.9998428798448658,0.30671806737551655,
 0.2860215929713734,0.06882156379007376,0.6638177334135362,0.585811547390393,
 0.515895187544707,0.9258717782431847,0.8466685610514454,0.21966096162109017,
 0.1339124581115242,0.3471741177751396,0.6655832373931557,0.29391909933019444,
 0.4167010329116795,0.1204396521968073,0.6047469461263711,0.17985024005620232,
 0.9337307736900893,0.4334088394049772,0.22346877103943996,
 0.43770702803051975,0.9164661460224068,0.8176982549227363,
 0.32408601725250463,0.45285031549689514,0.8210828058438115,
 0.8711358943214306,0.8940637264942095,0.5537851267215268,0.0723274519010515,
 0.35134737807866934,0.9002831939148588,0.29042370723128064,
 0.18910309626830601,0.4790090398690672,0.4358672012136109,0.6678674430026739,
 0.06651713587686137,0.029915942660950323,0.436830887097993,0.390613312714983,
 0.21208731068392161,0.7858497933505956,0.42340997361161614,
 0.7742743948974569,0.29339283360835855,0.8601631228001663,
 0.05650927162773933,0.35293558373842115,0.680732566593027,0.9670609903407218,
 0.1854906623646294,0.9885663682512438,0.034839726532388204,0.359963271042413,
 0.9199434347297881,0.14091117570989198,0.12819842511146573,
 0.8942453800129725,0.35109269370112406,0.7961276710305952,0.7117642698007695,
 0.35361917803735843,0.2507528208818589,0.10847783454792226,
 0.24250990946850393,0.6723851914929225,0.9932931673944709,
 0.21248579372114884,0.4377385104434842,0.6769448823820563,
 0.41811510180664235,0.7474290226588864,0.7971796763942973,
 0.10897040003331293,0.008110062258269046,0.8957729296480209,
 0.07377878193186116,0.5663926565791865,0.06179135210616793,
 0.04738527722579622,0.4088749661888409,0.6807770613492329,
 0.36480743485265565,0.25870378180253983,0.22596762030546724,
 0.1808975469158698,0.7686474064355164,0.8219327598993859,0.4885173752245924,
 0.07567113318490837,0.6572906011591756,0.16018154182011768,
 0.7241016715700113,0.24264402229632975,0.0013337143036292742,
 0.1849015337554668,0.1341957815764674,0.17930792011264762,0.2333212323613283,
 0.1996200115266944,0.10253271552031507,0.7599117355666386,
 0.02433900481331519,0.18424448116612735,0.6241056272008644,
 0.8330252571275794,0.6457575494016078,0.5557111320652164,0.2827412766796322,
 0.5593595724507269,0.49549959777341746,0.04149682352014383,
 0.7184168684101017,0.9877400033260114,0.8197392840365169,0.0725718711156218,
 0.5354933363517294,0.9858600252191707,0.05949321211906977,0.596042413857271,
 0.1455259199908584,0.18786978420153178,0.6439844708389334,
 0.10732855223145665,0.6538288469077241,0.1237969285280035,0.5680029843898864,
 0.09572601016533866,0.6972449565690145,0.6169617934245709,0.5009936060392766,
 0.9520385760666299,0.670267903080917,0.29741008767163,0.2277489088807737,
 0.5082304236651123,0.38023930851559196,0.03744779861065217,
 0.18746949048366346,0.11791052819819337,0.3660321072443039,
 0.8696303799418242,0.3987695225952814,0.645077358274551,0.8798772441092693,
 0.3432958754852271,0.1565430662093794,0.9914047353111904,0.9249988641237292,
 0.29664742112376785,0.508094241197327,0.47057146878554057,0.3701216757138255,
 0.021967333600740213,0.7311635078725011,0.003122468156072866,
 0.7310671008105594,0.5216368682185607,0.2080776664269668,0.266688522556628,
 0.7073881680974012,0.9996012395521449,0.6448479613382597,0.4365532220739837,
 0.5153464560249211,0.7079029891799868,0.3695570240381503,0.5827638621086015,
 0.8140779358094182,0.11510960233467726,0.7026298898768527,0.45077613821814,
 0.9287578527994031,0.3749483628466781,0.8576915450907099,0.4830438310909966,
 0.1136780807688399,0.6741067365518804,0.39622120105636505,0.5304965410196203,
 0.8687707118904355,0.6773876148950468,0.533925483334103,0.960313885378804,
 0.733003764887811,0.9608065563152604,0.6416494437854968,0.4671735155350806,
 0.39984950214863735,0.3149384243098309,0.8716359024661763,0.9774508237771788,
 0.7875279831311793,0.05897239059403303,0.37318339876970097,
 0.2222861350255413,0.12412140523627635,0.06497152477971857,
 0.022865927653251794,0.15438735418229432,0.49489553950815035,
 0.1019689045166936,0.6924096900684175,0.1889769442818343,0.484302350762987,
 0.015998308899493385,0.46406731991052363,0.4135865770677638,
 0.4467151988750676,0.822428824008152,0.7173920624789987,0.5037726164618919,
 0.8278668883544101,0.9385997096535611,0.9155551943110489,0.34776072714545203,
 0.3176594466426078,0.0817555364206447,0.5030898047051965,0.7615641042711531,
 0.29290829610559344,0.25890067000667094,0.07220586542456997,
 0.3553211260969813,0.3443503186596305,0.17696926874616048,0.1343021778217638,
 0.6238863685615826,0.45583937405462516,0.8211356315341078,
 0.026089993847269732,0.7483578196773398,0.2962460737608046,
 0.7814408839819885,0.4328712106284429,0.18799621679478395,0.4687718196085414,
 0.09158536714511656,0.0010765011581159722,0.43389421438361775,
 0.3431302903322928,0.017014332390995257,0.555320287449005,0.2229794064681221,
 0.8431947921057121,0.43589254602957705,0.9737714503877221,0.1723462223272445,
 0.6478417296747305,0.2402474294046717,0.9064438484450563,0.47281449392397745,
 0.5352318124290893,0.7923413550361181,0.8056822279727874,0.09251256350304271,
 0.75564010024024,0.3584927658952872,0.8461984288166268,0.7218988803461535,
 0.30343833041720647,0.26660938210591734,0.2647563604978067,
 0.8943480772304343,0.39229455276368697,0.29983565709397064,
 0.05454198201352112,0.14006558658172574,0.6168595687314935,
 0.37272106462688237,0.6640177375000719,0.20762953308848575,
 0.020736258217787906,0.28523449854658145,0.18031405095737563,
 0.006544813009426997,0.13137874929237525,0.7442474353168369,
 0.28472817882846047,0.7941847273453045,0.42367662848815046,
 0.13540354589219483,0.7199241141559645,0.6705677692597742,
 0.04146572631610612,0.7777852628899875,0.17802520669622113,
 0.7406485725236626,0.2402388840814932,0.010643023938572194,
 0.6740792377339307,0.5128593729408221,0.11164226058183147,0.8563549494571006,
 0.7607930820098782,0.7560082890466944,0.2928447998986252,0.8506752038641024,
 0.9453409912741461,0.5951459314905577,0.14929514308502367,0.7402063830490297,
 0.32550913182404717,0.6103043905308331,0.5891107127185463,
 0.0011473668321189123,0.7801427703429407,0.7332230134228114,
 0.9038438347494825,0.4557794582239555,0.6185092934677769,0.11617180586987241,
 0.1334457745585388,0.4812173173849803,0.6783736068579841,0.4516133263195812,
 0.7896275621122717,0.6689211572679381,0.9265009331117875,0.732010608370156,
 0.698916888549454,0.6078039631010208,0.46458437190325874,0.36655435085016574,
 0.3446386438801424,0.4681863074252949,0.9447469038346208,0.6488769403096462,
 0.3609963239278089,0.8069902830482356,0.3912601619518322,0.08748143170699174,
 0.705298581395205,0.3259235118480416,0.8342589968374609,0.24661244040199137,
 0.9831462328788984,0.46665834395705985,0.22868738770520558,
 0.00973175412312044,0.9145856878736226,0.8149018704039559,0.8427139997315578,
 0.7530406373188324,0.015632287740684925,0.16783481983311077,
 0.9648893298497504,0.3351486735625444,0.4638912228224852,0.6191402430781888,
 0.8426400172749832,0.49062327437507425,0.9188650623229296,0.79937159546423,
 0.31035414444621967,0.17156830843926218,0.4948098913020257,
 0.11154505175985197,0.8999592625000455,0.5154070654727945,0.9056057014391999,
 0.05362127742826628,0.8401950817720865,0.28011459800233485,
 0.8824227365120946,0.785040864828985,0.26688481803759323,0.4848827455670399,
 0.3313437739499534,0.6793103837900974,0.5214759973229064,0.5609275808930989,
 0.7348362240528694,0.4472689344507108,0.7047993579844146,0.6726568455864602,
 0.9154744343908902,0.9891692729864204,0.6011860932718396,0.6258355456878233,
 0.28221935331545434,0.40378517402969183,0.7088468251708978,
 0.8637987638846525,0.20489705516316004,0.6869349401737376,
 0.09578171838951588,0.9177310575380083,0.25000338042644554,
 0.03554544423510164,0.14402205864478024,0.9789536883042929,
 0.07513699556059572,0.7828275478023468,0.28541233355039275,0.689709975627007,
 0.6121344344562507,0.8522127913663392,0.41595031772769386,0.7982759056388762,
 0.7076190108875667,0.8069514560702562,0.4386378062544931,0.3984614623391315,
 0.7530012636377614,0.9121635763355073,0.2175560223213553,0.9715272215583978,
 0.8890113107473837,0.9461410934122068,0.4991897319078251,0.7329127883328007,
 0.11373025434668083,0.7223257121827972,0.6399634503710021,0.6473782331287325,
 0.18414026379781334,0.07972675434344889,0.8378461991108375,
 0.07358443935096737,0.6988048218555548,0.6684295017666064,
 0.46799102492469147,0.20792038547447556,0.24626353550897928,
 0.554691248846721,0.9590451738115683,0.7426437937351678,0.36918743446684466,
 0.6611238829346642,0.7943332060874283,0.38012269973927215,
 0.48755928447363994,0.63745896110747,0.9377386809860306,0.4214441409600983,
 0.44991917978965745,0.5390687439344483,0.572488646373694,0.6795922719339705,
 0.4488742813615698,0.2774166943943007,0.5727844141051941,0.45049560246151743,
 0.86324699997313,0.5177801637187639,0.2792636686777502,0.9436703321440724,
 0.7828000721122426,0.2641182415817017,0.054917845260236464,
 0.6912219429470597,0.47973576822386055,0.5908874621745522,0.7420638412885223,
 0.5703972490285468,0.26892296168695995,0.3784311420195161,
 0.17797624786197086,0.6640006149009631,0.610256199660099,0.30180597788696994,
 0.40883845877634306,0.29584151449215657,0.8104591550530288,
 0.8712259245240244,0.8589115627127019,0.6487430991855351,0.13403264642887458,
 0.7992877883254618,0.05720682176736924,0.23966969681983508,
 0.4139307873306002,0.9196590036123113,0.6888127765380295,0.6968690627585971,
 0.2781487901464772,0.15340641213259854,0.7547323749008862,0.7911142992201303,
 0.7271905245535416,0.8209521430640858,0.2435670860259299,0.2870119633778,
 0.19501738691262294,0.2871077547647193,0.3093032934327894,
 0.17446425982332237,0.1182850012024933,0.19786892297714664,
 0.16628443241108548,0.9901593715148181,0.4016598618785061,0.6950620611655309,
 0.7269646248126594,0.48118698421609485,0.39122778971369465,
 0.6026983990697939,0.9998093740074216,0.028383464485012166,
 0.15202129993432834,0.6769800552097786,0.4775178198966109,
 0.45871850963820393,0.4217012570592805,0.6210023038355656,
 0.10770335103366824,0.14086715062466526,0.044961964087241846,
 0.5651532517292622,0.9737131363276081,0.9394896953656456,0.0979327268637018,
 0.1467021990203924,0.1127980324383554,0.7696812549177425,0.03914098840482083,
 0.19571051906418513,0.4705137690278164,0.4954996583478083,
 0.10913412715714121,0.672375184016218,0.2941826809585788,0.644074693603223,
 0.3137207865830516,0.375983341729763,0.15316020941668795,0.9075974116656738,
 0.37834087221895984,0.2594527366387638,0.0700188554288379,
 0.040392127868622896,0.18352728780601757,0.7564691171588381,
 0.34756449263880906,0.07087625059619707,0.44087386563775455,
 0.8825505319382966,0.36762887561674684,0.48817147649717985,
 0.7131801503749647,0.7450122210772891,0.8551609772582784,0.7396623811615998,
 0.2154487135285693,0.7310614900226022,0.13300347259483625,0.4859348946287265,
 0.7322197612995988,0.22726665979944538,0.08873252210051141,
 0.9670159210492529,0.057462078278165296,0.10399586854789689,
 0.062424346317915824,0.6373944763516837,0.46681556012293257,
 0.060268145830782105,0.6636534155507614,0.29316456446777905,
 0.9322212795203796,0.3625522900516098,0.559695420013164,0.6222384323789008,
 0.3129965950949589,0.07734440308682378,0.7876262789097924,0.8637000412052925,
 0.7841697184373084,0.3399436666859694,0.8430123399513604,0.583889406846213,
 0.4677414033578915,0.35876105654199886,0.9667000805589554,0.8546898440567158,
 0.9065813924656408,0.15399823566814264,0.13721449201017488,
 0.019271599381683324,0.6640972220522106,0.6607980151088403,
 0.004049370575412992,0.13260896470788042,0.19728954950580424,
 0.28443039830314687,0.7485756313192452,0.5560121895863029,
 0.35031120834989427,0.9428452895350444,0.7199883136443104,0.6624737362151492,
 0.5369647973007121,0.8375523748521831,0.6187503989073965,0.3949019922210042,
 0.5939667099130684,0.23296861272424696,0.8979158894739178,0.2076924352935583,
 0.9293067614909444,0.4140227393573458,0.11391938761987896,
 0.16186440417053793,0.555617515061969,0.8649066055074456,0.4161509991563155,
 0.1321850550573228,0.41135369145386014,0.25046561978557724,
 0.8296194600272334,0.1524662136782764,0.7147307934493868,0.5090962686037545,
 0.314407465718166,0.6343758105062263,0.3471998644455293,0.7463840242140517,
 0.6229218124931337,0.4441626779895279,0.30485843121564316,0.6955450267898092,
 0.22330375821878512,0.5621791878027085,0.6072578005679775,0.7363020832928785,
 0.7905230205540069,0.14864483289610197,0.9881698574017734,0.6889999598507649,
 0.5423611815686382,0.7164362790663386,0.4666236313154979,0.3842035132487881,
 0.7121259557506494,0.6583609789918177,0.13175634248908763,0.9445444803655079,
 0.8509891613635896,0.6462637987473513,0.9655938757109938,0.4371279760956053,
 0.2699709048180301,0.2176301427205939,0.8135898499166427,0.5737086775336744,
 0.4202303281458538,0.944213645304905,0.16960641878286276,0.1812364899249277,
 0.2024279213570983,0.5429717881444664,0.01299193820053024,0.9033962422175444,
 0.6402276763925205,0.5874441188566683,0.5080379048242516,0.25863175638005276,
 0.6157774845596439,0.8560360269798472,0.6333892172576349,0.7244799810808336,
 0.39729359933377106,0.9503934909445162,0.04473001338998839,
 0.49552601357095893,0.7807083449489246,0.6846903691623539,0.8155495411672153,
 0.40496857211175563,0.15638819044053676,0.11755461176507798,
 0.9533219843000149,0.09802662042380872,0.34941943661121133,
 0.7888432656216264,0.5476620182474523,0.5963091797032598,0.9949544267905652,
 0.6816497267585777,0.8843540924784984,0.13904714946491936,
 0.23812206999593677,0.024266725091532382,0.39455315193840346,
 0.4107824167980325,0.5886494779957048,0.5883270957281534,0.14899906726303702,
 0.394149581954926,0.39261177396148206,0.7406140239848363,0.13893148271710798,
 0.05551895142634833,0.7383638773141293,0.20850956637494422,
 0.9543783138208779,0.12453837280002877,0.6125472430208583,0.5517738467811117,
 0.7095304378771635,0.4557274352440176,0.11116615327856882,0.713477184204829,
 0.41489007238544995,0.96129385874689,0.03504190493207082,0.5899903194752982,
 0.38887874216151064,0.20421865637774905,0.043803931510893124,
 0.5934601934090469,0.5177817274347271,0.34941706216438284,0.873811553932703,
 0.897902221025813,0.9396805690910399,0.27705500936139593,0.7932150185400653,
 0.8400712311246458,0.04733310400631641,0.9440879883146951,0.2449682021568067,
 0.8086311064816118,0.9980533975451504,0.13461388845210942,0.1412034546308112,
 0.47744416586198923,0.9780353656246763,0.07237085464281068,
 0.5758341681389336,0.03932773365438336,0.49080477490008667,
 0.06091612655411427,0.8025583251212269,0.13595912976946967,
 0.3288971969656729,0.8969198668892616,0.03697613473794448,0.3306490768717756,
 0.663989546239925,0.9638550744481371,0.8855637234703143,0.1470988366953635,
 0.6272098532500274,0.5511837121849451,0.5222997968407688,0.34880776263620716,
 0.9514025181701614,0.8546386480724694,0.4375266034055284,0.5499442594635464,
 0.40233164744265093,0.08521655716210286,0.39035452516985436,
 0.31485455560395514,0.3114638218753192,0.0952014924744875,0.8921241305616883,
 0.7758823997346328,0.9896638299941275,0.6749798427038612,0.8733809076386637,
 0.21292984576952823,0.37297631312305446,0.32498247786324286,
 0.3149920809101393,0.0546953969696754,0.4031561434154258,0.876458479313491,
 0.10597585916920627,0.2692179257517725,0.7167165888938447,0.3034122974200516,
 0.5356506174207456,0.17406528079654637,0.8169776840152839,0.6889779573059347,
 0.92889324161583,0.22366600711003592,0.0760865560948264,0.976661369182552,
 0.8020054772472787,0.0033965911727393827,0.7602681144470685,
 0.6436271956887519,0.046403306178526615,0.17661654793052495,
 0.11209466756657882,0.6359814248834431,0.9724813412516929,0.3506890069050165,
 0.30030556414817955,0.2967485082646064,0.4116394628103569,0.5265385877070184,
 0.9922726264121644,0.7610764814374956,0.8180206952563247,0.05889214104953833,
 0.350956421013748,0.4153306173926249,0.7694998284049929,0.3095628391166715,
 0.017869176144818977,0.08073738154906362,0.10653775273504529,
 0.05770877805614949,0.7856334512741172,0.3772353082441444,0.7582992984954121,
 0.15753439505409272,0.44826277825676175,0.5452442105825805,
 0.6215085836683023,0.5462187064227929,0.3672748444670635,0.1739019689547907,
 0.9178506318594495,0.217604320601392,0.09488114219238863,0.5090473190908104,
 0.739687035151603,0.6435440676063784,0.15744636300103,0.6429262243458602,
 0.6682236693101028,0.9981962280828705,0.46199867025760455,0.8647438630188322,
 0.44813989421058176,0.9967110877473193,0.4743906381820413,
 0.24605518083815758,0.362958830649463,0.9718999994892016,0.08844121802039928,
 0.08997468478767146,0.6082845166185719,0.9108116972047235,0.6537209591589852,
 0.4934792703862124,0.7560503602920006,0.41961195341428015,
 0.39557837620676095,0.9468825276286837,0.35764493722701207,
 0.5669990638086467,0.6489168102562051,0.5404318925110443,0.1977610686134622,
 0.6383888621914048,0.0523351652304912,0.15319471964774034,0.161868252210425,
 0.059058023641366364,0.8741815448500193,0.4062127361665864,
 0.5119595136700529,0.7107167809760149,0.601422122498315,0.5505058632921402,
 0.1795261197569462,0.6286255366029754,0.7922814553247768,0.3250624332825873,
 0.6760081857011959,0.35813499360086365,0.9325200302787751,0.4410604756861236,
 0.2984429793731538,0.2209196701632583,0.8461057749813092,0.34908253272279766,
 0.1658464912779617,0.43006673326843003,0.8808928544406784,0.7690889842212631,
 0.9421563081376612,0.9038999991707173,0.8272272730899869,0.7600594168334394,
 0.23350946022335972,0.03917137793043246,0.40340556391629945,
 0.627122276632588,0.9009122683857818,0.9121755086667622,0.31181887407675457,
 0.9976977310669065,0.529909646391985,0.5841052293717411,0.7363634683568677,
 0.9315228333490752,0.43841156940922654,0.6906600986359701,0.6783798222850129,
 0.667717662324544,0.02601415055477152,0.6133301452470299,0.8545023196346354,
 0.5799888019624967,0.6098008089662565,0.49908027309342373,0.8842092504995899,
 0.18648334001209532,0.6873136157403288,0.32515247616238097,
 0.9423426334570315,0.23029532112890883,0.8194655285055308,
 0.042706396388400014,0.37267108642015156,0.6997449001715851,
 0.2517142350535524,0.11533609044408388,0.042108609273880715,
 0.9051769845045122,0.4303487379876799,0.7734839915839201,0.2836249946780076,
 0.9174056412522275,0.287676275665943,0.037648977101357284,
 0.013334751633661934,0.1465988545430743,0.08638272810176773,
 0.5299981331370541,0.6738578432577662,0.8202916533000402,0.8438943427602581,
 0.10728098671866737,0.2632607923660679,0.8648031350032988,0.7042762304978213,
 0.8211655465366943,0.9457659759371924,0.8269017719665008,0.9986451152685198,
 0.08127497629298852,0.3806141813256301,0.4082603552802009,
 0.46386048144888903,0.4430514677453189,0.8874365473580788,0.68873979206769,
 0.15032381032060127,0.7512584850787865,0.33780913939420965,
 0.4480268527949334,0.32560847819776506,0.16022094160784495,0.27128370522805,
 0.9456924634497652,0.4862204868176626,0.541425672465201,0.9401971573066821,
 0.6153293593862788,0.034821219639292744,0.9627184407876014,0.670677081798327,
 0.08233102745430387,0.10522796994646488,0.0031726854900869794,
 0.4661922631766381,0.8548635110727426,0.15140309779799144,0.6493949877089805,
 0.4940096966802041,0.29202988403547314,0.31264052163283207,0.594750661463506,
 0.6317983999754581,0.07680581931789587,0.496887198318427,
 0.059061480720592474,0.49956659238842016,0.5245877270651113,
 0.08960025859986098,0.7107722029915938,0.3141585391143906,0.8877459781490411,
 0.44385358772287054,0.9768427332353741,0.6591816688380728,0.8368324016487141,
 0.035635245871849186,0.184563318010598,0.9875550751579203,0.5870960688249789,
 0.10222061422540729,0.1938008978903798,0.7781006733130202,
 0.07889123502477646,0.9315153411841253,0.5415057948745832,0.8822620203903646,
 0.9897496048516565,0.9791340842172622,0.8405494764251095,0.9605416057262328,
 0.5683710886227133,0.04762516129471139,0.8552486179086476,0.5167313374752343,
 0.11630817629922874,0.27388752230563285,0.8777730500350993,
 0.35981939024593634,0.9407102882735735,0.033789808203739335,
 0.3702879587107333,0.30975063012826465,0.12767233778903464,
 0.15600243313698547,0.21170623864429705,0.6748358396517916,
 0.7704602039817308,0.2626852034757693,0.3025467104001066,0.36600153895428067,
 0.8880205086802105,0.6726367841189002,0.3747063612304463,
 0.019766082459159562,0.5424099699598506,0.742615660080358,0.7205988678019732,
 0.21736215650990376,0.5046313879832879,0.8316082450078903,0.8731285737574476,
 0.09582428273957755,0.7023520314186903,0.7283376622649601,
 0.43301400781253685,0.29940869077397014,0.13274448742113754,
 0.7904262973097318,0.9423831076093919,0.8007263006482201,0.9413487652972239,
 0.533706385663078,0.5662536074268084,0.10255520772263216,0.05633111649506717,
 0.23539521862371116,0.14718258083642977,0.30829719596355054,
 0.49214679534012284,0.6189642042107173,0.9675199822040753,
 0.33561450073126586,0.39737791094706887,0.8445630560802355,
 0.1549110813779867,0.23563946328957086,0.8923017353735117,0.5988759603928042,
 0.4103807544762348,0.9872099579969517,0.21432070895146405,0.6484137200324942,
 0.2706656306264795,0.4040040375927426,0.4737146351363064,0.3067129124849859,
 0.8392586296452615,0.6605050363050822,0.7789497567925187,0.3026179938303516,
 0.1789043096380174,0.3364779840618134,0.43335284288710474,0.7649131262501826,
 0.350951860953975,0.48051412617758305,0.4881243352702225,0.7558724677385106,
 0.4716890846279924,0.45632855068840916,0.2720890044248796,
 0.11271184563185022,0.7029322612101944,0.4894450706529303,0.4471356161419968,
 0.3041266209396537,0.2628944891641538,0.48457851608188307,0.2117401111764009,
 0.0593883901597696,0.3485814038361026,0.9544590127023153,0.20180737691449946,
 0.060989849116234285,0.9714907870609522,0.40832900761218527,
 0.15214334764855397,0.7574427559994502,0.15005573267676153,
 0.5837093208609465,0.4739870675102392,0.40209090415393944,0.256241364090388,
 0.20834311682490014,0.186025645670709,0.6891911147660162,0.14094679166964064,
 0.362725234597689,0.5906251579335161,0.5282174869286931,0.058284247547498835,
 0.4549318389125485,0.9724947317750947,0.7022686777104071,0.4383484513553869,
 0.10034541558816223,0.10163110812118148,0.0768688809889152,
 0.5080200199433275,0.3162146612063289,0.5887152238001081,0.30535009633346655,
 0.7657945925180042,0.8599823028956718,0.5329340098001226,0.6628430828575349,
 0.9138706016274055,0.39653968496338277,0.292463199779341,0.4698084683639271,
 0.21552661448641974,0.37180700921914,0.688476092245051,0.4471494965446211,
 0.5691271544206512,0.9415392783414285,0.5950384259836657,0.299055047002482,
 0.9195994697930755,0.39469084415118316,0.46064642084673835,
 0.45171714669801666,0.8081404190836659,0.07341624504136568,
 0.5331354537601798,0.724368407074464,0.13138722538115255,0.808336672424469,
 0.12503813629404315,0.4399059894989069,0.5779043778698478,0.4927740891777471,
 0.6597276780122935,0.867797870731925,0.5649510672527425,0.27681913634622357,
 0.025861635076555145,0.9995816028158416,0.9016977767234501,
 0.0787863514469741,0.21638058278417893,0.8493442139431774,
 0.15293072210097503,0.4338729865714748,0.04634677436848844,
 0.8840237087801758,0.5154228228305705,0.16988926832066942,0.8689757857564601,
 0.852530629700945,0.6111721882903937,0.6282470610051225,0.9923106793342962,
 0.9365545781727121,0.6227196452569752,0.008756849572779446,
 0.2781260098817797,0.986059694562829,0.9032415703450849,0.7832412855522038,
 0.7048729661721126,0.11703126296311384,0.3205030584452525,
 0.43426854004839077,0.9519675809649029,0.7520977626743266,0.4331079844885437,
 0.06828422265724976,0.14258957403704864,0.09497805616636523,
 0.8993819456815938,0.5268802194403482,0.512144078749119,0.23028688391597074,
 0.48946204647014724,0.9045752075661065,0.1688228704557756,0.9935287921229768,
 0.4816978140527841,0.8024531188840961,0.9213077822665897,0.8567262785344705,
 0.9656443676164047,0.35350519986644646,0.8243234337487664,
 0.10921723379219517,0.6065748883551855,0.9233052751675623,0.5120058529749762,
 0.6589210961977292,0.009531190470438267,0.4941921138848776,
 0.30757863477717406,0.6293986034819997,0.10481378932421348,
 0.5164584965695798,0.8262491501209975,0.23756156352198787,0.6571859115275525,
 0.38330261077452676,0.7397201350407958,0.6462225622400292,0.9652595134312681,
 0.6409276019817338,0.2777061504771431,0.29918492123412843,0.672320132311067,
 0.9487027088659077,0.22456712018294,0.2025260986369959,0.976368760949142,
 0.13089872598186147,0.6974368628817325,0.05250406059478574,
 0.5826145368121234,0.7822436868435684,0.32000247138494176,
 0.30427888612897547,0.4699218840208512,0.277233363140058,0.9042525996191225,
 0.3718283084541474,0.8736410008808042,0.2243868516853451,0.20741110831685972,
 0.13254882047786065,0.9887260468465199,0.3802013417708947,0.8267970984749504,
 0.6159677691770379,0.7017640771873265,0.8323374042853695,0.4710825072908327,
 0.9565009768876194,0.6375487030119257,0.6561460226748247,0.269922094427005,
 0.6331935998693866,0.253024630343026,0.22884806993843876,0.4786584004435622,
 0.9971957019428201,0.9830836332528952,0.7528187136021197,0.6876401976817366,
 0.8702590142916428,0.4486180684748543,0.023384158605254446,
 0.2524105548435709,0.435263083396773,0.1537845431179734,0.8857159664516878,
 0.9942522670726414,0.3160081775019533,0.7522840692056179,0.12692318583221518,
 0.7299722819270271,0.49253071892158173,0.34790617623460895,0.952929520542179,
 0.2677095160527343,0.8914851451503855,0.6825963646338358,0.4925814651715723,
 0.5938767678956739,0.38277830275942004,0.6430534731957738,
 0.38402874651131425,0.48355118222786597,0.04904179039861456,
 0.9955398774808298,0.9731690907831558,0.6033882109467991,0.6124069343009756,
 0.5988294016782796,0.8640622564402238,0.15050873724086755,0.3701606706319964,
 0.22202897686806122,0.967659548102494,0.7527257157722074,0.440055850573412,
 0.7152439674170861,0.12679715570283812,0.9075768741963948,
 0.32470067618899867,0.7402499709012206,0.6840702561752179,0.5475994586954827,
 0.3750918571796603,0.38562373593717736,0.9583288079033025,0.5007392411225671,
 0.38090495177018324,0.08202512289933639,0.7474834978813281,
 0.40734494313875846,0.2778736767402258,0.30077092049301624,
 0.1638432043207465,0.35126828295475565,0.6828830800606009,
 0.05249712395448547,0.5867833957613086,0.589001186070204,0.6895904697865719,
 0.6056976197525471,0.20746387123846677,0.34541522091063126,
 0.18007101884539367,0.4620825318678742,0.5804174441395872,
 0.20501073108037504,0.5372656767314018,0.7305452285089045,0.8952530271207282,
 0.28212095043724894,0.2528704891935938,0.858580999568659,0.6525885968054256,
 0.3652762370832028,0.12780843235424322,0.11520759896215416,
 0.6685173962810934,0.15147627167181987,0.5797631541410481,0.5919763547348411,
 0.16531028549926163,0.6600474645971165,0.0354985033799492,
 0.11404083797997988,0.5346755196518863,0.6521425525008603,
 0.050245709004987704,0.12491399734970754,0.8621938606814976,
 0.9510382516295269,0.16324247887695975,0.970110121327195,0.5768822915010051,
 0.1439389992745126,0.43530668452629184,0.8996199987958583,
 0.44984620699055333,0.7371094732436786,0.17300430096665664,
 0.36890310014324346,0.38194205439331086,0.4050031975239762,
 0.44387720236715755,0.8364557791789055,0.8119331110426458,0.4446971887023966,
 0.9779131551347937,0.11131985969371683,0.8034808923881389,
 0.16320706168176446,0.03969944546545101,0.12047587661036963,
 0.7707013488865693,0.8802727469889183,0.14587832751626406,
 0.39828303591030967,0.1353105685940923,0.5925304936670956,0.8618604480601529,
 0.06975112930198424,0.7562084107158601,0.18972010092922298,
 0.9028082723419225,0.8776951969365869,0.1746519827456009,0.18784765328862496,
 0.35174393220198763,0.6170068907652191,0.07821373010242827,
 0.17055486948080534,0.028782942266912914,0.39825158645110914,
 0.5316148249368098,0.4171793393295853,0.2606277622911751,0.3381144887254385,
 0.6274708560908211,0.9134975881305454,0.8372950946163039,0.621479428920603,
 0.9050493516219072,0.8035878748334443,0.25001564894921724,0.6423434029760402,
 0.4219029302028542,0.7371317443306189,0.5544554733914062,0.11139940015794891,
 0.8997166726777494,0.19286954382860777,0.963665641675218,0.8666018765706863,
 0.5016993550807102,0.4145346434051769,0.6962756377747358,0.8898576915673206,
 0.3968494249055424,0.8840054821784926,0.9551799508792724,0.09178519115320749,
 0.7897476984105487,0.9987448159914448,0.3115389583226391,0.6802783774009881,
 0.6040368933586258,0.7091720228768847,0.46306521104507636,
 0.30458567078984466,0.14002647709276828,0.518641349247902,0.7358975063707055,
 0.19329468278452488,0.6444303721635445,0.8124424015672163,0.8925250135250208,
 0.4007094477856179,0.12274228857661607,0.38456775443610003,
 0.5105916221031779,0.7267900145284045,0.3475209880003627,0.46628853200465403,
 0.6664665218784328,0.7931564043058942,0.42189401062807763,0.6456106002474744,
 0.4666904116043372,0.00724282821867428,0.6758591541712178,0.3410445716772811,
 0.3497943948860136,0.8062383377157354,0.9720967926295255,0.038072775738348,
 0.9844427535880359,0.576825624463716,0.232602927917963,0.5707733979646782,
 0.2499551269700916,0.15973474216831052,0.6989645342865618,
 0.33550949147598885,0.9436222013485807,0.488784075697399,0.7695507022403261,
 0.328536888815101,0.07893677657964182,0.650063422412316,0.7162519121193857,
 0.38664011326453174,0.4298257532558857,0.8956110065775512,0.8219887185303342,
 0.8614910412622505,0.493864434576613,0.7548043942318496,0.13117107992355492,
 0.8068995816623706,0.24375731534182254,0.5686472434640428,
 0.23606796991486578,0.37446429954076144,0.6475271152141324,
 0.48995709748765337,0.33215493600931,0.3810851887844131,0.42304578071849475,
 0.8939182116422928,0.40586627749120474,0.4690523842126779,
 0.03980958283821012,0.09518695262322152,0.5100851100532849,
 0.28234623958909844,0.5793655361888982,0.6445993117996826,
 0.06022340807488713,0.20397556989740495,0.9992114324892512,0.639942331299112,
 0.39005503737524183,0.7720437191997532,0.09241900965070193,
 0.8276627418630282,0.5172860921679183,0.8448135155879899,0.3619920191112671,
 0.48438925065736993,0.5933041284152896,0.40092108669042226,0.356730117093665,
 0.5494418326573011,0.9060632369351802,0.6736363809890288,0.06670339577135431,
 0.6198787311619356,0.7415561535131844,0.19864772113983464,0.7192862216188857,
 0.7806136522910814,0.4217867049519196,0.14316229027191874,0.9137690169728199,
 0.2923261939353904,0.26280617205010204,0.5881354882615033,0.0715835796180383,
 0.4545915384614644,0.9347117072963304,0.3703448801833009,0.5125441774002877,
 0.5015518473330755,0.6333505334394526,0.3290182231727661,0.9396803888836118,
 0.9767984653322307,0.9260811059848459,0.6596988559943775,0.4063871955296614,
 0.17831080430266688,0.8187636033981761,0.059287563077299876,0.50160316079438,
 0.41242174850452207,0.7432076681387555,0.7191529400573526,0.1269266626895621,
 0.8357516438681001,0.7797298907159486,0.014344842229910393,
 0.23426021942264974,0.5759324525698358,0.36172389534838223,
 0.7409918889562144,0.536717618514664,0.07513387991422005,0.8735300490296538,
 0.5651305820984908,0.6203378937063975,0.23004931534533746,0.643930166504844,
 0.8058387269564732,0.8041029132213905,0.3920532501293572,0.01286077733115687,
 0.6398980737901393,0.4809267796143881,0.9459986669135928,0.7668685659621333,
 0.09746278564845323,0.4822435988027971,0.6843797059235416,0.278237331733311,
 0.19941337121848735,0.0857067086770632,0.28095536930052645,
 0.5494595579614232,0.5618117342545557,0.45241051741128047,
 0.06981650845260523,0.2896244954889122,0.2695422608533449,0.6528880856025718,
 0.4857101292323681,0.9176007512977772,0.4228310031316641,0.8943071756250343,
 0.7372593897456825,0.7568893636467929,0.4840930272320859,0.3336530954400607,
 0.008076077137429305,0.08105450714342854,0.7991309166371376,
 0.04505434958187016,0.36412164695627613,0.8176045113956304,
 0.0921085714837766,0.3554340036800029,0.6619666104798211,0.921607020614768,
 0.7539035415694207,0.5796561693312323,0.39234226865749755,0.6845884740974817,
 0.19597897911945195,0.04458802276310958,0.24220069700022917,
 0.20765952994143677,0.4447215823684314,0.6439433900318184,0.5941046416021162,
 0.8443013685849896,0.954022668188557,0.47695890090151105,0.8621249332845968,
 0.5840480991853748,0.9261466651055775,0.6590628743572846,0.19414897193642688,
 0.8842118215584049,0.8935681719278963,0.3560860819168641,
 0.020643810798668505,0.3489536921455656,0.35522682886494905,
 0.5355425871954415,0.8398063613926765,0.743078994479576,0.07979317395438756,
 0.4626991410560579,0.19494059983006595,0.4296962771845252,0.6099518737383053,
 0.15959173759589906,0.3779561523922692,0.6590371663614953,0.8756582759764289,
 0.9538853887150033,0.6791587492566853,0.9965045187577394,0.9174892984558509,
 0.32097482258647714,0.5272030975457612,0.4961085784894743,0.6500901714514469,
 0.0359377540526834,0.8151236079306955,0.11466843997625087,
 0.47941825210288747,0.7965993755679578,0.5563721997109774,0.7238190262463222,
 0.5121540097474407,0.19950465709979892,0.8422820039776294,0.4375249139470603,
 0.7101044687804111,0.595039906619812,0.6514615539609452,0.511219605312788,
 0.11034707510139963,0.6342419338419492,0.49802935384936564,
 0.24224480279557925,0.7492840167273596,0.49540515280725916,
 0.3939730317759875,0.26941801220361516,0.16663555628430748,
 0.09395065136094649,0.9311140005998234,0.8560157929362422,
 0.31483802950239503,0.05362429709520167,0.27546908609209475,
 0.801446703761246,0.23366677133019986,0.019368783357303965,
 0.3844149150227405,0.7310011006490994,0.26124534720555426,0.971715388197141,
 0.9881999547323701,0.052037247674269316,0.06727377732541096,
 0.8323062701163305,0.06543921653412665,0.3282757134867148,
 0.34536417250680274,0.2594412824321426,0.49950194109247226,
 0.24780013758048614,0.5124358513385106,0.5158683031261591,0.882968722209455,
 0.9344266126373931,0.5746034369187325,0.0806932721885959,0.06771684460023197,
 0.65387052387408,0.5965142589064072,0.04580819471220676,0.25678937711211736,
 0.30010883067117156,0.724543492451019,4.3605323083184544e-4,
 0.13029329804973433,0.46536103677728424,0.12192276910528654,
 0.15194939610433245,0.39967084493572913,0.389253909272279,0.511975136453949,
 0.7253030022514448,0.47853469308645813,0.6171307475813708,0.3233760537997292,
 0.24183611343913414,0.166416303939378,0.26966692793344726,
 0.38041472269931653,0.2600419910258034,0.4823286845602104,
 0.45209343273349634,0.3632875052649476,0.2925533086201848,0.5160535543532698,
 0.1352196440983754,0.7370082505004333,0.6095926389545643,0.2735385882216739,
 0.12088429287332159,0.9845075075472808,0.8862985341122953,
 0.49098565831029706,0.31929988953088995,0.711530442479664,
 0.013614467323087687,0.020788664995971207,0.9429769071915522,
 0.7680400918669963,0.47675467350262646,0.6854117418231896,0.7094819492273102,
 0.7565517582759754,0.6546934607959644,0.9462931654145694,0.8346298902279436,
 0.18456637335967008,0.17520752594177247,0.47477813065047736,
 0.3418185691433443,0.45039771530353867,0.10432427623552187,
 0.21517385981304704,0.7118274516762073,0.26257306639277933,
 0.2316834910322656,0.14918934772570758,0.30124797184902796,
 0.43516227615827685,0.5866935146017076,0.48705020817052536,
 0.7719772633794328,0.7140397957592812,0.18483734197827117,0.2556643870076526,
 0.36980696019131687,0.34032447598334403,0.7248863035874569,
 0.1514787077443126,0.8766177260623002,0.036783784487617766,
 0.37905619062496554,0.7116856492006997,0.2976283898101785,0.786284128816016,
 0.762458528041513,0.6461630066539301,0.8243363030615352,0.5997539292211177,
 0.42649533118908545,0.21282208222664778,0.8134014520521569,
 0.28685204451943813,0.09748971192470046,0.37289929751014017,
 0.24003677728732442,0.02720320495180495,0.25369270635085495,
 0.07359928316306574,0.39882855126003525,0.4643698854496089,
 0.5290554849136087,0.49968265846446236,0.3301576706487256,0.6789775431225582,
 0.7283266645783388,0.5648466477571921,0.042315818381777115,
 0.5640123503826707,0.5296148637353558,0.9102174186614383,0.75371783712402,
 0.4832616473116955,0.25052386006011407,0.8105119413060905,0.586921411577287,
 0.9079282052571497,0.5791628117782176,0.9673023942608702,0.6646220771885167,
 0.4503977105756847,0.5040377453527789,0.4783852174680028,0.09615291626454103,
 0.5187796854498111,0.11670994797826417,0.15854547853886158,
 0.34270562606402744,0.6878393078666373,0.7545889790127571,0.7524340127511484,
 0.7606117253520468,0.533201013975922,0.5114132405724483,0.6403203149194931,
 0.8239235573989284,0.11382540731257096,0.9441045059402948,0.5552835820171816,
 0.9824136732379396,0.9865931578127439,0.19691946863500653,0.446845771662538,
 0.8883167118173585,0.1956300111727216,0.5510315479691227,0.4490653456606323,
 0.9062960448402192,0.582835372582091,0.5244170477857771,0.5600173634338708,
 0.1182733291454724,0.3453622053775707,0.48077829443061326,0.774853782348268,
 0.5585041577871334,0.6830400044950218,0.5130766647143647,0.12234428739606773,
 0.26266769949981605,0.6643200649795022,0.28074768635167513,
 0.4378934351684658,0.6761708311344166,0.3798013005445945,0.42191585942476195,
 0.8143598055970555,0.4322654098767681,0.40316848467591615,0.2767707632514933,
 0.40208807384985157,0.6326669407539092,0.4424448016135496,0.5174579743534453,
 0.7887277497300408,0.43443013909447614,0.3236114426465092,
 0.08324407726240446,0.35314148265892475,0.40828849188207206,
 0.8434952825782024,0.2281149330229929,0.8967623447090876,0.1609388235482182,
 0.7450089091596473,0.6056298891407563,0.9257865329770287,0.6193858818015104,
 0.4273059472813585,0.867854148025514,0.9035819978448911,0.6515448339657564,
 0.9338771707843089,0.3512450202532382,0.6536575765899603,0.8542036331010554,
 0.7182697329958133,0.17730755394140696,0.5830395049565344,0.5811249249352829,
 0.40630216377569184,0.06522267065105924,0.19626161035295786,
 0.5280927776831286,0.43578239187085843,0.3335314334210384,0.1680150269387356,
 0.5010533628971923,0.23164408487489851,0.57488041071913,0.9669116418544617,
 0.6688825491323539,0.33189629907576534,0.5344741331586416,
 0.15002856411067977,0.46692176303593225,0.4474360579642621,
 0.7586996600669079,0.9538091331468709,0.2592549534813051,0.8635200103311141,
 0.47059120097561224,0.35901807301990174,0.551474626707028,0.625620599729029,
 0.5659023283717971,0.04385477092086032,0.3823331385427454,0.6487732708835564,
 0.9114434300682173,0.9430754055538184,0.37480276113379973,
 0.11460537019678407,0.8063813324692563,0.590986613649705,0.30712911312135516,
 0.470847536427367,0.43901459996765646,0.20032992538414907,
 0.38047802800728125,0.16700958026002766,0.964514410066849,
 0.37145860529429053,0.061247245047717724,0.9951325115826837,
 0.7764794201203528,0.16111393691060127,0.44556694519893325,
 0.4774722834906767,0.1128817935101315,0.4211094971547743,0.06631535781549736,
 0.9604297591297515,0.03301322599828327,0.6902577964353191,0.8140742811565737,
 0.536723994709383,0.12516094057527094,0.419176257632057,0.08521419768533178,
 0.9426582273884494,0.003366572556110392,0.025902256021067438,
 0.7806631353780064,0.8585628362088559,0.16684016246799038,
 0.49344688049703356,0.13196970029058164,0.3920324599240783,
 0.16679751282492983,0.7144492237462637,0.04535395189679203,
 0.3754184609485861,0.5595825515378617,0.647699408958317,0.21373442610986348,
 0.35852305291260633,0.31612895805430585,0.6444413930192152,
 0.9389834199816968,0.56510600002899,0.6232966370550914,0.6077273715030314,
 0.6774310121236846,0.874794336706137,0.06300418276216557,0.6548303937886127,
 0.9541555603209632,0.7073521411078647,0.5854896874579324,0.550929305038236,
 0.9323286046570424,0.19114396466940942,0.4474477562376711,0.0766370543570084,
 0.790032248900624,0.39193412408343686,0.24328907457229854,0.1813391128338413,
 0.29010950095202936,0.7372280727886622,0.19933836431601792,
 0.3806310002370037,0.6386260560952739,0.5586056560289724,0.5807849411562496,
 0.2710068770398355,0.9864011036366997,0.28394897641401107,0.8202908109920429,
 0.9453995150691749,0.9999485602270585,0.029626625962465347,
 0.4806092780659399,0.5801782961279277,0.9242895613451624,0.498608652891504,
 0.7135634038167818,0.21970503571149425,0.19629180050756534,
 0.3441601786878825,0.012435799789235258,0.27828264897981203,
 0.23036194132707388,0.36024507201439515,0.07151180863911599,
 0.8284776631488568,0.7767424253658232,0.9135777077646683,0.5521738193420795,
 0.5835577256903973,0.24690988090604704,0.7411030945154952,0.8436592506782323,
 0.9678704140987848,0.6122857215905211,0.224868469014343,0.7779793130843091,
 0.5787854007166914,0.6118702916463916,0.5979992694318926,0.5477449663832775,
 0.18041912885556632,0.3710058939210992,0.1258496032380776,
 0.15185047273367946,0.10220006661576242,0.49193891687503444,
 0.38145234462705413,0.18697307879417913,0.31042944246652626,
 0.6319930099121502,0.3026661956117487,0.6958151945454851,0.4720028428926425,
 0.4225173207803474,0.5425431475684149,0.3364224733827391,0.31516118898019596,
 0.35358201534740763,0.3323354231360254,0.13945959505240624,0.821124524982056,
 0.5995442773345081,0.8594106877452667,0.07448274027052915,0.9135191103719621,
 0.8056435240380996,0.14082346262317624,0.6897400083532939,0.9421818793206979,
 0.9782807244158976,0.5162923146045412,0.9973196595538656,0.1836342366106315,
 0.5008846914593418,0.7940993840852237,0.6420815686956507,0.7641287487289165,
 0.6712734127005621,0.051098913214068564,0.8352278048938895,
 0.6666408999160858,0.9229466979364921,0.4612215942787371,0.7737058063598792,
 0.4653861493400221,0.5165388691189268,0.9515454819208415,0.9845590743867452,
 0.21953848260293474,0.9244394649374716,0.5865709318116682,0.9062544493046248,
 0.7482734113830485,0.6754818783171812,0.9851413882447766,0.09105269137958105,
 0.4342613142095896,0.31890468733582045,0.347822128134626,0.4890843688097454,
 0.14459194582072943,0.28589417695657526,0.9521215608824429,
 0.3989130986489131,0.2436966709832924,0.4514002835752452,0.4688928415329704,
 0.12317156564056553,0.6211930162946107,0.5726663044585446,0.5333949471360757,
 0.8525781770697944,0.6504319836230718,0.8092582882653675,0.4901707363652137,
 0.44116799053131683,0.5411752216949759,0.5450623876210346,
 0.16803786970753298,0.8489122841883842,0.43653960297731775,
 0.6726526998346894,0.6706599875008041,0.21655398469327714,0.6279208110583133,
 0.9374719908910139,0.9227355647270704,0.03130716416758983,
 0.26063529396028584,0.7658278356101551,0.7438385032032431,
 0.28092642649430843,0.1524142639826941,0.058098226407377274,
 0.5609544179267039,0.9177346296878575,0.24444789218570573,0.3033084228205035,
 0.9222419227877163,0.10275777919245199,0.8951702116984332,0.9416701414712667,
 0.3918445167968211,0.6871619147936211,0.7428967114331525,0.529680049499765,
 0.3589061062915384,0.6675287481956328,0.46532686552772506,0.737463909718769,
 0.44928095088093667,0.4209470121849278,0.11911144912774663,
 0.7186775372174946,0.2130718360475593,0.29592601072495506,
 0.41799356517097963,0.25354614000032827,0.44641146148537314,
 0.502887184033143,0.6084547423372584,0.2193516493692219,0.2491015793430612,
 0.14141613213617787,0.8381162367346889,0.7759018337777752,0.6538905420688897,
 0.38163727747518583,0.7791373596980999,0.6535351471542941,
 0.21975295000767037,0.6069607783990296,0.6741116469777473,0.1056934678476914,
 0.12667097838083652,0.6318768936971371,0.27911449277189304,
 0.9201527645923143,0.6459157682587627,0.9465373153559102,0.24558549499753557,
 0.22272620342982663,0.8234130575033579,0.20880193122835478,
 0.6960474400615301,0.030387312372970654,0.7400839933075906,
 0.5510826651711747,0.5180170356116307,0.4042078740578622,0.9993015417363897,
 0.2989900386495663,0.5855068397721246,0.3499767934835094,0.6273561170174184,
 0.9361958719685599,0.8188991026842909,0.4196667340329795,0.9463348543581961,
 0.2145480515326481,0.08821272186317874,0.87907338029736,0.8688559670105562,
 0.021958955516207967,0.5628707812071779,0.9981561654274637,
 0.9343088928249623,0.6179055271777034,0.14264160967070383,0.9818032319238562,
 0.28427440728528997,0.7243532483818503,0.5058517954432085,
 0.27791030449937526,0.7507895867100792,0.34493479015515094,
 0.9479704422796573,0.8728385778373391,0.12512325444840167,
 0.032446877450931666,0.6433428780758832,0.7152404420692915,
 0.41108248062741437,0.9564394386088473,0.0620978389892638,0.4757168684588877,
 0.809689465756598,0.33981475001009176,0.792529897256754,0.9360499494131205,
 0.731571460183475,0.6561424509233935,0.7950790233851047,0.9621697158258876,
 0.550928943928118,0.31105749305716024,0.15045804865644608,0.0989758259845892,
 0.25744002369223085,0.22645276478502674,0.8060382045656225,0.511742845885749,
 0.15155438667523546,0.9621540384134748,0.9165371455871393,0.4896716806535075,
 0.954199501198631,0.16302969181912175,0.8565031742785255,0.4475091080143382,
 0.3824334079269127,0.46055875558142356,0.9958065375125881,
 0.02308958613184031,0.9215361182233264,0.20488237134542686,
 0.5125358237740985,0.685555177713118,0.18272715653054805,0.3214787150110907,
 0.04571480479419421,0.683113998315069,0.3860849996307387,
 0.003312602000107856,0.2174574500315153,0.5416475536676038,
 0.8611566784108036,0.9044409798146033,0.007914644755860456,
 0.06053420437707413,0.41518396937302604,0.7123616339160774,
 0.04171989694326661,0.12124008617422244,0.5223805228698304,
 0.18985874919073342,0.040811342796452155,0.8623654937563163,
 0.7235091041061854,0.7408251175349447,0.031166667981718965,
 0.7332747977829978,0.9710465536712074,0.3745307189372675,0.8351285503069711,
 0.5065949909606459,0.10501997791606144,0.17270783658646394,
 0.8568762566586483,0.19336876196161268,0.145656967599744,0.9055998498027716,
 0.024519809043473284,0.8007105638299417,0.9134492017224233,
 0.9994460568926664,0.6176279421724156,0.7986421884252478,0.3421179283962541,
 0.5024864999395569,0.35728512661511636,0.8530366863384109,0.7801927959635474,
 0.2074040546386009,0.920888030047629,0.3222028562147531,0.8351197787705564,
 0.7786918058603208,0.8234244184366092,0.29247093770003596,
 0.30083083541114375,0.668320038378445,0.5656556442768593,0.9479450736629255,
 0.392002183821503,0.6619976825112444,0.8864162517405404,0.8153896398857655,
 0.056735823622656056,0.303534616535014,0.29407194389751323,
 0.19687501687975428,0.5889440992736271,0.1376944811419728,
 0.32202898082240006,0.3965315595111142,0.1912358967507074,
 0.35338416350621804,0.9126277644770218,0.2458358505842747,0.9549174284837041,
 0.29837825809500673,0.23606529541214027,0.09039933757569751,
 0.5107408949212414,0.7259402998671644,0.23848207252430687,0.9472801141987643,
 0.4698842052699971,0.9777558271273702,0.766307411935804,0.4962416580105783,
 0.33951501661626105,0.637228994547755,0.6855518384787084,0.8051346357213068,
 0.4405087160435226,0.1910244092988621,0.49471715902328794,
 0.09970924196002018,0.030331257379868748,0.9184758958537161,
 0.47772423929125774,0.4955452887352729,0.8093350774123124,0.3458733633057459,
 0.4532835385081114,0.4841520163098165,0.6365310174320988,0.9941338445977445,
 0.9202834234554959,0.35468130283129473,0.5166591188952344,0.9366338646535086,
 0.3102629104780752,0.5612922654673769,0.19178533183901314,0.3414673090872201,
 0.5666791285978521,0.8717070134662224,0.33257908973696915,0.8059930842986929,
 0.9334301675342705,0.5801210281726882,0.22796446574467777,0.5497812970039959,
 0.8946116240049877,0.7378883042034781,0.7872810950613814,0.9613019854472604,
 0.8689420603437867,0.04377618460402255,0.4325680401599612,
 0.20874137497950063,0.6586242396388388,0.3973847933673854,
 0.03436218751956388,0.18015161675485403,0.13093181389572817,
 0.3632477433868475,0.13226545766739972,0.1255890487468333,0.6756812336662295,
 0.04196899142388866,0.8240905119196865,0.8093064146918163,0.5826000141102919,
 0.5263848788637597,0.5085990922918536,0.22719685397388012,0.453583719318576,
 0.19761053618392954,0.9873507878259025,0.6424575413965621,0.6515157855200145,
 0.4800154589752682,0.48510912407222084,0.8931721170596637,0.8345325601103186,
 0.7113194919961509,0.590089508087161,0.4971863427358927,0.40856162707236066,
 0.6323723095793056,0.20530440145472106,0.7721179482537976,0.891251144457298,
 0.02279929646679968,0.3191550647575093,0.5107496500720129,
 0.13264067731402762,0.2577314104203291,0.2769571265618618,0.8443216125056854,
 0.9657547431590637,0.8336302650829326,0.8781040054791903,0.731293062575025,
 0.969093271768338,0.4410063000535138,0.10624402798463062,0.31387912303040255,
 0.16531802531135642,0.38967988668857856,0.38774297867045726,
 0.28046323191744216,0.18786949309516454,0.8477211638826518,
 0.08094203158640245,0.6683706643456478,0.4592363739008001,0.4167970302525724,
 0.4927085218618267,0.5113274675904037,0.6971000633049614,0.25890341114599114,
 0.16971747747635313,0.4939128989855033,0.1420739156057793,
 0.41413498534318127,0.8041261230750276,0.6700764416357387,
 0.25103844732993363,0.5456015348891317,0.7058985811302034,0.4322637874145012,
 0.46267268494809943,0.42120298606637707,0.6799244454681523,
 0.8033434222514243,0.6318184468537437,0.9540525161434831,0.7692504201450221,
 0.5285284348770489,0.768488269072992,0.6001691278648807,0.9650134563715411,
 0.19077527741893596,0.5381696852544254,0.7245035162402745,0.8708868918483423,
 0.47682454017130904,0.4722838583617036,0.5961840625832435,0.9615619145399632,
 0.4508724926119405,0.7385073224584279,0.1334366133290691,0.9260746641983636,
 0.5474838039999075,0.13538695992547845,0.9372987088524962,0.8614728055081269,
 0.30376856908559935,0.4853407803862857,0.8107998502458948,0.7987153446362251,
 0.997861074315846,0.511277955457942,0.17805458669354113,0.6026555665578402,
 0.6946620049242509,0.12964886068503678,0.9476948154770051,0.5225905641704227,
 0.27267946587233427,0.9935373179035951,0.40837115746887975,
 0.49535342444567276,0.6203990331883357,0.2986136220440986,0.537931706566292,
 0.5685066857076247,0.6401424148627122,0.9538327102352429,0.6806618740314144,
 0.891018779342873,0.11231802134198854,0.15285028714499038,
 0.41860123958519246,0.5212804403442355,0.9234854746317358,0.9241143468121147,
 0.8642115575986726,0.9672575022536988,0.7391893251520544,0.9744505777811575,
 0.6599616295112083,0.15610715735060932,0.23908470146790517,
 0.8510408049857071,0.048571857239026794,0.6980860949018356,
 0.12041201422530168,0.17988673171759584,0.047760073658167146,
 0.19804260296947984,0.6316104948606657,0.5613876563437745,0.766096273261822,
 0.3262802006075983,0.1467693897350656,0.3359663697148013,0.09409082900065968,
 0.48778059011330677,0.6237721659489216,0.43194069210182806,
 0.6217605274023535,0.733543204113669,0.6723272088352293,0.9608454490611883,
 0.7793036235198139,0.5996431669867947,0.3995064162105326,0.18222773691453997,
 0.7310636984166377,0.6192748536409818,0.7256408501578998,
 0.020039596711848073,0.006615439981247473,0.20302831646339836,
 0.6574781821706523,0.9077088661768395,0.07955514791390939,0.4866545988734343,
 0.7412252718232866,0.8112360920051203,0.8605114263571655,0.9052742666182565,
 0.7412168996026424,0.2603066035872116,0.3221965972939712,0.13456444425696312,
 0.03157516516227721,0.16369157027630188,0.5886125292897455,
 0.9105611936698879,0.7775334637001254,0.3812225019054182,0.42212326984262316,
 0.17147332160641882,0.20355026506944385,0.8845174741080648,
 0.06200160317882386,0.9505668452350533,0.73131683388951,0.579296024811872,
 0.6771768593475469,0.7378358222218717,0.635441595091672,0.5035833826957222,
 0.8507944589593273,0.20862701233796932,0.2954775917176492,
 0.47635113120488737,0.9763834820873835,0.5875370698442075,0.2769239809322164,
 0.26743308585005077,0.1363268565063961,0.8623559627143345,0.6836621892051993,
 0.9614890495640797,0.5032851095638355,0.4182190843045053,0.7590626603796703,
 0.3748176449277463,0.9202194323271832,0.5844898994035165,0.20654048066150033,
 0.059994606263024663,0.48885723474843945,0.9793333191356747,
 0.6799287189507063,0.6520017628959947,0.44769993470219394,0.3544887886415814,
 0.3247656926659095,0.4029029144727242,0.3294424584390706,0.7489032228483995,
 0.5843180123302272,0.8464453905121513,0.019857298499037612,
 0.18896899390021948,0.8544582660968767,0.4646128120962002,
 0.21873672688865486,0.5046796237256213,0.8238592290990987,0.7057214440806857,
 0.6681052151151983,0.5630121721417162,0.37984372650429,0.0343990916981296,
 0.5632586199224611,0.43779871430851613,0.296034668970361,0.2655477071543402,
 0.14574712252575517,0.52662263053042,0.015017913002957295,0.4045542219689108,
 0.5074627875426481,0.635479837513222,0.9192395689662678,0.6892162906994407,
 0.8558777625367056,0.4945166177180035,0.9970373598963362,0.3619329223423944,
 0.4968452424316778,0.08558909596901487,0.21533539579201078,
 0.8972790663661718,0.2191385954160281,0.06648639806826706,0.597690526188331,
 0.8651226997683765,0.3336740097594335,0.0906932759092276,0.8695137613675789,
 0.9377194559379624,0.5972261675174313,0.01767599087112659,0.6561380788222708,
 0.5510556288815693,0.927186611459723,0.3242833764247699,0.8719926372647426,
 0.6761512833674224,0.6952428942530863,0.8248072553385664,0.8516912314326178,
 0.5607228704428047,0.41353272057024215,0.7520730895164568,
 0.22186904960851683,0.5053815175739287,0.8904414692986019,0.9575440587488986,
 0.4330574094082821,0.9804003741936151,0.2133296837452835,0.8208728909728698,
 0.9482710839052275,0.8988488380228228,9.729475752664474e-4,0.66358869984724,
 0.20619915569511182,0.0624850934221709,0.2823388544113864,0.8161374871297514,
 0.3480584050280532,0.3464380747242344,0.2494897787844943,0.9374011073578525,
 0.9125186653148534,0.4938748503079823,0.052065645056053045,
 0.3699741168994094,0.7549214798108633,0.24016702440810556,0.6695320419325417,
 0.880312033830082,0.08683579969512845,0.521345713803301,0.09179530734884433,
 0.27905957547163074,0.2863422771188472,0.5746056711041524,0.5051424240418304,
 0.7350105225763567,0.49952438909881636,0.30132467473556857,
 0.6700812187147145,0.16435960994584597,0.35621764838686043,
 0.5104045035500081,0.08166128895223757,0.8818479140426513,0.7456394465768446,
 0.2888947304237667,0.31600739927432975,0.7099101238533491,
 0.08460777789424989,0.2203660374023616,0.7947274436897249,
 0.028603019680688124,0.6890380481347895,0.15904349937185325,
 0.281933082233057,0.5229762848041472,0.8132685439676663,0.7345050660471434,
 0.6351055167123263,0.8750862369791714,0.7205410950916918,0.22938797430222424,
 0.623680847746924,0.605765376361688,0.924999110812339,0.37369933033085434,
 0.7360728933233137,0.015499217431393153,0.05569778774445866,
 0.8684671912229494,0.5733662133213501,0.4680564707873075,0.7798730743126907,
 0.9746875733414639,0.01787171209389793,0.9567364884739441,0.6185506808274644,
 0.19483867103399422,0.7528012010400982,0.1448541843811394,0.606292146572021,
 0.5284940681197048,0.23954665945089304,0.16060794265151168,
 0.9373017340848562,0.24973035246423692,0.8147030967554395,0.602763541880055,
 0.24542360941302988,0.8497687252511148,0.26721959187215216,
 0.11687784776183119,0.43904682809244133,0.9742883062577042,
 0.4247575800247809,0.3125153361537286,0.46103761544071975,0.6982181913754202,
 0.3500667231572625,0.571988201912949,0.6404011647400367,0.3762738473321364,
 0.9686969243757719,0.2751808960683758,0.5025619984155122,0.8068064543958448,
 0.30692858112598154,0.5061980159826697,0.41957138637800495,
 0.03865072586763818,0.6969022984471012,0.26356810151886734,0.662331769819221,
 0.052092601871939426,0.8715389018175134,0.4155640800449534,
 0.22367423750730775,0.22766576488944512,0.39746969207922744,
 0.6375708623659819,0.2494361357088577,0.20164596390535305,0.6599075849791392,
 0.7556849720212625,0.00820011695724232,0.5197342531767348,
 0.16245467483705067,0.8957573937803827,0.5376667587247603,
 0.07085703275858668,0.9320608857531861,0.6961205805763171,0.8069391731792942,
 0.3088714455109508,0.7027309932878396,0.10287002672840917,0.4321287859692774,
 0.00669978306434571,0.15942043733379063,0.15297283172378173,
 0.41076534154650257,0.5663419330727569,0.30423626888136457,
 0.3647649112671928,0.681048146767311,0.12437169786136926,0.5547297209928602,
 0.41386737434574283,0.3841334683948334,0.44704678805228004,
 0.3928887907780044,0.8904635379834512,0.008924144789430821,
 0.24840766940568848,0.3339628837397819,0.40006222272500613,
 0.4026023383856039,0.9333126626664954,0.7220329534352243,0.7362191926620388,
 0.6768975535971297,0.6001001734112426,0.40826171390623056,0.1534758001581198,
 0.3684887297500026,0.10962827606615477,0.06115978617075368,
 0.3319420949823184,0.39763623688240946,0.5326982491583885,0.8495822263432085,
 0.6252042028613658,0.3320390114455951,0.5337055109720332,0.42474086433772673,
 0.32263228863690363,0.6771944701235456,0.2383632510750312,
 0.13253866731698516,0.8724230203210083,3.0938662468771305e-4,
 0.16039221120802083,0.2102242576079203,0.1256816336812404,0.9975803627937916,
 0.9029695479479295,0.2114792150127267,0.7449463210055786,0.7350192967021175,
 0.14706070035035668,0.3313205094569045,0.34764516698573567,
 0.3704724616192978,0.43509634359462124,0.9512171565838285,0.844936151788061,
 0.7600135014472555,0.3446022541485195,0.5027755453980529,0.9619287159777193,
 0.7200834410637094,0.576387445357834,0.20977737476680958,0.16557741972343132,
 0.8743403076558963,0.5638381460387358,0.35141327560405833,0.2604497538595453,
 0.9565233075338612,0.631703765594508,0.2180259922704818,0.2919638783242665,
 0.8126382683021531,0.6122717691127858,0.5242651632811081,0.37910981652808573,
 0.5212328052311723,0.05577055977511902,0.5311932845089494,0.5115875667245208,
 0.766889284519391,0.049765336210262245,0.7457163617522508,0.871882246442886,
 0.11990781358680092,0.041234085749074056,0.472975697812973,
 0.20662901575787107,0.3604042959669601,0.6048494804008626,0.2268722464616465,
 0.323539706048958,0.1886512910006084,0.10395247819391706,0.04655610656482145,
 0.9782433909364451,0.44163230890791594,0.2194671969626777,0.961378972343714,
 0.4727167944450279,0.5006092421210356,0.05018275436259623,0.7530470887777174,
 0.7138556751069984,0.8120280593666046,0.06945926812960912,
 0.36986273021696026,0.748833969047682,0.3974304635739686,0.9817374512360263,
 0.642843042526082,0.10831146041325823,0.6389161305920096,0.43029400745118196,
 0.8270370256774195,0.6832633480967178,0.7173472598128977,0.07757539063908792,
 0.6424074260135423,0.6993928762440732,0.24609442998518538,0.586670702306042,
 0.5948879168098024,0.5697255107979031,0.18230666517425353,
 0.06844025416578092,0.46939905938538673,0.7586386547094488,
 0.3853279814882239,0.030687511160871095,0.791934174873126,0.6703909672119509,
 0.5226481051267258,0.9574034769801268,0.17436003037215564,0.4474914921816734,
 0.17209382648043303,0.33577788827880195,0.6433961372303008,0.25743517367444,
 0.7813667996965017,0.9435884724670763,0.25288421102235104,0.9924435421593004,
 0.5662349940391203,0.7118383282394523,0.6656461587076967,0.5334360121331352,
 0.5240768432175951,0.9816232358324871,0.7386292471193812,0.6411768183976126,
 0.6271617266701297,0.6386640640232042,0.18081865071636116,0.9362384665980312,
 0.6519409062766062,0.13251465903939952,0.3757812414492886,0.6395878873578527,
 0.13836047295858533,0.15555891811463085,0.5404431969606607,
 0.6663977730557898,0.818020573277916,0.47120499030664387,0.291734705795079,
 0.25570610911674874,0.34966937792516595,0.3374431802112231,
 0.24932235933058933,0.36104245858552175,0.3811631525261844,
 0.15816983565222353,0.7033570771807482,0.7128026298790534,
 0.10741720717003767,0.874541000291099,0.6158821551512322,0.3171741034522696,
 0.22553632112515787,0.9746303025510298,0.12799093020306762,
 0.6371168869237084,0.7412513798690227,0.5018762386565051,0.8081442187225305,
 0.18883674038687348,0.8156399924146089,0.6158836752654677,0.808145581892139,
 0.399951660161428,0.6785862302351255,0.4331310608514599,0.15168163564551107,
 0.3331319410069722,0.032772917001359314,0.4378593337405918,
 0.25437104401267996,0.4436896150325076,0.20438526316653183,
 0.16728568549530165,0.24001718475174705,0.3257074349802991,
 0.3627044978928434,0.5771664677627075,0.9699285058890422,0.1545225106765944,
 0.558790429398832,0.4144401671460194,0.2182042297529211,0.09907413941277576,
 0.9371573438042518,0.2537533261088376,0.08287636572282031,0.5206993928160822,
 0.099638473187315,0.4703901214465368,0.030036513777449758,
 0.09584447827917086,0.2843787529666725,0.6235750764131068,0.9109114935770686,
 0.7120027637376658,0.7384934729248939,0.05593086140605197,0.8343415297391313,
 0.543632431810992,0.38716600925446876,0.3589455692781598,0.0745271500081357,
 0.4736773283863699,0.1079015488961299,0.3813612598076025,0.2940218101733869,
 0.9977118476865519,0.9815367673791753,0.3397684406818927,0.8302619913171811,
 0.10609090033526458,0.28790623666715875,0.37198358086026584,
 0.2566686010155861,0.8502507592561385,0.3859539117523654,0.6181844715831561,
 0.7737036306805544,0.19973203143166307,0.9033312404096033,0.7421730383830768,
 0.9339647182935866,0.9930429416879497,0.684280119540089,0.2798483808052774,
 0.5679429198666558,0.5093393243364895,0.7465228703846555,0.39862082462455783,
 0.912559494472188,0.14965057437022256,0.6813779499677067,0.5761727703491942,
 0.5386349236260611,0.6822247958784234,0.24454027643589138,
 0.11326001853790768,0.3634319106500985,0.6764504533100536,0.4859049156268411,
 0.9625783427310156,0.3058818608513735,0.9631280508081469,0.32371836309582414,
 0.9286639135265322,0.13210574093346683,0.7980549829710708,0.5878729908298159,
 0.8696395869280125,0.5721509560677158,0.47914083775649796,0.281571685536272,
 0.31871756724753897,0.1760036406602662,0.9794236697574478,0.5391807910664493,
 0.8105115909118249,0.839652958711307,0.6286865534834727,0.7590992274882945,
 0.31665791624446005,0.832576950681787,0.5176527042808841,0.5280592085060043,
 0.2740146822552004,0.19375784173629762,0.21164087152031252,
 0.2772719498859827,0.4401475966106787,0.1476520463226264,0.3387102258964576,
 0.5288598160499765,0.240449323163217,0.5516581436375451,0.643053529864274,
 0.7129855330789004,0.508556467345274,0.7611467124692287,0.8826739989032918,
 0.14598128623415818,0.04754321335403111,0.8962292987314668,
 0.5980334459794242,0.840749431890989,0.8922040219806455,0.2907147728221662,
 0.3065038339404166,0.7309304496430173,0.6003868001820726,0.22008151449433133,
 0.82238344931616,0.7302696751790984,0.8660917059708182,0.016714473767770355,
 0.4966912421485197,0.8603357124451488,0.12745231292971904,0.9333482666679934,
 0.5913951026786125,0.910778826495612,0.8467664486699109,0.6706642745950215,
 0.8479047312804631,0.554007058799745,0.3961014825688909,0.3589615355238722,
 0.8069421204689395,0.3532584405983441,0.5625274202055919,0.5817269514062373,
 0.4371193759188604,0.790175932168696,0.1640749970502966,0.7308230491988279,
 0.6340132183945026,0.8312395746834585,0.5828663429184957,0.20112888428950115,
 0.8169004060325392,0.09896564614019498,0.24643452265085652,
 0.12535381101155385,0.5348654804498177,0.027648902027873956,
 0.8968565722573997,0.30732490050146355,0.819172821938531,0.1868655169735165,
 0.8567970759129846,0.10790879619041605,0.26869240254582827,0.323177788608062,
 0.7245548281032358,0.49647551491052166,0.4367438567367872,0.5739190038989603,
 0.7879242630894584,0.44406176975767,0.6755229112562043,0.8599577634911966,
 0.47959058154077583,0.9655432349932119,0.20260605166162904,
 0.4288364807443632,0.1548030661616575,0.41534841570406833,
 0.04710900894208925,0.6558090304746891,0.6195713153050495,0.6801083342645685,
 0.7499679324179277,0.37401433640729564,0.3091673815215614,0.6886443437691699,
 0.4916549010899258,0.772899742134687,0.7769472053246863,0.2791506205639689,
 0.5790491675020252,0.46027154144884563,0.16560915973663337,
 0.1460285220009917,0.34733549227851845,0.21470267067986937,
 0.30707201410983265,0.7644724382560506,0.8473667841855074,
 0.03644948341302079,0.5468711528808318,0.6064312256011083,
 0.23799041376590402,0.24329434931021887,0.6922443334606272,
 0.5369172361129906,0.16674866726331228,0.342488226730589,0.7938479170707263,
 0.9001369701303663,0.4772924181707249,0.5031844419959413,0.8935361715616934,
 0.009124665228125495,0.719027651617326,0.34187579000895385,
 0.5433348900245555,0.9380074943721494,0.4156969879014851,0.15117494024514777,
 0.5632859000526286,0.3942617464345832,0.014705639334474885,
 0.7957297468250377,0.6133717417773037,0.4707947795316276,0.11211789299465624,
 0.09159729395508576,0.8344478754863565,0.31506088812767574,
 0.7788340236927792,0.4417165369400561,0.5163566223150187,0.1165321920894612,
 0.29876873791878167,0.21984424155046378,0.67767173580834,0.08508257718124046,
 0.45529161209367897,0.7038463967501727,0.6907725676908761,
 0.13066650117024226,0.6898167462030516,0.6490466442488304,0.7223834754629855,
 0.04874040189601103,0.8594845274741556,0.7742637670271111,0.7602858432337232,
 0.7284959267820257,0.1845315324402892,0.9113800572897589,0.3863346625876831,
 0.6225345239337539,0.11548645023758985,0.07459431506608227,
 0.6651229800316133,0.37420905184637043,0.9938709585188235,0.6028459937167132,
 0.9613385442498243,0.3595336038271152,0.8541620663300301,0.12300436645125734,
 0.6217438273553628,0.6806395666467118,0.3190569402291572,0.5855385054055713,
 0.007701520031077269,0.5420325609145327,0.5668962613493,0.7800945295165255,
 0.6975327731874091,0.9554686788358916,0.5674788602381445,0.5335254161855615,
 0.0021350970288788185,0.7784895190698381,0.6680604969586242,
 0.3457292594444963,0.8783329850210116,0.18022344976769467,0.1110600834283727,
 0.9569212114414308,0.4972566640580448,0.3614274986294643,0.3680807657529154,
 0.32789440792363633,0.9251776562748044,0.9460288763757722,0.8908324138142504,
 0.09121671887790861,0.7199947252979193,0.8412442674452625,0.6007386834896946,
 0.26980973429530963,0.5079561835259656,0.42551432255221766,
 0.04076104145342985,0.647591581555412,0.33473032245162204,0.5382812927058841,
 0.7722804551697404,0.28958109057702375,0.8977430673654442,
 0.39775507309021096,0.3564533847122451,0.02160645516342352,
 0.023874708492073293,0.42846275699270775,0.9352978619603306,
 0.3315432177589004,0.08972778448884411,0.6696046757179914,
 0.11214178508354566,0.7875098768492933,0.25406976554877103,
 0.13180156045583402,0.5452839978612218,0.6095829374613105,0.9820012500272397,
 0.8894019887456261,0.35413005147924725,0.502763553640162,0.6824591857227453,
 0.16721475033660493,0.4360301022642201,0.8398703326172937,0.326676826527298,
 0.8522421134647941,0.7085692499066933,0.322343988240935,0.7639097519326462,
 0.1498918366352343,0.2386019662463016,0.6909779697606663,0.43381851371637614,
 0.5181274058083672,0.9335004660767513,0.8256861142654243,0.9275052095003316,
 0.7853247358561639,0.631894977739469,0.8641799463209601,0.09315247719500508,
 0.8763038743879588,0.5504833146548392,0.35816392253477913,0.6055892553552318,
 0.7071480672511011,0.2540526820470359,0.9629740261590409,0.382268702834186,
 0.16653320644729408,0.19372361555135897,0.33034403793958256,
 0.09064965489110244,0.5882900033948054,0.8517974421128972,0.1958348571113998,
 0.20922162967515567,0.8965228841902118,0.9523737787806756,0.895713260850285,
 0.26588513785788603,0.9523041401661195,0.1254089381477601,
 0.23149099440873722,0.0575625206380751,0.7760630599721143,0.9857663867803893,
 0.29297700614173106,0.613293433165603,0.4755212484696201,0.25793658508051975,
 0.11423959095162095,0.8328566857301352,0.765335939208658,0.16351325288070429,
 0.38732037492808813,0.651083109013531,0.3433006610936785,0.8610289600277274,
 0.7612417697551594,0.9603327872225073,0.9864250025196487,0.384326455112021,
 0.8057770718038402,0.3498430149426782,0.48306871143377084,0.5991582448965952,
 0.9330744757902248,0.14665889757432904,0.9966828404506037,
 0.28599402642249816,0.13459578593694643,0.02559045062834775,
 0.6265338475406339,0.04993065687697806,0.8034658138279817,0.6981605534944139,
 0.9914336597756793,0.2571553637062487,0.2873898044480012,0.9882214142131286,
 0.4367821211268569,0.03501165502142278,0.26264329521409335,
 0.20168706059885078,0.7623973295844132,0.9126649300339074,0.2511409570854741,
 0.6477474722070222,0.2594726940692811,0.07864276075160914,0.5283569812908242,
 0.9144523498119816,0.13895478440731923,0.2659384910057859,
 0.49950525074259366,0.5115304916027255,0.34053861679015507,
 0.17102162429608792,0.7456967219877914,0.5693707017475912,0.6667897088892154,
 0.8815069703338292,0.546973931104652,0.6551340067015404,0.3437264610873598,
 0.08939983577688881,0.9602749948139003,0.588096878942793,0.3186530792796275,
 0.9762941362084664,0.22704400243207024,0.6632727551165266,0.3276903341833338,
 0.5502448448191151,0.6481502047159606,0.41266205080338914,
 0.015248986918077412,0.7065833252619502,0.5305478936192143,
 0.08138501857498959,0.7965153670114521,0.7029734565833481,0.8226473771453419,
 0.49487845658041385,0.5831095955958603,0.15467904546821665,
 0.2554854989940063,0.8450672079115764,0.5648629967716639,0.5137978488069523,
 0.3267324686410933,0.2868779200640337,0.7563148542342508,0.05368394592719439,
 0.48368737306317966,0.6446164872923135,0.22960434992625223,
 0.5174611247350986,0.5600634749200821,0.910115641065763,0.8307650124563386,
 0.013575855846323592,0.03423512401649087,0.633235796964105,
 0.8823966123261315,0.13876874088558866,0.11464338175143485,
 0.14795532024507896,5.728150026527157e-4,0.4429057313923447,
 0.23938809312519105,0.7807982845434072,0.47107337315095044,
 0.09744986971262082,0.41842323434282436,0.5506860341976617,0.881704927379586,
 0.9754107940419039,0.5306146596095901,0.7261253619692911,0.24893211577961372,
 0.2799855468476453,0.1117667635734183,0.2989258831420638,0.13483184215378463,
 0.20291907095665473,0.6858384515632231,0.40973162899725535,
 0.6474045847154752,0.45734790724602137,0.6701015626484425,0.9267212911996253,
 0.8403728382512072,0.15954621591462603,0.8847755529415536,0.8128349028862683,
 0.16163346152165148,0.22256491088421115,0.04945110863782187,
 0.6365973445239826,0.33053482176790205,0.23544903731086464,
 0.36960217007036034,0.7513363795424751,0.9792017041822396,0.9928031094736792,
 0.31380816142587475,0.24531773608407093,0.08611552980103099,
 0.6543224275375714,0.32152811932791137,0.023827278621261994,
 0.07479867098097848,0.2231474150053715,0.02101020002891274,
 0.1945492910534733,0.4134653490836203,0.7704566651517385,0.477338683385661,
 0.37885745056630493,0.03971514586420022,0.6398109963684002,0.655838727133018,
 0.8817143772211044,0.5216082241352268,0.4392139439193026,0.23189364146049063,
 0.7289439079101181,0.8585406678683196,0.8361302148548502,0.4043732095038377,
 0.7629959947643693,0.5379658019807153,0.9816122358495655,0.24439713241249916,
 0.857034711187326,0.14555699327938898,0.5165165098596481,0.29131272396796415,
 0.40297074897057694,0.8762722012192757,0.2750461943319822,0.3958887908622625,
 0.02802540759229033,0.32736778884665496,0.6640253671646964,
 0.8690164061227109,0.009899064750426234,0.728741081702839,0.3883794719500193,
 0.18434406599669917,0.9538340238189823,0.00790894143872889,
 0.5938858687778464,0.10605933351316332,0.04352236836283219,
 0.27212376356014234,0.668409169447487,0.6681619090977015,0.4726642123032472,
 0.037396576279251414,0.47238774894597346,0.18551174452544295,
 0.40114930819086314,0.4198947096372101,0.43304528200501147,
 0.5026436591243328,0.15761425159679066,0.42420318667295753,0.764345451298062,
 0.20548389853296656,0.6314357647046991,0.43225254288948634,
 0.03196633917771341,0.9400246074050467,0.5200049703135879,
 0.06829915171979783,0.4618309973316914,0.009402428504327487,
 0.36886627920062653,0.6848318405512458,0.27837713364034233,
 0.9772012528249306,0.4427438729629841,0.4565788873857536,0.7045267094055322,
 0.5470754697076584,0.5633096669016666,0.9358325085638379,0.28674519325316206,
 0.5789334176447262,0.24669535209395177,0.4540753409503483,0.9422597402919272,
 0.8660710425297871,0.976159587059486,0.6313382379704344,0.6666471400745204,
 0.005870972459455781,0.6254598658703462,0.4235131780583212,
 0.7147736972716894,0.38203526677297095,0.7809302259715568,0.474059650165194,
 0.1261782984427846,0.9306755022496306,0.46090661687871304,0.7749290879161671,
 0.5478280690720574,0.18295828535532865,0.3320003058343568,
 0.43113902732371867,0.5940183885884208,0.002513606349083819,
 0.2551392373444764,0.945983960567627,0.5148367886034602,0.7068236139976323,
 0.9758589364548861,0.14152860658086153,0.9391272959666918,0.8329373915142751,
 0.984025650093821,0.873375007908434,0.4395952510205955,0.7004869890009313,
 0.45467497533415,0.3826542728201554,0.5883692231847807,0.8661775649449097,
 0.3627973071278905,0.12590446513076547,0.09959123325743346,
 0.45340623181090933,0.7490262816800082,0.7358805041838185,0.2257854154640232,
 0.13493202207963995,0.911820176242075,0.955722157007804,0.003735536879936996,
 0.96818219681574,0.08357586009833262,0.5077793067337022,0.13448546257939253,
 0.03879198468357303,0.4524585201000684,0.4350570130933401,0.807354509662382,
 0.4505034365192774,0.6483635845951894,0.057913842866745835,
 0.21371422055117772,0.769055203768997,0.4671329198105023,0.27068989542554,
 0.5253694793354797,0.8510110573257814,0.20812364289545415,0.7656287238313205,
 0.43907712924321585,0.4853020865676383,0.9567117808199995,
 0.08406377302966317,0.8952519246620154,0.7102661753257388,0.558615309735656,
 0.22190044960819688,0.8254950853740377,0.3381342142267121,
 0.36435595809878174,0.9840714003199382,0.23116937293686535,
 0.35035748633100416,0.6354655697994415,0.6671249772201514,
 0.18317803731533444,0.42144335938190547,0.9965950942838495,
 0.9380548125426567,0.799888700859607,0.785946986310499,0.6155452092335666,
 0.374397619360348,0.19974867492932424,0.3087661927388462,0.24895050528275575,
 0.6476763628745281,0.6549626323880826,0.3031272051914896,0.6724117876130943,
 0.9281967322122361,0.8306646366792334,0.5022199052629865,0.7565512862035942,
 0.7574641440733091,0.9844126809988197,0.9607768089303474,
 0.038732142196212016,0.6421037933974697,0.3362115082867988,
 0.12203484012555776,0.10129938043549791,0.3649104610951921,
 0.49810086950639454,0.7652035520655863,0.4894040966243334,
 0.06646896557454629,0.1617272069060417,0.8757875529430801,0.758807661464964,
 0.9708641164180377,0.6305320084943178,0.08603239531705931,0.4370574314334701,
 0.43928356018870196,0.39309396601925783,0.2795788837917893,
 0.1322634145985988,0.07334735944192505,0.8231619526435946,
 0.10338598749476902,0.1359851876169098,0.14358289391094825,
 0.2108878512977751,0.3020930420598673,0.5959117353366679,0.1993642534345842,
 0.6614743038587392,0.28662350161489636,0.9604552787429781,0.9175042796645553,
 0.02510874258572937,0.02594259424674661,0.6466514959417026,
 0.5010666566451444,0.16451068326230223,0.9052101346239536,0.5796429950248986,
 0.03472937164064305,0.03660382032158238,0.8459294243671323,
 0.6912028924774198,0.33877258105187336,0.8439673759619467,0.5913140883849787,
 0.999019499040902,0.06448470491893865,0.6930622415655409,0.20650075403057344,
 0.06594182328710396,0.5282433187908369,0.7117159912965265,0.9119577608940548,
 0.6469787314211651,0.3817574991413981,0.7419205593982972,0.9309731825012986,
 0.5797841437008917,0.33637457235894686,0.18427689545432668,
 0.8761440501278323,0.6429501178541778,0.09596746247962595,0.5168681157592929,
 0.4178641367830318,0.6300427031775792,0.6367201592271623,0.6972084535987293,
 0.6438133510086448,0.9942181683653442,0.1317886600919218,0.4682488160362226,
 0.26484289588243526,0.4748100629182671,0.02545582941856328,0.702319698828045,
 0.7069371034617502,0.9105423601642724,0.18615258462131368,0.2970192939265771,
 0.13970604652212004,0.3243216998600983,0.04703274750273523,
 0.37771954613318715,0.7697877492034172,0.7131704339054848,0.895058343140798,
 0.3144076485409625,0.08101265839612515,0.24170854915286322,
 0.8744821320972443,0.7748929917833666,0.7925162424223023,0.957312490004294,
 0.8408257130119162,0.9063217001387812,0.8998585706521282,0.20573896025949523,
 0.3252174276865867,0.9965525903869147,0.9877603188281704,0.411371272263496,
 0.5100998849122053,0.5391185542365499,0.7970158565990957,0.8043290197669558,
 0.0940635228995893,0.693526944969676,0.5418000046589072,0.8733168700217099,
 0.9948226608168373,0.773255917970017,0.7024568810007525,0.8388148342138025,
 0.30983264576652403,0.3902772348772844,0.8079442842207488,
 0.03429000129866999,0.773041916494769,0.2156969083362109,
 0.039120285807012434,0.29261406201769524,0.48582190507765843,
 0.9265793575089354,0.11710505284236916,0.4542126909715378,0.8872026424067699,
 0.38018491157779366,0.4144954691861745,0.015838808992013398,
 0.7655817602548627,0.4140528868844435,0.5603569983290431,0.8267401990088572,
 0.21484606636718706,0.542995381227042,0.20051356923760622,0.6992643647109128,
 0.9383493022620703,0.13031158138244103,0.17044860482066615,
 0.3378490508008164,0.6435084724049085,0.4640051354107815,0.8928230469452211,
 0.5249142422058292,0.7556625785262167,0.41250655822251103,
 0.17565149378112388,0.8460992575549684,0.6410138267554969,0.6280805650956147,
 0.6722048212601075,0.8702820345794474,0.7935406082190563,0.05158498421944169,
 0.20184715336393655,0.18497587283227312,0.47733659378771087,
 0.44565167620282686,0.8666597645674589,0.3753434958924262,0.2963328062463373,
 0.5830321419777618,0.4647920333773924,0.4113697539948459,0.22810060138037003,
 0.49276415800031537,0.2628206548945661,0.5876262459132059,0.616017660492453,
 0.26023472879678256,0.14469738111170838,0.7693535372432977,
 0.8529710734303755,0.7386841748443815,0.34294279359481905,0.3057292344980489,
 0.7120451494686857,0.4653786333098746,0.4992212079394819,0.1555747468918678,
 0.8635254454903278,0.45586619794580574,0.36690627064307924,
 0.37317755261555785,0.07931018738418749,0.5510651554747232,
 0.21415461043986406,0.08196685636310652,0.9032231144684132,
 0.6343722448809631,0.307336204781695,0.2569609719100161,0.8223251868577113,
 0.6489147028431447,0.16046023061116366,0.41713452882833946,
 0.8073836349279017,0.7868908262757955,0.5281619774545208,0.6004249554201193,
 0.7916135496824176,0.5847466975740355,0.8816854087004351,0.6202518522877951,
 0.2541793130464264,0.9634123645378279,0.4103747721095907,0.2708105462061495,
 0.712094276097406,0.9416966240805908,0.5243211322813005,0.30338562944891767,
 0.13166447848556884,0.9848944557924394,0.7398628942116081,0.7222850332913092,
 0.061355596100886034,0.9981019819151435,0.1585133553390492,
 0.032659439351604425,0.7472132642429494,0.5315006453117623,
 0.2842555397046663,0.5482496576100006,0.9251133898019337,0.08710794875132044,
 0.7649364021577323,0.05725377264091791,0.2732864118636661,0.4096902673880204,
 0.8789566283674597,0.10322555642532671,0.6615169017031003,0.694291901806005,
 0.12666944059693164,0.28837352682359463,0.14625591594939047,
 0.7958430072976095,0.6038960299902293,0.2088675476314782,0.6659996476845914,
 0.34883272653455843,0.21731860700001293,0.9603800364575541,
 0.6313884191185791,0.11773782209787509,0.022046153172364802,
 0.9805937318342713,0.8169860867228182,0.8350467382431104,0.8454851774555658,
 0.264689035360234,0.272841805779482,0.7106236761264644,0.43836337218788934,
 0.6045650286361517,0.9196788064048743,0.9908037071861502,0.744327140007462,
 0.9930198522339723,0.3002953978064966,0.29147317719232857,0.7806941192130782,
 0.3427160571249639,0.09356542197159678,0.17837883756985295,
 0.2995519167739642,0.4862307892006964,0.751238880114467,0.44275718508006134,
 0.6366611384470442,0.58169298610552,0.336840959864825,0.863767243331518,
 0.09740082769388125,0.7314877951595592,0.8085966630962687,0.6540563160641046,
 0.3677864511642306,0.3298796281242484,0.6757012865393794,0.6702336281295941,
 0.4951924918727375,0.008509830685667574,0.2587335716973861,
 0.3203125926686857,0.3353655477944002,0.14763980432758084,0.1406358935577845,
 0.4822623463168031,0.6780636487373656,0.37662243296233955,0.788594748203332,
 0.8019662658222421,0.22497280502725392,0.7116209983976998,
 0.15406606364207953,0.019130524169250318,0.5563587872654054,
 0.19267749495004072,0.46210630708543565,0.7860829509206018,
 0.7244386235853103,0.3007401192020558,0.6234562371446991,0.45207998142390915,
 0.20829920650419964,0.3400665019615533,0.08601306757156002,
 0.6523933301286728,0.6742356616283098,0.12730154624051315,0.7733446211659658,
 0.42682939067068415,0.0016351866334088916,0.758447532318464,
 0.631383767980159,0.8623550817158278,0.42208823647697225,0.5891350052397133,
 0.5510422925381668,0.9961691264674707,0.39068198299275825,0.9435555205241085,
 0.3910614849657471,0.0964415463491386,0.5828723340378354,0.6416764419789031,
 0.6802122449698872,0.22278235752802356,0.17084326672466776,
 0.5557475560199132,0.7518421069462211,0.14835730486433818,
 0.20101394130535755,0.1410629030229884,0.2580688337040635,0.9387862448225898,
 0.2874000361573642,0.9532922770315478,0.9615666209005252,0.9012171789590577,
 0.6268989419176304,0.2789671205680868,0.9804920482700294,0.5996642236148195,
 0.024547393142340335,0.8578517059017275,0.4079248122185479,
 0.3062880436830667,0.6301672452764575,0.6412192705893508,0.4586211366823958,
 0.470580838067728,0.706435880478178,0.7955961065125061,0.6200385197936155,
 0.9787466786493885,0.4660090873162428,0.39347639491181685,
 0.21731206333977293,0.08251436042944726,0.09950282401316668,
 0.08917922475027462,0.2813439209501888,0.08257210024611517,
 0.4390685624644374,0.07536657481968412,0.3909606149498759,0.617523277683311,
 0.9490932327387889,0.9294139119604364,0.23665173998249045,0.2896853271331945,
 0.9237113379098479,0.691092467389519,0.30779853929908585,0.6870945336541324,
 0.7360226816684846,0.7488750661042104,0.8079220210774383,0.3145964045433086,
 0.5270763677537883,0.05831644627439492,0.05825736178065877,
 0.9657560481833717,0.6055958634175242,0.7783052601439124,0.1439815874574344,
 0.6651639858150348,0.7742005813550439,0.42724278081124556,
 0.21887834365240932,0.45225829980598853,0.399345079089122,0.5820615196340453,
 0.27970074527467537,0.11767522210920267,0.23388847182283057,
 0.35052961438513885,0.557386737261246,0.5747594224105405,0.3372444310780023,
 0.10021321630401636,0.16369089248348412,0.48468922656887437,
 0.40654847982677933,0.7500548110669489,0.5272397633649493,0.0886047839951345,
 0.15775159518336024,0.1433741978193115,0.48500119739524694,
 0.4331073100279438,0.1360939349566015,0.38448078709988,0.019010690408916897,
 0.4443994893044373,0.43265177230193164,0.19280023532584667,
 0.24403059293261276,0.6837230228950328,0.803825032421718,0.7146213726426286,
 0.20341214930603757,0.24513456549631008,0.5897473988842252,
 0.22057093675864936,0.06556404103568059,0.6888267353084685,
 0.6029034115414625,0.7624682893026525,0.41432271159205936,0.6877694883131094,
 0.9013752957431128,0.7944185144385333,0.3207600224607777,0.4474187975242576,
 0.6339062989248984,0.09673767979358638,0.6062227618098569,0.9376797618956128,
 0.5776245948804963,0.7920733041746768,0.36405391240628704,0.9031263248951091,
 0.7718243300748439,0.02007009399349935,0.12924704843909607,
 0.09389163993210337,0.8829395563653286,0.9150324264501157,
 0.053563684948402246,0.3108961882277481,0.6784932793669597,
 0.03416301571531899,0.1431870456744343,0.8004225551095221,
 0.022785542681330373,0.419650158302319,0.7917419387076109,0.90921827756168,
 0.16270729364652037,0.394802301183771,0.36478451107023424,0.3551234781362882,
 0.40942763284059436,0.7869049531894672,0.26152379765831457,
 0.5305104407332113,0.7865619074177953,0.20117990263732222,0.8883004174370682,
 0.13938785514043817,0.6370323502988693,0.6558360080113119,0.7867691633925933,
 0.09581660296889749,0.9421712096233048,0.7333386240523295,0.3098231355466714,
 0.14732010797474815,0.6960704694221611,0.8981930271824545,0.8680605748969787,
 0.9163877358417547,0.26568334889544565,0.8869505065191113,
 0.38388675900853175,0.370253407951316,0.19499196194857438,
 0.44125241215967725,0.15392939079781676,0.19971792375775266,
 0.2723959689415252,0.9974896625346493,0.5340253395526724,0.07494184845552965,
 0.8678211665962249,0.4218013074146755,0.9635355143239911,0.17236926683715736,
 0.1326154239157208,0.0014949309618310647,0.12398986524347155,
 0.8450228345149355,0.5471787773490219,0.742577916932022,0.32939359215943764,
 0.4081334316534817,0.10244690392906097,0.8686006605417554,
 0.19587185830506137,0.893325947573185,0.18716758579365056,0.6726106321244864,
 0.8420011131745819,0.03294341424193081,0.8273939917884358,
 0.24341308712810616,0.7329078013201428,0.7044359515443509,0.8202619853019407,
 0.49975913652720316,0.40498363563733597,0.020065230966187397,
 0.4703927530678491,0.615126062218751,0.46663762566844913,0.39473223809718416,
 0.09351790146733807,0.7553722043307144,0.38525332258021794,
 0.8436859599387992,0.3155656470791278,0.713377634299866,0.12920459335834555,
 0.017476464351006138,0.5839479050624501,0.1086770201503866,
 0.09237426108240843,0.46418131440413013,0.7022969010392562,
 0.5561797795311333,0.9706308578716084,0.25898844191058423,0.9464225429128181,
 0.6087895902593932,0.15105476710778842,0.8548213573687784,0.9551441852015105,
 0.22321418864221998,0.9034334703300139,0.08372514114874674,
 0.36076473517545193,0.3585782472392147,0.2823745128788768,
 0.32561224653763066,0.655477027497672,0.5027693561386173,0.39891878815884463,
 0.9416462313307762,0.551961047761913,0.44681881933924705,0.8275853263628171,
 0.8301671870903313,0.9121732874756558,0.7480862464216195,0.10721241829106765,
 0.900154760880784,0.5473909433396668,0.04287933538935029,0.550573471959947,
 0.11661387653212674,0.35133253223953487,0.8250219529488404,0.55913174478131,
 0.7849487115066358,0.21436865639886293,0.8898954090922417,
 0.39499767654198803,0.3554119154498413,0.9404410536070925,0.4957654846336168,
 0.9747327841756092,0.726620601415507,0.3787335566175035,0.012133740550648087,
 0.9444117605150734,0.11753322378520448,0.3576598434329574,
 0.02356593672611551,0.002449546366050326,0.3910274089840615,
 0.1725811938338988,0.4060501615146227,0.9384230407213221,0.3837410364939654,
 0.9858735962993098,0.7595525458537162,0.20156080746815352,0.8855886080241232,
 0.932797302139044,0.430194966394261,0.020178559193763212,0.31889507802884276,
 0.21179491092651748,0.14713516951918648,0.9119016055593365,
 0.9425794640518844,0.7817841861499528,0.4929816039039082,0.8267349975016219,
 0.4033838753744008,0.43849886872573574,0.7354925114698295,0.6319046707151263,
 0.03579261145183765,0.8242127721510428,0.7916735668927183,
 0.38844765551804916,0.13715983982790425,0.4138295038633676,
 0.43306604273770477,0.19608443707911793,0.6501562481572764,
 0.3770161484349088,0.3820387622704241,0.6265808114259586,0.06926945136856144,
 0.8929564153690447,0.9455218062583743,0.10658737128152418,0.4984735941907318,
 0.4641174808583687,0.43410243350565714,0.7854175906338021,0.4818676337770732,
 0.07389064355306674,0.4486019555104187,0.06737306894540651,
 0.11481674895427252,0.23894391931657877,0.3531713353106892,
 0.7963632471276129,0.9246820740317181,0.25403151914825295,0.7442649141864985,
 0.46955101660937526,0.8189275498131451,0.8839845758418983,0.9109816906343072,
 0.12472924901056981,0.42986643227511023,0.22199592102291898,
 0.22601417965341364,0.1856906460864448,0.9341575340716136,0.8177340238161378,
 0.6601528462326978,0.3231909080043345,0.7750261500834862,0.06943245238737195,
 0.8271375179599686,0.5295025833644534,0.08948692036639061,0.3966889110406774,
 0.999076282714669,0.4939271362771822,0.9243022813654802,0.477499054757057,
 0.2833034204000442,0.712433535803783,0.26685707173423456,0.56814526067962,
 0.3511932546674138,0.6396191758342888,0.7090167011901636,0.8436769945855108,
 0.5063682535822172,0.2162707419777513,0.5764678103194559,0.26013895868102466,
 0.7999286860671264,0.0301345078829649,0.41362481049486477,0.740608279795989,
 0.8863557495321227,0.9712896226601327,0.6655327975790172,0.5315432403387375,
 0.5172595094575035,0.4821299599400938,0.7837297952081401,0.15996936524369287,
 0.8542393643006684,0.4583262348860745,0.5952679399629426,0.7182592565297079,
 0.49736961096769117,0.6477092374814913,0.6764698858553487,
 0.14600091324841546,0.38444288107274216,0.43922851562185516,
 0.07544779571573434,0.28247183435392387,0.05978796529032615,
 0.5378264061202178,0.4884717842913677,0.9938601790251781,0.9864142691300963,
 0.0955143688425436,0.5185829742872713,0.1652761001866503,0.7367886979660141,
 0.1158433675430881,0.9920283345356721,0.32825418583059607,0.5056546192660366,
 0.7406576724276318,0.11487634023386639,0.7592030836192609,0.4884926320544287,
 0.35192992712445503,0.34341309330480774,0.10003038414350152,
 0.1402689285855644,0.25460337828899504,0.7732808433855161,0.732390311304686,
 0.19281086374989742,0.6455387228160189,0.11362400914223136,
 0.8849736531525568,0.3159246494561926,0.5522210554489819,0.8108359783733057,
 0.4315072276859153,0.1927078613437283,0.22695326798501914,0.5826857045982129,
 0.8450609113098064,0.552976250240201,0.36653993204434654,0.24693873140497302,
 0.10174737086758168,0.514404610040547,0.06860795008950149,
 0.27538670132798615,0.3322072324501656,0.9089392199650506,
 0.02732960448832422,0.4204438600466246,0.7767528480890815,0.4120427972592525,
 0.47955408459093773,0.40508875408164313,0.04765149908611854,
 0.3325912840525419,0.9556432261915698,0.6202734225056976,0.887247715416428,
 0.6409763171375346,0.5012271351433962,0.24081517720243562,0.6951930541905544,
 0.04973613920787756,0.9909878845157668,0.13960109771989837,
 0.9893520833182393,0.6169847171052645,0.7204590910366564,0.6570684330270831,
 0.6148019892365372,0.8377072926945389,0.20738451686346637,
 0.18711373309868873,0.30252446922595266,0.8110508220985335,
 0.8927821299916793,0.13658009396437876,0.8829205696052417,
 0.46586510632743305,0.38536424241637623,0.42110758844134133,
 0.7727347750298353,0.5190931305702818,0.410591428961149,0.9760178566491703,
 0.16252944602654507,0.8387794803920192,0.8885190354689342,
 0.16952511029308182,0.40824816181304024,0.8579513076040569,0.594184070090333,
 0.26844722173972413,0.6504053324949018,0.8849216769235024,0.3405343222969873,
 0.7980153723970907,0.7893883020187197,0.11834355672817609,0.8965511209353463,
 0.27878935254915216,0.5930361392949673,0.5871177118131647,0.8611773366271775,
 0.3214737020802453,0.1735223224010064,0.21962369220657352,0.8971553848322753,
 0.5681486881047912,0.4567564564448109,0.4628410221596666,0.7732908451346772,
 0.40174990407231426,0.5067691436878106,0.6736864658339892,
 0.42903657094770165,0.6523876828262006,0.011802896565144327,
 0.24484339095888297,0.26613536296108065,0.5847056077070694,
 0.6839083118406128,0.08718508455775176,0.7249251723308838,0.610281486175021,
 0.9027530510513586,0.48173950668703314,0.07047961059724484,
 0.7115848635293881,0.0553821454092549,0.794195577250564,0.46116937182455653,
 0.749591880866474,0.9228297125358373,0.21760055537342848,0.6015463469704343,
 0.4898290388383313,0.5480863375404617,0.507549955722562,0.017376002048768213,
 0.5315512028484514,0.39277686547949175,0.2754790065440935,0.5169799845821066,
 0.7000294675919676,0.27709446514499947,0.12153245776046084,
 0.7468332284001236,0.5564496155332185,0.5477325339651041,
 0.009255712643233216,0.9821360416593115,0.6360545763250042,
 0.9173503922006631,0.14862930174162392,0.7211215264400923,0.6113647020034274,
 0.2295212487660927,0.2022374659484094,0.2940596945922622,0.1950415593212811,
 0.6004070101585872,0.28322021400949515,0.6164359868198104,
 0.036785046950197686,0.48172921659901824,0.3984409174761123,
 0.22511030405338317,0.9730031515177939,0.40097149403763255,0.802467383958979,
 0.8612193038996651,0.7596306467729929,0.24818297175494297,0.6670816725293732,
 0.8509167657091774,0.8223048967296742,0.10701434204282867,0.4821007739438534,
 0.06940677817770735,0.27345256461770173,0.7912775246984368,
 0.10248495374019162,0.8718441156591297,0.18738063373195502,
 0.6361295478746234,0.718168874277129,0.20325055470288622,0.4896411406684563,
 0.695190294082035,0.8114037908535029,0.5415934006288895,0.730447781606564,
 0.4888275194758649,0.008502774847566674,0.3779912498079391,
 0.5207685848605307,0.19517174955572958,0.7829216858444095,0.4603665756733444,
 0.8233260958995204,0.39129448792536026,0.7651633823575723,0.528951902481905,
 0.8802575732322225,0.028914017554991123,0.48183167706394125,
 0.25777819626110543,0.6305945821918689,0.9672273163833519,0.578235212297005,
 0.21075360338719706,0.8952126707341372,0.922965067809699,0.7406467205897976,
 0.022316457848456928,0.9352539212836851,0.11434139393332599,
 0.12221921920594525,0.9534050146336599,0.4666700989396638,0.867780095329376,
 0.6869999215574116,0.04736510510215308,0.9392661763480387,0.8357830840143681,
 0.4008065697450691,0.49029416429360106,0.800642934449013,0.7582986717874742,
 0.9223394600912573,0.07569238515017995,0.17470663206951176,0.697484223591301,
 0.16284628864368877,0.648693226652494,0.9660568094800117,0.7607145846796406,
 0.14826902975251544,0.3572879178413051,0.49321479334830776,
 0.08617698968937171,0.7648833556715351,0.8478024919287576,0.8447966088849422,
 0.11021788773498598,0.7509815602881195,0.8430548794897202,0.7159375244791552,
 0.8175085396549062,0.40388950879676455,0.3725854774312125,0.0589274692691637,
 0.08350367853474994,0.24717638198934355,0.4956101659855773,
 0.4875139508358789,0.0961230208116366,0.8503754743545745,0.7282904855409161,
 0.005113238778022278,0.22326739884765823,0.8818527164626278,
 0.745749958436687,0.32505138346160933,0.9564870325631855,0.25788972298829216,
 0.6774376124189373,0.4073025512138696,0.9960528485552632,0.447739223864438,
 0.7159650540464841,0.9253105889057431,0.7593813707772056,0.6800766380532879,
 0.36782783716035006,0.8048978741547633,0.953221372682503,0.8869066482500272,
 0.2565396129145493,0.4005305431568159,0.4572675546904864,0.1176931628372434,
 0.5017217976084885,0.3105095274553967,0.8217537438615986,
 0.017602610968164734,0.3547570658519478,0.9456550674562917,
 0.6507948200472635,0.9801241997899067,0.27190664770563466,
 0.08647238416321756,0.3812346660307796,0.08977459901219842,
 0.21240865898885575,0.399175228017709,0.5678713691565211,0.4719427607644512,
 0.5939624995094488,0.02588126867719298,0.1496768177694553,
 0.43003630919723335,0.9780502892970235,0.7619858609383496,0.9926292684616081,
 0.35162754160322596,0.9319594643659602,0.665717176083947,0.29264158853037336,
 0.022242038458525792,0.913302187342701,0.531038882356942,0.8089433764560557,
 0.8232586966849021,0.45993416920742014,0.41429240044819315,
 0.9352968359640845,0.6407376660549704,0.8356526211685259,0.7572397855692564,
 0.05171744018740099,0.5831823825144674,0.3999460828045214,
 0.48758513857550667,0.3376475775185548,0.9432505092716188,0.6428158821675487,
 0.9762605283294828,0.9305954167725264,0.5552099308949531,0.4086570666580589,
 0.6661811909799121,0.00978834707667775,0.17229067660470188,
 0.7571233422178041,0.19758249480955636,0.48775187157934885,
 0.06648378654019871,0.006204923239540583,0.10520152771459235,
 0.6775752801437501,0.7182849399118375,0.2243165834770644,0.9283844228782163,
 0.4320439158883531,0.9143109486021209,0.8721054627086517,
 0.030647201322087536,0.07912221799788899,0.9066042004895649,
 0.6043012734025518,0.5063553573113206,0.9106350989901233,0.5267622921401777,
 0.9522350458634496,0.9478907918694939,0.28761994972694604,0.2606052144778501,
 0.6605012686377273,0.18719274383165074,0.9564987191965212,0.4953416274483843,
 0.6702731882350369,0.42718021750326096,0.638095655474798,0.41605092474493355,
 0.22818878817969035,0.4558593869415509,0.8020850176771929,0.6535968662044771,
 0.3705882811017098,0.8976607661348723,0.12769075383674466,0.8298801639787186,
 0.4474107892894067,0.939444195906685,2.581580800490313e-4,0.3919456700280868,
 0.7342102599765097,0.9327078968265541,0.6201357109448576,0.9335317258214995,
 0.21304096558062924,0.04127262347252414,0.47025449761375526,
 0.7755430605841769,0.34267002682269454,0.5184230946051572,0.724801466094539,
 0.6105653577800231,0.35879489735583425,0.15548596109988633,0.973020258434305,
 0.7261849448903225,0.8162093108566966,0.018822888814419514,
 0.36487252673696835,0.12980768826026567,0.3940399036446094,
 0.12971326315334653,0.2785370562420808,0.8723791662643845,0.5961487561665895,
 0.2963879033163793,0.4261199764834258,0.2932454398713221,0.2502265281192865,
 0.10090596896497861,0.22287092079143933,0.9654684157936582,
 0.008470337828235697,0.9148503895543478,0.7265111293206751,
 0.7336181127469368,0.3055203840326103,0.3763935282394989,0.2800647028428461,
 0.5470058982294876,0.6673151573577009,0.584502453461205,0.5976295193024675,
 0.6350190436085549,0.913373666920902,0.5602178544426951,0.7287462732628933,
 0.1419492981235293,0.9806453737624838,0.40550887351468456,0.305052182359706,
 0.34707173681821435,0.3278573050957845,0.3428035118902848,0.8530645306595791,
 0.6885542436258832,0.7602311425634224,0.8160817798443986,0.16994274491181716,
 0.0502895280633151,0.5837137877810248,0.3214404199715014,0.4684868556653621,
 0.8666457271049279,0.7820607111723454,0.9016815098025397,0.08943765341414756,
 0.9570805416297035,0.017563784814836114,0.48871898962165905,
 0.1946549952700184,0.4911861903914244,0.15236501139022962,0.5367990943909857,
 0.7049334563529793,0.9669861110854934,0.988383149903168,0.8583364182845427,
 0.18308540851471145,0.7717476485611243,0.057506299209298395,
 0.048694519726534935,0.7253082847613914,0.5855347349091203,
 0.9094660423876924,0.06627548474493294,0.5999999761888004,0.2727870400468215,
 0.05032498316788958,0.42463106082423874,0.2794561939688207,
 0.5540011709891342,0.27460936568041006,0.4942765873041254,0.8583202981450159,
 0.1166705181332105,0.4068972589472416,0.022960878658333073,
 0.6308361841712055,0.602114667846068,0.49441286994003697,0.5974436657806073,
 0.9466823468938503,0.3360466025508204,0.7250219170249989,0.7534313377342164,
 0.49942278363438486,0.9612826381118178,0.5583327304724413,0.9122921178637271,
 0.47893018818479616,0.31025879253942623,0.538711230917733,
 0.35358012931645255,0.8974461090854046,0.9200180473522263,0.5136396426129796,
 0.7897605400218761,0.02253669711968309,0.3213183358232079,0.8816421643383962,
 0.16121333777340552,0.39607381214522497,0.6340658907657213,
 0.2974517811102311,0.1859227753179059,0.17988523971502945,
 0.10878740710317314,0.20847387438353848,0.05365500031205084,
 0.6817771913429651,0.8427082365707057,0.7079688475388546,0.670445543267791,
 0.4732601019973087,0.7583354627559794,0.83698740441562,0.8656610265148695,
 0.049226420847792274,0.3885277054587638,0.8479065965961597,
 0.4748605501706491,0.16767536429651297,0.1591370025428287,
 0.09170429231310429,0.20024174331017086,0.6829038296231317,
 0.19117357510856559,0.11164352492738372,0.8809687995782106,
 0.3023235504639079,0.49433645029423723,0.9749064942294099,
 0.39471669095766937,0.6055472925759917,0.828103635448639,0.26450652530737373,
 0.6222584193997533,0.4621834848604629,0.4691551330259105,0.6949488583967655,
 0.22600607026722674,0.8195816323074933,0.8069972949298613,0.9022832200808221,
 0.5878922686542665,0.8021795896678006,0.737952331163431,0.012502527195143442,
 0.09221887704784903,0.8841528447956291,0.1689672838065306,0.5801746305426313,
 0.1279248816173303,0.6802832733029368,0.5694054255316623,0.4276489833265118,
 0.4451987839077777,0.8780899822285746,0.9817801101025224,0.48217551497418454,
 0.6397501985981814,0.6201861933472393,0.72161570187439,0.972164794651839,
 0.2879825929137636,0.15051237701264197,0.37572290351053694,
 0.20963381229846467,0.12367880278626342,0.7983977828851598,
 0.49005882438374093,0.5306644154161686,0.10261046001397656,
 0.8550107020170039,0.31675421290133,0.4622571553952579,0.6516822477378038,
 0.5090401244483254,0.024018004553733152,0.08988918742218543,
 0.1556645999809213,0.1190422639719747,0.9680773236154245,0.888770845709594,
 0.6983438072835044,0.5421948150653383,0.8077614332287671,0.11210122160058256,
 0.4684131950244005,0.38017560636702286,0.8325055732307127,0.6582192685279757,
 0.6897246526379649,0.7649468889278997,0.12361104493508701,0.6475075133186878,
 0.8469919010085878,0.03008711244853224,0.8927112815547485,
 0.47219449039205363,0.4307694665281985,0.5979431518525725,
 0.09476678663918325,0.4725314145790741,0.5885057539008721,
 0.06063650692138367,0.476600722753468,0.3598094043942256,0.3926479343695882,
 0.1310403582903663,0.022325698480953005,0.3651639141510016,
 0.8535026650796241,0.21800679782087584,0.44595832484972175,
 0.18211369254755405,0.279251435645788,0.34799040764965594,
 0.03297547745913365,0.9208714436319446,0.9883702778157472,0.8036881105621416,
 0.8805409140282705,0.482249386852035,0.7202805404602763,0.7758417031723353,
 0.07543310599940578,0.5883891648735118,0.9914299066749749,
 0.32121731256655883,0.09859852394365531,0.16407294443403053,
 0.04078477495632293,0.3658711390502759,0.7755450608169767,0.9986789005209609,
 0.45062635015105323,0.6608290245765663,0.12765424499758915,
 0.9669985328141449,0.7142245857325116,0.04302385110398954,
 0.061914143223346985,0.6926883235541921,0.6936373472583678,
 0.1032129843363756,0.3225103486937113,0.03967858255589363,0.6265347927296228,
 0.6578603181854777,0.22091832466082328,0.1915145413326551,0.4031711384810138,
 0.5120422969487833,0.6629114903031789,0.9273042507998078,0.922851725896517,
 0.8889525802046288,0.3187145546097614,0.30387471105562724,0.3854473384907906,
 0.8208912012521189,0.9731421417213035,0.6144538619983606,0.40495216388226385,
 0.7622120450193556,0.15124756205869927,0.8668517419417634,0.9372191301020192,
 0.4482689813124081,0.6518333288517868,0.5080799003619523,0.856589536318206,
 0.189376269108967,0.9736392563710432,0.9838620318824365,0.2751641922788459,
 0.967777742351624,0.37355162261679076,0.8065817128225615,
 0.020535992791758217,0.00462032555248304,0.35060324069224413,
 0.417387134666346,0.5162841012659842,0.7958349098598005,0.14802155143884443,
 0.41357901366277805,0.2505672048793006,0.4257406400177415,0.9767393079881259,
 0.43684780054001593,0.6136433869583101,0.2507019274853918,0.7182352109820312,
 0.32794688321509147,0.3266525299401004,0.5308477551771728,0.9357327818259259,
 0.3013709155579507,0.7033555306023218,0.06179485468622237,0.7192592280343074,
 0.6975420514139439,0.4628507004553316,0.07240898313882027,
 0.31415095767018164,0.8512752593954241,0.7975364705340453,0.3096880951598463,
 0.40435551283634685,0.8231502439363775,0.7423708250414439,0.8943033472229129,
 0.19881587867527506,0.3884477102255801,0.5324120489473617,0.1453236961867601,
 0.135597538144561,0.02723528662893837,0.24461660901876714,
 0.31801779068011693,0.2355922452825946,0.2589750456080149,0.2924387973785634,
 0.3894148422903765,0.019027785151285226,0.49012805911480606,
 0.4970597636559132,0.6480577903204171,0.44510167875291273,
 0.18216670691614845,0.49512280708383427,0.6462470108974293,
 0.9075998713121468,0.009596883727343353,0.4960875114927167,0.352464980734734,
 0.6174730628157697,0.5570491276937665,0.03617051252692749,0.5066211261150061,
 0.9209515723007329,0.08426229770409088,0.4097951269196429,0.5185415325465151,
 0.28370686133702117,0.26153349837416096,0.06263590698439436,
 0.5771926918320578,0.8555085130197123,0.3621541397785546,0.233151165317929,
 0.4363639157864252,0.6757500477539602,0.08799298434395963,
 0.40533585714605236,0.4014147600122163,0.7791275534261828,0.8325499132393419,
 0.7771594038812879,0.9891992541645204,0.241728124889794,0.5521714085126158,
 0.7102951808647366,0.77946455615558,0.5091083090982246,0.5246565028031133,
 0.23669681835132306,0.03863544056492019,0.14632376212562914,
 0.9413961478839616,0.8640153961996431,0.520749240481026,0.8647984205140351,
 0.030785641844429557,0.9355549825096146,0.7481171380763986,
 0.42659946394012627,0.7711092911663732,0.9346317323404244,0.9068379301289101,
 0.9531814296824549,0.2896287430288883,0.004502258468572551,
 0.3067018536373721,0.2005031236080731,0.4535850146610666,0.43806144271133984,
 0.369140139517828,0.9777390418839036,0.07342657146055953,0.24107560731180144,
 0.4779679051683595,0.991058194495975,0.6561854095662822,0.31042037100258335,
 0.2540121849645416,0.5934105435672826,0.8399687967113236,0.9438571671169501,
 0.5897399419829986,0.6377109338298971,0.42924872260463587,0.6716164727901917,
 0.7352081204956423,0.5075330754298416,0.8274225426415667,0.5249577535646367,
 0.8347547546861724,0.32576726452117144,0.8848097490333038,0.9004398028936647,
 0.9416436376271947,0.25814268076030067,0.3730979422758347,0.6768926186162124,
 0.07797276862838709,0.7110765491094105,0.3254074143902662,
 0.48874234157018603,0.8373790772177945,0.8600084013129642,0.5000880522469444,
 0.9739186002750054,0.10165576997728631,0.022243336008675896,
 0.504131514363147,0.11498396715894521,0.3889624197620464,0.5880495683505048,
 0.4823328022482226,0.8792318767466832,0.20831684561621389,
 0.17471344255967225,0.5628057006950198,0.7034434151007838,0.895654871230698,
 0.7626898626640559,0.7904677061035275,0.7531020331093741,0.948409418169948,
 0.14518540715060346,0.7234771503272107,0.007915120298134126,
 0.33312491201393624,0.7887356010697738,0.5766880600958211,0.2547718293025085,
 0.07609197821237657,0.22928445924531327,0.14012592762178822,
 0.49545972611378064,0.6795514358071146,0.4726720231583411,0.4095097239620108,
 0.0795457614431695,0.06417886240236648,0.3350341027078425,
 0.09668876175067809,0.02331959995220312,0.943443082768386,
 0.39694691454741915,0.43431901255589034,0.19394326728043843,
 0.7472842262315845,0.826370087989925,0.7038593580850249,0.6178619834370312,
 0.12499684136120426,0.5729085283693747,0.3329090748788791,
 0.055861070722845074,0.9703397523505669,0.10255315519056107,
 0.010385441522237304,0.8800496159597179,0.06917960923225619,
 0.5808761332531174,0.05143541136077656,0.3193814647576593,
 0.10288182975264804,0.028127388248833207,0.8789154547980443,
 0.22048184693896866,0.7546102243585497,0.6757815679322339,
 0.45072618811026977,0.20133565285346133,0.8902692920919388,
 0.27811650484144423,0.4344819811955759,0.7672069656228138,0.9619872232120723,
 0.2546499316509163,0.4242290682698132,0.5340377334463848,0.129881651960555,
 0.6700025507626887,0.10001667796511693,0.7323779439421914,0.8050468751653201,
 0.9846648830218422,0.06300606893952887,0.6568329169497337,0.907828255762948,
 0.26073070941694876,0.12123707738144784,0.5447604023255944,
 0.0528382069623935,0.5246389902703822,0.504510623341837,0.013846461747472727,
 0.03923563186939538,0.9162047484327347,0.9637999813465927,
 0.39015433661113974,0.3511220975262923,0.8192325971254519,
 0.20712712017814794,0.2230239358738676,0.30742546388146297,
 0.8688921848523927,0.15883646649283267,0.8213115544004597,
 0.48513422027136854,0.43965525188657095,0.44467987386402474,
 0.11459657686815206,0.5844577480663604,0.09630026367582178,
 0.5574159928287231,0.996654666495169,0.6687096865096472,0.5239707781455392,
 0.2574785383390247,0.7750944303160792,0.4380109415186111,0.9313621866609884,
 0.6661492784024382,0.41773086779450463,0.687168681583284,0.6517858268443256,
 0.10167005587370825,0.5739757847065069,0.6097938310269302,0.5702123886161689,
 0.01525513119937838,0.4114609248104496,0.4824354635820879,
 0.21644384165937802,0.03440235822306004,0.4706144261907904,
 0.5570591281880857,0.655899693052966,0.5083613273430565,0.08265350872560917,
 0.7877734261113393,0.3513906899246708,0.11615438080238993,0.6638170865631928,
 0.802575108553037,0.6490121228746787,0.6178589949948334,0.6364281525564591,
 0.42685021318067984,0.5978322366214219,0.41042975425059924,
 0.3616640951383465,0.33785174587011957,0.3786440596754447,
 0.16720045124807692,0.23879132636088984,0.6530721015123996,
 0.2178744370263025,0.6258852213481807,0.3766733866951488,0.5920067268505356,
 0.3984310774688372,0.6388378502776004,0.6881802586941734,0.17391496571639575,
 0.7141144169189603,0.7993384463223184,0.03759252747993347,0.6728899852455692,
 0.2359244829601561,0.024247971688172276,0.7044822811862407,
 0.11954626077924335,0.6935116553102675,0.023286202854148508,
 0.1345733337590893,0.6202311121274081,0.9211171807318331,0.20865999596300722,
 0.08395335155664763,0.6542676010134681,0.13565857362385214,
 0.9244295091177908,0.41004686942103596,0.8020440815639659,0.1580746029864324,
 0.021670956758652493,0.1046323101549661,0.7915776258617828,
 0.8711884690599948,0.33402613245044976,0.33508506523836434,
 0.9062471011637877,0.44586145623593987,0.5549605607511382,0.3450933737354831,
 0.9321905460325652,0.8762257225459882,0.4251156617527615,0.13746750794804052,
 0.11146771186209692,0.21903011598032818,0.47531535582813955,
 0.9960928835035819,0.9016322064198067,0.48670570739765606,0.2963879819221752,
 0.39420016399175484,0.10055576626402118,0.21324093310095038,
 0.23128274218705147,0.05813986944974636,0.5943873050722408,
 0.2089999746454294,0.18754891202513357,0.21127389051185907,
 0.2221707147357571,0.825097349776811,0.01617480561167184,0.6491477422349899,
 0.042927305995631215,0.5852577355760106,0.9805921075984129,
 0.2857635644509331,0.009133659285497586,0.7005727592754739,
 0.32822229490066623,0.8276548493516399,0.7097361292582748,0.6781124521301537,
 0.5493768894023805,0.11569560432314152,0.20103604810103914,
 0.8361105482498241,0.6901468963431795,0.6179286355416751,0.18793613437453893,
 0.6087123517266138,0.526920883719477,0.5637049136233315,0.5890361435266414,
 0.2537844486192732,0.2166092923706605,0.9497890344380716,0.4470661006796899,
 0.33793975979264723,0.3931766325171533,0.7431520609305616,0.7093452632097842,
 0.43831857991875256,0.06620355985177806,0.2637442007675196,
 0.1269781714674395,0.6995467221265299,0.8270066506240531,0.6548541072443247,
 0.9970348436564263,0.3409103428876563,0.4981309077274627,0.4891734220954441,
 0.7069109655669641,0.8261644191087748,0.9881127580310775,0.7317391273695758,
 0.6908308696836413,0.9372432395203987,0.037786476939879576,
 0.8721265398710057,0.7136053898415918,0.2864137113691265,0.43541787190016323,
 0.21109994735818538,0.3862498157677312,0.07111795488247308,
 0.2519773795365098,0.9925243947478105,0.8788567655528231,0.1128277770249444,
 0.9323663289966067,0.9458978737744361,0.2955590343594223,0.09142778662869533,
 0.35561581390993113,0.810536643793757,0.8872385279104468,0.9992046137091788,
 0.6174914844472353,0.087944211442739,0.20369966588780952,0.4604468925923514,
 0.022311182440326593,0.5795907246410773,0.29728198932820393,
 0.905605554920343,0.5969005736231231,0.04012374781546757,0.09412996028924381,
 0.8006021182683126,0.24241482564757333,0.24259682212807931,
 0.4424929650277395,0.1402255637430967,0.11149394992973838,0.5792560118227088,
 0.12029501906780582,0.11507171164446794,0.12717209750848557,
 0.975974046270279,0.9023139328620029,0.5472875038544028,0.2185432626445799,
 0.921873124878812,0.41438059180780173,0.19353407258614963,0.9905938699962116,
 0.7578747724499252,0.425239457824492,0.06917648069122051,0.4984616877563094,
 0.34113602788360087,0.4151511594166618,0.5181583960281542,
 0.23335918976007197,0.29907853748898094,0.11605857793301144,
 0.9880210988600262,0.3470983860116561,0.3685795236711146,0.01800645383951116,
 0.6403475746363161,0.20671507671076617,0.22363251716435106,
 0.7906297012106812,0.7001063499780453,0.42031312207509164,
 0.11363654389651756,0.2758083404519923,0.8394531790405226,0.3339759030324023,
 0.541912103869624,0.4689081269492015,0.8117857935998734,0.5988716309020801,
 0.1444081794968035,0.8472318409050957,0.279167213729795,0.720412418659093,
 0.38976441813667284,0.5878393191731458,0.3114786242321156,0.5636032825463332,
 0.8018535387210727,0.47910292113443564,0.6893422307005201,0.5052469201466326,
 0.546395523694962,0.4437234198705877,0.1678311694292952,0.4231421138418231,
 0.6907440795370289,0.9679590591883572,0.5577791055774337,0.49560269000470314,
 0.0360326245260606,0.5770041590346718,0.5751411855617417,0.37837892421659214,
 0.7022560794106523,0.00740160579994531,0.08269220768282226,
 0.2679674068530019,0.41411203809513475,0.6509347823726502,0.5974308114675417,
 0.7975134737748326,0.47804896529005736,0.2795894434781633,0.7581140661374206,
 0.38220476130010983,0.9866087824656835,0.5567647674010688,0.5567248164794111,
 0.16486691731561054,0.8840645963997921,0.9962528840782837,
 0.19039633345369933,0.7935107946816453,0.10712989237825998,
 0.9574829343880727,0.5261420755084896,0.527566416687732,0.33659627289498595,
 0.7211442636521005,0.2558943576581635,0.25105738635037367,0.3383814253973041,
 0.5082207153997614,0.4208840895084415,0.7852631958987522,0.06782846915460661,
 0.6977648587394791,0.5700045349868275,0.9469917322399329,0.37718230721838797,
 0.6924708756242793,0.39576599314819183,0.11979829888119042,
 0.2344349271074827,0.15157595375441146,0.49442794532334766,
 0.2637713143016125,0.8881101509745094,0.6013167401833085,0.8387383089923156,
 0.7081957796509861,0.653697218378619,0.40921414974666503,0.3571268610890662,
 0.8972341057787596,0.3357856860025449,0.41140868492784854,0.9591371200513091,
 0.3599050271392549,0.7651831646008409,0.8458799644792222,0.1905797038761129,
 0.9748465823947213,0.7908338892682567,0.1178391233202688,0.6008422550671434,
 0.7498837772382089,0.9611884590868776,0.8674002929226616,0.4820780849982911,
 0.8851347684312932,0.21398475243954795,0.03149954340683847,0.510114026465075,
 0.9417320204956082,0.8903644999001995,0.9194859677218736,0.0844500432037314,
 0.8282031278169744,0.25923285956061104,0.23220045906390852,
 0.46537966938749475,0.40936897096233293,0.9063006963439005,
 0.5913666701178448,0.20702380178031177,0.6172325660963546,
 0.12607190911392663,0.13329019743653492,0.8300602954012294,
 0.9083755721739398,0.606117697663796,0.7461476274630487,0.7999915456671718,
 0.7965002957436309,0.9582217351843074,0.7191048826788848,0.5621637965142026,
 0.9020149622663506,0.392414090675381,0.48254882565937907,0.9607183722085476,
 6.252564047982689e-4,0.9959902860000359,0.34125821128870104,
 0.7724126329097348,0.24808944854313175,0.12039461379037608,0.836438889425196,
 0.4665141847717873,0.6812014667732507,0.9464166807211256,0.07755275189960331,
 0.18596575889009703,0.40621857942798134,0.7857453977922725,
 0.7342700825734955,0.7761932744831916,0.7264635740118087,0.49858367409397597,
 0.17482916419343975,0.209970279327246,0.6759783788464178,0.2971593010292479,
 0.55821602343516,0.3069787535440798,0.36863458032604246,0.2788287851007226,
 0.03455615399362921,0.35982836323147793,0.6728432463061877,
 0.9676719216475548,0.7527661807619614,0.34842925773422606,0.8181989267380331,
 0.8859096992709361,0.2841789468324867,0.6895927114378974,0.8250456136343997,
 0.6671988621947634,0.5569797708436666,0.43144668971556954,
 0.33110357342433283,0.35288134394988735,0.5836268171501884,
 0.5358112016632639,0.014172533905723594,0.1282677474763455,
 0.7610323645936422,0.604022321916079,0.6859681958233295,0.3121279997404639,
 0.10707156095870307,0.6524627498787271,0.021173631431201434,
 0.8402044095473434,0.6905900740607694,0.46445557332146825,0.3911915482262044,
 0.253595610245845,0.14524130347031505,0.07806530129984512,
 0.43016926522971866,0.9667259888061822,0.1391462956509456,
 0.36448054484324766,0.48805607011611274,0.7873687303531689,
 0.8060665821196634,0.6178738395707623,0.011716471847436183,
 0.9159593333659318,0.37533218579181427,0.5546187193880056,
 0.05521654818086397,0.1218892426319198,0.38458919556580484,
 0.21532833241612137,0.36625837082515256,0.9667830396899743,
 0.6476593332007442,0.43620311411186496,0.48761998626719605,
 0.32492109909092237,0.851303042822598,0.7370211071593831,0.8080352670808113,
 0.3583178698689108,0.4939883034548166,0.09392252959459924,
 0.08427060839518008,0.8108409616437123,0.310409597276582,0.8791457855483036,
 0.6620857714497643,0.3713479668094284,0.8924723820892098,0.4056865848044965,
 0.2795400244521282,0.9637002301500818,0.2751464395747124,0.4760325775942946,
 0.9199210754602645,0.9050089135022684,0.9009296205891931,
 0.008308119662330582,0.9368181634561276,0.04214145222647259,
 0.35406222487723993,0.6523054129297783,0.8722243221190585,0.8953926469030548,
 0.08800000175538891,0.8015048563258751,0.29203485996059286,
 0.2109868048678486,0.4720754700813393,0.4670284856134205,0.9386692155225725,
 0.07686781021372036,0.6193008608977177,0.0605827320667488,0.1711023557743806,
 0.9433939623119897,0.1127874521354516,0.9236567355039791,0.4236467889115567,
 0.1667499645220163,0.07988554154307792,0.7274029316940539,0.2801557121342929,
 0.44229799710960094,0.37065782103646316,0.9721087891632284,0.583476719757577,
 0.44502762628541526,0.888657321469273,0.4101692759729888,0.7571916256241977,
 0.9452521511547629,0.6995736419288524,0.4743864589628043,0.8274271381171072,
 0.9727202401585137,0.6139816608303488,0.400687473137328,0.15659824116275245,
 0.5044336444971105,0.0010563021740701561,0.8247668240842171,
 0.6263856866596256,0.08330704094316743,0.16825067471110833,
 0.21120121894819344,0.3535731448815923,0.5033902107553161,
 0.12858578668042042,0.370965422514113,0.4067728175210218,0.9229279334334243,
 0.5047310952468297,0.15462133099853814,0.9292384991658243,0.4936524569245989,
 0.8920142912230717,0.08453150215239402,6.255675538273664e-4,
 0.48080443094492353,0.14356121863372995,0.30170796831941016,
 0.6669337094297888,0.7863805748616807,0.8202196734492726,0.3191524852128824,
 0.24316808134245804,0.22135756058876732,0.504328222267855,0.8877181928389646,
 0.541291963763213,0.4471562190807642,0.7232778734357532,0.4812477294322065,
 0.9007869019321604,0.9749051523320039,0.4304944096184452,
 0.057909086612238925,0.20677299522705683,0.9905617891294113,
 0.8801599492371576,0.13669327254850605,0.273665531689278,
 0.007884871950164407,0.9624674828445768,0.9605639675654262,
 0.5079382075920303,0.5315684263018516,0.012574951935314771,
 0.41993362435311565,0.2955723201547019,0.3910187374802525,
 0.30474846109283915,0.016498900826414142,0.8534236073614884,
 0.7941942767505489,0.29815577783127967,0.23035518413024025,
 0.8894354510376714,0.9115137520073199,0.9280560799565194,0.576649867548881,
 0.07496796871162847,0.5094392381082432,0.8298708353970324,0.4221907602144216,
 0.362852278513614,0.12347379587157059,0.6192625229342509,0.9114113781196714,
 0.8477734401888343,0.27916213991217154,0.8990293365086187,
 0.19670943130142438,0.5020006554504461,0.7660047664053633,0.6561097261123042,
 0.1753043621255932,0.8891272668015385,0.8372014128121853,0.6986677867711985,
 0.7956465599112522,0.060770322274401045,0.7096321066205383,
 0.32328443315471267,0.672057915186999,0.8639195269055309,0.8133112973982302,
 0.011761611057119925,0.8392361677522605,0.8494031850212944,0.823939438812375,
 0.4252650452205674,0.5681539297643314,0.8041781702912074,0.10892321795963245,
 0.6727652482138772,0.8896035652165726,0.09690170391626152,
 0.11129941605324223,0.0636046017485643,0.9007633599045202,
 0.20694745220016264,0.6157065957065098,0.49486854783488277,
 0.8622527442866313,0.9696575309670944,0.3991448763315708,0.6502291752450446,
 0.025792882798742145,0.66972614527141,0.8006627241069392,0.688909436911722,
 0.06193349170789375,0.5860453110259661,0.635955756967325,0.7714089404377245,
 0.892521622461793,0.024280287794711364,0.9772445998581668,
 0.11445684079442575,0.40755289828123176,0.2876807054701478,
 0.8799129592990096,0.4825883201506225,0.7884097561584545,0.15635642721387644,
 0.8742622215705513,0.39163969479709526,0.33279895485486766,
 0.4807518075352246,0.44588668404728127,0.6099771709239136,0.8567501758958341,
 0.5450914521406095,0.19584319831433938,0.958328798173274,0.4490947370781253,
 0.2218687127196458,0.46369294939970473,0.26514386184520866,
 0.2931839133163221,0.03100158922233276,0.73642772994548,0.1820996610707274,
 0.4750975029473591,0.6469194332726708,0.023803636019539276,
 0.40798988371577094,0.04102315463149886,0.4588170402604047,
 0.9242591824787437,0.2789545133158494,0.13705294014416947,0.3326637794092562,
 0.2513086626600365,0.014441155463964406,0.44774356239758406,
 0.48084022711180574,0.20912951837156601,0.060465132609684114,
 0.07711321800877147,0.7617742714863356,0.04282343973588543,
 3.6850593977566515e-4,0.6655172785258494,0.4010372661639888,
 0.5353177716712273,0.961919027033557,0.4388296589164731,0.9802378409212733,
 0.5843795574859489,0.6411363470524599,0.5998731564985389,0.33919393697871814,
 0.8203650837405561,0.634574865258745,0.6985182813158655,0.48665471040929653,
 0.7983424609269956,0.16184691177092392,0.9973485214469199,0.6666318411804005,
 0.3733677472594319,0.6823092819695572,0.5481446053196235,0.7279044405157037,
 0.2772766895720491,0.26184800015658716,0.31672220450025834,
 0.8124323059219913,0.7327128324223612,0.6793192288323675,0.04789918364791568,
 0.4295390948693113,0.5841502322262293,0.13574357856592945,0.7178829688624393,
 0.4753686181866733,0.8688170952852782,0.3394711273337647,0.552667597357849,
 0.6421480546196503,0.38272226565008405,0.6561030205505817,0.9203486151458085,
 0.2826815196703668,0.5805369558778192,0.3003271831493478,0.8407538784038745,
 0.9316648938592758,0.5149459776319842,0.8870265224914268,0.23656299094527655,
 0.6263100907462334,0.10708540415499912,0.7501972953888013,
 0.05009570370824856,0.26866804900257923,0.6549083748589267,
 0.0801495086131444,0.7190228197189334,0.10106163784486999,
 0.11582521802352952,0.9777300752500081,0.07374095780014045,
 0.06498980581960812,0.32444623025625363,0.5027405532444786,
 0.8332763380911548,0.07192547117433379,0.9563036663875827,0.8745365382033677,
 0.29229247522886515,0.9909685454375331,0.21646147275587058,0.741203514058573,
 0.45420426283989335,0.03691807146705606,0.7919736055252171,
 0.37252587906854884,0.31809743000491786,0.4007490197144833,
 0.5890838846923409,0.7558072291646037,0.17986742689146673,0.7425549903118432,
 0.4689258360317433,0.7560681337664684,0.06544742063904208,
 0.33789994464714246,0.8901716762177463,0.7624560994213876,0.5413433917919278,
 0.2776638040912499,0.34944943658001915,0.8059324758215618,0.297816364601439,
 0.38606583655031246,0.6170839240500685,0.9147572669068941,0.5509022666753077,
 0.3490754524722153,0.9273146805310571,0.3877851074668277,0.3108396357876121,
 0.8872806760742944,0.8985065756311724,0.8444790415458933,0.16697333538143655,
 0.9082366751196167,0.6274212963901555,0.8912251109077762,0.6565408041459865,
 0.7370646473471109,0.07238938949573359,0.32283514287117054,
 0.4406336367901754,0.5961566220247742,0.9163765640592726,0.305779626032595,
 0.5554393917408753,0.7006060626864717,0.9669335596316948,0.43807427327662696,
 0.3992443860202738,0.7164315209318124,0.8039528994419752,0.9145473531021814,
 0.26423844500427296,0.1760664750470794,0.5743781361195737,0.8453515996136743,
 0.19607991384616108,0.4690163597051178,0.3081787059679628,0.8828958716544045,
 0.10547800809209973,0.5580447274883745,0.8424578719043059,0.5847279470279169,
 0.14950018237581186,0.14008009268462795,0.32886885082049166,
 0.15261170086387965,0.5715868431299125,0.6333311965489694,0.5819336623104311,
 0.7573004563305493,0.9972837396447538,0.6092862181460903,0.6488622375793901,
 0.5127345192640077,0.31016322727526846,0.7207438412628066,0.8861449986913928,
 0.4344131878615102,0.13461898539090933,0.9357156600715172,0.1895166151393477,
 0.0225024203528702,0.7512849239833901,0.08650158399905739,0.6833577246784126,
 0.4730004619884284,0.8172925638159149,0.4273678168670533,
 6.306130156374934e-4,0.5024158937099127,0.597951847925134,0.9830114186783239,
 0.281646949880776,0.6846182875697384,0.13973401956575981,0.1304209159460724,
 0.26170266880062276,0.457963107713912,0.5179681993567016,0.09824512182920053,
 0.2309999104885112,0.43947868993069805,0.6137207383392261,
 0.01851648046100396,0.15292738997377742,0.6229281069643083,0.539300978954533,
 0.2613422318156493,0.1055455427614177,0.5355340375018598,0.691888220535598,
 0.09810391293100496,0.525510031677871,0.6058242342797455,0.5498581197907643,
 0.7653958228447915,0.31081750214177273,0.7076025610971001,0.3635631884807399,
 0.7720066072640873,0.5035092374003345,0.5411480112101876,0.9280685061725698,
 0.478100865386199,0.1857097911545197,0.8209426673549408,0.6089923926105865,
 0.35863015661466335,0.27979768162897733,0.43767125251524475,
 0.6049988243741938,0.12329289850035974,0.5695031672310472,0.6332584545181128,
 0.366195350256875,0.6754992695904427,0.5501494544937215,0.7482947308666293,
 0.7122922254761068,0.28611765530037914,0.8636869569327703,0.9508534048255357,
 0.09773425602372576,0.6010898958710561,0.21952661277608454,
 0.9874580687072948,0.3731035312817461,0.8883778895263161,0.8028611076481385,
 0.39078174989889436,0.7822358276137895,0.7779424534001634,
 0.12851063806991758,0.8915919379216707,0.9897804285574392,0.279088141717148,
 0.6109235767671397,0.8608082773228067,0.3567420641659078,
 0.025040863362335575,0.7483777878557497,0.37794222093309415,
 0.6809607848102024,0.9726945902505607,0.30859556691966106,0.7431607665304679,
 0.7271640120885464,0.15635702363357595,0.132046789772446,0.7348316034250537,
 0.9976478090508879,0.7999624906641655,0.3435234649558787,0.7613303984156016,
 0.73044946614465,0.6647006125842827,0.9567009645368005,0.07813667929816326,
 0.7542716877422512,0.5016118166036194,0.08416459100550866,0.2512497703937755,
 0.3372008018824221,0.5133945777648092,0.2582201851535655,0.6448268580802952,
 0.6723364667863608,0.49042759759691446,0.08764484353466573,
 0.4654026725283037,0.5591832879551252,0.12700491160797933,0.7523385212131379,
 0.9051373270134115,0.983711036544389,0.5368644010291965,0.6230140459655961,
 0.6475767389846158,0.027684589123773473,0.43121886832226264,
 0.6162462869056154,0.2253987786339373,0.8704500089687435,0.7225999233296827,
 0.276539794088385,0.957013872481539,0.9234006391675879,0.654033086582314,
 0.2632676391716754,0.9524040397733017,0.9673012173116325,0.15756336343445465,
 0.4243777475267767,0.0749710365091355,0.22283794802430013,
 0.05317378396010464,0.7815280620645527,0.9772798271853977,0.4095251532319797,
 0.03324718621573641,0.17037289323935223,0.3307140839835536,
 0.6513432772628804,0.5531870529218947,0.2989408095752104,0.2595294477320085,
 0.7977914394512996,0.7685539554548336,0.3959476275473469,0.05012280844892536,
 0.6547811305629577,0.7524827737595259,0.008911845027621323,
 0.9590399427824198,0.665865673257813,0.5691427557534104,0.4264543810348408,
 0.557955620211175,0.626053194072532,0.8657147449200678,0.11338023623685478,
 0.8219204015166243,0.7445023038944699,0.02367307018550069,
 0.024186715192672792,0.4834336287413912,0.1644095146971032,
 0.40038551140709255,0.398696585178426,0.2820895731354829,0.12678554856970203,
 0.134559418362745,0.3242155182894133,0.22452382466326914,0.9085586907978814,
 0.7174186814500304,0.5288540786558215,0.248368145672608,0.4429672361789698,
 0.8202491880554408,0.656028387487197,0.3841354823189762,0.16420127415353847,
 0.13659305429589919,0.14480060986553955,0.46304888205745653,
 0.09855458851742738,0.7628369892838267,0.27638141662563753,
 0.5835802737281098,0.17735027536176484,0.6368226009814164,0.558696876330838,
 0.657930098879818,0.18535389589775575,0.7020199638940108,0.47749198148756666,
 0.7066289496153783,0.8804410055945826,0.8741937651220242,0.7818736064256995,
 0.7372735412011809,0.9333272011395607,0.44948551577421036,
 0.07601218945381882,0.218934412057593,0.46916975405159145,0.0500761500122654,
 0.7053347379729064,0.5228446219569095,0.4193591731842453,0.8142237053210629,
 0.7564843472862639,0.2772769642913211,0.13515247205923764,
 0.018936177905705676,0.6374776122893051,0.9353222000235981,
 0.17637116498340388,0.8563157104425081,0.31886329107198896,
 0.15727861338875582,0.7847770321549427,0.7222756040935808,0.5388516929870746,
 0.9821785000932305,0.9319128041199526,0.6971418594497333,0.17606751152925615,
 0.9525561730646589,0.7587192030432288,0.23761259066322482,
 0.11351118038334496,0.47645090596033524,0.4057308532716307,
 0.9493984462590229,0.9825901659670292,0.04987335671144011,0.4390350073239141,
 0.09327394527635813,0.9796595264399185,0.9401179152710738,
 0.12801879639038138,0.03727867461455803,0.461556651563807,0.7053911141697611,
 0.25074017441497576,0.8695763885383194,0.026407974188253247,
 0.3261578322055543,0.010875572395500632,0.7611580707509582,
 0.040038833211314184,0.553044862804193,0.522431947282936,0.0419056980700202,
 0.2672866918666208,0.995801098411507,0.3820455711241926,0.2855828008691179,
 0.5520215470552244,0.0641623011001089,0.08365856597541022,
 0.15118001283518767,0.29804343647735854,0.544848857482263,
 0.08057564121705785,0.8447231657688143,0.36179189307232273,
 0.8987025573357085,0.7966475632727058,0.12479412952066027,0.3866841719279359,
 0.6269795652783471,0.6851255321591542,0.18923439818493792,
 0.19982485316984866,0.38505799566653665,0.7747358947982331,
 0.4131343124219746,0.33098053626464596,0.5376092863704333,0.2010977607339035,
 0.8574977613271253,0.9960847264988049,0.4486070588982668,0.947781404383306,
 0.7485285264794143,0.19066537247680548,0.8600586734428606,0.4556079797604463,
 0.09942363997112713,0.4757659245550683,0.03422375284169343,
 0.061088544456358806,0.5006169382666741,0.6934902577335479,
 0.7035259970123918,0.9733337578495093,0.6497299919380851,0.9540589921037828,
 0.07151280813179839,0.6350608660749569,0.9250823659658766,0.1781888928958817,
 0.6850686093257019,0.9453187363870563,0.010697183925338138,
 0.7805185769609074,0.6764909597353629,0.30533220058449095,0.6041507268777667,
 0.8647375968022202,0.45576146352521907,0.21228562510324822,
 0.48274131558454203,0.7858005545569893,0.201943076684297,0.14654897840624026,
 0.7468527541365126,0.913728798566738,0.16459165293869138,0.726878344662527,
 0.26641951590845137,0.5025118898269735,0.39096576428007024,
 0.5669430072083184,0.3283305286143614,0.9108002190778013,0.14342410952019802,
 0.8687742711277047,0.5838980450872541,0.14926947365734655,0.9395743427983259,
 0.13196586714347802,0.4967614264560918,0.816515278994495,0.8115996465850827,
 0.360029518376513,0.011004996848433257,0.6168644899905446,0.4060770389000262,
 0.9421896326134095,0.4259368109089715,0.47103542973898604,0.8899565491958078,
 0.9492371468551811,0.7405162540401727,0.6360109711290506,0.6367377229786053,
 0.3129002640167502,0.2422653420080928,0.9525852278221988,0.311720000392828,
 0.04876289785317145,0.8439160653240221,0.38251248012327443,
 0.9246114702057993,0.3671156287199664,0.7437288817204274,0.8253886772180083,
 0.36672471450966304,0.07878863027484162,0.5115061987538921,
 0.7330055642628783,0.17966425124114105,0.7085289295053491,
 0.020673157561909905,0.9316089998656187,0.5378366626665714,
 0.5683170733294549,0.5035201298113487,0.8410312102077729,0.17769792980174537,
 0.623780156175187,0.19485766418783723,0.0568255114643933,0.5652752383898481,
 0.26898655844985686,0.942489345067782,0.09297178184854649,0.9884727472401926,
 0.16130032284464113,0.6544707004354308,0.13169954294247033,
 0.8069144075035914,0.4789799744755143,0.49676634979128187,0.7879914654418151,
 0.9235371036293352,0.4663127718426633,0.4926119815772234,0.11949795981844413,
 0.025391582681587033,0.7114151639758891,0.7934693025514166,
 0.9135695339111537,0.2149300953560065,0.577089161080655,0.6018985941417526,
 0.8836439881222908,0.32439543153141603,0.2629257315183592,0.4707153749692714,
 0.4775211023650663,0.3387391488454562,0.6419324504317357,
 0.0014038663911980498,0.9099216718712727,0.5500887584471208,
 0.6756382867746734,0.2553536453031968,0.7934241377723154,0.6374371914247879,
 0.062125417321792,0.12488046157998456,0.9165103032856508,0.8736717682919383,
 0.12836637017382502,0.6262871148592362,0.0396835038465595,
 0.26694255112164056,0.3184409716028713,0.08400841386775593,
 0.5959962709687665,0.8386475867208079,0.1546456700929829,0.7573362077661577,
 0.3523880551693708,0.9933882134876837,0.6982662882857558,
 0.042504735185645015,0.45874516672206433,0.5616622092666802,
 0.20043188604393702,0.9618277048961683,0.10136057881440785,
 0.3933285662545267,0.36875185525053866,0.6009246804306533,0.9648684526181748,
 0.9644083946407723,0.7243816656156767,0.7114671735699649,0.13214346357548679,
 0.5635560497471095,0.0515182966054466,0.5001564570579505,0.33003581440751795,
 0.18117426287688143,0.22889564132872564,0.18502495010437736,
 0.7504800598704398,0.17777838475418817,0.9956730111880057,
 0.48310833293950706,0.11125145020401062,0.6385573378557596,
 0.8669659965733617,0.43717944321034186,0.34945759813622024,
 0.005604789514110253,0.9406634792727179,0.5089053124067737,
 0.13852735072538036,0.10029903800450723,0.6177178375775965,0.028798366490144,
 0.723801466691732,0.5698330736707032,0.3542919897964045,0.5455790623731307,
 0.7557106535020437,0.9069820343267578,0.5565168576183073,0.04830828235502238,
 0.3370988797812613,0.40802228782890504,0.892197452761976,0.11564186587909275,
 0.28884650149233826,0.4044265827563662,0.4940416422382007,0.6069993725324422,
 0.5912239407008959,0.537243280960882,0.5504076008745983,0.5497261582240873,
 0.13658768814957378,0.39042934535506946,0.6018557542844698,
 0.40098751162107216,0.8505380234693178,0.8546630267909893,
 0.23823854551414625,0.2142239608333244,0.19008106361851984,
 0.3478512635970127,0.1044785389297247,0.5355666504399578,0.8039329160473717,
 0.17461284563199408,0.07817134692569683,0.3947735974392932,
 0.29491809445300377,0.6379629356357459,0.49059320960166786,
 0.3465934689789737,0.9679919094791791,0.6430278335450322,0.4327649334535455,
 0.4106280751475595,0.9237430065294949,0.6071422143987563,0.84735869454366,
 0.2851975580652152,0.4755430322408868,0.7668476497820279,0.18636800080135785,
 0.6303904389938837,0.9460602398477324,0.8704384143649029,0.6938505766594956,
 0.8725707489054727,0.4004099624458709,0.2779464505774327,0.37477456677973076,
 0.8784103380227574,0.4660525415802965,0.978502913002935,0.8471828847301024,
 0.8018124179341458,0.13756933121331072,0.5662429860376037,0.4618359674562391,
 0.3138468328735651,0.8548177253254671,0.5830279224416702,0.8691280406723242,
 0.7519196032420977,0.5624131327525647,0.1471601597045754,0.828994858954781,
 0.6668117858912272,0.16165160353042718,0.06129141369557012,
 0.2785476039414133,0.5191430033646468,0.21351095894487682,0.6938276671526675,
 0.2633059861442696,0.39117179845057304,0.5938359269215507,0.7103105904378832,
 0.3188513865271182,0.4370429499615548,0.6785727870541258,0.6085003080188487,
 0.9805209982699936,0.2511958105325267,0.3957891367483952,0.5715114862395715,
 0.3374733532787306,0.8594175843874748,0.5580943796102547,0.9371375086130521,
 0.9210578185618449,0.8777110809913325,0.7331219663309988,0.9312605784453334,
 0.4348809225085959,0.9986899426168312,0.2857252790211814,0.7341948424025981,
 0.6483697403113438,0.4579651914608477,0.21838613579531185,
 0.20754533012500453,0.4531586999972159,0.24890141855687287,
 0.9760178950219869,0.3795956887697346,0.5154363269437586,0.3506154606261517,
 0.8540441806355088,0.05708810785215801,0.05576916194888448,
 0.2536912155430797,0.03247500017915472,0.05515893334200184,
 0.1538045884745135,0.0937851833682688,0.4762367219832343,0.20499628833526617,
 0.17819199007300113,0.8219514364576652,0.0190436314810144,0.3272482866507751,
 0.7984639031710443,0.8674262678995457,0.5244460454207976,0.2987128266917338,
 0.8989335040756994,0.05752372127868122,0.0981324302906742,0.3828723517195425,
 0.2680866694985463,0.8375977741079413,0.19900183949209782,0.6809148930113019,
 0.3577449138334473,0.26304811295010355,0.48464107656602406,0.623731187893713,
 0.1935628504511564,0.6333702038561878,0.3804063504690455,0.841655384120348,
 0.516936775674355,0.6924454588198791,0.29881202260281037,0.591365632570237,
 0.5323652019906927,0.8396025930828689,0.2861028368794243,0.9271871273555921,
 0.32232645533920956,0.5111056942457644,0.5002525346136125,0.5985433387599536,
 0.2116022815795603,0.7494234174653441,0.6094431895024355,0.742363048684339,
 0.5790024756631386,0.6272587732934201,0.2850503567627213,0.45019816866644247,
 0.180628421533939,0.3498584505753357,0.47320334690855814,0.20548804584869518,
 0.00398612821899802,0.8061173788181379,0.8805638431915412,0.7283344232489138,
 0.22369965240795087,0.9855659393088783,0.6167526776936739,0.2466405695616063,
 0.7498839362733236,0.8028702647477456,0.05998617889032476,0.7228361879298782,
 0.28145679191310524,0.5242970308157013,0.707811738200146,0.5146741202461584,
 0.155932156803411,0.9269438976143398,0.8426078167250842,0.5236164804330885,
 0.9498115735830536,0.40751178345711625,0.2074131180969313,0.8095840588083638,
 0.43987007038230497,0.6192911134037866,0.5899262181896587,
 0.034925332948783194,0.798914602365904,0.4227190850054485,0.8862629350640496,
 0.516220867725956,0.7777058862583551,0.17170411917146133,0.39033404288029205,
 0.44804810634871295,0.665881564788367,0.7631557469223126,0.7407196119408863,
 0.5589242338348741,0.920481908323229,0.6996043599782628,0.9368009004329585,
 0.10613608549414033,0.8030437064044034,0.7394950743007735,0.5614339003655634,
 0.4335693283407498,0.0899569374256175,0.23134683781991328,0.4176164648324401,
 0.6417320202851867,0.29430433431529945,0.3938423491708135,0.5849158957948675,
 0.5313993617899796,0.6511076621494429,0.24529865221492142,0.9286757927614001,
 0.6718947255213008,0.3082922258984935,0.05056393056811448,0.7696822767981504,
 0.462188574109876,0.5556419563709799,0.9484864313271025,0.26450855454556954,
 0.7777827649376767,0.5321707417324472,0.6500636608601352,0.8829357419232973,
 0.966839527681834,0.8058003245831012,0.13785910639707344,0.9028348420421869,
 0.32339762335467215,0.8698052212390195,0.9571622550168747,0.4466819700446586,
 0.05661369140580663,0.9942169044354967,0.12593173760226706,
 0.3554203351266424,0.5080916796151409,0.7909673932073041,0.20035395908963516,
 0.817500065053043,0.9523543660990755,0.5942074321862305,0.30078909492354244,
 0.1452751701663776,0.7881694237259979,0.6293816301633811,0.4800588400074002,
 0.007462459283723799,0.04256085226111872,0.9950264053492015,
 0.007750777737729697,0.5861146039760617,0.1322605553069104,
 0.5751216524572178,0.7887862046359343,0.7450094181266326,0.17234716313893705,
 0.03339484334885712,0.5485583745755913,0.9168704286162194,
 0.0033621813211939244,0.1497554816080321,0.38034645809809264,
 0.606375619819725,0.5747313876383784,0.4094772923844108,0.6213297463383805,
 0.32596737831429223,0.22500195909115073,0.9965728549081331,
 0.5013974032757238,0.5508331707885263,0.02969390271157002,
 0.16898045771785997,0.903274784580191,0.6306246665619544,0.4235432101327793,
 0.2808733110605113,0.036207055916706476,0.5246611612337816,
 0.6550201633753275,0.9919643480865371,0.6453642271135537,0.5391261134611405,
 0.8199494828806979,0.7482754380491681,0.4752339821909949,0.3503585736707665,
 0.44599457190282765,0.8851740256198941,0.32288214056774056,
 0.5646968904144313,0.05613391770072029,0.40101967339358846,
 0.8447242863365917,0.698011531007807,0.9117135982884391,0.024602493356739807,
 0.9112666869066168,0.4799480243122203,0.7350817623944569,0.5561026396472636,
 0.021614519235089946,0.7628975963373196,0.1086884836710239,
 0.5032576164341416,0.43590431851643163,0.32548914486689884,
 0.6837624106782013,0.766011144362352,0.5684001722529628,0.01085656245509814,
 0.8483979077008594,0.778914649983201,0.9434755805944909,0.49836123883216543,
 0.934168626643414,0.37611466294219054,0.27905929849242317,0.7177898001104059,
 0.9714607433157054,0.8158076306752065,0.45352178696839607,
 0.33204983611225414,0.7208120885612282,0.7670730872341054,0.3933340740647745,
 0.3388040435285049,0.34314696399620015,0.09196795836180027,
 0.5377765523966852,0.8174681568363307,0.9762507481082244,0.9576000881321706,
 0.9689257020686399,0.28050973619786235,0.8531361181239243,0.627591649306557,
 0.6907326866184342,0.5907878637297792,0.15333911978802617,0.7436900559604717,
 0.8113918656161575,0.6824954019371162,0.3577061763897551,0.2268756662649869,
 0.9276053784326733,0.18052456154662,0.7333689223661797,0.6492969114811759,
 0.6436079861705096,0.07934575236794728,0.8994366278335293,
 0.25884908357727643,0.012166499560835375,0.038955893981903245,
 0.9885068076732747,0.2638488445921542,0.8492959843988078,0.3576260813284906,
 0.6440142249339951,0.5375060302224858,0.2949475116324658,0.12640955043895796,
 0.455250068793009,0.5955615050223375,0.08137128442485675,0.37458069756272483,
 0.8776834368123203,0.8565643325846118,0.8991642253548837,0.33628173605477296,
 0.9151582310643991,0.32156302245757884,0.4623553422974146,
 0.007596814134667618,0.5706098597866152,0.6566229313981554,
 0.1550665727311581,0.14014451096252678,0.24981062918396235,
 0.31681405599281787,0.2766898117823504,0.5064830836146389,
 0.02958054744568639,0.14211696754761627,0.6741201456386282,
 0.6236912441850201,0.6227201854432081,0.42059076939808027,
 0.14641708659353236,0.5982914470149574,0.7627683439552552,
 0.07221269733296398,0.34735034435293355,0.814237660357053,0.8003903974230893,
 4.401346946300011e-4,0.6354790308344852,0.4233757631009989,
 0.10298826182056064,0.40937823383685157,0.055987599242037,
 0.04318636756168126,0.22480689077400162,0.7164645849623128,
 0.014764729022235024,0.39944907997509216,0.7319986443935402,
 0.4183242223350623,0.2809717923309928,0.5153447736014277,0.6693677488046471,
 0.6524757550472994,0.31745387465567076,0.33337844870087874,
 0.9836793795505684,0.0758829888080843,0.6978496587787547,0.6863179540472327,
 0.759207219526767,0.05691072683573806,0.865233315005254,0.37120712413351376,
 0.38081931826385207,0.84681735467596,0.2364518196756087,0.6847812169594146,
 0.7433215835854771,0.227864530727096,0.0777074809228313,0.9501873210208981,
 0.5901913033335113,0.5198969841451095,0.5402455251947438,0.1353088447889208,
 0.963147253806005,0.5160802338294482,0.4731784786148233,0.5907865569517449,
 0.8959901723101976,0.19501651614986326,0.20754597113849071,
 0.43086403455838795,0.04022420639742663,0.4724940240557145,
 0.7067690945615808,0.11852863334826402,0.4085205816882175,0.6068836648419409,
 0.3487208806173534,0.5857315693820196,0.015995175184794874,
 0.4222550924347641,0.34396520203890335,0.8860522204580881,
 0.07204207886585356,0.6325669230079682,0.25678565257028685,
 0.35031079396578646,0.4706110553947036,0.6072269129156788,0.5225956955528842,
 0.7690911580252952,0.3581802928383334,0.6822606023182842,0.1621765164652782,
 0.6185189727924424,0.7833403843402165,0.3817458896710525,0.8924814553320308,
 0.8747203421506641,0.7643639349185779,0.8981759150213117,0.31776201862460707,
 0.012492321915897886,0.2186541567505743,0.6602710391952045,
 0.6700421268459686,0.5140397423364206,0.4983861818359292,0.4836873946117025,
 0.4100862923952163,0.16376947106858197,0.10224587930011864,
 0.39635127349860144,0.6056430231683634,0.6522663091082822,0.5263986573733919,
 0.8604576357353433,0.9431498611646585,0.430938335010561,0.026679697743803077,
 0.07524289613580137,0.7393640609328554,0.8892315693694725,0.3704967583163641,
 0.033887177467459395,0.7418254523415667,0.2541314935634603,
 0.36316986517505934,0.059633661971674856,0.10448141632874086,
 0.1092412138733776,0.7921620608820712,0.7324979650295709,0.18774378911034528,
 0.41332997833281604,0.6450711277184764,0.6137198577606141,0.3368441186068042,
 0.7380815394535378,0.9643218547815596,0.6458442557581366,0.845146424291166,
 0.2160038784565117,0.27612398631252577,0.266711709152591,0.27721060335538417,
 0.3782636182632879,0.9299460411563404,0.2579914819777783,0.5012099429770797,
 0.3945172837819878,0.8087132178409651,0.8637915671941113,0.2543505107506405,
 0.361656762130223,0.8828328366099452,0.8175787373594533,0.3482618873097525,
 0.559901470694414,0.8352942923589338,0.5673717813835499,0.2868532153698742,
 0.4081474570371504,0.6706223662424552,0.6811016122748181,0.07251123745677901,
 0.9852092542348072,0.5590274217834011,0.6920907636191759,0.7569028014927128,
 0.7021071951917324,0.27010879312014535,0.6931907066480318,0.612667349474959,
 0.9348620514097773,0.11301419163401161,0.21873467475555486,
 0.9082229854355026,0.9263405052091801,0.3820063710435314,0.7971392985163815,
 0.35166741006057833,0.6835296946048434,0.1582985077457233,
 0.17935124455173357,0.44035461977203816,0.022893072823422456,
 0.36808316930719354,0.8371340416836413,0.49898632187538683,
 0.9892580200589135,0.37069524943382237,0.43919197720516534,
 0.1306667985640646,0.1985239535097566,0.04336733860623798,0.3081059756818916,
 0.6192701165678789,0.23976273292555816,0.5220914618687108,0.7493810563294557,
 0.03945054817077387,0.8052259055566064,0.6081635729912239,0.9314674538287401,
 0.023129385857769025,0.9198095002535741,0.8879248535316074,
 0.5890707503507094,0.4644899283701377,0.5279846907475558,0.28470224078799533,
 0.20824504810210598,0.6922568520975956,0.8637933736812413,0.8422146037984629,
 0.0024196229807146308,0.22643032005888353,0.29432139221590603,
 0.7923217215622055,0.17043211254415014,0.12626924305125253,
 0.29224861578629524,0.11358902587778974,0.8680501606216692,
 0.8902358488076357,0.4642809731890619,0.8173013239089794,0.2972896076656063,
 0.5572734364467408,0.3018019301965327,0.04204938230521971,0.967866903934157,
 0.7561512692578125,0.7531305191987931,0.16954282586390512,0.7075732516596291,
 0.841954479790774,0.4647174982515664,0.559493959729765,0.45115240698240267,
 0.5973301019706415,0.2911351265863764,0.3899847757678825,0.9823261855176212,
 0.32339324854817575,0.06674297745395408,0.12470053144105009,
 0.7692475948599201,0.2855730961918346,0.5119526010324701,0.41134925400590583,
 0.39865357761505116,0.8259786374269634,0.6178828467131474,
 0.04177036952409141,0.08278853220591997,0.6451422876034161,
 0.9989498129223149,0.6918467717337333,0.25184466518814574,0.8579192998038612,
 0.7956735273162477,0.0121896176252414,0.46400328752651065,0.4920828427951678,
 0.21746493566646696,0.04058404175666619,0.680542765360467,0.8057258663699916,
 0.9261406386655635,0.9480008893170475,0.11855302549734481,0.6702747915667385,
 0.9295591685829935,0.5985356091089755,0.8721704632201759,
 0.010320218084797661,0.9523254920704953,0.788942076973253,0.5251715104938361,
 0.15024237628485504,0.37352786332927534,0.4292408663232907,
 0.5887274955795102,0.21659947308172234,0.8854593812632818,0.5106718502712438,
 0.7441221937291969,0.4226998787308641,0.8186851471311043,0.7626789348483142,
 0.5722249361865657,0.6349453275007841,0.22573320650661144,0.6679329662405079,
 0.555631906471711,0.0578172628881124,0.983898977853399,0.6580380323580741,
 0.49551941245351516,0.5544869516393179,0.5213472020917439,0.6177855639073224,
 0.91386955521252,0.8145155769077054,0.2997611827874008,0.5334865980005786,
 0.38824825866391344,0.8811955090611913,0.19885250953843958,
 0.9757784135561248,0.19422244559587132,0.4406558483850115,0.6560341442618983,
 0.7523723255194104,0.3354664147609946,0.9574143628350866,0.662975264382805,
 0.03553726718584471,0.1254406954177587,0.3075699500085325,0.5666110692893839,
 0.2548562382500621,0.17056740695315953,0.6685478660896496,0.6720102307308319,
 0.7314981151041546,0.7783258849313152,0.08081753398769242,0.5162251363604468,
 0.5726244743635328,0.5951404324929496,0.24928233620286755,0.6746476287246834,
 0.3085129798054973,0.701083096361325,0.3316564092509786,0.16652322130004704,
 0.8413065062408567,0.48855054303160994,0.49501372966917534,
 0.11031442412343273,0.13872658695048212,0.47145540678807596,
 0.6770011557533916,0.6326137114800826,0.6023966989434939,
 0.012550675363085628,0.5540261864342193,0.6019807140515138,
 0.24858682471555482,0.41898633392829476,0.42938118627827704,
 0.19637845776575025,0.05863009341547043,0.4824823168813772,
 0.018209178904510903,0.5505281939194621,0.050655328871507144,
 0.6059582036606463,0.11688228415553192,0.6391106027668886,0.9416157002741996,
 0.1638478782545103,0.10988493869158256,0.5240328881794173,0.4537349193300393,
 0.12278232849252968,0.3348560760539141,0.15398184062384268,
 0.2877099944886887,0.6105134183231754,0.7519844831358019,0.14485445090392635,
 0.5811786082664059,0.5325207985408273,0.9891445839402401,0.6529352217250208,
 0.488614562534059,0.03840516198341437,0.8226048268539088,0.1185376158096283,
 0.577543632627916,0.9309160338737956,0.5290869157060897,0.5190800618641543,
 0.22106818083367308,0.7094651062883026,0.05350524340809859,
 0.3615145592361587,0.39564363203847086,0.9938451954261229,
 0.32563145670413685,0.9865055477343743,0.8818181548859925,0.4475624166295671,
 0.8901337951843493,0.15746808398559486,0.3219750565012003,0.2126213865230533,
 0.613727830366836,0.5668542950156015,0.3945525220780648,0.2804084140555374,
 0.4803998314632393,0.34340149407650045,0.36454483874529175,
 0.3797239655723198,0.3485056914395118,0.6218350680501461,0.7917994530578112,
 0.03630283601085493,0.11959923666086958,0.7036833261639319,
 0.12501102575787948,0.9622540723778371,0.9781996043313098,0.9908632518397644,
 0.7303111440976721,0.17098164210512046,0.021780933146297143,
 0.5631544655705482,0.9470646148195335,0.18593949869590376,
 0.27302826803738034,0.36752437280235106,0.7338976579504486,
 0.8204972155527027,0.09491875705770592,0.7274106397942466,
 0.26470643415420236,0.07152569413357412,0.092526261826821,0.8798759065414009,
 0.21180221407641409,0.6636449750825077,0.3615422751938763,0.2513529086719133,
 0.39502847061840884,0.7272489035376828,0.8142172665909801,
 0.31138596339673197,0.5146306541259096,0.1385428863091891,
 0.012816466171622909,0.4949321842096346,0.7156089912629653,
 0.5482162978867086,0.8291540713597,0.11705465590018271,0.9196678072817417,
 0.7463585002128175,0.8714120563687539,0.7915850784526156,0.5260792113381783,
 0.9898753714934336,0.3894700766771324,0.2512305197994912,0.9111570254149206,
 0.19369222294175736,0.016762811840468794,0.9134999787771162,
 0.1927918949301951,0.17309539095927817,0.5422470322833375,
 0.16203676508145604,0.11153938489988979,0.933527097913044,0.314372967165764,
 0.039106503703749906,0.6022014493650761,0.5227277306731128,
 0.27198533793416446,0.6248468675285409,0.5211243228030004,0.8735604105722183,
 0.48854850596288346,0.9975061490764866,0.9349802184358545,0.694950412207257,
 0.4817746242681692,0.5801342132172961,0.9359370330649222,0.21539666451327666,
 0.4893676491395301,0.9964081199880771,0.9785216895846437,0.6541523969865605,
 0.42413000028348746,0.1891374867845661,0.9058941050573748,0.9276206677694813,
 0.3561440271931867,0.07891874868295379,0.5694220136891583,0.7074288515701457,
 0.6341666092531099,0.8975286989793246,0.598646990870788,0.47410481902911317,
 0.9545516323627117,0.38314593705421407,0.5783651930558893,
 0.04242791425525483,0.9413761466641906,0.8556975108754092,0.8169069458567298,
 0.33176632038843734,0.7009324230360223,0.5681626298158803,0.7879682470062743,
 0.32647990405647587,0.12845717996083794,0.2337651937072167,
 0.5352172102289688,0.9385138866929273,0.6937063117962491,0.5544720716123365,
 0.03958912264564063,0.5325986940716875,0.5014251806572618,
 0.01109572057389785,0.8086517890671369,0.6385812834496293,0.7849954445516762,
 0.3413909269910539,0.8386495954609889,0.8363537054982189,0.9528169272757993,
 0.3822738138253058,0.8685482757041553,0.8590372572460181,0.9366835634907229,
 0.5296598318217647,0.26019462790540815,0.6418703833271611,
 0.41658140940855903,0.36236762955609336,0.9645033096563644,
 0.6137301491367043,0.4719667658419947,0.35424466498500173,
 0.46711846422888104,0.9237549151812382,0.6646597987007301,0.9488138392390308,
 0.7991614251149078,0.16812327029272822,0.8058392026208147,
 0.04813211844914567,0.4393061831737971,0.45197530841540545,
 0.9871484036584066,0.6894389294753449,0.7087829971609608,0.1527563454186458,
 0.5470160230535166,0.4528649485341947,0.17229905671475176,0.256990919446213,
 0.2819068838604415,0.37517039105540484,0.380418413908997,0.7272571754093565,
 0.14216050615460418,0.9617066846815627,0.6450483376297471,0.9600720829701759,
 0.2469777662860836,0.6983959209065219,0.954588690448492,0.6937483447334905,
 0.34706090087087094,0.6357200333185067,0.4679806827138643,0.7174370340513117,
 0.3322005902591558,0.5831156437272347,0.24618996767976398,0.9559335564498697,
 0.006699709193740189,0.9601596838587845,0.5742085583967088,
 0.26642140977650974,0.7059349544225197,0.9669571820801464,
 0.33931755596795043,0.729825642425429,0.9756730289028095,0.8627587528534104,
 0.7846824769158123,0.6931187640708361,0.8519518607161,0.5074361920966239,
 0.5978373836882223,0.5979224801371951,0.32104499185346147,0.314099454537488,
 0.4740311320679814,0.20627008333088426,0.7066412045478764,0.9239584327671706,
 0.7653394769950554,0.21470680197005143,0.8719469823608211,
 0.13244463906143367,0.027427994085515106,0.4317669600983579,
 0.5315887546827864,0.2946660870697402,0.8640752716951319,0.9742610428433084,
 0.6956683112483337,0.4160432574079965,0.9333905893674539,0.2541563779338113,
 0.5373965588031756,0.20975151833087124,0.14855806002610306,
 0.6895507925891516,0.9841175808789439,0.2847923322714211,0.306928221052545,
 0.07103321482989411,0.4807787899701299,0.06958518804853797,
 0.8507355979501351,0.2529077257749157,0.6388143932299586,0.18379642405476915,
 0.3903409078329657,0.617733782819615,0.9398181823019478,0.5939070858664559,
 0.9177708087392557,0.5954544819972072,0.760411225999094,0.1714211124682138,
 0.7370586037295405,0.41454222232912974,0.5411544418676612,0.536406778891825,
 0.14481302252316686,0.7221224152734255,0.8233445882977081,
 0.12129690204014965,0.7559159400065532,0.8196652886823879,0.8255581159398012,
 0.2581307306681808,0.23863161517372455,0.12748835290510763,
 0.8696817525865113,0.7763875156120382,0.9068885246728515,0.20516339038309617,
 0.4884542384000802,0.25719707672782877,0.24488855671480747,
 0.6415893198335756,0.9646471609156436,0.5633089774946602,0.1740185500016045,
 0.5703706172972071,0.236155953583841,0.11472015813949388,0.3697043016188949,
 0.5081018573569894,0.5685362553094077,0.4488262364718685,0.05232330898522175,
 0.5103903204477263,0.18516166728050198,0.2759764705507771,
 0.15774622275990868,0.04021168202444525,0.8590046592252597,
 0.5068740455343053,0.35792246507785697,0.6776774966366482,
 0.16632891803527072,0.6840304032652876,0.6876594053444247,
 0.27768775408928603,0.3093194695811068,0.8680531503986528,
 0.19771102552083408,0.14045547841548267,0.39965948609469826,
 0.837132959426322,0.19364431911791924,0.8350327078891029,0.08942449577451028,
 0.800467766858288,0.2468114740681785,0.1906385908581636,0.2706350740920347,
 0.1629242011600206,0.17401198366168358,0.23205935262980903,
 0.2953127079085378,0.038788907229801106,0.9141666678007672,0.275049196289048,
 0.028991548785688104,0.7974603353790677,0.05780397416410976,
 0.6909608637915537,0.5922830594294749,0.16883991675268595,0.6652689269888382,
 0.3938125023312956,0.7977990682871545,0.799517059964577,0.21352823763412243,
 0.6387814274386059,0.4741427789771604,0.6563207951672663,0.9635840860790803,
 0.944311188325328,0.605199535639821,0.5307521480855466,0.6843682172080163,
 0.729892570694583,0.4720197632368557,0.6480202175464306,0.4128605427641625,
 0.0845666675456459,0.09053706368626568,0.3411728493490829,
 0.019362578357266402,0.6338890291041575,0.9604304124629909,
 0.5947898300209103,0.5076623329372516,0.8758801382043053,0.2886217671215655,
 0.39941474512271646,0.7508095795643507,0.757200706998766,0.45415347781929816,
 0.6458876395338377,0.9133198917291856,0.8732153358447856,0.7428983466060082,
 0.17461080281485364,0.7121654577161827,0.5353759851900669,0.7600913805465116,
 0.19008762572781823,0.7674006648518499,0.20722254309609267,
 0.5435761356094715,0.05985229642764989,0.5333818457882943,0.2450482672221337,
 0.280920151152491,0.4034901696567549,0.597790712344803,0.7753635493368246,
 0.146959534154393,0.3259537288041582,0.670557189297061,0.8940396984291604,
 0.15359740233982155,0.019716403309922415,0.23565456707056753,
 0.7872115490686524,0.36020956114436586,0.23296575349230375,
 0.8895849986021636,0.9183988520858328,0.03996385766711685,0.6610175709148985,
 0.5428193657350371,0.7615481885314231,0.5248453863650091,0.10793641707388535,
 0.3921889560271501,0.2548738702545307,0.21942107116480813,0.87442938621517,
 0.4328443133657811,0.7483515229865398,0.13010246692442962,0.7747110468428878,
 0.6537394833893323,0.8697115825762026,0.041949429177393305,
 0.10866892747794554,0.5872186330492769,0.7477620366447861,0.8758233059161027,
 0.9112328432087012,0.7983300359618505,0.22600644886592391,
 0.22967608103387704,0.13616045512488817,0.08520352944546061,
 0.7215157328297656,0.631611622218718,0.34501419969148917,0.6360825240627703,
 0.7484077969182725,0.4736142412346711,0.6633839762117687,0.4337020428219486,
 0.6214923445064879,0.21591451577634535,0.7932657272988649,
 0.051053478001159336,0.6935150919506686,0.19596791408843206,
 0.45273444776582306,0.9794923480538992,0.12158131839730213,
 0.2070766345164986,0.4275799342639196,0.7029783453841039,0.04402314011425412,
 0.16231125165113036,0.11448511915999693,0.19989206558002592,
 0.9315011668101228,0.23580657516495818,0.13007421168218103,
 0.8305914590453062,0.12966974425990596,0.5072536177905501,0.3769533150976062,
 0.6393372306979963,0.09472462642292401,0.8753260195973764,
 0.08815481366487699,0.40528799081884004,0.2961749928467129,
 0.9761687759946752,0.7745417860301713,0.49931785157238895,
 0.26738785026131817,0.6154531912123987,0.6501242331994697,0.2546414342708041,
 0.4448650997995732,0.06292148555846522,0.0610665038360525,0.5990028946524786,
 0.18523198677610186,0.9450071371710785,0.9782908061939019,0.7839942800097807,
 0.9557095379487796,0.8544029804235583,0.5683923558424875,
 0.038962575118322906,0.7929954390286122,0.21479437118392053,
 0.8326848528022773,0.6235738831241135,0.21107732314904482,
 0.25909962013281074,0.4938666754668325,0.5627721510403634,0.9559419902045059,
 0.16541550638149827,0.1802973773120412,0.8463687414777253,0.4470960378508645,
 0.6870771276918258,0.5089886361453972,0.6345099350773112,0.4778464106862552,
 0.8548600821897376,0.04497761225409602,0.11805886125791942,
 0.45742648296648025,0.26035050799198445,0.6627356913391116,
 0.6770129751394011,0.7359544081985031,0.09652914821710157,0.7521430054198892,
 0.3586040751830457,0.9747366478614096,0.822555102331519,0.9208578063343352,
 0.8197327058561185,0.4114706421060539,0.9626140753240578,0.7957338608415151,
 0.44513830238970753,0.2993781196446398,0.5687376166777075,0.7127158879929043,
 0.14002017695798186,0.7637990081887911,0.6816725361191378,0.8556604656746624,
 0.6651845313122162,0.09152808702456905,0.23800804613320392,
 0.20345719467816048,0.035240838238672634,0.8536662961775044,
 0.7171550010306366,0.1461269939923433,0.8675507196582738,
 0.053003576881347936,0.33548171224656764,0.5351455549852637,
 0.8889163857000801,0.6475494368342904,0.5609548275749239,0.8525689738052287,
 0.02526462475581237,0.8191173013242603,0.3178588748931156,0.4310519149143516,
 0.04509700677302586,0.7012561029232363,0.2647986277568505,0.040480789177751,
 0.5453358136726147,0.1236539063981541,0.5560582795089213,
 0.0025815130256795493,0.14241382081202947,0.5740005469541427,
 0.31150309050291636,0.5253613415636786,0.020876898635752927,
 0.973210458324164,0.7634294456474019,0.8602308302696972,0.44154403698088796,
 0.043168114373460686,0.7358145009066286,0.43228551957949823,
 0.7056230154035705,0.8437261809490886,0.1435532175629951,0.906656867531571,
 0.17385164191595492,0.5614878422645254,0.241362261101996,0.6737624392262009,
 0.3720092398849203,0.6138047787083509,0.09862252522486825,0.6147674583064905,
 0.9033169483744237,0.6957924644417424,0.3335432256466907,0.5028087932004248,
 0.9882757464346755,0.15640879240388184,0.8109045354613911,0.6674630446689945,
 0.4211070894353739,0.6640438041974909,0.0024531053637741884,
 0.8420221935937068,0.3558741465217081,0.428101387805636,0.33178951821077485,
 0.811938833088302,0.8075702754018131,0.5470743829275727,0.2267829050693071,
 0.6983566169839178,0.33546781966677985,0.7362660950775828,
 0.19300685689182517,0.7391990320311135,0.8877807007426116,
 0.021935403687589616,0.4756650473394186,0.6983442891444968,
 0.4451702757761302,0.7051201941229878,0.8741453806834899,0.40578301934681527,
 0.6722304884359139,0.8998673552425808,0.46646458287357084,
 0.40747760152474255,0.9562039925229837,0.0044780750442408035,
 0.23472726771860797,0.33294344972942935,0.9863239359150144,
 0.03551740749076149,0.8992468602128487,0.9252542318386335,0.4036532276674949,
 0.5823946008167881,0.8916516187958805,0.7348199889000032,0.3474577754712884,
 0.43592127317415674,0.5450202640695835,0.9516917110530755,0.664620139493076,
 0.13223798524337138,0.9681542507027197,0.019542885127640908,
 0.12429970712243144,0.533672738979987,0.035389728308629964,
 0.8599047081137895,0.21629880629532994,0.23910864543107957,
 0.3741086018587707,0.4709466177532371,0.12204223668764236,
 0.35303009941250274,0.15329781797864905,0.49911286352753714,
 0.9703122338189012,0.5159041144227715,0.4538135755003292,0.13182565254560086,
 0.22286935949198838,0.7326126429093485,0.13900518303204035,
 0.4777429114134941,0.5246235970977062,0.6232798233895908,0.910421241317845,
 0.8938119813096278,0.883793756695187,0.754737261517813,0.3501458189627714,
 0.24311919539409943,0.46528308700894316,0.49749157827823876,
 0.05634237074177939,0.0629920263311049,0.1157484930582462,0.578294723892928,
 0.6274645946543562,0.5411050786418315,0.8264620644681598,0.24475510275720336,
 0.0793649848740362,0.7515948559704785,0.7269511387507417,0.5141989624147019,
 0.8503763079614495,0.750667068019752,0.5624212039060077,0.882205179030202,
 0.503906308428477,0.3517126007198088,0.44162918423320696,0.6697026409416803,
 0.39866829509562696,0.8866478360552259,0.3210966703109014,0.8726659864734134,
 0.3648639473758003,0.7133126213799102,0.4938761301065818,0.30794788489042035,
 0.5815259314147093,0.6317718298555579,0.9960356892093682,0.5380653385789504,
 0.48461643375239116,0.24284846265997273,0.8864881641274669,
 0.4357364917481833,0.5271513837068525,0.439216495963321,0.6120585220443755,
 0.9398183712697034,0.18831558737610732,0.6830796682902976,
 0.11970424599217366,0.40096103899064417,0.017695842698307818,
 0.8978199505177433,0.11713384766715484,0.700388528451377,0.18328845190076992,
 0.4295383438621214,0.253047222571511,0.5814682950349909,0.6369470342580066,
 0.501175594761261,0.29658132199609033,0.9077404678723198,0.1721737542076528,
 0.2018819426522721,0.20167904680316406,0.6562243607861612,0.7833033266821251,
 0.11814519281824776,0.39839346485522076,0.22718796838701116,
 0.5002995930008443,0.06217934389929336,0.09910340575596155,
 0.2648144867567642,0.6376402504204037,0.5537131568405038,0.494044853256537,
 0.29597515398592555,0.1938304086260565,0.8353873540674905,0.5135301749506871,
 0.44922092672426905,0.6891190569557031,0.6635667952478976,0.9802214846085233,
 0.7435157985954357,0.16134413658466473,0.5795591620929555,
 0.037738557059808775,0.9074832911031824,0.7331798977702375,
 0.3977194399400288,0.12928050894654364,0.34482071316642693,0.259190920100666,
 0.5332011323830638,0.7944391221737332,0.9914574190755772,0.31870878996918994,
 0.24827535704569992,0.8736945907221175,0.807277570373168,0.9217833746583907,
 0.1257386410931708,0.5238673190520111,0.24031783183060595,
 0.07914015191995583,0.6695427955814933,0.5246774439669349,
 0.32506852612882575,0.9128267322694313,0.3995041796588731,0.6143900534415359,
 0.5726853021539184,0.7670982893603733,0.08252179178775543,0.6023425187802486,
 0.4386946777784505,0.6160701877034109,0.4411743146575198,0.7597785097342016,
 0.4248063077904104,0.9125571906437839,0.027062977677116185,
 0.2153017826930127,0.7103146025054082,0.9928450016658463,0.9731345849698047,
 0.6992920869164145,0.02197243509642466,0.3074404735703846,0.4118988058293491,
 0.9664061540435944,0.18464367115887814,0.4473720439343838,0.6932223287579053,
 0.09458840478043684,0.07699562137394489,0.48880655619622826,
 0.8323863923041341,0.1515845563289323,0.3918398672128257,0.09339445724740636,
 0.5976630245766216,0.751335946482727,0.6854827977463672,0.7353537716008501,
 0.9081117644248209,0.11816608258086692,0.7664923508212804,0.427181947669786,
 0.011127717539522242,0.7416459112034197,0.49223174897878574,
 0.9298593241607254,0.6555442296050998,0.7598249283195444,0.9359049732424135,
 0.5171249559138577,0.41735454486269796,0.8229307923356052,0.6574703308229985,
 0.5833256564068606,0.2554647050530534,0.7686421440844001,0.4516746980074883,
 0.6979267465876354,0.2545911175218082,0.08166796455095005,0.2785236431361141,
 0.6202140022068503,0.7070697752701637,0.5834557091722787,0.8187085656438072,
 0.6070782361600848,0.16650960649868907,0.32834648704163993,
 0.23803306466552998,0.5581927574684556,0.41388038675960026,
 0.9965193795772112,0.8165286240203002,0.045286354085682046,
 0.1349655022285957,0.5187882964662593,0.2565564878827252,0.3065618386032225,
 0.6258958638370911,0.18769177557295802,0.3752780854523171,
 0.23399170833355143,0.7974687030425007,0.4737798350110116,0.9377881927278621,
 0.7712528618311241,0.7464152943621416,0.34176427193091397,0.8760272395869091,
 0.3083421684372595,0.981271775154265,0.4240827231543567,0.6258351089571288,
 0.890067415197326,0.8113608132604746,0.06801087212844081,0.5711006502678067,
 0.6212855202195344,0.08387548858787852,0.9978970253842396,0.7231463246120197,
 0.16521110840126352,0.35995118736148635,0.4110586199307461,
 0.5865896913240405,0.1694414086335979,0.8570345751328778,0.721951339377637,
 0.42025125411854225,0.359386814759321,0.41013483690394903,0.9379783652121151,
 0.9524163881891764,0.16101999922074506,0.6078696391405747,
 0.26922015475952443,0.3505998100167572,0.2628358411871672,
 0.04126844355698389,0.4783204732813765,0.872437575352036,0.24354217692968594,
 0.17794184449695816,0.20585459342649937,0.9912648742741142,
 0.49571874621727785,0.8251411706049426,0.10261792149270754,
 0.3773377179491587,0.2289615132264914,0.9577093811632389,0.45825794100196104,
 0.9179601381767568,0.2591763569500797,0.8615270197505143,0.628749061862819,
 0.9390270582040255,0.057860155255072154,0.8323911349063375,
 0.3655155356176981,0.610106607658954,0.8604516388501592,0.46944852242824964,
 0.863972629126301,0.7905387624200214,0.41018236322073653,0.14434619440710883,
 0.8361006177542876,0.6085823389726953,0.3005363387844693,0.4928332372777423,
 0.0502166605890364,0.7122307388658168,0.08815670734463288,0.4211786136236111,
 0.986293754562691,0.03552668696965078,0.5375119967646553,0.10691777039353767,
 0.6943109011936893,0.6193495361702543,0.7463312859065915,0.16905287285393056,
 0.7275007504467097,0.4116344334421205,0.35910126181666,0.8083706475448675,
 0.33132602620650187,0.33063798875772255,0.8138043640252332,
 0.7602346608258711,0.0457330219385349,0.03607282891328345,
 0.42282253449228846,0.5949223058567124,0.5910053820167409,0.8567640909786376,
 0.08688948691183596,0.8836179170858697,0.7246878333807176,0.2987337957775781,
 0.3769754306715276,0.93887980641356,0.6475688733489093,0.11094239550790186,
 0.7759511017590242,0.4216323061389332,0.1317591964707896,0.7955312478245758,
 0.4061115654887053,0.6662443347493644,0.8339388805324229,0.6424733067357201,
 0.8241814843184552,0.07599706070660661,0.07370415103235939,
 0.8607982377256674,0.8105348844872008,0.8048589861042386,0.22135121823722126,
 0.6089922192564367,0.8984493819140514,0.18291747381382728,0.4905709291298749,
 0.40085355153757307,0.43156101675179714,0.7431280983154492,
 0.6549638191459292,0.08073193751739416,0.9503164875885107,0.7070492515751249,
 0.5200191386726568,0.7509099486451315,0.1696869213730129,
 0.020309424360845707,0.13208977702387625,0.7808827508420613,
 0.7945916219289889,0.5068364541986495,0.5537918384718303,0.8995388112187087,
 0.42621485270141024,0.22716851129183158,0.45485417955996355,
 0.02455638168159624,0.7951934718016992,0.03625735602442692,
 0.021615862506376926,0.5703585042439618,0.882356739647526,0.9260798634943259,
 0.6538992510586168,0.20396813459202567,0.879784197317988,0.6964903046434535,
 0.7912795432780015,0.19734428105384638,0.9829010300227377,0.812208740639225,
 0.7229041206452365,0.6826498201627921,0.2942266815306652,0.8718067541200204,
 0.25411169332047245,0.38692269703248916,0.5444527695187148,
 0.4376699207700895,0.9486461987258274,0.6480142051966408,0.5456610094651722,
 0.9984934897902008,0.26674742573442844,0.8318007670766333,0.5011765047136156,
 0.7243867636942177,0.6207087100261947,0.45417370733335793,0.181643150160528,
 0.08746849067015916,0.9064734330299423,0.9311288846234272,
 0.20888574269714877,0.23115212552262776,0.5856803179304121,
 0.3623628369822932,0.9731352861904771,0.06380231798893565,0.0787936616517676,
 0.2493637890294892,0.8549505111466807,0.24672463972098535,0.9362810393520533,
 0.4802885693100617,0.8358172939573593,0.8665036154129095,0.8655323314519101,
 0.8905158951955339,0.4888273858773424,0.32024023763591536,
 0.18619063741878772,0.8563600191463863,0.5964841917746618,0.7315541221067322,
 0.8471144113081371,0.17165334145404187,0.42056279052752443,
 0.9460968867307293,0.45185795690404806,0.28636905679928626,
 0.8402045091828013,0.40427709856022387,0.2675616809605499,0.5849675953677014,
 0.2793117498623193,0.2528846269526126,0.21589878062422718,0.6832141763625545,
 0.40444164826320184,0.12796051510013362,0.002880875706834707,
 0.7328187793813313,0.3755861723828321,0.8434496899107026,0.45072546471036246,
 0.8682418462005457,0.9504153704041778,0.7970820095510623,
 0.036084859880511355,0.4525043218451642,0.18453489151458424,
 0.828043684632039,0.1963082072328508,0.8098825536890824,0.9966872544653463,
 0.914853132251551,0.5968600016601735,0.7436730174028137,0.6555171070203426,
 0.3334811562102066,0.2874445525004069,0.7297926499256671,0.42869480769279855,
 0.04658298511442682,0.8762688379277129,0.2613033976594489,0.3129219378795719,
 0.738975627383064,0.5348079594756565,0.3457039560868156,0.8297854059902879,
 0.15045433735476454,0.2166918995276358,0.5226809836792456,
 0.40415302266339614,0.17372808704742904,0.5198266262469391,
 0.6252607028821989,0.6173124406641314,0.20830477003405945,0.6078959342490534,
 0.8862977967791539,0.5530366014247282,0.5239976373631844,0.12054222593461628,
 0.5617351490851601,0.7098063147826484,0.7409970512779842,0.17758364504152768,
 0.9523333493163083,0.7974111243716957,0.02756822909326495,
 0.053547184306520834,0.1257837755857737,0.7659362400729406,
 0.6819135614896135,0.2482435160817812,0.8284875163270802,
 0.024585159267773937,0.793823666381006,0.888697667627827,0.7860366934701607,
 0.13170255093030614,0.3812667199650521,0.3063238711302274,0.9043020410692535,
 0.30142015208310735,0.11519656623432539,0.47217105027081785,
 0.9035820026073527,0.9047871729653189,0.5577119094396112,0.45806184099437963,
 0.25722783523670145,0.1559274972320127,0.9151063748902567,0.6156687170671387,
 0.7241787392909041,0.09321399275206943,0.015033006484325373,
 0.46220417001254466,0.1633140918681314,0.8771473409140116,0.5121726566048578,
 0.4840382979363256,0.20760277035564068,0.29070894072648734,
 0.4531139572585505,0.29224477680681793,0.26803487413550897,
 0.32362631474004344,0.3434751742540524,0.1863690696219107,0.6992243732662641,
 0.2344945834328046,0.31620919372647693,0.9588622529531375,
 0.29377806497200565,0.9900965236210093,0.38728559207964075,
 0.0038384774692094004,0.6456756706465623,0.1995537512857657,
 0.9078229869326694,0.7137801717034399,0.6124767593847341,0.4517833080630138,
 0.48700612630738016,0.4619639737312431,0.4700213454773925,
 0.13168622873253366,0.005118080391061652,0.7349318451848443,
 0.7031667850766536,0.7323213908222644,0.4346172311312605,0.5021796040998314,
 0.3036958794240796,0.8674084635306348,0.4191264866636789,0.3777363824740416,
 0.9104410092766021,0.4874199708058007,0.9513082483970623,0.4161211095149904,
 0.2180246588244089,0.11970046768180187,0.4024282140060419,0.1445461130721447,
 0.1912406641645087,0.2365753431566825,0.2639649436458763,0.9640896374514415,
 0.29688018379755765,0.18297642123604907,0.9218988220115396,
 0.4839553224110271,0.7123118288001178,0.14360577401063057,
 0.11704113220607693,0.41391377322320244,0.6825129179911215,
 0.3888414890052878,0.631788406895371,0.2614559809807099,0.7660882345040381,
 0.6306755114744754,0.023248859910797504,0.9253851600196845,
 0.9920258502743272,0.22824576431932586,0.10147516620216634,
 0.8094883201479296,0.6116428997909673,0.8829621525435846,0.9591606006959813,
 0.4679065723513519,0.461120760089603,0.545149893094092,0.8411972091463006,
 0.5292637831391319,0.20026405409830206,0.0405536084098147,
 0.22198313068526343,0.9312674720044597,0.12884567035198424,
 0.8520306845547188,0.24048403770631066,0.644987017083633,0.06503445943823194,
 0.6107350845657642,0.024277023291731314,0.44794066043072744,
 0.016775267251404546,0.7492646864009845,0.8039114839930802,
 0.4441617151832479,0.08211966849247132,0.07972728143642727,
 0.2520198937029927,0.10558090263321962,0.34224066795687713,
 0.7962144740411432,0.3008612792588956,0.060332337153396254,0.918108109584263,
 0.7371972737667728,0.6608402711226848,0.6755063900991025,0.1324069546266895,
 0.2150658217315894,0.6329413343895409,0.678863256322589,0.6272419255613222,
 0.00818294730032687,0.32483829159655053,0.7632909364764553,
 0.9702989975997716,0.15911764198798306,0.13040206828166434,
 0.6774597738378343,0.5938246301341525,0.32874070805635947,0.3051375517217534,
 0.8974006516013551,0.31413125362925287,0.13383901259610442,
 0.5961103775149605,0.46905083651951696,0.8292331524343733,
 0.037887175683583796,0.7553401635202239,0.9862016253554369,
 0.32926304221364855,0.7685271993609826,0.2948845240073179,0.6537611052553658,
 0.4227663292235766,0.28920958773750427,0.5762146394240257,0.9184475611578655,
 0.6770878707868337,0.22465365720391606,0.7452523454621983,
 0.08761683239747153,0.19952077245936817,0.10011571661894059,
 0.04620987907350815,0.8400253301222544,0.3073377987161491,0.7095424040621369,
 0.19538087991062092,0.39399019838369387,0.5359799145894424,
 0.5372123167814221,0.9230572809429684,0.3893611753211862,0.8489510167491,
 0.33469099236799216,0.1408847115893651,0.9824980371372608,
 0.019468383503109643,0.6799071677996883,0.4978459870703472,
 0.4583758927385848,0.22037899365664515,0.3922992727471243,0.8535617356333891,
 0.2873387776886711,0.17851751529429805,0.5499522739390326,0.5906679393480058,
 0.6093939851245649,0.8423377225361072,0.8991855830659645,0.9970423230662128,
 0.06220424067187513,0.09212110868103496,0.6259156203423873,
 0.7272719179337497,0.4896409499694583,0.6404116205586484,0.8429522293199935,
 0.33687596206711934,0.7803794605216406,0.5041460994904844,0.3478151855766336,
 0.815649466045378,0.515147783952216,0.6811479356709079,0.08700048630414359,
 0.7504306836695919,0.9642781884402262,0.7830587648242997,0.06086756904131996,
 0.417450609064005,0.1408904289931211,0.4952262883945364,0.45471180073428574,
 0.4072823270250656,0.6572319467918456,0.8213065592370591,0.20216973461615206,
 0.1323025067615763,0.9443953159912809,0.9271026046633871,0.17149691478210427,
 0.8868345606450633,0.25924823828919985,0.970031266603564,0.9245407423174319,
 0.2769665335676149,0.8321026268935775,0.7461801784936498,0.16903908167304038,
 0.26224138008830633,0.6614257016778107,0.3582280153273102,0.5830644369571263,
 0.26305025512463553,0.9824361944799298,0.6494794184836791,
 0.14351213083551118,0.831401406439878,0.18323730588247755,0.493786821308688,
 0.19867791427662618,0.5824087986744958,0.2585426372781713,0.8145707437612146,
 0.7826345532989554,0.7449055156700907,0.05586881457794646,0.1794266065667096,
 0.3118631825882875,0.11007001436330788,0.906094933509409,0.7969081484537471,
 0.013581599461324756,0.21861031549904775,0.5828376973749522,
 0.6922443512075506,0.07602463727031361,0.7171829982349394,0.9720874178321216,
 0.8973751572664026,0.34761109748785024,0.33467530374135357,
 0.4033562054824785,0.9485161258820937,0.6252510204924244,0.319000196833079,
 0.4704671633782329,0.0980477165317506,0.6213303799943343,0.957420913996607,
 0.23809478034394416,0.22866920936018842,0.979606086725119,0.9022261147934709,
 0.7821113728004446,0.960272019768231,0.16644585510815801,0.9105621485775056,
 0.7024210801004213,0.15083302963032286,0.8444910369071087,
 0.19314610425284295,0.23013949779643916,0.018782249049329347,
 0.8171859753392894,0.2911848798061747,0.2918697586423069,0.092124083992973,
 0.4464162704911201,0.06265313290713037,0.7121577586308795,
 0.21074896924691833,0.6441094881893867,0.9525672045538374,0.9859127794635612,
 0.8542415794858648,0.12936688903300886,0.27402811783068615,
 0.8819087954817734,0.3056546621292281,0.16220948134042568,0.6177480657391236,
 0.24460640745968598,0.4725602080014719,0.8428775346583031,0.6026392612326893,
 0.661113679532738,0.16209842944676978,0.1752136356741716,0.4850755751181276,
 0.14161714237368028,0.7329499664712644,0.3804344790042129,0.7707667272362643,
 0.49652102162485934,0.44340227812944344,0.7555983213833294,
 0.3295281330028541,0.253409597531559,0.8512900028062719,0.1062826094639403,
 0.7820102834500514,0.14153222211769556,0.458565937964178,0.00833212867928923,
 0.6992608056802302,0.05946376677452703,0.5083712585061352,0.8221403599462322,
 0.4795520454577886,0.24098500321459304,0.4966260009020377,
 0.42717719598457293,0.22131261822233395,0.0980269072216704,
 0.6465147058343321,0.3507122009892697,0.04542345178844087,
 0.31170581616698323,0.56133766044354,0.708145667310757,0.533710381336399,
 0.06050971917516501,0.5412908057071395,0.09073451963695933,
 0.5732130561784609,0.32586609278497025,0.6860657579378413,
 0.13805031881306573,0.23437939829214083,0.3425179754006602,
 0.6066327048667048,0.05140806807044951,0.6881429526101894,
 0.11974210744179459,0.4776220458463539,0.5836920196888034,0.7720938273718937,
 0.7993777667625985,0.2737097748904128,0.02456494176885682,0.5123620177821506,
 0.949687633846694,0.2355027246812963,0.7567974197664994,0.4712299945365137,
 0.6407352278430767,0.7454619557887292,0.9648734077389346,0.6186725311078354,
 0.5602825192316228,0.7555312873260267,0.9267176546952237,0.640045401319469,
 0.7036362868420021,0.4490303794438856,0.9917806904517925,0.34581747775023486,
 0.910310957632527,0.43284540654477965,0.5917788648594188,0.675439618881204,
 0.6638469740652732,0.562805874747044,0.45475059491032765,0.6987464997540026,
 0.3847613103571216,0.4723835960116891,0.26951794985291344,
 0.03114601598400557,0.36504929352446647,0.5952544995962342,
 0.6382963422194518,0.9386173732742944,0.4697590947020379,0.34411240923836595,
 0.7259420278731122,0.3873012761590314,0.44153711000684615,0.9957953625540276,
 0.2448659477924361,0.874139049762351,0.7181828378015543,0.7466534283158002,
 0.4865618977278655,0.5070545408299894,0.07648178012335005,0.7797071294640223,
 0.9739131514138315,0.27597746088504826,0.8834518085279428,
 0.01689335282224369,0.7374592024529816,0.8805428484894791,0.9427354478444951,
 0.009873296261318387,0.4409087830311843,0.00587806839476146,
 0.7661294443566076,0.8699394521393566,0.842802683612407,0.4626049069770981,
 0.3689741416238621,0.7051027550920461,0.019079248560454065,
 0.6306256328621475,0.002406183246752791,0.6019637779613561,
 0.5973663757276438,0.19771175315345502,0.6415467421163069,0.7268001239368375,
 0.10050110379026378,0.6983622146979789,0.7970975442563023,
 0.23471942394462353,0.9247937293849969,0.5983776658732936,
 0.12240933363585849,0.6108042977194756,0.21244068412634853,
 0.9573154576481296,0.645272274919957,0.06312651989035223,0.0376595592723864,
 0.009925948308514676,0.4284591994514191,0.5422698951167023,
 0.5663873924365739,0.9998635542422474,0.060056756612508866,
 0.41223728563562834,0.18909297631471622,0.4174434102227723,
 0.8442578308909867,0.6988538402735464,0.56954831675102,0.4719803429935887,
 0.19557899488896713,0.04030502999295926,0.1415271900988656,
 0.03911622713194651,0.586521718691898,0.4823406891816897,0.3470812270830109,
 0.6101718470370702,0.19795769264981167,0.6448783950605529,
 0.028024914809585355,0.994513411441394,0.30455397862309486,
 0.24861011021415813,0.4574516905925212,0.38119063580253065,
 3.0755125493808766e-4,0.7359466312109683,0.8559763909010181,
 0.3051554826425532,0.5133873385736429,0.4255211827179013,0.10881785781249076,
 0.17056421583473413,0.48546065899297064,0.002522680503481256,
 0.6815294241096338,0.6045513133311722,0.9312565582252335,0.0812662967978397,
 0.21189214012606616,0.00488178193144273,0.4371061320013774,
 0.42233946806355616,0.8716713064612556,0.5423393517858976,0.7142065460183415,
 0.6713163663668467,0.14412612127794355,0.9691729086143416,0.187357012715057,
 0.6779576826837813,0.1403178282866555,0.3697916578051579,0.3222324740042677,
 0.9058894164288782,0.49134620544100227,0.5494195914396129,0.7842388053829524,
 0.31102885995335083,0.337668031460751,0.6914206478464453,0.7672466947027892,
 0.9782707817084201,0.7633939795290244,0.3182865878482337,0.4490158025206633,
 0.7223891688662842,0.07659615681787602,0.11252060560473531,
 0.20189236154225665,0.43242426914981913,0.97198977913831,0.09052038181327082,
 0.21248568728547657,0.8146134746131195,0.0714047100063443,0.6814309423318801,
 0.8685774224692442,0.2824410065351013,0.17169702978875978,0.5413829513738165,
 0.08926347296545689,0.23023993394903908,0.47584102166984343,
 0.25899241343452495,0.6410791269748072,0.018117694896899295,
 0.45795628131453847,0.11751043429054386,0.4638531660958999,
 0.5523693947046231,0.6378341857841613,0.7521125187593387,0.9670020734340746,
 0.49988460325549555,0.3463660054797374,0.5782263682969526,
 0.03271196815771171,0.6014800514379297,0.5951544228185253,0.5065785020639438,
 0.714092964439446,0.5176615893147157,0.8109315394673319,0.3311096679949007,
 0.8360346403715969,0.01609723772132665,0.6614469834930862,0.7263208507593697,
 0.7906933990450923,0.7163802957384542,0.3847475438040834,0.04537838752417889,
 0.36580120839583563,0.748777045195904,0.09606866011175619,0.5104176251801197,
 0.43272311807219754,0.11704932240514954,0.6211992233419047,
 0.5973932078219738,0.5544949730998963,0.4102400739086466,0.5619676818606373,
 0.8909728014237156,0.1863155790849459,0.1427249204238058,0.8384208118023079,
 0.17574561546347733,0.8502672206459269,0.9286067015943524,0.185963034328971,
 0.7076709753025275,0.013680313541627687,0.003047735394374973,
 0.5704357159095459,0.5584302424631247,0.022304597503960855,
 0.024172911424007903,0.6993114886247865,0.2888668508938077,
 0.6013592660579146,0.4605714063947881,0.22559383838517033,0.9387806387457635,
 0.376937109661731,0.657852064929949,0.6024590243491748,0.5697379355704111,
 0.5621631755811112,0.5082304627342479,0.4850524858108588,0.2517360783610335,
 0.28836321793495234,0.0518756383726906,0.7624175647131732,0.531600913904178,
 0.9421529190963303,0.9492253987007513,0.1673295061971688,0.31250922900635303,
 0.3267563866330473,0.6140422902399442,0.7894106504353832,0.00960372241617724,
 0.46377922448622755,0.1648186566552885,0.433021104378275,0.14932081587790103,
 0.9598664982655786,0.06078313676984559,0.9598247171524508,
 0.33216400140566815,0.15250370583919637,0.06919016204193429,
 0.483128863776075,0.8197226580467158,0.16179103666715688,0.3377333012015771,
 0.47066980119589186,0.5483186446709367,0.365015297761473,0.773687301255412,
 0.8979347928630491,0.2621156208112454,0.20566715233983734,0.9145311995689687,
 0.8986143416566632,0.6938933371704012,0.6138072839743551,0.258967682028294,
 0.19122636443783902,0.7271368907290785,0.8165994290241135,
 0.07330284644659635,0.7596811795514421,0.47733143671782363,
 0.7521385508661177,0.1481263107503844,0.19124553479219852,
 0.05742625740429319,0.08427265610212387,0.8467377716025762,
 0.12466533029182536,0.9721391600899111,0.5905822493490162,
 0.16651869085340865,0.6391468570890604,0.5857073853123004,
 0.042826761063825214,0.17496567902986337,0.12557435295732244,
 0.7384831369643566,0.21349327681956365,0.9732622001324049,
 0.26477868122347414,0.5532830977253769,0.6352111227869379,
 0.021763441512164583,0.44856137104054516,0.21368516927952164,
 0.5785643504265778,0.6128183518927603,0.2623026888123666,0.5295111425948621,
 0.45004343967803817,0.07612069745853378,0.4371859150006965,0.542711174448906,
 0.1680923291496068,0.6384123845874953,0.9184746762085759,0.7565285262721918,
 0.11513368068220031,0.767203432529428,0.47676512670046745,
 0.22178800708168156,0.6403306472603357,0.059143774005373184,
 0.1930537893269083,0.6770011565587645,0.30439303940954276,0.9288871667545068,
 0.7430985422426996,0.22084543492029562,0.9458515739893443,
 0.15395749438156048,0.7734593063582526,0.7205209107229795,0.9286837248046979,
 0.5298992406442654,0.21722047293075397,0.6818887167924769,0.8386421191664098,
 0.17186597377333146,0.5354327724288345,0.15713604130065617,
 0.22790137443976988,0.18798104246905634,0.24172183451947526,
 0.7712310135862271,0.9000448797031492,0.31275295282897164,0.8531783745876034,
 0.5581020781491519,0.7816494322726795,0.6630384331490727,0.08742669560604122,
 0.701150417540953,0.27917916891259564,0.22686635858023874,0.7711573607727177,
 0.8146246821423049,0.5732449783323758,0.3722725841634533,0.682901469771442,
 0.2968115278479029,0.32594913166083317,0.841146418675395,0.14718572549358366,
 0.43646127002527213,0.5603856604261621,0.3982612449522611,0.0940910164937474,
 0.31030053624608045,0.3786409840585534,0.14301834924056922,
 0.36776770136718306,0.3675447626406597,0.6041594269682506,0.8382481655076525,
 0.020206434325009792,0.39488230412941394,0.8789915468474716,
 0.9513153948130705,0.6324811727880606,0.08369973711120315,
 0.19033078813037652,0.8804107248413584,0.11487003231409021,
 0.4728131475376248,0.08579561418915937,0.008672278872720707,
 0.39898273838645215,0.482380264503503,0.2792828824868947,0.6713639419546491,
 0.23069168247854277,0.8242995840929598,0.4105448667639372,0.8217695496916476,
 0.5908546263927348,0.19181374546171615,0.24882312880989732,
 0.9248267964667198,0.9298466873473303,0.1837251579723641,0.2833835588173058,
 0.37131854240816176,0.4234986773302909,0.8230710950033859,0.5184196354661063,
 0.44942389842091945,0.12410030171373487,0.782947002078646,0.3773286033219193,
 0.9563583870841434,0.9529124955265953,0.9056487222331084,0.14059547296499797,
 0.475767799848674,0.8413331214043676,0.5636654857795322,0.6820025923549194,
 0.6908917780017909,0.0915027731012289,0.0767961308207219,0.4092396941230313,
 0.5172251727603817,0.5975263188817431,0.9811506448729918,0.9461976439163895,
 0.9555750308125595,0.8920430018437008,0.6527683421965196,0.4757947622443597,
 0.9979903608338933,0.39851335511789854,0.8630781615473975,
 0.32846407845376424,0.013408744707892772,0.32694903644855744,
 0.9861531626384352,0.24846928231536625,0.9730328999478719,
 0.46929126316947634,0.3393934546791173,0.7286183495589821,0.9085411701588272,
 0.38393136462234156,0.8241193885609527,0.5805103844349557,
 0.13275566119693472,0.4653236814898457,0.8717476353680352,
 0.029417368066267602,0.8508750169483023,0.35140615387567464,
 0.8323450350491757,0.9642196505434116,0.6027537870732438,0.5915710672228879,
 0.9593098777227398,0.7775511994137506,0.5677622312379942,0.9266081121278641,
 0.8805364916353882,0.16702118475370653,0.5151455669232567,
 0.07409938137281646,0.9078104009360747,0.40200052245605045,
 0.9726505436212336,0.027775001523638432,0.2917751290112649,
 0.09021026429353385,0.815901323198403,0.8548726079476666,0.2649540844463083,
 0.555110941095313,0.96404860618876,0.6810757850870957,0.5027654662214613,
 0.05940834124498451,0.1289849973059609,0.48926823686651544,
 0.27629963899229404,0.34969993068908944,0.6536042086265552,
 0.6759612168743401,0.7332312816563554,0.49650737084332197,0.1319620935226311,
 0.05779310462265197,0.7795024642412087,0.2557749588353537,0.4087697126216935,
 0.6430003652897973,0.7642075440120216,0.6135761060138885,0.04502046199555809,
 0.1452291862188193,0.7610662588295484,0.3536562772697398,0.24980923204504513,
 0.6024557111672191,0.3387631513416408,0.7336814939621689,0.4129918585130308,
 0.3013660126639005,0.9089106279187308,0.0503525491576402,0.1291606872904827,
 0.1463538714226742,0.9997241845856646,0.6705617078298001,0.3728455260610205,
 0.47685582250358605,0.7759127340890193,0.2793196979221573,0.6466999379710238,
 0.6340676158953779,0.3809214688161655,0.531523889554992,0.8260215507376159,
 0.6146344846938147,0.5812671547621516,0.22012591122320035,0.5466901583457656,
 0.3099706073346462,0.30621076173583384,0.5095955876053047,0.7582710261895779,
 0.4454833770164155,0.9403072620162048,0.48443357281040345,0.8773982706829377,
 0.27006918303214966,0.27588805616663703,0.6683551403444379,
 0.5950869337832049,0.9382259192520912,0.32323934881093386,0.7785457731918579,
 0.6525638579343844,0.12100727613691109,0.7009372623625718,0.7779508733817452,
 0.616047517823899,0.9430560235820888,0.9906451739988786,0.23275954215487515,
 0.35163835824581124,0.8535146821752236,0.8649811894404316,
 0.06657565726451964,0.8561078065854443,0.851630339180086,0.7928447964626856,
 0.5237167977801217,0.29729312181512335,0.47324237646474565,
 0.5086171618603459,0.499437638046226,0.012147009673523668,0.6685867301751036,
 0.24047135290355648,0.8314621391653655,0.6578750905088147,0.9894635268873422,
 0.9196438850145906,0.7513039279560665,0.9835771360489657,0.4336794166791802,
 0.6960523244300041,0.657850492630835,0.6408319155850946,0.14085141211867058,
 0.4955598243337124,0.7898597475544734,0.943828747489804,0.5269782890382938,
 0.9039210048384427,0.0016950190108364627,0.23205308768705712,
 0.18724964221063212,0.6659505985088999,0.6058489502293983,0.8021907713074802,
 0.5582995555056249,0.033281391439972774,0.048361154598193345,
 0.48758696136062496,0.2187256217894722,0.9175695241037113,0.879052973966865,
 0.23380902743749765,0.9655137610645249,0.9224770161658576,0.7476290637664244,
 0.3038526618571711,0.12182620867039795,0.6882849899741675,0.8707522126192319,
 0.9739999189377541,0.3511314115360211,0.04966178027849699,0.6963075704798022,
 0.1302078671673813,0.7686418221623528,0.8393851842197231,0.7794718457342708,
 0.6639292012080542,0.6774235950816981,0.015441544254026773,
 0.8783877137321028,0.3815060850366533,0.5856452442591399,0.8727464868978627,
 0.37206079066176656,0.760792467494113,0.8267300596704634,0.5422893564782195,
 0.5067537844241418,0.1884824238132191,0.885568501757996,0.7520876712963892,
 0.7083973066312241,0.20639551784631172,0.8808458891016107,0.7379471747009672,
 0.82525414144537,0.8223506747616831,0.49160651492303775,0.26004063780972375,
 0.7869931833705992,0.34820817291182316,0.509427457292377,
 0.005658769807060815,0.5058770859235772,0.6212935043604854,
 0.33468576365743286,0.6039900292880365,0.16864498875650258,
 0.19412648178029013,0.3840517843951884,0.5333631425452926,0.7596876631141891,
 0.2692842546010321,0.4813152899190235,0.9230767527342436,0.6045799044084181,
 0.060456612337066895,0.7221024699187429,0.78875154959485,0.7616539627760887,
 0.1586752562569258,0.44829432014538484,0.386430804861981,0.04523387391406053,
 0.44964912693341585,0.41848236162250796,0.9144297868270305,
 0.2543700275655302,0.229960389783329,0.39322567327303415,0.6764901096607964,
 0.3163607512993861,0.2773359527143491,0.4419374230432058,0.595317092391144,
 0.4951641264850368,0.49384470085203436,0.9233605559621475,
 0.025720487524222624,0.596574996435921,0.8438610723119715,
 0.24131978990289393,0.8765585062541476,0.3944891947002148,
 0.20937557555824282,0.7292482867336383,0.7608515856729066,0.8337441945469892,
 0.9865898094022361,0.9389876514112568,0.3415863005269175,
 0.020558592452361135,0.5853238105840437,0.8958535009032135,
 0.19923089658736104,0.8546984665720885,0.30641655588930394,
 0.2423807030579892,0.45354099593591657,0.4577495976483974,0.3196732111559767,
 0.1189431623477325,0.07970444261275911,0.7340009387475536,0.7852277789889217,
 0.7617309470442492,0.3718764087520834,0.6732167438138758,0.6427933628153937,
 0.4911167288796445,0.7384287417377238,0.4819390621337928,0.652415831588504,
 0.19693115809611494,0.11937764584645794,0.6883724755989937,
 0.4908869341379841,0.8043691403959916,0.525054605043659,0.40952831385223554,
 0.815133671847174,0.8942592418920593,0.3649754890506498,0.6760729818520124,
 0.3576465655465195,0.39165946432690135,0.8756320201054795,0.9063790096586518,
 0.9358130550369106,0.2535371421752449,0.89552312421086,0.8227723486843725,
 0.6583492719917254,0.28271385851208364,0.5324792201326146,
 0.42019391388284955,0.5992061073406065,0.23115421948903936,
 0.8420692155750436,0.5042872081678884,0.29943180049419693,
 0.45371740840636665,0.09111128049874684,0.5733038835386273,
 0.9919095114282707,0.7683143555983031,0.030406491064672858,
 0.6955392069807447,0.6959496005090138,0.6861353425469111,0.47016147937262553,
 0.9798587768637725,0.3084598941340002,0.007047386727785598,0.546728860912224,
 0.33009833327937366,0.23610259721741744,0.23361074306211982,
 0.7244626747998656,0.16715641601325892,0.2927357917246879,0.0954320378510688,
 0.47208206144366827,0.6315720371805589,0.18112748057004124,
 0.08002627587529498,0.8703239786859908,0.6913411230014684,
 0.08740594975555449,0.2147648204265481,0.2438317605520428,0.9667799482632536,
 0.260438350897931,0.7240744359923458,0.8160768960754116,0.5566618032438709,
 0.16913409432303872,0.6560619069813939,0.7826592411180572,
 0.42800809125548156,0.2827354940076071,0.23861397907337212,
 0.4929606529394289,0.7186129032396862,0.7809445451967828,0.5701275064228655,
 0.31091865249024,0.20816381956155872,0.47248355057589997,0.06852390048524493,
 0.7253559055240348,0.342023943840683,0.1762982252995139,0.3429463035605971,
 0.019540193801913852,0.4529972300469809,0.41565207451827413,
 0.43646670427071754,0.6679951873665048,0.8128147595758161,0.1647237781233034,
 0.669209507801932,0.7471077191143957,0.07647665222949562,0.09970943891412654,
 0.8828718655661937,0.542970196300319,0.6591477353688576,0.30426575923459676,
 0.8938136458845453,0.12699944434900434,0.49802273772239425,
 0.9633791009921798,0.8662885711337267,0.7631686709687632,0.8351109389714128,
 0.22918049997019252,0.61431496349356,0.06979085589308964,0.26339692160124684,
 0.12371196331734546,0.8536262649879243,0.20056098184333304,
 0.9636594954874966,0.8892221840154025,0.8397833505641847,
 0.051483884111214184,0.12541694024373096,0.66520427277776,0.6990567607824341,
 0.6870608792713594,0.579238135086055,0.902608216628094,0.6225082705826273,
 0.11543991997214342,0.896897506069246,0.9596364339330108,0.38666911587562935,
 0.9137793743229867,0.005142853571517447,0.2071354000157939,
 0.37926087702587497,0.877808599366278,0.7164001354660817,0.9098097827048888,
 0.8384055158184741,0.805573361049241,0.854615100002952,0.6047261694947434,
 0.9178164826424169,0.35833993495236793,0.3611790658954801,0.9609378593189338,
 0.8629192607025576,0.9536388633081966,0.5946461920286487,0.470039002359827,
 0.23309014180601983,0.11577269485749486,0.07760988263547375,
 0.6971684229820787,0.28518728363454193,0.4251347384067675,0.6859667964005305,
 0.9971129937492276,0.1113521444485851,0.622434092857798,0.24252487637487974,
 0.39220481928555073,0.28550408702619556,0.29936917911705163,
 0.9284253841178863,0.6560355786599416,0.10463564336824904,0.7481575039714492,
 0.8896637249481011,0.9436863776708727,0.3013640743083257,0.2871018973682852,
 0.3749612001543847,0.5625315603208733,0.04470068540505456,0.7897474892075852,
 0.3724546933177991,0.9205706000789691,0.5615069582591673,0.19804880237124856,
 0.03377956535145221,0.6336718812039459,0.8065084977379136,0.6109326559420052,
 0.9677922691888722,0.19011205268942089,0.5513769571881826,0.9446856632726413,
 0.7984185678163863,0.4642546534625278,0.10039047615427255,
 0.07232187777240717,0.9253581763057098,0.4458880473489256,
 0.45065494859338073,0.32347860074092294,0.7610236860400221,
 0.03623922657655365,0.7042963941949877,0.19789236148049383,
 0.47357481608480134,0.49210920562336646,0.25624903890302897,
 0.6580590290016969,0.47376154084239774,0.39205229754975646,
 0.9995695525072974,0.3315633638100465,0.6397834558829055,0.2242146221813408,
 0.5178839778059385,0.20319886551667243,0.005729430781719946,
 0.4389079210347635,0.7925308934721014,0.16656355976264514,0.7649226057050599,
 0.005195374788274032,0.6351633281944196,0.46941551914997115,
 0.8467960047577953,0.11483474426529694,0.03750054854025242,
 0.34322231329610287,0.04435908831687552,0.034868411994285586,
 0.9204407325982781,0.33128612925784484,0.9413396564569941,
 0.11303392233727394,0.11305372687129611,0.07810796286486932,
 0.7288342069065581,0.9473588741398591,0.9153066783585976,
 0.011092734651285796,0.4864078541464628,0.46738562284026264,
 0.6241347672924723,0.19419278272908524,0.7168927684362641,
 0.06375955665984356,0.020562682686497635,0.15004609499564348,
 0.8239858743916456,0.4160916868783111,0.2999537627338297,0.3291591338021773,
 0.19169184507718018,0.2130872628393704,0.9804846557004593,0.9674578123600466,
 0.06927508203403632,0.37229807383992364,0.6413902807351943,
 0.34242055074021194,0.22823651406616374,0.2235666315622583,
 0.1725451309019006,0.6519388929536973,0.61153398702607,0.19858571313752849,
 0.1950055157780861,0.9401337907698241,0.4570583767711839,0.2505599309014974,
 0.9546258350695265,0.006900428276694348,0.8330412486647767,0.710222514490627,
 0.11500370318504605,0.47143876146722397,0.2940635398732714,
 0.0727204127903931,0.036053275847954036,0.9145434692731237,
 0.8175653862049111,0.3970867196049612,0.8341877129089714,0.36853801315293,
 0.2576083902212958,0.5135410581810123,0.5083714499267032,0.3379209043972319,
 0.48200870867069445,0.6084196814746282,0.8519365579372584,
 0.33465321070594056,0.11025930906945214,0.6280485773760557,
 0.2544589225692163,0.9370099335580406,0.5181161170175804,0.1450768655577127,
 0.7562609390664637,0.3471249615303027,0.18839589077114705,0.3059698586131452,
 0.012651776348586607,0.9642207779014588,0.27557327626757255,
 0.8228381064095381,0.2696104258498724,0.8009006210627074,0.2370146385817976,
 0.6018199350814564,0.19221553298322092,0.9950935351193589,0.5358813093416541,
 0.17007635672232424,0.8604612765023968,0.6664562740056925,0.9413750780657287,
 7.969131269952578e-4,0.5768604498103826,0.006425099720182459,
 0.5353466782740353,0.4199086900369514,0.577978696470089,0.6923425466362121,
 0.7970904214202037,0.8629453276750219,0.33670924088271437,0.2756709114488439,
 0.7143762603174335,0.6180836266270424,0.8857945758068165,0.8345797867595579,
 0.8941145325932333,0.29749680735837114,0.09295962334950048,0.696185526603255,
 0.6753502824024665,0.052693256335099026,0.5384595904733914,
 0.08216007076683818,0.0819572430478065,0.7256558966249215,
 0.14584471027005674,0.7206732109708824,0.008593329317476583,
 0.8470923744147333,0.7916186543079298,0.15042663948565282,
 0.25614798551202855,0.3068231501318589,0.7560442694095851,0.2686199443277202,
 0.5052241651848465,0.6732980862359333,0.006176984603281199,
 0.034440283567688024,0.6083636626781443,0.7670055870405663,
 0.4840738311739168,0.814392291089115,0.04011000920073493,0.5081354962490311,
 0.14777948741567593,0.9267180968514077,0.16821478249797162,
 0.4823694462013668,0.9322821437146548,0.8028158961668734,0.9412476494851286,
 0.2993499448333198,0.5836687650769209,0.10968916807565465,0.4008186709899445,
 0.7430440679785855,0.04115022123671375,0.3401094910999487,0.9524555763571522,
 0.7460368357128162,0.8985448291371128,0.4419909000181599,0.43891442002719416,
 0.09636978589825163,0.8422484207826597,0.2702918811398918,0.6409069216865885,
 0.9370458198010518,0.49518613102302866,0.8303335268096556,
 0.23758491170123575,0.4020025195623942,0.9054095066606456,0.5590948187161909,
 0.6659969984686596,0.6809639114623931,0.6886386145337313,0.3240006678877916,
 0.23484633879555572,0.7613038301166432,0.17173676047834252,
 0.4225426834771575,0.4388625009917724,0.37754220134189953,0.7753094726269354,
 0.2258546222649902,0.6699984973553347,0.8842837005393563,0.7817632521852068,
 0.4295173621464232,0.6754709740757731,0.9390968860514834,0.22839434490459865,
 0.805425888063628,0.64077563995503,0.5248445209528236,0.10945310828586352,
 0.8728174532639558,0.8655967121350642,0.8747315275688603,0.5237245432366254,
 0.33141264860881603,0.7197401791337037,0.051285080086607704,
 0.7520737749172152,0.4905269759114287,0.015403465369134617,
 0.020827087797852695,0.6060457413712,0.4613746194169297,0.18238494808312888,
 0.17482199565820222,0.7105475428333697,0.6678409552361738,0.5937638613231485,
 0.24126747506667823,0.5076484706538137,0.01661264109848104,
 0.6807899686928255,0.10685700752659744,0.2050324313902775,0.575891480390602,
 0.9685104636090293,0.14800276020020342,0.0663316484659866,
 0.29853651609376153,0.7724861982990913,0.8299876030632279,0.2707333087772684,
 0.6376900631937886,0.14382851797251417,0.1264489781460173,0.9002065732513627,
 0.7288187883470109,0.33390596094604796,0.6341513638021332,
 0.19471674965908614,0.07057318547590097,0.5764759074653167,
 0.8262738031341933,0.11140887662828369,0.15584298142843633,
 0.8375552848254666,0.14739583983004012,0.1870361611772935,0.9101324488349897,
 0.9493053037576806,0.5629573841741865,0.6198930244065263,0.4812053948058136,
 0.6413527860037785,0.14795230705108264,0.4085610871554053,0.2966736436757462,
 0.39092999418045005,0.5714557333241822,0.22916504146211103,
 0.6516458250329837,0.603505993749977,0.6358923145298881,0.9258607954468787,
 0.05867451360134668,0.5009356023119605,0.585756291821841,0.707727442089417,
 0.501046522345382,0.6699540439255341,0.662656400964023,0.21663540435648931,
 0.3007406704905493,0.7970252745384805,0.837634949624559,0.7870968571237466,
 0.46830406062696583,0.921788796893521,0.9095226967561559,0.2957901850958441,
 0.8179160384035198,0.8246248447095553,0.36362919995016685,
 0.23925520621434626,0.7843249068876734,0.20727494970861893,
 0.6284949351573988,0.35521738301499883,0.07118926834657746,
 0.7556358580208385,0.049670622519922714,0.3569736015634589,
 0.42454342929939104,0.9512112901869261,0.47149201258054885,
 0.4468683511547695,0.04480956201440789,0.0954632454484311,
 0.27754845283895424,0.3245543137051172,0.06999898428952123,
 0.3200204788571751,0.327232954631,0.42883870874040375,0.11910092758020707,
 0.0580929352725873,0.40433746750290656,0.5150841644088257,0.532669818221539,
 0.8949610071263172,0.7829051793261155,0.04154563704579084,
 0.021990102344932816,0.05647404366530828,0.4822436117589177,
 0.7958017926657036,0.7500133267219086,0.5222035287391451,0.06476093268266303,
 0.4685795033391911,0.9302421843566188,0.731467251078542,0.7361271476022946,
 0.44187848569044585,0.06169583879855245,0.7299400705337217,
 0.02527233795313688,0.5560807565965612,0.40164129052522024,
 0.9237548311449659,0.8946181108827824,0.8666080084231873,0.5772713535355388,
 0.1513264392228073,0.27749361256547345,0.06544408546609248,
 0.14482613600681116,0.9638572632242353,0.6224248821616385,
 0.21074748908329555,0.1844146546625065,0.34784320173284256,
 0.39626720883418587,0.6575507073944473,0.588823813751946,0.38620415665970187,
 0.03278216755792729,0.31740868293431956,0.3323569227849129,
 0.37366932743697334,0.6848825848424872,0.8985332398599309,0.6338807275662329,
 0.4979906222929661,0.1214757532734041,0.45856239133651855,0.4669274229135334,
 0.7174004148237797,0.8442693527787446,0.49853217451139364,0.2941502764655627,
 0.953176510096097,0.7140868419212677,0.6272493667682909,0.5456901062640941,
 0.6000416486920801,0.9969383920661289,0.3918708857423938,0.504011060880062,
 0.9008929869178549,0.13095826151079493,0.11366188767414787,
 0.4343902069396528,0.1800368667930099,0.06250532168125411,0.9847779051722043,
 0.2226191980542055,0.7586387864499201,0.5086754231592601,0.8279805040809481,
 0.13933847365485263,0.06293917405080673,0.8717485353260145,0.709680210707814,
 0.28710880127266825,0.08280375065143186,0.4119437887200117,
 0.20612058201750194,0.6672959406912774,0.7619096303771378,0.7116283395185417,
 0.9624463217395489,0.5619593686814904,0.5453994389324757,0.8261033677558484,
 0.1402471242257426,0.8191911731232385,0.25789247919220315,0.134512114931155,
 0.7643823117878121,0.8898211108411836,0.15449743518957892,
 0.012747115946139531,0.03966131753701563,0.28049660068452964,
 0.27557065880158926,0.5708639796961421,0.022446139334392434,
 0.28630510099240714,0.2234335028510278,0.4339472184491069,0.9192727755542255,
 0.8878304736241146,0.3444157081941287,0.4853810037344811,0.8650823761372299,
 0.1512611771120378,0.6107857481651799,0.4369202072188886,0.6072257346343545,
 0.31086754257860516,0.6483082887468563,0.5495846788878488,0.906349046233043,
 0.8426887719101517,0.09586000467826672,0.5936115665197239,
 0.20926674229854492,0.7501193638391986,0.31731269913581994,
 0.9590143389588098,0.5430088050424665,0.7792757392842851,0.6546388466601707,
 0.3484356145059955,0.21318695755799844,0.24813010634546684,
 0.9797469431149113,0.6154958039862601,0.44880031412058985,
 0.01311033658542049,0.17693991128520614,0.39296677369815336,
 0.7216898119677213,0.6662564696355623,0.6108970707595432,0.17147870614041172,
 0.8660632742401454,0.5622999132476704,0.14265922261686637,0.7885622006248257,
 0.807683724850877,0.2746434314234336,0.76651902752911,0.14282149985784864,
 0.9576063971891188,0.4317669607028056,0.8557676948868929,0.8420097776973673,
 0.14579208573296532,0.9564778303493201,0.46586948990417565,0.999648094262147,
 0.775364944177785,0.9701482870616142,0.21674870708687244,0.5889817843358661,
 0.7049420994605455,0.58174697716281,0.13591378208476923,0.7135897924144771,
 0.024505442595584292,0.610887178894545,0.5365767115556617,
 0.02637157782523114,0.7239570774761549,0.48030777090251275,
 0.9344720004624173,0.8801517097902127,0.6271829174178404,0.06321836068203535,
 0.4130414408003885,0.42626749661216634,0.8727191522563637,0.7412297809835533,
 0.9054116678304984,0.896401333065804,0.6174012961753746,0.6603516843218491,
 0.9133378821743827,0.6937718854058121,0.8755450046107636,
 0.001603116965214535,0.08349801674840984,0.8206243019768391,
 0.989272450172376,0.38618429380884933,0.7981341052301497,0.7773073713794962,
 0.5105551132567914,0.15564863424175046,0.4159596439379871,0.5628877352665679,
 0.8463959874979737,0.6246955346020318,0.3410815697148477,0.7149663233514186,
 0.35075264642466963,0.8259976046109933,0.3836781725342563,
 0.15670609410522984,0.8632448929302139,0.0722093619304166,0.3466233693669283,
 0.7295250799089732,0.7808597855741022,0.5266347484141645,0.8995886439735343,
 0.9986748135483885,0.4794640272775662,0.5003478096792865,0.937931236579316,
 0.7012142214992462,0.6814946125734762,0.31276105917490493,0.9848009334235159,
 0.808077402037395,0.4825837937849179,0.14727424092248298,0.44492043189029284,
 0.2687222341823843,0.7241908264204459,0.45745628248257797,0.5802321023888485,
 0.8133772136688966,0.8503603836202582,0.7141772999682845,0.7117326884712789,
 0.7042566200713212,0.912059604348247,0.18017290630400395,0.5082359038260181,
 0.7834951066267943,0.6786377282507907,0.9360028232534753,0.37499397601143447,
 0.1657310803194656,0.09648768309419786,0.3857706006894278,
 0.25068901120952913,0.3553056048157638,0.4957704287323538,0.7923909232495201,
 0.385204023710092,0.8636978130369718,0.6793549116835302,0.24627132248384798,
 0.5670048798092256,0.8043947720214736,0.793187198753293,0.11101179128185001,
 0.6207661567562395,0.7181246689571034,0.8133212432972846,0.9170526162423422,
 0.19830922394359818,0.9944278057355938,0.7861984344173673,0.5479424121234847,
 0.6653352751631885,0.45569572725252616,0.9175873586424174,0.5778371489636434,
 0.893374521487813,0.4834209005127099,0.5571344134434886,0.59582338499035,
 0.9824745439118389,0.8246794764733872,0.6852549741385718,0.7165540356710272,
 0.21672552227293818,0.48182339566998444,0.5333330444500907,
 0.9426086146506835,0.8042843195661131,0.23043752042893162,0.9756080483291167,
 0.7382061395618776,0.7638366976747314,0.40789699365166654,0.95843289953743,
 0.6449774180387884,0.814639322773843,0.1920847825689852,0.8710916525979204,
 0.46038563523699616,0.06873933314010938,0.6927561604145235,
 0.48318748630755637,0.8174644975033851,0.5181731471511031,0.9148744232275815,
 0.6536284319427745,0.39098360486476347,0.5877146679703047,0.6150446142455533,
 0.6082179020000064,0.3272395348405317,0.4427249647262347,0.43620303386826076,
 0.09040655190805336,0.1488291823591723,0.8394114178025318,0.9194264438121261,
 0.39311391091423487,0.15788372240301984,0.09979193980438783,
 0.5752076272123074,0.26264145081181334,0.8228132712814661,
 0.39039173268269156,0.5912089843523963,0.8757881373570104,0.1082244219775701,
 0.48759835233777216,0.4850861042982505,0.09155630697200567,
 0.6157937796495935,0.7840376137259142,0.08059919175958763,0.8014366564546651,
 0.14417722125939458,0.5361064005848453,0.908731186644851,0.01616007148097498,
 0.295402761764586,0.8752701322559916,0.6836097947068329,0.40679981431932055,
 0.6265529385370437,0.5968186980090256,0.7958768475122883,0.6823079119356745,
 0.41809770222970943,0.27498140898611567,0.20135458095776226,
 0.5867329561395199,0.01200243838887749,0.1525436002735815,0.7819298638634226,
 0.48531332270827443,0.6365127453976336,0.972773577091263,0.7495029437247243,
 0.9195808273731586,0.8273176841962971,0.9663798387600443,0.6728719453264214,
 0.2237870194005892,0.9039794778355507,0.7560655378661472,0.2755206863467924,
 0.14685710360112358,0.9537390835063742,0.2946866736803062,0.6509344422524569,
 0.338889102402828,0.38603555433562364,0.2562822037785196,0.16988077519879452,
 0.5514317303490702,0.5528402619545557,0.7867465892257128,0.7978465804729556,
 0.44458726248091984,0.34974183305537154,0.12495892389128638,
 0.5228423125385071,0.8616168476212344,0.26912095469522734,
 0.48950353452907525,0.516642402586953,0.21473932318540645,0.7467271875333048,
 0.5187629992228696,0.0512698266816588,0.1716582144329557,0.26076680972607413,
 0.6404172361352852,0.288095522084443,0.49054526730032844,0.44544427174976386,
 0.8108622526988761,0.46468467767158905,0.45024869335010986,
 0.9444978291643882,0.784356814760726,0.8316512527723742,0.6486437285196712,
 0.766355614478225,0.9125376268118922,0.5845802499426458,0.18319195139395317,
 0.5686484038560775,0.22141162841961193,0.2349289444031366,
 0.35957059826414306,0.7742095000743061,0.43007432142871505,
 0.8789864575172907,0.15229109839865773,0.21407157703152468,
 0.48728421997561355,0.7846652213878582,0.605266632463631,0.45583922205511374,
 0.8683720652031219,0.9457011778460321,0.014214215453608459,
 0.36735359796825984,0.9728820464130432,0.5465772086438963,0.8198124205633103,
 0.7944448903228946,0.9930240877620432,0.6839749269719824,0.03355653532167957,
 0.3779752629236819,0.23704270524239235,0.368818088059927,0.1578396907272147,
 0.12545144138318431,0.9938727641343571,0.6204288583032774,
 0.44247008399368704,0.15289521373934634,0.25964512798115535,
 0.8889664318190778,0.6791960391362596,0.4334984928404113,0.49121575979165133,
 0.39543490744988874,0.8156225844264804,0.2206945299291021,
 0.43429032051872307,0.40423419127336757,0.21009453026255243,
 0.07979210779908141,0.28927958882613214,0.5169962347180693,
 0.5264065948810641,0.49622634166886015,0.7426375316466998,0.4226347621241683,
 0.2642296321556383,0.8958910323402487,0.6466411145287082,0.15027895186672557,
 0.6611349207463809,0.048487958563350975,0.15871352671739847,
 0.609885092096142,0.8499345379395931,0.21187630121021916,0.8098682223022537,
 0.7549671692001099,0.5704956814182354,0.5597602129699062,0.31886911636659687,
 0.8637200013008819,0.378440083678969,0.3800135935992155,0.30805972587495356,
 0.25377672079286584,0.9712927211524938,0.4682933518447404,0.8311763464332096,
 0.8950239900385955,0.31752026241933884,0.8812622480510004,0.88691122130369,
 0.4759635279209208,0.7449434188686455,0.05520487760035564,0.3619528591194875,
 0.46977835049982597,0.10897202001434558,0.5563368932262946,0.744782356112409,
 0.014974439506872272,0.2458594136376382,0.07677976028981193,
 0.6329151174110742,0.7087427398296413,0.07207127026698323,
 0.24235825097267405,0.25585548174721673,0.43195709074987687,
 0.6171864984361087,0.5882937567534655,0.42492610925917706,0.8905394983372633,
 0.5882983586042306,0.9879297953563957,0.15412426273816793,0.5931295987991693,
 0.7437165919063071,0.4902327706776197,0.7011202176043635,0.6939064786767252,
 0.37910742101388906,0.9901098301318488,0.5773136888359173,0.719154378283984,
 0.4803516397801072,0.2077312839600337,0.008812045392661338,
 0.7951417106131559,0.41691664070813794,0.5688566566970528,0.9237610114285962,
 0.6605427868501864,0.20057702769102104,0.36891422962043774,
 0.007992242435976293,0.7613103271040018,0.8857399558284774,
 0.5287229181644506,0.023660859053704075,0.1582675095908338,
 0.6237680261037876,0.7745372922228433,0.7397275202443526,0.6911829045698012,
 0.06594128892926843,0.27907948541993766,0.7145855670335821,
 0.9382245646480849,0.6027271389940538,0.2504820019085685,0.687126479078437,
 0.61103547573801,0.8017671577304193,0.8491381882845734,0.6152614469538764,
 0.6396692449708197,0.42370054618811237,0.9277988114213092,
 0.010249393863679401,0.6974558097423789,0.13498065225767597,
 0.3404967703936981,0.7374841026623841,0.5265039105356601,0.8375181187385091,
 0.06637010808457555,0.40911458892519537,0.8570564239872279,
 0.41986097438284564,0.0997399161988386,0.07878497120968553,
 0.3407706560593613,0.9912388222598206,0.49536987188459014,0.7058151627832939,
 0.5983848666873339,0.003397863971098869,0.2936307160009348,
 0.9533036325239014,0.8586956656798194,0.9602703958358499,0.4179675766734159,
 0.08687065417941853,0.6177903107317926,0.7946595885295628,0.7697137759391394,
 0.5313416193245031,0.9143449021609802,0.6492888319755672,0.8367956706224883,
 0.2939396674777245,0.07223829944671101,0.44190605507972325,
 0.6026377019795612,0.28931644922742983,0.11579192407980188,
 0.9135168147401852,0.8024406217239413,0.9524194299820287,0.11224746179175271,
 0.9851212413989072,0.7547597000510933,0.9902006577766164,0.655881134861936,
 0.27599925133868775,0.8878373352050152,0.0717757108404451,0.8855896072694394,
 0.5143052028801426,0.3727670948353834,0.2937199785183241,0.23748658466525574,
 0.582503999682856,0.014021833201087963,0.2944251566220547,0.7211878997240466,
 0.8005610743133051,0.32635558600984904,0.8254764803729592,0.4430320189935669,
 0.23146911449019603,0.8849355956105965,0.9521468986775355,
 0.06021620486230106,0.13986431584389358,0.8442895723757288,
 0.49476813869263747,0.8430324132446633,0.9027780288612176,0.4328546187439265,
 0.46547270939984065,0.9962900899925298,0.787054927409091,
 0.011998061681317895,0.4269400594838525,0.8516638865781,0.43541462166763023,
 0.5178132963375052,0.6681313444087557,0.19583249992547436,0.971006516066269,
 0.40873665321819486,0.7827631935188699,0.666003912580268,0.5741501206101429,
 0.9644236852098641,0.08736386419969999,0.00505039570387078,
 0.20850753221346963,0.26302572469549057,0.7008147670730215,
 0.35311640683031187,0.830735304879838,0.968570993079997,0.9765909284255341,
 0.8955316864871196,0.828759488433528,0.9650121482288139,0.41279926151534796,
 0.41450604918525413,0.9143496816085385,0.49313054903987086,0.579064047741728,
 0.39990686394046926,0.08078051357095106,0.4562878446531663,
 0.3508503647898177,0.31247064180374584,0.8579346997309568,0.3490807694772736,
 0.037307693034329015,0.7332976999477823,0.8742185513761517,
 0.4503657334181299,0.8377483089081451,0.20982565672989828,0.3737640992610648,
 0.5619592087698677,0.9616581333159095,0.20162492572371515,
 0.06014344373651315,0.5883039178208128,0.4541652253716262,
 0.050749070183874534,0.9290707415159795,0.2800951357041527,
 0.23878918717346576,0.01581314605378259,0.230010526942369,
 0.30433941489621896,0.17695525501494913,0.9391891504344243,
 0.6757737657737223,0.515920410510037,0.9286204948827503,0.025638375613999065,
 0.2802322876526885,0.7233496159602737,0.48401324529101186,0.3657128708616548,
 0.1505652055872143,0.45432232654905147,0.1929691966050061,0.6143873874767485,
 0.0418085407251263,0.09130808508248434,0.7002063738025549,
 0.14646671971237657,0.013067337090285847,0.8016551528859416,
 0.9636669945749778,0.599190578553741,0.669004366509148,0.792672928585687,
 0.16585243586372966,0.7274433383271447,0.2537368786149824,0.2517858392714656,
 0.25791039789755277,0.7026293355832669,0.6427100862397299,0.7777033468353265,
 0.6307200775435312,0.1314771406203351,0.09877321813690365,0.6780442038819161,
 0.4340972561511055,0.03118189170604202,0.8398809308239455,0.3694948555528915,
 0.04626697367215726,0.4341071341094409,0.2612473525564173,0.9759680823183684,
 0.8421433913041317,0.27970091602126224,0.49373034699452545,
 0.3555282561576587,0.8024688697668667,0.9058292512080858,0.7679478646084302,
 0.1612649401955042,0.49153449049877007,0.13822052221405157,
 0.5402250107103428,0.6234827085126093,0.9821666764202585,0.0694732397309813,
 0.7900669857372783,0.134089860021952,0.7035436336331881,0.6492856394385829,
 0.1723879799286283,0.7760369762363928,0.3462637733674503,0.48262370441111213,
 0.1792582624761302,0.05457565364715289,0.3892106228715456,0.9815671195935245,
 0.4941430977757445,0.24846774305747155,0.5392983789027308,0.8383073636115363,
 0.2882323364526178,0.6728063183821863,0.19804553812261116,0.6611314218167748,
 0.35341369421575086,0.22566149769667476,0.06101506087611264,
 0.5099211335740014,0.4406026931041922,0.8555608666828024,0.6215573943846531,
 0.6702219442893018,0.8262880152152593,0.5657444568365956,0.6421654442840247,
 0.6010676486277456,0.24683696055656756,0.0921982557617631,0.7541932813574108,
 0.421545101480737,0.6810495039293252,0.45492115797118626,0.6549177163927898,
 0.7902102592213701,0.4422776221048734,0.15703542279817506,
 0.47771335404901427,0.039295301013677286,0.725233336593287,
 0.9866752532991352,0.8139777472904233,0.5364275210164715,0.4438139201486694,
 0.34330454205622085,0.808926559862977,0.12455117846348618,0.2153279134671956,
 0.6148736659924209,0.47956623368247886,0.937662583343857,0.36936462069374565,
 0.2597726419592393,0.3575607233028919,0.7139481475829605,0.7845712969913778,
 0.6739210040311939,0.5655878976899185,0.19657230343846277,0.5798666311379461,
 0.056670452965974505,0.737692669162112,0.8367181358015046,
 0.027295406157603064,0.5297341474647681,0.8370268659249211,
 0.7872698186709374,0.2779109227734602,0.6830896596318184,0.6145492190371838,
 0.1103738336528729,0.2909892649475594,0.5541291911991574,0.6772632847691813,
 0.4720472455071847,0.7785134327407282,0.20516972120246435,0.8340881413442764,
 0.5501173468730006,0.006400710491710093,0.1849605450975289,
 0.05708765550008765,0.0695068908236759,0.17570667512756477,
 0.5041710752779857,0.9007761499721499,0.16119676747331013,0.2500041050903059,
 0.417330458834423,0.5942694607442333,0.7877120462314295,0.19188949762497653,
 0.3530517770718853,0.4010136085714465,0.22955874092430306,0.9545722053180818,
 0.4545312354917217,0.469302664975326,0.00759839055497169,0.4817085465047606,
 0.6847324908557127,0.9293734264476287,0.5425142123391739,0.025703226537099,
 0.2934436664211483,0.11393320889893721,0.7275682736760892,0.0925897754208862,
 0.08778882990944348,0.4197642445995238,0.9627634729809481,0.8918050278231234,
 0.18278723109639572,0.4226198199650527,0.5434713937484146,
 0.46679621672686555,0.6558777030845095,0.1093772956452389,
 0.10143829067754451,0.9587799774629154,0.1004667704559723,0.9096894902032839,
 0.6644497971576327,0.7050534308654848,0.6465779767666712,0.20227405437645796,
 0.9958016959616051,0.1230344820391952,0.08510838357148043,0.9463234906442327,
 0.3229330346059581,0.47307715038003795,0.4059245115103579,0.9856004296216818,
 0.5380429988585804,0.5403826211495236,0.013570526074511702,
 0.35452429071320135,0.1784809802235292,0.6596815110427617,0.8530444296687633,
 0.2864596581209412,0.6252996778850956,0.8940304871204077,0.4228451529144458,
 0.04115172054076133,0.3118289454471267,0.6945652985205862,0.8942746322303871,
 0.26182513095118753,0.7629722786656071,0.574469690228966,0.5734703795423651,
 0.6964228525038689,0.6411984163579213,0.17303796433840934,0.6974619822852061,
 0.46923379021256095,0.14762909218498943,0.1455056261658334,
 0.8220913223732207,0.8480955910780466,0.022902100037395567,
 0.13265964246214845,0.6023156379025771,0.30764160016062303,
 0.9967899398817309,0.4365518721070173,0.6099493047483812,0.4383864467671712,
 0.4883208678655012,0.09112933033422754,0.3598645617232683,
 0.23221114129279508,0.3234596152786148,0.28571091075485,0.9369730483591154,
 0.9979989724232705,0.5328803464639871,0.5633500404755998,0.6791305573183549,
 0.11109550492328046,0.9881008511722171,0.7301587976289414,0.5317636455217495,
 0.20947789832751174,0.5181084918208396,0.8017711383133558,0.4707513449243861,
 0.16648844801653573,0.8079688372710515,0.6421757242581088,0.8984208934525089,
 0.7572329789406764,0.6974383021342468,0.5847161200950034,0.9522666010597328,
 0.21258721503087474,0.6800355602081155,0.4734706019328092,0.6207324812439392,
 0.4307668762871826,0.35795369914887454,0.4726165230764581,
 0.07351943257484883,0.5531624609634925,0.07340886845792216,
 0.23620766778763302,0.9143490891255199,0.38889842268254793,
 0.2989203990061612,0.7906766828785113,0.10274017272808056,
 0.11924593022037766,0.9855953202224449,0.701227758169613,0.2896153083751667,
 0.7867009111566813,0.7859844240473115,0.7026420620567306,0.03519538185996984,
 0.21202902814090874,0.08182758085602704,0.18458285823786502,
 0.2844416320975619,0.29484096825699535,0.5658476174545083,0.7730146293364218,
 0.6416249400194058,0.7929088835714668,0.6264798703337732,0.5216146850426961,
 0.026457167115605196,0.44302042474987813,0.3659860788124124,
 0.6871429807216608,0.7405998439692925,0.9828182406866577,0.22397911392831305,
 0.9932339215255465,0.3221439375361339,0.8051302203300235,0.8267011060783864,
 0.03363003083849059,0.5110401858934357,0.6077783756109346,0.5518216319262665,
 0.2561935704317009,0.011515632902301176,0.6607418279479236,0.390120415930274,
 0.5807988957098483,0.20032683116603311,0.8029542236970223,0.802711244934226,
 0.3984485818654876,0.007192745322946448,0.019228425925833514,
 0.4853430920938846,0.12778166402477575,0.74337084828976,0.9686595132026081,
 0.3875149451399831,0.5465204852584483,0.7239473258930091,0.2867479128397381,
 0.18978050657400536,0.40544630387278724,0.48089711103077004,
 0.48570372711079013,0.8465954441445557,0.6044181268814142,0.4165140405811346,
 0.41352985519631946,0.4177467061374507,0.3177107307082656,0.3890396226520262,
 0.39070977415371344,0.5655538869077802,0.44120020623512235,
 0.9111138329571888,0.06304343807734225,0.3750954074573264,0.350901037514849,
 0.5096698499739836,0.6270691564848754,0.7196594348552054,0.6345596273929598,
 0.788179775993296,0.9235984828062924,0.8312186346621536,0.05569503170977452,
 0.7885847927309284,0.004677148844215107,0.5355370958803687,
 0.07730894831185564,0.10587616175129233,0.32697190078775074,
 0.7242819644671619,0.10606374705082322,0.004771469665652006,
 0.23106042602310195,0.48620703187460423,0.6253918016170886,
 0.7082658609715697,0.3099765764630501,0.8017190101493737,0.6423012547591347,
 0.44692078040561767,0.7768895070081799,0.8733580300987154,0.4130055582092801,
 0.8633822227626016,0.288862829876534,0.3558086807034364,0.37165912003324575,
 0.4101743636226983,0.22305725478857397,0.7136295978977983,0.7708670720437643,
 0.9146497122988504,0.7331914846351855,0.92058198601961,0.37344641668006207,
 0.06041824783476535,0.39227677579426434,0.27563156051950366,
 0.03804765091148421,0.6728835300342175,0.07569140209675607,
 0.6146325334014813,0.7377052881280327,0.7557770620172208,0.4376333264851847,
 0.05515526646193791,0.3247599541976536,0.12457669612841482,
 0.6892683256192961,0.8449332358000446,0.5835284688572389,0.20692783927901903,
 0.19306602821504304,0.21965572749699158,0.7112843503712937,
 0.8907551895383334,0.173635792646327,0.8591587901794149,0.11386301572822921,
 0.40609243445877485,0.9393758587683283,0.4209837505327828,0.9205976520272402,
 0.7405201983102279,0.17150431980941694,0.12061704885959246,
 0.7062954984604196,0.6044552062767861,0.4328481337567388,0.8913458076783213,
 0.801614237992115,0.0262907962347666,0.24394883262545097,
 0.004569799783349482,0.8925041072600863,0.2029383078438165,
 0.9401513298223101,0.9982444145701286,0.7635444767370348,0.7641130930504533,
 0.760067814767496,0.13831412273055954,0.14007586381246218,
 0.12226557393394732,0.18449373910424827,0.1854096848761705,0.13790525042378,
 0.721091500131233,0.2994486419857587,0.8050876173931427,0.7278178528826437,
 0.8148564147399636,0.5888185578928986,0.1425710415836542,0.2524803492966775,
 0.7184826874981709,0.4902594262156573,0.16068150382711976,
 0.19144503085823916,0.8984367932414561,0.8795007457431366,
 0.04823006722754075,0.9195476784204079,0.15414488128009496,
 0.09404357766560534,0.3631554119980873,0.5356490253281265,0.1519915992617875,
 0.14408100992590023,0.0321271258963175,0.9710917008538407,0.3884184148794325,
 0.7966111102510383,0.20536071448000948,0.37774213270538537,
 0.9200291522433995,0.35620842137152087,0.7324440648154253,0.1947431851919338,
 0.31573687357411084,0.9222029311415899,0.7609946607769579,0.5583260598844395,
 0.32904278077608995,0.9133622452853809,0.7325354586249552,
 0.23778766192821554,0.9219880484840342,0.26553641892197477,
 0.19951674302868871,0.39612683496464385,0.5698683290801192,
 0.6135051252867758,0.2783639302477888,0.19075659235430587,
 0.49443522557808084,0.8250683439218574,0.09570198772866689,
 0.8338717982486358,0.9773488260894982,0.0791776170259092,0.10031631586735257,
 0.787674827824021,0.38920305902400365,0.5759755746249583,0.3412555358641147,
 0.9057574251303092,0.06735596188868553,0.7887239838206055,
 0.06231648187711436,0.951066830203663,0.07851390192860597,0.6942765838879231,
 0.8033600289285868,0.1745037468068883,0.8951202514517462,0.5524573875574004,
 0.9278690164852678,0.6796140644097531,0.5357979759799596,0.4662865918626322,
 0.4128408413702682,0.28644755445836656,0.3140438890937004,0.9217347099327534,
 0.5478466792373214,0.9771805246660112,0.849267499054678,0.14762287848365807,
 0.35175832099467086,0.08557791391656366,0.940962990164107,0.8259429756083165,
 0.7894100887087963,0.937086567047787,0.24106586044242173,0.4231815229008955,
 0.9971568393150863,0.38333619594983315,0.3013006358453393,0.4917090286878989,
 0.6513681580480615,0.12616413408209493,0.019856288932425525,
 0.9021350923745559,0.2948459677869495,0.01451023002104923,0.2175667422799178,
 0.8785959156350748,0.74744311633507,0.31674428903123963,0.6109088593614329,
 0.03885501593106999,0.46103255538734145,0.039879715744044364,
 0.7581031855622308,0.666321205044612,0.09339028925296244,
 0.007252734953629325,0.15175603556660033,0.9738046221834491,
 0.13996633607277886,0.9907321381390919,0.8996527173957004,0.0911500627466677,
 0.8845635878019266,0.17755224225390986,0.15843164133066023,
 0.6957993861806383,0.24452461541224435,0.8430545698500772,
 0.04517233908785556,0.037510726965752084,0.28255645624991743,
 0.9350866275619123,0.1946182447179784,0.5382049931867443,0.6679186368242039,
 0.09187889999449295,0.6403575466679419,0.7331199245318143,
 0.03442554485978577,0.5461522004868142,0.9149445178253922,
 0.38092869614056113,0.3751540656588498,0.08872200162095734,
 0.1179686775614407,0.0684373071930523,0.14242784294357158,0.8756741222971489,
 0.13468778771327916,0.9974904892953004,0.5050995266196224,
 0.01855168108874017,0.7382198358741202,0.4241780802722207,0.936240848613729,
 0.21464969481862328,0.4858396094491615,0.1886647961599096,0.5369750852150602,
 0.8141754993460156,0.9345771181860201,0.37745520839125857,
 0.0023487321931013883,0.6291735122177242,0.3231844122844033,
 0.4557086892515132,0.6938162566394176,0.47426889416323237,0.8132689277106464,
 0.2774976548744942,0.18718779326887625,0.9872226446337411,0.6671016193132856,
 0.4725608192917682,0.8871933733687059,0.4991303966207652,0.5815053114314652,
 0.4961586335643091,0.5170093801644048,0.0651746005143673,0.5111844246354459,
 0.8364596621288484,0.9183746661541282,0.40436156256887434,0.56363690945716,
 0.2481207399373343,0.8854959834096611,0.5865257231885871,0.3450116997478304,
 0.7220444163618199,0.5403596911077717,0.7564131031353564,0.4726579770865088,
 0.12258127142384423,0.2032152457262646,0.973636881831536,0.6960275778963761,
 0.9730944075069763,0.3719694625708445,0.2525397120214985,0.8168495952723548,
 0.3001410823814493,0.6948981783290833,0.43420156686996914,
 0.44730691016098645,0.23444401292066008,0.029685117247995052,
 0.7575131688996877,0.8360656150084075,0.8355027129209767,0.4582419482364899,
 0.5668171057064275,0.9161111796276041,0.27563489197679736,
 0.0029864917103040733,0.5531900894169618,0.7030484103500473,
 0.0797472799332094,0.37912472905651995,0.028956268844637534,
 0.12454133836845516,0.8023757606279205,0.29395235455401403,
 0.5668435707007166,0.7872576053490232,0.8692724395436633,0.8556718697211548,
 0.15652861284364816,0.03483948188979347,0.020520284615588302,
 0.9810795986567481,0.09759487171208081,0.7149771920852478,
 0.16316557897083717,0.4976911323437224,0.7385271103653008,0.3435617738579022,
 0.36866562021497074,0.5304931994770017,0.6901851229573023,
 0.09743974429433644,0.5011741790057658,0.7852755714473373,0.8349974077686608,
 0.9120175346140933,0.14304367662891537,0.34032478492834173,
 0.7116905979870617,0.7746398022374659,0.7881283279181939,0.21701303459095045,
 0.8024591221088075,0.24950158321795057,0.6040707982801906,0.4682582949220522,
 0.2833341969454586,0.9044743424833382,0.6263471997212166,0.6568897984844692,
 0.20675544308616645,0.6616666914969318,0.26464202746890353,
 0.6110166553423687,0.5670851659775137,0.10163853622344476,0.8146170200179711,
 0.7225891842803227,0.24044092459192556,0.3266860404866052,0.6824672887680897,
 0.7266900751840142,0.7456316423572837,0.13200341953399164,
 0.28375574248175084,0.8181219193441349,0.437909096886527,0.8729245243767327,
 0.9347732565979451,0.8533593547878995,0.9094140223088595,0.2586236773183126,
 0.8100739788148448,0.7698601623607602,0.5969970022201828,0.4100926469630096,
 0.9409100794180305,0.8156128866136929,0.870260730066147,0.5835162486339452,
 0.899808429984502,0.5717227150527785,0.35339475866170134,0.06891131623102587,
 0.25417545360176763,0.9302752954881807,0.817290791777483,0.19694407897231203,
 0.8836401769993447,0.5930602409377949,0.9746126685674873,0.7566837414596245,
 0.3966514974088249,0.41116112866828936,0.06739403577480396,
 0.8557015798675685,0.5977965057072518,0.806890228210532,0.3271229849772429,
 0.6797473121989046,0.24125296189836487,0.5817892075053271,0.5093577483257697,
 0.8113200792438202,0.7127668194200977,0.8318001286376348,0.20702790248007585,
 0.6267133192448269,0.6680698368840947,0.8301917460308063,0.4010147508274633,
 0.1756553513062521,0.6586933796228842,0.08972104745412979,
 0.07182674535462363,0.03139415474249441,0.2784304181930459,
 0.22243928245665046,0.12033133634312843,0.948086517740567,0.5329820031092476,
 0.5110964945483747,0.07352714538877314,0.1918868185521333,0.9884778014949415,
 0.24440842051670264,0.9110684513188335,0.5345424426581713,0.8470879496588662,
 0.26000711733873705,0.020772396745806554,0.8954222662480669,
 0.07335078796160421,0.7693988551141838,0.3860298934424945,0.722757489507311,
 0.8409002123722776,0.9060676349592994,0.18112615391579856,0.3528873720175978,
 0.6535921418557548,0.2653941661477529,0.7031304403502766,0.9598960468607592,
 0.8354820343531411,0.27217846212466823,0.10551233223317402,
 0.8780543641336004,0.29484964386042256,0.9484650820236973,0.9531056038614409,
 0.4428650988755668,0.013765020984370047,0.8793087895910181,
 0.16817505417569756,0.8736384159352114,0.9933349736398354,0.9851731639049744,
 0.1785973604065756,0.9817509866176157,0.4621818767868796,0.2171899955030917,
 0.06971898373729457,0.014240342293841302,0.4101656985779969,
 0.7918017004815594,0.3379920126566959,0.1276010099856426,0.2860438385454984,
 0.2812266522211605,0.6793294129092281,0.6657093216116206,0.6823796332196338,
 0.7421994505286915,0.6984705531353306,0.24361474151828066,0.6297392876956097,
 0.5843078573941448,0.4294093046047248,0.4062626352525325,0.8681691684695682,
 0.98638481806076,0.9012779435322251,0.8905449528430145,0.29208025761439205,
 0.6496845376328035,0.7894056616014159,0.9001274350996034,0.19380217423707824,
 0.6879306998977591,0.8968190298551957,0.42437777224752027,0.7239890144136766,
 0.05214706948414605,0.15452632056700233,0.41974760462290783,
 0.7941483441091386,0.23754691615235324,0.1716166752268753,0.727468443037161,
 0.8904103437198885,0.5738255818429606,0.24492409187767528,0.5927548183833867,
 0.5131974390655022,0.5957417942963428,0.3570898603264534,0.7583442832072773,
 0.5117370613276324,0.8115389036396515,0.6146642706206039,0.8172080129653966,
 0.4928580768334335,0.6522525176414725,0.5671370698961611,0.10917090445889321,
 0.808847639043144,0.6311182684861075,0.9576622384459296,0.5711032791171277,
 0.9677146790358173,0.34283806368327685,0.9411131008733641,
 0.49669937338579795,0.9709596778376838,0.5239740247105013,0.3237754544860467,
 0.4565955833020944,0.46159387932762197,0.46350734490130574,
 0.8011859794351996,0.1568061811010386,0.2979918126881954,0.20608677618932691,
 0.6548419303588119,0.29065855547617747,0.6166136080690373,
 0.35219022562134505,0.4230813872037408,0.054010677282640085,
 0.8403829819730297,0.7624411189392492,0.7159191283484209,0.6133809504637519,
 0.26743020621719527,0.9711030002213642,0.6317378841896703,0.4844143089488825,
 0.5504658199380142,0.961726984022752,0.8655036928733555,0.2214634811122036,
 0.4523430083651915,0.9898718025443243,0.22086214133622062,
 0.36679295188907457,0.538288035182677,0.9755231951727605,0.49327191152652183,
 0.48269586966350053,0.4991595824331132,0.9077915485780409,0.8746437495411942,
 0.8559432747882572,0.34700791137127773,0.4494295320259344,0.634915195687774,
 0.4654219426957269,0.6589541215924608,0.5850490766095464,0.5425787441807534,
 0.17410017710017556,0.43060167978225194,0.8079036496218805,
 0.8879997925323688,0.6328418957276326,0.9119838203265583,0.8422334584567747,
 0.3027493418409861,0.9552213792411213,0.3551494608758441,
 0.007738468614844729,0.03473042353168854,0.9270051251935582,
 0.4607319128138946,0.30828564646661605,0.18694386151923192,
 0.8554082007425141,0.35726051014113225,0.1767184782791158,0.5964522081495462,
 0.5449194834274825,0.8573206686330835,0.5507304858755635,0.2384243523517846,
 0.6214597135488209,0.8321889567974554,0.8210649255802364,0.4071958510190197,
 0.5001297981028809,0.7785167128317131,0.07483436173387892,
 0.014205926249826528,0.2855851036313872,0.061128868092164224,
 0.48489381561331224,0.8567100558058253,0.604287417211093,0.5306588125203751,
 0.3132299079489709,0.9660989657401946,0.47261816247814326,
 0.10041366485621661,0.0824439116913196,0.6970977614332708,0.6888808500614519,
 0.8905844903317877,0.26222719621593527,0.10469317985824345,
 0.7670879633156547,0.8513895591529782,0.9933407032723138,0.4109974814038324,
 0.053220635103025216,0.14839993215535008,0.9894913172509242,
 0.32565144417368763,0.7284238636718761,0.5002754284548324,
 0.13431154964013658,0.9512978707437373,0.77526770533168,0.1684388655662954,
 0.19557205784957432,0.5098771658146385,0.14169402000746512,
 0.41033478130440093,0.12436967056766879,0.6344023178521754,
 0.5998357293237608,0.8926774378830938,0.47872518985923507,
 0.059905072140796545,0.4386422368965097,0.2773037034991155,
 0.026728728510163946,0.027478765510904002,0.8532832947867165,
 0.2072089047726241,0.10969763282630551,0.5933416977672473,
 0.13073191148160113,0.8526377716721614,0.8213932203340372,0.5209113939326027,
 0.9378320236591887,0.7804690076183743,0.7133291446756644,0.8125113818965098,
 0.8172127917408567,0.9469258514516615,0.056402860903013385,
 0.5372756915926726,0.18455211872339217,0.8821002136005858,0.9099365583653871,
 0.554952331987597,0.4696042062845479,0.8240683731013162,0.8963094447724266,
 0.37880237076195333,0.06841285412968456,0.8835267296396683,
 0.029597112321140484,0.7799834449415968,0.9199628629535628,
 0.3237859203915139,0.9304048878017338,0.2799394139883059,0.4131822401709573,
 0.5775956230881701,0.05289513871953744,0.35148967994712776,
 0.48912022105901365,0.33962791041278906,0.007443919848773017,
 0.43856514836602467,0.5371062535838768,0.7368232673535171,
 0.30325978676210796,0.37590681903323586,0.2924378384981944,
 0.6348002354213302,0.5558898511925993,0.2251457740215439,0.06580473104622298,
 0.5150156579139448,0.004045943671456831,0.6623712657097898,
 0.38810601375903186,0.9307346778129205,0.7992450432554966,0.1985003595172996,
 0.6872395968402172,0.6369832347842083,0.22448690425383255,0.8114398502839785,
 0.13514082266178506,0.6891860866357497,0.22937370047194194,
 0.47585618529300433,0.8607121099889057,0.027625812401321115,
 0.6756509392879119,0.7536904187546934,0.8801501179788845,0.1842668333339289,
 0.31145135864349605,0.4962707730178044,0.3560365597640335,
 0.24622288287306904,0.3481575385919995,0.7477263670373795,
 0.009520028027509841,0.14251850653664588,0.7541500380507266,
 0.9540323779662241,0.24780460441391128,0.260772888029439,0.9810962131898947,
 0.6730523537126946,0.8764200049497042,0.5594443641886567,0.8962719482757809,
 0.9322476448353969,0.9389077644104561,0.5968433056891012,0.5611765722571874,
 0.5704225646119502,0.5595793327551798,0.8363577555504803,0.8098315808270616,
 0.6776590963888693,0.019262914901773676,0.6595942351641639,
 0.06526569987612874,0.5173970000452486,0.21974767424944552,
 0.34222212772828786,0.5354736390098385,0.23492406921501408,
 0.9980137087204369,0.40955467677870905,0.9805958938745851,0.6263291065110133,
 0.3899464452808905,0.44042937813575533,0.46510032246273114,
 0.5245917835901385,0.2778473142708692,0.8753760266431556,0.5333493838653042,
 0.8526156137957281,0.18819924474930516,0.1880099783716349,
 0.014900740903873388,0.9290207057088093,0.5815357944302317,
 0.9208931983268759,0.396340726415887,0.3704695394747006,0.5601070277738087,
 0.09564640721042339,0.05617446703451301,0.22543124414694138,
 0.984014374873655,0.5727500986594899,0.31621417503199845,0.5268014809593845,
 0.4039494972529044,0.08762964390718508,0.5533211196738849,0.2172170715675239,
 0.23503542387350307,0.3265821504363223,0.7375588868663532,0.8593636352304442,
 0.20552314999073396,0.3333579284308612,0.8062467868009081,0.9088268685840452,
 0.6845321636635148,0.2284408777399597,0.16757742731290304,0.0757447412706912,
 0.947438845499104,0.1226767535556309,0.9945711819759592,0.45145769364126687,
 0.617563321486528,0.79590665222671,0.9321474241208227,0.6159476223421453,
 0.31196368482898096,0.06482866315576186,0.007950378139045866,
 0.8771140420157976,0.5246776980297524,0.9064244613808284,0.3143490896509621,
 0.6099984585813344,0.36678273240249704,0.6300841911354026,0.757159977648304,
 0.6464297549257375,0.707612441532779,0.23705700336945934,0.49500277271111714,
 0.5109598458671664,0.7602353534084222,0.15409999104940642,0.5002039767507716,
 0.9532663341780021,0.43100175282734654,0.4290150075653809,0.8060888484140897,
 0.38819802896972133,0.21955836831747644,0.11295885386927706,
 0.7674414158730309,0.17443742642455284,0.8770585781904018,0.9039763561707636,
 0.8942731238852923,0.43848507248672064,0.30782538514419844,
 0.5680956664963859,0.5318055613679776,0.42949542752029823,0.9873044584554314,
 0.3900899719919376,0.013230835110575545,0.5779940671644946,
 0.8641620946181369,0.9768575468267728,0.10307664941823846,0.8614533121062969,
 0.9075873296724732,0.9218326119050618,0.13905659599707698,0.9287135083646757,
 0.12052791730240986,0.38893963762229733,0.9599745688026943,
 0.3452683102563644,0.7904992991733294,0.6931327133301932,0.16099470032900653,
 0.5516242638633313,0.008497434123753478,0.4346956758214343,
 0.6510392073786262,0.3230799657865274,0.6982636786920409,0.37580177199007014,
 0.46069303394553907,0.8771641408951872,0.8327109404757883,
 0.27917662650153163,0.6892278859667431,0.9720560348324692,0.8019469650906731,
 0.15767168310478685,0.1740615348684884,0.11801043905146802,
 0.9057279726796509,0.8528392946878388,0.7838017209003962,0.553710291436807,
 0.5849279591813261,0.6720991116870847,0.845215957483154,0.9524852485068603,
 0.44178731239498326,0.5563695666092966,0.9865710756597107,0.4959434548661198,
 0.1544615813253014,0.5350314357280492,0.5969093840526819,0.5677490747302839,
 0.9176466829957244,0.14444789731810292,0.3951372100993329,0.4746329599993183,
 0.9446938936497125,0.10615346090663347,0.5280532800157134,0.5809242873216284,
 0.32713715318708414,0.10465506667188307,0.6343351219516619,0.927675483864325,
 0.1019725978564785,0.22232936908311407,0.7266394412023409,
 0.14071109560468942,0.5561163342289493,0.2652977973058428,0.4506439752997402,
 0.193080514348418,0.491580028880223,0.9855749316602335,0.5631333231937373,
 0.9121725110752844,0.6434285646052226,0.17963564438596358,0.939427226318957,
 0.7915578280364617,0.57518303920209,0.6325448756907943,0.9165548204190074,
 0.06656743078758598,0.3279416542247059,0.7676225599838333,0.9862969855113518,
 0.7704597031849194,0.888192325200732,0.6282717144194755,0.6324825267242092,
 0.29726150352287495,0.4186735169247867,0.2080517272889878,
 0.32784157061171926,0.3057905918937118,0.5223656776405435,
 0.014533004165185881,0.985000368884223,0.5591910260068446,0.7813488558479437,
 0.6021222286832695,0.6794797425518486,0.834500038066277,0.10239991087202449,
 0.08852066867431463,0.3188381434417139,0.4892751745382443,0.9490713882848834,
 0.23940794650780106,0.8927737202686822,0.23654445444715266,
 0.9358732409582522,0.37203102197519533,0.7578443723922453,
 0.028085678858149388,0.5556167126125205,0.3099682090270349,
 0.23204134663601206,0.6267786774644386,0.3740887009688444,0.6733098273043431,
 0.7253778265235625,0.3619284453308358,0.13370162482600678,0.3343890191636707,
 0.38556056565986396,0.7854713428225456,0.3908142209173687,0.7482318340586968,
 0.11057700139681681,0.8757227861627968,0.272825522176394,0.1531194185409389,
 0.40741144675376595,0.11224733518722207,0.9981758763053821,
 0.33756657590893835,0.5417915517501412,0.9631886844613539,0.8518623011839175,
 0.4790678914650687,0.2911041242093406,0.9139691893008853,
 0.009007381014471405,0.8379497034481953,0.9389914862374719,
 0.4713848523542068,0.803154192622147,0.9840669316385132,0.21234904035055902,
 0.013877400267365947,0.34866162054729344,0.5482275099100137,
 0.12679038874756698,0.9770361760876396,0.4166328617717233,
 0.18253427885952545,0.27785830413000046,0.5832490109610504,0.765558056418779,
 0.1395522591097087,0.12952459352432788,0.5165011745892507,0.6765859495423909,
 0.8948751392084,0.025042333340446477,0.5710684397948264,0.4121211670971826,
 0.7133329320945353,0.16040424843778056,0.06883255973482616,
 0.8373145542758847,0.4973402498056273,0.486304122040644,0.031315503067515404,
 0.5669707683091314,0.6685687334982522,0.46649778533771236,0.7963823886211343,
 0.763405090323095,0.9097837654500396,0.3525743382820178,0.4653560527720948,
 0.8556777564970944,0.3425837726971511,0.378823708523783,0.4595262002671415,
 0.7515663219199573,0.20506826773108766,0.519475612831333,0.63794441580416,
 0.26949992928814237,0.020997644999826592,0.032069169900850514,
 0.5852928682219942,0.6053051237888019,0.46846293396292227,0.6746543540655231,
 0.9085861987359043,0.2541719855355562,0.1469697521180966,0.7326835532694009,
 0.023952132457000896,0.8313850210331245,0.2448449442564703,
 0.35357734496429327,0.6557396482818385,0.4597425169715075,
 0.38891517912983575,0.8932639519904791,0.41231613722671456,
 0.2505319817280094,0.4201675715889822,0.8121054343257051,0.3991390375877176,
 0.9038666382792903,0.09891397816726466,0.09473242442166319,
 0.1821518026459198,0.17636506081557135,0.24833082102225879,
 0.38511369550189434,0.8705666832767717,0.356823556250605,0.5998606225700116,
 0.49027721203327734,0.4714239336922249,0.6901583561752292,
 0.22228289178933724,0.7662526535291921,0.470425321373935,0.9380145804699187,
 0.8052724441150216,0.5897259953314993,0.1717615588754392,0.5890482331725501,
 0.5270413408312065,0.3728352538883781,0.9287032110637686,0.5937522912109572,
 0.05674620838412214,0.4484312232557006,0.21353908804916832,
 0.7474271473684861,0.6982106188081115,0.7586291279940877,0.7179304875055249,
 0.8682951832497885,0.37452061538910075,0.7646148801016298,0.911798731918566,
 0.2058262426252907,0.6448211933891158,0.7889552448575907,0.5375854166836456,
 0.5003006862528678,0.6188803707336736,0.5525146077539891,0.27035708916686596,
 0.17480812442713844,0.4225808282140129,0.6972543517086506,0.5092969086919874,
 0.632760799098809,0.21458313818518537,0.718169829814843,0.12705843005550577,
 0.5169670824170429,0.27276884374922217,0.7648593889608449,0.7446991369049794,
 0.7021866067552989,0.5146181488733432,0.6215680546800837,0.7237136352609039,
 0.6650382308348599,0.9870742940521247,0.9640960048590775,0.6481504398467997,
 0.23491768631301535,0.8170075721543497,0.5051874069227623,
 0.31868360972472587,0.4412400767222321,0.5644749116015181,
 0.18752828473086125,0.12771233177542496,0.40819949067642125,
 0.6655641345622254,0.5507778643086847,0.7258217689980997,0.9246712392276342,
 0.821102932307587,0.962688399007742,0.16014148294495612,0.9544417555484196,
 0.3933496598206332,0.17924336114788308,0.7825070012360927,
 0.40061411193423857,0.7729231698316119,0.19719340969167654,
 0.4786943976917477,0.38942557487952334,0.3442566516857305,0.2815678798372703,
 0.3414454679585074,0.7048439044623138,0.7468981446114544,0.8462586353816087,
 0.9484686939854579,0.4450747760362007,0.9391529403799487,0.4297300959257825,
 0.999270536621496,0.9514409820733013,0.694903184118892,0.5158235740083272,
 0.04135650337101171,0.3799470386780691,0.06202546070137438,
 0.5214323852997891,0.754407421887566,0.9716694855354768,0.13852289105113635,
 0.024481709461282142,0.050099151263285235,0.7198425031592918,
 0.8763233016469175,0.8031286122551448,0.6048551233924435,0.3128079714094918,
 0.7596497285570151,0.7061370463490014,0.9535778145545231,0.5878602240787918,
 0.9356267399678804,0.20667144090445322,0.7364796072175154,0.2971711581598758,
 0.2447850901372467,0.18190689692680662,0.16437596993830572,
 0.21100472345700405,0.6010823576446683,0.8784489687540767,0.9739872230627172,
 0.5127477394395171,0.9688344145090662,0.8105423555774821,0.8893440357378763,
 0.2684718303665853,0.6986152318849446,0.5324763157193853,0.38349882878985464,
 0.807889176821697,0.4753619351880437,0.20656888096575665,0.7740336648697437,
 0.26538746893151044,0.7886731302601128,0.05232365499831215,
 0.05476817408108037,0.23307066839845614,0.5301307036499553,
 0.7396239584643887,0.22470670782215685,0.6626541958954055,
 0.33495380631981364,0.3522032234264417,0.9735476342439883,0.5036248975847811,
 0.3541528891085951,0.9808335170963809,0.04452225976626378,0.7339692565949247,
 0.34556031161190615,0.168172278072229,0.9252743107455826,0.7968278524191311,
 0.1485672177438102,0.5159018303907655,0.1071707578057991,0.644181411973286,
 0.4000430045743205,0.4892211841527546,0.42911414613631105,0.8494754446030107,
 0.7900306612256541,0.625285862300861,0.09679100088899661,0.2394779839449681,
 0.3021188653610313,0.35501743115048845,0.6879381753425273,
 0.10479931801317888,0.6630447248298397,0.4310188734498893,
 0.48783752148503845,0.25562459611719845,0.4960529715038441,
 0.7483984779670971,0.1939375843553478,0.3860030505930029,0.5523480559034668,
 0.26278004998609816,0.9259960251283141,0.6462175136274732,
 0.43882203200050696,0.7603319622767817,0.5970471984907736,
 0.006817032948551738,0.8331268784680846,0.7717772833383324,
 0.6279346609049177,0.7244712033270192,0.8368613612367968,0.4717318871209045,
 0.7050729733774853,0.5169071263037057,0.576325802059855,0.6349137715324628,
 0.3710381665431174,0.4202488068899519,0.7559773955908703,0.13498184766607324,
 0.5301945400648262,0.18074442998400742,0.1287971454438891,0.9165426590966268,
 0.5305667028623015,0.9145734373933629,0.4011657431544776,0.677687357083823,
 0.5740654682771676,0.6673161507657385,0.42704954387820004,0.8872949217051012,
 0.9031755992149072,0.059171815858459764,0.8736955751250015,
 0.38843121370733313,0.4450581840377468,0.4959198227470758,0.5592896613729295,
 0.43865510705399646,0.8538283904953368,0.0023716897530846603,
 0.5485520579876998,0.9505306414223513,0.9379258374907027,0.17750081388890582,
 0.9538573893024936,0.18867624152608864,0.07067135904129895,0.807009661631131,
 0.15107121023797987,0.36083556607660205,0.5347134334314655,
 0.9244110723611909,0.05881879229904008,0.36806766745912745,
 0.5131692783383077,0.8842790777252507,0.20531870312157618,0.550037235070926,
 0.32720395975316163,0.6126912847442827,0.09824392760915868,
 0.7169491690087588,0.12607376015993654,0.5010435510179865,0.8378707438998915,
 0.7974355016073557,0.05069722016446632,0.9329528758730754,
 0.20737412241012787,0.5787250023435417,0.7664057829265271,0.9334045659047994,
 0.8180010835589031,0.5650203771205093,0.03797694295253273,
 0.30336740093235903,0.13246527005965936,0.46881761844053116,
 0.06179867385601023,0.5800922971203807,0.2530969452880344,0.4940632329510948,
 0.03554241481763909,0.24471360036221113,0.47098685439490373,
 0.3713846140318444,0.2979482820921351,0.40943372232581143,
 0.009730698891124545,0.62825421200112,0.4956799411601259,0.07089309222942741,
 0.8933816548529588,0.7489198787585157,0.9160192723328088,0.15832906896241572,
 0.9464503212386974,0.6104104023400045,0.7879219809557402,0.5591117632255005,
 0.3258599110505632,0.23581393481655533,0.019591816012177077,
 0.16306669328766066,0.794731160384687,0.26970699907645435,
 0.23784918441043978,0.5050103291009016,0.21852474580960513,
 0.6221084206654859,0.5804895141809294,0.7046493526383745,0.3304880222769466,
 0.9120696805768547,0.6597793877309134,0.7629830092862123,0.706640955438417,
 0.9031143012203018,0.38363877183980566,0.29003885924731243,
 0.5576035850211422,0.6531991721695026,0.3211900619490724,0.4483948324678493,
 0.7757070405254342,0.42430897554889235,0.46553786397598695,
 0.6989101654367695,0.8381996673667451,0.3628066189744056,0.32143137914941844,
 0.43323248337288345,0.08762368110789276,0.7882049906053575,0.670841579485259,
 0.25316238100144794,0.6918372876413778,0.16526382260058137,
 0.29784939862874016,0.11526605363354159,0.46858450574563004,
 0.2122166306133254,0.6329725623926014,0.666882879893548,0.19032037921006761,
 0.49253585351366014,0.5286668008176185,0.1713524769461846,0.7364826815519886,
 0.20252692056363597,0.11300777290200825,0.9811942796976093,0.811363297891626,
 0.01206385706979729,0.1387732333140289,0.794667714859244,0.8376002759966674,
 0.39806219880283167,0.1630128774206009,0.4446067097026476,
 0.28997286631363595,0.264717765838135,0.7712135009873291,0.8220018969658688,
 0.5644729987756336,0.9471491204586103,0.6069754961200947,0.2397155828016475,
 0.6285616247526806,0.5238610583560178,0.74445586686943,0.8446803294632976,
 0.9538622489819989,0.3193545431816479,0.9796724126453801,0.4026906852901071,
 0.7263017015287273,0.6393833741432381,0.4653907010983541,0.37023357356854103,
 0.7737042805157038,0.32442057219365006,0.7697278711819817,
 0.15931943769921364,0.8706554634301944,0.11631956747921468,
 0.4643661286618216,0.9344525651366823,0.2038897865474072,0.03928645260140873,
 0.7698290718744912,0.7762365573807622,0.4071549635357271,0.797844023732202,
 0.273779508560969,0.08790856800000246,0.9349519521678322,0.43411256814654997,
 0.42041676572918396,0.09556428713008414,0.4507667356601017,
 0.3208704127298042,0.1173980096063344,0.3884931106305435,0.7067969241844827,
 0.138457158244761,0.26371100268735104,0.8537804052404999,0.3278210429470545,
 0.7383814621480079,0.046215703213334236,0.930448657395579,0.540542366216779,
 0.9185789560977702,0.8211600585521169,0.024556483492910286,
 0.03754412943080043,0.8937941060399364,0.1651677074822303,
 0.27318548672014475,0.5466870672113392,0.5007422794937152,0.7495515334169451,
 0.47919312976400974,0.12112217979864459,0.36697657173984777,
 0.5333402138930106,0.8812535701380656,0.8398707293274787,0.673746281871455,
 0.07860965203962555,0.22203762256329418,0.007869212662470537,
 0.4436190177439827,0.9850905125307695,0.08851455886563708,0.3354374308743088,
 0.5694851661619115,0.8253288098660516,0.12763680662014676,
 0.04370191083654662,0.19649862508705096,0.804589565017682,
 0.22920933441621183,0.07620205835924798,0.886652569630174,
 0.15650894315239505,0.570008754429897,0.8341557665131348,0.25247852325671305,
 0.9868157996282826,0.3939985830363756,0.6428906243984928,0.5599124299010534,
 0.35968555844769845,0.14167793553514418,0.35259255492327213,
 0.3184496950894905,0.6470934398005702,0.12139129704127682,0.4982130492952974,
 0.0974683962117473,0.4513701117634019,0.4451778353027531,0.6582736138794656,
 0.08642353639280431,0.7690367799639632,0.01631816208051584,
 0.9302762308054876,0.35537310485375984,0.1380628345243029,0.7784346229120435,
 0.16244100199249178,0.11239391271133037,0.20824994412226172,
 0.29766906344810784,0.31340165958819544,0.9554218658100323,
 0.3855897303953897,0.2187765125640091,0.626809413101885,0.9061063944231037,
 0.3082866886878862,0.6317016219324444,0.010230526508691273,
 0.9355364048782112,0.6157202072402121,0.6630625330175784,0.7540825317517574,
 0.7900459525110661,0.9128931420223062,0.4669373258240025,0.7179273401856023,
 0.7887017140831997,0.1409922867039175,0.892494795366125,0.7898501223309259,
 0.5619331285561265,0.7664828334296033,0.6285587569448396,
 0.012817229477231518,0.4501330404474905,0.22157509986351798,
 0.0166365784240311,0.367381583225474,0.646782339757763,0.07029933374444997,
 0.7794269959830105,0.5482389758522339,0.9009737149551117,0.7792657326055071,
 0.5126627009285971,0.7597163881818634,0.9913551065135203,0.751083710087929,
 0.38207084214867226,0.1921128510480682,0.81586401713227,0.1585962261225059,
 0.37882850592026207,0.5088235393782501,0.21317829336198046,
 0.40883960566270394,0.7984813964287617,0.8143022855180304,0.8156771915856874,
 0.35149498858948725,0.2877693851842178,0.4212930500294919,0.5808476040887125,
 0.0998298200365596,0.5979172580853953,0.538123111487247,0.7700974175997652,
 0.05049738265361314,0.21840807744592627,0.19774752073633617,
 0.7785189195877242,0.6251524283877242,0.5777852500732357,0.22909004770521812,
 0.04371160390934126,0.6761589670412737,0.910978903984265,0.811573842747953,
 0.2442832361995796,0.015200035212330376,0.804481792476454,0.9622621392542475,
 0.4216136344795761,0.5458447064508136,0.02781223329355642,0.7055255383830605,
 0.32247571808636544,0.8403092435842625,0.9981883916182357,0.6721868305593134,
 0.30321459728828726,0.28475311895679967,0.6755977668903337,
 0.7319697623827258,0.22997012002822603,0.7221458813887889,0.9503868977477669,
 0.9765083242493189,0.4973412827475032,0.8326901771371805,0.49769675647547607,
 0.3904354229795296,0.3706780400197597,0.3012261779783253,0.4260688497741404,
 0.9383343774512276,0.29304838718536874,0.8497309602129891,
 0.18313711924374254,0.5095970164131879,0.47623876636863127,0.980376823556144,
 0.9012811361603953,0.7741412836573052,0.8484099824917251,0.7166804894839945,
 0.028707883380958976,0.975121852072496,0.2609477801518174,0.5059978758064502,
 0.8144762039018834,0.6217986630835567,0.6273884935293131,0.7440523384716214,
 0.8887142506242622,0.9345096511423486,0.2438237428629969,0.8739651043188619,
 0.9265226069378448,0.6600384787648335,0.5705209323814381,
 0.013310196206256397,0.6258080271509016,0.21642159834440033,
 0.21907567712508536,0.6936598205461519,0.8820725831446734,0.9843972844722496,
 0.6066970764109559,0.11650405681414355,0.231767465093341,0.7613040930578294,
 0.13476370803559212,0.9689464336313555,0.08156594540308848,
 0.9575183355961894,0.43407375570670226,0.8858358696674036,0.979597543507191,
 0.04956422318657028,0.9392068119787023,0.5792044715269319,
 0.13853211475415994,0.7786151791695284,0.5137069320775022,0.6127175841874452,
 0.5863487678010151,0.3966425710309329,0.010879485035952707,
 0.7597532104803559,0.296270295009462,0.33413874902883245,0.20556764934211746,
 0.6900804754166688,0.4222117759389523,0.831674218908194,0.39690313057602244,
 0.6723519554134423,0.18706903850876644,0.6877021379664331,0.9145120866700361,
 0.0803140628939445,0.7821135445321018,0.9052655913537366,0.06685906722822532,
 0.20456591186677442,0.06150563531711706,0.16794292833130098,
 0.8736378355645184,0.9022484092318779,0.06266698462526499,
 0.49527325160665514,0.39764835089401884,0.497146370506648,
 0.23048825648838211,0.7788789012188077,0.33567702636131114,
 0.40669151266142134,0.687780819288022,0.027486114660987826,
 0.9006926280833412,0.35680324535842634,0.9164407591593601,
 0.06471806150694481,0.3655450852028933,0.9694393622122862,0.5513969183539188,
 0.06902520761460784,0.4613101464913276,0.34030507575075464,
 0.9654137239541454,0.8111305546155616,0.4542409739151314,0.889941361887321,
 0.06255733498001759,0.5220568768720124,0.022092129487668033,
 0.6756646341273227,0.5315303909496831,0.7112178810515937,0.9451760269458278,
 0.5451125072960387,0.8068078976651328,0.1724150177682675,0.6932067955273102,
 0.13813849486515462,0.30950380503707464,0.3968451587550469,
 0.3260192765693234,0.3193618967506415,0.08289109479538037,0.2942150425461215,
 0.2845913488851195,0.5331809264145896,0.16323610274098266,0.8542223916369044,
 0.9641092188187993,0.3417570585974785,0.6404687523781594,0.5684874860309456,
 0.45538644345385015,0.6844878491454378,0.7234900387995938,0.5415365979869788,
 0.8949680179371178,0.43488599111248394,0.9338079305942171,
 0.24713837500723546,0.5922758263906084,0.020359934266216406,
 0.1305924148219334,0.4547737067830364,0.5576312193676608,0.43972681439329064,
 0.6878988399777749,0.07950174133896393,0.45284651337014237,
 0.39968256924261114,0.6737284605155087,0.0840603316150208,
 0.40145018925756504,0.10676753360299773,0.5205931702292013,
 0.13289483442445194,0.4633733769296645,0.6988742313076537,
 0.19634282907732759,0.6563547263080536,0.28897499736974175,
 0.9288061362004131,0.6084437642537583,0.48213811139118057,0.9133927055558196,
 0.06404523350550106,0.4493090007047593,0.20904814040051123,
 0.7582444380378177,0.4140627706675698,0.6303603491331552,0.5304504352651769,
 0.24041189243934769,0.5521127516075998,0.04847061402499142,
 0.42028554854691924,0.7962507981775107,0.9523910891094719,
 0.08945010542020548,0.03678256639502375,0.5033595148277015,
 0.7885499487086113,0.9088400135523637,0.5334556856710098,0.4431342447361498,
 0.8818839390129957,0.011879378935462426,0.8799719339009601,
 0.08723703828083573,0.06456216329289421,0.8421879241859047,
 0.8314836414219824,0.9710553279195312,0.09842323158079058,
 0.014080613133514164,0.1947567145745458,0.1075869665013206,
 0.6307135848893707,0.3996776861019259,0.1819232188273674,0.9102086573465034,
 0.45460351449739367,0.5111883392050529,0.507846058049388,0.14397803263443576,
 0.39732918198146605,0.6733241288719832,0.7400144851016479,0.4197321700061398,
 0.6128168576524193,0.6341514904982186,0.5961790080186912,0.13923837759993352,
 0.6805822515361969,0.39957299509780286,0.5320660921803024,
 0.04525555882888643,0.751552558329743,0.08798219793646822,0.5796749444196352,
 0.253506341487035,0.3019236131216614,0.3729513138605678,0.587980151203125,
 0.31908692900706515,0.4459480332990845,0.175157988041172,0.7179252135054832,
 0.8801295895581541,0.9072434315488187,0.2327242190145683,0.4420384362244292,
 0.7060335854588287,0.712417725217551,0.19882140441955387,0.13771131191682318,
 0.8431299408888457,0.1855555215063991,0.04436218318676066,
 0.16521908799435214,0.7271753352090832,0.8737564622073115,0.5648725766692833,
 0.9248557626640742,0.07052740330438356,0.8900705985730173,0.3194309930975501,
 0.31887323945013657,0.3536140493820591,0.41147391448381865,
 0.3961861021714962,0.2875317845664809,0.08128129112007942,0.443048683979658,
 0.5855830359107975,0.839170504824835,0.7100272255121104,0.08456482252950903,
 0.13376259929919465,0.23778159352423878,0.8367560869399946,
 0.6689339011134748,0.9973687554750764,0.10244963032523458,
 0.07100603473721123,0.08790364362940162,0.9227525637958355,
 0.8015933738005211,0.6927059017116962,0.02518657574435712,0.7291524958184213,
 0.8688648237538694,0.448675782703977,0.36662050582850203,0.8219745373312799,
 0.45957770119177177,0.678754128438888,0.3550080555826556,0.7828572724838779,
 0.10996369955355656,0.02423021573871298,0.8221871067496412,
 0.5774482150081973,0.23739080089374742,0.4559794428548637,0.4987530314258035,
 0.7697486893857971,0.4710802719978491,0.029860004928513706,
 0.8601227081588216,0.8975353985050862,0.6512569744060897,0.8549457118924526,
 0.5341108619362198,0.7726623645802267,0.2089093311283836,0.8535993001089571,
 0.3956862068494824,0.2734043255513977,0.8428426976864783,0.1899329149927721,
 0.35907448312028567,0.9994263445195604,0.7146854902009975,0.4041530428216802,
 0.9266529266122181,0.7500940822400015,0.974877652417339,0.09344959223834837,
 0.278877325197877,0.11954960614265764,0.933061196047128,0.28601063478527566,
 0.5929863461858157,0.8262178123491287,0.03712022248400493,
 0.060045450717617466,0.892966831900106,0.42649718658110447,0.838765743606026,
 0.09877553050657384,0.10522493991859183,0.13779989919970514,
 0.4504081663093379,0.023115326415228843,0.09609720336945105,
 0.6944854328294912,0.3579837484933611,0.6728805227390089,0.9077507542749979,
 0.6174107494048463,0.5440753857146687,0.7758618371195019,0.9417886357010801,
 0.4951976662466385,0.6179904353120579,0.321891233630036,0.6465864436062347,
 0.27917980374242024,0.943452059213184,0.27976530319297954,0.4898744466789823,
 0.9682138804303306,0.408630451130099,0.7878438546120983,0.7036036253250497,
 0.9717542247194725,0.8218460319044968,0.8988765868891321,0.5339172482575066,
 0.1032380877128386,0.04459852784420504,0.5467098312546995,
 0.24302375627753914,0.9690361243371302,0.667041649110679,0.8847665809324875,
 0.8383886408713745,0.8047815296982348,0.4911259531422598,0.5634574595312194,
 0.769445369530806,0.21838011188099726,0.6802845761554484,0.32893619824276665,
 0.3827260154170403,0.08046721690855052,0.14764042227566898,
 0.8877352039255975,0.8206407164988962,0.1761082696578845,0.3025063154430119,
 0.6537468179364171,0.13006091753515892,0.7502949220597662,0.8720795484218782,
 0.24947825524382727,0.8595647951509212,0.2844282633268961,0.6441843395259174,
 0.29860660864675603,0.4839417232433515,0.8184165431465128,0.760952601872635,
 0.6973714220873186,0.6521714034244538,0.8071610201041343,0.7003043757914262,
 0.4825856755987977,0.5674003582811745,0.21378691888855827,
 0.30915009073465916,0.9329553963230266,0.6975481228854361,0.9634419946397157,
 0.6743007805072736,0.7735246274647167,0.27581595310381013,0.5133235781127388,
 0.48280646829865714,0.904351509501031,0.30708117543567703,
 0.49363735844617707,0.8784793742419725,0.9497983627832521,
 0.05557239040527295,0.20128406113975317,0.4052717162687689,
 0.7214738646437939,0.8435771424279048,0.015912580941963128,
 0.26533961124366734,0.5949387031701067,0.2946064055503239,0.5126463193975652,
 0.8156876849571016,0.07142800549270534,0.04501174518101525,
 0.05459328012641396,0.9962773847137835,0.3799518733767365,0.2542663716911695,
 0.5654389778541806,0.6050187194913169,0.6626114852951347,0.49231948390822333,
 0.21913712622131365,0.6619173450951157,0.7243460046519528,0.5368872236506437,
 0.7000692752889696,0.11985194757253381,0.906610368591579,0.629053222927362,
 0.15437947854315714,0.9129570385384521,0.06113229894509886,
 0.44106355559275534,0.516347794554031,0.04088805897012693,
 0.22718855621028355,0.6760852535263037,0.7492370340111281,0.5457162985307658,
 0.9336737852807387,0.26818620680465366,0.38041656131058343,
 0.7790956480613548,0.2964472284649883,0.6608328832931342,0.5970599267196953,
 0.6124154067093065,0.603852349147075,0.9137946171295597,0.2798357736260102,
 0.10462358217067891,0.6401227088702934,0.7977358710759551,
 0.16148653511394628,0.928190221087665,0.3827036703359089,0.20387118133271298,
 0.07814650980896465,0.5665495353699115,0.9069503162422642,0.984536023304104,
 0.3132636585707457,0.8453226513980927,0.7697729496866114,0.3987458929636183,
 0.20118880953923035,0.04028599660389265,0.29641970265537754,
 0.587133731264657,0.6000108452759381,0.37745289677111904,0.9656725853667014,
 0.7138640446877436,0.8669592585661056,0.5580036145370142,0.8887911139839344,
 0.9466474842934266,0.08866115380516815,0.1465150671959733,
 0.20405441106627054,0.24374097999562305,0.9909213948322871,
 0.2960174766195901,0.8841060333001115,0.8996276525241825,0.20316514989884626,
 0.8083198587693283,0.04403492244183016,0.17624010510193977,
 0.4686998766957148,0.19334148300228227,0.07461628868758907,
 0.2694871196094215,0.5492841232721882,0.6561186924992961,0.629121431261246,
 0.13781934632710502,0.358424199568531,0.8877788516207081,0.31959733801081724,
 0.6999073795820177,0.18745832730604473,0.769759630050036,0.1848553058423763,
 0.5640999940567265,0.9831749445687892,0.6569365595575883,0.8331957811659824,
 0.4061987203867313,0.6288334342066284,0.5046751487549135,0.9016341312072175,
 0.2393994408614114,0.04555578455571685,0.06757691993520187,
 0.03390098684170506,0.7647930044329972,0.7925979847225335,0.5389253901338288,
 0.8624674886302589,0.2725413796999506,0.6786439281960465,0.9645948638125434,
 0.5155031833630989,0.6391244608650332,0.9677323349351046,
 0.008626107654810422,0.5114211094454046,0.1802840089460378,
 0.5956301617802366,0.7436565627139302,0.9991073685199472,0.8150810610410533,
 0.28901461505733717,0.41162871482475494,0.27404106940621575,
 0.8338552452568024,0.3397146282509299,0.46554133935501485,0.7685770131823724,
 0.14606906071857784,0.08898538941609158,0.5623761956100135,
 0.3513288163209992,0.4514992342812709,0.03854966465585008,
 0.32789826789313714,0.6623735261179273,0.5349585049317013,
 0.24306089146893828,0.15446147198715188,0.004175705731217394,
 0.6195486643851087,0.5155599519972034,0.3342414915736327,0.213463038016247,
 0.06505385363313698,0.19327482655679962,0.2576994725711912,
 0.5451387019256279,0.1950008899977037,0.2934085901970942,0.3616858218000075,
 0.9436099622451386,0.36219126747299624,0.06050378266592926,
 0.11442668579702686,0.7898415322908868,0.9914763543404347,0.8571140810560772,
 0.3189306745062189,0.9821868745781359,0.6511363020648908,0.7706009915672016,
 0.5764691334929088,0.23429854012054707,0.8446423716962184,0.5434556022962189,
 0.7063587220318033,0.01768196548875811,0.22944707908135276,0.701306151358532,
 0.9398134235083353,0.7232285753598976,0.3845537568806485,0.996007731194763,
 0.38906779675515224,0.04057790473912992,0.09791709679948224,
 0.3042890279539491,0.6901959470888206,0.9961165126234921,0.8440422508524725,
 0.15365036641357321,0.027440903969993435,0.13929939223586119,
 0.6472673694481679,0.5228310102339133,0.5460602323652968,0.36570308381387795,
 0.255302810513806,0.4289805357355876,0.5395734393327268,0.45443135859726846,
 0.2299807907723479,0.3618497055063097,0.8401955794091599,0.576973692005148,
 0.7165973572615707,0.766179363902832,0.7478765084045044,0.07199892195193902,
 0.07815069842620403,0.5551669303270674,0.8384942526785835,0.582778819202807,
 0.5347622485278302,0.4625519066874523,0.4590364070357643,0.960547980805637,
 0.8428365867359989,0.9240148516907809,0.31619545547294325,0.8846269761987305,
 0.25305592524315,0.33870954203153725,0.1443919195640435,0.3759779605654203,
 0.17493367921187097,0.8450539324474347,0.5883295049355235,0.4076094869462692,
 0.1931956659054307,0.38273935670666615,0.3505980957653292,0.3962788109160611,
 0.13759545760039615,0.09837389111171557,0.4702735735637734,
 0.9593030414783869,0.6835904980628164,0.31230705091705346,0.679202286809021,
 0.3681460564348028,0.7821496507667699,0.04692036736442751,
 0.17590725103622518,0.6256739286756781,0.46087715558383224,
 0.8881625456003369,0.7667123792610464,0.8421009188650889,0.3086997358575796,
 0.0026921827367694107,0.6877511147113559,0.4376496279134072,
 0.42051468053901475,0.25574958503150635,0.8865930955494189,
 0.6527327037692419,0.25442013950837183,0.9729059248720991,0.5237015112850734,
 0.24701923060913822,0.3801244685598899,0.32574178841199075,
 0.6953496940276538,0.5058186474964399,0.1433359230545177,0.34723664417644606,
 0.8529782056866804,0.839180755346509,0.04683941591800489,0.4755137350611901,
 0.03572490587735777,0.4156987408879633,0.2398900259466188,0.6252601143759142,
 0.11090672966384907,0.4234428850311992,0.426918169842494,0.6942659147070743,
 0.18035498719567078,0.05316784542260877,0.07199280259991081,
 0.716757071607133,0.6620225190348483,0.3660082070143962,0.7220113856516489,
 0.7205574954331512,0.08099121333200066,0.8728573692638486,
 0.09965622713878652,0.8392761957071564,0.5344913422037207,
 0.07743253529770022,0.325201323059906,0.45301679362773495,
 0.23519322319623703,0.21634168552927002,0.773406044928634,0.808126101109247,
 0.061718750825817126,0.36722112731835466,0.6630910245506847,
 0.6141828284507966,0.05685472668768399,0.3021037897225445,0.9280840842569043,
 0.679690889669509,0.684987542026947,0.07907388248949965,0.778037572724856,
 0.27989693413990546,0.6709163548651942,0.06440695902428739,
 0.8291028721587687,0.3895394257506939,0.5272816329823539,0.7033197706863952,
 0.6503069122822882,0.6184849641400308,0.2355515255372781,0.12862764804937687,
 0.05959155885320344,0.6942270376150718,0.31772382563933677,
 0.7444127181751947,0.5384608961357211,0.3947639991694535,0.8757214018571094,
 0.1960602144299568,0.06064887683991438,0.1678196473215614,0.2155825521187008,
 0.2745142613309557,0.5787151369438563,0.49284296679387984,0.3530588589808048,
 0.9799684861699953,0.45218367002883186,0.8779968440796788,
 0.08967976979386638,0.7856065541273892,0.020817475813423858,
 0.025262254307399257,0.4818664710153714,0.7822945289601342,
 0.8337378688860345,0.9889155270962489,0.5690264325623567,0.8677373545584699,
 0.2672813979938522,0.5779413979586131,0.04473121785567835,0.7200420499877445,
 0.7897197730849077,0.0023967301039127253,0.2229422475175722,
 0.8600689944767486,0.8325754024559123,0.45332320184317987,0.8989339330301508,
 0.3076721547004435,0.4209136220505024,0.5170713767164676,0.8396243588321535,
 0.97454156032958,0.2685291834582688,0.05997670263769317,0.5711297687450125,
 0.20508894458576765,0.6097464005357017,0.4010917840069249,0.5916864103181627,
 0.3710818316448048,0.1705446491477326,0.22987555777706525,
 0.15755510067719514,0.06924669398200889,0.780386088461549,0.6212818297520769,
 0.7428694631003683,0.2176922174711925,0.8188991533425143,0.40449040069252273,
 0.6097970830171336,0.7135057319862192,0.730983355413185,0.4501377182659171,
 0.008533549108223237,0.08232462713631106,0.4755120610429817,
 0.09391418987578237,0.48568109893675904,0.6752420460125318,
 0.7372073809730376,0.29491007620461573,0.9092841901103476,0.8700780012715659,
 0.13831038153024788,0.7603695238534804,0.5856544283650308,
 0.26930922754279885,0.6763267799110672,0.36410722505676985,0.402721355444141,
 0.4138671047348992,0.7894986239293994,0.9452748799862537,0.24573957355082632,
 0.2908532417682168,0.3868864544142855,0.19954178909125087,0.7510255027173093,
 0.6949896511153821,0.22247035007414329,0.330423921103815,0.6810446227816861,
 0.10494908234678202,0.16886682419338328,0.44390863301616124,
 0.9709447661550601,0.03324330758123606,0.5085007239644145,0.5728727948397333,
 0.22051609539749006,0.10626655972061183,0.9361623422965911,
 0.18746127001733282,0.8229873379619536,0.3042830770347086,0.8374559907584643,
 0.37556239710417416,0.07864550917835966,0.8795456769463612,
 0.7304874381945612,0.2329142012089932,0.5553033189025741,0.9333746194492023,
 0.9753890263404437,0.1950500021727668,0.8655219778558821,0.8418927685638522,
 0.00914218993534155,0.7215682815159529,0.37583598786640027,
 0.3746564567200309,0.7636194665768622,0.924198814604181,0.285086116539482,
 0.8966175952659028,0.3027281930808692,0.7612985838722968,0.6052523806836011,
 0.41901922112989476,0.41380101805912517,0.7463599800819091,0.814482326374222,
 0.4662242383447355,0.9869620169104057,0.748912336672294,0.8673391837101544,
 0.9006105245628551,0.5399702023393435,0.3214102228790695,0.612760814580621,
 0.3497530065917539,0.8099070529762334,0.1598635301029575,0.9882505422670953,
 0.29090241361600044,0.7877665676906164,0.5595500981531677,0.9573275440917158,
 0.7811595111360763,0.3282063228548161,0.5452361018843883,0.8051560574837434,
 0.4184192562551712,0.7081371374878821,0.4860644091296372,0.28242488115011777,
 0.3556345512053434,0.6132952317543433,0.8521433353322334,0.26318018857723513,
 0.19383237499385408,0.4764370781383964,0.12080569797811758,
 0.16503881502193096,0.8293507873614832,0.7328524731522671,
 0.22512882137188395,0.5216915308404563,0.6255551710634215,0.4392734888229425,
 0.41155815710953714,0.6216174349196503,0.1159684489098407,0.9889508085317461,
 0.26079084279399345,0.24801835328631872,0.8765748578994733,
 0.28821971604350083,0.5709613129107035,0.994614340245941,0.4282340432063335,
 0.42566239207846057,0.9476016581027875,0.9491948353192369,0.3261982726074172,
 0.8716706694674023,0.6325427978611373,0.0362121799383317,0.5503440569899728,
 0.15182639293226674,0.042987011552142906,0.309888956563094,
 0.12007979885932829,0.8509696167846361,0.7107733897551775,0.9006675800205263,
 0.6635807378405909,0.29315656102781873,0.702279422157515,0.33850921084081487,
 0.14052155971034552,0.1376419520312948,0.3873504746233746,
 0.20867247697040148,0.5443377417159708,0.5437832323120083,0.8721216833323995,
 0.9196915103464118,0.954929741715513,0.5927049506586177,0.32492936792566574,
 0.40345941523650897,0.5494848751512093,0.7626786546276486,0.9515278288782674,
 0.6868246654369352,0.8930820311051327,0.0637958548121299,0.6063189221351222,
 0.9666502364319542,0.9397368942363257,0.10735283528711581,
 0.22821082585957608,0.764351902396821,0.2501657941087394,0.6295977955942136,
 0.5481211634417347,0.19724049199589244,0.32257912457568305,
 0.8631374234985071,0.04780692579716128,0.2842363537159891,0.8056075695142693,
 0.2192689449899783,0.04452230650665423,0.9483749802320386,
 0.36123536547403023,0.27122648979799224,0.6114177125516899,
 0.05201679934986869,0.2578271076544355,0.09588931492250685,
 0.5338415554850924,0.6313077999257325,0.42482747676208543,
 0.08860019279877684,0.18722117106042635,0.6480311201128149,
 0.2716248307658513,0.6834812999892041,0.4853327320180849,0.6368962994129016,
 0.9774148630332985,0.7865925772508436,0.8944328631354258,0.6538491332079075,
 0.7738854334917502,0.6746369091873723,0.0031854307224213896,
 0.4246341368868731,0.2825743200310633,0.9416432606773155,0.4643023537460733,
 0.9896455831907219,0.5314251429480888,0.874611409586233,0.9154385015097802,
 0.8436519512986311,0.13982442170175235,0.974741979066566,0.5869429845540182,
 0.13897143474977025,0.5892801112270878,0.14844316228203702,
 0.7413760347756158,0.674944907413345,0.6913274930579467,0.3578395846269612,
 0.6549736116922779,0.8064508659641534,0.47196290604064195,
 0.20190204325640704,0.8883908386625001,0.4931786031002864,0.9417659270881668,
 0.014870660834163107,0.15376498205264022,0.019696484920396107,
 0.4491946916797762,0.5265833349877229,0.3510508903447387,0.46369250776661974,
 0.05787541165308452,0.5998253065984396,0.915860212318769,0.6200519636059572,
 0.9755666184063169,0.334718560671729,0.4604315174813024,0.8726931616856315,
 0.15272845600189633,0.2058016082582521,0.8708874175254178,
 0.02129009102785062,0.5936261051808069,0.23641271641564865,
 0.7108477203596169,0.8429724599669643,0.2560905336282342,0.5766144277350502,
 0.6176154866312669,0.8164824056145763,0.8670371173596758,0.9411747094664952,
 0.9695988309868853,0.8408174716163274,0.7080512292808474,0.5445818940503113,
 0.35193103146018223,0.516508913516938,0.7054350157265765,
 0.057875197988906835,0.9552943800579283,0.7499008901561035,
 0.30395727931748084,0.8175839090317777,0.9554858910975181,0.5858289673008563,
 0.32725143395414835,0.813188137403471,0.27726787450627244,0.861318501234654,
 0.6544568345463563,0.2347198979804801,0.525913154400232,0.5816209098662007,
 0.7944745797135363,0.34127479575376296,0.5654737917124703,0.9693252863205706,
 0.6036281457255811,0.40834166620352075,0.10341153504460276,
 0.6056016857010325,0.5505383800975433,0.9223145130462097,0.38653907061575465,
 0.6942680368764206,0.5854216808243561,0.3441287377379729,0.29070468164647634,
 0.6948821109980624,0.6686788967671389,0.66099253538273,0.7456691397497766,
 0.17656606770384975,0.9150431783722447,0.22344066933018403,0.616663662738165,
 0.8655415244105519,0.32427886433417485,0.13952902424047786,
 0.3768038342193255,0.9414694923419874,0.4692732442474198,0.743738634461985,
 0.06283238779109912,0.052793955339016096,0.6891810168118038,
 0.9142824071974502,0.8869762494258729,0.06123387790430257,0.825942829193808,
 0.06277748121156213,0.7554629509328924,0.9924608940711687,0.679392179542592,
 0.8324317136117769,0.7337539192758612,0.26802033542529824,0.692795617321798,
 0.6682786012027028,0.87951385150262,0.20240114156361888,
 0.0045642686782118025,0.914784459362703,0.443062554680396,0.6809313673896379,
 0.8312488251572725,0.25628022245760207,0.16884282573406184,
 0.28948768419138915,0.6948391558301884,0.9353832261641565,0.7534575715770722,
 0.4159955746338322,0.3255128422175474,0.020474239351993884,
 0.009637686265357548,0.9711474083505739,0.4504527405589609,
 0.18162697968886454,0.6146306218142548,0.8067208718040768,0.9217882092386771,
 0.004103644958733388,0.6850976616256778,0.36984960070417916,
 0.8007751004792698,0.6832422920357353,0.42769186135883186,0.87230128477156,
 0.5345569595362183,0.781361491319953,0.30041417378038826,0.8910321303426423,
 0.834651706550878,0.34651059317036526,0.0629684780719233,0.25941514939372146,
 0.20135930744091346,0.4463613919060283,0.6674567830981274,0.5653751973053898,
 0.042368059561627924,0.7381962024256042,0.5522249262726117,
 0.4477983605447762,0.033904214414794076,0.47730930566977625,
 0.6360893740422636,0.4243599255842231,0.88285684894192,0.1795445744788735,
 0.36262546508316096,0.2690034269133954,0.8042292850326974,0.2827485503155057,
 0.2593215707078521,0.03490344860225125,0.455498411067109,0.7338398542880984,
 0.200797431422588,0.7204877585340137,0.32068324889916067,0.18412164588491742,
 0.18795399303197968,0.1235954926534244,0.9929920469708221,0.8912316136040455,
 0.8841924984082367,0.023157221951157658,0.8519554241083566,
 0.06469626859592181,0.8629759248349572,0.9701375823088961,0.30292015890224,
 0.2841678814020001,0.9352897650837274,0.0806806947229024,0.5938688339775446,
 0.35204385229593227,0.04741249355311994,0.04077484497265893,
 0.03744234654041412,0.3354031651901599,0.4804944231013746,0.7041581457268289,
 0.9036903091437664,0.038511454710687176,0.7735683688612667,
 0.15074763279572923,0.9482215865792554,0.8107267051335098,
 0.29334576758213426,0.7736303760807018,0.8140744223814951,
 0.13733630504661098,0.8737561356573017,0.5734329203228805,0.290616440483923,
 0.3167371282770046,0.4994970571373595,0.3413294751723721,0.9629278075400429,
 0.10665021122037788,0.0915051572316713,0.7159229834118509,0.84229050220179,
 0.8194528139674774,0.28083028863418613,0.8879328870628085,
 0.03406482633528507,0.32432953679607146,0.03869700877130078,
 0.003659187863954916,0.8487348664743137,0.8567467134272544,
 0.25001053126361017,0.9893905297534327,0.5797924164986668,
 0.39311291568763185,0.6809810243763849,0.8331131343410092,0.9295121405011287,
 0.608338797304283,0.9578751688405625,0.9601693100131603,0.05744528316859199,
 0.6758495465108495,0.8553278040452027,0.26612547644351225,0.6156746951995841,
 0.3777877256476231,0.9853016428733736,0.6128118843202649,0.3711113116858995,
 0.8772211782777235,0.16678220964783608,0.4169414365335895,0.1517807383140004,
 0.6341449811551566,0.4762736425751998,0.03157348647499347,
 0.19626313097967885,0.6847566451406808,0.2060185501599583,0.3219600217838562,
 0.3047263270802444,0.1695140726435166,0.48669027300767087,0.9533182216958711,
 0.3077492870932964,0.8161071005616444,0.1575418782318403,0.9375417663854303,
 0.0944248451168338,0.3286160241045729,0.8304615963363345,0.3380927438205267,
 0.18467861318109235,0.16629325719623633,0.4992733214361328,
 0.3384744998370395,0.7170316289318283,0.7594686212425784,0.18835005883268618,
 0.5408602044160811,0.23921382435967953,0.22201410257282905,
 0.48226423577520294,0.06242735804077337,0.08125110557657456,
 0.5698505046143346,0.15430022166697666,0.7625893211602112,0.266964184984133,
 0.6170970530924362,0.48216883158783075,0.9638242827645415,0.4221353067094614,
 0.7767874626253871,0.8397241734998091,0.7060310537707378,0.7147759671096702,
 0.3985664375372169,0.7688577349076661,0.131078576817529,0.1969458437051358,
 0.7834880255311297,0.5535173977540617,0.5987314565119924,0.04439734166175602,
 0.23159738896706683,0.10451503469643975,0.045836833609024685,
 0.4860847458259516,0.3881317352432937,0.4352249225587209,0.9463166245525126,
 0.27470456943413635,0.28675802792649874,0.1088139809387697,
 0.6251463862597868,0.06368525554812832,0.25236796283517415,
 0.08074150703720706,0.15241825902923822,0.16192100028590783,
 0.3160717764863955,0.00843392738907034,0.020266796693086064,
 0.44069522698389985,0.3819031532234679,0.8190780074065044,
 0.007534748636568556,0.14038629900959843,0.19232896822437984,
 0.9136965769918781,0.5184186683386833,0.9976606251176702,0.06737373385829704,
 0.9986741300823151,0.2606649510803749,0.10922464553268618,0.852969391752434,
 0.08839559961257415,0.5266259614712727,0.5862179998207161,0.7591611597249596,
 0.787254747063812,0.6672353249568587,0.2743556379100277,0.829847046176548,
 0.6466550902344849,0.5151928594145774,0.6474059094856579,0.7601424315621462,
 0.8118461998565405,0.5479807752247037,0.6244745353102692,0.9253563477317405,
 0.6081499763433718,0.8454090586138386,0.27504167513907574,0.6625427983305592,
 0.49828991511415555,0.9610767939648188,0.5396591328726248,
 0.32927742548975625,0.785213739676578,0.5994278507058253,0.14509904638840276,
 0.9268385373831691,0.46351336688694644,0.594173134957014,0.4980537304912728,
 0.820928225485702,0.42482666598182106,0.464214144061845,0.7082536923723147,
 0.4809012036257747,0.6442842763063871,0.7509193096508516,0.7539887129935089,
 0.36697123291613987,0.41683010375194574,0.3409644821545821,
 0.6712595163611496,0.2589529136817935,0.5761648498542932,0.19125916017456457,
 0.3682618668673726,0.8873764642039891,0.08925669696479677,0.0650109988003924,
 0.5287041745758607,0.5955996564739663,0.11174334674283803,
 0.13601845801117018,0.3210876272341867,0.7486828726699978,
 0.19025668117707495,0.6056298412424205,0.49432749011499033,
 0.21660240078709536,0.8563102791121048,0.2755680218992618,0.2868941877771378,
 0.9109013030275481,0.6370478304855478,0.5762013410222633,0.17805030283619638,
 0.40322136010144694,0.775281277526131,0.006487078684589642,
 0.7187599047777676,0.6653371486080748,0.2667368239676787,0.30375558029032534,
 0.10616422087787702,0.5552288345394909,0.4145212012706446,0.5594962896054578,
 0.5530331047844298,0.05850673065288359,0.5250651466098473,0.8296961642022489,
 0.8786601723517039,0.6193945787185503,0.6485259941759288,0.18428205659746266,
 0.2576480450636812,0.28501833826522305,0.8428267276553837,0.8696432371754134,
 0.36606238354483345,0.9158502206465419,0.465951500618377,0.3959240582268916,
 0.7800881857564126,0.6655302384272265,0.4932750562576801,0.9907330658758455,
 0.07050575405599768,0.15093064107362397,0.3594202242902239,
 0.5161284192917962,0.20117443079774477,0.33192798578934135,
 0.0016441266876978666,0.47826349335237006,0.39365589922682553,
 0.3553401701653358,0.46317580163219774,0.6211574288134979,0.8577591181443186,
 0.49072628337490054,0.15658767358081482,0.10016642336809711,
 0.9886174972804627,0.8796710095361946,0.7151224073822863,0.8314253571823643,
 0.32035779869589875,0.5650009557258168,0.6559450214222102,
 0.40682728398369516,0.02003394643105527,0.9704046394238796,
 0.04111015312785393,0.8393044124913015,0.8195581429585389,0.9794836526715804,
 0.32747521398029833,0.5828977789372459,0.5601971578484226,
 0.36703638998731347,0.9464433433257475,0.14010847834921403,
 0.5412333101061014,0.5412135901574137,0.28598188875526476,0.5586068958864533,
 0.915332312696196,0.2674077219192973,0.018629011638120252,0.4984644115472,
 0.39841554523258316,0.7096030127606012,0.4500246893267942,0.8782984588626837,
 0.6299712090297281,0.12837719836334716,0.01118539638893612,
 0.7699479864315784,0.17977273588963816,0.4683975381597534,
 0.30657841314783674,0.018821957301374925,0.3768120000080578,
 0.9291914092803543,0.7644318872195441,0.8676605240071544,0.7559518479407483,
 0.500885897977487,0.7396250883123535,0.40741923906138766,0.5659822044841746,
 0.5442300693115909,0.33508205968640337,0.5656872229863623,0.5302111657770482,
 0.23662403011913302,0.9869487797273687,0.2588013079230602,
 0.23021880423880203,0.8193678740249095,0.13320999077114604,
 0.3358209556108417,0.5284395521421807,0.4545952185913582,0.6775328284064335,
 0.24857323921654373,0.13766678429326684,0.6002616090926327,
 0.9441925269647137,0.8997373995910142,0.06321008259290395,0.3147771844127737,
 0.7705745141150873,0.4964189885320711,0.2526327166791942,0.8900716963855689,
 0.027681349524091403,0.26163493811651706,0.2424174324210312,
 0.8110598149521867,0.6261549099983087,0.2948473278768141,0.4870753580214623,
 0.4816995047036907,0.9590457780906144,0.5858982049581698,0.6837592047574228,
 0.6137723451806599,0.8573137326816909,0.8625689913315583,0.8753943121928709,
 0.6207473484936873,0.6064812844440646,0.7118815157783729,0.5717077337743774,
 0.18102272297925281,0.6075000929579567,0.5928804173859328,
 0.19011777303544286,0.5632712944772466,0.7435931296501865,0.9404186276779509,
 0.40957534723548805,0.4324356296352704,0.7263717151971654,0.8262900293670284,
 0.026831888529627368,0.7591341392171086,0.5178200592473096,
 0.14468926657163383,0.9300867607560671,0.9942312777987482,
 0.08155798455435226,0.4936115606355016,0.17905352300889465,
 0.7978745806674377,0.37118990168495825,0.7572125788792108,
 0.20355021098969361,0.1749533926797815,0.20324719819786907,
 0.8418828968716077,0.5961863382331125,0.840587038788919,0.43443715421776474,
 0.315719967586247,0.8837857812164477,0.038602188645627145,
 0.013490459312318848,0.6616707900394193,0.9797972875870278,
 0.3670100366625111,0.9105612693507364,0.521412048982892,0.11779000372676107,
 0.7159562214910601,0.16228223032507472,0.3647685930894624,0.6818459146917369,
 0.1262864783609864,0.7855423756212214,0.719649461975802,0.6497557799138476,
 0.7569895760317196,0.12843763591045787,0.26298211761708834,
 0.7990351039979101,0.6393404515980431,0.17343881872330968,0.96999224958029,
 0.41173484617732237,0.40965866462015976,0.23481047534373767,
 0.39021543954580984,0.7424546924320518,0.9078081970900812,
 0.11551709621622208,0.7436802087124949,0.19435046172192372,
 0.5219978103403372,0.980285281143292,0.8008965478666404,0.18823260327919245,
 0.7118288259006222,0.5218870108173483,0.6200062707251305,0.3307611567969828,
 0.7217994884729285,0.6816746301246961,0.9052185188175219,0.18421899129717723,
 0.7522975066788066,0.8048777740475104,0.4326129752080916,0.10894412467374093,
 0.2725696657951415,0.6059872462826041,0.756369057407008,0.15374955280669012,
 0.8402019975117347,0.02106396467707128,0.346248799532314,0.8997494102157716,
 0.9324086617823988,0.6044010713312117,0.1755971483792661,0.7325933009706119,
 0.22383029897077633,0.8718254294817743,0.780772760165765,0.7832104722675371,
 0.2467437873980196,0.5866056171458669,0.060377747919439284,
 0.15399651409465798,0.43574188666026126,0.848661758049674,0.8999503701269067,
 0.8399323630767961,0.9709509123528588,0.034651486681889665,
 0.13322853586663852,0.5891223613488833,0.8505012962254869,
 0.06923797545216259,0.4253998841214697,0.6082573867654739,0.3503555449581386,
 0.7566381227968247,0.19786151587970924,0.3667874167706823,0.9937761673752478,
 0.7448180952487196,0.5689703409998211,0.7721748597188635,0.7425942993366895,
 0.9075443536900292,0.7134837702011052,0.04887948636651673,
 0.11961853647854559,0.7214590280333527,0.6323106420506678,
 0.18236937497401084,0.07141286261658542,0.6027381292540207,
 0.5453506477664642,0.4576357606586201,0.5062903898624064,0.6246498767079459,
 0.4967112769371598,0.6300225516188425,0.22118119970110084,0.5382857893538202,
 0.899649867469337,0.31843021891148204,0.8951760956016116,0.773578738414875,
 0.44373618019480254,0.6774812667843505,0.35784270518077876,
 0.6178818334058852,0.1395388460006144,0.7419574701616515,0.3293956608112826,
 0.088193562338442,0.9302556408544181,0.16189654227991712,0.3889688361967252,
 0.34735761080427485,0.1839256830993713,0.162490042791686,0.2749885203041029,
 0.49768324239731665,0.05946683371302419,0.7178129692288437,0.9200929711977,
 0.5402335443869322,0.7176537375372063,0.3156045071930178,0.08081653625411622,
 0.7216282092376909,0.023276671259308657,0.830489808558573,0.4382214175801392,
 0.6562486370502679,0.752975802692751,0.5239157232094317,0.8734231574572878,
 0.5414758328750066,0.9998234815496427,0.6998280312429516,0.31044458813426246,
 0.6220870170954821,0.33872310904579017,0.5297827319038034,0.7132149563627856,
 0.47760000617733345,0.8975250271291332,0.6557455742251141,0.9683661673281393,
 0.7044040140200388,0.0305587183064312,0.839118095857583,0.8851763788384893,
 0.22787230055270646,0.709743292460624,0.7443204969198279,0.5511686583887577,
 0.22290746289446695,0.7690776511956597,0.7067996302479306,
 0.045268776425140045,0.2123834695549931,0.6816327382275118,
 0.4470674194647968,0.8492448762443069,0.3680470485191013,0.6151458961782491,
 0.804623779719389,0.17054354743375344,0.20896234161931315,0.3245063994627978,
 0.027462726010784477,0.5452275130096668,0.7784442772276177,
 0.9639313498787639,0.052707603114628965,0.9564527127599952,
 0.09591905110415089,0.5766797484097621,0.4691182065947562,
 0.39258732135801955,0.5938206521918162,0.8910002981399321,0.6405529229714414,
 0.9260173365600182,0.7541107248049597,0.9353223184271835,0.7736697564910708,
 0.6182235926326795,0.14661468841294234,0.5330078972105249,
 0.04338116876528675,0.8835573590943312,0.23319584782308767,
 0.21568512775341286,0.18791709820298108,0.34937575998050385,
 0.24826354293001152,0.48832121276619356,0.04923167147250962,
 0.46602157042343106,0.4911037281201326,0.32560919118278997,
 0.2801380397346448,0.8240590029288463,0.0064651187214859185,
 0.47260039627480743,0.10991525998724849,0.962052333924006,0.9781392610132069,
 0.2248341814460112,0.47631708989988375,0.6123343864662446,0.8322420135990538,
 0.3705107943617183,0.9693531109493878,0.8991629793993076,0.5485481235009827,
 0.2981682138444701,0.11277826777236588,0.02549503470847403,
 0.8293075729283834,0.9077741867908432,0.6249273239916973,0.6713288537943296,
 0.5560194038998528,0.5103347335239076,0.8220740081716491,0.6821719428802422,
 0.5827918459006082,0.6728509850964944,0.09548369564730674,0.3014724655311851,
 0.5289261378974316,0.14451368336706305,0.34270801870957635,
 0.4922414805505584,0.9663983558745642,0.7803502432521651,
 0.011807076804771999,0.09637868113847958,0.2904130448088723,
 0.857702856612078,0.038598073037705904,0.08793393908857283,
 0.8863167797007411,0.24136764691167478,0.5484488552561231,
 0.14409823811055344,0.6579114954827474,0.9108412410181437,0.8148217651460209,
 0.5713631742858085,0.9246100912257526,0.7022696414081758,0.859449449824033,
 0.4569993774389478,0.675878927034441,0.016061585717968563,
 0.45942861364616105,0.31860001326730736,0.6852592055439538,
 0.08874546577219067,0.7243442727666058,0.92647792553398,0.2794293290677371,
 0.5383536872057041,0.6586287753019497,0.13211889346538852,0.3926443239957448,
 0.5359749934382811,0.01085633959334742,0.6435197198503442,0.4752376136199733,
 0.7071686824293476,0.36947350332558937,0.5125115021610175,0.6132151572503243,
 0.6054687188078001,0.9640671522453265,0.44938123543868014,0.433485726122276,
 0.07679588518328206,0.7701200427570175,0.5247000855839485,
 0.38073990943915925,0.09335694591692567,0.6082681285830196,
 0.7453652383319205,0.05398038044261377,0.44548975091582954,
 0.8798237433864775,0.8861602963374324,0.8374585843625051,0.7701890416817503,
 0.41243422191710155,0.6658545853544948,0.5162636611988358,0.7772115174641742,
 0.22625052789132227,0.8678705418153015,0.35900028423140184,
 0.6411895439763541,0.382361614158197,0.6408935638375307,0.7470271471387901,
 0.8568525970083068,0.5476161654991794,0.8765921992231854,0.6498928088997045,
 0.5726107151047395,0.3978781568550793,0.40089801882019815,0.4713349844333088,
 0.07602735858027321,0.6480160730100373,0.8980730859420524,
 0.22039663013833166,0.02786397207595004,0.4744449676150915,
 0.43371540906510964,0.3171120158498443,0.6736184897778955,0.8857206548054273,
 0.5420383585175392,0.9446797867361383,0.3755416298262597,0.8161530321951254,
 0.612326544585903,0.23641868191154414,0.8811514577282261,0.3411508540596383,
 0.03290237173729704,0.05227475162785056,0.6561162103568882,
 0.30305781664381837,0.5061407717933997,0.2025479890195938,
 0.39177403461883453,0.1464258710663836,0.44122316820879703,
 0.08502284083973599,0.8940710435977453,0.48761368439976327,
 0.8686660334590188,0.8715809775929442,0.33959409522324635,0.1318614522123135,
 0.5721499483747264,0.46161530474415724,0.736085917951086,0.591420526292093,
 0.8101636997953031,0.7023367470665137,0.6867634842077034,0.34061166119249775,
 0.018229899657040916,0.8942634571461133,0.5092733436571069,
 0.8404223376329576,0.7759538856664353,0.42035869534428594,0.4128777578818603,
 0.22590272350497176,0.5236591706598527,0.5210285319653345,0.8230734710442436,
 0.16135654907588348,0.07749069466637493,0.996505010715975,
 0.23740163532913938,0.8665311013374254,0.5401768564482683,
 0.031024357084259546,0.3201317723806544,0.7087731475886189,
 0.5603120804668581,0.9064302477090461,0.2736382271837853,0.7410883037387417,
 0.9722628590491368,0.5592438912035277,0.9480554441235192,0.39312181827755766,
 0.10497875253288802,0.5264556555436373,0.9434023303061929,
 0.024987984762320337,0.765587943788931,0.35137133603045245,
 0.5198448945128643,0.5963272413796682,0.6176315095717375,0.08922567529735437,
 0.522237233537717,0.6304765460384174,0.3349098424000928,0.4187207219728307,
 0.9372174334198434,0.31003308744223146,0.36821783367421124,
 0.3694653123162802,0.7821481994579786,0.6404429914008019,0.2586181434178675,
 0.6785889803494338,0.9815835693021313,0.6933618100759328,0.3430334388462273,
 0.039802952823525195,0.7525568698842197,0.5833536818906027,
 0.5551649415409049,0.9788316424846845,0.27623614139161523,0.1647952287325749,
 0.2849770006807726,0.34615753614816,0.16923319161590777,0.15998620177506717,
 0.3004625791881457,0.009576904323740143,0.4641988401854673,
 0.6631447387159292,0.8512226370502756,0.9094816598210235,0.9616466573067363,
 0.5060701733233197,0.6459046692008701,0.8946488540555907,0.5628972214694512,
 0.7113096624407534,0.07075380360321792,0.02945220923223779,
 0.16768966503131955,0.23718066503100632,0.0031686305130594894,
 0.20013902450371024,0.30348730662179146,0.8993749142828596,
 0.22137771754326452,0.5649301139190303,0.3127715384435421,0.6618173560872518,
 0.5722770600828173,0.9533610324922737,0.8831159302446763,
 0.036724733372385576,0.6597696718542208,0.03646509620383609,
 0.7826131728871947,0.17249145989015557,0.9359937190178278,
 0.20955181982282922,0.09678819626882273,0.26107376740304267,
 0.9410619326834768,0.9142464558809432,0.13035878393338474,
 0.17783392274687748,0.9779873540656492,0.48229558717224785,
 0.7157846930276488,0.8099700177807346,0.6333012735930356,0.796567487221277,
 0.7018390611925591,0.21420984654557873,0.23990862278056735,
 0.9429139316538402,0.9853046862694226,0.5208746511729896,0.40154369262712286,
 0.17997831951418997,0.6094667080261496,0.26394599572993693,
 0.3178897625220688,0.1666275443611519,0.6249284779239674,0.4121104031288123,
 0.9607275116293804,0.8431748495367022,0.5764170912797902,0.9478579857723685,
 0.837375891359311,0.8177479487247821,0.4866044770399368,0.31933249252373197,
 0.17509399129144188,0.49383097049370683,0.8586324235897248,
 0.7995495562693801,0.44579157583363505,0.7767126197619814,
 0.34346885246408065,0.1210501896143199,0.6308079293613368,0.9697672980241441,
 0.4934299665462837,0.379061614478054,0.32080405274905743,0.5549188316046194,
 0.17573945134676638,0.5047391904860239,0.6863728606880208,0.6377967836876205,
 0.5658316132241463,0.5013402215863937,0.2902369968798394,0.8883154955737759,
 0.5119926826305871,0.9248443337801578,0.036409436122550654,
 0.24309237346236046,0.9971482302310756,0.8679218671881955,
 0.39920592966833757,0.5015151093442326,0.6675249109781151,
 0.030138284012483973,0.9659903993778997,0.42342186094497125,
 0.08430844866229759,0.9887248349833565,0.2576845753088831,
 0.10644427920156097,0.6855003705292415,0.556975429148309,0.6269695501242938,
 0.27411669553708284,0.28413737787800764,0.2375430162224208,
 0.8879468803475223,0.842601273315748,0.8130343174776012,0.3707982190484105,
 0.9716318727317099,0.9395450430426087,0.2152204756370638,0.01767646037432402,
 0.1278267024580665,0.7135811892962992,0.7157044939072634,0.936909619710429,
 0.943369091512355,0.3931761381990839,0.7966047596856917,0.1718235175046673,
 0.32409507452245157,0.88447888073587,0.8846583691466653,0.2476853942379491,
 0.04794440739095662,0.5968716731159194,0.04693060045182462,
 0.42670239316828773,0.26554552108076324,0.5139209265971881,
 0.5132405205065655,0.2128208892062744,0.5682622647072662,0.6535797418754896,
 0.38980348500651196,0.21877899454488126,0.26255988583267587,
 0.5954991765891029,0.6770461238346267,0.13522491375646328,
 0.47161907438797424,0.8470493988753327,0.812766199778744,0.8591556043684638,
 0.7360926289212375,0.9332728133905165,0.06652606477658074,0.7875497243875281,
 0.5092476313808321,0.4298169207278473,0.9554913986407628,0.25428018666496555,
 0.8149176879344006,0.12238637623919268,0.41687237372152863,
 0.19076826120848267,0.7215917343688277,0.46291706165180246,
 0.03682926169294998,0.5732892264209173,0.1290579368509711,0.942426815271331,
 0.43999162378612133,0.06414859208154877,0.6481104964394936,
 0.6618720289167356,0.3409840336004185,0.5275886286105633,0.06835475632530175,
 0.06005044546847815,0.8816302044260049,0.986687547188626,0.7737725931569749,
 0.9970983907724248,0.3974050784208123,0.7133462263172081,0.803692360111854,
 0.2076586442112912,0.22777589918851138,0.6379914427701578,
 0.44310981894284585,0.6357478507572186,0.8056218686061966,0.5850143267187515,
 0.35189226726894296,0.730522347451794,0.815905212578743,0.669591603155441,
 0.41890501722685913,0.9553914312819424,0.6646388515271657,0.9764593860740819,
 0.5536711412997704,0.8950447114872476,0.09659745727106472,0.8575940277674557,
 0.8535392995172653,0.3437403535659538,0.011751669505490336,
 0.7525822957752927,0.20652128126869573,0.9133734635978872,0.8970210313086711,
 0.5868311967837733,0.28676911832009777,0.9636990671806199,0.9012797980300915,
 0.4893991345714137,0.8379191649086754,0.3998758933004436,0.4236318086080535,
 0.6417522811468349,0.23296380966379826,0.5184631956566429,
 0.16867238242503957,0.2887780493214127,0.6640447645519816,0.5499023454233547,
 0.30062521201442405,0.05596714308908313,0.5027564087423653,
 0.22448626182104514,0.5358798993345929,0.056448965324066624,
 0.36857694263782015,0.4848792446110324,0.014089067878788053,
 0.771091814206616,0.6945714942024852,0.6663352158103468,0.9933157614506806,
 0.41205737527721553,0.8978625311159463,0.13207175886146683,
 0.0950557129368026,0.5805869088589003,0.9587182390214721,0.36569654851247335,
 0.5896476359727983,0.5495273050486947,0.2400210921772845,0.45916411550683434,
 0.6931756135407748,0.6201913449429706,0.7050496878789476,0.43799938814033024,
 0.5870130925871405,0.011747003014041968,0.7905166735758705,
 0.06233092372927296,0.8367764829152552,0.29846195986830937,
 0.8354961812452257,0.21713592911732205,0.5060574384416692,
 0.09496196145978042,0.09201694271819771,0.7128319518062423,
 0.8233464850427002,0.1161187084872416,0.9570767461787403,0.15208701875898412,
 0.7693834291786265,0.010036055347513728,0.9740791434642193,
 0.05463943033308505,0.47224341261350467,0.7076343571567425,
 0.0145489062875509,0.0808362018957629,0.9610695331742046,0.48725449780746766,
 0.7365410946946834,0.22575605848202374,0.831730807087741,0.6627671259731456,
 0.7253702884741591,0.0027765399429919846,0.2731031619249865,
 0.04557071263971735,0.01510941027617807,0.7702757576126238,
 0.7987501974166333,0.28838990907371953,0.7754653377004656,0.0898531639832475,
 0.6924591090150679,0.38408826718339395,0.4053985346002099,0.9343663242573292,
 0.8814114377619904,0.9974526257727054,0.8321233063036064,0.4854702375256994,
 0.6950136504806403,0.9040626945460862,0.7233014464156751,0.1734910679620022,
 0.6934223286407252,0.8969706396173387,0.49061324797288275,0.153515152096657,
 0.17762294867956185,0.391849429219279,0.9245045232756369,0.7572628579434028,
 0.36770160144396524,0.6301002552908315,0.8807606362901486,0.6167393819428596,
 0.8864764698284474,0.42194816690627923,0.0010826963106910226,
 0.23347043495573327,0.12777885524117183,0.724262204845005,0.9901756876495674,
 0.7502998154191441,0.9472700054375942,0.35134662825806373,0.1716251923473313,
 0.1835844901004594,0.37412643299712145,0.028011871882506978,
 0.2002245773478215,0.137272803485589,0.21758506777321185,
 0.018571630552370988,0.751476373915915,0.07433034094291813,
 0.9764907512534006,0.6055297358341646,0.13948297878045435,0.5740132025981606,
 0.6509983354293014,0.0817054604367351,0.31699698972848434,0.1342443379119571,
 0.2144439524196169,0.5272646120181628,0.1262579499772326,0.6993514667127783,
 0.9663482414693423,0.22449875722004042,0.3105528779976554,0.866814416289707,
 0.36569922503240093,0.7624444378421736,0.05661707698063645,
 0.7826551392840975,0.09364221152622965,0.8986529713102519,0.3396513676717493,
 0.6581834496488457,0.10250934361986086,0.37625231142458415,
 0.9394572993820934,0.986531252853883,0.5496774450597279,0.9450734484331447,
 0.5418660106493184,0.9780315859075741,0.5166261869170666,0.5563611992892137,
 0.5011630819200913,0.17014643522049544,0.6726838710060737,0.2610005226167543,
 0.18449202274005216,0.8493062738862853,0.3222731171459079,0.7630777756016995,
 0.13726470779371647,0.4265268186146013,0.4391554987363052,
 0.053832643951786774,0.7592293539688909,0.5451987257277267,
 0.8712977502594764,0.14857160337457953,0.16367816990436168,
 0.6550219444717115,0.2552028045570789,0.8794206464014852,0.11042044051308908,
 0.1910706808546463,0.5879640506555247,0.09467623442262851,0.3163424228980438,
 0.8777052111076811,0.784341690802793,0.4716769439421048,0.7515755829063784,
 0.09279303743435297,0.81080673707675,0.9904608478256078,0.2390508412195148,
 0.8534172838908338,0.7537011601333207,0.841549888587841,0.9426773195890937,
 0.10152595602543091,0.7576343472762465,0.21968094802744242,0.711343662159064,
 0.16638327289198107,0.044247693146599376,0.8416448834420752,
 0.13890762400220902,0.4958609703279001,0.00649372297454176,
 0.7070088402450461,0.8307675120298863,0.18741743233280117,
 0.26891087020383586,0.8302558821320203,0.879217944455005,0.06207110786863135,
 0.6439225136386096,0.6930881505941263,0.9501018154231844,0.09167001421406107,
 0.4081136193529482,0.9657456159363629,0.048868253777807125,
 0.16622925725427617,0.12212285105241283,0.7461608128261077,
 0.9984474308179393,0.8587214473313699,0.31847116709251444,
 0.10907491640911804,0.6912788471714317,0.43969776033354113,0.580112468135922,
 0.010514808283392574,0.003421065699630854,0.3024935843663974,
 0.790214786318357,0.21083251545270498,0.8101079946055989,0.41395707543322247,
 0.22391979399963147,0.5778726674312934,0.37574710561587743,
 0.9186648095020611,0.9034895784024032,0.8923335327495806,0.8300099016918735,
 0.7385177684280246,0.4575557266863609,0.3344172270016741,0.9646915365811304,
 0.3775670243601724,0.8390893707326258,0.943051599649225,0.8065330339846168,
 0.18446511666651122,0.6041780381432686,0.5881653596768773,0.2686214375229319,
 0.9558802053628807,0.82812386692058,0.6157540700628646,0.8543148313597078,
 0.07171868867126274,0.287407569248736,0.7399873359193349,0.856103425754995,
 0.2844977689019541,0.008136491293313619,0.6557419071170132,
 0.48083421293419737,0.5567197108497695,0.33066614353520496,
 0.6255243608938074,0.7916646472150902,0.9471958192605227,0.8036767296214525,
 0.967325639851724,0.18824148259323548,0.3407433383326981,0.6145244394445811,
 0.6107133889246493,0.24569680102613323,0.007710737776956966,
 0.4136959892295321,0.8957617069461636,0.21243330121103843,
 0.35407087016863237,0.7519308599159229,0.5234730488932338,0.1937005434480905,
 0.4423486951620573,0.013099716794920835,0.7016562360991226,
 0.7998798088255573,0.7213795930279343,0.9506438117710037,0.85952150880007,
 0.8010808447531265,0.2075992949018427,0.6179772558736464,0.43095280599609365,
 0.27683952777697085,0.216550787015661,0.16167000428099565,
 0.32205273848445404,0.39124242320370994,0.8811610070875086,
 0.9083344457545204,0.31099524217469354,0.5834656268803542,
 0.0060323857058420804,0.6005344587288631,0.29559351533027334,
 0.4299682230704496,0.8874227879588851,0.9272253342324579,
 0.0014801381241863476,0.8356640081949425,0.8387242131648465,
 0.1939321256687614,0.5131252570399352,0.8015884937380461,0.238615103081238,
 0.6747566621087682,0.820351232496154,0.642618949491953,0.05263934710157825,
 0.8515165806795532,0.38827943931718645,0.7448716777056674,0.4354897602736545,
 0.2211445072675703,0.596846130404913,0.9885075159583308,0.5932044686741867,
 0.6162021909540928,0.5753722383354317,0.6555836552431665,0.8138862838305608,
 0.671316841008424,0.03050619244175412,0.9873044693803894,0.5721932797843965,
 0.34423277209818126,0.2836000198507115,0.19070480024073777,
 0.8492492300056138,0.05646412452339278,0.08920489752823568,
 0.8587262831492913,0.3664767231144147,0.9585857992126519,0.737182886488529,
 0.9230378706047264,0.9003554149528751,0.9806314635588405,0.9681269956977838,
 0.07859646784344909,0.5425561251881598,0.43379264837356835,
 0.8334852882129649,0.13215881754193537,0.9653187443442675,0.4281735961604619,
 0.8057006618723064,0.43195410947961443,0.13210419470089552,
 0.9477817416971155,0.1242677625123505,0.3666910069024647,
 0.036048981317574724,0.14770943784809998,0.964969313864439,
 0.7565888520099212,0.9683847562467534,0.49968450959059685,0.5617087073097619,
 0.06635829040574936,0.37188597473342777,0.0010925083495354482,
 0.6687969648155803,0.29303660199666914,0.6421164612094747,0.826368820100351,
 0.05874475270963253,0.519303879826144,0.04501453428791202,
 0.25013326741631337,0.5997891630397351,0.5318847084710255,
 0.15118002307761058,0.5260500803149822,6.53586865725253e-4,
 0.6123706437325153,0.9594535583197348,0.7630729441412817,0.9195657717883229,
 0.9200149625361013,0.5546290203529759,0.9426629534449662,0.6177254345029767,
 0.466373814028492,0.9811591175948419,0.7843947847204867,0.1453241956831447,
 0.2958111626007919,0.5504106178386265,0.3954494934392383,0.2538508381273007,
 0.9763291158990652,0.8699461226167595,0.6452911799588325,0.8201059623915351,
 0.5763612755666919,0.46605392485003905,0.7661689707850744,
 0.14953444235925661,0.22103665448197007,0.334682547113199,0.2411398428560073,
 0.07786120839203459,0.6814258811603346,0.4957986543889692,0.9551408582729851,
 0.7028917084107533,0.0050217504246035105,0.34752117848687036,
 0.7233581664007522,0.2291809287571921,0.20007294035582146,0.4865321216411119,
 0.08109853110164778,0.670273618191142,0.6733798721474654,0.8518716222141565,
 0.30428872749482694,0.8254075779885302,0.8388786061886974,0.340512066104783,
 0.9158418872977614,0.19521070186286038,0.9706168262556683,0.8518317453315614,
 0.35384726782017273,0.04422152999587048,0.40937626398144766,
 0.7800358133167968,0.6706687249302868,0.2027309210096916,0.8958716050564297,
 0.8672516532843085,0.9309958963729853,0.30432676547751214,
 0.12990619817984395,0.8082788588135857,0.14822613453559874,
 0.9796566558994972,0.807493906014908,0.009823452231606522,0.8471718366257326,
 0.7956752711699251,0.4908154057799555,0.516170533910397,0.7555880938546012,
 0.0036344162417060177,0.8007823548220818,0.7344469823374802,
 0.3963648636356947,0.33080327721907365,0.6701832161974274,0.8639608739317545,
 0.18253391137514818,0.7583485121962851,0.44940972412890967,
 0.7846075188562527,0.06533338884714046,0.11074389833235754,
 0.5406197964060921,0.18656913639702877,0.3598635547701632,0.6356448888004131,
 0.8677231669907396,0.22466057697209163,0.8338177120123892,0.4905986865358041,
 0.42382015206861834,0.8406165919874176,0.04859403787099348,
 0.07105789875056223,0.1047538259879186,0.29395854333986515,
 0.6225314832438733,0.38666298436921454,0.7953448822509841,0.3413266071735964,
 0.6821381659595305,0.7331618678220714,0.8217314239581677,0.07910800136571994,
 0.3411465274179408,0.07462511981015518,0.24599369009717442,
 0.8230141256673676,0.28270882368766537,0.9974038014348102,0.653827560593838,
 0.4483991656644568,0.21616299997432453,0.924088995912163,0.735660532494583,
 0.3799392271648119,0.2458226071977292,0.9288990183674942,0.17038814908628874,
 0.9485130495892902,0.027503599354052444,0.05529631565658799,
 0.37053994550952896,0.6996926657730739,0.5730511144914296,
 0.37989347574250987,0.6869769127651093,0.22130362800069459,
 0.42889947831438324,0.810524964650318,0.8700830225341402,0.7470587378326333,
 0.34728982351605653,0.17119763207380045,0.5332276015206353,
 0.8237703605629744,0.08822267772814485,0.13941687231519329,
 0.37010360797194075,0.14164834191862385,0.8816946153911399,0.997526976748946,
 0.9833781295016227,0.8287551866728035,0.6337972131461647,0.8548064771500746,
 0.6806419086098213,0.538956644006051,0.31239115919508265,0.43112824967676666,
 0.18344672932953554,0.7132544058698873,0.7819910741618764,0.7585414020580828,
 0.6668284194049755,0.2430912026408396,0.17403058133953753,0.90095817804326,
 0.6343044897332082,0.2956478913374354,0.41238839647076064,
 0.18114042286893584,0.10178758358889906,0.4902576315793943,
 0.7237644342977868,0.20808877676677962,0.9199414340248063,0.3774650749981032,
 0.6931979121345186,0.08602668532908564,0.2977535026456428,
 0.03942584038243968,0.8234590417150629,0.8143507309206414,0.8405647277980801,
 0.6234351780049262,0.17544430854142434,0.11963514757830551,
 0.20450304926941287,0.43628204491717193,0.9180405499027585,
 0.7675446561721374,0.5026078756010566,0.09990781183544817,0.9883440386682493,
 0.5900430342955849,0.4580992840389895,0.5536940793339649,0.5990503083643294,
 0.7904925207883473,0.322805014632519,0.0579546947240972,0.17460624168293892,
 0.3065185380523222,0.5120363502501881,0.9127181465995972,0.667834077134211,
 0.9247468069653099,0.500693783980922,0.991012343586481,0.8028308093451186,
 0.7490422483611646,0.8004095285026706,0.39607574956718894,
 0.03563392736083659,0.4259859696082132,0.647794355474834,0.1586453498452789,
 0.20930421672907196,0.22965984455070731,0.6331611625251252,
 0.7655948063282719,0.9497121577546119,0.42730536101313144,0.670549626795459,
 0.05245719640829949,0.7577953485351037,0.5747531959741654,0.7675009106494893,
 0.038195692468716236,0.8639534673345074,0.8947165354305162,
 0.5287433480818573,0.1618854876532908,0.004263363890184113,0.937117933184597,
 0.7086284765884213,0.6370109505363798,0.11359692539376787,0.5923995466363435,
 0.6643739750754394,0.800546753632404,0.33133456905882297,0.6716590304096886,
 0.6949178185292206,0.668023774749495,0.9030261497114668,0.8999278751250211,
 0.8438104420781491,0.6298865917535271,0.35752130994327114,0.4935589620627099,
 0.5410473314503067,0.36732658491568426,0.032803791003788274,
 0.3330595158237982,0.6693490856702304,0.46939289937153783,0.715640562965999,
 0.8812301230328382,0.7807544911710519,0.5968108420497877,0.9447600276882433,
 0.14880704128103794,0.6004403490280362,0.7099485000316579,
 0.23599820487093848,0.5021625056539515,0.9249452513150527,0.8416131238983441,
 0.9568643106831876,0.6128501945177366,0.12505515333201966,0.3506115376947506,
 0.16458748681086188,0.04923666575970587,0.38147047205248363,
 0.38607706802211017,0.29209815377681814,0.34735706129595323,
 0.2306003997257986,0.2366369340748996,0.49044494491661295,0.9140526354572167,
 0.06840745928738923,0.7548029046463328,0.4787492271622257,0.9391132769156979,
 0.3335760816525335,0.39341616039990723,0.6541228226331103,0.8880961665465044,
 0.6429059545233624,0.22443740850080995,0.8206789797371515,0.508989103140586,
 0.4994179710602543,0.10092733041729274,0.9741501581288572,
 0.12553639186903376,0.8989597995694255,0.10711651658078258,
 0.025576263452077863,0.5994861965588,0.7121411834486485,0.12616965627851218,
 0.3904316001986713,0.8616287582661932,0.3380758257886769,0.7667040427369302,
 0.3669641006405333,0.9095447279971095,0.16431525963737925,0.4450102469162721,
 0.9466286118632311,0.9743770520520092,0.41167305867187,0.15606802677176157,
 0.23143349499349886,0.3117392102903791,0.38536303161205665,
 0.022413073313670373,0.8046088922448617,0.1602038944972397,
 0.3414852936062208,0.11933322933897006,0.083144294297821,0.46099625242846787,
 0.3193492699932132,0.0635902192247203,0.34944696225709204,0.7822048391697316,
 0.44937370950321887,0.8465073732222499,0.9550762644615532,
 0.25413741034964976,0.3683968995867919,0.5512464242326931,
 0.08174062159427542,0.3611080019218005,0.5665664999793469,0.4437892721612319,
 0.9331900841335945,0.8092031294405326,0.7332626628401822,0.49656243007805523,
 0.47466671483803835,0.36623545679154534,0.34578966841805525,
 0.6931093547595726,0.4644189439700318,0.47130733212876774,
 0.27605633996982126,0.5334079545360993,0.13955012741321704,
 0.23664942506757902,0.8472789092482794,0.191972395740588,0.7464420658497325,
 0.02359059047673373,0.45768412322479557,0.6980524452294894,0.719708010382951,
 0.21492486683979772,0.17309439780556557,0.7803967476802645,
 0.3968285967385302,0.61180252920896,0.48313664016003555,0.6487103661525658,
 0.5971452202627331,0.9383645151312417,0.14744284780219485,0.922836102679595,
 0.19252161769529696,0.8870294354721547,0.5480034520002521,
 0.09212833664035369,0.3102758600719452,0.8987845038974802,0.7590882756033646,
 0.3385515613824801,0.297729319468352,0.1587044154343663,0.13442498325142516,
 0.529296383707503,0.09664981102291126,0.0018621000878215055,
 0.4513904529515015,0.6099254761287648,0.810897452848383,0.8102919447227667,
 0.6315048564837946,0.7507457836149767,0.3378587289670574,0.7811776750806578,
 0.17536959138892994,0.805880562561672,0.0545849407166622,0.7055718636605423,
 0.5248281483843378,0.6165729405316128,0.979044283285461,0.3541323512120199,
 0.9752007606865328,0.057789129861248845,0.6503941014641906,
 0.14896116431583895,0.19293201182459918,0.0635780188014392,
 0.7717618306615508,0.9608045722416746,0.45562875566090943,0.9862549557528828,
 0.576587393962861,0.8851630379616933,0.3218352129625698,0.7981224931139563,
 0.6721763180157685,0.22083175640575037,0.7468776765354371,0.6597746355649725,
 0.6508030921409811,0.860304973550954,0.9583630852367647,0.3846655975291543,
 0.35088275011174175,0.7641082089732366,0.4985848558670606,
 0.45134206475891603,0.07568503314992536,0.5008287686011011,
 0.3285146712907885,0.38916809011026554,0.07057845688808773,
 0.030319634721401423,0.7037697523301951,0.8766823080487242,
 0.7389097937564459,0.7051197756768814,0.07180514745698763,
 0.49788379968748386,0.7096916351066208,0.12200764084357718,
 0.6952014566774437,0.0010873718840324065,0.2084832887251339,
 0.904119053988266,0.6820633793219504,0.8511098667397644,0.9408022139523479,
 0.03654591475614055,0.34074366495451447,0.30306243413967116,
 0.9887074329806816,0.19961369852541955,0.1258523235721507,0.7162246045158758,
 0.12987226635741322,0.1695898802133915,0.6396904467809226,
 0.21920648573463553,0.8449198606939647,0.06526262538314709,
 0.5747682998643087,0.22967099141146174,0.1336661913955982,0.965666486539606,
 0.033296972965497496,0.9520169832616107,0.07926368209274626,
 0.7564943268887905,0.6396576803851707,0.24203228642361818,0.5226163032835708,
 0.8516690852181772,0.08565877771661345,0.5826553718392057,0.8471660292760225,
 0.9495135292067155,0.28194087475009566,0.7998453764894475,0.8210226415619039,
 0.3196107478371275,0.12375127040781975,0.3618048746089535,
 0.14576726848859656,0.6568901164207217,0.914139702404265,0.01236136160815926,
 0.24250847442615453,0.2743066545252991,0.6527889376671586,0.2894753168821642,
 0.011849903687770502,0.2840005967745879,0.5104834341611254,
 0.12079673018537695,0.8962609251036606,0.8397956847039398,0.9148144644474242,
 0.8832763933677324,0.3063881007692666,0.9158732537276527,0.6179301910676849,
 0.6799705365964401,0.9996328689532934,0.6692039438245334,
 0.056890598172579354,0.5891405330708979,0.5478732982959618,
 0.1063269921323714,0.19520437980688798,0.9290595417646531,0.4055492952026998,
 0.6474398017380429,0.6738410860862788,0.7858199403841463,0.7051699315284871,
 0.5086959854831713,0.9452617359422147,0.31576967959606317,0.5377969270360603,
 0.48980138344890245,0.26867954316492026,0.6747155073227721,
 0.5767719892735591,0.5245249838064503,0.995876860989838,0.26253061808047395,
 0.0035375403993949117,0.8893622451174528,0.9183010603908166,
 0.4842541025925706,0.20278964377820174,0.5882070175768085,
 0.25717511602789156,0.5376052746676155,0.07255736304919835,
 0.8226403382555783,0.48685796335475995,0.2963964320913087,0.848491153152426,
 0.30418559192909855,0.8316267325829811,0.8236248069032192,0.2522519476940692,
 0.450504930438828,0.30333796933738344,0.8461525105022742,0.8785972163608826,
 0.5564271979126393,0.6983396983076376,0.4345881380430108,0.22090454786998792,
 0.31212368030450455,0.515959232257085,0.6027105961100199,0.6964916033360008,
 0.572831583913973,0.10862607322104012,0.36928539628755885,0.5731282891626611,
 0.8276593538702371,0.052379435438239974,0.14915388641072458,
 0.8348772963242117,0.04514813298595388,0.8007376902568651,0.5121247262843334,
 0.37291670617021644,0.909880173749621,0.46711578955760547,
 0.43127998509146037,0.534828817019908,0.4838121141570688,0.9434221210051682,
 0.9513644838800116,0.5903982734811416,0.8656535858160386,0.8397520865773412,
 0.9501062378799993,0.13231882430422004,0.9429723300375399,
 0.35335771758203105,0.5923674281054081,0.06311777357090298,
 0.5983420107263715,0.7894603844184194,0.09001619250700887,0.9622014915117447,
 0.5931222294297573,0.493367452851297,0.050956369407281654,0.5793779804135064,
 0.2409781116850115,0.16535451189796202,0.28470198906643596,
 0.5955527943309291,0.12170685919900126,0.8015770548784902,0.3752189440571563,
 0.5813721301179197,0.2932998685910537,0.6312979360905573,0.7886450070158392,
 0.18102514998457064,0.799677755335679,0.008041340175923173,
 0.9022343179309841,0.28124353099198773,0.41927955915748216,
 0.37280629873126414,0.21599520197556288,0.6496118179255024,
 0.5507450157261888,0.6801190736863881,0.0075403761536576575,
 0.4878243097515451,0.5169222839426886,0.5197506507399632,0.9816949571622015,
 0.1749534628113082,0.6444057177678872,0.03330522807801939,0.6373292971317556,
 0.3163045240958354,0.49013554235324375,0.19449246454458313,
 0.6667625703580324,0.45952219888888757,0.9638711256047415,
 0.34710929946239044,0.4652847059683378,0.46952584744595116,
 0.4505071940507742,0.17531007983396107,0.39475705360650315,
 0.9348398188522324,0.5087279830376226,0.14304867365303375,0.9284506639033879,
 0.8172088395792503,0.4280500999849771,0.4238327262735848,0.30750297085522715,
 0.11742554649455295,0.5718850769224096,0.7487418864803549,
 0.07235838633432745,0.6837736454506733,0.8692276316690588,0.2192518948848452,
 0.3530155264198235,0.1168390373586059,0.2777833639402755,0.05155021110235947,
 0.11376588041001467,0.36248019794638997,0.8540566690452293,
 0.09774191391035181,0.6210926913971118,0.045679694611857946,
 0.8792453383204335,0.20197500236548327,0.5353612856348764,0.5682583925453739,
 0.13273211500383386,0.2241110796195206,0.43448374015955393,
 0.14240329532815765,0.4512523923658279,0.27987184330466097,
 0.20838384434124502,0.36716196254972655,0.9196007643976521,
 0.38520517924733244,0.9362449120325668,0.5947100814845457,0.6969577332320084,
 0.6326821185909292,0.22144135066126436,0.6135459667878739,
 0.14682992517300797,0.16085663791176696,0.14519580646153962,
 0.12459939937117825,0.9500656279734379,0.4402911214763172,0.4757442841923242,
 0.3160655007658123,0.8558508644556382,0.5198880798107335,0.20194499311195713,
 0.6278153800903918,0.5736632355016489,0.6754420153621447,0.4260864094033452,
 0.3160835551315242,0.9030173947343952,0.021974166981190146,
 0.9764446026466876,0.96797937466354,0.6471766478469867,0.8317160460686224,
 0.939125440727163,0.3392938045079784,0.09786810059409434,0.4799622116936969,
 0.21778784428630626,0.1743597253510819,0.3740865983786862,
 0.34814270170819506,0.7291054867994152,0.4127641266994162,0.2155576929323395,
 0.0014980290586731337,0.9671961656007977,0.010260732661567351,
 0.04627565985263837,0.36928985177048934,0.6070789502519707,
 0.3949072718378247,0.8229952112104395,0.15467776738526584,
 9.301368102989649e-4,0.8245650155570916,0.32871425528540854,
 0.31237413667803704,0.863062668788557,0.13440525740459786,
 0.34782950801632884,0.35998489170243075,0.8119594218746157,
 0.3186831743630364,0.4163960600044988,0.06409552800483642,0.8159399322172148,
 0.6227228693250649,0.362975212863204,0.628983168102849,0.954993163136561,
 0.8967252228281102,0.36215732266975265,0.3588275781886954,0.4325134247676664,
 0.0900510170756259,0.28415505099293015,0.2814115659336155,0.9367037197934549,
 0.7979235248056588,0.550981358647193,0.4196992482801548,0.35712240186572597,
 0.45197095025948053,0.05352528874043028,0.834880983700185,
 0.27569760507118757,0.17992490342904777,0.447547790525617,0.7523840924345788,
 0.34830393034857476,0.012957812810166058,0.28815462307627504,
 0.7115724719045708,0.9082392014643986,0.3732604089825855,0.8823510756330798,
 0.48829584158222405,0.46935112761971254,0.8942504412015947,
 0.38621442511218984,0.8391407651853795,0.2620832349721979,
 0.20169155180291587,0.49181742728356204,0.16925987857472657,
 0.8400558066639334,0.7238353570337279,0.8380151326414245,0.6599484936918756,
 0.8961668798780739,0.38208895330783477,0.6193843915981189,0.4274039103487005,
 0.7418399766069629,0.613127069583712,0.8244327302271984,0.4948338365864453,
 0.44185202421353476,0.6087218108041149,0.8895212326195041,
 0.03071890702400748,0.552371757038127,0.22092254834124614,
 0.038813167194066134,0.8238734268307543,0.48542597744160476,
 0.7882964090480481,0.6118118736911646,0.17398959382394386,0.2924585002088682,
 0.8737037149612651,0.021739995182106897,0.4079038019438256,
 0.17581436964639174,0.09224550439781809,0.6163929065274166,0.837281099445645,
 0.6597428041556155,0.9146442138185733,0.2864019971580892,0.46107595651682076,
 0.7873815759382645,0.4446757801739377,0.8504864413928356,0.2938901046218442,
 0.38211817343405285,0.7836925949345881,0.009629590292181689,
 0.4404028731051681,0.929196289859689,0.6175954695677798,0.6522019596553463,
 0.385814827569002,0.8551134892698697,0.12101351315221243,0.6688464290357088,
 0.6660412647303782,0.3689462294646687,0.25721736133018,0.563833650969919,
 0.8276692706585631,0.9723688378088996,0.6278695512522221,0.5034615232556998,
 0.44072471823745507,0.06606697441208076,0.6816582433553556,
 0.1782747920992439,0.51600720392542,0.3113514766618045,0.08482367742182484,
 0.13663610402482207,0.13384871205236837,0.14989840967750512,
 0.5706622879654276,0.8066262880063277,0.47638780129539904,0.4108903634741715,
 0.7491474413897541,0.8490809104297683,0.051483935646405854,
 0.8246823933640365,0.6317693103893477,0.15954759041811606,0.9985365083622543,
 0.26096575759996876,0.45274635543786645,0.8805554212500707,
 0.6913943390837803,0.797016861103567,0.752250296789355,0.27665370805111467,
 0.29643524586413483,0.20383643829167597,0.6986735319724746,
 0.4133792274875455,0.8193511445900392,0.545673452694303,0.3734585520597451,
 0.6918763824359324,0.34005023779615307,0.2292813869289927,0.8331459990032566,
 0.3305605233026784,0.43770681944706213,0.6432240840749046,0.6952565774461077,
 0.8222525189766949,0.9763101856725805,0.2100819036701782,0.6020252313230091,
 0.3615264834256756,0.9832281667904613,0.47183605794250516,0.0293923761033974,
 0.17368793300216157,0.32511484469849705,0.7983092752675414,
 0.22453401315741506,0.8310585790348266,0.35035928051392384,
 0.2885788310179992,0.6144106831370196,0.19290944870380755,
 0.29233158623780076,0.10258127974556341,0.32189679586300346,
 0.9694879569636118,0.25400858901256007,0.680339079615321,0.41530973770363755,
 0.7292918128709672,0.5801910021487613,0.8652433612708978,0.9963075658894673,
 0.19780193277357316,0.15385321866046653,0.275636902208515,
 0.10706982477427573,0.8748002492014462,0.748379623126233,0.520247687982295,
 0.5112067639809208,0.7342566247875884,0.30429560932904365,
 0.22487466781627452,0.984394067957148,0.03940603863621828,0.8077767403361911,
 0.17650946934303724,0.742039807324456,0.0201111246317478,0.24007845583129428,
 0.39067055967824016,0.9243089773729031,0.2646167118471703,0.2293165359311079,
 0.36463176300112,0.21695369552266686,0.5682574254135434,0.049928168770546066,
 0.030856968892922887,0.42625485254378337,0.8647446562674962,
 0.02484711810321405,0.016259803906296555,0.49948540991240764,
 0.15207050791360333,0.1679546106998494,0.6030033482679471,0.4994779402084133,
 0.22258266489187795,0.6210039710387866,0.011320082044907753,
 0.7901038535064212,0.5356848827707592,0.7989070619369869,0.5621124361582333,
 0.257712994233007,0.10162730484694615,0.15234762818017344,0.8426038742503459,
 0.2491827411675268,0.9267338265541148,0.7315394677942381,0.8980098188802694,
 0.3229407298589637,0.7821434776899476,0.7232784155063886,0.20011384593455794,
 0.43324464377705996,0.8353275645990542,0.43179615875801547,0.296216708418486,
 0.43292602552240655,0.27797261031600606,0.8242444433899947,
 0.6200279467021256,0.7964007826748324,0.4118768404547066,0.9149096288665989,
 0.5698545815589661,0.4610263414392989,0.07774136205985738,0.4231543261126831,
 0.2214726333945407,0.726868586452359,0.8243659178051237,0.05147515430083216,
 0.9959624716813028,0.008639687974055454,0.5245389013199894,
 0.04432635408844732,0.9724840642291792,0.5482556468559119,0.891061651441517,
 0.9574395293214613,0.2826623630077887,0.9864014565415835,0.7891400135013807,
 0.9006726697314478,0.09048878291413665,0.5579294427826383,0.6228913341911978,
 0.38363607907802744,0.9420686184908104,0.3709199525015825,0.6308937924679625,
 0.8835530605931191,0.0954521442860925,0.9151599984622825,0.6234661669497099,
 0.5186866907405601,0.7225182251003871,0.06547028353650719,
 0.10788104610637617,0.4872293545912927,0.9426962541510321,0.5091371067960291,
 0.2439375502441723,0.5525742250534523,0.17585880975927926,0.4526935238262836,
 0.29708554641719254,0.17347072345758996,0.4524852728663109,
 0.9709189635211227,0.9558333182430004,0.010321512632328167,
 0.23714744455143677,0.09540686317409297,0.5964552852534268,0.846143287336524,
 0.27620035404286913,0.1130334745807342,0.6994838946568244,
 0.33861194543457485,0.1153883115824087,0.3321122075987083,0.4467700399049833,
 0.32151218745227306,0.761276823558773,0.8755197724439496,0.02428773995530953,
 0.5884178635081142,0.0018598883583952341,0.8917822536301484,
 0.33251485845425577,0.2635461917072556,0.8616797721709379,
 0.09725483696491777,0.8488593790082608,0.14654886654661003,
 0.23593757510746716,0.3731362773269302,0.8721467589756862,0.6545325183052306,
 0.34478042259937225,0.9805771102297162,0.7330119762942942,0.9999645785565362,
 0.3961569885277896,0.2424587516943656,0.7170399272550184,0.5941821867436435,
 0.8649959204262311,0.5037507973565742,0.04529142015872756,0.8395829225436132,
 0.7084899428406969,0.6052166162841831,0.031154336160373886,
 0.47002533780071354,0.640354984548066,0.8830128022459706,0.21364985121863433,
 0.5939963222258713,0.34821998027331214,0.16101278639349859,0.646980068093904,
 0.3112302354980714,0.9759242519189503,0.6421703473134579,0.3206642229371899,
 0.4233251297287439,0.5196992255837262,0.10983731160871235,
 0.10328536637668606,0.5364773634185649,0.623639774905964,0.287749346266568,
 0.20943333423889887,0.3646710657749669,0.1219821696704797,0.5056648827011145,
 0.40337405145935257,0.7424029616119283,0.7921462269599046,0.741688064753935,
 0.15843875163166865,0.4935351024537611,0.5988758304893796,
 0.36097135927235724,0.6008294184601484,0.6612449024865866,0.7391174262123577,
 0.17149091732228716,0.9050714248125591,0.32264851884370316,
 0.1498219164424872,0.8350315074532653,0.5244966385246752,0.10220722611443478,
 0.400762022879839,0.67802652418048,0.7707768078498817,0.7693539738031954,
 0.057395676292793074,0.07680068949916563,0.9363529965266942,
 0.5605313558326365,0.09734399349621403,0.5293284073247057,0.7320928594174118,
 0.6493241377827624,0.07361590383827132,0.025598660764645564,
 0.8235223947710338,0.2581228514859858,0.7044947363472926,0.3972948828379528,
 0.007731115972653857,0.3820391326803533,0.11560648478234747,
 0.9708738398791706,0.23608103791553647,0.15909457245110914,0.730947725943224,
 0.06549878350044469,0.39459469171036876,0.7775514422848204,
 0.001126970059032062,0.3566246849871738,0.19871708161353074,
 0.4587930207269192,0.2215941287060057,0.5494660687216734,0.7149632474607704,
 0.6637867563237576,0.06425819912314279,0.228089657678744,0.24601720302473573,
 0.21925678438409013,0.4155197391817228,0.7224505027033394,
 0.08900416744990136,0.24608996931140215,0.5213389552628579,
 0.7564258100643575,0.23118536947492752,0.7642315876918432,0.9964493828339485,
 0.01906188997832714,0.37472031230328895,0.32515418253467,0.1224826962585428,
 0.809221658169446,0.40124397930046307,0.3275720922218368,0.2542590068179966,
 0.2464404845734336,0.8277834448295918,0.2927341336966709,0.8729278063884104,
 0.4510679624952737,0.5756531195507566,0.5710109547855392,0.10139102787251353,
 0.41084763862864937,0.6866143161005924,0.686638339365916,0.21833082897076272,
 0.15410685890948672,0.5660483337307723,0.5836378098390116,0.3856134638609484,
 0.7610131835019851,0.016822494519089015,0.3301929165445395,
 0.4488779197642885,0.35096520854377955,0.2441299638627844,0.7186488547656645,
 0.39704525779062216,0.9034529952402403,0.46792613037521247,
 0.6670123941968595,0.42822017275722646,0.5852330945805129,
 0.19936695762260748,0.09744441048455021,0.07044829692474341,
 0.47155871498761515,0.08297928138235089,0.9092406817571,0.17655551115197543,
 0.9488090092770748,0.35599548480388254,0.798244043914854,0.7582650258743341,
 0.12451287255179921,0.7726842684983686,0.5007219052727818,0.6694472226154035,
 0.0247072262721727,0.5940331401285278,0.9090425632949354,0.8337875168291025,
 0.11637756020955425,0.480372695298141,0.5122784523211975,0.9196605430341428,
 0.24129247515849173,0.8284446020877452,0.6251265587939339,0.7942336161195702,
 0.37333203487538724,0.7100331903798534,0.5601063451332351,
 0.029832534592227122,0.7022064632274869,0.19806641301818828,
 0.31291278665834044,0.07537199391637661,0.5988471377433476,
 0.1639294248534876,0.8158685022312702,0.4296698746362567,0.6607209739758985,
 0.7030752121583285,0.2652102774674001,0.4223694647980507,0.7730071961757619,
 0.7695679825388864,0.6619771980495934,0.98716435425751,0.4085445422155046,
 0.7634991046285244,0.6197497405235702,0.04713293362497839,0.6031775170576549,
 0.009394633872996772,0.8445379445636508,0.25240905224344024,
 0.2600681428823981,0.626580495081417,0.31210829397651174,0.8167544417018201,
 0.7704491887659899,0.048228048178998506,0.2984518642495214,
 0.5341748735674609,0.04874214175640146,0.5650774285930855,
 0.10264731142263006,0.4156738204026117,0.980343385232866,0.12618863212749065,
 0.772522374632154,0.3572405667817077,0.31454016157718934,0.5245390133549855,
 0.7563561020882436,0.47999220887372773,0.7983339027003739,0.4239618115568554,
 0.9044863274111445,0.9549007555173687,0.01983858198072974,0.8836092108106447,
 0.1320961613429058,0.3308994580430218,0.4631118661521588,0.3150700553389497,
 0.6443410236763943,0.8904787245922571,0.8963239485813338,0.0909351629339421,
 0.16245437467532442,0.31977531199978426,0.5571990014841277,
 0.5849853620360778,0.27603145533957174,0.5925738833771123,
 0.12376767815997236,0.1211690249216264,0.23231070298655643,
 0.16636285512657156,0.8915815475775442,0.2812214040133312,0.275128997575829,
 0.3082564565271567,0.5826777900805701,0.7027308944554039,0.6746305285106295,
 0.6180000615720731,0.6603377688922875,0.17329792809732814,
 0.47835479492114985,0.4585577967146802,0.1259752710111388,0.6780193050395076,
 0.9808364553132861,0.2198362888137998,0.8263997864528961,0.6785381205462923,
 0.4740421744003762,0.2163648947745236,0.6138276738564713,0.9759615755834652,
 0.8632422882800319,0.2440429439795555,0.8173228033404595,0.34173562689316805,
 0.32342559738431265,0.334415457246793,0.5805118914243173,0.2828297603833766,
 0.4438665184265257,0.7028587580570151,0.2211263355136449,0.5292482870330875,
 0.8724368367450828,0.1109363154356875,0.8040899284855085,0.8438141005314583,
 0.5121643109223994,0.9624362052883353,0.16514559807965035,
 0.18928407459497776,0.6065740004864277,0.6243447639062398,0.9867166763071927,
 0.25214762100139865,0.32357234096705745,0.29604167344283816,
 0.676235442837792,0.6707099454372559,0.7012029857011267,0.7267680707297604,
 0.8386173629153604,0.7437240157451989,0.8889967619876289,0.8367776073755397,
 0.3850237441205222,0.8486417105977424,0.5299402414366303,0.3814737864319643,
 0.7472247839456075,0.6739228056999299,0.33672944245568137,0.5980342319410337,
 0.30601709845201064,0.49997643594139207,0.7438971869523456,
 0.4244302024515696,0.8245426494089124,0.29914425410212786,0.887443554756043,
 0.5240297649203991,0.5754151733702643,0.6458700429237842,0.373624165431351,
 0.7951863733914261,0.754849051071591,0.10582987522590803,0.6338400856525561,
 0.3953037145871521,0.816765083387344,0.8643677316462394,0.008414847607439047,
 0.5546905769120442,0.7657510042761023,0.6601227926838887,0.8630265912234747,
 0.7881220759298257,0.5720685749760278,0.5969800309752876,0.855459577653023,
 0.07917049527379683,0.026708514761108693,0.7128313727971697,
 0.5010373216468139,0.7289896308733829,0.7960588036901993,0.6317321578917836,
 0.19471325606552004,0.29376759351166015,0.12268923865985804,
 0.7665318691039248,0.009195210378657115,0.2377167742298012,0.645746083344764,
 0.7317464779481845,0.9841995529248855,0.29012050494378505,0.7170023804486951,
 0.5957281811683178,0.5535554648772444,0.3703057448736562,0.5068100860735252,
 0.870570718657849,0.30060609004559613,0.8674334868572037,0.09659866383551818,
 0.29099588587314473,0.9570662422735207,0.8082382652073586,0.913862510406798,
 0.3599653790497893,0.20756861748220679,0.2784613121260666,0.6780773424687082,
 0.6202534957069367,0.4251780638701945,0.11948929252455354,
 0.34029299127472024,0.6459783680757787,0.37719820727002673,
 0.47594405190160716,0.29558099329765575,0.15868418328204126,
 0.45259908485691036,0.7549168660005597,0.46737409840910715,
 0.15755164066182592,0.8752671655350702,0.5247656792749957,0.1411855976807319,
 0.2082656447629242,0.280150467339942,0.18767718050643267,0.4359121213507,
 0.6379065043490644,0.005600577707871368,0.05026349961878274,
 0.7614434076659341,0.6809529013852466,0.0228242591228085,0.04442032847163202,
 0.2950497410725279,0.2374093411916941,0.12341436750092971,
 0.34573422171150525,0.5851344046618366,0.17801397630396698,
 0.1076883967150345,0.6306980154110721,0.18048020417185928,0.3381439919781458,
 0.5259962360299207,0.5196963742270362,0.7791665258880692,
 0.012144371963781597,0.5888606932031606,0.3171459810012959,
 0.5540563516620093,0.554688818663218,0.810212599258932,0.2991027345096613,
 0.21511662288213262,0.582829271459269,0.368332878183092,0.7227961032426049,
 0.8104830711436787,0.1708383184308021,0.5978677477879608,0.17716317282614047,
 0.5487804914761839,0.28545012733903863,0.11765022930855773,
 0.5005759754973933,0.31412413988249377,0.9433838929516262,0.0438856584246875,
 0.693282245750099,0.02270288075575111,0.8427188031396025,0.6102226582656017,
 0.9900421325077717,0.26365126215114976,0.6064624534566116,0.3422491010550912,
 0.9976882665814162,0.35969557256357065,0.731073278216618,0.2814279118523171,
 0.58118950046612,0.24416676013631677,0.24774232748639158,0.13318835846003596,
 0.7920007710390106,0.21377198141094755,0.4670801743857228,0.2790337552523817,
 0.18627375908842891,0.3224474553901122,0.32370495903927043,
 0.3797914453693628,0.23574502437413591,0.3184876918676781,0.8302111677065845,
 0.8529675264165731,0.7754459467315697,0.6957941451054599,0.41994782796219565,
 0.8748028627496196,0.8014762033078896,0.13208902006402523,0.5317633970099636,
 0.9309646878316039,0.8481223491845474,0.11979716171840193,
 0.17828531417042937,0.9611425489799713,0.1493488342096726,
 0.005220819570575519,0.06216632588955273,0.6046375153384327,
 0.10512113526630562,0.5295555324171388,0.9766017811734653,0.9401514980922858,
 0.9230150623189797,0.5937795850445332,0.7099596629389007,0.8203104904803975,
 0.9914980725473009,0.8919530336219237,0.19117418343163384,
 0.25515263315936776,0.6241988259217776,0.7097263148079697,
 0.26318574088443436,0.6311004623966673,0.05635783636668146,
 0.3325593798129307,0.8042288977307785,0.3001545030173163,0.5402117502072514,
 0.7117614993227324,0.5910501731046646,0.49162154233200706,0.4696840038069323,
 0.5158614630558334,0.514254052292616,0.6498613320906916,0.9399954941223423,
 0.7044760027963766,0.48149151601005136,0.9424644665856584,0.70083919867999,
 0.25497164460733424,0.6965102709921985,0.7564836810737732,
 0.24256853891950514,0.242400632914876,0.7964634802154318,0.7578550027245938,
 0.796272108263365,0.881835056990456,0.6022093029840786,0.8025920252207724,
 0.09010901482025024,0.21679134722551263,0.36562101896369437,
 0.0725171005964973,0.4617237015859411,0.2653418967279687,0.9591650252825754,
 0.8269721541866194,0.3477165907451538,0.391779529139042,0.26555351537269,
 0.8759561522046235,0.6254513539833892,0.24843429794570915,0.8299986916790378,
 0.40063087027369726,0.8625309913820963,0.8953741785076174,0.7739194506218827,
 0.7229406731395843,0.18476892883817553,0.05596926777836575,
 0.6540054047058508,0.859623637882175,0.782003547903793,0.2259947523506437,
 0.21456905412365213,0.6634888895808457,0.3816099862551129,
 0.21352539290514172,0.3810385847010713,0.1527469970268538,0.2790898788134566,
 0.9720576269396969,0.3461035739825665,0.052134753020891766,
 0.00842968351537321,0.5244521845015604,0.5567862561092058,
 0.16669961410727496,0.27549722681096434,0.15914033490681034,
 0.6545475740313937,0.7876498220006124,0.31124541216477386,0.4703427611074342,
 0.4006356613004085,0.44301896038586686,0.26877983594308097,
 0.11465881069791195,0.4332895148248057,0.4570433599323621,
 0.22593549662857826,0.6618331481184769,0.7713093563732523,0.8024231496032602,
 0.8502918184213906,0.002892456369894769,0.30955163137003927,
 0.8812232469434216,0.8467463658736134,0.40058092986214067,
 0.01836694992368615,0.5474830058771373,0.3977621913688675,
 0.39939341076927126,0.34041321514605605,0.14452812771342938,
 0.7862388407461774,0.0953185346782427,0.9055869578803114,0.32705707628399017,
 0.5050791632828696,0.7458567257445544,0.2610078192085994,0.03994896850011409,
 0.5196119856317837,0.2608699740990118,0.3409469841787942,0.30299413203156034,
 0.6147608224800086,0.22751947942755246,0.04241911732296022,
 0.3899428910453402,0.6699937277780826,0.5763809851847732,0.8060693277627744,
 0.44400906916595084,0.564125878179466,0.4350456973081853,0.7477768644540507,
 0.17215708961369003,0.2696112439582755,0.5428192557217767,
 0.06666218210694819,0.3471107595564016,0.5626713328093613,0.211653277072388,
 0.8851194022689673,0.6968327883325324,0.3165842159655142,0.9699377701092773,
 0.07042712339872126,0.9227179990180706,0.6726420115242967,0.0646057508352107,
 0.6552300962169578,0.44807125687142113,0.9813753023703494,0.6931308068537774,
 0.7026871548582533,0.15378308263720974,0.8264476785424136,0.4342780363450566,
 0.8768023049051781,0.09368224957377058,0.9650727763459688,
 0.31446030178491147,0.26833105494904486,0.914975593451179,0.9952153300395683,
 0.18979313536803744,0.8994992971120754,0.5724523075288608,0.2161760439712388,
 0.6834109685979679,0.011770547262586195,0.9656383815798465,
 0.8810976831472759,0.1165550166845889,0.620924638488499,0.469986098868699,
 0.6095644428315721,0.8853594057439357,0.11292455255442113,0.3128307357395397,
 0.5710295340291176,0.5739600797257429,0.10183732971049708,0.1469800304582558,
 0.3744139224411692,0.0013764100233000232,0.2639444996281536,
 0.6254836350194198,0.44584867555657337,0.5676693176212806,
 0.20000563110365532,0.11833388062211125,0.13487415638019273,
 0.629218709225483,0.3448314221166373,0.25301328332732775,0.43824880799948884,
 0.3381993942581165,0.26290013223686215,0.12789932460166065,
 0.07835148792938307,0.3617590311544372,0.5093498604376199,0.6475663396229303,
 0.32053404200847224,0.21853774153191619,0.10042569899801268,
 0.3888371229880916,0.9525830478306675,0.6524624697552152,0.4084853502347293,
 0.8815548426174176,0.061579923893664956,0.4229906637999956,0.346614185170219,
 0.3949654546396706,0.1327968198690639,0.12031063499451589,
 0.23198378705390743,0.6296944176885623,0.7325423337276824,0.9537211420088182,
 0.28588256485638885,0.874241624091369,0.7535133806038409,0.2833004097076418,
 0.0435964180257189,0.24146982727212285,0.4719572428828631,0.9907409582588563,
 0.3367047308987534,0.6109202184124946,0.38899151090068174,0.5759064996147607,
 0.9298710561326519,0.25378667995198545,0.5736264867709964,0.5841254881307831,
 0.4064502852405456,0.7425759747660072,0.5130030152369462,0.06033509716248475,
 0.22385414319414687,0.12718019327956231,0.05442546843975804,
 0.5878281069308985,0.5581821620532472,0.1419442136506741,0.693517597667396,
 0.3343082615687414,0.7877097712118971,0.5852456411186719,0.188173822348488,
 0.2665147638735651,0.7151320405723424,0.04907094564731196,0.9949380372215889,
 0.032366638383903634,0.2419027792701034,0.5267439180716706,
 0.6583768690885309,0.04809862004751175,0.1390031545706506,0.0669035178484152,
 0.019692384349323078,0.426590926895417,0.013124770282827436,
 0.9987038642057519,0.5649901475691205,0.8771149264323145,0.07506733053234238,
 0.21633967005767762,0.6300789933118385,0.4621383247722961,0.9942997681220956,
 0.7384963590558095,0.4586917322701187,0.5280283553591029,0.11368185181493806,
 0.7907894002126497,0.4197472809806435,0.853301271638714,0.7649021401184988,
 0.5276809301684984,0.5425037645549081,0.6622127738674269,0.8841485956230971,
 0.7001415158835991,0.8238260407673924,0.29853468904779934,
 0.06720866381171331,0.6026774451202317,0.7664047287812693,0.9740905759991356,
 0.36197173590202436,0.31196142846307573,0.06648521205167768,
 0.5607813532045952,0.8578283829227581,0.5077942542485453,0.19892759017566164,
 0.3498215011985091,0.19525676403199355,0.90181094125188,0.5805692759275968,
 0.3412055970546952,0.33698404133228144,0.10225426131212967,
 0.018089053373850805,0.6292601396011045,0.6968218712510739,
 0.6060877114948852,0.24682463953457223,0.4683601892941175,0.7110254124047778,
 0.26463334050724585,0.4867317265302945,0.3669465249562285,
 0.27427145581017653,0.6635640224764918,0.3648274587698619,0.6499671177274903,
 0.17728374928300394,0.4141574706069131,0.3773469154917002,0.7932680305174667,
 0.5604845336151982,0.7445664780648374,0.28019754449679957,
 0.18011918985248265,0.8821186828095482,0.3005439075914582,
 0.15745440267531063,0.9469879104308317,0.40131367737796,0.6147638541323405,
 0.42656898237086693,0.9279787164180237,0.07107141253735527,
 0.6188838448084817,0.4424679220808436,0.648893594062093,0.4695436623269409,
 0.663939475708862,0.32292550957189636,0.4532310607456026,0.9944873525278504,
 0.04660615697460668,0.1990608974407886,0.6665639469155513,
 0.10424210997476102,0.6311020839301358,0.192150319656901,0.8025492840003582,
 0.735167888285175,0.90836196121926,0.5065091438956113,0.3914028405557861,
 0.12676442959847167,0.9651211071756816,0.8525472709801407,0.3154457574060908,
 0.5419946812298323,0.5684296007892051,0.9440005533361133,0.20274161146712122,
 0.1587820102288957,0.3865952850204837,0.9351034154382161,0.8519952399181534,
 0.3129925881282548,0.07517994458421595,0.3951116319758624,
 0.019229888630898917,0.22676024996464728,0.9989557344583149,
 0.053121869057640936,0.6326262339290771,0.3020316229678003,
 0.6685207249242369,0.0393646593484972,0.5415436052847289,0.8739250193355645,
 0.5950901042066727,0.36642028349490885,0.4377361505674259,0.7805583733805475,
 0.11719315793329566,0.5418011510275035,0.32563405001326085,
 0.5516386015568876,0.4373726127891542,0.19144907844155334,0.8834091689882304,
 0.1297903929805988,0.44309209628545154,0.7892615610988264,0.6790050640623522,
 0.9675968829500743,0.16630801354255353,0.06903765250065308,
 0.6320350179624379,0.17976219232844282,0.8521412403705843,0.2121792754068128,
 0.5699883923541392,0.8172658662532386,0.9772244663300182,0.18392561879035163,
 0.15084609822882467,0.7255150087747454,0.16508110946438903,
 0.7887517291570382,0.17478024468741427,0.11656200440467546,
 0.7041021692982817,0.4574524797348841,0.3333945171136482,0.9282996602916177,
 0.14823262480485733,0.8871629174600131,0.08900676964353571,
 0.02948737283239644,0.11980437835826041,0.5862220605853583,
 0.38744947663782614,0.6841699757999874,0.7656399887487988,0.2515398754949769,
 0.9305445866470778,0.026094511868483328,0.9755708254569395,0.328532521415817,
 0.44106724410703846,0.8991067149902672,0.6870039070057307,0.4257010220608479,
 0.3207087107059676,0.8271523063616971,0.5957656753587071,0.5058630794491713,
 0.8633762966237684,0.5357474423664569,0.3868210134201133,0.7669174992359696,
 0.09044995054546712,0.14768884409940042,0.05160652346267325,
 0.46320960220921215,0.14657828663111594,0.0849681791292004,
 0.14960850400136727,0.49262558223696995,0.2862295245215286,
 0.3340810192837924,0.1653125835786997,0.7955175088634313,0.4111772565450138,
 0.8249066674909623,0.822856899148574,0.21624596965085574,0.32705480544081267,
 0.7166974446531665,0.6967627063659056,0.4247543535642222,0.6799012518241464,
 0.7528583973951691,0.46068566007643463,0.4230282306763976,0.8416093686321308,
 0.6732622278362743,0.6354539447429721,0.2464197755480313,0.41856121790165357,
 0.9119364218494165,0.17670101387484338,0.11158439524716346,
 0.8504436271276244,0.25801840466017756,0.21477056926634885,
 0.08384507597713087,0.2878141862734047,0.8480673062826606,0.8691260295854215,
 0.9132899475870472,0.2817422585541022,0.9917698125052182,
 0.018695394458906445,0.7849262740190089,0.5016084878722463,
 0.18995846301382335,0.0352018886195794,0.6387450780537993,
 0.41379127354858625,0.8770745221068182,0.7628610430704404,0.8818333767941245,
 0.590512939788184,0.7966640929180655,0.44892781298734585,0.9244403698431056,
 0.778362084913676,0.11814398742604482,0.6442413660072459,0.8462802996540127,
 0.04821100061158834,0.05572439102789628,0.8458425025202657,
 0.11831526931628655,0.01801392858426709,0.6262755830789649,
 0.5484326511082563,0.28643053450260314,0.8053161685623356,0.8915861568212164,
 0.7698583567541166,0.07337332079819636,0.22258391452523485,
 0.05387182501036902,0.7775309637527781,0.3915958220287047,
 0.21950389058150643,0.4655178645236584,0.8513208810457665,0.50176001466312,
 0.08037270559673881,0.7948002310122408,0.5398738893964309,0.7402618784953762,
 0.7906665275296585,0.7867678964340756,0.8960295434129928,0.783732905647753,
 0.37988287290997236,0.3468000044620827,0.4964672805461585,0.8781695317415855,
 0.8030523138233393,0.7650198152404316,0.12896356841439383,0.790180023559537,
 0.49740834972932313,0.0408052441187563,0.8148452213169668,
 0.18952089674547157,0.38359694196924743,0.5858744282830315,
 0.5480342509575354,0.39557929650612134,0.2005396941205665,0.7570383380469271,
 0.9091141475575748,0.25698105445999775,0.30225445686594865,
 0.6526110617936617,0.790373477486037,0.4909310961690475,0.46630847423453226,
 0.839034614921989,0.0414453311822337,0.5726965880573327,0.4491467452241428,
 0.6325119351232168,0.9902390388061576,0.537011464965788,0.1596791972855952,
 0.3518845207330359,0.1537554098707581,0.8735083601776223,0.8751531236928126,
 0.8438924592266738,0.2887376371057051,0.9029839794813725,0.529507017590308,
 0.7509831238151955,0.27910979798798985,0.2816103329871734,
 0.20906960995287482,0.3015729939110885,0.4113304193759644,
 0.40790032366750084,0.7792094612839064,0.6357408476417971,0.3436183357511413,
 0.8724618285622814,0.9741110339504864,0.9587480972666137,0.24103792400482604,
 0.011387322220765106,0.6861934431724999,0.6395609731713157,
 0.4728835782306101,0.5413584752137296,0.5597175711959811,0.4954873881176375,
 0.36629866657391463,0.30804357904005963,0.6528715299797476,
 0.9590307951766784,0.3247637688946726,0.6904420150099072,0.6014160101655502,
 0.8818513216102933,0.6194978558642215,0.9291999037921803,0.06848766086715397,
 0.9300955326160052,0.9589657055727496,0.558156007525473,0.15035390497036882,
 0.4133815865431836,0.7610187987976272,0.7675861305036978,0.14781191656383152,
 0.2066962779493665,0.7028451779672026,0.061436760580461436,
 0.00965165260430112,0.1954934439764625,0.9770857876621025,0.331488781214789,
 0.05780230223784666,0.732281243862435,0.7650888253545532,0.38764899938445385,
 0.7221007760476524,0.7191461254517828,0.7462669934061713,0.35847550004671125,
 0.3017917910417447,0.1422133316519567,0.6696401931133033,0.6308801538845535,
 0.2806703529666722,0.4064400615984194,0.8515692463011134,0.884806283514886,
 0.7695054522994157,0.3831192127490909,0.1732135020912411,0.3717483518282837,
 0.8539808829773281,0.015848482075763948,0.7820673032365484,
 0.34675064269119726,0.7581746882971715,0.3992220257607111,
 0.16581767761270605,0.5145662800407934,0.9952398339158215,0.8140319201318513,
 0.7247067070243627,0.5915325996323871,0.09240669183071415,
 0.06710588779104298,0.09953866860338134,0.1752876873837984,
 0.8912031019700779,0.4069027656364951,0.6334712709048982,0.6879323060773672,
 0.6173231820035084,0.2630200839176361,0.8750649136620163,0.39633581425606035,
 0.22158681909915645,0.43978567281761594,0.38199332432867894,0.41047735343019,
 0.214102043406051,0.5955424233021726,0.21504157363334664,0.4549834525700769,
 0.42168176055548545,0.620329794438171,0.1542251769051438,0.89123514765352,
 0.860593369759063,0.5646121557470348,0.8651685104610075,0.9022540837285832,
 0.7521283602528265,0.6657320031763341,0.9472575802121942,0.3118797110161543,
 0.3855768200444557,0.4343423011633706,0.41736057381282154,0.4389650751281663,
 0.30126247779757676,0.3899042795587523,0.25845341611055406,
 0.9531589786887796,0.4729333047327753,0.3518221820452073,0.7422784927368525,
 0.8037324221799246,0.8660239931475247,0.2194988671637219,0.9865941391094315,
 0.9885998628147047,0.4873282304081641,0.7943277785743663,0.42673402314368647,
 0.856905184113263,0.31870126573219426,0.06396194495000618,
 0.26400415733066696,0.5226844529810231,0.6810750500424926,
 0.04458947133106661,0.9551110112983623,0.053233997427151625,
 0.4587820802853557,0.12649280845054833,0.06994675636166259,
 0.7246041947152824,0.9847226130403092,0.038140816889377316,
 0.13167052280241898,0.6011496495719912,0.05500643971431318,
 0.45436342106722294,0.11980611910387617,0.6184280140756483,
 0.36407528176767634,0.2956347627613086,0.3285222836064251,0.8484859729393344,
 0.11466838546876024,0.8302928670543591,0.6824013000222816,0.2980334127607448,
 0.40722032874777947,0.0052329022066294195,0.7952213374112375,
 0.2729726763272263,0.04885900528786724,0.4774305787530301,0.8492626075482141,
 0.7567390326772648,0.193301548657836,0.5307497040805996,0.0898461141013589,
 0.5551431556074982,0.3575290640799724,0.4251472149967015,0.6419161114983858,
 0.35559373180144793,0.3485918012735201,0.15664585116094465,
 0.4233109496649007,0.4727085666892037,0.9593146188535449,0.32856951912586463,
 0.8533603347073624,0.5265822910758337,0.9087350088197819,0.977458207409676,
 0.9687466569752472,0.08372900866554933,0.6903016406261562,0.9497786648373285,
 0.7153233346410071,0.9395533055367461,0.7472271303502056,0.7185575197811872,
 0.9675932019777251,0.4139666527967777,0.1826957466527932,0.8479302275535071,
 0.3621507622841025,0.6304088886097727,0.16529414771315576,0.7770458593348805,
 0.6012390934481591,0.7788205570168198,0.4449476034089541,0.6861963934940262,
 0.5379923278871541,0.8099176034135108,0.4111809940497899,0.23843250676100036,
 0.8364405783825473,0.07506776705291152,0.07246066091466097,
 0.5423657459215052,0.15914765924620644,0.005387698836582078,
 0.7886008116929439,0.6030478827122235,0.03282952463454303,0.5133133361473489,
 0.239940989958505,0.8271445145947879,0.5690465110419083,0.5741549840094555,
 0.9047902273255057,0.6307776717505381,0.9700300401165193,0.8238135355905591,
 0.12594788894407305,0.26572269678537896,0.48370816380961246,
 0.17445049313908756,0.7500112429439889,0.958024635041683,0.6273952982482726,
 0.8704791097619152,0.5897161219711071,0.5750898401824965,0.18579292155266913,
 0.886380907616451,0.42635939427290304,0.3973711127891142,0.09934901146071562,
 0.8545765261213059,0.13097778095367496,0.19719478113257394,
 0.5276094721407631,0.34170673644911886,0.39121318925317283,
 0.07740186956300155,0.6886783894128458,0.31594475970295777,
 0.8396134504238217,0.6025219962126034,0.05638650820149038,
 0.46044206597487536,0.5740438493801906,0.4073360354267278,0.4309858940208968,
 0.8440584073824392,0.112424845693885,0.8005752690786582,0.8922880376054645,
 0.15641250412965935,0.679437546568132,0.2778466947701833,0.3562808031033572,
 0.802592157897046,0.8748970344978457,0.03073493473026867,0.932334497711474,
 0.679033461244137,0.6625107795790818,0.604170658266821,0.14630041897896584,
 0.1653459933920517,0.46047310012226395,0.3636560317951252,
 0.040334146982449304,0.375686799985825,0.7993741181905025,
 0.06346253919137823,0.5678791656183585,0.6983895192671246,
 0.03602568095499037,0.48161075979251033,0.8273680545701039,
 0.16879158851114706,0.6216847850353884,0.4988264645082846,0.6028846081639974,
 0.9908553081334233,0.9384675421183835,0.33326233315285136,0.7163793652942927,
 0.11776171258966528,0.5336399332997287,0.4718264026771315,0.6185734627732933,
 0.7036998126104752,0.4800890430857171,0.14071595672505477,0.5408739406632748,
 0.25627681853488304,0.18627304098212205,0.47861644204627973,
 0.3785490780146371,0.7662238739997507,0.07479796318502163,
 0.04513148711012205,0.9851869808203602,0.8029702896640251,0.9186662611564004,
 0.2919773873292575,0.7027326080324767,0.11121904745604616,0.6458349334696585,
 0.017871703347314027,0.16141901779982182,0.46780039131368933,
 0.43057497606403106,0.6099712833015621,0.10784917148279438,
 0.03333798196190585,0.40782065768380027,0.6592037612999065,
 0.8803299502773763,0.6731165038500598,0.684832044343629,0.7410881414052581,
 0.46839992405397546,0.9882942500261693,0.3356929699431366,0.1783269880914803,
 0.2887801390070306,0.21188387890318783,0.8808349183545148,0.4540525508935166,
 0.7259985434166178,0.08010024512164837,0.8675380818342031,0.9386728261442614,
 0.7065356176041296,0.31122530307065543,0.36674915264563224,
 0.8121574996645524,0.08365065463699461,0.01273431158977767,
 0.11814408771705942,0.00581418543633605,0.17440625613847038,
 0.9714952636656828,0.42417643849352304,0.4900621965649685,0.9386040455637299,
 0.07477198396908113,0.07144274105447312,0.15748143320626473,
 0.1198348910499969,0.94395241377321,0.9953408717919136,0.9721346698339898,
 0.8338523864824082,0.938564981334911,0.6910823308942377,0.8955626209845174,
 0.5714377816947601,0.6325865619713674,0.3488781297271295,0.6830264642178653,
 0.66265916737246,0.7312813131363366,0.20325923712428584,0.7475089494998103,
 0.43978187879472264,0.8725129692368115,0.012027218039472043,
 0.5562730848721749,0.42234461826878034,0.06624457885191515,
 0.4080974345775489,0.08913686436392632,0.6973282114499361,0.5922214400847462,
 0.8738447108544947,0.20536564889111886,0.1345624181322621,0.0747722696316937,
 0.8422581394107395,0.2073305355932361,0.8177188624067897,0.7194175874160263,
 0.32641502581521564,0.9188370231331979,0.31700301851825463,
 0.5477866616849452,0.04912490708206185,0.5120645122380022,0.8497597943331225,
 0.9053322193768011,0.6055946870586822,0.43673278310466856,0.487772145878569,
 0.265065427056025,0.37767226841995727,0.7699567258515733,0.8031657456868033,
 0.9184743117541866,0.9420697928986063,0.5633705261065121,0.6441316695476544,
 0.8340740349782723,0.2662647218242542,0.3472152043807385,0.3653131693648334,
 0.2552003905400999,0.21310471475111215,0.47069656047631936,
 0.9236247454828489,0.025247103173403285,0.841058626303456,0.7779981834523985,
 0.789383017731357,0.932869551719837,0.16468197291125597,0.8630377310683399,
 0.05555108059950242,0.9715042065737016,0.6819517507874603,0.4078006251103865,
 0.5291264323949018,0.744714194593721,0.6197577417679305,0.12234893136066527,
 0.09922501025952102,0.7824739306610139,0.2885685630122624,0.5813797185382,
 0.7441729841528614,0.4913669610749285,0.3238014793175218,0.5786911689392502,
 0.9889421210376185,0.7483239973048169,0.8961366227057851,0.17529773932286108,
 0.30046605388987646,0.3933239896078098,0.40144152830890334,
 0.7547793266651776,0.22726557503633615,0.9482495243109154,0.2565595369692062,
 0.8294810144853249,0.774075070871103,0.32770410366052727,0.7832232408154916,
 0.16595619112543858,0.387609279464589,0.8898122597602451,0.21450097500139698,
 0.5698502150936606,0.7239944032061765,0.7868393688337811,0.9791981595942956,
 0.7940941376249904,0.3901788798155479,0.8714936495756562,0.6708503143788088,
 0.9961860794855023,0.7377967600607773,0.9424961879522369,0.09253301788501955,
 0.18401838203740006,0.5704134098211247,0.6767319054040928,0.6787783387775442,
 0.4975975068799391,0.6143381012171832,0.6291292779688717,0.4823511513900952,
 0.9006014638600252,0.06544045656259823,0.23628326371352082,
 0.1068003347601082,0.24201848785558977,0.06663079817958173,
 0.7437778627576914,0.6816377346305582,0.32534981029090737,
 0.06904701975414806,0.08305220859106854,0.3711743953662623,
 0.6910473245056394,0.09278905985036845,0.9328814254670162,0.6170470059559292,
 0.1013922207864888,0.5562036305976209,0.042744412163926215,
 0.24395734673847602,0.43832506979835095,0.32579005419848006,
 0.3721781796017135,0.17607066732148036,0.06528868304176383,
 0.42774546275542324,0.08404074820140206,0.5441022707400612,
 0.6991552279203969,0.7124331011974139,0.3299443735570783,0.4055870276092268,
 0.208927714370176,0.16692303390560959,0.5236140954217879,0.3936792203120739,
 0.3050760979356235,0.3816348558391328,0.2774149420301666,0.8033992653375679,
 0.9597516841488952,0.36101730650468367,0.31911346813611263,
 0.8267591841365566,0.8002300873490502,0.04682372126193812,0.7790209277455648,
 0.44546629282896044,0.744282142375946,0.5283789837416332,0.7989056239466089,
 0.872248729414897,0.515999042749763,0.11404981719041685,0.9992100804491417,
 0.5112977295880532,0.6176115628068175,0.6985171983272431,0.4538034772878915,
 0.9186594360191067,0.3955359394439626,0.8876118693321328,0.23925534831863793,
 0.21991943710088047,0.5586147673208816,0.6997213958665629,0.6620016352749963,
 0.26346959245245993,0.6605787314724757,0.18156138076626316,
 0.8826006412500558,0.244263954569223,0.5032267855039643,0.7006660353475901,
 0.9181191215103174,0.27484073569611356,0.17662795476918713,
 0.8922520936361982,0.06849402402457772,0.0775892351375842,0.7316981209863505,
 0.06343866994442482,0.9413332984269834,0.7412090784963886,
 0.48511397354367225,0.6372077495023066,0.36318719807012345,
 0.45477406562252753,0.8094849861832939,0.13170593728326185,0.709813694367521,
 0.194261856466304,0.8371050113874727,0.9741678816661066,0.5631456643440307,
 0.5760197576707276,0.698543317982024,0.9028512304890078,0.1129941388125757,
 0.4493946783403694,0.6631850728894486,0.6306976063331395,0.4579579870390912,
 0.714729311246662,0.8901068881299826,0.35446934794369156,0.2589172485312917,
 0.8267940871299497,0.6701407585032041,0.9392559268447652,0.37047530844865617,
 0.5798266068989365,0.4881960583854794,0.23774727833545417,
 0.04404197047073133,0.5288792659279253,0.7945101058637737,0.8230701176781955,
 0.5632758554093573,0.379173070639903,0.10698525343285514,0.6362517423064197,
 0.6988935793506905,0.2864133415522272,0.03421608011076871,
 0.10178329489430671,0.7753970701508504,0.1061587821815575,0.4714145336683613,
 0.569613280955427,0.9582861740692933,0.8613616666993094,0.23808186335233406,
 0.9931982062823412,0.8300513408856083,0.3941772014774173,0.2257986627402192,
 0.9028994039095782,0.12557178047826922,0.9142812307254884,
 0.16336042899394776,0.5709259783729461,0.4837300453601566,0.5093847819900363,
 0.121827327662833,0.7330528917453331,0.37359351418955467,0.9127388118485141,
 0.5879612358181427,0.7535166979103535,0.5682319216592667,0.5712640741127092,
 0.1052068627585333,0.9475324502973244,0.3216112621203566,0.7132614724997961,
 0.5598691424239081,0.45696764335948936,0.41124592310138575,
 0.5760847731983202,0.5567049375400518,0.10890921678183096,0.743647969852218,
 0.6270411084745877,0.996152042773954,0.9384535226327845,0.6290310927113685,
 0.5093657424692379,0.5362239604721502,0.5565791011147982,0.6899663043489994,
 0.4575972114062443,0.5626119745622302,0.9974395723825543,0.3457361087656572,
 0.1879132604638365,0.5084057124806428,0.8623404205087862,0.4348704762461042,
 0.2758521595166247,0.5166557425305202,0.8295900847191547,0.6616208559302562,
 0.13447546603179028,0.7780603742724117,0.12324918566355125,
 0.19528749677449697,0.8670186596152653,0.5934247523163358,0.7842539273986873,
 0.9428018357971824,0.4949401853677037,0.1594666495602849,0.33886045886239313,
 0.9092692409859652,0.755125430835835,0.018874019691471955,0.4643744011880818,
 0.5590466830038772,0.7467970141415826,0.19391337800206765,
 0.06957147151693377,0.0471026576723913,0.4478811957189123,0.5387515127230573,
 0.9876264808332053,0.5950136599425431,0.5153477576565663,0.2781993763207661,
 0.8491418176485032,0.6590373526875656,0.5638036945836646,0.11900154873609026,
 0.9480886206091919,0.43769323841937235,0.7323304464484599,0.6557124453271614,
 0.6247278621630277,0.6912281208915481,0.8406353401878347,0.6081081329420602,
 0.2978030690641742,0.6418284875651951,0.3751129822908187,0.6194178199845443,
 0.3615330993096255,0.6607525914856116,0.9612003985571723,0.2040812139542837,
 0.23215819671380145,0.09835946794233763,0.7319122237687634,
 0.8444270776562537,0.00299250607769308,0.3884771620239239,0.4414608070006252,
 0.8437886457540753,0.40163389180292874,0.6167790557920371,0.7933249100320827,
 0.3176849383568432,0.7077350595698336,0.8547066490732829,0.15855979096393957,
 0.3935659346775314,0.41527701986713716,0.749813383474624,0.6668030809486225,
 0.009279733426286407,0.23292713032358048,0.24531878660814288,
 0.1134479694493532,0.8772161940326508,0.5334165978624178,0.5557722919882628,
 0.1475477115775875,0.464639358456322,0.7860507283343883,0.059329438730989836,
 0.1964445401710141,0.5663730184602112,0.9938411717249884,0.5365573972495755,
 0.32144816284634137,0.08623970945401815,0.5854264285677107,
 0.7979631918384245,0.16254254394258627,0.6992978576719563,
 0.19015649746127639,0.8915482589393587,0.5293891794444815,0.7554115401171321,
 0.6152329324173281,0.25365448054094886,0.43055160189119657,
 0.46955287793517364,0.5743840187476834,0.6340439895095386,
 0.021362106762595312,0.7239665153867482,0.5753911293513149,
 0.8256146044599983,0.21518150561257376,0.8211687113007444,0.5404290846568138,
 0.9233091347904939,0.7411009236203612,0.37503711718784927,0.9575570445481895,
 0.7773454291852023,0.6900987993742927,0.3907682985232541,0.4349085667596212,
 0.9963638411472803,0.635028276122051,0.8220812203553831,0.27689269525218674,
 0.5908513238683238,0.13487371714104168,0.32791378027861495,
 0.5844422770665558,0.16992867814314172,0.07336609495651569,
 0.6918435367390395,0.14200116223449366,0.45820815087462163,0.567643929337861,
 0.5992200304251986,0.4383565830169761,0.5692832120238575,0.3470158311423561,
 0.21921940635134085,0.5338276511874407,0.0027166356462481334,
 0.7467942686079707,0.08270661557588888,0.8310177543098508,
 0.46401330750385605,0.10626308748731939,0.17449108262500035,
 0.2922456563752688,0.3944579946157044,0.696201581055647,0.9153996619247385,
 0.8599228554940569,0.6187000764090118,0.8231374959626747,0.8825452444188145,
 0.05189484632945851,0.6732900166348296,0.27920149992242194,
 0.9621799515484784,0.38950486495036485,0.15547552566635747,
 0.4988042582607981,0.9657625853252821,0.18036445336337026,
 0.03694266449665162,0.47848232180820416,0.7976802395409995,
 0.9090515672300382,0.8367144502450751,0.7334251046928525,0.13043205795564083,
 0.09957319028267841,0.4563805875358492,0.7770386257638903,
 0.04144316054073893,0.9304098593900675,0.41317452805200505,
 0.4159082616261657,0.9095116721940367,0.6115862261766634,0.17395015016377757,
 0.8011499707735789,0.5580813232454847,0.8075168240591789,0.3415373139630692,
 0.4668459796909896,0.6083880927981727,0.3834468903113105,0.4796422844659829,
 0.8814925779627333,0.12982788362231534,0.538617395049907,0.4843258823208384,
 0.16532336025939864,0.6605467954122861,0.16204786141461325,
 0.9384222264220056,0.6887849068725772,0.169473834420378,0.03393635788915106,
 0.9301110147940351,0.33289935776276747,0.14160178674931834,
 0.17804644647219725,0.9076516347484795,0.613210698088893,0.530220406412,
 0.1618533329224724,0.4863218958761638,0.8345544623301291,0.41396469612228504,
 0.10189740602070463,0.12077535084111157,0.35351368390993976,
 0.2240612078590939,0.27257143686793583,0.5817183579019157,0.3682414059746123,
 0.6300748462377095,0.0085067036770623,0.6470854523981704,0.24235996082589395,
 0.3625417809278445,0.8871671615496651,0.8036261115568502,0.22001495406527938,
 0.29392809072113313,0.9645592199421842,0.2440982156306042,
 0.27738370926600786,0.04051620544148671,0.23836623382646427,
 0.6939279517029967,0.5973029079064548,0.517486653140837,0.014289630950883647,
 0.12074808259858127,0.4698696310515855,0.785670764189451,0.44258523946427863,
 0.23077200675033127,0.6596638990888943,0.9892288276417359,0.5393408949198012,
 0.11202346574877065,0.5592674732439891,0.6929805229776992,0.9123226641045402,
 0.5038576182397431,0.5052754813767607,0.9963348040688211,0.6305121567883651,
 0.5415100188584036,0.17567304132709372,0.12384883678024661,0.314452345450146,
 0.7414186112803688,0.8350025270141328,0.657263416597968,0.383859969620549,
 0.7955328971363284,0.24984329596862442,0.33698160533822374,
 0.8844415568336914,0.3948643797252772,0.8806673155110456,0.7021027415789373,
 0.040054065269339634,0.798252751509417,0.6201120298582657,0.723080386874575,
 0.5115257048389733,0.9813172171799828,0.044017569980184024,
 0.5506075131298278,0.5290547684999761,0.8406257940427334,0.5169470040578206,
 0.98043386981483,0.20498677526014397,0.1718153717374492,0.9673073020156776,
 0.8911346670516034,0.6388280998400823,0.91984721313508,0.3053064940849244,
 0.8168048189810113,0.7427318850538313,0.04926132754617896,0.9217166126253163,
 0.9245352672397427,0.4747132709554647,0.2983060983085113,0.18893951809603704,
 0.9265100209476469,0.34132213711727566,0.8802157545032395,0.6593479864353424,
 0.14256652825798244,0.6925812323858294,0.07683244346767437,
 0.13154592764109174,0.7825304014335425,0.2698883017154283,
 0.037166459030973154,0.02070306565392166,0.9502194331707317,
 0.9038574061062739,0.27570012639205377,0.17388956377667486,
 0.8745204916293058,0.6587037755773123,0.8803662902573575,0.6680683288140467,
 0.8706183734415021,0.7330414435279511,0.25941703899210444,0.7140369654106514,
 0.7120434589572822,0.27624755249018007,0.6346206337277471,0.9451568586871105,
 0.1924889488320849,0.31378358630956427,0.3222228745635404,0.7540973834169116,
 0.8861972315932852,0.5491035887394722,0.7844390434347052,0.3711571236187189,
 0.6119753388450128,0.9096267121070138,0.8216262146523867,0.6216737684900887,
 0.6123697787619858,0.8454985461358765,0.5629021007079192,0.24020665805308694,
 0.12092508904190058,0.8113043793877948,0.9451301275253516,
 0.054340654543602596,0.44374742548268886,0.28950195554354774,
 0.8285228586457298,0.6001443328970767,0.011308333537474735,
 0.9425268255597714,0.12324286477359536,0.1204519749717049,0.8848872299967478,
 0.2706789025743732,0.4828257377007783,0.7356258159840502,0.5128775345157864,
 0.6483369785103825,0.6043357669379147,0.08267630319992691,0.4188493238811595,
 0.4341497296237016,0.20530561783698387,0.2802492306300337,0.5016491451228363,
 0.9499364639288395,0.2229492468502995,0.2671584858338465,0.3348618026802893,
 0.6106994429329884,0.49349253530946435,0.029245310369727662,
 0.09185840364645537,0.5083108432787845,0.06349169336075844,
 0.20232501776291567,0.25793294860897686,0.5840081697703183,
 0.20309635422793726,0.7888534368901572,0.9938172277216749,0.933345891580533,
 0.34219049781320365,0.3885046842613349,0.5539694251018832,0.9583340966810034,
 0.016928511496281273,0.355471391599101,0.2624412372611482,0.2269274546191219,
 0.13583790029672893,0.3578183851720278,0.9910104784534809,0.9701067056801053,
 0.3683219672488385,0.628511805379876,0.10508210536198537,0.40240549177192186,
 0.42256007869291323,0.3806503741728616,0.6301038576316702,0.5701999244735173,
 0.7765191699377154,0.7546844729077828,0.35779916866819805,
 0.0038984436607565343,0.5891853212028788,0.7127208990309339,
 0.4218146338344514,0.8081402368047879,0.24520786320186772,0.5131118638561033,
 0.12038029178789911,0.9205004221893505,0.6061900943631068,0.3363570111724008,
 0.7849908069752971,0.38649788874066515,0.7740423342448404,
 0.47907766702915044,0.10685269227800309,0.31240385331936815,
 0.7524326355365684,0.39028656468507494,0.29166130069700025,
 0.8019290098088002,0.32027314419856057,0.5435842489867315,0.425347058630292,
 0.8034018935612658,0.04531129179886795,0.9094895672167892,0.8539641714487862,
 0.6487294644956779,0.06655435116907327,0.3861780615960779,0.4380002495905556,
 0.0913704732311349,0.5795427799403586,0.37081928928014873,0.8795041624810731,
 0.2112770606693155,0.14194808477715548,0.02320631977165788,
 0.5372716280475617,0.6955439730470417,0.06177436197228181,
 0.025830659582504856,0.3070354132941444,0.5960605301494594,
 0.6337182011494674,0.1686927067863646,0.002615823969673192,
 0.1481865033204468,0.8032949858949363,0.20985633974200224,
 0.14143307987448595,0.7222870416470042,0.08396658280855207,
 0.38216457654552427,0.924480302642994,0.6027623020232662,0.5915608442367197,
 0.842943108428785,0.585420197146721,0.7855126534981387,0.7775651529935159,
 0.37978710749317823,0.7502299948605853,0.18252775766217444,
 0.8825621871868745,0.8938261736007725,0.7667730936297551,0.7424562756132674,
 0.8081769541143796,0.4956632536643667,0.9069058547195943,0.8831250817889288,
 0.8761920964346063,0.8456586680418289,0.932157874484865,0.3546222804879754,
 0.8760191676157395,0.6668012801810961,0.8497671248896603,0.8102621586674281,
 0.004486385373627444,0.8634277403208506,0.9034843768614768,
 0.9480252703307688,0.498465637015991,0.7066562625701597,0.4511077696853806,
 0.5634187275336114,0.8932412358134969,0.6325149759203147,0.6736384070648748,
 0.05714813484314585,0.07907972555184328,0.44086936087616957,
 0.8062777381249496,0.3477846322226795,0.5005533747728549,0.1530884472371079,
 0.3274235058960281,0.9991896177541998,0.2539905599366197,0.08767258687988355,
 0.45547808442625115,0.01437884131039513,0.8159405925364538,
 0.041945093878311135,0.5016277012143752,0.01746602471406844,
 0.06209655703849126,0.7810060027418817,0.42068751280751904,
 0.6212337683260178,0.6141882626551077,0.7599869769074723,0.10586682519460155,
 0.924495784258153,0.41076356464227604,0.9958956046398881,0.11661071438688186,
 0.42832682049079873,0.7459763112498041,0.839329431247798,0.14479371744955793,
 0.5688238093453991,0.4372763243379769,0.7767049367188217,0.14683633542640073,
 0.6001867799781987,0.8532309231274907,0.6961830442192416,0.4302902605022194,
 0.8428093473543579,0.4430212685599554,0.7021393511929336,0.04691210369869636,
 0.34555693249069197,0.7499530288314956,0.6750436447241573,0.2818824031632978,
 0.5307875583728285,0.7377983317807454,0.22760790928495545,0.9496364946875563,
 0.12800341800828696,0.13449640704133325,0.8492596504742556,
 0.4766991605972989,0.3719021311124606,0.42582510690611086,0.9764057688505877,
 0.5141021884047957,0.16359144077299703,0.7374438078733618,0.8479390285026558,
 0.22574470926542034,0.7105757666162977,0.5253618624622387,
 0.06420921003829194,0.0366909738962653,0.7708883634145108,0.2661432515668203,
 0.8304983252956134,0.4639181149589322,0.6866064258369555,0.6746471882349409,
 0.7248967172688808,0.030992353297352726,0.9760355055576477,
 0.3459373330985439,0.5433185132879081,0.4855705195141232,0.8926938433516094,
 0.8596921568417399,0.06114615389045219,0.9798935403639951,
 0.23677151389056184,0.30585388932505797,0.2748728761851347,
 0.8762704765960906,0.8545586622845434,0.8313714802334499,0.13317119769575725,
 0.9859283796324683,0.1106784328605378,0.31496256039173975,0.6411103169217497,
 0.40736591075407025,0.4770543788768574,0.6782013418399526,0.4522353657930004,
 0.2625827783069493,0.6747753610900217,0.04743065741220365,0.2795150621767434,
 0.26600463561285237,0.3680831802669272,0.13690148495577648,
 0.17372703404669054,0.6721561876055577,0.1255690234547101,0.9664866493891001,
 0.16416313319408604,0.4857950070256226,0.983529858375523,0.73006895898542,
 0.4556428324811781,0.675876293207945,0.12139235885402988,0.3389768247499293,
 0.9998196949846052,0.9932561973338799,0.8396192139369038,
 0.051708005034148075,0.03724506727202448,0.013681289467162516,
 0.8743691716973658,0.7205879306365052,0.6020816642942255,0.8517401484273142,
 0.3313582486412825,0.3677924144158835,0.8883451774136029,0.48921031419991756,
 0.4694868112124442,0.6746640679777656,0.7362515135613832,
 0.033259404678536275,0.8387658382712513,0.08909398038258165,
 0.7250806279211646,0.4488008901675108,0.5408593720998157,0.6822916902659123,
 0.6137843778089954,0.7415096197680284,0.6366730464690072,0.8477138294433164,
 0.5127090975568851,0.6111589312754458,0.3476398557131658,0.6388058026029881,
 0.6158293482761671,0.6487570468450266,0.07019896128006986,0.6364834708989151,
 0.6959344447117715,0.26836436008818276,0.738536599365077,0.9668873987910744,
 0.6426837495633584,0.551213582354879,0.7703038950001015,0.03473061225859686,
 0.7885868733683604,0.23598322891400558,0.5463205420410804,0.5651118008776715,
 0.8497502566160119,0.5720336763072956,0.5518661846614142,0.1223284573640061,
 0.18855849850275208,0.0626674850246044,0.9516596638695979,0.8026195688934632,
 0.6021381318264641,0.2783152896687804,0.42183928477399,0.6778934280381741,
 0.7241571994271194,0.18646886619334013,0.34227186807639465,
 0.8096041634604143,0.026809522515876427,0.35863154486857507,
 0.03708456803073945,0.9495169959837011,0.7592242339307633,
 0.27604481892963073,0.42355894980567343,0.5466618825389209,0.967804039264653,
 0.9722824357547803,0.12421129925522256,0.5601228148942556,0.8805748756767962,
 0.45788515996544077,0.6903701810785275,0.3029602787379272,
 0.14554133499416022,0.25355605990782104,0.2503697064835051,
 0.21728037499124586,0.6532045772303925,0.43597528854809975,
 0.3796945595865714,0.6983803870853172,0.140760010671646,0.5704560246905739,
 0.08364154322160622,0.6260565541136498,0.4441554788980505,0.3008426771652739,
 0.421792060669403,0.3241632868568628,0.3282515839760467,0.6893073089342491,
 0.8863868337987078,0.45233262994128776,0.06292633609284182,
 0.9551872016813276,0.14428409080809412,0.15013254996309577,
 0.4576342113057239,0.3377059212725131,0.6262503577599502,0.42891737721875767,
 0.6324388403706632,0.1491695101181013,0.5669084871451068,0.22866625230985527,
 0.4038976669690595,0.8471682354197878,0.2828951526513279,0.9052620675320531,
 0.005449515557311324,0.13692822765816437,0.5725930640385343,
 0.2830393861627476,0.40633899892929914,0.016850085187777752,
 0.7404176616868184,0.09953451661869628,0.4646133751352278,0.5808055122023454,
 0.6933084483247407,0.3915624590551483,0.8381821312535411,0.36673372264609727,
 0.08577863139634845,0.9634892061617688,0.41682880390457566,
 0.4097689246226177,0.21312951623258936,0.3404360484987181,0.7146506364189247,
 0.5616655176019233,0.21837260933188019,0.34013144500587555,
 0.32902389797913245,0.1685907514514995,0.03821453098884986,
 0.5951782036894209,0.289856773571459,0.12624093739730768,0.7072550424059757,
 0.020610735353430054,0.021923362303400418,0.6230095752273899,
 0.37387827909835125,0.6791842428732492,0.8936232102570838,0.0397340350238502,
 0.014006992965018839,0.6844619013348111,0.4004318418079257,
 0.9894496539211927,0.10694314431760932,0.5260309604048296,0.8403353276256857,
 0.5600614977548233,0.2924776528172993,0.21917745682257184,0.723040492930789,
 0.838068930076796,0.8456531604999783,0.5208842893012189,0.8136547481667862,
 0.3750468988674964,0.7290188856122866,0.4323689156982483,0.570903623866149,
 0.6948757543229511,0.6244116699568918,0.4066866487500801,0.6676846925620704,
 0.3531244763656931,0.7386729729380933,0.4534516755892031,0.9703338039180983,
 0.6474731437695119,0.10746252084010133,0.18610327129862714,
 0.46549618299556417,0.30312923254180657,0.2776055635966106,
 0.9308291658554936,0.1569828468293043,0.905536939880339,0.2163075014254674,
 0.8127118025040436,0.6172466682703849,0.6285490143076535,0.9625859896466983,
 0.6325530798496051,0.3020500789262438,0.9691582138225537,0.5433520145638983,
 0.6782120332349036,0.025822208674432967,0.5817770391697142,
 0.6806530362661674,0.3088678133660271,0.1558928515283594,0.5941365531016527,
 0.8827974785492164,0.26034697251231953,0.30332516224819106,
 0.24744280874358449,0.38327236216757066,0.4099626912427604,
 0.5028421450022247,0.014660423646401965,0.5835643992927625,
 0.49328126376512804,0.11633699204006376,0.01668196261173538,
 0.6405960592168006,0.09921816184475418,0.6458902090961172,0.8637485517346152,
 0.7611131862691702,0.7333148131219582,0.4045588396959343,0.4459656272831415,
 0.9192424892987737,0.7705561112902116,0.8090839301553803,0.09045093499030188,
 0.9042074834965331,0.8032843556597569,0.8707165381913258,0.5445420738660398,
 0.75038819064206,0.14004575267550645,0.3099491404412267,0.6203373038898553,
 0.6035490615792449,0.8798785852962329,0.1532046911130689,0.24079165343311593,
 0.8773004871086264,0.3179391887248614,0.5812590304423625,0.2702703615531947,
 0.058496973577278455,0.8036480082901201,0.19994384880545324,
 0.42487255809978897,0.7951364338409697,0.3406073851719682,0.9450311831305285,
 0.9549219508764957,0.1860198630389307,0.7259434584121708,0.6021685289276213,
 0.9767233801092639,0.5419602305392154,0.04783750387139141,
 0.09093348313201965,0.608071824695099,0.9220420768148119,0.5964236668303861,
 0.7630863547472553,0.8115056038590491,0.9689740192230328,0.9111213542387069,
 0.667164190761879,0.8626517268156102,0.36529727648202703,0.6076232056693659,
 0.9679161909047104,0.40045467059524764,0.4698178730593142,0.6230401949010006,
 0.04164941284709389,0.017521268955785874,0.44162974448032455,
 0.05536097142705254,0.09865686918044092,0.39829819817711265,
 0.47510878045866667,0.9808388214971999,0.7207224193942519,0.8080680322654294,
 0.3676402956191268,0.3087886872325186,0.2989069722610478,0.14504375043045314,
 0.36669867926316546,0.6253685383821377,0.893779705725076,0.5168274530194717,
 0.8779998366830073,0.8994058010673437,0.24864144335644456,
 0.05675919147082975,0.43284409446676686,0.5835515437121135,
 0.5946434859241744,0.5501827109478756,0.5158701174128353,0.937790974095785,
 0.2689006088322474,0.46911371117385814,0.06870346312940634,
 0.00384040236583183,0.3277401036324876,0.8000800772923039,0.4627099168538412,
 0.9984347328859218,0.9078083738620929,0.24373442413365876,0.9340200163452167,
 0.7840601788733855,0.1945869827513539,0.45474634539631476,
 0.060338272335965115,0.5279652429413292,0.8802545210590554,
 0.05964640723988235,0.0765801945467307,0.5778752161375289,
 0.44762984860175903,0.986738330988867,0.39821541640109204,
 0.02261502064814125,0.0638285296542701,0.7452067958677183,
 0.30093142569175513,0.09634462214588402,0.6455137332367239,
 0.2720435452141139,0.04638194848344468,0.8579618076494159,0.1886466531221883,
 0.2746436598185191,0.810540165912023,0.08721358473173912,0.958585006917156,
 0.12946755672933274,0.8666164974613929,0.17760932254723405,
 0.31449474911876507,0.06233624857018971,0.7503384732898077,
 0.23444401194725617,0.7405446953689073,0.5800792731783173,0.6176280452559195,
 0.687287949620715,0.8178581790693974,0.17075554313709118,0.6653260196997417,
 0.31438868691660327,0.3002813668398383,0.41053469075574933,
 0.7123404722994654,0.3611404761833332,0.36611247437902217,0.8808777673906244,
 0.3233340640045326,0.1782225382535475,0.5606383941213192,0.7207211265431022,
 0.5368257708849241,0.8847794653774992,0.5564446510928343,0.7248273348561374,
 0.7442172495147908,0.8639342482049837,0.4628507548819296,0.5814522909843802,
 0.6606935653744068,0.23138336034296803,0.145925571636067,0.5792197088614846,
 0.2893714405165997,0.9023765472266146,0.2413104893448993,0.6108779499513967,
 0.13973187153396105,0.04931302467870802,0.4624544871291075,
 0.7535039646962207,0.5953174646185628,0.4497239988106987,0.42464195295490414,
 0.9593016420020376,0.3869638733963586,0.1946560754424218,0.6002960249497721,
 0.6119371905801251,0.9491880933758233,0.05723080542199954,0.5068525353696457,
 0.0948372805267671,0.44319237328156247,0.6545098471072249,0.2301421200265339,
 0.8622344549199281,0.49321422212654675,0.9070236386149808,
 0.21447476267425647,0.6524217303890403,0.06153123660344251,
 0.8211695653535124,0.2446166560618196,0.8008519894885038,0.2709436562915635,
 0.6519738383097494,0.9575637349696594,0.07411982030190933,0.9749755395695715,
 0.6901868408937025,0.7649843691448622,0.5363776578997325,0.9798946862919191,
 0.48520553731659344,0.6342704202518369,0.18542885731351322,
 0.6096992483580892,0.11429120984760943,0.9610796228653692,0.0963718799383032,
 0.7524106742355708,0.7123176596727293,0.4859554242000129,0.09784808811312928,
 0.8927651189088045,0.5580520364581716,0.7907482086216556,0.9749362675615482,
 0.6495247466152236,0.48832709706012767,0.7567070919009405,0.2823604298774681,
 0.6134540934921777,0.7639343881169512,0.6922965185754741,0.10986236924250226,
 0.802164470495512,0.1363895615347427,0.0569040961874151,0.7016659639183296,
 0.9546965729319484,0.3548142378572171,0.9097461997292917,0.7624743733660178,
 0.005911239958089287,0.9298674574634823,0.9580549284102304,
 0.8919454623931135,0.6484900666325197,0.15170085927737798,
 0.12791827187870597,0.18308643805773783,0.4718716255441421,
 0.7910073631533225,0.24382047821170005,0.9434386980972795,0.6928845787326179,
 0.5575810540396582,0.0033424923746046,0.15827538336263047,
 0.32364622683324074,0.8309166099637065,0.6739897444573661,
 0.007471538555965496,0.34215066793825855,0.22784554393664413,
 0.9086547422553992,0.2842794892108478,0.17502978887273157,0.3155609141399608,
 0.981810088631275,0.5143249483381271,0.508947070285766,0.9324767842221123,
 0.43952078264408356,0.4416844488660917,0.19467417205534288,0.369456690960823,
 0.21876123695542538,0.7059607633863538,0.7800258738231788,0.8111206338047661,
 0.4378894854386589,0.8360166172091523,0.6048429955690301,0.3053413545903608,
 0.36205921937397956,0.5139786268528796,0.2850382388902768,0.7875268454733683,
 0.3682650022980519,0.39723024430193865,0.7238766021057192,0.2902616895120682,
 0.4369946576551538,0.821099572711352,0.19244205949408077,0.5896631668952993,
 0.6160071569580712,0.25573683994541285,0.14663754973214305,
 0.9130018709555578,0.3023009787331432,0.05072312974518767,0.1629141873580623,
 0.48156960198080756,0.2567789277781499,0.4396705853393902,0.3520126389688424,
 0.5763079352628112,0.07689802707545534,0.42395290087691906,
 0.37551144792938207,0.017902172881895106,0.8519800714911752,
 0.4871115350453685,0.4871533614904072,0.28309869363645424,0.7953529806327874,
 0.15787988410381515,0.9984692137399991,0.5749681795567037,
 0.011124392053461829,0.7032131233241952,0.4417077089625413,
 0.7759543862013366,0.17101600536420836,0.3378000898063157,
 0.37060959933890103,0.9301256971256905,0.9547821416378476,0.9201296769097992,
 0.854176843141166,0.22690148460341564,0.44575608637086594,0.5030644187824681,
 0.4843627991652524,0.20606352385686533,0.519287806393129,0.9505370166346457,
 0.4880149533429028,0.07687873650381694,0.3855306480668399,0.5619812594479922,
 0.8613316614703536,0.011412306898492708,0.911962882139426,0.1549110137195422,
 0.005536116701404414,0.4357400069327766,0.20888268824188505,
 0.7965956227443511,0.6932247122796509,0.04936628757108463,0.7169666642942771,
 0.3360127890150225,0.04644024611503905,0.7622530133076482,0.9964164252107273,
 0.03021285443646171,0.2594854013632648,0.7371299333954825,
 0.46820911684186817,0.5738498666505244,0.8612955211607711,0.4963962269941946,
 0.9104782041476834,0.00854766581584121,0.5348639584782019,0.8479245630869763,
 0.6273063625632214,0.1328497451407934,0.37614779835203127,0.3986946792852012,
 0.026146003659154204,0.24681635877463382,0.231565186217505,
 0.5698274754618557,0.11819989034998046,0.1603478373069801,0.7386035829981128,
 0.100855137347441,0.3564424240418027,0.17653978515253765,0.8212903721365118,
 0.05789954600815117,0.7910222729271168,0.26931463403871425,
 0.9807076257574717,0.869070057418231,0.9043820690471919,0.5135560810829607,
 0.46934206853509,0.26900452788366336,0.15986598479650704,0.6105317168022649,
 0.5140582811552448,0.4914880055632975,0.7108704551642108,0.38306131734485327,
 0.5910761264030825,0.046674068231216004,0.87378959254477,0.3205155185098618,
 0.3783658342874191,0.13584319244793752,0.9234117248421281,0.1388294007632288,
 0.09591188240620074,0.1894593637648937,0.036361177463332894,
 0.4297037731332163,0.019166685096208225,0.8931045791780597,
 0.16213652630489206,0.3132917017625979,0.21593853571689536,
 0.9017896082437301,0.34043186960642724,0.26737629275012464,0.078028194512201,
 0.4805247090220842,0.963258161944549,0.34654411006781904,0.8828332075377994,
 0.5623570004728606,0.7297326329597807,0.9040915950032202,0.15357295559468254,
 0.6806988251102205,0.07366677255897447,0.7245198338598415,
 0.26270467986153434,0.27790486680402837,0.2778506401138694,
 0.6971390703110317,0.8659434822191993,0.5591190615003743,0.7288693774959831,
 0.6658593633893585,0.034193451067043634,0.8551991362054076,
 0.5623957795360047,0.28520431710729344,0.7569211765825347,
 0.40869682233420734,0.6419173566066276,0.5335239734591395,0.2445941185789653,
 0.7014163262177537,0.5847660293169813,0.4970190384621965,0.6321599378743317,
 0.7783359978015167,0.4544465774937818,0.04924406698667927,0.274267545311045,
 0.07302719716703698,0.13526968428108366,0.06823280635322904,
 0.6341886796596093,0.3243923331041144,0.38487439526474987,0.6036871686663365,
 0.9199300902878945,0.8546908341326567,0.18077831588021198,0.7082938739215283,
 0.8403655449685521,0.6181781046454108,0.17114140460857508,
 0.17970899888125502,0.06292267283509656,0.33600698222034486,
 0.40390659347437663,0.42349041937983345,0.9263384444488663,
 0.10965659162542096,0.40311265836126,0.7596393587095183,0.6994883309710891,
 0.10138711176014525,0.49149546486962226,0.2737663210424416,
 0.7261349265171604,0.739508321970523,0.0957579874698421,0.542204883988184,
 0.6569493093991922,0.5853984873381436,0.5335392262112503,0.7856348474559196,
 0.2867804021991964,0.43311792192234666,0.3989442587294787,0.1633216300249981,
 0.9132279878900014,0.5752216993639624,0.018595498984182557,
 0.44367770745802826,0.3964238212467046,0.8334751002034156,
 0.08203366214207763,0.7454692965329138,0.47996386104655664,
 0.04196986276752035,0.22622903621151824,0.08588679686739153,
 0.7616919773863955,0.5541350628378104,0.00902220435428236,
 0.34069003660305064,0.057029408431937045,0.595541114468058,
 0.3891244833360691,0.23888731617211334,0.790188561095953,0.43203523817354383,
 0.20041571212243459,0.6603538265133778,0.25220855364890915,
 0.8480344733205685,0.028476930407045686,0.1515508698395469,
 0.7679236540280495,0.6207743925309819,0.14474475805485743,0.253422663566615,
 0.40596722473377556,0.9680203664763837,0.9115541607542694,
 0.32438629693060494,0.7424739954978176,0.3566331986091644,0.0887842960918449,
 0.4613403642110868,0.6344382255967881,0.06617724810045367,0.7105454575390489,
 0.8178309841651313,0.7159870559826934,0.8242154305467797,0.38091627092046487,
 0.33724831659661725,0.16243670468360927,0.6179572635140617,
 0.08223536227153372,0.7843153281774131,0.03368426787115597,
 0.47966333719298015,0.8668294888209205,0.6585005480353497,
 0.12948172682436088,0.9203659915746205,0.2254218521735608,0.8626765723137091,
 0.4991375325182119,0.5679395626786566,0.0916058875520287,0.5650860291784182,
 0.9933621608879128,0.7647062543557571,0.9879371064393057,0.1888707915016431,
 0.25598485092817824,0.6373210654176571,0.7131920345244815,0.461825040876358,
 0.2626492196298891,0.3534103198980665,0.9014372359468161,0.3957857928107895,
 0.5090855471656734,0.18406102096433574,0.5369073688935612,0.897924237726611,
 0.9041661416114892,0.5077871652435679,0.1336346618043034,0.693941940830436,
 0.813195454030781,0.026645173793121923,0.7936761117872625,0.5163381323035259,
 0.17047553481086686,0.20315164525691431,0.10641509656456927,
 0.4275366902328508,0.8648221008653432,0.5827562408670455,0.9783576757769264,
 0.7139153768287319,0.3167235362216537,0.9042029831186705,0.9447288357844462,
 0.46629779290085405,0.4172026719404637,0.8658119860592669,0.2070347552214623,
 0.3090396895963321,0.15095750514381523,0.43699202538598025,
 0.39668786931457534,0.3180987298049388,0.8520009299544802,0.587200184521817,
 0.6733482148854617,0.3968142218846551,0.8280949149875902,0.8492036955509059,
 0.7592196007814815,0.9003593085861339,0.0765031984607838,
 0.016583585197060602,0.004939563789326251,0.06321044776423168,
 0.33144157192073076,0.7155465056796309,0.030472051014608148,
 0.35427048527315197,0.003840762920068408,0.37388820942033896,
 0.7242793208894533,0.687099296296633,0.1517252531031963,0.006072396021141735,
 0.4703498345408168,0.7840216238438497,0.29859048168326463,
 0.34223955573508824,0.9082247878857925,0.6102045450565442,0.7481915562591096,
 0.26924012969891153,0.12330323018154554,0.8975229335057973,
 0.7544362376353958,0.8408010940644736,0.9377758047572574,0.9962978175343139,
 0.8774934877088721,0.9633411183455682,0.1939926232080259,0.8467811715925088,
 0.8232911891451987,0.979615668570461,0.2809148804488891,0.375942598295118,
 0.4521352238999867,0.3090106847613816,0.16652873186139616,0.4161850900019092,
 0.12865973702843325,0.151249895843948,0.8537342332836637,0.5419923223267349,
 0.7788682764210474,0.8185947366883615,0.930435831187719,0.35713924468040226,
 0.1482027210224014,0.802615786428778,0.9692555598819821,0.036450486260614756,
 0.9977917077116496,0.33193817194249664,0.42427325327882015,0.618963598236112,
 0.04362407518467437,0.6908070945984055,0.5402848207023222,0.451828621748502,
 0.14645765513493103,0.29843884457798875,0.883848875497221,
 0.34172752576984167,0.6471842568144329,0.44681432761123674,
 0.9907365257483893,0.7786574623345568,0.3462375627188925,0.7605336857512724,
 0.23510517237649053,0.6227801912463354,0.2150044967736917,0.9937033880537028,
 0.3781514990264102,0.25069315564050865,0.6201388637034797,
 0.30254942246558925,0.9437404776216791,0.21155713126136988,
 0.8501575758563785,0.1386248323421405,0.750634865474984,0.4672266887826042,
 0.7250320131907596,0.5004184829048499,0.07986676439831542,0.8040776064193248,
 0.6804921448575519,0.9583802887818208,0.6906181152453982,0.16330675565552166,
 0.5163556746404647,0.3837592351484852,0.4078275471908982,0.2878343481279212,
 0.3562611573406388,0.7804318349259045,0.08371435673102856,0.515271880854297,
 0.7625862986088744,0.8515739452716538,0.20082918175088726,
 0.11256427246568146,0.048607984942235616,0.8149490241594591,
 0.4572823629830819,0.23334921943016385,0.5615912296874133,0.7193704727619061,
 0.6452690019943512,0.6580321852122152,0.10687495646279532,0.4832165259006246,
 0.3537306421274107,0.07729285850238077,0.46376862079294956,
 0.023738063897159933,0.033456005799490596,0.1938271941832007,
 0.4764149318712403,0.5628595388713402,0.5573059527995237,0.40682144521434005,
 0.05280475161306064,0.018352538393894835,0.18266766623308284,
 0.6958535393815914,0.48551434127447946,0.6363387088824979,0.1367039075726153,
 0.8411549596297675,0.709116905206447,0.5385362435869026,0.26300960787853955,
 0.517020682130463,0.16506422331638082,0.20966237938796273,0.7223943998679744,
 0.21527428144120597,0.3630862253671764,0.7224323927926577,0.9281819111883847,
 0.28434595777021787,0.06146981812267582,0.024548966140011896,
 0.30411464037475133,0.17430879553048206,0.09287710146355876,
 0.9046790239988687,0.9757465442008524,0.7936351038542107,0.4737663244340946,
 0.5136278520502245,0.6823782608948717,0.7521272970942149,0.7157469528500239,
 0.16058020227420267,0.5904574155784544,0.5750486410941735,0.9092455702833775,
 0.2706310339841843,0.9854990263475596,0.6834520888288139,
 0.037934160698847585,0.3227656624715922,0.1408778179945981,
 0.10036194351122851,0.37206250746980396,0.368122928158095,
 0.18213219911047818,0.9836855959202885,0.9695213990228714,
 0.22909363296851837,0.5742829535994418,0.5435375671368096,0.5709491024301793,
 0.30775153423857926,0.48915782887306414,0.9847375017389889,
 0.7167185095873747,0.10549549758513055,0.16984543226106585,
 0.0563174339490089,0.17678649960102277,0.5945083631142426,0.8129223174024413,
 0.22928661186079613,0.2566182884784618,0.25744456368082624,
 0.1715105066049213,0.32251169445763417,0.4334847743464132,0.4225950901452926,
 0.49614485534858943,0.9733379943963787,0.4845440751199359,0.5562365887028422,
 0.3912625595056821,0.06938340028465406,0.34731999214379095,
 0.48029247250491913,0.32383719445077164,0.012147898416784098,
 0.4488036788537666,0.8109778082560478,0.6131562065099723,0.2868871757804604,
 0.983087862390382,0.8709365624866389,0.6349029182178044,0.4383267251534636,
 0.6561047799375883,0.38776566941056867,0.055985382134000616,
 0.7421995206598381,0.39663684018666034,0.6334738927661353,
 0.41677231838463424,0.42504809335415683,0.6725462359404202,
 0.3607069522248908,0.5282572870801744,0.6977479606799476,0.300039600331766,
 0.33315151172060764,0.35592128276283685,0.6341855041915627,
 0.3731722452359665,0.6286894055606752,0.050122454936508465,
 0.5151597569606611,0.5826541754595909,0.6038224639230173,0.8766957257931189,
 0.8967423150351572,0.4813789656816193,0.0014899813840560938,
 0.33569897440976026,0.400098292115328,0.35597680548562083,
 0.43703930046202355,0.8265604653298599,0.09214111096045652,
 0.9061472866854655,0.9491064406430494,0.07639235863161098,
 0.11120988000290366,0.39327231098294946,0.5878252636624676,
 0.12578044091894724,0.5106037943510147,0.7044637708577843,0.5217837265739194,
 0.9491488443608678,0.13993639972273353,0.9822452592633291,
 0.32187403722910835,0.9388931018506949,0.8423461718897002,0.4529574098960648,
 0.3964320535093504,0.8875835898867765,0.014852009353579665,
 0.005665322779759063,0.5926925783531035,0.19885264541232162,
 0.7383589419428922,0.03051642595418913,0.661842928831851,0.04577200945718862,
 0.38237008807375017,0.8180755778191406,0.758754770932669,0.169768854789214,
 0.17369289299503188,0.3374184321328031,0.6016131398212855,0.3717610434963683,
 0.6730275914119529,0.10695339664431491,0.03398965064669224,
 0.7642216998384105,0.7010895358015148,0.9211927381198555,0.8170761739803619,
 0.6803444296367168,0.5805530239557347,0.8919229765599269,0.3079452140412251,
 0.2343371587794807,0.5821228061723813,0.998984081647748,0.4347547554876293,
 0.5253728109388252,0.4932855376673513,0.3439732228069948,0.548254566895249,
 0.6105090453056213,0.05462554769177952,0.0941011789779721,
 0.023793258414603113,0.7495199506893214,0.23884114212456475,
 0.3696428449220889,0.3539493700536047,0.4069482406622218,0.2344530905496457,
 0.5448433180029154,0.3104464045019111,0.4356777536705022,0.6760265183763243,
 0.17134891343853975,0.8635261394545322,0.1606040480849238,
 0.46021014294649354,0.03571082820561178,0.0242774083691919,
 0.9020894859563288,0.5201189623654445,0.33999640842435486,0.662665088662539,
 0.25416420817885166,0.8647112160020775,0.9088475205972982,
 0.23893457078975755,0.23914034440292398,0.8394235347263707,
 0.32726087758866906,0.8106976135919876,0.020490427830998414,
 0.48542284754192544,0.016565697557969883,0.7401434622868438,
 0.18275487301111937,0.7156769552510953,0.5482568247573437,0.527498236424157,
 0.254175215561376,0.04814566786737051,0.67409366530629,0.5774545313243573,
 0.9089930858258266,0.45289611216664694,0.4171349837617411,
 0.00512918621077707,0.8707757770588481,0.24112548120529576,
 0.26750981252305595,0.5080895011961302,0.9157772110555588,
 0.13111621901445336,0.2949285764559175,0.6165235219179622,0.2480741089284192,
 0.13497581728465569,0.776637966806677,0.8145661338728384,0.2696887320934036,
 0.8206336175334659,0.4844277535454914,0.27383683352331034,0.5430184512053678,
 0.07761192541855833,0.6098790858580135,0.17540318183265702,
 0.02128380343139502,0.26050910849958164,0.03297097567775242,
 0.6069171585869281,0.7335659750375917,0.8550277587785644,0.6634507252083441,
 0.9629966411293701,0.7964535782693708,0.40950585573220744,0.8329591738725527,
 0.9287155366381049,0.9619898402918908,0.9139387036862208,0.848488320664638,
 0.14742664677763195,0.3453986549836269,0.5111285604885717,0.6669014217714523,
 0.6716583716241292,0.6966883007464721,0.8174199055812116,0.30917715506959365,
 0.09861457209473468,0.11101287371249291,0.5253655863028911,
 0.1899908779760362,0.3435420110682613,0.9629584746927923,0.6030557220372305,
 0.34845122179375587,0.748292780551984,0.9524409679740407,0.9270303517129435,
 0.34566469300466895,0.6787200145068758,0.19474265520589884,
 0.7770709955752884,0.6763622396655585,0.9550082944224706,0.9556989762368288,
 0.6324074809525085,0.5281278520642935,0.9050365194484675,0.9071291986466781,
 0.36925581340713653,0.9771780227293811,0.57705159670803,0.19219383923179356,
 0.39487371721985176,0.7305348986922756,0.461408495155493,0.41082114395489566,
 0.842326508167882,0.278900440452725,0.7253398636774177,0.3594931289957206,
 0.42661133599052325,0.8716701127723396,0.23288706144978555,0.68114974277716,
 0.5373758113155545,0.8034680974802518,0.5725999742986018,0.6868361687876937,
 0.7866728272826636,0.3180366799155401,0.8581357933771573,0.5466600513896351,
 0.25928716499757165,0.17121379915164114,0.19777938108360593,
 0.19218368042883316,0.32371470398665203,0.7711834967396224,
 0.9123100940665927,0.5384261324427453,0.6381866760474947,0.4638587734363182,
 0.7011891050932298,0.843797455124548,0.635062592476804,0.0010196641199827772,
 0.8487332556751573,0.24777082703197195,0.9659055045540442,0.2489405843134087,
 0.8246968819128562,0.18660720972807865,0.6506836856082656,
 0.41284387563756697,0.6586723166456435,0.8724975178852363,0.5477126591411928,
 0.665142603127865,0.5875414527662122,0.7305759059202093,0.6725417562981288,
 0.7268124787406689,0.1554110153320707,0.7396001011036909,0.2920951530197772,
 0.9442439637603466,0.6278400693616032,0.5746049748275328,
 0.016579795400847863,0.19087197261359878,0.8785321972632785,
 0.36817145715275856,0.6644645197300816,0.33581563448579566,
 0.06463506688136855,0.9110978204771993,0.6242394090843029,0.958646932185988,
 0.3102023558361998,0.872792881001708,0.8574295709752882,0.17389124095035224,
 0.19709988451955318,0.3345354572550202,0.053930782638036634,
 0.6423779853402447,0.1672807503990692,0.7238724514649186,0.7913249244224603,
 0.43527508399873727,0.5862056868355734,0.20943604678401195,
 0.9487640834257995,0.47535802810034355,0.12611818064367886,
 0.3043168177157952,0.9043529014262839,0.7384553525989376,0.6789174082486642,
 0.2885194322006879,0.9760124082417037,0.20183906730111245,0.6269359201245127,
 0.0825231636799213,0.34289804445650773,0.7879300915319601,0.7898066423803338,
 0.18298864942028603,0.20057826965846326,0.8776310462352557,
 0.1571939682903578,0.4355043733018049,0.487529033502475,0.27338773133330396,
 0.6489206699837382,0.7026223916380775,0.9232029322816294,0.9879894211146272,
 0.03856952887871323,0.5294654959059705,0.2232324379793046,0.9844565399289754,
 0.8015333498681122,0.026056313330409475,0.9390796367677087,
 0.2863323082360907,0.3879862025991496,0.27260853461147216,0.7049946929302676,
 0.8540145733319582,0.8205682113610042,0.24223770624864693,0.7756883449357956,
 0.1365874954925178,0.6285911366159902,0.8344123650717565,0.9279414547481066,
 0.45812946550354483,0.9963101103222293,0.4936814521364955,
 0.31024225340501577,0.3270778381254491,0.6526926372557724,0.3466718645477762,
 0.9063392857504654,0.025738137368712333,0.5499046342450497,
 0.7395980156415372,0.8194348271554232,0.36988727472105254,0.943636924256059,
 0.18782609140333928,0.3483945891098994,0.2326999236044227,
 0.24423637670098552,0.926216019929044,0.6404453950672594,0.6926484717570608,
 0.049472142404236696,0.3632257308373099,0.0676173920321188,
 0.08147091158029696,0.9303324745528329,0.9694600730675812,0.6619182892017306,
 0.49370411999756547,0.8546802986796278,0.10624311416493304,
 0.2853449288323203,0.22691329336315547,0.9213965016264504,0.8636821723196357,
 0.36531403548132335,0.10993930224770643,0.7849107326599855,
 0.5136724680575779,0.5036719163153335,0.8422583215198652,0.08545441637282691,
 0.5276002392295748,0.6256038434754702,0.8232987339733633,0.4399086626928026,
 0.15629855097506784,0.7310865620371226,0.10570618707589619,
 0.06847823579017309,0.015094840482281047,0.44608849700046393,
 0.03648701730913717,0.727897878428854,0.7923537548118028,0.536721284166182,
 0.4480856567709832,0.7246035044997615,0.2977301447311599,0.8501566513455597,
 0.1761826580503163,0.7794515452088894,0.30718624210080203,0.1777966697831609,
 0.007012138516185651,0.6947559205513688,0.846289084825125,0.8757232106258899,
 0.8213454238342023,0.8643480696493093,0.48280624090705493,0.7897349878106729,
 0.9009211602054545,0.1638826660083016,0.18899442292268487,0.4405747021530616,
 0.7277267954801253,0.06704030965939578,0.9290546178417959,
 0.13236170854458662,0.20207557966974266,0.7494929077857423,
 0.5177668984439754,0.2204796025537239,0.8993802362465788,0.9566575622359899,
 0.2716971888945676,0.7020387480632124,0.7076951465586431,0.6213753581080963,
 0.9482585935770885,0.9652906958401818,0.15362450668995942,0.4924094263090263,
 0.8719307289027051,0.2621369446602122,0.5878800959326095,0.7376171515770441,
 0.19868348119537527,0.579411359625254,0.09685985102310046,
 0.14124060061107302,0.008669775802189195,0.37941406095468655,
 0.0553698890118145,0.3694245667003807,0.4763122281599337,0.483453418230847,
 0.062011952444145146,0.6291246037770335,0.6600250831828143,
 0.23730365873507564,0.2699677486002239,0.759130655970566,0.261823745937944,
 0.941845976826046,0.6689685861883164,0.7843304122663045,0.29175206419945043,
 0.253668611381768,0.4639281800565076,0.9087694513234285,0.7207760349288336,
 0.03346138837184043,0.24668484472209729,0.4855228373503828,
 0.8254892459342618,0.15176915445881445,0.6327897466086044,0.6820937414328526,
 0.6319867651565382,0.5421000080577467,0.6183436344286387,0.7067394355575356,
 0.14943361490787654,0.843234255761538,0.7666078277535291,0.5249951184056361,
 0.8868295535289015,0.4650912263524841,0.8540334084002784,0.9878451794435597,
 0.1624370606478125,0.04037182245069171,0.38767416903666563,
 0.6484721645994053,0.39307391492970556,0.08611187123739217,
 0.8116583935047198,0.7856139755511254,0.8191558917378033,0.2099260395918704,
 0.7623374158390254,0.7333395476594425,0.6887041399644283,0.6899310010608621,
 0.3414799808139942,0.47745197426542063,0.7363053993176926,
 0.24388273566228413,0.3817621988872322,0.3362782121609633,0.9775556834064587,
 0.6894445977466901,0.8093929957108049,0.8250811966258742,0.6007127850300901,
 0.11449700830414655,0.9243061335616682,0.40249131360643997,
 0.8622291772046666,0.819693180120961,0.765020039260912,0.03029374298259646,
 0.7226521251682118,0.09294843491952787,0.8746256708886078,0.8248149520550349,
 0.505466194290443,0.9110446075366352,0.4047684305680842,0.4493080865497875,
 0.6466856023270675,0.4204790860207104,0.3996591556358928,0.4004865697024167,
 0.8711868257147831,0.5407571030495374,0.892298328125603,0.34998547744351094,
 0.5511517915509222,0.5296693875107511,0.7264770275917163,0.4122227483677301,
 0.48765317503792316,0.4811742439969845,0.29920737816871534,
 0.5199071522721989,0.7106581262363335,0.44486385790545857,
 0.41191418230173926,0.6925244963761227,0.7599542802263667,0.7917591463223815,
 0.7949435466649365,0.4923642703187676,0.03304719662485023,0.832554288761502,
 0.3225724405484446,0.39445487533088863,0.3691045353416115,
 0.08712478224515974,0.22065836423521634,0.6169386535153435,
 0.3046099444921596,0.485920462213268,0.40917029029623153,0.4220414488534613,
 0.08932265097648939,0.5161066613192064,0.228934447837015,0.7631466304783694,
 0.9542535946496453,0.7134895042557827,0.15321031023547615,0.5024520318726472,
 0.41839306147149613,0.027638018215979776,0.08740103307466218,
 0.5107544284877312,0.8206762188903102,0.41459446855067916,0.0576486279097711,
 0.38317460444325624,0.48791564747442473,0.2534003545930962,
 0.35655186994135335,0.2994821445042136,0.8301103591041612,0.6330723425795137,
 0.3206669522733918,0.039329136993276004,0.9729348234954796,
 0.1665225339303117,0.44113865358569604,0.9048961680327892,0.5515485308305399,
 0.4873444085311509,0.5677965135312192,0.270435387998883,0.7230729734609381,
 0.4183049667849128,0.5799340504183501,0.6965304502856564,0.7241717662036062,
 0.9787521275225717,0.4552707923717517,0.5029252539658332,0.4197880232112974,
 0.14496887156812643,0.49409184579807286,0.34277675139217434,
 0.5291315311791837,0.09271948020476917,0.7920276308736316,
 0.37521697134905097,0.7082515585319672,0.9707559758591695,0.6430279998832944,
 0.6170511145267042,0.11848622527868002,0.23640700659337543,
 0.6190782181214062,0.9414360632950584,0.8318611279383379,0.8363328570032759,
 0.7820867377689338,0.11633212724649178,0.520738104311419,0.8329986797330062,
 0.44472761947425776,0.8077856313119453,0.22526440947221227,
 0.6435462241051579,0.6760397772844153,0.3913360796519314,0.8763576784950436,
 0.5889564421230291,0.8768996812974631,0.723948844487601,0.20812089946069712,
 0.14862897624155225,0.03269542637430456,0.5030812829398817,
 0.23539680361819937,0.17130864323764872,0.4523841283578032,
 0.9096797125935292,0.07302976447818943,0.16194001068492136,0.60548090582944,
 0.08803169680012535,0.7615428591034916,0.16463072818550983,
 0.01467965307735275,0.9281956953649286,0.2889325958484822,
 0.018960904840226278,0.8416584494629364,0.6333790672032169,
 0.8152136660258904,0.8647695137637402,0.5111407234668685,0.7700067236582783,
 0.9579773449370541,0.08367289473168182,0.8389586059690972,0.6135252325446675,
 0.5829647837725376,0.28342618044437196,0.49378498915028113,0.254503038281752,
 0.8027269684086181,0.5307465495353221,0.5239485240568997,0.25367859517827895,
 0.6094491832763496,0.6065645094192327,0.03193438170246121,0.2748113655695634,
 0.3690147631798687,0.4398179703319013,0.5318804552868671,0.5154829980594712,
 0.31751154036814444,0.8742920132454275,0.8451832910798802,
 0.25872374888767435,0.10248889164613428,0.38332691146602504,
 0.07382979093739639,0.014834940230741056,0.8682156662009747,
 0.30139516993195303,0.6827992310156185,0.2915825038352893,0.4013507890979653,
 0.9003209011287012,0.4967965186370291,0.8541562618625644,0.4854965163311745,
 0.049946203543811185,0.7269935584873478,0.5733655145296963,
 0.4214664897338458,0.62257824090454,0.014074327972684153,0.48757203627523893,
 0.3202897524640773,0.5825260697546146,0.05615716483445976,0.6265332174985081,
 0.11079931918950048,0.1919352060767422,0.23933546749514112,
 0.8144114936460626,0.3499392854218265,0.689192391036981,0.9759384385835712,
 0.7003295024655093,0.8628269024274238,0.8912031881526331,0.5097403406550185,
 0.15490721762713588,0.8632913540741377,0.5180843459761088,0.5183255708836492,
 0.9009463397267147,0.7317620099366948,0.5911536088616274,0.17056617369263727,
 0.7661077288234719,0.06597429393270837,0.19145784461506887,
 0.2757091926984969,0.06160179889291251,0.8174305056585983,0.6953288394984147,
 0.2655320129754687,0.45571099413897753,0.754284162178701,0.2681627419653676,
 0.6881264507588583,0.5017554974231809,0.326289649785376,0.24636033669526292,
 0.8638955888100935,0.30264213215079716,0.7555676390901809,0.190084779480163,
 0.750621554922378,0.3082049813473684,0.4588563032793437,0.03907926496235681,
 0.6176827417446589,0.18942388608747418,0.06899794086454514,
 0.12736879904382026,0.0650611398736155,0.28089970271029496,
 0.8305095173883286,0.9740893244184322,0.5641825875078215,0.18966487463233594,
 0.5707426040699308,0.5047538093683697,0.11187556256455711,0.9674505752064707,
 0.5764219967144266,0.21870689172543867,0.10539283735947613,
 0.28848440035560474,0.2807366330861646,0.8491906782231258,0.6196872313552749,
 0.77421480011647,0.545269278526573,0.5269619950890444,0.8191687146664506,
 0.10400523988492272,0.9956428814642275,0.96360962828482,0.1947826542134128,
 0.4482801854218019,0.6229575024107893,0.03650345343115324,0.8314405011566001,
 0.5409912409599995,0.07591357098238183,0.4876568630333945,0.8390910453100199,
 0.46190334919103315,0.41461092604836236,0.9308408622558328,
 0.4635556545355949,0.23213589193536244,0.661473234137056,0.39761858467659095,
 0.9262446718894117,0.46986114745210383,0.8502146094511576,0.7512554018037026,
 0.9769877519276136,0.5201715755383484,0.5315986960430616,0.6883875695981373,
 0.6227950829193571,0.7072434887506662,0.8343535922596801,0.3057284387398006,
 0.13587338701413365,0.06768944403634658,0.6703809074759889,
 0.25359522001531387,0.031632011250134595,0.8510824558747969,
 0.29016992277244535,0.7769138958289799,0.7179266040623604,
 0.14530768033292707,0.7624079754288308,0.7417901406489265,0.8482632247962746,
 0.08544766329161191,0.5755930144390198,0.7013076644884007,
 0.35400074628131073,0.6385118445535701,0.6477369268990686,0.5101811679538218,
 0.45920730054562287,0.5969855326988375,0.8790631224276966,0.6023320999841268,
 0.3520543271414803,0.143239145781216,0.14491759190474673,0.08673021309438211,
 0.5603998040738193,0.9609328277209628,0.9186648436391674,0.1762615455630856,
 0.6864006189297372,0.47520128800727024,0.973028962378693,0.6971943355634311,
 0.4485389482081056,0.28728174275772445,0.7986616989193189,0.4882959645764523,
 0.8178994285651465,0.005048455132813601,0.8612445237509533,0.647131854258361,
 0.41265936325695984,0.40953060250466966,0.5735993577942136,
 6.074364232115936e-4,0.8760863427511333,0.8431221451058021,
 0.44668681203418203,0.40245322604857736,0.16272593551514158,
 0.9968969384508363,0.0016266899051269323,0.5093867353069622,
 0.7394456511270966,0.29530582014265727,0.2643804030925109,0.8294346830795059,
 0.5456662451072226,0.969032503156237,0.9852918446927179,0.7988133023393259,
 0.42584709218942207,0.01865438563763966,0.9220029669930252,
 0.9520878418552507,0.050281172689688014,0.7067992740391419,
 0.3657048742951692,0.24490650698779282,0.36901747978646315,
 0.10390539953821087,0.89487540462073,0.13905879991237458,0.7731127020200671,
 0.8474310788433496,0.17780596806186388,0.12840021916010824,
 0.7562873793212375,0.8494999407849535,0.3014918308519897,0.01835315252121572,
 0.6089068796535755,0.2024135753578552,0.6497735890544286,0.114408303599154,
 0.5879888991069127,0.35029166919763044,0.6486104071606617,0.5076232057536061,
 0.8572261682290749,0.6900018935873106,0.44850093462053486,0.8877836497673958,
 0.6849807756201878,0.761271351248004,0.21486610197055334,0.04128977218514551,
 0.7197034176526801,0.9284856505088095,0.3996769148917778,0.07772539291875757,
 0.7438981884510805,0.7509040598024531,0.97792682460706,0.3364799125698632,
 0.7623669809594775,0.19998262872702888,0.00884852015751214,
 0.3011032415702064,0.1576135876638569,0.3993098730756436,0.8498559764756523,
 0.48181378638937644,0.40021800944198516,0.8410251017846497,
 0.036276521143110974,0.13502991819355892,0.33100937421350185,
 0.3328944798369713,0.9125461241950825,0.8223094845675663,0.7443227002181227,
 0.2962916000936836,0.26724322667750067,0.021244194815191797,
 0.6060531510527745,0.9344098113906092,0.4460399922788475,0.03920994456757343,
 0.08053871222206521,0.8820983896879471,0.13353270432310904,
 0.04683737202362359,0.2802195759552629,0.8104095378680689,0.7564602318187161,
 0.8574534207914702,0.4720235648077886,0.2219825457589264,0.35839375476757973,
 0.5577464323987403,0.22027448556027596,0.4204501284980998,0.2634412091530671,
 0.9061910103103883,0.6491049820053937,0.701014697841239,0.3508178957932111,
 0.16650079385932504,0.9547376494627343,0.4216133020954549,0.6477794709215317,
 0.7396656482853776,0.1108123145126596,0.329766747814239,0.7424830661220965,
 0.10036937543952074,0.34319307820663136,0.7891203912402649,
 0.21400480965284463,0.7915383412744882,0.9169541211181668,0.9684553156208979,
 0.6706537598086296,0.3286475886106073,0.9677965721327946,0.38588590237361087,
 0.25975218454591775,4.263493900875659e-4,0.5898884110607866,
 0.776239759594386,0.029134754020318177,0.01538347091122938,
 0.9742547567003996,0.08004561575303293,0.27521786469060106,
 0.5426929377097803,0.26844294332064655,0.8780276446540018,0.6327057637364502,
 0.41939666541973863,0.8087147470061423,0.8863271008091702,0.6771932949228966,
 0.18882954386382345,0.5687449077735433,0.5147936203956749,0.8283942156692256,
 0.22234330326336982,0.09521882650160396,0.8696798658775643,
 0.9483222629760725,0.6417290354247177,0.12094195568405064,0.8100739463038986,
 0.8620428304861809,0.2088130320480579,0.7385774115466495,0.4764934583752779,
 0.4050159407111925,0.5471357660782984,0.67317544040143,0.6215142806075742,
 0.7111583703992944,0.6633052288325784,0.6291789853787721,
 0.015520643797303446,0.048876228001821564,0.061436817311421366,
 0.37266696148285705,0.1609008643229236,0.016105555194984134,
 0.6488659348778825,0.7671008754333496,0.815435081855607,0.12344988391133527,
 0.3729240905734654,0.9227077853198518,0.7986554795871572,0.38640886809239305,
 0.5273375078658787,0.2510953846292072,0.07090998296067486,0.8088000677309582,
 0.24266656149746935,0.6796710115773414,0.2700018299559699,
 0.29900226518352857,0.3770476921345707,0.7716868748773731,0.9451372731908776,
 0.8952403502355035,0.32235126468364794,0.18910313920827904,
 0.5080689548711387,0.6365867140901912,0.15238156486166798,
 0.26219054921359575,0.42303302475484716,0.4369135604125771,0.826658155653537,
 0.6527120467751245,0.17283815211704234,0.551995181036313,0.00654969872816169,
 0.6727382131934319,0.018940997112667057,0.07814980471436694,
 0.6056647009491769,0.36668899458657367,0.9600078339034743,0.1587948547618525,
 0.10307958315160626,0.2455311294151883,0.19845014248672666,
 0.6918024061783676,0.7862799109827925,0.7924789261232839,0.29989299996754837,
 0.6370462494678939,0.9081190946405634,0.5729571694283796,0.436250071716594,
 0.7752810314742088,0.4559255886263711,0.5146373884965785,0.94194828654457,
 0.21422738767463867,0.6557412481602062,0.04842323226859391,
 0.5850053677766802,0.04728255429014461,0.3394068471515297,0.3876241122630337,
 0.561624252653704,0.20402196963553965,0.870876867583122,0.623244123673605,
 0.7952386205123766,0.34683596667417016,0.9534872328778847,
 0.15584262112535663,0.8842251932739249,0.22583251592456977,
 0.8040189572101465,0.23812243468203198,0.5485536020583125,0.613184872107328,
 0.8089086189899044,0.5353696680644635,0.6305308920184806,0.12586715129623038,
 0.2488993391468306,0.023076893182854108,0.24648381629543514,
 0.8583901781109473,0.21707353412139174,0.10988886615614812,
 0.15325890238637097,0.6680741455573656,0.023807709137959376,
 0.5734335924477328,0.505685234410633,0.5216982085839552,0.7347826571775133,
 0.93089558266802,0.83242628089182,0.82104002342645,0.5756870073395928,
 0.1461705784260685,0.15027069677138571,0.942037264333685,0.1935603691921325,
 0.4170220205192253,0.63590173931931,0.713161682919437,0.7899738434749117,
 0.854562446103915,0.42700956750512853,0.8673375923114329,0.5014570968703587,
 0.9619312847456047,0.9617783191096518,0.6257751610718537,0.09952227233106059,
 0.2566690941411286,0.10003749884059243,0.48544841301326,0.8883296094013692,
 0.38943124007330543,0.8644897864587787,0.5105480799334904,
 0.30280225599575605,0.945011480700596,0.40019442657809634,0.9347979936107269,
 0.3561045860350329,0.45933157365033916,0.8702102679897203,0.9629192123482198,
 0.27412951403927155,0.5992892019699207,0.8975317184593402,0.9820247360884827,
 0.04810964245720062,0.9334497405046903,0.3430596473688904,0.8239841305555462,
 0.034944910078356206,0.3281909366213238,0.197636519272941,0.2928080236541133,
 0.1889325028279596,0.7291845315509654,0.6282065158555178,0.8528459456342453,
 0.33629424955518417,0.49935352768163377,0.6801531285667155,
 0.7271248187599718,0.949953366014338,0.6798732099798463,0.3112230112456019,
 0.48341781969359154,0.3387784456748353,0.637720482242791,0.6230560316967046,
 0.12837858697363058,0.22327373492652025,0.5135016913659951,
 0.6655194495458611,0.24030290486181394,0.6052231495107423,
 0.08537756412125497,0.04286278822138545,0.4155307030870812,
 0.9405424703366235,0.21285248514808774,0.7406179337783079,0.8100667403452967,
 0.5433700091533126,0.26937677985224584,0.43921839520073436,
 0.06550545501536997,0.8980313363460737,0.4908276876601886,
 0.21699139391001188,0.7769684011119676,0.818463872399132,0.26442734464050677,
 0.3377769962168642,0.6111806973404549,0.17332919401775781,0.531832087433917,
 0.5934636780222613,0.21358630139405455,0.688950524075268,0.49417482042379945,
 0.34550760350683973,0.6589693033080443,0.8721956324352417,0.6003060278538087,
 0.6973668156104809,0.15732627491359574,0.6552014929569951,
 0.016113318330802318,0.4922233619808837,0.9108691733234033,
 0.7948193532257072,0.4771262286692902,0.9369021033206931,0.9310749918918599,
 0.8228318960879559,0.3659654194689216,0.2634315075323598,0.9421080840137609,
 0.2558238758561294,0.4059901350359103,0.4459103692828482,0.19062115282326575,
 0.5249268314268265,0.4973786778299929,0.1163963864393951,0.45223746915678864,
 0.8999764121477516,0.20069494045040404,0.6470034826282374,0.7320763771790504,
 0.16517860799428552,0.8220534209955285,0.540562250138938,0.3518828272293578,
 0.49222235306594075,0.5173261383689882,0.27879616338135216,
 0.1837638541819797,0.4485179896592212,0.5558083502102191,0.42713924460080954,
 0.5977733817487139,0.49999883064235473,0.22290512348788383,
 0.4143412007243409,0.3087684370671634,0.9440412083390884,0.46546848521272044,
 0.10011181859861695,0.8980907073555113,0.5556226846832691,0.9927135658384281,
 0.06645125793820572,0.5695778734537197,0.3031210410690516,
 0.43386901657280474,0.19170107331088082,0.23292032914056848,
 0.23845952132412263,0.37683878941785276,0.6184938112705988,
 0.12354665128054754,0.5289580092678113,0.3404102531518819,0.3011947589900509,
 0.0747852770394063,0.3024279233535183,0.13430895187340353,0.9397859001033443,
 0.7558995141529069,0.4258589205994685,0.5727821973644014,0.5910028923754391,
 0.11461785873346708,0.18115909895562654,0.7064882140546446,
 0.6170634719395338,0.06609405825837822,0.16093940175260946,
 0.8550196383797526,0.31862590191144513,0.13652262123019232,
 0.42640332073715936,0.1489557312580636,0.12687496120617914,
 0.2664001611713416,0.4406705082578315,0.18120694106134472,
 0.18955634694070023,0.9732551225490909,0.316293405208926,0.8389462555355541,
 0.913192228852018,0.37061372892675815,0.5789369808226941,0.1574555219673035,
 0.07426626989035778,0.5102522064222627,0.3709567997907004,0.7042827999242668,
 0.5203554645928743,0.17244882229472203,0.655630575165438,0.5678915800638217,
 0.05226354676421596,0.33111085742266566,0.10264584778956776,
 0.20628596978686875,0.8530338548789672,0.4954443941025365,0.6664296691632063,
 0.9225214966641897,0.9009831118487669,0.3431039584009491,0.8901930101514826,
 0.47937428898584133,0.5086962799637667,0.5798459238495177,0.9906978725584246,
 0.02464962980545904,0.982944579035432,0.004595694974786735,
 0.6440359097009994,0.3883306012826645,0.5448218919887533,0.02598857342318306,
 0.42759379568798417,0.0014859130231942297,0.5103796690270525,
 0.2427482013755886,0.8363741750697214,0.22676246611839768,0.5083620611315367,
 0.24673719528403693,0.7546213634424397,0.6400034854019983,0.9612684856457614,
 0.3408793927547632,0.4728798033144015,0.4025045998890924,0.29554674116992996,
 0.3780316757984634,0.7498714432163911,0.0453815728606457,0.10038979721134156,
 0.28395978260657084,0.24635116494281228,0.8982447149780834,
 0.6192132744898337,0.8038460545232676,0.45481286002432286,0.7747178023725322,
 0.1736248813839788,0.2061410959374661,0.029359131800926974,
 0.6542222305996335,0.7904050184223589,0.03300445620013659,
 0.15807472426968672,0.7598999943991865,0.4665183945997017,
 0.24314709177563865,0.5409393971484673,0.17780150824528906,
 0.9651685072322804,0.730061943937433,0.7030676397492653,0.11224896867827494,
 0.1350759938668885,0.3356884434350379,0.925616240614561,0.06473647845435382,
 0.11087680144609546,0.17164754247693204,0.5693890976203212,
 0.6594628950344172,0.6194232721611559,0.3751183936909883,0.9714502561792702,
 0.46557435171227657,0.04172038688494273,0.7355833053929791,
 0.47008724223819576,0.22524602287899687,0.7593354587058435,
 0.3314819366350026,0.6981804805003716,0.5679382209911575,0.6104635990102363,
 0.940545901287309,0.10122345055773829,0.37968623865041184,
 0.28079892545906204,0.8108761275894134,0.28482823918369027,
 0.4215765235767538,0.5593241141686599,0.38065953165625865,0.9170265800597274,
 0.15680937053616328,0.3078446295652897,0.47896289457486374,
 0.04519552693658624,0.3916278820589706,0.05097616227462021,
 0.14288847970466845,0.502046303624869,0.06023548401598455,0.9459187846072338,
 0.15739976267615408,0.20891721691869403,0.4375071894377891,
 0.6905893087170364,0.615746162053995,0.03734962950803189,0.0539245086113036,
 0.4915004105260986,0.3435175095113544,0.8635807399132186,0.5002221152797393,
 0.0740095592158947,0.1687802583335557,0.8851739825655308,0.24069449502709384,
 0.7339482171838739,0.17845089470592734,0.33054273940215784,
 0.2547074268726357,0.8548302979974322,0.783706979686606,0.955272283258613,
 0.08629976978206366,0.06255076618840427,0.4602385720785971,
 0.8863263385274035,0.9518820725139442,0.46401947941633503,
 0.16942372519837234,0.8782922182429763,0.7080765380126885,0.3838592203529313,
 0.07405633070544604,0.23838475442865892,0.29020446008537437,
 0.7487930396327082,0.6013496740897176,0.6187107011633227,0.8348997548353074,
 0.7067139615625867,0.16815235965998,0.939924923995167,0.23884459996744878,
 0.4593200521241424,0.6716542444411339,0.6734987296244765,0.494853941443296,
 0.7581833291017802,0.16580438114231577,0.22735714348862845,
 0.8846242463057981,0.6247734933678175,0.7066674789838893,0.8994106957932745,
 0.7855627827509538,0.5307021550972602,0.040361008185353375,
 0.8803811219387816,0.45427762832783236,0.33233559187729056,
 0.5712340181950349,0.9839067695180674,0.6946619389776542,0.5341460319354857,
 0.2403191941699565,0.38379768522482915,0.43456607363047006,
 0.001203897667153453,0.6440058183415258,0.43127010817145495,
 0.871009883921926,0.42590062229853176,0.6448066090196656,0.542238125790183,
 0.9842386185431852,0.43016142189209217,0.7893514561887413,
 0.12829864603326901,0.04491647600766491,0.0012693447988290085,
 0.3137106538459641,0.27710779367051275,0.054031146857912105,
 0.5751711768512769,0.8388567048843923,0.35488671080552925,0.3544901876520211,
 0.33688874209535824,0.1267482365928866,0.9239764379054194,0.2844082281419259,
 0.7431311386571854,0.1539707895881538,0.5648205842074943,0.7753521243720312,
 0.001447379394746151,0.7135799624031326,0.6288554878199827,
 0.18924184084266749,0.40026521384721037,0.7042982708713318,
 0.7632562326411217,0.9471140959040125,0.4608830261873932,0.27220525985762467,
 0.12169640217917677,0.22890496318196585,0.9861507394327449,
 0.7562830475993357,0.5648132271471191,0.7897989482679507,0.98668208719145,
 0.9773353909883187,0.5311315041754713,0.14340954615424995,0.733748093409421,
 0.17401507091513646,0.9485263515124345,0.9321619629227541,0.8261006225633241,
 0.43131005531771005,0.1268402415503439,0.8043010862611524,0.594988674612703,
 0.572641892291822,0.1948173901020902,0.504900966477578,0.9356626002480984,
 0.4739633543382128,0.8706707521147965,0.6136603628452091,0.6618693846758317,
 0.4067684010834265,0.9650868518955006,0.16762808922972294,0.5598479654892561,
 0.4470059431935074,0.8096397453497659,0.7302753622231223,0.3396182941657677,
 0.17284029318785332,0.5995449422044928,0.34540717621009587,
 0.7234083689553046,0.0287161379094244,0.40065778563586374,
 0.08297217987559546,0.8593845513750242,0.041595380422029704,
 0.7883366690743492,0.2646513584468556,0.4481335228371206,0.6877422946409586,
 0.9697842763450565,0.7122827938475171,0.9049367402903985,0.4786311826849926,
 0.565126386344114,0.5527479745140944,0.518888014634744,0.8600645188569765,
 0.7867402860950757,0.8290622725141196,0.5451828551333743,0.15725223419286927,
 0.3480474620813734,0.7000969543748394,0.24561004372385553,
 0.35796226282052357,0.9733687808966824,0.6654885798088481,
 0.033577866975751824,0.1493813531704843,0.2806286541664953,
 0.9838251119330693,0.3666310451258067,0.7862348834404287,0.9958128795537473,
 0.528839590269272,0.5294133920362513,0.7938516118366623,0.9452218678852856,
 0.5982191986004329,0.7509825161398767,0.6054734041434597,0.4216905616240313,
 0.8049620617886423,0.8538165268893239,0.6756165099443161,0.9978753003215598,
 0.832593618103785,0.12031857558853931,0.2503146458461988,0.38233204654868935,
 0.4057028181041934,0.1650101571219642,0.8731172215577706,0.42568804548189054,
 0.30668039741070763,0.22184377356004092,0.9626686254367424,
 0.4991864664636698,0.15323682712984343,0.28309376280029563,
 0.9846393192569226,0.7895593412520341,0.46297008655094596,
 0.001885207455746496,0.6330213173567533,0.8489301484733125,
 0.4111049552799332,0.33821565863590586,0.8154729707945685,
 0.35713531038435486,0.09772523187917914,0.8863243488953889,
 0.6669537333764277,0.02483383303394171,0.9053392982402365,0.8663943464614373,
 0.04845889646176471,0.7356892052076074,0.2678437844283029,0.9764817101012345,
 0.637674561445437,0.3453825415437036,0.7260378199021857,0.9682212590976366,
 0.8463946405990693,0.27626370748397955,0.13889949813044966,
 0.24812341881418576,0.992450193610721,0.032169218247763864,
 0.3944002586050238,0.3083980341359771,0.9699276797774958,0.17799225379828865,
 0.10591706480292118,0.87913777515477,0.11949548549276345,0.6814073285843896,
 0.35386680248529545,0.754508700672872,0.9737875217991481,0.24974578984664197,
 0.10947741791985,0.29418944483879694,0.4481866551005722,0.5986503729044013,
 0.5696185453302005,0.8819183126729404,0.2956488712435531,0.6806393857833556,
 0.08447202261420772,0.6410276615318753,0.3195409387321094,0.7870781593236543,
 0.21659460126574004,0.06484944898852607,0.17442626426740238,
 0.6959967081983552,0.7315927195629613,0.7207024220276607,0.4402783661009533,
 0.5100672850518854,0.7087711527454591,0.5612934035205219,0.8471473005799328,
 0.48558680470515936,0.39620998570419785,0.575494780285941,0.6218178757369737,
 0.2934076716681834,0.18981026002204837,0.1261235258066169,
 0.24717454200261346,0.2867083352351567,0.4054118279154656,0.3833882664674799,
 0.08279781668834219,0.8370263557746187,0.015008678143701326,
 0.37049434162322503,0.7943385296550909,0.7154710822429701,0.2281709043639687,
 0.4483189224466756,0.9429313420945961,0.8585407418130724,0.3302251579583121,
 0.584395995138923,0.8030353143751786,0.7801295488519715,0.8551467123265857,
 0.20309591866812582,0.7455200650391812,0.449406637212343,0.608873012623746,
 0.7983671689013127,0.8551021013769112,0.541492553181465,0.07467135884973619,
 0.8240976845114368,0.85989975431479,0.06243070865259781,0.6728372385397593,
 0.7005648834241811,0.012046978109544249,0.01875226270064534,
 0.3982066933830113,0.5946579440959552,0.6562092133063812,0.27768947177325953,
 0.2025153582304845,0.30521334195507266,0.5547370836661223,
 0.38942454389062686,0.9569882424691912,0.3834572674288822,0.4382501571699132,
 0.35305645017270004,0.5498731087737009,0.5669090820403364,
 0.13645903432864126,0.1754844340367786,0.8944864088806187,
 0.17765838180630833,0.8277724648619018,0.9965039436470506,0.944703385420503,
 0.8555214881127201,0.13743436201735804,0.6812176320484625,
 0.14088797209742177,0.10403240636252686,0.6562943529572217,
 0.9471659634397859,0.43448879751797975,0.24675766764491913,
 0.29571161813187663,0.5880408640362659,0.33377532396763343,
 0.5628449649682993,0.41052645299926305,0.22962149651247132,
 0.27362149715524764,0.8598419876303638,0.6552456458838372,0.7101883407500148,
 0.11525865997611806,0.7683654527292945,0.8818230120077274,0.7872522261963047,
 0.05623886628987185,0.5666217691946492,0.7924771819336867,0.9433353239166739,
 0.6924684168625578,0.8550242155148392,0.7591371406110589,0.3502092269924514,
 0.14647695675027272,0.3662175358122195,0.5906532428113602,0.8695203790138916,
 0.34914003319552944,0.07239274465016576,0.02787682112047185,
 0.031176347202231502,0.1784551655170301,0.7977341184326001,
 0.9522239672235548,0.6774092363987846,0.14032995742521692,0.8935941453274762,
 0.1133592136489171,0.03951697214685179,0.12619052649319007,
 0.22268717800163962,0.5671896505341192,0.372632714102742,0.8150531608102193,
 0.2695904316811484,0.8425895479954103,0.7030390448074987,0.5581920931189628,
 0.8877793747728671,0.6433703234896637,0.9967634236468741,0.06388888851306251,
 0.4834753104442695,0.6213720815320577,0.8661544719188121,0.4459139960139824,
 0.5207008019750399,0.7380062861095024,0.6489312353046396,0.6869747049857163,
 0.23401533831117904,0.7991129888275034,0.3013077858901123,
 0.43565784047162437,0.6383153582901447,0.9119815772839739,
 0.43524655301681336,0.8261832100452224,0.6906593413462955,0.8511210976653166,
 0.13191409119859343,0.7413205615289659,0.07357634958270576,
 0.32592200060433807,0.7905460998470724,0.44223804050820126,
 0.7312868675632161,0.5646583196579527,0.7560068949081115,0.3994844751351325,
 0.20027936068288443,0.8113605525677943,0.8836114773343932,
 0.18679673632158156,0.9868721902067619,0.8522651894792437,0.2673346046499052,
 0.5637968363314976,0.22016829549807737,0.22771020616404802,
 0.5357454685511346,0.526642817339469,0.6004203919107133,0.3450496951489932,
 0.4945535038594745,0.010166144187808812,0.7428883418503107,
 0.25535280378659386,0.5003752437880155,0.4409556259174774,0.1936811616955859,
 0.8906380915262729,0.38517708219376423,0.5903687737020396,
 0.12093206772677112,0.6424775366412415,0.8021431073467863,
 0.17822912802484914,0.9657064893972646,0.9487267264851058,0.9015232006882961,
 0.2767723499360437,0.708179695072912,0.15640736704805547,0.9876948316033526,
 0.6777301330724238,0.7903568021366689,0.7592914924381402,0.9967503872473673,
 0.1901710554243199,0.3139149418196888,0.44810661324499357,0.6589556225285439,
 0.07181051517531134,0.5575736507564402,0.09022738073501224,
 0.01373558379909312,0.16237466559153113,0.7331855603630036,
 0.45646719654292034,0.9372258683097495,0.964672464410179,0.33486858838741607,
 0.49188651300527797,0.7796031817537554,0.49656799702512777,
 0.9899395910509434,0.4763360247257845,0.5561518525726875,0.11830448337644173,
 0.5242435931816595,0.33280278576546185,0.8683965013405484,0.8451481714214459,
 0.3688985441943724,0.3996362458144207,0.0956173806956837,0.8230258461900914,
 0.8722648555248201,0.9856959326436724,0.6885358998200604,0.32116866085481655,
 0.40246259974118437,0.900814157931415,0.601112715862675,0.8647707785069425,
 0.9997918976807517,0.8785908538186469,0.16368511613614456,0.5061963904722803,
 0.8011500229636025,0.23733180389228037,0.7865911791558361,
 0.46150801277357945,0.46977047892412205,0.6650343073520031,
 0.16766967534797506,0.1512983122445316,0.44774871648254244,
 0.4857075161492693,0.915244670071142,0.06980338492643168,0.5427625417429458,
 0.44671434533819054,0.6973461105668392,0.07752239542227646,0.754850716498403,
 0.0015297622353122042,0.9158168505992541,0.9377802477424915,
 0.42446502310100576,0.6642186833817991,0.3970392579935138,0.7744598069226045,
 0.12452929930151446,0.006245235315202824,0.939377966818072,
 0.9039400578386046,0.41530052899964554,0.5515251986688591,0.5134827064110546,
 0.8975058778910998,0.6065917808657132,0.8316065861966694,0.9718807067396208,
 0.37353682183576997,0.15837718434690695,0.5297083641529738,
 0.9718616612351754,0.8984350364911203,0.42476223088910514,0.5553605709912636,
 0.16415474749527992,0.30783357473097717,0.6783752905765329,0.497771494879538,
 0.17114825997799699,0.12227044726499459,0.12945684397737756,
 0.9471148610428484,0.8302399735276716,0.9991546207567031,0.784946986237897,
 0.7242519422061776,0.13952354170199222,0.5557255633688666,0.9750022771270662,
 0.03683400184151753,0.6641206352417426,0.4670815126792317,0.9144556811308528,
 0.030024774639079044,0.4905405990290348,0.03318304390366433,
 0.23372458242878802,0.6665801310694205,0.6814739645777248,
 0.12730679167889247,0.26767058205274386,0.20280269451128885,
 0.027225337866493282,0.4533133022056193,0.12481224626391374,
 0.6314918520719401,0.802381097714036,0.9785032544593442,0.6147492547359451,
 0.0645208196724193,0.9357536075411172,0.7750166607040466,0.6223768280181502,
 0.05946812828970938,0.05819238753381073,0.1565722158725592,
 0.8443569366342628,0.3937627808871179,0.044037901911467725,
 0.3618259969758715,0.0919403372994213,0.40107087369213335,0.5957664623285377,
 0.10136431532829959,0.3299388494238892,0.40356969308500035,
 0.9135613275198511,0.33193777616382936,0.9699503645054643,0.1170661127267848,
 0.06055185804958774,0.1084958620653651,0.03553182469159655,
 0.7986768412383135,0.2865494374218167,0.7572081297728064,0.6391960144745857,
 0.5223266778022564,0.5036997262246194,0.3421430804227582,0.9745722621983097,
 0.3513677812571996,0.078521629440931,0.6386552923635888,0.21505874098301614,
 0.5763285798112538,0.07038587570233468,0.8739275420270285,0.6770497396545244,
 0.37611220701771875,0.3344050077884715,0.9129831433299636,0.2921622562008074,
 0.3520914992281631,0.7504615006442822,0.011414974990380578,
 0.6462548062938059,0.6436101241286736,0.6415431015174462,0.3457206037989691,
 0.06537477895411681,0.8057797087016328,0.3470080621954994,
 0.23953916260552566,0.8686416707041748,0.8068374971321021,0.9828145674276348,
 0.5915307381583687,0.2260101110689835,0.41432149251639816,0.7707237718498238,
 0.3074020251444277,0.14190146608538456,0.03830959774957754,
 0.6424141142877174,0.6786213467393419,0.5700643537727282,0.18458816023384927,
 0.12080778245288615,0.033717732269971656,0.1302556249772291,
 0.28157598025647856,0.5789843516308604,0.7350040174780502,
 0.49006667443138874,0.6510851663147968,0.04248462042119727,
 0.23741824618701135,0.6765108909148628,0.5804661991650797,0.4153986828234464,
 0.33753589936110273,0.09338106676922919,0.37651960397931183,
 0.5614532694547816,0.5386946983089118,0.0028369121773652584,
 0.3824439749302959,0.13254880978141603,0.5175864255124952,
 0.22117040437530044,0.8190529519892542,0.4365876568283579,0.8819985974901824,
 0.556156860386097,0.5828387140748305,0.5331111725864124,0.04187552714627496,
 0.3313312057001805,0.10647091529713448,0.7105234058677654,0.2960449266521996,
 0.6614069697432946,0.8392294526880536,0.891016350897359,0.25948627932210777,
 0.7060009328330537,0.6892984401042523,0.7795235033760025,0.3453396994264535,
 0.277469192276953,0.5046805544299184,0.6363430758074766,0.03880039663830381,
 0.12698575544246615,0.28159364393363606,0.4503831038037198,
 0.13932563817178734,0.8013083058156754,0.6214180410721708,0.6706698669547319,
 0.7140330429683522,0.9969698800630922,0.11066654781307483,
 0.14930420716100512,0.14257527127880643,0.4212795373021845,
 0.5871617730972386,0.1871624055272071,0.8528473899092922,0.2684231585064586,
 0.32647686757614247,0.9287018673295561,0.8410985858032043,
 0.31538709417830546,0.8151763888426827,0.727121655351843,0.43565519117937956,
 0.3803399576433436,0.21951071161103364,0.22777060760937562,
 0.7811592752403845,0.9109077727357853,0.8738228638936539,0.07876884023279163,
 0.5161088409066912,0.17873078542766907,0.1496408254005892,
 0.12111404549486693,0.7385292513155423,0.32137461803278766,
 0.9803931500530947,0.07526754242987388,0.5284489472318511,0.5387601732468501,
 0.9363981094522481,0.4410366249918227,0.7436167582777127,0.05543882641236342,
 0.13311924573507206,0.9277310331110524,0.5203140104744378,
 0.21519868217438765,0.4922399005908116,0.9520916741739506,0.6729576799350041,
 0.883019465243723,0.704600610947587,0.4379151459602796,0.323243731064506,
 0.4869315405098269,0.14398566696333504,0.7380740107690789,0.4372851691658457,
 0.4545790618531692,0.029070788321848617,0.7081829705889807,
 0.7529010119882901,0.028445942766929133,0.5094687888516662,
 0.7647689536500574,0.786832290524554,0.8951712675669389,0.2874603467409065,
 0.19148577577255266,0.3709858441010394,0.7701171523239159,0.5870331829601415,
 0.9248882644651497,0.9449106578467865,0.7973475767771757,0.01451017989100789,
 0.17936865741186492,0.37467166833324284,0.3329952544875896,
 0.4139883789540907,0.49002464040592075,0.18684166914032863,
 0.5108986086563497,0.840407304482871,0.7880036262278259,0.570284621867116,
 0.14195794281577456,0.19945418596813458,0.37480798465956466,
 0.6846834955589086,0.5178975650232291,0.5522827971906081,0.793118847270539,
 0.32666916014426095,0.35224404022157507,0.5191596870143438,
 0.3325480228966038,0.6887490507376579,0.7078831637384252,0.9044137848750908,
 0.9805757679594622,0.47331241934198287,0.284357279165388,0.12437011105537665,
 0.18736212526089613,0.5061637269583898,0.597494261425778,0.7976654349017372,
 0.6887683729298737,0.8633924594711446,0.955297834580087,0.42155215036800786,
 0.43409938994578845,0.674594633389047,0.012473662412269548,
 0.6381310966701337,0.5156358042090577,0.9751307837027123,0.5939194624034084,
 0.2602502259715418,0.7961698079559293,0.16890639686662368,0.7347488722669496,
 0.12824194741552408,0.04132576256852116,0.45978195354224927,
 0.10826542509385217,0.18807416285578626,0.7063067618052523,
 0.3626102202056388,0.8174820290424424,0.2993990745880538,0.25923138254490874,
 0.9986891346652811,0.2106529031551767,0.9342727265437338,0.7384194684850502,
 0.6082989835218188,0.7632201526706652,0.6275009171175092,0.8924283864740591,
 0.8062645147381355,0.2127819539397412,0.5462012064696472,0.40834958866653326,
 0.4689654069120023,0.1846854416633783,0.7858191608647307,0.20191174318150362,
 0.1427161215437347,0.27748198391089496,0.469054069268266,0.36632963005213337,
 0.5206059260683685,0.32590533499028873,0.27638004077325373,
 0.4586887259410184,0.07156417158559103,0.49154896299579076,
 0.18258380177432387,0.3798238090242909,0.27400978792026875,
 0.5510226224740084,0.8807591313716139,0.9230725104694335,0.960235962806588,
 0.7197118226370178,0.17930722220044015,0.666526294758397,0.965162877047729,
 0.21548226382736613,0.8086047441673414,0.17093368826730637,
 0.7078483929505266,0.41884559277259537,0.5713606937839997,0.8924839720329772,
 0.997071685409812,0.45482927382328886,0.464226301378449,0.20724817157672892,
 0.6433400415178842,0.7306900486442298,0.5387446678459431,0.4620275075162775,
 0.7257649447480788,0.013356601901068244,0.6166211436350273,
 0.7252109665465363,0.28591605311141866,0.8827977309679648,0.50987529188737,
 0.2912247763841367,0.7742672765574085,0.5275989272512559,0.22617913204425055,
 0.7981047209180269,0.43902234522868,0.2951223018709679,0.3454061082379519,
 0.13705585692387479,0.7074950411034725,0.9056936937157534,0.4413579125979483,
 0.5645351110864953,0.4453212500841707,0.2583898924025998,0.7646676234678405,
 0.15508069280167003,0.40429470369319787,0.4936280166730759,
 0.35979761286013556,0.021402834864139075,0.7521628573314225,
 0.6866081827853039,0.172734873209465,0.8407870134105608,0.8952851778179785,
 0.8689487395472,0.904191459286029,0.8207646831041348,0.1830035584620796,
 0.3806128511220308,0.5000492863846003,0.8690030040789328,0.07374125128762643,
 0.11278213206219512,0.29652955293642613,0.5917114146935754,
 0.8929190815622826,0.8528077905165782,0.8180006963720166,0.7547961993828193,
 0.7690771736623184,0.06288190258387527,0.35915068981285936,
 0.39836646285244526,0.35666589434592144,0.16694871507177855,
 0.7381883750591381,0.39657497380485174,0.35673565835623855,
 0.6587046554306965,0.11711511299673283,0.7622754012803936,
 0.19434311235198976,0.7202363455120245,0.37425575967671987,
 0.2818455921890496,0.5853708714021536,0.3376805362768247,0.11124186823741933,
 0.5362674629012466,0.8298364628362195,0.3218829691652223,0.5226841628327746,
 0.056549485005344735,0.4689925918400262,0.06542517839128414,
 0.7505721705012605,0.9825557821001195,0.7385687623942068,0.9882216844113233,
 0.15722581072243658,0.3837688130467296,0.8580030675863601,0.5252843711231346,
 0.23373038219471676,0.441994929986366,0.3558087859076081,0.904757052534921,
 0.0372245317055937,0.5529567394782818,0.24729594132659838,
 0.33821345413941695,0.4691856499739919,0.7544965116835779,0.448509973602203,
 0.9719621881718373,0.001157265509250549,0.5936029993143926,
 0.6341469423364167,0.42430566475430753,0.03170759815271762,
 0.05098939077777431,0.1347164398473084,0.09929226309732486,
 0.9383601823454814,0.8707666268205042,0.4288389383391231,
 0.039918316793518316,0.7287746196032976,0.8819043129624176,
 0.13175446193507812,0.4869843867841438,0.4395543126278221,0.601801943010145,
 0.7741848378095658,0.7341950177222971,0.1531912160232568,0.7466895114427844,
 0.5185305067226265,0.26973868873198004,0.34127823831489246,
 0.6633988113995527,0.9855318021860598,0.895934053093428,0.9097182617150059,
 0.9206673518088901,0.11972142199435931,0.4257015753582837,0.7313980144301025,
 0.8513864289732875,0.6755013793878449,0.7245063793596688,0.914909969233303,
 0.976446338259763,0.726970825902672,0.14996882465644756,0.014052878673660563,
 0.10539809245626008,0.9223840595932508,0.40247568629753394,
 0.47286217195058966,0.9910195378498777,0.7726385153870444,
 0.10077045484177871,0.8201076295077833,0.5945724682869726,0.2587040201656293,
 0.8984532712964424,0.16441576503865862,0.885294739384097,0.9333277486355289,
 0.9591511320918553,0.3774073330750888,0.6730762741943749,0.4614335991703391,
 0.43885913308669744,0.7374444198939805,0.5329640140662151,0.8972372770235801,
 0.3232187119738654,0.926489917607542,0.5973432542138887,0.9958539050203135,
 0.807261347566292,0.6545123599279826,0.5020570849825843,0.29586687365361053,
 0.2266271695319131,0.9390223868558518,0.35137512100417845,0.6214495404202892,
 0.15044175146614092,0.45985337202854115,0.6338293183659129,
 0.7397058782311797,0.7759384979272869,0.7068263590658262,0.9796091705411274,
 0.834554205213819,0.12786146813471677,0.6558709192662204,0.9661088859158284,
 0.8168637631412565,0.5520927112828444,0.8007084815069878,0.5024710733113483,
 0.2728392275293443,0.8719382199007165,0.5926472293712199,0.23084062843743813,
 0.6201128140032964,0.36121477822330106,0.6615729806681034,0.4753205027969055,
 0.7002788030284243,0.5268239263463927,0.7868934352472841,0.911793488183923,
 0.9654062742827864,0.16394280562137875,0.10535369198732925,
 0.8978757055646119,0.7036979600407205,0.4797849315358108,0.8402898913604472,
 0.18549824545319993,0.9265696004977542,0.3125015877530979,0.213485548855885,
 0.0671110361692897,0.6752303669124678,0.1163540845035026,0.7418413873813205,
 0.6901622163512425,0.8769337882277183,0.15807749578777963,0.9671591196090645,
 0.19971142687095156,0.6285933375557833,0.8200448460448151,
 0.19515656309845464,0.040128302412113404,0.7862713934263412,
 0.47553919343395934,0.8523176927743369,0.8688130991759114,0.892169141702868,
 0.12722932149363708,0.3834885309613434,0.47922911075594876,
 0.6981341519294717,0.44651791527702,0.258796233474451,0.7929600004819679,
 0.12722979575408666,0.7788542760217224,0.027655384846651643,
 0.1223435123595068,0.20043217099860566,0.13528028552380178,
 0.8200764884537122,0.5706154861196037,0.6146970569756361,0.4708908185226687,
 0.21820042817895768,0.4809455131179202,0.14005285570218395,
 0.5721438945671625,0.915690744735341,0.026713017490403668,0.961649879617454,
 0.44224546298977585,0.6897557812866826,0.7386408420736859,0.5688311185867764,
 0.44811099087155093,0.9967588668392711,0.3690122736573397,
 0.0044280603806030605,0.03308451244679522,0.21731265308487002,
 0.43903240803805543,0.9340542100145128,0.42717973665463715,
 0.4375343218445247,0.38523048296455,0.07446490746308365,0.7798439876045851,
 0.4657673880177765,0.8652466074006124,0.16231664003346835,0.4799249586194121,
 0.8673801930615783,0.6492810652480047,0.2652470774624942,0.2774609274392368,
 0.5903397168060238,0.45175107257373825,0.7677952828044996,
 0.16957992162440183,0.4209521759985646,0.2222177053894726,0.7545083779427931,
 0.4322106909749308,0.9217062868159532,0.4923476794856977,0.4735521242507441,
 0.5139777441881404,0.9874392194659848,0.4722697969705367,0.08751289804221418,
 0.47873141022018295,0.27840190671319087,0.2215153265435834,
 0.8837013015741633,0.7513683593844493,0.9401005294818104,0.9683629890558648,
 0.205463734485255,0.13108323482732742,0.5234360080951087,0.33244816594275206,
 0.3417096783184733,0.5859971410812941,0.3107568099149304,
 0.008431467169912499,0.5205772855597016,0.08421904558591031,
 0.5602160632099434,0.0146040622630208,0.581932621087429,0.18601209550203368,
 0.00569471800681054,0.061661417565518684,0.039756144958458695,
 0.08577695777653516,0.9436938209537585,0.8194631740821527,
 0.019630498962479703,0.7407178069908401,0.36710054555697846,
 0.03137676266510536,0.9177250987476084,0.7577633911589093,0.531911130661317,
 0.86120434871476,0.5805056154123155,0.7675327335331746,0.15402531046489165,
 0.24780803208532054,0.9482847172615247,0.6943141012812194,0.5591226603808737,
 0.397859651091523,0.33372128848636085,0.7882932518446786,
 0.009789611880896332,0.7070789150264138,0.22802678309596458,
 0.9831934246525932,0.9454354750239397,0.6119664959655653,0.37699735661297085,
 0.5102213621796161,0.11597179608724795,0.9033589542921646,
 0.36064722536616156,0.18949001710911884,0.9298658569544577,
 0.9207336886022202,0.9742595324543776,0.4231299818659301,0.33697422566503643,
 0.2631264124738174,0.6167117325675969,0.4775482166381201,0.21397537092775842,
 0.5154718614436051,0.7895707305099445,0.2463098064165299,0.22993953346886808,
 0.29291630004925917,0.2576868968873688,0.5171179210526743,0.7387371605252997,
 0.2127110334305362,0.23543346374031748,0.493444938227793,0.8162720760861698,
 0.7151700655976825,0.35297327484716323,0.9049350163995261,0.507122774159352,
 0.6592643573480305,0.7758565709853844,0.7700482315272079,0.10237601473950009,
 0.8333623958580465,0.6642300455822124,0.035397006542704834,
 0.43717156063359386,0.5865989619615997,0.6423972048963171,0.3713802749404779,
 0.5523622703268293,0.06478976448983165,0.319537584544762,0.4671694859665739,
 0.11921182424595722,0.17056610230025093,0.64894503270223,0.22335747214605095,
 0.4040711248441201,0.5904948187837488,0.8345944561012172,0.8403827797962258,
 0.6410642902767404,0.7206860485776028,0.8790971776493229,0.5414410822038834,
 0.23756067830157057,0.9324399741479386,0.03409802919279736,
 0.9304811601708511,0.3032851471775565,0.7378985709276273,0.8359434735618595,
 0.4503585733870864,0.8559299783180181,0.09699994835444559,0.57621264418089,
 0.8473233838505914,0.08996491397832784,0.07995814047169547,
 0.15670373310794006,0.12979460279546673,0.6454905391182532,
 0.5506538476532259,0.744831099895387,0.20480104491765072,0.7264616545806337,
 0.009965778089206068,0.42292435462453337,0.02785044104682166,
 0.10686738028749276,0.7642960945984713,0.6740331668099329,0.3190953751565663,
 0.022542615010555433,0.062175846347416464,0.0035743185475652917,
 0.5469783109249345,0.5632128351911455,0.08887639850957707,0.9214317895869688,
 0.8621798291838718,0.11929156075864089,0.7102375702615323,0.3465118035654319,
 0.020423914456506154,0.7320865363259398,0.6467753358176204,
 0.3918177400417534,0.8226425461637418,0.7664875131270917,0.6130817871091202,
 0.9766601408416496,0.2607311004660148,0.562839520094927,0.29748944597162674,
 0.872448727901605,0.7747546940616601,0.5347454133599608,0.3156590157603081,
 0.361650573237415,0.057733482696647354,0.06174141551085732,
 0.3398044820041456,0.3399920355870881,0.8786584412418319,0.4939224626147347,
 0.14025518616156274,0.4413745432263303,0.8740525570541786,0.4515301070984983,
 0.19410555830428178,0.7475288289211551,0.680170209426337,0.14681610318558824,
 0.03627890752619556,0.5807305067058531,0.501501749152111,0.5952126152346671,
 0.6678790065695324,0.9053256828810841,0.36208807575598934,
 0.18363683637998773,0.8367604397424596,0.16775485384057065,
 0.6751769508997174,0.8246126769490756,0.6587375953526305,0.9447446407424209,
 0.014224947604575044,0.11357642651657562,0.13423817647441516,
 0.008238484817910785,0.622780444377946,0.5556384607543543,0.3591915993848491,
 0.5346510672285573,0.07282024389178621,0.7065790237958214,0.6945328160691203,
 0.322917248564465,0.05084082152971048,0.3443529151175728,0.8093104148130388,
 0.7270830106245555,0.590553440203029,0.1180434855895971,0.7414064573314726,
 0.3286211810827264,0.7886496880192281,0.8093210256425261,0.44502966968736135,
 0.9300924010941445,0.6758711937610523,0.08900450416951455,
 0.46846398481664664,0.2748178608846249,0.9839732675872501,0.504019059052334,
 0.6786465750542263,0.6902637985216188,0.36616118728672664,0.3349818134003584,
 0.2596439086211646,0.17406967652824967,0.0012876264536191284,
 0.25751303346423005,0.5758579234034273,0.9408379582607251,0.9409900371645663,
 0.6038145233491503,0.41517350079692084,0.7669635480882147,0.7127523913076211,
 0.9202515073934125,0.33091089389618467,0.3325080218289541,0.9392039381155681,
 0.7256913867274186,0.23951415762102768,0.2156600975977292,
 0.009932156653880142,0.7150790223813773,0.19515458010735254,0.87289332865926,
 0.7601795093615515,0.919874094779311,0.27116323076810334,0.40155726496411837,
 0.7998623567658095,0.28670562759354945,0.5822285322627115,
 0.31911917212788976,0.19125152085843156,0.23740619837179144,
 0.45036027680063107,0.7254755653822388,0.16317901044430017,
 0.6184046733166113,0.6466787569894271,0.16376201847593808,
 0.024297151476699197,0.5531556002803832,0.9300177615376004,
 0.9884604491497437,0.04143889841532744,0.7800051532712153,0.6338708859352121,
 0.05332187888166251,0.4799467555888626,0.7914398226232688,
 0.07055847442362373,0.04264705292088855,0.15029928429997041,
 0.6297641726547318,0.2907232992290032,0.648889500277779,0.4528697669009224,
 0.04284507093468648,0.9287088873449569,0.4852149412705654,0.9442721956058293,
 0.7287384908814716,0.7103883837411081,0.4413639235164135,0.48432654910542094,
 0.5799148555741263,0.557595201194248,0.6897245683315363,0.22517347595224113,
 0.20838435270293199,0.826497464204024,0.3795480634693815,0.8614909830896147,
 0.10434266462496145,0.6880222314657413,0.6665319449411602,0.8338081255766847,
 0.4208476257771989,0.056003006923248666,0.39389107627513875,
 0.7793009549052247,0.2745025268348912,0.22889623444901608,0.2052747009282906,
 0.6168766722503083,0.8765200247812263,0.8718766567968864,0.31218130886648643,
 0.8796773534401366,0.014157819358109891,0.455253896871857,
 0.05454649267436518,0.9344142170572096,0.45379012192818735,
 0.9812044633564092,0.15602876500732876,0.11681486397564378,
 0.6627048829286555,0.027199582016240953,0.026841542697725762,
 0.4315635570625441,0.1036185287868594,0.544714439393389,0.43873129883378625,
 0.37386264000346137,0.1269895026921326,0.3586139929931067,
 0.15920061805638808,0.5337591419140899,0.867125576730616,0.9727590464747212,
 0.03756171208871251,0.35774991161768943,0.7935199869716637,0.711952861381078,
 0.019934944987520242,0.8273734228156345,0.19975657806741864,
 0.8724818210206648,0.20919062123444387,0.42326509831296444,
 0.13448513891904423,0.6778630081560499,0.9069226556952785,
 0.059546415161561006,0.695962933413877,0.11411281245544302,0.642927608469964,
 0.2816208197230542,0.294280875257768,0.89735227907025,0.7165922298304843,
 0.8673752235209186,0.18346983359144198,0.3074208872552512,
 0.15207847051374146,0.7115598938137695,0.8587398176295836,
 0.15152936797378613,0.5461516635680611,0.8948527972370062,
 0.35796537255304917,0.1482743706439178,0.6495344512915704,
 0.13881808134993934,0.21883859094086566,0.873472617823019,
 0.041165654703066656,0.8466316887893457,0.603738030971297,0.595681645857282,
 0.6596889744184109,0.2764583782469201,0.3385037350061074,0.04366514961703161,
 0.916542061407292,0.2713475657798494,0.8146670616794712,0.8198869168950806,
 0.08893230691568255,0.5146034158103823,0.3311364346765867,0.6846088608596554,
 0.12769464808976194,0.06203227687514212,0.052008227046663524,
 0.8302453285091873,0.9822598018481222,0.3374702199358587,0.6622690018453348,
 0.7133856629855642,0.0595721545861293,0.033031763544260184,
 0.4829277948208698,0.08301520453781419,0.7927387081161894,0.5557713221734231,
 0.8735579458161961,0.3555919163145499,0.11439245685880817,
 0.15492211509869003,0.8280319100836785,0.6457463970703508,0.7027227851563411,
 0.023824182445825626,0.8612932156806572,0.7729880580547303,
 0.05984507503538938,0.510486084757573,0.7941207136334398,0.4705594152144793,
 0.6944629491505043,0.04526392249073752,0.5456307895551356,0.7516122278850188,
 0.6476041352320344,0.0010823679945017872,0.12201583019615525,
 0.6639936139201208,0.8945057484075833,0.21587300553592237,
 0.006906232160075243,0.9370037259651623,0.9047211152578215,
 0.5115772922826054,0.515081506372073,0.8491593237826546,0.6971423118206168,
 0.575871656313007,0.5986568922301219,0.3874339511385605,0.7916341236823423,
 0.9632913530317919,0.7146741580505342,0.8241132882696534,0.16625702032627743,
 0.4976676037858625,0.9113395672983837,0.21627578629104438,0.907368294313222,
 0.3387465994929344,0.031442388641539476,0.404795621872321,0.8119903994818257,
 0.5364619876261837,0.4083284901820111,0.1759006033002103,0.4423222485881888,
 0.9332486874680808,0.3407091377657666,0.10787421564191946,0.2583327314839624,
 0.258239526619642,0.4241486335323008,0.38494968297011345,0.6105511995403264,
 0.1663351199992561,0.5035441561417027,0.25630095453584156,
 0.11342042441231903,0.4872772721666391,0.29409807083038964,
 0.8375837932902768,0.5736009581733549,0.8177845791831436,0.8151896362954514,
 0.3204045130895511,0.8631033695808679,0.44992434475987,0.8683878428315934,
 0.3145605620110469,0.39891020451548576,0.7195336204639747,
 0.42075789180484957,0.2839384669693089,0.9330581495292201,0.7823825278700556,
 0.21158279588460938,0.9416132463256406,0.49831217473641587,
 0.2810421155382873,0.20932674948217023,0.5419872409374726,0.9863316274816373,
 0.5780159401702882,0.4298957749087159,0.77336097779981,0.5471933648858227,
 0.5738838365082892,0.8996050053683952,0.9917649418217915,0.9614464082214501,
 0.28022462452020236,0.0010845038895896941,0.4297503262940203,
 0.9396909706218184,0.4741998786355981,0.8641319385194224,0.9146881196573505,
 0.22587774544250605,0.42408506661780265,0.8063091072581807,
 0.2813236774502239,0.3328046890138364,0.5279778280888019,0.46937685764707027,
 0.0462425557181505,0.7140031296480033,0.10260838372010794,0.2591743053276936,
 0.8403000223322533,0.9084702475171319,0.09079503874847461,0.7447132979220351,
 0.9253407838505616,0.8736310966901122,0.690673029740701,0.6101559009836199,
 0.9296002329543468,0.32054376090299386,0.3579624881947089,0.8259202085041182,
 0.20392913020322545,0.8411957570300235,0.14249045120548465,
 0.12886330710055383,0.20654795428395156,0.7396595451581872,
 0.5530963835712699,0.03938131341847928,0.7035090757035709,0.1298470253654801,
 0.21173075673457664,0.149344991932432,0.5476494423254038,0.39604411023112684,
 0.0022220221651180205,0.5616063853083033,0.5105564499442896,
 0.3827813552864181,0.6434652409915784,0.1341303117004321,0.3064119551122112,
 0.23796633621394914,0.19485036591055238,0.051227009787798705,
 0.2463762756178669,0.07412091265131748,0.25158095977017425,
 0.2612991518208234,0.6169389640432245,0.014838659495647777,
 0.7753671085394073,0.25011824803502336,0.2116240416478562,0.5380664336306759,
 0.5124407708251846,0.5423947968636156,0.11099175947237538,
 0.46196056642934713,0.8426302731518801,0.3102878563081619,0.545311493432773,
 0.002324816032225674,0.37532409698914004,0.5521327507700292,
 0.12652191386662737,0.9874212365664823,0.921622307515486,0.46751202098339606,
 0.2959831241423405,0.7818186717065936,0.6551696366587219,0.6791814491593597,
 0.004200280744778628,0.9530570363567902,0.21584526958558203,
 0.5904342841369074,0.9703264235175486,0.4319785498806544,0.386655336616063,
 0.7455806794399896,0.8746883595738825,0.28450455115643447,0.3981086845054128,
 0.4141543728213086,0.4201182379892573,0.4705453551906036,0.2021292227011302,
 0.35238899612947727,0.9650566362335304,0.5301484268219744,
 0.48883796731683726,0.38815326008226203,0.5956669600057852,
 0.23671613237588296,0.736985071808423,0.856863105570731,0.7947992801814051,
 0.041233471739057226,0.42080725777077455,0.9870566683663207,
 0.9838197391402447,0.14068764272902445,0.7988868591391247,
 0.22467328168660117,0.893799763761766,0.8852194632927776,0.655428011766968,
 0.5830678756167464,0.7196622290680814,0.04124686130943189,0.9216229240389238,
 0.22473604233025712,0.15710036544008898,0.2365790633723044,
 0.6500725859666139,0.31549071441301657,0.6189709151772061,
 0.35952193200825444,0.3877914532474299,0.5636064868520954,0.9477808360001677,
 0.778408083559304,0.5247115409515102,0.3204745891260302,0.06812852983110573,
 0.09724072136023287,0.9886129533319137,0.1683070602859298,0.8793142062293199,
 0.944523601239702,0.8348287542866184,0.8217060133520531,0.7635912821611167,
 0.769705620841472,0.1402983254181439,0.18312368215413266,0.35555422650349333,
 0.20458290177529515,0.08640287968660965,0.8388553681945854,
 0.14198247353357984,0.03741496149584056,0.013317147086066794,
 0.10778176321812494,0.2196725526937261,0.9733203775608947,
 0.43841043813208824,0.03591696768329844,0.8999420096478207,
 0.005249976944933565,0.8429530851516347,0.8392550313857942,
 0.6712346286269721,0.7514180823021239,0.20177618776303707,0.8617408508781164,
 0.16700108698343985,0.6756100795372246,0.6103840812121882,0.5112092797590556,
 0.20961180555695624,0.868149987436951,0.5246625740492041,0.45540760379729606,
 0.2610804451227524,0.012698978809994821,0.24249983253241314,
 0.1796022297589166,0.6237552551103962,0.7235548687872458,0.2154162275638546,
 0.9596130647997345,0.8186003682780553,0.9355274804031233,0.29957449581592643,
 0.4510120002017365,0.3739809588847569,0.33971520724824855,
 0.24323438787355534,0.43205893986282173,0.42556525407919754,
 0.49045373613372845,0.6545705491989319,0.2669901446099897,0.8300274867427522,
 0.7375059823707256,0.7846479068093135,0.22368878370447887,
 0.11180174418620903,0.872108220273013,0.3291359339318043,0.8888653355077341,
 0.5357363672258625,0.4943433776705941,0.8857804465461763,0.5012604177396265,
 0.6660682543011297,0.24457444556892738,0.3611521004212619,
 0.12438087692471334,0.8072431440473192,0.3464948330831078,
 0.14721939466600542,0.5984878973352489,0.5378931425572376,0.4743239955554587,
 0.9570479255576332,0.07278202502076925,0.7633773184318013,
 0.04445748023783713,0.8122383092998064,0.37815870556430775,
 0.8306775306383706,0.718746923529017,0.540416895269355,0.42107830793263257,
 0.4718999870114331,0.984208105789206,0.01130952695805909,0.9973303806975238,
 0.20347487483409155,0.583719081908729,0.10225500336929505,0.1087268203230135,
 0.3934967296820425,0.04061216116303268,0.571704032920384,0.13016182203581994,
 0.19472268246640967,0.1442954044080229,0.47210118132971124,
 0.5312412595795992,0.0916249173364666,0.1425411423979428,0.4522119273973446,
 0.8705071738033581,0.7713842261215718,0.9223693556131197,0.8488572755204777,
 0.09695787808882306,0.6883023141303001,0.7156932786035524,
 0.49815285906833606,0.677617686205107,0.4723405494145454,
 0.0024513917521233886,0.3155156647756735,0.1245848892411997,
 0.49751226065617404,0.16650108588655288,0.7179262322479514,
 0.4513127567417218,0.9786176976291296,0.4287820633593684,0.12644826781066543,
 0.20917653246356904,0.3317368423480548,0.774993334694808,0.6653210336189879,
 0.5830434953636197,0.8812657072604301,0.9702050877692382,0.3392354061404572,
 0.49162849212659687,0.9960528041388378,0.7149266642784436,0.5006038224953384,
 0.6556401941900146,0.8097227801103557,0.8341539083413785,0.5030222102996238,
 0.4152286721110944,0.4068372770265841,0.5959099380764992,0.05765092162509611,
 0.620122569515971,0.83214139387579,0.04337896894547466,0.15026002492190038,
 0.5622002894056806,0.8723234939555337,0.6788140339876159,0.2759502231788429,
 0.1986915588483933,0.6363253951688079,0.815235951069782,0.3853079313638016,
 0.24169135773371297,0.7986507549807046,0.9160945809450849,
 0.21351109677098679,0.427258821922352,0.951978446113953,0.7919268680541154,
 0.3362446181990957,0.19392959512884744,0.17668976679466186,
 0.7363070231250262,0.07463817902290415,0.22227959643472905,0.696221216440435,
 0.32850272601442765,0.4410318868591063,0.6679688688865469,
 0.052052291152761665,0.2650296274519688,0.9163465793859186,0.225657074090871,
 0.6344636952353013,0.6737442860835082,0.8255751966043712,0.7629301848375457,
 0.22573160061935094,0.9863005627802739,0.9127802699934116,
 0.03550039581654729,0.25041249608563254,0.9477860752223107,
 0.4343866307871148,0.03093437232765439,0.060426388951497434,
 0.3313601876375343,0.37830785761315,0.1127715136350469,0.36656517201206174,
 0.47948797418448097,0.7599255352553652,0.9836872664509538,
 0.40127739741138324,0.44586404497589305,0.05593184577820265,
 0.6589135610137186,0.9450380509033295,0.7644897423914041,0.31820419050803694,
 0.8441142085444165,0.22035034069091308,0.02910425456724197,
 0.8598714760067961,0.8675565492784449,0.04340098141546567,0.8953803730245022,
 0.948246542572602,0.23772803611304116,0.487318776320703,0.6242486566467433,
 0.3111467202533743,0.9934213147077637,0.9243667756133455,0.412124618119899,
 0.5985627250961844,0.25348927734129,0.5327593885184909,0.7358922017353116,
 0.161748587490125,0.41748337595824325,0.46485354206019613,
 0.33459742324899455,0.6776676712571188,0.5956555479656404,0.1148015742636852,
 0.9175475951130254,0.49865441256816645,0.460503173052337,
 0.029108567091405524,0.39526836141168975,0.05328713549140329,
 0.6085984999618035,0.4788767438070196,0.8084577845953871,0.27959294372195453,
 0.7336388205761151,0.9999543933689865,0.19250164082846966,0.8637625392857798,
 0.27149417565211387,0.8040399836908513,0.8191792035064736,0.5691986602793235,
 0.34684234999086605,0.6434307925245296,0.9979670042413492,0.9250819834231911,
 0.0065088841241545126,0.5656249802556542,0.94189119724788,0.9666087579842825,
 0.7987834136102114,0.5451322918180097,0.5609067428790194,0.2616090832224456,
 0.9242750016480391,0.8074021448077144,0.030905997973427635,
 0.022574310294994904,0.9131276185424403,0.4857180251445794,
 0.5309280645474262,0.8515737650619424,0.002802311929336776,
 0.33710506585081723,0.6933233396755006,0.17394407111844667,
 0.5994396823137159,0.835317848845219,0.813827675985434,0.27930755701203547,
 0.03556959408981153,0.5348602443909694,0.3969162096441977,0.5612388871246053,
 0.9036295928783602,0.1178936832766101,0.9260804883252616,0.6772348746402223,
 0.04419204994155068,0.29721099692700803,0.07066230114085925,
 0.5313074254543678,0.23510559800038955,0.20620602461236537,
 0.01708923246603089,0.7142650087963482,0.27630077538586806,
 0.4617465499308323,0.9687301427596341,0.9056766471500084,0.5518750435190196,
 0.7898703317510294,0.43029956159245597,0.94010672505704,0.037627235030834205,
 0.20713993409086762,0.538815972185252,0.5672771037786145,0.6999666799351835,
 0.11829118879387546,0.5707609696437548,0.4479919452263842,0.6065438979889856,
 0.27302625270333203,0.2151893412502146,0.25397454853243184,0.943878051897869,
 0.13542950832208678,0.523720380543574,0.5696862900956172,0.279643671282774,
 0.25763009881684384,0.6110944412329897,0.7811574852438631,0.9934211856553338,
 0.39547023436599926,0.3784644146907812,0.4353259123211365,0.9902678880374637,
 0.8762415393912159,0.5176923888735141,0.14443807982002277,0.7544018296874612,
 0.45313140654417605,0.47068224535812064,0.36419021957510944,
 0.9053159367481745,0.06443867537653836,0.29636986522891207,
 0.9168296933650274,0.9223458824244601,0.16241791433958352,0.7454366543904385,
 0.8021650767665061,0.7327408529226174,0.3707170798490488,0.22280373843704682,
 0.08994823842708544,0.40369215234912437,0.10778666524034397,
 0.5347945337338451,0.7821296017298873,0.20979217596696575,0.5086662299490281,
 0.605939747911133,0.9981128229592322,0.8581253839459422,0.053777117413661646,
 0.7287617567055305,0.12387795886815811,0.9497483954444378,0.5047455068283657,
 0.9090198542511663,0.45026673664741157,0.5420968920130183,
 0.20026117878584593,0.5816393169482847,0.7819067248622295,
 0.36357356343360825,0.16364217135594084,0.9919584687788217,
 0.7851051017164213,0.4827517365518912,0.6622580381426119,0.2398331109110905,
 0.2783666195685892,0.4529393512276513,0.27731669768064604,0.6776829524394654,
 0.9686946882142301,0.09278937821908584,0.5391350721721546,0.3446046322151043,
 0.3157370634052019,0.18300053033052344,0.10785488385757458,
 0.4476740965169874,0.5706124647339013,0.19201826516102227,0.7479263363352883,
 0.38382646724717984,0.8200470555051558,0.3719904842024814,0.579005176091997,
 0.7791483788698538,0.6446702477826279,0.549755348114232,0.2828829951379186,
 0.366114918390984,0.4491237092269106,0.3513566359935132,0.6353837480161122,
 0.6342985902073267,0.13165993754733274,0.4086047453163497,
 0.27288289020157297,0.07726031105666953,0.685878642008696,0.9083885348754368,
 0.43821630636900955,0.8678305852191245,0.8969516142378774,0.9499187421521138,
 0.6061035272713023,0.11990652059285556,0.32340960491779047,0.658160883608276,
 0.5541407772755089,0.8986996528389011,0.02880315780738585,0.5276164378628212,
 0.36922636130497777,0.8527586070965062,0.1593574342352556,0.4632913501090026,
 0.33976250407398745,0.6480050575770263,0.4619466366686634,0.9837745723242662,
 0.11411561714474117,0.9954617553190789,0.14669839473654045,
 0.6985683940897309,0.7591132541106614,0.48585271336813285,
 0.22818580333054783,0.15101406396564343,0.9094395799358354,
 0.049424214965949065,0.6637494788044054,0.45776034303749424,
 0.42969493205569365,0.3498194836555094,0.09753789492592735,
 0.47157677477303694,0.5149419823585994,0.5066608313038832,0.4499715818481782,
 0.7025811740168986,0.15445896452487873,0.40789532650649774,
 0.20250089403281712,0.19914452010101225,0.287626120214291,0.559578183033099,
 0.9589150220251089,0.8875485790493365,0.4707544780228402,0.09862817143042446,
 0.35218130505320666,0.24189872967409887,0.8920959226974114,
 0.8625671305509627,0.15633507926094414,0.31699724211879954,
 0.19734343064084292,0.6011779152605612,0.83918255123083,0.5818640737046303,
 0.8616852660102072,0.03290523882823393,0.6038376319632042,0.4063220324871425,
 0.11378446190499358,0.7827013845469587,0.4332670433585796,
 0.41831324460325237,0.6781519284107965,0.6218745992433985,0.9090802831931856,
 0.9769023234275597,0.5064557906482132,0.30772711719285617,0.5019380439316741,
 0.8607426936928468,0.4301634154835807,0.29866750484484395,
 0.013684155364170114,0.9940450783636239,0.44835310189491784,
 0.5697706787461175,0.7868501707849628,0.3856025315594298,0.5953935630870663,
 0.25946966749818,0.7117311890274249,0.9276797848410474,0.6799106840719025,
 0.8824590761413649,0.7285247823603307,0.09005092723966224,
 0.014372706343926844,0.8553704308602276,0.2957337750335104,
 0.39099984499641516,0.09523526515895187,0.48610109852914807,
 0.20105293863400087,0.11634880279771531,0.3292403068399006,
 0.3838343531411902,0.4503214233986188,0.15579104055153015,
 0.28196378524549137,0.4141547720967834,0.1290051490148939,0.6772497341572319,
 0.8777171808823003,0.3176129133165192,0.2863933812812025,0.14610347222479958,
 0.4759159971653817,0.46525919019291395,0.12660851715426225,
 0.02008277888483201,0.7571367723314202,0.8883218327966046,0.782840390497157,
 0.05970735901895541,0.5191259265512095,0.36141733072720084,
 0.39875906265592964,0.11514928214036169,0.12242654718062851,
 0.038057640386563674,0.3164781164430396,0.9587687027030495,
 0.5357424460364177,0.6039668390626334,0.20212669639434788,0.8243444962623311,
 0.04494961558838362,0.05199127503808465,0.16623125711211972,
 0.4668825024911738,0.11804785795774353,0.4342621286383823,
 0.33273714395611576,0.29355949743363696,0.024368301115770596,
 0.8329662745730553,0.571730084012672,0.10598460335828752,0.5962367153182683,
 0.7389111083621663,0.11174969848165162,0.6500556041298139,0.946603329051238,
 0.4599422102030237,0.539030980434579,0.09384324661577959,0.4953767749597434,
 0.904990787362615,0.006517610944977492,0.6073983035485196,0.6276725770967976,
 0.4267577499044206,0.02544357543604936,0.34715068707625296,
 0.1607182375772963,0.6943921099900418,0.5487587891057149,0.3175299193166641,
 0.1448868550780792,0.17864918379939942,0.7721269063510384,
 0.004699427189101768,0.8275516852481151,0.2550208698589428,
 6.175101864378529e-4,0.776394703843702,0.584116292679224,0.775316207652959,
 0.11123132607481034,0.7882744776946047,0.008613767416345453,
 0.017796472382209272,0.23746473628990306,0.14525563882140236,
 0.48467965517125355,0.23660924595973132,0.060378589438252916,
 0.5461456371535796,0.6209598276730162,0.1327895920167963,0.03675412595488414,
 0.6736807895670901,0.3817286633465298,0.027716390877171815,
 0.9731580272927751,0.4520691716886661,0.561406347906494,0.28605654873210185,
 0.9985905404438908,0.2395165753197065,0.7953877626279655,0.5403225469027774,
 0.9490804947897803,0.3758838769362768,0.04105777677851408,0.6219762631446281,
 0.25224231409628306,0.6022306326100423,0.1667537162805317,0.9205366537789716,
 0.37450266292539625,0.3002804939016046,0.6891940492467084,0.4352948316206061,
 0.3678459473135447,0.19880854535562809,0.19116941484693828,
 0.7343988530973536,0.10024523939634145,0.13162694681623743,
 0.5359213331820072,0.540670600011674,0.8482464114603804,0.14031202853185631,
 0.9858462993261214,0.10580653187722167,0.1742903273086257,0.9366632649889717,
 0.8308062293318568,0.7927225192233763,0.964685502412336,0.7100171418613019,
 0.5944365920322889,0.5682453868626834,0.8647706743003823,0.5064061126248299,
 0.551871268985814,0.2176538454039636,0.9557229983070749,0.46692178402271856,
 0.296821530324747,0.8967539767237305,0.5120010055119754,0.8490406537834448,
 0.6577956447891296,0.846850106097275,0.2775672956670955,0.37432195108793864,
 0.9236659650333573,0.0419021742136918,0.5013240890014243,0.24197163827250223,
 0.29484992054563086,0.06455027082788078,0.4679296355075582,
 0.6489501628006237,0.5550318332303654,0.8869331798987327,0.12157011596514455,
 0.32087696589503667,0.9518250215187672,0.8158576994081527,
 0.45096095204600084,0.37241561253060773,0.6530162562853823,
 0.6498343207019449,0.8013774926863104,0.6579674931848508,0.2942259503175979,
 0.24484965463908326,0.1492736402260536,0.7171053558240318,0.9386364215895678,
 0.11834997668992253,0.05882056024485405,0.438347631693372,
 0.28774699648895596,0.3690154298571767,0.3149349961083251,0.0649605152841064,
 0.07923206917735881,0.44633496082430957,0.651997222365362,
 0.23343794044537902,0.998113248114756,0.18125084352055554,0.9639436664709052,
 0.7235684918244424,0.33761384560431285,0.12183796407648484,
 0.6796503936517017,0.09406861039198144,0.4896984210914642,
 0.37761330367413615,0.8462986958570006,0.6932079235391486,
 0.09593468953933926,0.8436315391958065,0.3794148271671919,0.9573333595476403,
 0.362908770381292,0.9478770982976565,0.7213737808165765,0.13885314174722707,
 0.24838775828628568,0.017146335350035735,0.09752103241964061,
 0.7050151647145619,0.9245182400559041,0.63350219297688,0.3407490782284668,
 0.09755262716812796,0.9650763877166151,0.19373087096316033,0.841215817040427,
 0.7013398311637723,0.3134235022009807,0.6973272700356979,0.3610594929440313,
 0.6589215520167284,0.14535861266230288,0.2289768429212442,0.9480613744523597,
 4.2159121351814655e-4,0.7003729901670169,0.2793023291730311,0.93539732949092,
 0.1049674134670261,0.6518087086919304,0.9478369077687883,0.7361946481240734,
 0.0813722684925059,0.6545921109625535,0.6928413154537975,0.7125926247981207,
 0.6507801013991382,0.10228913527706796,0.0036858434392009753,
 0.41933610980293345,0.3781155277985757,0.8679434264991706,
 0.28516212763023396,0.24867859011224047,0.7932719918114933,
 0.10822799270172967,0.144772934137678,0.4232607970960953,0.9851441415958115,
 0.9925481879887333,0.5411768402330098,0.9632452406653852,0.6004216917384892,
 0.2682289117229586,0.04875360458061828,0.6991599022210633,
 0.09046100780731703,0.6655366461901626,0.6486316355185017,0.698352678007519,
 0.6369265886938074,0.9022064359620559,0.3617474636501893,0.7049454124157152,
 0.7895690847878633,0.5655874326586297,0.4812993940258695,0.9603413727348284,
 0.13701179816187192,0.1522000761591249,0.10225065624596874,
 0.43283494404141953,0.43487426836636467,0.8957298879279526,
 0.26338494269869006,0.8046476992567291,0.7270916707970917,
 0.28140227199990486,0.19197733666770667,0.2619436855837053,0.72900783937851,
 0.02077228449756574,0.8981594766349295,0.08549386997360597,
 0.29351573712624845,0.8590346776791262,0.35010478125499966,0.982458521641437,
 0.1462849195630801,0.505355365982842,0.6921486513753348,0.48914851860643627,
 0.7823069928927286,0.23938139095471633,0.531557762984112,0.5414066051843227,
 0.8236425475932345,0.6793827231368487,0.9611870006127831,0.5037229903876941,
 0.3932925907814967,0.6260362520839087,0.6465314262877099,0.12153158106971251,
 0.8117313558274534,0.9123441133991131,0.894948738685307,0.12239459448639411,
 0.06294061791173378,0.7226936576688736,0.30035793887925444,
 0.1544890545294597,0.2626100247665807,0.8442521977412673,0.669257386022284,
 0.037776280070973,0.29057863016912133,0.9924009280810911,0.6522772570350844,
 0.7567301894070758,0.7968139098687443,0.768175536280707,0.9460005328583619,
 0.41451407626923864,0.19637746144287538,0.6375544550977178,
 0.5215238716359161,0.5982148155284338,0.30537211247931084,0.6051025527673954,
 0.30514932850605925,0.12058471430414774,0.21503952001260274,
 0.963287079804795,0.4375474437809539,0.5061703522676203,0.12003942498613651,
 0.14127814144011597,0.8265845288184156,0.7050457465226768,
 0.23092707056319384,0.41373051543296424,0.3254281747238088,
 0.05597023308419413,0.26006930496658254,0.8062239607396819,
 0.1322619156196807,0.04559446214019458,0.8390548776563886,0.8701204797489379,
 0.08915413430571972,0.1893543685445388,0.7711326738460527,0.181588546786573,
 0.6081701432024442,0.5087016402396444,0.24309635565306076,0.1743061724153091,
 0.1346495115960703,0.9892574001181594,0.4376241723482831,0.3120032698034587,
 0.15824247365025434,0.9534257830844374,0.41111362784975314,
 0.6673579775070125,0.4714955384569067,0.62668598408538,0.7092739289578265,
 0.052398863338660284,0.8713871552243381,0.5720859040268427,
 0.2947375627205907,0.2554665839980139,0.42305171505982153,0.2183557014609535,
 0.18763087473328766,0.5325449120916993,0.7093309502414289,0.788031139644839,
 0.6487853132437766,0.987119100868408,0.991413348610946,0.4791292475416562,
 0.030244443741461402,0.9789887784544927,0.43982358071368577,
 0.6335119609715973,0.02618862894021401,0.9849309791283083,0.3846864320285772,
 0.6530205513732718,0.11655304648775855,0.4811336627461712,0.7604292516653482,
 0.9123039723983161,0.0015499699754824015,0.5230297002634797,
 0.9758061177629278,0.526360781710364,0.4047636687723415,0.9244662916271685,
 0.9997124843591771,0.08591555378038407,0.5887717486869944,0.6074579385537409,
 0.31162369007235724,0.5439316018597995,0.2876286374008673,
 0.30858124551092847,0.3001703704788883,0.20111131935570692,
 0.6359581346438083,0.8160609625077142,0.3468867205705616,
 0.013609020293209206,0.7109320913277472,0.689801373655613,0.4626128707777417,
 0.9454085181344809,0.13262766659777947,0.6616924331096126,
 0.21585486622193517,0.4545627591071729,0.7713631027907346,0.5537258343019792,
 0.3155040548484289,0.6861625811888139,0.3082355403015592,0.8053212168882405,
 0.3068955984071896,0.3593021136263662,0.18036855039415745,
 0.03304636516756576,0.241531026594198,0.9600986840412102,0.5665617284275922,
 0.15310687945260104,0.5996814379352671,0.7020957167881637,
 0.06835347389334256,0.8036038809701701,0.6873262456907471,0.5433984230808229,
 0.1398820624355691,0.7009449983706135,0.5325102151676416,0.329674301299423,
 0.5972913020163113,0.9413691228847336,0.5584783773188425,0.06593055906001,
 0.42050967147002427,0.5631271906113677,0.6905662871846623,0.2257069964467966,
 0.6805990545465851,0.24128310202850833,0.3932960813240728,
 0.030179049561549265,0.9290337273680821,0.7189445158419563,
 0.6504398665821716,0.46498952709270314,0.601780743405274,0.2914842628325771,
 0.5057393681283013,0.96362065337559,0.6005678363126367,0.27732676754616103,
 0.5997893809150692,0.10584385734732304,0.11059615306345572,0.908037258261917,
 0.09839570390696828,0.6592712035093944,0.6381160660991462,0.4393175466450021,
 0.4356216753284011,0.14074678217209347,0.03168558495025797,
 0.8664402034608671,0.9014400455321565,0.5039754869337714,0.29083929170777967,
 0.9138425689093539,0.3933323080027853,0.0027056757689550848,
 0.020613612825228222,0.22121871534329718,0.7431981001745723,
 0.12584892364150002,0.23173202803818505,0.1981784022408245,
 0.04394297640350653,0.038256112641063544,0.6098689634208729,
 0.5988149111113955,0.4730993337449374,0.16526307019388875,0.8961074976262017,
 0.7653093004398435,0.9976002868657445,0.4853902597457196,0.2017745596235785,
 0.29179282792983297,0.9615162849111465,0.7191899832109323,0.7028961200879829,
 0.6516871165668984,0.8813390971315377,0.8327402720513482,0.7924491099486295,
 0.697962709503894,0.465144312688742,0.23901114083102148,0.465337813458067,
 0.3974297243406333,0.08550398390181102,0.8047642510869016,
 0.014090857355209208,0.590814772832939,0.934276372049949,0.13706387591681735,
 0.17833221730488913,0.06261396474054648,0.8734129718941662,
 0.49143500485333436,0.647198179119794,0.9027560461966173,0.374854885307932,
 0.283502333793454,0.8921526535805371,0.18588019030932568,0.4654958673472286,
 0.1762202249579261,0.552636307662658,0.9695486277042996,0.6775764178610993,
 0.48571745171712744,0.7397712035141213,0.1957878403256903,
 0.17498356925604475,0.11882408846455383,0.9262605267013813,
 0.3694138928039208,0.6586707725272924,0.24341248587272712,0.9243610997403722,
 0.20993845695979962,0.740077947168701,0.6213084245874629,0.06463359778647215,
 0.2608826131816179,0.701568636374166,0.9345612958623266,0.4348955213207901,
 0.6414031575476302,0.41985219474090585,0.7818361713553965,
 0.10506714337583223,0.676878189422046,0.5424076040015473,0.9126920040925893,
 0.5689397360059185,0.16749761984255596,0.7403830840087741,0.7464045358429988,
 0.4840809862457669,0.24009492424022238,0.7149645929857278,0.8784780986985399,
 0.985959473578358,0.17233351327391533,0.5227102361890498,0.348315494879297,
 0.6681744804055668,0.549665948098906,0.11510534673879191,0.16526857952852303,
 0.31308267155185154,0.4723027914565119,0.9055056953520267,0.4620070911150239,
 0.8861666578839594,0.42736837026148256,0.05787956890998225,
 0.28529443217311656,0.13085393417206914,0.33426410587254696,
 0.9644572726787652,0.18196575467459752,0.41601707736953997,
 0.6407956665917699,0.30564995880714607,0.9875805823243659,0.9115910403654441,
 0.9891944221657777,0.9162695714317015,0.646229451576645,0.8963724292201293,
 0.1667843212223602,0.33953349723042503,0.3046537004581733,0.6680697437834743,
 0.8807157587550708,0.36263242320418565,0.9368216678743073,0.8349307531450206,
 0.8888789700038802,0.6279596088213885,0.0546439093114518,0.9626936098074754,
 0.4089199744987836,0.38630985952051755,0.09819276754996831,
 0.050463527333540426,0.7678913405491156,0.200132195338646,
 0.014547448872853419,0.02745332726701466,0.9220418206918299,
 0.06762449036108731,0.07895957737527382,0.42617596674594027,
 0.1719526162459062,0.3373551116478273,0.38906986010789313,0.5546429977380877,
 0.34536209083279734,0.2884129701336495,0.43332633503892126,
 0.3549572773971881,0.447278429249531,0.427383969415181,0.47506827778426397,
 0.7805821868351382,0.7391911390330522,0.17782165876394707,
 0.040943156790081314,0.5492990636957131,0.7853051218797693,
 0.9234953055341695,0.5531973649826353,0.27669714499243014,0.5389930801723342,
 0.7227653026364453,0.49932204984823136,0.9691267856106814,
 0.13954674810571033,0.9572846644603897,0.026967035816493068,
 0.28917917044442776,0.8765247698994695,0.09753168026741199,
 0.3335233155036491,0.23693639855492632,0.5816478448172621,0.6078467192569275,
 0.5244071896814281,0.6056487173118056,0.8938160599279077,0.12377366649822874,
 0.43680721622947305,0.021091067696484833,0.9943576372447752,
 0.7655846082854869,0.7762821785391815,0.02115588593051998,0.5397859751261325,
 0.9631211668771749,0.35061678545007735,0.7225214829803592,
 0.33084804978970583,0.3562383421859726,0.37646931800366756,
 0.6243957249660539,0.6819881096342837,0.7082017393796931,0.04423745488613273,
 0.6451631338589516,0.903915490992296,0.47632747948710774,0.22535411094235624,
 0.10154653689871362,0.07218635546968866,0.5873980492005573,
 0.024070585266846023,0.21159746982044292,0.761835174387326,
 0.7752562608743647,0.07031381160778982,0.9538752843243724,
 0.20349772844355696,0.4057032262034933,0.8381611764768498,
 0.24986267572158394,0.6042441537928017,0.772317972844507,0.9657069910194049,
 0.12716004948471993,0.9866686372547526,0.24115356326795334,
 0.5991063573722951,0.1500768063803587,0.5978489264910877,0.2789164739318378,
 0.7589704860593662,0.6285703936158165,0.45379530064011986,0.6681770278985966,
 0.35873214084540306,0.2776492083680955,0.6700143585686724,0.7744152913930709,
 0.9736652877432334,0.0023447892390825453,0.2527259876392871,
 0.13577640562104998,0.2865537747301663,0.8989225192695011,
 0.08036825813160142,0.44272202562950913,0.5547446126661058,
 0.7401255280037135,0.6179322155585565,0.7100199735541677,0.5024339047060716,
 0.1307425676215651,0.6123500914205184,0.8482700871829112,0.3484444908197265,
 0.45273435954249164,0.2566110415196219,0.9314193497227592,0.9055422478066929,
 0.9962172299732618,0.41616758441301516,0.8670676144872109,0.7070778437883638,
 0.45523592256754486,0.1778680862029285,0.21257613664927832,
 0.08527007158868305,0.9709493923527042,0.8847982406373247,0.6033444091906244,
 0.6377066271285985,0.4871717337935918,0.8606897266812586,0.22827005271297107,
 0.8990597061249276,0.4540096676587012,0.4745490330351563,0.16908384328157655,
 0.7232648617545471,0.8468109104902021,0.398917367183945,0.5848196241213504,
 0.2544420216126466,0.8758214908484916,0.36576705444276636,
 0.31558071241194097,0.4238596369575025,0.3117203642899067,0.2448632962226518,
 0.25979420900461647,0.589397308132029,0.12958012158623322,0.5187065147790466,
 0.7405346344886385,0.98911972290191,0.9994097146425445,0.6479894422601834,
 0.01509713414166769,0.12062459366401301,0.2788192510095059,
 0.9490832437914182,0.9048659370883942,0.8616782202974032,0.6920185855375621,
 0.027186521914376427,0.30331354763752494,0.9266825210788447,
 0.08940661934473382,0.9189125309461793,0.057744050785602075,
 0.5945364862994783,0.844805614604521,0.3369935750524822,0.3338182790903492,
 0.8713310052442027,0.7724362975066451,0.07661660412788329,
 0.09561735132081162,0.7766259671302596,0.33600534487307254,
 0.3942949830240733,0.6922591690351877,0.4635838658241632,0.9358936853407093,
 0.423596203851194,0.6157878693048672,0.7171785430256509,0.4940002760208052,
 0.08705557225727056,0.7565180258968077,0.9570416958515932,
 0.034984775663626455,0.39064871235382115,0.1435561302448135,
 0.28989257797849743,0.3943026207119348,0.3718462400409729,
 0.34540564690973696,0.4992969316209891,0.3103446136835539,
 0.09523055933693503,0.3695670011748644,0.05921395978978239,
 0.9162852099751219,0.11425010071127806,0.18554761178684442,
 0.5772723755344725,0.4608513266529459,0.5389575343383729,0.02816195548068612,
 0.7367897058784199,0.5191300826867936,0.7293099314485605,0.1648320084503826,
 0.11362699508464413,0.5213862370294002,0.265108227600072,0.9343140820582962,
 0.9017021722240424,0.5023611446743959,0.5347850113374706,0.3659910762874433,
 0.435610597438985,0.02838315248333212,0.17132354907893554,0.5877649961180458,
 0.30637091671993166,0.8613977663915222,0.8319266532579432,
 0.27043987429743677,0.704037087664823,0.42004154663187543,0.3130727332291432,
 0.4787397075492268,0.7036820969505926,0.5579965893551012,0.7000241267268895,
 0.5865715915507886,0.3382302360752947,0.4639625290936068,0.04773614203382204,
 0.2437255197879269,0.1151521373558021,0.5544125335971051,
 0.002649064787643929,0.08031784562143407,0.10878201828396594,
 0.5448037915159346,0.8538931768597575,0.5005612926947185,0.25377948468940636,
 0.6841272971351943,0.45542070984555116,0.6771114347915507,0.6671718292050648,
 0.14788069468894904,0.6202297599636324,0.2989254473929962,0.1068661865715812,
 0.7708302215775537,0.42338501143845353,0.08580461388279925,
 0.24008244451234517,0.22084829896538405,0.29895013773902257,
 0.9992316976295438,0.8087372058898199,0.6660306218409399,0.22172273272476917,
 0.834671010733788,0.10117978080286172,0.5128080295184066,0.6837665528159381,
 0.2966283207472895,0.6303748397716964,0.1960398869006269,0.9333769166490139,
 0.8697524889657113,0.84532193137512,0.8671036833357942,0.20199866175647996,
 0.6266200891833339,0.43289036484719057,0.6725960492406,0.09862017768557207,
 0.5596404074665107,0.6116216185091762,0.9692311352329607,0.7811624796401381,
 0.7264154497730893,0.3269339172151544,0.10958753810086774,
 0.33434548139383347,0.6730473137819032,0.7435640920296135,0.2543319731827214,
 0.3196990668682118,0.6255918130162081,0.06613212398834478,0.1654339715575338,
 0.7496142611420792,0.7743512706371641,0.4709645654541488,0.9333233238378471,
 0.5232667340287186,0.9666317462369592,0.09299299803772687,0.6669947926218598,
 0.5525798558571272,0.5714498776826848,0.4398911795594166,0.4972173517190206,
 0.8193091526901357,0.08865624175069808,0.9231516251417421,0.3191493679107863,
 0.08567138652575157,0.004437170078914643,0.6180699369611137,
 0.7194242814652387,0.07175348943161708,0.9580748660702861,0.6368816817514491,
 0.6662211029055798,0.5038402527473002,0.7006785010626436,0.41160468211675894,
 0.14372708504639242,0.8538142088970229,0.9075144107292952,
 0.37668564670407256,0.13821582858759351,0.3573310450986148,
 0.7472493315446223,0.12887970749146638,0.11517414252122293,
 0.12392561820461756,0.9739551184831123,0.766517305100608,0.05017616821868576,
 0.36055715456681836,0.33291610695835694,0.13800805458930554,
 0.2618191745588482,0.5487084348881744,0.6551960485661175,0.3884073379399896,
 0.026552137536399067,0.8421039769923699,0.5349934544428158,
 0.11113382541303618,0.7483678336322047,0.50790491885543,0.7307727846449076,
 0.1075490293490845,0.7664293399370867,0.2660902272287422,0.20363310322262484,
 0.852983017816977,0.4912862004302877,0.548650333979865,0.7049539949775508,
 0.5788667187756719,0.9838735235552741,0.7933487691593142,0.5168804349092798,
 0.8720719305613481,0.16873101348369945,0.7744409088381985,0.7373538356067252,
 0.009306120029589637,0.32379009591844665,0.6171841356368838,
 0.19471806055335228,0.40408746444234145,0.06690504654697083,
 0.10673526851583326,0.17977145705742625,0.5186175614611177,
 0.6689451818545206,0.9620876761992863,0.04277867041922723,
 0.10730222424498181,0.8643612255591872,0.038022152051008784,
 0.6425943089585849,0.3989784885347455,0.22667903861468464,
 0.18669167244341167,0.46483316624544324,0.3777665402274685,
 0.25173642952575404,0.31873591695222936,0.4335955116416059,
 0.6808813033478909,0.9375845706649601,0.4451202554955751,0.6727993019500089,
 0.5991788317885048,0.4348765480932316,0.6435759842932424,0.7783011836540528,
 0.030796254281162838,0.9026555051292661,0.8851632882581442,
 0.5905210234642599,0.3904137338220618,0.06184883265776375,
 1.839407902289869e-4,0.9168165665834483,0.5854979285904084,
 0.020716850324066227,0.03131258011124016,0.3252940466052916,
 0.12174795464489307,0.19297103833252904,0.6195869866487476,
 0.9748332899394517,0.5603691304742536,0.0978627086239382,0.07864466401643044,
 0.9885042191027448,0.5880618660342221,0.06564818313714993,0.377297778509598,
 0.8929692240848064,0.8490329137184234,0.9682428041108568,0.6974263506438055,
 0.2790964111007571,0.12531797725163796,0.018036433516574246,
 0.7239606794451092,0.5362145277586566,0.19477691907156158,0.5723844714268187,
 0.5002289888859189,0.19409754084258435,0.3327606194143531,0.6477777208565717,
 0.4922139553588677,0.2523684058767264,0.31576671186773,0.44741023895313203,
 0.9391281184654069,0.868233883642336,0.7966424646254043,0.4298636100413421,
 0.2966886027801121,0.3614828802546358,0.7370534955229239,0.8845745501472597,
 0.40106329233683136,0.32406531086302914,0.6486699332740208,
 0.06391080212147404,0.07647686581845559,0.8227499051478686,
 0.2807179625760705,0.7518758644909842,0.654506609983279,0.6866064617841102,
 0.7474567953699744,0.46017853991506186,0.5727000580285908,0.7420007702286595,
 0.15277371335536372,0.5842288694367936,0.3055428263554132,0.8544640463216957,
 0.9450648798930571,0.3246316229641699,0.21403384373778334,0.2071879238013199,
 0.8677668235314744,0.12507579604536012,0.7784618726336151,0.2559871010330783,
 0.5212498091663355,0.46117471491220097,0.9000780676256893,
 0.49373881446978596,0.58505626457965,0.36734499959825095,0.41190892292833303,
 0.8136457197432295,0.9901594362008692,0.1807286027905317,0.5618153915234245,
 0.05890369635339232,0.13383540629470225,0.4005819599205681,
 0.9721943114919822,0.2385370445871906,0.33510938333682294,0.9234681619810193,
 0.06543801755488399,0.5357771545948408,0.9679159360559273,0.7512836058542911,
 0.43934493440274924,0.8612320553908974,0.4690663955521628,0.8584917744580931,
 0.3895520953056034,0.8112168817674621,0.5384496644561786,0.1930543831921886,
 0.0523460167434443,0.024619291341055227,0.6045346275686703,
 0.3748284122376252,0.26721231718217386,0.7154831628622654,0.2735185374265362,
 0.15729233363604755,0.6068853308235648,0.08141883519880189,
 0.7803028122981128,0.3635867432297215,0.8580831555299904,0.13127146713696702,
 0.1901306311870451,0.19436976208431522,0.09795423178417173,
 0.03772215717936134,0.7813851366264092,0.17958047178047754,
 0.30216301171028137,0.3915390780207747,0.7728114823788487,0.673995327760028,
 0.19368978180036212,0.2251907813079138,0.16787205606801292,
 0.8527013666407074,0.9719740857118743,0.8794817802926526,0.9879276635031181,
 0.7282709119041535,0.2805880376977603,0.23362693506688836,0.7786359706487184,
 0.3271592961779398,0.8606498556328978,0.3639813466860681,0.62131368858618,
 0.5750680934679357,0.6486583755252571,0.050214988633706104,
 0.8627346945569054,0.3652738580831836,0.7564559670849619,0.5438536227876025,
 0.1941580486002391,0.9844890555915748,0.292477012221943,0.6719649776991802,
 0.5548926299462265,0.8897940240784822,0.1469862021677043,0.5506649903346822,
 0.8577214978056952,0.00976280766479598,0.8573620168904043,
 0.18670635511804023,0.5518114816598207,0.9325773007547615,0.8938223223465828,
 0.5154025076348753,0.5510002990207578,0.08662334702831898,0.972681899063893,
 0.7070630782442292,0.9477444454272883,0.10668087386687553,0.3096065472983903,
 0.3567840136220778,0.4980352317310591,0.6557210709060591,0.3120311811873008,
 0.9978503667619296,0.13701900576589032,0.10462072901570973,
 0.36878646068912724,0.4351601938080135,0.7460499033782836,0.6531940291971832,
 0.09696683006348494,0.05807729478390489,0.8118284267219597,
 0.49976545828536256,0.53662603541523,0.481499075697273,0.21650941564769377,
 0.943438736873792,0.04516025421405234,0.03836952375273084,0.7705413667927764,
 0.7627067763792805,0.37871973531032765,0.44882028445218836,
 0.48421409719778974,0.06855344623044501,0.033211514840740586,
 0.6305945361568132,0.03408332452402041,0.9133440436568795,0.7615899154755075,
 0.34823451867165367,0.23310010103302914,0.45504855647768305,
 0.5089672020690179,0.8113412877137257,0.2720251260651476,0.4317258732628404,
 0.06314119667746709,0.4230230016859573,0.5253234325695755,0.4664182476174149,
 0.8045396746432641,0.12453188701312667,0.38137673940465844,
 0.7701385499945215,0.07398931510272944,0.7219172189825125,0.5387848321583648,
 0.438871946553973,0.7035691129244205,0.356117554050591,0.13339139387755927,
 0.8196060930832059,0.9530149435726727,0.12206863157002079,0.6122473450540256,
 0.04068604367978923,0.04145713537468765,0.7365883977463014,
 0.41093324898139594,0.42439247178843276,0.17935460572859085,
 0.9111329399692083,0.5245463560254884,0.14842914398443652,0.6278381132537169,
 0.6411910130105755,0.6292252405920105,0.48346069127612623,0.2656729688902876,
 0.8929451363875283,0.4990114239885256,0.9871615014592392,0.568171521262004,
 0.55015861727358,0.22914771087897445,0.008702839913813043,0.9584728480267803,
 0.06837200466102722,0.3305615859070966,0.15268203945899794,
 0.5301383012190165,0.5131059165885754,0.12828345623691617,0.653953230635053,
 0.512481046397753,0.2884331408640668,0.8069222958129382,0.29306077530798036,
 0.5563089390786722,0.5593407574974196,0.43689233592443344,0.8136843860391139,
 0.1676474011700433,0.3982001469180827,0.4656224225893979,0.5382007665411467,
 0.15812626781689332,0.15423129130203272,0.9331797320701015,
 0.049800855888497586,0.20435366896937135,0.8352189762692763,
 0.4800231202717783,0.06321906513919195,0.20578406243773995,
 0.09234180438868389,0.5234204867722556,0.31513314851837837,
 0.9753758972741314,0.5271192689225962,0.9686356602982522,0.4713643007467674,
 0.6733925571936348,0.9595771281475998,0.04222694242809899,0.6266905053882944,
 0.5573138533572508,0.39947791188145143,0.41490102668690265,
 0.5689785364611184,0.4174511282246869,0.12198994830192511,
 0.25195492732987523,0.22124034536154258,0.19249137844760145,
 0.7711567249612353,0.8463191124065579,0.9799820872245676,0.7398510182581886,
 0.8845148340555585,0.8643879922485799,0.8494632307524461,0.8482388892720268,
 0.47550469957665153,0.7443781494529105,0.8165608853487728,0.6224651343655023,
 0.26179880190127747,0.7384870510181222,0.2787663195124285,0.7868514391692479,
 0.2999489671237535,0.253602908488042,0.4871477984026743,0.46433381533359996,
 0.7899390240639109,0.4384432707974242,0.5478042727207773,0.34633368164168377,
 0.993739228557534,0.5950024843354085,0.5117972951235966,0.3586289505961672,
 0.37425165511324987,0.7464594866795767,0.9023186312325279,0.0568079415531163,
 0.09482115349125853,0.7274627986787372,0.28912145853622173,
 0.6780212485871551,0.008097181867721126,0.9958181465557145,
 6.365200589891762e-4,0.3812612764868367,0.5409127730864955,
 0.3565969064398806,0.5501951101321271,0.5698895146452032,0.4817575500428948,
 0.25913976473568134,0.88942618018911,0.22243933308172592,0.6604938588704862,
 0.8709557203833234,0.5160541731467717,0.6869566177136467,0.7123803476267893,
 0.6682689715329655,0.12546565191909675,0.0387488501721156,0.5439954752049068,
 0.7414501372359775,0.544888661505069,0.08736843070014233,0.4363884919845539,
 0.6839651559278774,0.5470295807136619,0.23949041535395044,0.1620461815303854,
 0.2681836855357541,0.849661675224207,0.8537559936511713,0.09984939509725177,
 0.9242506235609186,0.05485629103825085,0.7076069862670287,0.7500888139836145,
 0.6990160531600658,0.8155453664895854,0.5633282529842927,0.3164089465912091,
 0.974125475834711,0.05016002572476819,0.6620282264832836,0.6674692002457275,
 0.825136310345339,0.9077242995132238,0.3535332263978136,0.568610910375195,
 0.012361822495453478,0.4137364076615574,0.772162146837665,0.2142216474300177,
 0.3490317898229953,0.46876943864162146,0.9573651785522792,0.6172507695915022,
 0.2359315632978859,0.7792104200110629,0.7400313773291161,0.6928483545890413,
 0.8781839310486683,0.3418219310808416,0.037508528759216286,
 0.4229306719587981,0.5541615510381938,0.7733424463182395,0.6284671834053843,
 0.7369988370529339,0.48083466362979166,0.9717669336814101,0.8451025178302344,
 0.8979532195581275,0.5098470735649381,2.595431308569651e-4,
 0.7104157937333135,0.31007222142858737,0.6862106583273523,
 0.10302722196464664,0.4350369018405551,0.6119995016086656,
 0.08197965640390137,0.07676219769095372,0.8242203766138405,
 0.37288204896517574,0.9306294954773097,0.4112292734575953,0.9984331079602077,
 0.8247644636384367,0.47384337056547765,0.3123076714263685,
 0.38660542915916707,0.4410002551675546,0.1911507938230117,0.6767898785723974,
 0.9591358454797821,0.9567195780159897,0.32415955726130374,0.8284743513497628,
 0.6316410219025372,0.317794350927339,0.9012828419333153,0.15851795579089156,
 0.8072003183226674,0.5744570973005261,0.36899516707866853,0.5716394299493724,
 0.8443059925700632,0.8115771879244646,0.7196096514097471,0.653286903780229,
 0.5480754718721478,0.904949326877746,0.2699002949365613,0.1087507517864228,
 0.4415751079615369,0.8028728395355793,0.5470600877160221,0.8181757543527289,
 0.8024763875487052,0.351540769951863,0.48966650869666173,0.3439305849559401,
 0.8637634609861361,0.8833502970873902,0.5846380263271235,0.4412798199390001,
 0.051343006954951975,0.31397324746125754,0.26385087653672123,
 0.7730185150950498,0.5248874387101774,0.2923221483542807,0.39996877341237314,
 0.6665203159102434,0.7014063424881972,0.9391360510304931,0.7289313769668971,
 0.08120491943965535,0.997559655318419,0.3039118659144846,0.28385893927747974,
 0.7791251625479456,0.8823722913406787,0.37713301055248105,0.8532345320150131,
 0.6579006338992106,0.9943610640912883,0.958717039348055,0.4462807533921621,
 0.9805350097884489,0.6569046517670154,0.7326891199521561,0.6004628648186164,
 0.585326451419959,0.35567289896677,0.8387949194541722,0.35414101012143184,
 0.20696088732757922,0.655850261192946,0.0941051009481777,0.4660493561075343,
 0.614377676093237,0.4790853378939064,0.7153881109544958,0.4178806037352871,
 0.519159447941232,0.7284906363330265,0.5503431113949544,0.991209503836647,
 0.5867829676012084,0.71466489209464,0.31606499520099773,0.9411642252918798,
 0.42834124997182754,0.9341833632717084,0.7981563588898583,0.9799365242348045,
 0.2629995783064085,0.26988925056226687,0.4575216976351997,0.9792906853589812,
 0.9688342986738756,0.11986747833252409,0.9572038667177512,0.6362567293447301,
 0.23940087269560906,0.48415759756144605,0.37554126480170935,
 0.7697671550518886,0.9598299940395413,0.9379727838312324,0.23392713728041525,
 0.6615214075939767,0.8127816258591769,0.12195746452709788,
 0.44530619646966296,0.9757661285609374,0.7459636330255193,
 0.23994722653456957,0.07472234538878542,0.03647839568713551,
 0.9519167703410908,0.860476668485028,0.8327519336653163,0.023427195297799686,
 0.06008868646231502,0.26035236177804966,0.31515248944386975,
 0.36410406933611816,0.18522755301341987,0.11817118456484255,
 0.9675987623722941,0.9476442712118015,0.10370205897640483,0.8974368359030704,
 0.9199709733185395,0.6339706809633592,0.03138793922831218,0.6783905884257396,
 0.58182158646572,0.22518738168530728,0.9657495597041698,0.9840566594608134,
 0.5600571108330057,0.527546393513378,0.7835679590458907,0.8923869314996633,
 0.8747598484293444,0.11709692615086864,0.08813804019672478,
 0.5298024681391283,0.15396357075543532,0.3640607532872533,0.7624770661573146,
 0.5709329424265853,0.40102664184704606,0.44361651385228207,
 0.8105658456929093,0.03735663472430528,0.4862551284207841,0.2553693892972917,
 0.5912850106480552,0.05037586813448591,0.4572882285080996,
 0.38790614618124997,0.48357202986172243,0.12131184034398101,
 0.9530720451603895,0.03563953707702161,0.5975389148266712,0.7750836847006936,
 0.04943362674013596,0.8763569370778712,0.9024793299274497,0.8941006614402763,
 0.9920510643867468,0.2838575087023901,0.7677404881541451,0.6681592983691881,
 0.30224866324817234,0.2667675900335993,0.4147263590843395,
 0.010542245334119626,0.9436311171862447,0.0394507120716725,0.686745622801442,
 0.35239728572393125,0.6111739324670419,0.7951878688723846,0.6897686779979368,
 0.032277489000387405,0.8499998175744448,0.4116009882609494,
 0.5259210822063407,0.11596587740313602,0.7629246757540654,0.5141708028527859,
 0.30439682045704464,0.03950225724789458,0.24490617764445533,
 0.12325084098560135,0.1143819740428792,0.3377246018392066,0.6972415843367539,
 0.15372411545522024,0.6346380189183732,0.8391973958315425,
 0.28695671643355547,0.3291530201247467,0.569146790007278,0.639782433040654,
 0.010735414765313034,0.6121221947324144,0.42142042290203885,
 0.21529297212041076,0.5778199245584517,0.9971821603494824,0.9412343274083619,
 0.6838648161653162,0.6727305353849583,0.4334197907170121,0.17932045917793638,
 0.9306593555283023,0.7891490204498941,0.31477795488017113,0.5717019030137866,
 0.9054063318626234,0.007229856393999201,0.5923883722059219,
 0.07782118816226846,0.09952373459344466,0.589774086534776,0.7337646343836222,
 0.8154091532220117,0.12913967082536404,0.4659461084061072,0.8080774603778528,
 0.575509044628465,0.23310088906556914,0.4438681297328104,0.7119400400764411,
 0.5777666311102552,0.431809443550232,0.7682445614878116,0.03094730621095998,
 0.9265218123164716,0.6974403020026634,0.41624916895138964,0.6993526835059877,
 0.04140738500833052,0.1680682885844822,0.7849919745397845,0.5476592413762527,
 0.7090429699288876,0.15768245097139422,0.600783497264184,0.4373127811343046,
 0.2069413403577407,0.783061114205678,0.34825196508922585,0.8338694903963082,
 0.4008819895761717,0.5921431734132021,0.3228185210362938,0.15000467806000828,
 0.11330260826640404,0.3489642075611482,0.4742572441755333,0.6804259046416256,
 0.3136708617676629,0.5671783500312783,0.26313620658662096,0.4055565939217046,
 0.21033909212131563,0.7944145488157084,0.47699842448988894,
 0.29206951634912015,0.12209683639089208,0.37534686466789746,
 0.9092016956562055,0.9283559032101929,0.7828912704865296,0.8321057602544538,
 0.9832736181630332,0.09926877377032961,0.9692425042716354,
 0.42428430082192514,0.2033235552648196,0.26345676561326414,0.610919464956369,
 0.9178587834707259,0.26639767571765804,0.4311181122020108,0.6026206737670117,
 0.4726290767421436,0.9067032774174473,0.13658068232228116,
 0.11684407608143554,0.6896359410865466,0.32107557455521063,
 0.9483749783480282,0.6600676856535304,0.10563353656840613,
 0.19403627473846718,0.12221066609739595,0.26166262313117616,
 0.10146173167853323,0.5773136905687858,0.6910214050416094,0.5939177165203918,
 0.25847373542623575,0.35994259590975564,0.5184026864529624,
 0.8973211445035566,0.3231874513675207,0.5849334222373895,0.9653505765387157,
 0.3069381779192879,0.30809326912195467,0.13619068068050033,0.943404384629908,
 0.03900713510201559,0.7407303502864143,0.07172611607028967,
 0.3778849907086905,0.4862069053396464,0.7790793607210903,0.21274091983796495,
 0.9207242679745274,0.7922931392367145,0.5768773870143585,0.8938791308499635,
 0.19389496399858253,0.3727441669244346,0.7275375654291839,0.7146312560162712,
 0.9236416960355162,0.6456598261836423,0.3829040865602067,0.48312680796415386,
 0.15090067053580647,0.1833618551213333,0.3207797927561373,0.5406957973332813,
 0.32202355093983637,0.1396295140753333,0.3140873468654619,0.5278477359465894,
 0.280592439768388,0.6152511750355493,0.006973234145719376,0.6596414437744563,
 0.7894771480000945,0.5956727508770503,0.7408006811834393,0.3615779932353824,
 0.40510427620616163,0.2799185247812056,0.18489402307557523,
 0.6331467158074282,0.03424198328864447,0.3955578357284705,0.7894445018573099,
 0.7809446443960414,0.7021950097519931,0.23558791784350452,0.6808021806400146,
 0.1276760896175344,0.5757633260432111,0.7906530683220718,0.4879285408689412,
 0.5398471625262259,0.9016272182324674,0.5548274062738903,0.8981937456112667,
 0.34189767441524666,0.9129095034965873,0.9395235318398838,0.9855546481888652,
 0.8540499161090139,0.7531177775386695,0.6877692406016478,0.6550847499565616,
 0.8964409703276393,0.7215730260143774,0.7050356007059533,0.6592716145162482,
 0.2427239683471577,0.7602553442430486,0.6872051066996078,0.12475359096149541,
 0.2803332704537209,0.573074240614124,0.12414597497963598,0.2027061888428492,
 0.2150042774136285,0.5632617968234619,0.9720288206459635,0.7769812088437777,
 0.853666104816306,0.6671917362941775,0.7023822840211988,0.34254553531647436,
 0.40727977670439364,0.05540103543569641,0.06945597782791124,
 0.8727397337751485,0.415146809738665,0.3083056753902459,0.25198960307498675,
 0.3350580705656707,0.6123952641959161,0.1619017192780262,0.9792360711646045,
 0.2427414004445142,0.5678101251430441,0.23610763433955595,0.2800388232598693,
 0.840527383181329,0.6331597883694162,0.9381514611934817,0.38828416664392307,
 0.9368691613112068,0.7271165601376166,0.7609011841440122,0.5287945526475634,
 0.0033580319380243084,0.8143716812842751,0.33207170504909844,
 0.3030368602274236,0.2664068490621634,0.2708196569455251,0.35489085219105154,
 0.23822129824384664,0.4444287958998404,0.8184450314262973,
 0.17262817644080009,0.01725898226633038,0.4629311592434481,
 0.7129099484291703,0.5178865838093266,0.28357411576213054,0.9891292211203219,
 0.23698010875778253,0.5373127082930704,0.8216735856596348,0.8916842945078987,
 0.6617556242665928,0.6960603446520875,0.15641839352180054,0.6098591511007344,
 0.6255899110655244,0.7355880004207561,0.3958979076647313,0.14791290830851767,
 0.44897576694247454,0.1668878586508461,0.3288082605456928,
 0.04067901883508518,0.5820496525986045,0.2714533502244645,0.5379523601860496,
 0.2775232521555858,0.5005834653551926,0.5372601242152029,0.9900673550278832,
 0.9095625279926037,0.878577416354718,0.43621076039877793,0.81578022134441,
 0.8274540622114324,0.35553631951501696,0.47694281968847196,
 0.8099208477965714,0.33826838284216176,0.584348480575498,
 0.004174566638906696,0.987583807783314,2.46463355005222e-4,
 0.4360104299390146,0.9607997192590062,0.22113468085206667,0.2706863521916084,
 0.6635791836501064,0.6486661127748266,0.5678021689092765,0.2681302472405691,
 0.9298806032603568,0.0935630558241588,0.9322065902120334,0.5091555122552828,
 0.059483251514298674,0.7600821815488089,0.5217123099742197,
 0.6349303769162145,0.21708062191229316,0.8510317779559307,0.467568111009632,
 0.3313807596713235,0.2948427072652269,0.6915024572868854,0.8610489340978484,
 0.635452222561671,0.31806859745926397,0.5624244111262849,0.5785884015261202,
 0.9544712933623533,0.5013879896873769,0.8984115854195867,0.9533452161800536,
 0.31905443118202914,0.2781130603463736,0.350938721165856,0.41929775283405357,
 0.7328676784061948,0.23161324954207307,0.9211482695730255,0.9928098002853827,
 0.10576853837571998,0.94854761231069,0.5181433358426017,0.24958552721069183,
 0.012423086142125728,0.7989793231792655,0.1943000450809279,
 0.7470091645047906,0.23850788377598242,0.23686098705321246,0.573454873855722,
 0.16094517155087484,0.20907918711639595,0.44767577749290655,
 0.9316420173327234,0.30523325938104096,0.5193074366501105,
 0.14910602641317994,0.31793946498614567,0.15220112244344897,
 0.6399458876858027,0.749928027506873,0.5795399510858558,0.011338561108156231,
 0.683352296260675,0.18496105927290096,0.5781004352298108,0.04641297302390779,
 0.7316871292868335,0.7235777878175415,0.1382028809127287,0.26366187513013917,
 0.39543668252825415,0.5864727357857112,0.6656194809725491,
 0.27673741908067895,0.10979811905469294,0.7949502603811143,
 0.4514960779480567,0.3701525232762777,0.5188106174808623,0.91182118411577,
 0.8831292270178221,0.24352126314900535,0.6491482943439171,
 0.030314308930067146,0.4906093260200586,0.8472141640880892,
 0.9541581097729812,0.8287869247600333,0.5436618668322292,0.9556642714753301,
 0.7735464390907416,0.538845248838353,0.021293791782616367,0.5525103362254515,
 0.3487105484072819,0.20117991126999224,0.10526247552790458,0.765826641033429,
 2.985005067760449e-4,0.5366950010254583,0.9167826336453397,
 0.9571830620336591,0.48978459455561263,0.606456803858519,0.6789396608382607,
 0.9925348794554194,0.8088163430168527,0.8006750661737105,0.6541441995951789,
 0.8804444300575233,0.5975788187174564,0.9936974607849162,0.5608090802311175,
 0.21354783430084456,0.8666070064256165,0.8660418619581537,0.625168985750225,
 0.149276521906757,0.44718020600256636,0.059211332808863704,
 0.34158643385975007,0.6330308183427994,0.03912474535909205,
 0.4982611710645396,0.6819809469431728,0.3630929894763163,0.4221964206291773,
 0.22628709852338003,0.7471273799685271,0.8132292020072635,0.6453556258223869,
 0.5475452940519407,0.09115917125088757,0.1831410359115745,0.74965179452574,
 0.7931638930160901,0.3172330043822018,0.6747421585695382,0.08240647107822074,
 0.6816907607782088,0.003414596555844085,0.6648659891726489,
 0.13199785172870204,0.5319475632041613,0.6835453316644375,0.7246626412440167,
 0.8081621947905234,0.4266871638676044,0.08428524917769209,0.4304752959813747,
 0.9853064370096388,0.30004944279589396,0.6577053580392059,0.7714753826814286,
 0.20266481133127212,0.12551406592144732,0.5105677593596403,
 0.3205237477959365,0.5761546888410114,0.7934002759774798,0.9085507661091015,
 0.5725110321765613,0.0673513120298902,0.5400918836770211,0.8199042665266547,
 0.787462041547645,0.47029157072658356,0.6158349959381187,0.3201121901270926,
 0.01867653346294018,0.4518485570275942,0.37073633117775706,
 0.6313105190322968,0.8112888779533729,0.36803437639267744,0.8575295709992263,
 0.626467864865244,0.03603398236454558,0.3454255942810732,0.5872948264051664,
 0.853716738305268,0.99146199469877,0.14015140468387743,0.23821263709872342,
 0.8640082656226601,0.8650357605013478,0.7007558057271458,0.8381750629406338,
 0.78879057884192,0.1988817963975119,0.8317062807286338,0.8160528119700098,
 0.07008743500485826,0.20991464406903027,0.7533428652552473,
 0.3379928676075078,0.9312688426990309,0.3047056197022302,0.47027525034730244,
 0.6304762867008743,0.22444965103964631,0.3051669125608544,0.572254611712069,
 0.6539412948379811,0.8075049938826806,0.701632577646472,0.841034564880589,
 0.4394460368626385,0.4229455462250432,0.2085543418278245,0.7678150823225193,
 0.43909449692385827,0.053838939836399424,0.713176671683483,0.292574433715292,
 0.925340957790321,0.021186041888038076,0.814989068432712,0.1212033183667135,
 0.18618937935434643,0.7861085753118264,0.7553242602869495,0.1899542972039332,
 0.07215365660500905,0.16810933498904845,0.47682039278433574,
 0.7079558061841846,0.46745793313494044,0.6855481905967933,
 0.11141244030458708,0.20698718696143548,0.1624889616234566,0.926792112525355,
 0.5780566091904452,0.07254635423397504,0.5642576649476021,0.656905653932147,
 0.23769230917043316,0.8194103755006251,0.5732221705557538,
 0.002005922337739996,0.3974359211994748,0.12221560955718547,
 0.03812642149198597,0.39404934413511183,0.3820448456318719,
 0.8001928779346659,0.7756575198611217,0.466741702538912,0.8842795082425523,
 0.6922839174208991,0.766768870734732,0.44761675859466066,0.968135891455611,
 0.8135708015547357,0.6803529765001695,0.5464644106742415,0.9210350475933743,
 0.541405569035614,0.061133079424225634,0.21908106709934472,0.759279063614683,
 0.39441718869737685,0.18166457810482495,0.6884647765907544,
 0.12241791294381132,0.5311625406300009,0.22385276830212342,
 0.45203096104127793,0.7497116044954225,0.6314401451817682,
 0.42574416201861076,0.09546940719745622,0.4592631140126284,
 0.11821511071253388,0.31773739418362523,0.0192476679703385,
 0.6767574712253858,0.8090578575248162,0.42114289272172756,0.7783288736044653,
 0.8528245866600108,0.012895225187092296,0.07845026044734249,
 0.684325108164188,0.3200311044494616,0.008879272256804516,
 0.17242474452146894,0.3700153953327534,0.7135114144968809,0.4802884508816112,
 0.5895466772121765,0.9468253785068007,0.37053225191405426,0.3624024466672504,
 0.9342271757789419,0.407286865429216,0.40024124131582717,0.811919489455296,
 0.9088830024226634,0.7205673838844675,0.3476253022266107,
 0.042515006010461986,0.750734353967817,0.9045352221238451,
 0.17637101881231554,0.2143135680288899,0.06089900319624342,
 0.20798151907507956,0.3106154394712375,0.4019622704636455,
 0.18751013098150082,0.737832275304627,0.14456869729693234,0.8592410187353203,
 0.10770782787400401,0.578955140147669,0.3635836141025839,0.1914909074679767,
 0.6161674142732672,0.08231504030069525,0.15366502028078788,
 0.28780816297582745,0.5691388304567211,0.532449914510611,0.44981602589563874,
 0.041641072923254474,0.49598434174227746,0.9330382010390581,
 0.1913557331409681,0.6561474810315666,0.6862938017570902,0.41546856320454106,
 0.4697702746188064,0.6885604407193769,0.7459038167395743,0.8037334265121444,
 0.03860797370886504,0.6375464352753816,0.519219880646061,0.6130020999461064,
 0.11165157530042225,0.6235213814402545,0.6130015981291127,
 0.46310368912243005,0.7396327731356237,0.25062326783681577,
 0.8584584374330917,0.5335660178027336,0.40663987901075405,
 0.39418330803579915,0.821591622662855,0.5335845358911314,0.2187217135386671,
 0.7962536438066715,0.26448734459596857,0.9290261165052844,
 0.16761671237165443,0.27690724612469353,0.026569211060933018,
 0.8976850437242403,0.8183710210143795,0.5526231907349864,0.934666332267309,
 0.8631928502323933,0.7430795659693872,0.32198960368600726,
 0.41071879353088336,0.6806943520311254,0.545938081380179,0.9626938493888868,
 0.6234060121187492,0.8895896221792178,0.6377176340192418,0.4271593814743124,
 0.2500573096384431,0.4940271755957517,0.24696724057196884,0.3530939330717012,
 0.9533592279273688,0.6176217935394659,0.059142967464375196,0.646184199984859,
 0.14133967508080048,0.6171325993347139,0.4925151407379572,
 0.48637223192300805,0.6034941274833548,0.09667723463623623,
 0.7672648983336461,0.45553730839388185,0.9621166211003338,
 0.019305202637060148,0.6256971373141839,0.19991344672676237,
 0.6790540364904881,0.45172242265824625,0.6356397413919102,0.7074129035763538,
 0.5693125037122247,0.9094820008710458,0.06157122789545921,0.4083138104066827,
 0.8537788210462715,0.6237658690880168,0.5285721627741168,0.7208939325374193,
 0.013919703350013912,0.4793006623766686,0.3853295800664759,
 0.6599359425923115,0.4822606815935341,0.044818186365974566,
 0.5559073905823453,0.7366880577176164,0.881530432458857,0.6094433540014838,
 0.954579096288693,0.6740414068368418,0.31204987747569823,0.3510191985379987,
 0.0991045561083771,0.481212939480426,0.8919350859198406,0.4127963743214651,
 0.7379775727613944,0.42124749250452975,0.37547414418208147,
 0.18485583187618726,0.3455420369623652,0.8964150495488914,
 0.029282425983378713,0.25895607978853596,0.8274688220764715,
 0.2950274927797094,0.7479421298781757,0.7028498097882094,0.4138393705114358,
 0.9534370167758379,0.38917242697439114,0.07585259440246694,
 0.4109702634226262,0.5558164308176959,0.5972917906962394,0.069306649522623,
 0.8869313316187344,0.06984818819180982,0.6722793692925125,
 0.39618937112166563,0.7261492720896153,0.44410056803479114,
 0.40831207910798517,0.13744878514948755,0.2616931171029593,
 0.04030342540825549,0.17137257064955136,0.17884295237652337,
 0.21707563378497163,0.4747087267759088,0.4929736460475409,0.7581665428687013,
 0.43195631108320465,0.05891560446781574,0.14185453144505222,
 0.9631209827039466,0.5511361606674101,0.9984137743224534,0.537015468752354,
 0.635340230175667,0.3800939158642239,0.7410919235380007,0.5441278324456601,
 0.855662967617615,0.22160425228846792,0.7922591638055434,0.14032463556604913,
 0.3704120283210667,0.5062477574306653,0.3703750788714615,0.3334557898584586,
 0.821967652847904,0.5995840549616972,0.45470683913182397,0.4092829852879234,
 0.9030638384183438,0.4916756791310197,0.39409591777776565,
 0.18269470973865365,0.8986946514542936,0.2582707808402279,0.8120229900906452,
 0.30797972058534173,0.4236263197210862,0.3709639156542468,
 0.24458816420856444,0.1560110482145055,0.4002703069111624,
 0.21928426067640028,0.013916888267176297,0.8889939262463872,
 0.862055247355608,0.853491497263339,0.25963850453482673,0.5812068407270555,
 0.696253989017953,0.11900373399013053,0.7655699226431768,0.9772535754323691,
 0.5215322161757224,0.16138134493308154,0.535977688541111,0.5194556860952181,
 0.3783639970687761,0.6745671964652087,0.919976620512392,0.8805507462361699,
 0.5699072659309073,0.5149711552520664,0.8104638979582331,0.5884219340444176,
 0.7545963594619747,0.41717824501352696,0.10664767900312855,
 0.47823074693501244,0.9984437449383099,0.6237458887306959,0.2820675516538743,
 0.6789294526653445,0.5751612627514326,0.7323631176661698,0.29885072717290306,
 0.2011362122194117,0.0015386340102967466,0.8191292811639418,
 0.5674251849548577,0.666019372567716,0.421373661429993,0.6961827417242479,
 0.42072811056080284,0.22582577232479017,0.7924447099428351,0.676078459081287,
 0.5214775404674791,0.8752486218166831,0.9725042610470679,0.3154512226502907,
 0.01088953087698219,0.6400360921826517,0.3486226061212294,0.7395970349741129,
 0.7643963448203658,0.30933370276327365,0.11834828893774385,
 0.04444674221501088,0.11935010313658256,0.3334998349318886,
 0.5118329816216404,0.10321671531555998,0.764818194448103,0.7269400347846575,
 0.6397940342292452,0.14646372237484662,0.7327227157035141,
 0.27737107855739696,0.8016452029192428,0.5185562394405556,
 0.13372477274519878,0.6226224505808761,0.8113287485887513,0.6410990457502216,
 0.6903136076308289,0.1759470640105315,0.5132006195560175,
 0.011035948597170053,0.9314728203198583,0.6166463526763772,
 0.49480372036839027,0.05444484975950448,0.7308830634490208,
 0.9854452920597204,0.7265689840776794,0.82009405734558,0.8812060540470975,
 0.1621585007313775,0.021723615731385926,0.4384101350362143,
 0.29642116151347186,0.15368186371927417,0.5171831709618833,
 0.7439359814297539,0.5788822250713413,0.9889683752708704,0.5674664027802924,
 0.7197736186251901,0.9994937463756306,0.5631909419746537,0.44958889778660915,
 0.47395478027823623,0.050412009802309976,0.16792702557718642,
 0.8048743196792914,0.25199925711805604,0.8847196253439927,
 0.15113526686921563,0.4833412650856458,0.7807506127308157,0.681559330726266,
 0.16008172461995318,0.6606251425594072,0.1588116856440317,0.4665523104410241,
 0.19508924215807522,0.821891912508742,0.6358719707345382,0.8441626081807205,
 0.8614825828637902,0.05659608704942787,0.4160942965548806,0.2810943781093962,
 0.16212847422073529,0.9199010109633586,0.29618241203821005,0.066396261876696,
 0.7636756799933319,0.7371083875399994,0.5751485677126602,0.23420846196314893,
 0.031115042146098393,0.2516271713121849,0.18774343446621655,
 0.8742099750638663,0.06427599334036793,0.8940932382265881,0.8149561635019186,
 0.5647487976971128,0.8172141920802367,0.8236922915505562,0.3935581680963778,
 0.6694654184279916,0.857080951583852,0.9847197561964725,0.9818883984448237,
 0.007178413427938235,0.9825263561338187,0.5178915947881748,
 0.3311127830730386,0.4043419494502021,0.8589702911403203,0.5766170089822035,
 0.15939485582175883,0.3578914936574564,0.4638434422615051,0.5353514521604797,
 0.38932004041589574,0.8461338598505762,0.6112311115320508,0.4231703869156651,
 0.47527189731142117,0.49988789725056093,0.4143483181434986,
 0.35443840718970254,0.6283837950289436,0.33513109229960814,
 0.16722018241508674,0.7520926836359808,0.6007264092003821,0.7718868368220708,
 0.8034959118937025,0.96025917180411,0.3888854319922339,0.1385971786208674,
 0.32132527948533196,0.09191698200840293,0.017573615268048015,
 0.03933014526950218,0.47369954945615433,0.38145262944723923,
 0.7154102140319971,0.9596006750885135,0.88161350099027,0.3983202281147803,
 0.5452461111550958,0.7890213946955047,0.6970361803667403,0.9792287191011377,
 0.02310010187536715,0.4294147263520086,0.5606908386654361,
 0.22452398448524702,0.045193637986832846,0.04201822994669513,
 0.13466009348089358,0.6661230764572807,0.607228057456937,0.5743448939973709,
 0.09913275241116992,0.32138945649705675,0.6817050032011502,
 0.6019880696191187,0.6371799050749726,0.8989128936887111,0.44515269746730746,
 0.7659502341940763,0.012292874582108482,0.5078922834687448,
 0.8533408746626653,0.25696897498067417,0.09077390945225661,
 0.9454594811397429,0.5978847055005462,0.9822573219327185,0.6702078523806777,
 0.05933513965017356,0.9720355808687319,0.4709695156664382,
 0.49637958928827064,0.28947970737171924,0.8009166154307586,
 0.5724722795599223,0.3829540180410156,0.07587531883764154,
 0.35934761977332785,0.6144521798029723,0.8688749968137901,0.6014873096012265,
 0.26785081524589394,0.584367661030698,0.8302717462938859,0.8903878784030022,
 0.4191360723593418,0.5725021176845568,0.7429550536968635,0.7519943369435996,
 0.5661788382831998,0.7249899771077027,0.023254232931846697,0.723704385127519,
 0.5398482770411721,0.7725911617483053,0.9829562441789653,0.7669068463206224,
 0.5208896365836304,0.7243197965675888,0.7586675147266483,0.15817712426660435,
 0.48749439907953707,0.6565389243122117,0.17838986561578551,
 0.3718198835517593,0.17284932291822663,0.541920478193565,0.7975042121043876,
 0.9897852557851916,0.44317860123705477,0.32241937792803144,
 0.18742082529971005,0.6652342187079142,0.45441059559454977,
 0.7247896316507669,0.49470856487683545,0.4806828722546891,0.3966768108927874,
 0.9949347996307251,0.27579575356457753,0.3417452328035536,0.3784668019685873,
 0.8349985357365065,0.48236491544605464,0.027870577359759174,
 0.7438727710176022,0.043369034584275656,0.42620116484173365,
 0.2897600919275064,0.6326346359707129,0.6450222311364154,0.3399340635186381,
 0.186907140350263,0.8341160385043745,0.8064762311648418,0.1908338132084194,
 0.26543849916597795,0.23606813254097092,0.9037591700958104,
 0.4474248597951457,0.7095950729173621,0.03593943017643719,0.2515404107150724,
 0.030445545620016334,0.48999108720807427,0.009659527267408263,
 0.8377822260947627,0.23110929703945138,0.6331013080479431,0.6879602465327448,
 0.8598538871442932,0.28782662034145534,0.5720357982899743,
 0.30867926546436153,0.15932514460858838,0.8963490711814287,
 0.5454746778929183,0.09418488205697195,0.4933161156207464,
 0.42874345471723985,0.7993672547471475,0.6226577920715066,
 0.44254811295232277,0.8760881104169426,0.7623627196972173,
 0.050090007271779924,0.28450983491814874,0.14714946265359552,
 0.65980758420503,0.21337398556750087,0.3649165125302771,0.58649883499482,
 0.7043988435327391,0.43748080180318116,0.6773714767692531,0.748180610322014,
 0.6666605823600901,0.09042396429018285,0.0789072702896999,0.909675430698491,
 0.17519794411634348,0.38993548800629263,0.5071231091543364,
 0.7334048905487311,0.17546870486724275,0.5466699237092808,
 0.04006412238730084,0.6166052575369505,0.9230641308266515,0.7618707419743822,
 0.8817522331976535,0.9446060160495748,0.7630617988051148,0.08982616705523228,
 0.7565055003360077,0.6880371154787753,0.07878908381164718,0.5358524825323068,
 0.6856474553439269,0.927747673949935,0.8138708815951788,0.610881023610895,
 0.870615871598527,0.8094461241009342,0.05282532586001232,0.3794539480816128,
 0.5887723576039925,0.09140619235050218,0.5341301069777662,
 0.46342167810060897,0.8555924563928555,0.3413754551212693,
 0.06244065965352685,0.8509942414983084,0.7348184895767003,0.7147227741299326,
 0.5165387284885626,0.09001228428322694,0.8835266679997644,
 0.21595854490663557,0.6731909993528913,0.46885260391954797,
 0.6801548872821449,0.9116411935084328,0.613795968628286,0.863631345847558,
 0.6996562047199152,0.43016649535022067,0.5544677739418815,0.5886944721188344,
 0.16645721961560833,0.14328453493838267,0.7499788712543819,
 0.9206104788390402,0.3816975907028697,0.8209057869980153,0.7654980234038825,
 0.9740053665337255,0.21389259735852384,0.4766009634863816,0.5993512337756081,
 0.7990962605396504,0.01617107946094043,0.22688923997731236,
 0.4574960585911626,0.799068777241588,0.7000924264104851,0.10404858185544508,
 0.3736868682705451,0.01633588977228495,0.013487940422173383,
 0.6624311808041238,0.7877177495322971,0.27988931979557086,0.6320640978834756,
 0.9217522346264538,0.7236110718040639,0.9024717246645495,0.8015620513673898,
 0.07500678343707567,0.36823284585955063,0.3240074789093649,
 0.9279856875276665,0.5359324430620035,0.6400323529515322,0.5585840938417425,
 0.7215416015302474,0.15131407762324012,0.008421577795340918,
 0.849479737817868,0.5591647338603158,0.4942339526746341,0.18376364109832843,
 0.5838179363001692,0.1973701972539521,0.6065787468520817,0.7863207313225233,
 0.9616772147894896,0.8009736876596967,0.7713673387189628,0.6811977778722466,
 0.8190052639923432,0.23679184713013024,0.9959460970925585,0.7524936429626815,
 0.6704927533613044,0.8373399935030166,0.2443897130090683,0.5594122046814243,
 0.48139160955796023,0.012154737296524387,0.502835922548781,
 0.21196392687807397,0.0414118422752453,0.9501528869434334,0.3631279330576819,
 0.5629849515821636,0.3396881181007285,0.9194155411498705,0.6663973347417939,
 0.7360720570165841,0.9533250475197819,0.26464627753859726,0.9146863893699875,
 0.6378447623731368,0.2929615516617261,0.12237169740060294,
 0.006119216506281777,0.08649075533324146,0.5217926274707815,
 0.5062076711621704,0.5851207688649955,0.5514527241024688,0.5297771308515548,
 0.9461034594724519,0.516295871540496,0.7584240977639952,0.847278855493802,
 0.5613824838435254,0.07152467959457454,0.30979490436895785,
 0.22044432766255984,0.9738341485032845,0.17852139945221057,
 0.46743200970178267,0.31786186894377644,0.646025867881116,0.4625976709077889,
 0.9876342984693013,0.17013962306353647,0.9639536248911943,
 0.20225792304927281,0.4679393856701606,0.375180324770815,0.9838454117663656,
 0.5398725519320198,0.9818270636285944,0.9427292117688602,0.7881904602618566,
 0.8848812574253893,0.42796607215151483,0.6984516959037086,
 0.36242124858986713,0.5518015882479437,0.39421251587510564,
 0.7134664439855448,0.7361430791216041,0.5795265199552775,0.15289373201932988,
 0.18675301771927955,0.845001310229045,0.23737837293128794,0.7363111172382371,
 0.6021334252672416,0.9699310371471402,0.9272201471138946,0.49688238530632634,
 0.048651585652300544,0.6453302606112301,0.16468565713667727,
 0.2952649153331789,0.27891890686425747,0.9448097887979797,
 0.22300515851211722,0.15777450172044116,0.6615022792213009,
 0.5378131275161822,0.00791314590822001,0.9810460400846296,
 0.04690354085890536,0.4073540970924153,0.7033935456109744,0.4586254103912003,
 0.2233475837382869,0.7874340073796862,0.6841523581619419,0.06471122396405915,
 0.001239659275647842,0.9783275538485697,0.3540525577316995,
 0.8069050310231034,0.29869312591322783,0.08219081091112035,
 0.9388589135387623,0.21753706201072165,0.21612882498369146,
 0.6532611687166032,0.2894486165035004,0.9144593634671061,0.13964179352157624,
 0.8001636389915512,0.03473777206876405,0.5765338101130326,0.4560175743037699,
 0.9701104596490685,0.7875546792999638,0.5467975631075507,0.5795583896520118,
 0.25023688218879836,0.206337873929897,0.6384954779545664,0.8287447347478418,
 0.09478739219918397,0.6046825017133446,0.23077608129010174,
 0.7619025615229801,0.18701576208766735,0.07967235766308634,
 0.45818523550633616,0.9370305172808416,0.31192627065848244,
 0.4882743447084067,0.5028636887946024,0.5544680032035239,0.7434943389183291,
 0.806448250386579,0.013059875969023471,0.1880639979905867,0.8351646763342886,
 0.9451965404403591,0.5120243645754611,0.3577060645848602,0.8483411789515347,
 0.3766658655804499,0.264214206507958,0.2190748272824261,0.7792229553621738,
 0.04258635375698794,0.8805317381439929,0.619530430761889,
 0.021650019600924608,0.4987402267608967,0.2125830994197042,
 0.6281902616783838,0.6494968976408965,0.2978501037504174,0.7628292529195815,
 0.25117343390227465,0.2316876757093741,0.8576505740696152,
 0.19032671051122463,0.35416241414490457,0.911016397443003,0.1432389913299017,
 0.5003801738689181,0.3282736777842893,0.6085865416688054,0.7633835160002483,
 0.5379212434448563,0.49274042810398233,0.3421937737646831,0.806740395580929,
 0.8588563535266381,0.4107203382596508,0.020056867672796352,
 0.19261826581225305,0.3789156948416592,0.605240414055394,0.7710328591512058,
 0.5997387301307304,0.9055038285051066,0.19490592137949503,
 0.22962852622510865,0.10406019529857191,0.06443548403460386,
 0.19692011306676227,0.7593354989623579,0.4957091724607817,0.4281629028287113,
 0.6502549389808205,0.529406301273915,0.7468685282961636,0.9923962095480389,
 0.7246403122771383,0.6269756247743703,0.8791217787494342,0.2678105642025499,
 0.9526912263986971,0.4044926496007659,0.7219722397203848,0.32820273993754034,
 0.7491961730445353,0.5356280540480096,0.4147346188275933,0.3264980209653322,
 0.8138369125616755,0.15561949967081656,0.47776383519675814,
 0.9687813462268466,0.46848073360542797,0.3931394362736029,0.8642565717844028,
 0.8063672233120361,0.7335666766891206,0.004189104834540491,
 0.25214639362458635,0.4664062369502201,0.9835674844469772,0.3269211570120222,
 0.22500332213891028,0.6376075488814033,0.10718804402961823,
 0.23660646093683668,0.0334808653370281,0.432831921170972,0.4084657022306062,
 0.2522256207952318,0.5411832458709622,0.0023428813617667466,
 0.6095756111253134,0.20679164372151881,0.9196716217773966,0.423175204151777,
 0.7944363893225151,0.8927619771499816,0.8378378858974189,0.29371296736318975,
 0.19626972300367296,0.1661793939249454,0.6191708433274221,
 0.15146786774823684,0.23962481810710767,0.3438429844249119,0.711905272131389,
 0.5044399305794262,0.8033088058327763,0.14261835763979003,0.7010610750328005,
 0.15540260640507064,0.47538077887910624,0.7113026569038124,
 0.4059247769277292,0.2628882683974266,0.23731282090251427,0.0394914846403982,
 0.18672816648988078,0.2496788776440626,0.915588303138511,0.8278684178290154,
 0.2875636206740576,0.027058686042370272,0.9542314950540634,
 0.1947243297547272,0.008886523872217666,0.8688604797810742,
 0.29905694283782536,0.17389625774478934,0.10652062273992569,
 0.9022035345829673,0.12643957395018268,0.8817240776052163,
 0.27778418545119166,0.6593941005187979,0.07203889467299107,
 0.4625121825987193,0.2363695635269072,0.5675415210391418,0.8268000504232427,
 0.6081398127998783,0.25994493096912374,0.29683547164708,0.9985296507393376,
 0.8620550662058369,0.6502342040959819,0.3510340216825534,0.903073088778713,
 0.9354266062986264,0.8792382357733424,0.2879510931283953,0.7354231336792891,
 0.6293174073701613,0.41162281677870394,0.5732955037227082,0.8041517894815154,
 0.2575841772372436,0.12470389828492623,0.35963422025712066,
 0.30421315402973836,0.5717011061389996,0.41297593595158155,
 0.3790913475646125,0.4993066248880744,0.9646875701001023,0.4875808108471451,
 0.6655243272662722,0.29039161361317434,0.32167474467727997,
 0.5253183512118414,0.34931333219041416,0.8708608108225393,
 0.33041407399887746,0.08351299798730305,0.05866090579187244,
 0.1364089839572894,0.6449606429919148,0.608337645846123,0.8933106148741109,
 0.1674221848082197,0.6072597924978622,0.5421468339333314,0.8332358937272513,
 0.7026343574161273,0.662472798630857,0.442046792312187,0.01499628873995984,
 0.41954240180809577,0.6377901881696656,0.6106091904502118,0.2952679034940403,
 0.557565505479769,0.8172616317922137,0.28455427285692014,0.4435066807891169,
 0.5053708037987217,0.7882768975140448,0.11935316718352085,0.5290388746087904,
 0.7520695959943949,0.15440879780432049,0.8217427369257155,0.8269220932019185,
 0.1144711537483114,0.23306204902333882,0.12736158536290643,
 0.5474310309820792,0.8158397993184053,0.007710634531016725,
 0.08599184380513691,0.166791864530628,0.31314979576608426,0.9243698804473526,
 0.46371405348624617,0.6416037415472053,0.8479611280548284,0.3030445780315437,
 0.08957956914833098,0.012589008326562956,0.10226465656747163,
 0.43768824987356814,0.7098287555756198,0.7067522958508539,0.443460669777929,
 0.11125810608438791,0.26458191985549895,0.6718978858312727,
 0.8775533195586375,0.2408286787742876,0.36867781712146996,0.1398479628954603,
 0.0786704468341598,0.4347230562396306,0.9722558511410571,0.7764768568685306,
 0.6356456822138641,0.4005705374815842,0.6891969454357997,0.2450289464077472,
 0.5694578530894723,0.7290949088008645,0.9047778611430934,0.5017824084148221,
 0.3302783089936112,0.31510546697027497,0.2512686887827176,
 0.49250168022140983,0.22618379647364362,0.36890539080591356,
 0.5062935272741267,0.8891273408334257,0.9865645586704288,0.2922893251943044,
 0.6385441251956827,0.6816946880486203,0.013490009265015979,
 0.8105706688225637,0.7322960748243978,0.12817720131801258,
 0.04653888273019713,0.3506548844778997,0.9240867295900288,
 0.41065140571347747,0.26246666030074595,0.6015984116443683,
 0.31806534503673933,0.7198492714520058,0.795616787660353,0.7878233095833126,
 0.5891502962232492,0.29176263801905133,0.8256724374257349,
 0.27856289234028786,0.05956864405604234,0.2679228980424736,
 0.7336061652322765,0.05700888280362937,0.7904693642625488,0.2647133073065764,
 0.05293678839737648,0.07196084346932319,0.9032195429921078,
 0.11393720679341113,0.22094905554881628,0.07821953956206262,
 0.4978463195878041,0.8931074469206969,0.6919772664344389,0.5978715793055669,
 0.5503417961452328,0.8273271171789408,0.16692592738809275,0.7648253159946682,
 0.397259928294566,0.16016912285811236,0.04076351039877302,0.4287795525085071,
 0.3754685291952162,0.9826947500192836,0.4821672055811993,0.212965946580424,
 0.9385142740723659,0.6373807397386653,0.7108025762517969,0.5456806851404189,
 0.313169822927471,0.9064515888241832,0.9767734576666434,0.07428228213014298,
 0.36844610970477565,0.5549391944345833,0.4441704348653117,0.6496154111204743,
 0.10726088704220993,0.3492991743232894,0.05755180772941737,
 0.7004510829296827,0.6192156693964673,0.7404971425566852,0.22537723743550786,
 0.8958475535753258,0.3780893125320419,0.2197794741145047,0.4479073905588059,
 0.43899352839157635,0.9277366061600388,0.06673299165270308,
 0.7287520043545385,0.8870640833870339,0.33350844882262776,0.921761101078404,
 0.9766567375643801,0.019301205005302746,0.24410612595263037,
 0.7071341114217563,0.8880094919393283,0.1417856635630812,0.10676782387241857,
 0.17814086558520326,0.8666970299048611,0.9349058798186232,0.700521530548818,
 0.20665451095761456,0.6185268721770649,0.672792979761225,0.45717845127976997,
 0.8892470926996593,0.9867821561213143,0.16881760529886125,0.6774173940961122,
 0.13773016405073446,0.8362285600448187,0.20605736427530108,
 0.7143305641804243,0.0901233038606224,0.26530532723412326,0.5886166751001846,
 0.2428113205304766,0.375639957070226,0.6296014366734899,0.42003672391269753,
 0.6839215884450536,0.6564007722795256,0.1018079330838203,0.8644692838299427,
 0.9084263088256572,0.15339138237879202,0.829309898753807,0.9280775763442495,
 0.49329153787083047,0.5154823747573822,0.09863017793260764,
 0.9857300311161403,0.842696999228169,0.843640367937274,0.5356039273028901,
 0.1479341082101634,0.2996599371936759,0.2656510977748129,0.7212199061904471,
 0.06650105583472399,0.759777179101037,0.921619413438899,0.514934223920851,
 0.4963685362094512,0.5021387389728369,0.5276109420725673,0.519031657925743,
 0.844598512640672,0.04143233997573759,0.04102033714021747,0.1319382501792552,
 0.9977246164132265,0.10690016991950546,0.3311574274326332,0.5345721681342785,
 0.722829763994593,0.23509311950196643,0.15742596722261737,
 0.24904844599480291,0.7725801840711121,0.3073378397331057,
 0.47791873102260496,0.9074523892136541,0.26592216147847614,
 0.36015095487242754,0.9937879624896313,0.8024049939244223,0.8266883409560268,
 0.03556409622441392,0.4400452470195261,0.016594835861989177,
 0.9074513622767868,0.07211040279783987,0.5224343179973149,
 0.03843740029855802,0.15318253512006264,0.04096371046154701,
 0.17849038095348368,0.6196078656937809,0.3708441326853015,0.3160420766472851,
 0.2094995046255438,0.6621747452009803,0.10624936052181144,0.4605769067684832,
 0.6553380705996802,0.3467473168770985,0.4353651317449607,0.5318133819113893,
 0.14419244243916574,0.1191522374712487,0.4058747085637032,0.5035296026440595,
 0.17895352800612996,0.8982274557132536,0.6325675004177038,0.8986581126949026,
 0.026620309303259204,0.06221107175227969,0.6441637035501162,
 0.32848032907727875,0.6836990876303286,0.22365824767895748,0.300656301399725,
 0.4339328927680264,0.7092942883613715,0.7964410595296331,0.5227452384721443,
 0.6740798587353517,0.5237755388238613,0.7122034181692348,0.6132777354354693,
 0.5710452577458939,0.057697903227628355,0.47984384931975277,
 0.36762843028070735,0.27604280949231574,0.07442942876901171,
 0.8902073388543957,0.8183317755390629,0.49731760257527013,0.2961417421145105,
 0.3212602973289642,0.8182286119184545,0.7039874381537675,0.12605379476360068,
 0.697588612665943,0.09585176133534268,0.9571521431452829,0.2398962901328916,
 0.9179827149601761,0.7453231495396428,0.5651987631706052,0.7267764232640802,
 0.2970672154284312,0.5829222321036671,0.815361420194964,0.8634965994954361,
 0.46671358854171896,0.8563215804925315,0.07079436963475916,
 0.17835603464705008,0.7977622738157472,0.21211114403777387,
 0.12620123822412432,0.3967275549635845,0.23999893598401478,
 0.03843186699317125,0.5297870953183237,0.1441520100280399,0.6339081534025685,
 0.39861727002248715,0.04537494809171527,0.4419112766822537,
 0.38361386891309424,0.6325588588994017,0.7286479711605013,
 0.12024646996216448,0.6994329146594009,0.8622247591767986,0.5633840022214511,
 0.9276497532525365,0.13019177272721516,0.6704165993947476,
 0.01695040504933698,0.8375348447413836,0.17966194135018176,
 0.10571985379185722,0.9356884433515292,0.5120342961247919,0.8156660458344626,
 0.03910334299927276,0.41875305742773083,0.07935069561442942,
 0.6386777863811736,0.10070334387585178,0.18579079118129838,
 0.06333295395103655,0.029653698264476502,0.9727600237716159,0.93980934968768,
 0.25384431632671045,0.712252151683286,0.4743815685329843,0.4498345826662702,
 0.38483865150275665,0.8269143175979465,0.29196805937974524,
 0.6155861493985284,0.47165240054210034,0.39645089925176136,
 0.48792908868461105,0.4034747972603291,0.2265753828986885,0.858768567378222,
 0.8755651398903981,0.1790254713140289,0.7989839603556544,0.4291044642478308,
 0.4356876467784442,0.9516843170819891,0.8378060717751311,0.9952196424563189,
 0.17397603363122238,0.5051276064755782,0.7695021782487665,0.7822906110895131,
 0.27677179104846716,0.9488581413087357,0.2723382964317662,0.5410249327846433,
 0.8118504245680838,0.2956514099381996,0.42917695325921823,0.4024484833010926,
 0.3469131223035984,0.655992930301743,0.9542782581904786,0.6231415134770347,
 0.869262806011581,0.04333746516687664,0.6772740441745685,0.7070114497717563,
 0.13377117837643604,0.7687657647351694,0.18666750665510112,
 0.4493271989083518,0.39246454343807924,0.2899599558324617,
 0.41601997864090334,0.8534600124198638,0.2163740761210975,
 0.42892003484778574,0.513021113408089,0.1818931497616061,0.5388558937146712,
 0.15915374244814706,0.9482531831692168,0.1908745200093408,0.7516679958855207,
 0.7118015640266566,0.9014709034344544,0.7908417814066144,0.6980025216854725,
 0.7294555560391733,0.9410554502085714,0.5000595626067552,0.5087057890145219,
 0.719837809813502,0.6292198201745063,0.12983661765480436,0.6034386607443829,
 0.5938889986969735,0.2644133023605353,0.39906253149952076,0.5745243623192879,
 0.07901926278448757,0.3770092604097264,0.10962535027629294,
 0.8635464501104974,0.8936844625230244,0.622253924496697,0.8445033661524287,
 0.9018103891956984,0.6915920686117438,0.7771221341749327,0.09732113681977861,
 0.8686144098441658,0.15943700859917687,0.8914743473033547,0.6952075801728991,
 0.0862576407913942,0.291052616632655,0.7224632075452518,0.24848744839148795,
 0.8167393704241517,0.8466025212628465,0.7870281921111018,0.716611385440408,
 0.6348853708346184,0.34026436360120416,0.12616082693692277,0.840647974290224,
 0.6478869465399337,0.8531029581467734,0.8381795365795477,0.26744093126465474,
 0.9394745631633143,0.45832444254145366,0.2217970768241373,0.5316437188381433,
 0.6842737859213084,0.7437672383624448,0.8848420888164479,0.750383281613833,
 0.2179140157211994,0.2636760869486342,0.28109682221154864,0.6660739938556542,
 0.5561065039849323,0.5324696949482752,0.9340256255425667,0.8437400391374671,
 0.1853771897428046,0.028307596409998248,0.07018218557475453,
 0.8392355913485582,0.27851397132350364,0.9017431470008694,0.7571899814371217,
 0.48486175999863634,0.8494700330824186,0.15515343148186833,
 0.12042707092801042,0.7145896314773484,0.18907747099668093,
 0.05326525218224365,0.6996245198584594,0.6452311408895559,
 0.49173832754831415,0.16343073050332313,0.922833601566283,0.4951312768872401,
 0.4352820949406335,0.12326115833365592,0.45044314432394605,
 0.20037316584101061,0.975898550223585,0.9292032584064162,0.33256273108296086,
 0.22429237891689335,0.35417942134111546,0.43840774933296356,
 0.23620913021579226,0.4318816155458908,0.8585652983402172,0.6292549748845061,
 0.07341484722684377,0.27841934195404905,0.6227821080810428,
 0.40097949631380403,0.1132415978460597,0.8389190259250641,0.6266032349427135,
 0.14587010379537213,0.11849409094771546,0.16680062077650315,
 0.843171844456849,0.2122257077960329,0.6563342164812372,0.14182465595744398,
 0.982928254073392,0.005602622373402921,0.1988546212942245,
 0.17580203445185016,0.06973876499526388,0.07001676917349442,
 0.19943603474114258,0.5356169295533656,0.21984886162457817,
 0.07336489983598149,0.11660076909297445,0.33358975217746123,
 0.18723035743821292,0.1713013350449306,0.7083416461749563,
 0.42078777867663797,0.4852070574673306,0.6694268052928471,
 0.28745885524855685,0.289674931566131,0.7643350420584731,0.5746537176289215,
 0.3759377512851546,0.3243712655849569,0.8373952518633264,0.9551488723755254,
 0.7214154218251102,0.2073580674426645,0.5852939323148327,0.3463157275190819,
 0.37332088339867386,0.029927473412604266,0.32397875232224804,
 0.5838672569457043,0.8925016476192318,0.06685623389471473,0.97421971854408,
 0.3968345616610517,0.6929835064434493,0.9030691281877099,0.32274883245982666,
 0.7196990090606314,0.19763142681281354,0.22708233111055287,
 0.31220334063643496,0.43271997500774306,0.7879799833702104,
 0.40378290499797975,0.7472257887164583,0.15550937981267587,
 0.6672252465979134,0.9648167259860378,0.06179323801023151,0.8002517701930054,
 0.020222320104735236,0.411633565679385,0.403790304892489,
 0.022313748044322623,0.49610665089116013,0.6391690456746242,
 0.8998483241272858,0.8907837056396106,0.7963040443686336,0.3207245963630734,
 0.7044854617112036,0.8384007212219817,0.8412987639658229,0.46540304154813616,
 0.46375070028146403,0.9946894685165378,0.7332747619704288,
 0.11317955799118173,0.5669256521555869,0.6123226203694399,0.6710298620234966,
 0.8697663506667594,0.3322261514895868,0.7994372466394462,0.90850223679648,
 0.776425698156356,0.16458326899329145,0.4063591155222448,
 0.018358859758776824,0.6698987536566882,0.10915183570627263,
 0.09298307377808412,0.024323475552424423,0.11402676710382398,
 0.7739739995973954,0.715119972111946,0.8867953204160892,0.9870104532202677,
 0.5696729734177401,0.28907087837275636,0.8398019961335399,0.6170598914425706,
 0.02979722540192231,0.8408187273253617,0.6628866276828896,
 0.27039870582955305,0.6297569882575008,0.8711015001080336,0.9434243852924502,
 0.2973544896799092,0.34100525022680406,0.8753013416458651,
 0.05113663350411768,0.016613985793633157,0.3356779131754668,
 0.35645745280345276,0.4126452584926549,0.1425148834984301,0.2293457558683153,
 0.17812323235128824,0.14565250175568523,0.3648926641965997,
 0.9564944827540041,0.5886615693510142,0.776019216578736,0.5395666772907328,
 0.41587756592937075,0.9224500906279808,0.3899950320018438,0.0848100051393782,
 0.23447910743960254,0.22058895230505726,0.22694356825159423,
 0.20212919434417131,0.9969789069147199,0.44327166728490597,
 0.4760656249912264,0.10846772425261597,0.6768442579674492,0.7816069055281694,
 0.4719907632411253,0.9828166899713782,0.37973015583903447,
 0.030427586438475318,0.5088614376110093,0.6964654134483261,
 0.006842482118900373,0.32207102027596535,0.4494765109198343,
 0.06118072533452035,0.12136310318684351,0.17661605808381187,
 0.2289125636525865,0.28256796312540233,0.9289547358198204,0.5893212075003419,
 0.7582624740389045,0.551456351984661,0.9157669035615541,0.2085642464606049,
 0.021284880110046367,0.30781250885928224,0.5692549671470366,
 0.6890765618700739,0.7369955777016982,0.3702137787013017,0.658948751040779,
 0.9263630086055412,0.9911603702077488,0.33951822910266294,0.6645173895313812,
 0.5084725797976979,0.2556367929619636,0.6260310074088159,0.16224931915415097,
 0.7927825452240644,0.9396193588564515,0.8763714639541695,0.7350980704025569,
 0.4411621603385173,0.5648929805951806,0.6267903223199303,0.8332580869011116,
 0.351972914161904,0.7769958014018621,0.5625479705262093,0.687017779536867,
 0.3658453103849434,0.1981368999680465,0.6654421153534407,0.7178700225960348,
 0.6769102305307813,0.9111732944787974,0.21733563378105192,
 0.06599919414790956,0.34583774321170035,0.762158620705979,0.2594080706950851,
 0.06743147293881446,0.004568641697736742,0.9882648618582277,
 0.9320373001868855,0.32735464442428264,0.04819942911200026,
 0.4879635828057868,0.16392111721586355,0.4446175066390773,
 0.03803360796614763,0.9386125389211899,0.322779636964748,0.27699377623071864,
 0.7630338309485224,0.7344503411000007,0.9261339755900031,0.6264735115179486,
 0.14772275364580356,0.42304410462086695,0.2799960019179655,
 0.16110797300281343,0.1768977317878614,0.52219014451135,0.9815987389287555,
 0.81740993667745,0.6424375430888516,0.8506823699945234,0.2213905208046032,
 0.8332144754156862,0.9150111257070552,0.3811288487873393,0.4322281480301893,
 0.6399288569912625,0.8685229476893278,0.9728179864554312,0.41578306869597226,
 0.16096773408188048,0.21869110239375034,0.7408407765637584,
 0.3248590326353996,0.10147771286842588,0.8057848661749601,
 0.26866988863407204,0.6347472105721259,0.0033841133288577963,
 0.9291402096840468,0.1952249044430684,0.1982674483907767,0.4193765624978505,
 0.580921426804157,0.9638853224994416,0.8898166227594428,0.37723416617074945,
 0.005596694715415351,0.8499744583384204,0.7654900702077605,
 0.8402864634869498,0.9328590474778056,0.7878857132833726,0.8775859872242359,
 0.4339728725776112,0.9586680385888348,0.14162463535312564,0.2661498926090212,
 0.7832155686947595,0.42416792107315504,0.08321444662910338,
 0.6970063004780473,0.3663605148654153,0.045263116582009766,
 0.24075904524514402,0.11118757108767285,0.6383852032059387,
 0.6339895244637801,0.6908345691795679,0.6546005661430562,0.37290745301435335,
 0.38333566849467515,0.7468811488838726,0.7709049175408507,0.2070255946438191,
 0.88078025277207,0.06859825209379722,0.17404919970924027,0.19401297433180753,
 0.21542713688662207,0.4858330279291663,0.6225761020433845,
 0.48480448410025556,0.5858729387531775,0.07107122671734356,
 0.6037645122009689,0.1558421925344151,0.3952405623086235,0.4211999790227615,
 0.21717391773850847,0.14892431470203749,0.5307808958120562,
 0.3501282605734345,0.936568901071768,0.31466894624270436,0.23929157616313346,
 0.258065040482103,0.49259009225191586,0.5759296190276284,0.14632719368917158,
 0.19361091887655113,0.8374502908661985,0.8870349092193738,0.9858562336268779,
 0.08690956381831894,0.029246497074981948,0.1515990267674534,
 0.8831491958245914,0.2286065186681734,0.34033639675451577,0.4100172622425613,
 0.9372688550099358,0.7682857247477245,0.23182175490564982,
 0.09854645438340492,0.508592258629974,0.31866310529735986,0.3187059804831065,
 0.9445534787721227,0.7987517291431775,0.4200063473666972,0.9298622422031007,
 0.791901051729002,0.259868445160217,0.8199383899570343,0.24569372114575427,
 0.9397742767184821,0.590170454068172,0.1404388315185776,0.08816190996524886,
 0.9172331785465566,0.930473680098962,0.8866369481035575,0.09553049603532726,
 0.3341779305504561,0.6798347698846764,0.7505004008203469,0.10377525273931065,
 0.9210151339844165,0.6872187658389,0.48151305601322436,0.18854583614349252,
 0.420551319012591,0.8689739670778728,0.4001735269954029,0.8066499537658608,
 0.21304973246269598,0.9176829878866924,0.04026741803458411,
 0.8089017138411333,0.9988147911618122,0.4569560127824641,0.1933789723224536,
 0.48956326502806446,0.6922312007578459,0.28294745336172245,
 0.06644620976775628,0.7465750612834502,0.76075502662229,0.7427397038351388,
 0.26709986843667566,0.3543611381255952,0.794327402613111,0.4708148920933588,
 0.06165988522929178,0.02442957141665625,0.34433560703723065,
 0.49296234695742946,0.5760603410667727,0.6428812572283986,0.8462616953008313,
 0.24326302721045634,0.5932095581510487,0.5241870639095721,0.7693581239992893,
 0.19265411144833378,0.9784537941290944,0.8909452552002711,0.8206376987937388,
 0.1824457187681615,0.4438741301932715,0.11614784990127336,
 0.07583335684773052,0.215665540189365,0.7815582811121656,0.259005391873397,
 0.13785950203831288,0.7428072118196308,0.6076853848882653,0.8799091783184687,
 0.5084074171138048,0.4185690201422795,0.4624145975870094,0.7208894447484693,
 0.7696991404656659,0.004172883608216549,0.2730215092082854,
 0.6536004035500563,0.7329135455799318,0.32250399513242023,
 0.05031715197261544,0.14620205993855584,0.5976803601380587,
 0.16302510404213155,0.7467472276899276,0.7302356915116298,0.1786045205448169,
 0.3499980403519348,0.14273645842093063,0.16893620714535584,
 0.8596537034281628,0.3030776774451661,0.3236158410379576,0.3769981587112331,
 0.5365152182530348,0.3619677485449816,0.7528571933201119,0.10092260221670013,
 0.697284325112397,0.8580045223192256,0.45262607379919406,0.8032878169362158,
 0.947955677742594,0.5698083241213874,0.34492306256175176,0.47693362380089666,
 0.026646528884642584,0.9345527790880838,0.07184848420815848,
 0.7956538216769627,0.6332148763245886,0.8530698912929383,0.2365712553579422,
 0.05806938041715082,0.5466259274770774,0.9257734474221554,0.5464537036155659,
 0.0625336011292551,0.2950498568902785,0.3368163691222589,0.6408049687788123,
 0.5002796323821499,0.20849797053648245,0.4650892971459506,0.9115790336744929,
 0.1287782609817879,0.3021800508546765,0.7259938089353736,0.37610894310916065,
 0.5713296977757116,0.6409446458396757,0.3412898548480301,0.20917571139023738,
 0.43099547136588157,0.013072069325521518,0.940569121399771,
 0.08254198786371647,0.09949631515161406,0.14285687296796,0.05070255080760866,
 0.5450249388345207,0.6692677101174552,0.3714074415660388,0.8668064764603242,
 0.3746248069524587,0.2631916638722934,0.39382413426704066,
 0.10754421483585896,0.4974981906875078,0.33176577444011557,
 0.8099316822902163,0.20080671561844676,0.8744168304651677,0.9160814569762575,
 0.7114473015297463,0.7267051616560748,0.8479782601634064,0.7348793961827567,
 0.5694429930581453,0.4766985346879995,0.5010549070846669,0.7871362009990251,
 0.006968585482916234,0.22729308078796973,0.6561599563929204,
 0.7744145583962344,0.5664087752183624,0.26826011678178174,
 0.46419946445714544,0.10077778303949159,0.47285406465760627,
 0.6926093424981127,0.2119054972549288,0.9652118434462892,
 0.007468988402174914,0.1701350002349734,0.40622387233119983,
 0.2410690834220226,0.17303334358263323,0.9573470322631173,
 0.11286935247701835,0.22042317398905542,0.8076809638020105,
 0.9394022317386388,0.45554101160632143,0.2742590394637411,0.1776211159528155,
 0.1177792838102012,0.9341996683389996,0.3774900150001619,0.9195423042838131,
 0.6167684238640725,0.4042693937103613,0.7256337736552678,0.7115388042638663,
 0.8640250117402616,0.9717025471153538,0.9058588364017266,0.49849664609415856,
 0.8292815092789089,0.024345898236527774,0.9188306061546757,
 0.2662711216575776,0.2615911927057287,0.5516736032400987,0.38947408539522865,
 0.2905587931325452,0.0017513952837323599,0.7577367229402385,
 0.20925877237255697,0.10563332211222265,0.3942667221656757,
 0.2941805906698367,0.22370450792174745,0.7245886370843088,0.1814519446834406,
 0.6772380159633002,0.9598857595305489,0.07564578702316149,
 0.39669264802370907,0.231721966905446,0.07985867525978108,
 0.09503238772058409,0.9020427560988605,0.09167092344970162,
 0.2246500037881104,0.09971906421265442,0.3124442538463511,
 0.33953026567477473,0.4751760548811925,0.3189513145585213,
 0.47184125587394143,0.7326988199585807,0.7109150302526709,0.3786179295689368,
 0.3248351231489348,0.9022710708108004,0.2806981957840652,0.09645143162046721,
 0.9602088300153017,0.6640030500051752,0.3290848251326922,0.37903409916546527,
 0.1941727809797632,0.048507911698326134,0.04261513510629078,
 0.16847750346026735,0.49113745106715045,0.7938796915071599,
 0.42944604222130733,0.9490258377038328,0.1973254435486177,0.6933234191185159,
 0.9369715104861805,0.988286228215732,0.7819380915207744,0.9018498665436989,
 0.5607793680323909,0.501701283753163,0.7014574379455805,0.002814411796430938,
 0.49712814720996434,0.3702976321916551,0.359386077720136,0.9833503097398287,
 0.18991489972598186,0.5554643025491026,0.6280598395768147,0.5487675181473163,
 0.1976965301405763,0.038826879789226054,0.8699557174748916,
 0.4810851850851636,0.6371581017608009,0.1784916051807488,0.23175988967721795,
 0.3082374930428269,0.34781249454452845,0.44541863937722337,
 0.9691434338801487,0.43519272190970326,0.9483950100445645,0.520517743048097,
 0.23521154005910172,0.013609155590498578,0.11026034415809505,
 0.6833321460534099,0.9737193717491325,0.1979487951747081,0.9960686545305291,
 0.3789556345575812,0.15176598122796103,0.9958649786661691,0.3645828468228999,
 0.8516481017364195,0.8851290784263869,0.7354048867945275,0.5440455025850649,
 0.6102887333774643,0.06672242958085839,0.7075915086404861,0.8915676541365163,
 0.44186325208248867,0.49075539136416213,0.3003712621790523,
 0.9725226632900219,0.9229328579144701,0.1426214640983292,0.45822412625948605,
 0.48458709826739044,0.7387405477055651,0.9011107895753934,0.1604539076075333,
 0.9960032283584003,0.34689663634287216,0.43143227960701525,
 0.8810649060341198,0.6378659985665435,0.03266431714120821,
 0.22375695295033393,0.9953735222092696,0.6237334968766801,0.235297191388676,
 0.24217201289426016,0.4197783693412436,0.39059716338074646,
 0.044233790932830974,0.23403485918350608,0.8903027072490244,
 0.5819497184314729,0.5909929650434098,0.7570444109248451,0.40538303182292734,
 0.5461598265878936,0.12188709072891946,0.3337198917076385,0.5816473791494022,
 0.5401246258866322,0.9978542761597251,0.1047263930214275,0.31296508618419006,
 0.3716498655217648,0.5612733251705843,0.153203702153219,0.9362551170193611,
 0.23111713666729383,0.5803121547478748,0.8509850449771076,0.5845990032706938,
 0.6130085284486657,0.8663441726514636,0.008293071446815459,
 0.31838536743690027,0.42958416719931636,0.060217642155159234,
 0.033934120563283265,0.5665231411681283,0.5876382705562986,
 0.6836296184770251,0.4274529977942598,0.11077571582425827,0.6954630174300152,
 0.7071398462238493,0.3722851032964929,0.24490580498192394,0.2691927833087765,
 0.2085383814587386,0.3364796293251361,0.13714129129544272,0.5331008930758956,
 0.45973022316887646,0.6470726229709783,0.6987237948449777,
 0.45527462583840705,0.42340693468409496,0.6983154962142883,0.359537473458195,
 0.483879005726516,0.6340698849257488,0.9716972336079573,0.6840252825654553,
 0.9994169603435865,0.2085965059095627,0.8853444646403161,0.08061900556463253,
 0.8076632914741563,0.7174466874297688,0.8803280149762319,0.5037966707263399,
 0.7111479634140844,0.05727470542721047,0.30184376727396955,
 0.7857817952293566,0.1338257658639258,0.37422058639154243,0.9664712629152834,
 0.28031691632981437,0.2844969804663975,0.04678307103572399,
 0.4843904757945042,0.07238855477778305,0.9751780347802622,
 0.10706153572725718,0.7096629704975761,0.6908582191429469,0.479160063872328,
 0.8519574382824582,0.4808019332568978,0.6505487201280907,
 0.058836979773811526,0.9901103515959235,0.5995405444642032,
 0.21779792017362642,0.9038623045656712,0.5552103634792348,
 0.10987296374992628,0.7406223416343518,0.38735584063743644,
 0.37950927073331964,0.7145874969797903,0.9003698209731326,0.7621596355235735,
 0.2953029787255158,0.2557741065970268,0.8115625047553624,0.2687763312143351,
 0.663402953484477,0.440069583967186,0.22857632140344586,0.4962374010689439,
 0.8962446706767436,0.3243933904617585,0.5297913913071661,0.9894540324978078,
 0.5056142803074065,0.013634199309148665,0.8821061587281308,
 0.5747146761764006,0.38662361710918325,0.011885910982516346,
 0.2987347450016954,0.886207479074379,0.45528990204469466,0.48994235714137646,
 0.1762855684574659,0.930469536231011,0.38201107975908344,0.6953437670282888,
 0.24646479112619601,0.8187403313702145,0.38929885561242106,
 0.22759056162370184,0.43050086756654826,0.992652002264484,0.4771178828531233,
 0.695513837508838,0.670058549722532,0.8933343922134931,0.9664320554524836,
 0.6287204453142261,0.006269459720931847,0.2269766974506593,
 0.31189791887495066,0.9756284738081441,0.5840469809190377,0.6617316054085869,
 0.7335846626571249,0.7161421418020514,0.277349031182447,0.012554909033527628,
 0.12242721659709044,0.06747079847981574,0.9673408980631529,
 0.18493246426846743,0.6361659469499809,0.17638707799465603,
 0.11388813446640689,0.4658789006931586,0.6960711636544609,
 0.08018686597452129,0.7933684948327285,0.5815430640132451,
 0.17800784309752504,0.278267310588738,0.6083954912890628,0.12898468961654275,
 0.7276984567823847,0.40471571450173305,0.11736268985797904,
 0.9197892150553885,0.10088932790162819,0.4735548087061685,
 0.03736312495403593,0.21946395102234972,0.4109870482809401,
 0.3737741513287459,0.8013847686897505,0.44863714693804935,0.7021097095036174,
 0.9783860395187952,0.09844424847857436,0.6759546474232914,
 0.39211291058272146,0.5476723183094857,0.052246093595081256,
 0.6617784249606228,0.9818242543136964,0.47099952539465473,
 0.37294224479946436,0.1444549091958911,0.48882377588403236,
 0.6190051132958817,0.755020344365877,0.2401497452940532,0.8623021686425459,
 0.41883383147111986,0.4859445049926786,0.5216097553759897,0.6689788223370208,
 0.8161044483454255,0.6548619524245997,0.28759548394559176,0.8409723141509521,
 0.7382981055094384,0.6924660087356911,0.8550224113463833,0.17551167515412514,
 0.3678731675377118,0.7259495291596685,0.563535396581259,0.8740396483821542,
 0.688511139986567,0.4378171244993959,0.8707614740268474,0.18945788265632846,
 0.386159461982314,0.07007846421480968,0.6277251022299793,0.16518099208069914,
 0.7122347771598407,0.031278778533865514,0.3880634141271371,
 0.3074339224113869,0.07991827518075467,0.08874995272658648,
 0.1683183314289075,0.5846886015626533,0.16883784187758466,
 0.13567936875529474,0.49470835683443326,0.348405841889476,0.5223190114833648,
 0.5585809521041749,0.2991292457395438,0.8455807206697628,0.6642497353535585,
 0.32966047129999887,0.6622724637802535,0.9348244948259616,0.6018363257198969,
 0.4292308066939161,0.9684978551701432,0.6296677790276078,
 0.011594876818104116,0.9049835993881393,0.9248705528790212,
 0.6846008301725924,0.734640246340807,0.24465040595239484,0.09565907595943068,
 0.9881195081279726,0.9917899555831213,0.207662723797979,0.11951295042002619,
 0.7671428884596592,0.8098051863448897,0.42538017183390675,
 0.06411466894167428,0.7905273647213674,0.9498558800504845,0.975718055673981,
 0.5772543037604146,0.9675518636020564,0.8144817922409773,
 0.013065058505409954,0.9257037990399859,0.5142546338351048,
 0.5121025533233108,0.10304717429716848,0.20853161312624324,
 0.8713565763817163,0.5025380685811273,0.4073944454045585,0.1988564787096766,
 0.7598071353546613,0.20529349052021661,0.3806896159211388,0.58664955267314,
 0.09749001899235965,0.17685255452994453,0.6734970683025266,
 0.0545824712851134,0.41984884262769784,0.9396647140472884,0.2436625921265595,
 0.8415407554075587,0.3483669314454415,0.015469245769214712,
 0.21083782976423515,0.6643690402373444,0.3568889909076951,0.8180002588629659,
 0.22253761527142357,0.42489039996120115,0.5820696198135907,
 0.9464048036965548,0.6770966458717549,0.8118112258891707,0.2968902274332824,
 0.5779109312127783,0.07494887174882359,0.15573785915311478,
 0.02845627667069628,0.6225391211010591,0.8133654564364109,0.7976272569835506,
 0.26073119223219143,0.605230854671062,0.615629925021517,
 0.0029821012258940405,0.8433375975814017,0.15136899978775153,
 0.010936091576686291,0.5729146137995813,0.6429217962667741,
 0.16960977079240236,0.9348042846938233,0.9431500029473013,0.9951576048320638,
 0.5420987131226761,0.21736975397497227,0.7566707140019507,0.7816053134675358,
 0.08382589569291354,0.6915620826634737,0.7952330461785639,
 0.004697431563545185,0.5966040472870289,0.1873822246406741,
 0.6228350252208934,0.8570280115950518,0.5265223693019135,0.8874005061546829,
 0.34159010485529495,0.251723525438699,0.2937060002780031,0.2770588928614738,
 0.0898108342097067,0.15485523691407577,0.6982633188733727,0.9999411308436481,
 0.8359448177588229,0.45110935367620264,0.2869221070302953,0.9449892013033321,
 0.45437303514985417,0.3924915248473231,0.5287737954071154,0.7869733644392276,
 0.3275652750625099,0.7826589788772191,0.05582617659470279,0.7161006829868256,
 0.6482080219883121,0.11052185281600391,0.28736976999099517,
 0.2847029709111678,0.7850029203710193,0.3137250812510286,0.3635635962357464,
 0.7801533083817741,0.11527082688875911,0.04533223617443649,
 0.4167675881109635,0.6531884016378572,0.6388997842714472,0.7284119181213111,
 0.594657425817229,0.4197405144714983,0.6978253068541562,0.3683793585684578,
 0.42412066423885364,0.18637090061193562,0.056593691643406885,
 0.4997470480427775,0.9916105918473048,0.32476786706491745,
 0.06596012509453564,0.2423219157830635,0.7035239957596341,
 0.06979217593781362,0.9797140609597589,0.053885539299976415,
 0.8017241892277268,0.4863738468105485,0.3459194518688401,0.2503594225871649,
 0.8847529626344081,0.3276297404874057,0.6144358621217584,0.4944926346097007,
 0.09840128943655702,0.027967215977524384,0.46797423431607166,
 0.08459873240172089,0.35643624266594565,0.7345956367743041,
 0.6753239638369368,0.19673834506885068,0.41775903444483753,
 0.6788900134954022,0.5171666460045792,0.1707953146901603,0.9261684295978344,
 0.010535518835988023,0.6857629994753612,0.11188810230565804,
 0.646195973530228,0.5728321556034744,0.8173031312236195,0.5887765116029605,
 0.09364192822241713,0.8944401014673695,0.11767785549752674,
 0.6728354024968773,0.890089120329713,0.34275449095561983,0.16256726892917395,
 0.4585061506129432,0.9805011750207093,0.011267561516388636,
 0.11315514884007671,0.7739455134316399,0.2384003865868758,0.9197258475475235,
 0.5418868427651455,0.1866021891859505,0.0605908758865511,0.23889875071054678,
 0.9734109249029729,0.6631033266320552,0.6889929117059659,0.5690921668666845,
 0.9115116568782581,0.3607967816154115,0.9752590963732076,0.7224307294444465,
 0.9610718141324764,0.8792529167112456,0.2272068024251066,0.05880070222867684,
 0.7939391613052317,0.9107016978103246,0.7382919890779163,
 0.024048518646756833,0.14367355264448745,0.4430997969763799,
 0.8185271987422521,0.897466055876438,0.1427106744237545,0.12371347785302322,
 0.515399687793233,0.061596310170609736,0.7188114757306115,
 0.04178859371474053,0.9027739210915473,0.8449298750846523,0.7171188113686354,
 0.5863984500189098,0.18924232248466488,0.31987882398191814,
 0.8589160472324452,0.4659716216113592,0.7325472355199195,0.5215887614281847,
 0.8916066001048303,0.9618377689305277,0.3466831479135912,0.9718651232439525,
 0.5111843112596195,0.4265053150963032,0.5676013552259448,0.8213496013734032,
 0.08651723135034894,0.045787040740114526,0.014572596997755594,
 0.2752492597616474,0.42563195337768933,0.32238237196102604,
 0.28021197012628984,0.05710186929965544,0.7084123925556,0.25232315608205513,
 0.4826618285565766,0.26847937941347066,0.9837737357414809,0.8043667456098971,
 0.9158137453926918,0.5184168279965585,0.7521064812546769,0.28245434282366333,
 0.4478611426196507,0.7411379999111017,0.7041645446030946,0.551055145616516,
 0.9069037475244428,0.27085790372769436,0.5716365119180885,
 0.28211997787065424,0.7495774941564413,0.13814915971802988,
 0.7316623399798107,0.8610183424116278,0.1429081647296928,0.5376581997775562,
 0.3636544625871563,0.21365104418309255,0.8864542834307944,0.253942706941934,
 0.9028295143310345,0.07356533491895156,0.6731478339703021,0.6855868965502906,
 0.46090469371815124,0.18211842197514905,0.9374428807852107,
 0.9889587751726034,0.8476785752605267,0.26348225454714247,0.7626312508520953,
 0.6678613953541717,0.4149917654450763,0.8232634116598035,0.6213567563080341,
 0.5989058189167177,0.9817638766572605,0.6454823610302934,0.33372636505455255,
 0.7939831881492043,0.03406284381609581,0.11624319014088014,
 0.8317771480297851,0.4257292366845036,0.4217076250884413,0.9591033285802518,
 0.21111479584170134,0.8469632854946532,0.09331103354992376,0.979209948483374,
 0.876807979742306,0.718941536070106,0.041689843571364626,0.5006735645195841,
 0.8608104770759271,0.9489835972041758,0.15027995438559294,0.4594013761298652,
 0.45424791488068306,0.8904618337953192,0.7047552576257258,0.3644672761460005,
 0.6281122930470445,0.4180840104825144,0.20481944767977278,0.8895155545116561,
 0.6295956486211559,0.6377932978591784,0.8743860555475923,0.7231449045403977,
 0.8431949146765051,0.15075932133643477,0.5908836429834845,0.6444195853178338,
 0.5030486654648774,0.11896137901467152,0.37251314374049005,
 0.2796005600726713,0.9543931942927774,0.5397393038023502,0.6131105572556742,
 0.06316157695062063,0.050503405828628445,0.5447743385257541,
 0.8847634124031502,0.33183799505564593,0.48807563466250914,
 0.0239589827658363,0.7769864905901189,0.38087812558792555,0.2215820149718134,
 0.4050295557929662,0.9204722620028635,0.048294452673012134,
 0.9215674736058653,0.39398906707023695,0.10900584494160104,
 0.9659042377723978,0.8831314370566411,0.4229527760548888,0.9592857614646098,
 0.5143425408481018,0.20853788778258675,0.4713239153694264,
 0.23845697868101623,0.20110732219226168,0.8705231437261945,
 0.1807668979691669,0.055662514378872485,0.22259340378017312,
 0.7387238798406718,0.5702856179717103,0.8806232584749601,0.9009663547327728,
 0.7858941537356179,0.462878194122455,0.8071240995461408,0.6250015705784012,
 0.25615808905295756,0.864282358189371,0.05281402490378151,0.8919301404129607,
 0.030859797233345865,0.24078760749142347,0.25326537740752875,
 0.9302240696839732,0.12262627183438024,0.7064909242951722,0.6289026786063828,
 0.46444795796405103,0.5552709596809463,0.4853877843462028,0.8879390020197485,
 0.4055412248199023,0.47951651202928425,0.09967911647118632,
 0.3420462145556864,0.7971487806545822,0.4127571199143101,0.7266711343530403,
 0.05548930292624521,0.5232723400054985,0.15818679472798924,
 0.4089846164435009,0.022056679986888317,0.5059292420175895,
 0.5057691669223422,0.7404017761470585,0.8960517686068943,0.4692031607923791,
 0.29185232403544203,0.15054980790212058,0.28621947348931465,
 0.2104077932649352,0.9680295728010786,0.5051424050944068,0.5271822084484481,
 0.33654940180909165,0.555934652161052,0.8264569542729017,0.6795970980459739,
 0.34527967667816206,0.750072015847529,0.7793463637975113,0.29306493499603914,
 0.8825844835597786,0.6279789042910838,0.5007945979892039,0.5007207561973511,
 0.6009483912016643,0.570325181588116,0.5978764105845497,0.8328149970236253,
 0.7757338178361264,0.08356712281432266,0.5133339402450839,0.4248849158293154,
 0.9173294181619832,0.10313146260095185,0.13927569642071802,
 0.3273976266038078,0.7507642270315981,0.43927982126703335,
 0.26959525997071876,0.793417326340877,0.35269934211485976,0.5480503144141804,
 0.8491040386821123,0.7800742332060108,0.9534102522134832,0.7314808187194055,
 0.2702166251173286,0.6655177373162765,0.2554837768451612,0.5330521364696817,
 0.0450800768991767,0.20381872573271465,0.04539707720359265,
 0.6963370729515901,0.4171762172195881,0.45473700297732955,0.554340111705485,
 0.5426694125954571,0.8084344503281533,0.20094772931879779,
 0.07354724844446936,0.17116153720719063,0.7201365693791009,
 0.3937817041814422,0.6540158321500776,0.23142546357591665,
 0.01865185237034095,0.7104837688321775,0.8766708523851243,0.8631099838925039,
 0.39769010173744546,0.8601600867218947,0.8467661336313461,0.7847937979334104,
 0.6358475308903871,0.11917219361510256,0.6508012505713204,0.694433879027595,
 0.22966476453720142,0.3631699923485582,0.4573254720535147,0.9279258386641811,
 0.8071074277235212,0.3738802811133062,0.39273241693516625,
 0.057981944999263124,0.11438020419644379,0.887921330728755,
 0.20548823100925429,0.5644674299596032,0.7321692087386313,0.704501983368973,
 0.13507680304685477,0.9958291798329495,0.0798460375498562,0.6697927297669544,
 0.6068565241099523,0.4875287995238009,0.7658708995264611,0.6464563154686223,
 0.5752170398506625,0.7984694816595197,0.7305758982392465,0.8148224462827821,
 0.3766182187322754,0.21915523414260063,0.0987103852780219,0.9876229349102916,
 0.8299938876300822,0.5221173957736491,0.6347922508461683,0.2946039305423569,
 0.5790799037409244,0.1765963861994091,0.2532506561949831,0.8599994175720274,
 0.8826643449740088,0.278243222963374,0.7312550934060911,0.16213184240915013,
 0.9790613420901374,0.5020167190001174,0.15468367412835216,
 0.18046375999151176,0.30071975500710546,0.6633600112395663,
 0.48413901027290573,0.433661800892861,0.41842482241939294,0.743434773569847,
 0.47625253054492644,0.5130569353426379,0.7171182293553783,0.6689556609999343,
 0.8375993760192935,0.22286037980719975,0.277098255397512,0.8389578418338353,
 0.9411710759336771,0.9814588120326021,0.4834095268205789,0.9612740373915191,
 0.04301927833627095,0.14705773382641463,0.08646266548524106,
 0.2277902819705928,0.017061148827187234,0.8932324986319925,
 0.9586651079751689,0.03285417462730977,0.174978432300972,0.6303790441998737,
 0.6703006050369814,0.5773588341344316,0.788810170707479,0.8398565076221541,
 0.3741133838608337,0.4022078671117535,0.9691468466290223,0.7590772821972074,
 0.6337783630066527,0.40428426552278307,0.6414829836005117,0.8058779590748931,
 0.2529979476284343,0.12922822605328999,0.485519900319774,0.9307108614773637,
 0.12494258600197938,0.22434474427154438,0.18992909209671371,
 0.33686428706732285,0.29593334301065477,0.05858864027090627,
 0.834943855722891,0.42155728534387404,0.7629853949102462,0.4836579785282883,
 0.7669855374886817,0.2202592551624748,0.3198583524320051,0.9946822000496346,
 0.9637033447566826,0.29224184269383313,0.15385999601204192,
 0.6961439375132937,0.5953494861602003,0.811945535719803,0.7258769474541863,
 0.30452029218681387,0.9580569137362465,0.16498128018945213,
 0.20138143586886292,0.7943125873453275,0.2814330401194467,0.3538995300652408,
 0.8844576849631881,0.8433393326508097,0.8105894711296375,0.38242186442148085,
 0.6828858921604948,0.3533160498653184,0.2336045247365165,0.7097553423412192,
 0.7903917529295886,0.9102575063232672,0.40312437162449755,
 0.05752125692524146,0.8385548746251901,0.03047860274387537,
 0.1491122535597582,0.5102751974254564,0.49999236229703947,0.1819239984346095,
 0.436022303821911,0.5320927613098767,0.8441313192714056,0.877309642794256,
 0.9716488671649618,0.5018051684952856,0.2804703598257514,0.7486142738160501,
 0.9143959332498706,0.34009098173680374,0.8316363127206896,
 0.06345463376096239,0.6418985711340002,0.7841881621963239,
 0.01273575528240034,0.6505088551994006,0.5636120338457602,0.6106639380960432,
 0.6474679511928267,0.19931854304735408,0.01647998431218356,
 0.8039513480233847,0.802229115185774,0.9607618597784898,0.03089776987434001,
 0.30254986099854897,0.5119555044643629,0.04798755074097616,
 0.6288423823446809,0.2799314185456421,0.8873267602726883,0.3488375013810263,
 0.7940685551739062,0.8069473317078797,0.7114762933283633,0.6167539859284022,
 0.39580207215463803,0.7709600251039795,0.39156864542174963,
 0.5161923961850904,0.3936532414408185,0.3439198759105768,0.21192472417048802,
 0.7312553469312015,0.17184991938524352,0.18629130771708524,
 0.23297764008004374,0.40835209375115344,0.6592060054661277,
 0.7250950215596652,0.3617722540322955,0.02242954870454139,0.9495684875388751,
 0.3751113432049519,0.44796204140167517,0.24932430728991573,
 0.8855546604503081,0.6196380658833558,0.8612423797028591,0.2640139927262761,
 0.6018145099311891,0.5574684789524824,0.48093974927239336,0.7782067850809632,
 0.2829127678101069,0.5612746243112405,0.9126200055175795,0.9262422624260407,
 0.7647770774345652,0.11563392637186465,0.9063969357543147,0.4711988263278798,
 0.22332239289942357,0.16176319593641786,0.1621195293857115,
 0.22962240193735461,0.6262628478825519,0.3803334457141502,0.9620606114903344,
 0.40083386919009556,0.8099059007292264,0.1750005530093165,0.5100786609446029,
 0.14057109834466175,0.15719473471072976,0.1254713680326628,
 0.21260038449278607,0.31706372182421816,0.4264792930197975,
 0.8196699354766137,0.2547260542082367,0.5386459778941185,0.37698238235708215,
 0.990351947260345,0.8597775648614686,0.6220061223995252,0.5511052716167706,
 0.9292939309577994,0.8491765019347395,0.07027324677523439,
 0.41640671687505415,0.7511505533923257,0.5530799799681763,
 0.14234961333658458,0.21442698237418034,0.5745718597204456,
 0.6255830251817412,0.5457226293346747,0.7440431212872671,0.31186479772051146,
 0.7012204603401301,0.5474461876123526,0.5134251342925875,0.7058115152947914,
 0.03244979667484271,0.5103514853929284,0.7231355569941057,
 0.09677643133988234,0.8130597583808223,0.552825237919311,0.8294941840113932,
 0.8507445470311255,0.2313373140849604,0.8708274563288821,0.26808821300874763,
 0.7023470301769154,0.8126136881126619,0.5234172282334071,0.7020185868471345,
 0.19837599671764994,0.46842895467190426,0.5667860094126757,
 0.7546264006838744,0.9774947365050985,0.5272294926823643,0.05561471492367531,
 0.25025346570560913,0.3568401492660225,0.003392939537676498,
 0.6054433462882571,0.8947014828035417,0.9698223535489814,0.946773311341505,
 0.738083250598696,0.6119148964344302,0.4292333723845636,0.4446747144945783,
 0.8213991214109989,0.4474803294922085,0.43447369018557835,
 0.40777298745960766,0.5856090944000512,0.7278321131533008,0.8676921793910772,
 0.3537655839999867,0.8914049865877565,0.4453793538014388,0.6196066961209815,
 0.5339054290625863,0.5152110572165427,0.23268894633942538,0.5879285646754113,
 0.01770586883248848,0.11551490387396235,0.025804824454752406,
 0.6091969760700889,0.12845806250815583,0.7803567130614991,0.8432503609686354,
 0.04560003201425045,0.3311787996470307,0.16836281387894447,
 0.24429959905212517,0.9044083154390813,0.21038158913935823,
 0.20689453863046725,0.1893965245768232,0.9413563481252653,0.7439780284215808,
 0.8288966735123157,0.47173216856188116,0.23260481812702027,
 0.1268304574511957,0.48082074176229583,0.09471775392472837,
 0.20066114218151176,0.6234088238389215,0.9497119130863763,0.3377639929600177,
 0.7876951267508621,0.18412088594718945,0.57702685372281,0.6354043580687745,
 0.586831085746382,0.17410154009103995,0.8245176744858548,0.11594089499089943,
 0.9540731976438903,0.2961729531091788,0.8503524057980749,
 0.005316026514243233,0.6964450903525025,0.2991133724708116,
 0.5619872974595258,0.9101181674916079,0.545498223997116,0.27150756161019873,
 0.8738139655136538,0.22476383544811052,0.8570492663031557,
 0.09217872411408967,0.5509641530367249,0.5905923784618613,0.6538885805760454,
 0.48414663452196416,0.12875297458981005,0.1752620058014891,
 0.6463118005584937,0.0763079220310573,0.4455011435034897,0.3115794244156286,
 0.43715162101320737,0.8165427840179533,0.431300259186724,0.5190466743879999,
 0.10283568716848013,0.7311775082876066,0.9719203382741252,
 0.13103067581621408,0.02685343057752243,0.17791684313014378,
 0.33002083516691993,0.6318093546416962,0.8193900560682893,
 0.48283099163551824,0.8621766430630782,0.49029794194150633,
 0.1281308304224411,0.0763501595635584,0.3562459521680439,0.6608447749416743,
 0.3348464247073456,0.9143706395944347,0.3680014055562463,0.24670747553427075,
 0.3916325593853899,0.5263663419610084,0.02878853243581503,0.9047897415604775,
 0.6738118907508951,0.38687850953111647,0.5664986727237666,
 0.007361122173805801,0.017471049118587434,0.7203370827872125,
 0.6243556793855343,0.26325446545749975,0.9844126419588923,0.8651033060982363,
 0.6655093963282286,0.6453714927162217,0.8164672300652648,0.2689120840787361,
 0.3407627127538241,0.7439202882809777,0.17342844372309818,0.684614907090541,
 0.13885802437083272,0.442230760563303,0.10177465229728533,0.4030819969342124,
 0.3183788774581008,0.029579626185657237,0.8543805702682343,
 0.6279379607225956,0.1713559196312957,0.7873910387130859,0.16284480446158944,
 0.8227329754982287,0.47604374234095903,0.20091400341092291,
 0.4148543555196048,0.9783291862899383,0.8021473253070294,0.05957647491104123,
 0.3376739703556193,0.9233041665049871,0.8895567857717269,0.8538138190027882,
 0.6963372491318292,0.35477810307651736,0.2750956262259423,
 0.32715778397131734,0.6394375903081789,0.8006728271763518,0.5349049731662903,
 0.9276239585444672,0.07468471849054703,0.985423442418983,0.3685077388744761,
 0.26851980990473123,0.9792942346357403,0.051527204915558844,
 0.358397411797454,0.23062870505737587,0.8243789825745459,0.5020917639036769,
 0.4073109069217766,0.825936446224149,0.7504765449549684,0.6783757668484655,
 0.012995384506511831,0.3703972610563754,0.5737976773293838,
 0.3157770764440996,0.9751365602578954,0.21730813348203704,0.6574294198741971,
 0.47262936732659877,0.11750713630550524,0.994154524003217,0.4489909171172095,
 0.33208769635056035,0.882694511886719,0.3679599500293854,0.04926623224016269,
 0.25901285474779523,0.2898133973363174,0.48387253726146007,
 0.6633313984392875,0.0618716883456959,0.42000918134146803,0.851105950978971,
 0.6653051798031869,0.7753473591741309,0.2343583390297237,0.18338979997665739,
 0.6231471060806462,0.48100734131599643,0.18742001364960614,0.936977235147066,
 0.4234075716939637,0.47787786603775784,0.9814898617014,0.7653406771230997,
 0.9676862508520516,0.7311574740993337,0.3457781289633639,0.5421450766146576,
 0.3196647055578754,0.4980745367312944,0.5442385582424422,0.6890614127876235,
 0.3714583104267243,0.30779244991795374,0.17477083653324477,
 0.6334595958752142,0.12994791532764227,0.8564028452110591,0.8847067052441205,
 0.18202315138618674,0.3575371938919121,0.4939845326012673,0.8833390975223907,
 0.9438058619705767,0.9615005793718512,0.32324931031011517,0.1848311860433942,
 0.3787675783121218,0.06702000300149702,0.9903002469648431,0.6266449571398385,
 0.5398428793974712,0.35726904040310925,0.1069055143533858,0.8020479747735341,
 0.8791558657575625,0.4201909265195266,0.9140953869107427,0.2535319939004914,
 0.8687505126211355,0.630521694698459,0.691532101249855,0.7418063327788652,
 0.017816248630576625,0.8654377700507472,0.1852332019952836,
 0.7883785103691003,0.15492004728178244,0.08285944501671105,
 0.9923128679274348,0.9506581130197503,0.1280119989056656,0.7506254740289044,
 0.4540098628175321,0.5242911502629483,0.125382783034643,0.656501716399577,
 0.5772637600593734,0.7151282894187307,0.8565738357766457,0.401782280652201,
 0.41167282715742837,0.35789223322490327,0.1705530068410389,
 0.33876939244786664,0.7700516171602174,0.3477832548163403,0.6322554353529423,
 0.1476536959405712,0.3621198532149602,0.8246216876036438,0.8805694707055005,
 0.07692458362490717,0.2712341063658623,0.028907130584161056,
 0.9485249689593938,0.10601234203409438,0.1814308422791846,
 0.11642226808438305,0.7827724565850243,0.23425712568917678,
 0.9737138117697846,0.3702659612487105,0.29413882914964273,0.6232247454509353,
 0.9643259635462165,0.9145882709233789,0.6740779481646398,0.7570057252288727,
 0.14444364883036165,0.30096832720875166,0.5369761203926469,
 0.7500391487683695,0.647902885262794,0.7745915488045035,0.3205523676166583,
 0.9422974787077862,0.5499441349769906,0.8716478403094773,0.7882123680397367,
 0.8823160501997844,0.13436745215195134,0.6416081398415261,0.741570194224056,
 0.6669749859369271,0.6969828847514377,0.7247064181704319,0.6556744382918372,
 0.281241490313878,0.719603616004735,0.6385567593396271,0.8962506737142979,
 0.5269895307222964,0.13871197187823647,0.699889187732192,0.6916511324787756,
 0.4795480737485972,0.726352927013868,0.6263371717203705,0.17056890828417937,
 0.4382408294322486,0.22353448199453196,0.03493343844802721,
 0.5452378047040968,0.3422197417567946,0.7221464058270067,0.9145089872346239,
 0.18772192543779498,0.22688894371218182,0.6796929796065727,
 0.4670518636464006,0.9120074349471377,0.031903149895812444,
 0.9235163807007722,0.30311997680420855,0.874010873783358,0.1831099308350047,
 0.03403377757025605,0.061548442979797,0.31423870811288923,0.5432349733714529,
 0.6044755748419067,0.6658870874243236,0.6086417501052147,0.28742644914922966,
 0.5892400496406094,0.9270163105032503,0.7066779861598409,0.7403194325955998,
 0.769867309576905,0.2453354795898487,0.017791817219500405,0.9529207597837505,
 0.6287391790971657,0.24330801643536804,0.6912644656811922,0.7099863858981077,
 0.7265697279735113,0.31209402886451443,0.814960540698512,0.23071486833052057,
 0.6807565417599124,0.3900335899049209,0.28608572252375186,0.6537644760348851,
 0.6149958462421128,0.3715570169791581,0.7800097965754497,0.5070622822455992,
 0.70444545494128,0.7476431147511109,0.6411919429301022,0.039756887769702876,
 0.598579180111296,0.6704268905541029,0.4098258633685913,0.7236868991795491,
 0.6122110545591573,0.831802108734091,0.9900214642877428,0.7950121059060129,
 0.8913301617266465,0.895690115811001,0.0701896216254746,0.3299300654078545,
 0.9416372250619417,0.8234797272725364,0.5876516933236366,0.9071701381482232,
 0.4961191180147032,0.9983731087232728,0.3628981959927,0.44180075930328855,
 0.051096234864957624,0.9302491446620849,0.7946472369846288,
 0.014873945768571595,0.8939709591049805,0.9877979294357612,
 0.9368633646873721,0.7745060085259361,0.39469384684559794,0.5427957363620001,
 0.12193773746610026,0.015874306598555488,0.8076877792964223,
 0.9140085132498831,0.8677190834186163,0.53550980698036,0.908993418455267,
 0.5890721106217669,0.42099033314225265,0.45345788300494394,
 0.6258479064562477,0.9442755490967828,0.8325783224653024,0.09445252757293154,
 0.25713351985681243,0.7272199968274066,0.6844233274260202,0.4485732344889102,
 0.8676193728927104,0.11193608071955863,0.7371590007133145,0.4779129020358499,
 0.1371411958425628,0.0024464355562721085,0.6656956120300394,
 0.40864403219619017,0.8619408466278728,0.36225640637938994,
 0.7518194723291529,0.4542271151114745,0.651132698901495,0.7523602288967067,
 0.30152496566210363,0.8197613746783814,0.581758871388754,0.7437349850219968,
 0.09721798982592256,0.9100484080303336,0.9046787261594753,0.1281912663222592,
 0.846401717039465,0.7175987660443865,0.657293534499012,0.4190619966814464,
 0.9770852870839908,0.980987693724155,0.028610863353276228,0.4626679754564825,
 0.17711678938267472,0.32414026156655584,0.6564546882023682,
 0.8208701368889634,0.8438270297177572,0.8455604841519254,0.8751119192284122,
 0.7743749680329514,0.13291286405002267,0.5739253946546149,0.1951375401799892,
 0.46537879158516215,0.3812466751783953,0.5765377108041287,0.6761805281511956,
 0.17344451158522312,0.02873702904708264,0.3641533784548676,
 0.45696733751371577,0.051407229681253,0.07659162118685292,0.2987516238293997,
 0.8309482304472474,0.2847726837200244,0.25968670771245517,0.5582492026820005,
 0.973974593684231,0.7815151348472784,0.3942059863325017,0.45606678594056693,
 0.7925552570314349,0.5312830103684416,0.31888656322741227,0.5525338138784922,
 0.26657816943935675,0.02098018134265689,0.8794165698379437,
 0.7724805850931035,0.983510727424749,0.7822339045698341,0.16846108737942456,
 0.683370754309611,0.16598210206180475,0.08506943809647116,0.8509266252383533,
 0.36529263900977926,0.8138596790304771,0.44206231417530717,
 0.6121460044960088,0.1247444852873052,0.011019357976603028,
 0.8148748978312136,0.6021043569465594,0.3973263202157473,0.3002194259768278,
 0.30589462835537695,0.619891949986706,0.9015768756483445,0.19768825217977604,
 0.019345104159355753,0.7858437457075407,0.2887008195267844,
 0.44941045300719074,0.7520636207405325,0.6403324610418177,0.7796900268599269,
 0.07675563156553267,0.3506285843805975,0.456930366356598,0.6878114045192879,
 0.881505792311219,0.43785588843270284,0.3406966332173441,
 0.050009742607518226,0.20241048693031127,0.6631728187556892,
 0.9895494757339139,0.17280176523985358,0.4170371801410294,
 0.017798622671919784,0.22925239433467748,0.5807822966229814,
 0.8756447025853598,0.0783014641861951,0.5502166696426442,
 0.052465993487986085,0.4674828433917253,0.8106245392582334,
 0.4175920130795904,0.6236339893678065,0.5533871317119589,0.08798586361156269,
 0.7362556617941616,0.5561336516517466,0.24558588809910153,0.6446047896229101,
 0.7849375589050299,0.8126347093532691,0.8844330159993967,0.24123045884050387,
 0.3883609977277268,0.3144241688599545,0.10017205140749863,0.6614197073129762,
 0.7716185369497076,0.47382515807886394,0.2091102354246036,0.7217935659987275,
 0.9463507251106026,0.11893211561552797,0.8902465674537537,
 0.29441414039340374,0.5979096349412794,0.5468277713687202,0.3167388188673228,
 0.7051835771447337,0.04015525685041321,0.8190504111632904,0.8836351318821795,
 0.9371422165280997,0.6525741841538748,0.007180853668959153,
 0.5932907935170018,0.14096558347187305,0.3763767541634967,
 0.05259864379419665,0.5301025790985573,0.7104721438851193,0.5372075313522233,
 0.2611426651765758,0.5957808372781642,0.14797460717169575,
 0.06699701078698761,0.7673881062784689,0.5613056992530633,0.5133944068831207,
 0.7736802414821898,0.21852061455431404,0.2479073225153794,0.7633151343951452,
 0.9625711923388307,0.6620889553298976,0.3675897526033174,0.4222756783092382,
 0.8539414972854944,0.53734943800904,0.7609582971371589,0.7297645079764784,
 0.17709706092193112,0.22283058882696394,0.2162375668989931,
 0.1438720781166225,0.23334837812608433,0.7238065919108172,
 0.12408268925290267,0.9479927321115686,0.12068255013290885,
 0.34459180032295456,0.9999990798946796,0.02118264526307556,
 0.8263573624569376,0.07404812526195048,0.3706627353238978,0.6474940420457055,
 0.499822140074901,0.5087103207162196,0.7042455799801648,0.968110770258035,
 0.5037195755659633,0.83656573213587,0.709622377143993,0.026594903248741586,
 0.2016725845744154,0.11181802893034509,0.7644148417697226,0.5406854503870945,
 0.025367887243467724,0.6929227981228648,0.1507543470770194,0.95570411842879,
 0.7317863248515409,0.4224403792523711,0.24562646161439572,0.4051103614255628,
 0.4679601841090816,0.3738952097824191,0.8180697258924077,0.7074591725310336,
 0.45543253037819176,0.48456041495674596,0.5412651936726462,
 0.04241292875348668,0.043539780680787676,0.014204980035192882,
 0.7216319492575692,0.7906671321804879,0.8542677240633342,0.18046764308352403,
 0.15985812050655013,0.7344085916216193,0.011439730188136266,
 0.7776021922069699,0.6430258898825678,0.2582245021272276,0.13093297102226553,
 0.965488696497006,0.14029032439461964,0.8718131927516519,0.4173432592429217,
 0.6342420388186719,0.6405114131691619,0.8796887928757647,0.979346327547306,
 0.7655821446037447,0.5054920743497336,0.3218501313873503,0.8746347295598199,
 0.8819589040782273,0.007219949871487774,0.3945660348286836,
 0.11661110510402228,0.7477777577630459,0.6800597123683512,0.8751762910220339,
 0.38905126513106625,0.4852428173673585,0.08111523421857292,
 0.4015355672042853,0.5348013561852812,0.6481707041034032,0.3461046204179481,
 0.9988695646784277,0.43678916795983147,0.6921207906734835,0.4171438305910655,
 0.5384708063622146,0.26413749032833067,0.30939317585771053,
 0.2031881174531791,0.9943711789027033,0.059746987140638974,
 0.3683764774058087,0.7049380773584046,0.0681835024224734,0.5029812281118488,
 0.49238107401426845,0.45088368062115014,0.6793972331905733,
 0.6918480199002837,0.30415396120581895,0.5630544238323186,
 0.07368007622202377,0.10255700066521034,0.6114406253491791,
 0.48968170371022124,0.026462912688296036,0.9729178619671112,
 0.7767042529451628,0.8957691777026838,0.007104586274016,0.009685788386194383,
 0.9898123957765128,0.5466332496260808,0.9809306754947317,0.18136555208223726,
 0.594372075173487,0.7507321504566093,0.7468235795204213,0.9101256116564718,
 0.9530747327794941,0.8596698239585311,0.7924783996403153,0.41558332282921073,
 0.7540930864247487,0.2646454340404143,0.717914669394234,0.6757340574786513,
 0.12917120345121436,0.11434254176626146,0.6248981026076131,
 0.1364529891246753,0.8775239091733118,0.10267840900541048,0.9821882513774605,
 0.14678429880514554,0.3308049531352295,0.6122676190525664,0.6702027103262208,
 0.9391954926217143,0.20827160460260163,0.5834262279406733,0.7970269291279103,
 0.3306650541446754,0.10593791707634881,0.19690589609353581,
 0.1308116259124361,0.48204897969113625,0.6621110952039204,
 0.05301225655198294,0.25135887522056843,0.7451315624634789,
 0.1346805043384358,0.9523185760155368,0.2536902154812284,0.19629579359609206,
 0.4267077646804257,0.9331115862587283,0.8608337289971456,0.06172935940981872,
 0.7465730013746428,0.7611225124819052,0.1878227286051345,0.4747312770621973,
 0.8136670015716161,0.5498148048427902,0.9274964241374503,0.6909178011030218,
 0.05013755809794895,0.15876919344304752,0.6383609135466722,
 0.07821976215667026,0.7842021050562937,0.3173793336403704,0.5222435384628351,
 0.3533095352277451,0.8052666782743347,0.059840162209577885,
 0.4500053339936181,0.6718775801871993,0.966198726216878,0.7098600599726799,
 0.5353948795843342,0.8721353606140785,0.5370889289464982,0.8542920617386542,
 0.5281284177944809,0.8511550653021136,0.18834417156142103,0.5290031519480919,
 0.39345310761055396,0.8634418013348435,0.11000322147477903,
 0.8929278164748353,0.913664588393452,0.5314055223689668,0.32476288266063347,
 0.20647232027570672,0.7472993373762006,0.4661422281724076,
 0.33669963945741055,0.36165070034123836,0.11194954747757757,
 0.8322474767650644,0.621454507293065,0.6937544363353902,0.9125674916732348,
 0.29627131827114284,5.020385848121256e-4,0.5111360753831574,
 0.0017191310000420623,0.8081537382690441,0.8156483855005073,
 0.24002268998385434,0.8624435290851977,0.1107661514587156,
 0.016190082847125198,0.7157532389132009,0.6091838128261698,
 0.4617612993566055,0.44580358352476235,0.10270883419539212,
 0.38304284997526095,0.07256603616382584,0.6476739661466774,
 0.5181926796858797,0.08256273828845306,0.2974850947825324,0.7105880638961672,
 0.349903766554155,0.4950546686070244,0.9808936977347761,0.05122001718650504,
 0.7891267340182736,0.6865397561111741,0.45670988211305696,0.6098258967323306,
 0.16254752668103645,0.5529356389787121,0.824204374176943,0.39716956205194887,
 0.692042527115729,0.7243226315572512,0.08977892091153905,0.9681641823832662,
 0.940340938442218,0.6131319016082805,0.0012299725218305468,
 0.4874571837593996,0.29139642589436443,0.5378837080710022,0.2372309367380222,
 0.7300267026662692,0.7007395939578546,0.21634593164688853,0.9219158035945092,
 0.5699680454402448,0.8413050060633464,0.6446896668790266,0.8164909917897766,
 0.20283959627304038,0.1740534323636842,0.005684501377212214,
 0.9398123108068247,0.9534295255639949,0.30508303491823185,0.2239994697241655,
 0.6149913790363697,0.28132507607152324,0.42190585468364805,
 0.07811086683390267,0.5040975746798487,0.05077054274880666,
 0.2697468591792711,0.8076216063041127,0.826179063413582,0.5671550534801919,
 0.148349736704096,0.43020225626998854,0.8235376313797312,0.12310067767952304,
 0.46575799369662874,0.6633934797557234,0.3356025051135103,
 0.22392226227145895,0.14783879079623563,0.384207274199463,0.6526218704456306,
 0.05989069937414826,0.1590548352671457,0.7212585415034438,0.3167595025468992,
 0.7979733906078078,0.9195813620095616,0.41162522917299205,0.7482890918604982,
 0.7367746265706767,0.8806148523343926,0.8755910919816405,0.7765069849712658,
 0.46381607334658537,0.6675697272358782,0.6815407205197236,0.1315417413173222,
 0.04192333426500716,0.17046978511319255,0.587150377655145,0.9356605926274675,
 0.5673717707282586,0.9894240633518645,0.2682083698279615,0.9715354763400132,
 0.019204053148073497,0.9186067846023556,0.21603910728903164,
 0.3886993490313094,0.5088235846469437,0.42808759647929195,0.3835742557287245,
 0.837087867789528,0.6569655897309794,0.7500402076362832,0.9360002944048212,
 0.2060991692329247,0.31531215063607076,0.6586796033519783,0.7484527035806573,
 0.07503281771328218,0.6740455713201771,0.290156623462014,0.3938415420496004,
 0.288048579233784,0.6668653242355663,0.10073978694606356,0.6550978541056659,
 0.9478459773161647,0.019326923753364644,0.5964419808021626,
 0.7527194332218118,0.9740286005676579,0.8113882490449603,
 0.026182165830213933,0.7947725750288401,0.3175684315116847,
 0.5733816067102988,0.3510692409487961,0.7483037592930755,0.255772464822553,
 0.6556399938456906,0.23158014918064485,0.0645124962905429,0.8652296145573309,
 0.519384049187754,0.48839288484917753,0.4966741813415644,0.5015931482388237,
 0.35482524019009354,0.4758743196237011,0.40696011679413535,0.501553944597747,
 0.8596320995351051,0.9782290279705677,0.21123601673210834,0.8903800426187595,
 0.9549445147107245,0.3223631103280119,0.7261022648541583,0.13679564992101456,
 0.3849880769605978,0.3638799600426087,0.7983246591916794,0.272628800624265,
 0.9857173174966303,0.6183411503708952,0.7355383047394042,0.7132090327719476,
 0.15099726871843155,0.5230368561703433,0.10128487693428279,
 0.9861699736445257,0.3747195066166724,0.8469683053606214,0.1216287295503975,
 0.9297058165056332,0.3079782644206801,0.5968707218408302,0.8806385706380616,
 0.08935459187695949,0.21046605808759655,0.2732247109637701,
 0.8752455446814427,0.4552023329779542,0.5583568104297619,0.16570031168678134,
 0.17915995804595475,0.8916086392638882,0.8650593667276454,
 0.39918379877866417,0.5750082463924915,0.6864342964896432,0.5635278479514055,
 0.5671200580761392,0.09971396241517827,0.3996142761188165,
 0.031613015817244294,0.15423000662541508,0.31811231971110243,
 0.36438954577386695,0.3557621778531417,0.022666612325081048,
 0.22419641195783324,0.23968336365519038,0.12832616109263983,
 0.16338859403517847,0.522399357629708,0.754807020950655,0.10247165164427352,
 0.7267855294981511,0.1697011459563933,0.2726631497435308,0.7363403810580904,
 0.8395232525367063,0.8681816017961191,0.2595539980218702,0.6962016739590967,
 0.15041519531711067,0.06722987012922466,0.983288308234811,0.7920317122757314,
 0.35507600836353015,0.2497537331616323,0.874230951251814,0.5430334737683151,
 0.8575030747336301,0.5873999645315301,0.6244450326499952,0.9009926402355503,
 0.7609822558567033,0.534121436416713,0.9571137050113517,0.8697619619683773,
 0.5860742306548965,0.5100926260551656,0.2761182160744958,
 0.020642374523472284,0.49561085537980654,0.030124566190955937,
 0.7711790293495375,0.8395564333064006,0.4825350355049026,0.4736700709629361,
 0.5817887845185938,0.24411341101244566,0.8762231642201298,0.5045849256535095,
 0.6874748462705247,0.4334333580908142,0.15035586624674835,
 0.44492166226756147,0.7315371659355479,0.5875883980634514,0.2669240873057277,
 0.3904589961160283,0.21373759388702684,0.2617815116324895,0.9739827881513433,
 0.4128845511167477,0.7076946303188081,0.17145060872548257,
 0.18308028036490653,0.3306447593523688,0.642416171633216,0.1500463091355857,
 0.456154697246582,0.8768362454062489,0.48944281714891036,0.9993431707714582,
 0.9697876544775123,0.8866916324762668,0.9139563762448103,0.6297797855660046,
 0.8126690622233624,0.30416368241498737,0.8363803764847331,0.9074365990193364,
 0.8476271429593916,0.664958104362455,0.39151393360252096,0.05874285529332213,
 0.20164036585122735,0.05056093312906307,0.038591288715166594,
 0.30660096437723705,0.2960612138939964,0.3330605920079579,
 0.36549765406947554,0.6989776621442709,0.017185456473621574,
 0.1479054246525333,0.9558204139030435,0.4058040873089581,0.7844506776674177,
 0.7428829066439319,0.8698592529757213,0.8079636091732606,0.4143993775205652,
 0.4461899103810899,0.9619711517890411,0.7429474292028323,0.8559877219877234,
 0.582887665423262,0.08466458515869135,0.8348430745958364,0.42265140587978356,
 0.5797592471869443,0.5391887678879749,0.6536991436579572,0.6921004203282142,
 0.5882737063981573,0.24814229739378146,0.9839169908643569,0.3890949618661196,
 0.29062897551527045,0.8123034202787065,0.8243510878698759,0.572710234455929,
 0.9013317172154179,0.4953045073610951,0.40591894333950396,
 0.37544463537134665,0.11808143489059031,0.33612230824008005,
 0.23636707136604174,0.4090172409793207,0.3268064880294257,0.8316495451395981,
 0.7612762710260728,0.20885037828587316,0.8029178928733801,0.0916725683790176,
 0.6780831090138847,0.1255709603020767,0.3576244982126552,0.8067663262232306,
 0.23771123549625428,0.30674272859082585,0.8637199636206889,0.602911945673261,
 0.20990453951958965,0.25544828736003145,0.5613574361207315,
 0.04034424240787893,0.5296495107183591,0.7254065891649112,
 0.14161877037327586,0.8361135882921746,0.48881244380280964,
 0.9192607488762117,0.7140908963090697,0.09268339518107238,0.3219596936549157,
 0.6583704047466769,0.005095936645283294,0.6111004748601049,
 0.8976588098074949,0.10990391381027953,0.4350394851047783,0.0753829910093754,
 0.710713738362522,0.17471279083115332,0.26864583630306504,0.9603656802331056,
 0.330473631329652,0.1499416623817219,0.6843642948372665,0.6907806068760334,
 0.4178067333519261,0.3254414379232521,0.10259014528800092,0.6600819390867373,
 0.9449427677827023,0.7419772157990984,0.35118034934996634,0.1527151206560312,
 0.219444917496727,0.26631971201593585,0.5587170403211285,0.32139457510072134,
 0.808360523728991,0.2266925965930391,0.7815613606115593,0.47709084376027744,
 0.7874728637337499,0.11448181206603159,0.295856015505206,0.8590313401448987,
 0.33421100692426453,0.6789002030937558,0.4533153711036566,
 0.18487755890200208,0.1659271319751936,0.4535623461031463,0.6312303324443269,
 0.8461861981780429,0.5787181165702897,0.40779435971892775,0.7325466432233456,
 0.27097625324258146,0.14092518608857862,0.19151070822490246,
 0.12353724006208822,0.18263378535492136,0.9076450833094006,
 0.049820102473993044,0.8078932112492968,0.3150884279542212,
 0.5379681660668861,0.021568471323365257,0.040832921034387404,
 0.8853586870550256,0.5679404027360964,0.12479943412084404,0.7092059030028562,
 0.06558108032521048,0.4660697774592689,0.6260539422966951,
 0.22492521197692505,0.6050351986261041,0.9304569463084925,0.372504021918474,
 0.3503814131636359,0.02141558654498621,0.8622177134292479,0.6257268400131091,
 0.15858158747231488,0.9785880407881677,0.9890493187304048,0.9792318542428345,
 0.7854880977843449,0.37903566192511007,0.5550508825372678,0.1965841076251903,
 0.18485815436528474,0.371297684552254,0.11694370829460066,
 0.09887676157550318,0.3256742684983537,0.20505442407504404,
 0.8319233260251939,0.7287236898111549,0.6460586504575241,0.3795770832527898,
 0.48999075567581607,0.514567710067692,0.13207070669837162,0.5749572033834046,
 0.7651212741328123,0.030296906097332044,0.1837879670149205,
 0.8298872392335387,0.731369532136232,0.5012419083625219,0.39524616528750134,
 0.24685490349515737,0.5948003303063549,0.5823255051299449,0.5571751680090833,
 0.9076030794335772,0.25499150068811405,0.30533247404751473,
 0.7414926279509608,0.8411683402577291,0.1512792894766254,0.984798937595245,
 0.2151784348404988,0.8441803024924962,0.23205287909311367,
 0.17512324954064074,0.4908431328290328,0.41572669152971964,
 0.7333398258141821,0.519787693965078,0.2902102087083205,0.8588620171013701,
 0.19810842342877832,0.5809092844151651,0.8520090889396184,
 0.31503426230481013,0.5020377564467762,0.17407283812075736,
 0.13022229554304132,0.7665926260506842,0.6660215646933051,0.4155487028419895,
 0.9319125945104186,0.1301502961070603,0.9111932225981926,0.4474534216678885,
 0.5650528162151756,0.5148443394464454,0.05448385779027731,
 0.37145823890690366,0.22975542016247763,0.6329047915875371,
 0.4056043519903816,0.11173614984560076,0.017730906662206003,
 0.12161563459341906,0.6055843395551169,0.8310783664158492,
 0.47481791619677416,0.4141753907628398,0.6281913029396884,0.6531421542993805,
 0.9661308168614546,0.9030361304024347,0.8833576126057827,0.24215469322769734,
 0.43274641987731777,0.8403904906731834,0.8734447920586285,0.9276076843379767,
 0.300137806499403,0.5288510361372165,0.10388444108499761,0.6380239332053937,
 0.7322063107767995,0.2986726794500263,0.8002597605895312,0.21306532720277482,
 0.8406846158301858,0.6292631006641956,0.737698459770929,0.2357649013014661,
 0.028550440908262753,0.7768792230330223,0.6498359401478352,
 0.7425371406106664,0.4529922997944329,0.34212689512113736,
 0.011905222877814925,0.7730768508918107,0.43298903185931503,
 0.0310558776975125,0.02698138734123534,0.6976748854531108,
 0.22753214821652945,0.14116096335233408,0.08980896305822561,
 0.4443726424526766,0.48577504304107566,0.8707715546357492,0.4725258313563311,
 0.005373193561663547,0.8545895861250894,0.7163885521574006,
 0.5699365701925515,0.22759917054543877,0.8238160236536778,0.6972966225700932,
 0.28780728394515587,0.14202667879853215,0.7342852695119562,
 0.22760533934529392,0.7121420333218127,0.7581550192065173,0.7854201073898623,
 0.9503123528527999,0.41151383925280793,0.0331622307620687,0.9172683110609345,
 0.15300186334288746,0.8883163529310164,0.8935175558861662,0.3047456248524094,
 0.7855885060016247,0.02613318486104166,0.7444826349932901,0.421920450504638,
 0.3779397022124805,0.7480856157852234,0.928123605835033,0.022923488288331018,
 0.1955102655100186,0.9717290407421462,0.4338629258620379,0.3657215418508366,
 0.707482599732311,0.6999001772369217,0.29371641193755516,0.4611151978488366,
 0.13292949495105766,0.31717593770759145,0.5079043520221662,0.704456093132638,
 0.9971634874375654,0.767297501892714,0.06179859523614406,0.35034290040906246,
 0.02352826098674654,0.38254913746116814,0.14297110045199757,
 0.31433804430315937,0.7993069361070096,0.3216026235268784,0.6064219680208441,
 0.281703065706979,0.034336100858017304,0.6844102717575266,0.7209222651704204,
 0.635524820106099,0.8149462358537627,0.2200456839678201,0.41609677824418834,
 0.8189139021188196,0.9647271054584334,0.2181068080873596,0.21489987261478394,
 0.7006659590538956,0.8761342431697122,0.8523455074856197,0.186875034408705,
 0.401720381081196,0.3044017530331643,0.9850685190348017,0.7458911830062401,
 0.7034938357794883,0.2870525315582064,0.2806167825145395,0.5333154713595173,
 0.2706314407948547,0.3202452048783083,0.5462586525785657,0.3248475193791862,
 0.9144460404167287,0.5596755327603912,0.3865849170676793,0.8271270185186024,
 0.21555323262196835,0.6707229592210564,0.3932998310942808,0.5136175341411906,
 0.9504553573752914,0.4039224168751645,0.5867184665034659,0.27430754840765503,
 0.7619063655822129,0.8769423178899605,0.05604074486640609,0.4405107450020529,
 0.7251050979475634,0.32489037161237666,0.10298958252907442,
 0.9846902192267967,0.5523473468197314,0.9016022558893183,0.4126290443291528,
 0.4131155840187757,0.16862660189388345,0.897588212807501,
 5.758474371215794e-4,0.6956236960062468,0.14415465845221542,
 0.05874708525720562,0.7643436109301216,0.8845409979318689,
 0.07186993477933257,0.017034549051394787,0.8696029858389266,
 0.39748768932131884,0.07107552049622468,0.8766127662886105,
 0.5408071136579758,0.5368296451030387,0.7370225470920181,0.3787329797371809,
 0.19119513662928422,0.20301232592211915,0.4416409892906422,
 0.28630492387501905,0.9874843982448169,0.5788270752435085,0.9814297075517965,
 0.5698639094394792,0.8059402448369752,0.8136184391299087,0.424066476279918,
 0.1736493012513518,0.9071698932406029,0.4211981998576859,0.03575503835246108,
 0.2175141028432196,0.7625738309299418,0.8731114892537013,0.34007801984581254,
 0.13208792305500117,0.9008881676945619,0.8423127915827836,0.8748347273642955,
 0.9843869680932105,0.1513697402415155,0.9470642096629185,0.23800054432935003,
 0.2719102419155708,0.6761483919280157,0.8819515960189405,0.2027585936323425,
 0.7875831723933523,0.2947154135143547,0.24534158871545486,0.6890794620151686,
 0.4039238212511397,0.19303592569039552,0.6738844839660375,0.681643170481725,
 0.9169431967413334,0.4646430014606573,0.3159663347568934,0.5727319553773124,
 0.4609572591241462,0.11171373720451772,0.22191064152113305,
 0.34020867927828324,0.5822911027495037,0.9779304700340959,0.6586996755247543,
 0.2095882849292212,0.3262125187284075,0.5019548745090004,0.8855994131493338,
 0.25070222483458515,0.8776658533813204,0.9115058412325252,
 0.19775634877932635,0.41199937174717216,0.23782298007445268,
 0.35040019353974816,0.39981494181483557,0.1640400074300985,
 0.15832471399206116,0.09328173689190511,0.6232035734114163,
 0.4618438646932006,0.96067050689864,0.835082986089947,0.3835742816495522,
 0.8865013423075808,0.7667071670913737,0.4869193374283317,0.46255524460559083,
 0.671915549033874,0.9373333935012846,0.24757292854090784,0.9651335230831044,
 0.03864139662367383,0.5416573861322304,0.37796931632329134,
 0.3522849403798879,0.6750659597981589,0.9304896753483851,0.6990706481630765,
 0.7620454484750757,0.7600154723638352,0.29353634430181186,0.8647944554911554,
 0.38935925140396155,0.15576206968145434,0.7413569087576545,
 0.9637485333801841,0.982618992961776,0.8648366793323568,0.3352411071637449,
 0.6224729177960291,0.6967076716379569,0.27074725673914657,0.9397393215238178,
 0.9506113237278732,0.8582254438332506,0.14340914532524374,0.7828993630083526,
 0.09639299230656562,0.6157913963611972,0.14174364018789487,0.315597739217791,
 0.28869891567647477,0.5607044509516309,0.016504376818831545,
 0.1546326580789088,0.47830041827502323,0.09141847525823288,
 0.7486685185426332,0.15105457052297644,0.8144553946931475,0.8803101335197456,
 0.9463120697353797,0.08749678030596497,0.7747989067554806,0.9952121221005279,
 0.31307998785609925,0.8810567060174035,0.6706210666936823,0.2886770304869912,
 0.7136982542579697,0.3691589448487801,0.25155717411085654,
 0.14997494219200957,0.2897593909423536,0.4460684845974898,
 0.04652147263268824,0.0876555177802365,0.7420590897911533,0.5293887272459791,
 0.5803398635995777,0.8863619211025375,0.17332723356914248,0.6740884470790715,
 0.6100070415486452,0.9386055844036154,0.46567605205026197,0.9044454735495739,
 0.5040771298839853,0.18291701586736486,0.8462436191387326,
 0.47510777538670346,0.25086691520718707,0.477883334226599,
 0.45614778653381793,0.3787375696430326,0.03939090896032049,
 0.3565065685681744,0.3054869606177766,0.013963385582338894,
 0.10691344943009962,0.4139732695336754,0.4981352319542439,0.2307284757147855,
 0.3119990748831861,0.21015778120620499,0.9176617180810887,0.800517477883284,
 0.16797813167919728,0.7379445398408532,0.7444024761005466,0.8352054275279557,
 0.5078175017377303,0.4757853246746444,0.500824695689403,0.18449454041332536,
 0.09879477577201945,0.5293984971179678,0.154418349666823,0.9033822435895074,
 0.6023619766278815,0.5772955676243419,0.7654009621287888,0.45860049078660137,
 0.6628901995978561,0.8079494393249299,0.6340304543020033,0.14350645773778425,
 0.426515028148825,0.3470856052804723,0.2293177824113618,0.7597800287408913,
 0.7859505856522081,0.6713577678119796,0.050322583211670824,
 0.14976429120959844,0.7753290159895168,0.4333757653262713,0.9646717389548357,
 0.25513775161955454,0.651620689631145,0.891126562859994,0.7404928304675381,
 0.1005049032391282,0.83694406658603,0.3180241651251702,0.8416962342846803,
 0.40544804398495704,0.17476084363632183,0.5490751203065737,
 0.13558789776495805,0.8118676675096538,0.5063411459829572,0.9049404840907879,
 0.9989512843836944,0.4860158042217758,0.6660512507988591,0.291434340537822,
 0.15227019431565936,0.430137036984904,0.5479707019348381,0.8389140642841452,
 0.8211801352074569,0.6151789031329334,0.15495176936549848,0.9251922343258471,
 0.7180249677857963,0.7550843525364118,0.6803911079569607,0.9039329633800168,
 0.7119840802387162,0.7966916936719366,0.46516929484515246,
 0.46168172122976225,0.5520610225262766,0.9426075608094864,0.3055156800038029,
 0.28152022030596147,0.04003820021461335,0.41349041594111857,
 0.9128400085410533,0.4884090340207039,0.9152871831008055,0.5861144254788537,
 0.11747529319830696,0.4866167408513531,0.17014612355613057,
 0.7932173115744728,0.008217205909470171,0.13850615825152945,
 0.3785395582786425,0.033085604454641704,0.5223169007970299,
 0.4289809169516946,0.8102521339951828,0.40241985477842734,
 0.14759801561236707,0.07917519354321967,0.26713214259786966,
 0.8246251677485288,0.09586977727264123,0.915016565203645,0.732474692388429,
 0.4488344062463906,0.14105537565451387,0.983708416967415,0.0514410705753916,
 0.07636093089655538,0.21874559409701233,0.4629035644459356,
 0.3412899305724344,0.49225804274329477,0.4631994948290006,0.8412795929604779,
 0.1389074552565831,0.5657439463381879,0.3355764253268052,0.6283705141929926,
 0.7773942435230241,0.6445025513369674,0.6261268220033551,0.2869365365933798,
 0.5477049833503497,0.8106704806861417,0.9511200331882959,0.8799929279944969,
 0.05808003670010431,0.5645301430308722,0.6610126758304607,0.2632608685908887,
 0.25214914789774634,0.8241263208132174,0.38665761983761127,
 0.37045734669308195,0.7028652882035551,0.19207952215074642,0.597755484725744,
 0.887157102589924,0.9823445906256938,0.5112745791051299,0.031490548163147114,
 0.37021323806855055,0.7562181929469616,0.5937846729565968,0.8675065155659084,
 0.9420146198704031,0.7225824258141363,0.8967562725346532,0.7394621133386694,
 0.9443866214416996,0.7806830114566012,0.4874084388763793,0.1544156219219157,
 0.433039637233935,0.8310801709555422,0.882889284774047,0.8745369877479424,
 0.13637043897521484,0.9472719410578091,0.9795547561771073,0.8837808431907943,
 0.36676034593023465,0.19806150970188585,0.6042397638995859,
 0.21469051249563398,0.9494524561247285,0.557996509328246,0.9402256213539053,
 0.11658126259887747,0.7400339163233504,0.005664973398258111,
 0.9158046839364519,0.6524773508507392,0.8399647434647525,0.476920442428433,
 0.6955124832599192,0.9702994244864579,0.7149504186427795,0.8455501135050358,
 0.9404670727853031,0.4730801567732238,0.5566667170929911,0.9688506587389332,
 0.17594413796114228,0.5781155645322177,0.15794024803831286,
 0.47697383983262476,0.6571000507643311,0.2559977916951852,
 0.040385132115209776,0.9590453892165117,0.7744750709126844,
 0.5973167053437042,0.040201534810493045,0.3071910731549913,
 0.10465674698120053,0.008265830377713335,0.17944972596965503,
 0.13722198926237905,0.7799134307480322,0.8348440901167478,0.6604742140351659,
 0.32626833574887615,0.1740213599035565,0.5457180888029649,
 0.045369871580602905,0.37708690472528494,0.40438884794871277,
 0.42434595086726035,0.6769274264838698,0.1832432232755895,0.6973967298335,
 0.03375828191100594,0.09513769431175156,0.806219425275366,
 0.24821783677497844,0.7432375312376902,0.5608417970534998,0.7463167245587207,
 0.018840268956732187,0.6195023389619465,0.354059449704899,0.2893039384108712,
 0.6246753141168292,0.5147600770426037,0.6814699823980234,0.9283750263043757,
 0.07283521396358672,0.925550554713112,0.2718803100852648,0.3520061708272527,
 0.7589377886707074,0.9499425669526167,0.4526130242547819,0.6134961159225051,
 0.7328193441582963,0.6206646605768159,0.872635766589783,0.1474688490627558,
 0.5915018991573833,0.8823606090605135,0.2995402205070121,0.9336126986675662,
 0.8094968443135502,0.6671208676186722,0.5833983884663448,0.8621073750619929,
 0.5636891686599678,0.8055468337393467,0.3283359428605106,0.9172127480575902,
 0.07009601473518179,0.5583285620507297,0.012899371333146448,
 0.9992437828828059,0.38957051486242944,0.8966758575638619,0.5675754599485175,
 0.8274924139034828,0.2728434325206207,0.9703430682921499,0.581834831435335,
 0.0021440137638891743,0.8101193019382739,0.7092299995205402,
 0.5293686128595632,0.4837028808709447,0.15099738098291515,0.4633624048510111,
 0.006234692592418933,0.01531391048002484,0.789739135378814,
 0.9178108337376275,0.69066360051426,0.3422038533258218,0.1600150710748317,
 0.7193600174635553,0.3894950798932063,0.4124045327703275,
 0.015186414216043609,0.46054020510157245,0.5010767398183482,
 0.8728629867095166,0.24222029837146797,0.6789099942693195,0.7175998967925576,
 0.27821252045016853,0.6969456095838823,0.12967783562255852,
 0.22903549116279587,0.05942860686784157,0.8064528076829669,
 0.7790058694264301,0.011125781832421744,0.36620602743771324,
 0.8493083499753182,0.6467129104627946,0.7334209143887865,0.7614414171285305,
 0.4701454626283186,0.28109967834816807,0.8023137903959451,0.8655024277031381,
 0.7285896028503134,0.0442387935722407,0.8168886780617214,0.1043768144269559,
 0.12125787629684903,0.17789767935521117,0.3763505246635057,
 0.5054684019106414,0.21602325398201394,0.4585951503293899,
 0.015553250523321571,0.03288740845379967,0.5247854754980276,
 0.7249425225914767,0.7933655793331875,0.5036443301443736,0.9126410726786709,
 0.27116408159070027,0.46147802430864715,0.36816398185348076,
 0.027057417021499552,0.906568115646736,0.4027480133849768,
 0.20270140762506617,0.8136221154111807,0.15563034503698792,
 0.9480303576337832,0.612027077297322,0.2307418729808317,0.9477629564693635,
 0.9636795285434305,0.9582691933331182,0.1779577640084442,0.04954234950030012,
 0.9166842740604385,0.9602454239639788,0.2915921061675858,0.4323944659107183,
 0.19175256470599178,0.6838780008388639,0.2527551574920408,0.9520542500538615,
 0.6799385908145457,0.39008712921581346,0.7114546586839003,0.7472681116368806,
 0.843465332615803,0.6229483017788754,0.06255829585937356,0.9285613947598066,
 0.1995974768439911,0.4521069056843193,0.5200276082689044,0.16483093899187695,
 0.5869096400739073,0.06840823640689886,0.4475609821184521,0.3179257555484538,
 0.7641793729548126,0.06625637002377849,0.1339040638987531,0.3957804047904814,
 0.13852692192984395,0.17720858279754736,0.6452238452829591,
 0.5119730433143117,0.3675362013786668,0.8633666256596692,0.08379224481479841,
 0.0483296943607574,0.4409208700563011,0.4652368535614603,0.8230195547665471,
 0.013146981684678316,0.5729985400315906,0.75627302525355,0.20852621858444376,
 0.07930351131957258,0.08223574789959065,0.39455919405357476,
 0.6521724533858239,0.668332317489504,0.017486847765253444,0.5926133699490439,
 0.8832410103603715,0.12416152306934569,0.5484750600374175,0.5823852292639955,
 0.14965187025730964,0.8521380466875909,0.9254965342689528,0.3378810279708132,
 0.4097006668630494,0.6225130558093979,0.9203612192723998,0.26643047621179583,
 0.5729396297250121,0.8442079231673963,0.24381660925824733,0.9668568742086878,
 0.3820681388771189,0.16247827795338265,0.5583900559098707,0.7282399581189378,
 0.1254347814686867,0.36552512829064876,0.46141287618133986,
 0.9324342879836961,0.2923441194690368,0.854845253436291,0.42739334634354487,
 0.7939276238781632,0.41111472347877753,0.209523612218383,0.24893223613051596,
 0.598945297176828,0.5991708108421402,0.10362457341577591,0.6182339392689113,
 0.24733918438300795,0.281541408786556,0.9456406150499633,0.7718069705252601,
 0.727778100417256,0.8877728443515679,0.2477991983752592,0.3719752606756692,
 0.9574848567207914,0.6844209566861663,0.9295599166354279,0.7845584674277987,
 0.6923175549656322,0.2413730126220296,0.7193701976590059,0.04389017247654481,
 0.14910437487416672,0.41716571326861474,0.05650227288531473,
 0.639073801815864,0.2885008157730752,0.1908283243224398,0.6845165480915321,
 0.041645548810922306,0.2713344694839819,0.6301877962904964,
 0.8568895259592472,0.07856193358095886,0.6537998646065021,
 0.43708187969598655,0.054559938071902514,0.630684645312714,
 0.5705164109724801,0.9729155925506805,0.5113654649018892,
 0.017504019371099466,0.3226157787100119,0.29219785335988346,
 0.5821166514821517,0.7932785688034658,0.4544292601558284,0.30268313280110404,
 0.37450086389156445,0.0578985603137705,0.5403099816123835,0.5709000906957379,
 0.939488223061137,0.7391118965226464,0.3305269680652352,0.8686353372604102,
 0.8898949173111956,0.7611735990531777,0.8383348150776251,0.568431582335647,
 0.8264577873780592,0.4357097248118108,0.4621010770764138,0.8561948515416942,
 0.397362795060235,0.2603720045198409,0.6204602280254569,0.3513698536319062,
 0.8518265417618875,0.44337837315813733,0.6117745888808248,0.9046784581042043,
 0.5303412965809795,0.15712774859837442,0.8298255290319635,
 0.27808936479328716,0.22145975867740764,0.6967225950659592,
 0.2843162077833046,0.34903187027494553,0.7036314933248957,0.9955914618357457,
 0.0697548046623847,0.09398010265048562,0.4984307982420073,
 0.08231561394346587,0.6520065174799968,0.818246601437716,0.5404990288402344,
 0.23243372090101444,0.8612738071105706,0.615935948836664,0.18904482338924444,
 0.722086759041758,0.26064551407296777,0.5429618159896549,0.8088501676050047,
 0.7881635832841086,0.6246395263415505,0.3054244391331923,0.5495812757615222,
 0.6415382623999487,0.17071622169026934,0.5648456603310511,0.2640315904097422,
 0.4252336833201362,0.28501487676705917,0.1697611041997391,0.6675710549390453,
 0.17359762719151306,0.16833943367637982,0.5887113402768319,
 0.8543892777265024,0.12894692603602476,0.3176313300896981,
 0.42712327568963426,0.8042356878427536,0.13498936883637314,0.597360152390549,
 0.6730051956941345,0.5194719652379294,0.8807357424411543,0.51285459294558,
 0.6173535848044689,0.25262936461914953,0.23480386755598326,
 0.8725367412562489,0.8682472764603065,0.7184309597946236,0.9038427332123036,
 0.7629822921079318,0.003511274574280021,0.31951653354709597,
 0.17385000089740454,0.6656102451856374,0.034657302912453725,
 0.7409745993739185,0.09041323498977416,0.7293175293828175,
 0.38840963692318586,0.5112466337605602,0.4129176993714504,0.3773965608775942,
 0.023626427017773954,0.5675148657007388,0.43953370497009114,
 0.805787196604195,0.7961917245193059,0.12846812067515723,0.5171571959867868,
 0.029395154885274333,0.3135021841810217,0.8345833607426064,
 0.33300847788463117,0.7661122936174029,0.7683260667958813,0.8102554305164482,
 0.8352910147472759,0.17412757760138642,0.7580562283977083,
 0.24379039194152796,0.929497065097652,0.9813324866416944,0.5786333524742018,
 0.9588202516035818,0.529607488208719,0.9404708949495131,0.5020300582393221,
 0.3484707117472816,0.23817545078744762,0.22751421528919002,
 0.8268162720974497,0.9657028521026572,0.3302388390291857,0.953533662609816,
 0.5527529488077937,0.37884737252485845,0.6794364995190654,0.6373998942331001,
 0.14043180503057862,0.1969363832676586,0.9455197421559006,0.7943674703491254,
 0.49370707110879586,0.24277389827356954,0.6684578706671269,
 0.9147012927078487,0.8629627137726004,0.7720366680412373,0.7279705699164827,
 0.4114184592594087,0.5695286653798185,0.3805314009911276,0.7796678098161466,
 0.29841817295668827,0.6993903749056237,0.7104515252295315,0.7057866589206387,
 0.683241184929717,0.4820746553317472,0.45313925542480016,0.917602606463213,
 0.23851459176272383,0.7444324141585723,3.6936351078564744e-4,
 0.22102230497276987,0.1969710013615209,0.8014388197723038,0.270036478264556,
 0.5325678245070524,0.6027703721606693,0.4290642959828572,0.6072621367518719,
 0.9474705397368569,0.7081366853038144,0.5859288512932405,0.514011819586096,
 0.6836370050363603,0.5482478969416095,0.6376859251770384,0.5864506089319578,
 0.5089351707273502,0.9670941659229608,0.646349378901232,0.8065468873472483,
 0.9933778312320409,0.7364550620789042,0.9372918830049555,0.37583940511614766,
 0.9882040983630594,0.955848303577758,0.07310292257874096,0.16569215113846747,
 0.19663754037356485,0.889149689529508,0.07995431660111807,0.7105070275542713,
 0.2203226462691168,0.7572279340758732,0.007616094543948293,
 0.5365471781425438,0.39300688636042946,0.6032963324695699,0.2408436423109983,
 0.322022681735308,0.09205315078681042,0.6524735433049904,0.7030516899336492,
 0.6003591821771269,0.4588212640809173,0.6882019112698573,0.4956400625824994,
 0.17763230632638316,0.14673836289974296,0.6288421373490906,
 0.20165618887153325,0.4227612818869999,0.37633334851798117,0.758297049117206,
 0.4888089778492829,0.24476835932684804,0.13572129901808994,
 0.5654056150218083,0.7808573886653718,0.5005911382461036,0.5929022230953647,
 0.12886663196448978,0.460523591626414,0.3202039573862393,0.39064007765973596,
 0.7938475417472479,0.41251172929078483,0.7398361017748641,
 0.24603197579257907,0.5455360384476264,0.30644313388766864,
 0.45094435857052195,0.8208448650776912,0.0850827313549386,0.0894422656033852,
 0.14600134233695516,0.9117449272853696,0.30630743000797667,
 0.2524462681478542,0.7778609371873515,0.18883737530561484,0.9539250447358515,
 0.7344815595628683,0.5145788681495703,0.3910582436993747,0.9028555889599895,
 0.29636023982236914,0.2706607192535082,0.7172721535880433,
 0.022114169464635403,0.669712391814071,0.2139035296432733,
 0.37117353535877606,0.5331467858754737,0.34001349775020695,
 0.3783301100064307,0.7358666513394322,0.02640689088158943,
 0.14719514058964545,0.3119986425018014,0.9669584824061983,0.5661756450532379,
 0.8405641430136338,0.9475831845078998,0.6800892503091316,0.6457198867634973,
 0.6652096953203614,0.4551136542534555,0.8070748586001517,0.842589733518178,
 0.8028591880336426,0.967520117549064,0.4523717137081752,0.754817167015262,
 0.5767452336092753,0.7515347336346213,0.26999171988780524,
 0.24459052216110433,0.9663288641826272,0.30136536847374895,
 0.5722988035201355,0.4380613163592335,0.09455904960825279,0.5966035369039188,
 0.14458339271971854,0.06713578079235638,0.7872722784245667,
 0.44187740145307586,0.040979512810492724,0.23567781371543373,
 0.6556330710225908,0.2054434164015193,0.7081749024576746,0.7638098230770007,
 0.46917030283588923,0.040016646018065205,0.9801307292867872,
 0.5532694681321433,0.2045283953989031,0.8244213444110351,0.3804140060034443,
 0.277993539747181,0.5061080190050152,0.22249276010967356,0.17033032021200878,
 0.8455502468670708,0.3443779951298179,0.4798130703823761,0.05123041573165854,
 0.16125840993811713,0.4357109583620665,0.35422324918957726,
 0.06626665431964196,0.605373328346966,0.9488149924060273,0.9982807572815455,
 0.08368348581014395,0.5136421360932932,0.7525586535557417,0.788334187515852,
 0.5693390554762965,0.8945221583706409,0.5310760393478096,0.07469254933246083,
 0.2608270660385462,0.7721414169244366,0.538236374535757,0.8873437201032305,
 0.8800093470809118,0.418590627449425,0.40827264708188393,0.6686415174977026,
 0.8260037631002954,0.6550288496442823,0.10591764594390463,
 0.026493779525535888,0.01519986023556763,0.8830549113734688,
 0.8497495308009381,0.6743643530305955,0.7072470335345508,0.6135944000714643,
 0.22874812296250358,0.4828859002781948,0.7789030989843837,0.3359508179763756,
 0.07665840480512198,0.1841549998496257,0.2621768664924258,0.4974796860722893,
 0.11845124732316659,0.6346521238030446,0.2758482653744294,
 0.40775376784930106,0.7988472150096521,0.05353963792302174,
 0.6950496673037357,0.8922762224929027,0.535312808465503,0.4842888620533129,
 0.6830864553021716,0.4846927814919786,0.6058926286076085,0.5205260901716672,
 0.13021923545736125,0.2010283008248357,0.03526430386975432,
 0.7803386257800313,0.9439059796006084,0.6048497422332463,
 0.008665127378608695,0.25978501046303415,0.5369760924238588,
 0.6344082059129912,0.6571184637405255,0.9920359712265653,0.6394475185812072,
 0.1769861450734963,0.6139341142783521,0.42943414271342084,
 0.10780036994771836,0.723993200002125,0.5941554281727734,0.8698753315242409,
 0.8525480276367522,0.5299274533619722,0.7610181852381491,0.6527083118387701,
 0.00959030100944891,0.3725668329567402,0.8066162776453454,
 0.005843594882730563,0.47972262530640564,0.2534825959402087,
 0.8420193021822984,0.10861320200284152,0.5866633602799928,0.6455964841601171,
 0.9422621793485466,0.0037692907163342504,0.8895218208110495,
 0.4654158416833418,0.3579520021835263,0.5813002607738805,0.4542245712244294,
 0.5811489416406261,0.41060988095840845,0.26193994590895797,
 0.7978386436573129,0.45246608222738605,0.7915772163450727,0.8534771015085131,
 0.6747678209968827,0.9791604351568591,0.3303883408582069,0.9944582437929322,
 0.13474103711933205,0.5893171013129943,0.4741619177309193,
 0.25860088072447385,0.001356700330448657,0.6557011788482252,
 0.7562055213492425,0.008772897578818384,0.7228124847734445,
 0.5101146353021414,0.8017748815706826,0.7493858139075369,0.5264913872277321,
 0.020238679075911592,0.542477290840693,0.23115246742886542,
 0.7656556241893617,0.39745338470553293,0.10728728755846051,
 0.8806941341066216,0.013022656756229378,0.1310813338729938,
 0.3309940797395391,0.4127362043384456,0.6803826702906074,0.3306791206482611,
 0.7580987627910324,0.528093326986635,0.3876858824501329,0.6642586983003168,
 0.4514590289673136,0.5722659448834236,0.14507398425544882,0.882919373443766,
 0.6357986256419145,0.7966192177132694,0.6427647843497597,0.25534915090227994,
 0.6852560922997263,0.07486722130735624,0.7680809913246225,0.9776385282406119,
 0.2228523687522137,0.39617496710802924,0.5375762921539093,0.7575800195817387,
 0.004599481968653341,0.20585624617830334,0.928612123574461,
 0.33581254934288896,0.7125333084864343,0.99090425640175,0.5764421362448867,
 0.5941716143315099,0.18518973410814377,0.8297817635127588,0.9096792060534725,
 0.5619084736487991,0.13187758307397834,0.19302412398356006,
 0.5264472204741708,0.41628516919413305,0.46743370951363705,
 0.5059782894882862,0.9934234622388288,0.4428737800578214,0.5335887326361325,
 0.26941747448037456,0.8407736805487003,0.6023524184402922,0.7656300237669392,
 0.6907627584424111,0.985520004688144,0.053121580221413334,0.8244252238086558,
 0.4364144461656486,0.23965684294536493,0.9362975962268376,
 0.05608441765858607,0.6742430120596434,0.5715413523024492,0.2249619874632257,
 0.6674846514180365,0.9946261236813031,0.6331724150692957,0.5309879297439446,
 0.9579546126814092,0.9425963443374079,0.812089872634412,0.8315186217615537,
 0.3450406293493278,0.06660042305411162,0.4800153427505077,
 0.23436328040915344,0.4591476792339585,0.18587478072438735,0.653107022841062,
 0.44061260507684796,0.4509775694564867,0.4876749162136419,0.4908650966753564,
 0.35029151757216437,0.816883494044229,0.2232354871808162,0.6277229849257427,
 0.9636362972627096,0.08506372057994493,0.5197130511407971,
 0.005363705414411557,0.5337813875811341,0.6905311971596563,
 0.8112257156527982,0.9320642991618345,0.5728123277486911,0.1834131450910952,
 0.9605336802633718,0.5944204139914964,0.2716767036082307,0.5702747770551648,
 0.5946089478036716,0.175304873980898,0.6575758396684465,0.930525767190398,
 0.7606366268332034,0.37009156913984587,0.3536040722650541,0.9234029297868845,
 0.6358602757420684,0.11626911503358384,0.5305368605780723,0.5611030459049784,
 0.02691743748253239,0.9309314310357986,0.06296252077571907,
 0.5262640055930279,0.3813106886138392,0.9095060376038784,0.25299169249636655,
 0.13529748372595385,0.588000470377655,0.5653867103028598,0.5099062089318522,
 0.7797806070486177,0.6037111367543238,0.3536179909850715,0.4200118117634366,
 0.5144658020384568,0.5240495051234341,0.9231151283433761,0.9164970159070593,
 0.9590056240418755,0.2601844135402136,0.14842945170430122,0.6465663325126394,
 0.7921944181274246,0.4635226168930331,0.2903760709468529,0.148423878451742,
 0.9477692483567725,0.051381288025521776,0.38267238186112507,
 0.5556427669035933,0.7692511188450264,0.7353537033785045,0.5209089485031893,
 0.8634518137458733,0.1235987733920425,0.46528322881503237,0.8116737537957786,
 0.09197217153349313,0.3620877371350816,0.13822613271811268,
 0.7301901023656397,0.4770881985714367,0.8487846201095519,0.20292620430142316,
 0.8588989291836316,0.012134203495079099,0.19879933058430965,
 0.07015183533924296,0.11563752551784467,0.9320607947844066,
 0.40831871255829977,0.4957487751523503,0.43743738131804255,
 0.45272417310064805,0.9021942587995526,0.5376931622761719,0.6027164155643626,
 0.029245686959025075,0.09108559817048212,0.08112194827333197,
 0.5686839680935915,0.3779136006392637,0.6758471887163209,0.17275665074177105,
 0.07454619285416642,0.6539810456950956,0.07007323516405295,
 0.9363936412841479,0.3764029014271717,0.5244030513109755,0.8930468781547508,
 0.9592104127255786,0.8973190239666312,0.7025422236729845,0.3865094266605922,
 0.10274791660779714,0.314694991153206,0.9962542747660628,0.11583315074124811,
 0.4950456071646121,0.7774699002873732,0.47204748544462805,0.6747571416198263,
 0.04982810832347939,0.11302152028525592,0.9786616411981324,
 0.29717551630695016,0.4424013370677461,0.48598743151888146,
 0.2485518815059372,0.1883127694035337,0.6030504578845315,0.6304303872899474,
 0.7992321694822687,0.39315049417646397,0.8843915062340941,0.5655674086817029,
 0.8684559861221609,0.10604997514461356,0.9815763360643752,0.3736597612047544,
 0.7682883698136268,0.28318090767862625,0.46199307549686497,0.842025204769316,
 0.43566585746460973,0.7452549278750922,0.24521143484207653,
 0.1707346208261442,0.1730826589618707,0.9292433263325663,0.689079201988846,
 0.7287694653413754,0.466753142409402,0.15285880892177228,0.01215798981838867,
 0.05339992597807597,0.32229785537405997,0.650639230900286,
 0.46211655168470844,0.5993821531845471,0.32875289840702226,
 0.5230309500490388,0.2302832809969816,0.783725047365808,0.1255225266599076,
 0.6173524946015213,0.9346472309477256,0.5959502411960399,0.07716124795604151,
 0.0201632556024659,0.5552626731149639,0.5430638566608333,0.9315462745561026,
 0.23332443709544493,0.3179697781854818,0.8902106630865194,0.2579397318353812,
 0.4265169728879499,0.7942669649913834,0.7053347731977679,0.4139971842023241,
 0.9837940075007117,0.23054920510817023,0.9122215681863789,
 0.11494415595379126,0.22303611534370815,0.062223785581693836,
 0.37236184446507037,0.6373042599644894,0.9919553780929893,0.3177747519702392,
 0.5064406915949903,0.498121377447583,0.6450099198697683,0.21590396568418802,
 0.8911684461017546,0.6828027666943288,0.9815495348793423,0.7179579848942906,
 0.3723742544247528,0.3837346275051754,0.6830730970732255,0.20118966187244625,
 0.17069848254157982,0.12807022579518512,0.44740342927579657,
 0.633586716620029,0.08110169383571009,0.47325713632400623,
 0.44675319189314466,0.2760788186376155,0.10728214282382498,
 0.48573658764490435,0.9185581704598037,0.19370799251799065,
 0.07313491522774351,0.3955870960761011,0.5858322349511276,0.3976959357324531,
 0.15960464271960206,0.6053914348912287,0.1802875894888385,
 0.028633429788629372,0.1432176277601268,0.34175370750342293,
 0.055910381301852,0.012347610940092402,0.08481114994449579,
 0.8522308742402493,0.9897130734793198,0.2216529181619773,0.38007558903474403,
 0.07867679447202769,0.3633848974497613,0.6539064050058002,0.2947671341501925,
 0.6549710024050747,0.1671244894030749,0.6709743993718933,0.3200216362849966,
 0.2980357073136063,0.19296791603155095,0.30255397779031856,0.860469770922394,
 0.7372386621741628,0.8890383819355246,0.734840306141259,0.418134128942963,
 0.2819229583822912,0.7280298090350157,0.27031781009816047,0.7390370587922526,
 0.8849098032404211,0.1215908559233605,0.20896007707521502,0.828003817246524,
 0.6059721141012444,0.5701938844492287,0.7622053653500483,0.5756193514255095,
 0.969642272724809,0.3757911615240155,0.030449605058879947,0.6530622806428238,
 0.21686534076432984,0.1535587846916745,0.9056330370169574,
 0.20384767353686595,0.4577902463165029,0.2958816758645092,0.276537402747262,
 0.34004716606701446,0.1830098090815787,0.05350254001588961,
 0.9470052001379533,0.23682392736521718,0.24813570504391214,
 0.7191819410971118,0.528591956267698,0.616410442260105,0.05579648296035622,
 0.19494545011858389,0.38544793886864337,0.38510123014220965,
 0.7466117803055901,0.23074029048930045,0.15280720918433432,
 0.1642201810321452,0.12058303774204637,0.9884841455582346,0.4454406680761649,
 0.7791072895976265,0.8527616697940866,0.26670753339589526,0.1497105836464261,
 0.22441616254615437,0.79283698897993,0.44920629867619044,0.9521819242879872,
 0.164473183361038,0.8689947575580497,0.5752030372969478,0.24190079917035667,
 0.015401712053242922,0.11846887407872586,0.6187419172369667,
 0.8188157247081075,0.17576278846794924,0.5858769725157118,
 0.020518792029934785,0.09588110455789634,0.671900571676589,
 0.46536514708579446,0.057580505861856146,0.3339984891181146,
 0.8283051651127342,0.24639903225884885,0.23567094560160262,
 0.6552379925415125,0.2967839149814483,0.9784479655964374,0.8735228272647488,
 0.5363173278499211,0.07878587998597308,0.04151484775601921,
 0.25902149292194987,0.7419619798797774,0.19725446464496454,
 0.5341646813711356,0.5339265301867473,0.5726925925880748,0.9407916164057042,
 0.8647851179593639,0.6974845014287028,0.9730146666566515,0.7469078511095397,
 0.059398119308172626,0.998996836087046,0.2171109549363246,
 0.08031751696493283,0.047209331395926535,0.12476951355891175,
 0.8673905070328436,0.953737036857162,0.8278984513140081,0.2701721374420174,
 0.5283029142031199,0.26947837552953935,0.8734142221438026,0.8768042142482434,
 0.19216023952280614,0.8400227718200103,0.7348295323697628,0.8571056189499295,
 0.32022519330502375,0.6627014419047162,0.8907293524158207,
 0.19064132460379712,0.2250648207324173,0.04591648479233179,
 0.018612849354175842,0.5297973896640373,0.2349694730662989,0.56784403142689,
 0.7329525462380133,0.8580698767960917,0.881861404883074,0.8446789155074269,
 0.374773175033563,0.3587502081278925,0.48119174372332996,0.29264993158247554,
 0.25847271518975345,0.39243342840306183,0.79661115066963,0.9520245355068585,
 0.5528324651503347,0.9675290027436345,0.16381029440131167,0.8811706388292132,
 0.6576307389099091,0.6519400936645103,0.16362518240669743,
 3.6330450686794524e-4,0.2220999839132477,0.7603457147436041,
 0.9680112807702984,0.23788235846903838,0.13910273273150342,
 0.3942367632801359,0.29416971053221996,0.574453494714385,0.6903539530612278,
 0.12393116078904842,0.2404951605296528,0.9849413086550874,
 0.018290379912459964,0.06833863217730307,0.7024647089072885,
 0.6056722839508256,0.29041974497345235,0.9927507203535462,
 0.14244697962270503,0.14893378210317787,0.5261516315153114,
 0.3024660024176503,0.4999638239890075,0.7589650532882839,0.8487206887588796,
 0.5055027642392094,0.10664290477368688,0.569544734312147,0.2420022005632726,
 0.5461728685420879,0.42213390719763266,0.9931969445857649,
 0.056911029817502845,0.17889205082575943,0.844275554160792,
 0.5429590974906278,0.5730513846565228,0.9733004705323467,0.8230925555835361,
 0.960177761503943,0.5957603975039384,0.9093411429042765,0.46533173207611445,
 0.5626611902040625,0.06977360618797057,0.8894559327744067,0.9107022894840052,
 0.3148104388783023,0.30279931077734623,0.9572598765148441,0.3289518537767241,
 0.53369784570945,0.983020204819755,0.47672925838938973,0.7311648538695874,
 0.09619305527452704,0.33767045224383996,0.236373742059663,0.3164346263171939,
 0.8243910026459755,0.07735570508496714,0.2672257350340871,0.7620579408879926,
 0.9515480742594894,0.38701072552455074,0.8991508496173093,0.2187649084695783,
 0.819339523243882,0.06917431053968248,0.6837310509811991,0.39010453232835884,
 0.27330866328235004,0.9333351121983239,0.6475303775816131,
 0.02990659797442341,0.5466016837577643,0.45374169004510645,
 0.8942722398230182,0.5547935890587705,0.28694255192435647,
 0.38662750528752277,0.4172118477771829,0.941536186198575,0.7026922898324482,
 0.41150508473465264,0.7197176674847281,0.23122225267681185,
 0.06840433657054523,0.8130120251358788,0.26066842532344026,
 0.9513262475178602,0.068928081621345,0.3611070324319199,0.16585127011544365,
 0.969590343223231,0.6562913321615953,0.4765856577079065,0.6340289622125798,
 0.23495504168153403,0.6666033904413894,0.24104472040185587,
 0.5846681871912949,0.1309909653704806,0.25909976277229985,0.2482368634537122,
 0.47774863837687687,0.7590014997766196,0.8404224669606148,
 0.13503177909857988,0.8224660839042992,0.5042620207351597,
 0.12127919807451448,0.8485036591246422,0.36219617091150347,
 0.19470658903744442,0.4911811888142018,0.6464121841887173,0.2076682763807962,
 0.7890676247896757,0.34151858636662924,0.771743283142095,0.7925393309916577,
 0.9246415467622792,0.8614884330657093,0.10238501222813445,0.8696857351577784,
 0.42714259921126985,0.25198797277343754,0.06910240274769774,
 0.6344042490640777,0.9973860126539134,0.9907385136742991,0.5063420425918314,
 0.560532469606213,0.12749238159744392,0.20247878275779063,0.412044588907312,
 0.8989841155482151,0.11957883595129493,0.9635880470538462,
 0.061626435999277085,0.11989227213561948,0.939369561850663,
 0.3567759921588045,0.7454162382439089,0.7385908018013084,0.728527373802475,
 0.8810818527363792,0.7002297412751566,0.20790451832304957,0.9909220257976762,
 0.7923298961934269,0.08265428196728986,0.27998479740917637,
 0.5154924046794954,0.21940297061394276,0.5861694355610277,
 0.21199621468303798,0.9827118810912084,0.012443861814241175,
 0.8570091402642421,0.9903459737269952,0.779150108981549,0.8689245709952043,
 0.15841381576350777,0.7098943513294342,0.5653833778939721,0.3006934895835004,
 0.7290123744935453,0.8781655213804497,0.4512814242935792,0.8469015129193933,
 0.3404011117255592,0.6638447440372015,0.5953304832229767,0.17980738055515777,
 0.734683417729139,0.7651229836687014,0.6137726686176282,0.43718753344387273,
 0.9437247626986663,0.8231159561561706,0.06307618309787699,
 0.46982941007021684,0.6873028787092224,0.21802708800576087,
 0.2986335866776316,0.9552851714296542,0.5660298970069373,0.7603107276042561,
 0.5550695862817291,0.38068374823178663,0.1993335380338288,0.4539926870600042,
 0.033344383142473344,0.7473864634239105,0.13013896817073212,
 0.7697797384385752,0.5038350090841056,0.8521995055366093,0.2683610842002412,
 0.2954691221246927,0.6885642284168342,0.10349355805397964,0.948140972489829,
 0.39003338267446774,0.6422957219251615,0.3080238333647395,0.957695762667655,
 0.6157906503417101,0.03211221763395944,0.13763026650184462,
 0.3450213951638028,0.910200980247857,0.4476803905164304,0.7639554372286963,
 0.854886532293121,0.7603395340051186,0.32690265582315126,0.7105479106520234,
 0.9335176293451669,0.31410908349254796,0.2823955435729779,
 0.11611323360682724,0.7893492845213351,0.6388641567793371,0.9580738628667845,
 0.43223831744892705,0.6710409406593971,0.08069802047362307,
 0.9231426444469515,0.11050158789465048,0.6881883547585156,0.8983480689928421,
 0.34781648624164063,0.30976742786250544,0.5465445339203553,
 0.30744302852645433,0.5382674739958904,0.5988825388147517,0.9132714729311121,
 0.14192136277416068,0.49839325889972175,0.005536684676948456,
 0.9855632942005405,0.03829883442734805,0.898724714794598,0.38519379311248736,
 0.2104902807068687,0.0836872915309147,0.31831612079344607,0.2483183554549121,
 0.8951532934783721,0.06045134652205841,0.5838253093063426,
 0.13557856836780424,0.13518288823505098,0.22799138540119268,
 0.2453467094364502,0.7035745218600808,0.329848731134594,0.9792277006619124,
 0.002661991686968679,0.05518822537295898,0.7746240678109396,
 0.8423130021751407,0.7373536689366086,0.08832656926029958,0.7214281869122949,
 0.9318835568398687,0.22448448392452114,0.6993692523749399,0.7272375196529554,
 0.5411481708360824,0.6728318560956779,0.652861284057658,0.6399119779282465,
 0.862869668315359,0.6551270822857143,0.010853846043504367,
 0.20718633669692132,0.6068587441342091,0.8009959401522845,0.4313584212509264,
 0.09923213429021394,0.9873787283562844,0.3574807992966077,0.657068870347566,
 0.7960614122233537,0.15169973253109892,0.6529380827712545,
 0.26678123114489427,0.5390534546377188,0.838851793053413,0.29022673821005596,
 0.01261176212730486,0.6509527744639912,0.8768495211472611,0.264802812032482,
 0.3833921697546967,0.9518792891199588,0.11258741890664359,0.5939152358524376,
 0.8792657455176719,0.6762931539879092,0.3326831161596868,0.5297831204583383,
 0.06782520561106065,0.5382881523332563,0.8691119648527459,0.5011103035970391,
 0.12417209748280411,0.8329556324365083,0.8593345946712475,0.7725876746358611,
 0.40396540153343796,0.09886270724610835,0.8978656235477801,
 0.41463030863177586,0.4742332580340055,0.5158145713614333,
 0.38659529040392204,0.3278428356088765,0.07524651257307158,
 0.4694026979119381,0.9289034947120424,0.14581784568879297,0.7537017259284857,
 0.00367724452428031,0.40700250422620077,0.8940719526171762,
 0.35738243400158076,0.14376413711578284,0.29101799189615185,
 0.882507370635325,0.47758639909031797,0.7714456314686923,0.7506251221109306,
 0.9838504530248672,0.6052041786796974,0.6259058663949145,0.8927908973787094,
 0.4921918278598766,0.2900219252182934,0.7992282102321218,0.4820242455291104,
 0.8984166467963351,0.17389734331220896,0.35114395711998625,
 0.9771681692598353,0.2669210560125851,0.7230017758437128,0.5587576459947793,
 0.3829457922172892,0.5039013613010999,0.13184047441610314,
 0.11224900488997869,0.14873220240099405,0.26626017521538825,
 0.7072840356236602,0.4362588984322189,0.6675941450073839,0.5239844877648203,
 0.4771605511644339,0.3529679193535472,0.6102083973966733,0.6189366604386238,
 0.4730733201911378,0.18049475962003092,0.8433267934092905,
 0.16558092379751033,0.6325008698986991,0.3866547754301166,
 0.49617583601397075,0.42748256838769394,0.8148984404111276,
 0.12178766057095847,0.9637831589229444,0.5697886371878959,
 0.07838838686830296,0.5916888479823688,0.41420658926489806,
 0.43190847155384426,0.24368279294158246,0.2476677661233655,
 0.8178433688822228,0.712153001279682,0.7201750490327729,0.2747419894852462,
 0.47060125982299883,0.18200979228390257,0.4304692068349437,
 0.8566366454148965,0.46472553563480345,0.35823422206097866,
 0.28208276488880213,0.4504981924050383,0.932809857999447,0.8561637349105233,
 0.07218336495237898,0.3265720492672223,0.9122514674051976,0.5153208821491999,
 0.3307977751334963,0.8221583714960852,0.7989704256991458,0.1402716677264615,
 0.00789498265738764,0.3754472651483429,0.6504724339512601,0.5625407377201753,
 0.6979098814716691,0.04709601659145224,0.3569369043219438,0.3297260947713281,
 0.3543952708199536,0.6800458701547551,0.09449634881163793,0.8226286524401614,
 0.7165025802852817,0.26159815777130146,0.31055648958606596,0.742795847633837,
 0.8357117418588833,0.20600548888062975,0.5238416413694493,0.7869201577929559,
 0.39081061676068196,0.6938986480688093,0.17164544104034318,
 0.6765219651866925,0.9948265592254503,0.6990026478214757,0.23909147137975761,
 0.7011810335540619,0.7384668177710974,0.5875155945041695,0.9949922448410071,
 0.9911475353079818,0.4010987826232235,0.7648582217393063,0.9052007013327383,
 0.14963026700492643,0.633836030040722,0.23426874009081156,0.7714300679122712,
 0.838048297401887,0.26796889733774154,0.28714931273697364,0.8989385695027134,
 0.04681784759644425,0.6738303023070653,0.1946478277389524,
 0.30759759890780713,0.5010831228417489,0.8526141419788104,0.2941821885785403,
 0.25517356489449294,0.43944040389955386,0.11798379221569621,
 0.2936309121468862,0.3099501271575176,0.47642576543116855,
 0.42618620110010996,0.30597089073504324,0.12646990121460766,
 0.04311982251813473,0.012444819019941655,0.20395708330487616,
 0.4527753865059754,0.10894348733513448,0.546292820022982,0.31501490009718625,
 0.2501076893997288,0.11492946398765413,0.06985834519168754,
 0.10456359073804056,0.5482888690550939,0.5501039083719927,0.2816243614347931,
 0.3786979904431149,0.04542656292870195,0.13348049548838636,
 0.5178365911242119,0.3753987117936448,0.1066142739860767,0.12023897014344342,
 0.2154530204434144,0.8113035385425845,0.8032058679283877,0.4232288980213378,
 0.9130670103235103,0.4197103058636036,0.03870631821498627,0.8130140387923637,
 0.652199640151802,0.3399730862960407,0.4185654993305641,0.497803138567679,
 0.006549616149347126,0.6568277726965632,0.9457033004420375,
 0.4818746416212579,0.1124584783585717,0.5436265513499172,0.9154922044692005,
 0.048081897026854614,0.7814645766344696,0.761350184566688,
 0.31147366945687405,0.49026908168115224,0.5336110419812288,
 0.35502012610731326,0.10158602929099414,0.458941839371954,0.355166760332821,
 0.06093506296684503,0.5597540414396999,0.8225250867271992,
 0.21775280357946247,0.2530584695385317,0.7579680026053663,
 0.02416790159779547,0.9857532456443585,0.5163930532759968,0.9528008340269243,
 0.9282318017528298,0.94369866819115,0.5149786719274294,0.8422381408260219,
 0.03870582494077013,0.7114019692375176,0.8874691221149053,0.4843607057263669,
 0.913720046192912,0.7814093457611825,0.21579356525754267,0.16907613479183659,
 0.1117727950611862,0.9885958053090461,0.12992930510675282,0.5041333296514557,
 0.4887281449526072,0.8571924670196364,0.44814413451840995,0.8373496127115089,
 0.5256958758263316,0.170735121581725,0.7305630023824924,0.8040273266911672,
 0.7755606475814899,0.6312320753460777,0.9180483506366784,0.10024473031115566,
 0.6369450128249955,0.31368647016845286,0.034641858563844385,
 0.17381009570755468,0.6968741218353032,0.2656680624978598,0.5955331725316503,
 0.11401529717741032,0.890573346443286,0.2666999098312578,0.4939331405691906,
 0.16782371399668927,0.708755750052765,0.697064625101432,0.837442585955743,
 0.6890955569564359,0.4124292310783575,0.45897928824267675,
 0.38897903154386226,0.8886557462044142,0.8865597209891267,0.5692306942208618,
 0.20006784410852063,0.3353920622651031,0.29399813718518186,
 0.04955808254030469,0.6580855153386105,0.4851004031921595,
 0.15216726537792735,0.8198535605577832,0.3144889599482309,0.5320116972860414,
 0.1608366624407095,0.7951873854386072,0.08151603761551451,0.7380615694732027,
 0.5405733896357285,0.6792619471076375,0.4622395071221669,0.5363114767256633,
 0.537243582373427,0.851230363753824,0.9482358962782496,0.7638373725152543,
 0.8954227686302435,0.8695027374143525,0.8502020635867726,0.638276972515359,
 0.0420541847698761,0.8057520770720906,0.20644027263739118,0.5022535222766685,
 0.942862309774682,0.6934501169639703,0.9710381628302234,0.7274461125465338,
 0.5520072815267718,0.24424341385623294,0.4143693997440554,0.4118949672323048,
 0.9428484350647538,0.23901277411843447,0.975189807578641,0.3429083707423157,
 0.49333597444031674,0.09823911501665261,0.7124205794713505,
 0.20864068003662128,0.775112164586167,0.6441219922595568,0.18547849843903386,
 0.5425726039772761,0.5683641590463997,0.024726101298409797,
 0.46256206910076925,0.45190078944373047,0.07892007984671279,
 0.4065467239042019,0.18487818839512815,0.14975004938615422,
 0.9008305119309805,0.8512518417373758,0.7688941057154168,0.3762950191036232,
 0.3164400168438799,0.9336726132592801,0.6392429952528016,0.24473071153240822,
 0.8182692007300505,0.8836963425510735,0.49552355321704267,0.1272364315590977,
 0.13695683036515882,0.3913809371309056,0.46441630035373616,
 0.5922382659058011,0.663300423051897,0.9863865731389314,0.8896097567434866,
 0.4242121086337768,0.4859429895976981,0.9204033067065577,0.7504232164593353,
 0.8045201009490069,0.08273948458349323,0.5345796270464084,0.5744233680244871,
 0.11136465482867008,0.7354772081654751,0.5079124944719451,
 0.34186694965270215,0.4707042779241005,0.7391542173639895,0.5614200807812508,
 0.8335821392978392,0.6793879382357254,0.46913578228530073,0.7713580791093639,
 0.7439384449581543,0.9959108472183281,0.10963266215799428,0.2890875535849111,
 0.03211197827542478,0.021273153029157887,0.17149994774069965,
 0.22494531268961282,0.7451095947517679,0.5701287079875199,
 0.047158061183724564,0.3637529757703194,0.5201731320356295,
 0.8844652643187931,0.1175168687891468,0.948004908836329,0.8373180140684745,
 0.8754429880590001,0.8809370122143476,0.6553954900766701,0.06620007200277345,
 0.5993531406411006,0.37500597835459615,0.5488764716112803,0.519785491817666,
 0.6922110226467791,0.5340728354702732,0.32967277322128563,
 0.28807376731459633,0.87965980321941,0.3522907169131839,0.3942031162118119,
 0.7988388788713401,0.09317271375880476,0.2848556136040149,0.5479584100544095,
 0.972272769303194,0.06550723246039436,0.2268381566700416,0.23329657909738066,
 0.35085768346547797,0.4006658016504516,0.5103988691564062,0.8954296932700863,
 0.8223852187812503,0.3975841072055766,0.035306699068446634,0.561682232130703,
 0.25829538033165145,0.10890243962573654,0.7714645834619328,
 0.6878722425379272,0.9479642037315437,0.6369529754009776,0.3605857603115429,
 0.07029907444530947,0.9064259644121002,0.9894737353291263,0.5759556392893647,
 0.059426817642947105,0.6410426132347015,0.7430815787704615,
 0.3513866745392703,0.07638911559010775,0.5955288248867152,0.7886778304227506,
 0.142595100744207,0.9862757669626849,0.911537291834247,0.3144969659756405,
 0.8437925492892357,0.45977815020244195,0.32243025263114944,
 0.023942399569705453,0.7807045331313086,0.2901046654791568,
 0.5437726114467956,0.9652306718009435,0.8322137352111745,0.4240143777991373,
 0.8061936794419945,0.25870243464512044,0.726251496050726,0.06086155824225026,
 0.3496507767807854,0.7410493200140569,0.049684020364044956,
 0.4261887591937601,0.9248345654270798,0.9261695386503295,0.2592369960644014,
 0.13094663074594337,0.9639152884787356,0.16186944225650446,
 0.15858187127481127,0.48682197049144693,0.7343310991219967,
 0.7271199583096613,0.09531433426969271,0.010400569509422253,
 0.8580904805641075,0.5010868199340813,0.8677250449403169,0.44993391852904785,
 0.15397838788797813,0.010932259637586816,0.7060020881730918,
 0.4901988964466272,0.5282820630345673,0.8624785102045245,0.9340067651776144,
 0.6617949254372797,0.21825364364825417,0.11607080045880835,
 0.16214205143957028,0.0444065942522095,0.4705044630515426,
 0.009176620188017348,0.9659199810185332,0.8699744742123938,
 0.6651287569928407,0.02453344353014475,0.16209183800668625,
 0.6510055066903058,0.02346734410354845,0.9511642710094041,0.7105736866576755,
 0.13996218239420666,0.41607226224978255,0.5749600993915808,
 0.4765909616835655,0.6832626210599663,0.6204752606179361,0.11330278296195928,
 0.08437977248757256,0.2786556892419275,0.8594764561293281,0.5204635411059969,
 0.7551881325725461,0.5003156471551489,0.7544285746257423,
 0.021638391333191453,0.3015125316412628,0.8047898178305942,
 0.9335924800775354,0.6914367117432726,0.5794774589998387,0.9879825661981924,
 0.5436963966717919,0.6159659693691868,0.7530854367507888,0.08724037173219779,
 0.7413471353897461,0.3265036165340873,0.424491038049295,0.3586274178005926,
 0.5283508842927267,0.7393206615768997,0.4946966607573037,0.40114425439573476,
 0.24132006293763564,0.6275734578405918,0.6109945043000927,0.5457713298404057,
 0.2620949019764416,0.08907766062139155,0.9449452595995226,0.5580171527181452,
 0.70421751430916,0.15308052361192204,0.3754190977897385,0.6247626175525576,
 0.5806962768203623,0.9181204272294735,0.5354487814188339,0.08818619455556143,
 0.007620775920329614,0.4418698880351607,0.4132050156874584,
 0.5403785811650371,0.14761898473595358,0.5781122954824045,0.9296829111903723,
 0.7769995399835161,0.23848711804789946,0.7087608943992711,
 0.15541995770894546,0.3732947831898613,0.9672263463580177,0.9376148155992643,
 0.7315617706974114,0.14520094927556226,0.8404227449735897,0.609807919095561,
 0.6087572170121266,0.4888376050351041,0.12608764037025832,
 0.05160390034309126,0.06405516856407523,0.9170595640954723,
 0.11298783650995947,0.6073582361143478,0.5368676632290059,
 0.34373279864052997,0.5966705206832401,0.7385808010167414,0.369471046704279,
 0.08479220588148262,0.6017355014374313,0.8125998121375624,0.3076874636190947,
 0.28285873110331117,0.22261162441632143,0.7497189984380749,
 0.1626139696912443,0.9242600268592884,0.785822202069041,0.307910420601781,
 0.9391712264431019,0.6924897611801668,0.5662471270493545,0.12076995671846702,
 0.9139398293795915,0.07125229642203679,0.24266485285203399,
 0.4464072829704204,0.531127503757195,0.44244729317104414,0.08473101706633857,
 0.9331284649933035,0.32597127380657387,0.45314254499494167,
 0.7479277250779914,0.03472367945259991,0.27087506091560565,
 0.6154808556038038,0.9967736672425717,0.0060026226259896065,
 0.25830637685269775,0.04177103752683364,0.1654428922254656,
 0.9778798882110942,0.7492777048649987,0.10306689866409058,
 0.09581645917086234,0.8584422834312192,0.3202398663414957,
 0.08328427263426963,0.06724106969269417,0.7192340413953814,
 0.4357028528711475,0.7855332343892792,0.4127201620711385,0.22980312570434114,
 0.08894592919161404,0.047090148594913894,0.7204259297247257,
 0.9478137191771892,0.8745550255080804,0.7505871767617457,0.35341778182217687,
 0.930273333274469,0.7473381276124662,0.7287911244744316,0.9779206270278341,
 0.661953101229841,0.6081541753721309,0.48605443169928286,0.473616613354249,
 0.6314215488083175,0.27124040838549246,0.41448985699916685,
 0.6155355129433764,0.9128545124994557,0.9792600141847458,0.8460349965968768,
 0.8003849886703379,0.5364383070917322,0.8667627725613813,0.9881495218103303,
 0.5070664725385742,0.3446919940214117,0.9446191347040794,0.6169744552274686,
 0.2881206066105316,0.03322400946641346,0.8058583413339276,0.4708412231852144,
 0.5921568210282019,0.41074351879289484,0.32152885907578066,
 0.6944703294374143,0.10347310535374221,0.7579187015143053,0.6063607662171608,
 0.8517637390050009,0.4254638046904963,0.7826330026701966,0.8161942266152044,
 0.8854686370896906,0.29294492926710036,0.6129853412917686,
 0.11675486598656448,0.7825990000177069,0.5932363669140676,0.4916732215948739,
 0.2520241231402419,0.5496971843219829,0.37786819755134027,
 0.43542876433748356,0.3123437028955448,0.7505072069127791,
 0.04376320249412635,0.273166435525356,0.8270222034558455,0.15866679911737946,
 0.05649357903323937,0.7056010762574662,0.20284569923561135,
 0.9118228861419274,0.08624905797108273,0.9595444098124285,0.7296767845439663,
 0.3116303766027907,0.10950696677680971,0.358043145396421,0.22635290444089118,
 0.35507260599214874,0.09602080701223403,0.548539667891559,0.4628713707750768,
 0.465158633059241,0.4496533901436597,0.38721525882973185,0.6832131810647832,
 0.766821599950996,0.1624129979453205,0.6187514893613049,0.08285917216584315,
 0.45929070377017167,0.3985366451479152,0.4599788355428779,
 0.24601889444751934,0.9878846279206942,0.26826927077909646,
 0.32195312078998717,0.3827891939629854,0.4718260377115586,
 0.16842624487971292,0.5213581834683719,0.1612380704417158,0.5501569800321502,
 0.9484882090205676,0.5057623127910101,0.1340334148497322,0.7718378310133105,
 0.36256182606074594,0.1369479492162694,0.1876668753364541,0.8783822175936624,
 0.0962920369321939,0.2738806461305733,0.004620270772701662,
 0.06412695836504045,0.027593591325978606,0.40515020090747245,
 0.4309993009635448,0.49465208472160493,0.31227653007030076,0.799344573233963,
 0.29232393997922546,0.3977274964217894,0.4205001296531349,0.9668301169995687,
 0.25190839607737725,0.5858266649461765,0.6054808975264273,0.9376156147189677,
 0.412357997263896,0.6663105242526439,0.32844804807765127,0.4677938065696622,
 0.37209566149343276,0.3261690350469525,0.040321920371931164,
 0.5914005513115139,0.5685873740242201,0.6435498070771704,0.8912282616783256,
 0.05458929892769682,0.24699222537009669,0.6987704792480111,
 0.6706786257355001,0.40167217890583884,0.2372261639697275,0.8851211238768584,
 0.9868613389686753,0.14313617778972532,0.08232356730860912,
 0.2757423352785199,0.29949575234420667,0.10368306270113548,
 0.6368013197665089,0.920582595042995,0.5791945938632146,0.8159095364054122,
 0.2191467017697697,0.20703443655602305,0.3850741984439471,0.7658121336358957,
 0.5017214244564607,0.09562762515808199,0.9968299812175736,0.0527440885621171,
 0.695492560988466,0.8518077641308384,0.11190857676571675,0.6572427743263966,
 0.9469090100078319,0.38180960415028875,0.4763028145522644,
 0.03190630204303635,0.6918920089004543,0.7992784427943758,
 0.25213147726065666,0.5665933152074646,0.6702057467705204,
 0.18696556246426432,0.2295660433503699,0.2697986376253694,
 0.46650837694470704,0.05449332089155301,0.7902239068221707,
 0.9189051572228296,0.05600371473619825,0.2071286695813932,
 0.07160397904978733,0.5453249275822438,0.8017019747809561,0.9023392809226343,
 0.11731326386278429,0.15472517817788112,0.5585865064087114,
 0.40167541762462067,0.21936137406269807,0.9984652191273563,
 0.33500118528853384,0.5904622694589302,0.32859970075527445,
 0.6475526766798546,0.47785561734400483,0.5417873005950188,0.5487920331666853,
 0.38328034937823663,0.6023861313478508,0.11683000870103344,
 0.9093201724456743,0.4799977522598948,0.5133898205139399,0.4753284055740573,
 0.036275333326971304,0.8834072274567393,0.9493475907684767,
 0.4711181840945471,0.37992811915608404,0.08285684409123262,
 0.40772263033780753,0.5672036285400974,0.6475179389432807,0.5194762053304044,
 0.942989741081865,0.33266735775502276,0.673973564294474,0.7745954261298692,
 0.02123908424318366,0.26854469151583404,0.7709430362012593,
 0.9630371031363947,0.6306526765797033,0.30501098497898915,
 0.21830695015424784,0.997358877904328,0.5777071646097544,0.24468192193894134,
 0.9565903026846939,0.4853611603840847,0.8831693651496189,0.5515658616975205,
 0.6855612498899076,0.5006102064341168,0.2494735637320339,0.2868304486917027,
 0.34897130361166195,0.7246248141790717,0.3914533773369332,
 0.31677796667031444,0.17564044171400417,0.5848061882728407,
 0.45897731095004557,0.8457518963956052,0.1656201395954512,0.6561449934216304,
 0.3889358111749125,0.36781484318270474,0.5564645585284433,0.5498253661668925,
 0.5841349949567806,0.46171890204665644,0.17672421936593907,
 0.6370639213675059,0.7011127456672324,0.8024633166752039,0.5850774986810714,
 0.14286679175594474,0.1418043150652648,0.45156791931509044,
 0.8888679159213944,0.3941196877486327,0.05470816464199835,
 0.21913411942385486,0.8258025304937477,0.049736337927308805,
 0.23749742385790495,0.19069572104273624,0.28347613786307146,
 0.16028185655483274,0.11804307365394295,0.4259685492733264,
 0.9112207466799822,0.07372938429195341,0.6503598957791983,0.2278026427751878,
 0.38175401351417215,0.8188961968328876,0.966174635321539,0.6734471081333712,
 0.45812599048735037,0.7213034595361562,0.9247394912291866,0.5752582660352975,
 0.270185363240397,0.4260831676126613,0.1043200643427612,0.3006369966755361,
 0.6492736549923526,0.36346116176618803,0.9885962734724262,
 0.17170139612882074,0.6511430061713928,0.39324617756307345,
 0.4664029748867552,0.3241269775966368,0.24463500551055428,0.6312615844673167,
 0.2743141088475912,0.048110342810989803,0.17027192474916775,
 0.32605026690171035,0.014307715476658212,0.3852619011989956,
 0.541537162833603,0.4861392062401213,0.40620368611256663,0.8514005870476438,
 0.13690720954050783,0.14416488370360603,0.7169567903592504,0.544934565635547,
 0.6485499432368641,0.49432729981436485,0.031080643456413037,0.86618748231733,
 0.6046896446594051,0.46582744260951636,0.23775932953788337,
 0.5846917426342174,0.06747349445650097,0.5173820143425661,0.7833430258062595,
 0.9220057627192553,0.24806731377295166,0.2409557838873434,0.6282978569657394,
 0.4480954800137338,0.33997329668045684,0.36500472234387604,
 0.6571564982917388,0.3147158748311675,0.6110016178335678,0.3609456736353177,
 0.9357771047522171,0.37198484058064707,0.8982871226715711,
 0.30439818704904076,0.6386038248224618,0.9305099082278885,0.8751253396693304,
 0.5444307460211025,0.6966509449682089,0.8320994421091806,0.8336225633533743,
 0.6346731660585994,0.5733040875106499,0.692948496241105,0.37615481169703013,
 0.024433773785153656,0.8726329032377519,0.3826293957469674,
 0.4604316381377789,0.0945746957410446,0.8133563979844033,
 0.013360492652257916,0.4382520586033638,0.22951118345834887,
 0.31444348004709033,0.7023057645505711,0.37063783512365156,
 0.2667863582434966,0.9171659826429026,0.8032083750154118,0.6104187909008221,
 0.8156615149708534,0.3482451835412419,0.12171051718582715,0.8401948565380799,
 0.9321738986441304,0.28561530494875675,0.36951381766413904,
 0.7773026169867869,0.35554842168339873,0.18685242835186688,
 0.5168254826574904,0.89763506599406,0.6764417501875833,0.003199460313547897,
 0.7606713574318448,0.302903773565243,0.36108763632493623,0.3315937408999756,
 0.16541760760336932,0.50598771381593,0.3574895644597955,0.9358393074066733,
 0.012116537384789239,0.6093658739955382,0.18117353048869023,
 0.6973708499360033,0.9733899712624215,0.8462355722745419,0.02090746630297191,
 0.9560212353915933,0.240269095837218,0.019296679678960982,0.8655429994042872,
 0.6935778895456205,0.3303163686370031,0.011993432784843105,
 0.46869929457900583,0.06572194933810926,0.2317353734077353,
 0.4710778453927378,0.5018028100638801,0.8109447763025186,0.5872137194128352,
 0.804619342763578,0.8262750099966125,0.899652842747727,0.9730190570293462,
 0.1590784724816826,0.7557279114794785,0.22039199314997338,0.8452885014964678,
 0.06148139992138291,0.5259839233309946,0.5207247245368943,
 0.37870664437716306,0.6166780309547838,0.9299906557440719,0.7782708787914736,
 0.5034860296293245,0.036993553798568946,0.6521202558794637,
 0.7780161105530813,0.02767372126751355,0.3498921457644808,
 0.26501118433352633,0.9641862814377341,0.39225688588884955,
 0.19590660268452742,0.7902246180604349,0.7742949593621169,0.1257525205056783,
 0.4595171801702572,0.3372123421413379,0.8100336813365083,0.8497199655844949,
 0.8079388970165453,0.8513053315695294,0.4895113211300691,0.14443302255133605,
 0.47769793150642526,0.0853301929326633,0.273218918257666,0.43853312674545863,
 0.5527611655348175,0.2546183865028183,0.09423145085637608,0.7708236619383587,
 0.7611904332577715,0.6977606552321968,0.6937979166603176,0.8079554592864007,
 0.9985100155551406,0.20520977658394302,0.4594482131651386,
 0.49160213759378857,0.8869347126203835,0.38333945485492205,
 0.7921103655605585,0.0772361736464574,0.8277188191908758,0.1877622330070291,
 0.8041267402575724,0.45447423933070075,0.6743060152137348,0.7707496844487915,
 0.34961910193733636,0.8492242415532983,0.3462316621103858,
 0.34428227601145367,0.8021561942974943,0.270059583310037,0.6739767045568501,
 0.33469793596667796,0.5233116113842143,0.6424177665804072,0.6932533924186063,
 0.3454259104020393,0.40332975170186647,0.16553451198859992,
 0.9174371256017467,0.6712854659264452,0.07466512726223684,0.9239203815418734,
 0.11739444841452096,0.0707195274249176,0.23762535840293253,
 0.6845159876273419,0.25983693300935284,0.2511787641323745,
 0.36732191551604176,0.8031607269482823,0.014927716051788775,
 0.086098405105826,0.5475515296488591,0.3617108411817933,0.6430140700536576,
 0.49570548919765667,0.10067226770733406,0.1826380729155318,
 0.9970308590914153,0.16455077058665168,0.2154368224921055,0.5213813350818619,
 0.44262641582848106,0.7214442788392298,0.5763788343107453,0.9482710160550458,
 0.13533569758613184,0.9413369181974897,0.5943834819369059,
 0.35930430624894205,0.23802225737671623,0.7831669915763163,
 0.2781298700279331,0.27234098137366847,0.8693516147093046,0.3683898918324795,
 0.4821040275652124,0.01792107532000997,0.05361928170262409,0.993096759131917,
 0.5747063286605627,0.1415841742819729,0.6393867353613814,0.7670156962653173,
 0.2826328475657448,0.9099580698802283,0.09790063264818727,0.4585371351364085,
 0.04697013800083949,0.2783873673191817,0.14745231958789107,0.544724600969966,
 0.7073549414243533,0.8913115999972381,0.13471863876656853,
 0.002343189481068486,0.8838448426091969,0.2429274081029682,0.602844643274786,
 0.0733847102660633,0.04870422630678273,0.513441092177001,0.24417764366671346,
 0.704282528530814,0.48151029407627344,0.597973749110276,0.6979457696010744,
 0.2829700934276683,0.10436374195015496,0.277650995302675,0.24582004267366642,
 0.18166424465912778,0.20637990975115994,0.9303455032202483,
 0.08580195581566907,0.5020758136013082,0.20984841919098984,
 0.9804559221861879,0.09230072631649011,0.2747515687977773,0.7028784624383633,
 0.09076609377207157,0.1465726315275545,0.9311302277490026,
 0.06986456574608291,0.564322511957079,0.0026683989729641144,
 0.11876091003946743,0.9389757796695424,0.4295739053845632,0.6445976028274216,
 0.830032940267631,0.6302907637082265,0.8007992084486033,0.43094729341916993,
 0.8067692927444012,0.8727014469830356,0.07793098073905258,0.7541260050066566,
 0.8038526720968255,0.2975986591959886,0.9340767288444097,0.500097877691655,
 0.12482067512544726,0.42027100399834916,0.43654039575366854,
 0.25648543219807196,0.22141334728332707,0.24272320121571211,
 0.6429834601764695,0.5163076989072071,0.9774100236157707,0.1650497212186408,
 0.6180884478481085,0.7224590638007886,0.07678425025856439,0.969064789311549,
 0.07232545775213473,0.07817438859600123,0.4186836123932285,
 0.4386181038111291,0.7006697181272274,0.9860412880563167,0.9828729575631335,
 0.9842220967273976,0.11019027373013501,0.7104201331520429,
 0.00591883675536975,0.7860021380407679,0.5993661873080117,
 0.19105212087885237,0.2030955019502294,0.9414446064919678,0.9222587363242292,
 0.43181581158905114,0.9259952185241284,0.9903090984532208,
 0.015697208666159868,0.6017998196433547,0.31629376222717775,
 0.9393184983622266,0.5310541186665849,0.7736091270732454,0.8276074372312074,
 0.7784968661638549,0.8197890473325228,0.2654902829593859,0.6584244305551596,
 0.46950404360982534,0.9357680519673066,0.9572451797443993,0.9282381055216005,
 0.1849581725828261,0.8527940599437167,0.6168512529789475,0.17617543451267825,
 0.6902280803402628,0.15056046027577408,0.9104551012945128,0.5110202461579078,
 0.3931739081962823,0.15912415158602566,0.9050670847562399,0.8200791337038522,
 0.7304680346340162,0.002183710073306866,0.10148923920870268,
 0.6977106336050982,0.39610274133778844,0.07511098659759996,
 0.1423750850785045,0.1009754145939954,0.21294614734368122,
 0.12755743840120515,0.8669635785137229,0.1331883084535076,
 0.36671044064921676,0.6353046743833058,0.22760447326563382,
 0.9487774631853867,0.3152393391095535,0.6465788622509161,0.21147680679686165,
 0.9109686612834715,0.7546584777909549,0.855910236467832,0.20903337212991868,
 0.24748693446784198,0.42494401986268837,0.24239177891670283,
 0.9966855516821788,0.7417141794845871,0.6251800434012671,0.9491088549946226,
 0.7046548277751775,0.6427503497080533,0.44015644999603853,
 0.20469576140142876,0.6575810072966232,0.21514815236705276,
 0.7972937679542489,0.8238034873590891,0.4989191462388741,0.8470126753496591,
 0.5884402802639925,0.7046279206607452,0.8639626339587577,0.23905346051739207,
 0.19614801297213968,0.9119995779108988,0.1381139543790253,0.9443887566604215,
 0.24729252464465767,0.6519268024427293,0.5505828125746449,
 0.09884352700018695,0.5086448048505117,0.16288501084038742,
 0.8889302886597379,0.3131240853698759,0.08195058129695054,0.5614890089441553,
 0.3601472053282351,0.6746554426817903,0.7155591635339059,0.6203741915183696,
 0.7808303370488602,0.09626085329332867,0.32839416093696194,
 0.05352337538117713,0.545820391519173,0.7183701040587743,0.6791124363173298,
 0.3041413177220956,0.24021567146206435,0.10109231091508453,0.947262748643611,
 0.9195924427921397,0.9159789686224079,0.1553751744510583,0.20490235070124907,
 0.9370511548664634,0.9045389499800061,0.4627234024685547,0.4174181515150337,
 0.4152317555448436,0.9841105718926744,0.260175201420891,0.6271386304287991,
 0.5823716947796119,0.7709966630632651,0.4371884426781558,0.8771451334635815,
 0.18264276341162622,0.13540312509667574,0.948291676711121,
 0.07643197077626951,0.02705241316475815,0.41945701272129265,
 0.4247367601759151,0.12876718557409061,0.3669567411112511,0.9688541488777114,
 0.7119015234809439,0.6710259070195086,0.9111776529182584,0.306923738610002,
 0.9181323812114144,0.6982123112573335,0.8782976937795081,0.3402377351671414,
 0.9944796907390943,0.0445041965648445,0.7882927006327612,0.33446826354867587,
 0.13553651450918958,0.5342866487396875,0.802411116933142,0.6218672758219514,
 0.34125826274822113,0.3773128907303397,0.27852211817696904,
 0.9492852279943108,0.7911520609545093,0.9381831932031933,0.31945529471295864,
 0.9188081475120031,0.9914670680253819,0.24776010973767215,0.7115212568466881,
 0.5801210378587833,0.8395401005358899,0.7315013411047845,0.07804325617692753,
 0.7476932617556674,0.6690035092292663,0.06528253026526176,0.4097997484213941,
 0.004549858822835984,0.3322306297561324,0.411189221871203,0.5226880411297294,
 0.4797974064283146,0.29305692153209717,0.09173653864449682,
 0.2855358967566951,0.6357389310513465,0.9057684802218464,0.8451972926992571,
 0.7414052793054895,0.6129264432157732,0.48826678307501636,0.5122016406729987,
 0.5321229382744949,0.945135945128797,0.6947351144218537,0.1339491238693361,
 0.04913984916299918,0.09843131739385169,0.7045574568574992,
 0.2644965024005561,0.9685718657536874,0.15743487829723268,0.8111965434584336,
 0.3656666810102781,0.1672153370315922,0.32716266963581886,
 0.16113603267838572,0.3787263343281273,0.44434531083858597,
 0.03485231273457157,0.10080105360711233,0.8003766707065858,
 0.17957997599347775,0.8956886335491429,0.48691352641507235,
 0.7432151811847233,0.7467373294400313,0.4975577828722604,0.9331224940250739,
 0.6870484478843669,0.5061213177304407,0.5053026286453309,0.0886149656768096,
 0.27324211177874225,0.4205420639074786,0.5234828033259998,
 0.050293426465445545,0.7207714354261299,0.07688633991970428,
 0.375417048542966,0.9091883459094736,0.5399190945518942,0.44043425775945877,
 0.1466906864100368,0.21888139107303795,0.3469307083448211,
 0.40064972404820287,0.7459065030234971,0.12802129935914563,
 0.2672878665634907,0.248688117617608,0.47067443114792773,0.15512961202981745,
 0.7406637266627815,0.5489613638279144,0.36984891761607397,0.9962068460360651,
 0.9484813046898333,0.35532929987689266,0.13758419275576295,
 0.8010476781652546,0.7689437053523576,0.8722836288873748,0.44684201971109094,
 0.942129279329911,0.0969483478710439,0.8363353749722514,0.6070374999505994,
 0.4889637295354132,0.703315904127032,0.8085428557178397,0.7327573179772654,
 0.21197883946717266,0.10530616525809677,0.7089944372176366,
 0.3103971072247833,0.8424072899920818,0.0277763129761901,0.8340795666733775,
 0.9205248125688826,0.25761428485337867,0.22182489733506272,
 0.6959425670675651,1.8500842303204035e-4,0.028791320529679365,
 0.5355235236342593,0.03314351759563816,0.5284242586373731,
 0.15021498786987864,0.7372818493122988,0.3839476651633592,0.9702405975251468,
 0.4358829837467335,0.9128620678786975,0.004136541436820851,
 0.20949468999419685,0.5439981184921372,0.4660280036752148,0.362463705668708,
 0.638845110976075,0.01826774373855866,0.43811214111288566,
 0.47898224268030554,0.00636129762778892,0.8086019631229253,
 0.5021828505031872,0.4216776155932608,0.8068485546806292,0.04456701490193016,
 0.522445209141562,0.44582472265540096,0.9488204290222784,0.7680116836057208,
 0.2685000092328026,0.21329423145355864,0.05636209508672685,
 0.8167208218076342,0.3836423149889958,0.994913026995174,0.9168025260124637,
 0.4191818097255823,0.24323191055725835,0.47773079335289403,
 0.5173769688365909,0.9454052489650151,0.7127542706516014,0.18010790084559725,
 0.9940663962785078,0.06318411884357189,0.22271560781395217,
 0.6477081377315734,0.7989440213788876,0.5149618759885891,0.1601507225277191,
 0.4037877162948207,0.6644450576803357,0.5385693605066546,
 0.0036798139707714395,0.7029652319958416,0.944091522973158,
 0.9437387086593875,0.3031380780551358,0.9309262906385918,0.5357882614300317,
 0.8447930256962743,0.10825587667372338,0.8426345580592586,0.720754344936986,
 0.8147955323924188,0.49940058422674594,0.6577948274499442,0.9782859909411208,
 0.25781235621095844,0.9367939009050357,0.3586803369446615,0.7400886140703241,
 0.7507135343506272,0.4935060429854059,0.9735736590761466,0.4700490048972207,
 0.805226514312559,0.3464730413106025,0.703398780474906,0.9988010801655337,
 0.2760257573510808,0.818727623563656,0.6278881417784875,0.15030447620046028,
 0.8918890235382327,0.47532684989917695,0.45676590356208113,
 0.5745164436122852,0.7681868269867287,0.1087844040872683,0.5688746779384597,
 0.9869528550203315,0.33758309552316057,0.016411579008411503,
 0.20965334914657885,0.7720106648507616,0.4102610830355581,0.948362706933893,
 0.8303862005609243,0.6960523411097207,0.26313129441198724,0.1305080426910209,
 0.2045705484326007,0.29709333763363655,0.6295949723525189,
 0.12266998335875734,0.45523607237479324,0.7972495526735388,
 0.6814338633071926,0.14625908076657745,0.43800422669621275,
 0.09370789576894833,0.41569564182800933,0.2990551192723625,
 0.5461979892083574,0.4647501170853382,0.8753331546909917,0.5632930092849878,
 0.7147732488806025,0.16984212551474642,0.342101953136844,0.7163191822427009,
 0.9667738607048132,0.33374177098144975,0.24280835962671854,
 0.5819716286070455,0.25125854348495047,0.9050523734088805,
 0.46586758728668143,0.025182387398671513,0.8492129649673792,
 0.056339483267192736,0.7317226451238165,0.36647939421002995,
 0.6755145166699229,0.7961903923146089,0.04554555013788242,0.7527765357858553,
 0.4375357185127,0.5627023041040466,0.5390153671896005,0.5361939538870318,
 0.39434214948063284,0.847630987517476,0.5946103398687853,0.10087664135545993,
 0.3648860788413829,0.30032984816224106,0.9545125716080523,0.7178440909536772,
 0.33348124169208004,0.23384943362356503,0.9856700034114346,
 0.8187220116368369,0.2811759836377813,0.4458386631854365,0.9658305412693696,
 0.6701296961688266,0.31890243690284614,0.33437847617339633,
 0.5868228144194125,0.6218361750754555,0.4737381357389302,0.14662607640133496,
 0.8450609316736195,0.6616555392137203,0.7789683899612125,0.7189534442084781,
 0.5750031255399719,0.2717531079021325,0.5552127192428492,0.5895122197721028,
 0.09543446378735354,0.18194524541710932,0.777043626378553,0.9049932182925532,
 0.13199138043577952,0.7857456565502872,0.9182516798977581,0.7072562538151865,
 0.8191117957735697,0.25907900369386083,0.6929049474813741,
 0.19877534302308808,0.19689174167301693,0.7598833768849931,0.823133229939383,
 0.20433871782819257,0.8558495210759542,0.28455966137484534,
 0.7924918152307019,0.04721280922785298,0.8927872924183131,0.396528752484793,
 0.07268425075883711,0.1493337761767396,0.5038362646184151,0.8818971925082161,
 0.46628463717957636,0.1669603332683398,0.6394675736106924,0.441525897175296,
 0.09284471080121037,0.30849397509212007,0.5000542596429195,
 0.8985575163619162,0.591443896230228,0.001204014907305928,0.7277415871496125,
 0.9026666248155433,0.1291660971362003,0.9535848377132089,0.5606791947754368,
 0.13651582826314335,0.8966390523698399,0.7067260346846438,
 0.11568675913863014,0.4443864093047265,0.12978228214814314,0.630759310360174,
 0.7493875520783513,0.751769521411715,0.3410407808441336,0.6179977269503893,
 0.6008822161814497,0.41343590158727694,0.8187289160666384,0.9714810330225778,
 0.5558580972183045,0.04431337562338,0.5607111554505397,0.030864127304826905,
 0.5768659609857855,0.8372362242891231,0.523463735999585,0.7981155885739585,
 0.2607752565820758,0.20223841059988068,0.7197932447113005,
 0.18493539206014498,0.09038841848117174,0.330890686409802,
 0.12473837181640857,0.6463184839629379,0.40828596179633814,
 0.6948159320502889,0.6596390888456044,0.17543177074267313,0.6731635504581902,
 0.0759276743120404,0.6931106046239426,0.6017099550992041,0.05409595856496818,
 0.21974264499244256,0.4134022003958132,0.7874468752279408,0.6926301204799549,
 0.4470652181290006,0.7960327666229708,0.02554035149342515,0.6980525522173271,
 0.6644546787356428,0.08341906482109995,0.35070075747578033,
 0.3447186896365969,0.4961960198675254,0.0843174656426805,0.7994983205425472,
 0.4911944988403434,0.06495376647966822,0.6513233367246907,
 0.41542789673409863,0.6927481381555789,0.9565471154642705,
 0.24560467693139854,0.5853656185350571,0.6587816645003582,0.0520430005287571,
 0.33637705890019287,0.4371226341299449,0.4989274445311124,0.5820405069742352,
 0.5485572657727418,0.36841752376739345,0.5014026188382963,
 0.39899905869778496,0.7520649628459809,0.03231109938593968,
 0.48954318327661284,0.09015172030497176,0.31505467866491477,
 0.14932092902618133,0.08891772843572576,0.15482311549185868,
 0.520204342764612,0.7368338570771741,0.9352071017752487,0.16691076916326364,
 0.08434469823359003,0.5666751001915034,0.3435104156537713,
 0.16421766372968372,0.38999326192650663,0.989648872025315,0.744419327785343,
 0.9662323734932865,0.6725108688655568,0.7204463515061144,0.5879208748311208,
 0.8627090187697274,0.5934474948369608,0.5329855930517081,0.876521999558417,
 0.9203416197829711,0.4405559871438165,0.5489898768396902,0.15699917288395693,
 0.5462732359706194,0.32521332353285126,0.6897249422279247,0.5837661787418714,
 0.10714030562135379,0.14283277319119914,0.34168327562128786,0.84895194384057,
 0.9982633791077259,0.0712455105106683,0.3656411759833429,0.5412586777871046,
 0.37321805367193617,0.9087793072126433,0.36206554388092926,
 0.2765661973482605,0.04895988965222853,0.08530678432720928,
 0.29006938111608616,0.44383657754633876,0.16470131133904797,
 0.8058541294470151,0.5675484261836742,0.773304712781506,0.8082650753465142,
 0.23982127514065066,0.8306871535821871,0.8818492242535749,0.5728939077437568,
 0.7666081254986836,0.08523865027376798,0.9756736775290058,0.5027273190776878,
 0.12244424227769457,0.5349972024007531,0.0747029171378999,0.1935038121212007,
 0.8109368746408896,0.9697004958844083,0.7679566927776124,0.9357831021546114,
 0.09672826835830184,0.7510780548204188,0.7812222802001971,
 0.09868146241760123,0.9863653315979088,0.2191131187526124,
 0.24415490819001961,0.08161565016087036,0.39766130536221234,
 0.32640527201147906,0.5998880307350601,0.2611534266850549,0.5206062358161413,
 0.2857085453012521,0.18009006333400157,0.17998522818753704,
 0.9544381701009644,0.5394675772981024,0.5569477381828952,0.9121153800620629,
 0.01676659626351107,0.601154799146287,0.2298914408468128,0.9778974165217535,
 0.7679192661915945,0.2622832210397811,0.1718716861789965,0.23956398069908558,
 0.6456789440470507,0.449020099457062,0.9599569926548452,0.6754228908676122,
 0.9296304549771631,0.08650560895036374,0.6015075608980766,0.6580191148128173,
 0.9137503690240403,0.6663458282214784,0.7081204038497337,0.09350718214145248,
 0.43571614023794136,0.7260491986354102,0.7405093119181536,0.850112800074451,
 0.301836606553164,0.6886037298595371,0.030210793202452013,0.6883176530450628,
 0.09704248454065012,0.2496725722633043,0.8051506491854968,0.6714187915620756,
 0.03034415808894142,0.5286644077731639,0.29308626128138127,
 0.3768887704551588,0.5014297082717358,0.6289830118839252,0.957703396318173,
 0.3755439130619991,0.20093451571786114,0.3743219786243047,0.3564607970677509,
 0.6080195591032692,0.7314197720418935,0.9679057849887212,0.8044619943180468,
 0.25362929847830296,0.3404485079566416,0.3621068746809353,0.5920393604488537,
 0.7229147217038367,0.4301772951365048,0.3409544156346038,0.7955607024999797,
 0.18890926255712814,0.7134189329312379,0.43645184707915796,
 0.6791281448852631,0.2831682903489816,0.983753874270772,0.13188799442317545,
 0.9096105064476684,0.9880297881690172,0.42612984514983476,0.63287553215464,
 0.04538507890855803,0.8057740966900877,0.7569316265482966,0.6360754345498584,
 0.8029179830174118,0.10536575092998957,0.007986849862289258,
 0.6757149150160793,0.14651087205072955,0.27161724249860575,
 0.8433263627358408,0.025146935455746422,0.29383982849083967,
 0.22652114927058697,0.4340848134348949,0.4890644346120421,0.2177013581762487,
 0.8875470084338474,0.40776062246731093,0.40389800460055514,
 0.31601787073366006,0.32504148261767096,0.5870701165696363,
 0.08772476402796958,0.2758878078541861,0.35155421884237403,
 0.4088265835566913,0.2977381819131505,0.8090570111156627,0.9729554694119749,
 0.45910196127088276,0.8892091032642084,0.4710936972215145,0.5507540631269181,
 0.7295596091196972,0.4972437931142688,0.5269515543884512,0.13914178665066135,
 0.12966133594382345,0.7793732633476397,0.9303006893950447,0.7570277382776261,
 0.6810496827427948,0.5592202467559063,0.7821808015938694,0.42990324205780817,
 0.0647410962613324,0.062906690114377,0.3370830114153295,0.7273734274909299,
 0.1996378258077416,0.11371493622248496,0.5837215318695402,0.7482087764860801,
 0.9848862213440527,0.9034901649256245,0.05698984783556982,0.2401049293231936,
 0.13710882470145436,0.8345781321534484,0.3916981530940479,
 0.17628023492152511,0.6359051817844472,0.7690952603570232,0.8983775310187871,
 0.2444752727795103,0.6294991613064029,0.8777139655213556,0.12048808474003192,
 0.213055224558503,0.9524193426094973,0.943117470358017,0.7032818154194258,
 0.350473213979115,0.9379253936605051,0.10300616840181798,0.9033766754508031,
 0.8454001539971691,0.8597070768815804,0.5161455354184561,0.5126297921027773,
 0.5013600388791012,0.444601921596932,0.033474800862052456,
 0.23970043297274957,0.3284040824847645,0.46945802564007966,
 0.9562205405427536,0.01701577394960574,0.633715800951196,0.12380763719442578,
 0.33881620531842405,0.07691540799644192,0.842175469208256,0.9881188781281097,
 0.15121220666727853,0.8925485835864135,0.3047747296475014,
 0.40232771886746477,0.9573959869094295,0.817685393605415,0.7694697960444793,
 0.4258040817423996,0.3227707486939614,0.2383048245076823,0.3346693626687569,
 0.6375985059298034,0.46018973834367516,0.7573747512393154,0.5994789753282954,
 0.40225093689746216,0.9762537978626007,0.6863838566625576,
 0.12390160469896405,0.4097927016122055,0.10707161283533906,
 0.9931398336732848,0.9263460528265831,0.6249154428215626,0.13296981610554248,
 0.6085090458776284,0.9520754847865501,0.16077875810286535,0.9674235841952191,
 0.8433158277431062,0.16383907882364457,0.20344653922763767,
 0.6128587457768044,0.9503436942201678,0.6798539519286548,0.5435929098017365,
 0.5959409335769114,0.8128826582525466,0.5253973101571704,0.20912089268078082,
 0.9386627750244744,0.027594783253654498,0.5451402342769526,
 0.7292649330153429,0.9978015030590617,0.7877824115694246,0.5251373414241807,
 0.8923590437221279,0.6505128844850988,0.6878003708192578,0.23745027153612297,
 0.059422373248928295,0.46519497841795976,0.615316899384208,
 0.6943211678537694,0.8314188290214117,0.23287677829001052,0.3742044090434785,
 0.2446431120945438,0.18248663561464074,0.2840473685957732,0.508410627169757,
 0.6955509404003647,0.25983065450295617,0.6226776052283802,
 0.056397381460534746,0.7635571908160327,0.5870671487646821,
 0.35588026632774017,0.19320795222542153,0.10576678224393576,
 0.8858875063955279,0.2511251269562783,0.007707973599472617,
 0.5371902804146578,0.3243600709695642,0.8739927264709814,0.45399496007402174,
 0.5389508494464023,0.1283650873203921,0.3632766086689745,0.4021162551395362,
 0.876691869163023,0.09404645867772243,0.9535046617816181,0.953560762106236,
 0.031087972769307304,0.0668301031416656,0.35500449085941,0.2784133484818605,
 0.8291748381601647,0.8524932247709368,0.25067051105569915,0.7444656396872382,
 0.39490037375496767,0.6060401793487298,0.3103313219634537,0.7507316178947924,
 0.7030276478997989,0.9120194761338327,0.9732485743816094,0.6102364519063865,
 0.8368748985821128,0.5585225901324058,0.4297406390999646,0.9130419350383625,
 0.05955486827087386,0.8675710539765574,0.7489481204454775,0.976946958590407,
 0.10877512670099954,0.4849427663311783,0.009825195829421318,
 0.9580883053474822,0.11742742760197256,0.8078375859261081,
 0.011471875884639382,0.5445274667799677,0.1831861910549959,
 0.12653619425140472,0.20286214475343545,0.15734536106628338,
 0.9178131102664845,0.1445141651580376,0.10422292554013068,0.6718792795261642,
 0.31298182485327186,0.5525661033623692,0.8206536766834288,0.9721411379644193,
 0.026409803252110176,0.8229465769273119,0.9463971230401179,
 0.25500609593205104,0.5836738323485388,0.4617879268926479,0.3786817669865101,
 0.02349583584653281,0.047358086402096156,0.7616867134510815,
 0.37835637732011296,0.6516472500329498,0.8452799001227381,0.8892826170946175,
 0.0905127712599364,0.23986207821444794,0.46053359929497173,
 0.08607051773784691,0.45646902329060446,0.9233018173361218,
 0.09326792134462425,0.9422967340400324,0.3553546637984285,
 0.21834837752362835,0.07713791146632554,0.0011846791894336572,
 0.28771135958542515,0.7829790342806522,0.37221758813283945,
 0.09356081361824575,0.9136799174259813,0.5549422512557983,0.8564648135907356,
 0.502181759321447,0.10296524860217326,0.7820152522107563,
 0.0021292640367970295,0.6327373173827162,0.281993534565128,
 0.9021650097053366,0.4198690258023138,0.1488436951987343,0.35138827029103303,
 0.9250496649233242,0.26091733607598544,0.6506189793183471,
 0.15710639595365683,0.235412940838005,0.20145721572418518,0.5303915415534267,
 0.18053856689717684,0.1408600855513915,0.32467990437858263,
 0.21060443105713544,0.217913826852437,0.05434403054136583,
 0.04688134905856989,0.9690711951755665,0.7161191379421594,
 0.27556135889145883,0.2078261631644811,0.3522573759280526,0.616696278931644,
 0.7031486816523354,0.7530651832544266,0.09154612424953124,0.8821867714401787,
 0.5855117926245372,0.04346737683645907,0.3196313656687899,0.4560190505976849,
 0.3758922082621017,0.6925989016105967,0.8083284082257591,0.37763835239166654,
 0.7211325554511302,0.6222936333315807,0.4377513720940548,0.6835466045442633,
 0.967940847043255,0.31022401819302725,0.8598998271748267,0.4971058153195499,
 0.8636537177862078,0.18761317424635426,0.35190087408001147,
 0.6442062371243729,0.03755581418536158,0.8575705614512144,0.9981386029860588,
 0.47728874267297317,0.2454764057139115,0.6656988622383337,
 0.08776350395323063,0.1429980232224607,0.6700907078312693,0.6194540968359297,
 0.6862391348111929,0.13451053137479962,0.8584440716971463,
 0.016486449525170754,0.19658379433989248,0.055416463538583316,
 0.9716457975894651,0.8155371333065631,0.0959813444641886,0.6277085906471582,
 0.15697993628687246,0.5936299745104117,0.1930310085974667,0.9476878977830263,
 0.3114361756244546,0.9832656329106471,0.7911117873101733,0.38429721118316496,
 0.8284305198290237,0.5417176603439381,0.7810721500224423,0.7854202078275239,
 0.6003454412094702,0.933250823609738,0.30699681457984895,0.6371476236551028,
 0.8718469110077891,0.5976319191276284,0.6137245124520717,0.6641862747210331,
 0.25899667100435264,0.5477284611467661,0.6304125838162764,0.5950491960917167,
 0.46904874174743505,0.09358961443720792,0.620771310242898,0.425903836269916,
 0.5449832277603721,0.17129634987473785,0.831006139374072,0.42373810310989524,
 0.5267284256821817,0.29236391480305146,0.4365537023285967,0.6422291008780233,
 0.7370357231412781,0.22596469155126775,0.6436825499209433,
 0.052598973633706914,0.9700733718798144,0.10743665995881035,
 0.6274286863437303,0.7907910925836062,0.528611027188757,0.4932854800892471,
 0.8137778719057325,0.04424531071526927,0.12108213054342842,
 0.3776292434171453,0.7065831038963981,0.9855255286318475,0.4545109408886523,
 0.5881724239420352,0.5770003113751319,0.5547036940151008,0.14111102399379705,
 0.6855173823098111,0.2395758429179662,0.63257925618055,0.72566992761357,
 0.7825845230265019,0.4995515111335864,0.7209922830043665,0.1675751522786052,
 0.3742119402037303,0.15667052897702327,0.96192418021258,0.775249236068,
 0.947681882988001,0.9100020814473001,0.7872319290851623,0.5778567672755371,
 0.007599603339252026,0.4615017646986479,0.16976742519397192,
 0.08051706169228723,0.4239606508556675,0.22948095700916993,
 0.9117284543145454,0.08045836455860933,0.3020123484257319,
 0.47591318812874006,0.4840932263547554,0.47080163588749835,
 0.5875701789490339,0.9843080021951214,0.5091663986692342,0.9784416920490857,
 0.8522463971849298,0.1513811772417054,0.16052288135038073,0.3981048588672712,
 0.00835857485572944,0.8719030495035677,0.8335243399423933,0.7513130010206386,
 0.7886915365244765,0.1135483730131962,0.5188323779583461,0.36453251251042507,
 0.9514032841791135,0.6005566233990763,0.07128715557138199,0.7325780488554064,
 0.5606387373419227,0.5775667741894223,0.8582756312332321,0.8457764658749178,
 0.8377044068396946,0.6987502271567573,0.7873222120046729,0.18507319507474318,
 0.37703051827664935,0.2337673676735672,0.6422166090795035,0.7639907746937096,
 0.011718296113222948,0.43659657426172493,0.8325192914540115,
 0.9097086542176147,0.8646416903043908,0.8572985235046516,0.112895692032013,
 0.3324500632772712,0.51483145871806,0.3194332279535591,0.40276427624865874,
 0.056231181314914225,0.6508929813216088,0.2647376247936212,
 0.35448356812034054,0.18442682547878708,0.12368288881980394,
 0.5038612033342105,0.1838548813377685,0.7708882506103241,0.8576633411380837,
 0.19924702987972187,0.6929811916290901,0.1412160754552585,0.4649483230745649,
 0.08914348751523138,0.34352751811458626,0.2892824232188864,
 0.7816558655785921,0.9551452182856859,0.04405332051906574,0.8034081803994082,
 0.3401668950450336,0.06699502004431102,0.46229666910286893,
 0.3767062727896444,0.272859042857564,0.16011239335933436,0.9276387196445465,
 0.4202139785075495,0.9117080234980244,0.65955856446239,0.5324507737005928,
 0.00571947804435724,0.5123665238686081,0.5007711604125462,0.5057683615499505,
 0.8931419778460501,0.6446699422472646,0.5092586323002323,0.9369552841790282,
 0.13721473051830568,0.9426001355757625,0.570960949578732,0.8872777485681034,
 0.47625693850592865,0.2857719821610527,0.43687470041257037,0.867352207459862,
 0.9834137814283576,0.6175368554058366,0.012838054555584621,
 0.5377269780959063,0.30781345757331824,0.9664195275419657,
 0.43498317420939514,0.05250234478741811,0.564047019701154,0.607651699303935,
 0.24130482551181986,0.7202981683090399,0.9334303463920735,
 0.29561157295387275,0.8407221900692342,0.559589320601974,0.6936434717230834,
 0.382551103407227,0.7646353673068427,0.587767933899789,0.27466084450588346,
 0.026117075897181263,0.5365271399543142,0.3382307522075564,
 0.06850960018857588,0.757247947429052,0.4347034798862337,
 0.040055500519587284,0.19873143117697534,0.3958266940027686,
 0.9432926952900396,0.5064755145378411,0.4129866944299233,0.7053419383459908,
 0.6419182942254708,0.9941360368225507,0.6795890095101469,0.8128471195148421,
 0.763269655556955,0.5922969039180096,0.40067555672534083,0.559959462284373,
 0.6101031968322707,0.47671635583759175,0.9015168239190607,
 0.08149704413576109,0.061270700592371385,0.7611541155206552,
 0.34762541187430573,0.7125002512976097,0.3541854002722358,
 0.46302837413810194,0.8815301341508262,0.86336285934699,0.6872850222326143,
 0.2533871236550027,0.011339090167952315,0.689543387567462,0.194476648206505,
 0.5380576850120219,0.505681670824666,0.38248740874937215,0.9514530630303014,
 0.34069737381311993,0.5772151718609767,0.49884264498366204,0.978594352524791,
 0.9033261880225698,0.31793652578103493,0.1819370962090433,0.2106999046877681,
 0.25805059061362645,0.3306428852376605,0.8353781499257689,0.8311594981974781,
 0.503930134440291,0.2705005502386314,0.31319977647585984,0.37427220028344654,
 0.884943194738451,0.9074167333867389,0.7125469661684517,0.887906500556227,
 0.6756957589201354,0.7948134370780189,0.3633216847029488,0.29612228412525254,
 0.05568070340148579,0.3932424606936056,0.05405927572789282,
 0.48402241547734526,0.9719942751632331,0.8867045963627224,0.7576131019458283,
 0.7905679343598372,0.03156483582562719,0.7167104455976918,0.3915444056592339,
 0.9696381636515923,0.42530474016144104,0.45615966773339744,
 0.7198528095816313,0.2156093814381138,0.8462846024601386,0.6234734254432686,
 0.19477511850512175,0.7575575570154736,0.40620033253656385,
 0.8406170943060236,0.7520154879264517,0.3047989211630251,0.3065610667965667,
 0.3990845180020094,0.026090171564001396,0.5210300866505824,
 0.9090798392300975,0.36629036914021285,0.2417576560334953,0.7228150416842521,
 0.543745018802615,0.23584962093418294,0.7946279643351987,0.13720704515740778,
 0.7771666901817589,0.22220171012866063,0.7670557202389939,
 0.48958249588442926,0.5676470491587624,0.34338295089840953,
 0.7091786185889724,0.6842126304132465,0.28506438046033,0.939896432633533,
 0.3690942403651728,0.6722835433425062,0.34506153213059754,0.802797063095244,
 0.8772825450289593,0.38170607289868697,0.7617935319848028,0.8711637037996978,
 0.5976788827209847,0.3515337295878801,0.9565224807641696,
 0.033506470753224216,0.5700917545582349,0.26954673857049716,
 0.9793888923986255,0.6934681757724402,0.5260932656649875,0.10864894441163675,
 0.9514117397178794,0.6020311423196268,0.03937174404329469,0.5174535580364551,
 0.9205476014452498,0.6728644031080591,0.5495073198278965,0.2374574153485396,
 0.7893003649180147,0.5009503485743991,0.5086025089959847,0.6899786051493074,
 0.19046978014358096,0.9038292530936997,0.4795344590405003,0.7925666678167425,
 0.7790283987360128,0.2631529523206506,0.9355304523954292,0.7237574411925618,
 0.25975942540741825,0.25826511323417223,0.08044438961396039,
 0.683523667594061,0.2027713386465193,0.8375466994679908,0.03422720421143488,
 0.6283158211686829,0.08528278240888365,0.41636446321374143,0.825546219857771,
 0.01332838830294325,0.8952553330895602,0.4700659860583851,0.5846057834291407,
 0.3456511066025877,0.03816240014779593,0.553272165655552,0.11541872781917006,
 0.5009952337444314,0.5849925766714584,0.6490403281414286,0.5933741633425152,
 0.34451650358744457,0.23549994713175693,0.8226164084490476,
 0.3222556010100557,0.04979498348439626,0.09488792559031112,
 0.6301225175536945,0.8416850075901475,0.6591077262124874,0.5718159009951768,
 0.9754493367829781,0.8369649273502346,0.348438201332102,0.7009397584753053,
 0.8576744573455387,0.02404469267202103,0.857782928397713,0.7447034100410655,
 0.5417877230496632,0.24891301773504515,0.5185521806856103,0.6668522550966942,
 0.21890472832736652,0.18539391107655678,0.16494107555886095,
 0.007134287236318149,0.6181906240423138,0.477471485627016,
 0.17725721005812434,0.927015037526638,0.5952300976775315,0.1584548896087956,
 0.4225421543171298,0.017515035858083006,0.9223944455066633,
 0.19487788281114016,0.025695772156609897,0.03230428982283784,
 0.1301305367643364,0.33510632626192327,0.6047718009684101,0.3602929804118906,
 0.04264335767085414,0.13212042116196832,0.8160005368570776,
 0.5643191707937886,0.8009440556668903,0.09767705971557483,0.8374061875740816,
 0.9415163695905902,0.5400058127951098,0.25946648386734295,0.6341257820195066,
 0.6967207395923919,0.8170749279288633,0.8996725527110933,0.8431400835861954,
 0.6936619392651093,0.8389897776267486,0.9553750280537736,0.46393237571122437,
 0.6111855323000688,0.2406151786167936,0.14623761501760402,
 0.029600855388275726,0.4463829520160929,0.18695660464158115,
 0.6668740374141281,0.2775949637773041,0.6508557305339489,0.11470822466623298,
 0.0635608653330817,0.0481769542660484,0.5375499237142322,0.5638261523790887,
 0.8217660300954521,0.7905643845608835,0.5476281252325278,0.25010233634700285,
 0.2638176788804304,0.8154505840863275,0.9703236906963953,0.6059803888292111,
 0.7171909712840618,0.8809279295948856,0.07785793413381703,0.2957669806817189,
 0.49570979409644034,0.5422830466402099,0.7736130708479045,0.9485653549705824,
 0.8663305423825829,0.3025791098887518,0.655400840959463,0.6173360635447884,
 0.9765838149905973,0.4156612654771984,0.305602294921526,0.35787758915715107,
 0.039464448135320795,0.23590557425417014,0.9711442003904371,
 0.7047709001403966,0.44909742065039604,0.3428193165651917,
 0.34005005003265043,0.48070825994312283,0.8492937128798557,
 0.030117792498331575,0.8539338158457019,0.13653391874126086,
 0.13133223880583222,0.3276779522429164,7.844126817171393e-4,
 0.7311532611969922,0.1991522922035056,0.6651559060942127,0.7484663359822809,
 0.05704686354952282,0.2180572658396599,0.1176342650402954,0.3692737282285441,
 0.49716426550378,0.3029653207471451,0.5228055242247409,0.5944714009237302,
 0.3088402149326578,0.27782690137670984,0.09127907186691953,
 0.10536667136467259,0.8743639824984312,0.5419102578846686,0.97666388221322,
 0.9731824151162259,0.2239809354462865,0.1514593674136857,0.8965199148160172,
 0.8045133827693711,0.7639055908174698,0.0029223732732652774,
 0.09064474577820858,0.7576699482014586,0.06488660024014115,0.826600722425651,
 0.537829353069091,0.8447416122836602,0.8750976807881783,0.07666705284768993,
 0.9812094993481073,0.3945110138237382,0.38915780071525563,0.4092761463134872,
 0.0640244119322333,0.41650347040477687,0.3661730595227044,0.9065556994860672,
 0.6188044370736411,0.012061726076933676,0.6489490956121973,
 0.8888030378462474,0.9086789073324992,0.8917195282054843,0.3027685413759197,
 0.46781113461262724,0.6902318587316293,0.6313079092502328,0.6844841248659845,
 0.5733177209595606,0.6845147988659934,0.4577421933684708,0.6007616744855349,
 0.3311160988220334,0.8447995040138839,0.2125237767300927,0.4782731873914159,
 0.1578736840376831,0.39883026430043866,0.9240827887269886,
 0.011284211156714963,0.34426967211990933,0.937996237070476,
 0.4923762906638883,0.2148116524945124,0.6863602480023487,0.46906895131439197,
 0.7329191510436917,0.4236418362989606,0.7501666635726914,0.7767038155815793,
 0.4450959782919617,0.23062403666756748,0.1178897905256292,0.1473175818062189,
 0.34836351629014206,0.06625299289076603,0.9751661690587893,
 0.09837324211046716,0.8346854213843413,0.2587084525632819,
 0.29652250139794945,0.13440351123274863,0.9968337400131204,
 0.7409818232841398,0.8574447176305212,0.6260465251843237,0.1970592649873426,
 0.7008162277589709,0.9361790508715234,0.5945783864202726,0.8411348702454177,
 0.14185427805994633,0.8043383611879226,0.5814062373534583,0.0554849423044117,
 0.717169554944349,0.266944566766785,0.2240076127616839,0.8100789868841859,
 0.4226978188803956,0.28407457776965184,0.2915258374409144,0.3907608320771849,
 0.8123408963853085,0.18560677735173992,0.34245497886159526,
 0.6527894936277451,0.5368252254384317,0.9071398055527985,0.2777501090489327,
 0.8483450206675172,0.015139815376707166,0.1346124400792995,
 0.47228078694030273,0.11634017467245117,0.6659739059400179,
 0.8398623829706479,0.5604098831369491,0.844403150153209,0.7286150768819086,
 0.7479041568436262,0.9981117714322021,0.9419659421659244,0.7746766932486003,
 0.5153128294590902,0.7992921145313808,0.6922683528309287,0.38561319260628113,
 0.04475252067934443,0.5853474704108971,0.8529214418393793,
 0.10765501764498775,0.10719710215742417,0.9997676732580293,
 0.5618584357657049,0.6538004667950478,0.20458840071144047,0.7812410341381901,
 0.4312726145865281,0.564742116291181,0.009554656760889735,0.9340812882330494,
 0.9888637884434597,0.6540806847829338,0.7881025270802509,0.1971078885690044,
 0.1493563880416503,0.4092833812932425,0.26603474348848744,
 0.33744190051400647,0.8000137056323938,0.5876781158735751,0.5431226587771807,
 0.7406713620822667,0.3850631620408884,0.37363901203023575,0.7547988065110728,
 0.20399243741448125,0.036262947135752066,0.8603498004885851,
 0.2938085528701455,0.3448642020395499,0.8218351850675631,0.7148739338490259,
 0.7471127393234454,0.636211204437375,0.44429690243552933,0.0580725498857636,
 0.2241417134672684,0.7968997952951279,0.2015388216403795,0.22186888871670496,
 0.1762341165116592,0.8323794427166413,0.8180467605638749,0.00377376603697821,
 0.200292440718721,0.5128276723415882,0.14261488000416855,0.30787045384484035,
 0.3060915180675343,0.6853000718204909,0.058766794060278627,
 0.003659322025473255,0.7344485923858888,0.8934313112902664,
 0.9667280983312758,0.8359089225167757,0.04945723472307506,0.5506433643679635,
 0.2309223236283695,0.978690906134148,0.7655558450505447,0.7018347032025704,
 0.37203820336651405,0.6991939581664788,0.7825535527629817,0.5309493991143351,
 0.16006319753808484,0.11660836577552613,0.39560387107599604,
 0.8249130227412393,0.6145114663801302,0.433920748500471,0.8721632349714242,
 0.9660733011014363,0.6624425411912461,0.2747021951251363,0.9206168250866478,
 0.7384327804431441,0.14664941798371722,0.4779929454573484,0.4447594011552297,
 0.8862625774177942,0.6656714323689485,0.13753048767839493,
 0.09963275271155125,0.9681229766975253,0.32557282347017225,
 0.7904272270824925,0.10091614891725864,0.43553377376237923,
 0.5124494112832725,0.6548510045597553,0.5500052895766413,0.7586899566982097,
 0.5699842808301023,0.8645075506356187,0.30685829943485043,0.8504035539254975,
 0.4653952982371633,0.3601074664570548,0.3433407230377379,0.7485265677661387,
 0.173794043635511,0.303111938570763,0.2644879213807231,0.8040352682103815,
 0.16662378776520392,0.5150172366854157,0.24277980248125608,
 0.19712915694492672,0.1659156507058972,0.10123780920021275,
 0.8818577822161533,0.6733164454187138,0.2106741786408559,0.6532485425754726,
 0.8039671900802522,0.10214113902221944,0.5283342035598645,0.3995636645634859,
 0.5584121102792785,0.525032667683109,0.14176248482867848,0.8913604206899608,
 0.031953634049139645,0.6419187515498228,0.2693631356774733,
 0.6645061288474456,0.8149418723148613,0.25748210499679036,0.8506645984142055,
 0.3105154798890696,0.38362843823975423,0.6211171097235276,
 0.052784092726479104,0.4848166696851698,0.8471335135299312,
 0.9134677880683825,0.5487759468537311,0.5023071862464725,0.9432199781728692,
 0.9482507524751607,0.5802845065831779,0.4087104806865741,0.20680755198099066,
 0.7146064049051644,0.11618828568485062,0.7275591293393004,
 0.06774359163176513,0.30459835578819705,0.7912623457162048,
 0.7753572165655725,0.6134878116475951,0.8187446685787021,0.30031129340422147,
 0.22913921736302467,0.6015479289224871,0.11404074086977245,
 0.7910678282642343,0.5203405950544449,0.18862052810595797,
 0.12913128962744203,0.7055497284025526,0.6714716107017819,
 0.46610186185451097,0.12579366496414934,0.49984110535483384,
 0.6003834108348566,0.7685791635175682,0.81534656610886,0.21129958252770065,
 0.2865963162180746,0.9641741515869557,0.5172812685778406,0.23197432252495742,
 0.2893445760027624,0.34422235566653525,0.35390120576772044,0.129483607118831,
 0.9467852218353833,0.9823990307535861,0.9261023316071413,0.5662016199859411,
 0.6651695699346772,0.22067676471289388,0.8183953668660983,
 0.24761589673345075,0.9749462263894659,0.8073549637714618,
 0.21683356584782987,0.5683619661038927,0.33671359868757567,
 0.07717733206230704,0.8321523595434709,0.6409996448032872,0.2590140544044386,
 0.8802787387466524,0.36405236668128427,0.8524962456913872,
 0.26307385308265707,0.7045371051966565,0.40088439088757777,
 0.7069700012235866,0.07127806461178177,0.7952366598428965,0.1045861154023221,
 0.2215458707101241,0.9167824045435572,0.3608666235431719,
 0.061188147561065165,0.830019382250492,0.8271489592898911,0.9481690097388338,
 0.5660661943383597,0.348753868880902,0.22986169741085316,0.05117500484134918,
 0.6854128248305051,0.3125499584089537,0.039364123187300426,
 0.8654311271170723,0.1647574775546513,0.8200684864068482,0.34171221224750503,
 0.05098902250805126,0.23324694349934405,0.9313863887233903,
 0.2836099368066546,0.49135776011119736,0.2608127993553343,
 0.10055223093124854,0.3510677906234845,0.49687833215900656,0.947652203064442,
 0.2866733728178462,0.6819628907744719,0.5068886668257991,0.6604925039162435,
 0.6652537267957338,0.7718366111548174,0.7635717812132176,0.5393467662612523,
 0.50819982421856,0.44945418844846696,0.6801189501891928,0.48075962081295753,
 0.710368016167521,0.9940623599178011,0.24456978927443795,0.6713698154466817,
 0.6540280874204109,0.5966448472531848,0.7094841638380457,0.4708107782698684,
 0.6634207427108462,0.38587008452959426,0.7820493243278823,
 0.32535331858843497,0.681000032692429,0.5556511386759224,0.5122390235858832,
 0.5722886434561815,0.9360178390070106,0.9243768161303368,0.34648678637102925,
 0.3084062718396612,0.7180040753799005,0.5321192974102533,0.40520601704366566,
 0.35340516967340896,0.38946901811936996,0.38462889474393824,
 0.3580701234400212,0.7586182617402636,0.034836568920661226,
 0.5246226052099241,0.12722992915829845,0.8770722895474811,0.8561728511000856,
 0.5857370931618407,0.8570869590550124,0.374607446957832,0.3108894736220782,
 0.35361800305215574,0.5932604418565078,0.060112995299488325,
 0.08714548270850586,0.4143517780126331,0.8802005044291057,0.9504424583044496,
 0.9753955855748777,0.8647033912819984,0.7878822641221008,0.09184657585459255,
 0.7520524983495117,0.17257321244339263,0.22341353098594885,
 0.4896398163178307,0.7290609977946362,0.18560214214642212,
 0.10102872806541208,0.7064221579285275,0.4924436727523862,
 0.33739344218314427,0.214478860697633,0.8177759702114794,0.13880780072528942,
 0.002558366804847334,0.6348981119527826,0.795857307043606,0.3420590974742421,
 0.11375753922538456,0.08153050020143426,0.18950316625643893,
 0.15528311741816647,0.18996152379685483,0.7484453931169244,
 0.6833683051868343,0.10231637571628915,0.1449677998674349,
 0.39186525220958446,0.49392224297737775,0.5234015782815712,
 0.9573058690910365,0.0036484657677842147,0.03168050930111266,
 0.7938068263469897,0.29061449113872706,0.0067518583370673735,
 0.048430481201087616,0.1753221003105938,0.5720088357757196,
 0.06789503720508372,0.17318758495803066,0.6225874463660137,
 0.20428343439692187,0.8188349294666486,0.8599784664253854,0.6067692901876377,
 0.6444552726652968,0.566128749986757,0.3254784487511011,0.9933613026381721,
 0.3493062667128296,0.3818979242559457,0.3475384185939028,0.30092246297769765,
 0.8623187648339778,0.7576577343126865,0.07745441714237833,0.5342234676879554,
 0.6501062630066924,0.682357606017964,0.8235292914596745,0.6691284569007061,
 0.9706536915237755,0.9743440763676089,0.7929521988124703,0.7388601715345936,
 0.6372691423443072,0.6044041601289513,0.8395792451046821,0.4961706971297607,
 0.35388307744617764,0.13496920468223028,0.2891653255537828,0.219810359794956,
 0.6177390869149466,0.5055787480072591,0.19495613910183807,
 0.06577886771129782,0.5142572914951559,0.546504679177545,0.5036615781709664,
 0.6643493917937685,0.056078088636246304,0.5081210704297902,
 0.33821782889195995,0.023695828496348992,0.23190681386633427,
 0.8563009695983504,0.39944828480274663,0.08932451624333959,
 0.6348845932725523,0.8662569303660679,0.7026081028992582,0.9522604938954161,
 0.082567381632772,0.022047016664268226,0.2071063514097471,0.8189984428797589,
 0.04711135624173668,0.19713808493930995,0.09214668275643578,
 0.7431611549101785,0.4651748021277491,0.8044564654691612,0.8489638546473204,
 0.020962782271396474,0.25177157608816314,0.35169870887596655,
 0.8949039222985281,0.19859140498108063,0.740847515146708,0.5436885865081358,
 0.776667781211504,0.5413709326715602,0.49581536648779734,0.8462647909355444,
 0.723704945584089,0.12231045818743125,0.4044665570285064,0.358150789521013,
 0.8632680802329834,0.23744429129900624,0.4775668373268884,0.9205073573252764,
 0.06860076020819339,0.5324823919863377,0.1622499039630384,0.6846639213980875,
 0.8991538691829459,0.9403091815447826,0.9839625795153586,0.3561573302676637,
 0.16636081646341183,0.05599243594424275,0.30253140643205123,
 0.25844306683122886,0.7759705972736699,0.24230804020623065,
 0.6927636145795145,0.9500516218455816,0.5435305095763925,0.8043812718946981,
 0.7407802562741189,0.4400347666388511,0.40760501956250617,0.319549452817524,
 0.16655563291686692,0.6926261564825028,0.9720487499805609,0.523177795207564,
 0.9914771424219988,0.1921516472755752,0.5053122787807669,0.99801519700614,
 0.23590574226533823,0.22314786898457917,0.4724960416786691,
 0.01634290488044865,0.2558463056734739,0.0765567012185937,0.8730143533146802,
 0.2834855074160674,0.18563598963657701,0.9764134735568485,0.3046003697291757,
 0.8171943532918134,0.5933101385465152,0.6171805622012491,0.2527286491508012,
 0.6603805147549886,0.7352471912640103,0.8440970576749777,0.3928556269349498,
 0.7142034694832349,0.2536621091564121,0.5436546423527376,0.9195941894427744,
 0.30220107136190155,0.6031312791546527,0.6429598410967519,0.9678832574764943,
 0.8040451922184628,0.9274120337588482,0.2197143991858027,0.7809128099013994,
 0.699752702797674,0.9094881227426148,0.4814082734042022,0.3593880942233738,
 0.1814720216631126,0.2579600066063281,0.9683736329205967,0.9701539786721359,
 0.8719842505477866,0.21934155815131684,0.7459545657605114,0.8003114140498083,
 0.9323758653534757,0.4880674661310751,0.6956646032029608,0.5801080434006194,
 0.39610180886410995,0.9374548038628087,0.2939461915205521,0.5030241344091524,
 0.4271642283241208,0.9377913467267688,0.5185640860610828,0.6113097055748221,
 0.565144871605561,0.5267268960124016,0.8403010269840813,0.5111239337094889,
 0.07145419507415429,0.5408358183349907,0.5163061937772537,0.8235353046625511,
 0.7197186567224647,0.9152123807126546,0.9808637171359095,0.4293938964473285,
 0.5001647749291678,0.029853117467710577,0.08082003281237582,
 0.7148855948543811,0.16770746399179004,0.8833572410740285,0.6223033103680399,
 0.1898529117733716,0.6729395565265016,0.7452529198341186,0.16974522163737527,
 0.5009150644091225,0.43515295853284663,0.8649942650058192,0.4925510907380992,
 0.6201971539434206,0.1335856401167913,0.33638450665518504,0.5982267759806067,
 0.3448774362020506,0.994634046710112,0.5987727917536038,0.8144977458910638,
 0.53198967032913,0.2484892519698576,0.8292644662330463,0.8695051112652062,
 0.5147459868790957,0.43000194327799557,0.5660712105245922,
 0.062292812973936185,0.8068452966870979,0.4918231030044593,
 0.12458344049681325,0.9658848641665095,0.5031116298330127,0.0432086883495264,
 0.08459335778677735,0.9718087568404126,0.6740568164013421,0.8242945634320015,
 0.597787563680012,0.38649815707848967,0.8575883976661518,0.6043276614477505,
 0.10935562514990804,0.6981510726234924,0.6153369575716459,
 0.15760087911238374,0.1923427612358315,0.707996499860759,0.1027259045920671,
 0.23981707026306143,0.7497152691533792,0.4069748498930882,0.7433255824850207,
 0.951871822874844,0.6080068038605143,0.15259806523452113,0.20625568427966368,
 0.9212061828501233,0.19680944964469593,0.4158390967711372,
 0.30465835879446945,0.046124820970915414,0.3592611992512782,
 0.49096961871396105,0.2879426821461112,0.2749252598723956,0.3911588775090077,
 0.6718585069611314,0.07228398346930087,0.30125138552746766,
 0.3213124784445408,0.4271305011626054,0.8950134649981076,0.2947055488226322,
 0.14631385350975368,0.5721853277869994,0.6793112734004837,0.5665023915649546,
 0.696688612343077,0.46896239400583706,0.8165865058056144,0.9116026330537176,
 0.7535412326899185,0.5043134280191468,0.3632064448447261,0.689793698385209,
 0.1656454557690159,0.129770975441168,0.622556200954458,0.5168305563148013,
 0.5133022128112288,0.3767426436707648,0.4699883340800941,0.7861400872665708,
 0.05125399051108426,0.4490729279540575,0.3947048468450738,
 0.43297654862909996,0.48184633961070156,0.9686192352666712,
 0.6654940553463148,0.66716479465268,0.5963028983472585,0.34345424566880156,
 0.5322355649226195,0.8308440884667441,0.09678941010059705,
 0.49396377634334776,0.9568567669391298,0.35460803506643956,0.611717796591454,
 0.9122423471098744,0.5750037726758871,0.3096944762534527,0.180180975024407,
 0.7423682649704993,0.0251283753397753,0.5493002928177052,0.24559495899980788,
 0.700778948784045,0.17024527882896712,0.2163334409262514,0.6818939567951411,
 0.7574564044083587,0.0907320683911581,0.9473525966688614,0.4872589271811867,
 0.11766359796969317,0.4536914947311099,0.7205742897047258,0.7673823845228561,
 0.5472390750251114,0.12172338447558073,0.2696962562657039,0.8190972123123322,
 0.5648158863146163,0.8246695405692235,0.4255937740663204,0.8536245468357027,
 0.9573784643212929,0.0498598110018551,0.16484360862703062,
 0.15338497072822932,0.02875538417886725,0.44827652781169613,
 0.12666364555328624,0.7349012740240463,0.897663966980698,0.5081920924413393,
 0.38274672674983734,0.765611708187454,0.6573142771246069,0.31685652391971264,
 0.42567341885889953,0.8076373952785456,0.7258611212466315,0.9273213775845793,
 0.1467997905289793,0.5913972704823726,0.6038253442035859,0.23350782438003193,
 0.7953637514311391,0.660437855997133,0.7921407673781967,0.027554883410142006,
 0.1209197127920848,0.18760504879228412,0.7633379190994736,
 0.48328638025842785,0.12416639818876918,0.3135620736714777,
 0.09611811520717128,0.6870195127654274,0.5858704867293022,0.7005122915452102,
 0.8915339887438156,0.5817959953839181,0.18010809413013273,0.9391414446679003,
 0.009547062202972523,0.08490895561565903,0.5902999054580238,
 0.790643073527862,0.6453641524619534,0.1571667872570922,0.44581372568282795,
 0.3262557508344104,0.4016790441462902,0.7414650651787078,0.9746318382765735,
 0.4151866705030859,0.7394064063388718,0.13258438215980306,
 0.20734163562927754,0.8396054955257881,0.3027276405322429,
 0.008053220911466985,0.6095659828802281,0.9496331828003525,
 0.25380982128354723,0.7020351269084095,0.29524643563774233,
 0.5986803696925853,0.4351988167016597,0.567897319038515,0.2840202901143415,
 0.9463795042685974,0.5950428063958164,0.1606166843749841,0.338808026545389,
 0.7698117475348497,0.15162120492686848,0.5654365260897916,0.621167870116638,
 0.4275848744037848,0.09860389458520546,0.06976589605520023,
 0.19514830819602957,0.48710636921216566,0.02746755060327899,
 0.49846525407836473,0.8698088063293565,0.9336041006068031,
 0.19806841311179657,0.09274105902129082,0.3953554384675053,
 0.32041757906707047,0.9271590094901113,0.07232378819865315,
 0.5362967962610465,0.2848853595040026,0.8774728907050168,0.03495009045717601,
 0.5749970990008028,0.5606885146046695,0.5596620346285186,0.9257830591933429,
 0.4945903054532882,0.20893726655263778,0.42721582612160436,
 0.5705463054499288,0.41246914014703606,0.009205841256029745,
 0.07519568566444357,0.3574989017935568,0.17053196934007875,0.799567091764375,
 0.8135778005489591,0.15737764561721546,0.06668837912250536,
 0.6467783059225573,0.9308947752625549,0.7079359358680019,0.1623698567203763,
 0.2558771643819776,0.11750855896963575,0.529433867811854,0.9032850156784673,
 0.24170830092936224,0.09466690395770339,0.9616223741628783,0.783562678256404,
 0.37402290284107587,0.07002822798736141,0.8156942602988959,
 0.5907307747401376,0.7064279613708225,0.3971861752478497,0.687312155713022,
 0.0662942098489121,0.5375498396015151,0.6560196184928272,0.9623065076503431,
 0.36153516411708353,0.9549186774690622,0.03649627441952141,
 0.5415224422274384,0.9753828029630063,0.19586132321748206,0.3966160468457862,
 0.71514365137727,0.03703431695176773,0.2532629393468504,0.8015874828268182,
 0.21868522302626525,0.8020327937297305,0.3199868083531401,
 0.32348696908794095,0.2810963762704799,0.07403173457456036,
 0.9912049792771623,0.8103613605308461,0.7126127994580957,0.8459752543539729,
 0.41677248236269415,0.1661634607207837,0.8606501371896543,0.4564597827476108,
 0.23435390730455952,0.02289964860891469,0.397030466524697,0.5672707682722713,
 0.8420490923885174,0.58637357441109,0.571522766645923,0.4859308139109719,
 0.04991529922142157,0.9019051201144828,0.8173083446992094,
 0.13706632062369695,0.14811037846468067,0.6157970634957798,0.477012891032798,
 0.27855762306867404,0.6758993097626477,0.7676667506215639,0.8931482998896828,
 0.6952996999335497,0.34910483943450077,0.09899553609865208,0.892074892482249,
 0.40725210658930155,0.5861546361498642,0.8875183634549326,
 0.37231539793427115,0.36306966281259934,0.7504900771039276,0.962351187043205,
 0.2385025745006104,0.33363493772523356,0.3425311932642777,0.6830060054413676,
 0.17467424860879233,0.7879027400520452,0.33625073327607213,
 0.6628722329614722,0.9688540474922477,0.5223808364402888,0.9872782667375598,
 0.6579533790113307,0.5519726040412639,0.2198226342471371,0.470784318849508,
 0.7016401127129692,0.9019235317584487,0.544436238317109,0.3520989731981541,
 0.97826878404655,0.5831891018129679,0.3785700066988976,0.19525137021890115,
 0.3134584509773477,0.04846343240847262,0.6324355516148272,0.6154667947211799,
 0.8109875918528296,0.24132740248569762,0.18549584428514543,
 0.6543210390244296,0.30120602435457966,0.7598125740278672,0.4411364034704788,
 0.4171684491256463,0.713145586342294,0.5839113404613178,0.20302123707300435,
 0.23307324458971757,0.01821835796211202,0.972667227045663,0.9692845226685313,
 0.5437122808423086,0.39977000602361135,0.9359487797464694,0.9265762950397615,
 0.6430973210527932,0.010249056065795958,0.43945586050447505,
 0.4927391890134203,0.4832573416390932,0.519024425966494,0.08191547195939453,
 0.23620487321986072,0.0646508860643491,0.04515773804954182,
 0.049407572448782355,0.43060856828023,0.5164197015011873,0.37491662004226833,
 0.1892538253111129,0.07194968523943801,0.24537030198231324,
 0.2452773731431912,0.5634361737087106,0.425358332032333,0.018567246267785276,
 0.05304165745355405,0.03483648793286287,0.5821349026227844,
 0.09905268423601321,0.5475208168126042,0.4582942173176897,
 0.17633094466062793,0.6099684877112782,0.4415747084988496,
 0.13221947417774116,0.5239847745030712,0.9304015819836123,
 0.09937480365290385,0.003596150671560605,0.2763020946676976,
 0.7769323543733644,0.46489994662002576,0.855596511932245,0.28880060319252576,
 0.6218465012293088,0.14687794065153292,0.5555109973225565,0.3844561338036059,
 0.7885531442378017,0.2869670501554267,0.5872677598624684,0.5161049650393503,
 0.6688639207568737,0.8947666563825002,0.2718391652826937,0.9283653169146839,
 0.7568907066675273,0.09694327435858907,0.47175067225193856,
 0.8667640260055931,0.02464604103471879,0.6474191845382,0.7732132102386773,
 0.34672359980357037,0.9779108592123382,0.5259007474837324,
 0.25423513450091595,0.13101597677124832,0.05710099632912502,
 0.5398322147331769,0.19143995982295559,0.5288278079466812,
 0.020656401173614447,0.3778556322330153,0.48248119753280594,
 0.6668953981698497,0.8294043487784772,0.6411600204781727,0.6956429049305328,
 0.07183729361041735,0.1007492301001397,0.723793071992795,0.959913283690561,
 0.9889615627505878,0.7815547893499921,0.9552273790246042,0.7193547897505758,
 0.4396132847365297,0.06665931647745027,0.8990757138158132,
 0.07189763754196898,0.673774570730819,0.5227958798098779,0.35290998719467204,
 0.018405699913609297,0.0015239060789233871,0.263777682115372,
 0.998890449564817,0.37913320201886314,0.09399058912518621,
 0.29772778745018846,0.34746849592124185,0.5683149007700402,
 0.14231360429922513,0.2278096655165054,0.7295656703826381,
 0.03303465981300413,0.2666491551335365,0.7927013451333826,0.2114713072981793,
 0.15985306315919257,0.5804730642629921,0.408902825555017,0.8869290806058996,
 0.4646557059156271,0.011449966020409597,0.08400182444785154,
 0.20186366455230698,0.9965784183314973,0.48477278951835723,
 0.07701280946131839,0.36313416788454844,0.9588793592855482,
 0.24715416648176358,0.21971315435248906,0.4978719483292223,
 0.5850138456993591,0.010395507595851394,0.12284277787653364,
 0.5106660011973742,0.5164378847478136,0.8980513122201511,0.1805780908040644,
 0.6565808216836583,0.6077736103228919,0.09728487659438056,0.8775866097393181,
 0.695572468178981,0.9468762943112837,0.316842585569616,0.5038657958878084,
 0.6285720173083774,0.958193059282716,0.6523321508631126,0.8132087041065097,
 0.9847930991464026,0.5707827792190077,0.12970086401168013,
 0.04622597772165804,0.34465580228253456,0.2776761779742817,
 0.3217497101350283,0.14275997287368025,0.8215028126020497,0.8416015583329419,
 0.9027691651946095,0.6867595905588147,0.29714182096745845,0.8652094185849742,
 0.5565151214314904,0.43503503238334684,0.1974746912617268,
 0.25154655296828987,0.7163267590861913,0.6852037948700332,0.598474272572768,
 0.7368823317125988,0.8235611222912866,0.562153310862567,0.3051875709149984,
 0.1471417487740705,0.1593412694432239,0.5618354074198489,0.40392072596067274,
 0.8103044376766656,0.5967191641020342,0.0560638061915224,0.7419434934699655,
 0.6637169663393232,0.04484417211554148,0.14126007048248523,
 0.03437414544436945,0.11952436604002326,0.9730310576080042,
 0.22349447498770236,0.5571710243806695,0.5135211736868595,0.9079538391937229,
 0.8070033974799639,0.8928645935891246,0.3157057254951242,0.40485450990927885,
 0.051327047551097404,0.9043344334732161,0.09203645948760797,
 0.7617285422664405,0.9931245727459734,0.4751157642110586,0.6897627061611047,
 0.2467980772338949,0.12513437621423174,0.3288296827815844,0.9756469609897923,
 0.912602659863886,0.0013462657767336905,0.43372191551613426,
 0.3406698175052135,0.24539562122612846,0.80463929076122,0.6519605870787437,
 0.5566200459103093,0.537394767485308,0.20617369772163152,0.18570650464803995,
 0.034890212517343255,0.41196277515383006,0.15499236059938903,
 0.6143398403239575,0.4352096187668587,0.9003258386266172,0.9524771696390631,
 0.28061331424153824,0.9789444786370017,0.5303020032012582,
 0.24210414383625378,0.6072123308566388,0.4470895928088098,0.3928685499705472,
 0.23394712371127202,0.05280198910013434,0.1881119907848221,
 0.8546634415295776,0.17507617582220425,0.8017711785907767,0.7094365426201631,
 0.7845615117965016,0.9413755503237211,0.7084266707002518,0.0645814240603304,
 0.7555702159619774,0.9515079755008967,0.8570919864631532,0.7413291771406113,
 0.7605456592937363,0.29071369318435103,0.49401862420560627,
 0.08509984171334772,0.8217581888166118,0.9047334137382773,0.9253156897506503,
 0.5274223115272961,0.322855122652805,0.5832436515561772,0.8310144133085777,
 0.8628995043308507,0.10061590990592528,0.5799323770638217,
 0.42828384535823727,0.654471282111307,0.8222403139417791,0.5600684482514212,
 0.46019818951316804,0.54295629572458,0.10131524037774009,0.2344196152706277,
 0.4453048591603108,0.5443341588999583,0.544679036701433,0.1667797803599841,
 0.21416549775615346,0.3417817598550196,0.5704290218019503,0.8161839841681982,
 0.7396434461801745,0.06113453191136231,0.7350388445604261,0.4528573631566548,
 0.4837393837975442,0.6186076179258835,0.4360330816981782,0.40150489170673553,
 0.4395057883110842,0.4554876218519712,0.08803559156744356,0.0718092845408872,
 0.5708185600162402,0.5656437721111871,0.006274252618509002,
 0.7903939982218289,0.05195752379329022,0.18329604050026982,0.412354366967592,
 0.3550586875274587,0.7683528112922703,0.22525942136187438,0.1088710209695063,
 0.8741808857741975,0.568358815820203,0.7176521419961962,0.0193046979196575,
 0.5945107510511357,0.5436900146163393,0.031135639870206733,
 0.6203167906480573,0.6622964174810487,0.14587166505940474,0.6529344230591011,
 0.13745666644358334,0.23835988791280882,0.37962236613694356,
 0.08688916300317162,0.9011890599340957,0.8702937200315247,0.8419051257980023,
 0.8737542498535897,0.470790441867009,0.007703459846836069,0.5795537828434287,
 0.8523941701883825,0.7757671940110079,0.1507204777049812,0.5218669300808594,
 0.301143558201576,0.40991892588247303,0.2684230140573508,0.1996718120420239,
 0.28257468322443324,0.5217759781364214,0.7630945590126159,0.8346706685876805,
 0.364076763862887,0.8899582305636405,0.18938830194063105,0.8564847473916308,
 0.5509837986916645,0.4117470817001261,0.8824224597709985,0.2245377326110134,
 0.27112331994325445,0.9704632196617826,0.46403930610571387,
 0.6217575291485772,0.20280000162989964,0.42895968122486616,
 0.20877805797695748,0.9068129464726413,0.11607280374849971,
 0.7340741620246893,0.018245333829849386,0.8474275136421229,0.551690447785764,
 0.6813303105179933,0.5667652977402313,0.4220597728397746,0.15642867737135657,
 0.34767032044301605,0.236883638652729,0.8172688102549086,0.5787310513932543,
 0.21368471553690138,0.9817871186422851,0.6328802385792845,
 0.15496286851409347,0.3145614630651875,0.4780827675312167,
 0.09343053991180406,0.8044342152848298,0.07351997736828708,
 0.42616548911843677,0.4134737748018985,0.7784682006439486,0.3935681133368695,
 0.1629682962609954,0.01603052401449323,0.43622890522024804,
 0.3058573107311494,0.8169425415483876,0.6114561503801433,0.8740064062607424,
 0.718057721748859,0.43927068976119144,0.5757606931898902,0.0404020645649138,
 0.7015838702504713,0.11583692383446231,0.1011747238557037,0.9227498988377718,
 0.9002712835199826,0.5344948110956762,0.8005488462687086,0.48430586226713834,
 0.553237724254551,0.8059103623733295,0.24077687270463077,0.7718751337415666,
 0.5831402329743329,0.5309021962687991,0.15735810308772324,
 0.038351570255394374,0.034370050448474965,0.9172471068693994,
 0.4983879303743137,0.4967549832762962,0.16644058267468953,
 0.48936853677447834,0.10719037573553769,0.41183540661583895,
 0.5137335437440023,0.2704222323308866,0.9807691087434304,0.9568442149810779,
 0.7156499775401612,0.5857241356208314,0.062239163565258604,
 0.8283336907619274,0.7892457612833379,0.757641279871774,0.5759701353502854,
 0.8113860052457562,0.11129842854341454,0.7857463136433869,
 0.36217642316300813,0.3694337313589625,0.33373020562411304,0.608084456339018,
 0.7242418414159493,0.9065653178586937,0.31501039745501924,0.4741217305513795,
 0.17279909113350056,0.18219058837429014,0.2809546689762642,
 0.00827346898417336,0.08145723936766192,0.7052733005114802,
 0.3518422329469123,0.8868442369687709,0.36956390778515047,0.9170381989031343,
 0.6377376283549643,0.6363089904099207,0.8572595817002684,0.43921359938613214,
 0.15118438090334874,0.9520692891786284,0.14122060715635454,0.660899830875335,
 0.24191069503211537,0.4167615519494047,0.8837322732679036,0.7598566561317595,
 0.36827165815826257,0.1482765519408471,0.7528678680765323,0.7661534076454166,
 0.45925032315127434,0.693857989165589,0.3566404911083615,0.4123251659865841,
 0.013522953970267704,0.36318446575581476,0.43922510189739594,
 0.6783767853629462,0.47516756176558494,0.23829626317899977,
 0.23372300017696168,0.7321819572728392,0.13930315037728713,
 0.8037814318319098,0.5056421964148695,0.21945946444966347,
 0.46756406274457585,0.5104090986030746,0.8152859479762249,
 0.40785670684847675,0.33261161280177365,0.23658494258650387,
 0.6584797475559919,0.9747311489529195,0.8005166942907119,0.41022968240992475,
 0.7299820813971681,0.3618666135828331,0.9242613998657665,0.517673354232512,
 0.8555180753988229,0.7303927465732318,0.267306850200447,0.2535404684022403,
 0.29746295713239346,0.3632335778228928,0.1622746211959889,0.0999557136303304,
 0.5997340964562842,0.15413704908397363,0.3118266216323169,0.9360369022548045,
 0.9986332185041882,0.24320884793981978,0.2843991820569627,
 0.24900522662005198,0.45690874065430953,0.4587253248105405,
 0.7491799808702138,0.7408428277237471,0.4398863115771452,0.3392917992841791,
 0.35262142414560427,0.8974386278729394,0.540107370600383,0.8019040102131871,
 0.9764888245210882,0.42319408149054216,0.5426629760601236,
 0.17506625606144433,0.34312014787143885,0.6128967680525679,
 0.4046713985224438,0.8173782433600676,0.9613977405660978,0.04711381252321145,
 0.5117733138049839,0.48780370511247806,0.49809691469213624,
 0.47903847470802496,0.9679104921469166,0.24549852780817405,
 0.0024311926696721775,0.9305829414413568,0.5106924960692107,
 0.009940694171980957,0.6579826316233777,0.6794132679426882,
 0.17798368784168783,0.03516424347830538,0.737307570095115,0.4217569565760192,
 0.04173472592586536,0.882560841193492,0.5017495024152622,0.8757921273599879,
 0.7958610729168033,0.3225963067976806,0.9346351415410348,0.2975188576443012,
 0.18319044112328098,0.053212923873925955,0.2028634138950295,
 0.19277957423322567,0.9897124793131661,0.67297089323495,0.15050679138413703,
 0.8925648840013191,0.025512431546703795,0.896406140964669,0.6929480065498423,
 0.39835274472085636,0.6618400586328476,0.42286563669225885,
 0.2102371354678393,0.3531891145067729,0.38166289938853093,
 0.031115322360986797,0.031176227368631126,0.15692409659368112,
 0.33646416176229965,0.8247400275699731,0.9489156379919665,
 0.07207512371240743,0.5320904893212862,0.6126179581704506,
 0.45920171534771526,0.0313350467247836,0.6447393612814741,0.700707532543683,
 0.24874206301457025,0.8093562235057966,0.10499042297080452,
 0.5969128513647692,0.5652307869902561,0.20640990679023274,0.6071230012170387,
 0.051538946049556666,0.9923779559736395,0.13340209042540796,
 0.5002429112458783,0.6960494882537178,0.6470929998834746,0.32159962730870073,
 0.6175205089911843,0.2483072580422735,0.41925110494555984,0.2907555689890402,
 0.3518843022851783,0.7270695107028642,0.29428948846893244,0.813487260391941,
 0.7004479389663267,0.2463960507799554,0.16799520110469857,0.7437742623679249,
 0.3891232069971652,0.5745012019356832,0.0577054087097697,0.7340891146332984,
 0.5368969600349177,0.37458221862498353,0.5275664538650434,
 0.20286498918010432,0.5952014186704316,0.5873971953408965,
 0.040719886399953276,0.7443220038790708,0.6655934029312965,0.872742405452729,
 0.7515812918394184,0.6734398190659198,0.7945754257390699,0.5847208491003706,
 0.8379429917885994,0.7115159786022892,0.5049644218660022,0.9457780077081587,
 0.6081250379945106,0.8199024553010384,0.028266343952198847,
 0.8269546863014676,0.10800385810970525,0.8650176412222362,0.9413338807891154,
 0.07820199110546999,0.3206683751859717,0.6182691712247724,0.691743034767024,
 0.6772059685518348,0.9506456628686974,0.7436281686320049,
 0.028380673771486453,0.6978316457385421,0.4241123743395565,
 0.0028436463185741445,0.35473335864082123,0.34611945835056634,
 0.2082313494326361,0.6320973523561585,0.5655663360866973,0.697815360170246,
 0.7198404630871685,0.7623329005979687,0.6978137739636118,0.5791673622647971,
 0.5664260757614621,0.5399374340109312,0.5185935754206774,0.8668777139764866,
 0.46244110078633127,0.9699540562357437,0.3662097554587901,0.629932572003264,
 0.3695562734704533,0.9360050415716438,0.7357893010352349,0.8199307531446862,
 0.5186820725307286,0.8477991835076122,0.32824139945791964,
 0.022711149186986912,0.7164688962102502,0.6716282860561223,
 0.5791168849735459,0.6813289729577232,0.841054645085833,
 0.0016553404267092287,0.3920691010894971,0.7603227641973136,
 0.9968277801233422,0.7605455166109162,0.8987064472147873,0.9004435311444439,
 0.4153610191924807,0.835513821826209,0.11610958739788546,0.10018123878924834,
 0.3438739198269517,0.4493600339388193,0.07220889425930443,0.6047640213700592,
 0.37550392120801446,0.2621352993239431,0.5404060598434878,0.9850101701507484,
 0.42859179177153406,0.8490900777281526,0.6703425889350564,0.8866790788289891,
 0.6369921262918178,0.502944564033862,0.16099088860217048,0.719030411302154,
 0.26906631362110134,0.5006626657457801,0.9646968009489308,0.7579588876970836,
 0.6913393926740421,0.3935133497531901,0.581293123101912,0.24670033808260283,
 0.4960836208056605,0.07935955535162165,0.5434130599077357,0.3444553348716669,
 0.4149651735362392,0.90618505794657,0.9411567546856022,0.428882739109852,
 0.5349185485351041,0.7464266560971603,0.6734925447187511,0.03664879999757553,
 0.12281418067153527,0.48425713962056005,0.685465925180506,0.5036503227309093,
 0.9957693827416352,0.6925977789279552,0.45565074576997666,0.2549043217544177,
 0.1670716155570673,0.7676113307995691,0.24380648817472805,
 0.44641769501835415,0.24959329739292702,0.37617146482682384,
 0.5049437798302785,0.4394040183297696,0.41092207017332727,0.7253528176042122,
 0.741305372988586,0.8173241864100115,0.9422073102552108,0.8616066404924394,
 0.03068703212789048,0.06052084914104383,0.6565696036503315,
 0.17396817944587328,0.7156976651266782,0.6525720205306804,
 0.0029628749968536594,0.08815031715663191,0.47096508086414557,
 0.05988299123518159,0.7691890966937903,0.9861275262096618,
 0.36315275876458386,0.8367514402492788,0.9888543169445015,
 0.013034016245387914,0.7409628355463922,0.28654952168478354,
 0.04201882368990062,0.09830846226650258,0.4700414682326316,
 0.7037849091164939,0.3132803118158961,0.884340683179954,0.11386060249931307,
 0.9243785262809854,0.6008827536561251,0.5368252190016567,0.37311299537749865,
 0.8550852407529365,0.8616799484010659,0.7840509035627837,
 0.008042654689214834,0.2597192984335056,0.6060381964336538,
 0.6413524631405861,0.026040319292650604,0.2861556414449644,
 0.2961690571081512,0.6307634487985881,0.20164536263080768,0.4732461085619919,
 0.1726606348953994,0.6510109940737303,0.0025183021964713115,
 0.3020180803385657,0.3581664019631242,0.7952996753910538,0.6673481223630655,
 0.8191464334818791,0.6484445993139246,0.763741237841373,0.4572638590968918,
 0.7691788286022355,0.8984242026336313,0.18010959472724364,0.7690624835784965,
 0.5884744087319512,0.3597645256870652,0.15176775058852288,0.5834408048316801,
 0.4741470076740165,0.8235264492309661,0.25762441817201853,0.610793582042343,
 0.3629399107789013,0.012809195149649732,0.5177017696807207,
 0.45012363832418667,0.06346171044144178,0.4833918080270083,0.829979634430273,
 0.37242087798993806,0.3878987034889908,0.3511757552777278,0.766039830403859,
 0.38968779021338484,0.673622313409744,0.8162555792280854,0.15991104143046853,
 0.5997001658354663,0.577820377515633,0.9349084854329112,0.96916761069598,
 0.15813314344099005,0.43349461704156644,0.4236827014880036,
 0.5495846482046266,0.7630883243748737,0.0649044738873017,0.7982125599386691,
 0.42661997387243633,0.3422916592745633,0.9104103106551836,0.6555324996540949,
 0.611514144059764,0.874691445413822,0.21477306236292726,0.9395676351251449,
 0.1395740024870411,0.2829401292709711,0.7861278553047437,0.861147308528281,
 0.7309330796221924,0.862876738139001,0.31240335706282085,0.8166173102348347,
 0.8659017967628662,0.8700294493189935,0.5330305271422714,0.39912299936433837,
 0.18308726874296166,0.88292213402903,0.10252945859136509,0.6098070922308573,
 0.023600977735283135,0.731605149610385,0.7145076780181913,
 0.05113415619273587,0.8054627085359273,0.873337904300962,0.9313137354775157,
 0.3037617401760969,0.44586957291437357,0.48173511724335727,
 0.9387844592805461,0.9865142502108074,0.5645251622601135,0.8927927887418298,
 0.5893319910916894,0.9994077335048797,0.417515517696214,0.7520784508611391,
 0.39176258343881454,0.6030523058074633,0.12542343876179685,
 0.13051538783468886,0.10663713228087479,0.05088173250071104,
 0.6646785831620707,0.46116296620076647,0.3860338501957088,0.6747164386169852,
 0.7127146348640803,0.5376845875610162,0.8867647388773465,0.9207625189194031,
 0.9651075324496137,0.2964964073470524,0.3953637087004408,0.7416420986352346,
 0.6768684626247812,0.38343941711143414,0.917253811982833,0.8910279631157333,
 0.7901445024782731,0.7867766889009491,0.5865812397440658,0.2698538476519886,
 0.4827104422853271,0.4792559918312145,0.04076332448259856,
 0.39466073632032495,0.7481486174996854,0.7775055474571841,0.4939839219761649,
 0.9905659710918361,0.007837590779167125,0.6697541439955053,
 0.6431233391903217,0.31408282004709365,0.6269368114990285,
 0.21000211671580016,0.6694173982726616,0.823519935029121,0.07438906605338702,
 0.9680432859331976,0.2719247252800381,0.34330497477676725,0.3362285200906152,
 0.8158332139462915,0.9598292452030998,0.5053690914187446,0.9317642563950633,
 0.9060229806533716,0.9429623595659593,0.5531707942132712,0.13563285123058932,
 0.8077373472687798,0.8777626978160706,0.5816922299938673,0.8245302617015935,
 0.795485007998642,0.2726241926064723,0.014864170807595722,0.6249000169954381,
 0.28503085661554595,0.20046318252766648,0.7466544910726748,
 0.24663185175453173,0.6314022767046199,0.27648603684554607,0.775485571186597,
 0.6431835037288188,0.6732216877336984,0.32131320716475265,
 0.011523996887766175,0.6553379247286294,0.4272746888531991,
 0.08516269313601765,0.9789358046140152,0.9601131169853834,
 0.43264817750717666,0.8750378943594388,0.502461616842145,0.6318399858014823,
 0.5939493240904499,0.778289205022619,0.08649369072090085,0.13575861427161984,
 0.4097375339881154,0.36546374556951555,0.07635157791952174,
 0.24233149671736798,0.16169001211825262,0.46236634022771717,
 0.6307603527483231,0.25102313509142804,0.2975554033087343,0.3011201463608145,
 0.15963667329670161,0.3106777824135729,0.10530226320079783,
 0.04677752581717398,0.5008945641821906,0.7414461829489866,0.6431735656317462,
 0.4579311172112889,0.17176547764064232,0.1928282196529928,
 0.39803122506681565,0.8805349345499754,0.7744607775464906,0.9496133213364167,
 0.31673863437603145,0.3045680691567543,0.3775057767616252,0.0093771029562979,
 0.7321012048531435,0.8676014404074877,0.3321849342837703,0.49300659417984904,
 0.8213762008065375,0.5625146698728649,0.9303984941241441,0.43128712790572266,
 0.5557939798456611,0.1951846577901447,0.7969731622296723,0.5789324733506104,
 0.39453010145065925,0.3627987877299973,0.3309554166464207,0.3387160502525979,
 0.46770192527276533,0.30612728835150216,0.3044057488492995,
 0.06913727179730511,0.9392098499556292,0.02471817525290454,
 0.6101867003289292,0.32238952292492007,0.8444335438503975,
 0.30635103502203276,0.5315563221061494,0.8237260767644755,0.0388043381222275,
 0.36996605676985184,0.9515380722876295,0.3230565101958862,
 0.40231086807204397,0.008262085092405447,0.8015214280186221,
 0.2470049409889724,0.9684578511245715,0.22761182992072038,0.7280955840553254,
 0.7333250639257283,0.30988546097265846,0.42089424796486663,
 0.7738113284925926,0.7762330521838628,0.3527832170450721,0.7720524917504069,
 0.8491191065559528,0.20043282930097717,0.47858238812334786,
 0.7418217093496966,0.9744703882498766,0.3505857171917788,0.3955630908988651,
 0.9635459597222964,0.5698699236435923,0.7937865957912141,0.32807543897120195,
 0.17149377257511667,0.9448219390130878,0.6511672865787044,0.9804385633979017,
 0.7984067289453264,0.5860534723702098,0.9416879783313623,0.47068656096961725,
 0.4176524165062703,0.3226166424058793,0.873502229512974,0.9137434129606516,
 0.06720024453941731,0.20739641274036125,0.0746386149366729,
 0.6511046679388557,0.6894185142440592,0.5966201673370682,0.698135958194527,
 0.2051312154827718,0.096091089123441,0.6697971909922668,0.05409245024053844,
 0.05865948293431,0.8135173599358774,0.6529279817272681,0.8515950716981084,
 0.47800396440847104,0.5781527168168221,0.26569056676883196,
 0.0719275365360389,0.7352376532411627,0.67222957992213,0.47769754846578893,
 0.13272846413061223,0.7787417464352159,0.4398731557255484,
 0.15282164218678917,0.30921472981520004,0.6060153511023206,
 0.9367524875202323,0.9013933746605715,0.42752922181132935,
 0.41172239918457887,0.309115395504466,0.9585498999636218,0.8438992178671288,
 0.6238978182356406,0.24948224070118918,0.4649702097469779,0.6489519724607093,
 0.2989898907291818,0.44238678043214985,0.6286885661444388,
 0.21014189739631084,0.5024385962376792,0.7384696051478883,0.4316559379982625,
 0.30322520666096986,0.5308777630251351,0.13339014898762502,
 0.19326346658113802,0.48145969684171186,0.17561355347970486,
 0.6155995837530536,0.011637759093387268,0.31938122554104553,
 0.5483106790454982,0.3216946473419797,0.38901638380318027,
 0.49898243350786387,0.8256364891613402,0.10013233863879312,
 0.39213426950187014,0.9451035422075634,0.5607097163746613,0.5624100835096715,
 0.8977918118624019,0.7074637592669624,0.6037820649842485,0.561916869966022,
 0.5830541102010199,0.8048423653391555,0.03129431258712645,0.6829357420662642,
 0.5397826175065317,0.5455897645853115,0.9722159585669569,0.5183063628255895,
 0.5531803590885486,0.24883097646306895,0.6589614401082721,0.6296033277577692,
 0.3400943589769757,0.6249598940778743,0.09601259721481159,0.4051245687229553,
 0.9278048527901652,0.24589168437658226,0.1006503597472167,
 0.42982625155012133,0.809474325808315,0.07000924170632039,0.6508600666131332,
 0.43138771353944483,0.9471387871365193,0.45628769581600614,
 0.16765237134687505,0.2892959252563575,0.511649112877857,0.7272784681541518,
 0.348879820615145,0.0018817935500688998,0.1420928904811808,
 0.4385307477360829,0.0805009553384538,0.9551455115544295,0.15919586487515203,
 0.26384600599139174,0.6328076098049348,0.2838705247345732,0.746488976855065,
 0.6535090886881703,0.5816212718355334,0.5406835129079312,0.8233589869691086,
 0.1498957149522654,0.9839068049718933,0.3151101452819782,0.9380719765521892,
 0.20664282968554348,0.8764313674291667,0.5777678312975901,0.7148192336148919,
 0.6371495281687813,0.8778774262218421,0.6756888261438062,0.18685119490478663,
 0.8057592918834403,0.9589633768790149,0.46386383461124703,0.4578785348487331,
 0.2541625580825566,0.4942531735815474,0.5182452954968292,0.750656944011475,
 0.5291637094046968,0.14194148348766256,0.1328771918234103,
 0.25161373967457135,0.4712233134883871,0.31869500337830714,0.892088970020013,
 0.8656120277170192,0.41367570023267475,0.7941949676186486,0.6208389878525965,
 0.09296274275244198,0.22313016444999945,0.027859256133933252,
 0.8587442179657422,0.7244088307558423,0.9024447141599,0.12050801231731767,
 0.09728450017191737,0.7858521747886701,0.034246819678144025,
 0.8442117777973042,0.9149111660423135,0.20072695256216533,
 0.017790479107951396,0.7666295986649693,0.28801782732199976,0.71484968988483,
 0.7171599593844185,0.723025117888269,0.6837685616175628,0.1421418687019076,
 0.13930447842275628,0.21723829494089775,0.2929024727386358,
 0.14139070135836362,0.6507323280174983,0.6661622886084283,0.756517243694727,
 0.5912048405231979,0.02852946431906389,0.22114102694840043,
 0.20380969747221467,0.44902661516859077,0.060058566090103294,
 0.6131450272127383,0.402831453018314,0.07703841526343314,0.31572159178334913,
 0.4778869506029364,0.03771567292597999,0.10335458502870809,
 0.5343089577518336,0.43858854632810496,0.04580667807130523,
 0.16171634857081718,0.5702325194456664,0.5051779244364568,0.4924325618558446,
 0.5148765679558144,0.8226752970788823,0.4561602276045996,0.5644100433344195,
 0.14323716948854837,0.14694320527306992,0.3188820455673821,
 0.8526447467343115,0.6338290200507168,0.9884723376486748,0.46574648756285864,
 0.43984188927636325,0.8428953586789304,0.9629614267069821,0.964389994210448,
 0.768030917016323,0.5995534202632161,0.6804240086729542,0.3748227664689773,
 0.8827151449800873,0.388007138071204,0.5250949980977101,0.3121501604724457,
 0.2065001975236127,0.47282496879178304,0.26780036259289797,
 0.6285764506709682,0.19140405852906128,0.06331444853965507,0.775821520607956,
 0.6835826684777356,0.7575925099011194,0.25311296121473603,0.8675439675753804,
 0.23408335003761138,0.4008276070163558,0.25649183596434666,
 0.1469374392266427,0.8543938457806317,0.33992881380944706,
 0.19776611661610866,0.5489252980235664,0.344880115440222,0.43404554042256027,
 0.8315274558814625,0.25172188367035775,0.7367994262980235,0.642610422240342,
 0.6245750986478469,0.9214842606914407,0.46644683727210634,0.5895696895826932,
 0.5917917299143337,0.332582355695457,0.41338821157826566,0.49644278312286183,
 0.9624034609813175,0.5655208491237487,0.8862243529140322,0.4827673037416863,
 0.4235163384671006,0.8625046604060985,0.7268862646263836,0.04714428924787051,
 0.12782181578575014,0.8652766491472362,0.005165398014044809,
 0.6870784811178181,0.9292217103292015,0.5125831475214806,0.9275876380436655,
 0.5151763918145795,0.48497916620637604,0.16395102526409544,
 0.3280189632871231,0.14726372699812096,0.9075474260665596,0.2857808147137758,
 0.95358431226809,0.9692363704745699,0.8710858119437704,0.23472346910348696,
 0.4547310973077956,0.6902760970469458,0.47345307945903115,
 0.003586291231703065,0.7563624446772176,0.09166039464229436,
 0.5656362495150259,0.8272811941959335,0.4811384320172505,0.8651841317420359,
 0.03571279964482632,0.8285681559922856,0.06948110344303415,
 0.07515322289574633,0.7511874487786987,0.8265940790496912,
 0.02726979485291081,0.35286354016493715,0.7733327187169659,
 0.1251489975866401,0.6805333333122187,0.07101313667270204,0.3179527525498802,
 0.6137932834894475,0.9806718687602258,0.7749326959267071,0.956547103294089,
 0.22953644027306197,0.5597796907288736,0.47919963007418753,
 0.40989235166022064,0.6619962403244267,0.17486973998646627,
 0.6147460641500668,0.3646880893938723,0.3014719828735095,0.7447819048717632,
 0.5697281226139286,0.06877332349779175,0.31737464327790854,
 0.7796718512882406,0.9595353096026715,0.3964058994345421,0.5984235329095816,
 0.24002664324965395,0.47310467137978984,0.5914203845326466,
 0.6937205265033558,0.5040230184934199,0.05512533989463586,0.1099200736090511,
 0.9939308434609532,0.32548122310418925,0.5938934185392983,0.5771350793782727,
 0.6856049277539166,0.40550668690194824,0.45731217994202555,
 0.023299006264484445,0.8980072691196219,0.8429588284232649,
 0.48898802690421017,0.9924200318576976,0.7626770036924563,0.1669794943669134,
 0.05188787972602671,0.454324199457953,0.6137052122802285,0.553529983490168,
 0.8131071973686206,0.3537594342739794,0.7009156304659083,0.7405661282839442,
 0.36590717357294567,0.29225620139569386,0.007297583197649393,
 0.19993314283816832,0.17430724635246397,0.5397051842204295,
 0.6179014512650235,0.07873226897662344,0.4075990743196475,0.6197696160569462,
 0.7746526883613908,0.023095686826576145,0.679538763539707,
 0.21985078077241993,0.9269844670210318,0.713563232246874,0.8070814387763418,
 0.8232168319277319,0.5605935055273046,0.8172900711873499,0.49980408851523983,
 0.572822103707209,0.47999098990752465,0.03227162531794714,0.8949593702587424,
 0.4755556411075126,0.9162363875855881,0.5344080549684789,0.6061492865896366,
 0.9037870048991062,0.9805504840383019,0.1539629729300943,0.5619080389430613,
 0.45588150910166814,0.014609292322458445,0.9740659291788573,
 0.27773589453878833,0.5831293247183298,0.9845998590572161,0.711878710769777,
 0.5387343487772464,0.3854648098959177,0.31238228474288054,
 0.061184893109001215,0.3076631961081401,0.6478551729788495,
 0.4743167721709428,0.8071402492342994,0.5338529434120837,0.4648111496027174,
 0.08924141081538617,0.7138969996738187,0.3798998015767421,0.9177650924318596,
 0.47286457077292365,0.7111318227495148,0.5589412895818693,
 0.48991295062265716,0.7999604820168364,0.25660311384184065,0.748933031757478,
 0.28406445469087316,0.3636426960602541,0.5286566749774257,0.9616328369325584,
 0.6315330717414047,0.7441453927824766,0.5639410625789698,0.3820951868559195,
 0.7863161797964668,0.80613370194608,0.42055519557747,0.6003520636803403,
 0.8620800027978526,0.8231495438640175,0.8617442449608907,0.48745250701988063,
 0.5696041990713276,0.5607937646923691,0.3287346335941854,0.8665782677456197,
 0.25352342185841215,0.5658754107735308,0.35398591029593396,
 0.43384574541306864,0.7758892887926261,0.6277446402477217,0.6056870770655162,
 0.7910890128545662,0.1729277050058724,0.27543993583974236,0.5811978459877053,
 0.9826578555896781,0.4020349835186481,0.6755460976149791,0.0828791579369913,
 0.7158329208095835,0.3660079500031952,0.4267182391194638,0.11402895418556713,
 0.6335025913023333,0.7777741609700252,0.2965424413171366,0.9703877418332021,
 0.27251788270167243,0.3263259991768783,0.14099900813652,0.7103371731777343,
 0.5234143196790749,0.7635349382712925,0.033323115842324835,
 0.5890798320235643,0.6112106617567941,0.09591893959088937,0.0859963438859016,
 0.04515929521208495,0.5221852133413033,0.642330882229542,0.03899917878097281,
 0.5216236891775098,0.9601718242867248,0.7059485542957431,0.46103729568526497,
 0.19904748510001546,0.44253096052113716,0.6503335357547981,
 0.14797727487197088,0.5950770148442726,0.8272901591748877,0.946772078311301,
 0.4260570433511909,0.14384271860053144,0.8278885698534252,0.5818799386786243,
 0.07591539483088461,0.5688816031449833,0.016882707058582302,
 0.8322412228525464,0.6620033062348398,0.14047701183699712,0.5211187796892551,
 0.34504338117273803,0.32652161054437956,0.17906622780441417,
 0.1329863578501529,0.8230421742320619,0.1507711743464608,0.2672003367860799,
 0.9006393316104789,0.6446554792189694,0.08453762811939802,
 0.25353578164893187,0.21956928696214084,0.05112236377578161,
 0.9356383885295025,0.9847707369520022,0.7357531162652601,0.5179764890044543,
 0.012096469727644443,0.1664695107370291,0.09652194834053351,
 0.3660615205867537,0.7954910370979476,0.03738659361669594,0.6506956183514369,
 0.2946789520163814,0.8605172166509716,0.15275161031896367,0.5150794206480515,
 0.6860634698132057,0.6375029845256603,0.352070208561891,0.15458741297179746,
 0.8728089814587079,0.7448089470740175,0.9254236176532361,0.7258305166834944,
 0.3143491129436572,0.5671269805217669,0.9107247770714434,0.6648662636081188,
 0.8930363654815576,0.6256660164283604,0.3279663281686228,0.11321757577304292,
 0.14333694344213954,0.14662848414870155,0.2497952784092271,0.868533263501555,
 0.8906518411256407,0.6138165823243892,0.7006705755049897,0.25714741853748435,
 0.1038805510882469,0.06249287580863794,0.8599781831878675,0.6308047629798238,
 0.556822115976395,0.08569018241195159,0.30094788499820835,0.3571814524656697,
 0.09776443594387141,0.5635192824960754,0.5534437653352056,
 0.38722064597295136,0.055801510372465724,0.06254784263162327,
 0.8370298250133374,0.021026364711789758,0.15758379589743288,
 0.6927808106296165,0.0394159847047566,0.8240550091462244,0.05180241482636372,
 0.18911962069784827,0.3868674067822957,0.8754595795044079,
 0.16840745336056873,0.45374817805615875,0.7790310213522368,
 0.3217289225244848,0.40419739887314576,0.9077269128117771,
 0.07051626348324613,0.7782895906980853,0.7788947843808214,0.4547519620014834,
 0.7190626336797271,0.34605769330932523,0.10670153686712858,
 0.4593480957870181,0.4616471747829616,0.09886072428390091,0.3080109280278531,
 0.8071882936827335,0.47806315148989864,0.40355102888220584,
 0.6195378923108297,0.903100290190734,0.9217495731672961,0.09644321116860188,
 0.9304638398902937,0.9347319002952544,0.13330543954968666,0.8929218340795315,
 0.07152160288605869,0.9920567104691637,0.8255561561031757,0.4407455897924988,
 0.7638647495452559,0.026264282439882547,0.22279215473775393,
 0.16076497178167826,0.02656125250474095,0.5557834169005175,
 0.3789051395159717,0.43092287687480335,0.4105263380384184,0.8288869732829521,
 0.8458913684084821,0.6281945509492307,0.9851070736860259,0.11893893271535672,
 0.058952379607277616,0.9024198619856789,0.9972628947852339,
 0.09251730813521708,0.7524752497780669,0.4139060008918375,0.7590805243527836,
 0.8969472592341198,0.08644310164942581,0.8776970704266376,0.611121554205193,
 0.7378530312933194,0.7173729923129295,0.5620854509693295,0.5637724523871224,
 0.8446774909078988,0.4994093940437433,0.6163870328412213,0.2733780469230551,
 0.23464472439419604,0.09139927061012831,0.49374095117027483,
 0.06823970748547814,0.6644531106794693,0.40790338878775745,
 0.1714746343333967,0.42352604527070337,0.9936137946269246,0.1833075395648146,
 0.663627761699458,0.2716607175326896,0.12784978751784293,0.8040664051955696,
 0.9838611829749107,0.3932471840573354,0.7683735871649359,0.8518849505558852,
 0.5846142991643495,0.8330317730053091,0.8999011738500481,0.9946079909575095,
 0.8156998628506472,0.3749132647611685,0.044547537272440096,
 0.4503944213961827,0.4706128714333213,0.41582870920408765,0.8522943423537281,
 0.24962746277089443,0.12460726823379953,0.14796479889057235,
 0.7863875643125057,0.6002021224762385,0.5675022237417777,0.4292965880699908,
 0.7135496899991215,0.4000101351242469,0.36876267619043,0.4774065480101928,
 0.2750827726943569,0.7001268845695702,0.7011201236790416,0.8616035437442601,
 0.19342177544376526,0.7958813239010103,0.4507682176117783,0.740885466543678,
 0.555976174190079,0.2484168134074186,0.6887488545427587,0.3351216596219152,
 0.8006217799419311,0.7927957953545571,0.030578591593233617,
 0.42976428089630314,0.7358782941696934,0.36538787554405994,
 0.5776191283864132,0.14091508435119804,0.11398761594163886,
 0.5237198691144412,0.1996278158378063,0.9001786431150575,0.22178208718099146,
 0.28210678679882895,0.8856791024607014,0.10597198634010685,
 0.6032840675586909,0.5894007701853077,0.193321282401402,0.6913317576725498,
 0.8179367339237334,0.17239237935299734,0.25035391151378894,
 0.2558049032559836,0.2665445190671485,0.31785748432234373,
 0.35673024139986054,0.19539296244802773,0.24902701859903453,
 0.3409472832354197,0.4810094387825855,0.4705976482703279,0.05578438307219391,
 0.8446941978452699,0.996893805396924,0.4665304414405309,0.5412591514886681,
 0.7537658353230324,0.22472007469003163,0.03491188681225654,
 0.5939676936678948,0.07631970377414454,0.8016609837466666,0.7228213437631151,
 0.704672001990951,0.4315984024687476,0.7974055184661294,0.945424651431928,
 0.6628070247238499,0.2570141405714351,0.15686665565828872,
 0.37553967090983076,0.22132799747745435,0.6778554264922539,
 0.45412307823948384,0.45376358653315507,0.5036443934586035,
 0.46895034104342126,0.10866586450297278,0.2593989540721091,
 0.7127163716786413,0.037506135057697065,0.25106141513139535,
 0.8215218911056243,0.13461781198278056,0.9933111827460815,0.7858333881657803,
 0.21917785676966273,0.01943509977241853,0.11476150895544301,
 0.22764201319724098,0.6144601454679559,0.7287344566019247,0.5477718766204795,
 0.8182838167198628,0.3071002403975822,0.7411692446456181,0.10622177065180338,
 0.02407157511731739,0.5883112196320914,0.8103683970927946,0.6605564864312696,
 0.28640893331847006,0.21900254469339053,0.08626772737848398,
 0.8480579083669727,0.7682930402916704,0.7098337078883343,0.22454103670906855,
 0.3766997932243168,0.3852407734680715,0.7889528384887622,0.1760857150267301,
 0.15745442008014,0.07182754860583751,0.4202801246015021,0.7707728444786961,
 0.5453164772142816,0.26923879872658274,0.4397764543487671,0.3234347802160049,
 0.46938570607124264,0.563934285872673,0.43725869743653367,0.5631243812241601,
 0.9419598211361995,0.5940300389990166,0.8031358739982808,0.4500779046054709,
 0.8114038075905835,0.66835276913798,0.060137174715581665,0.637079372077681,
 0.04847241823124071,0.351797496773279,0.016352603194862425,
 0.6168575619036939,0.3387270252058132,0.04568480611840142,0.8231183508818771,
 0.540617030040651,0.5027764627497325,0.9457395244730045,0.3173998796171008,
 0.07906866598998619,0.6180075468121646,0.8674629813284936,0.90223720831898,
 0.9554682213122565,0.11317100122125423,0.5511665182678649,0.3524231361833916,
 0.7209992913707216,0.22385979144794899,0.7786470641799605,0.3605172273658125,
 0.9199702973066802,0.6270858847343994,0.80400804925748,0.49919860607821753,
 0.7633634902966728,0.18875315781930002,0.201336056075468,0.47086641545898145,
 0.9438618372467121,0.40046769843920915,0.23151074334300237,
 0.43622734702832866,0.13731056669573505,0.023281416048459014,
 4.628742913693751e-4,0.5191121561478675,0.657408413314422,
 0.12247427904208552,0.4569520919759429,0.8475464259582575,0.6717644848578849,
 0.7477967908996193,0.2942584026865772,0.3745375175224708,0.5175466582530776,
 0.5083931266505258,0.2649556373194618,0.2895951359419664,0.31228454140135187,
 0.7442380081997089,0.699684361974529,0.803652004220598,0.5419924979101379,
 0.9570417631872288,0.860068278177427,0.8795251249801951,0.6592331392123072,
 0.6966746198498315,0.20850756189871944,0.26941088245457145,
 0.11719606687366291,0.04632635537474272,0.6969439136226887,
 0.3023779961298265,0.9928411798727821,0.9256605186637451,
 0.017407522439552126,0.5931873571796772,0.5448585770796021,
 0.20626964085036548,0.13205727858006933,0.4965675075706417,
 0.18316207458647826,0.9208457173418493,0.4162866743224568,
 0.28734434951213383,0.0030908085499002924,0.031859157758317824,
 0.5390782651561457,0.321107964652377,0.6744304427649981,0.9814920964459187,
 0.45989435036236437,0.13641072384141262,0.4686881792770583,
 0.4648601615211979,0.7872882907516922,0.4572383857886553,0.6981721394330216,
 0.5998265644483416,0.5927930426645864,0.03793710209888246,
 0.023367792336716597,0.6565488855521221,0.9058736998356677,
 0.6071913510181568,0.12607967207030912,0.1354967907328689,
 0.23301339945873512,0.12083288129871228,0.8635991324288966,
 0.1607529387758827,0.24336356721654107,0.8964774993749636,0.8722500096468174,
 0.09518840003229012,0.45614379402602656,0.4854478196631582,
 0.36369280455169084,0.26284779670012315,0.42876331288501857,
 0.036012939394782895,0.6180042491060377,0.08865695718400024,
 0.6441898017151322,0.7120474204709673,0.17834377976170623,0.959038435994628,
 0.8498483402658471,0.2857217770259003,0.5904746588332005,0.8800248528052049,
 0.9593823576413836,0.3175898694609376,0.6017231769406275,0.16963661337038993,
 0.09570223885849383,0.18483530158031936,0.9630867056824379,
 0.8369284139249258,0.9045279645372617,0.7232633670105976,0.2981418812693025,
 0.35115952163347175,0.8430673012693827,0.8503543124258877,0.5017073666151635,
 0.24622584291839267,0.7100961908371944,0.3352136154723926,
 0.33089275744050395,0.5374141375355822,0.4223017002640228,0.7222461560830409,
 0.945072402042211,0.7749283246618203,0.1526386417700526,0.22989994575809292,
 0.3059853716757811,0.7658650609351886,0.3335258584040278,0.1338236075805399,
 0.5010226350886285,0.017949556679203282,0.30146536325102335,
 0.45994035693881685,0.46305691999766985,0.9234759042186476,
 0.5269007506262204,0.03835137824606538,0.26551376124878245,
 0.7458286455409808,0.6545435281351026,0.7396313702219828,0.5427174193965442,
 0.5666749618062932,0.5608165810644168,0.6958857662866973,0.9818761930660639,
 0.47833220667987286,0.5588819326650796,0.23275495802931434,
 0.04681675600238411,0.9289787641360755,0.024089291297481386,
 0.6624102645962692,0.46886489029677925,0.5740442404346925,0.9427378565480142,
 0.11815433222447547,0.7712668321275588,0.9764825834988461,
 0.003998628854466668,0.4205338079135027,0.37416966790841666,
 0.33837912211705046,0.4826968128893262,0.997221562527659,0.8036355163203128,
 0.07576954672498404,0.7037287348142157,0.5166924873480185,0.879805951424068,
 0.46755041797073804,0.9272956098010601,0.09143794813286976,
 0.8183907866888658,0.8622275951015296,0.8994634381393402,0.08189265778622634,
 0.3384083475421421,0.13475034992059753,0.8184029609723247,0.9565678343794418,
 0.481488601957332,0.417580388619472,0.5497656940436596,0.7713856561029173,
 0.007286924035060038,0.8624048378498955,0.7019354484524786,
 0.4060082602454992,0.16864359723182099,0.12020912692798402,
 0.6212972190312509,0.6516614231219987,0.08542750079330663,
 0.25388933199712294,0.600024249485402,0.5209476526319203,0.43018628060865294,
 0.5918900878724114,0.681831064490478,0.6991619419281668,0.2884659937630085,
 0.5881378905912686,0.8975331955452567,0.3236652313954196,0.5812412131893961,
 0.6990269322044721,0.348831057448433,0.17982091543108503,0.9226687881892661,
 0.21828858260472783,0.6673267296632613,0.7554937378610173,0.918690756343575,
 0.24778664689568597,0.648739051011847,0.19685168540568287,
 0.42736118751053376,0.11814918045647238,0.577714744058292,0.8522840354510341,
 0.5509294157335065,0.8708900199245686,0.028669656085049122,
 0.3259012389256525,0.8225640731213129,0.5658169784276803,0.9633964699307791,
 0.6006537968486054,0.3136664792373264,0.9355623256009256,0.9747060869696663,
 0.7847641812350633,0.17459427868150101,0.9591722750725912,0.925040206797403,
 0.11008132362323308,0.2438289734172221,0.3206856166695037,
 0.37835040462089775,0.8533635975928816,0.7864292552583645,0.5460898765866432,
 0.19370273888599965,0.8081424159018783,0.12218979520381679,
 0.42181815982187887,0.6731307079628192,0.8980215933495125,0.8050546805078245,
 0.3537364990603916,0.8473250594052448,0.12083851485122399,0.5516354381805006,
 0.6136411497112149,0.757610379747651,0.850407889895674,0.8729546292438083,
 0.20812355198713894,0.010035651059726147,0.3326367629252377,
 0.6330768042560729,0.15411667230315684,0.8961931348709036,0.5846923007764296,
 0.2956402524038918,0.16934464070354416,0.8575444673380421,0.6429172606723239,
 0.2089041162443006,0.021239948003108022,0.6659369990922711,
 0.4953974542439371,0.25939356231927146,0.7197956484315702,0.2878157888596464,
 0.05020799485294214,0.48020177652054075,0.3323109553647935,
 0.5765558615576767,0.0718211527306849,0.2708411364984501,0.4766500935127035,
 0.783701197026338,0.13231336205397348,0.09504422715219207,0.7052444216452706,
 0.46970341558002493,0.925941204913265,0.8432000453678419,0.7959603863452943,
 0.881722536465202,0.03655104562857048,0.5100663719122671,
 0.013821768864833328,0.7509749447946878,0.7860580074826673,
 0.7030619423712041,0.744525902813983,0.010135570220654033,0.9997560136873463,
 0.47568095796258214,0.24094905157342805,0.7287144604855267,
 0.20767808629320716,0.13904680354084809,0.9070572218191622,
 0.02708842606496631,0.5778230089361032,0.8912622532723037,0.4550288219516364,
 0.76055353611398,0.16544978240270536,0.53600628639235,0.7432863780212561,
 0.7823215096572486,0.3058034688334682,0.8604726260938483,0.40384307122076857,
 0.5599766337054829,0.19863920508813748,0.50520843246854,0.4798290749582116,
 0.02435513885035312,0.57359183085065,0.2608652309035344,0.2981872297792517,
 0.08756914217625156,0.748652429011496,0.6945867406681145,0.2423184792490637,
 0.11821523256921362,0.07629199577369972,0.6588711889655954,0.929538120040156,
 0.4374710025730393,0.0055965574426064535,0.2767945453106444,
 0.15301248133065948,0.6641662167670905,0.8935457480178429,0.4643876798532518,
 0.9258965142498021,0.49872736840385035,0.4992194708480252,
 0.22335622340793393,0.06738924462887419,0.5434607434381584,
 0.7375068072259088,0.9812742116764526,0.49453427375210957,0.8936243357557833,
 0.22955437982209947,0.02853224944746202,0.5368320080524354,
 0.5507708172243785,0.9632694191935572,0.6618730981855926,0.5443789741246088,
 0.8901222505374146,0.8957325287336542,0.13278053064643436,0.5531086073462325,
 0.7199247485497192,0.8798197515986655,0.3936312597490431,0.25659788165935016,
 0.426742811182101,0.06254291748052077,0.4753749250185523,0.09667956351038487,
 0.8418557976955839,0.41976374802706207,0.3885303117384199,0.7685024246055547,
 0.4313822425936742,0.2067195407416732,0.33741175115873556,
 0.16756869783764294,0.36358010551447517,0.851836636560929,0.6319544641820668,
 0.5378424616780272,0.25638391678948247,0.9522384226862353,
 0.18010335854520532,0.3779378352415822,0.6291691672955715,
 0.11050219982639375,0.686031387209201,0.6713671679824529,
 0.019356656865604838,0.5165705326830431,0.05178068615773268,
 0.1253524914263935,0.5388120936053492,0.8390353450164768,0.33485526151790534,
 0.6773867455201592,0.9315901280415537,0.5943968117761064,0.0485014006028236,
 0.9846989158024211,0.8130895676805246,0.4073569661471801,
 0.003207227787125788,0.7143419657161603,0.7967997513533649,
 0.9495972764525651,0.19014883169291563,0.5055745711034922,
 0.39754457372017926,0.5879729446590137,0.2236219732054585,0.9797788284477233,
 0.8311332689069576,0.4245630725792925,0.9313200542796759,0.2112610566162041,
 0.9947290232891166,0.6799275539780453,0.18793083199692628,0.8181131185867639,
 0.18220991704824707,0.8898551494387736,0.6522770548935717,
 0.16855676675084053,0.25128769913764437,0.4349719208384599,
 0.4176666690093138,0.21900775700422448,0.26418699830789016,
 0.33392628889806086,0.08696479285800829,0.3875785933400441,
 0.08847948227150881,0.03539489313932964,0.42655524594324146,
 0.25968548075698017,0.9028826086411681,0.48484321657556373,
 0.0014091316033839618,0.6917220793705222,0.40189571878809227,
 0.01830069168491344,0.7933424241266332,0.27107269448190674,
 0.038676620535538664,0.835264659777061,0.5853114945319974,0.5101674682220325,
 0.5633627090091442,0.81863620251328,0.6614284391167982,0.12357679846958303,
 0.14643473712410138,0.004221294324444402,0.5528530472883528,
 0.5720358509295703,0.7531218308772869,0.5498870304690427,0.7786675673775818,
 0.7525403992847233,0.34337347853737343,0.7435410552672473,0.6046360598755788,
 0.7778515417845716,0.8720909407423652,0.812793875166465,0.8700339925512566,
 0.30586862494604006,0.08261763851511583,0.4082154252943835,
 0.7005502161425042,0.480700019437331,0.9535461245489316,0.3826952633007489,
 0.7679349974577807,0.8788001961215465,0.9210563381539654,0.36799520327499047,
 0.04997239472595583,0.22782399020181776,0.08813410387029574,
 0.8174640759077028,0.6257040959760511,0.6458169106611539,0.848716003588909,
 0.23697391778441546,0.3684250909434069,0.7803556067747343,
 0.18072423604714616,0.18436159043108047,0.5138179456775064,
 0.28616210787740737,0.7423828175832226,0.5347965850327812,
 0.12597601406944836,0.19907575174330683,0.16539487976145506,
 0.46079073484634614,0.2495810250433521,0.49058309427670177,
 0.9585196187764309,0.6125580650395365,0.06314672679283018,
 0.06273758148178588,0.9150016127977361,0.9991524097338934,0.815106865331667,
 0.4371309504600924,0.15213940228756972,0.38441556126656573,
 0.9338805201693754,0.40113972184865354,0.24731228576679176,
 0.2375592185307167,0.0253658013425212,0.6910897426564409,0.1991817516309674,
 0.5607489207964661,0.45280820780960074,0.6765336458159945,
 0.12567988483783377,0.3944470327353816,0.6819210834387842,
 0.49828583063846554,0.02662969588044939,0.4715564123682894,
 0.7252417148170929,0.16646656633751933,0.6113227225119073,0.9426263477855739,
 0.2260129770846041,0.7407366263481568,0.32277200995552824,0.760089489964649,
 0.6230055825935648,0.1005629278462562,0.7389884783776856,0.2991493878075232,
 0.44486452732798043,0.8381913248120789,0.9106047146822964,0.4093926769237227,
 0.844340941053856,0.666741350674023,0.10582547349835647,0.10164938015272362,
 0.4696716514966527,0.7294150918282831,0.6255096024564397,0.4684380255940954,
 0.051651613831497634,0.7174237308659184,0.3324751580815548,
 0.6350674489147939,0.8102682315219389,0.8754592340879634,0.2708842347814918,
 0.9887163061111418,0.24284735042833117,0.37557995601554184,
 0.4460889975096729,0.2009455158271981,0.7961879010001417,0.33905294502704353,
 0.3112827353496619,0.1102191500644163,0.2409312454838195,0.5862194784349265,
 0.33899150436904013,0.2480143555634855,0.6570725726735062,0.2670281379425806,
 0.16656425776488915,0.8432897532723121,0.5409816284190676,0.8460638292322781,
 0.16120807060023123,0.6736391304171071,0.2247019439454323,0.8654030071685345,
 0.15289014397361078,0.5666691772201903,0.48016741488823134,
 0.9816572115496455,0.7324079908132458,0.11754152929494621,
 0.027139900937854455,0.245260003919072,0.1629748735022829,0.64781131243526,
 0.05352012185582977,0.35997669463088355,0.5698922478550801,0.764026998185434,
 0.03739448598427442,0.6861113951163653,0.4207746699025684,0.783743980856303,
 0.0199201810663997,0.8809610607040721,0.6386225660342862,0.9417314236522195,
 0.26709546496151737,0.26817557300108263,0.24593879102172822,
 0.843565694501958,0.2225268391497831,0.21984130605868712,0.1755169675365298,
 0.2015545850163103,0.9907365858886805,0.32319983336474556,0.9869340463213395,
 0.5926088945909467,0.9779513536475346,0.13826321133786523,0.7816298595458584,
 0.8913099046435374,0.8283676197384126,0.8556990318216094,0.8700190132564795,
 0.39027285093841224,0.25136002130770996,0.9542355989971704,
 0.5091103956340258,0.007648349815303601,0.35400679382342704,
 0.9706975827931299,0.24332173920396838,0.5793017307120334,
 0.45005713624025145,0.526516274781745,0.8497820757189095,0.5580722014462307,
 0.7747388975072834,0.2650153115139976,0.6923825304396785,0.09143677013989182,
 0.34757290684199293,0.8618510176557811,0.025157813974802212,
 0.15290183294758541,0.29287055464094536,0.16987647534380412,
 0.04010446623473596,0.5305262855931058,0.13955217593252622,
 0.0866608724764617,0.09079448725727546,0.012318429722951385,
 0.12360799507672082,0.17322009980643926,0.6375802897202407,
 0.16200273323849212,0.27532222779299453,0.05584222517755788,
 0.4213279363587159,0.9702482142762044,0.06683237720754731,
 0.41938867968199867,0.7242171906760525,0.7687490893956331,0.9623335930404797,
 0.479645222900448,0.6484648678946277,0.6858376089270091,0.4837752976912516,
 0.02705146017062776,0.047254022021898545,0.6543537385147906,
 0.13925465923541358,0.46023520503719806,0.41264488746659533,
 0.7532736230148397,0.31825822993457764,0.8686687589831529,
 0.33772322920033104,0.5288746190333652,0.464404481244558,0.8574132267514027,
 0.7963418266424869,0.10589634466974607,0.3120637076059114,0.9808400998726594,
 0.46353707028220625,0.3715147393194992,0.756373477760272,0.26697715642828546,
 0.957402347138354,0.31757564218500267,0.9014846819726487,0.6299685618949744,
 0.5845472684320481,0.23807735177115708,0.7912666065532349,0.5451906895052246,
 0.9409454259416479,0.6679712287787869,0.72843600420908,0.6228509576296264,
 0.2923361393818824,0.45704358152076585,0.39694897580476063,0.326723859202122,
 0.04569226115395697,0.510067445492543,0.5738243216667174,0.9579224457596258,
 0.01541581041180784,0.4119227129727697,0.9002627847198631,0.9451387506050098,
 0.9672699196867653,0.29346467417315447,0.7434547553591332,0.4639811538163866,
 0.9154462755396588,0.09183549601442076,0.5463120437984698,0.1729762990047765,
 0.1651810649174753,0.7746131181805913,0.9887197565934692,0.5444486383115156,
 0.7893431955350129,0.877358590472372,0.18170637222902997,0.34518694381497594,
 0.17313514957891663,0.23910234281800768,0.8543542972446782,
 0.5238604870229557,0.3493669512530758,0.40663552247734314,
 0.24412136952420904,0.459585312146225,0.3049332328400294,0.9564814969852105,
 0.020619719621700616,0.9645129787409504,0.3257650318200749,
 0.5778075545610666,0.0640005799457356,0.08373981280300691,0.9143380192844149,
 0.6066743750754553,0.8672527520473078,0.2316188461444576,0.47649905172036877,
 0.06420985987121164,0.9658932584214373,0.10056721728140472,
 0.29750202765575673,0.9616744143970182,0.6881710991233678,0.7914211535913436,
 0.708371166861618,0.973426970511827,0.013079923546846062,0.4563783999468711,
 0.24531648804131767,0.01126234423980721,0.7861218459527576,
 0.3315035859670358,0.43917666584724546,0.7402640554354482,
 0.18188107509335638,0.038766708063491584,0.050017651475035674,
 0.4293926284133507,0.39556004625185137,4.0802750989432646e-4,0.7445845002929,
 0.17070248650599285,0.6692985822528517,0.2655941399769014,
 0.021370985149404942,0.9748522919085941,0.1440403115678658,
 0.13293928368117947,0.8979129210517716,0.769842713475359,0.8430755294495165,
 0.8114963924600492,0.23069744063338704,0.7785588639862155,0.5438093206330816,
 0.6464095236207158,0.39867359958553694,0.42515746982153857,
 0.003811130945964969,0.7821189445557766,0.18772109181175956,
 0.17464630945259862,0.20538720849722303,0.07436464915232444,
 0.9781750210271136,0.5863780482984154,0.4920352960517761,0.4368056391984616,
 0.19402193293765746,0.40237420035844773,0.02145589690376326,
 0.048962133906520156,0.2525434151219449,0.25761034463055943,
 0.9608208581599935,0.06344668554974242,0.8859061852430004,0.6313352207760378,
 0.20628171064428968,0.537537033962257,0.9536160820490954,0.5605309061938772,
 0.21920289884243405,0.958798803680002,0.9618158081644155,0.05922568575522569,
 0.636363961937663,0.6564721421786786,0.45199632330682993,0.588598911629499,
 0.8866726184981981,0.3646817224329727,0.17660260198238842,0.1632660610630754,
 0.63703780067443,0.91041106609746,0.5309032211653496,0.602153604290733,
 0.4347830001612133,0.3919431254845577,0.29838671539034156,0.4000306115858323,
 0.5454704544761078,0.4194040920931815,0.32322321659976494,0.8660890905459031,
 0.6622610012457443,0.6018835469607199,0.7773098813657365,0.9994221791034892,
 0.7458930019217604,0.4620617067550209,0.8366079930175324,0.4201084407988721,
 0.7215691482809571,0.7097522036528214,0.10176145519977486,0.3068918605899952,
 0.481570694573749,0.10535768036875326,0.5457241111614457,0.5176913321903384,
 0.3001301419291569,0.7702603919584927,0.5899564880792849,0.39468628727208677,
 0.5792861113024601,0.11736626223960755,0.25356734560930183,
 0.18810879996963292,0.03929854956976342,0.753554917472273,
 0.061849824384323626,0.47351953282542114,0.9953520446933715,
 0.4561346281418538,0.42416182407069336,0.9790734030385232,0.9577717790262877,
 0.6755053197120376,0.9247382435269067,0.9728110851710353,0.782978130470914,
 0.8168620963793882,0.06853985026449994,0.6576136695782456,0.6051776422211904,
 0.6642536326497682,0.13890698119232892,0.5195604732309385,0.9711432973309618,
 0.29147304453287715,0.4851476672477816,0.43837422941877346,
 0.06543803019054695,0.47268164602645946,0.026104856553591738,
 0.045676448272934156,0.38778426096354657,0.3173155862576247,
 0.22846479258137253,0.7181063657167717,0.37158600014569565,
 0.023006211182082903,0.12837172006945474,0.3334532885924978,
 0.3978493355921633,0.3193755368758372,0.28629987864918016,0.8394260960655429,
 0.9730308816897788,0.5418520691506751,0.5400836028769467,0.49457136219204767,
 0.0018490082019403387,0.3138682921680225,0.8809822724113179,
 0.5610876792877131,0.7029307683522086,0.8320837795467996,0.527674449099959,
 0.19853806329206336,0.4249755203551959,0.06479399473773961,
 0.38787028429664827,0.3156470966797108,0.5583214019508461,0.6511349329630375,
 0.5306573151740405,0.4587224561918708,0.971679356586842,0.9683939707352534,
 0.7785338037190183,0.19534850113768543,0.9832023136046241,
 0.14293603469314686,0.7448306889656238,0.6382875381583449,0.5717174110502456,
 0.5754924508888336,0.40886432667694417,0.038496912543300454,
 0.5161448045594429,0.9234094893669459,0.37457688583292603,0.8406373464765462,
 0.6637668021175314,0.0963194093790325,0.9266924724645998,0.4813914029555202,
 0.34438951708374044,0.9812430520681545,0.5411942124396367,0.3186555717995505,
 0.04260792508348943,0.7245476567576199,0.410066687137596,0.0331094079083083,
 0.27120860213868825,0.4328584109939356,0.5838525394431933,0.2493718109334192,
 0.8440314809811131,0.9224955155135803,0.9727652589156413,0.6263698499418585,
 0.8626208549626041,0.8408349045724792,0.41842084306514593,0.2690448379929071,
 0.25706925123464663,0.26002672395537374,0.9238583667231034,
 0.6089928599851064,0.35795186413987856,0.05999806186177725,
 0.0745881640492938,0.3749013933625325,0.9016920559492037,0.6211740420125311,
 0.278476014155412,0.5139770779226098,0.7197965378057505,0.5468898579161369,
 0.9679757379547977,0.30011464386908726,0.7923790633621962,0.7591561970859803,
 0.9759216182645138,0.19583668993851333,0.8288811459352107,
 0.46171543295393325,0.03767786859692657,0.5466321946628584,
 0.9106861952913855,0.20849691689829697,0.5444723751461401,
 0.17763785801570386,0.7900492127493594,0.17238201512577545,
 0.7256785886895198,0.25259186862177374,0.28449662366672024,
 0.1633010660514762,0.08306413588806538,0.9276014980642887,0.8175872405522224,
 0.501469418083172,0.4930689872085142,0.8859976745885647,0.6853707022531498,
 0.5985078333082756,0.9751592801209132,0.8723728004375818,0.7433714574713979,
 0.4348769008924229,0.15808270491766996,0.9019465260907845,0.3844517245173371,
 0.8394217676327395,0.2606002984724265,0.33149675965906267,0.7481866468002196,
 0.9667748205682508,0.47021240874050574,0.31343318658829455,
 0.7736429309903068,0.597459754296174,0.6110523751872214,0.3854808961575845,
 0.21086357055697347,0.05869492875132476,0.10774040322500644,
 0.19872341255186754,0.9260852204945157,0.08864259543713326,
 0.4076932704116768,0.05158980166618632,0.49604589998365056,
 0.18628548258273103,0.4600233970121843,0.6023134924936059,0.507837325941956,
 0.2929510338954825,0.5244201589644646,0.589755250550039,0.43298211677566645,
 0.6839908836415913,0.7718348990459153,0.15057355703011122,
 0.11082463779190199,0.9088382628532906,0.9306250885340016,
 0.10531457317975401,0.9811501579533604,0.9649448905000629,0.9127536238771615,
 0.8786019507079992,0.12265379512446595,0.2654913377670416,0.9606527120788687,
 0.793491747251988,0.2470440376632682,0.6556526146614658,0.15915816872222333,
 0.8845525871439646,0.7291032077808726,0.311577434777918,0.5988646988608388,
 0.848149457316081,0.4574429347508373,0.19373308718426951,0.5347956345868896,
 0.302236560159187,0.5837801380197428,0.8055600942474115,0.16691394610020194,
 0.9513724029711348,0.2860204587202295,0.2790014741128529,0.569909305885261,
 0.9506690837174678,0.43102942166313896,0.8431969273786594,0.6189055754152755,
 0.858352974137448,0.9971374910532953,0.16073240008461043,0.5474523266629195,
 0.26037832140368633,0.8314448159701446,0.6755383119819253,0.6913487047879716,
 0.3728476562651629,0.47265156137696374,0.41834437161852,0.9284667752912661,
 0.8943409307106672,0.048708283269131236,0.7512454361830332,
 0.3787714645074447,0.08247780920812975,0.5125832316994676,
 0.007271126659053939,0.21009665208786443,0.5182771476228238,
 0.22310892128853643,0.26671405902336587,0.9307199379428985,
 0.1526032054130093,0.8787331094411189,0.46292154690258824,
 0.23265577259335368,0.4697847334035322,0.1009486604390708,
 0.09964456797299193,0.32160391172270064,0.4248240902102154,
 0.020103335010473566,0.49062437404752113,0.7365012042503,0.09870789796207058,
 0.8734670466328468,0.6982351345289418,0.3512230625027427,0.44927804733161436,
 0.9904299555670115,0.5416497022247202,0.02505785306175834,
 0.09158045137608584,0.73513552654121,0.015189210873546632,
 0.39731828773829203,0.8413312093885426,0.6353458696819583,0.7275789302667005,
 0.47703702684062144,0.11421933423057606,0.07000945734935404,
 0.11168473442357008,0.5786169625037083,0.28085770167035684,
 0.5122672670881713,0.6749559490918277,0.2900647899187859,0.648495746049404,
 0.1807005626170184,0.11408898567608206,0.2285761954797655,0.53684161602325,
 0.2085412753257847,0.8803241254207234,0.8708753303845632,0.2873097574397948,
 0.9920085231516496,0.8109867593120279,0.8885936229510684,0.47802571229339874,
 0.7298705989447167,0.764679318253669,0.5944687282341301,0.7548733333530622,
 0.6843428162387306,0.6863068139593961,0.5297372029243348,0.5503305045057926,
 0.3972378985962409,0.6775195362194811,0.7115925563607978,0.2365043483800573,
 0.8282988003918978,0.08421955972385342,0.2942168164554926,0.6067020407109849,
 0.9813473337171535,0.4197262739141886,0.5159634439790313,0.5199742249840026,
 0.5131981630577992,0.7210919729271142,0.19012605773350932,0.691056884992268,
 0.2903195210543049,0.08824564080962471,0.7635222223114303,0.5285451740947908,
 0.7026475985735349,0.7361796709266306,0.4502062152629044,0.5522074861717802,
 0.7348322354957703,0.7872046046585844,0.2164564294087743,0.7373877752181123,
 0.1976615622805208,0.5475126228108652,0.0627057160273391,0.40735791409644906,
 0.7372163968021472,0.40499743725855164,0.18716619873869922,
 0.04882668590777828,0.39787765188039415,0.49702972479565144,
 0.11003089572345537,0.2834927314407857,0.045713242897370554,
 0.7116996308030594,0.03120077158249579,0.9240200067349805,
 0.07095267563071428,0.08099904763282395,0.40017302009387967,
 0.09102230240442433,0.42261522667725393,0.07163702790607229,
 0.2675437492840499,0.9714774933221815,0.30076746826589495,0.8357330596652764,
 0.9436805617319377,0.1350503738970621,0.3069372502120504,0.5691424841095825,
 0.1497586484542388,0.8797266160739403,0.15926437520070458,0.667819921064808,
 0.019961107044591975,0.6898763506741867,0.7764550433245038,
 0.8088638140650197,0.6152072575197979,0.42641469996126213,0.4062118479049931,
 0.45375128813371357,0.33383483549604887,0.6633713858255049,
 0.9920168917444493,0.5356433372560132,0.773805590967052,0.06303433126071278,
 0.29798026168978753,0.8332025344360403,0.23883969329995092,
 0.6205091820068659,0.8510016731665836,0.9299597728427387,0.7606554940511725,
 0.034811513744564815,0.6781756615980722,0.47320620074007036,
 0.6842418123784203,0.9814860678613803,0.9452985578899943,0.673201738384148,
 0.9403422277276138,0.9928343744672148,0.7661647624625442,0.11674425661667998,
 0.21054326462624395,0.49902295532857666,0.6848121248550489,
 0.9635597437998157,0.6975775203919896,0.1523112831317227,0.7208725605887986,
 0.07610595898108674,0.9511609898790936,0.32775482165580694,
 0.6065425744473011,0.1166693894516535,0.5807687597817874,0.15514632755003033,
 0.7516124695735882,0.06397791118076601,0.995847651041167,0.2584190271738551,
 0.8831795236049869,0.3902338897098321,0.706008673744757,0.6081035936201149,
 0.2626646217997164,0.460791990516526,0.11161231554777096,0.539721972027071,
 0.9439718196850289,0.22361648927292022,0.47876941673017703,
 0.2693250245388612,0.8600756834976792,0.21338340945712042,0.3221915140065563,
 0.09898275346186192,0.09208309241976176,0.6230913465742414,0.207110342687592,
 0.5786617755928675,0.0207531679144225,0.9880015755110237,0.0493942888281349,
 0.8177153183135198,0.4270277714967522,0.4208580952154002,0.3489138065282915,
 0.16317382441406525,0.9510853142693617,0.37700246867764176,
 0.09892311910488993,0.6816282988418026,0.6289163269360055,0.7415254500396881,
 0.4399410540190769,0.6130305577991569,0.6635635037904569,0.16390847230065575,
 0.4611915434567452,0.9339476097237006,0.4812486495167482,0.5951704468340575,
 0.7901227721596714,0.7602937096132167,0.9093579288842993,0.4184437540980166,
 0.9597851891104542,0.7554485962406043,0.003476614650389176,
 0.9111022961395538,0.10707161049702174,0.840369436173841,0.835131074887228,
 0.013148325876442879,0.6753017096063068,0.24356343586222473,
 0.5734479353902835,0.7044008000531824,0.7501714444019147,0.5730970845325227,
 0.6174193999736524,0.5787415579865901,0.7625283919271248,0.6673708491067839,
 0.9033799302517894,0.5845295546816436,0.39630592812588306,
 0.18430114484373505,0.18714861909476266,0.29201424368145656,
 0.16464252118486578,0.3556901513080736,0.024373307991894855,
 0.5957211355214451,0.11741065428663677,0.27134448556712176,
 0.9120593597205933,0.4856226280102127,0.26711027886847516,
 0.43240979710800675,0.5364179250796927,0.7663327849801663,0.7083300510583225,
 0.47675213023327334,0.7232786939666,0.20198095974820496,0.2732310311840329,
 0.177139228159772,0.942652228918217,0.2021790726288586,0.232441128610122,
 0.4025891686512999,0.6579344138223412,0.7864131097076914,0.8139653766000393,
 0.3364862223179017,0.3771784440280328,0.09598220103773758,0.7011458185544194,
 0.6642531232722702,0.024651375551583254,0.8996725167436613,
 0.7347383476792355,0.35723321513478157,0.9982331261289036,
 0.010098430513070134,0.9385564939345652,0.715232674410503,0.5428312324480828,
 0.9108578905892812,0.7020422965285665,0.423439332549818,0.15375711294639127,
 0.1940649272099938,0.2415953709381431,0.3857296327297224,0.38688055558557466,
 0.44268321805580124,0.9804134425274251,0.6098581944891538,0.5954711275389534,
 0.4472430699515151,0.5275431538560772,0.42513196057073843,0.5411835085439621,
 0.9275632672529759,0.9585212373962296,0.9344571167717173,0.4656371347838083,
 0.6266535572847688,0.6816496821686298,0.6954840572094668,0.06792990754361183,
 0.14390364795240485,0.9336487819027524,0.7610990916551621,
 8.618733110840049e-4,0.9783210533452985,0.00503275662024949,
 0.6955652752470439,0.32426522383126133,0.12592348584652346,
 0.8494665563796275,0.6411095728190828,0.22614936067186453,0.9981626149709215,
 0.31842392811393183,0.9622537094910266,0.7214065553684724,0.975213546666943,
 0.6972678035721216,0.07633937390084189,0.27615801307307974,
 0.9110619706980615,0.6974459765181447,0.31849679955050547,0.7962922633238453,
 0.531555068490592,0.7053916878624278,0.5723969870181269,0.3640036920056141,
 0.3289635374821438,0.7541409180668315,0.38921146697964515,0.5334977373001281,
 0.6323667956713839,0.6348805370184426,0.7458842766952712,0.5322648569317795,
 0.10215569668719504,0.9368826716322934,0.1514132734338809,0.9357091198037804,
 0.4125343621136752,0.4631080491216548,0.7064384604893702,0.19742301089761627,
 0.5356656853409498,0.2084030976794763,0.7809956704508227,0.40528467755217,
 0.5855295425096935,0.3817936041779759,0.7411729320405979,0.6840460958095356,
 0.4056703618474844,0.5010931268539741,0.37046498963985175,
 0.004644818333055412,0.9542412312526428,0.6068741537970898,0.168308616457843,
 0.859635203319507,0.10142184251837338,0.5096117168553331,0.7536306978241629,
 0.05618699123856974,0.14889713298316698,0.5645436270798528,
 0.05741586570614621,0.7313074461597128,0.44541215025125036,
 0.5145808067194552,0.5462931974503739,0.17056686248302877,0.8798045051931666,
 0.6150166317625075,0.24494744172217908,0.13587118530314923,
 0.14463779866481552,0.3420720260205419,0.3878289581057859,
 0.14379656891393566,0.1477638979863165,0.563813567573591,0.489986029307782,
 0.622258293180941,0.39174278438354726,0.058828454077124115,
 0.7561374565482272,0.2770205950361737,0.17190069771964966,
 0.28220252171599136,0.42534248547080333,0.7629894866212447,
 0.8815989255535721,0.30173136487181473,0.057040451445485196,
 0.6672866627626163,0.1039516046317508,0.8544736865879206,0.44372835253559806,
 0.11906392935420229,0.7938768991511529,0.6870894413353488,0.66227090815441,
 0.30516220719584564,0.7582762563082791,0.6386345834344226,0.3646759342071777,
 0.13826819140154145,0.9175470274747638,0.33357533743759193,
 0.7697328565362943,0.8546540182705389,0.9919326193599214,0.929887448341588,
 0.3053868675467827,0.3466052442280493,0.8910263477598718,0.648184974583137,
 0.8421175066919905,0.19873723210582128,0.19256945015208093,
 0.6305358731330504,0.3989912645782303,0.48709130128116873,0.7834735306284086,
 0.8363750902214293,0.5653861795564946,0.15442305022859426,0.7248481172443102,
 0.6850390747902656,0.2704712512343096,0.4179249618961022,0.8825983031564931,
 0.3088118324433814,0.012521795167884875,0.7825367829644213,
 0.7360416578876783,0.9098533329314927,0.9780366801845342,0.4125756162797547,
 0.8090377213382711,0.5664254561433187,0.9341471002127926,0.3979996117466148,
 0.5333708483327115,0.45057637209037193,0.6177277146368197,0.691291992663703,
 0.308358865485566,0.7724488950263584,0.3034326415937496,0.7129265254217043,
 0.26173198951090604,0.6030316578854857,0.9752116519890451,0.5557155348392017,
 0.9323515402610355,0.105716376555818,0.11004730344347469,0.44615266410883003,
 0.45174267286516556,0.708734564761143,0.02469380222723816,0.9459806724919778,
 0.7823390106780795,0.8091124510756492,0.18590426034895302,0.7593104474898543,
 0.2184553559666761,0.441476282506728,0.26856731389606825,0.30257871735348796,
 0.4110714748664893,0.8835404324027536,0.6788920078209201,
 0.019086881540180012,0.639999617740393,0.58745035522623,0.5944825189550016,
 0.9778193071802931,0.5768803778726063,0.1347999097933742,0.6060874269360844,
 0.2642153809492114,0.8053114219730269,0.8817904813861491,0.9624949755789282,
 0.5718263435130251,0.9436863365052226,0.3581603733846995,
 0.020475778205894635,0.4503481187431717,0.24769721712554893,
 0.9448000382166036,0.006052391553455272,0.8244433303754702,
 0.9157175586066986,0.5266318511589019,0.40867150302375843,0.7046903926713418,
 0.6238002592986382,0.6971040158404542,0.03617173748945868,0.715261793429035,
 0.050136949644025486,0.5205791934192875,0.22966865980819184,
 0.38829691523893695,0.9610522031626043,0.8825988344105489,0.8098096017153934,
 0.8133821653957939,0.45094915189045437,0.9626684752457471,0.844121277236131,
 0.5578035639976373,0.9968409053018064,0.7791498114667686,
 0.015476192077084838,0.2324783431569628,0.8105939530783216,
 0.49604292394213334,0.881879555669434,0.20410592032840746,
 0.31510785449936196,0.6948988320150088,0.8054191634592855,
 0.49758843297479605,0.16887826307197618,0.05503830232178497,
 0.19609440599671102,0.032526049689469705,0.16780324180551176,
 0.7536559175826385,0.40130990108199915,0.06120947997357795,
 0.3148873114753279,0.6041176167070899,0.102830167097549,0.6725757496421816,
 0.9561615155647979,0.75304056888728,0.17934997508399708,0.4762777589166145,
 0.33264077238858714,0.021161520131694922,0.4341158059173996,
 0.43447424285206826,0.41243926618541693,0.5791777806943258,
 0.44684799853489043,0.5553020949185249,0.4765457753265263,
 0.053980379568229764,0.6792710833138158,0.8699736584110064,0.459082734684676,
 0.5460239596917732,0.9789531048587707,0.015349134394275876,
 0.9278329444809328,0.5775767646730444,0.22124033864992854,0.6610319147582828,
 0.2254468203606359,0.7425883246915603,0.5717848831977796,0.13637717654921355,
 0.5320266821010398,0.5495993704252143,0.45708083141128186,
 0.41988412858288404,0.16146699904102968,0.04278941753646315,
 0.6092714089933725,0.4080800035833002,0.7702278990717245,0.7992442307762958,
 0.02609014528625253,0.144712092702579,0.48602716666986856,
 0.48160689195791895,0.567998268778347,0.5500554960437884,0.04491724469464087,
 0.5974771097106948,0.8571201529170408,0.18723030625867298,0.9417776727786031,
 0.11607433611394913,0.1666893232924047,0.43186329395695766,
 0.9767465258619048,0.6288659604898704,0.36671547977339636,0.314642780567955,
 0.028891121335675396,0.10811280062978779,0.9942963039999673,
 0.5432128151106619,0.8065541984693392,0.38943820961908726,0.686904445647126,
 0.5483795269159677,0.0170936880115532,0.5198398569066877,0.7367202067379367,
 0.9728660116622557,0.6009154220237661,0.6971429053058098,0.2257805988162499,
 0.6176523750614175,0.8809715316563327,0.5440285023824816,0.5038434551705177,
 0.5424742854886181,0.44129470028113194,0.9568619048526996,
 0.02158414626705174,0.7330457308356081,0.623555662204047,0.7621271334936566,
 0.857622668570855,0.9478072795151321,0.8507736561785645,0.7410817486889215,
 0.5421354837084738,0.7339939830267453,0.6585055381393184,0.9743905265008586,
 0.4736103747972945,0.08110315989407435,0.3451837097224695,0.2559628573720837,
 0.3633629527767911,0.03351274296632023,0.16404152504945047,0.555314465005327,
 0.26550691652880043,0.5830668575792819,0.7701966232111384,0.4946370501337253,
 0.5378939327242633,0.7389459040327321,0.18050127739526045,0.6019418890768181,
 0.07942533835090182,0.9529609921663567,0.5959183230720609,
 0.36760109363146565,0.3428027251552783,0.2890932071427953,
 0.11124476180729082,0.8614530716497685,0.7733102708724422,
 0.34778825834012417,0.7259056265857315,0.186381910325629,0.5952492499234275,
 0.34361564217619456,0.12122557171851844,0.6744752708295817,
 0.03403436018892425,0.5385574221332234,0.003032197291423344,
 0.9517930398266394,0.7899155372370404,0.2152970760907401,0.7451554259000192,
 0.804752965024953,0.09283530000924878,0.6252115954037565,0.5478813441615881,
 0.5361099392959789,0.7412903479961936,0.4729397377120068,0.839495573084781,
 0.5410967810110822,0.023131297304675558,0.9951338950339607,
 0.8474089023691397,0.1275396967449226,0.3394334020459776,0.9837584194184655,
 0.44281046993875106,0.28233189516811297,0.44950625409264944,
 0.24957455198291523,0.5670202528004328,0.32132751104513524,
 0.024053135572853823,0.6665403077519015,0.2425503720885679,
 0.19183721790776143,0.8956750664431481,0.08663537833102064,
 0.5640058607614612,0.013699204932980158,0.6593417762654067,
 0.34257014380508577,0.45983231905811395,0.8982161397597033,
 0.7187216984858611,0.24712197194015828,0.5528311973163602,0.7763125524000409,
 0.14085539137564018,0.16744935489173396,0.6416006740233313,
 0.4822255120427985,0.547688571590313,0.866693834798475,0.9098577651212472,
 0.36189127713603875,0.7236823471017452,0.5853235021937228,0.9801449923340578,
 0.3169600555533233,0.6857379738887613,0.8349810668400298,0.05513742852442571,
 0.6240275924119286,0.4662365450703847,0.5151820682816577,0.12526596931151412,
 0.829778586778425,0.326280799543732,0.2375017070810196,0.6985570204114332,
 0.3159277091315632,0.6596856806884761,0.4258472075317411,0.5316428502104902,
 0.34788121341110534,0.3222754209432007,0.11529690732950659,0.878763437534342,
 0.6922016642057607,0.5873875557802571,0.33019442636463214,0.9297579895251665,
 0.5368363870496223,0.73309640113608,0.052334438583244824,0.6125089086350647,
 0.6301152490933085,0.1486085134732149,0.13458470463367667,0.7196326594511102,
 0.9130080036919328,0.07280329328590374,0.47084160730378166,
 0.4853178212784769,0.13777151988365088,0.6789658655782657,
 0.45320089384858175,0.0971768568833804,0.38109497262555125,
 0.5141678493883532,0.8268530385360592,0.43972421644680004,0.584908706794482,
 0.5166737414857983,0.06628512396532849,0.2124306722729501,0.8370521461618181,
 0.2762236211762201,0.9351826071576022,0.04880249700632777,0.8734090349204747,
 0.223654761470795,0.11921600322285408,0.39802573533413155,
 0.07997755094747516,0.9755677412018727,0.537884600113255,0.8546665843761366,
 0.7471649035557522,0.42355556689779983,0.5755340441777664,0.5163694310058717,
 0.3657388080170698,0.9803862310324277,0.1509035737705986,0.1622726207734827,
 0.9023509442458931,0.03521555442632285,0.6286147086118291,
 0.31115425199448044,0.7707039685336424,0.07905030227568499,
 0.030775075387553863,0.19806785331150412,0.5166421637889332,0.9263883712062,
 0.17921791377190976,0.30382586981562865,0.6858785406275411,
 0.050601349078128854,0.5966571402236882,0.5893235022298867,
 0.044100663674080276,0.043193870806063184,0.3278552076849717,
 0.09197390529535754,0.3091625234903621,0.7006200067142971,0.9350529758777483,
 0.08186799313255999,0.52640730810803,0.737804736977847,0.6748944730768828,
 0.8943961675753006,0.7584035940444629,0.5094266704131845,0.8166589711327445,
 0.5647702964745095,0.9907654754740016,0.9194695895476435,0.8553122845015803,
 0.10345673450438053,0.7013888430445415,0.3543290222302936,0.2788379112664686,
 0.2527984210017612,0.5141645081282687,0.9097686831782231,0.36241934431475,
 0.9866085644064669,0.8044123083360621,0.13629894366630113,0.6508698399623987,
 0.011467978071899587,0.3666714967133481,0.6410349618022126,
 0.1555393262805771,0.2164963058417585,0.3362157129883274,0.5050963291758119,
 0.5881563171428639,0.15318844922507924,0.32203994671156533,
 0.7697331313972344,0.44587984145522297,0.02829278542922853,
 0.11962741002857269,0.48644664920994884,0.9643728577274144,0.857842944948007,
 0.8897923104515717,0.8341126208410513,0.6371003174024418,0.12910824203697135,
 0.5992499432053682,0.8237302845145646,9.172557216728938e-4,
 0.019990563474769463,0.8198075899064782,0.6886715392438268,
 0.3598887052714015,0.7195515344682478,0.38924517548558013,0.7581751782280696,
 0.9290305047181923,0.10758030958447429,0.9883419595119952,0.6528527752561212,
 0.5533417703106713,0.2401678206224933,0.8554178031415911,0.06170146092954598,
 0.7116891038821371,0.5603534226952978,0.1800614290682745,0.26175440317715193,
 0.9349350779118973,0.37067341132311993,0.8791956829914428,
 0.10145335705436742,0.49920283273765187,0.515754028839854,
 0.20963946839981829,0.9855147866227982,0.9119200088263815,0.8829235814873114,
 0.9579469815345683,0.3742306008036782,0.3672214452576401,0.5245488381485309,
 0.8451740813947927,0.31705892575807093,0.5549466149627289,
 0.33962315131519927,0.7831268198999989,0.5819768729829343,0.6721810513614407,
 0.618140210101081,0.5933319242297795,0.35154252503349104,0.3915771199452436,
 0.7642361651820595,0.6600541353501476,0.9438883299676358,0.6360468701941279,
 0.24500497559832235,0.7980014429443875,0.30674687819959445,
 0.7148786442173158,0.13171858068257303,0.2684413148455589,
 0.36687125228057227,0.6684174637968213,0.5223518434516804,
 0.30716286894401523,0.6664190587392692,0.605796999741602,0.14548578571480353,
 0.001088952733378612,0.987028274421856,0.7579790966402182,0.4550166114941435,
 0.860871249033716,0.5500230901372005,0.8348969431300348,0.05029322539788317,
 0.9919033517918695,0.19329607050273778,0.8387086752672133,0.5375375262122412,
 0.5602700252272674,0.6341022285250917,0.5804570296391811,0.5514920990688018,
 0.6188267937498227,0.11604574378584265,0.3124632220003212,
 0.10881098660610744,0.051939814104051574,0.46145391520468604,
 0.9535756112574767,0.9967670207429106,0.29563713594497765,
 0.054484465230366874,0.7892142977670528,0.2602076341337096,
 0.04720803962038722,0.9302028611494274,0.7655577221075796,0.7541564265925248,
 0.4528444364296266,0.6538375013858972,0.2829160942415214,0.12392897584096085,
 0.9097392945627872,0.8061338320562597,0.8773618844329922,0.7652683673316584,
 0.44629338873449853,0.8840580977562384,0.5687877790323951,0.756202558155748,
 0.9075506506341068,0.14142782984636193,0.33645609494050244,
 0.1499767811407896,0.4924507717246309,0.846680258082677,0.04600115302867358,
 0.10361808543081985,0.7574599802638693,0.8290847467719183,0.7141600631606425,
 0.5829847508738051,0.5148281930319283,0.7121284419128591,0.23891366222810062,
 0.805763432183753,0.5552047029941887,0.10383127548307836,0.5431902595626665,
 0.05293248314234078,0.11698842412147026,0.0467722590617361,
 0.9744949461042486,0.6437496861433714,0.758847608376964,0.6662937035004468,
 0.16522473591107834,0.8976892321689698,0.08187845460995102,
 0.9093621080516533,0.8365989740146912,0.8556599921709117,0.8764492424195154,
 0.06610151599617475,0.1649117882555512,0.05496583937544475,
 0.7312725087299002,0.92386496609738,0.8189783199396385,0.2820418025689413,
 0.6240098620667864,0.07213738662822755,0.042882276296982824,
 0.8935586400439528,0.5198222606573676,0.8344801169729126,0.7394737597934464,
 0.2823186724785708,0.9206627657230289,0.8199490421177659,0.9372159949075345,
 0.2262922477792736,0.3568989376500322,0.12613776424311418,0.6613287087669009,
 0.6760514447932546,0.46277974571959324,0.15714214363050472,0.944226121836997,
 0.32789198509726303,0.24218097815935813,0.5447438437416849,
 0.7013130518568891,0.33673821408202553,0.7392492589991233,0.5925712237660345,
 0.9573728274109276,0.5528677383116989,0.009792047278631033,
 0.4600762250856203,0.06514606550673874,0.22700261823444756,
 0.16686648897337264,0.5416055505192306,0.01548495010389539,
 0.054178787260667494,0.949345264417362,0.6223538945601597,0.5570044656111581,
 0.8930306404805717,0.774500036823091,0.8445257409498813,0.44495900349060413,
 0.6253503707909479,0.8018446463923963,0.3633512713961182,0.5756686197671702,
 0.8320422235317754,0.08968142675656932,0.911047755395761,0.8038561874197986,
 0.22325388511077504,0.35582113131363047,0.06137705266916682,
 0.7428960333730603,0.9792732460809521,0.4271563124615504,0.1299423979611608,
 0.5493946978487861,0.418433087796271,0.46269026867991014,
 0.031585109564961855,0.6804980972257016,0.36838680304757654,
 0.12683694383606603,0.08186703911787208,0.8293922520575431,
 0.5226763194908899,0.5588916889001397,0.6599889544406076,0.1657070892772482,
 0.5208860608844318,0.6161837173579636,0.7351893678182733,0.9813657397690276,
 0.16499438743403705,0.14334466372031662,0.18206604847675456,
 0.25310208365237563,0.8973092862758569,0.8282056860881813,
 0.48405558989678965,0.8914796194710892,0.1874144887135557,
 0.41295832760231976,0.5410741932489457,0.15445869203974005,
 0.7651457430877805,0.7413940331492592,0.5452572984674471,0.5847717750916155,
 0.34901339685981325,0.30320553486352164,0.5739920636930234,
 0.7015002090974477,0.9134569887641323,0.33291221980046504,
 0.49608415174103304,0.8788324325700536,0.5839616177423358,0.5775009398117396,
 0.45154497253357373,0.11588122444237225,0.6473751765070319,
 0.4879756897202663,0.9332593525383422,0.4209560383588855,0.2586856490637409,
 0.47843131802138517,0.524888084707458,0.649232176952975,0.4907222299394939,
 0.5888939671360921,0.19229253852997596,0.7120870349916039,0.446887522488684,
 0.2582375737685484,0.09253513166154426,0.17772003233568523,
 0.03210276430896042,0.5930730239771194,0.9698584426454658,0.5908299952274966,
 0.4690704335639775,0.304792147565801,0.5407069929638518,0.648759699581593,
 0.3308511825810657,0.21511180265727192,0.702111472812661,0.3499134830200066,
 0.8071183676275353,0.6435334945497045,0.18328330565743034,
 0.16081215878779198,0.09245638863187011,0.14046984853076716,
 0.9461588728407013,0.8083930358441863,0.8645459888371405,0.6062107624733566,
 0.2544774605192698,0.9237935505328695,0.37978042971002457,
 0.38583086775599207,0.5872086870020005,0.1767103711745257,
 0.18287511358663433,0.6814755933558206,0.20052837214709784,
 0.9710206575004552,0.8336655823771664,0.38918428013494877,0.4269378855072562,
 0.5012674545967127,0.6222822483145687,0.9208439217185066,0.9078761438680435,
 0.46674860836491716,0.43068635213261397,0.8574518319265966,
 0.14377481641878243,0.7090194845483946,0.7050925904946631,0.9881231646195615,
 0.7604381455457946,0.14874363099940768,0.6262878327434814,0.6081196956755168,
 0.6077371515938481,0.3740264059525812,0.6128025384489799,0.37937052379498204,
 0.870508217391983,0.9965825101638927,0.6820428524018092,0.16842676972127157,
 0.10637552464034328,0.8424987977723803,0.3593768653953928,0.3304127647283973,
 0.23894051334240052,0.08122618862388487,0.39130426703966026,
 0.03370896375689869,0.49623877432545715,0.010346563994550806,
 0.5304611219670153,0.5147316747094253,0.14792395826122096,0.4025265814484308,
 0.07976591586017934,0.9057940161749388,0.8339360573327883,
 0.43235078225941415,0.47296056719336044,0.1684534099818482,
 0.8927322789322057,0.01426190003375094,0.47822826191539003,
 0.8625480554103963,0.7544211726268018,0.7514236077586604,0.8858457127433578,
 0.5394149182349407,0.39030539334575187,0.10096999007187146,
 0.8866087896670678,0.6322974078715251,0.3446209120113979,0.969506771176301,
 0.6783428817696222,0.7229656788277488,0.12798141029309384,
 0.15980348889311413,0.7566012139914953,0.510760614487616,0.6684423345258539,
 0.24560480835762266,0.13244647440531998,0.9158810964640527,
 0.8854755893598553,0.4107799008747347,0.8130870779273854,0.9334436679247691,
 0.5739161821357301,0.9545926432126838,0.19658783790261913,0.4461156091908659,
 0.8706785720029626,0.9492796157482475,0.10227965946507256,0.9102704787115702,
 0.5260117938034964,0.22356141031599563,0.22414597548245874,
 0.7511935472065688,0.8665635323644102,0.2623378604858624,0.5935456647961244,
 0.58002898279046,0.5867418135132159,0.44428344302938283,0.8513262878643015,
 0.9548294349461196,0.5105635670147323,0.6728245885058445,0.45236860392172973,
 0.5063407099844902,0.33048091422941805,0.9498720828609779,0.7503176737114278,
 0.4989520969517013,0.2157335107331504,0.5665606443241905,0.7484398256920337,
 0.6750598173570436,0.18525939478914422,0.08911503508829355,
 0.7040185861198857,0.7167486233514303,0.1819464270530935,0.23141074290523667,
 0.021778474797196634,0.5956703995931814,0.029010952557393388,
 0.21787366771909633,0.5584405473784485,0.4537963642339763,
 0.20847926928276195,0.7063893849141762,0.33065091092592225,
 0.34189554953030743,0.822022645485701,0.7992209482041819,0.8624499431587354,
 0.24727810726441968,0.7670648490766157,0.32624881444208154,
 0.02363344887148111,0.5120463810241966,0.43935013354154484,0.911269953938729,
 0.8195706378712333,0.37358335424931366,0.9812319286180888,
 0.40144287256591116,0.04985081005910197,0.7396002645885271,
 0.38294504847297417,0.16510341358060854,0.586743190849277,0.8160321696234405,
 0.48062947847538773,0.4253525501315054,0.5645782453781358,0.4680612411327908,
 0.2900685737299612,0.0168523544032273,0.5360192721371406,0.06307875185875178,
 0.6083028443273317,0.17701343491637955,0.6353444752775882,0.8311513440239686,
 0.11096990106127935,0.07774423140049702,0.6468967527244999,
 0.5916092344926724,0.9112323783614616,0.7634710577276329,0.6282240116747457,
 0.23535283729932532,0.6115856846201697,0.49127842597047244,
 0.6309838011924273,0.455079875952616,0.8612422558335883,0.876148827608305,
 0.15480937176178566,0.5138250112200088,0.1818723251260954,
 0.23727162852943817,0.43679955251513936,0.105515398043046,0.809733278001431,
 0.5276726280233094,0.383850438568362,0.8129564349738121,0.01957134501322555,
 0.8784168174494694,0.6190847157875824,0.756897627817942,0.4946812126357345,
 0.9874268751805338,0.3076001575792755,0.9211251783179853,0.5620371734787635,
 0.8745135131945396,0.4289904099622349,0.3200215872129857,0.08697624145549265,
 0.30358225227643965,0.4286764814397511,0.8903423685589136,0.666075541172432,
 0.19381593979080247,0.9900721338106934,0.9656562460386381,0.8979200321044069,
 0.8272808535653959,0.17866677341235926,0.7080619244746562,0.5233733746876675,
 0.5975580142564203,0.22558141813472554,0.659787502386283,0.5096631679511113,
 0.26487600542488154,0.09649906701727617,0.2867356084521495,
 0.3753426193514693,0.5987965364421208,0.9950600825281485,0.32250112654028407,
 0.4457359431576403,0.4983590284198154,0.22649936010273797,
 0.22396906086744284,0.8403247773938196,0.7094890140441968,0.4727977528566789,
 0.21334905024842643,0.20906271195336967,0.7963313986024946,0.522489375341471,
 0.8655431206413857,0.7622471807183845,0.9155865836479654,0.5524881298307948,
 0.09373172436615129,0.07582781765502733,0.9649091258928959,
 0.30601633727199695,0.7030656573928553,0.9017172318331883,0.9405844648238152,
 0.33867895953054106,0.41548561460718914,0.046090851125931476,
 0.31508262426243494,0.5213545523968784,0.3676729244118436,0.9205855057023579,
 0.2741704620065121,0.38428613298476433,0.5656372595299682,0.6719546177212433,
 0.49311562054192914,0.008872048816644718,0.41249943980692083,
 0.2830217004788691,0.8348093341185691,0.355149728913453,0.07317749555189845,
 0.9709785379483065,0.9590321638021746,0.485161879433072,0.2007239626180708,
 0.8189684073731898,0.40282632734792534,0.635448946160018,0.8868371398008617,
 0.5016782364761909,0.8397521033509918,0.8906565737698937,0.5724078813112984,
 0.5174054453768145,0.22004124967748961,0.7878540679361302,
 0.21407525099262692,0.11400725303095693,0.5331542340576975,
 0.08272945502037299,0.44428291686150245,0.36232436536682944,
 0.3324113795029142,0.7457850857291247,0.7450561070472473,0.46255424182620886,
 0.08746276233287309,0.5372693536198938,0.31970892516447524,
 0.34429944233585497,0.05241407347945548,0.7200054637816937,
 0.24648194764510623,0.3219776303306101,0.7250686246088891,0.2635774997488577,
 0.7254477244796461,0.3688471720910822,0.9651484973618758,0.8918271075634118,
 0.05302550335020295,0.9870139489185835,0.09150161701406212,
 0.7558892881002304,0.6403462915254902,0.0736922107084309,0.6457158736745149,
 0.5855851184074438,0.04053533847892177,0.3066699659968891,0.9565606522266346,
 0.9179088863219497,0.3759749607153333,0.6225042712499962,0.12845362372905944,
 0.0030165086047959644,0.3243551960103018,0.1898865971290078,
 0.33890882749132323,0.8041382023043075,0.30080284424506687,
 0.8686994000040502,0.16941077408365013,0.22321051861414098,
 0.5346341268882858,0.5248985419844319,0.5845290808034939,0.5345002162941754,
 0.19410313588063377,0.9137674262544333,0.46326730962092466,0.276119463965284,
 0.5730294436523773,0.034720267334812926,0.845113314229318,0.5633544970394052,
 0.9309925510258641,0.23754575723428284,0.35425024189258114,
 0.006943260935163842,0.7880158778457128,0.9484370258917246,
 0.35702266404097704,0.07290760749182712,0.3698541806225918,
 0.36595909357086476,0.8796686979153627,0.34370807224575006,
 0.5133178040768492,0.23740470164061678,0.6769433655000666,
 0.38221237683264964,0.0816153169154743,0.46033065640667126,
 0.9819955852375861,0.23815604033094073,0.43384809736842533,
 0.7947941464286983,0.14932108073358186,0.5163877869234017,
 0.016131290436093626,0.8852531729630537,0.8093003661248414,0.203818450960095,
 0.9231551833543231,0.8320098070222205,0.9416394824383422,0.6472982327465846,
 0.08322931803237232,0.8185057639291502,0.08485039353553325,0.255448379174277,
 0.16129594137758763,0.6620926755557279,0.17640860293678196,
 0.7515910692177148,0.33087967544491803,0.6719330729974096,0.5560466400136722,
 0.46486157813193374,0.9227474462983708,0.3731503716249941,0.7509891044230458,
 0.1506401337353389,0.8030127447825702,0.38188494376694493,0.9917280771105168,
 0.5370680337082591,0.9826326981012282,0.9897035414673233,0.8700640081557108,
 0.774689245207989,0.34693421560171667,0.23663895633545007,0.1014378486383426,
 0.8594846812310051,0.721701291627493,0.09449272467929215,0.5327910646513895,
 0.9445075537952919,0.4447087067661527,0.033079067434943266,
 0.33087186665486545,0.9966601539744662,0.29502012591138493,
 0.9209728011336582,0.29057011020634904,0.9547299236447222,0.8552014431058683,
 0.7398997507608518,0.39824808760966146,0.04019561614873102,
 0.5392530806817845,0.8612628867163048,0.048330084946181406,
 0.0599676249813762,0.8916608265400442,0.33917714558132483,0.8011302483717561,
 0.8184658711285284,0.3993885641700733,0.3065617779626064,0.6747218234472541,
 0.8982689693251015,0.4630179998725117,0.06110625001107273,0.5509629049269111,
 0.11838912409639224,0.4683210946010439,0.488578425854184,0.10758267252759246,
 0.1645174242344032,0.5881431349341651,0.15073381794185425,0.3720714587754349,
 0.3137047222630942,0.45751399381620717,0.4694314521308356,0.7693450423049657,
 0.9314196967428957,0.7611284192728771,0.9383358448390765,0.8638129443617932,
 0.9165258033068943,0.7551331104622573,0.405250790500579,0.3435447472287708,
 0.2424653984085816,0.08201011881046416,0.8791905851772497,
 0.47926463663572516,0.5671985763049368,0.4147331808813155,0.6461770455839927,
 0.09979302015230795,0.03618251052465671,0.19405739320003346,
 0.8867120836451825,0.04093156560817568,0.508441873993424,0.05521119454052659,
 0.24907286097965886,0.0786548778785473,0.06961282175689887,
 0.4234657465731996,0.045592987984355404,0.49717308157459994,
 0.47705447037361715,0.20810409853515477,0.32772230801181546,
 0.36158144428471806,0.8654493924082598,0.04886826208247308,0.553555541202021,
 0.532103301660687,0.5689139697755151,0.3603438967950078,0.4315979836235849,
 0.8715226451124035,0.3126652962851725,0.9873048568291529,0.3068875553761926,
 0.7458537677738641,0.35438858706401555,0.9690554165178165,0.3222093574465613,
 0.8258305946347416,0.08677187933606989,0.48531066637082565,
 0.1103742907311146,0.9452289916401985,0.9732599714976542,0.5662257002432772,
 0.17847224626153002,0.41957757992950573,0.16830668617886402,
 0.6448821018222369,0.6921639513864583,0.46392455987580006,0.7923295558642356,
 0.34401302551405566,0.6746409487890832,0.6543590748354864,0.174818778650756,
 0.18326218568166996,0.21769099691669558,0.19658849907893183,
 0.2699866037572699,0.43349737740283223,0.8388433133541322,0.6917400711976214,
 0.9675168339308622,0.5812660702845779,0.9434953313150439,0.3597377181271302,
 0.6128039057829183,0.876716122097566,0.4271772369827832,0.9377559478797558,
 0.0024458700589310434,0.7054522368280567,0.36229202547118644,
 0.9883068335639963,0.4283053908447986,0.7529098508050919,0.7302503506203039,
 0.32048403484451615,0.6560836334221152,0.8089249087107996,0.4182781219639905,
 0.9928882350798297,0.5750802419200542,0.8372935028590951,0.7034664553929353,
 0.5743749070768955,0.8023141838751123,0.9423519835179199,0.22850230935488947,
 0.3577945092729464,0.6393395964460364,0.10830587638994749,
 0.21754934494601796,0.2771581039869855,0.9464922936555767,0.7128105335273504,
 0.07437518656734854,0.6184898824649527,0.855598643413436,0.9581687757201008,
 0.5268358680226666,0.563433283132848,0.1589337099526671,0.17792132124584958,
 0.9924388494817885,0.5748320234351544,0.28478439413817214,0.4650040252218135,
 0.5287779288596812,0.16966688081834502,0.19129018981289658,
 0.043586950049556594,0.17651706807979617,0.12302630869171927,
 0.6471412668361196,0.009433084072201003,0.5819478524816285,
 0.8001722650110652,0.3709337318187734,0.20512355016539785,0.8421083532087902,
 0.28386038347105336,0.5065979345460908,0.6487214163485746,0.4833580040822535,
 0.7254592519498678,0.4488032183296842,0.37376225797956986,
 0.35431182612867973,0.12114260159147872,0.5866301702567351,
 0.49097636038307946,0.40935389605091466,0.7062762198199362,
 0.29005487011513975,0.5113780201781227,0.8536420510056484,
 0.30797230973263623,0.1083844426543018,0.39986379704278596,
 0.4045460916128628,0.6754809737732564,0.1718759549225296,0.837033507478461,
 0.4367297138260401,0.10282078139602446,0.324208870872602,0.7925715434753402,
 0.09615847079161566,0.5750546514205045,0.9373076874109553,0.6061672510604893,
 0.16551163404605296,0.09260567127455865,0.7319221075303255,
 0.19774399756913175,0.018576108933098467,0.011098037293063912,
 0.4892870455895858,0.48915598773491553,0.7922159913852933,
 0.14420129808806526,0.5016801115781631,0.8781412815233863,0.2821592587358136,
 0.1229893855433839,0.04756094056524185,0.6890432791769744,0.3572421843824436,
 0.3648000667620185,0.027101218664622118,0.17915164832710018,
 0.8142032893919667,0.25409595796256035,0.7801552381096697,
 0.024268248752193577,0.25859870005699226,0.44816523212861414,
 0.5414532696094041,0.7935668241557073,0.3942265618618127,0.8268647264716877,
 0.29370045173302306,0.04020108684877377,0.7823766962061864,
 0.5014805880052625,0.8081730354104676,0.722374548413922,0.34396867402227616,
 0.12041018035928941,0.15844213663869433,0.6976545235914967,
 0.8612991998176596,0.006680726129506875,0.2728170036062918,
 0.8830708438181497,0.3691807484894023,0.20570216627721882,
 0.48451190003554123,0.3162662442439974,0.7595236783017839,0.1146832139678825,
 0.8623110582020571,0.3192587282888045,0.28849434488754777,0.2941881076112258,
 0.31531292476261163,0.8189795704747889,0.6747145995462691,
 0.49467624009683164,0.0305904012601389,0.4492079922228017,0.8937992179988757,
 0.11561309658824248,0.66634969655964,0.9658171530100048,0.017498436089443348,
 0.4972642482580629,0.1698921814543557,0.701454892458511,0.25457618584948205,
 0.025730567674435045,0.8935742636006316,0.36736051288711413,
 0.8465699288289695,0.3676843917228345,0.6914934526433448,0.7362252978038205,
 0.5854694334733854,0.7592830745469795,0.5573831988117505,0.4600271833070687,
 0.6212607201837851,0.5474960689158581,0.09886046927340797,0.6126096836853324,
 0.5942269083780868,0.8215981524793173,0.11214283095913746,
 0.22016944837911623,0.9009799013251751,0.5461101163549743,
 0.17141514976488692,0.2178105826272998,0.9231660814921068,0.2584102648268326,
 0.18763987848677477,0.5596297981241234,0.9800722482911366,0.3402008591353365,
 0.39374875921677677,0.806131659935654,0.3309012272655991,0.5623618187083402,
 0.207121846730697,0.07300532961191919,0.2477462126070955,0.5879875809498931,
 0.6031808497722922,0.1841099022587379,0.1983817753121222,0.2984683431589704,
 0.4179755315973902,0.5835338178258994,0.008846386340371604,
 0.3116534315500833,0.9978974332334928,0.11265734313094067,0.4023238707477096,
 0.37860190488612744,0.9606520655400151,0.7471874529840867,0.5123362045440617,
 0.037604988769919134,0.17488838385679184,0.055072726476895095,
 0.056109857260696616,0.7194736466090131,0.9391548770553475,
 0.5726779100198721,0.8902032879252755,0.051893464545783585,
 0.1782281570610703,0.9379579958886451,0.9762418463433741,0.3626454225833373,
 0.012562709935287408,0.5378979596867413,0.42348203673400264,
 0.9720990321366647,0.8279312123153891,0.7449613428509935,0.2475743581513028,
 0.9440792444558327,0.5315512122070365,0.14604798185985446,0.9424025737367236,
 0.6624461882974684,0.6696815192965356,0.53677472800569,0.11438055487984233,
 0.8151853276926997,0.7486285312136494,0.08864409691988417,0.8319811230051931,
 0.012687362707057837,0.5228572968293812,0.7509500530764629,
 0.4184685403979056,0.7248783586027041,0.9109864983423681,0.9024923471818426,
 0.8789847510815818,0.9477371781437328,0.707305799772376,0.506869531133145,
 0.506422717112269,0.21284962061042778,0.5851526018341959,0.3528767350078369,
 0.25045370519233,0.5091490628618783,0.5921972664941952,0.5507304599267168,
 0.5096697101834076,0.7959881768425524,0.20653153268130375,0.7828936885528033,
 0.13087468646599343,0.9263577086683953,0.6537055057751634,
 0.34320997367512385,0.7290438184075564,0.10762959554287399,
 0.9358313240747539,0.8582465589777717,0.7354009178078261,0.6540424683613426,
 0.36834514566494736,0.5952715261855849,0.4825661232212599,0.6098709340676673,
 0.7929010978787041,0.7359045124876687,0.762407153080257,0.3566575130696523,
 0.41029107132027265,0.4312651664693884,0.581257969858806,0.34037140978031977,
 0.6189844577941093,0.23890959187157434,0.005491730149677743,
 0.6086040618436374,0.2662821014692873,0.7976256272331801,0.32683120635596064,
 0.0066034084556919614,0.01866473108556599,0.6350019159467838,
 0.9526311298954085,0.6412316444882944,0.8489524998806801,
 0.007421800342703455,0.23941539124326594,0.35948296948922887,
 0.8164792519727461,0.25503555085559915,0.09752517263616367,
 0.9070803575001931,0.1359924747782616,0.8148837007117369,0.5502215004423521,
 0.5407589943941588,0.7960312449923682,0.6182839950906271,0.9431414585665834,
 0.49845636186156106,0.5636550684187085,0.4944920711980336,
 0.45078380121029626,0.5839765215419545,0.939784672385622,0.8311990797465371,
 0.15962778041027592,0.10919570643997312,0.9439892295049642,
 0.2943582836897888,0.21272992400825386,0.29287573826934965,
 0.8987741856324756,0.28776282298778266,0.49290217607721865,0.749611643636483,
 0.917210730592365,0.2763358487298727,0.22746698407052957,0.740128062764548,
 0.6797218623205691,0.6476408454312671,0.49885243096507825,0.9916079897798155,
 0.3785697939602958,0.8148951952634625,0.24164901675894146,
 0.45009970799491217,0.7460064844329254,0.8143564387427855,0.9639182269884525,
 0.4728766132592961,0.8362194711190306,0.24844769877049544,0.635837839730239,
 0.08566249404619997,0.17837026849294013,0.8656942000612331,
 0.29861101648822463,0.5243582657892714,0.6917448707777264,
 0.17539165337627094,0.34426684211348035,0.2534341282519248,
 0.7908578348358783,0.4861477944783208,0.5944506165360761,0.06297266135109592,
 0.5264830257879348,0.15698034553016893,0.5002942188257143,0.0125104721799415,
 0.34299341691698626,0.5469853601571767,0.23225072449857354,
 0.9317842164869635,0.23968971988822396,0.65813034157598,0.49964432720134266,
 0.8011900976742232,0.6953014727654157,0.43730666268612417,0.9597223174458579,
 0.1540574312100622,0.3538804369133919,0.8906894465513527,0.7144285804218191,
 0.7283332793913596,0.8265127862690327,0.8154036666034621,0.00658256691813619,
 0.869542108187198,0.6688072281235431,0.061723220839107085,0.8854036108692496,
 0.008586703365386361,0.4691069143110477,0.8052915794114166,
 0.4814624995669007,0.6247409037530544,0.004495949636741603,
 0.21205381021027725,0.05458876011095826,0.6539417292320701,
 0.4311077987450097,0.3241010569186257,0.9360468010535076,0.6014222255758637,
 0.9905781043387487,0.09596148131971127,0.9569240354902017,
 0.07036774348812735,0.30046697114382614,0.8102208864389825,
 0.3871581696287525,0.9708874256619107,0.8684000323786616,0.40054354317263485,
 0.9694440520072564,0.3896324109969723,0.429698393566589,0.6110073615595071,
 0.43111982741435606,0.7276722292628247,0.588277386558439,0.795842746649496,
 0.8916168695923872,0.7985340651030339,0.03223406195883838,
 0.48371701388484567,0.40993848517781495,0.9541472702916687,
 0.4343011799050439,0.5366643376953504,0.6670999978722759,
 0.009868299102292033,0.10710408071029032,0.8613740284780603,
 0.2924696986030204,0.11770928374593359,0.5971270607323016,0.9443823392721906,
 0.9825761147467134,0.7326296625039147,0.9247924213206979,0.6880985751398839,
 0.19818422322069262,0.7035744612027767,0.22976699599157657,
 0.5212164377825317,0.17967850491921766,0.06980594236323667,
 0.10012967334817835,0.38617752585820186,0.40684249029633945,
 0.5961478816786233,0.44571056261148867,0.6557928006542441,0.5055463896473845,
 0.21490486299422606,0.5278109810668796,0.3419784293871504,0.8670636461749054,
 0.32622041783005773,0.19787193712657702,0.25244314901040343,
 0.773325611201063,0.8510763308802185,0.26009885358247653,0.7912438007710904,
 0.676066615400101,0.5975568929734597,0.9229308199746236,0.36033961533732506,
 0.5226875019672486,0.41771575464888633,0.18840085811916363,
 0.41806458103127797,0.3664514710628324,0.8633772806404628,0.3617916743303218,
 0.5374244007503152,0.48850394627899163,0.06034695334894413,
 0.4172227114953877,0.3897001514124171,0.6362571472916309,0.0583428446195704,
 0.3157106160831351,0.9320747784435867,0.749338578352712,0.16221427124792387,
 0.055926546987102976,0.24147358478690784,0.10178946565154479,
 0.3334006761945434,0.44264703884084,0.1280195341342769,0.6945667244075253,
 0.5908079321944444,0.3297254561566485,0.25338939259461435,0.5928238717195111,
 0.7457665065755886,0.4185398912304429,0.6861753374741058,0.906342967750608,
 0.38699092846258853,0.06535483615344206,0.5473721591921697,
 0.4653616257707939,0.558254192157512,0.05790408599998598,0.9085759002053257,
 0.810859180341974,0.5894970126825487,0.3868133113622436,0.39127707026370406,
 0.9212619620603244,0.6772275637461054,0.4044210105356538,0.9939141875255811,
 0.6419003759579557,0.3642560256087345,0.47958421415777097,0.887885416919731,
 0.8779871024108822,0.22068276322797542,0.45944659932294873,
 0.5652821466858733,0.6122555224056248,0.6577630375924868,0.20182878693221373,
 0.6535261264682357,0.5535368247312271,0.5368578326385034,0.40165185896883127,
 0.6826235378337902,0.013742699515283285,0.6850135601829606,
 0.8485631982318717,0.7853078492434205,0.03226271096499789,
 0.09401527529127407,0.9023460813003373,0.1621143574776015,
 0.31803929033639255,0.4009937994699393,0.9571915838711123,0.7122207108635384,
 0.7608039036399945,0.24334565173266154,0.3252028337301409,0.8907346402152092,
 0.6793795755452696,0.8653164534512688,0.2699309599746852,0.6973298478879789,
 0.9405384801489559,0.6858305287220186,0.5170542625048958,0.27952694179221305,
 0.7134575294318117,0.9892308729920667,0.5850352109320759,0.607822475898381,
 0.7737522216008424,0.4694555728603663,0.8031682303059661,0.5288474575316258,
 0.3614461093308269,0.14741305025658769,0.9662464462202548,
 0.07744639396534714,0.4232052612919157,0.9747631873013164,
 0.14696906256408904,0.8451645760693514,0.7281762286091686,0.4850640213743702,
 0.7681540756979429,0.22340037866100804,0.215712277165667,0.13929551022946318,
 0.3706665651328631,0.4077283900763927,0.4382645364132137,0.47736916330128376,
 0.07880683429550339,0.540854464699943,0.4250559222445577,0.41171095492490783,
 0.6927243717152928,0.1387944983944145,0.38147411541891596,0.8517250538911892,
 0.5902336159959101,0.09512937107223074,0.695744792112368,0.4070902078033478,
 0.17404593431886406,0.8532437585053557,0.30444904948591034,
 0.2880053585661927,0.9354705493103814,0.9607962444117962,0.19339235092837015,
 0.6180002871273672,0.2911379742353467,0.3428802581428484,0.653744900784433,
 0.19535555814866634,0.545359661271067,0.9771496594767052,0.4087266869428101,
 0.8833812312178455,0.09283762523288375,0.13825221531574217,
 0.9081337161638161,0.9507822135214877,0.1419147254608356,0.4205880723905322,
 0.7593580315536087,0.5667703590268904,0.6238251287395814,0.9298909495877867,
 0.3317431258377077,0.24367607426028515,0.4969969071438437,
 0.0035832117289479326,0.2704116974727966,0.04120978101119155,
 0.37696949772141086,0.08362461982518432,0.5508349676788384,
 0.4980450986134216,0.5700467566790519,0.9660574949623547,0.0882035801129405,
 0.9473978946467912,0.6790159574967428,0.645352075449575,0.8122813252839666,
 0.6704567497539844,0.8435238039050629,0.7387876852840937,0.7480463449001028,
 0.004347378177542938,0.22617581648089347,0.6251324906585866,
 0.4232612168258878,0.6173649603556394,0.16012559356775902,0.9320175615734175,
 0.6933137956775515,0.5228065596706829,0.07264302852474547,0.5161326301860473,
 0.94757483621324,0.7538328429287847,0.6790415788321172,0.8010706062928914,
 0.9462260093722903,0.37719369942398406,0.22697561831739987,0.540239374787994,
 0.9595031683823041,0.7143971301005656,0.14743770464755557,
 0.19879128433841387,0.11806543573051775,0.2336432728832245,
 0.38485279579934395,0.7192635649755713,0.5275489622790012,0.8797356994365054,
 0.03196737912648728,0.9343637637575679,0.26656029755061417,
 0.6680379623656965,0.4571831383689454,0.9038100792073144,0.3393041883650787,
 0.9002376366472531,0.27894721786592824,0.060221582374566984,
 0.8559321458721162,0.32562024890198327,0.6270999912377029,0.6971507237298901,
 0.6576541982138739,0.5437249623372389,0.2309177997195474,
 0.007804670471578179,0.02345436785597954,0.15968512835579407,
 0.48703639129175325,0.14034562882421153,0.6997262191088861,0.366567430764256,
 0.46568652842227265,0.3197111903389638,0.38132876814976524,
 0.0790102627589152,0.7314006479565477,0.5738008714968055,0.4229096985623504,
 0.6727595604204151,0.491416777200709,0.5205530135535111,0.11601293979574023,
 0.20941556849184895,0.6124429114868373,0.6170759704338756,0.9514280075444572,
 0.3069290283909759,0.9429415776407323,0.025489239427924026,
 0.5227123964579716,0.9776857829439877,0.03845943754258552,
 0.013764120850485062,0.7351876934041126,0.34303889941847343,
 0.7573102606254688,0.06537977483980217,0.35673598010702,0.22039162731631068,
 0.18258149431885207,0.9093794847648267,0.5469839599773624,0.2985552540249359,
 0.5167353088666041,0.8029917047780807,0.2798435470959344,
 0.022373806138239805,0.25952886607371284,0.8045781093177269,
 0.0332516412997621,0.06441701861183236,0.5483638744128,0.24980134419859468,
 0.4215404160630064,0.3612993831942045,0.3970021225014626,0.16923524752167074,
 0.7367897069009459,0.01981072397775907,0.8325696582994974,
 0.15806183793729012,0.8512222806402095,0.920749867070671,0.3700318739958559,
 0.6430584408140618,0.19631232912043883,0.019423247710267044,
 0.6339454024937602,0.36521869509460514,0.09180568643477904,
 0.36744388530602445,0.17052140682004113,0.2542285605295944,
 0.8740376181780031,0.4070018576057678,0.5078328811952285,0.6763201637954365,
 0.7272591991452589,0.11511835321447927,0.9087662771409656,0.4419563688600314,
 0.4376746664840754,0.29102058343570647,0.37345773715463493,
 0.6255653176017265,0.00670630638911518,0.6709943729694609,0.4731078578760567,
 0.8004167092127736,0.7274483692193142,0.0031125978622693617,
 0.13711403700026537,0.3730810552391952,0.3731361073993261,0.0879832041923595,
 0.3519901810840319,0.23565701555683782,0.7959686603899913,
 0.16655039081258582,0.13469690834062098,0.4930746325439471,
 0.6196082635904115,0.9647975482347249,0.784374921573128,0.9303582928847974,
 0.7011846874383554,0.32077135308423554,0.9293356577564216,0.5834957351842056,
 0.6078464921025751,0.04034182564717881,0.4563789765841043,
 0.18342760539001535,0.4033225442432359,0.28394773790383143,
 0.7180420116974271,0.9864756523273162,0.8541283284075155,0.9478737479130246,
 0.8955427417622797,0.11890703259727065,0.08219554258683393,
 0.21487837805812193,0.5261110255256807,0.6699515563527225,0.5929555926926847,
 0.9318093692796625,0.059383912290120744,0.1569897726941094,
 0.18040984154023043,0.5677716426811723,0.4805876853424186,0.9095735182272566,
 0.49625842589875735,0.8296244489885986,0.24424445455298016,
 0.19714311764927506,0.19575661149087475,0.6003250556271613,
 0.9029835556254925,0.40157616337729385,0.10027188611139726,
 0.6480501650372187,0.6084161050504016,0.34164553234628026,
 0.17868505450993044,0.2797315888109364,0.29758036215367245,
 0.9792589988016079,0.22659142311165392,0.8142338159335893,0.1917403751326353,
 0.7986805886888391,0.06444356748076552,0.37332665804269805,0.574070668929579,
 0.021360733703638113,0.5614280415868962,0.1945236931135489,0.293595841902755,
 0.823600475466158,0.4697247342127995,0.5309658279994778,0.2408646534046579,
 0.06134380463437339,0.8897682328858572,0.4278076436883276,0.4136651302168136,
 0.014862864121462316,0.744729310128056,0.5130380414437246,0.9657204389553065,
 0.33090428466676924,0.3110064468387217,0.2606943718390651,0.4321153814694577,
 0.22102490597517488,0.9134552041491896,0.6599526603286768,0.7434312413643522,
 0.2691786087416188,0.13307915775980772,0.007008147457859271,
 0.5613143961746214,0.3761923936232616,0.40206823600509634,
 0.46156427712100623,0.8256586585903635,0.12271988503404185,
 0.48358916455359535,0.6270934996538109,0.9402018988434484,0.2527453960928595,
 0.7921913130289661,0.17404996541536633,0.24300763732271846,
 0.6476298633328238,0.49059334050322057,0.7056731026670843,
 0.34763634632590434,0.2564532642601999,0.5950461776218793,
 0.14957445538732128,0.8258182695145657,0.9760182587740968,0.6315525251349687,
 0.44298408411108103,0.43237822962860517,0.0587186723314741,
 0.6943763288909498,0.7069377467036367,0.5476988575579634,0.49349347670970334,
 0.9318826568884312,0.5048652439671492,0.4126080690001662,0.6897272070171252,
 0.01499353152078109,0.9731348513914767,0.045181716904050484,
 0.020154772485704964,0.2102791082882295,0.4608033137203076,0.571747853586402,
 0.7391396033078287,0.6092484756438812,0.12146360768968523,0.6912786069824325,
 0.34389468243378174,0.9175654453782702,0.7964009409555229,0.3625729974344778,
 0.8675171373400744,0.3389057894052697,0.9262147916554966,0.43658858014242263,
 0.17534338487645973,0.33136340485780524,0.993422209897216,0.7964401315764205,
 0.4792650656162608,0.281025079344117,0.616809680192622,0.3510465608027351,
 0.9620162137248425,0.46682252409120617,0.6927597464735573,0.9712802732088858,
 0.9247422806358702,0.08570957174127025,0.36009362887398755,
 0.6039848889428002,0.0978411847607239,0.3971459592559632,0.7180389508412096,
 0.3510363583138877,0.060427581647178874,0.5508003034126703,
 0.7499711529690039,0.4368679166213105,0.0777326726876223,0.5610451193725705,
 0.1442024778215938,0.57263713155657,0.5842084054237889,0.8851972470368752,
 0.27168537005319204,0.38110968342639184,0.3576788869398865,
 0.9973241248384129,0.08999935873658449,0.05910799743998951,
 0.2352879839874391,0.568319135820476,0.7756869684271682,0.39932524981853035,
 0.26063162634184867,0.7290856706841933,0.34286314610364566,
 0.17967831600589523,0.7410435261787257,0.9480042136033092,0.4850177488138706,
 0.8128041751859071,0.4643730303722525,0.33991474382759035,
 0.09013819492726538,0.5810782450853443,0.3971039459450221,0.6859428905633213,
 0.6222122513329553,0.6623246140713703,0.8354825853189859,0.8261821632806686,
 0.6598168625836374,0.6317797834087178,0.23629500086599897,0.5124531109940134,
 0.2629420385429886,0.6898485718159721,0.784666150835409,0.1395860226151583,
 0.37512663717015404,0.3330960839921101,0.8120342697384048,0.7481238995939704,
 0.9366478609613768,0.5309341129334761,0.8615101213734458,0.7287403145533816,
 0.5442740603924913,0.5906203920960983,0.792501651853501,0.13066498584827102,
 0.9089387320580375,0.27553801265337985,0.6814868258475635,
 0.46603057676028403,0.6515795180871589,0.9191577400793549,0.9235909670509154,
 0.5588081645767671,0.13790707942160474,0.13387777259828149,
 0.5693376867494807,0.2390070418730782,0.7659199202726048,0.9482871553660871,
 0.8291468101058128,0.8988088050797003,0.04379521753560378,
 0.09100706115587287,0.5699529226868609,0.9849027085989014,
 0.18823369021523684,0.2600319970697177,0.962897727409809,0.11878584040065354,
 0.06457569042977407,0.007840758217634414,0.9351064579694491,
 0.5742049905952961,0.8591374670919861,0.9731656439117627,0.7732095086926265,
 0.9590926758853607,0.3141155731532843,0.6188741407025607,0.18599618625318026,
 0.324601687953746,0.7181487562792944,0.4599334829355721,0.8119636423189431,
 0.21247744558956505,0.6458814095944824,0.8590435530217194,
 0.15419673881504214,0.5443813760660395,0.06564251562688839,0.599528478030484,
 0.5761504820701286,0.4514985040856563,0.8702237999556296,0.6039483903116688,
 0.47865675829968246,0.32812850446882336,0.7458354867546046,
 0.19815389907991732,0.3025631173231321,0.791878846158364,0.18708579100723333,
 0.6223501530898551,0.4687082359223598,0.165802102035976,0.04239744831840664,
 0.4055657396523502,0.10646132746246673,0.7291756484769575,0.8088571073009045,
 0.6949899182472481,0.5790447428902881,0.6370409550896322,0.985344939713239,
 0.7830421763117908,0.6349730528712819,0.6766886363901701,0.40285555978795673,
 0.5844772056810075,0.6720408689770629,0.3198274062986838,0.22974529860523707,
 0.837678436799149,0.8166907257328706,0.34225613660829446,0.775277952269401,
 0.01053772842080436,0.49488957058843885,0.8859670167898792,
 0.18016393931714558,0.45070554577926936,0.10060210509197809,
 0.3588999580810648,0.772404756931979,0.9548114324112517,0.3362611746302795,
 0.7240511358546516,0.4956471086972556,0.40511074138038783,0.2513314650396461,
 0.014862289779890125,0.7119951110239642,0.9382887080624425,
 0.7040870922553564,0.6380197141931206,0.6997504296790128,0.6376364601191419,
 0.3149865031952762,0.7087280692069476,0.5913968160921057,0.8307462330057992,
 0.9740985873033997,0.8081276951885784,0.023679568873125167,
 0.04603417113707864,0.9229000919375023,0.12383781762004242,
 0.49302340182828774,0.49527923587453937,0.17341826339246935,
 0.34781262311227934,0.3600666097514046,0.048564954064067245,
 0.7016125541285222,0.2533749882050711,0.3438309098631168,0.32382972469868754,
 0.6011834634598752,0.1335653039115392,0.9973191681821664,
 0.013360220246683596,0.1715478586596001,0.9452459608207152,
 0.32623558799513097,0.5467468542553434,0.6373460065350827,0.6844346428234928,
 0.7805849240924134,0.6040877893822274,0.7791096370773025,0.68320858817042,
 0.047983975790496514,0.3364408480974789,0.9351603906545828,
 0.8407481411137778,0.8166564795549066,0.9274981444441509,
 0.006417994155576978,0.425989428339542,0.12098501107463067,
 0.5751394472677012,0.7653774286996816,0.6808217311023147,0.9767643087311939,
 0.7341726203956892,0.05008672435569661,0.2741793714056413,0.9162264707896065,
 0.5299962224032387,0.5230517865340787,0.16099748757265708,
 0.25793312133445545,0.40309178418968883,0.8172083082387932,
 0.019381302923193955,0.1460487000463725,0.015155529105701393,
 0.33442518789356246,0.8755888327479282,0.3389531392474032,0.5605936039140903,
 0.10496850009305159,0.2724792580778701,0.7333565765683834,0.6468812451318092,
 0.004911344046973798,0.5782683610013867,0.12029055743882822,
 0.12804571841262158,0.48986961307254584,0.007976632678142304,
 0.21668755822375962,0.6553602240328997,0.21770470496102057,
 0.22640120983201872,0.33729664614942667,0.866019898763599,0.2542286755278653,
 0.9317327235259557,0.31292479433398035,0.4790450049391397,0.9566782405793709,
 0.9220892835605768,0.14268465023063304,0.32956804260731365,
 0.8394560002941119,0.12327663088874952,0.6536342107582973,
 0.15792523153002247,0.47660405678422046,0.2553952074484991,0.892324906460737,
 0.2635472740354853,0.8134068035384097,0.5167776242526897,0.5695460016948393,
 0.9137655816327673,0.029132673974350265,0.5101555379243609,
 0.9366600674839267,0.0836389447530238,0.6385430989432551,0.5203669082250891,
 0.6910020692325534,0.3490483658452169,0.10337950568055165,0.2043363165704234,
 0.2869965605136503,0.02358878423349564,0.5305073718413363,0.6311340577279912,
 0.926198733285716,0.8508494066278774,0.3272511272645041,0.6367069320148109,
 0.5408574661699987,0.8099404265206965,0.41887229036505047,0.9302837863274335,
 0.3482482789102691,0.4225172001417935,0.8258943939375475,0.563500160702362,
 0.588560373551241,0.6773538230689652,0.7835148952434983,0.2408280582158483,
 0.09425553775367712,0.25505408344177316,0.7186899913205438,
 0.15190751221750265,0.47587599701580663,0.4197687840906009,
 0.7890281648995667,0.7285712910168052,0.9505060658320403,0.7031257842527946,
 0.478209165582523,0.22237207116861957,0.8107801332089886,0.08465159193811145,
 0.3787494757945383,0.8688265633878652,0.6020013639372266,0.3794511768683221,
 0.13316124974649968,0.9631894237458671,0.1730443360953271,0.4428291925800414,
 0.8130492862456029,0.8705441493791752,0.6399120566059093,0.8854448606338413,
 0.10592414751712587,0.012178919148029932,0.8489497270238626,
 0.8024286221333198,0.8966641254956603,0.6111001031391734,0.7566168496515052,
 0.9489578056074867,0.6298460011002168,0.5241657169416768,0.6663631805185523,
 0.19972858086175138,0.4374908622911615,0.3378420043504018,0.6362523950335593,
 0.5944534120714935,0.5455645298991993,0.44257099854986304,0.5232798571597042,
 0.6722148825769134,0.13499539898114177,0.23365483995681413,
 0.5727533000457975,0.6584675346869229,0.9441427679273411,0.19948221899079177,
 0.6147857631563041,0.8489532254381024,0.3678029396021093,0.23831450850455493,
 0.8766834194692966,0.28874437888981397,0.31272992914451503,
 0.13075320547547886,0.4084988009507623,0.6653852846744549,0.4899834852910988,
 0.6091946234077299,0.32382026484468396,0.8593105158549749,0.7308877894164219,
 0.28522025291334785,0.5273668793653374,0.6579633436468705,0.4764440047701891,
 0.3405181453380641,0.7590700731039083,0.8781735499276615,0.4104515202861717,
 0.70697483110397,0.07898949020473445,0.8201758135542289,0.09208421541987166,
 0.484613905205882,0.06253121046106469,0.7398084966000976,
 0.007865828375268391,0.09885251279584428,0.2482525224530926,
 0.8754793191809043,0.6444237442951113,0.3460207283021113,0.8312445534204126,
 0.2006775993711173,0.4382634277680225,0.22825208231176952,
 0.15694352075570495,0.3872040583262786,0.14511352968765479,0.247096228627793,
 0.28300491188237487,0.6117017883545416,0.9899570892438766,0.6104272766411715,
 0.9622053313079134,0.08437826143987048,0.9779793148276367,
 0.40318639577808046,0.07329483518548174,0.17271443363248873,
 0.7149108614921451,0.8344815449497917,0.7747178483810859,
 0.028369002558296885,0.8119498996835004,0.16569829418151605,
 0.6569833262942245,0.9536791768206005,0.2677844593060683,0.43896579804606184,
 0.5797370945901512,0.8735322059871807,0.2278258495872999,0.2839250180262569,
 0.6244452359827888,0.714355699370686,0.4150683288179864,0.6207665674336675,
 0.24914125592128977,0.27234748336526027,0.012131095344592224,
 0.5380624920126618,0.3243069025161328,0.5867539582557637,0.6668375815642615,
 0.7081684668304922,0.5628795687569673,0.02061223342464691,
 0.030315989792582765,0.4842082904707108,0.686457452262484,0.6036683465324002,
 0.18846535689000554,0.13525695360423817,0.8411143933160106,
 0.7213790796022442,0.5695822453805249,0.6765591263546747,0.3552290889094829,
 0.26273259428764806,0.1182265144172776,0.07249415571014484,
 0.7757237462703672,0.8797002685018831,0.22226230772466982,0.4668115733180227,
 0.2686104009095891,0.6522776056508144,0.06356465888406071,0.6113244254545614,
 0.24810491210008845,0.2453573600819129,0.8116777902742942,
 0.040253955935351526,0.7505103609338408,0.2470659436520044,
 0.36529761195568533,0.9763603287601332,0.8496558470589474,0.6174306950721198,
 0.07985428777065429,0.3736796435481903,0.4611510321523965,0.6477076120634151,
 0.5645082808104518,0.2388160544761988,0.24832167331965416,
 0.15656253005699095,0.13973259454691178,0.11611495740055289,
 0.9418978057509738,0.23572181275402793,0.4282871018102762,
 0.044786481248084264,0.003563673688561697,0.24638177987285725,
 0.18942426378906718,0.9441243438967105,0.970481004226478,0.739611762813484,
 0.08153575800691815,0.8166102923970959,0.4734266761108993,0.8104257314356951,
 0.21349777839542716,0.2724516737458397,0.9990060068251103,0.9940907624408352,
 0.6285892596136655,0.3370563090362968,0.7487944407242973,0.26925307151672573,
 0.004126093232692307,0.49960698889889876,0.4321614838365251,
 0.3258208437344592,0.3018530663153025,0.7274203040468173,0.317433105612705,
 0.4543974265093238,0.4099627306464515,0.6027981031902535,0.7696820620121986,
 0.6419087777289494,0.4497394050294563,0.8493878077695178,0.4652916703315366,
 0.22663104105758602,0.5873370807909084,0.7811302650787174,
 0.23069296482161517,0.08638976675888521,0.6640190185562441,
 0.4301078509172148,0.6159349140397645,0.2561045009817663,0.6582446393185498,
 0.3745346188900238,0.7495720967426448,0.6669558454967586,0.4389972040739408,
 0.581043925944333,0.09637645281272633,0.6975075425962751,0.14835170741810422,
 0.4931437312036395,0.49682620178659676,0.6761774434717905,0.4856619473255839,
 0.1848530836375758,0.3736505989469445,0.08643823635942027,0.3956002558128222,
 0.7269061646951801,0.7778573774704883,0.5646098732182293,0.13684760443276667,
 0.7265545043081492,0.384127218639009,0.9842753949696671,0.668512001436551,
 0.31614846971142985,0.17829961735805633,0.8736951320351252,
 0.6601853446193394,0.1179530246532211,0.040922043043795875,
 0.7831393143896045,0.2203077037821024,0.5575785222900691,0.20083494425838033,
 0.3836349461485993,0.3705584792684281,0.9107497120726524,0.03215292152222082,
 0.792368533719629,0.40696280614828084,0.19221770075263445,
 0.32820285153895834,0.1975445133514142,0.35901762099746637,
 0.20661230274765263,0.9537301120514253,0.40027432326813484,
 0.3784969806702666,0.24616448150158354,0.6711730712898966,0.1264762184998095,
 0.9798647378740987,0.7655240463037564,0.2938609302049666,0.14709190210673773,
 0.5826847488354572,0.8168620830527206,0.2421061944869356,0.9933619554887848,
 0.49558135197461095,0.8341274933710703,0.4836277728840567,0.5206181492616272,
 0.9910497937519741,0.8547347796370185,0.7102224553174864,0.1767120493894243,
 0.556278440961588,0.3798636178258078,0.640430626497842,0.9376852236388135,
 0.6695357647458502,0.566377146150366,0.0019205079522057744,0.491724071377997,
 0.7793944181723131,0.7361266215022898,0.7222575345303497,0.9042359974325523,
 0.6642097708969742,0.9875945367110415,0.17201990940830258,
 0.32170481278810303,0.0010286767512891482,0.418798021299308,
 0.044444901577775586,0.5433070881791247,0.22324267837238454,
 0.9693900058379205,0.1819007099503277,0.009248230455206397,
 0.09158130402661624,0.950544235737627,0.18750247722120317,0.9609712772598994,
 0.9369813026155077,0.6206179327106466,0.03201104870975091,0.882672820121735,
 0.7080779649559875,0.995927407828892,0.5118973123954085,0.3494735970759285,
 0.5511779468586635,0.8362506951628781,0.10954351234530535,
 0.42692787124327713,0.31523413518811527,0.432844246593877,
 0.27666567038572365,0.3723663607227403,0.5498680566095239,0.4566025370825909,
 0.0893341480400186,0.23744282208980694,0.9795005502184098,0.5943708423304181,
 0.6921243152428892,0.5368415223049906,0.7119023522867194,0.36190972419481615,
 0.21653410935502293,0.94931453812718,0.8179193169585035,0.02230156044501841,
 0.6194318337666098,0.5681230417927632,0.2107328307732671,0.3473880131494437,
 0.07420257520234996,0.9418472411429618,0.8439942899851071,0.726882855965343,
 0.9353777835876251,0.04032977806770721,0.05223728500022684,0.96138933227252,
 0.03501383295376792,0.31693028219345265,0.5212050707854006,
 0.26613457254185624,0.18625993838302002,0.8197474815813625,
 0.21837733694940065,0.3575887291432136,0.7241372386983072,
 0.24161532728764323,0.15917378501997426,0.1327804664998674,
 0.7313278922685155,0.7736764360775296,0.1737846660496355,0.03540735792399674,
 0.86092235335584,0.903889444940712,0.6635591925519979,0.1137488593523649,
 0.5002849173213911,0.5054486044519482,0.19006333010934628,
 0.28827118493387205,0.3812994760943412,0.2236581321675314,0.8533909120988091,
 0.6010754473884192,0.9802430037509624,0.7319760994655079,0.0614599714776598,
 0.6682340496996977,0.7342528932994106,0.22103609837333194,0.8064356977740498,
 0.6638968823827622,0.7264976823827887,0.6116937548447863,0.26261024430006363,
 0.7396677786941057,0.11406890896698951,0.7339155273582366,
 0.38493972724842074,0.5070395849234238,0.9127999731656019,
 0.27182222333462547,0.060427294282224375,0.9107904883914614,
 0.8849602620081835,0.5441609972330141,0.780096088069233,0.11790157002572943,
 0.24323055599162502,0.7618871068834958,0.38493601654488463,
 0.6201611530357994,0.07526701133670277,0.48094504755255774,
 0.6593625631135479,0.869259768224201,0.8824003081760089,0.3521956017534055,
 0.6058932494254549,0.6820229317422485,0.5337270091652174,0.5441069669285983,
 0.46439969731786335,0.4078114413127055,0.2009061282780874,0.8712184198964698,
 0.8579809848261561,0.1498434326762017,0.7970081643548057,0.1952126217063349,
 0.48576713024231777,0.6899199912199231,0.20142397528184663,
 0.03698302023598288,0.7032600654022713,0.21028060216193567,
 0.7704074879815365,0.12765844931332238,0.4211049553818377,
 0.17559260000459398,0.7690280463558731,0.3386556288435918,
 0.054601170527909804,0.7320727337406472,0.49997340335749785,
 0.7814800189706057,0.5694448972846888,0.7813752300651181,0.6446479600121203,
 0.11171505871589882,0.07991767647214432,0.23032409118987507,
 0.7443802268294201,0.8282933164499945,0.6383478579408627,0.4056561452403532,
 0.4391593559095901,0.6425328129052409,0.41080253832477254,
 0.07079318465967877,0.12911317893172958,0.6790946849152067,
 0.5513985964225274,0.16273369746674093,0.3212238462858751,
 0.23459598769653944,0.2264234302446716,0.18520332982045817,
 0.15076918353070556,0.49469615470735095,0.5114154619871848,
 0.08137847630465733,0.731210213495078,0.18167849129407954,
 0.35970769489097987,0.8775221429945462,0.34262172082992715,
 0.8728161076827932,0.9661623436214948,0.18703462633888324,0.9067908376582862,
 0.8417921117040194,0.7774455196613853,0.4260958126004578,0.2095792309873461,
 0.14541403711566347,0.575326684845646,0.21831879287285827,
 0.45282282104390226,0.6553349189472931,0.10571866798673113,
 0.12493190176567448,0.3932681057850328,0.8782268730490846,0.8598935698969385,
 0.8716488227380402,0.9809552753585339,0.4889017923679253,0.2398371457094136,
 0.11297295166255072,0.7064557793489629,0.9813361668525409,0.6175796085548413,
 0.9040263598802152,0.26083675943993656,0.8228421929842495,0.9345944532226452,
 0.1675685103634722,0.9162856221041911,0.27613860231719467,0.6690898106282464,
 0.8907082073068082,0.9450175880498528,0.036929115755870545,
 0.9877927589870964,0.8893772511453288,0.1382930575410195,0.8095917978053533,
 0.3034630591029942,0.9978079788092354,0.6305000324593968,0.720736707727759,
 0.9754832718831592,0.8395036297406571,0.6166825644655625,0.7588390040557526,
 0.2278183325576003,0.5653696681633097,0.3045312929537619,0.1332820190969778,
 0.8764318748974687,0.0830863514856428,0.7066964628578078,0.13814104244781245,
 0.43942891478639945,0.050048020407137184,0.7842000025932168,
 0.03318458865954721,0.47935418150740416,0.8101504078446985,0.190426381039694,
 0.9921475380746986,0.5364064452713628,0.146734116589886,0.21757639904323767,
 0.2679512484498149,0.13541401594240732,0.5745708018106361,0.6234009557690697,
 0.26849668107192226,0.8686937106145128,0.44993894661993206,
 0.04839065029601386,0.0973331051739148,0.30052215846872854,
 0.8304833968638233,0.6477045409133861,0.1796776756760914,0.5961341025318492,
 0.5723070375026063,0.7999732789659059,0.9302920334702216,0.7922447746333589,
 0.9708706919667722,0.03463043998979076,0.6737442147171114,0.181406807892313,
 0.27517942404937745,0.106170710676606,0.511796442643038,0.6465481558210144,
 0.6398760810785031,0.21333223563772208,0.42423178123339245,
 0.2988278941695959,0.2796921062280915,0.23361032864657993,0.3632178173145306,
 0.9578546955776379,0.17467646411101345,0.6228077605978569,
 0.08305571019807678,0.3592501577375088,0.8236490030090533,
 0.059229945442838994,0.2725730956678556,0.7655695704753654,
 0.04554697361013438,0.772229664903031,0.39007895934877646,
 0.11320222296052063,0.9645328961564353,0.8030131092529623,0.5553688575653415,
 0.9315939128484461,0.6163714065922221,0.9113879745585435,0.8866560806313979,
 0.20096730221704695,0.6923641108776963,0.789456149944629,0.21215844868646594,
 0.2595063248992895,0.7171535034135952,0.09128978839518997,
 0.21584898724830737,0.47677505636915385,0.19642807921348926,
 0.5094589389331319,0.3218015129878372,0.9283743191671917,0.22209173665962467,
 0.2257533428291525,0.8255092003626917,0.7638258797571109,0.949248128178058,
 0.4894113069103646,0.4283077083806891,0.3259290585583897,0.3423091729100751,
 0.6475773682152554,0.8777930019932598,0.05552590316972683,0.2647563358119652,
 0.8862969215655433,0.061288577482822126,0.21132300230787315,
 0.7636239249714298,0.9845356530333434,0.5094676176998443,0.27575161667686443,
 0.6366654841600827,0.9310653113805737,0.7566346492644552,0.4518305876251414,
 0.2299296452638111,0.02943144095155248,0.4730083869547056,0.6534929503034834,
 0.7381181187713035,0.6130248647908874,0.5248879759458874,0.5293042931992754,
 0.8655770325824397,0.010325074295928771,0.36905316129857446,
 0.37197667817637803,0.9050900991229323,0.6697271682520274,0.9798321650200729,
 0.2656251135622453,0.512527207682032,0.2571563889288324,0.721613641281354,
 0.44873115859301216,0.16989256900156624,0.49945079496151634,
 0.7828502781605585,0.8326349520977772,0.9642903020367316,0.31312079778444224,
 0.10629323266459811,0.2513822825110995,0.9462956792511902,0.6160841304043772,
 0.27465477220844403,0.0887668997256823,0.915084750404406,0.7790246712066463,
 0.3521973219927833,0.8933179898824952,0.8410592288585301,0.14820282506312876,
 0.2615265248540676,0.2598053954247883,0.2488319577804199,
 0.030254138850021373,0.24006208831170872,0.5155947747443046,
 0.7785358602900292,0.42520813733351626,0.3860296952266511,0.4331305622339202,
 0.24143476012351162,0.9166511291317359,0.09183615853417493,
 0.2737351908464327,0.8342981209633848,0.5478981444835022,0.5475992645562161,
 0.2595384513742589,0.31562772403186,0.45357001695825583,0.566992143959094,
 0.42912324629126075,0.9613952600736193,0.15826874010340364,
 0.3689295588382959,0.9996290985025451,0.10975125936480168,
 0.20782343906896072,0.17199498389655976,0.08696518134167464,
 0.23385331474803317,0.610542760691824,0.5745229953331835,0.4717264182156742,
 0.5086387639788248,0.37827037916315265,0.24091493716655754,
 0.7834701228366107,0.039230938153085426,0.908075331834413,0.8239286914946113,
 0.5923270854004506,0.29247120293338,0.28297463248945964,0.30841206182431347,
 0.6498996194752208,0.11606473722223398,0.025767963004462713,
 0.579371659705426,0.24375329119397926,0.6450101831724758,0.6237362185929046,
 0.6108211958594892,0.06886056400562057,0.8602958377328253,0.8846864374863914,
 0.32234154443415985,0.6599580001355493,0.005047254870915396,
 0.5518074296754845,0.1297717858325087,0.253995440341491,0.742193018914553,
 0.6033350414999535,0.6185957034875076,0.6293473036851044,0.12188524403221968,
 0.38292685008516625,0.2969275571654906,0.498178686548178,0.15646289905135635,
 0.6878321650544553,0.41824195137021436,0.598979425271272,0.8955188197700406,
 0.4908941756800572,0.5412852078083341,0.169094191768146,0.3049835455995582,
 0.31436456688649406,0.8814666840726124,0.2988365797286229,0.7535223132042681,
 0.8658431991226575,0.6545917168806947,0.2776261251331985,0.855504449254473,
 0.16897592065461997,0.31029226312456426,0.46472772232099147,
 0.5222948820247829,0.21825467366220652,0.8753117529515525,0.1524749950699259,
 0.023525787538386078,0.2531120924944894,0.47101322013917113,
 0.2445730350836679,0.8458906743424149,0.449790727135277,0.5605159847494647,
 0.1485832416196844,0.10344809864610993,0.192857558858181,0.8988488627988087,
 0.3140053109001878,0.9320445174293275,0.9446312144522631,0.5537048349021847,
 0.4844558220802674,0.1027298646902638,0.5818838480468673,0.5294306623698669,
 0.3380251710691873,0.9860409680199635,0.7184217312404293,0.7096187492626204,
 0.4837991614292748,0.5441044628069054,0.9400090090058965,0.7969495533558623,
 0.8225676461783356,0.10762812671411515,0.11090751085864303,
 0.5164699540542624,0.21451428630261093,0.2014634334023181,
 0.29683003583512846,0.0951417324470496,0.695094294445275,0.31566400514510273,
 0.020613279000698426,0.2601133440806974,0.14072625464298716,
 0.1226859000932754,0.07157988336675125,0.9583590805905379,0.5915034814755191,
 0.4723089121947388,0.14687403654777142,0.8518594362429943,0.7880554610812764,
 0.0794685540451997,0.14463209046455272,0.30225248235255076,
 0.9572718671976022,0.6457002493443637,0.7040454159176056,0.2898996533376792,
 0.5377847427336185,0.2864915167080416,0.45973608918290765,
 0.11157530112324521,0.3777509866660681,0.31173553252307973,
 0.6045473850248895,0.5771066860983184,0.6435287399119998,0.07583265707743292,
 0.5724505024826644,0.53651084092919,0.8435776990645218,0.026482701842688527,
 0.5273820391699123,0.804835550434196,0.11378189847075681,
 0.041779578169280684,0.2715980261016967,0.01357952892054004,
 0.35621442380713664,0.22140410306532554,0.5760942739570916,
 0.3788970138571063,0.9497045319724863,0.09053466795553433,
 0.46722272440151513,0.8714953967990465,0.2061601249281716,0.3046546954326834,
 0.7027586794362055,0.7110540678446249,0.04780153860050562,
 0.018463396820004974,0.19151980207433483,0.5202867238123872,
 0.4331449071590686,0.7072411521655517,0.19479465181678846,0.5643104842033306,
 0.23998693268447657,0.7926409933849775,0.4393774386638155,0.7088307660862734,
 0.8854351714768899,0.2868905372912335,0.8768880766623508,0.872873417076406,
 0.23639151696000849,0.7908381980084148,0.5881812221565419,
 0.20042472470159378,0.22595124442168635,0.9648448398642058,0.966223442213396,
 0.11754252259907072,0.8164674369454199,0.6550751173227253,0.8380497631531074,
 0.8724137090253657,0.21246002149773935,0.2411291200850968,0.5831353937621925,
 0.8075407159320797,0.9235019953010892,0.9882069074479671,0.6507793557887087,
 0.34502827964173544,0.46615331768857526,0.662313512226617,0.5094172693235536,
 0.30504654515257335,0.8523580841609791,0.7376180041986589,0.5736586410749859,
 0.7904555099550188,0.2741421620446206,0.3545817013295006,0.8123621571540941,
 0.1296999322111979,0.058637861384514545,0.5970450491315736,0.717546024142173,
 0.4476806957356031,0.21115529684573886,0.3010378825878125,0.5813664655449297,
 0.3200715606366745,0.768899106938137,0.150195483811437,0.4665182386037371,
 0.9344147862759952,0.12406550233353297,0.34420056379117836,
 0.2916468457295257,0.31299777321610234,0.8512457669732182,
 0.31846049565888324,0.4989964765581536,0.015845523790936067,
 0.288275551146847,0.7307248099210831,0.03340988909644482,0.6540990284355618,
 0.1733640539762188,0.4606335095862806,0.28015240396375796,
 0.15373251828487577,0.09653418618403686,0.44965895937513245,
 0.457212800757431,0.9699909492221578,0.7152406735918616,0.5228697404523676,
 0.8329395737539446,0.29531373312535203,0.2638766435779931,0.6036620054681763,
 0.3130258858074031,0.39125911390820545,0.25314867268334473,
 0.7880689741778497,0.5487472030933207,0.9190276859925903,0.3472990139814738,
 0.48272151684404074,0.5878503128493671,0.6052248822450101,0.472625138794563,
 0.03648634461967726,0.8365593681385816,0.9450741968316344,
 0.15952245051998615,0.04440940674126059,0.38609761328790515,
 0.3669338435560121,0.4954561759909645,0.1669650920764948,0.42620697298281307,
 0.0574771440421753,0.45314981264824583,0.10938485665694031,
 0.13374619141135835,0.7223559134553749,0.6851994286925306,0.4033937815028933,
 0.4854976908025054,0.7812559190352891,0.6875819419293087,0.11438760351558808,
 0.774132729321163,0.8099143711215475,0.9861924203063914,0.5875661729655236,
 0.445990728087879,0.8934421779559248,0.8088540547169697,0.8970398209969856,
 0.46208474365757435,0.3654693338164381,0.4521510664663726,0.6589718120992778,
 0.5487684037713866,0.8422683584754209,0.18525746886041006,0.8264528050119703,
 0.752549920496654,0.1571676298979916,0.9750401745510903,0.31120625420797954,
 0.291538608660311,0.4178885271888657,0.09745876596968195,0.4865645901011417,
 0.4416840942793713,0.04389365919649724,0.4039027051949803,0.9794180715991162,
 0.124866288218487,0.7612115863269227,0.8473391535743501,0.9241877583313965,
 0.5303463383989737,0.550193581178933,0.17650085020142248,0.2301637746911419,
 0.7306749798994597,0.9803345184299996,0.3808865334292112,0.951608013474364,
 0.9597282563387569,0.4833052830676082,0.4498967357811653,0.6260105688916034,
 0.48405350028790606,0.6241152329889784,0.8510132258874498,
 0.36343963184239025,0.6856256560828807,0.1859917443179443,
 0.05267138028244023,0.9111655218604366,0.9797983387663823,
 0.011562336681343766,0.4128181191622078,0.3014164770228266,
 0.7674745392351953,0.9830079596681753,0.3852545173351183,0.08725397911590183,
 0.6750862769527551,0.8898299474306542,0.6001605910863669,0.46270154311201706,
 0.131091058977175,0.4019747385018012,0.07011732151264394,0.7268022851288778,
 0.3210434124045429,0.2303402314527958,0.9760231047502654,0.18612894789966794,
 0.9544213160256234,0.13841160576958844,0.11773435277957378,
 0.7267821325062575,0.8440924663546241,0.2328534746195844,0.46797600651207716,
 0.6192324132103516,0.2373800980779449,0.34499871689757433,0.6070748970628846,
 0.7134376169792512,0.0812108494027517,0.37729175577562213,0.507253698257382,
 0.7306988710496274,0.5226048528473066,0.7608281306683047,0.6398717995702549,
 0.8551819770704722,0.3351744540600903,0.21096070541666667,0.6805851425293098,
 0.6842600197206962,0.11621972833087191,0.7666519437399187,0.469582552921875,
 0.8268289939568072,0.2119842943850948,0.9306780709285801,0.8719658857255681,
 0.2520059169256931,0.12230446062031874,0.4223552588904177,
 0.41083532447059035,0.8191098836009914,0.24635573679821743,
 0.013372614674831196,0.7449471437910662,0.7813663210915104,
 0.13509257422099408,0.1491409253171594,0.7795071806339456,0.7213521923498547,
 0.8691961393341494,0.8496845256822784,0.25353451328934995,0.3738584917112282,
 0.15985323291691267,0.8734958054689738,0.9474805258807277,0.9599742052597737,
 0.31514457039437194,0.39813706191830367,0.2039018621439065,
 0.4342520351221054,0.5386924544301678,0.7196975069963276,0.16521463800522995,
 0.20354960132411914,0.3943638804528141,0.013700097605711115,
 0.2959757414091715,0.33130183165135063,0.4474304948143111,
 0.34608735974782523,0.16210704156770706,0.40368259078942703,
 0.5369983382723518,0.9292959031916395,0.3777625483959547,0.24953957588309172,
 0.7728549871412334,0.35102350554645056,0.5578927323580889,0.6606008917521758,
 0.3196755160640741,0.039620025700825456,0.31284797860230595,
 0.05952189355699744,0.527640591215929,0.5249763315975868,0.8136005808140288,
 0.8233700275661593,0.04760823138701176,0.526958346826533,0.2904457155287308,
 0.7140492149688585,0.7175430444798567,0.4341975468792435,0.881756221649141,
 0.20658493102439457,0.2828998007506204,0.5820095237420456,0.451794364677107,
 0.507368298628231,0.4628946415779087,0.5745366345479608,0.4676269375345995,
 0.30280825128505384,0.28097970205782596,0.5069787478585663,
 0.18201631789524653,0.6074624783208072,0.8258378997590502,0.6726640758110428,
 0.13029057694832136,0.8126999385858755,0.44505478294411405,
 0.01325244025338268,0.8489016534089214,0.38841693821062406,
 0.2883057182984471,0.25175636078329655,0.7291739610423387,0.6748443707814555,
 0.9295897479423461,0.9383756275905957,0.18663343322496861,0.404716493085017,
 0.9327661542988595,0.9556867147883056,0.7811311645970125,0.3003272934535809,
 0.5996982001130527,0.08617912384697413,0.6747994388649599,0.7139589768481015,
 0.39754263542888013,0.4647752350005796,0.1327336038837652,0.9684752373984297,
 0.049738392383748,0.6724948980489978,0.3452735444745346,0.037314582019464027,
 0.5599011127321506,0.7152416804857226,0.9549169596180058,0.46277920552356,
 0.3187107634023463,0.3431087975403092,0.3100106232844799,0.5588452527843898,
 0.2174172331068488,0.00584581746723889,0.7356948476926299,0.2651652956040268,
 0.5477471883750081,0.30779042565466364,0.6677589167722946,
 0.26301640455174835,0.037474665465340506,0.09452183929098146,
 0.45673859353063984,0.9769470744418196,0.5282034386585771,0.9895976254027756,
 0.21321585143310373,0.1382838708846661,0.4521282729119934,
 0.48023182545487975,0.7944301138867074,0.9547544586105979,
 0.004013983296794699,0.06496638195846915,0.6474741565688623,
 0.9284957150389512,0.8558487028816305,0.1355200551423129,0.4089302137025963,
 0.40326638162908135,0.9534740114036697,0.998289064919015,0.24301276100376967,
 0.6195880186177034,0.44940189493400595,0.10179680423797355,
 0.9662244375015774,0.5068985340452365,0.6538225881092754,0.3320261228923779,
 0.33309400359165875,0.7838586759360338,0.013116107068546046,
 0.39647802180097924,0.08131310318920071,0.05032376701132191,
 0.11785373544540279,0.2700436941238096,0.8528490003347484,
 0.35717860578300187,0.2915823060683206,0.3061152424068887,
 0.34045634107793876,0.5173649541164346,0.01148354448155986,
 0.8065593886291529,0.8625201448888465,0.50947351737802,0.4025114581004252,
 0.2788526881190706,0.9739182566218162,0.3005079211501729,0.9124214354593259,
 0.5688637632061894,0.2726795565823833,0.08529176487787293,0.6611889765445633,
 0.28713445417956385,0.974709993570552,0.403751633335374,0.23991649267066117,
 0.09144103143020699,0.00989895380729422,0.06289157848676385,
 0.23730312268365095,0.12509250146797724,0.4878654704106168,0.727892535422825,
 0.40863013049992203,0.3333281158019883,0.7611152854481147,0.5723404685767096,
 0.4415573190593447,0.12868172053961546,0.2114652448059131,0.1467704245993663,
 0.6741076264575306,0.8113422764504429,0.3701032802603571,0.3892240009172395,
 0.018995658676595806,0.9705459222119728,0.2558342864664336,
 0.9426381318104452,0.9923873471378264,0.49025892002151505,0.2336229427651465,
 0.01917108186895089,0.7435106605975256,0.6510031604801775,0.4557840089652925,
 0.6178096009313851,0.7600211437079173,0.29082599332227177,0.6682986009454809,
 0.5161063905658769,0.0979811509958668,0.19240631404886832,
 0.059839737103484114,0.3096729954069737,0.6553013904523995,
 0.27566171187415456,0.3897601236005158,0.7971346842729296,0.2596246764023322,
 0.7236298095268222,0.27014400708337516,0.5255962795238742,
 0.27106054291552406,0.03110951630409131,0.4082788693243743,
 0.19797581926579877,0.5829438889214587,0.676664779209989,0.02391723880751273,
 0.15581437655646502,0.8923644490685461,0.9828600126059405,
 0.22394901212494156,0.13460101212775777,0.16927605947117186,
 0.20390744290466056,0.44392294705730184,0.5575075965677655,
 0.1665244609282116,0.537080052417668,0.9136567706371714,0.5919599002801817,
 0.8715267690345769,0.15721476699908798,0.14878798633828083,0.984572152309806,
 0.9095680174289865,0.8631492744185254,0.6334718859885707,0.44355947108382465,
 0.008051102622973372,0.024257158548710245,0.11219388551726173,
 0.10701020580731613,0.5315892705303218,0.9320921349975939,
 0.44324877882958746,0.023521163293087954,0.19567609212207626,
 0.3595410987323866,0.4845718949005797,0.3725716188762742,0.6655847032732952,
 0.22279651195704242,0.4676142832941024,0.7372372092121446,0.6992883852436451,
 0.3766667709447513,0.28662205287519693,0.6373704128832096,0.1626086075424702,
 0.2343816423995606,0.26082367933862116,0.9897533867986213,0.4444974773399629,
 0.7711570000780276,0.07609730980004703,0.09564126009408191,
 0.5302621018610876,0.2178303661499904,0.055011202532432524,
 0.6283200517335852,0.42769529493132874,0.8241612146106941,0.600785647168417,
 0.47450766120187515,0.5900886384968974,0.20242049454309585,
 0.5185035570682295,0.12546038764343537,0.8639242281023402,0.5585033241921155,
 0.2424190574510371,0.8903669162811156,0.09103311646256995,0.2012629858553856,
 0.9869276822038999,0.39562394141955537,0.8564656592676012,0.6912419371467345,
 0.14815198589552858,0.5414501475172495,0.5098537580833172,0.4733979709317645,
 0.7270903005679414,0.6858219899433543,0.40971048033170165,
 0.03182984501806785,0.7562283739734375,0.5632087011675033,
 0.11271765387986699,0.19484674036343752,0.8552963112785722,
 0.010310491907544184,0.5603485477226661,0.8485387974094718,
 0.9685902088506639,0.048625692539787324,0.4963365889830694,
 0.9521483796188941,0.4431438596634847,0.045920506745117384,
 0.15737364192324255,0.5421053368149228,0.22898297631614672,
 0.6222825072604616,0.9812997497407484,0.5375121444935572,0.8821352464353216,
 0.14879374562124026,0.85122085168135,0.21706685327265918,0.7422031882346525,
 0.03402244123653231,0.6743992741951024,0.05522120427743793,
 0.18083206856470246,0.8537010178148823,0.4278738488982715,0.5549353240891134,
 0.20238762953405254,0.6455321074040673,0.4249082430297313,0.5103468916021925,
 0.46705146455733315,0.6053940752679884,0.9654511462135742,0.6563849269942712,
 0.6807070336764054,0.13577172748406885,0.10917811918560705,
 0.3974817274682019,0.5962546484327735,0.8034558357005557,0.3537235943790811,
 0.6528842575028958,0.06766049623556059,0.7088637110407087,0.3247832210482634,
 0.002455199879443426,0.3390092714683681,0.48848536116424046,
 0.3322892073834691,0.2026850998657198,0.4388104383809577,0.7279552269740148,
 0.10018473585380006,0.9590443576898328,0.33676481651986523,
 0.08813159651229541,0.8016312292881995,0.3782860068058935,
 0.31351312599014125,0.1737361226948193,0.9691320409246975,0.9147926505701472,
 0.7362875642973793,0.26619040976040154,0.4985737716520111,0.8889141377985579,
 0.745856371856264,0.674168357011736,0.11102227055575964,0.9000216488401814,
 0.1549681125462069,0.9409128260057017,0.38643045500591244,0.7980999550718344,
 0.33533614012463975,0.10648132973351465,0.6669095110801547,
 0.6840746736104113,0.7748741664485834,0.618763111882132,0.24598748376978719,
 0.40301292818961554,0.6586803965313965,0.7340275695496604,0.4870717306585718,
 0.9852865582725191,0.6151091941791074,0.7899610577832428,0.17307295200532846,
 0.8880606745431272,0.7257825540450925,0.12246106371126064,0.4469034203370633,
 0.7800759309082104,0.873916424081325,0.07339263337692281,0.28278005696960484,
 0.13377915871251456,0.1319309590384019,0.9043549878141237,
 0.48560459690151037,0.795845732201568,0.8738869238773617,0.5120694749012665,
 0.5211456231068039,0.9263411212709052,0.9792769225292131,0.4321701070878837,
 0.19168312943113752,0.13179472457907082,0.11719100166423024,
 0.17344313450904913,0.8576438464780464,0.9279816578841315,0.3515871439770575,
 0.5102105306110297,0.23719674825204917,0.5514292399171012,0.817842930862254,
 0.688459333945252,0.8080239077341368,0.4813397761924544,0.15120362629128514,
 0.7505382178594279,0.35627527401041514,0.31132025421410847,
 0.001408388478708611,0.23944286157700023,0.6074040334824634,
 0.9384934644709961,0.1292543935011452,0.6915929104160313,0.25845235333137884,
 0.3225731413576993,0.28568512102909216,0.19278822340044666,
 0.4428885877077742,0.7394072749558954,0.5581701887481161,0.24168437934127895,
 0.13399740278202144,0.6255776991582183,0.47386669008835736,
 0.46020608319988754,0.07644336975811195,0.6903375885353003,
 0.45656298257143824,0.1733371203291736,0.049493771855345536,
 0.2414494763977646,0.6794210737386751,0.5266289905503063,0.44209442703688695,
 0.3051443262921011,0.7555244189788699,0.5323004630385081,0.16013908168959112,
 0.23620030509099366,0.017956745480103065,0.6020994835649706,
 0.27885813121097147,0.7942293792191613,0.06639670321894997,
 0.5012623075062915,0.4519979036867523,0.03496589996090238,0.9140700139857568,
 0.5132374622002598,0.6774301494085755,0.7751493312877195,0.8592920828977336,
 0.45711518903572235,0.3108766501141962,0.6008383828614957,
 0.46733085520530326,0.6069066633924336,0.5330375178308249,0.5600862920192666,
 0.8945881353668714,0.562057069509352,0.26478282977369927,0.8602607103182607,
 0.8594673334009084,0.9016827431062355,0.04287361212946561,
 0.38890833871813135,0.48130021779112875,0.05156389107989756,
 0.41397850341488673,0.40238580889171194,0.4888257458475068,
 0.9044115158129972,0.880566127214174,0.5193337935431895,0.026368152651038423,
 0.09111515110049229,0.5979085024615023,0.9408350973887704,
 0.40950162023867465,0.8232243135798728,0.024778374242266477,
 0.2436341457694824,0.1822412694287927,0.569335148618392,0.9535665659790943,
 0.6373874270588469,0.16815048137595845,0.7887230509602201,
 0.060423718357459744,0.8364682200753525,0.4768548684153755,
 0.45897968057237715,0.49580412076435143,0.7970806390364884,
 0.12364232274499698,0.06985878710577431,0.6339008767036772,
 0.9669074963396722,0.43137688844714783,0.6002559982571879,
 0.45117257815562817,0.7621125944915184,0.9157913599410943,0.731344283685068,
 0.7297079006948463,0.3208517197177858,0.5243129534650524,0.5993277684243085,
 0.4771127930800192,0.4677605635511122,0.033556725365889495,0.139811117019072,
 0.1678952471096965,0.6388638342395592,0.5401918973156087,0.4142739817454574,
 0.40238942182765924,0.3436702022808671,0.3295024410693922,0.6589050349501597,
 0.21804366537190567,0.9801600454864817,0.17756496119327214,
 0.05364171488381564,0.5160349781138063,0.8472022854478116,0.8624920936037664,
 0.9026949799403421,0.985416108497454,0.5115284709308472,0.9794232631271067,
 0.2959724039108357,0.7190084304879117,0.5845187193798921,0.6557430730823134,
 0.9901322698171798,0.17824399159105153,0.36309725546978955,
 0.4040349782847287,0.5441502830624461,0.6753238421372536,0.21166071158464628,
 0.7792620661890913,0.04528980511401459,0.24669167657550395,
 0.9322321446546522,0.2850489511248846,0.2439649606135641,0.8592539104150432,
 0.470416782695506,0.2682499978767159,0.16729725908859827,0.4508312582617159,
 0.34907559286220324,0.43857919324129546,0.9398581752225971,0.817767006062939,
 0.10114350383688275,0.6384148903639704,0.7623608827293225,0.5194587628534199,
 0.9211303428162259,0.5693609699187414,0.726388143701564,0.47302111413374714,
 0.857832219966644,0.6614458238769877,0.6420237526819734,0.30532352932078977,
 0.2600926794850892,0.7625971924448394,0.3845656956129343,0.1071530924128693,
 0.19045659877901056,0.15594601446198664,0.7729252464212159,
 0.19287834187187425,0.49262756867809854,0.8736648973278345,
 0.3197102120638362,0.6105172255322472,0.3165985421028158,0.9237610566524596,
 0.8640342440391726,0.5583475987853808,0.05678063506124886,0.4436168743573595,
 0.7430345448202759,0.4384511514382605,0.22619865956610052,
 0.42741870480717004,0.8712683414373151,0.5706103418576163,0.811373342671941,
 0.8602178649510385,0.18909360678067633,0.337423755341823,0.8361090567924525,
 0.9789865296697307,0.1762972658957912,0.08120806214337706,
 0.060316516287005184,0.7359398964436437,0.8078629444459033,
 0.5222295743777845,0.08930272316187071,0.7007046816029003,
 0.15841048853800754,0.583286862488372,0.27678310569208886,
 0.05134252985592347,0.9945377253905967,0.5245964397652089,
 0.21588126234166727,0.09606748012590915,0.3026483233964393,
 0.26696988000021726,0.2207540975974397,0.24208070105558166,
 0.01976273510341653,0.49430964475836914,0.3960694151983931,
 0.036687780298666484,0.316986066353132,0.25902524917166614,
 0.42612974472971155,0.06762181943851742,0.042010659313254095,
 0.43289122031674243,0.9539574863651938,0.26239334659760627,
 0.7437892964162699,0.19075838778213383,0.5313790656782262,0.5220685658558605,
 0.8313331000254608,0.239173766221353,0.4059581591532123,0.695180245454154,
 0.4955259655196883,0.17271354529081662,0.9308925235828344,
 0.42886160525242223,0.905963589114983,0.5698832648795911,0.14863123516073584,
 0.5818968791109732,0.9122977818571124,0.6286264728208084,0.2851273083391622,
 0.9049963805931297,0.44183019923243516,0.09043049374042811,0.692920941063688,
 0.9451130202207538,0.48637941668268414,0.9631518986228582,0.314337770468854,
 0.6897735703745391,0.9413276207774081,0.9857514499945191,0.4217497472473888,
 0.8608974081475672,0.006876713082058394,0.16896869745786025,
 0.24705219096355047,0.9668952533151751,0.11902159180305227,
 0.4271833254737021,0.2577407667379944,0.16690541180078977,0.9170741757903539,
 0.4012780717053057,0.6094682276194872,0.5401400128225622,0.18698093263566795,
 0.0361455154081195,0.16309630056288327,0.3219027634789998,0.3685825050111199,
 0.3341669247149046,0.8639065347308676,0.10039313030619978,0.5787259210448431,
 0.4207455808948113,0.30674874829928334,0.6076655316240583,
 0.12015788973820551,0.09754017093457368,0.5753127694429274,
 0.7981199363261802,0.5914702504314918,0.05628440287326075,0.5285487989674158,
 0.5306964321640879,0.8891981062357559,0.04862542676189341,
 0.0035548378994029184,0.45512336163251454,0.5648203616480694,
 0.1749497569633598,0.28053756949967723,0.21501658864671724,
 0.8741038130222037,0.3345947657460292,0.507865907421823,0.23205761800690117,
 0.42991385445769725,0.0797965119812698,0.49060870786586486,
 0.3169505406891072,0.7733494468263835,0.9293541677429366,0.12688989790621608,
 0.105390257056167,0.3509195848758988,0.996852118079906,0.3555059134526988,
 0.7101229540286157,0.19411605258348663,0.3845900080158453,
 0.11882122424144925,0.8348583048255881,0.36193184591097016,
 0.04568540892212081,0.06940209714441559,0.5373190196520115,0.974738760653903,
 0.830201443829834,0.7056866586641374,0.6809681380530977,0.38242020343795813,
 0.7548073292521403,0.870818567671017,0.5763229280861334,0.23308422285429375,
 0.11850196807035762,0.1249529397075595,0.5655764962039418,0.7088690860898662,
 0.7152384785503618,0.47584044249931656,0.46527891072288474,
 0.9262819134819373,0.44690241954812604,0.639767271019387,0.41912195333753355,
 0.43144901312912465,0.8179600223928836,0.8668667162631333,
 0.44905878306776803,0.834005303052213,0.5146076771356726,0.6471504646183022,
 0.43464255008415564,0.3358996573755829,0.02111297204071494,
 0.7202638264658751,0.4534658423340284,0.061009259803262994,
 0.6853446863701815,0.09432639811211074,0.19966920401215227,
 0.9219622341897826,0.7083111689258947,0.9739877466552891,0.3410291174619575,
 0.10528469790254313,0.9744989972883096,0.060327935041015524,
 0.17553374175507308,0.515609035798709,0.5197525222244248,0.6103954738983628,
 0.6007489438549396,0.4882856626689387,0.4443716843751727,0.23489127784884833,
 0.5806886077290145,0.22539569179759888,0.6596275444579459,0.9213466675734073,
 0.7670037994895641,0.3472051442237615,0.9091850464846618,0.36237189332647934,
 0.038769037704881826,0.7739368060060042,0.43088410184258086,
 0.1345835614571208,0.10196067912969431,0.3304234519516056,0.8254573538905842,
 0.27840665443972,0.715507769880705,0.600514847815139,0.5538713643117085,
 0.9844099373702946,0.9721881393941737,0.994462793252043,
 0.0011898456083514475,0.7186186098978812,0.493639891025349,
 0.7329887738139312,0.19112762314379117,0.6018345933893907,
 0.23055202377139805,0.2825944005379334,0.24398334193168403,0.921749810445935,
 0.7598965784106989,0.5860699443725046,0.8713183342860508,0.6137661888212095,
 0.058763877384652274,0.06796714315195118,0.7583868494114915,
 0.4774722915002797,0.09678280107004322,0.36980040787104806,
 0.023137163768326863,0.7167178501807724,0.990510858433601,0.1982634021413875,
 0.04604786536360295,0.9181386469321609,0.7242954973611146,0.7724205459807707,
 0.45018673055593317,0.16112398135727557,0.16115144715364682,
 0.18703987761768937,0.31739043460133054,0.667287649880739,0.628891494495181,
 0.38630273775502744,0.1785352657916951,0.2923718438989372,
 0.13655945278008608,0.8320301103322479,0.5896418398460721,
 0.25604402127551307,0.28901566123493805,0.02631872448488326,
 0.32559036156862875,0.1319728256984659,0.7430663443093144,0.8231955527207598,
 0.6762738696856927,0.664003019808378,0.26131675655685105,0.5961213002643542,
 0.3532758224090605,0.11456328247993541,0.9552889348649366,
 0.28422068055253424,0.7029274743373086,0.7207936795652875,0.578286290155361,
 0.9263163776513732,0.2393351360410626,0.7647961302584021,0.708785237907887,
 0.010630039688322102,0.8640443766242015,0.17950690009505765,
 0.5027085565162898,0.7586176310305102,0.2993576591870196,0.8738672551078641,
 0.2558514712701827,0.6697503398727598,0.8695349836492624,0.35406847926619833,
 0.6419824976816446,0.4774037888719769,0.4827036121024473,0.49661919527947473,
 0.7630522412793744,0.2504389361003423,0.6066069422899764,0.8876685825548255,
 0.053637502338163556,0.9652255034285583,0.4096795407861701,
 0.8372276954819269,0.838039185426108,0.4737957758766247,0.8313341855378715,
 0.406514527437181,0.3544993909624218,0.39056974870715444,0.21118140932625296,
 0.9635485881347161,0.6921244161635262,0.7660337040404561,0.6445232082490937,
 0.20877573366629276,0.03967654742863791,0.8210748975512474,
 0.25542013943710085,0.8870468116448709,0.22740356518235916,
 0.7180600352873971,0.07789463013794407,0.6133205398306152,0.5609082214228717,
 0.015693445586705268,0.30958640849049746,0.9413003654527009,
 0.6733921909490707,0.9024862088861075,0.2993946790504415,0.5688761564280324,
 0.3682720575943018,0.5435345900986859,0.826586846817034,0.76105489278436,
 0.44853446469596836,0.3016316922116371,0.5860313269538975,0.4049440697834895,
 0.5631628395307845,0.7805606174953656,0.37274238970418916,
 0.005501403770390656,0.46629725898012964,0.06082934805830764,
 0.7329170178704775,0.975186059244389,0.9156238496890686,0.41536624821656676,
 0.07998983516473201,0.4291960666878589,0.8362777854813189,
 0.33844325155592925,0.2893342837145818,0.8837690266787667,0.7124206705431277,
 0.4055049445064862,0.9603207673407302,0.7988937027850957,0.46218313849576553,
 0.7635340777856898,0.6895695662130373,0.55297554810956,0.30224040819401143,
 0.7196321372161831,0.14750690539055922,0.11749947261975768,
 0.5415536138702193,0.559367419259756,0.9707445626207752,0.3902988704728131,
 0.022355930169151605,0.17293278148581492,0.5673870079415634,
 0.44307301991091397,0.6541943207636292,0.012905990091764519,
 0.7085495982763022,0.04390629334656959,0.3572342824205321,0.7166068214869593,
 0.9152044602972843,0.8691703023679132,0.8701314165210842,0.5275786339321229,
 0.13328603868537814,0.09194704654070684,0.7522207385673542,
 0.09592696485232943,0.20548236608894033,0.613131696996311,
 0.12866470570665345,0.025641888207253105,0.6854899691652925,
 0.6880050562787476,0.5736608053319603,0.7923407081948756,0.9882896556388797,
 0.7774624821921443,0.6740174714776155,0.30299519360543314,
 0.47794415503186727,0.225304010750922,0.909627993067049,0.27647575960102677,
 0.6711352941789459,0.9556834761143475,0.22850689404948699,0.259049954633536,
 0.2660345758344783,0.3056792803661381,0.6780469450742811,0.6423615867766871,
 0.9261433684015364,0.8255327768818368,0.10072464626827216,0.9397354686488584,
 0.8886910872007107,0.06971707690965079,0.5055265613721464,0.4681176059330353,
 0.7286122908888637,0.9949429305182135,0.5562094475764282,0.9744899629631306,
 0.2971139833508154,0.9864654223281006,0.9506002510534577,0.5580706524707194,
 0.41329025154877186,0.7557243601635514,0.23567719877483817,
 0.6393506544450546,0.05126410720709451,0.03609292458273117,
 0.7697807313093309,0.40895848887626685,0.15999141102338643,
 0.35441683805417845,0.45419936616510936,0.16089091335871064,
 0.5883461775167997,0.7756074878205229,0.929665433670409,0.28049159040875893,
 0.024195075577554293,0.36316154993259664,0.47626436109271364,
 0.47753653633962734,0.29827231170474844,0.5477854847663666,
 0.0935749433419224,0.1348584201909866,0.437846077818155,0.48647497186931,
 0.7454050556887366,0.30168894020965853,0.41390353885950115,
 0.12139684215233904,0.38215532099715444,0.7450562183445567,
 0.15810137964714344,0.34539274492555283,0.9821111948054707,
 0.9652754506535497,0.9292614743201818,0.4754698048620325,0.0898431757183622,
 0.6542260395072694,0.2860128963568972,0.7060889018147873,0.17895334233735816,
 0.8905521073819691,0.14349474096171788,0.08537329286290107,
 0.3725137239656292,0.155413432538827,0.3176368003248722,0.3048673150709281,
 0.007198756558558506,0.14782929036284098,0.36643946499307467,
 0.6222876903285696,0.9850693575445427,0.17141171548092538,
 0.06955034719152053,0.5395309370930024,0.9074077607897333,
 0.033361498134118706,0.3415804164619626,0.06761016303816592,
 0.32634971728491025,0.0418182692222856,0.8466003928798984,0.7781286301723281,
 0.557253984126172,0.7526536211997565,0.9555229951596532,0.9585427815440545,
 0.03258089491492411,0.5298599579553542,0.3588632814856818,
 0.05887781310766782,0.779404440868818,0.72288446991412,0.7150291095434375,
 0.8473997668270496,0.004765650133529475,0.9574901543349664,
 0.5224918504059233,0.14431315898418795,0.639155952575674,0.05208094559233456,
 0.13336469747104995,0.9459065801716418,0.10839724549363572,
 0.6997417773925033,0.4934552054760777,0.06378244911994646,0.6371193258113248,
 0.9793122609651416,0.548130320752535,0.7798321766572084,0.8742887117535945,
 0.5456344967196989,0.8126411192549173,0.24713557484232518,0.819708831218629,
 0.38824018322640863,0.13285015538060818,0.9723942892203778,
 0.04724676682044915,0.06670439202623579,0.012474842289408317,
 0.2263536536234798,0.3407521166627854,0.829085171168577,0.3772242250561967,
 0.36251163965171296,0.08646867071296549,0.3877809990944948,
 0.5329596467810519,0.2953211879614619,0.5446332239734065,0.9234961425041112,
 0.06590569797321377,0.9028867720252933,0.8390195857088443,0.9583662390887481,
 0.7324073750927106,0.03471665265103696,0.6784697204824548,
 0.41529596213564746,0.38765358122748295,0.4274329865354203,
 0.23479492609584518,0.38305596058439906,0.8233868185278307,0.453608591479041,
 0.9435578921393722,0.27433218724607944,0.9910186391289548,0.8358674951319495,
 0.19757483203977633,0.5891048842015093,0.9320419183296718,0.6370338900777125,
 0.928343820016711,0.9725351632853303,0.13419358763378442,0.7642554449393932,
 0.3388897786447409,0.3015431266329597,0.7089836910907944,0.7642636395959321,
 0.6328158709108077,0.5690018974190737,0.6364015605124004,0.39955609911840084,
 0.3430637441805404,0.415043078393486,0.1814573812814948,0.5568521050597188,
 0.6834054060549549,0.8028555363050927,0.9074411683143413,0.11770533632672475,
 0.11687032959593724,0.9785150107566342,0.31957074676871633,
 0.8901965195415775,0.21493749550676589,0.035439730187781016,
 0.6826078833129894,0.31259149281357346,0.485769792371605,0.9015527537746607,
 0.2523504459354937,0.015924674102087644,0.012784298910621383,
 0.5916818225336298,0.8774677760141172,0.46628065295414245,0.858931996625277,
 0.7295045790291461,0.9916171185919136,0.8254977232041973,0.7845012813429534,
 0.2155369840744671,0.010150565635387854,0.06886105275336041,
 0.07504321099617917,0.05175674116667328,0.6337500422520312,
 0.5283571897584631,0.6599403219308836,0.88921508200496,0.9480091528681633,
 0.5932478535889277,0.37355267187516117,0.926519895450566,0.14366050234202898,
 0.6766892804807685,0.7925843603272158,0.5913269530135867,0.24475003312638233,
 0.03825646436428243,0.25464652976808577,0.9301832964543033,
 0.8212019274689908,0.7779525411467793,0.2044855729952958,0.88506805180379,
 0.3220118566641422,0.5256751845795472,0.43981269971676373,0.4411455237329327,
 0.5371247459180211,0.388012604369397,0.9989593832985016,0.9919702806293963,
 0.6485873322902951,0.43081866339737074,0.7625148376055536,0.4619700313121483,
 0.9276961610274265,0.3030524047753138,0.5231552712239405,0.5192113476129763,
 0.9616696324867072,0.09798633162254311,0.1830305828530715,0.8301862857304351,
 0.8564235456266207,0.4060408699862458,0.45317174186234643,0.5870451161182322,
 0.8288290644388847,0.18234075420785134,0.5203007783304724,
 0.42198986469899047,0.47146503618294133,0.4085242029972128,
 0.20720692991236622,0.5111708945007091,0.19482149125524773,
 0.45393367257003037,0.20273605528268646,0.34725579012799135,
 0.6368824052739464,0.29496561454326986,0.24516319582650836,
 0.5086627784694409,0.1371401474442987,0.8630947956818562,0.958055269876875,
 0.2641340639853458,0.43199827552306,0.3787086020210342,0.08423595437694387,
 0.9228144281370793,0.4820810525262754,0.9012179433875015,0.5994184013131773,
 0.016021284933367852,0.7930579748729132,0.023251353692400833,
 0.6572499452753757,0.6008742515627613,0.8601339874957197,0.4171561763120194,
 0.4580485918062077,0.35638125081032546,0.4683173862578982,
 0.21459885990590788,0.6247457896214179,0.7628042569135468,0.6202246272120787,
 0.03854251145591936,0.8163319157532963,0.582750265730956,0.04925922166180796,
 0.4611528904415021,0.3100754525850359,0.8935695291742504,0.5178324869224782,
 0.4157149385065506,0.6645448132420984,0.8104242910604399,0.06929839748947997,
 0.9126303248522014,0.22604733132018906,0.2364226143367747,0.6497590021833451,
 0.6925663262423113,0.17452212386403343,0.228488011846106,0.29851856110772157,
 0.541154695806306,0.49412304797033224,0.48217843035987473,0.6084840070977284,
 0.32192592796704467,0.46029763286542946,0.7096122745849519,
 0.7056272700754553,0.6430787994511487,0.5909319467148412,0.30696756169653694,
 0.5585837751078097,0.8541302062061766,0.9018602395954302,0.41297602991784543,
 0.10685250428573367,0.425154715189313,0.31420118902000915,0.6318267439704841,
 0.14876217388435098,0.4184185986565494,0.5425601114341767,0.3688063190274443,
 0.61795492817224,0.5978551685572107,0.7228947403864843,0.5481505148018863,
 0.5648468427326473,0.9026157784659307,0.3838444539489506,0.9508616897183267,
 0.872560790165954,0.9066935211789405,0.019348043320954744,0.6292334436847057,
 0.1286645088880819,0.2639843145662435,0.683080336211033,0.8155015336968098,
 0.4560852763139507,0.7059160753223792,0.5954141389482542,0.5110411712147719,
 0.28947856611910905,0.500130339396228,0.9350250693956665,0.9170133676751785,
 0.2665299543323997,0.08415392339434735,0.18275406862932342,0.94062085950334,
 0.08443106698939218,0.07085226163259228,0.11854756496197572,
 0.20107706373824852,0.6043935309648087,0.5971690195081295,
 0.10874660809532277,0.918338177496856,0.9473718015804637,0.33023498255050654,
 0.40987438901474027,0.4107521817669121,0.5446372706308451,0.6235505812689562,
 0.023231692828540762,0.4979978958749016,0.9710978932275132,
 0.44964374104335136,0.26932302281465026,0.8257492253857501,
 0.8841548941974359,0.9619540765191306,0.5635271699579028,0.5634298031383362,
 0.09468548623838213,0.4734140126768007,0.06037962264755892,
 0.30901797428116384,0.9311816981139552,0.61642797108468,0.9624393590641808,
 0.2345680950876824,0.5239455842718843,0.4642641581989758,0.992162913631199,
 0.3791921106225935,0.029165560996532336,0.3921784683374919,
 0.9212206879262506,0.6121216270985641,0.42903996160921654,
 0.42392173066779393,0.5046811526531206,0.4346956774445069,
 0.052913890497036875,0.2635126832027086,0.2998522317812703,
 0.30647824156557957,0.38630621003597,0.4960912850778517,0.06961068766670031,
 0.9117818270123025,0.11125261806680709,0.9014910269893381,0.6792116276821023,
 0.789725723360601,0.5362669712319676,0.2470771318081576,0.9532206484980827,
 0.4016923776549024,0.6156490495511526,0.40610070597310766,0.6935705824064913,
 0.9922004763658333,0.14414616140685288,0.6997916387090328,
 0.45021647033125856,0.7833984955760731,0.5769192343639243,0.8967264677397486,
 0.8659397407028858,0.11377876715325619,0.3577858146028311,
 0.04243638658498439,0.9708853311208799,0.6655755154667331,
 0.05755321086406162,0.0937293809145211,0.4284183523766677,0.5147168746569122,
 0.2265023143071272,0.6523983447202955,0.19133808262955132,0.4968302089865634,
 0.6986573176647293,0.6758940235080672,0.5512298561880922,0.8193903916884167,
 0.5078073510959067,0.6613266983539628,0.32110639667281593,0.216303234484562,
 0.8413985453747359,0.11182399737923854,0.9005248205237254,
 0.30106778559475134,0.6089708833254647,0.7981396915183762,0.5959269746745441,
 0.6031798151348587,0.7248494918329986,0.3014871508199155,0.49431643928968827,
 0.9832369721192239,0.3723464839610572,0.2801952038647917,0.44183450316790995,
 0.16278188472275534,0.14881324047976618,0.5468747040544818,
 0.5096253227552283,0.23053922887241007,0.1916145576168271,0.3619005740045197,
 0.26492894054058247,0.21721892639710394,0.03970329612505219,
 0.8818108912126218,0.6512429255700577,0.6386194652784369,0.21814727736770367,
 0.10751549547075134,0.09495443878510712,0.667782906520997,
 0.12774336335581804,0.49822797459399615,0.06782454434025598,
 0.49950683329010126,0.23651420379310317,0.7248531826567254,0.788182392945709,
 0.5592008252778046,0.8260987142656108,0.9412583050716119,0.9021098884473084,
 0.4325603453889749,0.9201670054424786,0.9993816346581699,0.8786615203915515,
 0.816861770669507,0.6115372759792724,0.3237802580173459,0.25598366559492414,
 0.34097892820379494,0.05001141891643801,0.08574184364765969,0.73262354492341,
 0.1749302140155592,0.5366876792304226,0.41105340250755384,0.769042418193318,
 0.2287551973434423,0.5250117231584597,0.3549506525314423,0.5368842009304902,
 0.9978096440231683,0.9353660081507281,0.4528431350461579,0.5212731617335487,
 0.8329345872335809,0.6409686160364768,0.4181737829442427,0.5557133169413715,
 0.0375357819507911,0.3482839682634863,0.8271267018558495,0.8158669602149982,
 0.5601437708940711,0.0854931601655966,0.43833612430600466,0.4144892592228404,
 0.7923775792994698,0.7695744186325805,0.443338779908777,0.10442124640362715,
 0.9407871809024128,0.5592986287619515,0.4256910527217693,0.853435947985874,
 0.0389652499356794,0.3779937182518496,0.9290452312916476,0.8719099366705827,
 0.336174366406345,0.30402435403578587,0.9971363426575306,0.283427846604253,
 0.5971308929579132,0.5417738634966318,0.6318602577975329,0.211195530890244,
 0.37771118566243156,0.531034137149804,0.5799338827598453,0.20907759927641367,
 0.4411367269149571,0.5216523489986284,0.557075344384083,0.13436850608885642,
 0.30697024614455504,0.972686987230891,0.7588980392222928,0.13574308278196667,
 0.006244935441668709,0.7832874057797286,0.4410344120216454,
 0.03344063915883966,0.37279129618607587,0.26570836288879973,
 0.30297961571151666,0.2654970430845356,0.1037911424167185,
 0.16968201411497752,0.756524813365599,0.7207961117851879,0.7420717534135681,
 0.29993106164424077,0.2400028377464123,0.987322602020343,0.7490549366520342,
 0.9462993981670278,0.21687360241848253,0.6498914144740338,0.6116831042549696,
 0.12142348927201296,0.6427581728735288,0.25717656588688764,
 0.9118138578581135,0.30404881597874456,0.38251824830480174,
 8.867624927143547e-4,0.2914386436091405,0.6171337818850051,
 0.45242335287599833,0.37232449996527295,0.5032220615506158,
 0.36219702079403604,0.2935918901200274,0.32617544041475266,
 0.24432162422087567,0.9070280904461122,0.05610731473255626,
 0.9689741547910784,0.46423322254278077,0.780328534497044,0.8245594832615484,
 0.2393462185842793,0.2850345345551424,0.5775843532345584,0.09846582034249285,
 0.7599023581561157,0.44100462112733607,0.9206696791548139,0.7283511678802016,
 0.7000706784656291,0.743290169050812,0.42163189633445497,0.23970109557651043,
 0.892078117117885,0.928406016005809,0.0015468947339767336,0.2923144505918651,
 0.1164888145376044,0.6427878109731089,0.9264771082380621,0.5731886412814615,
 0.4217388582044156,0.1354555238921351,0.44043147043314024,0.6678466787243049,
 0.377817229225627,0.16924971111869236,0.6686939376809233,0.04132874277966603,
 0.6019141525425252,0.23293401237133615,0.5302290723175096,
 0.22217964931788292,0.7577364110198532,0.8730191670562576,
 0.14075524612693713,0.2729275310246456,0.49926369349818556,
 0.10142638995110287,0.6669341373640345,0.4362769324082576,0.8059445126323028,
 0.2942930102752327,0.3222723206536684,0.5117106777362849,0.09121626655692805,
 0.45374224140909136,0.9168588229716155,0.05287776904917263,
 0.8368020225998033,0.62311374135935,0.20647168200065535,0.06898267219889931,
 0.8216903912288153,0.3842586407127435,0.9747564905474487,0.976412002457038,
 0.19312334188702307,0.0027302751749584964,0.04002545430233917,
 0.3107370146706425,0.7465232562341818,0.6005095239490937,0.8026758992265759,
 0.22200187774311675,0.927974648067721,0.06805746097972243,
 0.04843844167136924,0.3406690529628289,0.7941351267222434,0.2226313434615107,
 0.9766327973190404,0.29525299804816896,0.22400655672543335,
 0.9406332008669408,0.6272225249231878,0.9396216911824625,
 0.019049151074716497,0.8734488640502908,0.695504817970798,0.5320134790551828,
 0.4245808635271471,0.13179188340110393,0.7891882644659269,0.8862101645310165,
 0.7953538189974955,0.8115369564921295,0.009728730098897187,
 0.1784581574066385,0.0969156615719603,0.09547978778996491,
 0.16402059393996382,0.13930970485277872,0.5914430363094713,
 0.12651326048956568,0.8704390090160045,0.35195725799081135,
 0.8356091469293627,0.39201241397628894,0.8687339331367737,0.5139398291042268,
 0.8639720764818435,0.8230954614316461,0.15908261967903792,0.6927755070894178,
 0.9941249272657531,0.9074947811302218,0.6236096419913142,0.9063400146060844,
 0.08558710638661671,0.25285040308505247,0.06515332313809463,
 0.4922834992558288,0.9622380293976454,0.39270525862249017,0.7019681365640685,
 0.05764542261790917,0.4539715236427334,0.8015327108134073,
 0.32982653519259864,0.03197999628195536,0.8666003298667451,
 0.9935741049539232,0.9602670311492071,0.48797004936106725,0.4493462221184811,
 0.6389314462481902,0.07915276683473893,0.977354475075819,0.5344142943706597,
 0.8324829189602612,0.23704086539467129,0.9634868860735827,0.6149956022235551,
 0.241563632209029,0.7026251250633013,0.4507414312983321,0.1129058791278954,
 0.5623909839881553,0.2916504937251293,0.8534007004613722,
 0.014660838249934027,0.13055577117883144,0.217313163059895,
 0.6196637363697479,0.831653979347351,0.4625965172374026,0.815453952300385,
 0.6227000181697837,0.453025772555526,0.17572292435967674,0.6911195976470047,
 0.2971337871674463,0.60267888267089,0.49791562028844116,0.6492847926585331,
 0.08346672203755334,0.011573205961957989,0.15141837559990168,
 0.2147086586419078,0.25286922021416525,0.9204695063972459,0.837679352141258,
 0.052652048806653196,0.9890973415363384,0.1121828151965295,0.872405356719985,
 0.22076149284721525,0.42078513014570595,0.7270817760443448,
 0.2850128598070235,0.6360160937143399,0.5578630540576196,0.7747247644893333,
 0.2993556328865019,0.34966444904304084,0.7259370616346914,0.414056402408263,
 0.34294719544078844,0.6856272459536845,0.7180637930810155,0.2102388018656226,
 0.7376662502455504,0.9613450623196464,0.11984820511671124,0.8407902375795228,
 0.2405477909084175,0.5861583708696159,0.47811726372131846,
 0.07922893597617742,0.9070733085258676,0.37726256454990736,
 0.7551971659342124,0.5782635552111802,0.16800259613600876,0.3990528511351227,
 0.6396490357566351,0.31269066546118485,0.7568898969547034,0.7612291227367167,
 0.5781266330787491,0.22757957000486884,0.3309359051762004,0.8442964776792902,
 0.9069619635504945,0.4402680384176495,0.788565847969663,0.8893875648061528,
 0.645964844817194,0.8924698771287837,0.5811932010425562,0.15809656294255559,
 0.5089070941530434,0.21887318109204712,0.34922617946877077,
 0.2052533859740766,0.43601810046848766,0.3188346040164636,
 0.17383327484993838,0.9922899580821393,0.18682452229039148,
 0.0348752210752733,0.10888354476922657,0.36274431568057863,
 0.8930663704891493,0.8450397434965419,0.7151133450856558,0.9345812437400265,
 0.9052892086876969,0.264772419271455,0.46361023601559326,0.6429730242210749,
 0.06372495595346594,0.1945059050366299,0.08506645026504689,0.935105755915039,
 0.9031552652398954,0.9080395118898705,0.4495303439750036,0.4808273988121765,
 0.2176964753728976,0.035071957987827695,0.6686097177205825,
 0.06198146200196475,0.7291453359703837,0.33144516835562066,
 0.9785688409753249,0.9272497357775641,0.16480317429184654,0.6913216393508769,
 0.8735999816286912,0.5224521511665847,0.9385305225264965,0.4997006379816935,
 0.9010406550279367,0.4462202147143741,0.0491747860727918,0.8083452975302907,
 0.16870914230400025,0.43036278826465435,0.5539869854536289,
 0.3869251797545956,0.7894588266653442,0.48801172075771093,
 0.42506947564570485,0.2379484308780283,0.743184741112813,0.6336537507698852,
 0.14697291465285012,0.48529888410558875,0.6394946849817094,
 0.6934224007512511,0.04494552413926023,0.45987760047296056,
 0.4056562175479509,0.8195770553794688,0.8417180805993433,0.992217223393217,
 0.36844088893244065,0.8730276874250501,0.06747860306300979,
 0.6802955353694334,0.3056999698899059,0.5688974310721384,0.4715998693427008,
 0.027523296460697955,0.4236458584239582,0.5563561717069466,
 0.4248759641472729,0.5381892962427661,0.4094226068189055,0.9905118660240793,
 0.04487805215660823,0.29967275243543445,0.33216858865076737,
 0.6514499179886752,0.02498944078963017,0.7014563054953475,
 0.09891664528694477,0.9461819817522005,0.38272035342128663,
 0.2838225113640508,0.9189436411798686,0.9806652799345074,0.7516660287089576,
 0.434426646486507,0.4198911703676512,0.09934251879851086,0.6108456419281706,
 0.07229507116366662,0.4160535115349069,0.707314357055312,0.37189983147515415,
 0.9533063696960502,0.004495648513136996,0.8646033548898451,
 0.17417988546747165,0.41776532796002286,0.5167718346939186,
 0.45094114063968005,0.9648967072724652,0.06240136429114729,
 0.2958178374094519,0.6389105679948327,0.40149556391579966,0.8399403099591697,
 0.7602906694795927,0.7949048788120914,0.9080675846835675,0.15341450445755878,
 0.46912680744514246,0.3135037585806941,0.4123027314923913,0.6957679908737016,
 0.08347117009974236,0.7583099163315058,0.1546625170941629,0.9443519544804192,
 0.6666395507402212,0.5645251213156527,0.09087729535855038,0.935343844137396,
 0.6702355463192256,0.9333420055130478,0.27082321805403187,
 0.34714477546544975,0.20169042602588627,0.44358542465787,0.45788009951432207,
 0.17721888113031503,0.013063026609935635,0.10912308820269268,
 0.8749311495770031,0.31018943816340316,0.7535524573191507,0.2559320080154809,
 0.8980493330169274,0.7444863597134328,0.4768013190056094,0.16723240634652636,
 0.2211658378846506,0.8217776406934479,0.4785831631759434,0.11625910612995893,
 0.05041060779459072,0.4285656677019166,0.489778098800268,0.5699160560197682,
 0.6825404258143154,0.8590216918011726,0.879044570160029,0.44226454961375183,
 0.11538713257647604,0.9179250245216747,0.2133023614859909,
 0.007941028666508254,0.12011390078858186,0.2046484791297144,
 0.8075854393901136,0.7919529018221381,0.618416637548865,0.12407487765433056,
 0.33622681019110245,0.9027227305139739,0.5804620786456078,0.2399026245925856,
 0.20427699612387018,0.6671227132042848,0.27881671181019163,
 0.5415213480578867,0.15492623129158922,0.24314180693255627,
 0.6796811692777416,0.44074049009530825,0.43739825680147504,
 0.6561718696782147,0.3796637831579118,0.46177010933746887,0.2972817416981848,
 0.011329015703633094,0.8659205363670388,0.2873946610560051,
 0.39521468337466037,0.11951979186900508,0.04910885178046831,
 0.21670023405402516,0.09383532785550253,0.06264300688116642,
 0.22289070035950198,0.8741354899221057,0.971741980564924,0.9920348532130104,
 0.8967781678320927,0.6872488474953405,0.5467310112247583,0.46171337656881317,
 0.5308327423882649,0.00970844817476757,0.6660478764597754,
 0.15043384316096675,0.43415432704348556,0.3989243612132962,
 0.012865583755750976,0.7238485364563298,0.9779117399750311,
 0.2572811335810097,0.9851159688901446,0.6098111113996342,0.6775403750566047,
 0.2487202520730718,0.922886223954181,0.5880166482446483,0.8511527312637432,
 0.6065143455913572,0.47763705186180494,0.5788388765663094,0.5927854693217024,
 0.7184887583098913,0.43020748907131456,0.6180537820535624,
 0.18736716769864448,0.002217179604416808,0.07748725275915624,
 0.20653381758469125,0.6974088970670692,0.6708282286309127,
 0.13666720776761632,0.3263831336387796,0.5805535712056336,
 0.29510807176700227,0.9718408133801689,0.37967925239991274,
 0.8022727021005297,0.19282773507049922,0.7396317405518434,0.1930319903457739,
 0.5750798177717203,0.6382345358363998,0.037771521007666786,
 0.45786112647067867,0.8024984801125006,0.7881468360366016,0.9247812173223421,
 0.8185938284385741,0.33133072098553895,0.3027569256165136,0.9960547501290452,
 0.14487500326596603,0.9191735686265395,0.9007644971550244,0.7564939910349063,
 0.10583398129597654,0.3334574881960395,0.2076737761626114,0.9270086273353774,
 0.6325040421192963,0.056771828827810844,0.09465972664590327,
 0.08764597134679908,0.6070999404779728,0.3613909007247691,0.8529936512966807,
 0.6680862125695847,0.9471897248091707,0.12571060128465705,0.9191994304264792,
 0.10885354609763487,0.003080281292732967,0.08844168316026779,
 0.7447500059959129,0.509331251824393,0.22334969484019906,0.7269305785249951,
 0.8355288923802906,0.6085525787242787,0.5400516720962684,0.7163975454366933,
 0.970348013253909,0.01951423466272706,0.5163816087425243,0.901673521248046,
 0.633276010092462,0.23457983262370874,0.5160948888405519,0.6447205510210563,
 0.16246157248363946,0.4859107158484499,0.5484699909555525,0.6628521515652612,
 0.9001209641959531,0.24261918892993406,0.7276149313703066,0.9406178763510891,
 0.2816966188665082,0.7147195164795095,0.17284825667865256,0.8160454289715433,
 0.5069856736930846,0.7154637303725406,0.07339276677371787,
 0.20636895993238713,0.6104000827907067,0.9841675272634471,0.0806133805208682,
 0.49080627727059856,0.10787667652466826,0.5730832186217693,
 0.11363349309586024,0.4293903159927067,0.15314791642393633,
 0.10925102054154989,0.33403793082140365,0.5447676369889214,
 0.5586198371107338,0.49238068930486845,0.5578070558518333,0.9074126129530178,
 0.08332439564841954,0.7370838581631702,0.47110422193392254,
 0.24705518672942106,0.228155373273343,0.5307697824469105,0.2925696061003924,
 0.7119454743471881,0.5334507143593203,0.7489033270848191,0.11028259621840109,
 0.14266225845009828,0.15943100320054082,0.5073900034668023,
 0.36782135654060144,0.2597986361354937,0.29002288390191,0.3220252926935574,
 0.6796626556057176,0.1573686090295755,0.8807445961663009,
 8.071712235911965e-4,0.08272783469618072,0.38221993548678146,
 0.8071679845686004,0.5883249227855307,0.041904066600191414,
 0.7885237777476397,0.6617391987468205,0.616834778136524,0.32989048571053126,
 0.0435299532428749,0.20164327785185288,0.6064611059716256,0.7107520058767236,
 0.2715996459520754,0.907745316573738,0.6687126937108665,0.1659520228683612,
 0.376828631363897,0.3648259412445425,0.6098084461988689,0.860401856309976,
 0.5866866966977243,0.5687036809974816,0.8970658556993546,0.42295031072654254,
 0.36711492351592656,0.8591989798936801,0.9963640036950848,0.3201376303355823,
 0.46135201532318293,0.4052220208913361,0.7597912213576467,0.7601031250061243,
 0.2901077358829154,0.5416574224991687,0.5997394394842092,0.2749347901130702,
 0.8287445198958934,0.7106261855679931,0.061408384631650326,
 0.3364893984226355,0.5837282876693919,0.8326428381675226,0.26898564160818683,
 0.5745233347230333,0.48568793640444907,0.8342847727218516,0.5090854544933588,
 0.43459908707861494,0.005524586713373303,0.718488933515657,
 0.9795586143223437,0.9278236664392818,0.2839214112051327,0.5550823600741998,
 0.09774811799088723,0.6165684153164391,0.5946954721880094,0.946229116505508,
 0.8020669715825072,0.2376164418548603,0.1181476853277047,0.32839750650616906,
 0.7018308171098269,0.21472133923094483,0.6401569131841504,0.6755532566186636,
 0.1860445561818822,0.8779325413767998,0.7402841120235075,0.12452481303588503,
 0.6860632794074012,0.0711085589312902,0.5539291855392321,0.5484472784028146,
 0.444880828490958,0.046665268112485725,0.10253548157963577,
 0.6002664715945407,0.9595482198440328,0.8462584735363231,0.2660641536723124,
 0.9634133697762819,0.7146197574123622,0.8343911482482715,0.9822646029390105,
 0.8619457032630552,0.10898447093635577,0.87767148847695,0.11806239711517696,
 0.645356870288138,0.09384783845090139,0.7061223003901721,0.3545908192457027,
 0.5594154460902269,0.4014665815924001,0.03462354070362372,0.0674324855341848,
 0.6714677391263064,0.9105972021537775,0.5016796342393134,0.3511236245146987,
 0.8611711178068212,0.5616184337122234,0.6347229814299263,
 5.675997341429984e-4,0.9856902302353674,0.5651496637580867,
 0.3253736693733894,0.39560220014873404,0.9211182800762228,
 0.36430211521579525,0.6971260310485418,0.2804910961446988,0.381653100677384,
 0.7773193412522904,0.04165374902461405,0.37357418374244256,
 0.33855072763566285,0.5971389817251735,0.3983729593585966,0.8926055667340155,
 0.4559111692120932,0.6401344694218415,0.5655232561853025,0.8900690521772158,
 0.556077742777248,0.12585403597859812,0.412901681186008,0.48578096238283586,
 0.8393609537954968,0.893299740531496,0.7499022439590971,0.7656049251676192,
 0.3679384305158805,0.7375943157562419,0.22645601789060765,0.7317334617919589,
 0.4893768751999662,0.9359792430274401,0.524598853086045,0.5682543854567161,
 0.18451047897592665,0.6040335326855001,0.1565280123786511,0.6878391729901022,
 0.36596690264687437,0.7983789370163031,0.8802042680799653,0.680337955457189,
 0.5151669859569963,0.8038630639206011,0.1413932193916373,0.46514693237592464,
 0.8208385577361039,0.377352216415868,0.9142277419906722,0.5241636876226765,
 0.05265879101013815,0.8359044906983534,0.2876348563538811,
 0.019891669781130994,0.13122240382713424,0.6073830515154497,
 0.44550344301922995,0.7031123122589278,0.3694280924889459,0.995157072345739,
 0.8331836905419309,0.05470945605430555,0.6182016806045789,
 0.47487191422958686,0.9170888183893724,0.32113269733043937,
 0.9454703314987427,0.0610393612657254,0.2146527973500515,0.10891894493272813,
 0.6590978647784512,0.2052219027121096,0.3965439322527291,0.8184054438967594,
 0.5477578012342902,0.6551275761163806,0.3309589250560212,0.8460288770515975,
 0.4195321869519565,0.6989951424056174,0.8776178946879183,0.6899896729881817,
 0.8677122242825066,0.8824588210153647,0.22257191272648036,0.7158319413519406,
 0.7674621976718058,0.43095323414178566,0.4291952233164129,0.9693359076411578,
 0.7194663290269528,0.3695324340183984,0.04623451413798918,
 0.0075411511030416545,0.7710010747492264,0.06837645315115626,
 0.6916844866021283,0.44466807954758925,0.6115042578001657,0.7715249739930916,
 0.36435666547262313,0.4231233908148002,0.5265768835950112,0.9670377629954359,
 0.16627109895160896,0.530895337825632,0.0704849574845311,0.7605785420022465,
 0.4185930882658433,0.3903277953045591,0.4371386413046804,
 0.012305351324045644,0.4963053461419925,0.710857149955439,0.5009850890900832,
 0.6026463303135192,0.5364057865037751,0.7540350617248729,0.589541940225744,
 0.3527079244964253,0.8286564493407719,0.8458318798345358,0.37825651217607836,
 0.7035309526273693,0.6702876368600035,0.014193556447684164,
 0.025023103878984543,0.8562336595279754,0.688395476903207,0.7899846425198753,
 0.590299122044286,0.6941309140315272,0.8608441365151034,0.7185704601898759,
 0.24044753879539538,0.3512160037550648,0.1273865074143007,
 0.29661836283992216,0.9572568959864286,0.6683704549679444,0.5784224179386371,
 0.3928277008384341,0.03141573818647447,0.09121293301575117,
 0.03819930952295614,0.19769357084997197,0.22499563312891535,
 0.4882430678861326,0.0437990133388626,0.05763822855382161,0.3132385199169041,
 0.4666432559796111,0.017766399297957314,0.38872134359125854,
 0.6151551132848011,0.04935510069514748,0.08066047379629271,
 0.07458323369136943,0.08432785490899508,0.0034163672292358838,
 0.7256087433785949,0.43234060738597635,0.3965960531430587,
 0.08056345386622343,0.5099824679370566,0.9173248429803499,0.5538342086501589,
 0.9293896129512635,0.38522342284409716,0.44662351354082863,
 0.3067881039078011,0.6389407620906044,0.7981247426655865,0.49410202339555453,
 0.7915085175377682,0.4896382832458409,0.75322501960613,0.9432797151549146,
 0.4728733966799791,0.7032861187422682,0.02234101870276861,0.8057019736882757,
 0.7672014258779355,0.9810634558955024,0.7106633528640683,0.697314052752467,
 0.4700407941984528,0.6655628468212729,0.16146661518928906,0.3440323148636093,
 0.7525216555378603,0.9166634455241638,0.7534529284238325,0.17667118892328848,
 0.24044031691225043,0.787765252402067,0.7982737633492072,0.2630191131086841,
 0.41765786503594016,0.7473124564833322,0.9854498018664273,
 0.11016139719904483,0.4874994239482193,0.8693812938733205,0.5734665620783637,
 0.2719672114408972,0.7234861649754597,0.5340742715473753,0.13436295372042495,
 0.7383709677526625,0.7057876021635738,0.7441158243202397,0.24083379868526522,
 0.1256192135711861,0.25933878562090307,0.022261513079354978,
 0.4553357618382361,0.3846993155335209,0.4712518458248136,0.5226856568029121,
 0.018549382595294706,0.7534135590356664,0.6391577893317443,
 0.5716382028895121,0.39816223023145514,0.17063554077942888,
 0.3812114706637516,0.6617357904039174,0.8694869786059558,0.6833246531789795,
 0.35730002496327995,0.06198077264492774,0.736903384062435,0.8214293411247026,
 0.14989008711183116,0.9234789967226615,0.9150266615443804,0.4675295047568151,
 0.21387364733855851,0.591930386243732,0.7074879717317177,0.2216798187728629,
 0.380278412823155,0.8700819957218026,0.9526247651044245,0.6359215986365513,
 0.9707378402880935,0.3006757161452589,0.24801738796586448,0.4812862039898119,
 0.1035276940089851,4.82513825159403e-4,0.29762925345700186,
 0.29368800018483876,0.958365344631888,0.3357745844509683,0.3450810311303173,
 0.17909325044689894,0.5059763751333145,0.293730798243226,0.35363715230336934,
 0.2900141584071497,0.7248643602868442,0.49397145755997485,0.8913928050808712,
 0.9538651183529548,0.2966281105341976,0.16350181781204554,0.3203363885445826,
 0.18654918775634344,0.017035449705211314,0.1556179655083857,
 0.094283267849953,0.7964141251831659,0.38967474059819085,0.07109202322127306,
 0.20894381780680882,0.12869624297215343,0.22033958264790676,
 0.004216225492440162,0.3670323964361233,0.5198313463042797,
 0.4668544786562975,0.520311014928543,0.36567339701327284,0.4464054851097502,
 0.0608253920688866,0.3091569231528444,0.6570341114633657,0.9463893844004332,
 0.9564569253360399,0.49390973325278154,0.2446011183257103,0.5302112404823757,
 0.5878244596364051,0.4182698512544416,0.17125010800381935,0.7064560274553241,
 0.6880071325626184,0.5718595511188234,0.9550202028394376,0.2001589318106901,
 0.40898475299981285,0.6521679392844736,0.7951793463785632,0.8934196125844467,
 0.25147295772732325,0.34576648539925514,0.8521265586620055,
 0.6269547233996883,0.15934102482121826,0.3896603700448168,0.5665003749504212,
 0.30613729162778525,0.7484720386925807,0.0625465146419647,
 0.15395841060836768,0.14701441401405868,0.990421841956343,0.5485610228226916,
 0.3018976916268451,0.34036785136751035,0.3509047524956286,0.1383874802091688,
 0.9499094783569312,0.015440853622330386,0.5798051679928417,
 0.6628826069734907,0.5275955444633521,0.9723586059594475,0.2341938339640457,
 0.6793389994690799,0.6258393700291032,0.35358689801563425,
 0.36212627439831957,0.15289474872951425,0.3934604066534103,
 0.4440061313231187,0.36714723085730183,0.4223433370126881,0.9800272789549451,
 0.9837536681053404,0.7905738139559466,0.7387314450325939,0.4340548693104588,
 0.35573510453496937,0.8835301129432716,0.6879880162012704,0.9676201416009013,
 0.9532912274581072,0.11088315670724813,0.8669563220995663,
 0.10294226889273561,0.5940865772835817,0.24795123535513586,
 0.12753053059088182,0.3341490098391783,0.762702102380049,0.7898985071932703,
 0.5543640470322001,0.7159058292838845,0.7918231443912229,
 0.030520580162202937,0.29072641395026877,0.3225771664395173,
 0.2415626503830982,0.6114204466114961,0.1740660773587872,0.26402994344869024,
 0.6455406696793238,0.5056508328099932,0.33520284570197056,
 0.19476992017888295,0.743572815146264,0.6754459064658588,0.0920155467725644,
 0.3661018793616013,0.8816928361843384,0.013806476244380694,
 0.7081280140565268,0.1086408009341584,0.9470720214096405,0.5856053569637872,
 0.23127692837213387,0.038185104112495316,0.17798853850940943,
 0.5881755483287375,0.2853740593735865,0.583845881862138,0.33886832113528575,
 0.28167823185963714,0.4320806484745826,0.20594961185194693,
 0.3952957707499871,0.770950186057147,0.456308678855023,0.21986888956197226,
 0.6204432652945072,0.3267549986545114,0.235496967090985,0.18052920302491804,
 0.05089271545400176,0.5076982842383728,0.10401978205387008,
 0.6555873689808676,0.7185571268521543,0.32086769306345464,0.2428878668925193,
 0.39967286289413595,0.42810759999665327,0.718268937363494,0.5849700297803697,
 0.47056167299569485,0.8895000964186757,0.07546514158176754,
 0.7297630390750484,0.11400813820691691,0.4095139224582651,0.9136058814575426,
 0.10106575173974663,0.8774971462420293,0.5314200301230115,0.8711974228883416,
 0.3385424765381081,0.57887234565743,0.3215647226827473,0.08540265401253677,
 0.9400624609648984,0.9177232301679501,0.8143055261857853,0.3194586566306009,
 0.9604730580745636,0.5993072113904823,0.5901389517768602,0.668534538704703,
 0.291751866168108,0.5879946482587223,0.35142849301923995,0.7566820534040308,
 0.38142137403841336,0.44022672177773714,0.551738382665976,
 0.20454689441774665,0.9950684108315248,0.21591706867298022,
 0.8239561742938476,0.6731545592107196,0.4129451536139517,0.47503854788473887,
 0.7262502692033542,0.2456308547065349,0.09294658533687705,0.3757028227752044,
 0.9996626558654603,0.029782870571911424,0.12571803696186057,
 0.6763996398570725,0.42194905488848256,0.6089618702800588,
 0.11764787530753018,0.2364331269536225,0.6168815449653899,0.6317283165759929,
 0.638493523846683,0.3903364922453072,0.40455942406248147,0.14654515633617837,
 0.43919574093963476,0.8124326630006105,0.40222153213990497,
 0.8512055187736832,0.378577171187484,0.25519342266632206,0.9009697934843636,
 0.12365489510131955,0.06121367538153699,0.26994630764523375,
 0.7108132854481762,0.4128145302161036,0.9440246571765506,0.4679129199979516,
 0.3946270771635979,0.21094095179425543,0.595776163224058,0.3234467823582756,
 0.8096385120557351,0.9781215164348496,0.014396452700367934,
 0.09113776214254743,0.23453019677391207,0.6445352550882799,
 0.1698527397299715,0.8349932040336202,0.5125620281831721,0.6566656217490799,
 0.1767983517194497,0.13031786774273046,0.12959746870520938,
 0.4675072374788779,0.8477615102159608,0.7745519450941795,0.6013178636881484,
 0.26239519180773907,0.6285748826633601,0.46945460456264754,
 0.0026326588237607362,0.10085987592621515,0.8370674533524122,
 0.3163242877249337,0.8291205563518871,0.06772987335589797,
 0.025293036953798587,0.9262709053263978,0.5897333002239602,
 0.46592296564923463,0.1372516800050576,0.6055366400372891,
 0.44756844478604385,0.41685311335133357,0.6642063879077355,
 0.22954906563995803,0.7081616265424323,0.7944611032602235,0.7059025775660667,
 0.5938256877466398,0.33166897021206476,0.14952467149203363,
 0.3161251253055809,0.5764533674147333,0.9212444871331549,0.18932640182029947,
 0.25237492444324583,0.876703370148357,0.6967746221615903,0.3914854718692591,
 0.4163730782105669,0.29499848926497874,0.9952397778739471,0.7640000056130866,
 0.5248804456135744,0.5794922511619175,0.790364691626709,0.7673015922461144,
 0.38003311918040317,0.4374022628037082,0.8600145766501331,0.6920598579626235,
 0.8234647101132283,0.19999049019384563,0.25403943128694295,
 0.12488298426251254,0.21341109596741348,0.9433083421427702,
 0.9254520515552803,0.5436608359400119,0.5130624627168237,0.9595514820789827,
 0.3089691884202366,0.7197698410988547,0.3341651189284407,0.03896383335892639,
 0.43454931139034636,0.6071407585320232,0.5112592701400247,
 0.24582584201534596,0.7506109333179228,0.2890871097378186,0.2044930094781522,
 0.9976848754025142,0.23541589890825,0.4222352525480997,0.522561481326868,
 0.8731200178898548,0.43573770285919955,0.6095930942420427,
 0.12287811758834954,0.6395741057370851,0.9924177512778529,
 0.15048143333662078,0.1292459495675069,0.28426862363036287,
 0.4627816695710286,0.16283617458165311,0.21919685848073833,
 0.17724990143754416,0.9614452669790006,0.30905859060079954,
 0.28985213084912154,0.9757275774518454,0.47509209487918014,0.554407796350598,
 0.4490689971437848,0.9882970410221311,0.1047286844791695,0.7517961545920602,
 0.8259377218181266,0.4791391835368328,0.5552387409011457,0.06559412475239224,
 0.3438308428476824,0.6267501616719445,0.29553922593052817,0.7132823677430203,
 0.5211759114913566,0.22406288760576665,0.12220855598979308,
 0.9875121865257352,0.923498551803067,0.3667454274584072,0.8424020172299757,
 0.8263293594477916,0.64663368695011,0.26143164478299885,0.5639092680207544,
 0.7235354271305983,0.1550680962052262,0.4540473470317723,0.06265750448262608,
 0.3697249517592276,0.40312937491872725,0.1963433345243596,0.969522867554822,
 0.17744375359044906,0.7853404422893231,0.25174515909494843,
 0.25917716643085353,0.7926343982667482,0.48382833455577745,0.618466452919916,
 0.9744290958047453,0.20770151789736901,0.8777798724830905,
 0.44521958270359396,0.9269979259222223,0.12251686077727664,
 0.5979372014941652,0.24440420400141616,0.5942290680960048,0.4615523012136028,
 0.35604772377389315,0.08414889169426054,0.038488663249588106,
 0.843781516887839,0.16829203041987095,0.943708121550943,0.2122111989719313,
 0.03577992424410614,0.3520634692604139,0.21609214587449166,
 0.8164771855643607,0.3694161965399305,0.629340214941471,0.5716370839036232,
 0.1473435858602914,0.35308043413268286,0.9037730975318939,0.3463735667245239,
 0.29800514343188833,0.4850958924961146,0.611108171407115,
 0.019718145678123244,0.610137779270788,0.5660570249560463,0.5621762588761802,
 0.04412263572955222,0.6655173514514587,0.5812287948287974,
 0.29571236678425705,0.7303343749815852,0.6641667564612057,0.2505537403239513,
 0.28752319216060584,0.6177842119456017,0.9339578833949922,0.6443060840206392,
 0.3088974429838609,0.6917789087842016,0.026155356472512103,
 0.4838019205451284,0.9406343177716655,0.2149402257346259,0.930441516777089,
 0.9372579191982536,0.3110672167590982,0.43406887653705895,0.7023473586662093,
 0.5203561297022425,0.2746490834340274,0.5611605348574178,0.8344674852166385,
 0.7117124980612837,0.26476824220384554,0.1601478204240956,0.8634270873203036,
 0.725533147764094,0.36295090879981406,0.8335180942685835,
 0.0059127443127776624,0.9006738803775778,0.31614110853398736,
 0.6921869053788539,0.6406136340769149,0.7941592410208412,0.4815470138089549,
 0.2743570191260516,0.4782177491164591,0.9818689924068965,0.11528345058865774,
 0.2860169275274268,0.9692520501399213,0.12219156052451807,0.3084827150581665,
 0.03267982382414214,0.6031569322859252,0.13060307462156318,
 0.25319170082769027,0.9395297552281392,0.8596701483338941,0.799435175161718,
 0.8023580507584585,0.6879213726028889,0.9070059157096592,0.32679364425533486,
 0.18591619884909727,0.27041581460644326,0.32249509635756834,
 0.4204532184530584,0.7151353628650992,0.945020242512625,0.26515079332113256,
 0.13802451279894268,0.45494148646968025,0.3840120973340031,
 0.8441644693665972,0.8257694178174865,0.9480087642681646,0.15080632263992755,
 0.402145122049246,0.7125123085256392,0.9375482695708277,0.14365930068196253,
 0.7338077487127487,0.399551412569381,0.1336495475906312,0.09528576188914728,
 0.527515356914139,0.46345590993831,0.15630839425576415,0.32107312809382504,
 0.5066341755669118,0.16771958515937913,0.625401065666885,
 2.525138623735712e-4,0.8328079443589251,0.7491774600721779,
 0.3201330999789568,0.2618738510596148,0.8567760225354993,0.36421039507576536,
 0.15914324005686753,0.15458508237607127,0.264392404530584,0.3561265428132466,
 0.44854298831156836,0.3958267575657125,0.2902855214023303,0.357386572624644,
 0.15168053412182403,0.6220469825713028,0.9184838832289488,0.639618815102489,
 0.24917042950400292,0.757134411352501,0.9070153898961525,0.540221403807064,
 0.32750854619464354,0.7096953279118984,0.9654666763565334,0.7630071588378875,
 0.7985175783844514,0.045860370048793286,0.0015998596405930599,
 0.8218592024241568,0.15189521218036683,0.974541519586921,0.46203612138716654,
 0.3114402551643579,0.37064251088439304,0.36544978748582435,
 0.8282664683629133,0.18458455736050605,0.38557848565806385,
 0.8980173820139847,0.6104580621406089,0.5222373459522761,0.06343338223685469,
 0.03338283445299528,0.35306183557835635,0.30605047600583524,
 0.33320485613697226,0.12742446250614325,0.2094826261020526,0.840615553110211,
 0.30254447624067415,0.15864389380213217,0.6836517218690132,
 0.4116577608684536,0.7252053228455065,0.29214158187752315,0.7283698183853988,
 0.7687605720416734,0.8075750055381395,0.48749085448716256,0.4672641903077931,
 0.4870240258749995,0.9544634962677738,0.7103363048673824,0.11709222317249923,
 0.3690218624878303,0.8648808381155302,0.3685006003986354,0.5413620666225799,
 0.03953463489986553,0.31428787673187364,0.5257099810188559,
 0.030257050932306573,0.8403349540784222,0.005350784150253585,
 0.06298702455821192,0.833931078336308,0.19381147946114718,
 0.12694117785412895,0.6822203899848931,0.22708454473036732,
 0.8870326460422229,0.6148229142451886,0.9021245947137092,0.5180795724137177,
 0.5064162040419988,0.15155202397070588,0.5752814113041668,0.4792022156305229,
 0.9967909886662164,0.21891772386042874,0.4317725590674094,
 0.11673023504588076,0.5839853040479243,0.8611599228232066,0.9715096291835199,
 0.4323136111388256,0.5146573638015747,0.19824787026897484,0.3953883014680085,
 0.39939540016894926,0.008424046338418978,0.2585126934334593,
 0.7316719414542009,0.42003307866508943,0.8343109691131291,
 0.006207306981900862,0.7054168165342991,0.5266661938463629,
 0.5553536254123969,0.7781771023931667,0.893466143873618,0.3964191088292759,
 0.8362503942894829,0.14607440017227302,0.5963433200468343,
 0.08254591188537441,0.5469309290317673,0.5739726963006349,
 0.46439283029443845,0.7252234512684945,0.0669108552929516,
 0.46037261422781983,0.9506092926345342,0.09207399362095481,0.750471117943085,
 0.4269718611832184,0.05856298971035834,0.5312490323941875,0.7579951259464088,
 0.39369919513559926,0.8155602281597444,0.8586019179089747,0.491032430796949,
 0.7881554115657117,0.07450988197658248,0.682620050659867,0.7537996331346428,
 0.6344400266517566,0.27861194725249716,0.30406139902376805,
 0.8504667502440711,0.13652630811218947,0.3005405551995677,
 0.10456830956519059,0.9440389115837231,0.8037683230136652,0.8216626137017831,
 0.7561856082470084,0.445445886664956,0.3542458138819575,0.6331675156068282,
 0.6710219694739583,0.3352643873138088,0.4581807836519455,0.951238969713973,
 0.9842794029688948,0.6691035459821433,0.7595679983050345,0.6754994314068524,
 0.8013987775263288,0.7484801329400472,0.6078636041202307,0.6764978344612668,
 0.21290774813500246,0.9045320407838092,0.4455836557523576,0.4271845539205712,
 0.7388483575095288,0.5217159366890993,0.014807113452806164,
 0.20740275648174045,0.30753678708893173,0.32352984350005354,
 0.15721681583024005,0.7667413633115606,0.19747526227241619,
 0.9422446251300733,0.09235898599486081,0.8759022054206795,
 0.20514550901273154,0.3783859561573777,0.5062839170471738,
 0.24956949423797703,0.3806342034391401,0.22239744874813372,
 0.14797476941957555,0.2936949153354279,0.3380533671457461,0.8096772088334856,
 0.6418120029963986,0.3528431739902347,0.9546366008221816,0.9389418750286427,
 0.24400920684737892,0.4996520619297957,0.04686170841719717,
 0.07702670878988227,0.883987282284389,0.09860264600599045,0.6621119569088793,
 0.5603193764604293,0.8383237217097708,0.9713166427541552,0.28911209278741534,
 0.9090280324995182,0.7323513991496972,0.6175265398276045,0.22807695426861496,
 0.7088547543595749,0.2875590406068954,0.1418466549156412,0.629926849523671,
 0.13155002417488137,0.7482392850035411,0.7580924399313882,0.261908087405585,
 0.9551994535476501,0.858111891279635,0.8459266356152803,0.09369014870988557,
 0.6744333915188591,0.6657339734970948,0.0804114011619993,0.24461377371391024,
 0.49787234874686015,0.9943991331093875,0.40162585882277657,
 0.30770544470813066,0.9361049038547401,0.6294801963851303,0.7083766983658679,
 0.24104863792981268,0.7940847987066442,0.8141294678052439,
 0.02443755990216878,0.14123165955348083,0.6646955581695352,
 0.11496466564911256,0.9830476451530983,0.7483961194566237,
 0.10589119816688497,0.39341149577782975,0.6792926419457749,
 0.12035001592080885,0.18493280120012678,0.7125118008913042,
 0.7493201557705351,0.9372429701559328,0.9054946599463874,0.8995837348187516,
 0.23436740119016186,0.9931273359898309,0.9116481697455746,
 0.09216474480756853,0.12516281800440365,0.7604835315326801,
 0.14466121738082183,0.5504387074608774,0.7921719707932248,
 0.23021641002314153,0.32376818164125265,0.43329445070388506,
 0.07102551026681958,0.8240746813531645,0.8066833522592058,0.8506982003872493,
 0.8881588668985346,0.7963914074594408,0.4118069768920656,0.9555304895379472,
 0.7263345524723733,0.5240118092228646,0.34523515119591786,
 0.15995634656697066,0.39901371812569764,0.11897142513044656,
 0.3423063174736437,0.13158895173722807,0.7150675517888975,0.5777755607809314,
 0.9898857032733882,0.3627394092255073,0.37666033954479883,0.802449801589391,
 0.6943697080417168,0.9520562805938994,0.10752390655821631,0.9342741665775978,
 0.5161521472671475,0.0766233605786909,0.9696393561621345,0.6816444690468977,
 0.1643591652174915,0.7478521019626326,0.5301688119686857,0.6946298383756467,
 0.23493705452464997,0.7324099838223084,0.4502221469681231,0.9404235669263581,
 0.07043233194617693,0.7194607198285223,0.1933932481409073,
 0.10505690582128346,0.6012171242610053,0.7453780572422767,
 0.07467598766169847,0.9440042107035689,0.8070808188246084,0.7943571633127606,
 0.04272379270435234,0.2884960225010692,0.2924205741775471,0.5520225876166529,
 0.38925447184200634,0.9957571731206017,0.8443948208284041,
 0.21569314990172783,0.3491204979417687,0.7534657343589086,0.5349020358893362,
 0.17423210582240156,0.5541099884408274,0.4447915112494618,
 0.28077503125674264,0.6214165142633256,0.998695716175177,0.5476228175331539,
 0.15618072246139425,0.7007369341428137,0.09407158729173559,
 0.28662157844371383,0.4033348446410332,0.9883180784910605,
 0.16012215372171157,0.655117496673221,0.2865804252724887,0.7459588235456976,
 0.5535233270052674,0.8804875652824405,0.6562283454415085,0.49871753332407176,
 0.728431590403849,0.5798904570357389,0.3573516793266567,0.31140881412484334,
 0.5181798039167509,0.2922655306150742,0.7402834068000967,0.606510679210772,
 0.6180304855045895,0.19018762707572212,0.5361718409273781,
 0.47154187977808626,0.6473105049876888,0.3578530415957656,0.2721029519030791,
 0.6204129779615148,0.4292260032604449,0.1769541990983845,0.6813894492070941,
 0.9807577032887891,0.9187556112166184,0.012998488579088474,
 0.17984477457617132,0.4478484825749177,0.6197583122966314,
 0.10780400324965012,0.9884358399835271,0.7279174139404072,0.4463268868897643,
 0.2277278990892727,0.024130491725756764,0.7541858692572262,
 0.3576333065905909,0.34690363085396725,0.6204879594134555,0.856330829621667,
 0.8340416845347102,0.6558503931762542,0.8508432296190794,0.15030463066677635,
 0.882270937079174,0.6980515098747472,0.025886188275484612,0.5183916596892919,
 0.6113625067520071,0.7679933927601478,0.4830342065752832,0.11400575933216217,
 0.0535395906632562,0.8428761467541869,0.381518174139157,0.2316922830114846,
 0.5615796126399818,0.8328287072233641,0.8982416479373784,0.40300414302911836,
 0.20789117718525718,0.9989473737176001,0.7249072593982,0.2403592858459117,
 0.3832845763924626,0.5769381886646444,0.14478411790703283,0.8898664922618248,
 0.11557437258858849,0.8337572304396863,0.67481709454158,0.9144068069271006,
 0.18560651592948962,0.027773120307093335,0.11476479461779565,
 0.27370451852112,0.8766879055178638,0.13004825071222814,0.6989423986755375,
 0.003633676953190945,0.6228325375420577,0.8619165651930488,
 0.04740159012654055,0.12810636680436616,0.607833670488888,0.9514883832555018,
 0.5002236830421772,0.7212045849738059,0.4450997878139874,0.15078736577381058,
 0.47876181770760473,0.729398629697691,0.5084750931390178,0.4366224304164489,
 0.4135968983273328,0.4185230756240209,0.5780005069169027,0.4977386130006739,
 0.1779779215191406,0.06704970659142184,0.7805115700837015,0.4786104878320955,
 0.21347874656410326,0.5396246180199923,0.0021194930977822546,
 0.5443631743752424,0.27878062917058744,0.03990547597212679,0.923909988509874,
 0.35492180118144445,0.26997678052326113,0.7111951788127691,
 0.08133579584229311,0.5383689521752363,0.8923084795794983,
 0.023073910939858777,0.46323334089238355,0.7062613000677569,
 0.14375036604207325,0.2528840101619273,0.9871746838525104,0.210752395267788,
 0.9433108247545465,0.7860971062728067,0.9057177126518824,0.2502561286132251,
 0.5217747473424159,0.010036349097844788,0.33688666801902545,
 0.8096616101540661,0.4375023585273873,0.7366355969102484,0.5090680695556178,
 0.03593479123080301,0.6166409202526411,0.6916213572624712,0.325832401200179,
 0.2965782900405819,0.25914429055055044,0.8010604297508453,
 0.40165074177602944,0.9198641531366225,0.690941395996169,0.20046285127515218,
 0.22657612779517466,0.40969417527041563,0.8821811246568929,
 0.22018334978958043,0.17691471868729403,0.7771821689359074,
 0.3829948778906104,0.06743122032036186,0.411079276418871,0.4878697663684549,
 0.7056354270902143,0.029689812601657706,0.29360476249892775,
 0.491049693406885,0.2878822715515632,0.15203533341514042,0.8632460821479756,
 0.5899997096007302,0.2669491181400878,0.29177914039207853,0.9558368688808925,
 0.5132309366242955,0.8126297912430208,0.6314706113517125,0.8718105049409607,
 0.8558844729256005,0.3163966954544376,0.9448164891940881,0.14157395280454188,
 0.9847612435708559,0.4205639746277805,0.9771479367547279,0.9409797204474073,
 0.7244481840318872,0.5558001595264751,0.5346251529556741,0.5559594488917,
 0.17886546242138557,0.934470101144136,0.8600562189530536,0.7194779681246136,
 0.8190246064831876,0.12397913496680169,0.2305465469499568,0.8068175803871828,
 0.415126401876425,0.530212686823347,0.9677396858689653,0.36483500841158634,
 0.6530371738591524,0.9318879804321414,0.9728938115267713,
 0.003741338079752299,0.4355843613156761,0.5122210957258491,
 0.2862022987195456,0.7122610227073026,0.8253502084785556,0.2717431093102609,
 0.5306612629282301,0.5741651958322395,0.05424881919198277,
 0.22231451360980214,0.14761877637868936,0.875783322926575,0.951221972962543,
 0.7797820878678425,0.9338367644492412,0.41664108689987744,0.8515876703644315,
 0.50083960545675,0.4864644058983565,0.9522414704681174,0.9490816997373469,
 0.13087488242639,0.7465318668053076,0.18248180695972605,0.6038972810303058,
 0.46614935523352985,0.6463254854088261,0.1303853908680026,0.7948915313892645,
 0.21849221815820863,0.23469825664610178,0.8925352152391395,
 0.8439433125120274,0.834904811338504,0.9844815853200332,0.012746292756946498,
 0.5585378409237052,0.060441230626845366,0.05860377076688206,
 0.8953414237767835,0.20659927965203562,0.25831251972607294,
 0.5685713141155455,0.9107407854430811,0.9558217318316716,0.9512650010447671,
 0.7191319393083755,0.2858957627014318,0.07634778827009914,0.3768722353666779,
 0.8799510593066315,0.04940296307226322,0.11809205602659367,
 0.2816193699415299,0.5580921107911946,0.5937237523380197,0.9059020898967515,
 0.9178084107231088,0.3935301655529063,0.4007079526226196,0.671019314545216,
 0.4367228890114556,0.6315255411973725,0.35113927066532835,
 0.014957098754399989,0.024157668205714078,0.13081401110648117,
 0.42724671256175495,0.5009381517913557,0.29152642611871527,
 0.5113199079195574,0.5672617615923892,0.88792319264478,0.2258737654336067,
 0.40428234995340684,0.19464217530913064,0.19166384368258238,
 0.5627607220441309,0.29386270079324184,0.5329824195601742,0.2360415160147714,
 0.5255222720654672,0.06721779945461548,0.8201837410619353,0.2515276719729985,
 0.8017611023932978,0.14345187509831403,0.3317578496621547,0.4147736761381474,
 0.7114325381297862,0.7288284254981483,0.11373731610713655,
 0.30099075502063777,0.5308965307295226,0.8262877801463834,0.7730091784052228,
 0.03062796885834429,0.835886235718149,0.9190109274660283,0.08466640828053285,
 0.7027850134531114,0.8749612963917158,0.540586791024574,0.6443199687347501,
 0.9614828711158422,0.4311440174324529,0.22511839466660566,0.8083421367043718,
 0.06201833898838682,0.18147415954005985,0.7368866938197963,
 0.6397457912581069,0.806981351046073,0.1987530291931784,0.4906585046694969,
 0.442909070694693,0.6395928982535733,0.7752612958062597,0.5576994082302367,
 0.696498529582395,0.18373041231411125,0.7893128590631958,0.16426623773069282,
 0.2962994453632185,0.9718687488100252,0.7835346552523758,0.36479069117661245,
 0.5823216194703827,0.29738502702775216,0.5029575256555011,0.2605506880488855,
 0.17221205023042585,0.46606038748650025,0.6998569948298532,
 0.8076414990962472,0.9528420665151038,0.9035266923002345,0.02317486073946362,
 0.11426789606672505,0.7907472320474491,0.09231576416411147,
 0.5896884345538165,0.1655475674305824,0.9952596809146389,0.5832454306916404,
 0.8540331600067852,0.9200389350275114,0.13949512821694987,0.9086262608063564,
 0.5438901465774313,0.20903868970085182,0.3575109118869445,
 0.34427482169721135,0.2609986371273584,0.5637297652242419,
 0.48303131182520587,0.5617407598031566,0.6533259123477813,0.6919052336704402,
 0.0497286948427913,0.46139571320578787,0.019120153441480103,
 0.8090456869770157,0.8494442300031033,0.8555573743042348,0.2641906968669675,
 0.3783249801584311,0.6779475706747518,0.0633345294379365,0.5114818782576815,
 0.7342085057191248,0.6628969197253228,0.5794426291216485,0.733399574359419,
 0.7809804086207246,0.9977226844820104,0.13700204512587788,0.641474661412898,
 0.7912275941388834,0.5586822504352922,0.3116673666452575,0.21392156930973216,
 0.34708776028233124,0.11737826019012221,0.15851733452868566,
 0.06705693851761685,0.21633135776327805,0.5507444293529118,
 0.9933111472688391,0.06294571113394831,0.7714186333339687,
 0.24033667924974345,0.2957354177827065,0.20076298145647886,
 0.9623435893512318,0.09704987288070133,0.25073579730733064,
 0.5079632573681773,0.27386483182894983,0.31264684190601943,
 0.40866977429807205,0.06472867075564748,0.08489411893068799,
 0.7674534688955621,0.6868063253608265,0.9538069014063101,0.7854531279608294,
 0.774765427448734,0.634327494198486,0.4092482511378984,0.6244982990154047,
 0.6199056041086113,0.4366870184398528,0.7431459365439399,0.849445435929346,
 0.5861828894703263,0.7925077550836158,0.48582038769349034,0.3787470652949164,
 0.8829636354450406,0.007069155199744825,0.7244024337549552,
 0.8100250344936522,0.7164542082464148,0.6557350042593463,0.32863782674444253,
 0.18545469688142857,0.656148559482267,0.009490797433027165,
 0.09460918622198666,0.8028817926350545,0.17626969520536118,
 0.3942284349608508,0.3072807834408754,0.7869703290620766,0.1362847952207249,
 0.06947676557328486,0.28262763397803947,0.635495032257826,0.9008955186644021,
 0.3784776007149038,0.5566213576046525,0.7062886053045145,0.6980794297745858,
 0.4790494656557789,0.3386043653491,0.734931777758628,0.4827265470292674,
 0.02933648066994432,0.0525042766138869,0.2147244779922256,0.3829076067209143,
 0.8731033322713652,0.47066158039139494,0.6795132992872522,0.6464845753586295,
 0.14882381809699718,0.06097821748115839,0.9583256754938074,
 0.9687323956829565,0.38340353394397786,0.11812911480655475,
 0.22195736036337188,0.5372245783919145,0.3809839072415553,0.7351510053733441,
 0.8216868635290755,0.6397584642069327,0.7234732870323929,0.6742164287991252,
 0.03827454320177803,0.5910898722303015,0.062310382149407095,
 0.15934952485112786,0.9788313841624421,0.982303298644977,0.7697716729365411,
 0.6128378724860245,0.4855075782074638,0.9418205731145974,0.979249037603229,
 0.9298447095361476,0.5701986136082136,0.8646693047139068,0.9230079895960576,
 0.4121940753215193,0.907459049322865,0.0030744949376416697,
 0.6498441595556396,0.9033024426421506,0.6990255977277084,0.42814703933169374,
 0.5460520398506052,0.5814936619704217,0.3972693211014029,0.5909908172525434,
 0.8600596824181544,0.3496359506101786,0.586145369036882,0.6708035825021222,
 0.964328629218904,0.7418792554867567,0.08659902749768977,0.6892282570071129,
 0.8990485046440693,0.8697720353052842,0.5678715290099384,0.9771042742123154,
 0.24899116956271694,0.14599508412791373,0.06177998316394451,
 0.4478344301409012,0.052769061086698765,0.09743108186036942,
 0.8685382381934252,0.8177514740514025,0.3225813688397693,0.32219385991927074,
 0.7007072519002459,0.9497921695089804,0.8832229100765876,0.8557572858308454,
 0.14627346720661905,0.5080311639050176,0.5509979154284417,
 0.055085529466316285,0.9784792353823943,0.9153416379962018,
 0.8086173666964902,0.14912839958069724,0.26723584506476616,
 0.09211280546774392,0.304546348579085,0.5039098445398159,0.1569551198979523,
 0.6459235321542793,0.4616650792253909,0.903940923044413,0.6147059970546851,
 0.3638193595831051,0.9315401136095776,0.6898862240417318,0.47096935188586264,
 0.7264221523943901,0.0423460213808694,0.963893336074551,0.0735745470816026,
 0.1174192597318433,0.2794541514650497,0.7457713756809276,0.9519130774250445,
 0.4088752318937573,0.18267420745693996,0.31576671298015657,
 0.9424481119477186,0.8198154814709766,0.350876535810563,0.7377756131582472,
 0.6038670207237581,0.06854310321145873,0.3620891613010053,0.906344934328042,
 0.40693641282740567,0.7099119078740121,0.34779874971818225,0.61925831982346,
 0.7907224767337765,0.5478452680145162,0.7261352422994399,0.4927265071437915,
 0.6394122907209948,0.23249724407511285,0.032059382153960314,
 0.46632470281465777,0.0727684659270631,0.5894660854697394,0.910793325484195,
 0.8714453555701153,0.1598192781085488,0.9907365487916331,0.7770318585868354,
 0.41222411087250665,0.659818732560762,0.583121181454848,0.8850029223207996,
 0.7637210257747761,0.6286456670456766,0.36047278250175474,0.6923680686034401,
 0.584245857706377,0.3137343791144984,0.9745787837952645,0.02024000003235249,
 0.19020597175559162,0.22789670124457972,0.044681946189767285,
 0.6014423042638098,0.3279611857169311,0.1307666660593162,0.11631871619945144,
 0.7119989099113696,0.6431034139691599,0.11677601539105331,0.6977947008559637,
 0.8624424673541949,0.8430149198216501,0.18075076539637136,
 0.04900673867717342,0.7230973357663257,0.5752972129659152,0.7068140720242135,
 0.5488796034786491,0.9533994711010669,0.7250566853776209,0.14061425576829645,
 0.7265918332630793,0.16402033021491624,0.5492131686535777,0.5759141277596269,
 0.1489678261627384,0.8771949036054502,0.1207342224481267,0.5953509680012397,
 0.8065282735534814,0.43968608738776727,0.8565555436555177,0.3603367225939659,
 0.39406673256336355,0.28760153732725524,0.45215487166598134,
 0.37209387274236816,0.15264707427354474,0.8043851305266755,
 0.9052954779742169,0.07078510941835725,0.8861732989823935,0.2594379715679018,
 0.8141032340786697,0.08228145814887367,0.10641802214156226,
 0.7215021407177096,0.8320131765616962,0.5992637454658907,0.10691202548154621,
 0.27528865999935437,0.5774175658843859,0.554610957815561,0.9980113633300418,
 0.07714722764111448,0.679767324401132,0.5734672787671052,0.14379562819255753,
 0.8741441663725444,0.51714716827841,0.8961574653766888,0.045827747983591016,
 0.8940841060700491,0.9187034569956907,0.863875005672299,0.7322245964230596,
 0.6995522104761945,0.6322699875213871,0.4698885767587433,0.4337809477515254,
 0.416290000456357,0.2317130691976529,0.008481197485946002,
 0.03227812752470893,0.7777109271075734,0.08554291707726247,
 0.7603270165982312,0.7013657772443403,0.7497843884931574,0.5667565829666197,
 0.15187243020500063,0.35914924178451335,0.04246235156663236,
 0.3339027483271114,0.8989601782210583,0.08638124857423146,0.5328733134767134,
 0.8326194562036708,0.9334475440914108,0.033016341708622554,
 0.5945716850702367,0.5169729725526133,0.5479407524411218,0.7427425529551104,
 0.9466656095258514,0.5531844770356213,0.2606932631253561,0.46897661217417963,
 0.3557185139763164,0.8926875065329868,0.9450512966882402,0.3243913640428878,
 0.7281700682759455,0.6165762294987331,0.6256680634296581,0.1353551584595708,
 0.9790293115167409,0.20374749221335986,0.130409688850459,0.936698688964276,
 0.021599758528718094,0.005390841390362611,0.976380580710359,
 0.16102380589204512,0.6079847993484282,0.8209640356506852,0.8305471782992,
 0.9868858918509839,0.46561264088097354,0.992956376393126,0.8099492480014554,
 0.5262492417966033,0.7551973247101504,0.7390627605764424,0.573254107596106,
 0.8809427525131511,0.1601801341114908,0.23390880090176125,0.6359887598787394,
 0.9016655869119725,0.32908394719395084,0.9848840974941022,
 0.31796848129689614,0.7553546524466875,0.9294762777744378,
 0.30289499574089707,0.5861289650144972,0.007236049639861575,
 0.7113750336321727,0.8107043527930514,0.43734970187881506,
 0.08792321213533971,0.8012197161192923,0.8371293327407388,0.972552333736417,
 0.6282472037722991,0.07276396732739498,0.44203242760629635,
 0.7047737728230252,0.7020208456917731,0.9748090644708111,0.963227198674223,
 0.6784848110595912,0.7966507163394347,0.2616306979190265,0.3653682365295715,
 0.33731862779870236,0.8683572471552354,0.7324621513647509,0.7453411003457112,
 0.612015972149138,0.23603496035217497,0.3572655929283092,0.6463841415954733,
 0.06926981351421857,0.8935421725063896,0.31172304454924604,
 0.5328371919821998,0.7619207041333415,0.12244487481107069,0.8227875459129262,
 0.4128999156282288,0.29434746401170253,0.7907765465058453,0.3609346362419008,
 0.41546786367359356,0.9961113826953267,0.018213631503422145,
 0.43504515121065557,0.07615419027169001,0.38763659465578426,
 0.9070926885353439,0.8515826160125882,0.2624876655364786,0.35458773399717447,
 0.13779181070686652,0.9769727932095543,0.10625027043972124,0.511642760699767,
 0.25316732124871344,0.08048951875900567,0.33119741131006786,
 0.9885820629836588,0.3170129248495652,0.15184228376581876,0.8455398546202212,
 0.10335443766526331,0.4501326848164362,0.12460536301859404,
 0.5308533338741456,0.6288403983805324,0.18250498776258162,0.1429298998476387,
 0.8723680097509816,0.6993121358915096,0.5805467875511717,0.1500189497508153,
 0.8065108722228844,0.685652851979285,0.7229655724219437,0.5965767716970309,
 0.7933050324897729,0.9526995530127813,0.22541199397499623,0.413127321727235,
 0.47113181898912715,0.18041408335443454,0.9538780126587525,0.547629251974878,
 0.719217599444121,0.31124066357044244,0.3587707343865043,0.38572588691486986,
 0.29738167035788,0.6864909979055177,0.16411577308083525,0.9407521664771707,
 0.2714026434428626,0.009992627662593123,0.34151265866247904,
 0.19834224639051834,0.9431534545075003,0.40147272792893995,
 0.6475582157821749,0.45834624276668123,0.31512298890866297,0.8032330118804,
 0.5913316978885657,0.998268831465487,0.4459055852955781,0.6438847730869891,
 0.6413866976281591,0.10685700449775115,0.5218823650221148,0.7310235630831206,
 0.08232329945985895,0.3147450816855747,0.7489507634072079,
 0.29195325767817226,0.07444693666814228,0.3050421707517085,
 0.5193167925080897,0.00119858434269271,0.9720463026069478,0.5455450743646879,
 0.3558101620506602,0.5367759098694168,0.43848338419408084,
 0.37379516860819006,0.471255168512032,0.6014352815970736,0.5964713373220432,
 0.010913120028386625,0.7646120197875518,0.09171574321328413,
 0.7399374543840515,0.620929118198344,0.2056995827242465,0.8821735860870168,
 0.927643876971417,0.5181472385919249,0.637841199772829,0.5705243873354366,
 0.7009937270206477,0.38222263576901416,0.5532094703533534,0.6454962646390072,
 0.7974997757825395,0.43026727822319644,0.6753795495180221,0.4915067578850203,
 0.42300562444356027,0.06324771158158748,0.557678788512309,
 0.27481006626857973,0.09354174848268981,0.6453163970932143,0.558859939533859,
 0.4871417498071866,0.3624375213681129,0.6702524322962704,0.755936428000911,
 0.6395838493780829,0.718724269511437,0.22021732980090347,0.29749367128102155,
 0.5656335098520934,0.6345792600693709,0.34463675437735075,0.63914090891899,
 0.9001701880967483,0.8820259997872362,0.04200248521679528,0.2720300782529975,
 0.6786773077448367,0.9115832520567723,0.6256696163399743,0.9830081673417119,
 0.6712845467333132,0.30450529718334285,0.20586751043205465,0.976021945803492,
 0.062279846656808524,0.9738440991635476,0.3085110730728864,
 0.14643959610365287,0.6366182313664674,0.30317152309184314,
 0.034040364167017834,0.7874110787463882,0.5927018248342613,0.975703782274869,
 0.764530792127546,0.9250457673665226,0.7218938835764099,0.5511054648184461,
 0.8328253961199791,0.8747285421422141,0.23488366953173323,0.5010083032305854,
 0.6393891741403364,0.6179573230852033,0.20102451336902494,0.379287642529863,
 0.7249300101752639,0.9542561779832116,0.10840930456155473,0.1333711275260675,
 0.7975768112490975,0.9523206084516357,0.15184953578374194,0.634616324151891,
 0.21331580864646804,0.7593713231432195,0.7846938782719844,0.909031576286811,
 0.46305764455258347,0.12193046320644796,0.8382754631429137,
 0.2899456891002453,0.5801914174656136,0.5894990448014488,0.9924259522011623,
 0.17875633719597261,0.18589111575458628,0.6959993666344894,
 0.9352138760099704,0.9720776820301303,0.18884841174716227,0.8190812145939858,
 0.3257855335558809,0.7615293481727586,0.8306838109879672,0.48952580362233034,
 0.40155364213248446,0.9821529262645803,0.3448426116117932,
 0.29322294833316076,0.961942288459202,0.9467075192389064,0.14420955986970918,
 0.7069972407044043,0.8402413320037756,0.09069921688034888,0.0952804838206992,
 0.23334877918769303,0.6347179038154412,0.2919097134887374,0.771705640702875,
 0.156007267153491,0.16192045921331777,0.7271332789201557,0.39575458809884245,
 0.02836824473475863,0.7745453248222158,0.4782195934703567,0.8511841981110216,
 0.5697900805456606,0.43532888287896676,0.38961871191992636,
 0.27827364726172643,0.9334526271254312,0.6361831678846589,
 0.16421195297263902,0.45007862501677576,0.5146331478226074,
 0.6639337185774874,9.022609619683264e-4,0.5355080169472701,0.771141512261568,
 0.9375528582486478,0.20864052175508263,0.7290090386521678,0.4679297568856805,
 0.5040672898485288,0.03102306567514246,0.4932032619148795,0.4602486047291643,
 0.7617512681210777,0.6433388774254982,0.5294874579314148,0.8281872358516997,
 0.9669420737019792,0.5194343300101354,0.5605582215972903,0.746341666612262,
 0.6129550691130832,0.8630790020573712,0.8666918113547377,0.7182683211155556,
 0.37020015021475816,0.7439022105453447,0.1663656021082528,
 0.48542967733976805,0.7629754096136158,0.756675818546695,0.9390213126167021,
 0.8721607582244538,0.45089374603645493,0.9284564909849157,0.8814169010603711,
 0.6288334305175941,0.5340053055369974,0.2812309824905541,0.33303760585360775,
 0.7490998589230894,0.7260260236780607,0.969462164147628,0.05159063723604218,
 0.8902663191789281,0.5916368887924026,0.6638695460548784,0.47609768334097324,
 0.28745376872104,0.9287376514131851,0.5515574136851484,0.43654306007536425,
 0.09658081050159684,0.8709965592516155,0.8334409464038561,
 0.02227916828360421,0.14736773447488516,0.5888395115640883,
 0.8046067729285226,0.13694844555915897,0.7306715541086767,0.7226546852338585,
 0.44416683072659835,0.6983910685653552,0.3396041490816625,0.4671829088695977,
 0.1062741345830085,0.8865015398692535,0.6040304593831319,0.3487519562853215,
 0.14255591760403263,0.9638724094169433,0.5114386013815816,0.7140471859278923,
 0.8895957694206428,0.5363469464925414,0.6064888697018691,0.15960421510183664,
 0.49209415529154965,0.3281226472168326,0.7021444794912055,
 0.12957646561337777,0.5851210393586671,0.07435104139813198,
 0.5907798129263979,0.6235885212996483,0.9050601884773525,0.9664185866464698,
 0.3770092213389874,0.9794399615630266,0.7486423376547966,0.8519906537789848,
 0.5165570659606565,0.5898233486430158,0.3849816172497539,0.766834647505676,
 0.6063919882361478,0.9938487102309094,0.250110445561712,0.8251391723244338,
 0.7046355531744476,0.9004493390267736,0.16296685334237737,0.984830340421795,
 0.767751410874729,0.2704462479681792,0.7299530845354519,0.14522877096388476,
 0.2092856402531118,0.599664842832723,0.9328146072750619,0.10416772578470213,
 0.33071597679959475,0.7945588454756924,0.24078605995562097,
 0.41123783776104306,0.11867538243167552,0.01968952050673123,
 0.36381800098706907,0.5298364717692472,0.6176423974696679,0.9861597750489524,
 0.11931571934704421,0.6318869295933761,0.5867680706080943,0.6233760802440566,
 0.7066633125615315,0.00996599442660595,0.2204324778565735,
 0.17018164858308116,0.6216065414693555,0.3924286737531173,0.3890403299525671,
 0.878648763894432,0.7731697800757459,0.36935323150207433,0.8395248058008619,
 0.9228567633424503,0.8855356488437943,0.4718686150178073,0.5734321774806246,
 0.5094363786685404,0.16563500886490978,0.35679453109670434,
 0.8851186599999541,0.620525712033391,0.6119537913499267,0.24708500244567744,
 0.3774119302035116,0.7302207312396796,0.12830530628974357,
 0.31178338034214403,0.8242730876545378,0.8383810610868916,0.537741193226459,
 0.5898040395470299,0.45333269887312255,0.10113448725390728,
 0.3668779850834807,0.764197649989708,0.4058021284483342,0.6541837586570353,
 0.0023743035965275716,0.6743491382656455,0.93152355225875,0.9094510812608018,
 0.8862854538961755,0.17392581591016465,0.7796329699894953,0.9858621132504591,
 0.7897746371854344,0.06356781974657488,0.5245834263475283,0.2903393940847151,
 0.2367115992307688,0.5320040953852387,0.17930649621150596,
 0.07446740981878497,0.8645874687827287,0.19771140689087796,
 0.6371840974091403,0.6141901315922835,0.17586213779987125,0.7933060355120389,
 0.7262513491232089,0.7541512376098805,0.4192138156341192,0.9754465409861709,
 0.24416446369736278,0.333722931254275,0.8692947168502365,0.17169834540294315,
 0.5453186194523647,0.7389507656369163,0.3924444800221395,0.3057582767472441,
 0.45255241966188586,0.9814326138344763,0.9117735000197601,
 0.15766108932790013,0.1866072615978156,0.506491318314229,0.740951606390118,
 0.5974244775792212,0.8821646017064009,0.35208850797256463,
 0.05574675075337865,0.09144996410424644,0.369603977200337,
 0.10682641583016816,0.5051729969327481,0.6420532277258142,0.2892578113739628,
 0.6812344424642449,0.32769054157712774,0.5466317434710288,0.8298871395532362,
 0.5064493637425406,0.07828696971816507,0.06819066236204718,
 0.10128831807597827,0.45013870534600586,0.6373987750637237,
 0.19398832870405758,0.15790051219843204,0.5326328532365864,
 0.6478052389958181,0.3577889052550294,0.9594569580313528,0.701225608883415,
 0.6660139264654511,0.665208569516043,0.5634435602481791,0.8178963921579577,
 0.47164942772220875,0.7436900018502972,0.6646240118091152,0.6028466572647996,
 0.6464971549159864,0.3969788117505224,0.16320549997112854,0.2180267999965272,
 0.9354019099874499,0.0754089814435075,0.15492346532587276,0.814992934337391,
 0.7571004079466132,0.17303606031327523,0.4536074729305052,0.7848011415115804,
 0.11458753978707814,0.2747669838186302,0.46664610556767294,
 0.8277207696397362,0.5254572710113536,0.46006853182528606,0.422916528827582,
 0.176515629936705,0.1301510138709101,0.35077062044565954,0.8969030892359958,
 0.1259374323042941,0.8518841081972024,0.979909256469515,0.8832013988799128,
 0.5907219405736314,0.8252340237335545,0.8400476164785176,0.28775127118009036,
 0.42124295723806837,0.9542762908328961,0.5069548657121882,0.1427708310305411,
 0.6736604093607379,0.4518860475322649,0.3640965775491257,0.2081885938947584,
 0.21146647444578115,0.3597335625364413,0.48752493071799186,
 0.6087821203033238,0.5466770644959257,0.3209156424314219,0.25251411157972536,
 0.2542517179978043,0.7581873881574624,0.9725992480218467,0.7554880654828524,
 0.43424894832364136,0.7455624409721817,0.628077577929746,0.11491819393899672,
 0.2612184772528965,0.09288225885994394,0.6569606768063095,0.3374255687629313,
 0.4341088131383588,0.9295214289081345,0.17120726446543022,
 0.21095854049945328,0.8935376076614177,0.8244355352820979,0.7727208657829883,
 0.7184739564805004,0.29686442846871164,0.07085986924479581,
 0.8453471114836849,0.006684271488923144,0.5873929649930393,
 0.03958046799311743,0.39043672377541183,0.6715328340698646,
 0.8827911671262996,0.7931970796101875,0.342852978462865,0.05204195694172209,
 0.7339187879498504,0.36541805547432793,0.4332909932864124,0.1799925624857539,
 0.14865113618047532,0.3390077584886252,0.43927718652070247,
 0.42771868315187933,0.9819243347030788,0.5477482794420772,
 0.13733350041780268,0.3062598011664721,0.30502997431955936,
 0.10401082899304581,0.41677641872208837,0.71911627305377,0.8116496646673212,
 0.634207341093041,0.7020049382898933,0.7529328197141376,0.7682627198265748,
 0.9083488019874114,0.11024896587838884,0.7929835824987062,
 0.13759663161853353,0.15245218167701335,0.16885403691891576,
 0.7477865206356614,0.3325448948351508,0.5552204439105592,0.13570497155048344,
 0.1891742528408571,0.8678696378426473,0.9954443670642145,0.42460510706671095,
 0.35613131661371733,0.7409553998521211,0.01651632343295928,
 0.6015964401886783,0.19218075797272216,0.42076634513592237,
 0.9977660857264745,0.48765547967574807,0.8231151565757023,
 0.41935499024377343,0.5053883081096046,0.9033320107170838,
 0.37475681457110455,0.9407131973758597,0.8807599206798306,0.7576680358484367,
 0.9224122384168127,0.20107722451379084,0.6567465106805792,
 0.19243196110762628,0.9731198387043143,0.2355265407103756,
 0.27815760921277144,0.005365058793834421,0.8939131041068181,
 0.5593997120026527,0.8219754491240358,0.4395565414912491,0.7588757761409329,
 0.9912292202137903,0.8356816997208695,0.31555042031164093,
 0.34634997237706655,0.9393634559279243,0.06286940926775697,
 0.7572094091248075,0.8991549466691457,0.9299754953278134,0.07032228162649568,
 0.042187740539727514,0.2006002089026636,0.5999170500265366,
 0.1443591767708694,0.14970375791875057,0.5640390161461769,0.3053779076438222,
 0.3111778694483246,0.45829096226316346,0.3224662099746637,
 1.6635280926924345e-4,0.014916645698711761,0.24643566438453335,
 0.7548090494657343,0.7808207533426785,0.5113948124237606,0.6985594341347592,
 0.4266010833745867,0.4751459352232933,0.01037284591287757,0.6548256639721609,
 0.21416402612630936,0.05372837017943011,0.4877104777699759,
 0.7190108942721507,0.6546786226289406,0.7147530816186024,0.6316558097999924,
 0.50288473041039,0.7398282178209076,0.01718787460021809,0.451828156441731,
 0.7702589010171306,0.5961347069587305,0.3521730982900616,0.5600482347212457,
 0.09946415120582852,0.14194209950798353,0.09529485794789372,
 0.5364312566662942,0.23286682487350407,0.4083676186282523,
 0.45256124364984407,0.7477363060082776,0.8697135771695643,0.5130667353315779,
 0.4369030064970839,0.1169327257435876,0.47997258687106104,0.923651709616704,
 0.8499192035372333,0.546044867960347,0.3391771990139679,0.6133579474414598,
 0.8771626120739455,0.19726832290897356,0.1854169033348636,0.7545832461858294,
 0.5153613654479293,0.0010734702321200107,0.8668379576275563,
 0.4536976877497546,0.19812001576882798,0.39779740219979187,
 0.5584442290031362,0.03952879043072832,0.9774129992303973,
 0.43671104408826944,0.9027905706875881,0.9297217733073501,0.8115282556729868,
 0.23570929220791215,0.6872024767262773,0.4707385422670123,0.4666411903793396,
 0.9263079222587703,0.21292106911253006,0.3005572457669319,0.8992838843577368,
 0.4263227306954984,0.37489601614685764,0.017339881674700575,
 0.1988760111388933,0.6600292493948495,0.7043804788096077,0.180066188068999,
 0.6300604535387162,0.36608719601167117,0.6269122102451256,
 6.873425803695099e-4,0.596556040670119,0.21677477701961223,
 0.5517124293085176,0.5788805836621939,0.7519755729103894,0.3077645974988924,
 0.4150705359809731,0.30086551389380944,0.971411133956667,0.28028892155558993,
 0.6309170319493671,0.7416677327670185,0.6855649893870089,0.1371107821465185,
 0.30884950451136584,0.11285236321059255,0.9442437149370089,
 0.08213124257316284,0.08329689648084326,0.16668962078653515,
 0.03167653788859115,0.9911209693305414,0.7520176986992013,0.8553403709279649,
 0.7530419393920386,0.8346410348521828,0.8713848420034335,0.9559067329229084,
 7.327014070601656e-5,0.2719525259257639,0.6798900915892,0.38949707057174077,
 0.4741414806442832,0.29518466405004806,0.6274156252441563,0.5671223910006497,
 0.19412173024346469,0.8865843314931402,0.36603964929915767,
 0.3383751118639865,0.35415224037088056,0.870983510601907,0.4236833767032123,
 0.5449673003566226,0.30149757905080343,0.30403458571064923,
 0.22362109343833603,0.5877809258791966,0.32715618970088245,
 0.4522270304145415,0.6640113707413594,0.5532406589537009,0.09883034898159981,
 0.45723059484607886,0.8013768430784027,0.7049211198885127,
 0.20666246902565166,0.5979282579347295,0.5120208243762088,
 0.32611316898554676,0.774897871722864,0.1545322845039383,0.3726373013473061,
 0.7595788206611048,0.0908952662850383,0.7127929516086612,0.6046733897018125,
 0.5771098703304721,0.048160759379726636,0.012423587728078944,
 0.5701813985658617,0.32747172181305173,0.19335246746812973,
 0.15959711188681636,0.18638076859127473,0.6498573411086913,
 0.9162079320747792,0.4212598248513917,0.5689661003406898,0.681543895696038,
 0.4666484279269979,0.8514806456107177,0.6253004436520357,0.587481434497472,
 0.140731891859036,0.1747744016257632,0.9650361343541072,0.11181399818468274,
 0.49800479960109123,0.7342220464091688,0.8166295455796907,0.7768874047643339,
 0.7842777093443382,0.17133839320165523,0.01468409335400711,
 0.5892717505093389,0.47262097462188535,0.6318930936472215,0.4701840007387864,
 0.5964365690938924,0.3728340801767309,0.6189657016258933,0.8378561491086876,
 0.6832537373005673,0.14052459859190225,0.7192295790963438,0.3247890547766641,
 0.8552088860054162,0.561673418924787,0.11823918310475956,0.3529104836818393,
 0.34345455327163943,0.7881940711359534,0.758934598148781,0.6302543336405882,
 0.6208120304965326,0.22771329050494904,0.9158005709984032,0.974028443924018,
 0.8322062158738983,0.8156388002021945,0.9807614391896226,0.8783968593496977,
 0.11953206915507308,0.8389251252335769,0.8717035945188387,
 0.46644578668776715,0.9559497653974335,0.26365314599698286,
 0.4423069983005775,0.47050956373604647,0.9787221950335698,
 0.22817276165866984,0.5170638678888345,0.44663149130492386,
 0.32295257687061707,0.21839195742215545,0.26418129036212656,
 0.8103795191880373,0.012521002965498629,0.2663022418551546,
 0.8440339153719767,0.1828671107074259,0.5224486565340899,0.3277622561227591,
 0.933399214874377,0.6574057168340329,0.9655393365909501,0.25435821864336416,
 0.4045507080563294,0.39452062144198097,0.8870229021626549,0.9937869957180482,
 0.06737837097932309,0.5613343065786193,0.7601312355712375,
 0.49934502581313545,0.036383652312764525,0.0688213662820677,
 0.8236431637848864,0.5548043708229857,0.6729407531342506,0.7700551978626167,
 0.9003666217990453,0.4676637951528727,0.9335870430464532,0.9880643944952235,
 0.16071684363761674,0.1304554008094322,0.7411478975387537,
 0.26580158636348306,0.7930752365197188,0.5903904663992104,0.7966173986648691,
 0.22516874780685447,0.7692941816722284,0.12608224613706476,
 0.6054316225515597,0.5532905509427417,0.40116834671229273,0.5766297858965108,
 0.9187736733775582,0.2137652280815553,0.5468244461104612,0.5910782473702257,
 0.2114671733320943,0.6432501577475452,0.5835439560068902,0.43252430008231246,
 0.399007143287445,0.7290377441332229,0.7175631233390035,0.38452808838063646,
 0.3115282892668102,0.6454526925262127,0.7722317712265895,0.2581562849895678,
 0.7850950920970128,0.3889224448981289,0.0433651163935076,0.4814079681295148,
 0.12559700438723254,0.10919281892445165,0.23956625065438197,
 0.18565907395134007,0.6796220177027332,0.7006231106779279,
 0.08300582640823329,0.915674938859673,0.9610564082039905,0.28342564046323926,
 0.3715464469874892,0.5686266305747719,0.5142496619554879,0.8548023318041744,
 0.6962001822500076,0.8016524768224731,0.5042156942825867,0.10015122639901863,
 0.8845309630165121,0.7328460103571078,0.7727324168885896,0.24045531437002654,
 0.9496844703238522,0.6811031289705618,0.3763701460081236,0.21528471505703273,
 0.06647326395871034,0.22451946865691752,0.9641699731811628,
 0.5350287261774859,0.804101730105948,0.7554456753356456,0.9916462718446531,
 0.4606850374931545,0.9018816165293346,0.9727858359498941,0.9950976560533289,
 0.4051621794318455,0.18571520983580858,0.7669839209611142,
 0.04550882853742777,0.03849464050555018,0.322433657154946,0.2510170731262288,
 0.5767863468907853,0.26949667239864916,0.39975119910980517,
 0.21772812513453288,0.23201675639416075,0.16793959716494522,
 0.02144591560509368,0.6029251318684163,0.04266229900238816,
 0.36371458754638875,0.7258956045532993,0.1653089099325522,0.6717635211126172,
 0.69603037367259,0.9020126255165517,0.6082824336216416,0.7966779077056938,
 0.6386409042265143,0.9533353942501682,0.7237573549636834,0.21172473757270804,
 0.47815593695962655,0.7643379208714763,0.3422144136672073,0.3077546325941509,
 0.9166455875617361,0.782056910790196,0.39167508196054246,0.13488517472211647,
 0.42417891284890363,0.5972054853161273,0.7675132637369942,0.6941231891281041,
 0.48514554528402265,0.37775782975924976,0.1697262979836608,
 0.6073030057936135,0.9219532249630473,0.1431031747794509,0.6641945171766219,
 0.9568044521401355,0.341712202267071,0.7138854533762792,0.7684282465564851,
 0.8586490863268561,0.65062438162853,0.41814476289598923,0.8054275352990563,
 0.5933602682150175,0.3289048964361342,0.996985253078375,0.16673732796642304,
 0.009227783352201313,0.18634436282788935,0.39623140529042866,
 0.6697398578358664,0.40214938172784975,0.38834662218496807,
 0.5517150185236701,0.10568102663141188,0.5415276346703732,
 0.21788778202602765,0.3562885378832621,0.4105321722927098,0.5215019131283734,
 0.46109614206204974,0.11022591413152472,0.9509844278247671,
 0.43634616337644416,0.23666884761370877,0.0013479245711094379,
 0.12091050136855219,0.7150366207352823,0.9929026764458558,0.4523970941478994,
 0.7903080468519202,0.18666899823694616,0.8737262805226912,0.7147468469853494,
 0.5571540898278677,0.4385894919337069,0.8554387014914782,0.1123510343820644,
 0.027747158080458467,0.539296120041891,0.5264450917487034,
 0.22402449495326515,0.9013594924227187,0.07490673032095796,
 0.9606672139713808,0.6756669633800743,0.9148792537479243,0.34867961617722487,
 0.08119016911195476,0.14411472011778648,0.2658280791974936,
 0.23299952496258447,0.07951378871437653,0.09720498670554722,
 0.29536096997297634,0.9547182601317418,0.7969615138845979,0.5411513919222836,
 0.3117485592894107,0.9234878331798771,0.8889739680412655,0.9139167612731313,
 0.4430699194256267,0.3164532467632265,0.34081038490866367,
 0.04432965122315469,0.38278844655269695,0.05038512613362878,
 0.18008133848514418,0.5577330578804799,0.24956818000149494,
 0.5056761894805425,0.38726326547470646,0.3255017023447655,0.8447174516899923,
 0.3806622024788274,0.8907439009390729,0.685228250395272,0.4136553689301943,
 0.4220639133068983,0.924362402369543,0.5850968628847268,0.4261045788083432,
 0.6900169419282411,0.8499006538790633,0.934297197025977,0.26110728000201566,
 0.6280218213565222,0.6463289367820726,0.2819638545528318,0.33018904744059696,
 0.4849183212158117,0.9204161100508904,0.07009662511918735,0.8898792947144359,
 0.5275264709814054,0.9510256766938251,0.6782426087792578,0.9611654402189624,
 0.4912925122032079,0.6598728606400948,0.134046714990378,0.6394743123071395,
 0.11144220191837395,0.8999642903006057,0.722389085514696,0.8300457957115295,
 0.5496635238677424,0.7694845367015405,0.7275592266029002,0.23340259430303167,
 0.35624128713538983,0.051244363663436654,0.603015310900645,
 0.7851788192935525,0.7960867380824558,0.9872701579353415,0.2097797658136844,
 0.17990565903310574,0.09400979972758639,0.05022013142326931,
 0.27148891287160604,0.5796635264877784,0.32182475081397255,0.63096469347789,
 0.4695991841108884,0.27132983774035213,0.47259467559702595,
 0.8114448573143157,0.4597006362712598,0.304851001860199,0.565837832444532,
 0.7903311414239592,0.6812048698580365,0.7405789368386615,0.33792590091640484,
 0.41816170386587936,0.612953905730494,0.9055917443008438,0.995129514376232,
 0.8610602715453731,0.11874936252220758,0.7535880325174165,
 0.23896880482734673,0.9231382489129009,0.6181150542259451,0.7664708193912931,
 0.9507470421924933,0.7725639089630216,0.5788258341624972,0.42769971346032387,
 0.31664713703923286,0.32009780489847195,0.7959624738055076,
 0.7861077197726001,0.898667665563392,0.10594588305860642,0.08038543388452724,
 0.8035308915298255,0.17097148227196446,0.37066911133482794,
 0.38358336741419685,0.35770350757376024,0.8659706888015524,
 0.10169792950346213,0.06027382255717206,0.4046073314005355,
 0.7718054999895237,0.213587462800658,0.46028894714581425,0.17572702387172012,
 0.28234245854691475,0.5003375194511206,0.5212621999144944,0.3608881802895535,
 0.5836166450375279,0.16047342700727552,0.19628100823426142,
 0.6125912960513241,0.31380747797523034,0.4577011716575715,0.5268105400381059,
 0.9910581053497693,0.3922689462897442,0.7758118062043486,0.9289273567932118,
 0.9139475959909428,0.08604026895948536,0.308813448481156,0.6150391486593108,
 0.9770107119698485,0.17128684760576396,0.01944754509115454,
 0.3304990196217974,0.801928584937892,0.5359179902303901,0.04378203738489439,
 0.563506275844868,0.4502078740086902,0.3290891145093243,0.2632868079217964,
 0.8809883851574485,0.3724443327552488,0.0030287798394557086,
 0.3807772536656373,0.5205481069502419,0.8535402972481431,0.7341927507768216,
 0.03497351934609505,0.7080694256511719,0.5787484293444237,0.7552690036643577,
 0.5590194594142862,0.9946222582947984,0.11846758586054318,0.3289721717651216,
 0.3489339807128051,0.8294821994202113,0.1503871892282972,0.906766663111016,
 0.2892884394031161,0.194441638263254,0.9163492827274078,0.8572404830147948,
 0.978006818726135,0.1086960127752209,0.21053089721936513,0.5074068541925677,
 0.8730058079297864,0.7664913634494148,0.3195814700339359,0.6253748619555383,
 0.5902865014342651,0.3349725297789038,0.5459271130903018,0.7595977102725345,
 0.9983067316136096,0.6351521434289361,0.5650994148007993,0.5403583651641015,
 0.2733379341284956,0.22514389645360688,0.974648651678149,0.24874076351377383,
 0.4861652257904432,0.11730393900286595,0.7844940908093303,
 0.42279919353602546,0.9643424679140354,0.6364344136417565,0.566708438504063,
 0.9288620800901816,0.06662637771972302,0.5731739562105561,
 0.22754826161119368,0.9213511740749258,0.6136510588904598,0.1112994984638701,
 0.5222001618300177,0.2368044521798638,0.06859890891201859,0.8754726059662308,
 0.7685087691534902,0.9637485161741948,0.4913730566041117,0.5827412100466065,
 0.9442595681716826,0.7365022930957482,0.6293109275886954,0.5012256310436703,
 0.19683724687441084,0.7962226821977205,0.1824672347715054,0.577571206765275,
 0.27775829850323785,0.8591889406945645,0.3982887406810709,0.9435094270018002,
 0.4275744006775154,0.11939052431170016,0.7832039583298593,
 0.41826435816096186,0.24016811881871702,0.9845585727476551,
 0.13903684081965673,0.5574754785891125,0.7539033133785324,0.7232687245681979,
 0.6922619181501352,0.13405171151550788,0.4368244977270147,0.8036226683866697,
 0.5636222772831021,0.058323164579235476,0.4052869108876258,0.694898700148315,
 0.589852839771489,0.14689275603786867,0.3068262229804758,0.23062827004075892,
 0.12013653492787257,0.10729870146981679,0.35508701096801454,
 0.3144944297651757,0.7523450677765293,0.032850038134730086,
 0.2637989729964123,0.16147795689215338,0.7090694406273569,0.459173678505007,
 0.08274844516382462,0.23335654043904908,0.305035266693794,
 0.07747762128681412,0.7328905508928957,0.8901898335818037,0.956009399550753,
 0.94407601160566,0.33745878395415574,0.42822296522636927,0.9709100334089792,
 0.8393318979209842,0.9265961348106311,0.2479875673810399,0.4136037611405934,
 0.6352591672733698,0.9769305946128197,0.9876863701460578,0.07576237924989782,
 0.953755065225708,0.8264944553823868,0.9647132052717554,0.13172631470965812,
 0.18012313511213618,0.2853532227639005,0.7327327027731777,0.6715314115869774,
 0.7424686456132463,0.8225999841787036,0.7454611674113625,0.19011513152859383,
 0.369994703811386,0.9211970799907022,0.5100675153242693,0.9000694019573159,
 0.35606621131913874,0.6009902743945942,0.509002769482443,0.41481137801275125,
 0.8149603739201016,0.5356499073000093,0.3200988183202371,0.8683305227724163,
 0.022005880123117882,0.3697018976333991,0.8509344631185132,
 0.3497125091937292,0.6670852926533425,0.36886618996123033,
 0.13322755293903188,0.6009589665131712,0.7586894196070065,0.6651673133896702,
 0.13684334682705068,0.9597223836083709,0.07070581618338412,
 0.1145053118961975,0.2071011959318947,0.7161456247006309,0.8775343676241093,
 0.640473226943071,0.28802841133635226,0.12578869493323686,0.8385903651335016,
 0.3252107732235823,0.24421577660429117,0.0467427862035138,
 0.21677610759343469,0.8032842262206206,0.9259292511400468,
 0.004914438252754838,0.22280728305373743,0.08799897115865529,
 0.30877629258369255,0.3638787002398264,0.8643772154014145,
 0.23360708568213306,0.8684592167880163,0.4043748995313413,0.7753263956125899,
 0.8845170103037987,0.5688136853284103,0.5449530436449322,0.16316238219773305,
 0.8647514198566103,0.5372617674732002,0.5060877802425243,0.1501314847856382,
 0.2696960967882275,0.5881081879789869,0.014359589172502352,
 0.4291513612905169,0.8277847358140037,0.49760622586148673,0.7959336983190842,
 0.3377466885232121,0.7408161455601725,0.6473627700317885,0.4697786686805746,
 0.7195123196008971,0.7512145500275466,0.09606690821336084,
 0.06771270621061165,0.25734328547648266,0.11332576435152786,
 0.7369680576361987,0.00612791192864115,0.3110915634592203,0.4438193241099774,
 0.569596518209655,0.995544182325431,0.4751366582451326,0.010549451577312352,
 0.16261611685070443,0.42672790615026723,0.4562404715573438,
 0.4059895738552426,0.23873418240764488,0.4472148350861286,0.917754401997553,
 0.8857649326915641,0.5114979227992881,0.015711714382151354,
 0.04379120039551476,0.19138574424877886,0.7901078527818816,
 0.9444339534799878,0.7332563030280783,0.23837660858048426,0.2067706279980005,
 0.8484673776436709,0.6362192525387694,0.429299070423086,0.7813365888046119,
 0.39393778225217047,0.19614481708243747,0.01874569472337051,
 0.6641938872305492,0.8814231395906464,0.35535632182313515,0.2839111006740388,
 0.30416033381372365,0.5900862829387593,0.011348142287153196,
 0.787406389352858,0.23475650570936002,0.7945679840044713,0.7866430716258184,
 0.07275624682295512,0.23098897521688144,0.2556661445746685,
 0.5061157378015647,0.04586427077151389,0.3999580337288573,0.919304611967103,
 0.8056316320911467,0.9298957436854742,0.13316988200549273,
 0.20220889117008023,0.7514669552552892,0.35964713503229606,
 0.42856051168425735,0.9049524960792866,0.6561219027420199,0.876557229582846,
 0.8074893553183287,0.8072437207105642,0.5808556257795543,0.4899800659517205,
 0.3177363502375965,0.1581981608085481,0.018127774801380725,
 0.6985551384054264,0.6168863288916685,0.9689501648511014,0.9109621323299771,
 0.0063938844579819465,0.5518785722094801,0.3093698791143785,
 0.4611014840312939,0.2552083704283079,0.7947728232430553,0.31168104779230466,
 0.5940630526883542,0.1856096540147557,0.1521938047296464,0.6893797654943974,
 0.0789289414759885,0.2346206335527523,0.0330436244578074,0.3328396322058932,
 0.44333481629032223,0.18418304343145586,0.15205337563010612,
 0.060933932725339424,0.3400296151180635,0.4679160690575479,
 0.9735476227878607,0.33246359417867677,0.8943546254860917,0.9260452000954402,
 0.17682037819803775,0.07816011183061855,0.7213480857875137,
 0.07486063389479913,0.1519913139290805,0.45417381002397184,
 0.7208348940523375,0.025587577632636305,0.06153017428529972,
 0.29037588713011286,0.2461210401218219,0.5351496274688576,0.1895241347826373,
 0.2518913702577913,0.7523225056597943,0.9533468551797966,0.39187242300178915,
 0.5242272764616897,0.21122839857200515,0.044063541303780984,
 0.5297728295516226,0.36433340245168266,0.8977166264357853,
 0.19150218651103412,0.6193485243762713,0.46234814741433294,
 0.8179772070111992,0.6969549451139687,0.46534322005352147,0.354906551720799,
 0.40735658024820054,0.3830297795797868,0.680382122231385,0.3440926208986119,
 0.9044547284451776,0.5666762447858266,0.4059704268766404,0.2928856950419666,
 0.5096177301090025,0.18685912811509497,0.32566535264532614,
 0.7488938391579304,0.6961694849504809,0.2224360387930917,0.6528323171482946,
 0.11554297936723579,0.6411292842735586,0.46486116950905043,
 0.9798115496359783,0.01107840962162654,0.7229250616776212,0.7360013000601896,
 0.9952984571856388,0.9775683805203108,0.12471432725695708,
 0.06468239452148761,0.9911387453436973,0.6385416580272293,0.3090451185779858,
 0.5507055791942363,0.33997418406472746,0.5423843891322473,
 0.010092471619879917,0.3310971856987994,0.9955558118686537,
 0.0472108507804061,0.19267002244836384,0.7402835556233581,
 0.38622782645342013,0.7383733873977003,0.2763429608242647,0.3172486856103124,
 0.26915656130574017,0.053241760649490155,0.6007749549217022,
 0.49860843177477554,0.21902303976524717,0.8511779277787601,
 0.9408107171288804,0.10242182508880915,0.18055907912091895,
 0.1951706218897532,0.04513857091669804,0.9844143125097796,0.294000142013518,
 0.9896220133587903,0.3075014502766189,0.3203926699624542,0.40116881384537373,
 0.16169240125384743,0.1714130278112036,0.18319440364879502,
 0.9844099310221528,0.8516909331847977,0.3718554156533782,0.7152173537828442,
 0.5834333617249223,0.3566062622840347,0.9721788246647435,0.4234621695250119,
 0.2191476521576452,0.8088006987484009,0.9988868535404888,0.33155371860022065,
 0.054377642499707446,0.24345858907969364,0.3629905379740861,
 0.2138473609081356,0.7590561870649477,0.6618496343818726,0.2555874399185081,
 0.8544317960182259,0.9192050157086094,0.11419558458594303,
 0.06080361152373903,0.6264279820535745,0.6007393356380933,0.3666300252081689,
 0.24514772600030965,0.9738490795198204,0.9117792273403973,0.6643301358540673,
 0.6090797771594589,0.679263251425079,0.7579672553494995,0.36654928014135646,
 0.41227514984443625,0.2617254603735061,0.08156988192462211,
 0.4977925639971005,0.0691773225753276,0.928899664664367,0.2193429046169365,
 0.32502354746568385,0.27700902211436707,0.9609019723283495,
 0.9891500480578026,0.016044737280607735,0.4320568194603035,0.93930496687956,
 0.6563144271640502,0.9694669322499179,0.2448296818957404,
 3.767085732421238e-4,0.46382671466664327,0.5958597015554881,
 0.6902674595935286,0.6059169441187094,0.15106204476043916,0.6149838959887665,
 0.2598294803701928,0.05695521829833439,0.693932940429528,0.5863361168214885,
 0.023686327784104066,0.3855761534934802,0.944027329229776,
 0.44967326035286925,0.35759141565673824,0.28516356983694624,
 0.46483450519661396,0.24812570297530712,0.8640770269201301,
 0.40180546361698366,0.89268896063714,0.6046803924406844,0.5943362813360783,
 0.25922275212496304,0.22346687917737473,0.8131984896348341,
 0.48012552115860396,0.18235406032017443,0.4117207069619242,
 0.5590842419040533,0.11320165425809647,0.327074886652999,0.2627253878624345,
 0.13983586000322334,0.3827187022726194,0.8715042755561758,0.8594758727573278,
 0.07278262517737288,0.29460561351282766,0.11393021175308471,
 0.04870889292958047,0.46793648536025545,0.06217561441705821,
 0.036613167423537596,0.350772731400973,0.7505330997956267,0.3450847143309701,
 0.43824733368523605,0.21355084715100592,0.6725306102480972,
 0.8908019887443298,0.009047689794112745,0.8983859196282408,
 0.5066341127329659,0.644124173889845,0.554815099274222,0.5956427613104809,
 0.37396729722835576,0.7837985109490959,0.7291238206115327,0.777009665891211,
 0.7389532605670087,0.9386387258850175,0.6333242020799474,0.33792245787176534,
 0.6917323748171119,0.5197230625042387,0.9505715540624655,0.877316963179131,
 0.6826867476261532,0.6524857734629766,0.44443670093279086,
 0.04995993988352643,0.5167368674876721,0.291197951400179,0.6776577398771745,
 0.6182118206883115,0.0047906780636778334,0.9605390309912746,
 0.9374367211069136,0.3968472087336663,0.47388007864688797,
 0.18608030343101278,0.7312020117618433,0.9079748383891573,0.1448262021522655,
 0.9049287824077239,0.030792065912558342,0.23095415995528468,
 0.3692704936387202,0.9294930844975089,0.6497043934892479,0.4295538239254518,
 0.42487014183182814,0.7418179756180318,0.6316763864716672,0.2846384443889185,
 0.20542061091850727,0.5501678211268808,0.8425230268678909,0.7660857406242184,
 0.02013188218062134,0.6563678077942718,0.6193193548996744,0.7345898069292309,
 0.9427765479922268,0.046162104548303606,0.6782015477143509,
 0.2172761220931192,0.41538678814975083,0.31098707915557633,
 0.3567231526864991,0.8707838968556187,0.6329015976696543,0.10560431685631877,
 0.8613135374989045,0.11084341066931114,0.7575026779029664,
 0.15689746530212445,0.8108469264940363,0.8443989526183583,0.6972050308476154,
 0.25169352209501095,0.8555285422395413,0.3143043509477609,
 0.12105071047985827,0.7923629354132371,0.4139987563869876,0.5460162360942353,
 0.02051503735783433,0.029477224874973063,0.533780979201646,
 0.01983854411861541,0.8490182314497727,0.7667655034143495,0.3104706420848282,
 0.6489310367476555,0.188188517723351,0.33707346174913067,0.26229756082625,
 0.02039725166849904,0.14998340605353044,0.8765004253020495,
 0.12870023333353542,0.9258814879626971,0.02697854348768791,
 0.5048265660802381,0.2167519171486696,0.5314192865435108,0.05059528256584844,
 0.6629954566852201,0.0347387696985052,0.04496614354210404,0.6706468573849511,
 0.22841847781367475,0.9102241498421391,0.47411024767051846,
 0.9267255748022113,0.18838561594304026,0.47004096188069,0.1867343416152708,
 0.06116005925816281,0.35905616260294837,0.6592342595768628,
 0.7994921708796037,0.798114868064184,0.2224239152699431,0.7065458548256608,
 0.46613040072836176,0.76871883006174,0.2586126901794982,0.5394403095227143,
 0.09197255381280989,0.2201184670840205,0.19714550651868168,
 0.26459875563958435,0.1793304611077038,0.7622121982753122,
 0.13172893305863975,0.6334412613916842,0.9634971306507584,0.5242403797199116,
 0.6632600238971459,0.1144271726430035,0.9993218680384419,0.7456969025231462,
 0.817801071444419,0.9043851786765144,0.11884127833055635,0.6213555163655424,
 0.04915990158484829,0.7339029638822647,0.42053191419718206,
 0.3060801269033496,0.5506157565941715,0.5505897018365451,0.5763539130138944,
 0.49811983264742254,0.36748941194541984,0.2682208715380907,
 0.8528956338706428,0.022988877648898565,0.3507143327271589,
 0.8553256789878424,0.8064884432473363,0.9357729723138039,0.7643916787049063,
 0.6866127505636266,0.9441348053715354,0.07038954640671391,
 0.06359904147507467,0.3246352315501503,0.2729613938671751,
 0.30938544957313796,0.7390453245966321,0.43917389437361065,
 0.14264279202313013,0.91609518675418,0.3482364754883033,0.17020678987424387,
 0.304179987844608,0.18153487912219624,0.803582514222761,0.7832171011586094,
 0.08768967654612081,0.29605051839724505,0.08795093135321552,
 0.377358342961594,0.9695388360552921,0.0029458278066369914,
 0.8450661383557148,0.13527856907097424,0.04896506435993131,
 0.7476363984325064,0.1115214901146041,0.3093955785804646,
 0.027966400283476855,0.5944611007954618,0.6864942753845283,
 0.4796157591070678,0.2693861129192223,0.906720671255087,0.9473103545749122,
 0.8398013522640019,0.18081221628118105,0.6509077272140891,
 0.008728735748257499,0.7839119781871573,0.3277925962578274,
 0.6559543329146775,0.625561490883417,0.45619830200998435,0.11705001674789972,
 0.41619675362545616,0.005295839152911541,0.09849456481341645,
 0.052847671510995675,0.5508143209405846,0.36763744694959666,
 0.2593979752879032,0.5601429650668395,0.7046771378308051,0.8925534757390791,
 0.8619188285219832,0.540352904868722,0.4271845375968224,0.9437266938544732,
 0.9092998496440976,0.3095087749510782,0.061377590832426,0.6209168277778545,
 0.021165924253256363,0.8760183670820471,0.06815298974637995,
 0.4903118788149914,0.08660775943742427,0.31519413056995815,
 0.9190460124391628,0.20688020698064924,0.07803739200253434,
 0.13053870501320342,0.920387422606417,0.06915997269252472,0.63126586697061,
 0.48112551065494724,0.13480464994414976,0.885828331468448,0.5811577625789902,
 0.7162818919379521,0.36405966790594335,0.4828150817632042,
 0.016835957338880103,0.21049080553513422,0.6602973684354163,
 0.6296040735920272,0.46991746850964367,0.6365457352086265,
 0.19292428678019058,0.9011625037993428,0.022080984121888747,
 0.225528043524998,0.580195079882328,0.4938581669707858,0.7140064806007301,
 0.08632311163006123,0.31825029342371225,0.11010530876177071,
 0.4662041765778808,0.21288971235636933,0.574018334196738,0.7212097660453912,
 0.7406203467471011,0.873636745082661,0.7529273573068426,0.15026231176920313,
 0.21236958462811695,0.1919918529613719,0.37374258265022076,
 0.032675552593042445,0.0053440053591474745,0.6599108435715272,
 0.4415279013685609,0.7522248619434285,0.08875390129258764,
 0.12348502109955484,0.6346943633249353,0.9167429873493855,0.5869837862062932,
 0.8128437609415933,0.6625565420694908,0.05462257620576616,
 0.05164307896075693,0.6235212466568344,0.7569383724939189,0.3376942440375774,
 0.5793523969072459,0.3113915458628811,0.8344344085198742,0.3233029175692933,
 0.2567071470011134,0.6372894030474002,0.3125132304575474,0.1729619572526212,
 0.7428918523372274,0.8837930022160707,0.3111083699813706,0.741543660315685,
 0.7212729673423216,0.9408429002011207,0.015312729161895411,
 0.6858322692054812,0.8607843891174455,0.8044338899818784,0.39059474754298673,
 0.010888346106304736,0.9956734814519254,0.14418103062700072,
 0.04048692405252896,0.03403314966530857,0.4661167314174347,
 0.6981395356179639,0.35971790543931825,0.13042353106860216,
 0.7832854785488337,0.43835688339840784,0.5339995569565779,0.1305394010821488,
 0.8848001940088834,0.9402609619366741,0.4546543488808532,0.5560667680298899,
 0.23925889919967236,0.5077168614787094,0.27392474183037685,
 0.9972588057563391,0.39642274399015964,0.13711571039626602,
 0.09629905234885394,0.5252586048584182,0.4780063437110754,0.4999376323444933,
 0.7901588855292098,0.39337797053204815,0.2554524651378629,
 0.10171492399502813,0.6679435684763054,0.5408175472534268,0.8292753461075985,
 0.7653848924208289,0.6515427875290238,0.8333400549276477,0.4550066051115722,
 0.6538725704089656,0.7479838386042889,0.615443755150599,0.5614245857070708,
 0.6436008742825253,0.1662233391170933,0.7020754212762179,0.5114103416346453,
 0.9213012329683746,0.5086313152485883,0.13504086345139688,
 0.19232170239642754,0.7160180350937841,0.40275766578726957,
 0.48234075404481613,0.8375966602794528,0.13445107124703626,
 0.8666803987690855,0.43403405096576675,0.10810339866793561,
 0.5322108923599559,0.9224469545491979,0.05828497685234568,
 0.38287385855736533,0.6334751301564433,0.7524399263425068,0.6919676448688303,
 0.3921972496877987,0.7260670081545257,0.42749896399142584,0.0670096612390394,
 0.6046842863269447,0.8138494448322323,0.8673799208639185,0.5437824854146579,
 0.6068620904066063,0.942693188209396,0.2767079089800554,0.2400473400965789,
 0.6110285244067695,0.405411566096737,0.5754165663182425,0.08314054714705121,
 0.5125710608800957,0.9465273487283077,0.9644441618038344,0.9009883760930261,
 0.8880402348111396,0.07781671287759573,0.19830936584413617,
 0.24181912014536588,0.5399451904296063,0.32304356663760303,
 0.7272139367214803,0.3824633075989411,0.7939098546765126,0.44517712737125725,
 0.8625410663503652,0.623625179984922,0.042324534423924476,0.0892774037688977,
 0.0011365418619714296,0.02734453417241578,0.7180214852677214,
 0.6491100439609995,0.7552134854126866,0.936466557389237,0.7881642488359346,
 0.07995663770865258,0.7897039799472861,0.6957519530587897,
 0.22822422272464316,0.8659139882930593,0.1485047048428263,
 0.13278895033707516,0.05063505444127392,0.8100204675972833,
 0.021010820285108367,0.9715586098446685,0.13555051795609663,
 0.1523026322051253,0.750690875460011,0.9503334687496127,0.4725711620971331,
 0.17922766629447895,0.22398615083381002,0.8986450796960789,
 0.026552395540057172,0.35494555535486816,0.5267562205824232,
 0.46901517312458896,0.5414060877690163,0.386781109534839,0.1462558260965826,
 0.7044876030062248,0.3329727700270053,0.5473913466397055,0.08135302788487919,
 0.4876327069874091,0.4433193800104266,0.4189972639815074,0.5229551620216424,
 0.3996236929646918,0.5535171620714123,0.4517713290068829,0.9506255487753638,
 0.3135900310586428,0.4418655637714817,0.35876171929409884,
 0.17243850504791602,0.10246429130511725,0.06577185597493407,
 0.038508764929907624,0.7574679720680483,0.9438585894205286,
 0.9461044604600801,0.9118225891890147,0.8018003174368846,
 0.024948723250879157,0.8674579988543493,0.5699894138069946,
 0.7357717575993704,0.4509071130283462,0.14377246854214665,0.8914275929369297,
 0.4421257711593669,0.07768228488667961,0.16974091565176974,
 0.7330923482655985,0.9761712183226319,0.937945085306233,0.4477673599841143,
 0.810147248700078,0.8465829277423069,0.06734620157758298,0.5674170426824194,
 0.14644883018568078,0.31652589774582984,0.46603921873082454,
 0.3493966106751366,0.14258346478837292,0.7698489809515202,
 0.49497490482888007,0.7631614080676709,0.5350004782418699,0.7777628271540759,
 0.7275560512972835,0.4721567234940296,0.47649486222499027,
 0.05712395372133716,0.2905658330350025,0.2475845698073399,0.7941497347153088,
 0.1948568273960537,0.6865011509626017,0.6331868821304824,0.7558502631823641,
 0.748761785895476,0.431882492787932,0.08197123417728203,0.7959640454661387,
 0.9260305498479664,0.6294909654717954,0.2573410105224304,0.8892831792386049,
 0.6352462118658606,0.6427268808045551,0.037445271199741725,
 0.7269127879724305,0.7271651921198018,0.5037824883262536,0.29733023371348244,
 0.7643494609224828,0.5317710951128136,0.0033015106624653123,
 0.3275552681245696,0.10810361888062625,0.9982517250659383,0.3823930156692231,
 0.8582017193270461,0.8883828897933246,0.2592941642978913,0.7764458259952554,
 0.5296334273300531,0.8337484288943614,0.977597940789439,0.5283225280472268,
 0.8093129890348499,0.4767938154849942,0.5021531439699596,0.805222550703312,
 0.2556627723844108,0.021110345388493545,0.37915670698115855,
 0.6186643361646992,0.22927482806692068,0.42777400273084454,
 0.49304783411890063,0.7409947183023973,0.10616208657869786,
 0.7640014547418397,0.8900428943724792,0.5744621019205081,0.3246675147681647,
 0.4946802935922918,0.19026990770754126,0.5260070536286947,
 0.48644087385063495,0.4930529652248258,0.44772546735806196,
 0.17114199127536944,0.2284744060030207,0.030398954318576088,
 0.0389563456268015,0.7215315409520193,0.4024332667804651,0.4268303529976063,
 0.031402704801280046,0.17660548341564153,0.4811892054800685,
 0.6860456855832592,0.1827364425084529,0.811537922610894,0.6864448616699137,
 0.10596834862254745,0.5992939859124526,0.9545873665546614,
 0.07382063356224389,0.7342637490862651,0.732882635682222,0.2853859194361793,
 0.22589690320015487,0.2730993852708896,0.048038890092739095,
 0.45223714668169046,0.4130130411756835,0.16675858947995026,
 0.5641093393489268,0.027125184642799893,0.4952977728770055,
 0.18933532664409092,0.5357745706889999,0.3916820670013935,0.9414929332855524,
 0.0728023768366447,0.785942233620352,0.14559797348201498,0.8711397784831916,
 0.7877925680401764,0.29212426205822206,0.5017619805524927,0.4353736229192051,
 0.2177608248532309,0.12923166705521405,0.275597162129235,0.8833094848293891,
 0.32334206609260396,0.9780305527611077,0.16599223282209608,
 0.36188836490145715,0.47142696807926654,0.11533737218021956,
 0.10530105393430389,0.5827841216313531,0.7268226534876308,
 0.21347714634487813,0.9910215482351716,0.3863491386505612,
 0.27601494956735195,0.11642424135699903,0.35290101326973444,
 0.13384110309556407,0.5249213241433414,0.8958137791313188,0.388487975375153,
 0.8209800320046701,0.9425720573434035,0.24937076871180752,
 0.14717175337697963,0.9621734064761854,0.2074537915399155,0.6492558014756175,
 0.6524983144183695,0.8858617734373907,0.5683935675583298,0.48516069842300436,
 0.18438957989327143,0.6186742701067743,0.7631529143908934,0.9666109688200837,
 0.7189506387502669,0.7535605120089753,0.8747601701172443,0.26625866350067207,
 0.11131401005378139,0.8383771968857551,0.1275202744787598,
 0.11028831189063082,0.7352717864772715,0.9493945994659425,
 0.003548415841106878,0.2230196762961253,0.07815092747143892,
 0.826156149244049,0.9274852219542927,0.1918499009783705,0.17452481563629563,
 0.10648337061013935,0.014190362263285294,0.3091401748475143,
 0.849113390867304,0.4892552935225519,0.1761735310972965,0.015599047614117456,
 0.23781592395017137,0.9040818752193205,0.8671961193646839,0.7200153115353787,
 0.4771031696296364,0.20054173921028706,0.09286258647687307,
 0.4953133022592634,0.9694654560133148,0.4713390119361348,0.380509844384627,
 0.34148811740182583,0.23721934683428114,0.6269973239798987,0.998987473109293,
 0.8819094174041281,0.28049862220793365,0.3962570449747532,
 0.44948795747005144,0.5211636101420725,0.6617056740376381,0.7699378128214169,
 0.5521074780991478,0.05907920188875926,0.7987800048758787,0.6519926851320972,
 0.5429942629079388,0.7136108316462714,0.29471588047050157,0.644321857950482,
 0.46989769558126016,0.13243783246270846,0.9577868046871075,
 0.8571251795583735,0.09716689716717752,0.2010732591580674,
 0.15947307155843127,0.7121036461872055,0.5977315784609752,0.613066891239471,
 0.24870301423270935,0.5367617941763423,0.4187139017420267,0.6213990131601868,
 0.896590371641391,0.24668707150653757,0.8742429965932812,0.5997495977870749,
 0.9068381230478167,0.3914430983165097,0.39172449697277767,0.7019580068347883,
 0.11023430424003156,0.46243175782883417,0.9931769022920874,
 0.9989957431679324,0.33646168703880075,0.009082515971725935,
 0.7082190315611353,0.06708672759784984,0.7791601848604061,0.6049121925222665,
 0.3355929759318875,0.2793091786224835,0.5382004970487604,0.2733256106465123,
 0.7634960424631365,0.932466724305733,0.08990141089640424,0.5767776927833511,
 0.5901716988093674,0.6041377382034531,0.12903905986574693,
 0.10076836664912558,0.333191341167707,0.47267804297736715,
 0.30639098227629047,0.9211236931595429,0.0027329067126596884,
 0.2727151965182233,0.7312806155007155,0.5718649602077461,0.5194130241166048,
 0.39600288300175657,0.27220940079841827,0.2770383727137564,
 0.31473238559306893,0.9143636197295452,0.022549972108163008,
 0.7387882771436822,0.44358440849233727,0.39233825291951374,
 0.5830466983741566,0.03494577511815056,0.5744096553502879,
 0.27202434012509025,0.8569641646952217,0.3809869995180599,0.954364835685465,
 0.41848464969301746,0.8034161350241903,0.2571872064383881,0.4822364592494258,
 0.7753569325149976,0.938989501379731,0.5055524418651625,0.4213834265620763,
 0.5444264807872183,0.44382594273052356,0.8590457913123069,0.3475755564494126,
 0.544105387338028,0.20759265047639985,0.06150945423290144,
 0.31482002794789576,0.7918296894161678,0.2019831801315446,
 0.21410360838622988,0.5132714687773953,0.8273513713860929,0.6960610082867557,
 0.07201274745779329,0.08177821734187796,0.3553294182753566,
 0.14473515172011708,0.7065610108730431,0.40794728159775806,
 0.7699873695771882,0.6166101612693179,0.24398605541106666,0.3782215912693796,
 0.6738596852209546,0.8217016561161858,0.6910545145814633,0.673080387831768,
 0.7685729921775619,0.281425707941686,0.9574818579858915,0.5355309527388541,
 0.9042028974552299,0.011995786458355528,0.4670573597017371,
 0.5819405037714114,0.6555378962580087,0.9610552597132679,0.9274374024169401,
 0.7905322805466294,0.3836449223016629,0.7964577358149033,0.8915170714655892,
 0.9172172483513843,0.06825315075789185,0.06985830357170464,0.786664897674022,
 0.617323695632122,0.5029563491651775,0.7480814443861095,0.39316047931417475,
 0.22065966694672212,0.198458163964366,0.917026089429532,0.19870372268444592,
 0.04667991923942849,0.4537074351874584,0.45871476329799954,0.613006876823199,
 0.4449860730589289,0.1198230770121399,0.6647113168786905,0.9908543325895967,
 0.11512718568481217,0.7611006876665306,0.6068034437156647,
 0.36808795729181065,0.7394522554108897,0.06452504386877655,
 0.7879818878016247,0.8741294991787049,0.4371573927840433,0.14948772082942297,
 0.24677692781250415,0.7624884706163132,0.2219635989568387,
 0.18686284416689058,0.8083148537923088,0.23111606024784603,
 0.7865590584094491,0.56700410164891,0.1697985575580815,0.6965680647339942,
 0.9985399437706143,0.20956141857577815,0.6506295712190263,
 0.0037228088027255746,0.471034997157185,0.2864531284692603,
 0.8131214322116675,0.7665101088551823,0.565282404447554,0.720080778826505,
 0.44203326486854666,0.618689904042438,0.2912840939011735,0.5948738635623454,
 0.5723882760307816,0.8496353626696543,0.44745516589840517,0.6322145919888971,
 0.7572359937859351,0.2623430583686409,0.6613024397736952,0.08442611945513734,
 0.9039230066722149,0.5116553223942673,0.8892236236947306,0.17827350484369942,
 0.9117351607705737,0.20039119743990452,0.9842609029578486,0.5048719422513362,
 0.7415238762236137,0.9728274105403022,0.8757982369040912,0.5520646083505412,
 0.5197258660161286,0.6481897237264742,0.12359213167635774,0.1953809270656397,
 0.8978490547252461,0.373366198773863,0.6360008129106604,0.9027911438963854,
 0.681066914864495,0.6829492331173368,0.987377353105148,0.10049671078948841,
 0.7693768938515054,0.630858095912619,0.12452689827040264,0.5789600874209031,
 0.8457050532804826,0.888132946140989,0.49402992843365545,0.8910961888928962,
 0.3191875179282324,0.2921600787206823,0.24422396540893188,0.1349934070032992,
 0.36095909368913837,0.5158632473981184,0.6917005422662525,0.508624423929573,
 0.6970091002201412,0.3410715340359083,0.23497309073378891,0.8753151926961719,
 0.3711438546500073,0.17304400926488495,0.8690500508763496,
 0.24845368978017135,0.930790221318376,0.09237193427090029,0.7887061013339712,
 0.2944903751584559,0.11702723295359962,0.02419547182198656,
 0.4707152587761315,0.09342581960150687,0.0888315252247649,
 0.41570063024282866,0.7093091099125253,0.7552262109085165,0.6154996902754207,
 0.04484321602842578,0.5710783359094075,0.5520351665333656,0.1497924932138478,
 0.43879616318950043,0.9138859485049597,0.8647755095231918,
 0.07758652304577585,0.29090450395942413,0.5327480887661619,
 0.8696467896871263,0.5194187907841177,0.14203296539769805,
 0.27219958678811407,0.9419374619378308,0.26926219815228836,
 0.044770318233159134,0.7514286318242729,0.84350813737713,
 0.011006041634496144,0.7400231611488955,0.31713756904389057,
 0.5587211455570411,0.40242506499462805,0.6131896504393092,0.2349410415117198,
 0.29996195732925546,0.5262482454254435,0.8391813820244606,
 0.16778391750859312,0.10289339972561629,0.3265637475300699,
 0.2393269562727749,0.9332709413783356,0.22804097583245875,
 0.14094251518707201,0.47762618918055244,0.37393943536657737,
 0.027352490831351384,0.7827599170194491,0.3905090067605502,
 0.24685948884572972,0.02509848095953393,0.26029266453862066,
 0.30077287265446107,0.30110445021351473,0.18759880982144017,
 0.2532183525821531,0.6924448617460324,0.37205413519381125,0.8395941012946411,
 0.08234571114809308,0.47215743018405487,0.9437748841962817,
 0.30769232284665926,0.4337671890588455,0.07231674017071665,
 0.5301831915001972,0.7163872563480269,0.7839295239746903,0.8601206512481192,
 0.7170581022500572,0.5504729677371045,0.015568395024964765,
 0.5197640322659247,0.6144908935275265,0.3283151425317792,0.5086392071864124,
 0.7684138178304467,0.979435059637229,0.943235491572476,0.8890202064298053,
 0.24309890290928882,0.377590211321036,0.11157754322718261,0.1552854185599335,
 0.6573472120703484,0.6218254937778968,0.8626263462274303,0.2550441524649838,
 0.15123224411061864,0.1670352087402911,0.9100446701432557,
 0.0030534896875602513,0.7623270284094387,0.8292218115907494,
 0.10446239588179884,0.14985338675218018,0.3595205471125168,
 0.7286031931128011,0.3186282739801716,0.9002652814208413,0.06380070354623635,
 0.03667018853128212,0.7266573135368745,0.03805371982921557,
 0.07819897473166526,0.2743733839167646,0.7099151883299507,
 0.11462440390697481,0.23603140425209956,0.3065457027415779,0.445964082641661,
 0.9534828345806128,0.31120327048096597,0.9805920278873796,0.7518578758468281,
 0.9755981172520455,0.9540985241282139,0.3809993772490029,0.34249943751571377,
 0.10900155377474507,0.08164983109258417,0.19786669758577635,
 0.25598209466698985,0.18155518173797092,0.9203723709757683,
 0.12400607197034874,0.25284316732861767,0.09337647754969736,
 0.5788494676406063,0.30510943615606334,0.3381613088193718,
 0.017966336978163522,0.36382249094935437,0.6540908203844162,
 0.06530814283502384,0.9279658875353809,0.9574997172290894,0.5738381615987695,
 0.9886446491721609,0.9102771005527514,0.9828270111087324,0.6038068933255525,
 0.028767690387144107,0.8315009315617108,0.49364618638898083,
 0.9393256924381863,0.4409059026300859,0.3629700095311609,0.6420029417322427,
 0.376893463977111,0.8211231528689911,0.6176096371794895,0.3246911478761816,
 0.5596384498752581,0.6792081621993737,0.5497802267417436,0.09650249841860858,
 0.8781263223681192,0.6856308294706401,0.2502543208439604,0.11244845362262301,
 0.07259027598522838,0.3989130427247509,0.8097290978061408,0.4329171076870205,
 0.5127006472245544,0.8083297072735415,0.05535235687874274,0.9673864553625836,
 0.36532263721819147,0.35190909338341014,0.32689652231892985,
 0.11372420012943674,0.7946107947958156,0.20257120826979036,
 0.20711818707814889,0.45183851155219434,0.9123429074020755,
 0.5302991380972223,0.746698936925071,0.14870396277951714,0.6746024367563876,
 0.9603690342406506,0.038665230373001624,0.16317113925875437,
 0.8624326283930358,0.6424942944846712,0.9954970277572164,0.9124202301492164,
 0.9658219589272462,0.10809110097748542,0.15026647354742462,
 0.6489147595585456,0.7048780487691567,0.8637846254514165,0.8600172986998802,
 0.6254780239234068,0.5145327441553478,0.2858953314284438,0.2091764872001859,
 0.2886107530778095,0.22190965502811066,0.3243685689961635,0.6835661747480044,
 0.012072712755999748,0.09976800105067762,0.5469288831960661,
 0.8643559919107014,0.642586425493135,0.6274560616207386,0.8421346919282995,
 0.8448263891065371,0.8528080116912207,0.40896621573768654,0.2131678620739451,
 0.30928798028381643,0.9331989183023025,0.3798171927986602,
 0.25755553912429496,0.13043600576849235,0.32197531231477816,
 0.9154303656515275,0.4403109849383887,0.8632822331447637,0.3975139276808307,
 0.2645018037297887,0.5560864439144334,0.043721406368409155,
 0.39125844570641233,0.31852128585245654,0.5709695847478364,
 0.7674258243594151,0.17073612854849407,0.9525985593081747,0.4744605941527922,
 0.5985164336665648,0.2067207453318981,0.9907449329979122,0.9131608905359971,
 0.9398401778826437,0.17410620853362657,0.31152277143628493,
 0.08077706566148213,0.4521803402336757,0.3266283799414792,
 0.20573230073753623,0.1062878330831517,0.546843343326212,0.12150443614815776,
 0.009360563132865596,0.5819077223035473,0.18562532878384908,
 0.7247634048308302,0.48457789490170844,0.8023819183636479,
 0.07716485832357767,0.04246812913778342,0.41888925513019815,
 0.6839542142917165,0.5087189810117045,0.13274829088614237,
 0.35701022415143735,0.06581070050656601,0.18418842622881304,
 0.6066652494153443,0.37328618463468244,0.6916364194496522,
 0.24179961149373652,0.5002120233775618,0.4495846620760312,0.5150566665719765,
 0.8442974739496252,0.7385020797857277,0.38170824196958797,0.4992421606066638,
 0.5088282074185577,0.17424834732360184,0.2311185856126805,
 0.39211892977506435,0.303361230688871,0.8293299137942522,0.36359610272977716,
 0.4594480622061433,0.6868028988140656,0.16213627141907905,0.8453813272649369,
 0.6390765488196225,0.6650623324895314,0.7364998424312776,0.8906391934499027,
 0.3584229452467085,0.25529288319652854,0.7195411183959706,
 0.09271537905457383,0.9449525747973643,0.26111617107255825,
 0.7355558822408917,0.9778523407372455,0.30391503526893004,0.9470164883049222,
 0.536368522403438,0.7899993632202013,0.58152021230468,0.7175331864175418,
 0.25513176633579104,0.8515193447285352,0.5461866827125703,
 0.28810329338007257,0.10099851289399675,0.09838864289562232,
 0.3889465485497784,0.20666085487298103,0.6986846733324074,0.9600780391933774,
 0.515857031373954,0.10668232707686465,0.6654426403612502,0.6203303711344057,
 0.9059687875111581,0.07578188335736069,0.1445659251370095,0.3093760292119434,
 0.4466645652585489,0.038182995630901884,0.018051581969128816,
 0.7514063314431642,0.7505942697985082,0.30835722368555085,0.7249506854290341,
 0.8181056048767495,0.11473638245241902,0.5936420329545251,
 0.36747259526408294,0.569210871117918,0.8443530405907453,0.8278225765704759,
 0.7639322252162659,0.8622599469435855,0.9700239206366139,0.45615261513373084,
 0.4515350237932495,0.2067362487598392,0.6224043242814966,0.9566716430657125,
 0.7101548624552876,0.9635117072734357,0.33475563173520084,0.3764691091171978,
 0.3974255020262284,0.09263164241722466,0.3048676784908956,
 0.48808068922070025,0.258877590060797,0.5490996953356959,0.16921287139680463,
 0.16370641355522586,0.9133540730858414,0.8240468155840004,
 0.32221753306777445,0.8503855953276912,0.8298027582662371,0.9268263315830091,
 0.03326482953776311,0.7482934574724046,0.99997233536402,0.05139191961877321,
 0.04481173079702583,0.41176328285181607,0.9489788283169805,
 0.8093810715474916,0.20283135413987463,0.4362528987445907,0.9073738731301768,
 0.5812511206687205,0.207069486890592,0.5042983980433852,0.9444248068121893,
 0.5210222622701924,0.42931649021828633,0.14540996760155778,
 0.9660344164211357,0.31166925743981533,0.6072371976515768,0.2370515884625638,
 0.705930290337165,0.004785412102455533,0.9325579453722009,
 0.33163589565532736,0.7439199148045081,0.056898401304515844,
 0.4180374074850325,0.4970282356536908,0.022326938939916996,0.887617733365319,
 0.9718300692550209,0.1463751017545899,0.9831757823526952,0.8409957768019343,
 0.2115103537665346,0.3711227869144743,0.7504143906789759,0.2102513299065104,
 0.6192256642862257,0.7366497999228475,0.6697654295252905,0.16462048430059228,
 0.35567419017777346,0.4813341084236076,0.41952876257407523,0.418039747029871,
 0.742559070429988,0.427495597143733,0.04443774933134048,0.9044259205745123,
 0.9740117211535998,0.5977019733973886,0.016338758479190973,
 0.2509464940106666,0.295287647148774,0.07788656786756376,0.7756952302802993,
 0.8811053608887803,0.04225535558255644,0.05621406242474325,
 0.21875960453274512,0.932238017200259,0.23855317851158997,0.7311276542556078,
 0.4465129835047491,0.7181014226056003,0.6199951529773852,0.13850477297584962,
 0.5710617267741709,0.47794452343918414,0.4148619409301747,0.5576517462812755,
 0.8694873805810077,0.7438907171528809,0.6787335990346297,
 0.008559423947621547,0.7760579041720557,0.22036382020836132,
 0.3395744581827941,0.8460197676031541,0.0534956612125268,0.2801123525005279,
 0.8389804737071661,0.9123068708505194,0.1275955618675677,0.9264284675644526,
 0.303735666076141,0.4101458491497735,0.3822534281482425,0.36198924447856573,
 0.0038743697609330274,0.7622071872285329,0.871514116731664,0.620365003007795,
 0.911190533391361,0.20414417228406168,0.8618721874596498,0.3274544737065873,
 0.1400769696844888,0.8575095682626168,0.17904581167556455,
 0.20646950447508916,0.30854784898233323,0.45993032628833375,
 0.24010144910452214,0.5119648114463295,0.4991142801604067,
 0.024655291536715196,0.33484719412007,0.39319031920519953,0.9241322457590497,
 0.9512803814698593,0.49074081555333793,0.9312732672535423,
 0.051915764001472575,0.5329830886525447,0.9334402471546579,
 0.7966984926805244,0.6159064346529824,0.23935286119118326,
 0.18345634107750575,0.08418334825709894,0.043521850809341234,
 0.6877781293249793,0.21566305256345886,0.870754607662206,0.3011866302235875,
 0.34367463963729494,0.7089127849492833,0.6099407427840006,0.5111086584668895,
 0.0865677646794013,0.6450503942686925,0.8976510133329085,0.48239580701739526,
 0.9797149626206154,0.872271068765532,0.19803610688708173,0.5833078194462737,
 0.5504036541264656,0.7554386877064857,0.35598652263748054,
 0.06198435219174625,0.5397409952052483,0.12019167431995181,
 0.6302903784543159,0.6979595980239218,0.8414636050827604,0.2917890555104292,
 0.771115574757458,0.15978217985031506,0.10632174853905718,
 0.29079694154207436,0.5397319599163989,0.13008464815718657,
 0.6313073186068893,0.8612305629272529,0.8512021077939481,
 0.038262695545011605,0.3890916470195207,0.16828271059663336,
 0.16618849157301785,0.003426106281901986,0.5140243281775718,
 0.08685203248762363,0.9520531739571155,0.5243842210822034,0.7215447959036531,
 0.010889978156680247,0.1743447698408448,0.9518192046137034,
 0.6771708156079946,0.8062032599663969,0.40028534185790066,0.90208959552789,
 0.9728154978436225,0.7410601124649825,0.9518415851431098,0.4927157716192917,
 0.8697427156010132,0.8633113659511265,0.37305751805749054,0.8607512355087479,
 0.612485412086608,0.46784871235922676,0.2453322979602377,0.458719557063605,
 0.8605168796458772,0.1075637973154201,0.24569437433520658,0.8872536581455301,
 0.5940599355340941,0.23841848002810584,0.3228288312923159,0.8000188114928148,
 0.46490218954397566,0.5032487749490988,0.8990040284907304,0.617551567408724,
 0.5656854198189507,0.7078489117374425,0.7410744207751017,0.06833314076557528,
 0.6354680517183202,0.35904743285175644,0.015024289961127701,
 0.5631677741427065,0.5493325067244396,0.21700826721927324,0.5498658093274523,
 0.6070088291466851,0.47439182996428764,0.047907907138658445,
 0.5240836520626737,0.20556993752622021,0.7329408012162059,
 0.12017034314781239,0.4733909347589823,0.8917073092416105,0.7228919421571058,
 0.007367027843521501,0.30544267804545666,0.9853744939253215,
 0.8062958011551556,0.1411636137944441,0.37299624696880707,0.9801638860672549,
 0.8230091498755657,0.8051685249960927,0.9599620908717673,0.17201740067652793,
 0.9268547981218068,0.9315384154101543,0.6346141259001172,0.5186350143015107,
 0.9329873276273574,0.6428053486177678,0.15370355121254797,0.5159205491415433,
 0.7803858768037992,0.6724600529563287,0.19100853264733275,
 0.020684511864233146,0.3346920688839249,0.48439523224423553,
 0.023362051632034975,0.27106485768677613,0.2146229097855472,
 0.6630523276610348,0.7801469323601719,0.15441364647733002,0.7904962728135456,
 0.9619356887172636,0.30870386970575414,0.29632865843533884,
 0.9953774468304787,0.023190444559646695,0.5940183493409935,
 0.36751120556444017,0.805164704104191,0.6863113002038398,0.5312176135414116,
 0.3720239411838887,0.03777989624420153,0.38066525707102716,
 0.35511491184392263,0.18293066760924115,0.692762419788086,0.9918891934350623,
 0.040403579572765747,0.3199617919412544,0.2104337044782204,
 0.6596953172149828,0.6753175353705992,0.5993344655105352,0.17728417259429818,
 0.8237595277687043,0.23940903183693762,0.7258932960183948,
 0.08315003593744363,0.8023225851421131,0.9684823043858407,0.2438601444361912,
 0.27082292096888394,0.7889376723642318,0.40576684958461595,
 0.6936128803503862,0.11803454874219876,0.3792677796433074,0.8286018376348196,
 0.8472371417854351,0.5075855032875904,0.41538960163889815,0.7085031022221732,
 0.6962620055860689,0.5106634017017828,0.9586263273280021,0.4021979915270225,
 0.08927158027430271,0.3955175459218244,0.007031900958939907,
 0.11315073020498745,0.0820819452472964,0.1795850765452217,0.1346999252360952,
 0.5222931215212278,0.9539835879030336,0.16851042751865042,
 0.13067372719435644,0.8555012673565663,0.9115278479322839,0.3205233930762129,
 0.2063594693978268,0.8609229562074363,0.11933927738999439,0.9310395217420967,
 0.331959691097758,0.8138316864242883,0.5267580487247285,0.918666696413609,
 0.9043365660354943,0.24706515513299798,0.5732843046737226,0.9516425317939754,
 0.35707321518485247,0.2286521405207459,0.20237095448206344,
 0.044138591991479736,0.7062664691748044,0.46355907680303765,
 0.23650703149263086,0.6748873340922265,0.534390400574922,
 0.015091186764973563,0.043922250082897185,0.18472848737330205,
 0.27588817560330836,0.24642862295269286,0.8642469016370684,
 0.15430527232960556,0.9132627087962898,0.9054451558105474,
 0.03584157212423822,0.2419367854909905,0.33288534007442716,
 0.49498550344591496,0.5747507585194576,0.5911002204323959,
 0.18913513831761308,0.27533304763581945,0.3875544377040514,0.782848231219637,
 0.6628463779837965,0.10735448443116591,0.28844688816618047,
 0.12811103412357183,0.4991539887974833,0.12117242092101232,
 0.4373720586542268,0.07699704743761071,0.2772522208914572,0.1575264317759908,
 0.6200456455090804,0.8709472901535363,0.17474327063945438,0.8368643351416452,
 0.3404783772531246,0.3997516896826594,0.7991786187275896,0.3606211618818753,
 0.6004423751766726,0.4146151796035418,0.7792477844059188,0.42576391081842524,
 0.17059373297116687,0.9862467178820506,0.35938985987875416,
 0.19694724013310783,0.04854434566989507,0.25699299077897386,
 0.14879452203565502,0.9133538998083596,0.3241172892651478,0.0822288319764446,
 0.8076814146563189,0.9487329017122165,0.47513992518519554,0.6605919832295988,
 0.27793700082964745,0.8920789775003344,0.10833316155782369,
 0.5903478893907218,0.598401626767747,0.15326124595159218,0.08381725993364697,
 0.9429769751237039,0.6592932333686751,0.14670504425166764,
 0.08913682760875452,0.9229604948839296,0.29412801109811537,
 0.0945576434952724,0.8273824280782994,0.6377318363447282,0.24243874874537108,
 0.2537804741573916,0.4406972686761387,0.09652729232021229,0.8862095782965738,
 0.12172036525788221,0.5852914858499786,0.6502773763107654,
 0.11308091199205261,0.8731744791903,0.05341439824707894,0.948495029479846,
 0.06735493869253284,0.0024222921891154936,0.6514345883268038,
 0.24342508672017127,0.37111013370539925,0.4227558657655961,
 0.7569061709084307,0.41975444122642513,0.9554698589127121,
 0.16494579916388807,0.9141933572275899,0.8622194707419534,
 0.17642708109541827,0.9523056544831507,0.6173443379410112,
 0.42708361517600446,0.4800026121966645,0.006872642336356094,
 0.3225772418436237,0.07978805166213798,0.10119641361656995,
 0.9261717473085531,0.015101262804039561,0.18469391328986806,
 0.8649736108095857,0.7063054529380088,0.16581896796536777,0.4933857896175775,
 0.0781639187756682,0.9755413113202511,0.7247664622200106,0.5660147673030693,
 0.3025379598592819,0.29015980285512155,0.36069120004399224,
 0.7652624843637751,0.3896849318524729,0.5299544324751976,0.07211099181562808,
 0.35226249677115473,0.4286481184744372,0.48312243580373726,
 0.8778065871375027,0.26004195787202145,0.14097485625644746,
 0.29707626184857006,0.6534394620867889,0.4186793070442829,0.4431102356686728,
 0.9734045003743511,0.7623732970455651,0.11974388342387277,0.5288653208533847,
 0.7910285529872589,0.3917556460017667,0.24425596533884075,0.3597319180477706,
 0.7265459363402935,0.19086017660579946,0.4376684783480127,
 0.31536744995786625,0.33808636966288363,0.6424698236462697,
 0.008562358671823755,0.5084182775980984,0.5156753912138325,
 0.5676105860370504,0.32863326370667745,0.6282867988498215,0.8992979119110082,
 0.9755433229309267,0.13879955609643013,0.4091852286960985,0.9088341327347639,
 0.8145154071697476,0.9070854545011583,0.08852597754005442,0.8578887568660646,
 0.38851837962900937,0.6027102605634381,0.09682885626157134,
 0.6014132165343677,0.8851420157293678,0.5074443654834054,0.9060398851674,
 0.6549459680103883,0.6691702188743103,0.4165503693337569,
 0.0010065592519472588,0.8500771234479799,0.500864140328958,
 0.23289862679286732,0.5727527149091367,0.6751503317610825,0.9627446635404323,
 0.045863944485303176,0.787022011601273,0.4390905920458975,
 0.012022900867915909,0.44652475621825627,0.28332537190049856,
 0.06936041866263376,0.38462564866178506,0.17581857701949577,
 0.8466304683771526,0.5846980501257811,0.9817787538316365,0.3059737423824537,
 0.4658688615679899,0.7545189541131366,0.3988980381355446,0.7731166541276977,
 0.3781441899525708,0.659280056017069,0.5564681554371558,0.24888821924252813,
 0.19165083346202583,0.03440304379305903,0.6142677235211215,0.121502766282507,
 0.9865593107935053,0.7858809701610295,0.7679143819143427,0.9936596429549959,
 0.441229547999728,0.09506037021884606,0.31990073405506625,0.8720383339489596,
 0.115136441668934,0.10237034248088583,0.3388886381291758,
 0.040469373781153006,0.32881958724354365,0.6057547055102019,
 0.37200713337418734,0.8626785810443008,0.8186376735021293,0.4660885287847222,
 0.26890231566722234,0.5388914696567036,0.10939497839141943,
 0.5278247016494879,0.04870607891347856,0.9738023431801497,0.9824292916237811,
 0.5010120214466501,0.2198240374574144,0.8619219791792672,0.3707968369319786,
 0.2060761460233692,0.3417688414569271,0.7097084739254934,0.9886474135600445,
 0.7799483080490296,0.5111300436550518,0.5256429132518076,0.3767637557478275,
 0.5590683335627625,0.4041914243020426,0.2878806427934396,0.9061179829036712,
 0.894447110514106,0.5500822340242745,0.9846030528200166,0.08141459049677646,
 0.0366437847954727,0.04474853165084225,0.8562197522868681,
 0.08109227718500933,0.9498578763472062,0.2907011866204722,0.8693747573520614,
 0.8230722737102498,0.3429504349678234,0.09069686572841029,0.9824436166637549,
 0.33476939687871554,0.22971901772003345,0.22912800785697207,
 0.8620084019343216,0.6311029148479756,0.5233754399377776,0.6042413608075967,
 0.27379657671374935,0.926535101307153,0.4827864191880369,0.7448018198327435,
 0.9140719382571951,0.17138626420757097,0.3268379597008716,0.7419242479787967,
 0.9585456965424692,0.5465976378487679,0.32750476986460175,0.708444590790546,
 0.6778286038136979,0.2816767698401036,0.8171456592164058,0.3945226624206304,
 0.7422835021513006,0.17962382260514498,0.05814979780473817,
 0.4320501177777295,0.7220503177630823,0.04689320133236852,
 0.30192613135698687,0.7982369150799148,0.3161155643807173,0.8009206729370519,
 0.7251671267396013,0.6529337698467772,0.7638319882646669,0.46616123159427636,
 0.34184576686624013,0.9901875927439325,0.5750142175356479,0.2551998769515259,
 0.40179225741401436,0.6816353221089221,0.5287162499848531,0.293975166227163,
 0.26362331680952145,0.6912920551741916,0.41596416749340825,0.935669408876889,
 0.728221265053681,0.9935139130834076,0.6608703994605409,0.9104942192766566,
 0.21364127041705017,0.5337052682374424,0.493852002275583,0.2668420374454361,
 0.8028368475720178,0.5514181845327648,0.882865172141547,0.8292437786070603,
 0.33026725032103854,0.62736172632067,0.22148027993614394,0.29542780626208,
 0.40713670593705076,0.2169790199012832,0.03640282853907961,
 0.3471162796104311,0.07407571704081384,0.7646470753708338,0.8488701773193643,
 0.4786795353977795,0.3725496128617082,0.6811832906673051,0.09408881025358473,
 0.7200839371420102,0.5126208477612073,0.48377439109544107,0.60513060632269,
 0.9968334512388711,0.14167755448771568,0.043503521602272,0.8028460057395626,
 0.84073099089889,0.5751532886885825,0.9029084513961665,0.135737957387965,
 0.49773716849964267,0.36878125630739,0.3391201591911771,0.7489074148862958,
 0.6484198558602325,0.8844560340048135,0.6488639484222305,0.39450152223962065,
 0.3030802830074011,0.8337299998617277,0.39913970682790456,
 0.24188527431651352,0.5651377191966207,0.9344421126366554,0.5168428593502669,
 0.792721666045187,0.6733739454828026,0.07232549216527384,0.6709445112921792,
 0.1823443327512937,0.7781766696721596,0.40138014391478805,0.3353659365319821,
 0.7603794942361851,0.20340078669684059,0.001134888401716161,
 0.256581253979001,0.3469368463144764,0.33844494844252315,0.46690535196874994,
 0.2303747198065471,0.06921388362902503,0.0049135814443874715,
 0.35453050088210036,0.8246874932369685,0.42779436190279885,
 0.7196199362270275,0.16288253148858844,0.576481060040311,0.2276123120401884,
 0.029487895396503427,0.13088551919798386,0.7410776523993573,
 0.9130566030224874,0.09819178880999713,0.7531477021162054,0.4046663504121514,
 0.3113999979484289,0.5044030749846278,0.07505073287126407,0.8305741447301026,
 0.15946606891959736,0.6759547417063763,0.04253213158730129,
 0.46734220029771867,0.8923219748737184,0.45735942792829476,
 0.5423590501879784,0.5926186256015182,0.4998033935418842,0.4466954367832662,
 0.69386395714752,0.6698015109713376,0.07521247188343172,0.4531986846720756,
 0.15495185314666238,0.2505883312426953,0.8763288333115926,0.172445721863542,
 0.7300623274885016,0.8255608860826398,0.8340344674899961,0.2464775859508317,
 0.7271771635219232,0.5401792674717398,0.9167802409451576,0.7087721030310242,
 0.5404197819623925,0.8718888304136375,0.8189784353737822,0.6688198073584315,
 0.5843594229487346,0.04528122663654477,0.16633351212999314,
 0.8142620594256335,0.6312900736496687,0.07531351275382503,0.9600676627747634,
 0.12135586188478587,0.8020628307345744,0.3119191433863415,0.8255564496179366,
 0.7145892554360388,0.6427443376889146,0.33067079591086423,0.3287003146303805,
 0.9563176187753126,0.4159520413340383,0.0817430952898689,0.5983701643937366,
 0.8361965183790168,0.15593464030408444,0.4736383480931442,
 0.24608693021919315,0.23947311045091402,0.959799095487749,0.7250846185046057,
 0.8849499825290922,0.14200678604559225,0.6213240871641018,0.609508225342161,
 0.2068113453250493,0.565191349674472,0.20063332057715177,0.6949148054485881,
 0.2848539838504335,0.7951396386645335,0.4505592767027742,0.9272687258543093,
 0.789298960870889,0.44480669756068125,0.46392266712759767,0.4369155202400532,
 0.2239858334119802,0.45081767537268314,0.6201956029664567,0.4841331207741961,
 0.41506360691751176,0.8666559786201798,0.9466401490814402,
 0.008793713134521264,0.2447867382405251,0.4138497691706037,
 0.30045035156841426,0.02029368096406503,0.500840638218362,
 0.17464056500716885,0.0988369908731288,0.4254352067513436,
 0.13601005956736345,0.3835931788163589,0.29055024505550964,
 0.4866160368979433,0.32870249675780827,0.7024393142672952,0.4477249822427001,
 0.2083098108515713,0.6791154965562769,0.30504399782497515,0.8341509689784039,
 0.054327557833751206,0.645013845176956,0.07697949744748267,
 0.5483827900880696,0.2825878615911178,0.3072945909764385,0.9437903298858714,
 0.31209819475574363,0.21622648800867206,0.052228599584355884,
 0.8996528864576055,0.13313741799297774,0.08217389573060041,
 0.053573527448018576,0.1670148497131012,0.7859528249210085,
 0.24322907628253476,0.9730501703921492,0.5304334829783236,0.6703358686147514,
 0.8675158368534691,0.6056596823443848,0.8125931657574025,0.7628654977854592,
 0.9464618138317631,0.08548933990784768,0.3005102328284365,0.8192152415376408,
 0.09872501606054662,0.46806470163569014,0.26889162459396143,
 0.9307998530377648,0.14544093550860593,0.3701317034748832,0.6535549283817023,
 0.07475619379851384,0.9875766446476819,0.13307366243795027,
 0.8631278793870857,0.8961498789689575,0.989411598376056,0.17946707759086045,
 0.10532954554204343,0.6207381235568309,0.2895849340296265,0.9533269085602886,
 0.0606687293642042,0.7974249296460473,0.38544858178911046,0.6569615950969556,
 0.511555878096392,0.9748780132140407,0.7298832449316435,0.9056248513043156,
 0.91702678263594,0.7918271094039491,0.26157532398647365,0.580666828446192,
 0.5437187277615134,0.08096986683873753,0.6521203925051491,
 0.17784259465082486,0.037786391378007655,0.6708701672121291,
 0.3669025724591486,0.9152351356812864,0.4953974670612622,0.255376628005338,
 0.7683595442389761,0.07374481581338899,0.8927535972042654,0.6751949808963933,
 0.4047848710130658,0.2925111999201241,0.590462556334349,0.3356570601389216,
 0.7333690188679464,0.306258075890419,0.4613094492684767,0.09618766939374945,
 0.8646806754411569,0.5912244192331769,0.30938617818689795,0.5125355306673693,
 0.39074972233680105,0.8409223737934758,0.5687481689135305,
 0.16394607931134997,0.26182261895311243,0.4823725748513178,
 0.9285539829000401,0.5599669431285706,0.30228217978492644,0.8988946271194151,
 0.0065211133404061705,0.981437243865257,0.5208264202153366,
 0.08870152088701933,0.661857079515565,0.1406008752919663,0.9457099035425601,
 0.9685487532201748,0.615390071621425,0.6611103068870483,0.7975256072237278,
 0.0729864449271509,0.7895755619826567,0.5076636180989861,0.8516493334181015,
 0.5451209145638202,0.8562416859270252,0.24190718001137068,0.4085217129860741,
 0.756806762521585,0.761191621249798,0.09673604567750482,0.777172726519995,
 0.9428189026328333,0.14729782019389503,0.8687635842807263,0.8922281047248102,
 0.20943752683753725,0.7136619728122914,0.28310905387598473,
 0.8723094258367596,0.69020690857999,0.9255193232937119,0.2645814811114695,
 0.654733193836692,0.9455748654851988,0.4400799153863024,0.7569184757984085,
 0.022213937916339632,0.36191810737723673,0.7129167096933461,
 0.3565631455474412,0.15879638100252114,0.015135327055791126,
 0.27695583995220074,0.6169660370697525,0.56977742242683,0.5879949570115273,
 0.2686507196939606,0.9033596994833449,0.27222625748440055,0.6991963536210861,
 0.5894847169303237,0.25541511769968184,0.9509729896180736,0.9142090517071697,
 0.3617370943369106,0.913582922138553,0.666805080589136,0.34347412152114376,
 0.24189494550503987,0.6071054447696688,0.6424514828389004,
 0.22310602233387478,0.8679808578507829,0.25893119986091917,
 0.7151773515244226,0.7921224752492598,0.49702451753707666,0.7035973416694941,
 0.8961109738638311,0.6164560958610716,0.5477308806726505,0.17738092420700613,
 0.11112135804108547,0.0395521987652967,0.12341679282698714,
 0.7112114433478721,0.8957059987862933,0.35485664731848165,
 0.44102532576978903,0.12600622228090552,0.5199790444670688,
 0.7206629849867165,0.5668644586550198,0.578858059660082,0.8344999713972648,
 0.5463832327097602,0.7373146132446212,0.9733866693852702,0.08417600329566521,
 0.43587406091571046,0.20621089685558291,0.4857033406624409,
 0.38543304241095744,0.36820646260681045,0.7133862039729936,
 0.9627803863585846,0.6865983854773953,0.4343755124394608,0.9927910297924041,
 0.06792946122470978,0.910376405097188,0.4471392434611068,0.13198547086928092,
 0.24967880012288246,0.25672364147154125,0.6263568387164207,
 0.08884574321802075,0.02336357948571488,0.4567484938345805,
 0.7621525739863068,0.27939129159891396,0.10272540385950912,
 0.03503306442884668,0.688695154562331,0.5395528724427998,0.15769550021449075,
 0.7499598653092452,0.24134947783249838,0.5227891857205357,0.3222047481499012,
 0.2542248699063542,0.3096705622917556,0.930964591177686,0.683996690123661,
 0.3875743150723191,0.6864906222918944,0.914016477779829,0.8260778744428547,
 0.8635639515639283,0.8460819912451384,0.6598909525429962,0.8636627468947427,
 0.9296506060783494,0.5940712364541605,0.7737309898334521,0.43898302898399755,
 0.19352733950119072,0.7908140131618329,0.11744710585029894,
 0.6563440618253823,0.31558388707796925,0.014715337952745267,
 0.12562939800508555,0.8516172722270628,0.818178397848967,0.42622526047222,
 0.6378181572444988,0.3293870960931016,0.3146286932233332,0.9523430149380502,
 0.14875634871973875,0.15508126532375255,0.19793126744947165,
 0.4004511494965465,0.8091063321805689,0.9015410286366321,0.02512518571717315,
 0.14428498204032336,0.3375419581713148,0.5503543474909243,0.4164858478838307,
 0.9996353606228641,0.4458608349705502,0.795140291356772,0.5315271802823696,
 0.5539777945966483,0.4582620662205089,0.7020561239984494,0.6802490299211307,
 0.7032900204244519,0.6737244650877519,0.054680732680476574,
 0.8364752936778486,0.44410512298716676,0.9694218483141722,0.8446065107540075,
 0.9540433856702685,0.9273223318938663,0.7799579508423967,0.25703221323858594,
 0.8476971548545348,0.5265243834003673,0.04249685387582003,0.5770949653185686,
 0.036994029409430085,0.9048477477355734,0.1091607554582017,
 0.5945634314774133,0.72205165752067,0.16073324098051722,0.31058632649354967,
 0.10903267084452706,0.26963748409039723,0.628254192279944,0.4184505850648409,
 0.7320122615507404,0.47186062135721296,0.8293467846563981,0.4236145150858107,
 0.43238282468371714,0.08579594228927956,0.5210435711927537,
 0.7745018790438803,0.7352370421571056,0.19494592609685868,0.4186441836897745,
 0.24859547423670003,0.92746183553503,0.5925019463466016,0.5777337515395674,
 0.0781611455525566,0.8312325399689073,0.8324788836416895,0.5487590793337541,
 0.3422725928577781,0.054534646076387916,0.8331191875349967,
 0.40594414682902413,0.4705149478703735,0.6426145871120863,0.6212615122585354,
 0.723377860938567,0.04002285012083617,0.6176641808862626,0.41818435433929135,
 0.9734087084348431,0.5513517539714026,0.7538149605807464,0.04133808779751191,
 0.3891183222795037,0.26827003514707537,0.32335508088423737,
 0.39817936526393893,0.8957221423619337,0.1459028229988989,0.9218503831209054,
 0.6855279281856373,0.956097626329204,0.7145931206017144,0.8084190345191483,
 0.5316014326974845,0.8822032969467963,0.3981215096452998,0.2519373066145879,
 0.7465155725902723,0.3848174018655419,0.5866866761937513,0.9347533059217868,
 0.8636080528817126,0.8438642622117003,0.5859103917167434,0.963751456208882,
 0.06534876385983801,0.796257746078465,0.7640796583702316,0.9619712701543274,
 0.8575947542115632,0.37220009523668474,0.4322740170803103,0.4102842950906419,
 0.07947392493186312,0.7258296385823094,0.5447992649803071,0.9734942130849572,
 0.7614563878263942,0.7633746848274259,0.6395123582906227,0.24491241861088808,
 0.5026646266944367,0.28913146833103354,0.9949594873723006,0.1344536232633763,
 0.19889783698425356,0.34485009933329547,0.40705437998914373,
 0.33537086644312586,0.045422155406971254,0.10872117413788596,
 0.4464186945455867,0.1166919241315123,0.396836011195433,0.03384796579090632,
 0.8618188733685427,0.5623255029338262,0.4248902241893774,0.41632038779760416,
 0.4401250530358495,0.936902560924572,0.7528306645077839,0.3478374927187048,
 0.3804146804439956,0.980438536626384,0.09216669229669738,0.11986193358013952,
 0.6820764038312748,0.7839327071088376,0.31611160772278957,0.981882764898065,
 0.9546002302425307,0.9936845556490251,0.6054311335097029,0.42393925423548573,
 0.43256146270629336,0.6561372840536215,0.4400339481809955,0.7209623855246159,
 0.4615216690071534,0.8594648134583015,0.19885446991846079,0.7590290168880277,
 0.05734729627206092,0.1173764820795401,0.8126779292482587,0.9798882418691587,
 0.1962801141247592,0.6921007048946607,0.6801122335024833,0.37581812825099825,
 0.12649697502593926,0.8629703593501337,0.528752944136491,0.48411738157661444,
 0.296622572981156,0.8304463885878903,0.8315183185456178,0.44241712186571935,
 0.19849873884588964,0.16778778613267276,0.42203370262296924,
 0.008561028422844896,0.3579701013180182,0.6806064575064682,
 0.4981653058731933,0.43794660946278263,0.5476202454534792,
 0.19100294605901036,0.8855504967352097,0.6072700064645004,0.972417851502116,
 0.7150532228731797,0.17904111413572243,0.6052028440901588,
 0.005292905992547947,0.31469360301456784,0.41852506014098023,
 0.9166453616220304,0.9166892471516053,0.5536625810724377,0.3934462883360852,
 0.20931832628396185,0.1103152237982461,0.7809511614752552,0.5740808794172415,
 0.42525592400425094,0.3158021604145149,0.7389027126479374,0.5112159892675989,
 0.4616374633074547,0.06203104364142875,0.11607311013936428,
 0.39950895528588903,0.0074858078250364946,0.057480297511547596,
 0.7910326443058859,0.49921914920001975,0.11848086657907808,
 0.3190331592366884,0.07071911488057347,0.9369707584590298,0.5940865048375189,
 0.8527177155869219,0.1359575743124617,0.36667097928743075,0.3337379585484188,
 0.3087979847989084,0.21866600912254586,0.9939790520471696,0.4357087973293021,
 0.1482841158443925,0.02970118402157418,0.4721895828258609,0.270799547108735,
 0.9781044068875147,0.8774910956699143,0.34227398136848153,
 0.28306189151942973,0.6741004339826535,0.00431601619194677,
 0.34791678827776307,0.9663077070316228,0.4741000947605012,0.5350438633248114,
 0.34504079723775316,0.27262506815967613,0.8273213429501949,
 0.6042461886056747,0.21734365749540185,0.4808680755979967,0.8982049380491113,
 0.2480062083787019,0.023767850820734515,0.9344755107484453,
 0.46232490670556714,0.7509841900450942,0.11003489393267085,
 0.9801991255903479,0.33819480496400267,0.9662466217356069,
 0.36976842501076257,0.6433021942066682,0.08608662674215506,
 0.9195770452457718,0.5772583580066339,0.4240127846089844,0.15058845434099255,
 0.8797868326692264,0.7040777926841728,0.8588064132359066,0.8652677140586521,
 0.785632419203634,0.36161997715991534,0.4715948125763171,0.3898901178513582,
 0.3567897045153089,0.5029382338024696,0.7844496143657879,0.5955853975832275,
 0.6598604714807474,0.3872807757567368,0.8635317029583874,0.7533992304180007,
 0.7130673968938499,0.8502264063608397,0.5310803412573561,0.606017055368274,
 0.17175046271208838,0.2854787921509372,0.9783964158882825,
 0.26607821484160366,0.8758540694471383,0.21731859815711985,
 0.7649728406450049,0.5370806809280556,0.8508901297472092,0.5131696761205893,
 0.3442899523721392,0.8992139769443211,0.8790341584265392,0.91955943980435,
 0.11407369985834226,0.4632502132496088,0.16327226874981726,
 0.5899357603988917,0.5177055814563157,0.7715407878143387,0.4242188755143128,
 0.6049504174522152,0.6665292315225201,0.852923340860262,0.5207602620157489,
 0.8770166166202369,0.8947855643403098,0.20259283007850404,0.6682348486947751,
 0.7295658428622556,0.749552009144261,0.6894294053650045,0.44218616106537334,
 0.603277974056111,0.16979885960312324,0.4749502291738167,0.9673809152689995,
 0.013963927906942453,0.4528647331092117,0.31600676471360645,
 0.5358648032129644,0.5048505039854599,0.8405623125428008,0.18377244252495173,
 0.4572733077620058,0.6128017871135081,0.695754202946013,0.7781455617166483,
 0.6286285868467739,0.5065056912967449,0.888630492032217,0.28694238114862525,
 0.5487238721104485,0.11522534380013072,0.7970201564402692,0.6504393613210354,
 0.49557233551768765,0.105248538248508,0.1546590631569429,0.9140500941149712,
 0.41721516206361475,0.9415519766153946,0.19616320210210425,
 0.6952333173655214,0.9429922327564445,0.804951494023255,0.8619939818475416,
 0.05678754319996704,0.0032414979125801047,0.38284314138368036,
 0.8377134100851346,0.2799689263379891,0.00895796515732561,0.6739123905206788,
 0.09211730673145835,0.08408757689686186,0.24712202748510703,
 0.10169803833864188,0.5924666288369971,0.9447956349366584,0.515057840496502,
 0.20922654816772956,0.808065526122076,0.9541121974236106,0.19516112569024435,
 0.33348551158494744,0.018589938468846712,0.2948451556919641,
 0.9965021821602769,0.35559992579068744,0.8543819359437292,0.9410043457555848,
 0.8818193915477142,0.1658897454962197,0.1673667755259256,0.21894259584170583,
 0.5229400256566037,0.5284309915492653,0.8123816494374158,0.5345104734132053,
 0.36559278315966925,0.5800745935469503,0.5667365628621149,
 0.0024543052736104576,0.13920251364967084,0.1487446730655403,
 0.39955225482684764,0.4749332902472455,0.5092482226408042,0.6422079418041977,
 0.048929684505740534,0.681047348856029,0.7021702890712994,
 0.42110693323396176,0.8626899565819957,0.9541215236173424,0.4461993900137545,
 0.5222899395761034,0.05740916844369637,0.8164628425302719,
 0.009803012536065192,0.7335713476179482,0.06121045421931004,
 0.6495284569587232,0.9692609596912507,0.6262977289591763,0.5650941937835648,
 0.11599314597772759,0.22850455292666916,0.09282568174991057,
 0.8432876205755255,0.7385070888366027,0.24327561521120955,0.9311471027570407,
 0.26784162339026374,0.5516728009360996,0.8627262234948658,
 0.21908551522708086,0.07056459461722175,0.8792949247131687,
 0.6878886492012086,0.564136109380796,0.842571123816419,0.6531228190819223,
 0.7856362705179939,0.9146783052257781,0.2800534541777725,0.6000037946942872,
 0.16139316324378816,0.08296394607539181,0.05492209742772025,
 0.782641552439194,0.5561675078273601,0.11281271207878052,0.4745760022926542,
 0.15726712300600143,0.942671147799157,0.6399346737254752,0.16839224178056567,
 0.8426383842209064,0.38457292721083747,0.17150002676964393,
 0.47253099062717574,0.8940566000191583,0.878137624087864,0.4405331810742573,
 0.8641908794001266,0.28844073826165595,0.660339019764311,0.9215606075247478,
 0.39652928248984187,0.6155616394600586,0.7287509144297827,
 0.05777360463128034,0.8273775757580373,0.21561428901061275,
 0.5212194975876979,0.07418692810763328,0.18225997987702647,
 0.5298436432371197,0.682832048361905,0.4775585337844974,0.32497969041156594,
 0.4052780362015693,0.22991470325283225,0.8002667178231517,0.1463004782646944,
 0.3568469911691653,0.6042181556095282,0.4246158797855699,0.2725737549935986,
 0.722002918072632,0.025710789009807167,0.517876754738319,0.48300276965090794,
 0.4619193737237913,0.19301929125123396,0.9347148534990413,0.8844437486029457,
 0.5531265899277793,0.21412637266102352,0.9529464905777025,0.5484174518128041,
 0.45774765145408325,0.39695893496832946,0.4425078761523178,
 0.7922676105300438,0.6565075611935521,0.8261541478545915,0.732017212735272,
 0.02767297939791291,0.33973100159927894,0.36217144358348163,
 0.46022039880358545,0.597326130104418,0.7364715110805985,0.16453469421656164,
 0.8521808271041034,0.06356572368337654,0.7044609822052554,
 0.26562271673260607,0.7459426125226625,0.9499801323280564,0.9471378350272374,
 0.5212313108239026,0.056069449559389306,0.6197629712846336,
 0.2252331851006225,0.7420981070246984,0.32905175443717216,0.7770293471510645,
 0.9545886104930628,0.13825570433024237,0.47878724567024633,
 0.4647407498158047,0.205285443940167,0.773728728009275,0.3098371163154091,
 0.4516939553989039,0.34683610537463616,0.8153351398142231,0.6441589549547084,
 0.37526532294394,0.5919615027962997,0.3060586873568353,0.6053890605019677,
 0.2354504832794514,0.4700887123398091,0.1367588378635306,0.9414758720991113,
 0.35799106367322875,0.29195054017824895,0.7483869795275764,
 0.21068003506789745,0.24415142697205816,0.200439121959611,0.5407235107035044,
 0.29706989261462113,0.6012340620122391,0.7997492424973571,0.7885058227750107,
 0.17799901142902563,0.8572398502005409,0.6785863919670492,0.8436111687423672,
 0.9881007945264694,0.9101076097038425,0.08939606195281435,0.9278815688370227,
 0.9202179529820774,0.406369395534355,0.01868292302466734,0.7486019660000409,
 0.5864377738918946,0.7268599157211197,0.8218729050489311,0.8004663731163049,
 0.3447459393167265,0.19183359170156944,0.043691721861457755,
 0.6414588878412957,0.5872545999796438,0.4948834816102414,0.06880558904220113,
 0.17343264867968355,0.09908235899416207,0.2548469204558985,
 0.016994461104945846,0.36049334674042677,0.49072352578767764,
 0.12416532432761829,0.3151144509618199,0.5697978707367142,0.7555973527014443,
 0.1536528402173455,0.9049493305900886,0.8530934433264538,0.4189671807837698,
 0.7775201903867629,0.7172590464376168,0.48575196903826967,
 0.06922414368145724,0.6097542313320179,0.058233052535085816,
 0.6183277180584162,0.18945818185342145,0.41352464209336737,0.420340172935984,
 0.9499635020060418,0.1727364107856305,0.3749261046133807,0.4980050481348014,
 0.9308483971597501,0.9163783173298164,0.7024638462907836,0.5312420685398338,
 0.6473054345882508,0.6032006407641931,0.7463513906753237,0.9127391931368171,
 0.7681828788656678,0.45844101411607596,0.14723790802723868,
 0.14566649742900062,0.021577368146103337,0.8561291981092956,
 0.39935202523802116,0.17018767104407284,0.49737352946965985,
 0.9365809459064033,0.546179299111864,0.16748842920770268,0.42124289395822145,
 0.5101065489071491,0.8616101594127312,0.01568452008689103,0.4414661405295075,
 0.020835049824735163,0.2932930158233521,0.4404848456375249,
 0.4268817655850261,0.1083800068382812,0.7260902059249675,0.0863168233896302,
 0.5201036835137618,0.9317578903678481,0.9109447454624355,0.5834494590003589,
 0.5205438659460933,0.0475393534193852,0.7786654802524026,0.6805182354134045,
 0.6438853542717318,0.009953807878840037,0.2841437366311089,
 0.3231296070488493,0.8171247486346109,0.849997400661486,0.13498516524650395,
 0.017060005445163395,0.07779405527256245,0.3887187579182295,
 0.9534302427547283,0.6512851032667972,0.1678343208159465,0.2832096302937196,
 0.32418533892549484,0.7893368572833146,0.6439496151258646,0.5669896599442312,
 0.9120303498857336,0.40453156335626306,0.23228322479670716,
 0.23595155003879353,0.29015289599891125,0.4788740135883299,0.853502297944389,
 0.5316129910744811,0.5326733461543932,0.6152294306122112,0.5456865720496835,
 0.5617152652199291,0.8961202862477293,0.6138723618583558,0.24969717980937234,
 0.6920520708423369,0.5349168272343754,0.9825587793549048,0.2081091649264184,
 0.1232219977169341,0.084986314315033,0.3417607566281805,0.6691435839675015,
 0.14950689948692863,0.7624301272984814,0.08622119602041933,
 0.9736217398191428,0.33825945705438965,0.9304835447695279,
 0.15499871437667634,0.15662452321575915,0.16431830584512308,
 0.34281821908318666,0.7118041361982769,0.33209613748866174,
 0.30327486670101245,0.723440059666472,0.7241537573405499,0.6710886278225554,
 0.5006150057987608,0.8691928179360435,0.842772971156252,0.31236140147904456,
 0.6082866101633788,0.18654074344874316,0.1551801126886061,
 0.34946800351758656,0.8968785420135582,0.6430385522178301,0.8302817552943982,
 0.12857208078029558,0.7730126014496634,0.7437320173713985,0.7892872038003728,
 0.30312099227034495,0.12064848296171238,0.45644417779960234,
 0.6123724550616104,0.6471939080932222,0.4616691176879705,0.4240499946006485,
 0.990457001140622,0.48858398472071796,0.6223962219697933,0.914032748332486,
 0.1331527577986562,0.6723828103641537,0.3534055661969492,0.16722556547452427,
 9.19356979438013e-4,0.6824739947827575,0.10828895837562935,
 0.31113414509465975,0.7723658504929753,0.2895992700524712,
 0.15970785089457484,0.7152594665397252,0.8797697403792591,
 0.21550589684857002,0.22838064806888692,0.12180846289044345,
 0.9344904119393016,0.6058668602019097,0.020128030867959112,
 0.8482126208935786,0.8890809997530232,0.3553531081390038,0.5256577160053719,
 0.16357881513612282,0.49426245377948563,0.16932899012956681,
 0.43371146804207483,0.05454156068750249,0.47618295181892667,
 0.26786741050412843,0.2679367548123899,0.1605763636202273,0.3198872407262172,
 0.4696471000698139,0.9843584111417236,0.044988503080652276,
 0.8775634449455708,0.4054513963886629,0.9142312187847838,0.08458773980587198,
 0.709186529863644,0.6776814079708828,0.0049862818556049016,0.950475808894006,
 0.4638810729919738,0.245071183718371,0.5234605562934922,0.4101948132787425,
 0.5598858065590061,0.19107581946506147,0.4357173635590408,0.5826999511532648,
 0.918061238232922,0.04320109656777715,0.07524385975000691,0.8247960995596582,
 0.5348598651328187,0.28416434681591374,0.5328458166284831,0.7367126608191528,
 0.7446726227348126,0.20009869454464646,0.9811067179103649,
 0.24180036753590373,0.5383077713629235,0.21249254404253082,
 0.029937773765535303,0.8547880637510304,0.7496070558383908,
 0.0834409886345644,0.455222986298111,0.3357777864243737,0.42714487983827276,
 0.6005182293452811,0.6051097997517245,0.030003874266205988,0.764443686442259,
 0.3866886673637382,0.2647776739747276,0.1814980060383633,0.7233414658641909,
 0.24391029659811037,0.569268300843813,0.07247962051150691,0.5395704237187934,
 0.25639803117562243,0.422514573327779,0.3018309304139588,0.37931826708085725,
 0.9840068720991161,0.41250650954358103,0.9609207491947274,
 0.17127220375648777,0.18514178659592262,0.07194062116403899,
 0.15795898109050255,0.02639243237283162,0.41550785660067935,
 0.8339136775292931,0.19857489350026347,0.5500819290471162,0.9221793948801966,
 0.030875105047916462,0.3686898964788513,0.8123344651716321,
 0.5251159029213757,0.8381833376331252,0.6459239319739881,0.25052704031035067,
 0.8387679172836775,0.03497027448030543,0.09058248506731159,
 0.7427419400458559,0.8833170359849893,0.38804303840594445,0.6142644364299878,
 0.2887774630700646,0.6693510946900457,0.7922784083575838,0.9139758823213787,
 0.5468376618622701,0.4533777335333894,0.9821204838792224,0.6005483298404488,
 0.24748347105510082,0.847676875829839,0.6708847095257633,0.21201877296039506,
 0.47308045932748655,0.9861598349010396,0.29532238021245083,
 0.7267285741051246,0.17147785490124134,0.7399392445867213,
 0.36582849683170116,0.037492531282091957,0.3878412105060023,
 0.3910158072606542,0.706119675177318,0.5395163399612419,0.07801089002264328,
 0.8504831119058581,0.34929994311044676,0.3378521162520378,
 0.47609359993667133,0.305150242709301,0.7754147767870856,0.7728526854152009,
 0.9329844051555729,0.1226469058409062,0.8158923637574694,0.4760081428381492,
 0.4351495189770924,0.042695471446786315,0.28514385899585837,
 0.9333350683352034,0.5375207435982866,0.4677406134255111,0.7327473668108307,
 0.020499505948797903,0.959228030119989,0.9843297645492698,0.7012236167243528,
 0.40668536928258925,0.6144407649642144,0.8845741497769907,
 0.48877489057564716,0.19943785794226498,0.28927319474926616,
 0.3687326278745984,0.9164199227976071,0.47538790785638496,0.2626235292078234,
 0.11684441100494403,0.5006591254799325,0.11590777405823294,
 0.8904341793178955,0.7722696391769938,0.3616833636358141,0.07144444658331583,
 0.9415749076440025,0.5890018414356057,0.10952331790183878,0.6236612749538115,
 0.7010711283124902,0.4279121727403107,0.11753070508993924,0.7533767895087825,
 0.20811894927951025,0.9451438263288542,0.2713159612835754,
 0.27570456307073987,0.17470261084172722,0.32778955022460854,
 0.3637200073969573,0.5928642407001831,0.8323663994948772,0.9159409954313383,
 0.42447859552765355,0.4866222393549301,0.42846384232383516,
 0.6680965563881806,0.39970772808971344,0.41310876301054,0.1351359631738327,
 0.7885678718175279,0.3236361524125755,0.7837056762894438,0.9253613997341339,
 0.993819727151259,0.13933224999379712,0.4555051821374486,0.577141515823812,
 0.10081546885729675,0.4818956215167155,0.47125530436022456,0.849904551924194,
 0.44025230051457254,0.5417316764713792,0.5242793207111459,
 0.08672721627342395,0.610706046655458,0.9602420189871433,0.5428495447285293,
 0.2617682240398921,0.7386221922879368,0.845880641724198,0.1287306151450086,
 0.58803105226535,0.7302315125729532,0.4714635261682786,0.8995873496395648,
 0.08348730905728463,0.5635833185626866,0.7616168636817415,0.2628821388683593,
 0.8294126947279301,0.98686955803612,0.3907204085118873,0.2192859985106258,
 0.5388466614959054,0.7378754603909768,0.08916940958940045,0.6299561208731767,
 0.15940775754236558,0.8483889242043092,0.30841564047391756,
 0.7083089781799525,0.17387856835475857,0.4982648385823202,
 0.024809594667970836,0.8322415951533729,0.07906816651083204,
 0.7423335380774663,0.25364533963281255,0.8819317510931122,0.7509598795777386,
 0.5662396954880999,0.7052212768321247,0.453873842800243,0.02396892031723008,
 0.15903711690335154,0.8257154002665269,0.9130250905685646,0.2470514786659873,
 0.3881061893496036,0.11220612184892731,0.4356500378464272,0.5430148784670379,
 0.46366993379255517,0.6498967475053441,0.3339574430396376,0.6274141819990806,
 0.5897023341998198,0.5399090407332691,0.5396658610428836,0.5174290659069078,
 0.8150424938867346,0.7156651609000959,0.31820651834175995,
 0.08452417490097675,0.45901240886311534,0.05652700728276305,
 0.26602735938918354,0.5785539897191877,0.541682906130268,0.31287104828580337,
 0.7765798820317829,0.848989077853099,0.040445483073009325,
 0.014020234439399482,0.4589561024391853,0.9567240176320889,
 0.4824985526587994,0.2037319899066854,0.6558032455011491,0.9289899047558762,
 0.4538085074275091,0.8879165906997852,0.8744454082476621,0.5162103087631817,
 0.04863199020828313,0.6490718810585215,0.2854190672549137,0.8763570638351316,
 0.8167744300652249,0.40480059432401183,0.45034457202613043,
 0.41257921216591176,0.21335019006578126,0.16651100607657143,
 0.7444592408701818,0.8832331471256661,0.14423418815271027,0.4193329929660806,
 0.30471995941088703,0.9865608816258592,0.21585241148913992,
 0.9430448831266148,0.8601067828389767,0.9337063588892038,0.8016141867515282,
 0.9495404135312506,0.07472834252872684,0.45510563534639825,
 0.4421004809122111,0.2877236491049151,0.8913054999692313,0.8797552140050908,
 0.439519861088829,0.21531260088907012,0.5988049936227667,0.17365436879410634,
 0.19286735234704355,0.2297147916270651,0.2805640109798897,0.6236463624529476,
 0.7337064271529212,0.8209714460757005,0.11326633089830362,0.5566795495512628,
 0.9258988450117056,0.6866677491798461,0.47280272371294163,0.7676810924662082,
 0.6772007004800431,0.5801070227470689,0.24252774216580775,0.9344922566820629,
 0.8622924156915786,0.30342301082348144,0.12339548401495914,
 0.19422253008164714,0.888920820001354,0.6215804793741642,0.8932070950165546,
 0.30297644066055884,0.18629168728527812,0.9544282766435983,
 0.031778961126115335,0.5051821216912943,0.8389622223684746,
 0.5606474167685481,0.97974697609298,0.2438604159073553,0.2269618769356836,
 0.9182441765100531,0.2589076241082575,0.051218963217248614,
 0.6116244289483369,0.9976155609440311,0.2822696874327998,0.09672778850880237,
 0.2252841618340774,0.226312035940756,0.9291087645244366,0.16753007631493322,
 0.5556695830528887,0.16606160822783234,0.4137150972056802,0.8561625004235078,
 0.6558615979268255,0.044944860914422624,0.8036572466473311,
 0.9485488404817926,0.8944001099226534,0.3191499270049921,0.48101180846428737,
 0.5045895162187454,0.03584384009516939,0.47870921718454296,
 0.8504112686145219,0.5289427075356662,0.6505267614091917,0.2947677316512607,
 0.7614329491523001,0.6265880862934158,0.34186416671092634,0.5923115783988159,
 0.22309178892029724,0.022638052821003507,0.030372754569832683,
 0.6222172740346403,0.550937277378219,0.4761075570242017,0.5941204894048264,
 0.7397425529218865,0.9807575680879579,0.6552270481011387,
 0.040502354113090555,0.7370437040630644,0.509679034025302,0.9197736689549213,
 0.8019309389195527,0.12239055480510563,0.7789029603976971,
 0.49827031229314644,0.1212222136464679,0.11647894207042564,
 0.4758330560605839,0.9068379367214036,0.7937195677153416,0.8500728193808941,
 0.3842367844981519,0.9571280886891322,0.5439603000517018,0.38642194918811157,
 0.5133012332684426,0.09333367295697137,0.060210402166564814,
 0.23571828382675142,0.21638269450541636,0.4061372731682348,
 0.007892605689013843,0.1445773626068425,0.3632831026624539,
 0.16239938218797412,0.10312440219926378,0.637048216565215,0.4381672544094908,
 0.6796335239109017,0.2362638511108075,0.0680941572586834,
 0.021536748500730618,0.62521887519632,0.960036846919939,0.4657615308846541,
 0.8354422797154311,0.25529367573040784,0.8309083610971604,
 0.14098770043119924,0.10824522000655246,0.7729059660086196,
 0.9586793371359965,0.3638578763030569,0.4144586988933773,0.06650522810031734,
 0.5299595024871597,0.44507758364391514,0.5945355684770742,0.5989749764493522,
 0.6135899825778988,0.7923611542573008,0.22859135353966598,0.8209609312891061,
 0.9451904946008529,0.6489585111095542,0.3629332648674225,0.5778484140272622,
 0.2923116433296227,0.7976121749344116,0.3941412050573462,0.8673136973261426,
 0.809681892781619,0.1912964858520877,0.6799262831876616,0.2392432241238921,
 0.3823201736596049,0.913647275353119,0.7379967096966134,0.2370509437378776,
 0.0913725105772194,0.36719400563756244,0.4560257592742354,
 0.005249032865533021,0.33248510501618556,0.06587268082729969,
 0.9020599113773162,0.7946167819878365,0.047936428280512366,0.961710986465432,
 0.8795162070331756,0.2098871490389187,0.2931483357484662,0.35748395595931315,
 0.028241873490136404,0.17751880153663824,0.17951592079666967,
 0.4130199860096988,0.5133394820137213,0.28521927697183824,
 0.06810405981529555,0.28324215285189724,0.006452317563289456,
 0.5021872640966327,0.7304541218315236,0.4881105076390311,0.9407647412738157,
 0.7267300112149379,0.7375094117540569,0.3510854036040074,0.9970047997022918,
 0.8968370489322153,0.5981953943544583,0.2710865392054198,0.2578658911101135,
 0.35275402314448534,0.84003257801655,0.9694822292391111,0.7725906820859031,
 0.4512241083055185,0.8719269165036745,0.2793438889693838,0.19305775613280773,
 0.7187500876743624,0.5652663141077074,0.3614528715010873,0.9129005746863473,
 0.0889043233620419,0.9002418411101581,0.04993431140562976,0.3553756844342787,
 0.22030232895316104,0.8981102536806955,0.25049676768492646,
 0.3338055333477552,0.9222361242580182,0.2657006652880052,0.3429694402319886,
 0.43441845028746073,0.21195840190234227,0.3006206537119769,
 0.6643635027062619,0.4154668654228283,0.8667331103655112,0.383364623202897,
 0.26996681590943894,0.33949757560450633,0.4875234413692423,0.216280875301747,
 0.6485039776041872,0.5342882092227117,0.09506404964507731,0.3647972358970436,
 0.7827933270133651,0.8409268066249277,0.023704224880842117,
 0.9211470738286879,0.4327957515139469,0.4923754700758588,0.37341304223095095,
 0.36607624416735784,0.7835569599223307,0.12773828407812537,
 0.42373992608492717,0.7132353805439253,0.7239959705999885,0.1099209073278008,
 0.780885885906286,0.6155469365140149,0.7077298294988297,0.5409217163415894,
 0.11871994814222575,0.3474495890132414,0.11184334581867783,
 0.2928339119622594,0.7939618175252608,0.788364147246403,0.4329768687326013,
 0.23429995479105303,0.9248799583939719,0.5855655869331136,
 0.07329742153188101,0.8233552762670246,0.6288410741150947,0.6512354895715947,
 0.5818755940891577,0.044009615463865615,0.6500212198715168,
 0.2296330326711412,0.27065361843109326,0.3934973611009389,0.7921229863408806,
 0.992468871733012,0.902819702118042,0.025981287814054954,0.2431524030340213,
 0.3074433370800741,0.6217973018715499,0.9336710669289077,0.5796905008844367,
 0.9667624682544645,0.4726579672219665,0.2509466364625732,0.6339225997152723,
 0.1257214508201856,0.7538602452275048,0.5525796692891245,0.761930080068365,
 0.07533995830420803,0.10896205030310746,0.9862958917433741,
 0.12704395401051338,0.3404589232432209,0.17489170430072465,
 0.6974056601337942,0.8002719098364822,0.18500981662128724,0.6075968144239641,
 0.5709721708888984,0.40257592940207954,0.9870850644854324,0.8194973578511272,
 0.6901870021462925,0.24194263000484284,0.2015952228259008,
 0.44982299302379547,0.6569949711038957,0.6700231145333716,0.6212649354394902,
 0.7316855688408617,0.06024896124282575,0.34658255668907445,
 0.4927628263414736,0.7298506273277244,0.08677953847775721,0.6399959689844371,
 0.4156977965369357,0.1991487390488571,0.9266450248447263,0.7675030558577947,
 0.6629500297591443,0.02240921821817221,0.25699522682318343,
 0.2591036167572647,0.37587183703935745,0.316251097034735,0.8631082586741405,
 0.7236139189897163,0.811904708150866,0.9765995016308817,0.3489002238475728,
 0.05867411628339958,0.29687452032232287,0.31558804679396235,
 0.113018552340858,0.2908028224362489,0.8757343529835739,0.559231914638981,
 0.17444075358617983,0.2876080045954512,0.1005193316253512,0.2438011385549581,
 0.6050476697605538,0.055567047120277824,0.9362115800621602,
 0.7883863442069412,0.21637365910811845,0.03660417079679534,
 0.0697764120618043,0.9510422589430847,0.0655355931235484,0.8480738358118866,
 0.989399805469038,0.10372786968150871,0.5165656932192562,0.5005243524733496,
 0.3620601403665151,0.9298862943508042,0.669574890451106,0.269711687817521,
 0.4279038769627297,0.5312918046552034,0.009161171906580057,
 0.6359342958267798,0.05712590668077688,0.5799692615308902,0.3084236351837989,
 0.26996339122020385,0.627800563275523,0.6961660083442283,0.8054223996709723,
 0.8746646894776127,0.08079680832327929,0.6694963140139728,0.7163763261208642,
 0.7527439064982929,0.9993397807421106,0.07518344127047971,0.6993261228965251,
 0.27362326360779243,0.5087396335701503,0.4735331817031224,0.8873038065808244,
 0.7920276775781069,0.41858549857516614,0.6087940004177781,0.6104787310041583,
 0.5480450904673096,0.6042853720822482,0.2689635635593465,0.8054171361061638,
 0.9980771374790073,0.24670080486137647,0.7562754049550187,0.844341669512756,
 0.4261453190665534,0.4807045394343885,0.7958520900847059,0.8393092482779834,
 0.4669230788750871,0.30506884296552783,0.8253187960758565,0.4723658954438428,
 0.26669381601889475,0.07328266847897114,0.4620234908697155,
 0.2526775404393904,0.5181185064917704,0.9549876323881127,0.5921321588127387,
 0.16871926262229353,0.34941342257320307,0.6803112735531409,
 0.5950239300343457,0.041354792570226895,0.43959748073993554,
 0.8261699048735187,0.5254683491989731,0.7095380506281911,
 0.015360865826943737,0.4877234463705309,0.8819785440325723,0.444583290823384,
 0.24124204183230447,0.10858444282688251,0.12998372994404017,
 0.9333177492018018,0.40854427138362925,0.2171334307213496,0.5289513390977654,
 0.173244407318488,0.2825845321923969,0.5889847417269503,0.6891824642473835,
 0.6929834525802646,0.8354307861005836,0.6362207091690774,0.08494505830918764,
 0.8826070790600187,0.7711672588972579,0.7145076832532935,0.2659828516615128,
 0.6340339248964152,0.8993079457826566,0.29885351056208553,0.8620294540364434,
 0.3658589570118077,0.8286160921253689,0.6435822433996025,0.3084859289185147,
 0.9132827471522518,0.6708166814182116,0.37808803643619393,0.9993772473645229,
 0.948738617170748,0.08452414415879062,0.13555192048488984,0.9050617552647403,
 0.6500980301021632,0.744858688761716,0.7210856592013036,0.11404058126399097,
 0.4862827577000829,0.804034852110825,0.4889866216780815,0.5479736716382333,
 0.06561318258505433,0.8580565382842072,0.14038556553742565,0.588774137466929,
 0.9551961450522584,0.24378648355425148,0.8094251791380949,0.3223285889327606,
 0.2477776339401958,0.9125230080258975,0.4416716774317151,0.42858135568822686,
 0.20350709319826155,0.0695325716760925,0.13826411356579704,
 0.20391625690968795,0.28198742610034555,0.44890113532170095,
 0.8279597555031389,0.4079049704257962,0.029680115028316356,
 0.025007358074802366,0.21531246811056404,0.5236116290369737,
 0.42152831545048275,0.1633105308888354,0.8653296352898837,0.6578515749901555,
 0.8597879272257202,0.17833263384885767,0.82943828108143,0.4952463761815117,
 0.8600526526837649,0.20852898243535267,0.9372879718856613,0.016889779589497,
 0.049417317383039894,0.22550090958082858,0.7168049090837507,
 0.6917812735954447,0.3807440122713196,0.18820966460698163,
 0.04294475827702848,0.954649121884003,0.9483336451669682,0.38789170219723446,
 0.9189647454972741,0.31179054693192265,0.8842665898490387,0.3260330064692286,
 0.3058913013952268,0.4144582285734396,0.13440235028468916,0.8933241249263275,
 0.05968578622229814,0.6898931455105483,0.975033190328033,0.5976744148635601,
 0.6661372451049402,0.4399249744362782,0.24924324883209525,0.9557269343781047,
 0.6940147157187314,0.7489493356278978,0.9816289897490645,0.8860291309508243,
 0.9570743081977391,0.28701252808983524,0.8972146882101302,0.5067016537008244,
 0.12139824235892271,0.42750091867542217,0.06070526177004332,
 0.08025799641129827,0.8452334203311535,0.0972171187468639,
 0.17513048788902885,0.14255943517638991,0.6941130689098264,
 0.9120729137848059,0.5200425605110076,0.5070289972266346,0.4847233137593898,
 0.3168084144889701,0.0767060543427831,0.6117662468127285,0.765261307339391,
 0.649335967993739,0.6323252484189874,0.19647969226686246,0.07614863347101697,
 0.20870165414771868,0.03043075806693729,0.04886786080637373,
 0.16294764778587134,0.6747365054422934,0.9482971009176914,0.8886573090840524,
 0.6264432873862201,0.786768820335296,0.6253389516801869,0.690470130498074,
 0.4716117076085775,0.2597262657696039,0.3289676031726063,0.4540571839848927,
 0.6458372942495287,0.7200298421279598,0.1584729320633741,0.05440415390936648,
 0.8011250864064751,0.32963270291772895,0.31295149351965124,
 0.13365925198624906,0.43859042371430035,0.1884445123788766,
 0.4495429545921086,0.5142343053425035,0.9515200598130189,0.8839391672164075,
 0.564392997818847,0.45188347179719734,0.2931311037480945,0.35478780011176636,
 0.5541620658372457,2.4781342736834855e-4,0.8666256220127125,
 0.9446776291263688,0.6169702331301553,0.38544615954276307,0.5832379885033069,
 0.46774269439222094,0.323268706766512,0.39200062736046737,
 0.09587937106540556,0.013281951793910451,0.06730014833133291,
 0.7126923423535688,0.16305644675370112,0.7503766321417047,
 0.38699038976018374,0.5813367966109806,0.5377963914657403,0.9528966060845764,
 0.4302055101991149,0.8903758263443429,0.9561252206337408,0.40030540057678776,
 0.2561591332101163,0.999968161843736,0.7670525136381408,0.07033826686696298,
 0.8633379958368752,0.2195029994405523,0.18760223184768465,
 0.46821727142235015,0.7708649471516111,0.23191053235812342,
 0.11900678720661306,0.5860209102498936,0.4078787829424334,
 0.015339595547319118,0.7560657828558678,0.23666994186726453,
 0.9913827409971014,0.8982798242916484,0.28927699659062867,0.8053256290888106,
 0.6568362507264724,0.6098053768606232,0.0965700705916841,
 0.010465529793126538,0.47549961775170446,0.8984707893318739,
 0.37174713678622273,0.190881438277545,0.24905278364111327,0.9864619974228306,
 0.6556853980484499,0.52163948320386,0.5005528035815618,0.9634038387448561,
 0.28124784995707963,0.5229942378933389,0.01383834991456867,0.859672408878252,
 0.5944609328810071,0.2680080573220416,0.5174731834983383,0.23944414418139048,
 0.714189627207377,0.28670151463027027,0.9638989253283508,
 0.029529588710350274,0.9219083459359001,0.8559565275881403,
 0.7843991432855705,0.8053980236187364,0.816697140901569,0.43270478386868927,
 0.5759585639887412,0.6766178826982854,0.750635064067583,0.6404911153805175,
 0.4069898457969866,0.579415357973388,0.4239787122192731,0.6414621510114457,
 0.8487981860049729,0.8112837990444774,0.8958273675720476,0.716631367713095,
 0.7271146506949844,0.3912759608029478,0.066333165649889,0.667516186720226,
 0.24876955168885417,0.4638534495996567,0.17936234556147868,
 0.3322700830843257,0.2802667445337528,0.7494535975594432,0.3861615208635695,
 0.17050002044620471,0.5699120139642918,0.04177788085101708,
 0.8441929473146771,0.9913225730170174,0.3951351662202577,0.7166757042063902,
 0.15897156859972505,0.11132828435332931,0.28878744387935584,
 0.6466340498569803,0.5998310534280071,0.45741453835069346,0.611329470398093,
 0.3796401329353394,0.3976937543781979,0.6799793086737059,0.7766900180050111,
 0.20017546062849256,0.3448041620644471,0.5940051804586275,0.535311707008831,
 0.3144969643208656,0.39806475769359906,0.029589549847955365,
 0.6956946171599695,0.6190350852599723,0.7626458247316207,0.17338397367172187,
 0.8623118287071424,0.10872048506781762,0.2771545735377252,
 0.39796227644249293,0.430048391848845,0.04339218663472988,0.9024526693518977,
 0.36782797143218016,0.485936779748232,0.595035866473186,0.3457153185869838,
 0.7626716978132552,0.37545168510351135,0.1666485673667989,
 0.06476835519583268,0.5921600577869616,0.1892159248101113,0.7152211899199542,
 0.11659541969629461,0.9016120338068502,0.1154282380382674,0.7953483238851213,
 0.43353391107114614,0.5989164734570398,0.8536130316376791,0.2012607190814174,
 0.08275850234382831,0.9935908452063964,0.10260278346664864,
 0.7888568671390845,0.9606285731648884,0.4954678876826464,0.20596377422265288,
 0.23848050808045906,0.09068921001507124,0.044322766504861666,
 0.5551482622478159,0.9879889836520622,0.42575267509030434,0.533093700713631,
 0.34097387290736236,0.817834241114836,0.8459999362378406,0.46313000500944423,
 0.783502846475811,0.009834344319121224,0.7141598325499773,0.8020661937857729,
 0.2885260980744834,0.8042857729099979,0.25269606244223286,
 0.08457857484241937,0.8967944936011057,0.4196482429415126,0.8859513951763339,
 0.2993230565877859,0.7139622863714289,0.20204262357632674,
 0.20018689978298843,0.36043631845623325,0.7777948723235053,
 0.38488939809798206,0.022902184162117756,0.3434430659825961,
 0.16835292971176008,0.41549506542966297,0.21169831546132722,
 0.07552426372570042,0.747193238594475,0.2328078733691018,0.30646042340838453,
 0.3162953323391332,0.667659896049042,0.3103749337245638,0.3061388741571275,
 0.7059878238111423,0.5223491353640848,0.934115196872457,0.8046716650320581,
 0.8632678522509282,0.8264645573374674,0.7341048500067627,0.7626983840889316,
 0.09078669574334841,0.8121451524485005,0.7585910409726747,0.3050134157650938,
 0.992746213002462,0.07011702365262507,0.7542118944611793,0.5611556102977727,
 0.9270940061582947,0.31619615215263286,0.5708379361843025,
 0.25323622635791043,0.962368692940509,0.057384311702641466,
 0.8509381794068195,0.33868482586019455,0.7977694825587608,
 0.47759351894473623,0.6759136967484057,0.1411643098939107,
 0.37086415963208486,0.2197320756150155,0.6000036886618965,0.2899177843591778,
 0.14018012067838348,0.05658644723655082,0.414465301741457,0.6733782730020801,
 0.8677983384581692,0.05385204737013749,0.42711494615087986,
 0.8827919634204402,0.11793099342989521,0.551144993437981,0.7379489558615119,
 0.2635980966684204,0.26403190609691973,0.7280098689953896,0.7710887145081547,
 0.20178426450789377,0.44081107749350423,0.7719369874615246,
 0.9227940788619794,0.4957812634753326,0.33420481716392225,
 0.15990046883987563,0.5377583075539595,0.2461241021376679,0.9103230983244532,
 0.0526890831278809,0.7125143507950348,0.3907910867291443,0.2476347229706506,
 0.4460545466945075,0.11360516876449522,0.8036410046685669,0.2120905135695692,
 0.25649640992477507,0.13094533120486884,0.9833283990983788,
 0.5539504743087118,0.45661032416859504,0.8609214261405095,0.3648124690074869,
 0.4795202222911248,0.25973917969426896,0.4187902169120997,0.4489485845972514,
 0.2913598416386334,0.4820967861142671,0.2100577243614652,
 0.045277910198344484,0.5191355404039468,0.04471759763274552,
 0.3145021789542257,0.4538453474954617,0.37758590488816,0.9688860305146062,
 0.9447514888562746,0.5600458976537994,0.7159072633298452,0.5528861862956951,
 0.16708901868269155,0.8181167341910391,0.654947570424211,0.787986578030109,
 0.07627480786620433,0.505672101555899,0.5764720198512899,0.5975181130429315,
 0.9619158719740336,0.3563735891455255,0.22623179331177223,0.5445490639093379,
 0.14012977598037435,0.5532789562284118,0.7956322685821697,0.3664155005315253,
 0.6582995487638574,0.9313570741467837,0.3029416633823656,0.7459442438055083,
 0.762464755677324,0.31909422616002203,0.7516440352448277,0.9774556017046049,
 0.42835923828171574,0.4355022063954781,0.20664590478875278,
 0.9978115970436251,0.9225936481793087,0.15986314066156693,
 0.04390123093097398,0.5798020949286379,0.07316787643658695,
 0.9587657364828245,0.24927842598322403,0.5045931764128383,0.8916368118154838,
 0.4928030927320701,0.5180476867239556,0.5192182597393221,0.37619016132999405,
 0.8739301915354559,0.596676490931629,0.596670219611912,0.6544272508805383,
 0.20269932825146075,0.5315740172341014,0.5665739899833442,
 0.14477216817522776,0.46929349948997745,0.12472594550740712,
 0.8944614616603828,0.09828669054487993,0.14682591217630037,
 0.0840306296060418,0.2465777708927408,0.7805624123726288,0.378428851251131,
 0.14848246206409743,0.24960842803402394,0.7831764942565661,
 0.41759932553184254,0.2955493912828201,0.006123983116760989,
 0.9093576076148635,0.6417010502851284,0.6069489425352869,0.5227221331934648,
 0.9232447611530219,0.17157267599894144,0.49212618872155667,
 0.16316790870064768,0.7271742004423065,0.7119207844056717,0.7801864408819049,
 0.2911304747918201,0.08839335298146378,0.3597012270448403,0.6750337423606936,
 0.21058831405173772,0.3496774922494015,0.6112624996429145,0.3907116591026375,
 0.977393150733926,0.540483438503142,0.8119296208474349,0.8482708407255142,
 0.18349868278579096,0.3952070598172335,0.1199451598069361,
 0.026595119235198195,0.28609878520295373,0.3481173677404964,0.85542396149575,
 0.9014528129109133,0.583635291087498,0.2747666161065254,0.6173746702185116,
 0.9757161850904332,0.021323635224831272,0.10705116544758131,
 0.7174491967937151,0.5501941958513797,0.005805457554944771,
 0.3333203111245321,0.74579044485064,0.8139213758297186,0.16226140648839127,
 0.13344814632934954,0.7530303770704068,0.9414308984880181,0.866052553525969,
 0.749618082811126,0.46767142173626264,0.05532210786032943,0.9574498487912666,
 0.31992780210515515,0.659135423186659,0.40710927027616806,0.9566199897921821,
 0.7117824105968681,0.3936895612832365,0.22439143623872826,0.5602498064005031,
 0.10182249218572159,0.9893906805931927,0.44841536822637273,
 0.26166417955654697,0.09492000615992391,0.9071342636463079,0.703244915964144,
 0.06308331233794662,0.695864055662648,0.22500689117949313,0.9806869939219276,
 0.9974809382655776,0.9418780293522093,0.4196300220085414,0.8874563345890379,
 0.8859662824015737,0.7208129281502547,0.27861653736517455,0.4535541351462393,
 0.21419107820376992,0.8164687468112308,0.44435592629510134,
 0.6976832985734496,0.854049234557823,0.23937927052493368,0.620633753612447,
 0.2984411180781501,0.41709602306132987,0.20480508687519838,
 0.9660216082644794,0.18019028838378237,0.7806163554342181,
 0.17791031640505206,0.5195382362616476,0.40793870460186743,
 0.6106131894744877,0.3470762855684669,0.03043083975853733,0.6588056317711387,
 0.3569193529235106,0.27131376860076684,0.7309700750313299,0.6970496436684012,
 0.9398147032981249,0.30794113947427904,0.028550232141851595,
 0.1770161991107928,0.7891617270704984,0.5183480916438239,0.8032950318458363,
 0.8305957055060039,0.4114033165600848,0.7355418619443478,0.5209172133297967,
 0.1110612322096809,0.5440952748224434,0.4870924264843791,0.4718808984495365,
 0.16218014785400603,0.13070973954583576,0.33886924041419464,
 0.06218784855814419,0.2345091590601731,0.030967204754906996,
 0.8047160443350374,0.6565278813612204,0.6942488189791313,0.19767438177763852,
 0.903025321762807,0.3723903153434358,0.17681927438716905,0.6244885114864674,
 0.3573519799356273,0.2633630295944265,0.9201841954534854,0.2787665250250885,
 0.972090394399759,0.0026602828289349345,0.5870354436736701,
 0.7358448083927683,0.7083480185062306,0.5538308976939434,0.9561627414769353,
 0.02699526131555363,0.05941352464599259,0.18203525098327822,
 0.36986072512740265,0.30928062792426136,0.20202584994690065,
 0.212313829388593,0.9180436637605474,0.21498208386991013,0.9884797581596337,
 0.09412353857772793,0.6498653464703363,0.7916232969235836,0.5006586963648993,
 0.490306057779611,0.6560596141137041,0.8047106500188521,0.6237353954414165,
 0.4437111510767333,0.20308944011107355,0.31980071782999997,
 0.9025205713834645,0.4969086957783224,0.6817474662524201,0.2547239903549713,
 0.2225009612078882,0.018196189681372377,0.3665489520886738,
 0.9263201333800729,0.46273904113389075,0.9822176032419129,0.6282870782971293,
 0.18536144970672097,0.6092675204298565,0.5578969061787897,0.2556029295726716,
 0.3491216133291841,0.5755369812772083,0.4157195258755727,0.562511488287741,
 0.04198758401091762,0.08788283280501963,0.7048327262658315,
 0.18153267678345453,0.6790051073570501,0.8064149770510927,0.5586562702139032,
 0.47238692138840144,0.26375165585207616,0.40541285168303465,
 0.6680704905157012,0.8115424925153344,0.40081599492007636,0.8679558809910056,
 0.40828266263146085,0.6246862186530189,0.6325685667293574,0.9015121157818619,
 0.876038140672572,0.9823639026411422,0.8798088848711958,0.2972980339077208,
 0.21781631479366836,0.1588666429013329,0.8982089479283059,
 0.41628815244558803,0.507337342595957,0.7602952717219956,0.6986434606722577,
 0.17016153530549727,0.6976022089730399,0.41882386894312607,
 0.6293125990961324,0.5808294276767931,0.12362509331734128,0.3075815627702826,
 0.6608379128980173,0.7041057837550428,0.9132486807365263,
 0.045104599242808185,0.4407326084907204,0.9898851303972503,
 0.42237159073231434,0.24259831208128202,0.9572273401918874,
 0.37738935513685856,0.18291102322869401,0.35231568374878697,
 0.5406666799969655,0.6113811758993847,0.4800281249895677,0.3675046787609857,
 0.1335780227486172,0.0482402963240236,0.6342912866151551,0.1777440487502041,
 0.23117490960098475,0.6039681180118784,0.9188179572787584,0.9714898578813888,
 0.34767331187614714,0.651906355566882,0.7398024244223076,0.06673161098050329,
 0.5803760641986194,0.5029048634933339,0.005059726663442454,
 0.8175494659345892,0.6408275537190604,0.15812293471918415,0.3435821706315072,
 0.7940087727293303,0.5573649618283225,0.820968514782952,0.7658709383267093,
 0.9661684341317237,0.11113504078436376,0.12097632833233474,
 0.5265090980819765,0.5136942187036102,0.8757395801276455,0.7555012106512722,
 0.7129256066856907,0.49138940202837667,0.18915265736416154,
 0.3037570279466695,0.20792679088197163,0.2934241495391572,0.6202077903870834,
 0.17787858526707734,0.8687523015473161,0.3347589935032321,0.8801744312757908,
 0.6841663629453059,0.7177318264949799,0.3649303603346302,0.3914769010747605,
 0.4158609215758956,0.21874704991281746,0.24110677470215758,
 0.9693059750576859,0.8009622838391143,0.7350284103218077,0.7763015136863348,
 0.24224738497634035,0.4926985036357773,0.9105505046185183,0.9607283871590728,
 0.9982554373448865,0.8585877071125606,0.32233610180736993,
 0.09435985426996185,0.5866736884487783,0.7015210729700962,0.1703667092853255,
 0.005251377061122997,0.8111941040421257,0.3254531560342264,
 0.10545639731143308,0.6706096154924587,0.8912276847163263,
 0.021956169562876715,0.46497543130440167,0.851930285616437,
 0.9563520363023155,0.1724017535042005,0.11943174492846287,0.5566647616883195,
 0.07965268253473812,0.9308864390595626,0.9203026992995336,
 0.17576497608646213,0.13701312009661581,0.9571200251626129,
 0.8000761487355745,0.681538344644954,0.6648318952462309,0.5440687831627574,
 0.4781869801238793,0.8168385744885476,0.39036411278750716,
 0.21285055818163978,0.6032811076493434,0.48220960614118524,
 0.27107934802587386,0.7359075308582652,0.5125297146422152,
 0.48863911266310234,0.32997962607450815,0.05770582513420186,
 0.392353629423414,0.9815958815478482,0.9432658611340495,0.033787546512528466,
 0.7562375807326173,0.04456002236880874,0.8319050045404223,0.4389634543671881,
 0.6337018629349186,0.3079757464074986,0.9725269359493021,0.9252885178305319,
 0.815524310921291,0.23197426164390722,0.7922035372072262,0.7949993453206183,
 0.03267610390029507,0.4159250535088155,0.34592622066777934,0.675774700522711,
 0.3471311040500593,0.0610903295582198,0.7237350651870676,0.2565139544841003,
 0.714472980696804,0.8323549119933256,0.5424464612876949,0.5167906581784232,
 0.31339890357812306,0.8534804746234592,0.9801751424004298,0.7136284021621182,
 0.45970696906538766,0.934868242598697,0.641295579100662,0.7011343478695421,
 0.49980720682650936,0.6657381642999451,0.4615958877454688,0.7907801099931435,
 0.5079956326131324,0.08991281085687053,0.40209989905178245,0.624045121956628,
 0.43990008637432143,0.7649105673648087,0.4745295514519997,0.1626916178485186,
 0.43541097213465363,0.6726209012549337,0.9853001392094621,0.9567441034258889,
 0.7593889349114203,0.4315551387419144,0.2948332328174105,0.5559163371134386,
 0.22066223414816954,0.7909508697925592,0.4971852395296904,0.9604699966528977,
 0.885049363921165,0.8386309858358268,0.1704962525970447,0.32781493252486804,
 0.944471860452358,0.15338678727537192,0.5512281088233579,
 0.005846681902169637,0.5987263830966567,0.3352252936547604,
 0.5075521973388414,0.907168348756691,0.21856698868889324,0.26158746400122146,
 0.33443674088851805,0.7318900414996686,0.7152495298998097,0.721935138880988,
 0.6677090674673374,0.5041046890015586,0.552731668120166,0.3348285446468602,
 0.8988496044328071,0.7014158151336727,0.6360698980148878,0.3534607825312619,
 0.09018757246058706,0.7717066918953975,0.35872371051550334,
 0.32995643143009257,0.30586766399315557,0.7876895677829048,
 0.2746354149404122,0.5094489337957552,0.9694547426452806,0.09411750411152853,
 0.032855477614951,0.02139008171849155,0.3342928254146854,0.922198193588721,
 0.8623848005223816,0.08814311099497896,0.688150583627146,0.09290253302931328,
 0.230348670564958,0.9523102737064871,0.2806917336454139,0.9971183233560278,
 0.22293520709452053,0.8709989049471161,0.7707618159828022,
 0.13202662105863738,0.5191952571012692,0.4278241646337031,0.8490759713528628,
 0.00744399021899933,0.7486257629880871,0.022353188592891926,
 0.28669593955358263,0.87068876566277,0.7345822155267352,0.0587186328717727,
 0.9487917335912043,0.1835197443248907,0.8524666227946367,0.5720723116027323,
 0.2397171113746024,0.5171801838113664,0.1886236290689033,0.23903385424458623,
 0.35352856020253987,0.7626324335997948,0.9231989129628041,
 0.19023670937757842,0.35197014359066914,0.9303329565206994,
 0.32853589152393015,0.39694504450284596,0.5786185094701777,
 0.6625643957272047,0.31840965652195696,0.38266421057497424,
 0.2888103923610297,0.018169902691307138,0.5602021454045963,
 0.9849401470997762,0.6419300744585721,0.7495216811480011,0.4688162031498859,
 0.7948018274532184,0.5310539221671376,0.844058294353023,0.6808060569053063,
 0.2542018587006264,0.36024913983721296,0.26375229936606925,
 0.7647455871741338,0.2082329552829293,0.756551637023448,0.08759214275291072,
 0.581502613780689,0.0839582366175301,0.8400423134821381,0.057051883340812015,
 0.9639922092891456,0.38597720446770456,0.08039973969251801,
 0.6875310579044301,0.12305127718719822,0.43226248742719375,
 0.003055933206538186,0.29360773098285387,0.42565507024243043,
 0.06570261914420572,0.2534762249890834,0.28909586896418205,
 0.40497671778744915,0.9910384470599134,0.8721485247872476,
 0.37168621354071985,0.040735158737974175,0.46861306510699263,
 1.1864920691562197e-4,0.39013712838624537,0.19106515034483718,
 0.736617238039044,0.1342544055527196,0.4643127624595552,0.2553118908239016,
 0.763798888426189,0.472018918899904,0.4236025184001124,0.8792254388231268,
 0.7347048363356979,0.911413818669615,0.7645328373891376,0.33195128221911097,
 0.2585457018205899,0.7278399056487616,0.4589876428304692,0.35784270108914507,
 0.22614144889322374,0.751417865836071,0.9447261337207251,0.7683510025851057,
 0.05775680916389736,0.6144180835311539,0.6967540755227117,
 0.03564840891852983,0.06776072727977445,0.8940513875931972,
 0.7258402366042788,0.7613796645355336,0.6948564415375315,0.2043744109267196,
 0.8022553120790052,0.0677924600546218,0.9467290043910965,0.24677561236215073,
 0.887760506386779,0.8731595579800334,0.714134532532175,0.5111907077596347,
 0.6573234273758484,0.6412013839010134,0.2805744311203715,0.8459822228746864,
 0.11973326243837501,0.7282877020453482,0.22347898447245862,
 0.19843753238351436,0.14372728183645767,0.3274582840423761,
 0.9859288868852273,0.464088035042044,0.32037330493258853,0.47249213172627624,
 0.9026686816873064,0.4076020486410562,0.26052975660874367,0.7942534845060697,
 0.4692136151037902,0.5825171268590472,0.1286694992593923,0.31892650742631556,
 0.292726365654429,0.04269914886393289,0.17911450090316017,0.5817628526504632,
 0.596164798968761,0.40675972922820425,0.1991271629051894,0.36630569289099446,
 0.5685560772667224,0.19346162776264986,0.7221391923405266,0.7523761031847849,
 0.6659535301411899,0.38588596649449847,0.4007673049870424,0.6421460322990881,
 0.6925275270571494,0.10197202760152868,0.5855483610701459,0.530451129477801,
 0.391349139703574,0.41850713007883233,0.9420726665928545,0.32868223887149817,
 0.2114296951531125,0.4487671042524435,0.6328753737063573,0.7289856479256737,
 0.4756849461697181,0.7707216793557191,0.9133076115347125,0.09474750782083063,
 0.1416443789499081,0.8301983541803373,0.06344150120766956,0.4475555006479206,
 0.7533529274532124,0.44585889410527024,0.29558527723530903,
 0.7651057682220295,0.7856566357900645,0.4443716296461552,0.9696473989614706,
 0.14985279700263243,0.8995305594902686,0.3608225488706185,0.354416660767492,
 0.5190262603668939,0.572180648357561,0.009132167189835783,0.3289801030086399,
 0.435059372747572,0.12604196794902756,0.02024572898923016,0.7090539209983058,
 0.3204992028249142,0.22487747802944547,0.4566593071828374,0.8978487583641179,
 0.5854398612914178,0.21483140491126984,0.513068353150058,0.751948581375065,
 0.28368883039426096,0.10172190776812218,0.874647998044855,0.9962263874672497,
 0.44769053012498494,0.6013009166708136,0.31514118444166417,
 0.07646832128791459,0.049408274352423565,0.38773687823649483,
 0.05563909678141932,0.9095387061075356,0.5877386993719009,0.4738524324418283,
 0.8255104560219957,0.07917892124121662,0.9230684946875973,0.8683995761963209,
 0.3684524351981414,0.7222215598947975,0.867156682475252,0.5205954066562006,
 0.8040805821502977,0.8190689382123213,0.9519655045314652,0.39104871514628614,
 0.7433963347539737,0.8718427604137587,0.08175487334559384,0.5556568748489497,
 0.7394655884113417,0.7774485436133853,0.5475411491160971,0.269156513702815,
 0.48547515023183385,0.7333682791258107,0.060625018450278456,
 0.9630409460341701,0.6745416945065885,0.9254269756190273,0.7191851900190779,
 0.28114173905799855,0.150053304750276,0.23293583622179725,0.6540199220738622,
 0.15339910520515188,0.08625758583262899,0.48966215700566607,
 0.04068463860997951,0.3722284464636907,0.5980365810890457,0.7643199979796286,
 0.7820031520806318,0.3120950085490448,0.42137778350869026,
 0.40532552766349905,0.4500345442471303,0.07797299151590609,
 0.26504023269742005,0.27013426921419315,0.30006277764059386,
 0.5556417379958256,0.24176904618019224,0.5691999296143936,
 0.34846046778219786,0.8634503115262604,0.3895039874715511,
 0.22380079758766847,0.3229465514232347,0.3040657369926332,0.792369217597396,
 0.06692643637342377,0.5898837545776285,0.5816074336817607,0.999728120989373,
 0.4618241209664372,0.5362889013830924,0.44313869109663506,
 0.04679640981464317,0.4577470004645129,0.7544615432838705,0.8892475101164844,
 0.9418952790563073,0.03159533017250904,0.8311932200590484,
 0.26244389120190814,0.9871748052547253,0.10754918060178165,
 0.6159067387828777,0.7528126941502769,0.17210926241639202,0.3691116516054792,
 0.6899630750469619,0.7185620506383722,0.9257058021007886,0.1359836196923424,
 0.37633260550553893,0.1749956445852875,0.8480345655711818,
 0.36963566549768057,0.46340143103330433,0.36117999417948,0.986766370721234,
 0.9375156749668674,0.7100618824679152,0.05287334285863243,0.7513651282985976,
 0.006326472627930624,0.8209102022187411,0.3045401265705041,
 0.024222153577797445,0.8592274201508362,0.5590800794231938,
 0.11011982732166603,0.9758744517626747,0.4117295857609763,
 0.0058174469245276805,0.10652032338809925,0.30891771488504816,
 0.5527652819361666,0.7864762446875632,0.04673361076715754,0.6295755460126219,
 0.017210586715638687,0.5088659969990978,0.6782950238432132,
 0.3257294985659167,0.4052783208184805,0.4935428134965203,0.1991614186311621,
 0.9081515548201072,0.6982673113552091,0.048169693768131294,
 0.7389569365742363,0.44491615217770164,0.49658250683592475,
 0.9493407127744307,0.02832921837055713,0.8957192397412562,
 0.10310298174911059,0.5293106575614711,0.1396144269360413,
 0.01506742449165277,0.8888739925166345,0.7710116377305724,0.944724371435225,
 0.6390440407722806,0.28338961488740966,0.48695527713211784,0.869587728982,
 0.6488780189857206,0.8095720223508043,0.9393160691369443,0.6127018725675831,
 0.29593615794783157,0.459421649802359,0.04071379750433557,0.2786308373257107,
 0.5581983273489157,0.6225784319267826,0.3260159367894957,0.4328221914178876,
 0.4443224388549626,0.5368294596154263,0.5653656671423444,0.41634852108771125,
 0.8432037370823838,0.8777499488820287,0.2679319607465591,0.8347711530947594,
 0.9049009392108683,0.061874939833979736,0.7180373568133743,
 0.8022987442379221,0.9430245541879738,0.5203910085843622,0.8062506564046811,
 8.764712154429866e-4,0.46084885704687106,0.7127608033493953,
 0.8656285724393757,0.2189404140354989,0.1871348948965037,0.8028316382397281,
 0.9570306739920575,0.1591125954774133,0.1959716325333105,0.9090158584573486,
 0.9644746200544345,0.9240997340598728,0.4015999513572749,0.47583287747412717,
 0.09169393165352491,0.1748304761420718,0.72863182706147,0.43103264994231594,
 0.22864382130611094,0.6110573866033778,0.6027040383975844,
 0.16179775550626607,0.6981092123909581,0.08119306482370925,
 0.14614810747588314,0.8450794672293056,0.6547759908763636,0.7720131553017409,
 0.4786203910666763,0.7632436307589536,0.9296688011127465,0.7520564257521106,
 0.09045478289928677,0.6135502945046339,0.7954247310976255,
 0.23234281543959334,0.42025145287316323,0.9614769250361008,
 0.9432462908153335,0.01255436147935085,0.7221484655046265,0.7415106051065469,
 0.9428951377458175,0.9546090806035885,0.6579267858683211,0.7139569931221463,
 0.5168649893558711,0.3293152295010129,0.8457087172650601,0.712341621256672,
 0.47483020506624074,0.3132441334455702,0.3077983710880289,0.7161634878412365,
 0.30359326604095316,0.4937628050780991,0.07036049323154991,
 0.17583053611557298,0.6332173631963549,0.5177418385050097,0.1205332819941558,
 0.48661485500155754,0.6639995832861665,0.9747873368120733,
 0.13146454760223558,0.13609027811514163,0.025822031742989138,
 0.8225282969552552,0.20037728287843448,0.9226990271859175,
 0.15940829340931473,0.6703362822726072,0.2460108429115161,
 0.20057476996403245,0.7161862838442381,0.29694477112238893,
 0.8682100170500585,0.2830937228257453,0.07818049110215375,0.5411847404422244,
 0.6673159909647778,0.15960496471016883,0.3011141071012532,0.7771793589555033,
 0.45255750130223305,0.38466194601163284,0.5917771394045905,
 0.1135383168438584,0.5572694624486138,0.599202719774094,0.42027331250539024,
 0.5546431734174846,0.7788412285724648,0.03373291918253274,
 0.21774444151493655,0.11779059603691344,0.6064284586252828,
 0.28547604373584323,0.0665621454365053,0.2545773857782724,0.8722027478260591,
 0.9836552178012056,0.7259019905931153,0.7156198075407164,0.5110776366405937,
 0.5047322145756038,0.5794606346316842,0.6658488538279238,0.11893006345784973,
 0.6710953360857275,0.9472924230357758,0.5093413425105777,0.8202317311751227,
 0.02148291889796028,0.6036663905709321,0.16211522540122814,0.396239043335613,
 0.2214135698172296,0.2658720384865878,0.22063072023045494,
 0.22901667026023076,0.46131229149429576,0.5226161588622227,0.51928814017407,
 0.7231644706648206,0.6902759679258268,0.7285409433856338,0.08497110089492865,
 0.193506636965044,0.6126765174623408,0.9376101654764868,0.5817618932946625,
 0.9529094750970986,0.5261940932638067,0.41208882660947077,0.8292640748124686,
 0.839856839568756,0.23235758477164548,0.12960772466005865,0.778641894649617,
 0.7413478666185176,0.5854121135644828,0.36303097039739185,0.7700324168377451,
 0.09856391076573368,0.8581166594014223,0.15707596763128773,
 0.21290178141821114,0.10670099364128427,0.23512378623940644,
 0.8202068505574135,0.611274260838908,0.9004162053230796,0.2073143811692556,
 0.7147217505707669,0.9785206183255538,0.4325826893015844,0.8119481384068935,
 0.912101941791842,0.604614818059127,0.7203843933314562,0.9932688329572029,
 0.5329531717471393,0.2502696666680203,0.5017297602906979,0.6918513809636418,
 0.07506604099671033,0.746365095022349,0.34636783867731946,0.2604337892136246,
 0.22469494587178185,0.8021249422285562,0.5206881347845789,
 0.18631884742284666,0.6196122675903766,0.41420137445722616,
 0.3966920019650727,0.7887772759525066,0.11385326802147466,0.7633969950638755,
 0.278103873986987,0.3799612346431034,0.9356068151897815,0.45803813024571605,
 0.9270005769446272,0.6653734712655552,0.9431007979701387,0.10114352339306487,
 0.5335015392607535,0.3971084337296562,0.6705167545021109,0.6216577462525819,
 0.86456274442501,0.8810245832968865,0.4493083780288927,0.2405795370529229,
 0.9522929930236729,0.031917902512862195,0.8033205567427848,
 0.6378438439740951,0.5829618162172274,0.7134781165519015,0.5935775968555358,
 0.12660488471338127,0.8365192062627445,0.6385883065117328,
 0.013476069197093987,0.7661292523455562,0.11377342903636767,
 0.5516649023579921,0.9883422703516496,0.5553985159276684,0.06267754976656026,
 0.4306100829859051,0.6916053491505725,0.8230635813290882,0.13782311856779472,
 0.9664505803511898,0.46170421123049643,0.030459814222769777,
 0.5208590775650752,0.5965202696566516,0.9739533303888525,0.6336465309941149,
 0.06909859096336213,0.2033429345788078,0.369658556387261,0.23636711458384074,
 0.15762233703159922,0.045120091008206686,0.4445813086108493,
 0.9544718936209462,0.8771392894362064,0.8831889439275179,0.7219194125345922,
 0.8568594075609438,0.031938124548716806,0.7122706671843836,
 0.23419564385105873,0.4574839309149875,0.1234224929449359,0.7435645364971438,
 0.5529523248726872,0.9514050643801328,0.3543270476957474,0.2712829433689605,
 0.4382799717807375,0.4917811872213953,0.06873581135451534,0.7435331985074156,
 0.8320303057121804,0.27155035138438155,0.7007598594915052,0.338026656602074,
 0.04445048568957022,0.06009847806654656,0.2859832432201007,
 0.2830281015021946,0.9082907590759121,0.3930770680067627,0.06629757152055027,
 0.1129622781194497,0.9475133911950041,0.43071913028553555,0.3513291536292741,
 0.9182519480659275,0.2131577301382911,0.7879117088119422,0.943722273353366,
 0.8077348203958096,0.340679814187411,0.9089025744982048,0.20187748000178019,
 0.18754093751543277,0.4600531533053086,0.11795726032211129,
 0.02886345390701739,0.7993818665129971,0.6391358033079769,0.6906309153123732,
 0.6779747456976328,0.4443548160012162,0.5414352776540007,0.35832877246667627,
 0.1656570038149071,0.04961227971589244,0.2334773942782813,0.9683046772422665,
 0.02842857788482589,0.9601340687822913,0.047713402935446125,
 0.2585469426770244,0.6857417580089086,0.4350150734385957,0.411813662541905,
 0.27731858140457,0.9004068028210904,0.7194400802035593,0.44881422876926846,
 0.2615165348015056,0.721987667789184,0.23886123449067165,0.5190121837498011,
 0.1892079448053955,0.5654884952000876,0.4643805749138966,0.07340365338181021,
 0.011426879164786574,0.9289883509924627,0.8901798407363782,
 0.6979967743211649,0.40990595385086004,0.7590443073666466,0.8093555470521431,
 0.0251646569169649,0.7659453877824174,0.36502072252265694,0.5171813322454751,
 0.36353970778546885,0.001021385463871205,0.44111064525883903,
 0.3256045709796638,0.025429558159369647,0.8297395642009889,
 0.5202871944878733,0.9923200655758014,0.9888646494257083,0.13279476466683982,
 0.6558148391698929,0.5523201117664294,0.27875132072455777,0.18971226626673,
 0.329655657896072,0.7087519661210288,0.3189178646868639,0.9694054324196322,
 0.49518158796295486,0.32375918714155916,0.12653596525102562,
 0.11531042058871077,0.12364972490896897,0.26414431391711124,
 0.09358329464479831,0.6204555803495888,0.3150954195864496,
 0.48133068125232814,0.6704918252789244,0.910883904023046,0.02271083410498842,
 0.9485217158406594,0.09947556969263249,0.08353686804649663,
 0.6434599351681395,0.4544794893847637,0.824978211014639,0.7949647429620622,
 0.9099088911110322,0.4303060347097424,0.07885812105665235,
 0.09391063328768812,0.9492295709808137,0.25358728574638256,
 0.9788387101008907,0.8783383296019108,0.08043751813751565,0.6939105933140078,
 0.7258109131352455,0.07532128764657586,0.17903119404070156,
 0.11414288673160411,0.1774770886408812,0.8053091852070571,
 0.025595746100505967,0.2513559267437151,0.2627157814139681,
 0.8715103474221136,0.4731346261809908,0.7425433467943116,0.8171751943103688,
 0.5574069242457106,0.5741150982496247,0.04589480488121633,0.9095993050754505,
 0.23324066650764996,0.23160477778777822,0.11751192792555132,
 0.8368655551669415,0.6801547759542104,0.6061953873176046,0.5900277069481774,
 0.9213106635798386,0.04344040064251509,0.752743272219504,0.2761228323825743,
 0.9874785955376645,0.06526740412166698,0.11239509736048137,
 0.8553337375766199,0.7474049676622365,0.47648867781687776,0.7410465264728999,
 0.7030594763742496,0.22518005150587894,0.3417745187637501,0.4600863914118838,
 0.46188801707150884,0.015777831740014903,0.7103333233053271,
 0.14598015570734935,0.008109982484218237,0.43230448722408066,
 0.04561943887695474,0.7124743068721483,0.6742410832705181,0.7478729383141595,
 0.17283629218654395,0.38469485129795333,0.12171211702455187,
 0.5048545706012002,0.9114144328610185,0.4152238940899723,0.42417941079590027,
 0.014015710603818898,0.790196475001062,0.11974929424796843,
 0.13033089810697884,0.42214881248612746,0.3157263157699116,0.309108766398668,
 0.8376641980258003,0.4640415878853392,0.44803147367429386,0.6181790338467448,
 0.6487570621356555,0.18506874057664924,0.14362050799462622,
 0.23470259070923516,0.7110171083434833,0.24506540259320908,
 0.2379598258881721,0.3304461726035276,0.2590453298682208,0.7529566126134314,
 0.1670378963236604,0.8196067970059497,0.866536574942241,0.11532143751149004,
 0.6537697821524913,0.7005682588396095,0.3715090126980538,0.06527068047547635,
 0.8811539983051173,0.6064566999134158,0.15002533699801224,0.8218769300612656,
 0.758448998355392,0.2460251469913154,0.8252809599304112,0.06782685426761337,
 0.4222479596191131,0.3641725725861965,0.044746769129715314,
 0.15364188315956706,0.16824090939048053,0.9286878258653063,
 0.5127542941394307,0.4644916352135461,0.8675684171087787,0.2809029182517073,
 0.6074782831907817,0.883612297125329,0.9865390547417183,0.8135377903919693,
 0.3394609790547132,0.5240673837328256,0.8736151449948522,0.3515884131586823,
 0.6759703665294644,0.897434038335163,0.012817157886792274,0.969994375460384,
 0.8699352051575684,0.8884311275283205,0.5898454654434524,0.7883509090278671,
 0.1376163600377358,0.4969607723105939,0.5771792982797045,0.11584441903195208,
 0.6805397662036559,0.3251254599260923,0.9114150262146465,0.9192317428231299,
 0.6434607096314734,0.25405266898113354,0.515319672019517,0.4187797588244043,
 0.8950573062839396,0.9951780972089902,0.5987348080950626,0.731640320695988,
 0.922641891655385,0.9761741842445975,0.24726066790318746,0.2998354814728892,
 0.0832518738857313,0.1479031143212406,0.5549662043981205,0.9985573699579491,
 0.6196768884802144,0.2657859241794386,0.1408064361586412,0.5529374452105875,
 0.8030582219072033,0.4029348819753966,0.7603498779542885,0.4178571580386189,
 0.624837756450958,0.07475778668077204,0.3119806673200167,0.8351226824719788,
 0.2754865248908813,0.2753185452221756,0.5426980014612353,
 0.026171823454436804,0.7283324390491193,0.2864340363386475,
 0.5007204716311644,0.14780542507182182,0.1642279683824437,
 0.20123173554004847,0.020708307575142193,0.4087071030761409,
 0.7202038217982857,0.9380338741498175,0.559735821472941,0.9013938880258939,
 0.501499637326823,0.7231190232900357,0.20216954332928816,0.8728645318592854,
 0.00754716838313263,0.59551475329323,0.45410590439623644,0.11525328879136243,
 0.007487942459126318,0.039200967780771645,0.8684315546046429,
 0.16897124655259022,0.6749355370392632,0.5388285744281943,
 0.22102227874917724,0.7035555670910287,0.04487748681835846,
 0.4710576735672247,0.18383631441896786,0.7495258132796984,0.2963514347258569,
 0.33021278452575586,0.5263749239166743,0.9675173544692641,0.4857547838850971,
 0.46661886073284786,0.9938311346740838,0.3803960100265016,0.5173815454561453,
 0.1044579715267333,0.10677994403121716,0.09975937994620554,
 0.5279263732384548,0.04903191545114849,0.7046543087086807,0.4356036195775621,
 0.8191229494620021,0.3918746390134258,0.9224538699802818,0.2708654127997996,
 0.1280196194405807,0.9391612543834591,0.25381866365614414,0.3402925969525482,
 0.02406928541891873,0.16208892979944345,0.19081084096861656,
 0.45441970752501437,0.6817912480800942,0.29684831740725404,
 0.14075868255472546,0.8693508223620714,0.9321406476122034,0.894922818438323,
 0.33552381820027266,0.8273419274870204,7.139008967596006e-4,
 0.6109673509744163,0.11698897375931805,0.22317491299391312,
 0.8190163264480368,0.4117490899577295,0.09840011209130128,
 0.20520318861393205,0.07241057962033803,0.10392659288405126,
 0.5480239776566522,0.6610708102554015,0.4621882893500445,0.7145990879253964,
 0.47927163639802306,0.8508729267590733,0.9586125619499363,0.6196687310351604,
 0.25345790062602436,0.8842955260834073,0.890247214932889,0.8477995077956155,
 0.014730074531779325,0.9545957577114846,0.47639168284105793,
 0.4722244694342469,0.7150280289668645,0.03859167122138574,0.6065738339932766,
 0.0483623152429149,0.8685074675741078,0.42891857966400604,
 0.10445069951961783,0.3144182475054169,0.22706566500030623,
 0.22867952311770734,0.007676424566652851,0.8504389189962679,
 0.4549168719400123,0.07733574539685928,0.5443310804493526,
 0.27934296215129306,0.49472846606640963,0.4462663851378832,
 0.04555883370444358,0.5835489710795536,0.8026849046018877,0.1685231168117136,
 0.9625589865800671,0.8005108705099112,0.7325002152119389,0.9595310341251924,
 0.972537552707147,0.4946258220947841,0.6612500766201906,0.03174179581286107,
 0.7939520882303188,0.19127910891813427,0.35979675529776534,
 0.16894667506266303,0.9855201943220215,0.2115753514017542,0.2834273505244723,
 0.8780783678690841,0.13086787950727508,0.07974808950228873,0.30318368710666,
 0.43267370103209324,0.878320599718189,0.5218568785043407,0.9401943907451258,
 0.11590418879449005,0.24208512850829556,0.8484163733891087,
 0.9567236829789838,0.5806571867270578,0.7319531424497037,0.8128194491747943,
 0.7267645298217644,0.26621425327733517,0.7825931777708114,0.8901251608803789,
 0.10545605466573593,0.6270331594512272,0.4316606223928666,0.4620065663391608,
 0.12750967771123456,0.10707816364111522,0.4512587816711129,
 0.07139154339348386,0.4360099684729598,0.48549142730302064,
 0.40273100557780295,0.780240679697554,0.06261358070819778,0.6790997115177634,
 0.4592078439280324,0.977293067359846,0.283678666164192,0.7258035288492779,
 0.6045133316908793,0.08856579825430544,0.5948870683342327,0.4504430922137447,
 0.37544072761690594,0.5303841911709012,0.0792615090610801,0.4003989744499131,
 0.21339749357598958,0.5654986920645871,0.7204141796184831,0.546595276364067,
 0.1253594531602703,0.5557408089060747,0.8926439552702982,0.4224666977537719,
 0.1781261611038497,0.9625282975309357,0.7600527901599912,0.30149269995443195,
 0.6127505211050296,0.6306681186725398,0.222766447308555,0.8718258515946898,
 0.7275029021679897,0.5654895431640029,0.8344801500369011,0.602305099722562,
 0.09173257908908172,0.6559984154047207,0.1794251951121124,0.6097084518989208,
 0.8523314046857388,0.9698702978814828,0.5377429180140254,0.4001882227497152,
 0.28878425793913143,0.655357749679955,0.6959118115151091,0.12175110524259092,
 0.8947748139532815,0.5922168558874151,0.13882375771996258,
 0.45463089342771923,0.3042811693978552,0.9338043075947307,
 0.13060211224487817,0.31206766258493657,0.9331742874355381,
 0.7465845916873496,0.7104874672271831,0.64356288041793,0.32535126615192134,
 0.9432827972738017,0.2176676806331379,0.1301580893361287,0.5711758600437398,
 0.9444809090968944,0.7379827571949553,0.18793456986654655,0.3127377790928694,
 0.8835167330544643,0.2536612611438094,0.017355201266506692,
 0.059210651561217076,0.5200271811843703,0.6691301847056668,
 0.5036253855382267,0.044042580656092456,0.05013431756331843,
 0.3107559158846105,0.7888308829307208,0.31641288240045184,
 0.24339237063976993,0.12181589730500841,0.1988568388808235,
 0.011797317110873062,0.6025384955020158,0.3965677402965938,
 0.7258133680972729,0.5902600084209322,0.9391254324531143,
 0.019064558259221154,0.8392805267941293,0.5127604888679531,
 0.0940795328706675,0.05429487955537171,0.1815827341605949,0.3776289398651512,
 0.5093622717144725,0.15115997052309327,0.7554128816952597,0.876081214814681,
 0.7419771098282867,0.8084741798897535,0.4034418716962779,
 0.006470475728870051,0.2615675862804303,0.23091052538343226,
 0.940112560644746,0.2501784587595479,0.2711752905089727,0.8361194366498947,
 0.8270560831361131,0.07108739939251218,0.5507852529646743,
 0.42689012533290827,0.3452028486239853,0.882606171643845,0.10955672910566061,
 0.44742984432756094,0.15426557095357496,0.11946474670210716,
 0.7346387420728818,0.9418449464005036,0.37637933024244896,
 0.05022001187997094,0.3427286341441762,0.6165961085105869,0.8251368961137064,
 0.04359648629431234,0.6349998878712072,0.5804530751023689,
 0.16872437340201796,0.8783909928376955,0.3488161800042846,0.6585273881221549,
 0.32195728614648067,0.34851248895103426,0.15508281862289175,
 0.1799587477839003,0.9183802010851339,0.35275138555248675,0.4627036435072117,
 0.7297326941371008,0.9402401885518837,0.8727306250338608,0.9842907197627588,
 0.3200865147015459,0.17626224075047758,0.7199297779196319,0.1411948204310456,
 0.6587961543416965,0.6306825689764418,0.3176683259577515,0.601555527358565,
 0.9474087687320204,0.19982150039619762,0.7900973086256852,0.3827155088169939,
 0.5974348373485342,0.889590405514061,0.18754087365841032,0.6354882922555746,
 0.6184362021470505,0.3798735303578231,0.2917961735550334,0.40817245717629813,
 0.9548276184571102,0.567589841552043,0.1887350208131925,0.7941479313334461,
 0.4414265955945196,0.5987240222946666,0.5010010639586682,0.6411223249931327,
 0.8401700037246168,0.7037340986119934,0.6887545558127187,0.5177858585055568,
 0.690706279372654,0.6122173022341558,0.8354693704838816,0.8924045223296363,
 0.467875981301037,0.27418335022187523,0.874270339643652,0.1837071506765957,
 0.41290357475763684,0.8661441995208568,0.9924687153264186,0.6085278806564831,
 0.531107417039668,0.7632021670468206,0.41827770307203727,0.48369947349168907,
 0.4796559930469677,0.7550324282877114,0.8368890996349808,0.6668381345678445,
 0.9207960795642199,0.10054445773913134,0.06803206957289765,
 0.42852423732000033,0.19617087472660277,0.4578911721989921,
 0.13166781598949084,0.10963172249713282,0.7630310774449657,
 0.8089635790588053,0.013990091591579379,0.7932895760662049,
 0.4155953225812268,0.8846909512439836,0.18662990147176728,0.7690369131449607,
 0.6786114428262393,0.7016611191599287,0.5914077485377756,0.7806932959915968,
 0.28029303931132354,0.030851156391992873,0.5227403432364469,
 0.890724387748876,0.0019261103696422976,0.5240418555585752,
 0.8701233681871391,0.124560619506218,0.043224104288009046,0.9617994501993227,
 0.7316733716703264,0.3304385556700905,0.11958036647071368,0.0150500907030382,
 0.784241838090554,0.42249086202571673,0.7942815668002756,0.886715730661041,
 0.43468117268585893,0.8677474012585431,0.40832843382251705,0.917087599972647,
 0.08951835655257456,0.5318539012913979,0.4116057403601776,0.8177971941851319,
 0.345236491876722,0.19654682134182444,0.9661392118605674,0.5169672216080942,
 0.14899461152589633,0.8294995840965087,0.25532332651637113,
 0.21838166439215778,0.5665762054595199,0.21891868227798306,
 0.45437695700457714,0.31324217289165535,0.21967221521605573,
 0.3663146785708843,0.599306712001191,0.03904836045229931,0.9668888507370959,
 0.9689315989957789,0.9757125334075923,0.7343346956990829,0.8038475157537824,
 0.46865089416293904,0.44108503295037405,0.30989816225634537,
 0.29548659906537256,0.9678607460111227,0.38354175548682856,
 0.9874660712572467,0.820039739881914,0.7436980490860893,0.04908085769732129,
 0.23386418230442607,0.4949127037185983,0.21055819125044262,
 0.23390834338604372,0.1786347366702854,0.8046417390278237,0.543755915386482,
 0.6975572855855225,0.6556407793722819,0.7509977289395257,0.6841516808342711,
 0.6249239655723868,0.48826140104973415,0.5999429475136309,0.6529931156732147,
 0.8798811531373767,0.9741599605835556,0.4280056342136618,0.08298623760207058,
 0.9509069728537695,0.6693680615426826,0.4482255192588268,
 0.007199923444562217,0.10306680083276454,0.9359974841577807,
 0.18926504286209955,0.1774301513218699,0.7689904085473681,
 0.14073711891752494,0.6164897952660117,0.1555830897334286,0.4273225424063617,
 0.36235554609045295,0.48750546548295115,0.7974646880161232,
 0.9212290110727934,0.7340868124942149,0.8125879731338261,0.9615203473053261,
 0.5444639043604205,0.9062663890080057,0.913790317798794,0.013563533550658868,
 0.36021899526346035,0.5127926204695149,0.12662509457586035,0.920324826317251,
 0.7603508446703235,0.1402404631449119,0.2830272002680766,0.20351727011630505,
 0.937108739222476,0.01993918854889165,0.360019974533488,0.6420560315306242,
 0.9266027185702578,0.1343137295953546,0.49289683382940186,0.4482025568695801,
 0.010755031527264425,0.053439610121755576,0.626021926048003,
 0.45738060080338183,0.028478916192801518,0.12251225663555143,
 0.15899402227436032,0.20971627530739778,0.669479372036991,
 0.35605457792215534,0.5743130892270314,0.70113367348937,0.6285079036056939,
 0.4624135826162309,0.8878956778076608,0.1702260262409776,0.21700002233082938,
 0.8171730390080273,0.5440713287736474,0.6926057808365155,0.8256660492064679,
 0.21331894916195493,0.34979643266176663,0.5989247684380057,
 0.37050798710023813,0.8053143033530129,0.09012103632591617,0.708102220603437,
 0.3597958140494939,0.22737777187856545,0.506648529642925,0.2593088149632874,
 0.43349072621451146,0.9809234045974395,0.1721073277370997,
 0.25594888467854404,0.4024108082248087,0.8819522307893028,
 0.12515515362021157,0.261253188334143,0.2643558792021029,0.05049791713459362,
 0.4144930726807935,0.27080011967421913,0.44902133904613695,
 0.27936449704084965,0.2797133322360956,0.7249574442979934,0.900034231717892,
 0.14341266391829877,0.609768932464302,0.3775242850128967,0.5846002343205715,
 0.41584646078997745,0.6219958588264616,0.6897817241743153,
 0.24840810559756188,0.5281065728816886,0.6763916923022636,0.6588940371838303,
 0.470883115725762,0.8829394578829568,0.4975381447994581,0.8527396799505227,
 0.9006497057907513,0.5540985331869551,0.9569544666930571,0.5125323189570815,
 0.27341215139483743,0.07198709365975431,0.24676694587566672,
 0.1425682394174279,0.9609936582194979,0.5240332418919734,0.6672115983983574,
 0.7701164182225699,0.6903190987454483,0.3782817190415312,0.20744981216610348,
 0.8587300539518953,0.12183136807609951,0.26376434503587043,
 0.7009351325410634,0.905365027569245,0.9943187937268513,0.5821623077818026,
 0.49071553880492835,0.7430953082629057,0.3133008336577079,0.8855227101021744,
 0.48091711041940866,0.0626733913544415,0.05102610816817954,
 0.16888162370354964,0.04805566233737335,0.7884519429810084,
 0.34525445678481714,0.297524641140819,0.8812770528514766,0.32992398028682535,
 0.4824199372151782,0.9162054499527583,0.9999600550458638,0.587255503873591,
 0.8360071563305415,0.11679496600293304,0.8646401845342362,0.5391290644592007,
 0.048605579207454896,0.19811172975069147,0.2121838925247801,
 0.796233782515691,0.7824235885169091,0.30782456616837506,
 0.042105882321910815,0.4751358823709343,0.32768189750862653,
 0.1720397061306015,0.46336839977506716,0.956641043702275,0.9777205021429499,
 0.5176148206025534,0.9381018451542023,0.14080455199176045,
 0.39237085067466393,0.023233329448918205,0.20184744394111365,
 0.40998275829813235,0.2655081504035318,0.827982030499452,0.21938840307786545,
 0.10593979694198519,0.730269755057712,0.21152279590558398,
 0.37547130416137586,0.5480132781959223,0.8144852711356848,0.7880029029975092,
 0.9516656652975791,0.5449584260574613,0.33466287318260557,0.5967060925244837,
 0.7798438104266946,0.12477316327421506,0.23948441022876532,
 0.15499321723719695,0.9373434934628374,0.040098163844931034,
 0.3823554532000507,0.46569277961904565,0.05948143773339598,
 0.24611588404681195,0.6890627725641221,0.5798180756149754,0.7013968287690525,
 0.4138979933468532,0.49729937920968303,0.09992471797995917,
 0.5563206216047492,0.7754831116167551,0.8981807794725826,0.749933165790253,
 0.8787562822338919,0.13721276575237262,0.27517968413855687,
 0.31181344356989393,0.4686106376870114,0.07906085634090276,0.689581496131778,
 0.6208474710833866,0.9840707000046967,0.7950454001661799,0.5639065857218516,
 0.24621234853893725,0.8174104478215147,0.2728685516573417,0.6588932559846454,
 0.21900228747098172,0.38776504847443927,0.6509643469433072,
 0.5978638377295669,0.2400314693047667,0.7023946726113087,0.9201623752179305,
 0.34381112309642226,0.6252592598401924,0.45496252390701897,
 0.4905835484469401,0.7939423060552657,0.27312385000097983,
 0.28702500164585254,0.5406373441051837,0.8913895552092972,0.7688323783216782,
 0.3090224847345583,0.03123507581732976,0.5075649527262293,0.8450298424782668,
 0.4729317337611192,0.4009876651100086,0.06463361865880835,0.5396963562305439,
 0.04069690601168208,0.13595660572987678,0.9477290319785032,
 0.2861324215802753,0.44426740180506463,0.028277663434438605,
 0.29353988868815617,0.6138805629809521,0.3325136660817557,0.317875897820613,
 0.9353083777684377,0.28603517260958145,0.512425259476604,0.14695533091752544,
 0.7885417709804842,0.19185832095709998,0.5005262255784373,0.0910241645924641,
 0.48344673963991713,0.5942373598038844,0.16212381866231362,
 0.0743198415542099,0.4370310445575152,0.5402308614714181,0.34880590800833156,
 0.050228712888796934,0.3884312092181652,0.2616255792530625,
 0.37637612466375403,0.8619620474022867,0.9804894730718416,
 0.22398400570471655,0.5287887614680197,0.830919106571678,0.8176162574067363,
 0.84681360389166,0.9354777674573909,0.7126613244660043,0.5392241998706602,
 0.5179839441974771,0.9040029162742439,0.09790466570330969,0.8655811081555621,
 0.7392458707395312,0.43733610347348184,0.7325516906767371,0.2181961683583714,
 0.4582115930332279,0.16018111188426698,0.9168554990705438,0.0560807781845476,
 0.3857168679584644,0.5994478290713361,0.1896419601340945,0.34342540770523355,
 0.008096527790344976,0.3769051499550413,0.6751893028736284,
 0.6036393003105903,0.7026723023182148,0.01696877141897335,0.8960083652681087,
 0.673207380747156,0.8444797927980419,0.8014341744152832,0.3668445456742898,
 0.8874276705850463,0.5001272586586637,0.7132516979026631,0.94503389234394,
 0.1539878357126525,0.406895203032561,0.9859617308146572,0.9876795348202302,
 0.7932930353219814,0.09530528014041018,0.005978938553599322,
 0.5376328438629682,0.7739654702645508,0.6335240830222255,0.34434399530566817,
 0.1883315375921364,0.622422735015249,0.32358439262961214,0.1747766426732834,
 0.20341831495548268,0.3174302301055608,0.39716181340021794,
 0.6695905222957328,0.1496319981204529,0.03728362338124325,0.6577473835336095,
 0.8590063523112803,0.8004147716334405,0.32923522890120094,
 0.08675471173793303,0.6494849657659912,0.5943953786012375,0.6600668164067901,
 0.8089615405713964,0.20823932971872217,0.5930680440248616,0.787144249394903,
 0.8140908898261139,0.844083533843736,0.16336931247535758,0.4043341731969944,
 0.43506021121278693,0.5461261978093147,0.30074694593541473,
 0.6441359433745338,0.9052518983413973,0.4919428019912625,0.26740062662086994,
 0.8953931572223751,0.47847882310482714,0.4120169252794107,0.9167184481067481,
 0.45231039265133366,0.8485072588476916,0.2808501617122777,0.572778181432207,
 0.5823598669523871,0.7401804522571107,0.41321029545916266,
 0.001575950528066583,0.1912699846289666,0.8734264444186771,
 0.6178448479661514,0.07603847252757012,0.3144433405220133,
 0.19840354597434073,0.1681378277721668,0.2085122764659244,0.869319914804947,
 0.3489710649295106,0.8342685424218756,0.18338673219734059,0.6255960443608166,
 0.4507541243684353,0.7356599134211024,0.15617715644324537,0.8992145138886415,
 0.8671914604751622,0.1449361093673871,0.03103482226760712,0.4726733365125554,
 0.2526206886713044,0.11342298227098113,0.7912787311526478,0.2741733384713826,
 0.5709978133918125,0.9246071761343317,0.6757379173047809,0.19668922026882163,
 0.7352972135679059,0.14976501903932626,0.8931795476181069,
 0.21589255316367228,0.34987243553023073,0.07093810697816072,
 0.25766794941910065,0.474619008324519,0.947383536759282,0.6294938257882333,
 0.616174115908051,0.9258346157520059,0.9922624574570705,0.19735860005258576,
 0.8758544809799483,0.8822733842744485,0.732822050249269,0.9929913987056914,
 0.23559403049799055,0.1985836262764682,0.4529933738432258,
 0.14518666995883034,0.4128072508489712,0.5009138049433964,
 0.08556861848421293,0.2910689961521826,0.7376428400299035,0.6404875508177871,
 0.2558552133820945,0.07430967672048161,0.5949565325411024,
 0.21228810395710795,0.8769088456886893,0.7068033433438226,0.3735051046929271,
 0.005320193732563427,0.02374691314749211,0.16889114932489002,
 0.9150810542393722,0.5784368916429841,0.838269894581973,0.5539941710094269,
 0.3790330086719803,0.10926603690726122,0.56462063090106,0.40484510579861954,
 0.0010643678692838776,0.6100604785779833,0.4994182486237141,
 0.15337282827694698,0.6167078789634461,0.678464480371304,0.972301377817818,
 0.31940579610490394,0.48902115233519516,0.22467864343288524,
 0.4903283161358014,0.2961014532090416,0.9455264459924375,0.8738007382532412,
 0.872785676859194,0.4458904436160107,0.263360551419878,0.4535120193212979,
 0.3150269101347962,0.3597135277983542,0.3369115787639573,0.17913866620812702,
 0.3165210590308164,0.9548047899197646,0.6694107164524462,0.4115451305288156,
 0.7776760920492057,0.7589629583576852,0.5281771281602121,0.8956016147746001,
 0.2530392325107631,0.14660895188453138,0.2460897087579048,
 0.039425923215508196,0.6750680947664294,0.4257137318735521,
 0.05528415426454003,0.8565245170610215,0.6391470470612163,0.3245430896787903,
 0.9128480342175479,0.07426801183867304,0.7920658892497383,0.9782252970926836,
 0.5401178413957061,0.9607610744688788,0.24263532810679844,0.3545421952210225,
 0.9527448712914273,0.8978226510019502,0.9348575383702595,0.3295089659581125,
 0.1760498851471215,0.4503138890788919,0.6705410715400661,0.44547513515711035,
 0.8142346986961136,0.40092679420867583,0.007545362404636324,
 0.6820123998925718,0.7435639290587607,0.6460988344490276,0.23812029303726467,
 0.13943627993002394,0.4974960210618551,0.4095328491812623,0.8878057767928358,
 0.9284819528691061,0.9554852262451156,0.4066698532083082,0.1331671159631569,
 0.6961285756436707,0.9459459032817911,0.3027169738658264,0.5631546825681873,
 0.6139830831514473,0.7925602284490185,0.6164080479714262,0.9561198435546625,
 0.06131652059941073,0.06597035192969403,0.39499377238890865,
 0.6493357060980322,0.42164714288193883,0.3359818443725917,
 0.20306202215103486,0.6022971367004575,0.01926378931438255,
 0.26346668386029903,0.15385599584229748,0.5225324970562132,
 0.1408077092410308,0.7986318749422353,0.4967840358456146,0.47211635302917354,
 0.6990560018300523,0.39599256268677285,0.1374810805233232,0.832251669535679,
 0.5150853414359865,0.0017109288203771555,0.6479776961690974,
 0.09924256279658084,0.4354070015475684,0.18714127804122427,
 0.18499634204533777,0.06940138565157894,0.3490315949069617,
 0.08420645548020078,0.8781476352223166,0.551466818234083,0.8100685030889426,
 0.5588036286484065,0.05934591150127,0.5787949088072097,0.7997412523051984,
 0.8744890602380936,0.6370519333197003,0.4388472413713157,0.1313296641182985,
 0.5877521351559395,0.7962682019192483,0.8440109991131514,0.16733865389677016,
 0.41685490701697625,0.012426733164908721,0.1680785702417057,
 0.47740513396461326,0.11150457817153803,0.6640661178881576,
 0.7220350377931153,0.6969883867016005,0.5060131918075776,0.3831328985198085,
 0.6423131093635236,0.30619296265331986,0.46861557477763893,
 0.5045107576751755,0.859391249973544,0.11582909751981463,0.8436677832106164,
 0.4338971009269307,0.014348969983639104,0.729850718297041,0.5418903186058716,
 0.12001588580398537,0.37800243856148175,0.8563178724025455,
 0.5660899581119994,0.5861142823353753,0.5964590397445997,0.6206146332250828,
 0.885909672486622,0.24592241178839735,0.23354940811972635,0.7011899743023762,
 0.2526852031026655,0.2908995687341561,0.9087455503370059,0.926761421688485,
 0.35616721412059027,0.7753367144779251,0.3058603866722772,
 0.25995498317396404,0.6627468719763596,0.06915789320485843,
 0.5074389320832828,0.9136781573738205,0.26592512408335356,0.3075411851730818,
 0.620033213558008,0.9790422407634659,0.2918680453805341,0.013702898828648724,
 0.3006013802744829,0.5006948470044548,0.6109754282263633,0.07693834620862594,
 0.24816273030149694,0.12356095493472563,0.4945585816336171,
 0.7726639756297048,0.9737165061030152,0.3400780964294764,0.9851910304077837,
 0.4477516036062733,0.3706999425409765,0.1063160868895553,0.5250640155843123,
 0.5789253536972362,0.4999866453175398,0.17362473135992595,0.5957465639722431,
 0.3972523510583571,0.8368338723613514,0.7718497543066418,0.3732040852652283,
 0.7067868027295494,0.6980422317462548,0.5336019331282231,0.5441735722936567,
 0.25810877186892145,0.011454438203516437,0.37939940678955764,
 0.9574300777677441,0.05366365291320063,0.6485548385441908,0.0985172720476426,
 0.41431798297578304,0.32772664968676846,0.45494209944165565,
 0.23363783493587043,0.8226006712307694,0.11515236848061716,
 0.5268239759029011,0.42200236094617605,0.8883326558929581,
 0.12173450079919834,0.9246063319421712,0.13223472902543332,
 0.9056108232384863,0.14428275304153237,0.9167339363424523,
 0.16246392549047772,0.06601059390147901,0.7773049183257019,
 0.5692502306879516,0.8126258103150914,0.3475355523487865,0.2890644303417441,
 0.39611745581368296,0.19961146198762825,0.4564378160800643,
 0.7401175195647235,0.9799675121290791,0.9351512098813195,0.6333802250167361,
 0.402060084766232,0.0322369425989274,0.11500708196514198,0.5055695181054376,
 0.72421016078914,0.20118207321661852,0.8846333924703741,0.8265128791806653,
 0.17617357139100187,0.5603753779501447,0.9547891819278854,0.6273201519749707,
 0.8343943693385657,0.33294836784005477,0.3517769188593165,0.769601527881749,
 0.3085464322346928,0.32478946189627467,0.23763049897807798,
 0.8343292528432511,0.8329984751140842,0.14230080888042185,0.458216621683174,
 0.9903224567003086,0.6109120134840546,0.9607651414034144,0.19656420799024332,
 0.8077501056806105,0.8890128505353811,0.9283559566079052,0.24053290821227225,
 0.7038992367390893,0.5014361497318344,0.7933464584008545,0.5390822510518215,
 0.8768494520223836,0.9517834260012173,0.6294077907516497,0.11179204473281645,
 0.683101526638245,0.7932622744967529,0.7363860608757469,0.709825200009273,
 0.795156880880793,0.17655757973266695,0.48742751685803143,0.2946169041774781,
 0.672350561558426,0.39107364445433257,0.5892058569980838,0.18981192576045514,
 0.8981347525534731,0.41215379594420487,0.38237909119362845,
 0.9369515035858257,0.5134257747580957,0.7977663392343104,0.5331910824787955,
 0.11254908024946397,0.4711183976785528,0.7540585094466505,0.507471685406429,
 0.32057273307302414,0.4470098343246214,0.9635586369855477,0.7646491018330888,
 0.9468690336327554,0.29233485908768786,0.6488640761305422,0.5637057476516766,
 0.5218126358868096,0.08221072227951387,0.09711996892219166,
 0.12854519841587997,0.9520806120745176,0.5841669013501988,0.4826698466073144,
 0.916494309196598,0.09725573563357626,0.580707498582822,0.7683252094866903,
 0.19257614329428607,0.23530273435672278,0.47672080748629686,
 0.7743236555432599,0.9833416268802104,0.453243131118401,0.48265829984964737,
 0.1741338153777433,0.5700855959160805,0.7493844612463056,0.2605491508959856,
 0.8873467773254055,0.534816679112721,0.919738905814469,0.5788448942247101,
 0.8418221650317911,0.7473094068282317,0.41874490267432807,0.5953215113132573,
 0.5949132547788998,0.12181615080472741,0.3788344826837676,0.5886845787596818,
 0.3321277642353688,0.36034304695237984,0.24916454215010708,
 0.9644831370299237,0.20039952552378215,0.9654825126209854,0.7783369026030301,
 0.08564336444357279,0.7899403563570184,0.5565246882186572,0.7387461765559432,
 0.461962073156714,0.3355128228580797,0.15748631703473825,0.2545571703674744,
 0.8159691276360211,0.4240994321953151,0.29185739320080506,0.0890577408895823,
 0.11398585610527501,0.5526049401729118,0.1555697570253598,0.9871093431276043,
 0.5545581508113381,0.6989196237050237,0.9062990913902114,0.10417137954403799,
 0.3590355617244977,0.6257241541369793,0.5192580113999966,0.6123302950052807,
 0.6688322384238649,0.25089689619000377,0.37429361523078475,0.75445653882057,
 0.7978154146672787,0.15202418132418516,0.6944344230704179,0.3062335096360903,
 0.23509868672887946,0.8060152089824792,0.5832490420938491,0.9621564636840185,
 0.5211154882770892,0.030648260654030146,0.5597111767053673,
 0.9845857331252899,0.485912128825,0.5974545248163979,0.23328738768493873,
 0.621328353556867,0.8025393592416655,0.48574559912827575,0.7450657608525586,
 0.9453126537373652,0.8121125095149127,0.12745155488444637,0.4382485373258037,
 0.0919555067429918,0.4247488070147767,0.9993276451467004,0.7123716505102283,
 0.14496062266261456,0.12865796229218274,0.8165810117717258,
 0.009369075261345783,0.9642271038549215,0.8093377621861885,
 0.4921662385323555,0.3053194251480752,0.24789098561130407,0.6164062816664323,
 0.9010779747531426,0.17027007919520498,0.18256625930073567,0.194393663170629,
 0.4592764884122166,0.7995132885120118,0.4272120692408832,0.3431031430076341,
 0.9930842011109678,0.6024782363516655,0.5972861496555696,0.7601646148551912,
 0.04641284229071507,0.6254258534097772,0.04453199982902467,
 0.9069904341258425,0.9825823369833235,0.7298014507328113,0.8447058602449724,
 0.6632375895800467,0.9012167718433104,0.18888780882181677,0.7214534678242889,
 0.2182450487880081,0.10345829904727333,0.27194711025127605,
 0.8684626580934076,0.5263785900311638,0.4702324612552854,0.29135901515076346,
 0.8492686600979296,0.04055800763081574,0.3384866894562596,0.6802390400143498,
 0.2007443288855011,0.29729668255952546,0.725573891099588,0.9792117548168705,
 0.8962483105883953,0.2819719441981773,0.0275182315548832,0.2872836361824982,
 0.6371970114552329,0.7573046592272721,0.9441726442310594,0.5566133230239765,
 0.4723816103297156,0.5684632422406902,0.020457433394250435,
 0.47870774182667153,0.622654738954867,0.5422799790915789,0.4276927272884372,
 0.8237517669907471,0.37958849191773103,0.5453869751098105,0.563500907022796,
 0.4793974823257853,0.7972654334173288,0.9895717797779046,0.8762437327168479,
 0.6813578883996699,0.23536761667581685,0.5691217401213475,0.5964566352785929,
 0.4253626761743726,0.5397403122066801,0.7534848252523326,0.619566011274953,
 0.6876435282568838,0.9587229350969984,0.8110409056450069,
 0.058752259523116246,0.006069157857551244,0.1070599021272719,
 0.2583656054545631,0.14632003577396202,0.9810622928676518,
 0.0014671426830735745,0.7745764374382589,0.44817174852671604,
 0.014215124004539392,0.4179130040055663,0.7010066727544151,
 0.8969275810494728,0.5593571990158199,0.5557288422192284,0.9498561263913372,
 0.6843491355344865,0.1912968959011615,0.6333052677855711,0.49686508322286227,
 0.7817452138609452,0.5455693466083491,0.7863268728232369,0.06252787251722325,
 0.6812350528894924,0.2795062208970259,0.9971190921335368,0.5722504815607481,
 0.6872989361814716,0.838173611260582,0.9859504717258828,0.2075189419689487,
 0.3574183025376412,0.8620739593627477,0.5544705591509708,0.42551223023185736,
 0.5573084939580432,0.8568120388018103,0.9291541126632274,0.9526265968633238,
 0.6534781863301539,0.8864493874849344,0.6997703468600891,0.787232793808948,
 0.15035737076469635,0.7642564943329092,0.7437122402365136,0.9358184653300919,
 0.7902326225472769,0.2474384148222959,0.9642243436024314,0.9248880024428517,
 0.9817474245451011,0.4682243655630688,0.7072696926654445,0.17113537991346617,
 0.019245663729923312,0.36601581861911503,0.837845441432053,
 0.1408846956178188,0.03502482846537358,0.619025204531534,0.6387705891575556,
 0.156208379652922,0.4631170836733455,0.17730155623744082,0.8589172461437845,
 0.05821750402801573,0.6107723555349656,0.9492257470821087,0.7744987106097595,
 0.9495979096709691,0.6319033243233096,0.7659435837092295,0.24641560684822417,
 0.30487070416249396,0.06687423578048457,0.4227288952277981,
 0.3737013893379125,0.10783810026883633,0.12311435729921949,
 0.7444598564249887,0.04263219282256181,0.7139093969835654,0.5675217485400584,
 0.6268849506071857,0.2863944933855187,0.06561325697534326,
 0.06376096481607274,0.6090703134746251,0.9723611474764295,0.8411930412004153,
 0.28445003450737516,0.49328612761082336,0.44808030584515623,
 0.16351891901731563,0.7186059982515318,0.9421453710069639,0.995353095539035,
 0.24521917951746486,0.08266203481570733,0.3046717582526094,
 0.28961699751559866,0.7650543291361007,0.388884326491333,0.2881338432680114,
 0.9639172114576254,0.9270633964099764,0.9598348096121747,0.7652175571538316,
 0.40330035100091544,0.13876883016058905,0.05220678862318773,
 0.4283212902188025,0.4937923705117322,0.3731057046432601,0.43995227377859514,
 0.4889926225734328,0.8900590338403755,0.9291432966865146,0.08021665867888261,
 0.13926907792899956,0.3187881635173211,0.6148614690294381,0.4843805720400066,
 0.03464052123941597,0.05763368453822515,0.47027174003121985,
 0.1388297856761298,0.29991471477638376,0.4615089157695591,
 0.010111168200128251,0.07937461379907518,0.7878740212870526,
 0.46369851170565113,0.6079765303021878,0.5823016058427255,
 0.19748541743297798,0.9354127661719988,0.23859522692110868,0.423342266674523,
 0.9728947818761797,0.5258573956244721,0.6228722479635661,0.22892149062450817,
 0.10291422322467936,0.25804337010461376,0.7983868733414189,
 0.004819224860770355,0.8796516167568756,0.12609046352391784,
 0.4605738674089698,0.1816615637070521,0.9774304353608583,0.8742646008695971,
 0.9362098116774382,0.9573264653807809,0.2652551030471235,0.2544547324309837,
 0.12835550332826484,0.30202559583827826,0.7359286879417679,
 0.6349296054148357,0.13315384893524407,0.7645802437290612,0.8139161085776583,
 0.9124733331080788,0.7428507067587714,0.035554855009483965,
 0.6095306671773826,0.3699656307608885,0.7228791299266383,0.7157514799855129,
 0.3479099132464646,0.7675008469072329,0.8591273649725159,0.9859034857739222,
 0.11216670862394895,0.20105987155569272,0.16452197873033458,
 0.6464756288888635,0.7344595611415811,0.7583710959887731,0.886658324483981,
 0.25826432443338676,0.8880229160484338,0.6413720722151178,0.6623957154847728,
 0.17383455756928723,0.6619190542115369,0.5657829692908851,0.4770365944619187,
 0.480041732086949,0.5558844970773839,0.254052739039503,0.2968742147115476,
 0.7488451255710545,0.6008426415398019,0.4306039750349636,0.522315028141901,
 0.5562397751568681,0.12371561711849466,0.29846686280932677,
 0.8654065100433369,0.18588570484648437,0.007929600426181294,
 0.19698847202661152,0.050460797893247866,0.5881160990470131,
 0.5766639031173779,0.8940567013215939,0.6365294391099813,0.7336920205488755,
 0.9551926726546466,0.7336478797403302,0.43427269948239067,0.7338958167385027,
 0.9748554917793312,0.17970446643436133,0.3879463043287572,
 0.28616801027747374,0.9544730869271936,0.22372434875943958,
 0.6507828036113712,0.3307661728094353,0.5594608463042775,0.8377810140606881,
 0.31347485339504744,0.04949026995588768,0.64878779626929,0.6026344534739205,
 0.18967499952016142,0.022683489881185737,0.028943913625279816,
 0.7265361684059327,0.39717095125424207,0.2844603762702729,0.7563734913829535,
 0.022016404179153293,0.5759378295734555,0.014512306169449207,
 0.37669880128698463,0.3673402851635156,0.5617081938205999,0.4756496062010258,
 0.5572778761731167,0.7250494831282377,0.5526506100449116,0.2525553334043924,
 0.6771797597421231,0.2739761662064718,0.33528436550856466,
 0.26524935430741736,0.8317202137112658,0.3057571581555929,0.1299528051618839,
 0.5670786485947495,0.30691483409438336,0.7837931831897895,0.2808583668372918,
 0.9712153762533738,0.7332214480531123,0.1965358100531709,0.3212396737789551,
 0.7572600500945789,0.4640792078124276,0.151170753341883,0.43870278584787636,
 0.7538190634422711,0.7130426006872045,0.299758639777569,0.15596176653871685,
 0.27914962485903383,0.7155288336241278,0.7757201786276237,
 0.16729356273406926,0.5180072063427379,0.9523429531418368,0.5452793531385126,
 0.7216183502088472,0.8438054307115305,0.5994824650862434,0.7025098738146581,
 0.03303148913853482,0.9887594461588166,0.04109461430646155,
 0.7725596519864727,0.17171969922694763,0.9843732571539,0.642921361383732,
 0.2436854912570836,0.9934146982812667,0.2935547132964319,0.5145519170143334,
 0.807504419519413,0.3946248087500709,0.9085189840992627,0.1336008436553281,
 0.28600963226210907,0.9200132707523467,0.7513293468669687,0.7085647305424791,
 0.5664402373512538,0.23523016607302338,0.5079420806748375,0.5768496902951153,
 0.4982399383182441,0.5905651193504649,0.23031703132462888,
 0.11859895426061284,0.38133316980605025,0.24599805551762122,
 0.5764335002374973,0.6385167878196925,0.18315714221933976,0.7244505429539358,
 0.9391875294556264,0.9694029733983931,0.13579569411776415,
 0.31442225298992377,0.5399454539424883,0.7403855550734944,0.925000485944962,
 0.6408918759411301,0.35603978964111993,0.295881322674696,0.47086600121503774,
 0.24515484002226917,0.3982419252456406,0.12438835085455191,
 0.17286892122114927,0.8647706617264445,0.2023816545631797,0.9721092562797294,
 0.5885425019502757,0.9550666894892417,0.5994776547301975,0.07819380369399742,
 0.612261167007293,0.5941617835981441,0.6667883407523678,0.59358329905268,
 0.7816776784776936,0.6179912815348495,0.4488568301639745,0.3395329015403542,
 0.7970117037713775,0.9849741761371859,0.2520340668854131,0.7466165996580323,
 0.5165624001595057,0.4589818854802591,0.6973327183497592,0.17552255031576425,
 0.9922350839196649,0.3728742000662757,0.20093485161008018,0.2567256163049292,
 0.9572313609389592,0.49231703300033836,0.5401565195669538,0.4243390270981987,
 0.6264917173520764,0.5160272662086174,0.9310328109320969,0.6224255005978292,
 0.9552513667148748,0.4326262562289446,0.9290842402488082,0.4003748345622927,
 0.46042057266584835,0.8809134711300884,0.1032671343232534,0.5717620574038639,
 0.5706924640507557,0.03933043955600346,0.8070962036809619,
 0.14291383071536679,0.5589351000705307,0.8284531979008468,0.3648893596708075,
 0.6585501687851457,0.7645027083549196,0.13605681123649704,
 0.035287404303320447,0.829465608977009,0.3211062450050187,
 0.047658685309761584,0.11026826252331035,0.9110326270092153,
 0.010940897543933903,0.7883064140048787,0.6952153434051506,
 0.24681252916422902,0.18221504832847768,0.8810194252437726,
 0.6882498713773294,0.9727188288318811,0.8269181274285843,0.2582975943709177,
 0.5630499513790757,0.36698790367180245,0.918616301195265,0.3737650034872564,
 0.8530532218246819,0.7354935467989134,0.6338034841881512,0.5874099852822668,
 0.2136350432723233,0.4075311174282813,0.5607536704085416,0.37194422726968346,
 0.46837750272948786,0.42924222768897224,0.8866739690963765,
 0.5269805621392013,0.8072656491453635,0.7283069714486108,0.6032463414727026,
 0.9239465758505035,0.2864479909395672,0.6080382103873553,
 0.023464641116277885,0.9148644137578709,0.9332123783417484,
 0.9118128270358665,0.25537738381341757,0.849863023317305,0.8346154403749877,
 0.37193534429232844,0.48792462175726603,0.6266505403405656,0.713419395088329,
 0.15783719230174476,0.5384177341867257,0.07780089187111283,
 0.6420995357522424,0.20413897418125782,0.23301465814420963,
 0.8090751434219869,0.2103121462413423,0.01063070088448903,
 0.26092396151057706,0.7464583275840412,0.9699545214682036,0.8434585780436475,
 0.7682806323957436,0.2701935125682313,0.62266400663417,0.7515829151395523,
 0.552416277638552,0.43459167146324773,0.8915896240372564,
 0.001080544351127033,0.7813947303619421,0.009673856376473777,
 0.36660759847644053,0.18177404338243996,0.06684270509499979,
 0.49108369162459464,0.09059427056155822,0.8317800126970893,
 0.49266683941845746,0.45579335411401045,0.3316647588835271,
 0.37653236936293666,0.10961237237009414,0.07097091248017828,
 0.30864735739364724,0.12627874608505874,0.6470708052503596,
 0.7445093162043647,0.0025391857298866505,0.6374870303236557,
 0.6577342933684165,0.346497261056125,0.5938533188756214,0.8883582456576428,
 0.4414138636484749,0.2864516814579817,0.29398706386261897,0.6658658460747774,
 0.643168075489841,0.23279548684300555,0.03826663690780774,0.6408945268606461,
 0.01866973339048783,0.197750157321544,0.2903024964415136,0.4275790881174367,
 0.24969700763627,0.2888784352240117,0.7863111290121559,0.021121982024747687,
 0.6327278918086764,0.14707264252785646,0.2015141682621957,0.7529626340349246,
 0.8471947485762235,0.5566412471356638,0.4673587925390771,0.01881298440080892,
 0.6322987405188554,0.7073259852626219,0.3793553585916527,0.06328671514785367,
 0.7648744609009127,0.5636066442748835,0.07236250648106912,0.6382629699192419,
 0.6634576504237595,0.1761365604261208,0.9610930771112287,0.4168184020884468,
 0.949729230010228,0.40922581629130017,0.07186242117845099,
 0.37407639265941706,0.7714232629537285,0.5311227591458964,0.2889103803194162,
 0.7069054707659652,0.6052992664953005,0.5191084172451427,0.535549878768968,
 0.42288364006006796,0.5285889451910982,0.2585661938553454,0.3711130178608756,
 0.18361272399361395,0.12761867459315446,0.560959890078631,0.5033533032648287,
 0.70942961379667,0.7110859446602469,0.2068349459593538,0.7085390600036913,
 0.5307716163543287,0.8497156038110182,0.012616929334069549,
 0.7899028633668103,0.06923744178852043,0.8017591665069705,0.1556623006642135,
 0.277226465135861,0.34951643212854766,0.18894465415256123,
 0.44122038220614423,0.31774627150416224,0.4429266116033366,
 0.19235875561997018,0.17035446636297746,0.1535519062520666,
 0.22027937106028472,0.15612537699682005,0.6002711426837464,
 0.34676293261802926,0.7789743238666487,0.2067792171623143,
 0.17179061273677798,0.34945139325199337,0.29384430687519547,
 0.8830179097184431,0.09795210384273478,0.46051151754443187,
 0.7006208643859798,0.562193897786136,0.7795104921952319,0.06693881491759301,
 0.4342004565924831,0.8413879574101257,0.4443721447373159,0.5881593275859845,
 0.7448687889394432,0.49529194034236856,0.6021806163804654,0.8339704334707531,
 0.3472924287238722,0.31385556540894377,0.3740047132247043,0.4793733278549186,
 0.8179237376339032,0.3729501115383618,0.5162616670405122,0.7569921435501147,
 0.08368694431255563,0.1523407151785262,0.16847712999086217,
 0.2600084778706716,0.9176772371997813,0.9343573628781626,0.2403172116928458,
 0.5456271709293495,0.2387024109976279,0.17532988563282392,0.9094251084807405,
 0.361282964349407,0.1889376113473149,0.6435752900882576,0.42969695739686553,
 0.6714439740351268,0.0583031553162644,0.8262475182287647,
 0.021486976998425278,0.03720163364219997,0.19708449178373166,
 0.48301593008362154,0.6966569403177169,0.016609225890928947,
 0.5030274218911686,0.22318653290175106,0.32140799217870597,
 0.28539953132356854,0.23807511090188138,0.3317403820096645,
 0.6366932411990356,0.5817150622950663,0.2903038763320216,0.0830837392769388,
 0.8345232252586537,0.7831736016122214,0.09885972603440474,0.2684177028811875,
 0.2608783790574194,0.7731589907129787,0.8048489797756821,0.26292141928434276,
 0.6567630910444348,0.6001213422595175,0.12314596992470384,0.6658175828870647,
 0.7521827279107728,0.6545132379464914,0.11291756261614727,
 0.12409043657674657,0.7198199891989009,0.5959293340858802,0.6458039073121583,
 0.8860672379084894,0.10151993112815172,0.19851998173205077,
 0.35175961424587887,0.6356217049923248,0.7261611846802167,0.8080008045804399,
 0.19653453443244118,0.7026638858255128,0.3929220820088838,
 0.10324553400065073,0.8004027242030337,0.04871375889223939,
 0.9449978015558393,0.8288723140821699,0.5272746477651923,0.08571523764158984,
 0.17370492123645465,0.4975884451974478,0.8351476586472267,
 0.21874674145441708,0.8246864004515526,0.7859703724781423,0.4811927380012949,
 0.4068932253821116,0.8850265377257505,0.4153490248667362,0.09407746510931703,
 0.7378560113997743,0.1869401756214265,0.9856590804734032,0.12330835144820818,
 0.3595166215943798,0.7936226435210777,0.9345656885030323,0.5293236512915037,
 0.36422496721565134,0.6818477651639312,0.9459194074944896,0.9537989246288504,
 0.5849023549135759,0.6662956842934618,0.6624039144368652,0.1181062944980098,
 0.5505976133477878,0.7508986991249649,0.5246672790301445,0.7514415304486268,
 0.3359687163026337,0.5856698294255341,0.6418361023676895,
 0.032535455075657804,0.15888587428059775,0.24696424515865467,
 0.4857473655702068,0.6015190752196642,0.9854680048568067,0.5557685775371729,
 0.7943301802871247,0.04228804743546122,0.7463371684907547,0.0645922051993526,
 0.6628074019503937,0.8481573706173041,0.45126674537263656,0.7730362322661168,
 0.3509909598670391,0.35767492807052037,0.6790548000264348,0.6136419720369871,
 0.27802863845688186,0.1394834507794941,0.5744019794705596,0.538903649932391,
 0.6656907338091703,0.9829403390777653,0.8177127012268297,0.8087503831087548,
 0.8260034115203364,0.826132490569617,0.3367168375215971,0.531442068709763,
 0.6828066190364457,0.05697737800175995,0.7257190281358855,0.4914241811882011,
 0.20838401309164156,0.14851091462666022,0.81908772097283,0.8048050440120494,
 0.7942771565754656,0.18463708892472874,0.04503432619341474,
 0.6183016889109735,0.5664278948023769,0.922713135536858,0.23995889255585257,
 0.3482767979938787,0.8216432375030334,0.3342194892926098,0.8413507564694452,
 0.4813222400750492,0.502605228526044,0.5687860274556998,0.6427234592776621,
 0.029217334296273045,0.5336562241477192,0.0884316659755705,
 0.8709476337075491,0.8291914038398116,0.11510643556282718,0.7682119480420403,
 0.6891197097532072,0.7409905626297274,0.31182409739779515,0.9457806493724917,
 0.0822922738838523,0.3940839623708362,0.8977464109943404,0.64703992841696,
 0.6444738414099895,0.16531990487049464,0.9524667620607006,
 0.40532264792714867,0.31115471151745866,0.4421611890270456,
 0.9788782739128324,0.8088821164542902,0.006344624909984664,
 0.8019921924727912,0.945507618650949,0.49541844353194164,0.24032434609793696,
 0.3261801509166048,0.6181725647740826,0.3112949358568816,0.22597182206019717,
 0.7345725167570565,0.4979954902501418,0.13896555930001353,0.6752809515572337,
 0.893094940755908,0.524868915486114,0.09894391904201838,0.726312415082751,
 0.8352903523624493,0.8055424194581533,0.06551760286238506,0.3141626096764467,
 0.5455580397976774,0.09290633538518023,0.9646381126283134,
 0.14808557314269322,0.9704128259828338,0.6241345961944955,0.1169660111201547,
 0.7350022679051859,0.06772264887813784,0.9216617383662093,
 0.06354836780627837,0.8679741141919668,0.18741318764492898,
 0.5043260058981266,0.13653661646569903,0.23251658480709825,
 0.4327399805836867,0.6948035685268303,0.8864622418027064,0.09943311565825763,
 0.6433272853486744,0.6933401455441648,0.20200108110496962,
 0.024933626184406643,0.2611198937024084,0.2879361804462539,
 0.5976654685731574,0.07397867520731671,0.015568294902966962,
 0.38418683223218764,0.8533978974558813,0.42356926631665837,
 0.4560947339467323,0.2247158603858549,0.04989489292626703,0.0389130850764795,
 0.010398102216502503,0.4255648235154599,0.0337148282464067,
 0.7991521175294481,0.8939799961437911,0.6424678805978944,0.2589673141069875,
 0.7291613101528568,0.27552516471469024,0.2062946728352103,0.7545007692790384,
 0.42887905000724313,0.5343363389169533,0.27913999597052996,
 0.1531368333258032,0.8126028027538708,0.9490721775246109,
 0.030008617337798915,0.49889426019830596,0.22953815384295195,
 0.15068335780172326,0.06861308418845213,0.8148968078150541,
 0.7243059176154001,0.05692745722422721,0.7900046650338364,
 0.17864486354031872,0.857258098958921,0.4910922994783794,0.7540386446723912,
 0.25774982498708265,0.22116649623214735,0.7865726293929929,
 0.22662422188299047,0.1503595936798019,0.5936633835668503,0.7967199177808849,
 0.2142176545254616,0.7036201508604669,0.1193201852327671,0.6539231827488784,
 0.3964671250593793,0.7170695595568176,0.2401360753907965,0.41263437657733926,
 0.3112735381982401,0.275764067342238,0.8518275586483375,0.6301211922365605,
 0.007652946285366169,0.5681586235211458,0.7367494743863453,
 0.35265194010700407,0.8025363683455101,0.6765694435792735,0.7509949582815293,
 0.9116444600455406,0.5589099853048424,0.7428564735452091,0.8656316261831847,
 0.44717184466165716,0.3334033530088719,0.2921259257065858,
 0.48924046557791445,0.20562730731934176,0.5505487639700765,0.864512175547379,
 0.8703997672113879,0.9945317117778607,0.6866107595721411,0.6855968502832694,
 0.6207584443865326,0.06887962553946225,0.4429230489822723,0.5918901342474133,
 0.06999788656003547,0.983175604093008,0.5867491620324738,0.8973536318622507,
 0.8856543914607228,0.430214384694242,0.9559188076980177,0.24069674873274416,
 0.8965599684545565,0.32656815040724063,0.5735272517217397,
 0.05581205263116418,0.8115135747180073,0.40663865849440106,
 0.07737714316754696,0.5702616140597664,0.43367900269948534,
 0.5415578780255399,0.16678371020438676,0.4157885673428665,0.9730544657541724,
 0.5994457729572865,0.19732560410142808,0.3696517938941245,0.873031621649805,
 0.4440358341044005,0.42952318452666804,0.7248576291531161,0.92290044825588,
 0.048565704099347906,0.5552063192299438,0.8566383280179092,
 0.9881873419132658,0.3451308129536643,0.7667761424155002,0.8322255761157185,
 0.7485297778603773,0.8380248346118275,0.6421832565225745,0.6347464060791065,
 0.011523534494009335,0.960241870848822,0.6727774323035973,0.1329017504537775,
 0.8877930749366577,0.2282006300449566,0.63199660852581,0.2775518613430392,
 0.05067198527183947,0.44892601435386625,0.09886187193400309,
 0.41658061883937914,0.8321720071103339,0.10761463336377086,
 0.10836344669089282,0.4289669393136598,0.01592959180565623,
 0.20564232161656504,0.9953588919064038,0.013178989252949691,
 0.054911296081575856,0.04719080868030889,0.96401218417495,
 0.05554436763278647,0.09724599115410826,0.633517171233432,
 0.36537056666391143,0.7109694023605416,0.7988800827716851,0.8477480624911653,
 0.5746152165577952,0.21298011998444044,0.8593433055766911,0.8743556972758894,
 0.4933419559073251,0.5160785414002395,0.02616833216295711,0.8950123160921086,
 0.47105955148598455,0.1289135748859226,0.7938399652837808,
 0.48615366947067373,0.21010080864016034,0.4621050912701574,
 0.018967342484147665,0.9245946881996899,0.9429431098944148,
 0.7395954033618428,0.28999067692272906,0.9435265873655322,
 0.11034070626517833,0.12968466963207492,0.9126512124885209,
 0.01633577366609429,0.4872335299842374,0.2163597326111023,
 0.06735448649345266,0.29245969775570835,0.7402808401196942,
 0.6494893853636851,0.9561421134490122,0.004993799136655319,
 0.6940186699560045,0.43514077461305223,0.943019943336397,0.7279524597682312,
 0.04934353690668247,0.6912944297914566,0.0011242095484947923,
 0.5555664009343289,0.17034936104038745,0.8098479701035463,0.9674309263279984,
 0.414555781137653,0.9491242016644179,0.7994866044386624,0.847303387391424,
 0.9169617233271243,0.7083296128826236,0.6754438296045214,0.15732348341057711,
 0.2563486935057403,0.42692678178857335,0.968691926063049,0.31586975107076076,
 0.6298179437075655,0.3364638616902432,0.5515492905597956,0.3810449273547275,
 0.1618911430110691,0.379925909468777,0.3271897956585026,0.8694933352478579,
 0.4262778486280969,0.6748849334955052,0.6130561396689407,0.33252713410575085,
 0.8793389074649142,0.7706097399399178,0.0413844376803455,0.13465838265364904,
 0.39720921478926996,0.3781786536910696,0.006564758596198872,
 0.8947985284776049,0.0367539228795567,0.2212714498167262,0.49532452043952313,
 0.42036596819484073,0.775060065486078,0.8217951187248335,0.8476102934173102,
 0.10066162040968463,0.23716222298419554,0.900497519262994,
 0.009109251362161674,0.22995532260372098,0.4213588223789355,
 0.34489788635137064,0.8106466809774304,0.5905318553031926,0.7026540026934152,
 0.1429295643184756,0.7109219099911546,0.5902664425364189,0.31609916989314635,
 0.8732772856970983,0.6944010354334726,0.0026495503831579503,
 0.8346807178839459,0.8098273978992091,0.22580551306421826,0.7579026299752474,
 0.2536609356627114,0.9840814436583382,0.9619880321520546,0.11823773415912764,
 0.27286060055589945,0.9211241397474288,0.261415676491932,0.3433978380422571,
 0.7521443057781727,0.6049039722170307,0.012509817454427696,0.541672653345818,
 0.581981185091215,0.03590249717693328,0.7068818489295524,0.6658824707548836,
 0.17149550354212773,0.7540609688428886,0.5492018092291961,0.9941933976981661,
 0.6854007938273163,0.16614437152451067,0.17605918810539334,
 0.8480597751922709,0.8633054357286419,0.5960799611498262,0.9939327842167278,
 0.1382949315760108,0.4382303894640337,0.5013359702373212,0.6000256016582411,
 0.4721216062313852,0.16410129915271687,0.5771681109023344,0.46580387603774,
 0.0076519488655201195,0.7797749032640904,0.6102198560904976,
 0.6687405188133798,0.5010502270913958,0.48000170420989774,0.8175882291451099,
 0.26986169083905553,0.8707351266527739,0.5360756588369204,0.5534041317796431,
 0.4860298648582718,0.739109369302928,0.012062192559016749,0.4769961574925238,
 0.8672549330727036,0.4469375786189659,0.23131799886954318,0.9663805044410575,
 0.43013609903737227,0.9624848384153275,0.8085272660248705,
 0.33887195951238924,0.1139444933598337,0.8911268295811842,0.9901476405084276,
 0.3218092360141288,0.6215706424557854,0.2032744484370329,0.9652415385425357,
 0.5945100644094878,0.11398807595644356,0.3973463204925516,0.5856534741727145,
 0.8934756469483696,0.8886950629156438,0.36848456991622713,0.1817175253401353,
 0.06989421569575327,0.5962644604267852,0.7880990394912357,0.6270363950842258,
 0.811560224910279,0.316468886138326,0.45089070454740965,0.273943888291056,
 0.4513627443773487,0.8689102533419552,0.3396061985181863,0.9796498954356548,
 0.8171965902240014,0.15935694613646179,0.2726224060955149,
 0.012655304316577909,0.3040106998556231,0.3182642945541292,
 0.4117764935776065,0.9060640251192831,0.33304513082991727,
 0.15743083304417838,0.6549872295857827,0.1419387595422077,0.5799277675205771,
 0.1254847935463117,0.03138550079242097,0.7290385209309133,0.3498184339594582,
 0.5528010434577594,0.5372336227993141,0.8307259465221631,0.15062346181234632,
 0.877330894187943,0.8907945982952575,0.09488044358013181,0.5072620574259235,
 0.4022987601963254,0.5718995485434767,0.25436789473824095,
 0.12961014694972417,0.08360797321263658,0.09651291284922514,
 0.8461319545522619,0.26392615581253764,0.8145266983634848,0.8440284474827948,
 0.686476365952157,0.38243777147469826,0.47341666123607185,
 0.018983149125902044,0.48143909108368965,0.3386814204327533,
 0.8427090435613909,0.5542689210329133,0.1566978564670658,0.3332725295815834,
 0.16408466221169882,0.48787356564309503,0.14683079279305766,
 0.29974937442186766,0.4774931060790104,0.23834487913637914,
 0.4303392403201266,0.19842435579855222,0.19017486140952744,
 0.45730610222682666,0.22624919764643714,0.9359167748242254,
 0.5589264553525595,0.3080019612560867,0.33790978274790184,0.5756948031162082,
 0.856673760406992,0.14010762894946005,0.8659108108420145,0.3866001733400318,
 0.9320138261934898,0.22676298557379904,0.5358288137521094,
 0.46371803618791574,0.016351481749191987,0.40738042076562064,
 0.13378324344886305,0.04070205893797574,0.01737605308054535,
 0.7396265079785196,0.027917664732692415,0.4139190228616094,
 0.8479887868150032,0.43915622374201113,0.8114099748132724,
 0.0023318541072931964,0.44858723027509917,0.8628282824188769,
 0.345905081178575,0.024962783051992732,0.31494103398768813,
 0.44491728024443566,0.6488122767784342,0.6528911277160927,0.7296715779011604,
 0.9119796504358483,0.04025091149829707,0.7568614166686243,0.5936223712634778,
 0.21777569718382894,0.9772708542297228,0.1335515522052957,0.262348965214265,
 0.8130562202087238,0.20860211871139667,0.23606993972820334,
 0.24821287690370486,0.25860233994420134,0.8548674028443236,
 0.30759746865467597,0.513621695231904,0.019199769654755672,
 0.10017945547533713,0.8830102689415046,0.015102794807432596,0.84063584327837,
 0.14168251907646856,0.3514791433963348,0.5093243488390772,0.0635971938050709,
 0.8827730761495697,0.7550722753928512,0.5096598286367483,0.6075486626301962,
 0.8163362841024927,0.6512002963681408,0.9992966683167519,0.36552482102714845,
 0.6907345265506853,0.5426247487688062,0.6754070440364583,0.8384183083140424,
 0.7316296243864269,0.46347694328996836,0.9449716996799148,0.9181896755943701,
 0.657573480446931,0.5705963575719041,0.564216146815606,0.4941516477418122,
 0.05007808298665595,0.03315520709352915,0.8587453670554749,0.497499981761497,
 0.4603222313494817,0.9278682471884746,0.9234954016808957,0.8758955718728411,
 0.561057349814199,0.13757708686928538,0.6184786035932526,0.6352847921726108,
 0.7542677421651,0.35281948093444626,0.4761403836974444,0.14889377605462406,
 0.5468425966952808,0.530931764239856,0.573423017666649,0.8330850041345775,
 0.6285390038307066,0.9756041124347663,0.3124094584570949,
 0.016042176733013003,0.16111739733794972,0.25347433586914603,
 0.17944468960322946,0.023257878051772263,0.4458837404598248,
 0.19968664155292182,0.05180290304401591,0.8245858622517659,
 0.9583420333875964,0.35940370639633734,0.30141261276587183,
 0.4040097503228527,0.01789020427006327,0.6938319689425887,0.8947594434492976,
 0.2200970109906295,0.15542533089244248,0.00754928820189027,
 0.4689754454295063,0.24464001011961523,0.7196761826601487,0.8377845731046027,
 0.9904172248588465,0.8576527653749669,0.8347658703595582,0.09235691445094796,
 0.11330708377708398,0.1464561140082068,0.5385657973766191,
 0.06603056009527553,0.5151266726114893,0.8635061501595868,0.2729809157974008,
 0.7436355436992386,0.17832377868665195,0.9772997519422107,
 0.07925126852562225,0.7286436029987077,0.5308096293544896,0.5041699376211336,
 0.19291382734265206,0.5029171024219831,0.9752481552596997,0.4770371123312759,
 0.4941115308621852,0.1352273800447995,0.11637800544102006,
 0.34777278630890107,0.7531863021721593,0.10759248081699169,
 0.5271332244691842,0.7581035388523344,0.8227961888316639,0.394219325170696,
 0.43975687204447134,0.48168918371043123,0.6256319736956147,
 0.6858239866825168,0.1634627059564978,0.21468542009115432,
 0.49339156158623654,0.6068858997587898,0.29946942398928655,
 0.38273391988755434,0.7445320400779849,0.36958574157831947,
 0.9230613420919923,0.8723665858932759,0.790857069936147,0.434307650307056,
 0.8250521504527184,0.46040577489931134,0.1791096819984206,0.6645519312458892,
 0.4010267741745981,0.9999179585018485,0.01196251273899862,
 0.49837042369071816,0.44245125607218694,0.8010576751459763,
 0.8678582355570423,0.22251764100209936,0.27978319603609825,
 0.22894151523431427,0.937097182771086,0.37064159899261606,0.4852070445635781,
 0.2831807925733145,0.7905115944649886,0.9013564270705785,0.3804949837195435,
 0.33317085107810396,0.9203092249716751,0.09462736674248273,
 0.12090533655993263,0.8302002542523472,0.21462727765233514,
 0.5053639951611058,0.25595814294353314,0.5103914782793211,
 0.22171355512263047,0.6776617658006902,0.19534841297479455,
 0.040126768260398515,0.7617270849771052,0.599417893116045,0.9916554391519427,
 0.6691834007729258,0.6458399210306613,0.8260161388061135,0.06378969343556573,
 0.7852079349049671,0.5623885605805107,0.1657223766968554,0.9886528433086266,
 0.4781243562210895,0.631491622937605,0.5110843770327176,0.4934302035989243,
 0.5801775073090534,0.4058210965404473,0.5723725591951717,0.28467684666157844,
 0.7612418861286012,0.09191328336718207,0.1532731459990847,0.4488591762175036,
 0.8027187745365408,0.5587650543010152,0.7442868849781842,0.8443277478881286,
 0.7443846052370525,0.9761300169694239,0.941565348843015,0.5019323926907802,
 0.8904652444576763,0.33358834495055645,0.8621953825588511,0.6966954935329396,
 0.6689851681605824,0.5196761563621536,0.9685319751719472,0.15513035224023097,
 0.4145619563126006,0.2256115099021878,0.574015231637649,0.4331312984985142,
 0.3944866117511838,0.7456190954682574,0.35515310208735695,0.4866803631305725,
 0.4038261010978931,0.5046255818693441,0.5104076177833997,0.9501853089295211,
 0.08685876220223943,0.9046082569786132,0.77229917382521,0.47020291643699663,
 0.9865848010978601,0.358601337310128,0.2725662237699227,0.9874659307696493,
 0.3399772524235831,0.9957483581100843,0.7335010408075,0.88894262727779,
 0.10119907998302335,0.9801303830601894,0.19594749044222726,
 0.5150633770430422,0.9269428532015169,0.3703204501161921,0.13231902629734915,
 0.8921059758875252,0.024621459478405217,0.9058917760394739,
 0.29327553556867725,0.7004776226887934,0.8205912476967504,
 0.24400312462989204,0.30712007396016194,0.8793208948989566,
 0.34414070362028437,0.7674923748688113,0.8964952191294019,0.5992696543538945,
 0.8090502616836399,0.6936552695235528,0.2744242520224014,0.12268222947713592,
 0.9832345424432496,0.47847425404949917,0.8599691389158136,0.7119732031348979,
 0.24435944786313202,0.2655011947201056,0.8202102593153465,0.6852174538595908,
 0.34110375240672364,0.5497529056042546,0.012636494962076417,
 0.5712159017422217,0.43406464475172124,0.40322790055571645,
 0.8253812098552807,0.7639458901565459,0.918451274921346,0.5227033355014037,
 0.8956380883004685,0.12623408927069935,0.47805535321763226,
 0.3824221691134384,0.578555042254377,0.9493242408763978,0.2987635029657918,
 0.9294356900253313,0.32112424496910996,0.755506442677964,0.0314705689864766,
 0.41304696642570826,0.6768330666859309,0.030881923539070377,0.21486241157814,
 0.5016900172643876,0.44245653660187556,0.5020643144978294,0.4089288512068169,
 0.35525713514449875,0.7257615993804722,0.8909964807237069,0.6039610503777585,
 0.515963604549612,0.9142513185820758,0.014493376715242201,0.5223310974457784,
 0.4456725200771229,0.3654408147453678,0.28236429396718543,
 0.33112527956241633,0.8944006811232378,0.05886619333059184,
 0.08328492157917322,0.27590929140853226,0.4081123442258068,
 0.8778765852934233,0.15939877838061067,0.3796219263120285,0.7074635760203917,
 0.050349464645281006,0.19715964544756293,0.05134386569085492,
 0.30554495160001016,0.1697335259168682,0.8552014286587224,
 0.41438731764854986,0.8639797553657607,0.07179306045884226,
 0.8192817984214011,0.00882360582074071,0.8874113204570637,0.85435075655839,
 0.07618643906178357,0.2614176242151185,0.5993191426609573,0.7721786414518019,
 0.4813236341706274,0.254170112284422,0.8555820096808904,0.49145398178024413,
 0.9717948061491474,0.5150476996108608,0.879827829833779,0.4092149808040567,
 0.1629052091941735,0.1281170867615311,0.38131055629605015,
 0.11801420743071134,0.7829084409887697,0.8781015046430573,0.8365066774213845,
 0.2804739259629794,0.6100928204961941,0.08317907270830893,0.9305541889757144,
 0.26681123279975905,0.08499615542446182,0.09061057765626845,
 0.7479588025906417,0.8265707254592205,0.4514452501845966,0.04282876234440336,
 0.782027502257791,0.41658155728345225,0.263043741163994,0.5302129571188815,
 0.08614539825749523,0.46391576305294036,0.2605355688667108,
 0.22573542088831633,0.9108929138421943,0.6985147505168917,0.8524990830842917,
 0.14408461339332213,0.541353130889041,0.9881859610408479,0.9630434587556663,
 0.14723179546937226,0.3013162599665312,0.10479004053313445,
 0.007705845644788822,0.2844306456968341,0.8003077356581506,0.61272997618043,
 0.2115434708117282,0.815157582044569,0.5868338904970584,0.8659016529138592,
 0.5927018396963969,0.06142235637495419,0.008891090303735805,
 0.10864861642000057,0.16668494281480029,0.291421591923529,0.7960464406467088,
 0.06342655738018854,0.7100329569758892,0.5129549650348069,0.8238551411979183,
 0.027311813408407715,0.562683096552153,0.19128627124554642,
 8.127850317660723e-4,0.4068441230332477,0.5625665571574449,
 0.47717842660785237,0.5179132311969336,0.6115627672380021,
 0.009938881182739667,0.3714849320664523,0.2663212535134456,0.336483392038469,
 0.4103117312265856,0.44593334180597854,0.1681915610796355,0.3071731762860084,
 0.12693348111118408,0.4291307052678296,0.08214525133465023,
 0.4163168226350251,0.30412864704463327,0.2032335639138112,0.9501428964597564,
 0.6113886754689174,0.05790655429682934,0.953879525346604,0.2365783101616904,
 0.1602805689563871,0.3079360629026856,0.4912451913027983,0.8420099636093438,
 0.01728515089375482,0.4331884740870804,0.30636621926123875,
 0.3741363525289354,0.45841229925861604,0.7756343569445711,0.4085776306845865,
 0.3643596922537239,0.4614045935529839,0.46255618775100393,0.8190947103110713,
 0.8170666326795963,0.553047068417453,0.7259102102795801,0.022235518954285993,
 0.2930297828441426,0.6872681130637761,0.5819842156285636,0.23261657450692452,
 0.7667316801319582,0.697395309569764,0.31869645212302045,0.6061578582076225,
 0.23516366253354248,0.9078435524256521,0.6285045356999486,0.5768152150430397,
 0.8196436746861276,0.7902768705864418,0.4586651689118346,0.29163679954831345,
 0.22362915891071544,0.9167418863397765,0.6253144651991006,
 0.13093903707900556,0.682278939132907,0.12227948237828301,
 0.27601440238182384,0.640795865106628,0.4584885487194176,0.8982603256464147,
 0.7744290358117526,0.14062052228688793,0.7956050289826345,
 0.05187586868630267,0.537344482922624,0.46368045296789306,0.7129614439030546,
 0.20114675739698984,0.33038844235352594,0.9305467052192142,
 0.031090917209458113,0.6340226505922837,0.8175460130812593,
 0.008819306698219287,0.8246781901694548,0.44216792296830976,
 0.15365294293740384,0.4267741420679102,0.3962545353863445,0.8906010303212405,
 0.605422632047256,0.24227302506597026,0.5639782737254623,0.8277055631207481,
 0.6932852076186545,0.16498692807624982,0.3494835280338514,0.5357798826138819,
 0.18597458253331278,0.4375124071198313,0.6253266336083209,0.4635956724213246,
 0.7137544908560068,0.8301000395766851,0.3895099592277552,0.2533208613862552,
 0.6188087299383922,0.4261682518691443,0.014733488187930965,
 0.6174567237872762,0.518943691476797,0.5050822405978347,0.9801409619065536,
 0.14129744514869047,0.4036812622606092,0.3424323748954923,0.7284763805378301,
 0.4624539403311667,0.9360152348689262,0.35085966466604424,0.7357832201957694,
 0.5500686604542075,0.22289378411496563,0.7615485825399493,0.9045405529760434,
 0.5371559581714658,0.3969566970555257,0.9199147936051307,0.7808166630973976,
 0.33680746839660547,0.8342244021354542,0.9675993276246599,0.2696694026088866,
 0.3211861158577637,0.6038497798106357,0.9228216665207638,0.6328445810946283,
 0.4934570555368254,0.5026508978958666,0.10508997170156154,
 0.04164531311911668,0.646207652765419,0.21347192908087864,
 0.38874672138710564,0.5160726335289251,0.5760712357833504,
 0.09319406408427022,0.4443642734577451,0.24296684782045253,
 0.9743677237913936,0.35788418493547425,0.17417404879754872,
 0.8257927511031009,0.8741596436015991,0.25849102010357305,
 0.07584389830866622,0.47624447922658464,0.8807221679125506,
 0.6312208168973288,0.9543280538528693,0.5618147356292089,0.6055505717054831,
 0.6779762339714923,0.8247746990662154,0.8922181764159199,0.31610790907536646,
 0.6986742637532435,0.5037947684074641,0.4999704716481361,0.47358264771029324,
 0.8632180386970877,0.9691533340333767,0.059365244898765046,
 0.22166034341311103,0.7920426258583464,0.9349082707430316,0.6152562993305991,
 0.8140212970424408,0.00887551477868187,0.6482952766356094,0.2692896088547726,
 0.4488694148475283,0.39109108439099216,0.4292075504651045,0.9824042239345475,
 0.7011770902882849,0.11087379454184021,0.7212611773617673,
 0.41255707982578893,0.6684691670736104,0.9857723961273998,0.9939134576214919,
 0.5604073117551643,0.4451030619066241,0.8438646782679352,0.4585569317031869,
 0.6289420003127607,0.7048147713863344,0.9052423544846062,0.6315182550209988,
 0.2735340608188356,0.7869320534374232,0.6615505800390789,0.7928846882133047,
 0.9044290085566564,0.11236262441554046,0.2441710728090989,0.4701088909275114,
 0.12363271879154925,0.9912564073236338,0.5245605142504679,0.3665922940746511,
 0.6609727141067508,0.5724209082149144,0.1960680972290615,0.5974203413399146,
 0.14394290327836357,0.7335105098079792,0.765991785876125,0.8755605554937914,
 0.3482500190944675,0.5756233916742475,0.3047102955672406,0.4076132553005949,
 0.42057432460259936,0.9432138723657963,0.1003632092051957,0.5834062248767855,
 0.13534896681969444,0.14180285595839326,0.8742198483317757,
 0.004996316706299964,0.6742255843752925,0.10462009074069456,
 0.6986494198640893,0.7013558117073113,0.564256915335913,0.6771519626284932,
 0.8973198369206304,0.15369356384128474,0.32036810044845077,
 0.29684533399026203,0.7596028130079334,0.9908401593948526,0.6270572789021456,
 0.7003583976521356,0.450072745455689,0.5200288198827497,0.38156548876371743,
 0.9099963524526292,0.9500444803952084,0.1680661973576767,0.08847579586221976,
 0.3269879715047266,0.8007496433557743,0.7135929643630283,0.7993043504927468,
 0.05553390678693293,0.49881838282476276,0.6034206367105268,
 0.021330638927122503,0.42220982969829624,0.6429160629838002,
 0.7351613599947313,0.17878533495057614,0.24682548600738086,
 0.4522155267122807,0.9062743764215826,0.6313532162225564,0.7003056555272239,
 0.2347535006611008,0.9660066828401944,0.14008792696288752,
 0.039486474171081865,0.6674258254164732,0.010122902513066911,
 0.012828148220936808,0.044017820024513044,0.6945724447360108,
 0.5026361243697152,0.8373829804117064,0.31993103905886866,0.9342372657819296,
 0.3444314911125933,0.9365818604006976,0.985019930441532,0.5645731773237397,
 0.12066628663806056,0.007536629340383572,0.7431990188799762,
 0.7695995047896349,0.8173380198750783,0.26204602066243243,0.3415789908387439,
 0.5080505936907451,0.8268750605923924,0.9233556909204782,0.991593066397294,
 0.19620751767333655,0.32161801751329677,0.1750535537909942,
 0.21978095158355504,0.17291987450656,0.16436213763969287,0.385854028780257,
 0.34794048363227836,0.20432743370687256,0.6460359254364348,
 0.7411619354968426,0.16317977376251014,0.9467846943384755,
 0.17698347399654324,0.6897445931889595,0.4007206969929038,0.9713534359024383,
 0.4111189192448369,0.28026750459175565,0.22812123722795685,
 0.5929383047781928,0.11209323662562065,0.22744856192477148,
 0.4292679890269714,0.7096113976867434,0.7418489863922851,0.6914019488973636,
 0.9335787216276847,0.6317185923139637,0.9975002898769331,0.6263187608048052,
 0.47517220000654714,0.7021964898135399,0.9601554870355115,0.5535843336754558,
 0.807559533680918,0.053489125743811816,0.9398933730548513,0.2826334118197814,
 0.1402041324826676,0.7637277306640741,0.31265385760441533,0.4316047402105089,
 0.8669366551817481,0.6437089172462919,0.77565861577067,0.8362120373062363,
 0.18970874312573882,0.31585469714880654,0.9703005489234966,
 0.051588464080303265,0.5002290329281077,0.0012911431697965048,
 0.7786101467297739,0.9956612958577249,0.5363680703153232,0.7777682301173949,
 0.5956092736611222,0.223589033661044,0.053870776492457195,0.7747502170847994,
 0.5225199037227217,0.29580550650529114,0.5558277966643276,0.8530181317998088,
 0.4822371551197706,0.2892577689247826,0.18612199919809036,0.9407753923832695,
 0.5964762342946333,0.4966516883095552,0.3766027054829255,0.10699941859819084,
 0.8864953400923747,0.09213999534799311,0.03759996993323422,
 0.23037538215050146,0.6019717686061444,0.937667164876102,0.3312144561565701,
 0.6483588365933527,0.6548764058639981,0.41204472403701864,0.7674946341500055,
 0.02042256570053924,0.9991014048755605,0.22255487865798917,
 0.016727234238833488,0.3450494402939318,0.19648732493223942,
 0.8204158066455436,0.6412003843334533,0.13818960684079173,0.8873495888173267,
 0.8107926259121849,0.5700839290646991,0.012424068787603937,
 0.10354696146996556,0.36444143766558523,0.45232505611652873,
 0.5118238160399484,0.24579943143330263,0.9861652898277923,
 0.014378039948695798,0.8813594349822456,0.8433271689612187,
 0.5962474603172795,0.3842262512281246,0.581998706288417,0.131595008349205,
 0.9448629106078704,0.07381666960061561,0.07672157451318795,
 0.11247362459095134,0.24896598425819527,0.27065966942884245,
 0.7738951977558367,0.5728471006995243,0.4515377901734384,0.441068096097021,
 0.8219563062608038,0.4634500494993873,0.8753648116093634,0.0807662638366492,
 0.8178891623361206,0.9722521773600357,0.8816035958989816,0.7696565012637004,
 0.5333174034911475,0.24759222212349596,0.22545808466266581,
 0.6262564553150781,0.5691199694920082,0.5020708821372895,0.26054551611032684,
 0.40882064478452296,0.7152960010372552,0.056792587640605685,
 0.7894275939854202,0.42582872638769165,0.7915970330384343,0.3968076445926926,
 0.8036557527397259,0.13371192926181752,0.3543155344553751,0.9659467260114516,
 0.29298736849001805,0.20510589797318712,0.9397060439264042,
 0.9574667218740421,0.18886366842709412,0.6321393671157551,0.6277679329401858,
 0.907984050716681,0.8087066345666842,0.19172073462221684,0.5727357248764848,
 0.25541889378541394,0.9283676113756257,0.8095242091638944,0.986303362613233,
 0.7879289293088025,0.7841129379097505,0.1275672713424295,0.02720463796051331,
 0.3865187593660302,0.6137428595761346,0.40816293336029164,0.6097751672793114,
 0.6735289732435065,0.3776216773511495,0.5051912351411681,0.11119805478132627,
 0.040335128045639324,0.8238152569786237,0.7014595677943546,0.929671677141552,
 0.5410423617844416,0.67292367335556,0.2138848319943889,0.32467948243290623,
 0.4299203514178921,0.9863513021584265,0.8498027436215624,0.6787844109682414,
 0.6887268347559506,0.627361800450954,0.4653005993786572,0.6512033209237648,
 0.8850259763533828,0.3716029466907489,0.2075486173648271,0.21012222873672604,
 0.6642666551681494,0.07426597352082243,0.65591901947695,0.0331140195304821,
 0.8190431362473558,0.5423329022931797,0.5187574976970027,0.5358377873060262,
 0.6751649051549636,0.055128931410514115,0.06604328394821268,
 0.9861790107030406,0.4524803783240716,0.7550844384021763,0.40586738899854835,
 0.7636216679225555,0.8749585706821286,0.1333494459164586,0.8833002738185953,
 0.429370789283699,0.6082301512735245,0.5545924862428372,0.5510331139660402,
 0.9347106797030598,0.06580335592426279,0.4333993604483435,
 0.11511547151576107,0.734971904107984,0.5347818874162722,0.6830105736432104,
 0.9150385525032951,0.575340624282746,0.45016727340261053,0.5567001033249739,
 0.5741125132175695,0.9206234206683427,0.5620844426480179,0.5564826303659471,
 0.8278085621117034,0.3799401185618745,0.8282275752424126,0.6432681484996089,
 0.6948833239681109,0.2827863066452003,0.39493280644925466,0.2839196742406812,
 0.05941832879615927,0.25791896107696966,0.16765168220694837,
 0.9344758717204014,0.9868743948708367,0.45034938309251304,
 0.44172625712470326,0.880841263718611,0.2705499876441386,0.9160320237327801,
 0.1487929154560964,0.790449439308323,0.3572890636409779,0.8224441207551376,
 0.9822316475435808,0.487052353261081,0.46558933096791644,0.8376444065518593,
 0.15764651313008926,0.02310411890799835,0.3437248564190757,
 0.17178052270944055,0.9752393163896915,0.9650920293359161,0.3766717920676863,
 0.25006994541444816,0.42228047061311014,0.4454356099855751,
 0.6997369340648353,0.1507431440872068,0.635252025808293,0.9226661418034592,
 0.2838169973201382,0.7884431109146204,0.8254915409888977,0.35202722661990005,
 0.9438447129415846,0.4732368043070202,0.2975501589235445,0.7092216882313893,
 0.45608410180806225,0.11817193736932663,0.6445750620462404,
 0.8931693659530063,0.05441623658329475,0.2929448716139219,0.7106713343105785,
 0.29669819587111146,0.42526892355930945,0.5088321640180037,
 0.13761914046320334,0.05451040197163848,0.404640825420163,0.6942392747042639,
 0.31741530037786236,0.28011720545458807,0.18094008801604033,
 0.7627906520517096,0.08574154632187769,0.4221748516040357,0.643846569288931,
 0.4604535205811149,0.23388306631201727,0.7754375329821401,0.3251360899034602,
 0.5716513995237565,0.890642560175227,0.44007379434196336,0.19457848925132637,
 0.5626836140517689,0.8558484149581004,0.18182090870525647,0.58225760585523,
 0.7197890654187822,0.7956426729885622,0.22745498988312296,0.5466367829811063,
 0.003437808289291522,0.862513287915631,0.6558090650040378,0.3794315287266685,
 0.05767967472853208,0.624971756025145,0.24179710646015662,0.2696974979016763,
 0.8544547164794707,0.03376829790069047,0.044707075573647836,
 0.3403883803479688,0.226080745953552,0.21206768728392622,0.5766195434450005,
 0.9403294443110054,0.630054065243736,0.9123904894520761,0.31946238401064897,
 0.14821874883427677,0.14827712058409237,0.4868754009544335,
 0.08581863554996161,0.9598721135269427,0.16921008819603212,
 0.5716850004006259,0.4690017098443511,0.9211010011093707,0.9930093960545555,
 0.6220916152466467,0.4549255503591918,0.20840779470859083,0.8646015934334444,
 0.8816078158001948,0.6673187171422689,0.8952652397852063,0.935674441665735,
 0.6130674327078768,0.37225177481743654,0.18685812225113585,
 0.04166296733870789,0.1447579738625121,0.40339804211849795,
 0.8787700042545304,0.7432658039657021,0.9756864896493057,0.3574726557821061,
 0.7260628510264232,0.07620563570280425,0.7320186855528976,0.2765829475224013,
 0.6463571130735336,0.5097871614570245,0.5144541428229544,0.8181145301131985,
 0.895216388793469,0.3493610910217121,0.44968827894326857,0.2639916386135549,
 0.46774891148081665,0.9816998338514404,0.17708931182852672,
 0.03296646689950167,0.07421362682647459,0.16654993668809692,
 0.17654195963929398,0.5107912143073122,0.17551967699900417,
 0.1434155796771881,0.4676452407443924,0.3888373341328112,0.4499854814669184,
 0.9788568132327596,0.7705173207027965,0.16468344203356589,0.7641693428320017,
 0.23200055802213082,0.5049454001267312,0.829231115880322,0.5935285774169645,
 0.8535557459006909,0.45795154820313866,0.6546442745725987,0.8501018486344272,
 0.41982381767190224,0.21617462717026892,0.3407833997822861,0.829653183499761,
 0.986894217718224,0.9088050818860312,0.5777388786841404,0.19814155127374689,
 0.7891776495675124,0.8755290805773157,0.3227680732432021,0.5397858845887857,
 0.22688947721307862,0.6890126131420539,0.8545816372921243,0.1429259435832727,
 0.8597951856432223,0.2302492411727357,0.6479054483132813,0.02385658941546831,
 0.603657968490168,0.3254204911702189,0.27104767829227616,0.6460333857589466,
 0.38939764358469453,0.28263909807304854,0.8606185253542858,
 0.040190241146532246,0.20221188348120978,0.6163978596421891,
 0.0735891712193435,0.70738244417211,0.8312061504753434,0.5406554028450679,
 0.017950343787733414,0.5213900856405435,0.5742913766257266,
 0.3848138055778828,0.5221186857808854,0.637168391610001,0.796570519268211,
 0.09044896227309329,0.3672426382682321,0.040782186515292085,
 0.12867270311174261,0.06387189181890085,0.02750136150282223,
 0.35372534491925867,0.8666184317290552,0.09478092769697055,
 0.5382861057053784,0.5782841614539124,0.5894081634018252,0.9924306440744414,
 0.8766365973984394,0.07731732073423148,0.5427950120970574,
 0.03760681675324595,0.10968069465354324,0.18239804293953277,
 0.16839384928604595,0.3204751603697187,0.02349694456568896,
 0.8851205854438935,0.3946708442936907,0.4449596698551035,0.29939085104641805,
 0.6958981940168051,0.1719376222370571,0.45188577492415793,
 0.007696260179896552,0.8941178565459144,0.7292770593958297,
 0.007385967560115225,0.5401211787787272,0.9779119584368117,
 0.03872306360790845,0.13707901522485144,0.9144820117010832,
 0.6796827804475234,0.5781588462059879,0.7282047075383149,0.19471640544734115,
 0.15388311394043075,0.8437044016172731,0.19437474218997874,
 0.7755631322369142,0.6858526167955048,0.6658665421042278,0.7699602032312229,
 0.5639789352079979,0.13481932409764263,0.8909799517143329,0.700240242860317,
 0.07721765547200132,0.3947952570765574,0.5162178893691649,0.5010920482113108,
 0.5919602911061288,0.2963330807224337,0.09854664923456224,0.5025742130120643,
 0.4524720131186958,0.43192810078931476,0.3193205986008837,0.2709140644058411,
 0.36546361450083076,0.3075197711862431,0.293787226415952,0.9296736982305165,
 0.5357782398819143,0.7669432900662033,0.7094532458278123,0.37394551839964785,
 0.27024286108091433,0.8375586531900571,0.1520204265415792,0.9638187486115669,
 0.032590120713414716,0.17546484942875096,0.17431191327164508,
 0.41141795924833124,0.015151863593639958,0.21632313182590002,
 0.17161627154458947,0.9246376033583836,0.6724443274557861,0.7277045943066301,
 0.00874438675286171,0.1607678090090775,0.03854006112804298,
 0.8531057808230356,0.571042514979737,0.30147309477908213,0.676931003153859,
 0.906711681493804,0.7023294534490313,0.6365313568151559,0.9574952804911602,
 0.6790703755609495,0.4160199764611715,0.3399668613601804,0.27696058548906466,
 0.4037265978475314,0.03952546751334973,0.868024114150029,0.49949184425755244,
 0.4215486462252098,0.7828560826028534,0.9930533777392413,0.8106230813813063,
 0.6819257441682343,0.26297458864057444,0.8665616288166909,0.5971718230960604,
 0.8694429004734219,0.25882175605851054,0.5632236137169122,0.226448540657871,
 0.43734832858405004,0.8673260111349737,0.31057450652481555,
 0.1097231765456429,0.9976062762050147,0.8591039936638439,0.8009598017961184,
 0.048160782226140175,0.7038176040132041,0.9114796449739628,
 0.3321351954225259,0.15071079989250102,0.3759852858882301,0.6013986539467171,
 0.5372034674417552,0.6959046557805966,0.8888962590398339,0.33461213342757334,
 0.40078489546265206,0.48766553131006163,0.6884221385327813,
 0.05695186888454007,0.6393150664183731,0.44171696186027665,
 0.07165251162966801,0.05006319723821473,0.015437558313841193,
 0.7828517238889277,0.27330404944011344,0.06892443853141395,
 0.22023677216035809,0.9428780231246181,0.8141605678907249,0.3222568959801626,
 0.02085863689536671,0.02049889730761889,0.8115935570470023,
 0.8131243615001734,0.45974790438081015,0.3169707920250121,
 0.41495792822310673,0.07011533575205431,0.6685602686139478,
 0.6794988301747393,0.8897547316633628,0.7791319169401313,0.6126699749320699,
 0.9522172534740707,0.47421142100066227,0.41046295584532655,
 0.9808676318079341,0.3710684267827765,0.28248799092597954,
 0.12662315383544398,0.9010330725154657,0.31522899299348384,
 0.49374519647543813,0.9976606645957022,0.21083271503966883,
 0.09355520323025546,0.09611862722180009,0.9136103198140968,
 0.2734127337034653,0.9452670774805648,0.5022831077066429,0.8416747105622742,
 0.03614277501009966,0.7185557003239428,0.9493734690160001,0.1237803685236778,
 0.05128935874576901,0.06160224676980186,0.9734411269123256,
 0.005326635125414203,0.6580649102039462,0.7149595291930528,
 0.20626474975722875,0.09893190298304133,0.22109262019566056,
 0.8863523300770261,0.6681340701416416,0.23566454674140136,
 0.17759078070671286,0.7553285845041193,0.4311706524920583,0.981616338610515,
 0.7658689931992251,0.0178201425142267,0.11045941201387144,0.8741781903458293,
 0.1573908554067125,0.4226585239896147,0.7593118470636756,0.12422257604891329,
 0.7526604908044144,0.7972935943654814,0.0709805597808606,0.00984853736343072,
 0.4867836195355598,0.5716246903418907,0.36390380195622807,
 0.16101059233055248,0.28248780668263174,0.19370971605577836,
 0.9472432640249179,0.2942429206999895,0.2021833764663893,0.06480037122807636,
 0.5274983691883308,0.06515358668317939,0.2978545201165509,0.6038389899105061,
 0.32598455590648745,0.017112166181346145,0.4483085409470431,
 0.1939617740942181,0.6659634417937506,0.4663376217809715,0.16020010586909794,
 0.17873003181065594,0.1622625869825245,0.24620745516014608,
 0.7033959313183691,0.8928758929851839,0.19584830639795459,0.9958402783471404,
 0.18037161663637435,0.4324776758893345,0.981674555460575,0.14628478919986854,
 0.07963021930000136,0.7237116246263909,0.13917295528510443,
 0.47196912215292697,0.5648132061573423,0.27595047723863964,
 0.49931598764824725,0.1334095361698303,0.8379263647516562,0.3090574685389704,
 0.06331643017142152,0.3417554687725163,0.20293860046008705,
 0.3477944348503835,0.44173079839153395,0.626035166169574,0.20507413412308373,
 0.501796520155076,0.3790682963699299,0.039275108721053886,0.77954562173126,
 0.7187387187868338,0.022045073260723802,0.33308278663015534,
 0.5444818884791202,0.7482023615254658,0.3888552523795812,0.46058848872488667,
 0.5012569970326527,0.7721925452990167,0.9093886843547155,0.07923479297827118,
 0.6119665995280645,0.5291644076834531,0.3165585723949028,0.06869506822412508,
 0.15586381555478446,0.30794035071457737,0.988407959184482,0.9595020930053542,
 0.5611512487667779,0.28887630335767167,0.25041871783247605,
 0.7200749822565572,0.26469322572613574,0.17623493489220532,
 0.4645323832343644,0.04475556941472458,0.4642126563808504,0.2124879682449864,
 0.184115757309957,0.9713426875887763,0.6710998386725154,0.7323230212437821,
 0.9447463951251411,0.11254142032661829,0.5844688640674828,
 0.09196392169533374,0.943168931570818,0.6595697338692793,0.757813530273643,
 0.4899407100226667,0.5623328739769125,0.18420640424456758,0.4109940505092675,
 0.2914821938088029,0.9291752095054711,0.33066744507467083,0.4157873805733566,
 0.08866334918410812,0.773039148350025,0.3516656608828612,0.6837697450796285,
 0.41926081618130895,0.6846729077309254,0.7249352135106818,
 0.26656255259490513,0.08681421259788802,0.47701665069492805,
 0.3344311262651529,0.018903916101948193,0.915228748382832,0.6455119434179537,
 0.46569161248114155,0.8870411681093782,0.9204715764503539,0.8243977768868411,
 0.35350041355737516,0.6382889570971529,0.9689799218587398,
 0.017212374376572526,0.2072128448388879,0.5595467739483404,
 0.2955077530581065,0.3850131384255314,0.2806934274198509,0.7239556705741127,
 0.5854179979868803,0.009493373191887255,0.4393789903912325,
 0.48201508848807106,0.33938601336861285,0.05489913486949005,
 0.4739207228181923,0.6802989558258065,0.3696022614769622,0.9022488748071535,
 0.8247336068103107,0.278506171844338,0.13374718384868567,
 0.031294604249432534,0.2022480449063685,0.57091633193118,0.38613968959493594,
 0.8816791009877253,0.18539188774738657,0.49159070477430067,
 0.4628643595145491,0.9428302996128706,0.49193427118725186,0.9926974282710599,
 0.6109116385421929,0.2149527638349117,0.6099103707606417,0.805487533781969,
 0.7787545962135699,0.4380762675294817,0.34809204829169704,
 0.10796085773206887,0.6948962270063521,0.9218610778185811,0.5028006138592773,
 0.9450696976425155,0.16680311470033593,0.46046553400015566,0.866575690338007,
 0.896557312368823,0.14380917354705136,0.3083751973405806,0.8954676189189357,
 0.7192816446490427,0.4596296987913603,0.20920239750428715,0.663622571502625,
 0.677465255302206,0.1660328743850079,0.524849359581183,0.2855210346612779,
 0.907256660210979,0.5060876246440265,0.9305747111049818,0.24837525534623384,
 0.3899993863224108,0.674335433595599,0.4013031278296102,0.4489759934537866,
 0.24865646359289773,0.8236544956357431,0.18545729026203428,
 0.9013645240565535,0.9842524237542711,0.3475871923702921,0.26110072517977134,
 0.6178214254090719,0.9168408815482942,0.4535046577458868,0.8029031651589504,
 0.41481652618943277,0.6083513561676402,0.5642143033929897,0.7654501127542082,
 0.9842682592944842,0.2051502739863318,0.039893834857315236,
 0.6457490731815677,0.024475609267431975,0.13779555385658893,
 0.2816204721731793,0.9427687062547955,0.5699441346146333,0.5338028739477719,
 0.26779217688653745,0.2711070359167843,0.8722082193129365,0.6459485046344681,
 0.334072166749421,0.1687352923039127,0.4027654070549439,0.47804585053307314,
 0.46636933184266904,0.27469481348188496,0.7784236132941359,
 0.5875635641010162,0.9369302769867948,0.3720040835906878,0.8833793603668108,
 0.3883850931186765,0.39550078121569254,0.24693266786185397,
 0.6941129060981643,0.9665980254557516,0.42017381614127647,0.1744524723417935,
 0.14935905195291765,0.3546725862907708,0.8086586705361306,0.8554837016645573,
 0.474177600502536,0.6663865293346805,0.1685564577396499,0.18208949245237338,
 0.6380531479554883,0.09137486278117402,0.29967295348668965,
 0.5498756383623049,0.2503027858263971,0.14924929775759888,0.7624251328566826,
 0.3983956313207061,0.06594799626288994,0.7109091703548585,0.2709897411975982,
 0.12173609502877514,0.34441456831034334,0.5980864332044518,
 0.48582811149281635,0.9893020056012239,0.29202214104120927,
 0.24127261211025108,0.03655646990389183,0.5109012090901635,
 0.3855796069339543,0.6120955379351048,0.7463584811592262,0.04564573360698543,
 0.9075623618265312,0.18570345762764706,0.34387015774814567,
 0.3126358825392357,0.43015912807140366,0.8591522917513179,
 0.39281890659087837,0.25217279074678756,0.814617963398125,
 0.25723731077315437,0.3832967885288211,0.00709592772065859,0.787146947003498,
 0.7862227146219778,0.6037200641058772,0.31203675848157353,0.7449201876359883,
 0.2775682275470296,0.6763706988447751,0.723378738548379,0.8247645679394098,
 0.30268394384952535,0.5437679697448906,0.5612292781889765,0.4546035602129965,
 0.36184866786827896,0.5047579181526617,0.6789109946710518,0.3057254610173673,
 0.9204059406459515,0.6881928237486037,0.694486897991386,0.76633564528342,
 0.45941056004645797,0.9152322139574558,0.08534949485355459,
 0.06133672553358638,0.07281026326625295,0.15056410599490244,
 0.9767483807152361,0.10843131945456663,0.39490463399647713,
 0.5784409175805099,0.15123771085883753,0.15556426760554787,
 0.8598581181313012,0.04130146563741621,0.3398018811678415,0.9122765532886961,
 0.78790823693544,0.3316004079021545,0.8801012811356361,0.4353860447254464,
 0.2691107134129812,0.15717192611202224,0.8861949835815909,
 0.06740571973581988,0.4732216970583827,0.12660781785972242,0.276338165152629,
 0.5619365564408334,0.49640885339020824,0.8438726705032864,0.1443365673494409,
 0.8284438029340928,0.25401285013384833,0.13203334824133273,
 0.41802140943534083,0.8779677591727975,0.020653822418245094,
 0.04126790554282722,0.6009003444848219,0.9239870540585854,0.9948373837230048,
 0.01799769888593794,0.778476248478467,0.19844430686751202,0.3857432194432501,
 0.24412920169322716,0.3854815003758436,0.42548704640758195,
 0.3791926243070719,0.11495683635950205,0.48693852615684574,0.502292288178603,
 0.07356544467595794,0.6526056561730076,0.7986078721762394,0.7940552925379105,
 0.9345753742311674,0.027748475946638207,0.3436646573786366,
 0.9483528004623623,0.003445588490624596,0.614598292094892,
 0.38673487175311405,0.4557683624316552,0.5920484956606892,
 0.25430142498799213,0.019924371512261563,0.34131033962190527,
 0.8714821995180811,0.44553639746726315,0.19373600372114885,
 0.45986502475913227,0.6419261814562159,0.42249090868459394,
 0.9201525461153729,0.08126231744073575,0.9770305397233948,
 0.40403436867322107,0.13899055317184728,0.5907680162556465,
 0.5195659381989325,0.1934107230857136,0.6356518160988434,0.362811627979127,
 0.7008666791163431,0.5479404481830414,0.22972919885283805,0.3838276496046268,
 0.2503803809543995,0.22262035342370157,0.8617441243308075,
 0.15161392837925391,0.23585478131309034,0.48673335294017417,
 0.6774356286155182,0.5242097803758686,0.4841443476042898,0.6859700581380224,
 0.9918466597901222,0.5895222118869037,0.2727249837220779,0.706367440316695,
 0.8765440166967615,0.24331546559250805,0.39726475116985005,
 0.5020370943449143,0.5847206543343318,0.6889724051269206,0.06596980705197941,
 0.5626531025070474,0.6352646122486414,0.9973374470827326,0.8190742526064814,
 0.13712867236168325,0.08949701063783588,0.9227172629028465,
 0.4884523259217767,0.7198508363746984,0.3975387365259425,0.7146909020412336,
 0.08935164822063424,0.9763945335203876,0.14161036187035136,
 0.9607804117411406,0.7975925994059471,0.9454924875695414,0.8877634592271099,
 0.719887512140285,0.38787240083511465,0.9364206997396565,0.9839843726910541,
 0.6505405426416403,0.7169824139671632,0.8866040713500781,0.5487334297325828,
 0.7082587445289585,0.8343021656872897,0.8335874596054111,0.13475183535120894,
 0.5212689308558451,0.7197812166592367,0.23810399249737957,0.4723029465765898,
 0.1647969444941414,0.9624239608500247,0.7861377299617129,0.9321798987893821,
 0.5279263981230135,0.8697070488316809,0.44066307005633487,0.7303004275144683,
 0.6316743413208346,0.3605207255368119,0.9665772118960638,0.6534761890030093,
 0.07237247332737473,0.6165670576785112,0.9807505773973263,0.1375477416451134,
 0.21829071474112605,0.17284337525457105,0.4054641493665514,
 0.34580456623786504,0.07875570499177309,0.742277353212029,0.5640779167740541,
 0.45223891519772885,0.9514387229313237,0.4330248799788986,0.5164619354357025,
 0.9663124287766702,0.6554450345502957,0.3423841380376692,0.26021158510227593,
 0.076879894798499,0.2553243961961992,0.7231603953453297,0.7376257240184808,
 0.5658095322023187,0.6445147466937816,0.16829951457791215,0.997311530237716,
 0.15841373378174928,0.361233386518963,0.6168877025278159,
 0.029181400156971016,0.43289620168363674,0.9126242549102014,
 0.9093572270753045,0.858201756067163,0.5987081306704136,0.256352915938172,
 0.27949507631594317,0.6585997791278639,0.557684021684961,0.6477887056555395,
 0.9618255392420734,0.13246652919264001,0.49826837256280254,
 0.14096019102089674,0.6574237075212649,0.8422292864966333,0.4056565105249623,
 0.8969828407181443,0.7208724228791796,0.3049052602294171,0.6701734821020859,
 0.9115394925311273,0.17641902022389178,0.26793003184537756,
 0.5029205691761812,0.639260027672418,0.06643364786269512,0.03486243950639922,
 0.7821215053351199,0.8108736324844469,0.9572650290632485,0.5436149534896457,
 0.03917949515303354,0.3096122939809942,0.6357226473703351,0.6803954903143185,
 0.20107454221968435,0.8720234057131315,0.4916710684525779,0.3697239844252579,
 0.6394188262590748,0.678997723168069,0.07645539996445583,0.6543235432247793,
 0.6577041270458345,0.1887327595816336,0.9342147905811549,0.5723798513862471,
 0.3322846041459676,0.992590637409403,0.13637751890970617,0.5166885281490903,
 0.19936838881352215,0.7634937601608921,0.813571055867165,0.05249513436153208,
 0.3284397343933759,0.5990620725632237,0.3598285687636591,0.2511261990115232,
 0.03714782634605607,0.8750212228493655,0.11852616666246041,
 0.33906865939620445,0.6214156577313452,0.5127839500473096,0.4998459822742387,
 0.8121050235568164,0.7046383009727925,0.17833238586311584,0.9280849690011478,
 0.3902844378412762,0.5370336362291818,0.7915840268996984,0.5358754134396682,
 0.6911787849473233,0.3937082749457589,0.4178574787048577,0.6562237548910342,
 0.0869625097201926,0.5157185690451063,0.36867640193349593,0.460731182040521,
 0.03677352980136128,0.0866016239528028,0.5139566832778247,0.6248737974635733,
 0.1747624697060982,0.6940497413691914,0.13817950922678457,
 0.48276345230847406,0.9988329104440965,0.765627832205226,
 0.016289322350889868,0.7136241434294551,0.021332957709251588,
 0.16555859901268466,0.23509088467840122,0.37478836934205073,
 0.6657453877018393,0.13222299727450604,0.3056477780318484,0.3432938927322351,
 0.5334534526049096,0.7953511773674425,0.041445414779030965,
 0.5274419241848106,0.8560513014257803,0.9576295813808502,0.41827236485116237,
 0.3350651876365667,0.6920285137896559,0.037368257507500835,
 0.41196453219417173,0.32941881293020947,0.6253261631280603,
 0.8702628910959186,0.675487677665211,0.36869387363818407,0.39533213085348085,
 0.05811624896999723,0.6006935909063649,0.1897305087065625,
 0.20199770930155647,0.8594978954017274,0.25652207535932847,
 0.31409616631060555,0.226330073823529,0.648398258582321,0.8470775332403658,
 0.440520465235957,0.9208295926436877,0.14435808956773644,0.21152677382016472,
 0.7387574426312865,0.028614794582470293,0.11046146951425395,
 0.053454198391751806,0.31345052350697167,0.4169219950811365,
 0.15111182831328862,0.7239767450042407,0.8080730660553266,0.4819991844161937,
 0.8365845046461636,0.7850801184439448,0.4792746041433288,0.302085084946784,
 0.3018773358255066,0.22526110583585168,0.09082799045818368,
 0.8898618696659023,0.41343293734642583,0.6085654151510929,0.9899435308777621,
 0.4990488854754396,0.9240674187172638,0.2428962452669141,0.07767616207556882,
 0.6937784497142315,0.5254370432884687,0.6110534836051762,0.8398858867414578,
 0.5026289583207251,0.7324271217147804,0.7372657225747928,0.24207208048951812,
 0.821326145349017,0.6680838659410306,0.3567922578888483,0.1783981485669922,
 0.14803468211530202,0.01360742438264162,0.7809808439905145,
 0.8698528699750205,0.4524984110785145,0.539050130069728,0.9355807307706036,
 0.4726728287023044,0.831654643359255,0.9285237567574877,0.7121949840792291,
 0.5619563652862613,0.6195645033201482,0.35116093428775275,0.8581371027576706,
 0.3317496940959572,0.11798337708710083,0.14769953712225203,
 0.7100699260332384,0.5218839550232826,0.42473363515064644,0.5579212674009372,
 0.7767287098472143,0.034553227601326175,0.23796225015443861,
 0.4406164189854629,0.5059932494831098,0.19153288957549064,0.5060390620998967,
 0.5277231174039649,0.6309701197173213,0.1907341053857129,0.16046844631926815,
 0.0785953122215618,0.04169604745442346,0.47579581517493863,
 0.31039852412328217,0.43072343768157195,0.7492980411073441,
 0.07208866514009526,0.9069807612389583,0.13517179124027523,
 0.43292113045317393,0.4122780712205314,0.8910839266467149,0.4540758928332751,
 0.20931535395115297,0.132417819501498,0.8863648175636879,0.8001658568228552,
 0.951385612603584,0.36563833213213737,0.2428491686778832,0.8072129146880227,
 0.05994113413050939,0.9974730718841543,0.8738008059624616,0.4130695511141612,
 0.7728805935586578,0.5879308686225205,0.6345677248966546,0.4939575806745642,
 0.9097591543963526,0.05298950677276326,0.4864196577238268,
 0.14963698821309657,0.1486107356806745,0.8266668886821729,0.1273199819610067,
 0.074079949125475,0.7260126078281781,0.4302905173454573,0.08816940178720734,
 0.8430593750716505,0.40184833754266225,0.8124147788378916,0.3782489730216524,
 0.52251775739717,0.6528591061676348,0.41154332408893124,0.973880619899957,
 0.5655220961010573,0.5560608095454181,0.7260444450250969,0.6649960237951883,
 0.16069457994280067,0.3670601702482139,0.11458715130917796,
 0.11128990300067909,0.5888938697511896,0.8123711460505626,
 0.14075631665891009,0.04955130156992338,0.04880096852637106,
 0.08597910402275066,0.26632641333988616,0.9350413245338435,
 0.30067493048390226,0.7590213158961114,0.005540545243994832,
 0.8925502238642178,0.7438952665899834,0.06836693371536062,0.6563761145098154,
 0.5701777584452611,0.6737377518308394,0.45886337862466053,
 0.06230988606046328,0.3348335814876696,0.46616374877066946,0.311516562082079,
 0.5176964091391743,0.3722532645879211,0.9109358173109269,0.7403657633028135,
 0.24099896001774967,0.28574758297138436,0.22151904334128336,
 0.9514875785744142,0.4916447414305066,0.7284758549396747,0.12000856552574724,
 0.25463698762815623,0.3280048803336729,0.44963972725525636,
 0.48361885666350124,0.664864004879576,0.9020655612800289,0.08066731154585816,
 0.8891556378014025,0.8541808915874741,0.1763586366579326,0.03060975589059456,
 0.6820867550670728,0.4456596631615962,0.922190515553645,0.6516923705406202,
 0.41112527559211165,0.9013446466293593,0.6838537939911342,0.960586733565016,
 0.6832939760053983,0.2033445070246056,0.20773876578315376,0.7964688474312256,
 0.1696331289055989,0.5194003695688995,0.29540442590364324,0.530289289110318,
 0.20817763087165497,0.6383704585110225,0.11515734914904951,
 0.45486437437032656,0.22114231011755814,0.5912240246879658,
 0.5125642490663517,0.40954461111235174,0.770595519397435,0.6047401198014811,
 0.33701147216028005,0.11595801467929245,0.9554880090216435,
 0.9486616798044112,0.834793033820449,0.9225317678180404,0.9980774013294862,
 0.7955570702355417,0.5358478040701938,0.8983872405825425,0.4261027912339207,
 0.2857917159419029,0.7193794314591994,0.2883940027021916,0.04883521846703209,
 0.05641521824317741,0.4918468659866838,0.5866922400680309,0.5743664822434458,
 0.09692685313943361,0.32852880370764004,0.29045776521156474,
 0.7988562923013983,0.24531895610632448,0.2215620405063914,
 0.13532121397625219,0.045453075876929594,0.7306929468118766,
 0.3322394855979013,0.17160828547947893,0.9286809417724018,
 0.47976551119660793,0.09491279236609429,0.8129798698886176,
 0.6397660397556325,0.9542305313936077,0.892927404610131,0.45787529983434827,
 0.4631917332081684,0.032304277923749236,0.6184463524529845,
 0.03093331213677286,0.9535348618426612,0.6854774472758126,
 0.20495964271399414,0.9221378855270176,0.6710720307206949,
 0.14239236873516303,0.569589586078344,0.22805463889591682,0.3320899132259114,
 0.07249001856555526,0.5142627366201592,0.7878070820551877,
 0.012209924329659283,0.8740463922418186,0.20832918280705093,
 0.8861978284309163,0.08291291282650126,0.8662319677700027,0.8814219433159993,
 0.7029527915258192,0.45016335949945674,0.5823399213772023,
 0.46229812710966267,0.224484151418071,0.06907179757953008,0.8077257406011733,
 0.08392986634499788,0.020050145332446423,0.971994051227343,
 0.8754859387565151,0.5445845474271147,0.3219100915214008,0.1486984264658131,
 0.6349619280471785,0.001547237733743434,0.7683837201675712,
 0.8047207320556761,0.5219634996754572,0.8569418182438191,0.8981526572001788,
 0.9261759172568325,0.6635297133048265,0.928998794790671,0.7992935843878886,
 0.016422523971422298,0.10031230972526584,0.717206797639664,
 0.47844495498384376,0.8241525766978618,0.7800061153386421,0.2517122865329573,
 0.5146437910153282,0.6391217689086435,0.20662305347311727,
 0.13658137762081346,0.03857226918279111,0.2096504275062433,
 0.46581616267003123,0.05339813446987096,0.30282476737969133,
 0.03249771184404815,0.19279424022628655,0.8260759399000743,
 0.6670878319465829,0.9306027322561138,0.881213564499195,0.7942807297223077,
 0.6571994259365438,0.8884169342724675,0.7493240818720412,0.8671319862634117,
 0.5150074472550938,0.764879118961803,0.4343272100941935,0.7267117671945196,
 0.611005348398446,0.3260441647148289,0.3085558747140411,0.6957601841781595,
 0.9074194563811583,0.8397193301576573,0.35846384673616505,0.6214579151097919,
 0.6829312834186335,0.24516392188662373,0.33005390300724946,
 0.9919233715223106,0.6736613900920311,0.8829735560395016,0.924005362166402,
 0.03506547783667091,0.7287698617994298,0.9002972657242312,0.735247863300926,
 0.25369983398222307,0.7855061494669067,0.5436339325405556,0.5062565425818607,
 0.9030133915649384,0.41649551687294084,0.7122990979989816,
 0.03310826495566843,0.9561082983797472,0.3366358765036678,
 0.11135061286599524,0.06112513716207757,0.9105336503928946,
 0.39457259524954125,0.9023094333652525,0.9541900066494925,
 0.20824755598364098,0.17467899933065167,0.9901289270492396,
 0.7002377228266657,0.1755696589480038,0.7233811699708479,0.722223049569117,
 0.6745848251110016,0.42739705001568384,0.5406997658198155,0.7482040830997875,
 0.25906888389397353,0.2026416165011058,0.13056045595569377,
 0.6597212297472067,0.23433148409082039,0.7692104610662084,0.7710256963764385,
 0.6002196553640677,0.4922832728769174,0.11056977892610165,
 0.15431817686332538,0.8880139010259869,0.6651941445160294,0.4704115634820748,
 0.7532367214792755,0.5190101267839227,0.5913978671140543,0.60385171152862,
 0.04782495022627653,0.6821340646618,0.774586636018765,0.6902540446426224,
 0.3232463763088115,0.34146784890778437,0.09743126776072264,
 0.2521958162046233,0.0734155474786663,0.9271562625526423,0.6592925914633307,
 0.5338347251649528,0.4168846506776015,0.6597423991979352,0.3258211709666956,
 0.4525233212779811,0.7150878403354339,0.5511985567958869,0.8720866166166334,
 0.40420405177950713,0.6894382826991597,0.8700159098824203,0.7819894856328353,
 0.10011009077713318,0.5555669364798319,0.2497458276598178,
 0.37025985540468753,0.23678846443124357,0.762011277515956,0.0857532563323764,
 0.7508451408642788,0.8344110579563159,0.21153702633323734,
 0.18110537959515927,0.8545896263584809,0.772283383465717,0.4196514548436838,
 0.23880810452040746,0.18295391971081887,0.702796502182677,0.8277950354997441,
 0.913394501009456,0.22456687900222183,0.7077766212070677,0.35675172232122154,
 0.6554663295114588,0.23942921010167095,0.5156865041758227,
 0.25269637333478967,0.032254289343747744,0.21753588713061456,
 0.1475299062159039,0.8623267119323061,0.9726803422241135,0.23663535556144832,
 0.882339867687808,0.7010139849642355,0.1739649760906944,0.08935334225335634,
 0.5087472682626064,0.1999913458860807,0.27817502097804236,0.6944314028853744,
 0.845783414846192,0.09975470564521338,0.1418561566535398,0.49473645263935373,
 0.46662183244839106,0.638750608367472,0.10018929155200795,
 0.39283050543162745,0.10474805856247826,0.23371608043613956,
 0.2508421354237611,0.48048770611219516,0.7572555907620115,0.6029286975474932,
 0.6172125114537095,0.1867030563046812,0.971249743250865,0.3599060629207509,
 0.7350554695694143,0.5119320492023098,0.344111785879623,0.6075320448675214,
 0.179555211429141,0.22412141704056965,0.1489568388368946,0.11524652195834806,
 0.9688825524162865,0.19265524861260963,0.6386741290799571,0.7722796346540061,
 0.15869093012096824,0.8939717935066742,0.6315331952645863,0.8235807209199641,
 0.6147267055207379,0.3532234566048562,0.2922605844988331,0.34681362842424557,
 0.2714900735094302,0.14345380363570337,0.49169427503402585,
 0.9510853540648823,0.13725846191648727,0.032326616497419414,
 0.2330771807456623,0.41642205729518267,0.4200217244698745,0.7972513360514509,
 0.5221011717615146,0.6206884689047761,0.1487260066521192,
 0.043136400358114946,0.45013316812134674,0.9154128171616339,
 0.2060801441935508,0.43943688363882627,0.5281884670260737,0.9619021442325998,
 0.6096464381729402,0.013533752727304948,0.6786890952383118,
 0.05283533996336354,0.7929291795877275,0.9673087607178795,0.3166242133640519,
 0.9651352699490264,0.23631100643752445,0.573422924427281,0.6037880974957684,
 0.5175666463070779,0.8015223122231802,0.6626590278561321,0.7135382874958068,
 0.7117550450139056,0.22854904627868466,0.3587383245857315,
 0.012400078490157895,0.3203768849127886,0.8200781685984299,
 0.06799487636829649,0.34710099289440643,0.7363380958732123,
 0.7327251215448037,0.2493831259683985,0.71905195739976,0.6896051975289533,
 0.9506781156230167,0.571709339774849,0.06466553341602288,0.8870375063697635,
 0.498514824592453,0.7144466346329845,0.9214357350563647,0.5940878726764399,
 0.8695998894517185,0.5341723316985717,0.051985006037568504,
 0.7279186551793171,0.17906837610617432,0.8038044917582298,
 0.48933768704591263,0.9074667193248512,0.2112315284459254,0.8056850257184427,
 0.04036622133207357,0.19125665097327638,0.08582529122814808,
 0.33504926098215615,0.8879123990065689,0.11123304622167507,
 0.8090516074684335,0.2007826517577812,0.870277680428541,0.4261529444621953,
 0.7674331570443456,0.1069945523709539,0.7354935436807971,0.5232470777054427,
 0.1155002404343769,0.08605521275742523,0.20365637063545594,
 0.5153760801354718,0.5721340982955425,0.974288762490228,0.8548004712620223,
 0.3661845052344077,0.75256845910398,0.967097961691169,0.11042152200706068,
 0.8698657568836822,0.49060305810483884,0.7087005992775641,0.6834157354967573,
 0.5916334476343112,0.8049674769676396,0.6525766687824032,0.794153514333829,
 0.7078147868357433,0.510556208100126,0.31975279097133014,0.5807544117815939,
 0.15093019104818717,0.45197294691086654,0.413665586321768,0.5282514518515713,
 0.03204726917602285,0.2670101380845049,0.7730613319476137,0.1550996316977178,
 0.6160163670404777,0.993819405242344,0.8748754261672397,0.5540891079476008,
 0.6144524578568952,0.6394778294213923,0.40166876455521316,
 0.41955603874740266,0.9983641448476065,0.685389143128864,0.8381982958610815,
 0.5878295676001952,0.6843352295329259,0.24301129352571282,0.035808953842833,
 0.4027350111760183,0.49839326648849447,0.5728084479932146,0.1542001158734213,
 0.3919595648567301,0.3238178978234172,0.3101334935354677,0.0986932211268221,
 0.44428407076562515,0.5834901278272813,0.5226660817474578,0.9574456095374666,
 0.1614520069499864,0.26895470088685003,0.09862518524397612,
 0.01884297436940552,0.15045987107967063,0.12344884061305506,
 0.5253466800789827,0.9697040552092036,0.970481435430447,0.9764874197927108,
 0.5231347736110095,0.2842362335326728,0.1716718674993356,0.09244978935345549,
 0.9474546944551544,0.29762178631905856,0.7942773622550757,0.9323364050743654,
 0.7949402180938564,0.06022402145976069,0.6982715586102409,0.2567994309722851,
 0.4096943183278223,0.3279580284540444,0.7504292098164865,0.7159770372745549,
 0.890659738844401,0.7903635760620966,0.19333926068811247,0.14447346204897238,
 0.2617455037578402,0.5446848541623923,0.41727634094782196,
 0.41330505720023425,0.8677697998485124,0.7394509783794754,
 0.46739676053345414,0.4862400733313468,0.23077901855562777,
 0.6954861922247669,0.2969215921233709,0.31730003592628997,0.4429256689874833,
 0.3627551305940969,0.8777343625320411,0.4661158390836111,0.23862940960113221,
 0.9628910487487268,0.10064072616182795,0.5514737014016163,0.9267802038844863,
 0.932719324576308,0.19827104506537163,0.3289070198139925,0.6591624972048717,
 0.7181464986335806,0.03839118626997562,0.47479433484077493,
 0.8327150602143562,0.9309269359707925,0.8909817127824186,0.12312798039770156,
 0.9541387952149439,0.11194628936940232,0.08255967088265237,
 0.5860724764997723,0.734449917531725,0.618829449729176,0.7766479556177358,
 0.4892066736925531,0.29027626903558645,0.18356447907918205,
 0.7469014403436671,0.42712139725945053,0.48860965819461244,
 0.6848743437997955,0.7216852782101114,0.8616359160782008,0.6283345195927079,
 0.008011955761229816,0.50703075448768,0.24309541931812606,
 0.37215134791168736,0.4660370055890062,0.8965993872131718,0.614348507420864,
 0.45905630003285414,0.769089421193495,0.46636052154438823,0.4365075728426121,
 0.5337778807798852,0.07132877778641766,0.13049017710846844,
 0.11023670239383854,0.2505999388852339,0.11671408301305641,0.722033373672939,
 0.9265396880043246,0.025523294243360195,0.8417550903944485,
 0.7742903165669385,0.5212383781069244,0.14729955015762086,
 0.37644925143834884,0.49898289345985836,0.8362162319824167,0.279701437268559,
 0.13415716846654424,0.9249446447184291,0.4194920963788539,
 0.38856347466622654,0.37764260684133466,0.41754961407582725,
 0.30064270300851326,0.2852371703275963,0.10253318225260344,0.858662954911292,
 0.13542200981983032,0.7626058494190544,0.3108578657483315,0.7791984133578239,
 0.8834103593786334,0.8368676745004229,0.5894367847046759,0.8707429983822558,
 0.5522186873485617,0.06773409833727195,0.37337710414102443,
 0.6612059500382473,0.9063543040999525,0.09382207640414775,0.8355944365898175,
 0.8373948831039968,0.25703109483227093,0.4711065901288385,0.9737052777196755,
 0.8602737034475754,0.8673125965405799,0.5048656078419459,0.9895242502597377,
 0.313150756123058,0.6336425658960633,0.9041713775350162,0.6513740448083528,
 0.5986713188746722,0.4389102982487294,0.945390073339712,0.7121894718582257,
 0.6737711652152508,0.5734555611280996,0.06418112029905731,
 0.15876417718285096,0.07669478414110509,0.2247932021349227,
 0.5649508752429326,0.4388144710096036,0.662268962310117,0.10250469872526558,
 0.627727118271425,0.27572331952666185,0.28729677601003645,0.4212404471845933,
 0.45191000825482186,0.9364419326699422,0.5686957066232169,0.9698896116944985,
 0.8957781474103326,0.5302604100329841,0.8936195878094744,0.03558320056554609,
 0.4728548195110934,0.9213887977612385,0.17036404691987284,0.6574328983169893,
 0.28268589701877034,0.8181484267563035,0.13887751416291605,
 0.8465430670790027,0.980805722819064,0.8400267968046422,0.26908635258492397,
 0.13594655628072494,0.3293339894595486,0.528610185017205,0.1466851298809576,
 0.6131763672315801,0.8700972388282652,0.8625648744762372,0.10774158485100638,
 0.5237156409643567,0.8527949456365573,0.288985866854338,0.7329348831251101,
 0.020223161455340755,0.8489311285601756,0.2865492195890609,
 0.4921448912522506,0.979522764087184,0.8556195628511575,0.5099158209830227,
 0.38510066653543795,0.04273434667143827,0.00795138139466467,
 0.849048411615724,0.6432899772389284,0.16645547758368218,0.7647897215551175,
 0.18086559112180856,0.0624211076318103,0.42459710311465215,
 0.4847090641285703,0.0026261073270322788,0.27491046480400316,
 0.18874662128399167,0.20411877359620823,0.7926019016791623,
 0.3914148007962627,0.6094413942904725,0.5947968431046403,0.5262887673343148,
 0.8410815410236884,0.7725298921655546,0.03873837754327236,0.8788774815827327,
 0.4819340875861723,0.3425985087808431,0.48488034482410625,0.5409456599800112,
 0.879209618726855,0.8288527409982985,0.6887097980379034,0.686466328879982,
 0.37910573392455715,0.09850458786246141,0.0676640743524578,
 0.7156235539546216,0.2139058207337866,0.9361736746097598,0.06935702976485658,
 0.1423752488453418,0.2389157651003173,0.4745856076814474,0.24560938254697384,
 0.854492482942675,0.9899504367565862,0.1868280763976511,0.3856911916857706,
 0.06753561902271787,0.869128630709136,0.6058128986912077,0.7581781516782917,
 0.6959465173466842,0.336396664838573,0.3616598746766664,0.8877462222650566,
 0.5708233939052425,0.013160782240388635,0.4880513915187574,
 0.02896200284590089,0.7688886450464493,0.6673269427469009,
 0.32872774211216094,0.9928595530937769,0.10355838517148808,0.62849185154655,
 0.2898773562575748,0.7396920644832248,0.2041930668277998,0.7574096817599811,
 0.6952520265551623,0.6664699156812706,0.019691274218180732,
 0.07788519895626922,0.393820888785535,0.9287761216806891,0.0730980725763376,
 0.1366426208537984,0.8411349724703256,0.7131692069363543,0.8702382193677244,
 0.2911607350684322,0.7033795016538795,0.820303228751556,0.036901862474824854,
 0.4733499969881634,0.7959313499694416,0.060586710859441784,
 0.18662090190394387,0.1741530363239543,0.9508638048186533,
 0.15709586866490166,0.04663252934792772,0.5567310455948177,
 0.12620728663743974,0.7624541198746425,0.3732648309341171,0.8029800015980105,
 0.06689213058223897,0.4070979617209767,0.5232713161415515,0.9020850120680989,
 0.7300175552000184,0.05959652163828344,0.12475430541295562,
 0.7016972784914279,0.37416673328064465,0.29220357263844443,
 0.8719988736579666,0.6447973550255996,0.6860172900478869,0.6789979314025887,
 0.7534484845718386,0.2970622361206551,0.6459203526154249,0.25359686264797454,
 0.14814790074012807,0.35237477511734605,0.7244130037313594,
 0.6924633306158341,0.2775145929009296,0.7849155309188586,0.16722409515103365,
 0.7603928802935442,0.9073761623999175,0.4661694733534759,0.16053209307287308,
 0.6060509526849416,0.5375333160746272,0.774804059028315,0.7343397319199014,
 0.7743507748989016,0.9515464338456318,0.978566357589571,0.08703739119449005,
 0.17528568062077254,0.15257662682221174,0.5535763518081076,
 0.11504591719850765,0.2561637469224688,0.021014627245167894,
 0.8403705786375463,0.5664980908179752,0.6552314737326164,0.07728179835141391,
 0.5387093770980818,0.17609309003178264,0.8116601272008518,0.5025408429533749,
 0.9979311663023306,0.9884929829337176,0.915178638376252,0.6802254569935143,
 0.5345417235573945,0.7184380766313105,0.9290471949583489,0.9360871618444873,
 0.5611032523879331,0.5531539850389949,0.9771498865116273,0.6320012783088493,
 0.8977406259973396,0.060269497439748076,0.9563985191156358,
 0.4046292253804076,0.23271817632863812,0.36888003732364705,
 0.7015775697134299,0.9678266336677945,0.1119305075976601,0.29094136926196823,
 0.18323648478368737,0.6872044534309977,0.901160539364734,0.23200365668098555,
 0.39662782097658555,0.4342462023813791,0.14163051585607278,
 0.5065318768186409,0.3425933452071108,0.08739083628410038,
 0.13789571041026927,0.5806614563261993,0.1349832250466414,0.4188574429062253,
 0.6907824100375458,0.29568375695414384,0.18557335373506867,
 0.9047801223589929,0.27842312814448744,0.9000779933676591,0.9049319682544653,
 0.36036515786179135,0.09159254707059061,0.18407258740726284,
 0.9019453305395735,0.6797442373316925,0.02238427614562999,0.9083456213625821,
 0.8338907267582021,0.6883400207539341,0.09835064278357475,0.726135236192671,
 0.656312153332838,0.7779960985371074,0.8327256284149859,0.07054040351672031,
 0.5507345859351578,0.27460581447733656,0.6362196091131742,
 0.10151814819707172,0.2444755517202687,0.12974945824148632,
 0.6672072581741869,0.2858505113213524,0.5124271280053585,0.05698096825013155,
 0.8264171122357545,0.43085384101756186,0.9963424229039699,0.7263968082475691,
 0.9523949994509755,0.6347194311203062,0.20623609316994207,0.2820857591315272,
 0.2253665814730409,0.0200166368587279,0.7567820580483277,0.5447247457699588,
 0.18820668168422583,0.7212410928335817,0.1452661638666577,0.9848609453635867,
 0.9369677291859799,0.6224565107051824,0.22428936640126917,
 0.31842854098075046,0.641082104786872,0.23852266202719763,0.4199229141281533,
 0.42469482919037727,0.7903145280314307,0.037082869479159175,
 0.5170699374731108,0.3774632183726975,0.2580510279900552,0.13823276411981,
 0.10563255032226271,0.6034601164796103,0.7325522486188206,0.7171092093092218,
 0.10383549061628428,0.9046787846842563,0.5114564251013707,0.9229198285769691,
 0.4295856858291264,0.9149179163085442,0.9302990081033556,0.718618238569036,
 0.24672801130004873,0.37563442097815836,0.08833480660670245,
 0.3734457238594001,0.8428691244985606,0.07012104946499831,0.856110276144002,
 0.660658393521538,0.3257763806550904,0.8294563525855505,0.3078820904302315,
 0.6239535528082891,0.341546684024263,0.8084099488073346,0.2773289958337216,
 0.3889523816965569,0.9499128939581544,0.5666180002457436,0.35976812864951824,
 0.9147636642422204,0.9637816189186734,0.49199396683915464,0.7985951721519847,
 0.43188893824616736,0.8832407789087728,0.6847362974727281,0.7854197887019787,
 0.5020459081980094,0.1461496178589643,0.8614918819978672,0.7114191678059296,
 0.21902565977421473,0.43538727275427425,0.11322807863900752,
 0.33035727670664305,0.08533580659724316,0.6218092455191766,
 0.8425543610453649,0.17169960357503522,0.2360704899333146,0.8068670069905789,
 0.3039992811243257,0.31889058278021,0.7751128577557099,0.5295800123588749,
 0.7050637079954963,0.3590293820787068,0.3139517040327694,0.9930743426653851,
 0.3877097441985681,0.8130398358856659,0.9675073734162034,0.6952691372204348,
 0.3139935193352509,0.701855373723564,0.0019795864981492217,
 0.5639965719813169,0.8736056134281536,0.19540105035946487,0.8830703892949276,
 0.7145430472693741,0.36368107391706184,0.43521882300679393,
 0.27943002412368656,0.13768667832206427,0.6119943502155166,
 0.8649663295077618,0.7239859243053097,0.10477597695383911,
 0.13889164958823308,0.9313393912726817,0.8231167710259902,0.8005947363318939,
 0.3849015139548224,0.8014608763807897,0.9281847069533499,0.41972369415564514,
 0.38706746084876187,0.8674120759995066,0.32175942081943254,
 0.4931898133585305,0.12454335353820523,0.05611608298199611,
 0.049699683715678455,0.08589071767970657,0.010218501117781442,
 0.3179055316344538,0.37779424171904696,0.2880853992282456,0.7805359183393623,
 0.808643867351833,0.6064076518768086,0.14608567023713503,0.39891224583158813,
 0.8076277963754007,0.8732217315630345,0.3161867006484389,0.5508548474375268,
 0.44262716933116353,0.45692473234606823,0.49888009414438694,
 0.19222715572708038,0.17295258780922873,0.24152445718488824,
 0.9971667037213195,0.8859183819633843,0.03747207451175372,0.7700885647954537,
 0.5715873147138993,0.32978458376169295,0.6314804527251434,0.9449958092133857,
 0.562679385960669,0.4733682313122701,0.10751819204406654,0.5993269988605376,
 0.13135051335140835,0.7241053073279088,0.9861103570649639,
 0.26121435140119764,0.21336478371460432,0.41161017371559605,
 0.7269811504157233,0.39938235617222995,0.8536310892744541,
 0.42076415860355476,0.7467561473335435,0.42595135703812914,
 0.08076734570991317,0.30111481766422676,0.8750529309027623,
 0.15833061977406016,0.12792198185799564,0.3851979675636338,
 0.9231015439378987,0.25064607003599804,0.9858175605844853,0.763465254794003,
 0.9976223564386587,0.47484355513811616,0.530437981900893,0.7958179829490116,
 0.5233558991684308,0.08622256473666512,0.06376313945989609,
 0.6594654788649528,0.3725657937637038,0.9679687579487956,0.8684374354031824,
 0.23460726903376072,0.14538794037980285,0.453500876681989,0.6840140165724411,
 0.02973715178252956,0.3238427588501328,0.7645262297904515,
 0.34722581351353743,0.33973158303570306,0.37426078474601243,
 0.738212511632072,0.7392187714342501,0.06316430635239723,0.08036113967224479,
 0.3723914401419579,0.49146697825458585,0.22679905240477327,
 0.8937577332347807,0.9692690474356666,0.02672678355612934,0.5604551912493655,
 0.6969634542692431,0.5019675712711864,0.6083218855270401,0.7909813302990503,
 0.1958779210883178,0.13250232149657448,0.22101012038375745,
 0.1362528613788203,0.8884296568186357,0.590232515395364,0.13562959084853243,
 0.9336940510378229,0.3969216213605291,0.8519205431351906,0.871231100024564,
 0.889127759026357,0.6253683144241784,0.20286066685948245,0.16906490215592407,
 0.3860543387209501,0.4845326845231187,0.8588317872421661,0.7327281085373343,
 0.8495346055318692,0.949625228617087,0.46073257753503283,0.10947154036269946,
 0.29942556210999294,0.8751801726143282,0.24376667765870563,
 0.26971735024171073,0.34578231414170224,0.40300320339436413,
 0.7202571840580813,0.8946045967348416,0.2649733023296247,0.43094864074710615,
 0.9234472948917096,0.008218841119155829,0.9473732540113763,
 0.48676058758649066,0.01927661269072689,0.3116257709109541,
 0.8494116107430922,0.29078619998552824,0.8084822085964648,
 0.08081046133157477,0.49411861908053356,0.6413127354584285,
 0.38558643258310377,0.025768899171694115,0.9737581908211757,
 0.9338334626929581,0.1013132819523005,0.3200141135801806,0.6666434837744855,
 0.8433966764985954,0.45165206659990254,0.06568844572998411,
 0.26785140516597916,0.6420758927671179,0.5835815283057164,0.7162200260684942,
 0.579458756827281,0.1843051571025549,0.9866383218366788,0.7511068356398614,
 0.24157313813400427,0.23471911390082834,0.08715379184081329,
 0.22216914598968762,0.5095988390667602,0.03749743094045799,
 0.30398169471211667,0.7413005625244042,0.37004002963187366,
 0.1952044246872553,0.9984743384074101,0.5228134931544584,0.20109444500595963,
 0.965728021939546,0.30466147214631645,0.0015405278693093605,
 0.0771322643668827,0.02587359504733,0.2927453822284214,0.9144992846885571,
 0.4830829483455681,0.09194451498626288,0.8663436515480927,0.9208187400815253,
 0.8416582276649031,0.08833066666639544,0.9671901064707132,0.7323415901589916,
 0.9221190792017224,0.9750701453950642,0.24836327692677163,
 0.10581729595431177,0.5075892854217451,0.8339476876549121,0.5368674519387902,
 0.40161658266542877,0.2282711564343709,0.11351706024404928,
 0.7468908847692387,0.9585770773204833,0.5510413038733136,0.4640014504555441,
 0.9757752611366352,0.4841822351886452,0.3269306590028904,0.43524279003048594,
 0.13613216042289789,0.9718130012189682,0.17260374072446016,
 0.030876991065706072,0.28740302795191397,0.22237480302422608,
 0.8968251066270364,0.578719127788343,0.3899295645893296,0.28069530403901777,
 0.7791842318604932,0.4453533381950395,0.28953908780172133,0.4461583735058371,
 0.5085573058225236,0.9788536230261573,0.08345476984378086,0.2505280193316861,
 0.2552634064681799,0.5790360900326443,0.7289326940476203,0.47727074077327214,
 0.575540201547708,0.4718127918830022,0.6644535977933873,0.6936484656437827,
 0.8235357702567022,0.11088246328129492,0.5736452096755874,0.404370864233627,
 0.8703161373713533,0.3291533652801065,0.33021888476390315,
 0.14514032622628048,0.4522883626473083,0.6739647523791709,
 0.19275809620472506,0.43542048769675534,0.5689493933789289,
 0.6834372964774135,0.7744856644467658,0.10019718126449673,0.5701529008880897,
 0.6739170129773486,0.8144057886162277,0.2430451075229404,0.7334580196822417,
 0.9136338179575599,0.8653955895889388,0.6457600001331141,0.4030624012005073,
 0.8637976911620802,0.5247475184877601,0.509557554901709,0.8397488570770797,
 0.3139964017840611,0.04070801594544193,0.836011149714802,0.9680050468216809,
 0.08127907409551927,0.15308812562933738,0.234900714979012,
 0.0027512188289250394,0.15300664734994907,0.6843620601336827,
 0.2092079454778466,0.1880599197081393,0.3641885823283484,0.09186357608518125,
 0.507938884356776,0.2443214677227432,0.8999119645155373,0.3387571327564556,
 0.9076603293996575,0.8680344942719586,0.09345713309212322,0.7169115418501968,
 0.19719610833684142,0.6130346928397157,0.4278396479251946,0.063419506362832,
 0.1414139512659789,0.05841065767784437,0.6475898002120754,
 0.20267049472877519,0.7682508131539664,0.1810791227812133,0.6491942122147396,
 0.7443654709615469,0.3236533461919614,0.42361894558405233,0.6897825533794633,
 0.5370402330679691,0.03096018559729119,0.5144792372301434,0.2820311702135977,
 0.3583232956397727,0.8892936143987253,0.09138701884285094,
 0.09378247730792011,0.36139322475941005,0.45783731787786397,
 0.13133912669105907,0.8134079207004851,0.07077031003561707,
 0.9033780127841201,0.037804141937405,0.46898032556638036,0.30960190582189795,
 0.40587348640429133,0.30432229762755725,0.6503823775392036,
 0.8798940047758371,0.14676202430086338,0.3539607726547286,
 5.004049598463789e-4,0.9545504725319591,0.4916318286646051,
 0.5341987526535719,0.2975411021008142,0.6154514397633011,0.5912878141763859,
 0.9642605572796524,0.7348696082933426,0.40042993043773145,
 0.10802886233064124,0.743757134567854,0.5065452881746103,0.26858152680313085,
 0.5612989768009132,0.2621722996149487,0.847212797627507,0.407774319688595,
 0.7581057234143135,0.23903721949404122,0.5635875203721689,
 0.025019716542705517,0.3868890558369146,0.27879949054464914,
 0.8974641442145443,0.5163246857717729,0.8884117594064798,0.6878920307347481,
 0.9540499576639442,0.6737115450203122,0.8183126437542494,0.9147190277333214,
 0.5797129612405348,0.20168720005731133,0.7614204546010056,
 0.14072772095106068,0.8342921288995491,0.15258454747290384,
 0.3698090657649935,0.2437567652164505,0.24267059988243433,0.9229969380013627,
 0.15883571557625153,0.42602512749112365,0.16108665581544646,
 0.9210035751291831,0.4401204773705578,0.07700635121020705,
 0.08533576407955845,0.6700307070668611,0.7220414516628157,
 0.34063843476684463,0.1537952606662768,0.48928215553475496,
 0.8949766488865462,0.8269952914016656,0.3399156430944046,0.6831046161055635,
 0.357680795526072,0.2619986974039241,0.7779997116192027,0.9111231450981386,
 0.7053524552248535,0.0682055582320753,0.7654686135842248,0.9174324637729661,
 0.40070953224438366,0.15912556808324507,0.9348592208372075,
 0.8407296276647771,0.07356605572003172,0.8477739902425842,
 0.19541406092577496,0.08250667215325636,0.8661177604002901,
 0.12846573015566365,0.18687626147930636,0.26394114943890024,
 0.7839335167752881,0.9883677151814427,0.30206050569063214,0.609728110036862,
 0.36928822622927737,0.14520361236988955,0.5760568683318129,
 0.8305222299408099,0.9822362601965432,0.6263074263115692,0.6596855009275187,
 0.7756313153347043,0.5414312651130705,0.015645792376662304,
 0.0226499397354073,0.30456343097268024,0.3639820620594735,
 0.22381278562983198,0.3265615221862165,0.446837368056632,0.9933427295591198,
 0.6145255560992078,0.5844871601209988,0.7761508994046007,
 0.014067820849122503,0.12350987656267487,0.28634784948274805,
 0.5478805645582885,0.9692504977040981,0.55129527659123,0.9479585862457718,
 0.05813441703360456,0.43346686329604167,0.3779312818625259,
 0.9447795756010972,0.18923071181733797,0.6647103647076503,
 0.007764792232480056,0.6654223590818744,0.19952985162371706,
 0.20268924189612147,0.5347664343191446,0.10003412845965298,
 0.3455566765190947,0.06987454646326341,0.2942665468920981,0.194458676252118,
 0.29295779525890275,0.7664534643019504,0.7325021530880077,
 0.02977770107226918,0.9669131924882344,0.2806729817163047,
 0.13529042817013504,0.11961485582225184,0.421344307034349,0.6461834934908782,
 0.9814876058391822,0.5751716088695951,0.6150992617004643,0.5760745221083046,
 0.27531773142825444,0.6162148659007517,0.30718637924741576,
 0.9728657134103471,0.5180157780550901,0.6194796940926027,0.6016841541262061,
 0.8775817272179185,0.30725955955658246,0.26578788858404734,
 0.7589519210046796,0.9302363240553476,0.41153025868742843,0.5615695323683246,
 0.7771441471386192,0.048313857037819985,0.871579530223723,0.6181937174295755,
 0.08604126013229008,0.08878894423475447,0.7502828836469346,
 0.9256286096829804,0.2213981918104655,0.19453382384441953,0.9766735947927203,
 0.10546878813043303,0.253201965110034,0.08744073781507344,
 0.46353323584866457,0.9940040671155234,0.943771707661691,0.9422893715272602,
 0.26788290739643394,0.06497851477450467,0.9172222408459192,
 0.15842691368097028,0.9934040989343621,0.5318255959027133,
 0.46241576557348085,0.6545892928289525,0.4702092528998386,
 0.07996665498607425,0.6527624780325185,0.20985238883089552,
 0.9512956146148401,0.07175261224392859,0.021680904084139696,
 0.34501927590758485,0.0019105706498695474,0.32987667083143857,
 0.7108326378758721,0.9255850458682536,0.6715811142955977,0.43375360385897577,
 0.1373758442571439,0.9457146548586743,0.5079153142069585,0.472545144326031,
 0.569000792329207,0.24673640128741636,0.35856751347472104,
 0.22026571878435353,0.4980561371469475,0.7435551462573552,0.1315243470433941,
 0.5952077041366316,0.5934683056044393,0.35146305520261323,0.9762673507275647,
 0.6255201855061328,0.541445070671868,0.952187866952832,0.6440215352807634,
 0.8923825201908783,0.1511164660600327,0.9385418719059949,0.9828631884668216,
 0.19974163492329933,0.5839595824849687,0.4905530397177873,0.3667946252706,
 0.5491147776770241,0.22499235718881816,0.5659373206992824,
 0.25029527946340546,0.23710981733714154,0.42479982003012395,
 0.5401764675366971,0.9063509242956576,0.2213045260398736,0.42275822928246043,
 0.8553456503971671,0.5588089164198006,0.4378574976969176,0.8513430490703463,
 0.09038441983781842,0.8135747904259195,0.6851218962398932,0.2806049481880225,
 0.03047413195316273,0.0071579885231976,0.9009568326090505,0.8562252423084156,
 0.2192533753004564,0.46462620678514244,0.6239742530282963,0.3917452538203372,
 0.6760769633946427,0.9231336241539947,0.29653533584047687,0.328375384150439,
 0.07796975211077128,0.5510525618353194,0.8960176169742774,0.3291823714321964,
 0.16930496838316744,0.6831392008333638,0.008025707681014072,
 0.9270907647186042,0.73802221634742,0.6995745262009783,0.06644490050591101,
 0.22615414241113885,0.598804592017941,0.14816176039756201,0.7520838895873774,
 0.17553730586646266,0.5273759016426318,0.45799894362782556,
 0.8612277384273527,0.9374557992129251,0.8003154274845433,0.4553181137578277,
 0.32730962932767316,0.8514252572171888,0.9450454426484225,0.5254924083879239,
 0.17127354540407447,0.4088953330862397,0.5388309386825844,0.5979130170287935,
 0.5036599162400199,0.700108977011946,0.9513008323158082,0.4862432167437175,
 0.8933936940822303,0.212451227869201,0.22672926227148915,0.78368122440449,
 0.3640200529390023,0.7953570716647262,0.46382520546915507,0.6464682363115231,
 0.45599227291186784,0.33407055937555075,0.6510158724600571,
 0.9411134179356814,0.33471197724271873,0.16059686982951848,
 0.7419618957730874,0.7468226880283036,0.10521583962700765,0.2540131878801736,
 0.5105191418192939,0.1406760062269743,0.46286156527938516,0.4401311066824103,
 0.7670218182420491,0.5547594643276306,0.658779194280834,0.6859257187362295,
 0.7381444133824732,0.14477253980179738,0.4117299195415358,0.5376591676931212,
 0.8301517662705493,0.9894776729445903,0.3214604100342635,0.8708757607141275,
 0.10186269779231272,0.18354061552067147,0.1107082896127507,
 0.7986544444223831,0.042203588233824973,0.25725201259440555,
 0.002854904844768913,0.7119895651373438,0.4758755139695914,
 0.5285705222740941,0.9175645239316925,0.4560719633790621,0.2513519559490348,
 0.29922457627756194,0.940963745062466,0.28497849120338414,0.0207435380821428,
 0.9024059977508089,0.13537818309811656,0.4781152116040651,
 0.24379782613401157,0.5860755214873459,0.3531952937949393,0.4594466460463593,
 0.48868884195904205,0.14533155620180827,0.7856569894332105,
 0.7026628086185543,0.10492595243278946,0.6180921418845275,0.9795690071716395,
 0.6853202803068511,0.9804759316660366,0.9810761588878651,0.40727522320458076,
 0.0950693060100316,0.5446956041433285,0.746230612808269,0.9791190496066877,
 0.8143480503707223,0.7949410920055896,0.07555745597879515,0.626567818052025,
 0.9388641585458843,0.28382737149235804,0.5434658677980613,0.8659747405016591,
 0.41669842942111335,0.5373461666527737,0.09351177857897519,
 0.028922872992357296,0.222594789516514,0.7213468253501248,0.8739138288052329,
 0.9178169316699825,0.751880903589331,0.4966417033510365,0.4208036448817758,
 0.1269818278617949,0.5856162220401158,0.2620061386791315,0.09359126245250782,
 0.32640827322281196,0.3348598241260815,0.5688438498732045,0.7662078876328107,
 0.20710912872090081,0.7532716132466872,0.29967316032264624,
 0.46939543733503786,0.11145337859964954,0.4797552157653088,
 0.8989716948548346,0.018839938790680466,0.7958871509258372,
 0.8448591419927282,0.3096351321051939,0.5994977714969626,0.4304723098820348,
 0.5425820274436488,0.1957872167890038,0.011007287108330654,
 0.8297009495101666,0.9017952364814575,0.06076774253894368,0.7955027815883904,
 0.8626564346581451,0.4586015094698307,0.380353049170159,0.3282738546800549,
 0.011681885293340688,0.2909343440132326,0.5406148335682671,
 0.3805045766761952,0.9047640278561688,0.0915063889591704,0.5069016571660222,
 0.7968526907768801,0.3113733664725904,0.7353668505471528,0.8882446050508737,
 0.11128722065138874,0.1411395045999264,0.36013833357299707,
 0.35596764556967386,0.4556546467475857,0.7068713128771443,0.7124993090471963,
 0.7980982429509395,0.4969424832329965,0.7077517974383287,0.894727369546034,
 0.5999695186393944,0.3741060492788889,0.2372479467816584,0.9531210437705314,
 0.2925579229379579,0.8042346011824011,0.9781022477023614,0.8102382669915879,
 0.5407789668162806,0.2676114405177976,0.3384052297766196,0.2581170383465854,
 0.3653761881578028,0.7635956209854565,0.48043088085736974,0.5320513749406811,
 0.6200121714982101,0.8068074277147319,0.5133623556064396,0.6711620698528169,
 0.6674904318687211,0.9948599368654906,0.4260182490158303,0.7770874557536316,
 0.7217898231835186,0.3326602415267015,0.7586971130347798,0.435071581885893,
 0.2680172578931357,0.7321615578150443,0.5950736299617331,0.8784364701571649,
 0.08311033571776305,0.5581990544416358,0.7379648111912134,0.5873881437891714,
 0.5342601470017176,0.3355921692818332,0.19861616938226845,
 0.16120159088012564,0.5889190568426581,0.7521214391333658,
 0.20957926394451498,0.12863589067572756,0.43828872952334286,
 0.4133605778732926,0.4005875352385143,0.059588194555295404,
 0.5818645372430644,0.005653999541416432,0.7054313168435914,
 0.35239896106907265,0.652859363172216,0.43272367974827197,0.55855835490566,
 0.3608228490671732,0.3717470267606948,0.5334062767504153,0.8051402543767062,
 0.9827721956512232,0.4925758758436811,0.33466103866833863,
 0.30559076033064747,0.8907884190937952,0.9380486837297052,0.83136852413266,
 0.7963895455789094,0.7378235951794755,0.43565055783400664,
 0.07378945048097074,0.26238102083364223,0.807310618753549,0.2672159878259205,
 0.33749677201733075,0.7877885945272332,0.30062053154104873,
 0.8823421402833114,0.1509051588505702,0.6988967001637222,0.7452824291794464,
 0.7139672282772422,0.9889544852417353,0.749008890853333,0.5808826638340242,
 0.5524003324584978,0.7526884873846338,0.7588618410121712,0.14720854003885575,
 0.9397589620473479,0.355952373082407,0.0923711368495096,0.16446740976960317,
 0.748512542473081,0.08248123080110747,0.7068902160879557,
 0.003437221828696546,0.2896678173892966,0.049205147066108124,
 0.1410720925928799,0.2847955506214743,0.13663935775220215,
 0.14154620620335934,0.6930198273700902,0.9037093246587955,
 0.07878627939310412,0.9257677293516412,0.552152347862276,0.6550923136389941,
 0.9371656621254854,0.20367347044124906,0.39893314839614435,
 0.5845413498846175,0.47570625177430004,0.5326268249396543,0.5115513830299204,
 0.03319023289412104,0.8192424158549063,0.7838305314749031,
 0.17455153870789908,0.7362816116851861,0.9386552121207619,
 0.13204083804213196,0.4909896825059089,0.04873418355580228,
 0.23807499672533528,0.8644454993368861,0.32014618981222454,
 0.6021565406660512,0.6150675382593754,0.20377130652816244,
 0.19140148948400304,0.48873491151138604,0.2530879156113953,
 0.8108038292085549,0.18100268369948513,0.14576362697637524,
 0.14042721722379292,0.6109513945599536,0.6158002751872205,
 0.07687843559725627,0.99069490054666,0.8431569186330556,0.3590718168590441,
 0.19531453612747585,0.1320358142697201,0.7640641362447453,0.6585540990866089,
 0.4413508272555676,0.34897088969942713,0.7771599646474107,
 0.34589877528722845,0.895129114890138,0.8064344576225934,0.5189039656351286,
 0.5125785076747275,0.39450469424898094,0.30375905059540187,
 0.8570690086324949,0.29536378649375594,0.15410232876051322,
 0.5582478629703581,0.9984787589290598,0.03352736271590728,0.954198499789717,
 0.5629796862028457,0.823360413271041,0.13408113147050194,0.5882602247810457,
 0.4325915513825296,0.829503447138714,0.29661086395179914,0.686635753934616,
 0.27528668292449765,0.41136131129775955,0.7230415735056681,
 0.9169503067691704,0.34022580999404417,0.6187253153101747,0.6432214803121376,
 0.9378411969336227,0.9860308697960679,0.6133479620018499,0.2825083937931162,
 0.8687461409324644,0.23013254480865508,0.09655038303981989,
 0.015180477400700054,0.2834163520037508,0.9381211103507516,
 0.31049019834290803,0.8361847294661776,0.7700072032676237,0.6000519633696719,
 0.37189173711493895,0.5376074145167468,0.4948271019636755,0.7634105329575537,
 0.6473464612805342,0.47399686498980587,0.5713310008665349,
 0.44262823315264976,0.3322946043744557,0.8240656826512631,0.889483212638126,
 0.7820488702072635,0.5342055511493559,0.3806871516909889,0.7388572257284727,
 0.25662448825125606,0.68397956485564,0.7599998284835038,0.14886487469754628,
 0.44435759262423813,0.1576603812313575,0.6480278370571216,
 0.29131644298037807,0.44397320760027303,0.07391642914154062,
 0.17672981156421863,0.8906944967841715,0.4051565332229964,0.7609031094449548,
 0.8041778936502169,0.42841733112023017,0.5097779193224897,
 0.10016141845264781,0.8733016532553945,0.21869061305230675,
 0.8249512924440849,0.11397780287726988,0.38118254570950305,
 0.9834690428862762,0.5276820577028525,0.8963712803363199,0.21238610764409205,
 0.6931070327519376,0.06211612635201702,0.5213102471887443,0.750091724473454,
 0.5799053575363575,0.30558043273667634,0.14790039683486034,
 0.7053318096144829,0.5553792197318141,0.2494355294501207,0.8826977472738629,
 0.19441897501252203,0.387355637620446,0.791543854446026,0.9809130909850332,
 0.7341350564986306,0.3062797396665221,0.047808019971306015,
 0.7739242751022811,0.04400389367804758,0.5072166344700298,0.2512414587723234,
 0.017912941097000923,0.8191507881674315,0.016582473212526372,
 0.05044543249937994,0.313000800354188,0.07796194941072654,0.5481205910679731,
 0.26177743970234935,0.979934968619929,0.7001553222788852,0.7599446049531897,
 0.4279032497565577,0.014891465032343398,0.07017722587812059,
 0.2164172424712335,0.42418954207411874,0.5990657106150301,0.6689187016231292,
 0.06947666903322658,0.2653649496266022,0.8852789818839738,0.1951008096417276,
 0.33324853851377734,0.6376418374630106,0.9549047575968842,0.3545403247023704,
 0.8221541395156834,0.8840565573078639,0.09791470015892978,0.3419901722556572,
 0.3490913811301667,0.37089469805303144,0.20025910846049488,0.830260745448336,
 0.40891411176933334,0.01205838932934078,0.11395367363137232,
 0.06829893045769708,0.29345167382873494,0.7322303148199473,
 0.24978528409436584,0.7542842563293539,0.9202465885221685,0.6016475342904145,
 0.3599554140331115,0.8841848434072065,0.112322194132777,0.1545051166863639,
 0.8922596559348567,0.34958729339801065,0.12233413139743021,
 0.8393538766331072,0.2165419165090754,0.5701664428729649,0.4267687019713734,
 0.8264659192590216,0.2778530468625384,0.5718679333621861,0.9435605832251375,
 0.46219189720592646,0.6642283537838525,0.2565650326712402,0.5124965148098993,
 0.4493719831517756,0.8982728778515906,0.32620572458598274,0.9810917043819567,
 0.13750598054427177,0.6425368148603211,0.6088583504516655,
 0.44151398853224966,0.9581773379367913,0.29694770119669267,
 0.2506388670005034,0.8347151868151998,0.05705433868623211,0.7652704031673573,
 0.8040360241512459,0.6330709842728499,0.5783407916077405,0.943872889846133,
 0.9684896487247354,0.7283705123269243,0.5158078685525704,0.03056342712212967,
 0.5009919290789472,0.5760766371991395,0.22455270339748468,
 0.38346544084615775,0.1107680755905126,0.8836733959313187,0.1622530923281631,
 0.744571457150166,0.5778287714438637,0.9476064687028016,0.2599382662902797,
 0.04660622815491744,0.15669577104596966,0.5176761014475977,0.935703944727196,
 0.6641120819041726,0.17351291186092033,0.16479849317801398,
 0.2721944181352811,0.5583776808379809,0.3263566665593908,0.9771221893002402,
 0.7851470482191953,0.9895177002085939,0.9282038896625411,0.4415298689415531,
 0.9280782837024947,0.039412863250327224,0.28549189715461554,
 0.8874836059609865,0.03476592684699731,0.6813880526309446,0.8484018766699686,
 0.799845921957411,0.6354348371848142,0.1308172204512138,0.09447392171104385,
 0.5430790976923073,0.9652922888233242,0.05636945380374769,0.9057977206694159,
 0.07140239539126325,0.5209150256843955,0.469915540147827,0.8575645673204573,
 0.03226533154944666,0.4876451742084703,0.6360598850982506,0.6788038669192878,
 0.2072973209034381,0.14141160291531873,0.6216223987883132,0.6011978083922234,
 0.5972349626880206,0.8191736590199683,0.779960325153306,0.08995794891170361,
 0.7122584174499833,0.3811535884596551,0.36925479277850126,0.226453466342323,
 0.7721470770182097,0.7844815771838919,0.21815364178628416,0.9970903723519879,
 0.047682447141429884,0.8681590039724622,0.7261579639659405,
 0.7086201806073523,0.035010734732541016,0.46763095592110526,
 0.4654281524637982,0.813881844462046,0.24786856529894952,0.5868959494115034,
 0.6526269542989269,0.20431209840384335,0.45702411092740536,0.68119501443748,
 0.07395288335206884,0.18892946344775607,0.6800682274774716,
 0.8822784718117429,0.6055650242593552,0.8986596771520764,0.40886807463327435,
 0.6390934123475883,0.5023785815408788,0.5599503743907905,0.3634281691610781,
 0.16041901140481007,0.0691030269553149,0.8917860373887578,0.7002533392085154,
 0.6575129826031287,0.4906671048574375,0.9748582802871117,0.19367756163190264,
 0.16783552098644816,0.6840919334844312,0.011122764035404242,
 0.8328322924919199,0.5905002746303697,0.018785415601420885,
 0.9329292021997594,0.7555255137342084,0.004185462870877954,
 0.49702081292436784,0.96800988323508,0.9443968482882211,0.7382092417492002,
 0.12993444976652535,0.5830526227139128,0.5079146686145198,
 0.18508851600651943,0.517249003848825,0.9620539985833669,0.29992869243864007,
 0.9480019162417239,0.6900733470733109,0.8998818539748308,0.2767678238696135,
 0.23534480476810204,0.8218609972329145,0.8682881113412451,
 0.0021966909438000393,0.06278942092614026,0.6607915584841443,
 0.11410160335615172,0.6327869412754175,0.7511225594941779,0.1981748881627452,
 0.25616914205845365,0.7054617749505887,0.9732973483617937,0.6377686610582051,
 0.6356574586306687,0.08215678933437143,0.5173161413955146,0.6969049814945003,
 0.6076815075968894,0.12205646670598624,0.5818747958655461,0.7110685058866466,
 0.08859602018983448,0.386882867382242,0.1760220377271755,0.8446812908339023,
 0.9099460068994231,0.46426122091022903,0.3884034263676127,0.6251279157338017,
 0.6683491085520303,0.30684908081398665,0.2914608822248904,0.8092949740263163,
 0.44559647631805577,0.11177150068659314,0.5917833574080009,
 0.43842907785165053,0.7216263113538932,0.967831425451826,0.7507886766339463,
 0.2185190110681725,0.4603563660584119,0.07399063237281966,0.9273926034006948,
 0.9007740127629227,0.4100597324993862,0.7616063723027924,0.06675196620104995,
 0.23331630354906963,0.4366074725881306,0.8973312629713924,
 0.22197083427628606,0.7065673460313747,0.40719021800130384,
 0.005280586120615549,0.37038301780554905,0.09925136103801657,
 0.986486371102546,0.28257152809257813,0.5834709158806205,0.7066427051781963,
 0.8441776062109037,0.22145572577522654,0.8633359924728777,0.6758318615631733,
 0.8254875174560613,0.8640225166921557,0.38218009977274703,0.9561335723604465,
 0.8811579989211352,0.4912024268330728,0.15334424961085324,
 0.34242874440329785,0.7427155491006846,0.7237257412000565,
 0.20970822434448655,0.6406346983251321,0.28813667350417693,
 0.9443551170458011,0.35958790281140673,0.2627150551719595,0.8448390045028085,
 0.6947053883032821,0.260070986958161,0.9579411755762632,0.502076555883505,
 0.3110424849721096,0.616016042483694,0.1570473964972332,0.6484142710189456,
 0.8248357723336159,0.9398316234965228,0.03333743164652858,0.7171637225608798,
 0.7253681023066548,0.4752383048430876,0.6205816109258157,0.7611907111819356,
 0.3053444853784214,0.35938815129248536,0.8982671576092549,
 0.008325855589941478,0.7044854508864261,0.40401840096483355,
 0.7641690033601365,0.38990173028314745,0.48299063927760555,
 0.9208697490158695,0.853997829272847,0.4014123427487273,0.2309217751889716,
 0.020740851509476133,0.09398442214278913,0.8667704156146715,
 0.3669037996502996,0.34548274602226037,0.5464140885660748,
 0.29912871940993413,0.33197188280466694,0.09097318291471856,
 0.16616757624983147,0.2821054581915786,0.9514879353387852,
 0.20167272608828446,0.9551898099369044,0.8487597969874969,0.795704674424867,
 0.7466943628664923,0.7268871836092663,0.43854348435120094,
 0.09748330662991311,0.4950439069396916,0.6390487906855127,0.7785722392466191,
 0.010097493534992674,0.19869956785403653,0.1949888014762592,
 0.8011272003172177,0.2925517511331809,0.8653309088500611,0.6760882225563074,
 0.8237799027458935,0.1650818337983816,0.82685395684402,0.6850383590124951,
 0.6130705954905173,0.4727642270981415,0.6029767972684619,0.958574316666565,
 0.5042397971604089,0.5874488443329616,0.7480868550364055,0.9254901489910072,
 0.35970666104776716,0.25621918742314054,0.7927159006649879,
 0.5422135460083852,0.7876612575892097,0.006464108894689957,
 0.7569695531570151,0.3331865773762177,0.633239134140732,0.33635835617318377,
 0.1260994129777312,0.8157602643000882,0.4379827110387393,0.42334211090005336,
 0.18081873052783004,0.7526933206969953,0.5114381826131442,
 0.22118466332913422,0.2624590895606065,0.7298372483809976,0.4066819786066427,
 0.17075216519572722,0.09493614161574737,0.4543971730609879,
 0.5301833190961893,0.4763460367742144,0.7132075137620324,0.8174479071328741,
 0.6444650557632577,0.5272607039198313,0.5036925310901263,0.4332680335302276,
 0.3135063831796109,0.35125918732857075,0.6293183229711015,0.7123565531063223,
 0.04237030821093002,0.9764099307522313,0.8719225780138122,0.6555154756728409,
 0.30122470412055324,0.11692856432475629,0.3896344344899524,
 0.6597772892269185,0.19854955714547562,0.25521674910744374,
 0.6841488367156003,0.9392723487104099,0.27278770335938385,
 0.24528785932596753,0.03277457856742516,0.8448301799827566,
 0.03549322955863787,0.003959095389541867,0.07923740660367506,
 0.8983492513174066,0.8244195742911657,0.25996873466276504,
 0.11258223575323822,0.5902244339159881,0.1306121784740687,0.6192065116740675,
 0.6251411748164669,0.32192063139702354,0.4346907834163374,0.8448326606791314,
 0.2926639470040965,0.14871056988072762,0.9297522230883652,0.5966036194486809,
 0.07379784990899996,0.0837028983240804,0.596921431971659,0.01836843531936916,
 0.9734627390218102,0.28390195579661437,0.4884105021570623,0.5288585799272838,
 0.36299833593042374,0.06504437049764,0.4983634229505607,0.9043053477852148,
 0.21663066351194327,0.2073873046268303,0.44598929685403854,
 0.14924226230851145,0.16409119063886324,0.8229704781969681,
 0.6687930083854667,0.49807527875489777,0.12055338658050818,
 0.9017170315370816,0.7801460293279827,0.625982738306743,0.22479865789918252,
 0.44161221734004275,0.7296603261849391,0.7138000592454117,0.558451818068989,
 0.48318137660416904,0.5027137020557577,0.9759520445196068,0.9925380312862654,
 0.5627403666943962,0.2225596896445552,0.7272348983465424,0.5219582747466998,
 0.6877924599729829,0.7048711006137778,0.09628410067861581,0.3829955517472293,
 0.18005186842641274,0.3235085708366692,0.34196562171029155,
 0.9812417691363718,0.5658533300125839,0.43204489290995773,0.5771874615095721,
 0.7400905138914169,0.12381755583919385,0.9046554938596995,0.9697869715101007,
 0.05570595454436822,0.9874836214253635,0.14073321857522292,
 0.2835458328560101,0.3790186926505976,0.08748847948057481,0.9680041812252194,
 0.6964785869927923,0.6256730715781327,0.512994389254358,0.7094634456372403,
 0.6552882577442329,0.5073255188552123,0.03299116742413444,
 0.48754757592927134,0.030000884154529284,0.6505969454939661,
 0.6802437015890322,0.22467905356426698,0.5885114417034117,0.8949762853234741,
 0.0019390583810265927,0.4173302497897018,0.09340865610105342,
 0.3714515097594038,0.058477835759011665,0.6959617655228127,
 0.9995133397158483,0.39255481456576347,0.060324703320782436,
 0.8468967196092145,0.4275290206999104,0.024361048488773962,
 0.19146578241302037,0.1504123060113136,0.8206536695398347,0.6345298050365428,
 0.019708780099290113,0.8004191908135015,0.0027913398276386836,
 0.8418178291627239,0.3360042256526736,0.09358106495040508,
 0.27638036982897796,0.44751583587256505,0.054218800221851104,
 0.6481492489335658,0.2496247444541383,0.09748876156837816,
 0.02503283851854865,0.249926093416594,0.07490586729916693,
 0.08326086385637455,0.8421562959818354,0.5232968313025731,
 0.40130423885302446,0.39781426581862234,0.9825524760837331,
 0.23499300059187944,0.801635715963007,0.9879536021512171,0.25524641445091256,
 0.8201547917582684,0.6691208483463391,0.2040588098244852,0.6557541549675777,
 0.689896342638658,0.17503605621173035,0.3754770753217046,
 0.031208744511038944,0.7074932668355095,0.693949635116371,
 0.06378071581907074,0.10797904236143951,0.7318809137803224,
 0.6267932568738208,0.8813453332730464,0.5964939469291545,0.8156464721000088,
 0.7601404664459805,0.9094347467823282,0.26520769961736823,0.4659124173102597,
 0.03328433163264499,0.34361505102572343,0.45429582923684897,
 0.8112237195920664,0.7014395461307268,0.7492583369166959,0.6551386941258067,
 0.6303258015885003,0.3612254389120042,0.6234439578247198,
 0.023159453553607934,0.43346008489046084,0.6155792605541206,
 0.40935687223511175,0.8932177477472203,0.09133555712937746,
 0.5523794187665115,0.2741428853894512,0.7891031582838544,0.40310081361944206,
 0.21365246514414304,0.8327699442729674,0.6608488442205923,0.9963645233019145,
 0.6539368512118009,0.32700708030779513,0.28395393259919177,
 0.17448763051164218,0.6794556132717032,0.8644613247938517,0.6348928957717299,
 0.920603760899796,0.08714040855692684,0.9855442911203312,0.9286676340863891,
 0.9191322989622911,0.6495178565570452,0.09811718550022464,0.5808163765866367,
 0.42812288122840325,0.001958182656094598,0.5215342796774695,
 0.9893059741841279,0.5525277986436212,0.9767975187616104,0.34007869640314237,
 0.17850713799557005,0.9454755428155763,0.39015955116922807,
 0.7567753395238356,0.9563649937613139,0.672555311420197,0.3575474768012282,
 0.9031699541441914,0.7392801041008008,0.154761346881376,0.4181985943777222,
 0.5142419259630566,0.22677052135558962,0.3292211505671807,0.9949677394752899,
 0.518205387687681,0.1835056719740653,0.9356332079382945,0.932258272131539,
 0.0920029306530713,0.33237411191606636,0.345354137367779,0.882164460392631,
 0.3250239981234345,0.7251572077581938,0.02285048965588632,
 0.14727370124191586,0.8323248750152245,0.2540333528235723,0.8897208776656472,
 0.28613702157560583,0.544094124627224,0.33109388051567146,
 0.30702569836199356,0.07006998331342185,0.5403991657478956,
 0.8549563119475008,0.7914456414253098,0.987669390509096,0.9560550562376915,
 0.5749124683429464,0.9234871064868626,0.9709018751909485,0.33601015788706934,
 0.7890581085998971,0.8581880501121492,0.42584486865090043,0.6074234502366819,
 0.8668753041719544,0.34638574627279706,0.6176114153076631,
 0.08907830713045006,0.26997121381947164,0.5733547391781864,
 0.7967563797495795,0.8687805952397214,0.9922684069569122,0.3772845272333146,
 0.37700222699697383,0.9094092235971354,0.7377265864363389,0.9192020068348473,
 0.49440056674376476,0.5595565106688303,0.27637158552102126,0.263449080007068,
 0.9703449191634342,0.6077197996878727,0.15708989063337353,0.9319842753030914,
 0.3459353263159415,0.7051192500931384,0.3337120143001927,0.9592822317089343,
 0.9150969968007566,0.5109959698608697,0.8351846950591236,0.25082246521625484,
 0.5100764710632123,0.8469340465499521,0.05748167753000322,0.9187199645185334,
 0.05401160673431893,0.5487663909795337,0.2213383874854059,
 0.12439276998856186,0.8927039177126921,0.21264100641805517,0.498835251523407,
 0.5868400951808215,0.8833417672914866,0.994202868275688,0.9822522950911833,
 0.16434325142180883,0.6886944710064572,0.3290303990260731,0.4242508306127498,
 0.011363379987514466,0.880711959284752,0.7906404556446724,
 0.03190046744695618,0.23037970524512374,0.34118415384312484,
 0.266545497528925,0.8763003700218065,0.5583888985839712,0.1839150676645931,
 0.4002069589033035,0.9550590916600085,0.6813868214704423,0.08168280565932473,
 0.6047763558462517,0.3352416028857266,0.25072881344395304,0.9885936939473162,
 0.15389255385031786,0.3947372064433774,0.7348458426000513,
 0.24212509063023746,0.8467370644082461,0.25262992769129944,0.695775370607858,
 0.8599517585920583,0.888612410315368,0.6314637645802448,0.4453327981613927,
 0.8214668935388452,0.28992514141339565,0.6554910665336605,0.6858353795013669,
 0.8712430464689654,0.5672470791070221,0.6896990989756976,0.2197953465564927,
 0.10178013371185579,0.3760050035708348,0.6309045878811737,0.162471737428717,
 0.6218628465931181,0.30780401883627584,0.21057638975691584,
 0.9505836766154525,0.7314986745207095,0.9548957812419571,0.2835789443407327,
 0.4386749915670428,0.26661710882894685,0.3702696027699658,
 0.42761331194186014,0.8640398235033668,0.47113141183359264,
 0.5027542319530279,0.909579344360812,0.2909519706454986,0.3901660341081703,
 0.8791354505205086,0.8050418929577979,0.8450668043965658,0.19295030775763067,
 0.23667232405168237,0.1837076857155554,0.0010703396213171956,
 0.9916709573844152,0.9036970814377535,0.589295392409156,0.5122678186292668,
 0.7805797761557562,0.5023117524905267,0.16396475637537866,0.4519200812873103,
 0.5540371882067244,0.16337136500365967,0.6419974778571106,
 0.24147759310545458,0.4761314277845916,0.041469840518000844,
 0.8826193307137427,0.8049295507754658,0.02020074991128118,
 0.41553463969115434,0.13228883027644345,0.7098659507206699,
 0.33168306478162624,0.8179871516290226,0.38792721588556134,0.949446025667918,
 0.10698479727236365,0.22082401709532717,0.677753752859433,0.6874928342250445,
 0.033923939871754794,0.45963113996000626,0.08828828662665866,
 0.7334379740016927,0.07682854824014373,0.5361187904836816,0.2764694433682363,
 0.9835227733322196,0.32895182412773594,0.38022172168438384,
 0.9746652347097522,0.5970969489422466,0.6325545556132942,
 0.030085609337316188,0.1614537066517433,0.32736205338033364,
 0.11443121343477625,0.7538841678364238,0.36773268180497476,
 0.06808123728438809,0.11246670505035639,0.5571268223685769,
 0.4000064223443679,0.4886787785250647,0.6628359296630206,0.49812005861779296,
 0.9728867417890011,0.3600025415697117,0.8761649500020863,0.17451910012062632,
 0.9365906537209148,0.942409041091827,0.659507746651916,0.16091223861430815,
 0.47460558272896325,0.19746451579910207,0.27076036508541046,
 0.16089532505161608,0.6978823682061156,0.5454384166794779,0.7752744714381566,
 0.47316479382776677,0.9162018512394298,0.9410521243424518,
 0.14103226848970485,0.840085853259993,0.0478367817262757,0.7067752468790285,
 0.3210397627697573,0.8494757472707811,0.7522562424207921,0.36982819227492303,
 0.1996395003533734,0.1701978864396505,0.448948975377926,0.1291575139891349,
 0.8840184677024282,0.732742676304166,0.30290900676929167,0.4817581615889369,
 0.018940939522443134,0.6881940557952021,0.06007819969773387,
 0.1913424292273399,0.5891090298140716,0.6072260804705317,0.969471565382309,
 0.2144724113111539,0.656783660354074,0.18988658858525786,0.9677414590431549,
 0.09624530537777465,0.18256689166895523,0.0016042776949434234,
 0.24364307001639052,0.29331390039479144,0.13487378872874756,
 0.7781659174224224,0.38794174787737057,0.36057966856081525,0.699827299078882,
 0.6704847621683053,0.6698935802164042,0.9161520570752859,0.964221912965611,
 0.8125518588303123,0.33633972106531584,0.04970635426339165,
 0.38344788846008193,0.7419541276749433,0.7092990346528115,
 0.06896345500165824,0.3945471775141025,0.5738038499922642,0.470458259968042,
 0.9042215573269035,0.8463817880235526,0.8609940916534541,0.7048284923804775,
 0.5330108724485582,0.5920305611236675,0.9450969716107086,0.8949205080512749,
 0.29506085936930815,0.7911305554927531,0.5407059814890465,0.5224553033627389,
 0.8455304412958808,0.5279175790228293,0.2745241502369571,0.455366466906035,
 0.40955133283047274,0.9550798318556487,0.9508111570178691,
 0.16559723075535138,0.8737560662187929,0.1840512731057835,
 0.047559260226373024,0.8278014398195113,0.5805260509082916,
 0.18850953613731192,0.1406443347557197,0.9644908260624204,
 0.18426759757774658,0.5529603913590517,0.5370615197592622,0.7796357833734915,
 0.24150750688217182,0.22023097653051882,0.25973822924260925,
 0.7104987561918165,0.6914913624452865,0.38191251796874615,0.6822093794717513,
 0.3963768328014299,0.9617626751485273,0.7940924882243023,0.816938990867812,
 0.5471738582836515,0.8469229459612397,0.8777281612971143,
 0.013894105880711827,0.31050599796576084,0.9359117072435439,
 0.8329368513206791,0.768598593092495,0.5468156989270033,0.610807191175128,
 0.5903637887321238,0.4227707989649999,0.10729327261598631,
 0.09474358990021214,0.9353327383900156,0.594309374226399,0.6130053123947913,
 0.5713504185747891,0.5425822871988324,0.5200122476846258,0.42401548175655135,
 0.6730981396252745,0.7849831126022238,0.6411868661761091,0.6582927798756437,
 0.8407796063633677,0.1783596197809122,0.7902930189385832,0.02715478740105437,
 0.12905345191402962,0.6804919453629835,0.505566153881012,0.1566825578609058,
 0.3317469371786659,0.9574932356336617,0.2916515088263827,0.5905043471511013,
 0.7450234710115556,0.801310157057622,0.8825842821379013,0.9671943270139509,
 0.6489482398499743,0.886015941013481,0.8620850950000706,0.08258910452653889,
 0.41962802392467957,0.6826060864345525,0.5535489705758866,0.5430467033791843,
 0.8767553057625849,0.9508043321333113,0.6090647778511403,0.8823312584326128,
 0.8142957425867919,0.4216272882663463,0.4348021288095584,0.738442103459932,
 0.5508573406403026,0.752594653328265,0.939192609467124,0.6717570265252791,
 0.06365297910101497,0.02725785497922506,0.13346217538775962,
 0.6134615450378386,0.1304045105081384,0.279962138458581,0.988599400777989,
 0.44809471816929203,0.08969401145799549,0.5428236190842729,
 0.3055852913601651,0.595634542605052,0.11936532068029926,0.7313087858670783,
 0.147128850253619,0.706796516868498,0.7164555027077996,0.37229542144079386,
 0.1801787067299322,0.11019016515194213,0.6955617376825273,0.9331920188842495,
 0.9806249897939358,0.12751080810144078,0.9314388303085949,0.986921967216941,
 0.8337113791193456,0.8401013032846075,0.25133270031684773,0.4432863500320424,
 0.21926777531485053,0.11537723709858938,0.18794238878587566,
 0.04408616379190866,0.22917803718211271,0.2636486830429823,
 0.6002473249323682,0.3842971445830937,0.789160726191063,0.26303481597828715,
 0.7919196864168906,0.8098781774897164,0.6008779494952479,0.24969051765421213,
 0.04346528468141331,0.1626988581836153,0.8966008303023448,0.8354794415699798,
 0.5902721531808787,0.898975712972763,0.7994874340978334,0.019177787719726447,
 0.1889726153077791,0.6303338900514519,0.5425967163556299,0.46735454822633726,
 0.505813108924853,0.16806689049920887,0.18919747262802167,0.008365589044526,
 0.9786383291690227,0.5271749170114584,0.7973234098905827,0.20419679445638983,
 0.6856363810362696,0.10845687059458331,0.09707442166153546,
 0.4871121455840618,0.003012746653687559,0.9080768318005584,
 0.8385279926226583,0.5274086703466984,0.6950092894509736,0.48016219207183697,
 0.07029848137233685,0.5186348293981905,0.19401552195368443,
 0.1064136629394602,0.306281455137386,0.363826809604783,0.6866285143108752,
 0.3499216332135364,0.5067547099157721,0.41630909222852996,0.6390079242373627,
 0.4935622245400785,0.605137332295094,0.30308174684460076,0.34237056649027287,
 0.7030008647960656,0.35688457866457335,0.7861579581062805,0.7421552651067488,
 0.7334272594377642,0.9556321815106499,0.7833747211611485,0.7399127809968145,
 0.1852783143531357,0.9442712323291258,0.8117698059443256,0.7503789704389002,
 0.3806200231413359,0.20840717413042464,0.1397576312822476,
 0.27201234223534887,0.8525203814765226,0.9817953575682853,0.5088224147049923,
 0.4252154584901755,0.3505991951932862,0.953990181884161,0.586553378867458,
 0.6493091549695817,0.7466340076923302,0.7188658518847564,0.5266820106986353,
 0.25334216795266506,0.3778200671946804,0.2516751099153356,0.6361342478579578,
 0.5105717235564696,0.9856039135791309,0.35178005441562255,0.5762333732418774,
 0.4741511197924819,0.9810735565181196,0.7320892155976004,0.3245260290710429,
 0.7496381408517206,0.18720506489766653,0.114850384280205,0.5775495315332994,
 0.1311156984200239,0.6717382882161688,0.30246417847220053,
 0.27149854938320983,0.9149366583217629,0.2418431020888131,0.930728809107181,
 0.38760016744552395,0.49349463215587475,0.715839689563739,0.6190140177740898,
 0.928368643519718,0.07396450961628975,0.3221613055519904,0.1673252388377473,
 0.9838755551205378,0.6435925824360454,0.06078832426361669,0.4404136781040251,
 0.9412895216903279,0.1548418530949297,0.08289965553464707,0.8236238979611555,
 0.8382822447594062,0.2909952115349891,0.7948873724950849,0.7529667058840532,
 0.8516931814962765,0.17226874919923107,0.2151072722134738,
 0.32592393204575276,0.7329717283980496,0.7893930101491725,0.7198231554111528,
 0.9174428996767581,0.13763371028864269,0.9689205654558775,0.9527966045525342,
 0.3525356289659378,0.40829554640094257,0.2624647945863756,0.3109288598478338,
 0.29825930603354767,0.32493126816807794,0.28835316646559017,
 0.19458861898506719,0.8916725512056113,0.9550241925375864,0.8751096112655268,
 0.1828966754908582,0.024353190650711842,0.10790677499070112,
 0.5864627138670507,0.19736682326665755,0.1044563689551945,0.9564102554025609,
 0.9560733829126733,0.3014362517650051,0.9837293891442439,0.3762972293688567,
 0.5179812214557666,0.5235846162459366,0.2737949174212003,0.8314341180116354,
 0.11856725294078607,0.16542328395013917,0.3959809548454296,
 0.25030553433341685,0.5263732999392225,0.6708117884324813,0.54823181618721,
 0.828780097965844,0.7666204795647883,0.2906476967317828,0.8460888423289666,
 0.7359991018044771,0.09693701291171719,0.5097242057935281,
 0.009004310004927074,0.05285400084817826,0.5558408335645877,
 0.7385080163684774,0.2352434200973781,0.9853391573022803,0.2142814194662318,
 0.05680997165996016,0.9958223830534546,0.8555500949546619,0.6886102752295776,
 0.34919117222661844,0.12521014996042767,0.7033456105061499,
 0.12657449282330913,0.7883136417206503,0.32472861183245183,
 0.016744267783917022,0.374556631988443,0.3779982348987947,0.9777745377125081,
 0.2687293478362285,0.2614870375420929,0.8229674885052607,0.29036524752324844,
 0.6693192805783351,0.29667203849630797,0.649725275830913,0.4582515642970765,
 0.8716456563751396,0.9492853974426757,0.02685111362941839,
 0.28253550174708686,0.09636709742468041,0.8980837454728762,
 0.5676115156124231,0.6833423323423232,0.6329430406172171,0.3939817569646267,
 0.09252232929012694,0.6869952191479574,0.5513567729325495,
 0.10176802454520506,0.5301308717017111,0.4299628257931538,
 0.15850691702501274,0.009032021701451343,0.9391567475869579,
 0.3233703147696728,0.29527816297136733,0.8811929650211225,0.760033926606222,
 0.6953708649928175,0.08510802318590083,0.006399481389249506,
 0.3385238160524512,0.1511985946379807,0.6011834163554706,0.5540507538310546,
 0.6603065463708901,0.8660264062584894,0.7269106114221157,0.9949449342692177,
 0.3326155986502125,0.7299779054668217,0.5883336677730907,0.1358574030350561,
 0.4247367842712064,0.49695637467052134,0.36899098212183434,
 0.5513305789358988,0.2176906347841504,0.14888066688513102,
 0.07716388320777268,0.5770331236531872,0.11731635718454936,
 0.7058270663170603,0.998502620327738,0.33653845314093067,0.3074282364874168,
 0.9644752350424866,0.4185483274894396,0.24930751363654025,0.9818147480755705,
 0.406888134039307,0.5170517765751707,0.7838082346361261,0.014296129648033062,
 0.6801082562035567,0.7271731270987983,0.9188166744100388,0.12421111993705969,
 0.8664687968013632,0.9539282165647048,0.25231483723587467,0.875091309077239,
 0.9806338132983364,0.6538841073445291,0.08675599653178656,0.8888228267430252,
 0.567481404821718,0.046814698445210245,0.012331608674564398,
 0.9711450870109506,0.44647372094449544,0.3657219834401999,0.3617235945827135,
 0.19261762393499438,0.1302834470373454,0.3520152979371062,0.8301174697097796,
 0.5242027146871888,0.7966153622658293,0.5412207970564162,0.8689012118053021,
 0.7249124723183193,0.9523498259417712,0.7689931360477562,0.16517704138086653,
 0.9651399256304806,0.5894330618218258,0.15161750456989753,0.6311834112199862,
 0.9907888240004413,0.17655257291944848,0.3245064484424218,
 0.14551288933411277,0.17331707047995226,0.009632794050080551,
 0.8035600870913658,0.4025306766142045,0.8082725509199131,0.6271027812643527,
 0.8887578623228105,0.43172423039574603,0.698969149755702,0.7359503816196754,
 0.5072599093134651,0.015512900158945797,0.21894836643087212,
 0.8004892108075008,0.9510148693684599,0.6567258162149168,0.9817368666600591,
 0.776894378760916,0.6165456566082211,0.7330663155165904,0.6628148468317071,
 0.9700367219214584,0.7238098804763252,0.8447026187981915,0.6076567136388199,
 0.08843879587862313,0.08845576597790439,0.9303125466772054,
 0.6525870448212948,0.3026254377172636,0.5832203029646583,0.12882486938959092,
 0.15940589715908116,0.538813384036948,0.5178112088484478,0.7675217048312653,
 0.5836183547339765,0.30394568644364417,0.04079673242794157,0.976612577142523,
 0.156569191827458,0.17023175473330154,0.7198798770767171,
 0.023167345542909334,0.0453521559570802,0.0559098144622272,
 0.39581343123271306,0.9505674612119386,0.5778573559623925,0.7393183703832181,
 0.7332231450822012,0.587087910588737,0.859984538855959,0.40982590750178594,
 0.7552487418633645,0.6024461752798759,0.9263934112051698,
 0.047518117434961704,0.9233276789220626,0.22183858527810973,
 0.7729480664412736,0.22721827606411005,0.8736420739936362,0.7380765377332804,
 0.844729267244258,0.4927319524720286,0.13375764623101793,0.6056271670074556,
 0.9632140556855693,0.46603504066658963,0.8940871397124218,0.7717349573654827,
 0.895245731082971,0.4260044288304752,0.7721761549937334,0.6161690845583734,
 0.0026429771886882225,0.3629503432644665,0.7749824801961016,
 0.6496412860043718,0.7616763016045695,0.4883433955562272,0.7876564603631262,
 0.38856102540995185,0.5099755149301113,0.27289835980742727,
 0.23151654009909461,0.5634149598208921,0.5302317023649716,0.3139302034512338,
 0.49458240804861475,0.5010964886073536,0.12150882314571698,
 0.9684015787670145,0.5701978312755999,0.601601442893403,0.732452739440845,
 0.9231974354839283,0.8108828225501736,0.343393365657632,0.5346296656305195,
 0.8145670823356963,0.26068468232295816,0.8996689802892099,0.4761198662977274,
 0.2827227245697217,0.6488297151653816,0.5758952149955014,0.9477142696222007,
 0.5320285705681399,0.21474721408617503,0.3568771526458646,0.2925598177789833,
 0.6171942216534418,0.42324197552258536,0.6495331641063998,0.5922649640159583,
 0.8942453546016482,0.9981130454921563,0.06149464929673343,
 0.29014302397824143,0.4295497344703757,0.8425792060258249,0.6925475073797642,
 0.2242218883811553,0.6295373372529243,0.23377841434309699,0.6064691923096236,
 0.4045574818392932,0.052338196031371464,0.004813009624684006,
 0.3166834299745348,0.645982957405917,0.8030745776734659,0.19229317649189392,
 0.9832864015147751,0.17531467110201038,0.17670259728551296,
 0.47704272827123084,0.029723624373013013,0.24594535909250148,
 0.25774294674962384,0.8902234395922436,0.2647843479490013,0.4460802619767834,
 0.0852862731488997,0.9001200734142618,0.05098691694877289,0.7309597053252936,
 0.7621288452299199,0.5338199824623606,0.8011086389650338,0.18604996472496838,
 0.7959382291057687,0.2883261994192825,0.7959644433372158,0.2185449371386754,
 0.49680632870004027,0.13229168024695614,0.9373878727765907,
 0.4586416481336406,0.8165795476062881,0.6762666093958943,0.2235397838380594,
 0.8022530438346271,0.4394018705650876,0.8414025348844844,0.9904087610167527,
 0.0752570551692745,0.1783385440585571,0.9401382769798834,0.9677454737322512,
 0.21037694554479736,0.36945506645339754,0.030480065781963717,
 0.1241870206958976,0.19814688831160154,0.5660257970672206,0.5435537645857437,
 0.48543608540194316,0.9887094418393433,0.7822306720690239,
 0.04569574260466358,0.7185001414892966,0.5489995180391146,0.5536815846069,
 0.7688501769810198,0.9509538095751987,0.07753976663591944,0.3073513277791976,
 0.7076003924852666,0.8269467935649629,0.2150829597470999,0.3944208513445854,
 0.8701308208974122,0.12803417137656836,0.7195703369163567,0.2534749508062093,
 0.8744750263287198,0.5107576532677394,0.7487416956045161,0.9106703171342568,
 0.3026888883939791,0.29722614127195357,0.4981692953862332,0.7351464385782789,
 0.8109390242625696,0.593150796451398,0.6993414598854188,0.15363743154778298,
 0.9765768483155751,0.8070194789340865,0.10984103065456752,0.71643483170206,
 0.1562219631460291,0.7603055400633303,0.713024750710336,0.8562605614348925,
 0.6972471236575617,0.11184219929118777,0.024759157441671675,
 0.9727951664801129,0.9699194045925255,0.21571838432224866,0.8456464167956445,
 0.08123310803204375,0.12392046588178629,0.49456003857591735,
 0.4989529469813323,0.18609734226686592,0.40236655918497555,
 0.9167813943042005,0.10894432676634491,0.7625396956057198,
 0.42563013987876963,0.25071007678593193,0.3354872445004716,
 0.46223326682629107,0.8126424007011898,0.26982900042537816,
 0.9843382533739424,0.0039329713661052645,0.9717506242108825,
 0.5216621792803251,0.23204725585240726,0.45582231988677,0.2238820875227555,
 0.3266861122247069,0.5212801275183971,0.4930926105468225,0.7997566511252857,
 0.8183894574713125,0.8729847317098796,0.7480202086508799,0.869129538938877,
 0.039827710819403395,0.7443246832061796,0.6487353492695076,
 0.9304731668393755,0.5640280440428729,0.8678154686804853,0.38186191029121397,
 0.8235360445300455,0.7061445345519752,0.8359576682631397,0.5143991978019085,
 0.4926639150853218,0.7612162915832885,0.867275965680996,0.38913754396460887,
 0.36657029225423643,0.85327287902684,0.3781003550887466,0.5052515987409728,
 0.7175356074552002,0.7918681433580859,0.8556604170553408,0.7357437888998551,
 0.13807423943113628,0.12834900670231297,0.04822928129809423,
 0.28573118253162466,0.02080183483216369,0.9585509102381348,
 0.8905713686192469,0.1271875929166706,0.9612768912270455,0.7900082584821646,
 0.8904312635406376,0.2011810158736418,0.8145771971689378,0.8426175898568942,
 0.07543811355003904,0.44804676347613404,0.28443704032227624,
 0.44286146970005436,0.05299070395640859,0.1950163054364199,
 0.7210687862915485,0.17476751971966165,0.07448743026591187,
 0.4065513840176219,0.8458163684248527,0.8882259581932421,0.6533670061146873,
 0.9204947708670775,0.25151220586897305,0.3498437794062266,0.7151309072602781,
 0.10917878002308556,0.048450515856831355,0.9010203385044768,
 0.5088392486485831,0.9542995666312143,0.4578845608896034,0.7060808850058855,
 0.3182621497280387,0.5859742997701025,0.46429908308771806,
 0.49152165468854814,0.9874648939248347,0.6515681124120689,0.9258047363914235,
 0.26251279462518773,0.18601997515604418,0.3197606075651612,
 0.031044831340739876,0.046718582769067485,0.5816204252997472,
 0.8026079386677547,0.030150359022050988,0.16536511771250084,
 0.4753115732720714,0.16366415574731863,0.9648672973755824,0.8055886739017482,
 0.07471201948237416,0.8920761657351028,0.39953285004288985,
 0.5316635833577481,0.2274107324052752,0.42917959054228283,0.9451021585086388,
 0.27549842704070127,0.03123015457694578,0.7460663508330352,0.963883106620554,
 0.27236631826375124,0.0780284091016814,0.971416184941644,0.29453388190056806,
 0.6453941643120736,0.984564076022262,0.4247143734560953,0.22730031489341473,
 0.7848409306240589,0.20486653522464882,0.2290556985780946,
 0.41875428124426617,0.22108686713003167,0.5059458488903437,
 0.4190805209865508,0.6040274375450712,0.04916036460104878,0.2805871775404686,
 0.14876149508414482,0.51218816241351,0.12470463003821475,0.41778574137422175,
 0.2712727439482545,0.32980585784666816,0.23388812223411948,
 0.10420099127738647,0.46820670018290134,0.4585012759356616,
 0.08982188034629734,0.3146604993558838,0.6675690097704383,0.9560323160526558,
 0.7488161750270429,0.9718581443372676,0.6836534048013967,0.13126598654594857,
 0.8183238531123369,0.13476551394872882,0.16818700159018996,
 0.40944653795433117,0.35956264128259474,0.672804439202714,0.1791167277134046,
 0.1178990307158928,0.9761394507030429,0.7309722114895062,0.11709950802034541,
 0.1764616405491951,0.9955865306587263,0.7633597492576301,0.14534473185428354,
 0.9209559249593086,0.016438522481528883,0.13958471514432702,
 0.9992362448605813,0.6829458409088724,0.8483904580134454,0.3918839291952211,
 0.4366102962234364,0.716195331016211,0.6068827982647219,0.9779970789039543,
 0.7626642444743974,0.7370267892690188,0.5698389446882263,0.8005420705070919,
 0.7007979156223194,0.3517591952922585,0.910410809938237,0.6374277697562774,
 0.6223435766541296,0.5444440116291595,0.9429272436596695,0.7779385058959357,
 0.8012792869087157,0.25030116762297605,0.12185710514412051,
 0.07517826956572027,0.5729296009980219,0.09323789230809298,
 0.8569750770600174,0.04689850490093861,0.6008828744581698,0.562331092508273,
 0.9417181607881061,0.9516830185728689,0.1075084503938456,0.30682109843912275,
 0.17457422983959126,0.044929418363324314,0.5138887369208641,
 0.6950716500279357,0.7741908947100496,0.765623737216064,0.9209442000158402,
 0.20561198681411963,0.16213771844947678,0.28607827049310375,
 0.2795299426473422,0.24375827212312495,0.19586425578216715,
 0.9714069345458495,0.9019058921178558,0.807342444978405,0.36746991049663436,
 0.8430996892434505,0.23493679970899695,0.0665941967954079,0.4307897880086424,
 0.5012088048678631,0.4799552401254097,0.918270235518712,0.47582439906690455,
 0.7363566933350773,0.1332336624376731,0.7391126509253523,0.07353037100168691,
 0.5456228193611412,0.38704628912756533,0.28549884042789775,
 0.39181805754813503,7.23713799998249e-5,0.5606952013678087,
 0.6359153010039135,0.4118344595552087,0.48051741496842804,0.6992650448842948,
 0.2021876811536315,0.6220417108168124,0.8318738814438124,0.41781953551142303,
 0.5218005168923091,0.560545882977758,0.9878866901644553,0.9216050883918664,
 0.6570102220483387,0.3263716830336123,0.734870031122474,0.9063186678080033,
 0.406157281011404,0.6681350078555264,0.666674096368212,0.19567872793791918,
 0.5822324224191217,0.42479049805404645,0.01843263843730092,
 0.3640134179503062,0.08647912639490163,0.37158589993597047,
 0.8912532250790264,0.1869022898088546,0.7133981713971277,0.5248610369758018,
 0.7080877632804367,0.4095384588987826,0.6271634435173454,0.11938561643278289,
 0.6427966360433549,0.43881790575609947,0.7077453779810223,
 0.32791847258797213,0.5187856243276646,0.3355445211919842,0.5976676645039732,
 0.8805405260586573,0.8584098188051201,0.9031454263907448,0.3853492832126696,
 0.22789888193079577,0.21773707135025222,0.7841062431540465,
 0.8033173138855765,0.6221484178919107,0.7866738557251888,0.850251158404105,
 0.7788419025495973,0.05910643132113069,0.7727857609611416,0.3634891983621975,
 0.41406004101238036,0.3871813546466518,0.846983884909765,0.10960530231564591,
 0.29990688091405626,0.5459567937404969,0.21817390782368173,
 0.04441421224920972,0.6319613821389825,0.9663856323599141,
 0.18970112017115792,0.9545364073408564,0.40449884892277044,
 0.3043159163698673,0.5148283498003574,0.8089627729691278,0.3381404755392071,
 0.2074071081007004,0.6620137845795341,0.8930157393450743,0.5685219632710002,
 0.07286717076403748,0.6787474918779031,0.33311065937756945,
 0.5696233938420923,0.8039032678315584,0.545094784000746,0.7689355743307551,
 0.13582674225769853,0.7009717100075996,0.03006480816982593,
 0.1380923234965613,0.4230749820416474,0.15458364459011997,0.4662690179605057,
 0.5502629467575766,0.062298108768681026,0.3947370148680559,
 0.5143868456593751,0.7898806502083074,0.872625193533404,0.7001386341503615,
 0.26237650461963613,0.9024199061540168,0.6265114753523744,0.8077716619098897,
 0.5376665745288719,0.7273249460747122,0.231620407522675,0.620267226787246,
 0.9020929161164188,0.8084553256664848,0.5973906889963049,0.4508727357849237,
 0.8769789786815274,0.8275245182132573,0.6659369484741273,0.9769875838986577,
 0.6507630244279257,0.5051923700304999,0.16731523262823422,0.4626306595034986,
 0.14414796141039532,0.33824245469272796,0.9247419918409344,
 0.3873772183115771,0.26314526643619085,0.7169826912062154,0.858042316709217,
 0.9438518170508747,0.3064706668241619,0.8213765652338385,0.260875542733243,
 0.2387179848023323,0.8408166049359728,0.4769491837271921,0.1016830964387413,
 0.514994824187174,0.7934967959353437,0.8437759873008567,0.1096512450366921,
 0.6724963808022368,0.328371916800992,0.3929878112830737,0.8883296032319311,
 0.41925542452498665,0.7059994699646465,0.0012510850182451128,
 0.9889441199717165,0.34378753366921966,0.05476643839507611,
 0.02391462751631568,0.5093503155198646,0.440289998613274,0.43581301471340583,
 0.5689085100145804,0.1869473864432981,0.5378497167422969,0.24926319534994879,
 0.5320781353997491,0.6021263770922443,0.3795449466581865,0.9282630524983156,
 0.3426305953189678,0.035566817033341414,0.12358885715224488,
 0.29433111639981135,0.2789411955905481,0.8848209342036993,
 0.21439036989556703,0.7327187394947112,0.3471962049189472,0.4053865997862641,
 0.6301609469552779,0.15212290403943418,0.9599777635417646,0.2925131734091836,
 0.7375784161832288,0.8373437243991716,0.38393318535023613,
 0.05317380420997131,0.0019323760638412235,0.6034305618590743,
 0.9670330409698171,0.4655069385764784,0.4471287147363361,0.8165678998167384,
 0.735603075563849,0.94088096751849,0.58494014151763,0.8746922628124747,
 0.15319869138816145,0.6402585538137998,0.9721731595550611,0.3840598198605295,
 0.1069082184192548,0.37872356592227807,0.5681517791134888,
 0.021663652400906508,0.2086302693671238,0.3477267669316987,
 0.5644149697070716,0.2915976511197673,0.22512816980189176,
 0.14342618464057066,0.31819474658083,0.2218161737533184,0.7345500986247822,
 0.13775051907650593,0.13727654519249533,0.5769975107904917,
 0.06490400187715373,0.1280826635743686,0.1593765459739933,0.9363057365805231,
 0.071672696734421,0.5827657873855174,0.7177717429335593,0.7250624681951758,
 0.2511751024734513,0.6745746097423295,0.7363986000937487,0.5177797981006352,
 0.7578827549491645,0.29383382014337067,0.08886073052983634,0.355614190471118,
 0.6153699166429463,0.44556850257938074,0.37395885174494414,
 0.16349106035423533,0.38435181096879967,0.5270641635148112,
 0.07697846936994324,0.1867354962745963,0.6888423191169515,
 0.11319744864022874,0.18155101829645837,0.5015207178422789,
 0.28892414035088654,0.9694920712412497,0.09821001493586212,
 0.4056740916393079,0.42730253413579167,0.5018461751229306,0.913726940690057,
 0.7294312525378828,0.9620667383558504,0.2589718868209153,0.13875239742434087,
 0.6786289516945294,0.8158887530245132,0.9908020327991869,0.16206921968725085,
 0.3431010130776695,0.6915721985508217,0.8852068624955045,0.40104771869080835,
 0.8828373572580299,0.5022751611385897,0.6564128429910542,0.16942606126622173,
 0.9658468316996562,0.34834031696293266,0.6762921799331663,
 0.08440448103612941,0.863088284473813,0.14571046721199798,0.4602370626287391,
 0.05645373593418601,0.6458603556560029,0.6786021515980096,0.2357435745312424,
 0.15124328905328288,0.3853137947816405,0.0013853640909807785,
 0.4512644877016121,0.11516058834559673,0.8136677679779525,
 0.07677092229012106,0.3743382432322735,0.5163511777800339,
 0.03654830712939938,0.5741904855759014,0.5180196780876569,0.0903555584271385,
 0.9691605066177973,0.6851622869759747,0.18424363824659218,0.508110910682399,
 0.2976551156857835,0.16228589296251572,0.6909908294429816,0.2919627067932594,
 0.2950246248337669,0.7814879613614804,0.24383682107073212,0.7830358380281185,
 0.03656282552635348,0.4840386483940693,0.5762438195973866,0.9762935421249812,
 0.27916797780192726,0.8881999604608662,0.18125344115708553,0.898627277638086,
 0.808593444607814,0.44227615440302126,0.8213377099597662,0.24725936050712727,
 0.3376171539765399,0.416779219449562,0.663499875882568,0.15536851284023911,
 0.5412836034343478,0.01829450036361957,0.19170703808258238,
 0.7968997614264871,0.5600267205758045,0.528592631842157,0.10746154983303047,
 0.12032569674832483,0.294040368276104,0.5085893755060675,0.49438300042673433,
 0.5014018264642114,0.246706264960164,0.08054089321826874,0.6347995981295553,
 0.17189551835118444,0.5426738726687803,0.7644504772885294,0.4157949502028221,
 0.9609595688094239,0.23259158142677805,0.6879096307690103,0.0326036246954059,
 0.7989580687317559,0.7091172923184991,0.11497441860326396,0.885570561476758,
 0.8485797390376453,0.6691077487073791,0.015151427579420562,
 0.12242850372514558,0.2353837490815962,0.10980931082260859,
 0.2899269228471014,0.40428915564444934,0.009548359862557643,
 0.39206231101390787,0.46794338781990963,0.18753269088113944,
 0.6690990359310777,0.33478761477712893,0.3619382460536329,0.7272381157833998,
 0.7362846749883667,0.37435079826239515,0.8070179901216641,
 0.05978560230712593,0.9556368108236459,0.013880591589097824,
 0.049112287522619424,0.30247736229385846,0.4149171225203594,
 0.3389693201516446,0.72225600717215,0.617437649738835,0.6961568057315369,
 0.057723624871649615,0.5321855127457131,0.19838534039789535,
 0.26239053727662087,0.5208749845399863,0.8604520053329561,0.3572546216033051,
 0.35528315623337703,0.014452083285605655,0.4707824835161434,
 0.9223479530970794,0.36706128871668064,0.858293672646347,0.3276300419874002,
 0.9210879172834928,0.749688266230368,0.4456200273996418,0.831211905711951,
 0.8563446857223773,0.44660999428927906,0.6754237349963038,
 0.08815613772169173,0.9544738396749891,0.3293198297795502,0.7201214334482953,
 0.008642861170940153,0.7807743339499197,0.968668327302483,
 0.33142858284777077,0.1556664188392196,0.2084448403918755,0.184828819337443,
 0.6626851049149541,0.059532496018031456,0.5128597651895738,
 0.2690155797897047,0.03839682758618401,0.8179123356722694,
 0.06252800060300068,0.4318876350865335,0.13272442215886804,
 0.5203221748944865,0.21983718289646026,0.7041435212068071,0.867750711345338,
 0.6508991209451123,0.7385754085757793,0.5854607321906979,0.6017340622966433,
 0.7849099963633813,0.1785367095359095,0.20808259409301022,0.8467294755225804,
 0.7972376019874512,0.8661209119488478,0.9743674894984733,0.13636719296131905,
 0.44327293597898776,0.504143360030217,0.21525384583542995,0.7235496699556704,
 0.06882901228364136,0.5401904647948468,0.6430406275542723,
 0.26684915422113087,0.8077407898355546,0.24784877407766237,
 0.32017221777112537,0.689371846700172,0.5063844905389412,0.8239293826057119,
 0.4363611149253558,0.2019767864124703,0.7141099308986539,0.11273562285685601,
 0.13223415982953435,0.8221248565687157,0.7574709361661097,0.5475495759449199,
 0.06705749056090515,0.6940023020179864,0.9659018127538221,
 0.30212571520565246,0.6198218365509585,0.03311149401475544,
 0.3792839978501005,0.038825856026736605,0.5743817341025999,
 0.39808290341862473,0.6993786487123469,0.6934756189875204,
 0.38240790814255343,0.29686594312107917,0.7124976015779346,
 0.6630335598955218,0.7813528574384675,0.9525408634444983,0.3162035403803962,
 0.3283354951211098,0.20070123840483378,0.009512547942811755,
 0.43346074548784963,0.3838662783764699,0.63665585931946,0.6621990721806624,
 0.7011241403444373,0.34574913019625597,0.6312527161183299,0.6273412051029649,
 0.07846692109678821,0.18902603466165013,0.9263692733253399,
 0.13021874371052555,0.18595944761254957,0.04529541143033733,
 0.759818032751065,0.41076408101999295,0.6847197369922914,0.0544312540116737,
 0.2708424995272656,0.4670051188127108,0.9580713536420653,0.3368526477239496,
 0.6981139253164642,0.030187472579441765,0.9779920464188632,
 0.1041037481777527,0.10811165715040516,0.7350719907885287,0.7626006489300198,
 0.5618162807381379,0.48206924488563674,0.4192697335409774,
 0.05997597440982683,0.2853532439021311,0.8179640580514168,0.8755332059073948,
 0.6734612346481967,0.24859664941082493,0.37682171467461056,
 0.7413154383361241,0.1733482895862145,0.5884078302669622,0.8709415081664441,
 0.2995326680411202,0.02686729331716642,0.9380028291763524,0.8380232639635041,
 0.39929555695494123,0.5870355355147555,0.5465394719769169,0.5981246230139298,
 0.4926825657725089,0.29985562197227433,0.2216662503656991,0.4588551715688467,
 0.5273706614894906,0.06175477960658193,0.3016789114540681,
 0.25054084155170653,0.2119664193607218,0.9455002960282421,0.6994119698314482,
 0.5921362637441047,0.09669835451987974,0.9573700399901651,0.7368759779621544,
 0.45692987160840104,0.003711592969558364,0.7132060523819636,
 0.10812531396061376,0.9394143710451929,0.8048972286891622,0.5210175586386557,
 0.8638601138419267,0.5472423727947153,0.9434382911825985,0.3860550267665177,
 0.04089723319965766,0.09151210557408607,0.07109043119236069,
 0.44218150928801814,0.45873602690294013,0.05945529811164907,
 0.285345346257318,0.8028576015756954,0.9518833889349719,0.17062142690035287,
 0.7203892763002191,0.43749587693016456,0.7925822492883267,0.9494844879539811,
 0.33112955216857,0.3884749284123823,0.156571031032957,0.9318176925060926,
 0.6481722157420541,0.8068453686095838,0.8240227430196898,0.7942424189346834,
 0.6635728039442379,0.8007796261617152,0.22528321824218767,0.6987597646361,
 0.26840199223233296,0.776260709186899,0.6778603373477419,
 0.029674353482461613,0.7818147302254043,0.4723310120634421,
 0.09704185868569903,0.4321789990770849,0.8442506071156961,
 0.10594219007122674,0.6138339433801441,0.3251147333828245,
 0.49211475419394013,0.11637766439744823,0.5624881095994325,
 0.7170139791301787,0.6290500257269172,0.3108206644960131,0.956418634816448,
 0.6864962266957437,0.46159374378556806,0.03438964026928115,
 0.8495781708995617,0.9041201631263662,0.7791711386435044,0.24858622063481517,
 0.28550329845254296,0.8971752646484028,0.4462770478237841,0.4916588616422064,
 0.6913895413607578,0.9007025630734167,0.42222349248092983,0.6305658501771186,
 0.35335501427050064,0.8213430606223727,0.6187903560287279,
 0.14429012531655194,0.016403276472344075,0.6374744329610978,
 0.8366770108120616,0.945545367116678,0.7264280141081975,0.7194826610128059,
 0.8279033442373435,0.2647856598495565,0.8551246421197951,0.9251077886658378,
 0.12776033656378694,0.3492741738037869,0.08443288145308059,
 0.5101243758599614,0.5895545169647286,0.42177207635595615,0.8216477135203891,
 0.29915112179618686,0.266040172784057,0.8064110435399006,0.9740720335876174,
 0.9728227661200797,0.7055168766639229,0.8239076773114108,0.481696011043788,
 0.8603699941823612,0.26204185353802534,0.40262532335551227,
 0.5533576091338057,0.8831798482958363,0.15416527308086225,0.1093854690700159,
 0.34493158305587257,0.6149221816234345,0.005549175138264806,
 0.2590719115655612,0.06766646825795453,0.8628161560696361,
 0.06827887938345978,0.9161447120357236,0.7457508096211328,0.9608843625542953,
 0.3578001504561944,0.9391104895030951,0.7836803527443409,0.74025612727609,
 0.5500893256781001,0.18777356728910832,0.3112430930466964,0.5169312520761155,
 0.8219082127528801,0.332040615697099,0.0679587523486026,0.08306535205647025,
 0.5526944290422837,0.3934062947121485,0.46888607263120496,0.4144104875200165,
 0.055913755146702826,0.44806406322708214,0.8977768344363567,
 0.2180129242254838,0.18469867902940806,0.9001180691710313,0.8134764985283702,
 0.45518754134254136,0.2845066195160991,0.42272628314569705,
 0.12857300181009812,0.29895640528205325,0.9259301355264866,
 0.03089553990937466,0.039205519872118066,0.17468066792713066,
 0.927535164118467,0.4298835703106477,0.13959253372075398,0.25863214490965425,
 0.43641991206283826,0.933337953896937,0.9982879797123748,0.10608447894621931,
 0.5681196232852,0.6374814793388913,0.8839908030780685,0.9282146458244709,
 0.20069654344934895,0.42820194378572407,0.3900494099833729,0.662899447298856,
 0.3474401171794993,0.36916405138321096,0.880854891928474,0.6483889290689593,
 0.2602130570217819,0.8265308331203207,0.05530239165204576,0.8983459613882882,
 0.7417013804522484,0.4261220149531213,0.2954818141775505,0.05589888534359211,
 0.7103310997381154,0.5795979004307722,0.35439429277066303,0.449129651375473,
 0.13613135632890783,0.48111327518967073,0.6509159149155364,
 0.3603471844719123,0.06339417171150152,0.8273467183848306,
 0.06040496195892586,0.6810132085236106,0.7061076037323294,0.8031890763573616,
 0.8670091253128936,0.5622546689151376,0.6551618134617286,0.6141804447976331,
 0.7052209670949,0.5511278870141814,0.37916965652421386,0.3562638956495803,
 0.6097600422706477,0.10081325880016223,0.4169921139046344,0.6256204826801559,
 0.6551207290118001,0.10141695932531902,0.606713594342469,0.4496968572761808,
 0.8631400636604307,0.0512891701197673,0.21616000989360307,0.6464387098011849,
 0.09862590555368578,0.48855263107913327,0.5613276432208419,
 0.5759899046157788,0.4314431113853566,0.6855669940976699,0.1860961071016295,
 0.16799467664035017,0.4897366015077158,0.6519873635504917,0.8040279894553408,
 0.01996365246696885,0.3544128437038021,0.9986185532238957,0.5859777928526269,
 0.6551887865544227,0.40010963905265307,0.29324756811989794,
 0.23563507889542723,0.9686099352355662,0.5517415242357235,0.7850527513456349,
 0.8433663319225655,0.9725245779243435,0.649080368277804,0.048547988723415814,
 0.614849939559451,0.157346059834096,0.7380560113465979,0.5972228024291316,
 0.6967841639166236,0.7960323962969053,0.446407318598001,0.49925942685762215,
 0.6173190258367942,0.6795020797842334,0.49655003059433656,0.7890694006189195,
 0.43850042859723604,0.47576897248369443,0.0744596639570334,
 0.7377667414892634,0.12450625793924563,0.1338752022380728,0.5324586673480941,
 0.0565505880615117,0.7019831634828725,0.4511379999876143,0.2469027826367245,
 0.1994638926951382,0.9056888548093436,0.16709523591188646,
 0.23654433836198752,0.9497491191728719,0.48971272016650413,
 0.28556326330160287,0.5203275854283331,0.08994851853015484,
 0.22165052771071403,0.40899434164312853,0.6499427434879828,
 0.2708188892643506,0.07862298855374772,0.3414137740702261,0.8581974235826144,
 0.4911586119376078,0.18051222074630813,0.06227616713637396,
 0.3758778953027955,0.5815458005616867,0.03849852183157443,0.9285088272230423,
 0.718692430149915,0.9075104614034479,0.8071008493210066,0.5666366715516457,
 0.5455370481598041,0.445795010323919,0.809694241346298,0.29384088131502306,
 0.3791610406758119,0.8276256385178339,0.2812082589246707,0.9717934022987876,
 0.8680080491639265,0.32563006595985966,0.3388472186888861,0.7878137176870148,
 0.5739116569777947,0.8276141536344084,0.9122120926922931,0.4861526785499207,
 0.3120470374940516,0.9050198264812576,0.009463291203010482,
 0.7024966421440425,0.6754952283689931,0.1850717272113115,0.6303741599259319,
 0.6178804573629227,0.5691829957938982,0.3587955173059203,0.2781665193618683,
 0.37612744144885146,0.2879304000867068,0.09152212081539235,
 0.5218775012310317,0.23983144490766695,0.5798709940634079,
 0.32177584291992145,0.03734641567300889,0.8426514101806516,
 0.9011387977849793,0.7459660046770983,0.18603093254842273,0.727571839391161,
 0.18502335714606521,0.9962945346775495,0.4690851040766796,0.9703645546378971,
 0.841277069828593,0.41478721953154596,0.5841110109238177,0.8769127522084211,
 0.7352530526655504,0.0395561273020727,0.9829624314784481,0.2600489197127409,
 0.04846289561864969,0.4224772449348002,0.8852884689806418,
 0.25711603379894676,0.2762472011839292,0.3484303070052258,0.8161283019954136,
 0.6615286720678448,0.8787588849551947,0.608585179818055,0.7243233671574892,
 0.16460857964194153,0.906066811582574,0.09671525836007389,
 0.05480728558107384,0.37236663231994493,0.8771825339224905,
 0.6989193071660696,0.8906487607925161,0.7639376803441567,0.7997823911074833,
 0.4500045001942772,0.8395954890565878,0.4873257930385917,0.1829884821801192,
 0.8822458613648925,0.9402012843459476,0.7352214100637863,0.3524383604329241,
 0.9546630949280012,0.9171783148327158,0.4742643493164326,0.9113957970085125,
 0.19162913772418733,0.9340108626990875,0.2948037986278761,
 0.033932950185380006,0.0844568562605792,0.1931502714533383,
 0.2417778327292187,0.5011608541338327,0.21992416447991137,0.8994398633800407,
 0.3919704165279758,0.17711485671364102,0.8023951782229963,0.6413645739574467,
 0.07436007878687467,0.2394788865658728,0.7694797517762452,0.8732599011012839,
 0.9501559514857622,0.0869297156164831,0.5296378017129256,0.2715084026264516,
 0.9866884807812748,0.10807146730510153,0.6910591114308795,
 0.31488016431452814,0.3407344971825418,0.594229217073714,0.671084826051528,
 0.19949482896892623,0.47893674875754577,0.7705855409851373,
 0.8922677860846235,0.2455842272302985,0.4031184547230542,0.09614624764030177,
 0.12591380106721506,0.3940437465544654,0.3466628878746154,0.8950649286790522,
 0.23751006692067522,0.8731847837983204,0.03827276736191798,
 0.01754356106595578,0.2735969667403789,0.9163805216890893,0.9316383099186509,
 0.26088620289933917,0.401741318363773,0.48888044866427793,
 0.27633731707621356,0.5810748206241878,0.5590572879117011,
 0.07168241777601858,0.20552964285220798,0.581583995910703,0.8779972402904741,
 0.5020784362316744,0.27822076050873845,0.011960426504749155,
 0.3945884583538797,0.6488977479901964,0.4785036175925199,0.47363020668085964,
 0.5898452578815215,0.39609662548034263,0.6262942294967812,
 0.44173026324625253,0.3743853626774738,0.35893543076818246,
 0.4994715905002366,0.9443993504898683,0.17038367351413708,0.5191015229705918,
 0.6347247101536165,0.7598709801807109,0.8845985976471779,0.6926781762152981,
 0.5308472815094504,0.38278254172786197,0.6445151714367029,0.9563830899329903,
 0.04326932030302444,0.19747674105872715,0.8359261825657394,
 0.05098208570559348,0.32658582355791776,0.04575105378558553,
 0.6924127544421605,0.6782242415913745,0.05891242160213206,0.4753954474755666,
 0.4271510202046267,0.4588575237151187,0.9912048661780886,0.7325620571289797,
 0.771592792950539,0.26723001495922527,0.19685740158690834,
 0.022840727826032747,0.4076666261773917,0.6283725186895088,0.493659502019151,
 0.9497172969193707,0.6813068142298674,0.5362520740337118,0.941458029545935,
 0.519462553755348,0.4637353743136948,0.3527323663033237,0.41937173803119954,
 0.954479269024955,0.05114798350600258,0.6106063302645237,0.5885862818478276,
 0.7258978866272061,0.19813597878892553,0.673754516064154,0.1814424920721539,
 0.22417357705198027,0.5212088220062377,0.2781186908120815,0.8811635001376692,
 0.6727368664892365,0.43204408921823323,0.45280265418442245,
 0.7280335899636636,0.42277486740510317,0.05334665640187286,
 0.009521081156646183,0.525859746482068,0.9793083998594516,
 0.23691385167507628,0.06833713363185112,0.5405294952686902,
 0.8886109474723048,0.7347829963731332,0.6301099670306871,0.2915546533515988,
 0.805927745640469,0.37953808670194666,0.5227575744575709,0.2626786436342674,
 0.5487400828150424,0.6319427879814132,0.5892846866112862,0.4047314110296212,
 0.5236175969614467,0.21589694116688807,0.15850946320282366,
 0.9633872223574046,0.30126435225658776,0.04649705065562504,
 0.34587568543857405,0.3111493373306332,0.32350688362971913,
 0.014231967909742016,0.3116043524118126,0.35953900172336883,
 0.6892268722622408,0.07220006179973026,0.5486060677120312,
 0.36442404278320994,0.688699420888636,0.7024489093824249,0.06790169847073457,
 0.19109133548696,0.7553047615710419,0.08703531018885324,0.37171618994764133,
 0.9110512771399855,0.6333931334814192,0.7344086475817896,0.9056268231796368,
 0.9382610253061233,0.7251569654513497,0.2927069046347487,
 0.026595964112857584,0.2978577850709482,0.9470895286192251,
 0.17225420911093026,0.8171408688116847,0.12577547142576584,
 0.7200420889293979,0.44563379623140364,0.34619186297115223,
 0.12925552116864503,0.13763499067350815,0.7672625527685696,
 0.4398386124434751,0.4768947407469277,0.17436802332618861,0.2912370171021431,
 0.20053222533463833,0.8735384668776688,0.10080024747192373,
 0.8161779044651782,0.5222674548190577,0.8502297589555955,0.612305795707492,
 0.7430594094052224,0.6865968031596817,0.18895331396797177,0.6643734396553079,
 0.38893771462414717,0.8261692622516273,0.7695875765539201,0.5718799060464457,
 0.7152205606769062,0.11571528887541094,0.7216590812823582,
 0.08163746647858017,0.1466131324979364,0.8544291275790419,
 0.32810994752667766,0.9911443168624926,0.253072549907027,0.5673356147872796,
 0.4538029140306671,0.225244537688007,0.013414573768868454,0.738094322961353,
 0.3392066276886585,0.34417506149545674,0.6232522685810883,0.4739500380704432,
 0.7251714743437593,0.7167883630359918,0.6461612018839988,0.232470065372248,
 0.1075180759490475,0.6005969808674725,0.359686778655641,0.17881709651955702,
 0.255941526376003,0.7899281161589289,0.8493358227859541,0.29042853780889166,
 0.34620850948501736,0.9561818922671117,0.6819064374644723,0.9974440251982544,
 0.8735485407210498,0.5460310953023543,0.4562925622102926,0.3272022127636728,
 0.2853252629265025,0.09915877185127264,0.42201360490884143,0.923888470008362,
 0.8546944788126085,0.9178046864673405,0.84034279010948,0.5779450124355465,
 0.8600470447914507,0.326942052433179,0.021558002714497926,0.3133928399064668,
 0.7730955044506496,0.6019549684450451,0.22921898435595311,0.6924296627485047,
 0.49581624733634255,0.6882108996384995,0.9591247787720608,0.7528166212579662,
 0.6926846144077404,0.49805420898508623,0.3353509701691998,0.8898871649910989,
 0.06786368236856388,0.3697883380803466,0.07519796158573866,
 0.8293840486191782,0.2924562649589195,0.6760776044676068,0.27405032657346917,
 0.028781791594272632,0.3418349848587132,0.2986029391174214,
 0.8416051708342621,0.42456494125196076,0.237633065416469,0.6221419824240849,
 0.9508602636833015,0.18785085307107985,0.3203538287106138,
 0.40239970541801595,0.33701435824594905,0.9742518866564135,
 0.6355564432259876,0.22336176435277322,0.9189556298625035,0.8445271346327092,
 0.6465329661147894,0.4218465651300197,0.8005225293254429,0.13564623937829245,
 0.26969938334894095,0.23953292953972816,0.8893624493629473,
 0.5637642965564104,0.14105335235437977,0.4830628264401855,0.5390200164487133,
 0.13428996918620106,0.02462829078451967,0.6445837368932013,
 0.26090665817636727,0.5407740711086296,0.975962663570728,0.00627733082342985,
 0.6823582731262153,0.07239831092262439,0.2609207695050114,
 0.10318558944791001,0.4265161823258722,0.5272780111725076,0.5646515992378762,
 0.5714756028955991,0.7104972981049187,0.007124349184323253,
 0.07383566897541716,0.4736670573756515,0.11253447174343212,
 0.07327862469698221,0.13193958783911885,0.5430772389669697,
 0.9108447917347441,0.3889458611306936,0.443888810769158,0.23195025239458766,
 0.5930617706862583,0.8499366749596814,0.743543667559645,0.17335284284000763,
 0.074570199636526,0.6813545705631263,0.20350240265336084,0.27448914116816836,
 0.7108646442817091,0.5862925966365004,0.903808694880833,0.6306588602112413,
 0.04638056653570588,0.8810295521153928,0.38041994623631636,
 0.5552328176004091,0.2717619726868078,0.09430981308061548,
 0.40701678341419545,0.7709218563057191,0.48947223216000024,
 0.5221994204539181,0.5117048803212186,0.0505841968293661,0.3202527515203124,
 0.4407317032019634,0.08402682667798123,0.8232467996099242,0.6280004132512816,
 0.4363249016538905,0.9836825654870313,0.07762858223197633,0.9557754833531061,
 0.446464769369706,0.9659760453145081,0.9780727729888574,0.17347913940924553,
 0.358970331410119,0.3493843743177123,0.3699866643541757,0.8567485507334826,
 0.5567043603412262,0.7770946960151414,0.7443365580841034,0.09036662581091814,
 0.031768756491559635,0.8717360606177288,0.36311999515213444,
 0.40425279872749464,0.8433232102214079,0.9279672774376918,
 0.10249354777252306,0.4955382749804841,0.8987409273176501,0.9778188137833514,
 0.5227645223914869,0.38203254390122954,0.24067478999434544,
 0.19365760235428586,0.9226138765227648,0.597756056015267,0.254156042028734,
 0.5235508819234107,0.6755845423410114,0.07117598246250556,
 0.20537265818271977,0.3909890747399555,0.2515741396618658,
 0.47774770579036974,0.5428622244404502,0.3916371547073739,
 0.38187106214012567,0.2547727355642644,0.4819215187792648,0.4042518716886311,
 0.3241058249571558,0.8387038193464899,0.019278271049230078,
 0.7326134711439813,0.6018409466603389,0.5646342497872485,0.7608005125925763,
 0.3626242983157333,0.46885981231438745,0.5504544279780268,
 0.11316696522018332,0.13893880366132094,0.69416078899275,0.28729217602771606,
 0.6909008917097043,0.6720656301437062,0.5571061572857683,0.7631739705134561,
 0.5323531721955219,0.5459191170498856,0.7151540495781433,
 0.057597963274198194,0.6502586784406325,0.9999450019251646,
 0.21834108193042645,0.8006642849309873,0.8038484202794091,
 0.04787940521738199,0.23231922575533404,0.10184608975102016,
 0.5051101791016621,0.4662050099874976,0.34509023274840356,0.586005573648734,
 0.00997465393622221,0.6673917110713805,0.7091121580652906,0.8285051387056449,
 0.06102460845884794,0.6612581498768658,0.6291921630870383,0.5682343814911767,
 0.47405328478952935,0.15702833713047804,0.9818007859181036,
 0.8616280307960157,0.440076227556795,0.5235363181665208,0.40034432418033883,
 0.15255549417952108,0.026405467555326245,0.28439013747773445,
 0.3218078818415039,0.7657115175451562,0.6411103549280392,0.18684537466395434,
 0.27258795734939656,0.34207804555028853,0.4541767582188059,
 0.5679160583965828,0.21781041270840895,0.5799748122655216,0.9363886505844078,
 0.46988457821005836,0.18402750144321112,0.0688322430110796,
 0.07163106204351066,0.23504875401494263,0.21456492369543823,
 0.8749868488965807,0.31911298041813685,0.8595624712377822,0.8420208606664239,
 0.7133646636653589,0.4078102490599148,0.8295314850612395,0.10330176831439941,
 0.47469188016063235,0.837316557967659,0.5427477860955976,0.3030773760131088,
 0.026571747581609517,0.6769216021377913,0.247990336436812,0.3328456774641787,
 0.3777831354647677,0.6120191601118282,0.1360552020072454,0.1786202320483118,
 0.08069915964393648,0.2850910473026288,0.378117713374445,0.22800210865995996,
 0.33647637707490174,0.7644032863191478,0.8495026335149853,
 0.06304340194130209,0.1345372779383972,0.49804729743863096,
 0.010271291420735129,0.9877503925801873,0.47280458585545393,
 0.17409923018403983,0.23186910063901867,0.3029552521517007,
 0.28379353196142565,0.18880169640822952,0.04239762281314563,
 0.13325314546591516,0.18688902673576346,0.9907335286210427,
 0.7013427265711887,0.9599696397992109,0.8395374878569649,0.7909171174638,
 0.3782388192776327,0.6165081105399658,0.972957924509729,0.7235274982585816,
 0.22978629831804032,0.7250726673102189,0.1314605824418288,0.6617557275448436,
 0.7421467492948487,0.8176327167137843,0.822646010052582,0.9461695947549758,
 0.9491892543607074,0.04364507657824668,0.6667985461399221,0.0875046124735217,
 0.0011770536062756243,0.1640613324118334,0.14450907463415752,
 0.174400145788553,0.7486776714913156,0.7368805728414773,0.6765698649765011,
 0.5407905603316492,0.3218956898198926,0.9460692344739826,0.6795795009460154,
 0.6351065767137632,0.3790095760417759,0.5492855792477627,0.3196011502129469,
 0.9471975665815299,0.7867368770205437,0.8959690707827688,0.24872538626108553,
 0.08395220866433473,0.3497812531514911,0.7445518258386616,
 0.30082008494816115,0.11581982860099405,0.4015504362994966,
 0.6326815831844143,0.42379965202457703,0.5153325543646191,
 0.21160132287440714,0.11990959231028553,0.5426700584475501,
 0.2598125458644167,0.5300308514144297,0.17941531695665258,0.7015412296960872,
 0.11817623588297632,0.01883248866197218,0.9700692590983935,
 0.10316735172884095,0.7544772886153244,0.6333024571229796,
 0.05431968623092098,0.07481841439691295,0.49765513407215123,
 0.9368278175472535,0.9958560451105386,0.21553638749119464,0.367972664144788,
 0.8881543193049356,0.9830503336104688,0.783795838797916,0.16042922769466095,
 0.650758419770848,0.12597357306921186,0.877392419540551,0.28966051042701435,
 0.7099371013624177,0.7505637383759026,0.08102291134518458,
 0.24763585204565597,0.898408899286745,0.06690619609322768,0.881705577769082,
 0.3470132148901387,0.047166620625533895,0.31009152335059464,
 0.16876190301526994,0.5364764375696176,0.6383370754703857,
 0.10605006674318562,0.19218565011293376,0.3096129482696761,
 0.047578579826518363,0.8538620896207687,0.9348154452059296,
 0.28701189256894866,0.9491137954655444,0.5679422084525215,0.9012379627701954,
 0.3624715035413103,0.6363494199784294,0.38354387988712435,
 0.49791198091645017,0.27931576910932754,0.5793343457779198,
 0.16632408880300575,0.03617216183707117,0.19905390403124323,
 0.6617488779872126,0.4332035194388403,0.8237645599415675,0.7249908512029274,
 0.4987374853014198,0.8902806459384286,0.9237612241640923,0.8322447930041341,
 0.05844124538527762,0.5157627325474039,0.9248932941551407,
 0.28253786431990713,0.14021773714512753,0.6097130702074381,
 0.28359858635483737,0.5803060297147199,0.3031321618586046,0.7377076111686739,
 0.036409528058531815,0.4957738588846923,0.2119742344714285,
 0.1482752648065332,0.09306565714146242,0.7005331875136077,0.6761655253713063,
 0.4879094319627436,0.9055502381999514,0.25053408687699796,0.7859772095347135,
 0.8578577262146744,0.9739700915201385,0.5848973012718527,0.638734941535141,
 0.5050591947250669,0.8480201344081026,0.20445090453386205,0.5233739137125126,
 0.5746333942703978,0.558548302143093,0.9094821390623835,0.7874843176369836,
 0.2467389653393578,0.7895585258149883,0.47132106385032513,0.6131913971494162,
 0.9595177102496184,0.1979472716009265,0.7736604054820766,0.9365735786485898,
 0.0800661049108834,0.9437039106818489,0.031693116994517934,
 0.6570773708330542,0.8414633515351753,0.5804529428150441,0.8364694021697718,
 0.05132875718288654,0.9252388051840784,0.1017896684935925,
 0.12847858859760963,0.6481694217707574,0.0984072022767839,0.7982153427486546,
 0.5422400838506742,0.7109351470918932,0.3734448076537481,0.5304341912126203,
 0.5897790934643914,0.6505920558985433,0.9182818703050403,0.14793376267753178,
 0.11615366004601158,0.4618219278359651,0.782453612603209,0.47174121398786295,
 0.9959736453245134,0.10051250916268661,0.10491102831095123,
 0.9886040420993295,0.06510525546827728,0.5877048960126401,0.6791505756700053,
 0.3013064087864399,0.2508256603598009,0.8031008192092133,0.2031174359606991,
 0.9151193448724183,0.39872260564936146,0.8145157586203918,
 0.42381550437419113,0.9067122160367894,0.008759460397571472,
 0.3065705627879127,0.9065895811669062,0.4357976865275308,0.08888231262127289,
 0.8062214001794766,0.19754932044530993,0.06826337032024155,
 0.8678376086901989,0.3073271274248093,0.08672257070706202,0.7984511934863849,
 0.21357571237001405,0.16516081606526556,0.2717235402932988,
 0.8401417590275033,0.9667724727169644,0.8332856724882586,0.02538582980174886,
 0.3700360771466593,0.910035204995442,0.00770940855948965,0.27836726800905265,
 0.6242179818514919,0.6488987235092857,0.23344911739473329,0.4789551940066543,
 0.5886278439657513,0.993345729570528,0.4004618972025734,0.7784795403423405,
 0.0186816886902198,0.017862468952018373,0.6067327595237061,
 0.13518722554204243,0.8988053641851781,0.8097859031680004,0.5318173938836992,
 0.4390589737979944,0.5977127762207499,0.9602400031650932,0.37836150674703173,
 0.8498647429336472,0.20583708242187415,0.6196793919558785,0.73688191331944,
 0.5593275084050712,0.40857841683254625,0.783023647256185,0.9790628779600553,
 0.06923007776219336,0.5368143435388227,0.10683319489406862,
 0.29145886406087373,0.2500656102187533,0.5353113260146881,0.5484962041484266,
 0.17899312910477638,0.8148324324946261,0.5216230584872581,
 0.43398103497432017,0.33641219032140723,0.18323770562054753,
 0.5463827075730573,0.44900416129209686,0.15933906105728857,
 0.4622331088067859,0.5192570076753983,0.35454515599635394,
 0.006415024874182884,0.05590935628211269,0.6110236657075127,
 0.9790386785555877,0.18097556331173392,0.6651693313078926,
 0.27561845419046427,0.9946168253513304,0.005144428675347479,
 0.15976439594472425,0.8597533840386539,0.5616344294397201,
 0.44616119397420184,0.561793774804157,0.6625464602967284,0.186693852864371,
 0.7844680906441905,0.8755398663898515,0.2911835933226463,0.9977631528316713,
 0.5629464800525849,0.2434413202781055,0.11862700289125783,
 0.12357909600261596,0.3438987419221329,0.6181654679263082,0.5918570888168984,
 0.15878055967330484,0.2976095785888393,0.788757696864358,0.9709381282066852,
 0.253690962199998,0.3009530276886462,0.8822480736549588,0.33157138352569104,
 0.7709263091144141,0.09717922471701579,0.23263579578932647,
 0.01946464731398534,0.8813755246654056,0.0019965560351657974,
 0.013948868500891609,0.23165485436988353,0.23117975287181503,
 0.9280346866405574,0.5705273431922223,0.4151415031347574,0.3356438467121332,
 0.5195743380338208,0.1775903800785693,0.7054088679067656,0.30107225919170755,
 0.5452770769914312,0.7119103192265868,0.9346932253930288,
 0.061992862764912005,0.49851259175926965,0.05055374908743293,
 0.4115082686298821,0.16733130288237186,0.1806446731760819,
 0.10774517971392517,0.7320379654181348,0.008932268288035616,
 0.7532019619914126,0.3508185158494985,0.17371329306460215,
 0.20410844535168726,0.4234567661641767,0.5273736098259679,0.1159080007504395,
 0.5066636834088275,0.3386646849341146,0.05506291889091253,0.0481161953100232,
 0.34531153876382403,0.12877591245354036,0.9005744099005422,
 0.7330879357738499,0.18161697395033916,0.9992050311980841,
 0.11441128329084116,0.7416469029768668,0.13706756780887597,0.755185141140668,
 0.22299061010787535,0.3020568577885574,0.18815830233435993,
 0.4575283586725072,0.5731811874202718,0.6246962046480826,0.44927680075787146,
 0.3842840517001421,0.3759829502994655,0.043936181601579416,
 0.5174727430793672,0.9543101187951116,0.8621503938846337,0.3750407677130687,
 0.5692648895445923,0.49409515662072767,0.3061986511009608,0.8094411545402069,
 0.008542287905633383,0.24905480574284655,0.10555927056077896,
 0.9215074758063683,0.2728455549314117,0.24904621126938375,0.8912061037074585,
 0.5992535073650176,0.530162831227885,0.24441220512979833,0.5041214990459204,
 0.7802920313030394,0.3222535357532136,0.28089477885668357,0.1299235294047475,
 0.8511344187104672,0.5693602553934375,0.17529844579885756,0.5936637438119731,
 0.5990516914373218,0.43604021187602116,0.828829494420843,0.22933471484940282,
 0.6033796581215366,0.9408649281170702,0.9701721988527805,0.42676594473291096,
 0.6483664624126114,0.35608963277421035,0.41855398663259,0.5746387819336378,
 0.785544558242434,0.44632726902943964,0.721209022843782,0.7877390591187111,
 0.8803144924660585,0.9753089226951877,0.9670736142559843,0.5327001687598236,
 0.9388485756187136,0.3970935324557935,0.6574298335423191,0.9200676417321775,
 0.6158848133834446,0.6285522486861307,0.5041488246397517,0.24500031614059758,
 0.4980770666275639,0.9498898141090574,0.23196516949805612,0.6482887605493499,
 0.3513865988828341,0.5566086724292051,0.9275079398546157,0.2911469442505188,
 0.3429857377805222,0.0705895955350585,0.34460878228121683,0.8283912196308173,
 0.6170690614533714,0.41793679060265765,0.030151311774535672,
 0.17337632519868285,0.9396628011278358,0.3773725213901199,0.7685206515371936,
 0.3453890638669279,0.44301611741430225,0.4656848885237893,
 0.34914524201633146,0.20866275936367396,0.5293512447946837,
 0.16633523801603833,0.1287905332378,0.8274327526351978,0.05830967561306721,
 0.32681888380903557,0.04021025831082958,0.4023670807774401,
 0.5700156947420809,0.6683800939192486,0.5731231898560879,0.21773600739203813,
 0.4113708142156336,0.9003542591805218,0.9198344040533379,0.11037490389916327,
 0.35609230545221005,0.39243117883867373,0.18909517889614968,
 0.6401072785304347,0.9316744053204129,0.19528042379467758,
 0.32208766119567167,0.36289700888806575,0.9779218674783667,
 0.3157585917074621,0.3443814433851926,0.9738952421105008,0.26608360652441587,
 0.9446014618750469,0.74463985230554,0.5052408215348299,0.26408641323543214,
 0.016795181783478275,0.1599911576280173,0.04914740440538079,
 0.9050500399735499,0.6985847234512281,0.7019036606645729,
 0.008079857333880502,0.1749923981613838,0.7593301502825961,
 0.9236078426232163,0.5163099729417183,0.5536387479383961,0.2311332938866003,
 0.3131173900092954,0.8352384429751685,0.5763115878187332,0.344003611675338,
 0.8658543484376513,0.0028505072055708958,0.7870039663353727,
 0.4320383942439412,0.42261574445879213,0.8141820977482357,
 0.22168818714675065,0.8721531704270278,0.15823313599979305,
 0.7768209383273443,0.19009055580468126,0.92269972875699,0.12494979226501313,
 0.24338123988438598,0.9642005707687699,0.4039220666709835,0.8415066835998771,
 0.6408199182166514,0.8886185188174889,0.46956901030219866,0.3675809300885432,
 0.9950030939804522,0.9792611177291856,0.35850806786982914,0.9642907078525954,
 0.3736754319099923,0.22791028604188202,0.21582178305672073,
 0.9165092335776938,0.1959050627747383,0.7241178768763713,0.18692473298014933,
 0.8475769271409701,0.5651892102528147,0.7113601839751246,0.38290532750010997,
 0.718066316758516,0.6671898044731914,0.6038074511374247,0.6311016801627851,
 0.5865565249372715,0.03254752649185355,0.9209866042652995,0.6012387891118269,
 0.4732538554071032,0.36491009508037675,0.995329389831057,0.9998067205243388,
 0.16801731887653937,0.0801742662151056,0.43837639151622776,
 0.3963830246604946,0.6157591898838642,0.6452554306528687,
 0.030819456664106792,0.8680830180900939,0.9371162493691682,
 0.8965224925994808,0.9178348955385542,0.37526128708015793,
 0.27852272668943934,0.7088414604616075,0.9810166424861947,0.8429962917223348,
 0.4596166407035033,0.013625055639632677,0.7312448237833871,
 0.16179100895697296,0.19560790371031056,0.358833410488236,0.9340154776599788,
 0.3097400255584811,0.31392689487623626,0.15451819750789542,
 0.9340136840338222,0.2188342155736569,0.10530107965453528,0.6289432673969423,
 0.042364981858875495,0.5336576478102363,0.7338207112472237,
 0.07801010626688532,0.22764388472762742,0.36278539461477655,
 0.9055937574963898,0.23250429900534286,0.13221900749725857,
 0.3213414076299863,0.608944259955879,0.8321178284826889,0.3854569072557654,
 0.9714575687402826,0.8362060384344415,0.687644395913735,0.7103878686655435,
 0.4703119080354782,0.4384827876790591,0.6235499624054313,0.44168469543555244,
 0.8819358753236163,0.9759512829617244,0.2665663728698682,0.7083042961468876,
 0.6944921670147248,0.914500943138448,0.22246008098975012,0.744628931080696,
 0.0698032092032087,0.32438278625341865,0.12082146123429582,
 0.4579278652711136,0.1993294896876524,0.33767646858224154,
 0.10790117337652616,0.039930554676354535,0.7348444160291325,
 0.7464241357632153,0.5958043092569991,0.8684236639241624,0.0805180820356447,
 0.4639019138681454,0.242222105080105,0.6865827551011795,0.23371296672611597,
 0.4908780774954946,0.1519267579867143,0.42331934843782193,
 0.19981001757195394,0.1721292531445986,0.795883824695715,0.7782801538213374,
 0.47595486612734306,0.9945304515773578,0.27830258069030733,0.822332526105427,
 0.40261595050141263,0.22658266804506222,0.19398509498256677,
 0.680744441223136,0.7538814775061671,0.6831076980933266,0.40783250150109907,
 0.7410268559124007,0.3806423425869371,0.9673773421449852,0.8357312714162174,
 0.20726737410573715,0.6066174617903903,0.6701198529898315,
 0.19440662432061018,0.9487977928711262,0.8016497983653503,
 0.15669195400876013,0.18132312214106272,0.4112344719043045,
 0.5051145514371805,0.32994509777894765,0.6192910604346096,0.6350037218462803,
 0.7760391262257341,0.617459920780711,0.7023964884197617,0.9533804392705844,
 0.9328485152628091,0.042032603246019096,0.8829360804844074,
 0.8630246859815933,0.3824781409378032,0.6804461421426645,0.6688952321223677,
 0.6802524633563338,0.5224149955889344,0.14340873352246608,0.9484363002532343,
 0.8732584101613358,0.39575823563575085,0.21506427004652673,
 0.9457827981202165,0.19984269685455414,0.018503101730876437,
 0.15140714705197267,0.9538499206904837,0.32951973896734765,
 0.6988999317787249,0.9300118893326985,0.345151266060288,0.6753541094886996,
 0.5040452352121214,0.6019023433252896,0.45329112550905304,0.5403036349365715,
 0.8548680514868051,0.9785974085115774,0.9281027737699745,0.02071492811308351,
 0.684244818991042,0.06538146086278918,0.40369478820080473,0.6156046789403933,
 0.8057816444622827,0.8776881331256513,0.44664600643481367,
 0.47308410687993885,0.3004372672193364,0.00291577033393009,
 0.05675627090074542,0.9620473013182423,0.9685852856163016,0.2870407181638793,
 0.15358191143698463,0.78514762724873,0.7524942912724419,0.06509107242341194,
 0.7701851643853196,0.33838890468717087,0.5809422168079265,0.902668202271084,
 0.4256860119106797,0.1637804084927611,0.6970151142224976,0.9185905246585475,
 0.3177010705100275,0.05242606329409916,0.07036639718161153,
 0.18029991637440723,0.47263496914675085,0.259206020736232,
 0.07958348166597684,0.9952318912142327,0.5870376510628321,0.3164793589802437,
 0.6568339519890266,0.9775868881405224,0.25843374964014176,
 0.14816351291682817,0.23214552240608788,0.2306080765300771,
 0.030122095993069253,0.6408097059120483,0.8020089991390646,
 0.4226602504338648,0.4372899100549029,0.1732834153001579,0.9461418161144829,
 0.9341876162501205,0.2488322328102358,0.3427210617244869,0.7715109051616368,
 0.01462055957788444,0.33032533565935374,0.6992360950945983,
 0.17276402142005043,0.06047403325601941,0.44284607825393396,
 0.6256683763117844,0.9496999722296928,0.5256900348292125,0.4911439734432722,
 0.8286197657248905,0.44943205617319226,0.8505433154216825,0.2579964850492986,
 0.7804619077036525,0.7452212180659396,0.3372250814520671,
 0.018790325751639192,0.8918330780086359,0.3624140625731729,
 0.6489979969614593,0.75801619243767,0.3261049991239712,0.039874371659952224,
 0.3506940479826339,0.0104091097036777,0.7159620968087136,0.07012863360469679,
 0.2766838845279028,0.4877674054611634,0.4254686633792786,0.6573213115117994,
 0.8446453571300301,0.34480360437558355,0.2799821286455284,
 0.36482967220657603,0.9818663152690149,0.04690705913182691,
 0.23250649240501076,0.3182094867252696,0.636390996070028,0.31135769851592565,
 0.3518043720857613,0.53332898165663,0.19086713475411166,0.1564046196409855,
 0.08601310375712146,0.09448944848695728,0.012797834842020128,
 0.22347121217374533,0.9616539184155652,0.6818799621036035,
 0.16333591841322992,0.2627436531370667,0.9295328721735612,
 0.009605546190252445,0.9707783254607051,0.4525598605296708,
 0.45852689205785513,0.2426155558524864,0.19792580766001355,
 0.13625060048790705,0.02769521425637611,0.6298432705907941,
 0.23059342757742884,0.4887612629967457,0.6573604978784879,0.6750679822226908,
 0.42087157851234214,0.9482043662041796,0.055433921565485544,0.65346754014437,
 0.24116421405763278,0.6365696036702446,0.560421854118379,0.5899731872147423,
 0.9711396202810643,0.6663622771485965,0.6235717003711073,0.11989605977435103,
 0.665940401167293,0.5372594025581069,0.6826072666214682,0.5399217174922887,
 0.16896759521221094,0.5381221413605404,0.2871350211690199,0.738677309392914,
 0.7780181425016123,0.21277932002353828,0.7797526747926461,
 0.41993896333258474,0.9200506437531842,0.04024960118527521,
 0.27212989983649805,0.8648558753934318,0.2409489387672583,0.6994539248301991,
 0.3808748194009006,0.9510686269179262,0.37717947501508153,
 0.024401878240243646,0.8701429893837702,0.4483389308327699,
 0.2785734941178748,0.04976301407104622,0.7379253171064939,0.4497284607582607,
 0.46516691880382877,0.7783593979723255,0.2561133186169988,0.2183022415668372,
 0.3128817020446788,0.8885331146491303,0.49152656971988307,0.716342253526509,
 0.5254152851888094,0.554758170740708,0.5541888506694393,0.5310174515753889,
 0.37497493650661917,0.664916745365905,0.6975673872872964,0.28989706545452676,
 0.5927167027054601,0.2604008119050022,0.4895510128914504,0.16592168379438188,
 0.9992253412924711,0.38348416346851766,0.8490453657687932,0.9491014928199983,
 0.5908775944421863,0.5719446506647557,0.4558111843092384,0.12799959129163718,
 0.11239545466802547,0.4913997930950239,0.6767374870154227,
 0.49849953958897597,0.3756851501834596,0.8220936810433866,0.8383068349045089,
 0.9271135540755263,0.35548166365760137,0.8920988617286026,
 0.16283623043019524,0.36585848129156806,0.3753001497201579,
 0.8439360383411006,0.15342629371010053,0.5522360718219652,0.6547792073806957,
 0.2459149247463236,0.49603924620949036,0.6343823553962913,0.7878938856107499,
 0.34694015548523816,0.9033565698877845,0.36176618584547904,
 0.8393617076693197,0.991486565119222,0.7187192733007209,0.2328781208248315,
 0.7854488621333487,0.4394607514233313,0.2829866188494463,0.41640230027847414,
 0.5669769206933414,0.20908773469137132,0.8955830637872187,0.9795306406896146,
 0.30193745784043324,0.7535765421394736,0.1204912136616576,0.9772570738961367,
 0.0641889120759207,0.8484500560153955,0.009642272437988098,
 0.18658317394228208,0.07920879426544225,0.04444032440937473,
 0.24451175417989723,0.8502422974465849,0.8723353465032676,
 0.18037333277220058,0.05911915948045188,0.19050787844401962,
 0.7800288228298102,0.9519618710595404,0.5070986927758984,0.8832930317268299,
 0.20621684260905793,0.5471300145897765,0.8870038687795198,
 0.12846300330923344,0.8850750771786144,0.8763716045074122,
 0.04396801419429519,0.22416948111062995,0.21203017442622363,
 0.8776465339917324,0.9954102509733226,0.6534648946332424,0.3940157116215688,
 0.814928055443106,0.9771793490326467,0.9358331054436386,0.34487495082265396,
 0.761226391578527,0.33335180973856526,0.48865099757847397,0.5526464837129778,
 0.2683668615055934,0.3393968901603862,0.6226071747733892,0.24462002420628037,
 0.4580225442631186,0.7507621115128356,0.2653773157329079,0.6905213026021251,
 0.9478522970393272,0.602808839557854,0.18005562800370556,0.6057884377970866,
 0.5767498918975394,0.8574106925362062,0.41884271040576515,0.5414261993424476,
 0.49675813971051563,0.9426768137235298,0.8118354570730372,0.4925662966080856,
 0.6594466185291208,0.3220070782876674,0.02042023014100669,0.6096140826972697,
 0.9883259594622738,0.5633112371256704,0.8047891507302252,0.1407852266690579,
 0.9488769550635714,0.9460628850664015,0.29074538020780405,
 0.05320134254076181,0.618808539993515,0.30502897515709415,0.5869194291337392,
 0.5333586408543914,0.3340469376561048,0.7001170783847215,0.8424564408989064,
 0.8960642135593493,0.6469128062593154,0.7923612316717484,0.7334237673961274,
 0.8871849161156998,0.7521424503587211,0.6876952193142694,0.4933669618916724,
 0.6450038600115701,0.10635754196666813,0.8662336184406632,0.3331769163279562,
 0.9893068554036397,0.2364706833504583,0.5590911074881106,0.6930121937609023,
 0.23709924813302585,0.6036176483415161,0.1677996992701173,
 0.45746205423019215,0.9274054939344472,0.500163342823387,0.3266093836566081,
 0.2891524981521988,0.5726393764694879,0.8428517824092995,0.8300132694734349,
 0.5023258349169304,0.7459753934571869,0.12660127626646156,0.8751582228036288,
 0.4840669612205716,0.8270861372572195,0.543425687213777,0.8781047003499785,
 0.6628272562688666,0.8610852654619991,0.5966420016354684,0.4247878825554353,
 0.4690353698983145,0.8735049986047418,0.644730907883131,0.6212225487903393,
 0.33275923878291436,0.6444604348211911,0.5240239471015182,
 0.44473351599076905,0.3662971705007726,0.716471517239668,0.6799332171540222,
 0.4679999960162581,0.8849319562200854,0.5394333613718586,0.9591427063565203,
 0.060941607932478714,0.03669978651909667,0.6134896187118108,
 0.8992724357170435,0.13893680957198207,0.9715472554993446,0.883170936164953,
 0.8974838248942968,0.779928770308088,0.3635675282689078,0.7279732844482464,
 0.05945193052580122,0.21129716059041692,0.786149087263253,0.6855281566990327,
 0.020121690962664696,0.29146605865247266,0.9398598500478147,
 0.35558599466249774,0.4358850545974742,0.5964728283994023,
 0.23663603111264708,0.1728986053139897,0.47713868706669593,
 0.5313741849016895,0.9560536798672685,0.26238309225931944,0.9670423277443232,
 0.4165028213656491,0.7534084842599219,0.29195327658086434,
 0.43446912557745965,0.9288523127642169,0.2207542293174476,0.8912828935655001,
 0.36970661313766895,0.033031020009726464,0.6699859926709715,
 0.37088742964597043,0.8206911840876552,0.8906515834938424,0.8696452131349449,
 0.6042013620974318,0.08876640713928918,0.08369702077619201,
 0.2981141908595216,0.09416864202421782,0.7524595927223037,0.3838681715147414,
 0.1791205453301551,0.6259164970991863,0.17807429126673713,0.7608444183682536,
 0.5769958453260966,0.14420176751485714,0.0436527023936123,
 0.02283135865621222,0.5070349676670264,0.3987675548832167,0.9676583675570085,
 0.49159215961026115,0.7135295790080437,0.12045284081155594,
 0.4411017014739861,0.7159294558984629,0.6176681200923694,0.7901170061600563,
 0.12767499372205116,0.34294475668635505,0.6177463915567547,
 0.1521532794800382,0.06867883442400735,0.44607780717702417,
 0.27078501883546624,0.7272232559440002,0.5354453322232333,0.4698133205388981,
 0.37027062202330896,0.14741153351801706,0.858980403708911,0.7010059911310225,
 0.5833439159467171,0.43247978751320815,0.33126588118691014,
 0.11724509823489415,0.29654165624946116,0.872424055863968,0.6423599840615943,
 0.35862850673200375,0.021860534758735617,0.03160964470355565,
 0.5864417722153685,0.04951325457519251,0.1692451875075004,0.5551093324975169,
 0.6095856516703282,0.8024627659575213,0.35870872321556324,0.8923461941028087,
 0.3397367341407581,0.16709612255231587,0.33037447089330796,
 0.9727681113303807,0.7422334963200418,0.4504812446373426,0.9499939690590632,
 0.2694707437979682,0.7456028827863831,0.14191832542873817,0.7097360541989106,
 0.09575627454768276,0.895816733448263,0.38992414953143406,0.6536363340840714,
 0.30681628867251387,0.5941727628878232,0.732446733728286,0.6929427100680615,
 0.8322094520677085,0.06235775812003941,0.4919393075413063,
 0.04054247682640855,0.5662518184077103,0.051954539149716594,
 0.2656075272551247,0.08863843975151409,0.9545025657149177,0.8177756616647266,
 0.5230906269112949,0.14589332959687695,0.6726605391082933,0.625492906665345,
 0.4613376481983851,0.2502445221766699,0.19681337144700783,0.6413718017332639,
 0.7650796669018087,0.35070427780986313,0.9834161719765376,
 0.03009197765230731,0.5360996907300483,0.6321264064891918,0.9428547548161765,
 0.8118047389814823,0.6072453339674845,0.16146704698978875,0.7919742680658799,
 0.7233098995621916,0.7189210561291532,0.7406997097561879,0.9720731474406578,
 0.6094523675056636,0.8353163109515588,0.6582441328621912,0.10392408314792811,
 0.5149995832401734,0.08708955022247777,0.10246938553979723,
 0.22008334382549666,0.6010745829626878,0.4958766904046208,0.9903416461878874,
 0.3443980976979676,0.1151781771938225,0.6397619803652792,0.5284381422960012,
 0.676748473749524,0.07645722306013703,0.5310077666232591,0.9294993139242351,
 0.6260771941770007,0.1732465298269491,0.13086727649623953,0.8901699105308368,
 0.8836510595668633,0.4705743665837634,0.20146668502744536,
 0.10260841503966367,0.6433379434209718,0.3541932151266872,0.6248902028853726,
 0.04155736163378687,0.05682655719860907,0.4396971367964895,
 0.030055651064513023,0.8835655447821563,0.5057544594291744,
 0.19726027130772772,0.8865069991061315,0.45654912793333435,
 0.5887167931169808,0.04619342046168362,0.4003004676673262,0.9268585091366168,
 0.8810766897267783,0.401576924259595,0.7445191312970776,0.26501288392768885,
 0.763642057182879,0.21288529758204344,0.7889681402240736,0.25000587577991784,
 0.33562622485134797,0.29197826818830497,0.7323442456361428,
 0.4691545427051731,0.9005591019908844,0.46453214905084594,
 0.03913942418193672,0.5321585608469925,0.11719088158691482,
 0.45515800614156465,0.3590545843774722,0.2342122156596178,0.5321081623843512,
 0.43463685589920753,0.15693012030660358,0.32342254954631156,
 0.3575785012022482,0.5003053895701569,0.5087568830799194,0.10914689041495163,
 0.6721504557914662,0.6157896443266062,0.18239104662711136,
 0.36537999183226666,0.014543095742219347,0.3158799789319138,
 0.9313210244189984,0.7154157214070371,0.0576504290258002,0.0331249844090058,
 0.7706093677333126,0.7305826753211228,0.019664692043930176,
 0.3458040725850726,0.8699341400593568,0.9891566880530677,0.896056378184441,
 0.088546760806541,0.09872265281815829,0.7025227185387383,0.49919058774458724,
 0.8960894463402518,0.701592761816427,0.44515291866586804,0.6031776238977969,
 0.49048572080261255,0.10928496402809118,0.9037745484084907,0.757169420567863,
 0.5122306391032013,0.9257666322745266,0.6873803769382225,0.02978177791138492,
 0.061162412450920356,0.014841577428013952,0.4625733508106754,
 0.9187763728202923,0.8862067106470487,0.06044556875864082,
 0.12997355006967914,0.4150679842160464,0.09096057360437793,
 0.7023963041272445,0.6506308311979849,0.5003478886406908,0.4822620837574716,
 0.8512891349987294,0.5605042561660998,0.9244993094521231,0.20347495790049575,
 0.8545185645220622,0.6063619069929058,0.030691477969210457,
 0.6997447222438459,0.18588635452740854,0.3326876405981859,
 0.02565149436395353,0.5095980598507909,0.31776326777102293,
 0.04966560470364123,0.1929537054406235,0.9812461929767633,0.8436650940231712,
 0.49002190119605604,0.5478173146873498,0.045377712324828345,
 0.034349626150640966,0.25513358130895514,0.7888477521039061,
 0.4352580356657114,0.2544232062106445,0.3104961805467029,0.661729729059008,
 0.9812798816805227,0.9053328425128088,0.7558785873596031,0.811551161306285,
 0.12081807705847192,0.8422357330990322,0.8793134106258504,0.3387543251609729,
 0.30739250394231643,0.3245791481108614,0.010274540046234737,
 0.4982151330696305,0.29055292099902186,0.08707198219370405,
 0.7005988166710341,0.6660637161476397,0.6225219897593899,0.5402274027781061,
 0.8413238543599668,0.03583497260299562,0.6148263340574021,0.5341298133794387,
 0.35600497446322843,0.6770149722322522,0.7833833954059037,0.9304953372029782,
 0.7062822586137113,0.5766806085366555,0.023468803820277384,
 0.5556636212668776,0.2953678373980869,0.18652777898627992,0.5643793153759936,
 0.8706313531694951,0.4025888861823955,0.02373263899905831,0.7322575757014517,
 0.9622117035152562,0.48098429087762173,0.6890639456459751,
 0.43279026429102474,0.36570638083568174,0.7209830408518698,
 0.14539792369080862,0.8868092790275255,0.6925753287388577,0.5557115702744391,
 0.4092915641293008,0.5392235088455178,0.8405136660507738,0.7969732949687366,
 0.810988228953867,0.5540530506928074,0.11530799143607384,0.7051310721688572,
 0.5437837521378169,0.5512284067025089,0.16242685287057446,0.3444042354410827,
 0.041396395626077886,0.1977625729073682,0.5979919536105125,0.368754395312055,
 0.00848196546291402,0.2813957499093994,0.7772384234508518,
 0.43673882853584733,0.17387193705188198,0.07461847275425126,
 0.17233770121928538,0.027265687898462176,0.918118441974201,
 0.41509679982634606,0.8198311748746907,0.4611852306362024,0.7376001567997031,
 0.09520574221135225,0.6145701774751019,0.5790949258349082,0.5186999151937646,
 0.5066601707488826,0.7256436391949433,0.7448714278724601,0.3437567894512903,
 0.3645594415147486,0.3273825717187612,0.9983763893654318,0.7666578396413368,
 0.9675710056432826,0.5200784468696596,0.44600839075733323,0.3060302551091586,
 0.9983831750351431,0.07886569729182091,0.4194484731297765,
 0.014049196394973928,0.9889742390375392,0.8176587139487986,
 0.2061663910804885,0.8918905762717378,0.8323402175883432,0.5529490494524764,
 0.3268370545295398,0.8349612704195541,0.26069288600993124,0.5594645175432018,
 0.8967144999784845,0.06269086321346629,0.2323511556335337,0.7963409181026626,
 0.853282459428639,0.44971324258706613,0.5129115722439257,0.5287747644885997,
 0.5490021108044858,0.5435706291987497,0.6709390911916836,0.8008962048948969,
 0.17647212714579075,0.22847379912877774,0.541190064430596,0.2069119240802385,
 0.7231119400108509,0.6383321884067086,0.48524173490542744,0.8009532153985104,
 0.7576114682300797,0.5421528957385406,0.6351618569897979,0.5081430522416308,
 0.36272967548366064,0.8192566956871371,0.3577946201108668,0.4525302068535404,
 0.018281162061970946,0.8616539224943857,0.6737019798680779,
 0.15845385263517642,0.0622023192859541,0.5195924487056814,0.7360049989013004,
 0.16748448499241342,0.21469868961589367,0.21331091357054233,
 0.9468739509315213,0.1711240650063316,0.5912193914105693,0.9952887665211357,
 0.5569539085965955,0.05919790501264166,0.4705520961115104,
 0.42982236215110925,0.10681587833635409,0.18525390983053058,
 0.22617679428808513,0.5824453398463901,0.6516928486353918,
 0.18454589789851295,0.31103403196216584,0.810142193589193,0.6752708868057123,
 0.04399857974017851,0.6222976711463619,0.021857915465836708,
 0.8039826778833167,0.19275976628481506,0.7601559025847491,0.1428556966528104,
 0.035073141451455236,0.05820878394890461,0.2326735801372487,
 0.5119250879148886,0.9037937525805708,0.054465866059161705,
 0.5583898501494776,0.017637344666134536,0.49302524075459686,
 0.06627218177023175,0.7680595747412571,0.8987476277371091,0.163549088894276,
 0.4064094179247515,0.7186531093807247,0.1806521373976544,0.5635443215698118,
 0.5452429241176276,0.8730192726634368,0.8562255509386671,0.25977904540550645,
 0.2737485528449023,0.4867514651657072,0.15228752500575038,0.5100086911592621,
 0.10596566681468578,0.14573720221171904,0.7636350133534383,
 0.25890786441644087,0.3110562602350837,0.4315663387124058,0.3324979082815489,
 0.8208121469597484,0.8563551702424238,0.21014743182093432,0.8067609110391621,
 0.7687783580353321,0.6177668395611816,0.02329867818192588,0.4070378706080381,
 0.9968040941046152,0.2987442485645687,0.9841923468786292,0.4211352651850503,
 0.6662683436765633,0.8384868623826794,0.5543486373119859,0.9225159289160143,
 0.5429755249153123,0.9130593665720312,0.3857394236265672,0.43249825503222905,
 0.41192857359231194,0.28480279084952975,0.004379210804405398,
 0.7506637999110707,0.48405055607962855,0.11228398489625335,
 0.41747957248597667,0.027060857136351868,0.38219721127694184,
 0.8329450091223768,0.17492183794519656,0.5907655948481079,0.196055993624478,
 0.47152446095562905,0.5280351232635366,0.33183318617445223,0.673626610354941,
 0.3270915236371924,0.7149043259323206,0.6997867673530541,0.06835844442029815,
 0.5201323099886667,0.9084767711103341,0.5929213490941075,0.9677785019338381,
 0.31832252009269235,0.957427595093001,0.6578791143794086,0.7050220778178519,
 0.3548069616978491,0.5296496171446236,0.7086344425598377,
 0.009049867423669089,0.15886034138304783,0.3936712416318844,
 0.7650189447048904,0.8974505929912842,0.4683956304717898,0.263246864149422,
 0.762205241093667,0.7777197653180952,0.9996164014666202,0.07471734959705434,
 0.9536378670347178,0.02446168218368361,0.8263983720349802,0.6946239828955689,
 0.8839223273112506,0.2656505536908519,0.7844485075143669,0.6633324165959273,
 0.06545476427409036,0.22744442385688102,0.34085754470003304,
 0.7988586447109082,0.7420937916887852,0.16003124350967846,0.4934779702718579,
 0.5209227040917426,0.8126261016343618,0.07097796654901167,
 0.17899928018336886,0.21188120247844489,0.39575093668084804,
 0.719172723432999,0.841351055128375,0.6955840977750645,0.035490240835354125,
 0.9333062542199315,0.9925528503370595,0.8554639993832225,0.10304782862081696,
 0.9032640086710972,0.45033026030354906,0.6647891162270472,
 0.32524725515789876,0.7206169706114048,0.9534829179340791,0.7819050836281187,
 0.5912864116745942,0.9293817980018192,0.8644892572397972,0.4281780426311321,
 0.0493941397888894,0.2520952612572591,0.8358969856997724,0.8554379501482877,
 0.00935239357265738,0.9587499669097812,0.10009640059731018,
 0.24644821633219227,0.6513376723829154,0.10093352899446317,
 0.7969671112333342,0.06389850796476104,0.5425024351924076,0.4883035895332877,
 0.03447499822343225,0.13906376331890213,0.17038414476487773,
 0.7911399441728062,0.5392882073853849,0.4614550059975817,0.9697867706989052,
 0.5085669687645451,0.4557731368949136,0.3273696029155717,0.24389482304885168,
 0.3700128641410868,0.035451218398734996,0.8681000010864706,
 0.0914522976319081,0.08936820091099018,0.24862454061316985,
 0.8099449043442166,0.7925956254310162,0.14365057143360627,0.7291599161964037,
 0.4820375058506762,0.5357147852341405,0.23576619190730896,
 0.038612890967820834,0.6598372415310322,0.8921350319511983,
 0.1848293032131194,0.5583647663163225,0.00525109055065931,
 0.39309575926367024,0.9650781555683384,0.04313051653650857,
 0.2893612463992238,0.4676231223433809,0.9801878334397724,0.6883114483288392,
 0.17676831035933527,0.6442472192468647,0.39169856767577294,
 0.008170512512350903,0.7643820713444807,0.9917020161461373,
 0.49988748364170044,0.05351906699798392,0.6014131373866399,
 0.6574533712008087,0.2617517798253578,0.8604780297432225,0.7700629167550207,
 0.4099959160494131,0.11106333830015536,0.5084974276056484,
 0.15524863468053085,0.34866705827255284,0.6676117612572652,
 0.7009170709860796,0.7879838862208457,0.13076658179702472,
 2.14249553538437e-5,0.1083824798009545,0.6687339168885866,0.7997023671640261,
 0.3744252654388557,0.034159999534700725,0.7515438719112699,
 0.5763988414600045,0.079012435356001,0.7137872025879504,0.9875247720357165,
 0.6037615484536215,0.6215877494029213,0.37106058655628305,0.9057628415127303,
 0.6118309285009121,0.6820399363143356,0.6806936347334691,0.0702653910878489,
 0.4857770339578986,0.6849266434787431,0.5212675258695865,0.5243413251373296,
 0.5013923082168597,0.1902304568307548,0.993709366850755,0.35387056013718776,
 0.6547091515455687,0.015411804670244833,0.5523749454716281,
 0.08180117964485956,0.9896165696792821,0.953518991722249,0.6676201953713926,
 0.2024924483731333,0.8741332351249658,0.9327845627136414,0.29151888336306175,
 0.3498884555894175,0.7005120949615943,0.1280651337082933,0.9776699983712314,
 0.03524476886924621,0.6865413330162163,0.9086972673925183,
 0.41748503810668214,0.5166404953409153,0.5690293561462876,
 0.28145291840698483,0.3399880710950324,0.11247789853272538,0.546184017701332,
 0.6555014557384103,0.5506784482266852,0.42628391287457124,0.7179253291274796,
 0.3885246789457618,0.4277110762063654,0.8762664236593029,0.24537990438833512,
 0.40677466137729046,0.711009006778461,0.6993617717076183,0.9776887795566916,
 0.8694498883559822,0.014598132035130718,0.8817207494431875,
 0.9869788014604098,0.35437100809171096,0.3304656304712632,0.2167173650401537,
 0.730169558339655,0.9477037973388046,0.8836802644975267,0.7441872886976629,
 0.09682663862368235,0.8797608866600803,0.37595533642946155,
 0.1958613759271528,0.9100051364559834,0.5839719374713079,0.5500951950091593,
 0.8391361699651066,0.9716940517948955,0.19307423378421118,0.8911079516543313,
 0.5300022790353156,0.18430977642074253,0.857638775256525,0.17718458716060803,
 0.6823807110501443,0.7207822245906946,0.5514316570121639,0.928644791666605,
 0.7164562878221997,0.6322740770943478,0.3692164960784169,0.9902208551826557,
 0.7083819146020922,0.906788309070271,0.008874380962478834,0.2841884171414468,
 0.9818332517126256,0.14754296570860426,0.9385769027949589,
 0.013585827246692328,0.6901637973748934,0.24424220104180416,
 0.3734799550893644,0.16621260809383287,0.11709072879353533,
 0.5912495127334654,0.7429156910721808,0.4060681543330178,0.8692616409525254,
 0.3554043090461577,0.3778706201987744,0.9133168524310318,0.8101055072445738,
 0.14905582441006282,0.8600816033802885,0.7512417549655604,0.8819534222577459,
 0.9396054869924584,0.6898129395456838,0.21578486982785539,0.9256125042430001,
 0.7751131315221867,0.0036746712447389696,0.9872109974945894,
 0.5818373072381918,0.16185155446911614,0.9194513826420788,0.6234461391098591,
 0.710305125347117,0.39747185577583644,0.38364598371558867,
 0.21095954858838717,0.493379552013373,0.7354644486675425,0.08414931954068838,
 0.5705426182405431,0.8330405834794741,0.32629232022882526,0.9285966888530675,
 0.009927347955930865,0.15432027692252004,0.7271481658192752,
 0.9100395864669281,0.06165038147563717,0.7945763909893921,0.9134009117553983,
 0.2550721713239257,0.7687595677885963,0.39071952057753445,0.2369583471217952,
 0.069438030669561,0.5137125597594394,0.42763376708459633,0.6180353659456679,
 0.5720132924935087,0.5666855510942908,0.5380368314772723,0.5670762498586931,
 0.6023986564950917,0.7065368725616266,0.3350284485452547,0.8317200493332494,
 0.7624857728175889,0.3099107989306459,0.21144020702934618,
 0.23852729350147517,0.4296434289364206,0.7845049995370231,0.1623322939536911,
 0.10327355961626183,0.36496828910534507,0.0946954059539531,
 0.2844767009157818,0.5267786855259501,0.9414640631306991,0.3986099431073391,
 0.5462206786144426,0.8570789026578697,0.7248155933132066,0.8353917447971411,
 0.7247184116330604,0.7048622853438398,0.7582887504542208,0.11327639990758531,
 0.7183396378533177,0.2879890595177552,0.9081470931605262,0.08684281453006715,
 0.76370529392317,0.8172985194123261,0.7552393892653153,0.01603207983780408,
 0.9653268539789212,0.0040438039449191,0.029159175460425457,
 0.5218235437123897,0.22965045760897362,0.695520055149926,0.6670284036036558,
 0.5288572613264666,0.07071357217161434,0.9671731354439762,0.7928324698508734,
 0.0811255930017486,0.7563681470009348,0.14248576635767385,0.6055454356778466,
 0.9597970530312052,0.010867875417729489,0.03243249928232683,
 0.026596145169746177,0.8370276493220824,0.5243682113556581,
 0.40742946494170784,0.24702781236436744,0.4928245193113687,
 0.18684489000457694,0.10622998835939945,0.9340363300286548,
 0.6301988103359858,0.26588457486917494,0.7689690105297907,0.4638873479698433,
 0.5707784737216348,0.520429472848435,0.14884183934065043,0.8376873511611127,
 0.9894634507748087,0.5403329273651548,0.1772040151246449,0.9304911574777238,
 0.633657772724472,0.9894251171875551,0.08935844764604339,0.09626841673321662,
 0.8999296195269714,0.8984586610279173,0.6400049634803231,0.9716531300567934,
 0.3503141069085717,0.031746749508792194,0.5382971090350581,
 0.40507932017784365,0.17475597556358935,0.574824784655927,0.6551474410015812,
 0.8127399848051933,0.6786392084857193,0.7047827705566647,0.3181034633247045,
 0.6755179990084632,0.02101846543428465,0.8791478988992709,
 0.41121294886097504,0.20490647007864604,0.46064736764186454,
 0.7605650886712402,0.16579621231258745,0.18882613333950615,
 0.2544280824280839,0.010639510402229324,0.9105902554306069,
 0.9106591356359327,0.48269992578025267,0.004832669787752986,
 0.3989463631532697,0.32206479178121916,0.16715256054230165,
 0.2002326454377139,0.0384136235771676,0.4130008914204434,0.3128813331604342,
 0.0148157173509611,0.7342918719722756,0.4890323148542157,0.03499695137201653,
 0.5853492555154175,0.5736949500901133,0.8948337206848489,0.23553133524104142,
 0.5774741176063196,0.3426787142191037,0.8206522737216101,0.652463774714239,
 0.2439613919230359,0.9322552444629244,0.03286351652429931,0.5115371902821092,
 0.4885365469180778,0.05447201755099251,0.8381220280067174,
 0.17165548213840476,0.07419792788916513,0.09673624081254462,
 0.45350353611748995,0.3981984330192576,0.4946697177019642,0.9002142535863076,
 0.8548757908303475,0.05680380024879539,0.13831079214614572,
 0.14478446563409564,0.557935464027919,0.8356260397359869,0.9484393991589817,
 0.7717576522940371,0.03231309764441803,0.7580821104676725,0.3220125320482581,
 0.9203170118871522,0.8127454868872019,0.374986710087741,0.953652967333314,
 0.43673733762763944,0.46074820921857373,0.46161754806960387,
 0.4844533915367294,0.35314735229992444,0.7383492363119146,0.5623921941992787,
 0.04501167060271971,0.3035723135212318,0.8795324430395494,0.4700399409584223,
 0.4173214936619394,0.5189835973093044,0.03231851671478614,
 0.33143625050086767,0.9122229300247514,0.26710177265534607,
 0.9742723504511966,0.15244454286428966,0.8860510510782379,0.908971000669439,
 0.758484008516673,0.7263954129268971,0.2479550915165576,0.16754845636335458,
 0.32926354956347414,0.5434226187730388,0.3870891594489798,0.255962902685298,
 0.3635737782101691,0.9642530274025548,0.3286736040731303,0.9365459563685326,
 0.6172251319677559,0.14989484547013054,0.02301311654077909,
 0.36739019377352744,0.7828685000511719,0.6753463874026779,0.721675369267901,
 0.5332000035032054,0.10028134661306665,0.9048908592849864,0.8322588920437594,
 0.5040063254871491,0.21155674931561974,0.26108750793240654,
 0.9404963583810577,0.501547618926811,0.5437692322792752,0.045163016057859884,
 0.33961170547621045,0.9491043920312701,0.36471338226259875,
 0.7799079827883171,0.5181860125951938,0.14371733468480807,0.2655766445146659,
 0.16157083209615763,0.19138982068930843,0.02073573857159927,
 0.9819614890684835,0.9825023626885377,0.7152578013359764,0.40677469027561663,
 0.23890372241033542,0.346485705477555,0.35324902389153734,0.2362747670274975,
 0.8837220117940678,0.7226335518861948,0.08928805080901014,0.8260307634060411,
 0.5431252411580489,0.2160801097924664,0.283590212186854,0.22472255606243685,
 0.7376040423449011,0.32180790009359994,0.1735722129498113,0.4878087724833813,
 0.9003441912052613,0.6718719396207906,0.41425609018771836,
 0.032317252734117896,0.9166112425266619,0.10341254248553011,
 0.04363152139929816,0.5924844391829083,0.7113463245635467,0.8537725300880844,
 0.18567983634421004,0.9314676667456079,0.9152031560082674,0.7322228551227986,
 0.3184491029783253,0.8405891443378243,0.9284283402752529,
 0.004804808247585779,0.7862863572253894,0.6571574005196789,
 0.06702232881599002,0.08442644737215022,0.92486262096749,0.548880148189745,
 0.9033475332099308,0.4393734049810174,0.0656163569541176,0.7540300393959147,
 0.40228175731984483,0.40129373648226907,0.521606218015808,0.7286824149916117,
 0.4274056832217481,0.6188431049793621,0.9650260303599942,0.20079995512366045,
 0.22367711439756155,0.44208751529016044,0.09031194978078094,
 0.3029537060041533,0.01792667689339844,0.19414857945251474,
 0.4293581494371236,0.23719396452960329,0.20843539881157014,
 0.5870132968800716,0.5328777746265828,0.3052526255232183,0.24411493922151428,
 0.8852804814110273,0.5772699461490794,0.2111389869721193,
 0.0030461349985846375,0.24617587209951586,0.5561541239373766,
 0.33074870251059296,0.49556859433954137,0.5102865556131535,
 0.23999905925032627,0.24959091728721927,0.2304834412617912,
 0.3465724647854601,0.6669376395193376,0.23944000148335198,0.2518403842267992,
 0.4661920217706378,0.6457710810849433,0.37794859188175,0.9527963319023872,
 0.3685038770025445,0.9304225883822715,0.20741127795966208,0.7086265183188288,
 0.9235242651029788,0.017660540479812692,0.7530595234163626,
 0.2246538806870213,0.8485382994544096,0.9348456844838086,0.6273276211491206,
 0.8278462949573615,0.03551234871397746,0.8460677732508908,0.5960392289220223,
 0.6863730812496459,0.31611652013639246,0.10277162093391623,
 0.7265843641481762,0.16946828152064364,0.9201585845516993,0.8984185740020996,
 0.5064704822490076,0.4660778588782991,0.6035588811662811,0.17415942765738668,
 0.9118895766691977,0.9403456109234247,0.42013218217639237,0.3017180571610447,
 0.7830378026321628,0.3612481697899208,0.49780157501680866,
 0.05746139605513234,0.7592650132485163,0.1477224203601809,0.5136170344255386,
 0.11085685328978556,0.6898137192472644,0.8691389959462057,0.8124299841520591,
 0.3274043586848926,0.9305892525921029,0.14365034248165753,0.856417376342036,
 0.9790342612029642,0.6428117718152976,0.8426173675937061,0.916691942480845,
 0.4373359343755834,0.10989807151830444,0.4197381331411325,
 0.07282584124621316,0.9339206761014707,0.7242981350743931,0.8932638168448632,
 0.9490465699052315,0.6914495744983983,0.9016945602789931,
 0.023778070170261456,0.46547272946754503,0.4105985738693827,
 0.08216052632103799,0.9337519529939561,0.41710997208245315,
 0.11116593384096596,0.2388009820970597,0.5433844479860458,0.8874841692140705,
 0.1288377564016,0.2732734091140805,0.8948517917313952,0.6795657029670642,
 0.33514000465610005,0.3162837037490991,0.6487429140375514,0.5185567812932348,
 0.8125573872795411,0.9959576409918657,0.5386737942847333,0.21121952894882168,
 0.853407592351692,0.7613051776999047,0.01646938540641818,0.08719564511947286,
 0.4082498247207148,0.8785851975021622,0.4223682282150396,0.19171518851731217,
 0.5409230154080678,0.2019370167232798,0.18385655920328048,0.8453259483041823,
 0.47298184784293484,0.01268891054729493,0.142447378470421,0.3155465988538648,
 0.8329140215304122,0.7115364199368279,0.5219096764424761,0.3169144602732399,
 0.2965799790954742,0.9793021527371762,0.3287657786927288,0.48917888056259373,
 0.2458539157055346,0.11827170613216276,0.06420876751684368,
 0.5218811910102593,0.301597515221699,0.771075320154354,0.8468602127775415,
 0.9414558280689256,0.13340300497982882,0.6584748984691489,
 0.49350853421311636,0.3704245695370828,0.1536142355534259,0.3860892456119943,
 0.11281757171249485,0.6191396049387137,0.11031650031362994,
 0.4037832457636342,0.8415217806386719,0.8543410287175802,0.580108447481767,
 0.2221259631423621,0.22998306249633305,0.20696033293517824,
 0.5055041186527561,0.1904753982413082,0.2743607911687067,0.7952787064939681,
 0.427056100423835,0.9477253538682289,0.2659820889390794,0.4304337577442171,
 0.1272510334525977,0.8220468541868252,0.028109480050590996,
 0.015176658960656009,0.7565813500891623,0.6393971236134555,
 0.06858310749260521,0.05301024635638718,0.36596975928289754,
 0.16218401800412652,0.33623712425483643,0.42585080016413335,
 0.17565853363048423,0.9199924532897128,0.17132445718589717,
 0.8483296264334214,0.6358528247807668,0.6599127405023849,0.7520986953557741,
 0.5866863184451356,0.9598584233869943,0.9422850884813233,0.4506839148824199,
 0.14885999285614537,0.40770650192146074,0.7324285227631271,
 0.9411353408731992,0.4112719936324323,0.1846326006384056,
 0.011592375186744053,0.15434820279603434,0.21643130917111353,
 0.8216072072926265,0.04636588685730125,0.9244167460778318,0.5660299231546209,
 0.13946187965184142,0.4053361279148042,0.09682656651457688,
 0.26917479711815595,0.3889454581493377,0.5970734531052612,0.3478931404796298,
 0.32774482009573713,0.21192411914133324,0.2944036310466476,
 0.5215859889816102,0.14687884910009286,0.22307320876741776,
 0.8720848609768285,0.5378488625457198,0.7579066319161151,0.5555999110285956,
 0.40440870844273036,0.4973122281844571,0.06645672174164496,
 0.5464520772857009,0.0834846990226541,0.7118140967157409,0.8793875686108749,
 0.4796953217805071,0.5941077060027886,0.6635941669434344,0.06662594116795373,
 0.9158614258068578,0.007932526356079328,0.692889754408264,0.7418452415672928,
 0.09419930436336799,0.7490614057129946,0.5746297658116114,
 0.049610520213787423,0.9827514272708288,0.3834750490435408,
 0.21124262853414066,0.993412490891411,0.9787663179426522,0.7109099047977815,
 0.47636433217345975,0.5775760771390483,0.45576358101238523,
 0.7552118563442446,0.15494402470794566,0.014264545622003189,
 0.30290297485222006,0.20389384614355777,0.5246797316377326,
 0.0012336378051767127,0.5259146643374356,0.5147382500355201,
 0.9910952925441724,0.10011219040912445,0.5597625496395263,
 0.17578030640803033,0.43971034376817575,0.18364346598698755,
 0.4907703866351144,0.5135406446614954,0.2424370167108093,0.0770254383303196,
 0.8410205118373097,0.2478369335641748,0.6949988297249454,0.759742391430735,
 0.5191907534158064,0.3878777057361338,0.5161341052182009,0.01676669969244271,
 0.21361251754114075,0.44133962472579447,0.8202424862564732,
 0.2990769180218763,0.3344437445657539,0.912962502517301,0.10941385775387502,
 0.9359100749194955,0.6682288161296802,0.09004972560090685,0.6204079652659125,
 0.888967168867242,0.13454995688340732,0.0014810246383000214,
 0.2678771415526421,0.483087090231738,0.4408789717403939,0.4428933528019936,
 0.2978385803342767,0.11133403204429715,0.1824460972336216,0.8038916949397746,
 0.02036330157700017,0.1796810575660247,0.32199600781279414,
 0.8754980047415326,0.010627860370851505,0.2950691690225138,
 0.17162303005412083,0.8655778023145821,0.0878591435855196,
 0.18740605830850132,0.9478953931273704,0.8310773926436572,0.5417671968580682,
 0.36478624601897214,0.30307794203386007,0.5192333018322053,
 0.2835948194776098,0.15253304442713989,0.3844097658427533,0.523798575428258,
 0.31575362226463477,0.05164076223904113,0.30539436866532643,
 0.09341771987506187,0.8312431841350606,0.35570725072751475,
 0.3291668158459191,0.12776534494723968,0.3840743505903853,
 0.45678882759249484,0.046248851026804316,0.8256990256764523,
 0.9678674419559231,0.9537216155602487,0.4366926522791007,0.3216421539221289,
 0.2389354464913861,0.5795313263027775,0.9363887078932258,0.6333069443815367,
 0.36651336086138386,0.7900329047961069,0.18696497762075182,
 0.7599911854271393,0.02685821813666711,0.6622221740916087,0.9740299696447732,
 0.3353401486879003,0.6146032435339667,0.9092816442185403,0.6649804430080146,
 0.25003249284928486,0.05027285489325806,0.7050893683862796,
 0.9352283358162546,0.981255366018667,0.45989001619595626,0.8389358592021513,
 0.9505070327982724,0.3814250215230808,0.11368342909170437,
 0.24132059929941885,0.4524873881389906,0.36521261800808835,
 0.9091262922505237,0.15245508695799537,0.9019683584997228,0.6215575874287368,
 0.951904019994981,0.5971313609035205,0.942168909308451,0.2097433428616725,
 0.8375708815994846,0.507594393529003,0.9334728049091313,0.3809874103909443,
 0.948799364867377,0.7294735657283752,0.31684043806220974,0.5124350692506091,
 0.642373678594891,0.40261912059197136,0.3417512453888264,0.8108233474396542,
 0.6314937019335749,0.19771857092582368,0.9375841914031986,0.8618548080123357,
 0.3452137052305835,0.06379734099445333,0.6109415131074167,0.9654479247239047,
 0.3803239080237192,0.873229153083456,0.9766953949948999,0.8862383757635302,
 0.5374971413081677,0.02440934130707395,0.4350331186964079,0.9970512015468821,
 0.8236033535167128,0.21760962627350122,0.9641218163670565,
 0.046636415828436584,0.5092386620959908,0.4945228401070251,
 0.05185911897397344,0.7302183538383981,0.22725079319229757,
 0.4149463085417803,0.7636804226124562,0.10859101395505844,0.8881028162021658,
 0.23540031719444743,0.1634167277553985,0.4596252075615428,
 0.15781991429096787,0.2090583718109299,0.7761251239855588,0.8369367265275008,
 0.22928266380866347,0.9891503334205902,0.3064522587853279,0.7071990532136412,
 0.7200300955959156,0.6167168039314347,0.048033484503691604,0.857420449889188,
 0.258409032778815,0.8519901407843157,0.5928790110059653,0.22467662151884915,
 0.4941283062755021,0.0781030570052087,0.2943652378220948,0.853578446757907,
 0.8383329392157612,0.9950589656660838,0.06930556750004147,0.5706300749854122,
 0.4051519600307689,0.8709032428561333,0.05653798220231587,0.7042718101517935,
 0.10890705821671753,0.6851224795266102,0.09557617599476187,
 0.07080686590278695,0.8034718704056896,0.9821090557325212,0.1964605032029184,
 0.6796103385995284,0.5352797300000115,0.7538049640529003,0.07394353700656864,
 0.03908828495697758,0.9272958850503742,0.41323956184706623,
 0.9582520441869471,0.627987129426847,0.4140979594055022,0.7036962684306054,
 0.5231001676658066,0.20060034955991746,0.48595966622802567,
 0.5856795706415272,0.7411210132801617,0.11493847039646365,0.788996046012501,
 0.3407832970605682,0.8364494866804274,0.6044560500551129,0.19009766340503265,
 0.5139014998779278,0.661986652757102,0.42752668529002735,0.675900818749164,
 0.5159972037073608,0.1488833768455895,0.02321997283876276,0.503437749491289,
 0.45511610899112287,0.516131843639605,0.3260799627681654,0.24992907743670145,
 0.8987147857942805,0.0556394209468275,0.19494991245909576,
 0.09860619416671335,0.1560156977580497,0.777831007325467,0.3841565195261978,
 0.2611494571189876,0.9924172570486526,0.6466618038770665,0.6991604267084963,
 0.012858995085042069,0.13904265482423195,0.35159047593987536,
 0.1182759974963068,0.05097267989679599,0.3294169983407029,0.7901249168619521,
 0.858741852931701,0.32816823771634274,0.22266255856880168,0.1650315119013085,
 0.6590115309303561,0.5905253005743308,0.057308252625273726,
 0.3753436441956618,0.4560727079733863,0.7412197098705006,0.0647049749321883,
 0.24317293995245404,0.9080733433276449,0.9496814678152609,
 0.22006974592165385,0.9796886141015403,0.8678289242003306,0.8251704657738942,
 0.5823887957356327,0.2666014649971715,0.4899305979109111,0.3104115898765665,
 0.10626239165926155,0.9294170981197476,0.9466571046479969,0.3619245816493879,
 0.6424230349295598,0.7990124859970555,0.7446359511726404,0.7146105648164807,
 0.6497838223933236,0.7348562357953901,0.48775054386105243,0.8024490813533143,
 0.03207428439309956,0.333446859712221,0.9079109043810205,0.34556900380844435,
 0.83334093545245,0.051675263473117905,0.07975095166764667,0.4481203275188693,
 0.7750498542107882,0.4833577650159191,0.005612523587075535,
 0.3130421278903188,0.5273060527845896,0.158165431693439,0.3986756226525231,
 0.6369615642036017,0.9391872368475926,0.5140490935941909,0.7829527853195426,
 0.06940120480608258,0.4862642812325332,0.6658253622157865,0.6738938542581451,
 0.04649615170280019,0.05625680843558134,0.5894034157379401,
 0.3200694131422316,0.5824009222609563,0.0041420157527459756,
 0.5233428005647212,0.37481899789526685,0.8788800396419923,
 0.47991555639468886,0.5337337727868866,0.9310318940655296,0.6931204877605738,
 0.06715918124341569,0.5121287838224892,0.024728019597418882,
 0.7928028904045921,0.056009901636202564,0.8243075551277275,
 0.1488229628836134,0.7462067385917867,0.9231631454474305,0.5842597768329776,
 0.8929306704822424,0.28408263325791705,0.9159169935057161,0.3981633906209764,
 0.9758174453522394,0.7866055233962346,0.6566381654550925,0.5626984104366878,
 0.8996547696530109,0.5847395766490393,0.7946519661884711,0.1056990356815598,
 0.6331371214922518,0.832393910593943,0.9821943279492531,0.15839674515420998,
 0.6670138696698422,0.3592265350920545,0.07934350980071403,
 0.38360852046900185,0.1465883611863058,0.2390440781599108,
 0.37323941733120214,0.6757788294028019,0.6543821099985629,
 0.07528971993656486,0.3639796721552021,0.41237703620403077,
 0.46786633748029616,0.6164100728334314,0.4479370100638511,0.5387751098101309,
 0.14166943677405697,0.24149579849494052,0.4183906848781166,
 0.5884998019178388,0.22335306741357275,0.9237480961459235,0.6511927151446947,
 0.45983645983301735,0.8284505570966976,0.8289890128803329,0.3734554635424244,
 0.9063426907979739,0.4991144601056745,0.9937159446211848,0.24769715036152895,
 0.08247957617258428,0.5906755040072575,0.2680304196974054,
 0.43768049717892854,0.013310357136590012,0.37420966992815863,
 0.7563590429282212,0.6018672743734614,0.5996496139226704,0.838349304695737,
 0.07442133725102307,0.2960697480905139,0.09845506505174895,
 0.5956767838098701,0.762921396662619,0.21849742001943784,0.8678288384347235,
 0.6989663655602307,0.9381856768428524,0.18367684147215557,
 0.24182609889604456,0.2679731036627978,0.7588803659361897,
 0.15678093399896786,0.39406450431243967,0.3806497461203625,0.922183916814611,
 0.12161802123857368,0.5396265948700026,0.3982844883678943,0.9212980415386032,
 0.4118677913974055,0.8183571662665274,0.1387906832597826,0.7910892102685433,
 0.027913514070433965,0.9278576744125736,0.8270944505547364,
 0.8395140405488355,0.15493542277768002,0.3675852562446401,
 0.11817428735767466,0.8408257815459221,0.614384587639957,0.7997013349756981,
 0.5243680358434334,0.8796281035619002,0.6003141445943454,0.3548444844358236,
 0.3189082952744331,0.08585835107991024,0.5124029639711749,0.696855125553828,
 0.04635915994622253,0.16078148535932757,0.6469639846230439,
 0.23306708858461467,0.0419806494640218,0.6709807320111357,0.5932012455553668,
 0.4101060436287237,0.20321353371796114,0.250899042929431,0.6960662043322668,
 0.6613465275450066,0.5053725706499496,0.6858091065621924,0.08828401853180545,
 0.3122107042571942,0.9833844927442053,0.7712921661160976,0.9909973864626954,
 0.5654549578533952,0.7964414014294771,0.06489584772638168,
 0.06511246132956228,0.582551980146662,0.6122698321977454,0.46662241325202647,
 0.951683019215409,0.14080365855477395,0.7964156949207815,0.7104898201361277,
 0.7026706249220944,0.5649884846705588,0.5721791743848247,0.5567397018822827,
 0.5211897944306146,0.01779077197104839,0.9541696198991696,
 0.039427212816221635,0.13699275706589753,0.5477123365959504,
 0.5195549825683865,0.5279667179643849,0.18192996053752752,
 0.014276568658569966,0.05129041273161106,0.5642294349462598,
 0.8816551521991345,0.5950082759564468,0.6451770617239762,0.8485835298337914,
 0.6280211659238925,0.3752117418543137,0.2691203163466219,0.8047135158708091,
 0.8844756215184977,0.32443659998267327,0.8234754398485127,0.9309343183420595,
 0.2302713901246849,0.06445806842990254,0.5298862297115442,
 0.24334628055941976,0.6183394481693288,0.4676383752699821,0.3625730806644897,
 0.06740700286858514,0.8769224399319856,0.30370180695049764,
 0.8365998338097396,0.9047396934751811,0.5662918208427976,0.22205354175302394,
 0.6586484676683094,0.9068428380964482,0.9389454114262215,0.9492843033476186,
 0.508464631997832,0.26696878951596137,0.3277784032292559,0.8664550060966059,
 0.7808737888472661,0.7265235315654305,0.6257102666816178,
 0.059965890567409974,0.7208796478879601,0.644928796007601,0.1524752486604274,
 0.24287727912012402,0.9116231755775461,0.3289492238122055,0.5668413211956946,
 0.9446915069846765,0.5200418958611462,0.5632216178706271,0.2886536467087202,
 0.5464400410742605,0.5782916235483705,0.09443812267495866,0.5979556748195125,
 0.5904720635428622,0.4931373228049414,0.9464185581303919,0.5458174469372014,
 0.7730313810578846,0.9905422030824412,0.7838285605301486,0.6298400701276681,
 0.8405909582279761,0.7601505199217847,0.4420716683451851,
 0.021162181052213347,0.5580017122315428,0.7532391482064824,
 0.2923734298737578,0.7912100746644926,0.22547998286062487,0.6805957914182104,
 0.45710434502479447,0.7194754261038071,0.21364738890478008,
 0.5795919348537457,0.30838284194539034,0.1985530583439471,0.1782028762506318,
 0.1790755970138207,0.5981239038528012,0.7134914892261679,0.7338848075037779,
 0.13467264381121802,0.16883857698259208,0.362599399136453,0.7671662989542215,
 0.26541662848017367,0.01460175374901107,0.700951049844224,0.9321166983423205,
 0.36721376183740584,0.6848983276298025,0.3980279674716448,
 0.27015537189499295,0.9074702133461869,0.21097684846758746,
 0.040992353070668286,0.43744708105391794,0.6121392062466873,
 0.19415134390807243,0.39568953758350955,0.8728541244677716,
 0.8286460339250615,0.254705312186715,0.24168740869922156,0.13527302786872852,
 0.10754342119829408,0.5828633754863445,0.49666722858595824,
 0.9050288691321862,0.4903394571942411,0.14569217470137952,0.2097828699230787,
 0.08257996768496545,0.5499260380576171,0.08246331217975156,
 0.5175112782123332,0.882802038438628,0.6558899060222995,0.3319628522523834,
 0.09991578534294532,0.5563767371430146,0.31267297420737217,
 0.2727600583360409,0.0584550478142859,0.7422488130804916,0.8611422424187566,
 0.26385357048388525,0.7091041897852894,0.7157592293320815,0.6605641489813622,
 0.6845677104759296,0.9829972872303435,0.07973391825245602,0.8701854147188864,
 0.6259334934444614,0.0077678665656203405,0.28096678761370497,
 0.7178854917040128,0.6523430091273772,0.8118234131888876,0.06187891211838448,
 0.29513038394325397,0.9390299841047757,0.36981383390826783,
 0.43310111717719935,0.37064594573375975,0.03614207957770066,
 0.2220926843370027,0.8938901041194968,0.07441264788164348,0.4932873885546165,
 0.26589040408705067,0.9674232276232216,0.34896871034680266,
 0.7714872766901122,0.9307625950110706,0.8856476546392192,0.8924663975499532,
 0.5181138866321829,0.5026489759447548,0.9991205702124588,0.8313386211770076,
 0.7275052644790337,0.17280463619239195,0.37717301507088175,
 0.11511021461562587,0.021162712530291117,0.806823910534643,
 0.3404858250587298,0.9466387196151027,0.9653548149430016,0.5550665455289492,
 0.19343030644404224,0.18666319023108857,0.034338195999611854,
 0.6159094158265283,0.5505296764194751,0.13994837267815485,0.5109083138904157,
 0.3286911802955331,0.6758245742012428,0.08773813073160008,0.9499405446349188,
 0.2722099857599506,0.6697208008223351,0.19790461099892365,
 0.31557537622863363,0.9678194615974494,0.1254581124091544,0.130345095738989,
 0.10146802819948741,0.2947076202213733,0.9987239552654472,0.4788954959205749,
 0.6689840501590044,0.4671727800893305,0.3062539684275557,0.9335403923045726,
 0.4546836471684198,0.21198267853085906,0.6753068431765189,0.4965083035597908,
 0.5366698018902866,0.8302616905291923,0.01602738069039744,
 0.04909509300580217,0.607523824074972,0.5781885191666253,0.9114744939828184,
 0.5637734809409208,0.3354093493523356,0.2382727726053323,0.5721866776665253,
 0.6968166137737892,0.6450944768962631,0.9409486103687502,0.8108481417235698,
 0.783549939378863,0.5108927835120355,0.4260374181204618,0.5189104612394508,
 0.6828791310376547,0.692526133047113,0.6145069218925461,0.2346432465608712,
 0.7638116461038437,0.8603772550881653,0.3730365308777539,0.3877916928998789,
 0.9960018484960461,0.08037381119869513,0.900376156802094,0.867544647227125,
 0.7693332053316918,0.5633350544353857,0.7572611577080173,0.0743837176961607,
 0.5344528181964726,0.02904293485621423,0.3137066658788734,0.8596711047422276,
 0.9980056848170538,0.7437979512218006,0.1307992858094169,0.0387574086893272,
 0.43217553124106434,0.38761912090811257,0.9892185529756057,
 0.5705514372968784,0.625271088684529,0.4510729208160966,0.2457052869364167,
 0.408876885315107,0.8356574525983907,0.9742960646184405,0.7888450739411712,
 0.9223252641688149,0.7217944244270721,0.8674047958155477,0.6307532015753285,
 0.2945522837013991,0.9746768082541499,0.24491578390155078,0.6482442538256817,
 0.9796755803198128,0.09678455562167221,0.8315240445268732,0.7233713348545397,
 0.9690404433107677,0.20848641860817152,0.6639937908458926,
 0.28469768744590773,0.43828864981149895,0.38595316144504566,
 0.7106500605883799,0.05407856813388945,0.1608889971521391,0.9670843414780931,
 0.16098206606431376,0.0462458474881734,0.46288666291768255,
 0.5912121939329882,0.34244226250122933,0.5237279966648984,
 0.47256021698998196,0.5743822085335056,0.5429239033015356,
 0.01899619789376661,0.4573599259529524,0.4316006177486311,0.1527348483836687,
 0.3126335075641181,0.5885561418603829,0.5107937174097344,0.5176446825097124,
 0.7121041633709119,0.663349569257701,0.19134953300427082,0.8024926332148482,
 0.02564922811004111,0.9338575284964383,0.061987542075194635,
 0.2431720702587703,0.35062444296568507,0.826028984080708,0.6519470592814988,
 0.04269816810183158,0.5489887562552042,0.28099423232346776,0.670954265096296,
 0.23220239964822764,0.9226485902003638,0.846637630724109,0.8303359575565866,
 0.2321118308534229,0.7156105141825755,0.10876197294293521,
 0.29276046483857343,0.17156010774278063,0.8960239689757173,
 0.31423136922423045,0.38463361139509344,0.12272577439616639,
 0.4445535916123571,0.7524540728215467,0.4305000165792662,0.7675788946428639,
 0.17092652976174216,0.6550922221327411,0.006342910886789843,
 0.21517956676775019,0.5373174429543147,0.7538630737511498,0.369760581337115,
 0.4406768442104436,0.45970735693388565,0.22037997079678595,
 0.23630551619644513,0.7988041703021062,0.2385043485161451,0.7371781785802,
 0.962749510421509,0.9430027583707733,0.7254445705335424,0.3118706913292061,
 0.09966043121460666,0.3669107417488794,0.7787385976735418,
 0.029427095779764656,0.21962385305127863,0.4070163289085733,
 0.7663147029795115,0.8082790724287562,0.6434372873805521,
 0.005464881203226524,0.9420448593702363,4.202490806920478e-4,
 0.8154441902888916,0.6920218504913521,0.056846334692637,0.06282092849122856,
 0.08194152057162318,0.21880978663586037,0.5611954761921051,0.79603857519288,
 0.7726341195364888,0.5445821191337158,0.929381681828678,0.037244301272991165,
 0.56564141077805,0.10879165283624437,0.1777318377419802,0.562695051301195,
 0.9414779130174913,0.9998964559736694,0.9362142428046915,0.6217146084892197,
 0.6127757588005592,0.8649775296435638,0.5110139573268154,0.21352992160489648,
 0.8112865121422433,0.9233998312849554,0.8166777903735173,0.7918101702489577,
 0.10792714954974048,0.13829586784027703,0.6218568571916552,
 0.3788804444323036,0.0016817555212017554,0.8963617507226728,
 0.0692911623228355,0.707145754421864,0.18201873129668988,0.1043034965992331,
 0.07969014044327538,0.9560377496102604,0.3839561720355047,
 0.27226218039928685,0.8251843361471164,0.9090664113533505,
 0.030926226111106958,0.46298322259302216,0.8448169747058212,
 0.07530379434479373,0.6199730585364126,0.06045884099733834,0.891898970505027,
 0.8010923140966292,0.4686389620444822,0.5353344574329225,0.34584591600593295,
 0.42116094882065824,0.13800446714886894,0.8104437636195769,
 0.29832087205187363,0.27882464349817426,0.12266907012446948,
 0.127489085594515,0.6648306654747375,0.18791321432625896,0.6322682385538134,
 0.014118786756556423,0.552210382682388,0.4760211922656581,
 0.15056959246646817,0.1821015346215128,0.24249682082633028,
 0.2547330179285493,0.7975569130905054,0.9497766998185673,0.9167982190692361,
 0.7341102960876683,0.518127208770006,0.4959785115569568,0.2903495579916421,
 0.013545082883966386,0.5529246909284072,0.5954685094051193,
 0.8137156557697782,0.4203974560011061,0.04003596848547353,0.5113743775012061,
 0.6554951067382034,0.8191912222938523,0.6680340191771394,0.803046859121589,
 0.43875397753156586,0.8201279639400699,0.7750265477465084,
 0.17113905709296462,0.4061596059228988,0.22530574576824602,
 0.3118522952778906,0.7609225495410356,0.3347948036255135,0.04467443039227692,
 0.43433858524214874,0.4884147889873778,0.7751612993661025,0.1296480434167957,
 0.07728305362442811,0.10489395223580067,0.14264305002556021,
 0.8309166268856567,0.2845143239357033,0.31097883193285847,0.7099956282153496,
 0.8104686950025621,0.7673277796019553,0.643670122236049,0.19844072314450945,
 0.22397628453335494,0.42581278261602573,0.26108352421450143,0.25106426202445,
 0.9085775161167179,0.27208056483192444,0.6154982411156099,0.6345326337974874,
 0.05931739338483799,0.2929734495137215,0.7882857781229957,0.2930517363117824,
 0.6967462094066917,0.4803667039991709,0.4051106877771419,0.9221226198045948,
 0.6444109183028711,0.20069554105635523,0.27587693176701344,
 0.6817567277434601,0.6357066611549116,0.6364710410406581,0.6040169826492341,
 0.7878244104760468,0.8357913344016552,0.9088814244726879,0.8396766467282335,
 0.3895966858524106,0.9117940241278205,0.6718328862589985,0.4244566443661466,
 0.29754701721905075,0.2139026931800765,0.6137310302100288,0.5250760870245549,
 0.5041092325099973,0.31805202509662667,0.7590813069061757,0.6036370603347226,
 0.7172351313807647,0.5590074448570799,0.5104736491943611,0.5792316399060857,
 0.7532820109330559,0.22143042992575312,0.6500688023438779,0.6777456900083426,
 0.7796847751755832,0.9382397472421881,0.8910015415459189,0.29540343111904444,
 0.5510429013257054,0.708931792059936,0.8661376962179036,0.6254260585890474,
 0.19245400621966935,0.39329826932295986,0.5683920165612543,
 0.29865037646031145,0.2688437364628615,0.7712503717815693,
 0.37053135002448956,0.003474415763298211,0.23041243278958723,
 0.4523264795823919,0.8305124993495491,0.5089277801243124,0.12296223355577318,
 0.14201072236635015,0.5344105931936803,0.42323710046070684,
 0.37490131896820655,0.9095279763582871,0.2746693866287774,0.8235542556741923,
 0.15181546487077857,0.9297984141905995,0.5560770379396456,
 0.05355133195929196,0.05589561383820929,0.3449814490213634,
 0.8089468255915693,0.5249355930943622,0.046383296864921375,
 0.5778830873781551,0.833324759405258,0.5729540913425191,0.7301984231564738,
 0.836150873509718,0.8813611103964195,0.7003232711777441,0.6088866570735152,
 0.9800120326319303,0.0011785877505943176,0.38225077370205207,
 0.6556930646723562,0.43131168830423794,0.10176039742571086,
 0.03519150441258678,0.9619354109935909,0.8362980955998303,0.7082480152464096,
 0.7653773323073955,0.39163423638125716,0.14091744015049978,
 0.3404681290426439,0.9614273867776013,0.22398540779224363,0.3376974773235454,
 0.019772533119223334,0.7704323436985875,0.19411688075508515,
 0.9240500531374811,0.8149724546547149,0.22304463388878815,0.7253369982581249,
 0.8857268027043881,0.7491650797478562,0.07900724381350444,0.4590460718072298,
 0.7341458719006944,0.9097396091911538,0.2777541173678879,0.9521555307904992,
 0.996139520071428,0.8485816513774507,0.7121683181692376,0.25439674517088884,
 0.28670596634245415,0.7150980960551756,0.773695964830195,0.1842194146974352,
 0.5990857587082212,0.4883913517322287,0.3538482769656047,0.9743042117546115,
 0.8736109303088745,0.6559688054057325,0.25259118977006256,0.6891880246021236,
 0.8843971291112915,0.6369879931388795,0.07226367224376218,0.4601334270765153,
 0.09879881230938314,3.6335342538817983e-4,0.7214372876711095,
 0.19895153172678637,0.1101344522925316,0.14358468296370985,
 0.2057444857195404,0.037412542210025546,0.364402810901866,0.900324231617853,
 0.03779270038761351,0.7758395708870566,0.7781809692779373,0.8921183614483398,
 0.6427850197184481,0.3765414535212199,0.2589548295503581,0.5527737045077957,
 0.618946282690543,0.18820961176397266,0.16618321933908953,0.6803904864637896,
 0.34060934227718065,0.5494333852181479,0.4291978875640938,0.5967898318929369,
 0.3602701028050931,0.5129622554309188,0.2274835609013135,0.7832554656170773,
 0.28260958055582763,0.7758402190005645,0.25952183317635813,
 0.23929276026447122,0.2652827572277675,0.973852906317105,0.2021931887462659,
 0.7068177880426875,0.588709211563813,0.0644045459809034,0.6840960868381337,
 0.9313756009995797,0.9307780257367705,0.6075379914435387,0.21391147211202066,
 0.8651085679783169,0.6210131727206524,0.5092528804949402,0.6641778591873938,
 0.4958609007665058,0.6166075529638002,0.2911432838533433,0.5116408781462222,
 0.4933269154950445,0.9387902040697769,0.05754670382010574,0.3127591060678486,
 0.32110109196082026,0.2422213204643322,0.17990736728270018,
 0.006604472516891335,0.12901996957264128,0.9953877254198734,
 0.24061994763775307,0.4723531743435918,0.3264588852084185,0.9508936060093851,
 0.8236552330187796,0.3920910961103333,0.6387683954507,0.8392666182165357,
 0.4540960020553356,0.058657583326322094,0.6859659859269905,
 0.43659779590313075,0.7833641890364981,0.22793372236238085,
 0.6828027875403891,0.5582688619562313,0.2829394322256029,0.8590210461224764,
 0.3221532077941096,0.8639573490567778,0.6208545171889469,0.6727936167529592,
 0.9510587198408479,0.9248072655630819,0.19710488594590525,
 0.35559219195636627,0.6984529576323568,0.4442339847565614,0.7944513982425361,
 0.7640285378429416,0.6703765902733527,0.41345642036308516,0.9800784998992044,
 0.6005269496287547,0.41712285261533033,0.24163774168942087,
 0.18035372688823026,0.6038256103370976,0.8667381186217098,0.5348271780212441,
 0.40401380661854935,0.13104759047670267,0.4007980844831037,
 0.8372399573216927,0.11978115160970715,0.8088682713431881,0.5408168351821376,
 0.8542610006215613,0.779681609177509,0.9967033315742306,0.9074504523465923,
 0.8755121480593524,0.6666339727253621,0.7689680525950181,0.43864838006879436,
 0.8073081411083151,0.617397018952841,0.4782702013961686,0.9640347008516092,
 0.33080264847115004,0.5182922211596647,0.15342661112985767,
 0.6877149361162039,0.49357852463961915,0.4636061417261107,0.9746004233485038,
 0.7542086985097658,0.363315214877326,0.7834603516257418,0.4203399209589296,
 0.13429378861393348,0.8934990402802605,0.7437599328879075,
 0.04583114705266611,0.7486006941748946,0.6169999050110138,0.6606910205001618,
 0.5539469460189395,0.7453349063564092,0.9167075042518357,0.6484953819431335,
 0.9203260526269793,0.12518554743052024,0.9244196750815079,0.8007101938449879,
 0.9516777491797472,0.905096421567802,0.9751264630683941,0.9808001426107211,
 0.9177307291024036,0.020617434582443472,0.253962894323694,0.9808679262248285,
 0.6503196233487137,0.3249190918140322,0.5828877856996737,0.6718647190694045,
 0.04125206721543673,0.7456487946776763,0.19261244042874448,
 0.8233561640122536,0.025078512562301203,0.6717123862441394,
 0.7734468794495086,0.20791548198294796,0.07886177609354694,
 0.7254027660635647,0.2549678428039198,0.3227089658365123,0.44158227001931827,
 0.6225371693935828,0.3375071456508507,0.26770312275156094,0.7753188111871325,
 0.5280498046509502,0.9652412097997677,0.439182753705892,0.08686936324518946,
 0.6497133477220783,0.49971775086826276,0.6578068061355185,0.3664327611249275,
 0.6560910800589136,0.3185981263789507,0.3637803924216654,0.3526419217932236,
 0.45318608121642867,0.2229605279064948,0.07081970223765299,
 0.5933292000953864,0.2314508359054981,0.6499421187494714,0.8130823727447499,
 0.38928433565535736,0.16370719349284146,0.4060271866957137,
 0.2772275729233774,0.8286825184117328,0.12252990212438808,0.6264270522900207,
 0.22150580372341877,0.9294557073499329,0.9953894936745797,
 0.09569389354891078,0.6012519844699988,0.3297823523109855,
 0.19525097370132205,0.8306019272376323,0.2288078117720621,
 0.05754916284304601,0.2945865620679433,0.12628360648455428,
 0.40081766954313003,0.968115465712577,0.39948580197573347,0.9110097631864065,
 0.06118791042643379,0.22848409539402947,0.31280292143746113,
 0.9364041285038724,0.7322413484548433,0.9959521102602138,0.09537393668106964,
 0.24135679347698813,0.4891261159101662,0.4574193765048873,
 0.19012904173835998,0.3409777683915721,0.48755377653141596,
 0.3984172426290624,0.1622308441082122,0.29065075315136346,
 0.40414256358946843,0.435413937354142,0.9186335372471377,0.6454945192142152,
 0.5308380592701242,0.975586138532987,0.8377989966968652,0.3873597707407158,
 0.6154224993378823,0.13555491600264313,0.8748190882019323,0.7745205190036726,
 0.3445345924555374,0.6827901082528876,0.4041668639002093,0.7956976940882533,
 0.79578171104258,0.08779486189131602,0.45766868010146844,0.6149908745064134,
 0.7799587488146006,0.9775568730568621,0.6252460635956552,0.42300676383304536,
 0.05204471765738372,0.7988880189145868,0.9664243931158646,
 0.31822612213613466,0.27182663201510204,0.7670314509391234,0.178993859012604,
 0.4857633410764002,0.17115275665117047,0.35312283050150106,
 0.7278078067764115,0.1220701028633363,0.6422502031860271,0.1738076896460089,
 0.29464930954898405,0.8308451483619785,0.9868567318090731,0.8095191944499767,
 0.7237452146664662,0.32730571507739137,0.18005127494341244,
 0.6101684597961222,0.13689365036287993,0.11235866035635245,
 0.22956519757198046,0.13195639983529717,0.8970580700715142,
 0.1242661872305666,0.6166816536468795,0.6591320225761849,
 0.0012845626052671033,0.8219701647255753,0.5625878044078677,
 0.4469999789711473,0.041990960585756576,0.42060925462606735,
 0.10960554928481048,0.2721360275454856,0.5868499937391336,0.6298220556293426,
 0.0069920314725633315,0.9762331651695307,0.45188148388199734,
 0.28999906735105874,0.08256583147945007,0.10217678336678426,
 0.38818744006742933,0.3089661206197123,0.25137101699706477,0.507439484472777,
 0.7677473241269727,0.5230779747075623,0.5453057567748029,0.0816396371273651,
 0.8650149454302174,0.49260084075370014,0.0974152444328159,
 0.22619978979631317,0.07600925948699466,0.7763701357986994,
 0.7890833908535081,0.2988144072393695,0.9703377246521103,0.26319552536544677,
 0.44720217084101466,0.3628861006654528,0.4538986017502966,
 0.15759780605518048,0.16337750691344977,0.07947269288261893,
 0.1886522590382318,0.5043130151761541,0.049134780041073745,
 0.9174078987436791,0.03960565456694931,0.2949973797891069,0.78497205007317,
 0.7412078640254728,0.3618780566151777,0.9228322700996234,0.806677423144359,
 0.17447517078676378,0.395249222144055,0.6934695024599213,0.06139735278646741,
 0.48235545553801673,0.914664034197423,0.19771209991168504,
 0.10127271081617217,0.4444763526304035,0.21075337359038782,0.425563359533938,
 0.7606423412541915,0.39635501181883037,0.2549793385752921,0.6048673385085155,
 0.13574655328268626,0.13537060105850696,0.19227956254805822,
 0.704427140026341,0.6714769595673492,0.24542959475645199,0.34870754618690003,
 0.5794428244476983,0.5114696920200683,0.5931659844210793,0.8131805749441571,
 0.7433608163636928,0.7243205238432646,0.556175187857889,0.7106167465555954,
 0.4158853296324134,0.3035592682450402,0.5366279643060686,0.16906998301887644,
 0.9155194333677218,0.23940429057193002,0.07134623187468747,
 0.9195599464698598,0.5469270046872821,0.41781286868140854,0.5200278479401546,
 0.9375698548838725,0.5852926473550056,0.562788955043709,0.758840473081829,
 0.9066798242293249,0.8337708285279897,0.14911403772234644,0.5234200567023075,
 0.7898211037346777,0.3442162407303583,0.21944998809635508,0.1751839578838723,
 0.8552839019927507,0.03678764318404237,0.3638391067125394,0.8729787357630584,
 0.061573237299692885,0.77397869832756,0.07494365501301437,
 0.17984622669285566,0.9779242191242448,0.5050721187554379,
 0.06298401316189839,0.8120094414608228,0.4374641580906494,0.7966898704468579,
 0.35590697406302496,0.1665827968668293,0.028779108719359514,
 0.004713570262712463,0.9403083209392343,0.36851408973970057,
 0.999797820198306,0.7884555546004778,0.7721982388950477,0.4562062471844591,
 0.48452411095558545,0.5176772342275011,0.0519959565465723,
 0.06668941897664826,0.6358605021673995,0.9982725167170502,0.7134184594660121,
 0.9666983384985258,0.3197869382769808,0.9716351818887954,0.25829059563183754,
 0.18570103841856211,0.381831600473442,0.22199538035338395,
 0.25247702155714113,0.6729624208406229,0.15528404031131915,
 0.3952448867455439,0.1835333628088458,0.9369016684186879,0.8046819636719076,
 0.8919594275245674,0.39557471132008615,0.03543252456166612,
 0.4738571626321936,0.6588783146833198,0.5493645917515108,0.2783305507865944,
 0.18036965438504327,0.6457599006416274,0.965027555404758,0.13801246665813938,
 0.03924079159335114,0.09364108661099646,0.8491245200365007,
 0.7620394353571998,0.564920304606978,0.832333794179201,0.5154830784513564,
 0.2150894177380106,0.23027774210952867,0.04003537504224508,
 0.40357964911396693,0.32166262226700126,0.4426353388973695,
 0.8941934762846968,0.8265397065285577,0.4777787581486804,0.49223763176397606,
 0.36703159523482043,0.6263872188967144,0.8942896724390385,0.8027506729693855,
 0.7315014854604666,0.9925587301594466,0.11807285725880656,0.7082113176695337,
 0.6791324600196602,0.34800141178762845,0.7466934051634985,
 0.02957642130544136,0.6764420790142791,0.01913885917802227,
 0.8331422012457916,0.0506660476792804,0.11723923434394279,
 0.17933150220351413,0.7912081495117127,0.9457268994998923,0.0608389860396229,
 0.718447563719675,0.010218910914386425,0.00558687017179238,
 0.9755869663053293,0.8595492419356956,0.9762048008707681,0.6140842560841278,
 0.8611995059145031,0.09762617205535229,0.8087498300355356,0.9618421023054219,
 0.8964761294240312,0.7934681615520113,0.2363216517934651,0.8771323812056785,
 0.7688170547653693,0.8670622157433365,0.4150138469751926,0.4518863751825749,
 0.22412964933630686,0.32353646670532465,0.7581613759390766,
 0.06081803184937018,0.41468829765948334,0.3011396379664921,0.732555650656788,
 0.13716284425813163,0.17001055351719396,0.934454062226491,0.6906588490755207,
 0.8161146533197351,0.5509994099975433,0.33227870986187713,0.2981239010932072,
 0.2224512158642612,0.3371148542291176,0.19690583257248417,
 0.15321967630579314,0.5554672304795291,0.7601090538243228,
 0.47669605459029896,0.9465517502076944,0.09802776760075871,
 0.5593020633331566,0.9428076949151527,0.6066803654148417,0.33798891769738115,
 0.14275177294725483,0.7730474154574379,0.6753306743270134,0.7475385046343161,
 0.11483496441821117,0.6067852447157314,0.7063440452775236,0.5267492127551702,
 0.40452508780158836,0.7440910566075563,0.758799657055436,0.7929761707503071,
 0.32305438100431116,0.7057844674416903,0.6913808600155296,0.7263089991285572,
 0.4196875263682833,0.2471519461317453,0.7841675194197409,0.17469855100205356,
 0.9647749513350882,0.09234601197662651,0.792005758110427,0.2831335528668656,
 0.8155312629440252,0.83191577061322,0.6508290756902716,0.03403929557989127,
 0.4628810776803147,0.26230056857088135,0.7193859699580657,0.8736734947608327,
 0.8111995404800866,0.8700888836280997,0.8506604451303351,0.5577619359314212,
 0.406167523583743,0.007975647266833197,0.539656739602005,0.3833065344067821,
 0.9435032546642537,0.7883756700825528,0.4581127428177868,0.7807671137608498,
 0.712971539881641,0.5979389780044085,0.6950251757159608,0.8894775838272803,
 0.3624923821685261,0.24928635316136327,0.34656958143112915,0.566544935596375,
 0.7350087098514282,0.8033555406463421,0.02501716069275639,0.7515490428531126,
 0.7029169907090651,0.39030726866389354,0.9951272483415167,0.2015938124903326,
 0.7016664658730414,0.7289634753832661,0.9675776212662791,0.0207198403604667,
 0.23679150913508362,0.04411743102458965,0.7481169124537547,
 0.23684603661671333,0.18052176526114583,0.2903135687389323,
 0.9771824865067813,0.46352284256238196,0.853823590615185,0.37111231527331,
 0.04125934894169314,0.9961283436224508,0.8186563528624613,0.3368206754443458,
 0.6780408325262303,0.23988629230015002,0.06916710007277183,
 0.002073011846361106,0.33633385865314047,0.012325970086412119,
 0.8352661103707159,0.38317216388244146,0.5628780213808174,0.6304795805246862,
 0.327865908532898,0.1780701636922718,0.8432997917391062,0.7383275091708664,
 0.6561635275844057,0.05187959147555832,0.23900023289080197,
 0.7166524126669599,0.034361352256706335,0.8750468032359742,0.414598627343744,
 0.8395955919239428,0.23765065872918978,0.8681294393539842,0.5235485503462223,
 0.08137973371606666,0.37355339715609337,0.9502997190845553,
 0.5817277683848425,0.3411860615107447,0.11641834569696796,0.9871890640870576,
 0.5067532834470854,0.04973357081790908,0.7124768602072724,0.2918301072304732,
 0.49579363807651944,0.6437923830791428,0.10089031420533667,
 0.8190656056416257,0.6022478748749275,0.6852545628891962,0.559010971340505,
 0.2833924058979128,0.4899327818035233,0.37998498347237053,0.7378071167081551,
 0.2718463747558083,0.7133908688991364,0.38775964270826946,0.5522589418981567,
 0.24069306390563083,0.3642784682708434,0.2388924465303487,0.7025062863707288,
 0.48860541630605225,0.3732760249662317,0.28044244331621226,
 0.8351487118109718,0.8325439517874991,0.9063405278495184,0.9469450085940133,
 0.13147065052304374,0.7368561988568422,0.7361685070797839,0.4707100247256393,
 0.5823098813126664,0.8378587868088172,0.8859375245047755,0.8436922838619787,
 0.8279413179321032,0.5198840109717898,0.9815124853092276,0.6362623593538321,
 0.47730075431174035,0.022812825662458458,0.9057543430824537,
 0.17829735008612946,0.9847651322543899,0.8656639922782785,0.7162788248418354,
 0.42457418689623716,0.8452589785919933,0.9447464546309947,0.3758832418122313,
 0.25467224766020635,0.7998848106115309,0.7459924088540308,
 0.09086623270238015,0.3787080661341564,0.3342310741957837,
 0.19721308771274748,0.7279734777722628,0.08247639649246874,
 0.4458794041909133,0.6029631049385725,0.682972440681233,0.48159599233104056,
 0.2946669190687471,0.5852350209234113,0.9767944070690958,0.9961284009695381,
 0.8086867153507744,0.03023513344784401,0.30762805378623825,
 0.4178663396133254,0.9699720852175909,0.3470778269610839,0.7288356336052731,
 0.2873746381135798,0.8679765111948958,0.4183514212909213,0.568727463884461,
 0.42811582340913434,0.5468500803515911,0.6766405439811127,0.6154000600509695,
 0.5418667831669741,0.87546192180364,0.3078207805272025,0.7288133067851814,
 0.5354371964266638,0.41541972163733343,0.20087861545494545,
 0.3364008442509866,0.9941870650704163,0.5800467980170692,0.9424553402527527,
 0.8902268546175379,0.8969918656931404,0.5527226914929617,0.4930322560603251,
 0.4390093405797295,0.07095202354145236,0.06801343846969976,
 0.19518557220698485,0.21702633198172472,0.38606603283730445,
 0.9384103837001595,0.11452040688415477,0.18152337834948673,
 0.13330673005283888,0.28377085143196135,0.8527109240467247,
 0.40371192232024256,0.72457726628563,0.35701380204100175,0.18663761419001312,
 0.42599085667077874,0.3165341091090733,0.3619681241304854,0.9359139698648022,
 0.306204673354858,0.7711580940642044,0.7470932992236428,0.9741359478448212,
 0.2549534775664454,0.8554443279466708,0.47642027879989624,0.5342439747940921,
 0.8768214142672196,0.36328406317785056,0.23362840026532183,
 0.6648082918414164,0.5386895533206052,0.42780356295427724,0.890740324632393,
 0.7882081578076101,0.432036471038171,0.5132055921862587,0.6995295025533091,
 0.574560874150311,0.4562140134227246,0.2669869036618494,0.9509939181084314,
 0.19910578768206855,0.2363275001111249,0.10585245005079891,
 0.12303244768266286,0.048681621900267924,0.16543678538230677,
 0.7829189107973424,0.6090740890538422,0.24842256538227325,
 0.22700994748560233,0.04365098452407834,0.35621653674718856,
 0.8315637297107362,0.8974180505375374,0.7202106870592034,0.255723940109313,
 0.4672959747489084,0.19635104758949296,0.6553583953815755,0.6794803125502383,
 0.8842105766322408,0.8673969185975873,0.9292334255068128,0.9892568103893392,
 0.45935991635901463,0.4469695103902557,0.6217698467373882,0.8157347279616233,
 0.04451347593689581,0.9075473762994435,0.768385083348372,0.9082482247173496,
 0.9022822033324507,0.1521093878105273,0.9953672669601333,0.8847308502772507,
 0.2489541765378409,0.8313569704124535,0.6429514093515027,0.8047832311955668,
 0.2928080500397887,0.2349460532556863,0.778112813462727,0.6678107916083079,
 0.1651322602126165,0.20319847049941908,0.05419853402862029,
 0.3136734204655577,0.4256919826918458,0.6893208598454413,0.646299265576928,
 0.6225130496966689,0.17842362770806264,0.8513427004105721,
 0.14815006776628648,0.39166896316685995,0.39748372038000424,
 0.8630284140807761,0.7098044896066882,0.08653637375090173,
 0.49160123447691984,0.5218294177105858,0.8898789760376037,
 0.09921062510416834,0.9094742876664142,0.7620702171055026,0.4992331028943321,
 0.054887468227854175,0.8972232760834203,0.2478381580550142,
 0.4953413256183843,0.19680388445009378,0.33253897072826366,
 0.8742246113354696,0.40787921363657575,0.00453638170666415,
 0.4035806748425693,0.14057446223298276,0.041046442631830105,
 0.07847786026925596,0.6614739780415231,0.819349829153634,0.807212990284883,
 0.8023835659601828,0.3645156521668518,0.7984885748903697,0.48379160604729543,
 0.8550214099079465,0.901045744760666,0.8638050025868551,0.17632651041386216,
 0.09252773315241469,0.48405321125010503,0.07632059154498394,
 0.1866344421279098,0.862172340970611,0.4992505297411166,0.6794876735837366,
 0.12752767097199047,0.79799595170634,0.7543861033484323,0.5928841393997546,
 0.38527105317914523,0.7366011446238728,0.5273839395035047,0.6440371331828981,
 0.8859180305414157,0.9328480018813888,0.9098580417486136,0.3759982332785169,
 0.36155871693031083,0.4363064751451893,0.07471196753003728,
 0.9721174591757109,0.18969670349507317,0.06836697872094422,
 0.44281093263821436,0.6685207647563869,0.23691143822785132,0.40220949652617,
 0.6941438725471314,0.059486207375277655,0.516240735615871,
 0.12080236596072957,0.30354955391192173,0.4594052485988602,
 0.08320211253809517,0.7908233865328256,0.07315483050066474,0.326480885497615,
 0.05594639730817508,0.5869409868384058,0.622796364965165,0.5415713244945606,
 0.42441354718540436,0.07426342707806433,0.9529491594424814,
 0.20319006319800959,0.9180674267485159,0.45293638701606886,
 0.37207159818620794,0.3252443908341681,0.6264526403212515,0.702264514000908,
 0.46778728513348034,0.33294141804950805,0.40356417705141834,
 0.23199755361322383,0.5279220586550434,0.6346191450904475,
 0.00963791163776373,0.18505618343024222,0.845050961464807,0.9807385645235638,
 0.2577670617381619,0.06819237788992927,0.9385919180673883,0.8184809489945808,
 0.7017121866123385,0.14438021989703942,0.06306407822025062,
 0.5512181224174998,0.9620529090833149,0.956923307777569,0.32350775809262833,
 0.402256621746793,0.24472148045278141,0.6187990062297364,0.4280590308782156,
 0.012551737306567556,0.7390310265078103,0.7149905576811592,
 0.5757639788595195,0.43784662334188806,0.7679816025198013,
 0.25854852517112936,0.014905220522052898,0.888292702274078,
 0.5412355918284544,0.15851936046740522,0.12720184420540437,
 0.17376877281696157,0.03729607518820144,0.7339450558468278,0.703568658189317,
 0.05457049782076662,0.6247631848690332,0.18592722162038888,
 0.8400249827083516,0.34315515572084676,0.9430717494244282,0.7174206345161674,
 0.05460365505083897,0.38007333555846123,0.6076762111702071,
 0.21588871592852532,0.8925445047996977,0.5673795626228654,0.7142951549640384,
 0.21492182836723717,0.15248517067211587,0.9455431513322912,
 0.8253750649931549,0.4161506279321474,0.06076895715029573,0.1935192363270093,
 0.27646290472773405,0.7704591864316582,0.6590523342645301,
 0.25273638697940937,0.7830738609110217,0.4715033216630181,0.7706501325781807,
 0.42085685282746477,0.005099825914974598,0.2910840117229505,
 0.6146909858921175,0.36476753644176896,0.9026951134268366,0.8479755488643396,
 0.7395671078962813,0.40097207301317517,0.07609224398259706,
 0.7800657846113103,0.15928353671573203,0.3071380086428833,0.8291718744549499,
 0.12951680256506282,0.08146971817057125,0.8589196386830081,
 0.4139057145119662,0.8650911434372648,0.9102587137300416,0.9989318253978888,
 0.8333280334552631,0.49818829888933625,0.4995069019874566,0.9028005207349483,
 0.48518569873863227,0.37557209194688,0.7305088437642683,0.46038607589020064,
 0.017672520489652288,0.9841433910634998,0.49533901887954923,
 0.4641656923942018,0.5106030359572634,0.7720199660080284,0.4164866308345959,
 0.9793493932648647,0.9530485796748448,0.9862375608180367,0.7173571261511588,
 0.27645770900917566,0.11394354371266124,0.061786319091409414,
 0.3565952648528431,0.9283724258035803,0.7840378422669666,0.6287173461192151,
 0.44357463607603653,0.9710472777096713,0.6538016255309639,
 0.13545063175153793,0.8895137637876207,0.7116452491167083,0.9720715379655316,
 0.663934223180128,0.9682589255916079,0.1251853866166156,0.1189148957507774,
 0.7887307845948587,0.09997998025943344,0.538527025839766,0.6883321835738191,
 0.8186344677575366,0.03739265639531941,0.13696965163784647,0.983967691442974,
 0.6083039360075505,0.5434566546293793,0.15758343987763068,0.9762230071942115,
 0.05306378803723322,0.6522759325264411,0.3079851963978092,0.8669079192689719,
 0.7703118205578104,0.49772548334152933,0.7797018009621388,
 0.01601505790245794,0.9545126757294168,0.7440321104044717,0.4361098473734729,
 0.9756760774848974,0.885536296558475,0.9165639077075372,0.8315095376764154,
 0.488362835529379,0.7989525093089807,0.09962230889726265,0.41051292005104223,
 0.6948722586487653,0.7238926018056057,0.25755809772481486,0.5043343007301645,
 0.3172798345414074,0.29808100631364043,0.5842157462961041,0.7626944511150451,
 0.9137479535248887,0.2694571045417531,0.7762209260552402,0.6750190111665685,
 0.6320228795548011,0.30271072084538786,0.18253131948031587,
 0.07811895140342939,0.8829709720937814,0.5820081824099916,0.3046244024996043,
 0.6800684810602046,0.7695340337087929,0.8538914162292767,0.2311206497901478,
 0.38314476847652335,0.4813954957506156,0.8016525443619776,0.9026279750683636,
 0.14373061351268435,0.5227472708521583,0.6647332458833763,0.967393571917385,
 0.10751087424436456,0.4840106617725911,0.06491849814527972,
 0.4554030469754342,0.29225474192510403,0.70252789370714,0.8260140508226885,
 0.07928302202371096,0.706380817977586,0.7011321016302117,0.35827079091929315,
 0.2511592228687247,0.5644021549307111,0.7359192528528337,0.571992929610686,
 0.8796837847291634,0.736605088110848,0.9840259562498724,0.1690529845022375,
 0.10472555464728739,0.7728426826809477,0.6590089729722606,
 0.26183696294100645,0.8106225256062793,0.5798220619930856,0.6502543451794682,
 0.12446046568735558,0.9972822091636097,0.44526690984086825,
 0.22012211217321442,0.7980431554442565,0.6150352093235699,0.8883928385819475,
 0.9446414516599906,0.5896695234998552,0.30046944092447503,0.8842724276520239,
 0.07536270375041032,0.3618225649242578,0.5936518705522256,0.67170307930278,
 0.007665309358483752,0.7964445004172879,0.8606346481150475,
 0.49020308877348173,0.9555256929461341,0.14517714277910532,
 0.1101424102155405,0.22072784263937417,0.5813384536492093,
 0.17287542667559985,0.5964995176067939,0.7054954101509375,
 0.016175338253555882,0.7748266970242965,0.8590993742130313,
 0.3108292676549341,0.31390509377160336,0.9572448737374158,0.127229921453627,
 0.9544317564380097,0.531893616058832,0.6937321760344181,0.18747016051924714,
 0.550166517868402,0.20465535463848217,0.36773639091823407,0.8009180242267929,
 0.5198507124912912,0.2315196526637352,0.9536630199994532,0.6868104787900032,
 0.4956821132668793,0.6798588891460071,0.9414420014250737,0.41099364022276297,
 0.17800740813704552,0.5373958485615626,0.1709753389239157,0.978900263841735,
 0.6672686192266876,0.17071750952854436,0.390008883832931,0.22513772196887272,
 0.8454362001035073,0.3766569427354278,0.8113633076488065,0.33866857269095685,
 0.32742845346081706,0.9775137243172436,0.7068897967368373,0.8120588207392561,
 0.676439724543415,0.824272170252993,0.6546101278664931,0.7213263275999539,
 0.7115131013504062,0.9045723422813509,0.2797383229545861,0.16262791523975295,
 0.01577468927743908,0.23443076512922312,0.04364474360946935,
 0.7699019262639963,0.5717294283008167,0.7329397518045764,0.7728419517363494,
 0.11242456779402665,0.026580608298264252,0.159978746188139,
 0.4243045136404219,0.9478085063576664,0.3830882477005877,0.719592505181405,
 0.19732539946891192,0.13041805468863976,0.9610892006823369,
 0.5453443236922504,0.4781826197025607,0.9027152791045295,0.5837332680129355,
 0.32971952976149876,0.708935718919556,0.6315782697414142,0.9353528838376736,
 0.39943427394875275,0.15337279254550018,0.8980042691312491,0.857352360755976,
 0.5086542994888097,0.8613605149637512,0.2471115943858222,0.4381433736274283,
 0.9719757494548655,0.9712122995696477,0.5907267023184692,0.10887479078189655,
 0.021792290617122623,0.12154183036709476,0.755027886147724,
 0.7177945798065243,0.1025826656151223,0.21263240949960516,0.8453437352859661,
 0.1439654071328491,0.7097262939343446,0.2099235298171671,0.9607083251004536,
 0.007705120330082749,0.19899325604356188,0.9816176290500006,
 0.7526692426211865,0.847568554235415,0.1905183177501053,0.26012514112306406,
 0.9766750382383005,0.7457920627512183,0.8656595018180193,0.424004446389044,
 0.31306720226398643,0.23707051765460896,0.9040689596872066,
 0.7870669091997865,0.38021950993668885,0.0852349599165082,
 0.19006488396063637,0.9639959234662879,0.8139497970119126,0.9750485401732646,
 0.03845959035097668,0.6629067164703478,0.1714141175485867,0.3445903740949906,
 0.43729520493407514,0.19015313501898157,0.556313643893032,
 0.24463606187291853,0.2876188318506202,0.35376091448992675,
 0.5634564978540108,0.7801125035720096,0.24991770772236954,
 0.11613226212288907,0.7867923295615065,0.1077256734541685,0.9172674824560841,
 0.0543464575470356,0.36336805912022674,0.49188005731986517,
 0.5105740338894038,0.5479930230255938,0.3279691718025577,0.7011949185058289,
 0.45365715851123567,0.20765390471051426,0.05961436914055818,
 0.8715724371637823,0.9302108987784563,0.7449058818774941,0.9926789279478535,
 0.9815555051975285,0.30127422837294837,0.20041036514523414,
 0.15141713674735036,8.322787919267105e-4,0.3656579666038803,
 0.7918105284967307,0.21266586647950847,0.9652116606510516,0.552227583320206,
 0.8144915009273298,0.4785642716668299,0.22480627405846487,
 0.20646985932366446,0.5347391910041805,0.9671337867320371,
 0.33683498085461594,0.1359455884289169,0.24816661259698714,
 0.26765348922699195,0.10588149064969687,0.005474334210288534,
 0.6409010164811355,0.39722910423087754,0.43520913711125264,
 0.7876407560050822,0.29954115661494085,0.96859918689009,0.9763514793379324,
 0.36846292934300306,0.34028921498609754,0.43095048081242715,
 0.7673408366112906,0.09984008258060106,0.7655783565945655,
 0.10760056478894509,0.1137872530987557,0.14136148643357527,
 0.041879267416599086,0.5220615166764472,0.6847094679840184,
 0.5986677935635389,0.6081138104281768,0.23657199095365156,0.5849040600929901,
 0.5389288046423077,0.33902513901279474,0.6244406340612196,0.584648813932287,
 0.260184981293899,0.2529295216468067,0.4839470727007069,0.5247823322889337,
 0.2658373096246942,0.1573188196932087,0.14964144967564974,0.8507660515344738,
 0.243738950378787,0.04919547671194513,0.6883599514315378,0.24494044770171086,
 0.5623338154137199,0.18093770499930473,0.8766930170528334,0.194892766405508,
 0.2809810637910831,0.788239679604099,0.5371319734207852,0.2016551701696805,
 0.4321459634742859,0.2713522863514076,0.2623681916573294,0.2915194341790881,
 0.3246177120463696,0.8018097461993439,0.008503179782331505,0.533537782052208,
 0.616144473691907,0.8828557307461359,0.9224742556804192,0.31376418199200207,
 0.12864119612047797,0.8302362090802842,0.36871124521297327,
 0.8015234528245169,0.21299245504687891,0.8193502283862868,0.4820022374039722,
 0.09810400421299692,0.9593195364428766,0.6990215579976575,
 0.011052320114308789,0.7257266350473378,0.7223190458493093,
 0.5056600592621974,0.38719832349939465,0.6536673087669929,
 0.005238278518668982,0.7682959552132136,0.9325477165450615,
 0.08656767106390761,0.23541645094669317,0.6507203998387278,
 0.33766153175459024,0.4382602366320494,0.10035408494242382,
 0.41957098336055365,0.9152095074741682,0.6896838323132564,0.8215782212483832,
 0.934096300024632,0.9324483765420917,0.09579962273495057,0.48397653683846287,
 0.4430280669934118,0.17525956491793893,0.6832553564290214,0.6669752311332674,
 0.8610981611498117,0.7429345239190672,0.9704976861714587,0.5794147204481728,
 0.3653985853912144,0.0818603233236671,0.735209132174351,0.5596242266526826,
 0.24949192776103069,0.9637716634877027,0.19315320864002883,
 0.5741216357594896,0.36688391887472527,0.4449595271241602,
 0.015868502331376155,0.18976772843351675,0.8366609180696581,
 0.7868565749200923,0.18527777352551922,0.985802440447262,0.6289777868203561,
 0.7517671847335041,0.28254787647827107,0.1003154461981296,
 0.031202225991595722,0.34496821866558103,0.021910040208695247,
 0.20837834555795287,0.1662596569184278,0.3495384625880473,
 0.19955308370210312,0.5056982550467878,0.31860142055107243,
 0.46890294336158533,0.8453771326289538,0.6909409304306501,0.554317849080389,
 0.9559154395695377,0.06691037432912694,0.36475486700601667,
 0.25939434228078084,0.4023286579035408,0.6761693683200304,
 0.29597663155218823,0.7206233982129792,0.3205433502046404,
 0.10902223659531618,0.26061974090546947,0.3901360224229202,
 0.5317167665011205,0.5041336960350051,0.5129110395388836,0.6815682490246582,
 0.15785161097979983,0.38487145171847004,0.06733632365988917,
 0.7673442975985556,0.4679594119926844,0.7581529894245356,0.519077281108055,
 0.07028396166286544,0.24187516114298402,0.6726466799188033,
 0.4382773272478082,0.5679930157483697,0.0817112110004744,0.5834132777331164,
 0.3589549523220972,0.772260010909833,0.006592838597012096,
 0.10979545310049521,0.2654046440244492,0.6785786195285988,0.5664712699481529,
 0.6214671626893932,0.4341340586416097,0.8503391203553264,0.614832337740233,
 0.24703278028064757,0.43623608640128364,0.3332659417844054,
 0.5487530172315376,0.4101288336083566,0.8430361831525879,0.07040339503169901,
 0.6896561277282091,0.153009244418843,0.5597559488404642,0.08035078661250461,
 0.7041795344067361,0.30717287588753317,0.8541388903837448,
 0.48029173924690094,0.15795807837108244,0.8949683669310436,
 0.23922591360626244,0.6613560440570796,0.13752137587807334,0.635795475224878,
 0.1776764531478925,0.9687307603257198,0.6600793089164644,0.281127055737303,
 0.02721424518172766,0.007117504660120955,0.5639676297563117,
 0.2125043296280026,0.9318064868428196,0.13207964763730695,
 0.43556236405049964,0.992245448964561,0.553689580149902,0.871602384391011,
 0.07700537198834201,0.12108683279668908,0.5790710139807925,
 0.27971856493463276,0.21590260156628893,0.21890088488444182,
 0.23120080213906258,0.5658993688622257,0.15423570237767836,
 0.3868517030459534,0.4994660166529107,0.12834376045528706,
 0.26384068647368997,0.9029580181603724,0.9433954839949744,
 0.06286484014994853,0.35574875567971587,0.28684425876893493,
 0.6969515219652935,0.9797334434919247,0.502275738511423,0.2287263905410578,
 0.5953832244880611,0.7022157307497203,0.1281677691381926,0.5282822432389517,
 0.019598356873388068,0.6127941810430332,0.1994817990895731,
 0.6409793746300441,0.814736188177348,0.5253633426241315,0.1641683879776391,
 0.9893739525361721,0.7515562108287149,0.29308640686957155,0.6753074766042366,
 0.5531315139138827,0.36328721689471544,0.6553821477309262,
 0.33666272835701117,0.7319101454064855,0.8118061155817385,
 0.07780632201630766,0.8797457710452273,0.5879712069578182,0.3296404158882196,
 0.7543507464193382,0.06706386284303978,0.2985250181071568,0.8312023171324338,
 0.8254810150518191,0.5431183193474485,0.8700912357875634,0.9511447731952227,
 0.19244009399729678,0.09985941916492813,0.7817298266595826,
 0.9369304675821445,0.4089729900682192,0.6970742268119435,
 8.636873029830294e-4,0.2886823163484352,0.9450258202610371,
 0.7820159141075657,0.7191660326120601,0.6855531761428758,0.7474115001787299,
 0.9517089377525774,0.7451781456381036,0.9923859531575989,0.01943106208358991,
 0.4904094037336604,0.5260626850941734,0.4125012287908375,0.05830509122611638,
 0.4039449556406115,0.3329532269870449,0.166127788787103,0.687687408333122,
 0.5179552170283512,0.9964630395010459,0.15971759249945439,0.3843409812853582,
 0.3503098863651516,0.2879237495677841,0.11190735312755484,0.8408209454505329,
 0.2112936684743565,0.1753477797146924,0.15372040615553695,0.5611112098856263,
 0.2140210342709652,0.3626119493301021,0.4978235262207843,0.4814728641364773,
 0.89734490020492,0.34392456530455684,0.7771029295246596,0.21643502353270605,
 0.34202449138769686,0.6321867409278292,0.0335829461903947,0.5769556277026575,
 0.3171884941372557,0.1843754466265638,0.28123713509905657,0.5989090033459534,
 0.365544143950048,0.9925774119404391,0.46057170077116794,0.09009608758247112,
 0.392660519526953,0.9055680265906972,0.024556483765757142,
 0.06770141550863273,0.4754585198480278,0.7145334305757097,
 0.040887168034581145,0.9518010138393014,0.4695165746068255,
 0.04720866555395209,0.41341055577800945,0.632449862744198,
 0.042378216772619104,0.9274243348175053,0.4197081874578972,
 0.19143046878050518,0.8244011969522164,0.883577370711366,0.6706148229292563,
 0.7480580985328302,0.8553547548738121,0.6843752671709391,0.3268660518792127,
 0.06359227173216186,0.04658346165222371,0.22200616351647973,
 0.6713252194784839,0.08855877352406794,0.04599943394517614,0.851560787528098,
 0.3597577356164997,0.6667047289772288,0.5471322619243334,0.0680918023804521,
 0.6484348213444026,0.10142161823055473,0.12411508618435729,
 0.7052182063588875,0.27796276584020874,0.4172568455954492,0.9345032796910623,
 0.7755972366205491,0.801724958841919,0.2140679922379295,0.08219015460161982,
 0.054022397777067965,0.5505742696031113,0.8219656591844805,
 0.08453862449795102,0.5462260619670799,0.30148993854369377,
 0.9352184053195883,0.36670114890303496,0.23717668572402895,
 0.7134019495925403,0.05739681545005493,0.006142993449907319,
 0.04888500995804235,0.9232689530188666,0.7177755116871404,0.6920006524568041,
 0.11765602454714197,0.5380530471134695,0.9589099574651699,
 0.49658356352956035,0.20325564144973174,0.862165766248046,0.9409745937229559,
 0.4045573089117781,0.5770504420468862,0.9926254232864838,0.02696211329148046,
 0.9540516139349217,0.9807937203894127,0.5619925404964619,0.5035608189457047,
 0.6453185235817699,0.6453799756792515,0.356127735492345,0.06007340099065939,
 0.8566551809859664,0.29454978356306094,0.8129980841354228,0.1329423775652422,
 0.6221549566542874,0.7472411115369253,0.6584558490637136,0.9103220213739054,
 0.09764261153123388,0.4054487960097646,0.5908797277801809,0.6391640224086572,
 0.3860606158078874,0.36964588595361514,0.17128970754257888,
 0.7540730728918783,0.8265487486225421,0.11514830608063176,0.9952810930404905,
 0.10109735187247737,0.5852187575997985,0.5230785579974544,0.8977993186329589,
 0.3158483329113586,0.5584412033920707,0.43348518480333087,0.6495095359900934,
 0.8346397995703341,0.7713806810581553,0.1060854159222906,0.2821103302815873,
 0.563036428919698,0.5652087516732439,0.295467604451233,0.9577199935284771,
 0.8281218795733116,0.8772397990819494,0.9401824487493058,0.6908074903880775,
 0.6007826109430069,0.8818115613780472,0.2719859270968237,0.5697188782045285,
 0.7344369126809481,0.3412187177753361,0.6399673802691255,0.7069641818720529,
 0.26408878678750247,0.33295044354500025,0.2136857508769665,
 0.4222301505047201,0.3237239278737801,0.9011891920003567,0.40854258219535244,
 0.21283883689524075,0.5925969366318483,0.8393438976194177,0.7155014731806496,
 0.9077351988083371,0.1187672236088807,0.24876428338522427,0.4330242899303445,
 0.3171231723782081,0.5374051302574095,0.6259552910924753,0.20535381833871302,
 0.6214216211053818,0.4138506656672425,0.39209697955920086,0.858351690700265,
 0.8526491077016894,0.3380234247661569,0.9454994591653176,0.5645866742561242,
 0.919803436302912,0.3725094446643069,0.18319395815737827,0.9539308973123277,
 0.05051575014881449,0.7709444145720719,0.018995541883486733,
 0.4360054444305511,0.8877338760837267,0.7816883964189761,0.08186647587334805,
 0.28190715713719416,0.7117136328506121,0.7640281671937991,0.9132844598425243,
 0.9242766613797371,0.14169404724472057,0.580330598115338,0.5862764910501357,
 0.8106020137216979,0.03126278764793777,0.5251241096582385,0.7990565725795871,
 0.7998199939941548,0.7190443399516471,0.5025106297002984,0.050875107258049,
 0.1217422596932245,0.7870689487524636,0.6638005424754653,0.3654371288002001,
 0.38337634833829826,0.6009907546058563,0.6378602344521361,0.6085224000038045,
 0.2510211573554084,0.25038033303001184,0.3750079436604482,0.9122552596348088,
 0.9256099685642031,0.5654659346087272,0.4092771552942511,0.6910014214341232,
 0.29636812080288677,0.2675555439234981,0.4281410172894381,0.6061190028017495,
 0.7632991872472197,0.37157663958383547,0.14093336610285512,
 0.3515551772581592,0.9607422314893865,0.7546812763221525,0.6561097899974176,
 0.6322739293577956,0.9347356905044194,0.08614711770088301,0.9989193752526947,
 0.24552246056482985,0.7855564622391353,0.4500814025063882,0.5206050020403623,
 0.7454109847220498,0.9948503643107154,0.3820829661117824,0.08870815364843787,
 0.08826756721698503,0.31464724001064925,0.5748344542966354,
 0.40808969533671213,0.9826512966859395,0.9663596053779241,
 0.18921000618471961,0.2184681279016527,0.5139430256685211,0.4260438174932507,
 0.6481020178492538,0.3716763809928897,0.35778519935680264,0.543906120046312,
 0.20840405929780437,0.754368799461671,0.2675085427867262,0.9173992711424352,
 0.10259818386567632,0.6287664932682168,0.9385290081399317,
 0.31456117605833633,0.46091453399914406,0.4852727769673335,
 0.0016139602390347063,0.3354207971443224,0.6495173223508952,
 0.9912352715902257,0.33412216332108147,0.9714052189858869,
 0.10318195516642525,0.3989287825113669,0.6565253767038214,0.5101992879228089,
 0.058998102401150776,0.8994623944974052,0.61870114792675,0.945974703558921,
 0.6387478335103959,0.9943013936053896,0.33413711389260325,0.9195527942721523,
 0.9401801007068852,0.9586542446900652,0.061355826092801435,
 0.7495120151109024,0.3603919580245819,0.4589996502739284,
 5.174757243938677e-4,0.8626820386045921,0.08966927108472056,
 0.40318644472335186,0.09472744762429841,0.9985718571025242,
 0.2450014824019756,0.9868036038821354,0.7202647831368075,0.819644785844617,
 0.32904353473214354,0.21226020341108887,0.46439641913072727,
 0.4934651376989918,0.6694296743799282,0.3199729169490514,0.9148231741996173,
 0.22062954553805325,0.30030935696400773,0.9622783828440072,
 0.5407092673189795,0.5207666498690139,0.40977529553935443,0.8638875163285986,
 0.5983419908327039,0.5541430720193392,0.07766938895792386,0.637207168883566,
 0.06088284219227824,0.9949302378127712,0.9714850406523169,0.4056548261535282,
 0.6614473722118087,0.41958949163553594,0.8547869445960763,0.7912184898890678,
 0.8704208758582973,0.5371423465657856,0.26360418122499973,0.6213601510627957,
 0.4943922013953429,0.8547975340452509,0.39940480641543463,0.9486252169563859,
 0.9765563422676362,0.7854426787224446,0.0584566970565783,0.5283360375295212,
 0.19193396430651055,0.13165125962537771,0.5954315347789277,
 0.9090618382319516,0.9619816997054682,0.3965596016097239,0.1318719346678302,
 0.3076891961038154,0.4382202524401152,0.9979332399650581,0.12818527300988802,
 0.554536467508281,0.31897918179048823,0.4824351416049073,0.42766842316210973,
 0.4395210400668217,0.6795158902317567,0.4882220050966567,0.434780985005059,
 0.9123325107303886,0.9598134106084278,0.23721231888644823,0.8187992265325342,
 0.48920179282706344,0.19289857747440953,0.7359412785518037,
 0.4415257293914465,0.05071460429459784,0.8951472621647285,
 0.027858678369086398,0.753892140371675,0.6672094057484881,0.8567173544452262,
 0.30256766280164216,0.36952856240379495,0.9422563427507905,
 0.1643690310888044,0.21445818435319053,0.033833520745536694,
 0.22015165362755018,0.7834858128186687,0.45750231443711364,
 0.08564590708047526,0.09289217047893927,0.752092523193598,0.260154180305714,
 0.8207626943207232,0.6162852741283221,0.8479325830471527,0.31350057469798764,
 0.7470371090770614,0.29152248363653277,0.5615269903161635,
 0.21292299273647597,0.21318206031022813,0.4109694992864589,
 0.6756953950404226,0.7676025759941009,0.03986650574860107,0.5117649656455205,
 0.6058253660886105,0.38746017962799606,0.7856946907294773,0.9521596860270766,
 0.6134204963960069,0.8184925384055418,0.6022994911797436,0.11300684293324981,
 0.6706635530642441,0.7649405689087099,0.8914021300618222,0.7223238739105505,
 0.11149778500277363,0.23840060649135897,0.3515404948390537,
 0.7715880277067293,0.8516702898366499,0.033998732772354545,
 0.2388068587936335,0.30698683475469646,0.03248104016191333,
 0.04269949618653979,0.1942275297023226,0.0700974266867489,0.83773772095803,
 0.7138123062647682,0.02014157688755225,0.3697570571075851,
 0.26368273558437894,0.5474232310327539,0.3902554021872955,0.709614178664229,
 0.9937803987924261,0.17139816116948414,0.8734159238534914,
 0.22601093655511983,0.3372480236054721,0.3667010689732051,0.5881092534384843,
 0.6529199803296548,0.30904740826256627,0.48574741052786685,
 0.16404741016126123,0.26259652821184554,0.49626469106661664,
 0.8696530405778581,0.43355792095940515,0.24817682201805413,
 0.41790368487341145,0.603632051819681,0.415196007508717,0.5263422810185628,
 0.5282974560160443,0.922791936711276,0.6570945188495696,0.11832695290236928,
 0.7914868320183964,0.5889763844890605,0.8365236256642944,0.6691875444713739,
 0.6221328999864358,0.47277526474752385,0.5192214433445277,
 0.47612433097462714,0.8934232240617491,0.4697769262649021,0.9829290735687557,
 0.7762374496356387,0.09437966175982393,0.8185065140463207,0.4501374337948605,
 0.6940335825313682,0.5564681798004464,0.860078592847961,0.347473864098475,
 0.2669671870099245,0.3170384559234515,0.6485059651797321,0.06915946217769398,
 0.3332691710032434,0.9360292104120983,0.5563094663476055,0.31830443686692367,
 0.7262460346758629,0.7860653961448114,0.16220474287376807,0.7077688732714361,
 0.5193762152482457,0.38998699894237165,0.1147801009840117,0.8685447606349156,
 0.8356426647062073,0.03579908615927396,0.4315974729389753,0.5442557074679221,
 0.5058711413464942,0.7233529170967907,0.2339310106849627,0.8052515599576675,
 0.5604572346822567,0.22871515447042645,0.11090906647999321,
 0.7077279304745965,0.10869666243539056,0.7593833551898156,0.8971585046451597,
 0.2756038847698974,0.44856573300506897,0.9136312138445807,0.3076779800778713,
 0.6172659919031939,0.6391404247354975,0.1930831528567729,0.07547473134826721,
 0.8016784365540242,0.7686094384590906,0.6938276736355379,0.21009199492468889,
 0.5843526415905533,0.8421584545961059,0.5652160063252878,0.6152416350199726,
 0.7875138394625252,0.3463851542850599,0.2036767816276739,0.26099240560111125,
 0.8963534382740178,0.21933237935691596,0.567645620206333,0.7093783061391749,
 0.60755347387717,0.15015896819510888,0.5801776561450189,0.08923807786201354,
 0.26022672605343733,0.6866532005221916,0.6118392060797454,0.5108977762459317,
 0.6134513226841048,0.009579155910042214,0.11302850937772535,
 0.6243788291151506,0.2766366517978597,0.09944551013146019,
 0.22167351303867588,0.44349281842747745,0.8765431242724688,
 0.050814891751821745,0.9659260547768962,0.024481818968519753,
 0.34422285075902204,0.2970804077003537,0.9925045888067348,0.9197041680365963,
 0.17336636954635665,0.9227684759427767,0.4870985883081024,
 0.016130821025915787,0.8093491452198698,0.7710113194394117,
 0.7076252715596959,0.9681345233786618,0.3906759515632682,0.9216341328780668,
 0.7295916674898509,0.6378861242993086,0.08266325240288652,
 0.24585382691283042,0.383112417188898,0.7822034782801939,0.5913041096232223,
 0.9936180942385504,0.732084732798071,0.6999403231351811,0.396540552526191,
 0.5424485236841827,0.7576831832688787,0.12574244944823354,0.9428990113347778,
 0.2841961204376636,0.46611603766612786,0.41003827211847943,
 0.5227832750839649,0.6312414327503731,0.5437736200467702,0.675311302439194,
 0.41644885209007054,0.9886879512646216,0.23882548207234644,
 0.7772897205064884,0.15086886772757568,0.4910579688325437,0.3656605876660215,
 0.28349268772394787,0.6166746596964724,0.10872850463573769,
 0.3009804483358606,0.0905571182763838,0.11930772557677449,0.9672448410918264,
 0.45086062436885044,0.7057182565875314,0.30823635259952176,
 0.3788195050671581,0.11047803344010731,0.3898965983334485,0.8543577031946624,
 0.826832027349335,0.3934370446143981,0.04973341594565217,0.07290448152621287,
 0.5237331782317627,0.5143195526368346,0.4165299012594911,0.6235584999567663,
 0.7354798631171883,0.5448211438534443,0.49589402853315023,0.8779185668449272,
 0.6980055445825858,0.4929651045819249,0.243136275192675,0.8988770189366422,
 0.780314380442256,0.8701413854191724,0.2772681455871536,0.5915848336803604,
 0.5296431265142624,0.12335814591425609,0.039922912732540894,
 0.24959101927727234,0.17558669576541608,0.41314915429937527,
 0.7587269602605999,0.4290313323293681,0.7322625205315407,0.20999926675004077,
 0.7086277271725415,0.8616007062996164,0.8685022805224827,0.991592968953202,
 0.40795143252120947,0.14599757166023541,0.9417207082503511,
 0.32978978539546455,0.5373202150055857,0.23718957753858394,
 0.8189072065992186,0.4214452617325435,0.5704267543885044,0.5152859742772818,
 0.5117093778124433,0.697113667547701,0.666856206976668,0.4388594459198798,
 0.751889316232341,0.46298407473295644,0.5429603822169217,0.5888235412716789,
 0.3409308664645102,0.9935575888051379,0.07475845594479547,0.5449311818832298,
 0.14792942559174438,0.3443372567239702,0.09309082634911803,
 0.7503989537211915,0.21738729438589333,0.5033069872647986,0.5596940702887002,
 0.8297885981554558,0.9585974705555057,0.354374405280783,0.7716142689035641,
 0.7529776095535323,0.4060420517521649,0.05294276060358416,0.8673704600106924,
 0.316680910001486,0.21205789485714177,0.8269253524218347,0.7633547574274993,
 0.14483440360956268,0.3839063252518533,0.8351347812300194,0.9430366475747441,
 0.426483686331115,0.85400926148927,0.10535168694204367,0.09543534800586495,
 0.4585105266404864,0.6457632650253778,0.4375145870336543,0.16049650117478043,
 0.2759153251164873,0.15289466934714657,0.729480937081716,0.5285874859995523,
 0.2380431305285836,0.4800555236771804,0.12915787559025116,0.7511611162400984,
 0.13232041749874557,0.9964102040219946,0.2875923455303636,
 0.43350178921843563,0.379554749382528,0.9573491475607381,0.1026941589609508,
 0.025608937731163017,0.805771502474839,0.1753878381527858,0.4620150199367763,
 0.49476091735937033,0.19011277342352428,0.02458788647739174,
 0.4765580940558112,0.7661909751421894,0.8647796893229925,0.09895994154680243,
 0.9954802098725799,0.9799903562049614,0.23587973797757245,0.6330357936345807,
 0.15188658303410563,0.6335676880063223,0.6504408777140097,0.7891801905621467,
 0.9613540413641519,0.8083331200584778,0.24250592587523467,
 0.41578449478602786,0.6039186574516351,0.12172644630576979,
 0.41909413770918114,0.018755957562812675,0.17468694711114996,
 0.60330350656358,0.019452254729935614,0.715813722763033,0.9377973817124207,
 0.37421912860279405,0.6029174529695613,0.4758419962023238,0.5412703553661773,
 0.6102768416558777,0.5288726145722503,0.3768720768105135,0.6595012436425077,
 0.6270067592044081,0.22124573976854534,0.32729475676943376,0.930621913650415,
 0.6136802077470954,0.3685763253820702,0.7029460266366453,0.9977272633684143,
 0.6079765446791611,0.44806608568610373,0.819175445980726,0.34699624239197824,
 0.6281025227767342,0.5620096898934379,0.9829893861619764,0.9271802035872102,
 0.6040193531587605,0.2618469189800694,0.6513888407255924,0.432036601960628,
 0.5749757367424985,0.7232361709641217,0.5531900436491319,0.3097874860832649,
 0.06690557796035623,0.5216434390037618,0.7446559157919547,
 0.17650635498350153,0.7498117100687895,0.5331611046111779,0.8699886106789396,
 0.5088686444125519,0.8146008322543052,0.35488066669705887,0.7698485393608187,
 0.7899228911204257,0.5945849933597046,0.18310645763084266,0.991455610240262,
 0.27278898109744354,0.8761979189347794,0.4142805530376573,0.9957657325458921,
 0.22217080995149385,0.9062968307154712,0.0027028415199323863,
 0.21749536909778677,0.2854919148207029,0.36739867531583625,
 0.30473183340097554,0.10488835553675235,0.6716133390785135,
 0.8326755805102762,0.3091734941583738,0.39948292822238596,0.75781384624001,
 0.49243398025088203,0.37629727048765826,0.8536423784909427,
 0.5062663708967294,0.3661171236370192,0.9372076710977938,0.5355585984361984,
 0.27899430536456904,0.7926045056058487,0.2031630444415905,0.9058475022389993,
 0.3637527624031097,0.7563647741416548,0.9998702127976824,0.4868246011301163,
 0.03690271228834585,0.6330729373010665,0.7018752999889243,
 0.25552447291004843,0.6912692046640287,0.874679036530001,0.783898485346066,
 0.34531093519208866,0.9476781863908982,0.810118164631008,0.42309672698357903,
 0.9846493971778346,0.5137257624357201,0.811956241317388,0.19270969089660783,
 0.46722772149991976,0.9446021873250436,0.02442492107776406,
 0.9691789508489178,0.37995086119956856,0.31942533143401386,
 0.9022215280520466,0.7844181200520047,0.5357184920438738,0.9006006150622925,
 0.31176387214319723,0.3385363388548812,0.8425550385575294,0.4720250774158504,
 0.4189973909433695,0.45495666178029226,0.843711006217368,0.3510526111711709,
 0.08639527481406473,0.6627329895377257,0.5345372376694452,0.3635128419574425,
 0.1965280351928046,0.989684076262825,0.13784835399549755,
 0.027769053080316408,0.5199262125958395,0.3160002169365934,
 0.9347120528535698,0.274641563242196,0.21763666892652456,0.11246388830000398,
 0.89698651202632,0.3897572156347412,0.48756419399064055,0.9315062951045612,
 0.7480238892367131,0.8908448593725885,0.41100990110966007,0.5675123607592363,
 0.3546895209763903,0.38812181057478723,0.3030005651239285,0.9162403297182233,
 0.8801320134408244,0.6080329943528839,0.30121543113198834,
 0.36113951026253366,0.9706010659332823,0.792189912370607,0.1551894359596111,
 0.9235462409370344,0.12986699327338003,0.5987596522393969,
 0.051938034112153364,0.26546954608783757,0.7623868718936879,
 0.2686108931867417,0.05064289925995147,0.32486301906273973,0.554525014821214,
 0.3463468767898642,0.3381910209627963,0.7374023673809461,0.23144473454317083,
 0.9972473078292338,0.4034970222803629,0.9181864338833736,0.00997736028381735,
 0.8990663337116372,0.042138004974035304,0.3437958899775282,
 0.26582030720275773,0.5305283173981834,0.16024462486943447,
 0.9932242075927279,0.6765514512596511,0.9715399977980587,0.6515560718328655,
 0.22502166072035812,0.4699212132094517,0.4270997751148955,0.970756686641877,
 0.46835046161191585,0.0015501909113000423,0.9662035042575583,
 0.48184042021229,0.9567557657713589,0.4476899667419275,0.5605414118105543,
 0.006281600646326102,0.2100858505468396,0.6497153373483874,
 0.9747846512199124,0.849071250961114,0.2429382257732905,0.595967429386944,
 0.6653665961425207,0.18065091965802893,0.2839176848684457,
 0.059649581407178576,0.6830528676803885,0.5313115247155296,
 0.18185291041836416,0.00745320673304084,0.7134121097696886,
 0.5330780259856692,0.3362485354983651,0.4052457229507891,0.8568618816477283,
 0.21938158213968606,0.6422719366998928,0.9987964017440722,
 0.008558992753278094,0.560069250493853,0.9397980266225797,0.9597737443127524,
 0.20412481492671453,0.8139072231532096,0.060158419874721036,
 0.14311925089540267,0.650059945800885,0.5008134535018405,0.24506231061998152,
 0.3073531464163267,0.8812424683374964,0.39182299994082315,0.6803210262450069,
 0.43703091183823584,0.8955396555555207,0.2623816351264362,0.8287633548438654,
 0.8320874475939051,0.08943758114239753,0.5752302336083631,0.2836338886583796,
 0.42504688796834,0.9451674322651493,0.1661638633530501,0.2755561022794003,
 0.45482712855121377,0.9670753145812863,0.7454647675647887,0.4605906570693288,
 0.8448940923208427,0.48830320629061863,0.7586757543248941,
 0.07121667174109769,0.4432315341734817,0.5461266673944004,
 0.17393426333501594,0.8092994585191954,0.5269198739175438,0.765294257672488,
 0.4827562340487269,0.6357069274453768,0.05249928458906017,0.6606461498497647,
 0.22078349458234514,0.769713230880276,0.8880809630233858,
 0.052939301775177916,0.8948065436915928,0.6520253059174386,
 0.7773257041842517,0.17855997692336634,0.35264416033140844,
 0.3085368409120174,0.8786851632683331,0.7469178488574397,0.1027926754489803,
 0.1684163515376872,0.38397270907928893,0.1678220150247164,0.9744511980847342,
 0.4248793265353973,0.0067681644138016805,0.16650846174151557,
 0.03415964245164904,0.42970515897868533,0.9376712660608392,
 0.3700127056310182,0.4068673137812532,0.1700290372541392,0.35671215462599803,
 0.8570322465844018,0.5450854086301437,0.9217184598236564,0.7775905370456677,
 0.446317116552148,0.22814386977124745,0.2420775940036506,0.8264862085354336,
 0.16613188593991834,0.4733434923051334,0.10554731321172328,
 0.013913430027825413,0.34002347921307796,0.6932864073261157,
 0.03846403507716034,0.2367943756459483,0.927076059537644,0.1651760511945256,
 0.4263760589115234,0.866731687181717,0.31647813120365764,0.1219988217400364,
 0.07733886948939916,0.8632568602778659,0.822124089804263,0.952360835239422,
 0.23011761856696278,0.5052942175582876,0.24194862588832378,
 0.9360835579831471,0.7048510642080501,0.8600141477241597,0.6773959382598848,
 0.781356620920572,0.03223798917972509,0.47064458314129287,
 0.26630288837346994,0.3116953064239344,0.2313180305597854,0.7117013221935351,
 0.37218357875494157,0.3762590280491621,0.31398077128364943,0.581181909986954,
 0.5782816332950851,0.061218387695092136,0.49808592104458904,
 0.4219927991591106,0.3258628952054644,0.34374544129466755,0.371894820757247,
 0.0756965135693839,0.10234832716355036,0.07867036759034196,
 0.7458601091382719,0.7011209246589252,0.5694850847695976,0.1244180118766206,
 0.18691024461824934,0.9139022307931537,0.8899784066540163,
 0.02776106072733775,0.6886101230117948,0.8057686552925849,
 0.23436908752602736,0.8672559261755218,0.7500911468450399,
 0.05512282964224413,0.9792378009954568,0.03609402269840922,
 0.4855103428528953,0.7833256049446143,0.620465208787062,0.2688308935649142,
 0.236900869548182,0.816244966829131,0.2598919592281479,0.8825860643293189,
 0.9505181492542596,0.09207943141751929,0.2655421481118867,0.5533900747168671,
 0.2272577340365649,0.08768847328072449,0.550032047246059,0.5210477426400938,
 0.7133904933760471,0.9968690908669066,0.7466923900584853,0.9919466404206778,
 0.39765571049741555,0.9287852727474052,0.9028934543370625,0.7336361651942076,
 0.5794866578931205,0.04147902060036135,0.14831657661779796,
 0.6154621033230874,0.3083723283737627,0.8735367840925684,0.0614807677031578,
 0.8264757715176044,0.7306855154771842,0.9681857421633557,0.9813314936312805,
 0.47109531794112414,0.673400327679561,0.2027415581995714,0.1857264463043118,
 0.7942343570218713,0.42350446417557497,0.03287537320141276,
 0.3628820732628738,0.3020922050418786,0.6395622917161675,0.9210239878804031,
 0.13077787125092988,0.5580524137946405,0.4985065553867296,0.5198550034752385,
 0.10706887428264211,0.1790961499466026,0.8372344945171397,
 0.19682799018539832,0.48339634471473025,0.8503508556795277,
 0.5139268926155639,0.8192431230069757,0.21576868249004422,0.2346787177849302,
 0.953449553491485,0.06604789240663356,0.4158254191169505,0.8359903458114315,
 0.7305291528945008,0.7145980765658635,0.06751667977998688,0.802996597245679,
 0.6566150624496399,0.7696409899452759,0.03153872611172093,0.6334334998969646,
 0.9465373180549693,0.6572745172447968,0.28764479449573643,0.9797633984725479,
 0.8835598693496983,0.4021026370131642,0.8304509193725426,0.416255274628327,
 0.027695011547212323,0.34968232706891345,0.4356718733684968,
 0.24574723739516313,0.003970352268270583,0.5184738714662611,
 0.42866505933836174,0.041650743293711545,0.03364366917007611,
 0.9100717377483635,0.5669873114553107,0.915721805330435,0.2777109882330475,
 0.5803945436105967,0.3209282115434272,0.0985177188852212,0.8080740498892756,
 0.9497990397267924,0.8027080685838541,0.4629340928893253,0.9828204914782175,
 0.8705345289519127,0.55604442657772,0.8143240033953314,0.9290371337392537,
 0.23714150153608982,0.1937542269595982,0.8414517393280494,0.9389337080707847,
 0.3355357928781084,0.29979707350211504,0.16758452662332235,
 0.16257005296620697,0.6304236372164332,0.8547999858719003,
 0.19993283449899257,0.11529410917524519,0.9298374686664039,
 0.21641639086212006,0.6621997601909567,0.3650490558201024,
 0.33744419108722545,0.18591665691475745,0.6724032075099586,
 0.18930306977105216,0.5612763229768182,0.5659281741060096,0.3341778976793256,
 0.3384738240135199,0.8359806223278128,0.2785089604627212,0.08714506255361876,
 0.2858584190422744,0.06424350094180686,0.8108064983195911,
 0.07349216463425379,0.9333859747162676,0.1025483537447851,0.797113735918108,
 0.1948997682790783,0.7643685321120561,0.9240361817170609,0.19378112222897925,
 0.2313564147340853,0.3031942889187882,0.2539923376023262,0.10411033445729612,
 0.19828701955682304,0.554117856930933,0.20797940190610675,
 0.14591225463386004,0.17409484955112997,0.17330031885112485,
 0.3941087637230054,0.037442777442363684,0.32624676846755873,
 0.8802909347855254,0.07048521120839502,0.007372380279503266,
 0.18745368278579422,0.26327231408507534,0.8918711041228438,
 0.1972169190366977,0.41936911721934556,0.0226295010552241,0.711046895640637,
 0.09460484692368831,0.2365683154647783,0.6510150456537238,0.1457919065792802,
 0.9722211856789168,0.35618394361603245,0.37326540791408236,
 0.48549572438455824,0.6822167486351111,0.49706297963655655,
 0.5267094259966876,0.7477693437554661,0.2767854550367571,0.30075526805917996,
 0.2614756077571654,0.7798412700728702,0.49508730688560887,0.2518288583938586,
 0.5077553413868351,0.8320349538911314,0.03294434320292361,
 0.07523481824585698,0.35526774000022654,0.5459770616532511,
 0.4577596283226373,0.03011172511755822,0.5170505685835243,0.7999404887188798,
 0.9617003224540556,0.7065129735038971,0.06267981136924539,0.1320063333440622,
 0.6433284970208555,0.3399777467698334,0.4090943334670677,0.3932351477065671,
 0.387963515304386,0.40860270410720323,0.4458797252209292,0.5965384880447031,
 0.3331684255068442,0.5784004237580966,0.20265562493881806,0.7684734284766479,
 0.4375734581709093,0.37920041862352805,0.3822504893257691,0.8517650369911782,
 0.13725588939199374,0.7066361597817917,0.38218998853854824,
 0.17163663025553044,0.693845790222404,0.6940251320313001,0.6230268728893603,
 0.14310400919263377,0.11889251853819705,0.8648472891456512,
 0.22582124510091295,0.44198324278519996,0.4048099639879934,
 0.9756067448277275,0.8741808554563456,0.9328782523590339,0.4494512275444503,
 0.11354419571564223,0.020737723642637707,0.4128684545433341,
 0.9632919325379545,0.26922672037194817,0.556749620430071,0.4938433040315957,
 0.9370584723831001,0.35916050944414135,0.37956532679096977,
 0.6272184331009497,0.8183192132308943,0.8552746278853478,0.21071284244442834,
 0.862040547787055,0.8242239052470781,0.05391038056564079,0.08141934192216371,
 0.447871347548348,0.4749468665910809,0.23038662158727796,0.6152112053614958,
 0.09107083460694221,0.6697375080215437,0.7035250012913081,0.3500983938661839,
 0.769609121807561,0.7993708377098921,0.9669198177449821,0.5727731916815619,
 0.8213934593589194,0.2865505215139189,0.8699569936604862,0.18930780679730153,
 0.2395792541688907,0.3309643288744065,0.7725123708164107,0.04825416850046049,
 0.23729682907392124,0.7646704896143625,0.3864877453098263,0.845464622590044,
 0.7444194818593084,0.2879317500543279,0.9253113313700365,0.5716101481192881,
 0.8330079493009345,0.6922687759930608,0.7503589622337902,0.662780928450897,
 0.5786997857063182,0.17832239688867668,0.9028599565210147,0.8369145950612809,
 0.6639303536600929,0.37653502080843004,0.055873496805904144,
 0.44530688353813097,0.9494140054446926,0.29701863604528267,
 0.3405658029121592,0.37590941042906345,0.3434213752807952,0.7019450419494164,
 0.5262391411045416,0.422671048146288,0.42632216014664903,0.598116056104958,
 0.6579481294395849,0.6190697838434794,0.9692543023916562,0.3853976538707621,
 0.1480436541050525,0.33351099072096535,0.569193829183025,0.1474539115728173,
 0.5829147209815566,0.25150421250606037,0.5770970034189955,0.525482370310944,
 0.5376659027113941,0.5071354966871868,0.3112285050590443,0.9663408775560027,
 0.19769255492371396,0.1240768412886013,0.33192023985868235,
 0.9100404622035406,0.36073160151713446,0.7638678240350277,0.3943586283190317,
 0.5292730244194827,0.1871432623173297,0.29334210743799427,0.7141236248963079,
 0.3318666658426487,0.658691300271132,0.6448055400262337,0.37861866456213666,
 0.9711280861602987,0.5247459934638344,0.5633758399524348,0.3613136953351953,
 0.6046500307706748,0.3722979841985826,0.05561922251244322,0.6363166154200014,
 0.5564678859992735,0.5990775314577174,0.3076674865511698,0.9107658879558649,
 0.7617257566423461,0.5307307738999588,0.39199895556357456,0.6315638778644507,
 0.11730227942124072,0.4806958951175171,0.579560305333509,0.61305807432311,
 0.5713850579745665,0.8886156312773058,0.6184753241754363,0.7642581526774658,
 0.11019253194196665,0.6489854938063342,0.4905503023069696,
 0.22954739982073968,0.7476482335744561,0.827652155183858,0.8066154725665288,
 0.3555466390822726,0.7556660212102095,0.764627260053867,0.7766443024978337,
 0.5072761592576531,0.24868178087084236,0.955531453476424,0.5687904625776289,
 0.2783371002820425,0.4315093713883892,0.15241511318178735,
 0.48294611380200914,0.3314502121733628,0.23090702251606887,
 0.7056216259438058,0.09521219905085743,0.3307327509342396,0.5798540016536583,
 0.6465473546860707,0.38467477916844095,0.49477631115294507,0.768890918475081,
 0.27195020133655645,0.23015413246215988,0.4122244307049382,
 0.8925069197736297,0.0849361990945613,0.20638876763060532,0.3500261989953738,
 0.13368791016040016,0.438754737400185,0.8694095744142589,0.4704074197798018,
 0.16662770587972886,0.7607177504004136,0.12289884433433407,
 0.009329044942842035,0.7545313587106408,0.2519657367737467,
 0.2150459002906192,0.9156025663803518,0.37717974476395566,0.2826545238928895,
 0.9041567199086268,0.4049649733253723,0.013665446156100436,
 0.8568093654152228,0.7603995653226517,0.6477949368473709,0.8518932287393337,
 0.2550856202373065,0.5156015043363746,0.1506066279098871,0.4521408108301628,
 0.39730480470655993,0.39746771660461644,0.06398148458312114,
 0.2912212690170015,0.8839838087996181,0.16342441931086582,0.7164582105021082,
 0.23778654909399322,0.31134780691838015,0.7778109721652581,
 0.5022078572623465,0.8472582175853748,0.11176774057057215,0.4411369516778786,
 0.7478898777199383,0.3014118552590219,0.20897364684931086,
 0.40554449485037847,0.5894987674572747,0.7066179372986308,0.9434268913770226,
 0.7781749310457373,0.7368523407565545,0.06932539882043731,0.8479081537896329,
 0.6326139189968397,0.7043669825280525,0.24942670027548286,0.625627007547656,
 0.053849714854995456,0.878548028949848,0.08056790869622787,
 0.40550736912758156,0.10075166892886245,0.05768054208389328,
 0.6610668288411464,0.6919939727960088,0.14871032588611266,0.2569100622190902,
 0.135859353863926,0.3556720968274353,0.023880051233094113,0.2869241505955965,
 0.358636200468867,0.6709540839931489,0.780498610387256,0.393931762090445,
 0.5522667549843701,0.6246431937448033,0.1465434991445722,0.3894031920495242,
 0.81083022431503,0.5816631442320539,0.6235752133053831,0.14146430608549465,
 0.7231455925846606,0.3923206627078052,0.8315754897644113,0.8290599934530417,
 0.7015459527506659,0.21807749052476422,0.37510182075601206,
 0.9309520061431784,0.7683667552627313,0.33887951177029285,0.8148516911805208,
 0.21368955047609506,0.4166158358205988,0.5783135986772049,0.4250057137622226,
 0.3079165324346209,0.5062856486350424,0.9069401846943523,0.6109552910687148,
 0.16000105400686082,0.7456354400583172,0.8013300418569118,0.4211029638014345,
 0.7860380518094554,0.25518802521507955,0.6979720668291096,0.746929806095325,
 0.9284674530083523,0.8593808248406671,0.8344933201933925,0.09338239161163053,
 0.4328008990008718,0.5781574149331244,0.02801328960513172,0.9392864198718198,
 0.26231795599331564,0.2990080151823823,0.759649723994705,0.9508132026662881,
 0.7481101552647729,0.4214655829756564,0.01218529005565494,0.3785976862656878,
 0.8651185289772912,0.7623921364140267,0.7108554402591997,0.14289349230706672,
 0.4460065423573807,0.8617182475330866,0.13267087553097534,0.2822962377385978,
 0.5636489774092703,0.25375575477125467,0.1406597218525173,
 0.41953945376516355,0.36682308688388865,0.17453908230389004,
 0.5202858068204932,0.3225504830659922,0.6137802225835768,0.9862044549182765,
 0.20153964484891884,0.6359863565927213,0.9438695841822409,0.5641706109905312,
 0.1284417885339928,0.29202595238048,0.8100875760385723,0.006945770301659437,
 0.9525967808031062,0.6221514539937448,0.8966149492649266,0.01622751017915469,
 0.46753478449681274,0.9980542658611412,0.10173575643663879,
 0.6613264758793264,0.6104996526929919,0.8416212056900088,0.7839076049337304,
 0.10803582427477809,0.02025630559164482,0.6653868804190887,
 0.2542355246586817,0.4594940531248203,0.7526390080446005,0.08588842754377968,
 0.4666325043768458,0.24076237666985978,0.325052461206902,0.832617805085006,
 0.5658728995305707,0.024213804925683258,0.931341474191821,0.6200334001016972,
 0.390770375441183,0.493935990651145,0.23872041616349937,0.1791814335711206,
 0.24021070145990575,0.7234684807745861,0.6226493294713359,0.3216569919597393,
 0.9842442903298707,0.25001808101993883,0.4902845847003937,0.3555476195947088,
 0.5975079754982744,0.8574106891394292,0.7140649262382892,0.7710151895701394,
 0.8558997351182811,0.09883816474524143,0.2958801393637198,
 0.11428775530708024,0.8589187606970606,0.2992474119490698,
 0.46588688759394703,0.20297953236119592,0.21972009163996153,
 0.6751453604441398,0.47137191397506784,0.49366595249984935,
 0.7590728770364253,0.5822524432603424,0.9947898366374547,0.8563511727277229,
 0.6468934580231188,0.9179002504859712,0.7320973532586738,0.20214357493975943,
 0.6046582257623118,0.23562498515111951,0.2170629265744214,
 0.03152450450042299,0.4204665958553615,0.3562014644249446,0.5489538899644232,
 0.7770117949309494,0.7305687570407164,0.2354480898019512,0.5598911575001636,
 0.45045479186834325,0.8491110465349947,0.5910165004347574,0.8032579021335401,
 0.5037983262759694,0.09136021776791947,0.8003884546067327,
 0.32184651656167895,0.8418659873402938,0.8635288427250714,0.9926247077429461,
 0.8066747463466665,0.09584890839676974,0.7242844527354869,0.6873523949402388,
 0.06440292688416172,0.29984641979841176,0.4480434558904417,
 0.25366456748218136,0.31556262005154956,0.8128109604344299,
 0.2979819592373303,0.6390278948094475,0.5661296625964966,0.16963604135670707,
 0.9197400292503188,0.23502151284878825,0.5263560970151552,0.2289427571611432,
 0.278308740113281,0.8797674486547914,0.65333839064845,0.4923289669313208,
 0.4215537141436041,0.43137268714360577,0.3294573617759766,0.2137922959914672,
 0.7602519388564652,0.6433627498019298,0.4208129649798563,0.86758035076985,
 0.9838401930178666,0.7416979134837779,0.5376155255870283,0.9970042279753633,
 0.3096580433838908,0.4536862125063622,0.9779677184790285,0.1762267007004057,
 0.09620675640653875,0.44339667498360136,0.5838008437548329,
 0.7943529764261061,0.39320886399448574,0.31326380691610234,
 0.14323803477521868,0.06980831510608498,0.5232495814876557,
 0.27655149361334974,0.26011772934493416,0.4811043369340031,
 0.28009390241919563,0.06757626391537808,0.9407871086108285,
 0.9056058718761562,0.5994851809987851,0.9498730060473869,0.27640447174072014,
 0.7247877918722498,0.4005346948988667,0.045130698192676455,
 0.23837591867762775,0.6756610735625415,0.7512305976534273,0.5234927373482902,
 0.5022895939965801,0.9708401060817885,0.4064906721809183,0.7637041208974914,
 0.6241787523273536,0.43921861510841964,0.8403666383108699,0.6559279748172542,
 0.12818496532315893,0.05730693924163355,0.9343120065918139,
 0.9714078339297163,0.8423729045083811,0.07245670792292147,
 0.25048962969972477,0.5316208650310914,0.3208941192731565,
 0.47961474193224507,0.6580784803734174,0.3574208887646325,0.9492960752527463,
 0.5824771880911767,0.19898224199233427,0.8513558713128265,0.8494436603125295,
 0.4252929222250251,0.5163979410704734,0.49489444477506495,
 0.24855974769396505,0.2660219955088543,0.05346599340172409,
 0.9030680144518122,0.5865230856305668,0.22173628286014924,0.6802716729985382,
 0.4202178335882256,0.9373690220623225,0.6361305467225065,0.6932666282844888,
 0.48177721876396284,0.9705115493184322,0.009123195477678259,
 0.12410138920964331,0.4605634181724888,0.9494410027683948,0.7984371506129907,
 0.7412052230045311,0.9655164357158953,0.7153343704467884,0.3079896631461352,
 0.6723168284105208,0.6553192351071697,0.5788124586672972,0.9737466243005105,
 0.04344658902106979,0.26286027747437357,0.19178366742286335,
 0.0441557931226364,0.060517075758715455,0.579120240776684,0.9298457049367389,
 0.5490343694172765,0.896819442079845,0.7267639282205113,0.5190450880168634,
 0.7961117697715561,0.49539337572370157,0.30068589286079683,
 0.45422589351524345,0.5528831225414692,0.445314388561295,0.6847893755047061,
 0.7750082407261364,0.29978666393206954,0.8894792305473188,0.5444257333497958,
 0.39096482240811736,0.5069409901931146,0.9913228197882852,
 0.01548541024751382,0.6329256705911653,0.41374917627203056,
 0.1400941596413391,0.6840315924414464,0.8977741030816412,0.7465750538860911,
 0.4054571329855522,0.9692961094158349,0.0031269178326042146,
 0.5394260242626421,0.5096938325585882,0.855205448429224,0.9415277880762761,
 0.46868256210445747,0.13728960207098062,0.15894057216087154,
 0.14071815826517187,0.9815686022792665,0.7016802858948507,
 0.09622329536856411,0.4687029180587411,0.48239462413881795,
 0.9188141379126977,0.49253543470653127,0.6309267724263959,0.7554039684000485,
 0.744928511244243,0.47517738133718956,0.1980983948626993,0.4352241383188954,
 0.37382200616929007,0.4633697994425475,0.7794276420765364,0.9588509502531609,
 0.10447362802722104,0.643535195399225,0.07139412014988122,0.38170545120552,
 0.40252826495982863,0.28173913771131653,0.3490141557344282,
 0.7889667914816647,0.11223712184523627,0.3209988036665322,
 0.06181165977270264,0.5862618521757699,0.4152222456024133,0.9742505620263434,
 0.1829232729339576,0.7345710236264089,0.2609446836389303,0.11955523225331188,
 0.38916343847084356,0.35778558565485463,0.8029696428288193,
 0.8449178033272519,0.30757354101330237,0.46266795937512095,0.132100438800144,
 0.41932011107170686,0.39351221839638983,0.0687217358737594,
 0.3724488940299122,0.8377736835172291,0.012226421047904612,
 0.4455730928681233,0.23394242697839207,0.5605781960523062,0.6326391124023958,
 0.7070490938936904,0.026005966041061646,0.11806788527925205,
 0.6592238246703824,0.0806112198938701,0.7207941168352097,0.6980627112126233,
 0.997646384850255,0.8642876493774777,0.6714937675286432,0.06405236719325791,
 0.2812470899070416,0.5746102520836532,0.6402110245724119,0.28784142624978226,
 0.8724209649099763,0.48652279634386364,0.14255760818556296,
 0.5863162853875651,0.8532716872979585,0.022388373258373395,
 0.6727304762771301,0.6914024014022491,0.16778161305606687,0.7488792352456725,
 0.49828895185435584,0.4787883348513392,0.6361002990868142,0.8006157602050915,
 0.6653058742061458,0.7244597860234112,0.4282636388245271,0.7543052840388245,
 0.08594520799421967,0.3610056659847487,0.9471308318971583,
 0.19729312134031418,0.494838796535072,0.851721357140385,0.41627815796897416,
 0.16978410052865822,0.605540148793764,0.9082879090419564,0.3625153500817615,
 0.24123918249368625,0.5160198588964697,0.4424826794958634,
 0.28335732845755235,0.10180102782031641,0.4057455900382281,
 0.8579174742939443,0.4812379740959677,0.524962480410365,0.05460884574272862,
 0.001959119332430981,0.7613527752730617,0.5095202811154167,0.961624886434256,
 0.00703395706167953,0.3245003794615974,0.5594878272796129,0.5360865802121447,
 0.5810537734295169,0.5640745917290192,0.7965015382852766,0.7166406723601222,
 0.2654980459316332,0.27712471144762996,0.7253441469479815,0.7220681812148012,
 0.5750493472193597,0.48730055994057075,0.08809347728516526,
 0.5516366873962826,0.11189349375472568,0.40315060071035946,
 0.030160336304350666,0.9263744459557713,0.12496286950179891,
 0.7411468929149947,0.9028450724563867,0.5517977401808015,0.5671852275999738,
 0.17283919517263735,0.43164154662383725,0.3973152806322898,
 0.6970650637971963,0.22999659458887067,0.11952303203400938,
 0.47284571096489225,0.30163115563300635,0.40533542532465505,
 0.5417999823658359,0.7982359472601555,0.18190001895322716,
 0.41699970468101466,0.914671791732516,0.39367460118778663,0.9951412248708211,
 0.8812501123097269,0.536490347957273,0.7678467120475977,0.6313957128462323,
 0.5255195321898607,0.9849142337450062,0.6634888790640335,0.39908784565281374,
 0.6617426719501611,0.37032309813043496,0.9154137009455581,
 0.46013326351059747,0.44569213737756896,0.35301404168843853,
 0.3374423537931628,0.18769803652936734,0.017820820127174786,
 0.4446282183439766,0.8563797454332555,0.13970329410472704,
 0.11527966746240503,0.09183978226619971,0.5338301093455349,
 0.8253154995606997,0.032835842328525544,0.1662225580138933,0.266701220098544,
 0.43362049057348717,0.9245814489142684,0.9345438337993819,
 0.07643442054816751,0.9966833265070394,0.11676920867177076,
 0.05775250253969488,0.5685143890544895,0.36686531889459484,
 0.7623212331635142,0.0441493981829475,0.7093585957116506,0.30620326788168173,
 0.045632011232397574,0.38047438977687387,0.5957889304855108,0.2562709838131,
 0.3462052458871314,0.17862228619329068,0.09465116168558563,0.538822862386005,
 0.4318579221874227,0.1621421596188467,0.6769098135615694,0.2320865385159211,
 0.6057513414057936,0.030043464152309074,0.28150673905869406,
 0.9732816579281581,0.29252352036266316,0.6967813724989217,0.7027239101528323,
 0.5394770446656294,0.3535809853084513,0.07656521347463441,
 0.03112563463104301,0.627416353116536,0.2867791398062197,0.26509886030792906,
 0.4735512711278229,0.24632515277215816,0.9476839018801959,
 0.005460264620881228,0.43549338376241264,0.17706907065592947,
 0.8107368523355802,0.31126249142657947,0.8796701357038863,
 0.23760305712944496,0.9819274462949262,0.866415084870604,0.6923211091902808,
 0.012556053788810972,0.16156480301040277,0.18466184340493108,
 0.5424204597289578,0.22782675189057433,0.49959746908530345,
 0.4912193596659735,0.737823787097757,0.43939014898402395,0.1430316701158897,
 0.02293016970949513,0.6313912407559316,0.7436741599399463,
 0.36117828439951094,0.4640367735794896,0.6553036266198675,
 0.37596429740912574,0.3923805193622677,0.6364053896746664,
 0.18397349732594104,0.9109885510779134,0.964551490995121,0.675175900775864,
 0.8644251005591138,0.1618171945594986,0.9048293259441647,0.6198177670409317,
 0.058396073059739706,0.38542354448580374,0.16462157726474125,
 0.6651723772818684,0.8814148762329025,0.8020584762594428,0.4270600497300361,
 0.08558273106765368,0.6782469989055206,0.51738997317772,0.22937780308588818,
 0.39567793007359997,0.5198941470111405,0.8857273048032321,0.6234468400356831,
 0.5627575065525516,0.014005928794749711,0.4443913914962214,
 0.8061717583661339,0.9547661342066792,0.8078414332837257,0.9542713515343524,
 0.6953667804268503,0.9591899793099637,0.9249492064231895,0.12151674247395294,
 0.7472621463558392,0.4963831738344435,0.8606637541059275,0.7733801576553482,
 0.06430002987148176,0.12771163237282568,0.8688399088940827,0.633095458784346,
 0.6561223012604254,0.5199741304777546,0.2862205694787531,
 0.004590366155739378,0.05990458881346006,0.9050140154052447,
 0.9780493002346011,0.414001074789458,0.7539117776292741,0.7893751411613992,
 0.9506135570608335,0.31472438701631644,0.8553633744151352,0.5868486518613866,
 0.7157602044118073,0.1390678977630304,0.9708620799058896,0.26860640867108854,
 0.17794253811702043,0.4189765407471564,0.12116233860317038,
 0.018454820105628378,0.8617362598056297,0.9638539242152003,
 0.10272194686204039,0.31250957342861785,0.9094676521905398,
 0.05611135570381376,0.428732657517355,0.5288237857175424,0.11490314162533899,
 0.29026913354959794,0.29426544474605476,0.7650648433609836,
 0.9367363595760902,0.4953224653853133,0.9493149483609367,0.16335346456021838,
 0.6818635262574575,0.03627125593059688,0.3173924457025149,
 0.09718718628282519,0.33351745362943797,0.5625137777151389,
 0.5336729369462038,0.9456392110582797,0.7980784519447007,0.1095188746300153,
 0.261674539658572,0.4222108522628921,0.4022457532829137,0.3586678179173787,
 0.3822382723257677,0.8468972254312377,0.029591546519130874,
 0.011362423849586079,0.490430514201865,0.718260481516287,0.3003369167157317,
 0.9822498887894976,0.8054579971688135,0.25969309963848675,0.6721151744743098,
 0.6027574579389496,0.8258091803026957,0.7564031473511039,0.530324793217428,
 0.734734301822752,0.6286202789953494,0.6626590394826396,0.22340019254272137,
 0.32954694812339225,0.5349322786796769,0.7729538949094549,
 0.05654028882482243,0.2704887971644613,0.09610726224025334,
 0.22835871977125466,0.5835187670127693,0.3354732213011048,0.6247159318666542,
 0.8544199353764704,0.6803900485666211,0.16267637932069068,0.5385976334099675,
 0.7689607578059675,0.5742162648476792,0.24459104728807746,
 0.23272773762767163,0.7923991294121986,0.769944980716987,0.07411821528156592,
 0.5767583245581205,0.6393641218049422,0.08037676595755172,0.8511044884800684,
 0.5499980138024253,0.2866017805549892,0.47724499656852604,0.7499774651157822,
 0.7914812013337659,0.03098511329900444,0.9551818770632581,0.3760482123357125,
 0.4584343711710579,0.36380508265170786,0.868733334397236,0.4051483863036246,
 0.9208933151048473,0.9446846474808547,0.04262999337037521,
 0.37965441461374894,0.05023063223794244,0.680074198470641,
 0.11282252259170789,0.3412085066189966,0.2561320899287224,
 0.09989611445181457,0.6832738126600528,0.014271752638025492,
 0.3680819986115369,0.5624638764757102,0.21275467366731082,0.744397166768045,
 0.45206604856237265,0.218091936596291,0.5506422020773167,0.5066272390880846,
 0.3077944270741283,0.19872348013060037,0.9928830863181616,0.9175795135569976,
 0.6355605270498589,0.4633962415975108,0.20967507288973064,
 0.10595614287962163,0.24137353212237822,0.42345282740540624,
 0.9839650069108485,0.2988948046962131,0.5555540469974746,0.6584720052969338,
 0.4008164699602761,0.8494176225375275,0.7214059993349092,0.8058479157874268,
 0.788553539907808,0.6732766135901479,0.23758524310129792,0.3194261582030934,
 0.463369291914441,0.39725927791880056,0.7340418874416189,0.17543291361615632,
 0.4582341402665653,0.024251190698107106,0.5590785598777014,
 0.7056796045040297,0.42859074680673837,0.004409230011243825,
 0.769208108780638,0.8973899252022544,0.0963528544627531,0.4067621920649547,
 0.2882667459136521,0.21195050336244137,0.41688266025585274,
 0.4631983285332676,0.09060397010890187,0.7815974261461972,0.8109675713286894,
 0.6528516871790296,0.13150025203433335,0.7127178859524756,0.7688207417845268,
 0.13543332111510276,0.1581645248460154,0.356885718061822,0.34459928162030706,
 0.7981710877972863,0.3152400370251788,0.06298153141195739,0.4741508996820514,
 0.24289087723630765,0.43095666191697746,0.21924347303080471,
 0.6986288800964912,0.04509451882540594,0.9215126158336433,0.1583268463158345,
 0.6780926638493732,0.6440668285166548,0.89364512185149,0.4349872282491738,
 0.3600257910578313,0.11799875299390761,0.05857175675068971,
 0.5841213283365554,0.5493083842095359,0.7088257399916554,0.5845946460549344,
 0.901888961581317,0.007968852225058831,0.31132687555885097,
 0.21205520050567606,0.022603079331881437,0.7486957676247598,
 0.4519674090192828,0.3794060555014547,0.05729189900917098,
 0.26949488717270154,0.018280560662413015,0.012934780069468244,
 0.5147762192908093,0.1417832976457848,0.19095926023182053,0.4560887151522677,
 0.9531853564233825,0.10600699173613926,0.43245313853526834,
 0.05434469012534504,0.6487232565553047,0.7548327494490035,0.8007383012287741,
 0.07065122455016615,0.25236138930622154,0.15747467832724882,
 0.3374357992733781,0.8580935344578792,0.8453531495862463,0.6089652015047248,
 0.20370727004642275,0.09846788753527114,0.9031884186856856,
 0.03508349040406089,0.22632796309602565,0.7980724317940516,
 0.1417941531241359,0.9406574690551336,0.6147065345733635,
 0.056751881706142426,0.425116675814717,0.2195523315572786,0.8890490637590731,
 0.988644818347293,0.8784869287512045,0.34554994233037606,0.9369089802490074,
 0.2508540513733871,0.8027604837501403,0.4353394549023867,0.4578336025560631,
 0.7581727962988454,0.9119244445483783,0.9373967860990494,0.698308493243167,
 0.9313717648481781,0.46881479693429384,0.8204592328596251,0.3143078131161463,
 0.4211181481699675,0.665870944514531,0.2534404437914459,0.1661951035405138,
 0.16772783254565082,0.7622150218911701,0.8580335685426035,0.5546838576895808,
 0.4269833806177856,0.9201624322643102,0.12853970625307787,
 0.051606861649967306,0.9667803606032563,0.8981947916127526,
 0.6335345029575241,0.6983618059851939,0.8773662120475677,0.7162666289934018,
 0.8528597805908997,0.3355461484040638,0.5464890270414863,0.5244030302841557,
 0.057321345328751105,0.409429628796085,0.5810954779030469,
 0.11720407614059047,0.18295360615005574,0.7998242266385099,
 0.8255097106241817,0.6538508762795069,0.829989307721741,0.6409085570350339,
 0.7690947639675612,0.7495437698769906,0.7418127236844289,0.18744407595525725,
 0.19292034382724488,0.4179244569129007,0.7741156960739699,0.4237766422494712,
 0.8651542568777897,0.6745909590536878,0.9180087435324248,0.9742847471476125,
 0.01336136558716905,0.5009141183062356,0.29191134295982424,
 0.6509587142043283,0.16954592033823335,0.369957759531201,0.9999495563129072,
 0.5837422131112333,0.8319263561614697,0.5132836182291904,0.42088327424803895,
 0.9743283593554285,0.14477220888789166,0.47411939441664186,
 0.6439787853768125,0.8076753975108151,0.7042115475865984,0.6148194894150488,
 0.22811473773600277,0.22217360985890777,0.20003164212779234,
 0.8923889692524608,0.2122704413127665,0.13520171574305007,0.6131169224814395,
 0.9735410554368875,0.48442642491219967,0.8587852182187481,
 0.49046301245228985,0.4349527964548462,0.9769368210949186,0.5625667965513951,
 0.8805292377931311,0.36977714949969287,0.7665187281565045,0.9217227031259569,
 0.5645381051455833,0.6717643025462073,0.6914367502886536,0.46772578729783176,
 0.24157421876118423,0.9164059524467588,0.164852131903793,0.7577238928472629,
 0.48250617486042857,0.6115511901698323,0.5942405550444171,0.9197417481420073,
 0.7473071775547209,0.8666625351861394,0.5869285165326705,0.19445378725480245,
 0.18851866420938856,0.9658619608671166,0.524259006122747,0.813554659849323,
 0.8981701025330041,0.9314737211243349,0.16660331349584623,0.7051109025641927,
 0.7235627056832192,0.09278349123009122,0.9717314173303918,0.7453166140968197,
 0.23549327801761744,0.3080520923146749,0.7043300163937508,0.4151744570837045,
 0.6621863138372571,0.6219549992935579,0.07548941020647948,0.7158453754160553,
 0.3124723609499821,0.7102766695074132,0.1171249549065827,0.5053175090391688,
 0.4570577488075561,0.1888264615299281,0.1754291261136648,0.9798107248707187,
 0.36322707584637204,0.5421028501812158,0.23227167788991554,
 0.2696214122038233,0.1436871718194288,0.6901754448988566,0.8398203628315233,
 0.9451154046244034,0.7234021869418198,0.3018644015803126,
 0.010573759827752216,0.1268372523709298,0.8631026149561591,0.329608045914515,
 0.07378973052608884,0.150360863715282,0.7742544771969877,0.27570756115361983,
 0.3416051881643599,0.36635605507084656,0.30597809319762026,
 0.17054418127156679,0.6673368748435259,0.2918717574047781,
 0.07964216440964333,0.29840645579152836,0.855828024001348,0.7148252808772251,
 0.06381474336440895,0.5935718434661819,0.27377544109209095,
 0.1529110203923959,0.7345896752282307,0.5899446358168778,0.9138283902991119,
 0.3545997482194043,0.9815388859651437,0.23459534177673702,0.6343150383562745,
 0.008665279365759782,0.3872178272276887,0.14194659929475384,
 0.07945030663370467,0.23088610694382827,0.4390064180602866,
 0.6400572521323294,0.17760906566326118,0.008607687423529464,
 0.2831888175453403,0.9201326520877118,0.6474462094242089,0.697862619922511,
 0.9443295935936532,0.33948664651786054,0.08061657771206587,0.805686268877275,
 0.8689258350683288,0.5245344025217179,0.5027976286374175,0.41967374196539464,
 0.30645119058840886,0.6377661268400299,0.3400835022623575,0.536978227088767,
 0.026036095976031293,0.9208642295805021,0.19843797352349013,
 0.3175509805980251,0.6577390816131891,0.20854978882913244,
 0.47703152017469796,0.6918514221267804,0.24253771287300785,
 0.39451020798857395,0.9402601017720966,0.1788345818935999,0.8593141156111034,
 0.15073717646622453,0.09483819749390687,0.46867241124680514,
 0.7119863698914438,0.632014316170684,0.7677943087594984,0.22448317053981082,
 0.22021441674346431,0.9679632990206637,0.47933410822481026,
 0.4549085527978799,0.3672179012470703,0.4639436040094741,0.7011418251451268,
 0.3478928469258351,0.9591495891614554,0.8018507219312982,0.4702753221667597,
 0.5895470268462163,0.6383520062504293,0.7014468995118431,0.8458836038636396,
 0.4621567677161441,0.6002142633528348,0.7250462973354291,0.1319419166125737,
 0.44382900213963317,0.48485846243631936,0.9554142383244973,
 0.40315706671872986,0.5964753252082053,0.9520915209536972,
 0.19630927540186716,0.1958003054273686,0.3994320379858711,
 0.06471672848908194,0.6737314608739929,0.3033500674956324,0.9871638139103158,
 0.886722056912075,0.8588552374180524,0.8767541699079506,0.7077956883520626,
 0.784966038629362,0.7455187733455969,0.10337549875195928,0.6531533472919413,
 0.32929294356170913,0.730895192347188,0.0689127355447805,0.0721640420216052,
 0.17486217583538421,0.2051752661452071,0.9620070479768893,0.7376325874012417,
 0.6411356480033055,0.029041137817182983,0.965208898041619,
 0.09518391562557393,0.1364284735433151,0.923860166918328,0.7628986127554552,
 0.7161143921435299,0.7251743878993573,0.5940727196227318,0.2087573852584611,
 0.4080935206471138,0.7247138456293751,0.5194024499809513,0.39732451510750577,
 0.8486831884422398,0.702065763944089,0.2735760574999522,0.34581216643468504,
 0.8979524281301211,0.5549767015811692,0.7576562768053884,0.636080901953788,
 0.13178407467672737,0.1399682171582728,0.980325993527473,0.6864736558933833,
 0.7981769862953851,0.7096824723815093,0.5681711774359615,
 0.012906196877757625,0.037341463515943585,0.7577406591670081,
 0.5899231390887091,0.3847915894866437,0.8131679568695179,0.7034618982993875,
 0.5390242405318524,0.32412514165625284,0.3733006949171369,0.7046546895606037,
 0.3122429923425194,0.8514507798065815,0.8360713767552137,0.5357400745762418,
 0.6169547535160104,0.2273173273066651,0.438632647895997,0.27013655608842246,
 0.6110910085003977,0.317039958706427,0.008135095952609306,0.2414337537844522,
 0.705504496964067,0.2768310624099175,0.8361034614367661,0.7875052179316666,
 0.7026863275952405,0.6302735769559576,0.8174355949969974,0.680936137597546,
 0.3384707214620494,0.7541252543523247,0.2929586809047816,0.715513553579107,
 0.2633243664476558,0.7524838632264126,0.5899709696164194,0.4094120777484809,
 0.2581018416206392,0.559594076867211,0.2243944593981665,0.2719830267769052,
 0.35534111611026475,0.9046819812823782,0.09073961894533666,
 0.2957933063804169,0.5604896334405839,0.214201717346569,0.47647522089630023,
 0.2012805140227123,0.24836910693576186,0.05612823752813678,0.499324700485942,
 0.8470662857542733,0.011013473264067764,0.8787987593073027,
 0.048724798740361175,0.4464627260129944,0.5350850567315275,
 0.0979657954483808,0.16591133454822038,0.37269498527826306,
 0.9821840992314458,0.27445077608186474,0.9852235294206804,0.4783866649831634,
 0.5654901978950423,0.16525869851010166,0.9040866297743776,0.9513520907927786,
 0.15395800139427185,0.5013638012471845,0.728498489108882,0.9175664678686977,
 0.30408303878274856,0.4460008998828021,0.12693726497700308,
 0.12773695737466273,0.20098345615521607,0.49375594417487834,
 0.2013125568586328,0.23020429317152102,0.1938247091196853,
 0.03762978008831952,0.3679612925460034,0.5311864866367253,
 0.013956527075718261,0.12958021903743633,0.04650814203939291,
 0.22954394362783825,0.953422938883541,0.6610193448549853,0.18101057178036561,
 0.1129233787881353,0.4564827591518097,0.3448652418363022,0.4805364975413785,
 0.1998582122483038,0.5114218211808578,0.11884808280934156,
 0.01430717030987283,0.778315287377228,0.6591569876628821,0.3200535210411375,
 0.7494131616409196,0.5500212006405708,0.09291224474691884,
 0.012283227959248633,0.2939039756501025,0.8581792569369212,
 0.2985675754949595,0.4669079201755153,0.036758478725349764,
 0.3658843916002973,0.417360888224104,0.3655758005777415,0.956254338552059,
 0.37193318168344847,0.7318265110424159,0.6888889833883056,0.7374639434899083,
 0.001975449607179103,0.05157280732820113,0.755395666295716,
 0.8394936018259515,0.019428109212779487,0.09150317844841616,
 0.39426148286274043,0.8680344142114655,0.5402286367550846,
 0.28632516358530113,0.06994719776204128,0.9409739765376177,
 0.4600186898698927,0.698523848770654,0.8564057307173419,0.5253169316717303,
 0.0686162032238059,0.7129097200509251,0.4649031043828241,0.16346097571953844,
 0.991778012067382,0.8518993519438567,0.4523879333841401,0.9730817433271315,
 0.9494368037709338,0.6026201495741955,0.3935505695710454,0.7000126440654831,
 0.8421303958375734,0.21572668494869252,0.22686786304748618,0.506792256650024,
 0.47831510014826417,0.5276209850338162,0.6432903091141201,0.3808159103810441,
 0.23036704923521445,0.9805910308545135,0.3030676979880318,
 0.18798175858069588,0.8284150559668163,0.45077253294589426,
 0.9244427735720236,0.8832988973035415,0.0650574616176458,
 0.016093619129777137,0.3887802260295008,0.38837748015217355,
 0.4640043018312042,0.6711751309473446,0.9808200226538588,0.4584011237300972,
 0.3572627374996936,0.893258222840481,0.7568235115941497,0.4189360608657696,
 0.16546918026991053,0.06881543989520478,0.6941595688871068,
 0.6399252213648597,0.5767213630761783,0.19218206812592076,
 0.17473727369078718,0.37822190754169893,0.9383299702107082,
 0.3658452224970271,0.09556152209391244,0.6888813172034982,0.8851825252193046,
 0.6332212225025965,0.5632431710198161,0.9426443511877863,0.8517956412238934,
 0.7510565961134197,0.34695074337331333,0.758862908059639,0.6699746637424743,
 0.5511442632717908,0.25397001708437894,0.9661801258062067,
 0.33543122235635425,0.2656348050165145,0.45621366996009904,
 0.43751315447650585,0.6388900691716798,0.66219796535889,0.12044598901383285,
 0.16723939809265764,0.23000480708957272,0.05064785004986794,
 0.12066199205126427,0.3168666845793683,0.9836901980967102,0.6094511847908047,
 0.9566446964989003,0.08624781362894152,0.2644243544690562,0.9482331864173176,
 0.26196652010722343,0.4136645823646543,0.26819489872902647,
 0.1501297647120774,0.11701084301384601,0.008686803650937769,
 0.13322486433597625,0.08067915836010975,0.8744849361408817,
 0.6927208630841859,0.3863676282134447,0.32690966956974543,0.6449720556341973,
 0.8433869974982137,0.6850452600756768,0.8706055146659664,0.4246463645419737,
 0.5680461204135056,0.4251915192538236,0.15263697706546442,
 0.029372572215565107,0.7772706494976293,0.4209304727835228,
 0.8065149131589231,0.09705505848418072,0.22769164739650494,
 0.12595312313254203,0.7897680947325363,0.40577665381746764,0.520223001506715,
 0.3495827593450036,0.3282741080075492,0.572901780433744,0.15925504730145668,
 0.8224117589583333,0.15313110261884455,0.09872017584013382,
 0.6953489025638945,0.9413662878811803,0.8263341107633186,0.1352589916838175,
 0.46855902072372024,0.3040613245980823,0.6704839685016601,0.8822544059929704,
 0.059107362602663915,0.9640163035108459,0.039026882261727125,
 0.5054036494115995,0.7091489243914069,0.9070084676695005,0.748303419648986,
 0.5952074200006328,0.22761342048536737,0.8266456050964494,0.6998154315004431,
 0.7752781200515604,0.2153995105637665,0.10935153431087219,
 0.35279607896694887,0.002195415704145942,0.49395947909750326,
 0.8859239609941398,0.28119944405779373,0.6919102319514444,
 0.48643193874278456,0.49383871538704516,0.6158196115820872,
 0.27787548039680277,0.21994355061279103,0.39556873182379304,
 0.6313245283620288,0.3677579618594674,0.9335521653670015,0.24434232976744918,
 0.5835304697058691,0.9214440800173379,0.17685410920363454,0.564352255894606,
 0.36709759729827873,0.617931762457727,0.2102391844712076,0.4052941432962983,
 0.25473861885146865,0.24396696814041452,0.7647309797815377,
 0.8591408948061581,0.5725168534036492,0.35398238369256685,0.9747815690177547,
 0.26126960290166545,0.05648679711636295,0.9740988715656376,
 0.9157394329256278,0.03388807095167157,0.8594035701884167,0.3878325750101076,
 0.614873946576423,0.4355285875996713,0.989405085242285,0.21827203827226938,
 0.5643431127275219,0.30501713642554196,0.712329876827845,0.19556866131179407,
 0.5927638717989028,0.8636756887986711,0.24727662825159913,0.6264835200038263,
 0.9280157327782895,0.5332829877876651,0.8250229708431239,0.5632897661709352,
 0.6099102814444713,0.5967640996027899,0.0507778194278522,0.06329576925244929,
 0.4440209199175702,0.43621070831286757,0.9434843441419911,0.8577231538010406,
 0.4179322017127427,0.17318473783765787,0.028712581841790374,
 0.9122479573083706,0.05573446846440022,0.8929201099152181,0.7806409717996017,
 0.3216658756255457,0.17938107721461083,0.18588025089132987,
 0.3404601108141847,0.10093828865456278,0.20196744087114815,0.53700218115749,
 0.10079149322437742,0.1771428370402941,0.03891451854409966,
 0.4375190633840804,0.8536934313380241,0.08610951672259248,0.8292743809630786,
 0.7663137446047932,0.13851710788352256,0.682534729988127,0.8650553945093598,
 0.6306514188083514,0.2524673809010499,0.6097338914965146,0.2820377474128548,
 0.634936369416828,0.01585491007482731,0.6799839790895816,0.1830080248282594,
 0.5199678382956117,0.9158642986425495,0.8544221821398554,0.80849880471395,
 0.09910666615103636,0.49613820140303044,0.33806217271357286,
 0.18863548400955776,0.5354818492685788,0.2377693728923007,0.8505556657691725,
 0.35661981990851754,0.8799003030799152,0.5466629296093125,0.6363643856576837,
 0.3012524136475938,0.28566640843984203,0.12502083015699994,
 0.8543536994777652,0.6701766724255498,0.36044802694746114,0.6934300524885184,
 0.7147635139724245,0.3133881848962061,0.42126574730399435,
 0.23521849459523458,0.32007609704351947,0.6799174121878804,
 0.28378507124709107,0.49876018771528796,0.9393647466939802,0.329368050282667,
 0.40069383596674857,0.908095122825343,0.11419445343238699,
 0.031432142713677114,0.6827942074724014,0.6483832204132176,
 0.5957903930470915,0.4264520136843778,0.6831016408609348,0.8102615028223303,
 0.43115248464954126,0.09421047966668838,0.45446011247019213,
 0.47130352957040245,0.8619544665838802,0.05163739448178495,
 0.14179886603854452,0.2947641804110721,0.8737093370237752,0.6544683980502028,
 0.21368609406774408,0.6520754975905038,0.858615721582009,0.18180982838833037,
 0.20248390633300406,0.48438669067559703,0.18238074524660597,
 0.6195104398041051,0.6714605319690201,0.414285070553574,0.42565586523080234,
 0.29435512435614997,0.5782411507648382,0.25966597264389124,
 0.8838300111645935,0.7075262969243228,0.9951667243295127,0.27529132069840456,
 0.20931641910195076,0.14689188935979147,0.49958415838620074,
 0.20534733008703132,0.6307401620446957,0.4327337375024234,
 0.20881096199652294,0.23920252715842905,0.9453494925096141,
 0.6659675996535376,0.8989285290108413,0.08936617678570824,
 0.24523849065755798,0.9940562488086512,0.6976454938035829,
 0.19777709787649256,0.5473386895129628,0.9034260152307747,0.3452071306418767,
 0.7337889232112903,0.30318166000298497,0.689421702788529,0.9895108844185719,
 0.6910048555307676,0.5710962318119385,0.6136778390874174,0.9556171768988454,
 0.10171056347622764,0.7858075011570059,0.15145938236420153,
 0.009107894091414659,0.7849758990222577,0.4841206296573005,
 0.40520971833793784,0.43047607543516586,0.7905915088197663,
 0.8360245634905396,0.20638170409698164,0.9639680125448975,0.3602940244589008,
 0.7262432259647958,0.346046602193569,0.1860384394153516,0.17744709370068834,
 0.8085059695865823,0.038512577227576394,0.2865490512008865,
 0.2665075098430668,0.9209362270169433,0.4484411053651972,0.46166475254061934,
 0.676257435235266,0.4334257753011487,0.5058485300685024,0.52522528681515,
 0.6769375319048618,0.65093557041894,0.23460108772738542,0.7109339729996416,
 0.019467185311353385,0.5488581815831228,0.5842569011383618,
 0.8662439999413185,0.4661702581167678,0.6695542376250315,0.6370106648111563,
 0.1041903619110669,0.34695283313246605,0.3292050908515626,0.2870986941985052,
 0.26407816125214667,0.8867864722740375,0.5344838760972724,
 0.04388151520697592,0.457358792301256,0.86261747253629,0.24226675769180772,
 0.1844881745461424,0.2938553908088952,0.708451860206849,0.5340525646350969,
 0.565114127511902,0.5875281432361611,0.6054459323567949,0.12849679809526005,
 0.3508386107246052,0.9056835101878914,0.5491724079789988,0.7865903754079976,
 0.24735237140641297,0.8703219546510081,0.18688743429261045,0.875651822813494,
 0.8452801807528857,0.14755525254451396,0.29635493591513484,
 0.3317970801828818,0.22532675639594213,0.2050284151264915,
 0.10979827153639576,0.18757130109168663,0.06386820487743972,
 0.6633150344008407,0.4826946958682887,0.2204687297066601,0.49373390774968196,
 0.994466438412584,0.3982743594287701,0.19412725304916623,0.7890514365491834,
 0.059146560029148976,0.9388099278607215,0.36312216119248464,
 0.07634925639368206,0.3286066828681703,0.7347586873068296,0.4775130542226871,
 0.7050253837879916,0.39330580032437457,0.42212641898594483,
 0.8009118179787612,0.9453428476874612,0.5768123095859045,0.14511107941311074,
 0.24347180918783984,0.07129071196085801,0.10505743322558614,
 0.4686766105849788,0.040269731417561405,0.38479795581609244,
 0.41292321489069705,0.6333503712756465,0.8864431431448115,0.947740592476126,
 0.6487224817817804,0.5012221324487295,0.7630826089867602,0.06787574217793413,
 0.7882177523584336,0.15377027533579468,0.5921251706842894,0.73924919196374,
 0.5324929145887783,0.8417612832423019,0.2866912729322516,0.1327486532397012,
 0.22631817311682445,0.4235493618961379,0.8485980208730324,0.9316909815164589,
 0.8750342232233481,0.27237173523182345,0.12052318170568488,
 0.3144400472106538,0.7960385039530768,0.7691835916413703,0.46650638506623565,
 0.927146191497136,0.2233146958926543,0.12922620512182692,0.9809750400679512,
 0.9118766242764018,0.025294416145907994,0.04525174156188905,0.17126298612516,
 0.5057334109933298,0.7845917497464466,0.17764126044025064,
 0.09374109796244312,0.7780544678863506,0.1317355363187065,0.2070922828847951,
 0.45631493108710175,0.37128802772899316,0.5969204539589373,
 0.18423455370889197,0.0586857815891797,0.5904390578579778,0.4872590067116197,
 0.7438570414005566,0.9412004351309421,0.5760788408089097,0.6059855947113044,
 0.967731602955801,0.15973497392294722,0.04277275409085057,0.2870121740392342,
 0.9302314560562688,0.7283909441400666,0.38984289175720865,0.343856734595095,
 0.606624762961681,0.20552733088399655,0.5551175433142614,0.8338185996345174,
 0.10157852877047224,0.9909156206113824,0.6687459217883824,
 0.25156667273930167,0.8102189610762691,0.2709153221030848,0.3827448952977409,
 0.3518243774669535,0.11756941844023228,0.05729788159905558,
 0.9583686153154568,0.6019663770266376,0.21376151875537264,0.7071817223514135,
 0.8167620341390345,0.7204686352245038,0.6700890648947536,0.8686972262526257,
 0.3388161418935628,0.2952875579742038,0.08611468079228068,0.10577920701873,
 0.8587310749383505,0.06990723592980008,0.41117033513926693,
 0.9078858995195738,0.5071330493751801,0.21061087712009863,0.837183843955054,
 0.7050132033922145,0.13644935686749904,0.7270032848439058,
 0.16810825459292744,0.40415533994505526,0.5853880432007663,
 0.24750001537365596,0.7230564984929394,0.9633033279770981,0.9117540068132232,
 0.9564545253745916,0.38013384661135063,0.9414917662550423,0.965812764020446,
 0.9856272265395504,0.3111359516686847,0.6277504396296273,0.8945122843393432,
 0.49811338378128034,0.15272478417629987,0.9526872097719683,
 0.7055435089951224,0.6367500814970223,0.20770961432028356,0.3392189130209441,
 0.1997333843997211,0.7848800753085101,0.24598241571818058,
 0.49015011031841205,0.13640895442542633,0.10811080764527936,
 0.04425289913514496,0.2685308300526855,0.13548434530538556,
 0.5829141543353221,0.223926497731762,0.784622205214828,0.21152065927490782,
 0.712633436691871,0.24110960844435614,0.11922636527773811,0.5743141211708723,
 0.936563111671716,0.4734012798214584,0.4764904324685184,0.6689637314451248,
 0.2359294521183004,0.2326545208424593,0.7608637845592475,0.27858277813584453,
 0.25654683468800776,0.51274513673838,0.3218192168773065,0.17547420768834832,
 0.19907695777214562,0.030729462606200908,0.382266168663151,
 0.005883212845604113,0.2374503141935712,0.7990511329087824,
 0.05375675111918554,0.6944727981626627,0.8047181606823338,0.452647067709388,
 0.3370165913935561,0.7611431170695382,0.7788066254854709,0.7723955319911383,
 0.471339356182796,0.2308058086972593,0.3294505818580782,0.45983321890232576,
 0.9536654352570295,0.49096658439205043,0.011729150906827357,
 0.44283825203520744,0.1966368654098002,0.3448482482138757,0.2376807958371001,
 0.038069834221121646,0.586948955896656,0.29387580241413824,
 0.007365701625125287,0.9910859395892815,0.5625185770788225,
 0.6833999151788636,0.8613683008758248,0.7733081134014622,0.8107694016466359,
 0.9148641319294192,0.26692250726099476,0.6631367004834011,
 0.06267218006706232,0.9453158110729777,0.8741531892306661,
 0.37574791494678383,0.610458839463557,0.9857224504848906,
 0.019406089461352027,0.10432515510924467,0.8846761471072189,
 0.01887734521823914,0.4232920192581898,0.8834319348848325,
 0.08517579520770702,0.4072842113842252,0.3001772536237921,0.9792689285682392,
 0.44976857740235476,0.16772840862039495,0.24485875930728596,
 0.7887117466362984,0.07525217375459137,0.4432246198193701,
 0.46510248008636756,0.28188820620118504,0.6732621907070213,
 0.5939678110337252,0.47008501801074987,0.8747764322640579,0.5849455138897378,
 0.10701721123086683,0.012549734508613408,0.17960604678090153,
 0.4233735580665142,0.16750041438946517,0.6572014201807114,0.805406106318477,
 0.7996906909917776,0.8102413238283581,0.9052060957012483,0.6183086743046582,
 0.9011087280527998,0.0044026771138949705,0.17002282129877866,
 0.12400507701911134,0.33008523031031756,0.5674239124947618,
 0.17295821099461928,0.7183621287985569,0.3522219961389723,0.8925889585146866,
 0.3719174797997913,0.9967345922656667,0.47236121876259884,0.9155671131958575,
 0.5531766627097485,0.4102125278137414,0.20941325481145,0.6857443211276546,
 0.760116741432246,0.30856548845870124,0.2202583972278438,0.8356959583770491,
 0.051320711645181594,0.15360406355883882,0.6421197608189542,
 0.9086731063288922,0.9822967168501541,0.6493011944707071,0.33391672905590675,
 0.4358525525586415,0.7324860849162041,0.11735644137889734,0.4054646810584952,
 0.7646435547792557,0.7582257166773833,0.8590084621900625,0.48899072818921385,
 0.2973109426096181,0.5579946887637883,0.2160717284496425,0.4701473459281006,
 0.07453579146168965,0.15029715092561746,0.012430640828690387,
 0.26906919436373744,0.23143917470556252,0.3155161905778022,
 0.5320238929065956,0.9846225740481633,0.953613104998763,0.16460975333650252,
 0.3850807100886927,0.11594002792315261,0.7020992607913757,
 0.46515213781107667,0.03569386675002706,0.2704895736538906,
 0.5308536763606313,0.48447565238343127,0.19959205734463636,
 0.3187074972531645,0.25891215304257564,0.10979219319366673,
 0.4038433231695002,0.8900902402731004,0.6857010053214525,0.16566129899361137,
 0.33601594578730287,0.8065296595902788,0.33704571395059124,
 0.08794839001161303,0.11454987940694739,0.3891691825429151,
 0.29890800174128795,0.9992631920924946,0.037625808389738724,
 0.9505158255545133,0.40062763153523273,0.4029323765892474,0.7920431587450001,
 0.5534879261557998,0.4431112672917862,0.5898957630601834,0.6625832542006176,
 0.22904002084344943,0.8763002009491643,0.328113994111932,0.9319945985942074,
 0.5573533755554407,0.020872023981972854,0.4117815389981546,
 0.4629267968489341,0.6779428559376588,0.7934767290327662,0.8246899355264883,
 0.04956855745349675,0.11601193277745248,0.08737304378572341,
 0.5774393119810423,0.4670266967801633,0.3571629859173472,0.38619669663392076,
 0.7179858295240877,0.14901096895312327,0.04322404160875859,
 0.06654493508762316,0.20311707298762194,0.9054641605881495,
 0.31061727970257924,0.8811058398298902,0.37503539835015864,
 0.5326879357990087,0.9132886165956594,0.33964119528515535,0.2542330792327041,
 0.4363216310638901,0.6516106863404322,0.2329965493180195,0.6100153922423296,
 0.7755973974804653,0.6579352140464972,0.8028769851653139,0.8521021998238562,
 0.06432175786012129,0.12153372044386335,0.5423489813459456,
 0.11283941062872427,0.849326947184036,0.8053230473981521,0.5484374327558139,
 0.14158098559583943,0.6265552130441467,0.3383957012598109,0.4650188016380342,
 0.8332989346078621,0.7604954255020835,0.7619017502998775,0.2378264096178233,
 0.15250126842927914,0.9332075163411719,0.8587442296220356,
 0.19517617960289324,0.8995217672341868,0.7899558357652665,0.8608288372298055,
 0.2122325129830227,0.5842307134785678,0.36146708253339954,
 0.40040918745788856,0.6119634985221513,0.8441575783921916,0.7685586451221447,
 0.5425507292310923,0.7247943600323299,0.8308603777209167,0.609978615828397,
 0.8559901234412711,0.08937458538183496,0.39509285588897713,
 0.8498620136247113,0.6527878583688649,0.4627732967954926,0.06366769839024067,
 0.7849477147277996,0.21206917661309188,0.659206892660491,0.2841521373828588,
 0.5036031994500643,0.9455105443923102,0.8848409737921421,0.2644617100050377,
 0.17679212212353967,0.14879821083972866,0.5970957475975625,0.344230294688991,
 0.7405925072513827,0.5305834334336454,0.914739692846969,0.4364437389199475,
 0.5500168952902497,0.5800377413978746,0.4741272962781913,0.5015046167963259,
 0.04243188940312825,0.6632770938260444,0.2725760390990277,0.8894188958506163,
 0.8415351586341061,0.11810122738003148,0.8703392095356874,0.2433520292253627,
 0.5228119571811078,0.23811528402525073,0.35853793334333384,
 0.9713297826859157,0.2072008782333885,0.3122087078259148,0.9363351634124756,
 0.9131049698138318,0.783447653579992,0.2860373686356569,0.41991166997703155,
 0.9930554271348179,0.7895803372859135,0.7065228176307666,0.16240094356222434,
 0.23850541802473102,0.8563087477507036,0.688684287227463,0.9341605182555944,
 0.4461341832629393,0.2631203049730274,0.40531174258565905,
 0.12188184338735641,0.646762077005443,0.7193752708711083,0.39021333364320976,
 0.5731544763440104,0.17369080130490777,0.27189138490054043,
 0.8892075294510937,0.791706261671069,0.9573356913199913,0.9260547251052493,
 0.10870632149577453,0.6833547348964613,0.19920819455524164,
 0.7881274807732241,0.3957619143214687,0.245490479032342,0.5235175270116735,
 0.46675557298237824,0.7853015990111265,0.7631275542272917,
 0.33836878126500425,0.22099699435641706,0.3507111704681697,
 0.0356491214118595,0.830912662347126,0.2054166681202766,0.07056553750741301,
 0.28686335731339063,0.37952538763585464,0.3932800260414884,
 0.9934552349968818,0.750019997798501,0.2081844808075719,0.8660976904499809,
 0.2465207510977283,0.9505216551606752,0.7524240728245865,0.2644740274607579,
 0.7110319315203694,0.7032151068275156,0.9882126030997205,0.7288020692543736,
 0.9027115927373255,0.5939605307412427,0.21538192573094517,0.9619386765863429,
 0.5489677243567167,0.2694985381472771,0.32001780788704437,0.4187966335174972,
 0.5035578386118919,0.44884232957847403,0.1838904550380258,0.43839534615957,
 0.5704649710145206,0.6424670867119647,0.7582485012926098,0.6191352028655365,
 0.7934795146670498,0.5850833689000157,0.2437549446389372,
 0.013065071556091268,0.16868808123998702,0.027072042598161428,
 0.3512328363999756,0.6142909941929903,0.5836766581254528,0.43124646066637806,
 0.0730186223471192,0.7517184768093487,0.1616537665287049,0.6728695537921466,
 0.6366779727944263,0.26691023899873756,0.7020385408067105,0.2948603228976354,
 0.2584876092975955,0.6147429576502319,0.7928328385401842,0.2735354155800601,
 0.9793130408490305,0.3797142866178289,0.910312648526095,0.13593690648695456,
 0.5137081204341647,0.1360433295160428,0.8580446279912991,0.4179022241401328,
 0.018294417222033843,0.7418426921741701,0.9664150376317406,
 0.8599470509412315,0.946195871990871,0.759376728625361,0.41839423771170114,
 0.6175524484042884,0.3166091666318044,0.801989534642643,0.05601971768498515,
 0.9657067252327977,0.2854300780207404,0.41268837326921526,
 0.08295504741275761,0.6421676269881302,0.43459561038011585,0.889857454456301,
 0.9726717221300437,0.5090523546739262,0.6288069517857352,0.8171763796926381,
 0.666958787993211,0.11054590090098881,0.4819786817596141,0.8894800905461203,
 0.49066332384653477,0.36712162687051164,0.8688556845435285,
 0.9194873388201065,0.25962231462933216,0.45844023014071844,
 0.4212429369048871,0.24243940597806135,0.6015226036309491,
 0.26088054787663684,0.5320458261839234,0.17473708391420684,
 0.42607924504076333,0.3820254184021402,0.7228366948965493,0.465723317594238,
 0.5727002863854769,0.5038127844939588,0.4414056178769673,0.1603522838619248,
 0.5706037726150812,0.534074748490081,0.4273905431618805,0.47773400376727837,
 0.15021009250205508,0.5186226269659423,0.7469946541368534,
 0.22354561523621574,0.5901934119492475,0.3580800052505334,0.5689558985156472,
 0.6030746674570744,0.318813262503188,0.22589801994070957,0.39934936680312905,
 0.7879067015788058,0.03930259849099682,0.2283530206298091,
 0.013443965463545204,0.005844838631144711,0.19551267505386172,
 0.5807526395884974,0.8890662110533247,0.9541348171015311,0.42611101861701894,
 0.7626363030171728,0.3028790279716429,0.9501347401331528,0.424268524968208,
 0.5160671543861389,0.4710016094677849,0.6192351072143867,0.5235144767861857,
 0.8507738036353616,0.25465632723262055,0.07567498782959681,
 0.20073891509415043,0.7505667166830684,0.569675850286783,0.797262417142805,
 0.4610758792475502,0.8098677791285311,0.8925069792883789,0.42487226903428743,
 0.9807774293753077,0.20115081897006715,0.8838976180996044,0.3384397489130295,
 0.09737108608167877,0.846876474550914,0.9065528173065788,0.8334347921293734,
 0.06157900476102118,0.580407517551444,0.46912781795528813,0.4542254938670057,
 0.06398716763616308,0.3670446925383394,0.16741898492249896,
 0.2862924869574014,0.34804861202144133,0.38117892445290347,
 0.8857788411227245,0.1448191002622704,0.8422296345666649,0.17752424339565298,
 0.10409876659363981,0.9119214576481017,0.23925856073862628,
 0.24390855960615976,0.1145792404113708,0.7125877099491345,0.5287610269445985,
 0.672531074757817,0.03203102518873213,0.325720571934623,0.4088008520401507,
 0.31535829763232337,0.2668809070594492,0.40115743106287194,
 0.9310095315863854,0.3590220763929901,0.6150717918214507,0.6133511103200814,
 0.3712459617324706,0.5491727056215777,0.67982449202613,0.5486328329685755,
 0.904539546873351,0.548117835119444,0.9680566192628018,0.6401745872296916,
 0.11038478867678536,0.48650633715624814,0.3926332435989095,
 0.18686669226486408,0.9516542145095739,0.7346243651094294,0.9491947615470745,
 0.04699723093997099,0.5261772346804163,0.9626203110168371,
 0.06068478636377561,0.5771423943050955,0.25755190497988045,
 0.1588590703481345,0.7634139981902758,0.11832121627626924,
 0.34390847401254565,0.8765229002574607,0.5134799478574794,0.4980324785643824,
 0.7608547157700515,0.7456765674593073,0.26567252967167276,0.669169299224926,
 0.9609101080649646,0.9607889390671154,0.9696701951768316,0.7751395795810881,
 0.390597473811933,0.9631506228010942,0.5442615763060539,0.2655368431124414,
 0.5302015401045961,0.01129207356632822,0.008824358081385597,
 0.27073478269479245,0.45298214304080775,0.0614229522996903,
 0.5378203925855315,0.8440469275736913,0.12051721388520154,0.3013118637508133,
 0.3772824068993391,0.43978468548558447,0.6951043566969419,
 0.22608680555467275,0.4488382278402613,0.19523412773167037,
 0.5160984936196014,0.24381124537732568,0.9316704328493226,0.9011712600676094,
 0.6423901064937659,0.741521242396948,0.7502752761802096,0.9297887043333621,
 0.6875117314777515,0.08007704453333031,0.4302049288693589,0.6018184032211605,
 0.5090002200607612,0.8122883088221838,0.2429598506667321,0.1814974404567431,
 0.7623887591619196,0.136820036746656,0.6749308204270281,0.6369513679322841,
 0.407332308956034,0.766760111302017,0.6573139063691285,0.4375136351123621,
 0.1151113360691638,0.9397515306497792,0.9973018221154836,0.7556038851224115,
 0.7417710781897757,0.2092742397897257,0.9240954995597969,0.8805869467956707,
 0.8006170827758571,0.1891156896347822,0.5565106346562503,0.49157480258255815,
 0.8375279142266269,0.4363357201268079,0.43822684157274916,0.8949481753180188,
 0.6163002230608622,0.17941775098590407,0.04129794795497732,
 0.3849342664782007,0.2399332026238169,0.013984786580953035,
 0.048479133096037685,0.8225257594291149,0.7224870149321798,
 0.4407518262067547,0.7210076099284211,0.6926847463682677,0.3719392569638924,
 0.9395038971910343,0.7622817854154019,0.3032648451200909,0.4519260647170946,
 0.1739297334336266,0.4156529225121701,0.3612402692367832,0.8472973610532907,
 0.25163897247446154,0.3215481404039804,0.546719721878332,0.739031813713063,
 0.21541826514320195,0.20714512872965718,0.527834580626555,0.7312600346342104,
 0.46057216917626376,0.6438704200585332,0.031890832160409,0.26214777397984623,
 0.42886228253470093,0.029204434469688456,0.15202326633466146,
 0.1322557237977402,0.0936503645606076,0.9994506302427247,0.8918733248298144,
 0.4524391068381749,0.19910158101414976,0.4319491697913185,0.2514393825631791,
 0.5499606070634031,0.07288895356760294,0.45688025060205173,
 0.8369235845514927,0.2864833008071276,0.5668122466520611,0.957649694094226,
 0.4027408037934621,0.041442388372055605,0.07179144280413852,
 0.018497071665023235,0.44893590775436465,0.6245574325180628,
 0.4195308384599338,0.4883375926221315,0.9914346084142597,0.5280694383439194,
 0.9034879253775004,0.42498017621713746,0.05740330769799973,
 0.3692255037755645,0.6278602386279447,0.3094955036627779,0.9660224054877873,
 0.28276485919883454,0.6161008027844996,0.4474871919714879,0.1350577969235487,
 0.8836727145044925,0.42555436087775134,0.9965770383225481,0.455255812796703,
 0.5941435846824564,0.07561363156376222,0.13573722994163373,
 0.28867866556273647,0.7383954951541616,0.25047564355786356,
 0.7796786475138737,0.9227459503549554,0.8688854995306163,0.18010118125476038,
 0.19712978687314875,0.6642453546737592,0.4376773683259737,
 0.26619118999923697,0.5146825833556865,0.8749028464874687,
 0.40138705822014775,0.9464041993125275,0.5439612134844056,0.4564019729171014,
 0.8695242544916046,0.6149140590667065,0.4446361005570576,0.2634900408039379,
 0.26849902217674615,0.5000625007033392,0.35858574672546595,
 0.6989159126084387,0.6031088316739874,0.5365540730888447,0.44535196436607594,
 0.3938075845771827,0.3838699206164221,0.42351069080634274,
 0.01661562359681057,0.25873528274352653,0.8361686765612024,0.511627036074721,
 0.040363480936434826,0.438077694973973,0.007487252650787979,
 0.5961174791742918,0.8414365307137677,0.7594795125480118,0.371136792916678,
 0.8182187626805012,0.29608935771487543,0.9677157781927024,
 0.07663344921851056,0.03374172904186956,0.9769167431638445,
 0.8623187014787066,0.10940695405062018,0.7896585783374696,0.769757102714257,
 0.288885099296043,0.48533137243056623,0.33634789290341927,
 0.23417159480245286,0.30680141835191765,0.5097012189321165,
 0.09300301641491016,0.41513961216435447,0.6673231292955336,
 0.42060725647361397,0.3867443866833956,0.8209015004162945,0.5675837933696579,
 0.01080626870827417,0.7872629653307853,0.8300245860026554,0.6630919036128897,
 0.34777115660883595,0.9151209384044969,0.6507694959122975,
 0.029799877174445988,0.9104235984199468,0.4522399325726829,
 0.9258985124544922,0.7877786669394091,0.7721499219623588,
 0.024657583519671178,0.571897662231467,0.1824785483686867,0.4867911428832712,
 0.8526345731358577,0.7887174094505167,0.20463251064706867,0.5909822533654049,
 0.9460858500463908,0.1685778048366502,0.02789458376392373,
 0.19122391488542434,0.7854715588607091,0.4793951329177064,0.2257191033449567,
 0.6633323522666461,0.6905647002089117,0.9218245931514036,0.9123663236980708,
 0.49924156415568954,0.8801070896386866,0.3976440089036851,0.5909351231231105,
 0.8591347730568926,0.13781484526317866,0.8578585236959956,
 0.19195555857109603,0.7185881844229094,0.6981703471199053,0.4286990479940339,
 0.11890411750902152,0.07361181081340373,0.7975192288278506,
 0.9531786836458711,0.2959959637552646,0.7821960439577024,0.3616600521788975,
 0.6214336655426105,0.21790025576204186,0.12498601720414981,
 0.11906358956216101,0.9007240498008774,0.9900611854255308,0.778621377802893,
 0.9850005277972091,0.8297939002208371,0.035016795939172174,
 0.41516517149889176,0.5063695176684001,0.6979688782965955,
 0.49591692236111207,0.009864435900321045,0.9838029287140445,
 0.667361302154978,0.245712275416336,0.029935205491214867,0.45601251397876996,
 0.07383516172297566,0.38488847078134414,0.3774905034562567,
 0.3597685993121429,0.2842674890850817,0.13588874201072587,0.5012761178327473,
 0.15582803667398692,0.46061831517511365,0.8105062215996311,
 0.9261900459449758,0.8717949003006477,0.6599030234346445,0.07878550567841647,
 0.6508144755052091,0.8922212899067683,0.8272941941931157,0.8279164309392988,
 0.7076657708869735,0.14518144992604876,0.5905697765307989,
 0.058776797127803815,0.309761518498578,0.42530836943422134,
 0.37405984020254557,0.6372489026105277,0.7306370794733132,
 0.48421483517238273,0.22872747680173044,0.41172027686656254,
 0.5962123553189635,0.6811034233235799,0.10371687420524889,0.9547978100169102,
 0.43889314362310583,0.9961808322044428,0.5519653268209898,0.4356468826588319,
 0.7214015228266594,0.5531047871107163,0.4493723530147834,0.19160063457736876,
 0.23536513224135058,0.48946117702579184,0.20670890642850515,
 0.7275903401632712,0.5680777839455411,0.8400844762103689,0.15049862938244207,
 0.67161923181118,0.33970969592918054,0.44091961379160005,0.9203984067953273,
 0.557123048517045,0.9807491546143837,0.8021570309015733,0.7420920887511294,
 0.8854522065634836,0.2973803523755292,0.3540584721024328,0.6348077229308371,
 0.6207411951129779,0.6361282050133406,0.640003902723878,0.8814764020005744,
 0.25044533792720214,0.8149791440821695,0.616976620637127,0.8164814379042424,
 0.34664333492475363,0.740871020277069,0.772137142276452,0.6932285484028502,
 0.9957945944856093,0.49803363154221814,0.6028891891900044,0.2034077711240505,
 0.28798115779493905,0.5576030382045414,0.8369937278411367,0.670906748677296,
 0.8258707307745826,0.5873478545759588,0.7118986194965672,0.358531490431071,
 0.8129762276892176,0.741456684544514,0.6714611394912399,0.08148708111881309,
 0.3274370927504303,0.9318527860863854,0.49255408263680056,
 0.45640148770281075,0.05243780082642979,0.16540921720904156,
 0.8119867023308643,0.9821466190756623,0.32591212353045995,
 0.010254360202561763,0.4659219981443752,0.2062402773264237,
 0.39441596945711344,0.7370167490740139,0.9953396570050265,0.5625907827035077,
 0.419464898142622,0.2252117485958396,0.6057338664624142,0.20947080594716005,
 0.8992862877276795,0.7737453767662831,0.18963256322754407,0.3722981633751491,
 0.08143560380255621,0.9094121865773674,0.45032426728109465,
 0.3626206983503625,0.48778783729844455,0.05790307594389632,
 0.37997191283411336,0.988314715469435,0.20743121939689335,
 0.35512256916685847,0.5898744989146916,0.3053863702791173,
 0.07216300150526112,0.2997476639324569,0.0038528750506460208,
 0.7999565272610695,0.13569879448990463,0.5939913632282172,
 0.19689115243022925,0.12122468570982747,0.9663335779642178,
 0.08969844157533602,0.14608514797622052,0.9291599863255702,
 0.8159465128075392,0.9822924593305326,0.05172342030398924,
 0.37871571292441275,0.576088245565044,0.5731085012177333,0.6503909962946104,
 0.07651737492352495,0.29748506354836723,0.20426887038234787,
 0.49795191443035747,0.1687536472392508,0.732263018026584,0.7503088580337162,
 0.2137682743113466,0.5338645939287486,0.2238723766481049,0.9055767205827245,
 0.5204199973965962,0.02458460022687803,0.4644358595895328,0.7035734191303507,
 0.4517586178339248,0.8693361119525693,0.16022178916515017,0.9747343239419421,
 0.3699442975883227,0.6290812844577626,0.8873155667906422,0.9742021278980071,
 0.11968641368223132,0.8583143368071525,0.21850146614614052,
 0.6438715111044274,0.5016918853475225,0.8517909334421477,0.32028136595515067,
 0.8473037586534928,0.7897829459073789,0.4742098532410284,0.5749830813352546,
 0.9022570350181739,0.5610426292271412,0.9383040936386823,0.7041915120197729,
 0.9845699305393197,0.811503382995155,0.7771674714915482,0.8467871525360673,
 0.5155991940307759,0.048748466567219984,0.9315070698947983,
 0.05379072163634202,0.8622949492571059,0.8712702476779801,0.4716262773694955,
 0.6565428668088171,0.17906601541341272,0.9264164987475316,
 0.20828781856603829,0.7718417007669682,0.9711737359139595,0.3180094976521518,
 0.876237418555617,0.054053234836976105,0.08280947700713825,
 0.48295721786469237,0.38471073154161695,0.3448822448246107,
 0.6059505844453117,0.4347475441594312,0.6162147689429912,0.15415617679356475,
 0.6216657644128266,0.5704215563785732,0.23852481892420307,0.9428061973951862,
 0.34142740765122703,0.8194985199900209,0.578975728079537,0.9973547728157268,
 0.3184961078211659,0.20229180343739117,0.5287834012169843,0.7103745733053414,
 0.4412485988350724,0.5324889737016537,0.9787616552057627,0.11666034732563735,
 0.5593561073331759,0.22931787978372398,0.9031090800743292,0.4314815005944347,
 0.14084411731469737,0.7427123623669559,0.15272514744567456,
 0.12177413154876493,0.9322289170315032,0.48010232919976625,
 0.17158792778448684,0.6103403640255514,0.3676158059990118,
 0.027807794681794396,0.6254353623205359,0.7514959656812368,
 0.6464700221966811,0.12269892827412832,0.10347170434449349,
 0.8789269901607386,0.7885348235656249,0.6056696316311544,0.344822735429852,
 0.4651116885732304,0.07030779860510172,0.3417268656222896,0.6290163213398711,
 0.7798314865624769,0.49634004721210456,0.4691263603108248,0.9417944548718505,
 0.5920707373920018,0.36982805118187645,0.852406179267053,0.19739804844213693,
 0.43856674582163035,0.7905359488814075,0.8156241620836734,0.358480143667688,
 0.6054131866799148,0.22416098529938222,0.4596677306041963,0.6670822006521302,
 0.17175695143075465,0.2649156697906073,0.9132370338657816,0.4936096825206169,
 0.8919494940170967,0.9232161507890796,0.5522235782216401,0.24643189595505277,
 0.3316906592053688,0.18739887386924703,0.5187778879814559,0.5232532974211191,
 0.31880966010201406,0.17233662951275852,0.08249671330136032,
 0.6955276074992323,0.6874787766324358,0.3146713133423473,0.18082419593216104,
 0.3549144627341839,0.41423376235123777,0.718110093716728,0.7722522347543779,
 0.819794408865329,0.6541724510272962,0.18371865643894758,0.29579769272189305,
 0.5343526705840893,0.7173949691819422,0.34939448075930735,
 0.12143002621889576,0.5883900660762712,0.8158112984298849,
 0.24806050139173097,0.30232136017983324,0.38044217322713614,
 0.45521724508289063,0.38937639290626347,0.08901293631663432,
 0.8405727873591775,0.555012064032037,0.8407042947117056,0.8983166796971883,
 0.656388644095638,0.8399309652797129,0.858198421997531,0.1062343699733661,
 0.18789211269465156,0.8600092492919665,0.9513488799122766,0.5700197577219412,
 0.22700841842716024,0.4487183961651957,0.05765244908317657,
 0.7664765537251365,0.2626135594836245,0.01856198242839524,0.9615363952065756,
 0.06135083599493818,0.23623765825700094,0.512390418164146,
 0.11581592043351074,0.6806952732798562,0.018902711406880157,
 0.6099255378190105,0.5552274993147185,0.9005218159871065,0.9787382785900437,
 0.10189019963030765,0.8086420309758497,0.32273741202153705,
 0.4144932017827382,0.8859167550719347,0.411871133481023,0.9379345009034247,
 0.32800399584256335,0.22820750860011996,0.7416452234329766,
 0.8842046751042991,0.025846847545355378,0.5816997183359731,
 0.3362912240747171,0.31187086548989096,0.3585643897065194,0.7548356313619914,
 0.6072873367239253,0.9914173355133821,0.21232141829363194,0.16381594964681,
 0.5403019453955189,0.5287476487302397,0.005443959873307591,
 0.21796329924143765,0.1959349596358002,0.08125505623658202,
 0.021402578158645436,0.1483792063860987,0.6729124531500675,
 0.03114402092693047,0.8055767998960044,0.8947291899842644,0.5832161389758119,
 0.2493585804852828,0.1322694881882781,0.8622427240585857,0.3133382742945414,
 0.7058187324811711,0.11411930633716949,0.5985703374022961,0.215884003224738,
 0.33859102469643776,0.8851812383372363,0.005037664162717448,
 0.15384583829905774,0.6803296826639642,0.07127986417624732,
 0.9769352489739962,0.8069801877557488,0.6565461679517348,
 0.004945810951199836,0.9675211676101353,0.3840565628513908,
 0.5547392954248151,0.81420725973317,0.9190158782855855,0.8724258202888191,
 0.5496151995112722,0.20631773531930586,0.8182140263273159,
 0.30097445845396975,0.6840653251887078,0.1099132409395821,0.3860713532871085,
 0.10310453446546641,0.23847455088823766,0.31958193896858256,
 0.10306730145315046,0.6915025079840676,0.878249579192962,0.7124098943663766,
 0.31426123233864256,0.5927110692307398,0.22061427342364115,
 0.8767958520514629,0.6384886838029313,0.5148719152620461,0.541252824198757,
 0.490475025145388,0.4438666410204051,0.7056545302589695,0.7518526459975847,
 0.09569824769410784,0.5475124465891086,0.8103947033377125,
 0.19298716692067475,0.6303056750034384,0.4370264991055405,0.3897795002618868,
 0.6101103555892166,0.6190519328191891,0.6121853108837015,0.810604330440957,
 0.8782576170559667,0.33040527786366614,0.9834192194278039,
 0.007343484852014082,0.6419846226700722,0.9064363541632561,
 0.8831951062609467,0.203697021015798,0.9865127242698527,0.6357182758836118,
 0.24156393198092518,0.4355304322596778,0.14843449440119683,
 0.6990447953416473,0.6869177820571537,0.6249886738940706,0.7674195053598569,
 0.37440756912940043,0.37887875120095704,0.6611774489783228,
 0.20228040684953585,0.9464612669506015,0.9276868265184856,0.7358416187546286,
 0.57732068587321,0.941159081945526,0.1245781483896008,0.7078583975176316,
 0.6843149102983948,0.6786538735459056,0.4225366005303852,0.6336081909441706,
 0.3603865247728143,0.6001669148237041,0.03001061202576405,
 0.22752070849544026,0.9601007018901276,0.2922015701550549,0.4600074343032007,
 0.9279948954347964,0.9123289494724762,0.17166982964739697,0.1540616183144905,
 0.25177193374303164,0.9839495471309597,0.3186679153715516,0.2720956837113736,
 0.11839315312719789,0.780293427305061,0.6614058373526324,0.9881239265299541,
 0.4477762196477515,0.5364792116635867,0.4441626381622611,0.9062497715388202,
 0.9039720090914747,0.3150349888703774,0.7192183818032245,0.35594679843354105,
 0.6984484538523887,0.553664519235869,0.7049938445590214,0.8086205089320799,
 0.7155990294712606,0.8661231540507038,0.8371601092499564,0.7258339993518782,
 0.5929658638175288,0.42586772272914686,0.14045465472794616,
 0.8370502078825015,0.1953781906371208,0.1594220907149979,0.21326636638131846,
 0.9894697412769661,0.7602957734942288,0.33049725350089953,0.8042864106404938,
 0.7369964693376128,0.09119462349748853,0.9875700653941902,0.5031058005678815,
 0.9539927818649174,0.37688991273545036,0.024331817831138802,
 0.7954909013485055,0.41778267488087184,0.39597911001769004,0.569250413691341,
 0.11970053419765525,0.947876215001188,0.9878672753719494,0.9191417304007856,
 0.372042769398086,0.6441621514609094,0.06007786166137008,0.5013505496553451,
 0.12120059063637778,0.7787063322844053,0.8135938442446623,0.4939309673205121,
 0.34030067663406094,0.5489996510881783,0.6644317423851347,
 0.18229834735498718,0.6630265512684143,0.9000176216947912,0.9410507202455265,
 0.9211743641360602,0.7142804100175563,0.6615308300326966,0.6199635216622909,
 0.4968450132663961,0.8065709242653604,0.5569083703181424,0.3355629577792896,
 0.2009891208275637,0.6283539683717109,0.6925029303972038,0.29179397601923107,
 0.27956059520340326,0.19059650313245702,0.7722955108983843,
 0.8303142873361021,0.2929086981837876,0.3800591130935146,0.29325424352455387,
 0.7126060987547078,0.18164571641070093,0.5687890288713168,0.7262241215746688,
 0.15007282190427673,0.7780729409159348,0.7221133017842125,0.5398201600658229,
 0.9635805279893985,0.6471867151950086,0.058562761920998185,
 0.15546979916026216,0.12772526306240672,0.5244001476760956,
 0.4083838181197471,0.5877601894520255,0.22105808607862587,
 0.35510566567716584,0.2576803703353653,0.4657891633416029,0.9327663391325042,
 0.0866372740783019,0.20344946266820596,0.2959308338436488,0.5393609087967293,
 0.5976184266311035,0.6770044981765045,0.8843569360207068,0.5309910467181483,
 0.9492473492366139,0.8993839418771402,0.791048703325718,0.9263119514053131,
 0.8069498980009957,0.09764805867122828,0.7103624399675496,0.8031114885517715,
 0.7274307147379611,0.7032078664178659,0.1858567947865002,0.16960133002124456,
 0.426869644011457,0.02728156700287765,0.4469139724645552,0.31232157403147054,
 0.9552750601742815,0.9862094982635394,0.2835368463533826,0.712938310503659,
 0.58219466793468,0.388522010589166,0.9246082281287075,0.615007175835459,
 0.5111622715725542,0.9362080394106206,0.0015428808403360472,
 0.7346987968829275,0.17990632279083574,0.8169255343737051,0.686651647277399,
 0.9271873946522059,0.49694002989338226,0.6396740300868882,0.7970987962534523,
 0.6569773907211489,0.9455332498480193,0.5480971027994099,0.803920547807484,
 0.33201234069450736,0.07451469732714244,0.5902900291463665,
 0.9502199029829987,0.3451764467347893,0.3249414962315861,0.9981740137663503,
 0.6462570704921357,0.2419811111074731,0.9049556752543276,0.9591838565692453,
 0.21245426738489281,0.5897620446438764,0.5744408245920478,0.5716771354420994,
 0.6248212207217332,0.007314387480007567,0.04585942078355032,
 0.1654983330849098,0.7404447853851972,0.62816593784476,0.31614700036768306,
 0.8522330760456026,0.8002242927739797,0.3300521128592435,0.13171499746401272,
 0.5621273136329519,0.592453923155739,0.2789615421789211,0.6409670173076752,
 0.3893688884613963,0.7275975697309809,0.5703923284373006,0.5981428021760753,
 0.9269053984497017,0.2303182078031717,0.5504439216103659,0.5583695749589381,
 0.10863349025598956,0.17693108748192532,0.7786350993148112,
 0.49521153019750275,0.758250476789397,0.6781221871717081,0.6850528285422347,
 0.47089412712825385,0.23833613213886473,0.16789356528464483,
 0.24992530940936286,0.3687273199791191,0.9977465358651368,
 0.025846915651047242,0.020687647453529134,0.5083324249760138,
 0.9670989414664013,0.002187496792099952,0.8940919373824837,
 0.0805913072496045,0.05508498487616964,0.6099596249718611,0.4264466399412189,
 0.6688490570018113,0.8994552909512346,0.6862186683280408,0.2347256072787065,
 0.43362047622348765,0.5787478886775376,0.42530126193348217,
 0.3970578382846447,0.15892737636562293,0.5651775258873996,0.9369965751986069,
 0.7102322537223603,0.5297684634208689,0.9397736237787784,0.8153047258960497,
 0.7695375263626343,0.5967541017735951,0.9300622868302494,0.7926181467866874,
 0.5301078510020156,0.8446185066753005,0.4080112773133986,0.9192521316993828,
 0.4729780152169111,0.8868596431135156,0.3540325338104625,0.28983419279386036,
 0.13130390856845042,0.3222242872300739,0.7974956879450921,0.5034232599380241,
 0.3755877702813175,0.39910253598543666,0.32395798111451113,
 0.6721525682290506,0.16453131865082404,0.7230368597466948,0.6469186540137054,
 0.5379001998605736,0.471129620138657,0.5265971577744681,0.8979434884509075,
 0.16476097494011532,0.4585673349871723,0.30007788623057885,
 0.13893866363880591,0.0022175051331749795,0.29794776670264655,
 0.3776073461302757,0.9016056928306346,0.10982515344938004,0.971930805435711,
 0.790896536665156,0.9668626575912362,0.6347294644038379,0.10550196854626903,
 0.6501833950582828,0.14550882461796755,0.01240298100146997,
 0.25087695290032586,0.7819931798582745,0.2737802464754363,0.6321483044876975,
 0.9442092333060078,0.8159795851644102,0.9420083889684534,0.7391628197569199,
 0.8071143268587984,0.6221595430640391,0.7556363318466991,0.9955113004688072,
 0.40205400578795025,0.2299934751941406,0.9395374026203659,0.4576946680015036,
 0.2920739016885273,0.9037846920111112,0.5390806134318273,0.21725464079909906,
 0.8818327768230995,0.6939713528167495,0.7562585889666316,0.6452296334919151,
 0.16794593622567333,0.46589175355318735,0.029772398316983084,
 0.4360615041154057,0.027979963302456,0.5038863143101215,0.13204749073537259,
 0.8900353581534306,0.49053422649242506,0.3148800112843253,0.2689706255601103,
 0.03360028575369611,0.8096120761067855,0.20300126599113477,0.400351898839709,
 0.3550293937141623,0.03593735798962372,0.1685533271758768,0.960017857102351,
 0.6629244588865895,0.12762611321170636,0.6067146466596504,0.4802000026134625,
 0.9216689163012942,0.04477765081023932,0.5220260149191774,0.6727643554176146,
 0.36576810901455237,0.42717162200814307,0.17182398128471854,
 0.9927476643360207,0.2845998835521918,0.16007583560818195,0.8375655216187131,
 0.023153385373195667,0.9396874275976462,0.07925721938330665,
 0.6732301120839054,0.5231228891676423,0.4433964000410241,
 0.005311589772106551,0.34221929889843294,0.5758877095208601,
 0.5622942674524242,0.761717924620939,0.4543125413709974,0.4334708979003228,
 0.45972573056195,0.5181880746464052,0.8157043040287024,0.6501163875463575,
 0.7374724900089323,0.576963126142813,0.906976857027997,0.6310085451817306,
 0.9603626422767042,0.33675500443072703,0.7849447761179515,0.3565833714638963,
 0.6026981794463945,0.13460608434794485,0.6479587865789598,0.797740700295544,
 0.3141660040184636,0.046064022141605876,0.48920785734110495,
 0.2939020747873067,0.4994809903054459,0.15630680466617997,0.2931194604622859,
 0.22057227335969087,0.7791114475852309,0.8727117994893343,0.7133806103838178,
 0.11130178450722805,0.8395017721541188,0.7334062671291096,
 0.025866069243232426,0.0958579482871339,0.41021751904855763,
 0.10731888773349063,0.04952729408369272,0.00372424456470144,
 0.14989099443186027,0.4277068369427587,0.8837649798331847,0.8876522147005286,
 0.933666699103394,0.3419313695665108,0.35210774400070044,0.22438294674834525,
 0.7167892427101415,0.8333546929302057,0.04368451663698392,
 0.47705305034544976,0.1406630879717542,0.9806627852915623,
 0.008288625367793179,0.012143362185345619,0.6276728791420186,
 0.10704779696999467,0.15926352242818742,0.6973527455460626,
 0.5103163084294047,0.35107374564836746,0.9912759754337571,
 0.37200319984696895,0.6630227361732274,0.49879636025118945,
 0.1990242074258206,0.03661416955818042,0.4978337529179204,0.8313876656367059,
 0.23391795284960204,0.68478347948215,0.1098526662494248,0.1551360701851524,
 0.9905361344812714,0.5424061171570832,0.4217473408817637,0.06689022870621275,
 0.7683838074318756,0.25612323654641067,0.255478920990051,0.46903504902505855,
 0.28948502408106513,0.3927326696104797,0.5033223621357656,0.8238054055227378,
 0.4725920742942108,0.26492886190080644,0.26217366545783183,
 0.36775293593298164,0.48962872878243846,0.027061501112114472,
 0.03307231295378643,0.9416360254529501,0.09353280061680447,
 0.5168914004575962,0.6977321704806114,0.4024606984747957,0.21646924577744997,
 0.043616222432126595,0.9408145826985159,0.04716948319829173,
 0.8377661407783602,0.6509505199882079,0.13749260963393384,0.4216717430513388,
 0.12848115699903895,0.7627244785770148,0.4576594961791649,0.8159582870304672,
 0.3531437637248821,0.5788945535327281,0.4390895499833527,0.5258972977135535,
 0.6762356317597511,0.7402196874719601,0.40504541092438073,0.5193730787102667,
 0.4435788469590213,0.27756982955574827,0.665673155874676,0.5411674022895452,
 0.7886147250099731,0.33764071401011686,0.4361705956166424,0.7773658735360183,
 0.17922064909743163,0.7147366740524062,0.24095257486004074,
 0.5989611675652208,0.06378230560689346,0.5578177250546562,0.9462207686294178,
 0.6860012440869367,0.6481616999268991,0.2747823983083695,0.15689508869143365,
 0.7115520454072707,0.06610635254223962,0.5001007561917297,
 0.18846642648753742,0.35132138134707624,0.7530279093870349,
 0.18452347025882365,0.6944300184855786,0.6713510042700531,0.6585019163331769,
 0.9060915294559581,0.9407591945543684,0.41280774234072193,
 0.08633638190187631,0.5990129603387988,0.09691155148918351,
 0.5688043686752947,0.7066249565594703,0.2678010203704708,0.6600623174911826,
 0.16471429270261118,0.24086784403383554,0.4851543708546706,0.549939460143423,
 0.018127281416998375,0.5498689513052378,0.6311034543077799,
 0.014066543875904336,0.7028350997057089,0.27744436128482264,
 0.9917607571101531,0.4592655024666067,0.17769995491316093,0.3552877576318483,
 0.15550957598418225,0.8603042934090541,0.5031292769100231,0.5362976011849061,
 0.15109941484209977,0.3505400759857723,0.7518701770906528,0.6705216710011653,
 0.8007159064127364,0.8598833514182778,0.6518823891491813,0.7907170221371056,
 0.47180260647560057,0.14259538587089438,0.640068421200973,0.5092610750729081,
 0.5635837104572029,0.28811796858591376,0.4723252578845547,0.5038347697753157,
 0.3022494798360036,0.5997761346260639,0.029955932324116974,
 0.8337158326534001,0.9448168288995756,0.16660655403527347,0.928145791696394,
 0.3894296176579337,3.753698115934734e-4,0.7236412971993159,
 0.9468503843542911,0.7112013237824083,0.3413060775932454,0.9856945910776141,
 0.237335403728425,0.578960173871219,0.3427261903441211,0.23564271550789462,
 0.7292239277583499,0.6063938595934187,0.7548269596693071,0.8077738427027236,
 0.6714176826811126,0.2765104052168056,0.533677184096874,0.15991952682961363,
 0.6652152481310278,0.7650209400915995,0.656791161931763,0.5204124101615173,
 0.22306423432918088,0.7426803874093412,0.08030845401959186,
 0.18723151658188475,0.7841464968216352,0.4155466311664061,
 0.09554029842934464,0.8129398067166963,0.3505091962699096,0.3052317986304113,
 0.9837895989551131,0.34476667391948024,0.258852841325414,0.9475907336576831,
 0.47062859066812335,0.15830029808501034,0.16608481341868186,
 0.008497486178909885,0.9294806861432867,0.6648483036917608,
 0.7371123245993766,0.2976975832582933,0.1757306414946831,0.06965607685869313,
 0.08980474106194547,0.4051331088439587,0.5710833995689436,
 0.36774476685173096,0.004675540861703187,0.8983145758597892,
 0.29222866935025904,0.02950581607256353,0.6605763211348898,
 0.2771573984962177,0.26896126063112347,0.9127039377422543,0.6212146152270199,
 0.4137973806666816,0.2124087552159578,0.10234537084880779,0.7351255443141396,
 0.03528843412696503,0.926194692468738,0.5363243837766997,0.8272797041906885,
 0.10398738984561462,0.8349990800645468,0.5603863806907737,
 0.11541188751404907,0.748601902275326,0.580116149049556,0.7414894043993329,
 0.8090100481727516,0.20901942188459421,0.744583095946948,0.7750950552228995,
 0.4947324077586256,0.5459124634780758,0.2430164663275941,0.4197951805944652,
 0.19889792282071272,0.32505828237297063,0.7267123045239845,
 0.014263257271193153,0.972819068248824,0.6367391325861212,
 0.30250305063322513,0.6818254975823671,0.9466544968823186,
 0.41806039104866155,0.8366970745922918,0.5765014627777252,0.5904853195221732,
 0.5349321327076543,0.5913753290358369,0.3555267809421726,0.07411657960696805,
 0.41472546166591073,0.08774832403076249,0.6151263182971924,
 0.35043944373493296,0.3745772947356929,0.5520189215731269,
 0.39309477934846415,0.4201128931592265,0.8208523829018701,0.940271152420049,
 0.9081233726730934,0.7168841715981803,0.9778995753761097,0.8887976594493541,
 0.05597654319634082,0.18843640450997257,0.3713062500084905,
 0.5988191985687122,0.16159495797403245,0.07044626225319572,
 0.7628433765315306,0.5347770387592641,0.37080950158229464,0.7384681780426151,
 0.8924860069204084,0.5480124233948724,0.6584868111686627,0.03938519612080127,
 0.8375504176628102,0.9928251827253699,0.8128191591837421,0.42716117737912485,
 0.26091421076332255,0.025023729135707162,0.42785526125120343,
 0.9681821415059251,0.9646835319499277,0.8525426341142154,0.6641643991685505,
 0.9739679797040046,0.6605573734555334,0.6256514331830204,0.9661823562031084,
 0.2628945794909343,0.6024843673313588,0.2770382700955989,0.27808311697035215,
 0.27387411164307773,0.6353901364363553,0.8071726484268753,0.0669565441311194,
 0.21145923437432168,0.9200589327936801,0.7758908103201018,0.3449039119205357,
 0.37190226337399646,0.4343456742062337,0.5495711681652844,0.8962076010686808,
 0.08949641943973818,0.9720168888165033,0.30015433869788244,
 0.49796026396295945,0.9508165464140318,0.9322098885139478,0.7453093616605078,
 0.02527871874346399,0.18524807494059636,0.3920175292079797,0.582347232007251,
 0.08798428242565026,0.295878741503506,0.7100698756347084,0.5399634077482178,
 0.7147993866333533,0.19531161829235,0.8766187940437714,0.6923646534185417,
 0.05704595526237932,0.40441503891266384,0.09362036090361503,
 0.5762467991843464,0.31145493589069906,0.29789876002237625,
 0.26769311071938195,0.1729156359936358,0.4813885881388825,0.1378219594098853,
 0.7221595220208992,0.2776319192863621,0.6859283403832125,0.281838288295274,
 0.4736483129021103,0.29918760354092966,0.7664147896568025,
 0.10709516839192579,0.5717605903966959,0.2692941946253603,0.4253631029064382,
 0.8331107684681999,0.20276537767389902,0.34928216204416973,
 0.3309301863254567,0.15279253465149278,0.9396279286286217,0.7988866393660138,
 0.8357907631480467,0.257045850325736,0.45375815487404636,0.8690330387966116,
 0.21822718449138756,0.19848730862464525,0.14897181253481206,
 0.7209592381178275,0.36024423216814894,0.7908489870317779,0.9142778770246208,
 0.5664101078776803,0.8850939931773436,0.3553628809194591,0.3575599018909501,
 0.35470851083804467,0.043125206266529714,0.15261032146480447,
 0.3412830502408809,0.15244778478725762,0.03974855381533693,
 0.5791048152289697,0.5981060662412345,0.681776360172668,0.7079466151630094,
 0.676855983984234,0.7951269032073497,0.9868216674898461,0.5441529530022433,
 0.20350040356941057,0.061940936655582535,0.35918897534628724,
 0.2636381380341086,0.028208046237176432,0.9872241689687318,
 0.9947690778417222,0.57375646464968,0.25591385292875823,0.37018047936247267,
 0.4170177977697449,0.6486921303838883,0.6185507705233189,0.30302011299240916,
 0.02459762847936453,0.2577318293015378,0.5002040083736391,0.8807649011807996,
 0.4242199966529344,0.876433518622159,0.3183619640109022,0.43096028524936403,
 0.3291690842828422,0.8695402298209846,0.22913941825390305,0.6987557963904232,
 0.8923295074874018,0.376327015057188,0.37680759171785905,0.380442548087799,
 0.9223526527818647,0.4858516926377573,0.6297888535401441,0.5184678565518402,
 0.5814389896378955,0.48290024817989663,0.15870763761503337,
 0.009465470169403245,0.06466953104563733,0.24255489183936185,
 0.799694630406377,0.34935977537190077,0.8080764432274069,0.734364699374872,
 0.0033074661620920986,0.3356649737752814,0.09154990664924978,
 0.6134750187782004,0.7583953361662134,0.3871999759190887,
 0.017286906364750942,0.8947127324697336,0.8677317499200619,
 0.5532021184920101,0.15856325958197115,0.10792915378669321,0.639973577225553,
 0.7527270410701679,0.9084611297834048,0.45423766754938355,0.6317974525996475,
 0.20649697780747678,0.41999369322809166,0.009478624073056663,
 0.3933437717053113,0.14900620367873174,0.5396680907321192,0.877339071579233,
 0.7940392015411165,0.5662800253106162,0.6708139028119964,0.6493238463419625,
 0.9495981915415658,0.5347603417754593,0.5747672838630755,0.715701750064041,
 0.6594535231134894,0.7409597533325594,0.5554251335271181,0.4582975528717207,
 0.7197956551660939,0.46148926436381144,0.6765206450666013,0.6825703138378597,
 0.9906912622849485,0.5848218231977836,0.01804780832856101,0.7798088759059406,
 0.5972143895540843,0.8404604523916231,0.5186682082484154,0.15106708490731258,
 0.18781739521575402,0.5537745453919127,0.03616682740950983,
 0.8448343937925744,0.06816292191042428,0.5109522604483171,0.6505128631452861,
 0.21616148373986077,0.864095527059669,0.6482323605581376,0.731425523683989,
 0.7562704720871263,0.7995702864587755,0.560942787131295,0.14211410978933425,
 0.8018137816362269,0.5753883759077997,0.35879404276809024,
 0.42891781604499357,0.33929013046823087,0.7405577328727454,
 0.11167133786336048,0.2510781486519583,0.6691756859613203,0.6960616586166419,
 6.440988868989805e-4,0.5896300010809301,0.86174065808925,0.19374552406366896,
 0.06678074613159113,0.20844270227142248,0.2495310292461269,
 0.22389141393965517,0.13144709106438934,0.3032129399171143,
 0.9128824559471751,0.20288070549090664,0.7427725668838947,0.8656439686475141,
 0.2528886165538289,0.117848755735669,0.5358261285623269,0.10925519411788576,
 0.6833930511175972,0.549989250341455,0.27663328577336965,0.0791254560668555,
 0.8039282390633534,0.2096700100249248,0.41764424420357427,0.1161557850071735,
 0.3349208080317303,0.32191465872191793,0.8187638691052279,0.8796670840040333,
 0.29503265173540694,0.9954392284852577,0.8692764311807926,0.3460463471638934,
 0.2806326951878809,0.4709280054190125,0.26757852356001377,
 0.29738435906465543,0.9800666223950567,0.1536949200966402,0.9224779428550612,
 0.7147577878729846,0.6069721520053895,0.4726515365498585,0.9398685353049524,
 0.0033444240268817316,0.2672896083974712,0.0741947738554034,
 0.7950046518557071,0.7794688535911827,0.40012342054114336,0.3747736791120072,
 0.8656590209531999,0.3627957813794207,0.6631361902490274,0.8346322612033931,
 0.3979466851976521,0.8385785045932476,0.5226210637335194,0.23151543371201833,
 0.11284644095735685,0.31840643428124016,0.14863548617715194,
 0.8179587265393818,0.437739548199906,0.5694763227656411,0.5111888169882114,
 0.8377332036994206,0.9733177199619641,0.7495329005305271,0.22794317664312447,
 0.875519687354997,0.06147644904527094,0.10271964602621475,
 0.23178132157821518,0.8812102635359808,0.10712349256917131,
 0.5155531239864588,0.6257400884620641,0.09048995037980456,
 0.02422075838760329,0.45091309731835905,0.37642596805432094,
 0.33432265545603546,0.27985676967458273,0.5058344424503245,
 0.7731134519968448,0.3101402880473937,0.90350554710027,0.3876252881675889,
 0.33383235861956395,0.9319189144345373,0.11251418130444402,
 0.8262661271907874,0.1373950395983209,0.903234078034987,0.8961884847763897,
 0.8828556267122846,0.28324323925558215,0.5315345085384453,0.5044933890369475,
 0.5199648784988231,0.9862277984340653,0.722107189763935,0.7930408783898844,
 0.4997211280448165,0.6257572137200882,0.18506557103277377,
 0.08774419258098709,0.7468694298115006,0.2179078416777328,
 0.32180606466414086,0.9606941962170674,0.46886785097826855,
 0.6533340917413671,0.5541092403879317,0.4081151589138521,0.17752024873485395,
 0.27641529311587265,0.3877624532726046,0.28559797877810633,
 0.9041230041787375,0.18876025688495102,0.6504899056940597,0.9199641804996969,
 0.8698183853945696,0.7036530048668619,0.20236301737798423,0.3251956736691305,
 0.17978501841959482,0.2544241035657703,0.8684096495806507,0.8054185711782349,
 0.9232122028477231,0.8740562738203609,0.9600656579761139,0.3118659106575118,
 0.5811006623010475,0.7087834227303174,0.5891291382032332,0.5478617221343203,
 0.2368993524115397,0.7873789941692175,0.7911718476556364,0.9275154744686842,
 0.03411236303243126,0.8123989912901183,0.12215972207827108,
 0.14839246878906542,0.6885943648178526,0.5771936215666439,0.7417368372703305,
 0.05277942038080641,0.09407713782422267,0.2920299739237062,
 0.49619801655197204,0.9903988651283948,0.49154425165927884,
 0.29132284094033056,0.4661853316326464,0.14292351929363523,
 0.3233673514246459,0.696707928493289,0.0458770190204687,0.13305018772082344,
 0.3276077946120992,0.30842873421250705,0.06579625686105184,
 0.8649679611539858,0.9655895370981136,0.6612290907839892,0.7412686046264416,
 0.25207144642531987,0.4715917238759655,0.7470812434655151,0.4374454122966658,
 0.8139081711841099,0.08320242542245027,0.5937467580365474,
 0.11220764160790586,0.7312013332995287,0.5669352440321418,
 0.08690905010961325,0.6853708953214822,0.01782107468599292,
 0.3910500049596056,0.37452972530737594,0.4647877271844294,0.266447712210764,
 0.5900363963807906,0.9603220286749082,0.9392703479857843,0.8401124002294484,
 0.45856546364563344,0.849790109356363,0.7179011694019573,0.9532227886529454,
 0.146613090652998,0.7280559188143615,0.03127480541451266,0.6489106974400404,
 0.8513120943154109,0.1528163511487275,0.7523485051953301,0.1168078962432807,
 0.4326955656865984,0.0362518410278273,0.860741259166773,0.5231615024235027,
 0.9827135670290065,0.7112343665962451,0.1566722034898138,0.5760805421024509,
 0.7685315464214849,0.7526105345547246,0.9503617030412463,0.14106726313287088,
 0.9440293351425504,0.9507549196383458,0.4183810440648018,0.8018834820699416,
 0.6986069964985869,0.21552951824439537,0.8018015632149297,0.8367545956362742,
 0.0077291853081074,0.39911335696714345,0.281475810707423,0.3018835570603844,
 0.07368132626277468,0.7085105972966158,0.3188168922734077,0.7178688722007509,
 0.33821999546273707,0.4344332064457075,0.5149607324929284,
 0.014195661971411755,0.3098794668882817,0.3242507958315749,
 0.6481508063017267,0.21955634224161946,0.7183173000719658,
 0.38801118886679753,0.01749544617492249,0.694681635814246,0.5097028556535182,
 0.666180677838455,0.5507649414369186,0.5946225456083294,0.6127796331332143,
 0.7198080169887237,0.45025591910804774,0.5574742448436643,
 0.07268000164610111,0.26105391947354695,0.4066588038037361,
 0.7340959386471395,0.9105695886430606,0.9856906648874554,0.35335702210517694,
 0.9820352570607788,0.7187918956676075,0.37507874190586366,0.7940331806600661,
 0.1892956580569939,0.3262498976592071,0.8112397551144032,0.08544396136035304,
 0.05136028741943832,0.8977100482140571,0.7377636519567097,
 0.19225875604185128,0.4619046245699516,0.7135978550764942,
 0.15156340533208712,0.9550725808879809,0.9598946026094175,0.5292106862464968,
 0.26204222146227174,0.9980474696239924,0.32708727903589907,
 0.7849813882174075,0.7612366008334375,0.06796929909766414,0.4036471759204995,
 0.30320118693798803,0.6463676656526929,0.24517663431881298,
 0.7950733217014602,0.027084134782564995,0.2823848833871231,
 0.22704912319649895,0.4315422300329418,0.2016137369606853,0.9693909765021823,
 0.5651262221604324,0.49067723921885165,0.621181742608213,0.49484317444480697,
 0.27812039830515334,0.8852650450869652,0.3051637116333563,
 0.041619875612507684,0.45201174821228407,0.19136313074853806,
 0.8295511730365135,0.9423095968239157,0.013635600525183755,
 0.5538407489649723,0.6199491096210995,0.18841809498717021,
 0.10377203449463168,0.6594629409355705,0.2436876884472432,
 0.22174034497177086,0.0978015109455469,0.44088936851558536,
 0.20533646699487318,0.8180770149820896,0.47653288162510266,
 0.15069026473801128,0.5719373212720251,0.7990278679593851,0.6368419764751825,
 0.8625690582185662,0.7570671568437238,0.0403093996415671,0.10289519388423929,
 0.7404296998114085,0.3577012132556928,0.14172906808494656,0.7779109563077876,
 0.0767344698339486,0.01094187271210989,0.4052700667323922,
 0.12437831679538336,0.6516978152769104,0.04512440606295531,
 0.8002037198585512,0.18610065090237604,0.5334818105148379,0.5303007662568241,
 0.9517817140506326,0.16483680803650436,0.10045705464948718,
 0.5731897808093795,0.10585467057973263,0.38115954688646936,
 0.7743616830694127,0.3015269119357096,0.12372416805369901,0.1622737557821088,
 0.6521650715063532,0.19586976765171127,0.9784621956108979,0.4170703412607518,
 0.1661500901223496,0.29401618500539284,0.9094835158827457,0.7716337288951206,
 0.8869861202272393,0.252635920306139,0.19602688155743964,0.6277217393557624,
 0.38473986659640047,0.26756858938673955,0.811620423163379,0.8355060193814872,
 0.6351123344271963,0.05960823194110754,0.1057658009852187,0.6033243855611158,
 0.9049459627016162,0.2957815611684286,0.3734146857996743,0.7570444254060413,
 0.44270201891699157,0.7875841215748359,0.16455275667703217,
 0.7967538595122914,0.10077781542916275,0.0648067460369034,
 0.24536118175081256,0.7143658233693849,0.3252416721623391,0.9562078882575539,
 0.7063530081970613,0.4413519140360269,0.8002749189441037,0.373606641885482,
 0.053528963098724125,0.41729676618991696,3.588382620347552e-4,
 0.17498895621901345,0.7855695670453445,0.16007156731485395,
 0.10873345953766733,0.1041155453714947,0.6872901366944258,0.8796523229150293,
 0.38338823757180185,0.9119727192356332,0.5105370113615899,
 0.36814898966378495,0.8971212242100812,0.08747660524871692,
 0.3609963200380115,0.6090374874406539,0.7744280561475033,0.40135060670800904,
 0.4374479289152553,0.8146241013799987,0.5078341765912284,0.3467835781764004,
 0.70149887587659,0.3940275902731323,0.19036154872060973,0.49433889860280733,
 0.6984476538077197,0.8860386390878429,0.9383117499064627,
 0.011036273061676871,0.7285052274961343,0.3313553559174708,0.768043877623935,
 0.5348387700091496,0.03253587818269976,0.14003452168133057,
 0.8010760508715062,0.8113067153120264,0.5345613257815056,0.7025834600707833,
 0.4561004644463794,0.5633647380785858,0.27815333614775695,0.6987945265073701,
 0.27145613946824754,0.028625468698769363,0.01648890405839709,
 0.7323154287428195,0.875203219706319,0.2892048651987644,0.4309871338045024,
 0.37845107074804885,0.9541491894881344,0.6624290244064402,
 0.18620296305072082,0.06814096979272422,0.3906671017078004,
 0.7587663101333544,0.45524846279211084,0.2818909178122414,
 0.33398104979175625,0.24123099426767625,0.5343201141034796,
 0.9539464525336391,0.7982253797812237,0.2749800271264713,0.7690083040072803,
 0.6020161234608364,0.4652159993031796,0.28209930371432757,0.55141183919793,
 0.6138844985560072,0.4082733786125502,0.7298465067185619,0.6930354206598053,
 0.7511182885626416,0.4496117271854967,0.9338717845346868,0.3830134267295431,
 0.821440953614282,0.3843306484461758,0.9345850032989435,0.43076475179280527,
 0.07732710039583712,0.488396817218793,0.8862536324315056,0.3075922836055306,
 0.07800012886448071,0.44012467425436563,0.31779564724365195,
 0.4823951056368663,0.7534647243930239,0.27387825457891757,0.9281021008872679,
 0.038045535563165145,0.5679074400470923,0.2673622808493673,
 0.7415008494393774,0.6851511668088857,0.051300123839293055,
 0.9355870475178213,0.09357260222058361,0.4486675864074877,
 0.010869767471470948,0.27559952143432354,0.33125109149282106,
 0.9487211645601307,0.1518202622001058,0.7433469772078292,0.7997123455485634,
 0.6556408507715344,0.18627268326416258,0.09324106816683797,
 0.4236457792631497,0.6942696063885404,0.5918494055471852,0.7259971097479238,
 0.3773253243111939,0.2108740050802833,0.6671618202805382,
 0.021295442372311957,0.14204504553111363,0.8390677674015986,
 0.1919898341403058,0.9492480412089174,0.1120064957371324,0.6327820311881728,
 0.8317965014833223,0.4583494238045589,0.3258037156823119,0.7675832482804874,
 0.6436898922521022,0.5897481999242886,0.6315637610423311,0.7777915652013075,
 0.31954253988767567,0.7126931466720096,0.7170169025719734,0.9095807699466087,
 0.4290542975063537,0.39698379816454055,0.29324224466614746,
 0.7883476039693058,0.5447441599176098,0.02224394771368321,
 0.19369606211411539,0.7866752313725943,0.5413930254245315,0.8520175260982008,
 0.08875272454493854,0.8155490555683773,0.5162724045606251,0.9556150270441808,
 0.18883270224059057,0.7959090452903689,0.8085241641342934,0.929022326815895,
 0.14230727418128009,0.76526167739225,0.2847880578741592,0.4013004486780155,
 0.5455974871478476,0.9312779449979864,0.6039653515586358,0.7483108606378361,
 0.28633972224568194,0.6232688351794697,0.9276650134142317,0.6210810456892231,
 0.3151398329656594,0.8846137114609608,0.20666311455669562,0.9425673591352242,
 0.35431885396157337,0.6129883877039916,0.2905158788072204,0.4078914819942866,
 0.6086073866295235,0.2204912217540348,0.8445898717452519,0.6425228766033921,
 0.1406218126962432,0.7609885998448357,0.031234354353325444,0.841919595540541,
 0.5229737339808191,0.07143732591004137,0.2723909650687206,
 0.13254956570403864,0.2815828627321877,0.3961082197974791,0.8795964319255175,
 0.47899773142876967,0.8342213418893671,0.3340301588521907,0.6954583276381195,
 0.7353296224525021,0.518337738541032,0.31414228730474036,0.3749181734421716,
 0.8533719348435838,0.3616828227382278,0.8504670179834188,0.17460271198435418,
 0.7804843728141497,0.09804175110351454,0.43833316138859335,
 0.3316368208182673,0.29988394672179197,0.5925366288799634,0.4521691630684851,
 0.8135619025051949,0.9565415361438975,0.16283881777470988,0.2061035078258644,
 0.36506720459734254,0.4195997825426555,0.6509555614274187,
 0.43699989029710906,0.7660101882004599,0.2263541954929299,0.972364919950346,
 0.21965811260498458,0.15070870693104532,0.03350484118921793,
 0.04991055398175359,0.48866538158260986,0.44765822725052506,
 0.12009443032490341,0.6299131938850798,0.14422678842308378,
 0.08780894195388356,0.016873568937727068,0.8328319744686796,
 0.7671322877029702,0.4439538624163506,0.0327183722554899,0.679123242292753,
 0.6866682262160504,0.993467868958275,0.3318865078515941,0.024640957726049684,
 0.846817385026707,0.8370377349586265,0.23710486825431298,0.46912086379278584,
 0.22253205425493416,0.6157659490718229,0.35081421363069343,
 0.7511484505732295,0.1217700854654079,0.08111587751081994,
 0.07342125639218888,0.8212290617266912,0.6710079481072415,0.5435565513230257,
 0.6476090357885986,0.32811348465508827,0.4723827106070583,
 0.49866450090465264,0.8452219344813581,0.9095045172766658,0.7236531431725631,
 0.4979086552596801,0.005626018470041738,0.17989056993382557,
 0.731691789742596,0.5901056223688653,0.05682408573867215,0.04815274343497,
 0.1381760644799529,0.036831278497449715,0.7013351071918277,
 0.7447254934401917,0.5220056146340107,0.18993440408103768,
 0.42970854652654733,0.15481303075175712,0.4325005526575836,
 0.5225877307147889,0.05858608042855651,0.22235068868344665,
 0.8835099917118038,0.4984308601965989,0.7784202863129855,0.8411500452600601,
 0.7808456834736125,0.15740713451439514,0.6462728426124266,0.8725209261824867,
 0.8780032262204369,0.7534893607975555,0.4479368679861745,0.15036131716164303,
 0.18829262107214706,0.985667866873367,0.7728441782956392,0.6792843884939376,
 0.7533552791776233,0.867977572076906,0.4932315373156334,0.7715080743231295,
 0.36409645692544634,0.3018817439731668,0.21275548937136635,
 0.07751724290809958,0.04793561472749652,0.11231273932377184,
 0.39666598535850894,0.10888649382767834,0.02695393385009115,
 0.2552503345288186,0.980666491517941,0.8523254920430722,0.3065081774234262,
 0.7546469681217851,0.8045186825828061,0.7245448327705085,0.9633443605375643,
 0.7217540191935116,0.21436242513566894,4.2582258466372647e-4,
 0.5296702405790511,0.38565075733302445,0.20858705762796526,
 0.32875476706746287,0.550022916503078,0.9360188044673365,0.04427569798099851,
 0.20743039681797149,0.48052762713600516,0.2194601644731281,
 0.7347725544534012,0.8645825032089212,0.5724261932505587,0.28934172869564234,
 0.20954847481741312,0.6397102867575933,0.678590407714768,0.31225926714512153,
 0.8454951981550547,0.887585802958393,0.4884496437404282,0.5965254758376775,
 0.7904423029036947,0.33706559336498965,0.3918525577835297,
 0.14848551000228594,0.5771513009214343,0.6648396741696518,0.6183693456481675,
 0.32403413219374133,0.9054095135283844,0.631550394000073,0.8840355288569116,
 0.10923139751555144,0.5150823810651428,0.9251567867124227,0.7167058531015094,
 0.9758625567118011,0.32968155936192034,0.892684101736312,0.09316266411781438,
 0.9448048990918536,0.3125466296457017,0.9589598129001219,0.4760913722930785,
 0.24633011691516393,0.8533683196933315,0.6480719210435963,0.8506911380147136,
 0.6198592621830443,0.9181371068875906,0.004984294617642493,
 0.009889550494159516,0.8283734730171333,0.5346901665840851,
 0.19007279656034692,0.12741883441482116,0.7361902811348378,
 0.9593827995815737,0.33982785266055027,0.7239167567573879,
 0.44036035340674007,0.852581911943923,0.6597267641603107,0.8543592742797307,
 0.20505408842940953,0.8226312660404708,0.016931221842970023,
 0.07034930091299019,0.12474184461460758,0.7316147087082553,
 0.3191386392165435,0.405807040390713,0.5449355081804699,0.1890675276762629,
 0.9586806399392036,0.0029523383948391047,0.45292281290786407,
 0.8403755530456987,0.8950141705825816,0.48510565490456914,0.5361023331630178,
 0.8852690339705698,0.1063245304442143,0.8821544949182529,
 0.0036919784700294445,0.9735659970441288,0.633870499690159,
 0.15443362256883741,0.2927942615434086,0.8586253653922741,0.6544955691973429,
 0.2886781020214729,0.3242737500061893,0.7927914753975724,0.31401162452542986,
 0.900700787446586,0.6749219762504979,0.5512391819313407,0.6902884871571991,
 0.30089818430020143,0.4900379722522149,0.876272638064608,0.7130617611689539,
 0.9533458642474412,0.5455306518269007,0.8471696009360765,0.22964069943317744,
 0.305999581378788,0.17423024504554552,0.5126740008128334,
 0.052415872951706355,0.7214054200222668,0.6710256600251826,
 0.6271133998607187,0.9360788415504391,0.7498829824588501,0.6681782108734229,
 0.11455407240188664,0.3359957227229686,0.07364401775989027,
 0.39024703221943846,0.3816099168858672,0.8458072282126129,0.9992853847726727,
 0.0020631392865754217,0.9922120142422913,0.08711602970287136,
 0.4930411311829962,0.13324445649549255,0.5516496955256651,0.9853902249654657,
 0.6454727095721144,0.4111938292308994,0.2175844129820832,0.066352624577253,
 0.8248920040845356,0.14275058295428655,0.8504779530577091,
 0.43444078201022196,0.14694164626033024,0.24259035904781268,
 0.5892382918370541,0.2714831047318591,0.7898508015328577,0.24227864995407933,
 0.8091628117719741,0.9559439629727677,0.0659318807767163,0.2804445952011464,
 0.12323832745264962,0.7844767620484215,0.7394759730818334,0.8055689559742221,
 0.8790344070152072,0.058537816622035166,0.7243740807321701,0.781091065130433,
 0.8665734133873861,0.842963751609094,0.47248516320733214,0.5527051108268185,
 0.21854393483743406,0.8803354074392827,0.9259442809553401,0.5268306256474771,
 0.09904320496376195,0.07437089418209342,0.19597354460410465,
 0.2606401570952117,0.7073889460162878,0.35507404840086854,0.6950009959463509,
 0.5454206404260704,0.47121309233488906,0.6400558158478794,0.4070580633023533,
 0.565679028630981,0.2172742841171431,0.8825170582315816,0.9402050567707408,
 0.15883465895324167,0.8998880037533452,0.8796909806044924,0.8229072704039819,
 0.9196983539709204,0.09310951020063829,0.24041019988530898,
 0.2795411552117496,0.6241467895418757,0.6590813387794935,0.3664730817518034,
 0.41432671063524307,0.8003679699690965,0.7999399624899615,
 0.37704459271907953,0.3623409674427729,0.9226606537224079,
 0.32444125950365454,0.7843057027075375,0.21330598614750595,
 0.45137195902710525,0.8465217310739163,0.16303910722141246,
 0.7293323230262705,0.11809674314827667,0.3135301407984138,0.51175072516312,
 0.8115672890857927,0.9428099026356702,0.7564095621372262,0.9836868383774467,
 0.5080365310194678,0.12139085887326617,0.5345375895285414,
 0.44731187812552464,0.06420865733159753,0.2508520700611122,
 0.7654708124296252,0.4950194808874523,0.8270770455439176,
 0.029362340547275267,0.4807771028732648,0.8148631912787959,
 0.06869012553954945,0.5001013035432154,0.82489697088642,0.023685971681259632,
 0.2907438064871851,0.79793490071435,0.8964960034011433,0.6583549869699795,
 0.7907562786545015,0.07587717036722852,0.42836428343397215,0.732047749486939,
 0.7954624863274643,0.8473740503478889,0.5000374381379432,0.24317648619335042,
 0.2645261614897886,0.3266100169313728,0.652398796599146,0.5128707590817838,
 0.35009525328906277,0.4125380467516242,0.06150050306649035,
 0.4300791654396976,0.7795203768424838,0.9379292605549892,0.11838213493278993,
 0.09426648259963355,0.12224560396263295,0.1552858542870118,
 0.9318248457822824,0.7044532914312859,0.5353526436013812,0.6031862358253197,
 0.7570075034472902,0.26663006374152676,0.11148067074734602,
 0.32558315643576075,0.20972149619898584,0.11714612197443486,
 0.8896133321963002,0.3100319407485891,0.7618403151275348,0.41950258323318035,
 0.7119519459355389,0.8870208128189825,0.24507548928852363,0.6871259224126813,
 0.2729641693193551,0.8935699141306106,0.7807974787255576,0.34320896956945424,
 0.8775248581256798,0.3140660019846392,0.5917819333368404,0.19052196748912353,
 0.459184475090169,0.19491554661591426,0.9088419303917684,0.6671092323915452,
 0.40251131498397763,0.8401519111481061,0.11908617439085978,
 0.5197654888660339,0.8196070969963093,0.2055799186945474,0.7730593746623357,
 0.07338186620897813,0.8950876283410132,0.27194064469953816,
 0.38162366942272674,0.5907010677624303,0.4006144938470354,0.5028225122767488,
 0.5030945045929881,0.1800651315379438,0.14407450449024628,0.3554758133518835,
 0.5961120167659406,0.7664013038912139,0.27406062332785064,0.4131641925217743,
 0.22553429467074948,0.8823817112134383,0.4968962461638542,0.53335339197295,
 0.9612961867410261,0.9714311084241963,0.9784324039470016,0.35911796537919516,
 0.04188403857682199,0.5573038015189978,0.4279742246592446,
 0.47635225524426716,0.7895475314992186,0.9700993803053688,0.6145831303686464,
 0.6197519186027246,0.6985328742510595,0.8978275586487116,0.4463365444554054,
 0.5413557642404757,0.6816307028455666,0.7146630135379436,0.996478104499148,
 0.8892808671664758,0.37018692197794234,0.5707881303400624,0.9922092747946438,
 0.16750180044430096,0.960817693347895,0.7659170332338407,0.8179637248928487,
 0.11989118367458229,0.5765670051194367,0.4399530661260216,0.3966063672726211,
 0.19601779086183446,0.8793386178317784,0.1678110811275555,0.8100933243030343,
 0.9583058667205407,0.9199641203660804,0.16332231423930166,0.5604300553376214,
 0.9443560318979802,0.45463551763079524,0.38943496616406936,
 0.1386720503487251,0.12076149947711545,0.6859019801630765,0.8543568161619111,
 0.2528221152783777,0.0409007352594799,0.245609507577911,0.4476286448090513,
 0.8216480777929329,0.21730074730026483,0.752100391548358,0.9396610473595618,
 0.8780549087499441,0.9815814135251819,0.9983646320938384,0.2042037220755133,
 0.5254272895806411,0.44272026508804707,0.8896448924429051,0.2039672460761437,
 0.6494300560006379,0.6274102275448681,0.5521897832336481,0.43478110677745574,
 0.8602038232294476,0.6839638593817446,0.295832518749241,0.5962498428902389,
 0.09712693847911169,0.43302900220377705,0.22490092153110075,
 0.4236555193036108,0.7069998369046564,0.6664333800987905,
 1.234397658953501e-4,0.2113073478289751,0.8634846289106923,
 0.017862064926060794,0.47875567708558153,0.949995561728305,
 0.13289153318985913,0.04856531805604403,0.6694329836688282,
 0.7568388905949328,0.7829561905471017,0.7762859557754554,0.45931590707084047,
 0.54303886583591,0.4565718971655033,0.5252766145236664,0.5642422837271712,
 0.37804756124761785,0.16854933499102498,0.08593488312846365,
 0.8580581093509522,0.518400159865185,0.8171696369741461,0.7244446115391928,
 0.12856948653006306,0.34188814395392975,0.3687525615155314,
 0.37848512063084694,0.8333498373759564,0.12800718572103376,0.400485763377207,
 0.805606439611259,0.8589319720022689,0.4935589940051247,0.4663723559712003,
 0.5123834437556656,0.6522296129873288,0.4038022604288314,0.07927374390121611,
 0.6556335603740288,0.8130412072738592,0.32391176777223885,0.3468445680482892,
 2.725978104860527e-4,0.05762108126298604,0.5698140814767411,
 0.0031283447641259343,0.22002017128324236,0.3635087568147056,
 0.817138248574399,0.967848825446013,0.08137778825570663,0.2650634630540222,
 0.5869265233926428,0.7821501335019138,0.8024876846032716,0.0743381167606908,
 0.27534844831128213,0.9441881189628855,0.03716746509992841,
 0.5784012731026688,0.8003847710500994,0.5123133527322653,0.9903564294933417,
 0.3710978459555144,0.5663252148960716,0.2783217801377498,0.3893674478078186,
 0.8519949955270403,0.5362997924517804,0.3193019356900597,0.2566122005185937,
 0.178648979782828,0.9409265855693486,0.27763086269277937,0.395862041846449,
 0.9670747924848231,0.10141005509273293,0.4053306551047332,
 0.11732819757376933,0.06387181917973894,0.5260510483146349,
 0.6683104339525683,0.13144491312878848,0.10838498890915582,
 0.9963519923924578,0.25636710906179083,0.6436457724039869,0.9792456590277672,
 0.31739857320607223,0.9919621558298135,0.4667347071810005,0.5858065781149435,
 0.6863870508062578,0.9374706364469093,0.35985960939179895,
 0.005556564655586849,0.6015100165307332,0.7438986894059495,
 0.2851184232321641,0.674756067689643,0.22170299949499783,0.7576037556506537,
 0.4671597785152024,0.566707783251422,0.4921026588626628,0.19131112244784632,
 0.5730557323894613,0.9854687479306528,0.05907077634991076,0.2901696706783613,
 0.8916916635116863,0.9865350117031628,0.7305374480077054,
 1.056725810121284e-4,0.02757739694015937,0.4461105658570046,
 0.7170209380839171,0.6848726357073714,0.45612293268353654,0.9906743312233778,
 0.39138026742234455,0.7188608638163134,0.862446295326499,0.6160424818872213,
 0.4321242005529409,0.03720470873283399,0.7991506252912562,0.9794546941355312,
 0.4106259989941371,0.723091047130057,0.9851529021994839,0.48939596144095265,
 0.215124906587859,0.5993577896476053,0.2668547311397319,0.22598496924679845,
 0.6999439607112156,0.27597742028097805,0.40377937378345563,
 0.8956915742333164,0.7332850508157626,0.1506507029988028,0.21804384843808067,
 0.9352354176696432,0.2992844812317338,0.3833011493877766,0.9054500776514364,
 0.2275252663388787,0.7095608367110243,0.7272366531006297,0.2084986004271605,
 0.8674175369313637,0.43045331578257795,0.33004063197814026,
 0.9012094896508693,0.35409045305779885,0.160371177128376,0.3978529631092177,
 0.6337378285867765,0.9407723731249654,0.4557583141415146,0.04994036260841661,
 0.3683539788227944,0.19367332404561322,0.9974056747660395,0.8862042366364131,
 0.6627339089586367,0.11943775623673081,0.9726125683663082,0.7698991772994839,
 0.7828382484095336,0.21021660461537572,0.4337602805711941,0.4090286348264871,
 0.22717513918819932,0.06344029177868415,0.13812606540079309,
 0.6698168856632128,0.4308183644696112,0.6527777172485236,0.4712113549686243,
 0.23063664937583428,0.6209342313268336,0.7849005541710455,
 0.023611765569750087,0.03589334410331879,0.09208964761810678,
 0.24512761794610416,0.15033281962869893,0.8812482510732933,
 0.9682939943285441,0.6517072831051789,0.049077804922581514,
 0.4536923218618334,0.8970126316386355,0.7791708184830097,
 0.0031565505624405743,0.8830535197437348,0.7892272573690932,
 0.611314441472386,0.0638353862297304,0.4727520510407557,0.19355156574216936,
 0.6051237062154334,0.09935784400242076,0.6585808936312559,
 0.25009513076482603,0.37111760052740417,0.6089760941676284,
 0.7595186463422062,0.5408800987298463,0.1754464084723275,0.9655952598806864,
 0.49424992353414265,0.7805476533575523,0.7203079440158026,
 0.10305755170165865,0.9278500874868919,0.9776932103060735,0.476392936402122,
 0.2195807498310286,0.5743158564588743,0.8903625404522929,0.9010695073017205,
 0.010867481340704699,0.13214664036710677,0.49333449671538065,
 0.09116911235014247,0.8555328340931514,0.23965638183592342,
 0.050977110508997225,0.5894226141815033,0.8387128631343808,
 0.8199577802660044,0.657431138253551,0.4696422847763835,0.11324776035705586,
 0.1526621456563847,0.2277437532088077,0.7416433827575777,0.9639149994358492,
 0.011840729234917857,0.4395113443309079,0.2813810108504573,
 0.8229345422992346,0.04515614635941578,0.0940007298090676,
 0.27000165025950906,0.5134215191104314,0.16698333856997727,
 0.8277914849783483,0.5503336960134347,0.2194516965634088,0.5192489304637137,
 0.7283969765953525,0.6239167869042506,0.7956145798607833,0.12044570336768112,
 0.621525984633714,0.3660404894411826,0.3785827881336774,0.9958728685399493,
 0.4970379663604625,0.6656260005034974,0.22721338910748423,
 0.07990659231849317,0.8074399832763378,0.3120557442101989,0.7235430212844911,
 0.10386340600594701,0.8917132350698667,0.7901001438991087,0.7142481078941197,
 0.8194722243732354,0.10718287614426314,0.4184414976199269,
 0.11560162447354083,0.28486105616469426,0.7317887324615907,
 0.3756836408463088,0.029905651885966278,0.6127760202878534,
 0.14954426406165422,0.3398406590697126,0.28311595860279737,
 0.0934389453780724,0.44508826100310583,0.5003074515759945,0.7679107218402976,
 0.883114790400222,0.5141834047748931,0.2738415790544686,0.5924859539542127,
 0.23709511168868613,0.503869020254878,0.8184605950250052,0.12659069813291313,
 0.02781555133303404,0.15125645439158708,0.9818683747182958,
 0.19039411203572132,0.393628864812699,0.11593762974303912,0.3924385060864922,
 0.2756061189697582,0.8927753342565579,0.27280189490526485,0.2619671981189252,
 0.9486862659083918,0.5614857834884233,0.8355454281682836,0.5330180763442974,
 0.21137343114441176,0.39487113940068996,0.8352260167883369,
 0.9286836900854888,0.44502910124098927,0.07651234647936922,
 0.5382071334023618,0.1956541736490014,0.7812728153855707,0.8921837526537324,
 0.2612937237689438,0.3649779774170546,0.047337094788977296,0.779861462783067,
 0.13534433168919946,0.8941007472077628,0.6526147861987698,
 0.26217850007601506,0.8948120597191592,0.3277213691562453,0.4483054647850757,
 0.05989106017490009,0.766378924293935,0.3036791384386117,0.1843731080845774,
 0.8741662591321359,0.7153734827544846,0.042614714161993916,
 0.3099869966209441,0.8558737500553402,0.30422645785520075,0.2076965045702922,
 0.9998791696507412,0.1460339057729524,0.9171138454099165,0.7360105621492001,
 0.872389428001969,0.49663374351541056,0.7371937641640367,0.5867435928455724,
 0.8581652347455071,0.07906734142933614,0.5190988832555579,
 0.03553195617415861,0.9519576216597527,0.7339474051324448,0.4880253572644223,
 0.9721917224470739,0.5534836893486128,0.6461820541845184,0.1248782021498408,
 0.4423992889580596,0.7576124006047928,0.5351508341351757,0.7429767832736376,
 0.4477635752127599,0.8776110248600801,0.5623523556738371,0.7102557063049155,
 0.9146958517995873,0.10393589713049944,0.17740151034582075,
 0.48114437268895915,0.6848659679268155,0.8603979951755897,0.4311385023922174,
 0.5201572297834054,0.6281759676261253,0.44757757574793655,0.5313662233001246,
 0.1291026585289866,0.48738947036403724,0.6016982647277378,
 0.24909275291262434,0.8287959044783798,0.9011001871749328,0.7577742174101075,
 0.008328404824167857,0.6410614814061248,0.30117693846007376,
 0.7939495462750681,0.9067058660192302,0.24835542607334093,0.1426257840321542,
 0.3428272660559444,0.1591162579189216,0.2584178632256785,0.2686752350552548,
 0.2632723926268151,0.5868673972073071,0.4108078677312398,0.651482660812109,
 0.8828705979849382,0.2908433198240753,0.5131009871802305,0.9858463541388063,
 0.8131415105135217,0.7083701848380891,0.28842244437281983,
 0.09379198509045628,0.792055809572016,0.9320610641254958,0.04365984979017512,
 0.73607739128261,0.029729482187992984,0.31048074477555687,
 0.049297791290902726,0.7546766102517897,0.958333139402503,0.5749333067036827,
 0.3441251690051026,0.7657509331268697,0.03338938469017472,0.6670391920877954,
 0.1518199413681075,0.25919746352038264,0.7781483593869919,0.6237619740683967,
 0.15102478901203975,0.8080310960106633,0.28122397725412673,
 0.4853225412382337,0.9961968388872635,0.5703564627012824,0.5020932042801124,
 0.36005231063797727,0.2814981528147583,0.21711281481246025,
 0.6898903653299685,0.9790834899962048,0.5636741514932468,0.3067672238764516,
 0.7168355461891572,0.6502120902878276,0.6337709590735672,0.2726084559212366,
 0.8606132084211667,0.4277425183285295,0.28447868212937055,0.4220506904367505,
 0.5668083333723161,0.5691298750472109,0.327983479479073,0.20891569189985437,
 0.7114549051535,0.8877662722911335,0.40331986624293925,0.06627917167638031,
 0.18818593091478142,0.8310730994385225,0.662178395019733,0.7145205392986728,
 0.6818734631574135,0.5600964742127688,0.1963878102857104,0.6714736817379031,
 0.4878288125447491,0.5376231859083255,0.4298774707336117,0.6757535481433666,
 0.8502968263023362,0.7822864498645112,0.8880168732744966,0.6440953134256916,
 0.6739987365301179,0.9518391076615521,0.5971463439475067,0.12855484471570988,
 0.2891306461511308,0.4657942293751234,0.8822628550776571,0.48144891942622337,
 0.023969151605122807,0.21745296395195224,0.8416110370413681,
 0.7490360620340731,0.9297544655356076,0.8002364242190522,0.09010511358893991,
 0.3590177525390025,0.12357350753972351,0.4510237604663295,0.7868780496579256,
 0.8279702652695041,0.45214389841356317,0.4287080815351805,0.6055366475123782,
 0.1848197575123931,0.22098412065287543,0.9832930612614763,0.5574075937123562,
 0.8499954598404853,0.803951527382558,0.7513460893642292,0.8562995441761933,
 0.12668329541221757,0.18735723844007046,0.39245801038701,0.24307907675253027,
 0.851902803314708,0.3499280348998467,0.9695845492609171,0.023566799680423856,
 0.8316934292033579,0.07027969527497668,0.910791143249577,0.3362361186638929,
 0.5995502626605065,0.684557905988126,0.9537650473188517,0.9595598413022115,
 0.9383809979380692,0.05249819604905315,0.3392761996835192,0.3363783845719235,
 0.5409850305587584,0.8962325697518108,0.41588996892683205,
 0.08692793334732762,0.15247918365431878,0.09445921021494175,
 0.6465346530125279,0.12312232833754422,0.7689846011476484,0.5656852161100229,
 0.22806749366313517,0.9187315499103829,0.24132095507290907,0.31860424666651,
 0.8149142734774086,0.045310161349163325,0.7093207752314428,
 0.9537123126568334,0.2927230998697018,0.9750312453830461,0.11325151563952507,
 0.6926132193132885,0.6442720111863575,0.45520178920173604,0.7606891076997411,
 0.47557756092576786,0.3081467640515707,0.9770187018975349,
 0.36096087326112325,0.8492831734756483,0.8171237025056636,0.9880102720181547,
 0.0773989656628361,0.13666416822329686,0.03151571375395379,
 0.8358466000247147,0.6137003852297266,0.015709213541072575,
 0.4407944653427125,0.6508195460046503,0.1898831115857641,0.708648872510081,
 0.5794849047159741,0.24878852164267184,0.10815160746506947,
 0.8115353126184948,0.5908340061363914,0.330838539669588,0.1016968850665465,
 0.17634942169600187,0.8797265064519912,0.6853535691002066,
 0.22156632833718726,0.6705871006021195,0.19297727566752565,
 0.4758225880111253,0.19706146587310147,0.489828503591901,0.9154430130453636,
 0.2868581416332824,0.6144757692761121,0.7083174068079447,0.22268897120416375,
 0.06367440792619516,0.792618551645592,0.8242677591837754,0.4843661647089752,
 0.5880982793741388,0.9141607291384336,0.4479234619031601,0.42061317320532143,
 0.38644880330796083,0.5296378640649603,0.9426775030559075,0.9286858625273402,
 0.5885373962577876,0.9449793039021295,0.2622973242808728,0.8194311174709304,
 0.9227059535796267,0.16880871751714976,0.03717618988618654,
 0.6208714815286791,0.7980225899646848,0.2509304024507285,0.5072052812515345,
 0.7139239217541165,0.7947970900473578,0.43721299100357824,0.6054901747829525,
 0.15325289139300224,0.8316769167365183,0.8889579466224655,
 0.0011063662270529573,0.3256380146495108,0.09778155795871069,
 0.9548396901149138,0.2473338234530098,0.6439817019231909,0.02364861331568424,
 0.86096969735908,0.37845414117245235,0.8298521302011863,0.5614622498905011,
 0.06573906624140713,0.985780014775135,0.048137584338658,0.6214292387643647,
 0.7745426153454433,0.48351206196288954,0.8328642460971549,0.3985634213969177,
 0.8562026637618594,0.657508962796696,0.6122090979054395,0.3988256056157162,
 0.8107097637713007,0.4728342940127135,0.04034280396117662,0.8374521353362676,
 2.5134848734631454e-4,0.2081286671897694,0.34340270841465526,
 0.7913896123719801,0.712083810561624,0.0641437499313875,0.7812033688601131,
 0.5645674463454178,0.7201577602748277,0.1745059440191744,0.2643621032809924,
 0.97561009775952,0.8037234761761459,0.2981859449225469,0.4344355518190468,
 0.6292977852775781,0.19955218187834067,0.4458667367987119,
 0.06821861188509859,0.9997002918681706,0.9286557202467427,
 0.15896665549966138,0.7256490250971809,0.931969074527154,0.9134284434674371,
 0.5177125592681115,0.02988365656602887,0.35872430570683,0.15952793588539937,
 0.7288728809333656,0.11064981271838525,0.26269006785798843,
 0.7397125334797432,0.9383249717219295,0.10105068296759523,
 0.12606528888863977,0.5504910664172418,0.8383966268009918,0.3881516785873219,
 0.829767783015972,0.137995004366822,0.009555330326174949,0.3588024159441323,
 0.23687314793052439,0.8409375038199591,0.012903603096564464,
 0.32793898740995153,0.3698242164975142,0.647113229984745,
 0.062081550671659325,0.16603890350930217,0.1964258356101216,
 0.8069448133830747,0.12034658206539572,0.6766545261416235,
 0.08044074827731484,0.7952720325566213,0.8267622026515496,
 0.41707310637536676,0.5703380060501345,0.5503405878415782,0.9303430646256571,
 0.5424819473642463,0.3612894824965177,0.12526075558208793,0.4507463035154782,
 0.22528971975983558,0.8410222517917725,0.3376338543061844,0.4096229203676671,
 0.13362336289486165,0.7339180770025017,0.4478138715209763,0.3381815507007323,
 0.530131231199695,0.12007090104042006,0.2495869537480263,0.15432141924401288,
 0.7637930826264412,0.9255311834904557,0.48410249661595617,
 0.07654677657196218,0.6619198672275338,0.36269059809261917,
 0.10731702587814151,0.7741641592593373,0.1740689252522426,0.5995276664280357,
 0.667085045717299,0.5162165050833601,0.4411603262755427,0.6447634319923294,
 0.1527326547699528,0.511111175155947,0.6556761230466674,0.39575334479622826,
 0.16822217768561065,0.48651847534597015,0.5192515744326297,
 0.02326010251893451,0.582533777996605,0.35858844711966553,
 0.47998614578505794,0.03176096484853863,0.3736412716577314,0.250445824274637,
 0.7856509356845922,0.1649724593875992,0.4329978304711467,0.1885726355486067,
 0.23862359822047707,0.948863364471187,0.030254947850317016,
 0.5204304109796818,0.8829477392218126,0.3011260834108447,0.7692708580630994,
 0.2973685088989453,0.4992717654174994,0.707246738898827,0.16891443759457614,
 0.7738848279779269,0.01686723730678985,0.05230659546173566,
 0.5662633043786074,0.18881522165464248,0.45186066176224504,
 0.5894024982894441,0.19331148050539548,0.45861880064697025,
 0.9255753968140832,0.6127740158799866,0.17221671163011654,0.7386202987237953,
 0.7432672882220157,0.35494579863116826,0.7756891190454273,0.5337185886078167,
 0.2215555479126723,0.9717049835911961,0.43822333749535514,0.838481448624748,
 0.9202642378318899,0.021992731895136952,0.43133306761646595,
 0.31152215040985687,0.5969043049614918,0.7069864194730151,0.5301249997929767,
 0.15593506118133427,0.955453279621218,0.1996583919633057,0.8159972102482105,
 0.7542086156869698,0.6420665070465563,0.7415433720283047,0.2030362427914939,
 0.8840333515136427,0.8921055742676438,0.9004593074887929,0.762654958441069,
 0.7057283592172856,0.06536950573874445,0.33722828452092013,
 0.7173212216992982,0.38898995048460383,0.29773882399074525,
 0.15643141581179276,0.6403501750380924,0.6418416924291074,0.5904123609773848,
 0.015149125095375804,0.5410025415350844,0.030410871050127075,
 0.41832709192476913,0.8652747229382538,0.9418135173864267,0.6304597102149011,
 0.926731376489271,0.5900056247728079,0.4080245021040977,0.3811968156714771,
 0.7900684732915412,0.3151926850273442,0.19398316807086224,0.9392676497327437,
 0.3512834690496752,0.7798476268198299,0.14169936923848026,0.3467971610459897,
 0.1321743169307017,0.3337136359486548,0.3817808268486971,0.12894584198022097,
 0.7808186203393133,0.3473383219858619,0.7681980736373101,
 0.009807338263651766,0.8695984685272033,0.490552602170256,0.5932237803847391,
 0.5090185933898491,0.06941226925177957,0.1721658050554581,0.946107051309694,
 0.4940585754456762,0.2404661818063668,0.1604492040351917,0.8191089907471615,
 0.7733756510035229,0.884901614145344,0.7726013638238444,0.07547110098253362,
 0.23968192345976969,0.9281589024732962,0.5717513886336245,0.8396334543828304,
 0.09848790021164422,0.416651011533549,0.15130170332916892,0.7067718125117681,
 0.9294144416835474,0.8110076694918571,0.5941474711998985,0.4194997555044978,
 0.8072520137427271,0.013579924716723424,0.7518694204093658,
 0.01057599659737185,0.7628032531772055,0.34426671247428264,
 0.8283430450608469,0.011694636890115961,0.7920120511627838,
 0.5224394961459252,0.6928794252825677,0.34376369998913614,0.4786501397147477,
 0.6437923396697157,0.0859640879645519,0.4247609388010618,0.11294335823516144,
 0.5613457550719658,0.3657259630759043,0.4746988116889823,0.9979275606831306,
 0.08143525707408839,0.4471471175072128,0.38080364197486305,
 0.9373306982633314,0.9426425531978194,0.12432340786158858,0.744747806874139,
 0.3474434572392944,0.06287699255848744,0.7143255182433189,0.5942886461733295,
 0.18841837758395796,0.4543730947999063,0.30943292068773265,
 0.6783902811623459,0.10729336967230463,0.9868305133278237,0.7711733090819449,
 0.8523949313367125,0.9721474299042601,0.7855479989728978,0.2878782293055061,
 0.7237880249961619,0.500289461509943,0.20457983152393167,0.3158245647589558,
 0.6450771292669794,0.6412046827185716,0.40527915057117714,0.8199197943444685,
 0.3353392801368261,0.8826250589769559,0.2888753500775617,0.7906900585843848,
 0.4140519866130967,0.719505366102872,0.17539244871807602,0.47858459886326354,
 0.16382028758936362,0.19486623878021103,0.2636621599724328,
 0.7333646289809094,0.028011091622346607,0.16157537336810057,
 0.7702479241066462,0.047299486062142515,0.455977437881426,0.9805929867298095,
 0.9325790027829685,0.8810159761154194,0.3574574453758572,0.8169623716099581,
 0.8227894862523446,0.8339921962085032,0.017375084375407135,
 0.025223099535902516,0.27907597100606396,0.6447804628457864,
 0.10424579104000709,0.2554265207238541,0.24448688793868578,
 0.7101043525038587,0.6671703635886497,0.05141871070068005,0.5792353429350776,
 0.835381829553834,0.3512287106552946,0.8527635378342627,0.05765507418515037,
 0.10867520298833133,0.08276961539526873,0.05980765053263126,
 0.07844577005441034,0.37265764882443175,0.37209829938386085,
 0.3128213541448053,0.8406322814448906,0.5169815578824218,0.8431039281070893,
 0.024074104668494023,0.24039467455112484,0.908638267216955,
 0.7748539848861935,0.05648431248844665,0.2950220393017938,
 0.16666963580066518,0.6007108768121681,0.5078351545381438,0.5639218826585408,
 0.6648264220099867,0.8240482703410067,0.6161443587945349,0.7090786215561831,
 0.4943518872486179,0.7281655359104219,0.6283073113883677,0.6030190323602653,
 0.337139248588739,0.4953870074532911,0.18897895228692696,0.5746122624385789,
 0.6092794817783213,0.3222087344192548,0.6923236001309306,0.08727034485899832,
 0.5287658940822313,0.3356567249343785,0.012473974411061683,
 0.6748635546611333,0.3585893193336367,0.7342794225316572,0.2899201775908655,
 0.35480670646108137,0.6912502752582239,0.2075489403972648,0.8002675071493892,
 0.21388207360980327,0.7641412487132349,0.8384795464328036,0.5456400189627193,
 0.8036100151654396,0.8015141825209758,0.004362643006286637,
 0.35449919171043565,0.49120401992177076,0.5724415668598164,
 0.09627862764574036,0.0754302559869705,0.6412958653829953,0.1561566960033035,
 0.8968484175139225,0.7243972509264938,0.8842111483956889,0.38367781035856074,
 0.3148725145232021,0.04750319423291138,0.6798181802177232,0.69543413460521,
 0.6629649057707522,0.39913957953961277,0.9592238188964575,0.8585733644575445,
 0.09600018792538467,0.5547808425457977,0.7091181224197965,0.7818283697976594,
 0.9147936003888059,0.9202727756754161,0.22504119900028274,0.5897149562300744,
 0.661047942983732,0.8158211466519241,0.28231401626922836,
 0.018648592449493573,0.6503617183997651,0.9701435745758267,
 0.6230708377787872,0.1900537778419804,0.1766369852535934,0.6952188795971643,
 0.8625763906255441,0.3701334142667624,0.9413424124448786,0.35946573848467933,
 0.37685932695799995,0.16833021376458557,0.9407287674143012,
 0.8853007334361096,0.7165956677918839,0.4667505933485474,0.9145702181438907,
 0.5860899541398625,0.32630835539809844,0.6705999750561029,
 0.23057835933644688,0.8433521295132403,0.9783498660835356,
 0.06826380635077622,0.884329070940088,0.5271016299697128,0.3411000179778938,
 0.7157721377214049,0.2235056047488453,0.9117251737251862,0.2712362421588651,
 0.523896586810273,0.8339082796904332,0.8763191863153719,0.9182382831104023,
 0.16337892153986078,0.33662181412260694,0.396113629105328,0.5717297551147653,
 0.04553707766920867,0.5934457221231025,0.13042161688138976,
 0.7378321432330662,0.5945362811529618,0.16328460927883837,0.9618103238305643,
 0.06251908972258657,0.8199818592607834,0.00534920892886781,
 0.7873118054278444,0.6709104786284159,0.6331511155716549,0.3175990885302502,
 0.5162406290226894,0.4421163465673056,0.7972698087107162,0.6496167295003201,
 0.9675565388961637,0.1510263276681736,0.18988702606025454,
 0.31713678121860445,0.45314588194450245,0.4338318140831572,0.857018704801166,
 0.16467301155453273,0.4234346565674675,0.8620056367737758,
 0.16029172189969498,0.5886048504776629,0.84400415787614,0.5900702181914396,
 0.001849463559691289,0.0316957228572492,0.8807556080025467,
 0.9863822651735309,0.3149690459556532,0.9108887837129371,0.5721620816220586,
 0.9917015121687185,0.5398414673156273,0.8129002044897354,0.5629077959290945,
 0.6851146970906364,0.5368678804432396,0.9295972030579608,0.5620691835633695,
 0.5681373448484291,0.3760353557474706,0.39975871992376477,0.4521501817557476,
 0.16190614088848376,0.3609974758668518,0.23370596670042354,
 0.9600512516473312,0.6769347356269135,0.7903286701499914,0.21877886686246462,
 0.22969278991873765,0.4753654828198631,0.32091366459279547,
 0.01338969618750585,0.04249274450513074,0.7076778271715288,
 0.5195258903094255,0.5700384399190568,0.2082226020415624,0.07930889933858598,
 0.8862718896469249,0.9980157270575966,0.09602488312723012,
 0.24785940158875197,0.1384264411841265,0.352710830358938,0.7745590767651367,
 0.44758992968818034,0.6323902591652923,0.7160984797277865,0.274798600628544,
 0.9872026489121494,0.9523086779024005,0.669499355595865,0.013134585945451471,
 0.9896768043745791,0.19238064171304758,0.8497175649900542,
 0.12288201729808712,0.957631721419926,0.8412927703871289,0.07985468111363192,
 0.7096594347881382,0.29435432106759796,0.9988696517008796,0.4310683968310586,
 0.47098891435017676,0.6467351429340488,0.9053724448936588,0.7198618527448885,
 0.35339698917866447,0.5196564335099643,0.9595107945224735,0.0732051341240646,
 0.14206815146916774,0.8616288258651441,0.9803694135120046,0.5266593436720455,
 0.031150446351903893,0.7042149577499227,0.5885933265588539,0.935329589896345,
 0.2686322098689873,0.7501630802392438,0.32918034391398265,0.7422142738193711,
 0.8748887191110308,0.5761417895014096,0.19643501464964586,0.8551502593545187,
 0.3672713497755551,0.7168104633106296,0.5994857838326595,0.5329888848195528,
 0.9389626306321343,0.9932663219067855,0.4694444083847067,0.6896359677787347,
 0.043031876928879,0.6239004811367197,0.0061246428515689955,
 0.5635323989013692,0.09144599677011922,0.3289516766246656,0.4868074348809568,
 0.34208711339054254,0.29981757268114195,0.5059314613449153,
 0.5382133948267536,0.4352346201293076,0.7671142842255808,0.6606687551609565,
 0.3081641247707205,0.18844943528173352,0.49348470325303606,
 0.9553458631384675,0.6633613305519395,0.3387978467567482,0.3463461788505232,
 0.12731618640821574,0.48253019354165083,0.04867468780513162,
 0.4564396240296561,0.602125279185993,0.5407184985229427,0.22923953995067925,
 0.11279139718348907,0.5145558632120746,0.6418992092795751,0.9863795957783551,
 0.5891161009241281,0.15668404231299204,0.035823098187538704,
 0.8575844051822451,0.27402782125737646,0.19439552937520443,
 0.8159947786397717,0.3868720511928677,0.05369603485295549,0.9249297427598737,
 0.7888853941713138,0.9077116434036847,0.7612990355036886,0.45384720464572936,
 0.8132794318770403,0.5790941322037718,0.47680639138376024,
 0.39490954685940594,0.18200899434414342,0.2764312949721749,
 0.21230119054619345,0.11983313172048493,0.030521673012716577,
 0.40086182797090486,0.8394914572532719,0.5723535196758919,0.8829668429412907,
 0.4609154339766206,0.43147239811636906,0.5361528414437705,
 0.48262979058559796,0.9435901788978237,0.536084665212215,0.5437450608199691,
 0.1681932836667479,0.14982994838876584,0.2772026702551813,
 0.49291840647979424,0.6286747057900156,0.9156138536199248,0.4754484630024156,
 0.9081073565627182,0.5645498964359108,0.42457758085403674,0.5541655812865706,
 0.8100836720402684,0.9291365263418665,0.19074061789423025,0.7158251897085176,
 0.9477019644178694,0.17092980640394073,0.1608157994340369,
 0.12105534593564204,0.9634898286010911,0.12030974203326383,
 0.8789271613034323,0.1879271117125736,0.9237683033191779,0.16715163228503604,
 0.18779588567962147,0.7277947503264246,0.38307629433196266,
 0.8751455842914665,0.42449321701735365,0.9730746185715942,0.4930921323315002,
 0.11373611186971067,0.13093877795915831,0.7401649818342008,
 0.24274607356881805,0.6283043472435729,0.6923710151132476,
 0.08073594633139924,0.16716017031081332,0.0416287075252646,
 0.03817386034394776,0.2709110967413706,0.8984441973131827,0.5037680235800484,
 0.20694687166609094,0.3247697489198166,0.08130577182617404,
 0.0917926820227919,0.04175350339878381,0.1051123447990211,0.8320513140589694,
 0.644974970080207,0.3796786543591093,0.3868245488753639,0.6861582810060458,
 0.7511578153896623,0.9474385137970399,0.10521161492767472,0.6510556348412185,
 0.7469304834112879,0.7410499124447034,0.001671860754277943,
 0.4189040239506554,0.16193027959095496,0.7417870787996135,0.5501633883476118,
 0.2525561864552913,0.8742788411429829,0.1353489787136659,0.8721596802237839,
 0.8210802721179572,0.8504243656907498,0.6004400577588022,0.6819647053924268,
 0.2172846265668058,0.5348285370712866,0.9523744692460874,0.30563030113539624,
 0.388807873814183,0.12272194004284753,0.18758257649644572,
 0.02877045730910477,0.39913021285594374,0.6688980184900756,
 0.1051055992835056,0.38648019909149833,0.241135665309132,0.9090008889924942,
 0.5827569577913965,0.17232088188047145,0.27300394780355897,
 0.7162702335447874,0.9692683217555458,0.3318888722140404,0.06748905424337726,
 0.04598775852200032,0.39036748770527363,0.09182297714723786,
 0.24882256269447822,0.6871839246050429,0.2986081062546302,0.5720165057439471,
 0.10361288634924326,0.33165349153324286,0.5093945942539988,
 0.9546673277978408,0.5117201361133683,0.5357659623580362,0.6907712352098747,
 0.04795882360686565,0.8738814734231314,0.3437401280624861,0.6711151586132841,
 0.029278119960966187,0.1908014011970044,0.7742899024463962,
 0.8534451986269688,0.3721224233320144,0.7728664037649546,0.574246297994041,
 0.33689104022731786,0.8992513345247316,0.42844209144329726,0.66268819893046,
 0.5476042245476473,0.32468198315794916,0.11513642622116593,
 0.4395998041846624,0.9068982590600059,0.5742799876434085,0.05736037815812334,
 0.4643870230894832,0.35885045922291936,0.30903384927721345,
 0.9179549189452562,0.5576552212482343,0.5243745320110302,0.2430948220114959,
 0.7273928767012701,0.00922368690880948,0.09645902785435401,
 0.4304544254104131,0.8188605422860016,0.7101953740346537,0.41312127688142763,
 0.1414533223749973,0.14533323969285483,0.5702172373961101,
 0.29131292145816345,0.013051512682068389,0.3809822445878476,
 0.45275613190283337,0.1603202952018784,0.11344948300246582,
 0.16851361774113616,0.3028184751544345,0.08612837351444047,
 0.8842212583419793,0.5829027959116707,0.32936698926806196,
 0.06669956691876977,0.6967042160898356,0.401599556649461,0.685589838839058,
 0.9754868759092695,0.46991820232474,0.5522264275465949,0.4517591904498305,
 0.25236655282756637,0.3061942759754871,0.2209652805719411,
 0.029753540365627162,0.6812585688237728,0.8849210032635451,
 0.4506814144129394,0.23113688685153588,0.6177658449202579,0.2624995348937852,
 0.572948700799408,0.1418114338118257,0.20380123150712903,0.24061988954792024,
 0.04123954873971103,0.0898933052506693,0.45547957844603204,
 0.0920430183776405,0.908041605563,0.769142936800371,0.31594510227727357,
 0.9826280357932936,0.31215311410206226,0.7520929529223842,
 0.015225912419885357,0.8126487463371863,0.06450246796154857,
 0.40506676581036327,0.3886037562348952,0.08909598166147925,
 0.8681530208603806,0.14992874476401064,0.18224267031181818,
 0.18523123918319195,0.552324408097177,0.9420247939403849,0.5554120649849543,
 0.27376695391412775,0.876791782637163,0.45345778847661433,
 0.08993709506359482,0.23041334982765305,0.5262002313899476,
 0.49127685342877614,0.5673816926920043,0.20632252043035504,
 0.8545234016952579,0.1845868259806056,0.48056415538160335,
 0.24715207760085944,0.16853944007141541,0.3013307969585226,
 0.23884931059157544,0.5894367733038829,0.43195076806945965,
 0.7876559678697265,0.15728150404464358,0.33259841024558234,
 0.5422031280223887,0.37137843414331806,0.9653159259180023,0.8087436300805115,
 0.39246933181559585,0.5196597255388824,0.8076196384101493,0.4584710345933667,
 0.9464748252774137,0.9680428301882329,0.9636707198589022,0.648856534783748,
 0.9543808143457483,0.372728887724208,0.727917544271962,0.795482002149876,
 0.10774391673270645,0.8860330780675083,0.2624960614160464,0.8345955011398266,
 0.0331865446471874,0.09722859924067961,0.8652678865713593,
 0.24111336930136917,0.7658612221486054,0.8317065615549782,0.4131756505864391,
 0.20261154312242302,0.08114569612877875,0.24195105828901253,
 0.5951458884851334,0.39791994476849957,0.9006876825807516,
 0.08648540819781103,0.918238850879852,0.5801661918632629,0.5207342080481018,
 0.8375795845087561,0.1510880470383985,0.9651575931210701,0.39115746031673915,
 0.2711839962563908,0.1187417316150371,0.555588647441024,0.21147624299708345,
 0.7925442690462153,0.027103378010818302,0.9330278084707992,
 0.5801311752119865,0.25313297125787326,0.996806885619638,0.5670763084781754,
 0.7838447645681592,0.2459909524134709,0.7077565963148714,0.22589233190367985,
 0.7295392165746015,0.8814707067050422,0.08604033766397912,
 0.40184971958007043,0.456095071618316,0.7570865793714754,0.6434248966974603,
 0.1940784930641528,0.7688382041632618,0.5929343906009159,0.6696695160927605,
 0.1870932973535211,0.40914136798694756,0.15511897869406555,
 0.20213529023760768,0.14352193694379067,0.9970305006167941,
 0.12596223138760154,0.4649593617318635,0.13492892063105344,
 0.03268248930176676,0.8774716202968876,0.5335127611571472,0.9277147820446559,
 0.7490709844204374,0.8961681750652472,0.5562391771623523,0.9642019343958268,
 0.1120969793268809,0.8685340885683299,0.17937210119487046,0.8417558081136922,
 0.5243431469247137,0.7537389739796817,0.6346860232741034,0.6454882107909701,
 0.12745168444462096,0.7605698461183958,0.6042061449487756,0.694201632350131,
 0.06439219622239212,0.8272326015785613,0.45385883454721654,
 0.0721722558213328,0.7143308617599091,0.6843759855822714,0.20328909894551073,
 0.516955655128895,0.38126713522767286,0.12235631734521335,
 0.17999624965990568,0.9286115549515583,0.9558111485033614,0.3625989853129584,
 0.016553386350347754,0.8582833174741898,0.6129456093930054,
 0.4491937592667766,0.6785697380624949,0.8148915236852983,0.408772045713529,
 0.7501063946860951,0.6480393954980805,0.9913946378133045,0.127484032153506,
 0.6798312727163963,0.3683875674641006,0.9116470995071037,0.9568563450446208,
 0.6982794204760376,0.22505254687517995,0.6081879325368149,
 0.06450216162957689,0.08392404125010544,0.3267174213255075,
 0.08596183466103757,0.04018882257385714,0.9005557254319629,
 0.05522162129142072,0.1925305057893667,0.27117297580005284,
 0.08869859927699109,0.14305638779534424,0.03261008196151727,
 0.7433316448514569,0.19325665917638568,0.670974905793174,0.12343296420848904,
 0.4349351045856409,0.27930830712893895,0.22041093137774503,0.869611005557034,
 0.46495124989328007,0.5638561614132966,0.00560256638715495,
 0.16518210311759585,0.3783435066289511,0.6898403967108278,
 0.31592852818596806,0.4759203689392979,0.5387910667365478,0.8590622458508133,
 0.0705241905495152,0.01339128158955949,0.16761252523843506,
 0.8807884267722463,0.4495605681562662,0.8770760506360739,0.11550540829435929,
 0.04219993273197964,0.6691997675716743,0.4337303523235332,0.5983106147403416,
 0.11140362160195627,0.599852512540612,0.2183427485143098,0.2531299948103577,
 0.6784112210926209,0.3268142212443783,0.9636484466999861,0.08466948601475466,
 0.5232203687091259,0.31906804982394654,0.26392848395314383,
 0.7381922976187801,0.696227533629187,0.3419958590963684,
 0.0045657303164332586,0.3205079628629224,0.5059926361146792,
 0.017484746357001035,0.6335167404418198,0.988950861442327,0.3723475492241438,
 0.8256285741870286,0.8828909907169994,0.6904881887362704,0.35276601046155265,
 0.1480274707715864,0.8197558194365803,0.4398662470948729,0.343184274635577,
 0.18665429501533248,0.8715961109495772,0.34327495326019164,
 0.39866691810049926,0.4204681713865732,0.5433905208386116,0.5338341742810877,
 0.6149795177864085,0.7378003132322417,0.46906503816691814,0.8030966571709155,
 0.15101385863812955,0.570145773244025,0.6012983235281687,0.32452639014033857,
 0.7625210144523851,0.5200644886487265,0.9812415878858016,0.6703240166889659,
 0.690753224646347,0.30356496530055277,0.3441089837593687,0.7574855330240489,
 0.05450739581566022,0.6646391081974484,0.7946998947747519,0.7169290387051463,
 0.7315253689193656,0.6571249253399103,0.3181139339618797,0.607791493601381,
 0.14000769785499245,0.025142681993324367,0.4496529440923709,
 0.35301000248034176,0.40937954621793793,0.26730055440958234,
 0.38043795430079486,0.9274476214681225,0.5141319480879678,0.4399803733963289,
 0.3420752048493332,0.03849569907770467,0.7094785210141944,0.9069207537266156,
 0.2345873563283316,0.7179037012969253,0.4525783621965793,0.5837525680611098,
 0.18983521475251242,0.09322685798141483,0.35759385391108955,
 0.4224662976621937,0.026030176580690956,0.12343047295936005,
 0.8793719969449714,0.29352768672049,0.8749072109474431,0.9986326783812294,
 0.2947380048154459,0.7514964006192524,0.02982633437030424,
 0.40333325628050165,0.5738896636728983,0.9200591915734856,0.9085250811445713,
 0.44012181635654934,0.7367323502143631,0.2610159458761293,
 0.02697712464412183,0.936592934269602,0.40510916304144695,
 0.40301998781190473,0.7007799823408266,0.1277405890672747,0.8729729429430443,
 0.09404985929335197,0.43680130833736797,0.6300751047055587,
 0.21145310235487236,0.294366835233592,0.4063590135164793,0.5619994042557112,
 0.7065409202541724,0.5619760298347432,0.1747660306118074,0.8181974974005751,
 0.6495721776304041,0.02149374416341643,0.49213370463192685,
 0.42469125432836585,0.41754163889837237,0.22205333519401593,
 0.39909441328071393,0.21820023286528678,0.8434581627764259,
 0.3643408263564627,0.4077788013379249,0.7853131773470913,0.7301449851002967,
 0.1913594932474051,0.07386200992456238,0.4228701059949884,0.5454824588022524,
 0.4692700314128294,0.6350632435292373,0.5764774842571019,0.4662318833156044,
 0.7865553218961887,0.3362190654479592,0.015597410778999898,
 0.6069114390689679,0.7937862238034996,0.8186137592311292,0.07662933578027953,
 0.03943177955102706,0.9168169762015447,0.5449288355804547,0.5414068838643216,
 0.23002769026640757,0.15750830370646607,0.7208253440545351,
 0.3645263093551244,0.26269472211983713,0.34878902251720856,0.535981966157335,
 0.5178439660153515,0.7393951947061863,0.5206074423006622,0.4772553597795429,
 0.5917390884745696,0.38201581436095233,0.31633904555788206,
 0.8048855450043211,0.10201189699053737,0.7276073658886235,0.6222343954101094,
 0.7351913268702523,0.6818845299503008,0.430963451433255,0.9477970364864017,
 0.46937734087850136,0.9497311910607142,0.2490095177856696,
 0.17770880735459726,0.8293789129218707,0.3701207069814041,0.8693142598973533,
 0.010537045721853655,0.6770390317722509,0.2172893554283042,
 0.8599114167083115,0.5303529069997671,0.3219206135950686,0.3642075228717435,
 0.639810553572162,0.8673665424520122,0.5180757779495424,0.24106307493599044,
 0.13216323671234065,0.7747779248452267,0.12332316356966011,
 0.43649092128484324,0.8482406991423507,0.6066101553200328,0.9619050214074749,
 0.6535583404707522,0.0832880923748498,0.8366910079248457,0.494514594671001,
 0.20250673219942927,0.7375040670320147,0.44571346260433065,
 0.6781624228798356,0.2891567933865401,0.6764181043691083,0.11013844365418457,
 0.19637151272800946,0.6670721114227123,0.31653617394044975,
 0.9584121989298212,0.887223816557532,0.8420749428844024,0.061928453846815223,
 0.8438587550714487,0.948575328794647,0.8906279405526728,0.5656967940797526,
 0.9236732946126436,0.9734034511812814,0.018429728325318973,
 0.8696309202618513,0.8683902249008406,0.4682344420173812,0.7641228754996338,
 0.2599768471956868,0.6294569342329743,0.4180355237983814,0.3602898938628267,
 0.5003882574794503,0.7309542167481977,0.7662541793470479,0.6312237907331048,
 0.4393364807593494,0.29749319155884546,0.46406254391458834,
 0.36696343979828283,0.40759375912551876,0.17503151013187468,
 0.40649885961188204,0.565246441713584,0.4618378892931544,0.07097131480004581,
 0.25107634757363473,0.6747998120403205,0.06576047198809554,0.963227771914813,
 0.33040684326836967,0.31221292360427033,0.2748708480958504,0.842996568879433,
 0.8250243526402985,0.14654294949240798,0.941621489598371,0.428294868274571,
 0.21227285801423357,0.6522290246454401,0.26153528996668385,
 0.10781725247341856,0.01068514080312677,0.14256537235703892,
 0.7505178544715351,0.5909904726146153,0.6069737901024524,0.5703605359283352,
 0.5402637098281965,0.9403777881491947,0.05982951937561087,0.9043827416525239,
 0.1727733398954414,0.0516767602679179,0.27181245337905857,
 0.43772695538951867,0.6116102861944214,0.9188664093800245,0.6095483402026902,
 0.7819627002739821,0.34210807962548684,0.6426369576407021,0.9354368567848464,
 0.329537085058509,0.024975647169427684,0.23652897510315973,
 0.7497421085514305,0.03395654205100462,0.698442910299451,0.1507104373058109,
 0.11683979324662652,0.11455539296144557,0.11649449957683145,
 0.7796762288826652,0.7966498988793876,0.2583833516537941,0.0771081771747838,
 0.41374221406614586,0.4336932379630599,0.07406125767288263,
 0.2745358262822014,0.7256101674388806,0.2950524920343218,0.8848343174672956,
 0.006830295211679549,0.9048422574544701,0.4266958062929913,
 0.08820431961765896,0.11910823138012039,0.7367769399978049,
 0.5719613019745668,0.7166084472061236,0.1974805210874323,0.8827337527912554,
 0.3469185323571963,0.4042233304044446,0.9593243064332462,0.39898769190957006,
 0.8153885392004764,0.7105966975552565,0.1411118616679039,0.06972888570659364,
 0.07517838981846126,0.5139358505440659,0.23375456234448433,
 0.24363400987917716,0.41743110651291127,0.3712347264339113,
 0.6342721579611026,0.40253731647612745,0.8750771241440547,
 0.24443055366798072,0.2728481253910372,0.6916447955361303,0.8619886849340918,
 0.5812928069491131,0.008815723232224237,0.06469388807363285,
 0.20065135586639682,0.20194376161658623,0.9641051122362275,
 0.7940997149334079,0.44740461640842655,0.4644634960121091,0.6107502623975879,
 0.7952567822959455,0.44242181519522716,0.9267905912549297,0.5370706439076325,
 0.5067154582605119,0.11733314414690332,0.045486167060717175,
 0.13144046807798082,0.006295798341327208,0.6008467648242016,
 0.8257048732739385,0.7055020210506313,0.9004809968431126,0.7717201256858125,
 0.704718819482858,0.9731247769467438,0.988275198158529,0.012645605995632914,
 0.7898082707313839,0.18696487020737695,0.6614876185681451,0.6387824157451486,
 0.3542648927802716,0.36236656262616307,0.7994704688837657,0.8116270665000339,
 0.5470295470021277,0.20214067576625794,0.8618739440594926,
 0.38332603984403457,0.6339020929609982,0.5059566758584342,0.9209186258223157,
 0.9321448402296204,0.5614165400243889,0.27457584430499593,0.5779795183702753,
 0.4140064020368852,0.007809247531281982,0.3884891815169762,
 0.4965656151393908,0.9593022001444623,0.7466724934462527,0.2510375463384993,
 0.7718873219054962,0.7437388177516897,0.2895593132950013,0.7230523321190179,
 0.01337302604863655,0.782903102924479,0.9960324964773797,0.4352266775426773,
 0.43614978531893844,0.7858119138095172,0.20181898190807612,0.787740643320324,
 0.12151753129576026,0.35038799253137753,0.6805915380142203,
 0.5524032113818068,0.9177842296554094,0.17701124189912376,0.8992048654986238,
 0.46135531705119537,0.24447473046067736,0.9560416297412676,
 0.8135583915789232,0.7898354722115544,0.2256740335130527,0.11851207589986668,
 0.20760940976154518,0.42701766413772846,0.7422776233416938,
 0.3524040636279211,0.3376965924932879,0.12479380881251267,0.6543654526644214,
 0.7723036540425111,0.6482732657220215,0.4818806834632765,0.6746496817660614,
 0.32413570772089473,0.2775490406226838,0.18613699118929117,
 0.42833226240075717,0.8991288618864927,0.05867543056132829,
 0.3088979887861084,0.950130485816169,0.9148015646215799,0.4905819159787512,
 0.6923884880198026,0.14886824650353836,0.6200233316860333,0.8281461486884932,
 0.09710674522351648,0.5260759795061379,0.2536415722594255,0.5600631765995489,
 0.7234492516686345,0.8706165149152498,0.5194592785202108,
 0.005670567922010927,0.9314124537946553,0.9890804495986317,0.961946397022412,
 0.7282930500340107,0.04225611119319328,0.48047592720311705,
 0.4817429593726804,0.4013721905486638,0.4147511759288195,0.4003221048991463,
 0.11642838666645194,0.7767118553199565,0.7003752075166846,
 0.06046266888495799,0.35861441423822826,0.35610338593009794,
 0.07611228808141668,0.6792032881113503,0.20447659669200413,
 0.3721699893183865,0.1913660503902992,0.796642234279089,0.8647035932311385,
 0.9392405249735879,0.4150123562560003,0.7114661854166825,0.16932618256616194,
 0.45481027428968235,0.9778744019884854,0.5919340297156129,0.3983271772677963,
 0.9477984170856941,0.8972173459356912,0.6715383243102403,0.3682592257245829,
 0.4375628499604931,0.9139695872937924,0.7461284152499937,0.46212531791820965,
 0.7756592161634592,0.10323253641621777,0.7454960958330012,0.2855529303856198,
 0.6058578606294152,0.7904473752192479,0.028558860593886792,
 0.9361083896494792,0.9346394413923649,0.6795235339157103,0.9363613838423395,
 0.5082080267643594,0.7808695158183755,0.5033690534537233,0.5401739686576212,
 0.5813976570312099,0.09382034097124758,0.8137643201541619,0.2779663404506928,
 0.8641176012349155,0.4319643346556624,0.9289462506597536,0.15574726954726836,
 0.5428581514847426,0.34210419817366333,0.3657407653220648,0.8301119171103111,
 0.9236649873711857,0.3922375522143109,0.3535669515032466,0.8704536791226319,
 0.545166773542667,0.7592322690295623,0.7132688175465812,0.9122776011324922,
 0.7450528087485677,0.029108499442664915,0.24797959341449005,
 0.8222038358537654,0.9539146194773095,0.7541577193467657,0.08566520422694013,
 0.28199972452828115,0.5668790245433207,0.015721069844703006,
 0.19972524205243058,0.03633938825773819,0.5382726490423182,
 0.24300018549189895,0.18529316040392052,0.2718412163361199,
 0.9156527796355043,0.9584138519087873,0.7363109759965999,0.09438132329847315,
 0.5413141218616097,0.6092204451366662,0.7492036849144146,0.2923251018470179,
 0.16574773202006055,0.4842940108793362,0.8969196803143112,0.7534330835717792,
 0.2824714160686348,0.5538567989909414,0.9189021438403965,0.8366499669791303,
 0.3749007532825316,0.8010489476172131,0.22687197155743133,0.5764578277309267,
 0.6894521135055017,0.7575786295097573,0.7155029504914459,0.8795944976699523,
 0.7546756826366573,0.17934818876534542,0.29825698012279955,
 0.8415305254923311,0.3909986808706267,0.2108091102052193,0.7042680668279502,
 0.32691843044189905,0.9784799652735556,0.6496068757958591,0.4475470048954333,
 0.5497492937585056,0.6872324698604041,0.651344487736912,0.45676015127317315,
 0.0900852329906836,0.762061260491003,0.007982137030644099,
 0.44260180946841765,0.13738746078268527,0.8624332946054639,0.313792645460782,
 0.9614901002710375,0.8950048474196559,0.618488379076913,0.5445083977822753,
 0.7486354934971813,0.6351952309396858,0.21068861984518472,
 0.01506721438650338,0.12815845399008574,0.5625222387371005,
 0.34503966889866555,0.1335737268823406,0.8820660038446416,
 0.032522692587170554,0.1452713986463151,0.8207385419074478,
 0.2825217054398993,0.8957995955127863,0.5795787966584695,0.2056289671939665,
 0.9373146229097302,0.8265763286454286,0.8554666610067684,0.9148909620529874,
 0.9794337439749128,0.23288968027527668,0.27843607007618243,
 0.7115181305181527,0.7821365867161865,0.19840733339931738,0.5267893090387328,
 0.5886046851849668,0.5599929424371843,0.38805254549559365,
 0.44062846562642555,0.5720755645779133,0.7366214039711587,0.4118946021290534,
 0.9693179838809003,0.9509081503673585,0.737740539858777,0.8866924358104842,
 0.26176093836507586,0.5343993126801398,0.4090111449800107,
 0.24962789684586362,0.8987160187324275,0.6581407946823519,0.3476681921150667,
 0.2579480022532006,0.5603770866975909,0.993736359223914,0.6902312757202791,
 0.012434080375737788,0.25730290106393994,0.1527649719148968,
 0.08989353281513324,0.9076932164146717,0.7810715421624312,
 0.15661665089360366,0.14527300638450102,0.5325158285789936,
 0.6698598685066984,0.2745931358765501,0.23576623316930856,
 0.11019442871647134,0.9149230414221413,0.933509837310337,0.9825024588601126,
 0.5920949771281577,0.023079717718610393,0.5228180031302372,
 0.4927621436457792,0.6303640102311924,0.034330718945997596,
 0.027010534468202674,0.021849294908917183,0.017618298797552723,
 0.7168685254285085,0.33019707790969177,0.04037350980876775,
 0.16987013556301733,0.9341503344105544,0.42832140947973885,
 0.39622493955804794,0.943264955788667,0.3666983913937769,0.8999287678005915,
 0.9169847727776153,0.8950299821213521,0.29707186838217425,0.998066465574162,
 0.8537236058329828,0.794555828528228,0.8240193798221404,0.25353440064026644,
 0.3901982014879224,0.27213747195255633,0.3938970560127424,0.4249353261287807,
 0.46435195886616076,0.34229682984355503,0.30895497066347877,
 0.41259256843236003,0.17832204652824768,0.4601497766611876,
 0.0731193872168161,0.6747950937715617,0.41973711449995976,
 0.19812836402045353,0.9574566618032492,0.3767958179638149,
 0.15494404997818945,0.8141363085962632,0.8415954707107896,0.77687505041557,
 0.6721751783029722,0.18441100264678856,0.7064231115696885,0.6626094534515371,
 0.4368073626872768,0.8304231690875001,0.4176879223841009,0.8864629744180161,
 0.0011414371023414915,0.6673959069220058,0.05133812761867995,
 0.13996018836893953,0.3603641555428311,0.3868259603468439,0.4007270615069506,
 0.7185401022271062,0.939787116318813,0.9252731341763534,0.5782554030985719,
 9.718241930771754e-4,0.83475441276055,0.642985253602993,0.9781376713794318,
 0.9234724122942182,0.16194938668826886,0.9594399010715612,0.6771069882883936,
 0.5410504652227573,0.3860171697397192,0.911136349999094,0.175391037616639,
 0.9480316579654949,0.333105224551292,0.7301735676869595,0.7889763987735706,
 0.15502798493162995,0.9400258869714578,0.8046964352323185,0.5092827610976833,
 0.865168356568321,0.10340942812251175,0.822335805407626,0.9170247767310932,
 0.04986047467374888,0.8422451082544904,0.49817930871208527,
 0.011073673727956534,0.5479839164797556,0.14519071022915986,
 0.03857611843268682,0.6144732354418427,0.04685375400179104,
 0.7126944206109112,0.5912501734620991,0.5726409191339809,0.4214227285421853,
 0.33193727066247714,0.6060124182940496,0.6783524416577857,
 0.41533107614078646,0.1019522308715729,0.5985450152898141,0.9419884238169653,
 0.00931272264632177,0.9968559351153916,0.26291049264542554,
 0.4654774645883588,0.3837548484018002,0.31094374486802745,0.9480208999008078,
 0.501377001312306,8.142052491080243e-4,0.892497007629059,0.08106856306625532,
 0.9942368744482087,0.3133767341737651,0.08562113286891782,
 0.14958720645509815,0.017068060058075485,0.1501326339124993,
 0.5465406641024279,0.06595967046483997,0.9930435945319298,0.7014209995611596,
 0.6313958852221061,0.9114163131540872,0.8111580322080805,0.4217283264987357,
 0.8823666707506808,0.3159294618810832,0.0804343762805455,0.18822710499211304,
 0.6530396250636599,0.9951431583643962,0.18003650317427788,
 0.35652041486850317,0.771160277109808,0.2836041334352588,0.5316709315042094,
 0.07600291055916131,0.5196158695443023,0.14883938351276238,
 0.22752798811763508,0.7118424293320269,0.13646311618533735,
 0.5963296428069665,0.9731882290088706,0.03362266400935532,0.9597185873553321,
 0.5619331310991496,0.44610300853754425,0.06785370641700084,
 0.39736657768272476,0.5201230658921786,0.5870988771107165,
 0.44221897058637294,0.0030085376517496876,0.5360926152724442,
 0.4501611756851286,0.24136133607995458,0.104887919852243,0.6972613354475858,
 0.6972981690357233,0.4097269906518828,0.2252411552975474,0.8986397624709912,
 0.8553560823073239,0.7048132405015654,0.20018639422400308,0.1425798784408,
 0.17653927848000883,0.9118605381857661,0.6015211593904017,0.6207048162537894,
 0.0989793827437464,0.3203362193689695,0.9319330642248089,0.8099017434329395,
 0.05313677997156563,0.49482492957915114,0.8163499577297844,
 0.0609936395800883,0.08232898974065139,0.45941718909907847,
 0.0503161016596001,0.10786389790778483,0.5793938803817966,
 0.08439488458149269,0.7709774919266574,0.3741040689137747,
 0.10924297599572053,0.7731313115990086,0.740373651313735,0.7157042883762846,
 0.40319653553413926,0.33642548567564456,0.2772340210534745,
 0.9091491302878378,0.05296972805802924,0.6904492496507924,0.5817492615097949,
 0.6434480518360067,0.3397338069931257,0.23820069049469228,
 0.28134344040559944,0.375181068010908,0.4946163192819558,0.7334966330935789,
 0.4157768922302564,0.2901380566095956,0.042128935046144234,
 0.5823192960541015,0.2838447994309752,0.6144027480628554,0.9539819002207007,
 0.6031173407289411,0.1897836412909284,0.9420973058569533,0.08646496009115323,
 0.497740968026698,0.9214890859591728,0.08290529890128995,0.9634902672758752,
 0.9475613640732511,0.48051091017621483,0.18321376646978482,
 0.4455276889000058,0.9087586539627767,0.027810106114165345,
 0.48151542196267516,0.8755167645614127,0.2673605033816142,0.4935311934743978,
 0.23998403450235284,0.09523887032582523,0.7862950159386788,
 0.43454612757979993,5.768439812451343e-4,0.32116856776071967,
 0.9878671861783938,0.08036755330696299,0.8263982425887243,0.6521000412617556,
 0.4881840403362345,0.3216488477224876,0.2479868796632264,0.8753057687853211,
 0.5307076322721821,0.7928705624138306,0.259440919761073,0.10080562476474697,
 0.5182891412614707,0.7779912706213163,0.2853799891260447,0.3280257433080547,
 0.17722593146742305,0.4861823592378012,0.7603615567210418,0.7935143915640892,
 0.6174137584291117,0.7635187271291863,0.8501428850606276,0.3382757633215614,
 0.06127740511053548,0.9760149086032869,0.9174031190489969,0.6391916828293681,
 0.7717136145782095,0.36000483483935863,0.1515318241705591,0.5610970958901914,
 0.04426903405746574,0.10482613939890983,0.6128396343236923,
 0.2804819722867018,0.13579889956978775,0.15599251633757505,
 0.3306825518663282,0.7975102803886563,0.9829201631609757,0.8584556217410166,
 0.5617008868099196,0.16705776473761436,0.024561916892466673,
 0.6023002395942944,0.8056706892444996,0.5528567285112058,0.7174916689223475,
 0.22572244433048527,0.5708091213457437,0.8296328344912776,0.929053031784921,
 0.3392019933944018,0.682626917852888,0.6783263608008615,0.5189112358637639,
 0.6058941204654579,0.86506644386374,0.59191187837174,0.759995487517085,
 0.986661620195946,0.6658307723821832,0.4907688298765087,0.5552938833702342,
 0.32486994562235716,0.6777987967377835,0.4563109073686168,0.9871327220728336,
 0.331775230551016,0.22537315220033205,0.8187501765710048,0.836259445648855,
 0.6992097638535215,0.7900217887840091,0.09318856877079518,0.4036966449443943,
 0.30490072249329336,0.5900560589688287,0.1729775459898112,0.5177625557218971,
 0.5354756829153337,0.026374833185385582,0.6473308023920523,
 0.3287525665454193,0.04339221290256268,0.4762568732971001,
 0.010149439613810651,0.26805401381628324,0.03115745903646161,
 0.29277031008164256,0.7609661754291125,0.7111418643240927,0.8864766834337005,
 0.12357999096102379,0.2369578933270946,0.018875243474769654,
 0.25916268292254885,0.312777597926022,0.4483537026675889,0.6191138380492475,
 0.939979473237586,0.18316974563889798,0.39570604820400546,0.9400378591324594,
 0.7533184612222817,0.7327357277093688,0.2767382752773674,0.16983262282454614,
 0.627485527102099,0.18255663230926,0.2574466179231457,0.09588543295535135,
 0.2375114596234319,0.47619248471709974,0.6688445856366317,0.6219148656607725,
 0.3049072894516536,0.18460260589988386,0.8722717629631632,
 0.035457580106718734,0.6936634005204173,0.6091701089283346,
 0.6393640363103592,0.35417890398110796,0.3335077064777392,
 9.969824080813261e-5,0.47004349960745684,0.5691867303542376,
 0.8646450545878026,0.8677763743226946,0.7542026276291816,0.702873919245768,
 0.10028573284886144,0.41278213897138305,0.9562195111609589,
 0.4454746416342116,0.8476375064752011,0.34839399148793215,0.7114208780959798,
 0.8315011203615751,0.1693315861901229,0.017309242898088217,
 0.7301650926538442,0.16467079329738388,0.7634414733415305,0.5809325042665288,
 0.1455285298144302,0.5141461094020047,0.5007350109731656,0.9528579441401996,
 0.31093147296987556,0.16742079187306746,0.13674074165357442,
 0.8062768619845493,0.19329672434911604,0.09258713039174304,
 0.7560815927376753,0.053848716666012764,0.9356168671322986,
 0.3723671908597096,0.5809000120574174,0.3690439901928051,0.40342338545646195,
 0.5028685724905114,0.3792545673132792,0.9647616330039652,0.9321671272218014,
 0.03047889263867698,0.32343321538799774,0.4724834590687753,0.526890477054207,
 0.8037454463548153,0.03475717919119081,0.26052868769012616,
 0.9805787766349856,0.6181756084062686,0.3673759053144676,0.28483256814860747,
 0.2315032067135916,0.5389940225921521,0.1219468790200533,
 0.0043684858107962965,0.4974354776603849,0.5396594229775263,
 0.6287699777179449,0.8417644274087893,0.71826368952431,0.053314233676547196,
 0.021943795919547093,0.15432591278754992,0.5540941129089524,
 0.25405166054039474,0.3837844063746704,0.4644355429056649,
 0.18530529049384592,0.022237152052009157,0.5853353391815582,
 0.1423374835519181,0.9329782084553443,0.6313931546987896,0.9109760934835299,
 0.8598424867628964,0.42607547914932653,0.8343100055350914,
 0.026724428695436786,0.8572007689005424,0.1041951121516348,
 0.5787494850180332,0.45611803429454867,0.8286432517844929,
 0.17566180143734333,0.5735601054829099,0.48097060367163635,
 0.6266566825323102,0.17108034592605403,0.019104201005881616,
 0.22649904140980215,0.45783687098269643,0.7527304446506176,
 0.7697896781431397,0.41484399677384076,0.9380696184954189,
 0.31881024291787263,0.6104428506020527,0.79879923057731,0.9671016117602431,
 0.08389207655113529,0.8055067881544733,0.8049298276223859,0.5453844057782717,
 0.536976328636422,0.5446778391838253,0.7108694847811908,0.2696283177166615,
 0.6102588721609344,0.8688552808248653,0.3742420980433253,0.1197771062172529,
 0.17490827390681463,0.007600892104217327,0.7030200635206924,
 0.041623580616184874,0.8262937518579324,0.7620990606497617,
 0.30320780672354486,0.5545762935127532,0.678424828380035,0.6606213033085278,
 0.6348793904383495,0.4904601571756786,0.935385803938313,0.9313473419940743,
 0.9125603155884605,0.48992494374450657,0.09416089290415497,
 0.5832837452552193,0.3108029354644234,0.8631366131188817,0.132200390893899,
 0.7255702524255031,0.5130347591085961,0.5927917264842766,0.0457784403634347,
 0.14427367583577655,0.6610284390178374,0.17694796761952236,
 0.5910912758464001,0.4805628045284074,0.31559988632119806,
 0.18865386458899203,0.3675213466997932,0.8928347672578535,0.9193090650465175,
 0.5769684347151323,0.7814878887516535,0.6717933052966842,0.2703936376431515,
 0.45607535462060367,0.36201127180161385,0.635860234214601,0.3615701958468307,
 0.20474641894935297,0.27115587060176727,0.5254075269300379,0.91257048167499,
 0.15603105645689164,0.6481171882254163,0.7985249107473493,0.2093182745106269,
 0.3993047358197015,0.7549365005571933,0.14921887602710893,0.9590693583284526,
 0.19024720836344322,0.39535522231108244,0.698024045750407,0.6178196155098357,
 0.5557010286362004,0.19649627489992516,0.31153113931157095,
 0.6706640416551252,0.9193186908705606,0.49530158457905604,0.9908518994421691,
 0.6697155780994869,0.15557894755654977,0.04084361386921487,
 0.048268234801388754,0.6623289015483521,0.833944533207295,0.995188208365681,
 0.9720008394032033,0.35500243311676405,0.45335274128436565,
 0.009307709347087023,0.8020983052477691,0.7801806852922073,
 0.7570978998958889,0.6497063747841583,0.9855981694354355,0.7802390041247843,
 0.2678525501615777,0.1286558549629364,0.1779604636675266,0.7626182621519899,
 0.8250159875920903,0.13983018559407878,0.43205396355899905,
 0.1649927020444849,0.3108149438804707,0.9910182436150575,0.4651687615952922,
 0.3372403577701345,0.28373887438593426,0.8185796261834906,0.9599082341588101,
 0.2604083123853176,0.8764128684150692,0.8614678128829785,0.05016613414691817,
 0.25089755683188397,0.6864911526454678,0.3126639282435585,
 0.04531941282299512,0.4172124008360618,0.42384054096705504,
 0.6297258762313767,0.48376235471966467,0.5336682383455424,0.5557197769517245,
 0.3982247091289971,0.4133041644109817,0.9632416106229912,0.8557554034603938,
 0.8625961536141674,0.4255829534558463,0.2525109590793936,0.5144346088372296,
 0.8519740722421951,0.9498184836952769,0.5283289563772492,
 0.037301331858076936,0.6376794170655261,0.8976472337700583,
 0.3411419722917802,0.24006258299450023,0.7868492485836689,0.5037732351042424,
 0.19625393619618026,0.7139255227290486,0.9151968389540078,0.4604321297244346,
 0.06782922169799177,0.5217542672382546,0.6765301424630428,
 0.17130362613895134,0.3231079328190464,0.5817736511589776,
 0.35329733576115885,0.7023242860754872,0.032283382949253814,
 0.47397090531475816,0.716961147738591,0.8150395676423199,0.8174446629672275,
 0.07270928630498186,0.2878873062719921,0.34008813773896107,0.855162554133565,
 0.948680464445516,0.022402032653984216,0.10928474537575028,
 0.5097418392155776,0.07109962551380422,0.21127498827338964,
 0.5977777414928439,0.635596568561668,0.032792195591243134,0.6086345271280632,
 0.08467289970250014,0.09144764481807566,0.3260952334257189,
 0.5252186075996386,0.14941150953985483,0.786450742359106,0.7064328729430382,
 0.28656358714075325,0.38801035144044693,0.06665067879916986,
 0.8915045701466218,0.128686184867281,0.09687495947931679,0.2238083068308312,
 0.24842405885273733,0.5860971147637134,0.5828480557450451,0.653525409472834,
 0.2540072612643536,0.5050544634975616,0.7260108189409817,0.7859690900490124,
 0.7989496942092479,0.3245805764555121,0.07371347787136995,
 0.16865670656420517,0.9459280556882201,0.6176088128294485,0.384370575530228,
 0.19817848962642293,0.6829451575996442,0.38276987484937286,
 0.9848773343165762,0.3886582347438712,0.2760245922926948,0.31663495225167615,
 0.8540265943162533,0.038484493262321506,0.5367789102078304,0.739573116307973,
 0.7178758943572077,0.540042736915022,0.7633579754429638,0.11738187007276024,
 0.1694163111883168,0.8272088095202571,0.16824685921515992,0.5925014028447878,
 0.7965689503659094,0.4769924109167685,0.5644803578833681,0.5212614480465668,
 0.5033884165336209,0.8657522276388073,0.796524000807337,0.8216719304889464,
 0.4895268104065438,0.9100900448482021,0.7077561664253548,0.3234083699781597,
 0.4496271449454894,0.8781624838726058,0.6206297720639209,0.6718210419678298,
 0.22529650573130333,0.3304631048238542,0.035723292556368524,
 0.3398724084868956,0.36248347119044644,0.2161425668401823,
 0.27628815035818777,0.10066231963852335,0.21262284368007367,
 0.9421188115530206,0.5518081496225161,0.9654920304196761,0.06902326859956631,
 0.9614632373382883,0.8543182246919367,0.7795399675784587,0.06006525733899459,
 0.4079798053086,0.5412515442893524,0.9005366494275405,0.3277213878970874,
 0.8725195338191062,0.8468104826697599,0.3630939866495494,0.0880727412627158,
 0.6169765702098047,0.6932719215294053,0.6896089795302102,0.9842420685201425,
 0.37354570580166757,0.9896205251849222,0.6850928312136235,0.8020670530259538,
 0.9651222697984527,0.25240816190657633,0.7728707961209262,
 0.37972378700399556,0.18955977723109185,0.674278167649976,0.532289277116428,
 0.5069444533167884,0.11701424555750639,0.5221840018959862,0.8662109725033363,
 0.16374752646461443,0.2027216826078979,0.6130505923805347,0.4000231602570019,
 0.11963536322501012,0.4783606787978716,0.8716365017752897,
 0.41958808277482207,0.39213783403911073,0.5234193265027288,
 0.3642290519151954,0.6621305457173581,0.7818606881163828,0.03428815870291391,
 0.7222319700481161,0.8737160205277396,0.28055281317475567,
 0.23127068382025984,0.5791365991953241,0.7962371300035885,
 0.18975190091248162,0.3825667218965513,0.4238283961030769,0.9713991328922494,
 0.11205490310929822,0.10664588402063158,0.4352777228479707,
 0.27384566297141477,0.7224750005192576,0.5243708948211171,
 0.03717883479595241,0.9186224345680797,0.22618851240321802,
 0.8164567644310499,0.9106133879464184,0.6070943438072,0.7426561220215463,
 0.1898764349860812,0.8552882158134868,0.9338582628128089,0.6659114881257414,
 0.6535939381201971,0.6113827143675254,0.7695640572719092,0.1254242084165501,
 0.7886900371234257,0.16394323396617727,0.13999045096079765,
 0.28965366214305444,0.6757950903054009,0.01707183053116701,
 0.4757855401664167,0.6506842274562127,0.43682082795276833,0.4818934718726513,
 0.9614690595496855,0.8492743230004502,0.8844111307657542,0.5061862599194756,
 0.1746529765089776,0.42625963760425434,0.3268097007576761,0.8421529818433947,
 0.4424136705846573,0.6129898983808851,0.3395179183940764,0.1897907374294141,
 0.9288517094142685,0.9977495844741531,0.09145879006352675,0.5805456167835183,
 0.17689906902382968,0.300156558656553,0.43795590261569295,0.5011610051203567,
 0.37319316796844904,0.425921177965204,0.0919560693189585,0.36254513672764344,
 0.8053421195832505,0.04438949607126963,0.42635497644625797,
 0.8761656646277296,0.06704605607043235,0.33955843357313575,
 0.31821979702344716,0.31118377202264424,0.9183980365502502,
 0.33810804462074684,0.7143603866680569,0.2814329140486951,0.2026626171247996,
 0.49803288024370773,0.6255229526349128,0.6668522981920064,0.136275153177708,
 0.8584168731370228,0.9240422214250649,0.5485807598245818,0.4343412085987052,
 0.0418253911717158,0.08941687269683196,0.9612234191843771,
 0.19024678914595228,0.2717287422662754,0.6792529772795078,0.7234453523243063,
 0.6283491383051838,0.12872550254541126,0.7777346217592976,0.3226032413158906,
 0.9887706426969523,0.9152972503879729,0.56045917910032,0.681989612874637,
 0.3319335330154932,0.4808167152248837,0.08007939579260237,0.3766194675872456,
 0.4787532760316875,0.5076460394108548,0.36684498636214613,0.6292293972810711,
 0.9766303819479606,0.6507583875911392,0.7443817015442701,0.7253941457232684,
 0.5510321296059033,0.5098783210472695,0.8861785168661764,0.089973557233059,
 0.43663219826793265,0.4764126562621801,0.5976165062732529,0.4007131656429743,
 0.23502822217804242,0.8711913406303735,0.7644327549034383,0.5854696048407227,
 0.12104959600178211,0.24946879657213328,0.6745290628539118,0.505243011720463,
 0.5263161114872825,0.20001317323207246,0.4531802782351295,
 0.21509268612231136,0.741291841195897,0.39903237805766856,0.6216755569434813,
 0.6046200501326672,0.3137346230686321,0.565607139045924,0.7955564963612259,
 0.6570606495819385,0.21666992618152459,0.9980204418622093,0.8701168117877537,
 0.5568555524993953,0.6766762091929107,0.875085922898967,0.3118045520135784,
 0.48812672724451633,0.9995130449796541,0.9855732812573272,0.6303859323819955,
 0.7322646918742813,0.6787341316315129,0.8260623928071326,0.7215110333330258,
 0.895319602708331,0.19687604940541314,0.5721751181752667,0.9848613658402908,
 0.6044808502762911,0.6950914315111728,0.39839648717706344,
 0.09824592218887407,0.46732659491745643,0.6517819203571309,
 0.8840312223858316,0.8320375522201648,0.6897331642002418,0.3451211131629839,
 0.6301985880695147,0.4945344369643836,0.16587197803368425,
 0.23910635291975546,0.15678112380851927,0.9637442329217797,
 0.3571885879947446,0.8555050134093872,0.18616196408811103,0.7551513004539515,
 0.7157943946176609,0.9304476501034815,0.02489670932786947,0.9252935259591538,
 0.11880896281924413,0.8688472397892842,0.5940290862430828,0.3453783881975414,
 0.5124802913371168,0.7912221297665888,0.3163359194396991,0.4419173773710642,
 0.7110086081788177,0.6152179561685772,0.5204317020694302,0.6385616428961527,
 0.0811408545605804,0.6007582017135634,0.1261816562830217,0.2751923015727503,
 0.8844296194180065,0.29402005889339455,0.9340091902844978,0.4867891108776793,
 0.4781769646805949,0.9766879350719911,0.6749814202280399,0.9904428260990339,
 0.03901820971465231,0.840029242180989,0.29654606437175823,0.3283171795996016,
 0.1298490170009402,0.06093246572402489,0.09767256093889254,
 0.9743926360989956,0.22846795773661088,0.9492233309529317,
 0.21898862488473558,0.29535444720135395,0.6687726949774523,
 0.7519812244733857,0.5670246283256695,0.9950719876030464,0.8432948867374795,
 0.33899507995053313,0.9226753586672437,0.3914989322293221,0.4730642775690316,
 0.7066870134420383,0.5883538514901236,0.2864545575757108,0.1932160967888782,
 0.053246458568666166,0.13745949384185197,0.5514132314626953,
 0.21749292272944698,0.46140306793326924,0.971620125292047,0.446936148212159,
 0.6519133877070298,0.3391945788270041,0.7035098942077835,0.5100713830523107,
 0.42188501861489036,0.4760633013367961,0.48569769757342673,
 0.010767907722318726,0.36607483476888847,0.8526067696693848,
 0.8933608024421049,0.7136200288969828,0.7254074823585138,0.47338991256543106,
 0.7996625190362646,0.8004041184348275,0.048042175393257724,0.523257873495193,
 0.582835561377632,0.17671798573565212,0.9785671485398882,0.3995442345074859,
 0.13334022792599842,0.8328030657362071,0.2599656569136055,
 0.40355010358996546,0.04399563710512111,0.8129814891082083,
 0.28608502569766525,0.5216697099423658,0.11833435064906361,
 0.8882608556638055,0.8118196930653075,0.6938315795764249,0.9647951546429439,
 0.06987712090909415,0.3037395451936973,0.6108165515685031,
 0.21952110671443725,0.6079535154886395,0.04596077791910269,
 0.7615166142150318,0.3668238370185002,0.3927193549731355,0.3809615493024995,
 0.9507632618936632,0.8892327654204044,0.1621686141144285,0.6016109001380722,
 0.8932114801855147,0.06214636268063678,0.3207170774512095,
 0.03792278330394394,0.1552347618209844,0.5006739384677624,0.9866865041749359,
 0.304207177670209,0.3651108415603339,0.6565171563431572,0.017154703892876544,
 0.7268864119703772,0.04862738256703791,0.17856523920931378,
 0.9943157384505227,0.06372713430911359,0.8397352689797368,0.3767222238381257,
 0.02664549692713014,0.8961796373607488,0.5863815037802897,0.6826926534743146,
 0.6728506994719118,0.46382881921528796,0.8170994373021597,0.4251761490631637,
 0.27414209728437056,0.9540993049888814,0.8323284935330804,
 0.042923997818146375,0.20913971344782134,0.02182327220490654,
 0.6547933288636943,0.6868140749871874,0.06113542844895925,0.8898904551908929,
 0.7305545311191367,0.14683504816319237,0.4564418719134682,0.4116680273961143,
 0.7051746440334968,0.014290576038726965,0.4849162140737088,
 0.8746060548913748,0.8631549593408088,0.11982618946334078,0.1275725034352243,
 0.5620521580548464,0.2800226710778241,0.4856249320937124,0.4983231717961246,
 0.19254188339428402,0.617596074928363,0.31297958755098343,
 0.18156924346048675,0.06848797367369186,0.035282003559075625,
 0.13525379913912539,0.1721616701025035,0.10193461556689976,
 0.5008102033362034,0.5083315373101858,0.755602522219501,0.17266792149847843,
 0.8397926930494574,0.8716289249895112,0.6444611177105934,0.726546105974384,
 0.4997184315935994,0.9242335253945231,0.33642831483367464,
 0.18469984613944712,0.1864155498232758,0.3539918333262876,0.829865554806854,
 0.48469585751184663,0.9079931931082688,0.5648948327348148,0.3754667586409751,
 0.5186038275185602,0.3489640888594159,0.6188476478537338,0.9445144723306017,
 0.49292536171908363,0.4876444385607267,0.6431905202463941,
 0.25886652651610387,0.6474555773499061,0.4970560022194638,0.5062045148586797,
 0.4344126546117575,0.917295584072766,0.09329421181887476,0.3836098678747949,
 0.6549525100800107,0.7297840231085168,0.5203198519161384,0.4297462590247082,
 0.7316671124929962,0.837333509282914,0.6700873842100116,0.6644155506346623,
 0.14034131800768002,0.2830129620014553,0.4169252050175686,0.6175999800426164,
 0.0210415724199694,0.8511767281721327,0.9484549678142508,0.6268859155176659,
 0.6103144596788435,0.9157237429211162,0.4629615312026011,0.8282131440924355,
 0.8779686045047711,0.5827774072982146,0.5250472855993643,0.41489324868930555,
 0.7403592709723485,0.9608527434463187,0.1632224840419224,0.580353175883048,
 0.44380420691981504,0.9312289127683133,0.6336563483848965,
 0.15781278499615903,0.4756779350565643,0.7604747991836867,
 0.42260273100361345,0.9011631262940326,0.1508427579148246,0.4925657883242459,
 0.8048284082086538,0.4664353708652278,0.12633122133867514,0.5690715434068965,
 0.4711924128730903,0.8426594154463681,0.5479534344087977,0.5433344852863418,
 0.311083890124747,0.057297142643097154,0.7839278249323413,
 0.15654099272713973,0.5970540189273909,0.8789580033548992,0.7091314429811371,
 0.038202498583687294,0.22233936273653232,0.5553350928727011,
 0.358367114142059,0.3432295411277071,0.5032490470867796,0.5373418897605028,
 0.7315841370813289,0.8236275055187932,0.1887509897214339,0.06546775987058207,
 0.25317186940073755,0.010448146738507469,0.36145373780414836,
 0.6103320792747482,0.8822062327257295,0.9287059038715713,
 0.032198530445754825,0.7892604168815555,0.26357173520992605,
 0.9838008580960511,0.28736130862768927,0.4049486675871937,0.850741257532525,
 0.5031136855033945,0.9862401712520993,0.514312879056438,5.961004414574145e-4,
 0.020935798578286202,0.3805548119658928,0.22023880753335334,
 0.8081895554428175,0.022434516257246684,0.7021169070360141,
 0.25564992544228227,0.5535880198367747,0.44245659051955355,
 0.7772448909384797,0.8691763889930989,0.6771092720179595,0.6089980394167266,
 0.9749447733706044,0.709240404171995,0.5695656055360289,0.8396336474000582,
 0.6593073890823555,0.8001880725750217,0.9684530765080367,0.5045512661019533,
 0.16077613930097778,0.5131968589105645,0.2109918053066766,0.5837425023771956,
 0.31397758429438094,0.11475845633912463,0.8904890159393868,
 0.03478251951122058,0.7397711728222062,0.11082184376552728,
 0.3290570950184818,0.4460084179622149,0.41997904524620366,
 0.16526469075002026,0.8610957543258961,0.913675080594488,0.8882016444936764,
 0.6000678089386121,0.5130065606901499,0.06137733702039827,0.967684944549647,
 0.4216148158737498,0.20682536811806063,0.3192706664356729,
 0.16158855400378502,0.18492377569918872,0.7318044402291972,
 0.2496771420359708,0.4658419255160051,0.7922811765719361,0.9275028422712039,
 0.6298575552451136,0.19483225370038204,0.833333812120586,0.33151536517964675,
 0.9132014250347028,0.3535566847322631,0.553085884284727,0.7270535052693661,
 0.8365027893779433,0.265010375756222,0.1544042426688924,0.8548854637958764,
 0.7156619236829235,0.23989660293993775,0.577103290923819,0.18100465431790846,
 0.6589012695698495,0.6765858958501184,0.9892668157969571,0.8869843533854115,
 0.5060771165220618,0.23930826937018534,0.6426185136883686,0.5515855901325691,
 0.8385533283858486,0.721683809542591,0.5439541202829808,0.8114630535450708,
 0.051124419271182386,0.2527681596516189,0.6012742682313231,
 0.13446374398545036,0.5105248386177095,0.9768132918748003,0.2978612093100108,
 0.9614086867564504,0.6207207090182169,0.8002508463678989,0.8009577833794719,
 0.6728561263227082,0.6916826283739193,0.5796752781269794,0.40995805535701335,
 0.2582351376316251,0.9901726863086587,0.9817457729993768,0.539999434815295,
 0.8405293895761307,0.6997860220901432,0.6984998719934655,0.1056017864944665,
 0.2788273269531304,0.12341880143531259,0.013863174005889656,
 0.8728769035363885,0.9658166585166136,0.0918068305052886,0.7848504512706838,
 0.30943489652791745,0.355070265729709,0.9234500871981961,0.21317022292144272,
 0.8260463959150095,0.5325025994417181,0.24503593238744292,0.4097477162315075,
 0.26695800073182074,0.14038503753128662,0.7254219900167755,
 0.9659637698504254,0.6694765318690894,0.8106451682000527,0.49701148647045656,
 0.8298716146852252,0.4953639976962161,0.7627655542184486,0.5516946557048512,
 0.3728565473268919,0.8179405253857664,0.6823734956034494,0.4225131737615623,
 0.5922525606840655,0.15547738344915807,0.960216298749211,
 0.0041749195020013286,0.6986491590832112,0.48997649883763283,
 0.28015267020690493,0.5686663230874356,0.9985768496857996,
 0.07151633406997426,0.6264583459045174,0.9377985497506619,0.2722574399446307,
 0.20021787646599432,0.06900821071716756,0.011176004477671975,
 0.3583909646010658,0.7994237141431608,0.5488709776277135,0.6535039334764282,
 0.38739044216769536,0.7920134636033658,0.9255227149458817,0.7213201977574096,
 0.2729780398354731,0.1957283405780006,0.1660362696467801,0.04370795595689325,
 0.336462824946413,0.3529952455733414,0.8630314479161352,0.14203440159855474,
 0.8107881166369221,0.5714551265684986,0.9510123496508768,0.9761703882481347,
 0.3754562866701273,0.863876389144049,0.21548879239002328,0.26127117569064484,
 0.6545089554641971,0.9760763021321904,0.354641196682029,0.8189666716179438,
 0.7247161134624204,0.1577993055244843,0.2761496114037034,0.7493084295161423,
 0.3978995460282604,0.16765726348375087,0.414899456019465,0.9967922170949122,
 0.15631326611444807,0.655020208067161,0.48265257295524655,0.9311587964110724,
 0.7681959158639238,0.18249858280014342,0.4992503933373168,
 0.11121252560950268,0.7207681728951162,0.2025051945589159,0.5435896221039955,
 0.8099029787247167,0.5108225250738961,0.24175320645235598,0.5357399968803281,
 0.5022151547176226,0.009812916450163245,0.267272822495525,0.8462798906241003,
 0.620071376720926,0.3090519336563563,0.9923430080415566,0.18147689454372562,
 0.16129509952325993,0.6474054952967937,0.39067279345841743,
 0.13170578004470013,0.6161257743708145,0.571097186796025,0.8095516631542963,
 0.18695870310852059,0.8276060795995892,0.3083849540045954,0.6865762907901537,
 0.018917492781398093,0.7694655972255474,0.885639187814974,0.9365154079247756,
 0.11432274099650497,0.9928611098199164,0.680230044227899,0.3879707095122369,
 0.2050089527443839,0.8485652070108629,0.4660488559457068,0.26164702868370604,
 0.7514452007554842,0.8418899383103698,0.8325759002520244,0.9278716696129491,
 0.5574787803098381,0.14073820156953276,0.8937728535686397,0.6189506860940381,
 0.16907629837577098,0.5885880538534635,0.18630429142358074,
 0.1900480468011687,0.5800874816613785,0.21090931323285722,0.3854561802235894,
 0.7869416468145077,0.07439108787071524,0.7658960603637389,0.7530504327715889,
 0.3867640811314407,0.3833560392971509,0.9748685453098691,0.9724601588391877,
 0.24596606263453724,0.7529523146077244,0.5415596100350005,0.3459596380308697,
 0.339386635944144,0.1553375008336394,0.9313176754282358,0.02227614282926771,
 0.06489474768617565,0.47977428067383643,0.46432274925265293,
 0.06490753575860353,0.3229831340348738,0.5336156466162473,0.9370279276217303,
 0.7840711107100563,0.8676546899591702,0.7041742053143588,0.8669830304767596,
 0.8026127796841654,0.4074853602289974,0.9701454660362872,0.2321375729886601,
 0.04946183972805951,0.6719938508718973,0.20347817475509866,
 0.44316129454467923,0.22669966201778302,0.9565672023394113,
 0.9438211394120147,0.3803945221248328,0.2362891304331579,0.6251140026510655,
 0.5318860282241998,0.7213578833250027,0.5105919915310805,0.1870270043029385,
 0.49132631391334036,0.7651141610742082,0.8970525207932298,0.8575784060384914,
 0.8226645211005926,0.9336615260860595,0.9521706186870554,0.9426827182540558,
 0.3689746754172939,0.2544879254090652,0.8632376189245327,0.47548248885867517,
 0.34265785472325017,0.5523378509611332,0.885297307766932,0.5394576901309978,
 0.307565385144532,0.9340755718323874,0.14548816188928537,0.919558937304686,
 0.4857686163113867,0.26851114332380177,0.2266596070848591,
 0.41232020669942226,0.17750552216091597,0.6624576436253684,
 0.24827169929236304,0.18365339658843427,0.3872935624782492,
 0.8667300001731368,0.6028842374512121,0.41988008079250894,0.6663697135357696,
 0.23939804281191868,0.29209468512459347,0.20202680047210253,
 0.8759697238220219,0.24224900147052264,0.9861672723428395,0.0841455082959468,
 0.2694453133437179,0.5367451263212937,0.051904708993403714,0.938936317441204,
 0.3149486175936935,0.4182788170585324,0.8954052734174102,0.5157898701565039,
 0.690181898262488,0.7421896930200441,0.7412212299750041,
 4.8604374703975495e-4,0.11302964991464659,0.48531194068154393,
 0.6935559802311428,0.882109732996998,0.9559971679371297,0.05031251541476878,
 0.9310421249149483,0.49082570209938736,0.7270176976486191,0.9461374540198458,
 0.4948755007079284,0.15430587925946448,0.06470374455563055,
 0.36233007955721375,0.47390121630872317,0.876364419974792,0.3157512792131171,
 0.2538185544976286,0.909784139914554,0.5467076295180632,0.4296773566906902,
 0.8144179428411675,0.7062377312171351,0.17489189165545882,
 0.21091016716065736,0.8494600505334744,0.03444701380140536,
 0.6281024077494978,0.6722330503624894,0.8165447666743784,0.4877431472378706,
 0.345984692619955,0.2240424533442702,0.5807960103808979,0.2749055138872294,
 0.0051126493932011075,0.3457420559168657,0.40539153963640273,
 0.7886374218304117,0.3013843681417183,0.3489463473743468,0.8597513676147266,
 0.002402176533174094,0.9263863830800402,0.2748014444870327,
 0.5767764103849231,0.08311689038077319,0.37968186440343776,
 0.3899010433968152,0.7157867785278899,0.4520734608659849,0.3370614082152653,
 0.5344870980525648,0.311912518736861,0.03533404658657924,0.1799360402905701,
 0.3383959593182033,0.9736283141414015,0.5171798337024572,0.15977966003957988,
 0.3864486967578833,0.5601367875513009,0.6151645090069924,0.8752334117169983,
 0.2524955640132134,0.440139362791489,0.721008885781846,0.05733734176963334,
 0.5838033913348144,0.6689798348603531,0.42023329199119575,0.2685483627158535,
 0.8998433674616712,0.4058085124079035,0.17829333158562966,0.6853572459536239,
 0.06388382145365568,0.9178930715417051,0.8781516748537747,
 0.19589953869540844,0.26432466224070983,0.5978808197049633,0.302535269638472,
 0.7980590718525189,0.4175179784592281,0.2789394183117234,0.3275629484387904,
 0.4766644434341799,0.161453289016514,0.7215450225865583,0.4909071657937306,
 0.7094757418846588,0.4488909419088706,0.5478044155383676,
 0.010573937738434847,0.45340323312940567,0.947456734556408,
 0.07865169339849987,0.006277920122503944,0.5897596985799725,
 0.246521206425785,0.21387074934372807,0.12604448876820973,0.9539805786425296,
 0.6718973386722619,0.551999355785566,0.17344028405177614,0.5647077618394685,
 0.14824999876648526,0.9775235130698746,0.8320139363915939,0.552000619260764,
 0.8501081090073002,0.0412814328513722,0.6303564004196544,0.466885672761844,
 0.9056582284842697,0.6755761816541114,0.6684037408542605,0.4630803519261949,
 0.42072794192766316,0.32313122365847224,0.7189575217276493,
 0.06095383967669443,0.3208267791037174,0.32594411299632264,
 0.34090317324044817,0.17085417446964624,0.07167946887621879,
 0.2730131717146118,0.3582613769306797,0.21788247663041005,
 0.041775218229559874,0.8183064021957135,0.7601083947506302,
 0.6919877039980914,0.10010509025504122,0.31951401066923424,
 0.40093087651719483,0.09442892211425424,0.9157715580788338,
 0.7696878533602206,0.07001021794276796,0.386580112220997,0.18163208032927414,
 0.7242360913978415,0.25086905740871,0.6464421075945104,0.3947150647774822,
 0.35760100096732106,0.7434870943437323,0.8843199318948896,
 0.12403871165213554,0.06225963149313862,0.9719963570569088,
 0.0720551703883533,0.6414840393923953,0.604712988497967,0.32482135299262604,
 0.8339414083648239,0.7785032226170003,0.8018189766634314,0.5046471604949975,
 0.7292520547002872,0.43829906884826575,0.19643779920400917,
 0.025714058327224287,0.7197929201069075,0.6325612544272949,
 0.8664315117780662,0.1256301554940309,0.07949799891800358,0.556456431473029,
 0.836423174789945,0.49458226605205546,0.8738348414202596,0.12485697683375374,
 0.7435160617852383,0.9511855235432969,0.8219461348348172,0.9293283813195786,
 0.7692871848670434,0.30806697578496456,0.24066355311125398,
 0.31393862901144454,0.8548707009943738,0.6738931932376216,0.8697237412831365,
 0.9193841117184439,0.10045302593862948,0.1882878618202628,
 0.27594561899470704,0.3070839938647235,0.6094049698716972,0.6253260063647809,
 0.4878834435047654,0.9969735667455464,0.6443420204278956,0.9815393789515401,
 0.8255766311540023,0.5349216131878524,0.07600758547674713,0.5276081459369253,
 0.12705959059642247,0.9994100665353784,0.16206044652565077,
 0.48911307101927215,0.44194262697780173,0.7275466733080891,
 0.3708679394485258,0.5868642958791485,0.9749915533566316,0.90605167032467,
 0.9885951073467326,0.8713701419264682,0.595138301738717,0.07242531514275985,
 0.6380547479918989,0.391106431676161,0.714531509576001,0.433953000796937,
 0.2560889037457812,0.9562908642968526,0.7408198812272192,0.5417041481878817,
 0.5597399712396516,0.5875189778777931,0.8007637792295128,0.9902866098825757,
 0.976765143398342,0.49198123702675933,0.3025347325012703,0.7379606184726482,
 0.9443287111298844,0.2995492889515654,0.10134674127349075,
 0.47164176813273617,0.03517129548758846,0.7436650764542366,
 0.35608983658825144,0.5949640652225611,0.8783858633875535,
 0.31785466526720363,0.9800272760119743,0.7189631829468713,0.9968375187576024,
 0.03164725281552783,0.2867248780561855,0.15383088982258397,
 0.8519250021122263,0.9669078796309478,0.045261268541106014,
 0.46910591892268183,0.19410498958234346,0.6123329598002295,0.820595761743995,
 0.4138043014432694,0.01650606623410633,0.5131204362501537,0.9819849225298444,
 0.04105017996118021,0.22955159702273265,0.999530790333727,0.2087363251678619,
 0.8735829116005521,0.7482145083331988,0.3811412042776847,
 0.006252634267697732,0.955274356641168,0.5346621310573408,
 0.039119441852551895,0.27587432854748783,0.2151044983475483,
 0.9978045567274597,0.7976483862479902,0.9949868261446467,0.32635660744412753,
 0.17133435455693036,0.2736033082671039,0.1320274877120985,0.8432462917840039,
 0.14737604777459867,0.14881194210788207,0.5988707863138264,
 0.18438685892681927,0.48871650785601284,0.49576835351833093,
 0.9474709883404571,0.4489939864020137,0.4356107431416605,0.5539335643744537,
 0.795822801021831,0.438233451384344,0.38542831218544726,0.26061701393493797,
 0.2139806458701221,0.375429032245026,0.697199993182044,0.39603604156816696,
 0.8343829020130147,0.17872142398226853,0.8241107021023051,0.5968535790902921,
 0.679521253602884,0.01654213184243203,0.7413092991969347,0.472506968979853,
 0.6980375516118862,0.057291633939495656,0.3920215399233069,
 0.6261510954731947,0.4876215816712568,0.015291335244521087,
 0.27269877365361905,0.996672007478323,0.7330410200849485,
 5.555374747379105e-4,0.9330143044319883,0.19738202799909998,
 0.7796317439473206,0.9278882655977623,0.4341423541606765,0.4056690930664023,
 0.5589172224354256,0.36061073364283613,0.8380729649023763,0.2213722382258947,
 0.6795822465000431,0.7088885334773162,0.3439059177203492,0.3486642055526443,
 0.6737948625162301,0.846016602751399,0.4988135219452724,0.6119230820917545,
 0.5318926249484079,0.04427153625574343,0.5525650000934136,0.1063428183215529,
 0.11765814229229776,0.017129532898565114,0.13030872072830013,
 0.26695592121241074,0.25483818644105605,0.5498575958302482,
 0.16738857243614913,0.742510031460713,0.735366178317245,0.7671840417382171,
 0.5672558144028532,0.5383133379385564,0.7742699686120479,0.17709643051367263,
 0.9569785309209419,0.8628793591797931,0.8589230320773108,0.7932543144772093,
 0.9018947904681126,0.8216685807983753,0.6728555903924311,0.7012635184446571,
 0.24965361717719592,0.013587172879392373,0.38799116701211056,
 0.11254129597493978,0.1858705955095421,0.9860482623570294,
 0.047177925557415734,0.38552047136565204,0.7533893584645224,
 0.7554666765496106,0.6445932403161425,0.6826501345364905,0.871046694883858,
 0.24074817735966492,0.5912885745574952,0.25695297810881323,
 0.4522831819815343,0.20941125372598446,0.17295113406324047,
 0.5548892501860634,0.005087878654311262,0.8635171484055563,
 0.31397886528038876,0.12358086234510135,0.5915756130477314,
 0.32806800482320053,0.11679305945470664,0.786661196906389,0.1787197097765192,
 0.8121987949530094,0.18003355097249318,0.9761775431655288,0.9609038727390712,
 0.5759828676572042,0.5326163988059717,0.44662585280399447,
 0.34525946199387947,0.2641658110301581,0.458102240260157,0.19346153469692962,
 0.5390560874817479,0.6026630437346384,0.8553330151248618,0.835022184405115,
 0.8106234427520731,0.7887790892924073,0.8381445636162275,0.5117476124418707,
 0.43808068339746686,0.5332489353851158,0.3526808047119857,0.9129392487747789,
 0.1244055871262505,0.9435368418982983,0.9522788425838663,0.03915363353252532,
 0.18128074355597135,0.7916188815850743,0.44462376723504926,
 0.6582700270920901,0.41956942439728406,0.4461881472525313,
 0.15909463547864588,0.6076405814468638,0.7083826801032183,0.5384727219606938,
 0.5238619702551064,0.2221734746578513,0.09674342096330746,0.7781790375001139,
 0.9537351112586154,0.08675114074774792,0.5209253298192955,0.3666372466038117,
 0.20824238049595656,0.6823775904392587,0.4313004505079876,0.628921882003762,
 0.22181017645407908,0.459006606678268,0.7469966778194573,0.7839105053903704,
 0.8180617116044363,0.9162497415958936,0.25657944744764116,0.7189680576373235,
 0.707765605135559,0.8906323646084091,0.7801099916657367,0.16623610211945516,
 0.7482140972627348,0.6185099134823422,0.1770510916660072,0.09500432554186378,
 0.19556189635854693,0.7528770229405958,0.5465779921944605,0.6814466347846219,
 0.3926520689704748,0.9773803758453948,0.8883940119833849,0.840905318395478,
 0.14265295409092593,0.22460577469600618,0.9847370026280037,
 0.3169918479563464,0.6344552706472378,0.7286807361311682,0.6253621123170481,
 0.42795047908894435,0.4512399779509415,0.9291134313609412,0.6469750739959527,
 0.8742598694279797,0.4894845201309759,0.6117020712701245,0.6565488608478661,
 0.6886783008943267,0.7777857984316516,0.4195014624056148,0.5725272635355648,
 0.5996345859924086,0.49795370972147535,0.979676526013874,0.8180676610689288,
 0.6429115600339875,0.1532741995286051,0.374446068763744,0.2850312145717506,
 0.2275096361751181,0.36964767108061936,0.897509034064461,0.7959760575086626,
 0.8165328900931649,0.7701651650931232,0.017526328640511735,
 0.22269227157446536,0.335186402910435,0.20489910885300133,
 0.25301126220924997,0.4883820580927323,0.7063756155169341,
 0.02279741211665376,0.10532553522138499,0.990611740827331,0.7203450744521699,
 3.618429006907453e-4,0.915497850363945,0.16963785936675646,
 0.7729600491578972,0.4624635603106607,0.7660015827791352,
 0.009718024033139927,0.9039145717224656,0.6414139765931433,0.72576966907161,
 0.44927856259678633,0.08422617042091884,0.23647519202587364,
 0.021307037461816725,0.6820553827210899,0.10932069476960493,
 0.3222612065407152,0.6278512881283853,0.9389348520740971,0.27519840053547595,
 0.11523946891186909,0.33761318837949517,0.34270222356951796,
 0.8861780351730012,0.2103827040462688,0.9584088597370771,0.10893155672270427,
 0.1782471144762685,0.8511259383608079,0.3427434499922124,0.8886892366649771,
 0.9214027600517918,0.7548272987418454,0.32447948022334405,0.9062913212322152,
 0.5856408289368584,0.06153821019758532,0.7352752615920485,0.2897863906639375,
 0.6729914473211129,0.8854819686497835,0.015902776247275208,
 0.6482898766058091,0.6531812430267406,0.06070261716288872,0.85448814416905,
 0.5652415313716078,0.925013770410467,0.9884258986378899,0.21447136618557117,
 0.012858255482806946,0.5163927323957402,0.6131105985277924,
 0.9530145853773437,0.8099377125247137,0.8512896274456082,0.6720226220429806,
 0.7956974181085458,0.34222952946586727,0.3029345846205216,
 0.43087261157457457,0.9828679564782532,0.21576592627223035,
 0.06416476958364226,0.7377563611800843,0.6821098144689388,0.6008264563479787,
 0.656553679914716,0.3192820339550676,0.8812292290734703,0.668909194467822,
 0.14347077569355293,0.7800079266184029,0.8282464051458485,0.2907666145824219,
 0.4348650182808852,0.00227899506407514,0.030644295858395343,
 0.04192858659752852,0.0054710911940135665,0.23826767712592145,
 0.4819301360929218,0.033746098186844486,0.7089533993723547,
 0.45565778563133885,0.9495874542064513,0.7009257967404581,0.2015036776533684,
 0.47190288687132587,0.6462953737033906,0.5633467597686463,0.2850219222245012,
 0.3450145442646181,0.685519434810482,0.2963591433390683,0.06852622831078681,
 0.13825255314447027,0.2057299725131564,0.17933160007450932,
 0.4452857643203991,0.6761493784723412,0.5356004449861653,0.4994608091874345,
 0.27441044157967287,0.7821007579322381,0.22785397019645492,
 0.3327855096223564,0.6440670360635388,0.3089538640260092,0.276848600476996,
 0.2568912355227463,0.5698479367854232,0.4987109626698667,0.4958793042092361,
 0.311468213787471,0.28869719138891436,0.5731212911776941,0.356483587447609,
 0.7715902277231926,0.686182600857595,0.49443000851207397,0.3962165422451266,
 0.6466354395648544,0.2638337014672981,0.587284629933626,0.6138805388113222,
 0.9259334883027923,0.22606727297454343,0.1768286155105825,
 0.04893970703380157,0.6295884534142429,0.3489893789295334,0.713315741492738,
 0.7865875682380815,0.35244417968940756,0.7053942127996681,
 0.47479161595286545,0.5416120331683093,0.5281329423982242,0.6233333352300764,
 0.48073778834646397,0.6918361441382653,0.007847373732678964,
 0.1499661314799814,0.5126609863545856,0.08033417560026601,0.6423700480666711,
 0.9954352362187753,0.20397050294042307,0.871509307751709,0.4132712592615999,
 0.6307583394319402,0.3939666815961449,0.8936330668605131,0.7136557828063488,
 0.46311544394072046,0.10618623896847956,0.20825616693427274,
 0.4031852272884402,0.950448526477973,0.2719780146816251,0.4279982007541686,
 0.4189045234146672,0.8966228366458092,0.15701254960845334,0.8509691878132364,
 0.164589061658162,0.7497159825641433,0.6251397365080839,0.8747311556917072,
 0.627973052378129,0.6921996365271733,0.24481607109887338,0.4893089031465918,
 0.9039658511458418,0.17926104477570226,0.22851048121336626,
 0.6177073443416288,0.3637948388748822,0.5333866953371139,0.11455132461328521,
 0.5626966696782721,0.3198737193568141,0.3994207225877332,0.5151454706806923,
 0.1120548352552051,0.05289490944858177,0.14963540199641145,
 0.5955630866497611,0.980368418672978,0.17894889526715008,0.4911430218594611,
 0.8805046805314685,0.21806512265350664,0.30141944071333404,
 0.5452769558971096,0.8029941988000681,0.36100954446644107,0.8408190592165768,
 0.15218448203217128,0.8220242533805671,0.9101487435750468,
 0.04123511286292825,0.9949963179860661,0.41070999479343895,
 0.4370371016479827,0.07554578849653992,0.5095079303366302,0.5992916664965394,
 0.5021579541135606,0.8761379212847835,0.7377124564726946,0.4135293868768114,
 0.9779316612014741,0.3705596643983077,0.8605246043602286,0.4715173086197407,
 0.408351390558102,0.36694430431271186,0.7257756891629071,0.6472639057724918,
 0.9592923245233375,0.38836503349741014,0.9708840381748568,0.6672228953334016,
 0.6112498960584033,0.7276677409906653,0.4315409932597788,0.8151332942456275,
 0.2619343351052561,0.8232458358812641,0.034087156078471015,
 0.32069985296826997,0.6718286357869221,0.06001898573287967,
 0.04235531197012754,0.150940982042305,0.3103255437109127,0.4206279215604958,
 0.8368624818529881,0.4098574052948637,0.1052350400266644,0.6643779474148628,
 0.307508813910198,0.4862655638947939,0.0136282750953155,0.2788173151874871,
 0.3185453799802318,0.45904858036093843,0.273717940201856,0.5417203946416241,
 0.6427572759497444,0.36642245432905074,0.912660584843482,0.13053462910806857,
 0.01545962291690628,0.11094015521838851,0.9146380596081066,
 0.8582304471259322,0.49064136971473815,0.7125469196855847,0.9993551088842779,
 0.4611087478390894,0.29078672589090093,0.9983237097854754,
 0.06395712382487884,0.6037579773833729,0.4999905374056166,0.7181263281327662,
 0.10180662056147505,0.39998290682737236,0.41353387546217,0.7646105826403681,
 0.5990193405306182,0.4388142904730411,0.643505590915773,0.5810507427105958,
 0.4939575286538871,0.21618475004140636,0.05738159827239997,
 0.040963241938444916,0.7160780787782566,0.4293398720341268,
 0.32134601191985546,0.49093679423274994,0.6637945746650461,
 0.35615306395173385,0.011212354043829365,0.5748849797115936,
 0.5593369293267805,0.75664335339028,0.12918101788394554,0.3475677230193379,
 0.6861174014982436,0.905694735101461,0.06702396516219322,0.8580529464773176,
 0.33790178540810234,0.4067463739935082,0.24394530082479227,
 0.6132593068599425,0.43801682694507127,0.4163661008870446,0.6008336959399929,
 0.09454562925180965,0.7764873812207238,0.8981852126275068,0.9171861216897671,
 0.4134245426261328,0.8483480803201124,0.7472085520192469,0.4230264713532301,
 0.19864554853965755,0.4620220012912831,0.5843194124335461,0.804265065772937,
 0.6368519605264819,0.23672855856758035,0.34710891265098465,
 0.6328794508808466,0.6593565083973113,0.2651024080113753,0.11929430274224018,
 0.004841604934482602,0.22239733779488002,0.17082878720503958,
 0.34894028890600437,0.40198428047166856,0.4680860481137088,
 0.4774479872581383,0.8853634543758726,0.2563230200192288,0.3981981922767711,
 0.9584198095013163,0.42448401943700786,0.9797019367087186,0.8435580915315786,
 0.2653952132337062,0.5227555687428358,0.00573218620740612,0.9315851215146851,
 0.07497191537436565,0.09577700226653052,0.33467392944706154,
 0.06671017543751079,0.3607104437721492,0.8677701173756402,0.6923113163956339,
 0.17447652789583212,0.9222805061047825,0.8826680997352286,0.8112088695296793,
 0.9710255443720617,0.8655003526426758,0.6453831972518861,0.4052060852349705,
 0.9414364354673457,0.2613862060631471,0.3552109630444039,0.07734457931043859,
 0.9091976340835427,0.3147366910374718,0.09421455978552695,0.8400113841682749,
 0.3313514629262403,0.14110156149434805,0.7415320609967796,0.3439366414877889,
 0.062230165413575955,0.30202896600780815,0.5803217062898512,
 0.7524325567334166,0.833398422601033,0.23066392388831247,0.3127116388739676,
 0.6529871960064642,0.7987938886267677,0.9378965456290235,0.36529439331007707,
 0.6856200076550111,0.7453943666058486,0.6800567163001471,0.6656591785386539,
 0.8451915788703745,0.08361408202498177,0.8567122351068427,
 0.49147762729012956,0.6320801646523667,0.7519310579577402,
 0.23109177525819347,0.37152702700646734,0.34229963046383305,
 0.9092623171343848,0.8458100246222506,0.5049846301483767,0.7158695612603909,
 0.5198182360151845,0.46566222610286245,0.9023574694428864,
 0.30898127982738843,0.26195365418839567,0.24022364789093875,
 0.45825610996459465,0.9562935479483157,0.1907497683021382,0.8906327723721944,
 0.7780048395647432,0.3592290393479278,0.28450721047987115,0.1176735785486398,
 0.6887810458168004,0.535493186259884,0.31085671757067557,0.6330280983253085,
 0.9393136528460261,0.9636473481790693,0.5496023036455348,0.33016745313898255,
 0.04111037170709719,0.7572926683227931,0.9955694393114307,0.8586466726414793,
 0.7023126483641349,0.5438801750988415,0.922625441771331,0.13109802814427562,
 0.24971414318439022,0.30312985279863913,0.27761168379860557,
 0.5406893326886983,0.4287033900277797,0.509311070436301,0.22702541320992264,
 0.7075442289823578,0.9878713323944369,0.32802376533637756,0.7403324078366414,
 0.4674692227606452,0.10921266683307751,0.7292219120798353,0.8963446734850505,
 0.455544338123301,0.8338545486240739,0.4574823944320292,0.07054889837876632,
 0.454696952106999,0.5190783003235585,0.29687322708560027,0.21560394918968473,
 0.45381712289566645,0.14699333628485678,0.637528752108404,
 0.46922246193334005,0.8608073961420191,0.9821236361636942,0.8245543465073206,
 0.48642750257217104,0.7047513862644929,0.6376563862808841,0.5383354998842629,
 0.48891309331480914,0.14649544092929245,0.08044201599915435,
 0.792986939453749,0.35862017957814085,0.5835555088748279,0.9504603123098961,
 0.9046134988008432,0.3313141511846972,0.8578505564023372,0.529500861977322,
 0.39860127317125205,0.6166315497727346,0.21937962895083474,
 0.42974105397575557,0.5832423573504741,0.0548791771346161,0.7023318980358908,
 0.47476099212012235,0.6818070925806845,0.20635930806050007,
 0.0025295154520956764,0.9734445371720495,0.6304825359073547,
 0.37630341117224475,0.3257333423627419,0.3360812901588849,0.3500930351390106,
 0.8949289102832224,0.7323422287605896,0.14178111985823127,0.6410021995804916,
 0.2156815047303725,0.3386494630422249,0.38780173913570337,0.5873294568973356,
 0.254348154391013,0.835564113443671,0.3228073277744552,0.3314623069655065,
 0.863455744004017,0.6274695626422989,0.6897336738705496,0.8227145848121593,
 0.12187831153424822,0.7247991475101022,0.009763677794795367,
 0.9170191634671238,0.46256715229304746,0.29533222700059925,
 0.9556495228772117,0.9407918901371165,0.13148056924726004,
 0.36090375355933646,0.10581190124250928,0.3448537244719473,
 0.3884808484985005,0.6255362298598529,0.43314109867477846,
 0.33709085602160827,0.5423171363940833,0.7084257651679774,0.420282398202555,
 0.19937369736314658,0.1538191746154457,0.11220315126576286,
 0.5219386419232745,0.4607624126431008,0.021197518351713796,
 0.6866755121460119,0.42680221319292744,0.919078328155745,0.9435972729872181,
 0.7986824251354885,0.4111121109128354,0.2350820663855614,0.730319847533157,
 0.9180433681252974,0.604739803467972,0.41826078238930253,0.6970780123594145,
 0.9692146114663315,0.7584015871668668,0.926267600226094,0.20140593491249037,
 0.7344375317747671,0.9571717336476596,0.7426633291036862,
 0.010208823959849345,0.3538562283975947,0.06386393427040604,
 0.17511315911965186,0.5771975054238465,0.015378028155465961,
 0.42575831113866447,0.36095584957204674,0.14894304013749682,
 0.5776472768386822,0.9417050410074469,0.7471735280597658,0.9047350009578172,
 0.4361334461109577,0.6444580197584807,0.881393321682597,0.8897087728679636,
 0.2682155764212504,0.9885524457622802,0.714668753386142,0.2016718957218715,
 0.44280075504283556,0.014323417765815916,0.02927990244914158,
 0.25793176165810805,0.5700636064738884,0.8882416153333974,0.8511484236841469,
 0.09518125487528206,0.0282471824628574,0.6626018889604905,0.6730740133087856,
 0.4424482599135712,0.16900393803453828,0.2961387422500632,0.520481469537412,
 0.757299205228259,0.7950953515736503,0.5065448429128527,0.7596163689419418,
 0.6879178613528064,0.2325889287867441,0.7683374888359346,0.12212091328681618,
 0.4790137403481314,0.5297265603659391,4.2745729558979484e-5,
 0.14624966431973807,0.22295671857043253,0.20214928390323628,
 0.4932930594313909,0.07018285082061837,0.8137463603962828,0.2500695570799403,
 0.7766029495426707,0.3837637245413711,0.25890361259788497,0.6627942336179635,
 0.542683283844107,0.2393602636304727,0.9820130440113165,0.1339437056605275,
 0.5750263118948686,0.41212416002246743,0.8747258816684705,
 0.035502518531397076,0.2914863415297948,0.8517325848223358,
 0.7754500644489187,0.7091785113930619,0.481882927905685,0.8514689905958023,
 0.36098317531772217,0.7557198255250531,0.010537881581944464,
 0.19417785817995337,0.876944483758179,0.27227667598408967,0.4812716858536359,
 0.48873327586090065,0.17249636700543325,0.2619188514232924,
 0.14774398686575407,0.649235527931756,0.6475350243742458,0.5494853719600322,
 0.2290961262614467,0.7343058313458726,0.9877025613712213,0.07317720898123681,
 0.4127760570132041,0.4404360349056716,0.39352972158746535,
 0.17410381623066795,0.8830001830130277,0.20687247466118452,
 0.7493515309771667,0.42246260526136814,0.4752012674488102,
 0.47796959255923277,0.7172985759613102,0.1967595317702907,
 0.21466746206515497,0.08684696389089741,0.3060422070449653,
 0.33799586535664583,0.4233089056099254,0.5104762158337941,0.8353782226007039,
 0.45406471419675154,0.967692764869107,0.9002390312164783,0.3257624860683975,
 0.5758692619759644,0.3332392684457546,0.22744739137763081,0.5308495234217503,
 0.3135353190496104,0.9940325107215299,0.07251176566671347,0.7099432959085734,
 0.9309315644057156,0.481377582507845,0.962185003777572,0.06065312780153409,
 0.8854609718729548,0.6747842146177312,0.5959902903575071,0.06597170163123633,
 0.06615843043378511,0.9753468656596779,0.7354503611234792,0.988820879275877,
 0.9973733823627137,0.42716392426389893,0.4375593314627553,
 0.004363642474084495,0.9685158223424972,0.0016316104644934493,
 0.8588052199290408,0.6799136805298991,0.6325931624742862,0.19963657793949452,
 0.12728901426270933,0.2751150210413331,0.2407553083792895,0.9989268576073079,
 0.9491823106223816,0.7691036640186597,0.37268136990542666,
 0.28363684410869516,0.297412446910597,0.547451212280248,0.3237838380170539,
 0.2508892901689531,0.7675387099742212,0.2398933506671317,0.6306456975272935,
 0.8054866584858635,0.9992571769039982,0.5563048278986746,0.21388155837060663,
 0.05237747254200753,0.16496236445271695,0.14173526077802545,
 0.8250553133449844,0.5537547267761421,0.2778868104929666,0.6411970390158876,
 0.8004509195317466,0.9642159433594059,0.9143862805343873,0.1226244973980547,
 0.32557561522729894,0.22389450691380086,0.5475816551912229,
 0.6668495291741005,0.8304133538998384,0.1757707270426071,0.13079951343572827,
 0.06041434902997078,0.9348966770321205,0.43851386313615404,
 0.7930649435849209,0.29926450960734763,0.34924944797942825,
 0.21635206397583617,0.9052580036000999,0.760126310354185,0.7467433243403211,
 0.013002021797532293,0.5315275389136541,0.6147202961286824,
 0.1067242256065637,0.7978659441084118,0.028967008963330376,
 0.6266272308773074,0.884422102352619,0.6415336966305801,0.5544697655013997,
 0.38840839397578786,0.6317946487878815,0.5632740297623152,
 0.10913122672037401,0.32282137606580774,0.026177772130963617,
 0.5886861799958089,0.5053120639499449,0.9090944495844634,0.34528551953569997,
 0.8527923140876119,0.25410838810820524,0.3158573322927968,0.1788047248734651,
 0.7705195097358077,0.6129315379863639,0.044471807623531645,
 0.7005218623072477,0.10230107391428767,0.9176334534163026,0.8473475949280748,
 0.16331128059721023,0.7906736990138175,0.7290462203153594,0.9368010761328868,
 0.18459068644350152,0.2840058206503636,0.5731543322846784,
 0.18702774624050944,0.37658443128866603,0.8962076754750364,
 0.8836948525251848,0.47537885334054075,0.370354367898618,0.8805526638010437,
 0.4240377649659969,0.6664701206426509,0.2817696772255651,0.7089497720633727,
 0.9362839107154887,0.28103015818181465,0.21023769475805498,0.046579453865218,
 0.41151614105555734,0.2915015739706879,0.0917691389835602,0.715430360527062,
 0.8741964594713283,0.7347151452304009,0.6153408899497463,0.9855197760310486,
 0.2787558496077448,0.4484095896338228,0.5260002132164755,0.9298105870002734,
 0.5762048111580208,0.7560013281125135,0.7251779378628661,
 0.013969503427455177,0.2088894574534379,0.4076342839123137,
 0.3261674617004252,0.4706777395515269,0.8062611627474461,0.5432357133653041,
 0.7338620959679385,0.8146388278823253,0.785323030189667,0.5022271363247695,
 0.46414684084804103,0.7342661589140586,0.15564193872846843,
 0.5547147322000546,0.5140031451042483,0.5243681276901574,0.4245332651662165,
 0.43595493461848356,0.00602458992960897,0.38879520160911174,
 0.8118251333858038,0.13923458527954602,0.531572333673134,0.2948061920571168,
 0.3123711570202782,0.7224712357466739,0.8144645233417147,0.714532553136214,
 0.542540541126332,0.5909302397705299,0.7693117920207145,0.23295189761034896,
 0.4356535712480746,0.7969960924551959,0.8681452606671417,0.6376001522200981,
 0.45034547066388597,0.49582626486840775,0.9054898155909556,
 0.05621406253687855,0.9294163985487331,0.43583658873673103,0.765719949328087,
 0.6976571238704349,0.8887051037729397,0.08144541511826031,0.407837948222047,
 0.8417933876182577,0.8603934766957619,0.6731256057966494,0.1713939668736848,
 0.9198252798494172,0.3007062566839004,0.746980585674687,0.8798890354925981,
 0.7313153437587202,0.0659422714975757,0.4358030797092972,0.30432108658846024,
 0.08838771034157167,0.017167916462027533,0.007036321587532757,
 0.9242932012122798,0.6057343262491598,0.6934778255956721,0.3495182962436736,
 0.14330944015523606,0.8262472051210483,0.5394329595767522,
 0.23130325327226475,0.22380638542676645,0.32283304724377127,
 0.49217615175077833,0.630331930480276,0.8249025446817498,0.2500113433288994,
 0.8351136735956316,0.9599478903004729,0.4811860517797022,0.4643366427368617,
 0.8088786807992892,0.028367256864902934,0.11902676182849092,
 0.559264620526834,0.2704738974709552,0.9613081395913441,0.6040970095163165,
 0.9625098404982161,0.35090369900131946,0.15737807209053878,0.70804943407135,
 0.7259131282291507,0.026044931977927233,0.7011284066924408,
 0.9873290136979219,0.5375372369742404,0.29157194774494655,
 0.033315952902541524,0.1824344562334791,0.7384339268926131,
 0.8738673716369445,0.33126792789381054,0.9275745791619158,0.8515852199364908,
 0.03899067573348636,0.16944380241552537,0.33155129106876,0.9619766066225376,
 0.7236663589565622,0.23194356919406312,0.2867878064520216,
 0.24193934896042857,0.5787179151647327,0.04530018972760275,
 0.7048272462595587,0.9810187671292705,0.3816342039546208,0.8017761284635159,
 0.19864914906455422,0.567470200245741,0.010285218420797548,
 0.15589394858509065,0.4714114608682921,0.5250322045008385,0.8101978014327093,
 0.5790502939879972,0.7787630944383656,0.5945408961647497,0.1780043401432756,
 0.10069965557041427,0.75830413089307,0.34749713189706,0.601539120327638,
 0.36228192548316485,0.934314970530826,0.1396274927151957,0.11200842019794288,
 0.8313770480938579,0.3198528973851956,0.5187497623735852,0.9289281591925075,
 0.7828125878469863,0.906664278448801,0.2010600605487819,0.746647992718463,
 0.7629278802735332,0.7302279031506265,0.7610886239642075,0.9427001062785544,
 0.18747483786240915,0.15387975498173234,0.247528546493749,0.8800383732718571,
 0.23445296735439758,0.5825761471135855,0.1950190720910816,0.5603644584147461,
 0.7957368046449645,0.17241392886991225,0.33772105522754037,
 0.16382974298851138,0.9707275001230884,0.73302360900449,0.5772119621742797,
 0.31740725216976906,0.543887383812979,0.7414890124486724,0.5487021734719397,
 0.7129917563170401,0.752403962991284,0.15781552093634055,0.58593075367695,
 0.034966259571462044,0.8883898808543891,0.9571555721093794,
 0.6564438451751516,0.02547669130201069,0.4570774619680157,0.61430699220655,
 0.10293174041350373,0.49116694481306267,0.8870493801264286,
 0.10105121575646492,0.35068254528696496,0.8530039857304464,
 0.2954112766833946,0.025025664057832797,0.4500754211942841,
 0.3404188095002034,0.37905763601139386,0.30266810696416413,
 0.05756897214547807,0.6334867497727604,0.3585685905496234,
 0.46908440453174594,0.468166491216804,0.45936612611974337,
 0.28752149858795695,0.6138524962545074,0.8228277120958685,0.6281350904851216,
 0.013757609035896179,0.16259952663709243,0.12018028201355602,
 0.5082418636048358,0.48338148865603314,0.9598596062146065,
 0.06091857665653788,0.6326970496472089,0.7578971132110159,
 0.09140342027160242,0.9659392942814482,0.6218555181814384,
 0.11553641371028378,0.5720113294315282,0.06319337056252916,
 0.5663347143158588,0.4900022213143791,0.6686237626357597,0.47593401748313946,
 0.5081190449807931,0.7612703540352393,0.8344100230888651,0.9170026274589776,
 0.24155945986060412,0.8176021698502963,0.25706191917693,0.7741262910642939,
 0.3322186391610641,0.9641830871852796,0.49632044571935696,
 0.11807934548083576,0.5669801940857889,0.4040010009801879,0.5965432551251806,
 0.05777269586599576,0.8919692209654504,0.41912172040278206,
 0.3272164802123707,0.5377101542758603,0.12158747638206091,0.8958430733029065,
 0.6495801257311575,0.9748444176892387,0.23200523603425138,
 0.19059578197402216,0.16895014044513756,0.1109252282340174,
 0.2731990845240777,0.048310139426295695,0.9820714539776605,
 0.19321340821913957,0.9167474697782083,0.13485762033985682,
 0.9006574065825772,0.23739797238317517,0.02722569529913521,
 0.28671460289716066,0.4700022224682744,0.207416598780242,0.3263196178679538,
 0.9573814844598572,0.6098120298387972,0.10331301432837425,0.6568350068039602,
 0.9908003370487332,0.6554210231425887,0.2301207822953546,0.24126135579243757,
 0.3734325986880813,0.06745365542486259,0.04888909943152864,
 0.8123687536419941,0.9041536986549762,0.6131958005455044,0.5918849329787695,
 0.2869292267404684,0.048441291379410334,0.7094129493953315,
 0.036331211627529836,0.4019125179459714,0.30459826137683865,
 0.1457585222540536,0.2594235862582299,0.5100949595888096,0.21331341972510287,
 0.6776857036783711,0.6458306923288388,0.6635827610318268,0.4477314952871494,
 0.5588319924708048,0.901700439041624,0.3746652419229175,0.3662861310673545,
 0.12153201196170238,0.016396184309380835,0.7630863351161619,
 0.36342096548446046,0.2659059120005587,0.9847394941636657,0.3259371528385331,
 0.18504246433808225,0.13251998958650768,0.3725424383882875,
 0.7053333168647928,0.8879274259174423,0.287769850366175,0.7500282976983081,
 0.46014110346940196,0.9885288144931514,0.2587335985657777,0.7302082125425178,
 0.5633879198161815,0.9682872726861513,0.28389416663944445,0.4862792834662142,
 0.3780184818866287,0.5372429821064495,0.25686285054556046,0.838365077697613,
 0.5097952355791644,0.13189331703769758,0.014390661583466913,
 0.2351569088763893,0.28837416751722444,0.495496862341435,0.06320575208027757,
 0.09845409931655003,0.2720710137149772,0.7076285977249213,0.5065537748810343,
 0.7985861891450756,0.38591052315585817,0.3942906288536906,0.9574150770974851,
 0.913413412952548,0.7571925643972455,0.3717579532473765,0.8423002718973697,
 0.9303273892439647,0.11814233042042221,0.6036219132448403,0.8524304795194616,
 0.31626905821932105,0.9832732166966862,0.660653975794347,0.8827737794832599,
 0.7834271352452429,0.3370182395240858,0.8805553587085748,0.03991631235679738,
 0.8167119432772803,0.8103690563587221,0.46024554899882864,0.2551636505150988,
 0.008612962760962906,0.1530154052798821,0.24960991066674532,
 0.9214234843407101,0.9924256641788277,0.5373530146469216,0.9519723475179842,
 0.2349312699117474,0.25343143563706105,0.33550346919948937,
 0.9836373139875354,0.7139932318293374,0.4097374376423365,0.2548636388794977,
 0.261839183998399,0.9979961327486035,0.06813379969043998,0.10979118880271754,
 0.2515452834889197,0.32909213256826364,0.7364728412834921,0.9206922362917533,
 0.48343380237048605,0.11023977798844897,0.5734938366850846,
 0.41369829052022167,0.9151805363613449,0.20058767273526745,
 0.8767072649843831,0.046518787760914004,0.803688286192358,
 0.47461155613500317,0.5818534563264613,0.7736028819945993,
 0.34083173289765645,0.33815112018119253,0.5868781441904168,
 0.21983833841302125,0.928129865780994,0.048301909692322065,
 0.9590969770726256,0.44940074226083926,0.3875417170828208,0.7956540114316708,
 0.3297707861326865,0.9314025965439033,0.2927570417066565,0.183640554135598,
 0.7270503263646604,0.5610554738961216,0.18193494749899541,0.6550849022749512,
 0.6443230987522588,0.029819699910186404,0.8461677171927174,
 0.2126398294070313,0.30611765889780185,0.7293551008441614,0.9346553572208528,
 0.7613722373629779,0.6019541845796571,0.1206185764214176,0.13785831783024782,
 0.7192493768697076,0.4904564514685654,0.6379312370102985,0.2814881234194976,
 0.7727843168568925,0.9909564635703547,0.9652435874954225,0.12543369348373745,
 0.3731885372164746,0.7124240935711987,0.07855722036108503,0.6409263565609291,
 0.43192672706151336,0.7763189148162626,0.826657757285889,0.8518979968939153,
 0.7754615475316724,0.7582199040235654,0.31881269117584643,
 0.41406598580467824,0.9525508400570409,0.02647615248534041,
 0.6014899692729218,0.16908043860329003,0.07556291025858175,
 0.2713670738098505,0.6331280500416746,0.355211008017989,0.5038251431948543,
 0.23605123838315012,0.7171489700360006,0.5785999075566509,0.9887351796274894,
 0.5051775242555825,0.5420161339328856,0.272326637384822,0.07101808581498481,
 0.7077164371029632,0.47332614120082805,0.03682091976205448,
 0.27564063768039415,0.7121344862186232,0.6311556974388527,0.6192862881376522,
 0.7956303602136565,0.07552482310361541,0.05713820021054705,0.564501116445508,
 0.011997421436790279,0.12914932270529522,0.2920458173028957,
 0.7505903941661771,0.6385717988143389,0.6354840214031284,0.8951700296648017,
 0.780574711789091,0.002250754705212854,0.5029231837097456,
 0.30159332795122884,0.3492422962005616,0.2625095827595668,0.4593537979570389,
 0.3898498144080821,0.2688240391292057,0.7589793782367067,0.9477023839065454,
 0.37928536163995297,0.9223154055021051,0.653843617782491,
 0.030711962598392595,0.03104816753940709,0.37459670168421655,
 0.2669720316966946,0.593408186165127,0.854991798148329,0.04569753283468392,
 0.03446175028419851,0.9198370270593555,0.3888507203459868,0.5967586487114511,
 0.538371506065387,0.18536390052535734,0.594934380416004,0.45027228217879567,
 0.0017576488042296834,0.3129307001120698,0.22499542908600456,
 0.13306910165709984,0.7795384543077962,0.9789048167526392,0.4956194878547998,
 0.7736073997905057,0.024032990505087337,0.41062796022246884,
 0.34240452264000976,0.44998288570687084,0.14228633171252802,
 0.9488899283001875,0.21266527962208692,0.9050270056000095,
 0.40797483148264213,0.8169351432286542,0.7563235264131224,0.7024409830927538,
 0.3687461355613193,0.12264399318023356,0.06202323193318038,
 0.0945434834739165,0.3972047281873957,0.5805606933488133,0.21029368034197404,
 0.84620236231891,0.2553166110229753,0.6787780420417846,0.7110342709120643,
 0.6408484072300727,0.6290744127156718,0.04504609471056753,0.928679044622441,
 0.26390553127223304,0.4403612046465538,0.2097907372096358,0.6682164467054797,
 0.6160291130684904,0.16870478548708812,0.31968897661878015,
 0.19761748035450755,0.846418108690849,0.5201716952042539,0.6764676411744769,
 0.9978249744879191,0.6939657359113965,0.31656737446741967,0.468019761269754,
 0.3670440096411679,0.3080180079806839,0.684882046389918,0.7901848958122742,
 0.40394735280237615,0.3978015683706857,0.5024991056942072,0.9120656280177241,
 0.83624817483508,0.1947765190235513,0.6824215377422284,0.30985015230413826,
 0.8947752761402451,0.06918055978242632,0.2789386307324597,0.3481410601603714,
 0.32114445701390293,0.3376922235076415,0.6879804104824089,0.5416683461505285,
 0.6115879487203749,0.9156609679353406,0.7199123528169965,0.7634335057696056,
 0.7367744604367022,0.13907925985608538,0.5477765625025415,0.969570365988217,
 0.5653618161343618,0.6675593945936227,0.7740043975206197,
 0.0029260199042027146,0.14061511459991594,0.9899370696520429,
 0.22610066677100582,0.12200161068286408,0.7928106943140268,
 0.5539213871367383,0.4405358753211185,0.8926456215969456,0.49243480948076035,
 0.9004920865182411,0.2081929141423433,0.5104249324588216,0.9144476764350054,
 0.9721876128298105,0.928207577555275,0.9001194480806772,0.7093195754410183,
 0.8341104942020223,0.5967974120027867,0.6572794630639238,0.5577114365071443,
 0.07265252272156641,0.704531524249783,0.08926385910436208,0.64380288314586,
 0.3293976765696055,0.7775460757467001,0.5620194249804261,0.5234921500446708,
 0.5503026578460724,0.6483863409627374,0.4839556044840382,0.5290571733775401,
 0.8194269104712967,0.5280613185791239,0.031841482729046855,
 0.30712630694079257,0.9535149458674087,0.5666896326706953,0.7470728903507333,
 0.4644164268203721,0.3808799213408707,0.37151106100414644,0.6598638357797442,
 0.21442165604441288,0.537177817986527,0.0888780174811693,0.5782439765487901,
 0.7895187365845172,0.6120935420050567,0.49027416437752935,0.1527646835303852,
 0.3249196472878686,0.3261048414575184,0.379236813918692,0.6207487950913165,
 0.28779317612390465,0.7509656626805251,0.5525805352495006,
 0.10538240379919617,0.20465920978366758,0.10029935588467098,
 0.7911537754652194,0.004531925545843252,0.3589708527758122,
 0.021638441469832292,0.6005100989400789,0.8880153442181525,
 0.1069519014403868,0.43017584673311693,0.2114717864286253,0.7730918615882824,
 0.42655546730445304,0.836909850139656,0.731195421441617,0.2082168680317522,
 0.08226361928584036,0.2433520172803948,0.5350364197598081,
 0.18860572386593333,0.29186181430313296,0.9254991060452542,
 0.8763343384022126,0.4524263665943511,0.5143401076047555,0.6420264369778829,
 0.7852043310127564,0.49203508601670365,0.2651139387227399,0.1333461961435628,
 0.5736112138419955,0.880164419154814,0.6040328006351179,0.4684012282846812,
 0.4398719741643028,0.7038701269358293,0.3915699532456205,0.1495652524822806,
 0.1924162935788546,0.9843980620260309,0.7441608629486468,0.444042127296425,
 0.5288425663898039,0.025675848833133452,0.86812251151496,0.4116405764573078,
 0.8976518022837908,0.5233706762474905,0.5345106653071019,0.7490279612794783,
 0.39517106673700875,0.39942018070586016,0.6795871292222357,
 0.1429080491684117,0.8555172655360668,0.5376374972950171,0.08111773006918699,
 0.034782900334734856,0.9618909407201061,0.4171448885452226,
 0.26503792461496056,0.6300304964231371,0.3347234851990605,
 0.10315438487171769,0.5556884687247114,0.29333514376447534,
 0.06417378160501208,0.4540932635143069,0.4896255119172749,
 0.01693327765057817,0.6181535288137957,0.685281229248416,0.2765445504168563,
 0.38739647912322916,0.6668286851152857,0.7876759196212295,0.4770503394090314,
 0.37686281007244127,0.09071608348933169,0.933724556801582,0.3491547212507069,
 0.08194352960562035,0.8816624292471869,0.8215903698124535,
 0.17045078646025946,0.6671158442738976,0.08776078700566003,
 0.4313015878857226,0.6769090725045028,0.9480984337725662,0.5943171168974359,
 0.3619852302207861,0.28546983718514596,0.12058674383208468,
 0.19249050730075667,0.10828325845089748,0.8136320301211338,
 0.8776487428263152,0.3690494691359718,0.11658716942576464,
 0.11004190146495407,0.5013149212384408,0.6429798507628758,0.3162644665622346,
 0.7504163916434553,0.8172467191374186,0.1570760408451165,0.0896148369598414,
 0.3013144560184837,0.6436438802361052,0.3971091001139021,
 0.016983585572770576,0.8147210738094093,0.34729668773179245,
 0.0016555814648193667,0.7588459222003869,0.5436781779643727,
 0.6636335091494147,0.36134772781000146,0.47826813973901294,
 0.9537001820492705,0.4368136178258383,0.9440079646515014,0.9281452038770347,
 0.0019425513082084178,0.3460887729023968,0.7962663075510137,
 0.8439374879369309,0.0317177256858735,0.3619469199982942,0.1764676699490194,
 0.21678344405550587,0.2856760790708688,0.47967120431670907,
 0.4804265370717411,0.8428773426345321,0.36803367114168706,0.9415756475871523,
 0.4757052109860812,0.9569194868614939,0.3759377168071938,0.4676615919712289,
 0.5188151804258283,0.45755383109773795,0.18141404800702432,
 0.4243700449659652,0.27155316948547203,0.8678208215800152,
 0.43116711267674845,0.3507385133854788,0.6940343780749573,0.6172654344812862,
 0.13874699982999772,0.4493304281614081,0.4509655880523912,
 0.08717438506162833,0.4089165168722685,0.37352846336215895,
 0.9440453014400384,0.9039312341656714,0.6955978948163563,0.9238981866779494,
 0.09457590895763046,0.9005806242942159,0.7809652026712846,0.9825904950464459,
 0.679686591568287,0.5062623762243871,0.4744551197172666,0.45417294351347537,
 0.2357943791563245,0.4272545566610241,0.31768616070647837,0.2437191279582368,
 0.24243584442835187,0.32922174075584687,0.5207890965986708,
 0.7596837478465661,0.3682977510545766,0.6260302521110951,0.9568133713072025,
 0.11319908218783747,0.29451423179581815,0.586557136709282,0.5821608676808026,
 0.6583898746826525,0.915951314343065,0.2626650517117354,0.3337128509703382,
 0.6201429420400603,0.2702747494863096,0.11279359231271624,0.7931458858522094,
 0.977111538020092,0.8883969022124141,0.30262897600025995,0.17119278312338138,
 0.5137459921617228,0.8578800128250876,0.3728577601587143,0.3579513004625018,
 0.2665255330046047,0.19932933386968865,0.6851518670890325,0.1596087356085507,
 0.6639415451177292,0.5076174251354812,0.4924311768445966,0.7694950721609652,
 0.5253486493273596,0.23183535322511561,0.5325115238662771,0.8142684209048415,
 0.09613474666539568,0.17331172839591447,0.7555376015477709,
 0.8781702417513335,0.665040457437372,0.5272011943857284,0.7672380778545785,
 0.21355125165206768,0.21541191080022648,0.817126956345357,0.4646727968643548,
 0.7747282420229353,0.7382863591297163,0.8721179772498371,0.8429753630458972,
 0.5973793643495335,0.25365803910906737,0.23082233862183432,
 0.9842892387965863,0.7481180283943181,0.8748676932566016,0.03456112172960246,
 0.6219573780535861,0.7691279945758348,0.46451165133901595,0.1825140258435206,
 0.8885919077825227,0.12927838435978023,0.32086413842417105,
 0.09505433380507489,0.2890091595039983,0.38080660713937475,
 0.39842889118438185,0.995885832960091,0.1378613590046155,0.17187334277498667,
 0.9439753366502665,0.7897656068911473,0.09711131914137006,
 0.07821932882753335,0.509373632400285,0.38417549495874237,
 0.23701409835986942,0.2937630903908355,0.49396188072584857,
 0.6755028435142232,0.018720133366485037,0.7821421907442442,
 0.7036688235602789,0.7554556371339093,0.32714541690236987,0.8257912427321457,
 0.038338811512634186,0.5766572258435517,0.1861144177312073,
 0.09255204625975721,0.7618408281791438,0.5845290500061731,0.709645560036779,
 0.46249789307090605,0.3479066759482845,0.3762526794951522,
 0.48818425477997784,0.9837385131007454,0.47171493572969747,
 0.02473970974648354,0.9537215735548615,0.1328604743166667,0.8797657470502552,
 0.29987847234343934,0.6967052895726166,0.6472346418214852,0.5510498844860203,
 0.3624905358437953,0.32321892509172445,0.32316530064971716,
 0.4799419536251106,0.15936500916886254,0.5531020001939845,0.3929518402089429,
 0.3333396880886783,0.3871065759953566,0.16774808893000293,0.4767761666448479,
 0.9757839479381818,0.10831016819185701,0.2080891458149584,
 9.700456736352336e-4,0.20499394940358884,0.4173426749475079,
 0.7441707544047761,0.9635530807029586,0.8140313866718187,0.7274270588397427,
 0.9565755744659887,0.21241802767529983,0.23041878500197743,
 0.5465160136275868,0.5014284349636634,0.2520940667058822,0.9221068623838854,
 0.8393640452107721,0.07860186379702405,0.5215285039375086,0.6294470380878177,
 0.3381714269075008,0.03657590511986253,0.7893445166614624,0.9578133638289827,
 0.7420386732526673,0.1469167924555086,0.9422422752127404,0.9764574979277734,
 0.4081699421057865,0.9882551210542124,0.37397351160781,0.09257540630197236,
 0.49313615463955596,0.30861658988276774,0.17276395875874628,
 0.05724940494105457,0.5228387914110424,0.25900796621950384,
 0.6403936880021797,0.14141249235636533,0.029326062431981326,
 0.8373891302111507,0.45747503638615405,0.1436528949991609,0.7982235472338112,
 0.6755128470311854,0.6079201564901603,0.9867747708073671,0.06469295623954296,
 0.735716102525349,0.5628106251507053,0.8802546409195492,0.05993527455901426,
 0.22210309907477477,0.4505414876177529,0.5649445322206981,0.5063620938264868,
 0.3381559304985531,0.3983770474311037,0.3830369946539953,0.3451594851019856,
 0.9820453078376301,0.8984418727235854,0.767995394511959,0.8287873435186557,
 0.10651158706326158,0.6027069312480314,0.2602999454192504,
 0.018129802768726222,0.9117196732324312,0.6966629919526955,
 0.26097067394506734,0.8269373958687306,0.21490341647214684,
 0.6236696221501699,0.07362228164218454,0.24224470739895476,
 0.7659768813405996,0.6044656609194794,0.8701373132135327,
 0.012831163244347077,0.4499941240237524,0.09207428903884396,
 0.6617301028891843,0.7615849559611032,0.8568605223068122,0.37807141810899114,
 0.4432545797181693,0.04928710044753104,0.6309346973696991,0.9267032972605079,
 0.2231872922836473,0.41539313005617395,0.4308692182020124,0.570511556875956,
 0.5467481383398184,0.330858971426888,0.9629098290234436,0.7311324917284764,
 0.7894530146487427,0.5722001678117428,0.9884230508846463,0.7377926831251883,
 0.854287948079292,0.0432372386457176,0.4245145609476686,0.8334791142085881,
 0.7740318591918486,0.9494567319511528,0.5943573810717319,0.9525799791652948,
 0.539733242297926,0.3208939387606137,0.7307866424681448,0.8933114972458301,
 0.9422677371804926,0.033179200886129356,0.16534095913188962,
 0.6479010406072048,0.228977076158434,6.286335572881496e-4,0.5028001218653951,
 0.8022367510700927,0.29842470888369677,0.09271196923012248,
 0.39808613989890573,0.1351500525237067,0.524390691122719,0.8739892033936523,
 0.4773877219771149,0.008406374249643966,0.5756269423705482,
 0.44316948965786207,0.08572861706913193,0.9957603131988219,
 0.013235803542348101,0.8683666754787573,0.07496258233326591,
 0.07890160752301234,0.5513325311176515,0.4765774335929487,0.1851976045313677,
 0.12182886623238331,0.5490239596705737,0.4333136731909677,0.7258064557399896,
 0.074051786838642,0.08504924771083011,0.7864070991665458,0.43953201818994947,
 0.7743341448942713,0.9990711574101474,0.6356674896655792,
 0.028473924809551687,0.981403594198003,0.36361316177740355,
 0.6990329635794196,0.5550938552617256,0.8937647348391867,0.09165685904915422,
 0.6965416872340336,0.1945783671945589,0.6351656764125492,0.4543946839730735,
 0.8557754549159221,0.7370274656570345,0.6476079707046134,0.6361916891610752,
 0.7403163565754357,0.1821318962053592,0.28164551539412663,
 0.22202541466638692,0.5963987558562259,0.3205047086218179,0.5091675224711955,
 0.9711275350785891,0.5990972798119698,0.19910473883619006,0.4613764679950241,
 0.17186273881664704,0.9292024687716004,0.03270709908641478,
 0.9040534753471077,0.6363262707534073,0.9715129397957716,0.35662272219420954,
 0.7710023271177552,0.6881111284262211,0.5319889103311702,0.7878731110106599,
 0.7451213544841048,0.01437722492531568,0.12654163161330356,
 0.6321371408637348,0.5475941275214962,0.4054495126508709,0.29843888576558697,
 0.11149975261390965,0.9963654286492778,0.249793467160072,
 0.031715563276526915,0.2311548099932983,0.858053546896124,0.9361776688319875,
 0.747682040871403,0.8881543780423626,0.033859434990418436,0.5559510027813969,
 0.46407429580387394,0.29219330225221585,0.27935246527124746,
 0.2621035589511064,0.3305990652824229,0.2567293693694741,0.17804635544043368,
 0.3885147157100606,0.9086413231834956,0.9754296762427066,0.6389542537431623,
 0.25941419402132304,0.8472483002932365,0.08254881401148073,
 0.28584267986034584,0.4067516553739078,0.46697755051685697,
 0.4681216349103968,0.9728680141675244,0.15816896932715707,0.5785163937208186,
 0.31257520294870733,0.26446605617810315,0.8747787447161097,
 0.1300543845626131,0.6857556081338738,0.06237488825573301,
 0.04555888618764137,0.8835186919322925,0.6287419471308417,0.5209938528713489,
 0.8724256672108386,0.09817316146699218,0.47040597926550864,
 0.7483791785573991,0.042906583123560216,0.5429828799167673,
 0.26265570208926736,0.7819692703846065,0.007592168550278222,
 0.524915341693251,0.9623770765968931,0.280945181368048,0.23978581938447596,
 0.9627406659198435,0.736876734871262,0.3332060272188391,0.5864259852276072,
 0.4966763911510297,0.5063473800557274,0.5720023359236804,0.9987954791643107,
 0.9291238168916317,0.9297403368210775,0.5490050905120917,0.12991811080127835,
 0.003272965424864993,0.8779182309765753,0.055356913130857666,
 0.3106315374102768,0.39768288787517236,0.3379900008275032,0.8139632665566084,
 0.5137472525620216,0.6152480286762138,0.35798698725160727,0.2904770056085393,
 0.9565418436249681,0.45583693406901116,0.26617391849330685,
 0.9243685348225316,0.1203380605275235,0.6917850191121362,0.1350081318164521,
 0.8624883044894828,0.7647508628648091,0.704949886315889,0.7449642981723291,
 0.7831948900574693,0.6692616091214701,0.3062431393784192,0.8844320753390196,
 0.1895831812484784,0.20827596919743574,0.7582508277684497,0.9140529349547845,
 0.8253778246474998,0.38256598244881956,0.9703180935340476,
 0.19612006131261617,0.4561787811329362,0.827749078965424,0.6318131169698581,
 0.6522356163362854,0.7913133672521763,0.22176407726680625,0.7614126380906405,
 0.9029882344103669,0.3779014572880698,0.5770529589991985,0.6281759319261843,
 0.7848890238019339,0.3961754338728577,0.4769291731200235,0.779638067871406,
 0.10493193509432408,0.5436492049396103,0.9252581985018168,
 0.17319264233753207,0.13784614411167317,0.3662273593150249,
 0.017338814666922886,0.17397222030517578,0.8603017484918704,
 0.851698613508769,0.8121614542032124,0.26290045281371366,0.5522620019895798,
 0.8053202235672108,0.2178139381716898,0.39628063619539455,
 0.09717582811686698,0.9587092186869876,0.26604806998200725,
 0.6203194923452897,0.5368499519881546,0.40948772810663014,
 0.06742515277665517,0.24555370366574258,0.38616673161458825,
 0.04697275790599187,0.15890559947699956,0.8420589790403261,
 0.06623116403626694,0.4503019640602923,0.42818711045667757,
 0.09576180899356168,0.7661081745947863,0.1622024390306216,0.5860328758316606,
 0.037953724365562436,0.46179699519805695,0.1322390038361303,
 0.34566158061967955,0.35856838307216365,0.32677815046973535,
 0.415584484929329,0.19834397603462461,0.10537302643300583,0.7684938021807749,
 0.28583102604367194,0.9959527947778051,0.11944342030368749,
 0.8993764884584646,0.6357293049636276,0.8153951613215419,0.3950781247520958,
 0.8808000499557266,0.18378916921142663,0.5432671435707377,0.8798797267209065,
 0.2792458494531771,0.23649988512316755,0.44835672971021,0.877066143362312,
 0.4258593764976222,0.33374082019592866,0.7562966723527108,0.366659662789769,
 0.7866630015900695,0.41594843791382385,0.6579344241599084,0.1370800342057985,
 0.6212038853480287,0.335202403804463,0.7969410199648602,0.8775602910528174,
 0.4837193095872616,0.9692613234592616,0.6214421192220604,0.6106579037835127,
 0.5106800512630868,0.5541055223928187,0.7249519343903139,0.9226686791794279,
 0.16985534160252114,0.5032864878747235,0.26649033772568265,
 0.6828979657325887,0.6723066564770396,0.8381674162824255,0.1928796693292658,
 0.5045552329158398,0.4728857396744345,0.8593207531150249,0.6631135028550859,
 0.9823058973816636,0.05152807192612252,0.2916311336176888,
 0.056086640232424156,0.09336634519752307,0.3692512612632609,
 0.7784096803020525,0.22105637234286613,0.8238086153595003,0.9978888477768482,
 0.7890421318419193,0.04497270612371751,0.6495632612254625,
 0.08915266348288886,0.17118843313214338,0.876270056405516,0.2066825521055159,
 0.09129367569695324,0.5786228083836872,0.7854131005666332,0.3926265388247162,
 0.46295062764698036,0.3061610267027378,0.0907756945244601,0.5111498690614226,
 0.7448350867927174,0.5969447167515528,0.4193393904665379,0.7671391549840383,
 0.7950894738694566,0.2894225241800562,0.158030547347639,0.895998586854661,
 0.6385000766170675,0.41261586927169036,0.7946202767687046,
 0.030962074606538104,0.8399354190265287,0.15530513149374436,
 0.9870797502048084,0.9470400901692395,0.6922829527557183,0.6539621157901743,
 0.42619300791426384,0.5053280176470636,0.3673385936335316,0.1626053620349227,
 0.790092654044691,0.5594458800024366,0.4813825030464707,0.48364640092954414,
 0.7373399307390557,0.5025500880113483,0.2938633193093574,0.7642414292123341,
 0.7759404977309751,0.8085788780771702,0.5051967406160945,0.8730729232673723,
 0.16042248365790313,0.6226024872855777,0.761893240567464,0.5556388765868677,
 0.9068439420421608,0.6189119997312869,0.7458069516313534,0.5933838986313378,
 0.22715845438727178,0.08928134558162215,0.4941277446714898,
 0.8929300933277061,0.8920611846463028,0.04046446750448518,
 0.07133313383560602,0.7345961798419305,0.4169121646444607,0.8242341742121663,
 0.4868971334239528,0.21531568319540717,0.4120173199457128,0.6111593986050362,
 0.8468787648909072,0.23547555867562098,0.19686068678006052,
 0.8676919249300324,0.008246919065376113,0.5845801099333686,
 0.7237813389175235,0.7860342341222012,0.4471837507331857,0.7431773797934184,
 0.21379754731511358,0.8991354560759746,0.2754606170051135,0.7077444800343884,
 0.37306526752517855,0.6720421247862864,0.11032125092708167,
 0.4545246446342892,0.8464782002835152,0.3265616546202582,0.14401095759825033,
 0.5132167303046375,0.6929906076285626,0.4836609105937507,0.43042846189144746,
 0.5795360323786521,0.5680456638322191,0.6370487227200394,0.4450189461052768,
 0.18977026957568488,0.8889258513177706,0.8067264630684786,
 0.29608317700576003,0.12251307976070502,0.15463164914327177,
 0.362535673419383,0.7743934771420515,0.7797265600534572,0.7900563528215108,
 0.21722720371965587,0.38297905541625044,0.5042074083141203,
 0.03219463911334941,0.41890420895991565,0.12989977328790325,
 0.46915983733475364,0.024699813056988984,0.7073762360331785,
 0.9706218041663097,0.8087057808341696,0.3564219599396369,0.6143794913464804,
 0.7994373815145366,0.6928928867211537,0.38776242620154855,0.6512115861533163,
 0.3042225990456412,0.8175747725316995,0.957858995124363,0.7196004349961421,
 0.5390617627900175,0.3659430519086456,0.09208172196561448,0.8562931825784748,
 0.6662935087619738,0.9495003735483952,0.19494397704746425,0.511721084318644,
 0.4633151508275226,0.9358380598564368,0.06417895550724306,0.506322295415002,
 0.6899582787421044,0.6003648131350857,0.4811693507161471,0.26900202619611746,
 0.6865740531198634,0.10871891468126904,0.35904700538623846,
 0.4524462879614889,0.6495768436583246,0.4729277543428505,0.1354606905969612,
 0.884054785993664,0.9441707904417409,0.17747216538854282,0.1365361326650506,
 0.4098862265972464,0.36393880996203465,0.6730742889575572,0.4004694858429644,
 0.6819596026123117,0.3399846880062215,0.3396163587797876,0.7965190955252013,
 0.32372273056670886,0.49680923292755264,0.6145619758960247,
 0.9901565080009049,0.7298101523416977,0.4837102921628219,0.17450396621460085,
 0.4679222410370488,0.14595681009806305,0.577351013636223,0.9445050685124068,
 0.42255270688465074,0.6092955964069047,0.1297724040067444,0.5751090567332896,
 0.779813893820876,0.5255382004013819,0.6436138659870227,0.2743442971657192,
 0.3962832410224373,0.6246450032154621,0.9440683140809922,0.12131792435299726,
 0.6185489726317004,0.9611089201294918,0.21754843810485847,
 0.23531609592512104,0.2681663661837467,0.8172686439829602,0.6722036937779027,
 0.4331370333421126,0.4038649995330498,0.9672666344975357,0.49659809509985664,
 0.9392703275929399,0.8971750680234575,0.04243066248846428,0.4710565773330535,
 0.21419321722560014,0.8868529780009707,0.2946460431673542,
 0.24441155973903517,0.644694379598045,0.5112985049976128,0.8026464680090877,
 0.16583669463142214,0.2454321895527357,0.15775978252118528,
 0.9944743736025273,0.1791688779101539,0.5380572334185479,0.6651796478857682,
 0.07906818929827719,0.9568043282648816,0.7562146750918746,0.9221477560530443,
 0.10260289387418764,0.8512945267545938,0.8014650332145141,
 0.49495006424485843,0.5686363030714456,0.8506997955433359,0.8938099001136881,
 0.9552779684177517,0.23519214268266964,0.038079805978792236,
 0.2474669865170256,0.36982723197751777,0.6832026597276339,
 0.19075289840243437,0.04471996482067875,0.2887620549357437,
 0.07431929516446212,0.6085960610612721,0.932169392951464,0.7113072229399179,
 0.7182853805332896,0.9359716985141548,0.39894077433361164,0.5399178914000382,
 0.35497288954686257,0.5270664510450138,0.5981896846274842,0.8567589736215605,
 0.4998562719042714,0.7552600482443638,0.6444390121407191,0.7709504255881177,
 0.9391800853100974,0.981144785625276,0.04610268682823293,0.9798718100486988,
 0.9333631122928453,0.5173398517425605,0.8970737112404836,0.8588487787504667,
 0.35594585577039306,0.9622239526223766,0.8426479067385866,0.7008237042771965,
 0.9600480636919695,0.9281332667208877,0.5735006081411355,0.6202414326762534,
 0.1536515439580094,0.27548718637525893,0.7356084236469519,
 0.10097069827936844,0.16138156968416084,0.034056185984817744,
 0.17754487634753724,0.15986473524308964,0.16709568393809027,
 0.13491656738700863,0.9585875270942757,0.7972340184147009,0.8750600792256232,
 0.30850339317141084,0.5185626682349458,0.697048152113456,0.6567758968421832,
 0.08218008274109101,0.8347094023839109,0.42013006692671484,
 0.4042415153966389,0.24181855746848246,0.8802287406671616,0.6330626803155868,
 0.13807226351238744,0.047371393895385694,0.5340282056077623,
 0.8965652724216484,0.3757331741995029,0.534368420910056,0.5025175286196124,
 0.10381214768757918,0.8425750086901473,0.6231679769884192,0.3228431243663541,
 0.6335973488499455,0.4629821964169325,0.2609086937873675,0.8255994850536582,
 0.818106684688399,0.047659687487522584,0.9416188392962161,0.7212434616245315,
 0.08204895643902799,0.7243840655426631,0.630477578527346,0.4748945790954705,
 0.8430652396246966,0.09688225389271732,0.3575897604373448,0.9903612015962051,
 0.7655692649546464,0.35082762855042193,0.7792632042764229,0.7128385108226556,
 0.3006661170239061,0.9517611448401809,0.7845819807257374,0.29465963054217115,
 0.5186824997320683,0.9711138560423999,0.9052771008217291,0.4080513389076449,
 0.8053607831368136,0.47523021390385967,0.6952803129377534,
 0.022347372526288267,0.43323029736522134,0.7411298069557222,
 0.23039942168614513,0.5646821871582373,0.7135282831942488,0.4894049066505439,
 0.796648726823418,0.8366505943473598,0.3305524821909558,0.45809453644407805,
 0.9008160061840094,0.11929995771559054,0.16816132377160464,
 0.24286783958015934,0.5315287423598823,0.33179476111083805,
 0.06998928321041331,0.7595213743063469,0.5488715758692945,0.4111345315313475,
 0.12188755931530815,0.3686650245777773,0.8467525630209257,0.7561656522406364,
 0.3702661060491418,0.07476639016195608,0.793994686128712,0.8445524159287241,
 0.5781721234993816,0.8837456392404073,0.9710276503524381,0.8805777217884964,
 0.947510937840013,0.3496050708116313,0.7245913304882338,0.017998253845747714,
 0.6313995431320993,0.5360993133277686,0.6773418995729567,0.5263365190233168,
 0.7101572747733125,0.9642993350329887,0.05514168607670156,0.8001795860976101,
 0.7291864845805072,0.06494841840286314,0.854531963149008,0.1592671362858401,
 0.013921461581765993,0.395342499370572,0.6804151808612293,0.4021975235829106,
 0.22904131558593832,0.30004036063761086,0.8984820475488453,
 0.2115205972348514,0.3277757647619046,0.8515955942256619,0.19934751122285232,
 0.32465777070003743,0.05329590993510769,0.8619765145629268,
 0.1708698627141697,0.5737361180654986,0.9563125020351909,0.3938026125762537,
 0.2742951057312141,0.31406536032202026,0.7978632470477608,0.7417359628774586,
 0.9622153893334637,0.6812869424813183,0.7033689553751605,0.44658033857167434,
 0.7597384881047935,0.531996685060995,0.7386862347083653,0.7629172372539513,
 0.0028199358696913146,0.9029147574540197,0.5410949912247639,
 0.610021992114949,0.03957201786851838,0.06269862913498758,0.7194092527860322,
 0.20055718321258797,0.29404744749481104,0.7854775408323202,
 0.7382404521881916,0.9968264677904012,0.6571478217437897,0.7589740037098038,
 0.9885030210688729,0.012134803048815757,0.08988895855224877,
 0.5750477974630361,0.7829351080006743,0.7467442197098515,0.29531289714298725,
 0.9142213596633376,0.6325071549465382,0.5920141740617824,0.1775308085897772,
 0.01657420488543271,0.5283810559038958,0.9514789629559862,0.9007998223939835,
 0.38543671347636776,0.3216221909039506,0.30339894166636383,
 0.23567963833613637,0.697538571618544,0.8076773766767766,0.19300204095652373,
 0.597780808054561,0.7093037136385734,0.6924799425184783,0.23686647680809447,
 0.7615984617460052,0.7958026541699598,0.23744787700181846,0.9851003352543587,
 0.9052567143137791,0.11653016490343082,0.1603631070236161,0.3631182095580988,
 0.3407405232633124,0.7930894737759226,0.16789660497144532,0.716225877505525,
 0.8315380346276565,0.7758009309186252,0.5945700998844937,0.6173420037682783,
 0.8241596001543674,0.2394254814696024,0.5767084181868954,0.9876873622747092,
 0.7661365430868099,0.6166294288101224,0.3803485275915254,0.8743367077240419,
 0.03433821908887802,0.6164680693775404,0.3853278095211289,0.9345514962509792,
 0.06920910875332453,0.6619331445952312,0.23650980387943976,
 0.3429715979828245,0.21772420245295343,0.4617783722360741,0.4323672278142995,
 0.1169980468930506,0.9001446061944582,0.18195395819755655,0.153379613164065,
 0.39304536820870106,0.8093195169090086,0.070131810756344,0.29807816395655096,
 0.29106963482731874,0.11598456924371137,0.9347895567347583,
 0.6030713851070201,0.6643987069307214,0.45822361509331455,0.5629439968203305,
 0.890406290820802,0.16690247165858108,0.8281869652548914,0.2628599907769029,
 0.03916155145767419,0.22036855098232833,0.8007259781687179,
 0.49577427635171345,0.6570308714670521,0.17278043531909215,
 0.053346117442179986,0.796150521185266,0.5814267172779531,0.7423220655064716,
 0.4093330447812451,0.45959813112060466,0.49544854905490143,
 0.04464952172839198,0.9505116726382468,0.7114911115073962,0.994959953246124,
 0.9296146938320085,0.3162574119517271,0.5623598595836365,0.29408215469914356,
 0.5408098447822532,0.980177081004175,0.8256759658284119,0.8500008973285721,
 0.07660795592937131,0.9075381678137531,0.9559045591194045,0.6635742367888638,
 0.8083827352470802,0.5943492424819456,0.22016754795182125,0.3688601825542319,
 0.10625092871284714,0.15525145046056488,0.7426022005474677,
 0.45456026008445394,0.4406318864214107,0.33022100366823504,
 0.03809264494101694,0.011631725836717677,0.8575797702126522,
 0.7435402486337754,0.4141052391369895,0.12455515216691648,0.6105353552889966,
 0.04160673075093979,0.9761619800237809,0.9218658562677079,0.8052040255761362,
 0.7937033757782935,0.4275593356452878,0.5160906567612954,0.05028234995932601,
 0.6757686560824766,0.008706322473366224,0.13331989266448696,
 0.6790129150155951,0.18078294855755805,0.3508547200081874,
 0.25227783955558747,0.2009088018202675,0.9529345892043003,0.8141270077021564,
 0.5784389409294668,0.9472399297619033,0.2251685554438848,0.9837373305392246,
 0.19072567403299234,0.7690188876369641,0.9916521523380921,0.7734214167956575,
 0.9130242903299677,0.5532366449791104,0.6530008721065452,0.39975809236795123,
 0.5091664940149186,0.6864688660687466,0.3879663378963325,0.9513821159331528,
 0.23202199454946548,0.9760996844198407,0.3863241025169466,
 0.05375656445852195,0.8604681293492873,0.4316604953727541,0.400489890251844,
 0.4711968520875265,0.677930195738756,0.40060783106779274,0.37299103210464013,
 0.602352134977722,0.4186370405916784,0.4560952342033803,0.2084147025787343,
 0.3371083331730472,0.11606448135726843,0.4284207633707332,0.9047768730044884,
 0.6559321363433872,0.48085668482261257,0.3237456617232263,0.7020354349992659,
 0.6959939179728619,0.4951633902943525,0.721456644575582,0.6671171710883653,
 0.03923930457100899,0.31957267696732905,0.9413814666404782,
 0.32795859904826585,0.5988556565546889,0.7898743480014332,
 0.24262240900416432,0.31171661144025165,0.033578197659563935,
 0.6253799352927906,0.2604590654852518,0.26825961553790567,0.5615883589895208,
 0.5919360608084789,0.8137633709331122,0.856114914501402,0.39193094964502295,
 0.34321007598419695,0.8499487935120692,0.8787291663669572,0.802793584859208,
 0.5268479173862936,0.4812038827403363,0.8231348865355012,0.6172132635462635,
 0.4846346434159696,0.7191739054221747,0.6295369399226327,0.4229969154737788,
 0.27149222197471323,0.9626357647296613,0.5781082209844677,0.3449376843391089,
 0.3506719881406307,0.08215189913406129,0.047842436580696424,
 0.8451342904931104,0.2018965899298958,0.9072724671659645,0.6388009227481205,
 0.16090614384465474,0.706547303364894,0.9589857585415789,0.3194424537795334,
 0.49678514275511476,0.5456573888215814,0.6173254197633927,0.4036307817260898,
 0.5778885540029763,0.3017124439947376,0.46892203562023016,0.717149346589819,
 0.9644717741699068,0.39396635851890727,0.8126920193366597,0.8822205755003447,
 0.3795043665704514,0.0108659776505311,0.9979275109253571,0.12237833717598023,
 0.5325411020917716,0.08137602656705922,0.9994008625485329,0.3519042456105901,
 0.5960495396147159,0.07368473665160513,0.0786740332993826,0.5636330655314631,
 0.14338141390378756,0.24339622928899274,0.877808647714697,0.2319829796145093,
 0.0011570239449536501,0.26426248322576884,0.8341908103421667,
 0.9020727089139141,0.2462409226916118,0.787006654463887,0.2754541624641279,
 0.16328462420621292,0.40713457885676463,0.08143727298265879,
 0.7909115254873278,0.4379868105199425,0.7956399285706406,0.1362113488020925,
 0.30684355500886107,0.6906179799692583,0.9112401233517403,0.862202124296776,
 0.2906697022759749,0.43064028008489585,0.0068658089442109604,
 0.31207321199601,0.31419138953660775,0.06757547195081792,0.27506294681153587,
 0.9914438531563551,0.26265666078496486,0.9686824522774273,
 0.48933325186725374,0.5944522779612728,0.4021129376175616,
 0.025596991085361287,0.49964561428686927,0.563940808178065,0.481221492607392,
 0.3639827929253734,0.18985456260442413,0.2584517146409978,0.7704067163920446,
 0.9676957672616607,0.3170221873539636,0.5084729460086039,0.23976700475634172,
 0.5999571973821234,0.7823359619710664,0.21274091655530936,0.829300380372634,
 0.27598476525531523,0.2561919411492066,0.6106389453790669,0.8816114169887969,
 0.6483248349742088,0.006765680631486748,0.4276212969927444,
 0.18524477394239525,0.3943137787068256,0.45607659177660853,
 0.03763882938084906,0.5202268159789603,0.6604315452363511,0.8015717251178887,
 0.7901285396696375,0.2105882964539103,0.947600328304401,0.8256966509249215,
 0.9169578884724279,0.12913242994147478,0.6816034960817015,0.1535876959063568,
 0.7403303740684929,0.9379449523381125,0.7074048466134363,0.6984048051204479,
 0.8642906647267472,0.8164312642101238,0.8134607992068126,0.9309664521826861,
 0.5695386460570435,0.4517082616392687,0.8372633980831905,0.8496951492981227,
 0.39536288089145466,0.058893620800182345,0.4567217126062484,
 0.9421830451620931,0.9790874597298337,0.6866981651657101,0.3530832590143298,
 0.983398425566924,0.10998402036376664,0.7453494999027833,0.4498163377565798,
 0.0603356597917305,0.9805962868076173,0.9847477923920461,0.06912016781694463,
 0.9126337520821317,0.4141898918315521,0.9390630367156853,0.24836848103067943,
 0.34105323024941503,0.9969678918123147,0.3327579534767616,0.808784325840115,
 0.10535586055833845,0.9743107250310645,0.9328782789761562,0.8355470647349129,
 0.1515978633265711,0.472333699441831,0.7736676742585911,0.7230430907044246,
 0.5491178309714806,0.3656798167369619,0.3475949776492874,0.07488302267852331,
 0.17672884223991236,0.7577124940974981,0.35221077818581437,0.411606332987008,
 0.1193818818051654,0.6570809642739649,0.11013936472904906,0.5093030740567578,
 0.6497774063501638,0.5892128313519269,0.3232750163462369,0.9371767648189139,
 0.1297635823235299,0.8256531450480511,0.5150597351262125,0.3702066583102842,
 0.34340467457180446,0.5830637144120214,0.016942870091670614,
 0.20240715917469243,0.13745975137809896,0.5562855119480796,
 0.7310726180485481,0.5805328452442474,0.6440557019123936,0.7153124935737052,
 0.145001674747943,0.5641751142773231,0.02693660224536054,0.8054004531757493,
 0.2979705853105553,0.41096507075649535,0.8660085336791483,0.5581446995174901,
 0.7318126123274686,0.17289509007683568,0.62647423562912,0.9651312911971738,
 0.8653892313314535,0.0904874772100509,0.8795387921268406,0.8374507337815781,
 0.151027905820177,0.24204740366780264,0.14976693383913198,
 0.29340921255826213,0.6141148799683578,0.38016651845618565,
 0.9776868094325649,0.8504094361996268,0.6557307064522655,0.3019126675984558,
 0.578282389832212,0.8668087737852695,0.491092882621456,0.5190383660964087,
 0.24131578770205775,0.16188379874386893,0.8425149604477832,
 0.37246153077096356,0.08233805647223791,0.5153540555587879,0.268386643712208,
 0.8121415820452468,0.3112944633314968,0.32330399811699284,0.5672094520229409,
 0.9957693596799503,0.49903315718393226,0.6014672688261469,
 0.16747440824368598,0.7433305525475586,0.8229129866433873,0.7143523026718621,
 0.8046102240460447,0.9895635641175374,0.5680066914182721,0.5362810708653292,
 0.48964306424435433,0.986933006936791,0.4462389974708654,0.8574637711605498,
 0.5209287701395405,0.44553877347438287,0.5477084809110091,
 0.27952840377184573,0.6804880713076592,0.8161288565314,0.7689376007838236,
 0.34513945054524764,0.31407433406842333,0.1633717837093741,
 0.7065380657829855,0.8344674933225735,0.34827999071939897,0.6731476885958071,
 0.08434400259084407,0.5742325330284692,0.7822008246462461,
 0.27477797282847716,1.8337099386522837e-4,0.5673850207063148,
 0.5275029753527583,0.004562589778335457,0.06383652658195804,
 0.6308704994071767,0.7788429503300227,0.7204809256150618,0.04117182848592005,
 0.9489349471918151,0.5793357079263237,0.029919770723658035,
 0.16355152446850485,0.7207550698988565,0.7391258438742507,0.251489892328823,
 0.3139720687096835,0.9655207007444955,0.5607724666569951,0.0945962694684217,
 0.7987436679044095,0.1219555241678687,0.12675749091955546,0.3448593206891777,
 0.9521195086448015,0.6345132304379896,0.054389829070063844,
 0.6431974912543028,0.7257107271043544,0.41546849607416514,0.6041811202438842,
 0.06894107647837588,0.5014977415659185,0.7845768892339244,
 0.48462997249219164,0.4321169249175192,0.5117528584618436,0.8198325700262897,
 0.8728435185415716,0.707650966448944,0.6471024498732102,0.7002734626812551,
 0.6638762855779712,0.8677964694969916,0.30436999207051685,0.5585214556481627,
 0.24681215866172568,0.4673188096666203,0.8511680817541394,0.8153833094138937,
 0.6390924303984586,0.20136377138004313,0.2219470283762095,0.5638964139479534,
 0.6212076943218482,0.1072541777005791,0.6069496738222574,0.8370064752406747,
 0.361208591430648,0.45401168085803023,0.7495138952689694,0.37904449449058353,
 0.3659627315582319,0.4150126044409964,0.6488643489892845,0.926308390686187,
 0.8017817493839005,0.8535463481420287,0.1810715377749399,0.198436403118605,
 0.8422571535892522,0.8907902953028413,0.9081648828660424,0.7307687916814323,
 0.4139849166680656,0.9877114053754279,0.19792241922595533,
 0.46378172604222345,0.5782432991493942,0.5353759536745649,
 0.09826740906185127,0.46544012418899006,0.02894396746249417,
 0.7316829796265617,0.20654432614603935,0.6552117233546122,0.8331807112679324,
 0.7780106323865805,0.21655343296427787,0.9156733725345554,0.6979317983544512,
 0.48409490962050694,0.28129425831414656,0.23463582441079056,
 0.7561154194149294,0.930260391819954,0.8467941088333257,0.24179634431378283,
 0.9120946134350102,0.021249647649492553,0.6735042774751494,
 0.5836161670874975,0.04527061101862062,0.2740613801375613,0.9401792377203023,
 0.8425756333544924,0.7424377941379038,0.8166757725018351,0.5641143523762896,
 0.8743743685834541,0.46611194362496156,0.496475770961732,0.20374574576135807,
 0.8929114709289411,0.6120107503243154,0.382760457930322,0.6524426314592227,
 0.6158598814390094,0.31982483750621493,0.3455843322886034,
 0.38280938645684104,0.7057288700681343,0.5413701557284406,0.3882608013929423,
 0.6956363901924368,0.8390339448206238,0.8353256134448473,0.6193730954880601,
 0.11757306430937742,0.22179952719118556,0.688350424988828,
 0.36354200275290904,0.8952668956395243,0.8099916594481116,0.0495052152459482,
 0.2644715298910496,0.1083348838734508,0.01573866805143509,0.9885184201546959,
 0.3093548252727243,0.08232895357858772,0.2907886710764167,
 0.11788847797465807,0.48473251719826604,0.32554818832328714,
 0.8594435404249414,0.383476140169401,0.7342556634832504,0.809381101466909,
 0.23865778656838244,0.7507442087156134,0.9190426198748483,0.0128667180883556,
 0.08091248950949659,0.566289414825538,0.9336933275288802,0.9031707891063389,
 0.8821384892596391,0.12588275054294207,0.7530908220645199,0.794117950524682,
 0.6895638446933401,0.9615082745978772,0.3743677311256872,
 0.040659908605123696,0.9474338385769137,0.14753150509667334,
 0.9979511039837138,0.7805362443706877,0.749041678706058,0.3810449872778773,
 0.6091081834012163,0.8832128774724736,0.4121775313563524,0.40392682763823584,
 0.5030912212666784,0.6099119297240646,0.5209370499569069,0.6795660063368701,
 0.11197231339215108,0.9831008067827528,0.9256190490155021,0.7624086506004655,
 0.541874429955982,0.8765098964954702,0.09461302733737875,0.6205667939296932,
 0.48660720128779333,0.10603923552884531,0.8753207800753381,
 0.06119776947799116,0.3094501599140468,0.1400385219554774,0.9590691431204335,
 0.24651713799716124,0.40055115246174,0.8163945454867045,0.8826714825003916,
 0.8797711077661527,0.3469921802638216,0.5309724135023168,0.47017929546355,
 0.9766012157816393,0.8597848529889147,0.5335523160741173,0.1172165309666009,
 0.3479579276688328,0.3921985027887702,0.15495009111360258,0.9053636186076521,
 0.7707426249054418,0.10104079717054804,0.26431992097223767,
 0.9809760829966879,0.5458323091358145,0.5780888396863393,0.6506467627020804,
 0.09903291031916972,0.7767121309079119,0.3228121177924935,0.7847388841792504,
 0.43978417738036457,0.44537491831722786,0.3886594274883739,
 0.6327820929189757,0.405028131911925,0.012099532578824013,0.5034539298240609,
 0.42759935969536256,0.9622102348454893,0.3098991497444006,0.6078529135025329,
 0.4786328468195623,0.340017290996799,0.4664940461026177,0.29552989415768327,
 0.36199271953610357,0.8249795225568826,0.6621134153016272,0.7667556460853121,
 0.8031327322209958,0.7782747676130962,0.6409794729739369,0.5721310603847111,
 0.6017479431407392,0.012909856948039367,0.5741145314498227,
 0.7361507476146331,0.27064023790658764,0.5051711417196988,
 0.07093515164791464,0.7512559709388618,0.6700944369558669,0.6727355634081909,
 0.06286666842031363,0.5616649418734143,0.16261415583080285,
 0.37402609247667506,0.6379039938062275,0.285087911944785,0.7596719920616312,
 0.47956758760432816,0.7807029760324157,0.14482660750454812,
 0.10946832514647642,0.5007821165801847,0.3623955348418517,0.0843711820982167,
 0.5894807751271649,0.8885967202754838,0.28601605508976247,
 0.33662841607980243,0.4451838251909047,0.8127927411634598,0.4919033460868477,
 0.9240083277615874,0.14240152824351515,0.8691228081585646,0.5743961830264039,
 0.5046338784311337,0.029114832080522568,0.4775279670945739,
 0.015068260847033943,0.828519407028396,0.6048645620817442,0.2819089115554124,
 0.7926434081360153,0.3803262886362033,0.7076256433605245,0.10694604113656436,
 0.8050108612616608,0.5319004953212421,0.8083615591440555,0.8010674108297047,
 0.15752697700070684,0.5489183565854243,0.3012596473556982,0.3401743489547465,
 0.7352409952573706,0.9298930316889797,0.28099738254932316,0.6210636520776315,
 0.6885823799508612,0.04564172244939391,0.5644430588622485,
 0.18644796561223342,0.13172153187957725,0.9546054836683234,
 0.8521684397084279,0.5293941249243602,0.31674123370734864,0.5313318282730347,
 0.16684095775876318,0.5148201071736614,0.46998243883840496,
 0.37817267993914583,0.9040512453220287,0.33734626643900023,
 0.17447844227737153,0.5733916724527278,0.09633490083937946,
 0.8763820240781198,0.5713439435783038,0.6660262667066732,0.6815148666667273,
 0.7435993058141133,0.6293668305808549,0.2010207375729931,0.7150282265066481,
 0.99667414984592,0.9822270255934445,0.24144191168529283,0.8524225801786809,
 0.5772771538997319,0.9413377571210139,0.7317222263308701,0.31427217679741803,
 0.34315580724892547,0.06215675746651739,0.049974695173763406,
 0.12023103131215951,0.5645848735836971,0.9115228981814515,
 0.44252457912179544,0.4357850650671309,0.6016649382890309,
 0.05649638130193724,0.2838561472060984,0.4726571750337456,0.8477840600517698,
 0.22380154320205314,0.3982835689837364,0.9666058594681664,0.8402113992812292,
 0.9115933493721821,0.03424924609014279,0.525171927633019,0.21847053123726734,
 0.8710682451236273,0.8327900847875572,0.998481523889466,0.7943283545627614,
 0.13931457114088674,0.7702609014779404,0.017708207582150637,
 0.4577863407553018,0.40258421108219977,0.31749941569022844,
 0.6918387257067906,0.9052879507367338,0.4983241546682078,0.8912135895057902,
 0.3060597740923976,0.8611829281150938,0.7890781419228162,0.8532852280470355,
 0.76200639049438,0.9561514838376703,0.835329943519526,0.8396903580760618,
 0.1354031488390648,0.017425907617338532,0.7132063757518277,
 0.11638217636260695,0.05714746258979275,0.37933198482418284,
 0.1387842223215897,0.14735069500571496,0.8147463258974135,0.622068112402372,
 0.31987445614047505,0.8066181236917347,0.11477758394307114,0.661457545300385,
 0.9230106830032052,0.8182580812509378,0.4520411045749485,0.5894178583849515,
 0.5780640943258264,0.10204276389957434,0.7638087158471518,0.5448748120373045,
 0.48689260144171576,0.5334628816602848,0.6623166151479303,
 0.18650252600270556,0.7423973793569305,0.6035255128532153,
 0.26076383762753863,0.8279633451818171,0.4333553919018095,0.6664508659176444,
 0.6279055864506752,0.05975087968992365,0.5373964964580085,0.6938263666037671,
 0.576291007188839,0.7718844655360969,0.9543835867465802,0.772795758053662,
 0.44303707908015877,0.6557158350192208,0.07412049291109224,
 0.7254937756697288,0.14151024959557568,0.40993941313826787,
 0.7924404707569304,0.3682923468884708,0.42150861112087923,0.6796256193568433,
 0.932034649091625,0.9992890953698241,0.6132760985945624,0.03135481716954025,
 0.3950814588694501,0.03348830992508178,0.3741470940029077,0.7419516513177221,
 0.08787713539323994,0.13679888748860425,0.6390148145897278,
 0.13197792377700768,0.5863473629027725,0.6256300351636728,
 0.19392140416318615,0.6400665943113237,0.3148428693114501,
 0.13838200147509538,0.4390033918622541,0.7380949495902229,
 0.32126768768129654,0.8434988649053771,0.628419401796617,0.13271389892830177,
 0.899274914347506,0.5256426079103306,0.11838145843298231,0.1883981177945474,
 0.9615773325276036,0.09633443581446488,0.7456658047141883,
 0.04447872867037761,0.5725798947706109,0.7948276326431843,0.9603848481381453,
 0.6519077133426843,0.8430222593588836,0.6073668641513432,0.12527268076318632,
 0.7811390624824167,0.08211869583907383,0.3171846411474184,0.1900387024447776,
 0.012263943441054637,0.3100382309921156,0.09057407951380503,
 0.19020892656435728,0.3660339159627348,0.34238864854776696,
 0.4241258926462439,0.9118653074100864,0.019289529946154516,
 0.3616505569909855,0.4574916020431642,0.8399231635563326,
 0.002454633379861293,0.1470640052685931,0.30922874397553657,
 0.22807475049268722,0.21751116697693595,0.6593495534628089,
 0.6818067611855705,0.37093230622396745,0.5725679625937155,
 0.31596317609277913,0.9391497449746278,0.04814233392968659,
 0.33392492100482707,0.6554181242271131,0.5677798890081521,0.6673112520181277,
 0.7831916928657608,0.19275641888643313,0.3449248461776574,
 0.29928755285565833,0.41510193449546817,0.8397812710299067,
 0.3057095517166989,0.5807258975202373,0.43918077477855644,0.547600467814679,
 0.6953648355808213,0.8487376815458716,0.6226545898859586,0.6111212767422152,
 0.3372643875248993,0.3703372107533278,0.35756142784425593,0.1963799498165747,
 0.5413322535841724,0.49415192262216245,0.0870737413001863,0.831145968290642,
 0.37931572586683326,0.3763002303293055,0.8602128862337893,0.9515755943191077,
 0.9028247819085163,0.23809983047731675,0.9956590087147338,0.5173024378033362,
 0.0830119131161422,0.008312750357677534,0.4568242038099587,
 0.4903580059267262,0.8648603066365387,0.5341632133255357,0.13066036232256983,
 0.39002650238869685,0.5536443525692617,0.4756839601362257,0.7953133837697464,
 0.15839823929163221,0.44024832207104725,0.33781445584130376,
 0.2688740645217689,0.2372615374816741,0.12906356098189864,0.6025778165919916,
 0.4002154897981488,0.9880093614730759,0.4459443458624637,0.9191834292181777,
 0.19198400482880773,0.7812452457983664,0.056141661806926146,
 0.5202495576238739,0.6524986685358891,0.19746487560901838,
 0.033012833553136556,0.4995300273361428,0.3437042405575008,0.953935517967417,
 0.13836649550631175,0.5899490476547617,0.7302340150150519,
 0.24047302270220772,0.182642996345469,0.22275687727443494,0.116652755442056,
 0.7427365975093588,0.8942964013599776,0.9748536611435512,0.46550683459620223,
 0.45675977507248244,0.7935997518403185,0.22816885404074772,
 0.7317838032303973,0.7517332899510848,0.6138863391076156,0.755311919729986,
 0.2784559609647719,0.12407545372157058,0.6261184980985751,0.8587943879646017,
 0.5023745349912915,0.8757648614973071,0.6432719583575743,0.7779522982238982,
 0.05264900670345618,0.005604080129953171,0.06278455461511134,
 0.5174959828118657,0.3790498367896321,0.9521101902246687,0.9363452023980094,
 0.639770158525784,0.39037201325634296,0.1010871064495864,0.8333196329160739,
 0.6024498379064627,0.08304971085111645,0.33757939341331,0.9303103124821639,
 0.421488748032546,0.7758174843505327,0.5723043902728806,0.535488566483246,
 0.6081615154403297,0.4637746354015958,0.3736959895414783,0.6759674254208089,
 0.7650891340535494,0.2196845708278502,0.3777802035223562,0.04706639830525017,
 0.30751106641987935,0.7733759163059302,0.022429264777574898,
 0.35909151694901376,0.8665758154850172,0.5374591237463976,0.7216115249660457,
 0.9451688212490652,0.7892538921191888,0.5184032215624829,0.5678108132711516,
 0.4386868583173582,0.8101765233457818,0.729011760603994,0.04759466663768208,
 0.41931641041946943,0.9034431468319198,0.8562545971862223,0.7041768781581431,
 0.45209764140641706,0.38931149953616306,0.3900325508192881,
 0.06362264917135452,0.21506901458052774,0.11681003699116,0.7227341327639133,
 0.10547148000916196,0.21122914445144114,0.4256491145785142,
 0.38778549285347597,0.8857770063452339,0.83870060037842,0.7814995274302373,
 0.14386471055206917,0.368940570256299,0.6970082206481812,0.29673849468725844,
 0.6955481524616908,0.4382424517828334,0.13629138464062596,0.7555834024224714,
 0.6069177489102743,0.5650383997309592,0.2655349087689185,0.9466590756408203,
 0.455005368870818,0.4803273003376961,0.039395272192109365,0.6345665271920632,
 0.2685507051989815,0.8689377284593618,0.560737571219976,0.07199064407247846,
 0.3707126140235004,0.17297113651462592,0.5161166006737633,
 0.07926553298692751,0.7663641283300818,0.32828792871048595,
 0.5643413806107885,0.8443973409211613,0.6086863020731652,0.9383825468859037,
 0.7439367491762053,0.05136026193979637,0.6393491242478994,0.9753516375342763,
 0.3218914442075722,0.8200400804717681,0.802716077989873,0.5458808281992407,
 0.9533055191085794,0.14077362977425345,0.8446144722609195,0.7858776671066782,
 0.14056669728499982,0.33204924418153037,0.7832242118668001,
 0.5353172844326028,0.09618768353758933,0.8193400106308527,0.9087742364487366,
 0.38863969803240805,0.48991607543370685,0.7879467469692542,
 0.5675900308260514,0.43418058228385814,0.3460392763484126,0.788354925913487,
 0.15412995146325292,0.6331176176313825,0.5561843346328603,0.6183358528683864,
 0.9909440116889595,0.20404340580191827,0.7957358964527164,
 0.22669586187511648,0.6249390624816337,0.5527972501166176,0.6894019540421797,
 0.8667778037688122,0.8337129836966626,0.19377583258770004,0.669563779908817,
 0.12351335658415508,0.7588618642758302,0.9127251449513301,
 0.30212464149151064,0.29844187589596227,0.8174299598333349,
 0.5630069053209347,0.8606482411691734,0.9084519055252013,0.908267185523283,
 0.20921557473359154,0.6228024147273525,0.004316256843946209,
 0.6920652968861802,0.42209225143817597,0.5632803115711461,
 0.26849150450963066,0.7808516635148225,0.015874073003184774,
 0.746099690442107,0.13944453345480046,0.21536591108046232,0.8275112154876662,
 0.4531227093474295,0.3574936725423592,0.1867397689823368,0.3583766906805983,
 0.3092912768028818,0.8894076233480442,0.3160514437334315,0.2682792551811535,
 0.9402108933719249,0.3006339266745883,0.9120973451308946,
 0.018310500321796597,0.8072031405932613,0.6457085366860071,0.632078724907314,
 0.8725317819681264,0.5264616912752228,0.8367681546389056,0.9545475475935308,
 0.7046190267667298,0.19222684994043604,0.5336346264996943,0.7797373687748445,
 0.16339071423985863,0.0490844387935826,0.06031934034950548,
 0.9222864409213235,0.29730416426470485,0.06582126759288442,
 0.6542955822540012,0.8355737102588437,0.9447681795033073,0.7220285006234338,
 0.1081576568565985,0.1044219361415325,0.36773151111129787,0.9435625175204247,
 0.5451861317486801,0.5946506613826362,0.4229711068369849,0.7504959106545313,
 0.5233896846758206,0.5178574403798497,0.13176472185227495,0.2425043175958138,
 0.924966455430228,0.3122593176206605,0.6734825036505059,0.39128195660582943,
 0.010148152518295261,0.6612456886446813,0.14574048605329593,
 0.9590419163567276,0.6130958835924609,0.3120805575404949,0.8073424578164023,
 0.41770974366962277,0.5893033232981784,0.19922638766201317,
 0.8959590293545288,0.8007888106118454,0.10915948446117185,0.4130232156684871,
 0.9986683016995602,0.39524549479675875,0.6357052481296833,
 0.10700494508291303,0.06953835594138424,0.22589524565120378,
 0.6428786451531194,0.29967505667285776,0.603790611086493,0.7606039080190544,
 0.06737729548342342,0.4795632502031868,0.5798432564073497,0.6909003997143887,
 0.6142861768797512,0.7992281276728956,0.03988343666504979,
 0.31457067487944046,0.9709332901086564,0.9698147020302055,0.7575945471631033,
 0.08155842409200853,0.6415574879693018,0.7984255194942464,
 0.09820174839362394,0.16163524511736482,0.7523108465220337,
 0.20693579767347736,0.3905098068899432,0.1740506155057847,
 0.18700373772923007,0.5826123421604686,0.9447422267137408,0.7381227310630583,
 0.9294740649045665,0.5449056827652264,0.7022195936634469,0.5863283083971166,
 0.06115450852792226,0.1664070881634776,0.5788744963439146,0.7754276358315925,
 0.4986453844595259,0.2848130685171728,0.048254446484208735,
 0.1815298118655413,0.2541563180209626,0.009395239347275819,
 0.3665639071771938,0.45526329766895457,0.7853141663081046,0.7666463196531086,
 0.8148634008712702,0.590263072301951,0.22767216077043573,0.3343408027404232,
 0.8323656595711162,0.6526000976849902,0.5277614444039088,0.3080743747383131,
 0.466797772878758,0.35966152741206237,0.7044194107563904,0.24484271996855267,
 0.4820953262077472,0.2935157991720244,0.6195896368795462,0.7665973646845403,
 0.515612400710451,0.3453121194979468,0.18357877887072416,0.8841506752177829,
 0.7172339946611305,0.9451949372351898,0.952411405563528,0.8490972546206966,
 0.6942213738635572,0.7379055926627169,0.6744825804139123,0.2500695048116234,
 0.9946294051072735,0.2691364595389206,0.8721382088850449,0.0496751768989262,
 0.9696510896640737,0.4055298507338734,0.7846769694627517,0.26131436574864775,
 0.9543166838127638,0.7118171465270448,0.9665525850136535,0.47665805299230013,
 0.6366272644896328,0.2191261214383594,0.8454828124404875,0.40624312893659587,
 0.948748142457227,0.10709368045706336,0.47036191027487095,0.3225358478152077,
 0.2794601808333458,0.974783451906997,0.8895694176663848,0.9745280480516918,
 0.4411010552722927,0.8520179009465225,0.3580185084813262,0.4651980211120904,
 0.7446018907623331,0.9817549376058706,0.952493046147973,0.04353894252847301,
 0.6322477504324573,0.7215118525065862,0.9011979826534643,0.7199377099826081,
 0.6038114492404519,0.7180867016294377,0.6372247623968673,0.2106802764257193,
 0.7974692982817482,0.04115493755374955,0.01170704752916596,
 0.35619925946166076,0.053234561956286286,0.45022241517217165,
 0.033805982803938206,0.7908825882785014,0.9077279479234045,
 0.5777760979547049,0.26961579204035824,0.4984393056038774,0.8326575068658153,
 0.8348532056897976,0.24155875754719836,0.41456222948707644,
 0.23718221668542616,0.6404835201574561,0.48544483767529967,
 0.18714994061884827,0.45111766325334834,0.8062085866239598,
 0.10260883811993693,0.7853819170224849,0.2837042394927346,0.0431148202244499,
 0.4029715100855764,0.37929320503212705,0.2700113805651617,
 0.17586244132583428,0.05125875236107691,0.37533558848278936,
 0.8460766591471349,0.5966629109332059,0.3484819149806855,0.11256278372117923,
 0.36733168289977713,0.7573147944309813,0.7306732210030796,
 0.06406001654188831,0.37113067700746394,0.035255765517931414,
 0.6089390497618933,0.873236959353066,0.5955585196574109,0.8030818326896918,
 0.6742267791285522,0.4976608552091306,0.854355614856837,0.08497004132246189,
 0.7520926625827102,0.800698450673634,0.7150643042730602,0.6476233289226377,
 0.7645745245886371,0.2732769280983003,0.735020738454671,0.128803164160453,
 0.7641962532378812,0.8202711190738966,0.5896860975985742,0.6258468637352743,
 0.16531237025242673,0.8231309875951152,0.9360475826276138,0.5045349203511157,
 0.7559613491537139,0.2564352831339045,0.5833600500557189,0.6134687321941147,
 0.6220397289409452,0.45160164516407997,0.3335166793282133,
 0.07569412814190624,0.23794191115057306,0.2438594746154863,
 0.8772602827159547,0.5361824717893313,0.29519998965974115,
 0.47792483361135196,0.9397241526140869,0.14159476077406974,
 0.23317031012625222,0.3563153328719253,0.11337995775045484,
 0.7887312339754063,0.6551821433905685,0.4209724283447627,0.12227252978660008,
 0.9926826447899965,0.2706640355146033,0.7089595726189513,0.4478806466985755,
 0.985092859589693,0.4684392327085276,0.09275043130916838,0.3344402888811182,
 0.8904419300820231,0.47270597018939664,0.5414060250840783,
 0.47356465221789945,0.19841645432488708,0.7189294881228052,
 0.6917458653962641,0.3612515582887593,0.5298806956253815,0.5825081268463818,
 0.0469174240862924,0.7116823577537675,0.9293285102251754,0.22032140736349237,
 0.6697163831770945,0.3336156193813926,0.006131437831156106,
 0.4456873050597908,0.7444826534542381,0.627603617878756,0.45415546889612024,
 0.29124250692977427,0.07800125483901499,0.5573501984196968,
 0.7787594334865562,0.3201544830825582,0.35914765534861637,0.206188186981473,
 0.7980183804926432,0.872534738788652,0.14761670108058422,0.7042232433696934,
 0.5299800913233929,0.34617530834738597,0.31847754643888226,
 0.025693000775810937,0.829540486978474,0.6324922861217542,0.5162874319123678,
 0.6345896102167794,0.1793146736768867,0.5531113553986002,0.9403293018271905,
 0.138192887057613,0.9106491959750793,0.26053325687856166,0.8684719280732827,
 0.28556776278976526,0.3147596647982397,0.663272424689658,0.12490640565686939,
 0.6860518271784456,0.91286878631678,0.9367807458984799,0.6763937285059057,
 0.6973872715923035,0.9330885053141926,0.09986338028789588,0.61347313388861,
 0.956894397008687,0.15836665433330233,0.5299249995724074,0.6797311059626439,
 0.37303445563436366,0.6045595117472408,0.9547288673164938,0.0614647313632819,
 0.38286349538538844,0.19192396219990426,0.660661250923874,
 0.32227821216627583,0.1906523981426771,0.23973789478118224,
 0.8169802753602803,0.5296432907695277,0.7687979994185371,0.8581429286656801,
 0.31643193073529696,0.03072860315275394,0.8709588447405785,
 0.3170171966771944,0.7882555444747606,0.1531569964717603,0.08496197084617207,
 0.7540399771807523,0.5194659902031608,0.5303525839035511,0.2872097979239495,
 0.4930300929770668,0.007147101076923179,0.8885758239903864,
 0.13232044470767645,0.5365507133411814,0.32804228093261545,0.337147242028926,
 0.9075491050398548,0.684125221670954,0.5841776510005885,0.3032267648287763,
 0.33628436026335795,0.03537972896361774,1.1506285854256948e-4,
 0.14229808939085875,0.19199592601158955,0.5351009118164637,
 0.48611586565513687,0.6416063495481442,0.9647887964062782,0.7666725893333098,
 0.9354790571093354,0.8704599057814819,0.27262820020432654,
 0.07380816815558922,0.3581468142733355,0.05194903593038758,
 0.12998641078587292,0.6446678186662386,0.4267773608698756,0.7098453978306056,
 0.10103524594864721,0.6376629970493525,0.5315258205245554,
 0.009736418103691413,0.36973654436003023,0.7987726479188922,
 0.25024001468875645,0.5991889396897014,0.28082222451669014,
 0.4185360996645472,0.3428635845087793,0.8405615971472016,0.35471670908924335,
 0.5664008375088967,0.8629938378810763,0.4890122480796688,0.5074305301670844,
 0.7158446628171871,0.23468899079992622,0.30990938261218526,
 0.12261311475585768,0.5342505353277658,0.46578682684272865,
 0.5339396136150146,0.8490668428629643,0.26825854787644343,0.7912609541577852,
 0.01077924134635011,0.5609416773738304,0.051424186107387526,
 0.9518317069928043,0.30920014379204874,0.4057029221574795,
 0.022368334751367458,0.15162091647771048,0.4140649871963199,
 0.4724423472086897,0.6044668711349998,0.7405650972797485,0.17501408214288694,
 0.4657888393955749,0.20693090988755802,0.5232870639237873,
 0.27667113278312905,0.2562751267974739,0.28121322895885714,
 0.42756629629193266,0.7189503668424053,0.7323718220462148,0.6810012438618596,
 0.6186306421577245,0.2877115456209526,0.6638894804306099,0.8113237991646428,
 0.5703040049556038,0.2523275393081631,0.9421538349610526,0.06081193019882303,
 0.027252097241758144,0.9234674372092792,0.7765464203854021,
 0.8604665257723014,0.36048163638269437,0.7586196678178156,0.5215785141712472,
 0.8394064346354223,0.10787991989898815,0.33812718616131443,
 0.0964969893166876,0.2624982774504865,0.9416989379295418,0.21325613484857642,
 0.9717234944218919,0.12562781625738406,0.2373922596350867,
 0.25715683451538773,0.2922869849104608,0.636880716094241,0.3479720522271663,
 0.04847020617757036,0.7497923595902052,0.13781267245064388,
 0.9965447434438459,0.6192885303151845,0.46357375821951186,0.7929710518337354,
 0.8419217143955514,0.19166048302096372,0.15348198133768043,
 0.017511440500599162,0.9363843828901304,0.2545966802635443,
 0.11085524267760649,0.6843579148856723,0.8067130724110982,0.275542093900367,
 0.9595345289386166,0.3435602332337838,0.9253913538524127,0.6721718110313988,
 0.03945963195078195,0.8159554879565709,0.10703698920586102,
 0.6160743293210126,0.23610406553721164,0.030971968310584685,
 0.8656956634999913,0.7716589417147195,0.7941509036307263,0.17899008547506678,
 0.5777393354441537,0.4178458608403822,0.4812490822801637,0.9998617957436562,
 0.1025278991459535,0.5211184698330338,0.33684341750328706,0.6513010426914002,
 0.9406319256371208,0.1702604137446786,0.3346181208517981,0.3449008443639856,
 0.026839175015025307,0.13522210497308584,0.38887018587899136,
 0.46034112372887503,0.8663507150985597,0.16660835765511806,
 0.8587722616449385,0.8238481129576105,0.26095635156474495,0.2857539924265369,
 0.7711426270228429,0.848841519408312,0.9265359917173226,0.7811323758826649,
 0.02257564635212228,0.43207018123171104,0.17566409885142542,
 0.8073289407910275,0.7555984236558356,0.19381468906768273,
 0.06566968903456172,0.9652193166488539,0.7160320354972118,0.4629964959423415,
 0.623936198849638,0.8966209546297693,0.7470247096514129,0.3969281374816299,
 0.9389562294563307,0.32833292468754505,0.21455269924201892,
 0.6462093679970587,0.40809363809696164,0.11073007844232774,0.891955398078923,
 0.12637579705055546,0.7324980017836334,0.8676916952071392,0.2885542342767786,
 0.5577207099479893,0.26008908468493663,0.4046170503145464,0.2495637814639341,
 0.7115801307812484,0.4740692393032051,0.5981225207697526,0.5913276003908419,
 0.9896980666966247,0.8961680604314165,0.2594158633545529,0.17468758511208682,
 0.6672220889038623,0.3675551151421559,0.342223397089914,0.3753964633516663,
 0.9102162867946451,0.34304809586457374,0.654025787643632,0.4073669127376851,
 0.47295658151713127,0.8406180864799387,0.9107127155999031,
 0.19917727358559323,0.5164357261946784,0.7808736185817475,
 0.47925954264747517,0.2662397484003163,0.38826381421109035,
 0.4427167757025662,0.8458121391075866,0.27738532158838114,
 0.45857682609587735,0.3378590803475977,0.03453611349867125,
 0.8668321082255196,0.71842318776626,0.1895421035508147,0.7156527678091529,
 0.15681134605593117,0.7323668977064466,0.8354912002524915,0.686319177161161,
 0.4153331775041408,0.8955622086796217,0.9515719988968614,0.7530863996388775,
 0.48464117555925923,0.4391315293228393,0.5655583276042031,
 0.27025076529242176,0.3944408935979177,0.12145721864032333,
 0.16249170115392308,0.3070173779655121,0.08302836645675382,
 0.2580287636226102,0.6682204828407543,0.5004085107718195,0.6021467050005455,
 0.8741915614576634,0.9231457913773473,0.889583751702309,0.6428432649290017,
 0.03868396927038653,0.6482013914650999,0.39449735901794714,
 0.7164287255263028,0.3000432783355962,0.9721201628097322,0.13395881829608258,
 0.6553950424932873,0.027689751982183264,0.0049673802111718235,
 0.02828360017547671,0.4161677229823908,0.7108456793019734,
 0.49115770810152004,0.9486271828330198,0.7525395794791727,0.7448031898889697,
 0.8134927148100869,0.781550742420603,0.20856283632995476,0.3298306722002774,
 0.08197351340472325,0.9254340753178487,0.4141225303059519,0.888437937200453,
 0.42932548155427186,0.19464373919998568,0.48904847254554984,
 0.39267826428325847,0.9815388935903808,0.9349785931056285,
 0.07786548334065746,0.48076230734685044,0.48637646313809013,
 0.2653501766718368,0.8969218350121345,0.3401039354020222,0.7691623104760901,
 0.5627127762566962,0.6094487674099702,0.5628367964699363,0.24578231862921196,
 0.36318003990518677,0.9827255651043519,0.4109981098340024,0.5570152103267945,
 0.8091435828129091,0.08917991210877207,0.6409278335909693,0.9814836094167032,
 0.1982359127973733,0.5136255746247972,0.5676608562978426,0.2464795822745709,
 0.03510480025274265,0.8109704050927322,0.12961644011163242,
 0.14808642219825652,0.16502280749789866,0.841667033367462,0.7900518187500779,
 0.4696444513567414,0.1181352533555724,0.7731315858672517,0.35444126400537224,
 0.15390164364963932,0.19524113557079492,0.960454080052782,
 0.06129428211540078,0.45849043378610854,0.20408688041074308,
 0.29044012564523525,0.26341259343299406,0.41777557822436506,0.70750071807643,
 0.788064227918712,0.6840353942592367,0.9118332628999015,0.396079055742126,
 0.3435228837759414,0.3719622772774339,0.7613572637564165,0.07428068373317553,
 0.04435406176170564,0.5870321776309044,0.4604474395120626,0.9070742214740778,
 0.11247339231887221,0.6350689179417869,0.20684119609991192,
 0.017194834768817224,0.3072392813662609,0.2195498823250619,
 0.1445486543103044,0.3666797005019219,0.3292046227097136,0.14815569072362222,
 0.4811401025568651,0.3845192585588332,0.43156539388904125,
 0.23196936278669977,0.6338564224105093,0.09072033802281643,
 0.9733526386459017,0.45711391286846514,0.8885693192439286,0.9052403541121146,
 0.29265782886015346,0.5597786417391123,0.04291586592673169,
 0.34765427966737783,0.44702320535638573,0.6150357998970093,
 0.3432017196623315,0.758155781552202,0.036650262769493125,0.9006525971940518,
 0.1596846271613639,0.7657927806977712,0.3896188471357206,0.2758844389603067,
 0.95291314661272,0.16701209204664413,0.726009701397362,0.2594591074883501,
 0.16656139638985112,0.8095119636351789,0.307478980172024,0.6163260473463749,
 0.2861548976452032,0.10143883868438575,0.039846638120429545,
 0.8104271536636831,0.5007481073938215,0.6192184283396919,0.796514345989285,
 0.973650042459088,0.918341353746446,0.3484305654879948,0.9407170143360304,
 0.30007062550809704,0.23810369626714367,0.10707331994924563,
 0.9134778378290943,0.45809573496081746,0.5966260684159779,
 0.45350822302582117,0.6125781252538939,0.3999693112597671,0.5498561749035951,
 0.7040326169316503,0.5556394251724895,0.3827005688741114,0.8953364979621484,
 0.7903508568711342,0.30558456277995893,0.29774517539465106,
 0.47227727162215094,0.05696267208852479,0.15112004622085262,
 0.12593479920694517,0.3706349042735466,0.5566421413681343,
 0.0017930787330489029,0.9076320883773397,0.04659608781315394,
 0.6285747504669684,0.0532592091319225,0.9001715940949051,0.6840451907673736,
 0.8713997686238364,0.052649001728369504,0.28238342475756484,
 0.9685697810631855,0.20292615199137176,0.2693453207518307,
 0.07197449164965819,0.7157403138081649,0.563680604758975,0.7749896592437153,
 0.8307637243282708,0.6068007947603298,0.5127018581514656,0.23038107680772757,
 0.9853051783036147,0.580911652853362,0.9333984862039386,0.8392529583421814,
 0.3710016552106453,0.26817301333514587,0.1044781324317745,0.6113412673097642,
 0.3984214480852727,0.556250633675003,0.00374984454941929,0.6806240760972717,
 0.7908918358404449,0.7568111668112543,0.4178835707079325,0.9471113534410608,
 0.5870463782917951,0.011531983089006714,0.44745264013808317,
 0.48175429016541926,0.5682170423878304,0.5858143069512454,
 0.48747669509166325,0.6872904193460265,0.6950172008138972,0.6303115248141997,
 0.8087097957960377,0.8805150816765607,0.2603819321211024,0.7968154429455833,
 0.35456007988573757,0.2721170620205222,0.9521111394493216,
 0.20891287668971215,0.2326446712869331,0.18132615663548257,
 0.15551563317000316,0.0591116071751876,0.3525524444129735,0.5112792609260104,
 0.6340274608272262,0.31282516785188275,0.13082369902063795,
 0.7367196723005149,0.01015317896805723,0.6339026268452257,0.5840948788177484,
 0.10663314491377607,0.19383744070911135,0.1124064506516066,
 0.019131180889857102,0.8004232059653463,0.9182517321239625,
 0.04718523718622114,0.25497336298834405,0.7710736823667085,
 0.9647981081468617,0.9684828766446134,0.7988126918280409,0.8553279090882207,
 0.3947699763737894,0.5417382257045957,0.5816762987803359,0.571790656469328,
 0.4121931572322506,0.9985667419981341,0.23152397606689812,0.8475728962055192,
 0.8080432218202158,0.6548567716286371,0.47100720453650025,0.563243918397966,
 0.907888793617439,0.36570458245597637,0.4936089179692019,0.292304146650362,
 0.9104118781375887,0.7409519032635881,0.1424623258107357,0.37834458928507053,
 0.4334800140325349,0.6435019791187259,0.9881677288731445,0.5707198576377552,
 0.7838550713514878,0.7146281903979121,0.8232932595558302,0.9468590926255552,
 0.2677636206468442,0.8399156724747285,0.7867497669092983,0.4818585141270626,
 0.8690027741601416,0.5261593013641472,0.1489017568344736,0.27739642816366283,
 0.19780938303309392,0.25141080142495065,0.4679239167683079,
 0.5035635896224534,0.36965384788867783,0.19485497146870234,
 0.6770443433339143,0.5008887220850555,0.46787822661473455,
 0.07882768632512072,0.72144054852257,0.7825055232617965,0.5899525965847214,
 0.009442759761844943,0.9757276532685251,0.4596441134860625,
 0.06818229372937246,0.5397548965285007,0.9707347984253931,
 0.39809092935441504,0.4210735095727439,0.5631260975504717,0.8567153525267344,
 0.8561076975881816,0.18142571877039781,0.1700219050548094,0.3231001628950907,
 0.07637915454187394,0.7077481543577128,0.21637597860128321,
 0.26618854079942855,0.21374439640730525,0.5376361187062432,
 0.30345850477908454,0.1567991178268756,0.39379834339777364,
 0.029739182513805806,0.2243910456923407,0.4015493551823731,
 0.7813588225259473,0.15909592992290755,0.5640020590075133,
 0.28070774293156087,0.916682823347019,0.5519905052138226,0.40842942601529186,
 0.8463596512880739,0.2889098747987554,0.3454000167257867,0.8658666121726382,
 0.0024119163143153655,0.045590587100684976,0.6204871372843798,
 0.5064685116747554,0.9221721840833081,0.20926126242198062,0.5492320564099089,
 0.09926188524199886,0.1971380545461483,0.047208300047083096,
 0.6432666037127619,0.814150627145052,0.5499651503209465,0.3452058151009578,
 0.798013457558668,0.6867789370635959,0.20152612206725273,0.6042722186530699,
 0.17843682889933232,0.9553477391021105,0.1808354879273183,0.7978521857883956,
 0.9618687739125253,0.08342198113164367,0.9729706251997405,
 0.006322754565118327,0.7478416885418181,0.9786652360156696,
 0.01976746357164505,0.8582133383846605,0.40465553307024005,
 0.11508870887649159,0.7490860932891396,0.808726015885411,0.8888336016641667,
 0.7024930121751906,0.9290503534365983,0.5883950184280791,0.6769035505221342,
 0.5946926393019946,0.6241363498494679,0.7739864978634109,0.9301634062780828,
 0.47690393543750453,0.1684591625076377,0.5322041708431764,0.8326853879192574,
 0.06804185369414684,0.6307878168181112,0.37171367392184684,
 0.5837792802141641,0.8908239504420085,0.8290519288087483,0.7851189665860299,
 0.4548078193165227,0.6011850137654112,0.19198288565892208,0.4019729907217444,
 0.329014205666216,0.9120672876480416,0.9783806737222915,0.37677477508498813,
 0.8623622731946683,0.26950674455602996,0.35277195523774163,
 0.21684671419340307,0.9264768397892716,0.5652924867823959,
 0.33238646897567536,0.708679303432886,0.1662264904324101,0.7140840840584444,
 0.10789075909329615,0.1864553025101674,0.7986101889072366,0.7197364503185574,
 0.14544209429205002,0.9008148834010011,0.6339477992514693,
 0.34468616809701647,0.24717285840174907,0.16308438439667994,
 0.8286950723173988,0.9722553574748064,0.7441744946768316,0.3538893155775098,
 0.22711616511963195,0.1417689748935801,0.5098008231232338,
 0.43523753319855885,0.052354962882982026,0.03778312876892376,
 0.0450965317952291,0.4400829942541381,0.2117223452875484,0.198660246836861,
 0.7181357542615091,0.09764597033884004,0.15759088961360013,
 0.8395235146954081,0.07169483682188638,0.8623088081333559,
 0.21020105490285934,0.5550118315651104,0.1475451392576531,0.5237831013363368,
 0.14580131353376868,0.9705300447955026,0.16762707135060184,
 0.8809756112349494,0.976058085619286,0.2999603099873067,0.39745222113029677,
 0.9264074756001475,0.04385893207403502,0.8431888293144895,
 0.17789099038829992,0.44165364521589723,0.6763772018380575,
 0.9499196568985262,0.3027684278317121,0.23813183622038192,0.9153788085971091,
 0.13388542457861796,0.5575690548349167,0.04469056694609774,
 0.08926322611770698,0.37653941239264443,0.916672346703727,0.6285661438692383,
 0.4234229082896467,0.11678475852567793,0.6098973066647304,
 0.009262458687034014,0.2015046371218655,0.5140387514736114,
 0.7835779166958956,0.7174470277322652,0.017189448600031576,
 0.02778168182443308,0.3653863980595775,0.8675475777093048,
 0.46703804249034064,0.985599505179119,0.49541591696604304,0.9122836594924227,
 0.42066855648989465,0.6208952943089137,0.6369768794892922,0.4507035228735592,
 0.4089860158496327,0.8545779057193713,0.02634454308620815,
 0.17938137663574671,0.5965245831411855,0.2714972129371772,0.8451605425180906,
 0.10798325584437651,0.8131533946128094,0.5059274018936022,
 0.06441385943880351,0.9818617807652482,0.35257606319239565,
 0.38349719243221936,0.5927847454120446,0.25547482929924636,
 0.31084375885265403,0.7307980690637003,0.4144918732207261,
 0.47849417494095037,0.9645138532074604,0.17709163409644424,
 0.9938166826249452,0.3099268319611751,0.06160471882481189,0.5346457334915081,
 0.43567314637950205,0.39549848050978964,0.10708532670854887,
 0.45923898903561655,0.2679377226494236,0.4949089721582115,0.6890601841537489,
 0.955059018615105,0.2671677199536633,0.1924418469392356,0.8067277367843637,
 0.8899222235729055,0.16927480651167515,0.2621332473904552,
 0.19900393971232877,0.7276029449540288,0.14798933550556215,0.296997772270825,
 0.5038997907610622,0.5486018950712649,0.6305642668382236,0.9688090064331845,
 0.37328609566611115,0.46892079161047295,0.9252878126267403,
 0.6127390712938753,0.6344208093335721,0.5288190633700328,0.8212371830435647,
 0.8004200350983839,0.3312436572865942,0.2422942788182999,0.2558523616745908,
 0.7216932014162907,0.06256065936133115,0.13124064052251605,0.333211806161437,
 0.023629937816028068,0.22701702042171457,0.022197163980181123,
 0.1884803509051719,0.9334793999098645,0.7173540357210882,0.31649331935506064,
 0.3787123222609834,0.9379846130630983,0.11861930649232744,0.3414856373376305,
 0.8491930482751826,0.35024344764477466,0.8307836691799575,0.9726858605671884,
 0.8390820649437928,0.1670373794069674,0.32511324648986206,0.6014276754546094,
 0.5028709297498777,0.5512084893001977,0.853936927932504,0.1845259921989072,
 0.12792469434720877,0.22683056632110632,0.42045609647062576,
 0.9190957958173895,0.9807171430589875,0.5626859398664886,0.8501034150378847,
 0.023839765032763416,0.8568105015295621,0.13808949412168736,
 0.3863119882477627,0.996960057218664,0.05947810671951914,0.5909205143932579,
 0.038627544424178484,0.13749050976662092,0.45020457620890386,
 0.5751318935544312,0.9272326813736725,0.02553752657429753,0.9531118458349348,
 0.7375970702709753,0.5020779241778184,0.3409138690387503,
 0.027124099109852695,0.2121382435814647,0.8406819460456015,
 0.7625993973718044,0.9592298734252455,0.27184311293371854,
 0.23953799904389117,0.11926262930239084,0.7523059871366823,
 0.06692192541574205,0.35597281444956186,0.8258197746101257,
 0.8867373865930132,0.17187625025955888,0.6855587250489715,0.8737702872514806,
 0.1772056109392781,0.14425005291963,0.02768189654616393,0.6116675169203436,
 0.6386832176967817,0.48233517104658064,0.5370190422712204,
 0.09190545129327965,0.973983695184619,0.40693985863312565,0.7239500541056144,
 0.2587843896385392,0.3823681110832958,0.40161707172953476,0.392699684778559,
 0.34103520081418814,0.504431280247649,0.4461997089926738,0.00985918508555661,
 0.8060465438451733,0.695501195258558,0.7672549950642967,0.885704452224983,
 0.19423722794316545,0.36166577442252046,0.18963579380373408,
 0.605040088131635,0.908579721900261,0.9211441129824035,0.5329453109356556,
 0.9407326306601148,0.7210249530259932,0.4940953269355327,0.9253003408535037,
 0.5396974499264579,0.23307688231906698,0.7825236022833549,
 0.051447977322898075,0.9553412301992039,0.08803001497832452,
 0.475032277440131,0.2815101252367349,0.6322744190930132,0.6219194345136384,
 0.6462683883444363,0.461133899563838,0.6114994000137135,0.24928368097902143,
 0.9475592677548268,0.4272799321714722,0.46099332233491574,
 0.016532328615689984,0.1318624044853799,0.7752904032752327,0.908135177137297,
 0.5777451731714837,0.948330947034596,0.36748538914156514,0.5334103111074272,
 0.29002664337736916,0.21376090549259064,0.68301355760449,0.20141878428804472,
 0.3748123167411639,0.7791257684638907,0.8533950832154596,0.408505973060002,
 0.7273651506822543,0.9319720517305232,0.2882271091998245,0.28399896127742685,
 0.33531372592108155,0.34437688267499145,0.6737176411974891,
 0.11546108418201684,0.7396825402182531,0.4103337341691681,0.7419944275268503,
 0.9162829220645964,0.6834231931578559,0.15050264783985,0.8926711212564545,
 0.041092181303721054,0.9868091910860199,0.201612729126167,0.4546567313116444,
 0.26987374836595934,0.9641145557171228,0.4494179909516416,0.9885990134319123,
 0.9953975695643225,0.9319714004869092,0.4351024760887984,0.8768011077863166,
 0.8944421916279836,0.7752201680463037,0.3174718584438255,0.6589217083451174,
 0.564379625174167,0.3194949758912182,0.4620984062382998,0.06134948392853279,
 0.05821604749617382,0.4979731953244658,0.45461641450853585,0.454570731106679,
 0.8434982497595737,0.2208105299059947,0.26694374746883454,
 0.052757331160212084,0.15864284444313437,0.23565509931079198,
 0.3665099016445218,0.30542078402601414,0.8919424295135929,0.7124013545677494,
 0.46525318827257023,0.9196817791849908,0.8641075363822959,0.8006691653853235,
 0.9006922360584834,0.6704347461641437,0.027112692602829558,
 0.6987923569549773,0.9114852685030578,0.16079251715913,0.24520873310018587,
 0.6262440931955541,0.3471252698912408,0.5425001860614976,0.18337659355941494,
 0.30055513919338606,0.6711874742278944,0.339332620976475,
 0.033129809050450265,0.5171076406772208,0.873599142549084,
 0.047092528409326095,0.1364925104828142,0.6418560163553829,
 0.09165958212629066,0.7673000540677649,0.26255011858103594,
 0.03590867475683024,0.20143215758890465,0.30150089071938824,
 0.7774528340250243,0.14020282041966603,0.5054627914012906,0.8677194039147756,
 0.2500678592118257,0.5347622564740264,0.3860984327755359,0.0480561277412227,
 0.15622157061668973,0.754585149234165,0.6663310126209504,0.4710585893678063,
 0.23008785106699292,0.4091059627833823,0.3415271099782361,0.7922611898833797,
 0.011650263399355043,0.43057546734542507,0.7022796746814732,
 0.7584595575580441,0.9575904942458564,0.4854707881356648,0.6046276010704414,
 0.59753918097859,0.17185852380609123,0.5588484571073644,0.1929495712213386,
 0.9436063221622913,0.5871363210947591,0.6756447166027826,0.7409162401753139,
 0.562468958261084,0.6577032872859984,0.5177418271407432,0.5923555411157628,
 0.7087756306875984,0.5666865320889343,0.1461853481093336,0.11747294994812274,
 0.6714109493111183,0.313016093972073,0.6873731186847246,0.4604909717798368,
 0.8556381766714687,0.7276680074304167,0.3182349326432805,0.5309271272205215,
 0.9330702355560418,0.11315163421971086,0.798805421020076,0.9283769463489302,
 0.3951255071414651,0.75019924790253,0.20330786602374284,0.9907366267848408,
 0.6354595353254936,0.012918432356846865,0.858717833632005,0.6760465498809364,
 0.016753249906937162,0.6033930011932995,0.6753270220395857,
 0.2746169487576965,0.2374105839015227,0.5321019795071474,0.871896136246503,
 0.5423498864660106,0.7493362794351787,0.33066780989371947,0.8711063090220356,
 0.01195823773990179,0.6438888172925292,0.17797951624629393,
 0.49350159504506186,0.4882238696030816,0.20722282466633746,
 0.1156259296051898,0.03222740206734798,0.9079500243611103,0.7053333195768897,
 0.07059912031990379,0.15422272212347143,0.4150000330034296,
 0.9234969441243785,0.967650241160545,0.25370342037424676,0.6424309290721848,
 0.3287571998064365,0.74906971562123,0.5838294482509742,0.08331202998781928,
 0.9712720599507999,0.20682059967464883,0.39525025039617456,
 0.39205661218990984,0.17548644758719956,0.9941480612827184,
 0.8220760399870018,0.8953297085040214,0.19310103284993596,0.9285869899165068,
 0.48334503276233887,0.4765056044746362,0.022647042334764333,
 0.6062504204047224,0.8386675280153291,0.78610305737664,0.3547006640327143,
 0.1360395860551381,0.6089128200274447,0.7371900064773657,0.4570179445442748,
 0.40708580265661676,0.7505198568648123,0.49583640092437353,
 0.5351181928865656,0.3934626543655574,0.9394727358689026,0.4494772649627603,
 0.25713729444744593,0.13942110906667393,0.045388935501438965,
 0.39173498903769777,0.43489352496331857,0.9335447994940372,
 0.5587912972185571,0.48359923739415167,0.3010261177403988,
 0.22529992649392971,0.28440472956862106,0.35706033656053526,
 0.7864248754010851,0.8284328395748545,0.10355901613927354,0.5399335454888199,
 0.7626137305203212,0.32809882191204753,0.02196057085608316,
 0.3562358837030192,0.5200751785651067,0.12856792053915833,0.2047695755567711,
 0.6108640915614092,0.9284589385381724,0.5953788849027154,0.2982789956572144,
 0.4134563344062565,0.6971566405982316,0.4117098424468465,0.13054470565304044,
 0.22485582900231194,0.16065355132184334,0.384889232260776,
 0.37092980353997507,0.22737073343195258,0.4107570314524053,0.915050625773251,
 0.6430462205981403,0.7799948289489853,0.1906111309655083,0.7562417833699419,
 0.20793262961683112,0.5893915203232523,0.750183813191894,0.5456922012705409,
 0.27925458923922863,0.018997838953801582,0.8879787128186439,
 0.6968008734329457,0.19199499245261198,0.5773187862773231,
 0.41332699711027654,0.7121215979924961,0.45060627889845384,
 0.36268567537204877,0.2810042334815618,0.4479871600276858,0.532797840596602,
 0.1482854076681276,0.008849834766627707,0.6152943935955164,
 0.3451001495348025,0.4168482823323968,0.27593744165372236,0.6816795185012802,
 0.7844616397656524,0.05119442293332599,0.9964793706727455,0.4996058312605831,
 0.14655586347166483,0.9358500451823639,0.6845438530663145,0.4241744503808208,
 0.80314642145832,0.22659402968585163,0.8575331345224824,0.46306614602875273,
 0.08934259002332257,0.07428505213483527,0.0015298879113658792,
 0.8717692482323219,0.21430836397155517,0.6443974678840683,0.8977880462580067,
 0.2901263544898909,0.18284593528757087,0.660123638287523,0.4851905648634056,
 0.8819945144046395,0.8631006143775298,0.23596912427350758,0.1440032328130776,
 0.9673376303427501,0.041314017780058276,0.5860783577719573,
 0.6693364098078745,0.5052895440245396,0.27009281518804074,
 0.15180363209895043,0.6215147779371855,0.4259135848174107,
 0.01741219021957019,0.8237495814547674,0.6951151804505336,0.4203481652167338,
 0.674189691775701,0.14404947182999794,0.7874723545465276,0.5920974030768243,
 0.49599475860876674,0.6459331650621148,0.29610746240765273,
 0.8056778967470656,0.4332268617830629,0.14460280062522934,0.5186371967641633,
 0.5480868629389883,0.2021730691648489,0.9719246329941295,0.24684097173552044,
 0.6211680343349135,0.08265082842984939,0.9150062935062683,0.2834968531398583,
 0.7399358910189223,0.17471001973371203,0.12409619899380753,
 0.5765990236805378,0.889928025634352,0.45277110487371186,0.34118790018902223,
 0.34431891216647315,0.3342380320673545,0.5133935515920345,
 0.10576147019540372,0.19537453526868664,0.06414460113071829,
 0.01686475228712503,0.284289402564433,0.8781877757802753,0.21324850165529563,
 0.07985150952087638,0.013472436454215586,0.4169791729000173,
 0.9402938691375532,0.14574015015153174,0.4323472627013849,
 0.46415835632315516,0.4979233156469619,0.16346712659751783,
 0.9413193105202097,0.012118440381945872,0.6248966070594972,
 0.8096395420555917,0.17826652503826568,0.9645089909556264,0.9427398156983776,
 0.7284169535368774,0.22128452554637001,0.6802005130139955,
 0.04659473202115749,0.4043273012198829,0.09872259495971314,
 0.7451089675949887,0.9020951404761405,0.5550284359987568,0.08551290925331789,
 0.6650966232017077,0.04286972764307373,0.9048928021395949,0.1723396103139644,
 0.7876508587550655,0.01491093123438636,0.9522138440574839,0.30008369100179,
 0.29287283075461257,0.2803291544169946,0.17526017944665284,
 0.2774360729315434,0.5886486450205656,0.04101232584803327,
 0.45521064358297414,0.4102328246498472,0.7061376149778837,0.1758879137655016,
 0.12538581917725244,0.9645419860630549,0.5761466371055229,0.9348151562755379,
 0.3342239871253313,0.4411803996726089,0.97792854850439,0.28753332172471113,
 0.7650727373017032,0.4003357587415759,0.7515833566281834,0.37608516378093804,
 0.41729023851533864,0.6569671056794487,0.8006561569914247,0.8972875142665004,
 0.8733268970924856,0.9271664541863622,0.7435346131837064,0.2943862972882294,
 0.9966404879552954,0.8592496576123982,0.8019787087883383,0.9338776888653398,
 0.013700951821267737,0.3458692733476063,0.8154137776380829,
 0.1810307201576724,0.9346691710391349,0.6907231434641578,0.2942162206003517,
 0.8718847130884224,0.774781952521909,0.967939367680304,0.9217483844351697,
 0.5628218788949768,0.5573250223606447,0.4182925817146199,0.601541202581611,
 0.3156471909412253,0.4719242183838799,0.26389462237759664,0.7854872258928455,
 0.34012401395882264,0.5759306586446682,0.00671910346999216,
 0.20663803333783004,0.2607127309885362,0.659190000292466,0.03448327445280386,
 0.5851276188427246,0.6305146295951782,0.09536949103602865,0.9280389959469862,
 0.0933541075529144,0.5762882509205265,0.32570206125975143,0.9541055370422477,
 0.08264924534772,0.5926848207084755,0.42102204394311415,0.8282594631296767,
 0.3544017164968777,0.6604926660657733,0.9275130103895755,0.17970368238757362,
 0.7971094639918198,0.7814806178986484,0.08460426185453906,0.2277430252992486,
 0.5982563860797064,0.04176895455791563,0.19064835933308732,
 0.4482950385066905,0.08662805403826301,0.09220087453013548,0.574877388933678,
 0.6343873052752342,0.9719849642119396,0.8033129632680286,0.41750047885332675,
 0.28119231355478547,0.8503374752430974,0.6293980583903892,0.5680778220398356,
 0.703157109449834,0.6848746811155025,0.3298271818186581,0.5302537343647782,
 0.7954943927363473,0.16804321074253625,0.04907461005431901,
 0.10870722176789183,0.3901529170855933,0.14203838755904197,
 0.6452352543241224,0.3461280193970745,0.6733927996364842,0.560884532965657,
 0.27014131307667855,0.3946489336447261,0.7865549396672646,
 0.30957730643985903,0.4621182107221322,0.025653272791864357,
 0.3546102634182532,0.7429050893354526,0.2782931791289396,0.7143270889536647,
 0.8010971003329427,0.06474511422586704,0.10317042491276573,
 0.31299178187749443,0.8818824903372322,0.29995555455135814,0.348681612293838,
 0.5570829457629825,0.36634703550793113,0.4864078033795002,
 0.20247544581544785,0.2526495444817842,0.4827917457460529,0.7868364879820102,
 0.45557934956663004,0.24915922701408977,0.35187161618517726,
 0.7119035619819011,0.5824263329890534,0.10860879753706132,0.7222452283009454,
 0.6317526882834716,0.534744302075554,0.9834096023900564,0.5415558092622119,
 0.9877971884399829,0.15127703078090793,0.5136080953255397,0.7626369300061354,
 0.4014571159472635,0.4357369483370972,0.10296994689332262,0.8014692376016077,
 0.9385978618774918,0.16946005708542555,0.539159627854643,0.46923739323793945,
 0.8918020727979913,0.3540272898231228,0.1211832843439945,0.4411855508152588,
 0.4132929455688136,0.540993739916216,0.9148385343994266,0.5509184082298939,
 0.25959640619247404,0.9634300436865176,0.8158621644916343,0.1645150554580589,
 0.7304808667432732,0.7057198015649537,0.36404583895813536,0.7179560146024365,
 0.518746500677681,0.2735838703753094,0.9295055166466839,0.9735623067011376,
 0.7812505215369484,0.03915171007805396,0.5571966177087941,
 0.29607149055781135,0.2158769142221587,0.8523659818925042,0.9401083135770453,
 0.2813758659300416,0.7359246360596319,0.4039095230835916,0.9549818237119139,
 0.5978353718141368,0.9079756823292079,0.45537920155020284,
 0.14245713727301168,0.2853363223583557,0.6678539073192692,0.3876012123539708,
 0.07127820163208953,0.8280973878256707,0.06958623115633145,
 0.6527866817768063,0.5540476906515376,0.2687138389382985,0.5397597978567883,
 0.14822073747513154,0.5768738677941648,0.05402630648970763,
 0.03040021215140387,0.6060384229107992,0.548325403831049,0.01989612985489253,
 0.8023472905770632,0.6262603291158005,0.47883919802728536,
 0.35726599178936735,0.11108585363793844,0.7161849817741871,
 0.6824002221013763,0.33944289017995777,0.5195326869012068,0.7306367388654693,
 0.28833843525770697,0.5601365561967964,0.5234037132291253,0.3490474671172581,
 0.6878425990352445,0.6621860639393569,0.33425553235139904,0.6735254471416475,
 0.34381278694293016,0.5822683430305289,0.30771020438364816,
 0.9203504331910403,0.5259955206453939,0.4668800600713273,0.1755470366837334,
 0.8806626091656337,0.9010357714351324,0.33191245553761306,0.4295695206493009,
 0.49667388871025775,0.5390568944121191,0.17993104648095593,
 0.2450710453744358,0.6393341153302246,0.3255266620969848,0.06977086541074984,
 0.02274368715592434,0.6534981614013761,0.136481725310323,
 0.058962874852975355,0.7771407219582128,0.48153528124774037,
 0.9228736203532422,0.9187077489789928,0.0967604044541337,0.3095197576586699,
 0.7117578435175322,0.812425165687323,0.14806760362344207,0.7515522869625961,
 0.5088137029945028,0.2979402241118426,0.21284376777719083,0.6045829642560738,
 0.47777636677637003,0.4304842969666298,0.15725118222616963,0.271893887598327,
 0.0020320587362047116,0.9466647903545979,0.5234392624364973,
 0.1145739031275761,0.10254335490506739,0.7063283201687005,
 0.47529251240330395,0.2488653843423474,0.32948564440547734,
 0.8769706480097524,0.0880777122491041,0.7710024525428825,0.4323730345845018,
 0.6351907015234612,0.9704824459738121,0.3273931325110774,0.21002818817840607,
 0.8305417059594646,0.6364549990645963,0.9428596625754683,0.9197801937024637,
 0.3108100666998729,0.008937159499990743,0.8942363101270107,
 0.019077889228290235,0.1802166023838344,0.8987361815203515,
 0.1439838590922271,0.5062710624611807,0.47552833575223097,0.7167078827907931,
 0.9761795865799665,0.7464299509725787,0.7474498249004703,0.622682101465297,
 0.829165494813271,0.2824593821079352,0.05548292905016372,0.14674760173766777,
 0.6487419695396952,0.7774701378024566,0.8226397721829918,0.337620189741751,
 0.5023167336128623,0.998679783201714,0.8546883009537131,0.3633084132735762,
 0.5726878552897977,0.3287047827534363,0.983390098067698,0.8060270234802722,
 0.2676075391513111,0.817985549000239,0.9617795185913343,0.7563247233960974,
 0.6116788266766162,0.058558098207999176,0.7853545481875779,
 0.22228496067361403,0.36797810815057885,0.20183659226818806,
 0.21434926647228603,0.5916957169392046,0.6811683140061005,0.5512572264375164,
 0.9667997581801849,0.05076374133836081,0.3346102445732839,0.9246331330761827,
 0.8703872538381436,0.5812886164024916,0.17488566623354218,0.4107430224151345,
 0.4989431103851333,0.20140042226712895,0.10058475323325244,
 0.07565055018441147,0.5768326604789286,0.9880694123665353,0.6771341684977186,
 0.36099768689289413,0.16108759829486774,0.13756451964282168,
 0.00168303637950451,0.43593442638078217,0.5782883616961686,
 0.9427854525791087,0.7208512625346527,0.1632300504839329,0.9597524124302039,
 0.6790683403801739,0.6852202541230302,0.5308558834517931,0.9407530235071924,
 0.293987266278641,0.5231048972936279,0.9095894784202554,0.6603261286011706,
 0.9709937151384131,0.3951289387179814,0.828549449779254,0.2443057479127878,
 0.8049782153605781,0.21436425052954589,0.5942122946129492,0.4216608162767337,
 0.9806632410574739,0.8814320136232583,0.877101830362625,0.4586871278723367,
 0.7605617914399677,0.31582411393621357,0.17984162435952655,
 0.3355059657373921,0.14244388682028275,0.9235674597128315,0.7749362470177876,
 0.18605016767277127,0.2253439744260145,0.9927716478463027,0.4557795089862081,
 0.13607702450219883,0.13307970983349926,0.7253928656666858,
 0.6847917841988522,0.6758089402895797,0.4429411846548028,0.18917378789894068,
 0.5005454330483077,0.4811166198516782,0.5312202326728005,0.02431944137710884,
 0.26864385362450205,0.8169549141755066,0.27882507805053547,
 0.28406043009482473,0.8769226479278996,0.41817659958222353,
 0.4972452524935803,0.374523193901643,0.491444815607057,0.9286532720983327,
 0.9359117066323276,0.97628032886105,0.185094378236054,0.739097238239131,
 0.9821530271836106,0.5522753937857199,0.8350962170087882,0.6524319819689013,
 0.7932818266914224,0.22461932890912562,0.4201551984336269,0.740666652117458,
 0.05290893378845318,0.43729137192163714,0.7517949869748407,
 0.33097148653601083,0.625051692421167,0.09153650402021829,0.2017366465709054,
 0.05115278125555145,0.14307303517583103,0.30707145112411327,
 0.09914006820312182,0.8918832758997648,0.046425770023535096,
 0.9727769294037697,0.048086042347934566,0.5977638776407489,
 0.15750155964350876,0.8593362043979165,0.3394786581505831,0.9396602329313153,
 0.3797848774274326,0.046444913937429555,0.07317032804599155,
 0.6308782495602568,0.3057143833435172,0.7660413388418613,
 2.296560113070889e-4,0.7731560543369099,0.37640382738138645,
 0.3506469642582649,0.3490904307233016,0.8146626802669599,0.11810776182788829,
 0.36415178923278757,0.24055951504494277,0.11883170892521,0.08388124552704312,
 0.9140164882765381,0.8846956846581229,0.3332461000932587,0.7496877589070566,
 0.4432127683549777,0.625174701940347,0.3256647173021259,0.37461252200119255,
 0.9037474685707567,0.9021180373198144,0.9909554008190002,0.29790751102476754,
 0.27632737611281843,0.3427003026786001,0.37660819216410946,0.910967842753702,
 0.2673511029557083,0.19412720244111004,0.7680326882131904,0.6714608959852421,
 0.6921045659532099,0.9771597868560997,0.0013073608175802054,
 0.4566017837268893,0.19197387781082376,0.43292040232539764,
 0.09019639661521806,0.104343025407996,0.7679612542950928,
 0.016491391083233053,0.7228802962636954,0.15709970401648699,
 0.17171435590485662,0.5393785525028623,0.4398486470005596,
 0.48058342294549705,0.04044639846370757,0.814106426387194,
 0.44638133959355986,0.21518366386923393,0.2375483981243115,
 0.7662972331847081,0.7839516892605475,0.302436144819654,0.15493343918182312,
 0.8891030599663449,0.41235555807144053,0.047536838031348005,
 0.15005357594090762,0.8389211387378771,0.20259458630914395,
 0.3442280419220648,0.31944557989155864,0.009884548972522889,
 0.2758242079673514,0.47470120415680717,0.6507774850940741,0.430997700652879,
 0.1826375813609622,0.25033502420930676,0.9237150058183182,
 0.10669326071116969,0.9340674645075562,0.556570840963264,0.58425504480298,
 0.10018199135315986,0.31917333375180124,0.02593327432664938,
 0.7995793113910192,0.5075875484198505,0.6888800085150174,0.3367146027494522,
 0.8332262655178517,0.3083946535392249,0.13190362031691627,0.7016247418355666,
 0.5262154235415939,0.09114378622522845,0.9942031013531754,0.6100922740216819,
 0.810632930584219,0.9981472057086496,0.6924911556893415,0.7229954254981299,
 0.7954172609040955,0.4252884804798681,0.47569788069810215,0.9290931378630366,
 0.7884477741730275,0.8704093747524873,0.32065974162557376,
 0.18634416507852836,0.08576133935511354,0.4391988980400868,
 0.2091559607189648,0.7625674481910064,0.3587181067969364,0.524498229901698,
 0.7882625723728627,0.6634676411689299,0.15324774441806388,
 0.11974972578849985,0.7297114597170123,0.19515341767508754,
 0.6729598940280901,0.3441735713347258,0.6132458939677877,0.6895509155203914,
 0.5623832911548114,0.06973338765837844,0.2445167923009035,0.8891133962950758,
 0.1951643574427876,0.9907407249281647,0.6187936596686523,0.5916864681000449,
 0.3134451881655048,0.9568884750987492,0.11026176316729941,
 0.24612674904467802,0.5038798452318172,0.21300330010363377,
 0.7115967396085616,0.5603476345480767,0.2542991609237081,0.292331508968673,
 0.23298569030431215,0.07373894418037574,0.7165445517448559,
 0.6982657051879109,0.9874557867216109,0.5987598010088468,0.9470335937669839,
 0.08553650848654681,0.7330718914502066,0.11327779809201999,
 0.33867612569501415,0.9136183513801585,0.614390410240066,0.106692491720624,
 0.5027267731437173,0.4095163172066335,0.615994347464447,0.3760542814493347,
 0.3524669837204051,0.3245460122693756,0.9963876065244041,0.08854951568610836,
 0.09783094884407417,0.8241981891427151,0.37800633727527055,
 0.22563433338591699,0.8095837471477293,0.1819943843676266,0.9626458789905609,
 0.246707034840685,0.5079794454162139,0.7321051014476743,0.3850345142126089,
 0.7709580594352083,0.39965743513429874,0.19121483050382537,
 0.7243018121312894,0.8295533599278071,0.999820264944233,0.01931188894525071,
 0.1306713643148114,0.6844330141172129,0.9689935366717289,0.32266794353756645,
 0.8045800510331027,0.22739156005296934,0.08931347149053526,
 0.4540023276787022,0.36863724980461676,0.6976225762409055,0.8788439618781285,
 0.9762592279677088,0.5822941098715507,0.9251772720655144,0.2564379771475923,
 0.5221539780601642,0.29480137316244237,0.18630778204909215,
 0.9598329004712031,0.6934430170065536,0.7920373884432685,0.25748640869481976,
 0.9237499495690811,0.11959000251591545,0.24891869218333462,0.94334090377606,
 0.31254638614387786,0.6223655277115825,0.16966870530765588,
 0.15324329410005522,0.3793460990851133,0.5463256753433461,0.9204195883322774,
 0.9226269465929671,0.12525589390728076,0.9786760236606449,0.0181105734289444,
 0.5528973958414688,0.4378655687282388,0.5678883337420628,0.6247691952783018,
 0.8994227229241284,0.4591101775552531,0.1128254238278632,0.8390122126566761,
 0.505610238663829,0.72732290408226,0.4128362081880056,0.04369188113710232,
 0.7932014752676322,0.6756296221760337,0.7064754956760583,
 0.052444007583345886,0.17475259718512226,0.9222858215156721,
 0.998713752880855,0.022056141533071227,0.6314208168513756,0.5636820695231208,
 0.5154957755784405,0.6299043966973431,0.6385488325741725,0.4195374285586738,
 0.1629665814077742,0.32646246145344815,0.4370484824260792,
 0.14492809310864874,0.2082887969267011,0.47928595816331077,
 0.4821172467790049,0.7826358768332004,0.8707452833232129,0.7331877660256555,
 0.9911704972691042,0.6964814441408305,0.17485119228471302,
 0.09793147342144626,0.32250695675234486,0.43903991799534714,
 0.3934690711153275,0.6580120495506661,0.450097916600418,0.3310197724217413,
 0.13129776922635095,0.3155615876341552,0.5277930222390597,0.4784038562737537,
 0.47734833808118327,0.46114651766285675,0.7697551281979771,
 0.8499254140185704,0.9685756326005998,0.9766288888059488,
 0.002081025467243869,0.1457784159108506,0.5851650573287694,
 0.6124168355704143,0.459654190809449,0.3999968175288644,0.8133247766031032,
 0.616366879936581,0.9304127654410655,0.15628645816780418,0.8784403097366716,
 0.0817105535276268,0.9871730238099785,0.035561234070848435,
 0.4835742019172735,0.3028777647248385,0.47007270120632516,0.4041215272766705,
 0.24066136617079825,0.2994324161496198,0.8499068984008991,
 0.45014119606129865,0.7752773324671041,0.2602539502413783,0.7478706943556203,
 0.5062777857086835,0.13386751242992923,0.9495947484868568,
 0.10234046408333619,0.38581194246202644,0.6632145619274556,
 0.3322392052732718,0.7580689589858786,0.6157451933655144,0.8390834738658227,
 0.9378066662302533,0.7186630646088674,0.6058792136590522,0.47041447411507464,
 0.6476355765764346,0.8176474600132597,0.06344138277287747,0.4589208726356159,
 0.5399868883375268,0.9571863093672717,0.16331532160582685,0.7563146029425337,
 0.5451586255299133,0.6930018028570867,0.9031927113269199,0.8144806083959811,
 0.5996451097794976,0.3609947435144434,0.05935540616934076,0.1741716511933129,
 0.31089240654799855,0.22919521214105487,0.4711865071309572,
 0.8357052208070648,0.7576342699335409,0.7110881181120238,0.6236646642394377,
 0.25506161148026185,0.1706218798370288,0.5487035017768495,
 0.054590517166330677,0.47732813527985574,0.3536591480788174,
 0.038463997578455644,0.7744061682837587,0.6135997550068615,
 0.49290305640360466,0.4967176362613832,0.08239635951791913,0.881746674879836,
 0.15037448581354584,0.3179984627866207,0.8759615793359633,0.4270097461734199,
 0.7547136586665767,0.528214517677895,0.06746844560437315,0.7846881005556144,
 0.563155474447642,0.18928412879799517,0.04706250552603508,
 0.44754575218947756,0.2516070982657578,0.8410435687503401,0.9268001038711615,
 0.4387999275882447,0.942795900532182,0.16525311164170942,0.48239372863022456,
 0.2859705569695471,0.16059202975971765,0.1724235482385803,0.6704115113273822,
 0.5408074663336073,0.6810327932944656,2.640712746734275e-4,
 0.9480018420483304,0.7861843858699968,0.7161396222749878,0.7170771769873244,
 0.15702833727827714,0.7900140640370453,0.04232197979847707,
 0.5059146527787802,0.5193764733762953,0.17591369242939792,0.4553973764036826,
 0.06834319254446752,0.3452391099482828,0.48235824638754454,0.735442590767711,
 0.6444378096609579,0.5657624827305887,0.7233204948073316,0.7527365837246086,
 0.5038238697072481,0.9539492713780345,0.682342531045905,0.8720344575424583,
 0.3784984630888736,0.5014169832575562,0.02772006213572087,0.7898656344487192,
 0.26459754071687136,0.6469292373794931,0.5202516350818329,0.6398711272328369,
 0.9561477220817092,0.3651831701606122,0.9379736486420228,0.14596305520608455,
 0.43494755400172413,0.25443107280649446,0.8536208293809232,
 0.8974278452738842,0.11374812834983239,0.23585137805313172,
 0.9933123028176493,0.17869573882563305,0.32923386350731787,
 0.6624303822959097,0.585525556360617,0.07194444380539666,0.9155213004102782,
 0.8978802325177171,0.6881869932948852,0.33917383925789113,0.8482423128871847,
 0.7132498573424827,0.1941357336189944,0.8399705776125603,0.26871374865083697,
 0.22532323444275837,0.6533914748660539,0.5460696569277613,0.8532023715337717,
 0.6650416098652124,0.5441456095407046,0.24524834001145557,0.4540476201669118,
 0.7099705282325166,0.13347670086177765,0.04125959178921246,
 0.8525412152981517,0.779784634040774,0.1768497350955367,0.6507906185215093,
 0.24404308407923692,0.8088869809705184,0.9911198626515042,0.2851520148219907,
 0.22777898411217967,0.9262970083860254,0.5557389799680289,
 0.41902063641856513,0.39904856089971696,0.38862187072947885,
 0.8041127929078593,0.7129419092634468,0.4582567941616036,0.08164329701934292,
 0.338030016689181,0.8775049330034478,0.7902865600479575,0.7199188944820774,
 0.30354801004426724,0.8460564290577748,0.1050404530547473,0.6160234032300321,
 0.3692593438672013,0.3200173606951856,0.6745585951883607,0.46667316236148926,
 0.09097612236204444,0.4700701620237189,0.6652147003001334,
 0.44340692484823785,0.13065088404272795,0.7404593105208297,
 0.5395771148037253,0.9199489858365021,0.7151190425500555,0.085283508185926,
 0.007244482994745338,0.18429275646939192,0.691791702019425,
 0.14732961512742226,0.24965512312080906,0.41888308624735127,
 0.4583491447276127,0.8220291217937774,0.7672312989543222,0.02821905738942765,
 0.6730261242324858,0.6664080787708154,0.6198014542768344,0.08476858764056594,
 0.4306054753207428,0.3774390057623266,0.5044052947458757,0.7491988901868866,
 0.5971598687615386,0.5666058611733531,0.24873970035360216,0.8866293881929537,
 0.0452081654270553,0.331729356123301,0.6542797994525219,0.682887831307756,
 0.9437394365235713,0.29265194115455384,0.2557229009903901,0.2224722000711007,
 0.510500400049435,0.44066022174043373,0.6773230626626079,0.35672168407935645,
 0.35363706946418716,0.7207805994082472,0.560115185264874,0.8022779100603019,
 0.7938951262794708,0.0323175523439958,0.5409953755009024,0.3200961224666361,
 0.11585116504477466,0.386478238073253,0.4386031217323525,0.7483926351610911,
 0.7399831040862795,0.6102284218836812,0.08581645352233302,0.9069375646927751,
 0.22158831008646918,0.7609943533629386,0.9143424950329907,0.59830834088727,
 0.4980213813829788,0.538112509925865,0.6449615913954356,0.1969952897750895,
 0.512439967313245,0.525386125495026,0.21058671463749723,0.5982886303904589,
 0.12659482272966927,0.18916701656891388,0.911102860708836,
 0.07699539968490843,0.211925631264533,0.6418878856038626,0.5675074602676993,
 0.8200574078818025,0.8972038180096801,0.5178385317040317,0.8460605380495315,
 0.29699393414137,0.8467979700976099,0.8088810409035007,0.625997698054579,
 0.6037591035392461,0.5789304042276008,0.3748246237854459,0.9706587716800328,
 0.22355858956547814,0.47855919188510243,0.6577958155159422,
 0.9049714513294874,0.1710178301871117,0.7671734673507212,0.5550499824037893,
 0.9098762938363022,0.30627561329563646,0.1935789435735784,
 0.47660339146246666,0.05831235243099986,0.7221241688844979,
 0.19034784997297816,0.06252612852030137,0.584755192546389,0.6830820716130305,
 0.9293613467182192,0.032625825048456236,0.7500321096334437,
 0.49961580583759246,0.9770838431849647,0.09857953000565745,
 0.05734923074935694,0.7164700449053264,0.233447182507738,0.9814627308713821,
 0.4745831303210867,0.20918800852885855,0.04383929653966645,
 0.15079507712527285,0.5414827697801492,0.01591197580499415,
 0.8418113236453904,0.8765549822352066,0.7215856778873825,0.6726993076085207,
 0.8752016630338509,0.09628628492051594,0.9093594057526918,0.9212341988864291,
 0.7009088434069668,0.10961776236439869,0.40928063197360676,
 0.5661794727517356,0.9810656082450311,0.039608407346517716,
 0.4547715654779918,0.8208978819054561,0.7599315492185478,0.40147936937050277,
 0.19011194615349303,0.5335857058908358,0.5371829024787866,
 0.49592758694928274,0.9729580314498151,0.0633771181067061,0.7426047557556422,
 0.24295434570087504,0.3498111583634643,0.07432119481124722,
 0.3070234215892863,0.2505218944023726,0.21489680826955138,0.5653901358135975,
 0.7154134002317462,0.23916679518939243,0.09538995927079774,
 0.4020560579473039,0.10335100951773601,0.05731556069553578,
 0.19723459906513885,0.42284439752203407,0.761269677827303,0.6901062323802912,
 0.7686450374371349,0.09808947133382373,0.4379963274938251,0.6964913841448128,
 0.34360152738889616,0.04475178856960682,0.5568407293585395,
 0.6759578773311996,0.8211999621615302,0.5504599441729455,0.3620409501439158,
 0.7968434708783315,0.4594577807682342,0.8929267145882722,0.9820154225736253,
 0.5355080588931411,0.5191725573442059,0.4735713936245848,0.2680446687104804,
 0.48148103219989713,0.585043644112615,0.5875682683852892,0.3477551967524083,
 0.8950597940118787,0.892478481274082,0.22632593128881973,0.7864590356255662,
 0.7877152266244789,0.9984612491362612,0.42930678511123577,
 0.32869661825282404,0.5648740453228169,0.2651700195737262,0.1263365230675626,
 0.9952463147884039,0.5023757121448955,0.21644664750876952,
 0.39983288513960036,0.8653786249289682,0.25527492657037343,
 0.07030343080641976,0.9761509406542677,0.30049102954704976,
 0.22374670472092872,0.8458861988995476,0.2660897094849184,0.8921660954123106,
 0.5094276777789989,0.2549292831469706,0.8101725150457809,0.7223634154925008,
 0.08372741194194888,0.099764322898101,0.9033817228311949,0.2318636802354812,
 0.17409408115592184,0.07218080483439626,0.49814334718353126,
 0.024410260485204116,0.7089050258022958,0.6118259478482783,
 0.8385388955538595,0.7891755032778234,0.7381490504921346,0.7820467405435916,
 0.043402159125829676,0.6917158268854949,0.4200246752767597,
 0.029954926498316103,0.426202518034144,0.23525070490652567,
 0.9656666428013768,0.7381837721434409,0.24598463887908062,0.4666708871454798,
 0.10455797892359386,0.7450303502287945,0.6276080277006441,
 0.31676317385512265,0.07848848986962775,0.549787230061501,0.2273387094175615,
 0.7799905507259349,0.2751466327379489,0.43455083878398715,0.1281373058552846,
 0.017018949004866335,0.9594457761019111,0.4599849675000244,
 0.22701553178655576,0.874352549424025,0.8355721142714342,0.02336434225293227,
 0.8728442214652872,0.026798023113931646,0.13203474319187536,
 0.34758057229088235,0.5405448873662986,0.19918193091839576,
 0.8469354270264859,0.9059047509754816,0.940248650241393,0.414206049668004,
 0.07823612125129364,0.14025635034909356,0.07983993665559552,
 0.667073997614159,0.8994531472186644,0.39185568650507074,0.5586491490855251,
 0.8781956561997196,0.544516117852717,0.18167599400896495,0.7697561258653174,
 0.018318193286618767,0.5055246364749906,0.9233393124893283,
 0.5449604019846196,0.0638429518106266,0.7361155986662216,0.5950667192329271,
 0.4349201732505058,0.5016333655663828,0.3088794737302214,0.5606970566193058,
 0.14917788799486376,0.6847479966741415,0.8870993018132317,
 0.15485650672255735,0.42100192241610046,0.522477504368371,0.830220476129641,
 0.04353073702122834,0.3274721224998557,0.7366750129119346,0.7260952751573109,
 0.809308731911457,0.4066057237376792,0.028364694542602265,
 0.38427560242983705,0.26735551401035595,0.5994922348105723,
 0.6381233476496551,0.09107674437849955,0.04782178766820189,
 0.26605005621672495,0.5829204218302231,0.5607119523147684,0.6819714006242483,
 0.6679633572858906,0.5436613155445003,0.6230532718904517,0.1660705464158856,
 0.03261125990181224,0.0026100649335238923,0.019550602081136792,
 0.5864519126329223,0.9855432136148291,0.1373734016134891,0.7339319165778764,
 0.505614802486349,0.431557826972946,0.5809887226048146,0.8888415270672516,
 0.44357031692015014,0.6782035188857303,0.38730939048070145,
 0.8605778980531053,0.2738111330229618,0.8481258256138445,
 0.012277327273067185,0.5430040790702515,0.8139402365674377,
 0.19273750260719735,0.16079224867325526,0.13663361645223715,
 0.49682763538344554,0.013347900658946421,0.7968929600576864,
 0.6110138595893675,0.5126835936497876,0.4538270186212211,0.466881703879208,
 0.30883051919948223,0.8309802622294261,0.11315875664899555,
 0.01839062915273626,0.8573221622412619,0.980862164229726,0.7564215474714246,
 0.2748983071002088,0.012449065649953517,0.7592813374985206,
 0.25469861856433806,0.010209744186385095,0.45319978206404443,
 0.5787551625237928,0.3322554900312109,0.760695672197268,0.6943839177790353,
 0.012643692745616275,0.8396767193823071,0.07071842398089745,
 0.8445354548011837,0.5522887043611013,0.3916972738645482,
 0.011718068070726062,0.8303391405341992,0.8750884605742782,
 0.39097000892774136,0.5826299074925827,0.3751038233492462,0.8730886508947334,
 0.21655108399341128,0.06066441989736804,0.654874622329877,
 0.30171524515513604,0.2507985061416076,0.8340304246999631,0.9213359023473429,
 0.7118070812105095,0.4962802490285253,0.6574022942230024,0.8026462257535705,
 0.7271024138477604,0.7565944165690064,0.6731177816017978,0.8666970518008537,
 0.20755357474108727,0.8170593201077629,0.6013801827939289,0.9812830157576062,
 0.47389151935904283,0.8431131930914233,0.7220006494606748,
 0.27683578375794105,0.29262920731878406,0.36853816260602246,
 0.3694424973621865,0.49605512199016744,0.5545282654684933,0.399885354677026,
 0.1876898062242356,0.854233717634103,0.5573908887240978,0.7847144332663791,
 0.26247204512160094,0.3731529628236294,0.026130112595182275,
 0.46308665456881526,0.41896387414695335,0.10497349825331614,
 0.23129864575295278,0.5457963870854879,0.6606225004603575,
 0.19770874957466844,0.614840359080118,0.47766830547609085,0.720995551955983,
 0.13478627208916316,0.5989219641779989,0.1492749339670968,0.8488043599603483,
 0.6805446378855271,0.7890451569808967,0.6084670405830704,0.525542732939822,
 0.43487466333588864,0.047343975851513065,0.4186910479394832,
 0.09664040600942425,0.7283462548036355,0.43929438961830236,
 0.5167549667951389,0.28031471543499065,0.3182409335682914,0.6628079603646914,
 0.398991319184795,0.7866252085789884,0.04085417235361888,0.07434331755462631,
 0.6570318237846748,0.6269100050519364,0.9284905306177076,0.9601901318647085,
 0.19546410341549425,0.22800473971185087,0.6955517261021756,
 0.47706405850785394,0.42435214127873455,0.001793075513982445,
 0.4878656849628211,0.7829582748795229,0.7818774694551814,0.412369262078894,
 0.8112218056192616,0.21024190970566936,0.16392388927424262,
 0.39369409017087653,0.6599000477490529,0.7647641175669491,0.3696188772079435,
 0.2668515760582375,0.23880394436272734,0.14427239906003664,
 0.7301205939936838,0.3489278213445409,0.4691101148351796,0.2960056932467625,
 0.8114898338044741,0.315362613444375,0.906559506875981,0.5390192533814433,
 0.7840555153060709,0.7778158293836546,0.38760378243739935,
 0.44682430897202885,0.7745850244159039,0.7835913896473627,
 0.10428394926584916,0.9120124668511207,0.050099399911285425,
 0.017510682839715908,0.6828415392782065,0.8240533393520099,0.696235740633824,
 0.9198484287416508,0.4808081899373735,0.2171289890293434,0.39485846266895086,
 0.532249098175549,0.6591052505626643,0.40476273562826826,0.05193834819478793,
 0.18993315441128988,0.6457076355090224,0.9259142958295691,0.4418017404343606,
 0.9658688777803529,0.6911128575023179,0.8043245940146521,0.5372431492711423,
 0.8303188333864777,0.45634678157048947,0.8185608223772848,0.4438907056906173,
 0.8641265382587522,0.42834039745027563,0.5793728326990871,0.6982040055315043,
 0.8368873330083753,0.17189760108717667,0.46370863541607366,
 0.3226521671562801,0.23691642982321992,0.10000558393561232,
 0.49463378634062416,0.8535285836076598,0.6814582965452741,0.7138871692973138,
 0.21460447030480712,0.6845801032330402,0.23568163300651912,
 0.5693127187930186,0.7642535954311181,0.5335857384059758,0.41525903316944834,
 0.8810194436871679,0.3273391270525754,0.0786751015979732,0.6274729874065758,
 0.43925842747187605,0.04015187724405711,0.788262773731104,
 0.23471928589925772,0.7746859647960549,0.4747126234544976,0.8695200015691591,
 0.5595926789757232,0.45152401487134464,0.9062826419582114,0.6189077012075099,
 0.258682596493112,0.8260619661571933,0.5158596786211841,0.7035243483015119,
 0.7665465826159774,0.3752736132958213,0.8063299326250173,0.43221436764444865,
 0.3119521990260954,0.08098411915769776,0.5280522011108657,0.5457526853638743,
 0.22089737164151513,0.4858972762337551,0.7174201293141773,0.2948036527844371,
 0.022195421091671608,0.418867406628896,0.23364841082528387,
 0.024854645054558033,0.9467062447168271,0.7046517395900124,
 0.7718530040048954,0.6334601161962622,0.2700812414773772,0.47912941893861116,
 0.34479422000809634,0.6404526327524577,0.2728734726364954,0.3536044911608921,
 0.8670509861096732,0.49338220554948087,0.19458572040677902,
 0.6445835423485076,0.6955666186679698,0.595482561675359,0.1996451873201195,
 0.8362963019319611,0.3747261371087507,0.7077822568817183,0.5376337212577484,
 0.539286139912889,0.5564958633180875,0.09425384065930442,0.4602781445131282,
 0.07182202402839843,0.3983091470898845,0.9607198012113242,0.2948567144261133,
 0.04707780565983566,0.29508046959755263,0.8996655335181949,
 0.45222994117277293,0.9932718197576188,0.4307982152197124,
 0.04975998113464564,0.011488244218305077,0.4168228545600886,
 0.8878421892473369,0.09780507294395091,0.9167612733542596,
 0.42834549567767244,0.20130768865298898,0.17125462697355087,
 0.5030079751199659,0.7628454996445083,0.28018153826647596,0.909271173093015,
 0.5969399570918404,0.5751200765896235,0.36641262314688905,0.6160738054182936,
 0.4169401452386092,0.6301193471259919,0.09993891097201979,0.4558162187762209,
 0.7351863550337423,0.8023497414614476,0.7555058472581819,0.24683657580404117,
 0.6160909510107049,0.6582951971723486,0.45542838128594465,0.6031548950932087,
 0.027754108573778802,0.5970050571288837,0.11136875017581915,
 0.6845882509237002,0.17991948728478546,0.3661337562092484,0.8481073612796681,
 0.7213386312661143,0.34014077912060225,0.10114559742819484,0.637979269930706,
 0.26048075036472496,0.5463525304971293,0.006485628519966591,
 0.3659756244859482,0.7141473331160103,0.9497724474826548,0.430049633665597,
 0.14993372052237308,0.2510774480351766,0.04248491562809553,
 0.18281426291831382,0.892270630542464,0.8149068812166858,0.3522273109118944,
 0.37482728455833125,0.021820316123855377,0.3910215622526715,
 0.9344727939198438,0.16809094980478245,0.26216260006542114,
 0.6838456939695234,0.877676073920821,0.6453566082998599,0.09556389412453237,
 0.07181633241870289,0.5389203960559573,0.8205896207989966,0.9151048495079198,
 0.15784725191143734,0.7106634086741972,0.8411556332103614,0.5516805283053647,
 0.8879211162702954,0.7834770713947565,0.10763404051413672,0.4739757801637311,
 0.4539275906112945,0.9115772783807324,0.3986226426515024,0.21521813587022065,
 0.9209351878307019,0.22990617331582364,0.5962010704367167,0.5019911853066892,
 0.7461746444571297,0.7163344371270872,0.9778361470643375,0.4323331811195451,
 0.4919195865811644,0.8357390342210839,0.8770161689622197,0.20164645004992443,
 0.2972761238434345,0.3713999611193245,0.11554046931295592,0.8999215675472066,
 0.9861460768094515,0.3432737674553308,0.771208793772868,0.8389549985277058,
 0.7575384063566849,0.9596154224430962,0.5692709181540209,0.2701520614235656,
 0.09530485747639383,0.13335528967368704,0.2513798210129117,
 0.7134232429766669,0.01669820929983523,0.527504250931128,0.17540935310706296,
 0.6719557481463706,0.7139398136594988,0.7866863099106428,0.14023207252016623,
 0.06997934768842762,0.2530104037147741,0.6200034359811093,
 0.25426210915723757,0.87912408295317,0.21646325175733228,0.7026451670950825,
 0.9813301317698553,0.3214311455496608,0.4047204452079748,0.32817932634884106,
 0.5991032668826767,0.6381328116007234,0.13082349818577943,0.9711440202844389,
 0.43857401226399717,0.8272237536437966,0.7282203229314402,0.5721237911221015,
 0.2525140563424042,0.7539166231992267,0.27985904574594966,0.9771758449879383,
 0.5402466365352498,0.7927268169867208,0.3859300161739626,
 0.002446220572530411,0.8762519039138095,0.15887176851375018,
 0.2201008294073431,0.45408885656597064,0.18443678653169948,
 0.20814978655516903,0.11771532986726996,0.4229458287808354,
 0.36589608560639164,0.37755250546344876,0.7272107396135331,
 0.34878474699480033,0.07865948005531709,0.8211809877222666,
 0.09309802311958892,0.2909886087559941,0.25248009829338824,0.766594214531116,
 0.1531468020690303,0.09310796284236678,0.7722581155551064,
 0.41249355945653954,0.09917630697989732,0.2768593388495605,
 0.8751399059786056,0.4908065133255072,0.5465057605349803,0.8130768208192577,
 0.021113293309234837,0.697831938022127,0.30237422937666825,
 0.5926614223341612,0.05520102525224013,0.8941073385706322,0.5946235592260656,
 0.5655128542630434,0.12168919830464286,0.95841269727108,0.0374278673875833,
 0.550872360111055,0.46853287655222375,0.10329538659754955,0.4764560647191364,
 0.6528857180262597,0.7238663170718785,0.53254798023643,0.3722631063455005,
 0.1450961218306993,0.911959760662401,0.16081406719595115,0.42267080305724636,
 0.026148870940130298,0.49653612173324524,0.06295591259668987,
 0.9364167127133732,0.272295328546834,0.3863338855537266,0.7642734219035862,
 0.46524217538763024,0.38161577381546374,0.551349867523995,0.589891250116282,
 0.25159056325272866,0.6864941057147117,0.7095563611022736,
 0.42110673269069543,0.3655594676560854,0.9005120470370436,0.9234245940869568,
 0.3184146173676713,0.9486905225552199,0.0393171275407006,0.9707832609221334,
 0.7904696614496848,0.8465505889210144,0.11213031183747224,0.7025586604863898,
 0.858152981844291,0.5846883889526202,0.28080189167324154,0.6510605747169856,
 0.19145178489926684,0.8032044556021416,0.8585204465099537,
 0.32754778894788006,0.29403635465539,0.38394123512611,0.7731109247514194,
 0.12853459247513388,0.8733460973121381,0.11038856716014456,
 0.8722079617099286,0.41501587509434257,0.22515441350850218,
 0.44802172502874305,0.2009852130675357,0.9708650677393519,0.5300796630312479,
 0.6677546453721123,0.759644686060986,0.19032475154109163,0.7905433189264565,
 0.3598077048414605,0.32998083657851307,0.9845760355933467,
 0.04159355173236057,0.9872817928206279,0.018945272602649665,
 0.37504080114310556,0.3311934754791299,0.7294351378942404,0.8879357755416981,
 0.6645528792354506,0.1619699904976577,0.6746603830297919,
 0.034396461593526606,0.8914150791032371,0.4475796474657695,
 0.18664907433185352,0.10307851096420484,0.9346840949406793,
 0.4628535852635053,0.5718658552185636,0.5941495208982792,0.585274520791604,
 0.6076814413986444,0.6817528816106484,0.44846089728373173,0.5373740087256748,
 0.12488373019530219,0.626286248210495,0.9822140903981046,
 0.034734918238901846,0.8375866583158386,0.10231658446472769,
 0.7427391739533422,0.5603122126046469,0.14435841211510814,
 0.04059585695569223,0.23005798526964327,0.5054206468418101,
 0.5703089421729972,0.9671945729109833,0.11802637454627207,0.4987601277441135,
 0.11398784657072214,0.40005177887179977,0.5147802687457883,
 0.2116623589471135,0.3138809023370962,0.5907433713539555,0.1984551156486788,
 0.17735019618182668,0.7610202300187066,0.13924272779755642,0.981906533122504,
 0.292377200949137,0.2415871781176201,0.8860567083844945,0.4177880820901485,
 0.8006696126871522,0.27653580926810584,0.05909750582260487,
 0.3673518489234844,0.5840162481738485,0.7234469526735843,
 0.011004434379809358,1.3688653285914842e-4,0.1461810947174622,
 0.8939723511512727,0.667895233353276,0.9655681993430396,0.420165124908628,
 0.6442002054927196,0.7342884544895234,0.14644427148077943,
 0.04164404446225034,0.05697024094472847,0.9531405559625811,
 0.43473549035359893,0.7678996778417652,0.23726037852547943,
 0.9450035578417505,0.4637895934380821,0.06669167889801608,0.917156160453152,
 0.2103388670682248,0.4199460475544743,0.3311546159284723,0.45947081563694026,
 0.9259943712825877,0.8132713484230262,0.40818911815187575,
 0.13799501747468734,0.7595982548866295,0.18934308051208448,
 0.5718574872333696,0.3962554262882376,0.7058186188257725,0.9043982832270867,
 0.26064035562523835,0.28982882041737956,0.45095894132008374,
 0.356402059830053,0.9783891253625095,0.319017184426295,0.544957674093668,
 0.8961620739431134,0.15920198905926075,0.4064438999861072,0.678150577215904,
 0.7671554531903614,0.11273900636857581,0.7782051106166201,0.8328898443728341,
 0.8562273133960104,0.9958664460227732,0.7389469689334285,0.5732835227714814,
 0.9834354302212212,0.0076981922947783055,0.4369719032236219,
 0.22146427957716497,0.589188801952397,0.4193132803211914,0.7802593265656106,
 0.668419195693373,0.3457414236962254,0.9331831888196456,0.3277371095505026,
 0.4452256264236917,0.4347733314329141,0.715537560971767,0.007051151084621843,
 0.9011179210702556,0.4425465400827838,0.356247597501436,0.8901661069748827,
 0.6327715870839921,0.853964285540465,0.19966718277567497,0.5427602547900563,
 0.6142218014656021,0.5297505544638696,0.9404991149533559,0.0870957374041903,
 0.8632407229722419,0.06991578423316225,0.1644259645591889,0.6563997651256928,
 0.669194930398262,0.3678891633628515,0.7616767652005118,0.03564020192146833,
 0.35204729357301745,0.7595227361617345,0.8102160502454483,
 0.48911954770622623,0.4085742549187762,0.18003417819009493,
 0.3745767023012736,0.2620136519980597,0.27648441482268693,
 0.20506452208453374,0.7346407814458826,0.08663667772838313,
 0.4703598595965882,0.8464063242818626,0.08072251229074678,0.3918249246333435,
 0.4815076277436634,0.02288716610433783,0.6059859002941619,
 0.04851022545684602,0.2936199969809483,0.1253703797268353,0.6750458403932262,
 0.5834793098078741,0.005779752277320416,0.10430185607249187,
 0.7951853703090825,0.06461830350120312,0.18023051465461615,
 0.8629775823338265,0.78378725147814,0.09839846241304351,0.24784616110813762,
 0.7463450221935481,0.3039828096225623,0.20656534956581818,0.7116578962214526,
 0.6813914195234437,0.14678091082870426,0.6527139224255344,
 0.21582382552714552,0.3331305348659338,0.13422769709289983,
 0.40123267246294536,0.8479766142383139,0.1345654703144088,
 0.37610682415469865,0.023055694517015612,0.5176602648700478,
 0.3265282308549763,0.5745835772422229,0.9406358395158921,0.8984384474327163,
 0.9627702597902166,0.789893380108736,0.6344603803377795,0.49248720241862154,
 0.23471817910628578,0.017825930050158756,0.005325528667377788,
 0.5833898326483922,0.9378814948189689,0.7845544582334847,0.5286913558632829,
 0.4133071707163569,0.13482968510718618,0.7066118521070903,0.7751226494563229,
 0.23364803628644437,0.7204670324573835,0.04207482885570968,
 0.33442484552705665,0.06286501254428234,0.23737483404788995,
 0.7404146570249615,0.006036282244667901,0.8662698562131717,
 0.6083829507607016,0.3848760513622068,0.11766727173886771,0.8273956725161434,
 0.5165993931171763,0.3953855836674951,0.47955745346505774,0.6949111452935668,
 0.8335835344906476,0.5605963354060519,0.2626372963866659,0.7123579077977598,
 0.12263568456163909,0.8986876553355405,0.9779985069484272,0.8589072844336988,
 0.8460765720038825,0.1950100906845118,0.5542040953574833,0.18925533338342637,
 0.29641598736009356,0.5171923105406312,0.6967201831486726,0.5480348448128792,
 0.6584292096776067,0.02325271693300579,0.942469526894455,0.08423261853747677,
 0.17852857798619248,0.31265937304592095,0.9713981807740362,
 0.7350530090936109,0.04301666267214188,0.09043373455250281,
 0.6513660833472185,0.6094112765818149,0.6226663273029929,0.07400204141361666,
 0.5257116791442183,0.7080868637036448,0.4312749946832034,0.43706881023767497,
 0.9042597936587766,0.4498162650129396,0.7643642567350714,0.5853654153503174,
 0.955774808656311,0.4115867695334994,0.1748291356168441,0.39054711436449185,
 0.19975196384280047,0.3808967971450029,0.6854275361107637,
 0.008786013613408028,0.3761287356052003,0.9671276713759553,
 0.13741401819332544,0.2626115630100839,0.4750841858083298,0.0872873662627982,
 0.03369324850075617,0.5103165313698564,0.47009824528656685,
 0.6995640049867963,0.4966822088157472,0.6743532359948176,0.10237300835425756,
 0.15975338471355482,0.741929336529835,0.008715967076920572,
 0.31857259919563463,0.11193620286837669,0.5660465954328193,
 0.22201039372041864,0.3560511144390037,0.743445123572642,0.07214943858822742,
 0.402780806927419,0.053283323214049716,0.22141947381100857,
 0.17543527861910757,0.5025185593540008,0.9876935528051103,
 0.24194647531053615,0.5606283307518808,0.05120051869768305,0.730781403620738,
 0.07636254887120864,0.08010030537519786,0.09238051893670274,
 0.5688548280279412,0.9284275796105884,0.7215590904961675,0.4340429643370821,
 0.4125270747337648,0.1974401938118162,0.052450720629252046,
 0.6863824887755573,0.013878217078734156,0.7506888546031484,
 0.4164987225652955,0.2538374778819589,0.47314967622767057,
 0.032535969898674466,0.461491161239979,0.07000548462318679,
 0.8096409080065152,0.11724199399722446,0.5292725347917868,
 0.30811808026167753,0.7390571172458906,0.4914353121125511,0.7595138011314058,
 0.758490743279941,0.7563362691434432,0.34979480410704544,0.11438822248247571,
 0.39864442944965006,0.0236505900120092,0.3820283668868917,
 0.14177077146172645,0.8656185482362898,0.2971181928973584,0.6397645339351578,
 0.04659935413372451,0.5903613681277015,0.5651180814892053,0.4134472498903875,
 0.08887783108486991,0.8549993617949504,0.22260443223304005,
 0.12796668745035256,0.5411341416091956,0.673950417504747,0.1835681177352373,
 0.2500386151248337,0.9234867486054104,0.6473127413490305,0.7559719482672057,
 0.7105695015568079,0.3318451808482785,0.14060922518458785,
 0.07701235007838148,0.9395420389178856,0.3511780504624795,0.7961444197396584,
 0.09415355328421993,0.7491344090027993,0.5662180002051297,0.4120173582471335,
 0.4153566295976353,0.2836223232520625,0.8728109042993425,0.8476680017664967,
 0.39114066856301477,0.6580425122194115,0.9094177458199121,
 0.018797612287252297,0.5815685365758647,0.5251704057499648,
 0.8027051006108675,0.7219793319869043,0.6271245238018724,0.6978543633318152,
 0.23238697790068563,0.5291415447954243,0.820833233263438,0.5044239925726689,
 0.022856825430279337,0.2177915620884403,0.5714162353529533,0.971184428965809,
 0.8764381492314687,0.6286322041463527,0.30520338082123544,0.58900432306237,
 0.9020495144152169,0.3118851205322196,0.03765378409950049,0.8745639187352039,
 0.4129002697829883,0.03787803929498412,0.5273521835928348,0.9128310791352404,
 0.7666713449651836,0.6240725143450009,0.44897391302666934,0.7068467180031037,
 0.24626440343170308,0.49099523629182085,0.48894111599383416,
 0.6562280032598264,0.25841358297094397,0.031239455320767107,
 0.4960074567274223,0.9371453871199001,0.24178714518817146,
 0.16637239709390206,0.05853108512384053,0.0194090707842427,
 0.5914086769653607,0.59846320876394,0.8480715720798246,0.9484043378507355,
 0.5677253183228806,0.3697373193987529,0.9545978008534483,0.29203636698247915,
 0.6835923185131229,0.6437492955257779,0.17653331220331603,0.6133447873523712,
 0.3388669002119369,0.28521768861925445,0.08447508603208498,
 0.9000123448281985,0.843721990357393,0.464804032010969,0.04387847144831958,
 0.904784278748104,0.47882585662073973,0.308019360383439,0.42204275499962207,
 0.8857107178884612,0.8684439512506041,0.8919474599628237,
 0.055743471271195943,0.6751522714910744,0.05924557518312723,
 0.006081306333466263,0.48591924593994473,0.5215769731757034,
 0.003240567756805812,0.8180278828739729,0.5199041885060344,
 0.025495656927440713,0.6589542825604549,0.0031538671018895847,
 0.4941914987289274,0.35555098941390273,0.13889054393971878,
 0.15306259934832434,0.044201468067069705,0.9530631508120069,
 0.8854540419348245,0.28732467439079146,0.5568489634712067,0.7521902667879401,
 0.804361501071127,0.2594360755132761,0.5359615273988159,0.051502231645130014,
 0.15520072410152042,0.4413396957455701,0.2536139211873504,0.3616413328180257,
 0.5660500760422513,0.7801724867435196,0.12900016541959347,0.8324934795008877,
 0.587328409242138,0.5909013040081793,0.19244258864215913,0.06463102120985431,
 0.6034636894341256,0.3395473900164342,0.08423921615271956,0.9492278972634144,
 0.9208491126949924,0.46495959399801046,0.8045895410504318,
 0.19861070472296316,0.7177888713345185,0.00551693285983379,
 0.6048272551163077,0.22445340722607643,0.7245439050519122,0.5097643844843182,
 0.1725671525719632,0.006241017625248646,0.5488125366078177,
 0.43769963446936166,0.7078043494294601,0.3562866078838913,
 0.13680437156903957,0.36641712498169465,0.24892802865706642,
 0.9086006178596283,0.17246778335075796,0.6948205073728874,0.2396354359895213,
 0.8988793270278038,0.013028977958100985,0.8185799186884768,
 0.8102755439543471,0.23882935905921954,0.32902515158842705,
 0.2515024286904788,0.9361345620286889,0.37583916439336174,0.1242746004754618,
 0.4008471807337156,0.7163651653593714,0.7901156947192279,0.7288300392451508,
 0.8234314864074729,0.49452125595678265,0.7008518366128957,0.3481858095419891,
 0.7781607839372515,0.463073727025057,0.3386141797699559,0.7865849563534237,
 0.228143521021828,0.421065382773975,0.14141027509290627,0.3023497163371234,
 0.20400332684580513,0.1525104162347124,0.4699730574099009,
 0.17597757154113447,0.9868122617809224,0.8692683051571956,
 0.19629548507921413,0.44428306531529627,0.16380303027896803,
 0.0791975021252872,0.21532965637730417,0.34067747832727624,
 0.07946733676258133,0.5564692682504442,0.7445689813380113,
 6.692319613619269e-4,0.9490694024716041,0.9900527429271648,
 0.6437447255357882,0.4677536624664693,0.3297671578275303,0.1615334755508373,
 0.09574728122746667,0.777020576042041,0.7979531582700852,0.6958415374328488,
 0.675131971261671,0.5623510775293025,0.7939809281737644,0.8870504111768057,
 0.7299981623038901,0.3629874463348599,0.7446092251666372,0.04504726108686674,
 0.6390404301857171,0.2547659087419897,0.7331989516929962,0.5904685105188004,
 0.6869234476135879,0.1515510046769183,0.8937556136625587,0.22475228169783557,
 0.9021898116966026,0.27488742985360604,0.0598612412102425,
 0.08338819901182126,0.01682439459683993,0.7915298680479885,
 0.7272904724356626,0.7773510080881694,0.045154207428111715,
 0.37232246757826115,0.8818909201967771,0.3150945007099235,0.9153850725476592,
 0.8833568725777782,0.2811151983677539,0.7111999575248797,
 0.005312704491928888,0.5767169402486872,0.45995734649895914,
 0.13615635987630537,0.7078411554582144,0.3593670947495402,
 0.27095870939631406,0.44994672307078953,0.6036038437679038,
 0.06785803871383256,0.5262997794992873,0.18367859272511577,
 0.2874909493784378,0.49431022528417135,0.7923416517263634,0.8440460789547787,
 0.14707493600915666,0.9266994613234567,0.4451420739204277,0.7627425630265289,
 0.024853636388124323,0.02012678215360275,0.3967469111025762,
 0.7492148080573685,0.3527759877543287,0.2061229292446739,0.33693749886138036,
 0.7239928251447204,0.3542997419680729,0.9591397029811709,0.4976557518367719,
 0.7441700746123454,0.8676292936805542,0.9180371409503723,0.9365279401670905,
 0.6391847263339862,0.5224087658850625,0.14126111542158337,
 0.06319755153196827,0.044737403982173185,0.8831298042852433,
 0.17748701448066118,0.921510972450107,0.14898200642913118,
 0.06598415451343931,0.8866762303764016,0.8209122255797101,
 0.22756080724994854,0.9030729586793914,0.8681866450636606,0.9519423874716787,
 0.33693662533516067,0.6767472455040001,0.4334373960882121,0.728595469392848,
 0.5622406113835889,0.006489371637928842,0.7049803875396742,
 0.11235895047697364,0.7628397472990768,0.4773394063637103,
 0.25246755150029454,0.5587693602542955,0.16976754935943472,
 0.8770086515334142,0.5836081148326688,0.5285889049195825,0.5669013407192102,
 0.22529842177705472,0.6724805184878987,0.1527905202161549,0.5977324696386216,
 0.621096054323486,0.8921561049553061,0.4385038783121372,0.6773820619315509,
 0.2076017450944655,0.1562904236652729,0.8370247569985801,0.8075704241426955,
 0.30926988563283986,0.4243360668797067,0.23286039156609872,
 0.9941155036881274,0.04117403722034296,0.024933982216420403,
 0.1128704431128158,0.012647531189970196,0.804234797986579,0.7668352782675882,
 0.8507848142457668,0.7369141941298656,0.25091049411697897,0.6492154919411823,
 0.6283202637145807,0.8597121557366074,0.7520514415805797,0.9196951890652346,
 0.9121391069290002,0.17447581068983298,0.9385649427760351,
 0.20578912657758353,0.1030235754359985,0.3817791510153191,0.8863384391634013,
 0.7443353134069111,0.5450099840083512,0.9691098415546734,0.35296980210798357,
 0.8461689007904051,0.8887718699252718,0.6215659509949442,0.44570233134865234,
 0.9856879565779397,0.6288571921479374,0.9471198196553845,0.9654302000275576,
 0.11247971114521749,0.508221114121317,0.8010758907028126,0.6522670004664946,
 0.7714272866192168,0.2811219627164637,0.3400713465486346,0.5806167190566734,
 0.7061321087655309,0.12984919598689837,0.7334062316196323,0.7848332071012287,
 0.3775946871323056,0.6966905551635372,0.35412518327991327,0.905597188050855,
 0.7442773476358854,0.9281247741028398,0.3147335235351435,0.40892724955654935,
 0.45106219580472373,0.6966927136695107,0.2722605979147563,
 0.07408404874475216,0.09030497321817388,0.6434855753429823,
 0.26279537053041124,0.9488804741206903,0.5537007809281765,0.7766780377461946,
 0.453867484169753,0.9326448467724836,0.28056179862070474,0.7840591676868959,
 0.33428921365678765,0.09343620341094627,0.36170282090566963,
 0.5936672153248287,0.17006958981349551,0.5530358211408763,0.596198058731284,
 0.9658304397889271,0.2983917834954657,0.5039007372410191,0.34932725277328325,
 0.7737844902730152,0.6135269030647214,0.3074433128152818,0.1919448537911964,
 0.9194949006048981,0.03492901565408901,0.03264489613142618,
 0.039162844712194333,0.5615539572115353,0.6408856469345884,
 0.45099308564618035,0.10513613251091947,0.6533884447215745,
 0.4019345703534577,0.6716265213288041,0.8642822880177427,0.7497670547110813,
 0.5344469086998819,0.5708457737029101,0.6620988634296558,0.02651688489273729,
 0.9372324264191096,0.2995435685912753,0.10190659296382398,
 0.37343422807946414,0.9173104168925248,0.06211002051641812,
 0.8545708384668276,0.9566083029690126,0.14527284909049043,0.8618731395309731,
 0.6295948440682395,0.608997776836624,0.7971093518223935,0.8591204366846126,
 0.8291240451513844,0.375706994799049,0.6835064795072517,0.2335923310330411,
 0.03165649904924184,0.5443485290414604,0.14227205013185906,
 0.22674275774518216,0.3783459626131249,0.8492454337804561,0.7245390838023487,
 0.0869976483001823,0.08310782974598196,0.676251752433721,0.5059112493609486,
 0.18185509758739282,0.41153907066777207,0.9448762801576,0.30295329841528607,
 0.8120281419299564,0.9408466824607217,0.6830280752912354,0.7834683068537426,
 0.26649367707398985,0.26759867370131984,0.6272438999144778,
 0.08996619482277524,0.12761334474120833,0.9011580078387527,
 0.5918180939963276,0.4962344289130778,0.5647307653484666,0.227948686505132,
 0.6438268962130145,0.8069926056990194,0.2590740584606417,0.3398992935286054,
 0.838666405379293,0.5659576694387095,0.9954540707394641,0.40468473820844597,
 0.44229899264698913,0.0946363134929874,0.8827158917568648,
 0.07101586838937324,0.846290367473917,0.700041654046107,0.7626706873213774,
 0.14614485233947938,0.5004096083074956,0.71186113762235,0.9122503402383166,
 0.49340538740140283,0.6966416054428358,0.4600522682326479,
 0.30021052866308273,0.5377237751410201,0.25206643609955237,
 0.32911554900041573,0.5549814248437184,0.2010018258505365,0.6557317617493155,
 0.9820776348640051,0.6037835026474855,0.016474333503525562,
 0.26003571346267773,0.9324502665005413,0.8379429862416848,0.7986815438189193,
 0.3716534235816197,0.5423534608137557,0.6834250177933053,0.7440978284619362,
 0.683374820797521,0.5480369252079655,0.6039569504562609,0.5956938994098705,
 0.1607085601649837,0.43773619508548167,0.8121317291696765,
 0.23581001466483187,0.6293259523739081,0.8958660269713121,
 0.11216585424221004,0.41094251853127495,0.14231524627230563,
 0.7391956650375846,0.6204121871725552,0.12588926859754146,
 0.39572930680053997,0.576387396662309,0.6564652283983322,0.5164807386624696,
 0.676546843892534,0.2320432266966167,0.028743624150645797,
 0.15726125306451277,0.5879124758213012,0.8056689810729907,0.2691873071162155,
 0.6068051762697673,0.2652244167326365,0.2611137419070697,0.06912561290061192,
 0.26357652541141263,0.46802815462025693,0.3383770127799123,
 0.4297769911797408,0.9644564153602156,0.8083470161752739,0.6475700269583099,
 0.9619133232771774,0.653435217220597,0.47348227503653384,0.5936942246518723,
 0.5232012447707285,0.4387131415767631,0.5391944239753724,0.5716335708960854,
 0.4727320032041228,0.38943573140049537,0.8767240360248366,0.6744027296540974,
 0.06701005810369909,0.8835678179772869,0.6759132699559729,0.4969661145935721,
 0.13265028845765958,0.45183653124230894,0.008694505184666967,
 0.888663102123808,0.8033758440156497,0.34994994561654147,0.8995739658318903,
 0.796483834198179,0.8862680174084858,0.1283937578799379,0.46373640344003,
 0.7900321109307376,0.3319408289233696,0.09695097422067045,0.7878174527642405,
 0.6243272510697596,0.688765477897891,0.9927192627164785,0.6611762765044562,
 0.767534477460937,0.713576395946424,0.1346789722554128,0.762974769657269,
 0.1260137718679325,0.9141859728068896,0.7416552717017766,0.3628874387751807,
 0.007586692869228973,0.7160718824237077,0.4191570786126463,
 0.13319286570689481,0.7056719450307386,0.04361666441077838,
 0.08618180294308764,0.813394829189055,0.4078020456309184,0.3148992147561679,
 0.7731576212916821,0.06927379183565907,0.8647874348345642,0.4395053927493534,
 0.12094534906380283,0.3358874100018485,0.3986298737646431,0.8586697147530439,
 0.9666575088763176,0.29632646893096415,0.5351224906172797,0.7156688801830647,
 0.5400719658369346,0.36909251658281894,0.10499585398087086,
 0.6770964139860098,0.7585607387207529,0.21854954590312325,0.7463198865609001,
 0.32910582480997563,0.680815886507503,0.06650557087088671,0.6252349869892749,
 0.014145355312897512,0.10351214731976477,0.2864775536795069,
 0.7836960166442812,0.11053501761545415,0.6414243746825323,
 0.49450660072513164,0.9672533249498415,0.30701643200492446,
 0.4566495025532802,0.7458540934530379,0.8930153422556174,0.41153072211738795,
 0.37278290957451476,0.16669313991817736,0.09623648674576302,
 0.5498154411635157,0.5014917123991806,0.9500701923199366,0.6040502250837626,
 0.6966179025935132,0.8027995765433041,0.9620812892869128,0.5775268786309153,
 0.69809836376585,0.5925141323787388,0.6096635916441137,0.3522870180119315,
 0.4532879960334806,0.18893375620294495,0.5314536841038289,0.6781477882036506,
 0.07094232051694682,0.3306361535341048,0.15460162855059123,
 0.7309089138999694,0.19051898096561493,0.5622008199086348,0.1504815186883729,
 0.446255958559853,0.9470627707801598,0.6577713164660471,0.6002970018021976,
 0.9580133795357553,0.1733176181235936,0.40760106149766384,0.6623344492136305,
 0.6116344766936622,0.7609376005198789,0.8858152926646622,0.07466628103248352,
 0.712817781594933,0.9081718005579879,0.8586923080533829,0.39024760267191716,
 0.5682443658838235,0.25194025703168776,0.19495285047715016,
 0.5799229716793083,0.9924517310007225,0.20692210728959326,0.6093703709154504,
 0.5939003255288521,0.3108927348514561,0.8981329330958228,0.5049214735332669,
 0.4179956516070398,0.24859166485626716,0.6143362729779007,0.7907683173758759,
 0.5552597148080192,0.768759799299311,0.8138314219781918,0.09026371929589228,
 0.9195167593357133,0.21632337425932324,0.8439545288362164,0.5016696943806449,
 0.32153172887782033,0.9830974355389289,0.09716995998092104,
 0.9867817886078672,0.48690530120294906,0.8546914211048109,
 0.36099981020699057,0.4963337943707532,0.8593136303408265,
 0.11865281052464016,0.636632173742517,0.9879717829158098,0.9747697935627292,
 0.24270628346007062,0.8062816414090355,0.7551970467061773,
 0.08799652448255535,0.9309749495158665,0.559265540201488,0.2933866791652042,
 0.2534431863103431,0.3135371483219216,0.5384293694438445,0.5381653683311131,
 0.28723455877883064,0.0015508916601524314,0.2270126808066305,
 0.7193882272940002,0.3280422451657705,0.539917742875403,0.9197643512880059,
 0.17823354396364677,0.8509283828986638,0.3917647830102512,0.431819671869238,
 0.2067464208359,0.4239708966063217,0.1326841487448973,0.6027885097634993,
 0.09139992286896759,0.36785942502771096,0.22777072856256197,
 0.7010058393895089,0.4723283022828646,0.3010165959519675,0.9834636789490329,
 0.5335798462640381,0.5982831052614456,0.23012007381458677,0.4679244121649685,
 0.14292591727300397,0.38431965879367924,0.5456821028143259,0.657891001740054,
 0.2745727362430779,0.7016430760761769,0.9870434802811603,0.23618320033936713,
 0.5164871178307532,0.033521546938828584,0.10232775438754504,
 0.8928731134525479,0.525091090603836,0.27389316911329187,0.5881847758097819,
 0.5843322286682026,0.5429691261802584,0.04558446275675565,0.8096223514118536,
 0.167500449198549,0.1537389447270303,0.7677797605024798,0.31772730700833995,
 0.9616692287353461,0.5945888387590326,0.6244724693720031,0.10981251458390928,
 0.8403936724910211,0.33593508159455754,0.10269144614700321,
 0.3351070385346171,0.14267518122747713,0.5569588255219123,0.6537776743605764,
 0.19940324406474141,0.995202668211762,0.16609182713899362,0.581493558437276,
 0.9853556402401872,0.6159657925108362,0.060222885713179464,
 0.11709670432059394,0.4159248702666305,0.34423506531021286,
 0.6065864651237244,0.5749541229700393,0.7571860174484198,0.7700196916745264,
 0.7963149025627528,0.37890284693291376,0.8588573296162644,
 0.022965562141678664,0.6187395852347365,0.11985685578752125,
 0.1274511062995498,0.8539569238689738,0.005877105053575082,
 0.3400787501984561,0.7578899574822886,0.307657676397333,0.9355905854178017,
 0.21500215526217248,0.5254873780759644,0.6535436954742642,
 0.12846262864884506,0.8768184544058867,0.4633954856617718,0.9018291193530434,
 0.1778750658235676,0.08143345665169366,0.32850135900276023,
 0.17166511056949696,0.2701165253019292,0.7344325638707556,0.908693728809544,
 0.16992409577885415,0.6794628846459622,0.32278774086092443,
 0.47711291909419695,0.4857845747308399,0.584314431934983,0.7768283654476741,
 0.1760656521753784,0.5893994768965921,0.45580649257703276,
 0.39588186938361747,0.8324317729220919,0.44211164890546173,
 0.9163538781441797,0.31313113026560957,0.31355688827719375,
 0.46702846749645943,0.914182242830181,0.49486697637361254,
 0.09719371047711722,0.3200143739225695,0.2660700482178894,
 0.26606495071606906,0.5753401149074401,0.14236126018206163,
 0.6619067082437479,0.43140729350366214,0.6650408114359181,0.2581158680798106,
 0.7562962483452904,0.4583793376611709,0.2035107562340841,0.13107915686234517,
 0.32081615653946427,0.8100970831469655,0.9628119130089576,
 0.06695737277860969,0.4761569273239332,0.07448626736442576,
 0.8660469919517277,0.29058374727035474,0.5654218199376632,0.9129916708831796,
 0.3882190782025374,0.3807158888986705,0.8755217149883243,0.186556666814501,
 0.10156283539086497,0.4955401675006351,0.7268701759937074,0.4237899833943465,
 0.8152348842264155,0.6999703000418118,0.3427624397730772,0.8670786010139933,
 0.7621153507669155,0.663036981347003,0.13073472637015882,0.2181831268273231,
 0.48172682284484925,0.9457007176312805,0.7660640621313176,0.5215531979927512,
 0.6393171211736498,0.4104330450950626,0.32230851434396224,0.8751459719129201,
 0.45685982476334286,0.6393536299102345,0.7941302882720687,0.730016596663258,
 0.31525430009090094,0.019678979056160784,0.1894139886110341,
 0.1269288509977189,0.868121766697436,0.06582314025517588,0.69322048369867,
 0.38906060183181557,0.07030391712971285,0.7141827772794277,
 0.5025268952585904,0.6335173790294866,0.6837858018314666,0.6563442289322432,
 0.8326709076713197,0.6025912453832105,0.14736664617140482,0.5396178759381887,
 0.45007657781393084,0.9503150496627523,0.7329125743789252,
 0.16594401657191493,0.3296402739412678,0.9735767777362676,0.8592131059349394,
 0.723276789273836,0.7895377210148056,0.7426218234685606,0.8262905096431609,
 0.9158355692917716,0.8114221625029934,0.1562425143947821,0.33812271083811285,
 0.7964111886442022,0.2358338500849959,0.7351053380970038,0.07832904788792616,
 0.1232521179985725,0.09748269261987474,0.4430468769765603,0.4728631340503723,
 0.25548467858984514,0.713751440507421,0.9627768419114281,0.7130763711259437,
 0.27518033858333446,0.061047335013927895,0.31751162221675966,
 0.7091748800794064,0.5437375538146251,0.578770359472352,0.3461240515787306,
 0.854174779943365,0.7112661221081379,0.786589025182254,0.07465298232166717,
 0.1691545289684393,0.3856216309558961,0.8944686857438326,0.36066690884568064,
 0.29344179378407564,0.699330927650189,0.4247186051232701,0.31115898085741656,
 0.5699685766596329,0.9155394887593812,0.9870401161157061,0.6114863048970285,
 0.6060161417904439,0.9052244641634904,0.7454024515012089,0.49448767762026025,
 0.9283896963246893,0.15682576690815775,0.6201865563846338,
 0.03667706014221006,0.496234603160827,0.9856504353211157,0.9293160355313083,
 0.8799523552802395,0.06250514143301067,0.6030471318713644,0.6535362977871446,
 0.5936511015268932,0.3895661846535594,0.7475034502138027,0.8887787387475093,
 0.5096412221098852,0.5732798805155815,0.00799809213969016,
 0.28071010303381405,0.845157367682669,0.24185601971265724,0.8274155748450743,
 0.45516345009015347,0.7083731325817705,0.7056993388797597,0.6231820698519477,
 0.11151330364817458,0.3873778775281934,0.3532738678516817,0.4343227498732244,
 0.8772113845201237,0.5736697048174911,0.7399779440871197,0.06243433063553361,
 0.5307532700674512,0.8879811564555447,0.22559598926687208,
 0.013692355943761059,0.4185008588043402,0.14616394450037373,
 0.7410484564288496,0.1909228754483533,0.5388243645230668,0.29109936700858063,
 0.8920382707442713,0.2259384180096835,0.0934375603674964,0.5260320805048013,
 0.08892586062347774,0.619224225549449,0.6920487514346496,0.34612850930480576,
 0.7830324122648902,0.2494929262748221,0.6645042089001414,0.936823474090241,
 0.7047653771749364,0.9833798849408852,0.10540599108856707,
 0.0020359831199160405,0.46824319074272025,0.29830228858361185,
 0.5987678060337143,0.576088631181317,0.941464796034085,0.21467790805568998,
 0.018049759126633313,0.3885004219469512,0.9326259032785108,
 0.7264833712319377,0.2666421139713143,0.98559716122167,0.1646151676419062,
 0.6457097563381419,0.1254552934894474,0.23480109063080468,
 0.15802004662711566,0.38340545053837405,0.8671846303627053,
 0.36315414036454685,0.8985052786064582,0.5418397263361037,0.9389263576476378,
 0.3174047036119716,0.3975010664262545,0.2751409332566136,0.44952292231570723,
 0.607219948855303,0.5797045529033682,0.5866295577007439,0.5912899455757087,
 0.5099737566439815,0.37951188092074895,0.1630998769645694,0.4795631562729459,
 0.9296708768101573,0.01729517239924072,0.1975356730046013,0.7823676973151932,
 0.04237896485641601,0.5799356370471916,0.7386420354615096,
 0.25610074296134056,0.7049245665021973,0.8050850410888024,
 0.15763411771907299,0.060630724425112636,0.5810710127847942,
 0.8826377749477896,0.5960327123854476,0.5159791185696937,0.38133626339397564,
 0.719073865502845,0.42441055108264003,0.26349886364971165,0.8964148897229408,
 0.30541291038238116,0.9307915564785795,0.6082052956657263,0.9259983822815047,
 0.6145350826096405,0.21685392511916035,0.10565142758410029,
 0.6152762176839925,0.6570748464557,0.5514654642806468,0.0827152698425645,
 0.5730441252251863,0.4667239176311566,0.016376427708715613,
 0.5286664128566538,0.5093147865623546,0.7556368644844556,0.7657925279657881,
 0.9906991770810571,0.33003209249754994,0.39624568237681335,
 0.23787747578660534,0.6024805352880022,0.5135752189992343,0.9941385743667894,
 0.2924521739907062,0.8629932000207411,0.9336699368443269,0.7367882851323061,
 0.5466438089869404,0.12618082746643255,0.12063281659558578,
 0.5658283675425806,0.9029290375681737,0.44592671965871367,0.6755441707223033,
 0.3346272355561225,0.49585734632312783,0.8773681032709147,0.2527591548475101,
 0.39640961134896635,0.7876493789938954,0.7680000502044181,0.5851350219359656,
 0.5340275058666936,0.5759756977327667,0.6283249075978118,0.33083466310517773,
 0.21183203066089729,0.8559745502034763,0.7539918000917183,
 0.12240814622356089,0.10543245807934232,0.5056364525535121,
 0.19540603754473096,0.7444267210366106,0.30788542941334385,
 0.7858778829658205,0.36077849981260346,0.5071147068867062,
 0.47539436383257594,0.33007297016655046,0.38455898104702235,
 0.5620871437405816,0.58658684531146,0.4584363250169723,0.8359222743579863,
 0.7067366655259346,0.3228414854511594,0.42991550440796744,0.7300083070104446,
 0.8950892911316777,0.17411048257398087,0.17648863799267867,
 0.9098183866329421,0.07457877840350147,0.6482586264984188,0.9543353486949442,
 0.8226519160542698,0.773754901568453,0.13284460921346508,0.9499513589422757,
 0.8962434761534182,0.046065115108170906,0.8501990588409164,
 0.6836689385602936,0.06173812858573946,0.35807175128708446,
 0.5859762423275883,0.7349255513810391,0.2501060824431447,0.26365459749723563,
 0.6709257788727503,0.8302206091601486,0.6994174432313143,0.28878609532361155,
 0.8343828120765553,0.5475055469486124,0.5042889971154212,0.38112309167974046,
 0.34385139158132405,0.9945994691472653,0.21740373685900805,
 0.33293539720231746,0.42608972104318565,0.921645461165987,0.4724630177537199,
 0.7055206261456151,0.9026958966714044,0.4308696057970518,0.7962423533749531,
 0.9120883306026316,0.3541988888369648,0.4946951508169959,0.4101498716975478,
 0.8570747627832678,0.9068759736733066,0.6171215350008951,0.16977076843401073,
 0.09665725718413576,0.7745997775435843,0.10317570097676643,
 0.8278044750443811,0.3273520003205419,0.11446795977656998,0.4102844091510808,
 0.8343565601063858,0.8816909718562017,0.612435835998782,0.38771073304007897,
 0.3431222610274923,0.08232594238113866,0.25168119737916816,
 0.8780346905857984,0.9612139506443423,0.8969640093253368,0.11216515011990669,
 0.8758811036548715,0.4858687890015413,0.3816744631430514,0.10804001407145569,
 0.4741366968042616,0.9526207513568543,0.5983444759238636,0.49177899480438714,
 0.07800982747440399,0.3156367581170406,0.8348999616451231,0.5554369612554175,
 0.8455877067587558,0.12530660124942805,0.42149231378367624,
 0.9278886621547368,0.1984426926851347,0.2413930594821253,0.4388503216443036,
 0.44388078104524786,0.8361673022003906,0.039754131511391066,
 0.312110701951437,0.463117773834014,0.6371873929729066,0.34572132321433424,
 0.5971554949012418,0.2193408452469915,0.5376147080884071,0.8713524906119364,
 0.8443969075317735,0.10323592166488982,0.5987779038752957,0.5727725662235919,
 0.5943332687172117,0.8987540757302198,0.026440615663900102,
 0.37588550306569246,0.6392228223257195,0.8788571189196012,0.844492291655628,
 0.22860400393476077,0.683827090612749,0.4498482084388744,0.9513828561105594,
 0.4021166039777142,0.04144681994216115,0.934687701033614,0.60150433958535,
 0.5772353082391055,0.08815438737191417,0.022395678795640972,
 0.027691659220261733,0.01783351135411726,0.7506893123942093,
 0.05561322730891427,0.7514228566352794,0.6559841240009204,0.9299337941709583,
 0.6016217279900891,0.5063362271934562,0.5747526300069227,0.7969328148169506,
 0.3039243849944886,0.4640529288808548,0.4686009211197505,0.25099995689469723,
 0.6669002322900885,0.8773428869557409,0.7679090371544112,0.5451660921530581,
 0.5016262194251134,0.680038992754642,0.506158003592328,0.2426758655019854,
 0.6411841079001434,0.7414646520967952,0.6904460329555215,0.6272382858216771,
 0.3331884623625778,0.9930424487719784,0.6368675194653741,0.8302085592965666,
 0.9905792808261237,0.9922952353390607,0.6038424441372271,0.764793769113989,
 0.6019821217583806,0.7552291040318145,0.10371116273499315,0.3139064313909886,
 0.35508076184477444,0.08509840324326068,0.1025951356173126,
 0.38027402407244504,0.9976322565576117,0.6454736696599871,0.9164629891820779,
 0.7918172783647633,0.64266279025586,0.5343788908947528,0.8594866565551822,
 0.5828332833850568,0.411397028746167,0.004710644091902805,0.8707257739790574,
 0.5915005382031684,0.08662428877833217,0.6943901503617557,0.7917271470156323,
 0.49810827245478717,0.31966630812710006,0.6320766902732284,
 0.6529913745775947,0.5858241122669909,0.7573624523553715,0.07672608780659684,
 0.02163203158788518,0.4672124452353117,0.555937687998632,0.8609340873306129,
 0.6160835219051186,0.2967889776095689,0.6719911412754923,0.15488682278132637,
 0.12496572738542078,0.31131198908164015,0.06886671353829055,
 0.27320654525269483,0.9104385393436122,0.9158078480848023,0.6056380464003961,
 0.5178727301154957,0.5481230440783208,0.7588487564102602,0.37452842430313305,
 0.9395251721851766,0.32948710488488864,0.20161727642543437,
 0.32724063847007334,0.2946456312357263,0.9499551855918519,0.9156644665974545,
 0.21867129484340742,0.5370336087599045,0.7327092893122139,0.6804814620027,
 0.9900385670773542,0.9653112858374274,0.15972334496309304,0.8393258625660496,
 0.08496119855002626,0.9705302420424561,0.9980296870710988,0.7651673572287057,
 0.7055680165770701,0.08891845159539868,0.5513998308240166,0.1618347206196712,
 0.6769117508493473,0.9274578058129046,0.5161245838397789,0.6650303794751965,
 0.4729459814653272,0.17452309717717007,0.7798692121857791,
 0.31809268375965327,0.5760733234542963,0.06774121617636764,
 0.16630838023421657,0.9010960604317445,0.7179343994397596,0.5208873635910304,
 0.8594503692851091,0.6672516826540225,0.2925381018323261,0.47178265413060083,
 0.6075748713013926,0.5252367549188182,0.31401938902950643,0.435855408988981,
 0.4509317752641179,0.487850631937899,0.7375729197903697,0.6516182634738681,
 0.9801806886521338,0.7040776447349171,0.7938593853591505,0.15920099568620383,
 0.47556641779796627,0.7346298484334047,0.41872298999202384,
 0.2677705577690087,0.7723921472003618,0.3455851522818909,0.25493880957027104,
 0.03907277624728844,0.41492987683237803,0.32277306754761326,
 0.013511138459130323,0.3584191180033811,0.4855935984070947,
 0.5420155351072192,0.5428527313786848,0.710811888528573,0.9823854995940824,
 0.5163100771335519,0.3595829363699583,0.734668582455615,0.0500789365627643,
 0.7099836864229329,0.30392141089377867,0.6070522007767274,
 0.49826479328917517,0.5678198345637946,0.18122282683171909,
 0.16487263925228468,0.09426234480861317,0.21011323795533654,
 0.20797656712128498,0.8211425887914234,0.08216662088350757,
 0.18570144347137807,0.20479148912548717,0.8854875030353875,0.974758293062156,
 0.25337452790455417,0.34149781115105804,0.11419506315905714,
 0.5217068839319644,0.6898292836942498,0.7287568422588341,0.2798535113072802,
 0.027161582859193234,0.9004656009394474,0.04450686264532078,
 0.6374933868289422,0.07737986502719951,0.2690916017735113,
 0.24350969189692329,0.46124944018874725,0.47605693627772794,
 0.9523919678005466,0.05920643228710987,0.5379520441004886,0.7340882710402213,
 0.3993277033850141,0.7475791055526344,0.9828940957959449,0.6621603267948123,
 0.9252843640713033,0.6307989815236126,0.1873571469239108,0.9074098171358544,
 0.9818389719119343,0.4408332060038951,0.06483148976835551,0.6334928808103774,
 0.6664769340770752,0.03009501791125002,0.23534499802322284,
 0.14706047167361924,0.5619459453143222,0.6266497252783636,0.9761826283793574,
 0.716846407091863,0.7437600472067166,0.9311418328429941,0.8451809330659737,
 0.8502967103858453,0.05227503388497545,0.31767112999442093,
 0.21333601448639894,0.5689960452251892,0.7465454175358647,0.4935839210133871,
 0.9624874335681334,0.15807374346342895,0.09342802138320272,
 0.7843713245196017,0.014239529232814685,0.4392263756744983,
 0.1852667074214681,0.8747644797494734,0.6798747846665792,0.7906851785439739,
 0.9004300553898343,0.5108179052757416,0.8359994553850547,0.9555581526922645,
 0.9742530195822576,0.019630806603722628,0.48404289166694914,
 0.20468220069717646,0.1879559673494332,0.8725573403472558,
 0.15930334840778948,0.05822134951314162,0.3003061302348984,
 0.40646575004083685,0.13667038742804927,0.07232033790511805,
 0.13538336629106562,0.8510910501741443,0.8266762907626835,0.8726832171036195,
 0.583201293961269,0.007760275954656115,0.9729940282286886,
 0.24515342134193974,0.031968560406818725,0.2584196219146879,
 0.971663168284245,0.936872906918755,0.8998128808572519,0.043035380084107566,
 0.5703098236478462,0.4500606348691908,0.4237662413020614,0.4237101690130065,
 0.8755110684259745,0.9972678273632951,0.926447544448441,0.6611435387515873,
 0.4454320924388321,0.33946654910126495,0.10406718628631773,
 0.9500759160064074,0.32256423231269327,0.8600811092962921,
 0.15743719317121663,0.2616865914262978,0.9783381261255797,
 0.42149008297313606,0.621823640254231,0.4361899125516334,0.7286418015112772,
 0.9593952566717197,0.7482632760843813,0.9676516459056778,0.3325769577595654,
 0.8039376506305019,0.4850310867106241,0.8077161127673627,0.8801308236963133,
 0.2747732739528731,0.40695801517340924,0.8263176203865404,
 0.16780581504641734,0.6782366386209051,0.7439108348484156,0.5897223197365825,
 0.43187925382214876,0.870418959732654,0.6982761823258137,0.03771796999958199,
 0.08396544646397386,0.2863254004120622,0.8348343742712795,0.8523257315896139,
 0.9892141579837246,0.8491716263912358,0.9430527709585685,0.6563436755283102,
 0.10619179215562025,0.3258434786276647,0.7290625065059625,
 0.18028441394404315,0.8890111482907525,0.5886578541403105,0.5115177003804031,
 0.9219831431771379,0.6840629789890147,0.12485165948562793,0.7905814590877114,
 0.853040699773252,0.05502697218898556,0.5652651529194872,0.7575081935633337,
 0.5711801429847724,0.16149743473225486,0.15551316389108383,
 0.3297497486785339,0.02551941006660241,0.9076501586506515,
 0.08730748929055432,0.4137893433336389,0.6625124281270158,0.8216262854824375,
 0.889324033618799,0.8735007486787019,0.15925416472529208,0.699260480745037,
 0.40138720521956794,0.9530694375416539,0.2858276872048239,0.524544957361502,
 0.9511925315537293,0.9831062644442072,0.7733382734694993,0.64777307441582,
 0.9362580024378583,0.693838414843067,0.3217426898309341,0.8628585925596624,
 0.07504108424766553,0.9072079997964698,0.39046790289368194,
 0.30903635876391145,0.4614967848962528,0.8160727228492535,0.6554171262163976,
 0.6919475476883771,0.6913084499463342,0.6584139807335949,0.9290973599767179,
 0.6816510345487813,0.16425689815493283,0.7619040755363451,
 0.15937108969314218,0.5024462951714074,0.6020621193175152,0.6952145035231416,
 0.49818924975572854,0.5756517840983224,0.5038964911804588,0.1780564151533247,
 0.49047996114938197,0.6309602706651701,0.10970442778279565,
 0.43701829383024393,0.7991288855254145,0.5554026113217677,0.4293102357423544,
 0.7748245146720375,0.8229448248664811,0.3307659265412801,0.4448303984060966,
 0.9411996687065753,0.5546369543427805,0.6407048232958094,0.03927603088879417,
 0.7076406699117126,0.4462080303779332,0.8367082808918062,0.09792654103416287,
 0.16276209844783307,0.08689523596674009,0.9334975860263746,
 0.9431195180246369,0.8271074632758569,0.8793147267335595,0.19198835534006897,
 0.8048224320100933,0.5001903587161971,0.9467296255178507,0.20791633488974037,
 0.20337647812719473,0.7414429222056357,0.2361579926543067,0.8760120786811404,
 0.710355790936883,0.2544634459193291,0.6377162707695389,0.7031771406242537,
 0.7520786396268152,0.06734177225759241,0.5040901304255028,0.6824696106242429,
 0.9867088449709317,0.7611760413042216,0.2571652463199725,0.1871246510270438,
 0.3790560561895535,0.3035321945784907,0.21682457747088302,0.4289340305040249,
 0.522637603832557,0.898274274028992,0.013196220312041707,0.4620947658310539,
 0.27312610697829687,0.5143251289096855,0.386030169579044,0.6470825668867414,
 0.7081815659087246,0.19262788647924411,0.9184681833204312,0.911138348116719,
 0.2742587102413965,0.6606868514331481,0.27597752371893414,0.9682199539716433,
 0.9421254046576949,0.3464257304872205,0.0591289752820191,0.2764526815079371,
 0.8883717993303747,0.5591540410702373,0.5290938382584922,0.35727947973654817,
 0.03316782149542341,0.7468053645795913,0.22134030892361667,0.366941286464862,
 0.4426174770850443,0.5615369387711387,0.5878668699036387,0.26836225104857925,
 0.9968456888266805,0.1713175431435553,0.21953346939239937,0.946879934715961,
 0.571207800742442,0.697099231829731,0.9516974106748766,0.08703591527623089,
 0.7150198510093628,0.4966204689761865,0.7655370720368558,0.09914493420564185,
 0.5854789312337665,0.6574375180451526,0.6649235393970734,0.7270114061334215,
 0.10637943884186551,0.2091357239209639,0.4947623000812612,0.7268870599866952,
 0.9001365712053834,0.07399793606619276,0.636357997679277,0.29229368297678004,
 0.7147061750914881,0.34707033416672384,0.21314504596203765,
 0.32254011966025775,0.5478425588668172,0.8239425456120666,
 0.40599158707551386,0.8654843475447521,0.6171383221716696,0.9209500158869407,
 0.9392481443006555,0.4940350370462524,0.058274200822544975,
 0.7395182482414091,0.9248113614536246,0.4983181642672798,0.8994328280628326,
 0.10752996040685692,0.050521904540330675,0.7626058410990175,
 0.14047159930125475,0.7472658973757349,0.5162875155139555,0.3296244325142461,
 0.12280883709482426,0.015291417481541614,0.7372905708139319,
 0.44940922557801977,0.9933368218788958,0.8047976826099713,0.8284292387188286,
 0.03498749355352271,0.8497966585832599,0.19101779290479382,
 0.10228350585606671,0.17434735103174448,0.7317616347483251,
 0.45444949411519997,0.584276803310843,0.13861869490086465,
 0.014973084293709471,0.6505866217321382,0.8390872786631266,0.566549218204635,
 0.5844136526735766,0.23033611536103793,0.07967435373469345,
 0.5346109379867884,0.649952531503611,0.8633874295258345,0.3129690072037953,
 0.9782578401254057,0.2835357903304464,0.3972304751068031,0.45014885468151733,
 0.3154434491432435,0.763327740474068,0.41626842747965065,0.9625502510343923,
 0.43231117649732975,0.18237208193184928,0.2173063662932182,
 0.0925688926912347,0.4633770092181376,0.39432876878168777,
 0.01502147121931452,0.7395584192470789,0.3865650672562796,0.54487563305619,
 0.12631602591908353,0.3232064393162858,0.214286018630368,0.32346968246012986,
 0.6846808410917443,0.9865365451987174,0.5429266850604249,0.7048886906115005,
 0.4742297656354475,0.3881854261093708,0.43825619645895875,0.6131223914227635,
 0.6568117110938674,0.7464424514150819,0.3848470287497673,0.553721924694007,
 0.13547411928248043,0.13819197938871064,0.8401491054077332,
 0.9573028251377599,0.8583997688799996,0.5062769061101109,0.5777136936205116,
 0.1503562965392068,0.12634980320539058,0.47727160234621124,
 0.7085243215275369,0.5470047824294796,0.18215491626569058,0.5750917357249116,
 0.975590059845734,0.9477819919469151,0.8384184139405337,0.4172554430532218,
 0.6912839708449237,0.37572955923086393,0.971647816421022,0.4436514430712061,
 0.06244973033597756,0.9929993521023669,0.5642275158644487,0.6835197285910596,
 0.791439356372635,0.9120448993390464,0.37204162402652763,0.7286412767873313,
 0.937328164417907,0.11488106779344254,0.8376675234794008,0.6176964912630982,
 0.4602327755608082,0.5436582340411042,0.9815068279211541,0.16608519348735729,
 0.02254988367245203,0.4419169680951963,0.21251103551378114,
 0.9934369119009757,0.23856367354439767,0.9681880402784924,0.5826472302153421,
 0.9360360525985044,0.6744060514574447,0.8690301389343198,0.9911525696423819,
 0.8993876069601174,0.72762404948245,0.6462990215277692,0.7391567187790875,
 0.817793493921838,0.31981829623175306,0.9284767493488264,0.968886199944631,
 0.16619351269294314,0.9667818921707044,0.7592753067170006,0.282004431952978,
 0.9593720941414955,0.8861350589519759,0.27113559987610714,
 0.14487550488370393,0.22852905902443466,0.6443831062951723,
 0.6265664581953553,0.857394752666645,0.31982847927176594,0.22951309724941193,
 0.512376968764511,0.3012418752444196,0.20960671771883643,0.6538908751654862,
 0.7348840939920672,0.6057427792675848,0.8938497670230754,0.49794506787396653,
 0.15730340777929008,0.9110027582143561,0.63038318498489,0.5310603647070692,
 0.4238256824989558,0.167926592483855,0.5181531878361539,0.9113926622792381,
 0.7592762407066191,0.19798896787756004,0.979999109562357,0.14107617776356096,
 0.1761590675168777,0.4653900205034107,0.46569417696239335,0.9601079036634099,
 0.34079388137675004,0.08799604121908833,0.30188644280103616,
 0.2544068889523099,0.10567882866561651,0.6892248897639626,0.680457019862513,
 0.6393382510214025,0.34561036490424357,0.6535132543620232,0.7410302228692744,
 0.8999883966904723,0.3322936999898387,0.9934645492669775,0.10607839545960285,
 0.14490720912340793,0.40831047432614465,0.6341816454551638,
 0.6592050152195994,0.43731950091607685,0.1809403922560605,
 0.19693711571253447,0.6447865566141255,0.5443134070590882,0.9415744967483693,
 0.9150425931810049,0.4209300920597956,0.5885401229856901,0.1214364187914505,
 0.878339077041462,0.10573932311654854,0.6380699648328105,0.06656562014664946,
 0.8110157560553216,0.2849538372046658,0.2753350444375212,0.4737341667393151,
 0.9190928378397162,0.4525094086738588,0.3963515319895038,0.16552186805607438,
 0.6021538886585264,0.5361082077452027,0.4743063142880275,0.641891298058325,
 0.5923531859646308,0.0108472698719293,0.315259895398872,0.03136294025284436,
 0.146231016942318,0.36561315296954633,0.32812501693771556,
 0.26720077840481227,0.4993747168023438,0.26310055963142054,
 0.16613102087723775,0.7446647119001155,0.4093157298586143,0.5937267147062553,
 0.14779427364078201,0.35549758581162916,0.8790218452854108,
 0.48318003034931756,0.80120576242994,0.5962717770261257,0.8827738257306905,
 0.3299945531872044,0.4536497603484012,0.8419853836186123,0.21216289362138196,
 0.32421472690648523,0.3969718513297905,0.15218118357768506,
 0.40935834410687455,0.09912393043897405,0.35187274597906903,
 0.3928925155738269,0.3635307418632586,0.6175874338272,0.005956166138071484,
 0.4480245789280898,0.5111136554898692,0.9846511040701889,0.9429168985977245,
 0.07237477926041558,0.5964843126230088,0.9911890378973474,0.871117081691414,
 0.5789577088823177,0.4088296705351009,0.31479818721355524,0.4583438354581356,
 0.11801669439340945,0.14821712333183823,0.785716633495322,0.3302460107110333,
 0.47692894806079345,0.15863749497564783,0.34738609251099795,
 0.5797237940397821,0.4587235714008886,0.45339610018220056,0.9202150634067082,
 0.8940823421419154,0.09906611657744757,0.9923091092304333,0.8537366731414798,
 0.18439069114524342,0.3162674748380089,0.8659322491867092,0.2214311921057519,
 0.3456382005560814,0.12470808490902585,0.9253037404088414,0.7601004617393203,
 0.33739370049382134,0.11713562380615927,0.7611504800487698,
 0.8605753656618857,0.6463034911912766,0.6631325689156379,0.12184852281564718,
 0.0326221810462578,0.24380037000869903,0.7969670869212266,0.0482776750733972,
 0.5063810442771607,0.9625704102194171,0.2970684615304716,0.22895532859994006,
 0.9363007002940036,0.5289169094430257,0.34685445245528257,
 0.47649156209022303,0.2867833479526465,0.5837935632582935,0.7599334344560016,
 0.16507696167831398,0.23660637045421318,0.23994705132357885,
 0.9408856715551726,0.7159470589988272,0.20750163263935717,0.5746773353073524,
 0.2973613083983815,0.4388368668347925,0.7485063526746413,0.8833164920870683,
 0.9620702356067173,0.7425233670224352,0.6993422971130323,0.6768945602505813,
 0.5124546295969052,0.07881863497743202,0.7963105610495833,0.8007196388470725,
 0.3048045569107095,0.02703671450757328,0.36237645012585695,
 0.8128715327868963,0.4762446638493202,0.4523828243251927,
 0.020931248531851132,0.45746993538831926,0.8227877513292682,
 0.9964501474925201,0.3752976256566164,0.3167827658327489,0.6240259001565245,
 0.9279182315605544,0.0797353807714174,0.40974834043406705,0.5565268994296009,
 0.9492617353407325,0.523995828186661,0.48237761383966427,0.6647120394411894,
 0.6914479138439704,0.22857205722547602,0.5866067817264885,0.655385737252903,
 0.032424150928840545,0.8852644027839782,0.3690163331174675,
 0.42685814161150115,0.22954577721525582,0.7914213512698007,
 0.026148904821611363,0.8512689299205127,0.2546677470418254,
 0.4985797622088304,0.3011607795105291,0.6038363797660733,0.35537012980120697,
 0.4994010308787743,0.7396740543613134,0.802128787880387,0.8411439160833916,
 0.5806366644428733,0.44946064701793753,0.7666830555024539,0.742768227943583,
 0.016905683932987414,0.615118487895453,0.2689484353645105,0.8543123627899389,
 0.28145229889108114,0.4130255836111105,0.27652162319638085,
 0.028328042498750095,0.2527608423705284,0.1591383955151776,
 0.8525196912236583,0.5834734576449652,0.7206742619129953,0.8917187222019106,
 0.6135001274712717,0.43279463592732803,0.8694439473461973,
 0.12644921840246048,0.05639849776954775,0.19391058418096307,
 0.930567279439413,0.2252031926137955,0.6218307414426844,0.26625257537467717,
 0.8894581992144397,0.9379081916080988,0.9453719444409584,0.282368764744807,
 0.07508793492227206,0.4012990764632488,0.2536066781724332,0.9648726758816393,
 0.3295130136733664,0.8646467470278542,0.9057877310387318,0.08013940344691561,
 0.2593726377904868,0.10959220538829983,0.25830767889135975,0.964232439840967,
 0.6058789901129584,0.059592898737330646,0.05078425593124891,
 0.5969853380617957,0.39344710165489116,0.17854979920073888,
 0.09432332471012728,0.9289219169294546,0.10007582570397855,
 0.030883712132525054,0.26103622851454733,0.0036164098873588957,
 0.7965172543499222,0.7714756249783702,0.7781575463940016,0.8449839997860853,
 0.4035846344749391,0.5815848493182133,0.621640585391735,0.8741911085391003,
 0.019004015707100197,0.7687992972370045,0.2057833792703998,
 0.4767646078783403,0.30504733432501574,0.29748524688433997,0.707691969909443,
 0.7218939533234575,0.9188923343481388,0.19622852780183653,
 0.37652555842378144,0.39784277001218293,0.36476572022724096,
 0.5574733125740071,0.30585134713891293,0.3072303783691197,0.9163460286033486,
 0.6723679769617406,0.5980593041144843,0.34788636462124045,0.5772284680298977,
 0.9008354672557172,0.4361545101694382,0.5722718395541729,0.6639897391728984,
 0.46846882365091125,0.5271392553789739,0.7744762148520906,0.6712960135529829,
 0.2190463460071529,0.18530589805775344,0.07236485066651599,0.16613625960847,
 0.07610332926299745,0.7720921926504021,0.6368609047110139,0.6543246726624639,
 0.9662990321488409,0.6182784971513345,0.05910166181175691,
 0.05364244931449935,0.5854490566025831,0.7848337395949949,0.9294848256566483,
 0.5245754915404615,0.5365976081753426,0.2097140311764315,0.22916194505583598,
 0.7069087864715977,0.23269522901810236,0.45508668549156595,0.183178564580309,
 0.9217669899433226,0.6038988558709056,0.8655053133544285,0.7013455317900054,
 0.885194400339652,0.6805770196551271,0.7334067160274707,0.11709584633815362,
 0.046828330843438515,0.2407004884841415,0.2768180016636239,
 0.13013527916090195,0.43980075807105923,0.5440464771034891,
 0.7801579321449313,0.5135325893712647,0.8468698622680517,0.5488148022799858,
 0.415456779467159,0.4487346380083119,0.28440999349937457,0.9203982638180448,
 0.43473663687208486,0.7973607013816965,0.7206831119218636,0.6338198569709123,
 0.0622514987998144,0.5393865375027262,0.517774328898699,0.310343890680875,
 0.43407780387138395,0.7087209134009803,0.1850839999752979,
 0.06043287270543851,0.8409503568329194,0.9944617472828862,0.1207626008382282,
 0.026686937700885927,0.4608828998390976,0.8507357938873178,
 0.027964935475905173,0.11728432615163176,0.5625873321466249,
 0.6324259730844299,0.8421432283661588,0.2673161672959117,0.8269675393146497,
 0.9565130654556696,0.6030841822165235,0.4219808065043017,0.4198135088704995,
 0.8547900713787575,0.7723972116694239,0.1482258601137565,0.01120461830540509,
 0.45943556477398695,0.41448332859784187,0.7145895486893211,
 0.2604385744558536,0.693847589524291,0.1878662399120392,0.06931557331374882,
 0.7742448523630853,0.807150794088644,0.5102907281883274,0.6806583446689912,
 0.4207253578909125,0.39154500969988826,0.32939856593283323,
 0.4230101029806591,0.812862008090795,0.6728494880049454,0.06888774375640239,
 0.2958838847155043,0.04340628992779705,0.6107541422671062,0.8015185789750968,
 0.5378859933106007,0.23901192455978804,0.5740184571344639,0.7921619183698739,
 0.9976237225687954,0.8567463900206785,0.36290270201240205,0.6260200847699019,
 0.713884481075149,0.12080251495682737,0.33981320618836575,0.706396552750719,
 0.9488353546226117,0.7648490389885532,0.27678705330673103,0.8269617264402481,
 0.9983919278543313,0.4336369291535679,0.14314786385857947,
 0.10054853031910926,0.14325908391785525,0.34228572212614683,
 0.8927547416135789,0.5125485124208644,0.6380144202803979,0.9358289023764103,
 0.28979643917218134,0.6350380392386114,0.9527543310772125,0.1283991386363219,
 0.8645526664563545,0.6087236486602803,0.2796095921815931,0.4120479931249659,
 0.961764710964586,0.36463449458448816,0.4991447400029998,0.6112950777874859,
 0.6862637502120926,0.43897206300642133,0.3740916282207132,
 0.30368328494694996,0.1965797148464743,0.6896550194432916,0.3840540028070063,
 0.7862642996188314,0.25344134333346635,0.42088993943620223,0.665886770673524,
 0.7580048546636414,0.7120946472381372,0.9948016326979708,0.19775655781193302,
 0.7933288298229337,0.8899148617018027,0.015194662470170983,
 0.3302257670503488,0.5799578533935402,0.13490729379119426,0.6843364837545497,
 0.15814691520238733,0.8472814727087208,0.1538937262835749,
 0.10087836319800092,0.9448011336730342,0.2778148294841156,
 0.34553692027930594,0.16076745651138347,0.9808393330957239,
 0.6042226151677479,0.45238188352381026,0.7054865399849756,0.1995664017180373,
 0.5822470056778436,0.1307681630642411,0.3366090035275515,0.31476775569693805,
 0.7361053801944245,0.08463582676546211,0.1810163597586425,0.6034113882968848,
 0.9283600552896513,0.810834517847214,0.9672216232143988,0.22303178803840729,
 0.7836523151397018,0.6635403521520177,0.9825975392373016,0.09204596696088063,
 0.8390137171109526,0.9734023648403092,0.7817791521946112,0.9296811682649095,
 0.1015991874223473,0.5128427076107714,0.817772215091916,0.41284700922488926,
 0.46385497002808207,0.24230108047421217,0.8373739423567494,
 0.4285099254340915,0.7961756380454241,0.08981619970856491,
 0.19762710952542317,0.40520542238778123,0.4843108663098715,
 0.6333307194027326,0.6631374025026143,0.702088056992474,0.3297691701107178,
 0.1542067707172643,0.717090046162572,0.832144894874688,0.34203360404019856,
 0.3993241606662824,0.5245086089216975,0.07728483998886848,0.5228415921209962,
 0.94609352999072,0.2597146041314662,0.9670052430782079,0.9194392929113993,
 0.3458485613918335,0.8159399868704261,0.8684224202455194,0.9960328386722427,
 0.2034304274779124,0.49930242474654696,0.9178634870064765,0.4095379209916602,
 0.663699278264191,0.2055986207188103,0.16371479325518845,0.222233890919598,
 0.985061997608996,0.09759095418007036,0.6368952522478952,0.7036699868302179,
 0.8247225132867557,0.4937490278635419,0.25708792384317025,0.4653671004744785,
 0.8700590758259381,0.09184401548822074,0.7276814337135329,0.9289366720890453,
 0.04525698037120962,0.38955645606785594,0.378471204139167,
 0.41061415564879744,0.2070791298422746,0.097325970852854,0.8924130728248476,
 0.49347766098815693,0.977373775689532,0.4335580680835236,0.12157560563601111,
 0.6762082042942221,0.4892272597696895,0.685503925709945,0.22741853124532174,
 0.03103052829413122,0.7986302301354459,0.6406621050297667,0.8754870585357781,
 0.9353207075909712,0.5104069438317373,0.1341342220346322,0.4724628066958535,
 0.5328376356467763,0.18907109200568772,0.46168153976361115,
 0.8600931069250793,0.005728005845053508,0.03897080174561285,
 0.6717298679090511,0.11159664470696706,0.9616009785376851,0.9761670018365011,
 0.457675874873272,0.6042675953596313,0.5314928223703221,0.7371716769036232,
 0.36911492174161886,0.40743120742702343,0.9409667417715799,
 0.8733786097341331,0.17628487266098802,0.01813581783067142,
 0.6875703263704068,0.3209732655840195,0.14562420950192767,0.6841685621901258,
 0.8024244821628076,0.9271334309748468,0.024434792102452363,
 0.4638740589027761,0.7994200285060707,0.013525863992501996,
 0.0339762695285335,0.5690039753768192,0.4430290918439801,0.9743738149233796,
 0.7541543407302829,0.2483393078437549,0.0331975769566375,0.781444140488887,
 0.344174818108967,0.4593076082482087,0.14497360434612794,0.22971345481477645,
 0.5972913177324357,0.9866074231316365,0.6041494945875252,0.8264193792542307,
 0.06748326556763085,0.07454815578808272,0.8010163859580962,
 0.9462808872185693,0.2506682914528,0.7006827050207529,0.11863061582544598,
 0.5993495167232188,0.17742144995404407,0.8093585492736193,0.8067541123040183,
 0.2689354297158414,0.2913344048941465,0.13141848300633152,
 0.43348571693730775,0.5511510274762307,0.615189831642797,0.783391324534277,
 0.709224753090688,0.3508423793475193,0.8830774441953869,0.6003429429126758,
 0.8159288257946203,0.1602413025944217,0.9491762571413166,
 0.030392999513263685,0.16788945766903351,0.9330613417717742,
 0.5194749127909574,0.17977696565815693,0.7716020474742057,0.4767367291185687,
 0.7770544693693823,0.9246707362487214,0.9839787190208126,0.8806156485896646,
 0.9956776511585582,0.46709457904234264,0.2674586771690709,
 0.49042667037504906,0.6919091366860057,0.7901940543417162,0.2648016706811468,
 0.9831289377660487,0.9679664063221883,0.5216340118767121,0.09710623593272849,
 0.9903739718256674,0.6466044038542916,0.42860818740505036,
 0.22784929760296657,0.9807281571782791,0.051487915750320745,
 0.3090089856877578,0.08601988698712393,0.6210089363954838,0.3837146354057229,
 0.06871365408925068,0.20052861826510648,0.5841176607868718,
 0.9505509162590678,0.7574679861323113,0.5573321075893882,0.2688160266363151,
 0.9465033462524122,0.5268107275242533,0.1266429150640579,0.7756377802240336,
 0.8203156683712315,0.25115367376743647,0.9880212823756407,0.8313721041182499,
 0.28839155631247115,0.42532521033073667,0.22674373177737406,
 0.26496550086802306,0.644985027824583,0.39338411390380323,
 0.014350839457252684,0.0809483077599078,0.30824940544301693,
 0.014301510689323726,0.8455108264259878,0.9661770631838205,
 0.1711758997517091,0.7102338249951226,0.6134908607159599,0.6722513195833015,
 0.5119264466430073,0.37668408170122225,0.9843213190569521,0.7910575176410802,
 0.8038461027853564,0.911466086864815,0.6778956887146417,0.7849288832675755,
 0.010434413432246603,0.9695615292299706,0.9504409999414983,
 0.5502150707469254,0.4444681617897459,0.11759918338053998,
 0.25783632460270944,0.6198747881406221,0.5222076803063325,
 0.36733034190861913,0.4422644100316885,0.9043395054138882,0.9516315523851115,
 0.10246148426616297,0.529509615629076,0.09886837754546918,0.7695257094333164,
 0.32311359426942043,0.5411923003200583,0.9615393323559865,0.7285403561452904,
 0.1038977396690206,0.05009907727595997,0.5853144741785661,
 0.31673513706535183,0.961965713057308,0.28969526448675875,
 0.21554827579218871,0.7954996270870932,0.2949398588483786,
 0.22122905765932033,0.7260541375773465,0.5890784179288768,
 0.02473556987864911,0.030390338720808097,0.01620623011087996,
 0.07267974472934002,0.46772617887175827,0.6258216809307285,
 0.7371889620927038,0.31849044455702435,0.7088123711047118,0.8587918698414461,
 0.3695866129897487,0.7348234207460603,0.9991483133076908,0.2799923278459463,
 0.9879868558922643,0.5468219148579336,0.07007295852887063,
 0.44793109942317655,0.9361840867011961,0.5129427387125922,
 0.11004202144153064,0.0024538612173239382,0.938082483896545,
 0.22037220205760477,0.5415912287378633,0.6294843178996863,0.6057056089647672,
 0.8378766119980029,0.419534772566869,0.6821684406168919,0.41041982598049254,
 0.6044489506074407,0.054164877207185436,0.9173205419504412,
 0.7123403149872217,0.40397448104288414,0.3715184328215362,0.6788501669624758,
 0.19945318526602918,0.5859865824280455,0.3142468026531279,0.2873907110477969,
 0.9384517684415462,0.7248563423605499,0.7092824745492514,0.666694231552436,
 0.7585301223612794,0.48616128015719073,0.07537757893027563,
 0.6213762462689922,0.7349373617069567,0.10847921698987217,
 0.13317805681530737,0.9499195397485437,0.8267928427310854,0.6387531507562806,
 0.06585139126670247,0.15689835650901285,0.475445467665764,
 0.16466881697402058,0.9589087898740447,0.9318257461781152,0.818638739795969,
 0.504252958203105,0.5971106943897471,0.9376197184578628,0.3409949513684508,
 0.596983679759371,0.6350039293806494,0.22123084084140854,0.29760230715392044,
 0.9788152211378943,0.554274829833055,0.42894896798791904,0.937100139791091,
 0.2540259521394591,0.8603442468492125,0.21315234631037128,0.4965779001706261,
 0.5049679982742725,0.4898830569460705,0.8587066547393574,0.5959301961210456,
 0.9928485298768704,0.6379757667750099,0.9059898806398344,0.7638840203115386,
 0.24029478262746495,0.5100041739569953,0.031144513761863446,
 0.6253179775245165,0.8771268051146672,0.6881912932405647,0.13179270180466707,
 0.7369572389220537,0.5188389226281427,0.619724990735862,0.9604979940423835,
 0.7126075091932207,0.7036160854510884,0.9101743947626008,0.6268902338560673,
 0.4376066374104991,0.2788640264174962,0.2789774133992968,0.5111282090944287,
 0.6128257837184197,0.7320889555252328,0.6109869627096814,0.16453879004096872,
 0.3671321187116682,0.016152759379540194,0.03503324201336089,
 0.8428552858951769,0.3799438966546318,0.8301406702523448,0.8829129274152467,
 0.5803000394049044,0.09388960509001576,0.7345476467568549,
 0.14719690464336221,0.48766816088395504,0.8244584627827762,
 0.32556813886723424,0.6663689120387402,0.7493796169528933,0.704078108252648,
 0.48298154572677776,0.6087321318012837,0.4297609638021447,0.8209562677834061,
 0.8633356935426921,0.7427855702400687,0.3903467929438842,0.4986791788607362,
 0.9379840005103849,0.9093545417270145,0.5784269692938747,0.758736600379723,
 0.42987002298716237,0.1687348766751624,0.44614789700161006,
 0.18851213409775758,0.8262020368421353,0.8561050990856388,0.6776989816010456,
 0.32479215593677446,0.43048589531475157,0.9787515727273369,
 0.5333653455677847,0.3769933819849808,0.4134861892015397,0.3202782179610496,
 0.1046090240821066,0.37670995787007255,0.8929518806967849,0.7769631959911035,
 0.16126580822704495,0.7922002574642039,0.22017383093903897,
 0.13898433912131758,0.43408769013193693,0.17363169334347017,
 0.9928122582281079,0.29199324232200186,0.19919565080747392,0.957022108942436,
 0.03755523248147219,0.625223918346736,0.7725762467254719,0.96652304884951,
 0.705877790615934,0.9402281664140824,0.7952313774128552,0.4597230828181881,
 0.395465487970331,0.3365943139983909,0.30167373733678504,0.7261162205112667,
 0.5549229396136048,0.7292497489215335,0.17263413284274143,0.8602759623157457,
 0.13036116137463016,0.11258863926765228,0.6878595928829582,
 0.2154688903737959,0.35937321554610513,0.3282140831557371,
 0.11088730885304476,0.8245346998886,0.19664619388830495,0.5469689135020873,
 0.16061887125290075,0.6149046903544314,0.3537660939173779,0.5815660867832005,
 0.695710905347494,0.034104220325755,0.19570024661833163,0.8232097078203414,
 0.3247061244904388,0.1069688509065102,0.39155078469512616,0.7401722350712457,
 0.39218286157217075,0.9328944591948615,0.658086944047372,0.6774310602339307,
 0.5851395877768001,0.39772376165456336,0.7805067500806772,0.9660541939131523,
 0.8766042199170526,0.7739932198622272,0.39873580029745426,0.9329055902360375,
 0.3594727702923801,0.04648225867785594,0.4407356058990861,0.5698278087691884,
 0.5785809702143844,0.05824085801308543,0.23408216731893317,
 0.07289800214140729,0.06993351478804521,0.3425537235377466,
 0.8566377026695429,0.7187779018112872,0.17840076737849653,0.9598050326347869,
 0.8544462422260716,0.824712084248516,0.09064923564509952,0.0480518415303367,
 0.6697871536695186,0.3104566571033949,0.6331162304875745,0.07874538486399951,
 0.6848106004200764,0.823726520338459,0.21079430552684753,0.4665174630861343,
 0.19694761904804636,0.8644994943088796,0.9200898128462514,0.5779214913148601,
 0.806310909955999,0.3274548000300712,0.12110664196827903,0.4840682044421676,
 0.7773747890420066,0.7079823060795947,0.14866189434132315,0.2781364449865529,
 0.1418545852068276,0.047402494599386946,0.41939878297980004,
 0.10733441156489332,0.7436234948825127,0.30043369261884656,
 0.5791970389470489,0.994195484632328,0.38831264887200345,0.6340543752972129,
 0.8053798813819649,0.3306851757709097,0.73401905995199,0.22027271307452456,
 0.6089899607511694,0.08664804489026501,0.31055661057151074,
 0.9119835731994125,0.7943791833841467,0.8371157651189004,0.3148808280051245,
 0.31016260269729534,0.9734237973763333,0.8162171400270657,0.6482591674008437,
 0.5511801408240816,0.8732920006020114,0.6124492628521294,0.9314768652567603,
 0.34672581716481166,0.17474432415624763,0.8484089337721505,
 0.6135679346578593,0.8097187158505701,0.8218546883368845,0.8195342245674877,
 0.6697133447051968,0.7739595803457919,0.2632218817349523,0.6854563784882313,
 0.875117925056896,0.6285926550673475,0.896644482838589,0.30344638515068845,
 0.7772289304979314,0.7513450926275002,0.6523827637281254,0.5277485587091639,
 0.547602088721022,0.46118165460266125,0.7676563911716388,0.5310135930075003,
 0.5165398604251243,0.9042489207488847,0.20136946907223296,0.5223272759215174,
 0.5759997919881965,0.24317298289858247,0.9827569536702334,0.4347406867701953,
 0.8988281629736099,0.5012923458826876,0.8985259917008748,0.17018290829354688,
 0.9590221939531843,0.15642672313568318,0.8109766255741597,
 0.01762255236831778,0.5286826593734177,0.4064709481649683,0.3970328138632917,
 0.40166538454865974,0.431003776339544,0.43470692815211,0.6608157236561313,
 0.7487369321819222,0.369370277305483,0.2774255645986634,0.46941260090030046,
 0.46569651213564145,0.759373137245066,0.012143956606437856,
 0.8328804931036526,0.5730550601215032,0.34381228019991295,0.4160653683135258,
 0.6703102946461781,0.30477757428461527,0.2742330833850849,
 0.015567158193170783,0.34257260656343247,0.5389646269848242,
 0.9113746330637611,0.999035433801833,0.519494971550729,0.5521073737279542,
 0.968806765790871,0.642571463056197,0.3334096873628367,0.18662654679085044,
 0.5663281091409407,0.6844392432860996,0.9883937869974647,
 0.025760003405218113,0.9809063850525931,0.037490511802364956,
 0.6335107941516029,0.10138108217549024,0.2862057242045014,0.2257751025919179,
 0.32615642873516193,0.40297822314574516,0.4573516842514532,
 0.7246337744714129,0.059306487595418345,0.5052885155736994,
 0.1746760049811471,0.8852366552721074,0.15631728704158976,0.6428249301130399,
 0.049480096952486896,0.8758289048123129,0.8074602888905897,
 0.2771377901951114,0.7125552253348093,0.2947753112795999,0.8341269652095122,
 0.9079226802270366,0.4777427075932975,0.2955226561751605,0.3246762119348906,
 0.952008812655582,0.8835661014848795,0.4509125594735409,0.7571738951178933,
 0.3630890300397398,0.5036505591424796,0.0671745463864929,0.6555415277633015,
 0.49929933957125205,0.5855205479447609,0.8431213538376265,
 0.36086958890056675,0.8420619676415855,0.06491769619857557,
 0.4768899235462756,0.6098962011174697,0.18939610287560738,0.7308052557701767,
 0.6035343810037096,0.13743962924359732,0.7847367494508163,0.8715567618490748,
 0.9301512303150208,0.6629088014817013,0.3905861742508072,0.6165355840085233,
 0.07213923035321623,0.14998751362932206,0.33424570629645955,
 0.6871104114112078,0.04240767026730452,0.5664133795151681,0.3179421353950286,
 0.6635145468183312,0.8372304601849723,0.016385537324907773,
 0.3154581254563499,0.569374090566748,0.18424267444251408,0.665413276015975,
 0.261857008681167,0.03806579408785593,0.5403214977696357,0.6721749334014876,
 0.07742128278023874,0.08266552879406652,0.14953346927641376,
 0.1722700378066101,0.8316470832827726,0.8510717931510412,0.6578013225663414,
 0.43773213917789133,0.19366153587618296,0.4821041298176637,
 0.06030923593423698,0.19608107669264918,0.09985492599187729,
 0.22728127495618722,0.2803874874302572,0.7790717933160396,
 0.15037202025179797,0.25029966512922075,0.22742836195754446,
 0.8893680840792305,0.7776232237974189,0.47040284349760064,0.2287515342055605,
 0.6096088386155664,0.3467799126953819,0.43593726686850953,
 0.19434726559784865,0.800298167546837,0.3579019697011705,0.23043507450547873,
 0.7594668816437847,0.06289283775603838,0.6653804360289455,0.867750547911417,
 0.0682396582096626,0.7070159889161733,0.5436907450467524,0.19460257692804273,
 0.7275160964536002,0.9566372043067648,0.7761217740292022,0.22564466385769877,
 0.8140722714304688,0.12472857246006264,0.06194861292635023,
 0.6493529710376977,0.5990499708388797,0.47299467178309995,0.7955470622443341,
 0.21833308329247958,0.5305016379075543,0.7252867091866347,
 0.41773086640081536,0.3808843558826255,0.32011510851004565,
 0.5858970425714094,0.8605118533344932,0.6583896167526471,0.7269432362418886,
 0.0037038189225622187,0.2069439285527328,0.5393048283345578,
 0.04919452764026888,0.8441359886775428,0.6797089150027413,
 0.19702873897733408,0.5917222799934223,0.06480532311039056,
 0.1732509248343148,0.12570170391543634,0.1716839424826112,0.8731954415046909,
 0.45873871739677874,0.46981373541245053,0.7638428916303003,
 0.08524937657887566,0.5163617005857644,0.30709277816896463,
 0.6946634653990652,0.342901018322913,0.9286839633056205,0.05488349514910462,
 0.5326295943804839,0.5963433312280875,0.5697210443971004,0.2616856112748722,
 0.4549692579233382,0.668319689993333,0.7086429915524407,0.19794735320120138,
 0.24055152890947984,0.5667424399397448,0.8502180270475762,0.7202758658427064,
 0.5984128006970615,0.656208755553423,0.6802383478531933,0.2294923093380361,
 0.909996541367491,0.7224258273807486,0.02119246031126787,0.7886569847962268,
 0.1839287760815277,0.21250295420440213,0.5585619949818013,0.911851052179635,
 0.726008608492367,0.036241445502859126,0.43724355604976894,0.381016682876022,
 0.7048179152923911,0.7288829095544246,0.41831553583313075,0.612050551709949,
 0.18263121373363378,0.3283664944326211,0.9067630427173549,0.9104802716433383,
 0.8243552209052104,0.5878825612623944,0.949600732265689,0.5247070661001928,
 0.9808971992224658,0.14327008173139044,0.0553763812789102,
 0.057715071517239314,0.23330121261251824,0.07125791791131919,
 0.6360944189082693,0.03823262915084591,0.5728699073935207,0.3314732192266695,
 0.35543167988309377,0.388466826806959,0.5486774703772032,0.10038144570817087,
 0.6915430960685733,0.67745061540631,0.8385812864411061,0.536179991177815,
 0.10789212051029484,0.7744886922667716,0.2351575194416491,0.872627720745032,
 0.4182583998630136,0.5517917261688398,0.015139902998417165,
 0.30401485065581235,0.5368385371536405,0.785423237845447,0.220008861647922,
 0.08703738715441556,0.6563418446523844,0.04244397888811924,
 0.30716336280211076,0.4217198134035025,0.32719832532293325,
 0.17918531875902433,0.566157463514021,0.05755012829044781,0.7282787834921155,
 0.5342537732784334,0.19594479938546483,0.5184606483867782,0.5560715439541409,
 0.015356281888990075,0.08741967001157203,0.20429543038178588,
 0.8121720793862538,0.19637073935623073,0.7882231038046563,0.6178977469254096,
 0.8131038812711265,0.873027814240102,0.818401298054372,0.7965128085554327,
 0.3870662667027931,0.41364051239875443,0.5080749367995839,0.5715587174935567,
 0.2544091601516907,0.15014329011367544,0.8870914262518352,0.386334677033613,
 0.26033173810858656,0.926891394631933,0.8417579527533685,0.13206977650259533,
 0.744603249949669,0.2640716398078433,0.8746302643824939,0.4546819384517097,
 0.4592793548650487,0.569731327941879,0.21115110327315,0.674152871485557,
 0.3270274758076783,0.8851151056285211,0.9301166753564639,0.9279372150758765,
 0.3524889705805865,0.810994995327401,0.7681740319455986,0.018164184621049184,
 0.8083202447004237,0.9168910443053345,0.8382082400607267,0.5200814586814955,
 0.28888415314363736,0.3564663587881075,0.7934329459987512,
 0.11244707739006621,0.710784584756208,0.12642065995614193,0.6897737424063985,
 0.49209432695672384,0.5490092799075117,0.324887164352633,0.35041734389921075,
 0.8435134612583963,0.6442010331150902,0.13978992667266976,0.8675016876895844,
 0.179103236576877,0.7689612425252532,0.1867315102734458,0.85019274583535,
 0.6768718528837043,0.4013956234856131,0.12927328926845294,
 0.15417661182849718,0.32232916268449774,0.3539741132265485,
 0.8859587685535466,0.9632206217260647,0.4751998526164465,0.4528806289106051,
 0.20780942392922475,0.9969276166679244,0.8382411445811123,
 0.43445259015135185,0.22939390205788135,0.5110083534886457,
 0.39460187454426343,0.24843564285884412,0.028596696930318077,
 0.9134596117402928,0.9989715600155993,0.12024967152140242,0.5857809531965544,
 0.8778982070516032,0.674459969542531,0.7554945855773662,0.711894185740398,
 0.7122335270862074,0.9559759386790562,0.5482074819415496,0.26721136809843293,
 0.4654280335031228,0.3975337200280611,0.09981749723321365,
 0.11228575755773273,0.9557000726608884,0.5304579100112669,0.5756878058621974,
 0.11131479966743851,0.258147733283997,0.4669598541251566,0.37351984811397,
 0.6773031733171129,0.7811793481950781,0.7603473069636759,0.7040455323525695,
 0.7802450034493068,0.8724934199283854,0.9882133078479115,0.4177978406433627,
 0.01082640999354334,0.588981001238806,0.6024796251485521,0.3011104267700744,
 0.9041847567407314,0.6125135155387285,0.8736238310468428,0.1612622635168558,
 0.7666589813998224,0.9046674569625024,0.7734131174318942,0.27239092247923946,
 0.35616950427749194,0.6482948841534162,0.7437379189022342,0.3356423968600398,
 0.9299355568533858,0.5833015073786976,0.34004164078346966,0.9890763443664385,
 0.9248553621134712,0.7408348125545466,0.4745135595294746,0.17729335311500982,
 0.12959708585143226,0.8991197631084943,0.18504014949057157,
 0.49065334327627963,0.791712357394902,0.4876897559811564,0.3450103539327861,
 0.4040661794755962,0.6743842094089976,0.5631501393378259,0.6972431442440679,
 0.1301687136078642,0.39790491231517877,0.5975994701759948,
 0.006504295772659097,0.6189431754056286,0.7109059338827525,
 0.6289042943639814,0.9621346946688023,0.4366826082628257,0.2898633948914706,
 0.20580804210712378,0.8160819639175002,0.4735781344438402,0.5644718355735685,
 0.29716470306908094,0.9079583297257766,0.873613831237151,0.236256072461683,
 0.6825312688283202,0.4683534105882273,0.10757067246269592,
 0.013716437581088314,0.6353551403406679,0.7634540968774188,
 0.3067854210724039,0.355918987883149,0.4527085892602287,0.5882020492699449,
 0.19855434702389096,0.13071131680086323,0.9743299890073924,
 0.9738857402607995,0.08266582585016091,0.49755870065684493,
 0.5786280145352946,0.47363018891519393,0.367068156764939,0.04801774266370429,
 0.7821634811008636,0.9878367790185862,0.6354532168029694,0.9995143313151514,
 0.4500061120796379,0.5578861524699011,0.3305877221949409,0.34431839507332374,
 0.06488513693482068,0.7593268195160164,0.14207758394869252,
 0.4182139258727495,0.10783068604273238,0.7361972158715131,
 0.19893785873582337,0.5134517075511538,0.29462387171596127,
 0.2993635213361505,0.1925513995200231,0.6836255101008725,0.6544874188528298,
 0.726382082328769,0.5988251456330319,0.4869611400426912,0.37609312160976105,
 0.46240246400261753,0.28663847155837685,0.3168794334538241,
 0.9462890574334989,0.9473302097626078,0.8217511959432908,0.9264832012918256,
 0.1862011184256639,0.4324610257209809,0.7675178480774404,0.5704950533659626,
 0.8461742354054577,0.7520390401825906,0.7628247219886037,0.2882815239235508,
 0.4189493299374841,0.6339479433399183,0.7182929344133798,0.9404340132298912,
 0.39890228003209316,0.25782102434333043,0.9331462438346718,0.53090483640619,
 0.009999081204125626,0.008495558414291948,0.5833679999077835,
 0.5546762245164464,0.6028951030720078,0.5246716789268379,0.7433734404386521,
 0.2037683424088691,0.955617394114822,0.2192710828861396,0.1339341854192606,
 0.1334755818152995,0.019965964571912465,0.8144865915954526,
 0.4156366979281172,0.5791411066321976,0.3478452243979736,0.460335104048194,
 0.9365758563150564,0.07323129447727372,0.8325396399116467,0.9890115544368597,
 0.2013348847156654,0.34838101491325235,0.03344788520041542,
 0.37117503167766497,0.49208099549501627,0.8236099990255116,
 0.5860675227970595,0.5074982826329553,0.9511888395798629,0.9103103110348014,
 0.5823093535168236,0.5448513397466215,0.03577450844095997,0.2615684865546877,
 0.6163992965037363,0.8322182710267024,0.015638309749321122,
 0.4617393553816689,0.1194359267297026,0.3347577416630273,0.19980547281396166,
 0.4362728158717073,0.22101688131760766,0.6433281377475529,0.826889128182781,
 0.5160408187911388,0.8675795521511054,0.7826151658481313,0.7991372217423486,
 0.4685524683857386,0.53901834305495,0.8983151866679544,0.4925594854679185,
 0.5614524024776801,0.75130817647274,0.9605313096640409,0.46087956516728923,
 0.3346315835052943,0.17659881696668756,0.6153583902902028,
 0.11074391762149571,0.8277566382223208,0.7950851689023328,0.6439332215030164,
 0.48906843456870264,0.9071371146557718,0.5660837933805092,
 0.30042875233157285,0.0034120374893163996,0.7471902596943614,
 0.37668641210664067,0.9970429730808064,0.7768196703387406,
 0.34673383576239714,0.4993236517675821,0.8544418291785157,0.7081654583689947,
 0.5072935632536657,0.2188000467957557,0.27263347786481196,0.5314447234379504,
 0.8726033507972419,0.5387395759229209,0.5572992016301109,0.8938201770304758,
 0.762149944297124,0.1681452182041291,0.4581911493387786,0.2132532913499865,
 0.8675106003465175,0.9571054369964122,0.2447517373003265,0.6244966227528804,
 0.8817081075524099,0.5360413089303185,0.9600609607684524,0.04238229144426364,
 0.9659675591055333,0.44332973741428805,0.5232519011208984,0.7049570288498402,
 0.10293507788709089,0.05696707827365932,0.4938159207695543,
 0.15724823046945913,0.0869261150741274,0.1589151591909438,
 0.13717454826155573,0.8659347006757212,0.23174038019104293,
 0.49384125802163314,0.5054529291802958,0.09170248453813501,
 0.3923987021065306,0.8330324991504233,0.8434704791381272,0.38816545739717767,
 0.07706580845694544,0.6988146112736604,0.7660585544897187,
 0.10205927925435065,0.4518014453285757,0.2707387503851616,
 0.26244496110377624,0.8014620067588585,0.16749809323949816,
 0.6886735482107312,0.20733447263047533,5.349704832309232e-4,
 0.8904367153436776,0.8452803935267611,0.062292706850185176,
 0.5814454539195435,0.6789244065678339,0.314640193965795,0.6242277075298598,
 0.16452753043513935,0.8282709993164311,0.6338257605461451,0.4263384752417262,
 0.33185837716178945,0.5320925280938594,0.6330033206067559,0.0581727870067178,
 0.9693290834374692,0.7052135846383262,0.7904398286993238,0.5560568263355568,
 5.192542201599704e-4,0.6584629831491258,0.2769738540413462,
 0.2983834570914916,0.8366285911725888,0.021183918401718627,
 0.3104679058131654,0.17816665464922343,0.8646028983280667,0.7666143983350138,
 0.029916864227844342,0.6084395823275058,0.6661533525105487,
 0.3423290591593573,0.6845548035744444,0.3138273181720127,0.5337353395582329,
 0.6332724151810157,0.054604780777759876,0.9182688478778472,
 0.12569506712906753,0.11625879304102482,0.23221561735055862,
 0.0011015877944622332,0.08455215378635728,0.8018527992539297,
 0.18358167624911392,0.5105269857028976,0.05250059559332387,
 0.6935659065686806,0.14144907919802663,0.6499620217492325,0.9525728721925755,
 0.4799150332419996,0.2208113266316759,0.4690184478301933,0.3402351141971469,
 0.5366494417490998,0.7065514723496639,0.7443393214471207,0.8511591829706823,
 0.10746538645956127,0.08461648572680913,0.7187580705097163,
 0.8321426466406979,0.6885003881077931,0.810344873520713,0.7456264010063828,
 0.22756446359672422,0.08271677453308501,0.49844076122702996,
 0.6669918260471145,0.7730701512388473,0.48667152396831304,0.8489351204291508,
 0.181642304940676,0.37084208356181214,0.15390011725534758,0.547866489073303,
 0.4967456384327712,0.13366821655419492,0.23422276818655807,
 0.39119986809076357,0.6794214200439863,0.39257993446706063,
 0.9538504772707302,0.6324114717475491,0.23283938151107542,0.9490997584133095,
 0.8781934449999916,0.8454121726998767,0.5884731912518157,0.3649650524316209,
 0.7742059532176279,0.539945328611278,0.20564642812126577,0.7732666614431282,
 0.9871548111923574,0.11125815311358433,0.9105171677472477,0.4723037593657152,
 0.64452392413227,0.12473500076189992,0.08889552474890261,0.06972347853455996,
 0.22017194749460178,0.879596185039506,0.5527124661662165,0.6051701020340458,
 0.5330150325081574,0.4881583857144447,0.9301649525702127,0.23974643431825216,
 0.6505120722148903,0.9369768584350843,0.4765860152674759,0.9116402841262421,
 0.20529260980432362,0.8529521882102533,0.7499125028448159,0.2635200683377028,
 0.9410037071585577,0.8609237457038731,0.7180887491828754,0.8503500234397645,
 0.614348754763876,0.891032466716889,0.02560467862238147,0.8862410890221067,
 0.9985349808907318,0.5421999892495837,0.8041957713211328,0.30382654359131034,
 0.07460014585766639,0.8705891738348261,0.3536249576931151,
 0.21197917793463006,0.6195636525146732,0.3037966987183802,0.5938596180168956,
 0.05989484119113919,0.331969202375742,0.7869300110075197,0.8285588973500835,
 0.11812728542707795,0.010814050334656877,0.5866302631720092,
 0.8009946370874156,0.22258321436212358,0.6940278135663356,
 0.24424926946369452,0.8690878730354076,0.6024530491311543,0.3482171045320126,
 0.23385105136165973,0.5179850421051077,0.5473059462514523,
 0.043298499802217116,0.6511703153002936,0.5516063233514212,
 0.9290765801686918,0.08172512708922353,0.36630274368087534,
 0.5463096328472032,0.335729684954743,0.10575845997144129,0.35719371478460915,
 0.585478110980789,0.8071529589935547,0.09521413036395765,0.43229066922792214,
 0.5050411389705898,0.08900772550380909,0.6409147615238578,
 0.44712057318349685,0.65369639218422,0.297310391667178,0.24801671522534996,
 0.8389792792080497,0.7691563676002144,0.329336872375523,0.14965508419284124,
 0.7899046762876504,0.3712793893238836,0.6572653423699844,0.3079710409052987,
 0.3582158488950711,0.08532598029013361,0.485858105960964,0.5157486546073465,
 0.9275043967399575,0.29643438338425776,0.017420999815167648,
 0.5644387119276257,0.9571106341366866,0.4059742677790872,0.6510259204560788,
 0.1646971231005222,0.04866754189671496,0.8879765468722038,0.7971378116392256,
 0.7186080158299928,0.3409952218444924,0.0327978533396287,0.6006570627729164,
 0.5044090460236121,0.3785402938965572,0.9361699738353669,0.8898540378165926,
 0.45019090225972325,0.39752867005145576,0.2172918542387129,
 0.15699250996227798,0.6408496423546051,0.44051466994588884,
 0.8771931385690556,0.9400406476355404,0.4658246175736348,0.4459399583297585,
 0.8323088879297315,0.54037361098095,0.8987887042217793,0.9267307565356977,
 0.06097579561158106,0.5766758303380336,0.14416664155290349,
 0.15649042793208734,0.9187727649508654,0.24043933770912607,
 0.3929053895515674,0.8677847278207955,0.4055432766802409,0.01430209231742774,
 0.9629959739738275,0.8000419026508513,0.11353529047240785,0.9207740371486558,
 0.7385683572024908,0.020783659743096705,0.8535295678179736,
 0.9895114025895695,0.8729096993401309,0.9231423133483844,0.45441341462117524,
 0.21169447425214616,0.8561585537027155,0.48295464378328923,
 0.5942109859643391,0.8262736854283987,0.6866430055745477,0.4372839677622914,
 0.8618670365596363,0.041246879983291906,0.885946056029467,0.701822540384733,
 0.20403368153349666,0.12316184252590123,0.7692922764170378,
 0.5365964608202489,0.15896591381703584,0.8592768001276212,
 0.08011560270001161,0.7998006932239171,0.005089788068356804,
 0.011356452733284783,0.9980414392913656,0.42230099128776777,
 0.24410127830786998,0.5032072239809666,0.3365987529036377,0.3527243270824185,
 0.06488410529701227,0.9394940476694797,0.7048932854015436,0.2627754447719196,
 0.5341644571034851,0.01876958233775816,0.8506631408096853,0.8311083098651983,
 0.35412732367453614,0.71291905433561,0.19141752307736215,0.5411171412096909,
 0.056582498744655374,0.8018218837362117,0.6095460151972574,0.917816466905247,
 0.7389791242251091,0.0723545634065963,0.7915638905024639,0.8804528747555959,
 0.9996861909513018,0.3755856401579528,0.17195228369566806,0.7000656198394916,
 0.3867072796045614,0.8294015234663059,0.8586540919788093,0.1489617918289191,
 0.5223081102307914,0.08574059484383134,0.7288480283800604,0.6322228707069038,
 0.4602039834709395,0.48866559925004016,0.3344489229454809,0.8204739426553065,
 0.08519018935875566,0.5863091813874368,0.3124975374680887,0.6774685825190724,
 0.13565446233442024,0.5496445291026459,0.602012612826836,0.5552696892683093,
 0.6390748554899829,0.18062908655038656,0.8320521816480781,0.9224715347685184,
 0.3886011687410469,0.3112536345715934,0.8965301361139353,0.08501985211031304,
 0.9619087697528929,0.6975542342326754,0.7260384063647152,0.14249753975918722,
 0.1752988025454153,0.26279226055850646,0.5177233600114689,
 0.32414302579508303,0.09064135767220571,0.2217765768690093,
 0.8248734655787954,0.6602723669615578,0.5583374802265275,0.8987419097185305,
 0.3691936643933946,0.9867077012784816,0.7425052286163224,0.5487073640655071,
 0.36695005070130793,0.9866093112218145,0.11075879929021037,
 0.5656868673261106,0.310560102569472,0.4655498132999354,0.9585361183031758,
 0.6663346689623291,0.19568170305027854,0.46809224458524124,
 0.2052306464848448,0.7388500260047075,0.6069745941480743,0.8480613035124258,
 0.5675505730489071,0.024048302856134685,0.11225051322736745,
 0.6303802203702505,0.7961594737774496,0.09467058463257139,
 0.025875324710418646,0.3256847691566416,0.9612566902883651,
 0.7101001302820913,0.7056620121309699,0.8953325997018393,0.2548072181564289,
 0.5940246815119651,0.40739250277595807,0.24706837903953716,
 0.8917641674336803,0.17916194149101505,0.2277335266821905,0.800405715586632,
 0.07839993708721094,0.9303563820716766,0.1662338701690408,
 0.47754057499702185,0.8784793576161108,0.5784605793337799,0.6394483550395063,
 0.695235812843907,0.8235324519119427,0.15728525365051194,0.3661871722489598,
 0.33582916126210505,0.7669540801966509,0.6336868049696139,
 0.10652042530344108,0.6073908852768002,0.36882572897005306,0.706038102624282,
 0.5245432894544613,0.54654829101105,0.30793278894610543,0.522744468431723,
 0.29236425079438677,0.4672673744593261,0.9054961358728971,0.9889031457167645,
 0.6396495616876351,0.5727162777697943,0.7120892434757006,0.5594707426694271,
 0.07588838394675101,0.6877840380609295,0.31983850434491057,
 0.08534620264766535,0.08251228444023861,0.8785715956573988,
 0.021093345880140002,0.6232117577256762,0.7317626112232449,
 0.08395646397448098,0.3961823756046855,0.30890327569301357,
 0.8943061073663456,0.05151976778497791,0.5929068132909164,0.9767271998785128,
 0.8924907200543253,0.899929129434294,0.4038492440264603,0.5399848762938154,
 0.4120469670954523,0.4631586944552508,0.6061927978458557,0.6274547156087953,
 0.9256956070608118,0.058814539493273976,0.03237207104914097,
 0.03703951405051342,0.22178049932099742,0.9484553967026729,
 0.45312501476059375,0.35877653127094367,0.8490001700519737,
 0.5336592091998524,0.7433743580122364,0.8363307657677004,
 0.052878400470270814,0.7009628477953707,0.009540302465107398,
 0.13606813833853848,0.5472842600397357,0.5812048428585238,0.8499898851378066,
 0.7199744670540934,0.014970271503791488,0.43483170366296464,
 0.9867909639747887,0.6288873135369836,0.016133385254906485,
 0.13323050590916574,0.10784264533717391,0.5426625037768935,
 0.032843778080693276,0.7952464539227031,0.08462289140074675,
 0.984844044267279,0.5647970171721225,0.2379833533094392,0.06710574190890117,
 0.19753307167346656,0.17447864601467988,0.10602369760153496,
 0.6506432906865901,0.5463467957110625,0.740573735726437,0.6640061598440253,
 0.39900121472436345,0.0712517868462097,0.3062350101159337,0.5728691762347847,
 0.7010745514753629,0.7232200828821596,0.17031497971969378,0.1045201665128429,
 0.3796101362981422,0.6465752929281149,0.7309431407372664,0.48318922709174184,
 0.748987353736178,0.3413815145250373,0.2911302747294531,0.7724277619385078,
 0.9582727993997885,0.5105160045086875,0.31963157106979223,
 0.26861001473282897,0.8867602686660556,0.2553724681092362,0.4603710693042645,
 0.9083247881659157,0.05691843688056997,0.909193748111852,0.06393756559944164,
 0.3719625475014994,0.4796988352958991,0.7778222127482503,0.8870428899761098,
 0.6422794826209536,0.2804677850426418,0.12822855386971488,0.7791682618565102,
 0.9463255279626763,0.37015449296357406,0.8164221115961386,0.9479728760414515,
 0.2258815000810498,0.45251450391548986,0.5487951839134201,0.6470755855337522,
 0.07009704896467706,0.8270200604933768,0.8502289927161698,0.7130517509439648,
 0.9134769077344402,0.8568540899458329,0.5895555712421435,0.7522590673295579,
 0.1733764396180093,0.8525476032820261,0.48411865978563706,0.5868007320548375,
 0.6037569072183994,0.9649756014953867,0.9405505980350561,0.628401377050865,
 0.4645424173315926,0.7973870615233489,0.5836161479932769,0.6243635899468133,
 0.46572538957611553,0.2882899711192676,0.8463299389024217,0.4530516576468414,
 0.8761148650667189,0.11351064665829713,0.2911705606710392,0.3412252731913742,
 0.7644367283018366,0.7519314220561959,0.8280585748180942,0.2412442539267441,
 0.6456679429845403,0.4036333047588925,0.7002974178337398,0.6502378053725838,
 0.5212874071511869,0.08145020069829201,0.8710236360090522,0.5567551467286045,
 0.42120238870295945,0.17009403184742666,0.49057818115436014,
 0.9865908154883956,0.2010488926622518,0.6756509944522787,0.07393785511091022,
 0.9827220376423189,0.12647246544091473,0.4912496163452653,
 0.16372099666261852,0.8563020375948083,0.004341234016824735,
 0.6943144836977277,0.08086611711509584,0.4834150539042613,0.2933230111415962,
 0.5453343032994672,0.8213541859797895,0.795907114200971,0.1023901630242312,
 0.1273398847868359,0.15323540483160247,0.4988402892089999,
 0.22144139599420865,0.6589417323489216,0.18817954500254885,
 0.8144369678985407,0.8267282579016932,0.38524396080593193,0.5397471579929068,
 0.6724359151713052,0.8007721757085928,0.6115728877656782,0.2557321094907081,
 0.21196666483851911,0.650985857490461,0.7444053269512846,0.6045451487193843,
 0.6287799734298503,0.07524822354902816,0.02029985149833935,
 0.9159974315805894,0.32994136531104223,0.390025481351138,0.9844035378405613,
 0.8827255916302286,0.3283192412026741,0.444166174598953,0.7789139203422137,
 0.06930770860989754,0.81307019184289,0.7414442387297544,0.12355236580517315,
 0.40643964791151166,0.24049236296572496,0.4947977534387211,
 0.9244886375450649,0.04658750929829991,0.6808738554743757,0.988426106964702,
 0.07160646823642303,0.07234991889643161,0.943911027932835,0.5265617819807257,
 0.09348297963173213,0.21223970183034035,0.5076040733352682,
 0.6936369413804829,0.14132097221964535,0.5873396855777533,
 0.13163178684484378,0.3544298041966044,0.6866198000146961,0.511872607010472,
 0.2683767285029016,0.8083240126199815,0.2904735495025782,0.07610043879747086,
 0.8733461234353135,0.49306227777393785,0.6389001880610137,0.7092771761563581,
 0.5514324650680203,0.9431843931157672,0.5197047353440891,0.8021181143111982,
 0.7157709466245624,0.40976775316751424,0.5337135181613353,0.5636749227329781,
 0.32951758692193855,0.03184593893320953,0.9768984070049664,
 0.06405108896864486,0.928790655340707,0.8384945566615137,
 0.043763592773681026,0.16159536540883512,0.04834820004652107,
 0.7232626297621582,0.615045620690421,0.06756307833145103,0.17199648642586063,
 0.986020873289615,0.09232771052599542,0.9717248164113267,0.7007928612354725,
 0.6786111954546092,0.5304601433581864,0.37900916337279245,
 0.048053726062912694,0.11382518291016308,0.47456380424636113,
 0.3380404683396283,0.9067062424672679,0.4272022109696172,0.9412520789028472,
 0.41311071853426395,0.06905939749093115,0.1711766204210785,
 0.2201560108263365,0.999836857365976,0.9207161406864214,0.1841700768315817,
 0.6461116096406216,0.050361779923284944,0.4689631299718754,
 0.19227891660421137,0.01300845562332209,0.007305190719917087,
 0.4574186187916188,0.17978799858691752,0.10095226272597646,
 0.24151876023741714,0.9659837089777137,0.7363006658498255,0.8124649365692498,
 0.7887136310635716,0.6500176453397373,0.3911624987527854,0.12604882030284614,
 0.13050658632280687,0.4215725981555203,0.28600175379034254,
 0.3356136906919729,0.40130190122966725,0.21385196828274733,
 0.9611818118339835,0.007322904479930892,0.30740493000494307,
 0.8250761963570299,0.6793691466433001,0.5310951559345513,0.8694415104683418,
 0.19819422939635922,0.7168288491504843,0.9541595738379643,0.4091358592674246,
 0.3284337886596794,0.8852278026347803,0.745403227185884,0.36973905854430267,
 0.742279425683507,0.1309925891913637,0.8195487185075093,0.7625241192073224,
 0.20385002295683696,0.0832106785289477,0.5360692911757914,0.4029772393340677,
 0.7871051875760829,0.47789769136955706,0.8484227433461371,0.7737243663879417,
 0.032515862124909756,0.36017966266873847,0.8582558948920024,
 0.7848443143702493,0.3177222465342382,0.35768037624778526,0.7059208437426571,
 0.3231908588496534,0.20679853546462168,0.6497898348700096,0.5289612947025805,
 0.13448538097582274,0.6559059157052352,0.44769544645276294,0.928180879598163,
 0.78188963713969,0.8837510361952875,0.6495305929982582,0.30896460806703796,
 0.03630728203718181,0.7207916016545636,0.15355204175204562,
 0.9546365626245338,0.2206043476962002,0.7724724110727839,0.6838971648005977,
 0.8542214845002898,0.2699034429284598,0.8386938516783269,0.20069358471705634,
 0.2283457549068929,0.964074064925446,0.9492865529747104,0.7891095105286827,
 0.22951170601529225,0.23657900971970114,0.06457483909123107,
 0.4828442269138604,0.07355227737142522,0.010435318076666311,
 0.9889398105662496,0.5100782931861269,0.4666161113875682,0.15344530614537988,
 0.25225547283042704,0.029168744884388587,0.6352887267267078,
 0.2529040175875342,0.24861302145203068,0.12634754899156164,
 0.38537604049902696,0.9941547212313349,0.18503788943726218,
 0.5910569417855207,0.8413483267124255,0.4394534317080233,0.6626908262937634,
 0.8824745878060845,0.3769720843945177,0.8661364172538426,0.40937587003175346,
 0.882271888467012,0.1673225282184133,0.08208009975660069,0.32059768527509025,
 0.17098283298026995,0.3055196741132501,0.9376477884280904,
 0.30713837084986684,0.974876550426579,0.830873160977613,0.14915331355630324,
 0.2020313511460412,0.47810313132824844,0.5270792079442759,0.7328814193293881,
 0.6949910607982519,0.937722616066298,0.6133917865533837,0.2532690071667174,
 0.9865988481431194,0.09296975275054764,0.26008464723048064,
 0.035742457270054984,0.5628558198813807,0.37748427898779224,
 0.8332624600305354,0.8750691543100009,0.3581410380524611,0.699917280443759,
 0.9576296630334263,0.07075059566541586,0.9151843539242842,0.4727615748216295,
 0.5900334380829688,0.48299389008919413,0.14023293847414875,
 0.1195738020448811,0.7905108460075064,0.6446709256346541,0.44164063998981673,
 0.39876099616479976,0.5464607328939416,0.7231501822926094,0.717957076599619,
 0.012789695613249163,0.5846499741029577,0.4354231960815097,0.838994871735342,
 0.6797887341582214,0.7403225768481837,0.2596546554078146,0.526579676847532,
 0.2980810832647286,0.3035592074726513,0.285570854318105,0.27861132461873483,
 0.0320982211260703,0.3222761843291678,0.3995454584873994,0.3465782653060916,
 0.2438970810401645,0.1637606440855246,0.9973680530747351,
 0.020106734952066274,0.636933904035374,0.5435282108188515,0.363733594819318,
 0.009906076690972099,0.8650589174165907,0.3313514391702751,
 0.8514692093802566,0.1141711026047697,0.6376126162455588,0.30591042621571896,
 0.6362543488839001,0.25425818720927906,0.4100237063075858,0.2197331311755032,
 0.4358328844625834,0.690170725991753,0.5656200101073766,0.2851318554133131,
 0.38468721021440166,0.19371427728953938,0.08465623665496302,
 0.5414164294559793,0.4058387367183228,0.6726390495161628,0.2028938423587313,
 0.4348565380599899,0.5893553940551887,0.6019641313813263,0.9182170462305339,
 0.16152212627301543,0.6394160776451278,0.6505608028340444,0.2909311526302537,
 0.6782449541617578,0.3025701760764905,0.2605764518164938,0.6537124411438969,
 0.18379845024669328,0.6899215847985926,0.14116033838400155,
 0.1577332741603379,0.7261122014792668,0.48096002825051576,0.7577379896776771,
 0.5911639527146267,0.9366137903234917,0.23133030385414466,0.5127634196694254,
 0.5215211868678734,0.2749671388827828,0.7897644129091302,0.5855823464499129,
 0.9774625370414317,0.7724015626791488,0.15949064280156067,0.7078399010350183,
 0.23835980994439876,0.696805858380746,0.5295453452967007,0.14782824303278286,
 0.8509353065518331,0.06744168790717464,0.7267935387481854,
 0.19864418804551576,0.8767147906898477,0.9159788753929085,
 0.08620437980458273,0.35628703461671896,0.03562550158664213,
 0.31532053290352047,0.24123071964293197,0.9931763908081942,
 0.36559202593363904,0.5637605401960752,0.8540394611007834,
 0.06423486693188596,0.8061497444373309,0.6049436194151049,
 0.16157750766352408,0.39160451996860224,0.5054371027026631,
 0.5482110700425412,0.44827978494650744,0.5473593101673688,0.5360900889098801,
 0.9561923563871507,0.8370812934230105,0.42627073078576827,0.9427833799794504,
 0.1069001227517743,0.7042432204216209,0.39382258520386915,0.7609233356066712,
 0.5998918903779089,0.3146473844951946,0.4477742139152706,0.21281222350661,
 0.15810212112024136,0.3288024718408613,0.2541061070403431,
 0.46722529811729774,0.46567662029066414,0.250364439887109,0.5344694348496186,
 0.5783901395353037,0.707903671015513,0.6884628476404813,0.8727475099128894,
 0.4698916876422675,0.22591389005122564,0.08078635039804538,
 0.14664141173720757,0.6288365937064472,0.0026937034403526905,
 0.44013035406108003,0.677870697968224,0.425756076527118,0.36195524034962046,
 0.9645256545570252,0.3697485389806038,0.919696928895868,0.3784298688251321,
 0.21003739907603247,0.37350233524145493,0.6482995486169999,
 0.19803759345983807,0.6239915527318664,0.9112511372476579,0.619936032860697,
 0.8854400056921912,0.968437658250126,0.13751199559129734,0.9513832005078486,
 0.9006445095569807,0.2372824836988614,0.4699525664704304,0.9583268333029846,
 0.628189896896672,0.6863429952295993,0.38389110383124947,0.7154157149607693,
 0.7354573000929222,0.1812868354821946,0.04226420017647303,
 0.08981710420587474,0.5091269820448123,0.6766071981262736,0.1569544200575561,
 0.4037279283347265,0.7144348411092065,0.38673413571178283,0.4101250605598713,
 0.7665128996078246,0.6029954623033489,0.8503460154887387,0.14461595657827908,
 0.7113288761245592,0.47184162410020813,0.3026216896794667,
 0.44389126716857996,0.3039561449308059,0.7897054164534014,0.05385522286819,
 0.29068203400885007,0.11997388753157778,0.36472662241335296,
 0.41170606939410004,0.15638185116705428,0.5555662102990627,
 0.28863668562153977,0.6805428680581329,0.21713059619504815,
 0.22561566874532724,0.9386392465771239,0.3482906564593443,0.977159027217105,
 0.7100412007458896,0.21862058239458704,0.1852439086048333,0.3457779614019778,
 0.022075022929922916,0.21048906121667255,0.30043572928974605,
 0.09481888815146489,0.13754696522860177,0.1740846970391603,
 0.8337216137991179,0.34128570410821324,0.30038164906971454,
 0.5582049697037915,0.606072383239093,0.81030217373436,0.9752391622309914,
 0.4162539255320219,0.602218104841926,0.4154932159627909,0.21331180591008958,
 0.8581643841268626,0.5655778359849588,0.44084733375934804,
 0.30263389987621814,0.907806172514139,0.28759448331039794,0.6697006585310847,
 0.2278397689768089,0.2355542870595424,0.1289291794215447,
 0.012025219225886397,0.6231737051836845,0.18723654546840396,
 0.7836195788407601,0.11954554782594717,0.6108151380560366,0.5546613359500461,
 0.4770105286330931,0.1980591458232588,0.7092271259585532,
 0.034837788779462864,0.4420331304775721,0.44805909983158687,
 0.6944544891363571,0.39047029288665136,0.5289421904819877,
 0.13306987272339044,0.04494360682754861,0.5380865105259414,
 0.4668029084948395,0.30737055034517347,0.2925908578162131,
 0.19940952528452638,0.49949726603681655,0.37962530629023206,
 0.6725757255809347,0.09172163447435966,0.1731263889078719,
 0.024892123558122647,0.5086586125895912,0.5777466601559514,0.541874640882111,
 0.941061574792816,0.6289328365818627,0.9286797097470632,0.4935299638013644,
 0.1360388999656913,0.2957795998698475,0.14460900826645118,0.8618195623841731,
 0.8995400631347318,0.0933687214258816,0.1342549197286873,0.686573692764484,
 0.9343168962879504,0.513843599975587,0.8909992508621172,0.09961843302260898,
 0.41605219983202635,0.11462609875646079,0.006202383616837781,
 0.47440993335706083,0.7674013856271507,0.8662825526867621,0.5895809794907692,
 0.9127654496246336,0.6726231846662453,0.18582674825720158,
 0.08459928161335573,0.3085108571518116,0.8593240458032042,0.6140751228627985,
 0.19317853775101623,0.912211224348956,0.9830954734631242,0.9965350037607571,
 0.29101417596819656,0.0018845542596673814,0.49918364854073083,
 0.0747625574957611,0.9274911065888276,0.4000843046676942,0.697859448469434,
 0.6946344483802346,0.6210049258922943,0.8772336923972099,0.8859325014098203,
 0.7659173995554545,0.15170513564157484,0.7215331149624474,
 0.019933979377488642,0.4096811344198745,0.4616954436111337,
 0.2162040055941683,0.6758851345561524,0.044860930539550425,
 0.8351169285928778,0.4028590350103811,0.17131965377107006,0.5439552372887306,
 0.8335469924382968,0.5326604258424079,0.5096157071706336,0.05179356299747995,
 0.4887679756567389,0.6529220204913508,0.7035178048259705,0.7135920802535833,
 0.8811268514788094,0.5130924391753352,0.9374599678602175,0.42648831460783143,
 0.562208859293915,0.3957273378752826,0.3700599011981681,0.294608112653567,
 0.4280284929652932,0.7933822127112763,0.24071991584244923,0.7767984622545876,
 0.7190999248112607,0.923984197567229,0.9484405223942157,0.7857717237984108,
 0.682466517367433,0.5026717209558487,0.681172442993812,0.37341107894755954,
 0.4860208734894551,0.2731813807717589,0.5510804498177031,0.2840101697769387,
 0.7601093487988998,0.05842395647133991,0.13692589835459945,
 0.3306522131979771,0.04293628117075321,0.5245988350303161,0.6220719425456589,
 0.5990030883641542,0.8534483511587126,0.39580907918563446,
 0.009362345441592868,0.8438243449065842,0.186787646103584,0.0781310042864165,
 0.39387327271660455,0.44470798108483356,0.371560606181492,0.5490280914220368,
 0.9848383480310354,0.46485136666886906,0.32815371242430413,
 0.8838127020267997,0.8978962637764903,0.05160159552932575,0.6073195723710146,
 0.8802169168170709,0.7276908146569703,0.743968653767609,0.29122723423809915,
 0.8789068832005206,0.3460460293132509,0.5464149051446245,0.9146721254698503,
 0.9783733202767557,0.8135563006655361,0.8786963118054464,0.30082514770598867,
 0.9247469115145935,0.3916142755691583,0.13700278577109992,0.3591841914838868,
 0.9027234779465418,0.6941372578338674,0.3425085771159684,0.9707224729551591,
 0.9675227105015662,0.490395929733488,0.7444781111136095,0.9530003413976695,
 0.7309420350503054,0.14412722833774871,0.3036440784468972,0.7817744901497224,
 0.09246128533359377,0.9233824704181902,0.24781329645598194,
 0.5531600134094878,0.24782467408241482,0.5128003909892812,0.3882536038917186,
 0.3088012811700027,0.7642781137671514,0.03991575324447749,
 0.46673412790671553,0.2649020152772932,0.13237825762431454,
 0.3975978383268093,0.9864156094431579,0.6979504674746579,0.8061332095311119,
 0.07390592591581657,0.1768419997947156,0.6897943738902094,0.5029833813684709,
 0.8346965229116592,0.7206491992074086,0.7365670816328266,0.1319382530647195,
 0.472577604356626,0.1287224540624553,0.3010408893277291,0.4033137697580763,
 0.727761282503424,0.14992241783920257,0.7123840352272752,0.8118434517207552,
 0.039256902892147005,0.6514960028366494,0.9678587158219997,
 0.6239201347808461,0.33783930270979534,0.9436287942473899,
 0.23780839235931905,0.32637309421613725,0.6564627144252468,
 0.0860807530491775,0.8337381156464816,0.7167880421302889,0.8265726849113211,
 0.6123118137123315,0.2835066416315971,0.2125362823209782,0.973849035860088,
 0.1833505768723277,0.07720490789301027,0.8021176313601209,0.6625798767127422,
 0.5838934987438213,0.8291803956700586,0.8046211243109108,0.36058699487284307,
 0.44679433604285335,0.9668721962499991,0.45021789972969806,
 0.8651365671012748,0.3400152875886705,0.33678351547900665,0.8446419023691342,
 0.77021966812065,0.5138486778829983,0.18665573437552807,0.32644322554279226,
 0.02615603380896714,0.40845690470081497,0.6603589319464751,
 0.5615539773887108,0.6502648260531412,0.29998708160668663,0.8476445273680798,
 0.6333823026493948,0.9736260254939142,0.19569145000393506,0.7684480854028736,
 0.026782910799466153,0.14881216216026494,0.328473974301126,
 0.1955213596355363,0.857712076456396,0.40128023318046235,0.39380780342690547,
 0.6767087363143187,0.24007037247861485,0.5001877603320448,0.7114534941280971,
 0.6736034285217982,0.6711873645898144,0.3018117742804636,0.5679930062372366,
 0.9095425147042377,0.046934471280213996,0.6399226641493336,
 0.5255753999586206,0.830096281467863,0.5805227334894129,0.9037187636412032,
 0.5205195699003252,0.3566789007697835,0.9277058643306408,0.9118344640265091,
 0.12252589583788653,0.7549778148355702,0.15199386340724996,
 0.16945886056131443,0.06519051658721697,0.05615584734756107,
 0.28192595977782564,0.0464673495522655,0.13958734658413696,
 0.12867002330114163,0.785833117418774,0.1946285390995841,0.40723396713880455,
 0.5299990192384181,0.7655247090528529,0.8481532274190885,0.507995201344184,
 0.9441555291380724,0.4519621580815619,0.42481055670794043,
 0.17752966493896705,0.19443726163752095,0.7482648280168456,
 0.14714755021371662,0.5448506782524192,0.7735811168792327,0.7925464472621065,
 0.07785465732321373,0.1652528354465238,0.0472884673054923,0.35954939604876,
 0.24529510452634573,0.6260305303579962,0.44296416070350497,
 0.35277643571998185,0.340200826486026,0.13939266168695352,0.9089120887682421,
 0.06453574077420765,0.711931894232705,0.13859718088482453,0.0508544045533188,
 0.2919362036499058,0.9676843503762662,0.6172841546094324,0.7747512288657188,
 0.7795110467599471,0.5233223354950152,0.23201344467291518,0.8942858656808557,
 0.7219944998445086,0.25841756465663135,0.5397554598150118,0.3385433341000712,
 0.3679854871771755,0.5608203894966477,0.3787665338700462,
 0.0014856572547832236,0.9328223642443683,0.44119604012082847,
 0.2558605987952478,0.7251389755265191,0.3303225460846061,0.9228204081631762,
 0.09062483461832738,0.5364036368161917,0.35035668825057464,
 0.09483864680547949,0.07561054450310012,0.04083308599586777,
 0.9229349052192516,0.6197473442177505,0.5819951521164594,0.7068301687340526,
 0.8549105456022319,0.8578811827965009,0.25594540280040634,0.3279017405601796,
 0.79919998989298,0.5816984272231133,0.21500698206682123,0.9580376788895062,
 0.5098309194146027,0.11750031818474771,0.15220744277825937,
 0.06520116731096903,0.11785471332041897,0.2915852464011013,
 0.5123838504886118,0.02909033184095733,0.04596395545869425,
 0.03943771757570991,0.6600545711147073,0.23503463879493203,
 0.07876907385057597,0.023597167156911403,0.28470230463015755,
 0.6380351839515775,0.2348249601781881,0.8460206095048561,0.09746254620094386,
 0.016603313961667032,0.5936339868322931,0.9255134352471838,
 0.13568482027637885,0.36538747954838735,0.4108780373357931,
 0.7163376572010792,0.5097865226751808,0.06499422464143678,
 0.08610246723101644,0.31057033864939876,0.4676237810777696,
 0.14612233217923432,0.1411523447750932,0.42307039861551365,
 0.6567576642569714,0.0769328834272166,0.5609567303804813,0.1202426747537847,
 0.5144037952977625,0.28945817554366826,0.6980282589456389,
 0.13126762473367015,0.3542109498703996,0.6224386467786411,
 0.46492720436704815,0.426711561415446,0.678541467553431,0.7340860943385192,
 0.5646795583583614,0.2660321882635691,0.3253489653708772,0.8077607558673272,
 0.4155124070177082,0.5957678447007977,0.06548327517962393,
 0.026856336356277044,0.7115219471173844,0.6862605580603149,
 0.6183315236270984,0.2173486302446166,0.10551767670166523,0.5921871482130345,
 0.36902887977029575,0.09592642494245507,0.5453688511185317,
 0.07868726749725874,0.5045892999538073,0.42055177631681895,
 0.5251070919399766,0.3837575914322362,0.3440521820765702,0.02653980283291968,
 0.9382112145206445,0.2857742304407723,0.40880081013881475,0.5990361646163371,
 0.9296343882816483,0.7578644325254081,0.5797620965618474,0.6745825798606533,
 0.607431900133386,0.9291183458447899,0.5748594427902688,0.1064097337598966,
 0.352258857275195,0.8967451386984112,0.11177744855174265,0.6901018154078647,
 0.008850683236744872,0.7994327032141577,0.9697059758061621,
 0.8188045078052304,0.17642638467842586,0.51545162133607,0.2583886744136161,
 0.6407341967928379,0.5255168455986641,0.854929222435543,0.05993051697741936,
 0.4983909700054806,0.10835454595514216,0.5510034544545757,
 0.33669739163731827,0.8801797040533911,0.3503386046332918,
 0.07983318955289065,0.29271508367390686,0.7741052118172084,
 0.8993529059757424,0.10318529855431957,0.9019029949195463,0.4495222490230071,
 0.24892774479223245,0.5892954866692648,0.8964157283869227,0.6301705513144609,
 0.48966708806345116,0.7925825801340804,0.20842844010234374,
 0.5890855442237045,0.6815152275236105,0.5534013055625132,0.5087546177841271,
 0.5413686643567329,0.16966515488905265,0.36597074830155607,
 0.052331085003194056,0.9708278961554773,0.43563262990486096,
 0.6816930567426792,0.3657594245736535,0.7958178350995962,0.6489041716748305,
 0.13394747304993082,0.17355317458426622,0.6729347352766705,
 0.9369709475830781,0.3822839061953146,0.3923026798000989,0.45348730372145873,
 0.4410407206390917,0.9116319883798853,0.6292250070726237,0.6540482014136998,
 0.7603447167055255,0.3594322653414488,0.27254472158557974,
 0.30192216155886176,0.8674430408955237,0.8546976960492024,
 0.27356564418801266,0.4456510635821056,0.7365708732987922,
 0.08046639271030587,0.08276559656791016,0.5075367258064739,
 0.012551568065139351,0.9735457167900419,0.3898909437393051,
 0.7319455550060693,0.9423028361051305,0.13151979749790788,
 0.29851711003280856,0.9502287313275954,0.7552304103182662,0.325829787472283,
 0.4810781368551661,0.25676888358575356,0.43962656074600337,
 0.7760300409175699,0.8939473299486747,0.20596599272057292,0.431114034236702,
 0.2044320165713931,0.8208631417003093,0.8363296278340391,0.34770934750265303,
 0.5319758939309692,0.8370725610361828,0.6454941546022367,0.2462800172586712,
 0.7221920393196087,0.8458770191234083,0.5135761027738083,0.7289840821598822,
 0.07908834187633951,0.11086521047708775,0.5890339620643594,0.715890141911863,
 0.050964542012755465,0.7892156424551291,0.7001421254657552,
 0.20017289588855713,0.6267395090989365,0.5117633554955264,0.6259828951349279,
 0.2062512345808445,0.31267912480146554,0.016443827923829812,
 0.32237879949725745,0.8661828803245073,0.8990031837972337,
 0.26715953995898634,0.2741480073298549,0.40549127476795976,
 0.9341747523287615,0.43377176670154827,0.3492935930713684,0.3741334461868482,
 0.35050459920710586,0.4550113806962587,0.5379248768269326,0.8812060214715322,
 0.7782617528275994,0.849155776469849,0.8431034244065053,0.3533454490475455,
 0.8249143589863888,0.59498051750069,0.10235867457903414,0.7038123887518639,
 0.42436562853001214,0.2465210056114785,0.3884391797461584,0.5510746062519817,
 0.4070044875351092,0.5550925751351339,0.6533760035909303,0.17677912156101871,
 0.058388959353349534,0.9271688730219177,0.8643432519457429,0.857517176895533,
 0.7861787379677774,0.8750207036610818,0.42299235850715633,
 0.25708456716269634,0.727364490045206,0.9168879087743995,0.7651005511506176,
 0.47116526790913393,0.03685022336070387,0.8432660355652603,
 0.5908811894585729,0.40899585901157476,0.2754425648850932,0.3270618483575426,
 0.425740695429369,0.20322237411112032,0.48761183829490296,0.4781978945980495,
 0.6729778428716918,0.3389601945070334,0.8255413237105779,0.3699042423685336,
 0.5348470259396654,0.34749432598652874,0.3874300214702979,
 0.22732186746586347,0.8611505388343996,0.002232828559675215,
 0.932771512604707,0.47015332057864223,0.5641020100187544,0.9962729442644797,
 0.6713744029249095,0.6367014399478491,0.7843738295029068,0.8182779835158966,
 0.6032855865580282,0.26632271596688994,0.07904409506724475,
 0.34293367117421103,0.0642310304964776,0.8453171492050111,0.6745453706312553,
 0.1056864921213907,0.05148439250023562,0.36693948325458736,
 0.5355237886894638,0.3988662518292547,0.5328163279337942,0.08020318693844353,
 0.24296516164967863,0.34145645733447727,0.29699872992582543,0.87346741309453,
 0.12339576416487508,0.16213303721596872,0.5694795960273712,
 0.30915848110198085,0.4496181890269194,0.16695516279474,0.06567145148419518,
 0.318123397825252,0.19968271273976534,0.6670586628832185,0.8326156422507787,
 0.9863154736034997,0.49823953123773546,0.561638039099681,0.9673594708844254,
 0.8024316398717853,0.6942612998762917,0.842156162797229,0.1179445224911071,
 0.3217948572473308,0.6077710386362014,0.6145507238980393,0.8503584894074439,
 0.3573618994307023,0.7171150893189401,0.22691887522911314,0.7091176388342473,
 0.082188350489801,0.7627201535423511,0.5568207930508503,0.2811729599624728,
 0.8716353311375046,0.6153090257106721,0.6614247157279514,0.08853150577535351,
 0.5370511663588229,0.5716408909473106,0.6597765506323835,0.7272546587061616,
 0.5391791186770976,0.17117929768944562,0.19027357276796542,
 0.7835742316116222,0.4178385831832896,0.20015891788910967,
 0.27429533616207213,0.690043943471629,0.13871039363171678,0.1779897718075112,
 0.39800189592056023,0.10746925115707295,0.7032741645503058,
 0.9715689193936188,0.20971182991938886,0.006596168201089547,
 0.39202355489882734,0.9492445112457487,0.6205501458168394,0.6937493498309002,
 0.07937746510276678,0.975375198257744,0.4541748371817981,0.7385720937522727,
 0.9408619210964725,0.10629697431193885,0.46480919548409805,
 0.8195370078630305,0.820464195111488,0.45847885827393253,0.7870129744193699,
 0.8348464016376529,0.38753886999734066,0.9952585649896935,0.8592995635824296,
 0.0801125519358562,0.8854622473932019,0.4255001282580416,0.8784021397077865,
 0.4391630753980743,0.4448431307297651,0.372299017283921,0.6072996700743951,
 0.7368171161751573,0.414794923439583,0.4250668559311406,0.3666297727223802,
 0.6203067052141843,0.3250438246522368,0.004420123684520716,
 0.5849203783383019,0.1123968894215972,0.5513220972269116,0.31542967327877425,
 0.7592153786851583,0.030968006676242332,0.7425256823466215,
 0.7806934653752183,0.0562714037453943,0.5292094773994287,0.4760073740887669,
 0.6835243115661338,0.10812551479924348,0.2941872148252651,
 0.06335432880287639,0.7184843180211062,0.7993762917396336,0.8857280425964428,
 0.7927004160327203,0.5346387251912044,0.9515438217624358,0.8417238456832514,
 0.43330398069076426,0.9692575617776908,0.4565645756197705,0.7168812037747937,
 0.923567283973442,0.7907512747066094,0.7885832066470703,0.4485630337187021,
 0.7051745426848319,0.20809104469375483,0.2537685545346162,0.7979780317779894,
 0.110249146275059,0.4870226701716466,0.19604709942114773,0.7569329537327096,
 0.8991712199005024,0.5245845877170314,0.15797422980444542,0.947324458320555,
 0.34049439551845095,0.5802726384526831,0.5663231516560904,0.366911198318872,
 0.18162455014942458,0.6656781939090132,0.030913689368184616,
 0.5350462791214565,0.04000631114210973,0.30200079413552805,
 0.5020090212654306,0.4267441767100216,0.9654068884154562,0.14650238487464395,
 0.7137769109866775,0.5754121770652835,0.23719753438906999,0.3310199439996804,
 0.3635370146892355,0.8009000131767856,0.8935869918874073,0.8022337809196157,
 0.8774860732992604,0.15618415199269242,0.6470064761321679,0.7686136866748154,
 0.14394963084523238,0.6269855718556887,0.7320514375059223,0.687871586638535,
 0.5846263640172772,0.13264892836374464,0.65499864769845,0.9569415356087377,
 0.30905681146331543,0.24141166838438355,0.29606071000983814,
 0.6755483780381922,0.47019443837930686,0.835051281726973,0.8100999800271469,
 0.6009853056213386,0.387988474427873,0.6970085578140679,0.9933573991295126,
 0.9327712265585476,0.4732058638762987,0.3191454159760774,
 0.028258084472468803,0.8037069930410438,0.8232521493366661,
 0.3303521075211623,0.9044458947471392,0.0707022142814121,0.9555116623710138,
 0.8567349554254257,0.46041039056514765,0.5422636983407406,0.7028507737561812,
 0.5116709072670854,0.20790103562389584,0.13855219215177317,
 0.6197654751250462,0.3679737087419299,0.05906445419871442,
 0.11759983585236067,0.42348492781079516,0.19062737614339886,
 0.02389280455764098,0.3294902609548517,0.6167626270846465,0.5313747343995137,
 0.44505854469006734,0.7290009440397007,0.1279512749324344,
 0.11670477916173871,0.5142323161019347,0.37075224545701335,
 0.9448929440586654,0.6873121393300786,0.9436045447876591,0.8705376445825315,
 0.8179110117631571,0.9246200776849456,0.8776423399692236,0.8448307181859365,
 0.6708504847198066,0.25696553748586193,0.26371670959479876,
 0.29153335786967605,0.23561799611895073,0.42292768554962434,
 0.30237421180048607,0.4657210398031528,0.04359011294674864,
 0.6909123986299549,0.8721467571626433,0.32657710686815367,0.7456701288058191,
 0.9749407229876592,0.8502791309576841,0.21666473314890122,0.7116039596283847,
 0.6198860143183186,0.4909922551133139,0.8819945773197393,0.7229812111578923,
 0.3182362324188891,0.6769496289660017,0.8558483030112151,0.23635718768859604,
 0.02432878051686682,0.6357739231723561,0.15829966775903348,0.616083327585214,
 0.41022564921309557,0.1159226609403855,0.8857857659488648,
 0.15967038874873118,0.2608325612313166,0.2695885944411808,
 0.17211621930947785,0.9387750063317819,0.3737353784557883,0.5181700579554422,
 0.4438096166443851,0.7249513176246283,0.38466724794403695,0.8230913490360019,
 0.3504014409973044,0.5554009322029794,0.8595742277726663,0.526262303070386,
 0.09310971538807489,0.13141595034446396,0.0017283882713076126,
 0.19209264465228992,0.9462755048802375,0.4202527737164866,0.4770948646444976,
 0.23014141298936786,0.23304110173441306,0.1428338492277279,
 0.9559009342482992,0.8302880539900833,0.04958154869734488,0.3610430594351922,
 0.9456870895878428,0.0854736910697591,0.6721107632847786,0.9506080016342391,
 0.6369029942460281,0.3111755110279315,0.8325102906475523,0.13938066118237147,
 0.505202544873848,0.08939903939338212,0.6863941383668226,0.06438485608910338,
 0.47697349006487355,0.7797838266966369,0.12510281761699915,
 0.3963006999779879,0.5824214178509345,0.1099824037815611,0.7843886982787867,
 0.5125697910607518,0.5516873741620392,0.6517491067343985,0.4796337271340402,
 0.8040531000245468,0.8410144331985355,0.11451124887969011,0.3839350173253149,
 0.11857056233241425,0.836369467339796,0.10947014784410669,
 0.19369875848036355,0.6196261268162443,0.8041414405422204,0.6348840230215962,
 0.43110324431670444,0.5659306765020738,0.47757797456714524,
 0.7118536421991082,0.5665608067874222,0.9666846547170161,0.7494765466407738,
 0.21130669829303228,0.028484549727676223,0.8175393832378964,
 0.5147882213399726,0.9070202178632111,0.8492374396196639,0.31081210929246095,
 0.16287855942930085,0.8239431274311824,0.2775116408293714,
 0.16336620954060677,0.9966347463612506,0.08625930086170475,
 0.5107826015466117,0.0675328367503657,0.8711220102805337,0.23010569928906566,
 0.8264883130666666,0.819149192042998,0.3343509499157403,0.6276156765891239,
 0.8169662018846153,0.20179169571530065,0.6213715537007366,
 0.026068798613785638,0.4619144797272411,0.9911676218730429,
 0.7559861248280845,0.5689421158172133,0.39718854034095563,
 0.22052328074467986,0.8214303063817755,0.2010393018288591,0.6728812877500365,
 0.47327003328320416,0.47491088943735027,0.5972459476004371,
 0.5536969298019472,0.04406724912491078,0.5953620103118963,
 0.21075190569892643,0.6634898280808273,0.9947394363333216,0.6055071200973875,
 0.9485376702617502,0.23171642563964145,0.07214400559458378,
 0.9200221442909418,0.12321463303475055,0.3975414968575278,
 0.34499691593979154,0.5477710137975698,0.276885642947746,0.6221618193043394,
 0.4976331124325033,0.03930457981082325,0.17120363439557085,
 0.6344797214142072,0.6245634136818177,0.32239562590664095,0.9078953924457905,
 0.5184358536961868,0.5732970341247628,0.22169568337595802,0.4453773016437427,
 0.719853588238326,0.9167226836900507,0.8271298367614512,0.9540235280672335,
 0.425313738766696,0.7810742703795485,0.9497001180705574,0.7145116914803493,
 0.5112133590832586,0.5504136841239016,0.3538306438929505,0.9896792491694523,
 0.48325240958079285,0.5025472209882764,0.3874147890344477,0.7163352060110575,
 0.6666096881764669,0.8101648589725767,0.3543739144304162,0.929679212810854,
 0.5097123884882218,0.24236125882340565,0.20607460761905483,
 0.8994973422896342,0.2484203692019601,0.7351345269276708,0.26489547131118063,
 0.025471809830440617,0.6706038633335581,0.18025487313733513,
 0.2761527112173271,0.4830290651126016,0.05723858654438341,0.8644818064180283,
 0.2839756350761713,0.027533041520673596,0.4702450362779571,
 0.5900401609728605,0.009117497325930146,0.4979848505001766,
 0.5591585691096906,0.04533861217264279,0.029814723487824657,
 0.10591708000411226,0.022305770764776356,0.5736611946717998,
 0.8034856740611166,0.9111762364293864,0.25452181868833723,0.7526775303830024,
 0.2774806856637104,0.9825220676489167,0.5315709903505387,0.517210500758205,
 0.44673344434106843,0.10536234071206163,0.5087993334826008,
 0.4495603012296131,0.12255689529062841,0.15287238254601343,
 0.9751180503033594,0.6949122900299671,0.6337886239232592,0.7877353361104678,
 0.55706072817395,0.012290623156284708,0.924942095899428,0.2802326403914229,
 0.3058034641038304,0.8014389414345768,0.594516381529416,0.48192934321483205,
 0.8976748622644525,0.856810050298842,0.1538022643750787,0.9518487238813002,
 0.8978494469057697,0.7640828798170659,0.2621869300185736,0.43726711945903163,
 0.7970214556279086,0.9756979205726692,0.9028227654918052,0.8468598349188345,
 0.8894012162558168,0.8603161475076258,0.07871067867264459,
 0.10567865955885203,0.5826481090401004,0.875869314803608,0.26078864168869687,
 0.2899073893030032,0.6706916794381791,0.9188263165236673,0.5905689420874632,
 0.8426054260153395,0.9434819114061799,0.18928183612044736,0.5038363462636832,
 0.7886681702378577,0.6627111788812718,0.3873977513564857,0.7093605682474465,
 0.6390931391864473,0.17501379574104214,0.31363339111382094,
 0.6333468288456308,0.8363154116919567,0.8350817802499695,0.6749567045166308,
 0.6977760217872535,0.18568097926011162,0.29823814324249265,
 0.8384070929539745,0.20124530385361739,0.04995259259395868,
 0.16095738597203268,0.28949075958906756,0.8461250101111308,
 0.6594459116512028,0.26111231971337745,0.8942188258381636,
 0.20858261740014672,0.3074996969399,0.6109164304868308,0.5645217980754063,
 0.9694478305455938,0.9430282120586624,0.5211372272705141,0.16473136791539156,
 0.3542260626683055,0.5475015884788585,0.20672022071842056,0.9166338410196084,
 0.3240307767070114,0.2646257891642243,0.6820795364330196,0.4989850704600305,
 0.11239575971781601,0.31104204376661204,0.7831819619808549,
 0.5182358880444183,0.7349252612934134,0.3739096856023528,0.9690187966134701,
 0.21611949520714002,0.8222873964749569,0.7860051413527187,
 0.15674118670301884,0.5569154399992002,0.7227975337266349,0.5046247659455745,
 0.5998364949729724,0.389519637499001,0.2946547563834466,0.7688095068728864,
 0.9215638752061182,0.6352910675814701,0.37987474748925054,0.0276216754653964,
 0.2298536295434419,0.9932757864342563,0.10993237297242708,0.8500874012729271,
 0.8091520929048244,0.8788097712561318,0.5417260964829583,
 0.032088637117509866,0.9909217456970526,0.05726837347544278,
 0.4830130163911238,0.29845392904043,0.34795728985597496,0.22685942356795985,
 0.5067080025176409,0.1346046497136637,0.6103638094015258,0.9299825954090636,
 0.739328997461653,0.00553142858938982,0.5846603592992125,0.9073811881717438,
 0.606314083432876,0.5430508612734677,0.14085981477856557,0.15697518286893453,
 0.5287867586435224,0.7324028831297462,0.9749238074980633,0.09586357449886929,
 0.0828100075398931,0.637100603482455,0.6141063905159704,0.9482818856639401,
 0.8432532726820468,0.23085113216032382,0.4887850493290675,
 0.22564720537404503,0.20724729957663324,0.3840761636136262,
 0.6531910465640827,0.9773013173204854,0.48751784880557025,0.4437413307103102,
 0.12484894576719086,0.3526915657665899,0.14349320293799295,
 0.9781496907215128,0.6171878112627993,0.3948936216497012,0.19012404864613575,
 0.2772521124133246,0.4868888075788731,0.38281374463310824,0.6425725330856089,
 0.4649423498257297,0.6695044467935928,0.3747595613730069,0.9708183521557507,
 0.7706378997099752,0.67422434868818,0.30445349248718556,0.4823050985932402,
 0.30554874205616156,0.15952338339333716,0.1727186233717447,
 0.42691803239963677,0.3252458640464815,0.4282790613524782,0.3342317076220708,
 0.4537161663632646,0.14243872944968727,0.7281207222804139,0.847740199754536,
 0.18194932554296872,0.6976571724107509,0.5035822967583566,0.5513962095916745,
 0.11152675168258042,0.33043964036504014,0.6046550417755773,
 0.7512396933119307,0.37660084614789435,0.06042528950661685,
 0.2316327647178319,0.7934330763008819,0.5346433436489617,0.9466536028492726,
 0.6951382724245727,0.8978296121351348,0.6615622420797211,0.7553357520041333,
 0.5557270756606559,0.24402794939672068,0.7206036410967397,0.1768551362599463,
 0.5261022956454674,0.04628684699000363,0.2107107341174289,0.7504050381373517,
 0.8002346289483878,0.31815870860532536,0.9813166258659718,0.925411434707962,
 0.6002009534869966,0.7131848221866007,0.8775208385145512,0.01794170261310124,
 0.9130784371501269,0.8715469676390059,0.310374220203113,0.8368589128519534,
 0.7512227788166947,0.8320939195990376,0.4379660372307085,0.6472663932710095,
 0.7580284362614167,0.3691539531893383,0.7885747857835026,0.38442905395697835,
 0.31695956884597054,0.5652240508890631,0.8070239930964853,0.1454150492739671,
 0.2101994911914684,0.28545637689366865,0.8420368530089388,0.9044870227614703,
 0.7007316916445291,0.5674396375075395,0.7271992643676789,0.8219713091877265,
 0.14941943968369487,0.9833958071982231,0.8917412621344855,0.8520809309871329,
 0.09588183731695832,0.8839837304379575,0.7394868252657187,0.5225943848393019,
 0.5290783241399334,0.12586439670345795,0.07151098220046204,
 0.9852737793427774,0.07279646954002572,0.2712681739694801,0.3760349027722125,
 0.6869260763290028,0.22702137675029266,0.32744478847443304,
 0.4628425773110173,0.4811390170861338,0.475853988854133,0.5071666536384921,
 0.13065215708745082,0.4754463367115168,0.3953590127200324,
 0.060231588588336105,0.9438634702279038,0.35375089093667655,
 0.8600724853876762,0.8205367260212275,0.17122530427811222,
 0.34080120161022964,0.6856089657115505,0.6663553710441521,
 0.02699186517208707,0.2809323416965498,0.8205436612923642,
 0.13845379500901644,0.9910921954142882,0.13365779214156703,
 0.6659858991668597,0.3266765288363729,0.6775268669179112,0.7198739998651469,
 0.7587039694666159,0.5709569024072167,0.0951816231025252,
 0.009934562492605159,0.32196982282509656,0.6219160512131394,
 0.232410649549607,0.3962078705897396,0.9806703670408842,0.41439122544650375,
 0.38780666088144233,0.924358578170849,0.9871659693039924,0.5683604832435081,
 0.40831631795441103,0.3526303564173421,0.40260953802766575,
 0.5154460328816737,0.9033245137439048,0.6629314691188135,0.866946006522868,
 0.7894000108556446,0.277658905103925,0.8157653877755235,0.9983999700268825,
 0.021037662597585105,0.6083601633779544,0.9961113010046176,
 0.3454092703174325,0.11961525984027788,0.48829944982998597,
 0.2160541003684313,0.3500783901925526,0.4523746615488322,0.40181681930607105,
 0.7043711401500911,0.3378991914944628,0.5175924553387969,0.408556271950016,
 0.25073312533059355,0.44825038990501653,0.7203885430187584,
 0.07779328796171792,0.8195032038414568,0.45558758297250024,
 0.8438968801265133,0.43591144696154216,0.5991610784012439,0.5337075653514124,
 0.06238643094507601,0.8201425967761399,0.13435936347927313,0.631059114931162,
 0.8993375018786751,0.4167720694510465,0.5026014987939595,0.512714638207856,
 0.013512960672818375,0.20467879373919418,0.05428370050074949,
 0.10994478114371731,0.5549403900467594,0.14643135277172292,
 0.9181754590905509,0.8139746049484785,0.9597938414784914,0.6848755987055901,
 0.9205087560903487,0.6372053449975574,0.48834199530906464,
 0.11259492219268818,0.8579458633120702,0.0292334729020709,0.629492906219409,
 0.10148694631096689,0.9544258323684788,0.5006628848042726,0.2723875126810791,
 0.812923036107563,0.7004704961571256,0.4205308303687171,0.13400778115408618,
 0.16709319813468937,0.39526094930755673,0.704072121487723,0.5761169731259782,
 0.22624789603409956,0.14368638321084126,0.4969729201997607,
 0.10306427108144178,0.33139322798288684,0.747435367987242,0.9988338802897193,
 0.7078569543533971,0.47780865824296703,0.47080211678012185,
 0.7091133121111437,0.7806919715714908,0.6460719615818701,
 0.047532002240685545,0.7706240402554234,0.31529318859424704,
 0.4652546848630461,0.3386416098429772,0.16050798665040567,
 0.24719145780816776,0.3719445529944634,0.6775267812888088,0.4234566730207413,
 0.9971287526379428,0.19362359616789415,0.2088750875156925,0.8865975502708829,
 0.5357601277501532,0.40996339996958164,0.5475114978369355,
 0.35086376307739586,0.5023947892273943,0.4354493810408343,
 0.14134608242530755,0.28370229931700774,0.7405758589484974,
 0.2734981077292131,0.1889154839209003,0.17328435778075946,0.7706523665306261,
 0.44303124080282663,0.10553714606437503,0.5902407236508672,
 0.24730007228326023,0.7334962630298236,0.11178375778918503,
 0.31355182997244735,0.6782076317186146,0.09894632209307719,
 0.08553654465484839,0.4049530218362559,0.10050090348823093,
 0.3106802017957696,0.4980856840858289,0.7041364959313205,0.15483514008983634,
 0.5369473333493986,0.11191888268072425,0.18385973816311862,
 0.4322488909299519,0.07362156412147935,0.3166610442360728,0.6388959666059589,
 0.7023491567657868,0.20548890167862544,0.657075758508445,0.549375651196456,
 0.43921113642156095,0.5742000141670589,0.05298019423233791,
 0.4930431089165014,0.6963463810551848,0.15433362911915816,0.8649405571776843,
 0.09019053230477159,0.035058348089971436,0.648338267149031,
 0.04102724778732336,0.5060243375040614,0.5588969042705672,0.9963168951587673,
 0.14031970184939568,0.6021501182876874,0.28068308682190934,
 0.6730830497068103,0.5191759029860579,0.5531447384200348,0.3234755160892442,
 0.6669750887731843,0.8078655035089084,0.3093171317541209,0.21171248062473658,
 0.6405401026090789,0.3674956384585335,0.7863016612954601,0.7548959520056203,
 0.4672819484348977,0.7978025295941709,0.6480032926804921,0.4724841058810936,
 0.22399946647003421,0.32038783392380465,0.66442009964516,0.7976865903200203,
 0.7982517388604965,0.6153154673631744,0.5386149325583877,0.31763439967566365,
 0.5244766690878802,0.9396095013448056,0.4442137493202599,0.06851521268211425,
 0.02819097591195341,0.34886588541256724,0.4464072929175362,
 0.6558027861223402,0.5842275615980137,0.30932319903867567,0.8747109069888956,
 0.24705116472382926,0.6683240604534508,0.8117554829016739,0.5577514650882349,
 0.5847310758989027,0.12824690456087162,0.49454462676541266,
 0.1680960392488411,0.9231360987323053,0.7563084017214506,0.8397760940282195,
 0.0305808008186772,0.3683590870906276,0.16380231932604528,0.3804747585166196,
 0.853434185760712,0.4766926478042107,0.6009263003792367,0.4587661308616008,
 0.046706465841794165,0.15889663993786485,0.6227276590980309,
 0.9994581218207127,0.7901335447306908,0.05955068133342378,
 0.14979837278584052,0.27627038163710715,0.16722284565681222,
 0.7489991077092149,0.5838481842015735,0.9987617842201164,0.6568783091491953,
 0.909840538991672,0.4469846005693551,0.12262522289744893,0.38496775404255823,
 0.6383180022027066,0.11124905180390143,0.2922957202702725,
 0.43690238732411346,0.00801786447614361,0.9629518870267687,
 0.49047244189221917,0.8082712711457579,0.18374221781735922,0.24020703210812,
 0.2667265064340003,0.5770719005165528,0.9164213252751834,0.9578484409986189,
 0.46545763048170685,0.18845002668525024,0.5827382013272278,
 0.8317828075648922,0.45949679023725887,0.5677555594625402,
 0.47634001078935784,0.7418256867143153,0.0690605728640098,
 0.43731697963218996,0.28732171010908414,0.05444758424839091,
 0.15812819950849988,0.23458228846338758,0.9904933946033656,
 0.6037754862515028,0.7870482367026923,0.44644746873190866,
 0.28307901019122406,0.21117014536610823,0.5168258249569196,
 0.032021504238807785,0.2778434656757689,0.5555193135775667,
 0.4175682704854663,0.17542532627465546,0.008235338073374177,
 0.7310216340337043,0.1571238710479943,0.7718402940839185,0.8297669876627353,
 0.48150336459199305,0.7695944280528282,0.8769363617663971,
 0.47551409150421886,0.6751723849951748,0.8727575056464709,
 0.12440800949001296,0.6293375154520179,0.9093197930984633,0.172650276296615,
 0.2131635203963781,0.6809216946749657,0.12169858381490495,0.5840873360253548,
 0.7406759150822163,0.7159266986309375,0.16830202478076906,0.8578728026509755,
 0.14022280807500598,0.7262981903940222,0.302855999034208,0.4393813054028789,
 0.20803594860034524,0.44708351669883417,0.5554903347183568,
 0.055568735747801634,0.9023645093654074,0.355517573546944,
 0.039862810882366206,0.13202076026881016,0.3907238183047378,
 0.04050991861732567,0.5761977221867939,0.7487191477404521,0.7574695339778904,
 0.5831990423652156,0.9547036776887395,0.9931506483033352,0.4565473872716832,
 0.2670574266031993,0.5102755231893987,0.9907677022909346,0.3242808527561326,
 0.9860472244075293,0.7381264244944294,0.5923795881623082,0.04512320162757966,
 0.45015111361215976,0.8520375025041215,0.8354557222455328,0.6497121519044402,
 0.8645886715777897,0.16606431559875323,0.11565504699840978,0.491926140065324,
 0.5419137540780535,0.7646176067764925,0.05659805073778956,0.653626272959816,
 0.8405574681348162,0.749149042106745,0.6890265302221666,0.08850896493432803,
 0.4676510752483478,0.9685014576469833,0.472119726274016,0.19036785997616856,
 0.6233705183292306,0.6564826522390832,0.7478020747727553,0.01865140460371506,
 0.7758482166416519,0.003254564893993117,0.0895647772848539,0.634250604042801,
 0.13116000438238273,0.13258610549764782,0.3955521745487649,
 0.06054747161480534,0.9538260877639796,0.5843439472360477,0.5184631611034357,
 0.14372750389080025,0.4457933767929022,0.9538911840604203,0.542592868586803,
 0.31022733557069393,0.8094960489470479,0.5388676875390471,
 0.010236898793052673,0.7309680903482234,0.24212487643356373,
 0.5591430860554244,0.5754068572539179,0.5754244034915855,0.23928622701534363,
 0.9409846091154643,0.7192853626312394,0.13065981456541298,0.8387209421519577,
 0.1440583380867193,0.1581163648769246,0.01914114390683408,0.9354336701061438,
 0.8966939592688866,0.1555816773573121,0.5883092863218352,0.9003789272986886,
 0.5015158699871075,0.7275505007468489,0.8875843362443342,0.1027145179713943,
 0.36361904831395964,0.9220228723159654,0.4019273026043918,
 0.39003893100141696,0.5491999136950332,0.5515753107254796,
 0.12682037019219494,0.5917897461237246,0.2597438876662126,0.5071171055367455,
 0.2052781939811028,0.7033410351483818,0.19784281671144266,0.7724977543977146,
 0.5412289000835102,0.613339660054087,0.3932585852394701,0.8923704426997094,
 0.6329369542143458,0.9166964679032733,0.9696788043201306,0.11291186291910993,
 0.18866685600797495,0.18272416218025866,0.6440736589821243,
 0.6407996788339337,0.009921754727736709,0.1677510954994419,
 0.06156006878334419,0.5277431003097452,0.5421368729058212,0.8974846531963327,
 0.37855971408201283,0.06863922673723821,0.38550130903492863,
 0.21097534735795775,0.016379499785102114,0.5146610139069387,
 0.6174690532766075,0.11776082079467254,0.17188221969641926,
 0.7985750675413879,0.5563471970999838,0.20457398539826133,0.3798031771097886,
 0.4789072408603171,0.6518470957380501,0.19311522096222433,
 0.09051112800248384,0.7850948508699728,0.41012665695047335,0.427048117715378,
 0.8182825067848339,0.17332079268599077,0.809378026915502,0.9365868518200284,
 0.5831993516097951,0.13874185018273955,0.5967302000872177,0.8743767757358347,
 0.15757343462276396,0.1869400894053802,0.8339932406643592,0.6828048806562876,
 0.7120478999899967,0.7508109791973487,0.4546702668220487,0.23551573963700045,
 0.23281432591200013,0.6513384426295848,0.6598945850112753,0.5952135713419779,
 0.7830626411047729,0.8316377574302573,0.8102171685460988,0.2635009380337783,
 0.4122963880227851,0.40772647254709005,0.06413436982526843,
 0.45160057504594764,0.06664892241792175,0.2936338595661008,
 0.1429563781235571,0.19670211933101722,0.8929200964728359,0.622761348713172,
 0.3148896529392369,0.33903141247081003,0.634964710955347,0.7527811657584585,
 0.9558320447632973,0.6905203891535797,0.5252911741535286,0.7487741186622878,
 0.09426989672405472,0.2654181937949377,0.3780658969793188,0.886695086605826,
 0.16596833163208058,0.21195929520669687,0.1869922873281018,
 0.5674934273812563,0.4023894516500306,0.0321306628316802,0.5915333063365533,
 0.03546826824051241,0.49004464834185213,0.3968675764983207,
 0.2578443205679579,0.9396144804995865,0.5035319285401063,0.5320633133791379,
 0.3057419751694175,0.1379992060372759,0.1201535178689227,0.4254745054061554,
 0.016189686950027538,0.9696167465000292,0.6044453853212438,
 0.9400579903137273,0.9301952333486032,0.6889841298976772,
 0.028018048522584382,0.6001363205269157,0.16173432994299486,
 0.2965640408116371,0.4180436631452984,0.08454131408435983,0.6969749159509958,
 0.1165026455047774,0.5322800611133724,0.19261049950519005,
 0.22254203603540257,0.2642315439281042,0.710675913323553,0.03262578345615885,
 0.36291197189145685,0.4944496984049367,0.41604259685243017,
 0.28883887756245885,0.03249944146408379,0.5027533142341434,
 0.8238995502091819,0.9381352763577012,0.1051323578121528,0.5229178335412595,
 0.6659353948435023,0.8486129157254942,0.08921460854029695,0.7516558932327918,
 0.1080068394414373,0.5080583112628116,0.7026961730045614,0.18436036439365466,
 0.4185003886238544,0.7082611859520782,0.9028070456275884,0.43218726386158757,
 0.1573544441625233,0.566521802712906,0.7293106196988464,0.0748437060507412,
 0.8102788248202863,0.6455963619287745,0.8934332044171249,0.7487455873129184,
 0.6044320101757783,0.5281418929726278,0.49141379625951465,0.5539815244487374,
 0.19245612700573067,0.5511605632235282,0.4740082594568148,0.3443651547962071,
 0.2691567996536478,0.6158807258202907,0.9157633260005585,0.8487119749924755,
 0.9843730259798334,0.6985165247979425,0.7004819152307479,0.2757053977806109,
 0.9010642153204196,0.20806925592851566,0.13820894585871546,
 0.12748325887557233,0.558597032301523,0.819932457608547,0.8825643479725866,
 0.3623855530666231,0.969421406498428,0.7933408721093299,0.773159032449359,
 0.6038420496208133,0.8822529339843143,0.6949214779575751,0.2565318980030309,
 0.5067670196009543,0.8070126256587967,0.13980177488701284,0.8352095062554682,
 0.8955730231205243,0.28626611762322285,0.3872211632571799,0.773399135791381,
 0.31143490202358504,0.09851513077313201,0.3207690891865911,
 0.6962040269034898,0.019348884866087657,0.06998052823496215,
 0.5766874694965917,0.7839307843778236,0.543144809341897,0.21760944093820056,
 0.6134106134333256,0.08590577770691987,0.27428554918692893,
 0.33338770404640816,0.8294794577144902,0.9552551967537092,
 0.33723592547694903,0.2949856160301627,0.7799921040225436,0.535760970854208,
 0.6536818249892251,0.5217483150887435,0.5438546988288796,0.07870365650555589,
 0.3913437205602065,0.5588522337592242,0.6380707346833099,0.6873780395830525,
 0.9052554970619481,0.8045359740510283,0.38858091654718385,0.6565436877694604,
 0.6338909275305007,0.7048410424782107,0.30415894829368795,
 0.45741286021224914,0.23662255736109017,0.32461733294217676,
 0.664671716213943,0.7293583293319602,0.16892580777389898,0.6580634223117631,
 0.9923576151020147,0.10832236301876197,0.9157354579786382,0.2217321966969612,
 0.01560332049885782,0.1699644031373092,0.7280521589632182,0.988285656610663,
 0.8175218277293196,0.7823457443590643,0.10132527302535255,0.5074262859501464,
 0.9238753315442808,0.9059801704895625,0.41889828981545985,0.6892460043388936,
 0.20805117199134038,0.5601622921017507,0.4335477477781542,0.9834266537076943,
 0.5372191292710362,0.3953246159296996,0.32062187600321634,0.725620400617338,
 0.1197762317403096,0.19684477019657876,0.6531546407947081,0.482835503985933,
 0.10104656428719194,0.915978209103836,0.2932972646086447,0.7875995680753518,
 0.056985216712151066,0.789935031131149,0.608728738498042,0.33278408983983854,
 0.5778779447288965,0.003097647259151848,0.9348820701733567,
 0.8157997479513815,0.24793626480376707,0.8213635351394888,0.9529178337349324,
 0.016290272126696093,0.6620004768363031,0.3046495028305011,
 0.5621729297505397,0.7068422381836851,0.9566581440542228,0.8806662052021981,
 0.9298683311981016,0.8098017788784897,0.43706336012290103,0.8349585639756675,
 0.7736701179868424,0.0964307237459906,0.4430952273267351,0.8027891260035414,
 0.9161439317634547,0.05677425843630768,0.19550720339666317,
 0.9717408665674222,0.8883492283040924,0.6851413434901513,0.9411102234155105,
 0.9275833302675899,0.49025619046058744,0.14503709574585189,
 0.36902341678188255,0.37122895141424805,0.5598774876139094,
 0.9077253726417764,0.28472720092577475,0.5997602009749607,0.9084049441093741,
 0.7579489295529137,0.8113114748176793,0.8187743390575108,0.11885521132706844,
 0.014259488864843184,0.20014079767398252,0.8406032678026208,
 0.5925018250576355,0.5998128356802949,0.22650359758072036,
 0.23131857588517113,0.7257584121413749,0.24164823959852078,
 0.3580391819513613,0.3977341729809325,0.09973050286355212,
 0.12338113667108941,0.8571261439190101,0.8107964896598006,0.4253620980568592,
 0.4494761005090764,0.8998700165992817,0.8200587542845296,0.06397757369683887,
 0.9663840220926447,0.6052095900791065,0.7855778707393554,0.09127716373602801,
 0.8803777726451907,0.9217054653862575,0.041406434002370096,
 0.18927175115000472,0.04680971391794608,0.4937293911592775,0.228594846681893,
 0.6280299182316512,0.7024477780562551,0.2956875421744086,0.9344195169577548,
 0.2722809244293585,0.04958612052223155,0.9185434390319706,0.8226612790164815,
 0.6216620127136656,0.18932263063422383,0.31143633464592724,0.950387848680748,
 0.5417723357838902,0.6851354561187014,0.5373394878943741,0.45407237201929307,
 0.21669953716715207,0.33043862417071246,0.500162859819294,0.0750904494692538,
 0.8646242140665997,0.7339525296663632,0.15821336266652786,
 0.24153120255001614,0.15594765539744726,0.9733091601495572,
 0.44696810276922827,0.11424740339406603,0.07560454141629558,
 0.8306222922002842,0.8547692898583418,0.3175378720691353,0.41698696604141317,
 0.6335907509838442,0.3331913682212986,0.5297068064976533,0.2868514742289302,
 0.5236099351453157,0.9755631871060382,0.6441616655336763,0.5124027094548403,
 0.6647415230298478,0.7706146238639423,0.5803457899946001,0.30118428765234007,
 0.13568494670946252,0.322788391169741,0.7917659522711389,0.17784357987992339,
 0.5311676385216508,0.06901248467927179,0.10377828608607409,
 0.6640695821234879,0.17488303236212588,0.3913467854247614,0.5842878134467403,
 0.26961324417999366,0.10468570333294969,0.35251255815153804,
 0.3764607483985136,0.664045969590724,0.11913006575885998,0.3657728715727907,
 0.6549399286368918,0.6298844349927227,0.7943986431542224,
 0.029599952948399233,0.4025174530756689,0.02779102863851457,
 0.8375921692646697,0.610326611573752,0.9693022603643279,0.6793574255766331,
 0.36501659820536503,0.5846478387410917,0.014915475841314252,
 0.4857382850125681,0.4404257569080855,0.051801682186010445,
 0.11452636011336093,0.5082265236740608,0.4570755486043939,0.7002809130984642,
 0.8156534074278607,0.06393798461360589,0.4137810729493261,
 0.07852764479244356,0.7729786793612613,0.7027978211443561,0.6313084239773734,
 0.36001203863347464,0.7052486608603549,0.5362030526688347,0.4117507356663517,
 0.1630175387395959,0.20761237021889167,0.5451921571539369,0.7253380185879975,
 0.37632786571583143,0.8221272591317492,0.36366521718421296,
 0.8380078380999303,0.9955313301798798,0.6046166450130589,0.7861718724496306,
 0.0907618274054528,0.08518805680665731,0.5750524386825308,0.5366316151045245,
 0.30399146477272776,0.4074549106089319,0.6049012175010444,0.2830563926912284,
 0.10569873637840055,0.9729730265141228,0.3750075073130571,
 0.17106193194804953,0.6942049495073777,0.4521916586316864,0.7024002250021159,
 0.046784370411342535,0.9306573374243077,0.18889711242559404,
 0.6130167347408546,0.35872214434279515,0.25361605897620565,
 0.7826455371684421,0.33239716583152135,0.7307330046725329,0.7632330820040264,
 0.12199430038984882,0.5615686754287756,0.5210661785813888,
 0.10684050635621056,0.6217445110731189,0.942374411025478,0.11433052198010574,
 0.6563476598646875,0.39725573993748886,0.2322523641510872,0.5681878119823299,
 0.8329164384737956,0.024071330495217036,0.2847655865780405,
 0.06565205983481015,0.43349918587726266,0.8298208790507606,
 0.8600183724885176,0.8294401174047424,0.2943773169216022,0.5209279406729502,
 0.5582546778314047,0.23024193798148873,0.28836017348441034,
 0.9475933794796717,0.12807815592899452,0.9310640907881295,
 0.33857184674376595,0.7452924323922978,0.5481506304257303,0.5201297070309585,
 0.8979984209756675,0.8014027918164036,0.33649331401646077,0.0365243109523945,
 0.19630196957214263,0.3912526180001068,0.8426635531402499,
 0.42891628780715374,0.1871616970682809,0.9482058907709072,
 0.05357288329943888,0.6895916666488636,0.3407701151927929,0.6591228471066389,
 0.7011531608190446,0.5189293929461537,0.7761925013728194,0.7733998216438692,
 0.7951536795921827,0.43994355014951403,0.9525706132862608,0.17350571880694,
 0.030536478970529246,0.426101048306299,0.7877681351760945,0.6064632585197933,
 0.2990755587672056,0.2372509463593101,0.41717061374473374,0.6021320316791148,
 0.28650178310551233,0.6177698393636281,0.7591087971415618,
 0.23858470672790622,0.528324786686111,0.14019532147576308,
 0.007341804066236524,0.651395593281178,0.7774676484566571,0.8612753946337897,
 0.8855102863624257,0.9809824848292846,0.6166711751227468,0.0346878879535919,
 0.7154289544466866,0.2440104886931832,0.939852930798852,0.06495111835430867,
 0.5001033287962837,0.159710303985643,0.8125432442172282,0.5656441237442983,
 0.6627412157893919,0.14715474313108956,0.9285927315599141,0.5515922333884673,
 0.976258008704483,0.9071231261115537,0.6657016318797381,0.1009974258103048,
 0.12490935035570883,0.17955609931765515,0.7973635551275541,
 0.4871930384993103,0.08925445262114184,0.729891651484626,0.8848308862039125,
 0.42243978469025256,0.2862218073031745,0.9719118278051632,0.8289945849118282,
 0.04961293805044609,0.5503934640688036,0.6649257828067063,0.373892223230903,
 0.9669269664444836,0.2610019920574209,0.8298703391980228,0.5679184780291436,
 0.2934287059185833,0.8488910394516442,0.5002415343617644,0.4111707237304677,
 0.12302811832384308,0.8824542088874465,0.8374262194046052,0.9927984731088394,
 0.3903069401411069,0.6041226555721522,0.10760003777462324,
 0.11576386336115618,0.5887302026143012,0.1528511855194319,0.692189701541946,
 0.5683823396147487,0.4908671465615174,0.4431523151952578,0.3675601954043356,
 0.18717902880479398,0.47522747040095514,0.0461223990404549,
 0.13094148121838634,0.3069745762450429,0.7076366673835556,0.7589535390426466,
 0.17198888973257775,0.13648492339039253,0.46639640935182036,
 0.5799851059865592,0.3792746872776742,0.3402656205043191,0.27547519533507725,
 0.18241695172377426,0.6636924938915557,0.2116364590309714,
 0.40985841306036586,0.745518710262727,0.9213366255086245,0.12038237665981166,
 0.7360598151958675,0.5302911130505357,0.1821742147202614,0.46026839635395655,
 0.447786282830563,0.38360849520734563,0.29624481820888304,0.6627335530147745,
 0.35651525303485265,0.25323454472530416,0.5780346935763939,
 0.7588061399698915,0.44081268058934875,0.1297871754693577,0.2631002456814918,
 0.6694658556100056,0.03852784007418397,0.35855163600443385,
 0.8991423873697405,0.7263470752903123,0.8406228569081947,
 0.028803076943476724,0.1891877836788154,0.9409545338872026,
 0.4047443075599376,0.9080627756074998,0.8595708095376474,0.3537764510183594,
 0.722325030666807,0.11286257349660966,0.9927004476553021,0.8812106178651671,
 0.9212832586976845,0.09068184720932704,0.3982208063488134,0.3418095314392279,
 0.39868002405444913,0.05534739028172009,0.05567533211572062,
 0.15705034335474732,0.19343536118559268,0.5256426932217555,
 0.8952225156723299,0.7992596500316475,0.7787826262387239,0.32628292531667713,
 0.30703039875737115,0.48954185866643696,0.9683318578266527,
 0.6948776580386454,0.25985015952130086,0.48547318742794465,
 0.0738452023512498,0.3566390239889672,0.8522284215212501,0.9189174030963425,
 0.6074649207588939,0.09588353957146267,0.013854099409372878,
 0.39109950613399336,0.672670658576403,0.8320703354531171,0.9814912966818937,
 0.6541142432669331,0.3197320585150556,0.5512138047323644,0.10615165540922777,
 0.4274048531904041,0.7919545093968186,0.8472929000815987,0.7551742193327513,
 0.6433300274637422,0.03303175958848015,0.30909708623120213,
 0.2730168343677978,0.5015336932001258,0.05180197845376611,
 0.08397933665477886,0.3006222097620045,0.33937766384827084,
 0.02721089529030496,0.8611025108313182,0.07468591093350452,
 0.8722762564228869,0.20609952078303684,0.5746240576179585,0.7006961542706918,
 0.6317805795345148,0.438989837452067,0.16176612838128612,0.1417132155378762,
 0.409308005558756,0.23591775181275298,0.513821087295382,0.7867038813758279,
 0.19930736933616866,0.5513761516658209,0.2636168384205754,0.3461968276530618,
 0.6718012382799252,0.6805913224349125,0.1436231301178682,0.5298560838041761,
 0.5484247590491056,0.3116496895796781,0.8285905813168734,0.23720407777897867,
 0.67260781863125,0.9591941387296038,0.4070542530487302,0.31583834860473137,
 0.25239796387142666,0.003858790925123645,0.4316301759527992,
 0.7393037783112048,0.027221844370495463,0.27902502465694246,
 0.4935102651507919,0.4088232421849305,0.7178813587881386,0.6292692225830366,
 0.9099444995988093,0.3561399665888283,0.4972460847444504,0.6161545406566712,
 0.03550082561286805,0.48825577096380135,0.6223235529203354,
 0.10999391727874108,0.8009775600876494,0.6922625278043187,
 0.46722138680654557,0.22211358245779778,0.21080506250386044,
 0.33787963710318003,0.08574466191642138,0.6417376135480617,
 0.4333575922988204,0.8286118977124517,0.878032328434729,0.08817051138915122,
 0.6610173466178275,0.6193417167284347,0.6241553172665452,0.28447322358745664,
 0.5577419323597729,0.4339659589843793,0.8275618175659262,0.19459130389405355,
 0.4244863658390856,0.46769457982299145,0.5790917842027092,0.6402412986847442,
 0.18673048320406638,0.009828815613786901,0.5684398330172942,
 0.014917603874985419,0.04774218293532617,0.7547570823114892,
 0.8047901913036284,0.7381281380117501,0.7881615724376921,0.10043505863246582,
 0.5764164597697564,0.9313740255587378,0.5751762806101699,0.17894726957424134,
 0.9966580056670052,0.4753852367048492,0.7564348789766234,0.8959260075329389,
 0.4577515643497687,0.11041492933101771,0.09158731766546746,
 0.01981511530848301,0.29413260093220017,0.4195733433968708,
 0.3242107518681778,0.9252005815001965,0.40946313862943073,
 0.059852378808918694,0.1634814409956742,0.9811775286945891,
 0.6772068580762722,0.7005381012891839,0.7268514928811483,0.47344687996994883,
 0.921870175209881,0.47549425691908187,0.8967101355708819,0.7348954809897257,
 0.7068176367264588,0.7948191814916279,0.06001429185353546,0.6510503542667898,
 0.31524259866412396,0.4048227596061943,0.8292693731063787,0.4340564212890897,
 0.13885472187090708,0.12502128383467415,0.03367881595258593,
 0.14824184180160804,0.8349842649479063,0.564235447621937,0.3238594209982063,
 0.8277792991038881,0.32933807067402177,0.19507061950044402,
 0.1355282804761484,0.13219540606069324,0.6957227360140136,
 0.34138206702161056,0.4795631924425481,0.6093493508237333,0.3470897950191625,
 0.3281145965864235,0.8267701446899924,0.1950088682898533,0.7132454718961796,
 0.2465222955691424,0.469733326048867,0.5394763270340354,0.3384737371501776,
 0.6622247616127612,0.340596005142049,0.6601858761818292,0.252971707587047,
 0.25955814986836967,0.3599726755429984,0.4912564339602251,
 0.048365769222721955,0.7422742428561457,0.6137339842491961,
 0.07534729356844505,0.043949410757838425,0.9424674280825165,
 0.069916032483611,0.977849977894453,0.9855725364970864,0.8086143844851692,
 0.5861802706172966,0.5475030399517404,0.23801212777463587,0.2561407363625158,
 0.9645731862176473,0.1400266884685113,0.7663214958057686,0.22407218586039157,
 0.45816957991751783,0.5101527557301229,0.20021522985086737,
 0.5374229581674608,0.7390105915077135,0.929977157481036,0.5809236444412124,
 0.5423972615229019,0.14941319518497986,0.4612626124776763,0.1885660207772596,
 0.14535901744765323,0.052575423708459246,0.004807955477322046,
 0.40098962102251723,0.7855348130202854,0.394347177444607,0.3865089943265533,
 0.30268283985454136,0.637408106176167,0.7345266143059006,0.7017077941640447,
 0.1692000304576433,0.6954363822675601,0.9952091525703177,0.6790360824087035,
 0.16966464911884838,0.8899193104960859,0.6028084873191454,
 0.36929721630854395,0.3072146734685324,0.6459311938331522,
 0.13624466226869814,0.3034881485682478,0.7284273338866338,
 0.09252225622664945,0.18218524040814144,0.7414071734861513,
 0.25819919524877066,0.049420933880214735,0.3561104192236516,
 0.6631048467703357,0.6292843645998837,0.25155060514745353,0.9740401978204909,
 0.24161528916800623,0.123549561720383,0.2052127646348012,0.09396654752790068,
 0.6085095766073767,0.3448065624397727,0.8720729399225464,0.4455527876311056,
 0.1375497587941682,0.305830236466187,0.8595858488675505,0.9428727911366156,
 0.030991377544980048,0.6883729385888302,0.09953415127362442,
 0.3628338165772693,0.6013198146532378,0.6834166721350979,0.390757090893049,
 0.8274326278067343,0.9954029152208791,0.7724398543490913,0.20081565921268474,
 0.3860474385365614,0.9161357017677713,0.5871693657606422,0.6290405249570933,
 0.31101923060633063,0.42880135826759536,0.4700479630752611,
 0.3665033571188665,0.4748152558435075,0.6181215989617629,0.0896649015856732,
 0.009025337944185585,0.07067687653886456,0.857734213395417,
 0.09712843459184561,0.9935484755601453,0.33311122993380915,
 0.4702581898224669,0.09832915408438225,0.5499228995095528,0.6319614681515302,
 0.4076576740200898,0.7459270594008274,0.3793849475528702,0.29409370065022933,
 0.8969575431747049,0.039023439696684536,0.4753028184417061,
 0.19587304999842792,0.15831055517696757,0.9780026457327776,
 0.42758141293350027,0.8325021525872033,0.45340037904403385,
 0.17962519558028722,0.8260036053884293,0.09488455972221177,
 0.22414951323362242,0.8425749094885663,0.9176254398821998,0.5894904628817789,
 0.6618694614654869,0.5704926457701849,0.6168126963196909,0.28838025159906144,
 0.38203411250261,0.2370688991222475,0.035773059177426125,0.913954192676075,
 0.5544742674855135,0.8261870751784288,0.4467324103870436,0.5479816757151199,
 0.28497754411792897,0.7145201220234297,0.43998219911052217,
 0.9885856396729538,0.5105980700968686,0.4393358027283958,0.2005630187237134,
 0.806810531801609,0.6758899748126092,0.599020676257946,0.537462855601595,
 0.18582782815012155,0.16577536012426075,0.34661887610940434,
 0.2245796271639655,0.9306294150675768,0.03467086498109706,0.963821183512742,
 0.7722516956658051,0.7857900803536899,0.3788283084199937,
 0.004766616596122986,0.5092285397980264,0.58116365870105,
 0.015579008573518127,0.5824411649997201,0.24242562342536977,
 0.5550356262734543,0.5488861994442217,0.9067962872605507,0.3840986926850969,
 0.1684942432107629,0.9638772853418357,0.51746516782076,0.12170533417898721,
 0.2142965746688834,0.9016398418015926,0.06726049781211485,
 0.12136611677566767,0.22232952659903105,0.6998456997979265,
 0.8771142995033846,0.05235258181540636,0.5779249600919212,
 0.24373055549757128,0.34924102603605056,0.5337711781661785,0.68609582760353,
 0.10609603590681016,0.07861615204331851,0.6544230955691103,
 0.2913490737178761,0.19176488204081032,0.18222982407471888,
 0.3000259658287018,0.8611808812876944,0.5785386096463258,0.37839679002386284,
 0.13885802044910267,0.4733286898765682,0.9550993731601081,0.7014691262455073,
 0.8873475736397474,0.7498759413334677,0.6205438022682951,0.4661510582320989,
 0.051307625006945856,0.2733696266967133,0.7483038121766254,
 0.3626867321084132,0.7127832456971862,0.3411925956974545,0.2959074403595463,
 0.7356607071681299,0.8549072768193579,0.6962766161875461,0.9388757745027105,
 0.36829430242047856,0.3630719859322292,0.27678190279453196,
 0.06433723947250047,0.512996085045971,0.6396366824703138,0.15220567786863082,
 0.2646889986047716,0.763499837313489,0.6717612454265235,0.8756264106007409,
 0.32593647730499764,0.01737340726339054,0.39598135926167444,
 0.5925949684906421,0.5689014501235798,0.5181417253493751,0.5747309772124547,
 0.39840503230092195,0.1723811094727088,0.3256184647073128,
 0.16590655135187127,0.5978384394297507,0.9322469959275462,
 0.14501457163099707,0.9589918870822179,0.5350488742718562,
 0.48759771540010066,0.2904691817770365,0.9537244790430893,0.9328865868883157,
 0.8804199001612597,0.04609536923566526,0.7773385448120953,
 0.17437619827274142,0.1337971821770546,0.6081300872946598,0.962633411906337,
 0.3914924621678936,0.3449063793796546,0.659228646091538,0.5089327966646976,
 0.7480227935706498,0.48961259110897826,0.22010003631785802,
 0.35859330928442257,0.061552893154811716,0.5599036757074857,
 0.2669244724859823,0.5196755155581453,0.35540905689886104,0.6354134518999618,
 0.037694041878607254,0.6808017142224886,0.7559113106379507,
 0.6294999329689563,0.7191979347355687,0.6897071193221791,0.867288043610694,
 0.6141565584386695,0.29271827534400296,0.9217971650886665,
 0.49615005369923537,0.10183040869105653,0.41052195754292875,
 0.8807253680548919,0.7136060060056614,0.8224782902903501,0.8299654976058451,
 0.5467176629436631,0.5593809260857183,0.639680508945144,0.9587909475504671,
 0.747268472957301,0.1539235443257656,0.10506469004923302,0.18332783277049913,
 0.5422055807316877,0.6850866218149744,0.4280873440418259,0.7707371502090099,
 0.8986635797391922,0.45537023244925146,0.44918973159129394,
 0.4288558063962453,0.32988991815562174,0.6242298914092455,0.6682911865721409,
 0.5771517384076582,0.3801342216063861,0.4808333894651532,0.18844895687623153,
 0.2258725232322938,0.5586474661349256,0.05563306569523918,0.693640536203787,
 0.32272816938619864,0.8292048466424051,0.8365510678290257,0.4243634972615571,
 0.24237896268953374,0.16051250226824143,0.37397149614661096,
 0.7609196162261651,0.34268587797249217,0.883160648281333,0.5194665433231141,
 0.6804231889771412,0.748083632036305,0.48504614965530035,0.44499661917662825,
 0.9898019151989293,0.6099954167789964,0.9189285120663033,0.5146819266956305,
 0.5636584459548671,0.3821938171363578,0.8234836682043354,0.5005825578346137,
 0.4085836563427927,0.9710946685334773,0.36542857118378647,
 0.10595245388403118,0.5731441706939385,0.14077723392795183,
 0.7239167276776511,0.6643375352342266,0.5644545574316999,0.7560984033661365,
 0.09155848099567954,0.003445635886611953,0.8257571599926159,
 0.6609101055573781,0.2847481140762912,0.42851876305927106,0.152114317996368,
 0.8410161852610972,0.25786610591852155,0.6364421852526707,
 0.26053610056811527,0.7230062774304672,0.017280684353038978,
 0.8276085408869993,0.1138279135515865,0.5995217343921366,0.9055746364789427,
 0.6475549272708051,0.635445780895715,0.8107744246453029,0.3525282848076797,
 0.8933453119406233,0.7375218735164477,0.1181811086484178,0.22844026011503238,
 0.6428540210501807,0.8071917831211038,0.055394026534422824,
 0.03009106561967867,0.34805318587697676,0.7573222165848179,
 0.1962974880102324,0.44979084877534214,0.8334143356827519,0.5546884258154191,
 0.5142502307726766,0.025635498903289422,0.5885913389157261,
 0.1792463273069127,0.21363873713213932,0.5965168902219027,0.7781959241571311,
 0.6878953698614286,0.3907362009646915,0.16215377569911316,0.9077594363665121,
 0.5316712967504873,0.4088474369168814,0.4960895043012553,0.7291334527282597,
 0.581064849538064,0.7343479294874198,0.22610929852176986,0.6030569363758367,
 0.21937365307035306,0.9578562818747607,0.7551031341546018,0.5476124903722103,
 0.4914199282321886,0.6618943135298447,0.4934318107575498,0.6870224379078766,
 0.7084340795543869,0.7935458327185911,0.24190228041504247,0.8886840112803932,
 0.1483479515218562,0.7221646844751526,0.7036607038061493,0.20185831417914335,
 0.05706759112810211,0.7163451812962156,0.8835343835848758,0.7205123626947569,
 0.6159428605902961,0.8749871065118978,0.3338734280113953,0.2496885365361443,
 0.4403387233541767,0.006874741346726609,0.8593225798859769,
 0.8043556696446524,0.34857471980114363,0.06481052482919936,
 0.7976953864377729,0.6665495440277794,0.7791691455087695,0.42609576392166315,
 0.011174108448197773,0.4543903878325696,0.8757190659372049,
 0.8616523838909651,0.4548754995117429,0.3730534727554412,0.15833035633265857,
 0.9926902116508672,0.9507214341696774,0.15340221043480773,0.952953231310754,
 0.5369983707043183,0.5788925543830119,0.6686871903286515,0.674451018215925,
 0.7234433476254837,0.9309522336563315,0.6828905082220882,0.7649690421852722,
 0.3806445538682104,0.27047955935943624,0.7145383071031036,
 0.49651257083705147,0.6499073880453022,0.9480961100675502,0.7902556532502186,
 0.30395264972133706,0.7937969839548391,0.9820966171983043,0.6437545863899851,
 0.9498385560622541,0.2826397922187963,0.23067795771072896,0.3764123445855647,
 0.5897217416233002,0.39480596240085075,0.7138672228696294,
 0.38311567025817805,0.9094864619505559,0.2794137949102151,0.4938913337560453,
 0.8599221839321741,0.17306952337968506,0.5979973400378437,0.5376921231810087,
 0.16374471971351945,0.3667549613897344,0.8194619093089436,
 0.41035535451027894,0.520928678575808,0.5962322570521917,0.4075734178215581,
 0.8836340071784831,0.9534520285025896,0.2710966206762242,0.5296356558089281,
 0.38391659134133016,0.9442172654883777,0.9119425756130259,
 0.19739448641413126,0.5287725839875128,0.35330275266336775,
 0.5497866550590871,0.8797100518678256,0.08550255955024588,
 0.11812804317257586,0.9480240775400999,0.7668497033082986,0.7685173217375089,
 0.39573072099189266,0.7488660334148441,0.2689550113194532,0.5222758028536723,
 0.5188814189023095,0.8958528025936445,0.5378092300415604,0.21533610019592875,
 0.36408891215236605,0.7084599132206495,0.6562621698570236,
 0.46932911532710087,0.19873178968424643,0.21229431348502037,
 0.7426966190634893,0.6255575962669397,0.9259739629292503,0.7510722919063296,
 0.4852847843762961,0.396876895168771,0.5745649196650705,0.13010943674367392,
 0.7337792091623981,0.9660577813378456,0.36334438938458147,0.9370538882676649,
 0.08743633916994242,0.20996303345863054,0.6577633725365782,
 0.9090130025861451,0.8544362596576974,0.7773050882231027,0.05360138302918105,
 0.42224709090145474,0.8480783875544363,0.7584510057883647,0.8975200523834233,
 0.35433219616917566,0.23583622361348122,0.47311644459327573,
 0.23388417106203585,0.23822574005774444,0.13242170736235992,
 0.6528857908461693,0.8618127925898306,0.42876919673009883,
 0.016542263145739478,0.7912367751032741,0.3262225428600509,
 0.01840411741218828,0.3740738215809294,0.33745307041608164,
 0.7651320291050734,0.5433066232182467,0.1538132102819224,0.11261972940397935,
 0.320727266830533,0.27706555213671746,0.002315874776350091,
 0.6647966002912954,0.7002136993412323,0.2155768531879474,0.5101351402832272,
 0.9826280089362447,0.7490672966157244,0.24170704884632954,0.7660143398064001,
 0.9467696811236933,0.5758478007057023,0.2897493605398437,0.9557499586541631,
 0.7742715673987173,0.11643294505078927,0.500222354823653,0.7706480549871032,
 0.4627057251501714,0.3046311694184173,0.7272131977996492,0.7491857677308764,
 0.49896533118189335,0.1763716381009508,0.3568598642383518,0.0742496826337874,
 0.34624138375080304,0.24541566986436436,0.6946337648245112,
 0.6602855388891665,0.7202855139121808,0.6755577491690985,0.18203838128062255,
 0.4183413969182649,0.033531618355190806,0.5435821064613741,0.853380197874602,
 0.36341405625855827,0.9292834518398115,0.3364016347212584,0.8097089116847744,
 0.554104532536431,0.03697553836698009,0.8318722798322561,0.6280040454822962,
 0.21302300820737463,0.7441409769405468,0.7011764586355602,0.6812724430628614,
 0.6084482430640503,0.01686135382908127,0.9585438341599849,0.6647354656568778,
 0.07236262696452267,0.16858839147289417,0.40460711691770723,
 0.6118089498954025,0.2805452254508186,0.6690242079221829,0.6675655046670375,
 0.6684669075611264,0.43887072273940575,0.2443990566407579,0.1477209282437587,
 0.10441678519547459,0.7460789666951938,0.19439237784854968,0.853623754107964,
 0.2701106071949244,0.45055821296097975,0.5432599654423239,
 0.25110125289324803,0.0408054104122042,0.41487553794594945,
 0.9222663405704842,0.44251410367024224,0.01062622898521226,
 0.4858564471731889,0.11955750348424166,0.5758142436427404,0.2655273639379835,
 0.24734950839836967,0.6575374691449211,0.996633399435511,0.6432568320676785,
 0.8266172455858279,0.43628764040723256,0.38055364582841256,
 0.09147094644682885,0.13504154053205197,0.44195798856258817,
 0.4348942124153101,0.05213817298947665,0.9014235605837176,
 0.06680082499511142,0.2289730165282361,0.7050434649997689,
 0.11579908581712062,0.7446343693985932,0.5595821910752615,0.8439889755657054,
 0.11186191467643314,0.7162062704261672,0.16120749409524227,
 0.21756092873457633,0.5831310961833501,0.22595654031710177,
 0.2911230813017608,0.8865151836229272,0.7662195966475818,0.29786740046336,
 0.2545538476440996,0.17218455787221998,0.6051431906146394,0.264953522609906,
 0.9142990992858083,0.09252175523751338,0.18970398782532416,
 0.9667016801303605,0.618078892774927,0.7289044779245173,0.19325171379219808,
 0.4073213680119063,0.25021942502541683,0.03296840119337063,
 0.6224789405578661,0.05129937777594218,0.9020949014395003,0.9137152244167117,
 0.3926832312595123,0.16950660195023626,0.40968091509160254,
 0.9134695290296673,0.138952951346339,0.5517837882851931,0.3824608767749018,
 0.5606351951475071,0.9474627654157871,0.2704711164024195,0.8995258591318424,
 0.14463383234395943,0.3412713532207684,0.7497254670604028,0.3376645159950479,
 0.6306762582554177,0.7595842852615168,0.027946524952865204,
 0.01941046968459348,0.07924958676963445,0.48392058764939583,
 0.8641308039809739,0.7877470384391899,0.42603786798252896,0.2470927133461991,
 0.3450658015763779,0.2777452584071244,0.29293964717683263,0.6877011797820344,
 0.15181419099557425,0.9457733654012854,0.9151784887285179,0.9095967809724331,
 0.972226423359869,0.55564282866841,0.7621377416595586,0.9693327573002654,
 0.24239911834094563,0.10876826240363291,0.6072272809950787,
 0.49474570828271014,0.909821893728135,0.9879042633536854,0.5298240306914556,
 0.6886449121846778,0.4970558675077873,0.8320847546818637,0.09084937250343483,
 0.5486504245779175,0.6586880383896719,0.2723278191051558,0.8057258715177666,
 0.07274016033532116,0.6864905483544468,0.8757274174270067,0.8147228168447694,
 0.5096699121850191,0.8316805752411798,0.7473305164126529,
 0.0073207423182565234,0.7373528006070421,0.9563030095747596,
 0.41601635632548906,0.8489751622367211,0.8183542415946602,
 0.027520216633963313,0.9976036235030232,0.44084865359752834,
 0.41716568459145287,0.01346738452852203,0.68536551514327,0.3660520548003471,
 0.4899949239021417,0.46533303660086245,0.8867030321878301,
 0.09227909437607296,0.20025925877539485,0.34304801962238374,
 0.5787801609752055,0.5126289994107877,0.4221582824794946,0.849149030179979,
 0.7124881283358518,0.29637737714617407,0.1881260373150141,
 0.20056815892409485,0.7034189993388751,0.7704780275126736,
 0.10560595862704825,0.2573009615391847,0.5716940723692758,0.3382187685996221,
 0.13377575978159406,0.9932066833439871,0.2774741291126339,0.9520401625794669,
 0.18181994125484946,0.5612792545010438,0.8926136911854997,0.9018836315307742,
 0.9358307425253051,0.12639845948439843,0.35055473503663837,
 0.2585191403819789,0.9212816095549106,0.5110621834974052,0.7959929156640068,
 0.5231492221433468,0.5564235835609616,0.24979358575415767,0.7773415929406585,
 0.19136720713183009,0.7667211142077767,0.600413930687647,0.9433993501313254,
 0.1956911596930474,0.38750957739297076,0.9680686635967155,0.5624713677763129,
 0.8427730301338683,0.6979175175434665,0.11721478581640421,
 0.02174558271224969,0.02826095720317401,0.0726396858940983,
 0.7689686738212803,0.19308084650486124,0.4004054451483292,0.443235620676367,
 0.21370255880913047,0.6761414361089352,0.015213079902769544,
 0.6151107788771252,0.3256448456385306,0.5732904599444559,0.4876537903879825,
 0.16991878932832882,0.8844184486437399,0.42392079998536003,
 0.4179091132180327,0.8447929825960083,0.9092482679331594,0.4166433934885404,
 0.21647260437196325,0.6024921722672281,0.7189285143968923,0.2503339353511974,
 0.7539157874420791,0.6261537791311521,0.11336056397396488,0.9283126052030993,
 0.9660125641122321,0.8452682413890605,0.5988059381565494,0.2550711028346099,
 0.4072016647072597,0.033496475091646105,0.7502144307539894,
 0.5145283595627588,0.23761207697552278,0.06173522047845492,
 0.9369612723315308,0.05848430079546629,0.869251243828921,0.8831521130105501,
 0.5737141138861459,0.7008683007180988,0.09108454811017608,0.4915032798649033,
 0.6305390292068814,0.5232164068101122,0.0569045783460691,0.03685617311246303,
 0.4453541969770435,0.2985937178712086,0.7306337549809048,0.50119201544597,
 0.033857731429897564,0.8306409430636443,0.41879388603793155,
 0.46859673118378076,0.37098149716168305,0.02032553683953897,
 0.034532908768648674,0.0025661026491368144,0.6726387848420179,
 0.9658232759551136,0.5886963982852107,0.3147260725653226,0.9272895586994028,
 0.4674689186825647,0.6944801216930612,0.08915220523391576,0.3315200328665191,
 0.9236666501219325,0.41843530650283045,0.2701593770522599,0.7550055809986506,
 0.5344745716932896,0.8185759823862301,0.17035910181100977,
 0.13346089359769397,0.648330313554035,0.8994980939265875,0.7490729525440294,
 0.9690552626155039,0.36544074012239547,0.2633944195709701,0.7952155208555913,
 0.6306160830592386,0.13630318932288887,0.7420312560119047,
 0.38588916632233217,0.5420576726148886,0.858404157569661,0.5377142280073259,
 0.2171858566083057,0.6421343035589038,0.14331205258483537,
 0.17244606874241408,0.03453091562654731,0.634497861787946,0.4981806175069391,
 0.745331622280206,0.7613268816503115,0.5292753375647206,0.5215474021525224,
 0.2900399358800275,0.940722436157534,0.008528527634268368,0.4808978782866755,
 0.7733871078153925,0.35885517107154785,0.19970161371807327,
 0.9927347857454802,0.8738063451226028,0.10598206213491712,0.3111331177136969,
 0.5782697294684431,0.3726888363270059,0.9063838110785365,0.5994878503391307,
 0.49027734276452084,0.719138444035451,0.5489435664329044,0.3664110791676167,
 0.8618983324784079,0.04748243748135683,0.7282523712478596,0.6908291743429288,
 0.17343854266410497,0.4893433289362886,0.8865112662967554,
 0.08629403504709043,0.6386334286869513,0.8840454186139877,0.30947324035642,
 0.05430734480236765,0.6398555422682686,0.840936470619071,0.4812610151476233,
 0.698497293528626,0.08251590362119365,0.2783895869610159,0.7936111960676614,
 0.22967851307394105,0.7821889874727855,0.5379041627865799,
 0.45152720010434366,0.9535191161788455,0.06771756817233587,0.984935373787521,
 0.4604411820196683,0.8347347832300155,0.960975734949659,0.01620147629968327,
 0.8823745148706947,0.5498024774351494,0.6979569841726649,0.7870626586312248,
 0.9505416167117774,0.5331158809142016,0.773276669014399,0.005101898449364706,
 0.9194611518978316,0.8560805353108192,0.6562269105111386,0.4953156608571486,
 0.7243704502499007,0.5751522186883976,0.8183924017648722,0.4813049940393168,
 0.18855249281224085,0.9839207635149253,0.8632362928193145,
 0.39076731412900234,0.11377494941746014,0.6978330243995898,
 0.6046539386943228,0.8774806290155448,0.24103944735114025,
 0.46483751639491544,0.8605939196969891,0.728108906726048,0.5281619946126125,
 0.4149790868819827,0.8593859734345592,0.42053888018083296,
 0.020964841170093562,0.3471833667200871,0.48508343942281285,
 0.7713191568947141,0.18702708335316187,0.46444685324011703,0.450770312211615,
 0.029312155991813915,0.7105897101553058,0.494995438448925,0.0911435816305185,
 0.42486904680009474,0.6004192539592076,0.11355498209604553,
 0.3151350570570145,0.9247887699161984,0.4712073343767833,0.8622524702515331,
 0.30292711660139415,0.5046762208665044,0.43653304203191357,0.866563419197239,
 0.008624790250592707,0.7183878871064969,0.4372517291958784,
 0.5414042526474222,0.8925609094560201,0.22371195548242995,0.6211917831344013,
 0.7163132031699242,0.6745960059875196,0.15703632455644778,0.3233189910871991,
 0.25076234040329226,0.37305970753137563,0.43105834002849464,
 0.3293271129249584,0.06095068238120449,0.8164519510288426,0.2979863588837087,
 0.6114117682482035,0.24372766536415857,0.896644668986168,0.5792214573173161,
 0.5509247790770573,0.08609390096688552,0.941873947825326,0.23738459859468697,
 0.7100864475004292,0.5573693101327405,0.867693917775681,0.38567582362085406,
 0.10887013500681264,0.7698529643593507,0.9457490791367142,0.9375978784774163,
 0.5158213215233907,0.9949452006341666,0.47147500264679565,0.8008161481420145,
 0.6486732515879413,0.5610719955150948,0.1344044519465225,0.9416624895870015,
 0.4828836368507837,0.19615314236651038,0.5756058982283994,0.4971991145295592,
 0.4922454081916867,0.5942685818380397,0.9934036139573094,0.28454736135939773,
 0.4466887429928388,0.03434584609601876,0.17274651964925258,
 0.7027393290241702,0.5922142531973433,0.7672930807356693,0.36502137743783725,
 0.02805459214020578,0.45415843288423063,0.18175410245688128,
 0.7020952701641296,0.4701368546246081,0.11521970219426358,0.343274906663811,
 0.992875711951063,0.4787275594323819,0.40604770102716903,0.8735067130431031,
 0.992457287957724,0.9000523707642626,0.40138278961415363,0.6974572620217893,
 0.5629690818363008,0.20890830420365758,0.17775319426560188,
 0.6435538175830734,0.4207276573650286,0.8622364043505581,0.8048131174115262,
 0.9950962290401225,0.7176642064006031,0.5822601351797415,0.09725046874836796,
 0.8324185384182752,0.018149348448960878,0.07042851722528853,
 0.6211958734419338,0.9559159079236699,0.2888935857585656,0.6106536786073897,
 0.9885304818174497,0.1578174366681796,0.20284522666641136,0.8175306584973473,
 0.10204394695061825,0.07194092461617685,0.26990522046282306,
 0.7569698959044051,0.5810731529813022,0.15879895685797785,0.9838540110694477,
 0.9261941842770933,0.7560818675020783,0.06651294077495196,
 0.22895313061084488,0.8252844670082141,0.4637144026654887,
 0.22131865907066495,0.5240478533406883,0.3913576435006837,0.6145995986734134,
 0.7431194940253478,0.720848224548131,0.21887269710149393,0.15954372862663568,
 0.6886599195081258,0.3395143221015654,0.4818198540804347,0.6212867211295476,
 0.06492835503803629,0.4846127910031719,0.09404275053659472,
 0.10527814780468292,0.6481729389791135,0.6940270510596791,
 0.20830135966102648,0.23669456655441046,0.8991060211708229,0.515712082812312,
 0.917096293512269,0.6130476889034533,0.40671032628061254,0.9033174545994983,
 0.8681409842649412,0.9249049681086665,0.762498298949396,0.050417592508077314,
 0.4722143193867261,0.6381137124971618,0.22436308035261276,
 0.17208071206814257,0.03709161667392702,0.13412064218868258,
 0.3146470838817761,0.8979020280475372,0.07914787861367067,0.6912641710781944,
 0.9794064454218928,0.2821769356515019,0.8648054855269887,0.21904907147665453,
 0.5105794619469992,0.45119983286613197,0.014684353722090537,
 0.34499048269332744,0.03849264100305483,0.8156595939523195,
 0.46052415482460585,0.9063063065474716,0.9696003940228395,
 0.40956110688623326,0.1506680333280549,0.12877106757245105,
 0.6249251143559307,0.5702550986522108,0.9486762119944797,0.6796298145462588,
 0.39044329931465216,0.4819756418354718,0.831478512798979,0.18951277770409025,
 0.8578457695898657,0.16889726889231915,0.5014612226749141,0.2811604944103393,
 0.14297871996323075,0.042568510599368725,0.44439188616211767,
 0.2574442249498414,0.8392112003765099,0.3546796292443909,0.6233735971118557,
 0.045781951117950026,0.8032631512204237,0.7036471461733922,
 0.13812459768298357,0.09592348729135025,0.0066424219131601125,
 0.8483008319736827,0.31445694755687115,0.045890884071436155,
 0.39321909693938917,0.25125785207977713,0.259477021556736,
 0.34671613671349955,0.45479823226712124,0.9796493961555567,
 0.7963712589536627,0.16151359524049735,0.5701694893501134,
 0.10576763324660365,0.7958839345105937,0.45929442787713937,
 0.9030708592666447,0.9919892267411307,0.986071287311655,0.5619118980229463,
 0.5106379299703158,0.7111374846347998,0.8900309474873835,0.5550170396423745,
 0.292031040483554,0.049557745781234286,0.45356739584207684,
 0.22439359640659817,0.517091107149952,0.8184420399086227,
 0.004320797967879653,0.953144284638689,0.19459569259114773,
 0.7791588645362434,0.4124148628641775,0.3166643289585871,0.9318575958912696,
 0.41729718689367856,0.31501940053830924,0.5342566028824726,
 0.7855274267399946,0.7164565299188081,0.2419691789409648,0.7400122751014044,
 0.5602305861273089,0.3400605487527043,0.17736816323711713,0.5000464477736828,
 0.4793289924320442,0.3317917624697846,0.6108980485566089,0.20023747897394628,
 0.6809402008698207,0.7628089316467069,0.6560468524472867,0.7755594973768494,
 0.7824279489307591,0.39650037465671506,0.15509351678032945,
 0.5131055113901859,0.24140272251121653,0.5479336163385512,0.6113907865895096,
 0.3385481952361713,0.3746255591272517,0.05060686687057059,0.282734877579651,
 0.28133875768014316,0.0032602431182193747,0.4844480636102906,
 0.23506946830033648,0.8401111927010373,0.9904588538198261,0.9649518158035375,
 0.6640338387192353,0.5049714510702435,0.9154797490392528,0.8721879113644402,
 0.9678110517780095,0.396361329962182,0.350473811586146,0.7652630717400192,
 0.8372702127277948,0.2657218248395461,0.3526036442167416,0.5041085314229995,
 0.5279831820650143,0.5105127890207781,0.7823020065242479,0.4624063349981259,
 0.30134603285901407,0.6956855017361769,0.4855745283240612,0.2705092669627207,
 0.6648771069737744,0.9017591734695631,0.06353417363439418,0.9845169879027773,
 0.8260616204755737,0.901570288217265,0.1781278567741511,0.2880675598074559,
 0.46413954522989853,0.8944073334622065,0.20707884088401574,
 0.9641581737419247,0.06153551438404359,0.9344631546705263,0.8784199260746219,
 0.28772781294334826,0.31537854333612547,0.040107203264542424,
 0.22170052360148473,0.7095955896937561,0.9251149096374555,0.3783958739922052,
 0.7105181019226481,0.33870469500562606,0.7133636180762867,
 0.06351842260735052,0.4729580343891382,0.8051137204389848,0.9540607352685045,
 0.061920054610830055,0.3022765076369749,0.2678279045149703,
 0.0809262527895459,0.0199589832651077,0.1930724496030518,0.08666309133657912,
 0.4385062597378653,0.8429012589935576,0.8532132684690258,0.296758679183023,
 0.3545391497462219,0.40837568639312705,0.7453143558424699,0.7832887489984924,
 0.46022760550631736,0.022565738139421865,0.3682634797373512,
 0.9904509538670995,0.8551033603859227,0.08825683911666304,
 0.24344728060271437,0.18192528203756242,0.7435655446255957,
 0.07614359444444474,0.6945608294054594,0.38311715322382445,
 0.5422689120181139,0.628900520785716,0.9351770619467714,0.8878136917325568,
 0.6144693736287705,0.03945725561184643,0.0774802434494053,0.4960438531559409,
 0.39176674609821693,0.2218087300623265,0.487140383743619,0.0825912056175847,
 0.8877845464839107,0.4649920129173477,0.8736260135301929,0.2066873012605397,
 0.8854629463054853,0.1942587441113045,0.9300632925428372,0.9243092011751038,
 0.2697357562693501,0.8459608298904392,0.2003449707722953,0.4785339058430106,
 0.18252140845120646,0.015554203667156652,0.5914943042828016,
 0.7595377025497082,0.6532107158080275,0.0031587877716218182,
 0.35166927765162403,0.2712540159701605,0.5175263881388492,
 0.48818906278892826,0.4600554479207406,0.6243536779539691,0.8935699458483373,
 0.7559347105996508,0.5306611158068665,0.30876971693632915,0.504405604639519,
 0.8172864495647827,0.11969467130257783,0.010647345841572342,
 0.912736871979919,0.04219482241677941,0.49758266456396705,0.5512002355288381,
 0.1533247782677175,0.37471079871924917,0.7434115901432962,0.6307423568237324,
 0.19092878195435725,0.7502032791671585,0.6693586748225836,
 0.45886356061766964,0.24454061119422932,0.7242417941013533,
 0.7992369678899354,0.8256863554474707,0.4139484965964143,0.8083929503869615,
 0.8934037680983128,0.6799645280649842,0.8397576412870591,0.2750044804550015,
 0.11234554400651675,0.3895334593034374,0.1369811970919408,0.5194877263780817,
 0.9839809310778624,0.5083242527954959,0.30719490865964916,0.7093083500207165,
 0.9139927810248076,0.2889532229887064,0.9144245810019513,0.9264312679590305,
 0.7447072316857983,0.8520404982779775,0.6507748043510739,0.12046892920102692,
 0.45525851386302696,0.3611805365195305,0.15716364354869206,
 0.8402481570463332,0.8048012261698421,0.02218893868864613,0.2688270502343726,
 0.40271287331555916,0.9737885866058337,0.1449771008111198,0.6032844420591676,
 0.124042773204598,0.6404534145050412,0.118084812402795,0.022369904719446554,
 0.34642178138825563,0.8047571449731035,0.3934179760092118,0.948228283178348,
 0.7720032547214595,0.25380496611426684,0.5872287300677979,0.4124004841572264,
 0.04868004567023898,0.5943159185864458,0.9360948481403968,0.9559788690847035,
 0.10004458564705387,0.7316385544237474,0.22873366868029799,
 0.4431421039651656,0.2777234300368844,0.699480412734607,0.03242311769602535,
 0.12614499446287475,0.5046332008156525,0.06054730255393037,
 0.5865056283856304,0.33108672725429955,0.07593479927554514,
 0.06257935895884448,0.7906339124371304,0.1292795121144943,0.6763749997517001,
 0.6707422947054673,0.1079441050774469,0.6396575736745975,0.30616764103312366,
 0.15528739916859724,0.1507161289709975,0.540421671831509,0.255591474819003,
 0.9800469113388255,0.402820567818692,0.8613094433201369,0.8713854986576467,
 0.044014926174591684,0.8791339071417558,0.7120340114192234,
 0.17307635179031822,0.6454834260633816,0.9683254643152724,0.7509030813770029,
 0.08462954426704328,0.6893322226689008,0.8200275149729224,0.8224123513493132,
 0.06307296368823789,0.3791523759585166,0.023576719560177595,
 0.8059497515180564,0.24341836323874433,0.7140952295333602,0.9354903115912835,
 0.7245478881646255,0.32109371951986965,0.6080890900636434,0.848405807103312,
 0.12774508801067175,0.47782956970591295,0.735146331474832,0.3751579191852059,
 0.14638151255165976,0.6099659092574244,0.3088587465932996,0.6222376481411614,
 0.14756861615628314,0.4861654578901483,0.12994554747310083,
 0.05728011406656752,0.5190488045686377,0.5555799617943024,0.7307881857157126,
 0.36977977431814235,0.4142860158396696,0.6084310891987625,
 0.19622479889204714,0.3924170383520954,0.9689497322944872,0.7007083762368899,
 0.4322346959025912,0.3385525764269447,0.12833407141373776,0.5653012907798172,
 0.8781253202183307,0.06671566647860971,0.7535683579469429,
 0.32607222943545555,0.27406090123536275,0.39630278009967557,
 0.5120772031198427,0.3066359034141376,0.19689092579519707,0.1408105261031074,
 0.08594444472318696,0.03159575313787555,0.013314414680761777,
 0.19556872136381453,0.8382662117298783,0.6716660602029562,
 0.10886781697201275,0.245519165240918,0.6003632744517109,0.46677182471069156,
 0.8250335548939635,0.9339965043444541,0.062267174849727325,
 0.4043187806066779,0.9087258707562148,0.2484327393833674,0.7701253479530833,
 0.23790543529213004,0.6196088338945388,0.4213570425848916,
 0.17791930201691353,0.20501201794264434,0.7283387380084693,
 0.4672187769876689,0.47770549877691726,0.6345393612513387,
 0.017092941626112412,0.2278055906372376,0.11436319716643051,
 0.6045860763253558,0.4124224329910019,0.053301003306106653,
 0.9494510241775361,0.7556694148855359,0.4326639040318996,0.12524141621763385,
 0.1139298785012618,0.6891589179782642,0.4929915633690354,0.8261602189533566,
 0.09023782157916538,0.5955037651456359,0.6534472780552447,0.561530021365797,
 0.6795323616752037,0.5303262706132543,0.2694239918415199,0.2816747316142433,
 0.15771480795263693,0.5648747743564428,0.5855023920364469,0.6325030080143196,
 0.38950822760002235,0.00845012916645338,0.02602991819951872,
 0.7055234804245077,0.22020423236401676,0.7149261567632488,0.6434811966323388,
 0.7414654165201132,0.22532591255719026,0.6722177383790527,
 0.19337727126923787,0.1503021123995003,0.9775959021651133,0.6977538090160746,
 0.18455090529324836,0.522977761001165,0.5696553014059602,0.7102201584315047,
 0.46672828885371975,0.20956724407364302,0.2152452281182693,
 0.42083555224790914,0.38906556209252185,0.6517813446606258,
 0.8534519882799229,0.10072756324177645,0.7792723278662016,0.7675897603914066,
 0.029835754559153216,0.08181283216164925,0.1702388678538661,
 0.6774011596179997,0.6697660526008236,0.5547523091238205,0.9041504843725943,
 0.9105340209372794,0.8186843278224953,0.2619441124360269,0.15758472820831315,
 0.6713905396876358,0.5379072774250246,0.18214109488306796,0.5941373920436922,
 0.9789343694127978,0.7863901527276854,0.4339309015359012,0.9460107406486185,
 0.0980100935018896,0.9114731737531274,0.22057871219826308,
 0.39009117543735905,0.1659180022911183,0.41426124718417856,
 0.8612896031905285,0.68849791136895,0.3624565184399865,0.2368590608702409,
 0.4765380147934737,0.15714647402306048,0.11617407543719338,
 0.10721131338436218,0.7638943663577785,0.8426565948202313,
 0.10688398918065078,0.8595868402664201,0.45325699706254763,
 0.6116561058577294,0.19605534429198923,0.4874801089278977,
 0.008159844942599692,0.3603051184199826,0.45845034303618604,
 0.0380136209420463,0.46862739135761333,0.4023900671417806,
 0.06617411324904177,0.7061023863501668,0.5428203889754255,0.9401391517219992,
 0.9245577118972192,0.6779786410692626,0.11922424335489112,0.6712565115423325,
 0.2428522961401668,0.7930369819827912,0.8156721355759304,0.6162719027027636,
 0.6255726575773073,0.5375200375433274,0.2191239748105258,0.38817778108474443,
 0.4968452638429136,0.8205971761972743,0.24207649209437265,0.7782877242379256,
 0.6182226350429061,0.39512332197290667,0.639655697132456,0.6770645145067401,
 0.4360562988907821,0.9514279442464613,0.016857784039534418,
 0.19053811480534655,0.9650455321745725,0.09597409472288221,
 0.3619814598151717,0.5126323273157768,0.8735445528546915,0.5008922666087611,
 0.6282006313793367,0.7110592564344973,0.3132973302731651,0.42889658843913425,
 0.6922105374151382,0.6091320810415247,0.26487076604232573,
 0.20993315538567825,0.05645707118326393,0.9771999649121048,
 0.016297309826199258,0.1819240159850617,0.0016180712091690141,
 0.03718200841317898,0.2171526958240677,0.3781232537354883,
 0.08835908379267998,0.10539865926370495,0.21023415684509283,
 0.9158572323044334,0.17635463929093032,0.27318450778310743,
 0.37723591079056007,0.06606718800457645,0.2731010673946621,
 0.8945128025933593,0.1001663482187174,0.031652993284326914,0.858947162709873,
 0.7947604504839222,5.286802770424481e-4,0.9892852471429481,0.082997463295021,
 0.5702953551446875,0.6779106840615529,0.08295058712819292,0.1433449391478584,
 0.2304493736506089,0.013107135131764203,0.06583057882249266,
 0.9020089487291644,0.40500284049996593,0.9241998599072924,
 0.08228034102340431,0.5929874197989593,0.7713985370323574,0.1149695455721701,
 0.9945410546108492,0.33703177556382424,0.0062911504170212185,
 0.6645262992034063,0.1659774540731075,0.6985703360225785,0.5654682171930348,
 0.4131898188358494,0.1481978667738797,0.5761469858428538,0.726296379158642,
 0.7256802068481267,0.545089834010916,0.7584035631112706,0.4028180297154845,
 0.9065376230762789,0.35313723180036993,0.028342469402211745,
 0.5830842519887206,0.8590140204120185,0.36319335092195226,
 0.26774574418074437,0.4302770646411652,0.5390692252826059,0.7397942696744004,
 0.9992192527540897,0.5725115758653192,0.889651255454789,0.0598597616705131,
 0.17451320978907536,0.8711290621802753,0.9434613173588795,0.5157452997642452,
 0.30920673903125373,0.9718309339817919,0.7643067461000314,
 0.10829187539219998,0.38889664749553055,0.5847793844701716,
 0.45597345227563135,0.39145010546812353,0.4303239653125043,
 0.9826064702809313,0.7359573815962892,0.8907791632157855,0.31968687388527584,
 0.5724826533943088,0.03556419197434346,0.16185523279486125,
 0.5888184665586408,0.9106789561685453,0.6704906881355295,0.9910978669420046,
 0.6268039856175739,0.9832229316110801,0.9921861727456982,
 0.010448130979983317,0.6323179142047639,0.8975909494505704,
 0.8439215057918673,0.5623762487116488,0.953527958258609,0.7440419080642671,
 0.4197309675433155,0.6842289674518005,0.1330478888067792,0.2396110693273441,
 0.059411258709639925,0.718140767605426,0.7467223805751129,0.1904643567248897,
 0.8433801147355101,0.21463683406756118,0.5922770875125625,0.6200746300991284,
 0.7076019254549872,0.19014986968831415,0.1411506930893086,0.4498744057637831,
 0.3012544171381245,0.08176646300554202,0.0699882693414996,0.8420465185882882,
 0.7047843223124752,0.6762003340234359,0.209878434472496,0.036609192170956595,
 0.7338897970009571,0.23443754533758354,0.24794587624342534,
 0.6800800138556121,0.547343799240123,0.33901416243026516,
 0.007899375597644642,0.14293727900909914,0.8115285214370495,
 0.35654471701142754,0.4621163029133587,0.2711172160309755,
 0.05893544215698798,0.4479831927975617,0.20309143791198336,
 0.9497385088541538,0.9327592883736653,0.8994185543887113,0.36142359567527027,
 0.06388828975312677,0.5478600074299187,0.4424448666502748,
 0.21047869534413688,0.37415233429602734,0.3032783697001554,
 0.6953617647580548,0.6436219108811069,0.2446770027674927,0.3875501826365705,
 0.6895337458442212,0.970446230195946,0.18214052444535123,0.17855984097111122,
 0.8556061374327552,0.8060743771432338,0.22760344801846144,
 0.46601191073889725,0.5381725821399633,0.6882354994071286,0.5025459563859178,
 0.29608508362720043,0.6105568635743377,0.3735786288519929,0.5624008624644421,
 0.14814878559960987,0.18109722332708877,0.5195310898017851,
 0.3340641272839825,0.3540892406546463,0.8326150456300183,0.00934275891247005,
 0.6395375424979003,0.8003671022698664,0.5183628905683576,0.5167589966955096,
 0.49142859786619686,0.9127041255959738,0.7835156530101743,0.4403045910502792,
 0.02304002793930593,0.652675304202239,0.2717776857077525,0.9989043743660044,
 0.5207785120573101,0.44379852535304254,0.2795144720347652,0.2835995448238995,
 0.07249062357774372,0.9240505040091003,0.5838290785795203,0.2857764975130922,
 0.8501001255446247,0.5815975084485159,0.3535912345887223,0.15293942814339145,
 0.9218937677991884,0.9723164497761386,0.9471464953136861,0.06875607909073955,
 0.06374733666261645,0.6385325938270818,0.423680451104159,0.4226783412740185,
 0.17637265244985956,0.5482205130182772,0.3785372013035425,0.7995879930429494,
 0.30684455908801445,0.2838063098418706,0.8291711129457654,0.8168631633161215,
 0.9096966538240288,0.20996936363138374,0.145493667238495,0.21158276882459248,
 0.27731230159536735,0.19810456920371255,0.4433979859611703,
 0.7186507695509999,0.44409713341578116,0.27434137569969475,0.778276654220688,
 0.28651178177261494,0.6835039794740758,0.23821084490643907,
 0.9272142289753645,0.48323052074372386,0.10428907047905966,
 0.5398579744158304,0.14202730486015136,0.08200592559485942,
 0.7057916448486266,0.2725961463690939,0.09849002940398921,0.6780110982764403,
 0.8897788353739592,0.3223775815141442,0.645814630045403,0.28793083618005755,
 0.3447828031134008,0.6205154939311585,0.07516775574573853,
 0.04561667444585915,0.6264577943508695,0.17871968926232218,0.567262609465051,
 0.493584673217519,0.6134235908460914,0.8422235973317247,0.19991590219372435,
 0.29953251770870937,0.07699752374832103,0.4755956021889711,
 0.44382807238495947,0.30702683415341125,0.19287945343323565,
 0.537519145093917,0.29080561619485,0.8464482275841332,0.41363841958225456,
 0.5069498518091643,0.08431122995373408,0.7104082348932395,
 0.20993591646967635,0.3828630065450038,0.048498562442308324,
 0.7698609075614529,0.8613037409423833,0.4174778494982436,
 0.011150374822956333,0.27057256211387803,0.36234117599807036,
 0.7040547483603727,0.524186038881274,0.1687066256549583,0.7860725819231285,
 0.4242182931839348,0.7284634028768439,0.21741395612262537,0.4681155196841499,
 0.0434896746099922,0.6442929770711624,0.5228013126095301,0.7390586063944474,
 0.31448430660012516,0.2851386208514808,0.3375518759467202,0.6586128895094998,
 0.7975360219809563,0.8218276827344528,0.014345883398594106,
 0.1358143912465961,0.5358454353164017,0.3452811209546123,0.9881915667772007,
 0.47175322630644057,0.7510826590823483,0.39360830996961826,
 0.48200136305028296,0.39771462947309255,0.14659695578811982,
 0.7672836238926298,0.6175693445060731,0.6959074897825522,0.5333608497242103,
 0.9982338146408971,0.5694011901644728,0.9258058772155141,0.9382442877478979,
 0.8860022986136811,0.5023557123752075,0.28234868929529,0.09612519319948387,
 0.06783368402126988,0.216630301889434,0.8448986369244027,0.10929540917530711,
 0.1777437105602897,0.19708114301305402,0.6658750131264706,0.3125971994067096,
 0.6463326824118257,0.14050472742019793,0.692256659442895,0.4731005698412031,
 0.45629446271755936,0.7998535695055212,0.49411602422484147,
 0.3583032720042122,0.5247172289837655,0.5756337717441026,0.21013225425159587,
 0.2842515837681723,0.33331097434410484,0.9444568916935058,
 0.15570124132887253,0.35736549476233004,0.2041051639333311,
 0.6787220859719608,0.7462755752597136,0.023495839720433898,
 0.5322328951844326,0.18421233774899026,0.1787470078773189,
 0.11887578115347308,0.05191001599397482,0.6941475407642744,
 0.4550384070028435,0.28458170861167653,0.8578308508215291,0.5743622989026791,
 0.5706736329762834,0.3348660269192454,0.575305947163548,0.12953176912227415,
 0.6006939335979813,0.38068629452710323,0.9473569452681749,0.9776383150722266,
 0.8613732878990471,0.6173895318288358,0.27190676902178035,0.6693672550494976,
 0.4243574238397836,0.8632797787129013,0.893871214009933,0.5100959324662077,
 0.4741293672713166,0.4516674728024761,0.02978325726253217,0.4520494053532246,
 0.9224521656244511,0.525915359603477,0.27279113110717323,0.1601961301397774,
 0.15772081384076286,0.5892307083270414,0.06013435075332951,
 0.01877209333899721,0.7681087363244314,0.9411378466030323,
 0.11541238255361386,0.7394855857665286,0.746367991947226,0.7616794198939578,
 0.713529323517113,0.2240998753414416,0.8549540802516534,0.40240535641709485,
 0.30395357258095534,0.428410900589966,0.12426305129340143,0.7561552676299224,
 0.6851263970218456,0.2483877594089955,0.5493211996639185,0.355901628360052,
 0.8119058736977414,0.765651669292162,0.11473912572904332,0.9829092818155031,
 0.14618734001853717,0.31547943936434375,0.9233891075608477,
 0.7863808386079515,0.986164355016662,0.2897610350382209,
 0.0056203158040217804,0.2757299520001266,0.6036119701001912,
 0.33388935556228516,0.7935830608210959,0.1293674653277922,0.3849008604482477,
 0.1938321710869958,0.19457117150870462,0.5942237191414768,0.8718334295153785,
 0.5533984127083315,0.17637251512639673,0.9175865493296494,0.3947563314892669,
 0.8593139486231706,0.9711879772837471,0.17027507898151473,0.2658917912963308,
 0.8025788874959319,0.792165385333683,0.3862475349371626,0.4414514886757427,
 0.49774125608947806,0.28844413108980194,0.2514745604799813,
 0.4087791225800852,0.3453380825751753,0.27083831761055654,0.988781806374202,
 0.5833041749959917,0.5924692842049185,0.426519382744621,0.9310393573989776,
 0.6071497847147969,0.06919357417369987,0.13473933829080886,
 0.06533553566137285,0.5633381670756116,0.13223884491549354,
 0.8717551875875693,0.6914386659649601,0.10555653883047822,
 0.08457845787676388,0.4664049908384872,0.7019520201864232,0.928384809136499,
 0.7479049409311705,0.5952417485631503,0.6661472523030024,
 0.009026311313594437,0.27565005197239034,0.1497070413153655,
 0.016734178102910002,0.044009089313512106,0.30244585088715903,
 0.8983898085435682,0.005036181581521082,0.8059432998503484,
 0.04298595471879407,0.6176539325392795,0.5590254074848898,0.4863604010258783,
 0.33762152375114274,0.38966128603606465,0.49956748847614385,
 0.7074778396246595,0.5760890331541179,0.257009630801058,0.405675145935936,
 0.08273389519897456,0.8092763718229417,0.6233205840944247,
 0.43454344606148076,0.7919917207286885,0.05546405414121658,
 0.3213525359472069,0.1586437993124381,0.6062725906908427,0.41107911466634584,
 0.3192785003044025,0.6185353723035503,0.330134627176934,0.15343762486151646,
 0.3037250832960555,0.3228282996469287,0.0029956448786008227,
 0.9240743689175575,0.6636225730878479,0.18332534714727589,0.5650725466350023,
 0.7081140728458991,0.34490010868720344,0.32837372701753065,0.70247571826788,
 0.8453822952363599,0.9946459902185438,0.8477342544207321,0.46437114853908545,
 0.1306080676267919,0.5060381493730775,0.5386359808671172,
 0.019627614486531608,0.11036220286469911,0.6074705310500209,
 0.9222693667715426,0.382252522140894,0.2738504311906583,0.21819008300188847,
 0.04032971377971095,0.8596387690309446,0.8951869457986292,
 0.06555140356302569,0.1444231610435085,0.6319314666665079,0.504312275992241,
 0.8090965296436267,0.49410169417790284,0.6560162253415553,0.5867186219816679,
 0.1546448475742751,0.38570882586729294,0.3177375519752005,0.9445043494321108,
 0.023517546642241816,0.8644261157255082,0.36442606158701096,
 0.8243858130332293,0.7555104525433929,0.028369324373371474,
 0.05372153381584155,0.5674807505289308,0.27828560585640494,
 0.8856284951840496,0.4764313025578928,0.3714354186539909,0.45741552994727497,
 0.6454129208509458,0.23562564143816767,0.5967645693734296,0.8265493101141572,
 0.01682946860405443,0.08031081822013109,0.3817708052397808,
 0.23989835700660334,0.24272328639827112,0.4799229724998503,
 0.16853941886318058,0.5214079641126113,0.6814212756969743,0.8897799357318715,
 0.847465833389309,0.9278697930404229,0.45238892570492795,0.4390754545371788,
 0.5628342382443245,0.09775261574410343,0.7645030979050379,
 0.43184972174893643,0.7102589249365376,0.6690636819626542,0.9833772389022418,
 0.8767885945484224,0.49046480378743473,0.0970376794623502,0.0892463332068445,
 0.30965589582049136,0.3981403992670388,0.6109884254565142,0.6122475996547244,
 0.022161314880656757,0.20378913274584431,0.07149310452550728,
 0.7465957986009581,0.6162391169629988,0.4470755913503621,0.6840942570776655,
 0.34198936184604833,0.07427592521381832,0.8223150174459595,
 0.24906138176040904,0.8976387883993702,0.7378201981184889,0.4655664671053722,
 0.7162169015627596,0.06635405439820263,0.8343720089304133,0.7537977497375999,
 0.07181571817316601,0.5003230165580188,0.07218496603315105,
 0.1319608883297484,0.06700690788750496,0.1203522293109568,
 0.31281793457860585,0.10262193731989155,0.04126753446860998,
 0.5183215814631532,0.09384148721240115,0.8055008245957348,0.6374648200180523,
 0.7094935032009045,0.8609512640697041,0.5294282147584863,0.37903375166303366,
 0.7635756011449598,0.39474997670532097,0.636626319203754,0.658078503514495,
 0.6548064202480209,0.13920071770653109,0.714373030751236,0.8707518602981715,
 0.31169399847135537,0.5115610863944573,0.036630053327721956,
 0.3595962578325955,0.1668623342783504,0.7634117095135806,0.9266172937948303,
 0.4957353332022525,0.3915637098145136,0.4444569215617723,0.1353280568014179,
 0.859719012667796,0.5657441229041679,0.8564523254434998,0.07361402291223473,
 0.1174457428816823,0.7685781413514696,0.7544926456028812,0.4590253999088675,
 0.07518225707492154,0.46715079509180224,0.26697077692186266,
 0.9972893860041515,0.5050005494652697,0.4859829503808769,0.3168021213578054,
 0.6395168119249336,0.9321293794703236,0.2482627433510579,0.2959708449324481,
 0.9538078853605824,0.6394468860223458,0.8935597302805051,0.9135274438224901,
 0.23142203252158333,0.8477850129437423,0.4486592729085549,
 0.12469184365100416,0.015741436163449363,0.5527131848403739,
 0.01737262939926265,0.08077192656739152,0.6292040682641235,
 0.6774026353497951,0.40188548857625683,0.1814749090023584,
 7.672302953650734e-4,0.40811883950861183,0.9216942771534142,
 0.5435624702704377,0.24348079259552124,0.7846690857959229,0.7521818510823891,
 0.7016504741595025,0.770714822953392,0.12561025781166313,
 0.026705259510650947,0.9704909931012533,0.9481303598097024,
 0.3535442934934936,0.03352557580174165,0.3497368977779469,0.7955819888945077,
 0.613399245714716,0.009816729074185848,0.20903952248878255,
 0.050714657665552165,0.32359014892105975,0.919274660511825,
 0.46046718954406807,0.7026599489087103,0.15139781247751427,
 0.1793622343879705,0.43231366561729767,0.7334585094580013,0.7746923464442466,
 0.9744143415550669,0.6641322879000564,0.9627218545995562,0.1480412328518469,
 0.42471360850191664,0.5266147831086813,0.023278014126027813,
 0.13954887246815229,0.9310823937763377,0.649561547058129,0.17137091110093472,
 0.39061860832936657,0.48577773689294557,0.38637648351425047,
 0.5610325290616099,0.6331068428808538,0.4112409558037632,0.4884455667568689,
 0.5536895186423708,0.6054181478585483,0.5645947503441606,0.8197758963186913,
 0.5526927384198147,0.6808014432060767,0.6848327014488232,0.8990509108021328,
 0.6553427732020605,0.4919450781002488,0.5073102109486369,
 0.055398280370245634,0.8045891232337476,0.9657238765991695,
 0.2763404559938518,0.37497273217971616,0.7963432380427972,0.5792193413647425,
 0.25997648078229085,0.2547648116191633,0.5754112888404911,
 0.43178796011996634,0.15173032424063948,0.5161065598270581,
 0.7279864772795198,0.03415815067493777,0.8806337031266116,0.8792410777874439,
 0.3562090039191377,0.7245787261363865,0.17058636756919932,0.6927437261209205,
 0.33050399365819316,0.7495791943239684,0.2944531863051061,0.3772204683916214,
 0.9074203947301476,0.33057892682765344,0.5475407587859793,0.6922854575876681,
 0.7591585766238915,0.30910796672682916,0.29331762820723717,
 0.12103511287589797,0.10933586351656421,0.9408495783867055,
 0.3108963825514126,0.08838523369361562,0.9500596690972107,0.3703986815968321,
 0.7985016646921078,0.8605318249194787,0.7173719872468401,0.19740071631573974,
 0.1663631028767142,0.3073525595150435,0.9097715442941481,0.9740773931271036,
 0.4010639269580608,0.7086040380532713,0.059852094311889426,
 0.23769169556311343,0.748423258954599,0.49539695533063033,0.7215457724526111,
 0.6873953702501683,0.9977141644559836,0.935976666442552,0.2154184490478115,
 0.32839435970023323,0.1719084673347343,0.17253680187907294,
 0.40629948699403673,6.608294807866066e-4,0.9990826967838156,
 0.8496391436219994,0.1617025537151855,0.7524173353482122,0.6513042474891133,
 0.8314834641488342,0.3734636017417936,0.26700069460645126,0.3902095608923052,
 0.06572738005093282,0.14190914560511125,0.7927757102026882,
 0.7442468724086442,0.840228857397952,0.08936457107652362,0.22459687963474706,
 0.6629806136738041,0.07836282011474793,0.6408540856214574,0.6619411523020539,
 0.257671968153777,0.9582872049161464,0.114589959608704,0.23495539025000634,
 0.5713572089726068,0.15838997342526284,0.18360734645431653,
 0.8153364187627058,0.584538386615497,0.06059691840316095,0.8873012042182271,
 0.9719732498225222,0.39142346109304293,0.6484150473383199,0.2409041973993361,
 0.6388718502279682,0.6891350563710043,0.9406633165782935,0.3305002224631458,
 0.9553998131905977,0.10090283322086435,0.37693342217685155,
 0.17225182114471138,0.998657053575947,0.939940867011984,0.6488067436883005,
 0.25406636464585375,0.33434513893248197,0.36625989387476976,
 0.8428665034594434,0.13974946333278848,0.05447077538561362,
 0.07842627017702952,0.4234870920778032,0.3285274286890357,0.9802994604433392,
 0.21568329609342696,0.7635511944994536,0.23534226963668825,
 0.4320514504334405,0.8539630630132364,0.677676704472491,0.6601334358395102,
 0.21236297955056516,0.12934817187667402,0.009987498366639813,
 0.30034250964657117,0.23858182518300575,0.10281941682103102,
 0.9154292394987242,0.9022965963460435,0.9895851446626803,0.40020200761061986,
 0.9716354360012903,0.41427797124648225,0.3337148113718905,0.9581487772746218,
 0.8152076859927339,0.6489505391162734,0.7923375363788285,0.9518581617843936,
 0.8718134591741561,0.9179845935929775,0.8996889072252844,0.9140114201426275,
 0.13419024291209225,0.30954175826036856,0.6646103325541554,
 0.8050131930310459,0.08896164605020063,0.8821950355948736,0.8928380038758478,
 0.6410781339919625,0.93612034826207,0.2590950249235373,0.7462084928377263,
 0.5694320554852843,0.18408211432260513,0.5248301998725304,0.5961632711925074,
 0.7964642695013322,0.5245943659163295,0.15300633286305676,
 0.36525241269693975,0.874848340327472,9.119070045725541e-4,
 0.8849261973883022,0.4621795879962133,0.4074098797258916,0.37245089903226014,
 0.18287159389991536,0.9604584307878584,0.030648315982514496,
 0.7732434778015204,0.3429315491315936,0.08181787729730106,0.8458242449465295,
 0.2599600134489882,0.02426061360897913,0.45451879024522945,
 0.4159880699915256,0.5707473989109038,0.08781602380894193,0.545971214813833,
 0.3306204507848145,0.2476044411769689,0.4490553368084267,0.8454673443623172,
 0.02733667201324841,0.9840193938953197,0.2176635164082782,0.6896702460577573,
 0.025921927064841976,0.5527210640811347,0.6343555076306652,
 0.7289493240601423,0.311314917302983,0.15920505559395315,0.4569567400825362,
 0.8437246973879552,0.9854212191925764,0.3521573154672678,0.8314391657995025,
 0.5233332795560452,0.6719995883413677,0.4585118715898203,0.8880883662930895,
 0.905687532537507,0.33113337070888516,0.3275057310947732,0.42514450834900797,
 0.6633229405091793,0.37157809403552944,0.9574820506500562,
 0.44173898975735437,0.07386218165430658,0.7957885882323533,
 0.3258182480572843,0.22498430329680696,0.3202917211522902,
 0.07064708032240075,0.2866416761596684,0.055238220546248606,
 0.10816705859741949,0.7068687779882105,0.24659771855305068,
 0.0880635034286148,0.03396906980636938,0.41113688300908113,
 0.16640812394923554,0.19180078502713827,0.3582078805239832,
 0.3530913883540463,0.8193369826375548,0.43718405605647015,0.8676472822423573,
 0.6391280500279828,0.7730681399978913,0.2938417021220834,0.8264029846195379,
 0.8474446196879141,0.7491248328453335,0.9534217187724175,0.37617089417538885,
 0.9680746435421539,0.05917441399763146,0.7694048041294682,0.7430693794842373,
 0.8971412886666914,0.8024981926193606,0.11788689744662117,0.77211263623612,
 0.1673959398887428,0.19255898395140525,0.7273297315677112,0.6520683947067386,
 0.10050954601184203,0.5696353057842604,0.864214392412764,0.05690590579161259,
 0.7559452076927284,0.7668760681252602,0.18289218444758581,
 0.38687439901105947,0.7173568679279071,0.18126989084733114,0.456221379901835,
 0.09100124305263446,0.7441139138771208,0.14917104409655635,
 0.3121119324881355,0.10287712741826804,0.5390695112088252,0.9592806427053474,
 0.6941501233823374,0.5796295559623217,0.4779161556903029,0.18368206836862977,
 0.6212224885237276,0.0637606394468847,0.38631705970049335,0.758269408203208,
 0.42147548349313,0.5637605762357456,0.09986821648403843,0.6244427154539763,
 0.21020452711427318,0.6978963711031583,0.027372312432654033,
 0.6984158345588499,0.22704653944887543,0.9450411241118328,0.6514123424317987,
 0.5829575342495996,0.3957847498638468,0.7095329792734596,0.16933263057162184,
 0.8482196840690028,0.7127460461947737,0.023741184871872045,
 0.3479929335513954,0.8281566639028412,0.7075661745379163,0.00279202038442683,
 0.6365307409055305,0.7037048228051435,0.6049329139538661,0.607788539886581,
 0.0010200283573778002,0.7561291021595581,0.4542380387905698,
 0.6099098470516305,0.8315708425558778,0.6175359130437752,0.5885921639263259,
 0.9496732604553267,0.9603056012117638,0.5420718056607885,0.922802592706621,
 0.053473057311782424,0.8078425544018196,0.4859710211544954,
 0.8211039319216961,0.6808679633252595,0.8278436484713427,
 0.009471987643753677,0.09053196165365396,0.08228867392448658,
 0.934328371363672,0.765421646267642,0.8057967333747831,0.014213790625111189,
 0.5127246635694632,0.8453801849893979,0.4137944458523066,0.2838212492943293,
 0.4038348357834185,0.6873107347940145,0.5413813356471745,0.7640700159846359,
 0.01945303019994382,0.7337527700206877,0.8272544115824736,0.9684262202371827,
 0.6404319196447393,0.016419979558854103,0.6428349653277052,
 0.002025741888693644,0.845517043898431,0.6975565662459527,0.2506845046935612,
 0.37560432826749124,0.3914350614432256,0.9967997526153591,0.9588905754764877,
 0.38505990473754237,0.31630119932582246,0.5055178450306073,
 0.7094956305858121,0.13180878716993993,0.970817032874813,0.454616237820746,
 0.3735815634550088,0.17808440763527666,0.37723101873794307,
 0.9356730405649627,0.7422395216919628,0.1164225766811553,0.5261423395710005,
 0.83308270842521,0.3365431166052132,0.25841403669057383,0.3679991558571185,
 0.813989418910882,0.9695688071469096,0.7152347920126492,0.14429431632745904,
 0.9314673604647438,0.8384304530746655,0.6755655114110197,0.8495211509953865,
 0.1068921341821969,0.6792045761786064,0.7643369945241689,0.21103356043028088,
 0.11124494989461664,0.9081729924059019,0.5233637517672953,0.8351831163132647,
 0.04757097196194615,0.9134695466315472,0.7406051360538362,0.7502693034585823,
 0.9854226029240819,0.20153471430165626,0.3024074513152081,
 0.21259993633038599,0.7522161706476351,0.7784240107117236,
 0.17808674187326778,0.1748251853156726,0.5510851143563601,0.6190054504791841,
 0.7585947920300747,0.2855212689007549,0.5608849563273184,0.8120143177491888,
 0.8210602781427336,0.27289591803500346,0.975551728538628,0.02615632008387281,
 0.369974834983692,0.9277691775584639,0.25448518035539414,0.47766730341263897,
 0.4718012556213289,0.039055598984412576,0.03452345510681898,
 0.8942807555274097,0.5815649690216909,0.738294521956255,0.12620451284508083,
 0.9342362372130869,0.9126849394654911,0.7840592468531637,0.2860430796582205,
 0.09423246745942115,0.9075672560340667,0.04882115739965942,
 0.2686984101621841,0.5986389801323699,0.3035820620172325,0.04557655040532471,
 0.3623608867019201,0.041864716227159926,0.11332787162254554,
 0.5436267258944688,0.9320435415471274,0.06374674550736603,0.6990400122036279,
 0.8858206236496718,0.9009341960041207,0.3835887987106874,0.1458560185250375,
 0.9952317777829642,0.619189327726432,0.5162270522570757,0.25281390939940207,
 0.578893526116702,0.6497196231145239,0.7351539667281387,0.24859292472465555,
 0.42198559641679756,0.5936431169665433,0.9107942754602211,0.8489000117993636,
 0.9894096399205122,0.05084882225845033,0.02164606064789809,
 0.1345382399399112,0.040435342198722135,0.5902554774288823,
 0.28764072170187394,0.7142864222295372,0.36854050776969294,
 0.015934475819181615,0.12132285602492121,0.4238163678878708,
 0.28279374850013017,0.3460996050336037,0.322254734713331,0.7639973533543647,
 0.32667932678940403,0.7053845838987342,0.7903290982872073,0.3340346888642477,
 0.6795838704376312,0.2835015185963292,0.38183421398946216,0.348079405209361,
 0.2755898307282304,0.20018343451375364,0.3366153797350201,0.1455663283286185,
 0.4999755390405791,0.2763578776511303,0.9746333722538346,0.19866728697425573,
 0.45339691750644695,0.842789218117962,0.6982012563248984,0.16140406495717785,
 0.4163594681104992,0.5063052000722614,0.16069734117882928,0.6040908958356683,
 0.351946402626067,0.7086535707556292,0.5364000475018874,0.04856054898215878,
 0.5029057483643284,0.2830812561051733,0.7013192890128019,0.18664428074908102,
 0.2116179340094324,0.3175784707007472,0.2238495146730458,0.6646522288688672,
 0.5968273596048631,0.8351923211028959,0.19053988676882117,0.8893399275237864,
 0.14010383159645456,0.6394391828886941,0.9027287676858414,0.616721590344877,
 0.6665608732805278,0.47579420202303835,0.4388673462697814,0.9035251263607516,
 0.2471461552004396,0.5311577540358537,0.5689693967198843,0.05490949466339956,
 0.722318266684504,0.026900484119085633,0.9335603268676796,0.8923648886320169,
 0.5263829273424128,0.5615702381925353,0.28974965230773975,0.3832781528152769,
 0.06069050880941074,0.45880543362004267,0.49314924415766814,
 0.026030055479690728,0.04818842880733076,0.8645733686556499,
 0.10946553998943542,0.3310559034491172,0.2728057451549153,
 0.12173317946843731,0.6363329113098624,0.22000461192951393,
 0.22788569908820877,0.3468761078901772,0.1428361701264036,0.7597466924511463,
 0.6262523216359618,0.21729471285007396,0.1303714653422693,
 0.19226423676357796,0.33777964990324805,0.8178283439784846,
 0.44724770811817927,0.979869432250127,0.14686703682219493,0.8864383030787067,
 0.5312707991995925,0.661806844007595,0.7040254332477194,0.05115269288794333,
 0.6975663974647347,0.6246296227719798,0.7684605424126293,0.8825331740963678,
 0.33310737463501716,0.49355506713169073,0.7086581007045752,
 0.02918655029829531,0.8536237542612248,0.9748291176697582,0.6189839532191542,
 0.1772181161616614,0.0032883749245391503,0.34569361648279306,
 0.3419866209708392,0.9024381532433671,0.3467195538154235,0.5025115157962813,
 0.9882137845899952,0.7431205192187923,0.4931621999264615,0.2430752481927143,
 0.6424820435588275,0.3721716612895103,0.5695050916356098,0.5018714166216052,
 0.1882027725181279,0.0812847356280263,0.3896554937377109,0.2081098844792696,
 0.5231197405858657,0.9444714859392342,0.34040133106139836,0.7335903103127022,
 0.7080904765286647,0.27751684858777215,0.48055094451465574,
 0.7167191672816763,0.8014547339139555,0.8704911350598743,0.5843635203661621,
 0.22502516089909153,0.9999168946943991,0.13357628496457585,
 0.1996134083047887,0.1526186299051907,0.9753868167104143,0.10662597314255384,
 0.5862831195989096,0.058013469726634015,0.29949351197694973,
 0.1590176526340722,0.949996527327591,0.6213521904882742,0.7083664310375204,
 0.1615242128919866,0.6031384577595559,0.9499768234509088,0.35501862874655543,
 0.3099188310039852,0.18537243743775234,0.8481997371282267,0.8785087095585427,
 0.0517325132201345,0.5104838622435631,0.6213675772276315,0.9232721037460773,
 0.2940082392096881,0.4072503079884916,0.632950839256605,0.6089243962588619,
 0.9484981199518164,0.7633611151061137,0.20335947075314043,0.6462914133348102,
 0.6873291711830493,0.029064925024954946,0.8978582607328865,
 0.6329817910205596,0.6274112351776788,0.4573025935354377,0.23942450398163073,
 0.5815480393787168,0.7277097084139564,0.6457706308275678,0.3395744044243393,
 0.8191656472705755,0.47982406838890423,0.9321727381041962,0.7846289075547935,
 0.5114935626395304,0.8779112366282004,0.7862944183187113,0.3389133279171437,
 0.3929044086752399,0.6959478056004562,0.773299070277556,0.03847709193225235,
 0.25439340259952026,0.02021912969182771,0.9898448278536915,
 0.7281517875189274,0.4144162160482422,0.6821954175147045,0.928500949281449,
 0.23938664952818156,0.4057603975995189,0.2557837476114364,0.7056611275908986,
 0.4518105585458879,0.5759306574755302,0.3282566629433262,0.633104044289398,
 0.015854023728934297,0.07151484299233835,0.8756772724197143,
 0.6731509308707796,0.8938926186851259,0.9597002240664998,0.20643628930457736,
 0.268504003697812,0.5071185438064012,0.2618237609676338,0.646139487649637,
 0.9929085323381028,0.1863414412702855,0.5288532469494723,0.5212223457014382,
 0.012276778616186812,0.9177162998111508,0.364701939452885,
 0.25869603145576825,0.07973558947183013,0.17611223396686637,
 0.005180662110323309,0.09918158921147313,0.32537731328870056,
 0.32015040249248616,0.2813066410621401,0.7346851193511366,0.8329302101452487,
 0.34137234241766534,0.8998675344073842,0.02755445916036714,
 0.06539607700817907,0.12697750426834464,0.8690587375654674,
 0.37485305532395774,0.4308879299020979,0.4143208707710323,
 0.43866659688858733,0.5056543919596506,0.5686576967015871,
 0.16142599621682585,0.02530797701300569,0.7747005491160517,
 0.3505855054322037,0.17377664458117137,0.9889985684835702,0.5853735489785725,
 0.4765604204422559,0.846252664554602,0.7425477923465027,0.442652640348967,
 0.6658908542036155,0.40055139402979634,0.2670206787277196,
 0.004997843512441147,0.295923981794908,0.656927490296351,0.08294787367849732,
 0.9699018591576856,0.4262767682591023,0.6627902022589631,0.8284165308049416,
 0.43065790311070695,0.009293669071739896,0.9867525612999328,
 0.9297500071373547,0.9936879690467012,0.3050526529937191,0.4470351863722719,
 0.9220033488561235,0.04196303430370918,0.7563568004934773,0.9724456616673897,
 0.8558101342647751,0.6745226031330166,0.7291198059697268,0.4334861050499449,
 0.8010158024555623,0.4497546357023211,0.06659797713735915,0.2620526396911411,
 0.9257852928467708,0.8762785530823035,0.18019869649275633,0.5804085686321634,
 0.08446881785607652,0.5462170878432456,0.3156588926229631,
 0.23519238701117196,0.847528937097465,0.8614645483852501,0.7874338923423659,
 0.6081525089154688,0.32234350515795684,0.6362101247491068,0.0449895677746841,
 0.4258849343496922,0.09750322334777439,0.5632298126085074,0.8802272533236535,
 0.3038032737830001,0.09904648926812953,0.2717294193244352,
 0.28083434066516344,0.30577187784433235,0.9938900851892825,
 0.2846760240228482,0.4688426039284638,0.23199078060888234,
 0.08780793457139457,0.17058901346188426,0.9629859554050628,
 0.5548467042028954,0.5079963964380619,0.11457740623673973,0.9944980559816643,
 0.07991320703205951,0.3313970324350256,0.18380896238118483,
 0.14937564506749645,0.49869621313116985,0.5415430778861213,
 0.1378033156145163,0.3168326349418782,0.7691986355997168,
 0.035769690188942715,0.19888422846248666,0.48861018240676357,
 0.14160216892745248,0.2213653954387964,0.6258235977190414,0.5874449731323257,
 0.6019565812288904,0.878693444519977,0.4871760423029651,0.6257497439662507,
 0.9643656839435636,0.24581583539219698,0.46492173765322087,
 0.7191436043928628,0.7084150962108197,0.8176229903340259,0.4290218599152503,
 0.2616596522128496,0.16028889742157915,0.6584036088517226,0.8595079306594527,
 0.6914857964786393,0.6900957437661793,0.016317445432440847,
 0.6389683729284873,0.4766136453469654,0.16055827659836897,0.8666050726585239,
 0.17811792575147556,0.6038933391096204,0.6724351183078104,0.5159343876103899,
 0.16263529128275778,0.6886689317191932,0.19430672912718427,0.070592942930371,
 0.8987096366085499,0.029585524046438105,0.588903942313183,
 0.11434142841691508,0.32306645169891646,0.9685801624787284,
 0.9561064101923528,0.032356311490837286,0.41861689584389095,
 0.9449197384478731,0.9693683382499868,0.9755799057434451,0.14906053205896996,
 0.08441899774357997,0.5510681455567211,0.44247478901803194,
 0.9146554628265583,0.04954708510972938,0.007448701056663065,
 0.824035596969428,0.824551475657734,0.6444459441486077,0.06316746190159239,
 0.7229364661859905,0.37727482289532244,0.932106595286436,0.5108633894458499,
 0.6305221845180373,0.4811104745677681,0.6325095045103373,0.799836310199558,
 0.5373108899429394,0.8865227041921838,0.9630466604028793,0.8737747867001985,
 0.36454060607029937,0.11512639727550134,0.034483676395530405,
 0.8718679422068969,0.6313477695406979,0.5208950122262624,0.6326272058659035,
 0.28994007610800065,0.38509954537675906,0.17554114658711906,
 0.5730719383669141,0.6548079825633174,0.42008465652825044,0.9323130855919026,
 0.7788482621122961,0.8703752334071503,0.5569815106641252,0.5109396320495373,
 0.4799871883972686,0.9445102996073499,0.2003401217784181,0.0534961626319832,
 0.23971443731511932,0.807698511261079,0.7132360140912004,0.9757151292230455,
 0.39977107050532956,0.14254211477616763,0.6769706992358014,
 0.6301750020764622,0.16951795393333513,0.13129161379248078,
 0.9280121754955047,0.5145462809002829,0.4955271353789732,0.7040773631094795,
 0.19319718027518307,0.4584296632353635,0.5877644836285633,
 0.26979483351645683,0.5636661483137986,0.3789976505320901,0.1662315024379718,
 0.2707775579141717,0.8827784876384676,0.5240642788678942,0.7188078832225742,
 0.9695254780732414,0.9352739789415706,0.6365602786513704,0.6792938809966925,
 0.8717389464275849,0.2882004961898763,0.3684740276473173,0.4316837095739977,
 0.31280770635268174,0.1314804456048959,0.03944031327466524,
 0.9497699457994796,0.8353656434948077,0.6037069311189278,0.08927721399469424,
 0.9437393079506938,0.6960109690732408,0.013553426185652473,
 0.8109389742436711,0.010742427048028969,0.25004485610953553,
 0.07987776767045718,0.3724020493978194,0.5921764358906177,
 0.06279114830208621,0.8676146331345045,0.21602693494157643,
 0.6348458797726918,0.8302446337165001,0.6802554026124883,0.576314963135504,
 0.8502399224004844,0.7592157622113989,0.636458892312575,0.9613864043547874,
 0.10992319208856494,0.6621390501638411,0.3824893209665666,0.3697781949143447,
 0.5639609527257926,0.23962446032993923,0.49386974521050786,
 0.11402323252789148,0.10808997134215537,0.9246651808956851,
 0.3589279529879247,0.26824359437297596,0.5620524461923374,
 0.33882135200477626,0.7454134176309524,0.9325835733056265,0.7231189418155863,
 0.25819292692587403,0.8913932407494866,0.08016646131713467,
 0.5679185993978555,0.9447001666157939,0.42266671118324906,0.7363949942742993,
 0.3969957378073974,0.13493053222819296,0.5335974994314473,0.8504689107229504,
 0.774213980221119,0.6569393000375068,0.053511065942096536,0.6001572514989983,
 0.5632205522797655,0.21798244401834543,0.609108460197357,0.8919477468906288,
 0.18910437268437497,0.7989163360687356,0.8961085237914207,0.4150527794181216,
 0.3026761805211894,0.19744726689624292,0.7377611144201092,0.3097925249941784,
 0.5225833392898349,0.012407804263558697,0.6408177689080593,
 0.8273947102701785,0.7600637537693574,0.22614903926521823,0.7971623909017542,
 0.73556421750208,0.7608141732815075,0.37913906034056366,0.9044620490670435,
 0.4332612717639873,0.40160818059065184,0.7910716645114201,0.9302452764744552,
 0.7418437563102308,0.1085304275967599,0.5712617374130635,0.9668597479255281,
 0.8458377170306354,0.16692941721132637,0.6222580052383314,
 0.47173142735982043,0.14314031821703355,0.197655485232066,
 0.24940770118358868,0.709445072478554,0.742590886125256,0.40569279229788247,
 0.41979215892162913,0.913811979994696,0.9001946544802135,0.8858293422996923,
 0.623533374784757,0.4053506554467371,0.22203509089564266,0.19119081403813443,
 0.21942398703032073,0.25414091050785004,0.4587256130581707,
 0.5948305360586822,0.25063632860871776,0.9940749820001569,0.5494847179057936,
 0.9837747566578601,0.3735836513164126,0.04058811235388571,
 0.23226840953580352,0.9227148283295487,0.30359176076855876,
 0.6409881144181443,0.4113364416526286,0.17421004001172802,
 0.32373568958375976,0.2058563146066703,0.6619115659797825,
 0.33337159693661167,0.3884771267176924,0.6456498622279866,0.3455507162040944,
 0.9424398123973088,0.21694711925133103,0.5038930511770595,
 0.09773702567026665,0.16452711734612213,0.11940968172107713,
 0.21781797942463155,0.658236668418629,0.6171120877792946,0.4519820450785409,
 0.7847905775802279,0.5108070635641992,0.010881717286987058,
 0.3072943261537734,0.7435084027403044,0.3011574526502915,0.22426274027241888,
 0.08250531131665584,0.4903251525711668,0.27065108072309596,
 0.15697385011381082,0.4558710531101027,0.9448035275629332,0.8373174385107658,
 0.04089889894569443,0.7812139163550861,0.5547628321092327,
 0.44266983649753067,0.7640502450060268,0.1164458850218022,0.2638089700162628,
 0.17656370845175906,0.6215408222463988,0.9329076629966004,0.0711112361467493,
 0.538085474586572,0.41458402692774543,0.7808857288933502,0.454834340602269,
 0.28655859855967225,0.10972200055645953,0.6366070411161288,
 0.11716899660873503,0.2946453335323671,0.4857455667570675,0.4317792484302354,
 0.6269676723832581,0.5258223690003954,0.07535600580556212,0.5274606880591932,
 0.4190244933668299,0.2971512850160768,0.7443224496613422,0.5798355891571189,
 0.5496032922364198,0.2455514253988781,0.4354799967506482,0.0647585893471666,
 0.5038378863936116,0.16964877398534461,0.6598455761902056,0.8098315440512739,
 0.4420552366373315,0.09233677578395982,0.5599917611672808,0.4311182769582784,
 0.489623082379508,0.03710838779623349,0.5796648562036213,0.44484288496148905,
 0.20151562994861172,0.9824710234253466,0.5016244332330639,0.678301808054519,
 0.1735502029188497,0.3929404971859288,0.6943819845705499,0.5558514249117911,
 0.5522933137404497,0.644494380183888,0.749749400996274,0.7192609053449457,
 0.8723566488463195,0.23406083042770554,0.4313489716727136,
 0.06594858938743797,0.4010095103175698,0.6318300943625212,0.7286548068013231,
 0.9514011193011129,0.005156819488531861,0.06878250925133744,
 0.956626289680537,0.8140178994748477,0.8563116848150171,0.2738330501603561,
 0.24318070576274586,0.5933514437693317,0.6199507967634341,0.8176082662649692,
 0.28549241457266894,0.9954033415925582,0.8082882552903675,0.7375431039941754,
 0.48224482918218603,0.1803579376680433,0.9518790128345469,0.5139944097183264,
 0.3032116720875154,0.8296451036695085,0.8044360148677183,
 0.054701727937493505,0.03473237885791047,0.8480382396706772,
 0.7947527893329539,0.21705320498673686,0.8328637941379284,
 0.34130952195018527,0.45688524629040683,0.2701209539277276,
 0.8140472684209472,0.8972506909868049,0.5087792662235344,0.6988419435697937,
 0.9325144304695071,0.2266996927949414,0.0736107257116394,0.2823710470412667,
 0.8134891224289279,0.3547456572378229,0.7066258356627663,0.30063825351198414,
 0.20098844696939056,0.6377518683884303,0.5340516353783543,0.6496619432184056,
 0.6098804872430788,0.7477698570100672,0.5747728899092992,0.2528383314935402,
 0.02941991280629075,0.08956566037029223,0.34749569297798155,
 0.29906633310628394,0.16272592754379356,0.14947597028626392,
 0.04274226223529298,0.09279035829215687,0.5336654339482876,
 0.38020471395684896,0.9068748110465541,0.5149739246748103,0.9576429199347979,
 0.7736044227173203,0.6427615389925682,0.28188545712497837,
 0.39349215271478855,0.12194815619574473,0.3716059292466217,
 0.9918132993983567,0.2709069500302248,0.46193837716355834,0.8811796837314593,
 0.26215404361874917,0.0689592397997233,0.15051103290455425,
 0.3695597787656316,0.3977927464035358,0.09573394787976397,0.7539831659116919,
 0.8368398565715603,0.8739399485513297,0.2691325542208508,0.02326000923996774,
 0.5994556190488673,0.008985488337376224,0.46307842094549834,
 0.8369758787098854,0.35319604978571306,0.03707267331922126,
 0.37778731351818906,0.2365723947804984,0.6829886630812292,0.7945053598184033,
 0.7313741505936084,0.037922936881193414,0.618243932444178,
 0.005796264067876367,0.15234326601004788,0.029393359012323872,
 0.35801767330691703,0.43279223490703833,0.9049473313882532,
 0.20625954201600494,0.4832324713123832,0.5445360409497418,0.4873145873265864,
 0.9322679315087125,0.35851589392911043,0.8896903750720183,0.6649520983568065,
 0.3228965663851653,0.14965569560919012,0.8138052708588799,0.9071626316298371,
 0.6232901085310775,0.2476135758213689,0.6393223129270754,0.11054010136890746,
 0.8975354493693637,0.8632823235073267,0.3730909191820677,0.9666253460199717,
 0.8627238246387603,0.534721871218643,0.691542013087065,0.16777030094153977,
 0.3462809036718465,0.1955458405966075,0.0720168615535417,0.9080815582232958,
 0.20593680083760824,0.9171737209925348,0.9071994353537408,
 0.20875181323608316,0.38192785624443193,0.7447357672599915,
 0.5072212278303005,0.7329686357271513,0.3075882510738046,0.3263430485654065,
 0.9986119679804061,0.15585518370922624,0.9874466222950036,
 0.08609072434590714,0.2094608345944542,0.26960422091205183,
 0.5793833383056758,0.4721871430155461,0.10722813336001857,0.8402399883664129,
 0.6924336440466082,0.695910979379307,0.6571713943481219,0.7570920331989185,
 0.5928963060275142,0.11621578172472169,0.21454080457112845,0.516787573591701,
 0.6445076807701685,0.23055227812208978,0.571568027381314,0.20461436388739218,
 0.713069088551991,0.1149498896865464,0.02818031456208736,0.9857579283544697,
 0.1827620350999507,0.9474053945273708,0.7665173945739971,0.11661931160008221,
 0.009791335245662225,0.629812381073646,0.8680080594916045,
 0.13974775403238315,0.61563975825633,0.18339720155075856,0.14834999681053262,
 0.207383870623704,0.12422778829661096,0.4827743637812707,0.37589899788857817,
 0.0034675445869948307,0.8607357260301955,0.8176431902848784,
 0.20478866661927053,0.34276804916946657,0.7064391377315195,
 0.7293820263716022,0.23410868406918872,0.7151686406013251,0.9606190955980503,
 0.4857824294290377,0.966037302904525,0.0656060578539166,0.8881975224743485,
 0.3773348235614513,0.6314960127873256,0.6987802492302354,0.7215708127623723,
 0.1904597961113459,0.13506160277885848,0.7244498005170501,0.8946666822483811,
 0.8645037124918358,0.40359525817235886,0.24150836126423525,
 0.18061241570499542,0.8124229990352365,0.11193683369786722,
 0.5186375456465079,0.19279557212927856,0.5881866204188125,
 0.30398528482594545,0.7482923534261638,0.8218159829193047,
 0.03767511099104859,0.24013567768919575,0.1846928709202379,0.727944700152668,
 0.37984934605276455,0.7770030927489607,0.7439226133227055,0.1288921178673471,
 0.470863469513485,0.6615958301656417,0.8723627710791167,0.6355388599661264,
 0.17000451669753036,0.9712980081030591,0.05180200433496829,
 0.5539222336722842,0.9522047734209633,0.4329459051000476,0.6762140149041956,
 0.030562752213028066,0.37615282295337993,0.9486790164381774,
 0.12914831652941494,0.34916965974498915,0.5640695904267103,
 0.9549195947780125,0.5328034894905578,0.1207398599834828,0.9926600888800445,
 0.587880051201488,0.24832155411874868,0.14188913781442547,0.7056061451993433,
 0.524048066688052,0.6686822758250465,0.8882764940924689,0.020736222717600894,
 0.3059310198576667,0.3406874586552163,0.004387007260491238,
 0.13106009943246655,0.0350313845077217,0.22264746621266318,
 0.18114104533858766,0.6057107962821326,0.05128799261261041,
 0.3031845180290542,0.7441617858255762,0.24160238807903534,0.6977745446250448,
 0.47157441328650895,0.4925656670637165,0.8579263376086697,0.1049678756455883,
 0.07831423697360884,0.7430171341421289,0.9070824857261528,
 0.04599424640011529,0.14833650038869794,0.9643623889024184,
 0.9487949669584914,0.16640169754010314,0.2847556789724336,0.6748147419479833,
 0.1201476972065596,0.4147564870063334,0.7360579751210797,
 0.051882617035527545,0.3481487241171174,0.5671401533399428,
 0.7819871337215981,0.22555947154594658,0.46246874747672784,
 0.3165391306409935,0.19306430773495242,0.090044605790215,0.3386756507322626,
 0.9163293229908489,0.8395040291244462,0.6820612345602587,0.9490272486656551,
 0.5293234204846612,0.28084457414704644,0.7052243103747979,0.6685798602944117,
 0.8929040365631497,0.8059792764181014,0.6533747720202666,0.6640685768114548,
 0.7700853225507268,0.03461248348460433,0.9191350668138599,0.9207107789511467,
 0.7260051511613199,0.623715181492554,0.2181729643373138,0.6240291261749004,
 0.8073352320615506,0.8859828346028277,0.6911576560124995,0.2475028105816438,
 0.30478163817471204,0.11347194637112645,0.7630093837633006,
 0.10278545979478881,0.1815012375620093,0.7119657269591306,0.8857275329298352,
 0.14203375996513679,0.4601202152318917,0.943353461243981,0.5037591411380722,
 0.1478988866853055,0.027250135579674928,0.5780329532156463,0.255751811740654,
 0.3310504162098199,0.9816301912324378,0.02206289408809381,0.5076081989584647,
 0.8105607275003618,0.5234007015019719,0.12919999197365872,
 0.23215127007270664,0.22377037673182487,0.7715397953810343,
 0.1879841578529602,0.019226916580542097,0.005052329854790094,
 0.9070974642916255,0.17803235348024915,0.9163241324876454,0.836728010094776,
 0.620154426915298,0.7745387645725381,0.11318479496104628,0.2569016429601909,
 0.12618672715704782,0.7750985066136024,0.10039083157001416,
 0.8773367367512591,0.2942442712447303,0.8946587016184802,0.7903449994435329,
 0.8960237596380647,0.7542413450360509,0.14663014229068683,0.6432346956692646,
 0.2042783637992216,0.28132871299159024,0.9601042305885179,0.6378661731278227,
 0.0034802248189207408,0.013808840232936381,0.48203015599439525,
 0.5264967979105695,0.9083108691312587,0.024150059742163443,
 0.8514728738000735,0.7422541087783433,0.8615496321400973,0.904675850145421,
 0.8240377490538136,0.4352373903638692,0.5464378919056796,0.81373883602698,
 0.16814663760716675,0.6734120342503809,0.7297627565431841,
 0.16812238344038877,0.03156145088191842,0.07820310921399609,
 0.7060932928636785,0.08982197013966109,0.8617993348811981,
 0.48357029937704576,0.25132693341276413,0.7034341930651316,
 0.8589044892479871,0.868512579079568,0.64711234715274,0.15484787443069692,
 0.8552709999022008,0.17651917207665035,0.49495005594038244,
 0.46153127335985655,0.8277827040919238,0.6208489962529409,0.9500255950420866,
 0.08650059554984835,0.7825984164626963,0.923011033947781,0.6976770769807428,
 0.23443430182605762,0.34018220710445934,0.3829565960152109,
 0.22128572153241988,0.48660554950380364,0.9150227614721089,0.498460574371847,
 0.7793951178721501,0.026302236560800885,0.4022180815211349,
 0.12182664942927379,0.325370374541165,0.3934423039033249,0.32383753726198217,
 0.10145643832202478,0.8052307652465271,0.3035471741310036,0.5368186455560863,
 0.3155359102456369,0.0905620831388465,0.9074474516616973,0.8019654321585306,
 0.42719057806021465,0.3932889670106331,0.8414371239040989,
 0.49995731966602686,0.30228846935938747,0.305803887118565,0.7295840481004272,
 0.8697762773107427,0.5018003174370791,0.6170036640845384,0.15597564040471945,
 0.03748598155410787,0.8966060877032093,0.9344014873946691,0.6459098508248317,
 0.2040912758210085,0.7024216726567646,0.7126218505029391,
 0.042117901901506105,0.5012125996414256,0.2729866781069128,
 0.17760720482336312,0.35652880084710903,0.03763691064015939,
 0.48559358068104586,0.8589138546646958,0.8142997465209704,0.7882221878071478,
 0.7357428925224783,0.32332030234731934,0.7533773151694081,
 0.003920883980422896,0.30315604567901333,0.7496371077607724,
 0.937311459931766,0.5362589040785235,0.9196321337045685,0.1602018948344548,
 0.33612924029630653,0.23534471389934553,0.9178300351540936,
 0.40232718729636874,0.1489332411056622,0.34837536812758907,
 0.3259300083374471,0.4545193550511304,0.07280353878570867,0.8726141617188123,
 0.4371085387672916,0.24762487828476853,0.3133923935452845,
 0.28906539841139023,0.18457024622635865,0.3141221206081817,
 0.3364906097296102,0.7050092931101852,0.8101985354558376,0.28951609732386907,
 0.9460030738878918,0.8227570302907551,0.5308016294124307,0.6042946815815062,
 0.2245888275236918,0.11751985345366289,0.7588705083692991,0.4018757786197491,
 0.24567367009771746,0.14435367578494074,0.3739106796738588,
 0.11271942776862753,0.4029429037137211,0.17676265449962214,
 0.5104837053630001,0.14578150556468383,0.7124484366910463,0.7791062544926782,
 0.030355040195384042,0.3894203836089114,0.6629885979651781,
 0.3761100527202085,0.5701468418936233,0.6923279938131417,0.9455668129565622,
 0.10634945505953486,0.8218323884227086,0.4027471564394245,0.3918871574591719,
 0.28185656072446863,0.6565275896069932,0.033018603403160984,
 0.8593921992011536,0.20262825919570993,0.027821553038795632,
 0.5015643332017824,0.3682761626458442,0.27161878195858447,0.5921880196843229,
 0.08239059943490612,0.5770342080687224,0.34424619687055247,
 0.4218966305954269,0.8795865987358616,0.5290298939007648,0.5732499431434247,
 0.8239677153739481,0.26943165177640893,0.6494195108509861,0.8597329145580901,
 0.6210566926373331,0.35285379216801793,0.6122143856772498,0.7715897937158955,
 0.6176552582448536,0.16653863671256985,0.3011459430281489,0.9069617025533175,
 0.23605134832968633,0.8645374122472851,0.44643747212214224,
 0.5837313909248658,0.11960677760572114,0.9231771252244718,0.5527310439951121,
 0.5911825877883663,0.40231824702352037,0.5601324437321635,0.5557602607666314,
 0.9777834594285106,0.09674364089885334,0.9623767574013784,0.5143140473885442,
 0.3968207135447295,0.9437243667019606,0.9092810972867349,0.24950563747941368,
 0.6222315822086723,0.4307644226617785,0.6223379299706572,0.3280061669597897,
 0.7871765769141906,0.8743570351517992,0.8300615074213579,0.07984746774237483,
 0.5385216396751177,0.3944182256628028,0.16716433184704937,
 0.027072992933173667,0.6109574909402641,0.606713775408558,
 0.27484653584839913,0.5351731201309279,0.804120357233728,0.14562881600816813,
 0.998019979621218,0.9791818025863309,0.5027913337274751,0.1811960826446204,
 0.7136805630168612,0.8599937390747882,0.5407398821490628,0.7854723823486356,
 0.9144286630727515,0.2791726495212481,0.24199821518966447,0.7676677488606454,
 0.48864270220995587,0.943608698622711,0.9926301468221931,0.5251874346793374,
 0.6804785783629829,0.5490290955356107,0.6461556139057523,0.33677329367340814,
 0.053402332253010276,0.18211249938357255,0.455688821775166,
 0.20195625921748395,0.10264819832744054,0.3328241837864042,
 0.3685968336144959,0.1332510251183887,0.42427199238450486,0.4414401719181207,
 0.004679391984889536,0.20221546072050478,0.8327394702441789,
 0.7377717609604431,0.57778128422646,0.39894837117982684,0.694185230424784,
 0.7384306142212635,0.593915163589785,0.5304890397405196,0.87964695545109,
 0.05613160919889659,0.36561212469712245,0.8698996110486423,
 0.09406573685855746,0.3387509236167183,0.3585727303678865,
 0.055474406345989324,0.5209855883356863,0.47722219220327367,
 0.4251333591888161,0.16162865765645384,0.9809973482121672,0.3277112560257712,
 0.7653742988710145,0.680258786651371,0.5914195497775658,0.6484042488544423,
 0.8215444926584755,0.8627488517888069,0.6507876283845776,0.9128563094803618,
 0.8285906265499952,0.9927216721076046,0.11024117955291246,0.335386508461196,
 0.9738422133002578,0.6315111765741894,0.8815116303975519,0.0694551915502496,
 0.30504494487027456,0.5414137537637538,0.6048793780859009,0.6458608502316583,
 0.7500339969805483,0.30815626465369605,0.04348056745349527,
 0.8821353366209459,0.5072691541702474,0.6498688685935298,0.26205580980218135,
 0.9306457088963913,0.8496461965274429,0.4080676179111604,0.6547616546071243,
 0.7642401595370949,0.0563833403731534,0.657317917049218,0.7106999135811195,
 0.6106995427489315,0.1954368669863178,0.05302814255166033,0.2758555177583957,
 0.5714882750835715,0.08586084169507213,0.127775262698412,0.6611030194455926,
 0.9337247820627312,0.6199439765394407,0.20423855488342546,
 0.19323854698696685,0.9974999748719415,0.37016173554346765,
 0.006323940243148529,0.5392627373282425,0.33213228700811714,
 0.4866735123920124,0.2518024340827262,0.039145703367622575,
 0.28462345798298816,0.40654155339781906,0.8818794554353443,
 0.14864603777557694,0.3891425549396952,0.8008135799929458,0.8903934707683222,
 0.5107799325310106,0.030219541165511843,0.6901705872155958,0.583870167016943,
 0.20991417803491164,0.9971938882960834,0.8224111019556029,0.701996475293462,
 0.042789020880179995,0.20838282878129433,0.09451749688648192,
 0.2552916906509475,0.2251952849996406,0.8895959307344753,0.7707915186450021,
 0.2583041412308251,0.2404967084182913,0.2412645709608826,0.9489277453639586,
 0.9519123626502941,0.37273029923117684,0.1706773343174407,0.7804002827626084,
 0.29826958687450544,0.3074178485853375,0.3943354496406156,
 0.16336651056974272,0.27319422174727304,0.10479211073674,0.8467652386720206,
 0.5854516706459361,0.7327953750825801,0.43373119274833005,0.5517022873901821,
 0.46836901460789826,0.7631481271250411,0.1183631184794437,0.9418310370329711,
 0.40159739957093943,0.48888713507175674,0.9804934715951821,
 0.9304926915992794,0.9093748916009359,0.8794663526839035,0.4777408034587509,
 0.6221836726738001,0.6617077457880624,0.5861657588712128,0.7259059465455687,
 0.5569619146651256,0.9289060503839095,0.11408072726939089,0.4959112433558709,
 0.09461216879212486,0.8200687395025259,0.37233327861319787,
 0.4542175089886401,0.060655538815544086,0.46750784056187433,
 0.3289141838545899,0.8194556799542679,0.9280959128669737,0.1523838065310409,
 0.17065003746766372,0.4609051379725937,0.3310145085571744,0.6704982430042773,
 0.8762243961046997,0.5220103627019353,0.5414747797215185,0.5558900730025531,
 0.13945888822434416,0.5407569054705282,0.7949546464197341,0.8466826672750344,
 0.5394495555755167,0.4909433767884741,0.1836672186292152,0.691300002661118,
 0.45415918946990563,0.9281409312359811,0.23359580493813314,
 0.7425879496468144,0.15385848471064778,0.7375987967548128,0.8432783942158967,
 0.30857362283871514,0.9647936961014105,0.6073966414757848,0.637066576729463,
 0.5435756203000288,0.4906544363680426,0.31692188906244034,0.6449444719929056,
 0.3045113986755139,0.6855749425194194,0.7135623559134078,0.8332467869784815,
 0.8547518640305609,0.8582787810957213,0.4924427506499297,0.9388489296979882,
 0.9294566904740862,0.2580099440629786,0.7187718983955783,0.6320613300004658,
 0.9406747303151775,0.7755853966772264,0.17672681407560198,0.1267272865879463,
 0.267172300994583,0.180428396771127,0.13512712930534632,0.1015894982352814,
 0.47533275491143956,0.7261495066947221,0.5182258424309604,0.4084178387954567,
 0.2463538747942945,0.2944858954219538,0.6861923588922291,0.5258128289723175,
 0.6977120066681386,0.003671138794961726,0.21639042568654254,
 0.5885107742841603,0.014964179006077938,0.88112884995863,0.993900638188296,
 0.7511663148803779,0.5353874435393686,0.2721140157254135,0.36130871016016697,
 0.3685184224213194,0.7966732933617098,0.37766475958074963,0.2476194560314413,
 0.5034867656229535,0.5350991135251547,0.07782004602357495,0.9718047844134029,
 0.0345467737169014,0.024781636040103305,0.6264042884489514,
 0.4436009708455426,0.47978429663138566,0.5074445869919155,0.999465696962265,
 0.9681617147219979,0.05960888222392424,0.16996273149948038,
 0.014755554931712322,0.34759557738419,0.24373613191947818,0.3523371070587923,
 0.5597323456885219,0.2884716658271719,0.05044300597032392,0.5428309629542653,
 0.6236650053639712,0.8909779555180761,0.4277140175923123,0.1535741893868633,
 0.1712295174680859,0.3669424243819682,0.7446498778228084,0.3061530289561316,
 0.5102645379452558,0.47754160438249715,0.7268705252589583,
 0.16272953646441923,0.48678049223228603,0.4278109056562639,
 0.6867615635245936,0.2824262151315253,0.3690392127468445,0.3898337836416591,
 0.34652940493754414,0.8478507725550245,0.1589928081576245,
 0.40968454976337076,0.6883786441119999,0.19978083665994373,
 0.7851009297186611,0.8900073689514609,0.7845238453248294,0.713961551405001,
 0.06539912897440514,0.09464984401567178,0.08492013018082578,
 0.7617168519776987,0.3778976549895414,0.04202707696363306,
 0.06604356207509987,0.702705671189157,0.41960514617313616,
 0.21955961332369134,0.24662089587161984,0.784191680556422,0.7930972614955734,
 0.711719308717724,0.4541275001132048,0.17935816436366459,0.9598930596128111,
 0.533161259086593,0.07471451656517991,0.4829441865261531,0.4812809757794392,
 0.49169999341542026,0.12401654918231164,0.44067533886951504,
 0.43266669018442294,0.33511411397200164,0.9805562724044756,
 0.6381562630409174,0.0815382313778723,0.6117183272275866,0.7548363034042972,
 0.8551885489394598,0.17251139964444206,0.8873297709326585,
 0.050336010938005105,0.529195947177111,0.7602526636637604,0.3739865301054027,
 0.9158816182569116,0.3699226469878216,0.7664315946180744,0.9836604382956661,
 0.9168502701249616,0.6294922912265392,0.667400938756678,0.9920473564973272,
 0.4660605916266237,0.9365392078983965,0.42874164140719007,0.2026699777242399,
 0.25137859543726204,0.3430593840905021,0.5246629494122685,0.5454706883038483,
 0.3099901859549423,0.40970356007071884,0.9392874179374066,0.9567305500882788,
 0.5179904785112855,0.8438031501959761,0.10106891815183294,
 0.04470968645079165,0.43141246983849557,0.7881460183497091,
 0.7555151933397302,0.20316784256810738,0.6422788702912373,0.8068915418416026,
 0.3429304073023185,0.5299178848043878,0.10111043196649827,0.2301299700237579,
 0.4771743229460873,0.5148846183771055,0.8832383131603841,0.02950562748560337,
 0.8871960828522533,0.8720979193112919,0.269002836279302,0.4263489279499124,
 0.7770199592992467,0.7861855360108573,0.9955764028732832,0.5107910917555708,
 0.6103992928513943,0.28036342028157824,0.31310024399393266,
 0.43375750103429545,0.2724167373210744,0.36759719072529984,
 0.21146259440279203,0.19871447764046013,0.1928762732685526,
 0.8811489180671086,0.7383486907222812,0.7132685750004447,0.21540649098507036,
 0.8553025551979972,0.02194472989943319,0.7237139266631338,
 0.39198641945945345,0.987404149680119,0.4067739444525691,0.3702767475605059,
 0.2992510924962565,0.9786270371002126,0.4884775902715709,0.6624953921415018,
 0.3381216206208423,0.3870448491516916,0.6282343791731122,0.49245348724030025,
 0.18467636360303508,0.12344161069932735,0.9781242989575671,
 0.11510246888166464,0.13534943722613557,0.9014434107803531,
 0.08922810044839635,0.7227884063279949,0.3197684963587789,0.5538176135365064,
 0.5077432331365378,0.7032854294785379,0.26867308925911504,0.7618582617586823,
 0.42272801805379523,0.2077226544337647,0.011108984572425151,
 0.43327203973077133,0.34150949046231405,0.2267181679668453,
 0.6527161584699932,0.48165932682909396,0.8138872523455636,
 0.05298723973236985,0.8338374293753689,0.15794736031745726,0.797692030670528,
 0.6283332487953198,0.007802820139056421,0.3073385829721694,
 0.7908946955293048,0.47802577663387946,0.6230672006109064,
 0.0046904502799782755,0.6627652662033362,0.5531397417691114,
 0.9155461371025897,0.4645363773920309,0.10778957949107981,0.6372698754384181,
 0.33712408144420847,0.5422273478186354,0.17897245496072822,
 0.09689035446297667,0.20010684619420893,0.8010120089191696,
 0.8838182564286708,0.24482444790787516,0.1704749281466983,0.293986983542586,
 0.36930648497876717,0.5865024766719178,0.03434069739926404,
 0.12630377069982568,0.9367378533651477,0.6943101306484554,0.3405687213094978,
 0.5187104339067692,0.2334485057369089,0.24683034702349094,0.6613492736295025,
 0.03443710729769223,0.7319944850939466,0.44617075661231753,
 0.11442994030424192,0.5067790809310644,0.9247156860099058,
 0.26968034918358685,0.3737399513839259,0.5317899007424466,0.9172623291611339,
 0.22643703530692072,0.9913542574805598,0.1399503898447585,0.6941042853132171,
 0.9597858479647411,0.8524391785045984,0.5602888930004106,0.32411105393156536,
 0.13656279272498173,0.1534370436242659,0.3286917020555218,0.3053884784215628,
 0.8272936209208253,0.6860887130856429,0.33956456164917825,
 0.007037394497418137,0.8874195673665515,0.04400245726095142,
 0.38346952628268727,0.9532425942513412,0.6062735451890641,0.9458332685373898,
 0.43628691800256836,0.8681322904000287,0.8342669045791506,0.8853656701394791,
 0.23973030706422183,0.6230074847932238,0.42642235293343767,
 0.09407401652749492,0.28206965275228535,0.7497610321504605,
 0.36377250665860617,0.24702568043101436,0.7110228158215237,
 0.9341818715539443,0.585918631922387,0.6414301860184015,0.17244382233038014,
 0.45139140549675183,0.40034970070101483,0.11978173887422394,
 0.04330167716240618,0.38962195698916324,0.021540082409196137,
 0.7849055522160128,0.6576927213700663,0.447975893931677,0.4851761649553351,
 0.9162568028408802,0.2869183893362821,0.6113300231572765,0.7046356229833397,
 0.43626159809204623,0.2735379013125139,0.38843002125654824,
 0.46052987452910843,0.1413624882778901,0.09387031557196046,0.710968828882786,
 0.13757427869712624,0.710462175104726,0.739356628593332,0.4440113236563734,
 0.2185081146776957,0.2804688080564691,0.5532737333348051,0.780008700502709,
 0.9465988593662761,0.7281925155142113,0.699159715863963,0.5928213773851214,
 0.6689686561540252,0.6244628493016725,0.6601555124345889,0.5926364051067892,
 0.29445148424548506,0.8780869150583206,0.7460856512886227,0.5299973228455107,
 0.029511586488753516,0.03212596780276933,0.6814340272918286,
 0.37349437774213967,0.437403253437756,0.32701768944462195,0.7488011665617705,
 0.7906887368716108,0.8061056487472754,0.15770043751181928,0.5499814199962699,
 0.15359682598437097,0.8629952861188865,0.2956032671749732,0.3751795594129047,
 0.4664176176986925,0.07253160089934219,0.823749365082211,0.20179832410893817,
 0.7534768571148992,0.2231233483744458,0.7703836997226826,0.5124069071466257,
 0.5161695777843176,0.7680248600381054,0.14636411616753864,0.4700017525871336,
 0.9829182742641513,0.36762784692680317,0.8631874380541379,0.5484145202700041,
 0.9276961323407955,0.684350815956613,0.5451185159045823,0.46442486326284615,
 0.833492920437176,0.7345047638527724,0.8734897873169509,0.5097470513827488,
 0.8108187161340717,0.04622904290390295,0.12138601268615978,
 0.6036273415745875,0.477322242881416,0.859715896212942,0.6927233427911362,
 0.8536688844200918,0.06944651106276534,0.19704547604181932,
 0.5135045391698134,0.4692304673647031,0.13959710878719955,
 0.48210409170146595,0.7260750039477866,0.18399198914789872,
 0.5352309733800422,0.18085438510919716,0.3771704963468542,
 0.23891853943478436,0.255441225796886,0.7993633098459393,0.21915528613549373,
 0.7904648835897843,0.4688534524207326,0.7897766878018299,0.42854441416014244,
 0.9822908424915283,0.19083776325252189,0.0739002885146518,0.8946620101412912,
 0.2061409165406345,0.5800513176248283,0.6023037375588841,0.5877145616044778,
 0.3584829452391666,0.7360369140230208,0.9180752733464702,0.5748514445859817,
 0.46675722504002515,0.9918736823887125,0.8239384017876108,
 0.24264339345346186,0.5972211741312295,0.1583834947870365,0.5502077557633522,
 0.40083098486004387,0.6301271623509902,0.6636949357316454,0.1127612775694784,
 0.3113995255350509,0.8143941065473629,0.2738884616430023,0.4515341606472594,
 0.7007574766553919,0.30802359172968663,0.10010642074858711,0.537713349052485,
 0.5254407335007254,0.4735200508452111,0.4338772570160977,0.06264918259743357,
 0.6429072612782479,0.801267501065994,0.6333173655829342,0.4872685447721905,
 0.19886422305542184,0.5790771003815545,0.1453687042214279,0.7731571072061687,
 0.9861418901354791,0.18772901240549222,0.10058776331459796,
 0.37799720636466827,0.44243773687999377,0.6580745254079254,
 0.2794001995667226,0.12903441446212038,0.15006863064759857,
 0.5123057396117502,0.3366056309903065,0.9812526029772558,0.6588585034463881,
 0.3462845805442081,0.32963689682643615,0.9964394143911477,
 0.10350146474739641,0.46012356303027613,0.7096141949800552,0.947035425673242,
 0.21422668872249506,0.2358422346774134,0.9637942056722507,0.9618260098863103,
 0.20718154973978797,0.47347112820587645,0.9531379902419774,
 0.7755982540338289,0.8904208427441508,0.44352426435399905,
 0.49051868184600056,0.7383801184979183,0.3997302843201437,0.4631722020597273,
 0.9613905229516228,0.39658910046745643,0.9988356414487588,
 0.16537203295135827,0.2854025791320818,0.42756682193692597,
 0.6896582312638756,0.6149507488518741,0.537161899248532,0.17888700432731064,
 0.8537954651246196,0.39469784582327516,0.6029897538645407,0.9716886229317742,
 0.9236939237525624,0.7156670430859741,0.8478714593679199,0.4348109325589187,
 0.6368264859992024,0.23817612546242684,0.13027848565017153,
 0.7927737418104791,0.7049461372600766,0.27646663405748617,
 0.41518025820195503,0.6900698373959315,0.17204920497667586,
 0.35409207780129714,0.011811490333181851,0.6534745548316707,
 0.3597744860140313,0.9469763016578414,0.7175551854071962,0.6529271754681336,
 0.39884333814357553,0.47668758990687576,0.9763948203650318,
 0.38169138477577336,0.39723126676204523,0.5034023011262906,
 0.4857677976110504,0.4116645533432738,0.2927159274777079,0.2847326781501589,
 0.32085282707731955,0.4885948992075183,0.8250947821223122,
 0.15631569517460342,0.7652766745253325,0.5714421314722226,
 0.06938566041012018,0.5280598634004062,0.17945230678250312,
 0.48139334382795895,0.6425329380809406,0.0596996599246562,0.8268014982516283,
 0.7078111260550747,0.8974941562497183,0.2704558925048044,0.2027923191252654,
 0.3215629084923233,0.7052669957546097,0.4207176736785151,0.3593866661537235,
 0.520912045678327,0.9387770951119353,0.8662996322282983,0.8927483254075512,
 0.19811307170224213,0.4161586612801863,0.5850948702163953,
 0.08885858856691387,0.9326972348881956,0.839079483136647,0.780878793818723,
 0.2733154424787937,0.5391593988243777,0.3742408084657398,0.7465416994989404,
 0.689778136587315,0.741369141545119,0.5528315336591414,0.7401921021637875,
 0.07564645532863834,0.684029500304802,0.31055408324584166,0.7116502504775764,
 0.719191935062345,0.9014897236643497,0.31369305857241925,0.8138081438840098,
 0.08785026879886182,0.06778599500692972,0.22494945744614703,
 0.9714860574750063,0.6885478079257488,0.8851722491362451,0.719042089578911,
 0.003954634615562558,0.6260903214592212,0.24431825530423412,
 0.7811425499510034,0.5399005035892254,0.9853520340323987,0.5374061383982678,
 0.9402778869424662,0.9055128609725286,0.26132320742770165,0.975948009916453,
 0.08533100726179299,0.277009105634067,0.9674541865135036,0.2468318598505318,
 0.16513103522049954,0.49871975202806373,0.44555498461524246,
 0.5379168189050677,0.396551954288623,0.5370286973486624,0.41071322869241855,
 0.46441318866661163,0.866836430430657,0.05592570270408814,0.9960838900802426,
 0.39459365606826147,0.04334415625885224,0.31482156071183676,
 0.20048447769575906,0.1212259339470062,0.9014348749449352,0.0073133323811434,
 0.26392383263588826,0.8300779859920525,0.7996066384106277,0.1771739293268899,
 0.5499023828922353,0.4626072609188878,0.6260385232699965,0.3351133142951883,
 0.42714771870313917,0.44315633827635814,0.5465357445011869,
 0.7929953255651812,0.4725510605830754,0.6363930045475075,0.20070929116290015,
 0.5363108842538972,0.30476666235339467,0.6686603400204181,0.8064885831687182,
 0.8256380997070806,0.2971878152354187,0.32328050487801174,
 0.47950102778815995,0.6564441604566094,0.1576521085358482,0.0203913985059625,
 0.26599484877594004,0.884472092448407,0.6564950978729536,0.9204978569597275,
 0.06669881847555259,0.6834889739426097,0.5918129265819306,0.6023475198530377,
 0.3743558356694343,0.8672201372722758,0.42235606411051974,0.1331563296577437,
 0.9347343529449187,0.5285815112032206,0.6559003608713551,0.2634512715684053,
 0.8510470823056789,0.038701816209595896,0.961813714405483,0.8570332705490797,
 0.9853102687936178,0.3069209443761518,0.9577552528151198,0.9210473978388953,
 0.028486148830893065,0.11014146805765068,0.6350556915062853,
 0.1272727862850791,0.30459034098242666,0.182585829000261,0.3935128211393031,
 0.42570150287660713,0.7508280157542404,0.1633198076956356,0.5341815493502997,
 0.7883675167990354,0.22767390378793595,0.6930700046296817,0.6595571088364133,
 0.06455418428990167,0.07868724690062168,0.6435062180815909,
 0.7465813825518567,0.17005960232356188,0.8660814593879359,0.2079367535310911,
 0.9977117658704953,0.6337581044251079,0.2851246043823805,0.9270891192900538,
 0.4800830020496628,0.7248635299499213,0.5322403352696153,0.23851129704780305,
 0.06318524917480739,0.28734459374673893,0.9825935973939023,
 0.2935172997339517,0.19632859888469423,0.21595341349458208,
 0.24938464799306548,0.8981850350637505,0.2304057238524997,0.8963930918856122,
 0.17936981179138745,0.5899889619961267,0.3584132729013122,0.5448025887425065,
 0.3222018193173376,0.9730940769067096,0.021277583485114326,
 0.9424441559517609,0.12264151699723758,0.3155744979496463,0.2809212092822273,
 0.02591945778088811,0.33789139572817195,0.7383320673337346,
 0.6081062941139078,0.15413121846043898,0.3004267348278903,0.4734795209785869,
 0.10042469992913039,0.24795946575687966,0.45029030104408974,
 0.1258063230488483,0.01028605945539851,0.29507338824442586,
 0.4136961500341717,0.9126697379543465,0.6014248567295131,0.7845299300588864,
 4.39461267739083e-4,0.5978392471690542,0.369850917048988,0.3671148491323871,
 0.017600338200318255,0.0377071390276017,0.5088090657661195,
 0.6997518585438135,0.7635620639212124,0.14118041918835678,
 0.03829873100827719,0.036866945066816204,0.9854308990109933,
 0.12564405320243832,0.8470197196441112,0.33376726770326315,
 0.5568072186505865,0.658893051817755,0.41998189965272514,0.38776488319314373,
 0.6013251667684976,0.9727608996113226,0.252055808885164,0.4791901070075866,
 0.6044124147229224,0.9652249472629278,0.07885694075884897,0.1446357532877156,
 0.979043070780443,0.031414104310321656,0.5015541841403272,0.5310767167283542,
 0.5860007542901486,0.06517107669379374,0.5999035721778457,0.5040141634410942,
 0.18784366326353152,0.8706422681473601,0.2459802325897027,0.5568745362747095,
 0.6629468368792317,0.6450309303450051,0.9203532029333135,0.9265881508163377,
 0.17424103209346542,0.6877208363626682,0.9021780296612459,
 0.34141680233579397,0.7714702666105895,0.6399202996079598,
 0.09051433666671105,0.041347028908085726,0.30335064803423883,
 0.044981572364764966,0.9365827622833978,0.9069850086825975,
 0.8819049600191141,0.8010978537264658,0.09314513089831877,0.8535348929956086,
 0.28976493905230705,0.9026793992740457,0.30524574177387653,
 0.32384258316311376,0.34382487938501205,0.8630532477656021,0.589750796319955,
 0.7201890362513768,0.6907205588872272,0.10923033096237333,0.9486353841456775,
 0.5228422087379829,0.946813053593053,0.8482569508008693,0.7102297351737864,
 0.1632476442194758,0.6255064741357922,0.439750546619026,0.6091724333906542,
 0.6378058121823789,0.12658673715281643,0.22812047205598573,
 0.28279535604180306,0.13994867814368572,0.4746129079112601,
 0.08253498948261051,0.9425890751499478,0.6314108969866828,0.5613354323187759,
 0.4837904201260701,0.7158426208272375,0.8416539984436786,0.6248355910193203,
 0.7481287727751161,0.1824028316138241,0.11791337175316974,
 0.29594357352830136,0.3637175190481151,0.013385998241393637,
 0.9165227475938289,0.49450488273563575,0.8974941088177102,0.6823190620864509,
 0.5254657697853088,0.0026391966958972946,0.12341056855959343,
 0.7906938358179677,0.5227173277719034,0.9295418725209678,0.3558853230730574,
 0.7315098529387021,0.05739928714534093,0.46109207185757806,
 0.1486083585889999,0.826852729255948,0.8033637965162427,0.5942350104098495,
 0.2536463665700107,0.7066426287206969,0.8136009682415326,0.9726187161341969,
 0.06089765152547366,0.4127712565519196,0.989729595112491,0.25455119134342197,
 0.9800343566921261,0.5032280666737499,0.47442937812573116,
 0.36664526554638954,0.47212900974433036,0.8082518991099124,
 0.4578123861345177,0.8524897195298327,0.6251031458165233,0.647826871617051,
 0.9239558804968295,0.2923010912756031,0.6171065939295708,0.2010463516144365,
 0.8987871476438923,0.43201509549225325,0.7923138860560905,0.6216453219793825,
 0.15711714065887827,0.6462253505758542,0.01833296399034401,
 0.6589377678218402,0.212896371891112,0.8409407609340526,0.4496104747916262,
 0.449170886882456,0.10986149230330233,0.7076151613865304,0.8532192665958693,
 0.707497908994693,0.9226939914446827,0.8071206539176878,0.07752655709012668,
 0.9150593751411907,0.8479445927494794,0.19170979435666446,0.8461506302541205,
 0.6169849604965819,0.5040076066477386,0.8562409971082423,0.5714136121788842,
 0.9505727314295823,0.3197413887088676,0.02416958928538271,
 0.16398881463507808,0.050739916031389165,0.8777330410790711,
 0.15557786342146207,0.36925410059726405,0.7914762575162537,
 0.43154548342437593,0.32386386717459936,0.3799144692803794,
 0.7548812885525097,0.662693289599954,0.41404934050857134,0.4491535452144666,
 0.809807181248878,0.5056683201487709,0.34989848279516056,0.8628444524312392,
 0.26562117655321404,0.4427941524468364,0.5552083078388906,
 0.12390776952135574,0.685477199444769,0.37537320576464717,0.2984134363271922,
 0.5668680656379272,0.1683278471533619,0.6631752695966308,0.49217819179272493,
 0.599243586498129,0.03275885609484819,0.21103465665469257,
 0.13707935417922246,0.5191519204784947,0.42269252326058304,
 0.5316503400933534,0.16754924876648736,0.04809094537797931,
 0.5799032472236908,0.5458580966408769,0.7228792593759201,0.04136935212600157,
 0.36655914050665417,0.11163555646044565,0.700464413124209,
 0.19404829037248394,0.31540875469213603,0.25130809133298593,
 0.00957364548233297,0.9939727313993455,0.30846183171335795,
 0.7768696026523085,0.37592795772329235,0.992635621637971,0.24522909879141241,
 0.9272841782035901,0.02883756110229696,0.15746985743580133,
 0.8373551292188187,0.15358681613246072,0.3922049053697284,0.8266799358077803,
 0.8755760032823802,0.27232962149109174,0.27482539461840183,0.47089094849503,
 0.18814892248553705,0.6898088768236452,0.6602757540346166,0.7674516882360369,
 0.48800752999765573,0.41471438316027787,0.792125463760469,0.4092563743528701,
 0.4811334162638413,0.8797390796230493,0.8536366302630396,0.9071616862787986,
 0.5444189357982243,0.21297218076280877,0.9964601372608243,0.9769577466659902,
 0.8713159934630216,0.10859704084643407,0.39403704108709736,
 0.5733752364959556,0.19158934218036539,0.4923119539558637,0.7037318748326049,
 0.5724258910934384,0.225363507769214,0.30894600802391337,0.42796137172711335,
 0.6039974900793367,0.40842147989062927,0.1229925912629597,0.988820699048028,
 0.3723518362140459,0.6951477907150881,0.8989190237526665,0.6060898833212082,
 0.3615428875905242,0.3443412269352132,0.9892840531868395,
 7.350811403553648e-4,0.5938578964056577,0.37182881252234035,
 0.9784078342743538,0.9996600602200908,0.6200784995632348,0.9855365022668909,
 0.6711494329933672,0.35229605603569636,0.8435688966712779,0.4168327071295078,
 0.6899792516666268,0.9702888958827871,0.7865275379733431,0.34216967669609455,
 0.3030090731383941,0.7719035575574376,0.4593021131524023,0.7020513323335685,
 0.34070551376939895,0.5987826484075436,0.36915829658338684,
 0.36763114461946633,0.5052738513947228,0.5798739199534017,
 0.19904212786758302,0.33844023099182663,0.07546344088878532,
 0.05963913108523977,0.5035194233566336,0.25222881470034464,
 0.9422774858141953,0.33790822762335226,0.44681333724533223,
 0.09178959243691043,0.6633761109002853,0.9508295848585231,0.675178210955674,
 0.6885166053800398,0.7657515168265294,0.654952732302414,0.9027684072933839,
 0.11691549259463718,0.055261865051756454,0.05127595843813215,
 0.4108307159658472,0.4751174948757385,0.1043628350697069,0.17717156315065719,
 0.73465230694146,0.07237729380600377,0.48263850125333096,0.3510337759414046,
 0.09837226233230456,0.06476932323788298,0.7255217911226454,
 0.8627904733154393,0.29463897992557164,0.45158112513541604,
 0.9033095169974809,0.862518369727819,0.10996125102387677,0.5821378688313991,
 0.9158652874844952,0.7095542026202737,0.5079878586409468,0.03119502353460546,
 0.06944275554126467,0.36658695747130665,0.6538257551131732,
 0.46073779643903834,0.4059963112883195,0.4060618873177959,0.408280878505501,
 0.6026274449886667,0.7120035930983883,0.9210339261417434,0.664013036843683,
 0.3569329437046658,0.9015924015995397,0.3272205979978938,0.9139605452063368,
 0.3770173380919669,0.8290505354338275,0.42697733953587924,0.8419753375403339,
 0.7761580487446874,0.8007717489746805,0.9763734266360804,0.32445330842308007,
 0.5131686040278504,0.6631416381197425,0.8012540983466468,0.6712527499845546,
 0.5321698908011752,0.7113960646862351,0.27549828554054245,0.6774893337574386,
 0.8411476425617483,0.8343193444217435,0.31794791346440776,0.723613685290082,
 0.633273284399775,0.5874793705192075,0.9106227493806877,0.8364695021111026,
 0.34675137411360746,0.3441453200982034,0.6753799383802722,0.8891172878540401,
 0.051476525716872,0.1660172831204667,0.3011920599963165,0.0791300853526461,
 0.21080483170062791,0.3345265408319712,0.44643151208962084,0.833805701954068,
 0.3838508420768447,0.5030497696236556,0.9066622409382605,0.22381879936677118,
 0.2159134097684552,0.920632673673422,0.1705743463803403,0.8913154508785323,
 0.037908202613474606,0.02331942366000883,0.940116912852068,
 0.4517554924397824,0.688433762893101,0.80421392524881,0.0123216632902714,
 0.17659669251517374,0.5305252482546806,0.7178958668446216,
 0.054385331933931136,0.17833522020192027,0.7650364146729512,
 0.3894550251659924,0.6205892287311316,0.43903743554826347,
 0.07848267895759897,0.0893299384661107,0.8018423875821362,0.6567294210858031,
 0.8293630484254039,0.2708077147914183,0.8091372886857895,0.33955796982632447,
 0.06572144741421859,0.1441923732830226,0.25085297550415464,
 0.5380169932378049,0.09939595037989002,0.264204454850967,0.43435273012428643,
 0.502448965042809,0.6438160780794415,0.700765743073432,0.781091659518545,
 0.004292279075238881,0.5645606329093915,0.25013395356412615,
 0.7121654934183769,0.1652783100553985,0.6799676990460642,0.23837981923325657,
 0.7087184156556882,0.04650764423866627,0.4815026606536713,0.888899403325279,
 0.8849060506163114,0.6941269599351504,0.48014596261415343,0.6763186839767653,
 0.8789884684159709,0.24679392353497398,0.6210960373291022,
 0.045438619625843724,0.5469684467004194,0.7331963494927228,
 0.19739389871167856,0.05671386706069981,0.5483312091368473,
 0.47801847872170145,0.04970977215061456,0.6867779677674832,0.511532220989276,
 0.3132022327791044,0.5026686734976281,0.7180559668984209,0.502847504030071,
 0.8480082135141719,0.12106731087813205,0.8459999152738468,
 0.20317038467988802,0.7956579959052814,0.9541090191911218,0.401475758835833,
 0.0918727840676512,0.4318994178559725,0.5257165694640095,0.04528611128262927,
 0.15839213835294863,0.935458439012643,0.4436647692659309,0.690426443242141,
 0.24164929504960775,0.6690634911583258,0.6136439626101543,
 0.45476657391586506,0.6913567136338499,0.5104431830344388,
 0.24703773206178692,0.7968626856558321,0.5653921821938637,0.3175429943470859,
 0.21007861493100277,0.20031102571081993,0.2844891194242486,
 0.9097240368189935,0.6796524855153981,0.9118688007278891,0.26073470250955755,
 0.7928476528794433,0.6120696695516163,0.504711140297918,0.13878675465296464,
 0.4990062135068045,0.9759630162382192,0.16641837786105507,0.8253141867215854,
 0.9911616679348464,0.21273091450695503,0.8607852358319459,0.586452027322635,
 0.035156576880186585,0.795049232832379,0.23650071313051135,
 0.3485638325527063,0.47214974633181683,0.5691107961093134,0.5451367783644173,
 0.5005246193674102,0.3996331581935264,0.5384155933736782,0.38357965739335265,
 0.34296073409437966,0.72742929208336,0.3851059714520627,0.1643177629838518,
 0.7288107158934489,0.9766449804701555,0.42864871366265933,
 0.10961552438185795,0.08271098278231004,0.10718137581978093,
 0.5468451456770749,0.9245723513676029,0.28444285792226176,0.8838373365938431,
 0.1329882261408386,0.2928778589129267,0.7294227080994613,0.8073946129924642,
 0.6445165921976346,0.7003939836854121,0.34967523465415684,
 0.22803728879896534,0.6061193035224272,0.991442877172066,0.1716425458333335,
 0.06469514487842432,0.3290097556463343,0.5277954370116726,0.8042410988214134,
 0.4708722138271977,0.7901298515354683,0.30755388917251936,0.242920224976804,
 0.2583289041014327,0.8294952158932976,0.5084665468725648,0.03955417772805514,
 0.7930894851272468,0.5961448716605532,0.374524346214714,0.9706837556874506,
 0.24911465604167582,0.3393085100730845,0.09519312500326904,
 0.7180813270645545,0.9745441432059058,0.27901573111341593,0.7253709204875294,
 0.9698723460369313,0.13369775503503434,0.8012895493737628,
 0.43056881447459916,0.5742395290870662,0.2552572860014193,0.9359366738693925,
 0.6729977736600126,0.4266351316847291,0.5677289542183797,0.18502721677324296,
 0.5546417528735986,0.43218566851140894,0.8660823752638382,0.5973474796909019,
 0.0300319016445032,0.9935128967857113,0.452357397617259,0.628957231166443,
 0.8637049898722945,0.6130315154419431,0.8127053096239498,0.44875836176762374,
 0.28723148429328693,0.6823316634425834,0.6953963722653936,0.0710563471295308,
 0.7428262943336785,0.8450264665414797,0.7681575650822492,0.10557852518265398,
 0.43913113621164535,0.021684503940232358,0.8392203408140081,
 0.5933384013440651,0.33283610006188713,0.3907985193259792,0.3849471442921941,
 0.16824630908998295,0.016327613454188983,0.9452278196330037,
 0.9281000317078953,0.5515617871849314,0.8626227102913393,0.6617045049107537,
 0.8912949828545211,0.7336330220135959,0.7018742467337727,0.8644494736964319,
 0.5015881250965317,0.5706685110268332,0.06490760062905443,0.9557059146056949,
 0.9556920274665903,0.3605810062256045,0.07181669445155958,
 0.14021519128619875,0.9368614960116574,0.8484007490247792,0.7021369275265765,
 0.052817870600042394,0.15278447505359893,0.2515221544195513,
 0.6654012309510196,0.2729594652446753,0.9356564838984387,0.8251132738439246,
 0.1838335878994689,0.9734671105854092,0.8025384046774092,0.9788981928380934,
 0.7361986130949388,0.542158453678569,0.673425513277014,0.9298240780003831,
 0.8579019167915675,0.7883078333615957,0.3042432407158492,0.3500440583691812,
 0.2605584486401107,0.39042829994737194,0.754604445738845,0.9085360636946256,
 0.1758070944311506,0.3368699206836133,0.3136996984727891,0.21554210402550422,
 0.658262223106422,0.16407749487381695,0.3172740867675564,0.38864911186887374,
 0.6545741270591661,0.8851881797941941,0.1834715255665884,0.8782770885478878,
 0.8739291851023303,0.4939037049445849,0.7341447625071572,0.6138535818453574,
 0.05747804679911206,0.424069756010346,0.4415010965188617,0.9454828869356549,
 0.9652597351231054,0.7467632136984848,0.18925194212399865,0.9365021571638509,
 0.6758052729627175,0.8481741964978488,0.8850856975499201,0.6333937041309649,
 0.017698500149324725,0.10921791100987943,0.7806912844906949,
 0.7028017483783863,0.10640601268409489,0.5284463728684614,
 0.11759643978742484,0.5654063289562175,0.218290451890243,0.01940250665587906,
 0.8730968338031907,0.704703528691571,0.27826405685528244,0.46873348915182234,
 0.1932616729205272,0.49057526548298,0.8206439458192388,0.898838556536704,
 0.9891853894634532,0.7722081820772324,0.5914022047902696,0.07919383149201031,
 0.12835385176595393,0.7725258987556394,0.6276789881900163,0.9441137261128141,
 0.8996350206827917,0.9756031298814578,0.5008575730018627,0.9310089458522094,
 0.6174518966917073,0.45656155221729944,0.01943758937543849,
 0.8798704368291291,0.7483647575750161,0.06918552342370243,
 0.35514260321423397,0.9491803870488202,0.6722421950292956,0.3060819482796526,
 0.7171688967390798,0.6227902841914507,0.0908117896637578,0.8106798728259005,
 0.1506113824398121,0.07498096571841728,0.29377026707543763,
 0.35920688976057524,0.6480803138453757,0.06664181517613077,
 0.0863458620343176,0.92323571038506,0.2833070347362614,0.0776918403129655,
 0.48789882910207794,0.221679069751908,0.6719642210396063,0.06784062019932635,
 0.6382522493121103,0.26618972437112254,0.7631402584369125,0.5087958980683396,
 0.45573686088475207,0.32765382837577584,0.26995423969058885,
 0.35853438418741557,0.8700531097383151,0.13775963338089725,
 0.5767231208007482,0.10181157392424778,0.022671684822729077,
 0.05843589088070067,0.3843958795137038,0.06026521108692395,
 0.6584185290681893,0.39169308319848883,0.9684742970307896,
 0.021374584924115925,0.9840347364170849,0.2801068321244633,
 0.5026526271541644,0.6640596090689705,0.2935256729957967,0.24030551841116665,
 0.8158962285797255,0.2627969278754294,0.15289246332787365,0.9721527738671792,
 0.7110496709007617,0.8828146819646201,0.7126427173456837,0.4646677842184086,
 0.3824494633489137,0.6909094049521541,0.11692148197347296,
 0.08633712077043809,0.5483000608265635,0.005950808373652472,
 0.9063886893644227,0.7818491557453112,0.9865216310958898,0.38976952536369036,
 0.4514193979216664,0.4344743102786607,0.4119703455914263,0.734135117404868,
 0.2168646757153433,0.23426515428617367,0.8403348573424754,
 0.26534595439472364,0.7383167991721417,0.10164850272686654,
 0.12454084024115897,0.14322904367959832,0.9017798388191699,0.339421295048288,
 0.8699161619455152,0.282537086982971,0.5302924707780862,0.9419918240552568,
 0.898008218362542,0.5844769712511378,0.13346630173874074,0.23921728392933317,
 0.8587215654865819,0.7027842370656949,0.05082468080802649,0.6717657586198933,
 0.6307881994486737,0.8583090909620817,0.30875967526707415,
 0.03722323376899761,0.9399231385657729,0.6621660314880068,0.9216591521837288,
 0.04258468238592017,0.014101510854634514,0.6257669299120705,
 0.7720968002207265,0.7353509438112481,0.14313927633110035,
 0.08819529085288691,0.42456809707476173,0.22559256569023278,
 0.9400933348027192,0.24746783663033856,0.9092075793913645,0.6238771767025648,
 0.37499407452643074,0.48930756367777006,0.13309930702585404,
 0.7579588454602817,0.28739628860997835,0.39550715912255685,
 0.24262779543526025,0.04194375960864982,0.7627951329952071,
 0.5601397271174572,0.13229243979194905,0.7813703270033451,
 0.24559274804433506,0.7356521253506461,0.38853604827564603,
 0.1940218504686847,0.949603984402845,0.6910948184728001,0.4785993189475972,
 0.6318829195544813,0.6957580249458782,0.5182384675301255,0.20076026589432738,
 0.4950574612852824,0.009948129562612817,0.3652306363175668,
 0.0830638746700314,0.43362541665841114,0.05351292600462121,
 0.9500702521033697,0.07465246129927294,0.3084209168658325,0.210357286194491,
 0.791314218195737,0.19864085883828453,0.5466297511499039,0.13878827977245956,
 0.05767876331738242,0.2568854330302037,0.8268959395192518,0.8279533990055662,
 0.005749682099821962,0.007837834455667103,0.7220738544129849,
 0.6486018323828308,0.6602127022313933,0.1553424287966132,0.8091992045495217,
 0.8212470461880187,0.2582673186087625,0.7734432934594341,0.04001588819394897,
 0.4035089890304816,0.34353733801814523,0.8223776340193665,0.7318453927190902,
 0.266470646135768,0.9396724548432202,0.08417340871472001,0.4646043567275413,
 0.5869916771607331,0.4663470373205727,0.0717272823155275,0.462671293650889,
 0.23465168476348663,0.38754204899722977,0.8349252426373598,
 0.4685706991967322,0.19407816261516764,0.6587038733958547,
 0.46026447973625106,0.13496792878354402,0.9140042437937216,
 0.7403164661433979,0.013948462839752374,0.6179078818555351,
 0.8312406143707509,0.8350890986686675,0.672395447403084,0.8026187870469729,
 0.8605579205194696,0.9814964946434033,0.8211343621343505,0.6581885196808646,
 0.36794712395219487,0.7838054564441923,0.9150577830414678,
 0.38430581724879576,0.9659892854014701,0.31272374136374403,
 0.9785114359853168,0.17540810778667848,0.8933177016055421,0.3444378415155055,
 0.6605485988523916,0.32798038071386737,0.9698367464482888,0.1113345221972668,
 0.6774957129644766,0.3026591005081639,0.2952843408361273,0.9010887407256202,
 0.6074338269843154,0.6394513040046033,0.44245661972170636,0.4808232323355336,
 0.3484129581076332,0.213482654680487,0.7611743040533511,0.15963939005745764,
 0.947790568173057,0.4994580367770637,0.2835051181675391,0.13892711608410602,
 0.6422544180664282,0.741879344664653,0.45101993607809465,0.6235751062000923,
 0.8838870934134871,0.8347946811308679,0.2541857008620336,0.8559563088999728,
 0.34431459764767336,0.6134373995272299,0.6647840742286278,
 0.43180093474882186,0.34359477703662766,0.4151115794301252,
 0.8955002229308554,0.4430626605733813,0.7249672402110271,0.3217137686500898,
 0.8444061618338655,0.043172092313911614,0.05638516844586172,
 0.8870873917263716,0.8145499814529463,0.160419310140824,0.22335511839750222,
 0.6553232596086765,0.414005580327543,0.9717772721914889,0.4152619920594427,
 0.9862416368534925,0.24048342112443566,0.31587453588875947,
 0.5032030227924885,0.534979262896091,0.8218622389083481,0.5718209564746222,
 0.6606345714300008,0.7041413789560397,0.4719732077307398,0.2341607559190796,
 0.3055466985353984,0.47118541171738815,0.2980430932935547,0.8310048383128489,
 0.303420035111597,0.07507591207781905,0.5083826967104564,0.8547637873348928,
 0.2731102414429035,0.29642060674917126,0.9274695390802334,0.5908845175688422,
 0.5514377545870367,0.7217800811644717,0.8990494048199487,0.446971376173095,
 0.641572513792258,0.39627299026138074,0.9903191773005355,0.3298802808607544,
 0.21861781132417069,0.05436322083660927,0.4622245361383701,
 0.9497566168888412,0.7770102964776687,0.8344517947281055,0.21560726520108886,
 0.675017931496184,0.3979314693920005,0.41736808552738347,0.9997587706361882,
 0.268609519970217,0.23465036877330048,0.8832749501005204,0.08101075536339453,
 0.23436126833976711,0.20101547124793617,0.9466682221203315,
 0.8578115421157659,0.42614073119503615,0.7014975612388649,0.7095941018919738,
 0.770770567910423,0.9531651794740443,0.5165901526781104,0.1981462343684781,
 0.4612658331702453,0.8793712610248646,0.12451276991899674,0.3610508129070491,
 0.9155928470775826,0.5945065373917733,0.16196561110210628,
 0.39471535308006134,0.5074643792581994,0.31416866697477763,
 0.22807870739320812,0.13860313411291958,0.9193562540958644,0.409777597830428,
 0.019671085306393943,0.8543869246928175,0.7865631300053775,
 0.013420061546757789,0.6356154839028525,0.4361496595091754,
 0.5747279118379943,0.68774888482123,0.370893082338153,0.4542331863254483,
 0.7432976438832473,0.3709215618987499,0.6828313738765653,0.2277968041622156,
 0.15614460696475785,0.48618321540633036,0.6252025488087113,
 0.7916476011336646,0.6245704726507307,0.652597363249193,0.5791415449436321,
 0.9637332834642445,0.02430700694144705,0.7012295934432654,
 0.41889681084954067,0.7534561576630328,0.37525549684445847,
 0.5471774976578268,0.08750318804363921,0.5663130591670993,
 0.055644900327267544,0.9068936737557707,0.33752701816081965,
 0.36261482492392705,0.6682147668749712,0.9122206834268083,0.9053050776666223,
 0.9181611576972583,0.9172559173007571,0.23545064868320298,
 0.26449186718976136,0.3971660263583926,0.33332120065911,0.46696111715041444,
 0.5791837977383397,0.3461489392153452,0.15196248588742434,0.8332857363396537,
 0.7345925858065929,0.05897456698442771,0.7531911374082115,
 0.25105659614513265,0.2970947223420469,0.6943648526756615,0.4577853247505208,
 0.7574711024796538,0.5827190898060707,0.4930465396133141,0.9877638173534448,
 0.34887410662047047,0.5382449486636962,0.4965762768538099,0.4178442537451298,
 0.32657017938495847,0.023799341373384686,0.3657584393159484,
 0.721373332520387,0.6855090171364787,0.8920270656630269,0.8932887530280099,
 0.39877404657763504,0.3522084705747319,0.9501055815504508,0.7971051093091378,
 0.08908200447077042,0.4013581299352834,0.1924767577221781,
 0.44304294942935496,0.9380916455585301,0.9364534244954197,0.7658851074135811,
 0.9017628897612814,0.2909452542811939,0.30993673472695915,0.7009171821941438,
 0.9951384415857337,0.5698873849464189,0.1590570320882887,0.07638157996608486,
 0.6421766773389642,0.8541726402433827,0.6940096321522121,0.9314904213818385,
 0.25231561762849664,0.2294864592283513,0.6041608721645787,
 0.34380841169570575,0.23169323034327838,0.9231329416441371,
 0.3493442000355145,0.10876699293137138,0.6949742138517865,
 0.14573560157969445,0.6494067477148829,0.5751186343378536,
 0.28528874130494786,0.596757871242809,0.49331684475508597,0.9460813767434624,
 0.24665291535675615,0.8325561735952897,0.5740142495144563,0.7090680152476141,
 0.8203837629847013,0.5215368220766124,0.3128792009266912,0.8383845639013126,
 0.7142401295612845,0.9050972183102106,0.08440962255901541,0.7518669595919341,
 0.18084194763011752,0.0432563938361572,0.05458957761690397,
 0.9376683715132976,0.891483457780039,0.39970904482058844,0.5283167913428515,
 0.7358771219753173,0.09981834475637075,0.28475775588218266,0.175463390353237,
 0.024441690272751337,0.9368145511017029,0.023227946513215803,
 0.963657767630638,0.31507577340560977,0.30101642847191656,0.6913049114996713,
 0.3117965912797219,0.307231211442766,0.5024118969090997,0.25013481740556376,
 0.30436548393951335,0.7111340313588731,0.4734307656492135,
 0.011774536490123388,0.93025335219667,0.5045500762787403,0.547577080220669,
 0.33756811238811446,0.841994986506242,0.6688925510138122,0.8155210699982127,
 0.3226795235952995,0.7291978436397566,0.21631083003033502,0.6001675926040796,
 0.6219026987129572,0.2551210114930066,0.5149770633856968,0.8514111461529474,
 0.4662571020907855,0.6436174740585261,0.15846196360553122,0.8319028990609004,
 0.40688287979069937,0.8450017006453177,0.7497743132696695,0.3353957887623087,
 0.379605755568728,0.3283257361226246,0.7384863225732328,0.5924027840139067,
 0.975426189745807,0.8723199322179904,0.04990848274261539,0.23560873711310915,
 0.9197224276907076,0.5745807101621166,0.4353217240844244,0.10260309335453077,
 0.3200344628551658,0.9522595220382121,0.5823646507005201,0.795449294003331,
 0.4772981786033458,0.15163432208992023,0.16302991579751813,0.514622818358561,
 0.8181075628463248,0.9421009945036825,0.787011022660916,0.3260683429920781,
 0.8460561767955473,0.15725702938558872,0.6958447848672292,0.8196976801337862,
 0.3499569677628023,0.5464410299940454,0.2226601335372117,0.02049589827994225,
 0.7593824874119416,0.5338275273425925,0.1343123778869011,0.0460732530954272,
 0.4499955033322707,0.869568704736925,0.5023281384033342,0.45638200315229094,
 0.3677661164063083,0.5114311726962608,0.3140658009720929,0.25880009797814774,
 0.3543108669807341,0.27782438602535464,0.38625294474729355,
 0.3480150100018289,0.6931617097685615,0.3144223876634302,0.6674414781558372,
 0.39621763642499275,0.6949640088397939,0.7368619484770762,0.440091014748649,
 0.4181508705690218,0.5862387255709707,0.43491886359240084,
 0.04861115794903492,0.8126683112039683,0.9277509119244713,
 0.23179836386017738,0.4416242754857417,0.37560954625767373,
 0.04209198729353414,0.21510900123597498,0.05161560058459924,
 0.9632225211582338,0.8381575335690117,0.7027311465847993,0.7279061065920372,
 0.3302062595762514,0.14312537649544166,0.9358248356691519,0.7290852308293914,
 0.9689213700561018,0.5704372164084379,0.14749615644609226,0.3121001508214858,
 0.9754540665462927,0.2351338972414455,0.6832756727608893,0.8584577530745241,
 0.5558116102347417,0.27651805364540494,0.44213521588420357,
 0.7939625342951422,0.6407719475938612,0.03998125044726608,
 0.38891730234339283,0.31566770184176496,0.7343647991868294,
 0.5483895405300802,0.6192305942860646,0.9610939267139994,0.20524778754264295,
 0.181314378070905,0.6334067082446752,0.3606113149154655,0.8020969234821148,
 0.6987087373387362,0.31340120025882867,0.1031292533041207,0.9982230733878541,
 0.28879386446882327,0.9271204604348984,0.24724730209438028,
 0.6974546197120384,0.5113425072694467,0.6416131794327286,0.996284415243238,
 0.9495344345853036,0.4560150200055574,0.13117796564518547,
 0.033547438629330184,0.43758923076788514,0.8247939911709333,
 0.9639177890879407,0.6172176308002051,0.7481525768345046,0.24410487906077782,
 0.9113552070579374,0.9095875495147153,0.6293551583845692,0.1078533982418668,
 0.74531656610805,0.09056876771233047,0.3286444677947702,0.8941247030353,
 0.027499136088600107,0.3207809398396273,0.6082296112290587,
 0.4365356857630136,0.6177889142167899,0.8312766273731595,0.4241492954067638,
 0.1452558829299775,0.2979761537147223,0.5047270606219287,0.8894818494575378,
 0.992951246346564,0.8088221087619709,0.42676807938020034,0.5613192177571737,
 0.22734293989901233,0.6142401037580607,0.5001515786196543,0.7367037780959693,
 0.6887662640676723,0.8461500462730145,0.23983612190413617,0.9768172591804156,
 0.7686033201140738,0.4353581746714632,0.20229092749470368,0.7850204611976593,
 0.27823705640988794,0.9995611741989353,0.8583011833396356,0.4296813958622764,
 0.6374130962400001,0.3406854243705072,0.6818178578314503,0.03853987986557317,
 0.919619308284446,0.02161952176306703,0.4983873495916318,0.3875245143557846,
 0.5784813432367307,0.7504864789288711,0.9767541006917266,0.9866012402210457,
 0.3701709215914897,0.29557558667380046,0.07897203497750593,
 0.39195276266920587,0.6800711094945338,0.918485382303821,0.33944216498211155,
 0.4497060112427024,0.4123852470454915,0.11341928516613742,0.6716708429201093,
 0.535018738330419,0.6373623332017572,0.8789277097635293,0.47300691744752943,
 0.0242024670229275,0.3225259317036504,0.29688466823427373,
 0.11076826241317905,0.19987709385418506,0.949016698484996,
 0.17251623165268948,0.48647622298672033,0.19483683656433481,
 0.7357369471944455,0.7796561602663492,0.5755369204447613,0.3194172757964593,
 0.7399088087277121,0.3527106525720136,0.007444901081017474,
 0.9349836518065762,0.9561809591370374,0.9616122597929377,0.849679113590957,
 0.7944525953441878,0.7277784549280539,0.9933285473139177,0.10492670624262035,
 0.6007729090526869,0.9927132343565874,0.7679868522074693,0.23890920438682917,
 0.6698490246807991,0.07544665216343138,0.7527065224825661,0.9650348226632638,
 0.6271779214111123,0.4528537851686535,0.4640801782459649,0.31057246941247363,
 0.6205131422784382,0.674648320129004,0.02019664240385044,0.21270651061366797,
 0.4527815721099142,0.7598208735761699,0.271409528429967,0.7994276305202878,
 0.8287381992281692,0.7055275777599398,0.8855370708489614,0.8449375547905431,
 0.8845203409944252,0.553405408073811,0.12741993724942768,0.25354043960383454,
 0.10954219829673117,0.32838133672909153,0.5247753724406896,
 0.5504068603791151,0.6854596656862646,0.16175015357206846,
 0.26583937594675355,0.9478914140166207,0.7617493852504122,
 0.40831159971187636,0.6057666033690724,0.6302317107046196,0.9791602421855694,
 0.46776733053770003,0.3570658293662283,0.8688636425994622,
 0.018840993676480622,0.9602113505291876,0.631465227586142,
 0.22038397528817522,0.17326475325029078,0.5708018426202179,
 0.42830636929732024,0.7766994021410326,0.021697949546109707,
 0.6031336205251896,0.04220477093654251,0.9825774605097868,
 0.19065726522156057,0.02166214795451049,0.6814163129236,0.2779402081664587,
 0.39405207781340124,0.21455954748570483,0.6621067419058837,
 0.5227955639085059,0.05277833729468029,0.7921724984300446,
 0.25893801895053603,0.4492629953064602,0.08073578595882869,
 0.6892011312016818,0.09237819207206066,0.37699527457869864,0.492811725753299,
 0.453468415471709,0.1784044580927857,0.3573365584737779,0.13806820619928217,
 0.26571915733551477,0.3391606063885767,0.3822279907009085,
 0.10105842227333861,0.4566361890377837,0.1764838480963422,0.2132142594207902,
 0.9856885951013151,0.29550249877379986,0.011999745295567799,
 0.13600831719177775,0.22024606088684007,0.38179348990254747,
 0.5795251542035481,0.9378979426972184,0.7117404576782975,0.33675990414552004,
 0.7926878572860103,0.2725670062550859,0.709685275782297,0.5472463114101322,
 0.20936301172471583,0.20466035104490066,0.21503831293097908,
 0.8846609515451265,0.2582644943581507,0.32202456806352375,0.9271470376000327,
 0.13171931325889052,0.3077127547573043,0.1967448481343168,0.5725011501044206,
 0.6953598099846224,0.34261516548047055,0.656859538672832,0.9747611409422862,
 0.6469853870758486,0.25209612699455775,0.9559354924697152,0.986367132069146,
 0.5059169094044277,0.8424186563931224,0.3502596980406163,0.9771904139031433,
 0.9334076171343971,0.3020044892378255,0.8893944633930206,0.6147800164599824,
 0.24727324760373404,0.49616508317739605,0.7398088118342261,
 0.8532314717657679,0.7221923933304296,0.8177271947030716,0.9726393476055866,
 0.5991563956541069,0.4254636774393974,0.8420067109058929,0.03143693551379123,
 0.8679608456908221,0.9993694244403633,0.8847698522579812,0.3897896331787404,
 0.027286430606931367,0.23137797169341057,0.8199001508677296,
 0.4797885637968682,0.4549933899814378,0.20771151437973778,0.1891717563411175,
 0.3095050269158295,0.29234975012870057,0.8375550719634683,0.3339963985534895,
 0.5511017560058278,0.593119861367228,0.653705495156016,0.7516707520519431,
 0.41899273844207574,0.8502989425234557,0.433400508156222,0.7731989434254877,
 0.3215453382232243,0.13230647286888997,0.685554545171472,0.8940720322159416,
 0.2819475499512759,0.5875341481205777,0.5904356418355176,0.01984217481821149,
 0.2793401325434167,0.2942794925631126,0.6654574239237665,0.3620333046096714,
 0.43325842330935915,0.4805469554970966,0.7712734018235563,0.4492556482722072,
 0.3973859205717344,0.6613862261242613,0.15492264019483803,
 0.09607646305168704,0.5290046853976429,0.9749828829871269,0.8913313595489574,
 0.02101141013392016,0.08178832286997517,0.03264864457395844,
 0.07454175594146528,0.5745458325148625,0.286832987163529,0.9628438355776942,
 0.850514875855268,0.8983741125102694,0.3691170211091551,0.9559272651566637,
 0.1412593486328263,0.899533554742015,0.32440804321311023,0.39368708943391606,
 0.8817408312527887,0.1399150616567033,0.7909513057049682,0.5268971732912966,
 0.4891705462589725,0.6385713064223978,0.11472321706351563,0.5174201757367282,
 0.8944222158448228,0.5825132459509962,0.05751253497606057,0.9136367789390394,
 0.859764886465477,0.37621974912010403,0.5950934052676415,
 0.030786536431862577,0.8135173237565463,0.008683761942904655,
 0.7315229370738727,0.44761419750036713,0.16550475981555335,
 0.9059012622528854,0.9360194707628973,0.9725514486536604,0.1262186110928143,
 0.3405766005075147,0.5585797929259768,0.26043113418864106,
 0.20964079748587494,0.015914587548171877,0.304935246200179,
 0.45946810073952427,0.4098091678505871,0.4636074986464679,0.6383503311928108,
 0.7258478881951308,0.40117258453673954,0.7624046249066385,0.7658560698370146,
 0.6376337750992386,0.5853305443665469,0.43899998944992347,0.5675002318714373,
 0.6931136689194239,0.6106915073574738,0.780858659508397,0.9095861625928735,
 0.3976554355781119,0.8090717831910094,0.8008687751356498,0.37124896323315326,
 0.5555452386901035,0.8149512533781391,0.2891772899366556,0.3050135919827599,
 0.8680321398543452,0.57843794361034,0.17467631402806316,0.17592582745840202,
 0.00459716430402346,0.09048660895813154,0.22732931572063553,
 0.2889640514930202,0.6019129639289241,0.8863669656262936,0.9571268060686366,
 0.6628928911889203,0.17464722550856993,0.6888883467282954,0.5596965590758397,
 0.20519570269934329,0.48679848488151134,0.9621125458674165,
 0.2627032846289462,0.7698153715924532,0.5501221926840493,0.4539870197641702,
 0.5277427236361707,0.3984293067706821,0.7799978101607502,0.3944736276721885,
 0.4548606620946669,0.15547084610576245,0.02684912456141697,
 0.3253268771831853,0.43262404481585737,0.10108224127606269,
 0.7356491046150019,0.24739465814189365,0.9359514035454269,0.9318817334786426,
 0.41601256113436047,0.45022754213216565,0.21472731015501645,
 0.8036759851307429,0.0822878583223765,0.6561893095029607,0.3582328969768528,
 0.586467986305116,0.6808443653110399,0.5841087673534621,0.24780675717636658,
 0.8254512860474282,0.012321055396092517,0.8761715281015874,
 0.6124735275513887,0.8364800571963337,0.8880847528020074,0.31598725971694097,
 0.8973963219457438,0.26707957542303373,0.6709943160884444,0.9272028392857373,
 0.4818991035765813,0.1513987572216563,0.8202455219429028,0.7839970676275869,
 0.21538872224141403,0.8286255504854683,0.126443218610654,0.8305833953361389,
 0.292331670035362,0.6225358064343299,0.01532333107857875,0.03744288283852526,
 0.6038868329703215,0.40809236664998527,0.9903698258167312,0.5071319179340044,
 0.9975287313536509,0.23988493358362328,0.04759682972690993,
 0.018759274865325026,0.3629493230489623,0.653893006209322,0.5271550245335401,
 0.8166649861657433,0.17719249169018214,0.47498998269706505,
 0.0234445705377051,0.9786847851177841,0.776570183813157,0.4591878215022831,
 0.6249010726680891,0.8806523285775504,0.3004817187783657,0.8758655283346329,
 0.1844756971816539,0.33594382048661875,0.9355205656620226,0.9076551167693975,
 0.45739498311848936,0.6714769711236741,0.6080249995611647,0.5937488491011844,
 0.9914686482822108,0.7360909387655534,0.41401436730468344,0.6458564052920643,
 0.9068408343921069,0.9858275957408976,0.06693035555142124,0.6916350229745484,
 0.7915565537285796,0.24367787449606193,0.24190234327764693,
 0.1731032020483637,0.41808765264042747,0.6411224617497002,
 0.10884995849670343,0.7038181989503108,0.9839459090065749,
 0.16270526754861026,0.7784501314994986,0.9205447413477603,0.7320463510063818,
 0.4347644417551457,0.8564094701966514,0.39275191434126544,0.858203703190993,
 0.6221450851092185,0.6248905739400324,0.608654201284351,0.245981117543285,
 0.6644091690145715,0.962632870265711,0.904159431299748,0.9967078454985197,
 0.17950479032733757,0.11518603117119841,0.4434457427355354,
 0.8822252577150916,0.21245880101201164,0.9224084159957698,0.6449533041047808,
 0.09853450457844848,0.5221029294815143,0.48041671138021536,
 0.8812483239252873,0.9445394411488571,0.3226684380509728,0.7336651888124743,
 0.9276260872060764,0.6734386599319461,0.49845492162238647,
 0.06308944642663672,0.7522997073515502,0.6089730667866108,0.744133153372,
 0.3859372571647067,0.9359766332226118,0.22710934537518956,
 0.28104587374296386,0.7334400192554443,0.12722038340590025,
 0.17148809134323761,0.34119237777611333,0.8577520246236305,
 0.6547369406810433,0.6578593227230023,0.9575578762335951,0.5054292721108773,
 0.9878225982754474,0.12481956511645409,0.8743257688959469,0.9113263378526478,
 0.15169703377124721,0.9174391558155011,0.37125425949810087,0.518751399900206,
 0.45240195562113406,0.47055578388489905,0.21041303189873917,
 0.8018067194909172,0.7678862786697722,0.6897614863138439,0.5252811656265919,
 0.7619793131611775,0.27471359527038397,0.738821358054437,0.2200948153840513,
 0.34432777457778907,0.38360957015234876,0.034527045546050905,
 0.2016787636755506,0.2573167658352973,0.9946256315247974,0.6564955066146134,
 0.5184559848325876,0.7212210133010833,0.8062819361795454,0.3637148673712971,
 0.9139949621999536,0.2870913104676408,0.10994287108949097,0.6731778543988298,
 0.14221013200507782,0.4720794019389485,0.3110799115422861,0.7415666991523011,
 0.8609761305293345,0.44312322864827536,0.7697775081926671,0.5926919245169119,
 0.30374210594461015,0.06821221906015451,0.9778393179194977,
 0.4240055499339678,0.5531402184390309,0.45027698470859523,
 0.36997381798792206,0.2775114796505199,0.49522543028935706,
 0.8160183305514956,0.8142358152845782,0.6918180904528621,0.5222099631401953,
 0.09468817595362966,0.634640974221066,0.012918181424937836,
 0.5349486897065973,0.93225042599118,0.6810062315028698,0.09647523733317809,
 0.3305632289344056,0.8293089404521363,0.0599612065116123,0.19820565468021867,
 0.3160846220566421,0.43734244112589227,0.4326534352043475,0.7400496529301692,
 0.28277935151331823,0.8028936001452761,0.09721192525024014,
 0.2514321473236971,0.9851781596021779,0.954443686401118,0.991478986232263,
 0.5387224470729842,0.788290865101451,0.9308294791460406,0.735915040275997,
 0.05397825393700206,0.9035624705205504,0.16572859394225103,
 0.7584344381621003,0.17694027807878399,0.7629020830718178,
 0.21512361636631805,0.5955804003768473,0.8624653178485668,0.973961217757788,
 0.39735429075746664,0.216263387554805,0.8536033673097349,0.12357642399725943,
 0.516506211603776,0.032846041580956054,0.5069107953946099,
 0.01339675857317435,0.8591431220288966,0.16171466820934588,
 0.42532813490661225,0.5787918884910492,0.46554040037930466,
 0.7689156241712438,0.30397935696303213,0.9345486391765505,
 0.35809153472526023,0.0026204285009134853,0.37588536479859747,
 0.8421680007316734,0.0705488936879517,0.854752365336424,0.2297735508885077,
 0.788418481951678,0.6643740860238299,0.20962029886369604,0.6773427176914496,
 0.6380958526941137,0.7102580555067366,0.05040634192617588,0.8697512956335245,
 0.7093554294700477,0.3338335465891822,0.9651119477266729,
 0.035093140312280835,0.543766948946221,0.5501773173281582,0.5472467473254091,
 0.890031355409616,0.2891268134881311,0.3257422059370837,0.44382077476296133,
 0.05864403848144062,0.8117764961714089,0.2763205663360613,0.2315599017820147,
 0.015984342381462735,0.626763715342725,0.26591423405893166,
 0.7195113393824637,0.6280315676737188,0.4074151836609686,0.9068334287994239,
 0.3596287460391494,0.4690579230403471,0.1950340469556383,0.6533111837562666,
 0.1421683611583925,0.4617127819987036,0.47407046328381053,
 0.24580215844115727,0.3842952447426705,0.464756728552696,0.5683759472030845,
 0.8553812873642569,0.7666840403869053,0.8960432933901393,0.34224521510715344,
 0.6293329389498784,0.5197436954817888,0.720952977027142,0.5033275917166625,
 0.47979837099894473,0.00930110513360205,0.13811101855264962,
 0.2179177806392827,0.5514715836196209,0.4559193905215537,0.7714945733521728,
 0.06344239619759395,0.34908582696269275,0.4679691211082644,
 0.7725294568036021,0.24021357841165158,0.682098138774776,0.5102896715617181,
 0.38473616679346434,0.2082001178007563,0.4491388937403482,
 0.15913323595836748,0.6366833921871038,0.4925469602381922,0.4699986473211869,
 0.7134074216037104,0.33778805019136693,0.7124790847488183,
 0.29688387450816567,0.042473017664982016,0.08411348733650714,
 0.6656541082557652,0.10327041169468543,0.16042433495005726,
 0.10977429659841254,0.8376990635189019,0.2655581793385705,0.8282013463957714,
 0.5098433037238953,0.8137602172373034,0.3454291818987759,0.2601904678033724,
 0.7090051268815893,0.3924246794592646,0.3892455720086462,0.4649910753373837,
 0.616661559363695,0.9758193634674899,0.1939145039916903,0.8238988286498856,
 0.13616561557129403,0.32881368341117223,0.8066732977167064,
 0.22308194991987285,0.1307066199925957,0.4670162703919738,0.9399327389530049,
 0.2422810100434788,0.20472029261232894,0.26748567226634956,
 0.15353075500939684,0.6331226367385757,0.005060053256764507,
 0.22920106393879947,0.37735889747844875,0.4614923438404571,
 0.6374100475728994,0.7454468916384491,0.29808119610562944,0.3751551421162985,
 0.9166320833821556,0.5562508140842299,0.7833081864539869,0.7159254806995141,
 0.2553273569516461,0.3961390805196172,0.4109983233288701,0.737668056863283,
 0.9020373054785704,0.7476143089969296,0.029359545976830104,
 0.02752394995058305,0.7781915922752545,0.02611585499519331,0.502038807696394,
 0.6620215179504149,0.620141843570192,0.2997506042898498,0.16413854197096178,
 0.7440261457420945,0.8488043978577678,0.5003840632882671,0.08349496785830779,
 0.7934918570727266,0.37368976418813005,0.4754056945848496,0.3461227713957601,
 0.6037466232450461,0.20344168307526045,0.39037549990324927,
 0.010259035859977006,0.10913019151939696,0.416407836544471,
 0.9943905266636713,0.29265640052648767,0.04364577697934857,
 0.06769700423334424,0.6443879449122115,0.4759962612499261,
 0.10848008271389331,0.3757797276201328,0.7497808841809941,0.2429511390093103,
 0.388713810036713,0.6294858392937752,0.6862153325031263,0.8247524864117449,
 0.023819252458407658,0.39375059875324014,0.9616131646589426,
 0.7205938758676484,0.8825783690318196,0.7140919587765192,0.8273406482798614,
 0.48149044827614074,0.7468468008007816,0.21461482336135862,
 0.4982550122549205,0.5425197235230804,0.03025575391091584,
 0.061011264899510365,0.08440299787365979,0.41205363434687636,
 0.5103141276473508,0.6238072190550217,0.022549714967830092,0.685069920753917,
 0.20339342072630517,0.6472782478214331,0.18486997961010643,
 0.8405180200295989,0.09390533628553799,0.09250363746882995,
 0.6496120513783524,0.8538112921261182,0.9083303556110447,0.11952564207415017,
 0.5750930405316512,0.6371370867924688,0.8849076043834944,0.5252164763180893,
 0.459305168194714,0.7507108300393718,0.16701182759563327,0.5813308996333639,
 0.9982667400293832,0.1968903454156159,0.38759325529681865,
 0.35519547835986187,0.6891821554513403,0.27325605899551786,0.330450664543668,
 0.2715733216030318,0.5874521571558154,0.6187994140386466,0.2953903904717201,
 0.5145698879104491,0.9487935911783854,0.015315604577608943,
 0.15071721608046296,0.388490152317403,0.9488755599438681,0.7546256672592887,
 0.9438448156205262,0.5399625957974038,0.9910764136544352,0.25420941065930247,
 0.5481188044876979,0.6942170448325545,0.16848259537121613,0.3642963415018057,
 0.7844134202559968,0.34295513397858246,0.1889045754651817,
 0.48107599268846624,0.606563636750666,0.5609233264370438,0.21599791372276167,
 0.42146982958010915,0.034845823224934125,0.622000372307794,
 0.5130833194798741,0.07111335587319745,0.04164648389789527,
 0.18294699896701005,0.6959922648206293,0.38402874817030597,
 0.17281590741325004,0.21018016304756426,0.927248392742544,0.6905349341071492,
 0.9312093718225383,0.03628312408078216,0.3101230781558061,0.6467002212430423,
 0.9647582331628999,0.32158842538836374,0.6585939620440628,0.5655056365246595,
 0.5087761831702948,0.5104085797670634,0.42164289101893504,
 0.22488204836656367,0.4647928894918004,0.14870328649270304,
 0.8909355373349241,0.048841803815771745,0.44945445229754744,
 0.2707069326035536,0.09909745940927672,0.8706960252630648,0.2955821240578036,
 0.9140613414197208,0.844019406100513,0.04288706957613586,0.2609505217932092,
 0.319147910252202,0.22457526148397644,0.39103917000021116,0.7572754809961834,
 0.027946554025785053,0.14562908988647127,0.09132050088964039,
 0.7100473000087572,0.8690221960179456,0.5503247368165336,0.5672417098741733,
 0.4355835207835117,0.5065960570205876,0.35578954434108034,0.5806426697490874,
 0.7860146298367338,0.6976331927889367,0.5707142592778217,0.6338609244751698,
 0.2702335526870967,0.09961155022748736,0.8367333742573068,0.7782957022217172,
 0.4275119201297505,0.18185689941013572,0.7452448398169111,
 0.048058582500548175,0.9128428466177894,0.8053965493544453,
 0.6480618071593526,0.18788573864817038,0.39034139667309886,
 0.04470070969104867,0.752204138687913,0.07897893758194074,
 0.28371347154217363,0.7448816014560589,0.725698666495048,0.4062730105933946,
 0.43821971324526043,0.8083711567775291,0.8607617578933756,
 0.08970423690196161,0.8894053037271655,0.5749694178075071,0.7159968429423333,
 0.25553303946108463,0.669889096314209,0.034397948589674265,
 0.4090318483256524,0.027709705393179296,0.4298066270137878,
 0.1443122419011159,0.03445664217746969,0.4349365463907551,0.3624844934563326,
 0.4903964670785109,0.8903061136968545,0.7874070139706694,
 0.022797584685259054,0.8840014364868954,0.4908063180008888,
 0.4450316686494784,0.9431734563179007,0.9545202589110142,0.2308538452180614,
 0.9022681853085396,0.26476889834852213,0.4333767501651634,
 0.19168752114872012,0.3238954850505723,0.3166002890551304,0.7252367786981966,
 0.5910838540129133,0.6538750588092027,0.33085408216452505,0.3579574024322435,
 0.39237083872240563,0.39204230553697705,0.1614897188726786,
 0.2405379828828581,0.3947665674833415,0.2328426788166591,0.7837315582991551,
 0.3818558264505424,0.005676974299968052,0.8342625048867535,
 0.3649284845447709,0.765201408820948,0.8223131041348866,0.4972935561200512,
 0.2699810803037195,0.8356636063944758,0.19353982186348684,0.6825081211291655,
 0.6734862688029444,0.5099047472162801,0.8296615556749759,0.6463885778069568,
 0.130345211000801,0.632595616417049,0.5026768930790885,0.9232418425752306,
 0.28123846723689505,0.5215727352521736,0.21684500243143334,
 0.22316543311796122,0.4885540164004831,0.8287306399055037,0.8712184932649446,
 0.42255544856813276,0.42109698832948195,0.602976760781778,0.9721835077869033,
 0.7561523248093716,0.9340660238231371,0.4615025263173764,0.7194666984221102,
 0.33148640045675426,0.5334298801493438,0.5277809402484441,0.8736542861461379,
 0.7319542135474978,0.8889653398703811,0.5572720587520187,0.27205519811547296,
 0.8151039142483608,0.9517932464138423,0.5423954493362256,0.22411924824614493,
 0.5852657334291876,0.7250516221780644,0.3946536049531115,0.7709209359342218,
 0.13066404075332527,0.0745514742012493,0.40852649787037043,
 0.25105487656569037,0.29738419424240736,0.5178576136367767,
 0.36466299554171466,0.2872196289157196,0.7440448081417155,0.8048344702007801,
 0.9557964963358151,0.7673252298561054,0.8456719924275775,0.07436648787933564,
 0.2382030822195167,0.29149040175249286,0.12592713665552502,
 0.2905019878302906,0.08690565054060195,0.19905803974418168,
 0.48380074571781273,0.8860668760022955,0.963301365936325,0.15996080155281023,
 0.8644682387608592,0.508983101593822,0.40262797343660794,0.40815718266969725,
 0.4533247688856381,0.09764455824833307,0.06154256021659199,
 0.26393495842904036,0.07993991974348802,0.7387615015574153,
 0.9031328592009428,0.5542563252784876,0.2623954987061178,0.9361516930927821,
 0.05204042394448638,0.14908947089110058,0.47770189688203124,
 0.29954489569622167,0.6965195065523653,0.7543435488574984,0.8405264872371838,
 0.5776465858317886,0.40419031333859334,0.9185912402419439,0.9506773078724182,
 0.03891072128309303,0.7001173408696691,0.12508049292283163,0.548398648114874,
 0.1743136573450994,0.8522421214307105,0.21962250830524943,0.6150846252898091,
 0.757542143917342,0.7173106115304969,0.4083107888308817,0.05822359404906363,
 0.9375479551136109,0.6794152871635071,0.26871799324178847,
 0.41425949662620953,0.9385963809240658,0.6076980749111047,
 0.08184785869594435,0.9100175996184257,0.7161404963316687,0.4515533546285959,
 0.2896574829880184,0.887741194031022,0.9251640041167012,0.2035705631187148,
 0.9091266146227011,0.9612460939795741,0.2392450946420671,0.7486636559537724,
 0.9982214069226863,0.04288203246033284,0.10063122435076388,
 0.8424271804192215,0.6211569676115815,0.0929992628444577,0.3911607835618235,
 0.01317969398884189,0.4958508659929266,0.34841259742753206,
 0.4977999683002843,0.855566150055162,0.7437361755109996,0.8120302585656645,
 0.3473780018185745,0.6059403052735615,0.3777682118170077,0.7060410366936632,
 0.9115621213943549,0.9457168336600289,0.8130670072023836,0.8611775101466782,
 0.8877330118803395,0.6843014640034276,0.3358735390177293,0.6222749248520074,
 0.7508823570577157,0.17673557609536883,0.0744411617072056,0.5851087691298169,
 0.34645494556536216,0.779741481709722,0.8829938335150345,0.9228543550502554,
 0.7430130591275281,0.8479798516643603,0.45641544556503333,0.8147091155533078,
 0.3224436554229878,0.9858042609520216,0.2369304677003956,0.13634788987432123,
 0.4325845398029601,0.28750147654461033,0.08077945290254629,
 0.4519455802014758,0.9064578417438343,0.15133940814593472,0.7592204714549322,
 0.8526202209271894,0.09195578328816756,0.9850166678385482,0.5639642847946765,
 0.35245121916885047,0.6769455101386689,0.8162641489018514,0.685366929254173,
 0.6755566882865585,0.8380126537379126,0.026588017382552454,
 0.7886956112369794,0.7866818374142353,0.2507257682913303,0.7703552719797758,
 0.7914755386783967,0.9780335622930328,0.20147435928473434,0.5865501983435603,
 0.7391982348650068,0.14048469007348818,0.5605618015025177,
 0.48389374276303887,0.9809419587457133,0.9708568563058174,0.8834748660817549,
 0.7298873987234259,0.4445071502709661,0.8647418075152439,0.9858368649070121,
 0.21515427065462833,0.7533590440085677,0.5224810720309643,0.5833995332163477,
 0.5985363897582128,0.45099936955098596,0.45688060660318386,
 0.2804519722849016,0.2797840100477811,0.6491210913606725,0.2957234177146082,
 0.14401462633946982,0.3105540181216079,0.36485852631505533,0.333748906499567,
 0.6227682431085,0.9133133532567308,0.4619549456698958,0.8209060067606833,
 0.7401606196007169,0.8159118044193121,0.07234559826069664,0.6738108109379,
 0.5383266376960096,0.3576038044711095,0.768378992256497,0.22086597629381988,
 0.8676412521679637,0.27327765120011593,0.36365787785512305,
 0.3575997601555867,0.2631489365691704,0.46605749022389575,0.9598109499835777,
 0.5518991164603239,0.7363814568771598,0.8737415702090262,0.03465260621914268,
 0.18870636036206256,0.8249124877531735,0.9265107048250533,0.5724610915647884,
 0.26516827414452293,0.9651908442136619,0.1468957705305899,
 0.11124364987648261,0.013352785761443497,0.3036377150545665,
 0.15301396178561733,0.4350341133260429,0.16266501433761205,
 0.1383399863187954,0.13744980945523266,0.948152459752568,0.3300848319310069,
 0.34448267233289176,0.42262768242323334,0.37917684498966264,
 0.04048135654809648,0.9469177667504776,0.6280308981240514,0.6437557623879198,
 0.2748064465940916,0.4612148354881762,0.9247405023788938,0.15257376488480778,
 0.2705229048114274,0.18693728221329198,0.637608627026885,0.4162385021184648,
 0.25469309324477973,0.10753048378007801,0.6994156778519229,
 0.2441886781687217,0.8964877832494353,0.2908988053024586,0.4988803448840753,
 0.9195223832548418,0.2586678852272053,0.39197043023035616,0.3361544655062275,
 0.8015268487894663,0.6671465524602633,0.890345995544756,0.6783253025140643,
 0.38076831952610213,0.8409691045978146,0.28171270767000767,
 0.05980960986898343,0.6086924968930673,0.5955361432365963,0.5125233066424273,
 0.5611415826310002,0.28435257814645487,0.7354782277453424,
 0.19681001107592233,0.8979146303733041,0.2525324884206728,0.715883163103333,
 0.5223692810987962,0.04725611392283058,0.5217048426891594,0.5240711982780001,
 0.8763767411443794,0.8260505151420914,0.627392362455973,0.6991521335866838,
 0.8166947187477721,0.3396964607815136,0.39810924620535215,
 0.34970287881189044,0.785199721347224,0.698100374808342,0.314008903674475,
 0.20314033594885117,0.646229011051738,0.6494613543456339,0.5635220066518635,
 0.7373387806042276,0.9688754318300514,0.4553244795032426,0.7824725152217515,
 0.8164984322657473,0.6873971800209686,0.04346652850221744,0.9468000813209955,
 0.6910161816135667,0.3683998895315932,0.49982869959204723,0.4735896171585132,
 0.7387969980246187,0.737464831516415,0.2239350804938911,0.3893122579132041,
 0.09503510320803743,0.5690750965056574,0.12486267932436235,
 0.9865011587258385,0.1800851097004298,0.7581051569344778,0.3194329260493324,
 0.714923559061701,0.8720168794635115,0.6432635405961481,0.944697424167152,
 0.848607194000115,0.7022486078291351,0.07370664478063549,0.25563103681805943,
 0.037298375442535225,0.4976652967559708,0.031265612401677245,
 0.1992967822063002,0.8778972097212099,0.4733263556759787,0.15622386033160307,
 0.6172302951252641,0.6646530343520525,0.7925556009886,0.6801942705268998,
 0.2233219135959591,0.7899988784848954,0.5344444098746672,0.8169895735165761,
 0.08643339421195595,0.6994953181510964,0.5819121164001515,0.9446710174797407,
 0.07998999457035927,0.8431455437393127,0.671483033931013,0.6857001776243755,
 0.31419964172238357,0.5064635643591651,0.2771692521621588,0.9512583550676643,
 0.2138547418255745,0.5189814876310075,0.24741999068016962,0.8311694832033119,
 0.6350079536622414,0.36383144330986117,0.09310023779243148,
 0.1474369023265557,0.05898023260354812,0.23906712476968106,
 0.12286761499722598,0.9663015271681469,0.023861536017399154,
 0.934650118774354,0.8254165910170244,0.6415495017183956,0.1093850168122189,
 0.39157118451975237,0.4501944698800109,0.040607025782963824,
 0.2969142821505687,0.016784944040973304,0.10036240227287074,
 0.3553103830293085,0.8510507306743792,0.02970632621547753,
 0.28019833893327706,0.6405491744126255,0.683710901442686,0.5690979580480403,
 0.49531175555046336,0.9458561448668652,0.23343172294316483,
 0.5160605334470258,0.6725224495773771,0.5585602472336166,0.644803476290625,
 0.14127586989449126,0.9308579869014356,0.75595240288502,0.44526824153947553,
 0.1490040253512186,0.06675714258611687,0.42389894342582657,
 0.6903535318606009,0.08402528861435554,0.3906729767814945,0.550948202790726,
 0.38546529524153794,0.6158236158392504,0.8709472994082302,0.5281894058417527,
 0.29814914685755545,0.27147709403628106,0.06303378173136898,
 0.12561397864289192,0.5848118425186931,0.13534307577373683,
 0.39073433635985777,0.1560689400330484,0.9816711041852281,0.7907441365605449,
 0.4776064952730048,0.21448280279731746,0.5364772877493564,
 0.32435682622068573,0.42504458431370584,0.952852658311693,0.9667498801252208,
 0.08789133061026144,0.8877966337599777,0.19335406219564188,
 0.6062728997401181,0.8515465477461291,0.17592509286696278,0.6876362762998741,
 0.49370998369353747,0.12240304756064924,0.30770016393715793,
 0.20039017027872474,0.22237540691018443,0.4142214314489311,
 0.7050364288290878,0.8981803746514164,0.7664741036661205,0.856184562109995,
 0.5602761928541339,0.6109342688113278,0.7562033097833003,0.37315308774274936,
 0.009428279563134634,0.7209319627367576,0.8942688007307983,
 0.9280892674567031,0.7088433859298474,0.8231482342482449,
 0.045328025626562685,0.7074295190513711,0.9913916511068683,
 0.05851103843222749,0.006626284180136333,0.6062733255150832,
 0.5476940126957281,0.7858234067804591,0.46515920079897755,
 0.42005371776102907,0.15521597091536066,0.5607134661704289,
 0.06081307445011608,0.5538588227503745,0.47933657065593105,
 0.8416191666626832,0.11635431979964883,0.41630882828432403,
 0.8481388191352563,0.14921817477527,0.8556903454207434,0.28931534301925865,
 0.05510194124059542,0.04529317128378618,0.6628807953536755,
 0.9878072483303382,0.2274107093828064,0.5747129361331653,0.3537649976490854,
 0.4062844224001656,0.8949532626077248,0.6454465125285825,0.7496739697904773,
 0.9866098075947999,0.46071775126077463,0.9939081044380392,
 0.38703829654600097,0.8060722982200177,0.11542265983093114,
 0.8401860260215929,0.9771934213945951,0.5777522893477633,0.8221052722417345,
 0.3310345618686016,0.35160861064998605,0.31905350713948055,
 0.4171276782829413,0.9666220755515724,0.6298729166083746,0.08803480716477874,
 0.7635823410631167,0.5250611953129745,0.14547546673637646,
 0.015735643885959605,0.9554220075360127,0.5632979373884655,
 0.20510905679479607,0.6923739326065957,0.7357281543912362,0.8652403859942802,
 0.013066139461039294,0.40852684016749474,0.7135818900024702,
 0.5075699372418263,0.6338027105578229,0.9045736223863349,0.9062877567100496,
 0.6997355533005616,0.9671133995198182,0.07998941721840569,
 0.17240347894032648,0.7490826372264813,0.42165521090641933,
 0.28701473490209295,0.01478945428044287,0.310780208967841,0.5292566994110951,
 0.4976663121107622,0.6421315131776174,0.33163154931094185,
 0.10383320217703917,0.651950145456559,0.4952106594515524,0.860795732617335,
 0.4206032059733079,0.010842323433261303,0.5386739366184446,
 0.9260364004336153,0.12204659998467637,0.8672221936011214,0.5442675750784524,
 0.47763896788916815,0.8539924295876207,0.5302241737006356,0.9235930367539761,
 0.28315477599644734,0.4258561338206672,0.49667996264260295,
 0.9054409012573833,0.8902990587075732,0.34881110182714414,0.9296347651851293,
 0.49986495400832653,0.1858198898015172,0.4591430851183931,0.7817749476515481,
 0.7291773923919947,0.9058408187596035,0.21467459167737357,
 0.11386469990659576,0.9198823014386076,0.6175183605904798,0.5268556705364412,
 0.40380951227019024,0.8632894970594892,0.7604490798366405,
 0.17849255689125854,0.09717656146223241,0.5237292789679916,
 0.04796683369568444,0.5691867028872875,0.4838583349315899,0.17583039327285,
 0.27228280981355146,0.6553946674291148,0.8303805898565169,0.6300368660358872,
 0.6311273017114011,0.6906032399380135,0.547898684243798,0.9119084083871494,
 0.8166983343838792,0.10093059189481424,0.5375679404551398,0.7809524058828335,
 0.7865657607331297,0.2876783054380113,0.28762163643051475,0.7791697807309041,
 0.42112207872614804,0.048802945867984504,0.7211811069322915,
 0.5752172765848247,0.03853556817784576,0.09105803390248957,
 0.1315422838879925,0.15475063630246666,0.2716125907297746,
 0.48730299600831417,0.2797976908047053,0.4575335075022473,0.9687331410135002,
 0.6854613010760198,0.3919635382921236,0.06724726118839397,0.3275659095509572,
 0.28989623048225766,0.27323974804597706,0.5453393845278247,
 0.2197788858283677,0.984428900950206,0.8867525987451985,0.36202695720650724,
 0.12850698975266028,0.3571952411432061,0.33884129477628633,
 0.6890511834608122,0.7793157842431352,0.15661721534138762,
 0.22951053388013587,0.06399870597909119,0.8884921420947006,
 0.04774280677907061,0.6393029511315005,0.19098102181591803,
 0.6850839377367088,0.3346092298827441,0.61797671997686,0.7362000289382096,
 0.16391492458653012,0.2615627099517187,0.5236162401035964,0.5057211225871894,
 0.2599298517426549,0.6038843309086234,0.08708966100399318,
 0.29203827656992964,0.8429720323869972,0.71605749309706,0.8959830463298781,
 0.03776455523666589,0.2410756098281298,0.7482694712429696,0.5100655159478136,
 0.47590717623998413,0.11453120639650394,0.15258042810869243,
 0.5049301795250998,0.973079626690214,0.37025146994115177,0.25893297746852006,
 0.7247584281452208,0.25622178158213615,0.3389097747275245,
 0.048197273741987456,0.3716876107000542,0.7702785480158516,
 0.7967033056250329,0.07835296703004146,0.17708790835167065,
 0.1595476919336728,0.5252157594367647,0.3703212508640441,0.36039763140006,
 0.9078820493043542,0.8944505672004901,0.6259170177764286,0.3052206590027693,
 0.04005024166574045,0.14009130982015472,0.280268783760951,0.3024183840076172,
 0.5165660202040087,0.8541058331833684,0.008313319565905641,
 0.8537683804093222,0.2479920545957165,0.7515698982105827,0.8079097606233706,
 0.9779598586899687,0.9151688359528547,0.7832849993497778,0.47536395696872624,
 0.22511445959482101,0.459730863066916,0.28858329222874,0.6488083277390438,
 0.061418754777062934,0.05933650412968261,0.9863352043563014,
 0.5803372735357999,0.015387685022114295,0.007693187370140686,
 0.05631437870282474,0.26777260551756554,0.8847627270572894,
 0.039026210051479215,0.050237862944402134,0.3540593803386024,
 0.9003053237375347,0.17355333485293034,0.4165023302596673,
 0.45115653181805593,0.5925320136208663,0.7418492495478152,
 0.10123878206956849,0.4013497167176585,0.28527834632436755,
 0.27808425794691594,0.39874728774716617,5.707445227306485e-4,
 0.22489608066286026,0.3288894599303538,0.35278646920495105,
 0.8725438452931178,0.2528588576659949,0.5126087287716881,0.3812620721568969,
 0.7918971977018174,0.9483279607986811,0.09202795651859685,
 0.14940007866622917,0.5344578817182348,0.2862340780207038,0.8474275867978942,
 0.27345928051619417,0.5350127744198542,0.9142532024080258,
 0.00420499158685983,0.6090461455906024,0.5758596835964889,
 0.19605638912721957,0.23321960447390777,0.426921421598506,0.9526396893918906,
 0.542663765248119,0.3269090897355186,0.11204606745884793,0.6208777410282719,
 0.7372761860672891,0.766983733857845,0.9881802566887931,0.40992495640203963,
 0.019046090015171435,0.5800333751341871,0.8533614318646496,
 0.7700667838075316,0.21875147513566118,0.933320916866883,0.9319988087193071,
 0.11019084959452896,0.8142260969353575,0.6711338965938258,0.9362374526347227,
 0.5184511830765909,0.5057311115226867,0.6748828707585329,0.9736040568602602,
 0.554663699717419,0.6407833636647526,0.04944768756281204,0.1620641270606209,
 0.3063071921647107,0.9974356679600341,0.47806984362444094,0.3572973638711544,
 0.052400592381813604,0.1425728291216143,0.32362192570691506,
 0.3303746074121535,0.4728418334887985,0.09784703350905266,0.6718267288533868,
 0.1639192183555067,0.7861497365307004,0.579512715565511,0.8769142722236818,
 0.35030674899514447,0.06221206972644333,0.06960504302019843,
 0.7739326555373311,0.12544733533491326,0.324802163287394,0.742028326232344,
 0.8543845737190177,0.13740837933651373,0.3704712185987553,0.4051395311589543,
 0.6735043589085546,0.26734955237516866,0.9668130342609459,0.3434664762615012,
 0.9856484748431927,0.3797576510418782,0.6551215196227551,0.7198511014783919,
 0.5420693510501823,0.44950753212455663,0.9386525046424771,0.4113350954014544,
 0.8917374270556846,0.10536017931959596,0.009608962726394887,
 0.3939490754186924,0.4401927151979744,0.8848881903394435,0.18200323099493243,
 0.015030424037699053,0.4391330816373301,0.22617849204334917,
 0.7561114111519954,0.25199343892446724,0.8151534637416811,0.3371418440794215,
 0.6413753271644964,0.5851929018167302,0.6771235762815511,0.1830851284536681,
 0.5108226599686931,0.48291800386556993,0.8063365866684878,0.9714707887145896,
 0.3583482789977225,0.13674502749834216,0.17536261297828326,
 0.21143186805385428,0.32462902864179877,0.7791544289723266,
 0.33535006965448755,0.3626548594103126,0.1110764094204777,0.6128730807622487,
 0.24816204726174262,0.18699251436687592,0.490812194827981,0.8728508760101533,
 0.6237945894166986,0.12745510495814638,0.6272225148180152,0.7096762606687668,
 0.5901835864876631,0.6335769633090685,0.07314669829058218,
 0.11199969217157457,0.4242181964808822,0.3043836735736014,
 0.015117417996216798,0.4816684187642607,0.2508746767755662,
 0.5242520078188765,0.7154698059208218,0.5368735531587997,0.910083435246294,
 0.7256310616313415,0.7046284849230141,0.6357361432589308,0.3883551208502999,
 0.3320413680504589,0.7746026500931251,0.612743099941636,0.5047040822897408,
 0.7536582710353714,0.7175444883572212,0.9041603836797433,0.524634771634812,
 0.2557532330485207,0.05151143293804694,0.08142152570700978,
 0.6311888191269318,0.8589141971292947,0.7005303082795706,0.6766021557210186,
 0.12913505902078248,0.9761810008687118,0.0707300848582102,0.4776181227841516,
 0.3162545177564664,0.5403014508013385,0.3512853838006287,0.717308184517368,
 0.6421590464344274,0.3647681356531586,0.974362005490842,0.9952077147239763,
 0.11592940992717726,0.19165321511571565,0.17043231586494356,
 0.48724757100562543,0.17326013064690304,0.8174955133301989,
 0.11522640104300952,0.1559659532909693,0.6553908952144452,
 0.24840965387478842,0.3301238173650627,0.19126994788378393,
 0.5806209574017204,0.9695918647033787,0.6525794962427004,0.7599705937409524,
 0.6575475263637823,0.26018003375890586,0.2774238983534604,0.5801634586812288,
 0.7397984748977544,0.7806384496234919,0.5068963820712791,0.40347765619523046,
 0.7073277073033908,0.09612568636741003,0.25685685381165024,
 0.05077519577458822,0.7925827282861984,0.07396017619758999,
 0.3935879785961195,0.5332686071625489,0.43012281059220314,
 0.019527025894170547,0.507331491265129,0.008160983579990622,
 0.586889246787747,0.11793660053033428,0.4119598442146276,0.27012057265749523,
 0.8011768900280549,0.6515567403514164,0.9081035033428283,0.6616141590498257,
 0.3513758199414633,0.4741554198939464,0.9216327111892834,0.09511503193100168,
 0.3712848335618951,0.847104178153209,0.9777102349834187,0.823647611602205,
 0.7002393082084035,0.7644305858518838,0.737385373271744,0.9932804077352599,
 0.3024507735370199,0.938918961735515,0.18079551883161282,0.06173296887816071,
 0.27629038610621337,0.3436293530415747,0.2658753542111806,0.9015676060916907,
 0.9839243848130361,0.6393368262609463,0.9351318572882291,0.28725734678930026,
 0.6622856722726358,0.770472506346159,0.006265645271326781,0.5862719426069639,
 0.9756379721311095,0.36748304209085647,0.6151571315423844,
 0.015639109571561893,0.759516812105826,0.6537199512014367,0.5793080433498377,
 0.0638768259382756,0.8875645212649322,0.28278962945026354,0.7057868599684379,
 0.3385790087002428,0.7654319989797439,0.6883720505362827,0.6855668495117393,
 0.6753953618117032,0.27577420160643384,0.28764093457385576,0.77161394703664,
 0.9183078117031052,0.08323413170708815,0.644752332107181,0.6265957740039109,
 0.6213197481265857,0.9614685896707601,0.5430287570946204,0.7323958137582912,
 0.4211741887833712,0.21230033663652603,0.8849270040120069,0.7682306002351674,
 0.49417147001655193,0.26837898911987346,0.2549000195466311,
 0.7448814375739524,0.11385529395753224,0.4232426111128704,0.5579522044227371,
 0.9122249949610237,0.37262689500647894,0.49358220988681134,
 0.6555351258344637,0.05409919148071218,0.45532081982798145,
 0.27144558025790966,0.2163310173981282,0.05906248508039158,
 0.35785373069898163,0.7906422372169328,0.21588929814787527,
 0.34042564621956384,0.5854607690719321,0.5617605238461325,
 0.36915697595502406,0.9188696965946869,0.1692141501511063,0.5023185720830798,
 0.1004194003300869,0.32664995079926595,0.23029883368999415,
 0.42348429683783184,0.0027394029066076797,0.9153617111140854,
 0.9817630347764884,0.4313343198577878,0.4193007517180032,0.8878845346170212,
 0.8348375156377377,0.11242010101475008,0.5593538459957191,0.7786808673488897,
 0.7291055272323084,0.06018429374813328,0.6712914557599673,
 0.14325491584049577,0.24907411129024515,0.9369748431802508,
 0.7771025468826578,0.24524202629044434,0.10371893935357668,
 0.3929180928152127,0.48805609383102,0.09780638229489491,0.169412504093287,
 0.7477808602708657,0.2421563163339564,0.021923261605856115,0.709986307031992,
 0.19778726800157476,0.29682877614665726,0.478452496777493,0.9084220058686264,
 0.7611068990284703,0.9280466894370832,0.3719588198855923,0.6642454419010757,
 0.9911255129373912,0.12861973051493591,0.20082594846311141,
 0.7781061091734515,0.66042769414199,0.16081891432399675,0.4410319557102297,
 0.02357761389308599,0.5599483807888137,0.37314366241868513,
 0.5665586726657046,0.912692537319902,0.7615962405395933,0.7353955035382046,
 0.663102309323858,0.6226544493047157,0.5995637041889491,0.3442094741145595,
 0.9399264927600063,0.8285985818019623,0.032545148457570505,
 0.3374032455657602,0.5741655297078674,0.7496567242739303,0.6716787210674169,
 0.9568001041173357,0.5718125312691197,0.35801445846700053,0.5143615469811816,
 0.1977019972030888,0.9199889615336441,0.8174318446258334,0.9884199427274499,
 0.6478045875200513,0.5020820289725287,0.7225369179501036,0.4525542264879855,
 0.2682055008109542,0.44569285270926273,0.8785983912286371,0.6933757340553337,
 0.23606052031930136,0.625836181539731,0.679394226896032,0.21800956607042066,
 0.24435189453887385,0.19013575649092873,0.6373522741395306,
 0.5942339412145661,0.8485067791862542,0.8481425675554918,0.955829534799428,
 0.03315124427460181,0.17170759067877295,0.17744815060825136,
 0.1219242836889014,0.7823565080180684,0.6192123039996231,0.26089682150682747,
 0.748481254105883,0.7235817604254752,0.7454854527290513,0.8394904166017763,
 0.3514907285381703,0.8749301402419621,0.99576264384361,0.10514696377806909,
 0.3068252781979758,0.6652923285866852,0.3276616057803935,0.8259298918489175,
 0.6809355938217762,0.7756243584248455,0.5383396851144968,0.5619088540231613,
 0.1425709256279265,0.020619519947803422,0.6014127457416957,
 0.9943501465353297,0.09092467209097777,0.47473747102601493,
 0.3572499457142193,0.3006507140047969,0.3220732116779179,0.33049265505399017,
 0.27735937105695974,0.5968644092941914,0.3993386168771481,
 0.21475740541788135,0.889641297682272,0.6386582689184191,0.8762698775186262,
 0.585024914388294,0.2149483312367162,0.4979363166247627,0.0625409738043059,
 0.7420554981246111,0.09364467278771171,0.18043243088336047,
 0.8300367122405404,0.15399861078666155,0.8490932902705781,0.8440870879135869,
 0.5064824218013557,0.04376210137563208,0.127381155430344,0.8030535197368635,
 0.2615670538919408,0.05097721081407269,0.6369434910377079,0.7116003107415355,
 0.2509053138306413,0.5090218886626257,0.3895798136566062,0.05529024833690077,
 0.3458153956354151,0.5896668871079705,0.7425494746679403,0.5449684647536002,
 0.5789293955718828,0.5114753610721995,0.7182522879888708,0.6630113912787772,
 0.9172233822738446,0.18944517484350398,0.7167555584510963,0.9996604805708411,
 0.10996740360556112,0.5144408943782371,0.5148796527226851,0.3221369234273085,
 0.1674820340363503,0.7935663464720566,0.8897478508084865,0.8759810133823815,
 0.3566968081158136,0.9356106389211298,0.08349860527052988,0.5366767950528342,
 0.146210747403688,0.16093538950364417,0.5183908296352149,0.857752335577786,
 0.48507932516665153,0.16397793381028003,0.7489170400950085,
 0.4591054507397536,0.6440853312740057,0.6148949293480246,0.7438102372605432,
 0.9019992187030121,0.06016257162725158,0.5298269612015443,0.7693137503494,
 0.5503187915557006,0.0038506554183458386,0.40308675019924367,
 0.4965149889992939,0.8822009495540498,0.39985309409153635,0.3373239526287989,
 0.9512929381406101,0.9903387584185244,0.5428130634750776,0.5637116804547679,
 0.37383988437925453,0.5928059474434147,0.113898544299747,0.7064949854566489,
 0.3196950181348135,0.2332061892422439,0.6993441682086095,0.42952417682623834,
 0.5940646952589106,0.8160820199341948,0.8808649610267464,0.2159580939706629,
 0.9936011791243494,0.18064436360037295,0.23819965959464018,
 0.6373056440659527,0.25235824113756034,0.3436201325954302,
 0.15591933361791532,0.927200554538005,0.7008959728985592,0.23166652999875847,
 0.6534189917693727,0.14851187453895742,0.04363651342636654,
 0.8284630101288577,0.6852778250184282,0.8555020514923344,0.8529014443789372,
 0.24230299344018558,0.2093827311601939,0.14262439083605083,
 0.6417632228565872,0.44572976146518306,0.2805975325830067,
 0.32067476764662906,0.07112300121473192,0.43577039288492925,
 0.42099154939435124,0.23382918514872963,0.9080367788835214,
 0.34474595715893996,0.27191654558378875,0.27749811194546337,
 0.11660550452104224,0.15496440983842097,0.03142851573009997,
 0.956346781826749,0.45229926872216386,0.975865462266974,0.008372993533542483,
 0.3913327656569684,0.6781617476964312,0.32319783170185434,0.5467453758721732,
 0.038856150103636744,0.9244988822991287,0.14735263297308887,
 0.9313767373078832,0.22512511374181476,0.39464743715670914,
 0.38600681339386667,0.04720599875361775,0.855970806476991,0.7503957264096699,
 0.28925014070352284,0.9606185348982998,0.3848913060014836,0.8665454546720764,
 0.7164539048524085,0.6061455987974972,0.4594451516186637,0.7763433945948783,
 0.18041144031869127,0.457278834636785,0.16698032727404855,0.4122876990533191,
 0.06509263112383368,0.5095786260594457,0.9078836451995135,0.6496236042285867,
 0.8059436913101696,0.6682566122876595,0.8697628878397674,0.9023027008810924,
 0.9785509552586245,0.0787705815802292,0.18898169561077793,
 0.44330037285019364,0.6927261278833423,0.2570931393556105,0.4698674222690832,
 0.7758493573402427,0.5205424099472254,0.08928872165755108,0.7215730842067398,
 0.5123005522876632,0.857114979743068,0.3971033608284491,0.8495231203291755,
 0.854133164090044,0.11891449215597749,0.18987104596822146,
 0.16448405656004272,0.7784165055447121,0.40524336428764873,
 0.1375100155664024,0.03806692195744188,0.7423473547527923,0.6488771149286265,
 0.7224493227593981,0.8873910214253902,0.4885180731288761,0.26122321786436375,
 0.46351417095423275,0.29223986196448826,0.3167512197987036,
 0.7903728139885173,0.6976165085473276,0.3156680816393872,0.15050979133422404,
 0.6150780869428588,0.012926850843147153,0.5638046201373332,
 0.16080222958846213,0.2110456716136324,0.361079536832862,0.7516258872175536,
 0.003108267527631825,0.863167001199213,0.7270182489878048,
 0.24557147103049803,0.020613770351755045,0.5177902780725988,
 0.5904281315047322,0.6049555833420922,0.6762560221892043,0.8328261445686667,
 0.924431491882072,0.11078045166900874,0.5589066912626077,0.21504339672953932,
 0.9470929298033754,0.8489317158545225,0.538506208026361,0.02615136607345836,
 0.8661446603052042,0.675313754974209,0.6045362374541697,0.8875040581591446,
 0.42044799942890265,0.05231835100886795,0.7836293236390545,
 0.9430743402053928,0.007211394728221743,0.4539182719467798,
 0.3948771333756529,0.654794683384312,0.7691289861477747,0.9258547757711807,
 0.43918851147171367,0.09537670900312123,0.29276238502448115,
 0.1429817960409302,0.3018741271187084,0.05979169269874507,0.9389493728772091,
 0.20842333651194866,0.7039790373401168,0.8059066953755627,0.8271513357433954,
 0.1181264316058549,0.8947389964831376,0.7632167250173452,0.510196929916971,
 0.7720942154837634,0.14486662928985428,0.008025553500607074,
 0.8838776513073461,0.6525690378969898,0.5480207871859991,0.8954609302448312,
 0.0034207498205410802,0.021668527221766487,0.737933716948626,
 0.6362372526289568,0.5150335456435706,0.1782041603701482,0.6874240922421228,
 0.5524448969909864,0.49257578670567925,0.7046062023741246,
 0.44264917941739657,0.6935960086111908,0.9356198472408349,0.5008830545221196,
 0.2076546034039196,0.7191611246676931,0.24907862060745622,0.5750206953755221,
 0.5622436190693532,0.5077749688099304,0.17208602910612014,0.9507273317552175,
 0.9773537739445767,0.6126652233161806,0.5351927200122593,0.8924815480225514,
 0.6755811439781133,0.15908170001608868,0.14221728822723112,
 0.4207622070261242,0.03828884493072626,0.17544340918281454,
 0.07388364135446535,0.6669524590887097,0.8106946747169045,0.5773363198955012,
 0.3106958952686931,0.48845464442019915,0.40902360865377285,
 0.19836247333333912,0.8755090105579747,0.2590539907310262,0.0497225301481814,
 0.7537654541684808,0.36603203136807716,0.5115245695604376,
 0.01815988200917107,0.26151122850245234,0.23050527501610774,
 0.5393661457925507,0.8484109942951393,0.328633961261674,0.2447371796348079,
 0.24375555697044826,0.15733971807082459,0.9764148368678628,
 0.26146650147894457,0.9691352906318756,0.7756581094730753,0.3529744575520728,
 0.005510230965538154,0.3481857132294609,0.780037214047923,0.9027585262665437,
 0.004346356660659301,0.5405577847841051,0.6758185893320937,
 0.7097676420066027,0.2407123630757637,0.4437740514411844,0.3434068389265763,
 0.8336371672746374,0.4344842029897321,0.8472697738506061,0.8842022085634949,
 0.614187819238755,0.24657165070227716,0.3733216303989728,
 0.003715485237875349,0.9316832491494895,0.03710201695603821,
 0.39138263995675104,0.13461272015751158,0.6602751804516336,
 0.0456850923518054,0.630089829872855,0.8660265164812446,0.09803640071227282,
 0.3917515227644871,0.7395701974644273,0.48482015683956736,0.547597577118664,
 0.6635866208503404,0.2426814748413486,0.47470185006977383,0.651490390506149,
 0.7738096759216969,0.276723220288813,0.2071205898517391,0.5864666522217918,
 0.7701736013280687,0.3749900549909203,0.4386264211483426,0.05662182204282962,
 0.5834765679740764,0.14637425257579628,0.12836783832061482,0.876932643730867,
 0.17869885429391374,0.5810621797068074,0.442728628319765,0.8557481160293717,
 0.13583046307591395,0.5351970348678512,0.5265889943429132,
 0.05621986236688259,0.277553286088561,0.39448233662365884,0.7722613344925271,
 0.787454750612361,0.47457817972039607,0.8440223069421438,0.17315157961546457,
 0.22543092396778175,0.14672059536228543,0.301783714554257,
 0.08771924951736931,0.4513031003147965,0.7209695294584345,0.4923835219705416,
 0.8956847169929746,0.5936353233210971,0.5651663565418771,0.4039362473384408,
 0.7818229638192105,0.7346552992775891,0.6682983167010602,0.3760211587530008,
 0.8713279775161358,0.07522157317275602,0.4624793845732068,
 0.22293919027014952,0.8582494350883745,0.10784580110497666,
 0.12351730887117163,0.08915776810342912,0.22838559620417953,
 0.7394637843117581,0.7961909329289067,0.2829391961374741,0.8957289950824487,
 0.9265153950254472,0.5420061407300742,0.9675473025739313,0.40501496696738437,
 0.256718001796837,0.4594073148263269,0.899835978911355,0.5936817589145216,
 0.33097954595375867,0.08193398930063989,0.4025766880027407,
 0.7485663281881999,0.5007422390906437,0.34045311716984206,
 0.026438733399679393,0.4167086599033951,0.3242837506448947,
 0.1510222516697448,0.8397419289923557,0.08955877682691837,0.898993357092736,
 0.8767358717700552,0.03957144612767005,0.5214638725530512,0.509958896102236,
 0.9400028586447102,0.4757356796298263,0.2974186103674158,0.16318886407079436,
 0.5198043142030836,0.5286362244193856,0.5252773636845305,0.6956676587077665,
 0.4853905299580722,0.5464279149471387,0.03607317059399551,0.646270890289256,
 0.41265539792399153,0.3502388733200015,0.5347491018771239,
 0.37095968369360677,0.47911340975632455,0.9074078745493424,
 0.03240712675298718,0.8531432514583922,0.5953072937676237,
 0.10158513950725678,0.9336594579987775,0.09770254412889134,
 0.4139493533605738,0.3822570457658423,0.07402183409526353,0.9267803098489225,
 0.37449417269235086,0.688055912424849,0.14234978399965625,0.4496126200530659,
 0.1514286856485647,0.5640439814069722,0.16765938772935363,
 0.07952583161416904,0.2764911995203567,0.7116145794348246,0.6472687576181477,
 0.03312090022951497,0.9136050981122751,0.08560757231540539,
 0.9725972983807483,0.42665914813667394,0.6490395405737881,0.4749134326872948,
 0.40286902946377434,0.5018844960568224,0.37112822744565854,
 0.7903201120329874,0.9220141469481496,0.06888484073366796,0.819786644285262,
 0.9617965509829615,0.11682018406578754,0.31051014473642624,
 0.17379859049731927,0.509172575864442,0.4593951761030114,0.5760950585869328,
 0.8369595299045403,0.2617384400108622,0.8090395536617817,0.5655329191341668,
 0.37924801789514784,0.8138243531431362,0.7837468668580845,0.2845570000771359,
 0.4123816573411587,0.43142186578617925,0.4534535283813397,0.5394725059945303,
 0.03105929166475274,0.4325634484966705,0.33313505138442145,
 0.8168457437817959,0.4006975444374382,0.7662358832681241,0.2670222175832364,
 0.5853176066814119,0.9916735813247906,0.3322216163595211,0.3744804178038994,
 0.7355464514909642,0.24715084656484954,0.37414819833487334,
 0.06188401849847691,0.23502174584879376,0.18019302731116094,
 0.08488529416571533,0.9244816941888233,0.5855908960199672,
 0.27250496698122095,0.17755750166332462,0.8703012657829095,
 0.3802163433091734,0.8002468566391335,0.3469055184337204,0.8930130049512794,
 0.7464979604305393,0.9072961275646303,0.5657599749095098,0.15784018916218878,
 0.6776586078224965,0.8428959133943964,0.6598556242957977,0.3728499832426957,
 0.4197661092202143,0.37277450569052306,0.45677331318835424,
 0.5033848900917173,0.8457846335843511,0.5602543386265214,0.38897839879756924,
 0.09645347254811576,0.6514454929576469,0.6982671346768135,0.3769048340493928,
 0.10295389299120006,0.5854353993830033,0.4150749752962105,0.6373365310839714,
 0.8322335875242355,0.49722003032637974,0.7651021531327671,0.2881137540229409,
 0.4936746665233147,0.6763291483197247,0.45693171028525237,
 0.18013469343173893,0.16228127112291246,0.6280676368827158,
 0.15526363165472612,0.4091438579115204,0.5058516390220503,
 0.03406996902836579,0.0165960479065127,0.7126632460451194,0.8267027972863314,
 0.6292257647971718,0.8456342428606191,0.6524364275403581,0.09868354573263671,
 0.8971081613573472,0.7660619116976162,0.2740695274966023,0.9815652929451587,
 0.40035107548115234,0.8334715702294361,0.024855132120082724,
 0.16315157397257474,0.9349986824039331,0.5083199450690516,0.3497882594435211,
 0.5160705626320684,0.7301209541642582,0.5561687052431196,0.9352753597246612,
 0.0291785352559063,0.8893541397178404,0.6318164100972192,0.8110984932075881,
 0.371265077288255,0.9756304602644555,0.4217536943642005,0.4962342942595176,
 0.9967916587076648,0.6301440846909672,0.8906724684055348,0.24130686194614948,
 0.3848610856378627,0.24323216607855924,0.612914722115614,0.7494885590244779,
 0.12350486763169954,0.9117467564460546,0.10099438015826312,
 0.5224183443157897,0.8801576212874661,0.28432965829543577,0.6216073770379779,
 0.054353027591305225,0.274069559625689,0.8557854235887341,0.8229294666781175,
 0.8788797797577228,0.08667497921736167,0.32622652714837375,
 0.41474856707866337,0.3859416285483509,0.558717830346908,0.6917895525514438,
 0.8561478847068701,0.5828184462439423,0.175846905726452,0.8579024163264417,
 0.23880768401375174,0.22309538169747578,0.6411971974283313,
 0.32905890370243995,0.8208229842187327,0.2275572798978558,
 0.06407753731449328,0.755958575779425,0.3278565271910806,0.8363837917005058,
 0.6610630940140136,0.32904223712690417,0.07603022405910931,
 0.27206417787085435,0.21958391781959097,0.2564946088596266,
 0.03536319382576525,0.09179591425022782,0.623623628937502,
 0.18241728189388962,0.8958681372571058,0.712438108209053,0.5679487823010311,
 0.5231591920689391,0.7125758813427392,0.22753850834328326,0.9665938446146274,
 0.40272361285685887,0.7379569402610411,0.20690310035141313,0.733678372965215,
 0.575033249377728,0.12622653333133882,0.14353154969561066,
 0.20054658067430498,0.7864839155740307,0.26239641122736135,
 0.6116758766349432,0.21686849634234806,0.5252037575350695,0.6883323589836776,
 0.6175927912675472,0.34523656826877347,0.4812890049354559,
 0.28953074529125566,0.1845618571677965,0.4093448592032125,0.6318254624884061,
 0.5636733145738342,0.9512487000390789,0.6057195403280136,0.1660182055920093,
 0.45141088897680726,0.9305287640007086,0.25363399862087854,
 0.20166715990843354,0.011457096325214411,0.496457816395768,
 0.4133148825289277,0.7797957553045736,0.1506532329570206,0.11422885746251643,
 0.8944421637384106,0.1706704258510482,0.4687298210548049,0.14673913976812858,
 0.5843338433670713,0.3064027641952316,0.7507287126273252,0.9476456691925411,
 0.36944779782823456,0.21695390189642505,0.5115279984677508,
 0.4557669761728069,0.08656392432925542,0.42429555811455133,
 0.3528902360302253,0.6575501751671039,0.9127359067963282,0.06057117068798257,
 0.0510680752827557,0.898910504794597,0.8629119716339662,0.48699627295425574,
 0.3957113889833652,0.7397973259195996,0.025087266211064607,
 0.7421784362145573,0.722764607909531,0.7832183301322948,0.48392860188476927,
 0.053815365164437634,0.5653253711564379,0.6006446569385925,0.162981722910996,
 0.09823512489304453,0.14940131587848604,0.4484768522946254,0.315688059820593,
 0.5790915904487137,0.4760774069525747,0.3933805394149399,0.5995183757530625,
 0.5323763527364415,0.6553434892076618,0.5733105295492504,0.8942720814295007,
 0.41608616014858535,0.9759500094762685,0.5848563426244814,0.802871494013325,
 0.06938776974439465,0.38593907891158086,0.17287009875439263,
 0.2427038720160657,0.1695621985762178,0.22628527819517963,0.1911383698155773,
 0.5477613541509284,0.598765576592731,0.20409529748632937,0.7369616817652059,
 0.7204172980129339,0.7757497117651435,0.007308380658535674,
 0.08676416558012467,0.04856589526130417,0.934679808964549,0.8739556311622111,
 0.24986225005880403,0.14714396513700878,0.4725142919284776,
 0.33686406690017023,0.08551321659076772,0.22508963024063577,
 0.13077125083656238,0.12463455166602555,0.13849236534001197,
 0.8511921329011115,0.47848361753644053,0.09819687434834024,
 0.23962851099960392,0.6441072532601879,0.9427497260482011,0.6559044522534255,
 0.34502725325305483,0.23052463334108497,0.6120519260853218,
 0.25451277907569925,0.33050564749893563,0.4004333190397106,
 0.24655665071497468,0.22501099749222364,0.6383032785252286,
 0.009306839623392071,0.8335934312803975,0.06489239977334171,
 0.24888474831242302,0.13063877292758508,0.6188490392020132,
 0.5847250612095442,0.9969538281431691,0.45468089082810137,0.492231646132673,
 0.591438805446651,0.09556022401527042,0.29644014825863374,0.607760736194155,
 0.42875055123029704,0.11592117110178779,0.7020995212137844,
 0.11686734390289522,0.5190703906940509,0.8741571585816073,
 0.42767995981799767,0.5739611479672031,0.789933521605064,0.6810222333274445,
 0.21320077202832155,0.3378319498046656,0.8566146132032375,0.7614089020270984,
 0.819057548084404,0.7704957393413477,0.46636538873801336,0.9284554158621239,
 0.8609905687510084,0.8689734543770715,0.6215640504045112,0.8027759519764922,
 0.660892752745527,0.9400782058586996,0.24388908678109755,0.6218305372591406,
 0.5286300677146245,0.7307021246151322,0.5188388482660805,0.6302689169395148,
 0.7675563393695473,0.7236677522434964,0.4314443563389557,0.8613633549228188,
 0.5675192079610437,0.16220141689305623,0.33954212748732093,
 0.4654696935668575,0.22628257496918713,0.7083827789651861,0.2679113990832238,
 0.21410769412784536,0.9820154256542627,0.8741683031640394,
 0.13030114878155408,0.5610925356652066,0.5084156439155842,0.9367549377462635,
 0.5302810780637826,0.7844913652853729,0.5191023914991824,0.7983307688299175,
 0.6088242479051796,0.5688999964898578,0.5812353715628547,0.29216754573805526,
 0.784626443878822,0.48646023708982966,0.17361119680558268,0.4500500341815383,
 0.056219355890578915,0.37081649919545334,0.5082102980896233,
 0.47816447604010004,0.42428765471814434,0.3353651647638827,
 0.6397066859642911,0.018090855032245168,0.7544705589038198,
 0.45417543219644496,0.9507893532521178,0.18811568798714384,
 0.3794475986800936,0.008021820637142074,0.015430777311123967,
 0.7280962120918669,0.1889178754302444,0.7096453499912595,0.5696180580444938,
 0.8275949842732746,0.7768379800578855,0.2990982993125767,0.12115894598583066,
 0.08721636714666658,0.8542795166472832,0.5031537007739789,0.9081966075016147,
 0.654629876015209,0.21532332446258695,0.358980060654029,0.33348618957187104,
 0.12165397809697676,0.9709872957184101,0.6788764496343246,0.8461944623714506,
 0.7858093330948875,0.8297186849380741,0.35890713708586575,0.5418981998256618,
 0.604090704262119,0.27435492111307025,0.9211509761438441,0.24733181964315643,
 0.35467077147622916,0.505812711062364,0.2128313854556746,0.43344860606725966,
 0.7323735129020095,0.45249216297936645,0.1194672042939805,0.7503285890315965,
 0.7817227258626896,0.28612279667098717,0.15375508248150238,0.966154028742902,
 0.6563709251359066,0.3763235541821909,0.5276709765197681,0.7176416637368159,
 0.7829660546187662,0.963288101485757,0.5860137018768504,0.6032640434678087,
 0.7516580687796729,0.627729162377918,0.84410513780531,0.9458888661325214,
 0.6553702882819685,0.9499484962554181,0.9385740068752946,0.831536917433735,
 0.47165945859989133,0.44651453033309896,0.7516886239885666,
 0.11505475138371846,0.7012013377157421,0.067398087211471,0.7555203947012851,
 0.44754903626825204,0.6269130737363116,0.440276738086556,0.8900555151650243,
 0.02587873705099286,0.8889871595529694,0.7947846138876911,
 0.07407314590055014,0.6209646266951517,0.0367331365175132,0.7238437592630503,
 0.9540507826992786,0.9779880578483319,0.836441647309388,0.8704874898479942,
 0.9491939954794271,0.7917919185211564,0.9320491051629988,0.16267128231098704,
 0.39865728370917564,0.8446876097136681,0.37817246477061695,
 0.07155166285140213,0.2868324900262236,0.17813064782641752,0.142681780208327,
 0.6355600417324452,0.08159601265882344,0.13029690869971078,
 0.7162128509426597,0.28203752056120546,0.9953750968164559,
 0.07417171433913139,0.8603617144818174,0.6602365391068369,0.6713621095774034,
 0.5344638451335121,0.564093026384843,0.439234306622451,0.7465581546354926,
 0.9714033677931052,0.28801477987645696,0.5637058961954766,0.5703173233444462,
 0.8095349526733637,0.4656850627522703,0.7997245097535779,0.8747232651079714,
 0.922008046648154,0.8628326523623238,0.05091395994927861,0.45414856816255555,
 0.9558733720339597,0.6081194467808759,0.7344787894515291,0.22583169204340192,
 0.42553090902536583,0.44161750845732073,0.5229156806757531,
 0.7109463435095924,0.354585265484368,0.5064756663072786,0.35655418826142005,
 0.828030321905068,0.3849515756125721,0.7419956542554845,0.35735655291162205,
 0.445488197882664,0.625420784580157,0.5733895150117345,0.5619680809239899,
 0.4774874217790358,0.8184204041129685,0.2959332701887559,0.7146346371153967,
 0.2164354079104216,0.02094935107758822,0.6934101692158685,
 0.13246026722630633,0.41488975271031514,0.020959185494455412,
 0.43566674308232445,0.879998420116884,0.34558373897997363,
 0.20989831371003198,0.3537576593406946,0.7427176659834431,0.2391675338090472,
 0.8034853135334163,0.1125187148942276,0.8134494265093258,0.8547534275472395,
 0.5579565571529043,0.811376784361914,0.22078957472985616,0.8454961829261778,
 0.4246623075883833,0.5196437169662578,0.3394968724470887,0.28456404596944795,
 0.8692524897372312,0.9465454553310786,0.35517177042777914,
 0.46970076056213983,0.7168112690544207,0.8837242047438868,
 0.027707081698460145,0.8837928806515529,0.7040163869891101,
 0.21041213907879686,0.8188454238888063,0.29140170841708124,
 0.5768014369067503,0.4955266560388404,0.5832549575938561,0.07867420008652559,
 0.42126905137929416,0.5406219081473042,0.7834593073927454,0.4639102409442015,
 0.45553236133651775,0.9550687715736551,0.2567802220230435,0.515584782152096,
 0.08804330393055892,0.22795893527001299,0.3356035308307598,
 0.7541410371135192,0.8972066108045785,0.7586320230509719,0.8958761169542929,
 0.06967628336625387,0.5815352970615154,0.4815774171617321,0.2750024608991031,
 0.6090617010469491,0.5645848128019416,0.3315786461115715,0.07373459529278925,
 0.07439864816055441,0.27215256187640535,0.26813198171305397,
 0.8621886770703818,0.28721977182874414,0.9552160986589162,0.8071294674504907,
 0.6947882422652795,0.5027779355931715,0.8651369228327258,0.6406344728053242,
 0.7419160794354505,0.5231447172035888,0.3880406769315289,0.7686730242397812,
 0.9334196331786144,0.7785551657651619,0.036897810553338295,
 0.4513624033346847,0.2044043628537422,0.24301181865581534,
 0.0014029759274587494,0.12431981297921302,0.41556270038326293,
 0.07754671422490522,0.9199108936131923,0.11217870653100637,
 0.4903600291053437,0.9069557118112226,0.3585775784313042,0.9599019620103881,
 0.7377339099359419,0.2767437728037304,0.2822031834000035,0.09899121981571923,
 0.17329987435522365,0.6075351370444009,0.49263263909996513,
 0.7062618994179086,0.9617299046613685,0.4653781507319684,0.6702729962942151,
 0.2537813804145572,0.3923595488297379,0.06351674621917591,0.653478116842557,
 0.490025561604136,0.7962818913069183,0.2512261421723134,0.02791795659681684,
 0.7855155916737551,0.257224619531658,0.0743338933338118,0.9509225651507854,
 0.002153588562462838,0.42395311751384923,0.47963016137283176,
 0.5444542461084372,0.9755084156138497,0.6369952348856659,0.6392399377268421,
 0.8987814441848972,0.9723430287961514,0.21476284819077596,
 0.25918979398133457,0.5096894268113382,0.005314196450431075,
 0.8477962562970438,0.7601768203635637,0.5792603290901411,0.1916460152471321,
 0.2157739678704157,0.8141652609731925,0.7641212628244789,0.3236640826205993,
 0.033374891384397753,0.8605291260971557,0.875135603433021,0.6736021425416998,
 0.3002717294310475,0.7395212065630974,0.6096884442802303,0.9130070513238977,
 0.9841486247096872,0.06969819587338533,0.2622766604854033,0.6465444263487371,
 0.2335891445039041,0.991411203850419,0.10019349245695564,0.770520712360227,
 0.3800519548129264,0.1934474091665329,0.07948093365849418,0.7003751777360088,
 0.6754583368158004,0.1454682390063491,0.9044032026201074,0.40260385145648614,
 0.524178327094386,0.6059301011105994,0.8605473346416327,0.2616918414896664,
 0.2984876822761262,0.2070987514764635,0.8958430796426027,0.6607235078075118,
 0.38751541306279547,0.3795324503144455,0.9930753765484617,0.5815936448647668,
 0.15816017304560248,0.6502278595601482,0.039101943173307285,
 0.6330268305654068,0.6171221739705689,0.911734749347998,0.6961403718942295,
 0.8783209409375033,0.713745998211089,0.2761105300516711,0.5004515994133933,
 0.16002227480772135,0.4479938902139985,0.6036137890735593,
 0.40474517877030103,0.7912232080440488,0.18787579962532375,
 0.0666683796086005,0.4362690750893804,0.17579107865338706,
 0.22499969592387759,0.4274550726847278,0.5138571180400997,0.7896190174198934,
 0.9864096937090627,0.849422002824226,0.13606465578135107,0.05286674365931854,
 0.551337749967221,0.25389731926647563,0.078106247025572,0.9891170888179123,
 0.47333238819170365,0.7813973504392603,0.4282370244648126,0.3511665220423711,
 0.3148159801387409,0.9122307085597252,0.1959676347676944,0.2668906260337074,
 0.8888566772337702,0.036874564427272105,0.09755591638218863,
 0.18899122572727134,0.4944561104586208,0.37999839115285183,
 0.09004520002064242,0.9602264033489544,0.586689291604906,0.06682701902599575,
 0.6028945084430337,0.32302690291980485,0.9426073099604222,0.5066819061883395,
 0.8275798917044354,0.5856356207587882,0.34016086055177375,0.2750964633939771,
 0.5537268460038866,0.7670935357363617,0.9386062583765368,0.6772255201345536,
 0.948857466970293,0.15285180898818496,0.8308741737834762,0.4136683012360076,
 0.4546785992160235,0.6650652257400429,0.2029887737724595,0.48129423146931494,
 0.7582407901855553,0.11384568930099959,0.4336356861578248,0.8356242967945142,
 0.18981997878477397,0.4097273132959881,0.8052053778761351,0.7433159351860982,
 0.7940613085347455,0.07822028668618053,0.5551784714477661,0.5515623065393851,
 0.32728076963204344,0.7895617413322342,0.6105061478632561,
 0.26470275585857495,0.3321649911256339,0.15560842546436193,
 0.40179860512443066,0.990052555191103,0.8981866296999647,0.4958690357150749,
 0.4574002381488169,0.9734340684241422,0.7969826465752277,0.36854712742426055,
 0.9600301821849856,0.2302824039698982,0.025921426893703803,
 0.5877804839138779,0.3764047757673925,0.2682126082419378,0.25876033626448447,
 0.8203284854459966,0.5005165268216675,0.6527201669599197,0.31828390920831684,
 0.5339669422210997,0.4606171573850897,0.15714952640457658,0.8806329353185053,
 0.1184744510520025,0.1743925480817894,0.26924113033326047,0.5936331035289437,
 0.4058488148402857,0.35513278047318597,0.14420125794178607,
 0.9152980653068243,0.5393678967730174,0.4569489867910945,0.5893549678312293,
 0.08795613032715732,0.5451393111372081,0.1930043866515606,0.770267714017931,
 0.521768376473781,0.9789430922181149,0.8848718545333387,0.8097143113421406,
 0.021029837035377974,0.1655658555373699,0.28516749833243926,
 0.9656477111753322,0.7375481559947887,0.9042936693512599,0.6948674830262762,
 0.5410414140939777,0.0871833062509415,0.8042215251217966,0.5699452177935045,
 0.1692578466387339,0.01865831589788991,0.8428821330171075,0.7742430223295996,
 0.4086513044851653,0.7819041126452211,0.5292424271972617,0.3524255049308813,
 0.2931669078362241,0.9084714802927063,0.24170927256782027,
 0.23741086018994562,0.3225880637587073,0.9707165360607427,0.7025044874414477,
 0.22961864567054613,0.2982241970875915,0.8459539848245357,0.8570270911010907,
 0.39275200650317144,0.7153088296637626,0.041075716563885645,
 0.17157775220931248,0.2491421086441178,0.08464109811890863,
 0.9605690519487661,0.29654309967980164,0.8092618242971217,0.4098615013708341,
 0.4754589159183481,0.1478726840054062,0.3779941359450506,0.5835848633444997,
 0.890033929924655,0.8845565358203615,0.8014333938247545,0.7441554564564193,
 0.40859052589917977,0.013443877645639146,0.5558840773045497,
 0.9207813505961296,0.4702357589273246,0.23424675795641714,0.4549311846354822,
 0.330004986326777,0.6545825954411575,0.4375436818777121,0.22716113738321275,
 0.485477818078367,0.02977592928697992,0.897724761436427,0.21642322351347842,
 0.1422360062445871,0.7403511122723753,0.39891613964798933,0.9794635231958954,
 0.9075572722339403,0.9316167565055514,0.994374019401089,0.5706334171464712,
 0.13750239444202317,0.7825081242333892,0.830323450793814,0.06421513618905283,
 0.47920593346706153,0.44249728751985895,0.49940672597082914,
 0.6037490320316519,0.5644110867144015,0.8091764204371847,0.5954765558454034,
 0.5666335276515534,0.6843691874101577,0.770578190796257,0.19709384638595528,
 0.21678203762309323,0.6310657645747184,0.14143817806300707,0.255396045936478,
 0.048971375959302565,0.9416125480436314,0.09675748385280403,
 0.22123129255577356,0.9648250175496704,0.6398646613694251,0.8958072565307017,
 0.48429751005562494,0.3615588214451946,0.08722337895945931,
 0.7715939792136605,0.1353633219598609,0.4789625296111084,0.578966743323056,
 0.27616872167247186,0.7931762330931673,0.8073888371113668,
 0.27531186371476835,0.8651615347066244,0.3748086981017825,0.5614724880340599,
 0.0682232003621136,0.07804815056682968,0.17817470280234182,
 0.8765901795857205,0.5862031508301042,0.283612934971724,0.221410177083744,
 0.2200336163282236,0.7522308977976571,0.3644479803310108,0.7090198053091342,
 0.30691432162949894,0.5896310450929437,0.8648049174670115,0.6773840287438705,
 0.21941893333547025,0.04990417612272058,0.8804808158811334,
 0.1868770573352918,0.586892662532559,0.15752531595600217,0.7106691343579524,
 0.027428463793399915,0.3196079796024687,0.1750457284374196,
 0.027455767156970956,0.8436711686409621,0.1367627989100163,
 0.8489251714095484,0.6249342677343913,0.20438049115977652,0.8762842128114887,
 0.8278118270768471,0.34944123916985625,0.6107056765112893,
 0.43404872534229366,0.7490673307503056,0.17838463776898383,
 0.3438045345904732,0.719050123504885,0.8023888305873176,0.8918291785334941,
 0.2601928941861038,0.4696199345263635,0.10523345001276696,0.7784838139139573,
 0.7239697518736986,0.288743415930807,0.91927307203121,0.27624151531928465,
 0.366857447834693,0.6402313995225043,0.49055977552314967,0.4329972173705634,
 0.8206979408850957,0.44320976002064427,0.013569282416050932,
 0.4644880967381996,0.03189626971897097,0.07891632684348704,
 0.23612874851387422,0.4672197436380465,0.4856332194997044,0.3375064767087613,
 0.20087337020376295,0.8273891488759426,0.6319573148231791,0.4492300883865711,
 0.22970915663974634,0.7630394931312,0.5948202285213923,0.5906871075894604,
 0.838906222909212,0.8924865727819835,0.9681213000117729,0.16713695164320086,
 0.05212650915594297,0.23713810397257196,0.593500069268778,0.9827330048504116,
 0.1909537340362717,0.840233730610818,0.08576171103132879,0.2553569157678187,
 0.4562489550875677,0.11432669449151966,0.5743704812636767,0.9139788270051723,
 0.5753096312249122,0.0945946266184361,0.9364308331775487,0.5910916898058165,
 0.7396794940765974,0.8809919254169836,0.016945181933768372,
 0.8156151727414335,0.6859695977708336,0.18881330818723518,0.3635285891508888,
 0.14233474740149188,0.6748082311043081,0.8949097556756934,0.72040129093588,
 0.6803605008872144,0.7159360703635328,0.6652614705355169,0.29008634586834414,
 0.63436276523862,0.09494233482041581,0.5485342151868589,0.43487007654217835,
 0.1495631046753204,0.15525555754155884,0.49011684847894754,
 0.0177717383819892,0.2992577402175559,0.7791023559558014,
 0.0020197154165464415,0.2889622008987013,0.6475583875350581,
 0.6778482846798716,0.615131064216735,0.8778313070344265,0.05626577342791439,
 0.32932279651280905,0.7350466241163648,0.5318450588043657,
 0.06084507347833901,0.05874812413125152,0.19564544271558337,
 0.4699252667736211,0.44221862336288953,0.8215438038902165,0.3327584938312864,
 0.3133704781278084,0.29018599286555,0.1784123051451053,0.04265557301546963,
 0.09999831322515218,0.4201827577941195,0.38888737334610046,
 0.03671957265797898,0.5076771587752495,0.4003613779438937,0.593932887484027,
 0.9478499557092075,0.00870869300848498,0.3679913613468784,0.9898545782308336,
 0.5462370025573206,0.3848806147281816,0.40811191810623104,
 0.18365609537191085,0.8900143917359986,0.2672161840472155,0.5677060853860024,
 0.535712850035916,0.11997301715967945,0.7209421971891535,0.3856860059476439,
 0.18175972201228685,0.20713120186487477,0.34946745383295774,
 0.7782404033600667,0.2566617330459955,0.2144268802634821,0.6433601850145976,
 0.9931295696468355,0.8167400839526952,0.36811323807502494,
 0.40168206058881895,0.6681454395450999,0.23285421792666794,
 0.9283182593743465,0.20427484050561806,0.02438393994304877,
 0.5310529682735852,0.1531485059648513,0.6503705700182859,0.8880896529221431,
 0.5437599364391147,0.30005263358291034,0.2331123275059871,0.9758191588789225,
 0.8795986040203667,0.5833371308107036,0.6348380311706966,0.8142309848848819,
 0.3731860595578044,0.5875396799275793,0.9805268229633542,0.04998939416272341,
 0.9183661726163729,0.9739408251091969,0.7462370264280215,0.29477060690693735,
 0.23522646291702387,0.3840227347996428,0.35654016680446066,
 0.10860056055932088,0.8124816972001024,0.4806516441749299,
 0.06215865298611256,0.3004671916955527,0.9936745380498625,0.7032644368725466,
 0.4544472284456712,0.37680241233535505,0.3631338473890139,
 0.46590108568645217,0.21312699681841218,0.5952319947714654,
 0.23880165965743716,0.9488975552314391,0.37594918793384813,
 0.8997879990404846,0.7962490694919344,0.515018735957965,0.354165477977027,
 0.47947918952119084,0.32646613011904113,0.3582127860105023,
 0.5172217632165224,0.5130697423130568,0.8724856408195164,0.9917629042951782,
 0.9847600397118161,0.3292453468377363,0.7046294511515361,0.5884144023493382,
 0.34252204720000756,0.028533889441351157,0.2502646235303445,
 0.4887560158455776,0.6888397941965807,0.04905917083644418,0.5705871087896845,
 0.2776450254023396,0.575747849636206,0.8667597703593267,0.329718169794735,
 0.20761690780627262,0.14005627911658225,0.38738664738250517,
 0.18249255720755475,0.5864483956443627,0.1612890312240356,0.4918859028648279,
 0.64665813733483,0.9813799340426689,0.05445699541882354,0.5185386469585301,
 0.035609785179234965,0.9059549302087598,0.06275342827650154,
 0.9654126266067062,0.41564914734088243,0.052210882979923934,
 0.7804253468021508,0.7329573043203552,0.12860736564921404,0.3865623477329483,
 0.679044909598066,0.06509089139747815,0.13523916325636676,0.9412300180396619,
 0.9619718157938506,0.07257192724801731,0.5968715371093235,
 0.46374438524128114,0.5318173279566051,0.8933367888262175,0.8847641616208956,
 0.1468746303311107,0.7479056123368509,0.7746978250061126,0.18094060537722634,
 0.6222287127872546,0.3410370412648557,0.5425352608612616,0.6986800526356598,
 0.6045463980281858,0.8470924355012129,0.4665706910262727,0.2395003394715659,
 0.3203061773656374,0.050178797123760166,0.09314386310323586,
 0.7124535001578284,0.5496573583148949,0.693702446125541,0.9577005546736426,
 0.5651926129299093,0.1071879053651752,0.08512635130834201,0.51558990229139,
 0.26744021118652384,0.13905923519282448,0.6217246909542569,0.964659090704214,
 0.2070422759537408,0.9170892030904698,0.18512689134734361,
 0.10710355219109091,0.1690397983270432,0.10228243036191276,
 0.33101550274767766,0.46066986855948433,0.4641821906344944,
 0.20688078916170982,0.8317819051324731,0.08460347648542765,0.608268950882175,
 0.7874824153226796,0.6526874737412908,0.811929602657009,0.594381873382151,
 0.03952714812053115,0.9229977878466924,0.3411871077357216,0.9454086571717967,
 0.25019430826691114,0.23775708944635965,0.567956603878748,0.5538485996791942,
 0.9934744606600808,0.1218353312254753,0.2807495265733301,0.5107840655620169,
 0.13141710757320957,0.5721896655727242,0.724875396532108,0.6452379204960023,
 0.21673119071019264,0.7345735078370678,0.7555042539911291,
 0.38767273396169666,0.005053418065351201,0.7163172120615126,
 0.36106758932468497,0.7262382216419433,0.4802525502470132,0.556830819671555,
 0.7984912089958052,0.7372896975550919,0.2782620115564618,0.9388720887136532,
 0.4613812008588014,0.9271378577763939,0.08227952597225519,
 0.01839830980429713,0.9310416922812287,0.6033617773691236,0.3839460561863225,
 0.8623824231545659,0.0026330560206370146,0.07272795452826053,
 0.9902038913393822,0.9305090154598997,0.7289767173077414,0.5992375955863745,
 0.04301648148662407,0.5229104341466386,0.675689232584956,
 0.057734308788092936,0.05910837055736917,0.8976381051094087,0.70991062121328,
 0.1347090762592852,0.14977039422796978,0.8101543499421424,0.702201464782609,
 0.8079618038970534,0.8722018701023329,0.9457505120776152,0.1521366858538471,
 0.982492197285871,0.29315825629785053,0.590211073531084,0.6087463687857197,
 0.8423928954667218,0.37432741474148,0.08248067972120954,0.6619942789714557,
 0.5483524859603645,0.13500312457982044,0.5782540739676639,0.432993659637822,
 0.23246646900226575,0.5039872027378696,0.6682580154185433,0.859653745817402,
 0.4752466187326797,0.43190660210349197,0.6722503928670733,
 0.13331643999813958,0.2887375617699286,0.10967007403912266,
 0.0017008300594586467,0.6985926554650526,0.5836153218477669,
 0.645077779736281,0.6211528248936005,0.08742283199613621,0.3731247384961328,
 0.7372831931954432,0.1765920052104648,0.4870545970592789,0.8594461103266543,
 0.9045180945762672,0.7458674831727782,0.8862628357989466,0.7920418281447749,
 0.3225364365191391,0.9863677153478514,0.36146999645539657,0.2869025595594553,
 0.052924929560170364,0.9936979374584531,0.5172873663719979,
 0.41434133215073643,0.45725656839338724,0.816161053471924,
 0.45799076808240324,0.857654807165014,0.7154789302462267,0.6815386292162408,
 0.4512737127966223,0.1467774692813335,0.02960528111365346,0.8818857283249547,
 0.1254478632351993,0.37521665086104616,0.4123842192907713,0.9989045489941282,
 0.31512949883873675,0.123022114970742,0.4407640957032063,0.6894711496206083,
 0.11954962876461184,0.15357148206264293,0.4834204061618228,
 0.36953555522701287,0.962207511559453,0.1865137198273462,0.7889751179364874,
 0.9845544384865421,0.4729406091007732,0.5257735270026836,0.7532859093378028,
 0.7011953671156799,0.4806710540945246,0.6890123559976329,0.4615560622809447,
 0.8198904072901444,0.17522032151870492,0.4818080915011562,0.7519441752781715,
 0.31739267296246765,0.6244421544150663,0.36333330176969614,
 0.48549526831445267,0.9824861136412575,0.9812070311548299,0.9022533488965848,
 0.45316291923524965,0.7904413957825782,0.03860721480829554,
 0.6856957387855674,0.4555131413576283,0.5286771729400928,0.6208781145395792,
 0.12278354231894717,0.2613004578362167,0.22105726398797,0.9129408853885443,
 0.7591357238455828,0.20350204936583083,0.2147660391612365,0.8563717781840323,
 0.34859981465222234,0.5796444432458421,0.1371215233357419,0.9095379223904173,
 0.7873698181255765,0.9399183989928206,0.142360970611224,0.6806843676712807,
 0.7100467614779646,0.7108686295211285,0.08367060427932183,
 0.15054854716322807,0.24366681284817537,0.9107924581916004,0.781880493302662,
 0.8063126676325366,0.29771124998442255,0.6774296949659019,0.8593683849821274,
 0.17301382398279697,0.5253801570729402,0.23850498297533507,
 0.37202358462598617,0.8111517266318193,0.2675655168898844,0.7924168530432362,
 0.2994011279698877,0.8508818562717146,0.2168562946185818,0.2408090242069829,
 0.5003666966207294,0.8273745497775467,0.9169566683354987,0.8667926348236277,
 0.30152705167999416,0.3271992026718955,0.18784951637104685,
 0.6555942541679639,0.689401474019413,0.7308778130235415,0.9066393003946169,
 0.603122678645884,0.5330067549697427,0.32124001231166177,0.8579095377328889,
 0.2974293667879766,0.5733110114539365,0.09397195071381315,0.4031273876493686,
 0.6716663879401024,0.06484390760149006,0.11571927863775122,
 0.40282462078012915,0.2615835038127772,0.46520003934154863,
 0.1227431019353189,0.16860939507846995,0.6825913215231836,0.6392585287261767,
 0.7774985671788504,0.7144064380138225,0.4266698023016059,0.6284352003913475,
 0.7755780339184171,0.5662074945115323,0.19067239163818084,0.2537854488468214,
 0.5675297829846866,0.5911365739896899,0.5545073326093696,0.3385512535273505,
 0.9341862648336394,0.9719622399526161,0.44371770136765887,
 0.05832224560963728,0.20905346390907975,0.5889606551919101,
 0.4227704692495692,0.37934003246408143,0.35680107564156505,
 0.5690469822498906,0.8485663520215005,0.7198406035462167,0.7954137542199722,
 0.7493939644674442,0.6425013441722781,0.820706023863072,0.007313730192177514,
 0.08787123483030457,0.32793782493339363,0.49533658949775805,
 0.2215471225124378,0.029559790018163,0.6998666399452638,0.9528113771428015,
 0.10554568023149347,0.8675721393881723,0.2723240195220551,0.4572982412850297,
 0.6805667009261184,0.33692980151727514,0.5726998741852696,0.9175260119923108,
 0.07112411850526767,0.9507765561822425,0.19409743826722015,0.859624728001388,
 0.7887456731519555,0.8954202233400492,0.9075148417682826,0.738811913645477,
 0.5935911328839919,0.3499127457212813,0.4232811300381315,0.41693856380410754,
 0.07544927756775222,0.6053014121915076,0.4238035520493584,
 0.15510386039689972,0.8696286530709822,0.07510354129395946,
 0.6283141810747469,0.53542327604955,0.3696267977188298,0.5239859610310633,
 0.8358019951638646,0.31014617831265257,0.7245383317012466,0.7566278819520237,
 0.9826310133032169,0.3449784149680031,0.6971659636645926,0.02666481055171277,
 0.10137097939444939,0.07043541811764253,0.5471431170484975,
 0.20665938694690966,0.9234675474940057,0.7993624878137356,0.6804062300726607,
 0.13365682813645963,0.05291680199784732,0.13039961338519557,
 0.3196037391222345,0.17065211545887082,0.9368081430889915,0.201796584328634,
 0.4846096161939495,0.8646316843019975,0.1510090635097212,0.753406262773623,
 0.8448912556329612,0.3798052962701939,0.12247572717142985,0.7646258231659753,
 0.7879407970797219,0.9782089925241148,0.6687193966859657,0.41737151906373315,
 0.6658084554295396,0.014157694876528404,0.4198638666215594,
 0.10779314674848706,0.6543532107194787,0.7896168631024544,0.9431183080949933,
 0.25367356094195337,0.2045233137577308,0.7532967358529277,0.7231310505003697,
 0.0034623689769734955,0.4918937656732899,0.27712103132369603,
 0.5044872229324785,0.8251793825884486,0.35316484163262296,0.8920511703540892,
 0.0065939616957438885,0.6096929329857345,0.7370774577428258,
 0.4033773400891325,0.6685163196569618,0.8777444234098024,0.6161612619475486,
 0.6778954509362141,0.7711594708451378,0.6723971441051453,0.8553152866852166,
 0.0712584627424665,0.3314059349974382,0.6558517428333305,0.5486199028331845,
 0.06534452990531348,0.7941055401655274,0.4121252692809696,0.8040643804094179,
 0.8145356473651245,0.8130160544522778,0.09895528846231638,0.9373180053181098,
 0.2833935926429887,0.4323948568900633,0.8466429071347027,0.48302913934076386,
 0.40816265083877534,0.9571479466516692,0.35996638736507347,
 0.8569866314089771,0.7137144487871876,0.15231477097033097,
 0.24177591095009476,0.5747702996457305,0.7285049876140912,
 0.20031885376372227,0.5963743543879046,0.3357863491335904,
 0.03805367289202255,0.7307617441212547,0.6346274767316736,0.6600543693217376,
 0.5731954977394236,0.6983920230578774,0.6316318081499559,0.22045923968128367,
 0.9752450400498583,0.4931678735338936,0.3619640731431585,0.30498620495473816,
 0.3296932903867227,0.07165364757279702,0.10720262684782389,
 0.19614620432449215,0.21996989291736058,0.1672579341379543,
 0.0571381702235898,0.12832104933266797,0.6885993022474117,0.4831719642842962,
 0.170219077492738,0.9468524223449443,0.3511619810156501,0.988219478554256,
 0.6440953067277059,0.17606521202035985,0.5179114073528857,0.8894751731581936,
 0.762925614782236,0.1786786488551494,0.438914116918816,0.7594450490290956,
 0.15587685890353276,0.7284651508852088,0.9661639328937154,
 0.32731108587635616,0.12085775247482711,0.31575612222454685,
 0.44375049396932853,0.10366024337925617,0.006012863634578314,
 0.040441640998804096,0.8117040613222747,0.3301909993537606,
 0.8919901475574837,0.04657123219036652,0.8214723312388046,0.3821177624212816,
 0.325148544585945,0.7680840168827702,0.6271803125333498,0.4851443235182936,
 0.19775075727045444,0.33768687820597687,0.5569969459045334,
 0.6998997829059415,0.7812396433010097,0.8327720384785862,0.9152774156081445,
 0.6383573702513213,0.4329818118733444,0.39331882347499814,0.9424922946675176,
 0.6832348485476365,0.018221886580569113,0.6798062855466156,
 0.6509080000929818,0.9721196496525447,0.7246356652963116,0.17086913507461654,
 0.9321823654929094,0.6840995123353646,0.9855101813734058,0.18111352884749699,
 0.3438512790351015,0.1966069920821335,0.8942019249849026,0.795767723035057,
 0.2555524719422715,0.48082221379111556,0.9257124746115766,
 0.21569502629528847,0.28939354932397754,0.1775055483905309,
 0.032700682853690655,0.48688639031257586,0.5327369449330988,
 0.9458246000064668,0.5491977499862314,0.3931973711169896,0.4536607159026005,
 0.26848301842471656,0.5347373033792687,0.5781173951657013,0.6713467369387563,
 0.04357659329448993,0.4512688273219516,0.22557864917902093,
 0.19120692195021172,0.8980935058730357,0.18478744409731696,
 0.3186438952073214,0.08353299018608373,0.5188069360417175,0.5231435440575084,
 0.8749050259265799,0.8323153366457392,0.04846649678501924,0.6660091552311649,
 0.6483007942388617,0.404204375934872,0.5058247490721923,0.03410616803497091,
 0.6011221184006725,0.8857385383474594,0.9247470437334598,0.06347836382045613,
 0.7344946233065325,0.7944520120348294,0.07294514097839344,0.8316307265948164,
 0.8108504409160722,0.4998280229457642,0.1033922066272942,0.7286946317043264,
 0.30029805040978774,0.17371726129004283,0.3433503519544897,0.801860863912729,
 0.3463824757578645,0.9643349963983352,0.16329817677646397,
 0.08163352791070133,0.4113965816338885,0.8902297151697695,
 0.41838933006340884,0.745754069806599,0.565537086648932,0.03205108149808722,
 0.7211188700624434,0.20603010456848536,0.30434772514361663,
 0.19530888592657625,0.7285952566255175,0.6681722448168301,0.9174295966443735,
 0.02248527737253536,0.4089565515596202,0.07066722127052383,0.324084144469743,
 0.953173641491389,0.9017500054138855,0.06102240778098189,0.32427797150962767,
 0.6531255525067672,0.7453604364336569,0.1933229354108178,0.7607675857006743,
 0.6476306608035939,0.17477343072566154,0.4294788770463356,0.1463216222430581,
 0.7909990121190323,0.45946626037129923,0.42656500411263687,
 0.2962188842179774,0.9819701149531134,0.7694859274543357,0.6613960674669778,
 0.7219963427740742,0.2964227129349345,0.67774217777789,0.5365697627918967,
 0.04580895013661024,0.9821124376511919,0.07666101661157532,
 0.4367193978689745,0.42390441574693494,0.8704313065468274,
 0.47044363191562966,0.7104695839912477,0.9904941323889866,0.5877005848410396,
 0.578347768007834,0.9422767988817884,0.10770555782382663,0.3049934111553926,
 0.23818236104926926,0.6042747119363521,0.26295130815917966,
 0.15860194491999846,0.9401766467426751,0.15481067444343333,
 0.3173987329892435,0.8766878950645112,0.8910974960282383,0.07407016620386297,
 0.07993835994474263,0.3034406351801797,0.4115972065720749,0.7786871681211281,
 0.050055521279210424,0.8312947305806576,0.48986789254541163,
 0.7056830294621225,0.0906954466019223,0.45570199615668994,0.544933478526601,
 0.3144660434034696,0.3981051099795896,0.6960737357400903,0.24224060220761945,
 0.24285095920380584,0.2229264326150926,0.9105253658986672,0.4633143000643867,
 0.7745904813119329,0.25282677932269626,0.27238199050645895,0.323135161457905,
 0.5074810339725047,0.25157613796984546,0.8605424012073429,0.4535297171820555,
 0.43555726017447494,0.3902958094544896,0.9775026230076486,0.2008520209998904,
 0.3272940300356084,0.31331032114097457,0.8153123061559377,0.2599219097718457,
 0.5622818779227102,0.4731642452878265,0.9960791426979972,0.8754554658762939,
 0.5027505311260372,0.7046755013551362,0.5496172092835439,0.9163460693520948,
 0.36059402727358814,0.9766929643210748,0.3440652439336911,0.5114390969600233,
 0.1284833121222999,0.3955725552514737,0.8400885222558188,0.8799957319917354,
 0.5761980758363738,0.1061657663562331,0.7541713615893606,0.11005368952273453,
 0.8417710796189514,0.7132142328718857,0.045204658781678786,
 0.5833027836313993,0.12742187359368518,0.6225344768838018,
 0.32066215274993337,0.2824729957775214,0.31934658640096536,
 0.00448930188738994,0.2837802049962156,0.6120516649725003,
 0.13569628280467672,0.05722845318025882,0.7214507835832172,
 0.01051467779887838,0.517257134031243,0.05015387038758279,0.5484257784803241,
 0.024281519463835277,0.9940129791045805,0.8701597572930032,
 0.7135104765517388,0.6220128369728665,0.9859702511580591,0.14881540676393212,
 0.2796560281329801,0.5748139125707662,0.8113329291683999,0.8801536222492782,
 0.06229393141456485,0.5752452339966976,0.7943251576411418,
 0.16673997420218623,0.044428125611300384,0.6788150191613261,
 0.5472344076597273,0.14726153681781562,0.9484161660256442,0.5687418607362746,
 0.3124783392062187,0.44813944179829046,0.19398390087619422,
 0.9795869953297555,0.5345653823063918,0.9966070825352298,0.9735525231257683,
 0.2917942602052074,0.9078989463023074,0.36344347580496195,0.8114321007833439,
 0.823469214422684,0.05226422672375797,0.01596375969833197,
 0.14289544415699929,0.8562513450770058,0.39529051848683794,
 0.9709686933001404,0.7843558350462783,0.8620428415747989,0.09216439602068205,
 0.38336523729379,0.1433593157888441,0.8610956616992931,0.12227602401688165,
 0.6139135430262088,0.6627753374615434,0.654451713974967,0.23909364133172883,
 0.16429928121238202,0.3973535442136146,0.3713577161340519,
 0.14509736364520287,0.554139522211994,0.4463941601600858,0.7141458283579685,
 0.9669897909276927,0.995548518509337,0.42430385462129416,0.9285981824102387,
 0.5181364916597109,0.40542950321675975,0.5361284396927279,
 0.44604996506768313,0.6446913789889948,0.3743796335930928,0.6725235766934619,
 0.5095902363773505,0.7460963785706398,0.358015593025682,0.3633991084028546,
 0.8717318778636554,0.36750426393661206,0.18719886193702573,
 0.9964153138921928,0.5187381437783967,0.41057981302707647,0.5530532822609214,
 0.4555593331588267,0.013312700413803502,0.961465235489914,0.5813943504046222,
 0.6737374990610807,0.6745850076188068,0.650533836613525,0.6119451786505122,
 0.3204178872119947,0.7046457929330162,0.39560907783069943,0.7649391251133232,
 0.8755453246774949,0.1916531574564202,0.2507181562978473,0.11638374463691747,
 0.09368437082591075,0.9941609057944205,0.9904694146242049,0.5023597481195302,
 0.019988394268052767,0.47715640111520297,0.43399923311504596,
 0.19487698552269583,0.9855919658820715,0.8244395813977311,0.9915335222664182,
 0.6105366964226587,0.07768192049201161,0.8438790562760379,0.1521094693561733,
 0.892241168769931,0.5222228118410212,0.12293194449818134,0.22998029412686838,
 0.6699226592944937,0.3754891417448627,0.1910998938606333,0.5309213139195776,
 0.49199403165004874,0.8948355118211891,0.5934122537938126,0.9717209195863835,
 0.2837685392538387,0.6445138502825934,0.9693738271056194,
 0.021537257333059112,0.31629667680775775,0.69506296438621,0.8491941931603055,
 0.4524770865147314,0.17127349936167013,0.44117517379050397,
 0.9284495026550911,0.3004990081599678,0.5101595755236668,0.348006440475111,
 0.08776284987961169,0.1414250019756761,0.18258463775768163,
 0.8528580569661189,0.9607960007871594,0.05686950447053907,
 0.05757124429390725,0.38537138798201853,0.8261205084779828,
 0.02459105228554448,0.4606813490974081,0.6645000428500986,0.5547027264817269,
 0.1799443552976151,0.4049053847049424,0.8831502446085926,0.5089502529604039,
 0.8964102205857337,0.055436148344714,0.2506617275100599,0.13308202542694159,
 0.6651985753076581,0.05096785068023979,0.50088645224047,0.5016503012320062,
 0.8102055386593934,0.1512225220429707,0.618546306057741,0.4510764731123975,
 0.5393112494777638,0.6194010081810936,0.0978014664155431,0.9210372842900101,
 0.4717202055918358,0.24144463054452114,0.3572158026404655,0.4372499279670914,
 0.5323171650359131,0.7355186294272574,0.6246513762565825,0.7133253782948548,
 0.6430571587215113,0.5541476816007027,0.5435534313377955,0.8460317122847832,
 0.08930269539522406,0.8163083275413368,0.5229946196393214,0.9372548992598091,
 0.43804206999722295,0.7047405377914939,0.056211293286444364,
 0.3024768324687849,0.4025132862995924,0.667203078972145,0.5802619202970514,
 0.535142158195933,0.2361169480035229,0.28184882179997306,0.7154204607393768,
 0.9204647040297633,0.9149158322630696,0.6921539294324236,0.08572644025238274,
 0.1476198733557066,0.6442700431923817,0.13656855662861878,0.3828135302611687,
 0.49305807744083485,0.9228965972788121,0.9606699586019437,
 0.06371604956224164,0.5436450729110126,0.8546839882174513,
 0.21278752220393737,0.4141258256087378,0.8857889783603865,
 0.35240479571345484,0.4674695185515738,0.5539928335938802,0.8551182194015211,
 0.7875587031860787,0.5736577599995577,0.535253565897831,0.3235990759854178,
 0.8218474652994949,0.8228580880751079,0.4512044072434548,
 0.041442268782056146,0.48145822016325923,0.08790503356563262,
 0.1409740854853987,0.014549305272075985,0.03592929129254374,
 0.06360217929478451,0.4485620026809639,0.2607183651100128,0.2708236672454415,
 0.9551354566011411,0.9960749590087814,0.6975966926105848,0.6756646924886592,
 0.08543095664334222,0.6497005855839862,0.2870802216349819,
 0.08439153981901182,0.8916307947725178,0.45941367938345434,
 0.8699815667109403,0.7735370255393164,0.13663358135767345,
 0.42426203938585894,0.4533350100628545,0.2546421265219939,
 0.17893288940075858,0.6202102812302966,0.5323997399834781,
 0.020574733388386002,0.41676200329944346,0.14165868052624675,
 0.021762280914386034,0.24088019174694975,0.7741037286114361,
 0.14482885906189524,0.9544163352973123,0.22086602995258275,
 0.4724006605249068,0.31301805238907543,0.6684929398183013,0.5441874616607291,
 0.9707688345961892,0.42630434909588566,0.27690557168880536,0.402848546125366,
 0.6943162217661191,0.15053886396826344,0.30360261274906375,
 0.9263138544183925,0.9592615620878712,0.674167682244275,0.16509378491639104,
 0.7030020585596665,0.9699193376380415,0.6607567235217818,0.1490716967403365,
 0.42371062687626915,0.3130097558399467,0.1046691604075417,
 0.15811177109690255,0.6678092320932216,0.1867979923265647,0.237532711556532,
 0.7862651103329856,0.14486361598423625,0.5061076262973228,
 0.46243466184525284,0.6421359136739295,0.4191124706187519,
 0.12657928737638013,0.8543791082735763,0.28255142065785477,
 0.6443095322885062,0.47850519240863254,0.1758681759709858,0.7789047938383826,
 0.8495124117369666,0.21823157618379108,0.5735157086593579,0.899427807220086,
 0.19331226727766604,0.3905324589663274,0.7115742283182822,0.7346313042965671,
 0.15903429821636195,0.23634616220679927,0.14176508025677836,
 0.7448532160498055,0.5897305030915997,0.5678216550651617,0.06809155712395698,
 0.287370597114291,0.7299036141874221,0.21342540328121973,0.3951295590443086,
 0.8608387173779166,0.5916225343620904,0.4782582154764865,0.8850461865467172,
 0.1049600977507299,0.3378299734829855,0.7322633767812128,0.03490596607620744,
 0.27644064620508846,0.007387800200433503,0.17821781643606427,
 0.35811398347459267,0.1518635704528688,0.4479740302213029,
 0.21459346705473226,0.6677344949541518,0.14979798762318353,0.979865122350723,
 0.770027329562497,0.10196895859066601,0.10506027426362108,0.4130342594285017,
 0.6368471201710538,0.8954216877828598,0.37618913760238837,0.8108717114596139,
 0.24411993932080522,0.9216555685854102,0.8260685195102275,
 0.05532548135029702,0.8879035915740383,0.7996689290292607,0.937788809799031,
 0.6181636848035804,0.3560236036692954,0.633082624919349,0.27079847678370095,
 0.34916992736603714,0.9870844445752286,0.2989217088815066,0.8628784222912873,
 0.720912540796915,0.39066683967469473,0.7879275848739411,0.4071380237124482,
 0.04657614362113027,0.8526099794233608,0.8644420165696199,0.5676976683350192,
 0.4824164996758269,0.28350065568088656,0.40399488777145065,
 0.3575352783440774,0.07239607889176014,0.5951555310840138,0.3947171578696653,
 0.18331755676013006,0.9512272001759546,0.6234464069252261,0.5096686371201529,
 0.5276510199290684,0.8032942265413032,0.002180112487704111,
 0.17581716774975964,0.07763809548922962,0.06696932007689227,
 0.8235049890918654,0.713191091060072,0.2118300714661887,0.031878024916100256,
 0.37535524361192174,0.2097634253490004,0.05077542711904115,
 0.8269627962426088,0.5485539733111833,0.6142547089909983,0.80313239783095,
 0.6146560435373056,0.9849824546918631,0.2543176143624326,0.7870985624799983,
 0.5306290264806097,0.02867781484788723,0.4147604174470976,
 0.21987793840454717,0.44943281273657243,0.25776166097585695,
 0.6995718916133528,0.7488922522895807,0.39467153869969973,0.7635644279867141,
 0.5472411504322656,0.32458997698982184,0.3977390820330685,
 0.10729215977296813,0.007181024934728275,0.13840320194419253,
 0.5465840604271696,0.3830106009607238,0.8515190194620438,0.7589278075400859,
 0.26057807449584514,0.3349829460848869,0.009081625958462625,
 0.04971190059963859,0.5119898408045602,0.9956121538256953,0.8660559759867953,
 0.974665459626396,0.24850641775494653,0.8576335867108736,0.37079632151572317,
 0.005763767646193352,0.5975466605967458,0.8529374715818685,
 0.43538515983671056,0.02762166308721481,0.30448914090871393,
 0.9061663801752836,0.27264525445399856,0.8509182435992506,0.7877715342774134,
 0.05086997281105099,0.6721873386305278,0.7577901589333113,0.7483711868142724,
 0.1437184613048893,0.7870172494397968,0.4887561904370795,0.9313694723607526,
 0.03219351709256579,0.8985236506950037,0.9747654171856373,0.6501470967233367,
 0.31329483759888155,0.33020523248891687,0.22681434862181749,
 0.06521925570131426,0.3891917511954236,0.8397029320218466,0.5493364246913404,
 0.602929194271924,0.9448196901342429,0.7862151456720788,0.27659958952967745,
 0.10106863877202854,0.874645399224792,0.04503000217683728,
 0.18872505594643252,0.39804512143182225,0.013045810334255448,
 0.307484930844816,0.2907769479362229,0.31826411000595056,0.7131655687303445,
 0.41530488602373083,0.1336146982302966,0.9629643724325208,0.9942487900583767,
 0.35958919816097734,0.24946438983132968,0.8152758331469129,
 0.04576083146033427,0.04306457697957189,0.6448908136944936,
 0.16821732793374566,0.22923419022295788,0.04671987260596466,
 0.19628335211500858,0.13800336236084287,0.9774528706594134,
 0.4890090230188363,0.43522820126092865,0.8163091134066663,0.7272907017785029,
 0.15806345989338288,0.6703456332630173,0.49894959073849887,
 0.7421703052634795,0.6996479289735564,0.7060446765877743,0.7914003774867062,
 0.018692055460090784,0.22155906390713864,0.05539273709730097,
 0.7534751678202192,0.01117585860598802,0.6787206750297815,
 0.16547828248251406,0.2784068673416652,0.5050937114632329,0.9289614809267801,
 0.5986431003551632,0.7221945829970537,0.9166337340592762,0.20918574575089066,
 0.9534658658450164,0.4040307066374429,0.6116893026677744,0.1036122431222758,
 0.26580849989032973,0.5079526242466804,0.8668056987583488,
 0.40794931953805746,0.5915169582502305,0.2190360098835461,
 0.25621551604087744,0.8374428776509708,0.9576469382939472,0.1551790121342701,
 0.788024702581557,0.7290407976446798,0.09273049089981933,0.48194434448095436,
 0.217383515728243,0.1551141062123811,0.9789094531138581,0.3068481663472439,
 0.05554964137022633,0.9097495065060369,0.11963544132020099,
 0.4963048956143806,0.42187970932678176,0.011860948627151369,
 0.14284432339648434,0.9836131775949778,0.11280263518355527,
 0.4721433428595875,0.2233321984055494,0.6796459029133788,0.43135652324205376,
 0.3734622030322089,0.4277689069384003,0.5783299041193204,0.7282973471665268,
 0.4101967085945748,0.28051662382638176,0.018992409038694036,
 0.21023017327671856,0.8463614916503618,0.36331270776183944,
 0.7532824316385409,0.2836777745054502,0.7955379831940105,0.5513982869721912,
 0.7558157743171163,0.3724594930334403,0.08790042496384909,0.685956141117483,
 0.3424329529404191,0.8324754607293517,0.8485735500838668,0.11330030655012335,
 0.2907907836651853,0.9528279104992697,0.31263282871226306,0.556149246017131,
 0.5575709873756457,0.8216116125238065,0.5304074462629947,0.9327549728932004,
 0.9949859411304531,0.8068201156261685,0.4188371388937514,0.7807918984680287,
 0.05915163838954551,0.3939057421664983,0.3854242303321387,
 0.17895759358724472,0.2691540352910754,0.5176587813492926,0.2712696767239974,
 0.5903261643854543,0.15535983689302635,0.21052925518415344,
 0.19282172476204684,0.2960563194581698,0.5931509969562349,0.094728101997956,
 0.2274842939715418,0.5163244812396661,0.08864696967768093,0.7581082219714886,
 0.9382363879375106,0.7438006598698697,0.7306677542871873,0.7992660153048048,
 0.5560606047062936,0.9324652930662061,0.38671143463104907,0.6351589772130917,
 0.08186349239944701,0.878208588474022,0.49010331749376135,0.7779426655444295,
 0.20097625083739434,0.08488756629609062,0.03466129843030119,
 0.2952044633902524,0.7104088020574021,0.7323608892935897,0.31757327159477977,
 0.5997914875422384,0.8713925313360497,0.025895593122859384,
 0.6015284385620531,0.6191738674293175,0.05971840601795175,
 0.09098898645217413,0.6547955501529784,0.8699398507507227,0.2472455768632592,
 0.23765660776848574,0.9222311153629967,0.803929864926277,0.5260420039534989,
 0.32839431855168344,0.75105805508298,0.8948834588149965,0.3443442311439482,
 0.5899383117725897,0.16650477534413477,0.4806226126288632,0.7111366220041154,
 0.7790087842208249,0.2835669966681269,0.8590652824249151,
 0.014601721694630254,0.1703242547607544,0.2958170182187788,
 0.07393023989037184,0.657251484836715,0.36699049086296376,
 0.37890166213788523,0.3386510407627342,0.17655171509594592,
 0.20604395170196543,0.5542402433269029,0.14823562464127416,
 0.4278489850643865,0.011723566578618394,0.5514515725422062,
 0.24747651047287145,0.984466965775461,0.6932489247398316,0.986373933285468,
 0.8850079288980118,0.9670831633337275,0.4748605668392393,0.1267535359491173,
 0.7667111888807913,0.592185237505256,0.15646947669855182,0.9968175153615824,
 0.32269755245546494,0.663003091612374,0.5403094365882101,0.07152347530705017,
 0.5606627408011094,0.7244622111936699,0.6096541735895243,0.626499250102762,
 0.2281554468265543,0.9584961429719,0.8212447121150379,0.08803914843865279,
 0.8370015548736116,0.9192921899710564,0.14985519761501476,
 0.41420371559776137,0.7641656021478034,0.9981451962549159,
 0.24184564501382144,0.9302826036234442,0.2587009650051064,0.1780406411551959,
 0.4238142470924017,0.09242470868519037,0.9250423338515209,0.5415688122948967,
 0.7631134577614128,0.7992835903370511,0.5089841854058059,0.9802241312329506,
 0.36546128565897007,0.026887331538125947,0.9435804768996436,
 0.09689562883224634,0.058452785722356904,0.7556008907348151,
 0.6764308353436289,0.20465574059047265,0.5260835255028165,0.8223974880055568,
 0.16988390790056518,0.42639885957546253,0.46297947913561754,
 0.9238279030846582,0.2947929281723416,0.27015779453659294,
 0.48746354256731617,0.14196209330461418,0.1571081443256328,
 0.014731400288087837,0.07315505653509058,0.026495608990663833,
 0.6327298316858583,0.0768617684542604,0.020357433266695035,
 0.9468178570148926,0.8840201573338906,0.5175650788679139,0.38092964505050664,
 0.04114155625283966,0.831603442100113,0.24294991231731788,0.0694892728290799,
 0.19508908287492643,0.2677815303704777,0.2689778950494709,
 0.053144697196173096,0.22849676386160422,0.10798198263827319,
 0.16028369468845205,0.009543574407501532,0.025036749442013506,
 0.2946144616162123,0.8875634517184411,0.2533741149621245,0.537062981883767,
 0.5746095899267591,0.8789490655493334,0.27795897236919154,
 0.29306662822257845,0.34585951819013216,0.1004284379845739,
 0.3506561775250985,0.17333576518122484,0.7246307324510172,0.3061280669371138,
 0.35953779062911395,0.2104000686593276,0.14437900381264768,
 0.7110806884661703,0.7607776235412682,0.9532267805923664,
 0.022524591615077028,0.3974574758260301,0.7594950654737378,
 0.5364721702631271,0.42593739157501875,0.5731004255138455,0.7332941182907818,
 0.4549995624997131,0.30737777576011593,0.2445955571557471,0.9439966087956284,
 0.2766884559055842,0.4924800925593751,0.1597751211869577,0.1184229703435753,
 0.391579005004591,0.17781131261258776,0.22923924425386577,0.9795482874937894,
 0.17981469239972903,0.2210958730965833,0.3674268401825712,0.9664181272146951,
 0.21429973214202924,0.6502112401354839,0.6684766888087006,0.2772856855235014,
 0.3488776066918119,0.167498032914702,0.01379660622346901,0.8706611890543385,
 0.4575907090477347,0.11573237988039553,0.9235563997518718,0.6856875438147274,
 0.8426322232723192,0.34390871307113025,0.6729899380397495,
 0.16771321112786897,0.7868331410973852,0.9271640746587734,0.675641564322254,
 0.27872148820732756,0.6838163943089666,0.38283744318155377,
 0.8330519030815515,0.8759372478185447,0.054681603798647505,
 0.06978114339176322,0.020326550100067564,0.13348279605158797,
 0.06741024505565996,0.19526026740254354,0.22045009803578408,
 0.33946493809101286,0.6301881901158812,0.16332881682261313,
 0.6884429026351873,0.37567203704851837,0.7111123547463184,0.9035127403034701,
 0.5156768148856254,0.9935138342983955,0.22512821212339784,
 0.41698373110555786,0.793511468449928,0.37168897155975067,
 0.31723224018466767,0.7171429096679643,0.9102245951181449,0.2908184690036094,
 0.4305275808916328,0.9056788800193913,0.6957866158192758,0.5752323738257462,
 0.5824264844147109,0.2890585761296469,0.5755161679734568,
 0.055517677406544674,0.03142077423211165,0.9472303495433745,
 0.1968384398253763,0.2936123279764947,0.6523295425809765,
 0.052513634855960833,0.2686223671697212,0.09828365907857817,
 0.8032490306501293,0.24395604591461395,0.8074036034377216,
 0.24949629425313558,0.9727865192086723,0.5840080949262029,
 0.28926166861869085,0.91399138437472,0.7564809820486873,0.9946861611742094,
 0.7551003850506601,0.13723711117231618,0.9573879307959284,0.1190213550752024,
 0.04270957150658994,0.7494688167204229,0.5375345536565337,0.6559667773121296,
 0.34548455595894567,0.4213635120127329,0.3095458526271414,
 0.39035470154895024,0.07346697468646157,0.9781089797672292,
 0.7427140540388871,0.191724116580808,0.7779963759873915,0.6920472529244214,
 0.9058616576202854,0.12877107054084314,0.5316551827117117,0.7753404274962078,
 0.35813615054291126,0.8599964424193028,0.8608703669186251,
 0.39200087418752816,0.6497267572686999,0.6177474823219623,0.4890033251925465,
 0.2515609726922512,0.515968272042622,0.708225106431241,0.44495044413063956,
 0.7505728782608678,0.37813946294486256,0.7469665572794961,
 0.24005811370865338,0.8859487012638404,0.6800954241546875,0.5262141256530499,
 0.19849078640505935,0.34872362157369374,0.07249620353979369,
 0.547434643949798,0.05463653577354621,0.9665312842817814,0.5537269346741778,
 0.19100724495698151,0.28899642997035513,0.0338939643801458,
 0.7193178364980822,0.5144860766146471,0.34235832735210725,0.6610696959437047,
 0.7568693667315829,0.0696405464379487,0.5818634965262449,0.47134464513045526,
 0.320984231053952,0.5322831842183178,0.20911269223947448,0.16743041991467755,
 0.9493463074762318,0.18553043333661423,0.8068196288063928,
 0.34112846668462316,0.21310165434257655,0.9464530415036791,
 0.29049962409596475,0.6206593392970097,0.23971100643693632,
 0.4424365408178422,0.01844489140282335,0.6480819077005401,0.9091251204985517,
 0.9824831074634739,0.5508083415107333,0.8673946299352475,0.9301738879138675,
 0.10044687283998699,0.9662032220600353,0.21937114002276448,
 0.07340921867276551,0.4512874357531631,0.3033039772038688,0.5944886524920435,
 0.029817108316044227,0.3756181743533832,0.22385480161281945,
 0.349891132387255,0.16031498567313462,0.24017429282374103,0.588938571976819,
 0.9763415673553043,0.42862530558880474,0.9799978040232483,
 0.35082030710276224,0.9313271226244778,0.7363207664922599,0.4361772255810661,
 0.3845418780016021,0.14853232490437673,0.6416373707677373,0.8925774272140369,
 0.6743548421565456,0.9884555408429375,0.6235602500051769,0.4952664381151407,
 0.04528104002736455,0.3374197332136174,0.6622660672671041,0.5130879832881132,
 0.10885259887486609,0.736149619060494,0.7706142173470913,0.8236930378475015,
 0.2505241849748312,0.7665832477906893,0.2437726459575562,0.6316262291323225,
 0.45913211298389167,0.005699436841279715,0.6504073091938739,
 0.17821431273424726,0.3480966704402957,0.9014942637839461,0.5875503529041387,
 0.08970808145966869,0.7465980773213308,0.24986814904360288,
 0.5771308418861085,0.8980392005588188,0.8009385626757357,0.3904248642227325,
 0.6936089871073376,0.05060301270784506,0.038367929061821515,
 0.3437811383300947,0.2773046559561193,0.9031993599627178,0.8385337643008524,
 0.45188889121618825,0.6675327160874228,0.46759945246484924,
 0.2208981270075111,0.4750733416356684,0.5218350560701405,0.16761167971217705,
 0.47548101650613206,0.21862831539908612,0.7591666243689008,
 0.8501643535832436,0.8066461401799007,0.9560520761628563,0.24494093725214383,
 0.7866666672879271,0.3388021078253981,0.8809923215570329,0.4346986514271002,
 0.9874535378519641,0.3505540076598396,0.5323987978203008,0.7976593526191326,
 0.6577980407831708,0.3097770421545578,0.7895615379453927,0.45257614317659567,
 0.9930484079532529,0.6106382438957206,0.297154917825872,0.387210217086722,
 0.8512233183558984,0.6734069917632298,0.31365350595669395,0.4995132116189771,
 0.09032806479238986,0.8749060707431757,0.08650738528611868,
 0.7457584496111439,0.30035947942942454,0.7833764085501732,0.1463822722891076,
 0.49674834369696796,0.8354398039552486,0.7149758858006239,0.753815594603437,
 0.2739484917089017,0.025945325804738628,0.6589678409724655,
 0.25052111502137375,0.503451149285099,0.40142548412179524,0.8513299407735528,
 0.40493867623901025,0.4990280484682479,0.3051697339522881,0.3132655424970836,
 0.5068421552658613,0.23067214819403714,0.19749236284494098,
 0.15166607354759054,0.21522189780937984,0.1296830467845288,
 0.7043589803410399,0.521946298049326,0.017336319808155642,0.519046780607133,
 0.3682522816792808,0.8964702494842225,0.8927785972198248,0.7161875912023546,
 0.7216700076288985,0.03801212545176158,0.9678509878246228,
 0.12252425530298683,0.6104712199861348,0.10817979670513145,
 0.8255010292533308,0.689219804673978,0.9387850613000575,0.2881206914591443,
 0.3484610639463275,0.4234881101737331,0.7788971933022658,0.4764583684548883,
 0.14835838661685508,0.16419100632913075,0.5687424138964,0.5085596648764314,
 0.8190351088018343,0.32001796202577415,0.06405578838911752,
 0.17822781292869339,0.8907863500776722,0.6733168676830187,
 0.20375830616970625,0.47305738596523716,0.9390912249360347,
 0.7695985447493989,0.8677582730371185,0.23512274939106226,
 0.14014391038330842,0.45891451609986045,0.5269757339789605,
 0.9983329135409282,0.4056094822882046,0.37182197702674935,0.8377716549745566,
 0.10949569920948665,0.6447852237644147,0.8429655434834202,0.7872093032130869,
 0.5012305159637641,0.39811927824376125,0.27522744592804516,0.913661555407445,
 0.4874428590413673,0.7034754757035261,0.9758223680909662,0.6074057371959962,
 0.6736518772073087,0.7343265256245982,0.6630711939787688,0.13121458527395358,
 0.6980468940525039,0.33959358382609117,0.7681866501079028,0.9027552912655508,
 0.9241428162510728,0.7867787324768958,0.6957671089022699,0.5613894801842536,
 0.8781545556911108,0.05377299651918188,0.9147627174482459,0.7389610758780889,
 0.6658563843290269,0.9229865017114741,0.6383298437873873,0.9951426268222824,
 0.03857435190277814,0.800927252541505,0.2704598959506189,0.8595209113953556,
 0.2159906562440599,0.4871871582401447,0.3551700832805288,0.22218451777706982,
 0.9654792369224918,0.6662658755317623,0.817240759579445,0.3771738637942057,
 0.5694004857514985,0.655783526740548,0.9058194017610001,0.3855907594239054,
 0.17670750475115804,0.6090232117803104,0.31028665295809177,
 0.06484683613981745,0.5086760843032854,0.7357832096554111,
 0.46627556259060576,0.4682829877276472,0.06224418976248225,
 0.5426595322578844,0.8931854488157889,0.7907813909611396,0.9737698749245188,
 0.4283891234937085,0.8426689769193454,0.9817261214021341,0.36489486362721124,
 0.5231609736687638,0.6842655949496145,0.3612170728834494,0.8962222804636999,
 0.3905741498433827,0.8106403896947051,0.608589428812867,0.28432406227227525,
 0.7392806903414073,0.012908474825899963,0.9744643725519704,
 0.44015277805486175,0.30236493188237135,0.0585415567633486,
 0.8692051153970566,0.9628147404050007,0.3411573271095091,0.9568683970151545,
 0.42240683340418694,0.9721842847027132,0.44895261595936753,
 0.8518378399548339,0.7270460944302126,0.8343789230454308,0.20250549749919744,
 0.4467722847152702,0.7968409716240865,0.3657534436638993,0.7980651685799732,
 0.8709037822494832,0.9744957663282167,0.7160055731248999,0.9660841450989569,
 0.46517640633604096,0.531111480086081,0.2301083502418677,0.9495189840205606,
 0.4738171089350893,0.996858624223583,0.477224794468205,0.17675593176818905,
 0.8882966858997408,0.1553611690980924,0.7248759505512117,0.9698860810215334,
 0.6227061110878961,0.6510526802843813,0.42228475305076496,0.9124889471144703,
 0.32944361789285725,0.561187849743094,0.9984924193359896,0.4512544380324848,
 0.3851863939180262,0.14878921732356887,0.37515335972343533,
 0.4498004217017757,0.36091998342124176,0.7748070084171392,0.3353309574716117,
 0.4605281492372829,0.4826052984380982,0.5768097059409012,0.8071429613045557,
 0.5396870220787721,0.03582558029164107,0.768519725435408,
 0.025146133659697867,0.1954613282600116,0.921187836909001,0.9972736992731285,
 0.25620764519821715,0.38328920656378607,0.2482203808770277,
 0.7560270427995612,0.6734149416144569,0.9954333030825413,0.9642998583515608,
 0.31841131425747116,0.03158371260597448,0.28740822929358123,
 0.6237776580465544,0.7728951426280893,0.3908842496169366,0.8902702607172819,
 0.9215930950721869,0.14173256058210326,0.7548248915422828,
 0.17653087563445446,0.4107446481341912,0.09937450087489808,
 0.9673187493973759,0.5194969991475591,0.4030263730644207,0.9139706313651448,
 0.2599562522124321,0.12666301879055974,0.1502447565528221,0.7749849854275123,
 0.3347623238028161,0.7685184166777982,0.035518414916674734,
 0.1921651941058462,0.4172614348340279,0.44901192003336,0.8952794177421753,
 0.8844768824405369,0.3220841638640549,0.6818704889242061,0.06220251228460216,
 0.8805299172147126,0.5560174511080452,0.4515079920637386,0.11078931416907156,
 0.6858469011384251,0.620562786153132,0.9414604823200008,0.4104477921096328,
 0.6002633504018499,6.157015114306441e-4,0.8442822671546152,
 0.23308939895127234,0.6813995466866335,0.2144081750538469,0.0795261831577403,
 0.33328588803113923,0.15808185563205468,0.48565474911979,0.5629701004837627,
 0.2682018631063109,0.10901530256375425,0.8273546120836464,
 0.01857843023423411,0.6765175972471207,0.803018481165253,0.9445453032805449,
 0.9937900660060014,0.9759205539262547,0.2282674113907609,0.9807649599491824,
 0.27293975283312943,0.1209075578351414,0.9917174006338931,0.5275319659296605,
 0.22425688438709934,0.6982907540947779,0.24055771572025042,
 0.6788703017535789,0.9678936993335038,0.7070891864353888,0.21205863379133083,
 0.2749424658200108,0.20851253811698311,0.8196140693486984,
 0.08038326598659473,0.2704817718673961,0.36579712404969655,
 0.18881081007674894,0.5665749190591457,0.6575535632460936,0.4619714038174262,
 0.0011792018905680823,0.4003409115717691,0.11066058806075296,
 0.8608112928197446,0.5060023205159258,0.27881723994303376,0.8043275591506834,
 0.14928742757495572,0.325116531952651,0.07712104538945463,0.5449576443063573,
 0.026528031847895095,0.9742928415895848,0.6030721512644803,
 0.03161356810777105,0.6739546229083234,0.7582430635038908,0.8715010965033517,
 0.43981121703836656,0.8121482687617586,0.5047308171327435,0.3263761061899919,
 0.7798251838157543,0.3079778874875735,0.8524521803438542,0.13562690830592805,
 0.7817803879952593,0.7609656066086357,0.7015833487432926,0.9138162965510361,
 0.8555811967458411,0.8994157477567531,0.03661933994367139,0.3059974884169935,
 0.11326631700535938,0.4954176751171332,0.44596397236380836,
 0.5640638898277968,0.2774555187193347,0.22616453327344044,
 0.34256141737794665,0.7747282652806614,0.3682717635685785,0.4822168382582024,
 0.2575183354794318,0.7892143337895953,0.48178159435850676,0.9067250890929013,
 0.4325951099579577,0.809992876158421,0.13679156038373708,0.5103389161281346,
 0.82928487111262,0.03845708187957253,0.16105180694337018,0.12688121716021505,
 0.6862190308283062,0.9789336523215232,0.5841298834248291,0.9355670567577111,
 0.2323087722214594,0.12067439095742583,0.44434037515543257,
 0.11872723119341377,0.17192879831335994,0.6051195379499432,
 0.010276460112754182,0.7740796724415917,0.04566053309957396,
 0.45926021772611747,0.048479334782661976,0.023530061297979743,
 0.7214070680297682,0.15869234200744997,0.22854905800895953,
 0.7255984694846439,0.9342319058657004,0.8753904502368756,0.8809851378155151,
 0.96452353273984,0.23000060168997305,0.28962233692507044,0.9507706795688218,
 0.7304049569227665,0.21706518254959728,0.3674223349168052,0.5243945402964176,
 0.642296872400703,0.8411965497934304,0.9437128503518416,0.9299612249725158,
 0.8014116367128165,0.29099471541204625,0.7614455027830329,
 0.47506818523764194,0.37081321881504137,0.23073020068953032,
 0.8422045357580009,0.8641406899199695,0.5353943787382346,0.2453938740824334,
 0.64962263367196,0.4846893086541737,0.5912298208250052,0.870040696708724,
 0.4371343037008718,0.0918392708020106,0.14061076347573487,0.9677550059995461,
 0.6601159366122736,0.05737376045026432,0.39494033265402384,
 0.1823039233447018,0.2453538517533087,0.8541550024479582,0.12577678458538133,
 0.8976743465024737,0.7437931774480656,0.31695270763068417,0.6947814152347638,
 0.8786555769564831,0.13156784148886524,0.361300113385403,0.12301372768265129,
 0.11935660269521808,0.12951887864517952,0.8707763641434283,
 0.8918582635089531,0.7389849936913502,0.9229353932926354,0.9960848711635765,
 0.7776330681772675,0.5287506553883707,0.910720483785697,0.5594139790490122,
 0.13718529513220923,0.9013982578495813,0.5914322349772003,0.1198557868822906,
 0.13040864327877144,0.04586194689824863,0.6417072230658467,
 0.23305413415591714,0.8417827236507704,0.09923879448528339,
 0.5754227400861781,0.04973942518520824,0.40897610145958474,
 0.9754656743560579,0.7068732311556688,0.25180254542306846,0.6156127514900737,
 0.36863986867196785,0.7810062957741167,0.3823129685795853,0.7156072289067127,
 0.6059431257009391,0.873373482075099,0.36027154434775643,0.11961681738911023,
 0.10949626236570698,0.7440355012526035,0.18976961721048013,0.736141715020791,
 0.1718382755837683,0.5508261105051335,0.7480830757324035,0.32758757844095165,
 0.10542821768368627,0.2942634680175259,0.6237218599300879,0.4590193258612767,
 0.09472217470214872,0.4249875802421691,0.6431330799641448,0.9720848807527869,
 0.29865404931928297,0.4373775489446602,0.9478066867249316,0.5745650523115716,
 0.3712330519264925,0.6804258479437266,0.9420304661562326,0.9468315079250852,
 0.3149577681532776,0.9531020815714376,0.6790829661761805,0.22043655867821266,
 0.8587053611764037,0.6152661028382903,0.48072498201746805,
 0.026704938919647314,0.04785518564366842,0.5294254259616267,
 0.8925850985081362,0.3658933007542702,0.021684220788739217,
 0.8957371649987963,0.7910856898655161,0.8723694474015442,0.462237913868186,
 0.8257504207909611,0.3658336270433774,0.12799784312948848,0.3960279740503245,
 0.38929162962998887,0.19090967317101692,0.3120056397381673,0.976982625317024,
 0.635558567873481,0.9385339614818186,0.17695101054643847,0.814737080143092,
 0.37212226993255804,0.5531875222764953,0.03069141159485933,
 0.3789768819595126,0.2687422020870668,0.4156239307503278,0.06628091314153572,
 0.7741505600196528,0.09763030996539734,0.21926019138418196,
 0.12631378773265478,0.996190558478741,0.936022410561875,0.8938636223416913,
 0.24364216491857316,0.11343331621604003,0.5567442265332623,
 0.05508187987754776,0.7955467022918277,0.013829058385567028,
 0.4458553971669662,0.9421013165758383,0.8333856569450174,0.6551269990525554,
 0.13732599860237693,0.018158173500214758,0.08443666194164723,
 0.20944181499257752,0.4784000514224678,0.3961389186195087,0.6535638491254065,
 0.947721172141729,0.05149502885694546,0.6738978071991935,0.34792321483024924,
 0.7482179143774486,0.18040393183516445,0.7854308745848891,0.7657161552237465,
 0.9638505887249361,0.08684233886977544,0.21440728554136834,
 0.5344939179639895,0.44854251546224455,0.19406037027168277,0.884586073659367,
 0.5074274407993921,0.9134047372065212,0.6518242293160791,0.07406689486097262,
 0.8869490903022914,0.635692128038738,0.8558670040275389,0.6793412915876701,
 0.33033255463745625,0.9374509086632105,0.5634854399104895,0.8768875956019706,
 0.0424147781953742,0.06213411519664913,0.788266026454654,0.1962369795707608,
 0.22811188405362348,0.6302810663773647,0.4336804224524513,
 0.20489338229151077,0.24158270173394591,0.9515352344408767,
 0.4164776721100556,0.31422324941922297,0.24719109711638354,
 0.06862059444873947,0.7913634697211681,0.5925269990431751,
 0.16156338486184452,0.7703482465373287,0.49101954317996155,
 0.17797679416809853,0.5280833589492527,0.41376450512535234,
 0.48354909995435735,0.6586676588118979,0.8803770740794145,0.5840455971768804,
 0.9540951762102312,0.5831731136912113,0.4159236879169457,0.08979082115602866,
 0.7962881388140811,0.43145504184063144,0.7882739869358627,
 0.12836151343509017,0.0805756683436013,0.08357237807194184,
 0.4146448238205951,0.280679520392091,0.19906130155600243,0.4141701945356878,
 0.592255267383567,0.9687005901827946,0.8507225284093778,0.8489748141394924,
 0.6343423333622864,0.14552517202016046,0.21596760161680706,0.913415142839872,
 0.6129933867725967,0.16930735889848592,0.42284314207850415,
 0.9779516575218504,0.16777792961981697,0.055837524770952385,
 0.8380723760487615,0.5565396791003783,0.22227802342533676,0.5309864182282624,
 0.7797384212040331,0.2916470149162097,0.2560518711875498,0.7191043427669589,
 0.7673901640316734,0.6299740521565232,0.24873673900767423,
 0.03652630038036386,0.795040072877526,0.20801627295813319,0.7079880300425135,
 0.3213148629658654,0.7644612686257388,0.40768280721080563,0.4409148846944614,
 0.03647967560782206,0.667261090201256,0.4370800710474563,0.7867093564868028,
 0.7188873694290682,0.4841535864264539,0.8934633969576343,0.2004455445978428,
 0.31245442001469237,0.4945980207156787,0.7229369106180523,0.3598895050672031,
 0.5341280358534393,0.39118290604306016,0.3372384024059354,
 0.14290946556030804,0.7201139514887991,0.34388659293624524,0.921012038210893,
 0.10218158631936136,0.7943609060770034,0.6954732153797356,0.9990526713399069,
 0.7786833914124001,0.8217200633713998,0.1734385342844148,0.22943974944862255,
 0.9153598177134646,0.4963463678413689,0.7686905087898512,
 0.011326980482102789,0.7299157666802057,0.8424584368357027,
 0.11604834734384306,0.755094510445043,0.7654082956763671,0.6907636390425612,
 0.08554553516033847,0.45349326920490396,0.39291976535550177,
 0.05706437618457383,0.6621467387432534,0.8649228524291612,
 0.05193418454850651,0.23633865640469665,0.7799186642795041,
 0.6787105976794293,0.5275148961427327,0.5073796813200127,0.9938799997127222,
 0.0037963144661351533,0.09079078057842105,0.09819353319829416,
 0.4050921808056084,0.570495279490424,0.07929036231986097,0.39760775770752177,
 0.4592972451171963,0.5170756224167607,0.22454110950344663,0.0724046573004592,
 0.8043256690589167,0.9791401609576937,0.13644196684073862,
 0.07483573121634379,0.7357136512996382,0.82742037877896,0.6121598955593246,
 0.6476614276112986,0.5861372668941291,0.3953029128475194,
 0.004794338943985399,0.29647246910476044,0.45413585471088624,
 0.9525324188522079,0.4738323924911333,0.43311492322475686,
 0.07182642143506457,0.9575752578334717,0.6590215034960809,0.253060260730964,
 0.20293095135442407,0.9259290369691854,0.7715950481738942,0.7089123152909537,
 0.6308527092100599,0.33372300926178244,0.4831463158460563,0.5053247186527555,
 0.6281285824338928,0.05279989789585415,0.3577275866104507,0.8849089065256894,
 0.24007337464944112,0.8498273186689619,0.1338568944710511,0.5271148339446581,
 0.9997386602531865,0.17703331063531336,0.6623771195477234,0.9764819113465842,
 0.6059032240540668,0.955390114095467,0.3809821265455332,0.37326331584796846,
 0.7765104502508524,0.8126716400853659,0.0774464851097969,0.16687823547820424,
 0.6119885024677565,0.7184611025925732,0.30085058838972334,0.8192935510302045,
 0.3501117425088287,0.4295028668954398,0.9343084353664707,0.8968235408005262,
 0.5237284646126578,0.9058089685841775,0.41687800999681235,0.8155564220038886,
 0.14318380421183963,0.9518770667038785,0.2205142978044775,0.6902407792217023,
 0.1256709674312042,0.3706705336151469,0.19933165997279212,
 0.38081428884070434,0.40017740302299976,0.20811354618191213,
 0.7817997934249215,0.7763813007309779,0.9308529712130157,0.8888137145279538,
 0.5488919972159861,0.7408991691609516,0.4167710120795405,0.2743554726063092,
 0.39000369016894687,0.6237002664878732,0.4897978696728421,0.6391683032534365,
 0.6363537779486468,0.8285314243794066,0.3243754694346701,0.06028215779908275,
 0.058406223731084284,0.6645442062595329,0.7176708339352657,
 0.5938742964694019,0.44918859919858023,0.10294010179485125,
 0.8254514244079154,0.40736800944872176,0.7880385742650252,0.926822113547909,
 0.9646824669295766,0.39314675551908584,0.6694815019471285,0.8832787799745502,
 0.7827097125576136,0.49260358133848703,0.040721253121480006,
 0.908999075474402,0.7375390943897527,0.018647230533219372,0.894127468866458,
 0.21712266794517543,0.3594146399184571,0.14538710554202539,
 0.6267576714396389,0.8962648011804001,0.6534315868298813,0.1936708751853743,
 0.30380530646115,0.617719061519918,0.5830193569434334,0.8234305851308906,
 0.39544349050577554,0.15306546133719778,0.37451801079913616,
 0.6587164747283322,0.9509719853297386,0.38196369384279627,
 0.008141852736831656,0.293537231036363,0.7090423464157609,
 0.20079059589927184,0.27251348059665115,0.6768553395864637,
 0.7206260993345786,0.6408472179158048,0.011868690598457854,
 0.4310045239716622,0.2907526100196499,0.054099452626902855,
 0.6413472862710364,0.3225510192760014,0.21845809019635531,0.689337152235866,
 0.22319977249793244,0.040922862380064484,0.09140757536702437,
 0.19402325845099877,0.8039188754166942,0.7542855220213274,0.7289277088370248,
 0.9545217423968614,0.021804121440393787,0.17538012714395912,
 0.7998879773210525,0.35488095753929094,0.06303800039384888,
 0.022406350878010017,0.7956834948772089,0.6639260574524585,
 0.30578039945640667,0.9225571556669494,0.9265703042577257,
 0.10995191498139256,0.14061918604599,0.38468769936104896,0.6001082445027229,
 0.9027264965108174,0.16919049957522747,0.3048478473588496,
 0.14847701748274156,0.31819507761496413,0.7329223262126459,
 0.8800436381159434,0.647220387618857,0.7026757534847371,0.9515600476268606,
 0.988186230044488,0.5151390865590768,0.815267286397547,0.7031591590078703,
 0.5925439261771159,0.6929359162006391,0.8592536910864913,0.04507686971826008,
 0.1418749006973552,0.9206011241827278,0.6394142840298814,0.5584503309710536,
 0.7462922173921878,0.12494323899670834,0.4923546377945277,
 0.11388664659309644,0.5425571242955025,0.2553480650868334,
 0.08315480573056921,0.8259621415179781,0.7565526240009055,0.5394866226940642,
 0.8800627740621453,0.03962242083743217,0.40460863124878166,
 0.046119944422197356,0.026342003758647148,0.2709254296995245,
 0.13966680735651704,0.16784730983298546,0.506460721914996,0.9589816057781888,
 0.4506198467799698,0.3888554549100498,0.36725929573961735,0.2608138015509308,
 0.8886866613998605,0.7189990674398957,0.7993128905411502,0.13791878238523714,
 0.259026594309105,0.13956947326564884,0.32887307127587817,0.8516301922493696,
 0.13260629660070544,0.9888198113285603,0.683714679111768,
 0.004221524148715128,0.12364095787182927,0.7800544081116391,
 0.43339753212270105,0.3303887931948949,0.5387710440327369,0.3428259525781192,
 0.15483893108704316,0.9565351734723792,0.9552871535547588,0.4968905635228813,
 0.7194804531011684,0.9698212081485341,0.5110459711424301,0.3081479629988729,
 0.6708779679566965,0.6731979973019454,0.9471772014569994,0.7756184383604154,
 0.06639303835643284,0.4445069837086377,0.7258793205075912,
 0.11202352619641953,0.4166363697278821,0.5412796069271002,0.8504981231284334,
 0.9318478220145863,0.3042068750128878,0.2842683278949316,0.15243179630633275,
 0.3099488014294841,0.24253467807714002,0.2335922313341523,0.965782972247943,
 0.981038624063101,0.14059336003592848,0.43648116886951527,
 0.027485939586226804,0.8054654515001155,0.06398635840235667,
 0.6915321962427138,0.5827140646150402,0.6274199826569148,0.3561319173383849,
 0.08842813464986721,0.979309383197851,0.35262519867699404,0.3662914957082053,
 0.75473102373111,0.06384182347778844,0.4023236030187768,0.5343842674929163,
 0.7154296805968479,0.6804638207213464,0.9178733526967741,0.7155876528161407,
 0.9912880491699068,0.6159981719077754,0.20729837106966742,0.5996319043940752,
 0.4472329373173891,0.14174866868130864,0.019698748448018333,
 0.04438577472075522,0.2937279514795087,0.2833209733222738,0.6056495496591835,
 0.5450727528196047,0.5435781323383683,0.5747007296649512,0.8728618141257711,
 0.11064497621790026,0.015305645542483926,0.11292918359678861,
 0.3182628036414834,0.1735520730820448,0.7173736041855775,
 0.010912342588808133,0.25916335634383847,0.7834940579293217,
 0.3346019124018935,0.730814546925844,0.5963549489436583,0.6949378344441033,
 0.2469369847652324,0.9091611852484212,0.15708934541180952,0.8956964490782842,
 0.8425888780356536,0.21690531068164798,0.6341191050685705,0.7410483567892627,
 0.6406781838751288,0.9333358998598579,0.5788713527819033,0.2699358681020134,
 0.8182382779573356,0.713962483841656,0.47797087505376945,0.7259403328842517,
 0.13341112157920376,0.5612595837753782,0.27842561898259277,
 0.6841940801313829,0.8330432221049959,0.9461640855382221,0.3537696521630428,
 0.020320107399697696,0.04206981279232258,0.22090812349525235,
 0.48769264285632863,0.7107873063625898,0.9328590366565943,0.8368746524596561,
 0.4906965258331034,0.049266768384849335,0.03364967922893469,
 0.8352793948452281,0.029773797949923186,0.730643604026764,0.8640223063768077,
 0.8739102830770854,0.8308315470027883,0.41569415559030887,0.842103002563102,
 0.5452941198008621,0.4125383594659926,0.8022732779505466,0.41349295488024695,
 0.8819455199033216,0.9720511230552885,0.894289648341955,0.9091472868770417,
 0.9506934642176013,0.4068993209999521,0.9898892843240364,0.642473892924069,
 0.7313145556799412,0.3417205570556294,0.9835385303239447,0.1286679443109806,
 0.12881295898130207,0.36415191103030187,0.3044677962352751,0.592649249415381,
 0.010764493451464352,0.415616323860144,0.05018895194951911,
 0.8550936707250727,0.7940346359970908,0.023233881578921256,0.614569326667253,
 0.6607622864938358,0.9375433905341989,0.5312965864729028,0.5634708238536407,
 0.9857151767617542,0.7439576446986783,0.9266575649744404,0.3660260901103768,
 0.894941439992042,0.9275249757009738,0.289203551465291,0.06194086958663936,
 0.8375193565647385,0.920766931465268,0.4812025735575155,0.27330658504492633,
 0.14165532890618804,0.791507183014915,0.7273476531759414,0.16585720581058194,
 0.0612713524281312,0.9412917474817291,0.817207895421997,0.03874210668108935,
 0.6642707611692739,0.24395831114118127,0.6312779738246991,0.6965782076262912,
 0.5624817033448204,0.3544112456607694,0.5720385454048601,0.1385953544146502,
 0.780177171059237,0.11459703545695166,0.8254758737228232,0.44734316890493586,
 0.10866326859377962,0.954747534472402,0.8294224512765742,0.38612420989034646,
 0.5843097754902036,0.6225639386780351,0.7570765094224079,0.7278172056688256,
 0.5437279296156523,0.8493751864102062,0.3566233369928803,0.3507573153608995,
 0.7508798502307983,0.15276660547219056,0.9048398830434549,0.5332171352043084,
 0.7140438313928622,0.3370673858217218,0.039355027139790444,
 0.9683510240667864,0.1359872274545728,0.3115616775329825,0.4353732945191222,
 0.38545059260620673,0.6616806043420196,0.225985936404145,0.7833361268443619,
 0.22211209506886376,0.9929968833052506,0.6295050268492041,0.7247912771471281,
 0.28288493292468375,0.23549938521612235,0.24628211845552928,
 0.24093604790345236,0.6623761179689563,0.6373889243462787,0.8088299004391775,
 0.3051283167063943,0.01192989918334042,0.014584481814604588,
 0.379295264711231,0.9367798754320393,0.703340755629676,0.8071297802095656,
 0.5389504443493363,0.10547944247979668,0.8975767738575705,0.5432063850102331,
 0.7926790787899104,0.6395547473657157,0.5244615716041102,0.19642501144682,
 0.9966255698812151,0.7945648619076738,0.7594458366750458,0.29075246501997576,
 0.7517874349475162,0.29399238755194057,0.9357460462336509,0.767165163240985,
 0.9296899570441084,0.3893565802386042,0.28886510608369764,0.9218895763844164,
 0.9836727970875323,0.3688296418192851,0.28035264262072634,0.5462838477668567,
 0.17783869912855388,0.551923265403373,0.12943009387927573,0.8380689600201559,
 0.4882484378760139,0.01690727493320665,0.787984516821651,0.6176140073038019,
 0.18588168013948947,0.17604708619597464,0.9100042780054799,
 0.37207605861579174,0.29352544286118376,0.7206753310759024,
 0.27566355149320876,0.9445577470503522,0.4141633166094211,
 0.17207063654309807,0.09601046813575354,0.7227449386457091,
 0.9744784446238803,0.798240143835274,0.21531482384572953,0.2004160281000127,
 0.8190051999933025,0.21240336137317517,0.7794706983787644,0.6052093973832742,
 0.1750430110392308,0.33818498645286976,0.9933220736288225,0.267516026209909,
 0.7156157909820541,0.2451401301149464,0.4106556121105931,0.47660411012649884,
 0.5516206124175078,0.0783825712848416,0.06951201235114879,0.9541002018062182,
 0.8463412214121687,0.6149257134454714,0.7811838782205858,0.7901781444606847,
 0.662998634864809,0.9712358758894896,0.26853846543858884,0.02237290102823697,
 0.02874274539859345,0.04772588973752967,0.5945114781640619,
 0.29684600832308206,0.7994034432946021,0.8519399203742137,0.6676471450214857,
 0.6690080466237194,0.8495593437214203,0.749420019443557,0.24827002150934674,
 0.4139107215802953,0.5500819598167509,0.8158982085215585,0.4667525135950724,
 0.2196681545511786,0.7181488879804019,0.9373085906026851,0.09189021008793519,
 0.5562920504695574,0.6428579970906511,0.19154485640015972,0.6245226457148874,
 0.3756592383939775,0.3728591379594127,0.9407152907471646,0.8670072177472533,
 0.9831791175080173,0.6134897116429563,0.038272577330768165,
 0.3580096192353087,0.1345502312058957,0.4275964416780127,0.3636232512945965,
 0.9424513840239784,0.6258986040601412,0.4891765349216308,0.12868698544590285,
 0.6927807097935795,0.4461146610782958,0.8823384096797218,0.49907627128228005,
 0.6356390046541831,0.8830765287418499,0.4042414197991733,0.8688414964565677,
 0.9642685279536832,0.0973418351219345,0.6594223332341328,0.391612339828192,
 0.5959175169798459,0.5615162900622003,0.5538206539051522,0.1002551974774466,
 0.061986163410843065,0.1866736040163124,0.5856668881048285,
 0.4271431216686976,0.24169221204766156,0.3698599686414019,0.4924118240361569,
 0.6916094044599076,0.618780621968138,0.2813582783693023,0.8483624955696756,
 0.376356239493122,0.36904853135503224,0.7665164681030723,0.7817350277464337,
 0.2714423979373104,0.17673145689287684,0.8525563689023546,0.8349610891680493,
 0.3672186400279268,0.9846514451127079,0.33870882406569647,0.8185575647516513,
 0.3936053337487192,0.7168187986815073,0.40567895513186936,0.2942420430663354,
 0.6423288798764978,0.6042947444345025,0.5994042782008103,0.9334445868349439,
 0.9837255542216621,0.8825917573345917,0.7990892554902261,0.1794789600769663,
 0.8586867780602598,0.1688869034392847,0.8266671837210667,0.45125430764970165,
 0.3227586160400333,0.6755675281483142,0.5497498695583204,0.13263472202064652,
 0.5336504591034478,0.7843913980040984,0.69258762057354,0.03643583069139977,
 0.17097500652319486,0.2645092281621034,0.1265513430282733,0.371020281331331,
 0.21493392588949845,0.3401729789079425,0.38504057257804714,
 0.41338914869504495,0.22427138895173937,0.33823410661658315,
 0.9549638185836828,0.4197502567676069,0.6340684749980091,0.46791125365785013,
 0.7114475083477365,0.9945002819765659,0.9512035844249731,0.01945821872090292,
 0.7974433815522941,0.31342365735633826,0.6229370903215269,0.7684865711794554,
 0.7400830440990411,0.8953106961671468,0.3319019551027339,0.6196861591784796,
 0.17631399927661262,0.31384093374793876,0.15200058916497083,
 0.392933864061602,0.6122745909141151,0.7197909767519498,0.18806334094070598,
 0.0997953318511241,0.17658982625725905,0.4976507884118615,0.9739227913459138,
 0.39115102209772745,0.902384428416965,0.3782688963302745,0.8948216114719013,
 0.8988901617450662,0.9197655457831299,0.5963161490083148,0.13528361649316245,
 0.3474243867540242,0.9150267790022145,0.8112768908997645,0.7378260424165305,
 0.06532725884139612,0.2369484588398616,0.6892680258568663,
 0.005956440533348539,0.7595508937234251,0.03745041682664407,
 0.7548487172986424,0.2428053939021384,0.9432541679914098,0.7176571628684627,
 0.18935087201196232,0.956586047543964,0.7873771227762337,0.08114744208441305,
 0.4714450569651145,0.7675288924637786,0.018665552927343176,
 0.6497739033886827,0.5626625539120371,0.11397571465177514,0.6156199715401914,
 0.8359181558629972,0.5807922637411611,0.5625613702315211,0.22553566163093708,
 0.7070962163878808,0.09170613287751506,0.04673022569435448,
 0.6372410674839929,0.6513557042802027,0.5144721231654676,0.05659222397225849,
 0.2360557844965938,0.4306395897782741,0.9228580923203398,0.13983452647214412,
 0.2647125234713663,0.9788600313739968,0.41076275967954423,0.4459403658509967,
 0.9650718206202684,0.020141728923546376,0.9576269896536765,
 0.32664340111243295,0.459932185002054,0.5471507390624896,0.9250003840328413,
 0.48132341009135393,0.9667137676261308,0.7585131059773556,0.8548647793512393,
 0.5564982223370235,0.4958850207761506,0.582589120687126,0.768440676070108,
 0.06290636948136163,0.6890242090943562,0.6311790524026515,0.9021795399610648,
 0.5423411769457909,0.5463231355834116,0.9248361889090179,0.1574086969801265,
 0.14552534306142006,0.06101515896940635,0.5345141862474083,
 0.1717035275460299,0.6790371135187566,0.1435265181256148,0.13164344761299596,
 0.5760473265811997,0.5714244119407368,0.4765543691313463,0.5765358843880134,
 0.43985693281298754,0.3018001773563199,0.5119048167993261,
 0.47220046255358694,0.659044343586925,0.9887899358944496,0.8688556258924566,
 0.5017588790904928,0.3575037524052951,0.996995680596295,0.05880866865734813,
 0.1332608502962669,0.6763536175210264,0.43416603399399234,0.3509949366179669,
 0.8883599012686144,0.35976872187173314,0.4824271566716829,0.4912490955376875,
 0.0801678741302847,0.6031847386937553,0.6855048422988845,0.7361675894968243,
 0.13551982507830052,0.03199437370419966,0.03720589979853273,
 0.07730457979367211,0.9105623689443214,0.6396371212474963,0.4242982878071727,
 0.704548948553785,0.5900003205121135,0.3868984601628155,0.6167939983250015,
 0.6322701388091928,0.07016716808826173,0.41125459397350195,
 0.2610977529598787,0.5848892975803286,0.18273720729781706,0.8192497555210857,
 0.7537549864214043,0.9360530821917878,0.7849656208057961,0.18372377278532792,
 0.4618317266893074,0.11529783637146063,0.7671668306324633,0.9980279209864248,
 0.6533024860912524,0.3842188643017619,0.9419751839962744,0.8475318475328623,
 0.5306534623842192,0.30080191953112645,0.7080119199526234,0.9996033624834082,
 0.6072917916440874,0.6353458011473904,0.04459430099864692,0.0571440513569802,
 0.9605657807579101,0.12242507633035682,0.9301876338420071,
 0.30738507725611575,0.37099634101102097,0.7011577831092122,
 0.3120512552076218,0.133198128239848,0.14900802749764952,0.04376272856903163,
 0.23555691665088985,0.5651029081395514,0.7406871410477028,0.8951589253665345,
 0.74501441411059,0.5557054672979763,0.9984149770746056,0.09056289024074182,
 0.7728360995837078,0.00789363216499539,0.839310395478376,0.41524384233278655,
 0.10107239255296563,0.33042864929915994,0.9987493322498008,
 0.6899372335707087,0.19960842268938062,0.21859607438220585,
 0.7993860729197677,0.9460513597504318,0.34815214374619863,0.5923707912284925,
 0.3332279647223855,0.7603381620279406,0.3321695059233367,0.4270826082596132,
 0.9858330070745217,0.21617826645381666,0.9949401256597453,
 0.30071890514481747,0.9595761393158627,0.21314021765883262,0.889584159086026,
 0.3183259988597352,0.9148934864959375,0.8297821952606773,0.8783007468973968,
 0.9257336672882304,0.17434052886567608,0.9923895950835468,0.8074115784852521,
 0.9154279084745289,0.7292758849062271,0.611716116731481,0.37630884471667825,
 0.6851859673406068,0.791339604993827,0.6731200595389087,0.5042428766833731,
 0.4220033724689678,0.7107688945509338,0.2091618438758579,0.13353354964507624,
 0.6685095206395185,0.4216320221699591,0.11414276105346777,
 0.42846398704629296,0.5527950743059763,0.5004712839729613,
 0.16889719792718538,0.3516444827055518,0.334925869901904,0.9299306920256779,
 0.185644808327875,0.16559017318511504,0.9311411626332019,
 0.027965205066814547,0.07066466085828227,0.6405761808114964,
 0.9042609338352919,0.4097977426984366,0.6683659416923439,0.40020662169696564,
 0.49623757454951367,0.4153433368477949,0.09583882134073995,
 0.7954082639421103,0.4912096074926574,0.15323876625001032,0.7831817131797911,
 0.25224769076516385,0.684403535769749,0.5119914435867897,
 0.033030711319075645,0.8247583728186424,0.18254145656199816,
 0.33922883416478034,0.348334013668694,0.9135969873736415,0.9606125516186655,
 0.24167243150284634,0.5356106501235015,0.5842725270980029,0.6195896267635138,
 0.12336741691587494,0.4017617596278047,0.033549724666844805,
 0.7146490822700879,0.9880367256421231,0.4272484547204072,0.18924320758722712,
 0.4276949593108076,0.3125037927203843,0.544195537422897,0.8472850302200788,
 0.9559808514696234,0.845007246976256,0.9712446304386486,0.05099798848081827,
 0.5574273644469165,0.9538838237512057,0.8240797224754018,0.18539878474267468,
 0.5459711812214533,0.40043983679087325,0.5899069206469628,0.7038405220982714,
 0.02803402908076147,0.1934862841771473,0.5319346608560904,0.7682800276839049,
 0.11751939019563395,0.598447973804761,0.6312786629278387,0.9459851175032064,
 0.49068492703719435,0.26628352105079134,0.6801103359831961,
 0.7582805190314272,0.5228461713003546,0.7358768102087128,0.23296521942824555,
 0.5018934527518226,0.26259409966142033,0.7707904333048453,0.2768330536580572,
 0.9827386770852085,0.0205062200605427,0.036819363398048366,
 0.8747085074044207,0.5712132628991519,0.8058413436887304,0.9855917930786307,
 0.15076497658833532,0.2619785346465584,0.3061149168449646,0.6398232312781952,
 0.6034669091079224,8.835214148484871e-4,0.23287534780169739,
 0.8200211192294433,0.44129711822179485,0.003215175851101204,
 0.7973764057831815,0.7752798182428777,0.592483207506781,0.5958314707318573,
 0.7270804149268185,0.20549676575214282,0.958208624364892,0.38102199311498786,
 0.5944302677929387,0.9416078342482695,0.41309446752194523,
 0.42639023138308296,0.3439639809292627,0.3350652332564714,
 0.20192743471033636,0.4934506922509222,0.8923458041610584,
 0.19128806302157675,0.6257743895266437,0.7070984779473526,0.3102096296176503,
 0.4007095830198719,0.9434901961983662,0.9263312543926228,0.10602947054672474,
 0.22505258423169372,0.6370748350360176,0.9737851793748198,
 0.25573175867262765,0.2138317241305474,0.7642605292879042,
 0.45198819127208534,0.26993563070444493,0.13532707304815683,
 0.6976635504044975,0.3497867891184079,0.13637452789589477,0.5101926048654265,
 0.8160806244368812,0.42530190525573197,0.03149787876361143,
 0.2843389687107427,0.8332334300817134,0.6437487514015579,0.6328413785846888,
 0.547789364627455,0.546540021079589,0.5015451258805838,0.8271384794149003,
 0.2965685554218811,0.18901765496198875,0.35114107203718914,
 0.5947713028353928,0.158300890965693,0.0701700959860333,0.8631816136108847,
 0.07651020361139693,0.23830110700987772,0.8885858909087263,
 0.41475236372067104,0.008374195190997136,0.25897232855439867,
 0.16712103245046261,0.8151439833088058,0.7866088925550613,0.886566968589148,
 0.6718255022162586,0.5534595405140265,0.8241485193388259,0.26717194040011527,
 0.08690301130166178,0.48962796671606,0.7936485401804201,0.8938662224476372,
 0.44895811449411815,0.575352485981281,0.9289193870747755,0.9836499128705595,
 0.1237359525573658,0.648335879977503,0.16116592595164414,0.8442410989785093,
 0.06244170421859607,0.6934873466764943,0.8248972700213715,
 0.016381669401940635,0.7358481272265025,0.5437312007172375,
 0.057691669679067026,0.7737617232312635,0.8349748005090174,
 0.8021593981520921,0.9094776932546501,0.23963039757068993,0.7059582712279198,
 0.9997056313897745,0.38235891018223456,0.5048273688488669,
 0.12059364000117168,0.17279672047035033,0.2546982215418019,
 0.3920205669659479,0.9618790640101598,0.7038607163194079,0.43666216167533956,
 0.2896262461360819,0.5820341699835097,0.6259210767600069,0.17737216749140106,
 0.23107571806262817,0.5545022510610389,0.6171329376247471,0.9171739096723607,
 0.12036730206938961,0.13515944251087186,0.6493912010925323,
 0.7719630312940411,0.39199179329260325,0.8813605109565585,0.5712687012794051,
 0.8506571592453087,0.42230195261865744,0.6988862803394934,0.6096950148744791,
 0.646757897248429,0.2137532154454439,0.14375308156991762,0.583537820367725,
 0.4896783958146046,0.6190538455799072,0.27910285981091043,0.6707408241606105,
 0.9024533730892662,0.19415597077396707,0.6993052849067874,0.9073686143023664,
 0.5602711250364463,0.9321521743897613,0.33959373869101894,0.3983248996054475,
 0.19720126597087106,0.44731561877670534,0.5821264470225412,
 0.5788658942318329,0.3731299422738097,0.9987266673434733,0.47267295957832756,
 0.05613447516355419,0.049875179644793266,0.8824077020191705,
 0.25656839087229155,0.1800783265972542,0.0341970634719021,0.7503859490888997,
 0.36760654453938557,0.1938846818543477,0.6375305084902402,0.7954556566712413,
 0.610445659927305,0.15182827833958512,0.4690675445151339,0.2888308962988245,
 0.44826718225456164,0.6611755162403853,0.27850123480821776,
 0.5388237840742555,0.40304654284464914,0.2670947729600609,0.5127640108928138,
 0.32864234081265664,0.3987819965246969,0.09546199709240288,
 0.09612149903719291,0.21390884742179317,0.7826219084528153,
 0.24746260714702395,0.7709783337437881,0.5903852423155336,
 0.26763997550421204,0.09170919006123035,0.24354153968817316,
 0.49432701333323315,0.9074855392722843,0.21970558893675662,
 0.6233946092739326,0.08444421422878856,0.8261058567423979,0.3048342770975735,
 0.28448965199047005,0.3530933742166634,0.6963962985959723,0.7716040162980355,
 0.22033399529594,0.26352745144721057,0.6532695651333458,0.19057132458295023,
 0.04668740698541762,0.19795018862991398,0.46250076973528187,
 0.3062716284580812,0.4232907861623625,0.15096260479041523,0.9646177752608384,
 0.21879825285475596,0.03751459925314682,0.9074824519696407,0.887800981423566,
 0.523028465695574,0.8578073498623126,0.2262813871360202,0.998539203499066,
 0.5867781564567243,0.9061159629913338,0.27675907002960376,0.6083521047619331,
 0.0710559216301645,0.8792644923187078,0.739326799168457,0.15858038078178138,
 0.16081548337351703,0.23995940437824648,0.44009523828891317,
 0.4945912680987258,0.7011381388621556,0.9063022212884155,0.6022677485279642,
 0.5603393252873584,0.4789500899935516,0.06481088151029146,0.7501000413137523,
 0.9871221206451591,0.909328728526775,0.7821436219568433,0.30277618135486584,
 0.5384576589940803,0.061121144198167165,0.4424600427899571,
 0.13368485624077708,0.758973682142619,0.7232669801883344,0.35848644987997025,
 0.7045141666333485,0.7941271996492957,0.1695819223356726,0.23962377646999333,
 0.16526881353694145,0.6122957655874469,0.28354711607989713,
 0.7703119938541427,0.18211489378298917,0.494363863735942,0.7880365357438756,
 0.20529507962603788,0.19101641143608017,0.6125209117148931,
 0.13778923060584758,0.9009817721508441,0.14442504805278755,
 0.5779102169971326,0.09041731577818368,0.23649120476157381,
 0.9920849478197679,0.3878578166667839,0.8858104235951735,0.9089582967793102,
 0.42586607733435644,0.7076731613967325,0.6535767723580003,
 0.18550656267525056,0.045422614720948884,0.8021472314353814,
 0.19247864864625996,0.8426591160567066,0.39866224018136376,
 0.9600444119067697,0.35913692547991904,0.8425819900752154,0.7953540951721406,
 0.1377997997518008,0.5576664278115758,0.8621941497731271,0.24756661114111567,
 0.8676601892265701,0.07818644577728484,0.664126447793339,0.5602647256983523,
 0.8494461200054073,0.4359337745090993,0.3778936821451033,0.9918596278819853,
 0.046561678974747944,0.2564663027188916,0.6665059399204953,
 0.13423721272569367,0.5756574153585806,0.08985447235580735,
 0.12401309159866125,0.35607272701930803,0.5917457539750789,
 0.2627153601245019,0.42241120940928845,0.6505436717214432,0.2742485718757317,
 0.9313250837711243,0.004675623706860921,0.9838788385044483,0.686265518787325,
 0.4953017840120262,0.5495707647940822,0.7397338100319428,0.07322370396457878,
 0.8884669761684706,0.8353112889327297,0.6513607605485033,0.06821144143577385,
 0.5607019386621819,0.2353181724747201,0.968408361820557,0.12964899108752692,
 0.9797560695247618,0.38616082786550854,0.4640066974003708,0.5608049381419485,
 0.38259488573757683,0.7350956532219479,0.9473145525591996,0.6024376067693957,
 0.10204029353082056,0.056097183258840944,0.5394157989732813,
 0.8974935666772628,0.18202800523746787,0.427100709347107,0.6935122046198906,
 0.23456973740033216,0.5784969068349921,0.021668557366083085,
 0.7039570879805574,0.5198408567363516,0.9273804857524581,0.9390383948103274,
 0.0886391871526565,0.27385658171474137,0.22731044838036896,
 0.3138525663864208,0.8162743542058323,0.8138155255441497,0.2223910369416543,
 0.4434412107667791,0.4854794521254583,0.9993009340797772,0.9805511560920441,
 0.10509755102353235,0.8257772873591784,0.5576407902320728,0.1360179679796114,
 0.18410345598218558,0.9272152807984664,0.4703314660854643,
 0.13062910831608376,0.5623407645858612,0.06279315791213158,
 0.16206603996240243,0.5419744404554824,0.429482256470181,0.6576657496359817,
 0.8243347715686933,0.8461637910783122,0.5075052197918105,0.5945103852111385,
 0.9645012011699953,0.9839982782277507,0.3131959592188499,0.9294911790927741,
 0.2837066872357239,0.9908788533815733,0.003724088974984996,
 0.2508116175109185,0.173837997513001,0.6224654240568998,0.5457622007050343,
 0.31492671218012513,0.7391452917536685,0.25906812264044987,
 0.16254470436120427,0.13286934886998913,0.6369766006778003,
 0.9243691260176257,0.08936242408996165,0.6323159518690361,0.5643229021592824,
 0.6442814178033334,0.2672109361622652,0.07144107449613823,0.9880339951337926,
 0.2026363008790617,0.24880298907847653,0.5845464976163902,
 0.03203982151599061,0.5809128892075397,0.46241720939075526,
 0.8455467236875723,0.21520759990949112,0.2474807335095066,0.7271412676361695,
 0.5074493541965603,0.29565664655962076,0.980276246925932,0.9776442267212587,
 0.8455038380252385,0.4923439307258072,0.1971711270323241,0.22963335235590343,
 0.20407159949086018,0.7619082793955346,0.3656996450409714,0.4469123812843858,
 0.908282366179118,0.01316521449389696,0.6151260283974976,0.4010577225955446,
 0.9293584228771287,0.29187274582303135,0.7650477861757629,0.8869173510297393,
 0.5258420123290444,0.09838810030122014,0.8174743582056295,
 0.09062099766304665,0.4178629356925213,0.7403729642292173,
 0.17341035351418332,0.73372882152262,0.8049774822829704,0.5423516856036097,
 0.5584057134462259,0.3523658919076277,0.040978940737745106,
 0.6633675513504855,0.44917821605115027,0.8558086109358571,0.9177087472211045,
 0.7304143201471878,0.2804894585354727,0.4968548817971228,0.7841612649287727,
 0.370299814019469,0.05304206530742839,0.10872444295068917,0.9648457399856057,
 0.3231193647414108,0.4730328572401852,0.33122497632934134,
 0.06328886398877787,0.8362751996643824,0.8906197162416729,0.5616481479903176,
 0.21534125832857554,0.18354587727900318,0.5054199001809885,
 0.06164022669112901,0.6141670086256011,0.6162097294699272,0.8097748269945184,
 0.05376474602940051,0.8532817678178504,0.1776870946462069,
 0.47298854022211645,0.5419629069642028,0.09987165325354042,
 0.7075097304624819,0.9384095036178567,0.8098853705048187,0.04495938994755233,
 0.15603718967345526,0.8108172335939459,0.4860038407432641,0.5164735441969227,
 0.33847267700140626,0.4437987660045213,0.615496065133621,0.8102300595195269,
 0.9808327700181821,0.6040704757631733,0.28327072557325594,0.6692002587372504,
 0.2370304380821796,0.3084284055224277,0.7712066909441616,0.19858558929111125,
 0.8019663299093046,0.7701689249258993,0.8079885588004245,0.1685323334628932,
 0.2912394330793031,0.01295946353784605,0.008845193861420908,
 0.574627061318649,0.20549273928560563,0.7406918326114434,0.3065883188760371,
 0.6404365304857753,0.8135983813043809,0.5857457745039893,0.749358525809665,
 0.19727478991385805,0.006914506383522867,0.08636619858280992,
 0.520424451408871,0.6105272060813702,0.5040333703136931,0.38516832607802265,
 0.3109261691872842,0.5506574868278854,0.5541513787846659,0.1849175173087203,
 0.8794674401382736,0.2988069267669703,0.002918842856352244,
 0.9674627524909962,0.14758543553531822,0.5442188095465684,0.3284965743496241,
 0.13886851777610454,0.3916558828016621,0.6392769209061324,0.3623374360903231,
 0.377643141634468,0.21702212499987772,0.48030213846187253,0.765629477011105,
 0.14375858642716055,0.9023410565156391,0.7751985270523539,0.5215743011702386,
 0.5957435279862214,0.5895327876901314,0.21599269614418415,0.5220277178541674,
 0.2050331995456639,0.23344254424771493,0.18107039999919827,
 0.7300397948314347,0.7023116772054753,0.6193624283474988,0.6956707294799817,
 0.42537892726832716,0.21045421980053347,0.7387592774272518,
 0.7880241453416326,0.828109552847839,0.3374394682955312,0.7206069936085906,
 0.6443318344492399,0.5602238687051602,0.6223925817910424,0.6909847900086216,
 0.7092229810705699,0.8544981155639298,0.8035370344186176,0.8998508745677333,
 0.2047697115169993,0.38523360657629335,0.24984474176763138,
 0.6663287292753124,0.24811456867651005,0.22823914256654843,
 0.3638724172929514,0.7639535974427863,0.10343372171168319,0.3289527913442254,
 0.4700579020110043,0.6697401298803796,0.5287032647200988,0.7219708202572271,
 0.5830046265648996,0.40592080513121764,0.6278115534770188,0.5866479406337043,
 0.9431561567753403,0.7009925402171068,0.5029544922548745,0.48711650752584246,
 0.663705800028667,0.042231128776635374,0.8319781464281447,0.3156033695722975,
 0.016741226902840745,0.9012097238912306,0.9747334859799398,
 0.5704176663466504,0.4563398476163234,0.8827936672715753,0.8763358038036407,
 0.7369589064399942,0.3773208248911534,0.8822503243079169,0.9409833029106488,
 0.40965538422075265,0.9756723968201374,0.5256857166739204,
 0.19779876114317796,0.7130976725965715,0.24628530474282284,
 0.5922290013242053,0.20615065667309418,0.19061426235892465,
 0.05402117765567993,0.38284609060618857,0.7041433985719234,
 0.12553361472437985,0.8796802025513679,0.3643467918775767,
 0.48980336795302504,0.45290193849952787,0.06174816032145125,
 0.8359723414260124,0.9785055531204939,0.3546913064861851,0.8568865460334332,
 0.8432512335416741,0.2574280917185877,0.022081742115191982,
 0.9835642507208815,0.9403595772229202,0.9892105968765568,0.7623395654629402,
 0.27825298612130067,0.8943286554465589,0.45896428549978674,
 0.6193944245786239,0.6915216699755666,0.9003651347908677,0.9567246070711937,
 0.1789990266632796,0.453965791795767,0.5805686792315371,0.93543175702483,
 0.40453045683922484,0.895013165070491,0.29573832361911123,0.3877629283486461,
 0.2325298628708239,0.3290822296385567,0.16116378144678556,0.7422437586828613,
 0.5931376953328799,0.04938784735902724,0.011423625013223915,0.63873235839235,
 0.4437589122194824,0.9012985498332644,0.38055512812217784,0.945416287375368,
 0.6818812434618479,0.10876897703186161,0.3597664694130229,0.2837752387312297,
 0.48978277827504535,0.04576769475315745,0.5184469504055377,
 0.028864284115677652,0.40355973617112695,0.17471394170749477,
 0.8188945455848599,0.6022839937195674,0.5474914545766358,0.74483279309909,
 0.16577050227432477,0.7060245009224834,0.810248122126659,0.17400030094706087,
 0.6097174172546804,0.5730408396554526,0.7117993387102334,0.9242624974309667,
 0.22549202069495433,0.10404941204091267,0.34772700486047414,
 0.2142111010315323,0.29580878679453415,0.9145249959535389,0.4454346043794619,
 0.14654815043289426,0.5092103095997015,0.6073431828584228,0.7723752089540828,
 0.8296801669316529,0.6145889752425464,0.45539444892038006,0.9802488442716215,
 0.7723934878717413,0.3984137414016362,0.0453969580492255,0.19262737461220591,
 0.7842846443711298,0.836605824372007,0.6551820967113904,0.39328194305937636,
 0.38438852298656756,0.9055528850629695,0.3378039847983473,0.6286277051647791,
 0.6855282341581816,0.9388384583371052,0.9543369644339412,0.4581740859557083,
 0.4347971596387863,0.971290369538399,0.022143890934286037,0.5757336653476421,
 0.2566313854930682,0.1552542091947634,0.1813121802269787,0.6442629887809788,
 0.8214381101447954,0.9316524380735961,0.41179902554707704,
 0.26318270359315565,0.7921627632820126,0.749913077776858,0.4891909054958189,
 0.8741585419016785,0.05071396528962446,0.7887516321090803,
 0.12513654163008037,0.6371785440819763,0.25257271801349024,
 0.5377325828419349,0.9608607878239078,0.22501800184092713,0.7408221882195687,
 0.7873614779002784,0.7422203934873667,0.36889857568984075,0.8611722713542911,
 0.29020272024539473,0.6379572629714763,0.5651796964755313,0.7106220584185362,
 0.5181771739910774,0.6905494471594376,0.28132717293660525,
 0.49710152858899037,0.482575193665564,0.22542733606539167,0.4433119385935864,
 0.5292780768464711,0.9382855982626074,0.10180963710381807,
 0.18935214448019044,0.6629030193365386,0.7978174874371962,0.7269231301426194,
 0.8389750018630118,0.28041692489039116,0.746532705699669,0.9358580241482483,
 0.4594775807193877,0.9350779721952289,0.3168955599907789,0.8574947926527217,
 0.40459058448486196,0.6478837196732636,0.23345913118674777,
 0.0626043304091678,0.9699649137020923,0.596386326070883,0.11517892733002766,
 0.7879544705191479,0.6644273205908505,0.5577652899302545,0.1759480478942561,
 0.9211697680863934,0.38568011742123554,0.20252046598416185,
 0.2992836429833412,0.8762248348694005,0.6421758013772091,0.33351391877573255,
 0.749301041449972,0.02441803008865473,0.8733079943427071,0.5022923354540499,
 0.7849193940672915,0.3872332107574892,0.1286693431603878,0.37432186620644725,
 0.46621132015617495,0.5034319630538941,0.8568179924566386,0.7320137756463297,
 0.8289513790595842,0.8575131804463828,0.05518595788671943,0.9947769097328453,
 0.3136117730252176,0.4718694537026753,0.27244222854696676,0.7167996922509385,
 0.3649961941095793,0.29619652162765187,0.28963440707295307,
 0.5700206949560568,0.29546020692026387,0.5675341640185981,0.6660732693909823,
 0.38145036893174367,0.6673916296212925,0.15710421949803366,
 0.10077568863390574,0.3196417409842224,0.7264038371430309,0.7946786100694864,
 0.7921109337487021,0.5288523833689165,0.46348944900195366,
 0.08266833847822308,0.3654236912799286,0.7179647621339056,0.7176980884954943,
 0.0031704897341062743,0.9791527918788698,0.4030172435373599,
 0.7081534157139993,0.1706201630427765,0.5313069918672996,0.17178564533188478,
 0.4785811970042968,0.7870608278714513,0.017269261471738795,
 0.16395759171420998,0.46275548531147426,0.11408645577896248,
 0.9228113106858034,0.14519906752949163,0.4551706528082039,0.5227748640289805,
 0.3936421359904444,0.9984999200651146,0.6419214687597564,0.6799582616188106,
 0.9850723573639822,0.513747253652053,0.2852360766221451,0.15665615298732716,
 0.902907367321711,0.060235457636436784,0.08921279856590347,
 0.5518648646166954,0.04176400707715844,0.2204506296727844,0.5164259393319904,
 0.51438524961123,0.7727935487071269,0.893392319226178,0.15893011058647533,
 0.40506224429407667,0.12260424109372414,0.29617136118063836,
 0.19791247630313658,0.7491624989477985,0.22485237821868764,
 0.9545265118253194,0.5975706103268972,0.8605386816164161,0.8556771623156056,
 0.15374921809505282,0.002373853310829621,0.43462942613064204,
 0.8526079907889504,0.5127301483946948,0.4613985663418554,0.5771406785847912,
 0.8793216579724376,0.22927794362004905,0.03348404092912638,
 0.3403745849824291,0.41768966361342363,0.6615164518876071,0.554990747021053,
 0.16234324453026927,0.6991292207785289,0.25990394812756856,
 0.1986824443513604,0.8041694910072713,0.8318637417231721,0.7336455527523974,
 0.5010699982375745,0.2557606783911329,0.09664061297548687,
 0.22915507563308046,0.7572632603519611,0.48220601599162227,
 0.5998119709060004,0.061762733291406136,0.6085728836537181,
 0.9228510343789192,0.47462360085707767,0.767155687230847,0.5480806347555657,
 0.42347316020635284,0.0471000026747882,0.7512220007799453,0.3145111921329774,
 0.7955709002260959,0.7799049646361096,0.6474121659136984,0.08542551097744022,
 0.13799813998579846,0.24787007530830274,0.9536853093983081,
 0.05922761778235974,0.6820241907400336,0.853526575423102,0.24780531999709832,
 0.5545110746778669,0.5050757927544048,0.47411187527276155,0.522966650317445,
 0.25607521724015847,0.934395398103806,0.8161105287385361,0.6822570190650509,
 0.921373615375187,0.7868193782081306,0.8818961276617647,0.6480166018125167,
 0.5628751599247801,0.2561542609330629,0.33850836169044585,
 0.36837623015604426,0.10237555052463576,0.7083420194618574,
 0.37633138147825207,0.8029029491689489,0.9982892764227621,0.8181334659109945,
 0.565166829125746,0.372003156854838,0.704302380773044,0.192195411233529,
 0.3436938067723171,0.7955228422584363,0.6802770908057094,0.4408220970023756,
 0.37538280441093064,0.11910020244251218,0.473498633509701,0.3414309382340487,
 0.26161433971546033,0.8148665015257553,0.0023533242840437785,
 0.9249086329900733,0.8991243801674824,0.21138644284219987,0.7827461205773584,
 0.18074740285105007,0.036574450105360756,0.5808938592738967,
 0.2669912891252372,0.6116429750863718,0.9813257166702878,0.08604539785906029,
 0.9059950843327235,0.510523087100049,0.13864669405556396,0.1881381469495128,
 0.7185604634332796,0.8575462986810743,0.9002802170404742,
 0.012119346137389964,0.325256031191997,0.580460348123799,0.47908861287548765,
 0.2680590110410712,0.6420419779222666,0.5322451570577135,0.6835553996360488,
 0.9159312394994041,0.9985970466896599,0.8174126314025205,0.8596854496917876,
 0.08266541037795683,0.45555622321586664,0.6635873266135539,
 0.44402555518843323,0.23438297688615772,0.5884492777332551,
 0.7253167214225442,0.03413647220778948,0.6262741139359179,0.4405522215065013,
 0.7035492211946257,0.30337061221919337,0.438605131096808,0.49811521492474264,
 0.7633611649315039,0.9213826540104644,0.5859403856861772,0.7151993406427823,
 0.06445110412977761,0.6224671821941492,0.9477902577363221,0.6505124964764767,
 0.7465062282081641,0.8613843866852341,0.29664628801265125,0.4285052228643135,
 0.06120490855625982,0.12706321434021817,0.6166963028283717,
 0.8765679839442743,0.06990188054420399,0.3421310579616317,
 0.47786502831936306,0.14701474596564124,0.0648792077576903,
 0.018022010305887237,0.09733764924007182,0.3299165324329085,
 0.2347408402752884,0.6644553083970277,0.5538007773756527,0.6350387533578131,
 0.44991091700465946,0.3991209901699988,0.6168738718304745,
 0.39666354310258456,0.5213496972306959,0.33779914215058227,
 0.8319762724409338,0.847731790469763,0.8932317522941515,0.5312580408785438,
 0.024981007874748662,0.8329975872863256,0.9635239628320903,
 0.2659562565828214,0.27322653606869973,0.010514522867657528,
 0.4578231244161015,0.10409315201156588,0.8835266484286982,0.6003421365984651,
 0.3513740684629908,0.999593496060108,0.17787428861524202,0.9451212671609649,
 0.9149472492918864,0.21667044928757728,0.7419196070143381,0.0751737922283423,
 0.4095604051213224,0.5610333498847403,0.1949524294131345,0.36718499000499527,
 0.05471211432640466,0.10873678387998909,0.40710404090084573,
 0.5975297873142467,0.17956279686118493,0.1441346532023926,0.4502069197765659,
 0.5282587413528486,0.4403424265481185,0.7966162357240935,0.3327172045255816,
 0.22094566754441625,0.9853280980238104,0.8736861707973554,0.9084810063789945,
 0.996636692432466,0.33063091743772055,0.8561406546336551,0.7742021764571937,
 0.5588662401734473,0.8675743293627274,0.7388685434293676,0.09645105040570134,
 0.25790744922932174,0.07674585717216209,0.2516833955122215,
 0.030084586696601523,0.7502422977283524,0.5758962399266059,
 0.37960829514634864,0.2713508075334482,0.745561954662572,0.7747691643945055,
 0.4538336710450829,0.19471554276782255,0.8227023783184689,0.3504089519652026,
 0.5826862574986765,0.15280484771079172,0.7429084072683686,
 0.37140414319175075,0.11303372668184974,0.06534940892476981,
 0.35992380038098215,0.18587454409791082,0.2736814802407377,
 0.6188219137088372,0.20278733685490347,0.9344164811541851,0.8898007003402245,
 0.3438073936459842,0.6611848556360682,0.6596833817308342,0.43749023242353735,
 0.3925843494835217,0.17526251036073615,0.5184166892373439,0.7970423073942615,
 0.7987052054125845,0.3556336090325093,0.8401684332720795,0.4808681925626155,
 0.8648909925052722,0.8786173943567445,0.8631888553623552,0.6170148662357704,
 0.2586469382035468,0.776970128752838,0.8242398690682521,0.8562470549866043,
 0.2406289408545239,0.5880145566621201,0.5084928986212868,0.2737468040580666,
 0.8250935848984347,0.5408355423501853,0.6679411495066803,0.7476680548663023,
 0.8304413527492438,0.010228984204387448,0.20210310182678648,
 0.6600154537685312,0.4384234531108455,0.9567193059897161,0.13316598049791883,
 0.0037581800568888823,0.7869249954038994,0.6938542750966551,
 0.21360248067597654,0.9086009037090051,0.7118963150829114,0.964376847335767,
 0.2898348388365025,0.5640998814709871,0.7252508359544708,0.4431369921207884,
 0.5053356669073317,0.6673246120946953,0.8875337506919644,0.8520321561489261,
 0.4506594485248877,0.6590189916494873,0.266756023026274,0.11925629870921783,
 0.31867374565519846,0.307549768123796,0.4168851795383365,0.04932080434884667,
 0.41354100211043154,0.9751647801144315,0.7208815811651235,0.6448821611035648,
 0.20372541174652292,0.17056812675855682,0.30124757499634214,
 0.5076490929901648,0.909596920485776,0.13491246599068418,
 0.017186326080863012,0.38158889710807187,0.9995999765201137,
 0.5522435830819509,0.9676564294832845,0.04025067457498932,0.7470071128030012,
 0.9437339976421903,0.5123815336292653,0.798925615593651,0.6604793252052789,
 0.7360534764998085,0.36626291686492185,0.8729785392657892,
 0.41440673211767953,0.038281151354414567,0.5758868544923011,
 0.5398668205167676,0.4499982686561146,0.7637455041519542,0.7727142169089848,
 0.7124337412300396,0.6015614552388142,0.7239615751416862,0.6992489669698622,
 0.31637158988190217,0.9019618843500298,0.03368812121801734,
 0.4726898109150637,0.42964721509486914,0.6431585030209405,0.7471757738072169,
 0.7550952569882866,0.5313002688507519,0.6952423611960259,0.5823998067849118,
 0.22061103219871547,0.9960963794435762,0.4985182573868464,
 0.46805422144759734,0.029199989757104805,0.7087738724631507,
 0.07882162386911706,0.8997385221443684,0.8937913752730577,0.6484211042636254,
 0.9124287861069609,0.591977068974175,0.7798040151767152,0.1096001912865433,
 0.502149873387138,0.19704564393476698,0.04172795792611983,0.5091362463451515,
 0.8844662751786527,0.2765028842662506,0.359224339369813,0.7390631314745768,
 0.5493695951086811,0.327568873397852,0.5444975947574787,0.19011247111777363,
 0.6932730454600676,0.5180118076074429,0.08609118074604782,0.9561253445561034,
 0.30525477767187004,0.5952367120700817,0.41696820068118445,
 0.09407579496954854,0.032315429129013085,0.4253937276917805,
 0.4095192343221673,0.7316010838990724,0.5729592180235868,0.36202443897837777,
 0.02240969004327098,0.6006403269221876,0.45022400172528454,
 0.38172066736044263,0.511667656559281,0.23949885804218496,0.8734260999710532,
 0.28176485332063084,0.020099806210066795,0.2103357148334316,
 0.6062653503103623,0.5093689027459288,0.664446546160015,0.9446783607346011,
 0.9900792720221536,0.904744770671744,0.7758256313047729,0.48315417262277216,
 0.7949351617863643,0.29606846629064953,0.13581100869346752,
 0.3250826351578432,0.7441617350555908,0.6780434740160731,0.8809155730446238,
 0.5051404382604188,0.8456689004993746,0.3740409819095437,0.7263749163442146,
 0.535808325373893,0.5831158268380989,0.8335341318830254,0.17266497476912146,
 0.6496100249853548,0.6790104141719779,0.5381008779310347,0.2108422579644038,
 0.9407062663893695,0.02984831557310086,0.2754537705921605,0.335886970522676,
 0.9971709222214451,0.8161439996597056,0.3743382615460584,0.5117386699511953,
 0.3806911643938099,0.03417740288102333,0.31710988312011323,
 0.5874980224373084,0.16108642277523755,0.5372917834797812,
 0.22759731749952483,0.9745753476832486,0.8529845466482076,0.648567498163426,
 0.8715572684582913,0.29691265971204384,0.6693007619963077,
 0.29652884992746065,0.8239970270951814,0.11260345055158894,
 0.06682487220067368,0.6610158052122609,0.47325376539060027,
 0.9064130875390823,0.9333358446032892,0.9782840692057886,0.5735918497955701,
 0.7329663816964503,0.012575813506291444,0.8578091961241159,0.412534092988161,
 0.9963693359854034,0.8558951537565928,0.6957523669724401,0.9877000070246792,
 0.8152514136480151,0.559085782199688,0.24303052530028002,0.9605276250333967,
 0.2819434111049164,0.06065571795443925,0.06819061611858179,0.495619047295216,
 0.16533431738237736,0.1102374504936644,0.5526285009303551,
 0.48224850931673213,0.8416970278158047,0.06082673359338564,
 0.9886865263620367,0.2305967708078136,0.6064772229813191,0.02199658599477361,
 0.09724144051484862,0.8159145420080355,0.27678505029465506,
 0.4204966329893157,0.03650296821812504,0.05560409013020351,
 0.0373847533067897,0.19536000446041057,0.08483177633590855,
 0.013239007974579886,0.6297749342040202,0.09522019391291758,
 0.19657251000360165,0.11463586371856249,0.05600486689066275,
 0.056243531534047486,0.18830889333847567,0.9237360953454573,
 0.711028583512317,0.6855673035111922,0.5888670464595106,0.6517890226483541,
 0.6040337908635248,0.6698642609911951,0.4925448192395704,0.582776023906976,
 0.8206405125642141,0.3231413259482703,0.4727055737683943,0.7005737377774757,
 0.1973095970363935,0.38272991751784,0.846717056549414,0.08040668535592932,
 0.8977806850133314,0.10403649401533599,0.10042481105099565,
 0.21086683470851775,0.3007273756221994,0.7810721725250143,
 0.16014023283238854,0.3124924072610722,0.8742231335554166,
 0.10062089631272098,0.5226583980366922,0.6300712263363778,0.6635911775580222,
 0.3886761491997194,0.44932126091410074,0.9326562892695501,
 0.05337658166572423,0.7014128820431508,0.008437157396827932,
 0.804933012323241,0.8709249787123924,0.2605256026713102,0.6495608245984996,
 0.69985735115348,0.1204171786201842,0.3304796476404025,0.8325602225857031,
 2.58949083295934e-4,0.08431145381024652,0.7741959489242689,0.459300901354572,
 0.5647378574037808,0.4151460220202833,0.03912757386619026,0.8202472951489155,
 0.9335838586050661,0.8251955388321709,0.5515765167053458,0.2559152292909955,
 0.45231612710089797,0.775196514263482,0.303533779905658,0.9164633528526751,
 0.36354858615307173,0.7758614810244118,0.18234843325248717,
 0.19891606931011674,0.9036162434930425,0.11442130617335933,
 0.5759234829096466,0.6924774655256948,0.8690895027011732,0.16143954074680378,
 0.6738935371459266,0.3078619465942528,0.7624468876386163,0.8151051947689891,
 0.6691151262167578,0.9873052574375424,0.23389905330842964,
 0.21979038291048159,0.5413917042961687,0.6012854047034342,
 0.04672180296343742,0.06934829115572427,0.4054752952826304,
 0.9208441602563209,0.11149166772297314,0.02490569467436987,
 0.14606048939284622,0.6725386210918015,0.5282303047529396,0.7231059604667541,
 0.12510582512243418,0.3616248831124379,0.9979175763298693,
 0.026373794345725066,0.22360644662987383,0.1039215679619917,
 0.6672233611792057,0.45715968383611205,0.16637078741317568,
 0.5422939091810265,0.2003128324847282,0.9228536801751234,0.38111459293300176,
 0.9000729572656865,0.046235935911222215,0.17782911282657377,
 0.8410296226594622,0.006709490293673737,0.13827602631747538,
 0.40815974258387844,0.6608521480584073,0.911410762163737,0.9366914045227179,
 0.2727569505358861,0.5514769345356446,0.6701094312763846,0.8623671763718571,
 0.39933953650399145,0.9135177855904293,0.8009181643464299,0.9651299591131625,
 0.7654967206327679,0.9248934764617899,0.3797577163364314,0.6970988763313394,
 0.8291494326223628,0.49155692154773656,0.9092064420652353,0.6738664595748638,
 0.6259957367396205,0.4972008423437815,0.020556458442043013,
 0.5697920448449066,0.21135492543871837,0.8679839052836114,0.8449973077247429,
 0.7119483657660945,0.49349765603153006,0.9123020196597442,0.1408191095957536,
 0.8920017567520386,0.6936928171493109,0.46828055378758593,0.5128283800031268,
 0.18920519141041725,0.45197510519739925,0.44803806888854447,
 0.935460399589314,0.3461387384334029,0.3686674247269317,0.9722166629487239,
 0.6953259124681793,0.4453143997009581,0.7868351901096334,0.7173449847506529,
 0.36517233683943007,0.7751118603645505,0.4577221198742101,
 0.19488675068834405,0.12444856017725081,0.8830213268630828,
 0.5761789490994031,0.053911906504349294,0.3725031239303568,
 0.6115634016722999,0.3501568040725579,0.46827281518153197,
 0.19550881574735268,0.2546914854758202,0.43270239602313953,
 0.06032549091319328,0.8019642200508477,0.3223834192607675,0.8069831267318972,
 0.0972813695313306,0.9836750110522812,0.18240884745663022,
 0.39294321656516384,0.43007760635812053,0.43167547884375024,
 0.4259538284051453,0.43660654824680856,0.3927201513439623,
 0.35533092536969024,0.40420488229056395,0.8240569729740399,
 0.2254213047691671,0.9663903508532896,0.3073649539214516,0.6131603107248829,
 0.351766049147779,0.304387936017538,0.38444260815670594,0.6466334215205854,
 0.9237514838338252,0.33129996525526995,0.4327891041218126,0.2965478670741819,
 0.7789605892001594,0.06905872450815742,0.2648614988153103,
 0.09821732496564362,0.6168202813008518,0.8236755602795005,0.3977644828986363,
 0.5709447575168386,0.8349708354082218,0.710679904194819,0.05468117181741783,
 0.4922779451469733,0.09428667692452597,0.1585870175110934,0.9458165997729451,
 0.5824216429084004,0.14576774899801148,0.37756044587648097,
 0.5419792911042843,0.2819394196638034,0.8651922375730335,0.917901173156259,
 0.8502742108482172,0.04012835905073775,0.5857168870742315,0.5637408326979748,
 0.28624145186202654,0.7180653496840568,0.7551360216862175,0.6305362569412444,
 0.8268504470333677,0.6172869934245563,0.8052738358384133,0.9323108526561686,
 0.9681018644186022,0.3733444825438397,0.5084880760782354,0.30826179510731233,
 0.48780999415697446,0.48796766267674363,0.2712137439134664,
 0.33954871002278897,0.11856131954804816,0.005931305269369358,
 0.3430014324290107,0.11430286196295891,0.3076860430345445,
 0.14942509672255155,0.7305591719533748,0.01235237783594878,
 0.6536141121856038,0.11011324973971526,0.6765574565997768,0.5505811782051685,
 0.38523791785552663,0.8710566968228535,0.4752564415076057,0.8852747393469655,
 0.8681145401142046,0.11958395873590688,0.7958812639712887,
 0.16913077933708842,0.5386573096600672,0.3711962494295924,0.2012776707687728,
 0.5091465494570165,0.21704775508283636,0.09797547863863354,
 0.20959360384103476,0.4363064034854264,0.03345644279284232,
 0.3220549238434067,0.4319575763458815,0.21824824497111606,0.726396862163025,
 0.7792708268955574,0.5927447941476527,0.6237402193710383,0.7554797725369667,
 0.0315462393170296,0.7883194170661312,0.9102312723145062,0.37366726027920083,
 0.3821902915458546,0.7833414800577362,0.5581398012731954,0.37848511076534563,
 0.6648588391584064,0.8981225475090778,0.8341704193413616,0.4054068094207207,
 0.5993356038111094,0.8444675295870236,0.5062251315071129,0.8790558130798887,
 0.6604623682742848,0.46197914427514253,0.9612957677332846,
 0.07450425588529253,0.7289817870078055,0.19057413428848646,
 0.5085172501281804,0.745581450928171,0.2147149169993796,0.9654543436582534,
 0.4636414834647893,0.67433808780081,0.12059192889041226,0.02494975757102469,
 0.8484350284131091,0.24605639112683952,0.5834619971701056,0.3392469798591492,
 0.09106982710818157,0.6322423656202583,0.6909931701713363,0.4736823313252706,
 0.9098490443546333,0.15296930747427762,0.4352525578722043,0.874059057408779,
 0.7355144528762847,0.4045506061920767,0.18395000801041217,0.672724058662084,
 0.9273614078803024,0.4804939985589878,0.21443682176280132,
 0.07282649250826656,0.32305760560174324,0.9969027453374657,
 0.7629412879708055,0.9409951113015409,0.4510446560938868,0.7152691145916908,
 0.8174285303780413,0.6144262920696733,0.16435478887125177,0.9170239039771535,
 0.28271974162209446,0.3186942877241109,0.3718057833395655,
 0.07244289610436838,0.5225896698296465,0.6925463415885539,0.8172875973535725,
 0.3254684898869319,0.9404085963812006,0.9731916647705657,0.25141805724496435,
 0.4343699557692613,0.913667625601637,0.8476634884688573,0.6647574094712236,
 0.4158544464723024,0.5249432197468239,0.2887634984775139,0.4922755411179631,
 0.3225465390570714,0.8694896006029658,0.5014560142821272,0.8301478296853583,
 0.8003712296642752,0.883735474806915,0.5793478591857266,0.4834548475445709,
 0.7577026253183112,0.7227229904090365,0.21382850041469814,0.3129021150505923,
 0.9269637281470128,0.2520338551574559,0.7512626465659161,0.8150890958975513,
 0.7478033485449114,0.7064420033908936,0.6111465455916165,0.32276037613399844,
 0.5879275523198375,0.7121770162002501,0.6473508573004677,0.636796878029286,
 0.8286659815108294,0.5099019446739299,0.6715886162131877,0.0683505618423681,
 0.9254953289740101,0.7821912780573402,0.3529089617505139,0.5218464863393758,
 0.44006267684009204,0.4993482085857799,0.837850061559655,0.731789260934716,
 0.5980824598212299,0.06499469114869938,0.7945476336100334,0.2121720086981551,
 0.8920098273908272,0.20699451629858112,0.23470406195268356,
 0.5334356684854616,0.2105213578906241,0.5982172867708654,0.740297881179866,
 0.3585628904490705,0.9131446610782337,0.33929608686578305,
 0.14844660255203257,0.9858334448632992,0.5520869596032466,0.8142177264712758,
 0.3439481422971712,0.8305645517571124,0.6423536197978517,0.9071307015649633,
 0.4451949563291516,0.7601170188990024,0.5769516083086476,0.5217054935506652,
 0.5971583450318794,0.340685955287112,0.9621426233329073,0.9783677798233591,
 0.7143899998687573,0.20786001736596527,0.854748447722339,0.15468098318523793,
 0.14349766742244352,0.7512856402195902,0.1965502779815539,
 0.46890107348821775,0.825376152741459,0.828966527135745,0.6461311539919977,
 0.46164395619054577,0.02944250583129926,0.8176226625823548,
 0.9423312946034359,0.13703225637827043,0.2571761861938333,0.5462000671062353,
 0.39685889836512445,0.6043321630999301,0.7521702256989328,
 0.18514492328575471,0.005298740303145921,0.41478056902190064,
 0.17136044592012278,0.8377069825982421,0.3604775942216727,0.4023139545832254,
 0.9481364117756047,0.7574578750740086,0.576759660977338,0.17564820996760133,
 0.578442052055975,0.43122696041636477,0.4436453473590324,0.42895663890359526,
 0.04975475870267976,0.5326510746767799,0.09189594549825575,
 0.4625032175116254,0.3498424422317272,0.5245635259493036,0.7414345426085348,
 0.7076701981704347,0.08564073183874243,0.701581741867823,0.3277699602411447,
 0.1489546851503667,0.956434997652205,0.17948647444932675,0.09947571530666255,
 0.5749278174956728,0.34067679142347196,0.6112972881232398,0.5763908904086262,
 0.7643789473214694,0.8379115542772689,0.5932829207994927,0.16996039935984175,
 0.40434240265859855,0.7428048068187157,0.15762335016762385,
 0.9333414389663305,0.40967005476182816,0.43704212410651677,
 0.4680306965388018,0.14669818339294283,0.2145929784642161,0.6743996261219665,
 0.3613568257888382,0.2003297138445267,0.20940905244752095,0.3362742426094646,
 0.6114104165643817,0.5619060858586218,0.06333577231686871,0.8186308617763234,
 0.8621375547434881,0.09494571444476385,0.5541709738180981,0.6809535951146496,
 0.8223031734080759,0.863478250437681,0.4875374075561142,0.44772051539932356,
 0.4090761018119301,0.4058302411629937,0.6703522407986524,0.06952260389338116,
 0.4540220395820723,0.47783378923073083,0.43159292707337615,
 0.7440208566122057,0.1889601794504998,0.5597080490791846,0.6506357764125603,
 0.6287970363833846,0.36133875562062534,0.4979230420845481,0.9688610423890619,
 0.06970366716792542,0.29518817776328055,0.5000759035613458,
 0.18402467436292913,0.9081175741618513,0.2729425683169735,
 0.42186661983922624,0.6351944414336033,0.4442099430643385,
 0.17122777292819635,0.25998311067344326,0.954777364599623,0.1959064160355115,
 0.9155163184451479,0.7349955174838861,0.33579571084018234,0.6642242314890517,
 0.12485351726449712,0.08061171789317234,0.026575370060623826,
 0.00924856920595174,0.6199268713251067,0.918344696876112,0.27419790019981427,
 0.06081462206464339,0.5922126689875846,0.8977116278840993,
 0.44174213809813667,0.9497353684944215,0.9038682782055086,0.12230716564805,
 0.5302600448282242,0.989819280600226,0.5123834062688433,0.7698522945155393,
 0.3183941059114709,0.7924138096929636,0.32726759027432906,0.9922010158728943,
 0.4836857962008012,0.31509402646967044,0.5444703550226141,0.7108775947144952,
 0.005830657055270527,0.49586234144506647,0.32688415586615516,
 0.15310071174119633,0.4411317545386726,0.9912728479658222,0.7210326144816508,
 0.9582309468058547,0.6102089735018219,0.551115657866089,0.8667430764503271,
 0.7645400168070051,0.24142705350078775,0.4617862657732634,0.7364799242209957,
 0.5397168516149438,0.5432316353009935,0.9059135333664508,0.9397639728358657,
 0.12910716330088323,0.3013901237746599,0.9742863089818906,0.9488303297742926,
 0.5249159391485572,0.7444995546793596,0.07301708007103291,0.3601400656249555,
 0.9507561237554245,0.06801677479782609,0.8681156215175057,0.6879808561906505,
 0.3921296126581323,0.7398672358962932,0.3645086079057146,0.20721121867960546,
 0.6134354077813717,0.6281828186403248,0.2925525300235865,0.36737283104665475,
 0.614971495166031,0.8198019170291211,0.03779966802530543,0.27871875953101877,
 0.3152916159780559,0.8461712463513748,0.8693552534301875,0.49816038302906396,
 0.4696086457027031,0.8863573528647203,0.27559454282469287,
 0.17415926599982623,0.47739159119631513,0.24326917140063675,
 0.061024706185582644,0.6176817085657147,0.9560035108739737,
 0.08341621139881261,0.11984590345761148,0.2333634142015446,
 0.9189434263324684,0.537771936109039,0.34730956656986267,0.10551849848140638,
 0.854283345482528,0.3660006605341882,0.39061619075041354,0.3967254945725428,
 0.35567356993737065,0.6971095886345378,0.03266293637763895,
 0.38626680368247623,0.5808923739362393,0.37185610400410507,
 0.4270683103206697,0.98859075802411,0.6092050777708496,0.9323565079901174,
 0.45429300815078655,0.18970770928263114,0.21225578303246306,
 0.18947147388203855,0.2640312610371358,0.8491358061132332,0.2292835755642061,
 0.26524400254649405,0.5182597765373154,0.2868142494028867,
 0.02518653806220328,0.336033844226997,0.9962757828329175,0.9651738321777943,
 0.3986832878544666,0.34146618805191553,0.1146199871808451,0.5005807317471355,
 0.05322358623551582,0.24464234970112597,0.3460905155088291,
 0.6046637572899718,0.1068889022489965,0.864654170259905,0.18943157069353578,
 0.3003912777835369,0.913888451184426,0.4361354109665605,0.7311794925135323,
 0.46429000066276016,0.2990260730501162,0.7852540573797339,0.535509671631724,
 0.2764907637618791,0.029654670980158726,0.0613035950681351,
 0.5994725727810464,0.35631800128868174,0.10958944319399966,
 0.8707356329494886,0.26310604248389,0.8611105073800548,0.818597365901232,
 0.18251620494040421,0.8436325838853839,0.7422112895258685,0.844966739748415,
 0.4966086956034774,0.16568437288928795,0.3796363665657686,0.5449003328070691,
 0.011846664846662525,0.9584806158979623,0.4362500528445947,
 0.5920993326922217,0.9251144373593745,0.962073368542213,0.11133202897106897,
 0.029264401513867222,0.9499772516597471,0.07150061190438695,
 0.3694143407695566,0.42016649068513445,0.6166275566271189,0.6942881539258269,
 0.9512083867523454,0.28218663424502277,0.05695249064825647,
 0.6381126900670554,0.21432545322314045,0.696074273522687,0.8413171656936673,
 0.8504620796794461,0.42524500128481335,0.1659527135038804,0.8419362716232675,
 0.16510008987126412,0.7786630775462678,0.09373698118734508,
 0.8612013227844157,0.302031730413456,0.8676331989198621,0.8251412071129122,
 0.5469333904434215,0.3619374679623335,0.4520260854579704,0.8403458731191532,
 0.9091040401762869,0.748963790038451,0.9941125133614619,0.480813147110262,
 0.36371951659418766,0.37189451299531606,0.46074744399689427,
 0.8329949446425375,0.43435515988656714,0.1456489193422863,
 0.24924714034381812,0.11162081681632563,0.7304333995280088,
 0.3434531047946868,0.9758870766838544,0.25944938430180486,
 0.02719883713997584,0.7357036448731641,0.8088420659790859,0.6157304521265244,
 0.8763970448395093,0.9264182826782619,0.31572643979616044,
 0.034154707292893716,0.3068604448984886,0.08543702267250364,
 0.8162406642021038,0.45938868717178205,0.43072937041440973,
 0.5040398760118694,0.397072319149139,0.4586968029981068,0.4419288842190917,
 0.6200783124762044,0.6351192253613256,0.9671000753664416,0.1586192911514347,
 0.3117469398534194,0.1409106584837494,0.41052456244215196,
 0.20364043970824874,0.8843969765831516,0.6744879215303037,0.7432202515638212,
 0.044635205520251464,0.22995325786151732,0.1459273918961277,
 0.5590741064774037,0.002079635061101648,0.7801979488113882,
 0.27439274480478215,0.9162874400652432,0.6883513039082918,
 0.43262095309646154,0.5309231806355648,0.8295391915232647,
 0.12331063313033974,0.18445743402354764,0.22280001816043282,
 0.34809435784242526,0.9809599695905118,0.3627443273711689,
 0.06626263986657877,0.9705978854039796,0.08748349976884062,
 0.2994321712045671,0.7665461115944495,0.25049934639385196,
 0.15506592617939563,0.1014242907630164,0.7798149592392223,0.7826208854199529,
 0.5779031723120207,0.21674761010474697,0.3326927744352933,0.9808602526357452,
 0.34066829602036075,0.9982700724864412,0.4121143886260724,0.8069048833483164,
 0.16345569228597756,0.449639076716035,0.20015462900195435,
 0.14140568669211406,0.041819364350338506,0.8033734043525828,
 0.1513983833111977,0.2795696262170426,0.9107450716493473,0.24921422556593886,
 0.005570257118886257,0.4585706587813152,0.5818850974997463,
 0.7769890269496955,0.7656273555536958,0.0763712017548287,0.6896526389725127,
 0.40729656147079674,0.3687662183501189,0.5814097097114048,0.568543185901815,
 0.3542768887280412,0.008638157897500887,0.9543739077346893,
 0.9064815460374689,0.5840512186112309,0.5600643018250427,0.29053865591463646,
 0.05322208786682869,0.839742171724306,0.40740465159472683,0.2754995903712949,
 0.11879935152465493,0.3236103735535033,0.044994361461217824,
 0.34342330489598305,0.21216576061817272,0.9633495918758939,
 0.1074282058263184,0.6003782432748468,0.5186611735454616,0.04616095574036949,
 0.6255286587397492,0.9998573070318942,0.4686695377335086,0.2922492069798972,
 0.01258789599611565,0.40045034763588105,0.1982194390822265,
 0.033094269690374545,0.8454050182811902,0.5814710753247258,
 0.9489950211159688,0.7147677097402362,0.4378707645900879,0.6617137242222921,
 0.5139387352374908,0.34702066971118106,0.9741489320114441,0.8424880242020033,
 0.9488860031157493,0.6449513844945791,0.6716832451100199,0.39459055874499727,
 0.3753135774325551,0.9893145993809985,0.044194778103285315,
 0.05193063644808371,0.03751421557542045,0.3686910726449162,
 0.48876613829088167,0.5070858287099137,0.23374079637281164,0.564840751334794,
 0.7064843474249118,0.6160887283053087,0.8884638264267669,0.8634047894336659,
 0.9835769346343385,0.2034581778765081,0.008270932499178318,
 0.9230976057247976,0.6593386591215016,0.12119044715688798,0.9748241802303516,
 0.08459580280202728,0.9221320328704208,0.8937495630684043,0.9855139490659146,
 0.7520391794517821,0.3861560766101303,0.002989931988242267,
 0.36697130438636316,0.39564729278139754,0.33553294396228617,0.48453795116258,
 0.23853161434699421,0.09916070688291023,0.25500729909217623,
 0.32777424776110653,0.05813328709543775,0.5450993768649663,
 0.8640595338367251,0.27402890115612244,0.7493422558780408,0.6882858442738492,
 0.661063662080024,0.26950485080786546,0.08269383827660626,
 0.15029854944216414,0.33775994292987477,0.7910106306362947,
 0.7362637207239724,0.980537535105533,0.09888464020835341,0.8560485372621484,
 0.9013681119543767,0.8625073501713488,0.49893853260662424,0.9051671143510579,
 0.649205319575925,0.6113807728685541,0.16200808256349908,0.4595294436706706,
 0.4345488913355766,0.6587536695320171,0.6769419343955452,0.3829159372940628,
 0.5666785405124097,0.6238801680270306,0.8499058726752228,0.8231144904861104,
 0.3923073223090381,0.09280169236173064,0.16121785903290198,
 0.21912456672225233,0.9299958764674341,0.35223046981813977,
 0.6737817555254454,0.9373889516666213,0.0877434226209447,0.1008052563817925,
 0.014740984206585317,0.9241384517477088,0.8312439234921503,
 0.9607865871203479,0.3651511012506773,0.634729845548046,0.13502880768831726,
 0.20672175986713526,0.2718937029298668,0.6008152717183813,
 0.42066829345796064,0.9890539077174694,0.8339417904309379,
 0.13617276388954225,0.5032967335289698,0.03229137071177124,
 0.5168488137055336,0.45164919844499285,0.10439455320492252,
 0.9666886924053529,0.13618922648580045,0.05968104599730739,
 0.13775010024668422,0.21600046996293887,0.4927118634451402,
 0.0012563934144913524,0.5540023047356715,0.2944042538457803,
 0.06213917127807511,0.3752248414448356,0.8276535813353604,0.8792413490381682,
 0.9891258085398885,0.18379863206428926,0.9419569445382562,0.9581259272154284,
 0.2167258723382398,0.6782238205350617,0.027372755788372505,
 0.26466800841130855,0.6490340708162491,0.192334874450521,0.3526156462740846,
 0.963841287427921,0.05063999369455696,0.6153965653237213,0.4761405977444385,
 0.3449484494548499,0.1292451898678585,0.8213306313059279,0.7691924598804701,
 0.44334184803467136,0.3660037372132451,0.5633667210494743,0.822799151615161,
 0.011746791354824149,0.3124931691713877,0.4327780418067688,
 0.8086469402251184,0.5642463173021689,0.16661746162529234,
 0.26426460910799787,0.845751776797062,0.9421576319272773,
 0.029199827852404825,0.0500093416472176,0.8394713752163777,
 0.9904576655387411,0.5161386932815689,0.7506775611670276,0.38469061001935756,
 0.10829362467734172,0.053400619596649546,0.9668489136809362,
 0.9243269342039235,0.6283161216637915,0.48358186927980795,0.6441417172397066,
 0.4026737701362013,0.6437634336691512,0.38295537179359695,0.8469872226580027,
 0.8152266933173263,0.04626558480675447,0.5016139879546336,0.8549205479857492,
 0.5933621156349062,0.09750639834329,0.7851740070930543,0.7204196287851429,
 0.1273592467341782,0.45677946685653037,0.8297692870549441,0.7561518983476276,
 0.8562010085830579,0.5516147576431455,0.32237499454918395,0.2921508986926925,
 0.9565309309107082,0.9842304313769407,0.24187772831621157,0.7827365337082024,
 0.9409121964728832,0.9707095563167535,0.1942503695293687,0.4137565203906911,
 0.3663910612261878,0.4171477166798271,0.5323541917041824,0.49512477335394267,
 0.5915589402793454,0.06044957636323112,0.24976099483916803,
 0.7331220384333121,0.21608510516527457,0.8468922593160166,0.5758461979459517,
 0.13935983606251734,0.6919494046925954,0.9170397108261524,0.6979645948187674,
 0.3161100610730285,0.6892005196140694,0.304780775486148,0.8247789347191791,
 0.6994649471245074,0.41032594572105763,0.36907740337868555,
 0.8373957891079004,0.808199454265796,0.5256894165125645,0.9183414513801647,
 0.41406421206249,0.6836708268574028,0.6483049904456785,0.8030657356013747,
 0.13709768277105638,0.5478255236155759,0.8818923353711704,0.5553180344260951,
 0.5701024477597623,0.2108159583610819,0.6112478257502464,0.099582524980116,
 0.869751338513535,0.21852639919897587,0.26185661916337166,0.7971770144895334,
 0.716371606601256,0.8470472575544168,0.7183795993132285,0.8659465773095244,
 0.4403718111506322,0.47832640467210763,0.7640971050983157,0.3590584694800618,
 0.39259875210777184,0.033127960813878254,0.5938596793365112,
 0.48600610891721285,0.05682702395082817,0.5654903669922353,
 0.1226894562422749,0.02428165737634702,0.5659563354699739,
 0.47927384063562983,0.4957760100375751,0.880469728276454,0.9227492880639493,
 0.7938716281592555,0.14920348554486917,0.5069673997777197,
 0.044980529980427675,0.05453630209293425,0.8777106638297233,
 0.17480017270820525,0.8664078969372414,0.4467612313492749,0.2575557941567065,
 0.734652863169544,0.954679708708672,0.3494506105235272,0.7674870919351953,
 0.9898525423974891,0.044826562065245934,0.9393447317749395,
 0.9854979296601531,0.05498197377009428,0.47571801504085465,
 0.6023297827799843,0.7852019857186081,0.239633254758985,0.28539457758756637,
 0.4938010594810891,0.6964404276028128,0.8123483424821426,0.4953071228306406,
 0.7550652730235974,0.6383129120587142,0.2524101865214097,0.6803303136213374,
 0.7771287918787294,0.2924253502494639,0.7064661422455343,0.10506745993804345,
 0.9528228350651679,0.5227167047078828,0.4875938499184933,0.2900391732969444,
 0.9683427639497154,0.7572294351564849,0.0967092032477258,0.746690525689843,
 0.353392067646338,0.4084323950447494,0.6108824118136582,0.07671530091590562,
 0.9454647638137315,0.9413400787249635,0.6183146156955129,0.29771349178384665,
 0.2705435904682242,0.26237479773583383,0.7907540038615662,0.7541931074253811,
 0.8083799474910502,0.3670516446962162,0.8405682723892643,0.5720087527577589,
 0.2194480301155851,0.23398191472624708,0.5336745764757732,0.7567996127527626,
 0.6377088182053111,0.45335011514783374,0.13529346398916198,
 0.5237405214489986,0.11267752101059414,0.9505158588280476,0.4367739253047591,
 0.2770212898210829,0.5515214160576589,0.16499059128521754,
 0.030366297727587255,0.8301994543687541,0.548525847444473,
 0.39993763118200565,0.1507088682813743,0.0805665788107991,0.6569346722366437,
 0.7139330865108694,0.7312942371558757,0.2690217243655717,0.6280101000194092,
 0.970959501503283,0.41810228204756594,0.9674093929791298,0.825779944570791,
 0.19137644665791465,0.33186400319344944,0.02888815149984225,
 0.5653868925785466,0.19722798667810693,0.6341391243652543,0.3466189110870904,
 0.02387181638809266,0.30205295762635975,0.5469956648365634,
 0.25822515941783775,0.18852881580754965,0.1386877101211197,
 0.8986143419270118,0.29972522155374937,0.2469697041369564,
 0.10521698935404766,0.08113568152147232,0.2560022800220785,
 0.23177809507063962,0.24157736561097776,0.3742559360278128,
 0.010644879448522215,0.3558499633956903,0.6288756563831939,0.893112830463569,
 0.7632518796055328,0.9554444366875494,0.9778824894842126,0.544571034039289,
 0.5359142194372617,0.7224670703674986,0.3665994924208439,0.9361969514886274,
 0.34934961342533477,0.8513979804279329,0.8158797232280213,0.5311079234714455,
 0.8182603213316748,0.644274717422988,0.9346156873091144,0.10843774853323163,
 0.12221033986831864,0.19925774078211278,0.23216520137610264,
 0.5019324171167814,0.8629857853867563,0.9004110606921131,0.8986710875463672,
 0.2170332409919532,0.8430376630799996,0.9902040216063184,0.137246787536898,
 0.04121721039540982,0.4729598459948945,0.61800821678838,0.33423397209193173,
 0.7264705546217057,0.3483424015973955,0.2599572281793642,0.7333387390328888,
 0.4747000068107481,0.6570953251047497,0.7984325147291736,0.9207766434412441,
 0.44835577502486457,0.814102196944681,0.40497517421722284,
 0.24572860642916594,0.2962064498931075,0.4828895989370091,0.3692643163487376,
 0.620733982785566,0.41808951385505744,0.9405374193612569,0.10773526938956035,
 0.7730935925917454,0.08851102600207972,0.7892895830641518,0.7987301311501063,
 0.2822724551811119,0.21537935884497406,0.5744884224207055,
 0.27809435991744436,0.5939048326202592,0.7019065329751819,
 0.009688908800209117,0.4966822573200368,0.9943724915651247,
 0.7781861175600168,0.5706084918491102,0.7348072897897874,0.7177342708014094,
 0.2295436838709234,0.9044827393387098,0.32420250915529136,
 0.19211552690615397,0.009526505486085468,0.26014666406866027,
 0.21342766637730481,0.35704858757277447,0.030425682225837303,
 0.7570407503611226,0.054503646328499356,0.443920297344927,
 0.15772389398264108,0.24185619299525274,0.5850606229207791,
 0.9195285257800219,0.09160189065555846,0.7198368633380018,0.2546573556559101,
 0.16298851793835745,0.1507706528254481,0.2873856052989644,
 0.44431220223451473,0.42883579085639467,0.6057335479572323,
 0.20247068122316658,0.4978984363591875,0.22346557585872762,
 0.7412150759126683,0.10764194938213167,0.29737441574877654,
 0.14808799663643302,0.9138784695815544,0.016702801951946844,
 0.7847659872023354,0.5836803384450676,0.5768040153674322,0.2602878365252852,
 0.10759299409359202,0.8928873496876286,0.16805398223144696,
 0.37287779839485047,0.9598899394293109,0.8948934960336057,0.9675450590958244,
 0.19880174482355106,0.49580772480181545,0.6052841820666165,
 0.2948538444505391,0.7520192694851525,0.06916039031270071,0.5924371532145002,
 0.5800115345877164,0.18680644940565405,0.12520481638656256,
 0.39654892778633266,0.9527576403499176,0.6129770366535303,0.3279105588170246,
 0.1471505653468621,0.20103288374160588,0.7800639385972963,0.4120855005190992,
 0.1436421034576172,0.3739643832964423,0.373256074848031,0.17420921072426188,
 0.49797098918873883,0.4803690492387055,0.3562386986659001,0.513969719929002,
 0.26001444729078194,0.42087473192413294,0.5312271095287107,0.555552445289527,
 0.2997726678913186,0.49783276966799617,0.44474420640588985,
 0.9432167076062273,0.5047666730081198,0.21917491014585044,0.7291740301923285,
 0.17512364482044318,0.5894897484913459,0.9469915149255391,0.9807664292044536,
 0.758318426840053,0.6247103474789172,0.28247789409032165,0.8750702079884559,
 0.1031177514237771,0.44578937047096123,0.7609060121378604,
 0.33895386893732704,0.6292010593214328,0.9741996891377345,
 0.08414518968471929,0.7437771217538035,0.3267768026678487,0.2373778825184497,
 0.646321542845779,0.19168494561019267,0.8638609050929169,0.7291166913455133,
 0.4508938938499849,0.004787739682636238,0.6060726311174833,
 0.9780829434114748,0.23252568307291455,0.4595062740034024,
 0.09362560144352172,0.9210706237831194,0.4703963677021399,0.332202154672804,
 0.7531903916172047,0.40742397778788897,0.9521103321924987,0.6714517184482593,
 0.9882138986049545,0.2179880572357571,0.6304436908914268,
 0.035032537328068436,0.772737150735712,0.7784186242788087,0.6272916787762197,
 0.1927541174585039,0.8955859554530778,0.21349404947158745,0.7071665184070635,
 0.3207764544603128,0.5068783116128471,0.37307227190110426,0.6548965819845642,
 0.7475063790841792,0.4971799695645135,0.6620632878896306,0.08358017667401596,
 0.21751696930763453,0.23792150509820253,0.8493218506869158,
 0.7730441431797777,0.3220759463867465,0.656325402397047,0.3450549115675786,
 0.11899271194185512,0.34512963459005075,0.42118946924316125,
 0.49931497920473045,0.8883313117037822,0.4258168296034135,
 0.06663119652628957,0.9853008645021496,0.8585299395374149,0.5230450858438289,
 0.41971466120130285,0.6830457864433865,0.04432621141530013,
 0.9907921953492553,0.08574838594440104,0.24656211195630306,
 0.9685345367684449,0.05398034188809808,0.3218063310524458,
 0.40108545362086745,0.1470800866935561,0.13337334430296488,
 0.9964697624253376,0.4278356372671557,0.8392966167386152,0.9995833030853012,
 0.557633034413419,0.4966240546895546,0.6281143583879568,0.530234960217185,
 0.5382906559134488,0.2953200237681807,0.12371738168513124,
 0.39513661830393787,0.8730566758073006,0.913927998572499,0.19614235928379953,
 0.44263424825634845,0.04368685376346004,0.22056730591746387,
 0.8501081509867344,0.4567978932719785,0.6607873164983719,0.336907219946448,
 0.2265918039150876,0.8735641577785999,0.36443674782568614,0.2948844321780728,
 0.4874562092433279,0.8701704729110374,0.5083559777465096,0.9912668857097066,
 0.11200312216790587,0.6240400761229511,0.770462959789245,0.5905183346329151,
 0.6773839690997122,0.7838179378268668,0.6506515579228209,0.541443020274203,
 0.4013384571648374,0.7989556414308466,0.9559680307204996,0.20449743417320154,
 0.767788739357268,0.23761853971849944,0.040352668685964765,
 0.7253776576802912,0.5983779007752977,0.6058950614365243,0.1996869865389781,
 0.37114083943172405,0.31258278737264444,0.40064679623745614,
 0.04696989164524634,0.8426334331824118,0.5147421830420379,0.6379180532437541,
 0.016063176234462517,0.5389352518848236,0.8972087382961588,
 0.9901307377121463,0.3104595816418896,0.24636946897000545,0.3603856740487289,
 0.7740102625436676,0.8227653275836496,0.8879584594043533,0.417238334146547,
 0.8928748108251553,0.693826844408643,0.5061563259193806,0.12858963513704735,
 0.9968995984219132,0.5656582588350011,0.6543825499511668,0.7437810881955336,
 0.19739484172810495,0.7901972135548994,0.6603011359007331,0.8144074516041075,
 0.30839331198408326,0.8275989312123356,0.38962786016224893,
 0.2936885673570926,0.13518109219159535,0.7442283784392322,0.9470778073527579,
 0.6487045947109207,0.11225210141825814,0.4752502849312058,0.4801006379498731,
 0.8404736273245628,0.9496849269329,0.7319070184848319,0.08624601778289054,
 0.978471651875983,0.7881376263002396,0.5682442267025314,0.6458080464704198,
 0.517768081997321,0.11637477088372017,0.8275195876623657,0.6892365465938697,
 0.34241327814728184,0.8292336299748999,0.22095145291996932,
 0.6665794560488265,0.209922097368204,0.9657715163868116,0.5217607503805166,
 0.21171508758014757,0.1976103740936439,0.9833221341498282,
 0.09610937808743902,0.4901706354825106,0.07433727267795498,
 0.12314732722671862,0.2525949449873398,0.6145049467710171,
 0.01239908073564755,0.9870487451083922,0.6604922243537402,0.8012613800565033,
 0.5180112468757972,0.7916541734711425,0.17881072355165029,
 0.10021038644281166,0.2836365867510118,0.39348991995919824,
 0.2053130189354191,0.670230700990051,0.6160759838609979,0.8195443733544238,
 0.688455602870349,0.7786090164449425,0.6889020768462235,0.2497109944100866,
 0.16796727931297828,0.5583699067563554,2.896428687950703e-4,
 0.733163588060695,0.2148316782852372,0.8404497458266981,0.5582568229107642,
 0.9287877418254006,0.3005883876176316,0.9584060966324504,0.12563937229422795,
 0.007621392221675238,0.40547976063580227,0.38794511312718816,
 0.1718104078902486,0.4568380642757921,0.4473491669312647,0.5834821840229781,
 0.3116120519771012,0.2858441282567079,0.03384353851393229,0.4167397829647794,
 0.9347149789220639,0.10979302328681861,0.9625002490583718,0.6622379173584422,
 0.4726948032180207,0.3551351672880497,0.6099064932070012,0.8173788142257161,
 0.3232923371423623,0.1963358831622357,0.927513416262142,0.472213724855031,
 0.8782514869264492,0.8047071174341158,0.44132395272387803,0.8030499200372406,
 0.8305433109027478,0.11647019162163175,0.601235191569555,
 0.012962846031397524,0.7142361570002902,0.4487365717703484,
 0.4630356221900678,0.7739249427722736,0.8416227580402347,0.9552076187398048,
 0.1787532774649916,0.5398575697402658,0.20258578226949342,0.3904052949349075,
 0.4557073637948508,0.9914849208789225,0.3030717721105085,0.9105175404300886,
 0.2879471665272795,0.9207435274109428,0.6381509516977335,0.4991172643279842,
 0.06728423951834217,0.8949050387303282,0.9591814302791358,
 0.05503772866603418,0.9612370748065344,0.5336588564372176,0.5161777532472427,
 0.6177506194829403,0.6454122021748295,0.8683498758522591,0.8529604989092062,
 0.38882829421565335,0.4239401631170653,0.40358902573893585,
 0.9996984345988911,0.06878128638203163,0.19967551681507623,
 0.2415470539550707,0.5854141054890052,0.18664079651783372,
 0.29700713452455696,0.4119233654136458,0.14517012178381372,
 0.8291911529152434,0.4837634081295279,0.4790578660038848,0.9609463634597074,
 0.7468940468410277,0.8738721242621706,0.6232795013976469,0.38534259043841634,
 0.7385329632009218,0.4711751837256686,0.1342840687253375,0.8764330050055451,
 0.9386854053076289,0.9715977143062975,0.2893874842383334,0.20606778877002419,
 0.6058383918767182,0.06405119965286266,0.978914989388571,0.8707967294200645,
 0.19594616680185595,0.6681422663538243,0.8832107553635753,0.62916244349932,
 0.9769997430367462,0.7698203609930082,0.0949616025206168,0.6831593006970963,
 0.16172871650284293,0.09013397915605526,0.41809741646053167,
 0.9799916644848449,0.5007373986273981,0.07757446336145324,0.3350306852894238,
 0.7362202286523848,0.8742952123931631,0.8579088453522621,0.5064275006379773,
 0.027955196026882345,0.09742150941547367,0.08084973618035551,
 0.6063544815053171,0.45694718486543284,0.43934093401132124,
 0.8592509909189344,0.5857834050686133,0.43482900555596493,
 0.07956060186533243,0.6592655299519679,0.30280446536749983,
 0.9241529070561992,0.12054084575750812,0.47695172393159135,
 0.5183723269028441,0.655787560623529,0.10487441900905659,
 0.032224882325807225,0.1032614015135257,0.5276003959712358,
 0.7019842351053197,0.08874051888687018,0.6778414716591012,0.3763751173617441,
 0.9459422250931997,0.13779873144508548,0.3476966926021131,
 0.21301929222489346,0.14648784559669026,0.522346504150165,0.703865829575105,
 0.38829866973823524,0.9137160070888443,0.31856890065704035,
 0.19268330118235177,0.37196032066936535,0.7369889502219652,
 0.5738540686066379,0.6262782913513746,0.32011727166722326,
 0.02783866872930818,0.5707925998710653,0.9028848067658924,0.486595806151795,
 0.6553628305158451,0.4753586785611774,0.002765834760244501,
 0.9267643282808241,0.15521181160941688,0.48504440243622515,
 0.8297472238446455,0.849296040948469,0.813631498612623,0.2630732138671549,
 0.6847424152880088,0.29006652441029523,0.31134097163263375,
 0.8171991485428047,0.3656684033499965,0.27415615898319556,0.6353034294503624,
 0.731596620895465,0.660349768780292,0.8966171600724315,0.7101517015791894,
 0.06369627094355468,0.9204347844713053,0.53721637098609,0.9361175805500561,
 0.42882886942921306,0.5201039353179588,0.5437840114195358,0.9624376458475901,
 0.6873756175249776,0.13511354530621722,0.23915009917872887,
 0.31031449385593546,0.39926834534253375,0.2501109801365844,
 0.27122572333482897,0.619182431053096,0.010372610597612852,
 0.09574126519014503,0.9132152960866278,0.5909402711486802,0.2150687375367365,
 0.16303310152597572,0.6733973861551359,0.8834316490985364,0.2518539072182441,
 0.44373139537779505,0.9194130153244163,0.666965248172839,0.02879386318566357,
 0.019368331750057255,0.24577033738646015,0.8621891171806594,
 0.696951685636105,0.401370135647075,0.9849366981942437,0.36194243664260883,
 0.5776240226478807,0.89093521164446,0.47235002921867475,0.604936379593875,
 0.10616834625437932,0.19163786929753623,0.06122518544684796,
 0.6230945036432342,0.5410263595873064,0.9402985756899106,0.07044323189744062,
 0.32327789980767385,0.46685286228793577,0.16837322544578004,
 0.8772292925951172,0.7730771835477163,0.04824349221981261,
 0.13380916283200128,0.0015361065893171322,0.4082248631150228,
 0.25307710346041523,0.22802494259987272,0.7054185839363655,
 0.31964009091896084,0.6009414162163574,0.43587376105430087,
 0.9078257250171986,0.7207616911613975,0.29680108845579334,0.8115424201515354,
 0.9691740095931819,0.7842972690409239,0.7454232441526533,
 0.050218360841139065,0.9004589513192787,0.82248049858107,0.1307917026212786,
 0.8769201815595697,0.17861381418418598,0.746158631317548,0.38995774009369666,
 0.20854679975753632,0.5792292832123562,0.2106904097010034,0.2864354277366832,
 0.30067113897888376,0.31191149101504034,0.8201231017130597,
 0.3726048011643377,0.3980623749727603,0.6475306247306614,0.10287998487743288,
 0.7328928486892093,0.8099114174789265,0.2017830495872185,0.6286409411309524,
 0.9556847537469627,0.9005814659293564,0.9658353157908108,0.13135342490992163,
 0.08755609768227235,0.3347511937692391,0.42130023817843476,0.783738080465875,
 0.7530231794938383,0.611814213127746,0.003937551954177887,0.6068969142572962,
 0.03563828944384506,0.9428568116165233,0.18259492606756333,
 0.4439039197705732,0.5411240004293278,0.2606022726086453,0.7767801403674768,
 0.629848355640999,0.2665621453168232,0.6559982006954325,0.7563947317302516,
 0.9942347427271974,0.02824600013098677,0.9862214139319624,
 0.09778415078008329,0.4004041129473568,0.03072927281563176,
 0.3649162539167302,0.9866398413208898,0.3564058404451109,0.9402965681621824,
 0.7488406433271929,0.246815672088608,0.04583156371910824,0.20723082845056717,
 0.9353211307644204,0.030997864175967282,0.7910849900315444,
 0.7595259371050866,0.35566224688248593,0.6317340735357824,
 0.41542435692134205,0.5537725333644248,0.0015536407659385576,
 0.2206146238736426,0.590098283116322,0.26903178467378697,0.3181774264413626,
 0.39283721276921724,0.8437407219526841,0.8139388060716785,0.9262250956093402,
 0.11975411807199687,0.7139953680078592,0.20884904141035487,
 0.5219310164311312,0.6823094994505797,0.5258509665899993,0.2893560765827262,
 0.09751514551594531,0.5563198048164087,0.8669013203453855,
 0.26019651266806465,0.1658020204370374,0.6106875544291246,0.8513509166887172,
 0.3782363296577741,0.4663455939820481,0.4992851153899561,0.6882538827075212,
 0.9072797072927995,0.9836889507829579,0.3697709701481655,
 0.008741292395223366,0.6267376745891646,0.1291650888483592,
 0.6967793714339932,0.002949849215808986,0.1627955609397026,
 0.2514121004115384,0.20000458297395918,0.8642877264607451,0.7048419595699247,
 0.3181822721077172,0.5364458260823111,0.06833028439884203,0.5290256409755348,
 0.7901311411275466,0.2834941684461928,0.07272754397635006,0.7000467209275455,
 0.8610032546896462,0.36869907661729695,0.45093201548434403,
 0.4218281461774066,0.8113193197955368,0.9910945766432195,0.42997507171374627,
 0.7040810861718725,0.42063769368909165,0.37283702785805306,
 0.40735069570357796,0.3436441227239315,0.37982669722573104,
 0.44279818076338395,0.33137767996930023,0.6536336379914216,
 0.08259274442566089,0.3004704588502175,0.764593049325304,0.1128092282837393,
 0.9548649478944216,0.87193178001327,0.713853188409332,0.593721319023208,
 0.45388670160392286,0.30803527638624706,0.6911197237714075,
 0.09180509330761943,0.20312704846927554,0.24364800518094143,
 0.03725095584388305,0.9298616896980612,0.01600097563643421,
 0.6935614843336305,0.3177387951180952,0.4036910688725417,0.5121262351855964,
 0.6968905866628793,0.07882081805873864,0.5976569144363787,
 0.034119091298018045,0.8038532264887241,0.7428654325299133,
 0.6845183457756491,0.20990485290915117,0.05212752166602441,
 0.6242504474530731,0.9465866139491915,0.03973353423697579,0.9678511427278054,
 0.12996642341094733,0.07748270965197124,0.6247155315833552,
 0.0790017734045596,0.1864166091074364,0.14242442776054265,0.6873760923005964,
 0.9104045199026155,0.10364690504591267,0.20896778488025913,
 0.7003459066837183,0.9576027791447643,0.4875783171667514,0.8631601701706637,
 0.008565858985458696,0.3834639609002026,0.01971695351569358,
 0.5120694612357114,0.4034669464984564,0.7983776635310307,0.6851808031533113,
 0.19211951792323922,0.3891030414125155,0.2433877002168382,0.9930005661204614,
 0.9534469739555753,0.3631002700156062,0.4866134294314882,0.22367196199011463,
 0.12906587488880983,0.7286755030163072,0.6458491883550226,
 0.25187210641245217,0.8250018586433507,0.8266029802377787,0.8522482248699391,
 0.2579686700669631,0.2715353997786999,0.17504004782547555,0.7128987127290247,
 0.6272592125703815,0.27746139892770016,0.3097037233122919,
 0.13473918215266445,0.6180486060115005,0.4804672928938085,0.5083968238298079,
 0.44768955410711764,0.37162080216947835,0.30200045563755495,
 0.8790892660683984,0.1506450730791441,0.9835268106453054,0.45381460787462224,
 0.47763741686948036,0.30221554196844325,0.6097114208224069,
 0.5481598035015254,0.34673809712777737,0.40986252389913636,
 0.11437369287958155,0.9430965173573758,0.1652260966384811,
 0.41282413063525125,0.3385236084044383,0.7936297548375469,
 0.014249255454986787,0.0648038114827506,0.4822829124180711,
 0.13275926968700968,0.1550637223129141,0.952885855080666,0.34703366951760817,
 0.29386576941372444,0.003950924571064052,0.24287678844329197,
 0.13877687993434717,0.8368748881684296,0.01453779755713891,
 0.1813863870764877,0.281603847530012,0.9785967695930503,0.7520205461662735,
 0.6587912757310187,0.03554392259532424,0.9191576315300436,0.5727642649066255,
 0.9488809707866261,0.7818705369702902,0.8938545339130853,0.6046372949123382,
 0.1666359075317505,0.14240854053431762,0.1402117951556081,0.2548096075865859,
 0.9458371466828384,0.9557237088121602,0.7017467944880972,0.4869703417209622,
 0.42095272743749845,0.75895980806579,0.6150467420926897,0.36573756567398463,
 0.09792422302141579,0.1425428946661118,0.9551936447358417,0.5669701460279323,
 0.8306992472431556,0.9247039193091703,0.3396297997096853,0.2839884368515523,
 0.9553133200299064,0.3685456074644402,0.9200845400536611,0.19740272775691148,
 0.6956302139899789,0.6731302410088388,0.9521408696563086,0.06676003267334552,
 0.1596680651755843,0.7093221984411039,0.9278871870492339,0.6923126557053386,
 0.9915171216901972,0.031039109085026384,0.9156996405527232,
 0.9934179989182035,0.03058273263549305,0.2198594995183405,0.9672568986721107,
 0.6121372610455378,0.19992505303071262,0.18412385699101264,
 0.8035300768458803,0.8464079520324715,0.09832292430584166,0.796759576900477,
 0.08355110893448925,0.9323425021983529,0.4404327377978918,0.1416759053256228,
 0.25196460365747697,0.6458539193796385,0.022435042048660758,
 0.8897813062179334,0.5781801956948156,0.24430750141168622,0.6096019296049883,
 0.7142586556668103,0.12077844196020038,0.11810609284886675,
 0.3010840164010985,0.661568837542026,0.3607425230931778,0.9077441661600176,
 0.8976389759678123,0.8648291439808433,0.5662710184738909,
 0.058290179612289994,0.6652828531161128,0.2542466039683653,0.692177348815456,
 0.8693730941851985,0.6300468063683433,0.10328681528028172,
 0.21378849151157608,0.6356595456850136,0.42428372068793474,
 0.5102205813327743,0.7329737584800503,0.5821275064751082,0.3057080803157888,
 0.4409588652543742,0.3641353366101666,0.18039063120571552,0.9172210420016335,
 0.3416259336847054,0.24208201602470714,0.5087643158015264,
 0.15638040218342453,0.9004815064000184,0.731701739464134,0.9413913250056963,
 0.8457534154106497,0.7390784830757927,0.31404253188185516,
 0.08226432940561623,0.9139268290492677,0.40941005235966255,
 0.27537368962094166,0.7562149229369303,0.7334194315470239,0.9220828483963364,
 0.1411894202540882,0.9150723645491872,0.6935825103860905,0.04531888929147254,
 0.8850164770573223,0.6310386655507563,0.3724909904422762,0.40417409785170044,
 0.7651984035080127,0.23602331844925895,0.12169638099309044,
 0.3256167116716937,0.26579839049448806,0.3772821570762386,
 0.09795076057027108,0.5386184532570563,0.9891920160076022,0.8411509257672815,
 0.41841805113806896,0.028123030663096626,0.051707146217480404,
 0.2519532921621095,0.8290941871064281,0.6339625348850391,0.7463411983545553,
 0.3419464550449922,0.35461605388440387,0.26782378619499614,
 0.2396036061976512,0.40058258300213,0.7826890319260903,0.8365784825658757,
 0.21416291684389355,0.7683405616446968,0.12163680833556267,
 0.1343338148127683,0.736836759492217,0.9713938077316842,0.061556932696922306,
 0.8218651716345673,0.6953524549748898,0.27207515889917444,0.5200408340844057,
 0.03688789204280052,0.8108913977698567,0.4435216077125008,0.6537024005014309,
 0.9078121998473293,0.8214269466052041,0.028659443631832482,
 0.24985928345404518,0.48473649862995116,0.27319992073791166,
 0.6352648655747717,0.7056258440808556,0.08709278241358687,
 0.22835166352188496,0.33583547270311376,0.17167835717322955,
 0.3353805558771823,0.5880396788040392,0.2618081625401263,0.33130714982180354,
 0.2392184038355365,0.7820486111769056,0.42845228160327264,0.6985838787073668,
 0.9349074652901407,0.23864721801437705,0.6689256606474064,0.4469879272247963,
 0.8841137019003569,0.3813738438676286,0.33810011230194414,0.334088867324144,
 0.6352320719675365,0.1479711656753574,0.9543564447180145,0.7546536725431747,
 0.8873975031096716,0.5248021842644145,0.15966455627431309,0.8307806694549357,
 0.5311098228477042,0.520429477148127,0.9532210777789254,0.5329200091223397,
 0.16993630126595793,0.8297933242436714,0.10593424200051138,
 0.10587503447970514,0.26494745263698827,0.1633627193707723,
 0.3934205765390911,0.5089008699331302,0.31382034265192027,0.7078932143397335,
 0.07934332210406014,0.29498753685165613,0.11382934008958767,
 0.416507830075597,0.9069035253454832,0.28397387906882043,0.539234193598382,
 0.5007419249450559,0.6771523372951125,0.03484338624811634,
 0.08400936912531332,0.632074301923619,0.5372728964494504,
 0.0013556588392733548,0.03816992894580584,0.30593303849221354,
 0.4208492314053355,0.7857222874168807,0.04124522340527137,0.5582993152726026,
 0.8933479033871596,0.6936697248380749,0.6446377327294809,0.8629710792838348,
 0.6830951092244217,0.8963305344367705,0.03169430212625468,
 0.23205655190703123,0.19302667958737108,0.572697850009878,0.999031658456151,
 0.40545909083643705,0.8199884208472394,0.4787900910755948,0.6925781547077811,
 0.04907880354428751,0.6891053658268538,0.3765339716073106,0.172784004195762,
 0.8281517383381634,0.2738933241823963,0.24927794229516476,0.3896963600554628,
 0.42369374908537005,0.4969261750586611,0.5163975688025467,0.90166344827113,
 0.09583797251015691,0.7827360900713185,0.297731554717929,0.7241005442750498,
 0.09075662804334805,0.9584390515949612,0.17885538593960193,
 0.10733168423431139,0.9836703223604378,0.07698379501929398,
 0.6339719223058456,0.9011494399993697,0.6491065209602537,0.607121296458453,
 0.23088238267437033,0.5352377740744687,0.93325399828149,0.6678476768053798,
 0.5679452415358199,0.5972169068209339,0.2129686418446145,0.2577064320203818,
 0.008568272441062952,0.6458093958963835,0.5573279938178785,
 0.5542776408947387,0.4751102900027542,0.9581450525191273,0.5721835621225925,
 0.08264918047674297,0.4870074684438829,0.8377194906993586,0.8302473064224433,
 0.1788179147175405,0.17658157320368195,0.3654866988374481,0.8283669286552511,
 0.5188136353821342,0.23376950141343578,0.7895918667821265,0.5141232339715157,
 0.8122776693687856,0.8875707495812493,0.11505577540656675,
 0.43804786552272157,0.891781087937769,0.7412644869043292,0.7050591077966725,
 0.1816638387462166,0.005747602078370551,0.8895673012001305,
 0.34098538020733715,0.921438499959129,0.9647228720314674,0.37999028094750265,
 0.045393051862888445,0.8968799431933965,0.532609707722199,0.1883190076649487,
 0.02307051534917759,0.5247748960657601,0.12279598200650732,
 0.20939702951981665,0.18630499073803009,0.7257783482764303,
 0.5728562826905306,0.11795515227775122,0.8967299760818634,
 0.47289832172548885,0.5851173281085958,0.6043386537325115,
 0.13840386658135062,0.9479380824716098,0.27627735682613364,
 0.6897721415043109,0.05626832138442528,0.4541115872890251,0.3174570703882291,
 0.4910647601722442,0.28711060318226833,0.29176225312459947,
 0.057560581393156784,0.10017154967649422,0.4242830903204836,
 0.8213604151619406,0.798273569492327,0.47978875167058765,0.7450804916697955,
 0.9152959662573016,0.66231933381209,0.6752700345165142,0.5816874750246441,
 0.16068728835923773,0.9396147839659662,0.05005322892781894,
 0.8121927788501336,0.3379741004435617,0.19642602064032877,0.7795842267376278,
 0.2071651447715136,0.26700274050336426,0.7971798063378385,0.3928889077427493,
 0.10677841797570697,0.3960661174743715,0.04057520903312706,
 0.2347077852719497,0.4918982690069369,0.04479833656382004,
 0.15395906556988392,0.32740416078372214,0.20828430229331196,
 0.33198047882301607,0.5831130625041148,0.7100707805697803,0.8481112439507854,
 0.7074486604710886,0.44864043678616783,0.822830163394551,0.9503273774382007,
 0.09711915198508725,0.38476763977349615,0.7450782053416294,
 0.39552357332973453,0.0014872748603094177,0.406740256085684,
 0.10303045039780279,0.2405485423045831,0.4109263003334076,0.5269147224278619,
 0.3028415388239186,0.2888392339905269,0.6577816335504291,0.11875131067741185,
 0.4781676441676248,0.1504831669723855,0.02815737410180219,0.7915675197803719,
 0.6148401505374843,0.7444137133657226,0.3004561682514467,0.7075804291819854,
 0.7310077725145822,0.01027321411677884,0.09463023543774485,
 0.9197520558463694,0.2533126559359081,0.054496096120844384,
 0.7488305212029245,0.13885625499197596,0.9887356380179411,
 0.07354001650232178,0.1439479630430487,0.8450443006246553,0.3989205099737636,
 0.00660695422926505,0.0058281385844412,0.1004834231018169,0.6770889146721798,
 0.7738821120381306,0.16734242725293025,0.2537875368194328,0.9084134790296028,
 0.830250045307943,0.7512256651158403,0.4543372712132989,0.01604333073222497,
 0.4312382206749582,0.46016199863882024,0.91824884959422,0.45036458268167856,
 0.04785375424968907,0.5529714724418281,0.8955940161228444,0.6959807216121954,
 0.8261071620364892,0.9839664639750203,0.6079626809611297,0.6632310275452955,
 0.7613028588045475,0.7654928273577679,0.2687570445936568,0.3491292798283484,
 0.4456063780686058,0.2547750040046832,0.5788367484217568,0.1324062702212,
 0.2369263405267813,0.11896966005794518,0.4314920054213287,
 0.13440666634211706,0.6256711223931658,0.7478347609666703,
 0.33132470099412337,0.6544102378723425,0.10259605096869762,0.329352171562855,
 0.49302498437661,0.7679540079702608,0.3131517406896661,0.9776942642655079,
 0.19828063238875215,0.9339183730587146,0.6742470352827971,
 0.20880054783862612,0.2339474099611243,0.782396242024628,0.22716099304486048,
 0.8612278188498321,0.641874210961483,0.14394898345861873,0.5050053714192245,
 0.46117035388528893,0.021885851444448656,0.6132152289257894,
 0.24193138268078807,0.026508557134598543,0.5074394613100203,
 0.08565030553075037,0.7731366539701178,0.9104198870053937,0.9662922985719904,
 0.9379090010723263,0.43181944974323183,0.29765200229530375,
 0.43714393253060546,0.25046322311015834,0.5757244616455257,
 0.7346204502859499,0.8432600859434756,0.37385604612059853,0.5714655340371262,
 0.9505458270926486,0.9379998001636645,0.6243626612584441,0.5448907419677155,
 0.7943988350905532,0.26265193982446255,0.9579831059965064,
 0.26499286182940873,0.03363971136136634,0.8522556856993286,
 0.7836138704146608,0.9146564407006127,0.8541391540112442,0.8314518023269986,
 0.6005885398680734,0.7633545466581687,0.8464371924606997,0.3376150703795655,
 0.38562151156126157,0.94880229965387,0.09644919676349406,0.4070645145112598,
 0.8482268756352025,0.039785996012883906,0.30348236888122804,
 0.9034595135656492,0.4645880788784096,0.32275392818323523,0.8026367817198644,
 0.8740920044088585,0.20505824486293145,0.41699042870310066,
 0.14939972769676302,0.03680869172170742,0.44993970642901016,
 0.3632445548327675,0.7063425806167906,0.8386482515098737,0.18238095973855062,
 0.8514549591175625,0.3436704369859528,0.8327831474176796,0.5463520289158729,
 0.21363136465656274,0.9757733938833155,0.5021131899993055,0.938012699759575,
 0.34643529814284857,0.6566235357251855,0.2573308029522957,0.2877986608046672,
 0.24001400844978416,0.3086641050282144,0.9467026359849362,
 0.26318209378396795,0.08703681862082147,0.38623855673205654,
 0.09759290046556257,0.8371364974826553,0.730363823554875,0.19404919324661785,
 0.5668396035267748,0.1301613672022235,0.8153845818612987,0.516894110610354,
 0.18012624148163348,0.9511877671056823,0.6657173831399831,
 0.041109356015369025,0.1537608489767741,0.6987851437233077,
 0.5512326651260464,0.5774228711741669,0.8448196279068761,0.07685407315446013,
 0.6543184174423292,0.5533294338306625,0.1871391294037319,0.6154883550455091,
 0.1277364072200433,0.6282146712159541,0.24329433462454253,0.5919549867665991,
 0.44377396104317957,0.1939322637666876,0.918566736238337,0.45959430275790836,
 0.3752597807040864,0.8091039407051105,0.5180997253026502,0.3158328857405469,
 0.9548371877228299,0.22799847641509985,0.7503603465937513,
 0.31127399014399937,0.2554842397420626,0.34455995097334746,
 0.5637616877448584,0.19565495559428447,0.8551857508827233,
 0.31218259943851134,0.9013055910722169,0.2622229614224334,0.6838123460889807,
 0.7442145317904091,0.019079607946708932,0.8054333399024457,
 0.3820507226098133,0.7998644627169811,0.7368380726532038,0.30634312305492706,
 0.9762285873241996,0.536572781266549,0.8989185380262721,0.29448778522063856,
 0.2629675293154682,0.7225355816601167,0.123747212641385,0.20739527759736287,
 0.8393628752453329,0.6251675117006198,0.8805666373353147,0.8733314410776072,
 0.24404249458638705,0.01991242643727531,0.5014053859076661,
 0.20838296305045523,0.44927529109864905,0.4650749348333305,
 0.5453594694749848,0.15927173653100013,0.1241001933445679,0.6743946130379368,
 0.9816496443129347,0.04416903575272901,0.712170494600159,0.6140235972060801,
 0.7379127625462967,0.2174744547889188,0.3101904228358182,0.6439994385737088,
 0.005844972998495179,0.6399350627506147,0.39008497383067675,
 0.08018097560833903,0.644893112733159,0.7549071047589966,0.9358382398727542,
 0.8381655990806959,0.20763791683912813,0.1386948883420034,
 0.22397741584631525,0.3286878118468549,0.8271490757727407,
 0.23199815626900355,0.022721172812390722,0.7736844159146299,
 0.5188553587902218,0.14689439595791653,0.4891720128141335,
 0.004549023345459924,0.12910191966201134,0.45633289506768515,
 0.6683042819290288,0.5865345238052981,0.28585737564564595,
 0.03802736089500314,0.9557707411130493,0.9359301329890753,0.8689247984018269,
 0.6412537014089944,0.19099156706407583,0.12322732677877546,
 0.0637050452596315,0.37255156543472445,0.12059322963354313,
 0.1684632349634546,0.2066129474344338,0.2252659275204809,0.6312628688830099,
 0.3830607660054054,0.18769126962247784,0.6923873554632088,0.4436479689942978,
 0.14799867055720928,0.17595204446590007,0.3767455024291996,
 0.9674183475893744,0.8237944706762018,0.9051189697741857,0.8447803551528845,
 0.1327397444251026,0.43356141761473443,0.7483090031594524,0.9012767801992067,
 0.8297826373489406,0.28636125007027136,0.5779672185922871,0.9377046506880472,
 0.9910438447179994,0.12926750415457655,0.5694788352944391,0.5587607485155323,
 0.08900981564727772,0.6572167541667698,0.33458330480051757,
 0.12927994807919851,0.0077650634657890905,0.8916061218728493,
 0.22135039851850946,0.4501679648995025,0.451703152337282,0.40976786309708324,
 0.44492612702635603,0.9422940301074819,0.5224211895131211,0.861227378416791,
 0.3805470558584688,0.6759033684015833,0.034826841941986375,
 0.5123764817556176,0.47379638224447806,0.5943956742278156,0.1389534370116171,
 0.16115891048644004,0.4062329819308872,0.8001585011222409,0.7246313386975178,
 0.25911248924144026,0.17354567557362555,0.12308200612448239,
 0.7788204554061867,0.49055836058599833,0.6691611672838408,0.776062249637826,
 0.806197333283827,0.949472660484382,0.7394462005329189,0.9586237410238319,
 0.19308427420844987,0.33758884974620484,0.3190678836806813,
 0.41631156280758486,0.9208517521604297,0.2005588292114695,0.5529742878323858,
 0.5279740332436701,0.5282638473686507,0.07218177721121777,
 0.36213412312387494,0.6294687040594978,0.7490946851913091,0.7803107076990201,
 0.665763849139261,0.013124205197683536,0.41507767931527506,
 0.19116150496063555,0.8475770190501282,0.1287173545535104,0.6644987582162499,
 0.0030239240539656764,0.41798274168966776,0.77827210157923,
 0.3254746338365174,0.7435956806484671,0.5152926088999623,0.8242218683589709,
 0.6636670236716332,0.1625793339355175,0.3316680196951842,0.6298847498632152,
 0.5698032741206814,0.7468086149414669,0.39974833695441414,0.734515872665593,
 0.8902966752827746,0.5783118646772898,0.29072410709698604,
 0.38085582421223574,0.6632047900996124,0.46464919159629214,
 0.5581300370067422,0.5284332696113386,0.35321155167641094,0.2015054205366631,
 0.42043627085438506,0.3656709912941756,0.9543272584805088,
 0.23059087499616082,0.028814881754695265,0.6937113769185255,
 0.5997329461018036,0.8714512766778029,0.4669336129776036,0.7592415096827749,
 0.13099541799849657,0.860279826794497,0.2343936142569275,0.4169863449828325,
 0.3988926676956419,0.987468864134707,0.4718898822115697,0.6144850048968479,
 0.22964280659752678,0.5840270094745537,0.14993230281446346,
 0.10388289399515571,0.2156016488362349,0.4159940516238333,0.8210504181085037,
 0.03324497782814584,0.8871881842626357,0.21952497397248993,
 0.2878217037463152,0.49056938791992055,0.7020221784111697,0.9291374899768041,
 0.7374302182350999,0.5601873089314771,0.6211070838307868,0.8601636085340376,
 0.9318843700933997,0.5494138727120055,0.962355240279047,0.49992712710025133,
 0.9529434146560676,0.5032288931180934,0.4061859492149842,0.46442910423923167,
 0.819764012438491,0.08933369610635256,0.8576583701224758,0.25138932555946714,
 0.3377382970126025,0.09440027810799356,0.27986037746223746,
 0.008421082402572111,0.2041045201377566,0.9541717418929851,
 0.41167893200152494,0.9079368102333413,0.5317911382184761,0.0643476163505976,
 0.9936901533792573,0.4496366153375172,0.9363498118932164,0.9509707247644994,
 0.9718433770557549,0.11868254959063851,0.19284410423433107,
 0.04432225534381651,0.8408861153194541,0.14174443396341596,
 0.004201962330073639,0.33946889822056714,0.19763064498139415,
 0.3741846727926945,0.9230754509944926,0.3839806786971013,0.3759257825905644,
 0.7368585638944769,0.40094842964368327,0.9245124870213544,
 0.45534205118289417,0.6089450753941901,0.9964849446750719,0.7862219640731835,
 0.5570049732491742,0.40016592221647473,0.4641095862144322,0.8303509027087664,
 0.2790451898402656,0.38069501856335397,0.24350388706011294,
 0.08464245415374583,0.8896667818222006,0.8777850694037961,0.920399636966009,
 0.9690650795861544,0.16338392555373993,0.03293582095173431,
 0.5267465871426642,0.28752131747561793,0.1303006667726132,0.811019223903231,
 0.4332496987520612,0.5921741776000616,0.5369233059145859,0.8617779786539385,
 0.6046160995828407,0.2630790181599034,0.2516695061190727,0.33913409345413015,
 0.2527365133311007,0.5155160357241136,0.09077108128907352,
 0.21474529954508226,0.40355414284830204,0.11847887914976363,
 0.05980730284928315,0.23759418540784316,0.9568274015599159,
 0.5123179688870597,0.885204664933204,0.011007166771714871,
 0.017272290899148945,0.5668275375646638,0.35461907603867027,
 0.6373747670623431,0.8023978897940536,0.2915052843689766,0.3669857038302077,
 0.32229571113499966,0.08924211662244774,0.008228438976694674,
 0.02851032186483804,0.8002748017431712,0.12698272120557663,
 0.3401632711913466,0.7754836685033701,0.08246006833038433,0.1325688955347708,
 0.7686667285250951,0.5085706596427052,0.39417534829743295,0.4058189951528345,
 0.6972711892057076,0.3145428511984224,0.7589877186738112,0.885361178863755,
 0.5464944598502154,0.19782750897149537,0.6328794850044503,
 0.09600254214788406,0.4645126574034446,0.4803987866000702,0.2762651691502268,
 0.44524720083852076,0.4006448241660381,0.8269441245866035,0.7552704890357836,
 0.9659214760037916,0.37494982251772935,0.2794162681288548,0.5379642024564245,
 0.22310902834493018,0.120001903352841,0.9926818097632631,0.46871841350781895,
 0.266333136429887,0.02727794283712004,0.6018733150614025,0.943099541924577,
 0.8835059039591174,0.4574599133796067,0.9218256207151201,0.9865486462099039,
 0.9719176665900865,0.9611875147227931,0.42395042312750886,0.8200825337012804,
 0.03313776667916413,0.080501300690632,0.348172246352039,0.08672199777003342,
 0.14547671024259,0.8117729462622291,0.8292307814911056,0.3557605762143846,
 0.29321042796531294,0.3912465647152299,0.2919289634833131,0.952728527423862,
 0.1329076712079894,0.6923553155789598,0.574158472473762,0.8346251962735137,
 0.09587867615974144,0.8898076662103787,0.5975552387116176,0.9393368988878024,
 0.2776538172413182,0.9339660546133247,0.11066174967363818,0.7161025274315358,
 0.7904927315157726,0.6149968369179302,0.14460454970908165,0.5688922278565584,
 0.4959057797171035,0.5651596836721199,0.025268584905689173,
 0.34789111157103914,0.9885017317688747,0.07172219267996227,
 0.3827407858542933,0.8468246171472023,0.5552019267762865,0.7583039805763049,
 0.9780915153580659,0.3158742057920281,0.3717272084150183,0.40426963703518615,
 0.27506846581656774,0.8436884309471129,0.9520778280803038,
 0.17988910209637143,0.9988199002865039,0.8068062368545124,
 0.20648469142406156,0.2410799236349327,0.4533046171505264,0.9568915119721205,
 0.36444658541045194,0.5378660104214229,0.2865816258725491,
 0.30308495157422166,0.4094727416257322,0.7338195882482582,0.9633535288664217,
 0.11755721386946294,0.24730739303647398,0.316314485171453,0.9479269811482969,
 0.7233385947322827,0.706407209915508,0.7114974925722716,0.6225943891606838,
 0.8383023658900124,0.9602256461576479,0.05204458688160163,0.360027435010151,
 0.19809017025109,0.8948414542056734,0.8153512304706197,0.47145556560555746,
 0.456305802388857,0.9488190249493961,0.5750645991790987,0.3561825186559999,
 0.2748505244405688,0.4344414144471802,0.10022221107762475,0.470573684361126,
 0.46927443618414744,0.171890269423827,0.7349584955862902,0.3038114210614161,
 0.256102654017733,0.26256210476478536,0.752012758341535,0.03507952154462646,
 0.022606819993352123,0.3561510330343366,0.532721467014321,0.3239685498221989,
 0.3147943641162333,0.6555202611206858,0.8091346311080445,0.4180838177366626,
 0.730206603567916,0.32042439140418233,0.2758339730791084,0.791978652892148,
 0.6547518256204234,0.7398633111435442,0.09822325901660256,0.8565036888030264,
 0.9671257698676121,0.03428980877645904,0.7980663370598835,0.538557787149684,
 0.08301910538199064,0.34264201720471466,0.15473565015791158,
 0.8457615316952605,0.5669835097351151,0.7077610447765385,0.25314614520752843,
 0.9382145077671629,0.48389324805374223,0.05658282090155953,
 0.8811813304977136,0.019768521001210515,0.07984451842053253,
 0.6457343851775887,0.3617581887384719,0.8376295087325887,0.21831563094276873,
 0.29744193704120203,0.03605653583275914,0.19324277246349342,
 0.3064643750091607,0.9910375574789934,0.7817006970256128,0.5312701974918791,
 0.5414159109065391,0.12321884488717283,0.6560273298879719,0.913380980952008,
 0.5422667667653939,0.3136445023695906,0.41075783773673336,
 0.06980420618361016,0.17851040956537556,0.9621524373165691,
 0.8664620076503352,0.9110894353388926,0.3496450232885996,0.6557135394506118,
 0.5325101271822409,0.8316572816150557,0.0827196556780685,0.07037189203866245,
 0.830285531382976,0.6449612749044391,0.2713624426245247,0.37524198283759513,
 0.5333846958322433,0.8220019756759127,0.4054576835984329,0.5865930774370628,
 0.14657014849035832,0.34771321268885647,0.8840913159771496,
 0.12016958685427204,0.4840940086334111,0.819214675715652,0.12896688980660487,
 0.9924583706854903,0.3714055082175425,0.8432201089595314,0.501126812403465,
 0.26422997151521765,0.910816820753908,0.8494037201768331,0.8852689710548495,
 0.9488667821580704,0.9569074813841535,0.968142736866871,0.8983329989777797,
 0.586556544031425,0.006364544069467515,0.14936749412569295,
 0.31306334429573757,0.051635667751145586,0.00854583123780539,
 0.43192608803472765,0.5456950709979438,0.22467278651091993,
 0.9008196666605858,0.905005512672916,0.5904392605306518,0.3832323985425061,
 0.8278838219558791,0.518331139866689,0.006881746954945367,0.7934094709116852,
 0.047444319099177346,0.029265750645092847,0.3809228689650921,
 0.47147025535249854,0.12537231768217894,0.5055465589309045,
 0.6052553909690057,0.5929131014822432,0.661797100380337,0.1553075121363875,
 0.4534569968325053,0.6589281045836946,0.35751736698161884,0.566207646251703,
 0.8352353577286439,0.4463296293380883,0.815236838549374,0.43596639653992364,
 0.8021445866038294,0.6588848918194016,0.16957153131887237,0.2705798166578194,
 0.8043483147065844,0.7133261859488111,0.022073962736962605,
 0.38403796581990624,0.05059602833332277,0.6918375933517716,
 0.5726564592274586,0.4594206683161227,0.6494907588585309,0.9791935797866036,
 0.36197457855176784,0.10580432959452735,0.1193959606824998,
 0.6039499456630567,0.09518556357323649,0.8432364130789078,0.4194450522370924,
 0.2240040456145428,0.8648005352117724,0.1253206320728132,0.12191419734595288,
 0.5217589789588286,0.539212379807008,0.46750700693751224,0.14142581800776188,
 0.492844593860969,0.8864648556479209,0.07019690422950065,0.2870508237206416,
 0.8998409293348197,0.5333008734988893,0.22226479902774665,0.3167354919933707,
 0.9251865623683474,0.5070171616576058,0.872653610018943,0.32646487175108807,
 0.5404061644420042,0.2647221088240542,0.12899228369178006,0.6786124050892017,
 0.6575753471835833,0.3761153072148232,0.4479668377265471,0.5125954719792295,
 0.8126633864297782,0.3642240649319841,0.8239857492612436,0.7181919609713693,
 0.21064029800691397,0.61980370767436,0.4523651562492348,0.8767674650242808,
 0.05224591668706757,0.14512073334404663,0.9433475957545907,
 0.29065828723971887,0.4244219207490111,0.5844650105014788,0.8612681348815158,
 0.5684144368806788,0.35609413139676793,0.5247597489996021,
 0.33449764506157476,0.10966862371077613,0.35328671708373194,
 0.7615149277419971,0.5549603927206828,0.1108286786731082,0.9708384032519692,
 0.6693255010888176,0.1403117221378012,0.3685297676275291,0.9033130599154514,
 0.3538256794484733,0.4769778125760772,0.30128654969375634,
 0.23037983656445304,0.02668104492969381,0.4373885806691308,
 0.8640764510385709,0.07480487412952685,0.3397786047207445,0.9740131011447956,
 0.34155436836349895,0.11592195249056547,0.897353504752537,
 0.21600406971030173,0.9440424157719106,0.1727902912640813,0.512040446105525,
 0.7521563929503999,0.5994406587178595,0.3488088674309885,0.2323634947247588,
 0.29471181357156817,0.5178517837120441,0.8200194259459633,0.7316295001298737,
 0.8878415399096591,0.7239191319882398,0.46756273763584144,0.5343480551008446,
 0.39561312600036735,0.5926685507907538,0.057132704237858656,
 0.9400864221170109,0.22212151720975815,0.9864273363567527,0.5112744682891539,
 0.20492088642352502,0.10537861282266969,0.4426730789784209,
 0.004753345644663121,0.8730838653048216,0.02413028528801031,
 0.5870778841816764,0.6727857611109105,0.8807659908873924,0.9611176363428723,
 0.47573351494016136,0.23287246121067606,0.5929557566545893,
 0.9896834292167085,0.8373582858360271,0.3812379582351417,0.33162978134721033,
 0.5183535410418277,0.2746370828385758,0.26695826513837995,0.422006532398449,
 0.2455668866444024,0.6889034276483371,0.1621758735180776,0.673635313921969,
 0.18295093789429062,0.49518380581125687,0.3184169376431444,
 0.9590254324179369,0.49849127696555817,0.0027078702171299263,
 0.24807194011677236,0.8470922266264788,0.1566588756498506,0.579862543647103,
 0.6409993039789093,0.7716588557984665,0.1393105305680068,0.4251961713257376,
 0.0583556087137842,0.07582207606081814,0.0683258434453693,0.5607459529954718,
 0.9279440972407043,0.812650990696451,0.5437558252421707,0.20732222698601488,
 0.5880189280880393,0.10598373582923315,0.3121678865205102,0.6163142756627228,
 0.8504462078647472,0.9103076390407061,0.6824986239796014,0.45228261171200823,
 0.3402671644540777,0.38859581164847656,0.24663156162232158,
 0.3505166087858419,0.7230882548571297,0.044382413571207646,
 0.005884115919507327,0.5399169524927465,0.897844422342932,0.5343034185573994,
 0.5878528851208956,0.9492466588199641,0.8970037475860275,0.7426235339580588,
 0.6575252852069385,0.7260307441300518,0.7780300880169949,
 0.046134187147454875,0.1260438823232406,0.9879361126721105,
 0.9355888933751166,0.9682281048091237,0.19764112751837504,
 0.08091548639418467,0.6717153631512915,0.08515089495478834,
 0.6817771806798671,0.6420875456059716,0.6023141078665645,0.886039749248614,
 0.6410889137795881,0.01500206425525863,0.8633338445644992,0.8459486527763984,
 0.9785163302744403,0.18242032534745933,0.32849446512116687,
 0.9198814681651252,0.48244402008909004,0.43286297053984935,
 0.3656963007950367,0.0803183033859326,0.9080329327530844,
 0.045067112856564506,0.9518763128072603,0.7344081854481548,
 0.10844431369709817,0.6070312280336939,0.8261308607085133,
 0.06754109476355685,0.9889187524731305,0.8584900471655886,0.7416816826211796,
 0.1699917534001969,0.5991020297269178,0.35637984199076644,0.4035719208216815,
 0.3376454742366275,0.9383767171120814,0.3891946087406959,0.9755888489413556,
 0.9006522147962137,0.06034418882690851,0.6050221909288761,0.7685815583493736,
 0.24696569881829433,0.1985552154049608,0.564344695537138,0.8296092247324133,
 0.26627838025324524,0.9798562641076001,0.20827861535388947,
 0.8388745055824343,0.6704491933766682,0.5982553914853181,0.7481352874286417,
 0.5269952028204647,0.9689095996149252,0.5995848870075627,0.8399847923966401,
 0.43127733713674443,0.7402084725794272,0.9317736408328228,0.9209182782466864,
 0.7081160461875706,0.4331809214363992,0.48915014628180276,0.9300914704294014,
 0.8745647067471449,0.7464635621290806,0.16179155161495518,0.6162981729009309,
 0.5814377719564833,0.19994107519535398,0.3419122818861001,
 0.13895547121067486,0.5336417521498276,0.024239120179681328,
 0.7956641472530144,0.2870964503395482,0.2690886893831276,
 0.060104122069868926,0.27444884894341626,0.49796242499433796,
 0.15628573897508835,0.2666745928166506,0.21004939463316474,
 0.2008526842235736,0.20236956028086828,0.2777415886522232,
 0.10217488189491675,0.09236523045707046,0.7617411417530999,
 0.27314842693486363,0.8791049167622744,0.22162730020861543,
 0.5293741198510582,0.663820870820053,0.6045385662255855,0.05666442174122499,
 0.039994520471645756,0.04217818422995012,0.21447251349166974,
 0.6819808814486172,0.7020776889205764,0.11382814619386916,0.609695052933458,
 0.2130473990029812,0.03186769025479064,0.5049390437034109,
 0.42429356808896845,0.0778691014787356,0.4759418692691867,0.6455305721597709,
 0.06749860539052877,0.7664816150603907,0.24300099705616507,
 0.18561131660802144,0.2948654392528717,0.552687206973342,0.6997152598021323,
 0.3703883619688667,0.06470589810153127,0.8424670628344746,0.9281570466293638,
 0.5388635189119312,0.4514136313493231,0.3973180349306137,0.06228562342548882,
 0.1463264448186855,0.6597812675556269,0.9857776127567808,0.4269565353990118,
 0.12524743054293863,0.11525336781254703,0.175417654555588,
 0.25720954160970055,0.6872962647518548,0.6888341469615529,0.5337842184469115,
 0.0577143173355521,0.455649763101193,0.9662702837552009,0.7145229895374915,
 0.6190977133171892,0.6517400341916927,0.5519572251263077,0.5720807073410701,
 0.15085077112236422,0.6202833644510446,0.23304309321583605,
 0.17308415432543556,0.710329146289978,0.2589440490847158,0.8345284641499263,
 0.10351044918574026,0.015010746277014153,0.26606565995730647,
 0.3360945473334278,0.3132572560830409,0.30864907745166636,0.5369501894839879,
 0.43449681922729444,0.13962415149407503,0.8917550746431708,
 0.09670530888384898,0.4506837194492719,0.3198878937621229,0.9801633786763632,
 0.48668767269416713,0.7121229434469131,0.549825886867279,0.8713338579107907,
 0.9715292194725094,0.04560589236170565,0.8142061302017989,
 0.011074161688764739,0.5272208936439342,0.2663516417697743,
 0.9722594019161094,0.22815893816331367,0.06623818206738552,
 0.3302405731543866,0.26935672551465895,0.7107472677039023,0.762462999948066,
 0.9423454937530381,0.8908524036880112,0.8056764605452101,0.08358030401186156,
 0.7759083710932076,0.005081029469793719,0.5670230367999974,
 0.6819936362295094,0.5963503227373383,0.24046713863046953,
 0.49050579338249645,0.8372469164089681,0.7251581049924929,
 0.006183590166827679,0.20783364585517128,0.32570809722511673,
 0.31280107278629055,0.41537242859278967,0.7450126554644295,
 0.7412805054756388,0.3190178926510181,0.9442046071506854,0.3533254407236506,
 0.43726973878114117,0.5131458013022998,0.9910056626342103,0.9868734901264336,
 0.7466436044889148,0.5324957337277508,0.40010324623545235,
 0.07291206449468668,0.801052867638212,0.42573543391168434,
 0.42238785316397875,0.832120635183848,0.4413782862206336,0.16715297957286268,
 0.1321514867768686,0.7319934547635221,0.31704874633160773,0.3859525468098819,
 0.10640631633616182,0.21174375087591502,0.6128532467916343,
 0.9148071724811144,0.26020729972818524,0.9435953360928666,
 0.22444118979308114,0.30414227238261116,0.6798602190397124,
 0.12579285309996058,0.5326187779880366,0.6865583786503942,
 0.06270158922323321,0.4685974508303946,0.9792007993170765,0.5193155701879666,
 0.8963764291068717,0.6390563859855858,0.6452234580443371,0.6174472786379313,
 0.6955217147644214,0.09606738024622508,0.6582635769703659,
 0.45221834991585064,0.4989123153190369,0.33846027217428765,
 0.4879349155710546,0.12626540942568432,0.08731965971437639,
 0.5879444536626667,0.5534302971060872,0.2569281200714517,0.6507424649628156,
 0.40801394038706396,0.005335048121512642,0.7479815220796975,
 0.7854270677163618,0.48643563404486145,0.42355629871998746,
 0.6788822549570109,0.08926709811117495,0.8962328413654989,
 0.15189630792636322,0.8236001161496538,0.4494877483791192,
 0.14354952775288898,0.7334133451871533,0.19835037720492898,
 0.5481032150424489,0.8695882533834609,0.5750088894911445,0.14293734932356805,
 0.2386154287524097,0.09463663689178548,0.627185618735402,0.6728016609411862,
 0.263647481718851,0.8857187537019177,0.35332904938223786,0.3574537505954998,
 0.8429544100115008,0.4571504759885049,0.26445594930073524,0.6097146550495762,
 0.8794315576786844,0.6294949230703316,0.17475269940638194,0.8968971754218142,
 0.1818612118949785,0.6007906526235252,0.7853503046462451,0.17342804095943987,
 1.9243236004129471e-4,0.09350616612848517,0.36675224599044143,
 0.6074007539783286,0.9100785821334181,0.4337481637945739,0.45566908195166667,
 0.7677104822701647,0.14221738733453904,0.7411421513936307,0.9153983488319858,
 0.3851587582475454,0.47249529957607195,0.29004647211793033,
 0.30899099940298747,0.2234042808405654,0.5288112728851588,
 0.041953844063580825,0.0027500451771413914,0.8950774947789678,
 0.39950103287570327,0.23420114922254232,0.40347765748108766,
 0.6029724720340819,0.4950941060365889,0.010334390124168036,
 0.11458440193850794,0.5136018647616577,0.11967042941143968,
 0.4857775759748527,0.8505766538689242,0.8204517775577242,0.3405164403897779,
 0.7141922429887475,0.39612160496008575,0.46780938172217024,
 0.2657370887917374,0.8684785331558044,0.10590181941216958,
 0.19669703745651057,0.10037697926218891,0.6175230806173996,
 0.7714865850245506,0.6926178753596264,0.070385001118495,0.9721456910227917,
 0.9703491003991959,0.3608371543045066,0.3575898438629769,0.8802583027874311,
 0.17783721116613305,0.8492635993165238,0.06454190946193439,
 0.6638422176439328,0.24181943209062584,0.7369648431117912,0.4287956310649551,
 0.218683981623796,0.9585911591255505,0.395382654172008,0.292315608824559,
 0.8899290195229996,0.36138724520798915,0.22519590370628817,
 0.43764392989505985,0.0580217325500163,0.33425781132450616,
 0.9571069131421507,0.03037039004572184,0.5021347407845104,
 0.09964534367031841,0.47158147156102326,0.8083680722488815,
 0.44400118543696676,0.771578489017672,0.33102071560759405,0.4268644635373816,
 0.5316760451048891,0.34472775922996923,0.00678738083801822,
 0.12325131297245817,0.812011921417856,0.99247523251684,0.5843534158606475,
 0.5455093918212157,0.4954705549681687,0.21213760775936996,0.4357222094582206,
 0.2962314378665566,0.36661698146756116,0.7727077097002697,0.0947092985305531,
 0.9002676199128785,0.1810829446913047,0.42358454051008687,0.3047389381184856,
 0.5487574404145998,0.32786162632972704,0.6639915672698238,0.3836125013133287,
 0.8410722423697463,0.769793401656308,0.24118055857199117,0.6236076506741134,
 0.7487435162374687,0.631658404539447,0.7898917826745735,0.43739062844108745,
 0.7044040949956539,0.8056397540300663,0.9825548046559541,0.7288847962321974,
 0.9764471470790492,0.7435283064147197,0.9072686976073298,0.9141558251744195,
 0.5129406703371024,0.2451728705322458,0.6187339704007256,0.9828063311614041,
 0.839151011612165,0.4992293136117356,0.7373702261036474,0.7373839746243251,
 0.4961265713023536,0.9865312403232118,0.7888946923718407,0.9677154538581201,
 0.3237209722099357,0.2473450015376436,0.4137508134293837,0.6753026115575526,
 0.8852135973742428,0.4742739414675292,0.4124876916757012,0.2091403920289172,
 0.5105809497453444,0.5292628954108779,0.07138191218218748,0.433367149203846,
 0.9885375947320468,0.364974356677664,0.9730348017407981,0.7620521241978605,
 0.8131093255713017,0.059480799623415015,0.7543741527297636,
 0.5538542588762289,0.16730704776596905,0.2364744206084971,0.5684059993707365,
 0.7998605773188909,0.12175803418287257,0.8689473149279802,0.172005389076051,
 0.7918174362819267,0.8443194370931528,0.12707461966292255,0.1829726606738229,
 0.7039693201957257,0.42191590524485256,0.4072459737154749,0.3247822798199079,
 0.46693787126911823,0.12073737573010002,0.14639373468315386,
 0.603756030988492,0.33583016822972844,0.3421755588747276,0.8441112048461924,
 0.8714344171147449,0.8427007585056533,0.26945209653046764,
 0.45244016013655663,0.7307662246509846,0.8300496669032813,0.5233003339343968,
 0.9766938888073478,0.7374605340175413,0.13671994174428526,0.7747763383878634,
 0.3982095240264124,0.7546875061085723,0.053953661088904425,
 0.8276885742699427,0.4942090564598699,0.7717816663256292,0.1760347908408174,
 0.08343337599883127,0.6423166829404229,0.8366514159255545,0.6290707710858167,
 0.6846107772149368,0.7417509661327109,0.38282910060907305,0.4418318011581843,
 0.7647540714135637,0.85270738166905,0.3137759093729363,0.3806923290101748,
 0.7509125639888325,0.2740030303520138,0.8800517832112444,0.9444433619853714,
 0.06844647139288151,0.43818625982559645,0.3063794390175141,
 0.36781016306813763,0.3028826679349267,0.7730646813484991,
 0.20158829503731512,0.7746777990096212,0.38118155282427224,
 0.2853143558613336,0.5878660693108558,0.04676618662924692,0.6997012614748329,
 0.10980278525608478,0.8360639715311712,0.06904002865505865,
 0.5751070620335968,0.40944010697318434,0.7851865106585731,
 0.24430219206229442,0.6003301723006055,0.9827802078807903,
 0.24648316065653364,0.10660442544978133,0.7211966361664404,0.514561707805559,
 0.60855129619286,0.5664487572584663,0.25859677082487553,0.47908219219016046,
 0.5387609578162116,0.006236406274305328,0.2811477194546359,
 0.2598872951258725,0.4766235558235694,0.6729622514988542,0.9741173197878378,
 0.2633151138339953,0.7075123798010113,0.9888878997945828,
 0.027883274562969906,0.10607434672777871,0.9504583893000214,
 0.40487895105041116,0.40098273352939184,0.10274242762489882,
 0.8359144281744951,0.6613515212912863,0.8574480909601444,0.11712887627676616,
 0.424100550642441,0.35085149293516316,0.8555981510087391,0.6725701476031393,
 0.1760626804363533,0.025142175762653873,0.21098067959390954,
 0.06433177825376235,0.19786689983059513,0.1925740550071774,
 0.11356739498267765,0.314525317565266,0.10716634050231799,0.7787946749943795,
 0.46667179245550894,0.8028887503407598,0.18818294608261465,
 0.6063890424939005,0.759208578095979,0.6725513102884224,0.20720763602509762,
 0.9677999024452477,0.430079648107737,0.7278535762182989,0.49628711672109094,
 0.2791769021357592,0.10182066798301048,0.2610766572187728,
 0.24818559667784357,0.6048884435624255,0.4045103946330165,
 0.30038463736228227,0.6590697614963349,0.5400932677679446,0.5216352638296107,
 0.5050705629911215,0.6029186235772338,0.8097252650619213,0.9084037104790214,
 0.8836044576922243,0.7597652629088444,0.3828065790346118,0.9434620469898792,
 0.8166005245230017,0.2650619391554344,0.17050178089874923,0.970284069959616,
 0.7740711389025701,0.9505450827701946,0.9390479025747154,0.8711183560144391,
 0.2891665917940738,0.41486051623293185,0.8459820014884578,0.5609216010451294,
 0.9305538943380084,0.37033227458085316,0.2464456475381963,0.6632930599464995,
 0.372187055029918,0.5422403188952203,0.6476112193296217,
 2.0590965661260174e-4,0.2758310860514862,0.12141045435201048,
 0.1354619283814158,0.5464326228039484,0.1700105110477821,0.48974601487355485,
 0.39250912070819866,0.4385973268382557,0.06982443509239888,
 0.2173932531813073,0.022931121926662246,0.8241642807472255,
 0.10731972696530379,0.037671097378471075,0.8611882407028386,
 0.6521545932615899,0.02993593882164114,0.5282312818403195,0.4696736221224752,
 0.8103404000481564,0.16730071942937097,0.8037237271186867,0.8274000330747938,
 0.6896033688161605,0.085493137390941,0.5421848506027015,0.33093138947321354,
 0.45696624291370924,0.2715647138195294,0.9682615816710438,0.909654644655782,
 0.19499072862469102,5.908922448319975e-4,0.24208239734068226,
 0.14160661821913811,0.8259183625016149,0.18347152752359497,
 0.9507435096968108,0.38421623141754857,0.8612872312460585,0.3289737512826474,
 0.8209508388726094,0.33961091213291816,0.4393176131083071,0.4257914152945885,
 0.6528852720557609,0.012008340417545504,0.8025495401218868,
 0.013124633472850422,0.14852859927787387,0.6268995287740532,
 0.07261580475375062,0.6593249363147837,0.07531679890525944,
 0.27050791042911515,0.927727510979849,0.6428101420956088,0.14634769674185488,
 0.0438825208428657,0.3901491530498047,0.08401525804715271,0.9212770695011355,
 0.03873051366199243,0.45407061162850637,0.7869102616454468,
 0.6664740322309431,0.5104065029196161,0.9641343497907142,0.80215909348852,
 0.8297320896020994,0.9681703610347273,0.9688660437260045,0.9372902763853208,
 0.5535319112079398,0.4115685642048579,0.01443331661108671,0.9174710667594034,
 0.722118819468184,0.0994416128278589,0.8883099374511741,0.834216805143695,
 0.9673061111302301,0.9921650177995106,0.13234628672527105,0.7689832670155342,
 0.7483493270933403,0.9258849523285615,0.7348934158460061,0.612363865441242,
 0.4645885100716668,0.9165696906655869,0.6010423072399111,0.9619002501355196,
 0.35912388059996303,0.566867391367561,0.5025283251306287,0.4942487128416029,
 0.04514352158166335,0.7538794834340233,0.6740031486311374,
 0.43710976153724024,0.31137399701280966,0.6029429356365996,
 0.8070714183786788,0.3840956157422798,0.33163989915865755,0.6456139704301329,
 0.4265683392792279,0.8740053738222129,0.39758456679278675,0.9092438764526296,
 0.8886961508115638,0.6043729762115675,0.48700590887676876,
 0.47368575975071336,0.8391741285126461,0.31720430792417786,
 0.3930684834838625,0.4884067596836006,0.6938772979548737,0.885846694253687,
 0.07373802430651744,0.1312963655202325,0.8563788876307361,0.8419420018683331,
 0.04111981822896593,0.9721989678319092,0.1872536182672675,
 0.22002727731438665,0.685171411419834,0.5446763230900648,0.38167719119972776,
 0.6935075330043896,0.6671776762804537,0.252720006113492,0.25335042839357425,
 0.4145626959634452,0.5406133657507293,0.138665850181394,0.09568115319797077,
 0.7893959994938851,0.8739656009161517,0.24160739804843,0.6995971237441158,
 0.975526384280668,0.5272793061830737,0.8223038785106391,0.401461466332364,
 0.8380030933348795,0.14463138884313054,0.8930139797589022,0.8678284398279958,
 0.5738474374663897,0.3830387401276706,0.8381892914569021,
 0.013962173473307748,0.10470685843396044,0.31724532315161647,
 0.9486850898743221,0.3382423647276108,0.8263444409783681,0.06760367704684156,
 0.35718532469819764,0.6955175796550693,0.7206351287176405,0.4887176271655558,
 0.3509559451474207,0.9074330906263594,0.21904676678613166,
 0.17056670907532645,0.4887063676143524,0.5121275116864971,
 0.036358758793547774,0.8590345262596465,0.6356147235769227,
 0.2472669596417706,0.1373875566038283,0.6209854617597147,0.49752084633804083,
 0.700054183205785,0.668095424981472,0.6808243467331613,0.8159864270579771,
 0.08115443084541407,0.6687675162659396,0.3237554430099209,0.6505924918807527,
 0.22514551389336168,0.6744158170967128,0.8016490796880854,
 0.06983935041091027,0.1592338354291979,0.42143386960169416,
 0.31463861576244945,0.47946136824126884,0.4077184773936585,
 0.5527817281220014,0.7683085087660635,0.0034896151489132166,
 0.1772951195594875,0.8214511016568646,0.9159097120952853,0.529642473016801,
 0.004628359413842764,0.3990744196076089,0.578968806390972,0.953797997374709,
 0.5183039066805998,0.19583263778618853,0.5941134322283623,0.305187054237258,
 0.7027261024902735,0.5047319560398903,0.7169125921923888,
 0.009097586505109612,0.9934906865699862,0.09434544739233675,
 0.30757302446595125,0.6969828579535744,0.09958588028883408,
 0.9472939380247581,0.0035310232181373635,0.7911468349010274,
 0.8505833504655709,0.8065230327773286,0.7821401217161534,0.9297271963642367,
 0.3007423866579779,0.7623422335697958,0.5655433313852783,0.6727016387032717,
 0.8484968471238944,0.6770058887363618,0.5869596735466472,0.8970612099171436,
 0.37846128473900054,0.7862969041408528,0.23598876449856732,
 0.3614458924764098,0.21196455129619496,0.9402121918938424,
 0.31160536357632906,0.2894611446419747,0.8010850068250501,
 0.40005264565055565,0.06271844307172125,0.29841145359250665,
 0.9155813501239675,0.7906310851454615,0.9679290153375563,0.7524618666297639,
 0.9402218437225597,0.7056456205316524,0.8748596425483381,0.5919486483666776,
 0.9662703332958399,0.37404225802400115,0.18652043051674805,
 0.8827891205682589,0.6717600514448541,0.8014735721609548,0.4525325161422532,
 0.9122060581238152,0.3239108666999131,0.5106513641109002,0.07515159343807898,
 0.8118770407066718,0.10480343309916196,0.09321997370944435,
 0.1487914834297236,0.5747852042510149,0.949581901266844,0.4072451373809539,
 0.4373084336521126,0.3636974396681225,0.16613463465727218,
 0.07961063671032309,0.9133753899501509,0.12517350031947583,
 0.4541055432085326,0.4964493300834655,0.5657611834525712,0.4706055432217201,
 0.19444431266172046,0.1821475830057555,0.6589504079298963,0.9751913773288633,
 0.8415715841855714,0.3944198594667375,0.1815360594776344,0.311747099478213,
 0.5622581699606751,0.32064044333373753,0.459307088399617,0.7639218441543472,
 0.46898709498894864,0.8338523861228141,0.35539823782679714,
 0.6621767767569753,0.7070179016333731,0.178317879689197,0.6062567449512737,
 0.21486706099409636,0.7008869942015782,0.7315286036652969,0.6186020338797557,
 0.5332817297934971,0.6937054862281177,0.9330850392277386,0.8692247734709594,
 0.6431265553410667,0.04007417037244032,0.04979201493279639,
 0.37762354260493913,0.10094448846783566,0.5667433751685087,
 0.13215056863059937,0.6098120911365312,0.6262989960746487,
 0.08097466513037788,0.8780485988300919,0.25395813706274817,
 0.5545503794533215,0.36706554867201,0.16429497753714062,0.8650731912030112,
 0.8682326947200432,0.9206612254988114,0.5016503252220202,
 0.024411187541819035,0.5793733233327802,0.008501790498144812,
 0.4781833821516437,0.9018482854677856,0.4801254200464339,0.21479748040039848,
 0.9629403613442219,0.976535062197392,0.26379475949293696,0.9503899191857145,
 0.7097900530035427,0.40988078956237217,0.4324631619340662,
 0.03301689898648996,0.3160308638917977,0.4608077310492058,0.8758929241858916,
 0.8607816691077218,0.08035883052109372,0.22388577960223022,
 0.14459250301264936,0.30553863056499075,0.09642236442103624,
 0.7388505702728924,0.02938072928771096,0.4150461768587915,0.8938238692186449,
 0.09902570036772518,0.8177049016507842,0.3370572889795006,
 0.24592803443173628,0.010707192828754541,0.46315252388878037,
 0.2040634797645411,0.08535190676726256,0.6713310717329958,0.6840790841875184,
 0.7474996521187007,0.225600726534962,0.7752805872427286,0.07082164873617935,
 0.48502096668746264,0.6110665568792488,0.13230809135010968,
 0.22133402590711926,0.3607615179080006,0.9947216740978723,
 0.33958875358473894,0.09993498383245458,0.6356552681333796,
 0.4333389210448999,0.5218796081599483,0.056879870960127143,
 0.5865567979451278,0.40861295319489765,0.9750146804596436,
 0.33040483379513763,0.6402600313512113,0.36408595880237893,
 0.13665705020284957,0.4412124155203394,0.8493344347438676,
 0.14961482849230223,0.16911856207883846,0.4621051501607637,
 0.5288214199409135,0.2503503256540379,0.6988101859345413,0.5625299406819531,
 0.6194770075173931,0.463124982230246,0.737679671338969,0.5762801580540726,
 0.22532316940063213,0.5999290518061214,0.470528571396704,0.239326478365543,
 0.983306766802176,0.21569259783505557,0.302896919738744,0.18113569276905528,
 0.2760148180134495,0.06696405371374436,0.8938549788674657,0.6721646151491099,
 0.40703140886405464,0.4168030165164579,0.45085373105327153,0.68304875568872,
 0.5657625195481799,0.5391000077938795,0.5691466447621683,0.552665772183337,
 0.11158480538956794,0.3361452658914197,0.9535053187125346,0.6380162665167182,
 0.12479691817979777,0.7960125292674314,0.702595297603317,0.2823855220761905,
 0.1425739167698188,0.7731653326069474,0.10675427503615953,
 0.12119407844292629,0.3583673331720669,0.3720457428391304,
 0.10293081877426491,0.5740376249422567,0.29919944653879116,
 0.10254709511194737,0.04193387585337849,0.4931604976116818,
 0.9936297716331891,0.44027261775991444,0.15823852323514798,
 0.11621447361472492,0.6144285873011097,0.10459057323214316,
 0.6733318043460209,0.7361611096441184,0.15643009127119423,0.7730633362027065,
 0.22659639001639698,0.8535746759968519,0.6407483189580365,0.5806353634478879,
 0.29236449250924057,0.5078539362482873,0.02087447389778141,
 0.5243515130072647,0.9296335889173828,0.5630333219747791,0.23862438301259836,
 0.7461393079889067,0.38338060028517595,0.3075093219660936,0.6817508804546052,
 0.6204374794344959,0.8923408506190279,0.9261500284769216,0.06916708547852402,
 0.7884296754080979,0.7289134139126542,0.434987720318877,0.9908617324581691,
 0.27988309129253897,0.8908975504489343,0.37299231578563274,
 0.12288292405389689,0.6960686692467767,0.061440750995216176,
 0.9694045313640683,0.6114383310146655,0.32648700922864515,0.8040748964810129,
 0.3753190078256117,0.702147993577835,0.5559874215666739,0.4183105485718681,
 0.194932163020206,0.2634698293806792,0.30262855363864083,0.26011441117882916,
 0.09463648133312608,0.7570417918951546,0.910609555783834,0.38453574418607206,
 0.5845482838355305,0.003148048493784472,0.2836526643395476,
 0.10494425042175459,0.7225333707785832,0.10381854075816199,0.954432986504565,
 0.25967727429178034,0.9498412601932359,0.6092723067750447,0.6163190530389828,
 0.7128738996581583,0.3134686411077947,0.8791012489218524,0.9900523348313109,
 0.16254539111999533,0.3301309200471295,0.8328011981220586,0.2358570834687983,
 0.6779481808593075,0.19455144151595238,0.6197447453379363,0.5468129932045929,
 0.4470148405733195,0.04639853906412872,0.9096959842413007,0.7318209738645163,
 0.6796474910615543,0.7453747236567111,0.9873025758758829,0.8164582727893575,
 0.8992401897977438,0.15443563924213144,0.43275614371334314,
 0.3349597160406925,0.6197386207375517,0.11656431493705366,
 0.41516276612098846,0.343038672566528,0.26511755223569133,
 0.25085465384876204,0.8291790246000152,0.662046132229423,0.17498904859440345,
 0.42982015282128827,0.854217156519152,0.9580995380673463,0.24972228453414502,
 0.6103429370224559,0.17558003722405735,0.9717939771592833,0.4249133591541565,
 0.5501630530619378,0.3308280371450951,0.40154066992659165,0.5578769886287867,
 0.5379899387860001,0.005125148684720693,0.8949760288346552,
 0.6875668079703979,0.5514772914902926,0.27491309538567543,0.6029282156834841,
 0.593612977946262,0.6540403360040341,0.2504854239341031,0.8608255238344165,
 0.20012600677306602,0.43907001936621104,0.15119944006247243,
 0.016410341213416335,0.7472529579647779,0.1091110151330642,
 0.2021395341072434,0.8409203788519738,0.021316645856361505,
 0.5787345511769613,0.15685158005892563,0.21685747539697853,
 0.3419837504228347,0.1757311455129965,0.6816568151217133,0.5751981436822485,
 0.22147935759080872,0.3626577220849244,0.971824180421875,0.15904738826663178,
 0.2551501688176716,0.0526730995888004,0.5563935187327249,0.698827480168489,
 0.21353238920820594,0.24636624735446921,0.81859111887976,0.9086335840275395,
 0.3060459689014349,0.5540252917126807,0.867340835989379,0.9271644634780296,
 0.3560555532455937,0.8308014140353996,0.8166892635859262,0.16825091744073462,
 0.7174654915652916,0.10410067512155408,0.35323990656135706,0.76566203613687,
 0.018224354956129174,0.1581104748547556,0.4540299673040309,
 0.9584061469879133,0.6851062996053132,0.4567881491604171,0.9058131773851102,
 0.7180910917777823,0.4348528034290595,0.6485200100814383,0.7725916192160512,
 0.8611289068090859,0.157312265448027,0.32780897223616867,0.6385592891284246,
 0.9378777173020335,0.38683578120474926,0.8904369465226342,
 0.20146563454938549,0.20695978607795606,0.14592483558443814,
 0.6786284917112858,0.9088645751123677,0.10713440351817138,0.0549498714719554,
 0.09235864593063647,0.9017287078080847,0.9606323935226151,
 0.027618024414012954,0.5025841737546889,0.5546109086103506,
 0.45226494794803995,0.5410607468914714,0.46390120451329975,
 0.6978338470758509,0.8828024730292264,0.36711283967958963,0.9426502024865968,
 0.7762939263469179,0.635286775664116,0.7538736087811463,0.04627762657145873,
 0.49183386123464423,0.3582060429599111,0.31086084677325,0.33107129087327714,
 0.654239881630027,0.7722267601405276,0.8003143775433162,0.9102980227355396,
 0.4741977825998813,0.987746817216964,0.006623731561503976,0.1653294995992418,
 0.4701759407117716,0.4895149635375753,0.9784463012859701,0.3029988600145628,
 0.8651309261107037,0.032807298998588585,0.16673141190305985,
 0.5670810848846353,0.5271602547178615,0.3801759208701472,0.9819926536998208,
 0.5762391780523911,0.36263249428463795,0.11879800211178759,
 0.044768314414350696,0.0626667488928967,0.09028501830083935,
 0.5615258872388494,0.8031942765737969,0.5636531281494521,
 0.059467013977591465,0.1098668931937018,0.47162305301519714,
 0.9999875779317646,0.5811113063973042,0.3334214283813317,0.8590456227059027,
 0.32800881688552475,0.8346094132897567,0.1593709316660108,0.5955179328246727,
 0.7361507087566429,0.9110352414528378,0.9625290531577038,0.6059726170875047,
 0.5291367525371606,0.5813742330809304,0.15831028504666267,0.5475451143777564,
 0.6467926512336711,0.5229757047832027,0.027890019428698953,
 0.14543037824919647,0.9666746146997657,0.8823542482658449,0.4229757930424618,
 0.7548777118300433,0.3768668006796264,0.7335289303147569,0.262557283307148,
 0.9108182520774271,0.4198725873142978,0.527292138112002,0.4596504954050785,
 0.5056167339742744,0.6655597197326524,0.22162299086448622,
 0.37664763045969263,0.1969520261368055,0.7935633224227385,0.9154100072986497,
 0.07971780228638237,0.9639440011702595,0.6810948777674506,
 0.03989709760061644,0.46461396059502624,0.12254472771355784,
 0.23590723524240664,0.4572151165870282,0.336883012188676,0.2419041588679106,
 0.7225708496081675,0.6360056377087846,0.3083998968856777,0.32084596313845326,
 0.4791679531300398,0.2786134560346455,0.9035371081692931,0.3566236670133498,
 0.4124114418226047,0.15400744399664068,0.45205822381146854,
 0.47175245193000614,0.21858514824179365,0.4650506431459177,0.246191679039099,
 0.3776256417639645,0.9354067152838215,0.460074567422764,0.49130356870814706,
 0.7945076836798265,0.9854600287834532,0.27401603949964093,0.3518992888948642,
 0.1947052597285902,0.9643299025309808,0.8951158936188943,0.3901292181899172,
 0.5259741901646955,0.800763469515295,0.8256438979972477,0.07003225525690493,
 0.06944083638742482,0.13081846122557006,0.9020888416820885,
 0.4089994431257603,0.4215652966879635,0.8919746158755488,0.1890751394133935,
 0.5687741102783336,0.7425818676365218,0.11529366109152572,0.3084759063648438,
 0.7844247956892675,0.35246397177367683,0.23739531755205678,
 0.3680567536910081,0.360239915666712,0.2479146683153135,0.38814035572801486,
 0.6937478922490882,0.2930481335497771,0.14215010911839,0.15641676945318894,
 0.9967559338038184,0.8627035787888991,0.9984102736471941,0.2996663231366983,
 0.15453568011459584,0.6125952787766744,0.1548182364756714,0.5602776647230283,
 0.6514717367568139,0.42868658670148074,0.29622820586148624,0.441043005032805,
 0.34426350498081515,0.5292108703066318,0.24514012709068966,
 0.29784992283554823,0.7181901174659191,0.884571090725323,0.6887264334714842,
 0.2589144096902817,0.9416624935083149,0.2626545296926186,0.14993506412074453,
 0.5322819129702351,0.9256502081407654,0.7651606574606664,0.05512524450356704,
 0.9723246767501914,0.43333851748971286,0.49406222968759383,
 0.8872921273911142,0.5827932932050075,0.5745217163141199,0.5843394191564555,
 0.26575118503700634,0.2936824160921251,0.6381447897536078,0.6993235415365093,
 0.425439967541276,0.9490819079020517,0.6937297232851585,0.8769798702475432,
 0.0881256781489892,0.7385577612422954,0.2533451333558059,0.0918906319089332,
 0.1266818436132343,0.7717204553028564,0.8921307059272489,0.47267852514891473,
 0.8911899715529215,0.7822469437640489,0.7361613473795505,0.33554985964672035,
 0.536136276164564,0.016561891530191386,0.9734140200632106,0.616352411727138,
 0.3094044177529186,0.4021124607458183,0.799143551019494,0.40240907323958053,
 0.8460169244146825,0.8081063924998513,0.6538609243261998,0.5976643546476718,
 0.9859599743561929,0.3256044129518987,0.8075312738044762,0.32292263535903765,
 0.621080224832524,0.8708510625085165,0.13603571081223242,0.8973729186322665,
 0.45186873972231156,0.006363357068182696,0.29360855891943016,
 0.2981472352665945,0.41344360650206724,0.5775802207811006,0.8177776949425648,
 0.22628070087145113,0.7553294050107144,0.4968684432782753,0.5693862625260642,
 0.8775605521969587,0.526797245519729,0.58251773766951,0.43100704772487375,
 0.29020302318712965,0.5041682571239817,0.5743035339151531,
 0.35015906878327885,0.8734096237634689,0.42319803292961755,
 0.7359233703195844,0.6719150554711177,0.01675116530950571,0.9349685428508516,
 0.6805343562496318,0.5435522148838787,0.13793285397005017,
 0.16089386539713824,0.6176914730125179,0.857251857562846,0.7084233054358584,
 0.4697699509288016,0.2759729130579207,0.793686297204658,0.23904206666924344,
 0.31501370282783125,0.6940410399651333,0.15112118942110586,
 0.6707221178299952,0.08184742055161764,0.5264027248975008,0.6376290869132761,
 0.1542988235830245,0.7799795645199735,0.7970325992675055,0.8556781404108011,
 0.3629565309778293,0.587864674028611,0.8551375883491346,0.6322749219588231,
 0.4738574056815309,0.9797605496779139,0.4175215950662561,0.39272308557458957,
 0.47609397078489624,0.9917928216892269,0.388062243533014,0.9396383303115362,
 0.16149762278559798,0.17050239909499632,0.6763004042282771,
 0.2893772265388844,0.35082650690270534,0.3167730985147491,0.413951940368011,
 0.18510835510494805,0.452404298004608,0.7268741397916639,0.7676338776599037,
 0.34125626379067564,0.30285992713646337,0.7183806854244829,
 0.3878462275421831,0.06974474649496032,0.05125530097691133,
 0.10479422449109876,0.7989486130603547,0.4555348342414899,
 0.18346300934676596,0.6170599712495521,0.1596187179361579,
 0.11100125417908113,0.9649120860122685,0.3001102975831357,0.8995938989478373,
 0.10192264675158713,0.27792404568343,0.03285250893713232,0.9885382512220376,
 0.7089615993461919,0.758315696340278,0.5939759469034135,0.17766773571182015,
 0.1400261539630685,0.08330353263906476,0.9952137800052147,0.558160015751134,
 0.8110509573197839,0.7103338990404989,0.3435522940357816,0.6747688108041241,
 0.07390004199639455,0.5455937164919812,0.13693232745331163,
 0.16140443830287998,0.865877025427698,0.23671427926461996,0.6506962995344721,
 0.9944855540505257,0.6741286628359331,0.8721552615657048,0.1726945502192998,
 0.4400672280398419,0.5193926242960075,0.6883866475615029,0.5922724838195201,
 0.011243942352257896,0.5228180076421742,0.06074817233294538,
 0.17207775166174732,0.6054401635496096,0.16155158841139206,
 0.5831997130201324,0.44583132762446387,0.9828643613007343,0.5270916308729422,
 0.9695554133255025,0.9124510614460382,0.12180813582304373,0.5342469460353755,
 0.6522693588834843,0.04642123919324903,0.30198814696308984,
 0.6509725709068712,0.6401202043905095,0.46424368279016726,0.3742212977752708,
 0.01666925491422233,0.48361186038482606,0.007986933558446818,
 0.8242588740361644,0.6458877428613753,0.38917681484588784,0.833600626079721,
 0.4531671485818539,0.8752919750952537,0.8195231889214973,0.7188768107471671,
 0.4361745655709043,0.21172581448568017,0.6600601057829426,0.2838019817387887,
 0.6099125732947333,0.16733782558784682,0.21452531733765212,
 0.026858178702054758,0.6286900865286803,0.6748539991154527,
 0.6410724187553924,0.12591584536170652,0.31738860013424197,
 0.6588863281972887,0.723198336987175,0.5829033844997689,0.861168574005517,
 0.6586849315511987,0.33860002872845363,0.15212891091457748,
 0.3208045706558813,0.925985288625369,0.5057456922857686,0.15280804929852498,
 0.3590338209917828,0.6231456791391862,0.7411676570937555,0.1452112544811578,
 0.30923993539982964,0.195998917779167,0.13925556170422515,
 0.19818196619634143,0.6687897642538172,0.9598720398927408,
 0.22006887375648776,0.14937019751256675,0.013077923313179918,
 0.27303467589980823,0.6484034237392828,0.23955356520023074,
 0.6924079985901344,0.08025947690368995,0.5057777617888558,0.3824419781915005,
 0.24388087638432965,0.15237263115323618,0.21732554037881158,
 0.10789408903065101,0.8570658651855834,0.914946344329866,0.39607110045676763,
 0.6652748074797716,0.7359239665091787,0.08350262100670813,0.8194312747238992,
 0.7541513646150059,0.45588321932530795,0.47106463730168535,
 0.2555271540396584,0.11876551332746421,0.564847485461392,0.5779804291824331,
 0.18179099786001174,0.5063777941515678,0.3435835172273587,
 0.15650513542710298,0.26808102065851547,0.8510435408657488,
 0.6098937371141258,0.44169652988849006,0.07745035828351543,
 0.06548139930536867,0.5463610346728195,0.7601790007518815,0.5671854710827958,
 0.9992585047341443,0.13333398185189327,0.4945184877286707,
 0.29148861108074875,0.41921003908893495,0.5875323489055643,
 0.41131252031743526,0.9762976991141769,0.46361957165479084,
 0.39916546538880593,0.37282687332829645,0.5416583865868525,
 0.29485267158395967,0.4793326475042202,0.3870658743596147,0.8722597475687398,
 0.5306171209523064,0.20298696005918193,0.581729478506201,0.5102632719465665,
 0.35150030335403215,0.18686524387024028,0.9407050351768681,
 0.7419942086867063,0.7548927566245759,0.6455651722758977,0.40811805624293196,
 0.28161853170399764,0.5670521624956995,0.5406301768772022,
 0.47001735964339786,0.5920356912661304,0.9519377621381809,0.6791605437944697,
 0.38647597328498395,0.593398373458573,0.6947409878445661,0.791818577846704,
 0.8849841398942875,0.7978037926209784,0.8241077294190028,0.5116073366675091,
 0.3152417763720031,0.9498852035626116,0.8644193529129848,0.17522911398286523,
 0.7841486801516045,0.41495779759035356,0.6625961286215603,0.8452244493647116,
 0.40825423019275275,0.5803966129334764,0.06341016774987007,
 0.006310222413417388,0.40608781595650745,0.9651737395303525,
 0.5562883920457243,0.8325725559320841,0.34491599883980983,0.7371758806632173,
 0.34693318616247537,0.9313694352490623,0.007179040006843884,
 0.3761105363080768,0.04462286254608183,0.1701695956806224,0.8375761807383933,
 0.512900470473362,0.9088942174899032,0.6348883151319611,0.6187957685946296,
 0.9080817185931187,0.5539486985599318,0.10515548164282007,
 0.08435662107425412,0.36227426180579914,0.6122629372414731,
 0.2769587567518521,0.4214275240219203,0.7834832398327144,
 0.045402163240825244,0.7013952348024853,0.3375717391958277,
 0.012595585699789913,0.7431399395593871,0.737985789581582,
 0.04982481481711187,0.15655032365684152,0.5673585404986231,
 0.6593688424365672,0.7814404670533228,0.7630180168190182,0.596641933516691,
 0.6093651842059019,0.47736313667480357,0.4218191652133366,0.5182777400105661,
 0.043405998399969725,0.45174604199036816,0.8635882370401569,
 0.7962493296417857,0.9425296843329628,0.801231352646551,0.8073135564974239,
 0.48493068349625645,0.35869351108626457,0.6852336098297784,
 0.4705829836337796,0.5087336273657195,0.83003994045311,0.8280744384270673,
 0.014244231659397033,0.939739993990281,0.35413813369268454,
 0.17521301180775173,0.24299320358112864,0.5168945609325631,
 0.4369087737052142,0.5560058067025495,0.2540791245781089,0.5292627940080749,
 0.9536325709969941,0.33527767500479433,0.47249951706045445,
 0.09778366345542377,0.4155340452893751,0.5748989839113275,0.6817921814597674,
 0.6985666812384551,0.3473793392748662,0.1934590517330631,0.1627251214272072,
 0.6023379217551987,0.121912620899147,0.5053570924317675,0.13601281901168838,
 0.8838309880493628,0.6584152440734038,0.5701001835633388,0.5416159409118106,
 0.08839582732958196,0.23138444194578112,0.7851439508611339,
 0.7143738773784352,0.4638321889784953,0.3932449845685322,0.7129772528828616,
 0.4704714531763171,0.1761053488375809,0.5082806421595965,0.6589343967006079,
 0.9105393761307367,0.6933140442735892,0.6426702095604382,0.8330379531470352,
 0.4299583910616297,0.4605022207311781,0.055817104918052585,
 0.7015427207731625,0.39586466910397156,0.039569018681271384,
 0.6020571803267314,0.6708582549406298,0.47469169130739075,0.9831160234274267,
 0.7410079457063515,0.6454286091644059,0.703533790586741,0.5327677586977217,
 0.2503080846217822,0.38158080575216335,0.4114244643771513,
 0.09777752268962847,0.7432025759996076,0.6604247608351624,0.821501641961171,
 0.8930615829049184,0.7742257326275164,0.11316321813075958,0.3088165325884411,
 0.0028838520098023457,0.6292901720611679,0.29187002909379745,
 0.26288178577970134,0.4360584198522428,0.63932773639018,0.671175651397557,
 0.06022159493330703,0.5384251185839563,0.5523976064029732,0.5389756468506698,
 0.6306851443468527,0.6438771425160483,0.10654494293645334,0.2157813845771831,
 0.5427710188862813,0.5137633215138411,0.6468294215866963,0.12636891954755414,
 0.10141502478422115,0.4067955625510382,0.5041948042776219,0.5533632798178272,
 0.5277144095136775,0.559235938577089,0.5043440125408003,0.445613143361111,
 0.838840058058974,0.06791488269519785,0.9657684421986885,0.17365364687310425,
 0.726042098917542,0.8800392971614843,0.40855291365663293,0.49489231414492374,
 0.8650776129062797,0.22042323440259148,0.7138751010179883,
 0.35313864044860466,0.9670810825626478,0.188930217401734,
 0.022967346395550536,0.3453892143107864,0.8463610540192231,
 0.12985299860099753,0.9332928292342825,0.16960604051278239,
 0.6697728661792778,0.357317407571486,0.9737723424810435,0.0408361190477573,
 0.13932183184489133,0.3093166149925124,0.7270291640005505,
 0.005969871417273898,0.9646236311422056,0.438679089785308,0.4409277669357521,
 0.30586092994350655,0.9097507257813239,0.4419082920349521,0.6583602639219271,
 0.0852401400260957,0.6486178639900372,0.9629102386520352,0.508604506544357,
 0.7268067791629186,0.1875043837679109,0.4352697248802323,0.2767763161040194,
 0.0314348639983596,0.07942353192408502,0.25144230513871924,
 0.6534785152923914,0.2694881875672406,0.8462860592355277,0.0222874544918108,
 0.7683336496318661,0.40185668296652777,0.9772918598013597,
 0.03758842096915049,0.5462616888737569,0.2374929759203832,
 0.08400200378386824,0.7075788221680475,0.28849000591254637,
 0.7752188841027836,0.7081136929189755,0.9108847379677132,0.47831831659330426,
 0.6755367276321779,0.8418465710376344,0.8923595575655326,0.18106140313259833,
 0.4524171842521849,0.9023673394293154,0.3858207776573286,0.3699026181348051,
 0.8413998627700662,0.5553395262472471,0.13332516537895212,
 0.14265794771464013,0.8432531254302517,0.7125884266234819,0.504976390401075,
 0.4869804953287016,0.5228169938761617,0.8193797122647595,0.2704717544259524,
 0.7858244947474955,0.13841413630432498,0.924714857449627,0.25650080862070745,
 0.9739052941894598,0.046979479957780845,0.9590700597435757,
 0.21089978959848987,0.4299862235924331,0.6378964988301205,
 0.09349259331149062,0.38932165668714147,0.009759253674618829,
 0.639935443166748,0.33243205514388774,0.4665075996993703,0.15508538650586934,
 0.09632420647642737,0.2898529132489026,0.9097785118010789,0.8514744126829961,
 0.19078118916589037,0.8004027079910109,0.5104727033540489,0.8852870014287311,
 0.8929930757456911,0.8322747900887808,0.1800005958593931,
 2.5987478214917203e-6,0.33039144641238827,0.20355399013229758,
 0.1908278651053904,0.43936827585146865,0.7471180841743083,
 0.16782464721085444,0.47109763680145944,0.24420725704432622,
 0.5106111079008695,0.11816392772604856,0.5346610830711866,0.3367729321167835,
 0.34467255943527897,0.8915272119769156,0.07970158900498747,
 0.9396306366949917,0.8371421703566919,0.0617569405682199,0.9381387996427473,
 0.03253711605323539,0.7907194431754587,0.043800909208093985,
 0.7689189837988682,0.7163380375626813,0.48198848546676976,0.5764614111663581,
 0.49577934288814807,0.6668312804804953,0.09572980198441139,
 0.9543433420121757,0.2717079964152175,0.8800069745486173,0.9049833046349599,
 0.6566959408529004,0.3168020108166011,0.423277174759652,0.1658158492137196,
 0.5805796252027533,0.8657977238910847,0.924824952118597,0.8015135336421225,
 0.6299163899725115,0.2062601856609252,0.1422101754462991,0.04623254031473145,
 0.02247813116203845,0.09814160737359456,0.7367630101837374,
 0.36681712659880295,0.1220502440920277,0.766940876154906,0.7950883397820927,
 0.1794395700576208,0.5230239538614914,0.1456341958519335,0.7044899290777069,
 0.6745344501327438,0.23252585816349036,0.5980004403496314,0.6217759070428387,
 0.4946342312446045,0.7063328273867627,0.7374708278084443,0.383774320725899,
 0.22975619024400384,0.37564633061638764,0.719524342152833,
 0.21341005193341733,0.7063533944562613,0.7107389315941604,0.5478175027728486,
 0.9836225675990943,0.5272430911488157,0.8873565933440819,0.9744263675383753,
 0.7085954976312075,0.20192427097870858,0.6748883541554982,0.6385575653908541,
 0.9472900986212606,0.4854639620975635,0.35569526768755755,
 0.33717932109205195,0.6023001544858818,0.49704045746420833,
 0.6715010717849423,0.4624765413445405,0.1924877949125463,0.01777907840181636,
 0.01922396538076132,0.38496244366219534,0.1353878845462998,
 0.9255258573724001,0.20726780099293085,0.31230022772760413,
 0.6593922002552953,0.9324482142307332,0.020957175161807062,
 0.08820763530904208,0.44264927649550756,0.31328939886470153,
 0.3261582475019964,0.23109641102566425,0.2446774505304672,0.868610656833651,
 0.4886537773617313,0.6511192016777418,0.2771410836356798,0.5348957963445049,
 0.26914856514956,0.8765535534219845,0.1749249547690228,0.8903501410151926,
 0.3502874247582889,0.9477122490652193,0.8981566905071092,0.22597237190170327,
 0.6215961048050699,0.9420183786240267,0.8792816887432928,0.5714478185736925,
 0.47891076636872343,0.9136381323720094,0.7571956275156996,0.5519948653955591,
 0.7551961399749099,0.6531428161245417,0.23127231366440282,0.309474665959368,
 0.32042707802333703,0.6055024380261058,0.49412916714595334,
 0.8697751037119561,0.5906381214161631,0.12455482755207326,0.586134574840384,
 0.07712932167274489,0.2165075462622208,0.08816076328028599,
 0.2920512427918279,0.42917681075304714,0.12541968030896233,
 0.054874503124484386,0.6629227424710669,0.863304503768661,
 0.12357815607539158,0.733788421991413,0.9506157027764497,0.3517082652827992,
 0.8865299116707873,0.3955964650650947,0.43544020262421534,0.6747803620359081,
 0.7765999449870626,0.4011217146545968,0.6112045975237275,0.8947744135823505,
 0.7704688790292955,0.4069153565236916,0.7073203244017418,0.538568038808103,
 0.09977291453122239,0.42788796119587647,0.7008764754689577,
 0.9140485802581904,0.14154530155792033,0.5429421135320777,
 0.20308641051612164,0.735683131913109,0.8252175059148797,0.543143642924404,
 0.9323115153505412,0.6794895098135799,0.6554840550199921,0.7342748121239098,
 0.6622356279447379,0.29918710818180405,0.6029090494005729,0.4302010986981748,
 0.026889463643169575,0.39450395936306104,0.0981884540247725,
 0.5904141208800017,0.12645336646185024,0.8088270345919344,0.9976508083511434,
 0.7621473465099365,0.9381075813501638,0.4799772444292263,0.6404254965161602,
 0.769481611339215,0.5122771291375023,0.061145291996441586,0.7427517539393548,
 0.24793699134759462,0.2563089236301007,0.15589370496613353,
 0.7076276732081076,0.38471297870330223,0.22663869580197837,
 0.016840601099644803,0.09043733003304522,0.23511585454177997,
 0.01794457956343387,0.9661435104856078,0.7172425072621947,0.9760952197543787,
 0.49845972173650777,0.8629051201592121,0.3173855187404312,
 0.10457679280915688,0.546710179456452,0.9747185844126732,0.3671129110326624,
 0.21983719227743437,0.303783973628047,0.5129957773553977,0.12627528085824902,
 0.8678356280185272,0.22912448680730457,0.5295482179113632,
 0.07438182254920589,0.13748868223413668,0.9418539394585,0.07463956310172126,
 0.6359881938337706,0.4289036985858611,0.5016464805535811,0.5309493678730611,
 0.7960593631015702,0.540533119876492,0.9508229520899735,0.5704055645110877,
 0.19354723996968037,0.615065320935119,0.38241082195384046,0.9971658271998052,
 0.011627892076643498,0.5192092662143077,0.42402011012125707,
 0.6148279635604473,0.7796981200263214,0.44412028284244454,0.5541099191159881,
 0.9681761547610244,0.5381588564444677,0.4016227573539283,0.752027428668387,
 0.12697673145784016,0.2223158626874947,0.6648100954639153,0.4886414367383659,
 0.016628605157980703,0.8711226383525305,0.12238539532401083,
 0.17170600645404088,0.23035744383364565,0.9171624568773443,
 0.48425750086062525,0.39268070015702594,0.8995684427113839,
 0.24312796295194472,0.4455402894861493,0.7315725730518149,
 0.14198908321867187,0.9929479943424442,0.8485890567952878,0.3045036336769342,
 0.6207648288451901,0.9096445500189607,0.2821209429895897,0.36532874541483273,
 0.8820666721632968,0.22282933286727669,0.4111881056937615,0.2893668503988427,
 0.9883797200812943,0.8914869230273086,0.705861376649001,0.36517758465606576,
 0.8157697147981419,0.34213912998637164,0.9017289377648525,0.1391323917528351,
 0.41613970449446525,0.8728626377347335,0.12757357587248164,0.518963369090723,
 0.264531744703866,0.5796144644115044,0.46443318604655703,0.4571188210880056,
 0.6532202574933247,0.9468216983464218,0.6915772472103255,0.2615259080685246,
 0.7125038325916879,0.5061377841660184,0.14994048706045338,
 0.33731571126361526,0.6307848767915106,0.8807421671338734,0.8119160503338346,
 0.6208309172326841,0.015078976713728709,0.9053954267753547,
 0.7881449759639481,0.09025475164134034,0.7600323745304566,
 0.48209635176176213,0.6401876121344645,0.7374902613114823,
 0.06633702215594506,0.36373030485591673,0.4525888848149082,
 0.17702300390595838,0.6669216442873935,0.6573679310999039,
 0.33664069850933065,0.5574741900583786,0.514921783457285,0.10721445549160336,
 0.7393662325862697,0.7560321886836298,0.9541774048523153,0.03114941517306591,
 0.5980007757638863,0.3166554979294356,0.057038737215083235,
 0.1007545236675309,0.9919051940192236,0.8128924390748009,0.30592967317337194,
 0.6426162831011366,0.7865078618158177,0.8982152444377949,0.5024042443778642,
 0.5107410107545138,0.20699711029991974,0.04468872446647876,
 0.37655399884135554,0.8305791568175492,0.4271601082059717,0.8413885186274745,
 0.5675066501261179,0.8699666664868335,0.26688835354192075,0.4840385826562852,
 0.5911660166319618,0.3534105035562424,0.8315853229889179,0.4946153774986275,
 0.5906484811516978,0.3485077968097551,0.975410636163843,0.4571675247785365,
 0.34088696207198765,0.9342867584017284,0.17801168567053904,
 0.05398217093765867,0.2757963877244043,0.38058259575073883,
 0.5855628803087707,0.17101462580662996,0.09919261611283936,
 0.6693760720358973,0.8654814226030098,0.406840256073019,0.9218055356761832,
 0.01029251201906578,0.40907094391407484,0.18687556029724828,
 0.6488738067120365,0.733514355561441,0.42753554276818984,0.27264427803182045,
 0.16780403538669342,0.8399453703103942,0.5489387146422552,0.8310319310372507,
 0.28207574740127006,0.8457031573716209,0.32091923102992626,
 0.14659357647015747,0.25297737540354825,0.3260629595417178,
 0.6110480158022213,0.6358245943619367,0.5037722721213967,0.03587898707315296,
 0.03611444795199026,0.8912850868107015,0.08165541460236847,
 0.35350085879647364,0.791411888287314,0.2867619936922009,0.8058533348327812,
 0.5039417113045248,0.4399962888174088,0.9568829383854466,0.4165200706159232,
 0.39637574150837485,0.03833031281647303,0.10439388322220156,
 0.5110700253576063,0.6657049243447856,0.4740042729944177,0.15924918900804963,
 0.566499080687163,0.661023310290478,0.1063074533204279,0.5849176180849525,
 0.2613573549025402,0.8631145811357627,0.4047142223993373,0.19352105102588257,
 0.14838272741741831,0.21845519417368453,0.35846819100819427,
 0.7328286814298925,0.05478707763255464,0.08944846956846131,
 0.13697006186908256,0.8770110492822053,0.21945150518714618,
 0.6994180665762151,0.08525431794959293,0.8633898344040619,0.8818742247597331,
 0.2564324034594616,0.18569665621030362,0.11310195241606236,
 0.7306168840930108,0.1886992420583068,0.933066649272329,0.84892638602223,
 0.9518120080015534,0.7569621302026203,0.1830268670107471,0.09617044851425927,
 0.5529974642680587,0.3481760502765474,0.20280931377119515,
 0.01998223548040612,0.5936840116485694,0.1965796409362357,0.5462380361732606,
 0.3198054262145086,0.10311469548440066,0.22477866373797428,
 0.3359214787719811,0.6526553326496248,0.14725296379206643,
 0.23955599614386747,0.08828454924977458,0.46490978632333024,
 0.1491862574862285,0.3360324521572854,0.9516915473583483,0.24234562370737078,
 0.6791789147984707,0.1549544674275366,0.7993283905604296,0.22762839742447083,
 0.04462539108442942,0.2600828763230817,0.33976537870953216,
 0.7082310373387088,0.30973498044878334,0.8173305561299077,0.9037018606284908,
 0.6687456802006114,0.16576293963247546,0.013642258653535189,
 0.6236911136217207,0.7943010055212025,0.7018862788923351,0.12476713888274371,
 0.6099906544430281,0.2097212256266986,0.9264154948640602,0.4346376780586597,
 0.20419064229486494,0.22786042017535346,0.285616740500122,0.8217204964034829,
 0.9603821321837296,0.4789802356518966,0.6061136972780445,0.37803424766546023,
 0.6911630793745006,0.7868472147848155,0.38725967533118355,
 0.14056417011656452,0.36933602732004367,0.9266909035348675,
 0.49515665513617824,0.3252492650413652,0.992900182440452,0.6096256274224396,
 0.468701058748101,0.23387361207129198,0.7312642083177586,0.6362652628710644,
 0.5272658548840643,0.8861041448522867,0.24711562756912442,0.2390724767436675,
 0.04956085770046548,0.947036028266445,0.9046954338503473,0.8036585357717085,
 0.6001597578448903,0.48055943771373233,0.5016925953438882,0.6948508780181971,
 0.48297808414083565,0.7496401903349813,0.017481105066104452,
 0.6271024030760725,0.056594641814443514,0.1858424677867787,
 0.8061139748238196,0.603631043342709,0.7680035634629561,0.16230769111179433,
 0.06601544973808282,0.23682672842942976,0.1496075171086676,
 0.4387702832674516,0.9737702053669716,0.9596563246237813,0.8166807773505875,
 0.06076580121995301,0.03511815839605947,0.5015522042913898,
 0.04257884380719201,0.9266446166164143,0.8630207019190773,
 0.41581406567225454,0.9583186397245588,0.007367545809080722,
 0.5746622373452231,0.28037241016287084,0.05263602557928537,
 0.9821675747027879,0.07272707341909812,0.35547365545772747,
 0.14989992386411366,0.034228957536616034,0.433360191585617,
 0.6641787217494473,0.37630983711122046,0.9665579837396739,0.1108760576232003,
 0.641684776830419,0.6039292363280174,0.9319784848960871,0.6619665357009155,
 0.7104538062991359,0.8357353455512222,0.9869407106707917,0.5060183952688551,
 0.38899796662305897,0.6333669015112496,0.252268500321293,0.05510052282022837,
 0.887982313033133,0.27409067630460937,0.15113672241463982,0.4810135408997269,
 0.9757552787171232,0.8124944729225694,0.809183041216987,0.6868620155954834,
 0.7824172515075553,0.5146054675942311,0.8945146709064238,0.9174799126938789,
 0.17299928201947534,0.5818128052509502,0.30627319849540435,
 0.7901996163750202,0.14056283454200036,0.3053372480404355,0.570271237013491,
 0.8134294737959092,0.6670595521830219,0.9843490388288867,0.6811265687971179,
 0.39967865314657625,0.7094369508672543,0.042558020290536125,
 0.6427793865797247,0.13430797356967394,0.09188247166669505,
 0.14535225284701225,0.025125562334986085,0.9918328108444231,
 0.6638875684187389,0.8667944316726031,0.18372843930588234,0.5442949158298005,
 0.9995615580262952,0.7750420920671846,0.3596360444980172,
 0.033680349593233516,0.3521295968868817,0.6293225813512614,
 0.03515305154890691,0.7041954320992214,0.8603782534822484,
 0.08877916836955557,0.26143314609084467,0.2691696230841747,
 0.7442055634193642,0.9585551882268157,0.132215517125714,0.7403935423700239,
 0.06434767657199625,0.41249986965713337,0.4260029710953569,
 0.5868481612706986,0.8520938718569557,0.11402791897212006,
 0.17457402737014915,0.24984539302503805,0.9682212064775917,
 0.3569477541188829,0.8709650480524176,0.5098152834179105,0.6641507257598841,
 0.6809144936034122,0.5604588898778595,0.37498399448898,0.811196241374956,
 0.2910285521636211,0.9104918799900934,0.008096063441748624,0.828290672659598,
 0.42598174646535014,0.3287804999394194,0.3708064666425004,
 0.06628398157538373,0.29624954626977285,0.3984550564398167,
 0.5857519803657281,0.17489428900708892,0.8162168872904122,0.2628180054654171,
 0.9313149128691496,0.39553990350484547,0.2889869158039712,0.8176209289136744,
 0.9379350664743403,0.9592119248689832,0.8073969528190409,0.9737160496753875,
 0.07099411480184581,0.37815841944185835,0.09274317077539107,
 0.48458087255400006,0.8889096137014783,0.9543467026894464,
 0.03485562187258806,0.10038097714755567,0.9606111751585268,
 0.8272263952908495,0.9820854623464161,0.4176701338380656,0.5479724358350729,
 0.7789717682718748,0.38453778609442846,0.7175193946514862,
 0.36787494660280906,0.22029674300071367,0.05838292416463764,
 0.5933395638506364,0.6795191534523197,0.05782805426129667,0.4869795399794923,
 0.6603092318930888,0.1764487785007679,0.6449036248211809,0.27845418200623895,
 0.7031005857840293,0.7530744139210921,0.6732052395737799,0.06857830225628747,
 0.42878753116789026,0.7800300904401646,0.3633993318108294,
 0.45915276268657335,0.022585919370393892,0.041819386371656275,
 0.07890822103356421,0.3762029154026705,0.9370063294460584,
 0.07599010953843133,0.23327996488055236,0.6473336193714607,0.345224828419323,
 0.9689335417637753,0.8107740977194721,0.8634999071579083,0.8229410756724419,
 0.09669929182714043,0.19372023876551514,0.40677803910849664,
 0.9727758151191018,0.16987569227197807,0.008698720468825138,
 0.9100302286252973,0.48670630837369055,0.38630613779451173,
 0.04703376182078822,0.913584490645404,0.44272175875892184,
 0.18605184549611375,0.822805888059681,0.9541365655331235,0.6570536178878923,
 0.7858308071092707,0.15384108389369722,0.28410793797245404,
 0.3602415274693852,0.7283533565860487,0.4686547972492078,0.15249272804494818,
 0.9541311271120329,0.2928609228257556,0.6715330393581176,0.34729319930399905,
 0.32070495534030696,0.5092197292730294,0.7302637357902614,0.6035023658888465,
 0.770604137916494,0.3948345004564142,0.07046345696325063,0.7257232085435371,
 0.6637960903199018,0.26912019451394165,0.5378367424677669,0.5260208544147418,
 0.11504045701306875,0.6360892287760658,0.13031842290944462,
 0.7814322155971084,0.7749669204790833,0.7126140030147177,0.2872658730089165,
 0.05402696566812226,0.5244113178899624,0.08374659798989958,
 0.9284358342486991,0.8633151357505773,0.814660660085679,0.1220630606971338,
 0.7864786333592506,0.5112710644015842,0.36033802801133286,
 0.35491535397797347,0.8374024711350844,0.7503053331752093,0.5394953944806199,
 0.37567477169992514,0.03309198200161789,0.8690627879562722,
 0.36308824275393137,0.7725710554220232,0.30607870407949844,
 0.5674271875916105,0.8497268139642898,0.07436372067048713,0.5884905255804942,
 0.778469111193403,0.40376023005243145,0.020477491005521098,
 0.5285838415687502,0.9195505041759686,0.20798557693376973,0.5548263819719905,
 0.5169091695267893,0.566122577521131,0.11051616294084066,0.7881464214485191,
 0.9169728239709517,0.031319621439915624,0.2427456475620362,
 0.20259029374860693,0.24782246599710134,0.43238862570911174,
 0.28266501372082486,0.9816268199686243,0.9827486237364399,0.8269279470633404,
 0.9746670687014157,0.3976948567939371,0.49239338149903755,0.6392035019968069,
 0.5923165315578679,0.6193019753881874,0.13838716515158356,
 0.31093432477407024,0.5898678264151902,0.45146692239750097,
 0.4665476812897652,0.6142490613686357,0.8834466205046918,0.647807348735414,
 0.033920828961348404,0.3180477127872424,0.4320653773405898,0.576768426492474,
 0.08891878674660691,0.4773024275143958,0.727987302476356,0.9347689569588455,
 0.4998340279430781,0.6157360401858634,0.8022904861723437,0.41507652778589454,
 0.12526155856126908,0.8590654565556706,0.39137935969954485,
 0.7264534218875878,0.3509836948889672,0.3179277643415105,0.5204916394409067,
 0.7866212117830546,0.8260994890524037,0.6454376573273247,0.5006356284410705,
 0.029361465497402817,0.35238424225913323,0.9283809584644458,
 0.6231995574717917,0.21731964998871467,0.15006109614097052,
 0.9064179404918683,0.8601938929535745,0.04804086311147282,0.2176051846944813,
 0.6579578182068841,0.6975475704819638,0.4979561073353448,0.5770077103677768,
 0.6137579864612576,0.7245034503067146,0.8057356434381163,0.6487192465216441,
 0.30214572675184714,0.03277544610627481,0.6738384473271829,
 0.8116584182600726,0.21078138487650377,0.37866311903556316,
 0.36678156045176313,0.03811924164538005,0.10042783816555723,
 0.37961443859732735,0.01729818101281766,0.022343990421841542,
 0.05089891328145302,0.030920736288417516,0.8566549088177647,
 0.43933260470656477,0.7049531387380237,0.41195632001664206,
 0.33245567524995245,0.9822761417620408,0.6232940166127952,0.875850049910533,
 0.3439235817615238,0.11501712080957316,0.5867337887069519,
 0.16353316306953225,0.23941463203628321,0.9860855761020487,
 0.028389055935737706,0.9877226441423302,0.6270351151479546,
 0.09717311220148894,0.3141994477375305,0.6893490718320491,0.7786716353004961,
 0.7952028738887614,0.2968945595298108,0.1236499596134838,0.3044679661762836,
 0.1042575614030099,0.9809677990511615,0.2912039330759687,0.66370174244961,
 0.7806088213461405,0.011904438461750377,0.563885342427261,
 0.10800877501202699,0.42804904730611704,0.7846028066851145,
 0.27449506519757527,0.15363449605951618,0.7291504785241794,
 0.8726530709149379,2.018464292946609e-4,0.8340674911105754,
 0.8229317059506771,0.16835025742877818,0.5650715464635478,
 0.16306319132476388,0.11308074292646186,0.35049949010894577,
 0.4639594024569955,0.7486382778829777,0.6952090381736465,0.8200391910895952,
 0.3683708060835579,0.4692654665968172,0.017268415182349006,
 0.8970222879667741,0.1314693195601977,0.06872863537859852,0.7190664867631779,
 0.16854849055443855,0.012440263633459181,0.8863618028784481,
 0.7893573949161499,0.4271402141958045,0.45355491019817185,0.7219148045896994,
 0.9803390588001479,0.998533841151763,0.6513542961285974,0.18882512347128133,
 0.876190017411816,0.6776996048908296,0.28961469507350024,0.2403879726211119,
 0.37013724876845555,0.9283267751369524,0.15562423806915804,
 0.011153325204206088,0.6174006215835491,0.38105359218457224,
 0.9473556916586302,0.6807604182798702,0.7458264843025116,0.16288052618750526,
 0.7054646082291546,0.29342190317281247,0.1652517653937442,0.4153858332031102,
 0.6273002782557485,0.18692436574700844,0.2515071728559657,0.6212797107990821,
 0.6940327267562629,0.7769240597164256,0.46157599852402265,0.4177651595096301,
 0.3146695817679225,0.8078825422678585,0.6641377186407754,0.5367452804369898,
 0.21700449492732776,0.24604549325462743,0.8368006701089312,
 0.5282423763043805,0.3963286007333081,0.5213903071775499,0.16909396742565208,
 0.36325973165662984,0.993813434441059,0.9608193556456412,0.20301625227060427,
 0.3960036718904869,0.2571947679082831,0.10726579286826265,0.67022565792162,
 0.032895787349083716,0.8637576266333261,0.3644156113958257,
 0.38873714636642387,0.3039092224345237,0.6816639590375463,0.2383099494774168,
 0.1640552546817563,0.14446059818432122,0.8549168297738526,
 0.07752051148418393,0.7762351942942732,0.23234481752355574,0.781458486777765,
 0.9149926400765328,0.6159352323737979,0.9395731592879504,0.8712343550339814,
 0.3755000496523814,0.37621257594133073,0.018334312708957134,
 0.07381888679899395,0.8187379677273678,0.9139874607701433,0.3734247881691898,
 0.6727793184105351,0.1184301009474038,0.23408837509220004,
 0.058152505133028165,0.5530716160585114,0.5522424982804426,
 0.6268259912569633,0.19833626823853578,0.9628298839481064,0.7759237280770586,
 0.4878391249246057,0.19509047874481933,0.3094776891945249,
 0.21345205856149196,0.021130741114975815,0.7360736147222571,
 0.615084530911392,0.697093006377662,0.17640005063729092,0.0873765068705119,
 0.19766185955902527,0.38251065748761703,0.35469101074289766,
 0.718885187847197,0.9616058124869595,0.258255073286482,0.132199824644037,
 0.7620538610734374,0.03952887867091137,0.4372845451138192,0.1371557986136125,
 0.6563105313349191,0.44810373509234724,0.4536165205992173,0.2874427023064542,
 0.14325008488893165,0.18761024411035776,0.6934203493396137,
 0.02141435157351468,0.12688057191788749,0.10505645685147535,
 0.9990665256504555,0.8650277097619703,0.6989367527849324,0.23367486811402005,
 0.4674897133856454,0.03726017873286325,0.23237457503268433,
 0.7189079075648601,0.7166391379344762,0.5485215884595408,0.7652182540858783,
 0.6587804178213592,0.45596985285870884,0.6807168472801015,
 0.17896424836895763,0.6718789025791608,0.12865745465568346,
 0.6566016056124054,0.27741685423112494,0.3590255222642289,0.9550849747227158,
 0.6780760754609885,0.17275808580840557,0.0013894324593675123,
 0.8079121855523628,0.5970838746141314,0.9950737362424085,0.9305108342427855,
 0.048352006624008625,0.1625273185732221,0.9322556493782324,
 0.46717844305846223,0.6061722550443319,0.6992383294055662,0.5088007981227741,
 0.05844968832790842,0.8388263617048577,0.8359537746436174,0.8875088119194525,
 0.9749789339427464,0.8756236754816472,0.9669521665222545,0.4093374317245927,
 0.8244045254881885,0.36044597917309973,0.370998849924119,
 0.047016269809683875,0.3731598545353716,0.0011182645012740577,
 0.20529601994037538,0.9203127660851019,0.22611332312603494,
 0.3211471168757165,0.32875301617483865,0.21457609682812584,
 0.4797379802337072,0.40718647196218316,0.5020313129423289,
 0.37185787526958014,0.45885084748710525,0.16819660797602853,
 0.2775338570681838,0.19331260389713234,0.4390382017550988,0.7057973802513662,
 0.5858740667953743,0.5798095626813863,0.6293752578710898,0.7778926598007554,
 0.950875074669334,0.06801186458014763,0.15940169857246778,0.795680722784042,
 0.5677048162881951,0.040439110151123114,0.06145254625103025,
 0.6997309025886608,0.9267285920101039,0.13137213836440897,
 0.020190103576281593,0.71216551499922,0.9040206972485579,0.1646824089633988,
 0.7036698321282838,0.740421246418031,0.41879569192413535,0.9818194280903698,
 0.22030946287570574,0.9438009657874122,0.7825158403814851,0.6005961896788123,
 0.6826152388888493,0.3957412403679834,0.8456393694759772,0.30023087587239283,
 0.6167665645686355,0.81647363949996,0.960388068558455,0.968866261321879,
 0.7406346716028225,0.2440481433773587,0.49689248882697656,
 0.37988760783302833,0.8909951746598754,0.060354224015173874,
 0.36396709556414786,0.5113665281633004,0.9365318177623346,0.9551987715356718,
 0.6362684741449098,0.9139491571156918,0.69273181130058,0.8496135361972506,
 0.49283374286308124,0.3390723717490911,0.9280632601285393,0.7367179013394806,
 0.8149679352373453,0.49552540228951203,0.5027178057598581,
 0.11662021738467754,0.3336849059927639,0.46751699274840464,
 0.8896547212799427,0.5243094353962442,0.8970066570469796,0.974173101698756,
 0.700548190637676,0.09990056505878098,0.5175702649182816,0.9165627353902566,
 0.24216530214045484,0.2998447411495794,0.8373971702287952,
 0.046582549389203254,0.6026969888424755,0.39939627458038673,
 0.7670708033859807,0.30057741678438654,0.8161796292979808,
 0.17200604842583533,0.9145979228134055,0.13737159995660586,
 0.6246356308597856,0.32128612790675215,0.8605416790359629,0.8315737870796059,
 0.8000365536164903,0.6345742744934582,0.6610116158074004,0.6087627178767814,
 0.6754383799281722,0.0740772843294093,0.04636334758393901,0.5728648546058958,
 0.015177628835922219,0.0095389198094461,0.5147451840949249,
 0.40967835120747564,0.3523924979195552,0.07944577377381223,
 0.5076153289917434,0.18942584610677404,0.3817006403610208,0.703461619981012,
 0.636042961852489,0.633371412485645,0.8379819797163582,0.7258119618590045,
 0.049975414854213085,0.32102468890384916,0.5613735276528337,
 0.027406203564940723,0.38761740922838506,0.6349818165211892,
 0.49914315225658556,0.6584333246390212,0.5186172694126527,0.9486245013186063,
 0.49900334271547575,0.08699906373949284,0.7603726134068549,0.739965092327168,
 0.8459401488576304,0.9068218912187105,0.8252137457928135,0.27515259487491883,
 0.009397751896729756,0.7976817263201212,0.035759745121156206,
 0.8345844333011907,0.002186513202543927,0.06121179151893674,
 0.34536848788056396,0.27981064300636216,0.28422942958407804,
 0.5836192566130525,0.12923999571830558,0.25725706226986544,
 0.10164328771493825,0.6072787338883245,0.4539621642404119,0.584242466284678,
 0.736404835112872,0.7371179126501496,0.343297365378472,0.5298375164583912,
 0.13690511473647404,0.4957270690584965,0.6172369573014101,
 0.13544886007761492,0.049583132485487846,0.24775157486811938,
 0.16312214280930892,0.0710291354429805,0.3006845911610897,
 0.16020241227285137,0.9104231564853659,0.47444120334949935,
 0.8401389157197794,0.8595492612681849,0.06870571206075538,0.8245008948397864,
 0.4717802941911228,0.5933222806901666,0.05413922351339839,0.8593535951142284,
 0.05609240412516214,0.31710344216846653,0.9893037987217302,
 0.3240128754318178,0.635165654222883,0.029237502554612438,0.509908714519374,
 0.4185797075208716,0.6217636576296182,0.15060013567365083,
 0.028671963627512786,0.6264769378331425,0.14601139684644504,
 0.1373909275802574,0.8744170119514648,0.013833771428509944,0.845143411113986,
 0.8394412700423698,0.003265404411340378,0.13149787772097266,
 0.9880275936559827,0.4606331525333207,0.1585116764341461,0.683876932475126,
 0.02513643114550468,0.7143583285192524,0.7269846473840129,
 0.022566279279093715,0.3991293289376624,0.9252287921176595,
 0.46880235970441864,0.5847357773902079,0.23446275303454966,
 0.6200942808916978,0.35186878444441216,0.2781545335905279,0.6626967541460057,
 0.2674506099646946,0.8068091167630461,0.3862550383090054,0.04971077127439749,
 0.5002912044124864,0.20891253137664934,0.03519238565728888,
 0.9129204920173211,0.5355178612157093,0.20552210872668963,
 0.18393849529306872,0.4922848092630081,0.045857793219644116,
 0.12230404427931107,0.8330174014060502,0.7524643075384654,0.8417289948753343,
 0.4114966647886923,0.9755962499010851,0.5289783433853361,0.141550033485222,
 0.22747515574045574,0.29968592775880054,0.2904943903766193,
 0.3337963021823327,0.6810711153402963,0.701389759649369,0.6051612803850016,
 0.1340516698130646,0.17523827204953046,0.49192076661534323,0.501005280243205,
 0.7763616352260164,0.5459412664954724,0.6043829768920406,0.46849857476172097,
 0.6417597268928399,0.10538538615105952,0.6133139518346599,0.942869071363599,
 0.6085086658686047,0.03486703802130631,0.921126712103277,0.28573954200509066,
 0.46609365091245336,0.6216838097100963,0.4768635926369158,
 0.29997837588781906,0.036261783577430706,0.13993526377374643,
 0.5178967662797144,0.8144509398369197,0.7962431167100943,0.0970480081376921,
 0.34004856874477063,0.18273436128829312,0.250666370135728,0.149604817599152,
 0.335769754510852,0.28827052449607893,0.30801821179124744,0.6986515167639413,
 0.26092893070622025,0.9726953964403618,0.02629457211907793,
 0.9300176489948209,0.7872287380866007,0.23298487487074315,0.1075670611664794,
 0.06909593771206723,0.5912563442627184,0.22427515254658836,
 0.18192330115713973,0.5681073375804848,0.1076311079088561,0.3722291762066936,
 0.47580152665529474,0.9958468704935527,0.12805557774735554,
 0.06882094110423143,0.8652311697142921,0.9653533669360056,0.3482131455525096,
 0.14063077761006726,0.0605780438813186,0.9949591968636772,0.7981025723052404,
 0.7481396397057748,0.8029415738442377,0.1860952366534704,0.8011749038238268,
 0.17031169501701426,0.4927454023819072,0.46081793021829787,0.565670702491165,
 0.9372354793710208,0.8214118692393841,0.23743118847015365,
 0.02916481553054262,0.6821444156260555,0.5559947153437074,0.3934183892073727,
 0.21839439749320277,0.6979160867124772,0.29912743160790956,
 0.6054135669671565,0.12307343055773756,0.9836579043501364,0.8510086512951404,
 0.822905402641386,0.9621854675044975,0.36921816492560733,0.15863422083698941,
 0.3722335296186218,0.07727383598377191,0.5890034152658685,0.8239285058257972,
 0.9811360993650814,0.07246137413585896,0.13448469460059953,
 0.1736129281406653,0.28942606209008337,0.7550976941883839,0.9960006130123746,
 0.9784250272192562,0.5731820063941877,0.0747808871523632,0.1326424797916954,
 0.4040773977630523,0.9847443757752762,0.00828433647819704,0.352860448218011,
 0.39723065001311364,0.8038836831072482,0.9389449740276133,0.6627218091059581,
 0.7367690060209318,0.6153953575596165,0.6094246572601176,0.36778051039962356,
 0.5003287443762523,0.7650431313157404,0.0025383173277075954,
 0.22696681054043855,0.15270726084704844,0.7079076162658436,
 0.4921694968707997,0.7590606148597536,0.061943637865460976,
 0.20589463952067966,0.432181451897391,0.500850127739226,0.24410635407400927,
 0.7655280364684838,0.9020489748777966,0.3286209802872898,
 0.010965356816473815,0.966182940605643,0.46195236466654055,
 0.6407066135565902,0.141170282653312,0.5628885243332776,0.6135630376981555,
 0.7312581099688682,0.3968478661787296,0.4236417557605452,0.6853640606968272,
 0.2519290115932902,0.47474242932584465,0.5633821682250407,
 0.08746908719037672,0.8729531884744931,0.007896665697799676,
 0.007727225461104492,0.3301918224295707,0.6757329824428328,
 0.10859711284697282,0.9263088901883536,0.9328470066079957,
 0.14368569495820827,0.23603369496847315,0.9129644955943488,
 0.7259195684193012,0.86754107214595,0.39332759723322464,0.9382071550188341,
 0.5176582371600001,0.7363319285159132,0.8493807757576224,0.6276571728822905,
 0.6841291583825891,0.09549786842479735,0.3841157591164328,0.5735823681472387,
 0.9243926725330098,0.39575785649907014,0.10592714810017922,
 0.4338665244268112,0.6584196785396732,0.504749495085035,0.07369100660007677,
 0.1188815558583931,0.6262267592148819,0.45104107116952963,0.7952754106465818,
 0.7131333891352574,0.6672493230841772,0.014168165387148046,
 0.7914107910601063,0.06262284477649227,0.8948854234496803,0.7527149439692904,
 0.2550843307542561,0.8883014097254349,0.7999496777515297,0.2947174904613853,
 0.5175698595533582,0.6741408271977456,0.42184827240726186,
 0.06283661407244978,0.9123582433478744,0.48674207743328435,
 0.8785407035081584,0.06679967792181352,0.8861836136247776,
 0.49434255374021263,0.29298225118152144,0.9454746018879514,
 0.9267871265053069,0.11072010188729053,0.9047394617650109,0.9969622307558587,
 0.5048150147960775,0.6711783405625199,0.7016741499334983,0.4187108001892784,
 0.16231176735383268,0.36831190627502886,0.9192043639159727,
 0.5775974812213237,0.5776387318643647,0.11093130716485378,
 0.29408260565958266,0.07969757132691813,0.49619204561010055,
 0.19084907882240865,0.13371783698850448,0.6062701672504467,
 0.6499112723909274,0.5295695289524851,0.10515291747622624,0.4669225453862911,
 0.2591221564218952,0.5026136024247946,0.57670428897309,0.4452771880161682,
 0.9080872388759238,0.7591731365574592,0.3487455324710872,0.29128550005568654,
 0.6263880357913887,0.8770143164808526,0.035841654340686424,
 0.02554334851727913,0.2695437389261841,0.1688397118733156,0.442722348232505,
 0.06684354291518602,0.487734882129505,0.8832688647050083,0.17535414113939396,
 0.2732473824388736,0.12400864963551195,0.57925844204638,0.3214739996293293,
 0.4948886254440674,0.40263595925112083,0.13166709752847416,
 0.2817799792119574,0.0032931688003537962,0.18191525758889493,
 0.209285480134759,0.9346003305437018,0.27678936880017346,0.08184373892944763,
 0.07087691679441144,0.8382743341360861,0.867420673685297,0.5683027217355452,
 0.7551386526407811,0.3870449731494253,0.9200631400952924,0.9114745372528295,
 0.80178659660505,0.006349682578783322,0.7657932001839111,0.18366423905084306,
 0.1619899952516266,0.37567254493796254,0.6533497511180426,
 0.49486709894422076,0.3634009678553005,0.5495464405441666,0.1562201017963033,
 0.2560601052335857,0.8454815747034684,0.9461908274201873,0.7260541670981744,
 0.8038141611877392,0.4493917700038378,0.7144517832316414,0.6996902932113476,
 0.6366235226664386,0.48056076646847923,0.24873309814139244,
 0.24558642608901127,0.5751923801771318,0.49583456857485864,
 0.9556220289039742,0.054425821379674444,0.9728141487008265,
 0.2450643356171296,0.4598124420322427,0.5900497108242706,0.185627258086275,
 0.8529299983739695,0.5731912232618512,0.5468011477729003,0.8371087475745265,
 0.17660138788683033,0.8926923474481617,0.5626014359949237,0.3510603547104554,
 0.49818192718048904,0.9351926648501581,0.3916169096529696,
 0.014532599388870748,0.8673991829511728,0.9895033887835069,
 0.3882251436344015,0.9952960767540182,0.38676153923939394,0.9473251345510015,
 0.20879419171997016,0.0985932924562315,0.5432136000893855,
 0.14354869627957212,0.5237492755659992,0.8397148348277426,
 0.44497143296630537,0.6576646163071684,0.4756727966997685,0.4261152072257536,
 0.9120196929518976,0.5877030255017657,0.3310110709616346,0.8038113124889718,
 0.8331432198473466,0.1669968646714135,0.8246457164108599,0.28195572850774286,
 0.9448932791331042,0.7565079674850604,0.9781043886344155,0.3366611713185226,
 0.14734258915510468,0.9124728552558025,0.07805068925940406,
 0.9984729908639887,0.6407445001311821,0.5552678565507281,0.6341172127072613,
 0.1384637535526817,0.35264499786864123,0.7178591914602224,0.917281129736039,
 0.6118256949906211,0.7087207098335341,0.11848787839407626,0.3015142464059427,
 0.837212343326232,0.8079408050131165,0.821009684489133,0.7415617547630149,
 0.9691493802673373,0.4794269301936743,0.02297001157323264,0.7730871191576769,
 0.22144427988714188,0.5109388493684256,0.4558582985462569,
 0.34595811063137616,0.22565121070174443,0.8696997612578046,
 0.8813298807679071,0.026251454671478602,0.09032415875573352,
 0.40540035579942735,0.26217466490703034,0.8284943032890383,
 0.7288329980531205,0.3100633413537034,0.8101962435318478,0.9971866007203316,
 0.9496801013432256,0.5981130030941867,0.8598164030610572,0.7795668710700057,
 0.011563049909334988,0.8514114889429674,0.5804868446247357,
 0.05373864397598227,0.8768540450923997,0.7985020056214777,0.5836862584642177,
 0.9629330687802803,0.19693533103705918,0.3691936275432506,0.8546338075325304,
 0.5454853648417299,0.1132079857363324,0.5406183396117964,0.05759168611086496,
 0.939803194758356,0.15755410030158035,0.14013098208728947,0.9192775145903122,
 0.46523001310935586,0.14041251674290733,0.3206202201547491,
 0.6853220725557103,0.1273990098639104,0.47944384302691734,
 0.33580835035567536,0.5009843440552794,0.32648738383008236,
 0.6477306246498581,0.2590757461649539,0.7572807976124585,0.539168220180831,
 0.5879459901962506,0.831226798727072,0.21536792640170221,0.5963178891716384,
 0.14015073681743373,0.6653492717734597,0.875801633733083,0.07097747422695488,
 0.8527649020192809,0.8546652673987191,0.9104222763426296,0.7973766245325791,
 0.25649600203551204,0.4868193668864259,0.19848296791894005,
 0.28619438399820096,0.16426207099873302,0.5537580668023446,
 0.1609687821713165,0.9399595204167351,0.4044822493455812,0.24670190260141656,
 0.0750951296215644,0.5571653931398148,0.7689674798556178,0.41081018262515157,
 0.8069998232444053,0.05056348536813937,0.5333232490289902,
 0.25035336094501626,0.5678429302321424,0.20526287649820185,
 0.6192274901384962,0.7544915998231149,0.6942091831840287,0.40692971502586184,
 0.2585723874803527,0.03494198507750457,0.8650996545319836,0.375483325889463,
 0.47855297877533276,0.6251780752633134,0.2501075320667082,0.9325513678454878,
 0.2115544821329043,0.8418017583465243,0.7241189674394776,0.3864795863849535,
 0.04546101807170522,0.49397356346449095,0.7656266164972874,
 0.27420123805340446,0.236649515758386,0.5887398346884553,0.5982905034648602,
 0.07377031793937605,0.39138839218303745,0.06597013593756551,
 0.6845247278482713,0.05364137764020904,0.20526938784379656,
 0.5064788820291748,0.35913044956638196,0.853788745652688,0.292282270888255,
 0.31576325403151917,0.2953508184762945,0.3473450556840946,0.3663665307765318,
 0.8897948621501821,0.8827296781972959,0.021496392892206728,
 0.7025770100007264,0.048032709335026236,0.6415677625220931,
 0.8846226145780982,0.7697537228634425,0.3064234556557982,0.7137490883080011,
 0.8564402272514441,0.47659417994244047,0.7559126969369951,0.7939978369215164,
 0.05714003410390256,0.24395732197243225,0.5533410446174982,
 0.9916932924175901,0.6034104497515496,0.46098240249196265,0.9075478369784858,
 0.8742201550030795,0.40068471685509544,0.8560985018852095,
 0.028211110678315943,0.3035133617418383,0.6399726022632468,
 0.4702748272634548,0.9390104904049505,0.9896498533572903,0.48161103505280434,
 0.6810292220222048,0.5842493987683567,0.5853907161915214,0.6279023949823951,
 0.0512119709071267,0.7634956020321636,0.23000572267176345,
 0.05341276285960406,0.578679454340515,0.8864039091631772,0.7970971145713057,
 0.23262131391375118,0.7111959517129451,0.1238794769805538,0.6663575270311393,
 0.11083514713465759,0.7309053217015922,0.39132932932387443,
 0.8633020934418629,0.5596449817069635,0.4879935520916524,0.43075662261704317,
 0.2736860837417917,0.6134329704569738,0.8365221330044941,0.9336857755915025,
 0.15093070561109068,0.29758994295620755,0.7219125426211752,
 0.36197118237401493,0.7056051212125413,0.8088048881677018,0.5077259864341342,
 0.7842471959855798,0.5089462493969797,0.2369798596583228,0.4516460522092014,
 0.7084749595884695,0.20006637561072493,0.5617518023397018,0.6450515501114594,
 0.3395061711971181,0.8555530178922112,0.9622466780252118,0.6618839401716805,
 0.7951621367708857,0.2377357522033885,0.8374259140749833,0.06792440426277624,
 0.8516117293196253,0.41617150447188966,0.5571521327908995,0.868969534151769,
 0.3363593887984093,0.06597576955848505,0.7762880508738491,
 0.11337880449537185,0.3718416186766703,0.26826016975609757,
 0.01797767584821186,0.6279277570576978,0.25851869996245924,
 0.8402828459708972,0.49487229847711356,0.2896615703113645,
 0.23373549568246876,0.07476180172356428,0.918447457785914,0.4728029949332574,
 0.9912612238448676,0.5684886672645092,0.7993336725934832,0.9984396878764111,
 0.24105116231482682,0.8681281817863162,0.5759469264029233,0.7123879475724254,
 0.5248288292139507,0.8241966146692569,0.9517583297724272,0.26901991012709303,
 0.5522395926197795,0.7813753034870317,0.6335782812924992,0.2790136054798633,
 0.9108422946653969,0.468190263193454,0.13362818004410126,0.2096421958737542,
 0.3649224559101675,0.4996930349269534,0.27149244432328634,0.8374610777845979,
 0.6703441271393001,0.11483283443104297,0.1287482961870381,0.8515742523193984,
 0.22720675204749174,0.6055486378323676,0.09028522039145337,
 0.4154975386086096,0.14576004518400532,0.25150965877612896,
 0.8338288167983484,0.7294771543806526,0.536257281265973,0.381983823133466,
 0.10409438729329357,0.1801120712321428,0.4755709148655728,0.960863816382782,
 0.9734547924023776,0.5463519019625965,0.7525877615492493,0.2357851495631147,
 0.3188060611581235,0.17675998490810108,0.032071703567374654,
 0.621356848406797,0.9556025329981492,0.21232592738059552,0.868833730418198,
 0.9744152955227313,0.5396062739688335,0.8290104635694276,0.3109404561096428,
 0.919604216548504,0.8195706897777023,0.3614129730308817,0.43431615647739363,
 0.9623460268532553,0.7792755911378992,0.8625214019419534,0.01555944212503424,
 0.40938445981360927,0.054430395666063336,0.9834950890587257,
 0.3096602500276848,0.8165893108324335,0.46675269228294336,0.9733776344685159,
 0.6473312442447284,0.42041796277701027,0.23878352101011935,
 0.9273010735203859,0.28477193666308387,0.7379047315569034,0.8283999378142429,
 0.71467944979322,0.8763699620001256,0.5086596661015715,0.03269500723995877,
 5.312827519665664e-4,0.37726842609866684,0.9687545627041658,
 0.5019184368616476,0.6409149766778177,0.6939278847625959,0.06030775310309244,
 0.12938149315140313,0.06626416585505657,0.309516030784995,0.7945441246918451,
 0.5768922819653075,0.36862702593120833,0.3436802545109613,0.3681747486867468,
 0.9663490047749341,0.29162235502242884,0.929130620903452,0.6112394867822697,
 0.13339071633500565,0.7837963190144733,0.49997759627453575,
 0.21348617553525973,0.8134052988093341,0.20583211080126596,
 0.00953600707377944,0.6849226574441334,0.1674995463723582,0.767470968370487,
 0.9632626551577892,0.13297044599403884,0.0021688170227905923,
 0.8672100350492667,0.05236736868783132,0.048240824693789874,
 0.06067061601499546,0.9956175195706347,0.1360354577223546,0.7216379122860349,
 0.5358416340443803,0.11869152312622222,0.43663652044010604,
 0.5480403368832489,0.3166534959151328,0.6915396281233451,0.8046078460676467,
 0.8373069350828349,0.9325376223126045,0.8454673895685538,0.742825840691936,
 0.5295884479766013,0.24143252485852862,0.5432636934730094,0.9923171295443967,
 0.19673008860291064,0.6352056124002615,0.12932456057958575,
 0.19664270635584047,0.2200816345750698,0.12757451506322526,
 0.5772110631668003,0.17225626221101376,0.09259563623612699,
 0.28645739234185197,0.6478210655309183,0.8457947563524324,0.4270569484990643,
 0.34845983873700903,0.19741179997285918,0.13250416136852983,
 0.4600508842152212,0.5389671188512785,0.06444549870065375,0.622831079600149,
 0.11547584400740518,0.7028380308491144,0.17359636974028803,
 0.6647156212068249,0.8100063623193756,0.1561151491986872,0.22117969430270223,
 0.13413304101297308,0.6597395092911553,0.2439306450439671,0.7258011423332953,
 0.6139087668182185,0.7698412755451135,0.11177968076961231,0.8577436112067399,
 0.9605203543518721,0.08803908247109593,0.31116677973509443,
 0.3907224600585669,0.14734508782523192,0.5247675691329065,0.7643933479273357,
 0.34986430081540876,0.5831479348191164,0.6376533467745445,0.4887088936129852,
 0.4932614962742874,0.2074680877198477,0.8544038565473836,0.9993187347577539,
 0.16160564512746056,0.03918613434736873,0.9341801604911373,
 0.7775359946414684,0.7873215959420531,0.9512983848019511,0.37552846398023687,
 0.503606705668566,0.9510104687403131,0.2944354638815725,0.5371498167862805,
 0.038804731791317826,0.38885073926868374,0.9030938267172647,
 0.21146046845198962,0.4158914842204846,0.13546475756999043,
 0.2752866194034149,0.060600328524838765,0.9627352800375989,
 0.4248531207865429,0.590323046156716,0.3314830498020641,0.6094516315279228,
 0.19981027792457995,0.6997696695472825,0.6081764549745526,0.8362684852003909,
 0.9312091630547545,0.5145023372565495,0.49008972157372854,
 0.05289273209782719,0.8733258201545314,0.09344478402619816,0.574803603755165,
 0.19732149740436844,0.4227220017405592,0.8928532149328459,
 0.37583076927177117,0.7041765595220394,0.6317864455758517,0.8272078670626553,
 0.32504431090863595,0.9060420614914247,0.35866577680766376,
 0.6927316556123196,0.3560334245452106,0.16272682262112248,0.6841239941768434,
 0.6829932123211275,0.47568447894251886,0.9555007622049291,0.5681761316051347,
 0.6918362535508615,0.14674621937694243,0.5635057078124126,0.2995957729098274,
 0.7753178122184508,0.5474935532901892,0.22554352125648136,0.4285031632908166,
 0.23067857996250396,0.4730919673388656,0.7094377164354476,0.7312231794512848,
 0.3282780540530734,0.9671750537605118,0.7398195821762124,0.9971002118973091,
 0.25596714280150523,0.8839384011723956,0.0812032778440347,0.5764195373616222,
 0.654092233071412,0.3226058678837126,0.0033328702216027795,
 0.016377483422804984,0.7753216987972992,0.6731790691055075,
 0.9566251187730472,0.1313498699254495,0.061651073957856184,
 0.47402121210023196,0.763463771453213,0.5317049686119036,0.6160648430047975,
 0.27353447017172483,0.3562519156952988,0.4337689022550736,
 0.17741812296372084,0.8955472418339968,0.747632736243076,0.510835079490383,
 0.5590294640895314,0.21550687302507288,0.5536405069914794,
 0.45744039579379303,0.5334284560547748,0.3307896158554573,0.3852129581000804,
 0.4607831897436019,0.258563127821381,0.934033927567407,0.8427173040031408,
 0.1435099936696833,0.6013323703802129,0.21418397537964218,0.6388206874231436,
 0.7580326983940823,0.4792206193447637,0.022294240661777653,0.550501274546143,
 0.5586468207373527,0.6558317676247438,0.46820303388580486,0.9913283397964209,
 0.8634917200196737,0.22103207273653191,0.8635069646718004,
 4.6950632632680733e-4,0.7197378571617179,0.40619826557517547,
 0.754687725830974,0.7735318902701789,0.9037615198464705,0.49613641367424033,
 0.4151424022600173,0.5432753101050196,0.060760879238956345,
 0.3201378773502963,0.2609357554339736,0.13705863029134369,
 0.49775725904139323,0.7221868806204945,0.9899257904436018,0.8107105990794617,
 0.46442121395050073,0.8397137526127828,0.8506119079782768,0.2586333757183008,
 0.4556163174651843,0.04494009918284381,0.6734232692259167,0.8878018627145009,
 0.2125124776794356,0.22648197952589477,0.7620053308980258,0.9111214105395957,
 0.9437735380136694,0.6899241589985429,0.13446266719105449,
 0.09745804208048936,0.751221887386843,0.8519845710170632,0.24409006877849637,
 0.8927842493515261,0.6700559460371407,0.7415940185273188,0.8797688298106079,
 0.06967555002200776,0.7560642534424732,0.08867262139823373,
 0.4211698091823347,0.4237534577165929,0.4399962544867495,0.8753963699373268,
 0.32721873412430513,0.22935831039186705,0.6567630541954699,
 0.31629581055055467,0.728992022392238,0.6859765395543176,0.07074628697473717,
 0.8190673355731486,0.9694943378506236,0.23127376354815987,0.3046461834846804,
 0.42924631861746654,0.6945602056900237,0.669010341799807,0.5930088665926124,
 0.4403681312906763,0.02584395808447104,0.17514450404586868,
 0.8487849103978912,0.532867673579705,0.2724211858983492,0.2726166230926206,
 0.025044972893442208,0.5656120736438243,0.888381656302927,
 0.33929420828909096,0.6586009640903286,0.3062597118991185,0.7516393871599554,
 0.49136677639103565,0.640720362816661,0.4750253328924613,
 0.055703455392277945,0.9481232892633968,0.6740574212006847,
 0.26080013730065543,0.10798030009345239,0.4603971602910759,
 0.9603671228895788,0.1924429350213751,0.6666111857674768,0.8813908365972345,
 0.634023793035404,0.7828796615994591,0.3067393368428267,0.42606579408153533,
 0.2882526843631158,0.3605509409599974,0.30243660808078143,0.6353091562502744,
 0.009710139693123288,0.9358999542455063,0.585639400010917,0.2495044041597132,
 0.8642729259150937,0.6091989561887459,0.13866993577165765,0.5124498781104496,
 0.5072317355950479,0.18501504504504485,0.012947125977796792,
 0.7203269197131854,0.3563038990841164,0.11879269398654946,0.5521210586441275,
 0.9078268272556442,0.7928980359620468,0.8168971756810676,0.877623383113786,
 0.734250209860511,0.5543710331511186,0.47113823693447254,0.9585195113332672,
 0.315861755905847,0.6266178343851995,0.9170064459645154,0.5828659050585512,
 0.38384873290815413,0.528441753149053,0.6504683301068471,0.02140492040048325,
 0.48145827070770253,0.6909831132342019,0.07422607578378737,
 0.23203645473801693,0.5412839553020442,0.6046198812228042,0.9810311580511234,
 0.5697461147207366,0.8265157674655464,0.36706858856451585,
 0.40776271743217263,0.9478290113348956,0.05756805066019799,0.725933653589063,
 0.9872557861722179,0.6420628214526136,0.04804342222954905,0.3496807370757573,
 0.3930012630110773,0.46366760819142805,0.351407699401339,0.5602602987638474,
 0.99646141876809,0.6819684960625775,0.22557426735541253,0.4439781243395382,
 0.3661834677269902,0.33034664231660926,0.023675424653056942,
 0.22836076548681528,0.9231513546662219,0.3898906438906756,0.9147284128105314,
 0.4845891486293086,0.7701051459185374,0.31887339064251796,
 0.45117061288672555,0.7337983675610664,0.24273248305696515,
 0.3036033996843098,0.6399401969126072,0.5675928302986477,0.2249103958861679,
 0.7746804799102546,0.7936458982123044,0.19223157748609265,0.6093943513577342,
 0.30240743230945977,0.39614353135476355,0.5449987641218519,
 0.11934696775780618,0.3666898322407789,0.3924745474685154,0.599671680540205,
 0.3437912591529714,0.18553628968176017,0.496322267289639,0.33606316195062835,
 0.8984818020949732,0.4601111120373248,0.022038399027036304,
 0.4581624806732906,0.9454790749028905,0.7229002534340876,0.9505874710371081,
 0.7972911699586921,0.3296707616637683,0.08238962830654506,0.2940617075506555,
 0.801764343527636,0.032586662696795665,0.3661093418369722,
 0.35739907938523174,0.0817947542752997,0.4172502347531073,0.8150661305227385,
 0.5954971461431742,0.01626773883821553,0.6704493329761061,
 0.27426284738308193,0.6554086642947741,0.28738778031916423,0.945229185575996,
 0.2363256537500522,0.3804145726040782,0.1426863485642914,0.39873587643560493,
 0.853191619267371,0.32005892998711294,0.4917353146363447,0.47799759499505257,
 0.07607390598632668,0.9939301166691518,0.8622235822503685,0.8940528170689711,
 0.07132391477356459,0.7984929691431594,0.9159779795117048,0.5358635168273237,
 0.3601891849644203,0.19915182602585424,0.41328778863562343,
 0.9513818205817346,0.7370487416851366,0.3759739699414282,0.852541512774175,
 0.38644813491935515,0.4104446840099729,0.7236008722153618,0.3492677349052573,
 0.6943893288763106,0.5378803141972268,0.37747894814849403,0.402523841253273,
 0.9164300684495409,0.3002273269047053,0.30729286569828773,0.6921905107805232,
 0.930510444453036,0.26475107754285976,0.8348735910304681,0.5813894241353683,
 0.29759575395269977,0.6369450219488031,0.9476756153845655,0.7378738213193878,
 0.2931035026438735,0.1345366153035562,0.19751320641523373,
 0.49863386809120447,0.6601828924637329,0.7680387645675005,0.6120897989807319,
 0.09340174125987444,0.1571988154982662,0.10362716775895198,
 0.7881414848234203,0.773171005917394,0.07175164647016252,
 0.0033345646082373337,0.836114283033231,0.7285322952935007,
 0.8843731301953823,0.9182939435666274,0.4567285540722853,0.8854303719224805,
 0.5268982989402359,0.42087305255563645,0.1529074587478385,0.8945007304185046,
 0.28919193860002534,0.40691337488256596,0.3353856131131314,
 0.6563059197570655,0.4368863793572707,0.320417944860081,0.4488815236895244,
 0.3019295482708032,0.21342161838456097,0.883976038841213,0.8590647081621221,
 0.4662098510589895,0.7065884597021999,0.32248212311148583,0.462511119174395,
 0.3278697934026199,0.38783091744296583,0.6842769642044447,0.4973603761890325,
 0.16937676727741746,0.20366423943511702,0.6516744285536976,
 0.08030038527461736,0.5569221695962195,0.391444378348639,0.9966220538508861,
 0.7056652456009391,0.6238637473602539,0.02183118279423335,0.7707817591210434,
 0.04120194620778317,0.1160373747068103,0.9444191526462241,
 0.06231173853560512,0.7720869007017258,0.025351978811640974,
 0.3599346960976466,0.4327379309959898,0.8840135223772522,0.7455982060806311,
 0.3719828846688653,0.12205128529617826,0.00614921295314097,0.536372094339185,
 0.7894711222557633,0.954041729793302,0.05049704462732241,0.05319765154181266,
 0.5816413719019481,0.42019759367817866,0.7669561605091801,
 0.29004784741488343,0.3632967832075533,0.09096828454225847,
 0.2804359704829723,0.37033719436756873,0.25482498587406777,
 0.46570805850476604,0.6159426010977347,0.11945696596096633,
 0.6371292542688698,0.6169582173352379,0.27149396769885525,0.5933452232000128,
 0.138479260598937,0.7930346981699002,0.37448530660248924,0.13825612003076249,
 0.3189875626755978,0.0986731052699541,0.15794956186650766,
 0.18524632393379847,0.6566863865601515,0.6139637343718904,0.5939973512110228,
 0.7007689474533767,0.2670999549069133,0.9305951701535127,0.5438803832004208,
 0.9105395193995084,0.8740922298021963,0.5396208908093916,0.8447675927692049,
 0.3402476838353681,0.5985657030949293,0.525710541061552,0.1682605578600852,
 0.8474121272572552,0.5357848652794831,0.6720872899460011,0.30734855168482755,
 0.9177302570790089,0.5859244720687171,0.14001265616283498,0.2123129000074676,
 0.8731229548472591,0.529444137881235,0.0691263549627047,0.4132594556796664,
 0.9752686153732764,0.0037452338729453416,0.22792417550784272,
 0.6661996393113226,0.30859285886528276,0.9104127504863693,
 0.41213040706718573,0.21543810010397046,0.9094359700533352,
 0.16207804364047473,0.5599053652107877,0.9168901073746215,0.9773796988741764,
 0.04648932200483358,0.9183700581341554,0.40683742045093596,
 0.8836444390986102,0.777420486094801,0.04632884779992774,0.8354616698194414,
 0.5452539193637472,0.7289447644225219,0.35364380578337784,0.6925526932859565,
 0.7410054820722192,0.5978929770782277,0.7144150363736234,0.04178523579829052,
 0.15842217842334505,0.62650586585009,0.8618755236675176,0.015648722966878692,
 0.8240553641477596,0.8837365688731029,0.39461345805983705,0.6919082074747052,
 0.8361600450466372,0.10556851947979273,0.7584324955713427,0.4328167143387628,
 0.3225677262973847,0.3699709031836943,0.09054008861229945,0.8287442771478516,
 0.49357299876120597,0.9338994468770957,0.5054507411821152,0.5119261176107509,
 0.28392595671566556,0.7515352764059733,0.8792828115938659,
 0.13862261984758695,0.9853142959288533,0.742895870950725,0.8329880760101365,
 0.8958517335472233,0.9528879617560121,0.1085108693517135,0.7516335544597289,
 0.4228620451630356,0.2578209553433771,0.021057854111497964,
 0.1308398538230563,0.15491447402233438,0.6532623829270409,0.4254850222251849,
 0.14998560686219464,0.6694561212637937,0.3565406348863318,0.5102128016614198,
 0.9715738141701117,0.0011930411645934003,0.18671841489854224,
 0.24773859867234183,0.8672738724302437,0.7572742684824927,0.4123453080652921,
 0.9918842122648378,0.5722768072324149,0.7858430689202329,0.7653456971863808,
 0.18695868630354573,0.22459296624608394,0.030330374494047296,
 0.3795736078220755,0.9675212709464591,0.3273531215327492,0.6877964612327693,
 0.7124378795382615,0.25109858329870627,0.5431063086198337,
 0.20647326186111747,0.12364616857809463,0.081758763389566,0.9426595142409938,
 0.03415554161631751,0.1688404622297307,0.08789703132357651,
 0.7165675579120422,0.12426481875380391,0.9386029139572286,
 0.16335159002269173,0.08820422605664935,0.23981071577433966,
 0.128998320823708,0.8126309264296572,0.8561016856187813,0.9894350123185147,
 0.7702774302703639,0.11759905623551481,0.4383526557963995,0.4074563303671991,
 0.47375833697613134,0.37932645745028937,0.9275581963331853,
 0.4724585627339072,0.9710076912566943,0.36336540367309067,0.5180319160112498,
 0.6988258942325372,0.657081564654292,0.5660020853363438,0.5730383855300698,
 0.13208885528333592,0.2199615259092328,0.3888776818061088,0.9523827798191222,
 0.8951690947905975,0.46868452252380466,0.5036853140148625,0.6176112593753169,
 0.3783039598368956,0.7070317193618156,0.5094039629544279,0.6517232013423369,
 0.1857549448891509,0.5310671330138117,0.8146303265391287,
 0.047522364933847094,0.9469080581379978,0.32188403696361756,
 0.8564007968857031,0.16103938222048564,0.6340263506910816,0.3462402641148541,
 0.5102450347927947,0.4981287354540215,0.20411350339485168,
 0.42653880177703785,0.6435504674318494,0.09573971060851671,
 0.25619386939108446,0.9907943872147672,0.28527812456452906,
 0.8570966841333942,0.11016766300884684,0.5019972854236033,0.4856042127203104,
 0.6115452395825591,0.6090360722184687,0.42167925031073594,
 0.47559370637069087,0.8452205905826649,0.8512410467468173,
 0.08761367549495735,0.6472461073559954,0.4203220706171942,0.8878839122815304,
 0.36015005930440613,0.6086106088936003,0.684648880173696,0.38590410233601524,
 0.38296377240036117,0.01644375070011339,0.7533454189604121,
 0.46191382789359525,0.43381596507660325,0.14769444279293986,
 0.48602478792480297,0.8208882516136954,0.18195779627821718,0.921822544278236,
 0.63628873447971,0.24984446491651857,0.2540893126928172,0.31538595586156903,
 0.27888618059722936,0.18446629075326026,0.492130880814841,0.9808399342328912,
 0.44584143344637295,0.7594381305663375,0.6773184788871357,0.0789574586041556,
 0.3491387846795765,0.9302119420246451,0.08890415623822168,
 0.37148266603027347,0.16879671443409572,0.7417582358392707,
 0.7816817015201971,0.4732412993459416,0.1151869885086898,0.5445136998553577,
 0.24532670038196114,0.18020323459170218,0.6223616069803041,
 0.1047804164540257,0.18557832915641959,0.9506174573324865,0.8295762488327125,
 0.35092435105411823,0.1949585578803582,0.04424933300559908,
 0.08133554349714311,0.28451032426927736,0.35138309484032004,
 0.504583476987186,0.2110505052776085,0.9089140242422973,0.9773655799812997,
 0.727123140427485,0.5354985070048667,0.9883551355975327,0.7421597080097286,
 0.19494675483744484,0.7558251400554704,0.5095486690382137,0.4037118860170257,
 0.34645133338550194,0.11243288596707302,0.2850420123902584,
 0.6421676544212265,0.36930581585944944,0.7462595839083416,0.6832143932879741,
 0.26903365416493863,0.1307158878654595,0.20916047227925139,
 0.1973658003796298,0.0713858249911794,0.47186273703866033,0.8618578941613906,
 0.6339936184102634,0.3346923942476919,0.14494534502569278,0.9618875010716536,
 0.004769244860112987,0.40213885858245524,0.25149476556260464,
 0.4428342673862664,0.1461701249506432,0.42859082188550024,0.4167929354954154,
 0.1452380210274169,0.3634506061654462,0.48445510365442557,0.7298198040427841,
 0.5052207296491354,0.9008574102410042,0.9484006844386887,0.5395423059071618,
 0.1159900821353883,0.1050104638928584,0.9118145086668048,0.2355521789714995,
 0.6660829810151853,0.4382305344386358,0.9602998598238708,0.7294942229773171,
 0.5820108570302345,0.717248213245554,0.7083132203385611,0.9824872477949842,
 0.8532709036067371,0.6355433431934284,0.9599424829239982,0.0883382013424906,
 0.5568212061186978,0.4064096063371464,0.5203178497770455,0.7842538943749086,
 0.5987929969345942,0.7233946741522772,0.9885928228397892,0.01762120867873085,
 0.07866641926562934,0.4459291532881553,0.735844481986449,0.7784174708564134,
 0.9175053803966505,0.9957702395155611,0.3469336254929303,0.4912901706193752,
 0.37267439425639104,0.8685588328829051,0.7358666750657427,
 0.08682111976917317,0.6805771820358428,0.31288875944643757,
 0.6946320212142301,0.4001887563610851,0.2826362231953594,0.8405453503331977,
 0.26175390841567225,0.13632857532143916,0.40030802745047644,
 0.5467253687123703,0.4298678464940955,0.4595824040417473,0.4704011547607916,
 0.9090020344046391,0.20805292756801008,0.354885066944719,0.3924023707401876,
 0.29365667628045966,0.7021307237303659,0.8768741367564699,0.5624731675976045,
 0.38246330520994876,0.7056749253892778,0.6256281745198302,
 0.15599476486662867,0.43393287070997955,0.40293032315965405,
 0.876953389186632,0.18559264319554258,0.19971847233851714,0.6517547111290692,
 0.5841381695715616,0.5453065244207634,0.5524325576706792,0.12722668942007742,
 0.8345137061064156,0.8682173866221377,0.15991763021236372,
 0.16572650989548943,0.3435053541160683,0.48376685977450984,
 0.5218599688453734,0.9282491271196238,0.2602462550210285,0.7191361492073611,
 0.903792775292615,0.6007776526948062,0.7481429727874197,0.7833628370398744,
 0.38607338168582617,0.5238140806878939,0.8517384057987452,0.6260884586723211,
 0.17805642813524436,0.5016781621177184,0.07950950897575149,
 0.4000022914811858,0.29984818049739637,0.2756627908108873,
 0.13525908054921298,0.9917225717006879,0.4074153034966098,0.8470399172979072,
 0.11625860649993314,0.8502961194037775,0.9111294544560868,0.5433469854031268,
 0.8140950275186046,0.8750414396965684,0.9697786220376146,0.22971686957965975,
 0.7498977340024681,0.6656150035362709,0.08372153356706658,0.5613741042927447,
 0.812176084667506,0.019044876293638713,0.49389339425801226,0.577583352805975,
 0.4011329868873087,0.3384680029160857,0.19650785422099226,0.3510265517508935,
 0.7643276519220775,0.7421316489095068,0.0589164185647465,0.5636976900813861,
 0.19234833961309106,0.5521431897329846,0.4182319616013769,0.4488076386109284,
 0.7347284458709871,0.5645749442906918,0.9555737731891608,0.36977308321077695,
 0.2638707489843024,0.8807390337627489,0.34923030340709593,0.6471681133156685,
 0.9112469040427301,0.5802901931983094,0.08573981413877574,0.601996331996065,
 0.8872639250477506,0.6252523361795558,0.5881802158538445,
 0.002529619026133134,0.19608884162786933,0.14393344750646841,
 0.22697455786588805,0.08784370071017045,0.02766239493996947,
 0.8593209028964369,0.345308282445961,0.04227201238111189,0.9174551369093809,
 0.6285800452733474,0.39152453039855306,0.4330918972841501,
 0.07408321294316655,0.4042647752499705,0.8073414147057628,
 0.49752685628353377,0.6725421189423624,0.22094722673666378,
 0.7148288069877954,0.09485215665807911,0.3634652849405403,0.6396696454626942,
 0.7831185638248529,0.1129589994623339,0.729129983560958,0.6286275886678693,
 0.8710696975369654,0.30179098385140646,0.14868315897716233,
 0.47585859534904007,0.9855058491914328,0.19270481726569677,
 0.4597874359361176,0.10883050694688767,0.5819028536673303,
 0.37950790591253447,0.01697016213000624,0.09211710248153349,
 0.9656130817114423,0.8930713559789796,0.20334206672178412,0.669824189190047,
 0.4781527128393591,0.5739739145768907,0.10139820486550799,
 0.49004148710902073,0.2217650810233066,0.15526477583088272,
 0.7318864405969931,0.7392249810687059,0.4310321431418497,0.14955416689261347,
 0.9293740693232442,0.4171375663086363,0.8231039385347776,0.22867267122636437,
 0.6358314697029241,0.7737058307498837,0.08723162591472744,0.5105025904030235,
 0.4369558802676431,0.12187749544654414,0.5063644594705295,0.2619410947920341,
 0.9520351759215658,0.28943425118662647,0.10885858073302868,
 0.7430964417050895,0.28269840528988444,0.21910838894521178,0.269552669073023,
 0.7566899975201564,0.4384316332889401,0.1353143554813213,0.47463966326825635,
 0.9545064807898175,0.6140368156593954,0.31437716726770215,0.8113551085097717,
 0.4388014643130467,0.22471266711655546,0.2877522294266016,0.6113511146032714,
 0.4627031863270148,0.04891539555605384,0.08654203287927853,
 0.10972363850696121,0.23519115730706186,0.9817851027013185,
 0.06784304352481552,0.333930798252829,0.021235061202899974,
 0.6586583520283622,0.5071647592424822,0.3241895212894378,0.3056329043909646,
 0.6409094176993897,0.5698196392569183,0.2775912204928619,0.4666491189336165,
 0.5855414203007225,0.022789501586778393,0.4611908886442907,
 0.48989296565816354,0.7918430846591378,0.20395451430114142,
 0.39927736636852695,0.7805250599324828,0.5372548911779942,0.7287506134039611,
 0.17747755226322948,0.7149278610520685,0.6549343332434407,
 0.035617209877012335,0.03772503040349995,0.7410578217771187,
 0.20671413325379062,0.5295222356374455,0.5077033084115102,
 0.44063581039232047,0.2630541668871067,0.030092259763695006,
 0.7394643826942042,0.4867184075167019,0.9836173981000889,
 0.018915119906773525,0.2689919730045055,0.8023834485459983,
 0.8523002860739992,0.45420828801637236,0.3576606372757488,0.2919341490840497,
 0.26714188942568795,0.651693429975738,0.27430176310540466,
 0.45218704702096446,0.5005351426068244,0.732676755889588,
 0.013981164695779924,0.7788650088515794,0.16872109285494163,
 0.9681539700555993,0.23698964943271206,0.27922484740175135,0.873646398551249,
 0.7148504755637521,0.6082594523287697,0.661701577720315,0.2725720282565488,
 0.18975178560388328,0.5029051513840553,0.6298647215780017,0.4610465013602719,
 0.18612818680489918,0.6088735586635836,0.1505564238665732,0.8465399030729381,
 0.8369944511467872,0.890858962467535,0.9905862124938033,0.7335831877219253,
 0.6282433420644645,0.7944364608730726,0.9746385573540165,
 0.015493013372845477,0.4224405787025015,0.5233753079409968,
 0.4121814542146569,0.24633477959095862,0.3306111628145181,
 0.28636388512500366,0.19623543297240753,0.6074729050881177,
 0.14491985004456254,0.016417395908207788,0.21330048435283566,
 0.8554467111040542,0.051763335441936786,0.883357449578588,
 0.008980420494970054,0.6486210811975623,0.9970506216553081,
 0.9166783989230349,0.08889092138301291,0.7363851151913315,0.8872160474946263,
 0.5820832293117507,0.5630222650888403,0.5032531593471226,0.38102946107198077,
 0.5587066153154404,0.9650628906337508,0.7979948755970766,0.6468894806021256,
 0.8506644047834749,0.8842471737775353,0.6456182194942853,0.2922092847823312,
 0.3230661361785744,0.132083655410127,0.15790595554557962,0.6365588882231569,
 0.8239527538892266,0.07517463120031354,0.6448183489103368,0.9546525207433276,
 0.244782096793683,0.1738297002489131,0.8007716266811479,0.14016843330542328,
 0.31088493815255125,8.369605216885789e-4,0.5589167516536935,
 0.17687686279553816,0.5608408617020483,0.7335982207275447,0.2020596723674869,
 0.49056982868250076,0.4361669686259575,0.36251088357060146,
 0.15961707929341085,0.11810478057167773,0.22323720453135742,
 0.5924499805240417,0.9394847914910546,0.2512504613165186,0.42753579756648497,
 0.7794492597436856,0.41478676953405325,0.4787519534015262,0.7754712586554541,
 0.6802460700696731,0.44489601470204365,0.10107322380654526,
 0.04447948290397452,0.058983460329651294,0.024704367459368104,
 0.8507839996299369,0.4566707943226421,0.7342184129856797,0.574608911120599,
 0.9003751552087363,0.7742059390874833,0.9786065553805955,0.5408752779451158,
 0.027925868152613664,0.7598888141233562,0.056348783366384714,
 0.33430141175237893,0.5280571830660354,0.13829373865946026,
 0.5250201619113042,0.07471291254966739,0.2915976803079299,0.5119991145214247,
 0.4396721188851117,0.24252480712008895,0.22635684492480723,
 0.23273371559570077,0.712558683106701,0.5147846271788582,0.03878586623087088,
 0.8625676092759708,0.05008948568778693,0.8220396616592909,0.942438501013258,
 0.5148559136398694,0.7294461689223313,0.6948948956559994,0.8192155376015839,
 0.5131787570673845,0.7889278613656242,0.9545115210633527,0.23267443033355362,
 0.47688782877047475,0.637506530070995,0.15107810622275286,0.6308828632185344,
 0.5899221671733661,0.4013807469701196,0.595248281285799,0.16803336808355374,
 0.3663524155924641,0.5554811427808635,0.9393228631079109,0.6299631018120619,
 0.40594301570867997,0.7165088312668487,0.9582194402634847,0.8703510140528032,
 0.912240460807292,0.1344922726893829,0.9936525088296727,0.3104562058214011,
 0.4009295843654528,0.4687915221339367,0.6294671933626255,0.17463738501339476,
 0.07074762710615667,0.6713952097597979,0.3521604957948712,
 0.015529803227576466,0.35359458256459786,0.18291294461711793,
 0.7239654523979617,0.9979610463555223,0.22063051982940374,0.376280246096903,
 0.0882504799992907,0.2291869927568032,0.8392910213174577,0.6551516681309679,
 0.6761318819912497,0.708788589723536,0.20853687936716891,0.9696695740492264,
 0.23685598699675858,0.439711243913559,0.9664248390089433,0.9570462598255821,
 0.9370933490753537,0.21030769644340908,0.7449821757893582,
 0.19340796409677097,0.9146661256851799,0.9472054683221927,0.8539959677608547,
 0.1994925637341075,0.9886586410413798,0.5399266519111564,0.36175587906191065,
 0.7158224096902736,0.604749624608449,0.6820322424470525,0.6429819317302882,
 0.5615593775476816,0.6462435550930972,0.5885013141207129,0.999909557261392,
 0.21479913708497766,0.21136263626742202,0.39031542084356663,
 0.22391615057607062,0.5063515421974658,0.04627170225580035,
 0.01586221513402941,0.721667302580138,0.310803197209208,0.9331467288552385,
 0.5357190992856107,0.10413054675680045,0.9453481533094539,0.3624566129098722,
 0.7042273171104213,0.3256043347229386,0.5830143686570783,0.3516887949140647,
 0.14270297120610453,0.21376364872247822,0.49544209434280384,
 0.6527899640185616,0.0013045462813406905,0.9039242202884139,
 0.21791569598312388,0.3773485904079491,0.426375540886777,0.738583837192344,
 0.37313390593876605,0.6347243843869051,0.13884566124326747,
 5.893951865019176e-4,0.6407779876537427,0.984667982940971,0.5007026865544824,
 0.6595165893649001,0.7231956844524912,0.9025448682100595,0.30652154470638726,
 0.5133262245000031,0.6612794201966509,0.7000155562888637,0.17033422008210442,
 0.23324714066512087,0.8709709268663319,0.20870979897512687,0.994514007733527,
 0.05400520462046943,0.10493420360473371,0.1981550801851376,
 0.9627694035169806,0.7053479821022209,0.5823847735548368,0.4426613111661769,
 0.24709798673599948,0.6879481535079849,0.5973474883110593,0.7659193328874068,
 0.9080350335012148,0.840634380779115,0.9428323324969206,0.6425353065462628,
 0.955237336763813,0.43146633803671586,0.3005995325149541,0.9095390450779006,
 0.28308515009868007,0.7576007911676171,0.3456699004548057,0.6041822113470096,
 0.6480027432022366,0.6979050082606703,0.7316324479195607,
 0.012873170055223926,0.33718010544150334,0.5475982497052789,
 0.8635909317751941,0.012249314857667781,0.5577039770852642,
 0.045590169712545925,0.15571358358659726,0.019031187333366173,
 0.8554439807528664,0.8399588762204572,0.9614206439976976,0.4517685796650699,
 0.2708978563789961,0.08261980188122886,0.17669186293274008,
 0.7219959316689906,0.8329380188592733,0.2918999632480481,0.21849542263956168,
 0.38901584200574657,0.7571781164049685,0.5480004404335663,0.9383292619344128,
 0.5630632052771919,0.9774475730573189,0.5978865666037553,0.5669637737546306,
 0.16806010630452684,0.09683897973719857,0.5680125080458662,
 0.8773199447384687,0.4086655354978418,0.7586149330468858,0.4976364475247689,
 0.14235447090960685,0.9838425409951806,0.1005523187813937,0.7096626016563687,
 0.5473872994991004,0.6532394341303185,0.48858243562950543,0.9939475098113996,
 0.3635157047128136,0.8962032715023406,0.846453398192688,0.9831675484471797,
 0.592965036686824,0.3617657004674246,0.29106967967093633,0.8912268414731302,
 0.5989105445789799,0.574422602549733,0.9172666297957953,0.8523631929877642,
 0.9353989475776947,0.05229790078914709,0.13490633310164946,
 0.8079382560830387,0.187705685811831,0.8120144002889871,0.2428142214922795,
 0.6120170859847336,0.9427653259285146,0.9575173578986087,0.5661231308171485,
 0.03335835679807575,0.6559297450555269,0.9464901691685251,0.718542795025627,
 0.7472571489956059,0.6280079940764308,0.2636630773993138,0.5017751277724142,
 0.0013663527467271397,0.2691242649143931,0.11167142261967256,
 0.2858267364395961,0.6413603910721927,0.5033688249076943,0.8685971562984829,
 0.5840886433637307,0.2660186593358943,0.9097693168711614,0.9764319905283974,
 0.9686352833360321,0.9849615155491002,0.6575518770457616,0.09094350255538164,
 0.2563089010935805,0.22996386871368757,0.38567685850448286,
 0.3283494631972995,0.9771767879028833,0.4058533439355121,0.8315672606014801,
 0.22854762117423433,0.9758157393623809,0.9202707121011778,0.9241437970020673,
 0.16634087777411732,0.8545530988114814,8.51594101397346e-4,
 0.7434730873110793,0.10057040464541023,0.36700119268998155,
 0.13117405675579707,0.3717454232952848,0.4836768852655017,0.6682481868032843,
 0.4117446695005431,0.9938676369117134,0.7951063180107226,0.6927069849865096,
 0.4459004216833875,0.40213709879209203,0.6978332797548784,
 0.44059222731043857,0.6594920677922617,0.553437889664608,0.5318627151327947,
 0.7579104775911932,0.9896869505823349,0.9567253290916352,0.3609334625442573,
 0.7584509402821343,0.6032941736121816,0.48244269451569877,0.7942507794749734,
 0.8852970185436225,0.5728423976266362,0.868013363266577,0.5337128088176892,
 0.08506576882193684,0.9829808511155652,0.34675328476914413,
 0.9550792751909071,0.30335193456080634,0.7343770809363428,0.5727762964118933,
 0.745719261161159,0.4781733174888214,9.272582963917309e-4,0.7793119988191158,
 0.5667124663682118,0.4536437083154826,0.9104160729948186,0.34606820156414275,
 0.01331667011865778,0.6943036649476495,0.3728949187207733,
 0.13907683395995774,0.2264077357094879,0.003851163401672575,
 0.7485536325308065,0.7580593218257641,0.3931048046778164,0.2310219730862464,
 0.6504970220066532,0.43719093379540386,0.5141005340382799,
 0.08090257562437075,0.23518991081406604,0.4003517286545428,
 0.30330994311684656,0.3467167651961355,0.22003769590443578,
 0.5214916460906623,0.9346295159692763,0.37092994755414954,0.7945314961200988,
 0.7468244604527958,0.49028695440435843,0.6812918907642801,
 0.28467440403647637,0.6110286990672668,0.6255892317078029,0.8054993119243379,
 0.03770016936986975,0.8300077710527305,0.22835731278451377,
 0.0022038136019377763,0.13143657273993192,0.3094833248541716,
 0.08148328512069247,0.9238045838488356,0.882397459812986,0.8011374503391777,
 0.5910811955429864,0.9405573114130529,0.10373552699914157,
 0.10180101752235893,0.12711394233908313,0.22020226057149406,
 0.3473625043826677,0.08139982305129267,0.6419686211525674,0.8813103389649986,
 0.9289485094252466,0.27659843588065336,0.2142410275978861,0.9748126725059458,
 0.9072190220067868,0.3194122200829558,0.3508137228493552,0.5417513392399639,
 0.936556880669615,0.3599026654968155,0.6153352310406891,0.05785205403586091,
 0.5096659950926178,0.7559323790524539,0.9757352135300169,0.8122222825333172,
 0.7416740659835062,0.7180353664015316,0.584871235566174,0.13990166038984486,
 0.9938817887829837,0.7220169161534264,0.7525452733119022,
 0.0033616217174728957,0.2439118224897252,0.8238945743720005,
 0.7259338371440816,0.6416011489509368,0.8487594322747157,0.3370090900684042,
 0.7175283050203369,0.9071648561476466,0.637062958639067,0.841953680737589,
 0.005142603303305848,0.31455241694456937,0.38032674403228894,
 0.9190582347446199,0.4307103566097775,0.7112126441495485,0.36162707871354827,
 0.10949495960693356,0.3537129239038431,0.8878246130594036,0.6278604986439511,
 0.7460127351101957,0.08680686684840133,0.5967568199692801,
 0.05598503946087685,0.17457312808567804,0.5914420568254966,
 0.26240051663435804,0.5252140167084387,0.6076530178311623,0.6618700958164094,
 0.21778100715082538,0.5468500682167211,0.920649583662836,0.5823220459697238,
 0.12978624246743342,0.2019820637881381,0.008673247693745334,
 0.37028331932482084,0.4230088132306805,0.20919943421984255,
 0.34581591730334116,0.4017971971804286,0.2167148118247293,0.9151684980010647,
 0.08411238464599458,0.28771330323187505,0.9737278190948567,
 0.32913131396475626,0.7020830410382345,0.39525426607383374,
 0.7307278623649552,0.49536813404141156,0.43352414053427035,
 0.9325859009203462,0.7554855382004323,0.2912851842778984,0.904367637307622,
 0.265893724657965,0.7641075609333222,0.7094863489153019,0.3544192203155295,
 0.4650612677677063,0.460346740400164,0.08469422655411962,0.899431183319797,
 0.22662820257744176,0.9662057401275863,0.31882498410255156,
 0.11337903381537073,0.44998920705201784,0.5235361987485442,
 0.3277828673416062,0.9721334046250996,0.8707192963609626,0.9933322987238237,
 0.9617091098149051,0.5638933833826555,0.2306486816502593,0.8380380444181101,
 0.45328694284944615,0.7251898254700079,0.9653498912973387,0.7602195427923892,
 0.940740533402017,0.4399220031872304,0.60326504480408,0.29397170406682105,
 0.6543001866885838,0.05320143412798972,0.4213124508945002,0.9645383052531763,
 0.6815210121908295,0.5212263309492419,0.09937081740751808,0.9528966503522268,
 0.006715009552898277,0.43570639287846835,0.3943366585627175,
 0.6105726567277564,0.2306289413727075,0.65952342603198,0.8101642420930024,
 0.39004510420741634,0.7266909391659726,0.5523232612353286,
 0.22470779712558997,0.2626357544566361,0.41900674107950364,
 0.02625414063931719,0.1631462181284853,0.3493894247114845,
 0.10621287973786164,0.6690242890949112,0.0028876364292855783,
 0.012108168308852085,0.09526540077302315,0.016816785838698167,
 0.8035290222557925,0.29758828546495886,0.9502372091441993,0.6081866478853989,
 0.3011434926225739,0.0473556722128633,0.04746924824974763,0.5999667303759216,
 0.6547501880958746,0.4618030646994711,0.6504973087293157,0.13727527176166854,
 0.7839188825864876,0.6795471758474178,0.1109363676322872,0.05499741624201315,
 0.6329258660765545,0.4209664183036247,0.1032284306356609,0.6531389353779449,
 0.9061842801151838,0.3426867029004105,0.4339974785622085,0.12984236313907505,
 0.33195893914893937,0.679983684187784,0.3428030298279908,0.36542365955012146,
 0.9986878506865834,0.1185050947752927,0.5387914919530009,0.3294955057128597,
 0.590226084093914,0.7748939522239698,0.569235238856439,0.29541091859618707,
 0.22208054562928814,0.16476868714220594,0.9940556994080355,
 0.7194900771163804,0.13419338646686274,0.19559870954640246,0.245927863381334,
 0.5365409549856597,0.4267799092414223,0.635835518686872,0.16248391007427976,
 0.4956692095625096,0.5462162991140649,0.9912676787579414,0.3928724024596094,
 0.32276899779526025,0.35844572801862673,0.5110846205544914,
 0.10878717282644879,0.4965273300699795,0.8940409927908023,0.7819426313866785,
 0.8686193625862331,0.5131702848629176,0.4408285492316354,0.4563025990342404,
 0.905884603966459,0.7634230353000837,0.5616549253888413,0.8635554235044338,
 0.7949933370082041,0.5171976773829148,0.8786035896828376,0.7304502423003534,
 0.9454151210132934,0.0870982851455605,0.31478178835307813,0.8860793387187746,
 0.8495668573067507,0.12868113487767407,0.1139044829132504,0.8729709534133794,
 0.2899981400015351,0.7495818148218845,0.4522858180269166,0.14876481376714223,
 0.6818703826064123,0.45458144693604796,0.2914607679085054,0.8754603358222629,
 0.3084011299964109,0.7180150649143514,0.28293784458999993,0.8769392320645528,
 0.4237397465637791,0.28888681085613743,0.41969724874312375,
 0.27198778064101503,0.5624447302798304,0.3120518461179326,0.6099481326287295,
 0.8610515430431807,0.10575959845928284,0.5274017127347118,0.7623224596540237,
 0.04242204725734855,0.8520299993614019,0.6710760218682034,0.7279057314482928,
 0.23107450184406164,0.8663865058394445,0.11550879763091115,
 0.5571552360119414,0.12454805294206417,0.08882307758712782,
 0.09043015443951907,0.9914702388988743,0.28032452774029315,
 0.4363340022552866,0.773380135668051,0.055431940948242,0.6824604065663499,
 0.1969071797382218,0.6666870669649998,0.026013718184210055,
 0.7155707079625048,0.30363742333475796,0.5603445503029829,0.5649613410296754,
 0.8375071159409182,0.16531748669655022,0.40531011210592693,
 0.16881749701212811,0.8870635261119603,0.3880472493850311,
 0.12650525908633514,0.3354982835451569,0.8229828556244552,0.9212537257986814,
 0.678283756429494,0.20020956218758768,0.6470627804472453,0.5813758451034038,
 0.15645874946257643,0.683652234874932,0.16293481881264538,0.557058672467096,
 0.2775703963065861,0.5319453384926827,0.5622887070426894,0.5391790461148935,
 0.524003121708177,0.9159873485894677,0.733621800141163,0.8148679156389503,
 0.18029458450806424,0.45436603846759216,0.15775731339038734,
 0.7105539787968366,0.5292906212063994,0.23684897675459793,0.6120579311942101,
 0.20045197951999205,0.6901547594341021,0.08235645485751553,
 0.5619361953379735,0.9528904281956717,0.26091044327982527,0.4380953270487783,
 0.889721044339864,0.6183436677011049,0.5934489794853364,0.20859917337371892,
 0.6328952009643516,0.36845971313198156,0.2136166613107775,0.7420414278298606,
 0.4147385333791681,0.5500362556053984,0.042168587521781875,
 0.22107382388557706,0.4657766506050963,0.9454590902855795,0.3909461782486462,
 0.7913453854116628,0.4408585865092228,0.32825502833534115,
 0.13998633001750072,0.4626610756787074,0.5563287372588468,0.4211821860152962,
 0.6601533186832376,0.44532377492756037,0.4792296867285928,0.7639067939971547,
 0.34284288280075104,0.42346956424582816,0.680774080560734,0.7364087634800701,
 0.026787367065233214,0.003213640614904034,0.6745120150470875,
 0.5918356689049321,0.1444546244278101,0.6364928583397844,0.6471560576492833,
 0.14771714207714315,0.24433254928797643,0.7551617024829178,
 0.7563576913589549,0.3839591378906012,0.016803027289725048,
 0.7361126440752036,0.345715443814423,0.7937814145092067,0.9186152153010915,
 0.674150760393679,0.12258127276436326,0.907190730691257,0.5636139100921751,
 0.7626254548614908,0.8397278678186145,0.12663111318014797,0.9891559672138992,
 0.43899357657969873,0.21741722586611156,0.9060144955973324,
 0.5789388646779589,0.8731360776619699,0.890741122873178,0.8403764495286086,
 0.548902941887012,0.5112942691056139,0.3627435085819555,0.08508466600951181,
 0.9377350146316826,0.24759095092958394,0.0015139079662820354,
 0.9549095614210674,0.9936712767096594,0.0954383400087262,0.46621611972455224,
 0.016001994045719115,0.49674714731748415,0.05842340202712293,
 0.6858533856663769,0.12303176068457977,0.988601043632763,0.1974375611872614,
 0.9772646710513122,0.3873820004385047,0.3737337130422179,0.9468784272592309,
 0.5797033584460893,0.3165950318315397,0.13120501698396914,
 0.18577938241617897,0.8008008436038717,0.6898408276063875,0.9095061372807036,
 0.6830719995975287,0.12973909428472852,0.6745723910774245,0.4830446351417118,
 0.777055074496644,0.6256532742950301,0.6847414960897256,0.8708168168352803,
 0.600447922185666,0.0059480337776295356,0.6756325148257183,
 0.3504912511717284,0.08126666852894382,0.8388763448634793,0.151230868492505,
 0.6823746401303419,0.618753809540198,0.13520271869834444,0.17589688340218768,
 0.9971474436554343,0.9532430994062356,0.29170375834664564,0.6867314761537688,
 0.5922364603636586,0.8920639197261169,0.06882818414214664,0.5649298654492849,
 0.4827008472154203,0.7540757321731897,0.22359607443266583,
 0.08890940485391863,0.8063757215787318,0.7535744470040935,0.5346040705191482,
 0.4503934916746196,0.09679369857689457,0.5947728691778785,0.5263457037206063,
 0.016642056964208507,0.7910293605377914,0.270541400071576,
 0.32101457014636703,0.32338126189514715,0.6142122631384868,0.110917105873931,
 0.615887175925018,0.11640493229825244,0.07156436414149969,0.7787873199263665,
 0.23473996663029362,0.11432685573489265,0.629596576281171,
 0.28345139545846354,0.9079858313193024,0.3016543252671192,0.6040266088710523,
 0.33950385260148574,0.9824593237396992,0.85791419190628,0.03840116140926286,
 0.7038313420153272,0.9490143196482435,0.5907951678183624,0.4476085908220895,
 0.1247779333344845,0.7573049762948374,0.4359098483517978,0.11783691544231001,
 0.8520824940648595,0.040399907240657895,0.7630570304990341,
 0.7535698937923822,0.6743865792425231,0.1522746529886715,0.8703389198320769,
 0.6346118493654315,0.9105032829829676,0.3745172218245877,0.7252054672902267,
 0.7893361543994764,0.6222432632636715,0.2857922016993012,0.38303709871465685,
 0.07466717750888918,0.913305474721541,0.7108075517434889,0.28011263301356437,
 0.8065687731035875,0.7234232601382449,0.05166980258331322,0.7119119707148197,
 0.9665605957441971,0.8572108619531765,0.6028710428403073,0.7691077158814232,
 0.5375364663296472,0.5444951010049496,0.33423533168963804,0.5113633662088891,
 0.430718746441746,0.821577010925717,0.37302485492507387,0.4105437616942007,
 0.2327210873044122,0.434753432696968,0.9272702197676955,0.20606729112812605,
 0.6853972012717108,0.4490432255776804,0.8164105091268606,0.6955811159881897,
 0.9695387084313016,0.08380696410921373,0.6733038571539885,0.8228315212872394,
 0.6049159919976371,0.6703517772250521,0.9685477513374379,0.20123419272390797,
 0.5888863548515432,0.3469692994706025,0.10560084116697988,0.3859493473995883,
 0.6794084228284881,0.08539525170085827,0.3469043545027941,0.9053878770079072,
 0.06319875387310947,0.6293386622297884,0.989931839517049,0.3106889377964396,
 0.07486193649166573,0.6457689886643329,0.6701723051486083,
 0.0012438934479098718,0.5973568964609294,0.8803146711089133,
 0.7747774832038274,0.0838774573987483,0.8533538699630774,0.9338174830825737,
 0.8908954421810923,0.7755880884773264,0.9986923024645036,0.173092396818046,
 0.8540605974834696,0.2276117866401074,0.7278899019610449,0.5826324908941408,
 0.07452349708510597,0.5975704490088459,0.6650440863418534,0.969387911930545,
 0.1300886989641511,0.6778466701337805,0.8339723148102112,0.7978100223522688,
 0.190130980636674,0.735744294013614,0.12694622604055983,0.42488759325856706,
 0.1801030949349095,0.3363209218058898,0.5264708736119178,0.2564087423178727,
 0.03993363773267067,0.43445170188867466,0.754986948207493,0.5067780455864216,
 0.20658971795830217,0.4730030336690013,0.05249596016743474,0.752305648759783,
 0.44802916822210015,0.7424937941092988,0.5079506522588306,0.3184711126230836,
 0.45793038815227305,0.9761933134523937,0.42033040478290695,
 0.43796442968743765,0.2939442600744657,0.5080841800232581,0.3154277411666294,
 0.43770094143905647,0.34793451539561593,0.558037695916264,
 0.04868455690834306,0.6548858814848174,0.5754946484495318,0.21661847690545,
 0.5674864047155641,0.5895020409935693,0.4974038089189008,0.5810776489270537,
 0.4426172282817381,0.8314572140966169,0.3673998649939151,0.1580172028589909,
 0.9174385772415089,0.828016225442001,0.03226197154008448,0.6648554439405423,
 0.8326695595004808,0.7207892996810066,0.15758284693669233,0.2375074877217871,
 0.1250874543950352,0.5365565353458042,0.7452302794618106,0.39662631912127455,
 0.06538174739713087,0.670179271299316,0.9619467825043193,0.7955569620722909,
 0.7974933667778329,0.6390899703279802,0.6356454615746259,0.8192161752896263,
 0.4872367628750053,0.7959403403479894,0.4698010286050949,0.427555969003652,
 0.6673801529536837,0.04324182059628179,0.9948378549924451,
 0.43543000612775296,0.4048965035459887,0.222361618026911,0.7669808336935758,
 0.5257479525413494,0.5405757216586488,0.6803623785198866,0.02543003132193211,
 0.9056050375660502,0.09969731461510356,0.7929921365645572,
 0.09624355373697813,0.6388850505876935,0.01965032742980577,
 0.05580968007120002,0.4572449487758836,0.5298045605840765,
 0.49865075800862435,0.2195382004761509,0.897002998019805,0.43831073613411564,
 0.6735879889909223,0.631750015617159,0.4186389869536423,0.8383666971544617,
 0.22799278521567468,0.8580267453732522,0.14316749798848816,0.966906687573009,
 0.702719543869213,0.8742554240688584,0.23743052910190232,0.20763002463595093,
 0.6646621375771639,0.6829349477820184,0.9203622297996913,0.6197680824146803,
 0.10130433083138346,0.8772543407179916,0.8110445289993521,0.8295156161625179,
 0.8825334354912567,0.5925630630957122,0.8437486024407014,0.19875382675159992,
 0.6730299918699217,0.6192235211833883,0.0874339743311342,0.8819757248359941,
 0.8059012884083439,0.008911565356418127,0.08834418171020697,
 0.9830382387547396,0.33062889247242155,0.5262280067452612,0.3273711664022686,
 0.8936655038701682,0.14729016573145337,0.5106339743341649,0.6213681054257041,
 0.42635849704581863,0.33462397813769307,0.6851650429409266,0.708888665284417,
 0.29270901892206114,0.5115720219569373,0.44619661229839447,
 0.3962766734938953,0.6970198261344117,0.07540769957862725,0.6910641407925007,
 0.014414155623015401,0.667596192591881,0.6358748050903434,0.8990910136349635,
 0.9435453230637083,0.15785373044633033,0.5618731276693726,0.7821615262438634,
 0.20097698011522214,0.41106678914660477,0.5905553669200185,
 0.17248631432775396,0.7124368003724967,0.3449525741600722,0.6540419348719544,
 0.11983719925000003,0.02859036306712104,0.49852015659324256,
 0.2795775309015307,0.2389137032690093,0.6815202475747124,0.617038053109087,
 0.15818800459418314,0.6723123633099944,0.36361285954228995,0.874002724437441,
 0.89001179322364,0.6788529521746038,0.6890099802577137,0.37525021059121344,
 0.8198531156606997,0.05541841132910441,0.27112083375047813,
 0.060093700586892784,0.27824748029415414,0.3864236859156349,
 0.802345116856732,0.08951767974055824,0.9772526856464997,0.9534651223597348,
 0.10677932257575418,0.43670494911058366,0.8749209215425308,
 0.3756196020960192,0.050693949577456654,0.49344440165490355,
 0.843925876768505,0.5756747340043576,0.30238092088501756,0.6071254685438489,
 0.3180453728071766,0.44667755139307364,0.08995476453567885,
 0.1642749070422428,0.858141309919608,0.03376422432486259,0.7306601845080828,
 0.17531799421117233,0.42730617569805285,0.39012492330558735,
 0.9692681963066131,0.11245441921534749,0.41555905588553543,
 0.22701528317797248,0.7477995026057775,0.7994325807434047,0.6846473928344291,
 0.34966771319438483,0.8169572543230773,0.16911766086759228,
 0.22489952676800617,0.5149840791681205,0.7910587089666477,
 0.31819236140943863,0.8341694468996279,0.587653120148004,0.07458675232428769,
 0.03584747355659501,0.5079139256580786,0.07966558438318405,
 0.7948061619375746,0.5417140346936375,0.6534783008607911,0.8571111791215075,
 0.08946042704872381,0.06341966299710222,0.2953076762953585,
 0.04829184882572013,0.4408276733469255,0.05549223138576531,
 0.9932241154308069,0.07327390741452455,0.5437811565286924,
 0.09376979757354842,0.06599571254092884,0.9143157300255839,
 0.5285002174733029,0.4436949868798662,0.8869494867488953,0.6245097871432419,
 0.5136541767251305,0.5009761511813329,0.9879896008490864,0.21308895596648747,
 0.9909104568218146,0.8165358969408623,0.7179629498092145,0.6807600349090459,
 0.7299285078038608,0.6051752488559587,0.08728378178325524,0.4526759113544895,
 0.7798133354751435,0.14893723040990925,0.5971783539908689,0.1774815675828093,
 0.43183546988276744,0.7897912238667003,0.3743855435676793,
 0.10328739897258377,0.07117845688008362,0.6214188360024262,
 0.6268068339851621,0.2800080455407903,0.7498356675737431,0.14291911821984815,
 0.8116141887292062,0.5520820291174485,0.7407809438846183,0.43740141368618657,
 0.7228602423762174,0.5063620343287237,0.9572924463892978,0.9555391552524949,
 0.9465386678623382,0.8063755791371233,0.690812372976295,0.8410051334411515,
 0.7597996630826869,0.14982164851365842,0.30330912188704273,
 0.07404727443061898,0.10653864878072694,0.07924782803997887,
 0.5317824705646724,0.004073718587109365,0.01476319080832289,
 0.9415472035240547,0.3114022607037713,0.7706761994683392,0.15718549135425908,
 0.6794383162667694,0.4520111258394708,0.8095231927215178,0.8453450378027119,
 0.11104277118093331,0.9483519110350949,0.20892946487978803,
 0.9025158345298603,0.30096683841606975,0.49008000354778647,
 0.6905061146507966,0.9035383848746059,0.961181487356036,0.7503826005487455,
 0.3445422943180104,0.3384106143235309,0.7011014037832284,0.8835513578513005,
 0.3194136964070946,0.6460853600641897,0.24049447229308962,
 0.17107157403766127,0.4373091142585589,0.23126659000475258,
 0.6194588916130863,0.541360077081925,0.9431163443828685,0.9962286780105434,
 0.7498162189301206,0.8499798082563814,0.12989270192560265,0.546841203902181,
 0.9495446066547819,0.10004511145569872,0.8253268167559893,0.8105056226269209,
 0.7598260879610855,0.7675245477530959,0.8956060621290047,0.7780654689326764,
 0.21709339816614803,0.7256860607580476,0.07203026100915588,
 0.7638781165830429,0.6108277556195583,0.415907101610112,0.46658948763464747,
 0.42358530907997616,0.36319087805925065,0.8935177750655355,
 0.20675714557090685,0.8343160061384782,0.10004826947173973,
 0.6671258538736072,0.23165363600182953,0.6331123477500641,0.4912335086754417,
 0.4248288154251385,0.9240009440004695,0.9946241531853267,0.901420316263295,
 0.384997313326754,0.07146726862689534,0.3683281693595105,0.5982306488766261,
 0.08354082555271847,0.14749425289678408,0.1918217804919975,
 0.5924631448461583,0.7499699648542121,0.05856352803991827,
 0.21996920093735262,0.677921722793071,0.5685464464120447,0.6443276230778514,
 0.5792978598817676,0.10447206039945867,0.5626431188275542,0.448451807940467,
 0.6300644465785077,0.8348709219770434,0.26581443102366953,
 0.30613358452206263,0.8455871650841542,0.5195007558555371,
 0.029355428150031115,0.008860129702431552,0.5004701414065554,
 0.9936118615874989,0.3252210040449788,0.9321338829487147,0.16258603353306267,
 0.4381307535897222,0.6611686649154034,0.2709892915847144,
 9.372860619292211e-4,0.36150236715945305,0.3892837810140858,
 0.4111197689626688,0.9286943209527826,0.23069085294317881,0.9357772714965887,
 0.04381382566477843,0.10462912666389612,0.40913551430278483,
 0.9295288188953341,0.24252161454301824,0.3248746017831219,0.5639623994206886,
 0.6423705557243077,0.3651721052988963,0.3420265227560446,0.12528992743163758,
 0.9347266162485467,0.9515945510988579,0.9748371533258675,
 0.016039340730593654,0.39777926633946215,0.3811325931305636,
 0.8805153939534456,0.9554177805269541,0.26596754472079176,
 0.49687079980402116,0.4643580364839722,0.11139043283755379,
 0.5794222526494135,0.6424637072255759,0.6564526270295523,0.830786055906956,
 0.8078157543242681,0.5441581733219513,0.5258840149421906,0.42103342403165245,
 0.25745661462640945,0.9355994543693904,0.7771360320373288,0.7674114549984732,
 0.35955705106019453,0.5185088191543735,0.5176142573792626,
 0.49827211374377256,0.663169897392691,0.8580828232717427,0.7163782950980675,
 0.10202676853654091,0.6629729986427135,0.20661729036699894,
 0.8413857183522173,0.3629443387450325,0.04301005930492441,
 0.37981740613996473,0.15259857581462855,0.6690386174310475,
 0.9695976703641468,0.3884191573243462,0.14481202596038356,
 0.35993200493448274,0.36377088497843835,0.3529954203350981,
 0.4856582832637919,0.9761911711859091,0.350085709452547,0.02611724072930144,
 0.2451373035642591,0.13534450650280472,0.3807024844420678,0.8996310804047523,
 0.9409934629170578,0.2781999051808606,0.5133109470052238,0.5526187401630023,
 0.9738100888757798,0.39652429986971505,0.1967265942448404,0.9837034086864579,
 0.2751277185906341,0.14348006224151533,0.2514113176238775,0.3500402947312591,
 0.7140370024973897,0.36113135003203256,0.8394558061749816,0.6177115981628661,
 0.12034377456117451,0.9505786830962616,0.3214314477725563,0.9680433543620421,
 0.027345670662789145,0.8127173994678041,0.6692436037653978,
 0.7216873242055284,0.3820177110716195,0.5415788686908011,0.4561200330354126,
 0.36505873311695125,0.9678244652625194,0.2375169326597133,
 0.37567409943589813,0.4598176630892147,0.3751894432887469,0.3260692420776735,
 0.7647193169567943,0.6098941304655431,0.2520827458819598,0.20546140639357058,
 0.29895720360180733,0.6891308993937987,0.9719038064316121,0.4875057031119797,
 0.7498851551865723,0.3881719535790493,0.3758367199634033,0.41347488931996645,
 0.7091524706286705,0.284652816586636,0.792324310134651,0.6450810485457793,
 0.630903931133474,0.7716236074103335,0.5811174142405465,0.1840494550869869,
 0.6828180954788744,0.9132049916889735,0.2611986376565809,0.6131617756433382,
 0.5529731026633552,0.2628221057088338,0.7349781580566589,
 0.0019740413704123494,0.501196175927825,0.6371553598929847,
 0.7896314366645028,0.9890577533790258,0.3734265654847877,0.8294820211451203,
 0.839005441459146,0.23648865181737821,0.6865970659491025,0.7775352513831627,
 0.148352467956542,0.3150710896206971,0.38747153582239147,0.5258104246459093,
 0.8495089820241519,0.7878122175533288,0.05207361497044638,0.6529040120079002,
 0.21427655692588188,0.7211487888272752,0.11414588980641782,
 0.6954710702116228,0.18260735700574093,0.5450902737882355,0.6675151752427385,
 0.24965101257059508,0.6073393623343046,0.8345398982682676,0.4540891385251,
 0.8194530274907487,0.3281548990392267,0.9813946031593137,0.0969336845579295,
 0.02394584454821136,0.4078084697679877,0.8566832987160931,0.9671582254354879,
 0.6422145835290619,0.059413565162036996,0.6167758391167609,
 0.16274039129994722,0.9565499613892307,0.06591680161202973,
 0.8402613223541463,0.33477672127193725,0.4739824470278734,
 0.09211491101138536,0.6986787539326113,0.01102032275900755,
 0.9024219039898883,0.1699750548713188,0.4385153707655368,0.3402590191351961,
 0.8163062377212759,0.9217417314100553,0.9559317276353978,0.2604988151127142,
 0.3538990027830109,0.015048703948060349,0.04529526202939271,
 0.6187258164075284,0.7255366055894436,0.10838357972130974,0.6811417443015451,
 0.529753630124709,0.44677771310407743,0.9076152620761169,0.6933168279497274,
 0.5030171812559968,0.9727293573074223,0.10834494667868699,0.7503644922856886,
 0.9025936979831302,0.1806249442150063,0.21842814725527582,
 0.06576881736594209,0.5333201905463248,0.9620295491196973,0.6747129612640258,
 0.44579294892583254,0.9956387019472978,0.6033859722753825,
 0.46602405299662086,0.6537024002497228,0.8107738044508229,0.2662404754151626,
 0.8252579684903861,0.5764958297646511,0.8935012726791017,0.3253103291253343,
 0.524227931561304,0.4495356198747955,0.2762761346248622,0.6361486238433762,
 0.697919170789167,0.7126856733382391,0.3357773120698423,0.9873740553637091,
 0.5276092039770108,0.11942340695031206,0.6738327949782539,0.2598162398461258,
 0.15088955915298397,0.8927761954988388,0.46286297895154993,
 0.34443733907071494,0.8321797077331049,0.07822027625165884,
 0.21366984122889543,0.2997104184571413,0.35123064238649615,
 0.02648602348178264,0.5799287114911995,0.37738034750426097,
 0.1773700773252863,0.4642785297988875,0.23611002810758497,0.7361540946798578,
 0.5361737609369853,0.4451187130111498,0.8445792275662226,0.819836200878208,
 0.9736338254558133,0.059155039271942544,0.07339370795966949,
 0.7519251028676711,0.3254969186711494,0.9247106948621759,0.8501638631321685,
 0.36113154902094935,0.31034279057321357,0.21083208778514784,
 0.786474904654918,0.07166792688095258,0.37081186710279546,0.1972137987857645,
 0.9820087838981494,0.6305093492680083,0.5856034192552374,0.49636891116396775,
 0.892179276051048,0.585127524034002,0.30856620147378355,0.6333401675919341,
 0.8669940629533252,0.8407321480501636,0.6152949657596006,0.707654346661796,
 0.9364362796552204,0.39045029480933036,0.4016120790981781,
 0.07293113398696294,0.20443487834378682,0.7587741828599983,
 0.09703866274687345,0.42731774948928836,0.8619880537998937,
 0.25484757567501803,0.6209378795838095,0.5321268052697471,0.7209490878380692,
 0.019574481305413616,0.9079525626906495,0.36678717516616244,
 0.7069642745119848,0.11272009816360717,0.40803384463655135,
 0.3130228989781668,0.6997903563555181,0.8248667030260212,0.1478306986854021,
 0.7878493296250911,0.5159642680555334,0.6581577903761916,0.8989441709824811,
 0.3055150483009681,0.6947428430935179,0.3555318245912362,
 0.025161940941219152,0.9297277409178091,0.9121172165469221,
 0.23208914264421454,0.517127411613257,0.810176314983979,0.27150953919540755,
 0.7506748486049862,0.804108733069,0.8240472193027994,0.10312427299841131,
 0.9276707855759517,0.027304338441971887,0.5957689673025935,
 0.32467236213198136,0.8163521179017528,0.31968113677653853,
 0.7762591187936678,0.020699501020169042,0.3432577003694951,
 0.6842179655475146,0.7990329134145291,0.22274816448976786,0.4882959696943885,
 0.4439190940890383,0.12782574139152747,0.9463988099443181,
 0.24591098480972595,0.15844501790102083,0.16875640915798207,
 0.6995260229292007,0.3811659712256218,0.18069654221422038,
 0.27640444358319216,0.8740642271759316,0.0666879402986933,0.6657397030462767,
 0.8788956012333976,0.17463114796333878,0.9430032737862283,0.9305730987853024,
 0.9154423467081095,0.7170047945110366,0.24358191211825309,0.3468076688224634,
 0.0977948640900036,0.6268378448922104,0.605051848192448,0.4021409739516689,
 0.36322161302596445,0.6418148049052996,0.5378611735831192,
 0.13669908256170893,0.77519152676498,0.9448514639866685,0.35132581141750663,
 0.62863726612125,0.6202756169126178,0.32925029988919075,0.008548260449754874,
 0.8699051521257257,0.6455066828218565,0.31375175541855194,
 0.04930484419128556,0.13428397878078668,0.6106196947051911,
 0.02631606540486464,0.2938004208580608,0.7868821214705431,0.5079793335133315,
 0.0621915812397873,0.5738144610068054,0.44334812325059325,
 0.22517094072616717,0.155229737238476,0.15966737120867902,0.1804667197106613,
 0.5366027381498225,0.9298875368339298,0.306228206669645,0.3516527341577883,
 0.8965535825816138,0.08570132186383861,0.21803431923341687,
 0.9196596405999165,0.7553565086828429,0.33772983380346455,0.7596536901181694,
 0.3276959351882044,0.4823017105189482,0.4092968223674217,0.11086319395528443,
 0.2838027893101226,0.02760388278612047,0.026708593076419707,
 0.4299826933258779,0.6758894821219125,0.8456665351824336,0.5985223999481868,
 0.10433164428459019,0.7636008800872656,0.17233663963270351,
 0.9730726784308422,0.5578490905780915,0.889362095220934,0.4142393471864927,
 0.0035547299791559572,0.7348847179953149,0.9903207126947101,
 0.4198160877673789,0.6205119864773129,0.40273429193655663,
 0.47227493809249765,0.4914382737356324,0.20516890087876272,
 0.05485299574871105,0.5001187838348978,0.5868175865999091,
 0.07715932321907293,0.07648245066152115,0.9036046942593111,
 0.9241899256191691,0.10510321828727931,0.5584461180632819,
 0.04958401881662455,0.3468382399022468,0.44811820690545867,
 0.5229641342155833,0.7791095384828685,0.5930017183637222,0.1748249732892292,
 0.9281724807840454,0.5575740171604033,0.9400096908824894,0.28869045414685957,
 0.881757389477274,0.6789808809720719,0.1892591713867099,0.5878882983987309,
 0.9338037802564373,0.6657439739840296,0.5308203186971243,0.5953710242798226,
 0.22944825015786852,0.9679698062477159,0.2930557847720139,
 0.32983622218594677,0.9552569083445949,0.7899021997569847,0.8863789459167227,
 0.6250216873919698,0.6620813040751912,0.6976224557504684,0.4372667622158918,
 0.05191435143190204,0.5767244758993293,0.58205420968477,0.06250747983383176,
 0.6922129748189757,0.8552007574492361,0.2527217466350008,0.8766390262868893,
 0.8025908877258056,0.9350695885511533,0.705544289499561,0.004913933060968989,
 0.28784763334870644,0.15807907627818774,0.4104373462069185,
 0.09036430003363116,0.28495192526303526,0.15197679123308172,
 0.8980016721867375,0.8664495409680325,0.2781198715295278,0.9483254002509884,
 0.6697362360198577,0.2986967204967743,0.5257807097364732,0.7353248334977392,
 0.998417983373542,0.21740382555553284,0.5615000389321608,0.5170865368038225,
 0.31640757381476536,0.5005031253302614,0.6471727603217774,
 0.49762186552404364,0.20588207724147134,0.649542194209494,0.8201553803733969,
 0.8863957042628913,0.5094615572248319,0.37008966240071617,0.7955209069462564,
 0.18761903271700164,0.9927198877409357,0.11907853868177831,
 0.9089067799571282,0.6992308698937771,0.12729328053263012,0.6449161011518321,
 0.8943929765640263,0.8994115621056844,0.36247872614793875,
 0.14437627807471742,0.30447208691073846,0.6869585667603232,
 0.8179506226265382,0.07794844395063616,0.029778285679525895,
 0.7289752167523397,0.4723298818382171,0.6773063165847968,0.03347615340623977,
 0.43065425483358755,0.1225404629072856,0.1213232021706594,0.0314523066269492,
 0.659843107763383,0.9951129324339768,0.5083869325248301,0.667378200259116,
 0.28222252061115416,0.13793238975736566,0.044573483070130315,
 0.8933562784723474,0.7752824252322753,0.4512417693646884,0.12948101935252088,
 0.799110069411944,0.5004217701667258,0.6999752029061019,0.8240982294018578,
 0.6493556662772885,0.17666011669596704,0.30913713811676913,0.768765441106135,
 0.9850454512581382,0.9619948984247461,0.9587720627178874,0.7410846572204526,
 0.29078029566071506,0.31988349526710724,0.20167913092534084,
 0.42471337201383663,0.04325034840505071,0.13239428916574936,
 0.2931094326933569,0.21079937886884892,0.5548671326869263,0.498689722688366,
 0.3558681853591449,0.7050161362845097,0.6209687271580696,0.7289241566914969,
 0.8041389929728207,0.8463634339765231,0.6356526238764075,0.9396481169610059,
 0.5091050991312444,0.09812286350874266,0.6586710940579437,
 0.09757203577570883,0.6915185109301398,0.17471226139085982,
 0.5699515372948125,0.4746218159320238,0.654785860768258,0.6590170341448832,
 0.7923983192568234,0.16095652585130327,0.7302975460443508,
 0.038966762284071144,0.19767777936385733,0.6973164829191575,
 0.1738414997013069,0.6013763159741541,0.6129149506007997,0.5766440702577873,
 0.4471849753624315,0.1939005091409849,0.6299908707346902,0.14861792962676024,
 0.7963360330367606,0.8213708127427468,0.37854421053909104,
 0.11351597504717414,0.954984904372928,0.07450704086596205,0.5932358529366627,
 0.889644495616124,0.13010571788983982,0.17866405055408063,0.335242181166435,
 0.05203406727762028,0.5400735303650963,0.9407049335108716,
 0.29003713496731287,0.8675488200762849,0.860302179313969,0.9685511045122542,
 0.29072676824676347,0.48283139121090823,0.7344530743205435,
 0.8293021930076783,0.9224430118154682,0.9530571675493635,0.9129747020197932,
 0.26211287152862617,0.2392176608020088,0.04770751235380455,
 0.9478711375783464,0.7836167192156052,0.943681658749693,0.022335114257764688,
 0.1538009945729919,0.6617628974709013,0.47839139101104333,
 0.27987512926262437,0.8529386721261788,0.3983438389258668,
 0.42355112874339573,0.012805041728743971,0.002305079905369767,
 0.7356229022203297,0.7169463385558488,0.6865001206785021,0.8769400368960653,
 0.7781562816579056,0.03844635029215837,0.6552082058473766,
 0.13035153972119318,0.45275302071476087,0.6798187074257641,
 0.7405078223974744,0.5306848476952158,0.17143285647500428,
 0.11676359640669831,0.9562054735339167,0.9631866982390087,0.9911061584584226,
 0.3462849889846702,0.6768626796377423,0.14497432916236586,0.5267607742852431,
 0.18140411540517198,0.3165730445947018,0.06668659446820635,
 0.9581710582411591,0.978111500222995,0.9175021521337086,0.6995978601063677,
 0.11539520688599769,0.2253616850355954,0.8750907570634505,0.527987414871186,
 0.9392173936564634,0.5025358691536431,0.9974418142357285,0.9770344539599589,
 0.5140177700140417,0.009510632263888263,0.030431213653393296,
 0.11383449643455734,0.2820327295384002,0.23352989088439724,
 0.6961108747431444,0.4837536821527607,0.849974084517795,0.7926577788512867,
 0.39320971847785036,0.18031301192583749,0.0894594083722029,
 0.1556939282773977,0.8662834714029624,0.05846559385138339,0.5532828428393376,
 0.28405366021032585,0.4773559587205728,0.9482318808580206,0.6083090908009429,
 0.14342183337881498,0.5227821383980752,0.30119941038676046,
 0.7392702568088745,0.7719013333772582,0.5376509273771108,0.9056161471649044,
 0.4639783883817502,0.4546677346263098,0.9358886386288725,0.784539507467046,
 0.32387443997074883,0.5573540161757989,0.20676414734212978,
 0.3938003651849643,0.16590541291042593,0.06172283389993827,0.373727720304163,
 0.2187709418355579,0.26816151853588466,0.32746869239683873,
 0.9352219734054805,0.4963863409310564,0.7515331462169188,0.5338432094450062,
 0.9055477098546865,0.09682855097371168,0.4486579197415237,0.5399931482474131,
 0.3082049549942383,0.6864015012472842,0.2553446271948172,0.7128260548275949,
 0.02670424225116297,0.2672140626244638,0.8434340163085895,0.7332972085858239,
 0.5831073884303231,0.8200962296289315,0.5217100958241339,0.5249487370637747,
 0.3523751645291784,0.9983309000589459,0.9236178440437202,0.8866798693252034,
 0.03506140830043203,0.48639571577057983,0.31968309839817743,
 0.30593164342176493,0.9028987945635835,0.2651796919780314,0.6412788237296583,
 0.34350062457258934,0.1893738164474702,0.9615130721844664,0.3246425136594129,
 0.9442375192082171,0.15878897613211218,0.8193717328707881,0.9605201667512582,
 0.8527516604328222,0.6151837778244463,0.6312951666515397,0.26346265478015074,
 0.916672023101167,0.47604166877750753,0.022152004787404644,
 0.47331938147986263,0.15217267862759531,0.29698102396336257,
 0.9577338300301039,0.9024339048853331,0.7515901355532708,0.5908106415800709,
 0.31737799570306335,0.2357387427437555,0.5666601907924557,0.6738756215041677,
 0.8784742619937681,0.6249489548727141,0.7021076153944407,0.5185160857650415,
 0.268244706167464,0.8775843218479267,0.0018673607393852043,
 0.9691081508567071,0.2580336013306824,0.989544605541324,0.9811602852417585,
 0.641099113468141,0.5185238252222395,0.41645383703036154,0.9467072100586589,
 0.19280953919751898,0.14922965195675508,0.39826713912583755,
 0.6232193670411461,0.007562500747723022,0.8102270181843787,
 0.2646350488264426,0.9759865705656358,0.5051534752358587,0.32708311422630254,
 0.32738394274449545,0.2940627904741603,0.5141929495414722,
 0.19640436440726539,0.7182678476013353,0.45003736236619285,
 0.34841268746975407,0.29911489066522134,0.9968306460904297,
 0.23015390170254502,0.2256778370358964,0.2801975501823373,
 0.016368178983433457,0.9214348508674873,0.11293798816744516,
 0.6100137501406967,0.00841227847479531,0.1459309720797286,0.0524528856833999,
 0.9145092812929301,0.6389310746321211,0.85896247287903,0.7115585210920253,
 0.15545655313815843,0.9052802155723516,0.09582571714898236,
 0.32609767498922204,0.5464044523991433,0.5573644369251733,
 0.43072874067006317,0.6869774893982062,0.5912298325243442,0.2536366583195334,
 0.2035188044703976,0.6220388632213192,0.3830193847738367,0.14108214409418485,
 0.24006593681560673,0.6991332717838538,0.03406684969548912,
 0.1684396340208849,0.4648293738374467,0.6015630080217627,0.9902234854439398,
 0.3312200470735215,0.7615659165383865,0.004717600606653161,
 0.3119109393940446,0.5656525754481591,0.5329606672623147,0.4929427379671937,
 0.9745689499269219,0.4634166414195393,0.4315384953378659,0.5057736837661341,
 0.44078628421946286,0.25807124260337955,0.8136372328041542,
 0.6908927338188567,0.7676360369218017,0.3493327873370762,0.6549747985190589,
 0.5388148542029024,0.24185252159396842,0.2778643663573289,
 0.30206819212945113,0.1236777002251973,0.2695580972417094,
 0.0028863343474853087,0.05565148304398082,0.8304196298408407,
 0.7703189602396836,0.6860207736519648,0.29104913077331773,0.2851758707560341,
 0.9656042074322997,0.7835973062912479,2.259504640663712e-4,
 0.8072849534453401,0.8080137830759324,0.44177618054260315,0.6919849664042619,
 0.11829114463112256,0.5712237165080707,0.9228899631566373,
 0.18254534253610455,0.32887818218362364,0.11347633617817021,
 0.8355586564300893,0.8055116564894239,0.09288211490720522,
 0.02685712363028292,0.5423207797271439,0.9194385666944372,
 0.28984255238528134,0.6062688410966137,0.008947958270804302,
 0.028806207781877813,0.7147622715188457,0.4873484313667059,
 0.5047943100069432,0.5198822678204904,0.1095853230265954,0.5526840830089006,
 0.01519715907359398,0.9946532219388283,0.3413011935220127,0.1696004728702113,
 0.5462953640862162,0.4163569937849424,0.49365990244913416,0.7514348329020448,
 0.2159177923506309,0.23065573613024803,0.37658266278580543,
 0.5843030954951125,0.7758149112218635,9.714857030841406e-4,0.94067179913058,
 0.5957554009423203,0.7681714453082089,0.8885268504259538,0.5402595899974347,
 0.7119250097031696,0.6917486891898892,0.6140871958515477,0.6459720493823022,
 0.5639559861718764,0.6752466667612166,0.7380828255716246,0.46146367908289176,
 0.11043385446742582,0.7473694352852835,0.15958577302353305,
 0.4864841209109668,0.4625365185766258,0.8075917120371037,0.26488427673480264,
 0.23498540845302596,0.45339782849481625,0.2990520330729718,
 0.9571008009462867,0.5784370410746532,0.37549422054513526,0.4990129759372426,
 0.6799798662581418,0.7554873650213764,0.5214472308307045,0.12379204175870673,
 0.7053818756135423,0.1955188969352457,0.2364124126475371,0.9032878476389811,
 0.651766226329094,0.1315825109759181,0.7325165896453446,0.23745554033762772,
 0.7973924813447318,0.1890598000540311,0.07535163837784042,
 0.05021245897397397,0.8854519178373054,0.262031116520799,0.7515943323637063,
 0.2404910052532253,0.36825718976746724,0.21404910757096163,
 0.5557566631734003,0.10357658383230772,0.13699407452996037,
 0.10740991269379452,0.15901667946541398,0.5408397735055114,
 0.47382213863137623,0.7708568945516256,0.12066631563992336,
 0.2863103230759111,0.01910356104940525,0.2511345253651496,0.5661692109717902,
 0.8729576302576342,0.23771072475117705,0.6952132841412137,
 0.40560131010385747,0.12868594318824622,0.22060867458696964,
 0.24586186269156807,0.1421864597422544,0.9725025231714974,0.4785375367000594,
 0.665078489173825,0.2939506526053759,0.2228890367134111,0.11601551761414952,
 0.5413065353842856,0.303588849282169,0.8848371847630018,0.2553034132249491,
 0.0512135065727386,0.4480640753173992,0.6587253264384512,0.4857387353208472,
 0.7366485620585775,0.2707502018450043,0.4272534738217707,0.43285816775547603,
 0.21609517838923076,0.3628899806637874,0.2731468471249523,0.938048600045546,
 0.8055371369862845,0.9944283978552708,0.43403084402763037,
 0.025175143249976717,0.8645646255941519,0.6018785641772889,
 0.4064111703556397,0.5512510692699835,0.11652494426595994,0.538861618950236,
 0.5270158997255814,0.5225906914221635,0.05987431843149549,0.9525904250590684,
 0.4415689078037348,0.4709754698458073,0.1844407050230038,0.23964147386485057,
 0.8892562268792711,0.03637255089970104,0.5009575275634982,
 0.30653586274674705,0.9360551205136672,0.6804904679806941,0.7809334721746785,
 0.5845964049237476,0.12046955248615998,0.5721820346696231,
 0.46536244781999114,0.038538408912304756,0.42940405249686997,
 0.6294161447146057,0.538330974504803,0.31613045655193894,0.6768150843955352,
 0.968598714614104,0.04593679754068836,0.06756510939902316,
 0.05568428597982544,0.8545770229729743,0.41761490645184285,
 0.6614439798962686,0.4547713959460419,0.5614332241265135,0.20700333523077274,
 0.6549650490547322,0.8534901319871342,0.5629183643621797,0.20408073497146317,
 0.3805342386201863,0.6159802482783342,0.03540338128189724,0.3113856117729833,
 0.7538329138580593,0.05581610280578919,0.9951903093208652,
 0.10394212683608073,0.0010822895974029745,0.21651535367841612,
 0.29972050482855994,0.6527054625755704,0.508326246231538,0.5246926571891866,
 0.35291117857484267,0.7905221123180097,0.5061304543214173,0.9665757572443833,
 0.37925524355764006,0.7268007957580227,0.3561413267086746,0.7736241274496269,
 0.2791456715793559,0.9926000600890401,0.043911098643018454,
 0.8888910387965866,0.9604195232267788,0.7995855369970103,0.27706559885274473,
 0.7955279642564952,0.8321276043134626,0.9550410816852412,0.44213578990706026,
 0.4118105998929973,0.4003230168328945,0.1768669407257949,0.6841625082381209,
 0.5646716239467773,0.7500197722929978,0.31978658448010033,0.3674491443235509,
 0.5606377902284982,0.19681646956367183,0.743827890159266,0.7979451894057241,
 0.17574261758772858,0.5134134322566782,0.6810579827073718,0.6589477741123115,
 0.4888610012620075,0.985925714452438,0.19274303188571207,0.9362008305796964,
 0.31540328954223096,0.7065067717796586,0.14030189775013446,
 0.9132546302742988,0.2715527598145733,0.7333464911449173,0.8860568293703196,
 0.19886051286290363,0.2546541246200571,0.3335353293136938,0.503392552111308,
 0.46914735336400726,0.5262984954539311,0.29567560196849574,
 0.6632140668502332,0.21594999935589387,0.6688309922882755,0.5086367682398792,
 0.5268163404886803,0.6080132047803044,0.3535793112745371,
 0.054869523069067117,0.9713831983620396,0.5130613365171339,
 0.02319492366801701,0.03487905194761698,0.2813231971343686,0.574311218335377,
 0.18017129311317392,0.002599613772935938,0.8682829132650469,
 0.33242979236025993,0.8152207794539611,0.2704118402985337,0.5579464424800875,
 0.9078833322743378,0.920744398554081,0.7285663086504235,0.9840278962221887,
 0.2839534232782792,0.5681573985621639,0.16914384425945161,0.9806879484307764,
 0.9658392838545843,0.4455492996985124,0.6901516751113658,0.4164935396708588,
 0.3363785681809275,0.42022064820035643,0.46537137602185086,
 0.9173204799117717,0.5959766751716737,0.5007679562044886,0.4535708947819751,
 0.5721063285142621,0.14466459561725697,0.048431262626621785,
 0.6679876750860889,0.0696686416472303,0.0519372067307986,0.4257734905111388,
 0.6713275352324489,0.22914658311959613,0.22190830819938312,
 0.7374081830126017,0.44230143112507436,0.8758601235405938,0.6589785673223567,
 0.33952856256364716,0.095051777183215,0.9371083932683735,0.7396216622895451,
 0.09945994718140805,0.9819852203440135,0.9243331337640138,
 0.48410225468823254,0.010223893020255392,0.3096503588894919,
 0.9143599935218443,0.0714818371323197,0.1454630340484585,0.5544078302089568,
 0.907306327768823,0.30821712622530184,0.3306853825941255,0.11929590952264757,
 0.2186441234197436,0.3069345111108984,0.38463375809837097,0.9374167738995867,
 0.400026997580468,0.6827873930009833,0.7221242618198284,0.2264266789966105,
 0.8303345327546117,0.7184714309101007,0.6986146932072523,0.9439392062410908,
 0.302996564486366,0.790621164993445,0.13332682294142928,0.2483497211677238,
 0.06384515460878715,0.02524005036921373,0.20343356152799486,
 0.24783900514033663,0.20334308333545292,0.8834545423865899,
 0.9406308147594655,0.9342563489199471,0.7662193114677716,0.6859239681360132,
 0.8259730992586578,0.3164026887988446,0.633199314263509,0.5513838336599355,
 0.26470646336298254,0.5911963011321621,0.9791175593046381,
 0.45611596353396344,0.9639263930943932,0.8584074189637245,0.6022201807037243,
 0.9621589116558793,0.5997718701359079,0.8520337547761205,0.6504897906761249,
 0.04367073597759408,0.12764140206471453,0.6738927470013272,
 0.7262823505203023,0.8903109841960816,0.3026263468883893,0.8216171055678142,
 0.07021731580700552,0.8188167459611396,0.9732924837615761,
 0.11329631899170722,0.6449125109143605,0.9325899035305302,0.6824909110974622,
 0.48166950499896233,0.9382981576754617,0.5786571908017627,0.5314418214255857,
 0.5548475744775874,0.47443108914614873,0.39079413581334366,
 0.8022439306363125,0.15335884813131617,0.06514429658789544,
 0.47143559615346775,0.2866654564992993,0.8129150334244936,0.9755949627062063,
 0.24604881451809735,0.9604774401728395,0.1451223209780984,
 0.08186292495272207,0.2406575852013214,0.8087941300633221,0.8271412035362088,
 0.1668521496398806,0.7059767895542117,0.2606609170065375,0.9158428996684009,
 0.6888372386524703,0.8849791160940751,0.31221052496389634,0.8045810776956019,
 0.3752834907623278,0.5809832629414192,0.5147263446069628,0.7773488918861807,
 0.4267693316944037,0.30530451438314543,0.5665207375956051,0.9037898813984372,
 0.4936029772915401,0.6978914217668584,0.3931550047267738,0.4240661618096696,
 0.4941173405906665,0.8843538070237809,0.7888724515401235,0.7033791923702353,
 0.8588433637519483,0.9503433821417294,0.21426178581185318,
 0.38690231679050247,0.9319352579707463,0.908850813252878,0.6732227932766605,
 0.48320506565933663,0.328701544408207,0.3242445303096053,0.6726702511209858,
 0.6070376851200218,0.6490297443836273,0.9098907725205521,
 0.011775984631256065,0.9821350092765829,0.3554537186765018,
 0.37084801881061624,0.16790814076477611,0.016779938190935817,
 0.5010842554168028,0.5851357534527686,0.28140068116312844,
 0.029889931730091135,0.3909934654711731,0.6291839919297707,
 0.15423160941202863,0.3909565656188103,0.8186982110423415,0.9752947011636771,
 0.13081470269414563,0.2863785412368126,0.17609296914635764,
 0.5894973138872062,0.8263984313586258,0.26509842457495336,
 0.30399452828834195,0.353253118404275,0.50297757311479,0.1252205098470176,
 0.11027686505495149,0.6052854083842997,0.8572701961783774,0.6975629819395035,
 0.6757289409110007,0.7689453751258194,0.16382676542083452,0.9650524432294538,
 0.5553830156524302,0.9646174944046289,0.9172012697515454,0.26338250089798265,
 0.1681425761754387,0.9912100485867723,0.8193621864664751,0.8484741432458213,
 0.38115747404739353,0.34088995272498135,0.38588084839314485,
 0.9636474710521643,0.030485271452635354,0.9645196196770729,
 0.3183199242157507,0.0814747116683141,0.25101426625961276,
 0.36662993185292847,0.9073971972511241,0.2279011016385497,0.1561375359556063,
 0.1333635852860916,0.9377393157489736,0.47603842488419246,
 0.06855360900371377,0.9903212980604944,0.8289260522084231,0.8880004800829037,
 0.7459864108688066,0.711813733233893,0.8436142996043595,0.6355309977846507,
 0.3348671172533936,0.2707101015328778,0.3856912472676247,
 0.008982407634837264,0.04010327190693208,0.2795179050146912,
 0.44435086666298906,0.7973757275428416,0.38853737276839906,
 0.6906642729826694,0.9947566304556812,0.5954937816134962,0.4899807195313295,
 0.8821607404255901,0.4271269897503056,0.8074063275147936,0.1996469042676684,
 0.29238649037912623,0.06613568902938227,0.7995177092019263,
 0.9474143293133611,0.38597613224593197,0.39857446054138945,
 0.5415752626907484,0.5656335448855215,0.7694807822889803,0.43167465355908974,
 0.07526593611537624,0.021277655596769862,0.44811272395391677,
 0.5667201983533984,0.995578707918644,0.2107725026704934,0.26728036863074733,
 0.3750859010095433,0.534126321735988,0.8855899744626332,0.9672275193252806,
 0.5695952376486746,0.6672993559554957,0.5948751879475677,0.04912978322766848,
 0.07824612208925907,0.28593498599753087,0.5587087399265328,
 0.3372466808372454,0.4134452298587875,0.9389998575573348,0.80321474760906,
 0.06743433290337975,0.7133446299437267,0.253313851924032,0.22210912837008912,
 0.5930582282169388,0.9191577621542596,0.5976551898493787,0.6784203654852617,
 0.8381080199212008,0.7341926489067729,0.21826090185620695,
 0.19550273838627608,0.999770501133654,0.28929585552872483,
 0.15732062935593372,0.6947800519811203,0.5825102916260108,0.3229114959236259,
 0.8978247316648387,0.9256446886320321,0.45943898321616816,0.4265349557401248,
 0.10894771834770345,0.062083461056277156,0.5842326964646517,
 0.34688595953617807,0.8562915477911018,0.5732824229820199,0.4535734315265556,
 0.3919703167753871,0.4990645668355105,0.6934776415111151,0.6563244916577243,
 0.36292333868812787,0.9828912611597892,0.6621513641549104,0.1957899961620534,
 0.26341303153735673,0.07087693503559955,0.4291213221755674,
 0.25731558289173495,0.6579850392295927,0.7005861990738275,0.6628606270031302,
 0.00878818127935388,0.9410569280897774,0.1550486866958115,0.4743383228936564,
 0.19618269457492732,0.9517697476979502,0.45137111492556725,
 0.9023976312696672,0.6350525940911909,0.5850435849073651,0.37753428192610305,
 0.05482439368826386,0.11989588283062047,0.2539864099402589,
 0.9636610430449105,0.03243516458850282,0.5525176024835058,
 0.39051623668669966,0.19097703439504365,0.048433683281056794,
 0.1252383935015381,0.4044616142432257,0.49604066527838586,0.65832549708157,
 0.641006702889636,0.785934944431735,0.06623096489935387,0.7453119231594971,
 0.9048035277878533,0.426481365680793,0.8884772897357164,0.28209005261886144,
 0.5458853800286309,0.48418703143918107,0.39811001180468075,
 0.6361479198447395,0.9794456687957325,0.602752315501736,0.8316506060442623,
 0.9209127088835154,0.09271817016364492,0.3593871318661773,
 0.07835032262726271,0.9234631372658776,0.9752457546004246,0.810668541460298,
 0.5069093899076865,0.17614227285307005,0.3113944597585795,
 0.26522355401358944,0.2826907117592935,0.6390893529360421,0.8213692645822986,
 0.5479892874170391,0.6223019612684737,0.7868213360873351,0.6370452177192131,
 0.4102666065692352,0.3980100358991483,0.9203388748013397,0.5919907987834365,
 0.39134160868281387,0.6472712762923837,0.8210559780829763,0.8029022600308607,
 0.026538570811942552,0.6642930827029409,0.03158445788362663,
 0.052407495316185604,0.26101025071048867,0.569777816935722,0.315793448166907,
 0.903164351038311,0.498841504588983,0.3467018357514692,0.31914903334092726,
 0.6714645663881404,0.5789163956609945,0.17756425198209203,0.3327016272060721,
 0.9510080306534843,0.6485539953753149,0.41276171955201113,0.829370640963522,
 0.6323153263515883,0.9971280421786999,0.37089559709167863,
 0.04324142814904053,0.09800486682967313,0.5519166789332273,
 0.14776590538967704,0.8006278187136878,0.03266470272317923,
 0.38128644219922236,0.4350823695934479,0.8148465697759786,0.5797670402550563,
 0.04774014901556112,0.19702615453714223,0.5293970057852381,
 0.8033542089985343,0.2430685313258777,0.5814551983424396,0.2036353077383688,
 0.8357569192498375,0.03184634858197788,0.19666326745275653,
 0.24694601609205113,0.2820485312026061,0.4851528713142822,
 0.35013779450100535,0.2267702098748693,0.3545328228079979,
 0.22635677232471751,0.4403284303943139,0.9178658808099769,
 0.08770917151679736,0.7546477184585867,0.13232952473199455,
 0.5838084366259327,0.77764677557731,0.5219167603605375,0.9807675231237725,
 0.9896601274167228,0.3225068990945691,0.23513411113395022,0.9832170761806954,
 0.7757946969768009,0.866502013012964,0.06424846119881944,0.3186277220198803,
 0.7713512347381989,0.8299350971079474,0.4173936800573732,0.687441121047028,
 0.9617729175396429,0.058972182805195006,0.33434852364728684,
 0.7160739291611598,0.13001713700244355,0.5785152581554941,0.837532012425124,
 0.741485565869496,0.358741072704842,0.9612785799988313,0.14555788616830923,
 0.4334134928326516,0.7595760733177198,0.933706890988477,0.4940203503103884,
 0.6580857156223868,0.5262855470356337,0.454035846960509,0.6483745600676357,
 0.8536660682125712,0.07270575066166352,0.5690165101179,0.48531217612224853,
 0.6054400173869691,0.9900456783206216,0.17040358666962652,0.4655475455311354,
 0.25839170271071643,0.3414121136897871,0.10968528902961805,
 0.9366577813684791,0.7199670969706666,0.40590903885331564,0.8793598241385677,
 0.02507685521861691,0.6147377679649423,0.07212697349858077,
 0.6657005593030259,0.9567175331651135,0.4665423361653649,0.5999444909418065,
 0.36156922088137944,0.3907418667138943,0.3216592341431106,0.5809460872331424,
 0.38007136240506256,0.7555117930559508,0.5721255641951631,
 0.09863569694651952,0.43569456933754136,0.4003287498109048,
 0.19540888524120814,0.8732779788355142,0.3950590128289371,0.8178958734810109,
 0.828307159236208,0.8222491499680219,0.0017320113922322244,
 0.12283228849790662,0.08600814520784938,0.5815424808551242,
 0.7310571850684215,0.650181384531426,0.27767141499907544,0.36866030786773063,
 0.6175033911290048,0.4063385232379705,0.8911879799975012,0.6209324794462249,
 0.18992669575605092,0.07146743739479788,0.1757271070816766,0.473969546403593,
 0.6973751851395473,0.2843286729647281,0.8654237450300905,0.5409350457199178,
 0.08595522481241413,0.11446420405044,0.08952330934321628,0.28215167306316824,
 0.11854595762318265,0.6530575497499388,0.20378632556535514,
 0.8040535573817477,0.28039295657474717,0.667709712451849,0.9640400805885613,
 0.41943707338764225,0.9976845992299055,0.8025224931581194,0.8606283683379722,
 0.1258354277893282,0.45421928881174223,0.13164186342645734,
 0.9434138751085631,0.04972807554086567,0.12862179246922367,
 0.3585950680051555,0.7972643071889807,0.5401386687838005,0.03963946044357902,
 0.9625722847543081,0.571474086075681,0.5819475019811126,0.4968660345311803,
 0.1848247020520023,0.3413983710233911,0.6757027014184129,0.7460821657676756,
 0.5779910952253511,0.048189053141622806,0.3307236219991154,
 0.22803321936786758,0.059403864659496874,0.575336653282778,
 0.16108247738273307,0.5667478342039916,0.16842582950418683,
 0.08505412012817559,0.03363339847549818,0.5421359347859147,0.766182476843976,
 0.17962816070109233,0.17294417297247489,0.25332030881582346,
 0.33320053897522217,0.693182438310373,1.6366689350444386e-4,
 0.9625445971511153,0.7071375711512784,0.3526969441531904,0.5048212841120518,
 0.46983297006910396,0.0071462898288801036,0.4744926417870077,
 0.3440298301600647,0.27042613338295607,0.8550617179158565,0.5144022620948584,
 0.5955084699018499,0.6972278802598063,0.8571357714842387,0.4052293482220918,
 0.3879094929632637,0.43778392731031746,0.5575207962096351,
 0.22286883480623998,0.22782301630804236,0.9573373427876556,0.905320232209318,
 0.5634611408860276,0.6649621715110519,0.18635255461714084,
 0.36051751594113257,0.7596313475005066,0.14744969805190722,
 0.04309508615095203,0.42386885079376657,0.6468471227365263,
 0.8414493281457979,0.2050730248762831,0.527854011359696,0.371106827739896,
 0.17885818522983654,0.22650406452500227,0.7048198709503151,
 0.6737065856788488,0.3963389197932915,0.05194624792412372,
 0.09530497044352804,0.7847948413233767,0.29108936552781417,
 0.4836228093597785,0.37351291742052106,0.22032825158094516,
 0.7450181839748069,0.8485218725853599,0.7677904050843534,0.37590915235176325,
 0.06368912964558571,0.9578742652205311,0.5345677359917248,
 0.0013164173194153417,0.046108146912824166,0.1949344232157051,
 0.08192808634174542,0.918291476082756,0.8886025538076203,0.7154598574767783,
 0.48108506845420873,0.13979543923107307,0.2322488334514754,
 0.9469739074507026,0.3316661512728849,0.6926670351099105,0.7351110902599665,
 0.8444853642299032,0.005022288961593624,0.20313795320503591,
 0.24249954933609807,0.37187452513449093,0.36436549018559583,
 0.03615695799811103,0.2571382790035286,0.48822193146709814,
 0.5950347840132731,0.7612973444737143,0.7615603092048243,0.841971603012297,
 0.25301879713641795,0.9910588805090533,0.8077018380854124,0.9141001913488898,
 0.018769272250837687,0.19765707685863276,0.16759248441182695,
 0.5960372976690137,0.12367023216352269,0.5132418016117221,
 0.38368776171950114,0.5535615489114314,0.19851644135995072,
 0.4027095160950819,0.5408415700424504,0.32519089581421357,0.9213023371024582,
 0.9715812960180262,0.6135267092176191,0.6731505900260127,0.5278202205269007,
 0.4519758002170088,0.5374715049230777,0.38565682583382677,0.9445470014139884,
 0.08674361198109914,0.41046036339544556,0.34723206585202027,
 0.4569061756003534,0.47931556163368316,0.7975670161858528,0.9279030576353692,
 0.9754215071464851,0.03542363153886763,0.3493213738022001,
 0.024522960042189457,0.15826897705277554,0.09294447430865282,
 0.9782439056801129,0.09006286663777552,0.10312985763182303,
 0.4873603408438233,0.24129697360347213,0.9738172619912192,
 0.06146159953046426,0.18945489964687368,0.7024358021180548,
 0.06166665846874453,0.6220948668883458,0.7148406004504906,0.8124923497910934,
 0.8023731869346952,0.4263230235443528,0.8381617755972326,0.4622597773508589,
 0.24632713518372507,0.45474200674048126,0.49625064154179177,
 0.940291561742711,0.0834133812985759,0.4925552342861559,0.9564543862272702,
 0.20351609396573,0.9591651273067595,0.9439586450182501,0.9236363035078426,
 0.32767291944766974,0.276556547971194,0.9021975003742134,0.01411449569024803,
 0.17242533097253643,0.07025114605821325,0.705030380313665,
 0.48434937960334834,0.1780809002944299,0.4178940022290861,0.6289472231413704,
 0.8492970937519091,0.1602519611093146,0.03788913272177963,0.7341181886717753,
 0.9173046049096298,0.6243815612042658,0.06216282969439435,
 0.03957182484804278,0.22254767723263413,0.6494130493898292,
 0.3092180849977445,0.023027711647849447,0.11542408065372411,
 0.9459063334684952,0.07194802415327006,0.5214147809943973,0.985120241894606,
 0.665258895901663,0.23724196749439164,0.46656038000163724,
 0.24502067070086997,0.34323879587234785,0.9483694554072353,
 0.08587739374991765,0.023259173320903437,0.12537497616607807,
 0.21198152705704543,0.13162988500545258,0.2335016837219437,
 0.5432810188928975,0.10675415574882952,0.9646584062047664,
 0.09144372943858292,0.1295258377728007,0.27326458500219986,
 0.7183013612222056,0.5545031056407086,0.11842177736301396,0.1320166137870198,
 0.1664487668807364,0.2296037092736425,0.27596257403700064,0.5347630233807017,
 0.8912219685777113,0.7685571950109766,0.7696036148040585,0.26801584677469525,
 0.21073048841867004,0.7449506699038271,0.7025100760056038,
 0.10815204660331235,0.6078071799674989,0.31094036464308505,
 0.020188224010040057,0.9120437336324535,0.9446453948929299,
 0.7076657506630825,0.5866079629418411,0.2836732145520746,0.5599309500729929,
 0.10859768488558763,0.4327369717280801,0.6060701419745382,0.7429110094806147,
 0.685616137189474,0.6384587216122463,0.22792703666250702,0.49353184073219303,
 0.0939420216804332,0.18301902396586223,0.14708142100131183,
 0.9128088620394257,0.7063666884417471,0.4619167628391231,0.8613521161297528,
 0.7521977702118128,0.885256240149141,0.45666933402080767,0.23709001286581677,
 0.10071162625639418,0.03869940903353508,0.12854427847440408,
 0.13608664747189436,0.8907214140636118,0.24936433522172363,0.614513559352651,
 0.738711811041787,0.34808611620794017,0.6042170349023437,0.7909254427403588,
 0.22065387588303897,0.3503270842687398,0.04572355770184566,
 0.1686163591552443,0.5176430439935299,0.984763198330237,0.22386999896694992,
 0.3761644572442423,0.47345130276119785,0.2130111924490269,0.2588079795249829,
 0.8178831004995113,0.07229494637701261,0.21603449671904473,
 0.48535017188208807,0.2812105012868704,0.7921931818220876,
 0.13171657792792602,0.41806317087070244,0.07969217787742744,
 0.8440793470567334,0.8104862224068502,0.0924053479255873,0.8090498223130194,
 0.596339563264252,0.29413376010150327,0.7634884376491171,0.6618518809561872,
 0.15662831437785263,0.01027685953593771,0.8852228496987996,
 0.8860927986429473,0.8165217820219227,0.4639113878958133,0.8616981075083333,
 0.2627344862599851,0.8345322838480267,0.20596411721519037,0.8241004005552446,
 0.5074592285504568,0.13380156686437983,0.9766647334132446,0.5322196172984741,
 0.8523060172758844,0.5994579521480942,0.5147977895654898,
 0.006028996373940276,0.08404099549022204,0.8510808196611066,
 0.4041680046847075,0.8814019480748433,0.9507603251019584,0.6721445696079575,
 0.8449007439270877,0.4174422155351971,0.5823537854249722,0.13938359389369215,
 0.344598257555891,0.7343221914438878,0.11610720676780328,0.4865841194373871,
 0.5645512662573623,0.8518594675870338,0.6218966462599107,0.18513446946053946,
 0.4574938891248269,0.2166225992577977,0.5188938586306766,0.1985215449599692,
 0.6454290380366463,0.09816912034676073,0.20777386738338788,0.329684256173403,
 0.9999726732299027,0.7756323931208903,0.2860824583578544,0.6477302443798475,
 0.30362737434644504,0.6125187616992488,0.07446896219470434,
 0.8463107930032593,0.3020978224612578,0.18343258931266193,0.4599139463543409,
 0.5674155416111503,0.6408550022793122,0.863968502820468,0.1747810639830477,
 0.87689179069413,0.9458824090142671,0.5573419655764732,0.5249735833197834,
 0.7808925464933583,0.07542501431839721,0.5936875351076119,0.3301535651515817,
 0.7778393964038147,0.5811598377492914,0.7333182500177536,0.9594490120271499,
 0.28657548105916764,0.2554278317219354,0.24030141124913462,
 0.33581911714189905,0.3344604980937047,0.3567943050972704,0.6775643100290266,
 0.5440803285448813,0.8270460047712718,0.15363632331620414,0.6077620668739292,
 0.6868976223219068,0.537149669355833,0.10995281771407461,0.33059785671013464,
 0.9853950508247623,0.7968364718761919,0.5887058465123006,0.29549717127151065,
 0.9345524638136478,0.8485598643350173,0.6285652397488152,
 0.0026883210376824618,0.6599781952435877,0.40974138830005336,
 0.12917362016998402,0.19308589838348422,0.7381493493655058,
 0.08363511993952588,0.005166560300352008,0.2514262021050222,
 0.38327383315656227,0.5451599095092808,0.6345495828365438,
 0.20129484526612273,0.6485828241080677,0.8092864165599323,0.5085594230687116,
 0.6982028091677838,0.7424974265019317,0.6532349349413547,0.3172891576742526,
 0.2797047794314409,0.48177372360637727,0.4621700424852385,0.5911230377710261,
 0.5199400452640855,0.2985967559793512,0.2562196228618505,0.8334218403177047,
 0.04098745220294919,0.1705147881164707,0.15999316269253816,
 0.11554946480963924,0.7490381844486351,0.4397466432423007,0.0382655167675825,
 0.2812173491809573,0.7348183471496257,0.06447031338474662,0.4941508471426912,
 0.8012054400232113,0.6003272619330713,0.0148087689709816,0.31453389445496127,
 0.7099322375626306,0.6438050366603071,0.5966978862178237,0.35365443869369106,
 0.013332839369565641,0.40737970410412605,0.18768574708210917,
 0.21700261427604894,0.08143001764789959,0.32782388023331044,
 0.3283757928951435,0.3243973346558582,0.26474078587116257,0.3884473083248362,
 0.8716192263456294,0.6024009730610675,0.18473600763715348,0.6546542657195787,
 0.22353125930209172,0.9263133128999452,0.2327411494328797,0.76987053877741,
 0.8972680059663807,0.6046287282678935,0.41041454483145745,0.2450503661732445,
 0.4962607541887293,0.23027987830145813,0.13792823787424202,
 0.6733200547005678,0.9806230542715204,0.5294922335345416,0.33808037330520857,
 0.6634086092442462,0.8561002334899447,0.34575943126968256,0.7112982552104721,
 0.7406814380317243,0.5543657476884091,0.9193218677283841,0.6473944569650201,
 0.18219596005485816,0.6439925651508085,0.4174389192546385,0.6079962972248225,
 0.9496636145733148,0.23091434588585058,0.8625156982737932,
 0.45724721450526207,0.5237763019381797,0.7442198308244001,
 0.06834957653631246,0.7540244362858637,0.793678562785121,0.04565301269836719,
 0.286428575531507,0.6322689277861411,0.6055620271828166,0.5056508601973749,
 0.11472307839233475,0.71648483907428,0.5614112917078841,0.10135694037375453,
 0.10095955823396563,0.9522713294922679,0.2960906183482581,0.5761330889275907,
 0.9263738521578762,0.703078650275274,0.9564359766515663,0.12157554899627687,
 0.8662540183156954,0.16511803148145354,0.6173363296525488,0.9851451074734306,
 0.1574331623725218,0.9417634053440103,0.3083676211413987,0.5719061376895906,
 0.1852883707614521,0.6876162139993395,0.6401112400964453,0.6179331995868855,
 0.8849524657117837,0.9630493014870004,0.34854690603542116,
 0.022485699576549112,0.14900625654920918,0.25093027048718,
 0.00631987436008663,0.006079175228191125,0.9273277179112869,
 0.25226322473983054,0.49687690291813813,0.4980398677637251,
 0.47009684701892074,0.6667073725997292,0.045432279604482595,0.27113658853318,
 0.6270537224043872,0.3579047134111807,0.4515204850267305,0.7376651283997183,
 0.43479341320220966,0.6370320928656831,0.6287613172798568,
 0.19015339808338216,0.317749529304552,0.8360896606851929,0.7499202116637684,
 0.10118638930833312,0.6275961901734813,0.6779091821047087,
 0.044848510599415414,0.9046998089603697,0.7485184642221183,
 0.9071210375152327,0.6407279659195187,0.18301063391906525,0.5644611015067436,
 0.8700181602808024,0.8817408930311924,0.5245123594718879,0.03449894965142419,
 0.21568080860826844,0.45842397464569506,0.1614485920731188,
 0.7850862518591621,0.8044692691071118,0.4408651628753554,0.9315940474225812,
 0.3388396051279423,0.10183221386781638,0.20304148905168162,
 0.8386523715912393,0.33014096844910734,0.9155330286724315,
 0.17203235141931972,0.8556369389062364,0.530435660302878,0.31786285473311393,
 0.3104052924819758,0.8336192942235049,0.13386377786761872,0.4369235816785706,
 0.7731773963942928,0.7716012973351395,0.8083110955121281,0.5551377694644369,
 0.6729383698789927,0.8260631066116207,0.5173552487061497,0.9830508715975709,
 0.682488882857569,0.9166516140257474,0.24024530020694967,0.9665623236171353,
 0.8962374601954636,0.11888532067783586,0.08798695802368983,
 0.9781913242695423,0.14089869783207176,0.4309071629984189,0.395735844494715,
 0.9934352084285369,8.619900384376766e-5,0.8217990288876543,
 0.1290869797229447,0.7134155593862981,0.4827151021029309,0.9037865680928274,
 0.6655331146559543,0.0076576371290830725,0.7766794114480136,
 0.5861381223903102,0.2218950669851789,0.9186717079770014,0.5238996030745533,
 0.7211901888554647,0.6139037380107555,0.039590480493862934,
 0.08474171502357752,0.9020264239942156,0.76339879892714,0.460017254863231,
 0.8567454871754046,0.34874781715552716,0.22109148766985753,
 0.38539343653733615,0.9461877967977219,0.5316879623122633,0.470607993192547,
 0.6293978023502311,0.1241858769377806,0.8376449937062185,0.7905655497682575,
 0.38463243969525995,0.8053664337878831,0.6693433788726378,
 0.19548556709375475,0.5386159217461399,0.5144199658306681,
 0.33639338702494737,0.47819493810751934,0.7417790869736691,
 0.3700315094936236,0.383908129760137,0.12350172827126082,0.6198543612634597,
 0.19055399818608443,0.9081269002637632,0.06631515403318478,0.994663485130229,
 0.7210502189764554,0.15893080525891523,0.9530637526386033,0.5239506098574026,
 0.33959430769388177,0.648415609104684,0.8682933728721918,0.3529232384729589,
 0.1555855535976477,0.16167845815950233,0.8001742906992126,0.470515797907614,
 0.06947311538455925,0.7318801514672492,0.05994423187022868,
 0.32753185947227104,0.04305890466763629,0.9501571306623572,
 0.8114024414572049,0.5940922884886918,0.4666193240050993,0.22844936894671164,
 0.28986088048268643,0.07491001714813905,0.5828122992468273,
 0.7403187163403251,0.40741361544517984,0.41511738749965466,
 0.5804434053386234,0.26252900475375185,0.2039547850439538,
 0.09711646047244482,0.5657570300143812,0.9938790605304025,
 0.06702493633082007,0.4530653588236834,0.2479402549030929,0.7502062278748752,
 0.25921286506232466,0.3379675046907574,0.15196131556124903,
 0.7791539383067547,0.6210676991707413,0.7340487005645798,0.6750011135106975,
 0.7540100026499187,0.3330510843830162,0.5946932309435192,0.9948659963001166,
 0.7468413945825172,0.2476869930697968,0.996282714857191,0.4381340165223151,
 0.6758161973595542,0.7207938006171932,0.25342101400117345,
 0.030731115979723556,0.8865965865272282,0.9501112746569295,
 0.5145637233964008,0.9226776855478978,0.8518682624886289,0.9944869930892801,
 0.1870932729988275,0.7475608723717225,0.6133358946049816,0.591888800398628,
 0.8210488876466335,0.9655982733584609,0.43943581917825514,0.538510217252557,
 0.9399799271939284,0.8283741229601974,0.17487991930870217,0.8057102885559378,
 0.2882143670073233,0.969767403715211,0.9871889772985166,0.5188666628322405,
 0.660946365383007,0.6605281233245687,0.06862121552182365,0.9406339172228265,
 0.8031792244875393,0.8875587276731753,0.288663573318069,0.2777604531403385,
 0.6369255765740557,0.1875584554198958,0.7517657930093691,0.7505849013349674,
 0.7791389420793315,0.6323980164576801,0.7185575139297651,0.1473076274271088,
 0.6929097169064617,0.6647224945028868,0.8874268341800712,0.569828918279192,
 0.633619761652774,0.229742550183476,0.44111381445208964,0.23258639976387063,
 0.7666252621465408,0.682610125495373,0.7614351999004876,0.22987319464075529,
 0.1600074824530624,0.07205712484936733,0.4652082018859419,0.8590834650603643,
 0.021384265395171687,0.5937770931473072,0.8388942530777096,
 0.4624516700087907,0.7593002841037424,0.57443693297131,0.016554158203301306,
 0.6841800490238525,0.5091686766703628,0.4953548978294442,0.7296282780566876,
 0.2171695917392008,0.1374686408237803,0.5413043576880767,0.08104301593740881,
 0.6211697838235894,0.262489823703251,0.2761586224757593,0.5299242134580793,
 0.16394727912970142,0.17686419856244306,0.20184563905975728,
 0.21503462940850637,0.15750044853371614,0.42238576275307527,
 0.8468681047990465,0.8941189925838492,0.32963244465176345,0.9633476481083387,
 0.694807005949596,0.09599460802714477,0.39106940657149514,0.9952542467231613,
 0.4234217556626829,0.4902095086525109,0.14299163626269862,0.4984760476300557,
 0.5445638789644429,0.9941819083787118,0.8857808881038535,0.42679069926219815,
 0.8672423958257182,0.6175239114089103,0.03194785604381056,
 0.15648094265166368,0.7333012953128202,0.15415334047529716,
 0.19214831491206463,0.9704771224174676,0.38087222080590855,
 0.17404664512305523,0.725272723859962,0.1939117033889295,0.7659887687376346,
 0.12668049741497034,0.6193955387153316,0.0386962193637197,
 0.42567102719409733,0.25146278166625646,0.06527137064976374,
 0.38239112832396605,0.669878482710271,0.9334583829415345,0.1469730180109352,
 0.457380576701924,0.7883185568178515,0.10181108946262762,0.4572575665351605,
 0.40009881331034425,0.2764341731221689,0.36758568546386294,
 0.45783790302651306,0.5058658966012649,0.8964304627140464,0.6354710649646367,
 0.9948158793661626,0.4013373392911579,0.3556164827860845,0.10095211697258322,
 0.5737335286823854,0.4348974362940342,0.33288925813709835,
 0.32843562822749006,0.026270987542581947,0.3319937975396403,
 0.5317678852882157,0.3393741596090186,0.49215069620034746,
 0.30308192232570963,0.9122425062836824,0.24914314681859318,
 0.8225906958758619,0.2940545451046316,0.42233455043850976,
 0.41555506314804946,0.9024434685352424,0.5299362016573297,0.7495061961850821,
 0.10006871900746817,0.09220541910239333,0.07399721514782431,
 0.5639716622240372,0.18822128402386218,0.145249363368808,
 0.032165395445252765,0.20989737048787305,0.8284103586475858,
 0.29460966360061747,0.6444366203046292,0.9368781425550742,
 0.19362770840213528,0.9244325191730406,0.4962444281316122,0.608924522965585,
 0.5565318420188688,0.4701457024491965,0.6656991006646379,0.8429965519540088,
 0.8394984757303939,0.4941665629632609,0.3109211215975145,0.4572946644055529,
 0.6721320207759858,0.2625398407814201,0.9596301945974888,0.9017758931072172,
 0.11291393192015509,0.24701631000009794,0.7791183357231998,
 0.8899681894885274,0.16716166825971224,0.014756640600830151,
 0.38625296376995255,0.3658709746569816,0.9571491066843473,
 0.21229715507656188,0.9443456259078408,0.3205989515320272,
 0.08772157086452126,0.2990243558597486,0.2384556386343738,0.2532241224355847,
 0.25442015074895685,0.7055954951639283,0.6483410117466699,0.5493681273284977,
 0.8006755237979978,0.2532926038623652,0.8276986154915512,0.09726672275681347,
 0.9628912296530056,0.15795596013132418,0.19745470086803074,
 0.16975912601807885,0.3711470861601386,0.821198204511705,0.9589641520807942,
 0.7110424942922318,0.9159349468649872,0.6777801095528895,0.4485644451849372,
 0.4891894796226467,0.33941208401316614,0.4336100244201303,0.9694348259717236,
 0.8447228015388437,0.5353457429579112,0.5610823498294386,0.29306162362405297,
 0.008488331003228544,0.023549413644204598,0.26720430142749163,
 0.524877108497878,0.45633311470608284,0.9412547902315401,0.29282817335157474,
 0.21962576806779177,0.818638369160373,0.010372629456903537,
 0.6929247612918088,0.2758659367992552,0.42553830365280065,
 0.012489326052585015,0.009881388241007594,0.287303106793811,
 0.15600930387186818,0.08908304610815809,0.6894566026297015,
 0.14513134037823427,0.36166666483054855,0.09919769794014777,
 0.6930582832225388,0.5771904106601164,0.7257723709363436,0.21214365383553235,
 0.10445391681467475,0.09364471787625761,0.016379808449319966,
 0.24723594073045518,0.27910952493608776,0.7947972518058241,
 0.7236100986267615,0.6679503112410351,0.8844886274017897,0.08363711103674198,
 0.8327614689898016,0.5465728072493088,0.2778281366418268,0.08486138723165071,
 0.6185302591422088,0.11149329417593734,0.27069316932607146,
 0.7226499941828046,0.8587337607956483,0.030106756123199463,
 0.4953802235041622,0.9433732755951201,0.7740382450675107,0.37997978040026725,
 0.7933944086884379,0.5644515638601472,0.38042005091332476,0.6779638011162292,
 0.9605000363888696,0.7167932725968805,0.004708557720232487,
 0.7476273021961526,0.4095155600084426,0.44460221353759666,0.5403998720538413,
 0.028817680304340065,0.6721675971078103,0.8017459024895944,
 0.49971129684885107,0.010640966903591731,0.8242393190238178,
 0.49302415395644994,0.3541411628399479,0.7874913034590233,
 0.08905038005931931,0.7423987712704351,0.5346540642735966,
 0.17193941014731462,0.9196933684120955,0.6043493922894113,
 0.37484715024670223,0.11717904103182153,0.10609111385398817,
 0.03243066367190284,0.023786211943167768,0.5119253611943351,
 0.030931866283679166,0.9410920862496662,0.6844512170399485,
 0.01968292026127294,0.851058113628812,0.02760853954293363,0.9277185353073369,
 0.9491038043238267,0.23074555833459076,0.43884571492264635,
 0.2008405545373101,0.3199542298433188,0.2647002219032265,0.418259580979655,
 0.2546989454475017,0.4183093979188347,0.4928522558753219,0.6343117195323064,
 0.814829467653548,0.5707069789593234,0.17287850500645507,0.9358768429305884,
 0.40090715707239444,0.4050443355483968,0.5979183164750473,
 0.45155496817527685,0.6666806991173653,0.14493625965403611,
 0.1354244132838115,0.5056922083809267,0.034924136624498114,
 0.4772482809442554,0.45945501143481804,0.10653481563804879,
 0.9327881229869343,0.672628734607832,0.964307830051143,0.107856762069655,
 0.20178360126085348,0.9348563612418322,0.5592572467792574,0.8693003175525569,
 0.30290341553897704,0.6585001743938269,0.9047647650690248,0.8954344398203223,
 0.8769852161582078,0.10287761773058102,0.57427126226233,0.60486260171271,
 0.7559706991262978,0.2741495269540357,0.16806654057023185,0.7298910017145164,
 0.6506792178249963,0.32237824733327414,0.03963100387482732,
 0.1144627520067465,0.011620218930404658,0.0056914546331273685,
 0.3991772656089223,0.7489396567202918,0.9109920470542839,0.1974002291768674,
 0.30393351113867517,0.8764277416395384,0.13969957064306282,
 0.32848682820613273,0.35397796236649415,0.23188841229763824,
 0.23615633539635883,0.8313973188052272,0.3601768065284452,
 0.25160927442829795,0.9179326916702684,0.6721712067646718,0.637510441909096,
 0.16924955548534226,0.7992470356337751,0.1886744255954662,0.8775067714851655,
 0.9026582343809191,0.12159734980561543,0.2948744705852411,0.3146359428578238,
 0.8054733872551283,0.024999167931507582,0.8221717181438378,
 0.12951063452436307,0.998862356850898,0.3860505678050241,
 0.021283137086954174,0.22277211299926003,0.5589737887383652,
 0.28488432364929317,0.9718833764025235,0.95786254175164,0.1488572813785325,
 0.8659901984775854,0.21031044946707178,0.8682353749580813,
 0.08311924770346568,0.06506486411529355,0.1934100840220756,
 0.6010810268557931,0.5511251125443347,0.4137364813413822,0.9774756790849005,
 0.9539793120503259,0.8853156983048668,0.463791223160521,0.8209387542105145,
 0.96043004170087,0.6623353155767039,0.41212963614197473,0.7716087778297671,
 0.5628250109571786,0.33636431200963346,0.8498341745677401,0.6997447201060006,
 0.5597510150768271,0.7980608038417313,0.0031416241276499557,
 0.029569381365267877,0.9479491217037757,0.5359818422851944,
 0.15783885611378734,0.1740364319070541,0.9628850192589787,0.6517846728590505,
 0.12950134032873362,0.04722589136738686,0.35151339290753814,
 0.5460069488990474,0.40821214140451867,0.4753432849875667,0.43944791620795,
 0.4924273369348354,0.1349564855241565,0.1377026584042531,0.5362730227880342,
 0.36536816400455197,0.8158156053631512,0.17139915164945307,
 0.19882619458677753,0.6329797269701256,0.32718382005290203,
 0.8610630136162907,0.9589953824533656,0.09054705694723253,0.5343579265881433,
 0.8093929409726685,0.4287327315903495,0.8384778147428841,
 0.017434448432489202,0.6435393003524937,0.3123080683836691,
 0.27757912637842375,0.12639993638183322,0.27272601511211714,
 0.00430303132758747,0.01239361966059449,0.5408755583594246,
 0.2936866564777597,0.46198512307407924,0.4191864377512232,0.5523388818405646,
 0.6930718778438922,0.5795589902848501,0.8267537244354888,0.7034157323712886,
 0.543262432548151,0.26222237845627916,0.9394691324320503,0.8502391605618884,
 0.19753404777310746,0.005093959219445665,0.5365008593603168,
 0.737339905793141,0.34486807701461597,0.40171960879830326,
 0.21955703376011515,0.6371506938028996,0.5498526034387915,0.5368419095965686,
 0.7322098783686917,0.7472745677896925,0.2735240931284486,0.6703540340553693,
 0.6904780636909551,0.2728845948667479,0.7902437308875626,0.6054902224648008,
 0.7208379976924826,0.21657014918263306,0.08624366107240511,
 0.14696038671793799,0.1918055523354506,0.9446224167623805,0.0813502373627043,
 0.7488861402456596,0.07364494379265485,0.4296276857016159,0.5675196352629542,
 0.8911102214120546,0.29862861534348606,0.8624963042424413,
 0.060462147153326984,0.6520223216255214,0.6994957362795254,
 0.8130725882679811,0.6102755094922299,0.9297925927896805,0.2539199814053724,
 0.05210921616595965,0.7991186933451817,0.9395275782327267,0.7968775016148046,
 0.9257375293266263,0.4192014661261262,0.11234146912275023,0.796001405793387,
 0.5991132945640472,0.21412229126983506,0.11052086428541796,
 0.05873759150013813,0.8799713932764657,0.8880532655739045,
 0.03755182653417144,0.8112482128265194,0.1971409379371719,0.5175903121431341,
 0.16880016602308923,0.305444385593566,0.5048777326268616,0.7025356117768673,
 0.93438283964191,0.5099833912546392,0.009449897253121065,0.6315023975135553,
 0.0615413783756279,0.7300231511914054,0.49341050529335084,0.8023405239990845,
 0.6978624778788757,0.19032319360665717,0.6218064378963176,
 0.24779373159060558,0.6400880562787731,0.4653372274272479,0.3222861570832536,
 0.8340489480344668,0.5198717386122875,0.9440731402632938,
 0.013195220557803089,0.684160920208034,0.7379715221481368,0.9435097357656914,
 0.05183848382741918,0.2805759059112365,0.9925751435217013,0.7822890177186849,
 0.6778408764859524,0.6575948422226012,0.3665610548262379,0.5512337562704643,
 0.8264890025809515,0.3400627899724361,0.7684673027836368,0.8322574697546895,
 0.9088084212199116,0.3399124150688384,0.06427196582700401,0.6989950386373713,
 0.3912358877563247,0.42985092536950786,0.5875867856280012,0.8314293219386328,
 0.14824376305780007,0.5028580377969298,0.4871484891518728,
 0.009291531744132664,0.4779292244039205,0.8139396866933623,
 0.8391271297924034,0.9469726972168547,0.7504417127959766,0.10749389990304503,
 0.05733200734203803,0.1676651834328442,0.7146771237651484,
 0.026760892560492744,0.9448491796085309,0.43261376442099275,
 0.7872011956807827,0.6771616584379192,0.49370933067412126,
 0.06632509132254738,0.29190602100327456,0.23949236994012024,
 0.5586758385001744,0.7425251442991362,0.40279113671344335,0.7311575172947142,
 0.309160895240976,0.6411749217639231,0.9850194943869865,0.632688026341767,
 0.12233097324980235,0.10904379822915289,0.7077363216540811,
 0.4931487040265149,0.4143726449357985,0.12812937331386232,
 0.33071683649462813,0.7283852117009586,0.8640120231953968,0.5314649452952083,
 0.9427340011139149,0.24380270627024758,0.41322414348825265,
 0.8841696858775087,0.7752093463228115,0.7583268931911158,0.7774900160163831,
 0.6375205958088019,0.9622510305142657,0.5754640754626216,0.5853924805514248,
 0.7471638384182073,0.40001463825988515,0.9884024594728504,
 0.49036805531024363,0.9540022031691036,0.5470039173556098,
 0.13447339978374218,0.17065612056829793,0.9661889248738948,
 0.16596136080633261,0.017783132620718956,0.531353043418083,
 0.30949023897811523,0.13029233113827454,0.06081899365894605,
 0.057346367477513094,0.9150059226319943,0.26622607923389174,
 0.7390776638992212,0.22098229176612938,0.2455341019694498,
 0.35669216161957906,0.08576385521249297,0.3680644242253781,
 0.47080347052175975,0.8520893748254036,0.8387165350884415,0.1255068049834207,
 0.9945963690335925,0.016034001014290422,0.7920568576259671,
 0.4430567831513791,0.159144182396586,0.9313113304956684,0.5852252102801674,
 0.04477997685681834,0.5486175672507005,0.9754736488892832,0.2862443112812877,
 0.8857907752136494,0.7410339376976269,0.7062348953246905,0.9640015143140668,
 0.8471664235693844,0.5774306478701676,0.059971531854058124,
 0.7110009193748597,0.6621899493380984,0.8801490464420365,0.6278915630713139,
 0.5410815637744857,0.5667774915803625,0.6477180701034871,0.33392291477636016,
 0.04953166465218961,0.3888764700569145,0.5588899002048128,
 0.22314540068806754,0.8149396362443846,0.575087657818968,0.318588898954131,
 0.8502506751222602,0.7385745392993888,0.1857570404323825,0.3700921406629072,
 0.6937904923825409,0.6198093805727038,0.4359391965080246,0.4113542560690383,
 0.7809654387294198,0.7535434196986827,0.041572490336435175,
 0.1626564828649284,0.5709339900690732,0.2709857478627906,0.7327056058016742,
 0.7078121493273434,0.8941429936502676,0.4806906785206395,0.18886743405443307,
 0.8552586392869158,0.5662327608414868,0.31304633835347184,0.1445651631407956,
 0.6305499473814377,0.4029495694163656,0.624250200739467,0.9320847531661115,
 0.22099445999804923,0.7493617297400704,0.9570273151163262,0.7194627318564426,
 0.6660503723439235,0.9500888142903205,0.9615592775356735,0.16102133047585254,
 0.8601018733255673,0.6612818922687286,0.5501626037597235,0.5894661579281018,
 0.8840968733952197,0.3078869431812209,0.3346345672459118,0.607018264317922,
 0.501136929105487,0.28312057642026234,0.5067431194082617,0.06760281952095404,
 0.21995179256955077,0.8024808553589047,0.0787755944719184,0.4143197446033632,
 0.16546866030419827,0.0179360253788986,0.5075121532278464,0.8491254077891737,
 0.6927424787698389,0.8606505411636236,0.8602732578529566,0.23813062529578932,
 0.10317094297200413,0.1882313704504477,0.5329924144497833,0.5498821818867135,
 0.6256559109185854,0.6907033048080481,0.29867599377495324,0.2462279722446692,
 0.5799649789543211,0.5550722616888516,0.7514743239865732,0.4950429783969216,
 0.3112730859277738,0.16079736465999206,0.7700315744071355,0.6350149150741261,
 0.6471238362947599,0.4786741266282434,0.7689902699332013,0.7475144398287003,
 0.5937922095329308,0.5042594220303476,0.892473357012911,0.7645178201305236,
 0.06484652932202695,0.8842048444407254,0.6627689392583593,0.6743136950647192,
 0.47573972942078957,0.8791098468857383,0.04644924619981472,
 0.26577730925853915,0.49299223417381366,0.49428691777057654,
 0.8433273599335271,0.7174631829839124,0.38036058170842946,0.9711328261074975,
 0.7550924476891262,0.35729118340445476,0.049194864709301456,
 0.6154397343510327,0.48026355565737777,0.8025881228021745,0.5741995412125903,
 0.4282587004018481,0.8855562741795503,0.7977942959903923,0.7929126195903349,
 0.13479823630752097,0.2262778205182433,0.8679884819421761,0.4272423231247531,
 0.5244100790618053,0.19610994772154478,0.705189096795342,0.31531731945943453,
 0.263789570733172,0.9692767681634099,0.7651808861926435,0.29637028643074004,
 0.827696788818962,0.359608793051983,0.5693853831117123,0.09666084709384726,
 0.4346410311480642,0.8215965381469195,0.5426725333479125,0.1264358909235006,
 0.9634293987020232,0.06867603114039322,0.23288207188316612,
 0.35903561978996124,0.6805315629962273,0.8195343010252547,
 0.36270762518786737,0.7208495784717379,0.9356617230918011,0.4615533867458935,
 0.041323880191190376,0.022795929331649756,0.010392024099904806,
 0.15411939753552484,0.21248665456888183,0.15670617779683205,
 0.7319849266025258,0.4022986968485126,0.5765126807154437,0.2158831666723714,
 0.5213818678857008,0.11087244696287013,0.23993028977821118,
 0.7711972082941164,0.5822934827535394,0.5928064224028166,0.46579138821676047,
 0.17563659195488923,0.8213815022942949,0.017309195868224325,
 0.5828909685781778,0.8946227618865928,0.7094415488010727,0.03474461255593342,
 0.8585905629047991,0.3355326310848067,0.31723718800274237,
 0.16717042585514064,0.4432525958529512,0.6597749586624213,
 0.029846496174856796,0.8013316865375683,0.30327496434894974,
 0.38940155702810086,0.9745961438956715,0.4437231841885625,0.1681243393025602,
 0.910468981466089,0.4953864072598674,0.5796560419818566,0.9369191629153785,
 0.5829156318862511,0.0932591475987109,0.24888822255163623,0.5805743202302202,
 0.7491112000575717,0.4144929851228254,0.9098749745712682,0.5919994071408947,
 0.7630807324785318,0.6225926968453981,0.9088306891566369,0.8265385661742529,
 0.5785771816588385,0.9353303287493088,0.7671387847814439,0.758959134652561,
 0.7539356131091494,0.9526046031273366,0.24345460609272773,0.2440380583574745,
 0.5610952047791812,0.5398540342905642,0.9050656098013061,0.6741183076509861,
 0.11982088059651563,0.23534931735844855,0.45120089876687863,
 0.3010313774358766,0.43483026920641565,0.8531010188923519,0.4169302255962116,
 0.5739478010852357,0.45092384606649927,0.8950171329266712,0.8258725670908932,
 0.15989029773263153,0.9030592644360262,0.5574887395491669,0.8037806170145241,
 0.24875331353895946,0.37256444627432217,0.32449035385365443,
 0.731159285880561,0.4160059934170788,0.3670726408161836,0.700982422936213,
 0.8981747620542766,0.6390483831764979,0.5551844933841462,0.40557737192858945,
 0.9485840648878382,0.18141469786327025,0.8277175382841687,0.871257134382883,
 0.4649670397608202,0.1524036941961039,0.37707341285779583,0.1302618600569353,
 0.9708490285146685,0.9491683760635292,0.6027188300775408,0.1459535189984431,
 0.5675395842457478,0.005923985525050801,0.9502315105354787,
 0.10623819540041723,0.7095772929912266,0.6087013684989871,
 0.12392790180326196,0.19110347963455254,0.4179288789985458,
 0.10932685204119608,0.24804632092051726,0.4819361124626478,
 0.8613915495643386,0.224476071377135,0.21938688539774254,0.5330844410415616,
 0.6422692092340953,0.2623351641700944,0.9820023711246356,0.6037312216262938,
 0.9206630775168545,0.41879211163521324,0.055317350797230325,
 0.7733357359094852,0.12843039934954226,0.364052044778001,0.787477545478586,
 0.06151458794420783,0.23478257025354188,0.7065219919455595,
 0.44149238261249346,0.10757491079526726,0.71142009765269,0.30474427853438424,
 0.6908990655504429,0.757495210848087,0.8466777555773507,0.8350208875310065,
 0.4089286825374341,0.21595462893442918,0.5757330861686021,
 0.056748847800795366,0.4332691909247912,0.19015904084344282,
 0.7548714628470029,0.7700237988954479,0.45960392381025206,
 0.20183546408500708,0.5731600386159517,0.37889872416561254,
 0.7580058410185259,0.8206229370060686,0.7851908477920683,0.6044702052411236,
 0.9263473148618397,0.2023282791047487,0.5607872987607948,0.14140622908869416,
 0.5890298723719283,0.4935140585120269,0.8968188998068053,0.289649207655043,
 0.758522287837602,0.3035782717359904,0.40358037626381915,0.37980640743554683,
 0.15958653774009035,0.704638143794625,0.9329557464712613,0.04029094517171683,
 0.9550507905060196,0.09541258360748739,0.8898675986514468,0.4471354698835598,
 0.4310162727399358,0.46423490555923685,0.9193542746699777,0.5539995012037621,
 0.7948737158006335,0.9119400369382225,0.018881756196986066,
 0.14218492885330025,0.9851850141149963,0.35641650639718725,
 0.15953309074696986,0.10022651038209518,0.2396798980895164,0.783142097197958,
 0.8668376044813775,0.43150003357278666,0.5901388185713307,0.8485708178544602,
 0.6380018321164345,0.11156980110234771,0.6744787875392135,0.9739625650233419,
 0.7739241755880474,0.7872091285361844,0.8490563573363851,0.8650930216539328,
 0.5849699950078101,0.6744862659425185,0.8021889839419809,0.0578896361589567,
 0.9019099555705847,0.7887305293663677,0.015168090045995397,
 0.17337028532532417,0.3154883814880428,0.4862069652105169,
 0.30647867734295653,0.3342455301408963,0.7684157943171867,0.5750205875819666,
 0.41070473835127586,0.5023364877395219,0.7241565022364033,0.5772336920419026,
 0.8985660780828221,0.7854346176109815,0.7301454913173464,
 0.023180670778224388,0.9277290826596556,0.9692531127409467,
 0.18029368161583292,0.17820100343062728,0.07950907407136054,
 0.773937556776557,0.4681849322176648,0.0880474217927868,0.2115722401269432,
 0.9973378079959414,0.4738236012027659,0.8279405868184948,0.9801944106353828,
 0.9601096157307326,0.10610159354923299,0.47594390553875565,
 0.9297293536064574,0.6320531582072617,0.46136734180487793,0.5691989604939615,
 0.8020998174370446,0.9264121989283962,0.33771917111351824,
 0.29664409047976004,0.4928249918810734,0.023870449859372478,
 0.15448805040794789,0.12073683748540609,0.710000298516893,
 0.24679895216754855,0.7481079604808041,0.07835584815453445,
 0.3793036703417857,0.2555068180358656,0.4677150642653035,0.15173960039206946,
 0.015079973122119839,0.10107251919434135,0.3661305378034794,
 0.5204639873222231,0.3713948032647476,0.4837018887431159,0.2635902818053394,
 0.12166061345425938,0.3385409571776239,0.9788157100778523,
 0.20461589949458736,0.9532576044122033,0.3254552054487799,0.5696627391291609,
 0.626663532598344,0.7832192296014651,0.543438040364724,0.8350128849980709,
 0.6144931671404483,0.09470984566606044,0.5599245787782746,0.5823938303004546,
 0.1511282157995496,0.9453002432877227,0.49476975534625234,
 0.25033265898861334,0.5169529072582213,0.746377407868351,0.18572487368871393,
 0.8970369534797622,0.8384240701936572,0.78646580296958,0.24222034061255016,
 0.3850422210196742,0.3939456334813939,0.3207228880969437,0.40845593554225157,
 0.757241690677264,0.08596451421302831,0.17200470549892655,0.857740489698488,
 0.4525514193245288,0.12317782612132477,0.8075939213906499,0.9942429278452436,
 0.8911016601893893,0.3250566812494059,0.19887359567648266,
 0.14543202246959142,0.8807402892211056,0.6543344827021491,0.6388363698631199,
 0.5611723564704116,0.566692037681345,0.8221086018752775,0.3038633244547939,
 0.4504560465202516,0.3170751754167833,0.027674479512457273,
 0.5724516025360658,0.8177995669142012,0.9397231612714047,0.80991190464319,
 0.22673348413344185,0.9861000275240396,0.6650224340265543,
 0.08525118370470253,0.10802651622492965,0.7564740532858513,0.517339825239458,
 0.7715897993266123,0.6604035937069614,0.9458815567541498,0.4006010956163367,
 0.180974855030693,0.16751589908676123,0.8788376560596589,0.4087503149939776,
 0.5789536238114997,0.5463538938383274,0.49220951751277753,
 0.03537523675000087,0.7035280535938971,0.5843532882770439,0.4069167972376435,
 0.32714550373117424,0.8685696159680037,0.689638704715025,0.8670447136001753,
 0.4206304145071169,0.13896738529989194,0.5078623563029684,0.7167333433006864,
 0.5077470163493144,0.21349983598281863,0.9350287624787255,
 0.07358669153110309,0.03527076103517757,0.047364328317677784,
 0.03583101907425268,0.16086729940147015,0.5585347803900915,
 0.1661816820065226,0.09803586483470028,0.9482048143309803,0.2871010638371593,
 0.73533860541978,0.21353763607766263,0.9634097690302306,0.4160052271993554,
 0.8918880468365151,0.11498419180104646,0.7959497475344395,0.6622947749820611,
 0.10318024779949209,0.3511143566083247,0.33535848488552134,
 0.04482508862620582,0.6365477687105305,0.6526078092054929,0.9734900158324654,
 0.9046082407494597,0.6204716576443043,0.6402543066884844,0.6897270931036921,
 0.7190324779346101,0.16048956251838897,0.5219689727041716,0.5841758579418724,
 0.1257035494128812,0.23853783061941602,0.8564741472904344,
 0.29997728605900675,0.3892674538222587,0.3141774291278493,0.5586796536722274,
 0.46205972182610366,0.5865385806732514,0.9702997700306659,
 0.22161096695354987,0.9591066814919267,0.6970687521687808,0.9077276429795189,
 0.50728543437819,0.4443426078690922,0.03107473251847992,0.8678469207030404,
 0.0981855856824384,0.48808576663497294,0.8356430745252856,0.9926982547829032,
 0.03863269969495409,0.5179872868312202,0.8916352275319614,0.280133162380066,
 0.5052882170986753,0.00586883627788326,0.14880506466301235,
 0.5048837350758396,0.30379704634183935,0.7217201355340598,0.2625758027155497,
 0.2620082241325693,0.6175671127737992,0.20558114433332297,0.9076524149087566,
 0.7492323906512114,0.12446484300578697,0.2245361536045083,
 0.31905278983520624,0.9514711512452828,0.29150833316740876,
 0.5305367940331965,0.5511217495879961,0.5040660434804203,0.7934494200218671,
 0.8675784324665833,0.17861081775326826,0.72853867459574,0.16513206628987542,
 0.9002376255151132,0.9602779972892422,0.6900773780659398,0.9137912766445554,
 0.2788022328992178,0.25340872361336675,0.8828908649631717,
 0.36022995164334115,0.614087592284595,0.7419681133312628,0.5524839424745042,
 0.0508109352901428,0.96192637288589,0.4595528471289382,0.3051251988348439,
 0.7268096481882663,0.09695159581526991,0.25485547411914,0.02467894700017348,
 0.1923049106494077,0.5313739777690554,0.2826294325898334,0.04331552485746215,
 0.0434534920799895,0.7689980691826281,0.30834454300545333,0.5569150836109684,
 0.5018044034460557,0.25135430909730117,0.1827166102596922,
 0.05529018075982217,0.9793954398785655,0.8952068829023176,0.3413831600163324,
 0.28592476910855247,0.6160966201274299,0.7282414209735882,0.8015319936952551,
 0.6523994054081961,0.6724825541872006,0.7536916749070732,0.7295665477174054,
 0.25484636370025027,0.021159468482335964,0.6063826998868034,
 0.4779551989611066,0.2943964102849631,0.2632391880517937,0.2695557606794867,
 0.4996956868969813,0.10845980198284655,0.14116144220087434,
 0.4347865643348625,0.7875995907990588,0.01650495169664168,0.6992755277938955,
 0.16722715984103864,0.31000442789687255,0.9456442643678512,0.474735882012412,
 0.9502076298661297,0.7628631954103517,0.4097343977171538,0.23571177323684678,
 0.5348833465355787,0.22223174568024007,0.15121576865411446,
 0.07803225612465492,0.06082578340158051,0.2536517895261815,
 0.3739615447137339,0.8997440814536026,0.8060141107803969,0.8902674766444648,
 0.8301371377611886,0.2632293981395003,0.7341061766646149,0.39481313008131114,
 0.9703616948479956,0.7326619009222994,0.6186450368120495,0.7428106369218475,
 0.3487052660760094,0.4352548147577139,0.4147581974066319,0.03787667287913288,
 0.6562745739679264,0.3833044442809772,0.36359161018018205,0.8893503889425727,
 0.7203734983332194,0.8014941386178319,0.5046899936471841,0.12106448450529184,
 0.5420610121503642,0.5099641191444944,0.4942272576794182,0.8184798141897801,
 0.20432431750919833,0.5044505203031584,0.021356245201708313,
 0.8554349096492773,0.15824949528017118,0.14573318890729836,
 0.8497353819768222,0.7626427977160329,0.6155081029330868,0.04776240086138017,
 0.5327840935545614,0.6054763337180605,0.6895946818156188,0.9596771075015275,
 0.7409872272699192,0.24754023831436656,0.18911728867940403,0.889487660035797,
 0.6030811114314627,0.25207699715065235,0.9779300362894874,0.7006725764200693,
 0.680323367776189,0.9334641477263441,0.11592713458894033,0.3174386866108362,
 0.9963348890319562,0.49676730809156555,0.873924044372563,0.9493021437155094,
 0.9921644487873774,0.7152130848549286,0.2634401266498685,0.06991612528677382,
 0.8595830888849533,0.6393034758634467,0.6861856189161426,0.7023706025558107,
 0.8918439314792894,0.7342838113973656,0.1046197266381208,0.3201003347428496,
 0.23225128945794815,0.004559478806018036,0.5480740027850627,
 0.2988968187622296,0.584144626588279,0.762137204587403,0.22666198415536687,
 0.30102441097425936,0.6659343934984395,0.14347435897618144,
 0.2873069898155206,0.8956192408542201,0.7936230570779539,0.6932903234345789,
 0.23377237338151047,0.45905894521110324,0.7591492297126149,
 0.44687971928122383,0.7694591937010489,0.37893062095091423,
 0.4089995594991067,0.6050456052394394,0.26738626005756716,0.5597818172767658,
 0.9899863975609839,0.3308222043569301,0.11922376928325984,0.6190512818518878,
 0.029500778369013547,0.33880064664946974,0.8510694927614316,
 0.21984089696532438,0.05501917071843532,0.015390558006101118,
 0.29689693981910015,0.855231939056692,0.03261057220721475,
 0.45061482754702054,0.4443259817547748,0.1343178960605711,0.4364227404415886,
 0.04597518612104956,0.08592285153105139,0.850615637744978,
 0.37838378077428225,0.15237572579465142,0.28216538137183766,
 0.5019467769174859,0.6624199216072746,0.5665313496239096,0.47713156020109926,
 0.970860123141508,0.08936456017293082,0.0665692846387147,0.5513075339557882,
 0.7555693637953808,0.9283608484184946,0.02334006635739283,0.2748677562681334,
 0.2641788371194077,0.2514314924022436,0.23441698249996412,
 0.45377638083950855,0.7600943231066989,0.03028106350968074,
 0.03402252809855366,0.3000285683957048,0.2617965974855936,
 0.27366109842535224,0.43333135462190253,0.19080660936511762,
 0.6057393019300024,0.5962555103940677,0.09702411262336663,0.8884326243597513,
 0.09463389991126037,0.8702104296730349,0.08489030860770042,
 0.8008283484292741,0.005972475694994039,0.13158659249658644,
 0.35824171699325735,0.8383698165536052,0.12380600727078128,
 0.16716659792728572,0.7707470050321616,0.06546647594625388,
 0.9318375989460261,0.39178310818772877,0.016382830847776697,
 0.139761142666889,0.06119635060060791,0.5034807285255917,0.13110762689150446,
 0.6727491141640067,0.993533288083455,0.4682177040798867,0.4328894748541421,
 0.23732659452214155,0.016263204072512027,0.8196568187463233,
 0.6714770874907532,0.4814581049169653,0.9772158146882532,0.6213630868845753,
 0.6909719974317113,0.7504976317901513,0.07159130369409905,
 0.11998368455813957,0.8466263845171228,0.17231590404423036,
 0.9207496844230284,0.40752004971890543,0.2237084032414488,0.6946128734780384,
 0.45773099360456115,0.08648120723260533,0.18166397609344698,
 0.5087944577622197,0.05831428241170722,0.33059934547911807,
 0.9860382381073552,0.7045908151927688,0.3589731064242949,0.11993040159994339,
 0.5435940694212068,0.7037594949992668,0.6345542992595329,0.07908768642599628,
 0.33560717452545197,0.7005372467713818,0.5135148018817659,0.7565131153425663,
 0.9914241900427805,0.0888027606930396,0.8863977136975391,0.7653594689750138,
 0.44093147157463997,0.04883001902204476,0.24334841669570306,
 0.1366577890760189,0.6658100953308231,0.37823291803717907,0.6527492772064744,
 0.9713900929615289,0.9459839228553961,0.3091510229924762,0.817375789771599,
 0.4028644573003305,0.43840303424712923,0.6412584027694822,
 0.37685728496807036,0.4109221631573513,0.3417632419575253,0.933618636799088,
 0.10401016859313983,0.20985231925457115,0.47626790050281254,
 0.6079609385738745,0.17331207568793794,0.9802325353914931,0.4648052406892017,
 0.5808003128442769,0.23445857680880255,0.7968294479602716,
 0.04753705878875936,0.6178407310700658,0.8912792800436017,0.8324383795578715,
 0.9227175833559116,0.5284852023868436,0.6838459778620081,0.8994196770743437,
 0.047236244272662486,0.1367200820443606,0.8456198315800842,
 0.22543948083106646,0.5507645234571319,0.17465390720938,0.5513171615798054,
 0.2823070620332876,0.12395650116530132,0.8288610539101516,
 0.11657429508424955,0.9594436993745354,0.3074106441440918,0.4411267018803494,
 0.07571347910362358,0.15631533386595675,0.9511557321080126,
 0.6706046027588624,0.8979908761068005,0.7397541749641378,0.22179155947840812,
 0.46146385473321827,0.2848374328671779,0.3718239691487729,
 0.06615541352884324,0.9551505853070321,0.7433001193165225,0.9694762528387428,
 0.6866024033526633,0.647388260512668,0.966813155882927,0.9878552648624996,
 0.5129975962210042,0.6124187380178887,0.47487339631177206,0.952440960168365,
 0.6032094385489313,0.46553964532777103,0.1269858279004621,0.666931234556343,
 0.0526461656727677,0.17239302891409403,0.716553327654645,0.6188505288458842,
 0.26207969064611425,0.838294525983652,0.4046564508527076,0.16988739342657522,
 0.0349501799140608,0.2846483029648774,0.6361838275003322,0.9616249606519142,
 0.39584939989453916,0.9242689358936741,0.3436704149795575,0.2028252074731357,
 0.629090183299292,0.3956091308652898,0.7800204537178932,0.9097664511916397,
 0.886528604774148,0.09445824400599367,0.06208111626413437,0.635688881847835,
 0.38032911223142574,0.2682219829460055,0.14439793569703052,
 0.8219926950966583,0.9196602108346317,0.11860703142489115,0.5148851475174118,
 0.9141748409526698,0.5795589696130151,0.9703861735035576,0.18951797088383793,
 0.10274539015533912,0.07779132186114712,0.5315370388548112,
 0.9869487695137611,0.9469875140588819,0.36384804667443504,0.6533638454509269,
 0.5196064932083627,0.6494992193733556,0.6064480525123547,0.2602507910408649,
 0.5811691628418625,0.378434700672032,0.10157770314069825,0.1998904657752738,
 0.5873655441318695,0.3928858052930355,0.6233394175237722,0.21218238211341234,
 0.5059676564694995,0.6868716354219958,0.2222965758841151,0.0955832774685641,
 0.8284099985574678,0.9269485846174622,0.09929233870050935,
 0.38474895991189695,0.7902223096200929,0.7783924008394597,0.0408833773458549,
 0.36144820565076896,0.5604381848378697,0.5247320618298584,0.8424497828302736,
 0.03804716873166458,0.4322660977386764,0.7750009240704182,
 0.16379330713010753,0.7521772738910937,0.37572991244302634,
 0.9250854683653762,0.2408676575537727,0.9692047082246656,0.3242829054772923,
 0.8875360348273107,0.17249833452410746,0.08520184123978758,
 0.5304744988606673,0.9147119216462466,0.6059022565071941,0.7755247747276682,
 0.6342660949000438,0.5306014166255341,0.516920201164095,0.07615859190287566,
 0.34760345662349146,0.9433961099731063,0.1514888987587969,0.804650435771782,
 0.6896727474977353,0.7722706695489212,0.10583385738317541,
 0.19411035076054795,0.637191102372947,0.25870938162615975,0.6664854980117924,
 0.16076012266522344,0.949721181155542,0.4993621923725856,0.10344053293826005,
 0.13047897616966786,0.25705644174913544,0.3566739301063422,
 0.9827067934877676,0.6760517369139891,0.6024987106965909,0.15629286844406487,
 0.3002106330981368,0.9851672304885295,0.7113996908239933,0.8206873863844331,
 0.23422377421265939,0.3686780620691227,0.31154019641644026,
 0.3883451163776822,0.8089283641499282,0.853135348270163,0.42064151475698197,
 0.19492465210789034,0.2287325324117092,0.6817596312163728,
 0.007132193478293414,0.6678088252187815,0.40939646673547914,
 0.6170248319093466,0.33521573710229946,0.7033225318543207,
 0.060868495791495336,0.400411414457183,0.7864316430724722,0.5283890111159747,
 0.8974971674143394,0.9925024065871081,0.3829312455012589,0.2467574276029737,
 0.09305139183161049,0.3221148650785104,0.2746023587415708,0.7568458924725631,
 0.5856358372467352,0.12698728729581377,0.7988678715901318,
 0.42168624370882823,0.29063063323284477,0.9529761669232321,
 0.3664360657458875,0.19945387863048425,0.7297323356624414,0.8583499249418924,
 0.8840262049082055,0.2910360959859992,0.7236940711593125,0.869641873070951,
 0.3252043604824064,0.88331689787735,0.12944503273963803,0.024568538136241602,
 0.9901441214062783,0.7218652983083499,0.6287033181785624,0.5868313172699912,
 0.5541188198520469,0.07144943902719303,0.8492689287297804,0.6508626788175496,
 0.8377779833380681,0.658885485180351,0.7985564418825136,0.7714908657310703,
 0.030560558360382406,0.6679634870017416,0.4739491546135891,
 0.9497179257181118,0.468319417394375,0.6490279190301075,0.2647770599601834,
 0.5172193486291171,0.06301404933946886,0.7637903913530989,
 0.32450766777738627,0.3095770561360678,0.8503352160916434,0.7388875769964908,
 0.3876025429200681,0.5085863777390149,0.747045405407564,0.37775817472464246,
 0.7294542755128448,0.43293877088349064,0.946048285872012,0.6722371875558225,
 0.23865515172934215,0.8061700935229632,0.1660624315030168,0.943733686036315,
 0.784695795336905,0.24747505795685754,0.4554100463549332,0.3138273791248296,
 0.3593131604075327,0.5196659956560297,0.3118910341993023,0.03558455224779089,
 0.8795428608279359,0.2781573171392001,0.6860662997526813,0.2544424430701915,
 0.21428525252566877,0.543205074315599,0.5819044546627012,0.2292585375968107,
 0.8384129020228696,0.9176246133454197,0.48641953060174614,0.8169578735317963,
 0.3307603529320352,0.5585767640264883,0.3589632951530457,0.7529217430513462,
 0.9684614425744645,0.4695423623470387,0.76592814566058,0.03570985391236592,
 0.5363038629443083,0.5636024989736611,0.4166420996659038,0.5202202533230763,
 0.860838505213647,0.5167953381357415,0.8707362969539802,0.8573127482642302,
 0.34844636458912626,0.08996848841241167,0.3423440170612202,
 0.40341126937969973,0.6586776535274552,0.7524056820310976,
 0.060906220523814114,0.33058234159482813,0.27968714127546557,
 0.8618919170647404,0.9603169278262421,0.02716531799301536,0.9716808036154253,
 0.9558715446517755,0.6600083610759552,0.9332652920799187,0.6575778247452087,
 0.7541045994470923,0.15748206131066733,0.6098391580770319,
 0.25027123321580236,0.8703269263847178,0.942884790675118,0.06915254361095746,
 0.8258946573844526,0.2321957296744087,0.6060414512394456,0.5296094108877286,
 0.7550588012588956,0.24296368585502326,0.2543297270531819,0.5423606429819561,
 0.06800695995408301,0.7909676842995829,0.9950249957243341,0.4038087117489664,
 0.626822764224947,0.3385915417242862,0.9372834193060262,0.22683334494513907,
 0.7133046352987055,0.6955781171304308,0.7893437980329943,0.7970703445383325,
 0.2455875224757711,0.6974451216785966,0.9151144918653971,0.16077148696751076,
 0.8038539219009189,0.06075798825778378,0.5541071033539459,
 0.24481796705957448,0.9208845297907928,0.1294961542309483,0.5147761295580322,
 0.9907573145038614,0.07975437263451302,0.2663185546219661,
 0.06466829270475305,0.38595494784674644,0.5434692660127346,
 0.8125565716196629,0.030980486935067386,0.5875676576481547,
 0.3608314269062436,0.2815196601345282,0.3406268180682468,0.8546859151973804,
 0.6576155123368295,0.25430590417917054,0.23218649612301911,
 0.2370578406860181,0.31806783639699043,0.8896381008384057,0.7337446450759071,
 0.1691892345474365,0.19017360966498242,0.6900180823379385,0.6193260141968686,
 0.8338333359705553,0.4276606190011344,0.990991972479,0.9954109794494901,
 0.4157350353176428,1.8949545175894755e-5,0.04053832713033978,
 0.6249271563217546,0.7903998699902344,0.3310922857986437,0.8748020236483918,
 0.31450594077112115,0.1269106996719418,0.1887313664065352,0.5608535220428922,
 0.9681020487884865,0.7166656536299157,0.6636720068419515,0.5329006632716669,
 0.19555513503567645,0.7038218797786786,0.2876470508611041,0.680267139209973,
 0.9644613368539292,0.8542888452758093,0.457748373673883,0.07998255403557442,
 0.731183394736255,0.8406187240970063,0.9715568181450717,0.18562359935777661,
 0.6093850104483544,0.7218316278639306,0.5413886590486782,0.43704815289343524,
 0.7682017830992549,0.6761781910952401,0.7658533979335967,0.49944645802772836,
 0.575210101830946,0.3217070385349946,0.845064822850567,0.3600798055175598,
 0.7877601314609262,0.7081493043101972,0.22595035245119088,0.4430283668648184,
 0.9171814375052503,0.9009548432160799,0.9697511548526778,0.19746747925108954,
 0.8292434749046536,0.5862686898237687,0.7878613624266096,0.7614725999702477,
 0.21860649737165938,0.7839370491021446,0.7263927280149004,
 0.19999500942259862,0.4501751543865219,0.15203823613344525,
 0.7517240820650324,0.5983235115906184,0.6849241124133724,0.936718016831595,
 0.9016608616732779,0.275700005156186,0.4854954579022903,0.6995531224075555,
 0.3198319515453657,0.7175886693723841,0.6002388520378217,0.9384850553333104,
 0.6051073968640809,0.23496233348319118,0.9812482703603816,
 0.12336123712242641,0.28410093402280856,0.32417388207179954,
 0.1484406502608968,0.24925653905667455,0.14652761700867067,
 0.031312654605677226,0.3634212903318572,0.45783825189899874,
 0.4530878358078787,0.10288361078849462,0.9002675212208576,
 0.25835065543470637,0.5275325674016607,0.705263300061613,0.16475430115711098,
 0.47754681345773886,0.3143233624809869,0.4523835433411255,
 0.10453618201471992,0.14061111270450422,0.5244587038582681,
 0.9270653935890864,0.0358839126820758,0.46873824723941715,0.6716521720201706,
 0.5290616669425221,0.015297867086574768,0.09146985065833746,
 0.558641365065138,0.5759695897328139,0.4497399888910093,0.5274781892739551,
 0.996349814766217,0.44286023583263867,0.432273392581234,0.2289014979930719,
 0.2836329916461666,0.07949652520062711,0.9584268232053772,0.193482612108962,
 0.46804972803319533,0.38124352700812547,0.3623843951891973,
 0.9955140035613471,0.8199177894228675,0.601280520562543,0.5340621184947266,
 0.06851101625324985,0.8005214622175533,0.8722655209716023,0.7140197002628406,
 0.850235989416314,0.4077596151073126,0.6660572212323788,0.4566157297372505,
 0.03465583299748565,0.618918821488944,0.04613104117267175,0.7007679059656952,
 0.5858986685585923,0.21192517747907746,0.41239948933436765,
 0.07909278671445208,0.5584150140468985,0.3326723016843791,0.5035398148541103,
 0.9026528512871854,0.4971420692346463,0.1665100415720725,0.24586425257193534,
 0.7453075018846278,0.4539721694397373,0.6502909410619243,0.25242027786450416,
 0.21610877508277138,0.6603793336193168,0.32073621445069933,
 0.9447187411691884,0.30950013267098575,0.1624950539734915,
 0.10881067624814322,0.29212032239364283,0.00663426980614823,
 0.4257325410749919,0.5240021898871844,0.03953898595253602,
 0.44635663229132916,0.21538158703308585,0.7938008909180773,
 0.5969035203953647,0.7968422054721963,0.06871590124459781,0.5734100342129279,
 0.6683667543700502,0.6285388837918605,0.13652377526975457,0.681296347398183,
 0.14920868064847415,0.24797118796898054,0.6530324186652919,
 0.7115706605400728,0.02320992103226205,0.9409416941057615,0.9106884326271917,
 0.5364955133972389,0.06295093815885622,0.7921278082247328,
 0.22051061432604135,0.7599083963159279,0.343294813698236,0.6416703450840193,
 0.9227910627641341,0.6360915276666589,0.06324185895763812,0.4769437455501768,
 0.4724719019864577,0.4597516433018769,0.273829735549155,0.01737980078142287,
 0.1079104833147031,0.6429450891324493,0.6382817498515,0.780845783841016,
 0.8469456283725589,0.011361648023021642,0.8032456766138941,
 0.1912504716661917,0.20631529320513442,0.17139289926983337,
 0.6933639164803485,0.5199599879681971,0.5311314365840625,0.5326334907562177,
 0.11890747753636344,0.7152945699210116,0.40630080194351503,
 0.5680324580025635,0.5176433614705234,0.8988202336254497,0.5326814631925292,
 0.965935988737414,0.4042760594534681,0.3965201813646655,0.09692577480943709,
 0.06734939771976134,0.442191903211969,0.0812635913925962,0.6140971652595943,
 0.43785407981667124,0.2419522247937278,0.7446979818415037,0.621036087581703,
 0.8041774258348138,0.1338516704937962,0.8991934021970656,0.29034536660570676,
 0.9266671972322946,0.44913927136868803,0.8784034859914772,0.8713120525358935,
 0.6975879070916741,0.8118985064175327,0.656266701257282,0.10879600998821592,
 0.5221004178339552,0.0824307124454956,0.027570354470661695,
 0.1941182316922867,0.1458069959801318,0.5621786001942443,0.6466432910216889,
 0.8815928992744222,0.8162398145683776,0.5794611516085187,0.6333642700470427,
 0.4579193909227228,0.5105219390835007,0.16536102400441366,0.3787345362342812,
 0.800299127788022,0.8089096707336008,0.6110249125595872,0.6508084358925719,
 0.5491506288966704,0.6268329720596789,0.7810561521862197,0.4954361004353316,
 0.540917424018261,0.5462505146249308,0.2883178490165724,0.5977481972891063,
 0.06947419123685883,0.44666308633455665,0.573472765238532,0.7401327481883764,
 0.35776580372334565,0.7471401138535233,0.9660631747912369,0.2656947278037616,
 0.7959436609223363,0.13645168596050772,0.20086158975707924,
 0.4462573716380076,0.2841110845115211,0.32622239174020473,
 0.31613949660634055,0.42651573875833193,0.8221498975803496,
 0.12205836146166527,0.7743021835783838,0.3675858538836174,
 0.05103229129565068,0.8040689333092348,0.8549609146139927,
 0.04662352440174611,0.18120434935589946,0.36742133863964677,
 0.29975745944001864,0.7427133153472002,0.30339483254873967,
 0.7230821265256997,0.5851548133443875,0.3798468587521111,0.3521842652082219,
 0.8015133885019899,0.8577187782895903,0.0964177871366394,0.41176085198381807,
 0.2791386593378353,0.3134742882317758,0.07770093140409384,0.7380207217795347,
 0.5498024241357624,0.5782437609878103,0.5479579209347821,0.8354099645483918,
 0.6303690809229174,0.28362898691356375,0.4806959978187436,
 0.34036019813267315,0.8226677639366914,0.20620590845437614,
 0.03441539900884483,0.4825332837380034,0.10103544766040806,
 0.13529265028522763,0.42410084801847336,0.6803870774438354,
 0.7090828785600461,0.5576512971002706,0.9351474789910593,0.35295051744292205,
 0.7218596874234822,0.05569000329412377,0.20198979987252408,
 0.7945325870910538,0.582762872714288,0.789401332569848,0.8981468838602441,
 0.050760796539376485,0.7701020205062536,0.9729898118325602,
 0.5596732413094704,0.024434652926154432,0.8116337231154227,
 0.44253046117552775,0.16423362214635218,0.26838187740793285,
 0.49529377438482813,0.7150597179233185,0.009125373364183842,
 0.3285695821012803,0.7295711164156428,0.3950657634306931,0.2641248490964788,
 0.15488243110874378,0.07601268606112443,6.422080329242519e-4,
 0.9039806996497887,0.9764490624283654,0.7304227999391408,0.17647641090294375,
 0.23018701998841173,0.36085265627192487,0.37569181213424885,
 0.7606602398220053,0.3966188928940907,0.9864740212735791,
 0.012504845182851665,0.4058868371033447,0.1686485196420653,0.360561269385532,
 0.02010511980465668,0.18116233356149503,0.5995947202053776,
 0.9736703823792356,0.1549568675369264,0.8709152803425128,0.3643428591543657,
 0.17269591802083595,0.13182538387842768,0.5563367330537294,
 0.4062264572160479,0.4347111682756738,0.9851475373323975,0.9699154573858845,
 0.39920328697116003,0.40029588463561594,0.74088743835444,0.04173259260232909,
 0.5331985506156484,0.0805437612750648,0.9127214108501311,0.5567472927400116,
 0.768093286407628,0.6650411981293115,0.211545678828859,0.20724188093951246,
 0.38082593685581867,0.905263050358003,0.798676690272544,0.43666397244932065,
 0.6080084386513018,0.2525817311489298,0.8791187735064694,0.2297150835267604,
 0.5796721635232411,0.11846298112280351,0.8703835530493786,0.831575374198455,
 0.20927982229618036,0.8583421569908458,0.6170923382574964,0.6278335336827258,
 0.03624622682737977,0.6586515701792796,0.9879324156233578,0.4215836291410183,
 0.46536141686343024,0.27033131496103857,0.5830946623209781,
 0.10713891340911486,0.7496923790235472,0.5899504164980699,
 0.29362289148206233,0.1457980320872756,0.1490082010907715,0.1775076811137598,
 0.7335684280528469,0.07561900043445513,0.22534959490549167,
 0.7425586144500372,0.5930672027876333,0.03254593206041945,
 0.17678942017900956,0.7179278964231633,0.9919057682537573,0.7668869607433565,
 0.4105675856382659,0.7465449867889059,0.6152858486950321,0.02848595679632504,
 0.32716634426106683,0.26501881214171286,0.6931462160686267,
 0.8646983806315305,0.2245631883378033,0.2825993796256806,0.9000526173322807,
 0.8562966763150499,0.31600307402408523,0.19561729932467442,
 0.19801072464435454,0.26372081252701596,0.7805268135274864,
 0.9730451561084013,0.9272769830011226,0.6993404184530466,0.7140471531192373,
 0.6624256339460552,0.5343015061065943,0.8341046141508602,0.31946199049229707,
 0.9965605753280989,0.7106010065040596,0.21615613737959283,
 0.36182514733449533,0.10289519365046851,0.12731573089871162,
 0.21906241454056807,0.8060679550873363,0.10176850346718602,0.941445947765714,
 0.07989782567349402,0.4796550030189014,0.6134087411185793,
 0.06067603594091486,0.997550474521799,0.0019468971214484654,
 0.2803466261050327,0.8398168760782312,0.22004967180058732,0.8565239021061941,
 0.7231564050407776,0.15327616256872723,0.9066621780500661,
 0.028722682395558086,0.21447206258366902,0.29306627860787393,
 0.12250111258070684,0.46554277906985353,0.21648973167426533,
 0.17083058633818504,0.15955452744743248,0.8972407184213176,
 0.5710260907141645,0.8534389874657438,0.814632807617704,0.6972289919089937,
 0.30919378334322767,0.2579422767520836,0.02810189649472339,
 0.6145772936633004,0.8119142200310314,0.8573474689990975,0.6287551051099749,
 0.5525283716266338,0.960845590186281,0.17612215231826722,0.39251772058518675,
 0.8418313652105169,0.7494662174567802,0.7097611321509398,0.7403561618703457,
 0.13389458209526284,0.38450762015053985,0.04071976788865839,
 0.03721203729537248,0.38410222011963113,0.052454677124933013,
 0.13763663813414317,0.06687903561627784,0.9128208432217123,0.260382883159176,
 0.5893703836667051,0.8775975523710451,0.19573755180561525,
 0.10785359173904308,0.7003867361376515,0.2218757196309944,0.8475805655287912,
 0.920845832845947,0.9127675141176723,0.9822633041841938,0.914724154152654,
 0.5351606875657109,0.7015876337547704,0.7906821636269505,0.6977157584013801,
 0.27463130623608034,0.9162111482343349,0.8765908888931349,0.9081017757214459,
 0.614201210012747,0.2512650355207412,0.3273690479886465,0.8787329626084628,
 0.8240661796889535,0.2009268464317393,0.6288821534803513,0.620531294365066,
 0.039620064284942136,0.8524897570100827,0.25917386580011736,
 0.5229786824453075,0.8000255345769857,0.6456416077508137,0.7883503109027155,
 0.1358307333685801,0.9662072193573615,0.10053477853446391,
 0.19972842849593286,0.2978999310459489,0.7151020151243964,0.5094447803535526,
 0.030797566204253668,0.13025903985475495,0.7766578421309722,
 0.6794008563768424,0.6022375895768544,0.5491281936176687,0.45778969927610313,
 0.5653216574072916,0.4417718962866468,0.6958588021126135,0.6599356452072628,
 0.05310124907093583,0.5065760088330755,0.7260409056574827,0.7584357266826312,
 0.6409464874198191,0.02277716000779595,0.2245291227172267,
 0.28680522436086375,0.9790488765595341,0.8856511896559938,0.9156540624458277,
 0.005445972888428852,0.7247846079885565,0.7716814576856199,
 0.32504269087042736,0.948551299731993,0.8355905615636545,0.7860576110058678,
 0.8398943654312097,0.3425838765250895,0.042845266654715775,
 0.7030620716370432,0.09692763527907833,0.5855310152679928,
 0.16219477352647826,0.5707860279438699,0.6605199162324321,0.5877030948182117,
 0.5741536415951137,0.7274034241342362,0.9223402500965456,0.3488275938286053,
 0.04685540626551654,0.09593141295871577,0.25911746377864975,
 0.772396939306972,0.5590824951544144,0.6002215365192024,0.5944263550561869,
 0.6600404595592068,0.5099692187187306,0.2623459215246987,0.23435899267183202,
 0.28388874778779205,0.7297390436666517,0.04405977854189236,
 0.12662456434691738,0.8795536765710663,0.229841452128268,0.7009866325004237,
 0.8781631078319286,0.6346170979122022,0.9130678727554392,0.9115309493134759,
 0.630741282920846,0.26082374731880364,0.23620040937665132,0.5485265511142061,
 0.52649069094476,0.3305872994659188,0.3199909332860178,0.4982415772265132,
 0.2812043091634071,0.1349686667688289,0.7009307464124561,0.9414593804288112,
 0.8062797448011975,0.45010303978883004,0.5254791704266705,0.605274469278277,
 0.06067866027460722,0.8944973943153908,0.22516650615692557,
 0.16048585777505964,0.9163506059696211,0.4726783083852667,0.715657033839961,
 0.962042575900422,0.9248631189310971,0.05520279425739094,0.938650377802916,
 0.38061408651218087,0.47962118490166183,0.6518276236862905,
 0.07625369843069885,0.614489086478086,0.13456803942133355,
 0.47651336788859644,0.976275427503525,0.4663256654037812,0.7939025586606406,
 0.21168323914510756,0.8337603213111764,0.7998139042086876,
 0.21804443258701878,0.4815891626960028,0.09377265892331021,
 0.2753283142813455,0.6190663597265328,0.9196176295178715,0.403953532234851,
 0.9057177370447936,0.9245843157868169,0.5487431805109291,0.273748208189032,
 0.9669358762369444,0.4498770511285546,0.44774591090613913,0.8517502054400349,
 0.22155759913098394,0.39545950854608014,0.44949185247950496,
 0.3101031512720168,0.11669763049148041,0.7171848856669351,0.935872639665684,
 0.3150434979214246,0.21123920186764233,0.15541843727782134,
 0.9810101321438778,0.8373914445002997,0.039447305169330926,
 0.07091043220500348,0.47583123119006465,0.014915674419791913,
 0.9099513197067058,0.3827050651884416,0.3272862076715297,0.704875604087152,
 0.9375049180954804,0.9837497471482298,0.06625971415630927,
 0.43481406067015227,0.801733258772408,0.998782031495447,0.46568048563880227,
 0.7433762868355243,0.7892736644693948,0.059747255381740194,
 0.43960613106264435,0.4478928872274821,0.2963358592086911,0.599994381962246,
 0.05682124040184122,0.9167917396625437,0.08880842410733358,
 0.5415271282117985,0.5011676068988881,0.026703129422234073,
 0.6857898962250873,0.6364076619624232,0.7190573870784742,0.8018947781648652,
 0.17344288620924908,0.9202960329265192,0.2790452287643189,
 0.28410115400746894,0.6654567284794197,0.7824621630878007,
 0.49031163898574204,0.5706809155361713,0.013974963502735194,
 0.3010731672367707,0.6659834476276774,0.12707908934600032,
 0.04359858189188848,0.41997514990880835,0.6553720096930273,
 0.3340610607026283,0.340366671370432,0.6562154005956888,0.3181297252215627,
 0.3009705498053187,0.3941768710426732,0.35819653744425894,
 0.25672454739554895,0.6259496495548946,0.9919354319834024,
 0.29759395051106485,0.04648742506453496,0.8312427036994695,
 0.8526314499530705,0.3631304146458929,0.08284456271803697,0.7728043884532524,
 0.8429801296014612,0.5291222702697304,0.3443566609590496,0.7841677891818633,
 0.294396686295332,0.7623753003155599,0.8221299217356631,0.788956418440275,
 0.9553000817897278,0.7458352852359406,0.993661573144756,0.7850501268314276,
 0.10355596843048909,0.18365818150280944,0.7758037246451391,
 0.17327164617982016,0.890924243281605,0.19283099925759728,0.952163296003792,
 0.33360274985548977,0.4589097215357484,0.3851137209328275,
 0.16210696730957597,0.5763755585694591,0.0796878742111049,
 0.32568222550104775,0.5368912784347527,0.28937626739299604,
 0.3546953262565192,0.6984944807812411,0.0068325802205042585,
 0.593727754096169,0.38268333475370375,0.3986303051283222,0.5935510847792733,
 0.6677786455990098,0.778909519450365,0.9566359220381391,0.9207276143819838,
 0.6856993121734282,0.1639203839408604,0.6362994932229106,0.3613867377842207,
 0.8896873949301832,0.5115812885007625,0.1667893149694989,0.42463280754537314,
 0.7757037954367269,0.8807756434627397,0.7618166188781997,0.751645056725505,
 0.24222645887266636,0.6144899268402645,0.9340518583250806,0.1349339875229133,
 0.14969503325973976,0.9672605658083193,0.09760213453038169,0.531110059630255,
 0.07316120274216209,0.570752571611204,0.9296375288779061,0.7836627287045199,
 0.2194089497231072,0.6939363234877393,0.1209916423260502,0.9842482775504325,
 0.8723723784751519,0.8607098198696751,0.7845855686677939,
 0.0012748574496884713,0.24488132491635373,0.5269652674408105,
 0.046323364235680065,0.7380667904277394,0.33448996868107084,
 0.7097174377405143,0.4680277464876209,0.0038267319093980845,
 0.09359889248351039,0.6709267330292116,0.7708950914658765,0.5638187214907022,
 0.9621704150710181,0.5970659793994424,0.5881335897675188,0.5924593528318094,
 0.15082728906765264,0.3007511518164836,0.7992750794038375,0.897858125793278,
 0.5778023115314131,0.43526274184659075,0.15916369808559416,
 0.30412608282796183,0.6378185691865577,0.4481189189407627,0.43575379060402,
 0.9455598230750064,0.4575846367937715,0.8052162295476522,0.5175003964336368,
 0.9852870197711878,0.10269263750788116,0.6621435088364439,0.3559477369930444,
 0.30458186461424963,0.4963184741360468,0.20691379323139936,
 0.9428817597667102,0.8571120858546817,0.6229779515947482,
 0.020655851603891273,0.03139699806669616,0.9858338971818214,
 0.9928271642041784,0.16051690642840966,0.8556190420598376,0.5081873861249278,
 0.590397935823471,0.44038488156963196,0.2790986242021224,0.7546618790348664,
 0.31042353358567587,0.3522092461747287,0.4533131072793112,0.7714525877034872,
 0.924367782382259,0.5707374112916085,0.215243311606521,0.4649482743592158,
 0.005453869617934037,0.5788068185041956,0.5771377191569583,
 0.3188859751235419,0.7925429682867429,0.8629909454229829,0.8745502070034338,
 0.6670349689959314,0.9010319196376559,0.8589029962058216,0.9494566176188383,
 0.6664685683327016,0.02225770934691551,0.5727712720012446,0.8343281499998136,
 0.6094342190751233,0.5455998626812072,0.37841882875265553,0.2911877216418982,
 0.3819493691187906,0.1697319387867745,0.39727671004956167,0.9566545202789078,
 0.48130581593637367,0.8815691312823845,0.6398214479617693,
 0.21461966762077866,0.5426157052118183,0.41597308097006613,
 0.9884929234127547,0.3347629761600921,0.5831697062963599,0.7798944502027736,
 0.2973187845698744,0.7588669145303172,0.20772467469974887,0.3780006617760736,
 0.358687301216725,0.4099161113072778,0.7652785211897155,0.3212593568906661,
 0.2315292088516243,0.4040297573725573,0.42660977929436716,0.9259106529639777,
 0.04125028635834593,0.861907273904074,0.5596698557806099,0.23284311161477045,
 0.1616377717238784,0.2931794942050744,0.5742321402001466,0.9117894834302005,
 0.6038836486307454,0.23570897649294165,0.13007967041702206,0.847125248053629,
 0.01330533035888648,0.580668097207595,0.7574591513167631,0.46563893215327334,
 0.7397357901061732,0.21064525524383404,0.9444710989474407,0.4134189081659977,
 0.5222409369876024,0.03866271071717586,0.514070421538125,0.8266754205457126,
 0.23434029968790415,0.9470264265619781,0.6480468144173376,0.5031218494949701,
 0.3631456716762007,0.10755614289710203,0.2803782871594247,0.1957842070195096,
 0.007998905636392606,0.2537863519152731,0.2578146415792786,
 0.41158528813387407,0.8611759906497939,0.1302643643079182,0.7580786228398965,
 0.8328859617016514,0.6579859882425442,0.1586516995145122,0.2854023553374445,
 0.2295889739082756,0.5479132584314307,0.46273535932886156,0.8703004485213983,
 0.7913699230133064,0.08101309985210658,0.8661701939917961,0.9209390634177687,
 0.2700912997604662,0.42029279051440016,0.6186967552491291,
 0.34090594448346656,0.43941665346408343,0.2024618391055074,
 0.11438171167708389,0.8092631616361254,0.13722122716051677,
 0.21630753237063283,0.6936733956139711,0.45583393751427803,
 0.03430246640797652,0.31597447144251534,0.023241537900703046,
 0.15686365328224305,0.7153502254892035,0.7634184027862166,0.5265474052685957,
 0.03589460104253572,0.5039072561371111,0.0699788916071975,
 0.33307856117748447,0.9836787087132425,0.15417005428248376,
 0.20591875628061485,0.3888575462309094,0.16563353897991784,
 0.31711343237131073,0.38656263462225604,0.23527332214032692,
 0.44622374526102426,0.2915242898257282,0.2786678014124738,
 0.09769268593871905,0.2938029548563823,0.4948335075008272,0.6819815524228412,
 0.027053691207391872,0.16746415535921877,0.9302660869204228,
 0.4718998629874174,0.25739317392893024,0.09395162503929977,0.659653993722462,
 0.038651988059858366,0.48441592364162944,0.46999671134220133,
 0.9282656499398648,0.1814519487753855,0.8096033357654585,0.8798554649425041,
 0.9723713993636203,0.011729605786685249,0.42416876006522264,0.63344155189815,
 0.8677564217596202,0.05017883500465059,0.9297889551007862,
 0.23086965400409754,0.7591409766931707,0.6031782083347768,
 0.22027299526681854,0.29017232236397783,0.07267216871289717,
 0.25189883963112103,0.5226533237568437,0.2456055141765534,0.9271039288584513,
 0.5606732993153996,0.08648276229059837,0.43869593455661027,
 0.9855839878795489,0.21033806880668704,0.673787279505542,0.2869047364485853,
 0.479205409271107,0.9510730680807155,0.8207696708720995,0.479040215158402,
 0.42252591099765935,0.18529692840713463,0.6300747585961842,
 0.8166931131499453,0.6205341327482472,0.9266260015253845,0.22932768549514349,
 0.5284554880610086,0.25283126813595824,0.6149640981797093,0.7768165822241662,
 0.8746700495037761,0.7580163106938129,0.40940041514087344,
 0.17376893037919938,0.743168859565604,0.6417461224176532,0.4053826625186716,
 0.13192913384679972,0.4726917081952481,0.3947856956681385,
 0.021530179273188343,0.3455875548060888,0.3454316602028302,
 0.0708524210431778,0.697442875306387,0.9280354269427069,0.07601548376142697,
 0.9150558453623223,0.7698072867335549,0.014928887735783603,
 0.23957209482908903,0.3985453323508422,0.8237590589726512,0.3103367675694333,
 0.6850207628526535,0.24456816598705666,0.7453514357513915,0.9598918034332006,
 0.10041749143109857,0.7487287609680771,0.38296609074598287,
 0.33689047658995297,0.6049039148694327,0.5899537493230881,
 0.008189918256037765,0.28917491634196324,0.023077573686422825,
 0.0029554547888779314,0.5303233995820975,0.5004971587928625,
 0.31797246904229426,0.7640341821212026,0.7402367848868159,0.8102191706366005,
 0.08117411776443961,0.8357925061759008,0.9036029713211546,0.9472307985991939,
 0.14062815121773253,0.3953419241487285,0.7805136867746499,
 0.29902546594825374,0.3861486125916964,0.2749771286319659,0.6516102165290016,
 0.9536813869777698,0.6234712252800354,0.4494839230116603,0.4415496403676976,
 0.8712639494568225,0.4954845295764949,0.3063178304564218,0.08069688117743767,
 0.3577036447836571,0.536363257329566,0.5532129637292722,0.058906708871153235,
 0.5286416084101528,0.6571647717881246,0.4982476106029321,0.18711516172242615,
 0.3657037225510643,0.5782839541689601,0.9717162990291749,0.07795073575853606,
 0.11902940594969014,0.4142974657942151,0.8618691816813623,0.6084620870258213,
 0.9843924451842959,0.7591563996645447,0.4265195585846855,0.8657324204082945,
 0.94612964883396,9.427200505824906e-4,0.6551256805190745,0.6395509385773891,
 0.472002935250843,0.8167599932735495,0.7012026272886039,0.7893794101882797,
 0.6553960948089971,0.7952767526940488,0.3822888387421991,0.7701918883146773,
 0.0023374687928630067,0.4469836314333805,0.6527793345643178,
 0.4855072625869017,0.4534592529328695,0.32266590502693004,0.9611371601238862,
 0.648520434534061,0.8346549203672344,0.8386878065134953,0.5151157926171346,
 0.6118207618384285,0.5526606661991382,0.9704180453443255,0.7413307212309841,
 0.7140829293045109,0.0842590837501932,0.17258558672888302,0.7381625477559768,
 0.9475207169741328,0.9801605029442442,0.5085329966032852,0.8250356275507541,
 0.2949150398590181,0.03937552330498606,0.3554682489807547,
 0.17842133549836015,0.16176286814807572,0.5149829159844639,
 0.5342900535095151,0.46172419772544104,0.06898535496790592,
 0.3531450207463981,0.9082843354171198,0.4990402799221073,0.5840622879981562,
 0.9065406373540531,0.554156136468783,0.12211913988863854,0.6642718245848239,
 0.04645269307000632,0.9479256502889948,0.3747361527022185,0.9970257068898425,
 0.9531254758578545,0.49177127110540986,0.794752723747712,0.42155405476063323,
 0.8526563428522754,0.571391342051955,0.7622771433956067,0.31665751679831944,
 0.08643786036996182,0.05305172126590796,0.2718701541762222,
 0.3203133046089519,0.4288819907308963,0.08242083455972138,0.6853519425167974,
 0.8160595286044948,0.7789592337246887,0.049378317404868644,
 0.3004065771938832,0.9194336676414022,0.9485817012829119,0.28029630366897673,
 0.1495040426806168,0.23320115431352295,0.4596416382421531,
 0.46558625150640165,0.23245496140814437,0.9734379543230035,
 0.8017472678233061,0.9342538614425389,0.8164629201377404,0.15346101363350828,
 0.787859789411847,0.8296877062947946,0.6868748305352114,0.0716232431896392,
 0.35368252036547276,0.3800257343346142,0.513427206472805,0.5546244123396782,
 0.19822093280973818,0.6534647274129669,0.9351581954137203,0.5451824514817325,
 0.15754061722378987,0.3797450037020892,0.16023249682024343,
 0.3683784078240715,0.5408343759465921,0.8056903124788058,0.8672078776629625,
 0.9437117668655107,0.989579140351402,0.8761166807722831,0.8041055756980303,
 0.8221325268429974,0.1931077521787411,0.9536843849233186,0.3645825514563207,
 0.6639068216802163,0.6360617380755565,0.16368132587945905,0.7884313523289589,
 0.18644927821588397,0.39651459707638714,0.10442764196215581,
 0.2355016767663609,0.9838197370722743,0.14390113393432036,0.5293374956969715,
 0.09429118372474099,0.20643430805993235,0.35892187244028106,
 0.9030024982150895,0.29201844783458586,0.6243225501049801,0.3973875662080304,
 0.14956936851258973,0.060116843577394,0.760095131727919,0.8542449305306554,
 0.4442280659559671,0.2567256412349781,0.8556740742098656,0.44619042896799965,
 0.6494428159524644,0.8019280913439684,0.45183939106940374,
 0.013618911478555651,0.2828216286977995,0.2143844768568658,
 0.16263150743363475,0.2873841613089675,0.6212958480131003,0.496316676202577,
 0.5198581610017126,0.36771218953471696,0.4333790396513745,0.2820028570673071,
 0.15830708864285448,0.1907360189324202,0.7822736065882948,0.6273962353602751,
 0.1886841501102925,0.9402248433814581,0.25875792111664164,0.6162131244693322,
 0.8114732451995202,0.5562837164941986,0.007051790127220681,
 0.3442761936372999,0.9103062157987215,0.2768383709281894,0.1377486044281756,
 0.1566360117362957,0.01999092684315562,0.3420781051553786,0.2406431629245961,
 0.9738821926033567,0.3446810746055091,0.8715302303084879,0.2507866637344178,
 0.5734996198400002,0.6256059880733892,0.15161899656536482,0.7340968562039007,
 0.6056861444355376,0.5382695920927997,0.09384570050092012,0.5403593682225862,
 0.5580736417731457,0.9106608784694435,0.3854768477645256,0.150747314394584,
 0.49042382151540664,0.022217550311628087,0.6684365906684715,
 0.607033397488218,0.5327079471124431,0.41814037673485627,0.16093569646215067,
 0.8484076712017575,0.11342727158212462,0.9602416227459779,0.5555081348383777,
 0.9929352784175991,0.8029963399299019,0.4055046868614499,0.34116594909027986,
 0.8069784634838193,0.08737278804270165,0.9041762259383548,0.8286266615810789,
 0.3264247547920932,0.922202693901401,0.7903896619774906,0.7473294734798599,
 0.23911761104462725,0.21989834849368295,0.26673112043775093,
 0.6337916764498406,0.11398296508156203,0.5857501976104779,
 0.19212096446502913,0.8901579532200683,0.10677645896306853,
 0.5071460069456656,0.5886436048617804,0.1412199193528898,0.9156876398440041,
 0.21370807059710706,0.9015374311365031,0.9389348208927693,0.9769082323588996,
 0.6664731714345917,0.595852002170978,0.714928874823544,0.659865084301228,
 0.838531211899641,0.9143574336434035,0.0764522964117017,0.2641833027151905,
 0.7774145748182285,0.44163870793464055,0.05676312747395773,
 0.019787175065237417,0.35255794339232616,0.5736812326255385,
 0.8534677310510784,0.6923795154102483,0.39377678422187323,0.6602467498029464,
 0.6789727442241099,0.90129172651864,0.9541700913566009,0.5335326526279512,
 0.8678711041205955,0.3689080251212731,0.6362842421529448,
 0.015523182543352942,0.40802953995694347,0.31585717069184394,
 0.12389589226125808,0.7366898955500557,0.5499907311684555,0.4684394060594428,
 0.9327277298830335,0.5210168949913582,0.3074919490555176,0.36381998396276716,
 0.834984774437896,0.2421181835375109,0.29552286849584297,0.38518107651726374,
 0.897242057987951,0.36105818358428476,0.21780661147697167,
 0.11347523244416291,0.6331637572771601,0.013584992878349578,
 0.9212091833739183,0.5972810143085704,0.5129122250072505,0.8424025359356119,
 0.025444802619034146,0.34576208811720754,0.003366059609924976,
 0.3950769488079656,0.7669906491852934,0.023054227018046825,
 0.7484890796361787,0.8824059015097152,0.03985572188275699,0.4400717855055343,
 0.5636446848662012,0.2235780093781392,0.8543845868792811,0.6671052090184235,
 0.38278661364153554,0.8127909849573931,0.9708615035727686,0.5105588311802008,
 0.009380710647812651,0.8624069994613842,0.3949201976917358,
 0.18139499922277713,0.06640345233087064,0.7703291797243657,
 0.8404132775374392,0.6443379472792437,0.02514472930960876,0.6877580169395999,
 0.15705753894323493,0.8466889951630862,0.871346682118902,
 0.017340715730860046,0.850946649473754,0.4829887123916793,
 0.06484642083902681,0.17703629454091907,0.5349854703509305,
 0.24128878972480472,0.4235759935961675,0.35732614804425034,0.722724343445432,
 0.9348880965602615,0.7686030814371584,0.7078175406852036,0.6791116609038185,
 0.6846458327455599,0.20730719952938526,0.42311744502735793,
 0.7886478306680019,0.5408504426382316,0.04651099295926009,0.6522258540897085,
 0.5559926584417352,0.38557095075388204,0.7882768443233235,
 0.22329347454496318,0.9707399956387837,0.21249269966187412,
 0.20974559359627476,0.9179844499716519,0.7926332851654243,0.4079849789437321,
 0.359161853889129,0.30159401705299427,0.954892717522827,0.6307755605343677,
 0.8796157417682716,0.7409609229868027,0.9995798937083342,0.7272282141628879,
 0.2512707250126054,0.09300895049982238,0.8742441886353256,0.3457816413811542,
 0.6251027378669524,0.7755671868139075,0.4412995003193919,0.7002614782027066,
 0.3027684162140334,0.5971807379681865,0.5005714659221976,0.16838983408933117,
 0.15746100863757295,0.6317347209350179,0.6236013903099366,0.4425350250943738,
 0.7309283005049703,0.6516162854212179,0.33093635852555026,0.5414539681997821,
 0.13168228863173204,0.0525201980906681,0.6363534393406838,0.4308503965187256,
 0.7760304945753934,0.4257783893130628,0.6262081217305139,0.6228121896896872,
 0.5012342995179055,0.22219286299498286,0.008683520253837784,
 0.8304763077131944,0.135542054743368,0.2997733449299136,0.11007565614874437,
 0.5312472151688659,0.8187369851302007,0.6107540179816436,0.6454205306319587,
 0.3781183320161924,0.25806680455646,0.3965273477002753,0.5146754492236519,
 0.6775140286179535,0.6763006683471825,0.07292909996860542,
 0.30481182875312207,0.3510492335047882,0.9506736439338676,
 0.33456398367558016,0.6909538333988181,0.39078768913808515,
 0.17336355747045895,0.4448215242268585,0.020923342529455824,
 0.3137993081381796,0.8560810004145071,0.9280051662690441,0.7412218817893094,
 0.26377324817987113,0.7818706124178271,0.7084698090789086,0.6489568272841463,
 0.4896064789033786,0.8506062217526551,0.9329365741927472,0.6325462433614437,
 0.9364654933501763,0.024348417798941813,0.922417060935734,0.589519350750681,
 0.5889679614544983,0.5650264007546774,0.7213941205168624,0.10069375966797489,
 0.3861271464432804,0.4784233694633149,0.41312710122818685,0.6549420013770312,
 0.8775944189005441,0.5937157753002177,0.6913975573856519,0.9247515176381335,
 0.8348137820919166,0.49281641285310895,0.944649477741217,0.12369027492116502,
 0.029655467553459602,0.6518431560733363,0.3650302335757163,
 0.5263554623167832,0.6278325718225366,0.04491036013496141,0.4549103843416784,
 0.5688791463214934,0.5463632619337326,0.9519934158827726,0.2864865611289815,
 0.6958900688791452,0.7650835981967569,0.0069828510774725405,
 0.9421940128938708,0.39730202223675626,0.6159579010956372,
 0.37458345393373604,0.32633465654203764,0.7743737633834146,0.998934336304613,
 0.5753189362033055,0.5140682684239285,0.592374440694637,0.825117055816406,
 0.6227945422140155,0.3360873894894717,0.4334166533568511,0.4991156131557102,
 0.6127816158465327,0.46226424972672375,0.8849607194778569,0.9073894843947778,
 0.9313827728152964,0.6763083006998415,0.4195443812828602,
 0.030577874550773543,0.730151143166132,0.7889448059693348,
 0.19803362137959446,0.3847985465587228,0.16021576043153885,0.027236882230117,
 0.7036098972286291,0.46531208879036445,0.1262023871653587,0.6754489202574679,
 0.9821853287764833,0.8898016996132379,0.9352491046999925,0.7364289784496151,
 0.9266790194103516,0.7435646774088551,0.38330214343838775,
 0.41299925927992354,0.6310266227756018,0.8751867618938781,
 0.23144366653983306,0.9027422224938201,0.74658654181929,0.21083902303373792,
 0.850706438124338,0.88864953536942,0.9543736644365667,0.22216923133987765,
 0.12319115018797988,0.22547355343466113,0.8572490544205277,
 0.15436315844015058,0.5377717272397583,0.43088172508901723,
 0.04305038302621156,0.1652079384746301,0.5311871121363252,0.5741148137637705,
 0.4040305275866607,0.2707410447830564,0.1005930246962452,0.9147732850593817,
 0.6109869782424645,0.3823586839868527,0.6879373233549223,0.807058840801798,
 0.482523996638454,0.5671259558510872,0.06243713996855915,0.618313747789776,
 0.26437451377021925,0.4329055765761862,0.46450834247885375,
 0.47987905855428803,0.8479098292947288,0.19834782940260265,
 0.5037035528485748,0.9230776788903806,0.4517350501039983,0.5310906680970359,
 0.6114591922057728,0.47341292566501125,0.7582328915162218,0.6473038371574474,
 0.31884203943983125,0.8035828038423574,0.9620009789670756,0.7210899447383745,
 0.21280651863373123,0.2095515511779107,0.883488732735002,0.7357232873993762,
 0.5256732501481879,0.5837206829050753,0.6812381179034889,0.9763652462894065,
 0.8372090341680734,0.3328114157401403,0.8301844827214822,0.9166138311927937,
 0.8064562078886225,0.3855595093882459,0.9190792251293721,0.28947631684880004,
 0.7274083423963073,0.9734194746347973,0.20906377940989884,0.552505648476896,
 0.28462145074916634,0.41926384566877284,0.235161141771439,0.9222869006433249,
 0.2762855440534493,0.6700761736468384,0.3572001592921077,0.7282643157042362,
 0.3345572726833229,0.9376039776117479,0.030065394178470872,
 0.5863672428707827,0.45692972432789836,0.6781641951275914,0.8810478112696295,
 0.088535813033631,0.2635309773389276,0.6397440605015258,0.39827443810049246,
 0.6320018605116482,0.7509857789366858,0.09622889245037547,0.2564100161054813,
 0.25571131557176985,0.01900756041699503,0.14553934995223006,
 0.8016711779711451,0.4996920545731329,0.26007333806997934,0.7760127056942169,
 0.8036536446757825,0.2931895512460413,0.8204458907134642,0.3862516647763813,
 0.5565764704259701,0.5285014451691779,0.9798875020366224,0.4168804897154941,
 0.6363875039048873,0.9446566549225293,0.5930503781372348,0.8930203370670338,
 0.011829780766659748,0.9541092520201471,0.5561424627549204,
 0.21509082246881706,0.8667705566345201,0.06416032595577559,
 0.9773629847288384,0.9026303676378262,0.4007526788784028,0.7350041805273796,
 0.7829111139441868,0.17346507731819638,0.1650235860862994,0.3086821087449628,
 0.9990927328371304,0.4125944785463156,0.592070951368677,0.4628553121596952,
 0.29841406563621653,0.780185630893901,0.12406921238866997,0.8865786807929695,
 0.5501734169494057,0.9547669098993501,0.8800569080759388,0.8980490631544189,
 0.9563912786605844,0.9479642172824435,0.872459222616461,0.17756132975471484,
 0.6689556531830662,0.8297252448182696,0.5898567394124179,0.19042089771355009,
 0.7295674853944506,0.33385914266321726,0.6923874991465409,0.7183235384596367,
 0.4933758757987101,0.3998517607128257,0.1945536645373951,0.8708154927137295,
 0.27940459082064417,0.05172444956418598,0.4198933895551513,
 0.2577604905211709,0.2821555483212089,0.6036734868175002,0.5067248042513889,
 0.4871591654363885,0.8840736767651599,0.7056677948506359,0.48316585095108955,
 0.7570127940161033,0.7940372680948716,0.747649671078377,0.3213060775599742,
 0.033138422371720155,0.07342917831774154,0.3954926112908871,
 0.1724435135998612,0.5266044407789494,0.21722359299336924,0.7766688931610783,
 0.7939436757862726,0.20530932598642837,0.37835021166196003,
 0.7681181007374367,0.14637296206096606,0.009689728259318353,
 0.7809996098097233,0.4789939555248137,0.5020305141168641,0.6184229292066253,
 0.42057600022355446,0.14709099589645924,0.7387354725821407,
 0.02550250445241864,0.2533254664358766,0.5186488555853038,
 0.06381305132918458,0.37819931480195745,0.5108218075372145,
 0.03714828395297087,0.2435333010690961,0.5495947676311327,0.3050030769136063,
 0.7892560692314551,0.851985045742219,0.43833679836136263,0.7327946968625753,
 0.41732009926449143,0.5516552059965482,0.24186516865189656,
 0.4387884854069364,0.8435944510240017,0.9914806221309556,0.9273665133508016,
 0.0909065741978411,0.6453655865021352,0.15529772417336174,0.6125261983319636,
 0.6523003584437101,0.6955898687498678,0.6269928141167084,0.45099688638458946,
 0.9811301284757303,0.1864130146258609,0.6957261714907813,0.4015582827653402,
 0.3409513967048796,0.15388004331928284,0.4525895540843837,
 0.38575332145007624,0.19578393776218017,0.4287738832934068,
 0.4024936540675246,0.6907432126739015,0.2473373414188601,0.6297533052193468,
 0.824725203376546,0.5841929743107976,0.37094054350402594,0.603552549383584,
 0.41420928050840466,0.4066205138581963,0.06658551468837748,
 0.45327221988887556,0.6888793301041555,0.31742419818180034,
 0.18052479321522152,0.6579293747463953,0.7160797951762028,0.9788306613424763,
 0.006673026592008746,0.07333872716338719,0.6219237587962121,
 0.8271928340113175,0.7252846956286613,0.9588295594446817,0.1886038364728886,
 0.9012078412797487,0.6101391520425984,0.08403716951977103,0.4028434703242191,
 0.7061139644384933,0.2535869048511743,0.4499386268572828,0.6094409742080944,
 0.6667520279367652,0.3832170028800661,0.2972466043088994,0.2491663546706394,
 0.32998606790456664,0.2758992755537053,0.09991156369095666,
 0.3054400484026699,0.5519434381694789,0.2476731158339026,0.8641593341288191,
 0.1472157014846579,0.14288962074462264,0.5414089283060589,0.4684474683347224,
 0.10374302665020219,0.5336925155253484,0.5114725431336428,
 0.08201065217325121,0.5024080162778283,0.9546556707496833,
 0.23660192548075465,0.08653445804531068,0.05194238075129187,
 0.8411746245925544,0.24606862428326237,0.5307081078893738,0.7110687977697886,
 0.5759067803462182,0.5421178176100447,0.9975569602768309,0.8570404791789562,
 0.4557467523505929,0.8402291338067917,0.39769416835814264,
 0.20560218831046617,0.7172212791773606,0.10614227598030157,
 0.015533563426143204,0.44048676850436197,0.8799348304605937,
 0.08083276401888628,0.6417636374862543,0.06388722863019247,0.602591404657072,
 0.5617156640004985,0.21840370459143765,0.29237489945022455,
 0.5764412586220731,0.3405358622777608,0.07085737539318226,0.7047105352360099,
 0.5970508352704033,0.053877242025617944,0.7503505123904388,
 0.44412935824706623,0.8063971648363623,0.8158376027009279,0.2352341672561321,
 0.8763919982573749,0.563648252811325,0.8634389734595086,0.9713687740178182,
 0.8455678073005083,0.3554691936971994,0.46811884872726905,0.636213482249405,
 0.8396174550693453,0.44232643548096084,0.2776805772758254,0.3703769277442541,
 0.0884432431926484,0.8974444644486486,0.12050233942503996,0.6811817171001687,
 0.9840066323965341,0.4137332867116549,0.9489706302159503,0.7297698446115892,
 0.11475284336895608,0.8197877740138927,0.2119882920554832,
 0.10413354304406586,0.5717654122224409,0.012322989928237327,
 0.062065412481177695,0.08123296151106729,0.22880171940780258,
 0.969041465263169,0.5072935210132544,0.46522822444213263,0.45666770948209745,
 0.8280756635019833,0.9100002713461168,0.3301904782893631,0.36625867534106726,
 0.5193403080722199,0.4963577773991401,0.45077182757236123,0.9971983802556057,
 0.5706523509345852,0.23147326779626431,0.06734521158135709,
 0.8794150513866329,0.9316871878809805,0.5766569564930312,0.6364027904110366,
 0.19987731604871672,0.441478633930378,0.44164184864493905,
 0.03139247054578931,0.8735702730988756,0.814901433034263,0.18908523435135138,
 0.18457223847372506,0.8300971558185465,0.6402803873162812,0.6391819202999994,
 0.3570934544121016,0.3396505970129121,0.796754211099459,0.668685846593897,
 0.0949226071689726,0.7644897265299859,0.08399859640317398,
 0.12535948667308117,0.5792115860473974,0.5389229399240977,0.3771302154242455,
 0.028109841083733822,0.7929086717481768,0.12535750921452682,
 0.9984929941143319,0.9295517299522839,0.0378895099379577,0.28805576164113333,
 0.5159825115249131,0.019442146578793063,0.17739371923822334,
 0.5894796233815979,0.3583561146556101,0.7707225651717649,0.811866335543032,
 0.7941991588587569,0.9991175966892271,0.8892234066753898,0.8193122011521007,
 0.3841106959673788,0.6430052918543584,0.3350534277009011,0.6778564498026457,
 0.6116931125352576,0.505543750303125,0.9701717093454242,0.2726277092270223,
 0.7912681345508392,0.3966638393886648,0.8322536118428977,0.1993479334527154,
 0.038965982176308755,0.6161943998281844,0.7751549102932477,
 0.9610693594712665,0.1486255739992859,0.1204666476384808,0.4439417372895744,
 0.7905372145444235,0.7119697217123003,0.44506803727193656,0.6196171885735675,
 0.907756623557498,0.2701360329653171,0.13033153515879548,0.6082106733159287,
 0.981412635615315,0.4799812007267249,0.6493930269592315,0.8942230024422213,
 0.9136693487430197,0.3774873643827026,0.03573396119073713,0.7575592678112256,
 0.5125365641783225,0.6446408806042103,0.5521235473472247,0.1536816851675985,
 0.8913532858618749,0.047287507095195025,0.33592609286152186,
 0.7669938748089243,0.3014333292955079,0.9764319767680375,0.7858659499095705,
 0.6147183790765409,0.05096623487102525,0.15209933568022704,0.905811964258191,
 0.8637113469123955,0.7579682228769717,0.2945300514589002,0.02914629000968527,
 0.08312377091526646,0.1224290249533132,0.8071589322563829,0.4412800081718534,
 0.9366765497102251,0.59693677429354,0.5034568054367992,0.48921491244982473,
 0.949552929934498,0.9453402945890054,0.20632109944575872,0.8448367815054417,
 0.6084063770162635,0.6830486902444584,0.21040503326641424,0.1948746925355913,
 0.9068542537360217,0.8573739109660197,0.37041554988318326,0.9066094814095201,
 0.21100677289524772,0.6349362959251906,0.39358860115353067,
 0.8955598869188675,0.3097138512607126,0.820944697715826,0.14802675215414107,
 0.07030537969760675,0.36878499791078756,0.34898352306929203,
 0.3061396397542797,0.9744157269101971,0.9258407901110429,0.7624059319417947,
 0.23640690202250658,0.26180114721685743,0.7745000887630148,
 0.20282215898128386,0.4365068711247312,0.01837811941967027,
 0.8223170088604224,0.4172225210245528,0.514271947564983,6.519629482009126e-4,
 0.47157057084299925,0.4564627728534999,0.704669840219316,0.6580133680525744,
 0.9949402965850603,0.492700622147936,0.24803311809178075,0.1772115887624396,
 0.23942952375225446,0.9516509863967968,0.6471222770411895,0.6397163819125515,
 0.9941606960979573,0.7811589272162124,0.5693130066794799,0.12729132768363915,
 0.7744229366625557,0.5703674411477264,0.7597797808289812,0.3316846837541815,
 0.8375843655416375,0.13172658430493533,0.3682837898574317,0.7525350584110787,
 0.04961081418471058,0.9874670367250314,0.3362470459266651,0.8946996663936205,
 0.6737333054406656,0.6483058194323005,0.9124251374100085,0.5391864765706107,
 0.1338276795407738,0.290256775172388,0.37864319405444247,0.6257312482441546,
 0.6196585812830685,0.23342310879521755,0.7054130647031183,0.9248343477109555,
 0.7490519537407615,0.2749591760703006,0.1963051548600413,0.7218608888044945,
 0.8784388678659539,0.7140501462418191,0.23763046396830545,0.2825589538492641,
 0.8005620816239211,0.5578615878608395,0.565792439482854,0.9436603767208498,
 0.9171166045565723,0.31314084069368076,0.7378904169274307,
 0.041673771230086976,0.9373298256167566,0.6433563396365232,
 0.7144770727308484,0.12324052604557023,0.20247925424098057,
 0.5308137838315994,0.556253514610855,0.5979475987328186,0.43144251436529735,
 0.49967566036179845,0.29960665253042806,0.8699687282714839,
 0.5316175873480193,0.7803974298096202,0.32763759782670265,0.4076856846205382,
 0.857867624656828,0.7051390054133374,0.7411645828349935,0.42604752836461435,
 0.5023117537494273,0.4052409107292817,0.869977522469391,0.3451550842324438,
 0.2252293736166049,0.5305319757585245,0.8988497670152292,0.9112285817716577,
 0.8697330284936992,0.5532341968547868,0.7733540747067086,0.5228175629417259,
 0.07705965121789304,0.4276853012816877,0.9465258637208784,
 0.051463813077123666,0.34440740219532395,0.7332121725651547,
 0.12960395958314352,0.14802005244853977,0.0840851855396455,
 0.03506037841709597,0.8898528229333911,0.5865983887063595,0.8444778845550819,
 0.9047214354158424,0.8571434584573329,0.620401255162528,0.7168460661986709,
 0.49222091718008154,0.6497119269084779,0.8217739531444177,0.4354934615350178,
 0.7867459693407285,0.8202987080472892,0.43395335435083116,
 0.15695497691287552,0.9831572439799853,0.6506121370512412,
 0.15119082699636888,0.4438332302519358,0.2945926589524205,0.3114220589648168,
 0.6780359324707038,0.5199178765570499,0.7901959997187618,
 0.007412246787385057,0.6467022961198272,0.762520226071364,
 0.10565210075120712,0.5304224707086904,0.1256891262647314,0.9650825891902851,
 0.8984228346432658,0.7350780914794787,0.023921683972589203,
 0.07649037188163466,0.1282111966864743,0.09098791468971468,
 0.9067718659765828,0.6163953199194302,0.7787639055618365,0.7772080284214236,
 0.35902020827884573,0.557811841600677,0.2286298096435997,0.1527617078914878,
 0.4365621607668363,0.5507827931150424,0.979602591738553,0.18139793127026127,
 0.26674835486380055,0.3392817428905718,0.4691408327170724,0.8711450276796214,
 0.9570790444503088,0.5191073898655818,0.018422307406925853,
 0.05130320384635112,0.9773388994507223,0.9216812029912065,
 0.27526193006549404,0.680931583706368,0.6973881981996355,0.6487218728584845,
 0.7366175003500851,0.7889284715804045,0.44125777498338925,0.4788473984847279,
 0.9447523115372106,0.16890670357450954,0.5148554134543537,
 0.27953554284884496,0.0396897784835708,0.3494614113837653,0.8889461124935366,
 0.7520214556179947,0.9159877693339997,0.6708361983132709,0.1905246534967383,
 0.2846403021542191,0.9525307623155844,0.5797368442583999,0.2529282159135282,
 0.41603316535073864,0.2846416264458601,0.051761529035924236,
 0.28570800713428746,0.6877508276425223,0.004767900820974114,
 0.09774021567546332,0.505350256462363,0.7192379445203388,0.07601460655500769,
 0.1564620532899056,0.5214117183180489,0.9605824673333604,0.4189175458204085,
 0.19217940723709748,0.3904143162342325,0.09375793958523959,
 0.4785631510538184,0.2166304901327032,0.909492562730686,0.3751063475438059,
 0.6354118324182283,0.6242551468796244,0.9557666406498546,0.7077517915585974,
 0.2677803957929229,0.7718519718958635,0.9995182845165413,0.562111639416354,
 0.5884101584594297,0.10662680302962457,0.16901056064361608,
 0.7702282535532454,0.33896327746476285,0.8108083112379543,0.3216847517619803,
 0.11975111807560768,0.9371972128362221,0.37495300648002783,
 0.7278724505561404,0.041469762422906786,0.8900174562798507,
 0.7400292581096612,0.33676667242240843,0.6672413571646557,0.9823534994091003,
 0.9180053156904787,0.028529539158555495,0.6286961319698199,
 0.9370207109247167,0.28442097102029407,0.35270867279133156,
 0.15379568216521122,0.359239322565851,0.27730458172463357,0.9515257336729546,
 0.28518152819289455,0.7171607690276761,0.18017250973980603,
 0.20055146994308992,0.9706123491584446,0.9993959146130326,
 0.36588494770667845,0.16188333294619284,0.6063929035403057,
 0.9504519976354193,0.5948327141327321,0.5573353434153802,0.8791381478801387,
 0.8105680356049677,0.5995219277886594,0.21700807773623054,0.60353572694047,
 0.6555575134196456,0.05201718682417433,0.7199471116281686,0.7553940739507704,
 0.4695230816411756,0.05553985497907066,0.16849797208018147,
 0.0621253787982079,0.2986193156064313,0.35629606139901326,0.3724873296972012,
 0.5604325865642817,0.3382883885231829,0.3246485861038917,0.392152434290174,
 0.40007874983420544,0.7161262344465599,0.5534128719874432,0.548856116789453,
 0.49511143857325046,0.07715947791223632,0.8403982849623695,
 0.09141227548959607,0.21553821160210707,0.7877574287866037,
 0.34732009331837943,0.21468264542960735,0.09803834232535413,
 0.8288281934718575,0.3882685839118821,0.5028980923400844,0.6630559725672877,
 0.4017797984672711,0.7005201914383602,0.9386751775322472,0.17090355075217611,
 0.985880867528449,0.005161048774906751,0.33259607900906085,
 0.7493205242755171,0.626442841606214,0.4814571811536963,0.2049091352706297,
 0.6846446568591262,0.6680547297575505,0.8644377857852239,0.5841261120971818,
 0.4665213188580204,0.19796814964346143,0.9212643752447458,
 0.27792012723593695,0.7826933098595498,0.6417885138857963,0.998878475743411,
 0.8093717683004579,0.4173569009833946,0.7568569347105586,0.6527011572777351,
 0.6023982519705628,0.13398510770229077,0.7569411807211283,
 0.21020935376405803,0.24376867700068083,0.10447733362753275,
 0.2723686454720673,0.07266852275810987,0.6306794443069128,
 0.009199922802783145,0.4027610229479961,0.9892972282304184,
 0.7202878694204138,0.6988539833735491,0.529338434750873,0.3731479573222254,
 0.804602365621541,0.840408320853633,0.6167660375069561,0.04545431320731519,
 0.2973473230924988,0.053433749743659886,0.2299047645496407,
 0.48665620207422844,0.05747405469616862,0.968545738668073,0.7085378490572978,
 0.650164299048741,0.2934419553281319,0.3610030541539394,0.11038675312196888,
 0.1348970369119279,0.33352396499914694,0.5755087377193387,0.9573435682242509,
 0.7220526651460293,0.43346354536305776,0.7863852189521718,0.4998536545848359,
 0.04406494095523095,0.8740804656012231,0.5181391909257616,0.5508584349515141,
 0.5014088628556074,0.7172153657792504,0.9664052559600982,0.29913466098721386,
 0.38160429938606155,0.13044586934814206,0.7917506672575029,
 0.6138898265440341,0.03809818016829225,0.22633702395047195,
 0.6281460981168798,0.7665211496984284,0.13533180622064322,0.695317406452586,
 0.6767800358170135,0.6321002889360694,0.32313495853412766,0.5589535671649253,
 0.4895663469108079,0.5130201009323355,0.038789669724639375,
 0.8604476155902703,0.5500493397719598,0.028137753378625563,
 0.9321174701984339,0.08736649124958029,0.08168838502167419,
 0.4532381049153228,0.4138836901981704,0.6852602057326824,0.063693761422147,
 0.1858501400813113,0.5006911107106488,0.8342760387691993,0.324813544952813,
 0.2693860598863971,0.002210325096138721,0.1754421499641381,
 0.6897388434327043,0.9192480492253465,0.8926847006591142,0.11732806335087254,
 0.9258622428610931,0.1375782978155924,0.8843269711665542,0.7864221954532228,
 0.717416555617288,0.1570391068009397,0.1024574172767867,0.08787601302003101,
 0.5621619656356076,0.9374378284907282,0.8739110602990925,0.5142134977117361,
 0.23378099793775164,0.9093272810986149,0.18675070825994822,
 0.2644724670287716,0.30354928342389376,0.5811813545885863,0.8946383032837901,
 0.43859370104968876,0.5121983471539581,0.32992623248145814,
 0.1163990007066561,0.7189784055748052,0.7454190097872081,0.6672380863370165,
 0.7270951348488848,0.7447104039684402,0.025769692568637703,
 0.6859727408526387,0.045604477990150216,0.4950338693236592,
 7.674575748676249e-4,0.8474876145092536,0.7066493712407845,0.440371897205886,
 0.38803349300869705,0.07763298983333544,0.897247727319588,0.5722367015688473,
 0.7568703840018962,0.39904208689861553,0.4849500601380048,0.5446423460751711,
 0.6183024326302413,0.2687396374110791,0.5459805225887664,0.8886819858827327,
 0.7712408528912725,0.6237282461927446,0.3507688719297739,0.4351047131968129,
 0.8368744981965694,0.8523543978200162,0.07028937784840572,0.8044720102581838,
 0.3785081114982509,0.6160079413345998,0.6137735930570256,0.9113438006678029,
 0.17416722172513088,0.9213492338794687,0.5008362157266305,0.6524109490938654,
 0.6108024430983814,0.7294787906362211,0.3215717896769493,0.6909169341416687,
 0.29897831730943136,0.050253337116482055,0.4708247879536738,
 0.8238467429991941,0.8078339008628223,0.6855036055107362,0.9699238138313473,
 0.06842345144502238,0.6527436556564317,0.449524341543852,0.6413839877530749,
 0.3568296034381625,0.6953441801836153,0.8900392566501036,0.2535506099875787,
 0.7675060926363305,0.180365250838315,0.6231633362879537,0.242131810904742,
 0.6835535326576269,0.5391272996682195,0.8905763395693295,0.5138550803026164,
 0.31662849372213,0.06351152883613076,0.23508303782553808,0.592441056779067,
 0.004579743175772144,0.39856857683688496,0.8816174444726936,
 0.46197283662283506,0.22102525686300845,0.6045244355660302,
 0.6272507088256285,0.5849884514562401,0.9484799914337609,0.5324861698830193,
 0.38891250601959915,0.7440930776502024,0.654459533869287,0.7617538447728831,
 0.5471624631765682,0.6056060177670132,0.36296248372689044,0.6995722823041598,
 0.8014898223029502,0.49912620094982063,0.7065329438242192,
 0.37614718865687646,0.7788372146189928,0.42006630257987576,
 0.2774192769036564,0.19697512868092126,0.5324463804910907,0.7523415394708955,
 0.9391551947090087,0.4048329657180779,0.7853146066156083,0.06709039448471221,
 0.7328040705281818,0.3070852982688176,0.9374581032375441,0.00148380071987686,
 0.0901755857807387,0.6985200093515147,0.4407731177766938,0.5453511916121965,
 0.49405694417984436,0.39383842203075314,0.8097418934295484,
 0.05556825896388928,0.8287371809427929,0.6400625105614979,0.860135332742759,
 0.41008980187107136,0.08461403994001448,0.9514790207123742,
 0.5084450558083206,0.41998593320481126,0.5938707990020459,
 0.04594601640679097,0.3464445688495871,0.09678142255299571,
 0.7616124142424006,0.9878498269794618,0.4997470765308448,0.6795397630368362,
 0.010327284616819865,0.1197875295707942,0.6964433725981235,
 0.42136856849930027,0.7706969939300632,0.6885289687807307,0.8839500209409155,
 0.655865564131276,0.6556538761872729,0.22309068315677894,0.247302534265601,
 0.1750867881090341,0.5555052210896223,0.5594493328475705,0.8328475370436472,
 0.34733674337448894,0.1244460900203408,0.2931123967948126,
 0.06472703512687317,0.9699683023021727,0.1454582158709723,
 0.11144435654637908,0.9445700859621221,0.7281057537175823,0.9425935047493027,
 0.11706859847135898,0.147896336007719,0.035320121112493386,0.829516393675302,
 0.01471902402578229,0.24820117233071337,0.3327849685148456,
 0.18795640755718634,0.6474622645258971,0.654567999092632,0.8544890557917534,
 0.21557336172152464,0.21172099389863497,0.13027891645104683,
 0.18080307817400276,0.5594463329824009,0.5705327905113969,0.3399934093390111,
 0.032782862990372474,0.8177236429129725,0.6747698597565455,
 0.4831281496824352,0.41489391011737253,0.7071426610883544,0.6823228511585981,
 0.7333937169970346,0.42368059743170183,0.9484716680740758,0.151583731801739,
 0.8283637871294562,0.7474869291320642,0.038189868104918645,
 0.9809704442694407,0.0707257046773071,0.2399130857244951,0.9209327900663388,
 0.2581745713441841,0.25628110422956296,0.8096532591579535,0.9977738247480624,
 0.05320906273051729,0.34886213516601616,0.4049248460942537,
 0.9472606896676457,0.5337070047298478,0.6145620589358493,
 0.012484783888649265,0.255059667823449,0.14654668043215868,
 0.08346056991435902,0.5057508072288635,0.13653782268558323,0.505271235908187,
 0.06469516900365291,0.24944999212423724,0.6853082428259935,
 0.8491367565300915,0.3116051977320876,0.5222276798025536,0.5802013652160831,
 0.5767953314083325,0.1746819689456549,0.05057618599842395,0.5588037066741607,
 0.2285795053565539,0.8478074695163555,0.2832773544276599,0.8431407821152747,
 0.7567936993782398,0.02845149146946202,0.5228913316143188,0.8428713741556094,
 0.32607223399785223,0.1584421325836073,0.6125026461382261,0.6732659059540136,
 0.04700063143800093,0.5323422922924305,0.1971761132774692,
 0.15156962602811863,0.3914662145604544,0.8812830454490622,0.727013281742332,
 0.8298618167911725,0.580890695610882,0.48684984768663775,0.972830528579809,
 0.03867342890332781,0.6114322754991863,0.9518184517473256,0.9453524729695439,
 0.5594795042324544,0.09834083746239108,0.2616278771296423,0.5200654924295665,
 0.11116770777199791,0.796780479662696,0.1344690401394465,0.4858934747970456,
 0.7960526309616625,0.1925622798796197,0.8536938323015191,0.7535096857235505,
 0.08394878077802337,0.975359468959246,0.33164597499865356,0.9198325128506217,
 0.1795110826822477,0.7054418361761967,0.2850627409269585,0.8041215220053063,
 0.5140404920242325,0.32911159439827453,0.2490373065625765,
 0.17981003314403998,0.14034471506470514,0.28569423921846704,
 0.3112190898859808,0.1301452865984396,0.8803849602769459,0.7552394234894754,
 0.8816715696638076,0.20634674258434538,0.11998767785861009,
 0.0491204723664741,0.4692034913630917,0.10435484649142224,
 0.46838724685553856,0.2625371075621732,0.5679350914481811,
 0.16191008504516124,0.5735294806063533,0.5910213156735716,
 0.07741551712418726,0.44934445371108644,0.7172531092984451,
 0.8311471196582635,0.5541549254228193,0.6808050259290764,0.7601732381871157,
 0.8642563039791259,0.5248788631652623,0.15661575144892903,
 0.06633042328665717,0.4778978026796723,0.5511946494062615,
 0.29677678154588816,0.12217874711874088,0.055095292151476016,
 0.7596558127306242,0.21654431277776431,0.18338225785437579,
 0.7683702984619633,0.6206458535670911,0.9956444577921272,
 0.047133512553669044,0.4777496608228988,0.31156112479982456,
 0.2918431197175846,0.9395490882868561,0.7828154577772273,0.8568951237109383,
 0.8977621965374342,0.6002523370692894,0.25587540702223366,0.5671056117943872,
 0.06497411881919335,0.097206635462877,0.11699201948860183,0.5464676491106376,
 0.6806083734935708,0.6413405536899566,0.6920955633544281,0.41407099922689705,
 0.06812017082031274,0.07364642341776806,0.5683434943461729,
 0.9954657573924373,0.768987400875519,0.95837626812317,0.28726974253358317,
 0.5116056931470924,0.3534516109305891,0.04670084891068338,0.4899335029073184,
 0.2642081829365377,0.4449420671285276,0.8986033645880764,0.8959174537405966,
 0.3765688951548304,0.31586639208678113,0.7027129791429777,
 0.38757507427126225,0.9400363037720556,0.824881890225652,0.29040506732869864,
 0.10867328555535227,0.9617000954767564,0.3308569898449011,0.5592160179808688,
 0.5269492257011044,0.419967687294745,0.36950524803328166,0.25400459472041426,
 0.3817929738524335,0.4785188825245327,0.4659047488714009,0.1079530306621711,
 0.2745937816042867,0.3614788529682905,0.47049178656742474,
 0.08858964178626605,0.4432320467518155,0.046382080381001556,
 0.2765130884118494,0.7372309333063,0.22664673065017682,0.6074866308540624,
 0.9131216580610759,0.2295447609416813,0.2872794165908822,0.6705762956082395,
 0.5467820530212614,0.767351927897379,0.14712779225284633,0.3343885961490557,
 0.026254503014783714,0.4563133509417151,0.38927671508983064,
 0.7500711740705948,0.5135464940623774,0.28508397419069864,
 0.041514971120799604,0.4073892635051336,0.00982479474279263,
 0.8433140735158399,0.41960231201806975,0.21766145287092997,
 0.9910934554200378,0.7579950907190809,0.1171879098134827,0.4769729312855093,
 0.30114116891552445,0.6100079716450094,0.34416534672823196,
 0.6384447378507316,0.022709155505571288,0.580687841672512,
 0.11981234902803584,0.10436040453108997,0.5318836240905946,
 0.08454748321611139,0.6417021906814528,0.3702570315185185,0.6156864047046845,
 0.5323833913303294,0.7978450628340659,0.7613293862637706,0.2950021231349186,
 0.46164549491244855,0.23093449133465904,0.08680108662412733,
 0.3936973974126311,0.417891662910841,0.6442651771543725,0.293489242913769,
 0.12036709374510246,0.06855046371353624,0.4943770721743552,
 0.5859368754825993,0.8585283591812707,0.8155530204149842,0.6154806527679867,
 0.7822300501928514,0.948457498931964,0.5991170409604585,0.12774639958399714,
 0.13221214047243457,0.8021767292030149,0.5600582519759716,0.6984319928468774,
 0.6749589534680418,0.024027970078302707,0.3360235094455054,
 0.9398749391377839,0.8058500080496741,0.5928356956235719,0.8747793762000908,
 0.55493137863009,0.5210346454934964,0.5963286776590955,0.9592880793586507,
 0.10024839471319158,0.7361167432143686,0.7899454888562674,0.9827930605747893,
 0.20753990383736298,0.5831843546909796,0.9055668297889563,
 0.018906558351242664,0.6761383689487273,0.3001702445036878,
 0.7689730338195687,0.7556910207802766,0.4203362857938624,0.34333839034830793,
 0.24368982876432532,0.5282054294683474,0.9477283111500301,0.5600447192884365,
 0.9760165459939367,0.9106441957842456,0.0530541592517878,0.2844386968189536,
 0.5117362129708816,0.9573040504849921,0.08857454239146123,0.6534266396638456,
 0.33660362102636365,0.7541158776065059,0.38124901931287103,
 0.8740483519349921,0.6684123868104368,0.9806481665187544,0.8237780044561709,
 0.6153360902933097,0.3412064603151165,0.17580873459542157,
 0.22575911091099865,0.16606575625116538,0.5221042418961156,
 0.7298122611947931,0.2299127761902544,0.5726588681258216,0.5056584939517133,
 0.29500914056756466,0.651862260349241,0.9907513305392989,0.26364982610360554,
 0.6244996708240496,0.4913726850384501,0.8445140200343595,0.671077270793687,
 0.09128955105843661,0.26934889493237824,0.6455072426458942,
 0.4935580983632023,0.6476688166667185,0.2963774014951197,0.7347738910871536,
 0.2964819480611013,0.4523247497745869,0.6582301416300418,0.5700655080957694,
 0.8678560916686544,0.8956219678886876,0.10054892791411185,0.3760008369448079,
 0.48781814303885607,0.43522774393625585,0.6103487916646692,
 0.32774927758679995,0.7850573962062812,0.6635442815042012,
 0.03231385018880406,0.15575563881983578,0.07950919999792905,
 0.425530692509831,0.7364501171530557,0.0024246139579766623,
 0.08210764952864746,0.13986910469005087,0.1987035673519678,
 0.7762702522490403,0.1690366803444796,0.018914712562534186,
 0.34506623050927165,0.2642203170084214,0.9515185206121312,0.8886790690596673,
 0.6704634305292314,0.9202981574267589,0.04264676745313756,
 0.38244028364962424,0.5252737844132859,0.33031594107090334,0.254290412922191,
 0.2469242665252246,0.6226584966898898,0.6938682805412208,0.8898918758030153,
 0.5245064568699187,0.13256440862765662,0.23687330445301324,
 0.6466077819071736,0.45333923756746164,0.21552810254753485,
 0.7287143426100005,0.6737071003958434,0.4895826914856285,0.3423180953097026,
 0.09297229268832652,0.5846021230651528,0.3357381330463992,0.632012460250698,
 0.6342260880666017,0.5097502025941323,0.6804679968955023,0.7781744031422865,
 0.09836735687883313,0.005813684367482752,0.4072494261203342,
 0.8078358407685168,0.7819126306317048,0.3065250814419971,0.6793591321611213,
 0.9250719883001354,0.9383696711339448,0.3915339579775281,0.46016720622183527,
 0.9526787447867658,0.3241334997930415,0.7338432055292574,0.15454999773263922,
 0.07428502297326811,0.9440736314964656,0.5076402594428107,0.4482082787899718,
 0.9275156061014922,0.3911483259908144,0.5650474451395546,
 0.025234799309739264,0.4439492451177187,0.9447637033685923,
 4.062636945678655e-4,0.10319566057529339,0.7585198942990532,
 0.9184729021438933,0.2968995766814009,0.6541148686841712,0.06772914311471112,
 0.776438778286909,0.6187138241570732,0.7620074277441918,0.07195835823119034,
 0.09713217685746811,0.9152044971500902,0.06416091391763079,
 0.11922966758127673,0.2097477704017192,0.7700524151900906,0.9898282396736175,
 0.6104077760081834,0.5155343523882584,0.09669159805744487,
 0.01995567385458963,0.3408536077015183,0.7685663451655353,0.6131047037218387,
 0.5810739164804598,0.941984111892858,0.08256553473020933,0.06205481317687722,
 0.929766242865482,0.020425432452591763,0.8041297625651147,
 0.23884756929891682,0.7609750282243586,0.7238056468452013,0.4531657236074004,
 0.7327551327605191,0.13645550020631547,0.43140887860970145,
 0.7007961941284234,0.8812527465431002,0.2944369530146974,
 0.044102373746613854,0.04438975302800541,0.2939140353387558,
 0.6024214931873989,0.3803625626413275,0.16522143967131386,0.280535709348129,
 0.6666806617380399,0.5997708441121887,0.5837497714405209,0.23140254627053725,
 0.2503509446299804,0.46458013014945065,0.41634097366847767,0.905906496710944,
 0.40110623310214577,0.9353273906168431,0.6821204667098271,0.5655112794639459,
 0.5163654515176084,0.3640940732314364,0.37814178363433903,
 0.018218097097458585,0.7992644499578366,0.11683221216632722,
 0.9128407704982697,0.360886504174478,0.7878735606284152,0.05570484678304943,
 0.6705437181037968,0.2698771532909847,0.29722065241650886,
 0.34942660329415187,0.9541455470765156,0.6610263858338349,0.5794381837942518,
 0.16764142350389732,0.3226405298820221,0.6071014433895018,0.8773439161703998,
 0.09750604742961255,0.28078071197872556,0.80579198749957,0.3774332493186191,
 0.6509340730634193,0.5539032639499574,0.676392826291568,0.8696748099985154,
 0.5118754857985717,0.5056768630038007,0.08699725349726228,
 0.49273271413735953,0.8507260217721656,0.9956550284117595,
 0.13023034268895295,0.07216875377929699,0.41707586971708177,
 0.7542000548773211,0.3616329689456499,0.6665928437123602,0.7524676413422526,
 0.554066998415522,0.42427091473485057,0.6834113373333366,0.9053139262431146,
 0.7356609808079458,0.43318630331302843,0.1480826091091545,0.8609353214788042,
 0.26377129629789586,0.0862142319531719,0.5472474627283742,0.7434260746213461,
 0.6205286597752195,0.31634201456540556,0.3971703294817407,0.6108149716298216,
 0.20741120967262627,0.06865612152793965,0.6801503668247284,
 0.8662214791662916,0.19056103786090461,0.8271380018720447,0.9452341549969263,
 0.5498506764089474,0.691149157805983,0.749534072458418,0.6191374897429867,
 0.9921638642243975,0.16181692142138504,0.6948629303416233,0.6195067550313403,
 0.255772857466933,0.4533825422809228,0.9748138667706913,0.8438096736801596,
 0.11821337195198056,0.7976842047506922,0.4632859819519668,0.4166227657943371,
 0.9196348892271196,0.17433580072105226,0.46632579844641675,
 0.5328105708486952,0.11350359608618998,0.7352512250803201,0.5009085494187935,
 0.31009440352958506,0.5597429228575541,0.025708698887130566,
 0.4261155109887713,0.22892405444669817,0.545043839144738,0.8785278725277635,
 0.7127471341341559,0.7390811657256375,0.7103776412351894,0.8657056658971001,
 0.8988693178638729,0.779127388186533,0.3268819484844716,0.9431928429046597,
 0.20345698859795702,0.6195551435849477,0.11226239538599792,
 0.6004561285131952,0.0627390053422271,0.4134926026771619,0.5970062902933967,
 0.17669312614570098,0.8605662347660674,0.034104037234756746,
 0.038757103318549424,0.31833800168388193,0.2123760372791763,
 0.8029695342155516,0.6340245942225138,0.07728897904641074,0.7425351454801347,
 0.5667227685224089,0.6742931951502039,0.39879621152006817,0.5302371401024074,
 0.430157798218318,0.9313761807072137,0.02203377971010656,0.36262843155148927,
 0.6851017436062942,0.3303747072084895,0.3120049677359037,0.9719074608837499,
 0.25902939995161545,0.19807344734069166,0.965564405357698,0.7661820430882863,
 0.016354484830985228,0.3129597181344489,0.8185145999867004,
 0.6910801020781409,0.5095791225163953,0.8808107504050564,0.08076259938979102,
 0.05706816678689719,0.7070455209802481,0.046235389837287366,
 0.3019709327446912,0.593577112258641,0.6354335115334058,0.9672576485178105,
 0.851243174683023,0.2276126588360966,0.07463433598274993,0.9251244812727126,
 0.29021244608424124,0.027977026057610876,0.2994130540040044,
 0.309834327347282,0.19389044236056374,0.8046791975763847,0.982522352992608,
 0.33364525605352524,0.2613132630435505,0.3030248636232298,
 0.14636130865971342,0.28823893052942395,0.7837248976425091,
 0.5635716703976492,0.4566329938493302,0.9372554655566606,0.21823061426444967,
 0.9668184412904782,0.7043972804971016,0.2125173975500152,0.7068536023697853,
 0.33554098963748646,0.9068466788062148,0.8637045313497627,
 0.36967320140936044,0.2940403618340194,0.78625641852601,0.24742577400754584,
 0.06503409496858892,0.2119115024228182,0.516843548305876,0.10018054112620689,
 0.16747420124630763,0.20485077578256705,0.9505936825894118,
 0.7636265566400208,0.039576381918312586,0.5290777806993047,
 0.9999960431832529,0.033308105957945644,0.3997352794854854,
 0.7119705539325141,0.839328413982768,0.41065845885725116,0.8263302283226702,
 0.5907710599451094,0.7771013917897717,0.820302862846791,0.07674810236387997,
 0.4454379886712936,0.5858725642040759,0.9570114227467036,0.19765482664249723,
 0.36364909303554815,0.6361840851883195,0.8591773634179232,
 0.12078597542715142,0.19288119400795745,0.966181234706209,0.915226805063118,
 0.7378392727594046,0.4401759416615173,0.7575260790509287,0.8110261399181207,
 0.04507812453118376,0.0816570880696581,0.39821904268931185,
 0.10821700992303773,0.917407696854611,0.49099022813802295,0.8628033686682401,
 0.40171940388995,0.5733657351295037,0.8421195969465125,0.020903804500696843,
 0.025646078181119414,0.3629455895237985,0.11484029772287241,
 0.7360176165458951,0.06735685933034452,0.5619142404786256,
 0.30820114623913675,0.007214925485591883,0.04392015193343435,
 0.6837626147144916,0.3010305140704398,0.3613892742747158,0.36901650065139524,
 0.914085733423827,0.14715565009613607,0.4170913200364256,0.03735903556098519,
 0.17703417113839248,0.2836264932684178,0.0980033161516789,0.7139560559426281,
 0.2426513518758321,0.11717384884773197,0.6920941813492161,0.3242248325819863,
 0.8382461673801243,0.8346000197696759,0.03184197310492509,0.8950674996118022,
 0.061763706332639945,0.32509672641875864,0.5466939631424934,
 0.05450581715411762,0.4015307568181793,0.34452393470874354,
 0.6570769089644667,0.8397881617483651,0.6836953421053931,0.44503682542856815,
 0.633962587201139,0.21909720375448138,0.46349649468226517,0.9627910142190782,
 0.6242854910036492,0.11853146773233125,0.5355994055232567,
 0.48109293007449583,0.5045299270907717,0.17413044432918434,
 0.3236603353701415,0.9488023425224296,0.2534240908104135,0.44001049736387676,
 0.4357774060182771,0.5761371090369456,0.9334151397538125,0.7033385593968395,
 0.766722645867066,0.129616907209947,0.055754081200855454,0.2076386694619533,
 0.46913867815783705,0.9422098519075891,0.9214753927288674,
 0.30339945198135976,0.7613905011461778,0.6305027174910486,0.6229487145448236,
 0.24969937452915203,0.2876305164288565,0.41447458916907265,
 0.06300990891043556,0.7685317189917699,0.3239300363812545,
 0.08887919043731984,0.9746556189286891,0.5952540518161935,0.5406824408480854,
 0.9773306341123046,0.014448057165896921,0.5005478769747005,
 0.6196575787764852,0.5614825555186438,0.5936775745744419,
 0.017917955492058657,0.6101196564814533,0.7078997809022374,
 0.26780773693001214,0.94091193650589,0.8918644105047461,0.49120692267396415,
 0.17714855723888,0.3675505382446256,0.582413187381055,0.033470359583420706,
 0.7507570407650784,0.5984553447958816,0.10344311583302368,0.626721976201697,
 0.2736508953072483,0.16416523937731575,0.8023836708214667,
 0.41508958799359696,0.9566942633661888,0.5607035116607022,0.6384343826582216,
 0.20087578142503393,0.5595966976301301,0.962115902077521,0.24012901572632317,
 0.6813569303948144,0.6906322080360152,0.11042234363301917,0.7854279931657212,
 0.4175839092241872,0.700574107483032,0.105658377101984,0.648137765429774,
 0.04268872505528676,0.7626587080041163,0.9320196896662968,0.3439810372748757,
 0.47277446351804253,0.45333995072903976,0.30550184937787783,
 0.2066005204405964,0.021643384166796786,0.05905634334968213,
 0.20748663472413986,0.72758277437668,0.25717736854210105,0.701071116244605,
 0.7495894417298558,0.4858190846060594,0.8327929962909036,0.9667824720754968,
 0.6915162092572792,0.08479915132625082,0.30502611020626036,
 0.6658330554943809,0.6484407197405844,0.3356048691329667,0.10248269175503422,
 0.295519419055565,0.9184868418015891,0.8069118863055157,0.049185741919407056,
 0.3555395580394918,0.6751379548926386,0.34884975989977185,0.9123063436317683,
 0.6427662780428177,0.8171536313858621,0.2889784556208381,0.06671921508921475,
 0.8705660752093592,0.5744185127981,0.07030938119386898,0.4175096489668797,
 0.014832232078637375,0.9768261964662899,0.7047947936461346,
 0.5552720370376902,0.35501448873365093,0.29386128337790074,
 0.5016522239213815,0.0323706414447944,0.6757676969774838,0.03416015360188318,
 0.7281379163186563,0.43122346411269297,0.9923279350441531,0.9833280731074859,
 0.4640720437269358,0.3561543322344428,0.11915590081192873,
 0.04473063974970781,0.5680031062854484,0.037692501255664146,
 0.3901257123241536,0.33762427613483326,0.4851729567275572,0.5522516111275524,
 0.10546909091261047,0.1516535802906973,0.2205010125961877,
 0.13682353430155703,0.49633893374340643,0.21196701547970243,
 0.8810567962999941,0.8802855088157031,0.8165779384125771,0.5510264333853414,
 0.11075543964555112,0.9854022828072415,0.7269645816180627,0.3315462066696664,
 0.5306214581191997,0.95606968240797,0.1617697922186757,0.26333388349944487,
 0.3373688716379615,0.866483782705908,0.7354126748136166,0.6021334040189544,
 0.6244563185990671,0.7118222386222672,0.922033458947242,0.4843612182564242,
 0.972486045155981,0.002692945042778061,0.23553238523575626,
 0.7784075758128052,0.1270688293643495,0.5920904329108321,0.28167305745649585,
 0.48709072612502147,0.8112743439462199,0.9309762808057084,0.9525939568029307,
 0.3046371098821856,0.644031369044872,0.35293271799807024,0.6117417583758776,
 0.5057249396288059,0.15527949857043877,0.8446286945196613,
 0.09822066291298781,0.7662414629749967,0.6317363474269665,0.6890493681344575,
 0.3515543080978665,0.3729112180289992,0.6809306880029551,
 0.010615400740867531,0.966420554227383,0.7115146605537132,0.9238742162905509,
 0.7467261049560203,0.070390288488967,0.16797383380107667,
 0.046452506859547604,0.508484352137449,0.3530672446672306,
 0.39085974515263666,0.2911895532341202,0.8070137175160489,0.2960866045730167,
 0.9630118824335357,0.2489198099822182,0.47918019167008263,
 0.23538763049967415,0.46519243360027873,0.3158877705276154,
 0.7972334521602419,0.846104482366156,0.403295205069909,0.6544435992550978,
 0.2825371495940233,0.9494090767178333,0.8547127223515161,0.4888965798603232,
 0.31684226161633433,0.9941477288352745,0.6223938733746873,
 0.17638382785498674,0.5804930893143799,0.9441514187639393,0.7247953615445701,
 0.35834117789242637,0.9832736218238057,0.7442298537603497,0.5597764372871031,
 0.9151274206446428,0.8005567246600183,0.13343611533125033,0.6866696731617979,
 0.3677880095859055,0.8654884664263475,0.5112923733783619,0.9881167006643365,
 0.6540230945365282,0.6556995808613465,0.6746048934605456,0.18644548768481917,
 0.315664537497529,0.4594376004971261,0.9251687374070738,0.9167419571204176,
 0.1451253362166427,0.6310645391793778,0.017962661898413,0.06158653186443053,
 0.654414581111618,0.4315680352274732,0.49731489746639,0.738654870266576,
 0.6651621006876782,0.2374485613714915,0.1923086967428842,0.05470463660410574,
 0.001936269751161901,0.7786205161065255,0.5025367239161472,
 0.9160149216947436,0.5066016150160453,0.6754869831202593,0.5413995539464613,
 0.97868244991901,0.22373011542406807,0.6627695898262319,0.5689450660279026,
 0.4014561434535573,0.34716376669113835,0.32592756120795074,
 0.4869550011238719,0.8643131932791394,0.14382797417128246,0.1050437053466653,
 0.3068680959990856,0.24211135022948715,0.3895373330448231,0.6931602797242242,
 0.2531033865108875,0.47031472532641494,0.5757617191361802,0.5201728630976581,
 0.8726428307612729,0.7661256161953198,0.38579478889063445,0.3604055433527268,
 0.900079334228585,0.2292781550944364,0.7700077577653953,0.1438199559243879,
 0.7688023061167322,0.6555090293554118,0.6950791121645506,0.3797154871619557,
 0.605177493321744,0.6662459613563589,0.49112197320607875,0.03501140289306981,
 0.6722211188011463,0.4254451728143748,0.8555692232716144,
 0.0010088460283624867,0.3917330250115877,0.007410742989857,
 0.07733161043857761,0.394745044411418,0.4139880275307418,0.9149234818753114,
 0.3448371450585691,0.36766394253142654,0.13280509568587362,
 0.7341813814557274,0.35794939499748435,0.2162445220284076,0.0507725116433958,
 0.8468781183496947,0.8304181948451317,0.5554224510875349,0.9056909906890042,
 0.8627613751417134,0.5000232321792154,0.26142841838905784,
 0.47660152359245433,0.9159006651295158,0.7888699057857893,0.9756522262684806,
 0.10818260513578704,0.005258327030766341,0.8956569852000326,
 0.4931686111841833,0.1487670184851364,0.7976365427733552,0.7139317271078561,
 0.3558297912210532,0.14447276285015864,0.8255876478275493,0.383951555109478,
 0.4619811247381529,0.7798621040104543,0.15663047141573683,0.7474323488872188,
 0.43160390826816686,0.49632378709967895,0.07044853805288454,
 0.5662033706519057,0.6915554778094909,0.8295845275507396,0.8767578084671324,
 0.39441350334677927,0.28966171523473905,0.2701481395953206,0.925374890083053,
 0.8693197027554402,0.9025520118323622,0.5820164408526508,0.5868759601659064,
 0.9755547167339806,0.06733338883438811,0.9782115816176333,
 0.11319603839211456,0.4791342985021779,0.3736885776203025,0.575476281347312,
 0.18525717980210343,0.05107926771877447,0.8101890062690338,
 0.20567767676251425,0.9732855710720301,0.8609992244368211,0.9719204171491168,
 0.27291709644997453,0.6552681541453627,0.18295524359601467,
 0.8918338420964844,0.6851381364963745,0.14691257507499988,0.2560472328945145,
 0.9678080850351939,0.8133595786749939,0.3704307506471236,0.8058324857430227,
 0.7013566056253654,0.6067284072228963,0.4389404173311341,0.9111231986523973,
 0.19065683245361242,0.18932149042438595,0.03621927801691471,
 0.5666490806960145,0.9532381901497363,0.7317803415002906,0.24294229191074423,
 0.42041741452716974,0.6022520110444942,0.05641240454775531,
 0.20615084049108956,0.4069965280707072,0.938552610520174,0.10093448749407741,
 0.18794777186711753,0.7851754118711205,0.2826347546002824,
 0.41731663672641794,0.38054022800733456,0.8753426537378209,0.479967727150429,
 0.5293733821889751,0.6208665471998147,0.3565805575119545,0.6763302572323716,
 0.7601549436333052,0.6321602855521137,0.1688936877365067,0.8429329883066984,
 0.7669380294386194,0.27513402894051775,0.8734092207881692,0.5451889485193292,
 0.17886199833439853,0.9511125759973424,0.11786420952306198,
 0.39725537098108943,0.8109762611902142,0.9947428239534536,0.504528612461019,
 0.8445920886092582,0.9939577687841333,0.11395744562602972,0.5929031824696772,
 0.7610092876676195,0.3028546438323334,0.3651839752692001,0.8622369971361848,
 0.04658081248122092,0.06916062225310904,0.8415071694360464,
 0.46585159560997935,0.39260035814403726,0.919684566997792,0.8764711501658132,
 0.3429178637136111,0.8897305093724156,0.3232913336833687,0.16855068628584846,
 0.3001530244465195,0.7661420360917842,0.5810551090078623,0.28830707094811314,
 0.9284095806800512,0.22715542764822805,0.7633353446002011,
 0.32189786015946165,0.7156519984962416,0.27388773794208565,
 0.2675554034418768,0.004785940961098856,0.4160436291195685,
 0.016087636859244636,0.4815360414282194,0.45756986156618373,
 0.5318800554435235,0.08196996514454358,0.8415735610872502,0.5099656484662555,
 0.3853618275477231,0.7776855547331123,0.90652315785204,0.991200679998487,
 0.7031494226501857,0.03372465252245482,0.25232512431321696,
 0.20181356266789507,9.447488988971742e-4,0.7034725274891367,
 0.011177494888114148,0.15946842630665392,0.2064974828440025,
 0.9064053496164431,0.014515220327238398,0.48819569847676747,
 0.8793165277518272,0.14094087247157772,0.9777872208929422,
 0.08675260845475519,0.6498188967191507,0.6389051510326347,0.1977561310942666,
 0.8112502716689836,0.9934633548963419,0.7505315244814561,0.44681877803146186,
 0.8821576326651545,0.8816392966806277,0.9096278532267106,0.06880627879884393,
 0.39331200158686086,0.771158968214601,0.8646342779994142,0.4103234666087121,
 0.16288896711416812,0.0693010073243554,0.3412045393642952,0.2524488970276555,
 0.9184323592166475,0.39156092889659777,0.19883208478202818,0.854361843931524,
 0.007381667040494477,0.21362577910491143,0.8294167580257693,
 0.4233941951296838,0.21324364807261598,0.78457863235676,0.2650827151662507,
 0.23011132798027722,0.05733765565317939,0.1282425277933692,
 0.3221279931704293,0.8869162572665412,0.5380141051447163,0.792870530781908,
 0.571602521418686,0.48099278735895423,0.7930355860954397,0.05201855989809667,
 0.7861972876320128,0.4601061194802152,0.9398857415753873,0.37242822623341065,
 0.3277705699739386,0.2864898674678714,0.4641384742175383,0.5590469025359738,
 0.7475980454647686,0.6014273292098199,0.694991339880362,0.5195053126364934,
 0.36297197171734774,0.3874747715414546,0.08348222594143984,
 0.009408542706867307,0.8606713316628997,0.6753811897698351,
 0.4470618223899374,0.642617999213047,0.7022073739623051,0.7259925519767756,
 0.17357578612116376,0.8286324894217214,0.09031027285555804,
 0.47257546209952994,0.7732440236717216,0.24847718136820784,
 0.4952286918312735,0.7704293916335722,0.30300880732081414,0.6931754275837505,
 0.22355439671394917,0.9075255077356352,0.5395407242455479,0.6085197551063968,
 0.5821184583049257,0.437514876637368,0.7659236632241082,0.0611402162785587,
 0.011353668822153828,0.25659163044378075,0.6236514064743602,
 0.23812462456753314,0.8990354477316258,0.7355090521875913,
 0.22825817255229652,0.8761922420157114,0.2677950811026488,0.7172784711036737,
 0.515966920570292,0.8097413965402649,0.3927704770517474,0.9907309225938997,
 0.7484215501950373,0.9082522965144108,0.5925913627412238,0.06588825172412083,
 0.7336016122233967,0.9562887137028525,0.8940718052428661,0.5111940369046286,
 0.5934940451619095,0.3404317575733353,0.007544454760739061,0.066209562706388,
 0.9390791997339557,0.31971795107684176,0.7358002517122835,0.9286132593102995,
 0.5067465188646475,0.529924592706248,0.15839546755280776,0.8071187426686033,
 0.11874841458728724,0.9105829734236071,0.7781621238494647,0.642673575896376,
 0.7529664170730196,0.34944874686604244,0.5459044115214372,0.7768371069544752,
 0.22698022745233948,0.18494119917591345,0.7977012043856622,
 0.9383412935585173,0.6088309119618677,0.1770693742654692,0.8289658846757603,
 0.30700892993722706,0.7823960703846257,0.41620394353356405,
 0.06197396455568771,0.5497653721036637,0.32064322654914335,
 0.7475776379512139,0.5106227836531851,0.19464022739171283,0.5462036356475557,
 0.3411986216774565,0.6308417591053117,0.8858384760964296,0.20045853568114158,
 0.7249415062258698,0.017939241321659916,0.4165541473156905,
 0.5001039755305525,0.6282221800454592,0.08092154815582997,
 0.18183181312652164,0.45792180830044305,0.9674105302934836,
 0.49937701273358537,0.5618079848836475,0.14656332413756668,
 0.08929301629256481,0.294390899932909,0.580022565260337,0.22242842457507261,
 0.2541389777405323,0.8141990064388249,0.8513725571342702,0.06126776085866903,
 0.18305354122233142,0.6965089460902228,0.571871527261068,0.8344553514547695,
 0.02020020319321225,0.6074617223096812,0.8563802589595537,0.9666120194211635,
 0.04613308173233388,0.8328792339292567,0.7483610368763044,0.955964889419227,
 0.521460933762591,0.04919367031573929,0.005702686460769102,
 0.7492135974285428,0.3920610218095326,0.5722500317966964,0.679756462963873,
 0.5605803588787258,0.8080709794016065,0.7024393217918287,0.6917598877885119,
 0.7454859020481963,0.3591898459700482,0.3286216995891531,0.9154498817559782,
 0.49167387688968045,0.09445856187151569,0.6896193032466787,
 0.019700329193433563,0.1137631712440017,0.6680138146263962,
 0.6933065812120479,0.4722643935794044,0.487931783999482,0.846242787880436,
 0.9598170921457622,0.27015971232706115,0.43763544250109754,
 0.3150189978754334,0.6127674430489276,0.9661894725579676,0.4848131481183614,
 0.006208500923199578,0.5730942484598418,0.5233778283221403,
 0.2786933938680999,0.6527835027164834,0.3193916053716491,0.9277422787954446,
 0.009667454160510625,0.029152314945402047,0.2928320196602413,
 0.6770829735292373,0.869163538913663,0.03641222010439804,0.7767386283171512,
 0.3453325539342166,0.8906174356148077,0.1360012855983992,0.9899517572570711,
 0.8680538493704344,0.8776899228029493,0.9256504351070429,0.7449370902521067,
 0.0334077508636198,0.7017328203102235,0.18797761315836004,0.4376217894823957,
 0.1255255506026297,0.020658110074618596,0.07662074191389978,
 0.3281545493268827,0.5651235226065854,0.7386297857491938,0.45599752536260696,
 0.15675843732053263,0.1390618348761612,0.41462063517786407,
 0.6193822824943437,0.5949133224666794,0.4340341357955526,0.49620277547769254,
 0.7991840585082921,0.5508257848023539,0.08886673429902203,0.960711522534822,
 0.8251185489690648,0.7417723968689067,0.47930775800789116,0.1865482010692605,
 0.7102998752059616,0.15206641867204562,0.19748810270081096,
 0.8534170295292158,0.17505231596454607,0.8034523349597261,0.5478749433969654,
 0.19304948394602295,0.7953909413222812,0.2210436817795458,
 0.44247965967446723,0.25095842946497726,0.9939128386150377,
 0.9562188487553758,0.3596392160834464,0.1707744691531241,0.6753284007735738,
 0.2895028272815886,0.7416021236610248,0.06060629883048829,0.1103795856345896,
 0.7906791104663583,0.09074206557698294,0.4024937015539294,0.913692009049459,
 0.9621595204363909,0.1472532243852347,0.8026931403961817,0.670268494352663,
 0.21696229862997696,0.6179056664647513,0.4381876121579342,
 0.47777952798695755,0.38552749438318235,0.3668325462405131,
 0.7719443685174453,0.2678498521386543,0.7705214618331172,0.24562842563219822,
 0.5768299142964107,0.9956874155145079,0.21001845000358488,
 0.004859786208416428,0.6094895604009971,0.1773137062905762,
 0.4322839627992968,0.5717264013793004,0.16260513968638224,0.5128504561532087,
 0.35109104850332284,0.13713044550932585,0.926772550634324,0.5199387652107152,
 0.38329107372419424,0.7201538889636583,0.571212397874369,0.48672983415706306,
 0.5953926921194078,0.9365142278872467,0.4365972722826258,0.13481799619854362,
 0.5645094979426581,0.756157146487935,0.2057389257922726,0.5484352149824342,
 0.31007560764005915,0.17077903691948515,0.3393252461204117,
 0.9708688235421089,0.568780677583103,0.11396352943009658,0.07168172063333522,
 0.12582903746178375,0.6130855778388246,0.8750478387359346,0.1964101348774474,
 0.5678268174800793,0.011090542688339866,0.2642757122468984,
 0.8785170014447189,0.25668063064649504,0.5582883967782502,0.9280800863561427,
 0.03639259903403147,0.5806915146809317,0.4247378711353683,
 0.014423952291779618,0.879446065144062,0.3522886561189663,0.4832821210046411,
 0.2061828138309587,0.03657762371077278,0.17847421396577823,
 0.8385407729176542,0.3234043617932505,0.9683982455160242,0.35931016323106457,
 0.9473337985701672,0.3645486493229786,0.9761704019773088,0.08273565499760427,
 0.2176244304832231,0.30385103372847644,0.827240634823895,0.5747636911498843,
 0.5969245254640362,0.9472093031298341,0.5013887818021487,0.7064077165207078,
 0.020819468684192377,0.533313122436806,0.37309659376457793,
 0.11284685928769378,0.5202626020714758,0.04161917698477213,
 0.49514435862344397,0.6327340355513864,0.6917486983399758,0.6682550807407066,
 0.17104172649820637,0.7763806441592691,0.08925107788674524,
 0.7754546836277343,0.18518674292320092,0.6335037295190785,
 0.23343105914555706,0.3485966206683899,0.22665291288561518,
 0.6540883388201447,0.9101992577742823,0.19864483030309477,0.4263345186469263,
 0.9750073330877642,0.14229529527228268,0.29199458018607705,
 0.6537194823806572,0.8746038090213952,0.31319480422988044,
 0.051250884836423705,0.8524712547502086,0.42157123363006666,
 0.2742763737276941,0.12789597085889326,0.3231585053018601,0.1153090753605549,
 0.8497545380040931,0.22778818484312446,0.4831163282768448,
 0.16979323275223246,0.19850803652817373,0.01973227016683976,
 0.7160819093362557,0.738208465199029,0.7277072987461197,0.5322581524843082,
 0.6750117207566587,0.6628592801027449,0.07125329825099669,0.4232295704216986,
 0.3073892375947299,0.42436541414514506,0.2060694154948166,
 0.047260669448192516,0.5547328474116199,0.341652453614944,0.3596234490558987,
 0.355578451181056,0.7934286760854362,0.993386993581541,0.22464811979187926,
 0.5610473940994144,0.684260880761722,0.009314014258700043,
 0.35440855196591603,0.07875589272892025,0.985121424817346,0.2010554110992292,
 0.24249509885863363,0.5880039402071283,0.7958337799932774,0.6318105697852808,
 0.5639561891505295,0.8640067963723025,0.7136611701311127,0.5929456720786573,
 0.729264990841994,0.11281414126469846,0.8469076026058886,0.4167161949575484,
 0.4763427706104024,0.6480004367940246,0.7865722561182209,0.6761689206952707,
 0.7130459658337386,0.35848360314417005,0.5101648588323493,0.9189427311253204,
 0.5590460352383732,0.17652144667461678,0.789160497689651,0.06079093363417665,
 0.03379782520794383,0.10589520647559092,0.7224630174895794,
 0.8735241545436072,0.8109369881960556,0.2032181656831873,0.3070482502954931,
 0.5783271679829219,0.5470962935394506,0.9637135557296734,0.9537487499154442,
 0.718792006338703,0.04731905158887806,0.4633821714364801,0.1257098752246829,
 0.9628696988897517,0.4138187001685767,0.0687575040259033,0.8634300361564538,
 0.9967649141522352,0.15173184823799624,0.23846275883276913,
 0.4262375584408018,0.9468547645107926,0.48943413182605366,0.5466756922933225,
 0.26042939332859627,0.19712535397490782,0.7315578750001186,
 0.4576864696842059,0.22888528015296572,0.8339089850301725,0.6337529778064857,
 0.19184139049578142,0.6735910942200264,0.5128394929257333,
 0.05287386586886211,0.8663084848281516,0.5431407337003977,
 0.04060081551367234,0.8950391176196216,0.35166080374716646,
 0.31715577527668215,0.7125386436167667,0.9463061436198327,0.1090498685277933,
 0.7275685211027695,0.7830574260453078,0.34049446499831015,0.4921489950892698,
 0.3501462755801683,0.49119671114519103,0.8440247700578579,0.3062913997082306,
 0.41334817731198736,0.10495066922513219,0.03279996604602531,
 0.4904853874950792,0.6027212684363378,0.40728121293262,0.24882237347633618,
 0.9524655855962231,0.7697427563614359,0.07155768006122332,
 0.45654440541115193,0.5282330966374799,0.0011734798095324184,
 0.5226487492260203,0.2806197912977656,0.4049851297053092,0.9736777556842557,
 0.6688200228835435,0.7952796000837108,0.010788717587473684,
 0.3248217373137998,0.9817383324608798,0.8715886396658336,0.27397615687543486,
 0.37056343993663154,0.24030617127384146,0.15402159395823112,
 0.6114167306034674,0.8903020092977384,0.8164678367145357,0.1920801017412318,
 0.807358402147344,0.8757488839008081,0.848195794631943,0.015729938701077417,
 0.8016091347337948,0.5541005905392352,0.5484734691940409,0.4549607185872033,
 0.5370404096340647,0.9482314367016705,0.008421736973443061,
 0.49511111091954196,0.8681727770848832,0.3295738905672999,0.3973775445873369,
 0.826866651358593,0.9329490456754906,0.20260984634979173,
 0.050454698898490635,0.8190738489940109,0.3338661802168109,
 0.25124485836035093,0.01377479691249639,0.41042180205062917,
 0.09696514787239385,0.36096753015479177,0.10201572912940293,
 0.5601015532225927,0.8070136123032278,0.35846704543074326,
 0.40927446577011206,0.6586901671307088,0.1478607622777165,0.568235459331118,
 0.20782783922138015,0.537625025233194,0.7929869689118366,0.953000233789194,
 0.3084299973398761,0.8475702798884406,0.2491660908900124,0.09834908771953466,
 0.3133437419228652,0.02971870915069752,0.010145175511618643,
 0.8014609690046287,0.10008970747699797,0.6650937715174582,0.5022179287085026,
 0.5939870767880205,0.8326063084679616,0.7312341630112812,0.7344947501138996,
 0.15092482885797043,0.7082050761152164,0.11416648213589498,
 0.3820810272866605,0.25490846909624987,0.9673367553629442,
 0.12196112715474583,0.8603075153120306,0.6574734734798333,0.7635142468698237,
 0.17578519572432783,0.7078072884273806,0.7317699738725194,0.7047529600146337,
 0.8511492032683567,0.7626362147097362,0.2215908308252047,0.7659195905345337,
 0.16575206449352997,0.6045666008452014,0.8374522898643565,0.4868546353043053,
 0.9265069688031016,0.4151024830326371,0.8289163317431523,0.19464927490137673,
 0.08533385250506353,0.433203731299898,0.501451120607934,0.9324188171893365,
 0.7348801393268802,0.4280430663147873,0.39069870116694216,0.8351770666293895,
 0.2909342692329021,0.008091175570677733,0.007898819191210604,
 0.6859553836364767,0.7208594465391784,0.5588882583395862,0.9190539659403978,
 0.0746750916774932,0.6423294007105077,0.8321597526482283,0.7754890150062349,
 0.9396718337643422,0.9712290607620095,0.1801956152499955,0.1873594618329798,
 0.5555961542548331,0.48639337195397847,0.6057922906520294,0.1065497969578838,
 0.6068802248727745,0.2517794343174863,0.11228017715788718,0.3438236984689925,
 0.5152688521169043,0.5343768146173196,0.6352139153454642,0.8991425806732017,
 0.5157548667765186,0.3084254519801404,0.3676424650404586,0.7805599632853343,
 0.05228539887445793,0.480075437822294,0.028723524456912397,
 0.8148022232441298,0.9703273135805996,0.2665810095218174,0.9688680885968726,
 0.6451304773756875,0.607591660369492,0.8504175007696043,0.9291594173950509,
 0.5171456790225147,0.1639154342535346,0.5634209868226595,0.2962622115833333,
 0.7428445606661424,0.8125022845637931,0.9881845649345017,0.9000054725801927,
 0.4737846349615029,0.4593774435443272,0.36927083036890385,
 0.33975751316056424,0.9942947675284252,0.03406959750235328,
 0.5085450048187121,0.09456741168351368,0.44605510438379936,
 0.2029699748290521,0.16830891242106938,0.8692246173025322,0.4166851923649081,
 0.0987788374753713,0.9868475223820913,0.9635860715154563,0.2514895190108368,
 0.04231329820524132,0.7140648127626528,0.9773138855686198,0.8184107736051915,
 0.4479143503846119,0.6138766246700506,0.6941899624382436,0.7400831384120656,
 0.2325109629593527,0.21481977280641107,0.44676921614261034,
 0.2615004769985323,0.2907133828288252,0.1732689744986382,0.7274166905092213,
 0.968934239205958,0.7828033854535045,0.5206509670686781,0.5553339642224232,
 0.6020786886511706,0.7122858406913618,0.3232944684294794,0.7178699831560197,
 0.37723844267472906,0.8803739247107949,0.7093273906276939,
 0.32439073539586005,0.3959494955178008,0.3135035429089036,
 0.19578755073658782,0.5327136664486115,0.28142119079535954,
 0.3840596028068015,0.6201942534877454,0.3344974550378059,0.950125927558855,
 0.8284990210235297,0.11048734298328944,0.3581115834940112,0.4000576113735004,
 0.16541499618483202,0.08868328775912471,0.01893497350765594,
 0.011041095470963103,0.7637754959405444,0.16573260034183768,
 0.9095479424589235,0.1005454416224778,0.24721285575623853,
 0.07077113045103889,0.5037552642236082,0.6166601978784559,0.9959965098463012,
 0.08499429859262309,0.8861132261811622,0.6691682507657424,0.4703773901558177,
 0.09911206296212616,0.6803427073680576,0.5599514026867254,0.2272010771841756,
 0.2217779709078026,0.15924874193612548,0.8873363752252996,0.6720161330935571,
 0.015155139250895955,0.9934135342700937,0.818020333107195,0.5577361591623683,
 0.6956965448068592,0.7734226308803075,0.7914550150862979,0.13869266247037326,
 0.7009843730707903,0.1220234853408102,0.9743723241117153,0.7583808483048093,
 0.36403145769871603,0.6545382435382898,0.9791239481864313,0.0706413946911747,
 0.1863686406928705,0.31630612870678176,0.911071974644389,0.9901280927869792,
 0.880099748631077,0.16590775259666646,0.703253768704717,0.9414285716572846,
 0.4413869104711575,0.5405981421992331,0.9604468021485485,0.1235515287301,
 0.9706525792012931,0.8058258100523934,0.7357234827018141,0.14943587595302854,
 0.4818067612304021,0.7190496317958334,0.9403169064905674,0.5573465174200847,
 0.7886240408841988,0.29780085883582075,0.9002905784404583,
 0.18522319719290836,0.6668745736740147,0.6943709316676189,
 0.10539598318562304,0.5396141937123142,0.607278348121956,0.42255977007015466,
 0.1375637846373664,0.9681254232483211,0.2722366531366551,0.6305060859419817,
 0.14179953590219763,0.959952526681519,0.21309973252050385,0.7558964361858329,
 0.8184225758781859,0.09928003486775316,0.6419628392789092,0.1114556354584455,
 0.9783726659098452,0.11087223136648583,0.7064346895202089,0.3082208681891663,
 0.201652226722845,0.501816685893758,0.298135040848282,0.6376151097036974,
 0.4001270528995253,0.05567265540352451,0.6939897822697256,0.9427319496625682,
 0.5526636116235008,0.06968289371561154,0.10204505426575927,
 0.014975297139035448,0.5176108477996496,0.7632721824350257,
 0.16619627299909778,0.43805602115063136,0.4167532824037229,
 0.7285194301852109,0.48807783379943337,0.3226504974410742,
 0.35214812444575005,0.47231661511896916,0.8985189668586105,0.945327012687301,
 0.5307468808208325,0.5676332578255567,0.618523401560141,0.11330653550287695,
 0.836301489720116,0.6706664636062208,0.6425047746924432,0.14128297482930652,
 0.27524336711371267,0.2856289701067949,0.5466985863436326,0.6048264129511484,
 0.5410594631324612,0.7662845986617314,0.3400889413057644,0.2776960939589288,
 0.8662860380560956,0.09724664249968629,0.06228879691457834,
 0.5538314285652461,0.6855212640690899,0.5906543407775806,0.34332934336179943,
 0.10944533254091382,0.8499763017589701,0.9219988424033609,0.9210798863117332,
 0.15508337163207186,0.37664735723686116,0.3177030888453376,
 0.5011796811591106,0.6014509716704458,0.8754199063440414,0.9764293567355098,
 0.4047283696210149,0.5821653263684053,0.7522391461393673,0.47589991237129714,
 0.8969775232298176,0.8067215150979463,0.5532992032753977,0.9220357697695064,
 0.7396992605070231,0.30103741480223567,0.6540344592311316,
 0.048790292166584814,0.3520952894865417,0.7831202921512843,
 0.6034625744691281,0.14183932993885617,0.5507666772316471,
 0.20704783570373042,0.83175562766625,0.5123460674514596,0.04075149692782265,
 0.9306880090551999,0.118662356190265,0.04289402962509259,0.45823789166324735,
 0.18927518001143517,0.9457956948909788,0.5961303495672141,
 0.09612872337044553,0.4505705082878376,0.8756853174017705,
 0.49243969376690244,0.6579114592082304,0.31230874712996715,
 0.48311330045768797,0.9267833294820351,0.8384066048215068,0.606900529950616,
 0.02039120147972151,0.6019153226586483,0.889005025913658,0.488989133820114,
 0.2717519309170734,0.7925467726052323,0.27833533397283394,
 0.15400262916147145,0.2629955575682107,0.22365765987668518,
 0.4482331626785796,0.211649384580205,0.017067758484784923,0.8139521469771568,
 0.11656513289932935,0.20286887667617315,0.38222098888975575,
 0.05872168536424227,0.989681530738681,0.8018487566639384,0.33992788054414014,
 0.594187200943634,0.5483318159964528,0.12125007422235634,0.7531844052331865,
 0.6879885817295543,0.7092222099410448,0.6757462841508712,0.20017463154496984,
 0.7008272915436823,0.6115235031264331,0.8195020108036951,0.48257390599937466,
 0.25099704832477043,0.31049499372697364,0.540018314875192,0.9630253574545967,
 0.07712455892257042,0.7815788981214014,0.6074949129996992,
 0.17530102296951788,0.7327173944233066,0.24199354924411964,
 0.9440953834744324,0.7797570346454328,0.36510706614083066,0.6391334676952516,
 0.22945221766660673,0.493299635078678,0.46971334038612733,
 0.28118703547029267,0.8474651927512455,0.677739764771414,0.7085520752307033,
 0.946709297784885,0.7722128492270312,0.9289700224056008,0.15510174029273682,
 0.19294273896056202,0.6586423408826358,0.36512398368560184,
 0.9866989170251733,0.42804082459839854,0.4014007646235722,0.9802301572830607,
 0.13060292532553852,0.3544005234362798,0.495510685638037,
 0.012782267739842257,0.2824683077312443,0.3116367031210382,
 0.4562195188045707,0.2949532333169248,0.3821732141955575,0.216987492701611,
 0.29670698489268543,0.26839731825884705,0.9625395575740835,
 0.7552339999694624,0.6565794027341993,0.9543047457695749,0.6866201731067268,
 0.263847191119854,0.2185640793961563,0.2843881018799981,0.21372040632127798,
 0.003040008938753558,0.7876559955584622,0.3773370233004717,
 0.6443492338754808,0.4428670923965695,0.48978269794147633,0.7195045633053743,
 0.7196091418627244,0.6064808916920614,0.8488871188118469,0.5876343998030175,
 0.5321703398683935,0.7558255061445515,0.43204755170108333,0.615577657047379,
 0.17348363173637604,0.2465951985416155,0.6435490952367017,0.2628244754720568,
 0.17173014514747253,0.8395301658042081,0.18646277310513337,
 0.33572666428922227,0.769887500478326,0.4138796670587883,0.7727203866973479,
 0.27002805991395284,0.29453010033860105,0.742861361442616,0.8763485094988549,
 0.926750998264804,0.9114532906829254,0.10620435975025944,0.40155010067502006,
 0.5968499959922307,0.565364291336467,0.6369132769919591,0.04452821510493621,
 0.49566410438035946,0.5686572717402475,0.7247602602222892,0.5327844247468768,
 0.8078930684124197,0.6797027445350386,0.22979283135080575,
 0.17210431244173896,0.013105770124474891,0.34962097676299286,
 0.44063774687050883,0.5057660194966072,0.8557357200820047,
 0.11662063243552578,0.5370966104365958,0.40107954082567865,
 0.3419609565603463,0.653371421488615,0.3665383218595908,0.21724418819901103,
 0.6087704224326747,0.3346963098652157,0.9226625816069236,0.7131381351919078,
 0.5092732134174386,0.2463003792708358,0.24531394418413932,0.9721320449844961,
 0.7330408529422586,0.6511283309078634,0.6938729422599268,0.7661334297462529,
 0.09711179491403499,0.3306011554761773,0.6977680646064857,0.8325871833863943,
 0.5941074961963383,0.6825470293999136,0.06512355458200247,
 0.18194241729229277,0.09214880593012165,0.35404481739117233,
 0.29456427646112604,0.5572595824804741,0.07716778967852789,
 0.7062362626596939,0.43222561519839986,0.9279608146502045,
 0.025256521225329176,0.05537354800106631,0.3512839591639203,
 0.2779340384682135,0.4477724927870903,0.9388389861559959,0.8662986769407318,
 0.1690322782749667,0.06527123479484231,0.15277057676980899,
 0.8020216650428037,0.04854925087048845,0.20186733286320613,
 0.8858024705875216,0.5454499080970714,0.8083559605947926,0.819093357578478,
 0.29523788214687186,0.3859618624893658,0.5581189358875155,0.791643657962625,
 0.6448845446276135,0.26535836539742996,0.6073596921390129,0.9374652134047201,
 0.62453548561075,0.4859177941447973,0.03676373159589419,0.9278265835966083,
 0.472151207743881,0.2904146196015127,0.7677623251409084,0.9632521919274823,
 0.7349756621903177,0.4022856087377157,0.7720335797071233,0.11809558746893578,
 0.29308635580788356,0.9551182724745667,0.9505175049486453,
 0.21375529966578255,0.8124773531528611,0.969200375564781,0.3775847659860391,
 0.36167555648907146,0.36451349464711547,0.9855764920542479,
 0.4804587822512527,0.5685360945607991,0.52185417784573,0.09921252456697283,
 0.6255222757366756,0.8243066850652524,0.30823297517442994,
 0.23479070054497475,0.10737330280532953,0.6916085232761864,
 0.24947711983204757,0.34685845257912873,0.7554352948612392,
 0.1957639179786067,0.7317550744443913,0.5440867113189273,0.7184390758339166,
 0.7178814273322625,0.07093216391222712,0.6237750200555947,
 0.13477833530868466,0.4066253756495254,0.6961691201198453,
 0.04862864859110527,0.029888388209754257,0.9486045008216917,
 0.8718498986003922,0.46704581120195454,0.4946577852779701,
 0.020240755912152464,0.8418377047018138,0.17852099125723608,
 0.5418690532539645,0.15641748144121026,0.2906781002274702,0.2157335871003382,
 0.9738164678192595,0.562141121950934,0.5338203198018298,0.43659472341246197,
 0.2651750075702458,0.8789598622148691,0.03996942802977699,
 0.14327118106145387,0.2729249080274976,0.1857188035917937,0.791514199895236,
 0.10223799880276763,0.23017384152540732,0.41476909879568413,
 0.30809254282586906,0.5705027920921913,0.11360087369854166,
 0.6671616051945393,0.6522634160153161,0.7219679567018883,0.6666977204607415,
 0.6929780431276105,0.6216226013821289,0.6141710371577449,0.32313680220595964,
 0.27530554026739573,0.8429089198710948,0.12458155398721715,
 0.7215997460537596,0.8754886420207116,0.36084315765523556,
 0.06924365451364634,0.0824179019006529,0.0518775864151112,
 0.11455428333277229,0.42163309332144183,0.9218179829507299,
 0.04422959595441245,0.6868447362917242,0.30181743633672586,
 0.8638666221920227,0.6630636058037914,0.9911497409463135,0.2869222082858598,
 0.6748571941120234,0.34084637485637803,0.9860261855942545,0.8482601895588744,
 0.31922853563003695,0.14952534332073752,0.7144221160594696,
 0.9260699686430189,0.09344941406437324,0.21081241379641413,
 0.8684051216478024,0.19537691936382617,0.151462894021415,0.5739523897097129,
 0.9469937884508409,0.5108722488645276,0.3750659458937373,0.6080214333341281,
 0.015933471900386964,0.3275487797993264,0.7586695419334342,0.563749588676248,
 0.36649569712231234,0.34114382887370365,0.2450336900613297,
 0.26948311765338595,0.3616495614225832,0.7385925023982927,0.7266870747613081,
 0.5992031478732989,0.6772845877223413,0.2311507498311638,0.950956083839533,
 0.0665721329234128,0.6807295157427373,0.2557780014000677,0.31629352465930805,
 0.5013274264308261,0.45791887303085055,0.6802920340755862,
 0.37681966762659735,0.5516521587795681,0.27269838633317955,
 0.6408432926742876,0.3374782563098382,0.4627420714599576,0.9526991843040019,
 0.6762903794754718,0.5823879333268019,0.6366833044468551,0.46461945336856514,
 0.2279116897782968,0.1541294330685482,0.9309618231058667,0.26334848769546215,
 0.07188611354295182,0.06454072899643482,0.9013450105856069,
 0.4394883589166799,0.014717274476304665,0.06991133949801487,
 0.2494821912603784,0.3817158721452668,0.46239201743861924,0.8370020118691159,
 0.03308709328870263,0.2493529667873441,0.24183942810402426,
 0.9198994602687584,0.13264251031884966,0.625587911645355,0.09587088202531724,
 0.46898838907860585,0.5281325994058665,0.08682112765970318,
 0.6199720767164456,0.9935839504199612,0.017473007935836726,
 0.4111020349970236,0.01113127568406036,0.13311175642829964,
 0.23290460108570188,0.8423607208499462,0.016892987730037112,
 0.2174200866177164,0.3752691912414462,0.6750073668051599,0.21352648965386678,
 0.4452530181543556,0.23712847304795437,0.8395113083384019,0.4615852871772832,
 0.8027987468043608,0.5831571855304594,0.607794430416414,0.4253232259705164,
 0.2793433737140507,0.41579165224666814,0.7396102675373288,0.9765612974220795,
 0.3990285843618877,0.8281791895696614,0.5994226752144004,0.6363387954632297,
 0.32666929353767904,0.6681176063752818,0.5431460216975104,
 0.37344343859055174,0.17892068084509505,0.864998035423903,
 0.048020392024680114,0.12919210227273192,0.2804719286935169,
 0.3555406099671723,0.3451353285005335,0.4349931170739715,0.07682974791145447,
 0.7598218311525166,0.8921630322146937,0.0016310695521096763,
 0.030019564153185074,0.39900070442124314,0.4402006912243128,
 0.17084239490127728,0.48700676561915257,0.44664425376772143,
 0.18939396842952205,0.03590410992911497,0.8194619438110188,
 0.7725497585740387,0.44796425695068565,0.05414053011196718,
 0.2762310154441818,0.6959005938652159,0.10309321822395034,0.6871462295346945,
 0.11335751888744583,0.3919879100708884,0.6927893067142764,0.7490416764393262,
 0.5472941075028919,0.02094461064639963,0.4107592486686984,0.5700158770544312,
 0.5119916784244167,0.25791247828666386,0.18439254286408635,
 0.30302601665795936,0.9460611160313769,0.15796928390073406,
 0.5696650450272521,0.7272774363119916,0.5116083648811636,0.2422237127347978,
 0.0648041941579488,0.13420285387057207,0.011115230620821537,
 0.9480617710272721,0.30154396409293605,0.27772014854840643,
 0.7623984039784996,0.25104959628760115,0.9709226537923791,0.8092568514070042,
 0.10622979011171374,0.6388020640895375,0.7307412954008375,0.359806075858229,
 0.456161408339701,0.751606449119508,0.14155908187675148,0.009362735829100521,
 0.9444358401527989,0.5201505984315692,0.5435815433514674,0.4848721025467163,
 0.11134522149807724,0.8377439062233539,0.9615535192130668,0.5395005001269979,
 0.5505721776692887,0.9543581866436062,0.22882218530991283,0.0982187175050464,
 0.22841945266366948,0.19862231598650948,0.3761870370934831,
 0.21557638378260813,0.47041750376945946,0.9411627297138805,
 0.9114288772454939,0.7201197861427674,0.39401968514790964,0.5424963836290463,
 0.18237537280914962,0.8636148364572515,0.2362076995132122,
 0.12245985222052114,0.1614092455965218,0.7646613605409204,0.4489888801675942,
 0.3833315646049783,0.6390499106893078,0.7591835331085115,0.7340689126179039,
 0.27154696229939534,0.46201316076121446,0.43800891217573545,
 0.04450040309595937,0.4719929491736302,0.49097086327349626,
 0.1559754909691451,0.7161625294996194,0.861688976444109,0.20265773786917707,
 0.6035305163807827,0.9646511900326387,0.9737842940319894,0.24102977517984425,
 0.18611469308151563,0.264126520881393,0.9544573393055437,0.21584757701948842,
 0.6571591997769464,0.3260165290279563,0.09550108142241565,
 0.005572837146454956,0.9879786304470616,0.9370252579131298,
 0.27327335303472267,0.46580564570152316,0.0875475361572502,
 0.09733497343226882,0.043392212928339835,0.840758353926959,
 0.08270639856649353,0.22860379558174826,0.8580807269208858,
 0.34606183933295054,0.7868068479736793,0.9240205822947866,
 0.22970301061046405,0.5034659358367723,0.5778658902863882,0.2642573470670482,
 0.37704070019705627,0.2809081677430084,0.2871793138775435,0.257660094027789,
 0.42775230117258856,0.7220770164255621,0.0434167988673706,
 0.20969774517120543,0.8259209808564963,0.5905355940270308,
 0.10265054131200768,0.6183575269828726,0.9150082820321018,
 0.03371726006891873,0.7328922244176103,0.5511005490453307,0.8916156359493151,
 0.7839719023351469,0.037387738991845554,0.9445214230100185,
 0.16651228818705743,0.12612397504661388,0.5639095069626964,
 0.6671220669932338,0.7806733699633738,0.5428117537562699,0.27663611492117113,
 0.17759315160749456,0.08965461567170252,0.34458241440837456,
 0.8273048627374723,0.0766743412685752,0.7979801681766588,0.6333409319438636,
 0.15252104653009546,0.08800625443267951,0.3777809115079711,
 0.3906857979195023,0.7500089793254601,0.4879082228130063,0.24613245903746173,
 0.4133268345117036,0.7109291365293842,0.8135444286043122,0.7443073561406199,
 0.7913510092343093,0.8310276198095676,0.8345269355513628,0.4082828133020735,
 0.3752796824374921,0.7978328715244533,0.10881029941681009,
 0.06879746079884519,0.42670821552097815,0.573236029036101,0.7961645748706653,
 2.4345751241328095e-4,0.7169655578727286,0.26519463789115827,
 0.2059426666959273,0.4612873333903149,0.7979588554770688,0.761486998831632,
 0.5411247688958355,0.5483241561595794,0.060429935605336826,0.370942223088527,
 0.7687267246756992,0.10571623099780747,0.6387937138361197,
 0.40799663892121185,0.9095929327700232,0.13668054944545016,
 0.8829970292913596,0.9356939166331971,0.23788643618640815,0.752999457831356,
 0.27229533318976396,0.9802960048844642,0.22959029314408574,
 0.9460266533771846,0.1530042294695435,0.18062982110236647,0.3144357427694009,
 0.20275252339487937,0.29168222235324937,0.8999829872715712,
 0.5318990044019543,0.9887373259650746,0.6284621790924749,0.2694712692764435,
 0.9021814582039538,0.6045416904391991,0.2702711680094868,0.8913692088014828,
 0.2268316267559376,0.6489320999631831,0.16858664874431795,0.6279658239019955,
 0.9242790282636553,0.4009802579857862,0.030650079969783972,
 0.6475142864196638,0.4391628284348734,0.8193453315766467,0.5252373633494345,
 0.02270490766471167,0.013435234202592627,0.33838822972417537,
 0.5550922338687408,0.5961629396751159,0.38050125464711326,0.870307829186385,
 0.7658009328573658,0.9397342822338575,0.32024472275515947,
 0.004451450982462357,0.6159774769044202,0.6955415217435336,0.981283030464848,
 0.0042657792703291175,0.6815378029946403,0.2737506231857052,
 0.20721228783763135,0.701757603110701,0.8732449659473465,0.7514130871299612,
 0.9150547894368531,0.5828738404438341,0.5491589353745805,0.1520326935924723,
 0.46772218852469827,0.7940794603016997,0.42506286413665617,0.832991097675251,
 0.6116051340867075,0.03534182833359367,0.3916337186508483,0.5994531875190053,
 0.8125224956724902,0.7223895285403,0.7388238993225731,0.8525917185637281,
 0.13968970301796202,0.5452419150147941,0.12176510082667857,
 0.5751503954566792,0.05354671872787431,0.550982315716268,0.6000724381037038,
 0.737451742181916,0.20247799260691057,0.9357782972187806,0.9074173186048389,
 0.7586780972844658,0.9552498807477635,0.7318696405405969,0.12354290606951146,
 0.46266346917837065,0.47310806824963625,0.5956332179285623,0.381663731614561,
 0.21811634767820942,0.576471986267035,0.8121583703588495,0.3373603510770449,
 0.43279236952691813,0.35354912067794986,0.5163718590513555,
 0.7805055022624446,0.8506589715028688,0.7888939473436583,0.9312737143942668,
 0.5594322573289442,0.3411347490594753,0.39596788233138747,0.8994522392789844,
 0.9240710909634237,0.5827741769889787,0.7876540885367802,0.01124509907455351,
 0.18417659640704287,0.739933317957246,0.06933106695286217,0.5484304879454409,
 0.7273449259854861,0.3124003053117922,0.21936920698244355,0.5638675711932715,
 0.07556125723878482,0.9023168518570875,0.5126428519893681,0.6336531593970389,
 0.7332359418360849,0.15487200250644784,0.9949080191563128,0.3145695461994096,
 0.9075090904703484,0.493204621652371,0.5688519693992614,0.1479768877278489,
 0.6656913023269074,0.8749814360073466,0.9891906986544142,0.11398551986034011,
 0.44124594499690994,0.7098498405125273,0.32981960267735777,
 0.27165192455468634,0.804429118781284,0.4900324396987247,0.9991704352240616,
 0.5303865507348889,0.2871494806100403,0.6043751707397146,0.11589772313879076,
 0.5964891331662838,0.20515181078129252,0.8183226072564913,0.6515734976811598,
 0.5770187531049678,0.42473278687720406,0.8800887515824546,
 0.07173129075344276,0.9309836235149689,0.6517017680916126,0.5453765758755008,
 0.485542282268666,0.4974193596265749,0.5312300608867543,0.909258765787202,
 0.9526183780460239,0.37574870871533117,0.43972348398347294,0.324210759638322,
 0.14092796987710865,0.14926780183341537,0.48080126324732164,
 0.22106798923217652,0.010671952847313482,0.7889689545129954,
 0.7083437118813839,0.7153057200833224,0.34975983181442316,0.8022580187135278,
 0.9054695089898285,0.9949461427036725,0.7413800141611095,0.44418114066663283,
 0.3737987005648512,0.7373121049804133,0.4481695070915337,0.5444069030457055,
 0.3804813349467565,0.7578581530166522,0.286606104237376,0.34038839717716063,
 0.9579048333600289,0.7292174501956866,0.8239153490537597,0.8461663046582665,
 0.05598384554775515,0.7635964810465644,0.9511879326662529,0.8758737403898353,
 0.5216685284626906,0.9811955162246349,0.41556558353599005,0.6591981071246029,
 0.638432198417783,0.8071324554599342,0.7461817942013584,0.12122672617557995,
 0.9089877420400662,0.7206538506178746,0.8217501372811451,0.39588069183692987,
 0.5514178130573408,0.783861642584896,0.20478417895120626,0.7277504895318234,
 0.21094099111931375,0.18597334752513028,0.624825799469951,0.8953955128985726,
 0.490917921776059,0.10936896194657175,0.35221176430704093,
 0.04217002016689797,0.8367872844002958,0.5557713178547644,0.6791990125903695,
 0.023939440059238648,0.1744248528094252,0.8037762944528605,
 0.5806239363321211,0.8750863063612655,0.22923526816211726,0.9325961803895623,
 0.7278751889226726,0.10033982947367637,0.9304995544657153,
 0.41522101339616857,0.9475066223413506,0.020730442288083184,
 0.5572017924373414,0.3655099103637077,0.19334455598751032,0.7237553850756072,
 0.7495960170530432,0.5256026531906777,0.3524422875071904,0.36997637353256607,
 0.788042445932734,0.8033500358062363,0.2916543180311659,0.580206203589129,
 0.3819651443255806,0.6107348508027343,0.4117780192511906,0.4221482500791178,
 0.7839197163125879,0.3306253935983322,0.2824612449714735,0.4366811809900464,
 0.5362469349364913,0.19618122309695207,0.39313563164332477,
 0.5083236073487514,0.7577388422019619,0.2711197238569022,0.33491926778023773,
 0.6903192093113963,0.005176700164154835,0.639984023628224,
 0.06853985455423317,0.10534985126725727,0.08852651658324406,
 0.2638371029359695,0.11972895104036041,0.8086529658362734,0.9079848341132492,
 0.80560867257116,0.16662365594183737,0.11566785201172991,0.6008196428194986,
 0.9427521050215983,0.1206039974206089,0.4430549338751085,0.6757451669267336,
 0.7873511346905305,0.4495292187219859,0.823994814870301,0.08764109419203336,
 0.2747897692586093,0.41400765690007746,0.15573254583409768,
 0.5092435430280567,0.8366077528320875,0.8496796607955162,0.09910173126728217,
 0.3655937623449288,0.7944384959735467,0.863565840827234,0.897347149830858,
 0.4667178824197016,0.16113865107514114,0.1439673625293002,0.8789723754170553,
 0.3124192399375987,0.17483282312502668,0.10350439636081987,
 0.7891742014696893,0.22568083609045364,0.6402941094395407,0.6363529060763472,
 0.9520469991071661,0.28867833305632595,0.9551546503266215,
 0.07908308608517489,0.1750376884747331,0.3264493118193398,0.5238991980986071,
 0.47124705555920954,0.016192937721759115,0.938344797002184,
 0.9702809531029372,0.1862955392585477,0.8510603122118502,0.6560538936753896,
 0.4025391990567132,0.9855193822199118,0.5692418294139397,0.8377916626124063,
 0.19019727736512726,0.5115658189940058,0.9680956634160988,0.7616611121628054,
 0.3634452599074448,0.39676277699998996,0.9527638673845987,0.676146890336534,
 0.7428171435006937,0.02103309618667648,0.7672471865155516,
 0.014150995732056582,0.08103854075755834,0.13146385837599728,
 0.38219806430262016,0.5039802370383099,0.4430921330767231,0.6965911875660763,
 0.839675474623799,0.6575570560813505,0.5815825559201251,0.00597841584742187,
 0.4909892365205242,0.6590528524892411,0.8405534396835532,0.6279117858375518,
 0.08570427793186353,0.7237741120096869,0.47964656875777645,
 0.6695053261197423,0.21955955615099576,0.46135393364992416,
 0.47389049399408134,0.015145579833777623,0.6449195414403143,
 0.19179409732763353,0.8964089304993015,0.7800626675509343,
 0.02958919874637833,0.4120955245355148,0.3466807881916133,0.9375434060357692,
 0.41925279364654144,0.4242791278182161,0.44153294343506255,
 0.7639738769586781,0.9316518126088237,0.5504717174095554,0.3212256563414029,
 0.0636375173739383,0.4029997892368058,0.42569168082235886,0.4842273341195794,
 0.29017576595361894,0.790904464299748,0.7226078425454847,0.9335380283635751,
 0.7757510479294131,0.6610347058564232,0.21506825186568534,0.9554921006257963,
 0.055021914576746656,0.46003028907690835,0.5432689678797908,
 0.08550728124075402,0.9428542988877626,0.12683277876623344,
 0.7868255812799494,0.576344506204531,0.9584806614271619,0.9668337474167047,
 0.4706234045890483,0.22627483021652806,0.9301840301089204,0.603914946134675,
 0.1738036479755004,0.5043992701689513,0.49497467543474616,0.6741930947956896,
 0.9254366104935778,0.21187292984445827,0.3845411936337315,
 0.25379912177706365,0.6585511769767322,0.8203145512295404,
 0.04554996403619682,0.1481264353708862,0.07942760431466167,
 0.8190461374276544,0.2410903514843903,0.8784786393000882,0.3922071677368554,
 0.6032821735737911,0.132085755491835,0.6119384617667881,0.19208915776412494,
 0.5445171122059199,0.5534115210160602,0.9834203163539756,0.0849153683942907,
 0.41939062751399636,0.361138882288797,0.8008526755158193,0.3536084078196081,
 0.12685922255644266,0.9916041112067454,0.9788074564439778,0.9583468292285191,
 0.4847886017318699,0.13252172359274061,0.32537443144043365,
 0.8055303559143489,0.7202527727192722,0.4639872735644722,0.5591277322371474,
 0.2775541195684166,0.46345829519659076,0.35058874848776633,
 0.2583126327356995,0.867567775407974,0.2931065004969212,0.11823282756839071,
 0.23285432589831778,0.5489083097194573,0.5850872882973098,0.7392715717717379,
 0.7009461122087355,0.22911683357942203,0.630947577464643,0.3091544804047137,
 0.8846066124947893,0.9877758885412197,0.8135819536666028,0.8144422048040572,
 0.04303566317294738,0.6519342660303846,0.5883950441591679,
 0.11379661470141622,0.27487260571295413,0.1481463349665867,0.662132115629579,
 0.9850060220250226,0.2982171756323432,0.6897718938565893,0.03146636231213085,
 0.474824276482256,0.9161140626872243,0.35480834376956294,0.2567309794938618,
 0.526737041431123,0.35805338370186934,0.703276350035196,0.9305088783138378,
 0.019452817335915196,0.7750265142803875,0.11072842510898617,
 0.25188611267323924,0.3987984723858381,0.2592297869403246,0.2807791490739239,
 0.005860406845986321,0.8515333833549874,0.6516000953787325,
 0.28053070097481025,0.9974197298800047,0.6013866376822431,0.2858988439829353,
 0.6344140804087615,0.9181318058105256,0.720295487984973,0.3250859347220987,
 0.0036612993408602357,0.14760075332185885,0.5846753615007597,
 0.9377554538895999,0.38357655416995784,0.4032851161425076,0.8288665146177598,
 0.8873344109663001,0.31908949010829424,0.38770661200509926,
 0.6924826417101994,0.6980707893111874,0.5688557746406864,0.9194495522919159,
 0.8297861513411838,0.01984084661538177,0.7716045870789259,
 0.08539593432377757,0.32515672084179637,0.5775600057397595,
 0.5752797626376951,0.08514399354755176,0.17550251379047066,
 0.5579017165178513,0.8297739434236235,0.21398592661574,0.1282129583339091,
 0.5427595118001336,0.9188208215482835,0.29196899529303444,0.4554718453687632,
 0.5460895364083845,0.5851028663832962,0.8493343641916126,0.5160830550490466,
 0.5060159679754522,0.8121729725697063,0.02287917621382196,
 0.13198280325267897,0.9773220431763416,0.0914783306477962,0.1666105359847152,
 0.8510916047323078,0.462477852003219,0.8577671411273897,0.9393614914139512,
 0.7936709073958004,0.7975213590832791,0.6267618110042107,0.06354384550312608,
 0.9850081524319868,0.5024000418813729,0.44201374563941576,0.1436158519167564,
 0.8521737227698263,0.16072782083021242,0.006489258589868929,
 0.8395431840251756,0.19089150571973978,0.12597650995761733,
 0.3872357948015459,0.8210929978876542,0.5623765107751549,0.44164197365811264,
 0.19419057771530845,0.9735320229352915,0.8241911661489606,0.9246238382924917,
 0.8646993023066876,0.767758274314418,0.8768115056863335,0.9361912850791915,
 0.8823684830468541,0.9851235360969193,0.6340442576631493,0.36108346182748674,
 0.9685984539556834,0.1020479479242441,0.09065607218509497,0.5933775737804561,
 0.36648177048344,0.52000527060553,0.9725432629606119,0.965365084705143,
 0.33395979357381,0.9746922474140832,0.517281130135639,0.10090766776345028,
 0.5952901066288374,0.7129702991682109,0.29137885374259875,0.9829727794642661,
 0.8934282230174716,0.15260000335054325,0.7891634447749366,
 0.16832423308553857,0.631853046205358,0.006548396457512706,
 0.07058272208737337,0.392645347295593,0.10250608051377019,0.4298002831223471,
 0.3447252445989023,0.4778908366121375,0.7647078076992007,0.6152353129810111,
 0.8804266103452778,0.9384811404257098,0.5928066958261032,0.03458961551983475,
 0.22903030582019168,0.8662891823741028,0.9368259795798887,0.9701108843475301,
 0.5188692419124006,0.9456591682053418,0.7832631288765434,0.13906112562523731,
 0.16350354771891684,0.9360963817148364,0.188763193442535,0.43033421944105377,
 0.23451144893511,0.8387493348323589,0.17485489394686937,0.8075639550631564,
 0.9615924123922547,0.40792191884868323,0.3059499502086732,
 0.10611849471438162,0.07931129519838531,0.7893310463899982,
 0.5567710733886427,0.584436051300229,0.5635150864405866,0.9392032529024004,
 0.8757388835639326,0.0844493252334082,0.6084809354374943,0.6273343868137982,
 0.7456957094751525,0.8392392658052463,0.7702777879835537,0.8568306587005855,
 0.10524344519428641,0.6070630870177912,0.048261753088271186,
 0.8102153836691693,0.4976222225659611,0.4487052428193111,0.5401336002007585,
 0.818644637113776,0.3678154691646365,0.4873109999383335,0.061861966389290535,
 0.21432809963610666,0.2017299706713196,0.19333400244467924,
 0.6273556027492238,0.7421834397043188,0.7101521356780393,0.1973927007177505,
 0.5202856750453495,0.020042811662957738,0.2748974574263371,
 0.17861609339116324,0.050059373905355975,0.11161079232887583,
 0.5652860137817097,0.038979423010240266,0.5337927877427182,
 0.4743278890456374,0.2965504101747658,0.9713150860391003,0.7402712841430561,
 0.6927453776695841,0.8844363845550003,0.3912822978723457,0.24321661448372633,
 0.47949642981381413,0.802699550871405,0.25877216772955247,
 0.49966840785696165,0.3336187971902972,0.22752381153922863,
 0.7641557613724664,0.020905531236325214,0.5750163317478294,
 0.9587777177529716,0.25673998379164753,0.6539572712440183,
 0.34367072149361344,0.026860160262053734,0.6919624089923821,
 0.41315204444514186,0.860596854444107,0.6531601333271461,
 0.017192443805551183,0.7306647653373929,0.07857254800461466,
 0.5487409189506817,0.05766338100705115,0.6211951332776843,
 0.35166520023527437,0.5340827443546454,0.035960765432462405,
 0.8369061403830074,0.11252925648780787,0.4594677467666103,
 0.27421141393256787,0.7615357723647634,0.9513994581967274,
 0.06289340870741056,0.24556555717146433,0.3198750845150867,
 0.48238137967689554,0.03554316076845632,0.39240351015790653,
 0.5531793527179446,0.6725328691385957,0.5484483408718035,0.8715034757783677,
 0.14766558414661435,0.015305564599422783,6.351557705085442e-4,
 0.5928752531733744,0.011142306559326909,0.919310383780158,
 0.14978737424600908,0.33069347483726264,0.45445246147727336,
 0.8618656057990971,0.39422018524510927,0.13360692188685552,
 0.13357442109367612,0.9682436659471265,0.08170958834872588,
 0.6976266826450411,0.35127675248656864,0.45334074994499396,
 0.5660005560503465,0.8758352320938569,0.3734200581395205,0.621060705092231,
 0.5887465458325121,0.17301993826244533,0.6332733185966937,0.9726196294113127,
 0.18826550193016423,0.6811586233225805,0.48213729732989963,
 0.7314531137495802,0.07117740935834893,0.24127925598056643,
 0.9828571674576644,0.7721002761737845,0.0838233083352733,0.7135748455925754,
 0.21401494543461697,0.42302236191405773,0.4930495256806303,
 0.13470269382735145,0.34502566017326475,0.24572017913330368,
 0.9385757920009173,0.9929141132154071,0.9400257990944839,0.8557703864892006,
 0.7681836314557686,0.8490286618169895,0.19394200344115986,0.6096780467035158,
 0.5564489983568419,0.2959732972907676,0.17723814611576616,0.1949191830622281,
 0.679134064945566,0.8583824702646993,0.02627717231381066,0.9204579997997797,
 0.31479109433418606,0.25038076877973203,0.5139484828227074,
 0.6209329609802678,0.9719308399476461,0.4138900146593456,0.6791958942738051,
 0.7939332447393278,0.894142349506688,0.13037117179981028,0.47019960901254065,
 0.818415192672848,0.41649456883324343,0.4753147910554608,0.6933579384596337,
 0.6111777301396677,0.8631347119849394,0.7256106567299864,0.2958166754409568,
 0.26325130761613547,0.049055699329694114,0.0865115336469745,
 0.8845776733234821,0.28277696456404633,0.34069389380594084,
 0.8446141745416412,0.3597313775849138,0.7571491120845563,0.8070425745705091,
 0.9478715288115586,0.9969070917897027,0.5123597541348357,0.091658130807039,
 0.40678868787904876,0.5845289318938998,0.09509834282869178,
 0.6290799032707022,0.6575793015877699,0.4316446793247434,0.9670056462009329,
 0.8237165805095974,0.39637748828562114,0.1386759839904741,
 0.01206423419743663,0.9351789052612987,0.854650158488027,0.06018841330921321,
 0.8033629801446656,0.2965907094330501,0.8363594109101927,
 0.048787218366607155,0.24971156016986396,0.3100719299511814,
 0.03593286487918723,0.3343112212045505,0.5250366347748899,
 0.33423295764680705,0.32020006470560347,0.1791608014097651,0.611583971409373,
 0.036243826570340865,0.06492133440257353,0.593406869790377,
 0.1489233499548933,0.19775182616720122,0.39146838429899,0.8265913839021758,
 0.15661636059263684,0.1326194240018842,0.8068264097573636,0.798347505649,
 0.7877151159463144,0.5049147281412643,0.04632512419979584,
 0.16428518417063764,0.7284172823139631,0.8640635930719962,0.8876703682396865,
 0.40078476581236755,0.6593795630869729,0.7442806955455226,
 0.07178678658715176,0.11504198926653231,0.6002915268163502,
 0.15198199588985783,0.968204113423787,0.06823870842949553,
 0.47617087472236685,0.692128366199147,0.5568490875627675,0.948228623416173,
 0.8798016754786135,0.7033272245970772,0.15530265019220169,0.8202496270762542,
 0.02430074904701518,0.861397948188161,0.9717162471869644,0.10074708354367379,
 0.553275500211991,0.5393961472904106,0.8465141179707438,0.4968086732079139,
 0.2439951199793513,0.16801196040856947,0.8029172296976042,0.5112842004496521,
 0.2090363966155777,0.8159899530896976,0.011465758215455368,
 0.40379321540556123,0.25419652028941164,0.717768307151396,0.2845003601889867,
 0.42480703704738165,0.676600456728723,0.8609415007243894,0.7876998139878689,
 0.20893723809158038,0.4009101165188783,0.3041608916456332,
 0.22203280333222664,0.18112923296265748,0.720096854636058,0.4644425950372427,
 0.40238920912224196,0.49020730667442436,0.6060260686902498,
 0.9367606817501982,0.6680708944048913,0.4138935643808975,0.6951122272770456,
 0.471669252896339,0.06759695965684642,0.6654865031266999,0.7509284054376553,
 0.37011778290587094,0.30877604797823366,0.18614848219209723,
 0.787298967896995,0.10099566079338851,0.3902931613283537,0.7574626212669481,
 0.1998301708489716,0.6475936099636971,0.09708450830969151,0.223674804760278,
 0.7221578358406449,0.47284378648521197,0.5056401447994068,0.4506829658940853,
 0.279363192458868,0.9266381705872351,0.9474300669413855,0.843416151853263,
 0.43869390254847707,0.5014669348978956,0.017733787748859364,
 0.7047831360540984,0.024384817813296267,0.944291931300009,
 0.45658910455061374,0.03568367166942066,0.45167010135239194,
 0.6616910717507745,0.5467606095938854,0.1484557746041525,0.586500080509035,
 0.3913079390032361,0.1666633268697415,0.7779144428144573,0.6178711742431693,
 0.46045766157876433,0.215852471330497,0.8619781398289954,0.0924205402846835,
 0.5757396696172515,0.9228836560354932,0.5555674421139184,0.203874170714072,
 0.631993403416601,0.41402390088973773,0.34326831678162173,0.2552118849132117,
 0.23619170814869528,0.8452209282156227,0.4389331978451697,0.605647444916053,
 0.29800972504111456,0.9655046524871927,0.5000632246212829,
 0.44860960010308626,0.9285983308966207,0.5714870588893083,
 0.27678087091540926,0.9994897879668918,0.7164458794065119,
 0.13182197256167416,0.43702954997468757,0.7881177271282558,
 0.7717051839128032,0.0907579628568903,0.29125729699086955,
 0.14897595936597596,0.27099628120075137,0.39598258120400986,
 0.18298242301629764,0.4211146512263756,0.8397313993018168,
 0.35637984914076115,0.6914237614604628,0.07099113139123248,
 0.21645275440699407,0.6770340535540498,0.8406735737258069,
 0.06902653024958216,0.4137930344658518,0.27001987149647944,
 0.9707980403523764,0.7796161757428987,0.8185861864726353,0.9129705785702582,
 0.8678217281046943,0.11964330607576246,0.20228371281438062,
 0.23458075957113633,0.5778300862338636,0.9891028037907202,
 0.11936829912598268,0.7519477850306062,0.8880608187512323,0.9356930880628174,
 0.9578640637809539,0.9364710749118033,0.5816590270965116,0.7809501640416258,
 0.5492960338653495,0.738143282868553,0.5517494022794476,0.6174626526350288,
 0.14229872412266065,0.16585146168571252,0.529527707776808,0.7035305455936007,
 0.18377199796724508,0.5902142063238364,0.17866477349139298,
 0.37387861205045414,0.7094090902370294,0.6162403668100933,0.5725174123860026,
 0.49113744857414143,0.3686706919202961,0.9659723065822188,0.6355643477718732,
 0.467507571727207,0.5844820843776489,0.9433502163122112,0.029804126239929674,
 0.7867931621409571,0.6025581667619758,0.4709964896327443,0.7240492266237444,
 0.8587341843055168,0.3273361674026458,0.7140734452319223,0.6339780295428393,
 0.7324468824587355,0.7037623364703427,0.45740239965151286,0.9303950392131835,
 0.11699512391356837,0.40941025016053934,0.2337167293398319,
 0.45047565087889074,0.4695586431404545,0.4303976751544376,0.5167625254881054,
 0.7571716989743412,0.6238040058699903,0.921531631635479,0.5728592211769826,
 0.3792030356615046,0.12625820427550238,0.697232317703807,0.8858942121422981,
 0.1835469004917687,0.2345405297325276,0.8569308085026327,0.6984894433187302,
 0.892630412167027,0.7719251602322437,0.1975651829597488,0.6357820839627465,
 0.8491489602124211,0.51789878116169,0.554536384372283,0.2211277382336232,
 0.7558633466956787,0.3013667938192899,0.8149449035773082,0.24798963073363878,
 0.156468913568086,0.2461643008566592,0.4100076152616413,0.7299622181081263,
 0.0015560854594620643,0.21374396425407372,0.9236690641781485,
 0.8385108712138324,0.8315447964281547,0.005105823208264693,
 0.17047207519086582,0.02415666690438234,0.14628629122991632,
 0.7845224484117741,0.8359804278641663,0.6716549007813439,0.564145941621321,
 0.7648776326304572,0.03761164907431702,0.4595273157496874,0.7880681105433553,
 0.18396230058059948,0.8525672611496815,0.7219783441802898,0.1948831520122679,
 0.3638542476087011,0.07954076508086694,0.1676583375621371,0.889988640426609,
 0.9370344592149281,0.06284941540468436,0.6940655737840733,0.431903513281487,
 0.8993474293736529,0.9715234060721016,0.2783451696404631,0.38940948121382457,
 0.7139587834321568,0.5047996879613166,0.466583085214593,0.36822550113778907,
 0.8309418295489521,0.8414467580340562,0.6395233523225013,0.5202121767615628,
 0.6233376919321965,0.8208858133546447,0.12445543984243124,0.8247931629805821,
 0.9384705416753166,0.8927038296386256,0.42126977754295025,0.5708455159354361,
 0.3702639792398168,0.27433197760489625,0.5142983933344061,
 0.13085080701722862,0.6679856228532951,0.17670384773390557,
 0.37605750891388245,0.8410961604580571,0.6039492875365253,0.9499228030904544,
 0.20540599403956195,0.6712306060689976,0.7044236134942438,
 0.42553730134508383,0.5292336624013079,0.25006296015466034,
 0.12244433000802246,0.19746613595300855,0.5742390982158578,0.313550709472564,
 0.46867021604441383,0.7698017764320512,0.11407126354418196,
 0.26223831568864153,0.4946685063887347,0.5000733578253943,0.2571230994937588,
 0.8977266378318849,0.697114843621452,0.5240319675769018,0.13715525543321783,
 0.3101644508617428,0.8748620098840395,0.9155342651841899,0.0767379491522211,
 0.024359938922506985,0.4669892983963805,0.36993187874317845,
 0.47746038021367065,0.6083082128962826,0.5858942867911276,
 0.036545676139289274,0.7702015714469383,0.3722198961586195,
 0.8760180325676463,0.6791682908802715,0.7280598996045237,0.18194604941785353,
 0.08680875460807258,0.8188302640129728,0.3112548148302414,
 0.20879393830683812,0.1408138172715011,0.30318059846324574,
 0.26169992883677395,0.3121846056962613,0.6212803284105332,0.6676975669232078,
 0.9014092287167158,0.3363376190910089,0.12949279961170546,
 0.02311470517278935,0.0036721100935149797,0.572257050097573,
 0.15014395240232514,0.4816225303166163,0.32290498702694737,
 0.2547862853995462,0.002627713160029077,0.54094840572801,0.5919761478227319,
 0.17771224141451247,0.07782782460101301,0.3633385840899065,
 0.36642526162294087,0.5275882768087256,0.8768198902458091,0.83317434894095,
 0.5784914321984947,0.21801008003432576,0.23068386370698124,
 0.20015756070276647,0.12675801650298602,0.4099975507263408,
 0.019697252043150315,0.40074728201148657,0.31231928788116214,
 0.47558528012138535,0.321194665431223,0.446791691509034,0.32793541699239537,
 0.0710750264144634,0.9306447633416831,0.06908119156029013,0.7235899557429851,
 0.8533050454126626,0.041658974905712576,0.19955766826718424,
 0.23613540615873252,0.8047810725014083,0.020044303691206244,
 0.6341829468451143,0.36189461042130233,0.5232275178472884,
 0.46547557614294466,0.30458886938866603,0.5225070092215587,
 0.14944011452469907,0.3464835613262225,0.5045181633448341,0.607472840809175,
 0.6494296273207556,0.5737014036977217,0.598340430240186,0.5131497315919396,
 0.5122519509314822,0.5368180490660994,0.7546635266343413,0.4500288624404941,
 0.45952097810274806,0.8826734130504336,0.1195247948297472,
 0.14621840506581574,0.9518990226051538,0.3218735338886395,0.8669352780496227,
 0.33472007810771176,0.704567144142162,0.6959195943626826,0.8160776617614028,
 0.10563112487532011,0.8117979454139768,0.608653199967109,0.7296454475819643,
 0.9574886622113038,0.9244094824729808,0.9858498974950427,0.5060600272027272,
 0.825314225137128,0.3937458598279583,0.10233644255168317,0.8605766696044164,
 0.1464751505393378,0.3239578191927862,0.2527801426976465,0.5223375132020801,
 0.013818912235097236,0.14753698201876997,0.5883659435633934,
 0.35041146153480895,0.9730433185275061,0.031595568579293465,
 0.8788081361168761,0.30450715252583815,0.18837115846391694,
 0.5138550505741437,0.4127429880998913,0.276220353636081,0.5890474036679553,
 0.19928344597355974,0.6409942057702881,0.2581801904632698,0.928320548756096,
 0.6537733753824292,0.81749196856461,0.5867350217492896,0.6549268768263705,
 0.7888396952571735,0.7143369518137401,0.38336899074591824,0.5370743435593994,
 0.697892302089101,0.7631426344760885,0.57841195825036,0.179450162070255,
 0.16582406785988213,0.5316328073290014,0.25640009665120167,
 0.8889683757791518,0.1422350131025003,0.7121480402832263,0.38500560700152586,
 0.4669166812285549,0.26007740968646575,0.30758618398229076,
 0.03734068421360215,0.7880129991677198,0.024336941290887593,
 0.591226717542371,0.3161698857264623,0.594176437089936,0.7709041991606518,
 0.17445983009183508,0.05141645768546699,0.2526729524114778,
 0.6930858740579249,0.879150458151764,0.43184123028089294,0.5304841131052667,
 0.8107970192518428,0.46896794472431635,0.28199966560264866,
 0.9831619888965036,0.6470202481453328,0.042129288479222926,0.635977387479066,
 0.7150703943066796,0.07733364121867436,0.9118167686993344,0.1956140034176187,
 0.8045680946164492,0.9433925811793814,0.04058548338024426,0.9034000599458339,
 0.6206481127252981,0.12620573637287924,0.41102090322100526,
 0.6159717625197488,0.8504665035530284,0.7534421723408183,0.3778400082990966,
 0.4544364453165959,0.8971589206955692,0.2907809816114283,0.3547900905310166,
 0.9597611598179583,0.43902863329177544,0.20676493906201276,
 0.6943091752944959,0.4384068913582292,0.9687643003996613,0.5248293343633794,
 0.26166196751274806,0.5528030523210825,0.05797198224433664,
 0.21106623309333816,0.35809480687700346,0.13953062561260643,
 0.4438823926305382,0.1661283032586549,0.46774549647617536,
 0.06843598244285165,0.6288178726796053,0.38406213471429107,0.517207304104944,
 0.5379391297258497,0.24771887517688518,0.8673979398846914,0.5255041298379809,
 0.5115249204217444,0.4705113260596743,0.2905508683904914,0.0290248562340244,
 0.7672007099244862,0.32477834103691194,0.6405713460132033,0.635030011258116,
 0.7680281936717577,0.6287837935815823,0.5203142291377905,0.29719685217759206,
 0.07836183581146472,0.37923403490783936,0.2625391287641782,
 0.31698084882936295,0.3225089035694332,0.4779194744293812,0.6968128961008816,
 0.21312374310322968,0.7804255865358827,0.5786188240638178,0.5757513796349991,
 0.5592459811175878,0.1315507535683571,0.6751409735047006,0.6033337486750948,
 0.490160029179478,0.7949057558306563,0.2904918490768752,0.49324446630833385,
 0.3124809423499689,0.8166796690150542,0.5907895204313651,0.6365737360570416,
 0.49649124320233773,0.9238778224156521,0.787407991029239,0.5714919729666116,
 0.6664405090886779,0.0793709550403322,0.059027309641097925,0.157531317276967,
 0.12217709883398287,0.4182589435061652,0.23097414744328426,0.568397664061058,
 0.2451931005967274,0.4623355445560483,0.6631910440874077,0.3108215849725564,
 0.933444874650843,0.5060341250112201,0.07803487949571031,0.8890720912857595,
 0.3484301230340018,0.5214597166500995,0.2637639897664875,0.18177814381734647,
 0.4933304701033018,0.6124348072033603,0.7014137585336319,
 0.051127802483328044,0.9981434433433536,0.3058982746092417,
 0.9199331216256866,0.4464402191334014,0.48636598848547596,0.5447433212370502,
 0.1829564496815238,0.3590233441625422,0.37717954774610585,0.8431419191369114,
 0.9184228259436511,0.5782432117295303,0.15531818085023297,0.2908965243776439,
 0.27440973705287786,0.6195758384349227,0.46112155799055465,
 0.14049556878904212,0.7293125830762113,0.16132308279733754,
 0.6143491366692753,0.7721953982578578,0.982072933884425,0.6430513621017397,
 0.4171416583366837,0.32804912942788067,0.6023859034231972,0.8365172769500041,
 0.005898243058527819,0.4168425732891986,0.03689503842967612,
 0.7069858727902542,0.3517329927265286,0.8972975112789114,0.0689069544266262,
 0.10571665913327444,0.42645989921406113,0.866635264859928,
 0.03985428393458634,0.02689156560051431,0.797004811637632,0.7764483665813029,
 0.7283343543510372,0.5873781372141322,0.6025242134302204,0.42513935550611437,
 0.7681988633663117,0.8993316470298015,0.5494685117369525,0.6881407752728621,
 0.23620199152121557,0.8075381795821904,0.2641335808473252,0.9263245038554243,
 0.9066557745003285,0.48851060831624915,0.2215004696898598,0.6301227143982657,
 0.14053084070891853,0.30060180473420295,0.4908864001217391,
 0.5478392542543369,0.012921233966737167,0.24136582928266936,
 0.5429986978034016,0.12055579682860773,0.9996521441315965,0.7135447350968988,
 0.7681636954304417,0.15863979216283042,0.6412206617729825,0.4577589764387422,
 0.9877971062157196,0.6885484684339128,0.15077142545669697,0.6212177152866907,
 0.3846532845009747,0.1449129959795702,0.6810993423125999,0.9251772514250455,
 0.7975800539372682,0.5640735479058837,0.7693278002192613,0.7841966454009273,
 0.22414855603797423,0.5706150552110508,0.11632228337545825,
 0.3230237726075258,0.5847576799578919,0.13623617141108602,
 0.35685458337302767,0.14541214148478165,0.6022641465304144,0.581619540381464,
 0.9417862087127463,0.5722913174391915,0.5607014431552455,0.713200121695232,
 0.21256081882380407,0.38372987689793026,0.8569852712962979,
 0.05281981456222784,0.10045937471500566,0.5208796487679482,
 0.3638024682712496,0.43260171841254524,0.7928348363954436,0.9005371870432716,
 0.15257935535057088,0.32166284087949026,0.39560555381733975,
 0.621686583370242,0.04881483860813263,0.8798446140805993,0.4791507261031407,
 0.24032331765344306,0.3087439310121145,0.8587358256851517,0.3740632053796087,
 0.1910148155548158,0.513757393039008,0.18112754191372527,0.3234093166570191,
 0.7960781377478212,0.5915613061822066,0.557164183576141,0.9765017744292367,
 0.4852084303029486,0.5156055204383316,0.6927786753448077,0.651938531948134,
 0.028001609800871963,0.2369025504547254,0.4823100674315204,
 0.2996861252405949,0.41209707228831627,0.3865305572893418,
 0.039445361554545366,0.8206283799766576,0.6627495762008876,
 0.17018496819524032,0.8177959795217555,0.562828092341429,0.5074905634459863,
 0.425924261508862,0.7464587495899352,0.30032281129168037,0.6709976776111897,
 0.4636889849338681,0.2601897953403929,0.41569547863689993,0.8051139934797457,
 0.2273503688691184,0.050967402168522025,0.08739144790500786,
 0.4211451493063858,0.42682711290716757,0.15349783935458894,
 0.9225423258445469,0.11224254576364512,0.6355922848383714,0.8898687756203946,
 0.3368500829934734,0.43789829314344375,0.8960854188371729,0.8366567580716759,
 0.08285760597118685,0.7206381912534342,0.15760445756414299,
 0.23084083165767266,0.14432312848389128,0.44705084210663404,
 0.46595039616074485,0.432309928721332,0.5597953652808421,0.654230471399748,
 0.5121631653571221,0.9545020902172346,0.701717502200754,0.314850672614039,
 0.9934776018920543,0.9784963643822121,0.9462826208240444,0.07878325204226766,
 0.8935498633014556,0.6071466960943632,0.6040062811865834,0.6075835376723907,
 0.07108379798125564,0.5814005062042219,0.3940504093917312,0.8320763841081551,
 0.9890567041602885,0.09949780390888141,0.3002846801483149,
 0.08609255002478067,0.8360126995637622,0.9755783673622733,0.342254793194273,
 0.23479066642109458,0.7811499627498639,0.6095649812940801,
 0.12173029342383357,0.009079681824510577,0.6749265755451649,
 0.899277187852778,0.3770430040031644,0.3833595723419023,0.17990237696457934,
 0.3827419923256865,0.043606726580575295,0.5751472064292804,
 0.9062476748999108,0.65637012977257,0.9137485851624574,0.4103121911124814,
 0.20273313628947653,0.03517633819080279,0.16490535053817978,
 0.24682248198296663,0.3812669851077046,0.14461630606047027,0.61189563466945,
 0.13007659072704447,0.597876409246787,0.9185039706568643,0.20565221765884345,
 0.2006717496531547,0.20548661707126492,0.2014393857353135,0.7805506173769491,
 0.8342525946120325,0.38885626509051086,0.2674786187787853,
 0.12440353743408739,0.3933377910474226,0.225956561215375,0.6463766665810091,
 0.8934480392182585,0.7590971704054125,0.8788095088123807,0.3252553200029329,
 0.9954989805585,0.24500668372619938,0.8631896011449147,0.7972724081657576,
 0.23650860409555163,0.7943727974028207,0.18732832005999056,
 0.9058224327998943,0.5184901428253085,0.867853524079087,0.4057608505988043,
 0.6128991631304354,0.6861055636334699,0.36729533181246987,
 0.46454504728022195,0.5601002915939443,0.6020953337512311,0.3030245093515681,
 0.024833549476750427,0.17661867118353247,0.992791868845093,
 0.5990407092510841,0.6991318118031707,0.9924596578176228,0.04609776816264233,
 0.1727861114955631,0.021273998585868648,0.7778366092612243,
 0.1630751836613331,0.9457104369120003,0.825978435398795,0.12873614816482737,
 0.10634288285297966,0.11424822739267926,0.9808339124730209,
 0.2597564314356623,0.3572351769674853,0.4347769056476304,0.6206431040685051,
 0.12003464627476856,0.18610137959388295,0.16301653134264427,
 0.9663126636188653,0.49610412833448114,0.05872476002749283,
 0.5603771435642491,0.2891057970086385,0.2675193792484424,0.15854737211142078,
 0.47171540377271626,0.4357241250182181,0.4379351430291045,0.0952228107666081,
 0.9566756396918755,0.6146935119263827,0.6937606032117382,0.12698589797855164,
 0.21056776367388952,0.7315032633568582,0.02527835738185602,
 0.3032156593804304,0.9993228605155439,0.984232047509759,0.9385215800193698,
 0.4314295037487269,0.04279667149897526,0.9645240046272529,0.9554357652364647,
 0.9593436971511472,0.018002503337332998,0.16158031680661267,
 0.25303462130475474,0.6186970714049066,0.9787985989529289,
 0.09976467471500661,0.7084618522749554,0.9289563829253012,
 0.17687375483996126,0.3254948366618414,0.8075974893179406,0.9735036470929088,
 0.10347375477702436,0.9694216518549512,0.4610573279638246,0.2917892263693119,
 0.4979506717734925,0.6466287109291807,0.2941895056685344,0.09079902498598413,
 0.39297451819418194,0.052715601089388664,0.9491217817926952,
 0.31229804682874285,0.17912591922902577,0.2158988462560516,
 0.4585543673195396,0.7322594909917867,0.11729715966830179,0.6129930780018827,
 0.05797413721280553,0.15106833576132894,0.6576713796318139,
 0.5654500416451843,0.4442747668971093,0.5541765816002285,0.4232544576803523,
 0.06797803609952724,0.09824403660618575,0.1931065822246728,
 0.16618999125031308,0.10648176996766134,0.6174779817103605,
 0.3467972122593772,0.06693276268588666,0.017733949357204182,
 0.483612102764142,0.11297732349638889,0.7665365287026935,0.0035880263328677,
 0.6454103815099783,0.7390148881165021,0.31169092043713575,0.2529006892819724,
 0.2776186744869046,0.36481512044726205,0.4290632835063366,
 0.18411406405954012,0.8693062927229422,0.41242411281288893,
 0.8834771161160074,0.5456479853240094,0.861037799725905,0.2637791916829596,
 0.9160486745581315,0.6316246568825281,0.2561081222823306,0.2994765388039945,
 0.868121998122542,0.6401042255478067,0.9820947635241797,0.6725756874342742,
 0.39523482230853524,0.8686365780511826,0.25735499320496824,
 0.06369617291154384,0.2971039299083169,0.03783410658816955,
 0.8242107086772837,0.0834507811180758,0.9311476250782823,0.971687034880552,
 0.259116048772154,0.17435920578793895,0.28190033700057604,0.3098432517231893,
 0.20969394549041642,0.1937501277199557,0.45702565737772294,
 0.8533752562534697,0.1697587517589535,0.5445654854721255,0.48109828579613634,
 0.37890433542476565,0.9505690634230448,0.44617876138367873,
 0.007277020600543027,0.6561000042009177,0.8807408237323586,
 0.7243920823963588,0.39371764870557224,0.5780436385212994,0.502308653334643,
 0.9489732350755744,0.34508305562306685,0.8855354966469092,
 0.14147016521814293,0.46766510328629907,0.9751996238437071,
 0.7012189256061612,0.6647250545850399,0.2579350495751185,0.5023377505398833,
 0.9262860335165718,0.9667965922028399,0.540928277625071,0.11246160513705428,
 0.6314119836186799,0.27014228667358264,0.7875588413873477,0.9165383476342087,
 0.2515942509508454,0.8774651585070925,0.25825599305755753,0.154372115729549,
 0.9486905168954786,0.12066993232875733,0.2769125477423714,
 0.35227010373211376,0.5831340598351159,0.958289190296794,0.7589723022448294,
 0.2678793838134429,0.8932153027188771,0.8538714796245652,0.47139115793862374,
 0.5586827537926268,0.14216378211505576,0.7033100144153687,0.7396663855824911,
 0.20405145918994372,0.605822804029179,0.8689928868416864,0.9773014537002529,
 0.6792292974800829,0.16047839762653182,0.9505944364882535,0.8871440155616674,
 0.09201570728793063,0.8448616484228176,0.5165613980073489,0.485421146863788,
 0.9679050824788986,0.0995228282133298,0.23747240933065283,0.3970269321313049,
 0.2717470932863898,0.362119654529921,0.990451752853474,0.7629941401706745,
 0.6099698111040892,0.9728531691882446,0.3214133247641866,0.30735046531361665,
 0.7900961219020246,0.7793411080648573,0.8537791474695449,0.7536487773070341,
 0.70137635920514,0.5780454928272657,0.4984627580656227,0.4255737079050066,
 0.6218681373766919,0.4057007228681384,0.30351774689470146,
 0.07896596361267072,0.9102581125038913,0.9296289338347316,0.5685895275394196,
 0.8914013646934853,0.2621656666947051,0.3579852376613156,0.41775931910905484,
 0.44011171198862464,0.4565740946042587,0.39389746397134695,
 0.060416045469692925,0.2490007229214789,0.07878667780526083,
 0.6785799832219811,0.294087640265541,0.6294215482139361,0.417809170637936,
 0.6220161585882705,0.6146833232041926,0.9520805426813173,0.341153156328891,
 0.12869641833146017,0.6330191994998835,0.6437878633059364,0.9289102616016356,
 0.06299741954645,0.4242596703142335,0.9836637377602458,0.17125992756594344,
 0.08314323849236727,0.3540484342339789,0.3776755796021448,0.7451152945056522,
 0.9028150540385504,0.5582911776296265,0.7039918556588757,0.6200654512435138,
 0.08519300856225853,0.49479379898651255,0.6650527598489492,
 0.17364562799649885,0.11743448399382761,0.02972738371408734,
 0.22114484347515628,0.12002205710823566,0.9681184849928248,
 0.17282664704332729,0.67963684167429,0.8835181841829335,0.3007629593250143,
 0.18939920500302443,0.9425061460573162,0.526156534812537,0.43560995151643733,
 0.4890786844249141,0.8607345701790907,0.08881799153121206,0.5159549144306222,
 0.2853785901780096,0.1942932650820698,0.5607076483987513,0.34328378714703867,
 0.23761880539188418,0.15308944305017502,0.36370349806518953,
 0.08585932308083777,0.12211730421212663,0.1060526321254337,
 0.6187191871914317,0.10177820232264512,0.5793754944576546,
 0.05684286992659804,0.01458160077482007,0.9362435816095437,
 0.1403360320400041,0.6370714773016574,0.6744307747242636,0.6865632657317002,
 0.3762022534790995,0.38578480359294054,0.9237649999855692,0.1512164547574718,
 0.335737907981303,0.8973081728216581,0.7491631976564412,0.29149169547267106,
 0.4109211488906286,0.5821566560761318,0.37455809874246426,0.6443613469789631,
 0.5309795905305764,0.3895481606713781,0.748917522152186,0.39384412008934144,
 0.9917461748597107,0.7379034079600949,0.181124952172536,0.40450828936190575,
 0.008151386895930757,0.38287148678959537,0.23978822287578538,
 0.20352083279766842,0.43450105571614717,0.9411933764388889,
 0.19775794774787858,0.0047345024889579035,0.23643387107994418,
 0.8781043682898967,0.30877847391143387,0.4857383513573872,0.7223521842958517,
 0.5121473108580095,0.3897427239133916,0.1716232794503274,0.33089834514460736,
 0.352180327979292,0.3730219257075428,0.43454015841347216,0.23065891121473991,
 0.1416357763972096,0.10467200276924571,0.9320028726734704,0.8323039416180266,
 0.417791371104005,0.6511195513362927,0.033060896975882215,0.331002225234013,
 0.9277008761750787,0.7993666470021262,0.5425634287019276,0.35985493798663737,
 0.2090034461279725,0.448333974388066,0.906228882770389,0.03654704904975903,
 0.37439496252493254,0.4753835984152741,0.6163709631586363,
 0.35341361200739296,0.10424943291813982,0.7204317026961109,
 0.16266090677851308,0.8230672339432227,0.5699083320240107,
 0.10349390532976366,0.15668459533353818,0.407805107246919,
 0.014668261230356139,0.13275201133902992,0.6887196012271622,
 0.9824523213870774,0.37620024331182444,0.023172018043902565,
 0.6166415975669268,0.2038878248496374,0.08977378106883949,0.9620456402916531,
 0.10197506273598456,0.6272374535286088,0.5535481628579135,0.5178321021132689,
 0.3618167594389933,0.01143484004324169,0.01153951356237548,
 0.1601269604733745,0.6975503153396179,0.4018552750070492,0.7972368187351735,
 0.07203381226470373,0.2231084397682509,0.08339697237994026,
 0.6995811168837361,0.7781819172139052,0.2703783203760801,0.46345467762283676,
 0.7415386201299844,0.45583835830534825,0.17891748384775608,
 0.10824485819680896,0.9967944691634143,0.24008037465124776,0.799272786185158,
 0.8623436806701736,0.30703242104766415,0.1702146736238752,0.6352502112138625,
 0.07069428196264393,0.9165747174346641,0.6439163181556401,0.9486208287157503,
 0.6305991172646528,0.5131230692042899,0.03155147340131437,
 0.07539580156595882,0.7107298409659422,0.7778839660375777,0.669316599548459,
 0.009636522902110034,0.729537239130971,0.9874676653851477,
 0.27926646210607675,0.6588907137815903,0.9919579635164995,0.5049638060626032,
 0.6145093954790517,0.07521761122428672,0.3520752392954343,0.5178701283457141,
 0.040304601031181164,0.48210575510348186,0.14181905968114816,
 0.24386693072620602,0.6554593643381307,0.6113583626970461,0.8691985651138364,
 0.8481264951094218,0.6530502943203577,0.05164087247175486,
 0.24677988367765802,0.5917019985225529,0.4733628215013931,0.9968408937645659,
 0.6951351923382888,0.4912344526056419,0.14524950453574714,
 0.18855851238262755,0.5117707734494258,0.427435915459051,0.8124154352082069,
 0.5131643208081522,0.009758903066191538,0.37168316676199054,
 0.9242657714363379,0.6384719036769133,0.7167160272936289,0.8876268763682722,
 0.7208329324086715,0.5005880139072244,0.388453875559566,0.7773161421951019,
 0.8038629185235265,0.3261310272318598,0.015284877761668647,
 0.2763717584004942,0.4894201423850546,0.5949157101657018,0.7427456735024268,
 0.334051665445372,0.7404309955547033,0.7614633598720375,0.6458146214265192,
 0.8869739729872653,0.13059147554174655,0.5187206507623734,0.5835820614353585,
 0.2855003246192742,0.43000695529815836,0.2431501782583564,0.2347747471324182,
 0.8854508947504318,0.3993878484086827,0.05929034756494089,0.33692831763985,
 0.5394851039785817,0.15694477876442647,0.6820619110592202,0.2655216237709237,
 0.3304644243614858,0.1033259095582606,0.5306488881403375,0.2215026524743242,
 0.3482788698868837,0.7493054941514753,0.09503349977583142,0.596112604973128,
 0.6764677914100204,0.7127891127704128,0.18578001852069814,
 0.0036546952623636164,0.3538104987145888,0.25118617084213024,
 0.2598919446123822,0.0798473937510713,0.2742793913651548,0.4086465818151649,
 0.4381950763758935,0.9945988432485762,0.08678732156198432,
 0.06902482895800544,0.13214581273698411,0.3516697264671561,
 0.5388190530620612,0.6079780533096795,0.21876758274119668,0.9844733416332251,
 0.22804201545236413,0.7226412884267168,0.658195038304018,0.5196149564495088,
 0.3124949634526105,0.6379617354820394,0.17717959251755877,0.4329077306501693,
 0.974466638314543,0.317714601116298,0.07610485371366604,0.884768795728099,
 0.25712378963468685,0.04029804405656068,0.18234192923234394,
 0.777160244577816,0.9412373750462404,0.29040768228851954,0.16846794448262525,
 0.8673461925110602,0.5532118350054068,0.6331543696128427,0.6529246493055147,
 0.32083392182897286,0.02424816689707554,0.709263382006556,
 0.19205660762954346,0.24098096139783154,0.6220499899051093,
 0.8405425397075312,0.4173223128462834,0.3656339585957721,0.2643933354482001,
 0.06467205244950558,0.9821152220397124,0.17921533756804087,
 0.7873356029992369,0.41725660465887204,0.4456253282524192,0.7726627935782927,
 0.6285598517887139,0.4270667055146695,0.8202840108938624,0.8904725271593608,
 0.8080818707114609,0.8556917213939296,0.8800571747409717,0.23607979183999162,
 0.7262630292037787,0.11703001884199637,0.2900939752561147,0.5061318199946465,
 0.7681924524226567,0.7442700533871495,0.21889046128425949,0.4554582568760942,
 0.05779682494970517,0.5346978733105431,0.8310540361187237,0.1291721105586292,
 0.20175458653424705,0.40859193972846186,0.9518354351667194,
 0.5144349494181532,0.3878480387051457,0.8535578770645792,0.26562754060392435,
 0.07385516038099504,0.11393212897731653,0.25849237122206625,
 0.3483802539932612,0.8049090155768167,0.6689763287092381,0.9776970700899312,
 0.8066059125601498,0.28572336178729574,0.8098637814971478,0.7190005727991546,
 0.8383414292271842,0.40480484926980387,0.01298368481718215,
 0.06240801105777394,0.5597050143395716,0.8195245766551389,
 0.35077969340330484,0.48353918948104646,0.15078468589450444,
 0.3519245505678317,0.5992711588137545,0.8211843635701875,0.9440768378349709,
 0.4647473924979484,0.8247689466061655,0.2703211810676889,0.7997747570827244,
 0.9536351066250927,0.9527514020080184,0.008959196544596981,
 0.8412700673985684,0.8763595348679012,0.27343716401335105,0.5968004004070725,
 0.8104669883886159,0.11128464260628668,0.1550082555311646,0.652792882291072,
 0.8670635198648288,0.5185868679619542,0.5374284220476429,0.8025686099871655,
 0.5025949280719573,0.003494993920966638,0.6119919882316167,
 0.10598738522402329,0.24532646950655512,0.35605902805457157,
 0.02751029950120709,0.26741686947449483,0.299184743932951,0.8540791220728733,
 0.19573887723181027,0.9612587647089874,0.13227210559101255,
 0.004567033607850046,0.4413468192041904,0.7190602980772948,
 0.05624849999985726,0.74952872886595,0.016378963118363177,
 0.19170522074885543,0.6956077803342241,0.9125070594477699,0.6845732714166805,
 0.05378382200035203,0.39769903443034893,0.17314229809674087,
 0.43683565952729364,0.5385168081661622,0.7806578679859449,
 0.40823337903098145,0.3292862546378359,0.9463052623616963,
 0.21703401094321106,0.0571253201185038,0.43766710648663754,
 0.7563791701303492,0.22203700365647405,0.19979174505105468,0.213897254968197,
 0.33263925971415986,0.09031220517642491,0.7160642786074576,
 0.6244359375814525,0.006613553713884945,0.03542739980269627,
 0.09953126716825644,0.6067341264714665,0.1284590611858617,
 0.27298735968551924,0.1668331877747835,0.7719592299082615,0.0472197486003636,
 0.5667963648671533,0.12121293412062828,0.4201135416945483,0.9137553067713305,
 0.2296391787628942,0.17970676202707347,0.05379279069364484,
 0.2797433331915722,0.6722444943204763,0.20688864190927791,
 0.24399474579749925,0.5926607720091492,0.06710761831542977,
 0.007400952194248855,0.2717091528899507,0.298557893682889,0.8467978432950274,
 0.20654651369741184,0.4959469646113952,0.05521411094196593,
 0.38384006614392685,0.5223631944277488,0.9645460633371048,
 0.10116995541924179,0.41254078305616737,0.846370840358553,
 0.029124549023726187,0.3502788777771392,0.3571217742355647,
 0.39068404613960117,0.6018249988519795,0.9088943347852537,
 0.12908883934567617,0.9228454085396081,0.6119703686148306,0.9221393425260711,
 0.19796450372474927,0.0950723540973677,0.3606741204067675,
 0.04743019997191278,0.13702111733406175,0.29101083282446594,
 0.8670979829753183,0.3990987218599642,0.7661496257555637,0.15482752316187598,
 0.8040168840211188,0.24073458083346355,0.6763982970554674,0.9205313491389097,
 0.9511551222737393,0.21960371727398065,0.17023120180117568,
 0.37719755751816375,0.8191439787529922,0.523551520673529,0.45627435608018263,
 0.2208237986161803,0.29226614142703566,0.1002357465224587,0.719735731721442,
 0.0016300231468148185,0.23617460793067924,0.554713847658423,
 0.11060496364309258,0.8960558855716736,0.2479606673036101,
 0.23899545800924882,0.8204362847883132,0.9453641770772718,0.3911935416624104,
 0.23938141721143424,0.8376972424147622,0.29507590336004796,
 0.3548387673334501,0.6255575162280231,0.6395179259094573,0.7014433817495912,
 0.8675568029480816,0.8814146724125043,0.19234818002758436,0.8739232068977767,
 0.0926718114830899,0.7292582840356876,0.8781650029841724,0.2165171988230109,
 0.13560658337155762,0.34365106851877936,0.4064364353256881,
 0.046001837527635425,0.23688997031887615,0.2256074456653917,
 0.0308493919279067,0.7385181198810519,0.09348098076869893,0.5026498658877395,
 0.4982001929931952,0.10036482839440908,0.9119655293404897,
 0.09293476754584817,0.27785147489023676,0.303701596846481,0.6415203371046777,
 0.10570715502569639,0.8201759489879772,0.16749468172715132,
 0.7963639781750941,0.05990683878951408,0.9060544035749348,0.3304537354745807,
 0.9210420669968795,0.7270117127461861,0.7730766525721416,0.9398399165714934,
 0.8208949390092356,0.672049681141755,0.8854050094972344,0.940019600797251,
 0.6353831784370335,0.16237809155498117,0.6986974003287052,0.4647052680678416,
 0.8386921111630706,0.5188193295959597,0.9166390803176695,0.3969232785776621,
 0.5045870243153433,0.5839530793286377,0.379889522979621,0.8722394874322319,
 0.5763133912550248,0.2791421672905052,0.2022628497640936,0.5956203111927862,
 0.35045778283196727,0.7108014701508679,0.7770453340663588,
 0.08744728162795834,0.979834788360601,0.5739870100887393,0.3661638482971309,
 0.10956783983679674,0.9701367502256045,0.23273154622818604,
 0.47903906975731436,0.5732102275220424,0.5866716532554092,0.8463646101593961,
 0.34373749825637323,0.3784390926242138,0.6332185508297361,0.4440941868221877,
 0.7967440993214512,0.6967638728666785,0.7641823295259346,0.4321047065902546,
 0.18362836947193006,0.13144306438193787,0.9035937101791416,
 0.5611323372151321,0.4266788785849489,0.4271377455683728,0.6529122558678289,
 0.5474998937957966,0.13246924884371492,0.14080660169654502,0.455610499695,
 0.37330914642283997,0.2243971272469677,0.26628610213807136,
 0.9383712334820536,0.5039823735652028,0.6004495083969534,0.20908772483612204,
 0.08246931132975222,0.4554198746918898,0.5724903446672163,0.7531181507841757,
 0.4694225582257521,0.8693015838510707,0.09768902137840352,0.3951817646771999,
 0.7094683514623017,0.8622345614137072,0.46196498127088914,0.9128281924561076,
 0.34094159544806457,0.4849661855731682,0.8520736583221011,0.5667519179394384,
 0.2034022746558226,0.5663447855943353,0.6292344551155784,0.7590043833623842,
 0.8158553204239564,0.5901170134348217,0.9084737007197397,0.5426628781489024,
 0.7663432346436886,0.7159836817868607,0.4607096847261507,0.9150166381515588,
 0.8836628294391975,0.486635082215213,0.9293931453152224,0.0876639062130593,
 0.7198671790473656,0.413834603888229,0.5021060609247527,0.14754224499771973,
 0.6318280549637337,0.45642103614471674,0.3038692787708567,0.9162442129627111,
 0.05076100994671906,0.15960319378841692,0.014141781385195062,
 0.9553298617171071,0.7247175863585058,0.029184829797107126,
 0.3976081304656707,0.34687461374107986,0.03099366879095522,
 0.48630985685686057,0.13506890168410723,0.4714325920918433,0.864083864055056,
 0.1458028957055858,0.009215040540630604,0.3379372963562839,
 0.6014478765281619,0.9651971762932807,0.11840504565512167,0.362379067227515,
 0.14150415155036078,0.41912897836006124,0.8264742836980354,
 0.9710260573808225,0.42564364436418456,0.7993122041435055,
 0.21824555140280777,0.0609273425701774,0.0238605836188428,0.6703174872506933,
 0.638992197192966,0.4008699482632493,0.8454608926817085,0.19558727760266348,
 0.03493631667852126,0.4743087559069542,0.5059943579706225,
 0.31628760905662145,0.8341825478831377,0.13008773902148496,
 0.25967468944955696,0.8543927910657839,0.005528196474828606,
 0.6546276800735706,0.28188693818896493,0.0497262298886717,0.6884719008643915,
 0.3548474150814446,0.11439443739261812,0.04965612418723864,
 0.6886666194403448,0.21362482973748886,0.6546473405325696,0.7835553287481769,
 0.5477868535347868,0.8918600691232884,0.0857821622101036,0.06796163153842738,
 0.9370481315099879,0.4343332277781883,0.13730423357000576,0.8923033209935872,
 0.540883559186877,0.36917707181126735,0.8746902166795002,0.1701278954132357,
 0.5736718125733804,0.7602848246448899,0.7495467997405486,0.4662714914702032,
 0.8106482998599241,0.35762664475967354,0.8981808716013467,0.7919799905862505,
 0.6272260505650297,0.7991145165611082,0.1810160998793926,0.7141534754507546,
 0.6713712246457088,0.5345641132558991,0.06760161211693028,
 0.026073309812264744,0.7300456892643249,0.6588021879882077,
 0.9070249001628493,0.5519564332589697,0.10029537294517044,0.8323635777496546,
 0.8270314427138891,0.23798784179915322,0.3989790408044642,0.7587108246937199,
 0.8696914750471746,0.547580678974652,0.581141684485407,0.6176362176220859,
 0.35986149379811416,0.5850154352923824,0.8673732119450056,0.5661528715925048,
 0.034406193375231764,0.6241992116358542,0.944236944368173,0.5667551438929952,
 0.23404139321644468,0.24237236074631763,0.368822337102809,0.8644215173245035,
 0.9967946950158418,0.5597133285369842,0.5061755883070375,0.06399422874702776,
 0.8303138510440697,0.9378398931121756,0.1431034435924471,0.3294867400140572,
 0.45297010214681266,0.7939447190651774,0.30155293907391856,
 0.8968455289758619,0.2770841156777005,0.5793177192551059,0.537674932870808,
 0.7692346002908769,0.8236711098380882,0.4169136128947267,0.9521103623376568,
 0.7408145541742888,0.4682939038250534,0.3045078932236486,0.9201413562898401,
 0.5558564333368682,0.34497213357424394,0.974926619042454,0.5500749140894493,
 0.6555648279820911,0.4305855897463472,0.8830223172154096,0.38301514219125365,
 0.18844819718583905,0.1942731092415536,0.9577699635439868,0.6133080603711453,
 0.6366418344281972,0.755575620201426,0.9836975486147945,0.19816413824278833,
 0.9405849990120025,0.16874210544419965,0.34165485440063237,
 0.6745966818762474,0.22954336894569827,0.4343193031425442,0.5347115906335633,
 0.22942043649406607,0.7350759541498962,0.7738018569825917,0.6800594938703983,
 0.13490719413701524,0.11257426923680847,0.9620052437932111,
 0.6303668449516181,0.1050492639946995,0.021356922496923447,
 0.6147677668536721,0.15620124430268012,0.3570637168590911,0.2569901190392798,
 0.13013034286232728,0.5497926087764953,0.764930110941719,0.15889634344151438,
 0.7672736035368459,0.8221451285306771,0.4744951930251705,0.7866231666118362,
 0.3503834351451446,0.5592598344963932,0.6189032002119284,0.9425740327481351,
 0.27156571151263686,0.08959498825898826,0.2882385555402245,
 0.6789161128664659,0.9804187036830302,0.297329137278766,0.8431905502538211,
 0.26760530482839073,0.5306941205110532,0.6205440630634715,0.7572348442796637,
 0.5086765704155758,0.9655139199833788,0.3771750585890531,0.44623069238958946,
 0.048624583293004564,0.9652640891371174,0.5987396173084707,
 0.4139053487471539,0.27809606686348254,0.6726270552155712,0.453192681300542,
 0.1345036047252368,0.025319832242704954,0.040005743240949965,
 0.16114310868441217,0.1287377286783229,0.7671234455652292,0.6542924076614864,
 0.6187125170098496,0.7744317655415539,0.04744943558462611,0.2524238107860348,
 0.7687618195070565,0.1522038724521385,0.14103586434518112,0.729861288894861,
 0.662938718836776,0.12757272367744665,0.8338156793767816,0.6668411774787014,
 0.12655284456534543,0.12997426653427535,0.8317474547058742,
 0.8390202125391867,0.8041383589210398,0.9724062830059415,0.8095340604731137,
 0.46550457010888535,0.6572150393777709,0.9669187335313384,0.4747931599717996,
 0.17468745885824122,0.4601212489914338,0.769772001817638,0.16396063320218268,
 0.9209112446202147,0.30731324147215455,0.9663150724065133,0.5485849755261557,
 0.5470909096065922,0.007197511727703665,0.854229239779147,
 0.46028867643551596,0.850390499495808,0.013845326736929064,
 0.8653079711252718,0.16224240887280328,0.6482083371894408,0.9156135287220203,
 0.18865134869284117,0.4372087968424754,0.11625679513152853,
 0.44180146936629605,0.6405191498426104,0.7651215213494433,
 0.20034952718037946,0.1784338747543993,0.9601603425780493,0.5272903844719184,
 0.19667212377568843,0.5388363646969142,0.3028533104481106,
 0.28333087197045836,0.17202532914937796,0.49411558593042415,
 0.44195060012985476,0.9506183698702716,0.06750504193465567,
 0.9271656733525315,0.6289450354523521,0.20431306824034834,0.8686113083930243,
 0.10504072208420745,0.17820405726488342,0.4866633994491928,
 0.2640314867778024,0.5302371612650661,0.21253920188896613,0.2822262776326053,
 0.0892880315998017,0.70296114963428,0.04324036834355427,0.97315751896896,
 0.7757879710800255,0.7862109243988494,0.49776695643467284,0.9071373820025105,
 0.42785303159949095,0.8986815066710132,0.22751156451696408,
 0.28588377497967843,0.2572537080033416,0.29082624019972025,
 0.4979755715739983,0.1567719356352344,0.2612251245517738,0.49327343106125454,
 0.5621197211818952,0.04503147141649666,0.6978225629736492,
 0.23344065203080322,0.27826884447532796,0.8547012360559099,
 0.4105481210658384,0.9912942609970343,0.9910343123588663,0.2524378826252386,
 0.004543260801538995,0.4600124210999915,0.7139297984788887,
 0.14108192277035014,0.45749244491802266,0.5692406518979307,0.525293531110945,
 0.5776630959463295,0.5080182114466391,0.30748765585174476,
 0.21720590412499197,0.254331028538877,0.17902742102226443,
 0.08164410149830759,0.25126356318245846,0.08937983872592314,
 0.8393212459196275,0.056714982166058525,0.3287360930114124,
 0.47891076258066445,0.2238711381424462,0.7518891925571924,0.8626177773248556,
 0.796358939207212,0.9025399288037025,0.41138137436654054,0.7181895160049888,
 0.6805235259026634,0.7459356495399595,0.34343456016680474,
 0.22677776429863972,0.3225462300528793,0.20835429853065623,0.30336668423404,
 0.9239570331549178,0.05086666721121591,0.8056427819133283,0.5314034391155089,
 0.7726884177121588,0.9190545244592133,0.1410850414235455,0.8950554085318321,
 0.6114485436089483,0.02100896486354109,0.43349292944730067,
 0.1217338724065894,0.653773134086477,0.09522869249345511,0.7493520869821526,
 0.43323461810644914,0.2854923695904097,0.6722929169825691,
 0.43733771435001234,0.6647084253199937,0.9746472793450552,
 0.48824383325982146,0.7094094087868009,0.6639688242011561,0.381411712545927,
 0.16823484940321942,0.9112892014540978,0.8006579727504916,
 0.40515323411683013,0.1746176451821536,0.47073360367659045,
 0.19306046757144446,0.6864279204215284,0.4776066404205165,0.9305289115863183,
 0.8907297595219015,0.9032054131488502,0.4409789236851305,0.43417616801737113,
 0.7845226766487429,0.842246481028973,0.2597584059495016,0.8212132441182114,
 0.3469518962683432,0.9260848507318562,0.9294323694178488,0.08975078961953953,
 0.9948748501369281,0.09535313423749825,0.6485360919702556,0.9954569582180031,
 0.08009069897873733,0.6788061386836176,0.2646870868339727,0.7757844400804484,
 0.3196471350990897,0.880397998750868,0.6392992785264383,0.18846623936619156,
 0.9318985516247568,0.06565065785425672,0.921957552950007,0.48057738602020694,
 0.15841238996849927,0.16693136347381632,0.9223141093054723,
 0.26063103103516483,0.241211259026763,0.4284453098646135,0.11495897194743443,
 0.5378782099527477,0.5394738730532292,0.08390386003094275,0.9845510178208315,
 0.7234647464735311,0.12702387459080244,0.026547412264174453,
 0.6537048422331475,0.8309794241222112,0.2271329169611519,0.18203830271707344,
 0.5700156605642059,0.5600273037454326,0.21113405414312836,0.5891738553291332,
 0.23089213134496833,0.25463381527262696,0.4520079878948742,
 0.09222951798336454,0.5435824900784687,0.7665414542113701,0.5484824405000338,
 0.9221625484769921,0.4119760342338604,0.20295802634866889,0.7306671574235086,
 0.3884181524035736,0.05942526525724312,0.772810896537273,0.4174806254562956,
 0.9583710869570699,0.4156564503896363,0.9679486308412744,0.962225797313362,
 0.2441042642430883,0.1390192927778181,0.6509097138600085,0.14568441173272628,
 0.02155451722617041,0.09062569881426485,0.022084658631356002,
 0.8276312391614642,0.1237416541221702,0.4231859855368959,0.111141802651201,
 0.6550003990004982,0.5812437839716302,0.4433072266738467,0.9169779539580786,
 0.3317446665975823,0.8277390327801488,0.5856024195424421,0.8195164655595923,
 0.7283424946179134,0.5087921870003562,0.37739578564974086,0.8800215816540505,
 0.1429613315362932,0.8567301168721746,0.03867430044254283,
 0.22753622703357868,0.22259994706915054,0.13071990265944888,
 0.8821932072772111,0.8662608109255276,0.5523283046992244,0.44033186899803645,
 0.5842825098158235,0.025679609130772052,0.8661309721265564,
 0.6127407270608956,0.5684203557629551,0.26150709266365724,
 0.20459683120006533,0.7008804138374761,0.7158254968921572,0.2907732264345161,
 0.5737657657363258,0.1967266483375414,0.4317939559051813,0.19558160877087505,
 0.7813887603153364,0.6986030042477027,0.39289889875681494,0.867117930875347,
 0.22120781314118798,0.1517743346109709,0.08691046839088679,
 0.11010073291564704,0.07078058897594264,0.5155849850576196,
 0.3279579523552959,0.18009688013722902,0.6850866579067381,0.8697381881027344,
 0.5605847109352895,0.9617192946027874,0.877441744332698,0.799692881450166,
 0.4858198381747092,0.20049042221051783,0.20643694525198397,
 0.04492066504792014,0.5713832345191194,0.7405703184185879,0.3064228403995165,
 0.6227347704436976,0.019416684102901538,0.26104743503010563,
 0.6468130529584527,0.4581931716188422,0.8181815135436872,0.11526747382981395,
 0.4436445601447979,0.4981500857327271,0.24007804797643117,
 0.009297049053735296,0.8441829564973075,0.5147203341127707,
 0.09790764805135754,0.9288481914101117,0.9962098675104898,
 0.024173591177836906,0.3736089045860761,0.5563453927701643,
 0.3344940889557225,0.09876188968375399,0.23316292721809717,
 0.9867436572239847,0.03632903145978805,0.5354265680404289,0.518188456287397,
 0.4485112273425276,0.7543724272162179,0.03905085628894034,0.6773766267800588,
 0.5759253621200381,0.5037612094045748,0.2553317445978672,0.20419237510315924,
 0.5265773161773925,0.2606405253015155,0.49577067862665125,
 0.09764610946327434,0.9196123420512425,0.7067959500106185,0.3904120554531555,
 0.022458920496894463,0.42895749082742285,0.5790525948722522,
 0.5370427631605189,0.24034866686578593,0.03287416093123108,0.6880825906441,
 0.3615221468742609,0.8187292258998117,0.9778190161503935,0.8365401571406933,
 0.912095343821042,0.6490123157068369,0.510799635851401,0.4452984572034687,
 0.1234717904530217,0.9561333307540891,0.23718010788696775,0.2770689570605055,
 0.6403335832182289,0.7737874650738724,0.06348868870393776,0.5136755613647477,
 0.846938421069252,0.18242853858361108,0.05892401136736569,
 0.06536281464872784,0.8790697134419253,0.3413148899499402,
 0.056714025010461855,0.41912910545947757,0.3688409702051241,
 0.7342004758491858,0.4739860725114785,0.5546538075051956,0.09024258818829112,
 0.30788304289845303,0.46674625491473554,0.191094132144231,
 0.14848558569860548,0.8751070511473387,0.8959528202149891,0.4691460691316024,
 0.461657555459669,0.9582236221900092,0.581509032230018,0.9218494004386941,
 0.1825497352392812,0.23432843000915582,0.6500078637833386,0.7358565584103567,
 0.4539501688674954,0.4582784644872988,0.32196849118001036,0.5676431261474498,
 0.007814117913293672,0.5880649709743059,0.9074480636175417,
 0.6830912720889243,0.145597437503872,0.533409924050312,0.9954643754268491,
 0.1025646880354466,0.28146253316981495,0.11963005646227964,
 0.6534326176219862,0.9169857302678973,0.17093199254044344,0.6040461169159906,
 0.09613858218095195,0.4589632114948715,0.3825310259767726,0.1642721763881878,
 0.005560497645600515,0.32403168285787387,0.7644842525797174,
 0.14245563637356085,0.12967911534550358,0.1354721680836588,
 0.19926725979537185,0.08414121822486953,0.20255671860928715,
 0.6618619540521171,0.13485441503687667,0.574020136746345,0.5000283053398198,
 0.5968469879512921,0.22992649060001813,0.19857494545947296,
 0.11088829809395917,0.26596115625639083,0.5939278803628827,
 0.8880174775456231,0.06488221006760742,0.6665964838869238,
 0.051978946849199126,0.33378355821156225,0.3694104870007341,
 0.6513341018668862,0.816526226039363,0.2346239108250372,0.465312539388285,
 0.22284518462397995,0.4816810183779682,0.6968998076698729,0.5184742447273736,
 0.6166297563779771,0.30524210484216263,0.9862867794173067,0.7138512711696112,
 0.48806492276296676,0.6183108740239367,0.9804011686060502,0.7218570980396144,
 0.8066786981069622,0.8776936310936206,0.5243094410754379,0.15796692986475036,
 0.5215790121640029,0.20728080493040124,0.7800118728371136,0.9561444020894562,
 0.1539026145560939,0.16907637423887878,0.5074849464551968,
 0.26938009177519207,0.2223534864845892,0.46690612213647475,
 0.6352394112293586,0.5788205392967909,0.5043583941270053,0.07685356603872573,
 0.886399700976978,0.7910192266936815,0.4299070613124417,0.13084450977419226,
 0.6278399510643744,0.7886921956513153,0.8410722952440555,0.11863904906612188,
 0.6089177558387007,0.1907911303026607,0.07144666180144388,
 0.23286353158442108,0.8774740250366004,0.14385941091826138,
 0.30378159116250036,0.4587219501483991,0.5968253543733436,
 0.031246174435927365,0.10423473557626295,0.8573671607361404,
 0.44102870908953407,0.8440498924087918,0.2197879872377717,0.6786533223083341,
 0.0025525045739381858,0.04229636287492722,0.5476102274025877,
 0.20472720892852303,0.7456475404285349,0.8169032639182863,0.82387712342756,
 0.5743811539700736,0.888614208393533,0.16766977657614945,0.2665315277623469,
 0.3918866042126443,0.5600894239032325,0.8824207414528757,0.5498873857592044,
 0.2210832733670235,0.057418240696181955,0.3685290544614448,
 0.9120461094528461,0.6279509829658351,0.611119713577799,0.5592429057929775,
 0.4508883714935056,0.7268553552115172,0.9503340960192919,0.771825485267612,
 0.16548556360236077,0.44069971188229295,0.801587873291656,0.8103607853455075,
 0.5275323950453737,0.05712993009286649,0.1762029569557374,
 0.12043115652703018,0.19469166951941952,0.9393228422122917,
 0.7020770046710209,0.699672207976669,0.5109407370785194,0.9299414728009274,
 0.626930818589584,0.4109143539433251,0.598516625097588,0.5678385057221501,
 0.8357127269543567,0.6866489021565195,0.8078249045613678,0.3764514597836287,
 0.2546544516985745,0.39340820592542547,0.8541427670356061,0.8460623750005771,
 0.41896638435361966,0.09851892564262021,0.2488742656422861,
 0.6652460063719866,0.5522379935246425,0.3033322747001679,0.10790000387185472,
 0.4391879822447049,0.39915912769674944,0.22749809911423513,
 0.09639484331708725,0.25314348116431074,0.8863638169893261,
 0.048455911558824116,0.6369698412307856,0.2926192750999854,
 0.23638204225801418,0.5543821496607275,0.20330342322198902,
 0.22327420777747786,0.07525704141565759,0.6578577170563753,
 0.6082168636621982,0.20659664856606996,0.2610610097301984,0.859390333646962,
 0.3496704544935342,0.42748161987706845,0.09269909513521868,
 0.8597169107555204,0.6034261815693585,0.3664499565834648,0.2639092278139039,
 0.11345326179641069,0.15083410297820743,0.8222383899779587,
 0.6513695883897015,0.13049670502180455,0.43986762140658875,
 0.23892945894107287,0.13848554335527719,0.6668343452207695,
 0.5064617629356344,0.9710440018449187,0.15785901053067375,0.1242004875701288,
 0.022972450104915665,0.9229754212045335,0.838097145099508,0.6484247888366068,
 0.8829917525033774,0.7832327496708433,0.03718247697622368,0.9213733319746398,
 0.7323947557563009,0.23019323688794047,0.7346170719399017,0.993299815331135,
 0.9454342275202521,0.6982392216987079,0.9031585825242804,0.04712002441342611,
 0.35445977657422534,0.5037168019282666,0.7051146757323511,0.6087374635627456,
 0.8310781792899153,0.6101508992829477,0.0348936194982985,0.26231309445952444,
 0.862299140044262,0.37528407156962096,0.5178554660205926,0.1016429085702415,
 0.4457441781088872,0.723972859609479,0.866232307533473,0.9266190413141733,
 0.7212512667454898,0.7638013377687144,0.8987351893902412,
 0.058600620290741534,0.6811727059826467,0.2119430598935388,
 0.08850776512785186,0.4812550366743692,0.879393480905157,0.9446743145565628,
 0.8598029870919057,0.22511625699342763,0.7934328156054722,
 0.008300393746581869,0.8705776481242895,0.06878850748788468,
 0.9669058675338138,0.9760283090943315,0.3028216413646462,0.481439690490985,
 0.8067892542278494,0.9249975984246284,0.8470804046538731,0.3107104317683207,
 0.3668567746809068,0.9526226931686975,0.8154285364518423,0.16211633466959507,
 0.47875999888921916,0.46885754564394355,0.9298944724833643,
 0.8826681378808108,0.4143569915499743,0.8932583300635812,0.13923978378438984,
 0.23018828285137904,0.1636636494061141,0.937207752367457,0.6979831446151199,
 0.7202737458894605,0.3120469433284364,0.8209921599809397,0.6442954708089029,
 0.19433086699542645,0.9725722528274452,0.037099044711408125,
 0.012737969759574197,0.902861318421931,0.9856994373139845,0.5976305843605172,
 0.2448627706137606,0.716252496567858,0.23426077406494505,0.47282410676018893,
 0.9444637858740841,0.4502053653579192,0.81334288037761,0.9977740661745413,
 0.5593095723057504,0.40076524860046525,0.26489406805453997,
 0.28455306398909885,0.6642622331950649,0.12918433358724868,0.905311686772688,
 0.17001577070498786,0.7544374130748596,0.7145695511577251,0.9523230075385207,
 0.8818840500179949,0.06024252198921409,0.0879710417519809,
 0.023175926226023824,0.8689465245562129,0.290626298545717,0.695834169411512,
 0.017654382481016362,0.4798454458635464,0.5289084842295968,
 0.33624070737374134,0.1536427932426645,0.14905658469181848,
 0.7812605558209146,0.5192698211042022,0.17024376549148101,
 0.06619063184731688,0.8960347130116005,0.043895821191479056,
 0.9810304828094116,0.2699772147624214,0.09427898992097694,0.4350314357948042,
 0.9977135103724473,0.6895245819568008,0.10885733623235327,0.9949820286623707,
 0.9081816425189114,0.9207442896921115,0.27484355265607807,0.5721324027624509,
 0.40156191496065086,0.4897029900527671,0.6908774171683733,0.6931441625894345,
 0.16229975340227676,0.2854125373755877,0.279122911249721,0.9088779240568182,
 0.7907898832996847,0.6781237000250822,0.673535409923067,0.34162001104555784,
 0.9550281355507245,0.08595714239112529,0.5178595838681811,
 0.43295626131261766,0.9965089570368227,0.3014680846336151,0.9427502627808801,
 0.007472919861263194,0.28332484330296703,0.8098129298567219,
 0.9713384393899145,0.0033686559579370057,0.39713002001374476,
 0.4259497513323681,0.7987883019409234,0.1368754811829147,0.5545918433330426,
 0.8244720773411299,0.09523585636397403,0.36148593884576496,
 0.8226891040306233,0.8826574211867875,0.034321800304563665,
 0.2884468731761686,0.8836924266243509,0.39197606540773966,0.5770083608899512,
 0.5099994910123852,0.4390369995101998,0.9834836001296973,0.3730093440220049,
 0.8502388661140257,0.02805457169104375,0.7187339204981206,
 0.19355409854016914,0.3629876607819482,0.5972892168405043,
 0.20962176945865707,0.2007647877631923,0.3183271364884359,
 0.43637474367423923,0.5445061729958136,0.8201906841825546,0.5893083133495048,
 0.45191837751892916,0.30663742907032765,0.6175370057046591,
 0.5936849771018995,0.9921988615438533,0.8424486757400751,0.5770872232225275,
 0.6195923645623654,0.8180290976762348,0.39633291677725246,0.6054188228459179,
 0.13353964049570755,0.6173301688004231,0.07050790272083063,
 0.09516931008633678,0.5256646246610013,0.5299495143715152,0.6021758349796933,
 0.6316437326430745,0.19743233707477137,0.42663672462053526,
 0.24219248444352715,0.36188544871145656,0.35744066421939447,
 0.2875621050997603,0.4619137081137691,0.5200715753060043,0.6361255957471096,
 0.6414199727606702,0.6219630301189314,0.8377346854233038,0.34014268758757005,
 0.18753410140050086,0.45593300225180966,0.7836552456548767,
 0.09706427250744598,0.6284972535801939,0.11756663450198834,
 0.4694223919391458,0.1361235454873312,0.605257595392853,0.26218089467880334,
 0.13292956200869077,0.977935342015164,0.9588258369737614,0.5722384366742154,
 0.0479876062948692,0.9386235164685137,0.3546838083507532,0.8861598752287952,
 0.8642248301903602,0.8601815424481432,0.9898147443486457,0.4422365663734592,
 0.14224515335079624,0.5384608340487065,0.4306760708638706,0.6641387041518519,
 0.686243271304078,0.26581146584598314,0.9389838045114647,0.6214572439545492,
 0.13175062236703783,0.5922180149325338,0.9660109700422338,0.5359386979936296,
 0.8299937082293997,0.19889037542051213,0.9965703890183971,0.8805499219477726,
 0.6861583325581043,0.9412792225580695,0.5148947834704648,0.4312749125818255,
 0.46059483790253053,0.15983046090280906,0.28195729559639915,
 0.6164830186233029,0.21710565456571007,0.3492225718726881,
 0.13659477995012703,0.309612106905082,0.5440619737444591,0.0724529599230268,
 0.8180154191455001,0.7888157781723982,0.9203424675748514,0.19730012006937203,
 0.8175053342566784,0.10740953024795252,0.2612541331334163,0.6232362844217014,
 0.8041355848415415,0.4135734693661024,0.6769821327212076,0.17884543968977828,
 0.9247398296872432,0.5487829757887581,0.4899095404445071,0.8577964943912527,
 0.6855927158448532,0.7099430510545649,0.6556323966268947,0.5495065424450549,
 0.4838518118433094,0.2327349730370727,0.8771311719296542,0.725624514964971,
 0.873600255771036,0.8362254387957044,0.3583833905168392,0.2946671417417078,
 0.35111746001772304,0.5442095529882546,0.4305544893701183,0.7412881440503878,
 0.39467604146299773,0.8362617018743215,0.6439008099954097,0.2597358439571513,
 0.3918755478986444,0.47957721141461074,0.2963101823029042,0.1396820567581999,
 0.7148596554152556,0.8620800522496268,0.09651939100841112,0.9242808816081993,
 0.7082885524740858,0.6697136528433616,0.07381769670668015,0.6011801601929393,
 0.6191774031520194,0.46862740015262183,0.5850739342449235,0.5246064334313618,
 0.9230305341734915,0.45194356263680313,0.629305501166912,0.1248700945420248,
 0.9517845956207115,0.9951226384641225,0.003161173106999371,
 0.6573207509866387,0.5615956254465861,0.8513903779648193,0.0689437678153465,
 0.8530228912974763,0.6102884088163879,0.5313071773905914,0.6243637854089301,
 0.4273150085896049,0.3956575143530425,0.3550568556938385,0.6879287492985997,
 0.8185151367340264,0.9404196580656933,0.8766963708076343,0.1773876246474133,
 0.42961904073076407,0.9439250303709531,0.5999006864794486,0.6768884455205111,
 0.19437864161672036,0.14338591449275762,0.4137495471760939,0.608280637496625,
 0.742159591248032,0.4813909297398533,0.6434574080867094,0.2907916143324695,
 0.3446534107669973,0.8185433214156982,0.049155053573383234,
 0.12671912528012974,0.2728898710463702,0.3676701824367751,0.7265860120155757,
 0.43961154035814365,0.09423519978588979,0.9498262764602611,
 0.5036409533932142,0.16454052142649656,0.3294252405499135,0.7944504460921015,
 0.6919798587797381,0.803129858631229,0.5686099710825958,0.14511237281387368,
 0.5250465299893063,0.0446126554495101,0.547166652643061,0.13131156989760362,
 0.7151940927160831,0.637625396627234,0.7500377522569414,0.6844852567692091,
 0.43406548605755824,0.09612088423657483,0.9722027115438816,
 0.17041140553550316,0.39401271926436,0.7718791827161678,0.8972348390767677,
 0.9350793408990455,0.4566274020537293,0.05776992220314836,
 0.45282649858457025,0.12769777042563735,0.20418124330915666,
 0.6117304839472816,0.947032397487013,0.484908310740909,0.7293291887933104,
 0.7300931622968458,0.7190887235337311,0.4988287576052536,0.7512911609164891,
 0.24833392226804796,0.7687776312923985,0.6484229650452461,
 0.18772374370377787,0.8072469055068384,0.6704992309809755,
 0.19895199155835563,0.6353030219771776,0.510815709226372,0.9828764768447669,
 0.024894059998268547,0.05265647194764522,0.14612264579268408,
 0.43027795432274685,0.350145672302419,0.46761858897331554,0.523223026758354,
 0.1697001603830791,0.01524855991380758,0.46595474886675103,
 0.6760319696819544,0.01613533782104193,0.41199003398331,0.33796400888001965,
 0.5874288757706482,0.00493161484840321,0.06047147493759797,
 0.7134467278721672,0.6502350376623641,0.9121044115543854,0.8219029881765455,
 0.15907531802951425,0.856016483497815,0.5409932361794222,0.03658176614504183,
 0.9224128074375233,0.6266001026670025,0.6517520711613912,0.4379039461487362,
 0.6836144709182389,0.8813304457795799,0.7310264180670161,0.9481535650243335,
 0.27623332699783165,0.13411884071237967,0.46165035625790696,
 0.14448781491160367,0.4262590537652916,0.9292627605545303,0.8667867748871774,
 0.5062638227012255,0.009039365697892321,0.8188410109642986,
 0.6787879749914428,0.6642722466651239,0.5730561345761999,0.4228609839276315,
 0.6125664380575879,0.06412707140153207,0.14980782016638738,
 0.9204751077263754,0.6873048845379333,0.602262010366546,0.8632414859429436,
 0.14634648894920266,0.639131619909985,0.6108875420577787,0.6292358738851599,
 0.8869191294866161,0.8199546965660675,0.9066243437961375,0.6468152589630373,
 0.7902369201730146,0.8830609276698621,0.10236989603308488,0.8284732031058328,
 0.1332289946173163,0.5796036335095267,0.983081331660125,0.4398385318192971,
 0.5893697057714488,0.8506995733491914,0.34419744867720126,0.641923389468381,
 0.5568839209848463,0.9046179020134889,0.910379531005011,0.4038316889013783,
 0.053467492238185255,0.9320934363053877,0.500626338681125,0.1900886540443223,
 0.5972600605658503,0.5456667051218216,0.685061188877415,0.35277283905720047,
 0.3519367682265654,0.3805950731606571,0.1495987742352527,0.1459428353975648,
 0.7060968107639023,0.45265066535043785,0.5209393354117416,0.5181580383048884,
 0.6720855468509694,0.5470564716256675,0.47057630056888833,0.6195728866352458,
 0.5809221868231025,0.9316696178335522,0.36760964002036733,0.941191172816628,
 0.7048057910679378,0.8436734075305841,0.7236457834968357,0.7536271828049632,
 0.2926405832887662,0.448725529558442,0.09546881784116046,0.09721401797354667,
 0.394250251720061,0.9615438581225456,0.8715240909180385,0.16985178091951414,
 0.4178708652314681,0.43714635070521835,0.3528935441140617,0.7363219013652141,
 0.46477872884510774,0.43293978751290707,0.767637091903023,0.9601502928260585,
 0.23279385692440902,0.8983721664500279,0.16076916117550522,
 0.9073354952108139,0.17616647116667983,0.6694284453576322,0.9402025017230011,
 0.39045330248480725,0.4698693347227868,0.14220203185736102,
 0.7786493616330609,0.3299937030383173,0.634244250053227,0.6099092287638743,
 0.384218612841174,0.011307930061436489,0.37668188660211477,
 0.3507713530714931,0.18179111543871884,0.9285716301681138,0.4028941234027609,
 0.9536040039956473,0.5591833626479019,0.7677811269282776,0.5187508774487364,
 0.1977734055169077,0.6396029662546786,0.0466234702063566,0.6625381414114038,
 0.8157043473109695,0.0881926559520324,0.7150115122725477,0.16359161353324214,
 0.309659281989799,0.014980487445746316,0.6398826918649457,0.8368190133902663,
 0.37800894545506947,0.07283399592024464,0.07326657802228376,
 0.902044257184798,0.947117578906941,0.4424456224671789,0.022407955738046326,
 0.42055246770307264,0.7141591986622019,0.714234011563571,0.300085081338029,
 0.09957341877066506,0.013163868586006489,0.17382268038127813,
 0.555532265681895,0.6829904373954078,0.7522034776183105,0.28400756284377726,
 0.6116478926929702,0.9759289471577173,0.3057676821982238,0.5424817417519278,
 0.17316232360992811,0.2322923532352652,0.10745962781139662,
 0.5705795363350612,0.05382644268625747,0.9459856743105911,0.575536536423483,
 0.4690967302382696,0.34586714705083976,0.9806291767759641,
 0.43912563273871563,0.9963417701339015,0.7569276347507843,0.2083564007928077,
 0.2214160634128972,0.8555193923583297,0.8626535187058088,0.04426178870440567,
 0.6577715276751486,0.8430235291789305,0.33281602723491277,0.6293163261343826,
 0.22450048641309917,0.3533134526608849,0.9418856039931113,0.7231064291739117,
 0.889339070449965,0.024076408987941278,0.4190191120637762,0.5804789773720613,
 0.8023974519677334,0.4602640318237472,0.16077473077237436,
 0.48539039757702696,0.9035951184778609,0.3949201949431437,0.6133112154903192,
 0.1561063372127336,0.024584196251203116,0.7345749636198878,
 0.45928459400305566,0.6145623160730808,0.9403412185193453,0.5315041631014371,
 0.2430854270888485,0.8020872175713719,0.7364719392096488,0.14496772757798626,
 0.8384498534147891,0.156085253232087,0.7030687433438815,0.5934584987068096,
 0.5520874139692978,0.9412592188030343,0.43064604209804436,0.8148238267718702,
 0.7114599977652217,0.7539058020567805,0.3698407156730812,0.22409028118974128,
 0.6787719796618528,0.6486856733851765,0.43860711707193745,0.6873826470973512,
 0.9557320221562756,0.9982193936076919,0.8907790710464866,0.8020385317373844,
 0.48699836155307263,0.7111284799600449,0.7387404722365591,0.8607546091827638,
 0.07701246242339055,0.8577290242573118,0.0747415958962443,0.9926863474862329,
 0.2763899586516575,0.5217066257616487,0.017663234299563957,
 0.5542331913050007,0.3935461052122592,0.45634174344377776,0.973348456812973,
 0.6757789794261448,0.8730762275779514,0.4911808072676431,0.362792065699006,
 0.9997006583788438,0.9110035101505267,0.5583293039439763,0.39044697298939046,
 0.39804479506775625,0.31896492812570565,0.68148720154793,0.7629948557793251,
 0.7513815963212578,0.7989706610200327,0.6328215111745061,
 0.055283399012981405,0.9815472334252918,0.6501962508511797,
 0.4972544986010793,0.6313918660722954,0.1609792426888682,0.11253942660439298,
 0.4646171455894732,0.5517472895045209,0.9071283725051216,0.9572677384454478,
 0.1731844577504842,0.2243450569904022,0.07493588515821203,0.4395464700698868,
 0.6447376345182443,0.4862630309425551,0.7149309332398902,0.9225532130023919,
 0.07217619609346859,0.154057395544971,0.2170171168333377,0.06489820359944165,
 0.5998088753444483,0.3847693013069312,0.8034096806988436,0.33611696162965965,
 0.02674393354856308,0.1877515253421408,0.35858305167715065,
 0.49405001279761074,0.0839762242055846,0.29014332017786726,0.869990738294263,
 0.107729953728109,0.47559562227687135,0.5048237239059469,0.6299509702222258,
 0.05708679677700146,0.40768922471923075,0.27195135942037585,
 0.175035914505556,0.2404910134854258,0.40979781375915536,0.5657232724405923,
 0.569934615504117,0.558313582113479,0.41440620070708556,0.5580897944231128,
 0.2015718149209038,0.744793452699019,0.042549724943774736,0.2888137477110745,
 0.2486999215227882,0.102209344787623,0.05811339444295016,0.46923738169157003,
 0.7285548697008979,0.08260150563712831,0.4651937518330038,0.2965640379190053,
 0.6515475568863671,0.25047987366992186,0.9864014789256222,0.4569116850983381,
 0.8504929543674409,0.16635476267198845,0.42436901478567646,
 0.9541084814667935,0.9965148467274373,0.7262238984113977,0.6066961005532473,
 0.3889843101279966,0.6104674686298223,0.10887824142064872,0.9951761886361107,
 0.1971580412553816,0.8318454983749708,0.5260093188084738,0.3294446414174349,
 0.15146404791005041,0.16028665937491426,0.1728242023655424,
 0.6485648922163416,0.4573847928426318,0.8475012507147539,0.7715315604476218,
 0.01520481816727648,0.389634800107877,0.23048227252465447,0.8749603367876594,
 0.8619395261582639,0.30188532871440066,0.661204992433279,0.975854585878286,
 0.49181719437137805,0.725911188293778,0.12704534962844827,
 0.22016591552093778,0.928061519112185,0.12804531670930164,0.793170508078387,
 0.7238188505327054,0.41421185304220076,0.7374302608635028,0.5688354191800129,
 0.661093784331128,0.7798993366617869,0.4824075731901347,0.47453427011226923,
 0.8075414378814137,0.8861386963258189,0.5937913143458468,0.37085361593330934,
 0.785326152506481,0.46106158216330717,0.3877768757575579,
 0.037480687980667726,0.37858235933910245,0.5172318253653974,
 0.803703997743011,0.6180597132210024,0.028813775576625678,0.1584380091676132,
 0.6244861509008841,0.18060650461381234,0.24452158470766705,
 0.08073235507212362,0.947037372751911,0.03053959851744903,
 0.22907213639753754,0.5145948500052889,0.3571172110065408,
 0.04904674413819565,0.8385393353893126,0.6132970840700035,0.6953994256531755,
 0.5045074527163863,0.5054079449145725,0.35716373371162247,0.5866259317642235,
 0.8587781469212706,0.34828745505728465,0.28796642677225637,
 0.7579975781210976,0.07741809921285303,0.9744180449772291,0.6985610855730319,
 0.18191682965075007,0.480304900467725,0.011979732018594924,
 0.9270640066557639,0.6540710604497514,0.8028574992489825,0.4624282675142759,
 0.935742412484955,0.7798548236053722,0.9640521184532485,0.8429607135352386,
 0.17168257535652576,0.00864685761919659,0.23257224944622412,
 0.5502053495027762,0.5070910807797568,0.3496059506737158,0.7670459398732509,
 0.22166315507114642,0.14488286257053673,0.609227035558501,
 0.007926079974476363,0.9566591800159767,0.7209341460412316,
 0.29402230196437673,0.8528075440393171,0.33621899562542856,
 0.5230393758464351,0.12997745874902522,0.4633023245278526,
 0.25096944908434526,0.8733543719835619,0.07438349796640842,
 0.24285540563936647,0.034280935527939205,0.5759706037226731,
 0.5562740978126526,0.08999564662487758,0.919907805330661,0.1995497855395476,
 0.9077630704123926,0.14181793793771102,0.9348356466208856,0.4013128387553945,
 0.7627358640215981,0.8426510942106186,0.42810439140874,0.38257767041993795,
 0.6810721807386938,0.9154347087716123,0.42098457932424505,0.5257307351204147,
 0.8936347129570372,0.43232707543548987,0.08667000211996312,
 0.5525694994530258,0.8392118911570992,0.024433683303912046,
 0.9252350110854359,0.8854332299827186,0.8191376804833967,0.5113911543281469,
 0.3466957935743402,0.8837269797904272,0.3638796675811866,0.5387334840815569,
 0.19838358212912055,0.21196830993164906,0.5782295416658269,
 0.07756030351987842,0.08944939041767175,0.7393112452736389,
 0.002400760662720991,0.059786739717569204,0.5638308723259512,
 0.3822358674134958,0.12717912633860928,0.2835597461609193,0.4660628441733009,
 0.072051291833655,0.23403640254561808,0.14999754032730184,0.4803737918363523,
 0.6154106203497958,0.6351712303123476,0.11210275594480501,0.3144643227441123,
 0.7506235214898964,0.8864529055232933,0.641924523599873,0.7870380343027674,
 0.8931837775238507,0.11944911632099597,0.5341775752596143,0.6335489209654407,
 0.18223141253005715,0.7041883051954101,0.3924028490848268,
 0.24925302921248949,0.8943166524129299,0.16768839838907013,
 0.3875993118731169,0.00670249545227275,0.22493749313307865,
 0.9585062884113644,0.9186295579276206,0.3536534060190639,0.6307451585406705,
 0.10589180678303689,0.9291923200942663,0.5214471465558003,0.6254899547870132,
 0.4350667303374781,0.5819270115263656,0.13609561998712505,
 0.14147804066926395,0.13316365540285324,0.542147691118801,0.3144565432652031,
 0.13564995024265625,0.05205706208937111,0.26475822300718965,
 0.8242689379077847,0.49404717116261243,0.5358822910684476,
 0.16212664169933577,0.445852286801796,0.6950222741885275,0.47506420076254563,
 0.0219601662170561,0.08131120223937949,0.041079793293674505,
 0.6989197136372269,0.911310414602037,0.5727012591865661,0.9905062174897854,
 0.3930957236047029,0.08017565087654266,0.002944994035582016,
 0.35849438474854245,0.8943678383978078,0.22168049419817704,
 0.21509585441693901,0.7146668671112476,0.6744853596913516,0.3495870251197416,
 0.4072731811464436,0.3649300259818813,0.37239315581228394,0.7816859360603171,
 0.2149804244743435,0.7357296192330713,0.5684891364252807,0.6947342131115383,
 0.9935367445476531,0.182265572312861,0.28093754595422826,0.6088813783042559,
 0.7427019005171048,0.587367656384851,0.40237265125842536,0.7476408486731152,
 0.43454720374836064,0.5528416536053242,0.7784870852661114,
 0.32190679348951545,0.601811131415724,0.3196933031442092,0.9162023430822216,
 0.7511577627401816,0.17195957240714366,0.7789134184913273,0.5561025735930389,
 0.6900238375035928,0.4139259933027295,0.9048731034470587,0.8122596596906466,
 0.6439154590815138,0.7936530546418163,0.9383086710124391,0.08034635976612747,
 0.5747420699603991,0.01434880901856328,0.05401750064745536,0.591751057053258,
 0.005172463711355602,0.7964516994247545,0.7445059890816833,
 0.5210340862351677,0.9443071551823375,0.35167354313182486,0.6983223610344372,
 0.6852502504112745,0.41425720331132854,0.16604726803464376,
 0.6096488087982085,0.6059515300963934,0.07709872145232588,0.5516386064277824,
 0.8399892538952031,0.4145097247216003,0.24129989902404858,
 0.23991244736784412,0.1728266321970625,0.6629536402468185,
 0.34085158190173526,0.5457874480360939,0.5961864365557094,
 0.023354949385220092,0.851778946682915,0.022420586107725904,
 0.5422994289088249,0.1974299405368557,0.46718533380990435,0.0562906636109306,
 0.5235880973917386,0.08751180260099733,0.5819061613284972,0.501563860471243,
 0.8545145599728059,0.25463964601343836,0.888121774184003,0.6849104050206906,
 0.4691301311666378,0.24722350300730112,0.7318177437558753,0.6240626515858694,
 0.1792342030402856,0.2670172047676127,0.35250614425981597,
 0.17964720846795557,0.5606664249289018,0.7971763922984224,0.6060198503804147,
 0.0999132916766533,0.49770122252386484,0.6041259691032567,
 0.07024181126990658,0.13894392702370462,0.6007125046615867,
 0.6326099402634803,0.5187824563894498,0.9810445012448166,0.8305815237408589,
 0.5056291539832254,0.6993872104722262,0.288440518917771,0.11439447565200644,
 0.20961777448477004,0.061546298221146434,0.3446573083856189,
 0.44389338533268186,0.6327307804303299,0.11265949712807166,
 0.43037562445777555,0.730436060756186,0.22349620089054045,0.8919091998401096,
 0.11177621748993694,0.9390344623367521,0.18856070400175418,
 0.3163901139102664,0.3431937847589118,0.8079638564462187,0.5382533022309375,
 0.9079892878087258,0.02659980981450083,0.528443300378892,0.7861200871145594,
 0.4325466599178013,0.4683266705449284,0.865807439022998,0.006868815457862887,
 0.49501296523612626,0.16023562308327577,0.0041589056405901825,
 0.00944131712734464,0.13624640464468307,0.7344941431249741,
 0.7716514397954829,0.6684094933329782,0.2452873608160211,0.31264270241842507,
 0.12266933233287503,0.324523993509476,0.8818099548353063,0.26574132597491296,
 0.09382725629430677,0.0048865056403641915,0.10710075145332476,
 0.15936048300989103,0.5053192400260232,0.5243636457522503,0.7083221826242827,
 0.039641796525815565,0.9806114602829579,0.23384981409274164,
 0.7755724804211089,0.5264799045773966,0.19067005049102947,0.3915085626570185,
 0.7196935318865757,0.22448987898011408,0.45286847673018327,
 0.5644274488287233,0.09678192860772983,0.4989178895155715,
 0.33233530046567894,0.7114581254354033,0.5047562920093871,
 0.04989327100491592,0.5285937510845881,0.4847501614420061,
 0.23012833265834132,0.6064305111879471,0.6917873193750858,0.2758390938190488,
 0.36980685392424517,0.401318873823892,0.7639040531072083,0.6704472926635053,
 0.6224984713497442,0.9929828863645294,0.9966591118614677,0.7121297038575449,
 0.06148680848283661,0.42622971498134743,0.4193696570683755,
 0.9417144063507017,0.13278589759632164,0.42666759764677253,
 0.6030911039944411,0.2455380553613813,0.7363316894280556,0.4851285748151435,
 0.6017762597138943,0.4551191382538964,0.23760741565850452,0.9185770708625021,
 0.43933594867175363,0.9474810437447556,0.33753031018282065,
 0.3309502677777072,0.985327062743431,0.3868930089454834,0.4837276272277692,
 0.48342981750876046,0.5897158060978226,0.0785875071761094,0.3886340055427643,
 0.0824200538692007,0.5722595607574895,0.7550040547547816,0.8786810094703734,
 0.8436512337195025,0.6639408753368294,0.3457024339821767,0.3529385007275554,
 0.5322149397990098,0.05503077881690466,0.8535228808580393,
 0.20202800595977832,0.6762215466848225,0.3381899670153805,0.9502228573836812,
 0.05521863514816272,0.2587534518403255,0.0337971080371936,
 0.15570301382945373,0.1847766319160281,0.03905917752255117,
 0.6282332394045368,0.4647595818815169,0.9157477065068175,0.3763582158146733,
 0.7979856264464998,0.41617007633137126,0.6561089241517764,0.9498640768470007,
 0.3951294860144784,0.10368004568462841,0.4309739562803434,
 0.024365915304532137,0.3182535901615887,0.4479237370622642,
 0.3294475392165914,0.3682367269519671,0.9407964341542522,0.28668666832858236,
 0.6121634085746026,0.6060408572068303,0.06444563767082778,0.3949780942968243,
 0.8646348799797848,0.9928843571179182,0.6324326313926418,0.8081584154401273,
 0.14712042075149145,0.02453228356278192,0.6347099255029773,
 0.3908594215040013,0.42344297278837406,0.9237271754731065,
 0.34567399431318824,0.8244458786040885,0.8776696076939948,0.4098439946974697,
 0.29847877580658577,0.48253290722105413,0.6630453982398443,
 0.9965976531460689,0.30718686084770996,0.6892295955937833,0.3608131484611442,
 0.7467505770460127,0.6593870269450002,0.8585163173910032,0.4953844863131478,
 0.10155902791035809,0.9793270404997413,0.05033945627401071,
 0.0939847731769764,0.9094549888964795,0.07194075754691365,
 0.43546431884429115,0.6548978802112706,0.11357213340311101,
 0.48289258335420937,0.44126627236900295,0.614324470478355,0.508312516182125,
 0.7347267699625333,0.6211034931341546,0.8735146451728363,0.9159956785322168,
 0.5953118534321952,0.10260204870258183,0.27927271243138807,
 0.5575500135100633,0.4535401972083748,0.6549165442674494,0.7197482007613669,
 0.5140560476172413,0.9914037065134675,0.9452183642910781,0.01479257404467027,
 0.47446452689144447,0.997982786084905,0.36882313905590247,
 0.45646323600196137,0.23024538955060636,9.947470353843801e-4,
 0.16448723504958507,0.8422592116012283,0.4590662117956993,0.8757437510685583,
 0.553459482416211,0.7776312247467252,0.8595522475170204,0.8837787064366904,
 0.022144142136015676,0.6510707131259436,0.008729271794294169,
 0.788950992715489,0.7223884572296285,0.9140478365976384,0.48656485630020485,
 0.39975763613499227,0.26473706767270155,0.3656320630269738,
 0.9577162456660896,0.7479358389100188,0.08111561162719216,
 0.22078541435705323,0.5912361892143192,0.6469866956570438,0.3214058764465586,
 0.45641178243777636,0.7164028210799155,0.10601197072796076,
 0.18906825101403435,0.010223932965704896,0.7619759329167487,
 0.8823442743595663,0.2901673329874317,0.5466412223171491,0.3032468804031687,
 0.4251694607669244,0.8981762668483912,0.90432148367527,0.48015096944021873,
 0.7882914956680964,0.2812701162908968,0.7442783614260965,0.5320000881800953,
 0.45082210759629426,0.4652084507064017,0.005646593891307283,
 0.32907743787512833,0.39867864015144616,0.06815428024842818,
 0.6373911306645245,0.9639267238826426,0.8387471411024245,0.5287778209854797,
 0.8601944811652781,0.07439896115164601,0.46976980995235673,
 0.25499904007848895,0.27755927133483993,0.06689239671531322,
 0.44892877519088703,0.49391079658727133,0.42080288752330275,
 0.4600413401073131,0.01599443119690047,0.43255812300894114,
 0.0016709954985631992,0.8215436737291578,0.1252515665495766,
 0.32508487614265424,0.5837490736321521,0.34505761390940737,
 0.5294279819752222,0.1544044130243193,0.2492891377199581,0.41607629742091057,
 0.7539383399685411,0.3586246709392674,0.16636530503182678,0.5661158846048613,
 0.18962318630736585,0.2391205399124211,0.774687354618262,0.13308427156669678,
 0.41360776159490775,0.9028331718958061,0.6051599007689592,0.5113973190647895,
 0.690410693144638,0.21145866401861524,0.19606940347881285,0.598232444371314,
 0.9128088142009786,0.06847025206328528,0.22488517295648391,
 0.7077091313197952,0.6709234424310762,0.645382946598866,0.39081768454455557,
 0.5900646370162456,0.8147697854814785,0.16323545731571631,0.9036669794195589,
 0.21030998934774403,0.8800100593714174,0.13374904784515484,
 0.7492490522102375,0.5290297610235091,0.7344326933519503,0.428723979317982,
 0.6735085950221817,0.1530379820134362,0.20280168275049115,0.6056092237107134,
 0.5422103904753353,0.24806219000056484,0.9371363757186622,0.9544424932757203,
 0.13936479153998338,0.6867896088519847,0.8641418611829065,0.5105626905233905,
 0.4195017649409942,0.9336136686186025,0.5752361361567201,0.6045919488350256,
 0.886784151235284,0.6790327049711894,0.1610946814539579,0.553928818349384,
 0.45373157549965504,0.30330072099613903,0.9795198613083163,
 0.8636277053181747,0.6258366559735379,0.26061528683176705,
 0.35539411300784585,0.0952677638374333,0.9414145710141911,0.6094655484252628,
 0.13520934260049045,0.6674642818177898,0.5890291896973044,
 0.49849853206572814,0.8900458511663142,0.8786880150137549,0.6872887564434722,
 0.7652248857622344,0.6874030542182166,0.02400852107061724,
 0.09948879017703982,0.8294582514736119,0.7732121140548869,0.8484423335709489,
 0.39553681208584535,0.21152634555035132,0.10185803622909384,
 0.6389638908802009,0.5388105748549564,0.824527297465303,0.9854232213658939,
 0.9615461594843092,0.06153490314833665,0.9850850686973909,0.8165284364927515,
 0.00973870715251246,0.6173006782512873,0.2544315632751717,0.8263803329134962,
 0.23600609283747198,0.4961914059536008,0.9966823847071989,0.919648896941649,
 0.40685585029738514,0.6216836520287669,0.564893841383912,0.9390342783339511,
 0.2757120720019717,0.4532727599205557,0.8546519837800404,0.9266205369930521,
 0.993321009459534,0.6043978924046984,0.3626634539379382,0.19033852553059905,
 0.6790118801841613,0.5377428634255906,0.24852990601498348,0.8447196860618406,
 0.7412730920426196,0.7579607970782513,0.944109414685281,0.4407135361085477,
 0.5267980686186463,0.7104450498050345,0.9539252851569308,0.8725639584967599,
 0.0017598781912361305,0.1351582818527367,0.3146657144221022,
 0.0899943555241135,0.2348320115607202,0.7386069234071988,0.9836657774067582,
 0.5723770134877433,0.05883209146987123,0.7696949597415812,0.6760209851910367,
 0.6877774937379195,0.3167306379226489,0.8104284665835312,0.36479559304696174,
 0.38825721523426027,0.39698178227661063,0.09207498770154054,
 0.8127904344844826,0.7968853374612254,0.20445634390100953,
 0.23712442872330886,0.9561296057285458,0.8278250147973218,0.0611987291662085,
 0.5916124234147916,0.7524390262051868,0.018553916645499435,
 0.8616842296342204,0.9834142137544423,0.5977396802873763,0.5456005775099283,
 0.6073517656590984,0.8300834383672101,0.8928118911325309,0.30255123613005663,
 0.4409662257686361,0.06419697644323319,0.1889731302600639,
 0.06886863790208086,0.5005532455835928,0.2742033908954097,
 0.45939111305210745,0.3459507721073466,0.09505472599079601,
 0.43315521426665016,0.579714257346928,0.15968973147921273,0.713727321004717,
 0.9568003911957529,0.24677067922485618,0.6349068387370009,0.5309648027182593,
 0.38051896705150356,0.16101596093055837,0.38807440212392297,
 0.45865356067962904,0.8360090232705482,0.9206539660408769,
 0.09401682499422992,0.07174195734794664,0.7676693490546507,0.301849692959669,
 0.41029861746904295,0.9450802040832481,0.06878897195128919,
 0.039799706864186435,0.7245353647319308,0.0717799959058486,
 0.09049402787422989,0.6816787098009429,0.7337244096981278,0.8097700495385979,
 0.007300121947016014,0.32579858469042555,0.7859601130017697,
 0.6249797334315705,0.4820015680178721,0.19305504010720453,0.7275630210782303,
 0.8013376197288897,0.4984816360585633,0.1333627943136706,0.15626300174010277,
 0.04708818248513669,0.04881663408851966,0.38558144211840584,
 0.867498934287291,0.9338978256800934,0.009789730588543266,
 0.027253273614019635,0.7153607545788375,0.863178862089607,
 0.10564240809389747,0.5679560598706339,0.06030080151144923,
 0.2870810125642177,0.31420703818264406,0.7715544046915445,0.671255598525461,
 0.47472814605447045,0.03143567978644457,0.1173810856624693,
 0.17260982534793567,0.3485366983022211,0.5857021667159581,
 0.27537113563359816,0.23931027079723122,0.25527976417869847,0.99012910139142,
 0.2971102449480343,0.1711689923779982,0.1960369263966274,0.4171222959947224,
 0.39521408031423255,0.5809772916071019,0.15235788969910025,
 0.6036836272552881,0.3535841095707749,0.8329574269211231,0.6098299809385461,
 0.4848069463993089,0.8680762540146701,0.10532212210634828,
 0.047893119503227055,0.7627145782391235,0.1499041001235325,
 0.24394087385633245,0.9114585507680989,0.7162514078645605,0.4932747392834447,
 0.07515298294267314,0.8335872702087535,0.17066629673638012,
 0.5215052477576757,0.8209590293588409,0.4808997725809454,0.5935022797766207,
 0.9604110974146373,0.7318975378639783,0.20418659537032913,0.7916459801930916,
 0.8202397661142318,0.22968668897790112,0.7670792126891628,0.2854675201401662,
 0.6304802442177375,0.9562127743251776,0.5887409436309032,0.534646787618011,
 0.8102781071770785,0.612099118650545,0.5001505464527424,0.6469454231819876,
 0.006976458206220726,0.32365222041784747,0.0660792434062043,
 0.6930496410664493,0.5608265199715613,0.517930718240506,0.7239382641102212,
 0.9920900824204981,0.4130041018438253,0.8829592605824469,0.6136569030826466,
 0.3141539543368457,0.18970790621330758,0.15647119332801063,
 0.8365380766929451,0.6555038098559397,0.1809705320378563,0.8807830791927536,
 0.18352495332062502,0.9232044276996125,0.5812394423310437,
 0.20989699357073743,0.1448782208243372,0.6466591488847863,0.7408428611108915,
 0.39866739827687403,0.46975481370390704,0.5957764913434713,
 0.22711509858373324,0.3637083560041696,0.99464406861898,0.035891957784849216,
 0.9892129294629005,0.9437562737928035,0.12763481829965917,0.8286322221879251,
 0.17940956953612408,0.5123163299875225,0.76605638978019,0.7718837955581906,
 0.6536627157475564,0.05039617404826169,0.9825505751507335,0.7527564075718687,
 0.4199992348788478,0.5550570656522993,0.39448251330644135,0.7487737086847552,
 0.18415559670087267,0.7725006313734488,0.07522830712242179,
 0.15282229612698717,0.8225934917345253,0.33976580507104903,
 0.6339053410858291,0.020803477108422896,0.16342284921561923,
 0.07864046873939867,0.24291446706267805,0.9031110604419632,
 0.5842822672932909,0.13669391696578426,0.09568599035524927,
 0.33947589645961096,0.15820597070496534,0.6550099651537106,
 0.4451161801220297,0.8615943618227799,0.09958128939918187,0.6386017843917347,
 0.78523198230949,0.5078480571929467,0.5320821449951927,0.777812561076857,
 0.0052332592576620485,0.9114066589606677,0.9197773707369057,
 0.6382785455264817,0.49261117976067803,0.6734543529617784,0.5813302265797065,
 0.20455325504074062,0.24781504667729337,0.846743104571747,
 0.04942407968448237,0.8070347450353592,0.009065525425507315,
 0.6499566237750704,0.7274625318545864,0.22661392101197353,
 0.006458881435634223,0.32756354149175415,0.353407948387785,
 0.38455869965078393,0.47519884703779103,0.26126625614009735,
 0.15396586698778014,0.6529488941949695,0.468710475059276,0.6113865490457829,
 0.2452155088929866,0.46178254664797547,0.32020400606286903,
 0.32114487008234294,0.05474336870973484,0.8110534982497437,
 0.7335014067748465,0.46355836814370177,0.6527299568234056,0.7226514299233103,
 0.814641613182075,0.2851130606907242,0.4082047220178745,0.9143896475290323,
 0.9412935869517328,0.44701557364587163,0.19600304945384772,
 0.9702062707937366,0.7416644237248073,0.8674703806632078,0.8252395249125025,
 0.12099093661414217,0.8958196118238659,0.9385328888098285,0.4266657007679493,
 0.8798763969975474,0.0450990007534684,0.999586338956446,0.049772576012849346,
 0.6080147433227331,0.6697184467044848,0.5834660771177151,
 0.023307842640452936,0.9680451647248909,0.10941496210162793,
 0.9805144846742686,0.17674593818833295,0.4782636069153604,0.84159179264384,
 0.5795889257301388,0.34216993447006183,0.4392372674020114,0.8603942663600355,
 0.9297463701685895,0.641662683275917,0.36881120697686653,0.8036180894331938,
 0.03238876204998653,0.4043277068471389,0.3304808234604708,0.8098018063293642,
 0.1823020913257115,0.6368957518430962,0.5667590551334664,0.09831748050343414,
 0.9282712731272621,0.05851973349155282,0.20880117559176803,
 0.5592948525675246,0.06917277160691004,0.5144040544436463,
 0.23294974995752882,0.30068273993720696,0.8752856237217163,
 0.14755635136789702,0.1654291507670097,0.9747865800941111,0.8029043576942457,
 0.15029222266971232,0.7128614991357044,0.9114525409907835,0.2127723168429,
 0.9625345811534535,0.323461528331329,0.9497803381340723,0.37795285148632574,
 0.7785898060448938,0.13974249865469368,0.686365213847682,0.591144445680975,
 0.5069151744263803,0.18509970384302854,0.6610232995135832,0.5111296657816028,
 0.15403689499699613,0.05161795832279248,0.28017571723106105,
 0.21856271820259454,0.6633846449243322,0.15708682772476024,
 0.26327493201600816,0.5012715299413498,0.6508976495189625,
 0.09251441772308455,0.6831902132721815,0.17928545587114264,
 0.18048230279504007,0.2535384745803114,0.4414651065680836,
 0.14654307065602934,0.18379962543035244,0.32355588602935104,
 0.9203479089796832,0.5471651525134554,0.7515880589399448,0.8708389504778897,
 0.36115436738020545,0.9237764249728176,0.6155855337809082,0.8092027955436674,
 0.11367040485564583,0.6453034450108744,0.9744628168615989,0.7539185033197129,
 0.5357693006781725,0.3390966092899116,0.6934170030180947,0.5157663067296333,
 0.9779663874501396,0.3508621654116322,0.7587375005111223,0.5697644185678072,
 0.6927600934559026,0.25854282863240075,0.03662973077120346,
 0.5079667174811684,0.2990758515038112,0.2765941495911244,0.857073283501801,
 0.16239507054620217,0.6988974013705915,0.25476781500830137,
 0.7380797405468211,0.1414061644968051,0.22051483291200347,0.3791787817473252,
 0.07299866250618858,0.09928906442023189,0.9993050630888838,
 0.49908800581134305,0.10049902534194244,0.01995291615116468,
 0.3153175325263049,0.8523817332564818,0.4672102646790449,0.8222896016437937,
 0.7895308707960929,0.08795032407328085,0.6652132451947712,0.4453991876729464,
 0.14629285278774917,0.6856747925178434,0.2005777805379052,
 0.27224889383724415,0.8929529216412623,0.7947650272244391,0.9421150897326183,
 0.7012282842922597,0.10706533758574133,0.6639993123583992,0.9730492146534825,
 0.6890425884706304,0.2983462801093081,0.8256915563083953,0.25378065504029235,
 0.6827851544919177,0.6348828032720467,0.17465900839367154,
 0.019733934023713373,0.4776618993828414,0.5699077225259876,0.318491375817169,
 0.7050017242728781,0.6896588663268184,0.11074401800501421,
 0.30896174882056704,0.039894338462855394,0.46343278543503685,
 0.9673443332597884,0.5526826629014205,0.08518496629531203,0.6996823197567213,
 0.978981971600331,0.7462439000264817,0.5134649831557334,0.8399636193005666,
 0.4787711399547776,0.6130524392328216,0.3189425041593631,0.6403063015427914,
 0.08624972225672733,0.8091367696498469,0.48134684041883136,
 0.10565141502670194,0.5651853558944585,0.7970280908627823,
 0.17982757853297104,0.9476965820903329,0.31501964649311587,
 0.9542867624289331,0.7238735926138692,0.501644214106531,0.7002622860928005,
 0.14244688050217824,0.825885937425572,0.7246994333097282,0.41756139265494596,
 0.36860616196595974,0.3314896630656473,0.18077940841444384,
 0.45832973234787844,0.3255325021738067,0.007671931306069135,
 0.5995686764480964,0.48766527621525746,0.6477375375852655,0.6407836884686943,
 0.4576174806655222,0.15021559776153715,0.4124795551295025,
 0.49683517765571383,0.8156012451730812,0.7628081690441111,0.2785581940178441,
 0.09861372262295998,0.3404201914459879,0.7902719329759902,0.0883418657205991,
 0.24923017550656357,0.40305174679784095,0.32683741738751626,
 0.7776080778725125,0.7236954815666702,0.05350304432115749,0.7836256241118936,
 0.3542355451917599,0.888986556720619,0.11018510453493535,0.9621511795952432,
 0.1348191219127075,0.09845907293312472,0.31713996197703953,
 0.7231685501565169,0.26194997862398894,0.9544854586069484,
 0.39277598495889476,0.8286301121592708,0.1187945350997579,
 0.06710190263096938,0.11365542571841292,0.2626398107594886,
 0.8257356704649249,0.8993284171113554,0.05215599616079525,0.5159036119337727,
 0.5818753344059896,0.5121095513022659,0.38858611574689883,0.8756360869017686,
 0.17517190154457263,0.8128712641654922,0.3766286173880482,0.9324756117488662,
 0.9168978369802563,0.6755127250854616,0.017130139303741188,0.918813651195127,
 0.15610459344136962,0.26396136308909324,0.7391717237703493,
 0.7161351916904306,0.21636679821766602,0.09431708581626241,
 0.3321690434847381,0.34304658284265477,0.118025291484391,0.11044151972389415,
 0.0215142672581079,0.9939892187253588,0.2074943133051811,0.4502189537311537,
 0.3083057138083195,0.6286121913450098,0.7850770064236975,0.09552479270268655,
 0.4615535515113589,0.0613481256018269,0.6576201778488706,0.9378620535058514,
 0.6595553029468348,0.5543187364667019,0.32866966889024973,
 0.022678674894270223,0.4037777804990923,0.6198138859705707,
 0.9945507860035161,0.5475731997696488,0.7194331560065889,0.3777709704500487,
 0.589606507987736,0.625821113064432,0.8409193844068645,0.32708706411221555,
 0.6672695311153001,0.1378352471352895,0.10874922389801334,0.8355523533209832,
 0.9382513802908409,0.5137617061537301,0.4794823104104149,0.433165077843798,
 0.7399483097405536,0.2136273406677529,0.038100125604748514,
 0.39507548717661867,0.03671666057892131,0.46429902526443,0.9891586076770623,
 0.7314861524586668,0.07536786923453176,0.2467688883867869,
 0.16944505004201116,0.5468833878433781,0.4142699174163057,0.8654753919575234,
 0.48791317380250987,0.5082256644781455,0.09021412956639396,
 0.5792673111255604,0.4107574993317398,0.21788116929927503,
 0.14984864920675045,0.15090624598336821,0.8314868308842137,
 0.9151710257348715,0.14866305244347622,0.8988113148557877,0.3549509916858833,
 0.11185233726632715,0.841883346758033,0.43513107330810363,0.4613364868436476,
 0.32684505235480366,0.2071377215228698,0.45510661230061555,
 0.4724703232847417,0.34243424907238607,0.9260197137471223,0.9513472078679117,
 0.9644383986534981,0.4677453847546422,0.1838023849943038,0.8468694161515827,
 0.9430962548404581,0.9035335567447641,0.5175577346905289,0.6197864554617908,
 0.826922054322997,0.2867630698915036,0.7683044295728916,0.7145702215209785,
 0.8746633049584567,0.3631217426513885,0.22666598014636896,0.5532284921436413,
 0.2834337344371418,0.9470383145125392,0.7211684071453641,0.22138677510737803,
 0.5410401164355542,0.4938361778831464,0.6900373455234264,0.34158021575427056,
 0.7787664190983945,0.42230387917562195,0.9469066921218712,0.2691789816740202,
 0.2108494113049686,0.2757934424132634,0.9711068521159194,0.5758356979813654,
 0.2757752312451025,0.7347389559550044,0.12073771924938304,
 0.19211543458768432,0.10210662776253454,0.5463230025209229,0.535651539764222,
 0.46887749615322616,0.052351268972129605,0.22637647572919095,
 0.5050867861121324,0.20552413189210106,0.9989419108114567,0.5917106697567658,
 0.042224832427314896,0.06274464976895655,0.1856883178592159,
 0.6713351978930421,0.10706414673878306,0.037144728980913255,
 0.21688458140536937,0.5079488991593804,0.33471992422471175,
 0.4655058305023677,0.1806851944908574,0.934202437062503,0.09079807757078595,
 0.011165008245844987,0.12940186125787,0.17989782830038503,0.7167533928687558,
 0.4344400828759448,0.7904062540595532,0.3097352163947966,0.8618872306256941,
 0.2933690108077477,0.43958692686114254,0.5124750381962028,
 0.32211858150103967,0.119294302698437,0.09746657775114687,0.5881768016218725,
 0.03243590399939966,0.2938327627050562,0.9128376505202512,0.6720676693719315,
 0.8820777554943389,0.9162184926720783,0.8103762880880726,0.4089645592220298,
 0.6887065049963446,0.712370679650643,0.6167139051943364,0.24315584951178537,
 0.03402503063054907,0.47562238944454194,0.08834238007226558,
 0.7733624277371273,0.8310736528792738,0.7833883516609998,0.14270597596052892,
 0.1038651794141976,0.18265398076981854,0.48928846420278294,
 0.2854976975699809,0.220610396332806,0.32955273044820355,0.14264534596736356,
 0.49035719243245857,0.13337699999109043,0.9563678132907513,
 0.9929167456005126,0.8002642951018915,0.6546180978000118,0.6544086491621812,
 0.6777456991783883,0.35278053979518254,0.9886905548358073,0.5669109156350255,
 0.7920368031396924,0.3506863364316273,0.636969064049568,0.3868816800123194,
 0.5547847991461083,0.5821379035799135,0.5024710741761894,0.4788058024102423,
 0.5769990502617587,0.7438643525702263,0.2202509375999776,
 0.003313712596454632,0.649778224629061,0.32825021309479907,
 0.6540114118288662,0.36634765916222545,0.11361234080572191,
 0.5769626862703435,0.5820318559113464,0.013325948582222913,
 0.8688003885989416,0.8852349614208773,0.5612286836620739,0.3782046092876743,
 0.5687244411471852,0.9110372590764044,0.4680288061223705,0.07512213676444979,
 0.9662733500381135,0.2790626393921255,0.45508261556325014,
 0.34559844664500483,0.37153737221026684,0.6903587158969453,
 0.4935861125567106,0.1452229189491523,0.6761218828070759,0.07837641539944662,
 0.7717810603445737,0.7548786420104401,0.5683653998356166,
 0.012683808390005047,0.10418234452883679,0.4246835917246856,
 0.10483864695048983,0.7784719895883355,0.5802303427975082,0.5828040636125399,
 0.5858309308641282,0.20899828538056908,0.5819482471585168,0.8692203673943083,
 0.33243860630004507,0.9401846773480975,0.039937103233804394,
 0.1116480963212153,0.9693263373814412,0.9473084101285996,0.6036515448288247,
 0.336936177530907,0.6768688621762063,0.19611846181971293,0.07817446646386761,
 0.9295057147302636,0.6245459946404248,0.8047943074674274,0.1503769866570731,
 0.30574083381241934,0.2417740612273812,0.31275955431742253,
 0.21068526280921462,0.24994056637077267,0.02763561967655226,
 0.2696206455935095,0.03578724881922457,0.04833262498667412,
 0.2732514053888435,0.17584800293108904,0.6808602312901105,0.7582307105847607,
 0.4032052184225865,0.27196483966339924,0.9593576974007662,
 0.10324990493910391,0.9125804122453406,0.20639382479700363,
 0.34393779899837773,0.6580037104875209,0.597187075334332,0.7216464256347064,
 0.11060944135197182,0.7292211564852873,0.7341488026504094,0.9730456929583626,
 0.7510619548389142,0.380564219650964,0.25356761465646616,0.39214122354491954,
 0.5912224017074079,0.35763379213172664,0.09411669811193879,
 0.5015345901173456,0.49296788880036957,0.48202463024792785,
 0.8295274264168644,0.9650069548135396,0.32258977285836477,0.7898457862735951,
 0.08772380131811386,0.40813892048445377,0.26567864959516874,
 0.1515670024023935,0.9709857299904869,0.06992997582597948,0.8778513260913405,
 0.22466999363929196,0.006240553283149208,0.9732864092636859,
 0.024559217586230142,0.9755111066756867,0.5279261830414469,
 0.7067824720236955,0.6977072734006969,0.7227326960708713,0.4318319370658025,
 0.4882199806375017,0.9386241169571914,0.7554429540281328,0.5729235672452261,
 0.7644001955962912,0.7051694071065926,0.2429788639376027,0.04204550599169321,
 0.8392818769886297,0.7384341564172637,0.45541633463468467,0.5076907934962519,
 0.11840520200595561,0.5434187437078053,0.8963946791772901,
 0.30467690096372035,0.668115079781098,0.05485222341503326,0.3552387288466088,
 0.06328407976026507,0.9246742398125906,0.35154898339934926,
 0.07766446415560546,0.2635774150779091,0.4247307006576647,
 0.11112141575871337,0.5945413881326459,0.0836277236392724,0.4964360877998246,
 0.4638640974231055,0.3134416259118433,0.4887719601564169,0.7757993770877996,
 0.6049134611398337,0.012095151909331081,0.24186219397507025,
 0.4191841998003123,0.3445305703732129,0.6379752300312579,0.8765050266092894,
 0.8030004029344501,0.8490584567356566,0.5780402762278846,0.1861104103325859,
 0.2991330338949043,0.9563644863181111,0.9624056315765831,0.22812224743244658,
 0.2796462336021929,0.8340656519640813,0.38954885522468263,0.5063405099775654,
 0.5439839820529463,0.7252625614705882,0.17589964261857427,0.9950519291739564,
 0.22387308291783747,0.21655145190352998,0.9636264449889282,
 0.39135982500334043,0.12048726763675466,0.4040136905571907,
 0.12862843882257102,0.35807762886567185,0.527546317605273,
 0.46795861750757173,0.6358263635792405,0.4645466106214742,
 0.029695391874592136,0.4027072760537763,0.45595572095166415,
 0.7682787883739826,0.21517757569967544,0.268080927502657,0.615588110607516,
 0.36006157834041275,0.056257507306930465,0.21571628157630285,
 0.45013842495278167,0.494460630157471,0.4452263693061621,0.8581939408328244,
 0.892427999959128,0.6640600579382284,0.08824820239470932,0.4559094300974049,
 0.8271851734724136,0.08559221035514653,0.45412709795582296,
 0.9055008856497496,0.3055286563420119,0.9314287641131567,0.6709698283712782,
 0.4850489507051704,0.1975919542908252,0.8885764506961539,0.5136328908118559,
 0.632500065025731,0.03736878853762482,0.12190750040545784,
 0.28830740463602234,0.26107621169570894,0.2456120693790732,
 0.3081895234685966,0.6965088209373163,0.16521404134039086,
 0.06917535732498781,0.5954312591864704,0.12849513987245098,
 0.7150159989239074,0.2465340849140225,0.2637853858689979,
 0.028760503556784722,0.46292705229524445,0.9373157217283423,
 0.8084771872921207,0.29766168818948835,0.7120642761686299,0.8375668804993852,
 0.24644593994591057,0.1403832185820728,0.18567011574867998,0.71658640287437,
 0.07613707100405787,0.9485244637966176,0.7921785084354883,0.7483660397700639,
 0.8608821770815646,0.727867406148992,0.30752642955739373,0.03483474738722925,
 0.27882492394729197,0.9476279332431254,0.5308892851110316,
 0.35594476793931773,0.19831246982351447,0.7657443929248093,
 0.7363131051642294,0.5746373925496053,0.3237676986150677,0.821100219067298,
 0.06611534594176938,0.22841324190815215,0.5888626906264283,
 0.6020460067059861,0.9134311049274655,0.9620720667348106,0.7739080599562005,
 0.913735585980028,0.7234379617375905,0.39556501793325516,0.38218418905779306,
 0.07696831866968568,0.45180080361457275,0.8219231699701651,
 0.45928931576815635,0.2798336459246291,0.7940976074697792,0.8790583096677191,
 0.9930499559086506,0.9082243386440237,0.31596496905936844,0.7361721412512408,
 0.43122703061717604,0.10176020861782664,0.0902184754915587,
 0.35519990154095626,0.630904916186388,0.9854263008800683,0.8128999344588106,
 0.6205292441550238,0.016901288087148303,0.5111904085976202,
 0.6198519643899617,0.3958041565023045,0.27481206197495467,
 0.040695659980152565,0.812143606227907,0.6644982175146679,0.2426346690023865,
 0.20943439463419056,0.9753183868360737,0.3381590526374797,0.5377503773085225,
 0.8733597473505633,0.44663748055203045,0.9696413043530754,
 0.48972125359656093,0.5633614578790465,0.9813135940634099,
 0.019211311030091105,0.7163843521225989,0.6558034175174305,
 0.20207917846489143,0.9388935033049052,0.6585212945016029,0.9718043150036847,
 0.9250725783450525,0.28654606823531203,0.9766224876304391,
 0.12811576448768947,0.2188650408181343,0.9448557892692477,
 0.25967867441223014,0.3305371021329677,0.25975197568264374,
 0.5464634912810707,0.01431359905262053,0.5737842444061119,
 0.48875569028310584,0.5240908635610636,0.43343699765098387,
 0.6910279670193397,0.30117499899061895,0.3332302661106654,0.0661610015240236,
 0.043314540359610354,0.6795893686262309,0.04352284388252836,
 0.8821005460460798,0.9241274740975605,0.6467240453322466,0.7702541326277474,
 0.6027922120965028,0.24016595492003767,0.4104588281020546,0.9031013759055461,
 0.7635726378897423,0.3071554699046072,0.6933744651710759,0.4000484779096958,
 0.07401785988476173,0.40686787388560464,0.6914567498637982,
 0.17287967676447435,0.1274398087434372,0.9107447620523161,0.5309024250440446,
 0.5711004642844362,0.5849798757175043,0.3911315592924969,0.6888822931046428,
 0.15505673785141982,0.49924515644150014,0.8024854231123089,
 0.9079016431108619,0.5420805875438891,0.4884526409162112,0.08986707229990698,
 0.6101168377817437,0.4872749071506388,0.5443129837535724,0.15962770211566102,
 0.8173027260940697,0.3129105058140529,0.42889111856781503,0.8853949312407411,
 0.3774527197308678,0.6021449409027263,0.5561197250368478,0.4163980378091201,
 0.7396882235294544,0.5269119749408905,0.1837756127338951,0.639944491578747,
 0.8180587908191306,0.5105630478957129,0.23174032374253395,0.7762378785409181,
 0.678769823548954,0.12501988021878752,0.11519722296024704,
 0.18490950767200354,0.4650744556245938,0.9060005014608776,0.4793920020426289,
 0.5568981025520221,0.3522969974698523,0.4425141376663885,0.747697769675309,
 0.2998979511970348,0.4908200236853434,0.6797857242654295,0.832872030181015,
 0.4479558381335831,0.9160587875563563,0.6453077925988271,0.9802587506231952,
 0.205874584572588,0.23318208625370562,0.3154440848560367,0.3811148519238392,
 0.6783815763396787,0.4683030510575825,0.11030509577376724,0.4684724304503536,
 0.5191688854617873,0.33247741521921725,0.2042825496500993,0.6280069852699303,
 0.4333742290465513,0.7854140134795534,0.26391305279933597,0.2877147307778176,
 0.32296921830768444,0.8097460980094738,0.5939452558579786,
 0.24810049284305602,0.16299446086053548,0.40059539663283383,
 0.6150282996929405,0.9964650175034054,0.8494015169126253,0.09941369286046686,
 0.6139122134777206,0.039877678157737906,0.9872304606124209,
 0.4310407018199175,0.3037264563162122,0.867405422539839,0.2640888847547913,
 0.7942542978224195,0.791739284410073,0.8623313018875478,0.6528970127534733,
 0.32246567092114187,0.8859284723407362,0.16882659024863378,
 0.8566957763248813,0.32760924566493643,0.09160430187319435,
 0.47791625569463625,0.6803311303320475,0.9759245744405929,
 0.14338845709056247,0.4272855776328043,0.7738831760245573,0.3636331214320445,
 0.11688747790067266,0.17782956300565267,0.7072318464852276,
 0.8819372607578091,0.7144747432168336,0.6292474014857428,0.02800028992840875,
 0.24163935969007522,0.80803035852003,0.16925504611756925,0.9696796236207772,
 0.09566607423144846,0.16817539637935508,0.9799124246428397,
 0.9422604807831851,0.8708396439809079,0.4887075084102934,0.22263500897984667,
 0.08960618780467977,0.2680048358631384,0.5875860064434845,0.3340303833506558,
 0.9331298574425833,0.07888746284768045,0.5384433756558961,0.1975366308037817,
 0.5111440487471545,0.5514654030121404,0.39341822769793133,0.5361197151821201,
 0.9650297513541294,0.5559315145100215,0.3830364998621604,0.6885936768866993,
 0.5702421075973272,0.5023326970827199,0.5166775161046502,0.4437969722322036,
 0.40447819132874785,0.8663471287615451,0.882563648533637,0.6688923444209378,
 0.5697741313400561,0.4579037847299433,0.2528738104347734,0.15742817567940182,
 0.920524894154731,0.29237581909600685,0.9680265342166067,0.6822546085677726,
 0.8317246073086513,0.14987549216771046,0.9705695501702093,
 0.35466029027360957,0.3589589928328616,0.46301134120175025,
 0.40488273662294083,0.6485708830051378,0.23366355330373678,
 0.7927152112450684,0.612997561136875,0.8572989462015028,0.8753683170075962,
 0.4592494425553687,0.8879863097550744,0.3778930568110864,0.49267221618206236,
 0.48933252670623273,0.5242775211331873,0.12136291690451917,
 0.2564021011597917,0.8867935940916266,0.7216571996585344,0.4733699471932904,
 0.9061860315078772,0.3583271837432519,0.6862661585564194,0.2948119002193703,
 0.06242166766293311,0.7574430817092167,0.031817836369882,0.4765294674431917,
 0.7330571495141436,0.9170636545355872,0.8116834463105168,0.7391073591454027,
 0.6227146240373425,0.43879283089752064,0.18964082152572537,
 0.7545955641670363,0.19221580879947497,0.2365237318616351,0.9062910303239139,
 0.42998565205345196,0.6003480165972357,0.4429058760029386,
 0.25216598595746875,0.8054224821661619,0.4643264756411005,
 0.042704673455152053,0.7370985066911655,0.7406842276976918,
 0.36362579537924555,0.9214118518036688,0.5252449382176365,0.9604746686879182,
 0.3795993905788406,0.02549673167676103,0.08226246096906553,
 0.7780733164988385,0.02894557468676573,0.2873171748182374,
 0.42296927586639266,0.5627281356488465,0.28118807848085103,
 0.9014576880251826,0.9474208283153195,0.6987105254817719,0.5317384136791626,
 0.9219777759277845,0.09594569704517564,0.07807176217784362,
 0.22613838021546828,0.4731437927665304,0.03666559016720072,
 0.6515362711210646,0.0345573136750853,0.27010197714192763,
 0.07017095215807623,0.5531410897916279,0.5563983337876901,0.8684345031270321,
 0.039265384796035696,0.840205801895112,0.2655958822006368,0.9590306216044614,
 0.29944407885929025,0.4020944653939118,0.05211040405236933,
 0.09884168309062025,0.1370945128880856,0.4883465283480042,0.868198163920088,
 0.4535901519679567,0.5262937699628858,0.1048112410967258,0.42167933547041636,
 0.6129397740989898,0.7000206702456502,0.9307698897665557,0.3865976834211915,
 0.08986663262080752,0.02903872821512532,0.2536862685815162,
 0.1937188199634423,0.9259947914192954,0.3565711862396811,0.7988321973288416,
 0.24611325587929112,0.9120913392652353,0.0822277418259213,0.8432242763949096,
 0.27023550557487686,0.5624292821426677,0.5210794539872629,0.9594624795510074,
 0.36730420779988093,0.6246699395618868,0.9720146283424799,
 0.20239689877644929,0.9326258216615925,0.13409196275575996,
 0.14763752950178477,0.38291738504276707,0.21964090868938724,
 0.5720770620608259,0.17548207875888955,0.916347293521281,0.41623917769491436,
 0.3674991462542664,0.9690575513123227,0.5551388786233014,0.6607315680681499,
 0.10265479758447849,0.32308095009658677,0.13363236157998448,
 0.44614385341637475,0.4059918715405081,0.3669696117464376,
 0.37538154580284777,0.11384689194159603,0.2571704666022552,0.651676838942148,
 0.4746328293651285,0.5757514662777428,0.1762789866230955,0.902354998473428,
 0.9722000506147717,0.362514052567218,0.5153241767092366,0.19640225566147274,
 0.9361338151038271,0.23520219895824168,0.16567013574647493,
 0.38069096338677166,0.04405045418408937,0.36653195807440786,
 0.5211814819455314,0.5984844600887929,0.6428586468199701,0.5579996201437014,
 0.40520402685641943,0.35227195551444757,0.8406742292335847,
 0.1789562724674285,0.9632595026684511,0.2931226308459368,0.08199254585569105,
 0.21543796831454531,0.5117173037950424,0.9041704626966068,0.128819995199979,
 0.7570787734764973,0.40242545765917936,0.35580113439492933,
 0.7077644858288586,0.888682516288927,0.31876715540205247,0.20549754757520466,
 0.1749448482675805,0.8560429874699106,0.08328916526598396,0.7769365515216426,
 0.15022982866812462,0.7211057800138526,0.8650146541670984,0.3766082305710684,
 0.3646790777029333,0.3752310877419933,0.42592096574932026,0.8314707799127931,
 0.1192273460123291,0.33633008022813693,0.10319998774964312,
 0.18481933057800448,0.7238268278578551,0.5357144974301505,0.9738435135897175,
 0.9850726683180149,0.07573237376058961,0.8903006050801862,0.8721057136096626,
 0.20031363185817674,0.5590883847462381,0.5397098854212735,0.9471357793528119,
 0.5173193977078677,0.44729117721444434,0.8775346293699774,
 0.19086265065616104,0.15604775733396525,0.7298390608242417,
 0.4279055647609188,0.8484042757759721,0.9372986204823139,0.9091864246973644,
 0.6903325587773045,0.8791446871196579,0.06819661134380683,
 0.06963333889301526,0.26403255168023465,0.07392505206342259,
 0.8449278658514003,0.3530907065362221,0.5166156979372596,0.12840711908880853,
 0.39402584063109447,0.8014826147688774,0.1705649933502862,0.8363292464552378,
 0.589749916244633,0.5088638348811539,0.36317628453570483,0.09754070937221482,
 0.19258403238272648,0.5515394160017152,0.011998935636197028,
 0.053486967166789134,0.1365430164630489,0.18626379873177323,
 0.14492004739248265,0.06500730118696296,0.3827796601672051,
 0.22374807204010905,0.9332363198735706,0.42194838240856447,
 0.6922513896323051,0.44735428108357356,0.37067752760395634,
 0.7724192254682771,0.4722616264893881,0.7263647707971337,0.8040834511151909,
 0.34660230169167294,0.5672153001466469,0.045474716477149846,
 0.03322813591617846,0.5363777275193308,0.19316738974610115,0.72820794477911,
 0.9358786551057721,0.4326163702065208,0.7473503326064646,0.8412993499140393,
 0.6121439557902918,0.8215217767722731,0.8990423553173386,0.9999431324950749,
 0.9222338153598336,0.18668465515636135,0.14110459302178668,
 0.009499113399330072,0.7770479725288257,0.7341644557121457,
 0.7351694238702555,0.8464161189577772,0.12384137491279401,0.9523399799196238,
 0.09076772781796916,0.5599735808336413,0.25788934545353503,
 0.06855796461103536,0.9111044092466674,0.5961279133659402,0.8382180452623348,
 0.5977121493210357,0.42390277854642744,0.22688280127034177,
 0.5874393439436248,0.8676985747207003,0.40203488316945235,
 0.43927852701007886,0.6164043515855218,0.9211440131110958,
 0.32387080549545244,0.770023433731713,0.9379473310082741,0.0794069651234075,
 0.158743123637495,0.5877073148219286,0.9267096267961454,0.7822911427129329,
 0.37894164809738917,0.6342780995747745,0.9402990080204456,0.8269286074336426,
 0.7442481871023848,0.20352761000558395,0.28075027335168046,
 0.11715085904233857,0.3757758679936726,0.024189848498371913,
 0.49161237146009484,0.8123247844465125,0.020301859985420334,
 0.6367206852939221,0.33484180155895893,0.31498861433749137,
 0.5993156220730829,0.7145047092049711,0.8305221189269377,0.15972201797421082,
 0.19645739239455573,0.5718958666662403,0.8494585600428467,0.557744738393189,
 0.8389737218051002,0.3978996836648534,0.4324251593952342,0.6493206318380242,
 0.708280062428906,0.8419846955518066,0.651717053179148,0.04866839326104755,
 0.6178017193060351,0.2905362414179743,0.5490500954263362,0.32986482003386675,
 0.9884742119892136,0.7285153954675124,0.28917871695424124,0.153734422868116,
 0.21941869509927026,0.1821154896935958,0.6227170023815275,0.5480376691456127,
 0.1650257515366339,0.28251813132834747,0.45285340311754396,
 0.3338264783713374,0.9318841593172008,0.5716837688750439,0.5286837289980771,
 0.5107035902559208,0.8587786058773221,0.8405642480305595,0.9428800322586931,
 0.31678650262409247,0.5055241294147632,0.5111379084924108,0.6987106956930139,
 0.6646604193058743,0.5294111865438635,0.6097680398440393,0.37352584010535406,
 0.09908842705446508,0.17754343028871755,0.09455965487583506,
 0.46109686636901226,0.6043352802278869,0.7800031421864295,0.5518612266830668,
 0.04398231314565615,0.2018775937574,0.15295879848766147,0.29906894906484494,
 0.02105289999297566,0.9937002744899271,0.15245277533723967,
 0.5732312616753528,0.7531574326339817,0.9325001118601786,0.3679716083391982,
 0.7799961461032385,0.8443475534858268,0.6386953427605134,0.7952806923545811,
 0.2678689294674015,0.6556627710112605,0.508321119433747,0.504290471462771,
 0.5522104177770498,0.6138096506633149,0.09657508579729646,
 0.49663737429855737,0.2670387976787385,0.2694592849630054,0.9956209992567651,
 0.3101981766562859,0.6573640738260829,0.4830219857429887,0.9084876423125654,
 0.2933597840907348,0.3067481454919191,0.16967363082056897,
 0.003930015393247888,0.5543850853470322,0.9156556929649173,
 0.0017281534607509474,0.9171500472051788,0.28274957546024393,
 0.7188067181014246,0.2918501248940739,0.046367985151617774,
 0.28328207850552034,0.18531162571404658,0.7981609825887468,
 0.4322718821136895,0.5734911846757913,0.42383134450566107,0.4099292886604896,
 0.634840096355382,0.6929131352115585,0.6423991428367428,0.30635839819515864,
 0.4827185710373363,0.858162729893726,0.12615919804292908,0.6040453875519339,
 0.955601554765535,0.09197389252097554,0.509551204066498,0.03901515752914331,
 0.3693866109800893,0.09029737154463446,0.3341441273389938,
 0.43932351285506843,0.38045675827925296,0.04007720082028432,
 0.42115160584851974,0.7482516784921681,0.33555873309010664,
 0.47623541373543143,0.46832256966833086,0.14791648901725785,
 0.5952549096270375,0.12825827137481716,0.21377842491780652,
 0.5748945956756912,0.8919651927776965,0.12479343704635493,
 0.21404676507715792,0.13994444671568917,0.8538310079868243,
 0.0359021442607248,0.08150801126840868,0.8897068797910395,
 0.0026924774933763107,0.22564153622829464,0.5055176991474869,
 0.025804869169185674,0.047055429355250866,0.07850751128826483,
 0.870683184936841,0.0766709620357855,0.5087316982363161,0.2342129313684076,
 0.4750417823165961,0.5241278428303439,0.4223943279368795,0.3411069688752212,
 0.027014339280046373,0.16592065411137213,0.5351924076515527,
 0.2212881247442926,0.4017152724177936,0.18884203052377657,
 0.20021109603828413,0.5970521468049058,0.11747424670152595,
 0.8490828486499965,0.07932171057376913,0.193795019043849,0.601729782736726,
 0.29583683921484927,0.5806889699763412,0.4661248171504454,0.4246919959279657,
 0.7896386118517992,0.9328435663967639,0.9808015514062405,0.09680011890518714,
 0.39358342256865353,0.9752509179707355,0.9305982408897511,0.7730919491656796,
 0.0675803102388376,0.37457434614151053,0.5672902133146969,
 7.116217587366247e-4,0.3801373153146831,0.6975318366646713,
 0.7607289940365629,0.6357013784700787,0.8429506015813485,0.952128535571547,
 0.8419702803291983,0.4774338472110147,0.721896220208608,0.12130602741484153,
 0.7860461932155501,0.23495704999453404,0.5683030338672957,0.5631143659689939,
 0.4636116015403867,0.5254413738729551,0.07369931669630869,0.3823291781320949,
 0.4191652531677982,0.6978284634346642,0.7319081315253184,0.785045927863508,
 0.13911202326070926,0.0815646135445649,0.7287588058581885,0.360473184965421,
 0.6580612489686568,0.1685644068367671,0.9596071825330594,0.10256988627589458,
 0.6060560279269132,0.8970124953711434,0.013191028718375186,
 0.4322359662195179,0.12130508079208546,0.26412178162412625,
 0.9538160083152871,0.8978444580652227,0.10221057800714572,0.2217840662188525,
 0.6936858241044304,0.5830903677076824,0.4437359302822439,0.11335433744592871,
 0.4533588591775617,0.9048815369612553,0.534157776818195,0.09519641796627498,
 0.2844219734945659,0.029498839757257422,0.41593737848134105,
 0.10970430042081014,0.5062004634261374,0.905964819468727,0.22001482980649567,
 0.4407242276526697,0.26901427592529514,0.028874026171117118,
 0.5858107754498614,0.1912796997399122,0.7871998414741459,0.8523601156605646,
 0.5370535784098001,0.27587163066473275,0.9612047324915132,
 0.31488742089050226,0.21383570446474454,0.2839748321890805,
 0.6561186909755792,0.44187202003097814,0.478428048761422,0.24034594097989515,
 0.8602321789494173,0.5058509060469163,0.35573734056466644,0.7388003200884317,
 0.49773535194117924,0.29227364270452816,0.8349048034441223,
 0.9094250080296998,0.1348200984425264,0.4963347671191154,0.7218647163556458,
 0.2975659721137678,0.5942480044201573,0.48683374652994515,0.6686590710506195,
 0.27464324591724165,0.4756527846187508,0.9881813476116503,0.8149312188415497,
 0.3563782592645245,0.0013831580181564718,0.44927300958132976,
 0.7091009847582509,0.6521747452985907,0.1462488220530418,0.5854828128764319,
 0.7204334897836413,0.5657681913040152,0.4731303561574981,0.9904841983124091,
 0.18865568572253466,0.9127460181479635,0.9475558210063872,0.6477122105514556,
 0.40760968230337025,0.5585797556385693,0.707384660756654,0.3471375476776861,
 0.339544234485308,0.19413997456153653,0.6144715857403822,0.5610550504336544,
 0.08848006748144432,0.732634634812981,0.3501320955463578,0.05988161447002882,
 0.07347862695733576,0.24894982005062316,0.6043252145155174,
 0.7663861991052268,0.6510198372333187,0.5538322803636951,0.12273308453404064,
 0.9019360700793382,0.4801032445412914,0.7987760382433298,0.7303079331052241,
 0.10206116277455379,0.48169176209414133,0.2556300969192482,
 0.7070157315971555,0.17781878591996003,0.9937387454773146,
 0.14759556472100566,0.32432479581077356,0.2976012067578011,
 0.8230448099378741,0.2063281906053499,0.22279693594516814,0.844830152638493,
 0.2794081702639808,0.37642264120807956,0.749009669712201,0.09395250979503911,
 0.8121883279930411,0.7795800939636417,0.6644199080595155,0.12678174385412477,
 0.7538464841040768,0.9249477944786488,0.3415222249613846,0.18014127681936354,
 0.572135853712344,0.9026179756990828,0.35724587179594713,0.9165198342637134,
 0.15469249894384907,0.7384655138169618,0.22768371087574213,0.853324390395631,
 0.29433560458403774,0.7496034077811696,0.6616354013987747,0.2291229528889528,
 0.16021054995085227,0.25397515428947603,0.9388807125405187,
 0.3875479839944921,0.5371392663523762,0.7349356510762477,0.44513778930777004,
 0.5497186341980043,0.46390557930576926,0.43769159726817186,
 0.10862764205530884,0.9440031364215103,0.9158387439576007,
 0.08315491047278367,0.04522535221174384,0.7193238187090041,0.547551683807259,
 0.0012939224270491767,0.5468381075623675,0.7257494514416515,
 0.9029852083471449,0.7594743990943478,0.28209511733383463,
 0.31770498284359106,0.06436229649761627,0.6315719648305763,
 0.24967943161949124,0.8143502376221168,0.9952719099685821,0.8913125982375846,
 0.2853771987738004,0.5323605289759357,0.49645132738145303,0.9750342189531985,
 0.614529603266297,0.9918854183579116,0.5038080512655186,0.7683871247630909,
 0.9994426555237602,0.9722536717087877,0.16705690766827563,0.8957251644930185,
 0.4772087682877334,0.845011796115274,0.051227447200858034,
 0.12429190277286062,0.8205189409133749,0.8388156001125546,0.6854910597967832,
 0.9800303634727956,0.7286687676819729,0.8144072979276942,0.49457162150684153,
 0.8157417727109953,0.6249276922481732,0.4853876863577562,0.05329683084728232,
 0.5471254576704058,0.12305081527208372,0.8898930892398335,0.2891624532651563,
 0.2617234757219763,0.564040475979683,0.4535501887657869,0.20033984594956433,
 0.09842443764520392,0.6560611816982423,0.43766455805373405,
 0.9968678012823016,0.8540931396735679,0.5016378535993168,0.2664869902181547,
 0.20470309144218524,0.5249779021356645,0.7992896443666009,
 0.44963847860313344,0.08224818570036707,0.28466641399619286,
 0.7098335155230006,0.41057189274034245,0.4392071512076501,0.1142736027407606,
 0.7004517884655405,0.4557940366078176,0.5497136004117903,0.3607893787680859,
 0.5147649711691366,0.5170003551685257,0.19658414504694466,
 0.017768195996273217,0.8304705516092683,0.49762774833814316,
 0.44612359086790343,0.9903559475763418,0.5558380106347124,
 0.07582596291014265,0.7270385808264985,0.12153923122451549,
 0.36317665078482797,0.1002405957245045,0.9653106366921363,0.9533842183547531,
 0.07169193611562541,0.15677824416587316,0.87912296013179,0.24465492597112926,
 0.886034141099143,0.5294658027752167,0.6301678009525709,0.0561617478002554,
 0.8368959819247287,0.16086687361646823,0.287968749757084,0.42690734737809966,
 0.788157315811459,0.4224522782904523,0.047973474237869684,
 0.19858636986430733,0.469614048571275,0.4686530918014218,0.575581509144262,
 0.7233214586017291,0.8432096904153216,0.028329014392163288,
 0.15555146393152575,0.12937061137341632,0.14398361665151216,
 0.9487510265828992,0.025743069760469917,0.7500314483049992,
 0.39664301099368926,0.8490333715158369,0.9853281935264924,0.9618067992501127,
 0.6487336734699767,0.22463344445952815,0.7494939775527117,0.516159535022553,
 0.16594808190921106,0.8796888761519415,0.03144797400005739,
 0.15899992650540218,0.0029330176785342665,0.12697042339307718,
 0.8733248769965565,0.8140378834870919,0.6316076510570322,
 0.051409377942702394,0.7188348784252203,0.7598444553841184,
 0.11595918376392045,0.8254407983439589,0.7857439153147225,0.6760142854794554,
 0.7559818132767931,0.5750161258375985,0.3450097011146651,0.9136481111533507,
 0.7944627415233133,0.7774231392128862,0.9774772657028064,0.598021934223303,
 0.21715040317493794,0.0701637663841348,0.2545371028122567,
 0.45559765976588185,0.7996946166851774,0.7043433449317567,0.2536202882290597,
 0.9167331467958927,0.7165269857143483,0.20463700356369652,0.6566506881729731,
 0.09151071666933086,0.6484300910004217,0.9734053832995448,0.8490613777655024,
 0.33142248571317434,0.46669942425693933,0.6962610200337823,
 0.48167259876586044,0.7226889766032945,0.6117107359947248,0.6155035935027702,
 0.5755117078829468,0.3456592327679149,0.6627148621902638,0.7504503183638477,
 0.2415838200481295,0.8599138238638806,0.18425622949659914,0.9608172098090011,
 0.576490006658298,0.8485096342016996,0.5355314920566671,0.32291244452809087,
 0.1377479908560395,0.437391447648254,0.34547824741372746,0.19425846752482334,
 0.16450366932557636,0.48415360623037607,0.25397880424364516,
 0.5756273029161634,0.05226322037823683,0.014624528124596359,
 0.3689050708014795,0.8133665625574252,0.17308103615257098,0.7405982905044736,
 0.7879855312013548,0.7317573097366964,0.5032140182870093,0.7809679862690097,
 0.9442652209493708,0.7580864477084673,0.19810375186001727,0.3060764523357412,
 0.643486666487,0.7862999267049247,0.10306742359563348,0.7634483483003779,
 0.04122977616216261,0.4924064868758904,0.521708184408126,0.42083032200246673,
 0.6323273615629048,0.7775488209685489,0.5958830882708689,0.43736813111724515,
 0.32304046995133295,0.12158003259480743,0.6019953465848542,
 0.03151145023554358,0.1882621776422022,0.5300912148438279,0.8184726393349231,
 0.7980278441215289,0.48493088850044097,0.2527248834036555,0.3729039990670916,
 0.7892901425816581,0.7480265487804846,0.8175977536104646,0.1792075977099562,
 0.31741067458546557,0.9182162486272994,0.5115539372031962,0.7442867433432425,
 0.657523965004687,0.06256300460859521,0.1533877007841623,0.8254109859173141,
 0.9310621991888733,0.48570220047112844,0.5584039303609001,0.925881369415615,
 0.33368480576148973,0.4693708355811229,0.24457564224244066,
 0.15981645693127378,0.5368936881843179,0.6972422482181581,0.994987258972801,
 0.5798798026240096,0.8063920706380222,0.026033292514472284,
 0.03787449177968205,0.6050354126232182,0.08085306197921105,
 0.08253122535572166,0.13467537799678897,0.4729700119877376,
 0.10224820411070068,0.08177494167649035,0.5839760030556392,
 0.7606180247745256,0.085058606319098,0.22150982763212923,0.3924280710567346,
 0.4038680431839178,0.9215214112728122,0.7369829354547797,0.09214229658345652,
 0.6320102626979287,0.4906789859067531,0.6408536466073622,0.9196557670726337,
 0.13321715275472756,0.5043096338379638,0.7804016981235006,0.9960003308436078,
 0.4442964287455917,0.7921155716714244,0.6709143742965945,0.48579348616014906,
 0.03300178768959916,0.13387163061281704,0.8988091630179891,
 0.9667141963190246,0.7815890306722766,0.8092095369635415,0.3371053689158485,
 0.9737375543821275,0.19117485294399494,0.0316618136215574,
 0.06843754039017502,0.9444092649958079,0.9169526578394617,0.4785409494649444,
 9.45579345795422e-4,0.31019345795302944,0.34170600380261384,
 0.5400026492106442,0.9674514716698157,0.03278368978123625,0.9105734975833133,
 0.9250906091078843,0.5724731302453553,0.7969571824253301,0.39195613703503873,
 0.1262620108702237,0.9783001237165252,0.5307690829494297,0.7354634175488771,
 0.11293246488194697,0.11960284120423925,0.5792793663039978,
 0.42309054393919365,0.1338513939386381,0.5375395406713586,
 0.43346076179522575,0.7524145768649708,0.29927180113295815,
 0.8743019240871934,0.7544711791078024,0.2543692844958366,0.617434147090705,
 0.21987766927983954,0.9952321236290818,0.063891067958747,0.3289144937541836,
 0.7892475161652932,0.7671264487847845,0.44101029116402135,
 0.07208868016943881,0.704885860172092,0.027786124825753045,
 0.07474720701909499,0.25584596814471805,0.8990037206771926,
 0.22885108859832803,0.4734332072225871,0.7606845739841589,0.3216680756486059,
 0.2925612412860825,0.7977251131856344,0.5485398464113465,0.858436762851835,
 0.968423517861107,0.6671247930446275,0.8871338758317256,0.4508689441578302,
 0.16574021120530325,0.8375112799981231,0.1391629451743316,0.4460185398462019,
 0.8713343118783218,0.18818048992436898,0.5045013016508547,0.5748177434689861,
 0.26322871351296007,0.0778792635788712,0.5275797638902233,
 0.41471209095464523,0.8631779114024364,0.4812477245903284,0.5920568310333424,
 0.1098103442574263,0.48641328148034,0.0563252837215491,0.6383712873527848,
 0.41269839331851244,0.08775936144899144,0.846533805211392,0.5887756248715887,
 0.5048444821333923,0.31512318423548435,0.6703477250959421,0.7182790310771087,
 0.22999697297089916,0.07750969232897187,0.2525275816099788,
 0.4761409214504735,0.272084658948076,0.41956406082198594,0.6702179339835616,
 0.6307868422614966,0.3123683540395752,0.5285756007058049,0.44058689269094886,
 0.5228080111682457,0.3632568056501164,0.3143220730740127,0.14794214378495507,
 0.827277080150598,0.6161868242081197,0.38245427615732674,0.3667896253033094,
 0.8079216032441234,0.5757796119824045,0.17637804317780825,
 0.32124738916905116,0.17037815012573543,0.16737771742391327,
 0.6551781133850234,0.012141523509911378,0.36682299850291655,
 0.20986047740430147,0.025549837450080326,0.379090167602386,
 0.1412373688081886,0.6709017031945622,0.6716600367308865,0.7204620812713901,
 0.48217245590504476,0.6579001434804265,0.6719112538919689,0.1684990287541447,
 0.38918508898631654,0.4592887918247286,0.43431448410705087,
 0.2810104336110564,0.4658321066326072,0.09897500239037227,0.8051166267627198,
 0.8890787045980489,0.8720604053275212,0.5640234235712297,0.8748099692646798,
 0.36663145440430556,0.6505145833272353,0.45737168836110875,0.398329894609315,
 0.49169907100926813,0.03565535225027383,0.3172393731604003,
 0.5376591775722149,0.05128219948025392,0.5639404991660454,0.4184831002533672,
 0.9843044664167947,0.35374063581758297,0.2412755738381528,
 0.27033552343877376,0.11488094236779212,0.7497148706088317,
 0.5677271946412621,0.07092473625461027,0.9340387842107583,
 0.18848828660757455,0.15819607422218795,0.9215830153619043,
 0.6066495708819692,0.28194811386833996,0.788131054684169,0.4354359024507348,
 0.8913291909766504,0.953653247654643,0.2440877309669105,0.7337133842655323,
 0.10780071609375497,0.3434555705966247,0.14419475388589087,
 0.6812378024431572,0.06171930628525446,0.46065047719149643,
 0.5110246257858869,0.8142911913097589,0.06142877804676794,
 0.39476046725699865,0.40935197472102747,0.5792237176680087,
 0.6508092914352372,0.44491897484573084,0.9507635543252666,0.9477850736522155,
 0.3746293650889637,0.18800769106657178,0.6189193445568311,0.6134782933198777,
 0.8968774003249699,0.00645566400458486,0.66100658052569,0.44509866410147103,
 0.2948715439896722,0.39877765535894816,0.8384440329626233,
 0.30700262007695045,0.4818680715862511,0.7529880855767578,0.6733668071295167,
 0.7291139607220514,0.17213225289962808,0.19351319012201307,0.74242835714594,
 0.4845639947213166,0.5472090807744059,0.9334379082646665,0.09222512464990085,
 0.05014040827294686,0.5415754409558668,0.06539150359388912,
 0.45570495728671434,0.04370500718875148,0.8131881569379305,
 0.19675568753204153,0.33516625031580927,0.9246282372492985,
 0.3943370066773695,0.8068766823083849,0.1945490056298692,0.6257938475040137,
 0.4514012274341762,0.23644885455791553,0.6408456257187332,0.6626034947743322,
 0.9644312326722415,0.9302520547708724,0.8932918630092804,0.3721543363769658,
 0.3419179458383639,0.688424494565496,0.9486824989332311,0.5439696967034736,
 0.32459836631859507,0.6326215750156218,0.8001652536484353,0.3989561651501653,
 0.3538684887267999,0.8971304740936444,0.7623072128391052,0.8830141079604383,
 0.5212846855869904,0.218233893175381,0.46227595612632866,0.9248199892464393,
 0.25688859474605086,0.12288340955607724,0.06850857056065851,
 0.7828560578438929,0.3291805185855201,0.07335749776423228,0.9992198820421199,
 0.5727531211141244,0.44380238921717985,0.06085882232495188,
 0.02296260140125883,0.10617216206786417,0.5529361084896028,
 0.16865674513985685,0.6555474939686111,0.34474291609465446,
 0.37594959907237824,0.5230719478952326,0.6361157200201057,0.9359371936492921,
 0.7247213648975541,0.1282432098486227,0.004591242762885583,
 0.5014075586070971,0.8384347478112539,0.9038496714301163,0.4922116501982261,
 0.8962984095961205,0.012082667937734826,0.313805040316428,0.6096425691886344,
 0.8636913773761932,0.6441276261439212,0.7810942051000486,0.7966040244725978,
 0.6732063310277786,0.9296196851988043,0.4558234002070919,0.7574425973408738,
 0.08905761006370272,0.07536279261669887,0.665559917879272,0.4587506998141909,
 0.06000737253409705,0.4171752202475393,0.3360748961086436,0.5803246042798693,
 0.5214518995444294,0.1246615005109305,0.7115907671575019,0.6712243389082068,
 0.9519712326777868,0.5276793217530339,0.6725630781964145,0.9301393517426566,
 0.3386277111661301,0.41663675180891513,0.44499638435264555,
 0.32644900725766945,0.41536487893712437,0.3441073492619905,
 0.8722816938370619,0.3088379312257994,0.6109531343430974,0.4236724312877176,
 0.9689368004200228,0.9825987270751111,0.366599487185609,0.6383059753820097,
 0.7654700574004768,0.9167320128618152,0.27529050255020915,0.2421753481804242,
 0.6303351165047534,0.7928930648951239,0.7471971653319891,0.3778935399315483,
 0.3674080110107285,0.9423741492603608,0.8053289749974955,0.869729285395406,
 0.3330650518626478,0.6142013274122835,0.9535222007236794,0.43869479071263284,
 0.8374899381438042,0.5177721961372216,0.36549258609730617,0.6622402829333843,
 0.3426243952386815,0.5439712464178998,0.7406097927926828,0.5265160434667139,
 0.24689156424435432,0.215722376673803,0.2667260487312386,0.4981432870251029,
 0.09329112029465159,0.9860677706080431,0.3853538622525484,0.9118087063022507,
 0.006904706169470476,0.9227550591225673,0.7424774745577806,
 0.29818442198214057,0.41715846951609614,0.3023301602748121,
 0.03336992142550177,0.7966651699428893,0.4165471794596083,0.6411241872694882,
 0.45573187061662723,0.5776399172342853,0.7635913379672441,0.5531570562653251,
 0.26082045257097664,0.1622515713700835,0.627098149033034,0.08486923871570973,
 0.006759746095172714,0.4278703502658321,0.2281665590276175,
 0.7686313026866866,0.0654662090250846,0.07413203982692207,0.5425682399736835,
 0.5903704329776236,0.4235166180324592,0.20965779854772182,0.5166494430593683,
 0.3803454814834951,0.5359834823693597,0.10379531240411521,0.9366322860961258,
 0.22196771307656116,0.09650847933690498,0.46847328897944185,
 0.3146238473317471,0.2391489809562678,0.6784357812764845,0.16265681690334644,
 0.48122564213416974,0.5280995829867328,0.6648978837657874,0.6659905125725307,
 0.24667876317059434,0.5030087254389254,0.6429085253056513,
 0.25112094168711996,0.40618782394754493,0.1740586509625438,0.642401415258273,
 0.15937847690676676,0.15972307986920542,0.5824897251512285,
 0.49029290576258244,0.44264513824395946,0.46433639256428894,
 0.9453019489758613,0.5830600581153268,0.3602714537086925,0.4876343378666391,
 0.40761336395899705,0.46663763076191556,0.1298531231383837,
 0.07884880972916297,0.2625610581471778,0.8733620995866681,0.5716428791972152,
 0.8295286020132531,0.9865471049855348,0.04651100841403244,0.8355384559049385,
 0.9415647031898515,0.9661789097717104,0.18604264750463628,
 0.024969528287235643,0.1862907417406583,0.7538438155663785,
 0.36513225900882595,0.03871167425373301,0.5724898966306411,
 0.9005480997088928,0.24840370115784105,0.01898503949569519,
 0.40310936553616883,0.43713187531125675,0.9173255035221035,
 0.9234343877192184,0.2919262073892125,0.7292486255844509,0.9288761417196304,
 0.6434307177583671,0.15548666278464285,0.37432159009110433,
 0.8593903695354449,0.8856480150451309,0.24913257041303283,0.6337858276502291,
 0.5043634071279423,0.40310642864644874,0.1356786822724445,
 0.26748437149550475,0.35791481831303873,0.5454004158170881,
 0.2946331286395534,0.771306302787663,0.5676106769141166,0.6505446216781571,
 0.2788392439144849,0.7010809606520108,0.9060980879171472,0.5348154789506938,
 0.0023300196163911213,0.31144772479355454,0.3821791442489074,
 0.8426299065241806,0.39861312763870205,0.38629534962115364,
 0.2665630144861263,0.30924045498664654,0.4447591940184251,0.344382412508631,
 0.8916459742195243,0.7792364979174533,0.8407814869022733,0.8880458996907217,
 0.7462860706410753,0.35881756184713043,0.8217351318629519,0.3746815280044027,
 0.6317925994309778,0.06521672924745903,0.745562258270632,0.953358235929659,
 0.20211225235889252,0.9378160958579366,0.9497806620257491,0.6105594051127793,
 0.7459944054315114,0.11277113332240551,0.7353335295271705,0.8639936219302226,
 0.27973022649395074,0.10350086413702964,0.4058275895878607,
 0.5779328624067535,0.872813859042133,0.2422896960389057,0.827531075594453,
 0.5220034956552527,0.36766552860234913,0.3352647272032253,
 0.45323223058918194,0.328284046522362,0.07646819636856872,0.7612429185728927,
 0.1131850375913308,0.8196623998951008,0.320969713861063,0.6793004961959074,
 0.3254289693015874,0.5790608460541603,0.38329115746168485,0.3271988543024419,
 0.9095892112907074,0.04336511026944834,0.40270725964491794,
 0.3486356916592135,0.9962920672105129,0.26319167089336415,0.756138505187369,
 0.6303442457407125,0.6475157012881101,0.8433000994728159,0.16375141560619744,
 0.4828766922485649,0.05242675824362253,0.9085848192842306,0.5146288351274163,
 0.8336458586977761,0.5802997413707957,0.6309939297500919,0.10869163154970662,
 0.5247879781510101,0.7853363277395955,0.05590005025167777,0.8643149658043772,
 0.8500754559861478,0.3184084751525388,0.3204657822230076,0.602227161819188,
 0.7952295649076735,0.8575669153878506,0.12271858859176499,0.6135474236280098,
 0.30382620019337225,0.18208462754566246,0.7745162206979865,
 0.8158613435728111,0.14202173161968523,0.874115801636546,0.9378595873020651,
 0.8185875556001682,0.14824151100211924,0.2669646045178393,
 0.12146711895140105,0.40220792264069893,0.41292014059617943,
 0.10988235593858886,0.5303713872040883,0.3927885972712041,
 0.38982014709543344,0.46644476562527715,0.05619757007652404,
 0.2697398271728748,0.17170382376529392,0.9428263134316007,0.5320459738862399,
 0.6796921599233035,0.27208495773861163,0.9691889884678577,
 0.41486773507007746,0.49090741638188085,0.6545804320306527,
 0.5458952721278761,0.14395139083510533,0.8104684408610616,0.3053844542607037,
 0.17032247292294755,0.44640875733057184,0.5901696172263533,
 0.49151622027423847,0.41566806987290517,0.12333443271026823,
 0.05490799868984575,0.45664767460174505,0.8082920321756771,
 0.13978752526495408,0.6060132443746342,0.8865340639490091,0.8544749227139224,
 0.057026060035957005,0.49752975174106373,0.172678917929033,0.552130420561798,
 0.10496181172960006,0.412654451419739,0.2736391544254695,0.40602316350537915,
 0.16367757175112752,0.7766499635528616,0.06527929439679636,
 0.9929190275617114,0.9588955837875276,0.6238648446278982,0.8482887711981936,
 0.13150034608857675,0.8269953409384856,0.9773060270417978,
 0.39098143481366454,0.3472476010703226,0.5831918974678704,0.2755995813101686,
 0.08068525333911014,0.8535287002517321,0.46490398519162657,
 0.7602829324290926,0.5209428596360676,0.5967809487567245,0.21717340449929867,
 0.06179463400284746,0.5292636004582049,0.16318289368052796,
 0.32921707986817794,0.7044737151296322,0.6218454057663474,
 0.49618528674378226,0.3172615786119657,0.7040064210004826,0.4952346247357975,
 0.11809165423823464,0.8638956521241851,0.6960936214890848,0.6834673629774957,
 0.5741209633326873,0.31666577861984924,0.04702546106074834,
 0.6916879091544093,0.16649535118506043,0.5036357551835078,0.697857690825384,
 0.5910087979939327,0.2836799034139805,0.49222802473518745,
 0.060673341744261755,0.5628153592148748,0.48269223334796296,
 0.9065082265006206,0.36504096758709426,0.2644749912921224,0.7864698459697949,
 0.09083073768966365,0.26027374548914795,0.6668621375291884,
 0.6525950792205092,0.4782683891196866,0.4084905526920751,0.949060534195392,
 0.6778836276144593,0.904459695488142,0.11095998920024575,0.7372765725143825,
 0.8474831765620355,0.4945035835820335,0.05629366153084492,
 0.16543514383007607,0.9416664003797319,0.33671048067699416,
 0.1317820538822444,0.21928986000061357,0.02605963212918283,
 0.38316756301574173,0.4335426126621583,0.22697099639818674,
 0.8868817678044228,0.35617068675869434,0.48932076854035456,
 0.07766790030729354,0.02334014967465703,0.07083249750893017,
 0.7086113709755243,0.4269761017944219,0.11718815103269664,0.5689101805034117,
 0.03035657322359353,0.6953499281241727,0.1704183491848188,0.7744422628812133,
 0.9066072590404344,0.8749398994812848,0.2362978059755828,0.18083108321545438,
 0.039021751145770334,0.7338115008079306,0.9204416408312811,
 0.39888516095849336,0.7497771873205283,0.10065035266241207,
 0.6912923727600893,0.050943780926137494,0.9287913952867435,0.912641111687323,
 0.1994400575429084,0.47392513509840384,0.8101012880230156,0.6446294351016463,
 0.29726616548866713,0.5910766919494074,0.6637308218280192,0.7245289975754422,
 0.29371226270997586,0.5058542782443214,0.3578905845858644,0.5266537147635965,
 0.6799338658892314,0.41251385089366366,0.43449949251661935,
 0.7095060165638869,0.18088472479428797,0.8211955260673041,0.5764401622375955,
 0.6468223295712301,0.5240159946759763,0.3079781793744264,0.2795488196846948,
 0.24289540694365486,0.44139182311055913,0.5334596315981883,
 0.19978091821083166,0.07223781394738316,0.9436733615221173,
 0.6115469567876384,0.4048460352866412,0.6795949979159387,0.7815920721698107,
 0.4247412831947319,0.8617436883537367,0.649102052381528,0.4903344178349285,
 0.09607863009021922,0.45526018214510056,0.5494156259988501,
 0.6375032600685007,0.49930927417893456,0.16361518391696284,
 0.9416736929049513,0.4906743379283247,0.1084133718474859,0.5812468686015005,
 0.04398980000366892,0.4699801976882134,0.5086461038051782,
 0.11981312974787628,0.4004960897480576,0.9622981700909324,0.7003421392063869,
 0.45561162795776455,0.022827022479484982,0.18657670959250772,
 0.4164821894692129,0.22948091225182699,0.7819271262100997,0.648720014516592,
 0.6182266713143578,0.2706380402293924,0.6273256091079504,0.4869543449989229,
 0.7282318937906204,0.9051460296219074,0.5925125588820441,0.3801705335086305,
 0.15427057886860862,0.5871865017147039,0.3145085298688469,
 0.15920084464407214,0.6923157125529185,0.16965712872416383,
 0.5194972799205754,0.8953508963787784,0.3844543955118135,0.41330534516951545,
 0.7629123787706991,0.9601525029353133,0.19146850420032768,0.9693202758948484,
 0.4069898880906053,0.8893312720465579,0.7385385957176449,0.6111902951735303,
 0.6818458492367633,0.39776671794599694,0.4888499981530643,0.8913877133019091,
 0.9379456176779305,0.5866800167610715,0.5752467653630828,0.36485153926234515,
 0.49872663135815,0.4260694776591767,0.579684285176241,0.35506519062449615,
 0.8058629866657328,0.2577771436986026,0.5445841721785962,0.24260042328596543,
 0.8630051936496859,0.18028475234938335,0.39326165788936374,
 0.6108978308970193,0.33232465116897214,0.9935258778581445,0.570528879852483,
 0.6644960990915582,0.672987252709198,0.4085574214860427,0.8920414182632236,
 0.5100504674313127,0.9533351635002743,0.994363094587117,0.5564254407578012,
 0.7298290443830839,0.41791976876300096,0.2508679165932126,
 0.20830931120999285,0.8697172952746706,0.6889043957986462,0.5927371474181536,
 0.9549233002121078,0.47316986994845656,0.6330420917616065,0.597315250969141,
 0.9915684542911022,0.21346746959809848,0.5524434237920727,
 0.15035796415812874,0.5759433859419969,0.8886643917677128,0.7615941599843967,
 0.6882727027034494,0.5776111606469636,0.8986332173605219,0.7970126699100071,
 0.5374101975177884,0.8315109400903653,0.4673114699302219,0.5240925249828848,
 0.033150272815922976,0.24300262183738264,0.9750496163639862,
 0.7522181532319714,0.703564392479449,0.5550542376284092,0.8218177622813417,
 0.9085945892996445,0.9442388372593598,0.4170908562637793,0.6669112894270051,
 0.9149257371288442,0.5031972949530914,0.4132228517276362,0.7611423207818496,
 0.3882906301787735,0.08750953920565319,0.5567902407266004,0.9650976490616124,
 0.6912748522154192,0.9424515214214991,0.36550668914493,0.9936630110978272,
 0.948709256016177,0.754830985859461,0.7430926868331826,0.6976780384271799,
 0.7199953384617358,0.8724942889882659,0.625392329543067,0.8585139952884618,
 0.3068021249052213,0.867053334025303,0.29890818387731033,0.06302475245684813,
 0.12532444757462935,0.20770017413258823,0.7266088814089502,
 0.7955443623797075,0.2613392386364961,0.7773140623118102,0.7802852261533073,
 0.8127185924944877,0.7877723688146269,0.30982765652354605,
 0.014511418436857326,0.5640686103847753,0.8231560344929162,
 0.4003345126407676,0.24237841261134152,0.733755785118877,0.3279715097469692,
 0.007390807906334951,0.9879018619331805,0.4577693939850733,
 0.44430289885384555,0.1389146934068235,0.21451012645641943,
 0.5633876055762859,0.7555791581193422,0.6807378551605344,0.32228995040186803,
 0.7793133734607987,0.6045527959495356,0.3412526737799365,0.17453318274708285,
 0.5087736279072343,0.7847144550975325,0.43041328190404426,0.6325812087447115,
 0.18844156227687792,0.16885321586498592,0.0284771744568052,
 0.9209502934147638,0.7190481662867505,0.9003602334352973,0.979517486480871,
 0.6001872220140461,0.5233438311390662,0.9271705997051957,
 7.094763489581268e-4,0.3704330428066368,0.37108077081083113,
 0.7597619639950582,0.5967954954194513,0.631247316659375,0.184336823574777,
 0.968422712333815,0.5377808360547032,0.7784105838958458,0.8935242705976784,
 0.567422204668241,0.810848910546017,0.6116525960410365,0.4788709872121162,
 0.3585200003257679,0.029627190245504753,0.374239714874227,0.9563075021119885,
 0.6349348851403841,0.9159093327087539,0.12949826825453725,
 0.28162787083102825,0.4814745205834978,0.935092545025386,
 0.018360801473811694,0.42889980146583573,0.6059722721559135,
 0.24369151649216692,0.698372029215321,0.825721352204144,0.8551389299175035,
 0.6508684266050588,0.45515693013544856,0.4623365382145548,
 0.01522812542699803,0.20439446943566797,0.05827649311584182,
 0.3365274607759897,0.14115268170226625,0.09990628744812358,
 0.14420465354746903,0.1108629607983278,0.7840621209804739,0.8980127524135302,
 0.17436239929826436,0.8990993806323974,0.443082910458665,0.47979395863371654,
 0.5829957930728537,0.3539916888896876,0.5786446595677672,0.4131625026524076,
 0.56439339052259,0.11328561684163396,0.40205187304434586,0.8249954347753735,
 0.9147753411368297,0.4107483761663362,0.7990093101319542,0.924103157519485,
 0.4646511647530218,0.06612075171854659,0.3791479598148638,0.6688627686161669,
 0.3285559724911329,0.22312940111028523,0.09292197323996654,
 0.8294074943919422,0.1249245902613525,0.8478827239213754,0.2714539571587131,
 0.6247948235389851,0.3657974697221219,0.15803417668680098,0.8015051478824706,
 0.34896002647449764,0.540453710522139,0.868421908063645,0.0907285072245082,
 0.6801734779589503,0.15783567958349698,0.6712136968019083,0.8274120752591044,
 0.7847940804536147,0.2245235699209216,0.24366858045511963,0.7126335360884024,
 0.24117567267084283,0.22003002267254412,0.6603348879291877,
 0.2960494944186045,0.6507586487647292,0.03556191682521137,0.6385181342014463,
 0.9729065450713562,0.5175655056524225,0.6721452529059772,0.6041488642496695,
 0.6166816668370777,0.6267504094787371,0.09379149725099867,
 0.47802997494865185,0.4812160740320317,0.36426868652096867,
 0.6900964764389762,0.6539167735695247,0.833290435842154,0.1441377018031086,
 0.25352623743236713,0.9887325447776405,0.41616414963736736,
 0.4235063595246912,0.3876277018264477,0.1825998060591707,0.41916585909376514,
 0.7230492691387913,0.7921406778305244,0.5088839996679109,0.13824619062494137,
 0.9429249939940944,0.6273415556272816,0.9586452553269897,0.34003772742288363,
 0.7347967057459918,0.5351123313427851,0.8785516353093101,0.5510325079616537,
 0.6621538702081239,0.8546491019796362,0.7579988539643551,0.2557433556765837,
 0.1290932246948523,0.7928649008401758,0.46294148450263084,0.1466190695590316,
 0.7550613471559144,0.27119870549472846,0.32764027665437534,
 0.30937049171462716,0.2269041675033756,0.3990034182151745,0.5514658759381955,
 0.49453125973034784,0.9000673908276197,0.05160911207622798,
 0.33194446015086776,0.9858087041265063,0.05956916634396636,
 0.37235544649288843,0.5739056274601507,0.5308617504176596,0.9805757995664989,
 0.3689588801001217,0.39752851255064414,0.02347149915322555,
 0.6494155257584145,0.493086660429095,0.19074440743463872,0.8504165798408578,
 0.4383187824440268,0.5912556019589722,0.6919190580848049,0.1419797187880547,
 0.692208801820384,0.06716234148723665,0.3232020631272744,0.1761972720007008,
 0.9603716837394562,0.33907970336696447,0.8668731745730438,0.7641321918426601,
 0.37922990270290347,0.20390252069490133,0.6824386115483078,
 0.3017624812569353,0.5558754546594236,0.7736109345773236,0.79548788843424,
 0.5216364771640569,0.5364770089932486,0.5305119978696587,0.34606955452150445,
 0.1813928914291828,0.13666376997175134,0.2925535051542495,0.8417690049888014,
 0.8874663538624461,0.008262989167429957,0.7273786894649934,
 0.9835540006374967,0.4962733770504949,0.7786944547254545,0.42607557736973733,
 0.4835996161448969,0.8538416742453685,0.7314647064522063,0.5132471138854877,
 0.006819049624589346,0.6411732216941084,0.5387500915157497,
 0.8853718555452396,0.6376771620072452,0.9470347791041421,0.233455829044655,
 0.7239659110056311,0.30734682397707214,0.5147271866521983,0.9891501332732657,
 0.14159991601423927,0.16817094006159516,0.08270379971101094,
 0.44885188959734335,0.8657746014386438,0.6204505937682447,0.6928351477070048,
 0.12054908441642298,0.6615637370056293,0.8963696581275141,
 0.30378170617720857,0.27300347782760204,0.26893092951329534,
 0.20868572243833827,0.9073902729888714,0.7126166424689606,0.5036066151176137,
 0.6877684580855847,0.4862241342838586,0.9699868114241529,0.8835417097101745,
 0.07626378309557669,0.9018739286578262,0.46323295673769405,
 0.40016094643393396,0.5190032075950405,0.9219766701218393,0.5801334165142967,
 0.04562398356566266,0.15668895656892023,0.5104975123869142,
 0.015223114707257901,0.760432376921007,0.017184020732571637,
 0.4840136358524152,0.33301592565181326,0.2818749907987488,
 0.31777934365273075,0.32648544913829425,0.17651104313864163,
 0.8234008044623726,0.5890417773748713,0.23542524474702797,0.6985017943829801,
 0.7429473655762712,0.16069975898929056,0.15673375814914736,
 0.6462259476171713,0.9187828247386988,0.16860043283721304,0.3111982894776353,
 0.8695732749703893,0.2773418422554198,0.8272701361918322,0.4648208033792194,
 0.40558448966173555,0.6695458828008939,0.748838164680824,0.7305941371634823,
 0.37977982013443135,0.2530659277384536,0.8455543892108862,0.5091893758913997,
 0.7392435547736967,0.36981072831979866,0.0812341333169675,0.2907035041050491,
 0.5451380356541634,0.1633275872095975,0.5957137482372518,0.24066515373092112,
 0.2585231409497236,0.39765737128279055,0.047009469314618024,
 0.672208950418385,0.9750285513722958,0.10140314724800348,0.30630072951561027,
 0.5195944518654987,0.7303019805174046,0.22147483404849844,0.7979976534590231,
 0.7600566279114184,0.70922999591533,0.37698708012881865,0.17673928515946535,
 0.5060401281196261,0.5414795780148469,0.5340993592111032,0.8947323982360235,
 0.009854231031197647,0.2709570886007573,0.76526126640687,0.27191864599688675,
 0.9841369813315356,0.8344232966277814,0.21451842304137203,
 0.39365086643689073,0.724979820828262,0.8161050366739946,0.9471846460504786,
 0.15153363859168056,0.8077399086719492,0.26909382269710347,
 0.6655710228042565,0.9532430821581547,0.43447410812702525,0.9465495533452228,
 0.6810985342458734,0.585291610371504,0.726470360808364,0.2614785110971213,
 0.9806519739711871,0.9634963438993935,0.8809153254631764,0.6752128118639024,
 0.44232706095953267,0.893808382999255,0.2322315007065867,0.4769522697037941,
 0.2888929362980529,0.7605274926835331,0.33835193915162975,0.9104706484256246,
 0.9089730932786224,0.7030905307160735,0.12550458000075448,0.3604744309037732,
 0.6851149207340224,0.231965581402104,0.7548244824007435,0.5037328881128643,
 0.7977421948491912,0.6625135605706044,0.004087168348962833,
 0.016667358968534995,0.7211187686786341,0.41977037028488284,
 0.5635356616318931,0.7549814123367957,0.8737124950578036,0.730318919150776,
 0.8152287721846377,0.046486249735488205,0.018200195112246798,
 0.8530387493681242,0.5881909808626907,0.7179347969597961,
 0.037788187507974524,0.8383225503969823,0.9149922516386313,
 0.2587272690517639,0.8776928414564285,0.17607854610433238,0.2354951521424402,
 0.22766802884194937,0.17281728640837057,0.5421829771069614,
 0.5499063236408189,0.39590143819386614,0.9894090864249362,0.3492110413929287,
 0.02530195752282305,0.17488844859607955,0.5549681865340564,
 0.041398793111092846,0.6606558712525115,0.14609174669942204,
 0.9321195831277195,0.2866179272237691,0.06497937339612858,
 0.21307520791230938,0.5581574813554683,0.6462395874872231,
 0.051182963067037957,0.8525169451667316,0.5262236021041818,
 0.22384623774361823,0.019099490650277384,0.8763293778441252,
 0.5282293094780078,0.24836790931045938,0.5157309794592272,0.5306612584651781,
 0.6837853570279648,0.9947999099467109,0.8106061687771919,0.1952258856818443,
 0.09371107259127587,0.11394419828105351,0.17509106277831077,
 0.7477092065510381,0.30653885618561416,0.5375944708442585,
 0.27226715935832346,0.26001708127942025,0.502365562118916,0.5523947792578968,
 0.2696050405463979,0.1953173673808949,0.16404330977943538,0.414220804722757,
 0.7157472158832431,0.317646977644223,0.21643309877547023,0.4788364667554985,
 0.34646100292874993,0.4454574261566767,0.1356883795801085,0.9632953742106765,
 0.3432957754899697,0.7716961278017938,0.8331603248158271,0.2800573156690389,
 0.628613280959934,0.32663781078747467,0.24334136616932867,0.8971364798184479,
 0.4041125734964406,0.039080107992605195,0.1056517769527654,
 0.6368941705623262,0.32508847085683634,0.40135124809685896,
 0.5183983603933191,0.6600542162054226,0.6160519371353307,0.7317121735282612,
 0.37930576324947984,0.0060366215604398255,0.19679485700599564,
 0.43313068596033233,0.6761025391018705,0.6804448739399116,0.4290261440210501,
 0.4484494626106904,0.6141760022260759,0.9722618324761356,0.1533172303433321,
 0.19421293397336215,0.5398315232262751,0.11832543821371078,
 0.1900731452247959,0.8753627379286824,0.08248591915517145,
 0.45769280406825796,0.5525568919882676,0.035504184214014356,
 0.6223488407080753,0.6206860822610643,0.21108379207530625,0.5947062273805217,
 0.19848714388586286,0.2979182358943995,0.6717898865236321,
 0.01753311970959981,0.6791875856013901,0.02814055024871176,
 0.6952996503052841,0.762800555517051,0.3820808952711553,0.05928671204215885,
 0.4813245514181166,0.320194523631791,0.6957309876854131,0.7157692997611764,
 0.07388667892624778,0.25605578421947117,0.9112235525147432,
 0.9062836391763155,0.06676303276555595,0.8554766722297482,0.8413136768719758,
 0.4680590611304457,0.5215204704396059,0.001312999014371763,
 0.2011439212864009,0.9065354063370282,0.616271783762282,0.42611001118912983,
 0.5170048313216803,0.06277074474213817,0.2129571922975565,
 0.005085443248285526,0.2740316842210889,0.06822504233989535,
 0.1258091002351136,0.5250698014279854,0.3677325259337483,0.30298112153257784,
 0.8945205343088591,0.44580051666128095,0.2838837000643146,0.9100622605698857,
 0.5724218864663163,0.894979025978354,0.9516711299241769,0.18578345088765458,
 0.44441545836851215,0.41926417256341175,0.8897860674697844,
 0.6412967518280226,0.28908192699886603,0.27590864509929625,
 0.42318227932544794,0.5142627499771951,0.31758441436684814,
 0.21215087572948466,0.5977806598590356,0.7810503256145145,0.9246789996645611,
 0.8779304093030252,0.3402351553976496,0.15777186588734893,0.6759495908339994,
 0.9900114144578639,0.011653452092888816,0.3624660751079366,
 0.8403764093132806,0.41078373312224004,0.9344397061340854,0.3650907649298112,
 0.775641659746903,0.1474844159279427,0.9142206060539239,0.5032647257366859,
 0.7437042758307447,0.053861132653699184,0.8179220643258583,
 0.017238200550099947,0.4122296880610433,0.6550035040089915,
 0.5319435260161274,0.6220164406924702,0.9901771161406268,0.408723544105437,
 0.2292301924043545,0.31857326563657606,0.947267034821975,0.3062358580573675,
 0.8652795085752599,0.011938624333150849,0.0771676241464162,
 0.9965689353195689,0.5273267181075812,0.6035633715178562,
 0.050517154113104756,0.44378921150430983,0.6529308647425117,
 0.6260340373291413,0.4072484168731024,0.6228316679696102,0.348590251390713,
 0.0036716556897893327,0.19307933959491008,0.5530154885259506,
 0.7130123888414146,0.2234818584386319,0.2160345564586147,0.6299479830853677,
 0.9102645194871241,0.07557056341054169,0.16151375051262873,
 0.4060797917932757,0.7526003618757354,0.9345750562260363,0.18627773947375592,
 0.736518906920617,0.44175125674241056,0.7095803503515079,0.15057900688426007,
 0.2371280869307535,0.21644622967212845,0.07088973873771132,
 0.024916745921421346,0.28981215420183026,0.11221061067881533,
 0.4864003494076393,0.08142229881796204,0.7047014838916504,0.65230743365799,
 0.0045302632039057045,0.07612074103570843,0.5594397191321371,
 0.33895824021704346,0.5809662271546806,0.4678357062130616,0.2951050403298301,
 0.02439373816245549,0.013518207689347639,0.35455786988924654,
 0.1346475735257494,0.40335856724600316,0.5705420746940186,0.3013265665065449,
 0.9644458527725805,0.658034800000138,0.9682708560589123,0.8055182004825406,
 0.025307100923760317,0.6050222225436135,0.4090403631289019,
 0.8390437557282451,0.5259291306927331,0.5742200345706295,0.45143721754347643,
 0.5535563735928849,0.21296035975764394,0.8581540812984965,
 0.10986550774172166,0.7074700043005756,0.8164503876673573,
 0.27811992532766017,0.30770972809352015,0.09262881176822368,
 0.9594197036007076,0.7166066076209165,0.29201160125551373,
 0.23124734667778968,0.7703350994373738,0.08617781617385245,
 0.18243667818984655,0.4842304383552182,0.3417020038036649,0.8097153204585421,
 0.07467949351002745,0.8704936285728082,0.5618564258039959,
 0.11405812405908144,0.724011931657196,0.22344895407568177,0.6873951664571945,
 0.3689951465357775,0.6685570263743961,0.2428728239118778,0.2527096939458836,
 0.8130132395944666,0.5090701691357951,0.6221915693384028,0.4811061882211105,
 0.7922994496151032,0.17727172976723315,0.13686773816641518,
 0.25201723144440424,0.04787237255385757,0.5126779854099829,
 0.7026276296288928,0.14792575576020162,0.9277447415042029,0.1629467275800498,
 0.3674954533186029,0.38321904824337305,0.953878608718954,0.20182055500333285,
 0.7026507350400463,0.3609208902750398,0.04449321847817855,0.6082222182068887,
 0.7647183701844681,0.043073740567240226,0.47543236736357,0.1572994737867488,
 0.2676562760712229,0.4363468896266479,0.8770072739540147,0.9819200540667223,
 0.7704104451996525,0.18130700943662326,0.528836366509994,0.7114932027934201,
 0.11929236388397524,0.8914852146302864,0.8233737540011608,0.8678826342961792,
 0.5098246385899482,0.9997948018393628,0.4425956093289606,0.271327574206886,
 0.4627233497951935,0.3534507196860983,0.6858737172486142,0.5790297072364107,
 0.9005260279307948,0.3874448866683463,0.5566338566115879,0.975892526925207,
 0.025548727509739355,0.30633539083681494,0.3054418133416469,
 0.03020581169804637,0.39820782875789285,0.5849097501237515,
 0.43278911780795704,0.04545198043729082,0.9188463173494031,
 0.13236538811776954,0.17929724770484878,0.6866018679987769,
 0.7024776361395784,0.3311777632466447,0.6740024315020146,0.10347370671671641,
 0.4764852710533287,0.6306590025369159,0.1919736887331036,0.08567970314124884,
 0.8367837354159053,0.9296860581234627,0.5122432733583598,
 0.028996177558755387,0.5690103511881781,0.08401099735331352,
 0.7645562654608943,0.7617435435006812,0.048406763797416485,
 0.7709835870631079,0.08862036704581422,0.6652962832517237,0.2104825226382282,
 0.44162692122374325,0.3775115169088852,0.9835007712387992,0.8704284492970531,
 0.13447507171590134,0.8880479795864941,0.3414432732108219,0.9815143979482307,
 0.7878630745398147,0.8693869328802546,0.7040594366370236,0.6441373154913644,
 0.4441140876617957,0.11498612088922344,0.44704073098206987,
 0.4540329204595287,0.5215295592452925,0.3653169050730283,0.8006642186181792,
 0.163023148647927,0.8752695732499128,0.42768119761206225,0.8849404104372899,
 0.5695269181177597,0.5394722576933061,0.8480916620520462,0.5517301871986695,
 0.943073185531295,0.626486800285678,0.40444448070489336,0.39019511290470077,
 0.7625882023303562,0.7028616498050695,0.9189893800807849,0.04014279894448036,
 0.1301683177647821,0.0389166393580237,0.9827261815452492,0.5803383692877189,
 0.8775768212929854,0.31633567037919696,0.17452424213782747,
 0.3590765412028598,0.5450561756361874,0.7202109710538855,0.23309459662779386,
 0.5711161828345546,0.8053734464778587,0.008246940658176105,0.7650559611607,
 0.6283467568202066,0.6717855639554151,0.4362180723840795,0.5693090003655266,
 0.7673562056989339,0.4820790251101672,0.714615115089601,0.46406702437637337,
 0.3897888481020556,0.9836872869153875,0.38172861336358355,
 0.32858860835883563,0.22470341454707965,0.21327585953929207,
 0.7547313187479846,0.17867289967635724,0.6837229940807444,0.5840968924460798,
 0.24172685823157358,0.9680593118235127,0.4955574683105515,
 0.44206460094300803,0.4867470955672111,0.2574976901933065,0.8595594525125119,
 0.8103328817485571,0.535124381274952,0.9259013488941585,0.6943022768777574,
 0.41868131336276826,0.08335116875951509,0.0889097185855483,
 0.8241751798063439,0.9347290579463765,0.24719408578617308,
 0.05423388236382087,0.6102924050407824,0.6894908601210413,
 0.38953342196622653,0.5917705997270376,0.5449548479646393,
 0.08974390292182288,0.812743352810301,0.4758892447283495,0.915673875413731,
 0.007340950266598978,0.6339647318379819,0.8902879182280847,
 0.2770399155543628,0.6356353541894638,0.6410116574257924,0.24051817080000026,
 0.06713996738842776,0.47946256549309396,0.11551783132582194,
 0.7892064054497795,0.8374918604047866,0.4368408825200524,0.14734391132570535,
 0.39807121350914065,0.33675212234624574,0.7121995536398462,
 0.4391484285927405,0.1479658340637796,0.828454606617556,0.4940907229345479,
 0.285011512122735,0.9872257427683119,0.21782936327011748,0.6636095374752814,
 0.599306311638337,0.30780155637635076,0.4696613080267722,0.7819023776889172,
 0.6024724947449608,0.17566594834910576,0.5428341048554116,
 0.49198492450259357,0.008240569762519745,0.3112073077557296,
 0.8782608340279153,0.8881837807423116,0.18190404250930536,
 0.20423258950315515,0.09248323365237399,0.9178561848253098,
 0.6595965896182979,0.05849748121578513,0.7584877873410726,0.6528390725291805,
 0.34978770467528353,0.6575693515870632,0.8093222258361007,0.6672870015276219,
 0.27747940747209665,0.9520176460602204,0.13985029196778143,
 0.07327049217263504,0.9067434927359667,0.976591740744669,0.0787486954496065,
 0.9840031655901297,0.8756015654517042,0.46371481980559126,
 0.27519990959800844,0.7681733642218413,0.9735255994102218,0.555219339185055,
 0.39910109256771953,0.5883173314747489,0.6774713129689456,0.6540189875886244,
 0.20133460479409748,0.13635716177359924,0.3281524034440091,
 0.0013393369674721267,0.2681382248887545,0.046627313362649914,
 0.6019127628887941,0.5751076536715098,0.9097649381812496,0.02273021657357588,
 0.5261370069056667,0.9162998908558223,0.6504007626348806,0.39278691642850316,
 0.482656685496346,0.1815295893202573,0.6129768522212988,0.6945125585019902,
 0.2537169838480592,0.9492228206798293,0.09689402752193377,0.4279620011974433,
 0.7181202041337357,0.886648822196313,0.8752456788463672,0.593571849767043,
 0.7747325902658861,0.7560192756672033,0.3209145789088902,0.703630989575854,
 0.014017898802711426,0.7607664340461224,0.597174095935197,0.8595829377203275,
 0.9638630647897436,0.7200154667295852,0.9879903370452726,0.1570464902328863,
 0.3446030559739801,0.8261517207675673,0.1216707108864884,0.465042419204203,
 0.6252898302308162,0.728286752149829,0.6523265628829539,0.858690894269412,
 0.4859918901434046,0.845074543263116,0.7681707054738512,0.196075252691561,
 0.19601783575622433,0.16989426730675872,0.44521155801129364,
 0.3957532722838242,0.1469310708337419,0.250919667244641,0.818121289379267,
 0.7804746906454709,0.41943996338040423,0.5838654807488287,0.2606050758983386,
 0.11956536537959506,0.1917131967477942,0.040769732049350194,
 0.5803131443005404,0.20069288242096284,0.6816335846879875,
 0.026275948715553388,0.1421769732062087,0.3927855298865751,
 0.14349286331578304,0.39385119584502215,0.32412638441750885,
 0.6869932746288689,0.6138707352738949,0.047433799710440705,
 0.7080399650892452,0.47016115627462174,0.037752529415947755,
 0.5165569788117844,0.750266190329439,0.2675282227979321,0.5057779085463666,
 0.7757803149445677,0.942966240025323,0.26519205304684224,0.8095580622154012,
 0.42695135560058706,0.44092190433591183,0.4368011286810243,
 0.4041576687300159,0.07938986495502076,0.34529392228740585,
 0.17716772701044625,0.3878558731120978,0.2054378738536239,0.7391750421433297,
 0.7275754236558801,0.771387062105714,0.03903746420790499,0.8172582503856571,
 0.5193820048851402,0.9862805643907623,0.3737465065174812,0.880049445466802,
 0.9970728610825039,0.04642306873858182,0.5925356985590462,0.5154693927121033,
 0.9207542896679027,0.4187931440865369,0.1458848718233159,0.42673313698854143,
 0.1909578849567538,0.3267753924845669,0.7166016077475307,0.7735700329353306,
 0.6165010512332549,0.645764386847171,0.24324258643207752,0.38138158742833583,
 0.7542896996488905,0.7293402807105052,0.6232179321059443,0.1470356930140515,
 0.075499573401545,0.28798267907237796,0.014748198925677136,
 0.40484037515805904,0.9713849612431712,0.6611204390161757,
 0.11504772165673582,0.4136137816750286,0.3230572341515572,0.8852145308039435,
 0.3715139049298083,0.5084923607287682,0.31093672765744385,0.7362432336542064,
 0.3542128213968959,0.7958676951751678,0.6358606546252082,0.4633209467910274,
 0.17946668868862214,0.4729275127837136,0.9434014795171709,0.62665271833344,
 0.7031631536059081,0.5647194315445324,0.351271636843034,0.9811880476570048,
 0.14975537199134892,0.6082581829188861,0.6165609000377589,0.8329598857512286,
 0.9454681815024651,0.6521726961864477,0.008052615587216372,
 0.2446428575530385,0.8373187193556542,0.4606336459094067,0.19983127300916548,
 0.9742322469578364,0.0678544935246237,0.4199365985137158,0.5887102039758316,
 0.12468345179192164,0.22625755736687592,0.38243104390916727,
 0.860771403655404,0.2890579266656178,0.954755511124038,0.3292491441378129,
 0.9765217159832794,0.12682965595893714,0.19727605636166368,
 0.49607669256481624,0.07269624918869244,0.5355405482390887,
 0.36079965796192204,0.1234876615077578,0.5518825467011592,
 0.17013017044519063,0.2974922015586988,0.8023328148453812,
 0.14910535466301778,0.4897561805054219,0.6271914017594149,0.350391700621532,
 0.32312765589969894,0.350276083802466,0.0018510476504706164,
 0.3795956672413545,0.034290234672485154,0.7719127632437052,
 0.6892474302882421,0.3059228788903777,0.17181967502764584,
 0.24313352206703132,0.28517039464204597,0.6864569879841456,
 0.3129013964316151,0.42467486251131603,0.07092152091372228,
 0.6839984231294352,0.4197529016910261,0.7942603796550144,0.1326515462856397,
 0.3215777764078158,0.9937033972640238,0.22765033602368567,0.8675600520152235,
 0.061655407519694894,0.643098658324285,0.14722280476704197,
 0.3041477309605105,0.5408054695614349,0.5754774818497963,0.4515495078155102,
 0.8418941390139407,0.10066589107307133,0.5283016692636056,0.5656016131001262,
 0.5806304102178578,0.6290050298835109,0.15636548793817073,0.5852955772773222,
 0.3129405381221314,0.7063391632607885,0.9643989449294499,0.10795012622180877,
 0.7944758477414431,0.6675283187078241,0.4298027695930987,0.8215098421317704,
 0.4733332561311846,0.059053286663969806,0.9257194473953407,0.979433029889962,
 0.8572370397175026,0.9425535984861849,0.519116067920375,0.9037281671562263,
 0.09781992954504148,0.3526963459937714,0.3153100930257926,
 0.09956944128977618,0.062207148049224004,0.8477357709104072,
 0.9425882247634855,0.6189429642769665,0.45968878833199067,0.12802885021693,
 0.11348889438116816,0.05472972254976194,0.17811162312974782,
 0.835100687963922,0.6135148419683174,0.1285501325243672,0.3289293622214825,
 0.030828430843520827,0.5713366492390578,0.6392989366606209,
 0.4949405483629232,0.9304701420957959,0.3644328622233193,0.10553008670893849,
 0.22729512387000084,0.15635548501368968,0.9494523939723084,
 0.05154586809911421,0.5841887455536187,0.6962613018662648,0.6857929449418592,
 0.629125015013854,0.7610868921780477,0.8135482057013905,0.6404146880381703,
 0.3715453108328677,0.13046964936688132,0.8240584000342767,0.8509374513012198,
 0.06274016011024763,0.22400437122637173,0.4908291780381765,
 0.40248444717046594,0.7845174136120084,0.7523217820411807,0.8146585591657353,
 0.25193040430557734,0.6194487648913086,0.3743144818732249,0.1392433252813816,
 0.573550518868207,0.22005775453663612,0.33805357459219165,0.7788856486973091,
 0.714243929265522,0.5172266896140729,0.4490314464498891,0.07266838847738033,
 0.26326067683278587,0.6762946499167155,0.05057652364289944,
 0.6647345304722101,0.7091376602176603,0.30578761252206577,
 0.008940051559507056,0.8949620393929031,0.5246647896945682,
 0.8457692363558436,0.6096454735386193,0.18822180596403104,
 0.21332329564186803,0.2023069665147037,0.893385393091265,0.6659274244979366,
 0.6430889128984631,0.9995418235634379,0.46663890569987043,0.2845601654451759,
 0.5195771907687391,0.14259905339093493,0.32389337003783236,
 0.8427245694898997,0.2291052471341931,0.6856675327964501,0.7097970709520837,
 0.20493789848284327,0.9326255581202382,0.7281653707568768,0.5071406417255068,
 0.5942345219256666,0.9930221786935265,0.30614381116819445,0.8663493713249636,
 0.4728131539692252,0.8843893989167232,0.9731738756257099,0.9586664256502021,
 0.9859975128627183,0.36374553497286555,0.9265186764372962,
 0.0031810969838819148,0.7319686433334661,0.06983090527497393,
 0.11884682861926366,0.2157344538229059,0.9049921739577551,0.1151719727265903,
 0.7698273071596079,0.8402613511744556,0.05788705953836604,
 0.40614655691909796,0.7314134249580876,0.8604033423726102,0.6423396216020242,
 0.4112856180467682,0.6988125145041966,0.33069840415212004,0.7154777744832704,
 0.5580654808865255,0.6053891582084271,0.7956279865865477,0.16478523906147624,
 0.5698824549597932,0.9756661696658654,0.5802013312388715,0.34747621761673364,
 0.07069677273500174,0.24258848244731968,0.8126900896336093,
 0.5060887738659469,0.9513444401030203,0.10368534307760702,0.9037688777330858,
 0.6746528011274846,0.3327375591026914,0.36853956038191926,0.9544565053458197,
 0.9680221986482769,0.9096365524986018,0.5176582602959205,0.5971515465655829,
 0.8508254811240735,0.5078334187122887,0.9353152226720634,0.22751648356946153,
 0.16323782788598407,0.7186983736785166,0.3992216901297402,
 0.11051105771286451,0.22416526526664537,0.21384071800448878,
 0.4212954541969989,0.391712303577213,0.8813361170019826,0.47830779285953007,
 0.2063817542470593,0.2465136620056919,0.14441255515632911,0.9839182039784813,
 0.3011735359172347,0.6923023037921653,0.18971133674264773,
 0.12239537308831705,0.23676810008018612,0.2684946086209603,
 0.09765247737843574,0.6435389254906709,0.885398373367256,0.9293613602432375,
 0.19891891369720993,0.07851332563144342,0.6527180241578469,
 0.4154575174274814,0.2780995897869828,0.12260201633514278,0.8788799999665018,
 0.20240355852513836,0.9552049381411873,0.42930762833867797,
 0.00836606432132625,0.23561262311895792,0.4896984641299875,
 0.6237193564467828,0.6720157869836096,0.255172585021323,0.6532816918514167,
 0.8987594801040656,0.8328684386855378,0.699957865123683,0.5630674390360939,
 0.43860125830714725,0.5045866539115493,0.4917436633060285,0.515259956506221,
 0.43123361094860935,0.538119465660531,0.507783171493968,0.2416826477475177,
 0.6266127473772253,0.8245459751484384,0.29766515512804537,
 0.19134204757422424,0.8543426612094858,0.5501519616557473,0.5014622005334961,
 0.9648458658435645,0.8453997932402109,0.10160099809719036,
 0.08068699249672606,0.7295044353574947,0.3862062334161602,0.8623770928264726,
 0.7225226136897481,0.9861900586916396,0.8460812958108919,0.4657285195161004,
 0.34032111669330845,0.47305428221659396,0.9950565984909865,
 0.5103899437391977,0.5795608473364819,0.07113097460319084,
 0.08820907769650654,0.2868737373955391,0.09839379595928643,
 0.6608923268599953,0.8701310156112703,0.029390115525602556,
 0.13755734283974919,0.7122643701751137,0.3209860473817169,0.2311655039087871,
 0.8803938175431391,0.6725712800956938,0.182264145956889,0.18107599819984355,
 0.997148770417431,0.24780827546036444,0.047750694652808345,
 0.11706157360871261,0.9094813101850737,0.0987109249010194,0.9019098130135673,
 0.8408783880512609,0.41547005581652285,0.40223076710543615,
 0.06453382496000959,0.8362393203227728,0.06875158747064902,
 0.8998699747844924,0.47617862312924997,0.18151681085970428,
 0.9231074465558639,0.7613176992727253,0.18239001396929178,
 0.23227636615213276,0.7808593388131448,0.9425207619956597,0.8372341978160851,
 0.2851793573974517,0.7240921324738074,0.007458142849586347,
 0.25591148083165793,0.7413276671344989,0.6969782552068733,0.3992858798288391,
 0.831628338177267,0.18980736681401922,0.28465088040454545,0.9172521563319629,
 0.6918287776649334,0.2824995820826477,0.14437074423389606,0.4921228889321485,
 0.9491237396938619,0.8560299862417481,0.8932146268174778,0.8645779634750669,
 0.9580630848601581,0.8279600452185054,0.08942946867376722,
 0.34051867117011325,0.2960361861310924,0.3018749286690714,0.5092612636688473,
 0.9418606530824325,0.3780028336479845,0.12410322185115263,0.5743886351186813,
 0.1891377662439484,0.5791903356285271,0.9730315637649004,0.8873821386107802,
 0.7948498385116805,0.49105381709291884,0.06051360791627036,
 0.025669390146047744,0.205770462252435,0.7446566879381105,0.092045681199587,
 0.46453252050013827,0.8554054195757332,0.32177704811406127,
 0.07995280529295057,0.8156659112256333,0.8010930939058625,
 0.24882461722595772,0.3938424711637034,0.10991084292203546,
 0.41194003340440866,0.18338644938009685,0.6140491005483169,
 0.9335095680659938,0.8532598458311352,0.028109747284978948,
 0.6971903401826556,0.7305751257403348,0.3434567772839785,0.6674748646202517,
 0.024889677258311593,0.47099641106721146,0.34547519736459287,
 0.3702172891511126,0.5848635684566504,0.8902293608685133,0.1188951277489545,
 0.5171800195168766,0.03448186902513861,0.1908673688982665,0.8591664392393519,
 0.7136615682666446,0.3807538105305387,0.11860693416946011,
 0.06775798983827597,0.6188280865208537,0.8702965804858488,
 0.12292729611441433,0.08604909644066028,0.607538114386591,0.20981370570502,
 0.9643856130591959,0.37694069737786995,0.9469983497563249,
 0.05231721582883608,0.3827384607638069,0.5999216249846772,0.6332802759045039,
 0.672210150541026,0.1712866062442162,0.8323028270100447,0.2855447830665363,
 0.9045624509768816,0.097689536505843,0.10516120544923546,0.3946737771005271,
 0.5181447921143219,0.5336407037327764,0.9535608676600972,0.6867037052424821,
 0.30552049046926677,0.4386452761984936,0.7228107112193162,
 0.016104858239191833,0.21438482049056962,0.20459123919714017,
 0.49910233211488986,0.6171208698978324,0.6708389590660767,0.7200980776535337,
 0.012122975719164986,0.8706879184361275,0.9009393648875831,
 0.45347462009428796,0.2931672800318863,0.859716077684044,0.49630400885358006,
 0.5894689252712101,0.28095248435962983,0.6427161739231119,0.4005062622257981,
 0.3080761284440364,0.4206769323647307,0.1912683798991066,0.7484345959864005,
 0.5148750455211175,0.26908999644168186,0.8059432429782811,0.1669341021380344,
 0.9164761594315751,0.379933242808712,0.8577388992295869,0.8403097555965,
 0.4265868359340881,0.08247143018967273,0.09633925582711389,
 0.8722644855384352,0.37485605299705815,0.08379968323223785,
 0.7239704085025624,0.8411249398178314,0.6798835481089264,0.822908002033674,
 0.1715034182838423,0.4381670968169469,0.3808089280429109,0.43317005806593234,
 0.8244439256627017,0.4976077821306545,0.19643108007688403,0.4241810283973755,
 0.3880053074147205,0.5515335455820715,0.1570050648426431,0.7043727607752944,
 0.5240064499467556,0.41362746598801026,0.19856239915236873,
 0.7126822930752171,0.07657307100190502,0.25478748722778943,
 0.5884793950087482,0.2955684735621198,0.2035426676794454,0.4488696597599763,
 0.11492389142635584,0.22252967515275246,0.6670328594931577,
 0.45742596751312115,0.28386593496980583,0.43405110018982584,
 0.4798121201181892,0.03253340015396744,0.9529603161729853,
 0.27325940814140026,0.9192887624983663,0.5326669342473084,0.6004166272752318,
 0.4425805264778535,0.6759742853965414,0.3517163647423577,0.33190035680798013,
 0.2996620268517777,0.5462298502938965,0.9360585914623558,0.9767526949612944,
 0.07635272123608206,0.6701123550718783,0.3554285062902369,0.9334891285422237,
 0.5853896750061071,0.878550892328727,0.009840671172684612,
 0.09392210468771112,0.3657459668651344,0.003300618193714522,
 0.11185213075936551,0.7868706777824612,0.015655305299744127,
 0.8686162784500061,0.16610108621028652,0.3024123276768521,0.6338503074138664,
 0.47347246187167236,0.6513155733230983,0.2884085755200436,
 0.09882246461812039,0.07442236009484848,0.412007554108351,
 0.19612337515051248,0.5139569255575039,0.2762804283260991,0.4799615932108292,
 0.058113082128122606,0.7420875915926763,0.1530615728306851,0.734043696786023,
 0.41850390773731005,0.32436820540103495,0.6511252108500345,
 0.4680705199486713,0.4864460813552871,0.46701410135555177,0.6740326004867822,
 0.6463730627362761,0.8217687069058989,0.8258968699003658,0.9457156240641649,
 0.3945611745041494,0.3987182999255724,0.3208023617957536,0.17354332318577892,
 0.13655948477423552,0.004472136764232082,0.32554251333293716,
 0.0032605209635646615,0.6004281277368811,0.23351191091200574,
 0.18392214326647782,0.5876270775370847,0.23144123940829808,
 0.9940988486718234,0.4803846308394366,0.9905032631155718,0.7029494807761006,
 0.07353482537927614,0.135508228208437,0.553949298706839,0.6102037809497411,
 0.8207343443803504,0.3237649050872907,0.6369096368522653,0.21702187359281178,
 0.10296496670061162,0.31805985508518586,0.9030979265704825,
 0.2811457208062872,0.014773619781794922,0.6917982001942725,
 0.5363918852964583,0.8536123662595679,0.0307679139681164,0.7966509888673464,
 0.3754407538129474,0.8950791903482198,0.7566497258591011,0.4164399126215035,
 0.6081446485752127,0.3811382617453346,0.5646342978567506,0.2238898562666053,
 0.1008460853329688,0.8050535586356228,0.7656726803921948,0.04045160638750844,
 0.49068721481999733,0.6014803019785319,0.19652708368568594,
 0.46392724884936953,0.808648179891044,0.4318981123811285,0.12990243129896084,
 0.2400634866189475,0.3770592457307024,0.6631380167355851,0.5253282148054563,
 0.7671587957753835,0.5963762788785567,0.9049545773225908,0.3344908477248867,
 0.9323297886566646,0.6295144893757693,0.8245512747817383,0.7308253221864414,
 0.489279152814294,0.3249849586228073,0.017499657242433186,0.5093020071872639,
 0.45181540781388474,0.5207849078126551,0.7039309420882934,0.6287810697774707,
 0.072398703731815,0.709979203204197,0.3743417369292056,0.5975597084990601,
 0.447890518466164,0.7520512181986395,0.20655478392790227,0.14118408513269964,
 0.8731646919710514,0.11123720529050207,0.7343650934705935,0.6096172076935278,
 0.3068232784985305,0.16224076805094212,0.7813745220208754,0.8477784442622356,
 0.7730710330857641,0.44863768928353487,0.6029397243000219,0.5815599825637869,
 0.0478148471207972,0.0660888173073707,0.4018935021332578,0.2656088767026821,
 0.16850725067236882,0.15346916274371336,0.5011845219248547,
 0.5795702460340967,0.05552263650536082,0.4708433134714787,0.9548369232367758,
 0.19009456928706825,0.6997858600064571,0.5575279366884488,0.116827733626919,
 0.8196040259375652,0.790268745630092,0.2506061229693374,0.3702674483258793,
 0.6891730368359891,0.49554681164089465,0.3411432276113213,0.745706783065335,
 0.7384163495108623,0.9878437684551024,0.6417526982188744,0.06142517712860818,
 0.36678328497165114,0.4754629280670568,0.169944404516991,0.7033843226505745,
 0.6441897939460516,0.04630439866675917,0.9179753049152602,0.1286007960552772,
 0.62613042619774,0.22943167234004924,0.8053205355386819,0.48051394209533405,
 0.7129550104498981,0.7787315322991184,0.050122983518833375,
 0.4196712222826243,0.06359598459673244,0.22062472618970097,
 0.9965977684515543,0.3082544531641386,0.10968774656302305,0.920235516184013,
 0.9054037283307115,0.5437246180362055,0.4523042266955719,0.7840801103299533,
 0.8330954717590958,0.26767070025428563,0.06495711516810032,
 0.13857343918582787,0.42649173347436276,0.44550384432639933,
 0.2509199441552452,0.9807943304405886,0.4529826906842036,0.9683824365257756,
 0.1819335368860131,0.8343395009882582,0.36988086532875,0.45906195363837843,
 0.5295819297079292,0.3929286949456017,0.3974541020313125,0.2151620908298254,
 0.21696778752604728,0.8749611256309647,0.41722757142001954,
 0.5204203424260508,0.885154600825198,0.2714336550962888,0.023155511989784494,
 0.3739364597274,0.5527841176087246,0.5033371818352858,0.7057168249596905,
 0.6278081466325695,0.4656839607705382,0.8525070400257878,0.6279406447357644,
 0.08793667720115939,0.7268380899575834,0.18033767052302396,
 0.8372950469554583,0.5517501364150245,0.779595215373067,0.4002307189464491,
 0.9472576995411288,0.37407149856544464,0.9372821763370277,
 0.030101485903920144,0.9995191801392441,0.05882336280446987,
 0.12369977028479184,0.7970732727515457,0.1626690066401656,0.4811059964892497,
 0.5592351562829626,0.4737500460399433,0.9613043270961873,0.5237693527704805,
 0.18244496782270436,0.6141904567586837,0.6777031612907649,0.8488412536351878,
 0.004667697178826424,0.4428411158575627,0.6424721685431415,
 0.7263108836366995,0.6967074471560203,0.27484106886304693,0.123991429774708,
 0.0540928954436406,0.8337485177297276,0.7064859213861516,0.6082928323311172,
 0.678355437881986,0.3282974461131096,0.10060636000762724,0.9593261701251854,
 0.449980521725448,0.6177954502477266,0.11115268213198715,0.922221670369357,
 0.352210286688325,0.974578982744993,0.8554663699871604,0.856275202546915,
 0.7907923175824166,0.5556582694806389,0.28816193959928826,
 0.06877844670447941,0.27012026313638116,0.06521653090915036,
 0.5623934549305422,0.8658042812545585,0.2752176770348862,0.12388508815416943,
 0.3495806088639011,0.18865884923746323,0.9270466797042507,
 0.14038338432687925,0.35445209613380513,0.5113924844238403,0.501762783076941,
 0.4738131546053632,0.48512725517009403,0.4908441126971731,0.7324552068648016,
 0.6595945610050195,0.9313260649344767,0.8184104514904139,0.919546526599975,
 0.5822711262630119,0.27781129809524363,0.5267707137899355,0.7965868946108525,
 0.9485847844697572,0.7444482702151632,0.8729870894847447,0.8025246105061876,
 0.26483631542204134,0.9609904038261803,0.6027287974007209,0.9610053723927147,
 0.3597648961608061,0.006399789646373111,0.03602893213766223,
 0.8444936394560009,0.31575864926153097,0.11093936456383424,
 0.8802064720295821,0.2910981761244793,0.1824142945399818,0.28518339616712585,
 0.7769402248780487,0.5465010297204197,0.44727844689297147,0.811867234211508,
 0.40059998919713613,0.17570819443598007,0.6798057951996235,
 0.9970597514179448,0.10572322912540244,0.7512988423854018,0.7043810400537592,
 0.4655273310210941,0.17739490206804087,0.3331262159923475,
 0.14882924024741195,0.41896916084560876,0.33961025061636607,
 0.8712076952523023,0.12788015830582455,0.969371570133593,0.8232130126922219,
 0.7409112903754992,0.7777685762222146,0.47802404418027067,
 0.09130806891580312,0.742302416082095,0.6804878760619316,0.6065982894697591,
 0.6787090013793292,0.2729790529521924,0.15755296917125705,0.332822140245321,
 0.9231642669162239,0.015957846812397203,0.6959333748910488,
 0.6774404468718889,0.37143774371954363,0.2547146341273031,
 0.14550896954527193,0.15648604698678692,0.7450447851722732,
 0.26634353007627265,0.08705707156008158,0.38321042616232415,
 0.24143860125986139,0.28383513682186334,0.6836608435767689,
 0.7674354879871091,0.04634421368008135,0.7086708040316989,
 0.03464791346823737,0.8581741901359321,0.21077273144236697,
 0.1415601177936674,0.4512047783794789,0.22725267410452132,
 0.31735731291130675,0.926721784358216,0.25498747115059284,
 0.38472992368689396,0.6795827289345148,0.7190239448891789,0.9787036096711184,
 0.012821183868073072,0.6057896064938046,0.7717167163997052,
 0.0644813201672142,0.36922380117238807,0.30149056642997185,
 0.38220169957657035,0.5740799985373386,0.8825127370436693,0.6588924269427834,
 0.05162432461563993,0.949462507329626,0.23329765431981175,0.577350524062344,
 0.09239307023936638,0.6885798333598497,0.37758027785570347,
 0.6724143773165128,0.22970080566747786,0.23105818222584273,
 0.3695613663628725,0.03059114548522368,0.0176215459518565,
 0.26207691633279295,0.6519463766822062,0.6370658931918393,
 0.03159570609821283,0.026823002394395723,0.22123538301977874,
 0.3621409353186975,0.25243975476245706,0.3261874688226113,0.8785702990612646,
 0.015132072775137506,0.7512363959236839,0.8878547749673995,
 0.23116262479734884,0.5124341079254509,0.46362429541463945,
 0.0737989008949026,0.47315794256049204,0.09933545021822032,
 0.31039161742089993,0.48887865906757155,0.3756348492588768,
 0.8506357425318218,0.48701274196679645,0.16057502251527378,
 0.033335562278053676,0.7257597291732968,0.12004072853978687,
 0.7433532516656257,0.0416740107872815,0.0015832944090857781,
 0.4763903764920262,0.608134853254385,0.39136308602864944,0.15268219756394286,
 0.5035724669974668,0.6542532311242402,0.6781276280203989,0.9812806132050812,
 0.8366479465284933,0.1869914921543503,0.9727690815371972,0.5170484774440192,
 0.5437306521023729,0.7602441603408957,0.528943700564885,0.1971669045347546,
 0.39378700185684035,0.29731120581693304,0.4052260179538417,
 0.8004349698490257,0.5610475116034714,0.19116904800509327,
 0.24821536821044576,0.31462684195931157,0.3415143499657677,
 0.9800655613598211,0.7876982418195835,0.5315552174671918,0.780892159230777,
 0.7279923645110654,0.9304078199744181,0.22312662374678482,0.5802796454497536,
 0.5485867575933624,0.6602440421748308,0.011673342980962897,
 0.8212194183454417,0.33259369419198226,0.95157817679258,0.9384143662702017,
 0.6271477212261841,0.9555564799095035,0.5603179548598868,0.2755574213102048,
 0.9841340335050888,0.5509605525661406,0.20027320335930254,
 0.06038305100302532,0.5810188704592233,0.7227953332838465,
 0.26226205128342384,0.7282504277705539,0.7926216313595265,
 0.15594015384058657,0.4278117032802019,0.5419734291260431,0.8362032345652723,
 0.004844348292935896,0.4935777846281416,0.02852099002885289,0.94028112566867,
 0.8007278178825321,0.0241110273959958,0.4885732153236878,0.3611507811537108,
 0.9404829508258991,0.9774692259430555,0.5127602696274036,0.9511453103174381,
 0.3336232122600066,0.8406972031737867,0.5665240789063467,0.05919159298971355,
 0.7664555440706216,0.18866509311987179,0.7761503002589359,0.97367914226581,
 0.8962622961040897,0.6944775744486341,0.415650217639539,0.6449063385119241,
 0.22695826487176474,0.7944535454749619,0.6943281428350851,0.1905693307132953,
 0.4971194608250069,0.12462273793267753,0.5358323003528174,0.9106893777378967,
 0.08026415710237422,0.6017005487112049,0.9723554336973824,0.6818876485115836,
 0.16263220796346267,0.5497442324178591,0.27466599316975393,
 0.8714373109017866,0.5159046986968644,0.6581054245419936,
 0.015204235212348927,0.5484686701901672,0.9205792302433108,
 0.00355830344010144,0.49628559149316,0.5587271777033662,0.22087102501896705,
 0.5591673356174938,0.3947816310216373,0.2824397293027946,0.7554882261134495,
 0.27218469261494893,0.718495548653116,0.12179186868781944,0.5856736185276851,
 0.20854741538721955,0.23951784854741687,0.08671778128880536,
 0.34514878613165856,0.8963507939473624,0.9273365547872944,0.2815930832158382,
 0.6429774458467912,0.08177762850794268,0.2708398745445727,0.9357025536766688,
 0.05051724937058266,0.8405090480498502,0.2612175497388355,0.6809245181631094,
 0.5782966443056832,0.9814202191772974,0.8151434630055107,0.6612752137103609,
 0.2412770516342786,0.3389512450654637,0.8190745904355023,0.7561410801528476,
 0.07367373747203898,0.5714236993794759,0.8782633286951448,0.5721505729096792,
 0.6049570501582878,0.5653409971637726,0.18903659354635272,0.5034916558147482,
 0.15247878862632203,0.057201594086356944,0.1759697838138382,
 0.6121012189914924,0.8770224851625575,0.6351559538963809,0.13143458321633172,
 0.9769156606162123,0.5868977769166465,0.15892225665720283,
 0.39244508359958374,0.2888151065845692,0.11799861758984975,
 0.1283625056777361,0.1254100231300289,0.22588219438052493,0.1605641852882661,
 0.9045801645106601,0.2879929022348209,0.028773095503465185,
 0.30213160466638367,0.2594767888166425,0.7738170061121047,0.5320832009145713,
 0.7330575232169916,0.4882226262062286,0.3966765960059959,0.09123819612874962,
 0.9859492970262448,0.667253133271372,0.6649251898189668,0.7370112968157516,
 0.24151915492196852,0.8770954763605946,0.6425434224004648,0.4822198754960201,
 0.03644113561217588,0.6505257775170303,0.8953120204199454,0.8888429448717474,
 0.1318247881864184,0.44617183499283253,0.4671255365212851,0.7934852986345805,
 0.1264400320792044,0.7287877802895175,0.2615621026498237,0.1486241311417913,
 0.3964712287737947,0.25503803314425044,0.8605527630068203,0.1118458896797232,
 0.10989474779632258,0.7472756280893744,0.4550609395960137,
 0.09097680207145198,0.5988909763982174,0.9334963073067363,0.6257883201307723,
 0.5329730412903868,0.32077688411721883,0.49702740356692987,
 0.2687439666838465,0.09923658972649374,0.6632680732681632,0.5714519985903833,
 0.2910533042651353,0.7620190137975824,0.17841084315312095,
 0.048780854134491625,0.5115696186747494,0.8374347383913563,0.804433238883656,
 0.20219894770564462,0.8507493449486608,0.7501256395750883,0.4876800615289072,
 0.2734329716632675,0.1595946015123223,0.6767731289737251,0.2326766364865681,
 0.24921889017332555,0.09619891375132317,0.9221585905358441,
 0.7264328685796454,0.8163122817099279,0.05338894373447711,0.3941223395816098,
 0.32561682940309156,0.9007834274086258,0.6406680456411739,
 0.011357983008558947,0.10356466262749564,0.7437288399079627,
 0.9530742996894124,0.18188522198881596,0.34359428034055184,
 0.8677787875699289,0.5073503467432269,0.9204301901781827,0.9185586029646773,
 0.9639889289852673,0.3618100845893375,0.5697256555955652,0.02503353691821475,
 0.07744949961526792,0.511768836550461,0.4512560978222191,0.47826605661382726,
 0.868009580015829,0.31593347824858675,0.20224204387447142,0.3293570451457295,
 0.1262247114171523,0.7096248247443678,0.10094534047343473,0.8447901146299935,
 0.2692262993030472,0.0242713486718511,0.36540526431627374,0.9980852644589802,
 0.09409562390616644,0.9252530576560686,0.6629306490575054,0.7257011768364855,
 0.131255501877782,0.0344192526785444,0.9354125845751785,0.8190826897142602,
 0.22461068739462065,0.3336688347169682,0.8092807221477065,0.4133874848190674,
 0.7834086919338713,0.2620153833808978,0.43446452866985397,0.9554229455413307,
 0.7912396528593241,0.7540430808937834,0.4128680432037387,0.8285655108516541,
 0.08168894836906093,0.49135245802483607,0.9761281222373015,
 0.5367568518800385,0.5963829038023124,0.06495449347116555,0.6958860662740546,
 0.8008247300521257,0.3746955016522092,0.8665757551041159,0.3698900492952708,
 0.018208788950611732,0.8572675050023163,0.4925100182395463,
 0.5581371410869067,0.7755323504248296,0.029681513817491623,
 0.22338828296965063,0.4703562718198896,0.23130589545206814,
 0.37299348145076017,0.15752495944433653,0.1509218096646321,
 0.6420544132999764,0.7760574464446952,0.954494917135031,0.5270680029863326,
 0.32014632271064003,0.19820730756121863,0.005525784454111671,
 0.9703703149436086,0.7499355085885528,0.5218464051127629,0.777214352507502,
 0.4833883790948216,0.10164769240864391,0.6537587988900773,
 0.19967221625952658,0.6466892071140967,0.39733465295161685,
 0.9634780903241984,0.6199887284196325,0.9807217684734141,0.08418614056109686,
 0.7813350497141882,0.11787724471787919,0.8242854451044133,0.9131077543204137,
 0.3132110477004363,0.6390023302733452,0.5872227211007042,0.34316017798739995,
 0.6075852006101725,0.4987736751396924,0.6485205889469808,0.4921190269250506,
 0.5596940865723007,0.8686615767156125,0.44097343251727583,0.3381848464540165,
 0.6233822876034012,0.15083619132135928,0.6670938424476387,
 0.048402512736931746,0.25350311756872257,0.4057804825007325,
 0.03528161651717687,0.6678908404592703,0.6634722997094038,
 0.12225318216412862,0.051287907837262,0.8791187831353626,0.3349076107846638,
 0.05334268942375453,0.0744585514196493,0.47540170609954124,
 0.6411835520034744,0.009705549567625704,0.38781668928488844,
 0.5551476308440062,0.8589762953726461,0.3704074140973581,0.23657717183666893,
 0.7562662824961115,0.333587754893954,0.9405554520554195,0.9519418866221466,
 0.332840582922342,0.8788835642456483,0.5921103396541176,0.33910396889504,
 0.039689134804315795,0.7051101222152186,0.021618030188344184,
 0.5108198730025089,0.14623700508966886,0.8544593799850341,0.9408489182623669,
 0.8708048151094244,0.9615933878627188,0.7282576987357423,0.4167977250421321,
 0.10881822106274786,0.15220344077883485,0.19105985302917772,
 0.9337646946045156,0.6272862045388524,0.5510976302284271,0.03904022896228232,
 0.8844871834427483,0.4968416860995477,0.6624550300342816,0.6364071483365039,
 0.5854591566075764,0.588379136629873,0.07646115689343003,0.6850010635035616,
 0.9980580873893669,0.36791051224206617,0.08063657459517592,
 0.5147614573617808,0.8756245564891535,0.647224508127071,0.6832300471784792,
 0.40010559645851795,0.8181369454443352,0.22469493729784118,0.483623325414535,
 0.47999125943912113,0.8707432752591823,0.29305638553383473,
 0.6960554951353015,0.4142820247935729,0.5751187176170579,0.9459129739952001,
 0.2685380201596482,0.19190849958054812,0.1957243886539819,
 0.12183368338284695,0.8531845024455907,0.353346475147176,0.34980275292506735,
 0.345049530025069,0.3314347067272738,0.6138214996005654,0.7254258602142538,
 0.799576694565759,0.030934457941620863,0.9543040915441862,0.2940699003908923,
 0.5362014879480764,0.49741142210113487,0.3166911824224625,
 0.28243202327496875,0.8634117470736452,0.2583384558154832,
 0.13019619559943163,0.09079950273954696,0.5036863795624261,
 0.9889479209027193,0.8736140865475754,0.6521231386881985,0.6333400102288946,
 0.7957366101965994,0.6840522622984615,0.14838880892607098,0.8171577437186245,
 0.04307556042132843,0.6107968713041476,0.5112418431360201,
 0.40841397881551933,0.09156486583292467,0.6280778681041796,
 0.7624103417171798,0.8544809336357188,0.7289025453799353,
 0.015780510816893356,0.715230133582189,0.8749268870678951,0.9212660824851424,
 0.06718639031731255,0.6963885042023394,0.17246166587433842,
 0.6144057400129725,0.6371467075707401,0.7130346943979305,0.01409422768101154,
 0.390694526854674,0.7260274428582508,0.5119840051820812,0.7238400572654683,
 0.026465313237202803,0.14789997503630736,0.5226525313757797,
 0.3946795707099243,0.7943938354301991,0.5997436089945386,0.7187226541631387,
 0.3693285902849093,0.3891663489268472,0.34242660914015677,0.7571116577706967,
 0.20743849175704066,0.0897866815054652,0.8628475311550431,0.3595162389750929,
 0.27520095995640115,0.4508227414830416,0.20421670344793863,
 0.9536213074186649,0.36499558120934306,0.2760042936525462,0.3829670810726531,
 0.15304371074976175,0.18080463152802773,0.10075946856754248,
 0.5228096043823974,0.37543155833668485,0.8925073012049362,0.948699165866939,
 0.6437245844872311,0.44473540047938165,0.5410418951211144,0.8411465886196163,
 0.9191783932884213,0.017758143992174125,0.7554169715850964,
 0.29584124002352286,0.9727599949337762,0.31706574799415177,
 0.49747665203394753,0.6416441551770646,0.4551978930470796,0.9443464148512843,
 0.3948415435433126,0.7222633486306119,0.965012691840098,0.0591980606602891,
 0.023450087418119714,0.31852975923927207,0.9401908938437534,
 0.4495751870028152,0.8488048524599177,0.4368757923963552,0.5006770423946945,
 0.25523012674025536,0.801053517910057,0.3698510192189767,0.5881342657879802,
 0.3912598397674074,0.9534280618974013,0.7959597889435784,0.38608419461329957,
 0.5849342570941659,0.7202939564588889,0.2899038085196318,0.8255636322628459,
 0.07233348556423547,0.20227598124369894,0.39932725346068576,
 0.4149448995925116,0.5327884917710431,0.3850561250120065,0.0979434938700714,
 0.13423463642090072,0.10405547038696872,0.5260087056970113,0.33190871506971,
 0.6663448051180554,0.8068834046578124,0.28689672645159847,0.6418384309854274,
 0.4825258902906995,0.4450068456626368,0.3617125484356175,0.2958820322910043,
 0.7588929298559703,0.340375762564043,0.012846061674609,0.8209585316126164,
 0.6751755726580971,0.6196046420760075,0.16476765293256856,0.9813121712533261,
 0.5171234768731748,0.7959127927911503,0.9613043802722697,0.949224649954217,
 0.2571678785020861,0.45088969033801074,0.8182238123542755,0.6516850288813634,
 0.35614353543674937,0.08745374965363228,0.4666230012149222,
 0.6654840739031092,0.05260070172176068,0.6927103868049053,0.9860212600076405,
 0.21041626512628375,0.08365899245980024,0.5197018702941926,
 0.04412120393791408,0.6852957821898503,0.24597137918191847,
 0.4848405629086452,0.6349757407236776,0.18160802678862986,0.411648112296485,
 0.03374754640346633,0.5154652216658134,0.020527139647120496,
 0.3155666955752875,0.2770565631018037,0.7018627266561579,0.40420547555793374,
 0.6888906825065413,0.14446734994445432,0.013735079808143924,
 0.6142609883831264,0.8728336055891028,0.3848068586879285,0.2625085778984959,
 0.23491260127658042,0.9445903473065136,0.674823401122479,0.3414903675777601,
 0.7954518345367015,0.6200121497204362,0.478509561890346,0.5004513470029677,
 0.18019409921066432,0.09725339958769508,0.9458022109670484,
 0.1112802882509929,0.9121260373849495,0.9240152478277708,0.38930468581197886,
 0.2278918481201193,0.9611691818450131,0.15878189030783152,0.8476210260919094,
 0.30912187879391095,0.9532973368313542,0.523149788830775,0.7182855133847099,
 0.5852244738854309,0.866182184423121,0.5747970008233741,0.11327014100078292,
 0.748560439656225,0.3618009358922173,0.8566931144226959,0.5049062060909444,
 0.7932468906883665,0.1671729846005373,0.26053623983218266,
 0.37071481688711916,0.31324326811096226,0.7954744641766828,
 0.15477770222560727,0.7487656988776824,0.3693379888896592,0.773858001495889,
 0.09004209954219222,0.040757376253366284,0.37521419493508024,
 0.7955695092433342,0.4616692871481426,0.905014212929437,0.25894437567616924,
 0.8138171972693509,0.988693730307989,0.07547380139059978,0.5862165843405642,
 0.979641599512707,0.1271749391377176,0.27215265714217407,0.6952371930875161,
 0.3016068604754799,0.9585315435436604,0.912940957045634,0.8162051014266054,
 0.03355171907711696,0.7492729683637829,0.9831588229869374,
 0.43789438127799873,0.7771440889606539,0.5070702465975907,0.690154843678758,
 0.7315777420595494,0.07163533897649088,0.0060755221738137966,
 0.18672803434919338,0.0628238404821162,0.7396997042826704,
 0.03130107965805118,0.8834002618071283,0.036073906197429784,
 0.6137336722639674,0.10688980634916434,0.20814466929009956,
 0.4439646778907935,0.3262672104469292,0.5264227671453419,0.8409396601730489,
 0.03257088760392357,0.11144882558041569,0.9606521686981444,
 0.8794291938957738,0.11729441823286946,0.187046349608054,0.8394598109835832,
 0.5959768586762653,0.643071692690059,0.0996572143227048,0.11644305211796846,
 0.6209802485250107,0.6673536319839969,0.9861612320055854,0.03574146732614092,
 0.00229912409960642,0.46833890147523427,0.1530328472988508,0.521864853869827,
 0.08253775694085042,0.6756928986239946,0.1684048034368948,0.5067070441762469,
 0.9426988382389697,0.6815219759840043,0.604650434442062,0.1768497662351438,
 0.9954808132285787,0.6697603146220469,0.04119864258490291,0.5103334539222797,
 0.6895680057761261,0.6472790088664179,0.5090968260108185,0.7381421096378592,
 0.7666163683318098,0.18380422443024602,0.4755236315769319,
 0.20668339500304522,0.6849876344060131,0.7391033299131838,
 0.16942909622476565,0.8932911835194943,0.39719887232186146,
 0.8159432447811661,0.09237485049764604,0.1670737193259182,
 0.41213787085466247,0.13530414209984898,0.24071522647714727,
 0.7407990183347865,0.05604331878165847,0.9864134679627935,
 0.07104763969930894,0.9936470419999893,0.5247530325255623,0.7139906613815863,
 0.5104900072033309,0.2964664905702393,0.7863087547105679,0.5813940790223127,
 0.0655337533965924,0.7717937812259605,0.1947738857307042,0.12477104670508965,
 0.3693109122309257,0.8216760095276759,0.912031413489654,0.4772621578630385,
 0.7676687820717296,0.4956078445355604,0.058988230008353515,
 0.23634943101161343,0.8701929719087013,0.14489869268332356,
 0.048984265972304986,0.5838560576318844,0.9730911564018666,
 0.7226897238713226,0.836119850716182,0.19622469867258474,0.5335072281516269,
 0.997499197835526,0.5742496048526429,0.6115570507078376,0.9893893989630891,
 0.5269688089782723,0.4626259990433763,0.841743373103414,0.8556466135348367,
 0.2052860954429374,0.3552210152233506,0.40669806578687595,0.798211111423988,
 0.9630003887494413,0.34860719892993086,0.9375893960564237,0.6938530965843334,
 0.8590424375566277,0.3482703641769509,0.6554183659240277,0.19551731619826707,
 0.6841395110736241,0.4404994051443546,0.9676722470305614,0.9383750262529595,
 0.7154928333100075,0.7763371667435103,0.6422157098737487,0.8080378575574345,
 0.8384958992002262,0.8598702988506811,0.508519492404072,0.27962195435879145,
 0.15639883872961946,0.04500028911997589,0.518621003275652,0.8536348065107039,
 0.705896115321652,0.5744615098951877,0.3142906263568517,0.1757959903485543,
 0.652097266011272,0.4652761801138261,0.27455721527185306,0.3389620913297371,
 0.33285915480897277,0.042646364135391135,0.743586495687249,
 0.22457230805281736,0.6446192666536352,0.9656780856532774,0.7113338824456572,
 0.5335018629252938,0.8133725396015113,0.22437357291170945,0.9502058876894518,
 0.991442531245579,0.8053374311783013,0.41911758275670496,0.5588180758297141,
 0.6450219343127196,0.4069525034869055,0.5562728014989634,0.4841850882931541,
 0.6635286054698708,0.8046941628580982,0.9895842883799164,
 0.002775559290932872,0.47586913489728344,0.7613811628545016,
 0.24148443895846727,0.2786726772851902,0.08224991589399033,0.970038283623622,
 0.44335300085416385,0.8294120925955881,0.12784015079264766,
 0.13135480353105322,0.1577914105786996,0.6008440947689871,
 0.37501500636175145,0.08222757983996187,0.32363589550223115,
 0.20086192522583435,0.6142977892266093,0.4743632635453954,0.8864582901219766,
 0.0924128771427909,0.7298931269528791,0.5417149208933798,0.2693549963273073,
 0.211304157380804,0.534305842112303,0.4925020077090596,0.40749766274660937,
 0.08163516588341657,0.4536660713333782,0.7561679855695579,
 0.14396193382506128,0.35508796830340694,0.6374969089190101,
 0.6136744019279693,0.8559109627007212,0.454181480921952,0.25509685443398744,
 0.42400725473964507,0.5962736802362808,0.5199872750039727,
 0.11818132028630035,0.7523340843485672,0.23809973313187605,0.708132962266562,
 0.3087646688579727,0.8802055152691722,0.764810521494498,0.08812694171237423,
 0.8019695027412238,0.9731487389680396,0.8374426889923683,0.02202060616410284,
 0.06773248161248213,0.674617749009347,0.972612014634667,0.9214256004597877,
 0.2816479429097879,0.1414780438802029,0.667073534052383,0.4512913778145804,
 0.038544756154986204,0.20483058508487462,0.73316512703656,0.8479624415754843,
 0.2266776548331244,0.3240589879101732,0.524007124945111,
 0.0031093382113290824,0.28014440315447053,0.7955975862169943,
 0.07748251137965412,0.6617727602127503,0.5307842761814001,0.4154138175561318,
 0.4396254622001924,0.10915565071096622,0.695745105247946,0.614716346867106,
 0.3614190497259141,0.6964914530206424,0.5418903081285572,0.6199300219303686,
 0.10814608791199676,0.964870562394206,0.33229051510033625,
 0.25760237410477604,0.8097291233750585,0.7414397487402093,0.5807722303439513,
 0.34314125899318404,0.3009207348377698,0.8380241293353226,0.4351384551471146,
 0.9023010884626362,0.6918194343279223,0.4079787995260623,0.6069360646577844,
 0.9193277444807939,0.26401661145175315,0.7017344129733114,0.9925793233295581,
 0.7833569972734522,0.003823353755136094,0.10918400236439774,
 0.8780747628173697,0.8982302706673881,0.9359261355859482,0.7754715397278705,
 0.6860414524308512,0.751676361760414,0.25697505163714074,0.38270340321968466,
 0.24939112434233124,0.3821241737136253,0.2689419817192078,0.1723654436560842,
 0.30688680891177245,0.6655310354911417,0.9565519789021767,
 0.30334628939839114,0.5496605860341244,0.6150838224157595,
 0.020318386214763007,0.09430780823201401,0.9989000518687489,
 0.5772362528288988,0.3460595774752683,0.4317905293111889,0.5394424596553137,
 0.006461222292530655,0.955286781169646,0.9991223232813032,0.749776212367775,
 0.6042976888076381,0.5556602005792933,0.6376389812261359,0.849505559291502,
 0.6236090055312775,0.13124831834375794,0.7534534602338866,0.5611722045618738,
 0.3113995953747676,0.5601187545835938,0.5570453533075719,0.7693523230161893,
 0.5734335647981423,0.53581480575929,0.11467735584025318,0.20008247413931513,
 0.4147169803973816,0.620652455517895,0.7516647459200423,0.7886918807469756,
 0.17116686400933068,0.5871521302984621,0.8598200229901499,
 0.46871493626534944,0.1187413163544051,0.35874504971389987,
 0.6999600718801025,0.49651933278285443,0.7357313978683888,0.9142009978234232,
 0.3580284643769813,0.20746392853417517,0.013909509427193156,
 0.1372979215137775,0.9259558817521891,0.29581824081639707,0.8549127947413934,
 0.3219007118893862,0.4729112522598803,0.04304303212017546,0.7022829944811904,
 0.13479439177877606,0.19962663461156283,0.9664123466574773,
 0.8855780907643609,0.04897875174936617,0.9109948283877537,0.9964030214065492,
 0.9091754696071068,0.48212592768962037,0.6815266027914425,0.4423273587604576,
 0.7972876106942225,0.5077876409969195,0.8410325490705631,0.35318264904646035,
 0.8219917801253912,0.6408272774603593,0.9345129425028298,0.6275552290486135,
 0.39000931079290546,0.14618487639967748,0.09022554696396101,
 0.7700628323095603,0.0888499011872721,0.4074212476238558,0.10643598827719114,
 0.35195143502502724,0.2410599299609536,0.9903040746426258,
 0.22497880186020813,0.1479250223957128,0.12512756863233288,
 0.4979503353448431,0.38030761992713524,0.07509656936374065,
 0.4623151722060489,0.6216580469504763,0.3986421120693584,0.881713435656985,
 0.9328683655203159,0.6674963253913984,0.11851571245470827,0.5125472066427752,
 0.7909543418715248,0.8147275560723836,0.2254742796089123,0.7219093725489909,
 0.26135834576355044,0.9165124728624875,0.49602035527127697,
 0.29899396754722785,0.015784162954615932,0.10011280734100159,
 0.05152876450261101,0.6762774442302953,0.10849788228932133,
 0.17011718702936263,0.6611414624918102,0.7234535779373735,
 0.38274973685103175,0.11294772262617059,0.3183934012753641,0.743189817770329,
 0.4687279790846133,0.28656262491200735,0.48676078393184097,0.391346411620183,
 0.29196007397455515,0.48242372297863834,0.12953912115439636,
 0.7743933921951992,0.6954977643753371,0.47868133043999483,0.9630719341510474,
 0.39973924527733806,0.03978323612739376,1.2788579620237606e-4,
 0.1371402120336821,0.9727708313053558,0.26141713685566237,0.4955239021847343,
 0.7725116890062711,0.5289846024013334,0.6438535345959553,0.5801749188134553,
 0.08046242059229824,0.3468161971648588,0.550475539344686,0.7402573995666806,
 0.9869650698677421,0.8234601755673195,0.38366882443721473,
 0.25121821921168963,0.7352822817154601,0.5270942159612948,0.7816112293301946,
 0.7424455019562067,0.4387457225433251,0.6858805357060852,0.05548978111992098,
 0.945018854390621,0.8567178960805811,0.7782631007513289,0.3293416734018807,
 0.789096946582555,0.9514621901706076,0.20069961085112376,0.7974129848464984,
 0.7835587530252025,0.3717226086893941,0.773809384295234,0.9155465990488458,
 0.16471832668441744,0.88929049289349,0.626438734530733,0.6953627502746891,
 0.4577504948357425,0.4565406299034136,0.6378734294192188,0.9360643842802628,
 0.15486456509473,0.9664583049102535,0.6163246716331684,0.8127653206179071,
 0.8172593090293415,0.514721895646365,0.9087318545747772,0.9187046760387307,
 0.28196581869368975,0.6670732698078682,0.9115673162976732,0.9328566941090654,
 0.6421945354663273,0.3321042176251884,0.7838962851891124,0.19518902033866403,
 0.4779202722148025,0.480299789245868,0.728654860451476,0.8157725633459062,
 0.008590883217532985,0.531903363832523,0.27753638515221835,
 0.7477310379408226,0.44702958582359487,0.20153515631813856,
 0.5864407062946189,0.24116413071172904,0.6200248707744979,0.6847129848195557,
 0.8298652995451637,0.01987509739858684,0.4688333109323731,
 0.03815809767791001,0.7391616094997159,0.6565026589208397,0.8740579295268862,
 0.8832160437654385,0.6808637054563352,0.9560623927609143,0.777285017218718,
 0.4125846788616546,0.38597570140728765,0.7089265057740888,0.8655949200614658,
 0.8779831634276869,0.5107869398262102,0.5945525595629886,0.4661248604533794,
 0.08787291506926642,0.5820553347284527,0.3496293214188648,0.2400774430666459,
 0.7069090014080907,0.03786309879818406,0.11994823339200256,
 0.6822704930331746,0.9863755511840009,0.8473971340579198,0.5816121853208962,
 0.2227949047868475,0.8428248194315097,0.4070700249359642,
 0.004248280914805891,0.3938067207443141,0.6252987137670822,
 0.8341489565229947,0.5393322997518294,0.85505782880669,0.8545139611028861,
 0.5349416593716165,0.14944085885376168,0.650422059422417,0.26188883175149136,
 0.8450355764055778,0.003867125603898347,0.309761121352675,0.7445938569694868,
 0.9118458929605807,0.8518623127903632,0.5479114640941992,0.6701849635570608,
 0.6139169219006918,0.9537731065635157,0.4432455498340899,0.12260621084457324,
 0.7564578194180501,0.209136571056875,0.8537001124627285,0.1273068631152543,
 0.7506594321688767,0.4292553914764591,0.9681724256449037,0.733019950006288,
 0.818118430500423,0.19621188796949884,0.8761927745207267,0.13710067016024952,
 0.46529996450101485,0.5438912413590632,0.8678708805921193,0.9391108803070706,
 0.1344980647916676,0.8394901296716152,0.15068883963206936,0.7114568801597874,
 0.8795145439869221,0.36284693937632484,0.9527097595380061,0.5823305503150535,
 0.06819479153801955,0.19920609253962362,0.19837523526359824,
 0.7737270919697302,0.10275785641150881,0.21529112249110038,
 0.16673242897494156,0.32277517033146996,0.4223751379115732,
 0.10202050643583316,0.1994123600341322,0.6722092963337264,
 0.44361208162298904,0.11552679877149863,0.6114540575292455,
 0.6737991755829202,0.9283685392929135,0.5544521125990813,0.735987892017091,
 0.3135909615699649,0.006796679732270494,0.9071510812961183,
 0.9294621392890672,0.7749162419497219,0.7022730631854266,0.25886975242269583,
 0.5961491528903062,0.9880873205351943,0.5540237581597675,0.25891851727888515,
 0.647451345024861,0.19054881464603135,0.481673635628159,0.7615549881485721,
 0.6443044336869007,0.7600469049094598,0.6270119019937551,0.08944833287088017,
 0.8748231848547979,0.14081321957743553,0.36865807918486826,
 0.9070195243049466,0.11263164838634876,0.2745627228488391,0.5309836667152582,
 0.030562867711494857,0.6751421986009272,0.1802773119195692,
 0.2772268018284756,0.8673263446218503,0.602544934657412,0.5036131154573518,
 0.4965769706783403,0.21666011163173948,0.7863787870615697,
 0.06624782291742126,0.8797082815532351,0.5973963982999693,0.5233588359522401,
 0.29743150433318133,0.7301735274482738,0.011868560586389632,
 0.0978856962739969,0.0722448873471605,0.5991633947416494,0.8100026053226882,
 0.39233799901665645,0.9784276482453149,0.9218989254018131,0.3155921243125336,
 0.8050944941823777,0.5189606808016134,0.6732962809612812,0.4555500365248296,
 0.03689552222110504,0.11741709727341243,0.5009142696520528,
 0.6842219790996131,0.8463129835904801,0.5218208975067564,0.13986693871525413,
 0.49937722182569,0.29444617625426417,0.838075822588005,0.3148767264540526,
 0.9697119718377906,0.9069282633624635,0.8925710994446371,0.9799274755471505,
 0.5473500256277047,0.6627022702469745,0.5867378100724787,
 3.093763515497061e-4,0.21297616330409697,0.2791353850106487,
 0.704472415205836,0.4574620684817208,0.05359331237300269,0.36331892983269265,
 0.6127262770184286,0.3909474054677389,0.11030230205143565,0.5126170436192306,
 0.4709639587462351,0.8576029515230427,0.37101446695432105,
 0.15767281647279896,0.740678363288492,0.5453830957633423,0.20041259656176136,
 0.3663135706754721,0.5043922634592585,0.785436834662074,0.13221444478160682,
 0.9234827243877609,0.6630493762189373,0.009073160254591262,
 0.6586314163008911,0.5818312271456922,0.3710482013336346,0.3295798860202708,
 0.4552585930281241,0.7020919378445687,0.16822045193738866,0.779363603166421,
 0.4708808081828486,0.10987956448048741,0.45776961110721814,
 0.8099959829143758,0.7883210918709125,0.3749138466882983,0.16244047047164223,
 0.5567439057664708,0.3795866161096877,0.27488908808380896,0.6834103409561378,
 0.4095235993295624,0.8833158203601964,0.37166339234590906,
 0.28311740350723735,0.5661242766732458,0.07179158755082915,
 0.01395895174552908,0.5711122381824586,0.3907166127902266,
 0.12007534889853899,0.014029587805404642,0.0547813422160911,
 0.21075380426884083,0.47766439957658535,0.6085125254778475,
 0.04084537923880727,0.9495719152471602,0.23292610487533116,
 0.9182302485858443,0.8775288329484133,0.9131170351141106,0.8466571224949859,
 0.2939742818566533,0.2534738352327829,0.644483115210919,0.7400724093237976,
 0.05538847176148631,0.3207426910084472,0.7108824057080536,0.627729963680612,
 0.735221503652651,0.982029574498849,0.8186816311043965,0.45477625932982146,
 0.9396269078457634,0.5199349850716106,0.41464379250819094,0.4830837688107513,
 0.3454574808424625,0.08643687411086365,0.06269067004522966,
 0.2289961514343184,0.3941569010710215,0.7041434748192856,0.16665997859381365,
 0.7008989386457567,0.5516578661701481,0.6677803637562578,0.8085298284072223,
 0.9609578839481331,0.6850330028240245,0.6602546848445523,0.21537290033332257,
 0.8763134554058744,0.5320242156384773,0.402591477519851,0.6314402701628219,
 0.2923711422413743,0.4057671051602214,0.608184309328668,0.7035987569381197,
 0.21765958255143536,0.09194285865610918,0.39261381106978444,
 0.27393087900365065,0.3646981118909929,0.3913965950657491,
 0.18571449472697765,0.8256446550835337,0.8350358152534535,0.4455429198080083,
 0.3919087762274759,0.6180652888049367,0.7679827862121703,0.732721239887586,
 0.7819693081449881,0.9065853238300035,0.5007485400048963,0.9279665127003004,
 0.3241269066778092,0.2032421934070645,0.5564438192655922,0.0279539469231872,
 0.8906128996762975,0.668784094241552,0.9846637727094005,0.13783131997955045,
 0.5844274955236141,0.31448630574126146,0.18529235922948373,
 0.24078766829672948,0.2913590498311506,0.18354144389015414,
 0.8264308428401015,0.9400025723125655,0.8126943122619809,0.37707788581777435,
 0.13859278760976024,0.8111693069712553,0.2930586896888694,0.8730017215086706,
 0.8717663548234788,0.2075802267585809,0.33217430945709026,
 0.40732101194021986,0.7841422361692374,0.866720265796965,0.5090305284159689,
 0.931207306877426,0.18204092113997494,0.44295579559741804,
 0.49521140201449865,0.5279689149425018,0.12393623445919866,
 0.9237606156261303,0.9915773452494808,0.4858742813323681,0.5947505691979186,
 0.9575759060618521,0.6636275851624615,0.7910542800876524,0.30512919218050893,
 0.17072623654465535,0.7484862320254778,0.41079634637842743,
 0.26802309627709864,0.20795140241979282,0.8607961329266994,
 0.3347992204161029,0.31867715769107174,0.3696901144255075,
 0.23284971763456475,0.996270113765562,0.14576953515622237,0.5334992893330078,
 0.5698152544001552,0.11406696243403891,0.7602372795075695,0.140887799882773,
 0.4730096857286612,0.6136751939245311,0.408478155617507,0.18035680295789147,
 0.7640885238136155,0.2386526448652757,0.923879892373503,0.7833178360581703,
 0.3226445703205759,0.07754770028836655,0.11640155278713937,
 0.9357321973842826,0.8588161319976256,0.7691604325709253,0.757334784061566,
 0.29698446777660314,0.25138176930337497,0.8482298886069841,
 0.31253717508240175,0.5615707187889507,0.10474412086739959,
 0.027265801396763822,0.04961403428476108,0.4197376067826034,
 0.27013739789191904,0.5613282865687026,0.4599393012318529,0.7994361106181676,
 0.6326255358143831,0.7416257978774077,0.5130226346744413,0.5983167702801074,
 0.5225977368876832,0.7567266831479226,0.24746162805952043,0.9663111004854307,
 0.9658409227994674,0.616123622728203,0.03725146668736601,0.33401706306477696,
 0.22372205978186732,0.8989683663886283,0.16439893928078297,
 0.06284217450807783,0.37143581045074814,0.19117553452860625,
 2.9639956791349675e-4,0.3992948246413698,0.1931448770741161,
 0.879191711724065,0.2122039840025144,0.804839843025705,0.06509364487253333,
 0.5078502902064727,0.8917790445184673,0.9069350284525347,0.01930958653060033,
 0.2447756418409972,0.6459290728989292,0.17270274471623281,
 0.46629126393398823,0.9907811592995494,0.4401186323335169,0.6806663102850432,
 0.8115422260223729,0.9272704777703029,0.15847123671079766,
 0.11803055690969688,0.3373493375181593,0.8430517380049479,0.2455046947539672,
 0.6785495623125911,0.624436910639496,0.9742391853352192,0.37615972955463706,
 0.6698834994783905,0.2596690002928371,0.5065771991995732,0.8361049961682729,
 0.34838708308038846,0.32631275059979015,0.791126996749003,
 0.12164856678574532,0.02528855409321129,0.23712070901338367,
 0.7336306749009915,0.4758232497311936,0.1136700580331933,0.3255675098588118,
 0.8116978981169608,0.6416582638086423,0.015281674672302348,
 0.5741250388088581,0.9664547441806235,0.017405935420079177,
 0.7172476921520093,0.039350498974694115,0.6155029436799215,0.810197819010946,
 0.7206612845243345,0.2075833168295984,0.23980972662971456,
 0.18452694490418287,0.5079828513674197,0.07385371642289351,
 0.07671531487639482,0.5899529654946566,0.8190995640107982,
 0.28705755474532435,0.5796597454422907,0.8862932326544302,0.4392559822899583,
 0.057705840240865114,0.944414205124337,0.5334421198657794,0.9632446810865016,
 0.4778753467700142,0.6245825028116694,0.7726159257935983,0.3474818808915945,
 0.24987601133794546,0.146392782660289,0.6746583645381516,0.51820440782678,
 0.5562017222081554,0.6917475222868478,0.3521078790084905,
 0.039665503618019904,0.19094866650128495,0.9697349397016628,
 0.06568622422854675,0.41761394065793844,0.27148909201668514,
 0.13853771992953834,0.052869372872655473,0.1564318708439859,
 0.29404064142796504,0.6621716711257594,0.8754725213371709,
 0.33519387894429287,0.9704830500169418,0.026267849616203542,
 0.13091726519562696,0.13712885533749186,0.18692501387369087,
 0.16112976059457929,0.49188998534001505,0.6280838281615136,
 0.7086201195991909,0.15128166449267177,0.9299166859243401,
 0.04413923894473459,0.16236523293269522,0.007057868037923076,
 0.18013627582936853,0.6939819119374803,0.6018603970467215,
 0.058114484398143484,0.823301848068837,0.6909240912759242,
 0.35465390160988775,0.8593570786355812,0.4024889555172412,
 0.21794705154109584,0.22540916725196825,0.2235919198503057,
 0.7626421210766926,0.529404584550965,0.5246930845047524,0.24938363005702158,
 0.37838996611727804,0.7099329242644417,0.632403735684389,0.9050212795640388,
 0.8222377196430062,0.501694961765628,0.8403837821590215,0.969177284980099,
 0.17941329063015765,0.14654153266063696,0.38360282820202984,
 0.3902951598269587,0.9943253405082505,0.8215873205141508,0.34183988173605073,
 0.9069887166703124,0.0673539383384405,0.47627959113341356,
 0.49393736095252017,0.8009553814911031,0.19049046383568802,
 0.9647547019887224,0.5965057838604557,0.9866341159399026,0.07679335279589894,
 0.8878321467962742,0.14559063969404862,0.9751141711061884,
 0.37860937584879706,0.20665838710283113,0.8276180563029656,
 0.7693874427454297,0.6871556909975518,0.04725895965254867,0.7481260087876612,
 0.4072258145920338,0.12878054773687897,0.9049500054078944,
 0.05860290193890827,0.6294913569718183,0.4000270643734066,0.8312297825138832,
 0.8850102970107377,0.33675947300130726,0.7386123796392019,0.5711825184502693,
 0.8647916636493261,0.17501885600350264,0.7874702876744221,0.6014773884836416,
 0.5711182383130671,0.01346042219031296,0.6081754600329136,0.7438220870593403,
 0.3049320159980493,0.22567537091260959,0.23060997855926557,
 0.5110700329511685,0.523748816781469,0.014384397343891187,0.5526729464654574,
 0.5562662490871352,0.12081954905384706,0.49802916038815215,
 0.6134918483639537,0.9602998622052213,0.01648004360995481,0.769030275787293,
 0.6541744863190319,0.2818065512780411,0.8387386035842822,0.16994147201144272,
 0.04072112197817046,0.22407861821683095,0.5268236466488897,
 0.18108470404566934,0.9174483574475716,0.3492704185053226,0.7670270007468412,
 0.5127296625836323,0.11557419068623209,0.4315806939971011,
 0.053071924363007894,0.9909277869404908,0.025508381341791786,
 0.5709395412860127,0.12085990880782072,0.1463588963490038,
 0.34008877779958957,0.21607188409627343,0.6896067675328988,
 0.2286427871195793,0.5531371472472382,0.5652533354282687,0.43309455300169675,
 0.680968771265687,0.723622972510565,0.7845108526855872,0.30650881192462887,
 0.16419273570932258,0.5366556668857844,0.39963857929115254,
 0.9521904329532977,0.7557169644071626,0.676201743004452,0.5417002198641664,
 0.3918711306125834,0.9311759855896857,0.7324808314636775,0.1505088468114989,
 0.879304364387569,0.8226568675149406,0.8399308386121983,0.4170878323323012,
 0.9667600067677046,0.8341652330975915,0.441642012308461,0.8172142099634057,
 0.8144016578673284,0.4367296731374658,0.4169991030088064,0.0543116324525621,
 0.45122193826101675,0.36311600572837277,0.20123258632403251,
 0.7023065329092854,0.5645051720172959,0.2898136103265837,0.03975728758550201,
 0.8947670630921382,0.004987781269682023,0.8402105350859043,
 0.8204786678362455,0.19927164075045345,0.4636193565210661,0.4941013072013941,
 0.2634787820608899,0.32195781713503835,0.7936396097374314,0.5133207102418444,
 0.20850573613018775,0.25230865784775736,0.2511425083478034,
 0.033927905405537584,0.9028942810403235,0.4070725567094181,
 0.7193330953054562,0.6957389459535449,0.40098170774133757,0.449334933375497,
 0.8662662758908629,0.5343055149460418,0.9503164427066394,0.6601018247071494,
 0.6094931915257774,0.42107281278316433,0.8854676944363868,
 0.09805931962925962,0.7123860493173774,0.17190606310480838,
 0.0738251978886767,0.5901319456052875,0.03824185232976052,
 0.46711260465944404,0.6776361367739863,0.697352961306016,0.4472072951763433,
 0.0807166464319905,0.09035937088881296,0.24610672166824132,
 0.7795456804762398,0.2883654958674957,0.4146183618067436,0.46010506246148153,
 0.9574827353446609,0.2449747062453016,0.8054592807109828,0.4145524359016951,
 0.21473856559346582,0.8527645033924806,0.614321068018951,0.33302167034441543,
 0.593338941587459,0.5307008842448093,0.7432916047652741,0.9552057238749567,
 0.3021475154737343,0.957058642416351,0.45915892537069525,0.7578904266149037,
 0.6044457367092269,0.31551225897848256,0.9968972471265021,0.878127373748612,
 0.05531650032312774,0.22238838196149258,0.5569452427588146,
 0.2993832987419558,0.8626828324161716,0.6984267168233852,0.06862357224711557,
 0.7073726704076304,0.579400817171778,0.2510818410033804,0.62641037808342,
 0.46452875500979807,0.23138551850729028,0.3865210472810162,
 0.4420416951083206,0.09864127159404235,0.3940135935303485,0.9294061816964458,
 0.5075683717225781,0.007524163839273235,0.9814490135638868,
 0.04623853868889005,0.4460590820481175,0.6647302161263609,
 0.31030048653029896,0.5001915049728153,0.3638765969428449,0.9741082353799624,
 0.9279536000888255,0.3872000001562007,0.4031954396263926,0.6755145060219566,
 0.7338248597063647,0.54859527666825,0.8844439437455868,0.062265318603454856,
 0.006179834305947374,0.6985585502257606,0.8003071291441491,
 0.8132818376413931,0.6704175093029305,0.9574468766518016,0.9541681035200917,
 0.357215279626951,0.8625774017751543,0.7120911154917607,0.803785981872808,
 0.9432173683233532,0.816947878713249,0.11146877090904717,0.823876481267268,
 0.48802355633081496,0.17931210814083265,0.19801407501463963,
 0.985850830251282,0.9197165785811655,0.9059981449803681,0.8370008439083331,
 0.6997234106735583,0.17104613375932098,0.038117102312407836,
 0.024678795273565535,0.5058442536227441,0.3781772789984017,
 0.5550869727146621,0.9957229376385132,0.8612474767020232,0.04401635900545986,
 0.7098283328411189,0.937714093357285,0.720382018394718,0.40193752183590037,
 0.22900129393059354,0.7342733865087478,0.09115436884597672,
 0.8398314815599779,0.10947897149014596,0.2594717158529227,0.9129229779155351,
 0.5326810568799657,0.9968687087236641,0.13526027955223008,0.7284781764170418,
 0.41728079294673825,0.036449095500011075,0.6851751720603951,
 0.025813136901592038,0.3469379566089015,0.7956660648436394,
 0.8686007260264703,0.9699665117214286,0.17266077779930167,0.7853423410713467,
 0.7209031171272035,0.11913454233750853,0.47414137584234495,
 0.8172602427663768,0.6312307779283723,0.12665121756246078,0.7129427761232511,
 0.8183450916000677,0.7682741581536624,0.5971478272595734,0.5217833088797377,
 0.7651472156194551,0.5525545171194095,0.23803255132732715,0.5385441757167224,
 0.5216839348839851,0.17229569507143416,0.34484865869659176,
 0.45738904011036774,0.42312124179885147,0.7497227299946371,
 0.5988599651632099,0.03399374362732055,0.5165958607675075,
 0.37570966155893726,0.554727868271128,0.14635401321861796,0.8066689075451438,
 0.6180724397789944,0.4067947054975899,0.954259361780152,0.47712501117039885,
 0.12788607949067887,0.8073401510690823,0.20891832635512952,
 0.8365073787096323,0.13896950100646277,0.6726312205459093,0.6109093507265365,
 0.476778964541983,0.040139065801141904,0.3409651101606108,0.5370212473050101,
 0.5250964542294844,0.30501211361669034,0.9967479537810153,0.9451401479951002,
 0.9467568946461475,0.6491341528641776,0.3454469212279122,0.18049680002091628,
 0.2047593187424529,0.016294671158907814,0.332810748206509,
 0.17817216104753597,0.7756015460875111,0.2648708832282719,0.6476857762532935,
 0.3031265102480374,0.712289210902709,0.5094464400810111,0.8585449192150791,
 0.16731351636678515,0.5019718587762597,0.8294081962420238,0.5057549220467839,
 0.5727267466510587,0.8796639859277492,0.43905463695166247,
 0.18025680495386132,0.7756697751538066,0.2913266383899441,0.7597893843915452,
 0.7039986770813295,0.9105922027109509,0.8000986053058136,0.638467401368559,
 0.7842895859150478,0.2388008761642233,0.9449244151080738,0.18976319512519146,
 0.5910449424833132,0.7773290083889105,0.7832994593862432,0.3066118657611553,
 0.1774801113058081,0.11526284275751675,0.2543904826288499,
 0.32633017438050405,0.049119743752677025,0.5245450787578694,
 0.9416674963147186,0.11524502850177543,9.561520019323311e-4,
 0.32990537618892135,0.700847734155676,0.6646190267470633,0.11175215677171679,
 0.2117331503065899,0.15262525041483443,0.6207390304255472,0.8106149331864474,
 0.498653814837279,0.48489804978429174,0.030395512664789126,
 0.2502606380519685,0.5525777894443243,0.27494432573063965,
 0.10325262370548927,0.49171538014357774,0.4319522023966573,0.558338241459267,
 0.05519963883337786,0.6405978056233939,0.6766485626644038,0.7907122874864104,
 0.7271559426239182,0.4999776812376773,0.05923370601313882,0.6231822272364409,
 0.9034276559606811,0.32540501285834644,0.4809760621707875,0.6611941105128849,
 0.6223708032378131,0.857579206700302,0.45421361286488393,0.2887363700202492,
 0.21353923554331367,0.13780602998246427,0.7831086645000045,
 0.8193723676345568,0.7933277075678901,0.5391232769232537,0.7788825529018408,
 0.5367890405636102,0.1341319124342032,0.0336015892610918,0.5455660132579092,
 0.8219325159727554,0.7742299958886091,0.026107923532568966,
 0.0968181643998266,0.1116854855653201,0.942921971021417,0.6300857300604401,
 0.9818808774208333,0.6909716280163052,0.25111816354614835,
 0.31159028141389633,0.9572527282581376,0.16024870735749142,
 0.039352205256805495,0.5148895102779186,0.9291011941183704,
 0.6204783700811874,0.3915026251486363,0.7489943279406158,0.5415436990118722,
 0.8352987764835602,0.09134875980681323,0.4813677579389567,0.9510434569449049,
 0.5556146141786973,0.8793377216658285,0.5008801315509376,0.29810655542052267,
 0.06046279435415536,0.8318735255743007,0.4059032339353701,0.5055500336564097,
 0.03239245940550928,0.29443148008997155,0.9232425443643135,
 0.2716114154959063,0.3181510075596855,0.7803185279447387,0.6969183746070892,
 0.008063850392047822,0.20086540090817362,0.9280154510008146,
 0.5461905519179003,0.9637207869600908,0.38778553825222295,
 0.11776385657923727,0.33735043520601027,0.30776870752158036,
 0.4034597260261371,0.0729567559402684,0.5404152628830337,0.08504003885243294,
 0.6927567085007318,0.7672942184251462,0.8719565534489438,0.7412002998114741,
 0.9545666011093169,0.8036514690143407,0.09719152093308614,0.7700710467366838,
 0.09706585084144592,0.5503943742245131,0.3936425687810219,0.5714580877250057,
 0.43754799492094354,0.7794052335564758,0.07488944619490012,
 0.5543785166937513,0.32270928525292064,0.586278178616823,0.9876873928689677,
 0.4153940782599481,0.8355173843737957,0.5411246330768863,0.802028657788437,
 0.22493498309659443,0.8640159002365928,0.3151012538150393,
 0.35161559331863557,0.4656831818169618,0.27045081047512964,
 0.41666426087558106,0.5510200671166843,0.5463322276214354,0.7757825559508564,
 0.2531354160102598,0.1835083392841551,0.4095120475843368,0.8889361587359362,
 0.34021790133794927,0.9590695710386579,0.23894539799977244,
 6.651109007110856e-4,0.9462772459196375,0.5295827998371325,
 0.43988512171039473,0.4028699846809325,0.18008288648000326,
 0.3914572074514344,0.9936926429984985,0.18278786824991744,0.6498790374927111,
 0.3358614935621135,0.681892192131496,0.48532329090623383,0.9375983073304494,
 0.5653687011449668,0.13481457332933822,0.4572876829822956,0.8594403325596851,
 0.3589668545989356,0.8443154816145804,0.5624047934269534,0.01387149206222904,
 0.8118053796835479,0.24615177949864397,0.7440799367312196,0.850716678826584,
 0.2801853291550348,0.9589062484809737,0.7077112333166783,0.6347497501476423,
 0.64030941070008,0.9055886290206822,0.6902415747341801,0.5030848367372989,
 0.13004096646445285,0.6356056235321504,0.5067625155679009,
 0.0032257752302947695,0.8377683601269432,0.2910812471696528,
 0.22812055157752664,0.03965424787114469,0.48541233789336313,
 0.5243128845713818,0.8403678806338596,0.519923526284732,0.39504550625305934,
 0.611214503803436,0.3171655437571642,0.9402327031494205,0.8616219356979701,
 0.34420686204373896,0.8375077605253645,0.07794598536700936,
 0.9805544835776213,0.8733292090230477,0.27315736083945064,0.2171719251180526,
 0.6015471796576661,0.6405656596915914,0.27949825273662887,0.5408891526811261,
 0.5349350751150154,0.7934785220346134,0.8764602563976405,0.9987193816011193,
 0.19632495832436292,0.9085353300821017,0.1520571672834926,
 0.36885541605611705,0.9985955377176976,0.6782835396888635,
 0.13436883881413053,0.8707950681106998,0.22290210885803252,
 0.11924912745773497,0.15898982151387853,0.8719917893989396,
 0.6403254154100932,0.5741277524992896,0.3883324508531494,0.903383748514579,
 0.14455276324217248,0.29795487928524755,0.233998725462634,
 0.017397414431965297,0.86086585160801,0.08563365489257113,
 0.053896980230561664,0.3162030556533567,0.16599847613459828,
 0.7528979502111761,0.23288432779625756,0.3077894883382897,0.8287717568764379,
 0.5025167698253078,0.48989421701448754,0.14365422559468632,
 0.49435725991072876,0.8861433856898094,0.7364980074036227,
 0.44150879892867156,0.13910760493545649,0.16436128508625425,
 0.381631705752688,0.31870749133286513,0.8578452382359326,0.22700889663411272,
 0.8299426987788779,0.8781163414096173,0.2523418013992438,0.8368478680225357,
 0.3101501926671011,0.13909589461420047,0.29985802197661093,
 0.5210640167289129,0.23645502497586723,0.8477916786001645,0.6650047678627368,
 0.659412497418615,0.922565069778038,0.8746015211107977,0.36048390179804757,
 0.038752091049696014,0.7534493218212912,0.3931266055120034,
 0.015102735503585873,0.6531005328420244,0.24091033082513846,
 0.16811061852205966,0.9374352582825131,0.2412968413052049,
 0.18172135518096344,0.7479582160869446,0.3696671296438758,0.6666559403007005,
 0.6792585611889587,0.5908676266097104,0.5474710883316714,0.644180703963011,
 0.20502458610929197,0.683390816637981,0.08661690456292859,
 0.036870191446301215,0.9901357748326314,0.7342994030928638,
 0.6282662790909376,0.6785641145226926,0.09600213401472257,
 0.25378975754763167,0.9843065095272024,0.4157627563021312,
 0.28361265761528487,0.2839955060141114,0.6724203293988981,0.9928892396578742,
 0.27127685622497066,0.4246506344593616,0.8737854356612694,
 0.31728148462579453,0.45497122267495815,0.509593309248481,0.7406777488859358,
 0.7993566790741763,0.6284295890092363,0.7483166729304862,0.14829356752614886,
 0.5743526067173376,0.24037505634008383,0.6233843470632953,0.4145900467718,
 0.43797577390760545,0.6201905795803268,0.6083307737878455,
 0.00843467253865815,0.513836836850011,0.3095177775130219,0.43498888069998354,
 0.3414359014599384,0.08985269832606746,0.1251943744718642,0.7011020553063309,
 0.8011559188366929,0.9349705006591726,0.30133151345677345,0.5729590341800881,
 0.1519789863849783,0.7114491872804217,0.7034738248311474,0.7710414380145941,
 0.11973206016301119,0.3501302389768409,0.37644632934702427,
 0.9885357464331246,0.649628581687263,0.7947505610763371,0.298730398172649,
 0.8703114739747039,0.006340675871176593,0.9825093104137006,
 0.026513994298218213,0.3867750921815697,0.8576448432257173,
 0.1094970800365822,0.8543846638378685,0.9703430958055936,
 0.014748967022358728,0.3954556228207269,0.4050190382739213,
 0.23838410256048992,0.7570931420027532,0.646365791205198,0.8317023910020528,
 0.45651287342148783,0.06597718100963901,0.6603181453551233,
 0.5127335419253193,0.5124920573731188,0.6199225377664693,0.26046855274295644,
 0.8584933978593566,0.9901205315339832,0.1863858556115049,0.33645708491856197,
 0.9529960149440437,0.533466173984898,0.9697441775621226,0.7320426745666618,
 0.7545179197113057,0.40369368455732557,0.6295749173924944,
 0.08397130011549114,0.09259928744040702,0.32812530753473723,
 0.3883577447030414,0.03894201869135838,0.16774838719184992,
 0.6366913267958171,0.7847650971294067,0.36024821066331936,
 0.17498837898166597,0.9681678901432182,0.8884436210652585,0.8408713866728212,
 0.879797147316324,0.849975736203201,0.03886101957644583,0.17912810130386758,
 0.3821283084474575,0.5798806063227818,0.4083459857394688,0.9400880658778691,
 0.5808152742278065,0.41336802916628135,0.6255980976242177,
 0.34367308329079616,0.08323531529175432,0.5133802907475482,
 0.4685529604051004,0.2955054426986701,0.429270768251364,0.5534764629953836,
 0.7167577951976035,0.28661867167938815,0.5855712126270781,
 0.49022446577145484,0.3134108982403727,0.7386261044318743,
 0.11670762456834483,0.14033230197240243,0.7572403796699884,
 0.1521296284484086,0.18882507038719598,0.2715907190811382,
 0.36008817044879593,0.014860821199727647,0.22180106378736864,
 0.4545564686871503,0.46451434253721036,0.6739417700443328,0.7979230272151945,
 0.1154086460422118,0.07518758000797077,0.06247911380296178,
 0.9928129436954539,0.024293415800426765,0.22143216793931764,
 0.8879136006308358,0.36077805493476356,0.30247084326565865,
 0.3820175396646437,0.23092990231854915,0.4316606347194286,0.4863007438443713,
 0.19317243033175668,0.604768576504075,0.5003121174376884,0.633165593302324,
 0.4345262339809991,0.6790839931950682,0.01960019313818162,
 0.48922025235492206,0.04756932569610284,0.90715911784961,0.02793868949106515,
 0.6421006134697985,0.09145430595772741,0.1340869088702379,0.8642688521601504,
 0.3147989037741671,0.14156205973285996,0.8255034479204222,0.9368026788251739,
 0.25427140884049715,0.23939110863723823,0.15575721742136972,
 0.06281085311390189,0.5361601877022659,0.5358190929334301,0.4026855396828094,
 0.19208045570435395,0.4844580696962264,0.3275054362272446,0.3689621631612999,
 0.6469372668638332,0.11717234190084758,0.33082865262513617,
 0.48996278910994084,0.33061282075449094,0.8887204276284084,0.741417815826928,
 0.7568257806349775,0.8390388122487155,0.4769359826689038,0.205850106825337,
 0.39167885598662733,0.06674741758126412,0.5265953381969186,
 0.12237693860800736,0.023200485067843335,0.6140291005747653,
 0.15492692649522521,0.9873235979152284,0.8479187062876188,0.7628048114292854,
 0.4832378712888987,0.9735520804509761,0.7191870537091388,0.34412301196255723,
 0.9101915900619436,0.8302840243110867,0.2616952964948679,0.4216541393900348,
 0.9186803307210788,0.49448513616504863,0.24825406263410743,
 0.5256498847728861,0.7725418089242015,0.10881403186878047,
 0.27542032396974203,0.3845280649089501,0.41950478778680533,
 0.6476236884853378,0.5683466286869094,0.8244051444127735,0.12317621660657962,
 0.9270008585472899,0.9457817635988139,0.731629821454907,0.9093972717384005,
 0.1664113859431151,0.13427062821504854,0.8256694782156655,0.3258670684630188,
 0.7359573609918739,0.975854267903812,0.9229816131211912,0.10791001380896137,
 0.275499606604097,0.23979025207319882,0.45597020807114796,0.6903453890118482,
 0.3751711775085883,0.32479302346732875,0.1572203067709077,0.6402837235146208,
 0.8612438501601627,0.6123600852301144,0.6202429602153736,0.9827893465471369,
 0.4387428429119006,0.03129623553026473,0.4844694564274302,0.6090524661886457,
 0.03172520436862536,0.3142363522740722,0.9874617559617291,0.5684115945275837,
 0.6963487683743532,0.4289270866200263,0.18987930876677506,0.7274943400754516,
 0.8969044950679181,0.40993770390262974,0.38957491036872227,
 0.8916609988351556,0.2971946548240928,0.9506266220374866,0.5299640537852777,
 0.1750857271058499,0.13865505832764669,0.07765262032572418,
 0.33781933694370725,0.447795692239881,0.3535163159268442,0.7677536405341089,
 0.3992243849717193,0.18664054285695908,0.536182207977359,0.05529824163305552,
 0.8283449927349347,0.9996059289167566,0.9639153994724742,0.13781398007934131,
 0.3258638052214765,0.7571355695264312,0.3525308959929494,0.14169473270090538,
 0.7044524864467604,0.947641132681744,0.6277893770285259,0.7963573801137968,
 0.6700081010343102,0.5845449036609703,0.29884289719448087,0.421137775693465,
 0.6191760048878664,0.9808520286984521,0.5163820390045605,0.353983525435665,
 0.3739249194129364,0.7763273352784231,0.5404226949293327,0.20205635910778397,
 0.26159493746217244,0.821324683907991,0.3365689749655939,0.9881572687049531,
 0.4332931106371619,0.5959715288646438,0.7520583621633959,0.5770520118933743,
 0.3730607444091353,0.30683918309042113,0.13160151821987764,
 0.9377558494415208,0.7935297149440839,0.28980360981508224,
 0.11293147003466142,0.07772959697090642,0.7167892366126609,
 0.9784720299798902,0.28625923977600787,0.5104745389905023,0.6652802735441009,
 0.4477339839782585,0.7511346373427993,0.47986200858545414,0.7810988713833706,
 0.2769371939722304,0.7748138012555918,0.8039994148811299,0.8925534192565265,
 0.706922628077036,0.7968695822125401,0.8030842687211971,0.5759821314707949,
 0.9197682404277655,0.7583318585594365,0.6233075333220207,0.48001336865679944,
 0.3106381828185737,0.5876651082248076,0.8497456307891511,0.6113511971938738,
 0.5443255404620035,0.32750600587916656,0.4471669220499893,0.4234346547079698,
 0.8772885882224727,0.06224060132301745,0.690676045370015,0.724032398285496,
 0.8465937104916662,0.857987843501085,0.3695519551470867,0.8290565662651133,
 0.32150270361147704,0.7821876544998387,0.8244815036088401,0.7059186301496856,
 0.10376362063537514,0.27118345031894464,0.24605341966292582,
 0.6223203932475225,0.26405642513749783,0.5330144726632329,0.7026647299794647,
 0.9305007005444247,0.5293215010467008,0.17884473289189362,0.9631117042649285,
 0.8558118457737719,0.544186463785874,0.0938017667860519,0.9576490484834804,
 0.5916363196565572,0.7090921533959156,0.9224256066301927,0.8668607403191725,
 0.0929002602114174,0.6093027836525751,0.46247173347374926,0.5163233322450932,
 0.12197563825769053,0.46831177630269527,0.9829899073079597,
 0.37985631180113666,0.2966013249337218,0.21433570041117633,0.792772330957578,
 0.6579046018516446,0.6230949726460491,0.0559557658989126,0.6491111686439224,
 0.7514487175802462,0.7196666582512279,0.9567530356620042,0.407866421384798,
 0.10650939313406227,0.4620781277291578,0.20548626909705403,
 0.2909886325762113,0.24413319275318557,0.8698464416559472,0.6126392683742695,
 0.7445580855079585,0.6845146103253844,0.2694239200020815,0.8832228061254314,
 0.6623283035039365,0.557655958475376,0.7824980760759842,0.4530950407321267,
 0.942238155170844,0.9549256940389834,0.1736278049891784,0.4513856003130575,
 0.052835562777650336,0.36180012307034826,0.503113879659571,
 0.1843752242797213,0.8188332229037055,0.012775564512651383,
 0.20063994750822367,0.34127493113281093,0.28865893269891274,
 0.9317093232536471,0.9225777551559357,0.8605836296948033,0.3634953597851739,
 0.44423933054866116,0.49594756622412794,0.7237904911999221,
 0.10161484314103542,0.9518312310872229,0.8504826829605397,
 0.42900435183001595,0.10267662596396765,0.30373946578209154,
 0.6007925036317504,0.7924157397589289,0.7132482158117177,0.23370274587438022,
 0.48738890534972046,0.018975983876260982,0.7356252175374806,
 0.9430664923465377,0.07912413265355678,0.576562719396361,0.23306787135264495,
 0.19358778614255334,0.5902508459756367,0.6878119828675593,
 0.45213623925119795,0.4401928563735672,0.1971660218484822,
 0.10879147343387907,0.47414437729090864,0.867239688443129,0.6590871734856888,
 0.44179200311688394,0.049891371500881965,0.8751506840715575,
 0.5187913159124062,0.4164295930689029,0.6622918146644602,0.8395440846035808,
 0.7504246241774092,0.324037069609613,0.05711889983497476,0.7877749791448858,
 0.377420276491361,0.4887058125666328,0.19010343738779578,0.43754125665495736,
 0.39764029126702916,0.9628533841014374,0.8586892481885993,0.422851045438984,
 0.48462739523973086,0.31644663724298583,0.6769036160910585,
 0.5694880199370431,0.002166532017790046,0.43443084702398294,
 0.43721832857069176,0.5312847214751548,0.24194288767216232,
 0.5360821019435164,0.25339544744924136,0.4943626690170525,0.5381117535824578,
 0.9567798369198007,0.6389113895208399,0.43212910128146176,0.5394842147027565,
 0.21991230902540693,0.959514453638866,0.2843904248044289,0.31930729580967165,
 0.8173263644315248,0.12642983041853362,0.2088318069594738,0.8218815945056042,
 0.9327857685823168,0.5384318630508562,0.9405664721412991,0.39946448964168213,
 0.09908985296600659,0.47532752069267437,0.8567667655057074,
 0.07771654743765888,0.6061997130509591,0.022290304230428548,
 0.6523481878975823,0.3925710656804393,0.511649988629302,0.43830316703296757,
 0.10130167629881803,0.5481642170810491,0.9764211559262566,0.1912459934117462,
 0.7031022647554295,0.7821526244015521,0.2743111502019485,0.6592449604293471,
 0.1255858459091851,0.06333897592828097,0.4097035753861742,
 0.31811246601434395,0.5101746722087721,0.608324522119755,0.22170854330077283,
 0.9570568913251555,0.24824077457033622,0.8818224971745362,0.7596354445692265,
 0.6183835942129605,0.264430029675801,0.8421127264216488,0.4439844535063008,
 0.038325596774202886,0.7715236172556403,0.4902488670173506,
 0.38623088870410605,0.9444471450247888,0.42972789255628374,
 0.9321880358654061,0.6176591441432165,0.4764609702545025,0.547343250228258,
 0.6310954383645896,0.20870129573155138,0.7046320070029631,
 0.49909296286335914,0.6895409036910767,0.568189186639943,0.9362524745797411,
 0.7035301511952493,0.6110034109782385,0.6210164303310806,0.3021673483127426,
 0.6642112788743239,0.2447715177707711,0.7426822205257936,0.5870621955804507,
 0.6210325954185663,0.3916138738778361,0.21434686295848582,0.574892482188418,
 0.7307772162343124,0.8391676679372824,0.7821725657481345,0.14554567551992492,
 0.6914940254723659,0.2510494214120501,0.9900547835749096,0.5715151605613186,
 0.17911956132561324,0.9241639038516865,0.6353693703017026,
 0.27962925034458264,0.2742960775593358,0.4982116038661196,
 0.33430022879055155,0.058942341206493776,0.2075710037438503,
 0.6665787396556414,0.23660104076250466,0.6761475329337963,0.6278115646303457,
 0.712816861844447,0.9333700070583595,0.6665573386095472,0.13309120224879545,
 0.22908357685781766,0.03922515101269419,0.8886192138429334,
 0.9487843049660936,0.7771295926652609,0.09284290721052568,
 0.44961329706476105,0.324538030144939,0.02002886335190357,0.4528360501541159,
 0.06579382297296754,0.5508462958655828,0.4925737107341399,
 0.04019846417108397,0.3071510598344199,0.6988518294508951,0.3925115612446358,
 0.06360097681535937,0.09435408931935407,0.48193052527977354,
 0.8419950324790787,0.2992750703993924,0.04451683014106422,0.2961031500579633,
 0.4377680331248307,0.1371022731403898,0.7871469559970836,0.5233255097772801,
 0.2002975471963333,0.8732767535039433,0.26077480115663343,
 0.40101966827408975,0.18008347806308778,0.9072624549544899,
 0.21475010919055082,0.7931235637005707,0.8462273043180984,0.9596581490784662,
 0.13717874105882089,0.5425995747216743,0.6501036115600917,
 0.11526859100620124,0.9951575011267905,0.8913088637500002,0.7442803730434099,
 0.7134768340693837,0.763462145333285,0.6921353854640354,0.31790228091353545,
 0.2728762110741032,0.3097233248592931,0.9977537409280741,0.5353917316461864,
 0.43916964569730277,0.9668326671123714,0.2789805928359209,0.9466234397807537,
 0.4965949446343729,0.0018964477422359627,0.354633440445967,
 0.5406939139097238,0.9277700579452485,0.12344020768253594,0.3127759725503474,
 0.7291232939836274,0.09859810583760753,0.8185871018130846,0.9263107085621775,
 0.220422617750211,0.5917792438226422,0.21150893458067677,0.9352246822948461,
 0.9892996510508713,0.4182505044784408,0.3944687333884621,0.9197389452134737,
 0.7254122808227592,0.9363632332592241,0.5083431933623084,0.9689429061117972,
 0.5629457484109426,0.8108855295809243,0.6313152889194972,0.4406214687902126,
 0.5576331943881331,0.9074011069567579,0.6083104342255066,0.8614648528958453,
 0.054817326363396535,0.02969576523608486,0.8563933261008257,
 0.5674860317052705,0.7170863977783493,0.6411890636442825,0.6909462627636578,
 0.8058861155102498,0.3737921046782009,0.7177737935200834,0.4845423142541806,
 0.6417554063841702,0.5168008443337896,0.2596403548629901,0.7067431901725156,
 0.3403534502627128,0.3983356167082486,0.7897505516587427,
 0.002709862936747909,0.6688393796870663,0.8587228095640678,
 0.40619051409171336,0.20029231420260818,0.7849760662688519,
 0.7621423824151536,0.4904797033128174,0.9711808838460174,0.6672510734534681,
 0.6756230373216802,0.3952795820342022,0.34523127894988515,0.9108684222016313,
 0.7074025807752926,0.36219560962995,0.20285373209975044,0.8258450956751799,
 0.34452508702916207,0.8593364570272453,0.6553012989247655,
 0.15880970637704783,0.3766130028751067,0.239239712621808,0.6022516094288437,
 0.9451755564695976,0.001765978257357692,0.7197121685271071,
 0.6262296995952169,0.19786041737571447,0.12241158713665656,
 0.45911207557886524,0.09823828146165048,0.8223363401982167,
 0.15209679378105556,0.29467839294014364,0.617113373005468,0.9692374250217664,
 0.8542764791994197,0.31285635018937685,0.573316262244037,0.49036304232002115,
 0.004968685165191555,0.3114073134546125,0.4263977107693079,
 0.40876415945414024,0.45773014139635904,0.5326632761865681,
 0.7135810359355323,0.9226748010872521,0.08141329911101047,
 0.36932545495001823,0.37371333146506314,0.5769672676919281,
 0.27075264172933877,0.4003494054933753,0.556912428796984,0.20021921098382645,
 0.7028292926814493,0.8585016450967414,0.19513032099949912,
 0.38983130333544114,0.44885520550790037,0.541972666238724,0.5784810706693203,
 0.6490686255483782,0.14436820788433702,0.34279381501159345,
 0.27894381573779403,0.020710721090966056,0.10303138081681784,
 0.9077201244779715,0.8021229268466705,0.7506651552513286,0.4138942542077966,
 0.8235614634782132,0.3769830463507251,0.6352040592741549,0.8395574548347173,
 0.41260369358999516,0.21483465643510935,0.01832774749579902,
 0.20406351888198104,0.3086617566503712,0.3011868648874503,0.7198747154657679,
 0.575394992404622,0.850351457114561,0.47770327697668524,0.26994791590337563,
 0.5070197990082137,0.6999054104534516,0.4064456064209989,0.6724424878053945,
 0.9657519635890489,0.940917320746498,0.10211883272276356,
 0.004706418243087818,0.13502270297353192,0.005081317289333143,
 0.2716706509813266,0.5464352410206941,0.9468234716726828,0.8037966243377311,
 0.11881994959272524,0.5594646411706866,0.46130361405940334,
 0.28710865961841714,0.9516550568501501,0.7078566671857762,0.7489231106346602,
 0.3798719926063163,0.5350397546809701,0.5894105125614658,0.440033298261436,
 0.6065780310631514,0.26940892538744776,0.5692272858837848,0.8201492449367356,
 0.23676464382892626,0.3920717616257836,0.4916258284451249,0.991256247963781,
 0.4686276469535863,0.275029090159867,0.13833910733618937,0.9132313603296179,
 0.2094857540980466,0.5964270300259248,0.3956541528218708,0.8084948240603117,
 0.9970989805680918,0.766739485747131,0.4313100910184642,0.7233350387965656,
 0.7172230488611917,0.1348043787643204,0.38117920283266604,0.6069076707805856,
 0.2690465496673553,0.50528680403992,0.43382286968005135,0.8441950292995287,
 0.7128359312418788,0.7147598504477042,0.2901979806014625,0.06533312920457224,
 0.04266321761065417,0.5533897557769146,0.9742063539704199,0.1800951947435191,
 0.4403572757546017,0.08453679915345425,0.15233550119719497,
 0.9978368728074114,0.6290423348224481,0.7414619588882608,0.49021208886261725,
 0.6632221815021436,0.9449519653454497,0.46526134842182754,
 0.45508159006739024,0.1860743856468594,0.5473624440964934,0.3309075929019305,
 0.2400447980702115,0.5773260589870095,0.11143757860221293,
 0.35575488281922096,0.20610175153430277,0.09686745444713551,
 0.23242743139312905,0.22368106619615113,0.6451237524439193,
 0.32749649953690196,0.976221223031885,0.6738306596139882,0.2579262513028535,
 0.19142610249350211,0.5515366404157938,0.94116325713312,0.9996398989766163,
 0.4399567891834236,0.3943237632556441,0.20871477812232297,0.105194626182743,
 0.09629545093030967,0.5292996597797175,0.29363039006221225,
 0.27491118347331467,0.7421347797341393,0.05834697213998219,
 0.5174279604890202,0.4774022601914155,0.7083217177285298,
 0.013714178953203926,0.6476786085815415,0.9367571164801405,
 0.1429972228480061,0.8797032249583026,0.8364195632570106,0.20455756268251546,
 0.8477497449568605,0.43194034871356035,0.3654207890803368,0.5160588561970929,
 0.015995898372454187,0.009976544968650525,0.6078721787370583,
 0.30156371245762725,0.06852016702363584,0.8811956260807048,
 0.15140402955275434,0.5734224996029955,0.11245444720179298,
 0.7844356097249613,0.07008418631032953,0.061276705288678124,
 0.6643969229352141,0.8825226494955305,0.3218427209506942,0.46963641226383446,
 0.35167503188009686,0.30479048203587367,0.9743061986289152,
 0.7667274080447293,0.00209935756944446,0.930441568761142,0.723081739559742,
 0.2947556405394499,0.9499952246711225,0.09009253031703424,0.6297196101737297,
 0.3042957705907837,0.9656305745066543,0.8534700656602667,0.47679279821608134,
 0.6401918296870506,0.8102166516376498,0.845064948250146,0.27755623211701486,
 0.827260226035213,0.9935608550206734,0.7729898703969377,0.8279676097291605,
 0.903719138677923,0.4545594605685588,0.7182256388383559,0.5036155954091825,
 0.8331231618101301,0.08295824201602697,0.9046307219725045,0.1511265243069232,
 0.4857846798453861,0.7143339571635683,0.629784298257274,0.21418395035018323,
 0.5568982256301312,0.7683378495283313,0.026512267995529415,
 0.4937096594019911,0.5307157441652253,0.5808916765341458,0.3845742136992216,
 0.3329633597023771,0.4246712666044372,0.730081919235502,0.3124936414062073,
 0.006821372361566658,0.5008779042298499,0.4184144140623004,
 0.5756259304317208,0.3682227635242735,0.8412667868943726,0.12212138131960282,
 0.31648334940306544,0.4267121004182979,0.0859248993588434,0.240655564316681,
 0.011495256098760231,0.5285015898762091,0.11772314107078308,
 0.05393636789266898,0.32492857437452605,0.12483789154511227,0.55278453807494,
 0.32501675731840185,0.1704016520007796,0.5236885078143154,0.2672731071161317,
 0.13711975285281797,0.49883402036748536,0.20376352223254124,
 0.14986493235534248,0.9199787658225604,0.9010945357275251,0.7942410382407153,
 0.031694326019538144,0.12850956603295516,0.49519111867686527,
 0.04895704920817212,0.9885239516982428,0.30987118692041216,
 0.8153778486151664,0.4490408991665945,0.01095918800571316,0.317133589046534,
 0.7125644000913365,0.5111329183352767,0.46465863406544106,
 0.023163170441364067,0.3934058255710443,0.8436146480100317,
 0.042691226634408164,0.20929773528061146,0.42618436621760325,
 0.4532050579670561,0.7474513509318894,0.14095915802623926,0.8194490636387389,
 0.302723601635171,0.3858668199668548,0.2664982601763789,0.43701386686689414,
 0.8926813830703173,0.49994897821263573,0.10062107646883167,
 0.6027800154383326,0.5870314496016926,0.8368114095945721,0.5801782778863217,
 0.8637065126841597,0.6233030198625739,0.26611595061067994,0.9433582366403889,
 0.8633883095619082,0.7091630745589848,0.9955281454523205,0.4185810915304824,
 0.09449032673836,0.05160370910719703,0.3170656303896717,0.615936360357986,
 0.3618644472502456,0.3175409834543954,0.6754154093741555,0.09965458246069459,
 0.019632152591103846,0.6049816649811348,0.8771683259362785,
 0.5785975216088611,0.20855409473133368,0.8304824449256605,
 0.16903863420339504,0.14644802124047074,0.43641935221349126,
 0.1814960110220165,0.703342937902202,0.298501766848399,0.12684861911826317,
 0.2915076317451777,0.491222419250649,0.3068512801690485,0.3187285349842839,
 0.08461966182689595,0.2658737565811311,0.005605158591506232,
 0.3703544417039015,0.9654117286647551,0.5233640970878022,0.349811264856813,
 0.883205164571696,0.380912924614793,0.3862643772741552,0.45134287621374647,
 0.29106299662413015,0.7808209768796659,0.2864162641480785,0.7420823586352674,
 0.47804763665791183,0.15201551149443282,0.30970357284055605,
 0.2917777423992164,0.8932036855419042,0.777464628680654,0.8430985812381242,
 0.6077455256376695,0.5488853468357997,0.9254119605903716,0.1286005454565814,
 0.042781095626122645,0.16392669574390983,0.12482580390345144,
 0.18593592640315404,0.42515484258022707,0.759812273653904,
 0.26860792586429005,0.4016622900215139,0.5592033312390302,0.6114169044609791,
 0.9940205177047335,0.7303730690151703,0.17397669837497143,
 0.20681302236380628,0.7550438300133793,0.3197132689941795,
 0.036210860893004404,0.5130086130622835,0.5059731940672123,
 0.008550457006950007,0.2657493477281213,0.6501558296583779,
 0.2613263450086202,0.25688722165615663,0.9362629122547941,0.7782052782377188,
 0.7658869604022106,0.09757655322236114,0.9176844938251925,
 0.030334127990912618,0.9698156946113284,0.694961289009301,
 0.14430372814960224,0.7695543210207907,0.9309187553441894,0.946614049278116,
 0.12943661758346825,0.34063495370880303,0.2953104810676217,
 0.9288831485578597,0.30188470169244686,0.3058899610311142,
 0.46706769921272784,0.1211223075172495,0.43897013252471506,
 0.08266370573876047,0.23974718468098621,0.3723632144083523,0.705993178355062,
 0.12443337868445004,0.5248118577448843,0.6330001554912293,
 0.20050821771466487,0.41869031502115517,0.3550289647226955,
 0.029166865713689205,0.8625869637305666,0.281522044335898,0.5680571148882672,
 0.5051237942188386,0.6874606543001378,0.08232154302600236,0.6738121376556016,
 0.520692795805757,0.1589663360011463,0.4525004128417214,0.9103135105389497,
 0.17382466478240421,0.9061559153437764,0.43174551239738057,
 0.9257673340609625,0.4907381355289886,0.1798966722089984,0.5189612553910223,
 0.17682137409289045,0.6193225835121923,0.8408075361791809,
 0.01473283776247769,0.28332379360617277,0.1363302030497724,
 0.38806235779437404,0.4833155957662908,0.48815690077812024,
 0.08998723406630904,0.1499013569477764,0.798483302474036,0.5979752722081898,
 0.8680467863325604,0.4606590346404672,0.25210020613989204,
 0.48920612324361434,0.1406407327660164,0.6363639892737762,
 0.15836790879984308,0.7500931326712253,0.9101419449027484,0.8096295621929298,
 0.49930693953492467,0.7659961094432672,0.3298188320606391,
 0.38841090130321865,0.4248709466237609,0.6177605875726838,0.9137868587020681,
 0.9596658853369794,0.5040688403906655,0.682674765937964,0.8405273199726663,
 0.5404361921278851,0.04293751296368464,0.08728779193269454,0.386558602872091,
 0.7474951401503812,0.5842969554206364,0.9537539669363021,0.7385142808707733,
 0.8182879357483474,0.6006456596196296,0.31548581909841633,0.8111379600604303,
 0.4289205745100979,0.46927104359616256,0.42537832065126,0.5186058329547047,
 0.9955557895531226,0.4010454841142381,0.296041282631038,0.8062532671686027,
 0.2624570141767065,0.9338883756094792,0.5422182384285399,0.6094538664154816,
 0.993272528815047,0.5685542253920817,0.4428151846834676,0.24831690890069835,
 0.7256185614239777,0.9538693402639197,0.09814966032307926,0.8163767558606592,
 0.05908503646270291,0.8522323616984552,0.04871897018527438,
 0.6096245806460585,0.819526040343057,0.5477659670216753,0.9282535106743923,
 0.4776201708448503,0.2232319588539904,0.23386237243315933,0.6671264786068457,
 0.8057508070995686,0.5660044314715327,0.4257667779750247,0.908093911661401,
 0.6599107246771636,0.7414637561672833,0.3500765475712394,0.4005956168430025,
 0.5560189285774223,0.8248567425303981,0.9760533644245207,0.9116688035631002,
 0.44352524098222257,0.06364456617050207,0.5510079474701634,0.63518068541251,
 0.6406340335827335,0.7461197439702695,0.34103949029773206,0.7746324506547216,
 0.4930595215751361,0.9300092026497336,0.48228481647814714,0.6446620938451743,
 0.5532850872906044,0.4754808996644967,0.859328206080237,0.2607979830126709,
 0.9869601247600094,0.3451969996607688,0.3941671617720005,
 0.006378925292759741,0.953927956940742,0.23133044707943662,
 0.16805430715406244,0.44717472988867346,0.6764167516895558,0.48468963831975,
 0.0604131208874803,0.850555240375712,0.8216740447238173,0.2510308222261688,
 0.12053231836173539,0.43963444879936286,0.07614229030625286,
 0.5432216201922574,0.542774330905823,0.745042290085753,0.23809898342264058,
 0.9917776848880838,0.5003981947894435,0.797851881608644,0.6208405234094724,
 0.6485646464832423,0.7398452442390744,0.4992877645681839,0.02166158349501668,
 0.36202326294020437,0.7393528956130682,0.09761801556325511,
 0.1394535752698841,0.943249087734691,0.11122323354001029,0.7017306409790186,
 0.6548107299627166,0.03171297579540411,0.9273896663501203,0.6590625998428986,
 0.7958700136848572,0.024276331047557442,0.9194155036340123,
 0.5512000197114847,0.39294209888043385,0.1385004813611933,
 0.49796738004627616,0.43853682930824156,0.0630450629744509,
 0.18956791630653813,0.45915320287926886,0.9606075405655042,
 0.9701154128284077,0.6475914203215315,0.10470760419525182,0.3777967221285119,
 0.9885493258466082,0.3289679132670039,0.42355930048343104,
 0.09805134784500691,0.18132542594695966,0.5481022950769363,
 0.3398809225548548,0.2643442440523748,0.75018430072355,0.549569804257616,
 0.3616628859515396,0.3929420428901259,0.10252388858605666,0.7430584863409423,
 0.5523371724733738,0.11565771287437543,0.2991013817693421,0.8310600119669864,
 0.1974259832923866,0.4454476448815684,0.6622928174490355,0.6838448253340375,
 0.4332915140631778,0.2201715006798154,0.13790459094833063,
 0.23835394134924748,0.978086901585712,0.8690739155298159,0.06237215883244829,
 0.41562697113035163,0.5977661553950767,0.9725443162788548,
 0.10442760065029144,0.7350829368869465,0.678611448063178,0.3902672290492687,
 0.3451009417668147,0.8848987488991392,0.595132862461084,0.5893880831436142,
 0.9410986017028937,0.34675769735462036,0.11441125787421602,
 0.42657602313018617,0.8853133936350721,0.39247495729711346,
 0.38239446881678496,0.332274739449846,0.4432144474727667,0.9751970626376527,
 0.3443455925761447,0.4911957923689245,0.11123814055630765,
 0.44803155199989486,0.6508430134006545,0.17954373150135605,
 0.3219293126382101,0.5942413531507937,0.750037532325257,0.514552297343792,
 0.8190938457037227,0.22157580831330048,0.4843683316128492,0.5651451639545254,
 0.47847638419484007,0.41620248189842746,0.5879123991728415,
 0.5536070947884032,0.5997990351771317,0.11399615702743304,0.9229960549391346,
 0.7855641946321497,0.5335596946198029,0.08078285193186263,
 0.15535782041878865,0.9858743430544175,0.6150322959119354,0.2895105936664708,
 0.9681756399596171,0.1156945949382483,0.47906241028109653,0.350100600839214,
 0.8278977555274091,0.09193720708562902,0.04013827571978934,
 0.01303115833334978,0.17750754372936528,0.3136714877063247,0.136010509063751,
 0.35626705055875774,0.6829664354622874,0.008287912027214905,
 0.7799410842637071,0.005424331568510965,0.018807675682931624,
 0.4287026776317596,0.5242270845088994,0.3105707365004807,0.0897368687049932,
 0.5839006213378862,0.42129757442514115,0.5309218296593747,
 0.041538747423310096,0.5398313644936278,0.3896455356110018,
 0.05845452205890611,0.908503449853194,0.6546247937823175,0.15373783407770658,
 0.4743500653144609,0.6998120104863544,0.2783084400316589,0.7487471348101313,
 0.07179312615118827,0.04086294604233698,0.5226308099064755,
 0.0406699569164114,0.3354832887649576,0.16893406597025495,0.6234401389974352,
 0.14211726099800281,0.37106750687538337,0.5894326952080197,
 0.7300739678902007,0.14364722720728107,0.07927177333594093,
 0.07644524908011985,0.2678067181757059,0.16049491911970748,
 0.8741642056672059,0.28980998072852027,0.7720379215245974,0.2221253376927499,
 0.4884811604388021,0.28101883469912914,0.037675345208839195,
 0.8445122917445856,0.33302255847042894,0.526319468754761,0.34742054290022273,
 0.6784104499876059,0.7393870379720785,0.34951216460020107,0.3134054039399681,
 0.13699899418251693,0.10186296100852699,0.9625395622126623,
 0.9292147490670647,0.18873930994269672,0.546684296340655,0.3754198741056046,
 0.5763261191447684,0.939751148871131,0.5127926815068222,0.6594024020150373,
 0.37988986122521085,0.9675089277755737,0.6376861739678645,0.5858590968720259,
 0.428558681278116,0.18267386725639834,0.23502995878858635,0.508913550455142,
 0.049052078292532,0.6051362449311262,0.9638839814702658,0.10593537382532403,
 0.24502099418291579,0.2624803553257017,0.45215433618639944,
 0.8768330458758057,0.19655857440662583,0.5762756129046004,0.1742234142572372,
 0.0913210054775404,0.3512789048628444,0.48974287350019163,
 0.26721632521883987,0.5622112682335298,0.3353660487805664,
 0.27954540962874863,0.7755419173731032,0.4879401123634892,0.7471108851230629,
 0.9765065246194609,0.39797656676186643,0.6088655481454928,0.5689892317200315,
 0.582515138305434,0.2580414822786149,0.2919977268066234,0.18666815977558282,
 0.42474742087128003,0.9390790138591203,0.0022573634391036546,
 0.7943078119326995,0.3273368725680956,0.8105103573168517,0.21310584935722043,
 0.07358645571079236,0.9916969144930986,0.7403354122085811,0.8068383762299214,
 0.7123344974850151,0.4973022356204966,0.41985480520762475,0.4281661765875816,
 0.06241370362515264,0.9181803681322597,0.011925206243853959,
 0.21785488146557075,0.7714643210645294,0.562082950083572,0.6908189047812551,
 0.35350543151847724,0.429847745216255,0.87599912175476,0.8435834058405096,
 0.9550610576002306,0.8944894717641362,0.5088415019017897,0.514593403998934,
 0.40077846557719055,0.7353249948685567,0.6229968268551247,0.545317330815353,
 0.544146866648256,0.7734028990688233,0.6469867729817677,0.31834951118948585,
 0.7755633752220162,0.6103101205074325,0.5540193137351199,0.7729382261524257,
 0.26656608754784306,0.775784988932271,0.21939811496210182,
 0.23660123528270238,0.8113598544040925,0.9062155059995813,0.4856834823031292,
 0.34769729094364765,0.6935876868292759,0.7555216604599222,
 0.19054498270904485,0.9891443096059871,0.7383090982678342,
 0.40980658906174117,0.3487148449538662,0.9891980990216522,0.6231905584247995,
 0.29997771487441216,0.023546066754939754,0.2548224592145685,
 0.5850964970394712,0.8101342687822005,0.9458265517808114,0.8267711251101597,
 0.7156428931710991,0.4264581180320244,0.22728377641382647,0.7405681946249761,
 0.239267446062027,0.9033300278861058,0.5723357378422846,0.024497968275470572,
 0.7873595004059288,0.6060954111070395,0.8784572689577363,0.9818506142692159,
 0.9810004031453187,0.7374823934955799,0.5794726543459726,0.23582985625974284,
 0.5857347099646529,0.06422070793084123,0.7813895775390725,
 0.40189285443225387,0.5825803559374276,0.8843090736789061,0.6081320241986601,
 0.7429012032753813,0.27472018831113654,0.9264685262468587,0.8747617395830339,
 0.5868020311413469,0.2595064900287056,0.16465041145582082,0.5907714786575557,
 0.5085241230859557,0.5613515432895098,0.659316192926196,0.5329017360981083,
 0.48946513100570865,0.040118461672492445,0.5651276960094453,
 0.8735491092702428,0.8071795830614977,0.3064775466019235,0.8927467167345785,
 0.03237936281984377,0.20854718618449397,0.5832203398860835,
 0.8825203310816474,0.41968125679011925,0.9021050439846606,0.6181712651390192,
 0.6419096762988821,0.3721029182482518,0.10618888589067432,0.292492505156391,
 0.09179834895684569,0.9064638553567441,0.12605989986796262,
 0.6428992388334983,0.41297584877637805,0.6530271536880636,0.0806823050164246,
 0.5031907310329898,0.49769563853929466,0.06964546149975093,
 0.4099860743680688,0.7276771801814699,0.11595088832948597,0.2345886167636717,
 0.18865700796857798,0.6966504860991823,0.6582086508764857,0.2189803612457475,
 0.1250162421491181,0.16345155526794863,0.9546107195837995,0.7076701339352874,
 0.2881389127457228,0.6511235729064407,0.5133047239415566,0.9974093241652102,
 0.1384915721871578,0.240928484668677,0.3402681920442976,0.6061693606191421,
 0.22197705471627616,0.7716466795872605,0.12786996393286554,
 0.45785223152833787,0.8501162807139475,0.9663693857643771,0.598625365405542,
 0.338126523719929,0.9757834732182561,0.4666816424457979,0.5136965509098902,
 0.8928364014600074,0.35002891042895334,0.9331779434458025,0.6748928535675122,
 0.02531627743045395,0.3963883844245225,0.7019416332199603,0.6576378457534856,
 0.807756803460292,0.16727604374824723,0.4178534290153093,0.7602859044601497,
 0.17202649237602186,0.17064777639896123,0.4584023208557858,
 0.3688336149270528,0.9521575747093383,0.3526383423295135,0.30422030970537506,
 0.10922089254544043,0.6159638681341328,0.7724232048282307,0.8213543269016814,
 0.2599044707299336,0.45405114616386133,0.86609557625055,0.07470453637692875,
 0.5153951018021873,0.7160462391132683,0.3713886519492835,0.4811512874268741,
 0.45022594441278396,0.07378036480723671,0.4779277411708156,
 0.12195395407328735,0.29804366929633763,0.37465674147648353,
 0.44524148534869645,0.6438731720829347,0.0613188551181848,
 0.19615012217941918,0.5529654095625531,0.031834137728470435,
 0.29229101123736667,0.7400668779448998,0.4355924560798794,0.2927121312917721,
 0.5215235635368349,0.8979119438277426,0.8982676185899171,0.3110770185752252,
 0.7896095346248094,0.12438566928092376,0.11426906769808387,
 0.6880008066218656,0.3320802065132983,0.05220580656398521,0.1568754139605949,
 0.0798405926032808,0.01289643921596928,0.4887876338619773,0.3402330764807051,
 0.5414629076184411,0.31408513274515515,0.5426189075990155,
 0.20976428965109117,0.5899324638694715,0.8216366509735202,0.8430631042710989,
 0.16434526976647956,0.8013133664233127,0.7861179416571749,0.6578489263483058,
 0.442608484719771,7.434689360297231e-4,0.3013921659000601,
 0.22070051461606233,0.8002370305070371,0.2039009529780752,0.7290112563794954,
 0.09369940994023429,0.707601146040451,0.22499552937710932,0.1708054323359306,
 0.771735746614798,0.34009101625496263,0.4888957335553249,0.27448932272666915,
 0.35150688996009904,0.6831783807834828,0.8143619045467964,0.7244573283646084,
 0.4684318887934942,0.7711063607123319,0.27162998913703407,0.1840387990986797,
 0.6566709029783349,0.6543238213116357,0.9073851015226612,0.10422790514764668,
 0.774869666916966,0.6574987179086227,0.7615862093674494,0.43879869715939723,
 0.1894313593518363,0.7930741063621538,0.4046108232317116,0.4956824785444669,
 0.6590166918131157,0.652584177213275,0.9475611470920988,0.7744208019945409,
 0.35764901462413323,0.5769569531102949,0.6364877766335244,
 0.19836899506670425,0.4271800991077014,0.4929055729698735,0.4830066510626323,
 0.7137045782013821,0.4908905877887656,0.11144316718721725,0.7082754944448135,
 0.40538361995387584,0.6178595202301718,0.10138696621217391,
 0.6499438167953722,0.7509316539274026,0.49170440654825054,
 0.29276596227992957,0.7273028413741156,0.7923426909471847,
 0.37048475332439434,0.22942715600764574,0.36386245364898673,
 0.1692330577112343,0.44389064789303445,0.7486397960052579,
 0.38932541775001483,0.10655007218667789,0.21492595034655393,
 0.8779897695749734,0.8169758660066808,0.6411287147695313,0.9328869677358266,
 0.6704978403657378,0.7734726865703079,0.8691241525953648,0.5693799783046145,
 0.9594669961481691,0.22464087120921272,0.1444243451299927,0.5219154155639985,
 0.929857440471925,0.8052300221599302,0.8600005319551413,0.2130878541482275,
 0.7553765479533247,0.6217208997942612,0.8017958736019707,0.9060694011686977,
 0.28491215782439516,0.46115492694894467,0.40200765733177024,
 0.3156092267965498,0.7718881954506185,0.2325374624015203,0.6741947322823453,
 0.2853187904873955,0.686275828696718,0.9800575775954159,0.734718048534647,
 0.6576399000858889,0.8082996861369858,0.13368213763569936,
 0.26448697135635224,0.36577599092152735,0.2709923155616062,
 0.47797486422788904,0.8241491074771251,0.4805522503918487,0.3466603883558089,
 0.15342214067966564,0.05487779874843857,0.6063991680375904,
 0.8234167197294272,0.8988944278373814,0.3347647992884649,0.4671601272115796,
 0.7823642670155408,0.5466174035637021,0.522002656759814,0.10770204793994353,
 0.8877451376770907,0.819005493413815,0.47748170598920925,0.7936254778743591,
 0.41484101309774146,0.14679255043774975,0.7383533074938139,
 0.8785143036065556,0.8508723910764839,0.733358258671297,0.22954376305791757,
 0.7788799396309923,0.81080017542777,0.4606862906268261,0.730027078919737,
 0.7984306515323846,0.7509488866476155,0.19431062836037027,
 0.41511246011038627,0.31880126616450266,0.21710189206084074,
 0.6968166553779095,0.11003418515385333,0.9573331621644231,0.6122946271381888,
 0.463777963353475,0.9494103909786926,0.2516174350219712,0.40681680794959374,
 0.320839091580708,0.08450592530511891,0.9946373991267057,0.7624439445233012,
 0.17776506472344744,0.8844054324711206,0.07593455849130426,
 0.9918093853333891,0.647819388672506,0.6762472647943046,0.6817262877220615,
 0.8814243540010674,0.10122727682709609,0.955494494997037,0.8272264081663655,
 0.42619536362315313,0.8491510367612037,0.46521087808484574,
 0.23378647737984792,0.7093104958206562,0.5858873265439528,
 0.10174613615179984,0.06578398975685051,0.4110818577549642,
 0.1424038534932316,0.40957128824649947,0.805478020793727,0.6275880529213187,
 0.7996607785148027,0.8921398405954898,0.07442338995748632,
 0.21241875682088618,0.4772924048151935,0.6668361936901157,
 0.03636941276574379,0.03473185376013299,0.5115685301584985,
 0.48459349289335785,0.1773256034855586,0.963600818221506,0.4146259952679111,
 0.8632065622321593,0.808566322243682,0.5879319418852726,0.9593602547085317,
 0.6705378685748692,0.5891003474996267,0.16630631673303609,0.8613640862930332,
 0.18092612719620582,0.14460832759286013,0.9260731453312498,
 0.3681618037263288,0.7902437341575596,0.13077454375401443,
 0.49377406415869884,0.14581991734248334,0.9408356613656614,
 0.7967051951232931,0.235743627778219,0.7333881428677054,0.8400148215299214,
 0.6816515674660458,0.9541630761998553,0.5418382143705364,0.9920460408371014,
 0.2467483606714077,0.3910909493974667,0.4738043026731824,0.8454622008866354,
 0.49727826537595743,0.01761300516545483,0.3909447839773319,
 0.20542624045786273,0.8941398113167881,0.7609943525167615,0.8908007796116416,
 0.12605071272044066,0.8504147643765677,0.5558411940631538,
 0.42235820569271365,0.14346586937057582,0.7011663599457942,
 0.7815916781910525,0.09618373117679768,0.035877074125710284,
 0.12630867664280965,0.5350366156228262,0.9891151353925219,0.480640489722168,
 0.5572691618773539,0.8581197603583117,0.5406182645736026,0.40535440624338215,
 0.9552420749052393,0.19363752692870995,0.8596684415818907,0.3570174565635824,
 0.6690817048843695,0.6795066008988141,0.11156034836210449,0.5370962478038552,
 0.5524058795184581,0.9224410541473327,0.2604873065353732,0.835015362862998,
 0.9838090709948524,0.09805234119843731,0.2993003183882048,
 0.40589622546119575,0.09155521236428299,0.4835414843255704,
 0.8462806801852075,0.13380765183615506,0.7285581850739964,
 0.43738186432461146,0.6393953113282873,0.1601589461858377,0.7867371875264609,
 0.3970592812697642,0.7868739314613379,0.689840050762529,0.05190687207780165,
 0.7752889856483505,0.4050779668641312,0.12730715224185984,0.6029950097276681,
 0.6328222833615655,0.12189837686635485,0.8046789545017571,0.5707399603887656,
 0.1988127293999742,0.96195301553672,0.6893528275733476,0.9769700078389318,
 0.21704416410126004,0.883853247185017,0.22285016263769253,0.6358019521156897,
 0.9272380152353159,0.8617295975517663,0.7279935095896821,0.8320383658852795,
 0.5506702989208635,0.928995418287672,0.1407320471161687,0.1665979285814878,
 0.21829315351077327,0.09606925093495589,0.7281762345691585,
 0.35583739640295975,0.9448242710556234,0.8233105898255687,0.8040086270006253,
 0.6173643475382903,0.5978452293030315,0.08163062148386202,0.3508312820343167,
 0.7646255513768913,0.06707965776021885,0.10991164399131015,
 0.24045848231957656,0.548533328677794,0.011929549375351733,0.781714582854136,
 0.6462095768649057,0.7426518042070547,0.6459244795547028,0.8239066003362624,
 0.20247493569910135,0.5832985327706567,0.683863286985828,0.3162419801963108,
 0.6523410231900251,0.28849390223434224,0.2910519716810209,
 0.24056805506390488,0.9700925726973735,0.04817380485983447,
 0.09778869194997764,0.45469431536157057,0.5904884316527179,
 0.2439811378956631,0.8006540968165324,0.8286428480313818,0.971184197707641,
 0.22523324691764945,0.2483116138309246,0.2165778804273032,0.6355480285424645,
 0.6569961400903364,0.11316331264463475,0.5026587920331347,
 0.22639388327548238,0.9073464453826972,0.6761154790790371,0.3765478449873936,
 0.843326569745904,0.22539012127390767,0.892489876545517,0.7327587321227305,
 0.48144126139684285,0.5906172754725063,0.10468019711149246,
 0.37497167900255757,0.8353225755607953,0.565950838414271,0.14659459581082157,
 0.5102276522695588,0.8046737474178659,0.7111719263943489,0.08487392433869057,
 0.8078161675187612,0.6483779699975943,0.16596382849334257,
 0.39082645604532373,0.9495942054459068,0.977577574574673,0.23252423329513272,
 0.6600362649846686,0.8393365863378832,0.15885287200938236,0.5541944398556158,
 0.05373091559709975,0.8256968026637316,0.07748973939746406,
 0.5914772838606882,0.17820125531701847,0.13242411560044487,
 0.6841921362278556,0.8418135827534173,0.8183841641433497,0.8377769192070821,
 0.3598551326684172,0.3426864680400509,0.5330126768907113,
 0.012866452519669891,0.38650279392967823,0.04307022595656096,
 0.9705799044940012,0.9106622970083358,0.9845382078631271,0.38579470215911127,
 0.6537465454247658,0.5905461869737094,0.28172016468023686,
 0.19695489269878697,0.7011891738897447,0.8816900582626822,
 0.34559115237137394,0.7799345050156467,0.9415273835330628,0.5373947330065124,
 0.6659622514191261,0.653982792470799,0.7485852659160688,0.008993205344151933,
 0.05804044063486535,0.8981169822219186,0.9904599503785237,0.1514424158569897,
 0.9901736202455056,0.30214170179821953,0.9481520455971181,0.8410623427210684,
 0.746849749916667,0.08008532369320998,0.0716632758145328,0.7033693318799306,
 0.459256230135789,0.2900791985022223,0.45952216553266423,0.27806119035415655,
 0.6526140118573126,0.35739319550282433,0.21792113532902746,0.822939884266591,
 0.7755595368049177,0.6597596109271827,0.8433024517157394,0.9946251535829149,
 0.22777750927580775,0.671548637249396,0.24605075576768576,
 0.37095083189155753,0.44387120013358294,0.5242199400347511,
 0.14526460875890568,0.667007601674653,0.13322190092039776,0.830870231797079,
 0.04141095876758749,0.6627640246245645,0.5730764199565356,
 0.13729264365843308,0.9254643122121669,0.5053595627332584,0.2804928234451035,
 0.8597781096181707,0.7000097278551204,0.6001057902593042,0.6703702535319306,
 0.0669075750419098,0.6190161153257293,0.7775447916669418,0.06819079623208091,
 0.6839470494842599,0.3274673346876499,0.24600247938835784,
 0.44467280201388415,0.5668112669984764,0.0349611167686541,
 0.06602611810611736,0.9423001464088234,0.45212687151849684,
 0.8216319278591079,0.49679047221422423,0.927919659046294,0.31509193178660755,
 0.7662223212093061,0.3505921058982765,0.10288657369751764,0.7077027274049809,
 0.6065174297796677,0.7330640613655534,0.17984122841596561,
 0.27377620351942233,0.22338808980744862,0.5902064661645939,
 0.43716856115114566,0.29905596053795125,0.6482098455699687,
 0.3401074610442103,0.004823535492649911,0.5035737621827748,0.354196226359603,
 0.9414429680708075,0.07791493340199518,0.25511017451488593,
 0.9968276002451119,0.438782986556282,0.35611266620985005,0.6657384337413341,
 0.22607847234141898,0.26399227256517965,0.06805531789994101,
 0.7468949265226973,0.36920487389100165,0.4845507958183082,0.4690816903505056,
 0.4536959665875485,0.905637419064872,0.15028374409394518,0.9332878032375259,
 0.20039956543134008,0.7699032187898822,0.06401815792362586,
 0.2548745719715968,0.6284189597741797,0.9213693447880728,0.33569183088608423,
 0.33410121620260913,0.1830178212878364,0.6505127867323568,0.5088614467703132,
 0.008727746088267319,0.8842566818661542,0.6206324679870258,
 0.7990230320362056,0.7470723825711194,0.9473901011952781,0.945447322273781,
 0.9281039175369618,0.12569591734318641,0.38216832129482325,0.646267140539707,
 0.5031572437889857,0.9882214397979011,0.8602496193549546,0.8543582020840547,
 0.0646909669156992,0.3462652623846565,0.1272464258402629,0.27393047419907024,
 0.04002348871577155,0.9283144556784918,0.7848717783757406,0.8431847407651669,
 0.5571238046840172,0.02600264512657402,0.7648981292052064,
 0.36843739809175513,0.763428965249682,0.02992027735791636,0.9334707424398262,
 0.2106074971753874,0.9082816759787867,0.9080520401487654,0.8361036848878007,
 0.5143612120739819,0.7287671842274873,0.0025973610994937113,
 0.5419154004372736,0.8967104583873402,0.04649451830475082,
 0.37257643220360204,0.9288860266513211,0.0991464543355739,0.34489736291175,
 0.9431500350365648,0.34554145428912286,0.673421447911689,0.8297479716907472,
 0.10170852703001898,0.8908005048286115,0.9879776385983053,
 0.17830494223073823,0.7823581869891192,0.43728738304597947,
 0.6288305446883653,0.3678401774786946,0.6999851499707336,0.5113554212312323,
 0.05417631824573965,0.48193734735079685,0.5535138850344062,
 0.8262496470946352,0.4308427417015277,0.0416378045920055,0.9128785290393194,
 0.8348858615957299,0.33121657146126326,0.9998847285248608,
 0.18496734006014204,0.03929711437408634,0.4280171763636198,
 0.9879748481741473,0.28332692111282953,0.7793555501779429,0.7440825125936764,
 0.4296437935884473,0.0022658299061942833,0.3241931387615312,
 0.03677004449519894,0.880459764503623,0.18882633075048982,0.640871566269423,
 0.9287537354555652,0.21157246539657681,0.21391915573552212,
 0.5321048711592665,0.9627730321674429,0.6777674910665158,0.7082425296348007,
 0.9833014481238116,0.8010046049846796,0.655619107194547,0.4355758913543417,
 0.5871605934401672,0.7999988375323039,0.07190135538255826,0.8162391031057881,
 0.383351691307418,0.007843477351241135,0.8957172779429806,
 0.34651838809513225,0.40783149871300506,0.06084837388489173,
 0.8779195832682433,0.41789199065106575,0.9505385359190488,
 0.24263506657549316,0.8455338886585757,0.750331053131158,0.28260895090506943,
 0.5944606162201818,0.5070731914138152,0.8739009447935497,0.5740097003982827,
 0.3624161395018095,0.27508570236917307,0.11311419752112961,
 0.9169089212081111,0.9199932040150967,0.7188794456271064,0.9882842099870824,
 0.6593591138753142,0.9851854607925461,0.10967923549780512,0.7954376878514511,
 0.522257432501752,0.7683114270728686,0.4948293925953958,0.00534601304270077,
 0.5689685817798394,0.7310840908340261,0.8681583896875321,0.38836628809878404,
 0.8123015445462964,0.5540550683971127,0.2518848123062801,0.18459449350340817,
 0.20185689107843507,0.18823869392014825,0.08139466521097316,
 0.5081360172419287,0.7438781844026586,0.7661995979676838,0.02631997806725228,
 0.010229055680899601,0.7478184403858789,0.4725115926407588,
 0.6125082546738412,0.38614730392040764,0.9012177674795911,0.384803000415353,
 0.175804332631309,0.8492919455300333,0.2400100270996507,0.3057283626807663,
 0.2923547463701127,0.22154019943792314,0.5907162836150006,
 0.05718888694229696,0.7548185185667993,0.4369326151044699,0.9870865972098759,
 0.8990362865869845,0.27475153110949946,0.9629392722262231,0.5724676922483346,
 0.14295447608214307,0.259467213664071,0.956897054509756,0.9999558208108207,
 0.9961325148506504,0.45726044434772073,0.40814877712043174,
 0.04964225426667501,0.41684835647373597,0.002344545104543627,
 0.5086592896238546,0.8058494729462022,0.641102160610007,0.3762251020685764,
 0.5112532125504554,0.8658362751857638,0.21213233262278275,0.756104394657713,
 0.8901281458752327,0.3811008461269393,0.8645266524217612,0.30772966405613045,
 0.6384844187689462,0.40148732358087325,0.6432383669902321,
 0.30317918960370394,0.9302143391171014,0.9501348100467321,0.9769798413974554,
 0.32286655373460116,0.8272539625066524,0.7776680518247128,0.9408612858520179,
 0.9804418092036348,0.5000954546439577,0.43717605264777126,0.3120957496482396,
 0.039629839985297455,0.3838376419393197,0.7104539866409764,0.943564482772071,
 0.16852963444398572,0.6893655936211306,0.8675697598234466,0.6929840607019133,
 0.7632941608889701,0.6157077785412566,0.48738731976792016,0.6966616826866427,
 0.602080260253386,0.092242147720627,0.6995426327022392,0.582621787416298,
 0.6030183172163859,0.34233687053990847,0.6938107124532705,
 0.46468718286362576,0.40786629136577546,0.9652330566966022,
 0.13670903238944976,0.3759947755267362,0.09708949834618763,
 0.6088828551106403,0.19286520454765732,0.1692445411960204,0.7770833341637491,
 0.31233343471135866,0.8480491576918292,0.15689899043642164,
 0.45050484878857966,0.5614078369348421,0.535027712742933,0.7041842901574377,
 0.163914021888674,0.21564150638965662,0.5640514333645107,0.9308574838124453,
 0.8503256162035197,0.416421865336583,0.4061616987019091,0.005367779095495018,
 0.022317044013163256,0.0025751600098317118,0.3958041828029931,
 0.47458608625118426,0.8525054424567812,0.7870806035309512,
 0.34784564500920334,0.28107939911799373,0.8324998109353827,
 0.9342835945157155,0.9163398259000817,0.24517447525750935,0.6389834953822715,
 0.5415807109934709,0.33154222407916756,0.31850327305424475,
 0.6699376731591231,0.299484110156212,0.9127782258228685,0.8126793066087844,
 0.8583778422160443,0.06654646132402031,0.2201789258432697,
 0.48981267576923604,0.865796661526973,0.995337297603668,0.42937479347533847,
 0.5359783939541479,0.11569575431589285,0.9617371549441923,0.8186497452879924,
 0.6387930150949491,0.892688212055811,0.3320663129349619,0.49753263004099146,
 0.7794412861157927,0.4556309333920733,0.7837195980921551,0.6586484749843616,
 0.6165855690042961,0.8892803818463099,0.9954627231759929,0.7101706159910164,
 0.2807471421013632,0.13531282319299742,0.9815279541930294,0.3120893841752015,
 0.8423234300955909,0.7780622911946061,0.17559365625936418,0.1683142051399925,
 0.8464524237511039,0.09603758640017113,0.5130708792738429,0.8296918635287599,
 0.9079930074859544,0.5179780069866688,0.40349692101384427,0.356820651646103,
 0.27305270658927194,0.5422444967718295,0.8161747578391161,
 0.42575202023984093,0.6201589222075371,0.06426091473443618,
 0.8154280335894643,0.8319508130226557,0.0641808014250741,0.9316248033809347,
 0.7799203504250728,0.7038380968042952,0.7078628341851793,0.5523371074660125,
 0.7827497429677702,0.1178110724314353,0.7291666958198414,0.49338823612557003,
 0.07095457893796797,0.10921154598600713,0.9392177820531054,
 0.0019579024731217753,0.29660892372082714,0.7115035896068614,
 0.8100659539597809,0.36029325997823813,0.5184340546546633,0.5752082223887389,
 0.5647658791293594,0.8114756475779099,0.7795406743428884,0.48429709998643844,
 0.49280318809639256,0.4972448904794303,0.18793308519707153,
 0.13799088309136387,0.3953783583528656,0.010384173952474818,
 0.49644180887822065,0.38684483451028673,0.8217692231920024,
 0.8793624793766417,0.9976627161916943,0.6036536989671564,0.7520526326569132,
 0.8952331614904226,0.9292293674642569,0.49987379331833315,0.3069262095689652,
 0.914200627955399,0.8564760176061949,0.12447522886454099,0.8640073206178497,
 0.6360473621367817,0.347663132488381,0.4221959441904757,0.3562691739229821,
 0.2649709731790125,0.1526905199642229,0.265338567971266,0.577804712175188,
 0.39042142342758634,0.10331108152791524,0.3254475415864089,
 0.4095786763423086,0.45341512623284297,0.4390321643910786,0.7031833623992618,
 0.021293009655673267,0.5929287632591889,0.20891168952226657,
 0.49223139778901115,0.11636178546064702,0.09679429500419223,
 0.6754693610936939,0.5511874106756482,0.3157621061498833,0.7016588044415443,
 0.08664264125819832,0.02797203482877242,0.9557860769887758,
 0.055779581225911534,0.5517690349187159,0.8080651033738141,
 0.5871952593378746,0.9459890852507753,0.920420898960802,0.67978223703986,
 0.3913816577293159,0.05621007085649132,0.10195838804435309,
 0.9773336404249833,0.2117210721612115,0.5155106697748402,0.9572998193994164,
 0.892833535417271,0.4421392325791341,0.10961709672676156,0.6435108902975437,
 0.7880938808318184,0.5095259171082367,0.6079089171304826,0.1716201016930069,
 0.6076391247469255,0.10253582213377443,0.40990558672488164,0.962747085778601,
 0.5073510236944422,0.9672344495468709,0.7481040112831312,0.8097463968441458,
 0.8861070008599476,0.5980461991882923,0.17305390065475568,0.2635750385070811,
 0.6766709452225782,0.6294334492611919,0.020110561832249685,
 0.8745387654784305,0.2733670078118454,0.10627833630030958,0.414137540937889,
 0.3910491186178733,0.7659288010526749,0.7070771175885986,0.7297086634758098,
 0.9705415462589501,0.7611598462693382,0.8382539130676573,0.22005265858090073,
 0.5313779049887495,0.2523543203069215,0.21949151022408175,0.9474019786334659,
 0.2488994078518615,0.8584691130222384,0.4486343869749214,0.7078228426910232,
 0.3907698751396459,0.4588343894462952,0.497730310124853,0.4926772654232452,
 0.2655364952811198,0.34182381700790454,0.5614236113367327,
 0.06582218176558141,0.4054543219631863,0.39291821896776347,
 0.7974590387207737,0.8695667689905552,0.7759469866611328,0.5876630270541933,
 0.2439377188147126,0.009595002929236984,0.9370791719033268,
 0.8311451691084006,0.9663178475492107,0.040012033110665035,
 0.8003796565953331,0.8375656490460962,0.8111631007890688,0.9285781058728465,
 0.6603383330128014,0.6783586817895764,0.9356110381761563,0.4309428848274641,
 0.5632522451965345,0.08059296196965626,0.588622527314045,0.3235309808241851,
 0.15385569843509117,0.54287093346647,0.628443358628944,0.7387083676053201,
 0.8318257088963712,0.3657432931070719,0.7858946679118272,0.42044364964150494,
 0.6823921252634265,0.2655802740943737,0.09678235423926873,
 0.060514235480223943,0.611640039396368,0.07074288030762721,
 0.8852251147164863,0.05529777752105802,0.915672422918137,0.9343275333795826,
 0.6941852842885451,0.7101064497067183,0.3967123958207214,0.27449189233202276,
 0.5017018033934388,0.32137173805071473,0.9388561920370145,0.3243830283212993,
 0.6589383173680669,0.3055058889065446,0.24791991331821617,0.6799422203802024,
 0.6864896847010518,0.02321968357923121,0.9658876579111956,0.6882577696064507,
 0.32497507373442025,0.8023914001182793,0.5173982518966469,0.7458959452595311,
 0.4146405241165433,0.8080922760417006,0.7124804984696412,0.1626244365647863,
 0.760173093796999,0.48923426051393626,0.08434684193931397,0.8693928780331271,
 0.0434383228471944,0.14027786226045624,0.3250350637424194,0.3349104651501891,
 0.7316510905039291,0.22358825614522104,0.9645725091660253,0.7713059023327199,
 0.48480570512926835,0.06580565983020381,0.29344576507508935,
 0.31880842209250515,0.5906161448780306,0.4376191678571558,0.5696694152218932,
 0.7205409209039069,0.38327635024687645,0.9956615917078788,0.7764799460150218,
 0.10505709955027553,0.13976400591021476,0.35400468800189944,
 0.13983229963398547,0.8247238169282866,0.16662591593882692,
 0.9987145503618599,0.12323185648936841,0.22880993924868764,
 0.9024531989792779,0.3908125336362016,0.3978117570466726,0.6512095899855376,
 0.19349541596996334,0.15824307452505237,0.32553421218246004,
 0.5390325978710413,0.0365576560797678,0.5984131549561494,0.10458577524607793,
 0.9002862400538276,0.012595306174574605,0.7489371313447196,
 0.39270697912512753,0.8062624930464578,0.4123301949701289,0.5756779379451225,
 0.32563291774349745,0.6117702054651802,0.4101823704243497,0.977399281733791,
 0.19010650270232332,0.8570180106671819,0.8618456993678134,
 0.23799202688728482,0.7641964737705664,0.9516483210816455,0.600234891159282,
 0.020962844999704355,0.4410412010788235,0.12227406404270025,
 0.9182613279907302,0.4322939604362154,0.4474657003000958,0.25566587985471734,
 0.037631620722074155,0.097116105300591,0.06921836792760572,
 0.9321922069884806,0.9296275683783157,0.6929640770032832,0.8028528994450732,
 0.7381494682769592,0.2808220797709604,0.45412714323759296,0.3691001308338253,
 0.44850738324441586,0.6514397546752869,0.015376976288069333,
 0.7354534837007503,0.47334716613123107,0.5897009087434079,0.985464505128235,
 0.2711574127909143,0.702750833013343,0.11893576468159328,0.7586370527326406,
 0.8024753269331111,0.2087888530783113,0.5317701190400012,0.09494128511391853,
 0.053142490677239085,0.5977188864318123,0.5010598158719715,
 0.2425579735694392,0.3560340234188766,0.2508313017350501,0.5575307637598577,
 0.650861125468832,0.5719097775953788,0.4349188047140291,0.04981994441379822,
 0.9821329632302094,0.029181928397748247,0.5936454451588968,
 0.08190551328324869,0.7205157695328074,0.555575165979198,0.11432936405858807,
 0.9846896388098386,0.8053685912345034,0.6724502805649631,0.6395169707339823,
 0.17665622291891125,0.5491172983579808,0.6834408258156919,0.6668408683168306,
 0.23874696773844728,0.28301378248414677,0.1513522757754885,
 0.12663360462526074,0.1925940217240547,0.6231411038141136,0.9938867256879163,
 0.9424397923705345,0.9099845746771882,0.8240242830709241,0.0809746700728805,
 0.2510221770018035,0.4677307927257416,0.354154938725032,0.9176708110015757,
 0.3316754430497141,0.6486823377914386,0.6615358250492405,0.37028337040207815,
 0.5869464851859924,0.7430784649656035,0.8107574958943509,0.6416914990270615,
 0.35734082150765833,0.6934458961962878,0.07668888794386397,
 0.012377911816649156,0.5056864237852694,0.002234697505161254,
 0.9602599004253031,0.8502849590714174,0.9287634708430995,0.4469488753669906,
 0.5568799129242354,0.0777858648516514,0.5552328131069196,0.3256288380901625,
 0.18252577606771703,0.9805529079458221,0.5661100332933794,0.920362556784349,
 0.1875330505466195,0.15199036963319934,0.3710196748568584,0.9189285466473429,
 0.6958397304545403,0.36366898164349903,0.057544010809487345,
 0.4902628521605329,0.6576192798368075,0.3947313641522152,0.05137457966804704,
 0.953923650116722,0.17819879292766672,0.10400541569094823,0.2770262871015845,
 0.9859748761851678,0.711343458316213,0.9764821452952818,0.18059358189103114,
 0.7925521813284923,0.611015541078648,0.07520514088432495,0.530187317058577,
 0.8635249016854339,0.5167624975752211,0.9341690467177214,0.9718341048617676,
 0.7038280037324601,0.341712740466653,0.5459839180040041,0.04226749697204357,
 0.35043926397665526,0.7642736608556152,0.42484873261745915,
 0.1849902315454629,0.9322310425899438,0.9654778897446185,0.32631649084858194,
 0.3583043183018022,0.38143394461807956,0.22585032443582642,
 0.4647805448796586,0.9696393615193346,0.6006774986754168,0.4134389983954976,
 0.25883972404633093,0.7042318253372581,0.9514914535908117,0.391082809677741,
 0.8905608847751997,0.04760250618056838,0.22605117803164065,
 0.4035694017048913,0.14384602640745225,0.8013210154210696,0.9331630612580912,
 0.407875893255374,0.6990649749847491,0.9534417136998088,0.9932244657620294,
 0.9009756893657356,0.6075803872296172,0.04706542499816058,0.9786844681275033,
 0.2535066105132695,0.02681943411182719,0.7885928965427822,0.6862372418194089,
 0.9118550982654025,0.7751447832171879,0.7945680700398793,0.39010642014437,
 0.3944933844468995,0.458896014145318,0.03532717804839991,0.193148673594143,
 0.20131865701144802,0.19668644246624079,0.8656612328273184,
 0.7063485129791269,0.7994747373953546,0.20543058425470362,0.8095364309742485,
 0.580903560951356,0.09529911842782357,0.8398653568083488,0.31553968335690485,
 0.18592867407270086,0.1976958619222149,0.19650145787986828,
 0.4704410039486586,0.35317482156910385,0.25128198052224693,
 0.5001270862351194,0.04395462707479536,0.38324948688370253,
 0.3761070195281704,0.8482701017309074,0.8430174564142934,0.03667161910734951,
 0.4269401244743437,0.9333019269755712,0.6153746059791098,0.8631279370915712,
 0.705389582664816,0.6949228317941949,0.8499999974091601,0.03189650023447488,
 0.4722553452564857,0.1928071003249856,0.9481671389735288,0.4567123884935176,
 0.027734870742648465,0.2997907951182023,0.8816065029851801,
 0.7270931822197665,0.20521374957330663,0.7429532944828363,0.7991696045415693,
 0.8762724279829615,0.17061198687375256,0.765117847025723,0.6799877230119581,
 0.958208437587507,0.0850605564887551,0.5330518243753792,0.7910790482919823,
 0.8718938148477186,0.7388105390507732,0.9355260681287431,0.7832559862931509,
 0.3403315685067556,0.4739042634226923,0.6310036822890358,0.4060947980019157,
 0.8087279957049494,0.6292539548208634,0.07418583760462005,0.378253510304017,
 0.31566168179195375,0.4313894334730344,0.19574178983664792,
 0.2369668945573371,0.6396719204361923,0.13219691831298175,
 0.058502054457231334,0.40433084189976476,0.6087361124639854,
 0.5570639200085143,0.1368097166307135,0.46903102982721734,
 0.38170072629133056,0.7688717962727135,0.0181289537666679,0.8640883825885206,
 0.17111702032006992,0.07660864018292468,0.37572789921966687,
 0.6071872420076381,0.4261747097236258,0.7883993486969354,0.5015485192068996,
 0.47663662036700505,0.2675457490262857,0.5837986202234134,0.5278762259559354,
 0.5507670208874547,0.44464202951258236,0.4887484729117586,0.745735994221272,
 0.22674630198968793,0.396622865417697,0.420871530514182,0.8431389272494172,
 0.4596557105631598,0.17198953612804124,0.21486868151212757,
 0.7191696619070495,0.4522847427881036,0.5117211112773564,0.8405851132327776,
 0.8857957895013794,0.9610894212681058,0.6903932142942671,0.07278018085276672,
 0.34847302923428236,0.6865656611695865,0.3190016012878185,0.9036894254944575,
 0.9820564785793305,0.08508193845663636,0.541999227850775,0.5497908315412087,
 0.8481210802047048,0.8565109300827463,0.5035629882593786,0.7868836651844063,
 0.4525899760323904,0.7191847856239614,0.17952304492004934,
 0.28496462593984473,0.6801775012776341,0.05054965414757673,
 0.9661847832781804,0.6241380935504585,0.9753613918472417,0.4607670145004257,
 0.2869084207252893,0.07256721407828826,0.3258097407270223,0.7654337462901271,
 0.5570458984073271,0.6004923014812713,0.7169002036347482,0.5730300064300685,
 0.08069821702510216,0.6579655236718247,0.8743490759270591,0.8052048265076481,
 0.23066815935466334,0.051763011722544605,0.32094699167820984,
 0.9093289318125921,0.6529873342328858,0.5161944940348094,
 0.022781123015826354,0.49171575355142805,0.8256312526000855,
 0.22351555833482328,0.9333122249854844,0.9468809527419455,
 0.09550157190315323,0.35531930105375686,0.5965419106044757,
 0.15381171399696425,0.5248840775389706,0.5713172136787766,0.2663907130094473,
 0.4422623221301488,0.4375546133379402,0.751577498689018,0.9465522656039069,
 0.22954703231452112,0.3062058943450341,0.45136320249897244,
 0.9490037096457867,0.38745108873173373,0.36151006654313866,
 0.9482958386849786,0.12382600516594056,0.6522465309033669,
 0.42862136094774883,0.8651514867330303,0.1605023403591732,0.8523552358165026,
 0.7971107746825309,0.644226728012118,0.0908183701310894,0.49934389143359814,
 0.28112827316292455,0.6950785399579607,0.5839924080568867,
 0.46303173395290553,0.9043750363552191,0.205787201965003,0.9549549777867047,
 0.438398186027256,0.36358362737467564,0.9395763907737796,0.4743734575531212,
 0.11142291210792599,0.600212258281445,0.20122688713666603,0.3936820863595871,
 0.8741240967793803,0.794165033614689,0.6291766502256989,0.04496338465081107,
 0.571726898310847,0.5268301471568149,0.6234162141042735,0.648953413047954,
 0.42088559605097997,0.282227566894504,0.8665795061195545,0.6554977589681804,
 0.7307939557022626,0.35571268299510916,0.6796748828084743,0.5713559241056844,
 0.3327309187074867,0.546432086953467,0.5196071023363007,0.08006625010250978,
 0.7932100363018534,0.8243160812474208,0.3575971348754302,
 0.016653106635158177,0.25093353761098536,0.9323797975758636,
 0.5121318781845137,0.27549954794620546,0.6780613471581123,0.8814052578801692,
 0.42323578514951465,0.7813889070073053,0.7779920144433533,
 0.26214212121450475,0.35640199485103785,0.6515482304026661,
 0.8325361037584323,0.8128587140169325,0.22715540999112338,0.4868737743478281,
 0.27044839688666666,0.7427653445739734,0.9614836764790988,
 0.34680243549245815,0.4876616309111649,0.4982281409458674,0.8188307681057131,
 0.6471367272011425,0.6222312355485211,0.9881779746956596,0.05117378961371388,
 0.06193424294595418,0.28143835827560293,0.07237697670521936,
 0.738242006336819,0.4847170960955397,0.8209509869517355,0.644428978139189,
 0.8581000236988497,0.5865970730109931,0.10821314228521095,0.2519804129748512,
 0.5353113440576228,0.9585838561615941,0.7733361663682508,0.7811102565531256,
 0.9054009378808837,0.7778072979497841,0.6903656244067777,0.9257995781591668,
 0.49873645644221065,0.7089052978590974,0.12665786601612494,0.64840424063909,
 0.6743047362911951,0.09972204472564805,0.4000818660881037,
 0.007040892639167851,0.821134002520599,0.7185696380784299,0.5742851294610999,
 0.45784147385504526,0.6335451865103674,0.5444234337333804,0.2810275313348457,
 0.6001130544871129,0.40449786165125845,0.7256043166237475,
 0.49775261924304126,0.1697782338669075,0.504236383940865,0.9617190718702804,
 0.6620515435548749,0.8097872115327185,0.8761724189103732,0.6950801499095507,
 0.8137961659424566,0.1325136078144944,0.8379690281167234,0.484899929404803,
 0.9311397460311566,0.34426608675066994,0.3127182416104727,0.7957505736458835,
 0.9352340618251892,0.11509512587715465,0.1873136393513457,0.3263005991912974,
 0.37031088565062886,0.5531705174887224,0.6035812354449298,0.4372746967270479,
 0.3245824314664958,0.6704289314457855,0.7649985240963528,0.270985803814791,
 0.20321797768896732,0.1851595570939666,0.8395123993492256,0.7051375851380464,
 0.5199272966068188,0.44006848470642623,0.8248172346955854,0.7492007435015979,
 0.08919040652230259,0.5432264441572316,0.9247941020727052,
 0.05532628712326426,0.09947423240447373,0.6138421278635229,
 0.30230182888591706,0.6541193053441197,0.7961177364114829,0.4145260495260812,
 0.6815982685624067,0.5261741794506585,0.951105888793533,0.6267060130430856,
 0.7910086986253942,0.1780273109749535,0.46107078642233146,
 0.030666176699436276,0.8187172154131392,0.9135781417699,0.041876571650540706,
 0.5892304180334782,0.9632437048036606,0.23102979306263005,
 0.39521473831304466,0.5958680454830596,0.8777131053158833,
 0.23509339414999386,0.4371285609858222,0.3068800614056292,
 0.023403944508061825,0.7433230897845827,0.3309235859090932,
 0.4986274844722036,0.23488725403391808,0.5383322007277717,0.8636574434237823,
 0.44467106519902444,0.9347116392215615,0.11360798382028892,
 0.9308260521336287,0.7104830597054058,0.27150212027407883,0.4553026278236936,
 0.6888068030277239,0.8552545617254466,0.24959485992826502,
 0.33987731808250377,0.4028544017943031,0.061538513272277995,
 0.5736245645670188,0.25008405103266484,0.30441488546714024,
 0.7568144224422594,0.2072530565538132,0.572580906576514,0.49271835615271153,
 0.9954224078320238,0.7923968569536682,0.5547823476112124,0.7843828168233018,
 0.5214532935047084,0.2552287735156389,0.9079470055161939,0.30013763810636007,
 0.041607671010062686,0.2491661447319512,0.3063999722442382,
 0.8822311035259932,0.35896615532828646,0.7405566861792264,0.7590849849979474,
 0.6100841103023844,0.22142934519834434,0.05239060760024983,
 0.7660329001456111,0.7949851277506188,0.13506813208828994,0.3738291561431404,
 0.4708757550948609,0.27810634250257804,0.8925843237886629,0.5138886781939831,
 0.6817444251007032,0.061815389440554314,0.9723251694864777,
 0.38670760379430913,0.6999263897430021,0.7185266041843857,0.7972357325238805,
 0.7376710594521875,0.6816155054219564,0.615977828670228,0.08599970012412206,
 0.3299083673701684,0.332069031653382,0.6249761965446443,0.05579290384575697,
 0.4158581727545948,0.9023852339191173,0.02442654197458838,0.7525527312453191,
 0.1925966062903386,0.13071938970159414,0.0043522021384792,0.2636213013435418,
 0.2999377568021995,0.10018892461587325,0.724200604463292,0.5222406396479103,
 0.7124663589737414,0.09392966569594163,0.8750833201302839,
 0.37085596454872816,0.2722827062318176,0.9234434242929426,0.3221605806086747,
 0.7461673172234792,0.8152672522091463,0.7330754961993818,0.5252410821192457,
 0.5574880851271072,0.05679337462401157,0.1268583782542957,0.4430980118782677,
 0.5636952900880436,0.46186451973619824,0.04594367404034028,
 0.5500951062461685,0.9842600679122557,0.007810481926994184,
 0.5077516027900225,0.4708201451788533,0.6244894254702246,0.4186221992491418,
 0.8202933006471033,0.7363645761653175,0.9672306100557343,0.48524367631978094,
 0.4303063169777366,0.3392933131739553,0.9570481020742532,0.7078701243665919,
 0.9720412110560913,0.9220821229716948,0.13130148080184523,
 0.16727306935662678,0.3609503448351392,0.9384973884594545,0.8615562993727639,
 0.8891065541546718,0.610141994143787,0.9037575788245752,0.5306886672383418,
 0.4741464476405194,0.12666530021055888,0.7813023161435568,0.1506008209697407,
 0.48018878592914227,0.5095974636900433,0.2091247198713332,0.8184612532670079,
 0.5559065144032365,0.045552348094695105,0.9967835711740648,
 0.4147018322669461,0.3476254135359571,0.23359601137249697,0.9860068652404933,
 0.7816431299441133,0.4412931114658448,0.4169446952239151,0.47575751411617384,
 0.9580624245009629,0.5573596431414257,0.7537591894581949,0.8116620591095814,
 0.7103577333371927,0.5787541262205714,0.6889114754128043,0.3827345095858634,
 0.6111374864063406,0.3976053838371504,0.1463639120459782,0.16466025709668708,
 0.9976759516890359,0.943081778042292,0.9485258799411326,0.5081220646678951,
 0.015988860931222315,0.25221923769078136,0.36525443241937283,
 0.831337358857825,0.6195182604641579,0.7637195729551118,0.7628060640989182,
 0.443603895239838,0.5742469908197069,0.4536084942119375,0.7817637983495372,
 0.8141678267033959,0.9838135534668977,0.3618259995734533,0.6323389008146512,
 0.6635264962656378,0.8469960063013181,0.9735624669260561,0.8189650097042785,
 0.6366487665419517,0.06284493717802608,0.842665582131764,0.1331607583345601,
 0.042224089482585936,0.05223566423568715,0.12299800501288716,
 0.5641963196049788,0.4135888209360372,0.22841534969841393,
 0.08913562989219481,0.6974941087250732,0.5477587823352612,
 0.021777896328750823,0.4539764396225583,0.5334162083180918,
 0.019509908157096634,0.44783678084468415,0.5729174691670391,
 0.3266859363510731,0.017160941405710006,0.6552684396651222,
 0.5425673060889712,0.45446737589966635,0.6958775186307796,0.6649996865931579,
 0.824063540379089,0.3312713821333302,0.5022606118809174,0.22820273237735478,
 0.5539269263296304,0.3395539228392145,0.5919177508583359,0.9985931717953874,
 0.634984645732439,0.48952671196035824,0.46518279533046514,
 0.11927336693825386,0.791518496351293,0.39373209868365766,0.5033535573687627,
 0.15438061703477424,0.2675927385847585,0.10148663356899168,
 0.7270367293906908,0.5125719708153204,0.2094925899112765,0.8222774733004891,
 0.24707639581268837,0.25904386451464856,0.9330242912995549,
 0.3508925325025033,0.6654840628129682,0.33956702448081577,
 0.19427226964542088,0.12403021183377128,0.8415757574251823,
 0.9782312449687626,0.39678644722800405,0.20591649699162584,0.784656685628426,
 0.14210893052724038,0.5501405438968086,0.6068470610766622,0.3237058320482573,
 0.4249817738304672,0.9876820628764185,0.5670100840360558,0.40138075404707785,
 0.6487704309593801,0.7423479196528446,0.014176876764023572,
 0.054566599819854456,0.02734935930544835,0.9960908940964327,
 0.47108077347114197,0.09615529528742106,0.6345850755567598,
 0.07632704989354433,0.20855484871513374,0.3587689929810388,0.62949166554485,
 0.38898872727597056,0.7661325924636863,0.9146002127307419,0.9675895690399852,
 0.2809887548551695,0.07105808970757566,0.14579577234949825,
 0.6811087647507124,0.85266106505487,0.04099886302740574,0.8730475095145319,
 0.7311241338825331,0.8558597182562309,0.4563035477270765,0.4203594752255878,
 0.30929963646588465,0.8316246752291981,0.8749349243688096,0.4256192337849046,
 0.4972179971229427,0.7085501758752019,0.19353767881496808,0.8133290235662964,
 0.8718732214124589,0.3226854629479421,0.32706875183645046,0.1063302044365807,
 0.27338371565124553,0.4688092812166219,0.6599761492538566,0.8282575271765925,
 0.9572593449071569,0.3022210389313126,0.0877982772593574,0.9665960475075469,
 0.17283715949202394,0.5552746918433463,0.27282942718049197,
 0.42226658974893005,0.9429737362006041,0.8456687296814157,0.3742892904664771,
 0.932006644049777,0.29836447225197615,0.8892607696037478,0.8846980126711227,
 0.9345923963920493,0.2625983190130793,0.6496946238648424,0.14090334434664475,
 0.1503763417612749,0.45839101318944486,0.4919324261568677,
 0.15341738186987985,0.04759036876703493,0.12645572139097194,
 0.3204808966134277,0.6332944758057483,0.9506915043202391,0.8400071693389999,
 0.8651396361815458,0.9654458944945452,0.23410064281919263,0.8966849909299934,
 0.13612731896454688,0.574576839837348,0.40726360519978233,0.6116488911136831,
 0.8980095532226853,0.4608527742632902,0.4571253511995833,0.7490071916676543,
 0.37185872783063534,0.9657795296979073,0.36387049576966046,
 0.5715222316823321,0.266623494173476,0.8000289593781804,0.38262527829533544,
 0.22461317197605235,0.37855268832040156,0.6552096725784038,
 0.37885829179915975,0.10368878438977713,0.9456750156845035,
 0.13242558552377237,0.5523049339393874,0.6577124498893588,
 0.15143050575307138,0.1020358915475984,0.9004618730473459,0.174475199891566,
 0.9413702522692247,0.3777847743258129,0.6018376733092774,0.45226994756832917,
 0.46314993494026635,0.1045870132206792,0.2857592308864403,
 0.11247702463331344,0.3168776869276544,0.9944783857375046,
 0.33939168537469655,0.7427097600362168,0.4717458035355584,
 0.29319530876859135,0.25237250644400555,0.7346603367849617,
 0.39925134075583313,0.5019818087342606,0.6077723463555519,0.6926251033363097,
 0.43031449109664643,0.4248057864350576,0.12055263127921656,
 0.9146213210637953,0.7625096991849857,0.5792255635079437,0.4380162633688709,
 0.31803130390092393,0.11180115760692533,0.27518235153336623,
 0.2726826009349801,0.9852004270635075,0.10965796971748276,
 0.02606467310107885,0.7378331085615855,0.82537493721217,0.9662720280960562,
 0.9490555908441594,0.42093713732585225,0.10671966695952562,0.987077572076397,
 0.8395038142130586,0.5698718464904544,0.001872683412319609,
 0.9465529771435187,0.030361960099954066,0.41362928217851314,
 0.09781571492157637,0.2994788560208903,0.9838571426359387,0.1607078387269184,
 0.12189739350987439,0.33371255157019053,0.8484668566413921,
 0.8527210515643479,0.5239504350380837,0.4899943572446904,0.12385088027596625,
 0.38338236193799213,0.2971751610804334,0.48024087494761747,
 0.2718023103405739,0.1990124208363817,0.8990482860038038,0.29434710551751986,
 0.17915679321153777,0.8915705134057629,0.5573207558949393,0.8604236500667978,
 0.5561986465279235,0.5888141622635565,0.32419094381499813,
 0.09078344589772769,0.6480936455158252,0.5085967355053219,0.8768783079898913,
 0.5527252984393768,3.7355058777011063e-4,0.23835915718559186,
 0.3054025326983024,0.45542931930096686,0.6385162892050011,0.2812650462336227,
 0.11652857292225216,0.7025922144626634,0.741218050877137,0.936769077063347,
 0.6680570172868123,0.5868385988735747,0.5974621519137012,0.8570591862572664,
 0.8453554993373026,0.6604757748001538,0.25220920647519174,0.5834479446008103,
 0.8768572419309708,0.6429211366053619,0.6665388145905157,0.5912959410579967,
 0.2827052652656714,0.834371969757355,0.4761127519421757,0.916769593678289,
 0.892085410445689,0.8000430392019936,0.16656338806122084,0.9049886316856771,
 0.8194277544727996,0.9581915803800699,0.594831667805439,0.21133517921456713,
 0.33533030248353124,0.4458050799975044,0.06315492383331245,
 0.7608183198776992,0.489825372201083,0.7328334950051554,0.6928411053837542,
 0.03796146274880252,0.8671272926831889,0.818059857511332,0.02327146219085008,
 0.7631020387806262,0.354286936635665,0.7465999543585463,0.16708120675002325,
 0.6540413003306953,0.681897441941131,0.9095770785914409,0.461946340541845,
 0.029086264855041843,0.45975069881256725,0.6899375678244148,
 0.14616064731071265,0.3729800219174644,0.3703174990992456,0.9263505752418245,
 0.5727052210645495,0.9933204983680191,0.24481616017969077,
 0.28245670980095083,0.33122116937202806,0.35981574190288024,
 0.2853774305423925,0.8000322649801997,0.3839756179641015,0.2821306403180316,
 0.26963681060578026,0.044196623404412105,0.37287256116673906,
 0.2755968358918589,0.8490019349369584,0.7511223038738153,0.8928794347081906,
 0.7719961050076191,0.5937632296012849,0.5610362941429654,0.0597231763445375,
 0.31671423193395654,0.4868360101749065,0.8478142361902443,0.8191199044465778,
 0.9274130840929025,0.1385803346377551,0.48999597419568663,0.407626063070203,
 0.33373307597284674,0.33017388552047433,0.2858804051633501,
 0.5442132295136076,0.6748923317316179,0.43885195820665535,0.7636910884300727,
 0.09492607980753143,0.6817333398454958,0.3254035178794564,0.4727477173199308,
 0.6137872850814845,0.2441734830765352,0.34401531041852684,0.9524135060857232,
 0.8145406483363644,0.27490037371359743,0.2730264825892036,
 0.14394209120514412,0.436214651110984,0.089681856750085,0.43173340142247485,
 0.7377750358096699,0.13375593761451365,0.3604623055231908,
 0.07379474379268691,0.5966225251747719,0.33420694016515995,
 0.3435656102628406,0.07823213292686204,0.8421773276858202,0.7715519000525483,
 0.8212077205396782,0.36943165184487514,0.6398307133703687,0.2284370281877296,
 0.6878300000051443,0.29352855841179726,0.12674153397015075,0.699981012212774,
 0.7054299160752967,0.3931568069300191,0.4259841946718571,0.2226259365093809,
 0.3349151949030249,0.6590936768476312,0.6804899628449448,0.5075446526551418,
 0.5867390135278326,0.9555034168420992,0.10558924414400983,0.4061108619615146,
 0.7750036408216191,0.33008453878788846,0.9316073783065725,0.5582163762077275,
 0.9294043272081547,0.40637600281094965,0.20713156105295805,
 0.8031404849208682,0.862133105312425,0.111719996876113,0.13481279655850076,
 0.011586718643626703,0.618071280064291,0.19964000955976968,
 0.22294085057599267,0.26510622829499586,0.2543361878658086,
 0.36111633288351797,0.9497225065191832,0.20926461318241107,
 0.9048191055321275,0.7500211760273857,0.9476718846021739,0.7928321931123797,
 0.2629050739091485,0.22997532380008778,0.43211530298162415,
 0.22477578063510617,0.4020947170544915,0.7612331481525579,0.3140888409213898,
 0.11057387916821748,0.9508705558543638,0.5326865783448462,0.3924457684293303,
 0.21014707981726355,0.1483393375238874,0.7529660631435499,
 0.04691599967382576,0.8119256712463847,0.41239146783346325,0.771725797420416,
 0.07507644840647365,0.6976321365309496,0.08940671521453847,
 0.7270704421829872,0.4104316357105866,0.2745055029490914,0.35909467655030225,
 0.6848452559189883,0.9963945912737423,0.5239653912017432,0.2704838300903063,
 0.38047887796020696,0.8293893957062438,0.7916127195352025,0.1406127130260093,
 0.43237276550785,0.014802971726996406,0.949931633420176,0.9730962549939434,
 0.5750850840613704,0.5871800152659278,0.0676028717138435,0.817611167827722,
 0.5532765063495091,0.2885883110694162,0.7262545495462469,0.8572307544782665,
 0.973459596265045,0.8030034299468404,0.28009017635614597,0.2693813378041403,
 0.2955504123498358,0.08418917575012697,0.5484446062042296,0.2712337528370138,
 0.6267848728895179,0.1313783594848199,0.29468721015829213,0.8854316271637733,
 0.4398965149455575,0.07107136576957351,0.0879574713149942,0.7718872374711189,
 0.36145711075464393,0.5164701825705698,0.25142183964345177,
 0.9675702398305936,0.594009655972385,0.4332636446641086,0.6835521117983825,
 0.8327793554856749,0.9493601039451497,0.20831358266302702,
 0.10463173988440366,0.2726904962396717,0.6527159150940075,
 0.030980655400188906,0.21098900307570556,0.041095944198027756,
 0.12648217012138485,0.9258007414068231,0.4834520391626931,0.5682200902010265,
 0.2737578883616054,0.6372913567325182,0.9400858255962444,0.7769546448763056,
 0.16742679660792936,0.6238072081983039,0.8832833884755015,0.3951888601274557,
 0.2963284773579674,0.31188038724119993,0.0827786415978532,0.1554945263912474,
 0.058385320570820354,0.6755397888822016,0.9706016298658057,
 0.6706369822401766,0.11352440199002811,0.0924722877126275,0.9432301750970725,
 0.4742823374510715,0.911827008398626,0.9056621599422952,0.0615765412889534,
 0.7332247404490014,0.7982486384524828,0.6977701359555649,0.46395127146756077,
 0.3862515200374107,0.7525163994715173,0.26729844953220694,
 0.37617405431620454,0.3246757113899549,0.4748683957388875,
 0.031107382014016194,0.8213906659718636,0.2598386815987187,
 0.8629930644921885,0.21080526176587966,0.1838072860533918,0.8679826881733591,
 0.08265321663523006,0.41275130169542884,0.4333401363189243,
 0.43699519515872876,0.12143692698925335,0.9840777936745638,
 0.2133116523440265,0.9902482390148691,0.2557862175376534,0.28121122360546924,
 0.8599035439190559,0.8677449431851362,0.6630861516345208,0.3041675252133438,
 0.02965972232717884,0.9718807662555294,0.10573926458503236,
 0.5791534647555243,0.7808782191873002,0.8595980904031222,0.6205456581801366,
 0.12490010143753727,0.550846764694283,0.5028057169463311,0.5917169575605852,
 0.16810979163371909,0.5058507468928063,0.6303046265799268,0.551685537870971,
 0.07777621865284323,0.21124393224987237,0.23653819307377077,
 0.9278913888927643,0.6296374199364588,0.7654167502909119,0.45045100709794694,
 0.2171702313075663,0.7526683337630141,0.458565482462437,0.002462613582125628,
 0.7683218622272119,0.844393166996226,0.8018227979322213,0.6449427627427772,
 0.060014226027699236,0.586973356782729,0.6107381688884207,0.1122912150851787,
 0.8254092520324973,0.4858300193448296,0.5133439695331378,0.1318254573311769,
 0.07651040194913372,0.8831406122474519,0.3415188364953661,
 0.23402516810314433,0.5176338072354746,0.724499010062922,0.6776192322928251,
 0.24327832985255138,0.5642920861841183,0.93828418911442,0.5735562933278662,
 0.24482911599177737,0.9842364367087917,0.0502557558181691,
 0.10959766371217095,0.7973286885367322,0.5378247530628577,0.6729811744258825,
 0.30014908803708706,0.4045070699499621,0.49445609168867444,
 0.8289412871161747,0.5424348450703483,0.7390609878016666,0.3708916541623868,
 0.7628311564169995,0.5723367543198813,0.9761010989677164,0.13105216186039892,
 0.5875157853307207,0.9706207710744224,0.3936620083435626,0.6088409578560798,
 0.8685335931060099,0.21295753743031742,0.07001544797533521,
 0.8704278517405889,0.7328490615260554,0.5279284788559028,0.0278158928605029,
 0.1818252199309578,0.16899335531747672,0.3827992369071124,0.6841669287933885,
 0.7456129706477671,0.011426675439567435,0.11810051981356451,
 0.921483978872654,0.4669079137444999,0.17416782186994328,0.11459888133145235,
 0.3738573859548948,0.35670004028754987,0.6794533633478411,0.8210138165681555,
 0.48998971993580054,0.9338910321424789,0.9563977687710231,0.9853458810281341,
 0.6237747256792607,0.17165518455357454,0.23936037809360955,
 0.6422733795029282,0.13897267866644625,0.23999196053551897,
 0.7548112311179118,0.07262437268326227,0.831581702887779,0.31623863259823914,
 0.7182389984933737,0.1992835517829299,0.6812577996764789,0.4440189900170799,
 0.4821470420956514,0.8735954388490582,0.409563788168149,0.8578865160309208,
 0.4621317789883356,0.7130387492860142,0.22643713212303762,0.782951122320577,
 0.8568964651595418,0.03616749760967808,0.4684705479860727,0.789846035330748,
 0.5658632005470493,0.6408543030606604,0.6763878730109436,0.573042524261555,
 0.6512476543144556,0.1702003199898724,0.8579815186371748,0.422878890874833,
 0.8424844850618158,0.19178209331081342,0.632873082016763,0.2976767417292956,
 0.9708381342637528,0.48559121942649,0.9275422915415158,0.9490581787803003,
 0.6419940363222392,0.6054600288070342,0.5430738697901374,0.5390945408821053,
 0.25238303990395716,0.5136835581894518,0.36691343012705624,
 0.8937627556874707,0.750960877819675,0.6148658980099531,0.014587227234959532,
 0.1962555067922982,0.7545307455220924,0.08693133477987502,
 0.23616242785806374,0.6852907275378045,0.8939754851783668,
 0.24442641212611538,0.5747135863979823,0.5021166702750363,0.3794900000030299,
 0.2835124673074526,0.8171740810315631,0.35842962891819996,0.9836326328151918,
 0.2691267732577719,0.8556345567913223,0.43625656616890485,0.7722051251511303,
 0.4725616987817939,0.961552208560739,0.017193960820905074,0.3518274888348044,
 0.14656285995120344,0.6886584977795059,0.8604156085836313,0.8073396616146407,
 0.8618525922397845,0.33301986162936226,0.7851275658559009,0.9083259490308583,
 0.2611776972857336,0.15679132530733897,0.022581419141279646,
 0.4445319832214053,0.9710931418461586,0.0679987827724704,0.10414405518570646,
 0.08907994862361746,0.6212760364631863,0.3371539050173078,0.2865705860969501,
 0.6821112641640912,0.26244327552356583,0.4333262706595089,0.7259313799686902,
 0.9870394681915822,0.48585151668297133,0.004572890583750322,
 0.40221810308309747,0.646913439881212,0.3342388758502063,0.8347430692274317,
 0.08850982864587553,0.2957009161546692,0.6482461746916186,0.4512403119659989,
 0.2801124508517703,0.14299948912943783,0.1397582202513693,0.8247558403848291,
 0.2941117893834916,0.35585692516951595,0.19314569894897105,
 0.8153073136730231,0.16606613063415387,0.7477469167967308,
 0.13328319939661715,0.7951949967319418,0.8030389753683325,0.6792018458332969,
 0.19860037452520174,0.08875584303598738,0.46857830745669815,
 0.6675007257982342,0.14783579636962318,0.24759000293703637,
 0.40278407075282674,0.16654339015860442,0.558700181599567,0.6792835819943014,
 0.08389425986545895,0.3184017856718855,0.6341797150602371,0.9198378216847376,
 0.27655589536612324,0.09562065509225748,0.8798333120083963,
 0.33213904887124523,0.5547990685311603,0.9836913819380446,
 0.11774499184275444,0.42210435676607794,0.9765045356203254,
 0.039895193776640636,0.5947635356396016,0.8728673903249862,
 0.048699655029694866,0.7429502676275411,0.7752543439948114,
 0.40816238355154455,0.3641560248598731,0.26920719263161397,0.966219238154278,
 0.6856433442075193,0.5661400089592373,0.9499463461103489,0.999705520583482,
 0.9556273962538363,0.14417530803540812,0.6688046969128831,
 0.30365793432729327,0.9667025992918366,0.7192600689020313,0.6939533337557099,
 0.3927982791554525,0.4714041925791218,0.26638314650891637,0.9772084747591891,
 0.3195693843652284,0.7418977753635321,0.9767827653547438,0.23158903293917366,
 0.16483835590120544,0.6017833806554157,0.4579907725130703,
 0.07029976310306041,0.2839609788319919,0.84763198058157,0.18725754277155726,
 0.09433273688328514,0.6472033003354899,0.10270789739401687,
 0.3244380354745974,0.19002869611331408,0.2785470426785488,0.2474643285789634,
 0.28834021290142664,0.6201408907362779,0.6255298766575896,0.7322571672867243,
 0.19563882065720806,0.7421675818261296,0.863527521957001,0.5554815045239855,
 0.747001079013701,0.7093836019683669,0.7650545711370642,0.7940015513102991,
 0.5319633337903309,0.36883472643715576,0.8370546860595585,0.9178716304419701,
 0.8506141972031372,0.41930317815252427,0.3182927960479084,0.3056455400807916,
 0.42412517200155153,0.7642895182587098,0.3338402568787042,0.6248374474893622,
 0.9336766787522387,0.7894323742688375,0.7052322422429961,0.7465582995471259,
 0.7992608313192567,0.01921355082990761,0.9056491026963569,0.7641067490338849,
 0.6057889473288129,0.19994967628883176,0.5212872202790791,0.0723208109126986,
 0.06004906536831789,0.5738332192923036,0.6702356065282398,
 0.09005260901631351,0.9739393334978724,0.7210980844234304,0.6532543658180001,
 0.6660889410452033,0.927416201582106,0.5152171532299674,0.5247419221677002,
 0.8855633274272664,0.7573687652307114,0.67832876661369,0.6485280612302182,
 0.5771288151228839,0.927158698255178,0.925129904595903,0.49749748329698407,
 0.5075949062553062,0.3419943584518662,0.604836499266068,0.4062323335347676,
 0.3936049592039901,0.4551345472172398,0.44461965213215815,0.7492451842979182,
 0.23967984306779455,0.06417882000210018,0.4016238212880624,
 0.4358116975584624,0.9543685322452916,0.3899886400166611,0.32292304461041443,
 0.4157599732744769,0.37979983435090425,0.5326000820800205,
 0.38571552260161046,0.21031840654189748,0.20265426495358352,
 0.48547866882628665,0.9830274846119557,0.7283683822650531,0.8092180806287039,
 0.6311284866639996,0.8798165036895994,0.4779686424747964,0.09156228154495705,
 0.4730748645958389,0.3469943606625704,0.6777607521761942,0.3528954856296598,
 0.6314704640886818,0.5802449086103202,0.18039065943062638,
 0.22234537215097994,0.3908989206769955,0.6793145658699654,0.3668679950962346,
 0.17280490435467954,0.7104807273230401,0.6358708519730022,0.461898070411182,
 0.23900713029957987,0.21942802082664703,0.12095681499981381,
 0.4804022997782055,0.8459753811330915,0.7371041373736968,0.31132202731364955,
 0.3187754675480573,0.6106849861631125,0.4226362575386631,0.4001020864255046,
 0.7197667944302122,0.8897897105579379,0.4460572927567731,0.687148770813605,
 0.5289217875520332,0.642668921529232,0.29379651327717515,0.2089521342872891,
 0.46225076435376833,0.9933232115502268,0.6963008965890691,0.5834603462677499,
 0.9739457757436101,0.6774649254770699,0.9228781543740651,0.27673115208393784,
 0.7061752489443158,0.0682490179161176,0.9948618838786757,0.7534001459383906,
 0.5978855901460407,0.9846945973958361,0.2699421342662798,
 0.029222502577018084,0.5936107362794495,0.0990838070056631,
 0.21480469575788985,0.6459713020681831,0.5057472510577112,
 0.13058457136879575,0.6856345779515228,0.6885305554367354,
 0.21685785165330063,0.6402843614729317,0.8058782044677295,
 0.07158760805699238,0.6653813725424187,0.785793703428256,0.2468534439866188,
 0.24499183189482177,0.021274845447911583,0.5348599843817584,
 0.31997532258872874,0.22956896403218507,0.855593416403602,0.886915194007691,
 0.7832314696212259,0.32398042820074446,0.7188952812885584,0.75609234995934,
 0.7305953324569061,0.6635305664700886,0.133312782987151,0.8819095010232828,
 0.40425701787856416,0.6935741651229292,0.9262668330882007,0.1453331059752594,
 0.2433795481457509,0.6394327806781073,0.2740844286915468,
 2.4150107812970223e-4,0.16998423268740193,0.971626051894494,
 0.877987509618175,0.4121747866876887,0.21996576875770169,0.5571036693981395,
 0.9392933895871546,0.4678538060271612,0.6471703234062277,0.9067820412653572,
 0.24633906267821892,0.9731393259365195,0.27607753831476234,
 0.8853290857543733,0.5615911436642153,0.696981968568878,0.874170463412043,
 0.3879726967471465,0.16748089537439792,0.7347556875318079,0.5266612159514142,
 0.028620469489313183,0.8941884513793654,0.5867963389055527,
 0.12736538506596717,0.2690445673800206,0.22962526627197566,
 0.45903276233726076,0.9097217142077422,0.5893828015323646,0.5612195648134695,
 0.281854386224056,0.7187128599786742,0.8219721150333213,0.9838889406443713,
 0.14298013186898573,0.31938928565227265,0.47643116272231356,
 0.22904760738072116,0.11562658263806735,0.5029193585305788,
 0.6392234596057775,0.3544605354177257,0.48029198819763086,0.8835447102382216,
 0.05629759801214318,0.7823966328303055,0.48390943962134547,
 0.1646327941657274,0.49635292201644987,0.7802247846404223,0.7390386612996787,
 0.22993873977336377,0.1777312707156149,0.3880982418077086,
 0.13875108750843834,0.5040683679448812,0.3580931751704043,0.8148662715229636,
 0.5963070851443957,0.2827121755767493,0.15214362064160603,
 0.40773847697868804,0.4487079135995393,0.5301306047489246,
 0.07209634698537204,0.021847534083346054,0.05083844042494756,
 0.1855350387030592,0.6085334028825334,0.08574641968311136,0.5550744122914176,
 0.14613088145361885,0.3503786336745053,0.8353591586513256,0.5969637637299327,
 0.913576138120805,0.446359376576851,0.6811443697627779,0.9795342257151465,
 0.09742059444490458,0.6568782426295661,0.4643528025964073,0.766233562371592,
 0.053564410319784206,0.01344931284902473,0.8129224626339433,
 0.9701274655254131,0.9172197889473126,0.7041277110579759,0.2512030884500982,
 0.02796118654029711,0.9108397607556684,0.9397072437789055,
 0.036695654311920345,0.3702506764809793,0.021935983511671586,
 0.7693813049264157,0.25616876098870256,0.3203814826170054,
 0.21962873905823999,0.439446536468976,0.20063787606923522,0.4929401619117765,
 0.6040284804631576,0.20234229954499583,0.6169335953959099,
 0.05374081693321975,0.8531181830174104,0.2695076620736736,0.8485625486987494,
 0.9204261193017802,0.8455630242006633,0.44175160653041234,0.9102589903279162,
 0.08667896440396461,0.7820486182268845,0.6460828955724395,0.5251773420970691,
 0.1982191450173838,0.7451299447546487,0.6919274080390033,0.8439066435226412,
 0.5301209109815208,0.6459631330894731,0.9707303728635652,0.4710968960176837,
 0.6381642554784366,0.6794622415890136,0.10956962074050469,0.3924582692015821,
 0.685226210365399,0.9552689732151073,0.40992055928303817,0.676563160495676,
 0.587126974384054,0.5938391110676857,0.4943923243629874,0.5460419606300607,
 0.762832377192553,0.6445593704194574,0.5668350577413914,0.2654555237215662,
 0.423386417338278,0.8593693526762979,0.48254780275784726,0.18083560274185573,
 0.06412739893656161,0.3490303731441744,0.6039222394700474,0.7010638443407904,
 0.8130167971573014,0.3762310749735469,0.7163258217284357,0.06832653460935079,
 0.3557613267436769,0.2727241124773804,0.5362742677509726,0.9041404530988194,
 0.744249640273904,0.6312266210378792,0.7718040967161705,0.8449259188312558,
 0.9010493630606953,0.05345655336489297,0.34709934961547895,
 0.4936925398296116,0.40863654291164264,0.4853480521445398,0.6991698408073643,
 0.27782405695489976,0.7977164515300579,0.9462460012227848,0.2860678969314353,
 0.5656141040647644,0.3398221598147386,0.9737507127170726,0.36769748489573284,
 0.7595909082768848,0.912677694998699,0.1824012093228078,0.18217676477505285,
 0.3861584602518744,0.7469137637417994,0.6510702252170415,0.14361082375630285,
 0.961623241873708,0.09410891356069662,0.5808408459391702,0.9679828945737603,
 0.28332686425453346,0.756884240858591,0.3227192535381558,0.12719201548058678,
 0.7849918289182737,0.22066007386056752,0.166182056156171,0.10070181565156344,
 0.30816382371846496,0.19470068642452576,0.09186035204483722,
 0.4993773457986085,0.7878203797754838,0.251883726513201,0.411814913243877,
 0.23355691580925186,0.584212741398254,0.07433320826394896,0.3770333009744148,
 0.9341760818139457,0.6020167958454382,0.9563956171132239,0.1002841027520831,
 0.14429540255036866,0.915043943059408,0.7365312503592245,0.3525059426075786,
 0.5583268349685715,0.9238357573667574,0.40429092532127897,0.5344202685005435,
 0.8489131076315805,0.5868885843492131,0.784054217327975,0.2097440474832437,
 0.3660879095518563,0.10066844449390111,0.3637367925229732,
 0.22120384970748974,0.8563400104107622,0.834478400961619,0.3867902714217799,
 0.15421502627552264,0.6087112832314822,0.5621582225782155,0.7750159572781633,
 0.9690118837352679,0.612173293838592,0.40679756034672276,0.6328337472906828,
 0.6088612223824066,0.036114299226237656,0.3054466492296877,
 0.3907251795984219,0.9522018858358338,0.2275603318634939,0.09371894052447072,
 0.05072395250819639,0.8299391793323966,0.375462380351798,0.4386855763701788,
 0.2924594786717357,0.6501311159166423,0.36786240047875607,
 0.22164651931299528,0.014464327777176877,0.8240623453660783,
 0.5756085914594139,0.7585477260821035,0.8922510351291028,0.14886147314195974,
 0.625424590779538,0.47781678647922543,0.1584234792997702,0.5536324385243029,
 0.3160230732748792,0.4526592873100339,0.019886239262893457,0.695384297368815,
 0.7861743753900192,0.6722612970523724,0.10091582377590158,0.5260944580947196,
 0.86791803002626,0.7598511568845885,0.16406775235246063,0.40570014396895215,
 0.21387238610875114,0.3117503270205674,0.3479594443558669,0.4203546327706992,
 0.614620124932681,0.04948616251778071,0.9296876756869832,0.7968617044035483,
 0.5508193629452494,0.9082614451243216,0.2002464650338447,0.9291646492146767,
 0.1045551924854018,0.6756152979777763,0.5028810536311124,0.35038966006078087,
 0.6004741085185664,0.26768161700881166,0.49354549641135115,
 0.2414854432245247,0.1727259410280092,0.4554871020725818,0.650949438476706,
 0.9439843278757469,0.21186783391366137,0.5047447167222336,0.9510498448150936,
 0.02543070340376974,0.21616084592991858,0.9264160373924829,
 0.38800140782358916,0.9796604828521757,0.9676794406825959,
 0.0033029555083117224,0.34403922761656125,0.049358798958747174,
 0.7788070234107057,0.5837812436529177,0.7478941680613965,0.9495204647811811,
 0.6103259052740275,0.2537183550826819,0.5931147522424475,0.14198769497597796,
 0.39064199816501133,0.7647117005860934,0.31469126989711527,
 0.00827048021208221,0.5700963849263111,0.8986340200548625,0.7372576139411835,
 0.4208944996449757,0.6119235519493157,0.42396514446165845,0.6518689890694385,
 0.8749291357448793,0.6877880415977417,0.9673754012930736,
 0.016177179892968963,0.45236659755817976,0.5601379333489862,
 0.7593502481896113,0.7179495989991017,0.6383126676074912,0.9380846371448417,
 0.29555898555672444,0.8687939875837448,0.7626688095722792,0.7437710495386237,
 0.5260340845058031,0.2794827193517515,0.9191976856317932,0.8712370765617691,
 0.9778737862375176,0.03348963552544593,0.48394607511924925,
 0.2213019988280842,0.7314641134456336,0.8342457890192987,0.23186743809916555,
 0.07837376059358536,0.9184317232192686,0.9134174806762853,
 0.11692273257651808,0.9563038578690182,0.6165133803247825,
 0.13634219321099728,0.7457445379245661,0.008721686840651555,
 0.6801796306902042,0.382786416938412,0.15922299149029107,0.35500446587973045,
 0.1193172016115236,0.9304672490585059,0.3132015264231165,0.938986688583526,
 0.25715457795021324,0.6775348241389426,0.292968766957715,0.3644320026134876,
 0.6099245153448699,0.25557330911018517,0.39997449517533323,
 0.45510356451083533,0.9205906312237669,0.9574079174528431,
 0.18237850609957473,0.6486299530923861,0.21907799177779796,
 0.2894846330669967,0.6094342557270448,0.2677547264610316,0.5027967047691829,
 0.9133173694717356,0.5406194551969682,0.055494066046734125,
 0.1045010269857195,0.8292992310386882,0.4975781406021359,0.31384943697684375,
 0.45215086181049857,0.7107810481386617,0.4443420246716244,
 0.0019219089375670517,0.5628879010403004,0.18094684256425664,
 0.3738979982642944,0.2962296508745351,0.2157873237678214,0.3441598070804558,
 0.5221585480952953,0.7084929207904718,0.5330704172013845,0.7318165958496471,
 0.3727423596234426,0.22820506332210522,0.9541883041795909,
 0.07162841539842968,0.6103807547225623,0.2601836512439155,0.7682011090319585,
 0.3305112061387091,0.9410205512235408,0.5605229465960359,0.516473653022099,
 0.13754563520204444,0.259722094604827,0.6161927587874165,0.44076276141842563,
 0.3329961593079138,0.07799199105229837,0.4816181961748338,0.3088833251279134,
 0.29929899211835387,0.5069045817248433,0.8890491021633008,0.7643070495237085,
 0.4324739663718158,0.5273070679672874,0.6877614262998801,0.6632457550420647,
 0.3248099614184371,0.9778406573800925,0.3598050792732449,0.407578745356296,
 0.12519640074901928,0.8244580285314534,0.41723404466303315,
 0.7014954932718173,0.5114566000214062,0.670040410884012,0.892698839822109,
 0.5544443671987523,0.6688283672017619,0.29701905756725433,0.4134125603773704,
 0.08402088538805519,0.6863583001222696,0.224471625664733,0.6002316464406574,
 0.9182613598408674,0.04366651483872863,0.18614292623493855,
 0.6508011259078607,0.06639974304235563,0.393486821949387,0.07113021513655138,
 0.8843372761493635,0.8671318739842172,0.08457223680154669,
 0.43920092074658534,0.9947809888960796,0.16641284556341507,
 0.18242277451348266,0.308182117447094,0.2508634124392336,0.605560704924776,
 0.24174881197850906,0.8743754867333438,0.44036696723544677,
 0.37050137465925204,0.6838310491566632,0.04157521675609388,
 0.8813434855887854,0.49193017127600114,0.6777906157581741,0.298807011233709,
 0.09455105619008752,0.4987882710922714,0.4859250878187067,
 0.46929759174156616,0.19900765684816923,0.1389229851516054,
 0.45150747360743837,0.8698673679192157,0.8970845418692178,0.3782913090020713,
 0.025686709650159734,0.5053525570821416,0.4745346879182084,
 0.3898350131966235,0.3530981106447856,0.7643369063613014,0.46142154592664353,
 0.3659488041835146,0.840727675621097,0.29648923403697025,0.1506922879368452,
 0.8480413177815451,0.2422187933226866,0.8108295560630836,0.9045541092207242,
 0.465706426480332,0.34662707020229755,0.03789731354152026,
 0.20059346721685378,0.907597920821554,0.6446687971905557,0.6197234301264943,
 0.7865307127563236,0.6289456371960259,0.4522681535234948,0.6790059645631422,
 0.028438487203931495,0.5323247985455105,0.9316850167468718,
 0.2200073291567446,0.24785440031433204,0.6365320050564818,0.3660783834548683,
 0.015025619739829077,0.07166276838845576,0.20227994880558675,
 0.20668991015472427,0.1644737585999254,0.7515438014634931,0.7032423582705062,
 0.010351770390591097,0.7948507375731595,0.8152286005430083,
 0.4706481591607453,0.6951310608976259,0.9199424193765682,0.6906640470467217,
 0.6749521748059256,0.4128092239622233,0.6622360186962201,0.20596421451636737,
 0.7301481602199714,0.004350370244251223,0.2294678642287611,
 0.44887357545705087,0.44216156308206966,0.21517086800484464,
 0.40555512458239806,0.609317589562425,0.08234702284208584,
 0.08906354063019406,0.5274642362979135,0.058349996994877196,
 0.18833991003564832,0.1296335258758895,0.8748127750679866,0.9769650159696499,
 0.7891710494627535,0.5663042558871415,0.36918976306851237,0.4342413331765389,
 0.1241647214406516,0.7727271462470446,0.6347208190438802,0.9218325381838882,
 0.4616431326242274,0.7739002757058384,0.5633585855244965,0.8136666124517831,
 0.9733029873328527,0.13579368255963775,0.3711447782331513,
 0.44404928370570695,0.5092728636654138,0.9813310495809683,
 0.23768356590000594,0.4801830539766426,0.5143657056692695,0.6117038287336928,
 0.8259157667719262,0.9615222510598429,0.39429530075515795,0.1661120222319622,
 0.8515752995835371,0.19048108137291675,0.8826578725452257,0.7583894999847464,
 0.3815485533823988,0.27776935174296213,0.6893832818405179,0.2773568318007146,
 0.2850447376208136,0.11805919141564014,0.8093183308798492,0.7482543933649445,
 0.785708386401323,0.07090478232407893,0.6392181560444832,0.7721752669859347,
 0.6266654804007838,0.5301112870700033,0.7399785337931993,0.2763520136374603,
 0.29180442734207535,0.5444914564638713,0.9508871164026801,
 0.06267331708552337,0.5381519377004742,0.7157760100406657,
 0.19791583564770954,0.14527194505693186,0.5698993061483438,
 0.5909101109043311,0.18627899524731384,0.6815182321503356,
 0.24426687012377624,0.37646363573158026,0.5701222780478815,
 0.6703876307908418,0.3164569158076458,0.24127805333331664,0.1066955503273842,
 0.5607345515698932,0.9574761249981137,0.7707751474150375,0.21086582223003358,
 0.015236692939348706,0.6843894075198264,0.4762494845725659,
 0.15638351751841217,0.7189160237673567,0.4899793635279489,0.8656001036315669,
 0.8626157902806575,0.9880298443720779,0.30741423839960347,0.8777203437514152,
 0.27783746606579385,0.1418598196938523,0.7389570409596812,0.2660154090618041,
 0.8873698976676089,0.14826214989826048,0.32555660144505716,
 0.8842450661939655,0.9865958286554068,0.26454934206144454,
 0.16759092460124425,0.6291298607509236,0.5112796085813799,
 0.08011481919399355,0.4239905538931832,0.13741665017734228,
 0.19385947539426196,0.6839056068867686,0.7384654567465028,0.8819378722393725,
 0.003962291113026595,0.45630715055205895,0.3195417115170781,
 0.39142007157218983,0.7638003441066162,0.09034843711698415,
 0.16502782755122236,0.9831934706306668,0.19976544600047297,
 0.9454923276637935,0.704587258731232,0.04304423610493635,0.423875705528044,
 0.9077990305790065,0.800727983440676,0.5622485930152046,0.871773219620238,
 0.419726434322774,0.48310610758783634,0.9954325370599466,0.5402463589849482,
 0.993550768611821,0.6412132551233317,0.20881332525611962,0.3539850170323152,
 0.6836103290402965,0.785886733576464,0.17212238080314246,0.2923685300149914,
 0.41902415373351876,0.33140910436926085,0.703014558883589,0.7743012114156332,
 0.13757424524776163,0.7406038317696878,0.9344771376920267,0.9989043548320476,
 0.46354980318124417,0.9048546659266726,0.6963702443616597,0.3742760556631606,
 0.9045011433262848,0.722244752807438,0.11828399917299337,0.7225716372972578,
 0.760463485788248,0.7413320625477282,0.6060723579725293,0.8421881104132731,
 0.40204390961061653,0.31994890940046394,0.46943071046729135,
 0.032407487796806356,0.14986450204703605,0.9815399064373592,
 0.6997439180319388,0.4569198192702455,0.6762257654504009,0.6873567137869263,
 0.16486557759966314,0.11871105030340245,0.3017380781826464,
 0.23313900678312993,0.8648024981105699,0.1592946626702878,
 0.37771439615321334,0.5145428845099969,0.2609958131241751,0.9628533899566477,
 0.01626481573654226,0.7622966605446135,0.20622433003159812,
 0.8321675663088424,0.473462787838384,0.45975707835127344,0.20892675226165047,
 0.17359268332942712,0.300227515091374,0.6649853231985534,0.21541758885016216,
 0.6701772029366245,0.1825272443185102,0.7186897635675811,0.8468453596639389,
 0.8381589867347062,0.036597517246764966,0.11735006783296276,
 0.7831136637762646,0.003861070635724495,0.27745634176303624,
 0.35945654762880874,0.24521059449200622,0.7037064127037621,
 0.35234792452313557,0.05737179632388245,0.4846497490146616,
 0.1060807071995441,0.2948656314633671,0.8442343517363239,0.2640300098566811,
 0.7141619662946609,0.04552841522029116,0.3170592147745268,0.9195049536758685,
 0.9175117908457682,0.13616655738318628,0.18919686551247727,
 0.2981564281192922,0.2048027411570854,0.8003102702411445,0.2950351252493122,
 0.03234138191306857,0.6282304882120884,0.4754536131223621,0.5526257859889352,
 0.09964820821607567,0.44224788170996576,0.05205507573284729,
 0.37926032576950197,0.858865851220893,0.44422223954130524,0.9050609075987094,
 0.6973394545877186,0.2876727861240368,0.2710129788323625,0.26817744211000083,
 0.36581540154374337,0.6780767746370212,0.6590365070739942,0.2259727478101361,
 0.09819757518910333,0.08869070474879381,0.8053300272116675,
 0.05220479522399646,0.7485236173682459,0.34379144862138833,
 0.7834078073150996,0.4630071551441204,0.14249792941681882,
 0.27000295579711375,0.6014876986100806,0.04134250259360095,
 0.2963227257317015,0.1190263309959223,0.10149023761774945,0.916418422011828,
 0.7500793191332774,0.8641882874068391,0.4802616946915892,0.6538331682138342,
 0.3763371962834685,0.9050727688175165,0.03352190916794151,0.6263508980485418,
 0.8154178435663625,0.00857050580003249,0.2893775343369005,0.8520860567080412,
 0.8567419740703869,0.6911394372417916,0.07663033777065542,0.4340411679032342,
 0.6646809985087694,0.142932845355479,0.9230629346787232,0.582667253939621,
 0.07564119069187503,0.6982356072407021,0.7820433746651823,
 0.14083321547531136,0.08586386973502214,0.611731059379982,
 0.12485906881367048,0.08722330264239586,0.845715524948751,
 0.09952265348798917,0.7473870476559725,0.5580435455419457,0.7171862751981207,
 0.01953680426712523,0.5234502560538623,0.5893799688894821,0.1795487478175044,
 0.24811556329708495,0.9073356199525294,0.05259526864785702,
 0.6445880941694835,0.046769143612173214,0.17724780978641275,
 0.7494363687795363,0.08170103665807982,0.14508450788849658,
 0.9833749530147903,0.5920713852862082,0.5594539779427593,
 0.007987756662501955,0.8019422614821498,0.18775443066517217,
 0.9812767523309698,0.6130150269986367,0.5182200639067859,0.233918447017592,
 0.3511009970743636,0.15275895833638564,0.5692473002828289,0.6185907171568011,
 0.94436592351694,0.0750914207303559,0.7920303600847168,0.9275630116768167,
 0.05602087188384053,0.7186184785714866,0.9374879313683443,0.4322710292853811,
 0.9661414746637135,0.32075290124081446,0.27884993296778116,
 0.2085200221607879,0.8559010680916807,0.19466670077244308,
 0.49972983861969644,0.755298810294923,0.06938349790322296,0.9058408585090683,
 0.9611101493870471,0.3502689160315394,0.16227261650447844,0.9289482567038967,
 0.7477395726065499,0.053047438577312755,0.044986938405596755,
 0.6662851450889797,0.35963579819012004,0.40835903891821546,
 0.37004332436431797,0.37384835406272265,0.26889519607816736,
 0.6879354664446988,0.4462412434133559,0.6394226808189358,0.4911044672246896,
 0.502524077787656,0.08508225115655232,0.25461524536152513,0.4114268352615298,
 0.7918478165186124,0.3919289735809307,0.3394904260732762,0.241788503947026,
 0.591131142908775,0.4332169950710837,0.9034872005402018,0.6852279299294725,
 0.30011850128042805,0.35164036842778024,0.9198158867491183,
 0.2641607534982999,0.9224324521485694,0.8209368537998384,0.6655431446449706,
 0.6747657633891961,0.07378400770783444,0.14253198021687585,
 0.36316252194376264,0.733770275778207,0.6515186329106069,0.6283526253578191,
 0.7120678322402412,0.09331836583548658,0.6872884321911732,0.3433888628013728,
 0.44574666633090443,0.6485812883988995,0.2863698201829351,0.5665489984459086,
 0.608985976176893,0.6776416670094679,0.16005239074494382,0.5019838412792483,
 0.7421696614646498,0.08854681053180824,0.7945717768914904,0.4208135268045796,
 0.552554847051139,0.6466457453135497,0.2153839828232823,0.8333865357125215,
 0.05126337481082721,0.3609648839886418,0.08637882152603726,
 0.5616708786494657,0.06938473884860541,0.9432838384602295,0.5026970485744302,
 0.9971189918748747,0.9969640338420535,0.6777461225123244,0.5731056351964727,
 0.316420371725326,0.30056433447349906,0.6495656378676882,0.9758674690173932,
 0.37329130374731734,0.9613298818906779,0.017726734623341067,
 0.14428072861985242,0.7283865982559554,0.5728318543200207,
 0.03761446885911246,0.3852541456124201,0.3942906985910478,0.6168946122848324,
 0.1430544396249951,0.3054848320247595,0.2106407445431825,0.5503831025640116,
 0.04078257592254897,0.7589790035990546,0.7260482645278453,0.9809724004420016,
 0.6303420879256341,0.9230968012726786,0.3340610447216048,0.15121862828399935,
 0.5587763453297722,0.5469467769306997,0.6712705487274683,0.9621005014295572,
 0.9259129936708248,0.6419050784481946,0.028567545972695507,
 0.9928402682561971,0.30283098866222513,0.6577516346080273,0.9823664026423926,
 0.8789934811773419,0.10059701880694905,0.5249175461864828,0.9679591694107855,
 0.31488121618853715,0.695892082209153,0.886781150742952,0.6148989096231151,
 0.4828221521594863,0.28126118318484217,0.6060560176519054,0.7000685691858737,
 0.03197996326081842,0.9883745027389045,0.6164663819292387,0.2779332275217149,
 0.7892106764102164,0.9017086208288735,0.8169403664454231,0.9303236378472319,
 0.4319232315932011,0.44624703026149803,0.7442432948348191,
 0.26915599044866934,0.03954317206081859,0.07271797374615052,
 0.1758341235131553,0.3434337150066099,0.3970720854483285,0.162378495182804,
 0.12593968636915998,0.2747954764650312,0.25605678808147847,
 0.6966809238383165,0.47091325313393606,0.26513973161824056,
 0.5898698274847033,0.8645730498055318,0.276723411821325,0.6504899878559491,
 0.3003729758689534,0.6631806287942679,0.2800256380495808,0.3773871031533512,
 0.569945264404438,0.5794724781975129,0.2949097883806745,0.2928066888950579,
 0.17309895407879283,0.5680350517721371,0.09262589573033397,
 0.7919008293246087,0.4581891674861337,0.8100330209968993,0.8322006425622003,
 0.25883357247035277,0.660733291328426,0.9537640343059737,0.18246078279479028,
 0.6127881922943728,0.16525977261859504,0.13402920270093888,
 0.6617964098707494,0.9668964576278576,0.44107582633762366,0.2880628363905492,
 0.02514334302341492,0.9049240410372578,0.8793672201516077,
 0.10318526605402012,0.184559644071274,0.44727152793833025,0.6129021785093856,
 0.9937176041136403,0.7379531670068172,0.13438013648611702,0.6530000814386353,
 0.0917605419455867,0.6243620246034598,0.7406618920986406,0.6793124930657317,
 0.30523586765440824,0.029780122440615342,0.7008297658512521,
 0.8213956455476661,0.8880642245961505,0.11156736806884604,0.7989009666819823,
 0.46868904721214966,0.47354340924470584,0.2610734682392364,
 0.5252493256063919,0.5390951000134812,0.35923583804106485,
 0.009258009953339807,0.0022926388062916025,0.6416217838114395,
 0.1527651069611975,0.11027657495463528,0.14729827470556278,
 0.4951049238991647,0.07647740817526061,0.5142914074003812,0.7662458813124879,
 0.7600626987770683,0.8392732183782435,0.6472144078789676,0.5476022729025413,
 0.7275134490389268,0.7783541995451183,0.6472814326853563,0.03299408449366381,
 0.2960046413361529,0.9030610744690989,0.11391499996356513,0.8507356486342865,
 0.03651212687761074,0.8289513752281561,0.5201494013706611,
 0.22290257137934988,0.5587460740109237,0.1429698827373399,
 0.22479006071388152,0.5562538587459755,0.37643399084966256,
 0.6340707483099091,0.32239419864554864,0.28983000974423523,
 0.9944336223331138,0.9345235864759895,0.18314743970778813,0.0648672034866854,
 0.20350332194798593,0.6694356800590019,0.6983215339496954,
 0.05962004375145791,0.09063298220907445,0.8319175034147019,0.610137302418964,
 0.9412114127279191,0.3112217202729304,0.42403045119896476,
 0.18546806949805938,0.28817147759795536,0.7623185239184848,
 0.3701314047543195,0.11340154321146811,0.5248164485465494,
 0.49982901063046603,0.9990641446616174,0.9286587295445248,0.8678316835164397,
 0.5890007428229888,0.15201425320883932,0.5894758709310898,0.7948003274864766,
 0.4070802351691213,0.32135882152896544,0.6817904930747025,
 0.20431928195628923,0.405245896196071,0.6933951463767339,0.8027799680942728,
 0.5672547129826272,0.6656833720799188,0.29110956448474423,0.2842413856693127,
 0.40912429982607,0.4298983601768903,0.782953163592975,0.6772807980142278,
 0.9141141993703928,0.624138157294557,0.07658996840672005,0.89695595252825,
 0.5789318939136828,0.6240487607106233,0.5801158714265972,0.5878900603416838,
 0.19415899800058967,0.05785338095588666,0.5616631760946074,0.41953746822198,
 0.08634685009488396,0.7266173044507848,0.33945536722611447,
 0.7521765609616774,0.5216134987064538,0.695220864511322,0.8560690616426399,
 0.39813076934704206,0.5189318242516446,0.09847629198899888,
 0.6762575292918706,0.7146670311886147,0.8841603074452464,0.5271242297708209,
 0.13943880135979714,0.11570947272358201,0.09484273887822736,
 0.5368693059770522,0.9664301033266472,0.32971549042535586,
 0.45830318852586804,0.4082081719137356,0.44209794023311255,0.977858936464256,
 0.7365723899930322,0.4950449503425389,0.1345927466916551,0.6102693510118242,
 0.6916158446150088,0.523720339080667,0.5827409263289567,0.9431064932765584,
 0.1376639086057786,0.6587327136890817,0.9034365425779091,0.5254277999468381,
 0.33246902160007774,0.465155895812554,0.8912146118706008,0.19491153588650167,
 0.2658430505075864,0.9545157666695037,0.7440193486053945,0.9364567228084264,
 0.7290279804529837,0.7097798521962938,0.7171990865916026,0.9672224797186416,
 0.0010444164575695858,0.9462103056063433,0.8508201798289411,
 0.9188410172008554,0.23356029949934298,0.670745724806642,0.21863647623292193,
 0.5209080899074141,0.7346945755524638,0.8023492982764554,0.5318174185232034,
 0.29933683503914943,0.47412407792466715,0.1416418922547602,
 0.7120699917821203,0.308658836576424,0.8002854335288982,0.709411651661173,
 0.3006312219506505,0.9399276429183877,0.8943663831825904,0.7320813717319573,
 0.4434527461823795,0.703832338532492,0.5058963826904417,0.062096400927162954,
 0.9468654565930877,0.8206861992817032,0.21783965776410286,0.9005096714761216,
 0.21949599674538245,0.5150431709017486,0.830678571663231,0.6902234030910447,
 0.36718234997955346,0.5246682325259878,0.6627310444934592,0.8012173202381708,
 0.16356412283951238,0.2469800387041693,0.47975401715717836,
 0.8598587260075845,0.8590717852342019,0.18091955795389925,0.5117565522784198,
 0.6400858716774347,0.8860289185196208,0.5376982124009051,0.08644114278274762,
 0.025450483884527864,0.5801276647215438,0.5272176323550178,0.80858788586573,
 0.0957481339038101,0.42424809976443223,0.34795940384396384,
 0.6229040380944157,0.2842747234569071,0.20891594086857013,0.6133598573510942,
 0.4830209357837192,0.8057982790226746,0.10111503427751722,0.6224947126266205,
 0.6297944697548923,0.5254711250719247,0.3933600214508767,0.8477200758749294,
 0.07206694447292772,0.48423095973588026,0.8373413144738004,
 0.18177461573847076,0.6429344431447112,0.5161205086362166,0.2899445400702546,
 0.7340814904500274,0.6424961409705249,0.2842712273752439,0.4723797272886401,
 0.6897872246275,0.3342986326541678,0.7992298330219194,0.8532243730314985,
 0.4463032596696962,0.5752226798474469,0.5240310262554784,0.10943146966215533,
 0.37889738662576644,0.21140912537909673,0.6895622063549046,
 0.38014007981644904,0.18616240768216052,0.03342428312701362,
 0.7252955881581118,0.7759267398097501,0.211106277226198,0.7457933240692618,
 0.9995625729783133,0.31496700097657016,0.5363680157764124,0.7222304337611382,
 0.5858624602170737,0.4696329030043931,0.05760919923484287,
 0.36370074448118994,0.10888491641817177,0.42054008007032406,
 0.224574048227402,0.6140332018256083,0.9307696701919395,0.46018934837093806,
 0.5917966267470203,0.03510590221914145,0.81128247465064,0.5167092622972658,
 0.49302019441374223,0.8722277494096312,0.29066911195962364,
 0.4001226560780127,0.05282183946817831,0.4436782689540173,0.1456957213214416,
 0.35230656214015554,0.944140271077583,0.8577878069439896,0.03828662725243037,
 0.6844821232903056,0.7785433762388183,0.5757852708164947,0.14016907482406327,
 0.1861933869223804,0.6234952027490207,0.7684028613088597,0.7063686277427957,
 0.720954493854078,0.8262865434066542,0.06474666463002376,0.551105354747527,
 0.18168878508443975,0.8721330238792631,0.31516711458983593,
 0.6802424775742782,0.5371538459451488,0.7474180420577887,0.7352520817959873,
 0.028694825741301888,0.09950798009128237,0.22027581098197258,
 0.922991544874953,0.17711367188879823,0.19581383602503755,0.999058726938916,
 0.7086348041339076,0.9105966899293628,0.6841902453745556,0.6457149270659669,
 0.12822102425962356,0.9703060087989261,0.5832802401737954,
 0.17764957016379723,0.44622876767362485,0.06764351246955835,
 0.6651111747256737,0.07443150384112629,0.6062797460352984,0.9483601126145764,
 0.8590432466278788,0.3031712487886943,0.3442292288900395,0.8392792257991921,
 0.13060738215509538,0.9044259380896845,0.30531007930435494,
 0.7492542593052118,0.3500693547205139,0.6891983013847963,0.5338806742031382,
 0.8773677894470858,0.4407929566540606,0.9145716519494279,0.9109862794138719,
 0.7966081808765154,0.9060490509391166,0.2194032758708665,0.8538521945520114,
 0.22896103661643974,0.12040563020622308,0.083481665790053,0.81535035666866,
 0.7675592327114428,0.40199387275164455,0.037737551385327306,
 0.4593519370420591,0.8308076542266483,0.21797726701239695,
 0.11365735708927682,0.3065707649553965,0.8604886219785409,0.6379207956997237,
 0.3403892020090029,0.29957575562974514,0.012987824661541247,
 0.43877514098632375,0.14563524628699642,0.8753572510655427,
 0.17372961736037784,0.7437835287392716,0.9105068018598227,0.6981819052444586,
 0.444596963237224,0.43061304981539594,0.476185870892743,0.4863410875308607,
 0.18102691465669785,0.21122206376678754,0.9723081070660295,
 0.7201473396648304,0.7655327301947127,0.7982152688105638,0.719006113825293,
 0.2379083177888187,0.2754140099922664,0.6631895708720255,0.9221621624556156,
 0.8451210721531733,0.15595316014139293,0.12863639372254843,
 0.2839786941395953,0.516691520987198,0.1614458569845164,0.33499101900519124,
 0.1635728507733194,0.4032834071953558,0.8591022592266605,0.4129631579609814,
 0.03859208191880403,0.8665332473650039,0.5862556516919891,
 0.40482933571103064,0.5096386188220766,0.95782008733894,0.19041734053929416,
 0.4240686212412712,0.3824889109885786,0.24522932319970847,
 0.24312938683972507,0.35479846438641827,0.27139774648729653,
 0.45416699825992757,0.48291168925516814,0.9409210584565671,
 0.7461527041706979,0.8261746897004756,0.8943997001920797,0.31118352154288575,
 0.6417173304747175,0.25098040437214286,0.757283636428324,0.4901427500927005,
 0.973983336791212,0.3206834522436308,0.6306229416118946,0.1121417370951916,
 0.47310584787021526,0.6676579478202717,0.5784045542974051,0.7647236889944184,
 0.9739205245290375,0.6450538656281039,0.3960482918602489,0.3725389447831504,
 0.20894936869340075,0.6192764518555384,0.39417500195346433,0.974482857095017,
 0.6867989534408988,0.8663681395284538,0.07718210282312654,0.6618079093463777,
 0.046012126562556044,0.4299357934585737,0.7771853523610489,
 0.6592678528617988,0.08391241261345617,0.08336807269312485,
 0.6324950566733116,0.664475293963488,0.13960092875576868,0.9412413529194288,
 0.3241194844205241,0.5788447745571556,0.9573362944322917,0.5860099595544163,
 0.3833091021668299,0.621482001242104,0.27832364174990676,0.3843090867891067,
 0.24664031243423445,0.8410042761079096,0.7441353668116801,
 0.18078810116184396,0.8150528426370863,0.061238409032869034,
 0.48703537228775295,0.4409161167589267,0.9031621087453453,
 0.21472623032454186,0.08392181710155355,0.5701283133523469,
 0.4835589468670848,0.7344441591744516,0.06784504500531963,
 0.40636000545588724,0.6401818276859615,0.5685155137782353,0.3674758860922722,
 0.581661065057549,0.14575701895684046,0.9396649939749556,0.8275083401193456,
 0.7354094673994346,0.44129738450194245,0.9122272678390116,0.8956771116865926,
 0.2831501871984643,0.4220786245393865,0.5550582478929218,0.5442978222591838,
 0.8185638921442269,0.2608646851763118,0.4927271948635821,0.17673186198748225,
 0.1443866129970781,0.5784766475937363,0.6667701164364865,0.1237515328113955,
 0.8945419395486287,0.3652957338081778,0.344454410268073,0.9644387054084219,
 0.8706174874973199,0.7524540309213268,0.39273963758794217,0.8134931839805334,
 0.12536150771011612,0.21172625649022536,0.2622166597563923,
 0.5137840610618951,0.7358793061939278,0.7613313738037251,0.8101990096049588,
 0.6776865563456659,0.1882794965320883,0.28871232213213793,0.2930888998745338,
 0.15850324516349645,0.0019424712292650526,0.3467719670104292,
 0.8473017722974507,0.06458953689609959,0.28743740335099777,
 0.3437303261219581,0.707358136778294,0.765169176547633,0.08295850970187824,
 0.2076273051571954,0.20470506615881146,0.15495636121381318,
 0.16204418416956012,0.2490439149731032,0.9585524687883292,0.3862530419380049,
 0.44921874628737835,0.06858958157199735,0.13499777702766047,
 0.21617722114116544,0.37454539109773655,0.28078043664007624,
 0.4816581203738386,0.24888779237973935,0.9755346947093865,0.8715404893230995,
 0.0642699452420914,0.08748313931960083,0.8896126413103909,0.5245675282495642,
 0.6137863631013867,0.18973731999973664,0.6810695226865909,
 0.43928645690591706,0.8058055198861142,0.4024934195143681,0.6041868994756509,
 0.11332351652242789,0.9159214182729467,0.24762815598376786,
 0.6405561497934644,0.7695890085366235,0.08208305256524906,0.723303993473858,
 0.10871746359027123,0.745691213394828,0.5519225727304997,0.40370370008427026,
 0.6626437137931699,0.8568856373335612,0.7935901733217278,0.37866992357167883,
 0.2284068040603654,0.32392784076495773,0.6652175329865668,0.8505592658018974,
 0.8119541197338043,0.4605766083347391,0.7831144872081139,0.26843775473658793,
 0.08062727797604641,0.00550497102541625,0.25890722465662863,
 0.6409040327025227,0.3917718551968087,0.9085688818692291,0.20182687523809906,
 0.8703115252777409,0.1422369769056796,0.022847256324709075,
 0.044722087973975144,0.9039484202909256,0.9672515883136393,
 0.9127233261628992,0.2276118352822193,0.6934161090160585,0.7327982870213429,
 0.4748170078476711,0.8768426454772483,0.9765319039225762,0.1700984635344147,
 0.1987509464611361,0.685224382267118,0.7221763156440767,0.8611082240219741,
 0.1296565899923714,0.6874861063063092,0.44423921551191037,0.3983106665762708,
 0.33163463951050487,0.2236983635624742,0.0227631051651197,0.5052362124385852,
 0.905456951657425,0.37916484882615953,0.44042578886640316,0.8038160471474844,
 0.0701531624375169,0.6499262437995028,0.972533168615267,0.5447015466848327,
 0.3905553280561207,0.6623805579061031,0.3841503083269727,0.3938932314529263,
 0.4456138550990133,0.3867790072577635,0.13741181738152852,0.2422149543801212,
 0.2855772632075634,0.8900090646855539,0.7886251392306393,0.45030002904177846,
 0.21158961186028047,0.3043688193523064,0.49830147592288454,
 0.6459584545144872,0.21779154381788435,0.4648583823248479,0.7687713924836417,
 0.4433231805714064,0.04474594674900689,0.3707709474684625,
 0.44670034940428816,0.9361415116049917,0.03252827474227915,0.877558425815678,
 0.2157757339034485,0.016385540525594378,0.7688401347501586,
 0.6597030277249516,0.5512036414782839,0.25601336525212526,
 0.15171379567383458,0.5271022517078902,0.2575371099669016,
 0.35208015041886576,0.7261725178371691,0.46685525004111716,
 0.9413574678863169,0.3954356927988013,0.27179058115005816,0.5418672519011286,
 0.13933312391639519,0.43005393416537274,0.7005445158588153,
 0.2357139002812626,0.6860508746525441,0.7105793221197506,0.2970321209496112,
 0.12850072832482762,0.5298129909482495,0.45584835078246655,
 0.20738122896883793,0.3067671907224835,0.40100286742876123,
 0.3158300617452845,0.04258161741666033,0.10330745433996358,
 0.22130076813136335,0.7176404160949524,0.8688830230194803,
 0.31658663547977284,0.8880553690810192,0.9085521083974889,0.4029555182461825,
 0.2725986119741709,0.3882443555496897,0.2802409714827587,0.3608039707591235,
 0.5989168325908861,0.6507879670941986,0.43134212141253536,0.8672533223234676,
 0.18175518664777146,0.4087966424875953,0.19828264557990694,
 0.20844287393597372,0.8622770262234166,0.581691236614271,0.2686573246253392,
 0.6647712238128364,0.5567023089495339,0.2691721550381433,0.08467693817789979,
 0.03392650431748456,0.08734182549000868,0.5279886105542051,
 0.4287723437602985,0.5233399154347893,0.11743603405217906,0.8105611470316741,
 0.9327984215778353,0.43817401649861754,0.42545901847881074,
 0.4821587280655346,0.6423972187527222,0.9175856258376741,0.6051920144947254,
 0.7413087166383704,0.6321245009946578,0.7039453454815878,0.963214256055294,
 0.09145855227752153,0.9293009212697017,0.6549984165577618,
 0.056023906850520944,0.09083501209572686,0.9255021255744182,
 0.6061704031963425,0.45919919412396937,0.5625913132859762,0.3311395467741539,
 0.02246646879345293,0.9963084553091776,0.24697499883179008,
 0.49252145574132977,0.6649662768544057,0.6210593239778424,0.6489423818347853,
 0.456830716470467,0.06931526046054237,0.5557914022687168,0.705623770485548,
 0.5784451390478983,0.4357910833475057,0.5418644177261006,0.9535259258554423,
 0.8961325003513853,0.5752410668987814,0.498123891805639,0.9581178183387145,
 0.5126131986865952,0.5373010459380807,0.8125972963657827,0.7013917095768182,
 0.04460633287428184,0.8276829555117841,0.31875142814221347,
 0.6750891586829655,0.042553379227503796,0.3435136526962469,
 0.8069233956163979,0.09192214672681054,0.07379688657964278,
 0.5054156328944501,0.056928678951996936,0.8026619113632505,
 0.47742080977186396,0.26109708266451437,0.3578830440461328,
 0.8592120928341036,0.9247945751208789,0.37220190085319316,
 0.08613633966449519,0.8286434538930758,0.6978137896674744,0.4832363317832906,
 0.9310733373489096,0.4050666075144851,0.21500557752383498,0.4901256021543813,
 0.5637478369004923,0.3988041326575985,0.9567523773838081,0.775329807977761,
 0.6791735266986826,0.7200709206667797,0.3713827368323265,0.7214220920968392,
 0.226804057571522,0.5219709191201808,0.8653569290348206,0.9008785987705763,
 0.34744532792008676,0.2794308137745998,0.07111587016047849,
 0.6695483226949794,0.044842719595960645,0.6426549248787463,
 0.3993956982354048,0.9242388655582592,0.008842117823448792,
 0.9123656251446668,0.7934676814303526,0.48864348654608913,
 0.06128437238074369,0.7394004773906664,0.2574196177846466,
 0.35270381635975834,0.5851434068754249,0.3095946463334429,0.8870779995853348,
 0.7698788404939694,0.21632964369967733,0.8959377753634645,0.5410040273437852,
 0.7804296273488526,0.3736788215714486,0.276275351114877,0.6663895167413251,
 0.9528931494334818,0.14531991848609327,0.6994850183068371,
 0.22422884732053028,0.4552784837916325,0.5009584215700034,0.5404739950550297,
 0.8124684771177602,0.2409961115082726,0.45595646842117443,0.1553854943002937,
 0.664212151228075,0.011013125148604974,0.5304825996350114,0.8690947778998651,
 0.3562717846388255,0.8201461556413867,0.17092371647008264,
 0.46105873993434354,0.9435565119843153,0.8642088560223549,0.9037129064465845,
 0.6999092781949758,0.1690422280721675,0.5456139787074503,0.41789609944868056,
 0.5442164503067684,0.888404701516135,0.1934211732805382,0.449629075746829,
 0.3255148504309827,0.06115835251764368,0.5563460432132429,0.8430931546739996,
 0.2396316241558526,0.3576494014444608,0.7225916138567444,0.6896941991616969,
 0.3570435786505678,0.724363299007897,0.7972290854870653,0.43459809676322114,
 0.2594677053650094,0.5090314876370238,0.9479543490801358,0.43023965963146305,
 0.4701803338461581,0.30302490348829747,0.41313702460069046,
 0.6878030701752645,0.5453209534014745,0.6996632633172385,0.9231819296291645,
 0.8565882264442113,0.31913084202886177,0.12901104951184894,
 0.3448207165461742,0.7763397504222157,0.7297392847486899,0.410289196862058,
 0.03985011872619704,0.7800191105505585,0.15588898407107765,
 0.6508742940141764,0.8378733149828951,0.8226650183324397,0.17363881393323932,
 0.5456012840043182,0.429851845227342,0.7578368160159584,0.20977946324096752,
 0.7106108210244041,0.1452439139413627,0.7928074611351756,0.8105387996760729,
 0.6845950245898706,0.1313517615909382,0.6795257166557079,0.9405650402781568,
 0.4950661622712885,0.6299883606124441,0.9252114775782667,0.24368354784064106,
 0.10634512602782353,0.20391839114462773,0.3546262247821945,
 0.21877121062090732,0.8026688346188128,0.42856443496731367,
 0.17374033042437642,0.9588201663682743,0.905716854693974,0.742839894953994,
 0.10592723572616358,0.4602836632999666,0.4453789524506887,0.5323216991119242,
 0.3976468924983647,0.07291544635820946,0.15163384905884414,
 0.8452772457947099,0.16434026621732334,0.749362299530545,0.184620432702859,
 0.20480827702290805,0.3135483771218043,0.13656957322900898,
 0.19553809205261774,0.04796305471238871,0.06895173730990589,
 0.3834143161567606,0.8364599783923157,0.17554550191995377,0.485343200296226,
 0.55630503635494,0.49989872011561487,0.7071848616792439,0.8408695004504089,
 0.2652413283009055,0.04339622933883014,0.24920550993961643,
 0.3630386287912364,0.7722710387303611,0.12876255752228072,
 0.37651157260703716,0.2865936906407035,0.414627925045027,0.763960251017171,
 0.3348504970378694,0.27597341153157107,0.3626072486703573,0.4553037895340628,
 0.2531067341461952,0.6900047720439547,0.873773909374888,0.6342724164625491,
 0.8403963314457499,0.5832416395235541,0.1226782520468992,0.8910561590885502,
 0.9943618697100164,0.15525013141673083,0.41985821536943135,
 0.1248378371567006,0.16350019170323626,0.9720763105681383,0.0570830636006453,
 0.7287642099128379,0.6953659536133258,0.8351656504157718,0.8774758266289194,
 0.2769463832685847,0.8901312690992236,0.850090778935731,0.6900580447201089,
 0.20071745520741135,0.11324990120194178,0.623210707397654,
 0.14881467033715357,0.145744524211193,0.6462134801927963,0.12375178719996882,
 0.9768028743450747,0.8380821329093306,0.6726260495373788,0.5624250627647951,
 0.19249272346071755,0.5947291846777873,0.6607283631515887,0.8910727102358864,
 0.8457395461282075,0.3898851345288895,0.42553017676770366,0.8992976702358935,
 0.7694909838674177,0.9945753043364524,0.7235775744390017,0.6579716008670078,
 0.9139663729744446,0.0852301985279722,0.20880104431019464,0.6759345281699687,
 0.9771203420026097,0.5494578907932192,0.5154718965656064,0.05817170321590692,
 0.7728344418956222,0.7854771985145748,0.08265152417533306,
 0.027517751746586327,0.9410483337612675,0.07716236426610634,
 0.04059450295304745,0.668851762719557,0.13337571524631908,0.5222937104986675,
 0.49752859342026856,0.46127466430814235,0.6760211544303565,
 0.19178268331822046,0.5977137649749076,0.7620601565186702,0.9992874245677921,
 0.8555103515847787,0.28266739028017807,0.7551189327045038,0.8570238816947254,
 0.19229448863701903,0.3030276332713502,0.9351082689179058,0.5024842961468696,
 0.5568584721778204,0.12096109351023865,0.2841316262905248,0.763233495151107,
 0.2905818094951572,0.6596884294235794,0.06469349043833872,0.9600702817229531,
 0.6935771630949863,0.7125533173483075,0.14644033822198443,0.8995810699391175,
 0.7507491111718454,0.9721768144336145,0.8724081570040013,0.5626861446385762,
 0.6975724583607711,0.5867333584067016,0.6861011257026023,0.8732031547518935,
 0.16923837783327222,0.38819186177087217,0.18875451693072798,
 0.7689396836679984,0.20969011775408997,0.8527797636779013,0.6375231133141002,
 0.3154730416763375,0.08022536047137241,0.5375486877740056,0.5350930438532622,
 0.45745227225797136,0.6210299876790054,0.8195859716784059,
 0.23517033772518658,0.12216776796339346,0.95600001904219,0.892688111709659,
 0.8957434448630132,0.03150388169041429,0.07579248299047459,
 0.9099523153203334,0.4118016468515726,0.16058657856679537,0.6997053832065636,
 0.7840273400217569,0.8907108615389459,0.36232461059829024,0.7007112845133368,
 0.2386990819629905,0.060160487087447145,0.6013449861295185,
 0.42339733637254484,0.0466291098648427,0.04951292012117525,
 0.9348492404415987,0.5707403621188981,0.5298845471031401,0.18307226302607404,
 0.8519726422887485,0.5677469005109992,0.5410812779026931,0.6861292491600289,
 0.5009824334571181,0.21268858713355931,0.21498466396903604,
 0.9087094083266678,0.6753551265363087,0.16615283446120355,0.5219531909553854,
 0.8317777822784683,0.14048895668561723,0.3978828915741983,0.643765998421156,
 0.47410286288903813,0.04160340264152462,0.44804182832320194,
 0.7982053881128366,0.34958900778095847,0.8707686331789989,0.6879864224800046,
 0.7715039430000676,0.049567853501183756,0.5755420091573872,
 0.022579291148100533,0.4020729855938875,0.8946220455879439,
 0.4460994881250896,0.5930691224908665,0.7694983787281265,0.7770595992465137,
 0.43667979180955474,0.3300765079496768,0.6987700902058617,0.7276318903684692,
 0.6278718874543335,0.22899170276649095,0.30215104599973897,
 0.5958871188011219,0.9210622782166644,0.8936508782181153,0.06749676609020017,
 0.1568060879836981,0.6522192312152175,0.698380638043191,0.8510658104641589,
 0.12633908432361385,0.38165624277324073,0.3972482851142942,
 0.29517216263072066,0.5111066724740266,0.8962913448260107,
 0.08056453154077658,0.07642324915242737,0.5049387237070213,0.705576987769436,
 0.10390886572285285,0.505108836952253,0.4661416885091896,0.3508766825034997,
 0.34473450432221886,0.2446754400408293,0.8324609190329784,
 0.32920370794511267,0.7808265832020406,0.5697438882402284,0.7009503628187564,
 0.07442353306535365,0.4582213035441287,0.6478590978205334,0.3129443096047472,
 0.7477889890944832,0.0037799120322202784,0.7422735365537664,
 0.7574859807628301,0.6085107343486342,0.5625088790474206,0.34835624245440955,
 0.044323547540033426,0.031198750030294042,0.10210635989475136,
 0.6449793683825382,0.1982846005312442,0.7444008192714906,0.7991315035908576,
 0.9197627783656259,0.9102743529108953,0.9319872422409463,0.6810370206121535,
 0.18905855225138224,0.23697211653604833,0.5606877218318536,0.509575172170404,
 0.6479534342260458,0.197215551747553,0.6508452119751035,0.908932590123567,
 0.855015681868746,0.07231278657479856,0.3359280200182255,0.45680083191403653,
 0.10149607899143531,0.52399584607282,0.8803609463191405,0.8744843588599804,
 0.8732888112554748,0.18009805580592453,0.6008382599770213,
 0.26424394835672027,0.8949044825187485,0.8846802990914637,0.4912469937159244,
 0.027804735390363766,0.9673047592481063,0.6604940407351757,
 0.1806655275508846,0.055853915999993786,0.97409756792777,0.30443446822875,
 0.5833714229982322,0.764815469165732,0.4529655098270827,0.39691171000219705,
 0.9030252623977668,0.8949322134406767,0.8740511980751398,
 0.041640674333838334,0.26467886480041136,0.3960414660505279,
 0.43857713670631415,0.15599677760473607,0.093076655837391,0.8937760357357093,
 0.7747600534747577,0.03682191200927565,0.7961550889859832,
 0.45381696366443713,0.7624399645894222,0.476951030369527,0.6812177779432417,
 0.5547048920629465,0.030193925783549735,0.005181400377578527,
 0.9049387795079573,0.252901583572544,0.8324836342133197,0.26320483974856057,
 0.7148516875602124,0.8972668810842327,0.1263883586964013,0.44255461913601035,
 0.45878226004494305,0.44138468287858645,0.22720369635823956,
 0.7196327710876708,0.4832525913277068,0.9412940335829836,0.7933510346759637,
 0.5794336911225849,0.7429551500002691,0.7844445939989004,0.5870573772985559,
 0.551139567482146,0.14818116062596798,0.522724513808744,0.19210844674900784,
 0.358406235518795,0.9107334115444299,0.019106099803631982,0.3327230679193428,
 0.7572526016270816,0.1326993116482197,0.02271077587585968,
 0.007413961148545867,0.6095515708990743,0.6787765963560004,0.930166584019536,
 0.4885696120117483,0.8312515244871508,0.23211802011264082,
 0.20157866307369787,0.6740499727558966,0.7211938554025984,
 0.42442060341715315,0.2672759589794552,0.2168466664915134,
 0.28589840333958005,0.5448266106867784,0.42542952063972317,0.528467604687882,
 0.9087178845061338,0.7976941743826222,0.23918302388597557,0.5890034089780888,
 0.5644203805073892,0.2752065471077101,0.21106813064295793,0.32215051261092,
 0.5957776172127047,0.6144319575376805,0.5503673621871358,0.13352116404053194,
 0.6271851539931814,0.27713054975300266,0.22643681842919394,
 0.2560913057122891,0.9652254937157589,0.17741695752341302,0.6304824858765449,
 0.7767562189568488,0.9473420874514954,0.3137023593442362,0.05956234090223134,
 0.9018915661937825,0.6543185112849709,0.8836058365438473,0.5136145495125344,
 0.7792005186145802,0.1694930906258615,0.29174798158209436,0.6910013475308635,
 0.7233577200388938,0.4357308854060288,0.1422585268843235,0.0599690245319624,
 0.15631828607940101,0.1703161450657471,0.6587155711846489,0.5119102425283717,
 0.14008327049783553,0.6031495652873547,0.23521324277916011,
 0.6426442533851321,0.6256334113818629,0.34601980432030344,0.3032230053081393,
 0.18759060503811353,0.45646248581151416,0.8076855489737075,
 0.8835647115807086,0.9907871427016424,0.053666730555714026,
 0.32403107995238034,0.9234616106225026,0.840012311239594,0.9664164944122753,
 0.04004391928275419,0.9161758147262649,0.7393117099956346,0.7547125533675769,
 0.6531540133443274,0.13557933290452884,0.9263623375559794,0.9116332102216328,
 0.8549613208995825,0.9191865660353302,0.4857871616856664,0.3255543591782909,
 0.20200192462892053,0.5029834336287071,0.8721741272947519,0.4982349469286973,
 0.6946672954629837,0.5182070700138605,0.22429529954337712,
 0.21578758640508933,0.08329204580834537,0.23015941968207543,
 0.05786846517002364,0.43819297367276455,0.9341700556083677,0.737995650794637,
 0.668218512142785,0.9578801065171163,0.5488119023789422,0.7510301601669795,
 0.8753099812663397,0.4304421754153728,0.6317311411992974,0.5865248333668389,
 0.13849129541322214,0.9714571081523666,0.3328332004747545,0.740886207831126,
 0.8713982240251387,0.8045579069159555,0.12301087539208544,0.6485025635496673,
 0.2136113607611838,0.8608456331231059,0.10936798374750645,0.9708883778790302,
 0.5976887656541862,0.8151658598608452,0.2929761305970562,0.7613979815471372,
 0.19908690968440335,0.900595204988341,0.737886152598066,0.03608458954028437,
 0.7803619190351286,0.3922381315348348,0.7703301673697456,0.3016510674490267,
 0.6186601647499619,0.6275923759518021,0.07897336275383415,0.7218400936589522,
 0.16881484728328477,0.9324395023594438,0.8998034457993842,
 0.0037053725766089673,0.017675718896277837,0.4462698763061752,
 0.24608512869186427,0.2759573527010555,0.5122108838812678,
 0.13227019763562808,0.2667392218812279,0.46639929763499877,
 0.10484517021510065,0.3090164352647986,0.3819110280756369,
 0.16883714396564042,0.6738013151582323,0.6618362537963908,
 0.09239820877534977,0.9480587485877731,0.6602454823875324,0.9979258542131018,
 0.16038833387714946,0.9696633410056666,0.17014805794450794,
 0.43693141949345315,0.5419799978886023,0.07661507762265285,0.388546469345198,
 0.3822173927349054,0.06671993159602618,0.9754389793276006,0.3665747868476382,
 0.7579885584696443,0.11037459323393684,0.02118628172490067,
 0.8016497989671847,0.35487554927176335,0.2029010283085435,0.5054301742472123,
 0.5733554910071904,0.12025811844034406,0.06490330153671398,
 0.5400312880302386,0.4404171271119266,0.2514302134047779,0.16388134458038872,
 0.8732935566538098,0.2733121523399159,0.8606321371122299,0.18238472868957134,
 0.5163815330719612,0.018443680189013323,0.03671890920111276,
 0.05876245499458421,0.31584081169129263,0.004784999387923317,
 0.733775626205568,0.42421541154881637,0.42050614126762986,0.317425615233906,
 0.3627726835908679,0.5273608634653807,0.48166389498994266,0.4258357164166573,
 0.10718908871512878,0.500287990068162,0.1473299971745935,0.15797908554160323,
 0.4855313617247392,0.2980702425565187,0.32932719064314386,0.9469547311822226,
 0.23028370371450224,0.596328400979797,0.4503823050292627,0.5980025982540742,
 0.5965508916088414,0.6286721756307557,0.5431735113613562,0.558426795105098,
 0.6268278653834131,0.5123001404397305,0.45854434765073127,
 0.23936009302127215,0.9182586826193861,0.9870474378204575,
 0.15498847527312043,0.766693673021944,0.9111495526814605,0.559619245076523,
 0.6546130345866208,0.41591107715668507,0.08367863922880858,
 0.26558758547536354,0.5453588164637032,0.7049318501198027,0.8704839048663776,
 0.6969189509367266,0.2766707794250659,0.6921783049390393,0.14452045495628107,
 0.743356426202668,0.019172929403539207,0.39514380650967995,
 0.5286952450855734,0.886378822205443,0.9456973551578911,0.9102523238934905,
 0.5605708812287462,0.07406281939342552,0.5585992872747569,0.7574637589067883,
 0.006759706300975443,0.9804046758104875,0.8762843910267893,
 0.37864489661089995,0.4937106720763902,0.32877089948976623,
 0.20256482364154427,0.31433713517265094,0.30710274868296217,
 0.2132299313718965,0.6358992929406296,0.6185164084469367,0.8673826958708729,
 0.11494232504411184,0.5938928067640316,0.12182335413550738,
 0.37005775283096654,0.6226459600107708,0.4534311142429115,0.8653976236242987,
 0.41516871035316005,0.47880312816358905,0.6642320232063307,
 0.7012368166172618,0.9762706924505137,0.38665102019816466,0.8460191827404538,
 0.18272487872511745,0.737308330076175,0.9880273062208837,0.2627298354738542,
 0.8607843688187719,0.5571430677213436,0.2957790443166999,0.13753725860246557,
 0.4089812640435443,0.07145265434013337,0.7808337191375113,
 0.38180789213099364,0.42116197823251067,0.8934957086659963,0.971633743514098,
 0.7824176759762349,0.2256228557073846,0.7985620560184163,0.8707266383574518,
 0.906274649076625,0.38355056531256104,0.6619566820970415,0.7415090031851749,
 0.5362033454743758,0.743844183614195,0.06530887331108437,0.04392568370961625,
 0.024397129680562735,0.4888252924832691,0.850550676706128,0.6306012920570109,
 0.11740912013779548,0.940899025287454,0.8942167550129535,0.2863795378116941,
 0.10825840488419458,0.17804730787357292,0.8880232798082188,
 0.5467252063603212,0.8553980174592397,0.7753170962452713,0.5774935480423382,
 0.48652690431384116,0.8258166567530507,0.16980473303124521,
 0.45373932403683725,0.25187964948225816,0.58766501358376,0.6892014295198221,
 0.6703654392367772,0.5018963286547218,0.6413651965226544,0.6299943787553435,
 0.9887651685169624,0.6367541303790699,0.5214770896314533,0.6362957360494954,
 0.07752685007083282,0.23290237332153885,0.6069360797874419,
 0.49125531773188535,0.17219491317355096,0.5800657841114198,
 0.16046347362273394,0.1346845205293441,0.34246973453680973,
 0.8619109790636643,0.18070554094789992,0.3521524053343784,
 0.39832625735642446,0.20995458280887203,0.3825236560487456,
 0.2888222807412024,0.7506354907042541,0.46123797453530135,0.546308359837228,
 0.8117677935681878,0.6410617204446276,0.30307203051517917,
 0.38736948123697557,0.448056070192271,0.211442870717638,0.7263784657249619,
 0.08149735110346734,0.4162045680292462,0.5083020060260994,
 0.23484504360548575,0.19076234990015217,0.08522659572176994,
 0.7482666828980256,0.055005934799403966,0.8609506528234749,0.985619814171231,
 0.5397256378503216,0.5298048854179976,0.12262505559864745,0.1656522102669633,
 0.241270657138983,0.8264613674811012,0.17616543537339868,0.0196864100100973,
 0.2641751486665256,0.45986521599056096,0.10806060273638785,
 0.7603087909663229,0.7755396494755679,0.37430368292371285,0.5914752702641326,
 0.3427698274249621,0.6497729756663275,0.25246610053611473,0.3870284223881151,
 0.26588755835103894,0.003159865182251087,0.17935627089941653,
 0.1663580692320994,0.842873361525529,0.3722481963701104,0.01824131822989139,
 0.22434526450814962,0.7524076051791676,0.49746790862457824,
 0.25301971779914045,0.2566663416297572,0.17002133387185603,
 0.08993704872842256,0.3377996043868967,0.07276191645791963,
 0.8743016297610237,0.4709040813965317,0.02004109393644482,
 0.24110295215142563,0.5784378066187702,0.6448024163420228,
 0.27237636286288447,0.5438794680644656,0.2578531388482105,
 0.042876838561327735,0.9374508308191268,0.454946695209945,0.4086800242409886,
 0.5203039379713401,0.8661909382625681,0.9130415676430269,0.18143661882808348,
 0.4923789384327176,0.6542649399426848,0.7279822451350995,0.5654265696026008,
 0.9302229064695735,0.4041450969090772,0.3278714959144524,0.7036798614845177,
 0.35455916902743834,0.8942092732631713,0.0399811741010796,
 0.18852672824016914,0.32070198190689914,0.45389052629979454,
 0.8570032052938521,0.7912396925805703,0.6725945598021589,0.11755606872076507,
 0.6668188155556752,0.4019806123755677,0.21375179920861653,0.7586572314687132,
 0.3283112414711733,0.3837436130378884,0.8761403640228017,0.09079245528253777,
 0.8687017328463659,0.4696504524370281,0.9701806911406735,0.19715452995871363,
 0.18421291696589925,0.5488690459962746,0.5447644260406022,0.3242671750534678,
 0.8371250644625584,0.8921601624338512,0.2481824200972338,0.5136267285348795,
 0.46546353804259344,0.23652585638322,0.8799703703596569,0.7626154914450876,
 0.35600619594998073,0.15181438896459742,0.9755447415838395,
 0.19575961107400253,0.43126609206528077,0.9471446180826181,
 0.4724137194780118,0.432931697569197,0.35080749525522203,0.30285029867523505,
 0.09120880979468626,0.72234226260036,0.5958248872015469,0.7623218669828118,
 0.6416901893977989,0.8174103339215247,0.06577390061400268,0.6955679400986431,
 0.7707045431816192,0.3375173575893422,0.737198385996217,0.8517102318358976,
 0.6056872280623208,0.6778931706164305,0.5047118230584687,0.32176014006013964,
 0.12747523935174854,0.42263328011664913,0.37583073624297314,
 0.3802955253904272,0.03284071140029832,0.8686797007914495,0.6463231544344011,
 0.18546946237199935,0.11007236648176844,0.12773739538425344,
 0.05798792681135734,0.47088318200143164,0.5970233058023797,
 1.1202652679132008e-4,0.84757711595564,0.8748259787115361,
 0.45052602444238987,0.9958487112531158,0.18695370591308802,
 0.5265997484563363,0.5646323022381963,0.7260206202355446,0.9317337815825091,
 0.8145019542303072,0.7559690903689331,0.2936327182507552,0.5681361946056603,
 0.19967327733575535,0.437003284604916,0.8471406965364399,0.4293051572743365,
 0.1028854546417921,0.16744001096478378,0.25707426403615363,
 0.5915507484892338,0.15950029262738297,0.7959543007393924,0.770659342310479,
 0.5474591774443676,0.9845843536174442,0.20206593332173706,0.683642132579711,
 0.6240220192035296,0.3126674216841745,0.24914952071668783,
 0.09233492644051822,0.5612439754243503,0.8448930918867417,
 0.21897091653981682,0.8145909164070947,0.852289773936653,0.6273943106473392,
 0.27497607811040825,0.5938018963624393,0.2996230408892939,
 0.45086504782650083,0.4772749767949871,0.4956885978046013,0.7270830859378232,
 0.6746253425260013,0.05091777648109075,0.5741247374650227,
 0.035343122913791936,0.96598099466344,0.8521842479953942,0.7243459131638037,
 0.36832527371010004,0.6539304727002029,0.03299039325125841,
 0.7060212122381632,0.5813968095424354,0.8564499885447716,0.8707344708578977,
 0.10130624117839637,0.9911721176715733,0.3080018011995851,0.9300922794109828,
 0.9343180915206679,1.7965674108000318e-4,0.269815464211446,
 0.3259741448278797,0.1637120474712479,0.48552288770300567,0.2594784988717481,
 0.3491094043762625,0.31074259403184323,0.8329084136823536,0.696000160931983,
 0.07139242474318952,0.23967511687777132,0.06517547257703082,
 0.2617095217486749,0.7247013852395934,0.006657641649728596,
 0.014695629765773766,0.7024118906723552,0.6760461623246838,
 0.8136828384834094,0.31456853591338274,0.1548760972635138,0.7717477142025169,
 0.7503910883024429,0.7223370694177281,0.4967353365311096,0.2214235716805949,
 0.27317135258176606,0.17794683184223903,0.3883007839481616,
 0.47208357212941254,0.28109411050383903,0.6873262323162587,
 0.7112656141877717,0.4584620988796333,0.2089592323129439,0.5603172201125861,
 0.42205901341017293,0.25514090011175616,0.5685800390681865,
 0.8067156760072416,0.3496300400276502,0.5489837608648347,0.18025836514925775,
 0.5638118129046281,0.6519266314484882,0.14505946129307667,
 0.19147200239657824,0.018628980941446383,0.5950224265038768,
 0.5307461932989141,0.9438387648061257,0.3616662055108655,0.18855642398188344,
 0.5432295398136119,0.6403823726314344,0.5647383296175554,0.4957222490235996,
 0.6338673585096313,0.8817872221712308,0.053536155862355383,
 0.44657226718852505,0.5716080160379606,0.8507476231885458,0.40537121607141,
 0.01814299055229396,0.967049917154282,0.6611464024695765,0.3753182798211011,
 0.685416219413928,0.29300258049431904,0.589685722865992,0.6873013447207703,
 0.19456458402805588,0.5701098448697269,0.994035723490546,0.6226951130882961,
 0.4132046645490969,0.4630434955443219,0.7320940383541199,0.7337979174182518,
 0.8464569314053978,0.5953760463674798,0.3583646646778851,0.3507887748570191,
 0.7802364763390075,0.37008444632789395,0.05891238612657068,
 0.26074719893715703,0.7711917828743305,0.6074541912839376,0.4419643789298968,
 0.4799529191748684,0.6084964192374278,0.7278793121895248,0.9172952036105038,
 0.965184802476434,0.21952486733252974,0.12955267075113963,0.9919724690140407,
 0.566781072674014,0.8465932414932699,0.27511229444097807,0.7255252376142658,
 0.4412158416011762,0.07811246292313911,0.0015875375342622178,
 0.8274921498304086,0.4097921423562745,0.9397753734132517,0.5904208984765827,
 0.1737299292196277,0.9586953895648677,0.37027729086101535,0.622171964091666,
 0.7999935295676398,0.6265729205786887,0.3811914377227522,0.18337504711150487,
 0.8705134666167085,0.5643983743544368,0.17918134143916453,0.675914948674202,
 0.30925440909795987,0.24785883221063187,0.1615734454246096,
 0.14830015573270394,0.6318477820195902,0.2705347856286967,
 0.37937026301223475,0.7955539273808314,0.8194834249132131,
 0.30689716781635035,0.5293911289113281,0.7610255103030302,0.5424545070165845,
 0.5934021101757088,2.441539933910608e-4,0.9009699037530308,
 0.06293935399585382,0.17074539563631086,0.012138324411387624,
 0.8955138641915731,0.9373804098667103,0.9466161755905169,0.8761831361938208,
 0.37219067406728934,0.9984699737662813,0.3824840009315382,0.7400203102546945,
 0.49984699730798876,0.7280131920304129,0.47153170243627496,
 0.8245192830720415,0.775866236159693,0.5487273205655689,0.5208886798681204,
 0.35533051522011494,0.4037704921752574,0.9490330285860867,
 0.07836784223755977,0.28897477777730496,0.2604626757105498,
 0.48378457433926214,0.20303701192218426,0.3058416400319296,
 0.7724113828862529,0.43726886991201175,0.2762239872480011,0.9916339410121472,
 0.8713024918733383,0.3417661143246815,0.869989799718645,0.6442322306272217,
 0.7849988599235813,0.6877804568683111,0.23390526576571147,0.2432834981110702,
 0.31196619679377036,0.5067753796854265,0.9796975029668735,0.5233938726261167,
 0.6030411534796825,0.7638199111875524,0.8717782654734834,0.12863880156940954,
 0.26021953319594904,0.44951000164471144,0.12200127094490743,
 0.5098971929316107,0.4827216303908163,0.6433933328157858,0.11136642121520068,
 0.2314560843922684,0.3862748443875228,0.29779998071674396,0.9910364149331975,
 0.14776730081113354,0.37298827540585655,0.9213489259686836,
 0.6587676215399665,0.014011488444525932,0.6578733491939752,
 0.8717628801051968,0.26460119094052303,0.9252496872908355,0.9116246947323517,
 0.5537486582234386,0.4822156046754712,0.25002656344025,0.6945900078128807,
 0.9161777024527233,0.8019811895829512,0.8004206717473362,0.08752732638081484,
 0.1766281646151341,0.252744155006291,0.8243428488325655,0.6512130854388336,
 0.2538221092857057,0.7952040085207889,0.40188572144969015,0.4423775570238162,
 0.10630160591880522,0.6549973327798698,0.06425694299770279,
 0.12189963825797778,0.26229326850145684,0.6514532983365791,
 0.5839472535059963,0.38623736139404985,0.8355739813354667,0.7172411689291681,
 0.5219584032176493,0.4580561596507503,0.3565210470864346,
 0.027626988200512015,0.009046313263828432,0.38203173767453746,
 0.603117422570034,0.8857082680057536,0.6651773551886648,0.2493393188167019,
 0.9574844549232345,0.9411430778232828,0.7863313932103828,0.5381889890589249,
 0.5218677266951361,0.08955242906830929,0.6231618179937672,
 0.007701946145727168,0.3418086312481732,0.8385450075202409,
 0.062005189253117265,0.4264595523585213,0.6365203602908549,
 0.2521703772562347,0.9062340028788548,0.505310939293824,0.47343020680450276,
 0.35172084286061067,0.49773697768126735,0.9182075805356948,
 0.2141953540278989,0.12885676471930518,0.47543062052782026,
 0.3750214228687341,0.24506140139400734,0.5660314306326273,0.5091558822615387,
 0.9604091360741386,0.7591915898776939,0.4565499695627021,0.7197716764296613,
 0.5453593558743867,0.2678075985701348,0.0240858558750493,0.5873894486042209,
 0.21770778197214036,0.5945314792287291,0.19349206650622386,0.518417878554489,
 0.06944990972332876,0.9025731832035826,0.8775518049402584,
 0.05610664818477251,0.14302014806887586,0.5592541890154334,
 0.6656940212144357,0.316143214290324,0.9108214475564075,0.7641042229952445,
 0.08818759698499679,0.040928021201080744,0.6675968016447944,
 0.21322605239955772,0.8928329058416309,0.9446073687055305,0.3227357730002929,
 0.9973433617817575,0.2088250510830557,0.742263000650535,0.1868661463700112,
 0.39920275338238254,0.7717477118720006,0.8030754263784937,0.70159163933865,
 0.31937697431394096,0.5805225240289656,0.9991711307912543,0.7898456464712063,
 0.770014880504429,0.10352860781946549,0.13725185884269864,0.7044010382873143,
 0.5559047295845513,0.381451944680985,0.5351640180236036,0.9364705635658587,
 0.06854013928054592,0.3106435647895549,0.26130617402255985,
 0.6119775873076866,0.37917352266995086,0.6511700269541163,0.9253289618339011,
 0.7493261878685704,0.6001590428364096,0.1808117079178443,0.5164052535301389,
 0.09104034263953253,0.08358859699816679,0.8311500443495333,
 0.5688285602234525,0.49872142379007034,0.7262299066049377,
 0.13076555659556888,0.2479000801894904,0.31136349439006383,
 0.7213058983889656,0.28946595092602534,0.0725270421710068,0.5369298460954295,
 0.42293044174314376,0.5582842105440594,0.6387081066984855,
 0.24537626601774432,0.4155734735225838,0.6281545281287069,0.6632766198961757,
 0.01847057726397683,0.4701918432837645,0.042712652129535456,
 0.04172227664508754,0.9692826304560005,0.4267074382803401,0.2799525418568355,
 0.8832035057112889,0.8409682135615438,0.7472103007041492,0.24281278384998384,
 0.650479526805205,0.6391754053599819,0.3525692154753508,0.3035742257020798,
 0.2695756441149225,0.8983517537697143,0.44852697612863635,
 0.33507516759757916,0.823824041188615,0.5386998044949599,0.8739092090430863,
 0.9161338834757718,0.9613910214627572,0.5702890555572799,0.6745499932362847,
 0.5603432104626782,0.41982873249900576,0.3935445227088106,0.6417935735522818,
 0.29840645645856223,0.582348069464215,0.9313787884831398,0.717471890935254,
 0.22441273009646212,0.2056872606985013,0.9962751938233708,0.7331382070551014,
 0.9958712134064097,0.3944563700617315,0.5122248622806841,0.7177199743442491,
 0.6873532358174604,0.3437016041037012,0.8023902557077454,0.8077870625092154,
 0.00948482396861472,0.2641650847888942,0.27426093460861434,
 0.6221078492572305,0.6470465869853682,0.9424790643273211,0.8220525795143384,
 0.6477062482911671,0.56887582422375,0.91956262043063,0.5167794006311721,
 0.7122362627447525,0.5451794748312404,0.8365887030652174,0.08816681759970801,
 0.7826418336898711,0.6835055030722544,0.11128395385687817,0.8176790317201336,
 0.3846754159311625,0.17394591759544464,0.7414168661693773,
 0.46532528351307245,0.5838734041534686,0.38425974254840456,
 0.2341284699043663,0.2889904559038845,0.9052349549033982,0.6117998977833998,
 0.5853434684634207,0.7907670802465424,0.31612650268852593,0.6927643603518028,
 0.545810258650512,0.42111291868596135,0.12043856871625747,0.86073725492401,
 0.24276490541418705,0.8046773596434509,0.6360595996983,0.15911328717270168,
 0.7776026247693897,0.5357460577166687,0.9640771838924391,0.8127868219710959,
 0.9000624798873366,0.42920798717769815,0.7734503410666166,0.7933705162674729,
 0.2977840045510103,0.9274818803500335,0.13787450873829066,
 0.035631757851878665,0.8228775021602399,0.4226108247316034,
 0.6600054108333893,0.3478010092953632,0.8625079561987902,0.510746469289275,
 0.8981924211654354,0.8942837205204265,0.9355822284611495,0.30259175917082226,
 0.09532103908185874,0.7825725885787772,0.034995665193154,0.42899192660275975,
 0.8899101971694496,0.7661398750813324,0.07657380644394973,0.9129718090621275,
 0.1654616228150665,0.7492137161215693,0.32843224710441776,0.9964895181350466,
 0.24959217444117687,0.4872410690570067,0.08009871006983094,
 0.9825088280189473,0.5922472112190429,0.4888766563113187,0.799826755421337,
 0.5748410180685426,0.7021754225637894,0.5610655373482243,0.20275577984216808,
 0.8290572125681974,0.2663328967255396,0.65523993667257,0.49507545203952763,
 0.3957104200807259,0.85662550597375,0.13450252056451095,0.5292405332414697,
 0.7382124081270095,0.7539983601693218,0.7206835474406854,0.206762824916206,
 0.8659248615455837,0.07889674420296433,0.9260046767469027,0.6807790716540607,
 0.6349543766387822,0.3521858688655207,0.7235046404805692,0.544569884159161,
 0.012860078281201281,0.539913487562316,0.07412576122772774,
 0.7520945316698929,0.20988973124706956,0.7527569511165316,0.6051769289757024,
 0.3940182461116133,0.22276403793478616,0.2664496180979814,
 6.712434516189525e-4,0.697387540924911,0.36020280538843175,
 0.5566469517267079,0.7684587934292596,0.1696681482866409,0.5577865387892931,
 0.3401325249916217,0.8351583115637551,0.6842743260521844,0.9387992836962251,
 0.540020912540349,0.4637691579032448,0.10447691413727189,0.8034364327080794,
 0.7862675187845223,0.5287874852119707,0.6064525158505388,0.2604253604563582,
 0.11311617371019655,0.01897943722245976,0.2697293072964263,
 0.7805187819691478,0.6444197647670049,0.04107034911437968,0.9967983506058555,
 0.9655926241938858,0.378773339294724,0.4274680738621154,0.04473546344550894,
 0.664295444812703,0.13604596721723028,0.4163822075052429,0.5679743187115506,
 0.9566653688127061,0.34459013456074816,0.13691024610944125,
 0.20733503407829246,0.2779950691369333,0.14934070613717065,
 0.20489250452298957,0.9086913853204114,0.9134905823020085,
 0.02356980986181656,0.5669875570286413,0.6002679096396205,0.7129896627915089,
 0.28480475683238216,0.5728476412819025,0.7636402166329994,0.6842363299239728,
 0.02924533705850496,0.7902876918950716,0.41109998247505897,
 0.5482164156210233,0.9128834305712765,0.41149409014373606,0.72229494758135,
 0.7853715421370954,0.5348053087280388,0.8975725240489241,0.4707307649073085,
 0.46051101206146516,0.44920691457700357,0.14063705479912247,
 0.7291680458422277,0.6698895711070195,0.6820474274609937,0.5225827552539171,
 0.8251570822103728,0.7522016296946392,0.5193145539656997,0.19126613759989386,
 0.5046057288139268,0.2333867773238023,0.19961806422218042,0.3875542657307307,
 0.28441410992358807,0.18548702966131958,0.7110438117670069,
 0.12440337095224485,0.985172620390703,0.715471355815402,0.6167590205565551,
 0.18769412714350586,0.6970376623956266,0.2755427624197352,0.2674515182040882,
 0.9973236184575953,0.7844356552777337,0.36695617524734025,0.62137508288142,
 0.7345238696076138,0.015933037090701707,0.8556466890825227,
 0.1047817804786162,0.4328462762195392,0.36698421425511074,0.525957856564046,
 0.9737991016940916,0.09859146471110258,0.950345845216184,0.6049734812329904,
 0.4130069607915484,0.25733949092575925,0.5015428495906773,0.6375074318081362,
 0.44794139001188726,0.21138223246977184,0.07284020679041814,
 0.5247237333221705,0.6754713309889965,0.2544611000743716,0.32887659383104395,
 0.649429310936147,0.8471308243954111,0.3495979007647444,0.34971386084051537,
 0.7817356438452522,0.8172910940650885,0.5781944257074483,0.2869661349598378,
 0.15246075956883387,0.14501334803479504,0.4387272153344267,
 0.6199839786571174,0.9102482750593557,0.11196015978986396,0.4263432916424693,
 0.8085114293855652,0.3888821217908893,0.7315788136953876,0.49329403693562457,
 0.1691942516950784,0.2939936934015073,0.5747206360734296,0.8621227219915198,
 0.7494530682617598,0.7471372269415766,0.06054556362973573,
 0.16456279806407415,0.6530507123753134,0.26472560848258997,
 0.0935399776073994,0.11783318666122289,0.09287075393221178,
 0.5956169490380188,0.3647003738519854,0.7731146424159872,0.8270543683241531,
 0.13503880938483825,0.47394055248333145,0.23842937858415647,
 0.5636309464438104,0.11523847760327088,0.7764500183833208,
 0.36815483670950244,0.9493682134103775,0.5961299353252291,0.7726369851106782,
 0.7972117837750796,0.1355926073244541,0.4858788731217377,0.35090862476102436,
 0.547156961023692,0.0020485664431454387,0.8275129396395497,
 0.4373234889533264,0.6981737531742277,0.7024857623649933,0.7305993300486316,
 0.17405671953938806,0.5798599473156798,0.9380674909173301,0.915801913780552,
 0.12420456178389716,0.8101917068371645,0.5913574426096359,0.7379730106546839,
 0.1811806042710874,0.35287338919267897,0.6493828443537156,0.252987263299034,
 0.11895065923483783,0.2145197880694958,0.41712523084778974,
 0.8527299914130824,0.4359167911465497,0.050959006467695134,0.623143263239494,
 0.051362520038076176,0.1068649524382358,0.047290768934078464,
 0.4107059426950931,0.82266784997419,0.7042190913713008,0.7302633152993889,
 0.5089471608057295,0.1462648680555878,0.6062149423271104,0.07757853111532997,
 0.1805656331653942,0.2847734415220404,0.7066932374554945,0.730092902805005,
 0.8486124792304257,0.8053360740420393,0.8106947395315722,0.20451307685731157,
 0.16722346073304262,0.21551915733510607,0.00820344908687154,
 0.6756746694450033,0.7208564051353852,0.5740027881391334,0.7226821424198229,
 0.5066090852619924,0.42089078165373184,0.7668000183893272,0.7819980831053663,
 0.8732002140770474,0.9012891926826956,0.4549174393766616,
 0.0014294419742634235,0.36486189152088067,0.5416300152535091,
 0.3780961212076742,0.23365223675428293,0.306927616221054,0.23708758918383832,
 0.061061895623905516,0.9592581795558568,0.6049778302760707,0.519126042617264,
 0.11339978603741896,0.9652534977981302,0.44254510983670003,
 0.1290854249377511,0.06847999278725014,0.6620238834545914,0.9058868554583255,
 0.7802335779716352,0.7002100136001157,0.8327007440451362,0.36355558713044356,
 0.3434690646170532,0.43991805199485967,0.631799929342626,0.7763425749074692,
 0.7873898141585066,0.34529725417043544,0.71508842380399,0.33222941403730644,
 0.24500714078406183,0.6203171559037293,0.43185396774043294,0.613829130422254,
 0.10692575755381961,0.3243451548442041,0.271841294192529,0.6904564219887669,
 0.47415835678022034,0.6823317397018316,0.637739527779454,0.6846140785257301,
 0.9444278188652013,0.7166658976339302,0.6091922698441657,0.6773951187730635,
 0.4880067276778768,0.06359334011903595,0.8880716896303185,0.6291295307882933,
 0.4235596264290833,0.48379399310679905,0.8750247323366238,
 0.10889769277498251,0.821213226601322,0.6570063883119648,0.8547667963852728,
 0.5459235406920346,0.5581408861633772,0.42985178971978144,0.7910292841105374,
 0.29317922451351386,0.950879593039116,0.3392556669484986,0.9175658925311863,
 0.015727766106479413,0.33681750142584566,0.4011879274752277,
 0.9846059792685059,0.5697788785016495,0.07931075807416421,0.5485196893183018,
 0.5289282657812511,0.7815259612059783,0.08703529323110537,0.333463514431137,
 0.6257144829680714,0.9044778857063093,0.03753844948794871,
 0.04220338995340056,0.2874574691320617,0.8250305660695492,0.8148368137808776,
 0.5152973729366735,0.7753760893203746,0.291712349615425,0.7117274917387407,
 0.5124596384215397,0.16073364835277626,0.37153796928303373,
 0.6648744457035434,0.07255672850701211,0.3143850139596879,
 0.20623556042767133,0.8619861890965106,0.22971161007721386,
 0.9510125157969749,0.3292847417365041,0.5463051673548752,0.04940800703456194,
 0.8757458915330687,0.5221395144777498,0.11469676970081488,0.8980172785017524,
 0.5087384936739597,0.975576035384761,0.7448278218492034,0.8090111661095737,
 0.2625967781075511,0.11409681235256297,0.3573864771821743,
 0.35351273076677936,0.18917757862264328,0.3612119918987611,
 0.1527638450191604,0.9433674860996615,0.7432215726957841,0.6759313911377033,
 0.5995497126784296,0.8767505985168431,0.7152029667315946,0.36691585175943,
 0.42450632708650116,0.5744827523722456,0.23019070928691565,
 0.2901103482911629,0.16239157296932916,0.316243569603575,0.13974467912220723,
 0.9037136976096198,0.6279480580271201,0.18367404091626338,0.7082479804287817,
 0.7040027783683814,0.30672770432646645,0.478729123058072,0.01648880841431355,
 0.35038912222610774,0.13745118552028446,0.301428015692204,
 0.49631611852971147,0.9453842633358435,0.48712889701167295,
 0.12940207893684919,0.9706812280785428,0.9420567964825823,0.7750258523103838,
 0.9199624020797584,0.6754541208291553,0.09077848276990519,
 0.09001098501478366,0.27451959953332317,0.6122797199888848,
 0.7616463154022528,0.9331371810380781,0.2860970431076253,0.7767568009499259,
 0.18633859305650724,0.9747649577669412,0.4814257354344671,0.8130041832385981,
 0.9074180653857833,0.7040080928458372,0.7392627769380158,0.07907629778898762,
 0.6581007383786668,0.4090498279920076,0.3744226938414209,0.8672833113728101,
 0.6770203687749436,0.7520433647442386,0.6317833541657636,0.5424197274874357,
 0.985054477973024,0.6271153945992859,0.7460312872649533,0.4815292804694594,
 0.27317287637722076,0.6757714165041753,0.45964030709400405,
 0.20141751927711204,0.37165926576960684,0.26185856324487233,
 0.27071233954521057,0.26936487811900744,0.5162030863417411,
 0.47515378469929836,0.16598837266863686,0.18020105190554736,
 0.5769978405657428,0.2891995824722423,0.2934322740173392,0.22168039649933635,
 0.4412451516162845,0.5273662612638139,0.37839964767393575,0.4959002078904985,
 0.18758943305337106,0.7787618011565979,0.05998336189589393,
 0.1146803325308956,0.5459583724761649,0.9921615883541828,0.6278963477852411,
 0.7905511643433527,0.9047090632563909,0.84028648183149,0.6530492538898323,
 0.1687528631598616,0.11289516250862086,0.1317163850754468,
 0.0014825399690985286,0.859909462312229,0.4403895048358145,
 0.22074395931768742,0.5849570567201156,0.06392794206159969,
 0.08134541270487972,0.349892157852556,0.8843638831033244,0.07380903492681534,
 0.979849107431835,0.47194217846656294,0.08290172270356089,0.2846666862614303,
 0.11721860981606513,0.7002304866936372,0.7915607758466519,
 0.43719382602942547,0.646324123566291,0.41533623805524056,
 0.18892827221290243,0.2553383159662447,0.04442002665896139,
 0.16836168904663273,0.55632147171697,0.17504443224265176,0.08443669856110758,
 0.8572322228581986,0.31471496236096697,0.5150571801498252,
 0.05041912203808829,0.413717234439293,0.817993615688799,0.5361994969114254,
 0.15327981873413332,0.5740165340190958,0.7979162689131043,0.8888411100264086,
 0.6679087477223796,0.7601688975564705,0.4717785890731131,0.32379487626737236,
 0.5538223819629062,0.4928327520750546,0.3275667212687786,0.30194549163324325,
 0.29835277016182826,0.37216287396173053,0.9075837172737976,
 0.5751080595774117,0.9016565007835452,0.5002612808484753,0.7616440398309337,
 0.9060132885785193,0.3395982674760657,0.997216095275308,0.6554126400520677,
 0.43265902718598437,0.5414497613282567,0.7996285922179178,0.7233474877863797,
 0.7011158209369152,0.9284654605521179,0.062113920213442864,
 0.008043570715147164,0.22309101806180687,0.19032483010631895,
 0.3609911618308087,0.1011621308710765,0.9222189977298502,0.8251242462994325,
 0.32711577548543136,0.2877774075364221,0.020164732901012394,
 0.8205707511980134,0.022096143218769404,0.9449289797314776,
 0.4380944273615033,0.12211178441485004,0.4493191414771709,
 0.21387487855379594,0.4369299620633532,0.4768607283568338,
 0.36315531861862393,0.689224545049071,0.7700992796408566,0.6486218660354403,
 0.37965653338060923,0.09834319209863962,0.3237963838336767,
 0.0703085033608285,0.37689253990419613,0.8188941252722665,
 0.22607982680272343,0.12875125934480225,0.9565154921290434,
 0.9385370445919937,0.02839886548233539,0.4606494564330633,
 0.14337927356409896,0.9046755311612367,0.37629697304973564,
 0.32983498395779753,0.5168869144481106,0.8696759778847363,
 0.34043739084318125,0.30565464109311613,0.7185350750316698,
 0.2471436183168021,0.8582281478227036,0.899096834709223,0.32817167820150495,
 0.9373414104286436,0.9732787482263725,0.5945208897869027,0.900069109409392,
 0.34732854601418683,0.9644820845140302,0.48007056818744875,
 0.23709384359325036,0.22444568689239797,0.13779090264464344,
 0.42320388036260104,0.1398464306804036,0.2750004558870396,0.5588387622044059,
 0.2240919197735206,0.1804056539207507,0.013022394286348549,
 0.6603865779623849,0.6774463603488661,0.8336960420520422,0.5895476481116391,
 0.6657887109736432,0.9971927283351204,0.8445945893542238,0.5337981433729052,
 0.7639705408417415,0.9735982829208958,0.4255522954802582,0.9020522432842668,
 0.18449819569102444,0.6795409032660583,0.7812427507008598,0.4815762270177325,
 0.8198286156987612,0.5403496367881738,0.3653461546638098,0.04360081129000126,
 0.07279505546823917,0.08756470907894465,0.7083042695458025,
 0.7324611612017433,0.7367765808189427,0.6791869585530528,0.1640557183834156,
 0.43845754671136006,0.0808166893013359,0.5221616862208095,
 0.0075827443361881075,0.8021527295409442,0.016437945578516633,
 0.6167516388256726,0.7611475000941598,0.4037141526957343,0.4196999987558604,
 0.17435253723224653,0.859380589019397,0.9587160000391013,0.6587110612069746,
 0.5977267014049329,0.11932299366114973,0.8881356016201744,0.5952368540074243,
 0.9311925791893816,0.14236398457549404,0.34838983018266456,
 0.14978342796659783,0.35451979948734624,0.621330505021547,
 0.40318166713645476,0.5948914549922323,0.11150796921233053,
 0.011105436870983976,0.9857405241304972,0.9160246336127504,
 0.19985162613956187,0.47231517498637776,0.8738063123906238,
 0.07730873702789065,0.807276246422276,0.8406705101272431,
 0.0038001722469813703,0.37749463692239416,0.38855518356188445,
 0.980240144663929,0.7326376024946064,0.7318987590777976,0.0853594178588889,
 0.6213386741906068,0.7275341034100921,0.4207108951513102,0.08624144111803367,
 0.16205745393256377,0.5605371327525387,0.9876188968292464,
 0.03137799917022155,0.6849512493564851,0.753342864744694,0.37205039189381317,
 0.45152856371078864,0.22900148513102514,0.5992404862340933,
 0.4305883128747642,0.07767527722935552,0.03523292831822522,
 0.4412332036500868,0.07896371018230641,0.7085089362031287,0.6123947896733114,
 0.4731278706453589,0.35029562221335053,0.2202021942062048,
 0.10892029200034303,0.051244442786380184,0.7956198244724517,0.67630122129751,
 0.45415984739619264,0.9435816577590858,0.7377464312243743,0.1590108064307021,
 0.8587098521633538,0.4848940241140818,0.5498830031986335,0.4525830054524699,
 0.8888908991243668,0.2223616348247932,0.9611021538575036,0.6811408673532959,
 0.34230402959251716,0.6869699658879486,0.633607405791829,0.06253921352871661,
 0.299883399799732,0.884327978189453,0.2968130379927454,0.5132706638385899,
 0.9525359967137544,0.7468282124520146,0.20951257016373348,0.7730442204014147,
 0.4115445830177812,0.3635331446611044,0.0642620018210549,0.45255329796415966,
 0.8613764025445533,0.13458023966955523,0.13494642612930519,
 0.6535329360301607,0.8732870306690558,0.35139599249024056,0.7145564851456794,
 0.6899414151379747,0.9325189476162097,0.3451891634239348,0.8925337042104018,
 0.09332002111272653,0.8913765697166454,0.576968023604241,0.8771769913102812,
 0.21374028134310497,0.4669320804329016,0.16247202739808608,
 0.7596531957764786,0.011657202766495844,0.390614079736181,0.6059835098276929,
 0.8573455929962499,0.2273648870430981,0.9925045493749325,0.4764457877580419,
 0.8606531002042007,0.6913719621296421,0.9141501920058038,0.44315918963987155,
 0.23007104217630814,0.002480186779552529,0.4299857273165517,
 0.5406099919871543,0.8464876463419563,0.9597365516810261,0.3284737549667843,
 0.41829426521499324,0.8203401256285339,0.9961676416679739,0.7031149475800472,
 0.27394489834573355,0.28650350012977843,0.8490301724792605,
 0.002359474709772491,0.962626048913191,0.530358876228718,0.30312879662053316,
 0.12292473114720959,0.8592207964983706,0.5527421859427677,0.3733863018867367,
 0.22329927709027153,0.7766453565780435,0.3666494869521677,0.1041408860148123,
 0.7660760946223255,0.6299198224829077,0.06369907748016312,0.5198267369444686,
 0.12908774587821492,0.13572935262857422,0.8481803942506566,
 0.5255445742243022,0.037057330500699814,0.23611946702163444,
 0.532804841377737,0.798133642004541,0.36094225989521933,0.1376758317394834,
 0.9350535583121049,0.007040579427132765,0.42228961052081604,
 0.8942944893713888,0.16122617875114253,0.5416997148872061,
 0.36921471834510744,0.45196139993927353,0.9126065699922493,
 0.7369623396523645,0.39191041517240266,0.8254120957765972,0.6766463416065795,
 0.04598938071157166,0.43683289927168945,0.8239423101716055,
 0.6220874869371855,0.5163705398567928,0.6147430023255416,0.9400301197624368,
 0.3590313855162379,0.5545776676104704,0.8000320866008431,0.48742097770873316,
 0.6853006493823435,0.521539953464238,0.8587721199054233,0.5415666946098918,
 0.3956946833367272,0.7847515540424437,0.49173927514028015,0.717775975888091,
 0.6695712246825902,0.33959162429857404,0.133115914451448,0.16486069379624624,
 0.8876041919368138,0.11758641998347819,0.22500954928917405,
 0.10512860906828092,0.5748169954134331,0.7444756217184738,
 0.02884208552707168,0.14961490977325154,0.1319475750625232,
 0.6504791979767006,0.1387746396967221,0.028080823128369237,
 0.13682067917562313,0.7701616811227074,0.2858950641033826,0.9397911802962758,
 0.033026849756431664,0.28939047380017013,0.23607676407196998,
 0.03982913177338643,0.05263364862665798,0.2882532791686696,
 0.4628412040147285,0.4947361653748752,0.913542064250047,0.6656148455461336,
 0.23367247299815708,0.9741427509249597,0.4073844499306105,
 0.03073323140451545,0.13866154875032244,0.07758343606183182,
 0.609115289367345,0.025799597548553432,0.732237302154712,0.6862733880421563,
 0.6531753413954892,0.9015873576954103,0.655277068261213,0.9240695623426296,
 0.14364836122125746,0.02047180445249086,0.3549833658039291,
 0.3979683547719868,0.9632790848939833,0.11691698930547101,0.7276864938118235,
 0.8979441490637233,0.9668604301907732,0.19457610061150565,0.9478330368147524,
 0.5989822398566634,0.4605756517142866,0.054630486532238676,0.530228541436539,
 0.44012855595223244,0.07097718906207673,0.6590284057515254,
 0.8033426696153734,0.3163486912588305,0.30092548876498304,
 0.38615634951509326,0.426711401664582,0.30976930450187046,0.8857090132340901,
 0.5580131862338615,0.8772022110324191,0.1403626810898776,0.8188713545482338,
 0.39373952025032677,0.06544119060748255,0.2887144844767858,
 0.6203604441391918,0.8254389254786675,0.9900127494645153,0.32232387523969797,
 0.5212959079655388,0.9284075920606961,0.330437483292388,0.2999168861165459,
 0.2675577650237372,0.8113241474508895,0.08614362262886321,0.8834009973292228,
 0.25934329068400985,0.7724625911611166,0.6080605139361258,0.7115144968137519,
 0.131472515842474,0.19621609031364817,0.6205789570145395,0.2551245405969811,
 0.7236554565091121,0.47971595577665593,0.9673669485202763,0.3159140208873239,
 0.5871154368473792,0.7220713110314818,0.48108281003360986,0.4027641285235741,
 0.20671194662761772,0.8235445480675179,0.07333183730904391,
 0.37004013373738276,0.7089214354606462,0.2558904782772211,0.5954226415276972,
 0.02222683995769792,0.32067139991850646,0.47673152117413664,
 0.8128070476013667,0.10427813220440596,0.9444926470700791,0.5724982831776925,
 0.3978157343466794,0.75519500320997,0.07826582490211154,0.11529245544353939,
 0.6323992939373108,0.9938881030425887,0.9629531471206473,0.14708167114950066,
 0.7822912322133694,0.7139016538399175,0.3576807250968578,0.3229759293598833,
 0.35670740941049683,0.2522442918056864,0.5658859606497495,0.7334549739586301,
 0.2164150496831666,0.8434749289435172,0.13193314705126236,0.7810398015124754,
 0.925359033296214,0.720106462187624,0.6724854389617018,0.6386169631301446,
 0.9374329366395641,0.3606552506130195,0.40161603765909093,
 0.15907781774773044,0.9738000910344723,0.10113519353460931,
 0.8151528539118265,0.3266679817495278,0.3247679867310824,0.8394495952843821,
 0.4977925715653383,0.2614197498977503,0.9086204944329396,0.10446493448654548,
 0.03509705314649991,0.07666045339508232,0.33284809833983064,
 0.1233178264148791,0.4439626287565348,0.4209613071577488,0.8324437234852446,
 0.6851105613119666,0.007213468038267723,0.7472399642013147,0.948678978565975,
 0.050463533917881054,0.41367566225784413,0.5992390391158532,
 0.03310481374832197,0.5877101575496189,0.6574379696373309,0.089125011292428,
 0.6150562080949272,0.9850203263794088,0.15902484298701947,
 0.23835813356209334,0.6957538477702818,0.45292751988331703,
 0.12760640071534823,0.5768375758006329,0.6220582915318171,0.8098753707889484,
 0.60946393951961,0.9478714130021391,0.6091142759983404,0.44065027901902065,
 0.12697636725087014,0.13898718723465364,0.5612009962444255,
 0.7987700854417279,0.5076762383214879,0.8961796925617675,0.13521528602515454,
 0.5995277882626606,0.32894112919105833,0.2964730615141412,
 0.12127945722729194,0.4639997813962671,0.4127448699480062,
 0.08576286306386138,0.9424527191281337,0.2162147463266817,0.7390317764488079,
 0.5769025912319107,0.7604586698911984,0.7394805111322734,0.550355217441508,
 0.2261909036918277,0.7900583001226436,0.7539922044142751,0.16888772514465378,
 0.5892389062443601,0.4919383676712683,0.7253534577468855,0.28266562179622745,
 0.8552823629945223,0.6720012749607321,0.27746989428998925,0.5359364884821342,
 0.21499031251078815,0.5661000567128766,0.18448916014690042,
 0.8087342291554194,0.8450258258329425,0.03415538054849798,
 0.30591245063265593,0.10581278295383822,0.5947481365876193,
 0.6247660614370528,0.02825198160134712,0.7969204515555629,0.5416852820422535,
 0.7975888697530186,0.22793965110028136,0.053521157121902974,
 0.36217430006410567,0.3348336453528079,4.142788184468005e-4,
 0.8619300149562347,0.9712770691367923,0.8050243575620184,0.13801746614983834,
 0.6244406610006298,0.6210029679948993,0.024499245947817494,
 0.5793735303561467,0.6890589814009238,0.4200577273367362,0.993304305925899,
 0.23065309286270674,0.08816557530247937,0.3081952877957076,
 0.3295480866254251,0.5842046461363427,0.9187094841522577,0.5060598890413626,
 0.8726112815537772,0.8950755671742039,0.2889750212532529,0.35461532693312703,
 0.7192002955495563,0.3150827657512121,0.5097541343516272,0.5655954437379536,
 0.6704108962456958,0.14911043416184966,0.3309417932198149,
 0.25900315450860445,0.6939405959094903,0.7786257228803706,0.9827260448732013,
 0.1349458716590375,0.16483783169491417,0.9393405349084628,
 0.10748503475664184,0.7270175083134025,0.4414306743583506,
 0.048660281364007574,0.22316986859318877,0.9532458214272342,
 0.6960111429156275,0.045044145134573776,0.13199961875967725,
 0.8079853708536258,0.5156967377180677,0.9428561565146371,0.17077677892514975,
 0.5209194676775548,0.034947492054233864,0.9194312410395854,
 0.34945688091034166,0.18904168137929167,0.03703431203250396,
 0.8806557167750605,0.8883863814159129,0.9313022574862811,0.36640656206355593,
 0.2511543793337998,0.17808784933209443,0.8296600317316211,
 0.06495250590989854,0.09808246109453034,0.15109578641199317,
 0.9849138339002702,0.5709893315772365,0.12538637285481402,
 0.36802517909759136,0.40721656219405455,0.3169278754585746,
 0.7928542706971901,0.9994770801542537,0.32473921217914903,
 0.12725096173304584,0.2076871918424914,0.3939212761441574,
 0.24019345912516543,0.2481573702188068,0.5358609205539141,0.8479940384196701,
 0.961463085230886,0.7974986380205192,0.4343681902668527,0.1259785410831773,
 0.26906227753392287,0.8318862893813939,0.36093336544364074,
 0.6261234195365433,0.28103183733309645,0.4631908526549364,0.7840573124381804,
 0.05593046368315979,0.4265981875964233,0.844105679014179,0.4213590560288233,
 0.8846708801610358,8.469311727501427e-4,0.741503866289586,0.6638981379166847,
 0.803992098762495,0.8824082342782487,0.9542876366881661,0.060782785375038895,
 0.5676065783786457,0.9469367901022466,0.7550974538777707,0.28582171831059044,
 0.405710217872979,0.995221959932477,0.1143154670445895,0.04323826196083591,
 0.7522865371303934,0.8719987252974883,0.7207919550433038,0.7471407413624792,
 0.43522643174050135,0.030287364929962646,0.04310479622463548,
 0.05166728584568314,0.9732398829935796,0.6613838246802338,0.5289750875421358,
 0.4576416311108209,0.3332164553952265,0.8509422585527789,0.7410189283470844,
 0.3606088731750674,0.9736887544227119,0.40170446296785334,0.5148616721541067,
 0.8951461792158721,0.8158882037050306,0.3654988167262633,0.9608043000924126,
 0.9655767778195362,0.8021652048596803,0.9363783603649964,0.7183964388157654,
 0.4009954439624349,0.7848038785303171,0.35780477677317324,0.7095516326852059,
 0.8274074625422893,0.8627364568768062,0.08188608834450517,
 0.04967327622728823,0.3530628931603088,0.6099025280439001,0.6928477454870932,
 0.4937495927697413,0.35037267618783985,0.987901885685046,0.34376350788962784,
 0.8477791538256619,0.5236950740935254,0.2502478248486304,0.7691122852880634,
 0.4798901646034537,0.7900219854793803,0.5221188016050965,0.3764950538595667,
 0.7928174699509608,0.6951197684412752,0.2957947493867884,0.760016235487148,
 0.6767607158620862,0.6333625376788146,0.7718067761230396,0.7283170832262066,
 0.6182867815423151,0.8300952542527977,0.9961401715070614,0.7421498768081488,
 0.6402102039661565,0.8101273309659298,0.8513447871704297,0.7351630080269441,
 0.0649340464510223,0.5957730020977419,0.6782909037448334,0.7509057491342154,
 0.08578456584254612,0.49561332827966176,0.021387123856861256,
 0.46829754908730514,0.5781565248400174,0.21398333506707556,
 0.42516759846050933,0.32450404207562356,0.5792811557585722,
 0.9089855778344754,0.6971752442123634,0.6504044909903938,0.7733077012859457,
 0.4159664176622815,0.6581104516266808,0.29674165649670503,
 0.36909018441351815,0.7440535278530955,0.17350552772861683,
 0.6554935131939699,0.11419144337059883,0.3487740956918174,0.5789502391627644,
 0.5334054656098607,0.7139000427709948,0.1733826731783925,0.3869354430152282,
 0.23588106999372394,0.597462931233808,0.3723596052607062,0.25220666492701094,
 0.9783028726907321,0.6118692372078393,0.3668673808032886,0.7001367009876767,
 0.5436299621859741,0.29296232095510666,0.9519833533222131,0.4265227913699263,
 0.22775576754597193,0.5956135402664,0.3768622953191332,0.9165507959066538,
 0.5982982425989134,0.12724508559696257,0.555996291480713,0.3878234205446398,
 0.5623683220392268,0.3802455652702903,0.7205821401896183,0.09275254418565504,
 0.6887492760605528,0.5684712867484509,0.8322634854460803,0.3082116756395852,
 0.997203814769754,0.07173752888671592,0.4574324966674984,0.7804650563172889,
 0.7791998914703578,0.19828286687698826,0.5773862560392641,0.5013276545148706,
 0.48390448521142726,0.5471326594345943,0.20556249867894527,
 0.2747014846471302,0.5743394229251676,0.3309058174130989,0.6126773253782256,
 0.9998330597723153,0.6495087617287325,0.7265924047593675,0.5930156845800559,
 0.08721798589597429,0.08683669623705337,0.28136393713008123,
 0.763561797073022,0.8265505846458118,0.6060986698511155,0.4245270021371189,
 0.40481580082983404,0.17420084751601983,0.3373976648080286,
 0.39848942723163683,0.41434437825974535,0.5318840751191352,
 0.19039357397209322,0.976526763467977,0.7622324301920571,0.17630843461216872,
 0.5022420247295797,0.08323686538558628,0.6004297401382082,0.6472789032127482,
 0.2001658762615427,0.7085347965084355,0.12242399363931644,0.1579810154044976,
 0.91242329638762,0.24626073117874925,0.9956354838646384,0.6683408334886757,
 0.2872947377634001,0.12014890821325108,0.8742887832969555,0.1284890267640948,
 0.986584903770291,0.67547377204244,0.4971262420706738,0.12462231021584458,
 0.6311838367583495,0.860977501084262,0.7346030990719704,0.9056261349867456,
 0.7197033943207831,0.3998598481252914,0.5132834860178911,0.07822734854207081,
 0.08676545297321248,0.5383195453210233,0.862695937188792,0.06462206361644562,
 0.47541863911901794,0.5678792102677058,7.393193734834824e-4,
 0.9515095880885749,0.8408985605965044,0.9357069747510458,0.9611711836969168,
 0.023785525129020324,0.8603809465242145,0.3271723067143266,
 0.5586597286061401,0.49955487337104476,0.4462349215242081,0.2638165483272146,
 0.3324624491660869,0.5817534115205356,0.7936562520420094,0.8201804450433412,
 0.5976312819922358,0.5065794227498106,0.016675751518315685,
 0.9067000285858007,0.8638665551251486,0.31944570253097637,0.2100612382021324,
 0.2519724343425179,0.8381784058013859,0.8451136329063003,0.22119320845015178,
 0.6885205759383702,0.9158131893956316,0.5312210252740501,0.641081316334898,
 0.06989657436771146,0.3383229097134621,0.36371188448943925,
 0.9885561413402452,0.07433921724773818,0.6398452005660233,0.5118182336226956,
 0.44308359550371734,0.6991827006988602,0.7459718589281625,
 0.12756512143841103,0.9597628820135425,0.9675431043222925,
 0.18815451006438388,0.1849055271198391,0.5773629032869773,0.1422334807408876,
 0.4745267809253919,0.6485748892437997,0.8069575174568697,0.43263547148538617,
 0.2323569794172815,0.08109667204735782,0.23523547430539227,
 0.9897001158654849,0.4086893009077891,0.3439539901823676,0.921613313464369,
 0.1354776922480756,0.595319962516581,0.6259492068851109,0.19446342120880455,
 0.41322876493941196,0.23330176370402556,0.17320994888930774,
 0.016631492494687317,0.5350847481083749,0.8097321536801401,
 0.7180428328815766,0.0357129795219403,0.02443478547716771,
 0.12105401902610335,0.45980371239035134,0.4655893157227512,
 0.5660194048235571,0.44973382315523946,0.3822434018047347,0.6329643201182701,
 0.2747175136896517,0.013311063270887269,0.9476013512215686,
 0.6442525924092143,0.9549647897890258,0.36359652289502864,0.6451504660337114,
 0.10851587345737912,0.7926687781107525,0.6826181200168869,0.3900141928668448,
 0.7236937700424919,0.4093667305327754,0.4057868189535664,0.5191854919500708,
 0.5522749695013595,0.6768882057271811,0.8859063004371611,0.30740239446875584,
 0.853198005516998,0.16327822359999355,0.8673346886686613,0.79162997395142,
 0.9453690127550118,0.010000842024566747,0.8597740422775955,
 0.03942626197900323,0.20065302041113542,0.31278997946896214,
 0.35231815172729475,0.648864837175964,0.3961678110429505,0.42512494558733205,
 0.1542135597877602,0.45231542951662274,0.6297441739695421,0.0322235416830986,
 0.18051499638716617,0.5792263519222913,0.099820662549861,0.15748937907295324,
 0.37141686581931,0.9703298491353194,0.4055598190337617,0.836912897995809,
 0.5050603608529772,0.3752069794575308,0.4255398344916813,0.26516507865173944,
 0.12692236769564835,0.5119796786346537,0.24737297134738756,
 0.4465946641010087,0.1582705246776549,0.5358996428848115,0.15667645274755793,
 0.12568553492035905,0.08470152088065952,0.9273952743216821,
 0.06392581445764944,0.601520953534608,0.18845515868045926,
 0.29380697642506104,0.907884225476043,0.4414162680991698,0.48950952129973213,
 0.20641984810967762,0.8867556825168189,0.0033193501823125526,
 0.9189338001037207,0.6136687240472797,0.20730151149468645,0.7973311863680077,
 0.7871380258009848,0.7611414612945882,0.42513175860360486,0.5523587530806544,
 0.8156273820830742,0.5415084983189964,0.553914705837846,0.2924563100003029,
 0.5839935575597638,0.1553191149809411,0.524822331263581,0.9532469022867078,
 0.9237668192186647,0.7872118070317315,0.5059468951692018,0.8233305236951285,
 0.9861430205949668,0.6004933677025731,0.6451808080406183,0.6046154542472697,
 0.09810428465592314,0.25245859610324006,0.4263758843593184,
 0.8131257082597663,0.9507846147846671,0.7904878043300764,0.8987117854106984,
 0.5768916578677437,0.7973613864620351,0.901296309307602,0.9885581592866169,
 0.018940648493738,0.9934269987524806,0.731548912312726,0.5826182329170478,
 0.15281324140843244,0.3172864399368007,0.11140026256094004,
 0.08745752225014525,0.030684954578838353,0.05450373746210058,
 0.24244423950460947,0.23400469711826788,0.10702863255676387,
 0.4988630157885451,0.1087748468382832,0.2436945032175969,0.17938814822297167,
 0.9575647439825219,0.1757884988845041,0.969086847374239,0.173681692914047,
 0.707629384068884,0.9259825287841168,0.21361572844326837,0.4312753128091731,
 0.685294497775284,0.5548845536311958,0.8884994310148999,0.7069909433406226,
 0.14406424579008492,0.9497056272600393,0.7712105813984744,0.486807518150285,
 0.880280838784082,0.9802821128385095,0.7021540461346052,0.14524916128838705,
 0.4597599727225292,0.8910921573961168,0.8273582666641606,0.8349750774949399,
 0.9276333195238443,0.6280284276561585,0.07152707002916325,
 0.16895659024393117,0.48576667198791357,0.7275241580423603,
 0.8449222023988432,0.7690774087665732,0.9470075382333548,0.6251666299427568,
 0.28682510145706375,0.2206054575006654,0.8704549395530237,0.5458537422665383,
 0.9193586387980942,0.018419586988907666,0.11034446193787839,
 0.48475283879234843,0.41043781088872866,0.3844826346780852,
 0.4171034743750224,0.46917371280394493,0.6919802195079867,0.9937104235133065,
 0.8547859609975956,0.013421782477384459,0.6350327180837437,
 0.09156174551660257,0.7775157845815027,0.08550920126902495,
 0.06658585977741183,0.07983535156385146,0.42675979421613175,0.85955385490374,
 0.3316922114973414,0.6324397024416029,0.5159794906398302,0.8433831244849559,
 0.7607168359078403,0.6309888645377684,0.04767268245192113,0.5611082131595626,
 0.7605330087462572,0.6496385242547431,0.11459590932986119,0.5656796773911744,
 0.7072451572930749,0.422552726163101,0.43762137824172287,
 0.005836675954691173,0.4596283324665458,0.14151384497242658,
 0.9241410628656705,0.23461452488184953,0.7673718225038209,0.8871867908007864,
 0.272048696648832,0.41643732005362744,0.38579703174703583,
 0.19314973255356194,0.8523364303836427,0.07399037167368905,
 0.31868514426492056,0.6706691198175992,0.9335383581986884,
 0.05520926392705028,0.8180234980797135,0.46264677153698874,0.434088764622256,
 0.70658841411848,0.888285445537599,0.7138128567221251,0.5150429928546685,
 0.7093439544099834,0.6790536238534783,0.35706186728914013,0.9573673769944426,
 0.23788640365217417,0.3648094348247283,0.5353714720292412,0.8421170184847264,
 0.8269531192347428,0.8230091562514382,0.39782557384103245,
 0.24436362311956006,0.42510061534769417,0.8162140351768309,
 0.6650400668067699,0.6490596190723302,0.7211923692446838,0.5875422100767486,
 0.4205833453715525,0.23303416474937022,0.579769691930196,0.9910341687481231,
 0.44991785959798947,0.6948546272052601,0.8900881612941989,0.9309762658553381,
 0.02438732624003115,0.05546286333685013,0.9049412672027688,
 0.2637725332276746,0.3004610359717974,0.4914621518350799,0.31626647595387025,
 0.9276071660314356,0.14782400005915197,0.24710732545655234,
 0.6595676653108008,0.5147318167846051,0.9061603973232317,0.8970225095222348,
 0.23057357278396606,0.6418538740465506,0.7331275635705241,0.4636227103122249,
 0.08669807046584954,0.6892813272940296,0.8870994120860676,
 0.34697538238287395,0.3560457570493448,0.5826132020305138,0.7172120391355186,
 0.45131390419549333,0.8826025922784513,0.81978691660105,0.5883257674134842,
 0.6116234741435371,0.3177692735349539,0.6662213924653401,0.4629065047066575,
 0.4433218705448897,0.17854745199004363,0.5071202859345475,
 0.14776843893746738,0.013631680259787782,0.11775374736215649,
 0.5480742729098358,0.010652770766541919,0.91318696386709,0.4653108779078283,
 0.8972361967109105,0.9025163066672349,0.47379784287021043,0.7962159572465192,
 0.5199637146852549,0.8121997908982973,0.3819464208092277,0.7223624257617338,
 0.7189957597061385,0.31478798973606914,0.913612445104222,0.584857570656143,
 0.4636795776979039,0.7166070768619932,0.9741334668482505,0.25284944635129025,
 0.9853486722011539,0.9917019485317362,0.5138954717509308,0.570982223000664,
 0.3274934577622247,0.05794245605584869,0.8030849006561831,0.226011176806594,
 0.14908205418341935,0.49732965235037296,0.7978140299716341,
 0.5471181577130857,0.3993177173147514,0.9897931205365957,0.7170031968617377,
 0.34911036239180826,0.9523331007838676,0.9467746125331282,0.3144371162754912,
 0.7732198062745358,0.24252342354735434,0.026422536625125836,
 0.2396026394925277,0.46245056091910497,0.8620002804347037,0.5072144211402148,
 0.6599474182137149,0.2509857090485401,0.49789843523520805,0.3070344518644931,
 0.5977071631936568,0.654057333979439,0.44353274678610455,0.02230284141509531,
 0.6230544948406189,0.17573699675578525,0.7859771887370682,
 0.06479788224197391,0.45156868901207203,0.6099590205740267,
 0.12910631488593838,0.9381049242693154,0.19177909004400373,0.6007626046416,
 0.17983313395209122,0.3329041270041071,0.8896474462957019,
 0.19580009199629067,0.7057468206549486,0.49253441037512224,
 0.7299359549287094,0.8208308794942865,0.04653336096290028,0.7759622531841824,
 0.5691073961103041,0.9437336679781962,0.5230824341609422,0.9981558986561815,
 0.8661433572325574,0.7143186045407205,0.43883395680291315,
 0.015622672926879577,0.01663945463444605,0.4392283396294807,
 0.48760690567305687,0.5872083608638602,0.9373765909105581,0.6874920675004079,
 0.8997321943407797,0.28567987257922867,0.8247818405558702,0.6661524444445391,
 0.4576417206719894,0.5443627905415176,0.49798954980285437,
 0.22678962443237038,0.2880634564525322,0.011212982469984922,
 0.5258945774456867,0.010820975906977948,0.284719195936898,0.6282206453908565,
 0.953382580325552,0.3701877900114119,0.2925398096895139,0.9977766522374834,
 0.8256344116132991,0.16325849147876925,0.959453131950968,0.6452393656791265,
 0.4467694610502587,0.33560681380750323,0.901890756202035,0.6918008175455305,
 0.06163866438695709,0.2595285231121949,0.4657224410320764,0.067398999436032,
 0.41469686698866426,0.6468265462268198,0.5830065178345724,0.6953494252032021,
 0.7377874432824234,0.44265045948437154,0.6416860487705989,0.5250799597952601,
 0.18912421077651564,0.5292799212433127,0.24581458715937232,
 0.3918986349140354,0.1440479484256909,0.7892373678616076,0.6418823098386405,
 0.18992308487476695,0.7872900344856769,0.8524879271714607,
 0.006575027727532823,0.4535775067585436,0.776547438085928,0.9697623606167218,
 0.979365205153516,0.9568262770461823,0.8892889075485981,0.12195108621561734,
 0.7973564371216197,0.7730075349789368,0.7546368815276729,0.6944538904990851,
 0.3602306465412768,0.5526893440429234,0.19597392984995277,0.2659031969261385,
 0.636884637768868,0.5031736466688912,0.6875449998887425,0.514358626037035,
 0.04055646097855603,0.5253624155684876,0.8169319114386421,
 0.08968292153013069,0.758136166192242,0.43055255620569066,0.9434789063534893,
 0.006895635692101831,0.07799791180869609,0.9733729099261089,
 0.32811663921613543,0.902016125741454,0.8707966680147077,0.32785029877954375,
 0.6879711632370152,0.1170337930184221,0.25730303400082144,
 0.15238542253954424,0.750430257148957,0.8011438827303278,0.3135867674490196,
 0.9012847512225806,0.46023278429320247,0.9165512910175566,
 0.057471156738336915,0.6012004511640735,0.020487468974611622,
 0.0172534039063571,0.8161535396468634,0.8580115929461741,0.39643751555810625,
 0.35229105249474335,0.6949608258276102,0.12376964008975466,
 0.7896751542088857,0.7111598857180947,0.22476610729236912,0.9039602291477327,
 0.10575465769058223,0.1989109324289231,0.8768955010652894,0.9444841940501432,
 0.04941925996749008,0.24281830855529296,0.3887847364225211,0.830192065975886,
 0.8474115308064625,0.5619942502364698,0.037013499545105466,
 0.7615101207348381,0.816671813883524,0.36361854692354323,0.8049670083929714,
 0.07539740402680684,0.8629180596920031,0.42866017324575545,
 0.1887883651926079,0.40344180679543806,0.9863483170494483,
 0.20251306231037947,0.5326444625626229,0.9894824328976467,
 0.02519065700110157,0.040729084271810656,0.5037553863362347,
 0.8046113734311964,0.49421233144794174,0.5729637817539137,
 0.005087583159285924,0.9066718963338626,0.5573929613870906,
 0.02017887039377997,0.1001257735716401,0.6690294192845299,0.3426696604849151,
 0.9001665797318108,0.7678985994665524,0.47171616154857754,0.599126150770369,
 0.7298359643218055,0.3917633538940054,0.20838056926946236,0.1557285922484697,
 0.001525760775070828,0.012196534879562515,0.4597005446647361,
 0.6533511277732409,0.08623067452830802,0.06211987924973239,
 0.7194855322505509,0.48340769776367476,0.9229460893362563,0.9491467540604511,
 0.2352497922918566,0.9932365395635794,0.3964512324672067,0.8856672031353569,
 0.9978038903323814,0.3489105507652127,0.8299652723900023,0.8032082294967515,
 0.5167618153680118,0.654472365336497,0.9161881614451571,0.08659577605173696,
 0.9966897010092182,0.48257677150044465,0.9647159824988449,0.8303872294279158,
 0.0778400993888041,0.15224397021043345,0.32726299365262446,
 0.7274370940836186,0.15598688323277,0.12637779194809473,0.12532157175444802,
 0.3828743365397803,0.820699204855235,0.2800964603896867,0.8536160875705789,
 0.8808569477978331,0.6699249261410107,0.4068866714160151,0.9481292311262974,
 0.75160484146447,0.4881171730542182,0.5152244062611221,0.07879436247615923,
 0.9309108653494241,0.1270981699797713,0.3584809922947736,0.7577020993042616,
 0.44191559893641974,0.8282054049081728,0.6876978142241125,0.6794563820536398,
 0.9010691437077536,0.5788915402735304,0.8082934480850088,0.8810333029304142,
 0.9980395887465432,0.8784610707391693,0.7036522759970698,0.08822125750153564,
 0.6921626459181165,0.08065807723167084,0.5327695791737503,0.6737952234233647,
 0.2572886707717835,0.07296683141911542,0.7558039826145622,0.667775173191624,
 0.6680018356735917,0.6313661865015172,0.20511593623815627,0.2032580722922498,
 0.27373930330378227,0.13454086834685963,0.9736024931599049,
 0.021822710908201204,0.374951998426313,0.9202869383553139,
 0.26347025626811493,0.08031161242141549,0.5628223458869682,
 0.3546009938392176,0.008286438205848734,0.0451253151926404,
 0.7641410491605208,0.49482986513364435,0.17894347318908654,
 0.3368706220662223,0.9686168342386725,0.6478943896794598,0.36475929444755617,
 0.356713196892502,0.9120933939693041,0.3050666704970205,0.8357829950841577,
 0.320847615893016,0.645446526769767,0.7497566507298613,0.2571923794106604,
 0.3106751644188933,0.33503028012302627,0.7695240556162434,0.9020273115110501,
 0.7447569267663879,0.7416474794237681,0.7469211971631247,0.7173457334655328,
 0.977061767284116,0.6066311636927578,0.2612901899354997,0.6005510769196196,
 0.977419963819122,0.11872318324539932,0.06903964533503204,
 0.18055005696390425,0.8166827128116521,0.21064736141084894,
 0.32676440252683625,0.758889051300029,0.42344547664318666,
 0.21671732462679127,0.569842778381904,0.0373095238880895,0.3423171494233981,
 0.29913928532623446,0.8779956743314469,0.6039211982335236,
 0.48150724244064025,0.6118874922611439,0.12105336671371969,
 0.31847522884887125,0.4716715549108482,0.3836867214250306,
 0.15745244042979045,0.19152468283428548,0.252314856274669,0.5703934480263664,
 0.6026788158852115,0.6093374667144077,0.34781876899244424,0.8164384021806254,
 0.46764173848699686,0.16056813490230537,0.05706679842032614,
 0.24559797778692904,0.43974529241042504,0.9732186951267644,0.861293425861325,
 0.8928739059906744,0.11597323400404302,0.41158348040021575,
 0.7649980579345189,0.9317145577909807,0.478100918338232,0.7096667138209947,
 0.5355506949152629,0.5462841722511423,0.05173843804177214,0.5818012077328799,
 0.4781946356963295,0.19580589148071825,0.6543790073557688,0.7798537241666605,
 0.6429966661884923,0.6070385487595475,0.17890569810183965,0.8102382967440643,
 0.4856263209272983,0.9292984446387769,0.5749898927994844,0.1740145594717245,
 0.5167788229185152,0.5966203151764558,0.3601806476316044,0.47131354067775844,
 0.7958371863506216,0.2856663349616382,0.6252832847751087,0.06275669202796763,
 0.5610750469353031,0.05203585202563488,0.20282072008599972,
 0.7795219099206229,0.09808304666242218,0.9021407014876174,0.7764229398557125,
 0.009882449862913312,0.48151458492283084,0.4989731290894237,
 0.24088597618701968,0.021217198840218776,0.8987601435528269,
 0.5547432151636392,0.3185583348253479,0.6145973710742306,0.43527919963373185,
 0.7437300002722099,0.1637248025685888,0.9991868521817732,0.9946219966805955,
 0.326510218883464,0.741716299179234,0.10738339529785623,0.3217319935607338,
 0.3477139010367155,0.5222457643599662,0.3857631467916327,0.6799431574909302,
 0.776676447188256,0.6143871513203476,0.8254005173164947,0.7577330175316725,
 0.13555454119717036,0.017710358379258162,0.0738016725139492,
 0.7325347896115977,0.9569541392820475,0.3449175555716293,0.9119335423779555,
 0.4051046811006924,0.6389586529833797,0.6320501081181991,0.973972688989231,
 0.675944585991707,0.3181385084906594,0.7822249489933757,0.8107709882765346,
 0.2579597453492414,0.046825546509963734,0.9482744360323534,
 0.27676321938889836,0.1185498278620234,0.6578710128938178,
 0.26351522013661244,0.18635483296474675,0.5805463531609751,
 0.21330128710318685,0.23194028244530607,0.7906386169239086,
 0.07653977260553113,0.6271369670990256,0.9706275506977251,
 0.13140591988838357,0.14689671109877744,0.823377931545017,0.7520905729420369,
 0.054974059639839146,0.10448874147889509,0.43770630930922827,
 0.5021718895044203,0.42738314256884946,0.0380434567472826,0.5327551361616821,
 0.9800030931735138,0.43863212822739595,0.6563104561542166,
 0.22806417115372146,0.6897050476757016,0.6121166841813632,0.988207601697552,
 0.9044130377780077,0.8717962753338133,0.4401878873614594,0.0448961754280931,
 0.9086209243518234,0.549652844304485,0.17584623312651315,0.32563025798765655,
 0.5627966639085967,0.3599472509719206,0.05632953792792195,
 0.30054086373564803,0.3422273540585654,0.9812670990396257,0.4396566062135381,
 0.105985881507345,0.7130547095140176,0.44648894887113433,0.47526156755749205,
 0.74891445616197,0.8042614222205425,0.21286273671576295,0.07523373633904407,
 0.2538484615208533,0.08722504864224456,3.1695434100409425e-4,
 0.2707975357118375,0.5246788703326324,0.6498842966648709,0.9933342717285659,
 0.2893139989308643,0.5445321010459514,0.749222118146441,0.3903989524386762,
 0.9932962336146651,0.9127046944249507,0.3294124146508257,0.593864417362248,
 0.8504477044652898,0.005998621872298604,0.15197193238810713,
 0.4651624904010394,0.012659297388127655,0.2885866002877545,
 0.14160533069611525,0.8035126791559604,0.2077244701846127,0.8181565842831383,
 0.7025718491755792,0.6848397632846396,0.72928976302703,0.9307445195882377,
 0.289334250691782,0.7982759029269352,0.642685400845076,0.9615621599431405,
 0.18450513745514086,0.5169323803057557,0.09102302349837854,
 0.8526885458580856,0.9300428500943846,0.13606011170379284,0.7702247408058692,
 0.1253633094289318,0.03695275903073536,0.6513254065853206,
 0.49733311816180303,0.1262901603040969,0.7582132404770676,0.7252685835295304,
 0.6974283300302571,0.9359431125976891,0.6289581058926419,0.0194861059853908,
 0.2429343967502713,0.10413357968492265,0.005179564730152464,
 0.10391154180267637,0.1664904890980684,0.8927265639754797,0.9365045045916363,
 0.17071862352973177,0.2692135471195044,0.9187692198647778,
 0.07541799165390872,0.07490821566894756,0.477555027454332,
 0.13082037540983593,0.933970069998799,0.770447485630442,0.6801780869586828,
 0.5906768792601784,0.25450747314357736,0.3708574015738323,0.78752646373469,
 0.1105260326827413,0.13980504877820454,0.395178334090884,0.7343294910022848,
 0.4885821273817039,0.7013208841749328,0.5027904585055073,0.644807379028236,
 0.40877787472085203,0.45041282438554975,0.8437850822037949,
 0.7060238262856139,0.4501749398266185,0.2206765309304335,0.638785688765833,
 0.8023538262449291,0.478710890782603,0.9582588002002325,0.01766698133316491,
 0.31764992251504254,0.15761451109857283,0.5772031591519232,
 0.4293053993399215,0.7549364639969585,0.35103755787214563,0.8490452842654054,
 0.46251335446891295,0.36985575512417357,0.6058651629610073,
 0.17028202750085608,0.7009225996811211,0.7933040257024168,
 0.48446377624513615,0.5275048515756707,0.11894455073489407,0.772370523376293,
 0.883635620779333,0.41023183339341607,0.8860980984776916,0.41840914896538306,
 0.3293467864628966,0.22092923650086072,0.044601363335318944,
 0.6417806827385777,0.051497771770165235,0.42638540988764784,
 0.5284732957401791,0.4945362341855776,0.4920166349310727,0.9385344313917238,
 0.0012294072493314046,0.08567919423933645,0.9124033557924457,
 0.804895044319681,0.4942776986489651,0.5546939291019826,0.8028135030910958,
 0.47433058062929456,0.7410664180402885,0.4128812193988821,
 0.09710414959992597,0.22892167176728795,0.9924300675506376,
 0.28986744396560327,0.5152289082355026,0.3056555096972462,
 0.16391275491286217,0.13540793703274323,0.7378151072202445,
 0.9705135967497643,0.8079991315764862,0.21852939634579494,
 0.02198927737312062,0.11022392278057103,0.8740903444221682,
 0.29939448960844095,0.2177001322502442,0.7797421371383628,0.6269361239560721,
 0.3667127178527857,0.9751463082455742,0.6802175418623357,0.2233915679860039,
 0.3303555902257844,0.6866455855436522,0.04297243287615082,0.9415182333880633,
 0.6380518396526631,0.2747988684600462,0.4147847892554667,0.6342228362501924,
 0.5154275582680355,0.8920050622154919,0.1941042850337401,0.6382116935383079,
 0.8344368478609104,0.5235293159417805,0.7144115520744908,0.17864053478628417,
 0.42801406705939793,0.6362221048533713,0.7765585790303303,0.8183464471815836,
 0.6297882256332823,0.15488908178866356,0.6737126317319643,
 0.0030463504223153803,0.625301814711476,0.576003008903928,0.5019346153829511,
 0.7598384663811678,0.1854917948862863,0.03180677308193647,0.9124043441633671,
 0.9446091982901703,0.39808958772367786,0.5099170946630605,0.3087120461417501,
 0.5288050082362186,0.6326642426091502,0.7157837417290481,0.48613080283304533,
 0.9493507797000296,0.41507974667199876,0.5149915794229746,
 0.16414188867895052,0.10808738372080806,0.46583890427784946,
 0.45459109823020505,0.7934052943528006,0.49132677402383484,
 0.7391194004148826,0.7633233266164917,0.2302943167675665,0.22881625639102288,
 0.694727802188788,0.7855431506845405,0.9774840480014493,0.07198033073273868,
 0.4073550423018909,0.2356700819099552,0.20446851259368848,0.996008979245443,
 0.9842047488687867,0.7974019151393537,0.6455134076183661,0.265039669789498,
 0.6300853833053959,0.21197896639615021,0.8565659844170922,
 0.29168428963982895,0.22323037219940656,0.17592699307716353,
 0.2874123739950313,0.6294505998007038,0.23165924649566527,0.3033248757924496,
 0.44898412860361614,0.08954195053317293,0.8420956163710493,
 0.4010926748407285,0.4098197838938522,0.7069985024260763,0.962577817393682,
 0.6208572376823692,0.21438657406692652,0.958748806841141,0.7142043241966901,
 0.4484896589564883,0.7363071706564569,0.04808330683793027,
 0.24275883879276983,0.765606013076051,0.7769186079937889,0.6003267161987342,
 0.9786894416447771,0.695544494196666,0.41553748184562145,0.6096861557979857,
 0.04178071154534735,0.1163238419847551,0.8991120688099214,0.3688745781957824,
 0.3313714065503188,0.44111015606143,0.8156424299552592,0.8699605902390274,
 0.37652054142776614,0.6980479740313786,0.8752003567560278,
 0.42907303021144794,0.4503747010554823,0.9433480686642978,0.976524687505456,
 0.42937077436605087,0.04191386768320571,0.2806900401711323,
 0.18770544649063403,0.1293820025975202,0.2597912321566783,0.6793359328853776,
 0.7274112316284845,0.765256799688021,0.03898905175332279,0.7897060320486414,
 0.7024027421872154,0.7485969842320878,0.5363228419490448,0.3980962742325369,
 0.26986946513431265,0.4970171178435261,0.2042079225504555,
 0.25029140461882415,0.5859010518808456,0.26273371419389013,
 0.34786161336803745,0.5744364757375381,0.48629214428273115,
 0.27781550220075435,0.8934680303953283,0.1893536657676349,0.8898576907389301,
 0.569870606681777,0.4253833852964021,0.6315977247341396,0.09357163164581983,
 0.11166222775289791,0.635039628385354,0.328996070731867,0.3031735380442687,
 0.05934580395982736,0.14147205328507628,0.5434598795465949,
 0.8889153241491492,0.20093501212628895,0.3035693619642328,0.9963721765222591,
 0.9933119480753323,0.051274736719666536,0.2787883121733946,
 0.35266639381117904,0.057370739091514134,0.8096651664732093,
 0.17150312824115344,0.07738750482292145,0.8221922877813418,
 0.9341643303827745,0.5280264518165864,0.6297338441832423,0.3547667528745252,
 0.9610475773118053,0.3501568178699753,0.6267522169306816,0.9707556242165334,
 0.8942500417866845,0.3480940737050576,0.8555129350036289,0.4242085599676926,
 0.16556632128379978,0.02701484685125488,0.686540766685201,0.4763840299722052,
 0.4864367973628463,0.017300833843079344,0.8103157650376388,
 0.7715277504197073,0.6019356548366014,0.8383277789532073,0.9724204142966885,
 0.6251958462174042,0.7115780692600276,0.9784155067492121,0.554953753866095,
 0.8686470464648384,0.1868382337291412,0.07789654256254264,
 0.011933423136184396,0.08478374703108271,0.22638969358638872,
 0.23949506424126055,0.8804112151301804,0.3462231266909901,
 0.11316281973142184,0.4151980588130021,0.9533042710612671,0.3952424046972445,
 0.5354245687034065,0.8819429795855633,0.46229049854305104,0.6167240450849145,
 0.5745663323186647,0.6818253474122883,0.18971393656823055,0.8262020029696567,
 0.20653216111964767,0.7780862598159523,0.39428730391963396,
 0.7774460637012974,0.7191251780556398,0.9214511615386168,0.5292003472502642,
 0.07125752883325887,0.24828856887442174,0.39939214435911297,
 0.3491818826384505,0.8356976994151941,0.3805069515744989,0.8552858657635767,
 0.45853524481607766,0.9765829294696601,0.5824558581209054,0.3157128718937795,
 0.8428972228610516,0.4108761089053047,0.31846424001204165,0.4121270232650398,
 0.49298080256587873,0.9036516999199999,0.775368207843823,0.418999563888113,
 0.2670887165058936,0.06892072121486725,0.21305048987961595,
 0.28261607729679084,0.48403143211439903,0.24467739712110148,
 0.8181543634975007,0.627324636351377,0.8967926609273351,0.4548818717183213,
 0.1618684814083814,0.3520188680427485,0.7216805618103767,0.5340868406470846,
 0.7192513637782504,0.18986331111175458,0.532346067460819,0.4876300334353648,
 0.8270234661090266,0.11676965857816934,0.3434007314470656,0.5246013481245376,
 0.6012143647096228,0.4530909942759864,0.1406943342139727,0.586580383573409,
 0.48702138668869377,0.8150973390452726,0.36369536819631887,
 0.018596432822228337,0.28068914180366644,0.8374044068132949,
 0.3176610752086293,0.9473390089908177,0.2004378470307694,0.3607925050971993,
 0.4722840532405508,0.5757765259426244,0.7842537830565285,0.6739725868503005,
 0.41580371817487327,0.7950064409867865,0.9034224668186029,0.3793145789166418,
 0.8605162698103906,0.6104098535698854,0.5148153855908405,0.5869620615964195,
 0.6679200478464092,0.5078795969510111,0.16575278954245953,0.6165186252402522,
 0.8023088714063905,0.193562659302153,0.19831252760082818,0.8723241754372043,
 0.13177055745247623,0.023960904606996847,0.8638821943705709,
 0.6594008350849317,0.13025570566824907,0.979442668461304,0.2275459240197134,
 0.11550386030181903,0.2293839678936087,0.11174698804007432,
 0.08420336334502754,0.4327648776055346,0.25985193925537564,0.668961833107683,
 0.9158791188732783,0.2457052145671158,0.6963131099766066,0.36997807482717104,
 0.7678076359374546,0.5260973018082463,0.7423731603803485,0.7901693363235104,
 0.4261911563618731,0.24711046098247724,0.6702627517483777,0.8962795152612717,
 0.9507520735862316,0.9458737473650989,0.026814661143917173,
 0.6644790778286748,0.25929484495866406,0.6900914986696022,0.7558909264087599,
 0.5398739989908425,0.8019856695957641,0.9515278020361656,0.02373616175868498,
 0.20805766889408783,0.43700052169299286,0.2215613260661662,
 0.09529690042899175,0.9207873265787271,0.42957871811297277,
 0.37348518418754995,0.7287722703471111,0.39863711214363395,
 0.03002359313056302,0.7958475168483423,0.4115433171175382,0.9230086714261816,
 0.004539005366080984,0.9660237618477527,0.2454220394791281,
 0.7438309857910771,0.12558503359116713,0.6882868863530524,0.4570692124999387,
 0.9944546577531053,0.3708478397740683,0.9331846597244677,
 0.024950223547385186,0.6249371348500352,0.36763446574043257,
 0.7480205477351439,0.42113373936145737,0.4293566370240003,0.9625393577042666,
 0.35336922655493974,0.9320563343585702,0.5157093811312303,
 0.10733220216513695,0.014471888387608822,0.8889252987127351,
 0.38774702915436066,0.3191623013994822,0.6680621277253229,0.4265633706561396,
 0.44605751463558363,0.43535385937967164,0.9945296390121363,
 0.6920711886961587,0.2543700259880042,0.867614132692741,0.28259580136358,
 0.7997413625555047,0.43897270833010604,0.8633899615624563,
 0.034588218418538985,0.7048313949892296,0.39415119904135243,
 0.17778324866670536,0.6448239634582392,0.9518934117673987,0.6118354990555169,
 0.4155000722155576,0.7069942117853444,0.16014931326775195,0.4474747307918818,
 0.08659399707479165,0.12861245525060183,0.6716984028337757,
 0.6588804320869017,0.7530138597639486,0.1510573902240282,0.577418203023348,
 0.8627073447152067,0.12074334086699445,0.09941666329275922,
 0.038196288933347144,0.866919193620155,0.6431758060444339,
 0.19643283059251448,0.09421213846923737,0.6609007651540615,
 0.8275610305221186,0.5376411666731176,0.935766815461201,0.4512446066363127,
 0.021167247721430305,0.07278164292503764,0.19003278518687006,
 0.4122040220394584,0.17236956010151205,0.8613741860272004,0.907250790450475,
 0.9618941078974581,0.4477868161790194,0.9801652414292424,0.6681833607689353,
 0.5173390489693187,0.3925498049992159,0.1463846819710719,0.5513768671218433,
 0.11617326978800102,0.6573715540383193,0.07525398805683936,
 0.08373568095191108,0.5188819815676421,0.15879495392710952,
 0.25143331271906366,0.2818295357730287,0.46473706091349876,
 0.8650973558959647,0.7403121291748376,0.7708746579333189,0.15989895112854569,
 0.36228544925980655,0.9244694733365845,0.5392224725020467,0.7902592581300177,
 0.439016390912114,0.8069224759140403,0.22411949597679204,0.4645452181051143,
 0.11704581493866661,0.8897940951669701,0.25124462770333345,
 0.11976995383764577,0.7874725000083383,0.8538477468618741,
 0.21586169997713556,0.6207948825644767,0.6313259326268614,
 0.43106037070437875,0.7078302516400349,0.5519847134617893,0.810690621188564,
 0.625046959808314,0.7861274521271459,0.499708397854318,0.9932160149065654,
 0.9019792193255352,0.16484546403501354,0.10834024413151833,
 0.5502565382059108,0.9397505345679379,0.018997079213301626,
 0.9901514457817069,0.34562838099129534,0.8304029623968237,0.7818029419209906,
 0.7127119366564565,0.51875568837191,0.3354424505558733,0.2592782699011339,
 0.5541317546447042,0.2089427834477573,0.879408173325586,0.49062673617747565,
 0.3347014073101269,0.3446409125016515,0.3929235658160094,0.5013473069189093,
 0.853754913656706,0.6680265128830963,0.117071849721339,0.8769972456718839,
 0.536014196404348,0.8826321747058038,0.5975902812720002,0.2797857188355444,
 0.7465368059037675,0.3694305175896836,0.5988194584511916,0.03439494533625542,
 0.36745266869349424,0.9382401393469155,0.6231465491412541,0.9355023459740058,
 0.5855856194401451,0.7790510758221794,0.7039104030405721,0.7732278869930013,
 0.08625067486232152,0.7520555671828124,0.8376538372463658,0.7629988801302012,
 0.8430880416800618,0.22295185186433064,0.25525997920261034,
 0.8245851439650507,0.34217839630879465,0.338306182845555,0.26485550739731945,
 0.925129809672403,0.08842419417227976,0.6470232872786751,0.7556604098051126,
 0.3351973777035332,0.5271810289951181,0.18998806571525484,0.3572839892156007,
 0.8620828475463844,0.6071449163862851,0.9685288447378998,0.551451552236335,
 0.9670323214062424,0.77255820129214,0.15899865112402467,0.33118869263374284,
 0.47769856340664696,0.6311922704418643,0.9794915066726695,0.8409225202989437,
 0.011271724148630202,0.8608559068650787,0.7012103796398974,
 0.7604605330023936,0.20121498394591952,0.29750546380429876,
 0.9121278229318062,0.7582166730410657,0.5745742658688584,0.32281646451762436,
 0.23460901801921474,0.8824576148520409,0.6301938464074155,0.3033242367873351,
 0.2631175557682508,0.042492113556472755,0.041831320034534825,
 0.7821936141830117,0.7378220176602508,0.9250439649176294,
 0.036041272531101676,0.31579515457699403,0.9964048801352522,
 0.30004619893876505,0.41324634120959247,0.7836806860691542,
 0.1821990123119871,0.11805701243911249,0.4271850198392019,0.6508394510641051,
 0.24879096066080464,0.15756257139893248,0.7857237002194204,
 0.5509667668634973,0.7166584145639219,0.9451019699079056,0.40225652173503823,
 0.38119578673779175,0.4670567963037269,0.4060937252253173,
 0.06015299591759393,0.6556864456336086,0.16097915123457307,
 0.9081449874123511,0.3058105079425716,0.13299598098422938,0.7953695153960622,
 0.8426781714112078,0.9857956604760298,0.8603618811822777,0.977612098527956,
 0.3502373216157483,0.867770113151065,0.007490748548993231,0.7531010157122046,
 0.7586363539502328,0.4357619618260815,0.027551559538987913,
 0.04565300906312597,0.734532222163855,0.49985732838114394,
 0.30065539376198014,0.10640797974158334,0.566461549609093,0.8108582774303572,
 0.9912666481069311,0.3653371410658274,0.7169231845303599,0.1288321715423386,
 0.9334378334120639,0.07152435630217335,0.7813266610610915,0.8683815986507817,
 0.3057613630939131,0.22573549272378557,0.5021516050968161,
 0.15869784757408456,0.49648623537159053,0.14236667518382862,
 0.722814284211027,0.4396740838185562,0.006456000918088289,0.9540541623519644,
 0.26806232672539176,0.11972060696975995,0.7995871738834854,
 0.8055731978964357,0.6641674434892763,0.2679675281592284,0.6886012568498113,
 0.15235617572039872,0.15387729284600127,0.9541619137934562,
 0.023076348907375532,0.967260636088283,0.7424032758743638,0.6427270808092624,
 0.09796431249391757,0.09985929722010534,0.12445143180366702,
 0.43677085675789185,0.8368571775382033,0.5307344615741818,0.5902092515057612,
 0.7172235603234346,0.42414457147882545,0.5593234027450473,0.9839059871956017,
 0.6433145966795066,0.09244911343563877,0.2899603283513523,0.6911481516883206,
 0.16608605660638387,0.7731071723322734,0.79735419175886,0.49393240908719616,
 0.46114041139506146,0.18698977679676743,0.6106924311345745,
 0.9579196561245611,0.941143270488279,0.5956166829970154,0.5354398242286488,
 0.9362751395232719,0.07555363707142515,0.8942039764879385,0.6169877558162916,
 0.535116778990661,0.5741081339389091,0.32789728762900683,0.21846957034103753,
 0.4489753868433394,0.771856485947456,0.08024299619576702,0.26501778774920726,
 0.9391370696489991,0.9383991964984398,0.6828891178735069,0.08580012045585284,
 0.23996705504854,0.6578803942908524,0.9855989424443615,0.9931006629904227,
 0.9796882343311811,0.21251761357367083,0.11186538484148911,
 0.21108793056238673,0.874739058265235,0.43856078210482896,
 0.35712064060095916,0.3304858146714136,0.7896087495783394,0.6372351115257913,
 0.7383567974340868,0.9452309877802946,0.7017556553495157,0.40255915206033555,
 0.9645200918422494,0.7215831039717623,0.4291911509050943,0.18698666578462164,
 0.4561564985439527,0.5276999502822975,0.22762764222850806,
 0.15261715595319503,0.004283501389971023,0.5463068679315558,
 0.931457072184011,0.8051613920831897,0.1171937883999733,0.1348643797801201,
 0.9945669626250621,0.2386067447844128,0.5759163624162609,0.4962583042507729,
 0.8360580275759564,0.28403602088244484,0.05418919497326524,
 0.015255818226986406,0.10570897947257074,0.7416112643396281,
 0.31269178055596014,0.5901559243043122,0.9715574832914359,0.4334482096854454,
 0.7254392497872663,0.1807652114666588,0.6194773697649401,0.8018820039635728,
 0.3860983082481526,0.14884746257121806,0.06283158968222635,
 0.6006015524288578,0.4596795377623133,0.9367337505839309,0.31126233405492176,
 0.2708211695402819,0.8670986087747612,0.8438871110255308,0.10698940916643962,
 0.3821671100808114,0.27635289286939724,0.3452976813992652,0.4112158127072705,
 0.8831098500829181,0.24878487867886423,0.6327350815665783,0.5465199282834863,
 0.4929618971807901,0.027254160449814213,0.735695047518611,0.2673027994104207,
 0.5943881635400039,0.06258417449214149,0.7202207030998407,
 0.036088173798442647,0.8058783321019657,0.8639925722690248,
 0.11796473699141752,0.35975093819502446,0.17746663805895713,
 0.6696642348705096,0.8721802833307373,0.10540010868801963,0.9991560697887862,
 0.9680373541381947,0.3820253024389513,0.6657491189437883,0.433501462638873,
 0.1395163213310633,0.7780937651974759,0.43997194532790984,0.1636062239075743,
 0.5989643643059827,0.576354729789296,0.49315486595484714,0.6269716426097848,
 0.25430130827802644,0.5746299489103341,0.8976516559665927,
 0.06690352322798987,0.8123118286390737,0.8781840721719358,0.5258178713174356,
 0.8837428188960895,0.5417541170150932,0.7735032798519513,
 0.005563582095095554,0.6859752319859221,0.5624005435515984,
 0.4877396903124287,0.9000406448219881,0.4643586365589605,0.42805084991224573,
 0.2969249439517807,0.3056474046171769,0.15679427487124498,0.8215423558007914,
 0.9169717241069731,0.11656571852356712,0.47018372320805923,
 0.32011308068596966,0.04726225864139433,0.7267935014224349,
 0.6214976728650916,0.02602853831993357,0.5348814889377972,0.8451391161232086,
 0.5427597189386544,0.9395763488678013,0.3941037706429109,0.8160835655895637,
 0.1538287947824729,0.2609043283999555,0.49789260262125157,
 0.21723732631638792,0.1863576036514768,0.8556896265596083,0.8294871536881988,
 0.7651469565787328,0.4539262494030668,0.11546065905439151,
 0.14583549026086384,0.709474214722842,0.8706174700544351,0.3205123465064317,
 0.2898961147181959,0.5742614633695016,0.22338116900328164,0.9813504499783904,
 0.5853991218219461,0.7961820379869928,0.8036903150529662,0.2505714286520294,
 0.21407433146190835,0.15402511577525968,0.4952671118755708,
 0.7548022834003338,0.8633179041512332,0.36486360913250926,0.2421541736652819,
 0.40093312902309064,0.2823337595293186,0.21901828709038274,
 0.16964041718121892,0.139848379181295,0.8858978658167088,0.4729539642739491,
 0.6209406821063727,0.26414883032147396,0.44106558524185013,
 0.6185690896281293,0.6159725648900498,0.6963933294380605,
 0.044085866149849995,0.6807674829000803,0.6815626396013533,
 0.6041670959294715,0.8768740702762996,0.9409815357963041,0.09692873538027302,
 0.8619197053325476,0.7992267940060598,0.9515924454840929,0.25447348850091855,
 0.34817487649139933,0.5812542434658701,0.3389355918514029,0.3973830633191463,
 0.6975984101309614,0.35323227402424306,0.40441385501739524,
 0.07101399285983023,0.3166112984215841,0.43469273598599256,
 0.28530205165300226,0.804389091561277,0.10187202868869538,
 0.07994948148820835,0.349278022050006,0.5894193450545651,0.7031252554211599,
 0.5868305112023536,0.8693424123595865,0.8389901262113344,0.5064670230939199,
 0.5341137739906163,0.6549959402246358,0.21830200683773204,0.1257085293950394,
 0.28484391353376004,0.5157991853361129,0.4896844398528847,0.5740041493024547,
 0.5619005640971024,0.5581816948768925,0.832001562705116,0.6779417410314539,
 0.9413340474854851,0.2736216343329264,0.04219991952523705,0.1488304446161005,
 0.9125678521819799,0.7249304512222714,0.39867053171797706,
 0.48081564892246875,0.5709454234473914,0.8021385547676353,0.8771272809200334,
 0.9687327062951412,0.16529780454358067,0.22734393574807477,
 0.7723468506850615,0.47554208777630813,0.12927309016379462,
 0.5183464080593079,0.6677740069759464,0.8798624799808101,0.2534650394341561,
 0.8230203478542915,0.09356648495218367,0.7701993700036872,0.3847333825742443,
 0.637761149195823,0.026692105013689238,0.16166739488067805,
 0.1780088968463316,0.9552351033237546,0.605085767391066,0.8642460294895647,
 0.18954175077382418,0.640012837946133,0.21609716858823735,0.6184092858684591,
 0.456190042475336,0.4340094275413582,0.6312200234989915,0.9788246612539804,
 0.34723355659404287,0.9579186885153526,0.04836457696430396,
 0.18247648070363276,0.9667260999957037,0.44697861928669225,0.215154194591071,
 0.776487977782435,0.7191370594393565,0.7631188877097317,0.15974722878613867,
 0.8683250037564465,0.07306520111895154,0.2314682487853169,0.1710495659778215,
 0.9795736010909912,0.9446015847070202,0.5783488302332902,0.7870852246037509,
 0.3854413342007099,0.1325148579331763,0.5752356113694935,0.43706930945231914,
 0.00879029980762458,0.0693952193767462,0.9752180451977458,
 0.19240365170151896,0.49537779912178403,0.7243848578002163,
 0.24807851837031003,0.47354135975138023,0.3514244326523659,
 0.4150734886770556,0.3560962857988935,0.4197323919713347,0.14693972822718093,
 0.9255337602228407,0.44560517653945353,0.049536399762204586,
 0.9742055776595416,0.04324506737061895,0.9077039483529686,0.8655192228665238,
 0.8541257144848476,0.8955432898292848,0.37038159499958234,0.294514936835242,
 0.2987448469784063,0.26661641760867705,0.008970467005099292,
 0.5751410272531292,0.6602332930359307,0.6195413020094216,0.257838182554028,
 0.7263264986979441,0.4061988162994523,0.9809616274294495,0.5142404880058219,
 0.1927405617348975,0.6310079739702608,0.639972681740738,0.27406910741804424,
 0.4282041370408316,0.5493330728930353,0.333824085432619,0.97342797144266,
 0.8891577179513377,0.04049705391224667,0.745233122401229,0.6338910962115754,
 0.23701482062148016,0.9424876098422393,0.16743887016533876,0.796628816471611,
 0.9129138918041997,0.1740876490246438,0.34586306065170225,0.6629050712369178,
 0.7295249325642598,0.34059594005016636,0.4499326405226156,0.7658746043214485,
 0.9225928696967401,0.2246548441046008,0.9471857915709537,0.564006875632657,
 0.7231641389739422,0.9971760392938207,0.13409544483740177,0.8235593903358837,
 0.2905452729464877,0.09241529662304093,0.7766449554122345,0.35858268637222,
 0.22080538377526948,0.5328048905443337,0.933645901032014,0.973360816319373,
 0.37379377913447376,0.5385183547753191,0.8778317296062941,0.7745277590662029,
 0.45076873957930763,0.8531129594433683,0.9625306269473014,
 0.03421590854501855,0.5091746516437291,0.7908493973910159,0.1371994343521541,
 0.5183152275340049,0.15055014435415492,0.8590157611862406,0.7138803281595688,
 0.11238251316119363,0.8028081904839908,0.7171032797507658,0.804528248766347,
 0.15763572417668492,0.6780948607353967,0.009191169494669471,
 0.1424754983001344,0.04814947567751926,0.03335128643477181,
 0.031159150796319923,0.422537075434524,0.03549656648150168,
 0.5583439697728814,0.06574804346202279,0.27702866553291905,
 0.8839244529999357,0.9493633685751838,0.25867896514840794,0.6916342446850249,
 0.9865596953392499,0.6706232981910603,0.8399190901115721,0.3690738735888036,
 0.71284822829682,0.2500293136502314,0.008447038861929657,0.0345701452611894,
 0.2944896484818895,0.21689322322424298,0.23515127428800797,
 0.5157720654640457,0.4949915632392514,0.497961523111382,0.7187316947025504,
 0.979566865268067,0.8899400732296834,0.4206597241850537,0.48429291014536935,
 0.20421628855596008,0.8591166926525794,0.9623782857244646,0.2778253019626439,
 0.3878686999150113,0.8923001414751418,0.10020264605976126,
 0.049621277438422595,0.04176621395496183,0.3767708857248636,
 0.5567859708127632,0.3450444426176602,0.3062172446688781,0.7551209857199775,
 0.10910909246999767,0.8279029909768323,0.48047662035282834,
 0.02847646863232689,0.9265021886602826,0.4988117639219327,
 0.05765773215817871,0.9493012077657776,0.7719775967620626,0.5640638878480133,
 0.7045370743844267,0.9233837610926898,0.2842716487964114,0.9062463852806331,
 0.9450323363529696,0.693375964249861,0.07389460941350068,0.9330536544830911,
 0.6343299230851391,0.9501201751296768,0.6434208214421115,0.7858175972292628,
 0.697799644484658,0.7671404782776795,0.1723624470160584,0.23323634120853776,
 0.003254012635058312,0.885440851335297,0.27130447644056543,
 0.5535893298834994,0.11052266973573932,0.11776650071927774,
 0.7996695340721235,0.39748696005101625,0.6619716136676635,0.7005578822637936,
 0.6874121562390032,0.8414838134739646,0.7465900795119499,0.8372855728989652,
 0.40589671984187725,0.6665839686724018,0.6555803709723185,
 0.18337934467106298,0.7025619121040245,0.02535430609811351,
 0.07594269813103738,0.0033493379881937013,0.3532553427914028,
 0.3861965405864973,0.5981899357214631,0.9820266127714251,0.9963609404638368,
 0.8363404015201167,0.4735260373062592,0.5487570908250028,0.996436084932851,
 0.5870064576229599,0.9146672596998406,0.5828985456197451,0.2447212050531823,
 0.3515101415757206,0.6602568238938917,0.13897910065818464,0.3017005519510815,
 0.3480705907913766,0.2639867436202178,0.8163875883780892,0.3243341022735704,
 0.36218110367790846,0.8672081685158018,0.9847346700638316,0.519123590148325,
 0.19901249791761544,0.9395897366071507,0.15057290485166164,0.537531894776671,
 0.44674862380520386,0.9833550797775714,0.5698016309540443,0.6766544836228944,
 0.49869011926607787,0.4901488267275106,0.2598029088382403,0.386744627170711,
 0.6736976920430826,0.8474472580691234,0.8201134421156444,0.39108228287824454,
 0.967894947376221,0.33733945011299316,0.11223532254083468,0.6988001501553744,
 0.8922042920885598,0.6623023686822446,0.12812799355166726,0.7788969387043633,
 0.5082279466492369,0.793085186251914,0.879989128627187,0.6286992577018021,
 0.16192328861317928,0.5099451701096537,0.8480210899579964,
 0.17731980700018246,0.012398047460872696,0.42451573049208347,
 0.3285505961175744,0.6016138830674888,0.6802161583435639,0.9087889179494073,
 0.17951268544505417,0.7152961478729919,0.8858586728175706,0.6129605687544822,
 0.5371839293115142,0.3089317422946817,0.263396844196895,0.7648927910865478,
 0.8435908450827551,0.5129049820232945,0.28545784466660473,0.902160962399834,
 0.3017593281045461,0.45696247692783487,0.08256049817997169,
 0.6254018231176334,0.2921847456370811,0.6157472673882133,0.09831731396076093,
 0.11704374927517058,0.6402769455256567,0.5787898756085096,0.6440088433751748,
 0.8444916299536538,0.823564639263886,0.028791910306066204,
 0.38789550803548023,0.8086665656331143,0.2588259943531619,
 0.22550637344461277,0.7402049712481684,0.6544300564378971,0.4197905256041887,
 0.892524191963387,0.30816432195321364,0.24382043809779552,0.6094311942403218,
 0.7938053203887064,0.6944377737421804,0.8729976562349745,0.7942755094342405,
 0.5995391736397084,0.5634307556677826,0.5195975847847608,0.7726308056266916,
 0.37997986390911365,0.12221316198735399,0.4245372656148507,
 0.7379709320345356,0.38515317842172403,0.3447295917439036,
 0.37075248884831336,0.30782435561008226,0.7115715143546052,
 0.7766368080076604,0.25643307304919516,0.6492760396315211,0.1834571302610425,
 0.4133990737192563,0.3490683465143656,0.763361824761095,0.8860904388926562,
 0.738069639419476,0.12083865430873586,0.33591677283667853,
 0.18604002372290218,0.016389276450796686,0.12523362746118183,
 0.6105553285657477,0.4420412975892237,0.8135675036118365,0.16013180777187452,
 0.8631324054454625,0.09984189877682792,0.06453202095554378,
 0.8658197864059143,0.6694737463124827,0.3265113495534533,0.786485825104753,
 0.6086677412321401,0.41626400430888477,0.20666323694218658,
 0.7886897596144657,0.4657628630833557,0.9937633447196931,0.08998548662627648,
 0.8919371405340402,0.44550020218383546,0.9818681422586022,0.2159388499172734,
 0.007076557330069511,0.08826169335765655,0.6126018564069696,
 0.163144802421908,0.45047582758726123,0.4479165040576121,0.31862951743697965,
 0.723009555335482,0.2266827554646086,0.20200452739374575,0.22839985511924932,
 0.029389500698076998,0.29681607871352633,0.41271465412163366,
 0.3126175834247409,0.18564455089978993,0.09893495225973747,
 0.44662956927873065,0.6545577721694469,0.28975976090671907,
 0.49513519652506366,0.5946036160274263,0.9332550718248229,0.4305496482206679,
 0.7710599873426134,0.892505176587333,0.2690259465443142,0.5414426966662111,
 0.34461051931825937,0.3361996056894575,0.015116711085120249,
 0.7681579245087048,0.4500905155852202,0.6026537463699543,0.8635106761412747,
 0.16806690968756755,0.21395809560018964,0.26313682871706456,
 0.29545127647992564,0.9163854430902648,0.4886692541514879,
 0.03994332916245724,0.046372738413632764,0.2730119863821494,
 0.22758526397318812,0.1191167743850271,0.7238950642178033,0.5586217542226748,
 0.2619737673853808,0.2470865953392395,0.23920690540967127,
 0.32722501466345033,0.32910193166440693,0.21129889305416139,
 0.9541496380406461,0.17538283774785846,0.3661277495660955,
 0.33999298097802255,0.9736091617826984,0.05061802847762065,
 0.7113987312937651,0.7536642284141626,0.8238238094707255,0.6762009419656954,
 0.649713113884727,0.5883006319319265,0.7202472365515178,0.343896114442094,
 0.4104032795974467,0.8522860687604983,0.029926896822716875,
 0.15107149206868153,0.1386417415505773,0.22630820076559277,
 0.6937928939997825,0.7388712112673794,0.9804752524314942,0.3534771082398702,
 0.022663213803854165,0.12878720517073639,0.059656580647084456,
 0.713527264582065,0.27556064079545894,0.4587334241690235,0.7277589103502577,
 0.2924482222450143,0.8638418501738612,0.2463567289031554,0.2604143225995129,
 0.9537261803580077,0.48024787010960435,0.8315088473706943,0.6094854820253068,
 0.5390687834596922,0.17614094333245522,0.8923025866882445,
 0.019532176461986794,0.09077289197686278,0.6599616760578482,
 0.03068477053351515,0.9290166876771065,0.7156250259638851,0.963706266871953,
 0.14910260065409464,0.7800250751589247,0.4256036416879583,
 0.24304373733466567,0.40987598722116925,0.7893778950237377,
 0.8524151161599319,0.10774531585227709,0.7581509721206221,
 0.19626662011970164,0.48830087941596134,0.02961264515835227,
 0.39137664747010004,0.14523016887051843,0.8876647189366063,
 0.02330341043348938,0.1282010614521557,0.5961315008051911,0.3777307026676101,
 0.5742911626114482,0.4512002217409725,0.746146875382557,0.48777142164741727,
 0.18942476525655583,0.14024207265068855,0.6868297889530819,
 0.10191057270506054,0.39265645081289124,0.9423125847196594,
 0.26059912553310993,0.44192750119672264,0.3570340342942704,
 0.6654835651886104,0.09103751291853968,0.9446808996046184,
 0.13621529682744637,0.24476332090221964,0.6958805758406115,
 0.4712667728244222,0.5473686826036225,0.5854384643061987,0.3501217697045158,
 0.3147009373052502,0.9977535372570377,0.5522612244396845,0.3663897187996392,
 0.21488101486740818,0.339340622474509,0.5694320200840127,0.22072265288073312,
 0.022181777376642287,0.9775995885233109,0.86374479302028,0.28760191683325886,
 0.1498257546053252,0.9417816388255719,0.5183538033122886,0.2941952149164756,
 0.6792673800120109,0.12006978854171546,0.0937445148289473,0.4863710643185287,
 0.6174418985226314,0.13088441195017186,0.8254452473516336,0.6786326974149798,
 0.39598378551238145,0.6378059218946306,0.5998043307620692,
 0.43886291213953377,0.6575567867549265,0.09429049742262952,
 0.6176254748610067,0.06068661031413358,0.9053113482847726,
 0.18015550517318235,0.23823276515490122,0.289365828183454,0.871793523043233,
 0.8399339855374104,0.32485897830086263,0.181032894252096,0.20049893040855116,
 0.7806200386516883,1.4385601341315724e-4,0.6775383019198025,
 0.13157654861568635,0.7845998240982306,0.07375989539201788,
 0.8755651174380321,0.5253147359153557,0.8141633758266922,0.22704005091950308,
 0.3676146990117819,0.9294055786354,0.3031076454778102,0.402532486577833,
 0.2866900276387201,0.7404057450899173,0.11187916574470513,0.6883845205095226,
 0.6412901683624697,0.7173227426280527,0.3014560236678454,0.04644488171640504,
 0.6918960911164458,0.17549235280414854,0.8639889909171788,
 0.18718983354512586,0.07215953599873504,0.40924711378965883,
 0.8300284528546511,0.17397682170239848,0.898143506206058,0.47331395383349195,
 0.3483940394917262,0.21462605269902202,0.45051978265803316,
 0.6294443605941935,0.9884234478812599,0.5389803055449879,0.2118420650452052,
 0.9227640035099218,0.6962506112574137,0.7041306380779471,0.5509721130329936,
 0.47058722396707264,0.06546935212495841,0.10540050071125906,
 0.6957204047471307,0.587686925670674,0.0860540972400824,0.5575437813316406,
 0.5805165401938319,0.6401205339170234,0.6976642519001586,0.9785804121750942,
 0.2810212504457622,0.023772618127194,0.6286737395618237,0.9518519928481596,
 0.03586355258335794,0.2717278801693538,0.23182692934705662,
 0.7525030188270175,0.2347250574044012,0.4699070862461119,0.21703737689702352,
 0.7991490333338226,0.18082931299858251,0.8124254267411211,0.0714813994774055,
 0.38030905613270494,0.5554009188456537,0.9628713530717196,0.9482116496024966,
 0.9801923419953056,0.7022676019270772,0.9018563588639588,0.17339798043455001,
 0.07648563141202447,0.7813935486428386,0.294378086878861,0.31437764395702184,
 0.42711403905404066,0.1789936638574131,0.394219625835877,0.48480536559597576,
 0.5218667279523118,0.39491012206550846,0.42167744218035674,
 0.3318534679033446,0.49774703151705013,0.6420715301595501,0.3341986175238766,
 0.5409164570050061,0.9895462578335998,0.4156449909481136,0.614319390108899,
 0.27477817035921603,0.39635477170167344,0.22141214825042066,
 0.8364407413809877,0.552441686816598,0.44782432805053063,0.13199151186772595,
 0.23109139063455486,0.7645218863633035,0.9558003362198844,0.9909580482117988,
 0.13355010629833475,0.9308100566198626,0.05013334249036183,
 0.27555136640098443,0.40817194347465346,0.9562200629057153,
 0.9955749443261315,0.9052724404063194,0.5602113117124892,0.6616217055839602,
 0.5436681867412707,0.7968471396489036,0.280225599098595,0.26533075883325374,
 0.7784607645170835,0.6085342637537526,0.8071719331215943,0.5690330342681056,
 0.6912116070438051,0.6571638117711498,0.3758624727877967,0.3744876793768698,
 0.47526407112981695,0.36679293817202163,0.5440474738786147,0.659092140427097,
 0.54338283033044,0.2510310781980868,0.8292195165317835,0.17132724644829722,
 0.534482660118166,0.4170549682165562,0.39050080793998465,0.44579477068856654,
 0.5530584899307215,0.29023902687668646,0.06298069368529524,
 0.3602131011817279,0.6822452712179891,0.6815783669223793,0.9465883144925678,
 0.42191616139177357,0.47454128495261283,0.24276311319654253,
 0.5386812734529495,0.33078816975810965,0.14386369875956562,
 0.6627742972127093,0.999849994017754,0.6398853420037546,0.056184589531747764,
 0.6263376339493812,0.9969065317923365,0.939209090000918,0.7265677614554529,
 0.3398623898364883,0.9767101359413288,0.7648809431728707,0.4651144916336861,
 0.862433193952645,0.533700454589004,0.1938052102824649,0.20446023196463636,
 0.11091878347405171,0.25547009533522,0.13036400687184957,0.7740342259325231,
 0.5134872669158482,0.8871733094651428,0.6699517164504307,0.7985835820973354,
 0.03719949565697922,0.7020923999672223,0.9837831289293131,0.5318287919063971,
 0.5923404286884711,0.8506493256652735,0.05557475739852902,
 0.20813595177675792,0.6231131115009736,0.8216311237557877,0.7588972007666729,
 0.8875464302924,0.3364476496236425,0.11098906331045977,0.8918983431645519,
 0.9589095888949305,0.6269230408710285,0.4522294183015173,0.9957097467058612,
 0.8396187805154244,0.1157348381779838,0.041670267789313264,0.385228297213421,
 0.03788192075295016,0.47590563329136737,0.18255639769845822,
 0.1494238848886278,0.525478023800904,0.1900253176449851,0.1017855453973916,
 0.610509550579283,0.7496935133296396,0.12986844500758254,0.15442975122804037,
 0.2721780008559378,0.12914126274483484,0.22789634604366504,
 0.5276600600372464,0.40979908175029134,0.5164703836169819,0.3794871126437074,
 0.19227832836515457,0.14096598645881875,0.45317948790870977,
 0.2761947140990292,0.9503163441667599,0.6962088248276742,0.1511205273813605,
 0.6965283717555886,0.18033117238244234,0.7531030941081882,0.8171663041603413,
 0.573642493045778,0.5479660349936456,0.022698767456314717,0.6574487024234261,
 0.942119804175425,0.03271718839976756,0.699206589343382,0.5223253258776748,
 0.8313606829440151,0.36421902638417125,0.7430280851433365,
 0.15021258949861027,0.7158692978466233,0.7660271147639578,
 0.12186697124393764,0.14166273351206693,0.21294285425759307,
 0.32000364438934936,0.6911000780626445,0.8644171105465535,
 0.09818755536929835,0.11998082355810302,0.7964780137119457,
 0.9093063110185199,0.7633730108967736,0.8242443425916616,0.07579945785717446,
 0.7428266469370423,0.41253358858275935,0.1696832310314701,0.6169428188324583,
 0.7876148659536193,0.5409076216921925,0.6760091107207573,0.2168816715175348,
 0.33675407590876605,0.4747163050293103,0.926263962091948,0.3707875662314568,
 0.6077079746568131,0.8189329069948216,0.08548724373979322,0.8211993954032174,
 0.7693211177661657,0.006475577371814034,0.4962708246971962,
 0.11961657237705625,0.5141729549350119,0.921637266847575,0.7932281637582218,
 0.45091236672035195,0.026377418309534106,0.30604377537502103,
 0.8865194922961042,0.4518088853207719,0.6550522299842958,0.3381178253399354,
 0.3937149870563924,0.6006316536479963,0.5536796835769996,0.3887928112096902,
 0.1671873616906142,0.611120286140348,0.7711808891105091,0.5024984095125217,
 0.5455358698190023,0.5308113090991928,0.8888870322501203,0.8143394509349978,
 0.03718240148169871,0.6901250955656326,0.27271858515187486,
 0.2578166449867859,0.7487903362425987,0.6435440259054732,0.21466728518171907,
 0.745129673031712,0.1995289423711386,0.6738025400399617,0.3095529307346626,
 0.684514931987589,0.5051015307969261,0.5270053088037747,0.45183024914889136,
 0.26277480873805814,0.32325619581909326,0.02167296410180075,
 0.263880931545183,0.9671562877454716,0.1424456065177313,0.5229053231402366,
 0.01251102974667806,0.7761618298645826,0.7205748603373718,0.3262719114906877,
 0.559103800423708,0.6549832105839175,0.3907701839848714,0.7608839742108136,
 0.9279470818300375,0.08783640437102558,0.8523852170903908,0.3678824234340221,
 0.36957215801292,0.8549534002857677,0.9303397301928472,0.8292332151570891,
 0.14747855771873686,0.42985738023178377,0.20302748880792643,
 0.40839527704137735,0.794222077420726,0.5511366449555154,0.7283544463620635,
 0.8251994445059705,0.040727084286737214,0.8819012723899541,
 0.30466976860711636,0.6842707171606096,0.5342440906617487,0.6248486060280681,
 0.4143185700748172,0.26954439044766787,0.41604355651997826,
 0.16001993335452525,0.0345982906137714,0.816204035691997,0.6146497363081445,
 0.8026574078730576,0.8191868843342437,0.04676277210115065,0.6880550564765722,
 0.5518468817403758,0.46912880488423303,0.38231333198844863,
 0.13955662446756611,0.8054500316833192,0.2890491750268779,0.3998333476751723,
 0.14248371783495306,0.25022847902342726,0.7975981609600231,
 0.8400308975959829,0.22883756314119308,0.7767243389854217,
 0.46173491604947425,0.9925697637286048,0.6324451797273098,0.0978467908286208,
 0.5276210419775064,0.14809369783712856,0.37916759688504675,
 0.3447888164499393,0.38840626892203045,0.9452435811699447,0.8813966900557592,
 0.4921330786307191,0.8520218449005841,0.5971680837464773,0.22669065909387554,
 0.747836030058933,0.216589941009386,0.6370549673621715,0.905612390502502,
 0.2372926585075289,0.9353132543388336,0.5064911071982848,0.48368802931437616,
 0.528414360356043,0.4346574086270377,0.5099878295394433,0.7693562107079331,
 0.005338152535238949,0.25907666940625906,0.9304363184872197,
 0.4514018194526622,0.5686901780373036,0.2324928320839038,0.06600714212638015,
 0.04989998026174369,0.7350393103539129,0.08946706328929077,
 0.5096760823494909,0.33284701067494493,0.6239531327374486,0.7235391181837298,
 0.8333607435387737,0.7789921334041461,0.09851948052287385,
 0.38177422092542324,0.8095129118199484,0.2591916679085875,
 0.05787459283441421,0.6433351156944754,0.2686428071602197,0.6130226780883333,
 0.7085627067219378,0.1861850217809221,0.26749230269974156,0.7926533055836607,
 0.44288816314758594,0.5838941744221713,0.8278178784923852,
 0.20589697132951446,0.2509417448704645,0.9135519276232564,0.566676355615187,
 0.7125746236154531,0.4296473220836581,0.6002418904207106,0.11786247839180364,
 0.9141122310920557,0.18137570102850675,0.8111899809768435,0.6401183888585055,
 0.6085634593127984,0.5067357078312094,0.5860494863515491,0.2967466030907644,
 0.7074126813600935,0.14920606532654435,0.7642328302816797,0.6861177623570706,
 0.5177028767634458,0.3093583450155897,0.12655524297869736,0.8429179013320893,
 0.43015959465371123,0.46179660156920965,0.10140046599676888,
 0.14428900720753346,0.967501918211617,0.6166843499363492,0.9128027662875632,
 0.6152481947120879,0.7736634727439291,0.7063980139592808,0.671438895495159,
 0.8799239804186075,0.5457076544626087,0.15336150112105806,
 0.08027771151022689,0.5256188946513435,0.8053397337691521,0.3193866086236713,
 0.6479850325785174,0.44185934751096767,0.43083922085287163,0.628444200798975,
 0.8763663514521209,0.4282393431896201,0.5859269830160612,0.5390680221795084,
 0.1301791430307503,0.0704782343241902,0.26595339427626374,
 0.11436894722708224,0.04170955172848856,0.9451638239613147,
 0.42108699035654207,0.9438837966976699,0.26900173001733063,
 0.03441584891056715,0.742154303921151,0.2991322491574645,0.3603677099956113,
 0.747087306474969,0.003160849998572024,0.8507258093789669,0.4374920233298486,
 0.19218397838138201,0.6378765060033447,0.6511144102815497,0.2609682917012125,
 0.8792687599300196,0.10094765987384158,0.6049197275847357,0.6108044760675128,
 0.4671464416661214,0.09061841761433931,0.8021062136591764,0.8446545368045049,
 0.7186283917555083,0.06913544377839997,0.08059198432492831,
 0.026823723704080482,0.8064756282734918,0.41461934412172174,
 0.5184095732683853,0.5326022251385003,0.146564393296314,0.769297107748688,
 0.30116096881171206,0.6761608428173117,0.0813334088401021,0.8657023064614898,
 0.35917519490295924,0.3643176037928245,0.5097156568807418,
 0.44449501773328537,0.7984400575343776,0.5602750366892418,0.908148938670525,
 0.6383758483677859,0.25389542320551306,0.1175309374324528,0.2559832507935227,
 0.4435963174681148,0.8877989150880709,0.7354844281959273,0.9853702362046126,
 0.0017619790555057913,0.4627847372517343,0.30823131305585394,
 0.5423147352633501,0.25087717489120165,0.5700350730716165,
 0.09474928085471845,0.7882709685702054,0.7034909036779736,0.1890637092009938,
 0.20537766761648546,0.8139523988441485,0.9423665794411901,0.2759444130439719,
 0.28741637986987034,0.4266771193617964,0.37376596524396266,
 0.4604119077553982,0.8519183592496834,0.42949378048108355,0.8327068671588943,
 0.9794531064841995,0.04165160565993453,0.411440480713128,0.7957876716164486,
 0.07921922289972383,0.8022575346464038,0.07500130958988904,
 0.5707468347248152,0.3539727321720465,0.6710198516729898,0.95146854720761,
 0.6263188113674983,0.9530107474003038,0.1404721907165618,0.6774847717452013,
 0.9072699718213771,0.6926547510711946,0.6761067918098966,0.7362176801077231,
 0.8617885465852125,0.6869989160595961,0.5870446421648455,0.2334757545462086,
 0.19216401976097097,0.2185442663604258,0.351061133006476,0.18332137169068752,
 0.8372637484170842,0.006538431350568197,0.10781896860871154,
 0.9764043074854752,0.2874647482453261,0.689668651056787,0.9550604650955719,
 0.9260266222592866,0.11644715093011992,0.48786920703689984,
 0.8941465751826038,0.6632473082627167,0.5086967925608353,0.5548316620729739,
 0.21790512516335747,0.6762109770629186,0.7477512124629816,0.9755390935056037,
 0.30702344852884167,0.992937668018372,0.11068108999436532,0.590332760366338,
 0.6982156583048111,0.1482063411839718,0.3306574422403006,0.39278629035678636,
 0.21665442237374033,0.44216381388746495,0.6154686670012361,
 0.7410620997499354,0.030554341053391165,0.539079450265471,0.6545159965160052,
 0.9003945410062313,0.5087140130074987,0.16544208053309772,
 0.26655965798420866,0.7569725744142437,0.9429865291441568,
 0.19715518112308628,0.15084277356462283,0.18912276304666975,
 0.2535830193384626,0.3209803263890809,0.5869745782465732,0.16074076238625834,
 0.9901032769379576,0.7941737903468408,0.3172283572443023,
 0.039509586164784505,0.2791877270247095,0.9361658984586807,
 0.9822058294319487,0.021802074388961423,0.3336478468008145,
 0.18528685035413128,0.9459006812853691,0.20463550880891468,
 0.14786507247855862,0.10769232617897673,0.22735997319672996,
 0.14747226178728445,0.7866447862740577,0.882272704048904,0.04812894110912225,
 0.5899866921497077,0.1792842468785919,0.3025354617184821,0.935000232649512,
 0.14378160237139015,0.8440012607051737,0.1146998584834662,0.3989796119203596,
 0.17296958845696064,0.015371752817796014,0.414282918170343,0.730190946154104,
 0.3785172087618074,0.8043186212456351,0.7761084763687478,0.38006365917825224,
 0.1442513007256483,0.5486862358678164,0.4274823261934896,0.6052415659802584,
 0.630062873260167,0.36542953113925847,0.6256962047804753,0.5501953176854093,
 0.47865123919291575,0.1292047056207083,0.5091767143067758,0.7291724358416333,
 0.6924259467819688,0.011528048904402843,0.06354864319718057,
 0.6286561032185579,0.04520464226102283,0.9784038425860964,
 0.20277985688241895,0.9087131561797077,0.4094447610079557,
 0.14078697875910284,0.17552665451323157,0.4620945915627085,
 0.007212552955316642,0.2418894541156037,0.4277726927286072,
 0.7860837296890077,0.8154622202172294,0.057045833828937464,
 0.6779809050632192,0.49393795602477897,0.37271956123159855,
 0.7581248602360133,0.7883113611720126,0.4524847352329199,0.1765966294258502,
 0.7767997265892445,0.3361924857473426,0.6271067343392643,0.5852184043726046,
 0.44708159514523016,0.4563692762207402,0.8011227573360211,0.6143133733116316,
 0.9634240334303609,0.07461544040009394,0.13904762167973372,
 0.8113422078985222,0.9454083652136253,0.3075695009504976,0.35472051550780037,
 0.09408848297920258,0.4358662093870801,0.0916071212157542,
 0.025284743058651804,0.42821330894124965,0.3862634488063904,
 0.07793094042334903,0.9593891140736501,0.1178455297166977,0.5927278435281357,
 0.8185010644225446,0.641052477229443,0.310021919048747,0.157211420339472,
 0.2831807644091011,0.78802187173146,0.8793620111208614,0.09232764286006179,
 0.5219847966483302,0.0598212287141443,0.07471952644901159,0.7760298500046285,
 0.9005800740238835,0.8201288854556446,0.3458207405621243,0.8137670013089886,
 0.5897072732149996,0.5841699444717443,0.15217274790890556,0.8608928305545981,
 0.6772655335435829,0.9803683276707195,0.430190315041317,0.11436053699535342,
 0.31173900475983185,0.3894969551326646,0.15604732157933243,
 0.6652236061324575,0.6495490010061439,0.3022361609220826,0.12185304843471623,
 0.5296375314962737,0.25899428387003887,0.42111207045114285,
 0.4221292727717324,0.05154098007899266,0.8532912633557507,0.822983810748444,
 0.890902459020996,0.8655409628178306,0.5252033918692245,0.5968776036758434,
 0.22112828933165718,0.16636597497648342,0.8209216171080972,
 0.9862431350900572,0.5388261763572832,0.20551741417871017,
 0.26436161478175435,0.02700387276331251,0.892060277063814,0.309495334443714,
 0.49805789999655237,0.3600593060406114,0.7913428631549676,0.6160823392947191,
 0.6213808420812116,0.879941586060086,0.6174205848627431,0.8329376109935608,
 0.44660762885828464,0.13362946951132648,0.06399757147850249,
 0.9202149898053965,0.8356047098492192,0.13815725383188826,0.7398317161054835,
 0.3280057694991383,0.3774235315074397,0.5884765180016979,0.264806929736611,
 0.6839812841513674,0.12346240926221541,0.25012572214200524,
 0.8900410080913903,0.30012958221260777,0.42930227703061774,
 0.21563650638437415,6.408955583029341e-4,0.15694265448457034,
 0.7756836336307051,0.2709390411283922,0.9544470598042485,0.9999763317674281,
 0.09775115911410703,0.12266707516486952,0.10496880171540157,
 0.6559264907032037,0.22974799190602657,0.6914782191876331,0.2966173032106971,
 0.951308738537197,0.398208115844215,0.6999156212897497,0.22918949109453135,
 0.9554474007834018,0.16414087305694036,0.3069522414413458,0.2371995364736852,
 0.77411619131088,0.77368329760658,0.5841790436270387,0.9781760330959722,
 0.5481185554267666,0.900532734037419,0.04880110110901836,0.8118897178047144,
 0.8664712266226346,0.41264828116346364,0.41862234440892077,
 0.32453142870282803,0.7365549914184052,0.4635233879802495,0.3389222151585465,
 0.518635594531359,0.4270008640387254,0.851075901628516,0.6266677773686876,
 0.49839263436212067,0.918512917544039,0.9965301175257361,0.13755569433957593,
 0.9102587579177435,0.317037178193201,0.2710276830288565,0.2035021303659692,
 0.8921537300329774,0.17592882782539565,0.866852935783332,
 0.012373518199689904,0.7239367008865042,0.1504094852195682,
 0.5930652709333556,0.19619292256304388,0.25770672225942437,
 0.7973244539333938,0.7541197490980003,0.2731163318550338,0.55632590359953,
 0.5449833314301118,0.48368213941434335,0.5908627293853073,0.8147175916462903,
 0.491620453870124,0.5975489898808619,0.31663517576619327,0.8797896211179457,
 0.6434543847559847,0.4660456642779991,0.8488903760562209,0.6858383414648452,
 0.16240254650811714,0.4217411314642422,0.8888271156288153,0.5500939708448358,
 0.9901299254340211,0.8699407477522515,0.48251521815252063,0.5158796465141248,
 0.15646149002534704,0.6498296929552894,0.31106002437134517,
 0.8475280698634067,0.8462483886333114,0.5697258277378107,0.33195732220782936,
 0.3030516050212576,0.6756723189252828,0.21272681986829134,0.9844485283458273,
 0.17659480868932964,0.17786759845980715,0.06915400616215561,
 0.32958175984481963,0.979720265277783,0.09970526105994171,
 0.43768046002736694,0.7178463312642859,0.7371986614401616,0.6180561614049551,
 0.0011863775135382282,0.1135067279403047,0.7370680717033151,
 0.234681280546029,0.8022740972203554,0.4125453139872851,0.5982517150885959,
 0.9283830024473856,0.7554959590367957,0.45630973762555715,0.4461816095952966,
 0.17467526087340313,0.317273892303912,0.5479167863151965,0.8427658129061195,
 0.5316018460816454,0.44554808699530124,0.16591505878180413,
 0.8300637039401345,0.9920265400342996,0.09098486415385276,
 0.27150708074171515,0.963012311860207,0.6795351247555712,
 0.026803780692455037,0.527612807157228,0.8517043100944548,0.5486451228638899,
 0.17299755720534427,0.1787464868143256,0.5951105165084498,0.7299502367575348,
 0.9725006606320202,0.7864325561987034,0.6532901676459982,0.38443705545775353,
 0.059370788606414204,0.5349613726066159,0.7172400647546999,
 0.4428612509937281,0.4035655286727895,0.1216312333089189,0.07381367334099231,
 0.07542543454077522,0.16020035639379893,0.10632017087318557,
 0.1215831260222846,0.9159931235094861,0.3643693107379079,0.8318088980909908,
 0.8914019605450135,0.4089362951359168,0.5760997337630392,0.8123071072644289,
 0.6690057675896105,0.020316556652617335,0.3041688628335023,0.333747817357692,
 0.08548451892599129,0.010879588625202552,0.23133553513076632,
 0.12120738599899528,0.5651064347778718,0.8676749954585851,0.580363507041324,
 0.8476722793575123,0.19280171135923596,0.5536051202978016,0.7442520215083462,
 0.11780653032366417,0.6741355056727046,0.36341194833735135,
 0.7630276598432539,0.12305423195782428,0.9539112445574762,0.315903738292275,
 0.9441745946769976,0.22753433357219488,0.15854273381747808,
 0.8562561977683019,0.651559977748075,0.13053807021431219,0.8227307075446116,
 0.32949244791473375,0.7948084520251837,0.8460785783877278,
 0.10476679408517675,0.7135308673530772,0.9840828990787704,0.6543733832103547,
 0.6662921238077825,0.2488886670910323,0.5965759250079152,0.6747322802029143,
 0.4290834909026895,0.693533222455821,0.5842555792710294,0.39623696075669756,
 0.6134986397721827,0.27612272089053,0.9902237339034978,0.2589254178092525,
 0.45051613443617655,0.053406180181770346,0.8809199251253099,
 0.7861737161419425,0.29929547498874776,0.558190408409324,0.3646049240805236,
 0.91839332553976,0.8542202701985492,0.22392651407041442,0.04216077290288767,
 0.8202633860748614,0.5830444136282039,0.8308921373504281,0.06170530295584076,
 0.020982287076098394,0.8812384416382677,0.3249532688298198,
 0.7868868840661357,0.7386351709779442,0.1266574820983296,0.5466674072295968,
 0.8652505689453927,0.09931133026674632,0.2734953575002257,
 0.36615489739116636,0.2894724178102084,0.3501556422881519,0.8799850819185967,
 0.5801051606480074,0.29287970356310444,0.1969341674744185,
 0.27788110866044047,0.1799586762143751,0.8645470994583369,0.8226229125151211,
 0.7504982733343801,0.04111999355467921,0.1314777364270795,
 0.32578852758855303,0.17502338304358833,0.2617939525753108,
 0.3596825853081703,0.654513851736774,0.718253345747701,0.25377103053987127,
 0.31943052892768686,0.599822729673781,0.6558016863654575,0.8099413434538234,
 0.4869180478386691,0.13602551849443711,0.04493503993748271,
 0.6897934030689387,0.5625920950393785,0.15030554701194998,0.6252986605573639,
 0.5262307528999818,0.3580534727869831,0.890547920930643,0.009464339686107381,
 0.16403576101246375,0.3898815851261188,0.46190344783314186,
 0.6700762795020541,0.5904880037384153,0.6108825600236831,0.46204200516135974,
 0.9124954126581571,0.6370022847543835,0.4507673651591174,0.7475067651056456,
 0.7479696467976077,0.7657581707175621,0.2048595992062584,0.19207319504371345,
 0.7846479741329979,0.6203870916698322,0.23691127901586895,0.7369591361764823,
 0.7392666096316288,0.3361249990266054,0.8987494858371721,0.5650849200830829,
 0.09638492977535984,0.632322723925099,0.7203459217448778,0.7884847015596792,
 0.8462574814882037,0.04202606307789092,0.10202070620669978,
 0.1887608840526841,0.17105870505864862,0.05574681264679193,
 0.11390328901272306,0.5355907527300033,0.9154553555302293,0.7468264075828952,
 0.8012357741862881,0.7384707508121361,0.7928563108867219,0.2519798838357159,
 0.645274298844365,0.4035653721789756,0.6862255137434705,0.013343626458942714,
 0.594075094275908,0.8244590057529566,0.7844978423653238,0.8795434420412906,
 0.45802933535417933,0.12257021274810953,0.6949296344553932,
 0.5078933776127773,0.6018398107300447,0.6292764067016837,0.4615570041334377,
 0.4325965293863786,0.3011049233237313,0.4735439494970637,0.4064531444566989,
 0.14442889599133923,0.08433846247032639,0.15618427722854256,
 0.8843451638087279,0.7428677554157107,0.401518579358609,0.36112470570325794,
 0.003968757980365334,0.392283119305206,0.9957590873875191,
 0.23190435368737505,0.856789753280937,0.8265634363214683,0.536549160078215,
 0.6309637234420671,0.8890738278511539,0.23390378480576046,0.9227658678059656,
 0.19269035227873132,0.6745307606528499,0.7649908157382315,
 0.35834566151707614,0.9787733395119825,0.7929105526375323,
 0.006269222189066004,0.20765188087942188,0.6964526290797557,
 0.7142501278856297,0.17627784442729633,0.6426629574489235,0.6275865502098877,
 0.08826570865020655,0.6003393019127425,0.8202669548277628,
 0.038154476751873445,0.8393949720343638,0.49221529133944586,
 0.7913474016913065,0.13761554441725243,0.5774205665560322,
 0.03402024961824024,0.7752008584777851,0.5197172640469508,0.9991202703567748,
 0.5778301440844987,0.512373188428667,0.38730054957320437,
 8.288349180236576e-4,0.4923009158996057,0.8274379110980957,0.544999420387265,
 0.03895410871268856,0.6495841723409075,0.31895977269429143,
 0.28387155053662316,0.7623170867500166,0.07795140498310671,
 0.07574636717708572,0.6746585851666147,0.790857530038331,0.8789712468940516,
 0.025256149405566863,0.6606106871123916,0.08998410432818327,
 0.712858474462976,0.6888242998097305,0.6909907959279272,0.026351439096971463,
 0.556502982983419,0.4019956055796682,0.708553778177593,0.7114110939673602,
 0.17273308768501516,0.3012135801061473,0.7813483786780412,0.3341363132355297,
 0.5482121413111876,0.8034670735609704,0.3098802089010284,0.5889510177201903,
 0.915522627805435,0.11282953053288647,0.8391531683563401,0.07997360818150545,
 0.38635706731465547,0.4761873127295706,0.5861217186478213,0.9981331967839675,
 0.26009223377750823,0.1787233417764219,0.5155780837753197,0.0975085644915823,
 0.789062422577322,0.2610793169052088,0.845216045307378,0.8270736177284678,
 0.11180196984162638,0.27858736597411216,0.20980056248418855,
 0.5658191091209194,0.05158047524834486,0.6344667682927208,
 0.05020112750503203,0.16286173908488044,0.8480841872176869,
 0.021398594377935565,0.2932329363083087,0.7164116303994531,
 0.009206552643852284,0.3498025347273346,0.8456365087895208,
 0.2430361708823734,0.683616641412546,0.4146050818986763,0.1311439023515314,
 0.1717798851230966,0.41568513345157443,0.17941859015745454,
 0.9731581430330856,0.32947436759267534,0.47081276499977875,
 0.9951377771444058,0.33509673036000476,0.2575494869029623,
 0.07434623561626008,0.8552649693889308,0.5499042106064073,
 0.36311566074236523,0.6425350913995377,0.009951586119523848,
 0.6232045694193491,0.3302548815051368,0.5344520305228961,0.8957736832206997,
 0.9052210228050893,0.9174206508675791,0.6793914338622757,0.2505771654505101,
 0.6513562362055247,0.2949795899586015,0.5577865293507326,0.8664591659261477,
 0.021586385831246058,0.3240005956151282,0.626429215949197,0.2773010657523294,
 0.6915340401104431,0.6072348501316301,0.8979896140246912,
 0.035351391800473086,0.9752714683116503,0.23676173682265833,
 0.8197001589784152,0.7490464260394722,0.10212990721954907,
 0.20306464251760037,0.3131287627885413,0.746308206254157,0.8162676238913452,
 0.5370384430253452,0.04266294808100113,0.4929718309528265,
 0.08677445373584292,0.6673574089127969,0.42510882615385714,
 0.22596758483134516,0.003433516861260588,0.3006574208786319,
 0.6736787066240009,0.770290574595906,0.8664126242231561,0.6538558600418685,
 0.8503243712507393,0.1689094250544293,0.6826437383527417,0.05017284684940038,
 0.41703371214398244,0.5913803853976178,0.4245001726868858,0.4641463404661551,
 0.01676243029547342,0.5289618382246352,0.2914635176963196,
 0.015302595036058908,0.41889692640563503,0.4684518058571522,
 0.6542950444683271,0.7336468933881906,0.5249042393861622,0.6335178130576592,
 0.30708838138474714,0.29895584305751854,0.28469563165863576,
 0.7944921453099624,0.6948606811197928,0.5496767138996269,0.6053682325284668,
 0.5518676018015706,0.426999482256861,0.11585902055813058,0.8411531457886445,
 0.5876786251595489,0.16182555514588048,0.3772391955243565,
 0.33388740668090144,0.12895053716277072,0.17595963079857113,
 0.2676577802120602,0.15793651947224074,0.4471264538372465,0.5756535126898703,
 0.8577776797805426,0.20862140129876516,0.6125199180615583,0.7143005504306884,
 0.4962509076152094,0.11916467016342847,0.8104978713837113,0.5909887861347034,
 0.6487057946927776,0.45417396077325745,0.8085992170882579,0.6437174674080943,
 0.8227384260791685,0.5051870908269503,0.1669608333986503,0.9074256596161916,
 0.15641976244727207,0.2395155025572856,0.2051918798156228,0.7291155975550117,
 0.5990413728502039,0.35637307418421194,0.15795082063375787,
 0.3502238204413356,0.28883714829993956,0.4975169767773391,0.8873769202416165,
 0.15112521729636486,0.34509287473562267,0.9113580088233618,
 0.21686312155167942,0.7810113083480885,0.23989540815239985,
 0.25942123147983487,0.06206186803506497,0.09645318456701801,
 0.49435442152830544,0.02397275791744191,0.7430192535091827,
 0.10061103684422312,0.9332402747369086,0.04803716607720954,0.656816469933611,
 0.1430048493239242,0.2901466468937899,0.11314179111193146,0.7551944933243316,
 0.10776591594496487,0.1170758503838647,0.8516280786343633,
 0.08772741022817632,0.589148598724376,0.5592749858811289,0.0899978974627973,
 0.2533843070391393,0.5703976957953844,0.41107425479676074,
 0.15745617263621758,0.9806291584170244,0.7958912231620618,
 0.35062696277128225,0.028254033863700956,0.5013016988473235,
 0.08662159062704633,0.5970668499932246,0.9775022566866788,
 0.011726434561494692,0.7160224073017315,0.6805641775276765,0.769781036577795,
 0.05823175083183285,0.017863416292470347,0.36675777586019576,
 0.8377538661641469,0.8595960538813966,0.5141006861859196,0.2519260207189943,
 0.47277848576895243,0.23946955838125616,0.07394635001543226,
 0.7833069454879052,0.38461957205723074,0.8140460534634684,
 0.21416884608726716,0.246711306444802,0.571709142359048,0.5429920068162956,
 0.526629441347524,0.47614827094777556,0.1544964385803479,0.39860702261512904,
 0.050029079788508124,0.41413902997523555,0.06376469064985535,
 0.9851424553143675,0.8008131669625567,0.767257177111176,0.5195609730871943,
 0.25742676501541406,0.5150519791396526,0.5101630351339608,0.8759437385267334,
 0.7764400243094455,0.36252586779690754,0.94861454495455,0.9780420490792958,
 0.09139674512161933,0.41029771172274765,0.35501122185003975,
 0.22540578570314684,0.35325521896732603,0.3282635394039143,
 0.7216665827484507,0.9271808589731111,0.3692303177318639,0.6606608121647608,
 0.8313863269263263,0.1538387362508602,0.5079881506398664,0.10750414398463437,
 0.1823441314394263,0.3546904522276594,0.5174000785061915,0.5620249746304593,
 0.26019606957321484,0.7690445965692416,0.5941938369216954,0.9585451767711144,
 0.5113134188734741,0.32286792329468705,0.783825791865969,0.3814848726360973,
 0.4158060798231785,0.41415803154594166,0.6482150484444861,0.4693623475272889,
 0.3089953842387205,0.1959938534516057,0.6675990317411263,0.9210715188011751,
 0.45182931036574925,0.29883541129783486,0.7308207617814122,
 0.7200195790229554,0.29435934156106625,0.1151090794812365,0.8817163742185936,
 0.06198055794446222,0.6219182597345171,0.27288883453233415,0.630017701850363,
 0.16893029069212606,0.0582677353137091,0.8809703295238368,0.9974836409153696,
 0.5067573470506636,0.5323533520493927,0.8302945263998253,0.1324689318156188,
 0.5886489194467714,0.49682424897023114,0.07885851703225744,
 0.9860052020782545,0.7748288604866532,0.35570089036965247,0.4926927564252691,
 0.9999693247483142,0.34955189255519503,0.932028573701064,0.6624312068757018,
 0.3104093108176418,0.19450304944708052,0.9344278110297954,0.2762667852945183,
 0.16962439471439195,0.0840423627186655,0.156717290994684,0.29370595134551714,
 0.7765764882637074,0.9541388470224235,0.1632733004456688,0.5987367370683785,
 0.17225294164956706,0.7248677191529626,0.1098834668597426,
 0.19494235525404913,0.6950854910091682,0.9013902261746474,
 0.38720408442748144,0.2555682342447999,0.5183447680593437,0.9867501542881558,
 0.4917905363469074,0.9144733921653718,0.17429127873816408,
 0.045394870036943646,0.6797386475664478,0.8310387989984102,0.552784678117647,
 0.04393460651027281,0.3331552852987938,0.44547115564077955,
 0.4463133414305521,0.11666374815154956,0.32937038129451235,
 0.5085822938654064,0.054065239276034016,0.7223839117098313,0.872724541404939,
 0.7397451785661677,0.838717637665162,0.5856170266003036,0.8284105266080137,
 0.9148776127265144,0.07591539739407982,0.27825675383448045,
 0.9934419360914082,0.29874709450920756,0.8276396945961694,0.2111859143652315,
 0.9750812561575746,0.04260923465118471,0.11835263372139149,
 0.4283135246071593,0.7491102920922588,0.5856351308257575,0.2638999675258601,
 0.4189684870613013,0.75869651365714,0.14808217813265923,0.950223014371092,
 0.14254350903615587,0.9791601486339555,0.5012180324907359,
 0.31239778102632276,0.2419178708643751,0.4548715689356414,
 0.47685210751750273,0.15741485606754502,0.3725391619261096,
 0.8712360513298391,0.7425322325383313,0.19544061374262278,0.7663350285753774,
 0.055011411592475135,0.7738799073524527,0.7722810874164121,
 0.3760687947286021,0.1477067305432127,0.6935843051725655,0.7589011874059691,
 0.5036052739994311,0.6829395322114306,0.4219923962694242,0.3550130752334427,
 0.33273692024784074,0.18536669228803715,0.7270216733137832,
 0.6664303340846355,0.356644438883882,0.14701845746721687,0.12751453303545413,
 0.87750086169396,0.5141353562780544,0.02330983219649707,0.1716358415295014,
 0.29651194461971464,0.11994230239159076,0.3044373695327829,
 0.11608027183603464,0.14149132640741036,0.49172083294088487,
 0.28579388820947416,0.3714590312780419,0.8143767903724587,
 0.26979766956632134,0.6379522515362337,0.30398369318833174,
 0.7727194174407547,0.0879641255019622,0.8454488213537616,0.2561846419378315,
 0.24180772882437174,0.9708065256863095,0.7927241426314313,0.8276335713999177,
 0.9809680901265739,0.07572048073296533,0.7558640047182077,0.9545988357774091,
 0.5777370562291108,0.3579511468149953,0.9666473414734795,0.768999916153808,
 0.8436128024539592,0.938224295598651,0.9644076505424338,0.7871637519222017,
 0.8727747152771763,0.940865910302977,0.20095751409380136,0.1936599783296299,
 0.278093739398692,0.7238079438168643,0.5529740639437298,0.6853832496412045,
 0.604815751711156,0.8085183061792466,0.5146728410338903,0.35297998198650204,
 0.16485493769864779,0.6762792541542326,0.5564905191946615,0.9066230745725189,
 0.7038791262675298,0.23091623768069636,0.8514732877410145,
 0.18137432583274105,0.38226581519380687,0.38691309098811344,
 0.4223579683225337,0.03520927736549484,0.40992232506317294,
 0.8115652042456102,0.515009568293095,0.06044178170935155,0.7382193668743351,
 0.038302445004272045,0.7811279661864523,0.40011487131003654,
 0.1298394408948138,0.30328421367377745,0.20493070483434872,
 0.5345579680923458,0.9128036081588619,0.24445848645019896,0.2814334358644126,
 0.9504898650521646,0.09606516269227083,0.08641268030255866,
 0.9951350970242415,0.4501273450631582,0.7934474503285985,0.8125353936702816,
 0.2724421527111429,0.2987265284331684,0.5259662839214764,0.772072327950115,
 0.6759446158724381,0.07547503953951662,0.5352602354478596,0.6766037281131376,
 0.9623703996841478,0.4101195904299074,0.14813384558422593,0.5844637117180108,
 0.5778734943999287,0.9632163087222876,0.6091003106802174,0.9143089241636191,
 0.4884160874261416,0.9928018932481876,0.44147162246703187,
 0.36734067398673376,0.32120763874690283,0.569107519319777,
 0.07995188232035089,0.21281797273264458,0.6197516916593903,
 0.9425795665046359,0.7234386311024822,0.9413693274229417,0.02011459231434054,
 0.0971816406398941,0.11296574492531364,0.2817456633027896,0.6180360512326996,
 0.7878672430735094,0.4914209804185371,0.42261873546102136,0.4618920043129989,
 0.08263272938250665,0.3650217911097958,0.37132938265923576,
 0.020960006942358445,0.6424533516353097,0.07202637247595933,
 0.1908201800200905,0.3373346229680442,0.27538013257158656,0.7432145270257804,
 0.45166591358794206,0.48188405926965916,0.42236378334752367,
 0.24137326922090874,0.9540387545930193,0.5048111808200605,
 0.22249520214986707,0.9813658904274214,0.833677021298509,0.07079372580451204,
 0.43997319795674383,0.38037826478729075,0.6857473982124042,
 0.6632517277014132,0.08283812086496423,0.19723212618160335,
 0.17939288621561433,0.5128146809480638,0.6560071119918496,0.7562291008111897,
 0.2968863223474205,0.3428062677899757,0.4672620933761793,0.24678762656855102,
 0.5920112602042414,0.7746297808791723,0.6474361754862573,0.13403424556922494,
 0.9123515324807694,0.0708353931395217,0.4921375735125296,0.4393577938416192,
 0.4151413313551343,0.30419313200423814,0.537336829226746,0.5637567260360379,
 0.616390263847996,0.31817882376762285,0.711378483787687,0.09477899436995696,
 0.2059141717290387,0.9290215068637203,0.7797372133134288,0.21451702095647696,
 0.10446251260675632,0.6954890532008794,0.07301537281960901,
 0.49958229787282804,0.4829793092614618,0.6779857907276315,
 0.03902814438108315,0.20164755000138723,0.9149758799537534,
 0.6582972627653069,0.6392082363732356,0.043404984485493237,
 0.2069370934667899,0.38428649110412216,0.7397272625902436,0.8246370219649093,
 0.21087914413640307,0.2906192820798855,0.9445054262839042,
 0.18598474091384132,0.43422377635448506,0.07093661643385152,
 0.30134773891847655,0.9037779916101006,0.18655526097057085,
 0.42588373321395356,0.8251285553310462,0.06633064197215133,
 0.012557057718546094,0.017559308380334482,0.8870238370427656,
 0.31507061931291325,0.6922536324251577,0.8521952358513136,0.6443109374966937,
 0.8092100293395438,0.38446402282848247,0.40410919916594656,
 0.5317105770797346,0.68090680064602,0.7984358672091171,0.8902930082650337,
 0.03120779148929287,0.17923069727486096,0.5282654071131245,
 0.08150715884617576,0.9917988889946157,0.8462140110271088,0.593831268153217,
 0.6067452613359067,0.7725174582157657,0.866803809064327,0.8379670351759214,
 0.15706381939454295,0.26039089208121546,0.274836326207767,
 0.38520298871225445,0.9477421947125142,0.8341979477949347,
 0.02245564015304191,0.9087755473908561,0.809432313423924,0.2852261355870459,
 0.28829046478442955,0.11242195837274782,0.12884699403777178,
 0.7347940821871695,0.12356906177725524,0.228209761320181,0.7873320615189832,
 0.7278194583587815,0.311190550545279,0.007743347703918713,0.4338798147646319,
 0.16166182748685642,0.06615630839002229,0.1599055757812421,
 0.6412860770260383,0.5030565470916358,0.18470537174957835,0.8333244464654388,
 0.2915233596450052,0.3790519861156624,0.5461196535260251,0.4128479899330162,
 0.6770469485542372,0.5851136539107857,0.23957098298702018,0.8208174259084458,
 0.2768344622505625,0.8965147038542839,0.22192899940953714,0.2649917346728441,
 0.9960952644250215,0.7519120217358393,0.2334748432173357,0.24505348395098403,
 0.31846280751699596,0.9202884430543241,0.16293143294350787,
 0.9490847650328693,0.35050503378964193,0.7126410420049454,0.7846928638579084,
 0.7113892912566884,0.47926484579206774,0.9722642161383258,0.3141118204748391,
 0.4003261151833901,0.35623786949668834,0.8554677857901167,
 0.13646877701644866,0.180318184224308,0.20922351255974692,0.5927000609358324,
 0.04395573205415937,0.11772923965625526,0.1933894523629811,
 0.23607661009950576,0.4404459495866989,0.26150295815020663,
 0.07181232330165699,0.4340571003432393,0.3227724722549251,0.7618835528524314,
 0.29536710961958135,0.04976615179286159,0.18106088123384678,
 0.7500056678752175,0.31970769097478824,0.09420916351433195,
 0.2571822095622389,0.21175590962664625,0.7813083547517174,0.5849506430593638,
 0.7824249337860871,0.4178825106720726,0.5021203850892398,0.88602838363316,
 0.2569131441491883,0.10623938866362415,0.5413150190742415,0.463698072468427,
 0.9793323908258575,0.4649941680558637,0.36026373580362336,0.7691488728682444,
 0.15352474831102325,0.12798621649637854,0.819168053228226,0.6563596072082948,
 0.1819152253888725,0.28168276378993706,0.505265335700458,
 0.028888405630154157,0.4778948593710213,0.7692489445803592,
 0.11300981201445004,0.1083798435700779,0.38806483657079327,
 0.8171594146399411,0.4112872967484621,0.16614083839409843,0.6227324459839517,
 0.5203651060841517,0.6372189795403873,0.9039842265328706,0.47618227861358076,
 0.9068542853492937,0.885522759928593,0.009061443120894075,0.9680374026947494,
 0.5366625883018711,0.7991777590733776,0.2572494993922646,0.16419206123518515,
 0.9725087637655767,0.11677317875307891,0.04545645486805994,
 0.0825471435770555,0.5782812526457707,0.9929639933538262,0.6930414464213623,
 0.23931388021110478,0.44112859589272835,0.9639816196286913,
 0.7052223801007111,0.2277086408444483,0.7692088508168795,0.4805735446936028,
 0.6895196395590617,0.992343176552325,0.5146829048815565,0.5985311770089656,
 0.15137044916534748,0.24665464026265815,0.744568620558226,0.7498257550051859,
 0.9522642966198542,0.9402230333951673,0.4308664256661263,0.3470301712175995,
 0.814399580015345,0.9798264817954585,0.4255189650093649,0.33457620393961685,
 0.21425972579308283,0.7921835584545517,0.4712686413053243,0.4930895310003298,
 0.1765736150402787,0.23622196026949238,0.665916441172825,0.6320585600892504,
 0.7151702217446264,0.8530277290208013,0.9065746957037578,0.20745940993120304,
 0.582860098465649,0.0677783018833148,0.6923165995757871,0.879175052364105,
 0.2681780241456143,0.12421195265718898,0.8396546945081733,0.3371657551432796,
 0.9026838938277062,0.15575264529077226,0.2721064524465726,
 0.03059997576860163,0.5588310846101414,0.3197299506540857,0.8216614686267705,
 0.5122783050830578,0.21351008578330222,0.987025796521911,0.4137806881617303,
 0.844999197436476,0.9881102971027439,0.5323765647610449,0.20638939900515085,
 0.298419339533184,0.1331800720868055,0.9497908978338369,0.027298173443803586,
 0.3615889184638992,0.06722857135243465,0.05065503293520879,
 0.21132112389212054,0.10159045163130698,0.13238757887215402,
 0.7288330132784465,0.5319474935305265,0.7472327596114027,0.45757234017400017,
 0.83040435662753,0.6727819018144179,0.7650704717074631,0.42335436044186636,
 0.6237826356524796,0.13102747777652668,0.8039536536830446,0.6055496105756776,
 0.7037143721295511,0.730283862764361,0.7378772793187112,0.2658879132429707,
 0.9325660841916016,0.3807329178759674,0.5059923856214212,0.84367226547317,
 0.9824143757301462,0.8705889814486327,0.5114947115067692,0.1346854168785443,
 0.43395293604633467,0.1206089680001372,0.32644958233781285,
 0.5981686341360385,0.7416330593565986,0.20483402972148723,0.6024246966584743,
 0.11936651877753257,0.3986287367896888,0.723830252130845,0.2596041658598891,
 0.397037827962783,0.6111619182667865,0.7310604082296249,0.9671549991411047,
 0.7361149959608607,0.7941655093748092,0.9252173499912977,0.6082435325842722,
 0.399507463315975,0.147300057001869,0.5788257943048886,0.6285925267167738,
 0.23343810269874077,0.38445238950379546,0.3695679310030331,
 0.5085765026799146,0.8407302809680253,0.058126881933867613,
 0.4812770286710757,0.31882823390946113,0.5907959757705301,
 0.35976432649187073,0.15931922810675192,0.20844483367325684,
 0.07563448434401066,0.3502050671022263,0.252031971591716,0.08946835671725606,
 0.6173650187739173,0.9179463854453125,0.8853177422838971,
 0.012141211093224968,0.058622606721694615,0.6230996244094762,
 0.1639536986780662,0.18245157573099924,0.17324962583165904,
 0.16755161355671577,0.8405829666160396,0.7588075254635203,0.7667313687399414,
 0.002482834314577964,0.44605577202736635,0.33263586701628634,
 0.43608288356056524,0.5468651361414661,0.6998901071571185,0.6295925039291634,
 0.8386362737603987,0.7638229805129533,0.7690139645823715,0.3113075606625343,
 0.9597523251775277,0.06612840284782406,0.5194742688939964,
 0.10609711895984131,0.3653006077908352,0.5175205819726367,0.3684504685163793,
 0.40237581140410095,0.1273121248964283,0.6671418438480585,
 0.19110393788377156,0.020068672521937225,0.8980061780336209,
 0.46318019070111127,0.5843160049074175,0.9442888654623953,0.5985215221402995,
 0.2068693896943984,0.624882903536032,0.23429405983022722,0.8263587389410312,
 0.46435354800509043,0.41642392940902884,0.35137051258974217,
 0.5752526288852158,0.8948477440719937,0.7823844762340415,0.6909560954558318,
 0.6297858742101309,0.7507545129632512,0.7417684610456973,0.4578688799709706,
 0.8912708082290041,0.49027829047812665,0.14541468074635344,
 0.6163494052588574,0.2152279742148654,0.14492772754695138,0.5231887543257558,
 0.7869450340178054,0.21927703535136112,0.6956551036025539,0.5192303233221808,
 0.5563254293461164,0.9840698070963293,0.7902116449110033,0.7499350139669091,
 0.5380869390819284,0.05783221396070326,0.7229394505938238,0.9699679399110988,
 0.9422032934143201,0.975959158556027,0.40481279239650636,0.9435403673027066,
 0.3133650585443115,0.6499291997814408,0.24447975150117607,
 0.13207585764516994,0.440128473780489,0.6239603344078402,0.47415871854139624,
 0.05244876245744434,0.8718780505547786,0.5440204396389386,0.1527788248534716,
 0.335530705519963,0.39236175780899696,0.48335537993360667,0.4077584169860121,
 0.23084754418697717,0.7043667712879977,0.13046180834496157,
 0.2797907833167378,0.2589913670597823,0.9555436233472099,0.5600639460866871,
 0.6849215261486129,0.448335193390175,0.5413288728096568,0.9470982064769031,
 0.38763297889076975,0.47653114154401865,0.8439902752771683,
 0.14457451370694407,0.39884525491358824,0.4656598839263021,
 0.2840148954650342,0.3664978172423393,0.04070365006731669,0.7401907867095354,
 0.7761867598699594,0.13841787891690038,0.3465322265930484,0.5183067883153291,
 0.846914715834454,0.24214877692943693,0.8256570173023461,0.5085583234403652,
 0.6981396941182486,0.4928390736155229,0.3943302154010424,0.10894175439879028,
 0.881665898383615,0.911178044360694,0.39010095558030033,0.23397642835079546,
 0.3963951389671423,0.047766414156705905,0.2810403842889587,
 0.08369925745156359,0.6255701391915554,0.7525299129619247,
 0.40847780298770775,0.14830438174619065,0.07439875687806063,
 0.9441697879710955,0.4296019689207323,0.2479977426165897,0.1274210417347016,
 0.09112234550393261,0.9699529907148363,0.023948111216117507,
 0.21024016778470112,0.7770762489105435,0.7441251307895878,0.6321081192990727,
 0.4112938510752162,0.20332653565474113,0.8639610478400124,
 0.20606593211937263,0.07104732109392364,0.4483007278714072,
 0.10500648855845618,0.11188109357205356,0.4850296632795904,
 0.9690490023063292,0.06667953650368041,0.83135767199073,0.5641684324718932,
 0.620247752677227,0.46540035544845615,0.3372622682861478,0.476989154060264,
 0.7908567012080471,0.04603368269984831,0.517143435846018,0.6834826321802104,
 0.8751833645184489,0.744653423044346,0.28298404898053364,0.49358417287763756,
 0.10748088032797143,0.5681069374801186,0.7378674192230968,
 0.29217954861144324,0.5042605660682437,0.0455184679867332,0.8510881562305054,
 0.985744985282227,0.19114219591616377,0.4425176421316178,0.5696233966701039,
 0.06957279901978952,0.38229477289668823,0.24156437759639426,
 0.9264464384946842,0.995534141100744,0.7534761864978128,0.464105171347413,
 0.973138750430926,0.19808771046794837,0.7196615079383003,0.7238480674728451,
 0.712827183982593,0.5462977823015828,0.947324981224108,0.8106144958123365,
 0.9287588490606811,0.40500904473519594,0.8278271127806944,0.6457144572390972,
 0.09304866861447258,0.018971258002817626,0.5066837139441718,
 0.8628218821080706,0.660810154572774,0.2801910733635642,0.2722548515671175,
 0.3913291004954397,0.4734317939847412,0.9765937215539654,0.9995854390683739,
 0.4436506883476329,0.8716657072721536,0.9827411920223542,0.39683333478175264,
 0.5173531555675099,0.25453149404229214,0.5903818652074586,
 0.10028003983907974,0.9223230538253185,0.15187145871640317,
 0.20354708437583702,0.4672806736935391,0.23977968923568227,
 0.1901494121428986,0.08116636739223149,0.13079102013722455,
 0.22338794926915584,0.07953437351580828,0.28059667178511805,
 0.27338630848990964,0.9493199121225405,0.476624554361212,0.22602894149617458,
 0.976240315709855,0.05823667054571546,0.34046976745770663,0.9979528740956616,
 0.10643949079828241,0.8963577378614058,0.020869581116629066,
 0.07035731487573305,0.20237628585421208,0.5493307697234615,0.535333817228416,
 0.6675113176709793,0.70287921254929,0.7549426491581624,0.08337943587109353,
 0.40627442850685624,0.25938630328754675,0.8769957155348981,
 0.40350869785975163,0.0097493861445348,0.051599357635503385,
 0.048555395913547406,0.6558737191939655,0.6899312201415903,
 0.18057589503878924,0.970802540065274,0.9190000051963355,0.083328267397758,
 0.8719213567304732,0.4756207316251788,0.04637385170608055,0.788423324731929,
 0.11810224003107028,0.5943418613340798,0.8490137516578051,
 0.39496757485125866,0.09448616361995177,0.6818771898811994,
 0.6225365969744923,0.400989737830434,0.9122421510928176,0.12458041837415923,
 0.5672960628680324,0.8394348801382306,0.3090073107091764,0.9425515184118339,
 0.8880389039468407,0.45329593907771115,0.20923249408251687,
 0.5693401933446726,0.9184765892229014,0.4616367687847488,0.7919994288716353,
 0.6669254296934456,0.48658100949461736,0.3149129426529643,
 0.23243262456031544,0.6854143234579418,0.6506071487126724,0.8257092120701817,
 0.34263107020548267,0.13979429918752095,0.9153763068209566,
 0.7942990861837477,0.8604286936582837,0.6918261334304532,
 0.043394427129148094,0.4734245118615985,0.5640579683258278,
 0.4777301241045764,0.42987316087116856,0.03402139381344382,
 0.046240485468969084,0.8259652805035345,0.3685319623291541,
 0.9545135070914875,0.010044764090487934,0.0037588962661659897,
 0.4407365756608239,0.538455381358059,0.8907521555082841,0.7177208606654386,
 0.22891285530272953,0.3826647502192765,0.13974387604752458,
 0.06278988541458852,0.9636048781789304,0.6135235996950775,0.9397432252306036,
 0.97952324681693,0.5470328459712215,0.2617848301874073,0.3172331527461859,
 0.7440182708952671,0.8995121253140824,0.9316734507804344,0.4867189495593728,
 0.21755179365727018,0.8744444701175151,0.0674502395820964,0.9496275069263317,
 0.6827921699794104,0.029061733047221594,0.10675523227959749,
 0.5638161374612082,0.9409083183063646,0.3701015011940838,0.28915228317160013,
 0.2010220468263112,0.39373825803938167,0.5737806508853874,0.5780966384840115,
 0.46846806405387564,0.4011151084823983,0.06260979353140739,
 0.5635814807495717,0.45067692441098917,0.2230065548808332,0.86614319331829,
 0.14214036130362684,0.2257931413245,0.025777320696572015,
 0.056880074239606726,0.7483646733086046,0.851693741066928,
 0.09589697211315473,0.06419614859686407,0.44118179915315936,
 0.34871998944127613,0.7557730749705446,0.26870069771169214,
 0.5705132885681258,0.6398244619906323,0.6387661590249286,0.14657033613016401,
 0.7739897854436978,0.9962273568843936,0.09668425048250051,0.4617619288743935,
 0.9410662807422926,0.8271825447402339,0.8983025686186727,0.6720387196832651,
 0.3289966489335544,0.026982888770971436,0.1391291479908323,0.945827603791142,
 0.7030385162037978,0.8325964311470934,0.8627047980741376,0.4548051499198861,
 0.5745546499150349,0.12060555210193824,0.7665654816000391,
 0.032070802865957315,0.34126676018602187,0.850702513499244,
 0.5014077053037613,0.7166442255809526,0.22931083933154672,0.1782004617540468,
 0.5845012893211576,0.23989048845264793,0.9445268077892968,
 0.07952782270014191,0.5800661228456369,0.8062455217516806,0.0487692802848787,
 0.6522808032993499,0.2887762866898089,0.7019328270488734,0.8067915529112976,
 0.05296546883663189,0.702595061158719,0.8688385381531901,0.6507570420559774,
 0.044223733231628204,0.6866993725718274,0.164381461207315,0.7201090449233615,
 0.0013932916434459264,0.5330663376413908,0.8672569191511429,
 0.9025988079733955,0.36964588276722865,0.15605309999627826,
 0.8086858816032785,0.5175342420446366,0.19892414219152665,0.682042302847458,
 0.4729334993052463,0.5480932823404229,0.5087729967857135,0.5422351915779282,
 0.5672990182123364,0.5597158636321213,0.16824057209841847,0.2798231344198786,
 0.20651549130514602,0.828047779872662,0.5669190726896162,0.7044562786220263,
 0.537136417415981,0.34377912164970814,0.5334794327371813,0.09981774178517566,
 0.14038998194266572,0.7122184820839711,0.17377503863239474,
 0.21281765410722464,0.7500924113872849,0.5425021273410552,0.5304118972669162,
 0.47628642558602863,0.3043867696415433,0.8599159269584301,0.9422433481881879,
 0.3081407473953308,0.2034219069756037,0.16556833527528458,
 0.22758414858758613,0.6391982114662359,0.1687237984625346,
 0.22315047679227384,0.2181045751265447,0.04639923179840222,
 0.3479370388096237,0.8455638748787231,0.5784812616227134,0.638216742026192,
 0.4333334520622414,0.8559559853821449,0.5864002828219349,
 0.023833014532540053,0.5246888069812311,0.41121581813744423,
 0.7229257820345159,0.2134423081957636,0.269582335363012,0.4083613315749577,
 0.3315686569328047,0.19462585128615328,0.06922518585654136,
 0.5070193259163078,0.9773523993231906,0.6557020680109352,0.24542403415272107,
 0.4181065407821246,0.8572685183369692,0.444194683786062,0.05901237269111792,
 0.6078443629046583,0.2531048115525829,0.035306191385168884,
 0.24056065846167263,0.3037613944343649,0.6508811045060741,
 0.20389921423246948,0.4520187367826741,0.7916311372007038,
 0.03273182429771704,0.9295770353941449,0.806701083896372,0.2995390684335121,
 0.695691432684455,0.6774913331241281,0.44287478397159363,0.04471446384034383,
 0.7247673941581021,0.48034625738540515,0.9331008660557062,
 0.01621572068521593,0.5795285392785412,0.3382626802843448,0.1804329536026541,
 0.2485590147099066,0.8015146269111179,0.9693293141068423,0.08448731178183622,
 0.8881268225007961,0.19578995529529641,0.9784442697240566,
 0.19212559185675027,0.24206659789183038,0.5187631798239754,
 0.43579979458073015,0.6796369258827196,0.190907485472942,0.6684731760723414,
 0.21792302366088878,0.8455086031389055,0.2370953294854462,0.9658891876912339,
 0.41103461756965176,0.5942952753869976,0.21269777370235332,0.652670993991938,
 0.03794567762230194,0.7867087234373858,0.9318293113656659,
 0.26789877523399497,0.22142639646914386,0.6683581779136513,
 0.29592111444626956,0.5523793437504522,0.4131861815502369,0.609753773751399,
 0.9392450650020571,0.16903090549970523,0.14135717171273976,
 0.4684410801782114,0.2828683937850869,0.07947552728905738,
 0.03045805842157001,0.018417970480410384,0.6581796234984884,
 0.6086589335339104,0.7947048457386106,0.06286421213707771,
 2.8189586759852503e-4,0.4148506933928804,0.5040309238255994,
 0.9567295018708954,0.306589621161206,0.18078220431961844,0.6497325787815983,
 0.4127018510027647,0.2148940091972168,0.5978568023484011,0.1408781007809079,
 0.7534225130755707,0.7415612406620723,0.648633197980755,0.893510275525999,
 0.8000568151760106,0.24331512989078863,0.2005370073404692,
 0.27630505816297735,0.31042256273693947,0.7013940749744553,
 0.5930314467502842,0.8646498589477929,0.2671902996158545,0.5012644904459875,
 0.19562022350008812,0.3424420935635013,0.6917475430045564,0.3587105400972357,
 0.8308052561105634,0.3938006706149896,0.511014994408295,0.6230888802451793,
 0.2391186798910182,0.8222995413113956,0.8914201680184454,0.7245161128681181,
 0.24129041208443402,0.9896921083869666,0.15413584454004325,
 0.6149606535575609,0.6047705955500925,0.38464526597333504,0.6242106112576089,
 0.11690909211314537,0.93399883245019,0.08300549948821145,0.29556726402567246,
 0.7888710949223583,0.9536217399637837,0.5525673247532218,0.8447638791035713,
 0.8152456097997779,0.6962142113961414,0.5274827654302966,0.37840749098147564,
 0.34093930942657846,0.30600819222343256,0.26747969504818847,
 0.37641683706618545,0.9240745196656425,0.48641143367089235,0.736091220133749,
 0.004846574626601985,0.25577564230878813,0.5920229081422056,
 0.40206147517029056,0.8662406249893613,0.27868394304125454,
 0.7014253343184216,0.1790177859744715,0.33574097129817826,0.9752890333365692,
 0.11101680858636376,0.8489354859870579,0.7540427292359803,0.8629420143717301,
 0.9773524108789502,0.5083759296787967,0.009844984189874895,
 0.37374969323047447,0.8325993314443332,0.6832857536374531,0.6040453179003288,
 0.4137787668735403,0.6890635371849881,0.04891755244484519,0.8200719107559704,
 0.5492561733976001,0.40183575617060596,0.12211128069106214,
 0.8977551016234803,0.326765236809495,0.12611877410687233,0.8777678469003167,
 0.8594218502355115,0.2629629692464033,0.813002095199507,0.8063289639704937,
 0.6310287095555371,0.12375843509945839,0.2970549603917586,0.996580912219817,
 0.4794293271270895,0.5348023926159856,0.8080823463698863,0.20405315662403334,
 0.9745026707798085,0.29130121123170216,0.08212246009506818,
 0.6425198341783294,0.9566951188228118,0.7574005472791434,0.8737539516572308,
 0.016664954561567358,0.6599289494435485,0.2571566159513363,
 0.7232050699945601,0.38434176280707744,0.26565806282605287,
 0.7016532011855953,0.6143409716129523,0.5393209349792504,0.8772074008512597,
 0.15037834915986137,0.7114344221629645,0.17041995750549754,
 0.14787888656506754,0.6075060082253463,0.5157806574188817,0.3405931754186745,
 0.9350889233141332,0.6585846401241144,0.426783226388624,0.7124199316024545,
 0.7176121510326225,0.3121878974098521,0.34974541601559206,0.531006773644702,
 0.7246123732458141,0.3050634466011617,0.11459362678496532,0.6271592195062125,
 0.27944687086635134,0.5028524903663584,0.6841721758979149,
 0.24467285256870475,0.8924190044568155,0.6168645092418096,
 0.07660499283512123,0.7814337910992285,0.27268294970140605,
 0.6316468351622324,0.45333062042409233,0.5638213913975297,0.3350931425103637,
 0.410480603879511,0.876956781937188,0.7402644110427191,0.3244796980526097,
 0.11783339046491692,0.2667458341591116,0.23346203032317148,
 0.057088026032409145,0.5164826465060569,0.5044511364816504,
 0.29996499644962304,0.30370964601728345,0.6968269632029808,
 0.7575099223405999,0.7762761243642426,0.40578098402158047,0.6038898984998033,
 0.38455698666403537,0.6983205910839453,0.5591142001617494,0.5276568145516274,
 0.027089486740940694,0.25797879621401254,0.6149049021723049,
 0.7749027887951705,0.13041105661211894,0.8601968200765512,
 0.0037788740557180844,0.5720819750329587,0.8970497199411407,
 0.6947379677379476,0.44459776579982135,0.9402620661626054,0.5727209211167824,
 0.32676168788090443,0.294220893265685,0.7916430909641389,0.5893387309192653,
 0.8452249088640745,0.3040998329778575,0.04579857559546929,
 0.43450854397455374,0.4379462052000527,0.8981956710615966,
 0.09700621533279863,0.5717649563927467,0.8982616373602443,
 0.20513453129246295,0.19577213305255936,0.7577950216045835,
 0.027525330222942457,0.39672332028880286,0.9473040283380293,
 0.7432185743267216,0.13312045994874733,0.4999557906452673,0.5542090306820244,
 0.5329044093650226,0.13637101859529843,0.716119688664205,0.8274094530220746,
 0.8352642962781802,0.47348400255031553,0.28744681723352217,
 0.8398731637303397,0.8099844395332706,0.5682601030170155,0.36383343461208206,
 0.4108893395176816,0.29985430193825946,0.4808635936813759,0.7905403951991771,
 0.4127112697912422,0.13734209495685146,0.24444565610453062,0.642901266591185,
 0.06168178950951597,0.6215992035604239,0.025437857775476247,
 0.13567910516313553,0.10065063102680749,0.754154896625999,0.8972095290067383,
 0.03973326657997123,0.6504486534041325,0.7380735641449925,0.458132468992586,
 0.08582690386310055,0.32089393515815057,0.5299770970327733,
 0.08839017466316235,0.2393436445863637,0.7724924860936816,
 0.44453664178854857,0.9690008315131727,0.4494745052793241,
 0.15557891810915725,0.1586038422934366,0.4098440626993899,0.8866776990429387,
 0.9076860941098243,0.02006930789007466,0.21805557546925736,0.623283747691069,
 0.22855126532062708,0.08163052755628597,0.9628781877395746,
 0.08298155351933678,0.49028022575725916,0.47757316315434495,
 0.03660008732912057,0.7034884352364206,0.016963617823191446,
 0.05642977728377219,0.6790057461595379,0.3082439128042913,0.0473470465060718,
 0.9645240413807291,0.03293760508229282,0.5815949850806765,0.2927936703793588,
 0.31346189141954994,0.051106761121505584,0.927276715254228,
 0.20778592702661913,0.6645404013042966,0.7410189584214184,
 0.11909714028929241,0.7170925681513184,0.11705019237944059,
 0.021720836254889075,0.08568128455723589,0.9170697787174542,
 0.5557161287115479,0.517417246950273,0.8179451055684204,0.12963222268177843,
 0.14238668487535522,0.9596496201809225,0.6746914401482667,0.5602096430945839,
 0.33476682592349416,0.25714678237176647,0.9784952396266566,
 0.9615277873396615,0.17364258233475072,0.1076329561303333,
 0.19060780135248934,0.0631159640133323,0.959624588434546,0.8945558824771328,
 0.6014272776387267,0.8704870307547832,0.1330926841798351,0.6273964846269798,
 0.12497594889325203,0.6900185625058978,0.46930851238079774,
 0.9266229453542544,0.02403227336014091,0.3123719085818728,
 0.42964273456961166,0.8214854215532639,0.034579777138085355,
 0.786517682295867,0.9638649112792804,0.7549084376590308,0.07969468831232251,
 0.14159289519366058,0.18910034259405528,0.9984185565119874,
 0.6247768733167725,0.4970246042613068,0.20869750442000545,0.4004387068296297,
 0.9788085632545859,0.38234635053696775,0.9190456979715554,0.7773463536258749,
 0.7452369317272065,0.7182810337916932,0.25070596251536437,0.9723064527958643,
 0.6887081131188797,0.48706128420078076,0.055933156980094934,
 0.6637906513964693,0.5036743793619858,0.46633645928408807,0.8854138831251207,
 0.18505989406964085,0.942716236338399,0.9145814319895251,0.4469916562122701,
 0.4438875468422081,0.2647188102044101,0.5979792471757198,0.4047631419328023,
 0.2837186673477242,0.18649111475239755,0.34862050208036044,
 0.6748946922284277,0.22695219163214686,6.593913346712998e-4,
 0.800763321278949,0.996395929533448,0.8116258201346536,0.36141938607094093,
 0.649907040257713,0.045943990401568824,0.00675518105852746,0.566892767918132,
 0.4203792610801169,0.5945914820114763,0.2631252439894154,0.6864196761853611,
 0.3711475105458788,0.5213215286644285,0.47718743230365634,
 0.14239008841520584,0.5607657843737684,0.5658089688081597,0.4133382580346181,
 0.26325693709520415,0.2682353214575002,0.2592643994238011,
 0.18090037871311737,0.7332574089580292,0.44169224861222445,
 0.6004313381581932,0.47334477423763355,0.8662676835130326,
 0.08647592007686067,0.8834391361217266,0.45677975289712514,
 0.2588211383155862,0.053198624673380546,0.9244907412311776,
 0.04624299802593801,0.034935757040083226,0.5291197154579548,
 0.9311447515885345,0.08745417970403602,0.8647428409930276,0.9028912104994711,
 0.12390432969390597,0.8215318124939257,0.7903015618287301,
 0.07072897647544008,0.47765953368863034,0.4116334064336049,
 0.45428701067341326,0.2999966474559377,0.20404320134591736,
 0.7381454934987458,0.23434467528844083,0.413996467241184,0.42423783705442253,
 0.6921678476678381,0.07814357797931093,0.3747769132393167,0.2816388374716662,
 0.8320544857569325,0.39161594183188086,0.7395580499464925,
 0.09222939817391207,0.33678061299622786,0.16768562700529,0.6750175203735828,
 0.19435964841364683,0.8639949991968601,0.506165768086086,0.6827704266386616,
 0.5043682092469176,0.6171395602142451,0.9641160021361935,0.3938185801198978,
 0.5918622559431551,0.4501246088013071,0.058245512476548966,
 0.1739487782047422,0.11304225438956461,0.43694058564062566,
 0.45358603501851225,0.2318495314896747,0.5648903492351552,0.8271176038464708,
 0.35248469488980494,0.19310160607700222,0.10113541491472344,
 0.42294040629524166,0.3133827655364513,0.8765081387045187,
 0.016961420779400105,0.489467754257924,0.7124265023389487,0.7599531312894642,
 0.618602222898378,0.07956534953884653,0.6800997633604222,0.774452261919251,
 0.0919782800105754,0.3680829522046378,0.7904285059723856,0.8189766901869949,
 0.5565076453670854,0.1138482296673976,0.6455022396232311,0.22197302510680994,
 0.13154433935408993,0.17499076014202053,0.8742793835573354,
 0.28410042997517815,0.7814774548727526,0.6800758174531912,0.8600229193990131,
 0.7496427799572801,0.33461816929257804,0.2338795394029488,0.7193947841382903,
 0.2474034159981613,0.6320897557962897,0.1706631843625035,0.22007768667304095,
 0.9366955345408738,0.5032208044781091,0.9371149670491161,0.6969706785199911,
 0.25729375055396275,0.6904535142772305,0.13203734543850065,
 0.12836434757542747,0.10034483089319646,0.527394282638606,0.0562899522491902,
 0.364129207190271,0.33130280278404234,0.2239173802594846,0.4954556697018101,
 0.49668289495872964,0.9823001491487947,0.014096885490472832,
 0.6016820873242014,0.5793662786698015,0.2995579767810366,0.42982712623569674,
 0.37308806910150283,0.8888556260587566,0.7366330411421641,0.6566976958347019,
 0.74623621579741,0.0693478480932157,0.3677346042169132,0.041605021359485406,
 0.17859983517817923,0.7503092318859096,0.6131052102616269,
 0.0036349984385447565,0.8602705390953379,0.8396597295015903,
 0.829354974171315,0.17843623083145643,0.2538682228867185,0.4414484592978155,
 0.04497616850216457,0.27478200436832645,0.8101820449068682,
 0.2604443284350828,0.4862358052342829,0.32868360457365964,
 0.22233943134756995,0.20736363839484984,0.6818978595653291,
 0.15712536830249435,0.0649898627031189,0.6081642772594439,
 0.29064636915545095,0.4970339582320831,0.45389708477448354,
 0.9917826495304136,0.6500220997329619,0.8923347029101373,0.2111620091390769,
 0.5518718485190763,0.1997432779540207,0.20993636921988557,0.6080926195751316,
 0.677742607018969,0.5266163377525751,0.23121092020706646,0.3138402230558017,
 0.39984173637673504,0.5853555668253339,0.6758443191753909,0.3591445935137161,
 0.4212921498484711,0.26909215045819745,0.01700207742977733,
 0.8985494226808133,0.29990595129786857,0.9711356435193929,0.7617405309369015,
 0.052039961337338125,0.8180160532327196,0.2084725677024185,
 0.2741950837041516,0.05456072530430034,0.2559081181964149,
 0.24677355100117027,0.7529717400380956,0.06923041593326196,0.833849995513183,
 0.15002754950157704,0.6979663991179438,0.8982601076381149,0.4777205964555049,
 0.011176522602487804,0.052854796769590484,0.20367537852291862,
 0.9998363756159248,0.22743405436179598,0.7103347591760216,0.4015713610729422,
 0.2647739616685151,0.15555980982907236,0.924596448388117,0.7802652588673563,
 0.8772882645008474,0.9132917314648163,0.04551512988520556,0.8034347438635029,
 0.32109832793102044,0.48325532024525875,0.4352171077055933,
 0.6535839150325068,0.5743987852509085,0.7682785190061826,0.3156280944706662,
 0.2973551226129113,0.4086168285010393,0.46354775498737977,0.91988914865884,
 0.21527233293448278,0.6395693449003195,0.7876193480235425,
 0.03551712810281804,0.7629475879026739,0.33471637072815663,0.813825236068577,
 0.11173686175244912,0.41495504538938943,0.5481972625265203,
 0.8933625318860468,0.9284977406768373,0.9479005194500165,0.8982476314237505,
 0.4433223405273512,0.4056639399439935,0.41504604101778875,0.9382821085008444,
 0.35131126661978507,0.8661260979800838,0.27802739922608966,
 0.5075073028031843,0.5304740558208992,0.9913114940891646,0.9689143937999433,
 0.32259193399283903,0.7790307954971604,0.932399414869936,0.6903882783420338,
 0.34986967483580045,0.7931314210874588,0.5759636570347785,0.3335060194494476,
 0.9429314008864831,0.4351951701756478,0.6584631221390039,0.5133239281986303,
 0.5841973091221814,0.9548922535454785,0.1763926298987225,0.5206577726124416,
 0.9719749432650832,0.21182361222314106,0.06328803455099496,
 0.4286316654044239,0.4360286940714252,0.1461427472649084,0.9873381724649347,
 0.2403690351076574,0.7349992696763351,0.5814557919521486,0.16645706266917215,
 0.7636485970889003,0.3556959462300393,0.9966647043538986,0.5306100627160308,
 0.6588904140091276,0.2040646226092825,0.6997021789323093,0.685652079760611,
 0.939012669935248,0.9560121858087612,0.5047733807185963,0.2680339940520785,
 0.032152059380644804,0.7633369126682112,0.2339087168929781,
 0.21208891346810577,0.9591210370979588,0.8743615509757424,
 0.17712914140630942,0.36927132487102965,0.3366212051742765,
 0.7120321805270763,0.6341954946146618,0.1530855531651869,0.6256765782068331,
 0.03617625847744865,0.7544758623640823,0.9619661452815247,0.5320879691987884,
 0.5181424581922125,0.45814582593841924,0.9297337836407346,0.6397928179661097,
 0.07179941163125414,0.209749399636881,0.7154590075246762,0.23168699009683769,
 0.05532787622142499,0.8918823009354887,0.07739712342696392,
 0.7090234344924741,0.42972674245683873,0.24601701158353784,
 0.027709645348672485,0.10231119273354061,0.44588085874465555,
 0.2070343545333706,0.7114132366439303,0.5281486217940535,0.09252622774165598,
 0.35976893287736766,0.6109780404837613,0.6277943375511343,0.9895998312085458,
 0.32943483537867235,0.1291966237350466,0.36270392689528486,
 0.6598093520873661,0.03942119151032375,0.1143137413676556,0.7348681429467817,
 0.6595478225446153,0.3161275143493717,0.2844259008171013,0.9103829258778636,
 0.8210198981483512,0.1264537820005448,0.8356553419285693,0.15345538526770997,
 0.2394717055800576,0.06541766390868275,0.7651333067715052,0.172316914037131,
 0.9775873440317976,0.05808643641617761,0.07581185990001194,
 0.8626593990202573,0.3974591716605348,0.9014257292986771,0.960823357395671,
 0.16716045995629514,0.514537482083554,0.6691137983694155,0.2663826419246116,
 0.22727490060421152,0.5417819004859294,0.32006054007911955,
 0.7931859234774736,0.44795774280670064,0.5643257215618431,0.5803055529254099,
 0.7882740429938403,0.5588021228703502,0.2982433907039882,0.9995384393105229,
 0.01372854503392662,0.7945219793461448,0.049780390610997816,
 0.9138634803688249,0.41166009527298875,0.27872128340540725,0.249647996801745,
 0.49137816114741295,0.5281821987276111,0.06999135905587417,
 0.8776552236262515,0.6601926984054527,0.9165740705254545,
 0.030692918207425812,0.136201555633492,0.45288770088930574,
 0.19041560726528262,0.4293459226914943,0.5048558098016667,
 0.07067880970532059,0.4509274499026277,0.7379007941579377,
 0.15761187402630017,0.19883358516726135,0.736481222116617,0.437198855799634,
 0.9035837995416852,0.5952000426369677,0.059142329991260234,
 0.005160634749794668,0.2984853123602552,0.6795747790234865,
 0.7491829556076044,0.659692622478893,0.7254786936111745,0.4844710744772658,
 0.6037397997478168,0.31304708411282134,0.378671414863906,0.4629060691891138,
 0.27981272038690996,0.9465764217013736,0.35619041339860513,
 0.11892667164280357,0.21069214887713972,0.12226548540660498,
 0.9514927122973859,0.11396339027358615,0.18066462792563487,
 0.3704520087273544,0.5227134564446025,0.8658117702897853,0.7436990522200205,
 0.7161421094145388,0.3145861152042999,0.5798906552615947,0.6316976260507584,
 0.24567512511324197,0.6053524912959601,0.8324191982144421,0.559449546881965,
 0.2700938391198443,0.6864964229379132,0.16514425426607615,
 0.16213134831342357,0.660573709529241,0.5817649111656219,0.04870436672120515,
 0.7020182813376425,0.9614917608418072,0.6447484223628283,
 0.030146405539527188,0.13237340100755046,0.7181280217147621,
 0.6104752381404659,0.5634374079832553,0.6810275613016924,0.5758643585011916,
 0.37392686302129263,0.9122721176896219,0.8930183970495621,
 0.39841561294361305,0.3564803797465377,0.8201234872025992,0.6379089814105705,
 0.8228058629023687,0.2868631193007489,0.4798718992181722,0.6256446888255972,
 0.9222568986646605,0.28734431603562227,0.8672042379983451,0.6485212464803023,
 0.28385581789141745,0.8493000021178488,0.6113118111217907,0.5488718088732547,
 0.09025072652944233,0.9155863172446466,0.9893510191678034,
 0.29243137366760275,0.8346528196679738,0.458992857025722,
 0.016003217076178244,0.9436893989025116,0.4940034542787083,
 0.9543078996840383,0.7812925475382283,0.8355983861425361,0.6694049753811211,
 0.18264867537230456,0.24220587428533835,0.9182989527898454,
 0.8591829046372631,0.569068635802691,0.08438642209891889,0.7745841455551048,
 0.6796090700662624,0.32092218541336925,0.6349988082996981,
 0.14782712987873725,0.4201021413086008,0.568490525675934,0.3150895821581523,
 0.7670706212406828,0.12958367019470785,0.948161368051359,0.8035210137173521,
 0.6707287151870468,0.22315118369797482,0.880911529314974,0.9565132052567771,
 0.23595156690360763,0.1334739090043684,0.8243691825127446,
 0.27490894845711134,0.14947250783167831,0.8503154666113474,
 0.3059867056955482,0.6469844216476028,0.48103883245350076,0.7592447364959756,
 0.8966111442313756,0.08318723689341123,0.5091111894113473,0.1891242652786278,
 0.7596381665793618,0.4535289307756052,0.4928809421455306,0.6731614086083665,
 0.9956397161106185,0.5249544786787594,0.9352049356524534,0.9964772056601721,
 0.18053220421409466,0.8807532217378216,0.7102208744014289,0.6943276307263939,
 0.5637938138083289,0.8140871618792572,0.33389518839896093,0.8456472378927834,
 0.4116743422187271,0.4067450429606334,0.6546532348504202,0.1278731303135836,
 0.7357657465805282,0.1992176680776543,0.12191596932616944,0.8890364555050491,
 0.6504155233874163,0.2807542336316664,0.22543114173979095,0.980134401293588,
 0.4813254301473173,0.3400830219642196,0.9890013070100834,
 0.015046447796386309,0.624627200538493,0.8458608181067208,
 0.12150877582271835,0.8301003410093636,0.7661702908299305,0.2669489446675305,
 0.22068049491361696,0.11535290545159127,0.5903703062798467,
 0.7025645896315149,0.5608761287550914,0.5750233796157285,0.18315014137343133,
 0.5819154948266677,0.17938226782122046,0.7814936107100324,0.624773682243645,
 0.2897963759880505,0.6975954488877988,0.26414009357614465,0.7035114842772793,
 0.16217195223145775,0.027924251161513536,0.20482938432375764,
 0.9962886278289946,0.4869266403066941,0.7506406539946192,0.25740029216710936,
 0.2647664648503094,0.5374660143790274,0.8710286592617674,0.7415086137915083,
 0.37653084019128913,0.6077450360656866,0.6809246582189997,
 0.03249018846106844,0.15989138792019109,0.31869128450800577,
 0.5154357198270028,0.20062666052714484,0.7569249653949104,
 0.32618138451320167,0.7220694605934292,0.3155747800777212,0.0345407795042727,
 0.6890359694936615,0.8204455062637365,0.3721075691763902,0.08321129219026835,
 0.12879100075240824,0.48763399924423523,0.8681521086906164,
 0.08648828295878996,0.2231392133361365,0.7973585111262758,0.8516252914988135,
 0.48380718431144787,0.42337248164147123,0.33399526197440266,
 0.7959043846485835,0.9307503344136013,0.3327766415229928,0.03876259481602862,
 0.8544008442050423,0.7574984047121931,0.8928697787922888,
 0.023374482029820065,0.6662397195021504,0.7718253097325495,
 0.15625346661504647,0.5199932502336119,0.9156582464714356,
 0.03930041368379478,0.1140025245641495,0.4220815151483076,0.7029030894172329,
 0.5719981371206446,0.7685916410446033,0.6234495734677629,0.486426215270651,
 0.8199056129844853,0.25146470736923354,0.3198210399716823,0.8180903507300541,
 0.42192399083526866,0.09174528626502276,0.47971962271225155,
 0.3864620928431699,0.7847025168045512,0.5913684045238342,0.8028463080910115,
 0.40611649983511955,0.0776310773199409,0.6304449626541965,
 0.13852787713153913,0.5840217782244103,0.6201912766194019,
 0.31044403787543007,0.8673644122442267,0.5577510139842395,0.6467104252137247,
 0.8808341025053564,0.253525075385697,0.26454186389604395,0.6890069153592414,
 0.42250207873026735,0.08510343176083324,0.29602052214137153,
 0.6364690509956096,0.7969957844837926,0.6572850982771571,0.9389946224677548,
 0.963293067302222,0.951858807956957,0.9987593545688316,0.3797618468584042,
 0.9194051043931781,0.912341386293449,0.6702323732057397,0.046184128378887745,
 0.549835400867075,0.4608809271793328,0.1660446143323151,0.6302895711532324,
 0.9784548666475715,0.04003963612518513,0.8707760005286245,0.9317318427291056,
 0.19422354509846884,0.13173374511152314,0.1422091547297416,0.274407762029519,
 0.26581195510072164,0.17086992788838695,0.6035659895129039,
 0.7620613220956357,0.6237261792483514,0.9035863680425352,0.220056624567607,
 0.4865749859412912,0.198841978417013,0.408180670108951,0.9509949072679074,
 0.22004090750071437,0.43013943995144155,0.30147803787151195,0.56992137122067,
 0.5933480166374984,0.12611745791635354,0.4897929128314541,
 0.38952402718134416,0.6468366466249257,0.6908525534371904,0.5001854141338016,
 0.269323164981609,0.29681451493181454,0.6577024211600748,0.8032498118821091,
 0.8698598803445255,0.1005232651903013,0.6960203221522705,0.11931299228457115,
 0.015910573170287212,0.4884388590265075,0.5363855386435769,
 0.9961473841049544,0.7776648671660232,0.10465523714054326,0.3337933296506137,
 0.4286604237430156,0.8709045593114255,0.6905350567858122,0.6278853407487197,
 0.7352416586189698,0.14200327056656292,0.19171785229594773,
 0.14935857836972222,0.8212709354505238,0.660994163455378,0.8693013172293034,
 0.30388045001888875,0.7133128116371498,0.154712913316295,0.5905454125408582,
 0.3211522488990495,0.21131016669684555,0.1308711891564348,0.4080522866278996,
 0.8792659663487113,0.8587325258892826,0.664040666548221,0.8763054562756303,
 0.7981694034601912,0.17028430132053107,0.3623696689496202,0.6312310103611658,
 0.3718716847782306,9.91040850008118e-4,0.6536066330120976,
 0.22005229679949623,0.6924033522490336,0.7108140407678479,0.8556857489503299,
 0.32714465013305905,0.2801460977359189,0.26856646884310575,
 0.04757865990800847,0.5679336427737056,0.7044055103000428,0.762458798047278,
 0.1944623476697821,0.33425794654967866,0.07449542497112682,
 0.3914286440499417,0.19125136212128258,0.17709801336313524,
 0.11668176864671409,0.2651979631882997,0.7094314606914316,0.662537864020546,
 0.628875298328894,0.8214405467414995,0.4616480253140517,0.2942811297230533,
 0.9071591808354019,0.922954395792224,0.1272009994735286,0.4168936421280608,
 0.08179836099038695,0.9264791723620264,0.7545971766459777,0.6005279680833955,
 0.4954322953689656,0.44007123865983844,0.5552010857996353,0.8016840446794455,
 0.3505487111079332,0.9060243977215521,0.26619871725481503,0.7311307640006146,
 0.5906477883207172,0.7991347813676455,0.1634182978852463,0.41189386427659413,
 0.12359121177445243,0.6024632295278769,0.9828818663337717,0.272941776076863,
 0.239385707561102,0.14801127867072494,0.041430570631676544,
 0.5318457566658978,0.05286100309523789,0.4354110959040175,
 0.46748081828248167,0.6649873131590253,0.3777149697257465,0.5455124146020742,
 0.33652955975174903,0.5560562331538631,0.07357446714248428,
 0.37692871803023775,0.5598475234883183,0.2726025960470071,0.8974466047936994,
 0.27215315330206113,0.38133509650312436,0.3670953613912967,
 0.26374339312649353,0.06828685798721157,0.327023211464339,
 0.10183020168997758,0.5858662731240752,0.3826462736523689,0.7223922318858329,
 0.472655537364047,0.8703886920304436,0.6376429487136482,0.979504307534761,
 0.22217325960362033,0.3841658618705178,0.9621547174647498,
 0.40803790074159196,0.5580357403568998,0.027977735356408906,
 0.5103672364455076,0.3263183517991646,0.4573866361853671,0.8994324099529811,
 0.07174110364905273,0.5080125124072422,0.3900411513708144,0.5302627372441742,
 0.39394237707494595,0.543112651233188,0.25192240602692095,0.5838404597962288,
 0.9269892643449712,0.9349220150757377,0.6049705825779627,0.7538203540615487,
 0.6581048757130205,0.35197412336065215,0.4591383495750698,
 0.41004656968258624,0.9068028809041088,0.22933150102423772,
 0.6671752615104488,0.29858457134675287,0.2700961253416433,0.9652129600639843,
 0.4578569866021761,0.4411839052938531,0.4030716532211286,0.03833303752025485,
 0.5182453812686613,0.8419674982812964,0.6376861829740821,0.23138066149552516,
 0.44727475891746593,0.16670109195068827,0.7281087735178605,
 0.9870267669235782,0.08821280652496066,0.9680868708823336,0.012631611926469,
 0.8012716635976144,0.7084124410088704,0.2843362876652966,0.8614602384690107,
 0.29849597422482155,0.45619119377006945,0.1752020289570484,
 0.2979842302720982,0.07624688809769053,0.8882235583620677,0.7799723404289024,
 0.31863727461463076,0.8705229097315653,0.6111432039988554,0.3805074345035695,
 0.9836566208049045,0.8569549909681924,0.762386160984575,0.0313393365630904,
 0.2277302196769816,0.961218318684442,0.6183782542293943,0.5494390539723987,
 0.23831223279043856,0.6093541597116815,0.9051867860342272,
 0.10088893083405326,0.4689323255738569,0.4235692987499766,0.9395968144479404,
 0.19983376573371536,0.6130283952865581,0.26410476016011697,
 0.6751713632766488,0.5951044344128658,0.10729422117252818,0.6447830208162513,
 0.44161668527930975,0.2652755635889801,0.796318122825886,0.9064960587007327,
 0.16824080068332603,0.6499663081412865,0.7321459709894914,0.5147403225990688,
 0.6541295794161409,0.6999376549468727,0.23387886036621464,
 0.13360669955117999,0.9872173762867023,0.7778543643690093,0.8705167724808933,
 0.48054745513683705,0.3502973971958242,0.6447751066122289,0.1979129953834059,
 0.8194365785753551,0.23823074360347474,0.5666914698368163,
 0.19215434053471425,0.17793716270963156,0.9862574721063924,
 0.47715926863023206,0.8087535751851305,0.8009371490718528,0.4427725491093638,
 0.4122840126235503,0.34088739499404297,0.3410556257627879,0.3336724990248052,
 0.6641241932089679,0.08867469343262291,0.48062043276046396,
 0.29626769362308325,0.5565234565330032,0.12105353701318922,
 0.8961080208561782,0.4643995018245126,0.016917286651664365,
 0.48450909585133894,0.8067428384979095,0.4031681776524796,0.6650490308606913,
 0.8691756069510767,0.19854336739163858,0.0889206301684019,
 0.18356103392119383,0.6575277857865475,0.7431879568306246,0.7811348107725955,
 0.9326909626704546,0.3036693975381134,0.8892675498094456,0.10103103504361188,
 0.46434533723182314,0.3688570405234478,0.31774289079441564,
 0.1041696450632843,0.9511452393128942,0.5551610324799384,0.18808393915339228,
 0.756021399671379,0.33098021218509366,0.040253230871503454,
 0.7603656368947642,0.05534372807675936,0.21091579149512762,
 0.9624317432192695,0.7196593136854909,0.4337800125810427,0.24832907018218486,
 0.9111351854800979,0.5918215181923181,0.7066206748636529,0.5392115404567114,
 0.3181635451278424,0.7550627392784992,0.40954009045185413,
 0.38655156330913154,0.13361330691049966,0.8867594039586327,
 0.29572059084267555,0.7163239749402905,0.6419331184463976,0.5437401423948948,
 0.6972318145867684,0.981933621663932,0.6529604605189155,0.2358147602133095,
 0.5368286584272495,0.9695991049307112,0.8668226335752444,0.07644147107889654,
 0.06975352368692234,0.8399555287392815,0.550754064165181,0.9954926256215563,
 0.44632520240906637,0.16756447670604924,0.27452425138670866,
 0.2641696310187296,0.275157507308865,0.6994942536037595,0.1320026434974254,
 0.2403889730171832,0.5849066010544494,0.2839986363140521,0.2481048432170876,
 0.6071107867825738,0.8052839642513548,0.5862869325897988,0.5476295304279302,
 0.9182713499605916,0.5621477751493095,0.0282427894704238,0.4079749410456466,
 0.9615223993678159,0.2109610970362401,0.9284306958517734,0.05727624015170507,
 0.5369930265687718,0.7860448519276978,0.5511398773600907,0.7551623948132167,
 0.44786595299197385,0.7643379615253911,0.9551797613895429,0.7028412155375385,
 0.7963250254968118,0.39087902483159687,0.8833655945487359,0.6226269251717088,
 0.15822034329557944,0.19304257217186593,0.5120130782009167,
 0.060524942925956116,0.6957549367960377,0.923397998606983,0.9852460115581255,
 0.25818091071342597,0.6520398329908131,0.6603648827027969,0.5615277441832043,
 0.48985052181981725,0.2122804539852572,0.6449806917852384,0.4351956147336763,
 0.4167716155603832,0.7411639911109233,0.6950040686995964,0.7431390269044486,
 0.17339149311301827,0.7508388805023755,0.010274404963659611,
 0.35601848081615894,0.08968904566005387,0.30422230676012807,
 0.6606495066760149,0.8822789360299702,0.05539161040505258,0.8871542277748764,
 0.8726379359190498,0.4708829406047724,0.9426596900147054,0.8652102486877226,
 0.8400238640934738,0.26951395566653913,0.3053556912621843,
 0.09434029073598937,0.868090735895717,0.6089237675806155,0.9335652915692049,
 0.06519759354740795,0.11011338092128631,0.8448828648517934,
 0.4567329958531844,0.36750448707131966,0.6601824253499272,
 0.40402957022869623,0.2641153560959366,0.803057728843035,0.4761205592819162,
 0.639791258932851,0.8087429569294048,0.818727312273428,0.06260606242401168,
 0.8420448833100407,0.21667037661389954,0.2679554961291267,0.7661160119003124,
 0.8351023979822662,0.21884415828182358,0.09321435265437839,
 0.24680269342910743,0.23508865489504582,0.9577995385530756,
 0.09143361156411178,0.30112913952076825,0.5212352667207464,
 0.3428554183232245,0.08703210024390529,0.5354268449889668,0.4495799342875101,
 0.6672445141324299,0.1399649263475976,0.0238942482433121,0.514295811481671,
 0.785933863391965,0.9181016845395499,0.6351015004197532,0.7824634426194538,
 0.5658174674695963,0.7444050450285071,0.16552943010461219,0.2222495843289648,
 0.7538506648210812,0.2752757853803658,0.42477012323414665,0.6691336805567761,
 0.40476925376552675,0.4100706925351657,0.5739710147495423,0.7391928103264056,
 0.05484206303436612,0.6149600909359146,0.4412477815104726,
 0.18356956507008082,0.6051289515102101,0.8614399781199531,0.5452627196519981,
 0.6132048650386979,0.3492117553315899,0.7435076050677877,0.5654237388408221,
 0.07811527553499753,0.8851004633851218,0.8200499928042371,0.9130622205118625,
 0.10330916349533159,0.0044792276913108164,0.9986467116923129,
 0.7060306941124963,0.2313682629479995,0.9684294105203539,0.4053411972580401,
 0.8478810749933053,0.6892967261610389,0.04928969155565255,
 0.46107357918688296,0.6417724108495338,0.5135261091291204,0.8036926253562977,
 0.7558329192059262,0.44714647592972645,0.7258174166728478,0.12838505431699,
 0.24849199980078873,0.06751597086302119,0.981446589179488,0.3672786677292119,
 0.9170986673783565,0.1124096121696172,0.35721730698619114,
 0.45550352917682524,0.591374565807106,0.2984566248677023,0.7136742874219724,
 0.2221304491901448,0.3370686168737608,0.1390360408223258,0.14735507589212493,
 0.3757170036049837,0.433725612997263,0.6170296328182575,0.7114362516054372,
 0.6179112951107577,0.2998229833842758,0.8415137241988025,0.8018715827218857,
 0.9661896421742724,0.22258221346344254,0.6406889482526772,
 0.10126600189860513,0.775452162635307,0.3343805512233321,0.8249913445062091,
 0.9940893679318689,0.5223723850619607,0.12331018651803227,0.4419325829874108,
 0.6553516819502665,0.6416006127132922,0.8676616022212167,0.9046385357759987,
 0.028363009693834096,0.9347048430409146,0.6919160353410259,0.756535001442886,
 0.4642218878406684,0.39722539323414663,0.561880245306567,0.5510131390213541,
 0.32197957718220016,0.4930955329872563,0.04649204275645946,
 0.7847596703049805,0.7403660317088026,0.5129112584468182,0.5228620188112312,
 0.18787155360740404,0.20696994847504369,0.7098378953132527,
 4.468165976001526e-5,0.22772104522781322,0.7188342451442215,
 0.15266951339636126,0.19545008720174772,0.4995330243664914,
 0.3530961166965665,0.35086084566059905,0.2179978627269924,
 0.46167302146297917,0.601018750242624,0.3392849765409546,0.11954952890220605,
 0.5101867901896885,0.8798127095697718,0.019066770207805783,
 0.42909783082765796,0.7923758844664991,0.9253435668524705,0.9303864571794541,
 0.2800504082315699,0.27292369774128833,0.6768430344656662,
 0.28419847296652023,0.08201493996144082,0.8588120436235231,
 0.04657272604730622,0.7471527065689872,0.17741637786434394,
 0.23271834800751479,0.4874428364113549,0.3168001646984656,0.4025617973979332,
 0.12095641238359256,0.7475911603989706,0.4908089294059983,0.9631018502987303,
 0.23506940841966062,0.9820206632521115,0.5607510716005942,
 0.47040307256319236,0.45014041774429836,0.18338311254788264,
 0.38717252369101285,0.11286629617713573,0.048013712908453576,
 0.08276974408313831,0.6501114741608489,0.8105678029270654,
 0.20654931491317963,0.8956066998653688,0.12770136788737574,
 0.8226246397108039,0.31427309181211827,0.6792365140561784,0.2327550227895756,
 0.5812185596252935,0.7266122833351333,0.5912854977743094,0.291916183731995,
 0.7569780073239588,0.26247113776162656,0.6764695449533477,0.8162178929253932,
 0.29101776836888016,0.09360742005578848,0.015322705763773925,
 0.8116810164718772,0.398528152196268,0.9435470767885002,0.5183602058108068,
 0.6527498253223276,0.6334122225042934,0.04702413812564421,
 0.37543955474872204,0.6770078458209156,0.7966085947536623,0.6248789195212161,
 0.7335755220903089,0.36490160939807625,0.1391055500768973,
 0.07913349039714745,0.9166165113865423,0.562767366227553,0.16378254070343323,
 0.9749946818624721,0.36904523561645086,0.8444357298979661,0.7858365143500133,
 0.483541473296481,0.1789967023749024,0.29078147481120276,0.9883079529743952,
 0.7268695070587475,0.3974248417579671,0.36092358663572,0.1115145622372018,
 0.6973332028763677,0.2413454373790035,0.6757472971523063,0.8573371287317404,
 0.5340333222906688,0.5846714449019054,0.08845686034790778,
 0.08408051955945306,0.9024181540124853,0.9385072427340393,0.9385630590253999,
 0.9810122815175744,0.71474865547123,0.4807730277446016,0.4053398279007059,
 0.9001034544793213,0.32521348107266146,0.39175180013806676,
 0.27414355324290873,0.3772944306132209,0.40658598566643134,
 0.7943214019057443,0.6820567745998593,0.9836043100123266,0.11396057663267589,
 0.0281086166096719,0.3790065645446875,0.038494449494057004,
 0.4368232871075537,0.3381759344224263,0.5241759350277313,0.26299586581920575,
 0.7069693170562676,0.14298767371336485,0.4175977654188423,0.9411640061326394,
 0.05966195203188296,0.8419216726046401,0.7744299414551314,
 0.06636955449888715,0.6783708647253597,0.26172328963802727,
 0.14864158741268185,0.9360505391223924,0.9661612495761269,0.5577929928317911,
 0.8438369009391901,0.37520854832563233,0.9219265177193907,0.6092027927604726,
 0.16344293283131006,0.2630529720800341,0.25438601300602726,
 0.9007557051630989,0.37137418305983305,0.038788023648545056,
 0.32283183052772335,0.8150556006599449,0.22324917687753087,
 0.027893642185325196,0.8591592905113417,0.13790189664991892,
 0.7652766858084603,0.6108176683037094,0.08200509667841294,0.9253761913188622,
 0.7794065360573881,0.9355654599135959,0.08393975509615526,0.5293169816673047,
 0.31506164603352627,0.3945577344488653,0.10848378885606402,
 0.7694165440864936,0.950105230315559,0.8184018743725034,0.11240530713133401,
 0.41420730496799074,0.18303747476939203,0.23860956891962426,
 0.549117487010175,0.343456315214,0.9063304959636486,0.6235228147402138,
 0.8893277671479689,0.26333022532783734,0.04489671071565815,
 0.6718843585665207,0.16449022751407627,0.13085954615047346,
 0.4835222896871647,0.8695704198979571,0.746290918062026,0.6936858107690955,
 0.1539471427034128,0.7091156531036159,0.8247177627330176,0.19681706941345034,
 0.5364638105625252,0.4706316033484973,0.2567756254123823,0.1541666100800232,
 0.7814818723534527,0.5908922759826353,0.045656078182926385,
 0.20759229888558717,0.46216373180478165,0.498317571933089,0.7799648682609407,
 0.9679568384289597,0.5880874234174385,0.24498708605535768,0.2248127780963025,
 0.916742595499675,0.8120293029891489,0.6242280181971711,0.46052975679835473,
 0.4059784764800103,0.43386740371742916,0.49969765508773645,
 0.6645706983233362,0.5532718429018135,0.28503276797240895,0.691217214513488,
 0.20137844041894815,0.040049882334300624,0.6954757261514987,
 0.29489374394199397,0.24481213270164104,0.6366135154923619,
 0.5286374648773856,0.7149424262847413,0.8519113655421882,0.8877883209508627,
 0.43808579212143284,0.6436107210452605,0.2274454025942345,
 0.15149896380523875,0.8567818034658635,0.7175739438499406,0.8239351934163711,
 0.19154176236421883,0.03656253210726623,0.9726544829826493,
 0.7817641955426861,0.567955664006083,0.10685931601879561,0.08836167164603737,
 0.1349969002836121,0.9775107692604358,0.8159581275643335,0.29840438285035553,
 0.06160398644645815,0.21903806546018856,0.37816506266652383,
 0.2466987377584674,0.3427761005185286,0.009412525160054419,
 0.4180243405849621,0.4971957956054982,0.4976304348568481,0.09737222221393438,
 0.02821896031318316,0.3532762995854749,0.8528871262020365,
 0.19636881032061126,0.8881414730587596,0.3133299862387372,
 0.05202590144086505,0.6451519746983876,0.9831242971146976,0.6893356145322727,
 0.5495906349086753,0.5368421597854757,0.016484111163463044,
 0.048817671036725185,0.5144884888894494,0.07183828574773476,
 0.739324618670608,0.7038708564923207,0.6888031015318612,0.5580554993568079,
 0.7198371613342626,0.7547091262281571,0.12035468577979314,0.5263980438835498,
 0.9462156618760427,0.4441318815334404,0.5022327315044725,0.8373553741145899,
 0.2593569481270528,0.14880243194127574,0.8292957496318426,0.6064098335261189,
 0.06029710811335487,0.6125889817203256,0.5569406882270226,0.3608397042203948,
 0.3936430644007368,0.9497626114411006,0.71914087500968,0.2595172306965159,
 0.5412930104208109,0.07303840953196938,0.2341510314376365,0.430530741898732,
 0.4266052749853967,0.1533612899886192,0.6036495617557562,0.14043487910640984,
 0.4898563968852393,0.9943723174939805,0.26397373162283444,0.3824058668404794,
 0.6447890112359562,0.6324417485025252,0.8454310439811101,0.282237633711348,
 0.6470185535831746,0.01622064871168305,0.012780354554105644,
 0.4721846834330722,0.2468693195643693,0.971399598509651,0.2270058243982147,
 0.28589485001240944,0.7252638244361413,0.048374503348973974,
 0.660332129595723,0.026712365957246353,0.43170116234506806,
 0.5760775645160741,0.8098621423056854,0.9953533662522641,0.4135981588789355,
 0.42020682904951434,0.7993160169711808,0.5834303145463067,
 0.11619670226471912,0.6997499778597511,0.6071741328128641,
 0.09600283328987369,0.14664458725379548,0.6722107052441211,
 0.8244209251435836,0.4201599056760278,0.7347079920840938,0.573995503449114,
 0.6674348617561076,0.7474038494489288,0.4323910331779651,0.3325403537772207,
 0.4075732114359155,0.26509874213168416,0.3654977767554106,
 0.07454643331994015,0.6279802726907896,0.48560362494824405,
 0.27542624035940544,0.9524563050842126,0.20152672932961602,
 0.2543699701320773,0.8003281316473945,0.16970036673418876,0.388551624268764,
 0.9766264303603783,0.182185371046365,0.15185156636968467,0.5875078072415043,
 0.2157118556341252,0.22019051217415542,0.3819338607346926,
 0.06859165392343658,0.34964042785088456,0.7391336662979997,0.93074429465329,
 0.5338237197719649,0.899988698997087,0.14622510034505531,0.7793900482236062,
 0.9430983803966333,0.8149082634608376,0.8088374404542428,0.22291318851677122,
 0.09984192635302613,0.22609066876003747,0.9572512385704963,
 0.3506223359861296,0.22631375000637632,0.3618421053274381,
 0.20945635620646186,0.6310799319116556,0.5992337646613378,
 0.06698670830228515,0.20488235044595804,0.3130392400640105,
 0.5847236222750855,0.023147586665548725,0.8583590277926453,
 0.33764842562553055,0.06627827611081227,0.9541161372478875,
 0.42080056089590745,0.6492064298585705,0.20059550034881202,
 0.43493804739174013,0.7168137709238721,0.9190975500910703,
 0.32786120462909296,0.004811565744260893,0.11520897901742777,
 0.6427439298956149,0.32279834029046706,0.8780375300996166,
 0.15204447304261737,0.009260102701677764,0.9491004422732418,
 0.7526907898626835,0.14852057836316246,0.6487837897011265,
 0.02756744305325942,0.4551864418628735,0.3758026570906502,0.7533501198594219,
 0.8969460450423135,0.9918995186067274,0.9701572370096827,0.11826635709537481,
 0.47224773923833097,0.5193829363703578,0.9914060880811102,0.6213221803755705,
 0.888761408240936,0.6582167215754334,0.6465912597956689,0.5380453934796556,
 0.10427095572421496,0.33267193194132927,0.746472951017142,
 0.26054551223657474,0.5006674091742261,0.8974939697584805,0.8599960738438807,
 0.9740577886217706,0.6507795812340488,0.7008814976797829,0.35918923406185777,
 0.7364217293293632,0.14781445501853363,0.3871835884024185,0.928110100785482,
 0.15884377116137327,0.048747790170251326,0.903986236782984,
 0.03393592346947283,0.032083628914858386,0.7801428066048115,
 0.05919499126482963,0.5181346919588424,0.5776908924202891,0.7910547608174465,
 0.4859203120149602,0.45512887308374494,0.3752089283175848,0.9429247021160102,
 0.5574674760879496,0.6495487216101279,0.6533415859455991,
 0.021454484477512192,0.13976524076795882,0.7466198604746656,
 0.9169384133874668,0.43279937561582116,0.6704938608546249,0.5171077699218222,
 0.09824840404940816,0.6587761517551312,0.5089914319707126,
 0.39387931908064155,0.0935734670424242,0.6036907581966724,0.720230366620747,
 0.29890375799403524,0.1626104270673887,0.39470203340060284,
 0.04873487350802641,0.22352175612770675,0.2097756104073154,
 0.42420945182723646,0.24289901912816747,0.590733648281567,0.9371366598194257,
 0.47333052261775965,0.41078259212721235,0.7813586639193907,
 0.21966699511303345,0.30034269173433814,0.5536618189068583,
 0.1324294418048808,0.0383959410878163,0.16522137269967996,0.7238227264868516,
 0.35209713252212094,0.4383163128309303,0.09957696792326487,
 0.7522095618292679,0.49706184315846813,0.3985286790875109,0.0488951476346734,
 0.13294385143717435,0.20886970087125467,0.2408333035924458,
 0.7382633509841509,0.4224864338536495,0.9040374788956104,0.9549499634399036,
 0.15534029687593065,0.3087831441108947,0.9952890489841099,0.7114439986178729,
 0.008779398620310075,0.3595354726436567,0.5116888006090704,
 0.7262777111402925,0.624499617859003,0.3667312351477401,0.2824289534762392,
 0.5136167355400627,0.26218597882803807,0.4269688518161644,
 0.15147413652875819,0.3300836550230065,0.29021022912430816,
 0.8906637024839557,0.6453006158067993,0.6366892741549368,0.7966150343360546,
 0.4413203004875491,0.19874866320754447,0.9772266528315117,0.5310885456677104,
 0.9563440012965623,0.844907320950364,0.10566558529712367,0.4070432573908458,
 0.29306702934066653,0.035311231919897934,0.12284877637112168,
 0.32888742291563156,0.8261616251690734,0.8975274964922089,
 0.18729839686727467,0.7342813290471831,0.25756153122064274,
 0.6597177376466575,0.5795912996354962,0.37601295696174963,0.8618472940048713,
 0.9778261958538632,0.04829374047444979,0.24953268237286652,
 0.007483413973425779,0.2893471154621754,0.45523934211208705,
 0.34782403350455704,0.37202692962429784,0.39304576084129716,
 0.9781295022213503,0.5578136474784133,0.7832438401637675,0.41409984704245517,
 0.8281424352433869,0.9151359470123502,0.95438259821912,0.12196494982917216,
 0.6624776399709953,0.4844132501754579,0.5712324768550492,0.18049126257577586,
 0.2546487382078316,0.06657267140240875,0.02926882369851369,
 0.6099477815342796,0.47097496661214955,0.9061429691908992,0.7299457987899094,
 0.6329138657827522,0.7809120218982908,0.15838845961637527,0.6489196935329167,
 0.4783050062924411,0.28864747783729716,0.9191927662958883,0.3805703445743259,
 0.1959255662109567,0.9497335208225028,0.6094981437036704,0.5856915097274027,
 0.027014010972988167,0.1463636931342277,0.6595853241405103,
 0.2986159203742178,0.025427336485263963,0.7730110087620965,
 0.20265684783735616,0.5183444708248602,0.4899528878085483,0.6603676132357387,
 0.9397397741685622,0.2408320489755046,0.9902301253316771,0.4847467519734756,
 0.33820717597558536,0.7593449498307433,0.9218017004669926,
 0.23784279150992416,0.024195630668899182,0.3887463826826171,
 0.5205207484749441,0.9488816017904181,0.08513725716385778,0.604785534485164,
 0.21243090196943348,0.8378509789450134,0.23291036076105154,
 0.34752818423575527,0.21657067484432435,0.7292042779659962,
 0.028506534458754063,0.2954621259382333,0.625319290718213,
 0.015060150666340144,0.1023405545368008,0.8711768020259077,0.823221149593737,
 0.775214050190344,0.8167848866446802,0.38868793165491655,0.6960484190066589,
 0.670441112202019,0.5737207702389334,0.9879479158628515,0.2094391812075166,
 0.679985420897186,0.48847705960009724,0.7193500094479187,0.2709975819718161,
 0.605959271542126,0.3352645330188324,0.39171783062255494,0.6149634221360474,
 0.017917684716442506,0.9199385047695819,0.22690106835157708,
 0.9032834042008006,0.7609103288681002,0.35941817960225575,0.7966794347325414,
 0.26191397076321865,0.9878290728315948,0.20137919225284917,
 0.9944043587481242,0.6813356152002976,0.6647883420880097,0.577618328270851,
 0.7680679192590769,0.24727424525664565,0.5043971809901804,0.6611105636129796,
 0.7246836582218256,0.6304970756667653,0.1166108364646048,0.4745914650113894,
 0.7139678810946526,0.3217968364261816,0.8377279318386199,0.5735364130805549,
 0.6672473313756866,0.008799500494256263,0.5278285315100437,
 0.15590708280971266,0.686769870764085,0.614762026214423,0.7535363123971798,
 0.7141547360945563,0.47576615413269974,0.671857349802373,0.4814175687868456,
 0.10390347845175552,0.9859551665907241,0.1942596507580836,0.94474944334213,
 0.5107143942787098,0.6636794172329243,0.6322507551748178,0.7173319508289175,
 0.4250827975880467,0.14424086283958693,0.0808096878461847,
 0.39312230423101424,0.9437467497230336,0.2803303516615234,0.5554624636976144,
 0.475358720341139,0.8573867459895486,0.8780721504353759,0.5298079551948776,
 0.19410919106551283,0.19414413632394578,0.4309129800947715,0.15037497169421,
 0.6400612099605716,0.09256427800833023,0.49881118114423817,
 0.7652229973759527,0.8517641719955016,0.95622860049775,0.10444049437826552,
 0.18767597920975732,0.8211594772509899,0.7111441531479191,0.5784440406480548,
 0.18326574781960348,0.43492492021774976,0.7594830915275467,
 0.44744848032863926,0.8171855943669942,0.8397358543980186,
 0.45489686347336944,0.3447553054836121,0.02077520455438775,0.945746083423889,
 0.8572094832293135,0.22784887981709934,0.9503467097657288,0.5241690389862576,
 0.37996735770381573,0.5609730057724155,0.7581112608853393,0.259529816917917,
 0.7130836945701283,0.4251879277023517,0.08180329071464565,0.8616800322017238,
 0.0678813011446574,0.57595268629155,0.11335725331978241,0.7632249875418967,
 0.3230584247384025,0.9842887061874236,0.45060369782551557,0.3025725725970607,
 0.21400679410726353,0.40312320658570455,0.08965633674755924,
 0.6673806032076001,0.6363731180571919,0.7898902593355586,0.03905115771175416,
 0.7507057011800714,0.5271285028876319,0.9090036197600434,0.3302395583135803,
 0.3748742148179647,0.7071969025561959,0.3293894112642496,0.1203372365669747,
 0.291767610125122,0.15277724890968702,0.9946603001603052,0.9552709899347125,
 0.7231413198087351,0.6445531824893552,0.14163752347826086,0.6422947341132946,
 0.47272680264771016,0.48033089370633,0.6348924832298509,0.47958949870603773,
 0.5791352385075503,0.9485530491251114,0.5706863684165806,0.7852421485302621,
 0.838348594243447,0.6390543118993883,0.03510034365696213,0.39487705285378827,
 0.3610678251169176,0.7634320599294525,0.38539056304396013,0.7362167159878324,
 0.442965579235056,0.3865958533233036,0.20330619877497458,0.12244458291382099,
 0.6832771762172285,0.45704133103287126,0.8382384502715589,
 0.44061652589549916,0.7519516521012275,0.4596499908176681,
 0.012207210400322022,0.15553669028465422,0.519149244191066,
 0.7708317450721704,0.19540126077688114,0.9920658687815802,
 0.24264813680239972,0.3011728520928989,0.48529329685308364,
 0.0039611306548151815,0.011096799624207732,0.4937105595735861,
 0.06985325658903474,0.8282504062282653,0.9722022352555959,0.6069531598438613,
 0.5253775975795179,0.7419825040633665,0.9316378123399509,0.27526653882187735,
 0.8561308748597503,0.8424307896021042,0.7428419756636656,0.41890046348649934,
 0.12473674309869087,0.4840043153666268,0.7258871463623852,
 0.07616108725320148,0.7345524916534755,0.32127435866847565,
 0.15520416013625604,0.6785475529487086,0.04119967124466084,
 0.30392942335387096,0.44375108443457945,0.10493448879206879,
 0.8400475462011875,0.10533144556771712,0.856954001284283,0.5122801506270036,
 0.36264015436312613,0.7494941338673864,0.8167539651713942,0.5726315144566697,
 0.25476677815564086,0.2523868409259389,0.8842012056259012,0.6945354448686338,
 0.5394414535521959,0.6765571014790209,0.9802968899082796,0.2942631779573627,
 0.6080107319450955,0.6463086483190636,0.9007986367729461,0.2259298326348942,
 0.9966856385893285,0.7946345313751717,0.7372191626716369,0.31978840797931884,
 0.03582317386307898,0.926541010963374,0.9320934975556949,
 0.026960331536561633,0.3734131415219003,0.03405001201072311,
 0.08744067058588434,0.29496699756905564,0.4406489187458902,
 0.8852264600598275,0.1600810139336376,0.7286112570031624,0.3171210771918096,
 0.06826679895716958,0.3345540860872487,0.01823231434477912,
 0.7793229079900877,0.010668199564599856,0.23816011714968166,
 0.028034759406669663,0.4880813225060979,0.7793565549494399,
 0.9155949417417614,0.16860142620652618,0.6104733011221916,
 0.11408856478407003,0.07713623751113874,0.5337484724214858,
 0.009268512528811224,0.4000276674756623,0.9308345927422227,
 0.29783009234687186,0.8113324604435278,0.2022732224780348,0.3800348660145513,
 0.834688717094496,0.635173902387798,0.45213524800528426,0.47884043096310125,
 0.2565413262344397,0.801338028630041,0.2886295832565412,0.11853847944157325,
 0.28129919307185425,0.9803042945039404,0.44196776474766897,
 0.9775150989038464,0.36117869493382004,0.09275408097626947,
 0.9177816424295548,0.5948129603189249,0.2822662997418788,0.3117825610276075,
 0.43981849168559417,0.9396683283022081,0.5365960503684329,0.3466805343178574,
 0.7672171292281316,0.16722417258512434,0.6200267247619159,0.3765730153743543,
 0.9632085394035812,0.6278084902764357,0.2783678567573701,0.9645491107098324,
 0.491806971792328,0.4671579200584465,0.534017888413101,0.0665433662116297,
 0.9795692119231598,0.21858156996934774,0.8478299777558252,0.4321327947692678,
 0.5145970527736788,0.9996218404005811,0.19586719257617968,0.3564558282753677,
 0.6932787351536046,0.5250853866341898,0.35546027121948454,0.8923599829642007,
 0.15551877295537841,0.4025664666434531,0.7619984895719909,
 0.11525435302924547,0.037379881506048096,0.841094441702668,
 0.47028943326955386,0.42186005168587193,0.6018335322796432,
 0.8125044193012203,0.8258907968114103,0.8055799893802545,0.11676391093913352,
 0.3844656998372382,0.5438519637270687,0.8944097162171168,0.47856429600758377,
 0.9375683764409577,0.7665794077875697,0.5475151109865263,0.9268326658033796,
 0.6706177287125654,0.9860138071875224,0.2852206897683861,0.4706323533727871,
 0.7418890500404334,0.7497859964674696,0.3410653130746959,0.4667042701928592,
 0.2833084576754271,0.6010923984556301,0.8812807517050482,0.2659687898464397,
 0.90504219701311,0.8325155523057405,0.3174971336874858,0.2917990077876027,
 0.9458532730177132,0.49739454048315745,0.148288631659532,0.478916945650783,
 0.5999718267367148,0.25199127429126766,0.7859127034127054,0.5318542514590192,
 0.43585421941192204,0.3358122273892782,0.013516190137373929,
 0.19580710365221965,0.49556664492985847,0.3791703101992354,
 0.40395000495150957,0.6252271144745625,0.5385874911616311,
 0.31630196991914517,0.6905155733863614,0.20173460428224366,0.064872682295949,
 0.8063923303857059,0.45225489863075996,0.5814641938886452,0.9860263384021124,
 0.03474731978705625,0.6172195898774742,0.7150053276233987,
 0.23569787223883565,0.6828927570094082,0.20694571528125938,
 0.4894782731867774,0.7110618139543202,0.6383563917930368,0.7889577070645918,
 0.1911767442155763,0.9769961303320428,0.2952287393563039,0.9968111826636894,
 0.009347916399959688,0.3796182690804041,0.9880951414253932,
 0.9163183020522809,0.8542686466875439,0.2321800778221348,0.77462347859732,
 0.9673883164683723,0.7863176826160009,0.8511224758726247,0.21845663847870245,
 0.2844229456645173,0.0859925933768535,0.27111196426989326,
 0.46806933689690233,0.6266765323319822,0.6460532291074531,0.5299668093237254,
 0.5523243940417674,0.7148585415250722,0.9380904478528996,0.25597478119255823,
 0.5367920004901686,0.25607644657363193,0.38270426149151926,0.781968166649287,
 0.7041947611549909,0.5938289632495151,0.16073873268961636,0.679880795976129,
 0.809857237873792,0.796217088292563,0.9107457188315283,0.29215748465827285,
 0.5249673206652847,0.8970763766729118,0.8763327628399491,0.6121445026116792,
 0.6258443314043083,0.10847582291821012,0.9837742980410651,0.3306589860873175,
 0.4229084444807525,0.24298752810749424,0.6437840538054814,0.4501587807623908,
 0.5919420163098525,0.13704006558469783,0.2372308360019697,
 0.0025741300493087316,0.35950136005393096,0.9256677237540618,
 0.6943067800273334,0.09277650910258639,0.2089726171052193,0.8840347590936841,
 0.2607851841004718,0.26475958572996183,0.6198763258798939,0.4403003873342015,
 0.11171613465329211,0.3471260380456056,0.40912130626442955,
 0.3376222543026014,0.45071755324191487,0.36361107143205507,
 0.15713131681356818,0.13715530430028,0.4740995649687654,0.27864808034320976,
 0.11167458825367393,0.5367723792289194,0.8752375067102517,0.1027561174487428,
 0.44002424542029495,0.05634109627184092,0.988007773829055,0.6854026083530816,
 0.035102710200502485,0.8794280468196883,0.7360175250147265,
 0.26025874364273915,0.9967202146720527,0.11408491425588063,
 0.7415201066269117,0.07275689005125885,0.7959395154591826,0.6989591588465044,
 0.8549755586639076,0.9777678742090643,0.1182632705488903,0.45086053603638776,
 0.3302887712232254,0.05634797289149718,0.2678168472740572,
 0.25568042466316987,0.9737801122861016,0.25341542780073767,
 0.30324329839260744,0.19406596939699627,0.3667430779174684,
 0.48294749736083487,0.27937633148719554,0.9087998581687012,
 0.26828317620442854,0.044013632388724466,0.6072033523224201,
 0.9671720052016202,0.7001299249040837,0.043596118476032064,
 0.10081710736072602,0.28010144493607236,0.622059680918234,0.3252597070655687,
 0.11766977893247532,0.22047312120870255,0.11521744506275222,
 0.9331407395045985,0.1802048454222157,0.7927729351746986,0.4668453524388808,
 0.7461793517542703,0.45002803293606874,0.8874818709293515,0.790701313111841,
 0.19071426376334144,0.2926360918847768,0.4754970313255593,0.3066618462437697,
 0.0452158892325385,0.9047397563985964,0.8260898339230938,0.452279328379778,
 0.7538405401318135,0.17659805239463877,0.4642704944093857,0.458457050591081,
 0.5674809253585897,0.8282233199757078,0.6756884114087529,0.3145615687160189,
 0.14832312532287684,0.3645119412088166,0.27466782275641854,
 0.017298422000904368,0.8521614714239889,0.1632795645130074,
 0.36647396473273164,0.33088867420207324,0.5211191370327319,
 0.5360485936730214,0.17333807757706154,0.6026430781435738,0.378027646767389,
 0.04584489614712606,0.18634228958969312,0.856599017819817,0.781401262020502,
 0.055817178344185425,0.2621733165480873,0.20336896551327177,
 0.2000234131202383,0.934480181728967,0.08832477838823505,0.7704236575493473,
 0.7430251931798195,0.22857098999606051,0.2681713641378106,0.3556348202688332,
 0.020181438231392623,0.05969524720627095,0.315624372159229,
 0.34085291164782194,0.5635213849266607,0.6578537996081305,0.3256008110468508,
 0.176454941538556,0.9516573360589978,0.7646993398041927,0.7954745203016175,
 0.3060287563290376,0.09364416588365054,0.14693230352564135,
 0.7009840229723886,0.5714022336324381,0.09428235509600147,
 0.06048150667980712,0.7045391716230301,0.6185927898447547,0.6727047530761471,
 0.7183409463195773,0.4066012833613849,0.9695908166123519,0.6177090967957152,
 0.16003208752177733,0.5325207946995114,0.5108712766579514,
 0.09810903959973594,0.6705512389448262,0.26563509856423395,
 0.4846569753650287,0.8486576006635248,0.4042897736050054,
 0.032447125789756504,0.23961576600600054,0.33301218304066404,
 0.5938219045117432,0.5282492715033571,0.96745064812912,0.35782652828168593,
 0.39129529127453166,0.18668731219187118,0.9205804067703418,
 0.8087464807180167,0.5010685574767403,0.363162777133878,0.4542592208334081,
 0.08550525116683516,0.09847936047671924,0.4279198968259138,
 0.11322634815337196,0.8333313045995091,0.9074974677946107,0.0263714919908965,
 0.59617369270007,5.676264740142045e-4,0.6210324583786831,0.710867191629955,
 0.7627064631543222,0.8941941893832283,0.7579934799554685,0.4299312423319338,
 0.22018842935189464,0.7040561072075039,0.2846083587403996,0.2907552086519518,
 0.9538898455625362,0.3980564985494043,0.46896446040946094,0.39579682811176,
 0.32237362015948756,0.28272605078981894,0.1618716022711184,
 0.8526875722886622,0.7757020193000849,0.8615283404203716,0.9668493817627352,
 0.36563604848866316,0.27518922353452724,0.20289193208821754,
 0.6546745851130288,0.9043993362082592,0.526403167670177,0.042269542478747035,
 0.2948245358946532,0.21459000619452995,0.4610965723658158,0.4743839144614913,
 0.6573913488661479,0.40485810147667145,0.5788107990394071,
 0.33594543013747524,4.945454679360362e-4,0.4548159969093666,
 0.027533674332121394,0.7690775741137932,0.5609744888478125,
 0.4145942612473117,0.8152643813992093,0.6640625241690737,0.9233474392262454,
 0.16070148522166972,0.09098672447515943,0.8408156164654939,
 0.3065881483805887,0.9313845069254305,0.0017068163237261569,
 0.19115994677628478,0.9674440860381337,0.24490433527074784,
 0.3482338147996197,0.4060681392225871,0.363934019676744,0.885183793628291,
 0.5776176272018075,0.11345203452933894,0.593721738476354,0.507715193364897,
 0.22728232592103448,0.1584537934082615,0.6049495625206498,
 0.09144233161972792,0.9540833432637444,0.8155386645245972,0.8845030041684457,
 0.22915863342940412,0.9224825814464674,0.6991147423902639,0.9409234592203933,
 0.3981448893895797,0.16431916536250513,0.11343947505687169,
 0.3175462127558927,0.9735188983188676,0.09253692338190589,0.0203903851468199,
 0.6560820919142921,0.6971123162425753,0.34689370281163745,0.9329257590189663,
 0.933877022457522,0.4384035154882807,0.18094313321103972,0.7914677037847366,
 0.267518954514813,0.9217462032820382,0.3964825419859105,0.9294485155082252,
 0.5483810227109711,0.7641964563789101,0.7617222286824364,0.38626002575162866,
 0.9694185513076407,0.35455549694358557,0.2947034897322246,0.5446916212306951,
 0.2622667359002574,0.8224266112111126,0.6104177562629924,0.7448997371694085,
 0.4510991639198414,0.9125717853815627,0.9129803418635981,0.310963439866207,
 0.4113253291754366,0.9936912617349858,0.4506005172688746,0.7908048773157959,
 0.4785502701458665,0.20293020184645016,0.8295002024476801,0.5380247235191501,
 0.4301088945557695,0.8598184092673535,0.12803897797257136,
 0.26846448822861513,0.16183511499971948,0.6694078113549013,
 0.3150829125502552,0.41843709377718274,0.9782430788424127,0.6765458020128953,
 0.13713828680002826,0.7808192757520491,0.43913588798059444,
 0.03505995118102834,0.3424376608447368,0.12069249890478273,
 0.49996912469838195,0.5495115351034581,0.7729077747584034,
 0.27809344743336906,0.5514565393709829,0.3401415832677751,0.7272342815248615,
 0.16515467819804297,0.6367377060132167,0.12394249376454491,
 0.44666933225011896,0.6708990851145021,0.6949598141791989,
 0.26589822949124864,0.854270517854213,0.3597811037631895,0.26723953258258537,
 0.9276587007130557,0.5022968731912768,0.6522379374189041,0.5837203743420165,
 0.8180032972592067,0.37978647958633427,0.03302974105667045,
 0.22731300063208837,0.7745903520136354,0.20975703066609874,
 0.9027743793709573,0.26549062131648404,0.4541550891189975,
 0.43102260636356005,0.10623028769580534,0.8996445176314714,
 0.4340831316043726,0.5432466148012882,0.3232937448301789,0.7804408080074355,
 0.18409962276358882,0.7495843213332215,0.6009639449232924,
 0.04785299641140417,0.19177850152708054,0.743762978533105,0.1852922186428403,
 0.27798103255033124,0.25676368370185554,0.7741311994481591,
 0.7603840851100651,0.9276151269415189,0.12825923426022368,0.3752070301794701,
 0.8173442240062269,0.42705539558104044,0.9190730402050239,
 0.39137646397869685,0.35477242766551353,0.13613309757713965,
 0.8201542647697944,0.6084168765971798,0.6543515001284471,0.9911245983612414,
 0.20700891723847137,0.7308233568742507,0.4477040078145734,
 0.05623395032966694,0.34661334401300525,0.05728266426131312,
 0.6604174045605244,0.33535828692043645,0.09631861530609787,
 0.19690422222897375,0.4172396226382705,0.5399880184159898,
 0.05004869210787399,0.7408076030510674,0.585643708191675,0.3901505308513503,
 0.1450892379334603,0.621636502288377,0.10796078550074051,0.06956151952897816,
 0.5673393472766721,0.23171366527774495,0.7318790761454198,
 0.47080061327591993,0.7360397989267757,0.18393139434137873,
 0.2599507672723187,0.4211971878775753,0.7509205707286603,0.2587863507344702,
 0.3196435193872682,0.6331257438607856,0.9050568040969321,0.5088325509784944,
 0.9815420725368484,0.09405769136062703,0.0662150944166463,0.8149921003186951,
 0.19689851452701934,0.2518233865378805,0.9902517549500985,0.8897801883427868,
 0.7814609263000575,0.8732705338871203,0.36515219579157454,0.7066953773678544,
 0.9918362989044314,0.018355107194848785,0.6734226733317719,
 0.048515100623931495,0.5444639216044743,0.8216361212986013,
 0.35855399872539473,0.08539894936041426,0.9765535019287229,
 0.5446694240885408,0.15345550820445542,0.6144610895928576,0.7512730648364985,
 0.7860618895551347,0.22240364138798585,0.3119377662077699,0.7591454349931561,
 0.9836828961486203,0.9298667215817613,0.7195709265692746,0.7911491333209131,
 0.6910399338278841,0.026656275039836586,0.1970395071197103,
 0.14188988390457746,0.4293720695775449,0.8701426313888533,
 0.07015952007430515,0.4390533992015485,0.4858490686220983,0.6133239082675912,
 0.6537239179083271,0.5230953972728284,0.2798980077346095,0.4332196026657942,
 0.4723848335819718,0.9294800180774277,0.8003324020527537,0.9381539245112811,
 0.6407210345857769,0.6972391835909737,0.9953022414333045,0.6959968707529939,
 0.4793496542299487,0.3617534049980673,0.8816997954685479,
 0.014387130795572634,0.0029492159469671764,0.7200686074091542,
 0.13181407002634316,0.6086473372356883,0.5495203260479417,0.304367930724706,
 0.24655386149601055,0.07604372914175994,0.062115884883275796,
 0.7906689343894499,0.28145541853649325,0.6683261494217475,0.6918532408607514,
 0.5781895893860631,0.6906232868565156,0.18678977031141508,0.7798398190687551,
 0.10791292387863971,0.738659865674547,0.07681075947719718,
 0.04775710918894038,0.03445100967722914,0.6695891003123237,
 0.6618923472105567,0.5243714705198048,0.9546755500798015,0.11479390322814265,
 0.8499180920629587,0.3741146243519028,0.6050074510428977,0.7079089405059628,
 0.23694294175754038,0.8040099256530747,0.3236980223808166,0.8732504303662592,
 0.3456687600871875,0.3547059085641857,0.4417795417614554,0.4557742789423075,
 0.8868082691764193,0.04847729842451709,0.7510085739290694,0.8063929485428947,
 0.16875317487302965,0.5377478960364251,0.14612712919379267,
 0.4674309649002546,0.4433866614852803,0.11961972343223015,
 0.10398591283378056,0.9014395124507775,0.2778921009271407,0.6371236670003132,
 0.18208784952152512,0.02347525421340202,0.19114379410314797,
 0.7908088847813849,0.2514891702630886,0.16620301752880395,0.8755360924003746,
 0.6202113830996973,0.19691412491494875,0.18331004903692916,
 0.8299342149445312,0.6297410748295927,0.34853517881220764,0.3127914092190496,
 0.43157653969376564,0.10086883343286335,0.20158202274394854,
 0.7636587151425483,0.7761590314995206,0.3506191777347968,0.9354946978139551,
 0.6360023793472046,0.4097344287656086,0.02690186160521124,
 0.40889798138331657,0.009512396625088404,0.5740789552881959,
 0.12724968418211924,0.8504225605030113,0.8879210589843028,0.6152597611941053,
 0.9554726644881176,0.6530048537142332,0.9993108203368912,0.36372502530702644,
 0.29502641104135197,0.5294334481384457,0.24857745795955288,
 0.052806938397448766,0.25129042214656894,0.04597170617165747,
 0.018664465619445236,0.07390564414412104,0.6075142772788327,
 0.6189444544800682,0.43262059416346255,0.4755226929622991,0.9117326618742535,
 0.5041326594829492,0.34134683860855286,0.007845543743212002,
 0.6001291081245794,0.9313813779048505,0.9533403591289856,0.05544857942877013,
 0.6029256538388083,0.4152188420994637,0.04243266752627839,
 0.008340481180927672,0.6908675686500282,0.29246724678773084,
 0.32518556299232404,0.8615767168820849,0.4224099713949411,0.1426828518853226,
 0.9129042446289607,0.2054713571977136,0.9809622640431289,0.4666935770914269,
 0.5671160530194597,0.797273322255652,0.1915946759640189,0.4041684211022727,
 0.9039093979059991,0.7677032756613339,0.19725245491214394,0.5522501642944143,
 0.4156661717162996,0.602795087932193,0.3554943363797135,0.7777146489971124,
 0.7848232183443539,0.35071026340147193,0.46358813303832,0.14480102726292066,
 0.9648771796076328,0.8601577516698842,0.9186714723296175,0.205002750121134,
 0.18391555424762707,0.745885267769322,0.7150009021126129,0.6911493079501655,
 0.6868408200582568,0.4938271665976318,0.5538097250686251,0.2181230025082983,
 0.11143469900564906,0.11548335114524932,0.8098142109654569,0.692240379141587,
 0.5737172966071262,0.20751702883427203,0.23009953479639111,
 0.8820927092272453,0.8663533204368029,0.9239443264537512,0.30757235596070764,
 0.9076563719752585,0.7943247158882293,0.42352757905367844,0.3254940843639901,
 0.7375418893344292,0.7840107103577345,0.876662522296295,0.597001722581428,
 0.38721611610270046,0.5972041011719467,0.2666150307293663,0.4307768281199823,
 0.4518832751245847,0.5015817743715081,0.6806084192412227,0.0967117423998094,
 0.8399060283424371,0.363518879594656,0.9558202643083347,0.8192812577902104,
 0.09521012124798467,0.4692686807959727,0.784233139750647,0.8180538183865592,
 0.9543685890214113,0.74772467922971,0.7314088929343548,0.23682923743131035,
 0.9791711865799237,0.1831259301957726,0.7805930483062361,0.7848660475721994,
 0.4940471814296521,0.006832969583368209,0.6469103340541659,0.850369064831265,
 0.4538180859704468,0.9811546375144656,0.8463326566451096,0.5057133321589867,
 0.30520435280376534,0.198217397596399,0.942230820722642,0.4584561826424074,
 0.8749687629253002,0.1796721251149147,0.8770088033687531,0.2107157728042175,
 0.9652456322091177,0.18102382403057837,0.17165312158200818,
 0.4726394727168586,0.4091896506372291,0.8436079811741483,0.8708105469240291,
 0.5790115994187456,0.5541971232965536,0.015645400159079115,
 0.6437273352716992,0.7733779678636143,0.9957741900572249,0.0328750928517999,
 0.16171394659992877,0.22422974396608242,0.07374830883428174,
 0.7911661620772816,0.9065946473356074,0.060293529981373495,
 0.4240468092378026,0.8594084857604871,0.4432776576196591,
 0.027540693087317658,0.24681513295101976,0.7714730669032092,
 0.5971612052866065,0.05179362666215759,0.2912328363262483,
 0.14564467935961467,0.9314448134512555,0.19436313186207732,
 0.23176805328888517,0.4781815660132832,0.3901385523056863,0.6840524528892673,
 0.9761351101724852,0.5940625813529921,0.7267812176001984,0.4226439162418385,
 0.5438734401720792,0.4918104691160251,0.7975820679724599,0.6936675517690145,
 0.25713040557535527,0.5951741916409431,0.9631232919713635,
 0.06745697272515339,0.936515908833865,0.6006849917076342,0.5026695790320337,
 0.34630967367621324,0.24534856248193015,0.44076071687525087,
 0.5724047664909745,0.006206660307469969,0.1887729904499844,
 0.4501401884533296,0.1231444140426794,0.261547467931763,0.4531260641228039,
 0.4743774082573995,0.04918440366384347,0.9989857873995792,
 0.10945619035263887,0.03899824402023233,0.8938680591686347,
 0.8520617157900735,0.7252485423089012,0.7359494921311457,0.4694084208466326,
 0.5575693917134747,0.9612698063969785,0.09583438672461564,0.9233004046361,
 0.5351929417911612,0.5560552580990176,0.730931650641562,0.5119144520106168,
 0.39054931403210325,0.4685765126240953,0.29545368068014366,
 0.8859881613110819,0.8236789133065219,0.8865543759126292,0.8797589749476378,
 0.15841441718430116,0.352066393579748,0.9894155267330793,0.02193258600832193,
 0.4258498180852861,0.1414440918805464,0.43542746777647756,0.7038501050787104,
 0.6950004258344065,0.6867158586665325,0.32671803728431636,0.6681551082947937,
 0.09839048870494682,0.4835013657764067,0.27264953393730385,0.827640501240449,
 0.5438084554551019,0.9790229249516121,0.8614311849279794,0.8058518599853373,
 0.8431720763052234,0.7403086730364513,0.6137004731860924,0.8570712611190109,
 0.4514935666621106,0.2929070439597514,0.33962842665334814,0.2637023500856025,
 0.7123065561533031,0.1165950781313092,0.642579223372763,0.3239770600721238,
 0.5129141585473678,0.9578799097193272,0.9717999276013355,0.3356224659823881,
 0.52920765088943,0.21820205096952783,0.11046024453216619,0.3054350035107941,
 0.6562127585240473,0.5354763217585344,0.4167750634327271,0.47625352030913737,
 0.7150735552174523,0.06670907569499751,0.7119518749763488,0.5436754716722239,
 0.13163456495901604,0.9303290231641406,0.37820435164610333,
 0.4379022413042025,0.43790362642705694,0.19084489372019486,0.906839428412534,
 0.936439801527602,0.9359530053553294,0.448571836043282,0.9084826873238657,
 0.27520013646362806,0.0468967207195925,0.5108009131570086,0.7192244061605262,
 0.032096826800738,0.2131131048118281,0.4937363387853435,0.1523961618432642,
 0.8281881205249463,0.19057747014029358,0.13128916676883862,
 0.8263936855445833,0.1404717954420699,0.22320877497008296,0.9012217847891075,
 0.18908091789258552,0.9209494387738184,0.242986154766349,0.26271157828907643,
 0.6205652942375842,0.47717312011919377,0.6525536054505159,
 0.11879446936160598,0.18883303309467858,0.5642839799367927,
 0.44920737314206305,0.589961611082731,0.5574846701175211,0.5656378877369843,
 0.9007127914585804,0.49894205986204687,0.5743628226346973,0.4416930763875351,
 0.9594732161805948,0.29330578437571186,0.5746189999986868,0.2613947333210642,
 0.1475117133454611,0.32182647211073623,0.6609507296994256,0.7115236517878838,
 0.32707763285642233,0.7297338120868905,0.043526633898599454,
 0.3319971455625871,0.5529835940872418,0.7709999743841169,0.530349954313276,
 0.6922266702507045,0.17900108439262408,0.5343499714183723,0.5078040028189357,
 0.9377835446175931,0.7593906055431701,0.867831483745318,0.17429678951772642,
 0.9094591165126961,0.5674212557033065,0.7915153866414577,0.7317618622494996,
 0.9802081883952515,0.2340902793345827,0.6087287395317432,0.5178124914737646,
 0.22045732282011243,0.6673028933162736,0.6554967055183135,
 0.44670170875687154,0.6429138270333943,0.7724234483844403,0.8584463081183612,
 0.8816593524410755,0.6473940090793505,0.6103282766762841,0.13800822890646947,
 0.7778566891390426,0.18755166499427156,0.9462687590474098,0.2910355272883425,
 0.8095320239186782,0.7820019772654808,0.8210577184040752,0.7636459590761782,
 0.9040414162513157,0.05019038885717997,0.11297121189223269,
 0.9809656068885925,0.02714574307994877,0.5087741466126552,0.9974711236137089,
 0.5129492112544132,0.8938681702603103,0.6263047493032867,0.339777423044606,
 0.8569416272853637,0.5770109415136846,0.8600818571005618,0.40201992945777143,
 0.3878111767810434,0.796787387596557,0.6349105598462222,0.2179866927918056,
 0.7956991421425068,0.6149926537916012,0.02794222432222182,0.9975486322003343,
 0.15492893898436122,0.4066406891435842,0.05759508832583837,
 0.17821172903666316,0.03242821458847234,0.6950451188758393,
 0.5054854353936419,0.5085206304136114,0.5733791376352206,0.5246035852746982,
 0.5495631725047868,0.170865438449761,0.21773411695566802,0.3200075361278524,
 0.9478448836990834,0.7004385514548267,0.2857442185150081,0.12168439375364526,
 0.9790260907201145,0.14229504094168155,0.3975836242977395,0.1183649492954002,
 0.34097304096087444,0.7879419109595156,0.5026014511857886,0.2689877523971963,
 0.43144174952163616,0.06404020915738928,0.3169560931570996,
 0.6422275624250033,0.007941258955426855,0.20382125660167139,
 0.454732204982919,0.35605619502799213,0.05906265849002945,0.8538043978602907,
 0.07130596472594986,0.30948207411294215,0.9701035042300991,
 0.24928819217509357,0.2694204959111186,0.3670990084431722,
 0.39182605875622956,0.6037870775220258,0.8770006784401262,0.0796267126542547,
 0.8880233987282787,0.3061128426819416,0.6690349988407918,0.20966050912986722,
 0.29970793651712524,0.7770886677240391,0.3062206650422583,0.7739183942917794,
 0.006099031924955756,0.14054084493078245,0.4323465186897696,
 0.34098386519060486,0.008778039773817636,0.6867861562686571,
 0.7513801783873173,0.13591075931611185,0.11287599171419849,
 0.9492585153206686,0.899047486912641,0.252061323432928,0.24447264253797996,
 0.25531414947968656,0.05615602129521047,0.49327328957806493,
 0.4873761697194916,0.36826923769972,0.2561705808475606,0.6887239489973613,
 0.08623200796092234,0.14180400698023032,0.4946379576162441,
 0.7403503494003001,0.3161334768502645,0.9833604336991211,0.4630570508229882,
 0.6205465908060619,0.018875378468075632,0.5855732252490001,
 0.16221200580463757,0.7786520029361892,0.40458877681473493,
 0.4694255204459772,0.8921718804546073,0.09783165185251441,0.7850870744984181,
 0.526329048499085,0.09359784448688935,0.6318164367504882,0.8399244148045609,
 0.46741889032177053,0.10317761019851024,0.8995781624118309,
 0.5529572978509129,0.5055174148795072,0.6647334282799158,0.6504308768509385,
 0.1985900649959691,0.21606124997767384,0.2589965897217301,0.5604026020234654,
 0.1028475801785258,0.4639746608881168,0.478081389551957,0.5939823988554883,
 0.017299819025389684,0.3724648406150358,0.5256015423412994,
 0.2057554882444006,0.42562082727888506,0.272746236623159,0.8188378448643826,
 0.6842646390645665,0.9806445257628468,0.11763135200748642,
 0.14828021471191744,0.712199366011374,0.22450887853720014,0.6969923294286194,
 0.862880615256749,0.07166199140468799,0.30154364306495296,0.9358229653435761,
 0.8243696933900394,0.604730492097864,0.3147866919981658,0.7846277937319849,
 0.9272356717546957,0.03160154263397008,0.029088729258176294,
 0.4076842269206168,0.9175836790235632,0.01291841712513253,0.2703914342415671,
 0.6862947582731207,0.6445334534387976,0.18007993384317666,
 0.34244329553692465,0.8284456364899123,0.3717402979367054,0.1175675784664112,
 0.12646201976035654,0.07510868603439191,0.5382948477608313,
 0.06347465063163626,0.7256447173676589,0.16415098348321133,
 0.15952496595939447,0.02503186144480818,0.1631751028456775,
 0.1448971468823591,0.33557364795239375,0.2189375397683886,0.7700698822834529,
 0.7986700104900762,0.2738195368079971,0.10747218537272452,
 0.11485448748926386,0.32373598500958845,0.2988433816171596,
 0.6653148807217233,0.09233673136355658,0.6763024176138697,0.6853984570444894,
 0.532378048445429,0.49412454406437545,0.3745084981338632,0.3612408203111551,
 0.1247341569606013,0.13702969790127195,0.5723384799040827,
 0.16819958077891195,0.39417916117861984,0.879035730149072,0.7450705677694747,
 0.10428627804519841,0.5179841045898839,0.30391311120990017,
 0.8526688481654605,0.4973360554088465,0.8767244722278892,0.6135877025109625,
 0.002429231541312138,0.6735216650061007,0.4374802644012029,
 0.8058796822674463,0.23881191967884874,0.3674558874178835,0.2618605394488558,
 0.5185481578542646,0.5351471973013187,0.5376403122192678,0.8312613409994467,
 0.6219524591428711,0.48176497450248956,0.06484761342142753,
 0.36804328859480984,0.35347823459874905,0.39700578154620225,
 0.6793444090532692,0.27853944375779016,0.48798324122964065,
 0.16631573978061465,0.02721641213233006,0.05777452033751207,
 0.9632914685228147,0.8327992900710883,0.7482637355804493,0.7900441020309095,
 0.7499554243804765,0.3591758034127116,0.022261133473775674,
 0.4972900759536446,0.701280232473612,0.8399133022614745,0.20600299514424147,
 0.1339420893670863,0.46183523512232216,0.6840489262810278,0.5152966985243429,
 0.09503983619496104,0.498483828530101,0.11313658057331177,0.4316772680102926,
 0.6649951029204909,0.9789971526619266,0.5122862362118373,0.49431263276991944,
 0.37212369806556567,0.2417768912031818,0.8051936713259014,0.8414101422970587,
 0.6377240971780973,0.5706236737534571,0.1302758682421029,0.4129651441963348,
 0.719347630084401,0.9874599741699871,0.16393667667722567,0.40693838007904926,
 0.765398270023701,0.9922643356289007,0.9069563966822926,0.9495166301768223,
 0.019538938973767017,0.9993307579502808,0.3632882093125773,
 0.29029162192461455,0.2337677104499074,0.44485465660580736,
 0.2999732453696745,0.5696110552035132,0.5486699828663426,
 0.004314219513997353,0.1870098702150511,0.82183101475816,0.9096472549305863,
 0.7041762093019287,0.8274983454409014,0.22639379038869656,0.8526528573362657,
 0.7146381087930046,0.12132898053230123,0.6643546359259136,0.7991203730893144,
 0.24070834991388257,0.7615355023904641,0.14109893200756496,
 0.22026961778464105,0.6394932345401767,0.9328162247087269,0.4587908768916481,
 0.08191927497533691,0.37640186212706883,0.1872721936268893,
 0.20486174105084287,0.6686622914254096,0.6932749647672286,
 0.15668938098109897,0.7858627246195331,0.2293910554265517,0.6271611387495889,
 0.9616127263924923,0.8159619929447927,0.027141829199062717,
 0.4170637383958473,0.1398820804983314,0.6153982006996225,0.40992474815527635,
 0.9715527449116516,0.5545618183754175,0.04794067202325092,0.7603506593091262,
 0.2617774280363815,0.6180225197019713,0.5043324441472677,0.27674285908685836,
 0.6289288946218026,0.5116598212830809,0.8059690374133585,0.4222372420167808,
 0.6419680843783533,0.7246813172153749,0.3731930520535144,0.33410307370556336,
 0.09543267888883655,0.017657829154323124,0.9585268966615925,
 0.09502403196000653,0.5359158754258083,0.7483159480351065,0.5255131844398171,
 0.29705206244701854,0.9649259925123884,0.06859230616708589,
 0.3461165865721224,0.5137429738402962,0.8582598126699583,0.8138226965407868,
 0.1893892213677788,0.3592591767604111,0.8730759789650641,0.28681752464145227,
 0.7479870284657979,0.5169643216424686,0.22845218912180354,
 0.07960364614369275,0.6771686323863818,0.1901249319547863,0.7205491680497085,
 0.9222105309387661,0.5821031841742487,0.8276654783309242,0.20514116134930271,
 0.28205069943014616,0.7858394740351474,0.4443654177548746,0.7855723153241846,
 0.9904535774210412,0.4632972269129183,0.4237870894118574,0.34533589192533964,
 0.31127210205523703,0.7597122885455053,0.8602323048109728,0.4421301882609111,
 0.7116967927005867,0.5624308646209483,0.3869750581442468,0.1854744993256746,
 0.7322474569862853,0.40682630763065997,0.3727630466777191,
 0.03837587250566454,0.6800312064860902,0.5932305366591366,
 0.24761714435415583,0.940544397296254,0.9657707297925847,0.3694431066339299,
 0.9015028388691552,0.615014765850723,0.6198359790214898,0.028749577203638355,
 0.4360121284903945,0.761199630908351,0.4292506183947976,0.8942640860273166,
 0.3338338886571016,0.8014481086603872,0.23707397648277373,0.5431673092794804,
 0.3324535865274978,0.6573249247146147,0.0322089563421426,0.5403677038132169,
 0.5029232040945505,0.48400784677782294,0.5057562890685487,0.5389888981951086,
 0.15976048383328345,0.33072631362991567,0.007178915682546139,
 0.5773630804955018,0.904876973457439,0.1195872244945908,0.5423161770935336,
 0.09429824038447643,0.23820710299693182,0.10419060107452971,
 0.2391653305624588,0.25200931495502865,0.013110493252587063,
 0.2662394754046412,0.694886952725232,0.4812955434927808,0.896795793730734,
 0.4071043430553627,0.03373148856989627,0.3402359120981513,0.7438315186006575,
 0.696788823451947,0.6112074483320077,0.9173998665682528,0.42640652720867045,
 0.804680023845285,0.21752907634537222,0.06261120731507341,0.1616131363681903,
 0.609389112183548,0.6637327664786514,0.9301382783059906,0.6003061350248743,
 0.9763941196774557,0.9008516416037596,0.588158810311184,0.16011909051663775,
 0.9503439379911357,0.37355137286606266,0.6342538889191751,
 0.04060703507445684,0.5821772511122854,0.6373003168966888,0.8716020126274735,
 0.7758648983274288,0.9306933386598183,0.4729700567541857,0.2089138677186736,
 0.643288953622232,0.5573686789532262,0.4233301288519199,0.8736360203323417,
 0.6593343803555735,0.8036741438583029,0.7096183786295561,0.7174368905462561,
 0.19132951332710713,0.08174811829769457,0.034984831749490075,
 0.8918720822784242,0.8456260073644294,0.6350255620014653,0.6579194202314909,
 0.43867586924386626,0.34169552959135485,0.13873841814703936,
 0.08564760285959117,0.9601604471454618,0.5661816407206657,
 0.20327407561953847,0.04456296564944506,0.5864130835637049,0.58422017994769,
 0.08396102620744594,0.83808849384139,0.42738073066002447,0.6123230702555855,
 0.21744550502952187,0.04638599864817494,0.6845378624681357,
 0.9363939499178949,0.530236536441181,0.3073461653855286,0.5971589164218676,
 0.4737459492419285,0.11680681095395784,0.5861518220881973,0.4250096673319891,
 0.44077012761586787,0.7593240083444297,0.4992038217341618,0.7115993913320762,
 0.6287283170851831,0.8205970133494973,0.7285217912776616,0.7068394497709031,
 0.6394355315627768,0.5723767199285162,0.4313351838181234,0.7596043376715407,
 0.10204559055485762,0.9933932333174363,0.5996809752203287,0.4726695885003559,
 0.3895347417661982,0.9663050665866574,0.8638896595374631,0.9480594423560064,
 0.7221145156094829,0.08847213675490495,0.3554674482843124,
 0.43889483590878897,0.43671797473479446,0.1181807092767374,
 0.7218813900490434,0.060257622696542446,0.9399881573557247,
 0.33693816970758506,0.7655481938778944,0.15810597089824807,
 0.20816371007781298,0.5839255396300825,0.061947798980472646,
 0.36931909258124396,0.615894697144122,0.13725783408445968,
 0.29675275586478844,0.6171864918499439,0.9930229342118104,0.9243961520049749,
 0.5941245600407364,0.8180271298019759,0.0019254166900903869,
 0.4157112670745381,0.4864933252836451,0.8780364072641013,0.14319170746510979,
 0.856947050968185,0.5390716126429865,0.9281780306765504,0.9378619921998719,
 0.1794462746600447,0.8784836471018732,0.5508916641411605,0.6717804647653413,
 0.8486238282135751,0.01613099202901913,0.023934455449641012,
 0.31993338777827374,0.08180085138482596,0.45873541762080683,
 0.1528293825811884,0.5116890901520988,0.4369976387101374,0.7875039534552959,
 0.9068937689867241,0.8784862096722716,0.9132897045490296,0.17588462670352356,
 0.6129846923360162,0.9770903928512067,0.9474736229208528,0.31795438172322044,
 0.5070866493659301,0.3185974795033363,0.6995465244827531,0.6972220015654507,
 0.977629091934344,0.8374573648332718,0.8060473932231873,0.28736210340857293,
 0.8008302965822739,0.990630128868218,0.556975812842975,0.7259810402017308,
 0.5822709550066169,0.4399311095342091,0.6298907878953668,0.18996216827225276,
 0.4377107837006018,0.31678377179989414,0.05028749373328345,
 0.04557255515810987,0.14181344078045632,0.09681393565843244,
 0.26422923080385274,0.27373943511126897,0.8129325256654499,
 0.12740089740588634,0.6738554645097314,0.061896223401649686,
 0.2639123306288693,0.875374983986867,0.5053880359539501,0.840382596296341,
 0.32199040487962016,0.0627696466628751,0.2504325561188472,0.356721453460385,
 0.7389182076168275,0.017589057575599476,0.9939977344488004,
 0.5714188147981678,0.5610858602470797,0.11823643984379817,
 0.02354807225176636,0.00883377873136193,0.7061498399585651,
 0.13589919640758508,0.16343892550064354,0.2539833314054786,
 0.9514957702676831,0.031896480435417485,0.9168646258051275,
 0.24804977449858523,0.8246433328749851,0.3650125104503592,
 0.13078184376155033,0.04944178458249948,0.3016334226965046,
 0.9950208315838536,0.9240555768666604,0.37342397612178835,0.2054551090853607,
 0.9185265839980904,0.6067154763421289,0.6564649035099636,0.0252957290924396,
 0.50858870812416,0.7076654300878019,0.7830922374695474,0.33302278527069396,
 0.4064726407481477,0.8396516259349109,0.12896506388171602,
 0.13974700087545744,0.4641208476025026,0.9258503163773224,
 0.11787884455748388,0.5913850746924733,0.3020777584883514,0.4918395116961566,
 0.48788462237856756,0.5448711684482274,0.38153470641667975,
 0.007423513172659213,0.6849992045900644,0.006472620131997231,
 0.1176023352883766,0.06608766221396534,0.31338005645057565,
 0.08470457816575072,0.8759493541051767,0.6810933077914455,0.3944320404872562,
 0.6914226084403235,0.2819331551903552,0.9864242627209571,0.7780816406816623,
 0.3018076194326371,0.6273572427320919,0.34823840339628365,0.9969815493929229,
 0.8723902888369173,0.16377881190500554,0.8092708644929041,
 0.11478025856205998,0.3315356606972626,0.7762979025608824,
 0.14297127889701577,0.5421043669448983,0.03550904040036096,
 0.09510340109679649,0.4841214335430146,0.4553135478674718,
 0.36065645159231163,0.1485557341081144,0.3384183924584142,
 0.10704889544546936,0.5902913721823462,0.2860871809025263,
 0.42402077857192977,0.671791199191043,0.7311483981048947,0.07389829045001206,
 0.4166295222428249,0.9449916358606724,0.7373949924959202,0.8062313897915337,
 0.430324075661063,0.20737461212344088,0.15334833613058085,0.9446884219457897,
 0.6798365596326036,0.4744334343620994,0.9461674540622853,0.3258413582940417,
 0.2774079761942715,0.4688086816062056,0.40528168696945566,0.7188227946282025,
 0.41258319299125945,2.395540956932063e-4,0.8682546788264519,
 0.30031481015388095,0.5624633081766258,0.08064657317496837,0.442338510673402,
 0.7201267418879284,0.17507438000590958,0.46901278565834736,0.859194788420681,
 0.5676196030964955,0.8869132217120613,0.8974942775891527,0.1572870041329556,
 0.25330647823533015,0.7918330020397748,0.27337113944989666,
 0.5405286563538404,0.2830102692620683,0.4249074248366287,0.32641634975958034,
 0.8936731825441291,0.0851120961673798,0.02529185938685452,0.2131573940707715,
 0.6908785321934566,0.9239718302213175,0.8514906149607728,0.36262027882846326,
 0.26506637680526746,0.4930827261841043,0.03300930208449604,
 0.6803984061086839,0.9214503932348693,0.5014525942269138,0.7548708987136693,
 0.14953304011173607,0.7579787194718203,0.6021392081636261,0.9276412691769412,
 0.3652912773724051,0.4550279053291617,0.913421148274245,0.6188862072917004,
 0.08524595592311379,0.27116026222722256,0.4139786775609825,
 0.8592216670564677,0.2652446853277488,0.3418292975852746,0.04344685429141426,
 0.22319387765941567,0.6911864512818138,0.70970972357857,0.6298578262261068,
 0.6982226934899225,0.3565512608032655,0.4696937819040127,0.6170624739711728,
 0.7087953777171563,0.5443243969357493,0.45754591841576564,0.7042771912921133,
 0.23849720686006082,0.45312839698359975,0.5584273954531127,
 0.8611439175936386,0.49715293477973876,0.08812207316938936,
 0.9564174423614998,0.49107771734407224,0.7788368065097033,
 0.47995988355189767,0.8432318539735809,0.9803458938506726,
 0.24806773266744198,0.9003004676806448,0.9000098906057958,
 0.06501933093881573,0.7271739425500271,0.5361696987113652,
 0.15836810027599835,0.35787111400244487,0.5888859049213075,
 0.10232755801588289,0.08192925299030884,0.9497910963760724,
 0.5149431228787916,0.39569381721596086,0.09434935960435487,
 0.5742035036364276,0.2775857999866833,0.23535102019021958,
 0.10833089408316232,0.14744333025784417,0.3917525439058952,0.820044714718561,
 0.7743843302221368,0.8457882022700367,0.09541411112362796,
 0.25944530144389544,0.48573703970044246,0.6487461274561098,
 0.6937584454843116,0.6123929358453694,0.6905310078398808,0.13190442135421787,
 0.8703538816098264,0.36392311876829964,0.3543271598304373,0.2021277986835187,
 0.053613530359370776,0.6357615917653088,0.8822187631904325,0.795305216167389,
 0.8395054382902685,0.4374219987383301,0.2739724385407155,0.2641168324989558,
 0.6146493717744376,0.6355977953524341,0.8449099389483344,0.4421061331457865,
 0.04046844264341154,0.0030803860069369415,0.8219481996026139,
 0.62156747593752,0.48526078517769866,0.5332993802589401,0.4407983344921059,
 0.03648192966032571,0.5318216456322363,0.45028211412169517,
 0.33355941796115074,0.32888439193612384,0.18531700450858024,
 0.45998061881365815,0.4584918022130573,0.7199589969695692,0.8018077897491024,
 0.5329605130935551,0.6206863288256946,0.09369272542673501,0.5071564377405753,
 0.8349737208004097,0.06779352892155732,0.8637318263812543,0.9138209194186698,
 0.9537472987862025,0.6309154266920534,0.18489077879356441,
 0.44471219599685063,0.631483498847472,0.12541468737069972,0.8415286083034771,
 0.7162392984501835,0.1745266131371026,0.14511367148732623,0.2868872535875777,
 0.7684614626254576,0.013155298653143244,0.668322829216214,
 0.25326161704289696,0.621931632585044,0.5787616028982508,0.8040164321908446,
 0.6822606612930994,0.9760365427203098,0.31221942816000314,0.8282958852235477,
 0.3303522654809641,0.5542155175125448,0.9072699059380783,0.0661189396883739,
 0.8816374578856792,0.663453260769663,0.07756143615838262,0.5561419648253623,
 6.723289964094548e-4,0.4499408101310922,0.1834364229430816,
 0.10847297669989031,0.2803670584060527,0.6726408308401796,0.7138854363628737,
 0.16435661479722086,0.2888010650897742,0.9534314751896045,
 0.23399592692721027,0.5006398854071368,0.5839351986211765,0.6436887586649094,
 0.3020753402556694,0.9262354262657527,0.08061654293946452,
 0.011359259257534449,0.18259383498662796,0.8227431906610888,
 0.8547649787749785,0.6547661020870786,0.3115211628773664,0.6921497736576242,
 0.6051710405800697,0.17660911378304922,0.5632258456498568,0.0490327479240984,
 0.6272553800059183,0.7949157210441444,0.6706631512500205,0.5352430183107565,
 0.7807094233663362,0.6955440618114477,0.718753561127558,0.0615498775317479,
 0.2802731534979539,0.6492772241038752,0.8272460853246166,0.13375434909865547,
 0.8930446661858089,0.696222032241546,0.03908508149275469,0.6786502353519391,
 0.760861896237803,0.6007085550373054,0.6900704462604325,0.8140845963211703,
 0.4931189750020841,0.5942888847621144,0.3267527229760353,0.33210719177570214,
 0.8885772311220175,0.7309032516544951,0.7566873027847131,0.6345734805323812,
 0.6193566265645531,0.6287722683914444,0.002150723696613843,
 0.1236310342739474,0.5782758345635073,0.3513290130737171,0.3737692364039731,
 0.5778642184312309,0.7107115359244933,0.8620893192289761,
 0.011365783249903894,0.8794761265027116,0.5040818714633973,0.831989363327638,
 0.7954786448862744,0.9089257314557673,0.6725979959914502,0.02889299168567594,
 0.7821636781903832,0.3714272751313308,0.86480511777867,0.6186473712697352,
 0.6387360327705652,0.13775997894966063,0.37461950808648914,0.529684894016928,
 0.016866292649031944,0.04307331499919331,0.9565531772871783,
 0.5643778536325822,0.9246961936889457,0.39864803670469684,
 0.29304792277198777,0.7666637989230575,0.5471277501372376,0.868528306574859,
 0.63424794320867,0.5162843484156493,0.2641423007712064,0.710127599200318,
 0.149645449081242,0.7599546216830222,0.7601577375508864,0.4967074302055695,
 0.9554748145690585,0.6095563651234628,0.6243215447855406,0.9548958832866042,
 0.7647943701631101,0.6504871575866018,0.213495844651033,0.9797746048306942,
 0.48681057895744984,0.5187268848296521,0.03894454827022675,0.564614919412926,
 0.3256647404833317,0.2214097497029024,0.2377455317557826,0.21133650385170044,
 0.506070824153413,0.10521701995251476,0.6714859304265518,0.9218788792494989,
 0.4271608486481485,0.6005284647223184,0.802851399134641,0.05392314647617247,
 0.663934210005747,0.5856737035044534,0.8813803796209904,0.8456745804112648,
 0.688067616587511,0.6978172485723813,0.9507215173132235,0.3541054670582535,
 0.1864988068712975,0.826740638428542,0.004546517880934076,0.3313822786010209,
 0.5649904949079325,0.810550246178387,0.5533844540022723,0.1638736416939408,
 0.528662476283761,0.7766015547598343,0.3909342551601319,0.10410429096329255,
 0.18214247667042205,0.5377296472365753,0.794219750295837,0.773135387311346,
 0.3840699714169796,0.7001836932467592,0.9007544644843851,0.21471743335323434,
 0.8545018116888228,0.7680597240910171,0.5978603443493871,0.4188692567493272,
 0.729534358985414,0.767928980604625,0.2471534851165551,0.901433706962904,
 0.2683730749878328,0.9582516833292134,0.9583219537876243,0.9228673323083518,
 0.9363114808091151,0.7828823590691132,0.8224264765490364,
 0.009991724677848701,0.5769778887858688,0.7905781668342884,
 0.9704936919658326,0.7614122217713248,0.7430147606511021,0.15099036723432113,
 0.5229028053474882,0.12893980574022523,0.9527247413219776,0.7581122084410467,
 0.506989239441369,0.3392597857201164,0.8408286689679562,0.9378176646856766,
 0.15959786863608738,0.681443713128475,0.16199021738541974,0.9161028206654451,
 0.27678512791210574,0.028873912538432744,0.27256453465636066,
 0.6993112491240662,0.4831383290628357,0.5697156231157099,0.27504901243892654,
 0.9205121705852675,0.8029704280950347,0.6796526251871449,0.8979970407764821,
 0.5127864019851697,0.42381021802210217,0.5066361320341429,
 0.04089066269353503,0.7099552333150153,0.802166741603068,0.6805975377615359,
 0.3961129563609984,0.5800162248873102,0.05097616088702961,0.9264402469644535,
 0.2601845234660852,0.918557862439796,0.4097100829393868,0.71992105866298,
 0.27743959246098093,0.2218113712165044,0.8768629684758881,0.9099170209371702,
 0.6738602553288112,0.46373738748821935,0.5036676375601087,0.3174773856432781,
 0.37269842439773315,0.902166922845854,0.5682576996948038,0.3468606898116805,
 0.10044253177405982,0.7225042735534899,0.5636156724945454,0.7190397488763907,
 0.4465585249713223,0.4726522624917159,0.1308187146968891,0.6347117324664598,
 0.8026020783319151,0.0866137980759637,0.642385211664428,0.5049012010541528,
 0.12814363881623048,0.6234180003114947,0.262615164946038,0.9588349114614344,
 0.3424528402336061,0.4476567925506726,0.6832918349694359,0.607191613686923,
 0.43189969724567034,0.10109999162097671,0.9452881563780282,
 0.8610204699045588,0.976339221992543,0.9896928006005449,0.31357064257313894,
 0.83223915076892,0.17131612094975945,0.793884527636868,0.4415474366801627,
 0.07016538481009627,0.515987489954389,0.10922279511505384,0.5580625828354143,
 0.653252549852688,0.7027742528278998,0.26446268143343876,0.815933684638623,
 0.3097645170336034,0.6652801150059442,0.5692145922696716,0.2712681888287908,
 0.7368661463476261,0.8108981450052907,0.6831409258287116,0.9048665446532478,
 0.1842281235413491,0.854469039131991,0.5177572141251521,0.8525100354278552,
 0.11246711538455822,0.6185705072514764,0.47482377989479485,
 0.20199065457073706,0.3462444171260427,0.8900320755572719,0.6615800468727959,
 0.9773509670332595,0.975019666494791,0.41352719156315365,0.7030905750090066,
 0.7905687280875103,0.0709929636871457,0.6233727925552754,0.3679660047705019,
 0.6430784466430711,0.8387743233913709,0.35561598039706444,
 0.09976346891240995,0.5788521039723058,0.24274432071919572,
 0.22354163977805241,0.26798048716568135,0.393828397800123,0.7242445197656452,
 0.5036720634796863,0.28438769454814583,0.5814148286792531,
 0.37294286545793875,0.9303579054527166,0.541604926816728,0.16357733943781005,
 0.17013110917460983,0.7854165210465897,0.9791214745550612,
 0.45060023497925694,0.48867718451551456,0.15116332597803916,
 0.8374567901750709,0.8904049602872011,0.48181716359763094,0.6853074988325482,
 0.19681271032605252,0.5505472368271774,0.7782633369402367,0.5696696871846754,
 0.4134010540430114,0.3614559426582956,0.7984441086819287,0.9366074007455637,
 0.9880022738515031,0.13945666266867163,0.004415300806236311,
 0.9404010579365828,0.01790160122211637,0.06259959269074533,0.946348408898523,
 0.9294362709631665,0.8573601076037993,0.43950737781896965,0.5228517074218806,
 0.3953356329605423,0.7094579616237355,0.3025003124465818,0.5134757864435079,
 0.9451083001558729,0.8559985703004972,0.6199551885433431,
 0.056058518462603346,0.810610501657719,0.058621488063669625,
 0.7975884549756939,0.38958963098420385,0.004390466712856256,
 0.7631497518082317,0.20839414633735143,0.09444841591572684,
 0.057638467084188294,0.18925742409843882,0.4946243266057849,
 0.5295353081859322,0.26357024686261743,0.8368525958220727,0.877862554268267,
 0.7478766294566329,0.19380738257474361,0.24871147755982403,
 0.5669744883330743,0.8050891486628629,0.9014803450548508,0.357694929130133,
 0.334270620629769,0.12504167557758417,0.41914165925425684,0.458362868626235,
 0.4250321436695912,0.5614706486411869,0.08341872283750273,
 0.22749886688022658,0.9513572587272884,0.28841165646828826,
 0.6842392440650151,0.43291515399066893,0.416447008492825,0.9467941926644153,
 0.6792249013818112,0.5497772741119188,0.458508529303739,0.5031047144773947,
 0.4471298148616224,0.6854470415675,0.5710568717050656,0.7916228521032755,
 0.47285357197649946,0.4725453496977975,0.2422103426954989,
 0.39502630328739274,0.2244007308087702,0.7282623336983121,0.3817759633360457,
 0.9299867195916115,0.8314470772347711,0.9139541173005612,0.8283239217426781,
 0.12039720740061555,0.23672644082383687,0.6070775917708543,
 0.23194706953300914,0.3173560152446032,0.09941481679426067,
 0.2890100304157531,0.3915673355667667,0.1173516275840969,0.8269931232012455,
 0.03223280410141505,0.8891120798093856,0.8478060884490665,
 0.17946045747888595,0.023423858064157588,0.20748615145300053,
 0.8520213929566598,0.6076574177218566,0.15678645538137503,0.668189384144624,
 0.2717517158870262,0.17320283253974111,0.3361829051286259,0.9459847633590099,
 0.22581694997701696,0.23183370964921968,0.32036205581545596,
 0.8634716682792974,0.03541975376103712,0.32486700810183433,
 0.9747183931299054,0.8898893834410091,0.18183210864533006,0.7733711932341377,
 0.18379942571623864,0.042463428881817644,0.6526277751894792,
 0.5725069230679372,0.23379240059531858,0.7794048819485562,
 0.33931168286872526,0.33369093493733526,0.2660669962551837,
 0.7797379476959702,0.8560386663362018,0.6850005328614731,0.9665013470772273,
 0.33655903036253176,0.7622447339282193,0.5002742040160884,
 0.053666519011999925,0.7495682569841201,0.8593538420468727,
 0.42858654211227865,0.7307301550952469,0.7165761764452014,0.5595633187856808,
 0.25841008099171037,0.036259485240292544,0.11865235098939486,
 0.46012823706159267,0.6886853510675556,0.983457337360624,0.9233683587981536,
 0.13256510961961887,0.0354611405134988,0.8123338745533712,0.7856610497652425,
 0.6549307485074701,0.45459933115958495,0.5377382916110206,
 0.20313950608897235,0.32661700733511423,0.00427427665322222,
 0.15444865440666722,0.09096805914920181,0.8281471947215536,
 0.47325599265587937,0.1254191359663639,0.5201616861891619,0.5398132752865198,
 0.11289379688345758,0.06786791950622628,0.5961296349546381,
 0.9784663241200668,0.45972918001561114,0.19790071417023702,
 0.45774117563745587,0.8233942996892311,0.7554733252780809,
 0.24972993570598956,0.2376284585744548,0.6698567990925152,0.4442651214509389,
 0.05681057492398711,0.43431725467703897,0.6936528926221693,
 0.7124720295709193,0.023049910972708476,0.5707697655845331,
 0.14662570522285534,0.7356804193351373,0.8165609199286117,0.9632856776261559,
 0.06327325888394464,0.661706907818751,0.26637091653444034,0.7540545512448735,
 0.651829853335054,0.7623019300329014,0.9965747800602394,0.8196405785884043,
 0.5755630033106259,0.023849067284958725,0.921334387194693,
 0.13347596769114478,0.005947281130308624,0.03504618251691394,
 0.05628516095493874,0.7963519659157821,0.7838067702562526,
 0.35798274943173936,0.5110881294142116,0.7592446382480766,0.5909423042602382,
 0.9578523099911129,0.18486941480311037,0.11961244735564658,
 0.8904143770582755,0.4666541145295703,0.374345983595325,0.3482099365411724,
 0.02796167383581194,0.6633702598692137,0.33426891383755597,
 0.0949426022481159,0.5384896459532751,0.5765778614989416,0.1742597076156659,
 0.893349950068609,0.9917518164084619,0.34382863748291104,0.9667800696218096,
 0.4192832464004931,0.3610420912192921,0.8647515748446217,0.6313526137508451,
 0.06411857869345505,0.5063979112276288,0.4636363329972948,0.3896541094549012,
 0.9167851133641555,0.773303312518613,0.525424170210196,0.21656669864281686,
 0.759775105969335,0.8519101081736721,0.053354259434624574,
 0.15694068153533125,0.13447546530958143,0.45739209690533733,
 0.5471744591830708,0.9547946155400527,0.7463206461853896,0.28080274003209427,
 0.9806891314595279,0.07427044173999686,0.6529394356292987,
 0.37464396253205323,0.48263972815761935,0.5809320685518123,
 0.4863528415828243,0.303308993212966,0.10156367959098234,0.1161184190128639,
 0.9046479772841166,0.8523184158719554,0.8888785107521996,0.6738235185255232,
 0.5763781691488048,0.9426836963346108,0.2861079578415273,0.0574359307110337,
 0.8768118058574049,0.7685033932512119,0.2666716833602888,0.6533991800987937,
 0.8185328685423586,0.25565599357782887,0.5394396514240052,
 0.40381522961471905,0.5020072880891134,0.2591410849880663,0.7028564992510588,
 0.25164970670498954,0.23555417968325387,0.47911486249297297,
 0.7993310188969536,0.35133366280235423,0.12547077367312787,
 0.9518880656358357,0.1323399053212735,0.1717303193432932,0.8246285940709627,
 0.3247912784293441,0.5222109762324979,0.2620582857284326,0.7922737073101097,
 0.8502807653320527,0.563576717176645,0.6241251893277969,0.853962710965233,
 0.6342456224982093,0.7558613618736423,0.23118875743281908,
 0.41821261880078875,0.35771635675494506,0.2313448280464694,
 0.3151176642395167,0.005264107374497784,0.5634024437515216,
 0.21604699134370764,0.8194544895426522,0.19528443672186713,
 0.11608310253594922,0.6250119778283416,0.5681106311227497,
 0.26348555790814776,0.6600976260744144,0.42814195580606074,
 0.48118072120075095,0.011984725994016787,0.577536479088687,
 0.09408927049110849,0.06907958837660222,0.7913784965292312,
 0.4348002359223049,0.885831158172013,0.6429281676667884,0.21469050838585257,
 0.7679286834803997,0.8222353945149494,0.7336019356373675,0.9965791284996656,
 0.6362180759408366,0.7290509800818614,0.10737925665050674,0.8139954585368302,
 0.8233985364307478,0.9536931974488776,0.3098576664857924,0.26819295438028845,
 0.22816450578409442,0.6857800123599417,0.3305398230434108,0.7039188315503598,
 0.4684564950649104,0.7158335312530122,0.8962219426955422,0.31724101539942007,
 0.9370849970313828,0.035492599414153614,0.9863685277549658,
 0.9001917854849903,0.33721923943532195,0.5701621797470646,0.6263646617854055,
 0.2006300041220428,0.020955714719953833,0.49798648324801753,
 0.5454213640924561,0.17544348472064897,0.9986955912047715,
 0.060650120552380615,0.43623467787170445,0.7141167737281249,
 0.623802276250042,0.1612147048873377,0.6794816834364826,0.6611344828355394,
 0.663704307714996,0.3145760372581953,0.40293067943743,0.21234301106341624,
 0.4482689288788678,0.666147600405122,0.6405991361234404,0.19799315101973503,
 0.6332870182236127,0.4681288592821239,0.5600392756564344,0.5760265483150835,
 0.9465161951328458,0.7473741701924191,0.10780726858910106,0.2169269970202401,
 0.566346879065065,0.7477168164632043,0.10608762893807466,0.14508035228775862,
 0.9534117688301911,0.5557022796817088,0.6918522358510086,0.03427597756202527,
 0.611595410403068,0.4750285870525398,0.5154830914521213,0.21214179970634994,
 0.4611952342613991,0.2191184673056027,0.9144932886355522,0.8731878959311226,
 0.40555203633636916,0.9615175028507738,0.5816579147163957,
 0.027370477910478752,0.7565308712860469,0.8292557328764039,
 0.05990657020807899,0.9283275008835922,0.15924399336886275,
 0.8110102555747395,0.5956539518092795,0.8530864230047908,0.39478414459453015,
 0.4217654159794674,0.36144418689848257,0.5093771116341975,
 0.03719938738339357,0.07694519852496373,0.3460218002700024,
 0.16741756474041325,0.8547099540451119,0.42470780793162266,
 0.48808699166350955,0.6558384104771072,0.7640187742565578,0.756577733846365,
 0.8159325807558127,0.7065252278539543,0.6280077543038621,0.7589909688492829,
 0.7161827791247103,0.1820164904380246,0.6444300999352267,
 0.039128240261584124,0.4998228660055565,0.39077035519905856,
 0.3752436691289739,0.1412338035037769,0.7116014979845606,0.5285152861820466,
 0.7920964734081152,0.527714335705504,0.05543833516280683,0.29175826154439455,
 0.07944790719005468,0.8282735540547719,0.8824616287653112,0.8603791885800185,
 0.49582514673442735,0.43189444863935,0.38597014065020296,0.4715555146835392,
 0.6372780216260551,0.1757754907548491,0.6439593216466833,0.8397089274114413,
 0.29804384839723574,0.006054275853308977,0.413718495548112,0.813055482650149,
 0.45672386791683484,0.1817678375736348,0.6840172937819212,
 0.38982916000414514,0.9022304268932768,0.10802393861961368,
 0.7330589626095657,0.3573528390927556,0.8390611162188024,0.5456626962663818,
 0.3252291028539406,0.6850440468648687,0.934661099475901,0.45860439737961534,
 0.21050939999855312,0.6813797123128074,0.39743150446912623,
 0.1606635255128528,0.5911097162535863,0.27264767829629233,0.5126827410920708,
 0.7816256666921433,0.3447167658704986,0.14689588156260713,0.6928182403176181,
 0.8879347132084072,0.29843664026430605,0.8125301686699847,0.8834084073815415,
 0.44158548703416,0.08253907349050527,0.4938489318774215,0.810805135612547,
 0.2938038204931236,0.7531270256252818,0.8867288386404526,0.24441330316038246,
 0.28442200297998255,0.0266763809433177,0.36398741630867804,
 0.01754948024540215,0.6942223530994176,0.731916095837762,0.9648881342637421,
 0.986547617608762,0.26310130326989345,0.27948786528038017,
 0.41112604024886923,0.0558932511373148,0.6979980914301848,0.9715167639479368,
 0.6795550527070436,0.9281077881357301,0.0741117160771756,0.37945477740519307,
 0.5502280619250658,0.25272598958566184,0.9161051616971552,0.8217732981846919,
 0.1510880072791948,0.4065551904912392,0.6772530023778427,0.966914213939059,
 0.06524500438197078,0.6321825070373756,0.33322348263164403,
 0.7845456466144325,0.7422119448477245,0.6013516966712448,0.4354444046252025,
 0.46057972597966224,0.439955868206961,0.9428354541230167,0.9461666298611363,
 0.9452531483473438,0.8410638143179691,0.5424689592976277,0.24918744443215302,
 0.2712168059777207,0.15994582821114223,0.5610869733614763,
 0.14253422482772138,0.2924000839068356,0.2198654965243234,0.5305402591231466,
 0.10064647779383884,0.16984434902202872,0.34489437036575266,
 0.9948467789464076,0.42449189126170417,0.0741624063163776,0.8886719822290476,
 0.36580705328301655,0.6847268779660408,0.9632191540159966,0.2672795566596188,
 0.3277075835397849,0.5563145267911164,0.27168360685226967,0.4482876829835558,
 0.7983047626967628,0.13686905156467988,0.0839877534713851,
 0.25128688392775644,0.7076767007725189,0.8037884288497059,
 0.27464906994141536,0.1633187023980721,0.6656271561749297,0.4671518679453659,
 0.4001140631728597,0.4241191636142959,0.7022485174386485,0.35916673051927117,
 0.9780114814138026,0.6453438046363893,0.7805332898912694,0.29970437549038054,
 0.9367795024976068,0.7244433896530891,0.777841673194493,0.23547546115699847,
 0.2909055164457658,0.17753531929432254,0.3012142788281955,0.3680105161836821,
 0.22813380606368105,0.34998679707126545,0.3072872520630887,
 0.08083000836562249,0.7298053714811681,0.4111130912969845,0.2714046888649462,
 0.28290954348044417,0.9944575698833956,0.6792198594807547,0.9864895995112571,
 0.9318987480258806,0.09307220909226888,0.21320303868613033,0.8784193600182,
 0.5988510378084291,0.4106933547930236,0.9365028201008364,0.960259590073577,
 0.04869616646109376,0.4472790313553129,0.5872054492967825,0.9210524783847903,
 0.793405750491522,0.5824878373839297,0.3963062915001929,0.9954143363147153,
 0.2437775686519983,0.19076598217084884,0.14923923644138726,
 0.5158908140070781,0.39744235283394824,0.6028933700449096,
 0.38666902962385785,0.35830876281082646,0.6706655217768749,
 0.1106643744159378,0.8508924982588632,0.28233782023921494,0.9288074410856059,
 0.05433491674809532,0.11275590002889158,0.3266330678154643,
 0.1528797260797089,0.7827910014127536,0.07352395587986982,0.8785488759672693,
 0.7760706993706422,0.0694667312732179,0.10480193268525928,0.4198418459062139,
 0.11084945477179386,0.4635447382322493,0.44819983634306926,
 0.5069599568015165,0.17626739170229744,0.13725279436361348,
 0.8863677089888291,0.31608173506648385,0.9392711013380743,
 0.05033137807054422,0.5876653072814195,0.19520469791737216,
 0.6932597160556563,0.444208025357838,0.1919683892453179,0.4357089705981755,
 0.11738981911137203,0.8805878784841473,0.2517801988090228,0.9726900920965899,
 0.37558610938415626,0.36595460651065903,0.9035215879015088,
 0.2336058829152058,0.9830772265457401,0.328732466598728,0.46447995422083543,
 0.8799135050825522,0.09614473310394178,0.7850668707006572,0.5345121084840164,
 0.9721206474961427,0.5073723060757159,0.9854182966086794,0.336616699956503,
 0.8860637151858216,0.9061639049381913,0.9744886905353314,0.32417681056735126,
 0.8198759794095468,0.8927877633568901,0.9251457210395129,0.6320828168518151,
 0.008582331392405163,0.6475559277741835,0.33532958876607255,
 0.9401970821097202,0.2825512707871871,0.18689697425138563,0.6876478649822712,
 0.596485759273737,0.05373264180569659,0.662452514256819,0.6849051552108281,
 0.9937313859334551,0.3015294478202296,0.8496606202216125,
 0.005554491246623328,0.12621148149355899,0.28515688991762345,
 0.8148284878734137,0.9631971599084118,0.38966114017953035,
 0.44459729015392646,0.1357785516561455,0.9045956456652534,
 0.44035508803026535,0.19936063593137499,0.4035175181054027,
 0.016595484434272167,0.4741485248900059,0.3464485738933286,
 0.5970886524957196,0.9395007156275289,0.8332541953756158,0.4925531975272621,
 0.7623730190434304,0.7010518054705989,0.47220201731725553,
 0.17952404062698235,0.5819297724231793,0.8884089634142323,0.8950639940612131,
 0.9178767542144177,0.06516076347822264,0.8380181287934537,
 0.21291797373816823,0.1756383971530373,0.7649748279900814,0.8030531447386097,
 0.30888970043225983,0.9772174718495393,0.6309130871797958,0.7372118954360152,
 0.7238739841495503,0.8604771977489754,0.6797596627813697,0.09137645670217809,
 0.7763479493534239,0.7437859843637501,0.36752194188772436,0.9954930306771285,
 0.5079970500552008,0.4402101841871737,0.6718178248331482,0.8622454106869408,
 0.014345524491186423,0.1354877742182813,0.5785524853775639,
 0.05551536985076433,0.35244584854398264,0.28378408705025626,
 0.8142182512885964,0.31944718100886194,0.3327173173857919,0.6261363539058961,
 0.8852648970865602,0.09388198313902185,0.6846986899038795,0.7403217762937063,
 0.6233946710836211,0.9878376099452993,0.08705627666031523,0.7106055616040885,
 0.11560163330779016,0.5951408922569419,0.9740611936351724,0.5639320426091654,
 0.4905350145461509,0.8042995263653083,0.30912606935843734,
 0.44464134532687083,0.6990005524718944,0.21934511535052326,
 0.5242196585435723,0.7294144369691432,0.17386196189506742,0.8289294181181913,
 0.9063065138884793,0.8777601399074114,0.10856009922784915,0.5666050607592696,
 0.8982902686608524,0.03537545240632689,0.8170499970924054,
 0.08925119537890547,0.19783974602130516,0.5696603760008806,
 0.8509152832138456,0.28723961666492426,0.16040581758451433,
 0.7761272937317623,0.849322780609033,0.17716132602054735,0.8604399882064331,
 0.8219742687304571,0.5923313809897516,0.8518154735559743,0.5056471148135985,
 0.38655427687906874,0.3417371495385766,0.16415313967142997,
 0.6514207621318416,0.917083614218233,0.31778192899146795,0.5835692858943815,
 0.5610319003537486,0.2371676364216515,0.44706455591514827,
 0.33720122919670936,0.020567762416789037,0.08988346932807068,
 0.9231437620063148,0.18142340779450505,0.8240846056632002,
 0.42936645753886693,0.7082792310651063,0.778130088798707,0.6404807077937773,
 0.6126370006685459,0.9053602348129101,0.33521960293822084,0.9754451026639781,
 0.9421593351414502,0.7774654468063016,0.6049648925605007,0.9149004946030065,
 0.14000628856799602,0.5125976065258956,0.0711663661170262,0.6051953175554947,
 0.2911994542419003,0.2822686355770553,0.13944435022084312,0.959693158257023,
 0.6808101713156786,0.72819393226105,0.08290114658541337,0.17353544528689047,
 0.4757650724313799,0.31585154296031415,0.9577640329984908,0.9892753458230581,
 0.9382635712766967,0.7606633649960146,0.3751002634314765,0.3139938127982803,
 0.8496437724494108,0.38927606174321217,0.9040589652942188,0.8005903695016097,
 0.6449249210749479,0.8447570630797722,0.2705003490844372,0.3719532552458398,
 0.019954089066974823,0.7275169046710426,0.984641806225576,
 0.002071140120287973,0.6979533516254293,0.4564210183694414,
 0.2348758169189269,0.14270598274189583,0.060617413064221304,
 0.17598292918552105,0.4139266166676574,0.030033103478493528,
 0.5919116389249246,0.4935061269341181,0.5579299470255008,0.7132625925940179,
 0.1800969124081574,0.003519064107961878,0.1356574045257577,
 0.7187689589332942,0.5536962852180579,0.293984899896911,0.51158218665475,
 0.22431541909645603,0.4783814953476587,0.7952162245679739,
 0.07941370174517992,0.09185367064046324,0.28927047833198305,
 0.4070244350979665,0.9968165980638348,0.7164083922965937,0.2578539283524839,
 0.26639886822373504,0.7707341050120873,0.8416026140089614,0.6169433021142348,
 0.5102633446529085,0.9269917527767465,0.0744432446286325,0.206991992769605,
 0.5846462307069888,0.5931001323391407,0.6177319558580467,0.5756089047424734,
 0.9880443920130182,0.9304210385714697,0.7323149976749597,0.866421379091549,
 0.14188295726555122,0.451421368042045,0.8064951861335138,0.33612678821491726,
 0.9457438177564098,0.13407010231336713,0.4903379540440832,
 0.19774067473365875,0.8140834641653758,0.6819713979360735,0.4540272222666689,
 0.3035330359524193,0.3266366945285686,0.1377300788293918,0.8363798446974675,
 0.8817593695485568,0.7892953332911984,0.14596016446854776,0.1034669602798991,
 0.6853544139559631,0.2035799030408949,0.30261653100355457,0.3152929724365052,
 0.1324726583814806,0.6546112269171261,0.5382461560202051,0.6284346227915505,
 0.4314702606124885,0.8276998100578538,0.06296016233709822,0.576601422121323,
 0.5907904534333072,0.40192004622378175,0.2523530589902311,0.9049707910247516,
 0.10910860066614136,0.8179637614096514,0.5537484942391466,0.6011699826590104,
 0.8997137569140887,0.28312249427694447,0.7799297728283849,0.9020985356403265,
 0.23051646653574642,0.3310413454842742,0.9432131193455852,
 0.41053359122634747,0.19161946282295073,0.22884944896123816,
 0.2324377798466789,0.4207016013499356,0.47235872946244095,0.7335178297693069,
 0.0956928482535051,0.19161490426765015,0.24417692751214037,
 0.7013034411241328,0.2774243823573097,0.8700446102155882,0.24707052044626132,
 0.6566458309039888,0.754594798374875,0.8683630298410118,0.6436211821953629,
 0.3916644622201533,0.9458769961167428,0.30066498326783886,0.629944965846933,
 0.06006671692581966,0.8782620592233743,0.24334256196658333,
 0.019648798629630537,0.9480568629099755,0.027191550056334668,
 0.09151645961838406,0.22102372803549863,0.7146666707818063,0.665331611263622,
 0.7545552105872663,0.49656574610404824,0.45762987358718465,
 0.31890345203939985,0.26642297287648076,0.23122250496470764,
 0.4875623348453033,0.39248617157945664,0.30945071310792727,
 0.5043537590850576,0.3924536892645356,0.18078808214346542,0.9521401744702402,
 0.09797544577954531,0.1810078258908968,0.45523609402928567,
 0.4745877397657935,0.38001296611656665,0.7347444722629873,
 0.18829413913147486,0.41071060190025166,0.6515659961017759,
 0.7343956751086386,0.8316515100684873,0.6937248666229738,0.17996735102646355,
 0.9949444788560772,0.6858897437181753,0.48287212783448163,0.6042203352618007,
 0.6394972725684399,0.5575602340870325,0.986921411981287,0.8390507322435065,
 0.7538857611867454,0.84795586055422,0.31259024998605234,0.4907529629613082,
 0.5882624542521203,0.7181692909856787,0.6797621813434943,0.2737478065531932,
 0.06015692997969424,0.8645919669683162,0.9689976062327382,0.5129604973835589,
 0.8910763301996442,0.11963914240195872,0.4302674149527018,
 0.38303249110035353,0.2642730278475667,0.09989483303496693,
 0.24336556875706084,0.6132155901754479,0.7171309690744108,0.9516705110660285,
 0.4449029012844369,0.4015214665028052,0.023074401068139627,
 0.6971656165672601,0.006932336449759324,0.7214042138729392,
 0.8515237139850873,0.3655209721673828,0.5564871941133692,0.6440715941616193,
 0.15317574632958297,0.4836065924311117,0.9283504339132039,0.5095203194697503,
 0.9082516051281444,0.7422426828187059,0.21785607490130288,
 0.18462226155876027,0.9773411455583016,0.9710167803241851,
 0.33822371781627913,0.8381446169830243,0.6296347532312145,0.8059286708926213,
 0.8028340970025512,0.2867253197904479,0.4207208113209925,0.5100326862857161,
 0.2775061721769584,0.22712128146614352,0.626869766075612,0.1947872775930184,
 0.7857783987267309,0.1621497927058534,0.6617382748873382,0.9103248881787956,
 0.849974460075314,0.20251042758146487,0.6355599477870529,0.7770606487873557,
 0.39994412585863826,0.1880511892558372,0.21073750074034392,
 0.2677397086285922,0.8398333110201164,0.23626013875916496,0.3113766862068076,
 0.6025205780093444,0.6041010588549308,0.6388191759237034,0.4658609625726752,
 0.626782082176151,0.3544720544000154,0.8806967245382122,0.6598613709151385,
 0.6892221111046275,0.803947568722597,0.5738604499518024,0.43577429776344834,
 0.23032711745030265,0.13433535299202037,0.733775214540816,
 0.24163158405133878,0.8127997173909304,0.9075477559725125,0.6162181720281696,
 0.9284934137209931,0.0025582646514262786,0.1393318545765586,
 0.9685476778237856,0.9985442321711313,0.9222510345843163,0.06624650402185983,
 0.02122721326586463,0.8378151949543629,0.9924007544029437,0.4918010122945019,
 0.806196776702065,0.16288163158789437,0.45905485278999447,0.9070075671591235,
 0.4284052057436951,0.85562727637338,0.7749784719058904,0.49094799970372915,
 0.891084536768832,0.24400974726946323,0.11384569306470349,0.5321867234569428,
 0.9325972382528275,0.5898115476707875,0.4544381856256514,0.0960009551876082,
 0.8732555622861394,0.11131282840141454,0.435467575542947,0.5667327263864909,
 0.1651388286431864,0.10329543953289444,0.9300529160962787,0.4982134157370385,
 0.8327929683103648,0.37177576798560263,0.5778346187872092,0.5097931441870794,
 0.29229465961013823,0.061351114244938776,0.8344597019538471,
 0.4992813050688625,0.3458692335434922,0.5822924207639927,0.04796309484304406,
 0.904726859545535,0.1793049829454455,0.06333735485233305,0.8947906244863212,
 0.8617037014222532,0.09778674178882418,0.9354758344039558,
 0.11369247679747363,0.7933459050406045,0.8987871161330003,0.5503369055449329,
 0.14598331582177715,0.3060718557752602,0.18764054231609106,
 0.9456959452708456,0.9689128447125414,0.9569079445834053,0.8434870298425952,
 0.5464497426778692,0.13272550829710206,0.04497095607154289,
 0.7372021473697551,0.94027486666199,0.5995961975049557,0.07815040097038506,
 0.5967794675452931,0.5931604829887096,0.09915265257764472,0.4231487541817529,
 0.2838954756794251,0.00573594336230987,0.2883041620704849,0.0514517877869356,
 0.9683299816562956,0.5454099444228732,0.6748285135574379,0.6060885443088242,
 0.9076171457936587,0.434487213544519,0.752457003982885,0.04345369553774481,
 0.6701410927122733,0.6904892858861228,0.47893168971613653,0.8606620018049173,
 0.9427030982320554,0.28622224825491716,0.029417380349644806,
 0.428454075306013,0.424933065841551,0.5491058040704468,0.8005488919625042,
 0.5404102647873167,0.4220954571047474,0.20619679096302934,0.5691561374524637,
 0.4491862679347317,0.5072422066915815,0.973580486169313,0.12001698579434794,
 0.3525653868157498,0.11723842485707459,0.26436267564176874,
 0.8289516418172003,0.6783551907043748,0.9376879234303978,0.05541533265035703,
 0.561425792036008,0.2958210717630805,0.9718373078921153,0.532898333514986,
 0.8414042334327037,0.17862869279171967,0.8943326714852649,
 0.05490378117562855,0.6676020408189539,0.47081520669378674,
 0.24890999672342662,0.19320516073384708,0.46763721391803337,
 0.8742078731690274,0.7733296223206818,0.8769967205001274,0.6650364856851744,
 0.11507041528888795,0.173594409521993,0.8906126515013147,0.19816953055100328,
 0.5079021634137774,0.8227260426923421,0.6983092911004963,0.9523655063791984,
 0.4678512706822302,0.8591004379258885,0.10604369958331583,0.2780228423548988,
 0.6187773283814524,0.5021129707502273,0.6749667047840564,0.35498129674269063,
 0.72427942483168,0.8737377790866495,0.020076232825443685,0.7216047644916058,
 0.28211939994586077,0.47728395735136964,0.805795800423549,0.8530379810769094,
 0.9524068137147103,0.777748692502195,0.2409760730968874,0.2672567013815256,
 0.11446688509267233,0.08818809710666731,0.11553554927517407,
 0.9750005839755072,0.9176675628694277,0.5826443714502662,0.5973927479796278,
 0.4453463160140658,0.3085934003097194,0.010149389551786903,
 0.7297654797384211,0.0936331927448919,0.020545903607604665,
 0.018496555405992154,0.35137670113822783,0.8314577574659827,
 0.3838662617726525,0.6621186239131133,0.5861254050858566,0.17536187995720776,
 0.14263888100150213,0.5256932283790315,0.688967788340588,0.42128825281619897,
 0.4651821884704407,0.30938180677018134,0.41033384947768703,0.422528960730709,
 0.6671879343363321,0.3661981385325985,0.010156316597259574,0.75147071331143,
 0.6913536223539879,0.22408266877929395,0.31677209699653053,
 0.9795766817220688,0.7088963997592027,0.8139715035168974,0.46285948540287125,
 0.5262754183270725,0.9791886205718365,0.40224585697285664,0.714582861877449,
 0.8490394673159231,0.23601033815280592,0.45688934599236897,
 0.43072001897908496,0.3896658855588625,0.527736573172019,0.514093299984593,
 0.6750328242295013,0.6898428644995686,0.8523905186104032,0.6961085972480232,
 0.4253830931701412,0.2888672427083846,0.7271206736803336,0.2721170370103293,
 0.09277659366412583,0.580946357737968,0.21721051600026653,0.708143951770564,
 0.8653368820723043,0.2028414872493789,0.861618850875382,0.6716736380848273,
 0.5279405739511881,0.9389934668346285,0.2831879175612101,0.06594277917141456,
 0.8688433617549299,0.817213878767488,0.09877845099921356,0.9274186191775087,
 0.7121839722418889,0.1160577386743199,0.6103957440227215,0.2814866599259066,
 0.5318235119717776,0.49076441792267755,0.7705250060589445,0.4600237685349666,
 0.33100815134763684,0.9195091697008753,0.822262766587624,0.32739853374425,
 0.9899264189172075,0.3353517255113364,0.02573097594246998,0.6126374522691825,
 0.5907617562707431,0.44974406366332154,0.19250904861348106,
 0.6381672781929234,0.18846303506639006,0.630021244298867,0.07977176075774428,
 0.825659847188701,0.8217761007465878,0.3443326600591895,0.5651773588029941,
 0.40100446140927415,0.6210185046179155,0.08161850929881398,
 0.8174469095127425,0.14320763806559866,0.29946987009529147,
 0.7769444628781861,0.7507320718220768,0.002893849992992248,
 0.9301752416871649,0.04081734369288459,0.3913003524402866,
 0.17067535522489152,0.913020742603364,0.8429174526964436,0.8109385531233967,
 0.608016771918174,0.5065999695473923,0.06311125129899087,0.9738328289137392,
 5.1344133186082e-4,0.09035300978349392,0.5869132469098663,0.2648746502786905,
 0.8581204481180845,0.8453362064590565,0.8671096618221634,0.5230647484007527,
 0.1869832315057941,0.0032434588998944136,0.9997533009007975,
 0.31849391736323795,0.10882030678135635,0.670332247876095,0.2367823781193339,
 0.04924315258580092,0.37130434911194043,0.24482120021593912,
 0.8117311987050608,0.023160931344378355,0.07266838348663796,
 0.6367646320175621,0.4741593357602497,0.8010914120802529,0.0879999525956422,
 0.7861668436499911,0.3286342900586595,0.6327104096348819,0.09761854409386339,
 0.30075391157095244,0.1569733984641778,0.4718727155991046,0.7397936178994031,
 0.5247503076444897,0.29797748262656776,0.3521271771256498,0.9637846789594307,
 0.9295251736180026,0.9001404823405167,0.17813526394690338,0.2060939483343689,
 0.3405981456108238,0.7224860741846065,0.9338670752757701,0.4147529851128785,
 0.4195178141444833,0.15476230884185826,0.8304187454915392,0.6228546026115096,
 0.35785398108824207,0.18602599672100617,0.8784244924323898,
 0.06367978754660863,0.11348808193162141,0.5895778469139141,
 0.6175490694801716,0.1761711538378693,0.8320071208089261,0.31295020678028906,
 0.800377311750004,0.4449237401878008,0.3437806597279106,0.7343043175953384,
 0.9092774614179038,0.18243796152480662,0.15393814722581534,
 0.22741261283607683,0.4873631365063593,0.3216305843561392,0.7339593450286279,
 0.8253543909394391,0.9809639508378658,0.0021182406837337853,
 0.6164516170050558,0.6093530265726729,0.8250563224513331,0.923670988019581,
 0.01871700451986913,0.39675557449964527,0.5936208320188161,0.740940429965254,
 0.6059351409826268,0.33223217754937273,0.021246412421587424,
 0.7654357313422644,0.07150576175684598,0.5970168236733475,
 0.039429164203029776,0.628640355285321,0.760487753465038,0.7157042419597345,
 0.08113059764886954,0.28291588015272584,0.8665559768438845,
 0.5131798972386796,0.11962655617724871,0.47481466937275774,
 0.47665297010509533,0.5675869922526936,0.5577919439431283,0.8525312369834946,
 0.6131783274967119,0.6642746057751164,0.49295239530629553,0.7014451016007837,
 0.2593032087418584,0.7236821513947304,0.20676350993257497,0.4975767719010965,
 0.9560074486724792,0.8164141806410666,0.22024576117413275,0.7414366812358183,
 0.5102008443461248,0.06424234998123457,0.020971131889577332,
 0.3999209519792455,0.3384395175956585,0.9422737014443607,0.9711116862110375,
 0.5166797639481004,0.7723192213293187,0.3582760518418976,0.7735833481355561,
 0.009521551014793017,0.09014305838298775,0.5661675169402901,
 0.399469439382266,0.8154638633097016,0.2999003985313401,0.28291388884901103,
 0.738283496410996,0.7801047221852772,0.6181714616014188,0.6411095347063808,
 0.34884196550141966,0.7994108677341563,0.6451318176177677,0.8936627979378494,
 0.2196407471339713,0.4308247547754349,0.2919525853072483,0.16143032180547423,
 0.1492234736091297,0.9623175637208051,0.41594959426770384,
 0.16359595328511334,0.7578562881879077,0.49933371232044954,
 0.4944249992726354,0.3532839028883693,0.15149483490630145,
 0.03920084071016705,0.15183368404471642,0.9868823285673831,
 0.041994026805180606,0.023741550938954914,0.7882512040392606,
 0.1369646831498621,0.437636874844623,0.9799377005613179,0.7056268463511721,
 0.34078943849283316,0.020340513415425643,0.6993495326115107,
 0.12005237746752961,0.9931553045852963,0.8637395586570613,0.5054719507882041,
 0.7481947935431265,0.9865307107235454,0.9543140143209453,0.8924375921611339,
 0.8357044949409816,0.018597690753803353,0.24340166382163808,
 0.9955832837648612,0.6578514405315189,0.23187039295532053,0.548741245296731,
 0.2738067090266205,0.6244120725167278,0.716272673122555,0.8457328902119392,
 0.5191904042648151,0.7601348541480827,0.11262393276148353,
 0.12614994226388831,0.5612266720044585,0.4616434971268121,0.654077860403134,
 0.11174963679411487,0.7166379573400192,0.8412027962102174,0.8706969052651115,
 0.46844356558879996,0.8783846743372822,0.21798678398354698,0.271637720332209,
 0.7364598457729129,0.2962439386514504,0.17395884924384075,0.7462815268258387,
 0.8298652344761224,0.5463379789513225,0.4634891843890505,0.03459985710145974,
 0.43640006625470285,0.828814992173082,0.7188727811640806,0.7390466213104457,
 0.6307248062449908,0.3733696142894829,0.06752534321588932,0.5134934494642428,
 0.25338427776940264,0.9454085661822326,0.9474089063880979,0.558274004003134,
 0.26854143399374997,0.9606429407820952,0.38709103048006965,
 0.2614788128196277,0.4193277393562478,0.22964183301031205,
 0.15425249829953802,0.18280153495045037,0.516205998707092,0.6664114352979702,
 0.743727860535188,0.4856359156110931,0.15973930335395115,0.4973580195767412,
 0.3346862749928211,0.43824461279240556,0.4207026305693018,0.5105811574898517,
 0.42361246934210506,0.8461171721438565,0.07937160884446093,
 0.5796069964738678,0.6971315899983576,0.9270383364189324,0.1488215011220969,
 0.9092932827725939,0.4637638698542097,0.20053163468412494,0.6941796613881691,
 0.2876896638305504,0.01812920980588162,0.2258463207340451,
 0.026641973463921387,0.4463381229341472,0.27320428310336586,
 0.8883360884284215,0.9283545938550635,0.4549099989752402,0.3370468452504649,
 0.12211336857183797,0.8880930331902127,0.24611242915188314,
 0.6731264814178398,0.8089550923647211,0.9098654218947139,0.7089055791141353,
 0.7203701862795813,0.12074164160927203,0.6186563325822674,0.952796815241899,
 0.1463771092377476,0.34870705802009494,0.007287210457423199,
 0.1606789479470916,0.39950848658160165,0.673438296110312,0.9300073478946552,
 0.6852576244366444,0.7318126160448005,0.13485919117966472,0.6928452504228159,
 0.1786797072536339,0.811447125162072,0.23280503253378415,0.949454759594906,
 0.8938420296087224,0.9962387469777892,0.5072693933027822,
 0.041138842884215765,0.9600356255529686,0.6947819138450716,
 0.31914187175218867,0.7296042311411972,0.8004674845714398,0.8216390760522876,
 0.3117840939939994,0.9773083914023984,0.49742004512914895,
 0.10984194504441969,0.35556232199970683,0.027762694657614606,
 0.5729505207564143,0.25419657601939494,0.25822417512317664,
 0.16944339238723216,0.3709736912478693,0.47818354057062007,
 0.09977307010669201,0.7422424362734721,0.30119230761224747,
 0.1360173989238308,0.5815619046438075,0.7281880752829611,0.6978074793264649,
 0.93031112526728,0.7442794715334758,0.021045047662957894,0.6689860132512715,
 0.4306300174732445,0.2019420368405811,0.5500404902272761,0.04753928411749764,
 0.9280614835469891,0.2980759405738005,0.43854871493751924,
 0.11455318016155114,0.4069444599886398,0.20346127571619943,
 0.5680367081055203,0.9127164676555021,0.975466124846436,0.8480219159271173,
 0.9880950990576735,0.0028240615122185897,0.07082539422495104,
 0.8729354733605105,0.8552106054118828,0.816344299477167,0.8425962437750999,
 0.9479946268814117,0.21170132225707905,0.8988645910887443,0.6152962247840699,
 0.4887480355029057,0.2962918056796323,0.4497228622988946,0.9972454531887042,
 0.6918274590912479,0.3650445725253858,0.477967267570066,0.37253773138787705,
 0.43217827036654477,0.8399161889432109,0.22022659764162622,
 0.5047473509039971,0.19777986630988065,0.6753560741957645,0.5838790822141435,
 0.9209918326710678,0.9161277492902761,0.15031841914421862,0.5231792004169237,
 0.6659987695448272,0.6949734257043584,0.8286021667965939,0.7451918384589157,
 0.08615717884475727,0.3003412913972203,0.10456988486170105,0.627943249295369,
 0.8467158201952814,0.0467739992183579,0.866028731049646,0.5344056649049198,
 0.1993810967274463,0.7905108478858239,0.49262475160538277,0.506139195500698,
 0.5814151652975248,0.5134385161042248,0.5677951982043717,0.912991577468143,
 0.991856890911244,0.7773009855402888,0.8803148607521306,0.8248238935642641,
 0.2892278213880155,0.8417300223798576,0.3142035634761099,0.4008287633876695,
 0.675069664488149,0.006505223066725185,0.7189699851729402,
 0.24198717526574975,0.010473454245853664,0.6299242938142151,
 0.580064284360194,0.3780931036662146,0.9163520883377924,0.8884523487345962,
 0.5762830400030516,0.34437355469277064,0.26237651241310356,
 0.8303615679897691,0.4633685669307459,0.5438012599707973,0.32111357750172576,
 0.7097736159000441,0.21356558214059862,0.5298300211191579,
 0.006977153965490479,0.39486610923317766,0.6385268251365334,
 0.06367093453669792,0.521288209238379,0.8141495424360847,0.45704490456086666,
 0.628603048882026,0.5142509701858561,0.9376728827925864,0.6400519345963838,
 0.4554692345698369,0.16753063795925194,0.222957768869245,
 0.016349781254113238,0.346839329593715,0.7785934903197541,0.7570133259649379,
 0.9604133844972275,0.744893768564624,0.9221797435238969,0.19851923834714869,
 0.9736300889617573,0.3850839822274148,0.3770898385930055,
 0.005595564769951711,0.35227994849147903,0.5796925115593086,
 0.6558898136101478,0.10815513329514115,0.8489042314490787,
 0.014353041136899503,0.5432795020762761,0.029807939937189776,
 0.2514645810738415,0.5748566104611894,0.8002542397188079,0.04082155798828291,
 0.8518379801995266,0.5113503184248956,0.6873423017501089,0.26052434183001816,
 0.39358112009919666,0.9632451992459511,0.29840504816646984,
 0.6310934521535628,0.5775241764978988,0.44401506026061055,0.4645973702878097,
 0.2202966561477715,0.09955168736437892,0.03739521027309456,
 0.4355316583227957,0.11378344706243204,0.07013600509150475,
 0.4760439474522623,0.333403537160328,0.18334815150155603,0.00428520265243415,
 0.5704950495419786,0.44355945289887144,0.6886341236708768,
 0.13470962366241246,0.3552923510969743,0.5926816679744227,0.8879754608304253,
 0.6484548005395111,0.892841275954707,0.25053522602554656,0.5835316871561107,
 0.08637439341098818,0.6845134640007349,0.8116310928453125,
 0.11558703325614816,0.6252632695904843,0.36364572326100697,
 0.24031627582010362,0.043042094205118064,0.721165940353728,
 0.2518816480929351,0.01835137865854286,0.7798849770519438,0.5247963514805809,
 0.34745398445460407,0.21325385934544716,0.43585962106287834,
 0.16734214033299477,0.44798502280638186,0.9754715821564659,
 0.23717689296448397,0.12164473504273177,0.5903843617605716,
 0.43883480355779825,0.7055540089939426,0.5540226175481134,0.5544670168995235,
 0.17634696515318637,0.4496017504642361,0.03145408663104765,
 0.2629419056577955,0.07086221420702266,0.4499688755877329,
 0.08974678283605719,0.053003033604103966,0.32532986282418574,
 0.2803275357491476,0.17950739142900463,0.11534519251573427,
 0.18641422340535674,0.0733617376288983,0.1670039911043033,
 0.12470261210928524,0.8643487771037287,0.05626906549367339,
 0.02850607591000609,0.872803049889816,0.6498800141045238,0.3265096391409765,
 0.9489067023320263,0.5728922372465278,0.2244130285167384,0.887651888105274,
 0.5675663280770923,0.16129243000440086,0.2617965572362905,0.8052554858522212,
 0.16128475094250805,0.25375473798231307,0.5073401190295128,
 0.4690856697006317,0.14158057057663453,0.13151037220764095,
 0.9894660710981285,0.997914536118131,0.01009060467137135,0.5698052795947448,
 0.5373965486538751,0.19586088149323022,0.8337895076006956,0.7679725220513475,
 0.9121535337422884,0.7990303328457976,0.5798801455229269,0.9136054792575677,
 0.9077587993218027,0.8245542544828678,0.14125883919485482,
 0.05500048418562209,0.08878921001004214,0.471124081001304,0.282816744102259,
 0.7515462029567056,0.8419319105264741,0.10882771144262704,
 0.10244187433420004,0.38761857564959046,0.4286080620709506,
 0.7190040434171686,0.14929786295385217,0.5759076634913056,
 0.35924978151332887,0.757816122247868,0.5221264727945117,0.5778118832918165,
 0.9430328936579141,0.34960647143989987,0.617767666263442,0.985324312282807,
 0.5989304481333951,0.25447826538075935,0.7857003517510356,0.6781261988384542,
 0.11544458577592498,0.4603161415634057,0.7725107524038479,0.9861840507144221,
 0.8135655073997395,0.2976762372559465,0.4095307198946093,0.8591500375583211,
 0.4114780011305933,0.8786529929448389,0.51914630229373,0.4447647621797465,
 0.4540837136817114,0.5225517444908803,0.6022259871483442,0.6180889804551953,
 0.3956345357247022,0.2352971488126121,0.5294489472454671,0.2629189740544664,
 0.9651321517712359,0.2819438305341532,0.6623680797865049,0.6793388071077915,
 0.8957528544868435,0.31062548965909653,0.9887099720432723,0.9196023893431848,
 0.9703974198897464,0.5734026561585243,0.24849373096273863,0.2677853520489081,
 0.3653190996096972,0.7689163465329945,0.7398178096411803,0.8107063444148478,
 0.18386577543576677,0.9723621747101415,0.3179147218221249,0.5590803845102844,
 0.11328039216954755,0.19267603310830295,0.10298814179097093,
 0.8616093697318471,0.1001882022508026,0.9720833975546332,0.14622716604530106,
 0.7552466031403212,0.35638211480605664,0.4716085573077158,0.8481416757596485,
 0.0727651197889061,0.43374429331100595,0.08242820682029939,
 0.2531807903713248,0.9151548579864358,0.3459606312746959,0.3607100250362494,
 0.5865155258235982,0.8180298080884522,0.6994302416305097,0.763368575298136,
 0.164140376501812,0.8741321839220738,0.342599524547029,0.5507864737282431,
 0.7874353959826014,0.08055563681245503,0.4540832080799434,
 0.42757813236866304,0.5631433893268318,0.11987678745159791,
 0.39091857005116126,0.40903551513666125,0.3856255577293566,
 0.7992781049040716,0.6210775352321881,0.5598911801207883,0.6567110604754995,
 0.1355127072192467,0.22664002640927383,0.6193912204282673,0.6145345615928736,
 0.7265511770510945,0.1898700571437878,0.8446184123031621,0.9381621713266678,
 0.3028006380688133,0.9541743923925132,0.5179030813855553,0.25133746453817785,
 0.6885125674711381,0.22725417244232893,0.7904589403443626,0.4681911564491984,
 0.3743965403269989,0.7714138785118337,0.7986471626839112,0.5859997780532139,
 0.44322988077646486,0.7862208469184269,0.0355890369514672,
 0.38052903174511576,0.9812678005053631,0.5439744817672988,0.4338497498496844,
 0.22934918174055452,0.7335272564460709,0.6354992214690935,
 0.06751940056656913,0.1694014087433805,0.15895955674615525,
 0.4004980345028679,0.4962699163395722,0.7244661032904225,0.7370660603685378,
 0.8261965304629443,0.3535891124139815,0.5164371661152474,0.8059291747883449,
 0.7999864760142885,0.5260619415082386,0.5323322233975237,0.8093804821436131,
 0.6393567762936083,0.11529207278031384,0.7299556667781438,0.6354694521228311,
 0.08032120739497539,0.22508146949257224,0.1770285744531439,0.696263743282633,
 0.09842692648164342,0.3247900446869013,0.7441159774194677,
 0.08060899938669552,0.9655019076861024,0.49806809929832996,
 0.03922734996071808,0.3212763581416719,0.7313845605588555,0.138032547008358,
 0.30014168423775944,0.8069025940761992,0.26675145806935474,
 0.8644400761643178,0.17829237535420006,0.4707636669838252,0.7512340206827299,
 0.8649801310716816,0.01577577342521408,0.2613791591907392,0.6203626408524726,
 0.07555707697550496,0.15675190776649817,0.20536707495917494,
 0.3683789808490884,0.41829337417006407,0.8736361331544817,0.836549892249515,
 0.8271517847881774,0.30957122327981934,0.6596312836915377,0.9875003202548418,
 0.9686333740413672,0.7760439715834581,0.15546945092826414,0.5459521129554421,
 0.5301166264789894,0.5245141590171032,0.39967943997234734,0.8260426861246147,
 0.4177379367477102,0.6311046959810638,0.8878297207985287,0.11337885621090726,
 0.8993024243417179,0.24686500951789736,0.03569487078219247,
 0.9040166285687598,0.5849801432342682,0.7050825420303075,0.0809019184119294,
 0.8747714507999681,0.8993375603731126,0.7748384001610555,0.9251025289050454,
 0.19156015262045034,0.05907500255617126,0.7407769209114758,
 0.39218703307131353,0.26425464233801144,0.7376492707352401,
 0.9105290929620709,0.2484302289854634,0.03550968014734712,
 0.17055349720245627,0.043284339356509105,0.6698510349572251,
 0.8332844300497693,0.4885888051921343,0.6774004415248158,0.874284838918668,
 0.13381187269313166,0.11727659627390696,0.6528683992924378,
 0.2905416473984468,0.21258134007208385,0.43612314975836475,
 0.05922075621756029,0.2260976128785911,0.4826001317080626,0.86070500542388,
 0.96189076516848,0.11888475551475663,0.2582299740866435,0.2835134150519103,
 0.19943769057734617,0.4668962291996299,0.9281718569077728,
 0.14616935187534352,0.04311045896185495,0.13968863135106657,
 0.01877642098770793,0.19460047481378062,0.21313849571923216,
 0.9181727370748884,0.03468853383955228,0.9924834770209007,
 0.07235167088498606,0.7796463611999249,0.47821031437804773,
 0.9001613812942388,0.7222478690483358,0.3654443371034487,0.3054022295769344,
 0.08329179080490623,0.1252171463461571,0.18787768653254933,
 0.6933258247479193,0.10935446784776193,0.08222700363754132,
 0.7899542962058803,0.14503270239962163,0.9783535112866248,0.8079543375940159,
 0.16303680818307065,0.6850516437662374,0.4761228683940061,
 0.025557714209280213,0.9745665237119862,0.7106918918598217,
 0.5659110032195593,0.01235281440790692,0.05721789072481365,0.978996263038799,
 0.39243849622075255,0.7938381078147014,0.10625064423025032,
 0.19829677471837814,0.15735894517509474,0.8089340420616354,
 0.4999027894596063,0.14459001129703342,0.9089255796843805,0.6197719724940236,
 0.97901690648137,0.2413847818143574,0.7442143252026211,0.5289774604499412,
 0.6237418329220145,0.046787301425659344,0.6099389946289006,
 0.2650602047791881,0.7127953076657311,0.7761152052486149,0.7620930050721513,
 0.7600433089628577,0.0028414681790708274,0.5565972040534304,
 0.572505296756576,0.16184491606796092,0.5055497714175586,0.340317519851316,
 0.3865058337660324,0.13300722171266532,0.23341317116371174,
 0.9227621831281839,0.4927822748630434,0.5130110054698803,0.29193811261380465,
 0.16560134738379173,0.1287484302760229,0.10165756880012977,
 0.7677855187479599,0.8576296475670236,0.5791237098102526,0.925109974502674,
 0.8681652427842047,0.21963996910688466,0.7430220361113246,0.9673375026410937,
 0.08675167380068782,0.3506580266452243,0.425002732893935,0.7406762871151675,
 0.12447963771519677,0.5153785849229375,0.7374499581409989,0.3327741600216153,
 0.43627255268536025,0.6663410198528847,0.05277842428404034,
 0.5098970277032439,0.612597574868524,0.7605848015517394,0.21911982089764626,
 0.509025139072202,0.014423015881415813,0.30480794999765914,
 0.9949180555430327,0.34671391385058137,0.6487931886648758,0.9550754324578662,
 0.7018015906270293,0.6209808810947124,0.2925671687652396,0.556081056363141,
 0.24099060374945713,0.09567172881911723,0.8232811148429391,0.705164391702123,
 0.07996875260613567,0.48788823873400144,0.09632398249508078,
 0.4609865856252505,0.6922077767466661,0.850658719363682,0.12654197087193364,
 0.017747579003792202,0.27417244918214123,0.6035116639875405,
 0.7461484440465446,0.05184094616508117,0.40394277067659556,0.845173402200217,
 0.757188408532039,0.5190103634011499,0.3230615851505704,0.6792006793125202,
 0.9396909949651684,0.4810578730358276,0.5098050739678303,0.8340127951857976,
 0.7696294780969654,0.06697724553291551,0.6829141185568437,0.8958425832357692,
 0.9141148097801973,0.8620328807659248,0.4180863870261068,0.6991996399205379,
 0.5951489899919223,0.027716881850119335,0.16223266192218633,
 0.831398376293684,0.27956929402325126,0.2111697626788117,0.3891894240014464,
 0.7113296486632366,0.6370670313984995,0.2548277248791002,0.5573143923319189,
 0.4781455402113739,0.49824998294053446,0.7804250306341431,
 0.48449299323913475,0.7957591122674368,0.16210343669919458,
 0.49462873197346924,0.414967239793932,0.32561307453925803,
 0.23088750162117488,0.13969059854048071,0.5792671131581244,
 0.6106464817438063,0.30968526482978664,0.36199939071361453,
 0.6777178059634092,0.9951677587223934,0.9485647613626897,0.8118105266508302,
 0.7812548010690257,0.5935182835363108,0.5761221012589285,0.30023707667733557,
 0.4276544562413239,0.03315183207094474,0.8916632448392046,0.8789832618250764,
 0.30407835168864206,0.5818286876369774,0.06665420408163747,
 0.011570456692315267,0.16827493692780793,0.3151656383690826,
 0.37968529058574274,0.0831957242789304,0.31299816936348746,
 0.12224393110122633,0.69634618951028,0.546118233632843,0.7801721640626263,
 0.4515993793038582,0.3449208678058243,0.14532581111682252,0.9606603443319036,
 0.0340800104544845,0.8292932601905573,0.3021037475834768,0.1859905020094248,
 0.12925173582193417,0.8327196511999903,0.6113046888498697,0.6288052825845314,
 0.9542516565267388,0.19797577315383963,0.82313236619867,0.06109935220950036,
 0.49713745047554814,0.5488831189292153,0.44535457185946303,
 0.9720435244631096,0.9939949818230623,0.3422333750245923,0.5018044836568163,
 0.814206054955585,0.5745510733950173,0.7643849459408252,0.13196073545028852,
 0.32413409504307766,0.2649599672527887,0.9537887029654175,0.5512788711379574,
 0.7227552455440642,0.5484069182237881,0.854192396054593,0.8412819626303968,
 0.41913762232781593,0.12010322284849018,0.17300330340571046,
 0.7755613946400063,0.2891963456658364,0.6024298071333584,0.571560361419499,
 0.3624806192953507,0.5979268979912675,0.9266016553832297,0.23880702052253955,
 0.21779785368951465,0.5351921860332651,0.5435461729341221,0.6099749171871591,
 0.9770968840939848,0.6610101967687753,0.8425428134872888,0.4631069152962789,
 0.6639493527425442,0.9882475784915805,0.2924129815182498,0.551500422973002,
 0.5455051678368408,0.5161370060551819,0.8397099395665657,0.8278990436457825,
 0.7318283293500337,0.29040724625134384,0.09374431193495192,
 0.012865873530473526,0.6358689424505258,0.35457194553752147,
 0.4865657812978834,0.6169583743034283,0.3112470203498905,0.7122127227199605,
 0.12906932817528882,0.7032876171519673,0.3552930256830702,0.47811164930507,
 0.29350033731739245,0.016899041975095508,0.5838430255417529,
 0.12358818297850194,0.2387550916134138,0.7981570287262022,
 0.017341417568981332,0.4060668059346506,0.42162329771995277,
 0.3026245274787943,0.3956792338688334,0.10932849877284123,0.4029333720548762,
 0.9978512123753597,0.2193090562228982,0.8654096884037802,0.5614474925166351,
 0.7668233395680357,0.14539892815935673,0.8437863120175011,0.9809452954737763,
 0.5959183383747065,0.8430967157753838,0.317012562072425,0.5366929713899539,
 0.8662770378849185,0.9660325501601152,0.7535381628997161,0.39560158761084196,
 0.6836809099204838,0.5671036860516399,0.26043574778119094,
 0.005326110565506137,0.9088950274389038,0.8046036137174634,
 0.03170158648524912,0.04870892129911264,0.7267503669750166,
 0.8962893144806696,0.16411462434580892,0.8654126276464362,
 0.013430886629875904,0.6631774877478889,0.9217061140489725,
 0.41857108966666334,0.6300800322976025,0.43484352881403665,
 0.11882150884235798,0.14510117566929348,0.5201596171222174,
 0.4225758657324462,0.058783828188602616,0.23507899120477027,
 0.8172482434466875,0.02882925935949887,0.32054084941628547,
 0.5283956268191298,0.6207024999845875,0.7904069981240971,0.45245463402885766,
 0.5230265611064783,0.6430452435329204,0.19488978530206935,0.8017177431637613,
 0.051036865580869506,0.053586656382288544,0.015069336438797487,
 0.5534666213726287,0.691832725610713,0.5945424320962949,0.687178038576877,
 0.5491145571056589,0.8128574205914649,0.7228826793001607,0.02922974624520691,
 0.4798288552523965,0.10555800991826125,0.4834692360254981,0.5084680223305171,
 0.1221889424082494,0.7704293607194814,0.13930436346732566,
 0.07367144539638293,0.053029852610729566,0.14762567016844563,
 0.5550336219531246,0.8545869231806554,0.9889495518548047,0.7526155889702631,
 0.8560412355267295,0.7958601580973598,0.5269454229913678,0.3280902844552378,
 0.8184672152596082,0.45642428575244276,0.49026990779241386,
 0.6886441929509625,0.14626981423835372,0.12040163676359117,
 0.3986964019738446,0.8422631233284238,0.45771637957137934,0.9031525608014424,
 0.6058922335418971,0.15987197142623666,0.8191739948505028,
 0.0020005604231402296,0.11765009292400785,0.2581767992939342,
 0.7860826692709058,0.22166257844439086,0.8232579176749399,0.9573288111963673,
 0.041713266749043965,0.02071648087052569,0.002405405357096435,
 0.08023574866664662,0.5484647088046993,0.43692020490814987,
 0.6914122031464723,0.09781458044641511,0.557251899997755,0.458712979720047,
 0.677282919026728,0.5453838565191168,0.5861850114739374,0.8605408512945717,
 0.6987913033073057,0.7235024906502758,0.10445235504461015,0.7474545165561512,
 0.9786918164871999,0.06394112112024553,0.2989115585912161,0.9800102281469699,
 0.6628138004391271,0.5864360393249238,0.8648402968122233,0.8896083827803313,
 0.41958136034269167,0.5211157299952751,0.7314547298066482,0.8182950683200416,
 0.9134298029496452,0.8809673242124788,0.16687674523009366,0.9895703990698349,
 0.7958350117890323,0.9207409998381025,0.30946046817491124,0.6342252182301721,
 0.8380658587751609,0.7275291065880237,0.9186189644536793,0.5678571204257655,
 0.7561521302849935,0.9485890209929925,0.7096712905825264,0.07344306529044697,
 0.3009587504022977,0.8498038497218027,0.6494617373917942,0.7274388127190817,
 0.4644505132705382,0.9858720442832143,0.9325205380313036,0.722136827552196,
 0.04470795591745158,0.39009949696967605,0.04015850418830025,
 0.8568133360914955,0.2573505106715023,0.24596702080063126,0.4974919566779695,
 0.9993674247958099,0.568141193907228,0.1142832565191847,0.7918968286104378,
 0.19352390019240262,0.5111582114427514,0.2843725691268203,0.5935525835941962,
 0.7384322285612478,0.4672036589592533,0.6422495184844051,0.928571172632898,
 0.6463189191617987,0.3343457904607664,0.9046085576701757,
 0.014151436285415486,0.7828172058850277,0.6619211726309961,
 0.3530794495392938,0.5931432700230616,0.03020516992823985,0.955225410015693,
 0.4001038465771075,0.4942418461641168,0.16500060709067943,0.7574119340623787,
 0.5514400101145751,0.5641421902819568,0.324887420105042,0.17946227632590528,
 0.852775909423197,0.5998641943691381,0.47426408060936465,0.5746348684658591,
 0.21043555863033614,0.46327569785365896,0.19756099002732974,
 0.2257721746258886,0.7326181915583317,0.8143065336077314,0.39746721362042103,
 0.7882408574981622,0.7317024036131549,0.8279685060863333,0.37833818464859703,
 0.364450560176668,0.7454267741228938,0.8703516647376628,0.3777410962801445,
 0.5160208395795817,0.9376018616512274,0.4935257287790845,0.577489289374513,
 0.5870920506515901,0.03128943633695225,0.5091803784526518,0.8569875515075536,
 0.7372545933552036,0.1999850233520143,0.9189458825428437,0.617548438378131,
 0.9981264216427221,0.9703598932946138,0.8658962183785788,
 0.012108597709425406,0.2977043756100727,0.14270836876851,0.5430547577275444,
 0.5748502244888168,0.25500468089422423,0.5048137392516143,0.6625361481905923,
 0.6017795000423384,0.18960456173754048,0.4423617374238552,
 0.42084537431568136,0.1923435124470848,0.7174957574822874,
 0.10662613498394391,0.3479170813763055,0.12991853176648016,
 0.7555903102414124,0.6364573414127753,0.1041790805080508,0.8049047561928021,
 0.34025617113841955,0.5509523368393516,0.2110178652574458,
 0.16778541719854023,0.41776789961143646,0.5635340624266758,
 0.7582959529281186,0.8746985649673332,0.45923743765795233,0.6742145461461578,
 0.9041857458349027,0.9815014023699323,0.9243893128091956,0.9598590585112803,
 0.17994432911655067,0.5857415160907014,0.6291912349103613,0.2708023418094172,
 0.6924405138171428,0.009344140812180046,0.2533319972577702,
 0.8704964899854447,0.4880998688399615,0.29905702918453136,0.1625740957295566,
 0.8999114362085815,0.6143152247732179,0.3677216917691485,0.06323943922643815,
 0.3935391350826972,0.617444878672488,0.699770731306575,0.3146539632760742,
 0.11722229269147744,0.6770551516724163,0.6926090794897747,0.866175893764966,
 0.8113968381729872,0.9480008421784026,0.6231920566525259,0.445309005417733,
 0.030879433115803323,0.23651545200424529,0.01653558027851698,
 0.805833502339889,0.9022120617481302,0.8595996038554338,0.6581221485098339,
 0.40652727359073626,0.9017327982424879,0.2118772707208082,
 0.08413704773772879,0.3776063747657379,0.580851017062387,0.12778713347519377,
 0.3979747083261216,0.03816555421128354,0.5923888152603322,0.6148276770360189,
 0.8377444158514078,0.8931880506523939,0.0793771932892966,0.15739446136157842,
 0.09977111443970466,0.5883121341100787,0.2913126370841326,0.5223186054974495,
 0.5285944116961773,0.20414861010295426,0.5592865425352563,0.6028381252576699,
 0.7439835404087092,0.29814986170064195,0.7487190353208993,
 0.11232687749751069,0.4986345869140578,0.702841781496381,0.6004889641160326,
 0.2804382892436351,0.8627279864188823,0.8522067997323854,0.5923447469357525,
 0.39912504278751637,0.32730889996891854,0.16333244880786002,
 0.3175074773167642,0.9960593668756619,0.2856875902787025,0.45968868888983483,
 0.8875149308136059,0.12621479436902427,0.7274702116383338,
 0.37157990731191703,0.6986952507488978,0.011208210593890433,
 0.4105448079833156,0.5085680400138981,0.3947902018668161,0.29273420689387086,
 0.03634262403450117,0.546463923674595,0.06346056706849479,0.9629091694833606,
 0.6003950790415302,0.8319596828009045,0.5713499724890336,0.23048014582512577,
 0.2517909928412899,0.0032718814995955636,0.45299645872185446,
 0.11389467848419721,0.5527550379768549,0.2346655081782817,0.2370120144608311,
 0.6453477778607919,0.7411685234959049,0.4197683086238768,0.07861593906589504,
 0.3902390385637623,0.6618772421748967,0.49018128564151153,
 0.30413655755319957,0.7598599126889506,0.8320045551327824,0.9300041955810162,
 0.34722676125899643,0.24568807654179337,0.5126055418926334,
 0.5646260583557817,0.6583815372057766,0.39141652427921236,0.61507770283087,
 0.9041143293056048,0.08604717262803552,0.5138518630943945,0.8477461113597295,
 0.9807735332014128,0.3095122376332242,0.3742092232726638,0.3116985194784414,
 0.8159875220900936,0.2659894973631023,0.10034213506800027,
 0.46097926899698416,0.5676353647338631,0.545357391410787,0.11317812608279032,
 0.5354970038603908,0.17298967696946532,0.19240884944224268,
 0.33220629487123965,0.5764376715410627,0.42274666502705915,
 0.12835980806054048,0.27969083492656466,0.5662600273516837,
 0.2474819515623785,0.4104716391220371,0.33533191457362144,0.9551334168496177,
 0.5986383770272037,0.8873258984038326,0.434783782872154,0.5245904378647072,
 0.3512126378288285,0.07927685077707414,0.9349657646109761,0.7707503579106154,
 0.6535886177669077,0.6494663664440041,0.6221935322690271,0.5382473133991694,
 0.5741698533582822,0.9285949216848697,0.3422685979296922,0.4804661410398825,
 0.19033148252374965,0.2633690303055388,0.33364680226779897,
 0.6593776617743566,0.7306090196820642,0.5335669627501014,0.09014034993379583,
 0.9871163271592286,0.8781233853060776,0.10912068901161576,0.5380164986780591,
 0.01831501814583847,0.8731807028094779,0.1123742511317044,
 0.23834224630363599,0.2718671907015189,0.27314362861346164,
 0.5214226285173352,0.42227582203677916,0.10879479057161145,
 0.2513469813968706,0.3595704850072172,0.9607859978335741,0.35159879815340345,
 0.728338504711859,0.4402112042915517,0.22531417775284845,0.7368801036276073,
 0.1548850106485702,0.8084254003895797,0.16081796485387645,0.6275276119937516,
 0.1375914902445583,0.16576194523406984,0.15162830780497083,
 0.4683660297376846,0.3580243515882191,0.1584401685368032,0.18840438496434253,
 0.26651586821759266,0.3087072476054542,0.4746755202154621,0.1172400545947685,
 0.6337231165126227,0.03544683067028542,0.9613597825456861,0.641592991166093,
 0.2547937400082214,0.4931549797268664,0.6057983517787606,0.7350752579544041,
 0.45549827724423997,0.8316855646491702,0.5073862208438941,0.4673445256745168,
 0.9001355199659369,0.208033393832455,0.8214140050120223,0.7675031104625398,
 0.9303476923100106,0.3984834333005519,0.7981298698409892,0.49187801755951077,
 0.941871235146345,0.5612221175718249,0.6394131103717171,0.03887139333383549,
 0.29858289480910805,0.3857086973774787,0.3149161342967838,
 0.015491790868439437,0.3443732428987427,0.7215665983270083,
 0.8017440046562401,0.5403692073111522,0.5266460178509266,0.7385170240113306,
 0.661496466152719,0.7440671472541851,0.8973631533763056,0.7855452686426954,
 0.8832283370541474,0.7021085000789331,0.33175410683492834,0.5361635934344888,
 0.7244153175292517,0.2563011277547338,0.6048989670869843,0.5847765622696594,
 0.4413042921942303,0.4194297217391224,0.774086264484412,0.8324824714266127,
 0.15604152068070343,0.6217963239743975,0.20728074334602198,
 0.3896868381629305,0.07265798334112283,0.5433548683715753,0.6054114421537181,
 0.18059315195628345,0.6042105080163916,0.03726844627389636,
 0.6734431039745573,0.9914141869904121,0.8323682072737066,0.7144868073207246,
 0.2221755348983646,0.45783898981703075,0.31753226148038005,0.636624004792733,
 0.27804863677911096,0.5086398057840826,0.2110398972536185,0.5300335906137104,
 0.6554356398754886,0.9539882100230869,0.8079524021521413,0.47368470029858534,
 0.6069035445135933,0.08353671487980907,0.4079317500206866,0.6751673524873589,
 0.7137009151993793,0.255559536144232,0.9077120273742116,0.22718348933223587,
 0.6824208018108495,0.05387719851280626,0.7213921632616682,0.9990639956658849,
 0.17649978861381954,0.12966645367355378,0.005036053471292101,
 0.7118000855133879,0.9726806787372718,0.4451819343779114,0.13796253685531368,
 0.40800185925218135,0.03886215018634864,0.1947164928269074,
 0.30564449450140607,0.925678288261326,0.3685371587000722,0.9320654059900121,
 0.2669339541478264,0.06300550671428062,0.4147926408113618,0.8231364964516544,
 0.1241187254342222,0.7249293196597649,0.399020590696475,0.07424636263496442,
 0.7598247133516443,0.20760665926120314,0.8471774451175306,0.6906381645999644,
 0.6769038397650325,0.4791693079589664,0.4314654740783709,0.9822082437549725,
 0.5881394724858852,0.9051032648437576,0.8969871927921237,0.6621518035592,
 0.4521407773490329,0.14579656730702673,0.22986876602131467,
 0.08785915667515176,0.32718495116275736,0.8683390077299457,
 0.6363877077120945,0.1660984604404322,0.12971390103889613,
 0.14028143949060334,0.937890959096248,0.2820800094118612,0.7678738239180657,
 0.3542584204498137,0.07010236219890564,0.6797364917133668,0.6322319164334774,
 0.8817351092016064,0.08515569459561556,0.43521579141090483,
 0.10970557024763383,0.9510865935563273,0.9504634969753364,
 0.28252907149479944,0.32472061305256716,0.1390919056751827,
 0.5057288878028985,0.05787396515646659,0.19757740055920014,0.532363844287329,
 0.19366919835796215,0.12938001551382095,0.3813536448957622,0.695923812146113,
 0.7311847711507813,0.3983955774305127,0.9400327077212002,0.7612962477863927,
 0.9272815598755217,0.7968342251518431,0.9885544046762572,0.4639533201431828,
 0.5128082638703128,0.7559801350066991,0.3517517795456282,0.3883266550752934,
 0.6103380163626849,0.7439778158944469,0.055403927817535115,
 0.5591229360157979,0.1399502270002444,0.6022924224626445,0.9838585582289595,
 0.9415550456481808,0.1925435781788315,0.9568526815529153,0.6993912124058553,
 0.6869970148400482,0.5161107526871287,0.9521590007009665,0.7579399134573499,
 0.2500928864851567,0.5429524707161477,0.5818640720684413,0.8992820558526665,
 0.2388509511567296,0.7894417677665656,0.4160901322008812,
 0.035210031100130657,0.12380607165995772,0.7366125687490308,
 0.5775396727726838,0.03304090681016181,0.8645887940218646,0.8247746944793135,
 0.3231753277223929,0.3788181642038665,0.6626674854022149,0.6298798077216404,
 0.5309611053568704,0.16503077337340644,0.46343085858620114,
 0.5846872489698723,0.9721270597505943,0.6869086669662268,0.6895985086930415,
 0.8179553931739643,0.9230944419596449,0.5527481486561957,0.8020146841637698,
 0.9992141414105906,0.4398012532868524,0.8259074516354432,0.6444121786736514,
 0.972433832851117,0.9586141417713302,0.35704933947355233,0.8381476770095891,
 0.9136752316691069,0.26297974981351746,0.029600418981165055,
 0.5221998654727265,0.12253174435950498,0.18594235079382082,
 0.12363350811390017,0.049888640609415114,0.09601547454649939,
 0.8733019433139998,0.2601538497964918,0.36256161476797255,0.678006219159701,
 0.8424136626243565,0.932424176965199,0.9563809093706362,0.419554897131339,
 0.606666925980229,0.42133014325126794,0.4527426583380023,0.695562126842391,
 0.18300835536959115,0.4310144944737825,0.04738096578043871,
 0.48730988199049674,0.3162603472746951,0.24450735603168605,
 0.4778067556874186,0.4017303893264008,0.9092827819470649,0.5779839006700704,
 0.36139397677513685,0.4755841496929206,0.627450136188353,0.20630858284014242,
 0.27017929203019553,0.4545690429207284,0.5840779065536201,0.9524346614445053,
 0.1792932637348713,0.8081468285747616,0.5764753639846095,0.26923633792842905,
 0.7362643369107866,0.38303230720734416,0.9608914254615712,
 0.11511380120544734,0.9835454673696433,0.4787536635549736,
 0.23604985927271305,0.34354510471081556,0.47759739234252274,
 0.05642633961972776,0.7834597492641722,0.45409788564052234,
 0.7177008287436476,0.8037963419363428,0.5167329603210787,0.05641202958858482,
 0.9721466602644764,0.5627903799765265,0.9569500013011063,0.13647558256128445,
 0.2769336152151932,0.7784036414809338,0.06009095171611767,0.3763030279098759,
 0.5189497309692483,0.3116212344516577,0.7331208151991883,0.08280177824221058,
 0.7758514248624718,0.3295729519555981,0.3578074947614762,0.6491672489272319,
 0.5357645773732728,0.6318837382547072,0.12908541392486061,0.8005621635858323,
 0.9116526486766321,0.9630066352439873,0.38839620275189557,0.4855326537787462,
 0.713260281958276,0.3264884713272782,0.8486017687094469,0.9219322715650334,
 0.1898648505903694,0.3590584677015555,0.6893869151648973,0.10327328634240696,
 0.7524165001252052,0.9538202274000559,0.05068837980604102,
 0.32743456635224943,0.0027164490807193298,1.0180266964687767e-4,
 0.9966192132709574,0.8252537007124847,0.07259435626615773,0.7054255126358047,
 0.004501525646977922,0.897971115520748,0.007205689715868213,
 0.937238191128503,0.11862094714819271,0.21381255719845405,
 0.017103587023445388,0.8543030672644135,0.8837743152462666,
 0.7768011046540538,0.6412330708995225,0.988559014214275,0.3790581167004151,
 0.4115446747184046,0.18792902164825598,0.6861873531602873,0.9429096887367445,
 0.003033642945860171,0.008689490071018047,0.3950459193871,0.9032876509318972,
 0.1629381352188487,0.5346261989679798,0.6506877173032528,0.3168567510992273,
 0.7536699401040494,0.5889722910174308,0.9986861735403217,0.34701522312823185,
 0.6486809336514235,0.006638024834480394,0.7139962686943959,
 0.12461856888660972,0.4806992940079068,0.3737567632641021,0.5728991988301466,
 0.8247020872401457,0.8517436398414777,0.5341719483408577,0.8032426595767488,
 0.2859573413408779,0.6658449965838742,0.715792365292043,0.9142235075081278,
 0.7477306653616868,0.5889164100832799,0.9200635297300096,0.1202461059866401,
 0.14367264961830928,0.42626536465393206,0.14640966614546902,
 0.6221144712890214,0.18872513567237914,0.8214990803435971,0.6839794730477536,
 0.7390340080245756,0.15043371089381374,0.23797200002762953,
 0.5580800980231876,0.9882762827466088,0.7379657104105972,0.0569224125728649,
 0.7799713152133423,0.6229643828950363,0.15378824598521013,0.789315184527446,
 0.5795973705625703,0.8806989911862173,0.6757498932652949,0.24665372230093785,
 0.15779320288328813,0.26641104049774655,0.009638974484081242,
 0.20299197123506252,0.3737487096591837,0.3132771208002101,0.6417742631266165,
 0.7753939285652707,0.8994363464078928,0.49715967358654123,
 0.22706692557480512,0.8487524505581467,0.6583268502079317,
 0.056296197164863004,0.23499302457317628,0.28994386603367006,
 0.39855375350269284,0.04811671423573893,0.0904091025070679,
 0.8935584387504201,0.2023174420358418,0.23977365117756244,0.8535300598770507,
 0.39597867492088346,0.3436543245209107,0.3815789058706053,0.8442931791759762,
 0.5372423114337503,0.6991256216049249,0.6160674430808732,0.6141477900266946,
 0.7284950500051864,0.5500681757176012,0.7152518014057444,0.47679539814443694,
 0.5898383783614836,0.44693417257348045,0.2807248299309635,
 0.12627479975153932,0.663892728476044,0.5390110016442347,0.5181816683074647,
 0.6972900917350945,0.43557458668638105,0.5813854192446913,0.6592541401124165,
 0.12132281801497047,0.7294111119562511,0.3653063486424366,0.6902316629408797,
 0.4119590364096508,0.17777938217806644,0.4610231228919879,0.3551480474013824,
 0.59316864751438,0.2718198297641633,0.7420812722309889,0.8926261011117274,
 0.9624512588071577,0.5864160746853639,0.7179236389629038,0.27099064141860774,
 0.7639323379338961,0.18752239239686908,0.7582071406709263,0.9908757471891888,
 0.9708486002133278,0.1712067487647324,0.4603997797577746,0.3891974992580992,
 0.6006137203713386,0.8207958500770561,0.9227630420892335,0.1764014377908819,
 0.22979669785801837,0.17445432219112877,0.00250931575914759,
 0.9169184831610124,0.808354456353086,0.8423906958638372,0.06299566351670516,
 0.40944253326281566,0.7057522530028717,0.007366791518502613,
 0.5445368854861695,0.04957879651165564,0.3653600160454955,0.2605402845057069,
 0.7299641821307709,0.6086163792257739,0.8142446801430082,0.7155573113009333,
 0.05877138790143355,0.8120885806626095,0.5902021861318767,
 0.05653654854383938,0.9476664971059952,0.2484606218713603,
 0.18177132654244466,0.35719443025243225,0.8411592720773632,
 0.11226800636832257,0.5539251381659714,0.3973256958795225,
 0.16971278235961662,0.8344101974007399,0.7490075131267602,
 0.28771923623787843,0.6750023428377467,0.12759648792318534,
 0.10739291483862967,0.5945824651098327,0.8463281431208599,0.9244472026254917,
 0.6194074456084926,0.11881882894914964,0.17705125234945607,
 0.2500485007240155,0.33623592508617095,0.3656302187702608,
 0.21825796975275213,0.722176298898824,0.7745005180157569,0.680042357440672,
 0.82752256580517,0.2568335565135299,0.9119082953226613,0.2126428511687959,
 0.04324654801206773,0.9249277201908213,0.537226391934129,0.04775439787903246,
 0.1017510337418388,0.3024347995027683,0.849351860169854,0.09811636120165168,
 0.018235353442881075,0.5103977432675956,0.6481648308845167,0.668388822319577,
 0.6594250964982111,0.8301007566304974,0.060635741982662195,
 0.9825577621171481,0.909480101645286,0.27175682939670187,0.9102492668111198,
 0.1578229034560945,0.8126745454083173,0.6622484201590456,0.49721103896377505,
 0.7754234470377921,0.4884196648699012,0.3613073905864479,0.7712380515656572,
 0.6803407293907162,0.06816472649953698,0.24800966873658803,
 0.6166997355314715,0.4854666875344299,0.21093021968684145,0.988138324077617,
 0.14128235434910574,0.2104958804346364,0.10718109187376035,
 0.7876559609160075,0.5151956758033255,0.2834820323777849,0.7595885541430347,
 0.7531155953256213,0.7243392089601187,0.1227544288139446,0.9332033096493838,
 0.5290330259744979,0.5793494912408241,0.9615458480029245,0.07219238686301199,
 0.2947179581169994,0.016123265563083855,0.8131922563560148,
 0.7836072798770247,0.1839075517623744,0.08781835265632854,
 0.38865606142982423,0.3199061630459801,0.5314389098508209,
 0.0022207331684230835,0.10005269377539017,0.7902241887686233,
 0.5784398976618149,0.7894717465811071,0.4972140958043594,0.6094501948011969,
 0.349420976778129,0.7031033343367327,0.3205439995175372,0.9594202372468278,
 0.6539287768105239,0.4339864736017711,0.8385795544037792,0.7922781243822178,
 0.818393627395007,0.05014630451234614,0.4415509850348094,0.45398694750495283,
 0.47907633374211056,0.8921955713070968,0.564891055775608,0.5565165658861921,
 0.4647052098232606,0.2884507021346463,0.8224511207895608,0.4822562273971511,
 0.8596252995648374,0.5415178759676486,0.46102242434492824,
 0.42495765647244044,0.6187461028422027,0.9159713367121258,0.7042093582793006,
 0.2595366730556984,0.6583238343528186,0.9679174490088742,0.45495726476150855,
 0.935946592416417,0.25626806699351856,0.1657126803074278,0.9367769545887811,
 0.02944473549439519,0.14976017698342214,0.887528372794389,
 0.39923590735889725,0.8375654445995555,0.04483910065817287,
 0.33371187232242605,0.15965662867975405,0.9189813857235236,
 0.0068059545987128045,0.797575750965881,0.5392546636234241,
 0.8139057217579703,0.5797547547597164,0.929005675956152,0.9663867843269922,
 0.9701667521388124,0.6969127192741235,0.7484902995529462,0.2239255669677871,
 0.5328745002553401,0.15734531353146508,0.4170250952208474,
 0.29980064919555494,0.06119536501867506,0.9699346344041597,
 0.15420155673579483,0.3753458064114613,0.11393405327668671,
 0.057854435385236824,0.8729207578168029,0.4898438491568724,
 0.44466312579313283,0.47193017167117834,0.9017647733200141,
 0.9149408297926774,0.7753181095213794,0.732641038687021,0.7604179904549897,
 0.10748633484242387,0.11352504591500179,0.14405354887661526,
 0.8948115732048899,0.8470570295422056,0.9194336796189786,0.9215560554141169,
 0.9485725158796265,0.9947894328952266,0.9116962917597506,0.8453033905441966,
 0.22967139388458357,0.792951006122819,0.26870154236091015,
 0.04871617291382635,0.46687865078388,0.19310947343701834,0.6084605552285949,
 0.0517113084548807,0.8580850408478491,0.31312985713542474,0.4787806390896152,
 0.20175125185790477,0.19126656566693367,0.8839920798110685,0.220212603562556,
 0.4175825960753001,0.3749839592224169,0.5128489164027428,0.13511960213773644,
 0.013355445630053642,0.019782371949855992,0.3951467594218995,
 0.701732508936105,0.6382898791512155,0.2839402020439721,0.17811171284617222,
 0.6685169101932599,0.7269637172844288,0.777857741564487,0.8614720250341028,
 0.3885229235944948,0.9953377605769984,0.7498298695611347,0.20788387484926918,
 0.6549179687434546,0.10713078835249235,0.6489877654080337,0.3624535705115931,
 0.7943708774201652,0.07339999273753706,0.8237575871210571,
 0.03683381045958667,0.03638289126993843,0.2661300148048411,
 0.9113576039334237,0.8107152518429084,0.5904562969691718,0.30873176081883236,
 0.8912839807281812,0.5786373319035761,0.32347565778688314,
 0.38193823510294944,0.01663417719198612,0.9896035431998353,
 0.5749599553885648,0.6111984125406006,0.7145079724897413,0.3236099877973836,
 0.384944577202331,0.2612571461641451,0.8283213094539498,0.05067684188319965,
 0.6278288051725774,0.97769696161176,0.9677566035548165,0.6207878842031889,
 0.283245882863268,0.3791195720681426,0.6376927365659801,0.0919116401100778,
 0.6864943517443637,0.43729229931499713,0.6068880532107421,0.2916203342444532,
 0.9566750837333344,0.7185012593378223,0.11858612602965801,0.9769288681695083,
 0.2735328878050698,0.4029745132522522,0.9716006605302193,0.20407957520976083,
 0.9291497858976925,0.47087316816952485,0.3844277695718088,
 0.40793073178060957,0.7895233283677102,0.9307894864027193,0.5787832641337586,
 0.5793523340207339,0.7331003231413606,0.02491687743462201,
 0.36893088823179643,0.6256958452047994,0.3432490927501538,
 5.889238688714382e-4,0.5538385743933053,0.24946291961647804,
 0.21464703223853177,0.4658126599251624,0.8155356874642451,0.2720542968587868,
 0.24662272681495023,0.11586504640534634,0.8866970384255435,
 0.7814103008392047,0.43850732717274554,0.30670988796518717,
 0.1884831137014351,0.24914080604177902,0.4057742011060139,0.465212218425059,
 0.6617943782663294,0.8185542664719507,0.632761982023931,0.4547060453629156,
 0.5765101963617598,0.05449692123026062,0.21068932568164622,
 0.6826655335765061,0.90737999900624,0.19667522962269224,0.9302879070323683,
 0.019094726675575102,0.5034649320262516,0.4120682655344323,
 0.8074206407118869,0.3186832581163027,0.38227560826098095,0.4593833660125892,
 0.3168625490413831,0.5967165472617729,0.8352267819158126,0.2728096010820573,
 0.0038459866780959917,0.980370594158521,0.8718582017058469,
 0.5999550184012887,0.7384267237633863,0.15144195167771857,0.0708409266650879,
 0.7844003808183517,0.1071713094619372,0.5706877347621384,0.07885968907674357,
 0.6810290223159118,0.8669551727746434,0.07080270578486447,
 0.03210763827263985,0.17092328060550277,0.1285343813940788,
 0.3436345974784757,0.0992640881094381,0.4908468021228961,0.8653397914003248,
 0.855689007205347,0.6124109218929986,0.8306337482075966,0.7883092021193331,
 0.7191097560171527,0.6167903278672044,0.30135388248708395,0.4674699560192441,
 0.4873825113872413,0.1825914617131421,0.24356822588435767,
 0.21328865304723033,0.46798759818241453,0.13319039160938484,
 0.4789876617817529,0.5661316802410734,0.059589590313706364,
 0.45204976840700417,0.1464491944041395,0.12452057219468038,
 0.8680526908148681,0.03963171128021148,0.5646593252207925,0.7015081165948375,
 0.2676440013964686,0.25216442569947306,0.5340839728035686,
 0.11687742749692986,0.9175802130572706,0.7225829235393966,
 0.17335151409568028,0.6408449970001188,0.3300496565291601,0.7318610727568557,
 0.8917983348208114,0.5733036501266493,0.8016123757289626,0.31528954262728137,
 0.603018344038591,0.8107777423688529,0.6941840856964169,0.40980947823969727,
 0.9462485320529171,0.6607438204249032,0.37732547053908916,0.6730063608822543,
 0.2511056089393173,0.006052793334550555,0.17906838835933492,
 0.7944764066970563,0.9452364655087131,0.19628890288256828,0.7762225843048748,
 0.2908078207365642,0.5130372382303476,0.6798551524507093,0.2580229906932626,
 0.531396211305354,0.2627451840937408,0.3178446413833236,0.20853885047917897,
 0.1138272346315351,0.739004477113113,0.3929849126462517,0.7788601368258298,
 0.022020200563332204,0.2596765407518735,0.6353002395110522,
 0.22684008065321792,0.2909205055052986,0.10428116272347188,
 0.1576452210212983,0.31346414396049826,0.8953018044836151,0.5480975837012088,
 0.9988589204809677,0.4409264634163971,0.6666006787163409,
 0.0011820646699626236,0.9324688953733098,0.44203523840549586,
 0.33359521603018605,0.8386069055880274,0.9681728635548578,0.1987071513839408,
 0.2682990258995863,0.27417026670616496,0.13172949697923875,0.807210212255192,
 0.26550980927663526,0.04667437020762866,0.9496351352802385,
 0.4082381585900068,0.7557226870729838,0.002199606211931493,
 0.6490927972852869,0.13021487272755827,0.8456070306225181,0.64856724081984,
 0.8588234093233262,0.635825006855669,0.8603808915629857,0.6099609688546038,
 0.8631940335658788,0.7598398844051252,0.3456346477230574,0.4222762444121875,
 0.4096614299945577,0.537953747823794,0.9772590306356134,0.7028863608474754,
 0.33723560555138443,0.9990593208785662,0.5893515329669431,0.2406427423728008,
 0.14900131706411612,0.24521019672061461,0.40052034423837657,
 0.2938605065774891,0.47147761211452255,0.8154200296397595,
 0.21437459442931728,0.4664999614634858,0.8712796755272879,0.991081691929564,
 0.5328163495418275,0.5155971620720893,0.2754765927221915,0.7113140234628615,
 0.333282181761701,0.8197733400391778,0.9617161934035683,0.887229871742873,
 0.5721273899101983,0.6377544874290552,0.21089681356295198,0.9597077680194503,
 0.387000645950184,0.18025448947361367,0.29898815045378824,0.5610681471226056,
 0.11551709592567172,0.0031632730194547376,0.9398033539265302,
 0.6137098654346927,0.8477849455961456,0.7447723068272672,0.12081853261592312,
 0.2778121614083152,0.11456593285287242,0.9174099526454362,
 0.43360235001316916,0.28825914746132564,0.7497870980750874,
 0.14075950729042996,0.6947209809195861,0.7312090413617334,0.1778268459288216,
 0.5038974313108644,0.689872500422208,0.05229890774194501,0.09587184390005432,
 0.34509365104583734,0.14090128149475656,0.38445380615245917,
 0.8494764592382626,0.6983841402024917,0.9657587009923135,0.21923921683014858,
 0.08284201147812209,0.6104139105362604,0.9099976605523783,0.6096316518671518,
 0.8469079059174268,0.5419208423804958,0.6966064276598182,0.6480115490002241,
 0.47987933789406245,0.49400528278875,0.47314426401059095,0.2805236587029587,
 0.23009920371437553,0.06640243427225045,0.39846669503875665,
 0.28103502505734734,0.3052591061625981,0.27790949559776745,
 0.18406889896505274,0.3746321403859758,0.08498055856559272,
 0.9500606186751653,0.6130614066672302,0.0863487075763496,0.9047659409795725,
 0.13491158602256392,0.9279289000759112,0.6681818977399443,0.7763274950881404,
 0.1849928259532656,0.18004307957625454,0.09701630496053182,
 0.4307633985269551,0.9657238625325167,0.6104076198980717,0.054949761335641,
 0.3911330803877032,0.9065076572034672,0.6987990677781708,0.49525510039801524,
 0.1330215562825583,0.7207029403713413,0.7389598284622447,0.8425922755046545,
 0.8459113481431432,0.8885057268158713,0.9307044081961435,0.49941156680230026,
 0.6755032977096073,0.990734087281024,0.26158161926385026,0.3324319608051769,
 0.7800751129251507,0.9557585411079963,0.8553576179628276,0.615704861964789,
 0.8180166881786292,0.11706020507240833,0.6410063880672724,0.7847651303306851,
 0.7120396437721394,0.7060322846513079,0.5541095959088691,0.2394405386358932,
 0.9816144595734859,0.060513239701064814,0.8118009488763962,
 0.9332877735606188,0.7342041400353173,0.07165930580902158,
 0.30035621571961146,0.7789014752875413,0.0821882336311226,
 0.13773091105488489,0.10059473221072757,0.5733917339162515,
 0.12794965161339134,0.6112103180759717,0.7448835084977635,0.9598270106574118,
 0.44729791035703004,0.6518126053942109,0.4698543732701992,0.3862132893751534,
 0.5510475001668014,0.8705864215903008,0.48106678174712714,0.4785558318612435,
 0.08814640164383336,0.5159069435177582,0.30497118454495,0.23047258580086927,
 0.5083791781022784,0.7372595529266001,0.43766128048443154,0.7697962591011005,
 0.599103504736735,0.5051439968286908,0.23995833967712832,0.8300768014535556,
 0.30325580385627104,0.9109166738846255,0.21094718212749453,
 0.014718983698782884,0.9872711924991152,0.09645378388508774,
 0.19319997269725997,0.7251661472670988,0.5608183954892438,0.4734157823211437,
 0.21543418946001514,0.6089933550854592,0.931073311246024,0.5031460238990523,
 0.6912252439672573,0.9736991704299894,0.450976097420307,0.47989856057973945,
 0.8153812021501806,0.33689177256359915,0.4392597898257812,
 0.06905020743393442,0.0817412892506526,0.9608377295914637,0.2961122170546636,
 0.5558789318967478,0.620016960943995,0.48944399188523413,0.9544200274404885,
 0.9266909091312875,0.6270801071766016,0.6967547029490534,0.8559405754240518,
 0.186944725571487,0.10893676062569158,0.9390231608834991,0.6456252887903892,
 0.1923110322574607,0.857640473157738,0.764771205527459,0.5905572594237377,
 0.6642047058789295,0.12242745120173437,0.9435271524801089,0.5990771860308312,
 0.35402041056296196,0.20516299824257622,0.7535828648370977,
 0.24371781466734588,0.20583766391620284,0.2335269338480892,
 0.45902974688716536,0.8915171180286883,0.9883145857409763,0.7949951131681222,
 0.9146474187535291,0.9783701772268889,0.25447266009354896,0.3868459446365303,
 0.7888625873973293,0.6793097693798502,0.7551208090202183,0.5253203261641808,
 0.30618781517687876,0.3872653333127051,0.35467628433065856,
 0.45623157300283856,0.7225436601388983,0.19743194571375522,
 0.4542854045534802,0.055330148628671916,0.15978963960916437,
 0.25986333801146777,0.7781356603588159,0.94774726199292,0.29632107399203234,
 0.8413281210194136,0.9396883275868214,0.9219825137792002,0.9602820733988783,
 0.3201541284196944,0.6517198837939195,0.4691539307677661,0.3525785471579135,
 0.3289700627118419,0.946792551111926,0.7406536417972287,0.427642513600883,
 0.01744626346009459,0.401241880359512,0.06731082983717007,0.624794725364572,
 0.8277920087374235,0.30365718200410474,0.5226021793789433,0.3279628428703667,
 0.32239179955203445,0.9089420870545685,0.9885676450941072,
 0.25516763049117797,0.6473045267566443,0.7202706167364954,0.2706733356854918,
 0.47831995751616685,0.35429535327908823,0.894583094378736,0.9062401141067019,
 0.3537372893105537,0.4386711576097696,0.0863093332981939,0.5749671232751844,
 0.368401995225043,0.6871790987624833,0.5906597080344667,0.6844310760185749,
 0.6520581564792106,0.7034561485766171,0.8263796230797816,0.7448021590354348,
 0.09455283833608707,0.6364368492332754,0.9842562334927035,0.3220713486624591,
 0.41369172971893087,0.767632536795315,0.8862001021998003,0.6834847424816916,
 0.04133436018052339,0.26379099086612323,0.08116064034369008,
 0.6687631914121467,0.5057380477646135,0.6196367680963993,
 0.002300058654503667,0.25784034058591887,0.40264086080315664,
 0.8168957202597206,0.6528621141352015,0.08714731640107609,0.5062072500302418,
 0.6572529615392144,0.5632123081140256,0.0108161386327158,
 0.007894311795831777,0.04507650564469878,0.43160715299389885,
 0.5828180890240259,0.9548373663950288,0.5574503396829616,0.5200086812513706,
 0.5121141304039029,0.3839508501927251,0.5974557703694287,
 0.018502538848255368,0.9376016190966953,0.06611365546358483,
 0.5679399469983245,0.7834433698347961,0.2150635598908548,0.3579922300006184,
 0.19151070466702536,0.12452963160027297,0.571455224632537,
 0.029383548778326496,0.995864415522608,0.19617913556124855,
 0.8622394271409547,0.9324784605923608,0.21378374809732892,0.694354351739228,
 0.16361128604115371,0.06120511684862184,0.2621718966853873,
 0.09262604813515296,0.4074299334949514,0.2258491497504702,0.9254334098333631,
 0.5090209037398022,0.16899344866091126,0.7622138461060106,
 0.07262414746580159,0.8074710584988247,0.45009280691942355,
 0.9148893287300814,0.30750839437611055,0.605554564618835,0.07107626001950507,
 0.13138210620438273,0.6976933892490377,0.31107432278638014,
 0.08270587905187565,0.2427664935296766,0.18161284083472984,
 0.18567739220524726,0.44620480787737726,0.23442803913163657,
 0.12065294165702833,0.36086459049774333,0.3001014133520086,
 0.015571006020946077,0.9937865254233803,0.7079510540695164,
 0.3607496668566217,0.25455913314982237,0.6306273342658065,0.8272907667870972,
 0.11350979178808163,0.6420134926572589,0.3533773665755109,
 0.44600550284986085,0.023089261635438896,0.4889553601522527,
 0.544522933135917,0.23859783375644616,0.7124987855077405,0.5432909147896559,
 5.743417062530343e-4,0.7158590267773643,0.957345617731604,0.7586748168566465,
 0.6599667130697269,0.2910580873927642,0.9087343348716862,
 0.034700097343799974,0.5778450371063278,0.646477015236675,
 0.23595376393486123,0.20627875446587274,0.08515947144658276,
 0.34601766417929647,0.18287237524884858,0.4229390237915369,0.987652891166837,
 0.7322077735778256,0.002539295036968925,0.755684087065181,0.2804372459780363,
 0.6381853986102968,0.49557339506490106,0.7356447441481035,0.6673141712267988,
 0.32361429217165627,0.13432993899755985,0.8529290288283718,
 0.38752005382000954,0.8181164978336425,0.8815291817180425,0.5635676192100771,
 0.2807509686345957,0.07014689749862968,0.9810113368211112,0.0592909855759689,
 0.011355490949928715,0.4225940082591776,0.8115764046861602,
 0.028882623377827876,0.3534792744745141,0.45171541414618355,
 0.2135216689456515,0.9777547069187729,0.4816318582337271,0.12571692865877204,
 0.452006491237522,0.9350323194047172,0.8212120409982352,0.6366611740072097,
 0.4789904914288823,0.2804658511217587,0.89856143553731,0.0854535616875054,
 0.412238836724361,0.9479820461444145,0.12377306142818012,0.21431538976441455,
 0.5243594333542071,0.5461915104309009,0.7068887456308484,0.9209187962983025,
 0.04886067808744521,0.7485533987188303,0.4850404436441531,
 0.33774763115553763,0.21464775360638455,0.6259091235146986,
 0.09870167843692679,0.19166945916702738,0.8846607010432542,
 0.7934836847188149,0.5491629680913935,0.19114251862310516,0.813555569598316,
 0.06222463085028784,0.3989844654702295,0.8271878568785751,0.8844654805453964,
 0.7818051259899822,0.2778862605161383,0.5994730303813554,0.776607607304124,
 0.8624559562732954,0.18300281352976455,0.5837063733215082,0.0697743332463745,
 0.21957616585734974,0.14054669620916294,0.5574883586996329,
 0.41174714242170907,0.41812689066057795,0.6755591225054737,
 0.13815918641170555,0.5981166278388661,0.9151717235934161,
 0.14207758871995202,0.26684247128214356,0.5175283122435572,
 0.0758218300377933,0.20456522608340066,0.886135262940137,0.16753551129804112,
 0.9484469299813607,0.21202534050338284,0.8832721875764458,0.7956273579260384,
 0.8298617065977039,0.49028397188731665,0.9147555589254094,0.701494557526803,
 0.6409453217511638,0.8160185231103821,0.7343078627442823,0.8192613452331557,
 0.8847131348925565,0.1913155941070701,0.9678060167751581,0.7036906141968203,
 0.3870197987018362,0.026763918031273204,0.4770463734018414,
 0.19146516129026414,0.5728007526818664,0.5691749215466306,0.5429545716667521,
 0.6461937877660227,0.9291990010112215,0.6922475432890057,0.26867475158891807,
 0.7293859298933949,0.3853056771214657,0.7937451063301264,0.5395374512440816,
 0.9080244974963219,0.3531856055210907,0.6058975729800682,0.4568738249514869,
 0.2787865509305514,0.4359363084558525,0.6321928080427078,0.29956625332062525,
 0.2453336354893375,0.5296943783457935,0.8363143014718517,0.38656117082485764,
 0.4092433635047591,0.14937595633899658,0.8894567467316224,0.7887952281276209,
 0.593671430442589,0.09905847266168522,0.15736715199589224,0.9942276172237794,
 0.4536618005303279,0.5559541743783333,0.3534716120931638,0.9469472417809538,
 0.06282928711294178,0.24617603120631248,0.10711870531879086,
 0.010793611475626075,0.8562670854322394,0.6916840482559335,0.328204735770919,
 0.0717591185673111,0.3658138718803222,0.844637391071428,0.8869439272457577,
 0.9870906118709565,0.6194916098106241,0.4883859892507033,0.08892936540220886,
 0.6535822204063019,0.22683230832654022,0.9832558081589442,
 0.004358024728793541,0.7852649769948516,0.31159982137546105,
 0.6223852365706238,0.20553180578265273,0.9838401220102668,
 0.33002885711293073,0.3187099993857605,0.3371323801081547,
 0.22159929882546214,0.8357698090341474,0.3242548055679587,0.3703742732475188,
 0.4168088552354934,0.9671939566826925,0.9149116136939888,
 0.019916273479348878,0.288377512637922,0.06965369223988582,
 0.1799945742606882,0.7092881886334697,0.40087856392481647,0.6164176308630157,
 0.05402196160348516,0.4912048889701588,0.6133902382681486,
 0.35686758886755077,0.6981657890416008,0.8371874352836914,
 0.34164158300538594,0.877260464737062,0.7907545209962125,0.16382725522780572,
 0.9955767725669501,0.1355790617481819,0.0994457190731276,0.48892026528245147,
 0.3011598953117074,0.6080438354770883,0.424810154227361,0.9462042014903552,
 0.6043866876421095,0.9007059856607742,0.5167649363019725,0.9684736844636702,
 0.985456839357553,0.7142417417440324,0.7444323865572489,0.2449372086308187,
 0.48215224915808086,0.9672574445989782,0.5304964128077153,0.7518252769238563,
 0.6746454549443744,0.7287118657853474,0.04359079444385294,0.3807812979635644,
 0.6225143522655155,0.697740918722181,0.8110407389706118,0.6672236884258971,
 0.9132684249189134,0.007230561314335082,0.325832229578196,
 0.11670455453384976,0.006923422098592424,0.19596953044630216,
 0.2665880124584854,0.38396492600557763,0.5167580119211608,0.2471230246260222,
 0.6024597247065302,0.3113179970745721,0.21476994712483788,0.6822391337463587,
 0.7694045807404417,0.47687872384112207,0.27983990608068166,
 0.2023790028509066,0.5091848726240011,0.49743089413806396,0.6288614742880764,
 0.9884731639846408,0.03725677124967275,0.04446076032305857,
 0.40113791644503793,0.7808954708010022,0.03899472391380221,0.774091475463445,
 0.5047128161184316,0.9861761873021974,0.23324687828120616,
 0.24304570249385193,0.008301610799423376,0.6804012319776834,
 0.11148974200825201,0.03740368798254645,0.9920538341713552,
 0.8967869784520541,0.721075824286564,0.8271732746285025,0.8973840134081504,
 0.9893249171920961,0.7782743412886788,0.9213193264089624,0.2194412844353777,
 0.5136346004202417,0.7055724210085421,0.11897171881391977,0.7777182400852817,
 0.31709370130082537,0.26726413137275373,0.9131508787488377,
 0.1908860103616462,0.8845596894213903,0.11144755151328833,
 0.016353818335452686,0.8566782478225017,0.08801191725885116,
 0.35660193816922425,0.5276850654401598,0.41446100341199354,
 0.3019598054835265,0.6038287848026538,0.4857371768371227,0.6518322383714954,
 0.46212881059095523,0.7811325764740696,0.978239518443652,0.3705538844366998,
 0.6925068946538739,0.23615227145426543,0.8778188206126764,
 0.030400870368357924,0.9152538704565756,0.7421027729577275,
 0.04478517972872342,0.8131278686733963,0.1462096115594692,0.9970424718602247,
 0.4055534175662797,0.01812154181318615,0.7856782393242068,0.2806060757071961,
 0.8686776795753258,0.21509115063016326,0.017051834275160105,
 0.2790660861204608,0.8660021255968685,0.18021006671841888,
 0.30413053763059805,0.18117722722151575,0.16896152930474995,
 0.9730926765607549,0.10024476224831691,0.3592316742404975,
 0.35111493304612207,0.38520553556657056,0.17819313192333952,
 0.560603442887729,0.7450869014145178,0.6522000294172237,0.9342861206361583,
 0.27079724725067356,0.15315058797527858,0.323225932702054,
 0.07116482133832647,0.8184896344182186,0.9186164740879538,
 0.25600176216732873,0.0037764721182161898,0.894535096736254,
 0.2546820100774143,0.7611734054294477,0.08864838711446144,0.7154052607653483,
 0.009909805620283763,0.3013197370748033,0.8085209547344167,
 0.002945560814908821,0.3282244980867868,0.9146274074831815,
 0.2875122433866295,0.38803831769674835,0.37187699284692877,
 0.9226545922648031,0.23248271426196754,0.46511758100362643,
 0.8618986549940534,0.7658836039344697,0.331218845728373,0.9827530142203345,
 0.8877424831692307,0.9599440500940647,0.9938704277552558,0.9822350010107849,
 0.12984691470271603,0.7041444385818133,0.033715626870940985,
 0.2628695122038859,0.19529178875609698,0.8533492186471432,0.7099612712831073,
 0.6378216911376293,0.5110115843334728,0.6980707874243985,0.20384775761556195,
 0.2426571728820045,0.028416736196958614,0.6804164084468948,
 0.7770191508511697,0.4873182714409582,0.3626597863693124,0.6296587471804286,
 0.6372714595780137,0.9233479957079569,0.8479610425331017,0.1621793675046338,
 0.3347672340513397,0.28005163515778686,0.25432635860016195,
 0.6465693217521992,0.38669063157303707,0.4361021816381687,0.9552702029447835,
 0.761082501685688,0.14873857220273867,0.9129944684103538,0.4436486742446666,
 0.19073734259723707,0.06765859738670088,0.593935774935207,
 0.42648121320206955,0.361220813386642,0.4844590662800339,0.8420842664998627,
 0.22608319558199297,0.09321759127842211,0.5892379551361323,
 0.7517262533442782,0.9050146477515394,0.441185292287433,0.5831420031792035,
 0.351319783696629,0.40428760891887006,0.7136672683761098,0.7816859071225106,
 0.08508951539362863,0.34230224752933724,0.15360066921040283,
 0.5885415824394036,0.9606484902884825,0.4096366007657072,0.33921913442076346,
 0.9422649833085537,0.7203627662968975,0.18135936983501755,
 0.09913386854506445,0.31737344808191903,0.5796449971391735,
 0.7113847518573067,0.8400820949371662,0.43219105227633503,0.5695193780941653,
 0.5087107279890009,0.0018408227871317129,0.19894850778690742,
 0.1125927289765637,0.07487600740100442,0.15557292485234742,0.910228476824655,
 0.8260327322994118,0.7811857332978802,0.509050187462599,0.8405968478953857,
 0.7802469212293215,0.9634956518239286,0.6190295770681433,0.6744763325001677,
 0.7745609555813167,0.8076177950833516,0.06624113929962139,0.3307520139924387,
 0.2378075886213019,0.13519001349870374,0.29370650790501496,
 0.7130251756368376,0.701318948745593,0.7181662652478402,0.7729453640151842,
 0.5690654944525091,0.7592375725233816,0.6898123849192164,0.6197586208811162,
 0.374286260518899,0.6417907947542635,0.7498830486772605,0.8982022394188768,
 0.08440002319169226,0.5684425512866074,0.21675128367394503,
 0.03599196355010881,0.3066995728373241,0.5246817053319681,
 0.16616770071473252,0.2187028313065379,0.6007018376933536,0.4810240826251484,
 0.9306363411795707,0.6289790672881067,0.6812548456378699,0.06451474634395893,
 0.3703071890849128,0.4172435427583244,0.39878739706990707,0.7078125921351749,
 0.23775315986043155,0.774738505917018,0.9611687688658496,0.766497775114651,
 0.09386971822695245,0.637302581253204,0.3096924479552656,0.05835893261513059,
 0.5091160977929117,0.7986610708421246,0.07780968698066837,0.5607863886424697,
 0.45684151981635557,0.40575425185473,0.9662656483572469,0.4920619771672883,
 0.7349031461476623,0.9371576541422879,0.8322183242056076,0.2701096035592787,
 0.5878013316535683,0.24155392052199265,0.3089673120611095,0.9510400252597604,
 0.4886129717746672,0.23230960485013796,0.1775383602314664,0.4450512371711943,
 0.14770970590390775,0.7343022708294348,0.3636747778757461,
 0.47878435106463835,0.26817615931499705,0.2853922600566208,
 0.9379790758212633,0.7857648668635604,0.856235278823323,0.38348064366087653,
 0.05155141766936633,0.19706481880302573,0.4547489219927685,
 0.8242077636136572,0.8099328631091463,0.44185241360159544,0.3222651136017335,
 0.7554720032808792,0.3414945198906931,0.7352683979737891,0.9561441611893061,
 0.7497695794181313,0.974427136909121,0.7246895285768844,0.29976821933643716,
 0.21777352613368206,0.8665417941809707,0.8393539649038666,0.5356881471024765,
 0.3826963495878215,0.18267954539940978,0.8304734867241981,0.7624660850071027,
 0.5431436072876853,0.7604526184474746,0.6821807549823301,0.833965647042655,
 0.3737631382839396,0.991296131729801,0.010381875691512166,0.6852195688969291,
 0.9911454765294248,0.20258105652418823,0.16584311538568153,
 0.9587353553842637,0.9404852748883612,0.5698061966767072,0.904284783901131,
 0.10920398772010365,0.8072715087690853,0.005443149109552681,
 0.5425675509794998,0.8239268061303587,0.13291319309824268,0.8731719520789346,
 0.4504467321155886,0.8861488139566173,0.15250168910163098,0.9283875981565062,
 0.9433361241816866,0.5615237301337423,0.1122274343071149,0.7617896700772282,
 0.18363624935314282,0.7056814169429532,0.4434122103053012,0.4747760104313645,
 0.17599668991616946,0.3858996675193689,0.9151990404348537,0.727317284961092,
 0.08085161917685135,0.039727954418523836,0.8834621231285775,
 0.06191046176323356,0.9179667798387705,0.5723006281613433,0.2548115414334966,
 0.2839398797752615,0.43098542080151936,0.7347592111939701,0.5510941739251056,
 0.6705944619787774,0.7339292467719118,0.8691269064404925,0.5279665094053416,
 0.44169466525624446,0.27264347126103017,0.6071157818918245,
 0.07232578308437976,0.30918681391752656,0.46389569429114075,
 0.43201894346307157,0.020831011360279095,0.8235695716083206,
 0.3945412708541446,0.757057167095135,0.5412787795377202,0.740490300913706,
 0.7980143523977331,0.6510938128237588,0.9190378242098265,0.6531640548486717,
 0.1738272783609539,0.13992497805410364,0.3734872984144644,0.494105226177897,
 0.5255966885659897,0.01955619289535493,0.6226964886410308,0.7110193807496616,
 0.284398161104119,0.3813815050307924,0.017511975553134773,0.3532746606995286,
 0.24457850988753815,0.07709890310732548,0.5853141996501771,
 0.8708969469082314,0.2674945535682579,0.9757343285230736,0.2702939486159923,
 0.24399883250856214,0.3226487073904917,0.9724014281199722,0.798758129307492,
 0.8552057743919311,0.6278952023758078,0.5709887984325492,0.7236208486283966,
 0.5968784680950625,0.2485906697258503,0.3799649150633526,0.2782901965446184,
 0.816566026464693,0.6658317516165143,0.4242436318530186,0.8102528124563684,
 0.9237054319773544,0.09180288587115226,0.2006319044937417,0.4222867714405296,
 0.48238032700501154,0.6537781401268518,0.5356523663175712,
 0.07155171904048552,0.053597286116475296,0.3515646584410117,
 0.6097886003391977,0.2097755007954457,0.6235530878399288,0.5923180022940786,
 0.6107987811218956,0.10537337539507807,0.6569223355560833,0.544783500576256,
 0.1014972090051669,0.5400919834685224,0.8657437681934528,0.8536750259172075,
 0.20054907582897374,0.10880479988193092,0.278831290426238,
 0.10173190939283139,0.8559309487779392,0.5274918868363767,0.9613552434885018,
 0.6644622741088279,0.9415406433522133,0.725595384075863,0.07634281773165674,
 0.575391637621477,0.4569847692962994,0.26556283536725966,0.47133172377526955,
 0.273706687633295,0.2082998120736812,0.018735663116024148,0.8333649290302052,
 0.34524515802052913,0.7764345109209874,0.5367228290543047,0.982929166000516,
 0.9964436321199979,0.2233511799189436,0.24835479310737918,0.7207826200617722,
 0.6691728572024037,0.11863373574496339,0.8396884847055004,
 0.32607019680233795,0.18399393862232682,0.49899614663090097,
 0.3658357234957823,0.5125630736964929,0.5039187118320286,0.4393247336590196,
 0.4223374674167164,0.11593354181202742,0.2886426491441836,
 0.03175526396217088,0.6270923636332655,0.2670057609601745,0.6814570476953274,
 0.9262341317329542,0.9036395373495998,0.7197500639325555,0.08243336631833531,
 0.5770406774352317,0.3233382862789985,0.752565866188279,0.8688807001100732,
 0.32509465487526046,0.3807622033596737,0.5246154796022107,0.9239484637752752,
 0.6583936089540036,0.26322182643064684,0.5996631531747222,0.4038344387680698,
 0.4563609615154617,0.04036565010013371,0.7342823439860604,0.5197725856952455,
 0.13775717026529477,0.14525131441671502,0.6443017432665814,
 0.17473873679958174,0.686091624339431,0.3592209283491248,0.5812090756866228,
 0.19477128907372854,0.6502463648876031,0.30173814222496,0.7699386053391116,
 0.24544112167844023,0.12242058816387169,0.5105656547740183,
 0.6659998358089998,0.6432832756345079,0.36138387034726627,0.6622707412281748,
 0.1176851111079501,0.5612036091394927,0.22669907634894804,
 0.20265374870137454,0.29982768232617396,0.6716168187025727,
 0.2327516203081642,0.5441551323638759,0.7895883669732193,0.10917383020417182,
 0.37198135415095734,0.4635403494182784,0.783226468044409,0.4714775718659431,
 0.17628351213168691,0.40755794378443266,0.6959039525833324,
 0.20068175088857476,0.39337769436161274,0.16683866515544987,
 0.7274468571330869,0.37366538284591155,0.4860206917830887,0.5982726859760771,
 0.6992709924932252,0.028318966740424445,0.437753579626631,0.8747788435824344,
 0.785791323045378,0.6198403407844238,0.16801302689292585,0.3032344573535567,
 0.5465966959632703,0.4458821462297017,0.6569670758117457,0.3743522194659368,
 0.8040408331854079,0.08703405944200482,0.5783094374282213,0.6021844335603367,
 0.27230634089426253,0.2857081118624213,0.6525643537011763,0.7637765066569959,
 0.45749297558141955,0.9796979708056894,0.30811263279091616,
 0.8269481919613847,0.42829791828494235,0.39185988412603523,
 0.8178296549870299,0.3900304196793507,0.6126952333330373,0.3868740629063019,
 0.9712494373328774,0.5431177906152952,0.37982080920649075,0.7680376393725572,
 0.3068364485688485,0.6223737042943501,0.9382453465649762,0.5852071238003618,
 0.9674677564061707,0.056359952093166465,0.9943250928682777,
 0.33468694039740376,0.1638749262612813,0.8887488922211656,
 0.48814310049773013,0.922833853299259,0.12055845189765757,
 0.03116610369174111,0.15830908478168015,0.39459349472515015,
 0.13718502346970085,0.8977588367513356,0.6216095312701406,
 0.14114200045085679,0.4128118841828531,0.10978615276289094,
 0.8137027652145467,0.1313319769703355,0.010552189198868467,
 0.46551643205328286,0.596909634384709,0.9915298399852441,0.5363800317027777,
 0.6934191137173611,0.8733342755589868,0.03471290439370944,0.8607352688710899,
 0.10270049543984017,0.9940693932555182,0.6885527939420534,0.8215568700896779,
 0.037185790856089085,0.06172157746943874,0.6235710145845322,
 0.7102239899853283,0.2658848334330476,0.04202554805036973,
 0.09076515999594037,0.8927370125294141,0.4044019615360157,0.9168836894803026,
 0.340753871518058,0.44151155166419775,0.07335192343811237,0.6790715567050738,
 0.9793123308271025,0.8135679545772952,0.3488602870293357,0.41411923223525027,
 0.5653586996176435,0.9106962524893315,0.06419515294668277,0.975039542856879,
 0.9944110883746649,0.38703190799823173,0.1486332252426098,
 0.44550073923603883,0.4522268598150381,0.4795924225767716,
 0.17875864126981644,0.48021009847453167,0.149991115000069,0.8941042845314112,
 0.255377494113453,0.7435448305083509,0.7149921500817112,0.9824056800056579,
 0.6118861609012418,0.15180382149919003,0.37648327673942705,0.593493022640197,
 0.08415545376070543,0.005085817569331907,0.11970779558470201,
 0.532020698791474,0.08319318934303777,0.7900631886745776,0.7855847970083335,
 0.7993636856571927,0.2871787834174798,0.5394841526925208,0.7533901955259575,
 0.502273772811632,0.49592807076644074,0.8672423090940543,0.682282021122007,
 0.769160806380669,0.37142840899279517,0.0048127941947953,0.44308407657407045,
 0.44177145074789237,0.2606118502695274,0.26129161119129907,
 0.8328633673124305,0.09840776439870857,0.3547965994247707,0.6078033365138641,
 0.939474889151506,0.058806910169738824,0.23373364674060493,
 0.0945753800901532,0.6048758399047743,0.2007712942436447,0.9306625321230588,
 0.08608315653710252,0.9638223312076785,0.9394199767607813,0.6193466774371584,
 0.6673284544412014,0.8283305672999592,0.058138776811437864,
 0.5255979270705207,0.5313206760462378,0.17986782955009673,0.8263924209915036,
 0.19217423626451757,0.7626223861743548,0.2911905468598529,0.7906628061600001,
 0.8799551283423446,0.3777974592709137,0.5184205809867417,0.410339619271345,
 0.8867278691573549,0.030883538690214207,0.24302315063027924,
 0.0352623729340259,0.42636816447334447,0.9476883834349923,0.756310900513493,
 0.8417898698224239,0.3936942431906063,0.20707659272747236,
 0.37174008551341287,0.9755052958030427,0.20633006554954503,
 0.022324994762141337,0.2163466508094506,0.9111697293592066,
 0.9968395324548412,0.12935808530261306,0.15219016942529073,
 0.03152811265931543,0.5852403654152152,0.1346440290628026,0.6774534081310697,
 0.6928879599713529,0.17774802764564768,0.7273365224200246,0.7366567747308541,
 0.016638619648428388,0.9833277386207242,0.9353168877986208,
 0.8103879426891629,0.9029827560089329,0.5122271285836656,0.9383608813013155,
 0.7291823549817658,0.031940390364507465,0.1446912290307384,
 0.15241281953797425,0.07499720809174915,0.7893230508141033,
 0.1638976478218782,0.06139205241000634,0.07406158602571722,
 0.6614041109792862,0.349749376222394,0.8185267702546027,0.9273461437320398,
 0.40826230307297995,0.18166583924650748,0.47197881906680084,
 0.9834388934400244,0.13731258331228524,0.9548145820652896,0.6640208371570554,
 0.5950267413378677,0.7144026677400528,0.5059777553879998,0.3584006556180568,
 0.17556128106456814,0.8390534178387518,0.7079386849395414,0.5271554045311461,
 0.91372580693896,0.6042146063909238,0.5790527466856902,0.761109793013,
 0.2525550341883561,0.7354224026811355,0.7435766142071071,0.3580882690583479,
 0.7760474322518526,0.5487105983754523,0.9997137732989411,0.47868234793947084,
 0.30334036627477656,0.5711649819108462,0.7150089893054055,
 0.13223558741442154,0.07535282190947756,0.9625152719250827,
 0.6613327614580752,0.8050090333702652,0.14148331614741838,0.9665382148103538,
 0.7212052265190108,0.430718621704714,0.11149238272680573,0.8960529112305738,
 0.020456128013987773,0.1670205133396323,0.6150494529359295,
 0.2666351994550912,0.9123994309437192,0.9515767475431921,0.1069595142296218,
 0.3556632386766413,0.059472057506735454,0.20062077861226124,
 0.9936902184821083,0.30871027750225566,0.8977959585135308,
 0.15452923650443462,0.825471422169663,0.3243498507671495,
 0.032825840046559196,0.6673251358581006,0.2614880082710983,
 0.0025230941762353076,0.09625648656237673,0.06971066307541729,
 0.19031985711090849,0.9027960437319554,0.590868613021204,0.23052901021261507,
 0.5724004261264588,0.024095030230854242,0.9590876428932663,
 0.28031917423100905,0.056936978161187146,0.31757545238855445,
 0.14897769585230614,0.21690926812346512,0.04143973115024968,
 0.41335265364055407,0.2568004146023374,0.28555875591314717,
 0.7271917489668451,0.5139020082424401,0.40129069496092507,0.7930889639721763,
 0.318027766774433,0.49156287503264584,0.712473020047795,0.7790975220435966,
 0.9778340675200232,0.4604828914685801,0.735948572342137,0.19805131686642274,
 0.3340479171704638,0.3074347965569816,0.10840081052002026,
 0.48442918864953644,0.4626756897327242,0.8735474394908257,
 0.44928832127997487,0.9130354178296459,0.26195456712446785,
 0.29502021111810095,0.47732126469929137,0.9077524304121911,
 0.8797253192868804,0.545131966988484,0.37644708970517304,0.6144585000597347,
 0.784089030090201,0.05739628846130529,0.24771245818761278,0.6351526296790008,
 0.9336995849265726,0.9398848309927701,0.34408202500083085,
 0.07446401881058595,0.9906959140552422,0.848396475809197,0.9703291800018089,
 0.9032721395931339,0.630884997233629,0.20586971616167715,0.5513486608829631,
 0.5658726238151683,0.2683274843617345,0.7733944907506943,0.7216573581981146,
 0.6405856458349718,0.10265545382263053,0.296914670727325,0.5708039487136481,
 0.71484297085856,0.1521282350263391,0.26620556311043175,0.02258040612380363,
 0.6635604811941338,0.8316979584377994,0.09718086594022379,0.1065198261986896,
 0.9648303285894453,0.36079619811083063,0.6744573628649055,0.6635646064205438,
 0.2211062744233221,0.5089102957690369,0.9105117220070691,0.9630485372140287,
 0.10616333745712103,0.9332702854912006,0.2663567276892763,0.8633170545113433,
 0.01609255429473322,0.05203907278761355,0.35238606338880174,
 0.3792134848875638,0.584464421233111,0.4877480474822026,0.13337202227487788,
 0.37543663797487203,0.5332250884978099,0.31855730562490236,
 0.19462958601386493,0.20872165300077083,0.9180442828032264,
 0.7630229969427041,0.05384371547661493,0.9087076208786515,
 0.17590934360145472,0.8335881854638942,0.5476801727165802,0.9489417315376482,
 0.14493756425490667,0.4835238554533122,0.46848371560922275,
 0.9808149480002782,0.3700670573181901,0.23254737990812369,0.846762734834926,
 0.5156575493497539,0.06463090843671637,0.02545054434807803,
 0.4034303049281508,0.354386688172712,0.3329835089474874,0.6246041336688787,
 0.04124206123531138,0.8179577109735249,0.3872783701250919,0.7655100656969456,
 0.7593956037578723,0.4856774638923408,0.31157966580609975,0.2478442255618465,
 0.5879103115815593,0.06395094090127673,0.8656443080904779,0.8695999108174901,
 0.9810515211268372,0.24876466446394863,0.7223608199371788,0.9815987376322877,
 0.05432013640666944,0.4745971191214816,0.818421954329305,0.3717255219136595,
 0.04734672991064526,0.5898584603688295,0.7916577194809413,0.5296508915582651,
 0.7090303258925142,0.25356794149629325,0.7013808346812036,
 0.26314265443076523,0.8500760780772193,0.9219614924490358,
 0.38174799871301934,0.006616549358899881,0.5495474284915612,
 0.06900048416075011,0.3820763545269107,0.9250227030455453,0.6197836758378309,
 0.6026880090127145,0.3667074477490152,0.741399347868745,0.3487471972411188,
 0.8750901273098517,0.8324339789650474,0.21373677078180087,0.9429689481474334,
 0.6506424900818906,0.5530026362125356,0.0176828638038935,0.7275117115528218,
 0.6952959189728355,0.43812104603066215,0.14909573662938602,
 0.4481049079944953,0.9018637005293908,0.13647440262099353,0.5000259255145356,
 0.15571717087466097,0.7024304380132393,0.894112194329321,0.8842104611448531,
 0.06814123768751712,0.7274558090398937,0.7502143207033971,0.9537660260517288,
 0.7159036024620087,0.9012959614645748,0.37572010342613094,
 0.11856125794860506,0.07222940349277907,0.44503157626555645,
 0.680535761300771,0.09834392245742829,0.4464356923441245,0.4423900868527466,
 0.3237402764423567,0.06492600238007218,0.22181434476435147,
 0.9764416310109835,0.46850885502160566,0.7429207552325894,0.973662912501049,
 0.6314973330385507,0.7589351298947034,0.0801252191378029,0.5006463254101633,
 0.35778613480172683,0.06754895716819553,0.16265373980058084,
 0.6791486181515445,0.8997780847493965,0.5590819570942462,0.3034589942753778,
 0.9780731209131638,0.7935354384336959,0.9752159769486982,0.14831264782535825,
 0.522563935141927,0.5294622112225258,0.41351080041968613,0.8920487454490913,
 0.7002361651500478,0.2501389227437921,0.05124982641860676,
 0.0027445139614297176,0.9879333963286415,0.5001467490020761,
 0.9045470502983916,0.25479998224902145,0.7277804889450921,
 0.45718649159383973,0.8164718602502271,0.5041759308232783,0.6352294750534,
 0.37202240945265896,0.9142202339735346,0.4270900307300679,0.742722710096795,
 0.14509826508116885,0.5110172039759194,0.3983108404155081,0.7306360487038283,
 0.2436211996863995,0.533746061942251,0.09966918623512844,0.10315431105593798,
 0.19017059162827,0.519102666897563,0.465985759331597,0.6373865978329573,
 0.46704222215097246,0.03605640133828125,0.24040072998884998,
 0.5503959829379346,0.49443912167116044,0.48457188614613134,
 0.9037718909708797,0.6065946677632825,0.7380822243147203,0.8969865654301318,
 0.6001276225363296,0.6792276425897222,0.31682964447130846,0.6982059359032842,
 0.9314623800349563,0.5576241422197726,0.6045550388492784,0.5535975145518662,
 0.28954255077921054,0.5227907782393163,0.8012870012821052,0.2520845680996179,
 0.7328865526863574,0.704218668630934,0.892766860786978,0.5469173534779037,
 0.6392272311723621,0.43525641481113075,0.9084570632070148,0.2478205682395408,
 0.015710302499101436,0.677284415453145,0.4264610414675567,
 0.45870758932862243,0.28995115740654276,0.7264895502000255,
 0.09508059729019414,0.6958804649149147,0.604246398315803,0.16297650426402832,
 0.20022034880334683,0.7697468846028531,0.037387182327757174,
 0.3037530199375357,0.04689515348869144,0.8627102261157523,0.6078363239833278,
 0.018084241748159813,0.6065210838534024,0.6845750496754183,
 0.6265318026986739,0.5118181573903046,0.8746337613465286,0.7156526672815593,
 0.9528081545275227,0.9760491672449443,0.6087662335534928,0.48508284242110633,
 0.9163512716656367,0.9367424873350297,0.9782693060678082,0.8019086023427023,
 0.05433333589193201,0.9629564743320491,0.7167082605708651,0.5589371012872073,
 0.9446397816204489,0.9841890715951643,0.8274265599133689,0.08774083377317854,
 0.5608879924228611,0.16809498487010277,0.3603961628190534,
 0.005084000420886747,0.12453765748965107,0.8097544841127454,
 0.6371330656710581,0.37792224913031125,0.5008682734396207,0.5935098673358281,
 0.1457780053191503,0.5163752310303107,0.23582311419355284,
 0.31335241215546095,0.2833088214489623,0.810151368479975,0.5040720558513735,
 0.16257657044753082,0.6700947005793934,0.4853819768696165,
 0.03845923358446057,0.7574951145474713,0.3237735501839196,0.5114909849362351,
 0.9469030564411146,0.14002721292587272,0.1977742568081885,0.6288388028541673,
 0.507437938194866,0.860254869219277,0.7454026631855855,0.026002382532030355,
 0.5918883154401621,0.931577039909135,0.17163115378888505,0.4085326760955251,
 0.4591048891727716,0.5330893818186754,0.049889027228426475,
 0.5144448185273081,0.0864598530395767,0.7959573163251878,0.8385361682712099,
 0.7854161053707763,0.49811033300921903,0.06362375713647417,
 0.2471653287213541,0.9516875637636096,0.5080451721428705,0.7040533651953724,
 0.21214656352466965,0.4684017482317847,0.6744485765867313,0.8492384910423538,
 0.4551734341215842,0.21731046584472258,0.6244105452409727,0.7101669030876574,
 0.9887848681049579,0.7704189887783598,0.5953923310168331,0.6682972451237014,
 0.6786955890689593,0.7103913159925149,0.9022765681453995,0.930174565594788,
 0.6648804888964515,0.5509120506686835,0.9177548180156103,0.31301924376868373,
 0.6809178798586578,0.19662867449036925,0.08436898494153622,
 0.3106887622827308,0.8135631259890654,0.7333689060451074,0.27333793620021574,
 0.4231971143081623,0.17210525726108417,0.8593225493097273,
 0.04379851627491116,0.14158223273180526,0.20330722112830246,
 0.2838207930829686,0.4523450598278871,0.5814005134828616,0.45311478793469007,
 0.07965101928229168,0.3215501648251371,0.443282526428013,0.5251538291470692,
 0.9660277778864228,0.2068166853813469,0.22257814953169808,0.5203310941528834,
 0.4951987668104265,0.20522528576044397,0.06127624509283236,0.52147318476259,
 0.6704247264893433,0.5398302701926583,0.8338140525621263,0.256715662194938,
 0.9716275649474766,0.7464943957591678,0.5597165134307378,0.5522804617409477,
 0.1029441857947454,0.9498509521254511,0.6402696479098631,0.7453288994217064,
 0.620535333136403,0.072771293542093,0.1841673799712099,0.23434787548426417,
 0.82156326211703,0.6891556277881837,0.6778949959442095,0.03575352425915146,
 0.9286002927422831,0.8438011674016227,0.7506550485443803,0.33223898137501573,
 0.609113631855735,0.8478262338833317,0.9287351795210523,0.7317526076916812,
 0.5514656915578521,0.6350072708926048,0.7832362816111132,0.7286807836530267,
 0.19866871901801153,0.8323331555651445,0.6033720683409747,0.9583754474467296,
 0.9306407923310492,0.7289645905018424,0.36393387805129407,
 0.31189469125808744,0.975354269363465,0.6921206023048075,0.2923060293026365,
 0.7881013057611418,0.06009949490808353,0.968245793492823,0.9963119622028856,
 0.7782620986557308,0.031307988375539386,0.048768514153314735,
 0.017914713984080066,0.770238175014904,0.10664422323817346,
 0.6234627437926691,0.08433670237913371,0.9363786795015719,0.9708832116395578,
 0.40140001616691,0.37694800692258634,0.5988094401150861,0.30312833389252625,
 0.28883340213758923,0.6439714283357961,0.4405118216905288,0.9920052992156512,
 0.022505089598017336,0.2549820162358869,0.8199115285720179,
 0.3739106913038267,0.8277255231307798,0.739471574945977,0.6522269698912757,
 0.743061407362836,0.8991232352635321,0.5344668747718415,0.662935197875872,
 0.338314807931155,0.5657621008145427,0.2309217176398315,0.41077693874115084,
 0.08651110936056983,0.12070622481933457,0.7368310857537403,
 0.3421811093340501,0.9661540085672462,0.12729853261766222,
 0.26567957825645505,0.24327025914126787,0.40643397892922795,
 0.786817277254552,0.5787555848008954,0.8492181141303013,0.46148684820742314,
 0.42663727120533024,0.22184091931420868,0.881275999174243,
 0.25314690600614465,0.8941877392968661,0.5498694420987269,0.4989901404075715,
 0.24554578580183795,0.35391591382578513,0.6676896871961826,
 0.6057137349609955,0.7039721241413291,0.2128701082111537,0.38196124033431533,
 0.14369279597239648,0.379529471828029,0.3442992465112934,0.6167100961863465,
 0.5159194373415191,0.41025762571718294,0.9647062344166237,0.3094382538780871,
 0.19210001526279086,0.49051720101124374,0.8086334144459332,
 0.4911201726399954,0.7758900863856822,0.3594945279312314,0.870144536031199,
 0.8475591330090736,0.25549222720324305,0.49421999267789185,
 0.7797988581112077,0.9700174146513496,0.21571494143489967,
 0.17664983824252145,0.3162167169025951,0.20939306003677594,
 0.7275554992129797,0.12768920183947519,0.843797552806125,0.5127246309583849,
 0.1236306777130255,0.09414297849778663,0.31491100079924983,
 0.7430831248410656,0.7374551061891826,0.4031727262131205,0.5994252904179831,
 0.014518561457379509,0.4391128047587386,0.7287856830804723,
 0.5665717042456723,0.19202359156784,0.7134186429881957,0.6426585515766123,
 0.7827298437181855,0.2229821779281156,0.6900186432186194,0.7970078249957859,
 0.9670447892880605,0.5557030339491795,0.2343316188150708,0.04101058381280731,
 0.7327116698710296,0.007309694667194888,0.978339695918267,0.5398305478365335,
 0.020955693390507846,0.4370578720701135,0.757391200659873,0.484654353834728,
 0.9298482254190299,0.6511082243556198,0.7499735924381852,0.14250804017990548,
 0.3381496945120769,0.8853652212718809,0.3986942100781067,0.8691518079955893,
 0.6828149711566335,0.4493821212628524,0.6233731503737125,0.21843687854692984,
 0.9684590833528791,0.575787576246805,0.6509444928016109,0.43993962161078304,
 0.3471797198491341,0.0504717590446051,0.6609108080483055,0.4201436968976333,
 0.2037137913842284,0.40654345366052125,0.011945871918812023,
 0.7998586799952624,0.4108797240158135,0.9367752133935776,
 0.005876438621004287,0.012632561589587143,0.08855645444659344,
 0.481402876889801,0.07874532283619862,0.09827858761887276,0.6104220862213461,
 0.45123009165767825,0.2758422211196503,0.8361795781198288,0.3454931645198771,
 0.2678410325709033,0.3873531040460384,0.14516836212233153,0.4928778594458947,
 0.18740369086459707,0.08529472934713123,0.6558814823108065,
 0.7824029835632255,0.8044394828319659,0.5663173506986948,0.5023021027522836,
 0.4768952717010888,0.39972955778407315,0.2606489215098061,0.6839634139775405,
 0.8328183586765207,0.31258648919061205,0.35746285722310744,
 0.3922847340782928,0.1166109236891405,0.9831122420666424,0.546615154395715,
 0.803244638801829,0.8470450587254386,0.4073611132076447,0.6054506984428754,
 0.02545160316311379,0.30630600403488917,0.783030751185236,
 0.45976178307429105,0.47437552275197326,0.07587676927958587,
 0.3609564460275895,0.7339964988697951,0.7663335264586543,0.4149901017318245,
 0.8153510057188842,0.18493356288846896,0.26714107018754374,
 0.6591903359679329,0.7842214511389359,0.5344040331152042,0.26782572133363514,
 0.7791815626905731,0.2644725489001961,0.14177140300950586,0.6630234416443085,
 0.28662578703182706,0.579222492484629,0.3120495387716635,0.06728960124382333,
 0.5679467089938034,0.4069950292938813,0.2974648361723715,0.5382938574111729,
 0.3209880136883415,0.9732931780707024,0.7848512795209495,0.5183030052967162,
 0.08809836111780145,0.6693391218947751,0.9640144712741691,0.7742005337039151,
 0.6666684311481379,0.009989207466939765,0.07538642381385885,
 0.5245423519096158,0.9110051817735935,0.39880588132063144,0.7395906364383109,
 0.11523140324323367,0.9121140151748912,0.7761299761269727,
 0.11977377305283021,0.125217369008787,0.701519396499293,0.4964198673751101,
 0.03726156244513856,0.5200460176553874,0.605967092446717,0.21251685909584306,
 0.963646997545691,0.21431748556985786,0.6263137091181672,0.5990638863849118,
 0.979623088823241,0.8304013457806809,0.5367389604689062,0.7251849254473866,
 0.5191022528708443,0.611801873058115,0.13594803625069019,0.07473731023464403,
 0.012206065579654757,0.8059252359094589,0.108510232549335,0.6837775311791636,
 0.768424915525874,0.04709557158915256,0.5800390764078279,0.40062291328283495,
 0.1439778662662332,0.9896827546909339,0.9452054151754684,0.8197326143501124,
 0.9151237987161044,0.710447576209159,0.8822307750064391,0.29010691669104316,
 0.10882921013986824,0.721407046873736,0.001731349616266975,0.855533802032647,
 0.1693998845505279,0.8616135495584442,0.5754309334894414,0.3714538985271927,
 0.022670000202724805,0.048221524278113215,0.8685118235341914,
 0.03948272841436995,0.4815988018431925,0.636642502604708,0.09469912129757507,
 0.6976967321570094,0.6278115921673044,0.3280575330755021,0.3765300313929877,
 0.8322008831263107,0.5079316133674201,0.651290122976411,0.578094940564227,
 0.4636587114123264,0.1991852898357359,0.8463093352137372,0.8210280832799061,
 0.8886055225538623,0.3460241531680275,0.6514131518757168,0.7167456022119628,
 0.7831786067608418,0.7771167928917204,0.7771553221974306,0.2888391274138773,
 0.3775248520170159,0.5647785332169117,0.9764118985013196,0.4131672118610705,
 0.05701277725293008,0.9527499209063482,0.4255655033185767,0.1845869008187273,
 0.8166278932888204,0.305857901194617,0.49675895647031443,0.9172983934716746,
 0.09138684583694179,0.9320940524457647,0.6433168990541314,
 0.020697694854102444,0.38457081927128056,0.22229926246144083,
 0.4529073822091487,0.5826064496282298,0.7321932374385763,0.6963600078487024,
 0.8575444425957709,0.20674756910091274,0.18730847621324354,
 0.39313097431392097,0.6216993463637928,0.6078643135033668,0.9488617984979317,
 0.930398652005134,0.793645081411553,0.680210633476625,0.29919849116313313,
 0.4187510411856511,0.7573166019200537,0.41184132995295963,0.5594352601582707,
 0.07508846017893189,0.47962632582980425,0.5991492379156834,
 0.08726739743478362,0.6964816765403954,0.3859090156222309,0.9598494959784314,
 0.35072824330703334,0.4471097604588872,0.8069278980183159,
 0.33210110286490513,0.9653406770057028,0.05773744246304868,0.984096690297078,
 0.3589951257771359,0.44971471671142904,0.06666438732227076,0.514082932750967,
 0.1418809468350236,0.2974210542428163,0.29598216325774906,0.7956443850352642,
 0.8131132252336456,0.994592491745678,0.9843943694830717,0.889539074651467,
 0.406566350376909,0.7688759172713029,0.6204661559006659,0.9648640159952263,
 0.9458426790563322,0.07727403646770015,0.32784720277704527,
 0.20239286560476888,0.647002520102683,0.5089408604632498,0.8085838044422433,
 0.40577595902594854,0.10621771272040614,0.06830527538965558,
 0.7210100893309948,0.9892480343622824,0.6510520219941646,0.6819108089184075,
 0.47027489925795996,0.13129162275392592,0.1772125377048941,0.804232069606648,
 0.0521853198285559,0.8103813261517167,0.7900900232982565,0.14689199212710313,
 0.6435700806920805,0.673938627275483,0.0783091888545222,0.7329757607887456,
 0.19124554996804843,0.8930861275540928,0.6309519379729273,
 0.30367644755915557,0.2767796657313004,0.5783383229935305,
 0.40979543059261836,0.7774589958628086,0.13662090687167805,
 0.6016800187954806,0.08173085924362011,0.011198125781736845,
 0.6949322614860273,0.7115715190573837,0.4752627730349125,0.03153335493727227,
 0.9819470103671667,0.12456267129892507,0.04072388681282202,
 0.9089534474289029,0.08542262773384846,0.6594100286343295,
 0.27434453923314006,0.5424878899570997,0.09761444144634679,
 0.2600103475260175,0.05733689013481169,0.436913406821662,0.9942899251289739,
 0.3451146038502583,0.7218796703865175,0.7266255085128326,0.3654568301840927,
 0.8239282277484742,0.006105032604715843,0.3623752158440101,
 0.24904531709814481,0.8644101339555043,0.41777943876831247,
 0.8538049704248958,0.9471836281833361,0.8463174052004794,0.24565216292724146,
 0.41894136373285107,0.36324146725743867,0.026490490444851744,
 0.4964717161028279,0.54176012925868,0.3222990163427798,0.4750747275649758,
 0.024360883352139928,0.07356239426663125,0.15327224719893884,
 0.7596672545100803,0.10141352456344299,0.02107452289947731,
 0.21378454887479403,0.8750844601826405,0.7119500729885706,0.4327355329522937,
 0.04235016421683946,0.569538013495767,0.2253384361551054,0.3058210093845989,
 0.06663315667393088,0.5722743991763599,0.7117594135545989,
 0.16987481217812872,0.299876682284806,0.679098705088767,0.11416896647327324,
 0.5005605021493298,0.5163626027230237,0.7433791033792879,0.15762812908090462,
 0.9830774426387299,0.1321717928888475,0.23902086889257523,0.9340508285942558,
 0.03588380689043358,0.49843591132769727,0.8577235521221424,
 0.2148827693527814,0.8097838986420121,0.6185175850023026,0.6722722704298565,
 0.9496033992293866,0.3667648497639173,0.6332029341734521,0.14629411065383224,
 0.720609814196571,0.3881911390501239,0.4901695758943839,0.5960468031860805,
 0.5559904911167691,0.21066282284065307,0.6837206244992742,0.9011516811108389,
 0.3005421888165267,0.6776426306183638,0.11682892358525143,
 0.33275002291021993,0.8122922433174071,0.912845287922393,0.2475858418792034,
 0.9701166837392934,0.08049000860801037,0.6800610942024914,0.9249448739196272,
 0.8729718412228816,0.9256389628275925,0.8824280470448163,0.7850700574118227,
 0.22169032414483902,0.5616797287849586,0.04046698424169792,
 0.3525820787934173,0.019236988452568604,0.85532618280808,0.2815140262837118,
 0.3839591005349404,0.1608961161998732,0.35724738832696445,0.9898784939666532,
 0.3203635195106601,0.2106834693195213,0.8125117015018304,0.3893282113134996,
 0.8623944442466032,0.30097993652046473,0.6290406887527018,0.4534725918509569,
 0.42041032885152474,0.5362024192437228,0.7566243708935491,0.6773926796186509,
 0.43451777686797366,0.34826462976495576,0.37355738601568333,
 0.798819255933245,0.9291701797426943,0.42835299383336,0.533699589675251,
 0.33367686802934204,0.3042898117448286,0.1095506962210342,0.3805901505346875,
 0.16695286116487917,0.7112159624208485,0.3592708112454913,
 0.05714833067755254,0.15805111482031853,0.0375890957582341,
 0.9363860141667361,0.4713719144323696,0.3699848701619327,0.6153101516346949,
 0.12016569389039089,0.5983623694140231,0.44848903384823013,
 0.3564086970111796,0.21273546698206725,0.3723023747404326,0.9908229969914275,
 0.34282580873107715,0.3718909391065637,0.4684985798306375,0.8831815560820256,
 0.45824131729753015,0.7355184315163782,0.22673313544552465,
 0.9236732090452667,0.21985522685054726,0.11251677181856656,
 0.5580630357156483,0.9193265300885773,0.04483453175449914,0.5254325453808769,
 0.6252121227629412,0.29954464097642264,0.6993352871941596,0.866978980490547,
 0.9969168944006072,0.265567397438192,0.33464344292848613,0.98446349388283,
 0.605503088001197,0.25618784990693744,0.9981027403740977,0.7534368821847901,
 0.8197284641900563,0.7690858973796092,0.5866328414681994,0.00900815108526487,
 0.5374619934930313,0.1950096989872927,0.45982934468784276,0.8988007615502933,
 0.46652996691217985,0.7719144559735478,0.7087656900371525,0.5864100611120718,
 0.9115761279374954,0.7105141416484162,0.3053598981502964,0.6861558452864794,
 0.6338950207424476,0.8030693064237022,0.6463344165971728,0.25044718070923144,
 0.32750007923848,0.19258696425238164,0.8472329848630806,0.950065213330929,
 0.2785029272151649,0.4860516913904832,0.11322804726308389,0.9390583962784241,
 0.5819125049164371,0.5183669582459137,0.4223125807394794,0.698216042436897,
 0.27362671935206406,0.20796652662644188,0.2314829913401385,
 0.6939626588842304,0.19214782723207058,0.7197234659040352,0.5960977942316785,
 0.38457536087211397,0.09838045739819878,0.7540486786123743,
 0.1920538711190135,0.2120604552849763,0.5648863874630546,0.23522853845238434,
 0.8270815108803464,0.8229288255963393,0.6697248175109706,0.5654871871932744,
 0.31549502915112926,0.8623876384355896,0.35411340091906707,
 0.45449377270043745,0.9485047884790799,0.1453410895434818,0.7429417262433858,
 0.09290551296866123,0.2123088432782253,0.524151576646262,0.5693146562074027,
 0.4935569543459123,0.379352089007262,0.1257139450599457,0.3513813571170994,
 5.390777644180833e-4,0.14954338665730615,0.20873475979641454,
 0.9727549839982298,0.9617162528913912,0.012658503863856252,
 0.8022386753415198,0.46732402364326553,0.517822709160318,0.15698327066723194,
 0.937383059632165,0.2103910038632607,0.9475648305460371,0.2351525417741508,
 0.4217981119660775,0.2695395050264121,0.6381066388791308,0.04766073304537488,
 0.9521308178617169,0.0653395249130948,0.7732301332442182,0.28302406749524567,
 0.1514623795068113,0.38622942953427464,0.5563113752962542,
 0.49488495413011735,0.9436663083171476,0.9068435841863368,0.1256654891449236,
 0.26029799644500806,0.46318942348248493,0.8369161468803845,
 0.43439360952345607,0.8863518117411606,0.7447565420716048,
 0.38631184795363915,0.9007950324699645,0.18680311631645985,
 0.27494518835237924,0.7372888785077187,0.3624478815665363,
 0.13553653505710384,0.8556524156631794,0.035230641144658015,
 0.6654657956480847,0.7806411198876131,0.0969298781911575,0.28525501459588576,
 0.461146926324845,0.9731825289171655,0.6514549180851525,0.12197464833226623,
 0.5949292559814192,0.0032165011790719067,0.46623275450307844,
 0.09249124231050265,0.5311620304507114,0.6315440458935679,
 0.46177709727616434,0.9086543100608081,0.22196431817767237,
 0.8053063281086419,0.2348713060766915,0.5430622458269728,0.6996545697351761,
 0.027831375111218137,0.5888115319072501,0.3830344797703179,
 0.1898011873539024,0.28000664712211554,0.20638698074895012,
 0.35700862302031644,0.2628396027575095,0.34439597183652704,0.641273047153385,
 0.3623221609049958,0.252562569393743,0.15743085140486934,0.9584807704841677,
 0.17827590187090336,0.5037571148308645,0.788315035606695,0.14903457392788855,
 0.17515803691517995,0.5312855689418916,0.02458516046605752,
 0.7509864355958948,0.19048573739734276,0.9677854583609734,0.7691296024253013,
 0.04629483414418767,0.04446849860452462,0.7990593737908992,
 0.8501620254595805,0.642351735433985,0.6277711051414898,0.4814576503448963,
 0.5226219473882697,0.2884174522077232,0.08616637970124608,
 0.38406347822129694,0.6651694039792734,0.29826981870807645,
 0.5002236132202742,0.24827704422959074,0.3257140472412179,
 0.09421250692701377,0.9269105024343097,0.7217374363681169,0.959783292146014,
 0.9995337093188513,0.5829223569173271,0.4517238754448981,0.8415323360163408,
 0.38288075724988824,0.07832720075069399,0.7414393873719206,
 0.33282795164318113,0.739043210720766,0.20861004097456848,0.2370165804372103,
 0.495607861584752,0.2886002867603118,0.30568209147480707,0.07314976841775045,
 0.626471155196763,0.8495179520125984,0.7189559535151713,0.3895322908496601,
 0.604250997039848,0.7439604452415616,0.9039363071806348,0.4698062618756258,
 0.8630082476701593,0.21530454362702656,0.06245778975927052,
 0.6997517347506497,0.9726487073260053,0.3447338800381299,0.925645385823819,
 0.8363475358935164,0.43608414166449716,0.6297653800784616,0.9295041961952495,
 0.3660973287363881,0.4218806156303979,0.9189002891019707,0.7621129780029303,
 0.3014211613564113,0.7338964980968152,0.008153318083885752,
 0.4769962213203005,0.7681841218480666,0.9641771684890779,0.24179988244131334,
 0.31193496114951214,0.12778769231708564,0.9174108156601951,
 0.7734874569952037,0.6343061137523534,0.43156830535631074,0.401634277687996,
 0.3740038371782144,0.42671735385945586,0.6416575067094086,
 0.08230489148876408,0.6417484743362405,0.9062769146948768,
 0.010958138811274831,0.4272249503190896,0.8386354210263695,
 0.8811080495908781,0.62339021484167,0.36333008137585,0.3488161228221407,
 0.544344916027717,0.6615125296060842,0.025549008631121195,0.8857040247185431,
 0.8108642230646528,0.8884025309030809,0.7616495823284452,0.5559256385424844,
 0.375582805920032,0.15303509841548824,0.6084395465306529,0.6534846305039158,
 0.8917367452364391,0.7085221583561248,0.18720873380916947,0.8454096686854123,
 0.6519436593424155,0.8793746196089637,0.6366321255011369,0.4922683371110431,
 0.5666933801696196,0.0695212312750334,0.9600311808596523,0.573086611010497,
 0.20146543302357256,0.8369554097237624,0.05424582607131656,
 0.03445077556450371,0.6131908174440043,0.8551568041075107,
 0.22151443624916434,0.20378098637812458,0.3642622401751935,
 0.25586152178519017,0.34078065065669083,0.800286308912125,
 0.49123776863194424,0.5199946134584541,0.8806300572186132,
 0.23635628089936545,0.27613091288048475,0.0056806403441010955,
 0.8941738039803657,0.3549168561622156,0.4633187611827385,0.8811558996528503,
 0.318473104907649,0.6447014708775085,0.5652858443266249,0.6392552523181863,
 0.743892130483758,0.7651865056625871,0.7751034143108718,0.21205651334953413,
 0.9775219393191792,0.015760853141717934,0.20040509729787415,
 0.934762746296483,0.5835540857135558,0.7700661305508494,0.7176073538506246,
 0.7860656905843497,0.31418369350420483,0.5010882542619365,0.4002559789062594,
 0.38155030999875617,0.6025112003879881,0.6687625096040953,0.221329360492445,
 0.9523519902764126,0.787051372807685,0.47985014226412637,0.18357823063727374,
 0.9862870180436557,0.8195905024563859,0.7779939686680039,0.8490849183701843,
 0.0405173288908961,0.24785880757521972,0.7656585471590779,0.1906872686865716,
 0.928609322776621,0.46872215517106053,0.26169416125874856,0.9922146204069091,
 0.26566130106074404,0.558048058929282,0.6147462753744044,0.7166711271907465,
 0.15533964586605264,0.05103594711033821,0.23284363262461272,
 0.37580276809679036,0.6455393026528171,0.7479718107522966,0.7875631686075891,
 0.44218655764199244,0.9524609073042786,0.9502734190122619,0.609741237020174,
 0.017391259659012337,0.8655022947047405,0.4278495362510214,
 0.26198791448602965,0.7771518996861517,0.7907804764058124,0.834557918864006,
 0.06993962826843936,0.2799663043524785,0.4504971872708613,
 0.052864034698021034,0.9247399514042874,0.7555491163609485,
 0.8513843493078952,0.4523480302232603,0.1440030438576726,0.16887344404076732,
 0.7998928539607202,0.2729690137549481,0.7478048553104734,0.1472491615690127,
 0.20178748080964215,0.5363318312526529,0.09733221714778773,
 0.5194328818663657,0.0704942684585127,0.27444659017993833,0.8218229408959362,
 0.9480339826650913,0.3361764923469571,0.9367515013423416,0.6404723341102722,
 0.2666066147945464,0.42978961786115055,0.06919162776089904,0.890910483179323,
 0.2415553715810439,0.96051599826408,0.06042362424364278,0.14640627852836652,
 0.028376591354214442,0.5990138376857469,0.4695425745141897,
 0.43890476207307505,0.5756726891798533,0.5162547631280544,0.6213708110300029,
 0.19870755934649686,0.46632862541294096,0.5356448338395535,
 0.7380906301518578,0.8115552818640863,0.8698660369777879,0.1676396964329997,
 0.32958717191823983,0.024257271645534817,0.1240328524232317,
 0.23006054303756185,0.9689629991307201,0.236629614109783,0.6961393349439159,
 0.5232434660672463,0.740554371002156,0.11867752254108777,0.4245959101793705,
 0.534517999657956,0.9136318909242428,0.8416308906204378,0.43510428331453577,
 0.5912163808589372,0.8408364958274935,0.35414781282869645,0.6408008690802124,
 0.9141878697887265,0.8003039932836697,0.4357606887389146,0.5710107548494756,
 0.057667043563187415,0.8327759431947592,0.7465976651452806,
 0.3688989466344712,0.15698188833144577,0.16563092682086822,
 0.6443307587476647,0.4364657731989182,0.802062257266515,0.8907015745776747,
 0.8942306993179795,0.8886008721057806,0.013679081967466367,
 0.8447805583811168,0.6941903891993886,0.5780625615500228,0.34034386249630444,
 0.06250018992427142,0.5845586987941426,0.09914856557219665,
 0.29347066626849994,0.19882308082862055,0.08484352062612044,
 0.9579239631491366,0.3712794368116088,0.18472564816641823,0.1164336033952994,
 0.011844543480193481,0.765248337051521,0.1171037965091204,0.5402610796095413,
 0.5720086482946649,0.8172282452901204,0.06177933980274253,0.4497295597514591,
 0.27074524869882444,0.890833669777826,0.6401446136555019,0.6349029612922685,
 0.19411159478768958,0.3802732704184266,0.04740604561466344,
 0.7675448753272311,0.9935665367675097,0.7515542512162744,0.7772718813785439,
 0.8902557061350179,0.7555531528766194,0.13878759091005677,
 0.29603556772196404,0.8972764114144898,0.07403070637004605,
 0.031430985923338595,0.009578986881547569,0.4900700735949848,
 0.5014699247543866,0.714132211414908,0.22211907690891042,0.06717240991258833,
 0.27074308641219513,0.30498269402313916,0.5094212975863048,
 0.7912262156170636,0.4660556689249613,0.152455080606412,0.16305487990432233,
 0.9438902018776715,0.7134173438789573,0.15762477536361397,0.2106094367105249,
 0.3318603682503727,0.42608320119049425,0.6183269284677493,0.2872297098474844,
 0.16890211514246545,0.5881387993409098,0.6196203643788498,0.5559342050212551,
 0.12988061710874965,0.0543989011391639,0.9397893822421423,0.2203129480881506,
 0.9208475077927282,0.7038268293828541,0.6828953998229994,0.30195786617322695,
 0.01889431312452161,0.5217422562571679,0.48009739569969934,
 0.5477114636573742,0.2755296685267439,0.7307758215746708,0.9236629557691698,
 0.9597780426195155,0.9427594212712285,0.136239754374595,0.1959859620811667,
 0.11231369570668859,0.0454374326175393,0.06722704669397317,
 0.01047834639726053,0.03960357314431584,0.9550677121221433,
 0.7303252129428035,0.9923875341089093,0.17045842103088393,0.6509309267019787,
 0.8007526776657602,0.5873212651403515,0.3760015552930549,0.14986648278323222,
 0.4088877954902638,0.5623947938185306,0.7524011354693796,0.535618867651771,
 0.5587265445661591,0.3406283543228723,0.8621272224335534,0.8490904982164585,
 0.6754486925468749,0.020052065730330693,0.24873854111956606,
 0.7035418707559848,0.26281658388658613,0.31430004257277677,
 0.6392552049667269,0.16369479462599523,0.2276045777897161,0.3897576700502775,
 0.7256265238356535,0.18700057752048238,0.6694348677738333,0.9728494557965135,
 0.5520137091744792,0.35794595912471117,0.912528747240988,0.8729483519997577,
 0.6490944669916339,0.1245903146445233,0.8625239006395449,0.4120483993172823,
 0.9608297706365824,0.5286452695737551,0.4131522826623035,0.7787898037994336,
 0.9179593443249155,0.061591418429227884,0.5788852668557474,
 0.31558032799395586,0.2608838300792773,0.35362068458610485,
 0.16050659146604762,0.10019994272080512,0.8118358569546921,
 0.4375570693917319,0.7741137955640556,0.515925950213639,0.6070519393020398,
 0.6341557443761382,0.9925386942211498,0.3530366499991783,0.9462988983078618,
 0.2703868319214424,0.2239361639078834,0.4526124836585439,0.07123902185336495,
 0.7525318000544304,0.30201553816711235,0.07528353325248138,
 0.10434754249957212,0.14409024520719305,0.3847133336324038,
 0.18281991807364173,0.5789252346887229,0.5609743607284253,0.8574372091465035,
 0.8554581844483765,0.19284327386540145,0.015062254210843506,
 0.46372737938946196,0.8870751111695239,0.18484382701005597,
 0.38505718862017524,0.11858847034847286,0.837851770376154,0.4775218285632278,
 0.9738444194711086,0.5683106150091347,0.1656732729451802,0.3503977109810369,
 0.6769685239854996,0.4122900304004835,0.15033250565551282,0.6879687838124511,
 0.585395754023437,0.6984084854231734,0.2303024941219367,0.582416705353507,
 0.7375485459085507,0.6240664413158075,0.416861286815263,0.3506595684735391,
 0.21945220878272886,0.002259352216207988,0.3320728019416145,
 0.1374687912425191,0.740056622958791,0.5737867023113694,0.2155549258114925,
 0.7324259484808319,0.6918149411581553,0.14575715850917825,
 0.21785420581783543,0.6295261516083849,0.7701810412023291,
 0.48593080060855254,0.17649590889992994,0.14340683101056606,
 0.6650077940496182,0.4174049804413478,0.34699876665752627,0.8058331147234351,
 0.7052928874211972,0.8545453919948002,0.7235918101362272,0.9648142648863065,
 0.013533385767347972,0.004034146554077744,0.7140931679194291,
 0.054928252629191765,0.8969434357210266,0.4698247032440308,
 0.6551632869201949,0.863885495374009,0.2392220256151787,0.2124290210490969,
 0.8062674038701569,0.5943844054764,0.7172629066039237,0.5421295797977206,
 0.230977099649204,0.9587381105703339,0.6771326013448447,0.5432021222094177,
 0.7921914326004074,0.3327325246057077,0.4479076552589305,0.4221794362251813,
 0.5988755549244688,0.6568950764163907,0.8234864959745356,0.07040883721438529,
 0.09146128475264215,0.10106983217404131,0.5249392458850051,
 0.49903622863898245,0.5248900797529571,0.788716692060481,0.9317008289743511,
 0.6658985156246486,0.8528687112776699,0.6447238477069591,0.6665172743315808,
 0.6927210947095822,0.0533928392408165,0.6261851355765748,0.2649641998409107,
 0.3584808264042162,0.7258533829657768,0.6440537994877219,0.7639013003149843,
 0.8587259258293717,0.4450296628384345,0.09908294678843843,0.9122045460726398,
 0.1403113513709352,0.5418934698422211,0.837219880961733,0.26416946397436547,
 0.8551802809838295,0.9885794166792247,0.3244197416416619,0.2144434975754289,
 0.04791196342298687,0.5657207198845225,0.8478743589074154,0.5388617180693418,
 0.6297497267426949,0.23251988115657196,0.004406448475804514,
 0.3978143058054824,0.7384137647099478,0.6632401567616073,0.2430052074720268,
 0.2232125029105757,0.2156264929380287,0.6194031620347487,0.674357854079759,
 0.1367014419447261,0.24495295724590505,0.5852435747083383,0.8942281963426248,
 0.8808277006013704,0.45110749208853695,0.25236742098524256,
 0.8103214102824938,0.5352039371321562,0.2607815177540801,0.11599340720131401,
 0.0910554595296702,0.28292912042528284,0.7166958178456585,0.8030458515051367,
 0.2505183630280099,0.6300965370668482,0.5211797685211922,0.6215485897397098,
 0.6582479688114121,0.8578640954482156,0.4306776945833536,0.20350947571559885,
 0.13551075890937903,0.9596446965195667,0.931168623659109,0.13696238747913092,
 0.4733168783474213,0.3097477549644504,0.5300391634914025,0.6856922277135293,
 0.09874810245175392,0.05063785404725085,0.409099599507451,0.9294375287410148,
 0.15363013363624045,0.16770573429586555,0.6753755932337062,
 0.02311033954509878,0.20195226739083172,0.9186383822991773,
 0.01662117433938426,0.1014914500568368,0.688534674036096,0.9924154834954826,
 0.6990868814267829,0.4675886970237162,0.10188898143907121,0.6721182359955497,
 0.6073386806252604,0.9723545028383151,0.7646295834791429,0.4308964135399851,
 0.4267050853361819,0.4996484001507908,0.9007289020645023,0.42910848089687337,
 0.6237363229210781,0.2527988791460972,0.7765455722442062,0.44645862607071474,
 0.2762576846915389,0.18722669553829696,0.7914224306126968,0.4382081193371684,
 0.2475206943619429,0.7404763786232772,0.16226096358683828,0.9640790467577944,
 0.0747951815750656,0.6245484214871495,0.8075821917966675,0.2594100592109956,
 0.34017238220043955,0.7048066649104273,0.5975560059303484,0.941998517875091,
 0.22641697576638242,0.4702809980416962,0.45840029358301493,
 0.44876995865600056,0.526163264571927,0.3520095226627036,0.6449895369932452,
 0.635599357341472,0.2482292317555621,0.08785601059206116,0.28694359997425944,
 0.23375188028864757,0.834019118290919,0.8527039733607219,0.358188867497104,
 0.18713063959885035,0.49584450613268405,0.6074334171295739,
 0.14644603721032556,0.34068113637328434,0.05343137401634679,
 0.42641640125184066,0.7245958373582162,0.30389899441741786,
 0.6183600006540084,0.8259817035143943,0.7027213564134372,0.4512012967232737,
 0.803242076943672,0.17359411538109526,0.9876890631436808,0.5571319674603878,
 0.700790880009704,0.16878690921688022,0.5983930016252967,0.9818876807954547,
 0.5513509756754735,0.4110577415334603,0.7457242416874063,0.22891575104661366,
 0.43400022126581317,0.3701489245732287,0.11899658126584034,
 0.6024279378033379,0.49692942321074773,0.8763412073028052,
 0.30329480071307535,0.5768426616364425,0.44459147133407273,
 0.4938072930174846,0.3363212016251149,0.2216894392469666,0.4609288661811376,
 0.18988624703071766,0.2878384931573059,0.5666517331073556,0.2386086851471545,
 0.7906679468578913,0.6187663315274466,0.8074811566447548,0.3981869374599596,
 0.308013617957297,0.2633281400045767,0.3447290150847778,0.17347026877570682,
 0.7033523237373865,0.08199817095803885,0.8542199691520651,0.8682674128529074,
 0.17567200969291352,0.7411379184111478,0.0906194533804291,0.9436330347628464,
 0.10238369029538152,0.4352391765814342,0.0016336862633005156,
 0.01142321132172397,0.8419615425059938,0.4869478685468541,0.861478023990861,
 0.7565260497595439,0.8935423854281894,0.030842239883437483,
 0.7583266702391367,0.08027191197203187,0.07714009463334626,0.748696767205946,
 0.6241038367068944,0.5905925086336615,0.050634624663742356,
 0.06596691334540095,0.31020460775962255,0.6871835628868159,
 0.017530697581458088,0.13542308172646145,0.5612970177752844,
 0.33557759266518017,0.6940849411214834,0.1999400745655232,0.5360202684379208,
 0.5897011777078884,0.6671888231116064,0.32227277740587257,
 0.06646607271468463,0.27760521518841486,0.4117666432439995,
 0.3386037537946688,0.8938911521615329,0.3746726177994214,0.5077894181188831,
 0.8142643744639246,0.20313885268538823,0.435980039764698,0.10362820280456986,
 0.16502638825843663,0.8543167671445322,0.256831932588523,
 0.012972601882259238,0.69884438618708,0.5691120661260518,
 0.026222461319867385,0.8514722303244036,0.06658768672597559,
 0.9605538410327965,0.6861003371757421,0.8196743821416719,0.21390582088807164,
 0.5619290179314399,0.615200132808989,0.4212693930802969,0.5331916315326833,
 0.5948207919739924,0.3054302022434372,0.7701326789745211,0.834018432342724,
 0.07760120098357959,0.09850245525275936,0.3587756434568917,
 0.9530345822367071,0.9963006340554625,0.6862850672078388,0.4891746122862456,
 0.522787908902578,0.2511248603931401,0.39582309731871057,0.1321701061363061,
 0.12511985115626278,0.041917461495096564,0.09837930966247777,
 0.4400373511271224,0.13230812534200098,0.27065827740366577,0.882527873248681,
 0.7391793904011081,0.48523757474907947,0.9528521913362978,0.1125164685198663,
 0.17355039257864058,0.9460936342066113,0.9542069585672074,
 0.037203526023151245,0.7402808118947579,0.2709173521627487,
 0.43218697482190116,0.45570486918771114,0.09833106037001793,
 0.8569658865006557,0.6010623525166503,0.8766983624048389,0.5329919255881667,
 0.060982477193251894,0.22663128164014745,0.9980607939984387,
 0.017780232805409968,0.3621127048930439,0.17533771621556493,
 0.9809919809327728,0.5309452128383625,0.24830483431980066,
 0.06341070948766447,0.11405817288308018,0.19973814835099657,
 0.7021425902133047,0.4649963601997864,0.82719170520201,0.6407919855987153,
 0.8392449278250188,0.9509546254076553,0.25747011063405933,0.6708578320291917,
 0.3156037813221424,0.9969695797091005,0.19825103104620578,
 0.011265626293104236,0.07195117985018207,0.29271297486966297,
 0.19479764560668344,0.044175011199895864,0.9547345789742427,
 0.5951587983153006,0.7022744776341431,0.45078784044419873,0.2881948851452055,
 0.8931793846102648,0.44820245299050576,0.0712462229211045,0.2930011902931505,
 0.6593393784014754,0.8292864102633611,0.0731702780718857,0.46119414556892224,
 0.698108096211754,0.7467171061240603,0.6785114315010985,0.39048577697577036,
 0.2352002340321263,0.2253872670835364,0.3473152512455685,0.6340018891124183,
 0.6607368210620997,0.10872963678584369,0.7356835469659277,
 0.38829684610339554,0.043021313948317874,0.28487773602141375,
 0.5450005209042693,0.48945332628368843,0.11639432958442908,
 0.5282796707946867,0.2719199609549795,0.48924704255578066,0.7287131118024271,
 0.15448081027942062,0.6298530610981725,0.8334876292268354,0.6976986717676436,
 0.5854247339188319,0.21690972297933786,0.9705595932642361,0.9333707425389257,
 0.3008012502500014,0.4389654191713107,0.5868804376906653,0.27371207781534024,
 0.5874889629235421,0.6895980709713997,0.5020164415304907,0.765982504291328,
 0.7598477597001896,0.5145302715933526,0.9478395906294208,0.5372998733569652,
 0.9215668632737716,0.7363070226579874,0.8155650636522,0.9549475631494424,
 0.05781549520068252,0.17604005630748687,0.10635201876261724,
 0.9297286772263382,0.8856078599300572,0.05409512111918302,0.6013006239866503,
 0.9117175735830948,0.015526164230226569,0.17603017368240026,0.28243007389009,
 0.6291039298750227,0.9199145292456691,0.21840041307576863,0.8876967489737924,
 0.47732663579425605,0.39176692995626683,0.7847805761697425,0.432573868990207,
 0.477141918159134,0.43012605483064137,0.2761014868211119,0.5564175148423711,
 0.27951922684438046,0.5367290931120304,0.6377310309480066,
 0.16501785209403486,0.8663030293432127,0.3777709104695276,0.836701221390143,
 0.6856248575773555,0.9287197675396958,0.9079575725647259,0.8268894757766039,
 0.6554274017442453,0.9036151231325384,0.8178826651272837,0.9081525407390817,
 0.3910375739845814,0.72308905653877,0.7620814489060185,0.22502361257720893,
 0.16197295144385282,0.13776108889412098,0.19523058842470342,
 0.33509851888243036,0.38275645585423523,0.7829058343097298,
 0.4280595892339729,0.5867018878227968,0.4258173089610221,0.2975854634134102,
 0.9526800210539681,0.8824640845191767,0.4406525542726102,
 0.015935184327305785,0.28402735747419994,0.8261020520475291,
 0.9133189983350823,0.7049226755456827,0.36891883615244625,
 0.06365594616300263,0.060711460112589766,0.368953597406221,
 0.6082364109827432,0.6094872021957667,0.6956775577913411,0.46326335282740294,
 0.9008822423831719,0.9836429175343608,0.03468690983366651,0.8311102628410401,
 0.7552966296932064,0.39492977421844355,0.7828945126180331,0.673664550732834,
 0.8453455496209308,0.6649710172844104,0.2696625389249505,0.4712732946367516,
 0.6783137221333335,0.9767331911708899,0.8482357239379168,0.4541951036395222,
 0.26551139986845995,0.7464722199755858,0.4994299344692291,
 0.12585249644385033,0.9293596942501111,0.6532859348602512,
 0.18216194369085148,0.8091747387306083,0.24563336669804603,
 6.241347714683876e-4,0.23353457882667183,0.9553147674360695,
 0.6295997972997216,0.2166827398726363,0.40284006141650874,0.81641079344312,
 0.6834475185274806,0.09780479209542237,0.46608219745000967,
 0.9437645495096709,0.042599287158455734,0.9531125353862744,
 0.38250996120310843,0.9571750587465143,0.3085358762663928,0.5287665932566245,
 0.07579161932181555,0.9319612161110359,0.6475431353214565,0.8208373330153736,
 0.5710291571302238,0.2218481127790486,0.4453017508105618,0.8431066462389885,
 0.8061357295783502,0.986650479476729,0.5609440004700244,0.27399141480710854,
 0.12226878495302551,0.2686804326806299,0.12586528345578074,
 0.0040651623821631455,0.65676543808521,0.7461575830978583,
 0.24755183092289623,0.09209939374392895,0.9135898712852496,
 0.0030238069049811633,0.7946801459944673,0.3930179969284251,
 0.09379955228743242,0.8611015061704498,0.9757330929281951,0.6022247617407244,
 0.474395914195782,0.6819061027580288,0.14937834171397368,0.4958734670639442,
 0.4431742337688459,0.07290616353910484,0.24260247927459544,
 0.3139558599369675,0.026964099562417077,0.25208295022699145,
 0.11650146879516732,0.9662057140957006,0.10186835051452225,
 0.25093337752011957,0.3993206468548057,0.22305069293389146,
 0.6463288719927772,0.06600533296600952,0.5717868499453393,0.5092786942027394,
 0.2577653857142982,0.7978684547452973,0.13440938218005138,0.9520586070031519,
 0.42901165018360166,0.41034160754438354,0.6823341710478594,
 0.37971203707099754,0.6283744819489142,0.6106612613614247,0.7473199461326082,
 0.043625870913253983,0.9982614029730407,0.5240069922420648,
 0.19795622217393205,0.7356778103537779,0.7416802258550399,0.5119363767767323,
 0.3654254088396055,0.6091417842940698,0.3870840158596355,0.8749970101913606,
 0.15800859516499832,0.682365767590965,0.2532281308875789,0.4590016402948497,
 0.7205832950558615,0.4166294257476446,0.8778749686937297,0.8750312565731444,
 0.564091336474975,0.4347002053476694,0.19220338181445917,0.854933111192211,
 0.6020977665468525,0.026906572225736003,0.9926423216968611,
 0.7179855037847083,0.8076800440554431,0.48598535500694406,0.7171674118838806,
 0.9633113562902513,0.42273448190339424,0.7983893094725172,0.8396946118681141,
 0.25743037560852333,0.19714675315216168,0.745526033327686,0.9077284653794784,
 0.17665949267806114,0.03179423450215646,0.3231172117531149,0.329527947349451,
 0.7846227948627639,0.077939784045026,0.17257489904353984,0.47125418465081004,
 0.5474182762286539,0.38335315368272216,0.5924587192149393,0.5874181434771311,
 0.9385923381938134,0.9073984356625685,0.6046839370238376,0.7664224024333944,
 0.7468352286504083,0.47128204583740607,0.6081751629345034,0.5390593192350593,
 0.5203424091218983,0.28681597205100273,0.8062321783623909,0.6353158975947004,
 0.03828894616648526,0.9877778286128132,0.18289905368870352,
 0.4046928167670508,0.20249944831593802,0.21864481512423573,
 0.7253724208728585,0.6781372390344832,0.04507646680267885,
 0.29287514044651575,0.003609083649272815,0.9647819550644862,
 0.3857122279394921,0.47417556320061327,0.16751600003692424,
 0.09590213441734763,0.3930449643733025,0.793368957369553,0.3120392621005902,
 0.8708662317550538,0.09598094609097485,0.0687730295922242,
 0.17434538848036818,0.18171055585559537,0.22346231151305318,
 0.27905963781512244,0.5271644737661748,0.39857718619916915,
 0.39637570433196245,0.30406053302586467,0.09039102758524498,
 0.9156959447133864,0.04317597458914002,0.3665086643999903,0.5305022322872048,
 0.8072537855285513,0.9898539735874342,0.3688702721622681,0.7996782866389134,
 0.3060892490878955,0.7946911063091912,0.28985592278316097,0.5573160020936455,
 0.6152290319771215,0.6898631756680497,0.5104574449551854,0.6578865185066141,
 0.04569159474850548,0.8351899293863293,0.163027966918453,0.29058679393318465,
 0.4293514305682695,0.7471236983645825,0.9988128121721269,0.8602675097477107,
 0.991229855860135,0.46544566376393537,0.7407652473341974,0.5743183715659077,
 0.5661462454664287,0.5813148685604386,0.42952930320513805,0.9421120646486123,
 0.7041977823443322,0.8704393094871326,0.5318058256980155,0.18762397967994182,
 0.33726012201469013,0.9006559002703985,0.4441547936238125,0.345355529613255,
 0.8218070581359945,0.0530181465443601,0.8230475773963201,0.3714506928099035,
 0.20990136562304318,0.11820411673802389,0.8911022381022011,
 0.4670372430463381,0.6948206075979982,0.6736742961068355,0.05216401555395678,
 0.7540173451392024,0.5997153727164578,0.31508059254433796,0.3625356286535547,
 0.9286632551096377,0.8752864838975628,0.977267868348587,0.1145009993880719,
 0.5021926512361474,0.6517521595241687,0.40941155157477394,0.706868725161774,
 0.9368888266554577,0.7709252563581235,0.0841615889177999,0.4966795172407643,
 0.22311956818059064,0.38157367235587836,0.5613145449186239,
 0.5421431716302545,0.9153461032219782,0.8476251219150719,0.6202255419822604,
 0.7092860610163303,0.5837636026683184,0.8498534319867053,0.42997428193538934,
 0.5196885485199734,0.3988347343666573,0.9904774723901822,0.8936180638373327,
 0.48130733623390654,0.6502258297254259,0.5488791872983256,0.7364584191983284,
 0.07581913039872334,0.03137000576886062,0.2181792123373245,
 0.6572629445675178,0.6626354551385709,0.02672779873668385,0.9122859880313633,
 0.8703354256219984,0.8099484321888205,0.27625653093637215,
 0.18840186971214512,0.7363230996718264,0.8100710633337318,0.9151053596979254,
 0.5346714762535155,0.5293656642047506,0.5791925541099301,0.9626985257418395,
 0.9605598087319839,0.7255865263384469,0.46951161307993006,
 0.04080479297805817,0.6224640346795738,0.8702127374416593,0.1378081215346869,
 0.6334872680079879,0.17838324969560093,0.49382696211049315,
 0.4903858143888624,0.23852198294123028,0.16479062109175802,
 0.6068981552335215,0.02890001959771893,0.839444534254088,0.5280903814310132,
 0.8825508163969641,0.18285005467839277,0.8002129188158652,0.2574879042213226,
 0.669776927101172,0.8005708977888832,0.5455462329422498,0.5621117742438944,
 0.4167707434518466,0.13703595066987584,0.22970399530736385,
 0.6573656611254404,0.8909505080781815,0.15874999568254522,
 0.12397464595841634,0.2224191796053102,0.6432509326695988,0.3652839387571152,
 0.34133910240528975,0.1752339062633046,0.06653958767989565,
 0.6078455996078256,0.7871405651382611,0.3330703575023014,0.13415280176964883,
 0.3570993974822606,0.16374266271888893,0.6560491575314358,0.5671632348832395,
 0.3477099316634562,0.5437948266515898,0.8715560564543656,0.8030107562629835,
 0.28199541209486,0.7605204936889729,0.3598056441800469,0.6381559784355726,
 0.7744655358054859,0.006307783377992315,0.20796061172869595,
 0.5874268199749608,0.8045571671530183,0.39093233747468925,
 0.07847201654777836,0.4791622143556479,0.612053573540413,
 0.043539693900282894,0.44964870077721175,0.9202784591685029,
 0.6732344601706864,0.026109694648569826,0.23550712061004853,
 0.019759817861424667,0.32740460929393667,0.790331469425696,
 0.9809640104601647,0.2387680018394741,0.5171214357995252,0.4535412591978123,
 0.6268619226678737,0.24081984602371653,0.5663866261012823,0.8060249515866442,
 0.06471933838886668,0.8856337161175922,0.25872580175815496,
 0.25840346708361395,0.5735232542102757,0.38230587227108626,
 0.6938457133937799,0.3268041756600213,0.7846557976842925,0.8665550689768721,
 0.9056676917090812,0.4062036350743965,0.7748216749645821,0.623714060830963,
 0.7095135799158848,0.3783147547551785,0.04080253990681049,0.5832654161398865,
 0.22059951529514732,0.411278165210314,0.5810356509640994,0.06856181726357846,
 0.28038637607185757,0.4148526811175993,0.8413768821879606,0.7168354210768537,
 0.1485367851436331,0.8028616387772645,0.9247014342560509,0.3080595213882823,
 0.8418950898901117,0.6221837915679798,0.1319638490022691,0.6560425486165519,
 0.958444589463378,0.6611484293254064,0.13197595347609048,0.9116171102800112,
 0.5340477863596185,0.7582836017280005,0.7444586460279587,0.7825940561795734,
 0.33529753861049005,0.8109700614020408,0.8384357284631797,0.9690521588687032,
 0.5049378482487679,0.7172168028474187,0.5008796886072537,0.8391452015076091,
 0.0485442445488069,0.27885609915049325,0.791093048960489,0.9543856283764801,
 0.404582633609744,0.23689440655581895,0.42352880726463404,0.8781505158261598,
 0.6359144950044973,0.29427726962298184,0.7565569277786812,0.5062599229862877,
 0.4550196302706513,0.31289888596186555,0.02264585227285698,
 0.0775660737392625,0.6520156234468321,0.6518549937175465,0.9351707199146824,
 0.323879968386828,0.6154251714673664,0.2062629305070186,0.12368568024111737,
 0.010714660273677068,0.8973544499717622,0.33413895598806176,
 0.4047757288315347,0.47907654901055197,0.8841278425356375,0.7217166507391706,
 0.6078510492815429,0.9917604390598777,0.6563150602844748,0.14442708771045576,
 0.5044133455626107,0.050489283253637884,0.9822728318176426,0.92815998899739,
 0.029473382420316274,0.1784470436518103,0.9113094877983015,
 0.22573915323029448,0.07904820248646371,0.1249575615752051,
 0.6558717640438373,0.03603493413523218,0.02390720108762312,
 0.9165877585738714,0.5760058604485174,0.6134446183945284,0.07340684008392795,
 0.8994885323143164,0.44213172572441617,0.1283525964821003,0.4780351788716065,
 0.45992207570874977,0.10343862238693935,0.45867925227065753,
 0.08327657068339811,0.08948008082894832,0.7580970165869158,
 0.011436618536718846,0.36757567112716416,0.32902052448952124,
 0.029813240700855892,0.9750417996459271,0.5370853667329353,
 0.6684435472052833,0.5098716315264488,0.7064306403436846,0.14014512638722432,
 0.34955469957642704,0.7498222021130114,0.36552782145075746,
 0.48717359123617054,0.47585797943819674,0.3030228211703131,
 0.9865667672451145,0.610797725834464,0.34058034315832675,0.08463858521011103,
 0.8823624432024029,0.1744457078834215,0.1492961411546042,0.8977439477191731,
 0.9700263488920288,0.7781963857378911,0.2574229211873399,0.43760884406848477,
 0.6440715379835338,0.6838674056354678,0.3149398650435272,0.2242904996388012,
 0.6038250177887983,0.22460583781915655,0.7246239518141953,0.9231583859621013,
 0.5019573858520336,0.5612850512363371,0.9803474170948943,0.07021800064777173,
 0.39964968585767346,0.5570762122602211,0.13355755743610076,0.789848017318901,
 0.4321577114160252,0.36463509673787187,0.571421139743962,0.3477229459860517,
 0.8725118874944053,0.7700931642535097,0.5310038549523192,0.6462993562270896,
 0.44523804018416635,0.46034048257366766,0.29680478382862163,
 0.9038472218771703,0.8929925428787713,0.631955719702732,0.6867435951434501,
 0.8478084960597216,0.47951765116189604,0.14875968015513363,
 0.8018334758325236,0.5844281238910805,0.9312270642829619,0.5014404889479074,
 0.5932059670854967,0.24720564765794417,0.5762280594974197,0.6323522288538178,
 0.7467828091463468,0.9447800351705705,0.5546325844580591,0.23612885776131698,
 0.0935840239052037,0.8211062140222859,0.22418729895057898,0.1956247074443216,
 0.0681200710390274,0.5525873587043627,0.4523787622396567,0.817630134588335,
 0.3694066503926169,0.3699728440415455,0.03740983453147417,0.2101744402529513,
 0.732406208307843,0.6112531180207054,0.17665474684661808,0.6100991653837643,
 0.5198009655209266,0.519737061122294,0.7811034804697221,0.44680154268727057,
 0.8205578645643646,0.40739702751652473,0.8960776136083903,
 0.17884302248287365,0.07693543641988432,0.5321067562094888,
 0.2683248541000939,0.557525801649755,0.1854940657735522,0.27468140785359574,
 0.4770723198717497,0.78689766184904,0.24389012679113897,0.7684722239690698,
 0.07409567095309144,0.5636046199811282,0.05107359877152429,
 0.9243809504919377,0.37573684039399513,0.7941182442639038,0.8466718964075542,
 0.49760579942692273,0.973563335240399,0.2672618334307494,0.9798059781482543,
 0.3304547130569604,0.3711632639579546,0.4231843716314645,0.6745241068306468,
 0.9820161261920557,0.30415550342772835,0.4968288165585083,
 0.009637303143654852,0.7414489256143008,0.5080534124736081,0.272693053864563,
 0.194421095914723,0.032726696875407524,0.12472957793308526,
 0.7325636686312481,0.5259048768670723,0.8516425132827549,0.3054409049711516,
 0.8751798195022505,0.25498699390546165,0.6368646222254221,0.520090693315056,
 0.2793084574957344,0.5205926300856545,0.05363440566045119,0.8524049965289289,
 0.9101783307231228,0.8400424652517064,0.6513086749873799,0.8650094553514619,
 0.8156365633083582,0.13791901028813447,0.6295081423497594,0.7467655016311263,
 0.3345069389377452,0.7047539842746985,0.1278318710334655,0.35045703894151126,
 0.5296352989495002,0.8129241276315792,0.5915581216100172,0.14662462784286412,
 0.7841281123669519,0.3508437350184922,0.26098675902977275,0.4676927881399193,
 0.09030027696672027,0.9059045887152839,0.5956678002376692,
 0.49357202848640047,0.963770103402677,0.4327506090128028,
 0.017044424544474968,0.8333166806984892,0.9472049583613584,
 0.7324337086150562,0.5489055230494618,0.22643013412883417,0.7119264597995698,
 0.9874812818723836,0.9428405124945768,0.3727280121293818,0.737536383983344,
 0.21237359920581178,0.4839501068498733,0.7880740720044543,
 0.42640001121407345,0.3211602295095779,0.8349824120828446,0.505024371251706,
 0.8020336706029287,0.34515785251812836,0.2855965330341589,
 0.44210097134487136,0.948899196898332,0.10787925399705023,0.2713023766157535,
 0.4664732084668586,0.9894586957451325,0.3547992212435219,0.6163979454162598,
 0.31903236014284997,0.8629127709913068,0.33470946159967296,
 0.8607012770881945,0.9117056226993949,0.8311296097591389,0.2828753131046732,
 0.27859184046687446,0.09255817960840329,0.3985514900381416,
 0.03413977026533277,0.7532963646759944,0.9925326610112841,
 0.10696941957412165,0.36451791256279953,0.14205145053309542,
 0.08445650808673566,0.9768206349723177,0.5117857506549982,0.1284949038873877,
 0.7584507044525224,0.11159909915076771,0.6784129299049357,
 0.32309553035042404,0.5285291679386543,0.2412517252622376,0.9298989139427927,
 0.734390911274433,0.13059931149113035,0.7638221316100986,0.9294368094623764,
 0.1060948948922329,0.7161079608936406,0.9365497551078596,0.7867464454304433,
 0.44409695062623256,0.9649573284582397,0.05328613704542495,
 0.9287857754119528,0.6141968322075239,0.6383208010684196,0.6657483055146528,
 0.3793665471509934,0.7992509286074496,0.7014679849192886,0.5210618619406636,
 0.06032260619132368,0.8820416757220343,0.4376007562196308,0.9887007406163006,
 0.38510236780303164,0.16684243266251486,0.5019364329150947,
 0.1095256481222755,0.5429245371176578,0.8119874565887946,0.6918557090464825,
 0.533891244250656,0.516920575598269,0.010320136453622175,0.9205329998413165,
 0.5105458834119143,0.7712188448246072,0.5040096489751038,0.0718127009043712,
 0.5553955263666022,0.6398387989234603,0.4944131379281421,0.7826963806457099,
 0.9562932664401833,0.5468434886766015,0.11713862865796743,
 0.01238203723819542,0.05853055729649925,0.8046311990154718,
 0.5905219045414578,0.19950123566487932,0.6195184567113653,0.8042259894461725,
 0.5634689812653882,0.45058848394803097,0.5194322452743343,
 0.08595602761585108,0.36009580384559725,0.337879399806858,0.2772835362931756,
 0.13485662851543334,0.10323397245992738,0.7442441946832784,
 0.2849913213597002,0.560179382719935,0.1305951504580254,0.8362303712674264,
 0.32491198643733854,0.9384395105847325,0.6902760562587233,0.1211487198279293,
 0.567562027574433,0.34582737726876955,0.2994324928530401,0.8265398725248402,
 0.1335820499630731,0.38269131479071306,0.4729354817732341,0.9209152224937173,
 0.3247418694894506,0.09883673585116204,0.8542366302474438,0.6728837459772071,
 0.6800617559485588,0.9013865986798402,0.8115971972810628,0.7569490019765714,
 0.019153167352547684,0.5898722913761855,0.11373962552694727,
 0.7030053904984708,0.9404047695238043,0.14373049944518146,0.7183948363112287,
 0.5176217030806054,0.3486450458697756,0.9704665909039523,0.26105815612171257,
 0.7179957945174754,0.12209385305115461,0.4653332046067249,0.4917422695376219,
 0.7109296287914907,0.2007803369228729,0.8497475138313324,0.4490727084586419,
 0.7373936603644398,0.34094057901737707,0.2611494435148545,0.6370396396893183,
 0.8782075383005046,0.7794732830904476,0.33931266522350934,
 0.015007901342265817,0.5086952369526377,0.22843262583187096,
 0.15506107128114333,0.026432631016441888,0.17021983506406235,
 0.8860736495207353,0.16546688595340586,0.8657233028263213,0.9634221179908037,
 0.46553431864408457,0.9440361096218111,0.7259988299708762,0.7020048706435346,
 0.2997531250847635,0.6536581388584813,0.5308577431790054,0.1277014059651489,
 0.43185210434962973,0.3661138307349344,0.322534970998002,0.9822431725026747,
 0.6454166922029978,0.6734990247697202,0.1438590247272895,0.48125013302073794,
 0.7800091754341464,0.5589981967661821,0.7111189628874239,0.8686187584835329,
 0.5332629857556448,0.9031820490003641,0.8913586502216058,0.42737267436658843,
 0.2937279622215129,0.9551575236489152,0.9429252073765938,0.32426666812620664,
 0.7909003113680553,0.5844432167944006,0.4362551669510969,
 0.024046256783751052,0.9750963827141678,0.023266542479522645,
 0.608560678912323,0.824165411601502,0.17970886182526902,0.3517306026617074,
 0.48462382144765226,0.7232680153047681,0.8812323800049616,0.2643571961523783,
 0.4259929056670042,0.5324337456920941,0.9722010638831166,0.7225283485102597,
 0.35218997356710513,0.4331192813410041,0.1338161623384756,0.7766322977467045,
 0.9000299033697081,0.2885847621130747,0.014814710752831362,0.880995305908325,
 0.3232010092033123,0.4501337931738836,0.4028500437266179,0.37348488602470886,
 0.2729081336431408,0.8886544290616939,0.8660314450195505,0.8433688341026716,
 0.497803362190928,0.782309328998493,0.8092888356528136,0.13351649108612407,
 0.8368735267168108,0.6303923880949422,0.7997906626891946,0.07361613920092958,
 0.5908588641837733,0.4942936608915368,0.1999653112856813,0.23370763000987815,
 0.5208773596415097,0.3987460142681384,0.23344597797666522,0.9061611614379073,
 0.263578305530504,0.29283997657362804,0.028580741325930026,
 0.07733067489875223,0.3057731388030085,0.8787307964061146,0.9629992424371815,
 0.7849413984659591,0.08491242632724605,0.07296025192607569,
 0.3236558329421627,0.609051483610295,0.2688316680824442,0.5854792381768524,
 0.3489891840820081,0.2861227697707962,0.11916476950522958,
 0.31442841128531807,0.6608677700403234,0.9673994998953981,0.5220075862670738,
 0.792715478066441,0.4389846480563595,0.6176662085758872,0.74973231709459,
 0.547255713771361,0.05695395822949667,0.17260145986030284,0.6145100977415074,
 0.9222652863139127,0.26213896525625513,0.1220754317156224,
 0.07939528629176373,0.9247488767198034,0.49124471616922827,
 0.9362607625297372,0.8305825748739697,0.9746389798800412,0.35660352282762287,
 0.7036148222978785,0.17755332052769934,0.388386068712272,0.8284339940647311,
 0.5945895564242141,0.5336362207714372,0.7973275910654627,0.7832514342778961,
 0.22066216986645404,0.4869454944213617,0.11438757720769821,
 0.06181436179294442,0.5299584804590137,0.6312789113161177,0.9082615821435637,
 0.3704442856605138,0.6457369117259352,0.9553449250172155,0.10396468852908414,
 0.818869707904049,0.10645102917239369,0.8551070499198867,0.05827866861532527,
 0.9806258159790124,0.226560658649818,0.8538507271169085,0.5425749499747912,
 0.6623963829114394,0.02975200719831428,0.9827453023761179,0.573022975848999,
 0.21118104304903007,0.20402161913761363,0.5496175048541045,
 0.8774378752695978,0.6988283876848752,0.43816898441594676,0.4635263739818485,
 0.9505439377406169,0.13141898919025852,0.015878571251044526,
 0.7055680716631176,0.6918738845743971,0.8691254389416114,0.28533113655105813,
 0.8967554083093522,0.44898223725900266,0.9962641592009016,
 0.48024314370028964,0.6978325107352101,0.1989298533344398,0.5551131908101729,
 0.6975220950386426,0.5399398843665478,0.37329029363019073,0.9942151021078888,
 0.17178146891246016,0.06882290114618883,0.7688203108896388,0.723390176372136,
 0.7627464106765742,0.9395117508866583,0.5534876232828508,0.49911584825024335,
 0.03083320825866842,0.1656197775993642,0.6281184666180686,
 0.11103473440521061,0.6712453052076767,0.16265103223544608,
 0.7106349741706048,0.6158364197081354,0.09250915378985036,0.469564536010908,
 0.754329047758346,0.21671425472543038,0.4014671661913407,0.6057529652130683,
 0.7579696327754651,0.5081334662217487,0.5872339059118059,0.3793190364794392,
 0.48643656356771614,0.10167052610728933,0.33798721952832156,
 0.3677301619796758,0.05999416669207047,0.47146236980254363,
 0.9855030009985467,0.3912770136310485,0.05997298669363915,
 0.05274537747530794,0.28300117904641875,0.8599643927303575,
 0.2863572107186033,0.41493575049064557,0.35929357371715065,
 0.6860487867894406,0.21996622630102924,0.15898825575227016,
 0.1656186592160641,0.7267575963453963,0.1416653282471081,0.42258943563659557,
 0.23292610287793092,0.3147851525124409,0.5212886729717048,
 0.34886169071840045,0.9927782492833019,0.2981508962844043,
 0.16010774047538057,0.3353529441953538,0.3089713502194489,0.8563467043077829,
 0.8169001429390369,0.7346874943827464,0.5308922139743483,
 0.028458701514098794,0.4901541813199791,0.9732683775789085,
 0.46148882238669087,0.44348686280263194,0.6158677561588641,
 0.5080723122777694,0.37861720068500615,0.8174283057981444,
 0.08917037209205314,0.6816367315274643,0.6922916608220459,
 0.20993543345911203,0.6525356289555129,0.9969367134824326,0.0279986499037711,
 0.32800401904845655,0.9520074157434455,0.587597429879025,0.6631441457482738,
 0.6989168964908322,0.41445055030508104,0.8663370331759468,
 0.33159533037998057,0.7434222345901564,0.44424864161416155,
 0.047629354857894546,0.11746907543517693,0.8259416591407293,
 0.9469688888422872,0.49366751868100145,0.4137525339519005,0.9347339037376811,
 0.62349621851474,0.9615720107365348,0.1104115548522171,0.8124952377756566,
 0.7503394143788487,0.02546964555959086,0.5183538949513327,0.5900217254000927,
 0.4973964855229426,0.09248806684462074,0.44393528212128974,
 0.6990361775775098,0.3284374855911637,0.6361913079460291,0.02428587922095715,
 0.8290568026362977,0.02698431266921386,0.48638016431625664,
 0.6230600547068266,0.3595156524752995,0.7234529057739777,0.8924298545003898,
 0.5164401126419982,0.06996886367753052,0.4634134406186997,
 0.32869852419942547,0.6076291605038426,0.3640250651324165,0.7560763048381661,
 0.7468827386880337,0.2377937509233068,0.13656999056063868,0.419944872687369,
 0.2110907298786946,0.24415237183770966,0.5060435104262657,0.3019824836923577,
 0.10737884408662557,0.7418850374422346,0.3346782314750433,0.8876376845661483,
 0.3626494438260367,0.6394434519101283,0.8157445607159645,0.14077266737925942,
 0.44488963760916955,0.19153609362805823,0.04168937387045224,
 0.9267711769213581,0.15207675732409043,0.8163857927445138,
 0.49861514286614106,0.349566631390423,0.6183033965514013,0.3654409842686226,
 0.5166323537665723,0.5092060191977642,0.758734796553135,0.7038681529589028,
 0.3210063733857561,0.22421510397625688,0.27073237058791244,
 0.28845185524050343,0.17062346564675213,0.5445017178345313,
 0.8056689990891647,0.9262929950321371,0.9135266274972295,0.26998049372626765,
 0.7405794436315046,0.045062232075991915,0.8267475501830872,
 0.6184583765299994,0.7817217405995479,0.4042897698701827,0.47487272430404615,
 0.4531006311222292,0.09203920110554153,0.0579908191142684,0.5156877465096934,
 0.15757940424525585,0.24356710686379912,0.5593798636806366,
 0.2077936654380491,0.5679650010690767,0.20857161807429192,0.6532021480714378,
 0.6603207924889065,0.4191162054412313,0.8472387009181522,0.18885671153722372,
 0.8855381065702941,0.6035109699372705,0.34814936971629773,0.8347253688577174,
 0.24345791527186356,0.0054509312400392584,0.906457677443388,
 0.8552312097849367,0.19721769450659554,0.33761949054283313,
 0.9051906092684414,0.9056653715466136,0.5840657370849034,0.7967983800755536,
 0.48978151700377215,0.20774685144488758,0.42836754911434116,
 0.07681087198177317,0.15803500085845734,0.7623790367387723,
 0.3878419126475736,0.1388080109789812,0.1799250432630487,0.3261059546340501,
 0.40155525637080747,0.46051564520199106,0.36280949595165457,
 0.7188069049159587,0.9148085852617409,0.17997186814713562,0.4609239060496144,
 0.9039828632653069,0.8993593234649495,0.6455771978659067,0.4149530134811199,
 0.8485358997343861,0.435427473607073,0.1817831578732696,0.41027655163199006,
 0.2311511154688305,0.053342016887031884,0.9494950739734419,
 0.1899645768372762,0.8234325066559034,0.6941536807396341,0.22405989339685572,
 0.4113384713153686,0.9036899777358385,0.7580509114111134,0.8850508580586366,
 0.5278209533017448,0.42016405680221247,0.12128438082611037,
 0.8882715087685636,0.2521054635928497,0.4961960012858635,0.6461854685725621,
 0.26611726674246483,0.5566651808457352,0.8889585539236489,0.7086772578625467,
 0.7547498402052315,0.5410153197668618,0.26374783441803784,
 0.07065600546546669,0.4353334108855895,0.20193043863717464,
 0.3682405546692603,0.9695929706032638,0.10329914515535099,0.0673134207028595,
 0.9368829365669101,0.6733753522993512,0.964596994152486,0.7857318647266633,
 0.6156489071210234,0.4039621429223721,0.2929859105318451,0.3538548590646907,
 0.6048917296040509,0.9893114904608197,0.5722617811868489,0.4230419412022477,
 0.6076622705601757,0.4175922672833231,0.4024145709944754,0.3895032088249242,
 0.4345241856748413,0.19284366214920423,0.54235985356613,0.2866066392795983,
 0.45233376784273294,0.7352510595562629,0.806499123162554,0.4447606338807678,
 0.5166899356474698,0.3833776869242129,0.27131199279361895,0.6123568273366576,
 0.8811172303393574,0.566968688294298,0.10131048725570735,0.3242886358646725,
 0.3195393275723555,0.155507887186088,0.5397172139645245,0.8913729095073304,
 0.43181410233136663,0.504806759308363,0.25041478296146735,0.9921248040414191,
 0.37406659676216514,0.025562907925425593,0.09006560925327667,
 0.004807594290063277,0.175046525226361,0.9038255723411945,
 0.43109468851733646,0.18204136105064495,0.6345960170155908,0.919618312535418,
 0.8258855615623971,0.4266347208312131,0.43025001138316354,0.5965155698326204,
 0.9769075490902762,0.6159036117623674,0.38111304292893655,0.4365655232028699,
 0.7305320495478954,0.9534003116047279,0.3209519104490486,0.44084158556397957,
 0.6832816163227502,0.847974831559711,0.8337538110741397,0.9341734238431854,
 0.1674721092689554,0.9044310865724685,0.33801866363980104,
 0.11089892896738329,0.33027631926561185,0.3441470106607324,
 0.9339618695763873,0.12812079365116413,0.9193181288535375,
 0.43035994488591545,0.7718036296362042,0.3451990062114233,0.6114494388291286,
 0.10650593283803422,0.8521719211211877,0.8158432678894801,0.7162599090579611,
 0.9440687088811653,0.14611740932591422,0.33505030711143435,
 0.1756122240057667,0.0416237162081754,0.126180163885393,0.5963607650244289,
 0.02521016463292658,0.1060189352059,0.8217100015497713,0.4299257173606369,
 0.09916029068846788,0.1399183104216638,0.9504472882129069,
 0.47792640404704023,0.6366631668304054,0.8491361094608827,
 0.10608032013896618,0.09735777431465786,0.01676224329108411,
 0.3725244421385888,0.24461095799792532,0.9431257096606098,0.7040826457864893,
 0.7957656023129888,0.5243221081955887,0.3834857177033356,0.6621035563584122,
 0.41573540154084043,0.16141633679575873,0.236585683540259,0.2185605578715626,
 0.07773239536081733,0.13526692213701852,0.02577449780848151,
 0.540656688154022,0.5942403255163529,0.9165775215848286,0.9970408812509796,
 0.8442181884532817,0.25988253253904536,0.6232631776284746,0.1719244670600335,
 0.3584645341172963,0.8305733210307706,0.8255358829061666,0.28205681578368835,
 0.24920818435550407,0.5124942939225,0.8079451839502239,0.15237979395820767,
 0.2535880245765817,0.17658479393281923,0.48969683029017763,
 0.3796256861499131,0.06058356911711271,0.6505523365105227,0.3105494775572757,
 0.6142477704505739,0.5515708848522162,0.884847659901624,0.6856523731585626,
 0.020752170875105502,0.2368027622469423,0.7481032397152538,
 0.3200252933699457,0.8532638291031367,0.1181038226171034,0.5000380983937953,
 0.5099868516290452,0.9990137967860188,0.5269864840125056,0.9983863458784259,
 0.06445977770660494,0.6864559204057298,0.5238840264855044,
 0.0029950445551079063,0.11750632631275049,0.5069755364346535,
 0.7687578379530209,0.7090200182998135,0.3499947364651963,0.9324762246596022,
 0.44997613139314874,0.03793270028949358,0.3591817573143401,
 0.42749769542544547,0.6649724776068845,0.6016299216776014,
 0.08318422965015637,0.27659349048211956,0.031690000657634276,0.5771864477387,
 0.43678758371974014,0.2815627010207159,0.35321457941323187,
 0.2544451708404102,0.34942212383209115,0.8596163935775752,0.7076461662295422,
 0.15636939295816088,0.6773145020586957,0.5511538246651366,0.2746336376515288,
 0.560558290263744,0.4623944685639456,0.6062917836551834,0.242619702939042,
 0.7869023648831412,0.14352928990433034,0.7806702430503956,0.8349160013229618,
 0.2849272907375948,0.0248156069886718,0.7359617165390566,0.04342840921216573,
 0.46302524934609857,0.025011096735373384,0.8901245097130047,
 0.993938635075298,0.47501817833984605,0.3912745204729575,0.8752902054458908,
 0.08525683358580005,0.5843162341243648,0.8945734094419698,0.4322354888276687,
 0.5656787778046232,0.6869801413511464,0.13821059223390497,0.3648139673028593,
 0.3095769104410011,0.23783172268996344,0.927713243148697,0.8719387343987328,
 0.5950509223417833,0.5159965524315243,0.27923743723079086,0.8182177096678678,
 0.5324505383550409,0.2339012612828404,0.5499453446764949,0.6084775840224953,
 0.23271491372332076,0.5060820980746524,0.3572226297049903,0.3151477260250434,
 0.40626471164444344,0.15089014222348784,0.6270657607936528,
 0.5732289228944228,0.42548198827198414,0.6870858534454156,
 0.0064840756900941265,0.5505975051754562,0.27195278189806904,
 0.9657986515051434,0.7226849404801885,0.4962320658081478,0.5094958768473942,
 0.07734258968589536,0.7465447274119501,0.5595299639861941,0.5613265349670502,
 0.8278749319502666,0.7296871084538503,0.99691783875619,0.10263807681644643,
 0.1386423193489792,0.006253129764517196,0.459726788952985,
 0.09141985327477586,0.24307452917012595,0.8858569491842712,0.587076493235543,
 0.40140407926847566,0.6815122853609163,0.7905207600539031,
 0.17082395564430564,0.49440455869474553,0.21569566458292744,
 0.09612851930562727,0.9159783371437926,0.5410840202232344,0.8718325249753358,
 0.3953693179158806,0.9721283987024557,0.8213115046616932,0.7480808111633566,
 0.7171103073433861,0.9007949831354605,0.7454599183261549,0.03429855112564151,
 0.6818682644054679,0.7145231590696806,0.9004629518446913,0.9510565219163901,
 0.040667137193669545,0.2774587560201389,0.9397552823658064,
 0.4499118788012604,0.9922526427729702,0.5626960116523191,0.5701864931156518,
 0.734223298244345,0.7674792322964497,0.9014559269209192,0.8168466166169581,
 0.24898848647784222,0.061169638016874806,0.3033297549820697,
 0.02300221362017063,0.20631745469027107,0.9441191177413409,
 0.2176638108333364,0.18186717662155705,0.34404163448636205,
 0.9834079480928614,0.6104275093350351,0.3984111423199306,0.6862788225293046,
 0.8257543180505171,0.8199444947657732,0.729922505948522,0.6993657221685755,
 0.508510308152256,0.1851194127554484,0.5277524852408786,0.8929236103694578,
 0.460398141649739,0.3260106659197688,0.31567579847363914,0.12187275274766962,
 0.5977621614599714,0.9156085965706027,0.3876131841605186,0.30493218321433124,
 0.4749758720205598,0.744327894101587,0.7369900523681491,0.2932027039944465,
 0.5020098629834171,0.3676782724691563,0.016608710092088308,
 0.3468858742038571,0.14835274806457188,0.548658630052238,0.5064037733550538,
 0.5032785391999668,0.5864621464343709,0.5085286367492459,0.579978489687399,
 0.9736529346171237,0.015604328921801858,0.7662267081534566,
 0.6137303589326141,0.6358204448389762,0.5354015152816942,0.2351602883893794,
 0.039721450729715,0.17944532497290144,0.6762081353788638,0.19648986104047228,
 0.5961270407401844,0.9689491467806723,0.8382778187821777,0.3027536643577684,
 0.4147726819133708,0.6403748886238463,0.2503648902837037,0.33847051645286774,
 0.017254231531515063,0.44021807880211905,0.846534578225615,
 0.04974532661754383,0.4324495985572372,0.9879352163097986,0.9389666958264162,
 0.09005916457267649,0.6902163892144957,0.4082045915252831,0.9788432900750991,
 0.4948681753102462,0.04302651854724193,0.7919848521702192,0.9576597726466815,
 0.6928207114205093,0.5177550920497386,0.7169290018511789,0.23158725664847613,
 0.460073669277401,0.7645545739955396,0.18215838286408548,0.28651897824156114,
 0.11315436108040444,0.9277777751235826,0.054459685266214586,
 0.40819525915651855,0.495687225795163,0.5625866261069088,0.9442828456758093,
 0.9942796167627416,0.08007062782159635,0.9137173893321399,0.8235204168116682,
 0.48808128269339957,0.3159934259852283,0.8624366447081784,0.4443934696593006,
 0.7222360370691117,0.7277507641222505,0.5465426147565718,0.17466410870206917,
 0.9802227453090782,0.868545726772584,0.03903820149048032,0.577003129292285,
 0.49036780459685947,0.9273254146675675,0.9333935699726079,0.3263522631387451,
 0.5318516887534254,0.35744064519871266,0.3848227498679837,0.2978366105923794,
 0.7791739094476378,0.7091134523751993,0.268930330735598,0.7004477580459493,
 0.6450449970941501,0.5125829985013265,0.02294759664820567,0.960278859903466,
 0.12008971159506365,0.5712005182962193,0.9791330901992985,0.9292898825275445,
 0.790016062940757,0.6840483703972331,0.8102373671761209,0.3597602152877044,
 0.6378030121876938,0.0663344510561249,0.3018836801643163,0.5325046331178575,
 0.6419293023859026,0.23979300013366767,0.29480638148303095,
 0.7544854749148838,0.9231063326180169,0.5950259763930988,0.305820017742804,
 0.8485433220904147,0.48471202224715526,0.24315498999218876,
 0.8874140522752724,0.6574632099514125,0.7306279937529888,0.46090345860837445,
 0.3013804658716269,0.07037712795504303,0.8779232327540835,
 0.13383537978710014,0.8083833153152107,0.837124694830273,0.4540263552404199,
 0.29254997096414204,0.5678984920497311,0.34964553933790543,
 0.2458033241967883,0.4820991135033881,0.762592274283366,0.05959671351934559,
 0.6924495892270237,0.7886542505576413,0.836004068940167,0.12390570527394906,
 0.8721111465622533,0.030344120377935768,0.5573391192939923,
 0.2507033138628083,0.39628799531659065,0.36736152845755277,
 0.38121678300574546,0.6890631021704096,0.3255609093304771,0.851277331095875,
 0.47171729205253243,0.8596025467203483,0.9106534712655742,0.7744645767793613,
 0.4123667794881596,0.8407663061180684,0.6367229580440321,0.726452941199171,
 0.4239395468250624,0.7538492736279968,0.18658666624479192,
 0.19580850809940298,0.3439568216031621,0.7664250722042346,0.6578824574582413,
 0.0034656860627406694,0.2593525449157317,0.2680686281911745,
 0.9332623796820111,0.5412209996994551,0.5189161905248145,0.42622587536896483,
 0.38335348387527135,0.032691906055681064,0.8775896471778959,
 0.2712941447627808,0.05861291880344777,0.18723559867820094,
 0.8759360926664544,0.04777154555714036,0.8798798936828836,
 0.39549471477339815,0.41336733145684756,0.0946975450647145,
 0.7328357560021452,0.2058670690996719,0.14253084000863614,0.6189281744717671,
 0.5888772928511136,0.014469427334690588,0.1380511145199773,
 0.29842344457446235,0.8206537462510093,0.17514059800110782,
 0.34612660334389256,0.8097979213600821,0.31089319036391894,
 0.5629057930533279,0.31354843732918003,0.5458112183786289,0.4752533309254583,
 0.27484160500203136,0.511050867242197,0.10314338049721217,0.1766685553916193,
 0.19483027402228636,0.47506885936837995,0.22837489120315935,
 0.908719608170701,0.5766021898535303,0.14382156782910704,0.1698273429663073,
 0.11930601185675704,0.04402403198942584,0.9438758082916172,
 0.7948182509807018,0.3340349726055303,0.21658662028650477,0.9385880250849793,
 0.5811178317000343,0.5417093189484506,0.010398665374536153,
 0.23091826939524063,0.903107083495263,0.08889165142338684,
 0.34516803689570563,0.43041977364326334,0.5697134651510476,
 0.30332385835653153,0.5747225873199987,0.23683006815924024,
 0.7736146831842334,0.08255090866762183,0.3046926956457212,
 0.26925618136279794,0.24109831138917415,0.870820572860191,
 0.48905357161354657,0.02504391085747748,0.256901772224561,0.7265393188054687,
 0.8442243522349854,0.414365193645613,0.8109000636031002,0.6778631593617428,
 0.3233107377240648,0.6653394431912918,0.4004752960604695,0.6445317142632853,
 0.033470432533092676,0.12442434637198241,0.8004006004838137,
 0.3162883526855179,0.43339820595109624,0.8047418322051573,0.5730060292687431,
 0.34101426157827763,0.9919471773600644,0.08166666851687743,
 0.39916904304725187,0.044126356938397615,0.6928523738001049,
 0.40436034831083423,0.5192347167073275,0.6524329543385236,
 0.047358930577396596,0.310472318764493,0.3409488458029126,0.949632458271905,
 0.3235704958023248,0.48989998079458885,0.08214187415236374,
 0.4435597608314559,0.03620044867563477,0.37520526700799195,
 0.18407165259419456,4.777824019940313e-4,0.8287255586523747,
 0.5250722573160426,0.14249117983301585,0.4814569635451649,0.6884751344372602,
 0.2531910688642377,0.8709868676310467,0.8446543663952556,0.4498430768070394,
 0.5257099721121511,0.29976142105707615,0.1319648506051967,
 0.39918040273906896,0.8251796629766984,0.5514846901745586,
 0.47735555742069924,0.8048981148176979,0.9252476509226025,0.7724105226010405,
 0.22470232796604217,0.6941038221011505,0.6624109508037092,0.5059596219117528,
 0.14714196501554888,0.43048121237381254,0.17167942801478442,
 0.5189223564842044,0.2306605061833833,0.04221043015258863,
 0.08103088535283953,0.21002234038856005,0.1234257995916157,
 0.15496448861418421,0.6976035112943272,0.19804985348984128,
 0.15557493067575878,0.4524353932533466,0.1648776480531503,
 0.02918315307327546,0.8980750661763977,0.8597521520448679,0.1619931287162676,
 0.16514818271991705,0.7551107839663412,0.9155848655461378,0.5121757868413437,
 0.2249516857176893,0.6343545107207075,0.18473116066871798,0.7495614160776313,
 0.002843319640240094,0.8221743461385452,0.11735344982169216,
 0.7404083007671106,0.28709269705982843,0.15180203393469072,
 0.1687800007803658,0.21599248282869676,0.4291346597574287,
 0.34737206132474074,0.6556801061516603,0.6751126863589083,0.8913955162704105,
 0.4782303270250384,0.8564631459342443,0.7956959912703335,0.1935737528597974,
 0.41582970781326023,0.3813276345575388,0.6979024293537966,
 0.05352235905068925,0.14895891557343377,0.07464246523972262,
 0.6429204813695845,0.8592782048855464,0.654876057279857,0.49419496962797804,
 0.7002457481941423,0.39528766828553275,0.5870329182531185,0.8385859506945549,
 0.3391187254029483,0.1058230400105713,0.025301705854676948,
 0.7456998441436439,0.39310111708214035,0.11438121767364406,
 0.42534157783807336,0.9865022222780147,0.5154996340222145,
 0.18919098702711135,0.9746980212589763,0.7255523164268701,
 0.23714122949864613,0.33969681718011735,0.36299776309980714,
 0.04624026417142657,0.27936805639519835,0.2832898589688664,
 0.4103703291174141,0.9468082887709836,0.23949381214094245,0.8153549207250922,
 0.5869600992758682,0.8987277518065231,0.13029937134993264,0.5862729456284159,
 0.7583291017811051,0.6669357903783673,0.28509624094563213,0.8788384853656714,
 0.2839879267096196,0.6958817907174293,0.6532914743905369,0.31025924382536896,
 0.2126709575912551,0.9241651033119779,0.42094200752951716,0.1408378850194083,
 0.7773161313128724,0.24668239969789518,0.9076853863093295,0.691651775172921,
 0.30185125559876247,0.6707795629058093,0.32790901095835345,
 0.18508251958393218,0.1647739352204569,0.4010702985221065,
 0.22668864935167732,0.9471169662581639,0.765003675099382,0.2665232255764135,
 0.7279805109854165,0.7553187100501877,0.35668381434542085,
 0.42018951589668097,0.4513425596066486,0.3401541128948331,0.6940747470107097,
 0.11883907997079157,0.718356036164281,0.8636934015786817,
 0.045292123918980964,0.6833751819325927,0.3815534661914155,
 0.5787942089413197,0.9685940922135879,0.4872343221045472,0.33799827354866263,
 0.23453326074241732,0.6310511328476979,0.6364745889097196,
 0.29818387605453334,0.9973212142579714,0.14556599167460693,
 0.6264203186526411,0.7801995588640409,0.7946161511861449,0.13812426380252218,
 0.47310102582341873,0.04395293691600077,0.39531035608735376,
 0.7453273839326767,0.6074482171379112,0.26114457960840987,
 0.39780762088183297,0.2652452886254111,0.6132146947472401,0.888389320249315,
 0.36134031686690105,0.22830320681231986,0.031915464423296624,
 0.3355529991829064,0.9786093911150276,0.5625928845201486,0.9281546370637006,
 0.33162675752792425,0.5315767050815867,0.16924067601668324,
 0.9274452743329686,0.6754305089676526,0.8725187835985467,0.42016643735362,
 0.8745717955677996,0.5090715983717363,0.35863119535926535,0.2966298439104669,
 0.15623067298529403,0.3032861924373538,0.7661007354376914,0.3363915800865672,
 0.6027225365026998,0.09716372202229617,0.15506285415232468,
 0.8981557921471761,0.9349820888918283,0.8221063445912583,0.8279321082470034,
 0.4199010069701202,0.11337712077684226,0.3844203251159086,
 0.32962619739016785,0.5744146988557306,0.34656226713325966,
 0.5402412432386253,0.4416091289011864,0.016115342629034357,
 0.4782730067563393,0.28486373157120515,0.9728645058073949,0.9936137321637812,
 0.12765462711588216,0.7160606673837822,0.6068631081320054,0.8296860762597776,
 0.013275590426314876,0.6364484093865053,0.31802486641068817,
 0.4462790018397975,0.5499480465401742,0.5728230689557423,0.7437936350179276,
 0.31794745589309203,0.4752957791626756,0.0010133309486070052,
 0.44612526148085274,0.3064801527256197,0.8308460915809566,0.7224273645205453,
 0.10299831660539949,0.7420825262043641,0.05707573831395685,
 0.8503333977982162,0.2589645111576516,0.9512771225487843,0.7426202812038123,
 0.6297754499818988,0.09753513393987123,0.6823683299548249,
 0.061320907200338004,0.38669322006878937,0.6323859752538152,
 0.04759544714913788,0.28695919227296784,0.14838528403741524,
 0.2810580197669871,0.1863971996967465,0.9206538538042414,0.8226292737074011,
 0.5873064729749746,0.41526956825948746,0.44393552302572015,
 0.7758952771497649,0.17102124450669565,0.36979042112577976,
 0.9541175703051129,0.8082001207292054,0.9038159484611403,0.947585281022888,
 0.555446305996225,0.8284582420677765,0.1355122991916713,0.5310831766675856,
 0.9644087114172656,0.3591017138679313,0.5222136082124594,0.24050281353660385,
 0.7081005011031334,0.7488176341265677,0.16901009489361563,0.7175505042609188,
 0.018278061020140735,0.8157045524345671,0.30849751475541975,
 0.19993082181726174,0.020671031198099654,0.3087981323145501,
 0.8199860565322971,0.19787962798214742,0.26133177406343566,
 0.3686908865919686,0.5578643825213937,0.8209458080455486,0.186767271171975,
 0.9019865049873266,0.5102659202499089,0.6437632915672958,0.9316929231195566,
 0.8522112928472088,0.011120671943858795,0.9520543602412064,
 0.9352411392215129,0.9029635639703321,0.26376072585077714,0.6256547190260254,
 0.2908920028952666,0.6704060684086998,0.2563488150502915,0.2735909389722049,
 0.2408950095893535,0.5289270118716223,0.44202619399194154,0.9022514744024942,
 0.012863018801636428,0.3771758054024329,0.1327690156497735,0.443904628271544,
 0.5019302100593295,0.8617751419715459,0.46735843656646203,0.9391258472458948,
 0.5256216556169465,0.6422223490889896,0.7069227035093588,0.1836602696420463,
 0.2643220466033618,0.29341555243854145,0.6067580768750244,
 0.18708433625462073,0.8802777992418623,0.11752455972063314,
 0.37449907128195037,0.12074884042015788,0.6965266040387977,0.847148091299747,
 0.18801424254076127,0.22587799734775083,0.41057145336938794,
 0.3172776837486584,0.9559698467675861,0.019904604302327766,0.662823116520024,
 0.5964554050151074,0.2168025694877116,0.48738438400114403,0.6975861962633296,
 0.6257802904169099,0.5245793851991031,0.8325729731537757,0.6285821676265936,
 0.35038191531279894,0.9114568001688403,0.7990885505080861,
 0.30680687852363386,0.9616982758274265,0.5340574259721174,0.6535848722799255,
 0.011441267651315368,0.10874482495891324,0.6030229570245684,
 0.5813618960406782,0.8913699057337112,0.6792868972054732,0.6274341175878091,
 0.7216681705254167,0.7787183825719782,0.9813975436101855,0.3835285204864123,
 0.87376769105539,0.4204045212376979,0.664536855047451,0.7715870656540841,
 0.32904215766435296,0.5927403048567325,0.5599963625628768,0.5029276500347986,
 0.5566117636750775,0.4142235536591201,0.8552691136059777,0.8991535280652339,
 0.9481494082122094,0.8101316595288443,0.3487805278142374,0.10183654515196638,
 0.2593507738580363,0.5584511663219789,0.032746132685853246,
 0.9004489900775623,0.8292008016715673,0.518524353251906,0.5878813885137962,
 0.9545429196279065,0.2956844683422436,0.9745893132485977,0.16905718160341254,
 0.85976801386522,0.889290335294452,0.14972421823164017,0.052622500243154624,
 0.013420996570897015,0.8175496515624336,0.8044803239932463,
 0.0076468643013127835,0.3779900680755274,0.8900839965867711,
 0.9908502324601237,0.08014942562503602,0.36710384025566956,
 0.7993856974337168,0.23775342672628563,0.9245825324342651,0.2445440162908703,
 0.5377348908877035,0.05758580332415408,0.8522787975428592,
 0.08874971910238949,0.12905511463409092,0.34406394312977295,
 0.11852202584929317,0.2985540238413448,0.1391843714373976,0.3801155092388788,
 0.44614839594217137,0.4725799581669906,0.6742272210050513,
 0.44519603151468157,0.83583544088597,0.5409418324045342,0.6231015575116098,
 0.4410345897204563,0.12413569414992376,0.5264753367717633,0.6375825372823984,
 0.099398070463528,0.3149860243474356,0.973460981961585,0.4541942992390716,
 0.7155446217919089,0.19135401942694363,0.7121156905317113,
 0.045010559457537735,0.08084484362481059,0.4796286359419548,
 0.06237814273534936,0.5279063555390355,0.2535857156905813,0.5619531776888742,
 0.17130694809751534,0.49846863673797315,0.14696637074668695,
 0.6069779095327857,0.03117319890876158,0.9013776245474154,0.6168751107655119,
 0.4369978536438137,0.035724649907973305,0.8165492283143021,
 0.7266572608898197,0.6507017618153839,0.6796080130365054,0.08196966634293013,
 0.9538991160876641,0.7570638105733949,0.5697316334915546,0.16097236003727433,
 0.7992893862623895,0.22853508073091588,0.2572093242663608,
 0.10744404334407376,0.19776481192265039,0.4131658877958876,
 0.8912571010938836,0.6253210351934361,0.5922753083453032,0.4388914017807619,
 0.8859942262363425,0.47918246110046214,0.8854551403873498,0.9554578616713554,
 0.7073835394318916,0.5234833684069077,0.12360381770612827,0.6398679757896519,
 0.17201361048575337,0.284288169217333,0.8206765826066648,0.794337442854592,
 0.5598834929857154,0.4591594139753996,0.1844808159202148,0.6847731805121773,
 0.689205145351912,0.7077463328740452,0.46954240508685996,0.44722671658142776,
 0.32377017352433624,0.903570776311294,0.997574700691652,0.07744158227454911,
 0.4524918018688918,0.5259218278032723,0.2353676696525333,0.7253233702241458,
 0.31957589657774566,0.8806066623283206,0.31677026398944264,
 0.09114816516127777,0.13454007150327552,0.8165043859357216,
 0.4060530437045484,0.21630341237749906,0.24651804213076667,
 0.9657179451880691,0.909908141162595,0.16713600543393248,0.47436659834628336,
 0.4976970206148551,0.45567610091070854,0.544749894078584,0.2852400923398173,
 0.8744937251409848,0.4470203766178076,0.17157832451381982,0.4588509217680954,
 0.808010783775666,0.2118187738329581,0.8543909758410347,0.6191824991475565,
 0.49222783571644324,0.6974747329116123,0.4558440529781044,0.6796491809141689,
 0.3023271367925314,0.3714382443391646,0.9157625634790422,0.5548461853867201,
 0.572607122706331,0.9346333661114604,0.8968227577522025,0.44849862238457405,
 0.7593573471314268,0.0990552804459569,0.15434560772450034,
 0.023437550028685306,0.2957386098412442,0.6305473493014357,
 0.9298197078003059,0.013510681084818787,0.8379328149252244,
 0.048857598890917675,0.9208195301090522,0.8774165233600192,
 0.6204462069383531,0.9382646523797441,0.36733955602177837,0.2857835514150455,
 0.0661196580259632,0.9084183066115262,0.25644482397388224,0.9389700905860479,
 0.7331531187723807,0.646140240895602,0.5070571762302918,0.3086520347244157,
 0.907643113922602,0.51431937340912,0.55170977326333,0.23582500560542985,
 0.03641399844007798,0.45170370155482464,0.5656827074960578,
 0.007936258300545895,0.7999471194832627,0.2515456315207619,
 0.7618753846599626,0.9812589131475166,0.5252074776592944,0.5538832488290594,
 0.12474665698362641,0.9369760055069175,0.6641854373125529,0.6193688941361882,
 0.27212678115566025,0.14683059969312184,0.46954798681430043,
 0.2759065886110691,0.10633258317971706,0.6700790681012551,0.5942807574385603,
 0.047775828234216444,0.4605192624129226,0.38904359357220086,
 0.7679629867260231,0.2964136833330633,0.749862046423386,0.9328326617064379,
 0.8944836823020124,0.48620720673531204,0.18127430683025225,
 0.05252589635157556,0.90242379249936,0.9248595224606889,0.5399922386963342,
 0.2224587882129988,0.5679111570836858,0.58204699651001,0.09209976817080656,
 0.18961885342540963,0.40774294412230616,0.1909836124762122,
 0.24038932214807074,0.9878706537764622,0.9765397929080804,0.9506641470798419,
 0.2170038205289666,0.5170898418703415,0.3521014763129813,0.24756864564952152,
 0.6280274432536002,0.00995069428755424,0.33811140809222007,
 0.6985504480502612,0.3258000225230657,0.5842719050369991,0.29817498438191614,
 0.23654935237083974,0.9169134300463422,0.7628767834929936,
 0.38411591681259927,0.1899739687075752,0.9766860548559869,0.7695809731893546,
 0.5552852368550738,0.2342814391096989,0.49154268377367505,
 0.26233112573630124,0.5142127182076212,0.1542136803368045,
 0.22434666083306642,0.3499534274042828,0.11281492666289195,0.127487888243638,
 0.17921532944374086,0.9916924358346731,0.4404896995524167,0.9694128221526913,
 0.9080435108462209,0.2637428618222666,0.922938742826275,0.657809597320879,
 0.4231032791756586,0.9391543562478207,0.48789846535361203,
 0.23325288750337092,0.5132965629972653,0.927814038758464,0.9568933179141578,
 0.12419301735637334,0.12538832534802347,0.3190830709979585,
 0.6060518033217029,0.4452861831778534,0.26307429845773345,0.6526236909051668,
 0.5859156813455814,0.5536835222359652,0.11802790648348493,
 0.19772903681641996,0.5382131032538305,0.6914693784704793,
 0.20801296031883243,0.028943562278395008,0.15111191031601678,
 0.5889100626451197,0.367573099909832,0.14937534464219704,0.23822929115649705,
 0.9976608646251365,0.08485192551984044,0.8396032657883173,0.3061187473524385,
 0.9552186854190728,0.24525605772297654,0.2868585727126167,0.6545807191442781,
 0.4983215488191154,0.6566893806435122,0.9674445079471857,0.4611547191012533,
 0.5759999782488306,0.4933523547976757,0.9379245608364885,0.8918824977483681,
 0.7462150287198648,0.9962198721256792,0.15962172256575424,0.3077747019454673,
 0.832150997226305,0.15843991980955474,0.03924690788333407,0.5728077368572437,
 0.06297546218784589,0.04123404075878612,0.9021226315918506,
 0.017966143437095572,0.9322712871699027,0.17144751410120618,
 0.38798983831308953,0.7585037486641671,0.16499670303735103,
 0.5671239160787855,0.863671696948963,0.4589483495888166,0.1570258273569929,
 0.6969225334240813,0.3818558501945961,0.16699309193437528,0.3446572125737958,
 0.04600209567559921,0.3533115620776427,0.42408861414011045,
 0.5427341854052844,0.8172139784755638,0.5159903619720411,0.08025210904213842,
 0.7706011355012098,0.33708155265506234,0.0038796996766615743,
 0.02886299841176232,0.5683038950234771,0.7949244603938646,
 0.38063020222277033,0.4103016073949025,0.6150401638549903,
 0.26593768125534856,0.20110191383107745,0.5854669885687926,
 0.09105084476869307,0.3371002024976776,0.23469361596999538,
 0.15554307953326396,0.8957909985263915,0.8952699216214818,
 0.16987844200204505,0.6635602783445763,0.9534609711111112,0.477888363312129,
 0.15825705942888257,0.2709147226051508,0.5154802465152069,0.531791789494172,
 0.06236696794709806,0.452602544244335,0.5728253840968746,0.49934468852567604,
 0.42803895442903006,0.7324168561657972,0.7672252605470063,0.8197145533354074,
 0.5607014495359992,0.760860766322217,0.10409421540374064,0.8731351920706097,
 0.9139789037035692,0.15144371794661515,0.5542375078114408,
 0.27928152626607294,0.06493264169386881,0.33660919726715677,
 0.3276394324383233,0.46430980482903983,0.23070520056752075,
 0.10791639646831797,0.39444764449361724,0.7279826625652389,
 0.9264106579170798,0.058877723730763276,0.35082314360493105,
 0.04181819848922308,0.8583723928497154,0.09667099100669763,
 0.09553669625824734,0.11794180653089215,0.35746415752738503,
 0.7437681383775463,0.07238729291319534,3.9508933545717717e-4,
 0.1564183899532462,0.7961859740472859,0.1275983243386274,0.8446300989892752,
 0.08304637684129157,0.6549561968909554,0.7006098739435457,0.5081982666491482,
 0.08922300189835353,0.17055337115709057,0.723865088563281,0.5041793520394869,
 0.44255751512644526,0.40019887143916577,0.18246732455417447,
 0.13389809636471195,0.4923578726419572,0.8267992037902987,0.3715705983336193,
 0.14848887780175857,0.031071772195860325,0.8992890476045053,
 0.2795599990389511,0.7764807186473179,0.3377650609118842,0.9295193337739004,
 0.0031326608079539353,0.187962191904136,0.18104496556310878,
 0.36645344924766343,0.8760518443613018,0.2592708085842106,0.6490695699500414,
 0.5055220199261761,0.9374544127148274,0.5074692373347339,0.5101087570342185,
 0.40265998906087064,0.12229155318423612,0.7990516171483286,
 0.5291323443758912,0.9254500971597327,0.8217752406828255,0.12621003859365698,
 0.7039780887503623,0.6645852231365879,0.674109314208083,0.8367332414343568,
 0.459555775705328,0.7997765277996094,0.586615281225545,0.624063753028772,
 0.2571119949784316,0.33954179328371115,0.9408367846830775,
 0.33557128481212206,0.28221183587328635,0.7831264269296911,
 0.18735190241236288,0.27508769415867773,0.7279454361807077,
 0.31994714317408035,0.10836494226300564,0.8299394869061741,
 0.6254815537951562,0.3107628631208207,0.30249390918770414,0.3250277593087859,
 0.5695440222586108,0.7570689305071773,0.7977422611886855,0.9381853417468382,
 0.9488408834103812,0.5647301716578023,0.18791901699814206,0.92213016750569,
 0.3278255991981428,0.41135452869906675,0.8649000853130638,0.5639637623250051,
 0.17842352491558144,0.9485607252194631,0.6851006944247805,0.1406501769752837,
 0.5729554167340288,0.5557143200932764,0.9971175786306682,0.21019683806861744,
 0.027980221249554127,0.2954619218718616,0.5683894244119881,
 0.7703152819661607,0.6414521962621508,0.7681375497512771,0.7199206331804603,
 0.5768003092321791,0.8472063635845124,0.7731754940517874,0.5550613967024564,
 0.5397904841176671,0.9778313505275482,0.005260532838260112,
 0.3872154943017755,0.7998111346613471,0.2545231926833206,0.04157621229006714,
 0.5893709330237562,0.8669198561528267,0.49032880576664983,0.5947485382829948,
 0.20157649766059305,0.9819913958693363,0.9654334465542745,0.6654359615459806,
 0.8724834369283556,0.9351012380683769,0.8320965799913423,0.32576257167712974,
 0.66730017095286,0.01227434073696465,0.9149514075302115,0.9143839405106883,
 0.07201532372078745,0.9322662644093226,0.07070372769341815,
 0.43485378021883325,0.5395494495181477,0.9095695815106543,0.9982286464916854,
 0.7274955764039296,0.7202500142175786,0.056204430579265385,
 0.6881192605411238,0.6215898181265379,0.9909574226248455,0.8346085265026297,
 0.38552241715101443,0.5069682387194427,0.8078187615848904,0.3766132236121119,
 0.4925068164384945,0.27435325769083696,0.7519749830223744,0.616615714882892,
 0.3226997622523453,0.8907064054346979,0.515000536166552,0.2852867000954673,
 0.3409943158238524,0.47996920013353617,0.5494149938606605,0.3260823558447389,
 0.2554023583890539,0.2459145982992179,0.6096564893347196,0.6356643613461208,
 0.8995967791327999,0.17916808556795638,0.8751565114742186,0.4528949020850487,
 0.08411383009015783,0.5549828253928499,0.03704080405472254,
 0.4924777029652163,0.19925916802762322,0.5715196869380328,
 0.38898315863185695,0.851724836658551,0.6890969695738994,0.9700610792565054,
 0.5641709996987297,0.5606053206144028,0.030270770405369407,
 0.6269307247053477,0.9808468216861705,0.09183784469829248,0.8282417225655354,
 0.5039096973191507,0.7300083086891003,0.14214702001116353,0.6260346073115708,
 0.2762570821086773,0.5712532554757218,0.018807183142490858,
 0.3193601955032149,0.8854747637934213,0.4309626574897718,0.27659251691689046,
 0.9297962002659931,0.27835183346905734,0.684792028126952,0.7349856342382052,
 0.9232373622506986,0.9465268479808794,0.5113873441570774,0.819235453621257,
 0.32595728859648454,0.32826344867160506,0.7766084675137894,
 0.4337572062924768,0.6884860287564036,0.87022234065981,0.411495670131659,
 0.669057335175062,0.8606112899106915,0.39798006348152226,0.9765753921158125,
 0.43757660609969307,0.4554088936783498,0.4820714109533809,
 0.05594178338866285,0.4849037895000359,0.11038457197494844,
 0.9269617807689858,0.8509777752226714,0.1878594531641128,0.20273774697785374,
 0.05451836955975087,0.5417864332152533,0.8232224689988114,0.2631402976589434,
 0.9396732562938017,0.5549947533226114,0.01589254609898172,0.6739509181996736,
 0.7563689982028658,0.6484199877935948,0.217620792772106,0.3972704101955322,
 0.10038707393644608,0.39281533191997253,0.5199916235014963,
 0.4393931081432758,0.40197855193241305,0.731476310341059,0.43342713673806865,
 0.8467515020364327,0.2683978960945237,0.7225007246581858,0.32782970721727345,
 0.9151961254393458,0.8170228440349274,0.45463797791400906,0.6249763333509436,
 0.6179594491944272,0.15992445823371937,0.6180931587412263,0.7625182609909978,
 0.7340330341411383,0.29861225434208705,0.9120629981481164,0.6347106263992497,
 0.8896103613146766,0.9367317696257633,0.707842658084104,0.9473300332848531,
 0.9406047518726209,0.42552416630238543,0.7096618086389307,0.992334070914854,
 0.2418928344268988,0.8608844011946062,0.595882647813349,0.14582970409134088,
 0.07342429555153951,0.8538052944778345,0.09142731824012962,
 0.7335025080412017,0.5985354112898537,0.9510765947840465,0.8850865185729296,
 0.3336433933421161,0.5891076788145772,0.7169484959760294,0.8560293572337278,
 0.7485355236006872,0.7865576558615182,0.341803704644722,0.7840010502965807,
 0.8552368087690783,0.05223467370934398,0.6781178184127792,0.5966798468788961,
 0.01304175472909086,0.023350347073809785,0.5625992539044467,
 0.6094871016021659,0.8880143273174925,0.9893142301112422,0.34124758131266764,
 0.8058191199489589,0.5486283629073613,0.6667005734290337,
 0.015160833634093374,0.7320335965860556,0.3882148370313283,
 0.45525778035895437,0.10279020881675427,0.6519553176023336,
 0.05401385112846924,0.44590686735023266,0.2832529524112054,
 0.09952808151950965,0.8053765992602232,0.697576843470487,0.20288309151312833,
 0.5979531921662844,0.5831899206309633,0.05407396871469061,0.8620723927288946,
 0.7643652258008934,0.3779755060255001,0.6419107562128836,0.9320268146302653,
 0.022158407045922424,0.9935943661689809,0.4525326489468048,
 0.9570404725654894,0.4136045798943856,0.9225530478577529,0.13727820225871978,
 0.7957583175467111,0.8270272129637364,0.6782852856191167,0.24973221977717386,
 0.48256532759538795,0.8611471946360533,0.9179223805282369,0.9431802501683545,
 0.5390475554673707,0.4151232923918954,0.06462845860044919,0.9589100503123777,
 0.8142811433486643,0.757362921888109,0.804498213838905,0.6992976505179083,
 0.6588498628987144,0.5926055764907473,0.1479881812133501,0.7231727265911637,
 0.7608520666487195,0.7733328416117343,0.07536419968161479,0.5330736845452141,
 0.9273346152908752,0.3018300207041068,0.0677603187991932,0.7544157396263876,
 0.28542160484723733,0.8299987424045159,0.8406685473477145,0.5168298630249395,
 0.3624473786670813,0.7385887955509121,0.66623869494916,0.31607651120588653,
 0.6790634782623133,0.6844576521953709,0.9830690440779529,0.8437246547236221,
 0.24026118556124243,0.8981177239084054,0.010809368576730538,
 0.27594563903050406,0.0028540312140196145,0.6285070999157542,
 0.4263158399521563,0.44357342519731324,0.9992551830287331,
 0.05435911625835976,0.6496908049448545,0.8700711437092281,0.9345314432366496,
 0.32391437240218435,0.2773164676339056,0.6139205562621403,0.2829462590063334,
 0.5487381454173391,0.361783503501926,0.7444956721381452,0.04625986024060924,
 0.4384782258264577,0.9471445935417522,0.1632434496566213,0.5230359816981862,
 0.8940280199525913,0.4318767810867675,0.5917893676519362,0.6601975250956299,
 0.871650564361221,0.04355020875555582,0.10172654524331426,0.4285154170071166,
 0.2950755567075696,0.6569084219161162,0.9191489981668719,0.2893329752158492,
 0.44110219911260107,0.8243064919799935,0.8654367611399729,0.8666417276996468,
 0.3146410857638867,0.9333549982233138,0.7996358660243126,
 0.026832249599496927,0.7526423073991528,0.07031152322118706,
 0.014549847601636579,0.9874528071221201,0.3305150047919345,
 0.5979305981378067,0.18410487291433697,0.17176846516715527,
 0.20487320491144612,0.013264160840734895,0.06683089501405093,
 0.02177411631885129,0.8497128097531871,0.9795370425935656,0.5896257443128586,
 0.5206573254306922,0.19162612181454042,0.8698876846975427,
 0.03887952202035805,0.9276394814708925,0.10082929613563474,
 0.2510009908740938,0.1154410936151039,0.4439580601097112,0.9232014490319502,
 0.3017703911546782,0.6915906223474328,0.16291277468863452,0.3955145279547756,
 0.3716203463348464,0.239769796690377,0.12225562288056002,0.4108460614469165,
 0.08974588369703962,0.7386955042825024,0.1885500056582422,0.335681177506453,
 0.36440986223272165,0.5730420382394227,0.7046819309618548,
 0.28420693864323776,0.4490688027522757,0.8664205777317495,0.595864385382969,
 0.543849614017893,0.6475311228850502,0.9091613578809784,0.5033041705788831,
 0.6568857676563499,0.0016461055291965287,0.7132341696151395,
 0.4697189153624406,0.6454589496218074,0.8624821547359401,0.9931892650447964,
 0.07343488222078853,0.49221066687835313,0.337250600235371,
 0.28807979839387876,0.24880101423553158,0.06200054309766956,
 0.3896409258413742,0.21276732056227476,0.4241060149970267,0.6556516991131776,
 0.5360529421363517,0.41586290357320177,0.33636296633432294,
 0.18000296236541657,0.8631784170686058,0.8539278035040933,
 0.23544876335713916,0.8674468696058637,0.09461148777250317,
 0.33788440854099844,0.3607051025590595,0.23464166897869765,
 0.06983433215987866,0.21194350402963424,0.9226420869776603,
 0.4853766633326143,0.9970468335525385,0.23332176504987456,0.9536087113139493,
 0.7799365981396562,0.013632098521770608,0.5634311810049698,
 0.8742190332664151,0.919283365068506,0.8828750968578194,0.9202912513943138,
 0.6416471828642554,0.19214408360746882,0.5870973101800306,0.7846584654633247,
 0.7612942270768865,0.926050260405654,0.6925815404057748,0.34046279176301353,
 0.40051936023981416,0.7926949336290872,0.9202688528174491,0.9436130003129379,
 0.05387428509562908,0.19588787200923152,0.04223223356236383,
 0.09582450814352939,0.8050112594749895,0.6231130321365745,0.708814990932248,
 0.45721964322383635,0.10220252531122642,0.5804579374284922,
 0.20424906136557863,0.14427200394776307,0.3980092769776602,
 0.14197563093502874,0.3971774343351293,0.13134124458257013,
 0.8592682907687978,0.4902278125977285,0.4141174053976109,0.559475506642321,
 0.5825534643646262,0.7279598026645876,0.7996472672946163,0.5866645317718999,
 0.9299679586615015,0.5102081204485347,0.872851131469676,0.7188377936737439,
 0.8674868302805617,0.6956801455778441,0.3791881575637668,0.21707702088330794,
 0.6185618025341749,0.438022071974551,0.4158201709450138,0.6110095687641988,
 0.4210571161837414,0.44171094330170024,0.523890484377682,0.9713138746879203,
 0.5628865988469894,0.878619975822979,0.9990825619151581,0.6181636731001396,
 0.6224880381116312,0.8054713181818981,0.27998782512144316,0.4824803353539937,
 0.05361095294422902,0.5730595309762128,0.9007927772271911,0.9679468205942677,
 0.6964494514599225,0.25484128295313935,0.9576196399156623,0.8648892120693918,
 0.2275088428822507,0.5365790670406454,0.9734875699613751,0.6658156358286551,
 0.7934339709800172,0.012742542555371328,0.4337008699759781,
 0.25325610893365336,0.11767241407890994,0.8592292145887662,
 0.02620516149707408,0.20808257031390764,0.19069474896366911,
 0.6465827349951399,0.26434723141150973,0.1242576394486059,0.4684976725846003,
 0.7108682614834075,0.7398306876843046,0.611335658385772,0.996968588095338,
 0.8688825786386674,0.10701301980163291,0.8109759576972762,0.8030724463142285,
 0.5517405533950962,0.5241599438255315,0.9333566289642036,0.8352481229703581,
 0.5753807407997095,0.717923721877855,0.9922008167558485,0.1484942712510624,
 0.0018498507297829425,0.7737557344524592,0.26211927758060083,
 0.8763813685628239,0.018828480163702754,0.28385533758860937,
 0.4044846735174181,0.1613147831296915,0.9802004417670297,0.2253865453135172,
 0.6130756085898033,0.2673776684649506,0.22000138436763794,0.4026331086118412,
 0.9711675272937081,0.8307629648170074,0.29063823146984546,0.5729061027507599,
 0.9482263830230278,0.0028208152932941433,0.04024955256680096,
 0.8469001317796665,0.5128910350310335,0.5220239917242322,0.8196170496667372,
 0.37411508125274295,0.01373391195008411,0.35395923140999075,
 0.47453267584307823,0.9773085202538039,0.27246122731133604,
 0.5349933387290523,0.45806198277247956,0.3386864766161879,
 0.16164417859650937,0.8095820676155491,0.5472211241460839,0.6953682473730887,
 0.902132272132922,0.21846033379261698,0.4873709036932756,0.33830620951111934,
 0.28654934628555584,0.0895848338444104,0.04847745505216439,
 0.6582523630424973,0.05428786115021034,0.5548731365403418,0.6763203686627263,
 0.498578333998552,0.35886192449927246,0.6047204554356997,0.9979855309768456,
 0.2958712386161435,0.03389656010578035,0.9176794304163471,0.3244763819475418,
 0.05901080596208619,0.0019515271714692917,0.6553441352272331,
 0.9948668510509705,0.7004017790930489,0.35558843008508767,0.3808260654545058,
 0.07906519506778398,0.11825954362192448,0.059129585587694544,
 0.03782207164399998,0.7781909380500435,0.8223983275206737,0.1383349226341254,
 0.18043725962098323,0.6426054783313019,0.7647513215174403,0.545855932490374,
 0.8211535241571474,0.8514648646100577,0.9436525457996429,0.9544493046013751,
 0.7931865117320039,0.6971759653185128,0.6097580188298236,0.4827254270508349,
 0.7789932717658287,0.09388556915482715,0.3782817758709752,0.7339734867840797,
 0.538669705249432,0.6713290125459024,0.49833271886508845,0.3006357158716948,
 0.4449544090458303,0.9571140940604503,0.7728069710455355,0.8924116674542449,
 0.6630535432163811,0.9840412113447098,0.6358147561968566,0.5266671741618487,
 0.8352208136666331,0.5152601194585728,0.1019744378084847,0.7169816376017821,
 0.7254304773124487,0.7209752215684164,0.3979691825400381,0.2931367854910044,
 0.55143816931271,0.6502802725776865,0.06335648876665101,0.42651173420340105,
 0.24836148856804985,0.37329918715494204,0.7830992304563066,0.859912402942915,
 0.5114859947165752,0.4010480886272856,0.9567421031310694,0.9026421828784842,
 0.1025823623088008,0.8665721324485854,0.7692051991867672,0.902806439186297,
 0.6329290652061003,0.00226306754322958,0.9742395127980189,0.5835922341854588,
 0.6328522701456013,0.2730728328936073,0.8596445453199548,0.6945493924774357,
 0.39482609061189033,0.1543626628590018,0.3159570299791796,0.4719618024836775,
 0.015645604698957816,0.1547948205716969,0.040027540817092144,
 0.8146476326884272,0.09036949724989729,0.30283108748582677,
 0.41137890190808146,0.3857474214690828,0.1659966782169786,
 0.49483803711015795,0.7335394748596553,0.5526930822806166,0.6750922562350864,
 0.9220406022872312,0.7402004602447677,0.3225839368796184,0.04674814000091787,
 0.2338035327951259,0.5463807161327341,0.3482433647164328,0.30506284044964116,
 0.04299133595612137,0.07669507024688627,0.2616035235373213,
 0.4617671835025856,0.9010261648639513,0.7113985997520728,0.7918375607978405,
 0.28625559105039544,0.690573643853609,0.40653751369686086,0.7290905388944199,
 0.6691051456697065,0.5387590334152677,0.22189235283310282,
 0.04903220308149736,0.9663653288897971,0.171725457793787,0.6908733273891801,
 0.6211816224652127,0.2631383315867726,0.9155050066806104,0.8587018152071719,
 0.8301599125008423,0.4475918413062969,0.5943543318316915,0.5693438177188838,
 0.3291207415143209,0.9870083155951794,0.6301506153676106,0.7298600325164244,
 0.196868346688982,0.8798037242916856,0.07726118115894842,0.9545403365226716,
 0.29663692206932324,0.12795269701638767,0.29527029735005295,
 0.8745113011573898,0.0709015374449089,0.8481986374485178,0.6809898993263931,
 0.40673122490397584,0.7530932345333521,0.8929257396766345,0.4219283272683235,
 0.3295427095136251,0.38655300145235216,0.4664001158090919,
 0.19790086559662978,0.5145306066987778,0.5965702541376965,
 0.45836537722062376,0.12076162739909779,0.1430164230460771,
 0.14872057739291522,0.7597099086023215,0.43506864510721666,
 0.38829019383370866,0.8787054897282311,0.7838187684617963,0.9947380234265061,
 0.4213261317010807,0.7164375079063593,0.06261802960179441,0.6596956564702905,
 0.8989946244298092,0.9695085074338252,0.060252166584546796,
 0.11845569686430701,0.7747258081924808,0.8586852974388345,0.329137338479366,
 0.11368267546660704,0.4149866030987327,0.4120682598282944,0.6373135401741536,
 0.506808846201954,0.7198313644829994,0.16676498948221719,0.33110498065787974,
 0.1918418230913813,0.35726846355256625,0.776734967643609,0.20694308593201516,
 0.36714350472582513,0.47552493433617815,0.20613777339633776,
 0.8399007794621509,0.5032672578725136,0.942135107504518,0.1463913676081693,
 0.7602052075649854,0.11748358918092405,0.9559149824173405,0.8051005140373899,
 0.06965104428374569,0.039625507963923945,0.8429345551822974,
 0.045999306693226494,0.019523200400471508,0.3943149796203729,
 0.14268845075884318,0.21927044676420482,0.35521420146892924,
 0.9313900940737451,0.342298142418096,0.6804189014721818,0.7476896558636723,
 0.5001195549822526,0.022070730586673082,0.5367018446712605,
 0.3741128991900585,0.8171052563480338,0.9433697483288723,0.5841772615513341,
 0.4359128467033079,0.4461312891998168,0.4202879955981368,0.15312463127104692,
 0.5642424018114749,0.5014343143153397,0.08998162179607894,0.7853136029433697,
 0.11681435574601096,0.9816514080279459,0.44100278087925515,
 0.4124298176334149,0.192162609726358,0.8142313118737798,0.44111259543390313,
 0.0676162713686197,0.07324098291216607,0.7998300693722364,0.5120777687765324,
 0.5935483450633162,0.014397760246620539,0.7746662775106586,
 0.20061883708556583,0.20378547588265483,0.503695469257772,
 0.23631467862448308,0.5335024037550697,0.18321759858935394,
 0.45101821392544406,0.3357335669964001,0.09755984148079067,
 0.4467794831362999,0.3240024962642035,0.8744850962299628,0.6668823894229609,
 0.6190316589352696,0.7415103826883664,0.3159856371971682,0.5144776619832119,
 0.6339403927917974,0.8604605173811598,0.7026801858330195,0.20798382437930063,
 0.7705235935516349,0.2943052674575889,0.26839377139735576,0.8496271667745033,
 0.49062504302552445,0.23778950130831944,0.06914921536861796,
 0.010274685825705787,0.4190053163616637,0.18326306322852237,
 0.37636529112133477,0.8838866142163144,0.910950562311062,0.9474478414337235,
 0.9203495479193767,0.6629457382593238,0.5082644397017428,0.825998494562718,
 0.807961432963771,0.6346717768190584,0.10852786079220555,0.23014620698563093,
 0.4387746052363245,0.481631414526124,0.4318310297405904,0.2570386541475459,
 0.5156986625239182,0.5528773786045345,0.4627226551585746,0.0999799516753831,
 0.8305080718642311,0.7542127037482774,0.7958701728339675,0.7565955910188074,
 0.10822858992017192,0.5779320095840225,0.9770879806865623,0.8770721420947993,
 0.5058301104040007,0.9624774061967888,0.1819421251457275,0.1700940765818414,
 0.2460615621109179,0.2775424538399991,0.455849017638014,0.45029405009730505,
 0.9929733392898608,0.21631786218937799,0.2722774788383677,
 0.36572624673098686,0.7239833383924744,0.9167917210363613,
 0.04416549644687051,0.27512151376011573,0.6485627021012742,
 0.5323779633363838,0.262150364327162,0.4788111643287588,0.12243549045407431,
 0.33093078348664373,0.2524362864623171,0.1038600300646556,0.7746575754301177,
 0.9469395884002038,0.4901155435242539,0.5359709956197416,0.8559591199006343,
 0.060861497427452504,0.14035335306537788,0.6429139708320023,
 0.7984399452279093,0.13925177234266173,0.8635208895520247,0.8098956053257418,
 0.3049209959513848,0.5658208976212271,0.839989935748086,0.6754932661710397,
 0.7756396791826028,0.7641040462406162,0.9372864866868671,0.8885828936452641,
 0.639007775791056,0.3363068491269047,0.7906019761191893,0.604889750026272,
 0.7205275786426807,0.7995186703734606,0.6661094396387667,0.3107582656724708,
 0.8553041910605196,0.6116003854867686,0.8969264304722482,0.5934783584051025,
 0.07771670902079308,0.9840243595882003,0.8034506957866933,0.6630616385228464,
 0.5064199578334444,0.29476600954195664,0.883131899079471,0.8778424623726426,
 0.1361915684824887,0.5068520840489857,0.7229898435747207,0.03047749386005849,
 0.7558094075214282,0.33140784424351344,0.9984188563623518,
 0.28918605750322524,0.8693447731221213,0.24142017698535378,
 0.13224247905593367,0.792495962968043,0.7311537438582281,0.3195916194164523,
 0.982817665513737,0.8279733664075896,0.31106905812965924,0.01700875258320811,
 0.43620229062020754,0.34923362526130686,0.8094906078297632,
 0.9886673583738105,0.9309481830157886,0.7300461942724904,
 0.017732845736073366,0.08625040739457823,0.8863273912096312,
 0.21634082992131354,0.8965185786375184,0.02101592351348669,
 0.32868765549113865,0.34744369574217526,0.4837920509551874,
 0.01537684682409124,0.7473322792994489,0.3188450304954138,
 0.16352455897814366,0.49958457503675946,0.8310130352386957,
 0.012897706114905927,0.6103870542881576,0.33055309546192635,
 0.9829147741512582,0.3830252679334779,0.21256899958102748,
 0.47020389056330203,0.7651492629535154,0.38183957228225307,
 0.2370624380065156,0.5321998415307791,0.08003634682583294,0.9319856208330075,
 0.2968894472571799,0.6133522767150746,0.4786829219646521,0.2781675372383242,
 0.9725888070321214,0.45555992005838264,0.6306911078481986,
 0.23935110522812897,0.41118927400358996,0.11033406986706118,
 0.6482463186493489,0.6466958865669427,0.5210975282062937,0.4021305503857676,
 0.30673483491117104,0.542074542925591,0.009866427571202041,
 0.5362510008412152,0.2563035654442215,0.7927017504058035,0.26690028969299184,
 0.6856601724774049,0.21565788755822302,0.36193260268501426,
 0.5597212341528708,0.6827057326016933,0.5974650406781448,0.10278833436398871,
 0.36363791413529023,0.7418346079552456,0.44571820873605106,
 0.43657298978063497,0.9187421324709218,0.9900960020456041,0.8634761650763139,
 0.8793861421647928,0.6078657417902767,0.32656962552968305,
 0.16745516578122943,0.9374277459802356,0.44411990733196927,
 0.28431813911564563,0.3152663357130018,0.790612971251026,0.5444377453279422,
 0.5918444862506935,0.9463574797493066,0.8150069703273344,0.3050205601064684,
 0.6533666142570353,0.2418906192679593,0.652166977241012,0.3122560287094056,
 0.5192612304262911,0.11979682499152322,0.24898251312158437,
 0.4420673164124197,0.40598690663353854,0.8253033074540409,0.5158606814588118,
 0.9738661743662508,0.2820721985887391,0.7414804092727412,0.6725698808226114,
 0.9389462893251695,0.597216428089221,0.032804532559133515,0.7253444704264874,
 0.5424753126744174,0.8781427357130436,0.7569559559171828,0.11931549549530562,
 0.9151248762982355,0.8497601593439725,0.5706681992121714,0.9595731713184774,
 0.35864801362310195,0.9053217209515628,0.3147603035163359,0.417159084552586,
 0.515119481916537,0.6251957606236767,0.13688018326194373,0.5677730780845948,
 0.6133753944842875,0.7803341510234983,0.45463162386516964,0.6809332411331563,
 0.33461605747403267,0.9703030892028154,0.7041819903049806,
 0.29990990389918826,0.9727251991912002,0.6791629564421349,0.2440325041683189,
 0.8219725431899755,0.9253952021233922,3.9962035311014166e-4,
 0.4959571828203283,0.7751107204451172,0.976331532912177,0.6400892956111314,
 0.21059687048294762,0.27288105482618463,0.39244074182130173,
 0.6093089870167573,0.07093650165467824,0.28896181609999916,
 0.2959862189077267,0.08169225926588708,0.3406532377694398,0.3313370205194125,
 0.6261122058242685,0.7092846889245832,0.23630487561223867,
 0.026826442875043766,0.8407073708298407,0.1430079461109235,
 0.3404188317807644,0.21063133076235951,0.6972027239530563,
 0.34195435135494945,0.7272238938678587,0.28655100289771784,
 0.7830566678487392,0.9517442830897245,0.2639150633134144,0.9857415148776423,
 0.8335044589042887,0.7687249950294376,0.0034707757896366154,
 0.2118993473010955,0.1579216768965368,0.6772417586227122,0.5329597657533792,
 0.8491202460168644,0.6885059866923466,0.006304294383243425,
 0.4479969201998413,0.4856283708302268,0.7203327540005642,0.9985694077923069,
 0.5006583974499067,0.21940293070000938,0.9384562946803937,0.0894098361243344,
 0.1523420195188515,0.3894206070678262,0.867977828055738,0.993316633665594,
 0.6909019137033866,0.5125006782103346,0.3117519954770628,0.443068524267932,
 0.5213423814063347,0.4073764162086715,0.4699525946041132,0.25867489357220563,
 0.33838811088982845,0.4307262509130304,0.7744788571254572,0.9707642348467871,
 0.6411091301955972,0.012765967395563904,0.2906295675735716,
 0.4214591397302949,0.9183713324870454,0.13977126976296972,
 0.08747422733710342,0.6793296100153864,0.4251244800952815,0.3318931768218819,
 0.959905176646095,0.27565068391627134,0.5183944645554367,0.2907128793624745,
 0.8086329960836685,0.8734258016808049,0.9915086050103994,0.6844977065271487,
 0.1775026323132013,0.7252381862267301,0.7611565447509935,0.1829709783685154,
 0.21740026937837398,0.4662545916775478,0.06191150156750502,
 0.2798826944149112,0.9862932056769569,0.7805228553339719,0.683247947563382,
 0.8718343807464874,0.3959771779445126,0.6696426020636059,0.6714666649177248,
 0.6696026231420389,0.33700349481359493,0.3037081494343389,0.2740855289430484,
 0.12014252410090509,0.8538917207692912,0.9702200109301148,0.5948069261071198,
 0.9771818606709047,0.9847258428961705,0.13733758088785653,
 0.18632724454831395,0.6079943049652986,0.5184098665108957,0.8956202284054975,
 0.7570104436238776,0.7875581048524227,0.9957453257752031,0.6847998498607742,
 0.28263860497060156,0.544119166742455,0.8297208069182248,0.6625890758866742,
 0.02814347826414343,0.8102044267743062,0.9655622269027021,
 0.36447043921690525,0.248701649666496,0.3476178076025781,0.3384509299138412,
 0.35592837970752345,0.7563334048570945,0.7553806196667467,
 0.22941173690231498,0.18862708657007554,0.5671402432290761,
 0.49760221002123795,0.08933536993437963,0.8687274512980226,
 0.5869758169632379,0.7722161135757881,0.9791561544117858,0.732623106209729,
 0.2367214999256686,0.6933509887294245,0.09557555040536747,
 0.44082983892947303,0.03312758931841808,0.7629916330082126,
 0.48244415298999244,0.3273526489883012,0.6169585028572963,0.9862756500908036,
 0.4470604738669596,0.16712588058864364,0.2647637870702565,0.3199518630623962,
 0.8706614138889853,0.7649665390183289,0.8594510680493052,0.08203059562432113,
 0.12584437315380848,0.8760082155945988,0.24927326105345327,
 0.06642505607445837,0.045194513743525566,0.8322975775540843,
 0.6009831002867324,0.27554464929609934,0.8944893237386544,0.7466215264710817,
 0.12252808724803188,0.7115392091998161,0.7669403436365676,0.2682133842322689,
 0.3944232565996024,0.03459851410819337,0.6280732467672279,
 0.44847835457986085,0.30760824819383237,0.6901158277748202,
 0.6495920584242063,0.6792678756581967,0.8315877768794385,0.5185097118981328,
 0.5446592190308209,0.1876839443709074,0.4684227529430507,0.3726510159309091,
 0.7718659535135199,0.15287406961210315,0.9003103305615682,0.734778506943521,
 0.22496516231847397,0.663375924225291,0.8748455855844942,0.9454347693107299,
 0.024189368290491453,0.7716306812551705,0.7721560936705574,
 0.5865426328222243,0.37219896283837217,6.840167215589332e-4,
 0.7838687893567342,0.8557010714353015,0.6384806523008814,0.700342366423919,
 0.7979402711321709,0.040050234219994874,0.17797731236950476,
 0.5655569561686875,0.667293889503044,0.3967275199932744,0.3706428791347236,
 0.12008958012804205,0.456667313966857,0.4755967923673524,0.47568236222007054,
 0.25721127388766674,0.9003744502592489,0.5023060711136867,0.5277200649650133,
 0.5691252943215754,0.5797388371584498,0.3924674249077741,0.09652382606825016,
 0.012432887472383025,0.7155037801282323,0.6465920698371007,
 0.7286040247587687,0.11559483092735301,0.1348817263503438,0.2725116355572622,
 0.002806036207876894,0.07861455618786872,0.6945952381208188,
 0.5077224081760057,0.2770417745121717,0.01119393690194781,
 0.16263934300292338,0.759570421182222,0.5815800468031809,0.15544134354227612,
 0.44362197903050227,0.9015481117100175,0.8874379364710482,
 0.30225456617771784,0.39270740333453535,0.06116178920168858,
 0.11161384890579751,0.5940579621616732,0.19639320758548706,
 0.9562247156659346,0.4168036909024174,0.42503356516461255,0.6413637986775947,
 0.922661273670727,0.7058721794474073,0.5966552463821063,0.6979868232143446,
 0.335602968772134,0.7618771578825037,0.5269396237408888,0.4569534390549389,
 0.20990569626400402,0.29975878404739953,0.4797121052172948,
 0.6793149783987409,0.705482113055975,0.860760111093863,0.5393035718646279,
 0.04354980399001285,0.9811795538190845,0.9745764681191467,0.8645536761541682,
 0.7403471044499672,0.8596764424422085,0.5269481319505509,0.8940475973208213,
 0.1501526529967021,0.716871116525021,0.21626353577530888,0.14140413098752724,
 0.1358464422052421,0.9734511916622042,0.8104467023357067,0.5546128939280209,
 0.19881991314297065,0.6608000911813336,0.672936240265206,0.0355013921909455,
 0.45025698734729747,0.4168327208031303,0.0966527045778166,0.9045946369031773,
 0.5261011354405353,0.7996203364369548,0.45680879671676355,
 0.46834352671463875,0.12156539424329826,0.49320299235545106,
 0.058937577027945665,0.31048306112177304,0.007570367903294017,
 0.2983538219419237,0.29105973166990584,0.6454608694007951,0.4047710475550571,
 0.43358759175991446,0.7085963405022068,0.3226435149381248,0.9534375530186862,
 0.8870588477818896,0.6254315100544877,0.9554853617351825,0.9512864481309209,
 0.26952443503437806,0.6302507729176663,0.30112991148347834,
 0.19628571421755292,0.38737809127246436,0.7876675819932424,
 0.1538967423344485,0.6879665644759216,0.18956367801884588,0.4987716381938382,
 0.13568886967374882,0.17994170955124733,0.9137615514971029,
 0.9458762083779315,0.8743443256394499,0.9951157749257404,0.19869251824292666,
 0.5861409105092006,0.08146149858274154,0.35679186032138,0.7622932299488616,
 0.5433427560678409,0.5349768288403153,0.9009814952096364,0.36271897294086963,
 0.023041130058979786,0.5783036889441358,0.8202218735774351,
 0.8457917355537529,0.46131648680755666,0.48710112585477117,
 0.5673582903177149,0.8439772854317532,0.4490106652185374,0.7934162049112596,
 0.06991418329892929,0.8366027350493999,0.005049712902055137,
 0.983354479138396,0.423601230866097,0.10359157167748967,0.844599118020862,
 0.3124436828130246,0.7802538005901878,0.3355558600171038,0.791443611475835,
 0.11895030940658602,0.2267899364906354,0.18027250093090952,
 0.3530987855506207,0.42639009569424036,0.973540948147337,0.15373832280491673,
 0.7046494300637786,0.6535818864394308,0.9193665954490389,0.2917691442286472,
 0.3685260849216393,0.5225527236900571,0.019654936927162914,
 0.12426211772349438,0.516623501095591,0.13580705005604,0.4089826541640412,
 0.48527738736620285,0.8681790357441131,0.4363209458457469,0.5761206720657163,
 0.5408889376553052,0.7829626448205954,0.4786052713690667,0.07227806615277743,
 0.09422791455742663,0.6240199028779384,0.8205175030840772,0.4753651723201471,
 0.13601512622771517,0.6130880520241583,0.23998540938046675,
 0.31440414354680357,0.8869818598290696,0.7017541346621375,0.8862935599244774,
 0.995712689477074,0.6605156865770734,0.8722318334894297,0.6873042500934061,
 0.4195782484513704,0.02056986344211631,0.4792517376775607,
 0.01993335065500912,0.3865831434653271,0.014904614052277054,
 0.30724180080549823,0.715981268212424,0.3764355342500003,0.16118088495169147,
 0.49194529501907347,0.08755797181592428,0.6943004368544451,
 0.9516967618672452,0.2589194401806024,0.1374481903764787,0.04205255835472954,
 0.5795026506759976,0.8450645638762241,0.5997831903271365,0.0507818178421966,
 0.2440668401108793,0.03235419473788437,0.779024940124472,0.4775096986763758,
 0.7358421282546607,0.5727571112933731,0.6758459523417362,0.9256213396461546,
 0.5803973028366516,0.8993560763650533,0.810848101427382,0.6010850264550326,
 0.5196063493463502,0.6432856770935621,0.8717079367400247,0.40693731930666965,
 0.49093568934133436,0.6952001753224167,0.6097798767895398,
 0.04933527441870167,0.35627530450330247,0.3508021551463222,
 0.8774101000586908,0.639077937694101,0.39579263851829627,0.2493750620767634,
 0.34704023509714377,0.8334897496680531,0.4895166899799903,0.5390389352681084,
 0.3038211862102258,0.18447096017839637,0.8736172352802077,0.8558495072046561,
 0.6082630483582482,0.8005120552649949,0.9747522585660323,0.7234940534473734,
 0.2152429871801229,0.5329154645719749,0.04996824233974506,0.3477356950710436,
 0.9052384075760669,0.6876588642462217,0.15799365917541697,
 0.14263978820174483,0.6594391365011083,0.7794897039755948,0.4797925781140693,
 0.14135738895938676,0.0050118042948530395,0.41353141554828177,
 0.747225728190578,0.23287929921165218,0.1363313394015424,0.8094107557635627,
 0.18398324896730145,0.4582284378955226,0.715950519251172,0.45991084069319355,
 0.7039507311282445,0.5460326796354357,0.38555714603411495,
 0.16977916316207808,0.33408032627797635,0.1716098326366432,0.192872277639772,
 0.7862616439472713,0.680480679027492,0.547007824057841,0.9773205845061128,
 0.09338825201442624,0.026536580476121063,0.9535574566871075,
 0.16297092955300574,0.2489043640952865,0.645542920878231,0.7584465214798907,
 0.3600981264601061,0.8087182247187888,0.8783110040526049,0.4255457440763297,
 0.39876629946020603,0.893315842928018,0.14964237949028814,
 0.0050375364562236236,0.5905252806432792,0.4624495442911274,
 0.9245195154507961,0.015561729934576607,0.5630872387419061,
 0.37197715013145294,0.42577432356239986,0.7817995484138288,
 0.8572532421168855,0.1689912108445618,0.598992020058707,0.3748726854699488,
 0.39490313493870843,0.5659837590412776,0.002795268082492197,
 0.6393577440350121,0.8783687492592585,0.09727218012672434,0.9527855143283283,
 0.43037178395631526,0.5386031683738253,0.7191365289013476,0.6529554709617991,
 0.4553532142397817,0.673486578655728,0.2272366827624852,0.7700071026753211,
 0.22197750167743624,0.4446844451709995,0.5393052402577376,0.3156824125759272,
 0.3044005203264327,0.977619717652347,0.7401176352397488,0.6198709344234815,
 0.17543788118367243,0.7896831623707531,0.229237792043957,0.4897935065765472,
 0.8110198158440297,0.0066666713583789505,0.13205651510907201,
 0.5555838184819331,0.4858318499186465,0.15894709610453261,
 0.023346726458337796,0.5484553059074501,0.005075186776247631,
 0.754736714289828,0.3312018026110619,0.3342383350776439,0.7049152752009121,
 0.15549811328376817,0.27895932267666157,0.935041916566374,
 0.05353791467840818,0.13897653201828153,0.6237457828262256,
 0.4257615054503552,0.9899191861334281,0.5221468352894226,0.6089593585365781,
 0.2009307551830759,0.6988059054672413,0.014382253970751546,
 0.18177791324289405,0.8329151629842113,0.5782718438794492,
 0.13453532463572881,0.7169443973642539,0.9363035922118693,
 0.33375524783875476,0.3771837668772624,0.32609183888817017,
 0.30803555633140156,0.26236698463352703,0.7357437957356747,
 0.9228088332906854,0.47399469417506157,0.6315737797258295,
 0.03177680677012229,0.4477606127225153,0.49838082366874215,
 0.2560779474816537,0.7120687290834574,0.6068264907901655,0.34485724079670266,
 0.664089608239835,0.30630208510053736,0.46498813506299586,0.8680822032264558,
 0.9327606116200032,0.010366781174444784,0.5355272817476442,
 0.37051322725116986,0.022153908694859004,0.8258760011411203,
 0.33989358950136583,0.7789074350195906,0.520763908827829,0.7833186472127877,
 0.932227723533701,0.8547194644578602,0.2553259253154856,0.12197916081134019,
 0.8860173617421406,0.24466508131424336,0.562730621534115,0.7057782937367503,
 0.8748099679233605,0.3083968337329104,0.4597911214204867,0.06244195828450738,
 0.840386387181205,0.43844469991276935,0.24617309214850425,0.6150212182877265,
 0.6563522901387441,0.7414748564205564,0.10717515526362487,0.5382561880138825,
 0.17906966452449613,0.02767643914368534,0.14576725781054123,
 0.3108457145571837,0.1448338641278618,0.5923778658001441,0.7841613314947249,
 0.259898173427336,0.8340491572780021,0.282851643145701,0.21747455018431006,
 0.037941295341466286,0.03617543673808343,0.056114741295614334,
 0.814197498127801,0.657088846356228,0.47304503622778427,0.13865210191618937,
 0.9457162225032605,0.8626371215390531,0.535483130714646,0.19463618354862533,
 0.5159266524277941,0.8314790301977798,0.910719715005068,0.5677569588158283,
 0.49571938943599236,0.0755356714014005,0.2548865762727236,0.5813195042249669,
 0.9041500898789657,0.25351989975206946,0.20790789209942984,
 0.9189959327702382,0.39456251850839497,0.4939547121079183,0.6205550835968776,
 0.44894548979762616,0.03382128636359838,0.15352625646935925,
 0.7878005801475246,0.9305627196895786,0.06679278361328811,0.8461793693282474,
 0.4949920733066898,0.6096698940690539,0.5424192414858713,0.24732878756620924,
 0.3023938854091557,0.6509803400434067,0.10816978871055904,0.901141576063702,
 0.30697773021380825,0.5306850253818116,0.18282778361085528,
 0.3757655634284923,0.94038940911437,0.7350157424528616,0.21868900114524437,
 0.6403161217685517,0.6538005622645185,0.891958536546591,0.7273537225063197,
 0.8775688403263369,0.0662986548715796,0.793232117817826,0.33595708831496807,
 0.607442275340109,0.8533058377636692,0.8408828822496548,0.7700059431800458,
 0.6758226025549687,0.21622260549039896,0.3707438130854527,0.729071895531346,
 0.7281648710135441,0.44927272974748234,0.7566394502227042,0.358456990443207,
 0.3101010564205161,0.34789139847056705,0.31687180786021973,
 0.8742488766618434,0.13327379946339413,0.48219467894014323,
 0.5416599602703828,0.3882251880384069,0.6003202218216873,0.7822403283283226,
 0.23338467180762967,0.6950838441533175,0.24499962676053588,
 0.32627904086422355,0.8446259095537468,0.5431622641546271,0.5731269731788398,
 0.01294456997564375,0.5537386722929026,0.8586367652563899,
 0.08131251317272714,0.9733538743271833,0.17517729428062,0.27726426728884257,
 0.13060795042280993,0.2002465797415537,0.41661861359043284,
 0.6051158621124424,0.5438281376436668,0.7936123964358356,0.8215835313216093,
 0.12806588253590112,0.035188372106042864,0.29704220932051684,
 0.5817681788961485,0.48752306161508585,0.13712667766790032,
 0.08616037730656456,0.33184151895965686,0.7118517394927485,
 0.6609205652956984,0.46675347113863475,0.40910439899277307,
 0.6485517363394288,0.8622865693564542,0.3101678791290401,0.16814828201519283,
 0.7197326634443504,0.11048001129407914,0.2903595416207748,
 0.06477076058188636,0.9354296017320595,0.35788076771396116,
 0.6955971575265368,0.03949897057249485,0.40367676843519174,
 0.9025743027386273,0.8683704084217153,0.5250226656784338,0.7792118222173677,
 0.1630011939808046,0.5778239629063762,0.005972539600959226,
 0.38939949547862374,0.6973319183978862,0.7477790755382823,0.7787571826123901,
 0.35751327749425055,0.9250302826533895,0.773326602357714,0.22724239639761346,
 0.09135970541086558,0.034721345936085,0.7388750935239065,0.9566951350072525,
 0.9796609557577666,0.4734847056166287,0.7165312955203083,0.34798770151247527,
 0.7747870660828221,0.5571291819513385,0.9247002761510498,0.2355613682990204,
 0.6821815100801524,0.4779883754264036,0.9802878135780894,0.1859654768768101,
 0.4937706315419157,0.1771882962285939,0.4814914670549939,0.7397203680639379,
 0.2788922340044129,0.49632480090947195,0.6202106399832796,0.6531578724670364,
 0.34565164534198245,0.6357048708516619,0.6552458232458147,
 0.12411195606702374,0.48739452168476527,0.2487341064521431,
 0.4480137681692502,0.5444528559292566,0.5904753423546385,0.327652760355835,
 0.11888752464027752,0.4453797104466789,0.10910914893980339,
 0.3067752253683321,0.8075749113186362,0.5980139296839583,0.7201698153586473,
 0.2910156408882669,0.17693566100996894,0.7922787203504997,
 0.03216815907322401,0.0746461047961342,0.45430886823267747,
 0.5667322505511873,0.6762160960945038,0.8787340704552506,0.7076919101854907,
 0.8744102811315142,0.23115275554280879,0.8820377182885704,
 0.03916083937823367,0.9088592114194578,0.3497533442933408,0.9361792175839352,
 0.8645451465231605,0.1577834196432979,0.15710242251665385,0.9086810059924788,
 0.5960386636960765,0.5787433058202441,0.5739303608068449,0.5256452545226314,
 0.7420281739210731,0.7343747669959425,0.06802600078847987,
 0.02131620581063376,0.5308016193453329,0.31143445334384245,
 0.5456988175328267,0.755932352535541,0.3017474857303961,0.39209608153758513,
 0.10864758768006755,0.62186682512791,0.2604549730268704,0.42081941767571396,
 0.3374786104966979,0.34720916136276014,0.08942575287922161,
 0.8037594677315619,0.6372408676894995,0.843665486581391,0.28292303845415767,
 0.23585603327731408,0.8686726086625118,0.10949063275297777,
 0.7382615110760075,0.30210675442043944,0.4876890857050371,
 0.036723313736045426,0.4095093821045134,0.45856777816662253,
 0.5914967771279412,0.3730581632901644,0.9943544741272975,0.08893906838099219,
 0.9400583949167741,0.6855125768632662,0.7887542770502223,0.8720818188029416,
 0.0031368961181253985,0.8399042265378228,0.15098640348762826,
 0.1455473311423099,0.7556065716149315,0.7025717572179079,0.05989877671749344,
 0.5797400396606739,0.5718854904558313,0.440394420622527,0.6973669173911974,
 0.9097516564559959,0.10807579742975748,0.45478642119964796,
 0.6405083256515641,0.35217167773307856,0.3607706924377041,0.6153169259388822,
 0.5517651757733417,0.8146382329518799,0.4993443926578244,0.7549807770968732,
 0.010497491970740569,0.1452279949237677,0.8517353894100141,
 0.7833611711767496,0.9925328761947767,0.9388692878575333,0.20145981946916502,
 0.5072948485313218,0.8077348106520419,0.1183965483652697,0.3110042169298546,
 0.9256479632039026,0.05243358078327953,0.8763579053221983,0.9426021828446258,
 0.021181798738859614,0.6510831538515068,0.3475088557607007,
 0.5052500638697578,0.874752754403514,0.8581081463987904,0.3461294334255195,
 0.8647193850950453,0.832943126372877,0.8990225675973271,0.6911895772972252,
 1.3632703972654703e-4,0.5402018566663521,0.283919790886653,
 0.5015698385720613,7.10270964767834e-4,0.18197355562558593,0.639882289979258,
 0.45865861520827056,0.6003266768632295,0.8190835422645558,0.7890742415392705,
 0.7951567968614368,0.8642819445368973,0.5850146371553933,0.7732766687469932,
 0.2920250791709822,0.723196996148656,0.44242147524565045,0.16448564145350708,
 0.5932930061580275,0.8355770568091981,0.27284544628664853,0.71765472447749,
 0.19874023257209195,0.9005077928246777,0.017649079941246537,
 0.5278338990356776,0.0699946395301807,0.6922425045577091,0.9652530450489181,
 0.12356102530731328,0.4707956259576964,0.0127226714493901,0.3929160257950415,
 0.3547236517616097,0.3131729448656264,0.03645717497167611,0.8966039473800246,
 0.6034146262439679,0.14545846954836728,0.1820830844214395,0.2038766938298403,
 0.421179732521912,0.007026538874223998,0.8026174095996549,0.6712321083790515,
 0.4458194692120335,0.795217567276504,0.9995516858808011,0.26490635094003945,
 0.2232251006945376,0.6454129897668163,0.9471544004375121,0.4779796933770202,
 0.49990640841414347,0.763646493435185,0.12336226294590324,0.2374819482602044,
 0.8644804025369228,0.34965077014819834,0.22742885389655654,
 0.4966599523944655,0.6835069203300633,0.022584847424202636,
 0.7891426959318952,0.5423903717386562,0.8277766586101281,0.13386623652856633,
 0.659579548873422,0.15742657876009503,0.4130598918836851,0.38148317167114465,
 0.9798172292432792,0.46362267980214666,0.6040548517382267,0.9937444532829283,
 0.40519210812809126,0.3410476655395085,0.28349844680363634,
 0.5245401381577883,0.47116625852038807,0.5568292838017229,0.6307574829306236,
 0.2624488327375547,0.3344649206035448,0.21028944264532545,0.6684674348617387,
 0.21620277144062494,0.37024513259083447,0.06757072336183056,
 0.49114719057975875,0.9312546283622922,0.1947526793898554,
 0.20102921638513305,0.8758439416964053,0.3058602931700465,0.2296708224799413,
 0.8004065554167386,0.40442117411790945,0.49249472085582424,
 0.3418245170749604,0.46095192649316685,0.6970427900114093,0.5246887669604934,
 0.5641389950079656,0.4618004635535904,0.32304462300494663,0.9533579677238423,
 0.452222477465126,0.5904274262271237,0.6010768498627842,0.7712713188947331,
 0.8215241719972918,0.5690080506501398,0.26042220744625666,
 0.16815552917120635,0.6511400418725823,0.6055134320444246,
 0.08950465531967622,0.9806132004382367,0.8823734321457588,0.2246134517278262,
 0.3033812887034735,0.7091467197126566,0.6080687897593299,0.6841789531257754,
 0.06621464359637697,0.37933127764142494,0.14687352258786446,
 0.16057053544121658,0.03973452319034976,0.37866363951929827,
 0.26072012273649103,0.35128810419629786,0.6733145971418414,
 0.6348555620543702,0.6307260047197799,0.3645075133498904,0.46847637775003204,
 0.5637359789772065,0.5564107414596209,0.5889952394602121,0.8063409423645641,
 0.5017993121135619,0.8314259066082058,0.6610834974875608,0.5830775246728402,
 0.7386324793926916,0.29989396817042413,0.4474346422363663,0.4823041446496452,
 0.19714602954543337,0.36457305076960833,0.692847342646513,0.6559666982396831,
 0.10408962335271221,0.2398786389792723,0.7775366731150657,0.8249387631483721,
 0.21312968574688373,0.18338767180326487,0.40953485723786986,
 0.008694798980174445,0.883297490786791,0.9307295061265344,
 0.29300619163055386,0.4835040746266137,0.8489946928805244,0.8556934735655013,
 0.5482463575136154,0.08861522062555571,0.3572450800092326,0.6123967288789813,
 0.6990357381526262,0.5656791347947204,0.9922561658902185,0.23198545800514558,
 0.40074742749328784,0.6176095927636597,0.1130823504066839,
 0.30583134712931337,0.9652095249537084,0.17759253796283603,
 0.18523359389081018,0.701166164800459,0.7717601926796888,0.7614141613202735,
 0.8206005704226065,0.7150210439815811,0.762865247730749,0.8632862795340126,
 0.6587056399786632,0.6736266994693199,0.2249767075404625,0.34682385079430256,
 0.39203942332156305,0.5957883234417473,0.8474298676280969,
 0.47619281344307796,0.1825015873549063,0.3834881905063394,0.5381867584446408,
 0.8828602466080887,0.2578006087003699,0.28346052089285323,0.6915370823011087,
 0.8326010075199053,0.8950751500057036,0.6013978906935824,0.9075663431869023,
 0.20134056096589736,0.5029837264160386,0.030229506696568542,
 0.278066903262307,0.062445533390951935,0.8657807268459096,0.3901713845233923,
 0.08772257990099763,0.24222963539369413,0.09254414332794803,
 0.38272105014623614,0.09680867636839585,0.7588531782995542,
 0.19833429379842604,0.30532813534317915,0.7289586507521385,
 0.3003589940846021,0.7678287521166255,0.3375067617872064,0.9582100669753365,
 0.6520188071792845,0.39381206176898886,0.030403086111542965,
 0.9077121443853625,0.5947812691357571,0.6369243562352709,0.6925107585569802,
 0.91045610376704,0.4353588438372421,0.0446649495719178,0.2811537059622853,
 0.7121423928616419,0.8289801899167597,0.411404233704737,0.8539667503488928,
 0.8001955127728495,0.13172425042872016,0.287485743748911,0.03953413304906783,
 0.7372001347965889,0.3193954418852657,0.39799387504313266,0.7664624719530129,
 0.0027602616128046265,0.28331399402047597,0.33270041042235077,
 0.5840857529551825,0.2951156738725492,0.25798530246302476,
 0.18947827405964768,0.9129151818220845,0.5738017456970969,0.9156190651209531,
 0.08214199718899429,0.6993651967825839,0.6213614276098448,0.9987321791049849,
 0.9559783445393625,0.8149538884392215,0.9890998501867849,0.8149115381531894,
 0.6025698878936973,0.9726797714027312,0.7242523364760374,0.15575166060916246,
 0.727934329333553,0.038433991873226714,0.6200534685690527,0.7975329390644145,
 0.13156238598267522,0.5016581391826189,0.9418777034322813,
 0.14075213014882915,0.28331317015927737,0.4572589687297479,
 0.16731858767381547,0.5048888015758387,0.9394216335922626,0.9282607422487633,
 0.26733735280664683,0.7569565908494567,0.1592714107443466,0.776703295878749,
 0.06497672983120262,0.114740188680997,0.7650633146024174,0.42777972262683206,
 0.7282806249358839,3.0337455021567372e-5,0.5327268021940982,
 0.8069146363231623,0.4920794874155707,0.6025702844670089,0.03160553782653541,
 0.3280266732095062,0.23561565273090568,0.4680711032937248,0.2834195676955803,
 0.7673683898063527,0.5939750516067868,0.33203665902386414,0.5494557358359635,
 0.6423440497894978,0.1492655856819326,0.9741508728444149,0.2614759002277296,
 0.5540359951637293,0.8863139375612824,0.7646865896903561,0.0318192272564064,
 0.3706493078108558,0.7129787722607351,0.5968832535347002,0.3754749427769499,
 0.7895458533167238,0.8583693047309066,0.8080673031160197,0.8985137003451523,
 0.7292481754800311,0.298159198598246,0.9881688827088445,0.511783049292896,
 0.45667624385874794,0.1684570400330675,0.23160396120626825,
 0.12911162776009766,0.05156861629548315,0.3493016588332868,
 0.09675273221252434,0.06620599863761789,0.686479479595128,0.8671735536809559,
 0.8043454338273518,0.8378965466764421,0.25858504891818246,0.5827025609926929,
 0.27495176065947613,0.9449213703485877,0.6430049171501608,0.9163720092978997,
 0.4711458989213282,0.9571314997696293,0.8771730211966039,0.5344283118027233,
 0.08732010338431262,0.9678553108370541,0.5098593832493624,0.8223159596220486,
 0.6693502204665318,0.9560986336935163,0.9423113008420495,0.9438721336233757,
 0.19632789610913892,0.5695191371478717,0.3435328991408929,0.7865227919579957,
 0.3700296307029066,0.6065678742713971,0.5079402468451685,0.15207078860704804,
 0.13191985233814196,0.12430606235822095,0.2573257829677076,
 0.14091604626366738,0.8874808875876946,0.2993121370051679,0.8160205691063427,
 0.6607947502187371,0.9683444525995498,0.953117459333903,0.12214828195318361,
 0.9824202377433666,0.073375792265165,0.28835231586638055,0.3891570459662429,
 0.6861046307688003,0.045534570737734814,0.872964922740384,0.796038421836481,
 0.3898575415495614,0.9820599419734162,0.8937496442499645,0.2016167460967856,
 0.8971449309515828,0.029630046052471104,0.9798797931586416,
 0.7142608925515282,0.25359467282001014,0.41706949935401727,
 0.36248461863318227,0.18806819631115745,0.1577203963188949,
 0.7867929363549552,0.17646161258239856,0.9801576595129137,
 0.28630037342747283,0.13289352926934261,0.6496808671126438,
 0.1983881343640188,0.29055638376675086,0.577873973062476,0.7064538903419666,
 0.6064469280715256,0.23976012421859494,0.23760472718805714,
 0.6397608512158207,0.5141403094849537,0.7022799379460605,0.6427573795088597,
 0.5677923892666571,0.18980475318514856,0.08702446128061792,
 0.35410585656235305,0.9399047911940049,0.35259964272595046,
 0.8257656191896798,0.8855589713465761,0.9722220127715187,
 0.044353938167033924,0.2781949433048819,0.03646873376488968,
 0.2866998168157645,0.11424299036484309,0.2564391398221292,
 0.017487144883686745,0.878852460550179,0.6685222758192411,0.8451945657678597,
 0.14225946221511176,0.9707481320980966,0.610071366952884,0.48175803934990347,
 0.42575286774604626,0.7842193596069442,0.6374775592904875,0.6018532352327566,
 0.23545434900784756,0.7815870785469178,0.6212117285756166,0.8049795909028825,
 0.2518144104937139,0.8785956965806531,0.7023268325219904,0.2562299382297867,
 0.1808824174562823,0.3455136541488485,0.2160614288811762,0.04978009194330779,
 0.6831680789575241,0.9773198197549927,0.7990594676360326,0.07754430751904617,
 0.9009377384135935,0.49129562193113985,0.43735810996578794,
 0.5876452846224611,0.22678761690172833,0.6928022948944618,0.9003216751786217,
 0.9668263402422319,0.6521917889590185,0.9093942156620289,0.7302269456739999,
 0.017565302654093173,0.5662870498991722,0.053193884275915204,
 0.5239899068915195,0.05593112156104818,0.39732550162701386,
 0.21351153218050967,0.32749404047631814,0.5455246087574513,
 0.8113766039262236,0.39066667444998027,0.23376918052725637,
 0.29214119293856566,0.06549641553816288,0.48964851189977887,
 0.5190537107625706,0.10696704794132772,0.5361584794879388,0.8050684205316118,
 0.8363341165452215,0.13519891293333608,0.5929780511470973,
 0.25573550759361197,0.016509604370230524,0.5146223207522351,
 0.0729237460390707,0.04707717510240983,0.3624082676569762,0.974394383803526,
 0.7535914454296023,0.9814038097436936,0.38141071059771114,0.5364941111206618,
 0.7804416914065526,0.1694574498423027,0.6421162088983104,0.04053172998053001,
 0.8663283320717312,0.33675932678119425,0.8474225928104061,0.5402447908349901,
 0.15211925061455212,0.3630167586809986,0.4361241034639872,0.8989875816022085,
 0.8350952213507707,0.30571630824844087,0.574327446669442,0.19841130560029185,
 0.546443180377129,0.8226274336085976,0.7582360387941761,0.8104777819137802,
 0.3422921161055349,0.004325873767866306,0.600916739711441,
 0.17505294440779906,0.5537752622902427,0.6170489655791338,0.8786660949147366,
 0.1579450529039641,0.6343359884114379,0.3777940051362262,0.10323190140004967,
 0.3793627027251478,0.6237195367516075,0.0712774713627159,0.7313289232358003,
 0.7302942077306719,0.22007345291253877,0.23661478745911058,
 0.29132708084543146,0.6836579718415231,0.42155216374545446,
 0.9959066091499874,0.12146737217440573,0.23795855810181932,
 0.5949903620951812,0.43631482911830244,0.6942645416167285,
 0.32977229447491774,0.5483075451218075,0.43550085523567317,
 0.5947851616887827,0.3352226864464737,0.9138860206541488,0.3852628523755507,
 0.5287679137583258,0.45759609910420906,0.843139226035159,0.1408866061375723,
 0.5404199496615375,0.020783514192199926,0.5945897149684515,
 0.9014532763070702,0.7348552281281229,0.68423791533745,0.9024969900758897,
 0.028148673030392923,0.29916992178001856,0.5725026365799611,
 0.34986291176729745,0.5278585671903119,0.5667945961191208,0.5052378179417608,
 0.8478604590545156,0.4334934135262336,0.328388191926802,0.5192679136152395,
 0.9706730390318968,0.7760718417534398,0.6876308201437572,0.3726399986014999,
 0.2752185014541697,0.3699741526264132,0.8048933892234585,0.3738595517063219,
 0.9349580861352907,0.2936434883383774,0.23185860498213395,
 0.24477185909818144,0.21125621074533318,0.307329867557372,0.9793808667810753,
 0.4353247462667439,0.7343408143897919,0.7136006879841386,0.17841677814638346,
 0.6808711541539192,0.1820179530226591,0.8543495199966739,0.32825488963075455,
 0.6468500711668619,0.07753370144893612,0.43655915086540975,
 0.33782109867987187,0.732622140197754,0.7028011526579049,0.5606891808006379,
 0.9740186824363385,0.30938374521164114,0.11386026325155996,
 0.6202537140204134,0.14551230868559384,0.10249342706377174,0.82516498104338,
 0.8522944534860468,0.7567260290470041,0.2258492086465127,0.35444913309498494,
 0.3063996370098798,0.25171495823457346,0.6668212649151068,0.9773278966296812,
 0.012498552625721748,0.9127642345243132,0.7626458181103332,
 0.49455498094588923,0.10416325734169374,0.1007235602474682,
 0.08175983286561206,0.16926034482722185,0.2913048184063963,
 0.13061033480402928,0.42563163326548437,0.08507894408841032,
 0.2508855833847198,0.18826134139033313,0.6117831375171268,
 0.25575368840418167,0.9522094757857475,0.7936138868762995,
 0.07537616232559441,0.6544093807931647,0.7706349978679541,0.252097304192323,
 0.4501191444278255,0.49401067625551154,0.29560057035212217,
 0.3665480231456635,0.8176447258136189,0.89034782333842,0.3127108457336685,
 0.5885463347389401,0.6011757219557144,0.4694805155335834,0.4943183422821251,
 0.5118052305579583,0.8354116997909262,0.843382051876917,0.37893043472646926,
 0.08045558730464186,0.4086464041942861,0.749306556053471,0.5030727364975115,
 0.8410323754680613,0.27595339586477796,0.49197058956048356,
 0.9287240835618814,0.7502871400696204,0.9676470517945148,0.2906732605347869,
 0.23911629471283558,0.41895677607613013,0.8101295686212068,
 0.7742417015099146,0.5231279279310896,0.1894229826504329,0.7986977161847685,
 0.36407196127877095,0.4159619074324765,0.6799183543665652,0.6143841192100606,
 0.04071563704691217,0.8232690135795605,0.4021246175576898,0.9468298542014542,
 0.7104679103178279,0.36542798248447106,0.6453641562560357,0.9048605818290006,
 0.08159603241246638,0.023752486137662454,0.14322279198219257,
 0.44384524568037387,0.24631583913805555,0.3279227249907879,
 0.3456345362065062,0.2609855766907001,0.35114800288153136,0.4812960682968196,
 0.42247114256239393,0.7826791997738541,0.9400885706011533,0.2418887034465883,
 0.1473559464928793,0.45369623864662645,0.2738940114033048,0.9397593814733759,
 0.3690543844432914,0.9783059912424068,0.6235711203034754,0.7329284075822711,
 0.9281429915209404,0.6585737179856885,0.6156763960475794,0.09992288584710995,
 0.8511319686913464,0.5419174700948729,0.5163465341939812,0.21387472697295828,
 0.9550697198483404,0.4941920946534757,0.24629237384931046,0.0650592788694272,
 0.8775689615220708,0.7742994468671754,0.3580586486499916,0.6399781429631837,
 0.38866897121921007,0.8765727544685789,0.7361013210152083,0.811995614659844,
 0.9299267955996622,0.2940194238673157,0.6245676802037027,0.27581555932192414,
 0.4269190156213184,0.9531678619614092,0.9215870388919144,0.11034946296017856,
 0.8170580466952524,0.05416300262519047,0.5950683941035994,0.7867950754742623,
 0.45063133375062714,0.5964084181838898,0.19845445092805747,
 0.46032645044929843,0.8749003016974216,0.06892159852093815,
 0.40103807090353605,0.8088750286855528,0.9004356329413498,
 0.49234563073675164,0.7737085842087501,0.559940926096797,0.44753146823042744,
 0.7642725069319886,0.44008345717359665,0.44446816610320417,
 0.7163822682374383,0.5008658618149892,0.1701566319823259,0.6667484578488314,
 0.013010847222147959,0.979881307799609,0.1530983722752468,0.2170439376320259,
 0.3094197468978249,0.38550885957425185,0.9919697730095114,0.7042736632182486,
 0.6111747269524406,0.35336964691227535,0.30754499539547475,
 0.32381702290418213,0.014570852760529718,0.9331507361785454,
 0.30290213479862493,0.9247781083680707,0.6103259320532456,0.7369432678489248,
 0.5872312145648378,0.5316652319105094,0.8583860435935134,0.6007050917201001,
 0.7671003064102283,0.5893236063159984,0.4405447915159879,0.12303288321309269,
 0.6819676608427675,0.8673794591801958,0.8542529788917542,0.9314240445685052,
 0.781263761688032,0.024989011876630385,0.4074107823017571,0.0937853555352347,
 0.8286996761789134,0.5691243896210875,0.36570940086975434,0.3197810588968347,
 0.2498369584754505,0.7396066490857401,0.21528698225610843,0.6780934004478697,
 0.7055237312580817,0.4542816623075071,0.4431775703303862,
 0.010830308074798012,0.6769585854743346,0.08972531901621073,
 0.6181979121206417,0.1821075916738708,0.06370980084716871,0.9861164822136754,
 0.03781953289432316,0.09284444844008055,0.12835552022427932,
 0.9488467412939713,0.2920194425632907,0.0012206811510218074,
 0.2757683224019194,0.546415509031323,0.7027882265948541,0.1307962571806267,
 0.43545425432005613,0.17779956439228173,0.6142709977653708,
 0.20732064537325123,0.8120127719991769,0.5229289228053537,0.8360497157209582,
 0.027437530204506455,0.6157924192055163,0.7408446638046026,
 0.5256727176205244,0.5425733901311358,0.05010929633963457,0.7362141886989232,
 0.27677361227754893,0.2756780879489157,0.30013769815032165,0.561507396316179,
 0.10172682145386613,0.754275362315922,0.05532629434799918,0.5025088667584858,
 0.9383372461825841,0.005870899675330499,0.12914348483383353,
 0.8659342778636356,0.9425366986154218,0.8780023433411573,0.13928067425260993,
 0.28909289583066955,0.7722440113244599,0.2576661547116311,0.526244440500953,
 0.4819936280405992,0.21009142672969627,0.6678142530780471,0.4879713438009411,
 0.6079358653133998,0.6910531548121138,0.41436944985929847,0.6414826141899544,
 0.6318752217848956,0.14222740755959118,0.9247432432428405,
 0.36861007071626883,0.19491341165941067,0.431333979608331,
 0.058636541271160136,0.5739726086930352,0.3992472971938915,
 0.5486580089094181,0.589803510779152,0.349038977434572,0.2698408157477695,
 0.7711277685599073,0.9820470360492093,0.8088534216347193,0.34107889651952117,
 0.2316152615575523,0.9290425018243538,0.052297280047170824,
 0.9838486192346162,0.31788048043587214,0.8653475328789639,
 0.019531592479206772,0.8680034597918381,0.15285398816008455,
 0.9976782687222538,0.3297517245284667,0.744140670035598,0.6162587397865098,
 0.9889218604345413,0.20445033283893344,0.4364919634592962,0.6423495326219921,
 0.5832431868467471,0.6228600248985676,0.6012512192043264,0.803987458651872,
 0.19101814865265143,0.13806245629462766,0.45538116142315177,
 0.23846227200682724,0.5925722025435646,0.5256946720832251,
 0.32394792498077163,0.5581323535334516,0.5963752151749994,0.6315299172972921,
 0.774284124511276,0.02701637499713605,0.5220431091397962,0.5047936958724926,
 0.3573428663511059,0.7744257822305771,0.24866256487581317,0.7024720102284738,
 0.9232419638465166,0.8270728961051933,0.51962818175265,0.4444747931150014,
 0.656549021279131,0.5889056195196145,0.9352377978456379,0.6751296039016553,
 0.9517953753476585,0.6499896191653194,0.5941469434908442,0.5132142899980733,
 0.7936834976296354,0.010484617393814566,0.858154857287554,
 0.23400353901463167,0.3834960303550323,0.1912096228155884,0.8070969728120821,
 0.24027767202433492,0.32038610635721887,0.7684213747493636,
 0.3983957805245516,0.2424174201458661,0.3970780009084399,0.8575562754535455,
 0.7264988516021406,0.11560096019197541,0.5033431891802685,0.3229030797075274,
 0.6964981862618048,0.8131133491520621,0.5352820902973232,0.417665155205604,
 0.8583746029730576,0.886260292286139,0.47282834201485024,0.5739532823639009,
 0.04527292295635488,0.5224257502433274,0.927229503248911,0.2526062641290996,
 0.31336523020253015,0.5479629489285579,0.40347738949011025,
 0.7132547118030884,0.7927163477934067,0.3297197934261803,0.6397293610958162,
 0.9727739096882354,0.26770443894921536,0.5613672432044817,0.6176897510753364,
 0.3217821269120217,0.8046668542735572,0.8083748406874316,0.9056307928456508,
 0.2779468780144153,0.9082976900264234,0.47320446852227227,0.9967689548299006,
 0.42220872337089443,0.2572761999056896,0.6706249883881731,0.4543730341448534,
 0.7578972755552529,0.6832396207108471,0.10653218390885155,
 0.47570022153090896,0.11786787096779971,0.8830858699738638,
 0.7833534771660597,0.8623236055280175,0.8901372035787474,0.7278474434758412,
 0.517762851864219,0.9665767694689521,0.12358860309608577,0.43967460614845444,
 0.4443176216800523,0.4503654584382709,0.5821915381743444,0.7122897752841681,
 0.46286269919068557,0.9628184236681075,0.15525987194729718,
 0.5895514427313104,0.2636978848056697,0.5217703217503158,0.03464808320375057,
 0.8013276573570223,0.8542015754261908,0.8024964637867966,0.7518559566474632,
 0.0749985672452832,0.12169962258303979,0.09243396181264207,
 0.48893186572480174,0.2868714642894372,0.22303836538743327,
 0.24307496681256613,0.8882989092048469,0.15396633648974078,
 0.04687278144902096,0.5756016721716761,0.05126872077492017,0.635572945321278,
 0.8776015105887318,0.6236806358920748,0.31843258536023755,0.7298575500573227,
 0.7149208658527121,0.8704650988392068,0.1427714599954798,0.8374555588645877,
 0.5210399916627292,0.3382325775074523,0.7057216758438121,0.3442527554105692,
 0.7337532060722625,0.9367512329535698,0.8057337328139327,0.37545400034835896,
 0.4742192433363418,0.6538104326244194,0.5481873943156665,0.1775482069475307,
 0.054625134616688054,0.6158784980301094,0.8791638276519075,
 0.17042744420868527,0.23618353544119763,0.1436905439116053,
 0.4388133721697022,0.053599126407838815,0.30264954474476036,
 0.9658014866936434,0.8348066654370053,0.8528689434628702,0.31667982412136675,
 0.06847279299946574,0.5867108858252786,0.1642290069327792,0.8663027947927401,
 0.579239746906632,0.7564275738060052,0.827273149361918,0.9198992731487153,
 0.9790044521573459,0.551302098335111,0.2861080828884378,0.52483971472795,
 0.2725148782309966,0.785892763132545,0.9169511706382407,0.9152776646712835,
 0.6820608564196456,0.6685077444645121,0.9176891521808445,0.9344183942929866,
 0.8527036091351008,0.7406138648064124,0.42918088213389083,0.7009509255636516,
 0.6407159751837471,0.8716440100632873,0.4765155372158404,0.7654231986307735,
 0.8095437607248874,0.05814378647846241,0.02631976234669342,
 0.7368367792523806,0.17625284104822525,0.3050389684713668,0.545060838981576,
 0.6675573769243724,0.23349870471528378,0.4815648819860886,0.705906937540864,
 0.5723458730147339,0.07250383318526621,0.15475233409141087,
 0.8525829610905938,0.008189780007729519,0.6236341068837743,
 0.05660509908317268,0.14357698070347713,0.13503065627645894,
 0.39928938464894437,0.9703759582726267,0.9116809445887146,0.7389971439376595,
 0.9528675193528983,0.6450324955072038,0.3485920889365154,0.4572206475509044,
 0.7129579493152608,0.15403802216737406,0.992530546629356,0.41177903476918964,
 0.34929891091011345,0.27189988830618705,0.9892781464855304,
 0.46148649562880806,0.9185911075662345,0.4430865664175716,0.7159345670385282,
 0.4927302966327409,0.9207289956033451,0.5472585897106149,0.8578568037697071,
 0.07137634190640418,0.8355657347617428,0.6009450293059044,0.5657143741074785,
 0.40594631644718515,0.664264509741122,0.3482523609985315,0.34487723396222225,
 0.4112896591137214,0.3914162258620195,0.9635891838322472,0.6006029005535442,
 0.6297775100617988,0.5564506072751865,0.5306375509676005,0.6594476689742523,
 0.9370222173155723,0.21526334438759986,0.9588308814604205,0.9166188543662633,
 0.6789678384257729,0.479008882652323,0.36839478441758533,0.11798247877050683,
 0.6057030362082684,0.24830782922731265,0.8818274555565666,0.7144985048258967,
 0.688218112455729,0.3230869753478287,0.16701800756067853,0.7937855912806837,
 0.7040398926250042,0.29609310440697345,0.9119359895611119,0.7752083678982503,
 0.30282909366731614,0.09442518191244875,0.1719213325277834,
 0.4161807161817547,0.34413956743766083,0.3016230840124785,0.6539170303354437,
 0.28418376906149045,0.5101132251183217,0.9753723869860456,0.6055113800429439,
 0.07605429066262537,0.944481104167472,0.7588005441378859,0.5069533269543789,
 0.29264554884149885,0.028090404445635597,0.5458232893703961,
 0.17568742194503706,0.0807089436535926,0.743883364008878,0.9717069804556092,
 0.9582159932048371,0.5237994037140395,0.3827625546556491,0.33465454093837166,
 0.3987666260264223,0.9060555461257628,0.8235191809365601,0.34520092777239264,
 0.7278140988984558,0.856937368146961,0.1387697951213469,0.10966459580213084,
 0.24999856208382643,0.21461827598513927,0.164321483390656,
 0.45461192789963767,0.23434131503108702,0.14249544719988483,
 0.634655748461326,0.4429614169155368,0.9709311134579552,0.8640377413771121,
 0.5761457302539003,0.04096011011395262,0.707603427494905,0.5574124173733425,
 0.22591427738753134,0.44940109143846985,0.3537833128299639,
 0.1586664130065405,0.9026903725871325,0.692756233460646,0.5017561194306748,
 0.11418018496341731,0.3138109884088913,0.7110857846075261,0.8457694894490978,
 0.15834880393900175,0.4461455418014084,0.229350784274445,
 0.005293177084549683,0.9308728853468944,0.6629427508951482,
 0.7081535621383463,0.9442836926718294,0.8565496355775882,0.6751749640751046,
 0.5124254335691506,0.9681324506491366,0.2776757430991049,0.4114626138386521,
 0.7447812478638302,0.23325190535854512,0.6712874113541685,0.7536389556119708,
 0.27860653672193303,0.3603541460836358,0.11184203020270811,
 0.4599726413924108,0.15484152989521505,0.3694611202678383,0.7373413170961489,
 0.8370312699544852,0.47652430651899114,0.7733269872802083,0.7856430813286874,
 0.33391340793537394,0.22739105888686084,0.6505534341616199,
 0.43789705892430086,0.3206482279092232,0.12308902783627429,
 0.45278703126362296,0.10096569533675648,0.15224583366475974,
 0.16446003288159317,0.5073957458710475,0.6844428702308539,0.7771782043731453,
 0.5324502476236747,0.6839489813223499,0.6568382564870408,0.10910147596994313,
 0.9585802903067545,0.25691323957978374,0.9219649101495397,0.3070682031691484,
 0.9647601073288266,0.18226618212904522,0.8421523679443896,0.8445434501182049,
 0.5879172161680454,0.22555561984015726,0.34798128859870836,0.89630392853999,
 0.9412884443054399,0.4363478114523076,0.5781896037787041,0.9849211198342819,
 0.954145311613906,0.03786829436461703,0.6680155993182374,0.3546339426590842,
 0.11103131448368175,0.6518146878321465,0.12599128139938964,
 0.13381535831879243,0.3037189008976938,0.06891678781172161,
 0.32078454728937356,0.6821900642155035,0.7631438755753823,
 0.18691212076296104,0.5503067940649459,0.3668281467836635,0.5636012313252515,
 0.4923247130449684,0.49224702953797883,0.48080615747989697,
 0.8752340486948179,0.7687264163509034,0.6148717079807517,0.23415458465744843,
 0.4181052729079069,0.4414520835023267,0.6047453221235362,0.12960275357465678,
 0.019634977366678075,0.3539689793327825,0.9684134984175498,
 0.27120838259111346,0.29398777782770635,0.027331909582470337,
 0.8353747540730262,0.06739852975803129,0.8726460772971151,0.832017694690855,
 0.3756962279618087,0.6088779305213403,0.5439422610557642,0.2821416825739097,
 0.7973489004848858,0.6027680333982498,0.4915597179516368,0.6363816952860374,
 0.8604020763554034,0.2118266287237125,0.758211776077947,0.8615233169592015,
 0.6133919282975613,0.44325092507147623,0.769665174981018,0.08876396460422997,
 0.7074416130961194,0.99450589667761,0.17996948193358953,0.14080789500936608,
 0.7305311021234899,0.9950649242240392,0.777339994188622,0.9236998974333352,
 0.5160236815172338,0.2032030480857051,0.7704370721641202,0.6199912950895369,
 0.9031875079424959,0.8424827088733573,0.515601475477855,0.42171807038070697,
 0.6103931254728991,0.09543772988979693,0.7072101637626437,0.3022773623573456,
 0.06588082707525889,0.23592441295133704,0.9217114962992345,
 0.04954712300224229,0.4735223938189934,0.678314039761152,0.9665008507742839,
 0.9252110289171211,0.9019738901769212,0.28234120296719367,0.0343278223036263,
 0.24980837783023968,0.8726739122242168,0.9469889524497243,0.5962986749598238,
 0.5442998487454039,0.24192619573467966,0.8131060250604728,0.9703951967250166,
 0.4455538371032959,0.3121561228109889,0.8300000620294324,0.03712896176723102,
 0.4883292187997048,0.372852258167653,0.8193330054523034,0.5418461464724491,
 0.834414625131252,0.22815112754362277,0.46243288099783153,0.2913956053103157,
 0.8660322888768797,0.9325698968609922,0.4773522661478755,0.43152733868016235,
 0.3240128023661013,0.2879535843125012,0.1339577828469687,0.540009627537306,
 0.2371717080009058,0.6757707791723693,0.663498075389852,0.14530868769508953,
 0.8159375297528688,0.25492054713156065,0.21167965322532922,
 0.5820462142999808,0.585410889961205,0.9276469699889304,0.46414722154314225,
 0.7448484503332676,0.2208106983722371,0.49400775653651274,0.3072301082069072,
 0.8036136641298524,0.4532797635780331,0.06136012607464658,0.5781600785554517,
 0.6637237916148615,0.9458162621685573,0.5343333421423888,0.2632256757927388,
 0.662363416478123,0.8959156697744053,0.3798419479532569,0.174542465885289,
 0.27449111432711915,0.2517231339856414,0.3884193859936975,0.8874295496856669,
 0.176468965426812,0.4161067509206049,0.6653153554996802,0.9168816390171345,
 0.5064607206520828,0.005521982672640813,0.3520776080269422,
 0.8027067819272419,0.5235137687947495,0.4690403989278221,0.36454222713038864,
 0.2916695193089762,0.2082011253713416,0.24323071909170957,0.738343605417792,
 0.3722904683473047,0.23576908479100378,0.3211448159542296,0.718913330810651,
 0.9802045690387555,0.41539262138821964,0.393131219219906,0.7299635341907715,
 0.7075821874242552,0.6194645132934158,0.1025552536663199,0.8888941019821818,
 0.4527608585848484,0.7988194962442253,0.18811516576247578,0.7750630027925752,
 0.8226392578747319,0.3512825511751072,0.771737585834886,0.13519674344539523,
 0.6575186805671753,0.48089396445895793,0.24811066370236956,
 0.20101444718954486,0.8307062927918752,0.5874930460851535,
 0.36390798236996735,0.2265511635870302,0.2666671948306144,
 0.08733749069578378,0.9283549211144893,0.9877980993247281,0.9394201126531224,
 0.6958030190213883,0.8145263973635096,0.573263915128254,0.5555487953241179,
 0.19949277375302887,0.5157958593685479,0.8450357699167319,0.8688234207393434,
 0.013311377136586922,0.4464088680012672,0.8506105404160972,
 0.37780039257730336,0.016995386724093953,0.3646614263996132,
 0.044294778041668925,0.5257712077095493,0.32429997323832105,
 0.21052270667218587,0.1499319038366541,0.25349858220068067,
 0.7947558323175947,0.7766327070303389,0.690392891847988,0.1386107528275976,
 0.5482231288914705,0.8332112164827251,0.12098144322069315,0.7088848741230667,
 0.8402722796692046,0.3122529000456603,0.41157729568131374,
 0.05919588905705597,0.6525676476478252,0.2759751178659625,0.4997093404180555,
 0.8095407218625641,0.8006827163507397,0.2641426226393162,0.8640383379500588,
 0.9138731580131572,0.8537729803166499,0.41079443726118814,0.8534196549778181,
 0.8784722744266704,0.8811874582443345,0.7431294247459632,0.6837973589809632,
 0.5660336427909877,0.13286165298588393,0.9517777075695908,0.6762976736869637,
 0.4818495829597349,0.6340555919237021,0.4440408655588658,0.3482398938499963,
 0.4714252012004493,0.42680628846887547,0.1275880363152211,0.5143971191997859,
 0.844595864470195,0.24817015904992257,0.13241481946724765,
 0.05407843792791733,0.25476953209525943,0.33106607857812753,
 0.022972885165542634,0.10730337594328476,0.5355710836930301,
 0.5868865651365353,0.09457167949250223,0.3159880165736766,0.3043712847438651,
 0.384632601188693,0.7483318104297126,0.17357517972675018,0.4825326255053971,
 0.03148057143075511,0.9795365204106102,0.867630396256239,0.4626312308752605,
 0.6941960104894324,0.4739921392502082,0.7906870984281942,0.5875852709484204,
 0.8727222457949293,0.20217871907923135,0.7563535537951972,0.917116370840925,
 0.9389235923761243,0.21366568346260928,0.07567028610459314,
 0.22758820806921398,0.28158304534252077,0.9291192589381096,
 0.013071092039007914,0.6749150071697547,0.4578774566153999,
 0.5092892270768269,0.6617945067862318,0.6848861238346056,0.5329042866210687,
 0.48270746527592034,0.8769594682813279,0.9611116745350453,0.9674794050916102,
 0.4133207399343176,0.86934362909265,0.009763968313126092,0.064928774738092,
 0.42785513400714326,0.5009066116500037,0.9597161820927153,
 0.18760127486643163,0.5050528887440986,0.0303366104850753,0.5901324563470248,
 0.5417808101894286,0.4891583526517027,0.6455321953302453,0.29146791731789656,
 0.33673991484300236,0.3151566495911182,0.27991393244823015,
 0.4994899759078939,0.9164645818950472,0.08192596468227364,
 0.49011133098372883,0.17241814338032224,0.6849554294689402,
 0.6826374820371478,0.11302508311952575,0.8903277585409484,0.5672355028063322,
 0.7302136198181225,0.4080975293778394,0.6954513738100865,0.9148170837659846,
 0.409866893855904,0.45331208870524065,0.062375691042677484,
 0.3834135986640681,0.7487695891796983,0.1277359673267775,0.7189809695767286,
 0.10399840518367254,0.5489150623914606,0.2670293096334877,0.293912720294339,
 0.38480386152812973,0.3602820457397107,0.31523092696897714,
 0.13338750569911317,0.6876964138160824,0.4927646854023726,0.6009634951871349,
 0.8465357057515069,0.9919293567826314,0.468121086297544,0.23392290032047636,
 0.3423434497283322,0.016048178941365343,0.7443369221552185,
 0.4749364596333122,0.33278851926837416,0.5526258848882263,
 0.35454980676651604,0.062070207435426905,0.18773987402974646,
 0.1520800130013128,0.1774394161083913,0.40616519311729904,0.2658414619392774,
 0.49268056548869943,0.269889829940816,0.13352259341396677,
 0.38131645103120604,0.3492653444043665,0.4206411050875317,
 0.044454109320772095,0.4187223827023254,0.9391574585725485,
 0.5366288890077331,0.7313174510032244,0.7576184520556786,0.6018669846565643,
 0.37584621512989824,0.7410260573807438,0.1241446652234649,0.4606921894961218,
 0.7703743548516642,0.9012694410530716,0.19238980877477296,0.4673673599650078,
 0.09766054381541367,0.982573195120331,0.7425364267112192,0.07488350470159988,
 0.280312124374693,0.6609266760089864,0.9236401162049188,0.21044855391050987,
 0.8473885092313053,0.9283097366753209,0.28994798967552426,0.9556639406684909,
 0.10822306438042217,0.5153709599379621,0.4182708504363395,
 0.48060117104015976,0.050745363728648796,0.5628379322958947,
 0.6584188244155694,0.8471243983158094,0.816825460715799,0.38568127333525537,
 0.5743397338167121,0.893834188261139,0.2545423173808693,0.8756494744030483,
 0.536320670820369,0.6370344948214388,0.9587000013417514,0.3732991293324541,
 0.355474910998549,0.6152626235729411,0.7959907032006842,0.7897128617653923,
 0.655570324946779,0.7386130699788089,0.1360035193125283,0.8509927373299854,
 0.10448469703448982,0.20080332964757275,0.8264479289819123,
 0.3334474349755572,0.007530689971137594,0.04460268536485412,
 0.40087909433374946,0.7601162999937583,0.07598981646925729,
 0.7580362237229473,0.3235455725394156,0.9061120999411667,0.2937259006410844,
 0.01587215082246063,0.05513319749333989,0.9614132126842176,
 0.10428554630554576,0.11125787372274742,0.4394310615278034,
 0.6983441255350579,0.7503690257501111,0.3619023651172806,0.9772327982642409,
 0.9770917259532916,0.2641070724072,0.09189837154779812,0.24556970375911835,
 0.8003923434190447,0.9558624059569687,0.8884391727933542,0.12271775844910482,
 0.147218442040274,0.7096287586081627,0.3020743321269578,0.7241565299034466,
 0.8421020598845659,0.7584147099783826,0.7350242638059199,0.7466257619117337,
 0.17284902059628804,0.8662459612707557,0.3655396369529481,0.3197229437758503,
 0.06271002225065037,0.6975248497565614,0.3587257844333197,0.2866089196523631,
 0.7900206008770077,0.031210594887614684,0.32879456471179414,
 0.5539597178391316,0.3900426701556895,0.574399994650699,0.4611677534201367,
 0.980397895338847,0.10185897571034852,0.25223913868071557,
 0.47916023874465996,0.39745924308074054,0.7568168221511349,
 0.7228317148546965,0.5529834507475375,0.20171047466567504,0.318032205281223,
 0.2873021814697785,0.3716834776417496,0.6394782059006546,0.32247901486649355,
 0.33565832393679385,0.7923769695654315,0.9212529325604126,0.7434722477855413,
 0.8599972066595671,0.7288500407882496,0.061693869885645625,
 0.44497424238611516,0.36165283776496937,0.9936597926842168,
 0.9595119175418975,0.8129008199272356,0.7644672080315592,0.5703153449877119,
 0.9063020398570858,0.4297654628029659,0.4459220368288993,0.6134039819381238,
 0.5515543428665518,0.2759129711599275,0.8947245055544173,0.6783810089424112,
 0.34929177461455696,0.9084636541180137,0.8734429302219658,0.10690672469314,
 0.7915141963395581,0.5343408708064241,0.4341679554512279,0.14601626192696737,
 0.7752936802191299,0.995631740166462,0.4212359208482461,0.11992581640405331,
 0.8453108895623654,0.9456399442658141,0.8288177414991472,0.6650280169629417,
 0.41029899464107766,0.5057149840874137,0.9046593394273924,
 2.5124283784794255e-4,0.6212620213063966,0.4586816445060602,
 0.8411761529525354,0.680370483922866,0.2519486633921104,0.10836074671291018,
 0.2895239907014717,0.37494931714691737,0.1486606275218214,0.7631766895795051,
 0.24852042025236476,0.8012153028452597,0.08484334190343845,
 0.7232628632927107,0.5680807679505739,0.26043638977614236,
 0.36823545335600116,0.9884590557079856,0.30833475268732824,
 0.1558775726872108,0.051270775639843724,0.9859325824834659,
 0.5113135058486991,0.7780563449335334,0.23725361425538838,0.6703243714375724,
 0.2939964478255622,0.872310979486574,0.7340539593047806,0.7895425658909532,
 0.2842631970683136,0.3201151932067864,0.2713309398524427,0.8963945318149392,
 0.7973277956770287,0.4199351338279873,0.9690301230636377,0.7021374687085902,
 0.8481838140696847,0.1419492404920344,0.2768405360996764,0.06994814614861844,
 0.04041238965880156,0.6200369162421465,0.5821266189838998,0.904940779927279,
 0.5739490682116865,0.6043408446136076,0.6388216113996309,0.35789740569761586,
 0.7356951813471536,0.11506570279066197,0.10865908949004799,0.368927588703726,
 0.4374363054603455,0.7672508800573055,0.3532777902436639,0.646529658273961,
 0.9964554301416553,0.29948873870331083,0.9056819384454209,0.9756693521653158,
 0.5336633437057393,0.8769064502674192,0.1464886189367064,0.141918162983382,
 0.38629147750210446,0.918909772317587,0.545341356289787,0.3466583006332744,
 0.4588746580376286,0.23442222337226304,0.27776956487611615,0.035592341560715,
 0.16520395205630967,0.26382413016055317,0.8665384538346944,
 0.40699554299939456,0.8881274364368053,0.07092189878478461,
 0.24786112857039377,0.12240937999158552,0.1778216523505738,
 0.7395118285715686,0.6242302211176315,0.7969606742347237,0.7090169789478715,
 0.5655776600029453,0.48108223984427845,0.4534873052822588,0.8943132171099853,
 0.4432310581274549,0.3250768829423779,0.7942620086337419,0.24896596875294486,
 0.6633628410615892,0.38767795209569855,0.6694487875283314,0.1366916198952468,
 0.3927641707352413,0.24109305753578303,0.27094650404836373,
 0.31537607766005504,0.8459267250339347,0.8973764835122298,0.4083441858662802,
 0.8921293404872245,0.9119890067212191,0.031536125599308296,
 0.2671881552988168,0.5690208452613271,0.9208553590369151,0.8161789207271113,
 0.8872338275415084,0.18908460747949885,0.21326764519818475,
 0.42945178243920334,0.38870659572915167,0.20430601227763967,
 0.9101303678715185,0.26561565293393885,0.40815592531903844,
 0.5274382942213229,0.0688550634898829,0.7621431679160903,0.6598157730874511,
 0.9805189951321956,0.20763092303397312,0.04226295167977978,
 0.4843121647005725,0.2879529493679558,0.5850409321672516,0.1340566050108195,
 0.5258160611532011,0.02777823327078366,0.18784668586467856,
 0.4748752245602339,0.2325787592976455,0.10444654640906581,
 0.10999855558330696,0.7180685425847584,0.3569540612819877,
 0.44665376315630034,0.18839134345851538,0.3352853148024192,
 0.41603778229162924,0.01483642175747557,0.1242499658172096,
 0.0640699226438286,0.14891835340687387,0.9366337750875348,
 0.20165443331449717,0.5319517744820094,0.5253054546021104,
 0.12199347028504504,0.22437874215314724,0.5115829007270088,
 0.5883635238742047,0.730671444556877,0.4399982824179749,0.6922623692856009,
 0.3760967132426313,0.6797329131179334,0.5042905044995014,0.5904747637864973,
 0.2363390567208783,0.8229157636032822,0.8662950478108629,0.1525515122308585,
 0.4699069613432878,0.6897706334124342,0.43556122904345695,0.7352461005417812,
 0.36177001921017926,0.26744560647500937,0.16412064467681342,
 0.9702977230032671,0.21462339891013626,0.39202624582739287,0.705196785593997,
 0.2687064395214548,0.7459409973410345,0.8666051963452103,0.798222266835251,
 0.37033053540175975,0.1773991210200252,0.7067343164719004,0.6665006943973673,
 0.8271876600498302,0.7415882925751398,0.8816250825482197,0.5000238560712369,
 0.5038121679578119,0.749186493315988,0.37046288984248543,0.42420101443182,
 0.6694317873541954,0.8841214025902409,0.6692696263079845,0.23504397775000974,
 0.8974684670044506,0.7535081367270937,0.6994629695944771,0.8825546714442082,
 0.9251108281520083,0.7481983738607997,0.13339997548168436,0.7034417970949141,
 0.3804280698128065,0.27980448113300405,0.9178432757134706,0.6378964159408081,
 0.6948900567747918,0.9985076271024875,0.4014733183069713,0.46009953988478025,
 0.7652753434377102,0.6192727956509456,0.8028218976416551,0.4576566400659723,
 0.5384431951709477,0.3610083634122182,0.7475962972090822,0.35868742264405196,
 0.8624021182265339,0.5934420502804705,0.9304907200679423,0.5173021899248293,
 0.1278446219318976,0.7260574904982082,0.15111050461818554,0.3145059039376169,
 0.04237865044991973,0.7694961583536204,0.7971349542637127,0.5395679023466063,
 0.0900505019317186,0.9204581150300082,0.8130857685303922,0.6477796346702293,
 0.2467248220698527,0.4598429462271929,0.4129525971879432,
 0.001672128543468987,0.938393775510627,0.7503021690015513,
 0.21098135342966673,0.2778334660273002,0.5259321490394074,0.4485314975380811,
 0.7478137756040942,0.600074798308464,0.0741397363127041,0.013916494186778428,
 0.680778320289806,0.2011052025636244,0.3764194235692476,0.6862019222938283,
 0.27142957068314877,0.6482554821268985,0.43051226041332513,
 0.32819820366595387,0.12580068723708304,0.4510866029367062,
 0.6595474162714051,0.032596145520186304,0.7476884322947355,
 0.47647620721685036,0.5722807753185452,0.4475778756112532,
 0.05089254117483755,0.04378617114647565,0.07615565290674742,
 0.33590376388922183,0.25129169972585497,0.15669073031460956,
 0.2504393626125505,0.5873388283577684,0.8527099454282734,0.08515719992437643,
 0.9740095088934112,0.427384726842735,0.036754007891076634,0.6885773319730699,
 0.7340916742378741,0.18793709905401135,0.7853221608053154,0.5004562185209758,
 0.2977791201983715,0.2332147285287609,0.9780683787024679,0.4521626748290638,
 0.28942076103675585,0.3854010617519975,0.1457195536189958,
 0.17583982262949993,0.48363865711546294,0.5456738876809631,
 0.04065092868309084,0.46547089871875214,0.9535574455807924,
 0.7059305144907012,0.695009589654171,0.5076233909743904,0.3016840390948713,
 0.9440794981361155,0.2966555934722127,0.01707099104296561,0.4934979686616767,
 0.649491149922583,0.36124109663411974,0.017092307851664823,
 0.2029080885895903,0.29664875790826617,0.3453705888243188,0.5301389298234435,
 0.0239560516691395,0.05656766323688289,0.9811035204309406,0.8526742190791384,
 0.5621674951376705,0.7096134814679337,0.27922181692807235,0.4917278420111504,
 0.26971205884127814,0.4181672618934551,0.4158010885540492,0.3244128506911509,
 0.8588786836370433,0.5409427592131448,0.571575989629806,0.858627738101009,
 0.33530406814733327,0.8505291540075817,0.7338689343981414,
 0.47847843725161066,0.587458477755634,0.7105349230255845,0.5489263804019886,
 0.5803711264785266,0.09732445704609805,0.14942132677889164,
 0.4478810367321383,0.6660858830035226,0.5744150642743453,0.01705364552803912,
 0.40359696994658434,0.5485210210928931,0.6820903795457326,0.6033170079756199,
 0.7614664934472277,0.05116076556100857,0.37437696574741697,0.761726866105927,
 0.6939047667133392,0.4698205871158021,0.0357580360497427,0.2480982281518943,
 0.6355389804762913,0.6042886951911499,0.3873855813318555,0.1360867372653607,
 0.7936582238872846,0.31525708740071934,0.5082569776307293,
 0.038033016612280424,0.5583422258148345,0.6580846351188816,
 0.5100978277857057,0.2911801509889641,0.346165081933979,0.870301224743229,
 0.5252256553312286,0.32981264516553455,0.9858536630166822,
 0.35182050968039413,0.632439882909738,0.49748307522146484,0.6881012291077581,
 0.09958805352167033,0.05030000116238509,0.11167115221869017,
 0.4369390146504605,0.01837676171611946,0.6819568636930888,0.9417679750856697,
 0.3980799342857697,0.9842829992875938,0.13599532808182724,
 0.0020523796145301976,0.8708667712276882,0.07051558676211878,
 0.2123593644677838,0.6677289384287073,0.3067790104189988,0.8138091204323276,
 0.7834582487138092,0.7729756312688213,0.7907699810840473,0.5763291836843796,
 0.3512609050532076,0.20399593095810575,0.6673249387148621,0.6091041271041103,
 0.7409642092650176,0.5056465214796062,0.4636970669528071,0.6261673358733507,
 0.4171455468685399,0.8703015447184241,0.7560920450251178,0.09487185025602185,
 0.04858093525151952,0.4883287182582795,0.9497233151334508,0.64211553038777,
 0.9541400778159531,0.27481618794600526,0.013883956951046938,
 0.132636484257051,0.870832853898353,0.8632324405462097,0.07831850694224585,
 0.9053626862589155,0.645944080150665,0.7964790636280933,0.5373783733332381,
 0.23963070370998985,0.3214379098747888,0.37171932217389025,
 0.10414542469099242,0.011320462884714333,0.10998542516350407,
 0.14397548777979807,0.3753973027014026,0.6376028407693904,
 0.18964197588428444,0.9411933915976761,0.8762232175750462,
 0.060289232574983975,0.37809459640606957,0.46011286251760675,
 0.13846073842297446,0.9203927886846361,0.5525958835865051,0.4140760322212843,
 0.5254887646129749,0.734956349740671,0.20501970370919165,0.36064160384947064,
 0.0025406183725132925,0.3168298523288837,0.9527455420235446,
 0.7707874065011595,0.752841791569892,0.7768257029100253,0.5889636304940098,
 0.6311030054220924,0.4337093989127775,0.8527210411741283,0.8907038306982291,
 0.9937967793301129,0.4198140598437403,0.3973662600043856,0.4399806276105752,
 0.9655748267491908,0.7727758998729755,0.6231442340684378,0.40441838798960295,
 0.2339138591624751,0.7442900626456407,0.27556813247993217,0.7975413549378962,
 0.5199059534815358,0.580756831525546,0.32873402357261794,0.06833074898774782,
 0.18387587187908683,0.12702920110005067,0.7401486768948208,
 0.5548105105670733,0.40759378845533845,0.9159488287105975,0.9051326896614975,
 0.38122439459928015,0.2366935130896607,0.05459711371344067,
 0.30832117707357587,0.1518998635912847,0.7782370866717299,
 0.12390631404259944,0.47929607840371014,0.2962332068370397,
 0.40555105732892205,0.7791966555171813,0.23203972198337597,
 0.6060417988724761,0.14741915232720093,0.4728069659087478,0.8485008898280824,
 0.6114893247493749,0.4828200191243279,0.12404455793118185,
 0.36152025519052744,0.887550432173527,0.6840519191204487,0.27126970810348083,
 0.8461257703543122,0.9135296436038812,0.8463342718630926,0.8915197666947581,
 0.5870711895046773,0.3213991392804846,0.9051796440306092,0.951361952989357,
 0.8669575986849036,0.5040738592433067,0.278392793827566,0.11555893620940216,
 0.06702892906349112,0.9612753856772831,0.6419516389859529,0.8925513956399493,
 0.3098661027793522,0.9331197217383427,0.23121956087176332,0.6035413474008856,
 0.725700214352379,0.21039764719986076,0.9804629707831751,0.24079050371314792,
 0.09754675261578705,0.9004735751286612,0.6637904087247736,
 0.05706260333743718,0.36672480080893854,0.8892301471734086,
 0.2528247512688454,0.422579922174912,0.6543514077704534,0.7664723871927895,
 0.8753925765510464,0.9204265054516554,0.7297812030470086,0.35433280895516417,
 0.25137893558809543,0.8414664476841603,0.46156792221558895,
 0.9990535745202501,0.8878877005871438,0.8887373583803353,0.7599465536414308,
 0.5110801622558375,0.17845021365490765,0.3005921046894794,0.5114327893254615,
 0.18513421768996519,0.7863812778952372,0.20424612911160756,
 0.4446583118787062,0.17807792319409466,0.4507352430999454,0.9952986883036647,
 0.6429124500700709,0.5543113140809484,0.08901940804778208,0.680339588641845,
 0.004038152919635829,0.10199689555471514,0.2153641739739509,
 0.31845429754523624,0.9178604504176527,0.8484971735767343,
 0.010928593281597898,0.30885840604547876,0.5713417920082065,
 0.39801313205482236,0.9785237549128741,0.22113440991851196,
 0.4546843077172882,0.4774324992027861,0.27438143537199544,0.4665775154518177,
 0.18250886489546214,0.29415777353618133,0.20389872417457922,
 0.24399562470073344,0.20858005079292152,0.48925529227921916,
 0.026631277868338965,0.7931768446611693,0.0677035890766754,
 0.38347536918982617,0.707983580034653,0.38709783190707625,0.7411926660845007,
 0.2027993727951184,0.7917630715639007,0.1541216309813277,0.9760031467265293,
 0.6518348810104904,0.3243104694076744,0.6969335424211076,0.8790120085257069,
 0.5888293470812047,0.23482380240212708,0.48843797117747023,
 0.20630451224497626,0.8218533787108884,0.4089531423936562,0.8113068125895998,
 0.9651470427317815,0.6939035136426335,0.8438342406204972,
 0.047915460670579724,0.3502680912228524,0.4853733497349887,
 0.34045349775699174,0.1940509584821274,0.47486763324228054,0.724567518545188,
 0.4945053795965746,0.9779619776626594,0.38181876127717473,0.3127722973894541,
 0.24091347110803996,0.18089628968632898,0.744031259300932,0.8260560381257719,
 0.1764174595138991,0.1244505555472053,0.014906940648736544,
 0.11228075659650183,0.31988951664295384,0.4231054538687802,
 0.3565429375369199,0.6605363262166072,0.1555816992028326,0.5364968667224136,
 0.43295964311316404,0.18323771622109464,0.7603111072871837,
 0.05713725394479985,0.5135089793455688,0.9983471090351583,0.0888866884112951,
 0.3792120357476456,0.8658610603194669,0.2252946932707397,0.70167107974361,
 0.5655896648868709,0.07733883283708154,0.009349027428765977,
 0.43945495839239657,0.5034420268736512,0.17550393198351544,
 0.5343763402207724,0.3940415152233371,0.6917623718520748,0.8512067653003315,
 0.3018779044646467,0.30563492154881833,0.08795837502792192,
 0.6640564124073671,0.7634451333930707,0.4616389464768018,0.8479339773441045,
 0.012677544110701544,0.4223286580953489,0.11566455645526363,
 0.554256788852209,0.4202620033411043,0.2905002289699541,0.3656383973611691,
 0.4786682493424619,0.9601390765987013,0.22774551074830296,0.207237806365676,
 0.9616835255524017,0.10633137522631242,0.018776217031971076,
 0.4271644582226084,0.6873787108064127,0.5439221046898401,0.7140976459697518,
 0.26112762029992254,0.6685806993661182,0.9124877144027512,0.9298715653604591,
 0.6343084013123265,0.4416282709401492,0.6751361687560077,
 0.008899262506720462,0.749853900759789,0.050206128502637726,
 0.6063128495455441,0.6830596095312519,0.3720641712878351,0.8891923910313728,
 0.5153280867604648,0.9343973015431787,0.7208767093063606,0.0923862586194788,
 0.08499468912613894,0.15784651573897313,0.20594379452778,0.9684768759981811,
 0.013266028295971233,0.9181924058192823,0.48591120181994407,
 0.054497550643656134,0.6341694295381876,0.4193994613900197,
 0.7379466873547772,0.8817000685080767,0.8041106759221239,0.4010587899946224,
 0.5243697766142734,0.22245334310950404,0.07078988728574331,0.374387580788475,
 0.9826431494794057,0.3344906296081742,0.451831830897379,0.7033345855952741,
 0.35155302443804937,0.48274211708293946,0.6634651049104209,
 0.14974848167172516,0.1813539471085308,0.25447324529640847,
 0.028224603000340243,0.25970416936997165,0.018426581016075083,
 0.896054268911134,0.7780673143909873,0.2615020601082153,0.3401737986111468,
 0.6153068643781935,0.13222299521087322,0.6375175811336209,0.9847381716230486,
 0.06635081785667873,0.23806549722757286,0.05654224819160303,
 0.45753199711482173,0.07294506308373272,0.6544847394640227,
 0.9329148000388987,0.5552471557534385,0.8545908469716419,0.6333860587472412,
 0.65148132474562,0.8342557299576928,0.6296416602864652,0.3686572047817934,
 0.02910317452955935,0.09797748798788919,0.7598997243957131,
 0.004216316941156983,0.37806164841874246,0.9212499763128636,
 0.8792825191997622,0.6656165301533652,0.1198188973226395,0.6377324731169851,
 0.40013167292229734,0.8773354354939529,0.5453598952606618,0.868463709532504,
 0.3627575297879144,0.9114395472620691,0.04488478051626421,
 0.24999918009549082,0.710175296768547,0.20404956855006662,0.3040712255344411,
 0.09120354418723653,0.9232960488313096,0.007371780345611145,
 0.3389066471245521,0.5208289721788403,0.5145139596016819,
 0.025150826563997586,0.020396937700888906,0.7529077853155391,
 0.04907495968788822,0.9802143945017712,0.8966663532323929,
 0.07314456597206098,0.2567730003226135,0.10414884913506273,
 0.6802674074510299,0.5255835601131409,0.13020050241783587,0.5931245088007365,
 0.8583571971946523,0.1972201828759107,0.5895967746837353,0.5424647893048066,
 0.07773242118609291,0.706669158652438,0.5169617988601055,0.30151530835513984,
 0.06444706229875274,0.7773609489141862,0.25543645273212223,
 0.6988013446337038,0.23317569154607432,0.5728954629484542,
 0.12147068783543868,0.23656120224309873,0.45229732348164886,
 0.5931040330151408,0.03651720772016542,0.004753753711445485,
 0.10861505300324104,0.09640779690140444,0.5492138648572131,
 0.7936516114186859,0.5565125551659654,0.4395023633455102,0.8879961573852759,
 0.29604201229984595,0.31494141112871443,0.8542095428130062,
 0.7785907872468207,0.7094157961579157,0.6026600576923021,0.9850136756773826,
 0.5150489052972762,0.6202729199698187,0.5070675640818925,0.6916640062266126,
 0.47226612879164376,0.6662597096437315,0.5892904250691716,0.9894034661937617,
 0.7153488272309285,0.11117594118771423,0.3869120771686051,0.5470326290591719,
 0.8205821565101901,0.012020103258783221,0.49603506045175716,
 0.028590395919409928,0.18963632429726218,0.48169493033602107,
 0.05870343481667162,0.4517017385030616,0.7276419822873273,0.3945669868357865,
 0.9035083090969228,0.19686532046912597,0.6923379004016057,0.443162765466555,
 0.5137384693269589,0.6230346485264218,0.4233316763697432,0.769169736060621,
 0.3370631417258634,0.2570788532607863,0.8428695001933154,0.21864937176787735,
 0.3073974341531208,0.5635253159599641,0.7115672188839126,0.14991813787758712,
 0.24462793569149566,0.17906146706289172,0.025530250678766908,
 0.2809190478673781,0.31629977078622207,0.40012801061590875,
 0.31938316810657297,0.029263415123153447,0.9296052634366916,
 0.25820198189362464,0.7084663692217529,0.5492247372512992,0.4611260762493623,
 0.02725311611668202,0.8315774091414807,0.9819274358274754,0.8690807523597174,
 0.5107814822610343,0.033873868976739385,0.04868673549353564,
 0.6779257317589562,0.5023811380537371,0.44041155940106225,0.7823951134746223,
 0.5446186445349818,0.7981757354531901,0.4830925054075411,0.22071989612507426,
 0.7873222249538899,0.5467062831192169,0.653472355823887,0.9754448505935714,
 0.13658897656832247,0.32472176794361496,0.10624860238768885,
 0.7619228333897671,0.6330987397454866,0.7504286864145334,0.6520574359031667,
 0.08669685180025632,0.7279191900787602,0.7472674493981059,0.7652620069926972,
 0.5729662052847524,0.3672498438182724,0.9884318207122688,0.20516146602146568,
 0.2166194956353945,0.9072253997894364,0.6140725839371092,0.22606540623117544,
 0.16833947966092933,0.3918542446126544,0.11362094847083748,
 0.2941404880799591,0.48074254639322855,0.9746454376709943,
 0.028448756750512727,0.735202499232859,0.3344045695818355,
 0.02149425941365235,0.7218199718515889,0.13402745375773983,
 0.04210238346297912,0.9621024995243632,0.564534902640882,
 0.029644081553516943,0.303714355953683,0.07384589580271006,
 0.44215173010337416,0.5401604118390733,0.5754338807651435,
 0.15950012948852121,0.05230894407232234,0.3446409982278883,
 0.4031981868615264,0.4112806849476268,0.3098689664028441,0.2352135648089566,
 0.07081482010411444,0.7419874155038998,0.7135028469081213,0.8577369200608327,
 0.9093653685841532,0.3932812477136117,0.673284540998936,0.15031270946132858,
 0.17401169482119028,0.8660126786305412,0.27905476265945894,0.643963291532657,
 0.17798927439417966,0.4942608305382449,0.4384544776782633,0.2825007472451768,
 0.47651067724600626,0.5366238028397583,0.5605339240625802,
 0.17340793340196548,0.5062762768521079,0.328115252620106,0.8860170640566583,
 0.7439311278273003,0.6013617981448365,0.6346802162115672,0.9516203422293674,
 0.47896983286831885,0.2569985765739151,0.8623608189359574,0.8528078455606214,
 0.14604692887993853,0.8669106372235769,0.7710870484260034,0.3192162328691561,
 0.3365316748831362,0.6510883225752864,0.5618278905076889,0.907912179882451,
 0.7087264555102399,0.29365334241561314,0.4162151550361206,
 0.19918502956421436,0.4065618522313851,0.9346620729698287,
 0.28097437599128006,0.27612773111073996,0.5555189287235949,0.779450413767553,
 0.36968705123103374,0.04744033882709764,0.1707928073962791,
 0.14715871079698128,0.9086249158051324,0.6940605723659039,0.2304853914434688,
 0.15266212373588128,0.7351886392698933,0.3732794681897733,
 0.28098400986769256,0.2015744975310405,0.25289239223908966,
 0.010204009608262066,0.8458510611491802,0.5234101873782002,0.598629136207521,
 0.7518821191020172,0.34551025800950896,0.5682833310125768,0.5481099318365997,
 0.7831500263504407,0.06936800599473503,0.47265348035510435,
 0.40251103928799836,0.5716007983064713,0.44159244693996835,
 0.10183174437909681,0.5420870905636519,0.029642622602777058,
 0.1731693762689792,0.8421480747340939,0.9551792327090071,0.19886427098951442,
 0.9142076999720617,0.6315536843550132,0.14788060052907204,0.8708058552094352,
 0.4602878596015082,0.30274836196021726,0.1715997468201268,0.861770039277862,
 0.9341739442959023,0.48658095655553546,0.9216838229128657,
 0.46475313030758514,0.927042642273932,0.426108252255042,0.6253255202709722,
 0.8086682068819215,0.6401224724776005,0.09564444078825718,
 0.49015823456299856,0.6255087529293019,0.9799217869501133,0.0897575145170385,
 0.405559336957388,0.725510788166336,0.896717208115396,0.3974004530834562,
 0.3434818559802252,0.1903275835709216,0.6569477738599789,0.36869047649461517,
 0.7046545009706583,0.6213295345989821,0.912485254260394,0.1614503160139822,
 0.8014057322389844,0.41439082313866304,0.13417133940580828,
 0.4040554820328516,0.37959054737518627,0.5522141247641256,
 0.07106681754461142,0.5662469541328021,0.36928202802855203,
 0.44347941056863616,0.490964660226584,0.3226130434772585,0.4484376627716492,
 0.2625854001262091,0.8082462009308577,0.008982341381535863,
 0.2588228552985776,0.6154418442889722,0.10308547148788005,0.6962897187306387,
 0.2873520395998108,0.24696904891117677,0.1726001425546022,0.1786170847295574,
 0.8696656179466338,0.4892591992368348,0.9399881701393697,0.9341659223734161,
 0.7749676901647309,0.026930504580761294,0.9081264976427921,0.964157866101175,
 0.06951141927792936,0.4937983885557826,0.19880685492414008,
 0.4884883239107235,0.419961760277901,0.4518709633426329,0.680553608541752,
 0.9576032206133492,0.3124266658834236,0.2957131629269585,0.01143018402167173,
 0.3768213582539731,0.3038574668527598,0.15001654724790525,0.5161163837446443,
 0.43413325355163235,0.22132065696263725,0.4357708731831581,
 0.37718509389169586,0.14271762678367383,0.6180751214445519,
 0.06726397180441923,0.35683043245304347,0.9535731765288664,
 0.5068369938725049,0.4131334125401629,0.5432921800585588,0.1748383391060757,
 0.052537032140562934,0.13396583529519146,0.8972738690619854,
 0.5969208065316096,0.8018114498142027,0.5756042894467502,0.696664882781852,
 0.6895961467816555,0.9358724924163518,0.9072857605468427,0.8062597118739261,
 0.9260946230130638,0.2877668716698387,0.2794225613892928,0.1593875162162638,
 0.1789836823653116,0.5834213126649779,0.9907268417848121,0.11348448645842335,
 0.2773110967334581,0.6524731265290721,0.10456968440113557,
 0.17403721795985472,0.16005834184612722,0.46485439149239827,
 0.07938677518279691,0.2620337784335265,0.18982623979573499,
 0.8168203138803883,0.6548471148186281,0.6423207289945974,0.2555477530471292,
 0.9690861351961297,0.1386489491548305,0.5868973949227744,0.9506673427572724,
 0.8736811859344207,0.2860923236651729,0.48285151119451686,
 0.33604694927979173,0.8365456343501898,0.060109510180350156,
 0.6938991998652041,0.36187623528960966,0.08658419412487417,
 0.8925972341470252,0.35090739779407987,0.9927180026165265,
 0.24321778805700067,0.5370438952026074,0.26434312180366004,
 0.7422948795472524,0.7462931385912444,0.44550894310793143,0.7892516281893464,
 0.3415389950909721,0.28994313957422013,0.2315984818893343,0.5933858357097397,
 0.816785932762965,0.43824515200917424,0.21245131004458184,
 0.07096478415644936,0.634226958020558,0.7929203222712169,0.41314949220479136,
 0.9471917374985124,0.542010335541677,0.03226923629876155,0.16319885356099706,
 0.200744984456806,0.9520312107774564,0.07902335469448463,0.8755090421416087,
 0.4734769981002971,0.4064328985985788,0.6452425419325587,0.5447209082164514,
 0.8103670438764701,0.24426350055906532,0.8501815640706026,0.3559184958988433,
 0.14894179454425438,0.3647663845163235,0.6580446617531598,0.9210832064187738,
 0.0714351803321559,0.1086309925748562,0.6894247354492898,0.01646611692246891,
 0.07595900630678465,0.9063767426498701,0.827191849434235,0.5068372319649752,
 0.5305900225946307,0.2741319623923615,0.8386769486699434,0.40096596364086456,
 0.3136111384955116,0.4758163378173278,0.44313576908715624,0.8142684893877865,
 0.3264293973493241,0.9005099788665354,0.07082008558771558,
 0.23549927341111943,0.4295716059088981,0.16412565837591608,
 0.4157745706313455,0.7885160859682273,0.016241064274436057,
 0.8120675501987809,0.02364620516798044,0.6277695984633801,
 0.13798908467445115,0.27858259583440614,0.8939049073512123,
 0.31117556853508355,0.7361726549555137,0.004058758025823117,
 0.019856895959234677,0.15097392444124325,0.13261433248095,0.5729720164319299,
 0.49995543761881667,0.733149496264804,0.6427676530368136,0.5621489941075055,
 0.527551589110665,0.4562629688322717,0.131966293261504,0.7118599242882884,
 0.5027077326078531,0.5422769696192058,0.9212496376648213,0.6093499331076502,
 0.0346440272391384,0.5421923301890377,0.12115350777269207,0.7940956137323091,
 0.9335383866882686,0.7163423671785704,0.09624247556592125,0.4557529484619339,
 0.44309383962182236,0.5963542793728742,0.01597221409861982,
 0.6341500574375573,0.1443264196372427,0.30479791364199116,
 0.46392177381439004,0.1622644463932802,0.94606793244247,0.91782251558308,
 0.7741386070249967,0.3985607020361478,0.10374614152583639,0.5850346295844913,
 0.5206457920461781,0.78182256819935,0.17392184219757523,0.04879999371537069,
 0.5151445087173274,0.8313890173279195,0.5858377071795052,0.9712489096767927,
 0.7607172751824846,0.09749659612050399,0.9746722498302336,0.7183897352724298,
 0.023945088369003775,0.8933999642824133,0.42224092811723635,
 0.21068915645551567,0.9957358820457178,0.9932444163891312,0.2252983449880258,
 0.37313633060396434,0.9647478266151266,0.9812772690580784,0.612593603564748,
 0.4353810363900933,0.8701069698227072,0.005165492034098107,0.977889407770588,
 0.6824999797040563,0.1070152816979153,0.35292604050756093,0.8276280448578214,
 0.6254896312522822,0.22753903878509607,0.4060574542566354,
 0.06032244311229229,0.05828358641228881,0.706770840045493,
 0.49885843241798977,0.05004373664950257,0.5214422811634506,
 0.3514805555743318,0.7300469266026791,0.7402367135445357,0.6725029442295484,
 0.5931520506207046,0.0945676786985048,0.7131995294372231,0.8921015844431852,
 0.5233733805673783,0.2678523062110999,0.5178424928233365,0.27817814377469263,
 0.5828164361558066,0.5385318918952819,0.34963616282856946,0.6135736251594287,
 0.6676001969890771,0.6723158838660123,0.6158824263498468,0.2505512363126363,
 0.10141001933878424,0.23898882541160937,0.2767620458958576,
 0.5797040381390177,0.9099009779562552,0.010999766027285363,
 0.03639509679516384,0.5752730424334094,0.8178940584626084,0.6394911064287331,
 0.5755761991138911,0.6918749744109015,0.6949781934736416,
 0.029631964246367826,0.30215913632820424,0.009275382561311885,
 0.22569454124676203,0.181058687064502,0.35984695935988675,
 0.062119635033327625,0.5393891300543993,0.8538528641362769,
 0.4653418438628406,0.7604140567560971,0.8030833007561211,0.20398792736629845,
 0.5074944993890711,0.1405009159076206,0.9166273517021418,0.7416264102040466,
 0.13688040433092075,0.19006009610972652,0.5160449000365162,
 0.2643794209408945,0.12673156335757585,0.5907367855379257,
 0.30487306123268176,0.8924846786501303,0.46695319518892364,
 0.6285749471882278,0.5491664723005306,0.4613973324425229,0.1708598460584313,
 0.7925843184166295,0.47463000592606375,0.05284707433271518,
 0.46973591909567247,0.38969935393064803,0.9084192039377373,
 0.31141265761084125,0.35175946783559087,0.718958050209418,
 0.04389017166025366,0.20490704437687268,0.017287932736262945,
 0.4104406913075792,0.36358588253842017,0.4797278793962242,0.5451798696472768,
 0.7620144739311104,0.8279134330883109,0.642434227272222,0.830779262101267,
 0.47670935661742997,0.7004337743423261,0.6948419745517165,0.9302939869353969,
 0.24289850845479344,0.6484983823025599,0.8946272473111396,
 0.030506776915175227,0.056593819019981395,0.12776771495049732,
 0.4906115676266941,0.0914838331944644,0.050530689461245726,
 0.7906546894286343,0.5443571082987673,0.8676755406416745,0.5333111348168987,
 0.8585435830547449,0.813921290459729,0.09940400570335362,0.1431338410871933,
 0.7847005485575349,0.4588003578715859,0.015351808351306562,
 0.4698274368220131,0.5405453140666538,0.34937764067809496,0.9102323828839834,
 0.7588716904591987,0.7699356798035494,0.9046331089147907,0.32161430269049573,
 0.018991029370473145,0.6895244523636203,0.6799540149261278,
 0.6570353025232577,0.7112553553784863,0.47183130718716915,
 0.22834773951693266,0.8391547768594184,0.6194604727195191,
 0.49700963136997667,0.8679726608694149,0.1345211986298076,0.9861155422091837,
 0.8760931989627092,0.07930343441741772,0.6002170178615872,0.7602445407555037,
 0.6746833829149809,0.2209933525749298,0.2699553510401881,0.24716432813974631,
 0.8277471149630783,0.4075745426823528,0.16112703239593162,0.98793287221602,
 0.4437785685481822,0.6557237792775205,0.2682521166179681,0.15976674759728415,
 0.6980155988778246,0.8575814284944234,0.14145611921772172,0.2149944420959372,
 0.135653450696824,0.6805722092174917,0.7689999388129914,0.5175562906527926,
 0.04855837814400432,0.3745052727015561,0.3678299966715026,0.6892223170104241,
 0.298126042341304,0.8325410852123012,0.5362871841842738,0.8350721311691509,
 0.4198577258421242,0.6694641014581106,0.3925477034218837,0.5512484706745726,
 0.8596908113188462,0.6357366433715133,0.25215544374686927,
 0.21815091439971512,0.659877585256647,0.18375946884140437,0.9305913664017775,
 0.9116292485271232,0.43468476722879346,0.23464785846155212,
 0.13280416122870053,0.9109194193403076,0.5533236372496569,0.7752694728876278,
 0.7342764959975469,0.7116664631147334,0.5259741824071944,0.7897646153242274,
 0.14417087295239994,0.3565246820254361,0.30461852014559254,
 0.4098486639157406,0.9041557802799378,0.9336820030184589,0.1462756250232924,
 0.13732798684555747,0.8452705144970625,0.6700066701183335,0.8158544297945701,
 0.7774961762979754,0.6510665488193826,0.6036509794199535,0.627815309618042,
 0.8187673801958829,0.6639524642680451,0.16897706191419015,
 0.019615388685652824,0.6849141637549288,0.8808298171962324,
 0.13373793975051074,0.8003086053490734,0.5961546420587801,0.9626761721712281,
 0.7555396353748624,0.6702105357107628,0.4447178090145161,0.6308922288133479,
 0.15358956716158956,0.1982574794994848,0.9605749398734008,0.3222443100724741,
 0.20937710257078213,0.18953330972876314,0.669609845138243,0.9977807322055617,
 0.5319874854231395,0.41368882366603077,0.23423310801492248,
 0.6617420496094418,0.48575494913971606,0.3674081209419664,0.4730374077570074,
 0.9988997108787487,0.4386908770849206,0.7142495425326221,0.7655534458928988,
 0.056308028663827114,0.3392009107506827,0.593532727129757,0.4732224417438673,
 0.09933038979101205,0.07278281513685347,0.3883047409028618,
 0.18411863897687553,0.1991507137532258,0.08120116388408383,
 0.6065654659275004,0.7284295305890631,0.3859461962069557,0.9968263574564155,
 0.576456060567579,0.998467600931117,0.6956483173494232,0.8876169922722227,
 0.8689302572473364,0.5725343596694887,0.7012404103859705,0.8253999522964941,
 0.45386403723269464,0.7621653094605781,0.3196787485211917,
 0.44133990675602097,0.48905370913404667,0.07202814725046813,
 0.2098152733794756,0.3069159689759756,0.7590823926083162,0.556075042884984,
 0.9670256294719651,0.6683602380422906,0.009908834183267268,
 0.28375511187177405,0.007306918226757886,0.6165100781714862,
 0.8790510594687088,0.18287711417204777,0.8684869425310143,0.9373076040166594,
 0.43311645328734416,0.9255713773171815,0.8801872708935948,0.3209264790071348,
 0.8118095588556578,0.01383902863781894,0.4904854093912372,0.708056744993851,
 0.6984473443955127,0.35891912972885764,0.4028244164989597,0.2573525311795357,
 0.8007722496344335,0.21013435829247906,0.9602170464487723,0.6853648378320196,
 0.6264350111388783,0.7676354940018958,0.6942771405710478,0.537796707545358,
 0.7185981128980264,0.011915485523616098,0.045592841210843726,
 0.2956885497841252,0.29853810139306713,0.5064422859341063,0.3285742328953992,
 0.2641508386661211,0.6658628996987361,0.3767110032342258,0.565030203457002,
 0.6971501637645123,0.1034673690418706,0.241690999078631,0.17077273100366563,
 0.6923379409687114,0.9854324512824604,0.5651394619288165,0.8691222925816972,
 0.2063757512091875,0.5808928998800754,0.6112679911690022,0.4891466921613251,
 0.06528166573789596,0.43945222155493624,0.33919956580447874,
 0.5137155991424781,0.7366450198446153,0.6082213334783502,0.8179884010646299,
 0.6380659373622264,0.9216676353032651,0.36692930772392796,0.972926695010638,
 0.1262416049461762,0.754157519702314,0.9361201673941072,0.5872114608133004,
 0.737802788033103,0.4651867282355677,0.3642361673172988,0.9060489145733315,
 0.34572083549958954,0.4692608671052314,0.08530911453111267,
 0.04890677231940055,0.711604294730323,0.004629644957278023,
 0.9161950097037026,0.03461218668344035,0.8017177919552861,0.9496538583311561,
 0.30717692199879987,0.7835929491809479,0.15977362868348532,
 0.8891453666472696,0.6096255696338534,0.9667652856356973,0.5758485417092214,
 0.31617778748353687,0.8616223773696302,0.023812448506984718,
 0.7157559167898444,0.16535801553845086,0.5088578956302106,0.459331536600849,
 0.9461552162198041,0.2790200508332418,0.8444950695888342,0.5008859664423821,
 0.8960089603797946,0.07756155600015169,0.8667944962082941,
 0.18144730358061634,0.7875080299272065,0.817654730002667,0.9878087565643929,
 0.5192250859792154,0.2114687985831314,0.1944505223025228,0.6911705210148134,
 0.9813576680064815,0.6508736830845058,0.8291373564004318,0.2560309632210638,
 0.7272340396167956,0.16254255709770715,0.8544009939644397,0.4626541623372826,
 0.9764150346463228,0.42738229816170215,0.4619515731507755,0.7988948233623597,
 0.4931944794115293,0.09257350705394729,0.2540258644511455,0.7624787786676422,
 0.5891737281859912,0.6813839065101295,0.9178823987075292,0.5483328830941371,
 0.3938521953412891,0.8191371500113553,0.8878179917921157,0.30640132682079313,
 0.9819699787053403,0.9395125514495558,0.9299938057469851,0.6947173257216628,
 0.9255343424311582,0.37283441155127905,0.6780098239185209,0.649541896336024,
 0.1736111084986518,0.6905723409161559,0.7711781501598592,0.6879830775235372,
 0.4273086024500665,0.7714266170869665,0.001041569308819268,
 0.028613266497398704,0.1453117345999143,0.7592393042879613,
 0.5489201079379721,0.05222059575098692,0.6635513560399641,0.6701419368539471,
 0.41263196781122524,0.8810613312638352,0.3366385665658731,0.7006011783129866,
 0.14760828855207708,0.4292725816244236,0.05392837048184984,0.419944414080888,
 0.42461364604088003,0.7805340009245412,0.4334376190590312,0.4570264015357771,
 0.6455079123245747,0.048723004244208834,0.5576788602520205,
 0.9622654937587662,0.6209765798970102,0.4418848078247106,0.7930091813656548,
 0.39538450291525395,0.8526711270088593,0.3860184027021911,0.8200700735836554,
 0.03368624299045275,0.6516620491497981,0.8559300876780479,
 0.18423387373370048,0.9561552300285387,0.9647571863164297,0.3226831068914072,
 0.8210328563277725,0.13446324633703877,0.6755240612200482,0.5810496573185391,
 0.4154566861318668,0.20341347234881335,0.3472821305877418,0.6030222038133526,
 0.5578002288259746,0.7350640964413258,0.8431794263316277,0.1791718747669957,
 0.19382454385977976,0.09392545652957873,0.5338298339794686,
 0.40172042432329524,0.929413982846202,0.8096224249381729,0.5521460262528458,
 0.8937485350546754,0.25470836034700217,0.21932861556534422,
 0.19548516960772344,0.40975049743606407,0.42328694653703014,
 0.5991881135854034,0.6712169137994625,0.21733761222711068,
 0.20520875533238592,0.018529018887714188,0.08511073866518337,
 0.3139097514448639,0.5288538339927061,0.5699634777182035,0.9645077579047986,
 0.6557942603764464,0.3836869824307171,0.11377699925084006,
 0.46482020732040785,0.7497228050708465,0.8009033110813633,
 0.33467556576111623,0.18529163417203676,0.1005631985584895,0.554779448240469,
 0.6445596586712633,0.9789315370898117,0.9763838319954591,0.9723161225038676,
 0.42644939027454865,0.8227970510419756,0.8071963349971303,0.5822493929500627,
 0.6906553493966705,0.29258709230556956,0.4827870941678304,
 0.22627367739186532,0.26009017858094696,0.02086071212362617,
 0.9245857357161531,0.9189714594914329,0.9760230706364305,0.9442668519974861,
 0.8213374923467388,0.7506798429632701,0.4759747350080482,0.3481766263146684,
 0.6208887900717248,0.13181734494739084,0.623799665066403,0.8993154968968298,
 0.20393210682746965,0.6626116236834284,0.3624137178389075,
 0.32774069177161036,0.13439256134514554,0.8372657816380354,0.395502807778914,
 0.3552540530450642,0.25570121933362966,0.6018856330710796,0.4302141414490591,
 0.8273230441072592,0.7485624589830683,0.03508588992383088,0.9341818695416811,
 0.8895474409015023,0.00687484881656919,0.32525954267066026,
 0.2708326292487655,0.35454919065410706,0.33355371079420726,
 0.3456819083124274,0.5528386956944593,0.45082882763208254,0.7135499900926663,
 0.3616194359078305,0.11250406757030473,0.46069937765198954,
 0.26699244059418625,0.2375377621566651,0.35769507143760393,
 0.0662081134803102,0.39591591815845617,0.4976768998539831,0.6890207663230493,
 0.26004356085502756,0.1491427925173795,0.9692376595000076,0.856107841907731,
 0.7469094947909479,0.16909561904691983,0.5321778884257475,
 0.40545760380760165,0.35298539215739577,0.45312524276854693,
 0.5297330242684724,0.925066065882132,0.32999643581983407,0.7729457020759191,
 0.4712581948687157,0.9255465972785094,0.2849278607510064,0.39254009169534554,
 0.9467860920818656,0.8357871453395637,0.8086965798939099,0.8317089266558867,
 0.30484192740601723,0.49774852763369926,0.9343687460669711,
 0.8369336167181134,0.5639361303183751,0.6001503264031685,0.30172670681817004,
 0.2790117826155244,0.8946805798175765,0.7979290274521011,0.6957375372374598,
 0.022402745590668016,0.018273735248687806,0.06821918462779197,
 0.9001712243836686,0.4187797971851478,0.26643314277401886,
 0.01194192807068517,0.8178759819122607,0.9628479619931809,0.6395267806546222,
 0.10318413106408897,0.644949488851541,0.49719259665596593,0.3516713859760473,
 0.7748487025236309,0.9976918961601222,0.126118897272531,0.34376360480495205,
 0.7872318410631619,0.13564744490534664,0.9378981122076702,0.3761962723968624,
 0.15631452252919742,0.5886706092488903,0.5716290405151402,
 0.13332396041158778,0.5627834228090232,0.0838177588630844,
 0.036730335646274814,0.2584754975787841,0.9656176005182604,
 0.9118799198376456,0.5123357684278939,0.0704711272128018,0.33904726058227275,
 0.06937112470694384,0.10619184138380156,0.49162329065389454,
 0.010090861462265677,0.4302257334286067,0.1139182701140442,
 0.7701972669536856,0.362218400056321,0.043479927048401756,
 0.07982786079271409,0.7208985784296592,0.31535168608321584,
 0.5879112856252323,0.3713311419447629,0.6912563340560146,0.5269531925755642,
 0.3540493120501279,0.4015510430772493,0.16281380558443603,
 0.41180610015274854,0.17832232923333458,0.3539346876271634,
 0.34370211717834553,0.22874843414026003,0.5370601575127676,
 0.9204984506222647,0.9438198264534602,0.017625463457239587,
 0.9263516381287711,0.28095642726610304,0.4681872751404861,0.4248449893394364,
 0.16234508330965625,0.12388113913393717,0.1504952055344143,
 0.5159173769331319,0.8049581028062759,0.3358410472743061,0.5986142267750628,
 0.15185926627460244,0.28251264689858036,0.5359487702612107,
 0.5624215685009204,0.9984243519306315,0.8685045358853991,0.7816695748469168,
 0.3678161296731307,0.12139304737592271,0.30882282119427296,
 0.39851120019457276,0.10683990680036548,0.8234659741025917,
 0.9044506478714133,0.7497496208758114,0.09728718634663014,
 0.03951987279309288,0.7222152816780019,0.7206518604722841,0.9333072714117167,
 0.8353639484876424,0.7717677706905147,0.5778080511738403,0.22962608704572918,
 0.43249050002886513,0.021578103551285976,0.07386041081425576,
 0.7550165442671971,0.5382476450701956,0.39852540304156425,0.2826916672205728,
 0.892317294982203,0.02176120029317119,0.3094381195696171,0.6165187384405239,
 0.026614485572149404,0.5025750421262654,0.48080411193873274,
 0.3576639224917395,0.2949407837087915,0.924825940844005,0.7797484151988353,
 0.30868567291186366,0.8443232619842995,0.44901058871861277,
 0.34631354709845685,0.48661736794621513,0.5348993706055915,
 0.6722808075657872,0.3654640293392962,0.05025177273464354,0.8445401804457775,
 0.33142096218673445,0.3698799845765127,0.39194595015656586,0.919484119929288,
 0.22343488504814268,0.23510934463653776,0.24384516236792964,
 0.7604520476082287,0.6149975300453511,0.2638834732166876,0.4887150112052827,
 0.3604027442859,0.17743380449773627,0.8649070282577896,0.23740248379123952,
 0.9387983405962967,0.07049360022121864,0.06182635367014877,
 0.42943305541267585,0.024059025965405434,0.6200586964431418,
 0.5055149641057637,0.7956984569998171,0.3953168010776308,0.05211854532049076,
 0.7858872425845617,0.8102963233291041,0.9466806785618879,0.1271357502306616,
 0.7237200300206926,2.028651558186878e-4,0.481215340888428,
 0.45854291306165584,0.5850436998033818,0.11984707302982356,
 0.33290456771548094,0.05560277770033495,0.5699762745111024,
 0.7678652863209633,0.014942192964636947,0.7786781272006891,
 0.39485390442680446,0.07764630497463854,0.4196014338856846,
 0.5955897833910923,0.9778712809806603,0.5156533780470784,0.5289442809451543,
 0.962134567577924,0.3494925735606641,0.4929590100882828,0.2835114492015872,
 0.543422091120699,0.6655623693902495,0.2531129968390605,0.9433348261695379,
 0.002823088274868102,0.5292816726882585,0.4133927071409733,
 0.7351284347144487,0.550088150890704,0.1806455442309417,0.48003924945204157,
 0.2985764407908629,0.48034385662811796,0.33993430423509596,
 0.6754816381648128,0.06344155377487581,0.6682471811398741,0.6793380945843602,
 0.7358214138287236,0.620345173488581,0.6813351232245397,0.4037445691922563,
 0.0430133241865589,0.7794042699212882,0.37445134351921483,0.2635212840395339,
 0.2958013116326905,0.8291909989070954,0.8709892343317754,0.08408654485054756,
 0.5942859113784256,0.10218781431084112,0.4886350973006677,
 0.17268156278043834,0.16729378913131754,0.2522191243576647,0.823849043459176,
 0.4396721744371289,0.0598001434620381,0.33424485874653376,0.5597248282706625,
 0.411318576944268,0.8553179611529848,0.252805529978381,0.8684778503969639,
 0.8349438962850901,0.11043237381986315,0.9972353239873519,
 0.048027901649780325,0.6608298507506009,0.6764228417143819,
 0.7971048285007225,0.5931959265589851,0.025077483276790824,
 0.5039279084111091,0.6490169470641353,0.8959477314209906,0.7115745125043353,
 0.43295163388750035,0.13469878606129082,0.40710474914240513,
 0.9412948437146109,0.4857688090930897,0.47725465271668377,0.7983546889879938,
 0.34832425323359106,0.8296915650298878,0.05644173870596192,
 0.6281569605810023,0.7789760476039131,0.14419560074820403,
 0.49392808995908244,0.33524655715640117,0.23639209204945988,
 0.6854680911853462,0.4451561456715204,0.8199272332084844,0.35156895647385655,
 0.963639087445339,0.25476479712735467,0.40759087046580766,0.1633276184310849,
 0.3563545149866725,0.01835783520851142,0.08427617740845184,
 0.8194381185241199,0.025634048004984034,0.2420110570640921,
 0.16376800974973516,0.009893259627518414,0.16397523938751302,
 0.2121301794240964,0.4740751052347091,0.9199184848779409,0.706918897394253,
 0.9445782451852671,0.5061011980391177,0.04999873822250589,
 0.11499410065718463,0.45236877324567526,0.5258972597049006,
 0.6707894702391279,0.5309784494184583,0.7174172494477212,0.3894546560322111,
 0.05660525007602546,0.5866599236068787,0.429869616759603,0.10966624238458023,
 0.5770936274910805,0.44606728269473284,0.9629849130488668,
 0.29124235242388297,0.2995654886515453,0.7713743516975315,
 0.06968219235723383,0.8536623469189618,0.7197006914418119,
 0.44881457959276994,0.25283087298854534,0.3306137029960823,
 0.3508786833818496,0.5664068171096439,0.21876584329908244,0.2595094545734591,
 0.5845608717969789,0.9973881452961743,0.8145827863102545,0.1924086082908104,
 0.9910438635508619,0.39753728410299316,0.4615085458859537,0.8861476262459341,
 0.15342382170351077,0.4635027877418132,0.9118299660888842,0.7914611341545791,
 0.2036619068434954,0.5038216403210894,0.22284466182731233,0.7622446629175146,
 0.5996356510881328,0.5681314534995846,0.05747193940029682,0.8189889936257899,
 0.21731129958796958,0.8893051843618593,0.7693327216114759,0.4772127662842155,
 0.3450816617092044,0.2659503759877877,0.3896879080415694,0.7914628198077029,
 0.2363329970429091,0.022010049913950835,0.19969022358859279,
 0.26884379683265514,0.30108228778437496,0.5841469309092013,
 0.9259291460502936,0.14483899423268787,0.20256820150032728,
 0.3533831243187521,0.3198349389623738,0.512092943215123,0.7361034285873229,
 0.23510765047322835,0.7279070608758633,0.19822787184819768,
 0.9696672231128168,0.27483686592771517,0.02481710577582874,
 0.08887840726869456,0.21517778774624652,0.6217643282799933,
 0.6029043295482531,0.3168294691732223,0.3414660857164351,0.7735398829854323,
 0.4705478282298521,0.14250885688656367,0.9657318483978268,0.9957206169419343,
 0.03592169208739382,0.47673989269245887,0.5946829453712328,
 0.31468486337502066,0.5734921682863775,0.45389568940662617,
 0.14346231362529438,0.2499404778269778,0.38900509866963306,
 0.8233441537989242,0.7367351215121372,0.420195053424795,0.511132555004346,
 0.941040802751034,0.7874820884970465,0.5168730914249292,0.25189783582945946,
 0.8172901030237846,0.624538502478065,0.7501577578844569,0.9867140784517838,
 0.9988455444185378,0.7104205863882344,0.4961936232336004,0.7832767388318005,
 0.3711639995654783,0.03770863208693753,0.16594188912144192,
 0.1451103740786046,0.2771008626901895,0.9850429851656681,0.10990972404113986,
 0.83160203769837,0.9896913895358272,0.09455002395120404,0.2330567902620556,
 0.10382687140784974,0.7531644594430896,0.5128607640576626,0.9945470072281395,
 0.8926025463793064,0.31997670972293946,0.7780236666545154,0.4286378758906383,
 0.4180849166946964,0.4205836538413379,0.4840981463963808,0.6469471506176168,
 0.4752216172716718,0.721169216108338,0.12042718821357035,0.46698801592916483,
 0.016118391454052916,0.46854982515491606,0.3065515467260972,
 0.8808290342271371,0.4221647768421729,0.6974623455576525,0.7434339576636408,
 0.27951805574851074,0.6677687390161638,0.9629313886778075,0.2797394689977305,
 0.06704561934993525,0.9848778083102601,0.9196688227023623,0.3711678157403564,
 0.5053755398710926,0.4764027739183304,0.5067966369411616,0.2774394328724903,
 0.9148004767063097,0.4122600944993411,0.9615531522197406,0.5670913260460705,
 0.8279931139191301,0.7220025166133894,0.8738469157964021,0.18608612297366522,
 0.062098942961127834,0.909030367044775,0.29476481577796954,
 0.2308370705948587,0.622035697304904,0.3928719034486623,0.9057937459588407,
 0.9483069134850429,0.8023855079527301,0.2833630523311199,0.2327776014731423,
 0.16637557261751545,0.9573269385950363,0.9953049021899306,
 0.12467639110214146,0.14172067143663014,0.4362041221936269,
 0.5689754229453704,0.17646685004323714,0.06789274306324766,
 0.19054131423185572,0.42782056645762245,0.338887233747616,0.8439126322054329,
 0.978485028510849,0.949904044234266,0.6848175676240494,0.9977159329207064,
 0.7063655273718289,0.4708604564374893,0.6675171787974268,0.608650216418924,
 0.8246514506222926,0.8103962681094835,0.5647306340955944,0.10126556290748956,
 0.8360504806142783,0.6797004835507451,0.5648927803338322,
 0.054299165510575564,0.597019688425137,0.5506944438587908,0.9574761639254197,
 0.984791942373192,0.5244237635847572,0.08094200438643029,0.8106226350358054,
 0.2561214895002262,0.7214623727168601,0.5106950382042608,0.12830039380104763,
 0.8675446056336785,0.2537382744042649,0.2633214102680863,0.206807515958919,
 0.27862807606108764,0.32793464820835483,0.19443792991253228,
 0.23403385498290197,0.47618074421322365,0.41778507925493924,
 0.09919569942177975,0.5004060871804435,0.4308639637002851,
 0.053691732190575214,0.06915247740156283,0.2594006756980731,
 0.029658226904388352,0.4155702624395431,0.8058619442258754,
 0.8058938750124706,0.9652701450362895,0.715737287171384,0.39390263930478164,
 0.955635537993868,0.06206884339842145,0.2532484627431255,0.624778151616704,
 0.8088919157467405,0.9065103770708884,0.6011672667683591,0.9583536831190386,
 0.13471465141106875,0.5451134968134419,0.11937698964604837,
 0.4025242990277046,0.47708599006466845,0.5987754047838062,
 0.00859951831599648,0.30482865291429917,0.5094496707577392,
 0.12653684258813236,0.8164350629372091,0.6842242673314487,0.9881873307227272,
 0.2165579247069731,0.27722373928892863,0.738708016113546,0.538541893091148,
 0.8076650372610927,0.027002061455909,0.42067053334304205,0.6005835722191399,
 0.6976085422962168,0.5251160301889387,0.7939477588930174,0.8844578080207193,
 0.1116514333898081,0.7228592492042485,0.3513392069663106,0.6185577050209541,
 0.8104729622698297,0.7421884540779258,0.7384751720355551,0.7588537527586748,
 0.210654903867727,0.37162665968885567,0.8485464436278741,0.18796084604429653,
 0.9008313315033645,0.06195554704929929,0.26774540206564734,
 0.22151676141458165,0.708060681422507,0.9299553062066328,0.6585728376413426,
 0.8924728736310403,0.916523593853786,0.6766147520107363,0.24097975820504636,
 0.5595896245826621,0.35040593191873026,0.24123798066244284,
 0.7344528915255155,0.9536562271139456,0.02934895325353959,0.7406720735191052,
 0.15648602467750639,0.9602717316173458,0.5742079169866087,
 0.08892002157147694,0.06726188499976427,0.3800203324045399,
 0.6690983985685935,0.67815285200962,0.08598746937136093,0.00608004517909011,
 0.7883551101541475,0.23237380158912058,0.574201852797871,0.5686585724852372,
 0.6567556043989696,0.5054263537794143,0.6395719140299073,0.5305977259282295,
 0.013211349038900422,0.23401574375139522,0.9970818154108859,
 0.4426758863881648,0.703900678716319,0.26979128160860677,0.8485260858745275,
 0.8527144206679571,0.9767246483096728,0.4003624736860808,0.779259270332262,
 0.3161294595634576,0.5797705480196528,0.20150717126585416,0.7361787926774248,
 0.77409982174147,0.7637207642904859,0.5134826169796903,0.011706877299194951,
 0.20446648224123443,0.055152909792677285,0.42778628216956394,
 0.15591729338676608,0.9573490906049767,0.02731293328673856,
 0.5224865837005878,0.37076836219772513,0.11804593463928414,
 0.1736379244767845,0.06776170675071524,0.5670257674330476,0.9519961184364232,
 0.5884380685022677,0.20015520020460775,0.819599746139225,0.5787373565343417,
 0.11016957285405327,0.7249754811209972,0.923284123223903,0.5049704238864503,
 0.6818812977914278,0.03024397682909641,0.7727076473606609,
 0.09519595118247248,0.5942367133319614,0.3458019322618402,0.7592077133286067,
 0.1463921228547107,0.9193191983868995,0.4497105596665274,0.2538774080921543,
 0.2290792589956956,0.3424911977321009,0.17406392098515933,0.3721712874268175,
 0.07411902491700584,0.4397246609817881,0.3561983762057097,
 0.43368586956290545,0.3895185161521203,0.08829326273166449,
 0.37842798354981255,0.5019914155946487,0.3667429332118217,0.7980406616101202,
 0.49880856512821037,0.15417656421407344,0.45498487756548867,
 0.10136284298561515,0.25513671715521136,0.13893946406018876,
 0.7262981596320648,0.3108364012507827,0.2945454386553883,0.594022571339397,
 0.2787037509487269,0.38095000417373515,0.4617744471145404,0.2345400180196131,
 0.4962422898702592,0.47471828562139773,0.6419991652026485,
 0.22841294068136364,0.9003707451890456,0.2072070989932845,0.3605217466821157,
 0.2679851800987,0.5708226560232277,0.5175231903855697,0.11592594195575268,
 0.5096269315420495,0.5587083141785012,0.7354674938580933,0.663630416229443,
 0.018811426281871158,0.6868924602925486,0.15870394544759558,
 0.6378856346798488,0.6134816766506158,0.284994779363693,0.5213529224061156,
 0.440969348077932,0.8391274332438702,0.9228303538338464,0.28973053565218043,
 0.8451923823018719,0.052940182442911765,0.6270679341066151,
 0.7060334283168705,0.9766918024596438,0.44759578590145455,0.9908887401663514,
 0.037089616399665126,0.7117543464780276,0.31311106793312504,
 0.657950944752399,0.022706253948602262,0.3728375240617633,0.8548022018443667,
 0.5097018005972618,0.43099276674883147,0.3301579184029989,
 0.13978034368681092,0.19149623150822914,0.9320466083383914,
 0.7781675031984305,0.3105085935183777,0.822622885957764,0.18701176202357794,
 0.6489016102550379,0.7759781145577919,0.07996791696365835,0.7394918896891368,
 0.7638517094258023,0.3808817845021414,0.6886618758953834,0.7064188869452265,
 0.16132895584948492,0.4186586811806907,0.4443600874032013,0.819465562231893,
 0.13394587617980902,0.18854811000252503,0.028810867128860385,
 0.8038992959583537,0.4098599247235858,0.4771954954625728,0.04001620559585772,
 0.013863580079675142,0.2742452767233199,0.08597327902971863,
 0.3981706784964548,0.8789674446281209,0.3040667684519366,0.5938276660995492,
 0.9605646526484841,0.8687511339603442,0.11597899959982716,0.1998821996280783,
 0.5446509545711692,0.6192070631691112,0.09764552537399429,0.5593512783255745,
 0.2609430136008235,0.7080219538171547,0.4000332560883656,0.4146103130616593,
 0.2696195551365268,0.9959953764040286,0.9010471902861039,0.1533719016902082,
 0.9486725591739701,0.5345966578826097,0.8441545188840216,0.34697236871661374,
 0.030683504271910023,0.5184326151002787,0.23476296586847267,
 0.3077490996976746,0.23259057684881934,0.7610101188569747,
 0.38488688731565723,0.9958703339871636,0.3718659289927796,0.602848709118711,
 0.9496504285210597,0.41795287761366373,0.4482821073000033,
 0.12921983637068013,0.3852342902248902,0.5134593267621749,
 0.37952145641599644,0.9499604119199268,0.16323055403053055,
 0.09038074617746583,0.08198149104407981,0.5291384065079054,
 0.9893800643186178,0.391749827235055,0.4649925373879913,0.10780691160153899,
 0.10964692040563406,0.6925027962042598,0.1394917706682477,0.6392541354748101,
 0.4781960426997145,0.03879056310328022,0.4301944583294517,
 0.26334265925295264,0.3056559191570354,0.32245265179352955,
 0.20049172999920706,0.8481518636471846,0.9643740836833854,0.5542364266966954,
 0.6161167484099352,0.5571831320634687,0.04219510110163982,0.6003109631114844,
 0.3303588736216072,0.44780757836630236,0.8060352724785959,0.2403825643197428,
 0.20003373402159108,0.5663336883969966,0.9422028638084984,0.3120343105335085,
 0.6512113508547214,0.43929126347590075,0.6633514663382389,
 0.10358824024652413,0.533349480961288,0.4880945598956552,0.03769402591306259,
 0.3249088060520209,0.8453670388504979,0.8026195617475895,0.8732673702397046,
 0.766384878446364,0.749145921540605,0.9972827555372118,0.8045000546818135,
 0.9773049946541642,0.5352819907986666,0.4663187512709359,0.7204514222512423,
 0.0013811991258081413,0.5448916196943351,0.8870161862953573,
 0.2890762275869172,0.6043644177744845,0.06506211138566542,0.3570427822900182,
 0.8103987418051394,0.7429359667269189,0.09988126336634451,
 0.43778188767558424,0.1430238900738533,0.3239963418944788,0.653862265191041,
 0.3372355359626734,0.6699764878679593,0.6964062575705608,0.6164895237002616,
 0.8631119229107007,0.8386059570558053,0.6094486794635172,0.7729813775262311,
 0.7623323775670884,0.5448929293977279,0.8169326745654166,0.6666285079731897,
 0.6814260209880196,0.02077073401947449,0.9725063761637236,
 0.36828399386624633,0.8849735845666978,0.40162900268649615,
 0.9734432097928657,0.6052699047001798,0.05175048722838238,0.5040396200951198,
 0.6429349829597732,0.12573630530239366,0.2792334243791308,
 0.15640088321125012,0.19466019030216142,0.9331120410563744,
 0.4110152928799523,0.9808737236859332,0.10776884115617957,0.5032640908263188,
 0.16051874284129497,0.14373788774591745,0.2682983462901092,
 0.5407469079558657,0.7369678693650993,0.5014225251404876,0.8967501010329422,
 0.16890083673294154,0.049784309558671413,0.31918674256155455,
 0.12149746509350612,0.06430446914698373,0.8925047678390009,
 0.5875792285567965,0.8760159771823331,0.9313169330255007,0.8832151353506322,
 0.6221124757594808,0.6768103232866214,0.6072144378974673,0.5144088300360657,
 0.5857302740857532,0.8109542522019136,0.7463489070929765,0.2128755262258043,
 0.47245518681042686,0.8954097354825047,0.7444659200578387,
 0.40516869021560886,0.23977852768917396,0.6038964810023346,
 0.09506964464530077,0.9912101476460246,0.1517402373736002,0.7050982879190708,
 0.613053628903138,0.8439941575006608,0.269966360900707,0.5287424590943977,
 0.8521350227906965,0.6821016722596339,0.651232165155953,0.4809555457385799,
 0.6566175625248556,0.12687592242602597,0.4576872984045356,0.5873397938603866,
 0.039585984063043145,0.552470935869748,0.5669655167246663,0.2633967858714492,
 0.04546373972768547,0.40142974942590226,0.12474581751134117,
 0.8132170316644148,0.2335926905853164,0.04273943189057916,
 0.03499798011266664,0.05532687808287595,0.1722754311274366,
 0.9440437366214288,0.48903866941162666,0.538146933919225,0.40697887409691014,
 0.7186831714178212,0.01422635525770366,0.9843671294650972,
 0.06629323420532518,0.9754608674348919,0.7448744000489266,
 0.15586215061807263,0.21001203529831203,0.953458744001633,0.8796956760906901,
 0.14652274968236423,0.4950639514915266,0.6081233392747746,
 0.16909615818900525,0.7005675526881215,0.3914614530485593,0.9599620048836861,
 0.43444249954928704,0.22602439120373397,0.27488946001161496,
 0.7890439882372187,0.7256148689944165,0.20555228197830788,0.922233746151871,
 0.9684078574940792,0.9245738814935865,0.2404525539051856,0.6922007469069137,
 0.9763482204190316,0.06357830466396641,0.9476468896709965,
 0.32527602541013545,0.6264631108960788,0.8242021329399755,0.601952990056964,
 0.5750914437440866,0.7772508494257109,0.04212050549301216,0.3750124549477175,
 0.18235065295391173,0.7365614594523211,0.03072401134219105,
 0.6642257578434781,0.6286785762295964,0.47696567106676435,
 0.06682970334676719,0.5064389117938267,0.6143240328070477,0.610782018862313,
 0.3847981763458843,0.04228275995578601,0.9138018148611498,
 0.48530687787967064,0.28664172138283706,0.3941114082149937,
 0.43310738617363576,0.262274429451663,0.16505611768880013,
 0.39900852739701764,0.5248487871082719,0.8765547473594775,
 0.10625864758546422,0.030831941572765253,0.9086616643503984,
 0.01231574514178857,0.47203808972452654,0.19244633851940396,
 0.3326361094632627,0.8418136830000941,0.611467103351266,0.6861186191722356,
 0.21116965505852348,0.12609636339611663,0.005029786747890341,
 0.22968802481563255,0.7432933102632809,0.5794560514328809,0.2636407974446333,
 0.6592335947367782,0.7008401517251074,0.10471789875742177,0.3789572384860804,
 0.030464233580478006,0.37886829719713855,0.5013309707638314,
 0.3962131588705553,0.43467849070493414,0.6885668990468654,0.9504832832822213,
 0.751156410169486,0.56348666230933,0.1578502406677056,0.1397984193875923,
 0.44611822213756425,0.5437787982380271,0.46054453843487464,0.97361897261645,
 0.46669215973100453,0.9470828571906235,0.20590191594829732,
 0.07733510267433308,0.772803562710694,0.33173286688576353,
 0.32969445863807434,0.862679300359498,0.8078835183693779,0.5525622046102724,
 0.6617467497239782,0.2460560483930735,0.16175351379860115,0.8353693380476398,
 0.2154005289936679,0.5630878213916436,0.03715514851613855,0.3448131817616952,
 0.8694839684990322,0.026741217528047256,0.7073840846983496,
 0.9540302632900207,0.19779675960338683,0.6425086460535544,0.9106159958103412,
 0.4188716836453833,0.5214449665029788,0.36186488713109266,0.6721765699107023,
 0.4922232030500113,0.48911506439249297,0.686050196615652,0.543776922838218,
 0.5340687733130207,0.8856963358775968,0.819824980997228,0.00544748366094816,
 0.7944346907191733,0.5426232519734848,0.24952746658290448,
 0.12419661634077428,0.6172778663141343,0.22012581539726173,
 0.5255174724225303,0.6588881279663632,0.8374742259998488,0.6571050653848564,
 0.7499526417983436,0.6853204464045958,0.9180859456430642,0.9583111625299804,
 0.24444229304422882,0.8554102159062091,0.502618477567525,0.7136829668630759,
 0.12660615201199732,0.2560281391010829,0.889502523533889,0.3557117566349268,
 0.09194442817817472,0.07250950518290988,0.3724359203594123,
 0.27081941286735733,0.9034480599711345,0.3615436154795949,0.9263813454921539,
 0.03701918187470288,0.6180710468135221,0.04436014500854524,
 0.2414756051249839,0.9444213483015541,0.8867845579672688,0.5093032109159166,
 0.8361927831314325,0.8958103978753901,0.3743480910339614,0.6466068644607308,
 0.20410491272138176,0.8820360601411177,0.8470991187521132,0.6288449295013946,
 0.9557789019999119,0.9434217307010436,0.5614985553933325,0.5620923248974623,
 0.5558480973956942,0.39279861157761276,0.94020704550284,0.8524956598678839,
 0.8631690844090293,0.15642440026652205,0.8465037944781659,0.2127116687982744,
 0.5606553588164883,0.7240653415337591,0.056270508207501346,
 0.7732393644112198,0.2346457319648656,0.28119752430086487,0.1970317366798543,
 0.6821107563966763,0.37388708968111806,0.7538567108955698,
 0.12340220736924445,0.11221267276450442,0.00222348627679414,
 0.6085475217165909,0.7410449225654214,0.34693418512270935,0.3891376247929571,
 0.8245900698158015,0.6042076195590481,0.5236928347267703,0.8433902715615856,
 0.5654697051384137,0.8552610719356023,0.2529478690103847,0.9543450765910338,
 0.7917390383183436,0.2279783811112679,0.24521916595738924,
 0.033084188399718495,0.18538644477660682,0.7571627931304513,
 0.461339017356429,0.1868673279704508,0.819742276755543,0.8984808358869043,
 0.5346098615141718,0.8566176255439756,0.9901628215223239,0.3885657661099646,
 0.8530462354776994,0.014793358076772178,0.2186115600439349,
 0.7158085442681327,0.9779068649527654,0.9643619682901534,0.19599081622149628,
 0.110230709307512,0.5867273217957595,0.25090884717270734,0.1771922097610269,
 0.09191057315249496,0.6320723700016644,0.6101917007629054,0.8137918299461329,
 0.3521929072585136,0.44076976713585025,0.3730505861597988,0.6726703788182788,
 0.6422379749798106,0.22269949015207036,0.5705919964993678,0.9605074977234126,
 0.346581860874715,0.30579104210965913,0.1329187715507384,0.6160910830692312,
 0.002549043271513396,0.1864034568294234,0.5867181516052036,
 0.29605216038656956,0.7492549962540473,0.8598573006886532,
 0.10075308692300133,0.3273122025606975,0.12444748310212639,
 0.003597139293020968,0.8354410419467031,0.41159478828017415,
 0.4398438845155058,0.07291148224372013,0.47170144649175383,
 0.4876878978184114,0.5480323819029197,0.6676986055027473,0.2276752313439997,
 0.791932211311472,0.6362144434586653,0.04720202601991119,0.12953382713486938,
 0.5101193865274208,0.39343118052682347,0.65438312596499,0.9809002037159215,
 0.14731657070103787,0.9071821891242098,0.21862350943239073,
 0.16896639370606892,0.026933668955798518,0.15047219914691312,
 0.22721906769929834,0.5313305032783376,0.9073765173490174,0.3317027962971959,
 0.7990360655549851,0.16701270944548907,0.1276975644896765,0.7010853036770444,
 0.3709815552124788,0.27152645804292164,0.2197871328462272,0.3577242663154405,
 0.5642305698055542,0.05192342511639214,0.16375592499403235,
 0.49746895852417783,0.4714470006020949,0.8042188590585304,
 0.44532409737528933,0.6626856387450587,0.6043685213107203,0.7483651781534257,
 0.9672135633134266,0.028225832682040153,0.010883157643905239,
 0.28481628955088534,0.4024428627127199,0.4190582709280066,0.9679023120684849,
 0.8123113781794267,0.6351018410917165,0.44595041297407156,0.7321802599263051,
 0.6657958950101784,0.2652824151121591,0.8646864221709112,0.2770110748534478,
 0.6611141352339864,0.5585380277393471,0.1861469029395445,0.20878769828245658,
 0.24245936036266846,0.7879744108250207,0.2990402570144445,0.9721311312964427,
 0.06638384823509702,0.6718875106897121,0.2546379569204753,0.6893588923749732,
 0.08811338819667547,0.16754607785740994,0.3548281189085215,
 0.6549773041370937,0.9381330395118621,0.6004600123513284,0.5899562584249831,
 0.9259854553740896,0.1581437969552577,0.19127105073184547,0.6748196524486584,
 0.38786717235599644,0.9421254836649949,0.404554857934883,0.8118695589233433,
 0.4368020893173783,0.640674225504179,0.6010885777545083,0.23803175286368872,
 0.3011720724782613,0.6615103416997911,0.916543134780242,0.8458494998082138,
 0.8643051624831095,0.8337182230089184,0.8396004888881273,0.1133586048604539,
 0.24065831863277298,0.9643919513755875,0.8979929615144401,
 0.29500886929134473,0.2683127126078251,0.8736838156492123,0.9280175946584374,
 0.07850476224369651,0.21556126266403808,0.8425758893640926,
 0.16803340065409167,0.17580370263329892,0.06219479650529236,
 0.5842242233466692,0.39203605496506455,0.7488087942088205,
 0.05387171805639246,0.6326897515821993,0.8745849299720274,
 0.33781441657151057,0.40029888127138047,0.7624237131697799,
 0.8359051537726166,0.11133896618057548,0.41356596983012284,
 0.2464888102518642,0.6552300843066012,0.9476569723299837,0.4812611924925351,
 0.08197599737090655,0.7036497277534547,0.014353923843245964,
 0.4088699111344559,0.292345582208553,0.3319414534820483,0.7355154152894078,
 0.36595561907154295,0.6034073878389798,0.4473372603916298,0.489005541563045,
 0.11628114839883696,0.09602300096203198,0.539324501292504,0.5693986711295995,
 0.19466093586672784,0.5594366409486173,0.5003925574122581,0.8583323374505969,
 0.5610497753843823,0.6778339238362893,0.1064245651166339,0.7606800199024167,
 0.5883687650957949,0.856174104704584,0.7302192317605601,0.35013132369447686,
 0.29651653068286976,0.6099625376037152,0.023361826905503702,
 0.3252595808027273,0.4990998975126336,0.6277133394016073,0.13961462510690648,
 0.3339171194311652,0.7337673568213557,0.3988451201571793,
 0.004346036017003674,0.831311893185048,0.3995840574783248,
 0.14982594255812476,0.16955405235110244,0.768444723856859,0.6898455416652282,
 0.32635266170809896,0.6503093696107077,0.8578102843851265,0.6700800935523751,
 0.36108542530334686,0.9817008531150473,0.8881890227761007,
 0.47278448221633684,0.9872853071722779,0.15811701340471374,
 0.6615698779430609,0.24756310944548032,0.2542156136369519,0.7941859792087196,
 0.9146432097405811,0.11877669536463809,0.816407404562654,0.7043478194584796,
 0.9150040615161867,0.5547731142606986,0.9406815044282894,0.6639698798088798,
 0.7740501020837025,0.1403569951872915,0.354592492603894,0.2191642016493065,
 0.3694199302998654,0.6787920165818103,0.724526403278948,0.09002212011502242,
 0.9837258905420138,0.39252624935951974,0.41567070213824053,
 0.8387776719587996,0.04402377151692527,0.17468425065494664,
 0.3456156523269549,0.9117770806152511,0.3421999704488906,0.4114312425300347,
 0.2574358322404142,0.6804375497646791,0.4451508688623551,0.9556903199134407,
 0.05894827115889856,0.7084225479702275,0.48903799723185704,
 0.17998355041183045,0.5951178650962007,0.2509326946521281,0.6747352615761872,
 0.7412085962631162,0.3314317606892647,0.07800723784537289,0.6668619595834155,
 0.6727063795181484,0.846504246825871,0.3203073595536955,0.008815121642791413,
 0.25388848301227396,0.9486328429207389,0.8968997823075877,0.2444183363792808,
 0.5970378234891378,0.7871215146362751,0.9682973372628558,0.5292425832107425,
 0.1712896590089953,0.021453684036178977,0.8849054122891634,
 0.4652001765489294,0.05501867854298159,0.7033479124078317,0.4424131868162495,
 0.8924501898004621,0.24986429212969252,0.9414305149756952,
 0.42249945575695724,0.06624120484877527,0.2126364961797529,
 0.6832061939837658,0.5922408667919824,0.01316346781944766,0.8848174651256228,
 0.5832779685439251,0.9050376260521938,0.11700459642358241,
 0.29094487313310946,0.8024025843632913,0.4775837580708394,0.2900466543814437,
 0.6325870554451992,0.5564373623120109,0.4177553815506665,0.715050943946625,
 0.9395653254762263,0.4411145859700699,0.38908969314492936,0.638820094145394,
 0.9829597925993624,0.7469184049351769,0.33490733341055134,0.2089542588872897,
 0.4746532088274561,0.86735440205101,0.26091691516448223,0.9080078945186609,
 0.011507352815119054,0.08945113433056484,0.42930327050071615,
 0.8076913823762679,0.7423775764411449,0.661403451908428,0.5848183989076632,
 0.2543090655055782,0.18709046352045777,0.33608445863046055,
 0.45929381136581704,0.6923734730335888,0.6785949754112436,
 0.07246356072733673,0.6661292104555455,0.6665230549572657,0.2411969979040497,
 0.7049818668418556,0.29734237357530335,0.05607983306977504,
 0.7426136975319689,0.22393583496932978,0.04278497782233093,
 0.7733876238433782,0.994879118189584,0.8222964623668452,0.24578427723012186,
 0.7827024623699763,0.4597483159398894,0.8255569399023928,0.820329099268859,
 0.46600931444480886,0.8995193047693824,0.030574240933844976,
 0.3018277723284586,0.4391250102243289,0.31978189507427435,0.6933217086637486,
 0.8365911111825941,0.48919710978098196,0.1638620965718639,0.5716585589075511,
 0.32211607778605256,0.292666053523976,0.12713966043754454,
 0.030996575718746477,0.3984216317924302,0.6718114068379326,
 0.47550697919326634,0.9858464033152453,0.950616221741508,0.8190504605800569,
 0.28185124385605287,0.03998752850071452,0.004308753559443157,
 0.5795242270885058,0.19874711057173666,0.7223781110947636,0.1450458268722462,
 0.5693604119740645,0.7125739312670388,0.4509360532182246,0.7513091977261364,
 0.02411255193398687,0.39664930496251616,0.7696764134546286,
 0.05830166476727294,0.10239582569331074,0.8699255794757116,
 0.22573445165170714,0.2426808997295109,0.11724986731006792,
 0.9199670298144443,0.6305149119317208,0.18610230038537356,0.7391629815108656,
 0.4830738813968587,0.1448666744307231,0.49392110610746165,
 6.672146271485824e-4,0.9637304870099542,0.544322992508458,
 0.39363929752151827,0.2585336223034864,0.3500068534594645,0.8833132001353089,
 0.08849949675695379,0.49826810729465987,0.0930402601862742,
 0.6345913442604905,0.5559809617194121,0.9474522978894434,0.8423721929655138,
 0.8392950171099886,0.23852134623033283,0.4498075926154662,0.8573322359111584,
 0.43399914103007264,0.6188711849218412,0.7655316136389679,0.6969949438615868,
 0.20372927208252112,0.17891316663635137,0.8833968222036377,
 0.8658190936312364,0.927693891026413,0.3917380037695949,0.23684750710191282,
 0.32122464147657515,0.6584150869527939,0.5577402893667952,0.2358830911779457,
 0.26750267745082146,0.4384997632876084,0.811655617509117,0.13591303338994198,
 0.8869332877534822,0.1802153766814566,0.41657377262623896,0.5202260447497511,
 0.5394697721909217,0.4267689651644129,0.22002645453878322,
 0.21344420835677747,0.6258460404264531,0.7190843807692116,0.2801767602001518,
 0.8011312947815332,0.7619559131407513,0.7252593457243199,0.7600958463490999,
 0.0319006809485336,0.7968817872006402,0.044197231171963836,
 0.9488168230064098,0.8595367346121812,0.9081672102843867,0.8712904064605052,
 0.6288463562405259,0.3499702033922678,0.5895406693314675,0.18536438586035453,
 0.3912694177001521,0.9549533005476086,0.6640928426837599,0.9372885992329426,
 0.42681145345880533,0.4099257612453602,0.4581951242327923,0.8344768446589927,
 0.4587296320392765,0.58421572380179,0.6598342481891976,0.3303021509817843,
 0.980223530179552,0.15126794417351141,0.5539930700069955,0.7464198337168677,
 0.3727627878228552,0.7009163518154325,0.6913603999738284,0.5042022294096198,
 0.21353954519328333,0.5305292535431834,0.6548553972506855,
 0.12986779822144412,0.670346609297464,0.1056663843625274,0.9318163041592592,
 0.7829754361325535,0.8156711284161547,0.4825785683265347,0.7482385467025788,
 0.3643233274646831,0.9504329563943981,0.5358679958471041,0.15349065663295824,
 0.39339153447002384,0.35081905162692506,0.6535322897189795,
 0.02567927177538698,0.642675375113795,0.6094930811337855,0.02296395530141626,
 0.7147106074806326,0.24649444715632696,0.4824669603887498,0.9392056964382733,
 0.2604477603468127,0.3520780964348885,0.5064628288453903,0.1884833286290154,
 0.8048019401674471,0.34570563230897944,0.564901995556534,0.4125237388825782,
 0.7408531377559292,0.01905098079961387,0.3248886734239751,
 0.20075242881564348,0.9860200524895514,0.2874312702551315,0.9281866182318885,
 0.10093999673209408,0.7098030535539208,0.6962678468384499,0.9245495457018522,
 0.373380199009694,0.2711213614493727,0.44643947479681,0.26938964602317783,
 0.06407654809338315,0.765534578015176,0.24812386292556987,0.5290552485943536,
 0.416656631672507,0.9232490362022738,0.44448249873577184,0.7828820746907222,
 0.7290540159724803,0.4994242068019492,0.28730502120366974,0.9648202328011258,
 0.4220478808169136,0.002753033034406438,0.03168842733154198,
 0.09808491775722061,0.860611931724233,0.4858982567073442,0.5641464935453975,
 0.5860972169050344,0.56713044433395,0.9113382632561722,0.7814532461957143,
 0.2232630771589451,0.43551063424980985,0.7136249282084881,
 0.28285999684784346,0.2178627764961495,0.8966648731338301,0.7978134260518824,
 0.45613403528486707,0.10277036686219354,0.7736712764250823,
 0.07277907757307978,0.6595144398322398,0.5739122561714223,0.7900263269415844,
 0.3271819940396369,0.0956146471497975,0.5409518053246802,0.21578176423781348,
 0.6701455884537549,0.506765699742238,0.7212494759126966,0.848309821808916,
 0.03288170356328268,0.12492447467410439,0.4135782121994155,
 0.30134058612954384,0.9026224866383171,0.6926763221296139,0.4613322257526299,
 0.6627413870451861,0.6837957124544858,0.8475367320823746,0.3296883899232407,
 0.6967974964403959,0.6303498400939526,0.023986492701798645,
 0.1210320104558541,0.9900040709954666,0.10308353197814113,
 0.18121766902405734,0.18048979973804313,0.45148180742000343,
 0.6593229334672397,0.7064973803870125,0.9486555365144074,0.7104887661849947,
 0.07747754346662583,0.5943606200247117,0.9159994171389152,0.3641210412391105,
 0.5472677966659638,0.20705608031362444,0.24161541454146895,
 0.5007749383174305,0.47610513144862654,0.18810907805260335,
 0.38926593158074163,0.44313823804172947,0.19527263128742145,
 0.6117966692350861,0.10972762222415433,0.7754185644088046,0.9686372808380732,
 0.14948726463075912,0.01953159342204147,0.9950538693444396,
 0.39445152934021344,0.4845758898200664,0.4268351241040038,
 0.33495033866117685,0.37555222619720485,0.3796484404212197,
 0.8476838289256159,0.19649140276174304,0.9842699291319585,0.5328425833878945,
 0.4483005728119229,0.37949926288789193,0.45588373874987054,
 0.0941327453464148,0.5705222392400807,0.8642844579685032,0.8930486331907062,
 0.3235912658665153,0.08820863410365698,0.20713576943064493,
 0.04753612034155963,0.7545662775685937,0.7041105460756715,0.7610253678735861,
 0.8984249358705719,0.9774224192080098,0.23798579957139665,0.8928128510667732,
 0.33575194762793126,0.8666575462514761,0.24663729556681724,
 0.8319996933622534,0.49284107359724183,0.2516979243770233,
 0.21669578302886083,0.4463851938357907,0.029660184907167264,
 0.8700766861880845,0.5108448867069836,0.25652376577488867,0.707986802968421,
 0.21449678140788908,0.5573961191919895,0.6844187231244573,0.7812422347982084,
 0.16067574131270546,0.03424537878300804,0.43558659137759026,
 0.6602632005777818,0.16193606169567354,0.41364668714392794,0.073622880813206,
 0.9443905648696584,0.6668358216671824,0.7116151516276056,0.6901615989289177,
 0.6415618221462458,0.7748790146424246,0.7723783327165135,0.7115508002610577,
 0.5615336555673947,0.23836882678147608,0.23783779370901803,
 0.8544220588068588,0.444356796971349,0.028083208849504993,0.7556013903125016,
 0.5467219516627518,0.430537710376226,0.8067842320723498,0.44054001466822656,
 0.9311486689962151,0.04175941442814557,0.03749947267318399,
 0.3986138007703629,0.8950578254098195,0.8735497229855912,0.9593814877100746,
 0.9420793750978564,0.5569202916077043,0.2669468371115663,0.03855083728437714,
 0.6774888017934022,0.9871006375791606,0.8970014091293337,0.8242655238210566,
 0.5322107759192907,0.7897264523071057,0.05273847257639608,0.5573694209547316,
 0.6104512420825176,0.1661558323038862,0.301831883534585,0.9604415582734837,
 0.10851460697473492,0.23421872557508128,0.9583620482106691,0.635828173616999,
 0.1305806049954199,0.3377858084599453,0.44693803779985763,0.8197560032769476,
 0.36086421689336456,0.9827425874065195,0.24653870390081412,
 0.5973255377880192,0.7480319785220758,0.41242498783947723,
 0.10553795515171338,0.12001189261182044,0.14802571094600636,
 0.3172707697011805,0.241725260018555,0.3688301599936801,0.21614103932219764,
 0.7145727704239263,0.8162122103101993,0.8382376217244714,0.8724198679114014,
 0.6037442002370592,0.32934299303898984,0.7313356457120712,
 0.36649271290652097,0.0021544834060511775,0.654311187946224,
 0.5744073912991728,0.3267812611294787,0.34816674043899165,0.7390988983482736,
 0.6859822415633774,0.12127805924425894,0.713947193592254,0.9503758404599618,
 0.4845284919024172,0.9683633356573396,0.6282377342606726,0.4177703716630863,
 0.568487351665207,0.36192086195567585,0.7629844968874996,0.6969630562166416,
 0.8822332967246338,0.04836020887935444,0.3455112443101538,0.3092584155179434,
 0.6758023455951219,0.7375364994166346,0.4216163039274803,0.19276213385099694,
 0.4489986941958666,0.373327095735109,0.07475458542768187,0.03189395567313125,
 0.3054342468219037,0.1830523936065933,0.21313595349367198,0.5820239616702335,
 0.3795917630066291,0.888042161659453,0.11196043042817017,0.6343088071836863,
 0.16055088797382067,0.07795547416334359,0.11015994818317987,
 0.8201392418889256,0.6993561230088139,0.061472529766376405,
 0.47989594530220114,0.030499377483571233,0.4140355010893575,
 0.9596563836450521,0.12113994207086776,0.9422521576485715,0.6643087246343445,
 0.38317911312505826,0.3070521772909852,0.9316896603128964,0.8480795721978799,
 0.8049772470105792,0.13665022617438782,0.27377873603953695,
 0.48521800042166385,0.26318959643179574,0.03584422308588853,
 0.9441589652258343,0.8111558426411707,0.2791939612379707,0.03294404472975376,
 0.37245530772657565,0.7182528976667619,0.14823577593288118,
 0.45737509822310873,0.10915243916654771,0.3641916412739652,
 0.8330201896543161,0.44946367605783033,0.03751151255374929,
 0.7485927200285116,0.7383766663553896,0.41684148814723554,
 0.31161208008191765,0.03865537009321807,0.5153179690187734,
 0.3680542306958656,0.5553459752620202,0.40498048141746035,0.6398293569759307,
 0.15647906401087586,0.6841580268072113,0.01720918148298689,0.842916798687308,
 0.8398374974609741,0.5821746573735821,0.4704414275569463,0.7708669597972475,
 0.22868981615286366,0.6741080272469153,0.7435758868127422,
 0.15819422446419873,0.7057626979767081,0.3324789748501584,0.5173057977743695,
 0.36831927703744993,0.4017225661461675,0.16757474129174954,0.924126599357355,
 0.8907213276750225,0.8316025367663147,6.341844524320184e-4,
 0.44463001017114867,0.7081426944695606,0.8187515911170484,0.7506543783462445,
 0.4299176141468335,0.3557267546733657,0.04023491014814251,0.6835750782548282,
 0.9763923172277034,0.17042446211201379,0.6809342010013327,0.7136968950039576,
 0.19024492700945483,0.5113205051470001,0.05834954414201132,
 0.5714188723328844,0.1618648569530483,0.6445961622323352,0.07179871228071588,
 0.94221211147908,0.5148077858296524,0.022983988056889504,0.9754055178723244,
 0.24126853488296385,0.7864657137668122,0.49152929465623285,
 0.7320918453414821,0.8018984996966921,0.4312088673774802,
 0.002850923625368207,0.3394171451919642,0.3807604959868879,
 0.9366364580847084,0.9478931589337817,0.468786696469651,0.4279021471234923,
 0.2212126417273803,0.4743148005746729,0.3553362704779961,0.8917115797566175,
 0.12781607734121048,0.19702638920262028,0.48552796993243164,
 0.2402202371969141,0.7177115492218882,0.8400106591486182,0.5864465377997147,
 0.8452858993110727,0.15161030755699512,0.3819328928121116,
 0.18187677566682248,0.29980064810222784,0.296960591514949,
 0.05606171500002166,0.7040156349411181,0.6082549204649452,0.5249876513205052,
 0.5916643497277729,0.1947809677294864,0.7547138969523168,0.9084702929066154,
 0.4424431944352484,0.4770873270449264,0.9297189138527935,0.5745058478850985,
 0.5932575722478225,0.9706737196472123,0.8659774241625147,0.7045178995528224,
 0.2970369785066955,0.07057521802538103,0.7990054374968281,0.7403954041822015,
 0.680816652857917,0.9604828198623482,0.27889765037394076,0.7616007453667589,
 0.20826367754472197,0.2521494940048259,0.9429670939711956,
 0.34519623386643306,0.2307203836186399,0.756175052863893,0.28456050808167177,
 0.6739415037867809,0.17948417394253935,0.7300799153976865,
 0.14099122716486245,0.43123584260564696,0.41680046680770944,
 0.8990781097877055,0.8897960079045509,0.6999141140053526,0.49615533301549486,
 0.062355676748086286,0.13099004069888887,0.8811801292130006,
 0.16340943770621374,0.6980555773952014,0.17530318801054634,
 0.37430555767393414,0.2021953659337301,0.8752857908354652,
 0.48300064643256235,0.4990986559543321,0.2295884577402112,
 0.008092577105928034,0.33256628044567804,0.2966853885097567,
 0.9071751514414608,0.5118174784693742,0.6131605496929042,0.27514004642037027,
 0.527458437300385,0.17377767044316073,0.9075623469807652,0.556142954783494,
 0.1837134412706637,0.32762948508712353,0.6351685822506228,0.1241278159612701,
 0.6764273556945576,0.7629051530146297,0.31596431505300526,0.897368432361681,
 0.1677509132170656,0.18984093583049266,0.47091699192929304,
 0.21952867588632408,0.8929345825979139,0.0860454156415883,0.3079117058987436,
 0.08153354397109347,0.12604330028136168,0.29447945590963287,
 0.0804993635577218,0.33012601296492894,0.26524585886878294,
 0.38147466720439793,0.5714471365509385,0.007397382878694958,
 0.2697632767083815,0.30895166367465354,0.22454630334991643,
 0.6366817471203091,0.5860641062653306,0.6057804582739829,0.8886590983311139,
 0.9044424865763033,0.818665832039139,0.3165414209499088,0.6795358431234079,
 0.6810409209140954,0.6131555927031578,0.9292792664058788,0.8560941326775391,
 0.13130108102966143,0.2867749127290663,0.15168561529156,0.30116953386373435,
 0.9019631310379775,0.9723431952683917,0.13704480586997336,0.6206209680399158,
 0.7393863528304526,0.47833005386013194,0.013078143492114114,
 0.788830920662417,0.3766502971190936,0.37693149080875843,0.918593836576814,
 0.7823158479852663,0.0421445509346996,0.44709302675204277,
 0.48549201091431227,0.7454200336723494,0.9041494348268861,
 0.25129755841267565,0.8847962283155469,0.9888669976071145,0.7148691663242893,
 0.9195202261207887,0.46309922420020555,0.1180962796118592,0.8880882909154069,
 0.5900627866045277,0.9055802382843249,0.9220506523699484,0.7770433972467233,
 0.04025781097799053,0.8878692404679167,0.2582124252590753,
 0.36639749451641934,0.8855823623322765,0.8609814451986615,0.6868721765155018,
 0.022016648296202868,0.3498593438767206,0.7311253800494535,
 0.1483726536910236,0.1115353267383038,0.2624918413538597,0.8682289985959026,
 0.8816404366962834,0.1113928090938231,0.3998743674587777,0.4637429236941477,
 0.9217784066785757,0.4260751992499411,0.49503423803243796,0.5819064813716277,
 0.9685911229280522,0.43613342795899246,0.2390686149264517,0.4827761587929199,
 0.5393111165224835,0.6076282195448749,0.7019729856088299,0.7241289063965327,
 0.21413056331901092,0.367782430954529,7.030937164956841e-4,
 0.4412308092853713,0.16919615301810975,0.5552915114423892,
 0.14893774660799808,0.892195888845254,0.4115985515461307,0.4602497608506855,
 0.6241241078099189,0.2244022728270061,0.6040129974160107,
 0.017313935744705855,0.3291430438714489,0.21569805831029554,
 0.04130091914319589,0.4025529681326597,0.7237054960508585,
 0.049121398449460174,0.9825246573755408,0.9996579245101157,
 0.35317921450932976,0.9271217674907475,0.8307249779813192,0.7113225125729729,
 0.2292547752301931,0.8224377767121058,0.8985035526529914,0.01813179230343076,
 0.33606799784810426,0.7043139768479005,0.33468728077459753,
 0.6383685597037994,0.2605380031554171,0.2801094790713482,0.06471101174990235,
 0.6861612064819914,0.9281579672164637,0.8568427759744797,0.5063566858175973,
 0.6610177171170564,0.06268667301858522,0.05065280326707744,
 0.6853995160674891,0.5731803888785056,0.9307514298279143,0.6372849808317282,
 0.36664472748013577,0.7743670141025067,0.3351151790267719,0.2209469349964036,
 0.30941588005476794,0.9382467758545253,0.2873468283184266,0.4566231446445308,
 0.2643432673563991,0.9289857600537303,0.7183924488064467,0.7978640666266454,
 0.9074970790639082,0.8709777591592367,0.4639056230181796,0.4880265393378671,
 0.8233663188785343,0.7646287056662456,0.4653338875628672,0.7740884340727634,
 0.5852703453658864,0.4473966719118869,0.43763809010044974,0.4908114662692429,
 0.6632598401417045,0.7800356415476006,0.7526093790053341,0.0832216703911891,
 0.16165262196000119,0.08445702242477371,0.004065479576581765,
 0.049043246556658815,0.2172359156795901,0.009832748128914881,
 0.890658541364389,0.5055275544722334,0.813118298088031,0.7116959445480134,
 0.9259944781946928,0.3406615436910626,0.43748683102140073,0.7385823776950656,
 0.17374993970644703,0.31675607207443035,0.48103998714516516,
 0.37920488570953625,0.4711334419643485,0.504340532547713,
 0.041086165852464496,0.6676268378087776,0.9571281516625514,
 0.5289569135775846,0.040837471171945805,0.39631729714575525,
 0.7446006630301755,0.5814812020665612,0.07418032812322073,0.3078087521077958,
 0.4076188503353849,0.3116404623708905,0.9233742591834995,0.8421911668786048,
 0.8604473271281236,0.9786538952019428,0.4359785129169028,
 0.025138399906608044,0.43796620453635693,0.7472792758831983,
 0.9970689780967661,0.15354036017469674,0.23911553277021302,
 0.4148953038732154,0.10536932725776493,0.5813642284319731,0.815884593832208,
 0.17587375535233052,0.32997383238579003,0.8445185484667593,0.750529504271678,
 0.7876987723299809,0.89389941402993,0.5568520790421654,0.9802960978840908,
 0.8939193925637393,0.08670247718426782,0.024946126559849402,
 0.9664472948527135,0.5922680482766662,0.2709365565805746,0.4065551627370937,
 0.0035943495863957065,0.9375019541052542,0.4734775769738414,
 0.4116375793978795,0.784486996495246,0.3349622119743002,0.9853317122735209,
 0.7576833477618251,0.5247310817543405,0.44158312298546065,0.5705717700455785,
 0.7976821690299724,0.515756747975886,0.042966250500794945,
 0.26192977538175544,0.43689118117431514,0.14496147953676308,
 0.7599357755138363,0.6089521025940314,0.09185659153879788,0.9612355363454789,
 0.26490045799800555,0.0838040621158872,0.29058634995315924,
 0.7947651768368702,0.8616870136904292,0.26546102505017943,0.7088274672720942,
 0.338835677361199,0.6982774365249297,0.9209254910571646,0.27108006572672805,
 0.033520351504073065,0.48002947080658054,0.0443190520508171,
 0.5381502652632316,0.48500936458682653,0.7381563468532978,
 0.25275951860697965,0.9990189266828626,0.7894551201967775,
 0.26282948793470273,0.9567110509958519,0.0838742200841276,0.5071720064328985,
 0.5978799719873015,0.6812109738015849,0.4250414153353279,0.5526190962536099,
 0.4407659595246124,0.8326815094517,0.6870028264447124,0.058514903469779944,
 0.26247945732582945,0.5303640980246571,0.25289843314973437,
 0.18610889549068033,0.062205308059212316,0.4112809939668324,
 0.0853117287383377,0.11633282075925222,0.9330982795215672,0.5938122726465983,
 0.8225864512001635,0.17535806764082995,0.6394919714483674,0.6607409677012951,
 0.28919875675263484,0.6061621994929101,0.5317756670155056,
 0.22926070287160727,0.3794690719526459,0.32816973357680923,0.501038182452469,
 0.9291687926862674,0.06212057055863629,0.5148196418344215,0.2121130256894349,
 0.7319417443467345,0.4412236878992869,0.4572369554758282,0.6080987331730837,
 0.7567132608604827,0.48210860260360633,0.10837355899583312,
 0.43582193818957427,0.13380775346737273,0.1912556405605197,
 0.17240877430175872,0.12735889127530542,0.5032509882533099,
 0.6122492714589541,0.41530235571539564,0.24320287343718738,
 0.48702419138872377,0.8498753597514783,0.46946224887485966,
 0.1372173500937599,0.24500032205854194,0.08131120172749096,
 0.9499638940033842,0.22907497082031492,0.9859107443622445,
 0.14792329124526105,0.4688705475941748,0.35414003235559677,
 0.9213986219249358,0.9812691045898113,0.4355006924137984,0.18846937760647775,
 0.8094063598554525,0.5506431537195171,0.4204437700211976,0.04448351010750973,
 0.13457661752593109,0.5228341096165625,0.21835784508192868,
 0.4268156401731136,0.31374234906088794,0.2643697379927342,0.1578646392815768,
 0.16951790021244995,0.18307442101879312,0.30386478759736313,
 0.4809672367465665,0.7703171122534177,0.7564254979692403,0.12702987221667905,
 0.5419368912979189,0.10640892414706671,0.23246324639772464,
 0.3417670193350957,0.7027039574003774,0.28003390818440865,0.592526445062899,
 0.1414478445193651,0.08894270416766814,0.10335268239230566,
 0.6963250739110564,0.5896914754814225,0.5356370440529891,0.7185316270551567,
 0.023802044389270804,0.6410676170820271,0.08921576630105599,
 0.4361412635196653,0.37276013203691893,0.960420962661062,0.6044090278989904,
 0.001771127643017345,0.8205032819417658,0.3924890420436231,
 0.3375148828486797,0.8264599832835645,0.31727959861450283,0.9417129451468278,
 0.3335010033800314,0.1557761836308016,0.725174785742596,0.24165846556677162,
 0.7405788856330292,0.4857833623374199,0.33077931058771226,0.8600745092688433,
 0.7353921419364117,0.8127294589962246,0.5076912619391285,0.1986576350542587,
 0.13293110219555182,0.1486142513576847,0.1164954153335982,0.7595101120751332,
 0.10654373368163894,0.929443127453313,0.5840338452691429,0.5481237314440706,
 0.5697464324530721,0.4886270815590247,0.253441623063113,0.06274811892633614,
 0.619783219138945,0.24507781855077215,0.6806585425180939,0.19881501279482328,
 0.7642820691968281,0.0994196133402605,0.3063931064579515,0.08264671805646584,
 0.46692380714637216,0.5984031542090271,0.7160688706264606,0.559530268270223,
 0.31948297396502434,0.844739346385355,0.10318831699821018,
 0.030486576297113488,0.8761391726419067,0.06968083645200096,
 0.6005589904121025,0.9250560901940709,0.9604450185153058,0.07963420639822161,
 0.1319802018213142,0.2103592519958174,0.9264121341887325,0.8790062393786324,
 0.07708661503253733,0.9827632168762122,0.3729008255503785,0.6651933034513396,
 0.862312441252943,0.6977265411638941,0.15254556959875332,0.03746289233539857,
 0.8068039765468048,0.3625846926978318,0.7132858216816574,
 0.024261624921975034,0.4781388037887552,0.8976621634649262,0.647748353410433,
 0.520965791961337,0.5237803154658446,0.46807738721161807,0.9415087183441546,
 0.7069345304197634,0.15962080010759827,0.9056988727894612,0.5201775066002248,
 0.7068993440947845,0.8381155592696165,0.8896979441918174,
 0.017233481417657015,0.5582559853217675,0.7623212486265348,0.828688129518678,
 0.41421801163212413,0.6388464128053475,0.30072902907970844,
 0.47471329330897083,0.2598387008563019,0.2988428736400063,0.4008583899746252,
 0.0939925110545089,0.24073499587188352,0.9618181666467063,
 0.26404446073472143,0.40362503004217465,0.45335288431603105,
 0.6913063016102952,0.6974786890002267,0.5783145250906421,0.26283927796143824,
 0.037908969420133,0.4344548206646035,0.6964286553173764,0.09067071729070464,
 0.15702547811752043,0.38041497433389604,0.9644746431040522,
 0.8208777955029296,0.3141429459930909,0.5487748835532615,0.3241316667624615,
 0.8267938191724579,0.39856132987172743,0.914307029591288,0.5340964549713162,
 0.06337243817657356,0.8684303305363411,0.9740984761099544,
 0.18597023449632333,0.5413445287365648,0.36528034778111473,
 0.7597222228429286,0.7314528966163187,0.4924074701570328,0.43783386222626863,
 0.2896283579733321,0.17594551882688536,0.2252640631360795,0.7771818067706032,
 0.1056244896616575,0.34226643529348255,0.8315167536130067,
 0.02525362707048362,0.4087226269190254,0.99450919707122,0.3429321480500477,
 0.5926650237803635,0.5056314216757463,0.5829199731831807,0.07814556622477975,
 0.47989994003456027,0.21012658746178448,0.3521497089726411,
 0.7576791115514817,0.8774708041199247,0.4712255604576585,0.5825104780670238,
 0.6586701990340997,0.9172214383820055,0.06293801633595142,0.5831148039339668,
 0.39081915510625476,0.5354604631845934,0.9357447093955269,0.5880280355906933,
 0.00658471633654234,0.7694958257556508,0.9814849764517793,0.5788415676305285,
 0.809501147520128,0.20669601267492876,0.39474975370853294,0.3410389928825964,
 0.26909128034610386,0.6701528776218919,0.8640558793617341,0.7099635908227173,
 0.562420638851509,0.4724407102191055,0.6079204150385784,0.4064474913327448,
 0.7538418632143544,0.4134673842577402,0.4189569321877117,0.5905269821639767,
 0.23834961714861935,0.8168670674883535,0.6107721903177235,0.9327881021074753,
 0.3715210882898049,0.40891104221255703,0.9443954817100048,
 0.32908104815556327,0.22355175517582015,0.04573740496202172,
 0.09823752077779813,0.6917254326243348,0.18193297489057159,
 0.5580781995848555,0.13105879829172928,0.7214136030163515,0.7222512528358151,
 0.5854387030588285,0.9920834399992855,0.6311569544388644,0.15563054170609314,
 0.4263977142913131,0.5336813803453528,0.6517424363851627,0.6460059819924422,
 0.3353434350811102,0.4886601575964249,0.7091874883714107,0.5400388186996259,
 0.18672519342698912,0.25445925105005185,0.8550982020226737,
 0.5658551067145826,0.7614273992784029,0.9997671701769831,0.05872600318821808,
 0.49518997541232934,0.7895472379705232,0.4482005539359888,
 0.043691565778887886,0.44315894078568596,0.25947891672137025,
 0.39990319455925083,0.836822217249796,0.21840555186786048,
 0.11144891463659001,0.0484483425435045,0.6529304993612428,0.5843638793797874,
 0.30045464358448526,0.6095919598198728,0.5951779677417619,
 0.20303452268722633,0.5402329564809064,0.3350247001661306,
 0.33991151888795845,0.5030939754602437,0.7917307526383011,0.9805725782688812,
 0.0028445013814124565,0.6112575478816017,0.027448175267149266,
 0.870744815515943,0.716507964031423,0.8748902108233093,0.3723580047963332,
 0.07456049212106997,0.8993262476579348,0.8725618055126966,0.664799012114112,
 0.9881612936863883,0.36452178330422824,0.40761963244663213,
 0.6933890530244637,0.8955946397192879,0.12497325150935856,
 0.08958766686793618,0.9236671149768853,0.4834775368804569,0.7664508351775022,
 0.717179433603198,0.274896179583094,0.8416997085011355,0.4011671766555086,
 0.5537716918366627,0.58480771742946,0.20257703212871003,0.6069618685809461,
 0.028604256165781883,0.5534316285630886,0.8902923717870885,
 0.9737283555447224,0.8474946341651804,0.7843556298515251,0.44755248588761454,
 0.359088050475303,0.9379677165018792,0.2149991848300914,0.313822631078998,
 0.04316956404269512,0.2686908453643614,0.7810987489277371,0.9068370148103594,
 0.6250628030642243,0.025706707961809294,0.9562469240711878,
 0.8257769495915608,0.7686422201078056,0.044616318015066936,
 0.20839687626398273,0.9274691705895867,0.44046112605146404,
 0.21269671746571428,0.6613543689097316,0.464570774024024,0.7836802245795464,
 0.14340950857630685,0.19450593752187406,0.6925604286473508,
 0.32526793625124184,0.15077942933461275,0.5031659862973686,
 0.3561440847191967,0.06291471390060077,0.41455274865502867,0.479407284986331,
 0.47402916508767023,0.0776455853841348,0.6565696508837545,0.520164437531721,
 0.8193059708326719,0.8664400442932919,0.3856016777964608,0.08176948649618787,
 0.4244114822818479,0.2023042602278694,0.8056375906406923,0.6165187967778167,
 0.4185362243774252,0.5733883368146743,0.18241916113766077,0.7507061327022561,
 0.7081865479837325,0.4700299086296118,0.4068188927751496,0.04773396445817557,
 0.09368917274060695,0.6091017015196025,0.8211515329048045,0.5250821184060086,
 0.17322011061486398,0.20039941941189066,0.39708644522574577,
 0.3674540869876146,0.9309466187344934,0.05598412013532539,
 0.41782792129394253,0.23613262133917723,0.289382400168945,0.5928142364401957,
 0.30189961780616836,0.7299209135905724,0.828997447992747,0.17522927453272275,
 0.519436349749069,0.3225498857217761,0.9211806969569255,0.9819213126702624,
 0.117597914951176,0.26424195484181434,0.5289726526498546,0.8556100010479476,
 0.35147652688974984,0.8449015097270057,0.28211980256388125,
 0.5583905539607427,0.8319375989187323,0.3378176496636387,0.7641108722218038,
 0.5767971565893963,0.07917515929718921,0.48831456458889566,
 0.2757216920888199,0.383036989280343,0.2537881771224211,0.08016958422682186,
 0.4268897616953834,0.5429313315715681,0.24629307785203458,0.6220275962888155,
 0.6692400902140516,0.5329396170682672,0.7632248567682569,0.5618988600634293,
 0.2359967443751224,0.3093789469747238,0.32536114189501575,0.9500861301767033,
 0.7076642461932021,0.17077892475891132,0.2380240603940068,
 0.30821302425850183,0.4147578312625204,0.389665208488586,0.7106403804676832,
 0.641409087420761,0.18500904185747313,0.9466961183376569,0.38162273344061715,
 0.9803358394091756,0.4307468988478037,0.053449557721617524,
 0.43818812947277697,0.3076811709431888,0.572512904471317,0.1651720920451054,
 0.7830669267760025,0.013123344355340905,0.9741222070202131,
 0.008172719273256468,0.7618930124631108,6.254811645898251e-4,
 0.48307678883875316,0.3849554070541925,0.4448744857077167,0.8662899233339996,
 0.24336595005179995,0.8683462130560718,0.5534496478796171,
 0.08835943600343366,0.19211980809897167,0.9461474264636486,
 0.17587338422166254,0.4923871352227165,0.6120273297547504,
 0.41100639504587155,0.834102239036706,0.9959210956965716,0.525431111629162,
 0.43665453096723694,0.025687360641778123,0.3656108719966171,
 0.1292240975275768,0.25408513403257416,0.5470033738328165,0.7798639246331973,
 0.7242092996830327,0.9713489796025574,0.4990059002202901,0.4622960382913267,
 0.7774238477115542,0.755687634949966,0.05312158445858706,0.10419631978392718,
 0.06927358400391481,0.7588563387594961,0.42437128304648264,
 0.8718701847339974,0.354827179729279,0.9427988457520531,0.4361662920121415,
 0.4923509020762049,0.3753878557012291,0.6308442247770247,0.7569588929829228,
 0.3152966181595338,0.5648228810934424,0.9195488147545358,0.21869345893105985,
 0.9833834405972317,0.6674256479976096,0.5466199853312005,0.25827343549864556,
 0.05976458428424891,0.47325370630634034,0.37062416469071857,
 0.7300338344108103,0.7158043571306236,0.39320535981407834,0.8821965671266745,
 0.2272710219541556,0.43904976425536035,0.0829256712420493,0.560747197947002,
 0.2771243434147329,0.10000874675720506,0.45313376442581843,
 0.31029843358802145,0.16438838260903343,0.2492391439511361,
 0.4284381330813567,0.07937419037419857,0.5852292980971977,0.5096792152094413,
 0.29116097534831875,0.46809418635775335,0.5728147700364853,
 0.34953793800178423,0.30959991912138973,0.7375528178867498,
 0.2742786572586443,0.28748029119114304,0.9971456744006788,
 0.009503524852563228,0.8583169211389438,0.9755523410911598,
 0.9574068980314798,0.15493273240838068,0.37050778890808644,0.438103764293051,
 0.5803109196235735,0.2798629273926935,0.9919224135877991,0.19439447312939528,
 0.46639638227470004,0.3655636488403906,0.7741529187642828,0.5022768368134878,
 0.1935273667819195,0.7858161884797442,0.915111291326541,0.3577420057463485,
 0.3039408343232699,0.005952632246319589,0.656475330619899,
 0.32651455392185114,0.7001789123307331,0.3328475689145294,
 0.35255697689408894,0.17355684167260377,0.702970084196426,0.8114065179946075,
 0.3957320891412397,0.1237931341471814,0.7338670295604102,0.20292434353847522,
 0.7825780361784581,0.4591369180448013,0.7930615614992059,0.4115987243035183,
 0.6920194528787056,0.40067893532818144,0.1338670077029256,0.4150316102312649,
 0.3445026107175847,0.8012952979211329,0.4259415019251348,0.9841411273572757,
 0.014349296900527886,0.3648708279719458,0.10618521307471185,
 0.3020425566055861,0.7455833175304343,0.9654152679154796,0.45486608508717796,
 0.7836734019935504,0.7604642006039862,0.26295834378483063,0.8860742019939626,
 0.5073424151301649,0.8914650622286727,0.6832811725641421,0.22754689416702134,
 0.8277312595329658,0.3636791925205227,0.991722604459482,0.9226819377015616,
 0.23556986070962083,0.21701344284000657,0.2500857202043788,
 0.8150654049911003,0.4352376341900398,0.5642155572869452,0.39964608417052394,
 0.24504561250020718,0.6678760206797134,0.9305185137671074,0.4940271220763187,
 0.18235793648222043,0.6487081199965393,0.8663296133176659,
 0.27274259290432246,0.7960343809500392,0.598345686513463,0.11519054860392941,
 0.9568924802358247,0.8088917296307837,0.012484009950509023,
 0.0111999997017449,0.12388730828923544,0.09479624790066388,
 0.003212190743796417,0.8199182586123929,0.5761556129723707,
 0.06958308852999928,0.255842455577143,0.8682852832616141,0.5880118895102879,
 0.16585172388770897,0.06354302269020728,0.6562356183328351,
 0.3967669496521177,0.5052048117244637,0.2929727651959051,0.9938352894514315,
 0.6917879776648546,0.9235061120877466,0.638327793636974,0.2381001581519634,
 0.18721758470178984,0.1589276455258618,0.45772640845048185,
 0.7562585330099177,0.08092082819384905,0.11402036130891768,
 0.8999411071432557,0.6636077505295263,0.17264701648542136,0.8908827386134549,
 0.9352255747504745,0.5245132345981296,0.981151810920534,0.052146358610247856,
 0.2165099333037538,0.8298101497279112,0.6579550289372554,0.04507732989751434,
 0.4352265154868763,0.5001198360027772,0.5246281792249622,0.6005982830134452,
 0.6831981555588539,0.6314057815176923,0.2946321974851561,0.2015152317853175,
 0.7708522522735851,0.38146375645179664,0.2666746650850318,0.8889973916449978,
 0.1672870452170223,0.05643395709805621,0.2731589843198057,0.7994670630349149,
 0.7598990101981192,0.5056448077890663,0.7952174413031274,0.981491603392205,
 0.31024845200100115,0.06293448389165235,0.9760703966410018,
 0.26815667209632943,0.5377638675642951,0.8673081877746556,0.3262576234833161,
 0.077539335761043,0.4694559792368598,0.9024691215234396,0.34275531147346605,
 0.574247857680411,0.5290626271130914,0.8841886806690907,0.1104397329034652,
 0.20987806777327078,0.7699355811590567,0.3825111214560398,0.838303139073256,
 0.08953825109373814,0.38032825894709477,0.188947336895128,0.6147825804148422,
 0.9433660192156429,0.5567256980240214,0.763709397766971,0.7090568774543402,
 0.8576058082027165,0.7456536993732351,0.515474028810908,0.6896399695851926,
 0.3648976626364435,0.488066534990916,0.06828294091983178,0.7866574631777311,
 0.8112337245329269,0.4005658377030624,0.35008138017115065,0.3389351262538305,
 0.4382268468296471,0.9779076796367988,0.8782259220663339,0.7878262275933485,
 0.38553164765191694,0.1934655462296313,0.0833803794040846,
 0.24540155360534943,0.0030272828807309526,0.9641819720795678,
 0.47117259072756856,0.269480253041524,0.6234803574592681,0.19734951045610527,
 0.8408127619890586,0.7923584521964075,0.07918422103941392,0.5634715361043451,
 0.9426556441739462,0.20843687381186538,0.1383377826242298,
 0.060739033291899336,0.007974417612661377,0.45242799569271586,
 0.7347822659855399,0.7607794830858853,0.7925172619679914,0.620318517633571,
 0.40226363257174946,0.09861099024975362,0.19534669666728244,
 0.21865821780696804,0.018198283214454825,0.3741964345016763,
 0.47074969365671937,0.21700627982067322,0.09486171166660218,
 0.13479423586733663,0.9855137866034146,0.6502755977446657,
 0.12350695741232998,0.04927037045614291,0.05083952468036734,
 0.6029955521869484,0.5434496696502275,0.42132007045160846,0.2053623449099513,
 0.6979460012744929,0.06899927228844605,0.9384388043738338,0.1546570038529551,
 0.565351753487862,0.3324207441142738,0.744742295059466,0.5070267373193923,
 0.7965251988385464,0.7442582872343094,0.08312183043763322,0.6751199378145814,
 0.7913373619889051,0.11104177936281268,0.6602987720977171,0.2831750286357886,
 0.11128651270668977,0.7157704916679483,0.30879378588304385,
 0.07857529078144576,0.3141658219004523,0.933988848519608,0.2878319496731079,
 0.9733964513672575,0.5569672983099384,0.2963645871313405,0.1355394351472342,
 0.5521187150035229,0.31768241180807244,0.6146244971581825,
 0.18053351727420175,0.44692360024942357,0.9745109602578647,
 0.4053476471883475,0.17462887968733376,0.03616178900039857,
 0.2790759262891228,0.19030689014320523,0.26012766370693474,
 0.07184379572408695,0.993141950870209,0.007995520361382513,
 0.46990756185112537,0.7536219661881881,0.8577518396497478,0.5637413013761282,
 0.0895670573275269,0.7620344155956786,0.5570836657798964,0.18416379636485625,
 0.753952916519255,0.054793603056161655,0.10369186733581714,
 0.8302815998059047,0.45160688156885676,0.16803689936895316,
 0.03433331156512498,0.7758025837221209,0.8157143293662077,0.4518283083686635,
 0.9465596143326444,0.7812056805037529,0.9624467990124761,0.4922836193713932,
 0.2948280739971719,0.17866811171539776,0.2522416757433362,
 0.47054558139119107,0.2930828149593442,0.3833430207745768,
 0.45516036794637493,0.11150077622817267,0.2063322403836182,
 0.12837991164353524,0.8092270779827613,0.7280591210784728,0.6820947789569558,
 0.49808406113798676,0.5998877865973498,0.302294635046204,0.3971539878311049,
 0.4831923417301418,0.2639568938589789,0.031067980229299463,
 0.3561913118665624,0.9674327886811683,0.9723154758399739,0.538919808770167,
 0.2673674911931968,0.268790353715762,0.5967878056916925,0.9568885115327296,
 0.011638438931800943,0.2884667329912841,0.17485471908831052,
 0.5147225810010303,0.8423198720840607,0.464275702522841,0.30978410850044835,
 0.10810142775422826,0.6581505862320334,0.5081402279033295,
 0.23224899476296057,0.7881577815293943,0.19683814455943638,
 0.12355392134550203,0.1521924339120898,0.43525107849975075,
 0.4344051087855211,0.618189084674698,0.016675846111271708,0.7328239288909333,
 0.619317455283456,0.23781750465193796,0.22562498121428698,0.9052732549632838,
 0.19064905051704983,0.1324244329186447,0.13695039245683516,
 0.10841461075485381,0.12628821177104865,0.6627658249797082,
 0.4816291147824513,0.4456932362581705,0.4524794600797245,0.7753273883068036,
 0.10299260146201616,0.9670649488232332,0.041321264692735005,
 0.052145937601109416,0.6400385944343261,0.10101181898741152,
 0.6918070410601632,0.8501435013346669,0.898944514535561,0.6012496022521593,
 0.002403525248687033,0.1669489348865817,0.49257783746784667,
 0.34246665707751134,0.25215092880586765,0.21363428543230567,
 0.30125116801382457,0.5641218058055338,0.22633685260355785,
 0.4212581345659099,0.4851870804357975,0.21046657858580498,
 0.41120854177489685,0.2584190472539465,0.05417774441933654,0.66440252509083,
 0.8581029042028699,0.14002093316041153,0.09524077420096377,
 0.6874566358925244,0.12109465383471896,0.45785453868505355,
 0.3047803889392322,0.5533588110802496,0.7213879753858955,0.34951315218781054,
 0.47830744584437146,0.5391942071296119,0.40913895827632274,0.583478422154644,
 0.10781976785171665,0.8917728243386591,0.7039924525396305,0.7367424327701696,
 0.9827919268437252,0.9814147696908033,0.6301684833749363,0.626023185521729,
 0.5692025641824922,0.5683615300157084,0.5183680844904782,0.48277299047003064,
 0.6535813612004003,0.8007936284133486,0.6818136045124351,0.5630512864186451,
 0.13330128109500838,0.10374130758579025,0.9707228095486169,
 0.4560382534025499,0.7000078031396002,0.5297393003710429,0.46685927234007973,
 0.8218611572884226,0.004102891680857312,0.8197436107218468,
 0.10352230211668367,0.6240666445575975,0.8007473702263973,
 0.03810829779118463,0.7084299547816907,0.4523847945157825,
 0.02876495131128487,0.43966473228795167,0.868073621314569,
 0.009080111176516525,0.04964074472655533,0.33824165747888313,
 0.7140650306372807,0.5571913559156726,0.40773417288777103,
 0.16360881101729818,0.36552302759816346,0.19602951516754608,
 0.03965432912568345,0.897442546897417,0.9616356612962526,0.5693817846326779,
 0.9506411142015273,0.6645471462000757,0.850304546377638,0.16508680201495995,
 0.4298235841719166,0.08079939777039435,0.12637045848257944,
 0.7896996492773474,0.38088430660241357,0.9320278195143453,
 0.05576984219992687,0.8192516313700606,0.6594439748266742,0.4532037947747496,
 0.2691948980939781,0.6638444104115598,0.7563828092917988,0.252029498575952,
 0.7496903322512996,0.4041485737022943,0.16565149959518177,0.6001962984929587,
 0.41050099191425216,0.6704309841290778,0.1954290676254623,
 0.031012362720154507,0.9372028396148737,0.6864688687438446,
 0.49129193706548635,0.9553183296342792,0.9971890996462024,
 0.019693228053735412,0.4698061514513956,0.8100842328539953,
 0.9753312869840338,0.6034960196853626,0.48465565489949436,0.9172633524287369,
 0.7788414935239117,0.2736464561734465,0.7050438735773312,0.9622207415810853,
 0.916750196081179,0.9422916342238737,0.8527258808917673,0.5111789673699262,
 0.48557186147461495,0.4898108822078956,0.11189988455240574,
 0.9983332930091877,0.921544791229613,0.6815135433473105,0.18982901535391716,
 0.23373303485598962,0.5624312393806655,0.5586241804248094,0.7367822260688358,
 0.8227673299311201,0.5805801647146215,0.8347445860951523,0.506004701487294,
 0.2848017337803722,0.7433048096328562,0.39864249017456044,0.3655068314894562,
 0.5891477372475111,0.24193159747162862,0.3552469995743567,0.759873216347871,
 0.736197846138186,0.7367392096064329,0.2372133434967515,0.01577031319819533,
 0.7303189765960523,0.42108575625944267,0.6095218127163881,0.6639974801722518,
 0.8128823468111466,0.3703562821848292,0.7331381391041683,0.8355926365185481,
 0.8823068442263194,0.1560330787525297,0.5849040305760023,0.6976880772743222,
 0.08371795102891022,0.044154262315254544,0.08881588995918832,
 0.38658731656506606,0.4556334698686514,0.06173103636782151,
 0.16287855326751954,0.7539753448573459,0.31605303744608093,
 0.2137492643973149,0.7644694375407306,0.8389588458177875,0.7945166421314295,
 0.21924726341755685,0.38936057870955276,0.7564697115569053,
 0.3020312522400336,0.8973531007615737,0.7539431843608053,
 3.314469944792897e-4,0.5161379294727513,0.6601022114011814,
 0.4729594901466848,0.4721285533772145,0.1852362789075721,0.14795859087897845,
 0.4849342457464574,0.3345155887829162,0.6207940076424693,0.8913341438303704,
 0.49988010792621673,0.06550563785287866,0.13058017390931254,
 0.0011653414778906024,0.5199539238636837,0.8518255079668169,
 0.6850610610372974,0.5938245885739313,0.6750646889940062,0.13737425658187974,
 0.17047723061054898,0.03947124415738201,0.594667072488843,0.649528029690731,
 0.9020855402599703,0.599087118254009,0.2221495701150804,0.32187351344169923,
 0.7208755506806742,0.8713544152826234,0.5003760980170942,0.3786891559382616,
 0.40597470571471495,0.8914934826035882,0.36581167289571126,
 0.9224083609765248,0.49273261978046745,0.054815149988429845,
 0.7097849526426061,0.5108275754241557,0.28472543346384727,
 0.08889865584934042,0.042586646947034845,0.9291366272974776,
 0.33494206420660444,0.6100145936213058,0.7792246788636236,0.9388234657858571,
 0.09373085193675934,0.2525233939485677,0.49036337249819606,
 0.35889678327026175,0.2676864307148573,0.8621643351100475,
 0.49848404127308554,0.13412155226121902,0.36344953013385106,
 0.21734411894072514,0.03942055829778379,0.058863938408054706,
 0.007373458860113868,0.9390957830463905,0.38139804286535783,
 0.5926665172654738,0.005862129451524867,0.12508357423641858,
 0.5315063544140035,0.6829739503934245,0.7372302776926487,0.20881469656330653,
 0.3486420872503678,0.05850070913930572,0.7800358521102077,
 0.13801127942457136,0.23558280895877237,0.8141929814151609,
 0.22752573726312486,0.9384287514050533,0.5682027074619704,0.9132169248535571,
 0.6894690825749268,0.50471714236628,0.6965792996572121,0.9534836956769386,
 0.253721119412182,0.9355145638625677,0.28020251509508454,0.6973915840871305,
 0.5202400511340077,0.2713659152017244,0.416678910818274,0.2129402727859918,
 0.6294315881639231,0.39596953732373674,0.48462255098095064,
 0.6158084745660872,0.6467391378275616,0.3609619885628007,0.15675957068799806,
 0.3643826079041196,0.32056447151341927,0.041865395764475344,
 0.6498346523739608,0.8071514600154923,0.28932737930786845,0.7902118341242985,
 0.8900676420940826,0.1369276212713182,0.9902241383069582,0.05469572701100245,
 0.6577470481099593,0.4890049019542686,0.38565699214180493,
 0.35287958323628377,0.03097529002605548,0.9620071070323609,
 0.07971900790953512,0.8604877919266978,0.7413446819165714,0.2672927203886457,
 0.9094246587980037,0.47910486084347614,0.3481436452020603,
 0.08027679584714076,0.4626246287696161,0.21322627337314404,
 0.16182187033458273,0.6539219999305613,0.07511570128965228,
 0.5345826938077222,0.5249678832104246,0.4539361062236753,0.9313240661047414,
 0.6488440796004425,0.14915666779018188,0.4206663403467018,
 0.14133157825089115,0.5200572124201283,0.08687550384832321,
 0.3222938325531012,0.9126475693704398,0.13532914594018208,0.81428794273268,
 0.620028017807761,0.33544266334290374,0.3165560053056995,0.4353143977753783,
 0.15610031653384493,0.06295989562615278,0.3649613138386325,
 0.23233485548463417,0.39492989851219584,0.9345968880139792,
 0.6960331480875438,0.9779893256161212,0.4091398093287093,0.5272745383623801,
 0.5152317502779907,0.27292369098571245,0.35362738045236997,0.596823607992612,
 0.2293060903279911,0.3321029498756153,0.5583208197913855,0.9201065745526383,
 0.04358545222614829,0.3921193951778884,0.19945204115678905,
 0.8009820459468368,0.595480507739641,0.02356416266008765,0.4359078763405849,
 0.9776141857128133,0.2440010049842246,0.8078721680538047,0.5416350516220573,
 0.6822582541910697,0.041377883987255726,0.004094373310438826,
 0.3315170506977596,0.5832703710746218,0.043339581149352546,
 0.5006310897956299,0.5528194172138057,0.2987877761866048,0.6897697324307337,
 0.49268213988304566,0.4471954983589467,0.7941396666296445,0.5247841898796758,
 0.6969154923803034,0.7777806701825725,0.7469129744318791,0.1849159327111144,
 0.8441968045981739,0.4080866726404627,0.890965727447336,0.5476229794243569,
 0.142975879979911,0.9130663687398276,0.3987338199132524,0.04085258961765181,
 0.3398266773142379,0.5535795150979008,0.6649200675597465,0.3373535458485184,
 0.9506746697493232,0.9469865598222611,0.5912575352831784,0.2580736176966991,
 0.7440722819692009,0.43735986664628246,0.8827381450216423,
 0.11689754810506037,0.7493079671610224,0.0186109823412145,0.7520855163550745,
 0.9354067874908413,0.49059429297787294,0.6141027821116434,0.7204079125912453,
 0.010758037416529143,0.21941075099421437,0.5981618650773417,
 0.33437694520196926,0.7553203084968886,0.08174259631085135,
 0.7614203968906603,0.3295959981336032,0.5011445380032882,0.10976895018207511,
 0.5628680299125335,0.3594904556342079,0.217385023215378,0.33885011978376123,
 0.5019603253795764,0.38747232558251754,0.7892206824367033,0.6251817109071365,
 0.9640395941275921,0.5394118249638641,0.5736932968935469,0.42346193515424513,
 0.8019551055663583,0.24945995581174507,0.7266351007028773,0.580345976349056,
 0.6108874834062988,0.7359295257267077,0.4430014053146495,
 0.051286373464405455,0.1848525375134905,0.721701964409001,0.6731297003047363,
 0.08726331389631048,0.30427553775482785,0.4541475731767267,
 0.04409296136219831,0.18192182927558032,0.5923408151189946,
 0.15595248088492875,0.31297803692253845,0.3943848924408546,
 0.06598572513523004,0.246935061320815,0.27235560896366595,0.7832695728988731,
 0.9288197774071291,0.6661065650456305,0.4902105626941353,0.3059853635380454,
 0.06989692331858288,0.0012025867370899501,0.309356836616711,
 0.025007948443589112,0.681464694861096,0.41003952825127077,
 0.9738405225500213,0.0667329242615301,0.5971783691733457,0.24359616810817608,
 0.22489271505455588,0.893309212039856,0.09132764940952998,0.5660451963074608,
 0.5838419220018157,0.17599278638765825,0.6202869043546129,0.362350821653288,
 0.7333786915301357,0.868199119091363,0.7394631496016651,0.3449628895746002,
 0.4343441190886802,0.9814811610860726,0.8828971839825792,0.6693850259460912,
 0.7450294980469929,0.41972638404240525,0.03973108773979617,
 0.03561006762758134,0.3169824477053296,0.791565633207836,0.5716366378471893,
 0.9897217082775807,0.05144021480113359,0.22112269131800755,
 0.9172408818406221,0.22418955626087245,0.6388445222606385,
 0.17170725474152682,0.013114677212867631,0.928860582482781,
 0.7183405879413789,0.6245324178230207,0.22057586471308832,0.2604418775924162,
 0.5940505970091862,0.09494039693498058,0.09267187449141712,
 0.2564753732854701,0.6425932427229593,0.020260507265518535,
 0.20266266011684775,0.9985812918303081,0.16377436573619442,
 0.2325166857646721,0.02347553433356575,0.5287617333020634,0.2074879914116763,
 0.22342926512785322,0.1638860124417827,0.9661200122642939,0.7210610814527341,
 0.8315509609226115,0.6135865611522006,0.6804973673136644,0.6856229900336216,
 0.4666125837151146,0.5578672643731071,0.17380094907571297,0.769735343944801,
 0.36282630195734167,0.8440631067928309,0.434845201756089,0.09576030226329202,
 0.04360874634677603,0.861688394563279,0.4686189270151979,0.9632279039250318,
 0.19562545689753652,0.36557277275662703,0.7177588862054699,
 0.1733980508019045,0.08577683183749973,0.6169161375667103,
 0.45993200016956726,0.29159105269712715,0.5845786443352329,
 0.2875512962829472,0.24709258465237638,0.3168430032176671,0.5280052556733817,
 0.0507293341429883,0.9442740898275994,0.5786855231151159,0.5419035818584723,
 0.03365556057557906,0.8395134727827462,0.721208216788759,0.7412038145684596,
 0.7268409665584822,0.8576695531516744,0.9151468837543779,0.3717017113880534,
 0.8184075232485173,0.10274766150927084,0.3716686973372111,0.3181212521110407,
 0.10912572162079359,0.005093965363100139,0.13229554285683576,
 0.493793505617096,0.5568307898003867,0.5029357754233493,0.5346341535851367,
 0.8141024291818869,0.20609769363076802,0.2017860699070545,
 0.24549463839471986,0.050302244651428296,0.501185477188682,
 0.44797725187542237,0.04459355051560199,0.2984242588802345,
 0.9023652165957952,0.865800855044988,0.9168610084877924,0.7447206433033144,
 0.16355072425007866,0.586635627808967,0.016084230205787087,
 0.3051694826671263,0.027418053086831096,0.20731094005701556,
 0.026434045429670205,0.12032181860820268,0.7168407968635611,
 0.6669475386528755,0.9930523207566051,0.14352819913489823,0.5249040712253972,
 0.7437885235462054,0.9572066925295446,0.18617726677287738,
 0.08126140728174513,0.9863962152810973,0.23290410517006432,
 0.5075496735431091,0.7294604395593158,0.8412765006840814,0.23362591576967184,
 0.631446142973435,0.4189967269012449,0.20202843140097404,0.05585045581461878,
 0.2328481917062467,0.7772355584391915,0.8734080100230318,0.8396472609383964,
 0.724833530866898,0.8036245137578143,0.810183587905435,0.19749598434782645,
 0.41797721801115284,0.5711409973424505,0.35380711378075,0.8943482654467083,
 0.5695536330110826,0.667454614036296,0.03714014448515035,0.6739118533852421,
 0.9016398270758031,0.3416107473894371,0.7603555954110971,0.7613192521200226,
 0.13744279662895853,0.5488098869059613,0.44777200967543396,
 0.4841841767935763,0.4807044231140515,0.4072530595339655,0.5302323771480799,
 0.5672876381690835,0.9825326075209538,0.06857442614874709,0.7643993233600728,
 0.7063408268759861,0.8358285534671627,0.24002853247614597,0.2758248830238653,
 0.050618628476518546,0.4450289008500872,0.38425558150116157,
 0.1550290558488383,0.35602423243980375,0.4298722388963291,
 0.38385245843545346,0.6874922930449178,0.4329053699024338,
 0.11327153519241584,0.7188897209321462,0.354709816062178,0.16526933413301959,
 0.37962860697709633,0.9701311497910707,0.564376004659344,0.4214847877734672,
 0.6724998449388925,0.8421615934861524,0.6064486987193634,0.10073652550144896,
 0.7356003596794406,0.08746946071295036,0.5487877856924427,0.8690073162865279,
 0.3609330943376079,0.32320177526079175,0.8346111122829875,
 0.30221221308421475,0.11945483566830672,0.9682280258030547,
 0.1915883168019188,0.23109358051090823,0.5947724025540134,0.995977100709993,
 0.7647451710165871,0.7432387808706651,0.6198032081914054,0.02068316908739576,
 0.7849417519529881,0.6926276370018191,0.3841641854829295,0.9313690294698265,
 0.46651696036915646,0.6016041858496616,0.5501003907274967,0.7789083667640806,
 0.14368975736305645,0.41489940649692825,0.09595835749248938,
 0.7555249382774495,0.23545509464924297,0.05620050515890673,
 0.5716673522067921,0.37172654652371095,0.8810833678324431,0.5696031956635854,
 0.6566414842338278,0.7514981035055118,0.004708363701026452,
 0.3678466443329964,0.355810177938349,0.24515644431772965,0.9536593547928557,
 0.566171515525665,0.6124582894308297,0.5544759746586885,0.0725564820642598,
 0.44195367834258026,0.20976870165436623,0.5801334471258567,
 0.09456367065897475,0.32557791317559526,0.7773479073224389,
 0.36691623173765786,0.45857260100008035,0.008526858052878561,
 0.6404085617301638,0.15956969448052494,0.21129785966454528,
 0.03327884223856081,0.34656331926867345,0.560916496150084,0.9276914944534915,
 0.9868233282851291,0.4769469621700638,0.29452550745044437,0.7760392045096094,
 0.4349052533612944,0.35824558633872694,0.9797693727231526,
 0.03393586660703163,0.32524787468821925,0.3212003092478998,0.106989321131856,
 0.6797451337676963,0.15404639456769798,0.1060405168031332,0.6176071716646385,
 0.10327310727604444,0.5092025006691816,0.33155267778241004,
 0.7394648988668568,0.5896774593161201,0.33126447530473946,0.8308821929957314,
 0.7882808328337422,0.12361735031546672,0.7122253617187717,0.6797190259089846,
 0.39004308145459754,0.5106154175163727,0.9131101859457879,0.6574324635031844,
 0.721102534513759,0.924109895470006,0.28709881065172704,0.6583229921313949,
 0.5441707425520945,0.9734755031684867,0.048710726974995344,
 0.3626667054243803,0.08963631395881566,0.5672237040713682,
 0.0022560921311527427,0.20175512222803005,0.5130245022680725,
 0.40337150149623957,0.8192040964831047,0.4795184390511681,0.6144141690047041,
 0.16688452166362466,0.5935064445431806,0.3344622354131993,0.7724370632171854,
 0.508495048489922,0.9967700141340429,0.34140903885779283,0.2611348532165516,
 0.8251675618535391,0.11863720082496687,0.7110914584120166,0.9827321813775783,
 0.6933677771316538,0.6107149386070063,0.8640781709396701,0.0883381548627944,
 0.36678578272999374,0.19039710670047771,0.3608467693943941,
 0.39277883137259306,0.43668907037826865,0.6575257756468414,
 0.6924673703867441,0.530431887087744,0.48882753874859264,0.41032371574324045,
 0.7025899839049699,0.20667281960692485,0.4296854551101834,0.7366237489187716,
 0.9484641380910056,0.8130269981338445,0.27117215092680336,0.5514670794975551,
 0.9278222714798134,0.9885858074749708,0.30785368234631094,0.252121714100877,
 0.5990015284876158,0.976558209805536,0.09753017689379051,0.14850173846143233,
 0.1678568380799068,0.08094799158707511,0.9596639753241238,0.5603491927339638,
 0.6740739602530922,0.4208171827542878,0.01950784394849636,
 0.32230447195474554,0.2903826267545727,0.9441758368662241,0.4548406288377461,
 0.4654229262258536,0.33365196198292413,0.9196478363318241,
 0.31082849488369824,0.7071069457725139,0.12818723966527534,
 0.6773296055082553,0.8870839737408561,0.3009361746366739,0.8251273441960987,
 0.7799042039118452,0.33449722705297824,0.6466856847539763,0.9832454880009275,
 0.4368329585110352,0.6597009148427749,0.9943020903429947,0.3215492440861172,
 0.46814030995637934,0.09563010937045091,0.4878713131008082,
 0.30462266279043626,0.7758193374919025,0.4513206090391213,0.9480035100973866,
 0.3155080309757543,0.5292525159559375,0.44104662316660903,0.1397431924232988,
 0.24199370918803675,0.12185512621451977,0.8923493538488161,
 0.5793899652059229,0.6601926794359979,0.28414467297110213,0.6799818298902969,
 0.386881771761611,0.779758697223272,0.7806202490652016,0.1415326211654806,
 0.8485624909496177,0.4432577865455236,0.86305679947641,0.5118883902486272,
 0.7177134386621458,0.6335222825320335,0.14631631987622373,0.9483518352776411,
 0.3123910817443576,0.036461020858016835,0.5957695282039303,
 0.9024539387244301,0.8936335099978071,0.297456146526357,0.2050293035511278,
 0.6727519521122571,0.5871896890713035,0.15606906345694427,
 0.46930337922186927,0.8290430432350085,0.44963843034413975,
 0.07647977031510111,0.01781836874842868,0.07290150309144305,
 0.9496356773885394,0.30734501642333456,0.7462125903822482,0.3638064492319635,
 0.06481076177737133,0.004144477556065174,0.12034224356978473,
 0.8694092241021456,0.8925496074364466,0.7908526357980246,0.4194151103555813,
 0.7990193231991096,0.13880499464419926,0.7775949948823775,0.5434009485624444,
 0.9238455219979275,0.954209136120721,0.7675151763002852,0.8929063096690915,
 0.27397474888011986,0.8942285376888878,0.6831535906703924,
 0.11322113570925452,0.6596417862169466,0.5930470785863299,
 0.41915006694335266,0.2510420625856381,0.6800100860785001,
 0.10581089078407424,0.05344997847350208,0.9589624615289359,
 0.49133512143475744,0.20616283107895816,0.3689452830924298,
 0.4575976591558697,0.7482435121008658,0.8460740136958009,0.35608520946991473,
 0.5938950922243881,0.07322551452417692,0.7274679551325945,
 0.34543881100753804,0.8289984046447296,0.3073853469109099,0.4593205798269927,
 0.49656282575156596,0.19168199612343073,0.22245872323567195,
 0.661546002420592,0.7783939101065159,0.1680713970764085,0.7237614498743309,
 0.025136956697663138,0.49317996421511867,0.22203584092991857,
 0.04872344620002056,0.8290914793229777,0.4741697875348133,0.8460739632773291,
 0.917124123715687,0.3817139924618894,0.980192927388028,0.917089765615255,
 0.52255401661238,0.5668306446954284,0.8688931792140027,0.9374022819028255,
 0.48686972884007196,0.020503930650758395,0.34951297076111,0.237216056033137,
 0.20209941204502124,0.8936808339726182,0.10136302651603246,
 0.5739384404702468,0.9233485757991445,0.0393916765265766,0.01903398210480367,
 0.7941467768674708,0.15089008479516208,0.7763936376502372,
 0.19512991531813872,0.6309388759700779,0.9686000162216295,0.5395955693814218,
 0.1963320195293322,0.9726602982494691,0.2862595538214471,0.5274498532425956,
 0.746981268569282,0.3439715955153332,0.464946208208061,0.4460758230592713,
 0.5935431820712893,0.37379046792964776,0.529231486983999,0.6546426579147351,
 0.9543733058587539,0.7963620825006309,0.37983515270440227,0.5330221991670727,
 0.3557941075066342,0.8081624969331079,0.3086056417318297,0.7036982484942074,
 0.7523538552181223,0.6309417612382417,0.7383604776741449,0.9949797552573224,
 0.3280876022944127,0.9577260611657653,0.3361658532110716,0.8943195922400451,
 0.4969187007677509,0.11626303526301374,0.7136404317112885,0.597039077693024,
 0.1118200699730798,0.7632351628898045,0.7628337656391598,0.6526330935383714,
 0.4397720040357772,0.1326867280188303,0.8033767345174767,0.8109771060994377,
 0.19779250667663084,0.10038922500150693,0.574297125822671,0.563418418130393,
 0.9896110592744506,0.806504291515245,0.024191169535861867,
 0.07078801742719443,0.36941044157147584,0.6856018337033601,
 0.7071603771876234,0.45442513983031163,0.718210335638736,0.14942480566826444,
 0.20834150425215814,0.6616779481269283,0.26841176548569157,
 0.8670047349385002,0.6136380513368407,0.9887427251517025,0.9043577219722142,
 0.05821489283806269,0.4296098236769418,0.7034437911108029,
 0.22770610298175198,0.5637631539454714,0.126885690699288,0.4749190570400764,
 0.8589621370019348,0.7885512976971916,0.5471449222042979,0.4564950630885444,
 0.9988446115386951,0.2783008032427172,0.11664087013120683,
 0.008514382988798808,0.2503223684321546,0.009389449104618586,
 0.6330072465600862,0.706575069594836,0.5194165218440872,0.01046035104305476,
 0.08085455018852394,0.18327340653856616,0.19333056415597716,
 0.9982012475131583,0.008444503970825079,0.07915460489248438,
 0.9085516895468573,0.36155077822234616,0.5992366640899531,
 0.41826068916389203,0.7797796542990433,0.4028579130129911,0.2225168215237524,
 0.37138877474254806,0.3567739234089623,0.2661533252051539,0.9724454044748692,
 0.5415336666345392,0.560324868305726,0.3351037746129174,0.266782050234434,
 0.05818962371018632,0.7731755795862045,0.13787703480539482,
 0.20271759158468716,0.5566805243148749,0.21708594099521072,
 0.08186738552017947,0.8788609361040737,0.05247946628217026,
 0.18988983792530023,0.2918831625271845,0.738924519054672,
 0.005804635481675979,0.14136669499473653,0.5868528513084447,
 0.5394615920960955,0.8300109269942857,0.988799250541902,0.2529086121288966,
 0.42523954445150025,0.30186160617532165,0.1854000485848437,
 0.8760199307696651,0.4405304085753369,0.7413863353589252,
 0.034165132097140116,0.7193556147505339,0.13185788632141127,
 0.43964611337836035,0.9719560839189711,0.24640534779637524,0.611480467408477,
 0.1497885974512635,0.8783748707129017,0.3350886919838061,0.10978829639594956,
 0.5519214181087257,0.7146804569825774,0.13814177110898873,
 0.48242631674251324,0.7336475092251682,0.044320805771673855,
 0.22516190823167015,0.5096095369856901,0.6792412308009429,0.1384721084509446,
 0.39427371223671737,0.873035726969307,0.5653916974034222,0.0949527703013886,
 0.9010142882368279,0.2907843914363478,0.5881855708476019,0.777437490443687,
 0.9338948117470962,0.44544176439274963,0.06452817274659417,
 0.8194236368100932,0.08992025174817453,0.6565178484167914,0.2364980217186744,
 0.7491334299555004,0.004170721784712672,0.1497561114718331,0.739240979607525,
 0.7555618670910715,0.7833942173918064,0.658761209609452,0.9336073301934789,
 0.445033382123583,0.15823831575593073,0.530075839667937,0.9506418868444112,
 0.6448913821467889,0.911662387239576,0.6185434005478846,0.7972263784410357,
 0.045089569348063074,0.7890986149721345,0.3173253680233471,
 0.3144716591460247,0.7326433493104209,0.03522769127461467,0.6640764602663514,
 0.4450490190382119,0.11024917890372787,0.5097697067493921,
 0.03301885494701917,0.406362512541816,0.5547598884722343,0.6192886224642065,
 0.42623790575662857,0.47841238878322156,0.16837349351520892,
 0.17490225475255583,0.15396279567361626,0.8430780717294942,
 0.5649351353069744,0.6030815019696408,0.3186084529737774,0.09725339728341587,
 0.9971960918990599,0.6228617704196889,0.9976244299625325,0.3722820112234305,
 0.6367452037437595,0.6055048015601473,0.010646133787592715,
 0.06805722908353906,0.8328476427827722,0.7579483451430621,0.4004645616539122,
 0.2774419917103559,0.7976938739969675,0.2082247077498598,0.9243401585070933,
 0.7205550649670434,0.6369593025327507,0.10442605885225897,0.9983505400424403,
 0.7290153035255263,0.07220900058839053,0.3546351655863247,
 0.03537816096449953,0.00787806378641398,0.5731884307789694,
 0.9607504905124117,0.6416255309263029,0.7083324789290787,0.06100053857067389,
 0.15969913107188338,0.4134478844452941,0.5157505574027198,0.8346030952828853,
 0.6527686139008541,0.41749844050881346,0.6868351207182176,0.5121504753361983,
 0.02678970828773164,0.5430252440616931,0.5204915060746743,0.9909524595803667,
 0.7375683202797791,0.14162500651332754,0.616159835764684,0.1707611994574496,
 0.11098451271921272,0.03631531227388718,0.32668830931033455,
 0.8684484309992077,0.7363216689018324,0.48597603786368126,0.639485311283213,
 0.5267089146291974,0.07250582894033841,0.04374818685272264,
 0.2772899328790248,0.21850250250904668,0.6688535948788072,
 0.28470219398092833,0.5543601452170795,0.4154737414309553,0.8686589616891867,
 0.038446657726709854,0.5117241621399817,0.25530773977630705,0.86525682530551,
 0.5291293892985731,0.45578689933366334,0.7842351567137194,0.2803169154062929,
 0.35908318244588,0.9133809527438759,0.7479336720418716,0.4122943950128749,
 0.3545697731880407,0.39852531482872366,0.4779815946016739,0.6828388188243689,
 0.8418969694582972,0.7446418105000296,0.4185158562614427,0.7661044239732892,
 0.1951383197472759,0.22987955507779045,0.45503348545337907,0.477314386926286,
 0.9642419915778058,0.6087379788235904,0.11516803545244148,0.5295423839398697,
 0.15627592396109335,0.1337149384736389,0.582428346494127,0.43694891926664714,
 0.5105818849012405,0.709191350114042,0.020164426818566006,
 0.41556298451773155,0.5704965468266806,0.6911888452353171,0.268571251329516,
 0.8146246436010247,0.04422593345825909,0.11470451965538475,
 0.16888841668589283,0.8252537760076961,0.871949615101852,0.9378555296292949,
 0.9948904700257387,0.34093089587095726,0.4287662300835351,0.987765073387121,
 0.917500752526312,0.7869725078787652,0.7064022963879718,0.5396592566957443,
 0.4681822499902685,0.9894814061195027,0.588071448270377,0.13935281805115962,
 0.6953358422607334,0.5575457102388255,0.41071872460891656,
 0.39440892701407204,0.4781733236731309,0.097174248660168,0.09350101065688687,
 0.24794179230225355,0.6521861610184619,0.7467151472549264,0.6317668200181851,
 0.7473349595318077,0.04718555305914729,0.8060052320475248,0.3359041324959954,
 0.2782176956040142,0.6000841599290274,0.0776599647070384,0.8912905979800092,
 0.6987906248174531,0.7702234796106914,0.33541773894488236,0.6452195523299943,
 0.9377328549891115,0.17090405803575437,0.8780759250781615,
 0.20917727109094408,0.643579780593061,0.23847693777033285,0.4177485915714281,
 0.6434569575164801,0.17429998132564128,0.5934386745977737,0.5441499532189926,
 0.7383064281862591,0.8802283898619403,0.8636432063515965,0.5224529196408766,
 0.8271243345806489,0.9122184371741584,0.4545018366475594,0.12111315907944975,
 0.005803857874121854,0.22260220506363637,0.9335423672121171,
 0.8204179644296575,0.842839754984799,0.8916865481443849,0.5721467019165103,
 0.14107521737686657,0.20667061139044174,0.9146824894019329,
 0.9081594289069378,0.055358480424941736,0.0038249535833470016,
 0.03898873983342266,0.44455478695863526,0.11622382934412745,
 0.7169634348109429,0.7577023790028176,0.7323807190273615,0.6583198842746381,
 0.42447678606605577,0.14724360178353413,0.6385443799346009,
 0.01863417381715371,0.5546087989542129,0.7863656442350878,0.7968947641912568,
 0.35428330102907835,0.030131340339215273,0.9700385978557278,
 0.37179529966036795,0.3744133138036919,0.4513608774750888,0.8265187811363873,
 0.5497568416592598,0.5795613945317967,0.35152793165359186,0.7427081591224542,
 0.08625201446805042,0.5137149421836862,0.36668059196462455,
 0.27751345091599944,0.02339633646627448,0.7984596787347011,0.244472506480795,
 0.0753951714489226,0.243721363927638,0.749954334056116,0.984379523684796,
 0.20601600988539248,0.9420125015541736,0.4629082761438642,0.7475350088340131,
 0.32793050456279804,0.19004343780570188,0.5272039323046251,
 0.9120595353143214,0.8371890923306169,0.4708032240769452,0.13977357301954663,
 0.8956810778187492,0.31426458118586054,0.04519129412147893,
 0.5082533195380357,0.5216044652536888,0.07844115782318184,0.5856502567832538,
 0.11721226199963364,0.17554925907950958,0.19526790311795106,
 0.27098363308884565,0.8018121579518727,0.08972618632294649,
 0.5556190194411615,0.6938053228426478,0.3339403885051506,0.21840588069438538,
 0.6568435509847363,0.3794980317500456,0.6586878662157082,0.8759839138510904,
 0.5655096329876743,0.5387628489270546,0.2156148414950495,0.2617241460886742,
 0.4268160444828668,0.6808034616655299,0.6620079980847542,0.6708454693956046,
 0.5616695744856625,0.6921547843813615,0.45424185304504305,0.4290184089198258,
 0.17551469153491195,0.4252940940949982,0.4494932406166581,0.792345644854502,
 0.8373213265918648,0.7231666776186142,0.42489922999576846,
 0.24265074323348734,0.27904287630293945,0.1658534550850148,
 0.5452707231635735,0.6274942999209897,0.6485927335804025,0.8600564905264644,
 0.8888979427287281,0.7837212250531191,0.7151206175239688,0.10721085064946034,
 0.5656261064363574,0.2129186557772541,0.5450913990057971,0.20359482615124969,
 0.1539539739506267,0.3367298548460891,0.8568835071776086,0.11815058651302723,
 0.1644308620782473,0.9353745643395758,0.04235852157128661,
 0.16139228741039902,0.27100685418696147,0.8086685500429841,
 0.22075190505862752,0.6002044397266495,0.19422578189608086,
 0.3784480865528135,0.07921971974401831,0.2981271239461959,0.4348774293881683,
 0.7379156654897517,0.9972449723161543,0.9314724257443391,0.23478812677648064,
 0.3330948621722488,0.11438875544880522,0.3323583234427606,0.609434479743603,
 0.29980376126831854,0.16689821591911125,0.7057794292839175,
 0.12363501002528443,0.582357625321031,0.266482150115373,0.018752427483918033,
 0.8187428559938145,0.9281565057417531,0.352957914875922,0.8466253123313622,
 0.7995863654002865,0.11575008391892994,0.7565629682880065,0.221528028636943,
 0.8051416555673295,0.5199743504084464,0.9660231323595246,0.9872161808679248,
 0.28785770405901645,0.9690132872682982,0.8947020253698258,
 0.29718443457687904,0.9003858282376883,0.730751598396997,0.2825835745851738,
 0.9804935592743621,0.7726238008823328,0.16988877457323448,0.4722122574473713,
 0.5512407700330602,0.9023498737607895,0.07575040440541825,
 0.04839647950314774,0.5775796306204687,0.3431673373061478,
 0.45503612506910385,0.7224213112799225,0.08239849266472854,
 0.1634730425683404,0.9234133412662732,0.689485780591259,0.4621305054140399,
 0.22099133202761678,0.8680841006705708,0.9690956506984929,0.2592199866332202,
 0.2912567628383367,0.5911813836856623,0.5623156731304854,0.6066483695376783,
 0.833863758329729,0.396723784905618,0.0967117533580526,0.6683194118473376,
 0.8157969846453322,0.47763931426882056,0.8175299839561381,
 0.46005285757707925,0.20473718131783314,0.8536894420112068,
 0.8092568328915714,0.11186145146332427,0.30707326127653434,
 0.10000853364117257,0.011853574821489299,0.03286798690689585,
 0.6102472347375412,0.08943784923430087,0.20748846584604952,
 0.6600846272835896,0.868603823590122,0.07459382618144617,0.9559307460521203,
 0.33574529343031156,0.7688311011745365,0.21155877099278708,
 0.4497119800892997,0.9240754847573538,0.332801186356142,0.6303346527303114,
 0.15280524587736688,0.08778528835208732,0.2840832049877785,
 0.3159260434437291,0.07725231465013582,0.5707315739182669,0.8568422358445308,
 0.5069250615301714,0.7294423272221423,0.9415381264064759,0.3956821840143191,
 0.9485642626723868,0.5818023415346241,0.4221576146532239,0.6586833765499905,
 0.8715597817043202,0.04046523002608793,0.9797621486794347,0.3242868957883187,
 0.3768607347607045,0.4364651339414394,0.8348665317482491,0.42456739024128176,
 0.9007711387070823,0.9728440268650309,0.1408055371252862,0.7473218158356528,
 0.8079033575818605,0.6861637164841256,0.7619343977551503,0.07949317724105054,
 0.02132000614330254,0.35478932007955555,0.3921341901688413,
 0.9570074038473642,0.1653312623950467,0.3467288121484927,0.8866699633777166,
 0.5774304074174793,0.542531200389498,0.8750458627405062,0.6327520336892661,
 0.703704884536191,0.5913545667375792,0.9576536121615351,0.06800823294130143,
 0.9794647130031523,0.30060038902954167,0.9904869314427449,
 0.09679995683919929,0.17635673470051194,0.35400425035231864,
 0.8347219367945452,0.961032175478613,0.5938596738575801,0.3391676734042447,
 0.91500223154964,0.8008047638027479,0.8414271664670522,0.3137781949541494,
 0.0942863742031892,0.31926757095561276,0.5520201355868882,0.7776965964222712,
 0.12797396913427794,0.6842905124979526,0.49736888684986613,
 0.26684849976238356,0.20102706760762257,0.9963829292140459,
 0.23341255079187395,0.05562412377497361,0.02031869130114272,
 0.641602526806238,0.6870019102512233,0.574514472123094,0.6449745433689851,
 0.6235692032034179,0.4482347226116681,0.6236671231905105,0.31283694288187036,
 0.9453880377489305,0.2635248876563122,0.9337323376670572,0.15330867100409995,
 0.40817078590430067,0.39155435938261995,0.8139438045238605,
 0.4085419700303311,0.5940636981120163,0.3811512627263025,0.2177242259034763,
 0.5356624972041647,0.46025956609837504,0.4898657402387514,0.4498880204277592,
 0.23795739545223216,0.046418200782369246,0.2778432850848862,
 0.9437881179637586,0.045502121180395094,0.6560347202699525,
 0.6536754349855209,0.6610591729938855,0.7768436504578102,0.06563782373625193,
 0.6627008202275085,0.8967950807753069,0.9114858133261478,0.44452723609591716,
 0.8669810964029906,0.17996280307038792,0.6666865234771545,0.8149795766902592,
 0.1690897340171117,0.6928818224056094,0.045170823765630974,
 0.7333692192005598,0.7488847209100322,0.5001370795629871,0.23305477852315604,
 0.19504944993044893,0.3816875553265956,0.8150216282336719,
 0.48746311447999524,0.6289976979644121,0.1666836836990866,0.8419754067240943,
 0.7996173890271363,0.9819631382489005,0.9354670997616746,0.6074904600034835,
 0.030403919249992084,0.2272823371658279,0.6321097064562551,
 0.05882159527367348,0.7965652361257457,0.5442376090594524,
 0.020264398454000854,0.6643654716710513,0.7965558858895381,
 0.15870218114572976,0.08260677896600899,0.17543930506796623,
 0.5496545847915344,0.13499519677421334,0.7012818832191915,
 0.03640578874342493,0.9033246085801943,0.3042546451353485,
 0.21148354864650565,0.42152792517060034,0.08649037378984215,
 0.22941662719036593,0.05572140663254033,0.697029546608565,0.8874456896936015,
 0.9606277245342565,0.8484333660732836,0.8668668296455028,
 0.014173894542623078,0.44781784290775595,0.6314182700865508,
 0.6778116300238837,0.747405382127313,0.060852456841361535,
 0.27120673406564244,0.9970681993985112,0.36619738367960186,
 0.40041368447397163,0.754586975495213,0.38128625383185966,0.6528869006292675,
 0.3789266156624578,0.6860587662081284,0.5475359133697885,0.49140092122597134,
 0.4401033957560512,0.9765372631370326,0.4662813250084722,0.5945248507610456,
 0.3781392485750328,0.19668679667407218,0.5436064611371032,0.5702665218994696,
 0.3100121788242153,0.2740512509533547,0.6665938505757993,0.7507990872368248,
 0.5975717388650412,0.22737978447780327,0.5224575202850074,0.5533281217316139,
 0.23964842380926055,0.8358461767588664,0.280150122145688,0.8828012582062651,
 0.764419074257596,0.6768741225766685,0.3827000231205673,0.25062077767209523,
 0.7854034517474615,0.9433959885997506,0.37863569141248477,
 0.11987710562836651,0.9488553714971064,0.08911113630862655,
 0.20451470265326854,0.535582292028889,0.297669707409014,0.46042662676961477,
 0.15563769762134272,0.12612459365509476,0.7634359673915534,
 0.5429168387237138,0.34867908762912536,0.17712724720505335,
 0.7320899393876092,0.11480735443279322,0.7951776240143846,0.9509398020943421,
 0.28075227879394316,0.47358450601512303,0.09845898319210133,
 0.3734851056733599,0.9275282455443152,0.42442875273572855,
 0.05874243965277115,0.3984495937745949,0.7775481982460958,
 0.08516961818022661,0.013029158281270381,0.6332369709007557,
 0.6057685450577941,0.3112304481882604,0.506268988667519,0.21596000526946935,
 0.5384578852573578,0.7341939181883116,0.5772971159712931,0.7092836105861154,
 0.8836321532304471,0.5193726755064777,0.9907749447476778,0.00802739700804167,
 0.8865696736489166,0.02539697542972097,0.8504919299652529,0.9471143032739175,
 0.6753128608021681,0.02055004452812359,0.6427735098601473,
 0.21695623299597777,0.8472292756766697,0.9807366631801501,
 0.16390491537851604,0.10182914371749918,0.7095170559963239,
 0.9832300008421933,0.5632462000980525,0.3057457588022283,0.43869807930228544,
 0.35846067671422144,0.11359844837588373,0.8996483684916909,
 0.5862081747303614,0.14872744500359092,0.03241560813574251,
 0.6069753680643587,0.22074932595532637,0.050165489857508794,
 0.7599324818204026,0.8020449640498235,0.44827704206323116,0.563468967190043,
 0.9542606125013007,0.386337492370057,0.296479068522264,0.9362056447038319,
 0.10763852520025319,0.18654759757538497,0.8766979145516297,
 0.7834045962265197,0.9662972181522747,0.27008997999734374,0.5332221307307234,
 0.37833509834439427,0.1807488453203615,0.5956493728689389,0.5775020264114095,
 0.657094156951238,0.5211708839164518,0.8046029737684863,0.19175917713002277,
 0.31874256143084356,0.9344710616302696,0.39692861627114495,
 0.2231915477116141,0.09077230216811039,0.9945030020656012,0.6672980402081722,
 0.09908408012278269,0.4964922067865126,0.44709881791208583,
 0.8342346778733788,0.20959842237984938,0.4510186670741242,0.6322198457855861,
 0.6987475853198526,0.07473092105118984,0.4580332133865994,0.962743284393693,
 0.7937639190928379,0.18159834069351777,0.8152312674110402,
 0.15071450128686426,0.44671749700217833,0.43347838158756624,
 0.919721723629414,0.20279617988120158,0.6953325613332185,0.6525990428125986,
 0.45415969786430455,0.6913426272331649,0.6433011307603583,
 0.09775248456820274,0.22521414452145394,0.261215667911316,
 0.20249687654842874,0.11212412502040547,0.7925854779504747,
 0.9771312501165004,0.07620083960733204,0.46563967843214127,
 0.3718054937326737,0.2637353833246876,0.5291212352203958,0.5966109720005791,
 0.38182103307314996,0.982342533751966,0.5210737512219306,0.3519903440948956,
 0.7394770770620931,0.873846907343757,0.967527205426584,0.4557389217796177,
 0.9111948460573203,0.4697514261722814,0.04141531905574336,
 0.40933092930096027,0.7543570976075883,0.9317340016616794,
 0.15843065353290997,0.05304017438815334,0.47028197202384525,
 0.2612008838910608,0.34643842699778704,0.14381894587275668,
 0.06341330442999171,0.24494788268055934,0.0029216409627064177,
 0.47150940095747784,0.15763379687977608,0.18585966217010885,
 0.10535356288837916,0.9956197421403683,0.9631964208463041,0.6273181557090842,
 0.5461218625603235,0.4394258082090893,0.4001917722900813,0.4416300799650513,
 0.10424639051233664,0.09061000765395688,0.4379414245284551,
 0.7369359940390444,0.6986699240434548,0.32782563461686853,0.9837176675270777,
 0.2134858473929231,0.4834533847244703,0.19383039915884537,0.804512293072428,
 0.48941786166558854,0.45152773171381977,0.6812784282780446,
 0.2816690430625919,0.8607619307841943,0.3040919924352228,0.7795293274362888,
 0.6430154063734471,0.446551888635639,0.29054458207750467,0.11847652017681332,
 0.20928445289794173,0.9977365812769272,0.721182912622356,0.3441781476542958,
 0.32812659068531813,0.6447748172179965,0.5785314082948884,0.5985746505038861,
 0.09582931157981589,0.23763005271195492,0.0515383159455306,
 0.23217901052364498,0.05198841719402392,0.19377938074599288,
 0.9011463651387933,0.21102922053573503,0.03463268814672349,
 0.6521990617453073,0.9234702682516323,0.7857579243589419,0.9756224784156156,
 0.703709873352279,0.6056968085636146,0.6846841776074316,0.30969875399620084,
 0.6286310564425638,0.7323726206705132,0.9963986958358173,0.16633994464033308,
 0.6017676686984111,0.25604774899792493,0.09163120046272943,
 0.29744646552595133,0.304191739187086,0.6466078789862314,0.8243279539153626,
 0.5999659057488518,0.42346451851571587,0.2021185409335804,
 0.17285866136150752,0.01789630206936088,0.2781381329839947,
 0.32474877217981857,0.5785330430002499,0.22987544015328443,
 0.9452833533277227,0.5315542664765269,0.069710288827457,0.08121184094156386,
 0.6314185857223609,0.44640960915571226,0.3094123500353857,0.8565473053420339,
 0.5770800334928475,0.855805890014911,0.7461037306135193,0.42776485536457287,
 0.40306153831984526,0.9935740414594681,0.7440932732600051,0.7938827872510383,
 0.931371423878073,0.6732840165306608,0.3924898704967946,0.42688644833284306,
 0.7150329283568191,0.8544380731121866,0.26319895947439476,0.5415551749388102,
 0.04021488114602678,0.6659331589947757,0.9940562936707613,
 0.48199958774066076,0.21240727867406095,0.4154542226784199,
 0.13810495863306838,0.9354848162971614,0.7388094579653864,0.7268171442671728,
 0.48564103507669176,0.8330663564099418,0.17232384286357072,
 0.8871730970104192,0.4491053591905735,0.49128644057157256,0.6575492735464145,
 0.7514589859873976,0.12004204047769285,0.7474658203176423,0.7264718840705979,
 0.31857492756472106,0.9691324033053017,0.8374356394441038,0.3892157114531333,
 0.8743223739642254,0.2553720837287443,0.1891302287429648,0.5489739406919701,
 0.04173740888307609,0.38943899073680643,0.8163904375300355,
 0.5891666565346036,0.38313563095642045,0.7050957928309614,0.9886704391326357,
 0.18409207083387702,0.0176578544130942,0.83606622991961,0.29514842499707716,
 0.02024004199153029,0.973319772950151,0.727393471566788,0.8812623635231294,
 0.6816943872717779,0.5135519939539944,0.3543120099360725,0.7839792085393393,
 0.4579656777420026,0.1427900276346281,0.7422747385862918,
 0.026376229698395925,0.7630033950433833,0.26303588363835695,
 0.6743419510614909,0.9617784115756413,0.607655134174292,0.8887292892817444,
 0.03730286724496479,0.34228717315722934,0.01260486008759465,
 0.8776755664726439,0.517307690976263,0.29090185294879767,0.31495141390957426,
 0.12339964395139913,0.9136232138770961,0.8847613378655351,
 0.45382166013601766,0.2094683578573936,0.8826431844749584,0.9067067457600634,
 0.13789018474852066,0.028408623870906924,0.5180959588338314,
 0.6558921952776341,0.07489149100114845,0.243855392720574,0.968997419745137,
 0.7969449507078155,0.845510067383649,0.31675906419276334,0.9067729556091232,
 0.8028017517821349,0.9186492894928456,0.08770339752639555,0.3161301089781027,
 0.4837424856946809,0.4008162423000766,0.2611191734171585,0.6461859512893614,
 0.13541760761382515,0.7400228038811092,0.7199839418611155,
 0.44754241436294306,0.6437578343437694,0.801324517582304,0.605398130310084,
 0.12290168883442476,0.1474331115346057,0.4579192164780065,0.499424947441788,
 0.7144256341259254,0.8389290364065607,0.46177143621660677,0.9050373746777275,
 0.15919354888646964,0.5849713035762628,0.2979991391038327,0.8029625357045718,
 0.33782164471465703,0.7863716164500366,0.584771404167039,
 0.053045967614695044,0.18273621472877177,0.24864549787164458,
 0.986509581762147,0.7311962594597592,0.7667426623996257,0.08005619872654146,
 0.21742768329117712,0.2731815622208601,0.5360484443585826,0.3904834144413164,
 0.5611521782007066,0.34888577421612854,0.010814724486494787,
 0.011466710745353215,0.7863807680731452,0.6988223572939604,
 0.7213046341571517,0.4575775954041844,0.9857991973531564,0.18237180369881645,
 0.004290993593979997,0.6830196714957837,0.34055799153235,0.1249573477309166,
 0.6760540217287675,0.2818047034704826,0.2939609621836542,0.33144743597217374,
 0.556395740261649,0.8106105497958113,0.4619112753341542,0.016188769957887317,
 0.050479664855854645,0.2315175911703259,0.67257330950514,0.9466494999059853,
 0.10427651564664042,0.3428557111613266,0.6803829359480364,0.9730716746489982,
 0.6109616434728145,0.46171620299301774,0.7139395296329746,0.1388281664728147,
 0.9536309708157128,0.8202695282439564,0.48731341403517936,0.9407813817228218,
 0.9051904243983536,0.5858841193602041,0.7162232534634546,0.8640267334296156,
 0.733211983066222,0.6574263090073272,0.621786063227655,0.71537362811084,
 0.9596324395432627,0.19651591282637582,0.27632943093790785,
 0.7005720337856687,0.3198914804119459,0.007974671308087822,
 0.7481144439935354,0.4623673750638603,0.4144044225675423,0.5104686214610539,
 0.9436856878151587,0.5707835862357696,0.24284420818689822,0.0884539023784724,
 0.6567976777132106,0.5924079790476025,0.019884908445627625,
 0.7916962917531186,0.6960428460546025,0.9323573721468317,0.1985323877670987,
 0.6266701987430975,0.19441926646090546,0.7187551903915161,0.8465947198791053,
 0.6380722988494808,0.7892345647170907,0.3584855109964531,0.41248657475140005,
 0.47615557138293185,0.41861402735262543,0.2835891955832839,
 0.40639246088943093,0.8154664367167999,0.9313342767224184,0.9238111222292985,
 0.13287629250245292,0.2769932817586711,0.9892143072734955,
 0.32012605524454374,0.13316080002848163,0.3638277300915165,
 0.25252437970831954,0.08674952170793071,0.06428660391921015,
 0.48014260204668047,0.05821825412908144,0.9538266259222726,
 0.05612160251699838,0.7975921218686306,0.7219379376216033,0.9376849617344882,
 0.6257014776809825,0.0034063021529939785,0.6251973762630626,
 0.7063278246052672,0.4776344785878619,0.12065070931846111,0.62326071011031,
 0.46600647457837974,0.19313681871530042,0.6127894445914144,
 0.8944636466083671,0.9381676722840122,0.9949599479782195,0.5870228122555651,
 0.04338911807370738,0.36957809496115535,0.5213301119982194,
 0.7098028423616071,0.6167176595345829,0.47164293982249506,0.6109797724669633,
 0.373488133245935,0.2691299779809819,0.028906368351487632,0.5992978150098365,
 0.7516140348354621,0.4909135373868536,0.8016350018839588,0.30902073441595224,
 0.8366448211800717,0.2340709010441857,0.2911919070405049,0.49369763313476933,
 0.5808000946723623,0.435200905720758,0.5685912626359316,0.3921027563723042,
 0.3901784855026502,0.2539317077043157,0.3570818393070754,0.572873002912186,
 0.9477583916191455,0.3459268404168656,0.6883557504583759,0.6269387204349102,
 0.36767620256303046,0.9864418160460431,0.16753192771439807,
 0.2149908707515864,0.13405229773491156,0.39858168268698335,
 0.17311805002031555,0.9012632958136272,0.29224453765598535,
 0.16747544921986623,0.6411201074772315,0.9745878367865584,
 0.38055319256779274,0.12374356680412335,0.7067707948806078,
 0.08853479020674648,0.8493395863198716,0.7771520937236361,
 0.47970029536435177,0.6701024152993849,0.5559011268579411,
 0.12034963654304709,0.5634750994580668,0.32979178559229283,
 0.3088926521359979,0.5169727034467579,0.007899862318938466,
 0.7348725924571639,0.9527886584400383,0.6936963664392107,0.17667956960383668,
 0.7943202496822911,0.23881634037162647,0.6323925231709234,0.3576228380476223,
 0.10277325517301794,0.6085188523296118,0.1099551809204864,0.6626672533556847,
 0.5459290174496256,0.08907941421940901,0.6527933698725348,0.5448616072438052,
 0.5698781530259752,0.0162777117698939,0.964405540394898,0.20663418362483887,
 0.7190429837985642,0.24760516123074383,0.27460627373993973,
 0.04855315712163988,0.509786038930431,0.34667485158717837,
 0.04889413505271001,0.23364511934259224,0.5812189166761462,0.596251795471616,
 0.6171054932092138,0.45082582261987114,0.4642073060376004,0.9436193990641689,
 0.9606969182838869,0.040102029204802325,0.9192580107847962,
 0.8796938272245685,0.820377844614,0.3067888051308981,0.3433427073062927,
 0.5750926056343777,0.7850919343802499,0.720960578833244,0.2871721038642223,
 0.9879892709632204,0.885452809399894,0.9122613791825535,0.7352582128606924,
 0.6810988497207388,0.007894621671289626,0.7868425503808799,
 0.5358969406155714,0.23584546247486815,0.07554615864343583,
 0.2624582232851447,0.3932754262342384,0.07788666862362337,0.7397825482352421,
 0.1540928787069089,0.4421551774935506,0.05435718436878656,
 0.17853711682339024,0.330421942127442,0.4690215310854903,0.03265613570345971,
 0.44787713086487513,0.2745951404162361,0.2801700325219728,0.8813263966769632,
 0.7882467595455689,0.7326258130304536,0.48119798956511106,
 0.12069299169006298,0.2830507539800091,0.5741508872018516,0.6770146864682366,
 0.6592459403296116,0.04078841395627697,0.8356186561032511,0.6682921814054678,
 0.9594863686127457,0.20390437231192138,0.00652394186126648,
 0.4163560571331564,0.3195030987053594,0.7642988180935937,0.11538092770098696,
 0.2289352943236035,0.36005332594536277,0.2580601128506065,0.9113817516254826,
 0.32582854685292395,0.6295613711460292,0.4257488513411275,
 0.48531854165400823,0.5527769157102009,0.8945944954200308,
 0.00764870227701453,0.3148925146490048,0.8887413741341806,
 0.062398169445293084,0.0725836675473921,0.022753066742753236,
 0.980489176906031,0.67450842415893,0.6381295605269371,0.5645997843976039,
 0.7476723636638739,0.6355101327635523,0.13328554253263059,0.1148165215692396,
 0.593656433287606,0.21368645155331512,0.5755784841494755,0.03902302489522025,
 0.02370325322425748,0.5252585017560754,0.7731367797081573,0.7596372878047482,
 0.5078105352755433,0.4248465916421814,0.18209232866403857,0.1613136705825733,
 0.9119227472466936,0.8836052692840335,0.12139678959677869,
 0.25415792271654436,0.36907425158510665,0.9286878418310437,
 0.9870015227074456,0.8403245285493334,0.32381178015244483,0.9569212523622335,
 0.09918261309547183,0.477739988613304,0.6340877987629363,0.22876027482363104,
 0.12781997402130252,0.7697901200474814,0.7348808966625693,
 0.01930532098394777,0.872756465158752,0.4822874468189706,0.15253465725507787,
 0.08895856309813854,0.4852525939984257,0.05393568696469331,0.593910650638341,
 0.9335201470424247,0.31898440308063947,0.7893520989968877,
 0.08744945820889405,0.6173724984762119,0.8517890223763958,0.5234580773420735,
 0.17493016664681826,0.8572394931277947,0.4481927754037207,0.5712419112418565,
 0.7945591091595019,0.9358972934429004,0.31829614714722876,
 0.056512789028382526,0.22465517475964059,0.40287792254549115,
 0.7564332752835012,0.7526755510931257,0.6551961250863043,0.5466456125290717,
 0.436224921853968,0.5006859104551109,0.5368549735927838,0.32657418115948367,
 0.7332902860266393,0.623240772580091,0.4354905453466261,0.485430581099347,
 0.8166536797614197,0.37595830486496973,0.1544711541604904,0.5585276772264508,
 0.18709005740826268,0.4155506717428701,0.8261579261060382,0.3582541110457581,
 0.4716638083507324,0.801693307432508,0.23566254878550386,0.8694263410000483,
 0.35591638606921094,0.42172955775202037,0.2389804448497257,
 0.3399887287416792,0.8732837568824628,0.3265588518054364,0.7682159588599342,
 0.4550868043310927,0.14841366075640217,0.2768845597953046,0.7116780197411458,
 0.2704181057746735,0.3396383207245274,0.6056338532001188,0.2341622147037663,
 0.26888937734523766,0.8782865597101809,0.41780129374904174,
 0.9638425579204563,0.5041604610705394,0.5521689609009285,0.9266147813276682,
 0.002656122806121375,0.8077143555994619,0.6679983782231778,
 0.8683246352705324,0.7635124333982741,0.9433918155468954,0.4487435027447294,
 0.43215199099093105,0.6704466385102748,0.378009764887124,0.6877670903044459,
 0.16494761542448633,0.15461640279044542,0.8372315812301203,
 0.21316758367716815,0.6782987102465753,0.9108423002914082,0.7129454864145641,
 0.6666379609321752,0.5497543771352164,0.5997828543761808,0.6271495220983588,
 0.1901016554716134,0.5720408191945264,0.23483929440850693,0.9158934350859544,
 0.3800985479451071,0.9052939846165389,0.15793774636246538,0.9463320819073575,
 0.4007244043773017,0.0266782104029476,0.06946478686996826,0.9030160797179635,
 0.5115903527199515,0.40968936522825206,0.35882528470470687,
 0.5159938878956157,0.036535930077740586,0.7555886032061077,
 0.17491583722041715,0.6722046635956312,0.4243967196742586,0.8275228504499351,
 0.1751558351334621,0.6601696191591752,0.9990827618468078,0.9527658509735178,
 0.30767727424889013,0.06055627574595568,0.4183371486503443,
 0.5812631987201355,0.283737390285972,0.14095446325690353,0.7412576157158313,
 0.11798998821520967,0.0689971106796271,0.2038053395866769,0.9641602055391075,
 0.9393024067699635,0.9795908073382926,0.19031843626698508,0.2863804071347237,
 0.14361984228130575,0.006198158658419439,0.18428129436083385,
 0.23479708026439705,0.6541486086283123,0.913750528490962,0.9452275579221305,
 0.8780363986126308,0.7056145550767239,0.05332053612870957,0.6408373725697971,
 0.9612363343723898,0.6870435648745942,0.4594728444659877,0.2843384243754864,
 0.07439593801174549,0.043150683184686844,0.3296307687683092,
 0.3919401781584718,0.3586747848943106,0.831360009442576,0.41356491677654594,
 0.14670382912443725,0.13006021601969986,0.9076765690101075,
 0.20862354137815808,0.16364691281159127,0.8650438913376758,0.803556963174331,
 0.7692849689334293,0.5809498215170891,0.2600298660535334,0.6005942401821001,
 0.8000865828448855,0.7324905257269113,0.28599194615840107,
 0.41712971521551845,0.11644763458623386,0.7291972229542321,
 0.6559514469548686,0.16926095774673366,0.6914053415286675,0.3099781869064181,
 0.8031225733171155,0.8460445284317712,0.7969639506536854,0.22199875973491212,
 0.8096621323787403,0.022383582249803413,0.6131822292568665,
 0.8665459215681517,0.9556779204383744,0.6364021034340774,0.0397191911516348,
 0.3344414067420146,0.7767741371524862,0.5930817195623281,0.8269749223876665,
 0.12737675812989135,0.04040347708727832,0.2505721034962366,
 0.27372198476998355,0.3532551247552598,0.6775853195235395,0.8302095701996131,
 0.3126646906583146,0.41248651972708306,0.7417487255315386,0.2896065702347509,
 0.16590422459085885,0.32280935913301,0.717534192734492,0.12771949715571307,
 0.7837182601630779,0.1960359325090405,0.9801932806512217,0.3684410940501627,
 0.38442897748163596,0.007853662174778009,0.8193648367309193,
 0.7236369687403532,0.3142505659611201,0.4692173210008713,0.2993526309860396,
 0.3832825368915598,0.8234948320277671,0.19006384459927228,0.1960645128266838,
 0.3958051060039516,0.4642749145828219,0.04912127922602483,0.5990027446348472,
 0.6892726024326513,0.49611350866782233,0.7572051958616488,0.1595369952113942,
 0.31624869060709404,0.5777982934602589,0.2819954959022699,
 0.29682263407115805,0.8329778535894282,0.9352950996941103,0.6128585450054626,
 0.903315715015752,0.5084276179072523,0.6966449500566843,0.36994758698768926,
 0.615628528526376,0.16378758478932087,0.6266037823972437,0.6372009718527292,
 0.5252991140763675,0.8538012289924384,0.4095469292566898,0.8545908768283006,
 0.4213089967057909,0.2998387159390201,0.9326532196393363,0.6149823588268533,
 0.8365266792893287,0.47108166645939953,0.6586428368196292,0.5494303295059884,
 0.7127080241522601,0.735119301206174,0.8976926564115205,0.7497075581813576,
 0.3318922775642007,0.19416296603728234,0.27131345043221844,
 0.12952758885557936,0.4716886177511693,0.9679247126448098,
 0.22396621905555558,0.6222644785132591,0.4719911538417416,0.7437636728409827,
 0.6332987580860124,0.6137680840649746,0.06799616598321101,0.9396952830778599,
 0.04223909813901672,0.5266044312923567,0.057890323848877046,
 0.7600841093795812,0.33894726914990847,0.6491750275393814,0.7432969559642361,
 0.6748709774640244,0.12952322014836726,0.3845226094610279,0.8006727963552405,
 0.07811211493601422,0.9441426501592936,0.044684316025095816,
 0.9760111788761113,0.4102845005480985,0.5837753278810659,0.1235510246282483,
 0.6016213415766231,0.3798529813448729,0.22808068977128126,0.7304049575117855,
 0.839622662280771,0.046246913928508215,0.509505095597109,0.7508554275763845,
 0.8976761306220628,0.40983041519148744,0.6941108874662545,
 0.025328282690625703,0.7898214041871604,0.479417163811789,0.0424381329428235,
 0.5190588808221273,0.017891486408013413,0.19384016499897794,
 0.06304511236530264,0.9782908311705805,0.044120047665488604,
 0.5471075703866107,0.3833966652594306,0.8366327678694176,0.7764759524928481,
 0.34378292017113943,0.08827178778018885,0.9106665710980949,
 0.7829109151265397,0.7099679426879194,0.9455488853950681,0.6818816293722305,
 0.5449472855199959,0.9454096654307798,0.5091669280151874,0.6186768080766052,
 0.5435665666635425,0.8626237465444617,0.9335366632126132,0.5948382301690316,
 0.8885196012908981,0.18429927256930656,0.07134196194067344,
 0.0707020148169345,0.10789246455436263,0.024572604233892736,
 0.35447771450532606,0.35738113243721015,0.03898830790077745,
 0.07344123696336857,0.49134101040776657,0.4863252271833982,
 0.5629198283330329,0.13527738699432112,0.0411662560913999,0.8986639560451647,
 0.5102390229993744,0.3407792710415156,0.6061185411623844,0.850294655171396,
 0.9348262862457551,0.6091932733358794,0.4255053282339101,0.48225554050874075,
 0.1328233948722064,0.9386091338150911,0.8063041417312307,0.1126452875092766,
 0.1148415944488076,0.18448404817033226,0.3834258810426885,0.0819847170662853,
 0.19503537455798203,0.3768885017426954,0.21622776495858165,0.580496601875245,
 0.7017700380832712,0.08108975569793142,0.1925813664256023,0.7494492342530196,
 0.31859474510278984,0.904246882623228,0.31985134378708135,0.1198711519186364,
 0.28686825273216066,0.6083390995955701,0.9276952435121879,0.8649406799465292,
 0.4638914814278223,0.0022237356445617973,0.7027703241682035,
 0.6619121355347314,0.8197941848347698,0.308637428110485,0.21826461118617868,
 0.14821007869403613,0.7471293157238451,0.6302368517074295,0.6710214917621915,
 0.2060422254198232,0.549693645070788,0.47663542488183974,0.7844439899702819,
 0.9482387394357829,0.2970266068434748,0.6108796170172923,0.40351238216500185,
 0.9423931313111182,0.7975269465650296,0.7503191184133097,
 0.021092229801631168,0.3409568084866702,0.1710127432701537,
 0.4716631240165797,0.3243223999735423,0.597375476876207,0.3712570380370642,
 0.5237231877945894,0.07612644732577845,0.10443998832208512,
 0.33559599064255186,0.31196285793497047,0.7427584984281957,
 0.6542045739237432,0.2875300450325692,0.023880344226653527,
 0.8657091061876296,0.8345025619008876,0.9764702846135763,0.4434400061976299,
 0.10000331439460575,0.8180156448119167,0.7614283376631783,0.9674660880708502,
 0.08782121658285225,0.8071275144400076,0.024761255440560848,
 0.9123151902214908,0.3967121796768248,0.5167420762027293,
 0.033347854820210965,0.5608085509259687,0.6936703748521028,
 0.8944087155227091,0.15526713876302134,0.7940003820333257,0.7221634430814508,
 0.5550347607725108,0.3442529872161474,0.07730761304564904,
 0.22901573753940574,0.5801651885493845,0.30337518322019275,
 0.9958536839051948,0.09639374099387521,0.8135179482603934,0.5377003637524307,
 0.9774167763937924,0.5330518121851476,0.836983777407746,0.30451235373433583,
 0.1139246658182076,0.8374068373163734,0.9478858632204865,0.5439854005115209,
 0.5924246829857429,0.652470543995824,0.15561385338980405,0.6455797958206798,
 0.136155287270451,0.35069970558947283,0.5311526897968774,0.8121459181400366,
 0.10989564066083324,0.9515731509230061,0.22160523092371698,0.509448380246417,
 0.32232433568944974,0.8531952414876093,0.8446690382490851,0.8111374623077172,
 0.3759201959061764,0.1836453944009988,0.49221869361023773,0.9478432555954838,
 0.35072036788450156,0.037135867871878725,0.487063339608528,
 0.5078244235865724,0.5148895839502723,0.4361432905343925,0.7836684201855044,
 0.9818815190499092,0.05804314442078229,0.12662474301269422,
 0.8523116202510056,0.8320461642602386,0.4667617064761035,0.5812739982150281,
 0.37789779343221364,0.3785490999242207,0.5088684630403254,
 0.46942630374083694,0.7078534429749463,0.4565011987764459,
 0.20308075081523758,0.16672039070797606,0.4846422410620844,
 0.7454707780890337,0.6796382781824982,0.49876974918449724,0.8716909446967533,
 0.9932729836351176,0.9565816445203857,0.1422726604725184,0.48491990702508725,
 0.018521010345080913,0.17725888570081483,0.6953720059569862,
 0.5383179223790923,0.9830431728641907,0.7403632936718353,0.5090737474092452,
 0.6350763986286211,0.3142463612452505,0.016078837400212986,
 0.6584647423903887,0.9534922586690808,0.45032338079314094,
 0.25320896949524085,0.29061028062153504,0.2758621359299632,
 0.21683869308973114,0.8639971595810095,0.07062839808117327,
 0.6817516021842374,0.07626025240156231,0.20522711528944515,
 0.6172170102074499,0.6647366843188602,0.7976746355919129,0.4302531916788841,
 0.9160953084360719,0.6524631306586902,0.7859690613052406,
 0.024844221756410434,0.8143885673256299,0.14022535584392237,
 0.5609038592926782,0.4285070959173558,0.3799060812298959,0.4068198749997909,
 0.9468120923689498,0.683806240774454,0.38856110279253875,0.6519680878813133,
 0.09810489118338628,0.3944555138722564,0.24230030860148488,
 0.3219580823900391,0.5675065928896451,0.49219099437857294,
 0.009130249091065301,0.824366335006774,0.8268760237553411,0.3281993394681192,
 0.25944770837077946,0.4471978658532909,0.7332626407206467,
 0.04880165309470419,0.4124605458854963,0.3299713012822596,0.370850534063612,
 0.2432112998972169,0.9903519051481083,0.26810009048687433,0.8775238820143126,
 0.1408224680373431,0.7145776124164616,0.1515058805328967,0.31769857979310456,
 0.5763674393928827,0.5705472951988768,0.22140161662914104,
 0.39217684660825125,0.9047015223959815,0.7583938283917471,0.6878550596396296,
 0.21804890868113613,0.5515061686995298,0.5402870210836176,0.5027176469353233,
 0.5597545422003142,0.8088562061586324,0.024005428090795533,
 0.6737254957442557,0.7113446580784181,0.31383478225959727,0.5242132392936973,
 0.4291428219463834,0.38049563556021737,0.09812323360203767,
 0.8168681049402003,0.9669653181400965,0.13217967932456065,0.5413556325755333,
 0.14087035519184754,0.3941887906146381,0.09997722628072025,
 0.012823376749799165,0.443679071136567,0.8928279737999406,
 0.25524899848034865,0.9799827338466073,0.2258831182192308,0.5854606224455755,
 0.9121474304332994,0.7117526033666806,0.8876118876431569,0.5993719515749125,
 0.11957380843914533,0.453143090086947,0.8918174616472633,0.9281538481401592,
 0.3159589635082126,0.3833551233900655,0.650220735774297,0.2963255562882864,
 0.09937842449865564,0.5711897944930636,0.015390349103470258,
 0.6584824486393763,0.6447599620990492,0.9365390012496198,0.36704321198561995,
 0.47224519506617446,0.886891790329606,0.9131584896886654,0.804923680865889,
 0.23667479818366866,0.4093145338385723,0.9557001825620195,0.5660182598329611,
 0.41445125124819615,0.2752359942261484,0.7304914342075333,0.8561528688627156,
 0.7855561999749052,0.7436546094197222,0.8396997269326655,0.13227123472211344,
 0.005602508033414888,0.47437027276633015,0.016959809242193646,
 0.745261662675235,0.21636208590627581,0.543529158333457,0.2991852442215176,
 0.21563730022299832,0.9469166477402198,0.2273970382604774,0.0701161867683664,
 0.8691458466395658,0.4103076132629597,0.2592390680057759,0.9242015658741303,
 0.5448413430054614,0.687807677823035,0.1405910487519776,0.9310392898961977,
 0.6612875964258643,0.9356731479753863,0.9013856334755681,0.6624418747516487,
 0.9601686319384547,0.2078608967001757,0.006282017935381856,
 0.2530089730119047,0.5483579326181044,0.9312776755860519,0.24464782189292478,
 0.6206318242980565,0.06422953007951648,0.07328129721591092,
 0.7841056738272596,0.9029515868854248,0.26813632094686357,
 0.05386993225826686,0.42987606762122466,0.38973233908474547,
 0.34367612530593017,0.5531780570645707,0.6925864707418025,0.6416615154325117,
 0.4536768477255223,0.09283536399584036,0.7106195679292292,0.937318492488266,
 0.8519835167671538,0.5452581371054519,0.9317759449463354,0.7149421379551134,
 0.9498055688173596,0.355826037967235,0.65169121426161,0.27102748284568834,
 0.5458056028310542,0.9944936436093312,0.456369905307343,0.6039873833842773,
 0.3059817106600352,0.9918511700277292,0.8572037596549349,0.5695263550272579,
 0.07705555770227845,0.17761449393877127,0.35172737431790635,
 0.738460454623289,0.8177475006199703,0.16387261738790393,0.8857115197703805,
 0.20476137282915652,0.2856865691997459,0.165564688090589,0.38040505268709834,
 0.9557277453161033,0.7250687572807368,0.8579342196372322,0.5183249831829294,
 0.24189831964865705,0.4101150804367997,0.4723036615142956,0.5490147448269789,
 0.7967358313742804,0.5206833161352302,0.7613706398758006,0.5231932917194269,
 0.23005229949046868,0.9584649192063557,0.8110651038005431,0.3752103148322842,
 0.6986655079199274,0.9428976429478318,0.1165426371245597,0.7861275028920884,
 0.09590541335262925,0.9327163551500457,0.06517186188968571,
 0.3620084694311073,0.07130432791501529,0.9907900293910153,
 0.22302084555721124,0.8039033326388652,0.07372381579527809,
 0.7815620233078153,0.24109933590044963,0.17739379313697012,
 0.7797985241088025,0.038287733623796893,0.04919475763343695,
 0.06083393571128848,0.8208576522268565,0.3686784389980239,
 0.47087983129943367,0.39158700483581654,0.7625138127961022,
 0.4747691468546349,0.25175724584692205,0.49569651800916614,
 0.7417352552740728,0.7447574341579603,0.08892467926012382,0.9457722514690003,
 0.16494101319721277,0.5885113120615982,0.02364808005939767,
 0.20230258751134922,0.4031959290989876,0.16225538202663126,
 0.6293695211916546,0.23672527091627615,0.6084064155159054,0.2214518145129929,
 0.4928235522384331,0.27908176126667905,0.8658167052056359,
 0.04966871180902055,0.7454954311394427,0.9497916077029881,0.8901796666095301,
 0.7094465189494168,0.48514591488436376,0.45323882814460026,
 0.8014595201895285,0.8030525175613313,0.6769812048935779,0.7291913809257006,
 0.65743003451199,0.6237159707424573,0.16338640254828696,0.7127014955108785,
 0.6098249120397478,0.9125117551372729,0.46637812442029547,0.5880912891446956,
 0.7861842473520803,0.6142743799617606,0.7140211176570839,0.3054736974318152,
 0.2528184087184966,0.43185697366316667,0.03444748471834236,
 0.021269972933949277,0.5208375237068247,0.5604856042908466,
 0.12873297752143842,0.5165639498031859,0.6190029043542467,
 0.009625215085345862,0.8740397392504908,0.2847851864004626,
 0.10651190174673986,0.7901355478033673,0.6186832057774501,
 0.13053895804500437,0.8552604551736782,0.5934062042983236,0.6137240445189347,
 0.29377373311199506,0.4147795202677702,0.8079280740191236,0.5600271100445922,
 0.5986423533059457,0.7512538390572729,0.3303706191562976,0.42370189118994905,
 0.20238551683288863,0.14919214073497078,0.1231731997918486,
 0.7297696713967006,0.27703702052052204,0.10527967440296304,
 0.10413809780249084,0.49624362682129075,0.2809772818917172,
 0.8137141235518531,0.9023856234047009,0.37768315927832385,0.7117682595420225,
 0.4120563547766042,0.48466812764752687,0.615610172613103,0.10662562042971913,
 0.6212562355558231,0.26975127187249137,0.3493458672901778,0.9376680773739404,
 0.4932343630992808,0.5958225754521786,0.5666581270958801,0.371600253538955,
 0.23859690863366456,0.8666597102326421,0.8793192530709911,0.8135723655272175,
 0.21493128095402458,0.5270248658586512,0.6066240208152551,
 0.32327343925597596,0.8242541154585876,0.4409840518469569,0.3127756168560406,
 0.74023559114984,0.7919905446818415,0.6699647813566196,0.6350575964900584,
 0.5597652588544741,0.7887020637411482,0.918245694003727,0.9349704882960992,
 0.6515661796317165,0.4760536559565349,0.23850986378136307,
 0.06596469554129558,0.7328010840734717,0.17005249800477207,
 0.5496690598216757,0.8840562803786665,0.8379647570494426,0.4110172578442437,
 0.10942717289448078,0.37265076933640096,0.40821749135120333,
 0.2038088770486396,0.0989381425202206,0.17052416267299153,0.1302465928863069,
 0.9525380738257448,0.8756584408756996,0.6599285943221629,0.6546729009204614,
 0.4913901012928439,0.23360107305258948,0.28066497208174646,
 0.13651266767422998,0.7744394457564692,0.008135260651202514,
 0.6690745164290718,0.6666218318527469,0.771394259493553,0.1899984087590304,
 0.8669805519351613,0.8468565872584397,0.11880990913912648,
 0.05491182908387682,0.08393849416755061,0.9016984427035255,
 0.9198473540944471,0.18451638854922348,0.6379031024740108,
 0.40374404025902644,0.10493356189308711,0.6044595557225281,
 0.8063231327747623,0.4791328389161358,0.13763215485286762,0.729532938204107,
 0.05492707185056622,0.9777786271412632,0.34354929689240565,
 0.9707945508181964,0.04567558012619233,0.570958190312148,0.48176011515681083,
 0.8832206770015573,0.9952235449166866,0.1429838232638886,0.34057355975512593,
 0.5402124587954565,0.04088309925380995,0.06880402198046232,0.941489982065534,
 0.7218375967652498,0.7592684732733159,0.8232952874085538,0.9208022239178554,
 0.1971772549327484,0.5834711236053355,0.009592135680688951,
 0.4299977016041018,0.9594675102512127,0.9867145875791434,0.7186649754348181,
 0.6950018060154716,0.8192511091400498,0.5570720792745059,0.3635937863487875,
 0.4498227428042596,0.32082458055026875,0.487697247085907,0.6034266769462457,
 0.7636538258575545,0.14220511926261747,0.5457215540642827,
 0.19104679295979377,0.11354694767018803,0.7797186710363655,
 0.8789671108860098,0.030988350500815054,0.4174586458852826,
 0.17069593217948775,0.9156122977594534,0.7473823433389798,0.9559512819372462,
 0.6585525708706217,0.9533552348755276,0.640217271338138,0.5045948040189089,
 0.7115448729857901,0.36687861863781135,0.026346488483683372,
 0.894602603045228,0.972071877849331,0.5441988551611575,0.8051100714934356,
 0.7232139921500785,0.5552744223572746,0.770427903060667,0.05959875382818591,
 0.04538692626537644,0.7993406115663461,0.3470662209800699,0.9742351166880051,
 0.2596605169700945,0.32630793775051103,0.17089128322685077,
 0.9846662812757199,0.2199168123103129,0.34323563620397035,
 0.07434470711303032,0.4637974348746038,0.7542225101082519,0.6656196685758842,
 0.527226788872257,0.8232838036714492,0.09557315533720234,0.6353322072983015,
 0.18220881058359695,0.10989916674957767,0.8492399172609016,0.519145696846402,
 0.8524575825000328,0.6596001568013916,0.15594147060607433,0.7106907366034538,
 0.3863542942038892,0.34914986959153327,0.22715562449714344,
 0.9610651193961179,0.6124250730185933,0.9486748977934903,0.08243989366399329,
 0.48089527640917107,0.9126955969688291,0.5083845718991074,0.8965410571022261,
 0.5135079116202613,0.8420572057541192,0.6052711662980266,0.4023868249101965,
 0.021879684421652557,0.1906992249784606,0.7902463702130037,
 0.018209141864703238,0.07168472790652647,0.4616682741095489,
 0.22318729655452807,0.7381799665041405,0.3424581892114954,0.7800099662735677,
 0.3913773260500222,0.9884646959036699,0.12906014137234878,0.7592173027496156,
 0.3772750772641288,0.7846067070668488,0.8115527313048658,0.16524586366376925,
 0.7031156716865926,0.905913094104125,0.30430736272616277,0.16200266001661334,
 0.005133851585180316,0.9343432529799093,0.9180870791133373,
 0.01821121416439253,0.860361104967627,0.4863233960302471,0.09613441060823347,
 0.019970491180222205,0.9150008732100686,0.624306527417523,0.9684185429126942,
 0.7388323071923073,0.9804015682358486,0.40775022985785525,0.7996514395305191,
 0.9768062556413057,0.7210315922059538,0.6046050607471369,0.6598831600904522,
 0.7911075622720503,0.16197160313300585,0.4427373868070682,
 0.35125933736824855,0.6313642683206914,0.36792015376742226,
 0.37919858407815876,0.8108780986256697,0.5440054970512772,
 0.48009333761454454,0.34018332107444005,0.7320026844289814,
 0.05153232953574627,0.9230289971552096,0.8183917950260998,
 0.37386585868872624,0.9101511788374944,0.4442060167591213,
 0.13682389234298897,0.42932854406951126,0.28018887002301984,
 0.8150560310071105,0.36409006701170776,0.41611386686066043,
 0.5775444082408686,0.7629720036218137,0.7400522223409723,0.6712455092435751,
 0.17285770412009094,0.157336654332756,0.552909916430881,0.7091618236874706,
 0.8211755089735983,0.22755529719001377,0.47356107746083964,
 0.5964247053593064,0.9596183894354416,0.04711473025561408,
 0.06705079460488583,0.5423753354845452,0.40761397606186,0.1772995576115709,
 0.2134750374062937,0.71723018972653,0.43507365565130474,0.2828695451428531,
 0.7604978313177653,0.8866876646420067,0.09104739811980478,0.3633476751037249,
 0.12439634512258535,0.8454604170121587,0.583669922940486,0.20392605349746284,
 0.8808031080654521,0.34392527788838556,0.012957996834398133,
 0.8419437390897485,0.6388331261699868,0.723370330668528,0.38763451428405415,
 0.04489435293338995,0.4332597547483955,0.4343177951743771,0.2501770940750002,
 0.5246884783701773,0.21654955392570807,0.11625441303832451,0.322075591658359,
 0.8322688593637855,0.19007157802711028,0.2765585003463623,0.8530557970482359,
 0.9252333708745873,0.6094825647482094,0.8719876662147892,0.04380032017325308,
 0.9267380651519445,0.5471488537752202,0.27347874691503893,
 0.23408336616273706,0.0726254759270385,0.6881893665653676,
 0.13830939188379565,0.4180215041009736,0.941260906112347,0.1388436765032448,
 0.9487027461926248,0.6038716004792195,0.10363275102089908,0.3917076599895951,
 0.44036389474475646,0.3745223968574263,0.0632255564430988,0.9530016188244381,
 0.7356452236635206,0.5952193656926599,0.6230218354396798,0.09155734769013357,
 0.9622504028128722,0.3975759380931554,0.36943006163966907,0.438220176090643,
 0.3112695740035002,0.15724364222199827,0.4597873674731727,0.6173161747841763,
 0.9839494098572529,0.8003859690444793,0.29916546790123655,
 0.21596729219802202,0.02553401514656828,0.3275134981702652,
 0.8974801182327382,0.37835643095085125,0.29627231033394597,
 0.26279783531623635,0.8743402873174969,0.8887880675764565,0.6554560270291154,
 0.7689130778317824,0.021474997429707487,0.09008391728805476,
 0.1408628975347438,0.19128310219019873,0.09536723910360412,
 0.0815336151171997,0.2233857087054696,0.8025239053132577,0.13885965614015638,
 0.7211713036610897,0.8807630927115417,0.802540377828799,0.15548074107064191,
 0.5473895315003391,0.28235380804366605,0.11389805019022758,
 0.6373898679720026,0.9921917593977209,0.4757744065554852,0.8225567201027849,
 0.171068032176258,0.8111864388521765,0.8420613257622859,0.7867084966425967,
 0.09496317043850233,0.9805296689431865,0.19054398079899004,
 0.00747502557876667,0.8821408841301764,0.9143697242595212,0.6706118112592185,
 0.9519151837024226,0.15719087167310075,0.482987691655969,0.645179905480014,
 0.127791193361651,0.9668472330361663,0.10917857209064641,0.6414666039614385,
 0.5046399093313844,0.07740468303880221,0.32053994737159075,
 0.7209361121777185,0.7371096583701144,0.6606940177118844,0.18040924177516493,
 0.32221242098204017,0.10325914542589709,0.9903602964029555,
 0.48331117397982104,0.5079350146748574,0.45307915114155206,
 0.9077505464600787,0.9176731282658503,0.2567640524435352,0.7045012545452003,
 0.5534343212635932,0.07738408749166015,0.33794820688640004,
 0.8455447303417741,0.04619760186394373,0.5456849043403015,0.5778311216939729,
 0.04384262677614437,0.9092953267042186,0.8762721351567309,
 0.018084705873497753,0.6389181933002396,0.5737816694472715,
 0.5572180683559683,0.6973229998322112,0.6930839992372148,0.20096508806576474,
 0.4254676234421634,0.22269538269878075,0.5471767984451855,0.7400684222663638,
 0.5919483594428363,0.7530127469612329,0.1459203636230777,0.3074311210373376,
 0.33177268559105,0.07318964800178718,0.6638074529355659,0.10025759751620889,
 0.8866778311646014,0.8627591051819757,0.7856101375995831,0.22759508620263957,
 0.1899092479274328,0.5024052105168847,0.3777622686041585,0.1598628798274634,
 0.1754359108985153,0.9319085181088815,0.6276501801499281,0.6720945296610478,
 0.592244762381553,0.33144826429171714,0.1337336221945702,0.733442618858262,
 0.82297131295022,0.30675390270530123,0.24772017846290184,0.2641682372931503,
 0.17871810583432934,0.7135851890608772,0.1147067287897412,
 0.24471412287387562,0.8351956152593297,0.6317135232756225,0.4413050039206412,
 0.1487449280973866,0.3354936523492428,0.8161819984292312,0.1693789462158526,
 0.26843704917020106,0.8575111374446713,0.994149693302162,0.3227887227167864,
 0.0857513175552832,0.2558427930821836,0.7331926468722626,0.17120594727595584,
 0.2629673893667758,0.6450074158318632,0.4842964786260817,0.7240214109756338,
 0.16689135218201734,0.1414049428064219,0.967373627897056,0.730813964227629,
 0.5448468982176754,0.9764519216205247,0.058436744488135295,
 0.3220382386739218,0.3790882484843312,0.49199459058518147,0.7377508560239409,
 0.8270526320399314,0.433454571457634,0.7698555763066874,0.3303225516047946,
 0.1711627128289841,0.2777228396932252,0.5344154466530411,0.6130735482092955,
 0.9073820199551421,0.3178573913607313,0.4864682068631808,0.05051000126459504,
 0.6442408230023214,0.28765895781768014,0.4255663348087336,0.4911578062357246,
 0.37036781188590284,0.11601313954373005,0.9964784333394854,
 0.031143228879091822,0.6918910095517888,0.7530961689479133,
 0.2887558391179833,0.6764845431133494,0.2595236522079839,0.3347199282213703,
 0.15863187177893912,0.785718663736016,0.20754960999080585,0.6267067999432095,
 0.061553666678887176,0.5765038872271205,0.7342282513113995,0.55271571269412,
 0.4915322662162851,0.695057382935621,0.536017528924069,0.21318673513459863,
 0.8457714701294985,0.1623979599597899,0.3286244221646548,0.6523372277619462,
 0.9719921190167111,0.6007962097059644,0.42243598451025866,0.6122858398018861,
 0.39903220827441976,0.852093978152821,0.5467756417728632,
 0.040474801102926894,0.6353383354371054,0.9334912918469586,
 0.07353977363800357,0.8132996263909407,0.2239349145674573,0.5244851832032345,
 0.2694967094147922,0.9408741247949055,0.43486091617552425,
 0.37452470073845745,0.49718928229210235,0.07669996493989373,
 0.09229408209709233,0.6401674249621696,0.09132397336915987,
 0.8016727768625067,0.4989353666207259,0.4138403915816151,0.6749538050827962,
 0.5038979408526038,0.35584623118823344,0.37485133724370934,0.890379231768408,
 0.590662353814917,0.29949668216486214,0.4572650749935133,0.7900651875216445,
 0.8834024510069179,0.8511344986214711,0.574584797849239,0.8478583460868938,
 0.39230559070899784,0.3760133975651304,0.7861271848305058,
 0.05500948442291076,0.06907591168502536,0.21833324703214974,
 0.8001329003661867,0.04534023136417109,0.8875360606382541,0.2206123114744094,
 0.24989684351181418,0.7365015508368165,0.061817338366488395,
 0.6346005143645013,0.7912852702722004,0.5992105615876828,0.14446934820584145,
 0.9119125978149262,0.41174489340500486,0.14429234542163638,
 0.48326538798953145,0.6631950580502821,0.09608904954191544,
 0.4968661414085177,0.13540816247595056,0.07285917041001999,0.072143336791338,
 0.03339124337233701,0.5937955990301012,0.7461487217914827,
 0.43851393614583745,0.6853934561698125,0.8211374683900249,
 0.08613583330595376,0.6669169291738031,0.04511672982111259,
 0.42824425123338705,0.9100161091203588,0.19712491100302643,
 0.8813149851691532,0.7321835201717944,0.8182399014481032,0.49110892162565645,
 0.5640020346686658,0.8185026051578731,0.44860231458832844,0.2851749146453575,
 0.16475807058101288,0.2787966048356332,0.1925778887391112,
 0.16850211066853804,0.8948221872288727,0.8523800587262321,0.2929829654609072,
 0.5850565431568789,0.1750286302393943,0.5851692356079027,0.9977390273241135,
 0.6122894256694014,0.5362500648993815,0.36786130771827696,
 0.23714841418645172,0.4735399374691186,0.8379663010583942,0.2744848385541597,
 0.6357130551269985,0.7369522182313767,0.5964926152920532,0.23592696912147104,
 0.9085223091037654,0.17364444611691543,0.04863783024795798,
 0.5427620795406424,0.2797478711699213,0.49754920037134165,
 0.39275397775858856,0.5819909359223446,0.6417011260084642,0.5181848449907719,
 0.41790679275134734,0.21970356387313372,0.1251769024349948,
 0.9452845641223235,0.8702675981859159,0.09523307107943957,0.5062561776874797,
 0.18374405460056986,0.6486956460816454,0.5313935830822755,0.2873671470227579,
 0.8337415755474283,0.6966098941677592,0.08924537202607241,0.633178856700159,
 0.05215905600043347,0.9215022056341682,0.9149469573969161,0.5189991775357424,
 0.5119852093616162,0.5947970591663562,0.542233205166917,0.2528489471604717,
 0.7553472725599671,0.2580436764116082,0.42628221253574805,0.594522462414969,
 0.47532942455123073,0.5417461527432144,0.8006000601555652,0.4628088264239122,
 0.6864994767769609,0.6749803097202299,0.9334182620551476,0.4838219401366133,
 0.9942197703877804,0.0700036772869782,0.6967366287769953,0.8097806064473052,
 0.3353672139140176,0.339341870535696,0.8012301613834961,0.2354885983367071,
 0.8380962929670852,0.3376799577127275,0.21531273072000856,0.6844565331576223,
 0.2244710870099913,0.5710128712557452,0.8280778138443884,0.22373487939361913,
 0.26681404164008116,0.12516137633707625,0.8164461760705093,
 0.009796768799668398,0.9388105450749159,0.46633019088216077,
 0.4588374010092081,0.5028852195456086,0.7449569812994666,0.354699737510315,
 0.55846148842222,0.8979471610852039,0.5677268195842714,0.35013542671191744,
 0.09407010515993097,0.13036288771373705,0.9255186357274681,
 0.4152831272550259,0.2531795521397905,0.08248120435436601,
 0.38849527371106507,0.9325042545058261,0.12746210999550778,
 0.9749363912293764,0.7458260580825725,0.03400459705961978,0.5146913602678926,
 0.43292608257572496,0.8593252080417582,0.09322888145192587,
 0.4202940564552794,0.3620278528576939,0.4089145704942909,0.9495770017598786,
 0.8312506853824846,0.6761800761934347,0.7739279627991341,0.360350087801235,
 0.19027575590606394,0.1925649907650172,0.4877514512941279,0.4439469988618845,
 0.7785169198501667,0.12085367703298289,0.4302802264931581,0.559127539945667,
 0.8424694526839068,0.2271228275165772,0.0347744118795138,0.41017124920947223,
 0.7491432263160763,0.4426490956115309,0.0862232034188054,0.5125063026430237,
 0.6358956317143897,0.3268843285030374,0.6795903721412273,0.8327879751293834,
 0.47728372306804,0.31336190328932956,0.16341648267537012,0.45161276599485234,
 0.6222798555668886,0.44694433469500505,0.02533977217824035,
 0.23478462171602998,0.8390723926300513,0.5501149995588801,0.6663068632439372,
 0.3662486177513212,0.14584400499631156,0.9705154349325835,0.6576320268731497,
 0.3183163657115827,0.19491004163618453,0.17936852263063408,
 0.14909544046402845,0.993500542954991,0.7175498607533239,0.5094868999093146,
 0.8237936394881376,0.8318356739328622,0.9123868765215354,0.08178101224034873,
 0.7094886064354083,0.5860530640077815,0.005781602332724622,
 0.3519691288622957,0.6084077894234118,0.03829496846403324,0.9337988535843094,
 0.7512577143079877,0.5808403904138064,0.3418299865740272,0.8472997702058847,
 0.3316653468387797,0.7360722062303772,0.9775787822371346,0.21475343326110985,
 0.285490197166814,0.6366793032539508,0.19391746383888286,0.40520587972860656,
 0.25817705748711584,0.9541619852304988,0.17413549277157614,
 0.3893989799659423,0.08163910682902709,0.11384878751384742,
 0.030445805608384502,0.06680988173423397,0.7107232135800235,
 0.10829302261114127,0.7929174659356515,0.6359226754532923,0.4529812126688171,
 0.1847239785959902,0.17572617512445954,0.7451477409376098,0.9208665598429324,
 0.7807013292770963,0.36762565487393417,0.3116304802367418,
 0.07854801848227222,0.028770103712753348,0.2254606822508355,
 0.27243305247105665,0.8690810363474911,0.8657468197174103,0.8110061397569054,
 0.26838691712172213,0.950300723923003,0.13145353377443403,0.3182445487683666,
 0.3604821437024861,0.18862003578786402,0.35441156612130365,
 0.26222798201945874,0.8748811352103635,0.23274588591654233,
 0.8895246120220432,0.12815123155825192,0.3399072199339366,0.2665683142100035,
 0.5396013339584357,0.570404134161755,0.37677469444717127,0.9582717495794495,
 0.9782848932571826,0.9309704201750207,0.658407155451971,0.15155741811570522,
 0.9812945585894672,0.9008854639067305,0.3304478002966935,0.505687981956395,
 0.7194467763307739,0.06726767441520942,0.1976590491965261,0.7871981407577793,
 0.048429961723601256,0.749753441842983,0.13366441148333197,
 0.7049228050885813,0.013234591721749345,0.47379740457331365,
 0.9803210240981401,0.1821138003786168,0.61171303630913,0.610589019494956,
 0.3061066583273043,0.7487812561771051,0.1103592596155597,0.6277268496070748,
 0.8144929975432055,0.7664108436834404,0.2517696724706574,0.27277301849624846,
 0.08598987363294419,0.5543263714269827,0.6828849699761191,0.6826604493555989,
 0.5542776488475001,0.9112448331026689,0.6966824667668826,0.4728499484172445,
 0.5926992379013852,0.21522685200696856,0.4456646296088438,0.5213784282324312,
 0.11120160328941675,0.8029674442560063,0.572699239884041,0.46313428888512,
 0.797085878217227,0.8877857067193002,0.44959460972949494,0.12972924337421288,
 0.6112470144226289,0.8148350647564889,0.40119131696366406,0.7845166022839796,
 0.23497462437944494,0.49846487298176423,0.5460514317065382,
 0.1756927239402225,0.32522286716747506,0.81072190633093,0.418309846896463,
 0.8522471199893863,0.7033657751529544,0.5010136269649017,0.08618434301895617,
 0.6334240654417909,0.5516501405052376,0.49726850907294295,
 0.12470534309318648,0.5582765428173334,0.3789423811708623,
 0.42399924238321474,0.4017264163783295,0.17474845259541538,
 0.6546195951744147,0.7180470582145662,0.8925010317643042,0.5343993637556329,
 0.2922376350337005,0.23735305337754042,0.37814693023341184,
 0.4462677166814709,0.9500202185852076,0.7381097064300228,0.5857925166104819,
 0.8709851602995989,0.4433811215757326,0.125315739639135,0.8114412580266267,
 0.9662530804839157,0.007213303767424906,0.13822698985903015,
 0.42416664144114835,0.6134387580506209,0.7440932293272902,
 0.45549783654404796,0.9442611478560076,0.2295487039973454,0.3118359817905688,
 0.17660853836062207,0.572022183236968,0.29169600658948114,
 0.05756556291189252,0.9034948044950569,0.7842453810896721,
 0.35297311296685085,0.3386959756875354,0.7417722740963796,0.541192696550479,
 0.9972060627737361,0.8157498401226901,0.28139092988537595,0.6154286652832505,
 0.5987548442697784,0.8207006817258775,0.9654626310342624,0.22145741045261946,
 0.5071806817485248,0.3321179867134887,0.9165381286402562,
 0.007899460000950986,0.5217819937767961,0.5380187925676161,
 0.6653442923014344,0.31794680671181164,0.3951872020933034,0.2562579686877058,
 0.7292014088358773,0.6979149592973626,0.12199782916952895,0.4272105075498618,
 0.966827943140874,0.7205600184617007,0.8692013274481094,0.3230208601694107,
 0.9673388070714253,0.12713117890722847,0.4109600097192131,0.7751392170109309,
 0.23105693956493822,0.29322361675869935,0.1797731086762302,
 0.7220469550988735,0.740557975471704,0.38125067789921674,0.2967353810444501,
 0.01063138662785601,0.4582798772151877,0.38665352371724204,
 0.7260817929389041,0.18213864486194742,0.6952596893778846,0.8664919803667469,
 0.7988921035526073,0.5041671625952157,0.13411883812505498,0.6274878127567182,
 0.6657745914408909,0.014502930153757476,0.07366641420684317,
 0.9799483143737459,0.09397997750489373,0.6609951113651802,
 0.15188141185087156,0.8361829213617528,0.6484155284913206,0.7981667970347897,
 0.993730942104368,0.011074390126860845,0.030152286053772848,
 0.9729380705539837,0.4437537558531518,0.044039681247073914,
 0.40569323956171266,0.4131368253014661,0.6226623689552214,
 0.48290893286127967,0.5508899932016291,0.6833126145258699,0.7929680658222283,
 0.413715991940908,0.7360632516068452,0.379106865420737,0.3198972501056646,
 0.14557993863174934,0.8633551187924846,0.8808990878520644,0.5101011481598559,
 0.43698705663879744,0.2289882875737782,0.5715392013382883,
 0.14476905577851717,0.574430755177804,0.6850632042842899,0.9594915969855734,
 0.20304057251620633,0.9044574132857038,0.45061914905772305,
 0.39221152782078605,0.49842240822967443,0.6542610578428424,
 0.06590528014672947,0.8999113772043971,0.763207617772249,0.38597000448674523,
 0.010058846461126758,0.34064425683131727,0.8344283841594567,
 0.7277514931965557,0.24274979928811913,0.5991828956346775,
 0.43629484813354513,0.387697935196188,0.8094015778182556,0.6344405987841945,
 0.05517213945192745,0.7015251692800362,0.6378186378977522,0.4664722789255007,
 0.17439008433804537,0.03523207199285916,0.8879175548417637,
 0.9069134338037811,0.211872608311635,0.6639497559170405,0.48798552576115006,
 0.8780863790817279,0.25677552435784545,0.7156187470965958,0.4343439773846406,
 0.9725788928754733,0.9741624317380723,0.6763060018120541,0.2704133544409918,
 0.6344278916972428,0.6285017744344575,0.08277909596906641,0.7170632318195962,
 0.8284931311175388,0.0476595241070813,0.09155530998310646,
 0.07506420787609269,0.836666291337753,0.521477243555096,0.7615967610974075,
 0.7577211794408352,0.96413625224556,0.8260114414536585,0.9620448798935838,
 0.9975816963841272,0.3036000018102635,0.3039197729452656,0.2401014391781573,
 0.8501203401995814,0.004418157292601088,0.673867180856114,
 0.47648450216886373,0.3752989441915682,0.5034596201330503,
 0.11126564252997606,0.9434884178473677,0.32292971383150937,
 0.6122804578078307,0.3372163338139804,0.9968364240635328,
 0.024840168424411946,0.35753883982770296,0.3154209078702732,
 0.048120466954310714,0.5595217190458353,0.6217429734585427,
 0.07577050774471894,0.884965798568878,0.5359524514057075,0.3179862793057492,
 0.8393498213068264,0.35711700705971117,0.8277650887753121,
 6.347454191236723e-4,0.8871815267538089,0.11100582269867798,
 0.8069588865020848,0.7757378996064871,0.5371944256162565,0.9375321958380816,
 0.8337017265832127,0.3191192518093414,0.9917268735396456,0.8865896822398106,
 0.4742160484055602,0.7638298385737441,0.5002142214395097,0.3295267755814485,
 0.39628796531589683,0.5641715035127978,0.672316188621606,0.4997481111504577,
 0.9656948088985383,0.7236173455549345,0.7412281257522491,
 0.015093311382884278,0.8915894084011264,0.7356531689730514,
 0.21191817331478058,0.62650817242725,0.4121842131298753,0.08838151686186368,
 0.9765636884611165,0.6674296112666234,0.089934577947135,0.3973001576788088,
 0.38743091725232826,0.059272568890836985,0.4503395716247206,
 0.4571365918937714,0.7143078342251699,0.6226294519294242,0.28996832252556703,
 0.6411184528106612,0.3863347054451379,0.812019666680035,0.18680034409507518,
 0.22254820895889516,0.9038992203554506,0.8202023250560121,0.0229016780888337,
 0.6414135431328677,0.9116140156532304,0.1253744206596772,0.798458478056119,
 0.4610344516792636,0.01943721148976152,0.8322932850669345,0.4450396466149902,
 0.8336966982472683,0.3136139721267095,0.7116714959318542,0.22950468067445673,
 0.1335467281123035,0.4002065200119489,0.9831631891904729,0.5771525932339882,
 0.5134328316333975,0.35917958491095203,0.5583464601562649,
 0.10250225412289204,0.29763554525862534,0.31033074136215777,
 0.9025200491765144,0.6863414584351273,0.2783098598125844,0.2558896797042022,
 0.5144997868462795,0.8338135546485964,0.9369769177686798,0.46755193506141,
 0.2881809791691826,0.15668469106835958,0.6838283816362328,0.6061788787863686,
 0.4203751770899493,0.46675054501584734,0.6918301222645782,0.9548028666930307,
 0.8076763345592867,0.0736177734741984,0.0513818964653473,0.8356367717629145,
 0.5496025836104713,0.04327578079953931,0.66429277906516,0.291214476304278,
 0.8615981215650313,0.8040130512842172,0.10584262770137487,0.5128978443104091,
 0.9405233156039364,0.9803099322124891,0.8228796703324515,0.055959516585529,
 0.4506605520151038,0.7190929613116881,0.967367830632119,0.1172722136989014,
 0.26997023940115206,0.34384856931781593,0.6292601651947386,
 0.009657062108211711,0.5051837906029207,0.14879135021386347,
 0.7235503168740455,0.08481629456446305,0.4901541510805817,
 0.44001284930198303,0.8150671368648806,0.6077809514708951,0.7992775404969936,
 0.9243750002167372,0.6901197385247748,0.21046115245764707,0.5617701383649091,
 0.09231314552328695,0.45823641009533855,0.8700153661414307,
 0.16281907921959282,0.05562578151875475,0.7514609878964712,
 0.31042744615558937,0.034304462364473265,0.2409134753969715,
 0.1319896227940145,0.8078587561719462,0.03764067303993601,0.7697542462960093,
 0.2918742979703911,0.502871377226694,0.5344222951985138,0.9733779645659608,
 0.6192932615754849,0.5112687340027132,0.43939739260053046,
 0.43515710079722847,0.11244867855616736,0.4627916735206612,
 0.8417835281193474,0.16098073208150931,0.5351502301570364,0.8747294396564557,
 0.9503763346469161,0.8999609335446909,0.5442155348229388,0.08227713294687289,
 0.32758552722241474,0.43170492320547127,0.9504685635721015,
 0.3566368863014411,0.6442517623030908,0.727225650354412,0.16906178542689898,
 0.2087630329808301,0.49540425592895576,0.9667847105024128,
 0.40356292176275155,0.9903887138656858,0.015300019522630826,
 0.4065681998262477,0.1394923476668516,0.879439350176256,0.10178366446317544,
 0.7122374240590971,0.6412709673881101,0.6748776054623875,0.6752668708317955,
 0.06016324577332355,0.4008152875870491,0.5735521229871086,0.7578842526022339,
 0.05387492543907124,0.7989669250911919,0.7245460438463246,
 0.19816100781585255,0.8064424131834496,0.1579847870062403,0.1722911111286104,
 0.04356594547092352,0.044799620939329876,0.4127099479939753,
 0.682394387900507,0.08174329161048732,0.9308506939945667,0.5142741545605345,
 0.7794486811751649,0.8564942142883296,0.29853073648371664,0.7539776518012195,
 0.8463675757154987,0.222276658342749,0.3557838945598737,0.5137098561990263,
 0.9426048114005775,0.08918664587437974,0.6353303776412517,
 0.08763883056496491,0.8363339159233711,0.13496637023003943,0.573816962489674,
 0.4999045262245677,0.04102146084915026,0.23415731809273033,
 0.6614632052323799,0.9811219340083915,0.34199310044982933,0.3743814612380648,
 0.7233790123468613,0.5975174091317694,0.5345993664405886,0.5816066183929424,
 0.8082847516485132,0.32640067105915116,0.7945359103640333,0.371747729773072,
 0.5748944581559531,0.7853102996167348,0.9543941456654815,0.7441846205911345,
 0.9237723269932523,0.24049210963406986,0.32182494096027003,
 0.9950748156774953,0.88175905420063,0.4728030467476969,0.4685345067947857,
 0.09474922848030731,0.5476174610792848,0.41932163212666607,
 0.12865349424429795,0.43133273547537665,0.8769030079183214,
 0.2011547962053305,0.6149562653523487,0.5501149390059078,0.2955663683796198,
 0.6205910856427989,0.8028505022123889,0.9998535300062883,0.07748079348535086,
 0.42841167373187694,0.048388739489798716,0.02175270876175439,
 0.5049786249516861,0.548282674583481,0.658263839775869,0.48086990062813684,
 0.2580655435838469,0.32949836681734124,0.48956587964431464,
 0.0038607110752155727,0.1335353036751168,0.6455173903257236,
 0.3273467895415825,0.06936286837820171,0.01377024355269263,
 0.8273056094318949,0.17867019843832155,0.21764053127916827,
 0.5192449015257086,0.23052438277369014,0.5937242994746805,
 0.46823529812156006,0.08053948920055853,0.2562265148025997,
 0.9588647510877485,0.4187535458646079,0.12616701020347987,0.302945084083616,
 0.9150005470311261,0.1833909841894289,0.9813589808035228,0.30007744013745974,
 0.4191921729870324,0.5125800484050194,0.758852762637686,0.5504002161392343,
 0.7328667206460645,0.9131585652576141,0.21789123627396412,
 0.13825473588643233,0.15174875394364284,0.10453876026810904,
 0.410238217465298,0.49658515429073924,0.6735799551590901,0.42445651026536135,
 0.9599410852038337,0.067105695063855,0.1146629200398731,0.8530384841173946,
 0.5926740112191973,0.24336861472445037,0.23731078209891365,
 0.6513186775428654,0.4866328407138185,0.8768508892489311,0.3036290307569389,
 0.8321786040297762,0.19334586242964646,0.23448059764666107,0.531151540846089,
 0.5950156918050462,0.8120035269957435,0.5694125775119315,0.6953524440328671,
 0.16045827210346586,0.212284128388449,0.1408727226746911,0.6806015675485814,
 0.5336524540157193,0.6818403439127844,0.9328264520388362,0.5949434864912012,
 0.49126224798122153,0.46305467763588615,0.05545906086614172,
 0.8935435877668497,0.014700133080841837,0.7575438330498475,
 0.25812944307308616,0.3235245530429236,0.7034469518874319,0.068129863671476,
 0.4972494328113033,0.20575289872340075,0.6735964344129458,
 0.021104781936519768,0.9302437903175155,0.13736289308369265,
 0.4012422923403792,0.9913822952515295,0.12062025729134396,0.2102160232510234,
 0.7149463351004248,0.9829649434711685,0.8044646930099266,0.25438099156333416,
 0.7452175426444241,0.48083844003665743,0.45368783526479717,
 0.6303239097321428,0.9248325193063329,0.669284835422163,0.5007678414724422,
 0.8537224671636078,0.16600915590151122,0.7653531086390493,0.9603052791676484,
 0.21874379438025138,0.37626968661021964,0.8647319572415485,
 0.06718110092024454,0.3863540940067719,0.575925833412115,0.2886436777556699,
 0.5426580600852721,0.2376288102275106,0.023639352821805848,
 0.05909101360485525,0.24549392327890462,0.26558009265416926,
 0.011523063463464989,0.7660365415925672,0.6307825287002481,
 0.31891478938403706,0.5617765686947439,0.6468847387041289,0.6880522685732473,
 0.021579512657930122,0.527955353635284,0.025496290120310405,
 0.2711552242711385,0.47672682036325353,0.3801108467025416,0.4556465398651587,
 0.17668314573078292,0.3290192136719382,0.8418312420242585,
 0.46043855595499306,0.6102411751271284,0.1486956305960464,0.6142454600930174,
 0.5226995537744498,0.6687774505107559,0.818627791576244,0.06588753684029858,
 0.4470842926179551,0.5490085063182728,0.717759563637423,0.3026047323876403,
 0.4404362920098044,0.5499875699717424,0.9199624687166958,0.05998312564228736,
 0.9785344785870566,0.4767377218237361,0.8233151340464078,0.1623255534422915,
 0.524931614020308,0.5107531698843262,0.562711276651636,0.9727663461488956,
 0.2112792922854212,0.09302559490509477,0.7661061511976579,0.0750490406880866,
 0.014837462665844625,0.5817171106720788,0.666611298089181,
 0.07961338121540151,0.9804332018480678,0.2957662376810143,0.0954459812815982,
 0.7754557833506386,0.10861481947824214,0.14003939509630492,
 0.5456150068879702,0.6102549017724506,0.10338859260324285,
 0.49262103628679477,0.47542266934316313,0.6303136009158948,
 0.5690689468923077,0.9922552269506387,0.2837752190812236,0.06617569880144558,
 0.9729434584085662,0.43722211263862687,0.8185806810722351,
 0.13819716801692106,0.6682322614470779,0.36145733835637384,
 0.9298381819434745,0.3592498003215471,0.13100700346679772,0.9318958798471682,
 0.973286765036113,0.6131889521662436,0.39801976220138324,0.517061512526543,
 0.9388445668278268,0.6664817246581629,0.670576119693401,0.5244919624959771,
 0.5123721689784977,0.4122169904581955,0.5920372602945947,0.6975108335039926,
 0.38602858511189175,0.518181488463961,0.2350450900558383,0.18003292837044138,
 0.6487426004169121,0.9429175382111653,0.4954466650367829,0.3518235089503483,
 0.39334076873002266,0.7224307827946186,0.05173218995783935,
 0.35120553179683056,0.5480798360776808,0.7722742136961317,0.9841294862749861,
 0.7087099924598058,0.2742796178165541,0.6305284772988644,0.43005805022008436,
 0.6855002992134704,0.007480865247252,0.23292533452366115,0.3155812740362193,
 0.6121390499228087,0.1514026926537333,0.7063079664391686,0.5407198649949623,
 0.766311595827655,0.6214893632782668,0.8011237688750277,0.014504716494542769,
 0.46800491890972173,0.9239930077574903,0.601046480365744,0.8129362520593598,
 0.9502738757766829,0.19493946914102078,0.5515204952132825,0.3806858035840085,
 0.43572768904871906,0.06172833206396211,0.6963499886230964,
 0.8011579655639884,0.1290405287025549,0.6839871524411424,0.9693825405204493,
 0.8782265000022965,0.23221937682822347,0.03806540088024524,
 0.030207576366548805,0.23102099486790495,0.14201548477930626,
 0.2699122364888745,0.10335657291114952,0.3967172206938461,0.7635594386393093,
 0.17082666163336413,0.09692438925207403,0.6319655279750416,
 0.08717565167681673,0.5831398749712534,0.3451216507269924,
 0.44080106802241503,0.9719464089949098,0.7040280265821273,0.8738303483110749,
 0.8010981118302017,0.6149199685473201,0.5789231673731586,0.18434419977007532,
 0.3033188609870776,0.7706622636172735,0.9435313688840604,0.7050206811446336,
 0.27492025561101086,0.23566751380668727,0.8788781328018049,
 0.4591030918562028,0.44737659518058537,0.3274652671132462,0.9034528381744137,
 0.49834641303946825,0.8294884258316154,0.08807502393938393,
 0.15605704158957512,0.6675179906940926,0.5387086079440804,
 0.32097058703952164,0.8312677109577374,0.9959588737942993,0.7774092150736155,
 0.9502643767464963,0.7755179112645031,0.13043415102460165,0.3199066913940356,
 0.8034273928127245,0.35327527640833445,0.3555085873800484,0.7593837194192793,
 0.9372533819180786,0.4664120514769923,0.5396760246189185,0.8678758951843263,
 0.8350264290794176,0.9344990476109608,0.7363466822007932,0.3575557853152047,
 0.6857563457071171,0.014989067770218134,0.3829075068901253,
 0.9488339406440195,0.0610445638132987,0.8972769004566027,0.19398989805934885,
 0.9837818243261379,0.29447238674671095,0.24810693569244613,
 0.40554197800052805,0.9626332402931159,0.9918574312041091,
 0.42281202999129897,0.2991710518069446,0.7654929586040866,0.9029162698350222,
 0.22672381877013204,0.6600394030004822,0.42453720402594575,
 0.26084211919975187,0.17783382190559915,0.5622600598608589,
 0.37764800707921586,0.4674509642081516,0.5085174595934892,0.7072332081334145,
 0.08498083125071354,0.9646435481144046,0.09390958331364141,
 0.9443836372158804,0.9684074257586647,0.3638856393371156,0.202683157950723,
 0.7252765243477169,0.934795869321372,0.9794443077220317,0.07868421631989497,
 0.9137240264723254,0.6809455777602985,0.5074655240117709,0.7153318437312461,
 0.9320816207773474,0.733959217149769,0.9229109783155213,0.5001817964356308,
 0.5928624548314857,0.2959564098794767,0.1228355244503061,0.5914975100368318,
 0.8971478955463916,0.0663728511808932,0.5211374446389189,0.2635533394835302,
 0.8041112818506606,0.843866194117961,0.19771572530003534,0.5260912748754331,
 0.04777902961193381,0.9011704808337289,0.9857642009164089,0.324090551210903,
 0.5588962230817248,0.8279180588365436,0.1924708823042972,0.40217199731114806,
 0.6538185210434881,0.30278308260373765,0.8977139839188996,0.0698411944337528,
 0.4226352030043391,0.8336423378785209,0.7701730366527129,0.8484720823884156,
 0.033782998320701974,0.3177008926909366,0.25903031749636374,
 0.6326262095447983,0.32997052936251037,0.12866147578375942,0.511901391470267,
 0.6188490355050844,0.1906581985236987,0.3640526004286623,0.02505477045526905,
 0.4593779003044487,0.04640664897076752,0.16252453077411977,0.369252246899626,
 0.42971464749973765,0.851075227260419,0.7064820687474325,0.6882739809266852,
 0.5532845492702906,0.014569868781150741,0.9539498413732146,
 0.8339604901773413,0.5552364508899135,0.9004185347336748,0.7510733048694075,
 0.9067367543724142,0.12416021426144808,0.6039907540222742,0.412116006689041,
 0.02510342320626735,0.3618165905167192,0.46914633876837153,
 0.06959182285158405,0.5608063690852498,0.6215784795846793,0.6558112026468502,
 0.22093085434969817,0.3919094528263912,0.9577490234324721,
 0.26875027107364535,0.9480929647149692,0.588599377392604,0.24142463718457297,
 0.0698613953307965,0.45007758358177263,0.9805273132322471,
 0.21264922942831332,0.5009647178143882,0.3267437453015395,0.1015645608612904,
 0.09155069696497309,0.18696210945919267,0.8799240620310325,
 0.3438620200995378,0.9198071403094763,0.1576405859090746,0.3081524166026909,
 0.19211567408675478,0.8559589963713281,0.5128311500641487,0.3316865536444612,
 0.7625647247853412,0.681463005502685,0.12243080803766349,0.30727362202172104,
 0.7135750843907183,0.03336631907468457,0.1937204111324191,0.5916294682341573,
 0.6217916244907806,0.11504781668636699,0.2251195430121532,0.6020310058695372,
 0.7686311672550167,0.01222306044107302,0.41384348254945935,
 0.5829149517126939,0.899611969280831,0.2831268233685599,0.3866806324214692,
 0.09611622811928777,0.4022310985511591,0.9074465828068621,
 0.02754886463360895,0.42405779986177417,0.38738519855117526,
 0.39874359110614976,0.08076088547001237,0.37079591429514014,
 0.7405512735510469,0.757755483376572,0.9369799848881416,0.21704432793006312,
 0.5942281200266641,0.11813332528882536,0.5657345293813044,0.3548891449806556,
 0.6524151015885838,0.911104124291538,0.48537915194803716,0.981228195435835,
 0.9724670393982436,0.9647312123287026,0.39204437757987676,
 0.22929166310449967,0.5669878929041304,0.1702142366359427,0.3344535045166589,
 0.398124697007348,0.10206917012212025,0.7111332942171448,0.6288241251746876,
 0.558372499785719,0.8090585310856442,0.0542067013898927,0.034767802209633936,
 0.6415980164577312,0.01284992632999915,0.5330041580637609,0.4261862532543321,
 0.3122814109453317,0.32151392141223056,0.8307968547073088,
 0.24337565311516385,0.9664382623645904,0.2008592961489737,0.168221115912488,
 0.10591874665429013,0.08412123595374033,0.0773600542094629,
 0.2814909706367662,0.45069860966336983,0.019977172450432068,
 0.5276120851662844,0.4112506620804892,0.027767345702011226,
 0.7629556399242289,0.42966206936569484,0.9173710401847739,
 0.05621067050843265,0.45402806451672706,0.2772802938545458,
 0.4162636874682091,0.5688811975385092,0.100296905495177,0.9878699276201297,
 0.14503425669686953,0.46266121647404257,0.2131925199034388,
 0.17042885169816202,0.37806103385830536,0.3738398347772399,
 0.30054259965189667,0.9022765239134364,0.33288123929569124,
 0.7620162057466585,0.4792347887160716,0.8792382085020793,0.15487569170552118,
 0.8681603366290581,0.2697191171199984,0.2924005927584685,0.5376708986854981,
 0.6050841110514676,0.24917202509062242,0.18820097607102948,0.289215119380392,
 0.05309492504199198,0.2941990833031355,0.8861899904677892,0.5843635318168765,
 0.15656327954531135,0.6861171913342113,0.5198958965346436,
 0.054529241357240354,0.40270082335912283,0.4021114846952224,
 0.22920251876165487,0.05702400002072894,0.7109542810094585,
 0.6272479512538441,0.08586948523652593,0.20794092822349397,
 0.8827767911411794,0.3597667421068955,0.82509907095137,0.6407287106618748,
 0.9618423086360143,0.2290786344029545,0.7832760573183324,0.7662516031715778,
 0.5329576857736577,0.5949436206760401,0.532602214342822,0.16211443704027328,
 0.4686759031507416,0.15218822169201884,0.08030967012149426,
 0.3285215523061572,0.20939290232028274,0.5364172966483616,
 0.19249607096305277,0.4334204914522146,0.8416502146655426,0.4326014986488236,
 0.9981991359266692,0.28384757955704554,0.16873407142449137,
 0.07253694293594948,0.5647579829543049,0.8007655120367005,
 0.39962068561897723,0.006921687280653344,0.5894759820728682,
 0.16656826443576156,0.47780966580939543,0.7422392653244578,
 0.9401970440974462,0.05427452620716644,0.20142383012643428,
 0.5165837893590851,0.645803138962639,0.42881918645975703,0.345320257373088,
 0.4473304117793523,0.14309526551753082,0.37838254484771017,
 0.9950360332636374,0.28973697582713,0.8674114726925106,0.7644543258442758,
 0.7888939278890346,0.9695693226501061,0.24822799154080788,
 0.35003910751712286,0.40172739482540665,0.5884734351767569,
 0.39914334461888756,0.9885187051303878,0.1286335394443624,
 0.24462370198146888,0.06578647784129488,0.33201129374599714,
 0.5678891449155661,0.9097038473089484,0.5266329044889304,0.6100857882022855,
 0.3140112648870186,0.6628535030749,0.614182352189545,0.3744462852986188,
 0.3508723356082126,0.649148200493449,0.31825635044232914,0.1507301204182261,
 0.7459478363979208,0.0661312534794114,0.12225623270441321,
 0.20776464393601013,0.10531854395151274,0.6837504027795591,
 0.5976679020253051,0.1541945010796021,0.5823989373406315,0.24623732324846026,
 0.28165951274636425,0.7122829775853253,0.8484632748998808,
 0.19900512095140743,0.5533610784777304,0.5296760292534385,0.9426195630949489,
 0.4537734932812144,0.9245438718689961,0.056659040922028114,
 0.8973985465529619,0.69765652312562,0.43649257210397674,0.33047167545590783,
 0.3615164106536817,0.36300444077213145,0.9658145703540267,0.7668440958972433,
 0.7357780433765373,0.9325239805803422,0.3696262026720947,0.9241414276145066,
 0.6267332631661396,0.025801782939566364,0.2614189287066464,
 0.23236467349776246,0.7072398911436191,0.18183242997401683,0.986152096097702,
 0.842916969821435,0.029937491179686027,0.9209950370834263,0.2658845294613075,
 0.6504984487721372,0.3065360194142248,0.676382908371724,0.995544009429928,
 0.4732884878493617,0.4295684286974032,0.5297965039407322,0.7318208594902232,
 0.49379630211290315,0.8655230049226468,0.5780758979084947,0.9103468479304035,
 0.34376359180181737,0.05773509939497257,0.7459806156605308,
 0.8038771710984663,0.8815037482272168,0.7607465508207201,0.0697082638857337,
 0.23413512235010225,0.38322428074196,0.8523524947581319,0.584488538540962,
 0.18307861884060284,0.24277105872372406,0.1318897963370248,0.5790497699353,
 0.8057177407737776,0.5578163857211319,0.8287085985489312,0.5943688664766213,
 0.46824282773766757,0.876832280454154,0.10646613514622782,
 0.48470318329195905,0.991866673769166,0.9322838827270661,0.41849111190169097,
 0.3345419105842443,0.8838493020144516,0.16494847189348927,
 0.48494017152074276,0.7294118061029229,0.038021887758270645,
 0.36579560522817955,0.6829376731547822,0.5467513689113009,
 0.009821920719462418,0.3329476613503879,0.15501349944842213,
 0.645807973101106,0.1426592454864447,0.01019491439576059,0.12318607673786486,
 0.1480696279703474,0.5862342648811063,0.585771962752687,0.7938373772914387,
 0.7849799712594474,0.6672000796323356,0.24104851080845557,0.723095254546761,
 0.40591493288301894,0.2835231819912879,0.647990580560319,0.7495017475416185,
 0.8077121600032147,0.7643755303832852,0.2038568654308761,0.06747111377409465,
 0.21670934284728305,0.4956826142302717,0.9478634473478178,0.9669882057864858,
 0.8437353849664565,0.6303281690670757,0.9992785655355578,0.5869659137491751,
 0.8604563776981977,0.8855550806383585,0.08887733907567275,0.7277332136183089,
 0.548569965204689,0.07437519316166685,0.057164361948860876,
 0.9299370892903563,0.3499002897638377,0.3056426388112661,0.5055882300418574,
 0.19941390061282083,0.4151567078868337,0.5133742541912383,0.1323168899357674,
 0.951360752031259,0.19821310423036098,0.03403091569159333,0.9299772195632028,
 0.37661762508700836,0.20650130153762747,0.2966495769518984,
 0.3281322018376337,0.24121723114032123,0.8197576589528597,
 0.004266854356111516,0.358552816913989,0.15080148270415328,
 0.06796610507076895,0.18315809262172777,0.7441379638023577,
 0.3199848860984289,0.5764938615034525,0.6367112732136565,0.31860101901009086,
 0.6285367072710921,0.29124914912068156,0.4784045178831401,0.7072182426465036,
 0.45451313574624597,0.6557131133548812,0.6747917830077784,0.546345101188159,
 0.2813354324204562,0.0687380033605266,0.3657926255977807,0.45278973238947895,
 0.9922177642961765,0.2961499612980075,0.2830000641997261,0.8120286125615674,
 0.8441137399290435,0.8660768233747375,0.12077566673114348,0.3515193433451179,
 0.9486264617020967,0.45351530435483056,0.47988962198804563,
 0.6148915788548502,0.27739093223097944,0.8119971240614512,0.2368107459920442,
 0.04330184925952585,0.5088181785309253,0.3681980384171344,0.257304600597556,
 0.09906845804520437,0.2145815946806191,0.3162708774882643,0.6418244555915017,
 0.7169954961453789,0.5434883041068548,0.458113940119117,0.46355583586127036,
 0.47595316212139993,0.15703105468145917,0.09205640029402828,
 0.46274353316692407,0.044152601841320305,0.6415738438105032,
 0.7899913637216647,0.3379479799485715,0.5159210296583,0.6866853799309367,
 0.662209706831085,0.938657004121685,0.9331586588029712,0.6174875698642267,
 0.7627990808191046,0.593699874562435,0.49994497169250873,0.18981965823615587,
 0.8547965624132545,0.6109964920716997,0.9666683675818404,0.8395047207739814,
 0.8371354985824855,0.21815551557820778,0.035400872972948805,
 0.6549822652176837,0.050335976271356,0.6484793856397681,0.698124917271154,
 0.6484979630371789,0.5165165857723839,0.5894924298615215,0.4215008407268368,
 0.2602111472521962,0.015368326994984804,0.8125500035834374,
 0.8832765966866344,0.19685372784864663,0.8109989831115801,0.6490263349661298,
 0.6890788306887189,0.5715824311309038,0.36752491757707406,0.4682011184662007,
 0.8027043273455635,0.8390239158231739,0.9924446108407311,0.9585902901970336,
 0.8201120341623456,0.6881653271301934,0.34195757451257913,
 0.09378195808308853,0.03867207216294932,0.33421184692013106,
 0.35616038030386954,0.3985561928418697,0.6310412550503426,0.3427312734359177,
 0.797970452082862,0.4217885960654214,0.7296756607611306,0.19032698697114847,
 0.2487313394194619,0.8848346863162618,0.4520791126704262,0.3236368691356991,
 0.5911606547493071,0.6460544244417101,0.8271563417674352,0.14747660090882697,
 0.09134621215790595,0.3619585096817839,0.8562786060892515,0.815651932247525,
 0.6194759463022329,0.11582500212854929,0.38581914608820533,
 0.7137429312950108,0.46389076223738046,0.994160534880588,0.43226244572914585,
 0.6531891460837153,0.5607160916453711,0.8230319452322664,0.25178830645393,
 0.3022187347119403,0.27555812003030655,0.39045310617005136,
 0.7055303435106909,0.555965561396127,0.6621529378517089,0.9793076492899107,
 0.5112282538006306,0.6906308290169948,0.682162803805258,0.9198944438429862,
 0.3469664404228142,0.5240891259792714,0.30740743777579627,
 0.08754316328698808,0.7220071092609399,0.6577280427761563,0.9880406580383366,
 0.9031029844599641,0.3455493200598938,0.15193493975635686,0.4418596054157088,
 0.6077976337336211,0.48978123302403287,0.33881793104164504,
 0.9986095811291884,0.998465581364128,0.034025298376318736,0.7179273327015101,
 0.7727858904824931,0.7627382473786236,0.9764021217862173,0.8431396028697585,
 0.013459298733906055,0.7028598082036823,0.5906218687875373,
 0.37528584316731883,0.6146499351816518,0.074316553430265,0.984694170150789,
 0.09461151510692178,0.6415072160737084,0.5544363864531239,0.766491238801963,
 0.04880229700087879,0.49378507779280223,0.7352009277531024,
 0.6872624566041762,0.6088394383391231,0.4909802148170358,0.7006541092824574,
 0.21592592657901954,0.8271701189581352,0.8108972625958977,0.7425037355812927,
 0.5797569307890696,0.8100405745063796,0.4973416158271271,0.11296495229932835,
 0.8039629013444984,0.6035853751381892,0.7979065892937974,0.166413254514041,
 0.8806307064064062,0.3753004293382727,0.8433149301430675,0.5054907622969533,
 0.2678147146617128,0.5235317958566752,0.37203671398413884,0.3038436701494992,
 0.12150738717638176,0.23573791693297652,0.07345044773561193,
 0.05656834344321826,0.0567007724780757,0.5506118741672897,0.9906507786114671,
 0.4252216838365329,0.508436009588197,0.6305998053135984,0.08353993077144539,
 0.1057219005801675,0.7869283587281093,0.2189720612876468,0.323359630920366,
 0.008121120872034981,0.2909880915452928,0.35523498025936573,
 0.8325752528011546,0.47643258939956257,0.9756270086312656,0.2891957055827775,
 0.8101297119450939,0.9647223527492509,0.29700297370827344,0.9388942616208034,
 0.28276103738391145,0.29728163743381164,0.834661337655011,
 0.47500730860940266,0.07627694016688624,0.12862301515798746,
 0.7725526616131373,0.4421562405458208,0.30277120237852506,0.1371096223299012,
 0.9571726239087003,0.4388072864838647,0.06951377858822494,0.30273147392557,
 0.46846951203536447,0.5284708600756198,0.20684965904894603,
 0.22635463104527265,0.40504253375878707,0.1285976600352131,
 0.9520589612400467,0.314361749668549,0.9662008454462193,0.2487665263314659,
 0.36327753038267774,0.7161644291760259,0.902244557095835,0.3880360254064814,
 0.8497238264728598,0.1548871964389168,0.10735926751323288,0.3813707952779283,
 0.09232020265976482,0.5564295282721271,0.8437089540558185,
 0.39030379190143827,0.021605509844582205,0.6407279221925246,
 0.8903461369924895,0.4212112318259551,0.8854058568350399,0.7881872398794152,
 0.1418810036139454,0.8366914245177786,0.22304435836556535,0.5226138571581264,
 0.4305300369262848,0.970625285684401,0.6133813656893083,0.2517093472938372,
 0.709166907169451,0.043583500671028785,0.5044769216870119,0.6711814117242072,
 0.1843240246550556,0.2978147674148185,0.42969865925014616,0.8440392651776526,
 0.8208497376810006,0.27933321137023825,0.5028687880960884,
 0.21607522741623286,0.9336017454866867,0.09508345706462162,
 0.8236086684965496,0.9248731761664883,0.5672332389602178,0.8383218811098772,
 0.736961684027809,0.24114199418582127,0.9495711150836808,0.4698930353233586,
 0.690672566393123,0.19706243260224077,0.5325222321745992,0.3273750451273373,
 0.002434459896098584,0.15345223939632413,0.45008832049862846,
 0.10104958166463474,0.24894873914289406,0.2547141077114967,
 0.8775405896006869,0.05904118897450805,0.8980295751733651,
 0.31428377528015716,0.7240086294039136,0.5022858834308321,0.4806438393187489,
 0.4342152201569127,0.07066144323956336,0.3684034472068093,
 0.10361595147127023,0.6452992796771262,0.028999964657642208,
 0.931226868402966,0.5690127687972879,0.5881746794832156,0.1654699206403354,
 0.30425903562373224,0.23940036637815532,0.1582769055597314,0.635283985152303,
 0.2598688116344635,0.18099014441126282,0.3612908347320868,
 0.23582020952240623,0.08158515903224295,0.13472903522676305,
 0.7361895283261003,0.9725234916328608,0.7299496867123314,0.673798753581453,
 0.79100782511378,0.16826242477632347,0.9356936973860357,
 0.0038757481937122806,0.2458947626786695,0.3850391506125552,
 0.3732517992947345,0.5668005394096125,0.17761627812124814,0.5272814621612547,
 0.6969905711781453,0.8628225293525422,0.8205371036675565,0.5688671306871727,
 0.7827429835118702,0.8310776331395372,0.03756363256749351,0.4751572520661199,
 0.5792736513225513,0.08621839417999266,0.4052300698658823,
 0.33052687203740216,0.004663700535150306,0.5554932984174442,
 0.8183369323418046,0.11290715456473133,0.5563220987058978,0.9267567749025193,
 0.40342633489513546,0.07211596428824485,0.981168548766943,
 0.004189975746625163,0.5265529053671797,0.13110534579529565,
 0.6630399763026469,0.5283520221931018,0.058913173316064604,
 0.23722379657010706,0.858553227298349,0.06446103160418726,
 0.21387221593758177,0.9515838381710756,0.5913330530839452,
 0.47826428423211553,0.3782787812123105,0.1559684370200488,
 0.43985984518038657,0.22781868383050918,0.2589086932753213,
 0.5819048593680132,0.5529965094988119,0.7928116147550685,0.7894564747785563,
 0.608059661178759,0.386060794063786,0.6055934625631983,0.84054948219178,
 0.4194264012175799,0.00825830597012267,0.27395194033677395,
 0.9111886049405621,0.43273008924319845,0.7310433301646581,0.4349872687546834,
 0.3727481501621288,0.034898001314617955,0.5123293879305387,
 0.6318386027874396,0.9613019345337552,0.6648259710278426,0.04690919161569318,
 0.33215206125006413,0.6418076101724159,0.4793631420752935,
 0.39709625559393524,0.6300944895109901,0.702032056848943,0.6616442764390328,
 0.520065289280787,0.8378857367116119,0.4700945971841658,0.8752649144464172,
 0.020369935165701403,0.19519640651436532,0.32497710402817037,
 2.806763471152607e-4,0.32551485362087007,0.24866000222118567,
 0.46271538141972324,0.8438012597591333,0.04207695289864366,
 0.7924997679842967,0.18970841555339224,0.03020234327891891,
 0.5546753305359384,0.5266305454087412,0.06195162863094228,
 0.43851891422459466,0.5257650306081477,0.4141024623828887,0.4268794904350217,
 0.48567776234025883,0.42673235242022944,0.6647293375127448,
 0.6534506283534117,0.8193665444203718,0.78368064284099,0.730465180417032,
 0.9209302596411455,0.013373969086834747,0.5326221265180381,
 0.6009793598500933,0.4969545152879159,0.8573944961620379,
 0.029083232453981034,1.1583942799242575e-4,0.19373597989513935,
 0.028156141004150093,0.3519771982901152,0.4742019113040068,
 0.8114409108959313,0.993582290879281,0.9751744569646684,0.44664127188053715,
 0.2100643437860341,0.2343845721638088,0.13910423550142637,0.5353595250922133,
 0.7499278428647982,0.6210041540833302,0.3824660209216718,0.5966589436129703,
 0.8427063985179567,0.7831942272566109,0.42337410293478905,
 0.42759582596512713,0.739061700421333,0.5538494931873156,0.16165084469219204,
 0.8403591259542775,0.8062655595426037,0.0375359222324394,0.7587080486406172,
 0.9871918218384195,0.8804266460930298,0.3979252717427646,0.22412431517538356,
 0.7649458084833426,0.3152819367716353,0.17769850742779714,0.1587797391360437,
 0.8583150929990501,0.694466042115909,0.4325531378252808,0.985072231365889,
 0.8116725414483817,0.47084926787717574,0.5827403300089079,
 0.47741456503117896,0.49949836904868106,0.15221080532540976,
 0.9633952745560315,0.8179945836635122,0.5476696659065164,0.4002258119698342,
 0.9449240798488934,0.9700697125391248,0.6573647188725498,0.07693216197625186,
 0.9401207186524408,0.32377746626474013,0.2495205632940689,0.9959148218536058,
 0.9322628994773732,0.8466780547970288,0.4535978621523332,0.08012950580779354,
 0.008349091299783229,0.3984898835301538,0.1963447050169942,0.760649648293418,
 0.1547808033708311,0.8311134044467968,0.401118588366539,0.12006957983410549,
 0.040608062029039216,0.06374764453665716,0.3354486937271389,
 0.2113129867903285,0.7523983194284891,0.09504562361812063,0.7615812623138993,
 0.7916153201889593,0.9946022374670158,0.649062042158442,0.4121734517169122,
 0.21745830516385078,0.5556532713283806,0.15780985437961237,
 0.45410633281917434,0.5704977372134981,0.2743836127308905,
 0.15458096268306276,0.6365442371419493,0.1419602709616411,
 0.19739439715886864,0.8880180080328951,0.09312117267759068,
 0.7376780407162088,0.3061071203862197,0.29249807462266575,0.4460681068212604,
 0.9522965111621994,0.2757254908108495,0.0587497986479113,0.9576323515565386,
 0.15600805418941432,0.5631104958174407,0.6599591539469403,0.6703476372934869,
 0.37836505463267023,0.07856407465859816,0.8787729293447766,
 0.44999204535156256,0.5494682668904152,0.3725505203389772,0.6185606929774021,
 0.3816272217688379,0.7495550863485382,0.0432973725641862,0.7926364410740476,
 0.6976199424272322,0.6026693155191095,0.9047568218723915,0.1517866014316811,
 0.33417098504709797,0.4533093578304579,0.10775005838202012,
 0.5968404710950524,0.8376779087655455,0.6467873815068546,0.4954968706631776,
 0.8515129827561416,0.726738197158035,0.8658472460030388,0.10594580805880027,
 0.6007495296141958,0.37117073010897383,0.61374623898204,0.5679328526983177,
 0.2477210148060539,0.9903717022385252,0.7215459096996183,
 0.045383680933927084,0.7606621396323274,0.5488764018728021,0.83367487509659,
 0.09525369103036074,0.5925957756477528,0.5049780927920317,
 0.26591743206438545,0.6002970322809019,0.4241564309926106,
 0.13271962919532843,0.4284648138922049,0.22886443104088605,
 0.12580025337456047,0.9243296250253461,0.23829643365973252,
 0.47781582987402793,0.722259855266455,0.23244776096388842,0.2410140618163903,
 0.5602736493591783,0.17763479921452285,0.40921206444033054,
 0.038647626257729994,0.1990476335944713,0.7350611383036615,0.733345758160507,
 0.43408536958174293,0.4949519974968142,0.5379084977413996,
 0.33650685154472726,0.0429317512620806,0.3881519306180421,
 0.06075728776053124,0.4332072605590749,0.2690758911065697,0.5368993571972056,
 0.6710580305470337,0.5189496957768374,0.2510774621846351,0.2221186716694552,
 0.3045990191528607,0.4527400711205316,0.07765819108390393,
 0.07757081726244497,0.5765598012260714,0.26766487336840594,
 0.30055003878428965,0.1729153682349357,0.561258705233322,0.08084587499867413,
 0.632569907082242,0.48188208020724665,0.7684666186201414,0.6649964691361886,
 0.08653621725851579,0.40117869945011875,0.8817579312637148,
 0.2097833729126093,0.20895481289873852,0.6557032517276705,0.8413690068926322,
 0.8196586360616929,0.4234000824454107,0.7791288775947922,0.9443450331143457,
 0.94461318171491,0.2176949971730725,0.022274920302093437,0.614543069418652,
 0.013814462382087611,0.5882438367768279,0.9739410324431895,0.961726922564852,
 0.5845911040111601,0.6484917925981579,0.9580238721655685,0.936770667261536,
 0.25802610455021924,0.25921956525036405,0.4265119541163275,0.952026885892829,
 0.14230362873569047,0.9861344201856541,0.9741324765795838,0.1117840250957487,
 0.5654874745178596,0.36242561628256065,0.9326508691706211,0.3382843793951261,
 0.9453101309077843,0.3158125516605783,0.25002325382410306,0.8148052234203301,
 0.4487213759483939,0.05771431825515971,0.7690400010569797,0.4174022362719493,
 0.9971505656674278,0.35643408127066967,0.1780481463665834,0.3466210455385489,
 0.5462894653907254,0.21115274100719583,0.27933117703417476,
 0.20543767090356335,0.9268322482749582,0.259567767037556,0.7147962645743529,
 0.4220207613317457,0.7367066174839231,0.6107900584240932,0.7053077696775994,
 0.4728003296197527,0.9054878068763131,0.1868547035408542,0.5403974511984551,
 0.5077123444438792,0.6963642405184806,0.35161554339955836,
 0.32397634387594454,0.5312828429733158,0.31030017229499385,
 0.7798769932514601,0.7082023437094543,0.9880881210642503,0.7128021434004461,
 0.2973413824969422,0.36602573033849695,0.7395813310224639,
 0.37598295154165207,0.41080590662923056,0.6764314156916807,
 0.5922368790508864,0.7811383518983142,0.7575286855699734,0.5176645872940313,
 0.4514505751702691,0.12317258942129161,0.7590746470983035,0.3975203132858224,
 0.2164118872544547,0.7796463187853633,0.16638004449310284,0.8635208559527197,
 0.10656224346369503,0.8631055085900302,0.12945917633856918,
 0.3605539167872289,0.007338853733784956,0.9594672421142125,
 0.4827707816351223,0.8348858944404013,0.7843643388812336,0.3071490137884181,
 0.029315007070597088,0.5766125280944564,0.4905175399014302,
 0.5320759893338496,0.18137943308487403,0.9783921136431595,0.6570193794621987,
 0.025997441064783744,0.14677622076999086,0.48574574208522014,
 0.3880345745694753,0.9305897460869014,0.6608636212145313,0.8558671648136947,
 0.802977864974558,0.8427904404939187,0.5549004971610723,0.9035008863971917,
 0.9518647347610247,0.48982312653803517,0.9381335336050677,0.537144731072848,
 0.29355300045498933,0.6050133191120723,0.0981250306356366,0.5022165467350175,
 0.9571350334243551,0.06941503184778053,0.6650510842825258,
 0.14327726571798927,0.6772683779537014,0.4014845905597474,
 0.05902095851412936,0.4757879959027811,0.20555513167027872,0.107840087279647,
 0.951735532237396,0.5717407522027137,0.8918229284440827,0.054252433836781355,
 0.7357684429568576,0.4592965293697969,0.3733604092200984,0.8576280685131789,
 0.10255749624559884,0.9986921470143467,0.564375834985221,0.33017375999234966,
 0.729758971897058,0.6723963757182609,0.5985274065099857,0.5641040947175544,
 0.4577116374927328,0.7074932788047098,0.8719715395970133,0.9705668599066702,
 0.23042313984291707,0.6397185997060909,0.5722081208815976,0.6260765142372755,
 0.4107287050884847,0.4553823125357186,0.5386868334680504,0.8695233312220904,
 0.5124007056687432,0.9615079517857241,0.6575185720790687,0.32254368684788426,
 0.13160369638893066,0.02178164225150736,0.6587232466362628,
 0.9130461338272373,0.8370788976011129,0.7000263964908153,0.516628349506486,
 0.4888021639516006,0.7737486801369139,0.053503395089711914,
 0.6343560980667561,0.8845136903739979,0.7561819729677948,0.63113277557763,
 0.740411840034333,0.7658485594858832,0.9220921275002063,0.54901292964928,
 0.5064428456337279,0.14408978755642143,0.24472071728989409,
 0.8660221371659405,0.09676858531969634,0.6317370885339402,0.9205086173906187,
 0.4634845067541663,0.37828799247858225,0.9879208350702905,0.9490095912712073,
 0.5107441388433471,0.7785071987730863,0.21795225875049973,0.991808964533679,
 0.5775068695379748,0.906459230002751,0.4128285637036324,0.8863657706675543,
 0.8575657265435435,0.8240882086271054,0.4580027002281666,0.8357257724300283,
 0.46008548724675513,0.9016346887894174,0.011400187176873078,
 0.7069623098046052,0.15809788186785856,0.5763582174605124,
 0.44209840260937916,0.849282550365902,0.6785948509794504,0.16952583751176087,
 0.5387004395726191,0.599294544699361,0.6547105667821352,0.36869720184056076,
 0.6435785761678842,0.07212313589102937,0.07693269946587289,
 0.8781483445965091,0.44975913284672764,0.5416034009584968,
 0.15681820588175355,0.7593703139354117,0.9181924943761469,0.4747942427490809,
 0.30007955359363747,0.6322944894929347,0.43979472678060616,
 0.41249494441340984,0.9732864622530376,0.5319317429886259,0.3847705750735103,
 0.881107070304463,0.3368842125488162,0.266550000060797,0.9147259594825871,
 0.23080363660441006,0.093549060355539,0.17517554796198997,0.6418317220964889,
 0.39237996993244073,0.007627588506652461,0.42955353980698574,
 0.22689594514389966,0.17050475433125412,0.9023695957312838,
 0.6779253551454684,0.32681817356537146,0.7601457832382673,0.6521808093373644,
 0.176583600215919,0.7490243785218592,0.05642713618181605,0.8126582942496041,
 0.32607809765506446,0.21632999104627737,0.8090289988286052,
 0.29222675983233615,0.7346687449282958,0.5035052469646851,0.2680197711801088,
 0.21616071785890345,0.7868502581256085,0.28034788407019784,0.682792219247729,
 0.9973592455798286,0.5155614528374436,0.301414646135269,0.07727279157651779,
 0.8076943524189698,0.6458440918918705,0.4836751793038031,0.38559485068126087,
 0.0684765406565675,0.5079663758419809,0.11940487814779555,0.9222597351494362,
 0.5622594720027756,0.793669807150795,0.36843561779899603,0.0949188348014699,
 0.2481093463205507,0.40028604029112913,0.3358172314012333,0.9237993019339126,
 0.04776090776764619,0.07259055040505491,0.804700159485878,
 0.11390582377354286,0.20866989195592567,0.6514846335401332,
 0.5422253096419051,0.5994537913924004,0.34690846990192037,0.1521833114533676,
 0.36778710907790035,0.3474922689181269,0.5008214013116483,0.53540066424466,
 0.9304589126903143,0.45641048756626756,0.7708167916611333,0.6989274749368484,
 0.43773971252268107,0.12294751270012494,0.2631364978563293,
 0.033834635404834223,0.6731289178816131,0.4925728102972301,
 0.7041794822044414,0.14743848460011844,0.08850837742687967,
 0.8915598495228847,0.080967935434749,0.22239670451493532,
 0.017858213880246776,0.3135973501131373,0.6789710115000035,
 0.21792296507831665,0.21661891949579448,0.1825477496446688,
 0.21000306364622934,0.21310979228601123,0.686116014931636,
 0.022662990005300365,0.12816428370425392,0.46086386567692494,
 0.14974813941002518,0.8120513522606053,0.19136092425168294,
 0.8120029772077911,0.44820721756717175,0.8339138735637946,0.1424658901333764,
 0.36587340823175485,0.027615234393270294,0.7620054958697733,
 0.17285097331961952,0.33194576375841645,0.698520785017128,
 0.06188402686664041,0.45975494066329703,0.9140670525786049,
 0.07646267403379381,0.13995314645430834,0.6950981046397179,
 0.8079555487069813,0.8402721242822999,0.28446546101094783,0.2947065791831658,
 0.31393084150659656,0.8895187328149969,0.854092079815134,0.05478023267326304,
 0.5298345994512278,0.822476771362525,0.5348250534993981,0.22428265826088511,
 0.22895626848724093,0.661802887355918,0.08613893410505924,0.5269267478891312,
 0.22942472800981484,0.3088572425780328,0.8987969997188346,0.6594247069141324,
 0.35369215496810746,0.12238818312523136,0.25115782444473456,
 0.6920191400319725,0.10473808189191214,0.05439423277447941,
 0.9848700685136921,0.013081864208296712,0.32583241453843104,
 0.14451478469875012,0.4175644044971816,0.8232026958700737,
 0.08157719893276727,0.9636539194244357,0.1536618487661019,0.8005532282587656,
 0.7289815132746426,0.5781090037329726,0.13978425235938174,0.891754354484698,
 0.11266249176040444,0.724374136709738,0.9830816252669695,0.6636692695547964,
 0.5793183758438986,0.6407595053496383,0.0043655309033862055,
 0.1457816277920978,0.08370204183164742,0.4519725449715922,
 0.038201082459918156,0.14940806229250525,0.5858159386608316,
 0.9618256683927371,0.31921825253518543,0.7785919806136525,0.5770343627370397,
 0.8937498464695232,0.9625117713592117,0.6581994913610654,0.09582537369806055,
 0.6730331348001161,0.27400643639921296,0.6932083044337253,
 0.25091169296103855,0.9653976921170841,0.08732218852042428,
 0.07245364306498114,0.9858698726849637,0.5548284981075791,
 0.012946563431941227,0.9850327274503561,0.2725621467078019,
 0.8447898229752426,0.19562698941417045,0.2724944760484217,0.7216626762587763,
 0.9873403113434029,0.5109932480322448,0.6197070984884222,0.4723533651091636,
 0.3065275293328772,0.4263476571359005,0.6586252840105737,0.32686425626460136,
 0.5729631489020092,0.5699850576278195,0.44392556288712903,0.5879882508911406,
 0.5207527844782005,0.16081036600609333,0.9859692119333983,0.7816379135977447,
 0.7952679492888032,0.04975451230730599,0.9663731683720674,
 0.013009004887973408,0.5950787926236414,0.7362354558099041,
 0.12676308348844478,0.845583315257102,0.05288231980260483,0.6154148227492456,
 0.13741156874079186,0.06328353563166567,0.004733145718766263,
 0.9819002173234265,0.6514401233286923,0.1312350907157438,0.932498621567986,
 0.6540369420497008,0.7883814459572955,0.06382365679152846,
 0.48914716976656436,0.7648871086719434,0.19722591138174894,
 0.8220592441454878,0.27423409742033966,0.5496086219298201,0.8387382268774931,
 0.7232858258549394,0.41251443472157157,0.8184589013761495,
 0.08867744729149629,0.30495823731418603,0.43298713084475693,
 0.2196728970041283,0.2863291107271365,0.9540636814593744,0.436163264056055,
 0.6902000356336151,0.8509236199310644,0.8180053735901845,0.39961919064213713,
 0.5621540742803691,0.710251106876869,0.8103036289831538,0.11591601763117576,
 0.03787375283001304,0.6481550506114827,0.28328127920079516,0.781577654274126,
 0.3196273304910986,0.7440971845935083,0.7456940636598215,0.0632820740349409,
 0.5908190953895269,0.9099799749790601,0.893024658690491,0.7829261659681308,
 0.17585307869942735,0.31865607978812127,0.027110247355642603,
 0.3535284793700685,0.942676620418528,0.17814485907220756,0.6917623303924524,
 0.2900663926863687,0.9433630516525857,0.5156176585054706,0.5399807918378968,
 0.7154375608712459,0.5513722679181007,0.19388143489460796,
 0.33312253543590864,0.07943623125830956,0.02298358716223403,
 0.18449808134160262,0.2966090449741339,0.9002290241174825,
 0.02725863939544304,0.6738972345918667,0.5833167393855102,
 0.29208743367989354,0.20184729809203072,0.8304254743263277,
 0.8246745435558288,0.32000351740480926,0.564357871783648,0.27777030777120904,
 0.32367921191271576,0.9671497926967393,0.6161162551023781,0.7875103405793544,
 0.5164726552433503,0.10469302157603,0.008971251565821459,0.7825211645098178,
 0.11857559903216108,0.19088078445341172,0.5692711386307926,
 0.9421051857545147,0.41389355836997854,0.1549972221036855,0.6780126129795421,
 0.22882099513597254,0.5300355175961554,0.3506282821975101,0.4259061056771758,
 0.8620856412505778,0.9699485026681034,0.1305136147789676,0.8282948118957597,
 0.5602192781729637,0.9271830957995572,0.03389679191928574,
 0.02760626276802025,0.7903142451708558,0.7397042305274506,0.7524188765704719,
 0.38749791650224186,0.5755448795278486,0.10676908280225261,
 0.28208810495879566,0.44849379565761294,0.054238203052725864,
 0.23728644765024642,0.07188860541110353,0.042975277913486765,
 0.878597401289685,0.8263266475488394,0.4192311817459742,0.8017895712247747,
 0.31300565472858544,0.04010418282960859,0.647057851467662,0.6107684633915118,
 0.7265728688484395,0.13841555911788328,0.3138368978064223,0.3174854549927615,
 0.44396992278241665,0.15041262740427008,0.42952544358113653,
 0.9987382526645883,0.12874198230354728,0.23846153933871428,
 0.9435738361426881,0.3605897369798101,0.7388575966297948,0.5870818622865163,
 0.31903705516787295,0.49879460230271633,0.2196386644028131,
 0.7938157264761804,0.30748584004662327,0.5112713227568996,0.8794924549326153,
 0.9375383573826384,0.5968062870320368,0.21561950709882538,0.9370653142318149,
 0.8302267346538171,0.7085501203825002,0.4809488421869883,0.28935875571662795,
 0.8988559121385122,0.6604310577546988,0.19414028754863,0.3391720051457915,
 0.003893400306761663,0.5125280192857381,0.8749750604876649,
 0.5247473211284446,0.6999976710748006,0.47575500470067755,
 0.22905855173821799,0.26759430398314843,0.9807578121655803,
 0.7772877230140927,0.424491719829746,0.5428635257780934,0.6414804844797749,
 0.6353717373700698,0.28673073400583526,0.24556032115936965,
 0.41390146999910704,0.9223376556658239,0.7651358482275473,
 0.37121855726706976,0.498510480683104,0.4225936122334396,0.5067895839276635,
 0.3633149139113818,0.8385922435152842,0.972000446810252,0.8020064122451349,
 0.41376618164906054,0.2599292482330632,0.19645040943086822,
 0.08405809862545899,0.06301821356950255,0.569207275121194,0.9246156654097396,
 0.7495327798862113,0.05451155804951813,0.15534271737527372,
 0.2902456057705419,0.8339905009261844,0.9590551275836637,0.7149836196369171,
 0.820763819843718,0.7907508591991458,0.23171968177908053,0.5401810846498568,
 0.18505762794871505,0.5124369313715771,0.24579610576158872,
 0.4877670092269384,0.3179971020886001,0.6927916930303372,0.7071924584286017,
 0.5595511660037127,0.5964365473284844,0.11425231793918345,0.3071638858996071,
 0.7135366631265354,0.6629895911750859,0.4004956056325386,
 0.033866373626095236,0.5626813633161134,0.1227565556848984,
 0.19516687499539442,0.9560307562722424,0.24300396002544034,
 0.038369425087918296,0.9481978810831052,0.42142035362549013,
 0.3435610852021703,0.698914208633596,0.3477853605887571,0.5457372340742849,
 0.8474749752627764,0.44852921178920113,0.6801858038238734,0.8674763613490232,
 0.8487740909445048,0.5854106171011012,0.07921529841440722,0.5779677651641433,
 0.7418489700102061,0.16049657429202902,0.028940558492740354,
 0.0792670804465615,0.8628799567519069,0.462349819498151,0.7754894239685168,
 0.8220610203003463,0.6282857870099035,0.7766019522642736,0.9041905726929634,
 0.8876842069635973,0.4357398940386701,0.22691487958247714,0.7735331281032592,
 0.13560829569085253,0.7462086737021418,0.4286206154005905,0.2491271312927419,
 0.9501656344397403,0.5255368476421395,0.5595670258959236,0.06651340393200322,
 0.35314577797895264,0.7661353090810701,0.33258676718033187,
 0.35685212871581506,0.2504777701760872,0.6444036200636436,0.1217721212280991,
 0.9598025598048225,0.3106226310710939,0.22714238778344098,0.5581269950884034,
 0.535842784644215,0.4913631354253031,0.7763672788406399,0.6504919087865503,
 0.3902555677529558,0.7991756206551177,0.12557497669714113,
 0.17159803606437163,0.6402899661780141,0.4438816478699925,0.7023796568480637,
 0.6961610886151124,0.6789290654547843,0.0811479678506285,0.2648638749727731,
 0.06252252622647836,0.7085038182806427,0.2706106428404176,0.9174889333611923,
 0.9916278717392161,0.5783915969904633,0.5768773374987989,0.3749680890468128,
 0.47683135386379627,0.998578366856459,0.5796943087884162,0.06299217904953391,
 0.640259953796128,0.12455844039414565,0.6896854439121398,0.1996310995232502,
 0.15780794465849557,0.8456501926129014,0.07298826237473888,0.628771440020799,
 0.7891298248820419,0.3248169891503847,0.3223216411519967,0.8671473046869519,
 0.01871683410141234,0.7788564041192407,0.30044386611310203,
 0.22973836235142286,0.2255717015821722,0.32714913162717185,
 0.7671829766234154,0.7092826765011826,0.4253681532239213,0.3671351120252345,
 0.7961079145590104,0.9609281558881031,0.22421468434276248,
 0.26578937759684074,0.3098110822013397,0.052706108902757065,
 0.13372750625741414,0.5396187042470693,0.984852415426008,0.35308269666382486,
 0.6021709082731961,0.9665308520836857,0.6378257879514124,0.38294393839019025,
 0.2267544527716282,0.30529284152517167,0.8871818781029547,0.6766190064474379,
 0.7256225378579091,0.7688860330943948,0.7036317573982175,0.27749629523561836,
 0.4890167931926336,0.39469166925948285,0.09456281954887136,
 0.4991340527897029,0.09462225479140907,0.7610250773453308,
 0.07898541895932387,0.6863115108482545,0.953477359973646,0.6547275445499382,
 0.32803871072991597,0.12757018537537468,0.07247152981262073,
 0.0031165866473622517,0.08347738976772068,0.5113700242046237,
 0.7141780067651037,0.8757573894196676,0.8493670144592519,0.11543351000076918,
 0.5965088812730697,0.6128107199762008,0.567396236375323,0.38894295475515217,
 0.5244249602775991,0.5297137486516776,0.00835730793974554,0.9920307916955792,
 0.5882927211938151,0.9536752405064909,0.43130108177087023,
 0.10608119326131638,0.3472159918977734,0.48278732849992556,0.780897006120905,
 0.38290630296023564,0.27617088112136823,0.6104370717348523,0.611174306392318,
 0.8817598821526298,0.6314060944819102,0.42174236527242437,0.8884953190835185,
 0.5094960977447849,0.2980505031989422,0.7797869990042332,0.8405596305044417,
 0.6503255594734465,0.6188117787562664,0.3129633153814737,0.5136806657774523,
 0.8693060367205103,0.584139674098885,0.7685458681133007,0.5474181107573484,
 0.36385241884790076,0.46390754242033705,0.5193457940613732,
 0.3978504886308555,0.6241103491760678,0.7646868941928432,0.9714526800732916,
 0.23082674609011788,0.4607194889886803,0.36902624282685725,
 0.7391610113454622,0.7005001689595783,0.9716385209046444,0.06644234794165937,
 0.7205844290530458,0.34784860038841126,0.31135535982280915,
 0.5750181715452316,0.12813325765164973,0.7809399160636402,0.8989424023605139,
 0.39690127778829487,0.21836235395254566,0.880149054151057,0.7602100851261272,
 0.2596003701365607,0.8133560363707806,0.48911845748716254,0.4751128356782006,
 0.836576201564431,0.14288206960982863,0.9824161345000086,0.5154662836671362,
 0.8248162071395717,0.575066081545768,0.38985285575356854,0.02332911533683757,
 0.4086511462918053,0.21222934465816012,0.3735531010474715,0.7202068967208597,
 0.34704918943090035,0.5571696418842135,0.7496828158843338,
 0.21555019417653754,0.1387040730893666,0.687876673740133,0.7737846289235655,
 0.7543766388210539,0.9372272763975006,0.846314648785023,0.013383673044750077,
 0.39183199589013773,0.5865762874551489,0.30103402095117515,
 0.11325239836673395,0.9479057823625183,0.9211068956031372,
 0.19379117904562693,0.3830046729872638,0.27578783054194644,0.493829862115275,
 0.13125262998040255,0.4389664752699751,0.08077691396581699,
 0.8794458364133803,0.333140973645189,0.7205790329162062,0.7172378804125498,
 0.6199899695877079,0.03380679988000701,0.870721278032516,0.11433958898040142,
 0.9362368464906081,0.3045089252737119,0.42225687975026516,
 0.039355139772682124,0.6336217250856644,0.17494853582806347,
 0.8977019227754636,0.21959506173272725,0.16644997446250598,
 0.9373542952938307,0.8652223955720852,0.8875741696816191,0.9929573206889611,
 0.1945614795187034,0.7453691449442889,0.32001049611102306,0.7146681378565982,
 0.8414127556348562,0.7554849051031416,0.824312479058035,0.24109842484273614,
 0.15116943502508284,0.27524829352699276,0.4257170944606049,
 0.5761874443261689,0.39529236592963246,0.671979359183264,0.16800971152455257,
 0.3803918555556013,0.7356880530253378,0.46977263602710306,0.7470783887305689,
 0.06403534805690225,0.6500888917481819,0.5852051333820758,0.5859819222513486,
 0.15649098654289018,0.06762149673908346,0.66338339164513,0.8606260575625094,
 0.9939464624292952,0.29410404916248867,0.7254931233341612,
 0.32331691084904113,0.744588787860335,0.5638156238408355,0.5107027998424867,
 0.46771320298735275,0.21179459056716987,0.5038915048417548,
 0.18217080483451564,0.5811719685471903,0.8979054652656713,
 0.024642618266979555,0.658672657117272,0.4683162192196705,
 0.32202232134409914,0.8977412648575089,0.7244806990536738,
 0.027062476391123025,0.990264900304161,0.8454071246493197,0.4006035867400902,
 0.8506809213033236,0.36348341281388186,0.8351875408001748,0.3947585713800238,
 0.3894049375402595,0.21622370177731998,0.5631714251193612,0.5490974657121588,
 0.5284717837779652,0.4502714074549705,0.8902767864574288,0.5634350667197014,
 0.1517130036423634,0.6702609272307565,0.2196539572985715,0.07719553814219537,
 0.6008376454225425,0.7446560562189927,0.2323156039278198,0.7141996638892295,
 0.8177072518355245,0.7485937001205868,0.9538445245746006,0.7286741392089136,
 0.3204151996704093,0.984800864808659,0.7886040637874312,0.7131187309480853,
 0.22625081390947788,0.08670788496292992,0.6038820739265081,
 0.9510886154510807,0.8004928923638261,0.7230923274649339,0.09251868155803145,
 0.32946341642544497,0.37023646520464537,0.4323940233858308,
 0.7308909754919218,0.22333909486299253,0.4666446783207222,0.5833223144141435,
 0.30805949560398704,0.8430231581394976,0.3962530662035555,0.6062022071217891,
 0.7701734514688384,0.6024302700372376,0.25469777126472515,0.9122387847724863,
 0.15268364296878834,0.44122678634137624,0.669508566265031,0.8769326220063415,
 0.9854569849225465,0.3201804152670029,0.6193244882695094,0.9536558358398346,
 0.8705715467701122,0.1354224072492226,0.79855027675874,0.023340104150993857,
 0.37905242730878186,0.9410020909340875,0.7843060904353794,0.8408968503585136,
 0.2864623453934203,0.5950181151032183,0.1010733578665669,0.2977740629110366,
 0.5439289324750638,0.6999832346306468,0.10610006033714747,
 0.24756011265170874,0.1629064575947583,0.03206045819644776,
 0.4904902289241949,0.3634922802771511,0.8925280935292509,0.46742610554616715,
 0.17947316344314612,0.20555173224013135,0.019485152977995135,
 0.9993566008371773,0.9368524081751223,0.8953708516075796,0.19171595599237068,
 0.7186321043052915,0.8689462519274939,0.2731316833094024,0.23087284665738017,
 0.8826741464199365,0.5980648080342217,0.23285515858076034,0.4742842842927678,
 0.2499354005543416,0.7099464652553361,0.8963336191969148,0.46328525546782795,
 0.5707003465079441,0.5579095229618435,0.8066670387106616,0.5833744787984886,
 0.7670370085796026,0.6982207574218433,0.9041919513789591,0.916832481555168,
 0.6336679391019622,0.586498858644362,0.8464507335249378,0.25186752548200353,
 0.5682563604578036,0.30475536897881317,0.4303783079924093,
 0.04931527569479921,0.9458121825447189,0.86057083909744,0.955195658890483,
 0.515160722072543,0.029553398954993204,0.9184810448376276,0.8233922250053872,
 0.39553906894941526,0.10439066796651342,0.8127395331410027,0.60210116001066,
 0.6385173685309535,0.3270023059661019,0.4363788530110452,0.26377012475539385,
 0.18645413997943983,0.0953262907930732,0.2024375735833499,0.9072345196442568,
 0.5362502312327644,0.3292650368691208,0.8709461865341327,0.49863487900079784,
 0.7630191417399906,0.01596034203050367,0.22414836922890735,
 0.6044420164511611,0.9058973998621557,0.8228740976935426,0.9843506246073179,
 0.42032583653145406,0.6862134039502934,0.8391573043885306,
 0.06267274547690282,0.020438227286193378,0.18979192841896697,
 0.2742529355767731,0.7751950384472319,0.6503487479412505,0.15997695827393466,
 0.4106291853897779,0.5490863193538025,0.6219893847699833,0.290294269423655,
 0.3764018179285722,0.859768355997393,0.86830273560475,0.9971261009627662,
 0.6386227705555173,0.07091344130526389,0.0022337185084974065,
 0.17580517106465032,0.5876854389551929,0.7550820516007501,
 0.23968504487939857,0.857046945956799,0.32141555972686364,0.3410419482921393,
 0.5790435449171373,0.26763356872522637,0.4910047026676624,0.8486353567915105,
 0.5354882284489831,0.6828803799751744,0.17932090147574986,
 0.44760359388140913,0.6958711903231944,0.1784964753133641,0.3332219601184053,
 0.7843302515630879,0.27483457106898035,0.46145089743727374,
 0.9827138377840746,0.9986286656266694,0.1918159895758298,0.44685164236763486,
 0.2099438312029771,0.3056450308060765,0.6150785797291808,0.21920005329065928,
 0.6849888262455357,0.6756927251844258,0.5093348402991921,0.9815240471986196,
 0.17308497850777615,0.12135554304636798,0.1386933702233919,
 0.3115320326713691,0.8408623947649051,0.24077340462825136,
 0.31522462637649307,0.9211429226076696,0.5443669930026809,0.3590801019125123,
 0.3003881429688935,0.6526651375017962,0.631800793580307,0.381600776650053,
 0.0896546224738114,0.1078241165877526,0.4636800354790527,0.27584747626793404,
 0.8289976942337115,0.46811039050250125,0.417939422127103,0.8194960687436382,
 0.13556210292147963,0.5342580673289299,0.24673054895759516,
 0.9577816395247234,0.9938065543552556,0.009478918603604058,
 0.5042295965760838,0.006666545925474221,0.6050193043447325,
 0.6207842076884271,0.5682691359497625,0.7872313436010557,0.3954965807104551,
 0.6776423316750294,0.5679865829662691,0.7703687874471675,0.20085998904511637,
 0.3916321335708879,0.9633432956342373,0.04418991758788493,0.7678648468414137,
 0.927277543274792,0.5817555735591791,0.05541493548976495,0.510259524215588,
 0.9313835581777611,0.934246873334,0.6292123397249574,0.6008107055510982,
 0.8794274380809173,0.4561066210936049,0.339650195213955,0.017475570284510855,
 0.04701302213195202,0.43853066942248986,0.08030685894021594,
 0.19977006879594328,0.3864402471257492,0.7634168542850823,0.9459606385436363,
 0.031262837999185056,0.6672525334743957,0.08916437845049008,
 0.6911297638967538,0.4322834712016389,0.7486631713717101,0.4447486092018029,
 0.002066699247974979,0.32142141823154036,0.03762335013496709,
 0.8978119522689639,0.8825520426305805,0.8630599193627901,0.6188009932169565,
 0.7761565292191185,0.4893138886538918,0.8614011255292341,0.08687021768413683,
 0.07363015868584177,0.8359694163308318,0.356194718360196,0.9084748097044621,
 0.7411197656119217,0.07380321914200128,0.9785413143801989,
 0.14678819418089128,0.9206303773081945,0.18474536053918822,
 0.9717579390688191,0.43427373655791357,0.4422720310263092,0.8648395838577589,
 0.3357199727358712,0.3062474623216803,0.14687675527271227,
 0.19359699919006934,0.09579525939768974,0.08837868443684416,
 0.12120585848237786,0.7723145853933764,0.9841403255748268,
 0.004835382019857337,0.03936882842181,0.3083783805279867,0.9518437582822663,
 0.06586739697090571,0.8577838814559902,0.3941388006195289,0.9156760643302484,
 0.18103940378032957,0.637415827696373,0.26212235126008165,0.6408140829907989,
 0.6759090480196057,0.9301512616918505,0.9588363945515204,0.41313271685966146,
 0.6085154423155992,0.744773740006241,0.675168865253123,0.07543554784190776,
 0.6753922074182064,0.16541415160746686,0.43884765326768815,
 0.2758826937213149,0.4430918265692736,0.27288448164328816,
 0.018529527806209423,0.2510010953259626,0.514013501417218,0.5218718972266755,
 0.5579427571006702,0.9243831081273991,0.6452337601304398,0.37303804733115586,
 0.9300291961137981,0.11075800874384345,0.021653458813645776,
 0.49665597599700384,0.8804699030938105,0.9800805528714405,0.8822355463815952,
 0.37841746436856083,0.7948517481420038,0.3302880675876412,
 0.28741903918843825,0.48750691318319117,0.12113268250864073,
 0.5374057288103369,0.3315040752038424,0.5240585592520725,0.5272981231779109,
 0.14582710278912447,0.7892031566188205,0.10462230684695284,
 0.26332229471420543,0.6820784443133618,0.2650771549983324,0.7454828696800063,
 0.041856476075269655,0.2709687832931803,0.4578966426041574,
 0.48025163025530493,0.036194180992647595,0.9016301080248793,
 0.3439115679273935,0.5636991838496238,0.011422805717557605,
 0.4999051178592765,0.5753872398225853,0.921191176568254,0.6779483043700416,
 0.7019877981819127,0.7499592571934707,0.48906102126107576,0.8296913382915113,
 0.30097271438088424,0.08848516340859613,0.5587290245328228,
 0.7021626880417078,0.5540626341310613,0.736946121013048,0.9940842278855174,
 0.6965830129383425,0.47009448933261966,0.5697260230561381,0.9177113920788834,
 0.8000569320343768,0.5869044170375678,0.267293517871254,0.8954045472292728,
 0.6668055083852463,0.16245504710328207,0.9594005110917382,0.9235273597850882,
 0.9363954738680156,0.9334923726578505,0.20656578806152393,0.7115328763539371,
 0.13752205341361112,0.6117084023202808,0.6876501790229362,
 0.42223700838161426,0.7893113472377847,0.3230973123253724,0.563314460393137,
 0.6483512515224577,0.8603013908968447,0.6094965705753892,0.7568837524123283,
 0.6286328798916888,0.3728975503361762,0.9132044264870889,0.5470609427963581,
 0.8635851983850233,0.891304099564978,0.3116315205180519,0.9723860257981279,
 0.20981953178528379,0.44209602010155524,0.5484284095407822,
 0.32415253734033966,0.09525229052910977,0.53058966322568,0.8226885071799505,
 5.619551734723505e-4,0.38972876484880725,0.9313880282485837,
 0.43712692945860554,0.2687557036496593,0.3789104051447444,0.6829125401928191,
 0.47397080318044393,0.7159431607988059,0.9857304508038264,0.4540784008422031,
 0.31301046717344383,0.9493998113077101,0.30253421202587827,
 0.7729950990456523,0.9413150659292248,0.8998183113093671,0.7026105769273201,
 0.880293990812321,0.5838854369381126,0.5766062877569745,0.018135850436956313,
 0.23650626825853394,0.2528421671939166,0.22067216259499978,
 0.9198318453164629,0.40489824258624174,0.3382069807254281,
 0.15025134618589453,0.18706649039608425,0.5271450113838173,
 0.43728715792242046,0.6328284070154572,0.7782291692815525,0.4261954158978546,
 0.4412145135799854,0.45354777410358127,0.12218490772478763,
 0.7598443324915283,0.5851049242022567,0.6309029835598137,0.6555936436359782,
 0.5497146688806959,0.14057208807562305,0.013093578841535392,
 0.9797288775311863,0.591521715925647,0.750639818464755,0.6844527713800013,
 0.8004107300955017,0.25559487310816154,0.8013090777444418,0.6861138599516393,
 0.607650222139509,0.6509275034097279,0.6419222277988469,0.49220083679425386,
 0.456868572605137,0.3438415658627324,0.8878009485541156,0.026087562240049045,
 0.8626065105833606,0.6888640066634665,0.7747222346460909,0.8869273841139895,
 0.8237887293919175,0.6712339630277255,0.5114426358728181,0.1311160306529231,
 0.9114545689723103,0.7897927708835717,0.8506646172348861,0.7308251668884164,
 0.57481536180227,0.654938392695575,0.9643830082859836,0.01443583382426672,
 0.1741278942656711,0.7988326539335993,0.49216956215092944,0.5602569715070493,
 0.19530604550385888,0.4945504248995938,0.08817241284650423,
 0.07280014634923793,0.7978756717983367,0.06059672294945373,
 0.2442384064076154,0.8036553000122362,0.43451597229569905,0.6951794688807191,
 3.446058019677123e-4,0.7836580752626973,0.7832926765232892,
 0.5590001276854414,0.5136668212954985,0.7885631957596693,0.4857994636264362,
 0.9969283051737097,0.1335042469642197,0.3838158043402471,0.2728461105160268,
 0.7561588304274743,0.032184109708342845,0.03937369474868868,
 0.6052909354808022,0.6028319895073653,0.39473959163383254,
 0.059024445422559246,0.745592347330715,0.3240885188298499,0.2731535334300179,
 0.7958100316082701,0.8751437223995169,0.1114475271466413,0.2868934773748635,
 0.8316232317930479,0.4407262035125783,0.5818139671335674,0.47079709376906886,
 0.4528797769183156,0.6267090532370071,0.8286001068918929,0.39958696174164776,
 0.948146941104442,0.8425883039741088,0.18947674173744788,0.19910715181748895,
 0.7943509130127502,0.032962969168941925,0.6185065051197878,
 0.3885268827064703,0.2541847486051183,0.6805787562378393,0.1871114470338091,
 0.6789848027729541,0.8329868300969656,0.43382308853246865,0.5926906427555906,
 0.6019730978614728,0.4055304281084412,0.04054018412430249,
 0.48575858143541273,0.876654680169188,0.9476208378317674,0.3495432506384748,
 0.6091928546523613,0.3972397017630658,0.8735986193285095,0.08463716077577843,
 0.6773074962047613,0.5559526417029643,0.46584789183159403,0.7255261009806967,
 0.045338342502250994,0.3981235274196746,0.19168516166318206,
 0.6713322800546029,0.9073797170276319,0.8238845982185846,0.18157341541080463,
 0.18707891887348294,0.739018766189099,0.5038118750959744,0.06998310483965675,
 0.6227615932152955,0.9734606880616691,0.2171380065673557,0.5623441149138693,
 0.2984829809995315,0.04107604988334246,0.48981307476326597,
 0.7109673999179302,0.10948317758090687,0.14928954550963813,
 0.6007914010818032,0.6320421185839039,0.9904960392825812,0.7774156320210039,
 0.6274973381649456,0.7674931077662923,0.5741471891848032,0.4231159984567211,
 0.22861539013492815,0.35767913996301226,0.6111706737942311,
 0.7094665132362707,0.8029280454273948,0.6714130775508602,0.4117308408335151,
 0.8111431917404379,0.5913874931798218,0.31911782349335827,0.6122152519079035,
 0.9259974013597959,0.7890815061996923,0.019733641060094897,
 0.2989507587096367,0.772590635777189,0.5519420947116984,0.874909311646821,
 0.6200160740531607,0.5030796931639954,0.38744333881751414,
 0.13099152981902218,0.9935355062757706,0.9832652105243767,0.8021747002791358,
 0.35543881820367074,0.7069484559214124,0.16572145082437428,
 0.6364295791540595,0.25776107865829634,0.29261963240540256,
 0.5611614064741405,0.17671843085430539,0.3365645075080982,0.1864641424028679,
 0.8681117115943023,0.10656649288764952,0.5512282108527734,0.8192192576437111,
 0.8612216522779925,0.416396323252502,0.11444258652580253,0.16783494251183084,
 0.555259664464797,0.17508099412744527,0.2213373374127965,0.7476590171045889,
 0.41703936557442656,0.38426425748304605,0.5772977734237846,
 0.9247375757052916,0.27665118111886455,0.4183359271667513,0.8315595716750147,
 0.8977076969617888,0.48695685810302936,0.877436874682827,0.09294279240232739,
 0.9533066182200919,0.7371292745891876,0.4830605683353295,0.5361613489405358,
 0.5485909362555014,0.6258000826981878,0.9322517771811555,0.7938859345778062,
 0.4458143593559307,0.542371468545905,0.490440927544879,0.5948806081301082,
 0.18747781952629006,0.6657312985284675,0.28669655636933067,
 0.17959747476429977,0.575608899433619,0.1316298215903301,0.6612134343949571,
 0.7146968841028489,0.8656352519931371,0.06410429426315378,
 0.016632989338501325,0.8325476469931351,0.0656018722308438,
 0.48840456647621433,0.013124282965642498,0.5766452843406572,
 0.47380540208855193,0.38895063737243774,0.20035669637976228,
 0.3980736894681445,0.02699485748449637,0.2700266842447826,0.4698841998852228,
 0.2772962390611532,0.9569147079636853,0.19491296280462334,0.6515801121715862,
 0.7140636782771959,0.29724932157660133,0.3289308156864241,0.7221377977743719,
 0.26261495910805555,0.7943746380202871,0.16869616961568423,
 0.6349473557665031,0.34976074945433255,0.6453554680197031,0.429218979416787,
 0.6172683411398191,0.34430822403343264,0.87982696379525,0.30491036992188303,
 0.3839441820229097,0.07993095228213454,0.1877790192093597,0.8736549940579655,
 0.32809492770651527,0.06474530079218965,0.46036388623357294,
 0.737606677639546,0.33300599031378086,0.7528614112150651,0.6579332989617761,
 0.42127576377400566,0.5475581321455065,0.060268873328638395,
 0.020178745277354526,0.6412661226882972,0.8828253589569922,0.190457749338097,
 0.3917403232297477,0.17542706192864577,0.14521469947446175,
 0.4556827120798408,0.3755962698073272,0.5214015804402821,0.8266771384868239,
 0.9011212373275623,0.9233819140778475,0.2627125800496739,0.9401915872821839,
 0.9599164004732412,0.11459426690569918,0.047597050612592784,
 0.3808853233157663,0.7463918874214084,0.8532570962730037,0.5239177790290059,
 0.05369106551062841,0.23039458969952542,0.7536786757224998,
 0.7138980094355487,0.4637921842713323,0.07363063274005266,
 0.08297829865115247,0.06222751193779841,0.3097173654667965,
 0.11625939390778273,0.490558850226663,0.8258877025137195,0.6209955790045201,
 0.9808896093162724,0.8307819333882842,0.23919791933369017,0.3095697889835779,
 0.6199470999012625,0.7199116874281225,0.2693330509015166,0.5409442294438866,
 0.2950045894971812,0.5558837882154612,0.3502723013770318,0.426426562054128,
 0.8362298836514876,0.0655363209976555,0.6831667435248736,0.8873643687205661,
 0.4503415944627355,0.30939291032578364,0.17103093290475524,
 0.4093572181492824,0.22752205754247257,0.2603023142455472,0.7467154302999153,
 0.3631301730325297,0.8853646282062358,0.5054230821559048,0.8056834213703583,
 0.6346264156593386,0.964241685331493,0.23605510865234802,0.7825921873784821,
 0.23968768441765465,0.4703591643004592,0.5009873774792492,0.6675038170083368,
 0.6085126169047774,0.4933876696102304,0.9876450808858382,0.8814930763469296,
 0.30493839354951136,0.15467464808212938,0.7377425131082297,0.105306531508783,
 0.3770281579993473,0.027610860568632156,0.12113313102795553,
 0.365907287728955,0.07139327028369158,0.13198862334601869,0.8680599231018289,
 0.98931958394514,0.6371068839640847,0.509652464489888,0.07112775389227743,
 0.6357773036866867,0.903912267203606,0.749629773118151,0.325163592763824,
 0.863252465908058,0.3302233615931336,0.5721478530993788,0.42783275241772556,
 0.3391431431433364,0.031805913645643225,0.19777777763909443,
 0.6675705719644902,0.29802178180764893,0.6987930937168343,
 0.07307851977981328,0.7924695691499732,0.27787693478801323,
 0.7046713305800633,0.49429928842392834,0.09863076979081276,
 0.9690573843892658,0.4918137420950587,0.6414208985491043,0.9082059409959712,
 0.46445416382805405,0.7761734940087843,0.625759195751324,0.6043728949486841,
 0.24686034852356875,0.06529867617522056,0.17011238234552217,
 0.4882079449419332,0.8649784825909043,0.7329684345010091,0.4171657186712123,
 0.412150827978524,0.5682122920793014,0.3999339764846035,0.9464959857361044,
 0.5473422974067422,0.8601831072552092,0.5004384815280788,0.08415594075643174,
 0.7967488600826053,0.3926509624911976,0.9024485394720908,0.6031737630564181,
 0.30842696652565527,0.7987945561239072,0.85013562720137,0.07554516940096234,
 0.7580623008128701,0.34150770765870386,0.03038118627205555,
 0.09765254135424939,0.38336458394269934,0.9015391999961329,
 0.3882511361941289,0.368271548235252,0.5321359187280107,0.4207927395409571,
 0.6302811094183919,0.27335417753593505,0.7020745915726808,0.3113144346838421,
 0.9049073150820671,0.5483875507113846,0.10991524032858535,0.269184246237235,
 0.5372478555463547,0.9645861119328393,0.9498348502873243,0.13271649281650544,
 0.1950397619889871,0.8729838801767607,0.9386718395619426,0.41967570291647294,
 0.3999724566828263,0.024712046982942826,0.48527361290722115,
 0.7172343735394605,0.5817319359384064,0.8130606334695085,0.4827918788387985,
 0.7867642180712876,0.38886936454334764,0.2901450124422852,
 0.057111276853598736,0.8994504393043725,0.2002185741306819,
 0.1664695704303708,0.8218465067351975,0.8558693575978807,0.25033480815048315,
 0.011587158673298403,0.670675865485355,0.3114061522412842,0.2506854894929882,
 0.1362500938917658,0.16795560655058106,0.5352434699391889,
 0.06037100530913464,0.7274076648620582,0.5958109028466276,
 0.03669455039521041,0.48206654079672717,0.23571367820138667,
 0.2954487944673956,0.5922997231030336,0.32174205079527873,
 0.33247372602913905,0.3421260616912225,0.949560486569908,0.8695907784220782,
 0.0962267382575489,0.6918004961697245,0.8953882239250106,0.5897399024517149,
 0.9874247853351208,0.08217307005904018,0.9112952163892655,0.7948541083227084,
 0.141876617922588,0.1018915665616873,0.19242545362927155,0.9060477919985277,
 0.42736100738223104,0.6906761968974978,0.420787106691348,0.7750354818336526,
 0.9952319248603406,0.5974747990440376,0.18840827616810574,0.8943157102799807,
 0.6764520130794371,0.16999564093579178,0.13551859517529008,
 0.07394445325255472,0.9428619889730785,0.9243732761278746,
 0.30766441036294834,0.028282821135854008,0.2043671309113564,
 0.7622066314868743,0.9384943103515003,0.9804365201391578,0.44038468120408947,
 0.45873308007504,0.2641425292846229,0.4372441655686944,0.9594388964674675,
 0.8521280290850259,0.8804272316179885,0.5609637014409228,0.8673643724101293,
 0.03938061555533001,0.6700107960749888,0.9787091285244106,0.6216543771992005,
 0.9154074119244178,0.6876592295558906,0.5244688014635069,0.31339288598654325,
 0.5137148124447115,0.6815940062948623,0.9453326533740827,0.3002262882669041,
 0.053457692294335724,0.09211136896316441,0.6260905398494184,
 0.9703537055258292,0.38932440437416604,0.12287131769222948,
 0.4504988203601913,0.5989036032591535,0.14578893403590332,
 0.03459519640753217,0.386859687613869,0.37510988572857695,
 0.09680199416675528,0.16507605085646937,0.3135049007885641,
 0.010518040329304146,0.8940722846689219,0.8369465640706302,
 0.3740434353039338,0.7180016357027439,0.04732699017998321,0.6655351816828096,
 0.5455743708772309,0.5461538754868268,0.9361952342151167,0.7349290570158936,
 0.7118263298425197,0.9612074068762915,0.260928303334671,0.23787994955317038,
 0.29938811799012865,0.6653751363639953,0.8151618331620174,0.5190841182786524,
 0.36049944429895364,0.204682353167146,0.13446810190667646,0.8376940802023152,
 0.04481954066015914,0.9458103692616274,0.8332290529247312,0.5806096474628841,
 0.8630219765813765,0.18511865946077455,0.7998702338894436,0.7737897098451092,
 0.026464386058531075,0.2022979398689272,0.22734694197373284,
 0.6217331465318926,0.4218044561313804,0.9340357468631163,
 0.047271371819122776,0.27113574010046115,0.3363518623033128,
 0.2583415085373332,0.7564633845257176,0.7961923945459051,0.8914258848774508,
 0.7619137613217224,0.8320203464889241,0.8499157366231009,0.2883833014668514,
 0.4273610980004302,0.6179077212475586,0.22424617858125984,0.9698317098420768,
 0.1744033754818084,0.28326500440002556,0.7241977026660682,
 0.40315783246988113,0.3348068870544316,0.25502121686549617,
 0.6292172464323583,0.3099569673194482,0.4927112827228095,0.46843246393448557,
 0.4060954127590204,0.818783903847942,0.041531898240587894,0.6625638202264438,
 0.1621199502319638,0.6312099488444204,0.6550393965756932,0.34242330245307817,
 0.4968606781823842,0.1363253415848723,0.09052037204411445,0.2622119860633533,
 0.1732382785104244,0.9129362513078771,0.15327033646341715,
 0.10479435907420664,0.5844293019744881,0.09401136524790532,
 0.4006232883413917,0.12262615304272906,0.8114931268558704,
 0.026975551235734585,0.007410544335482716,0.06092045509652466,
 0.06015551820370624,0.8106740687582539,0.48838341336860414,
 0.7107799799644479,0.9126487200794082,0.7628765213869925,0.43341280448385167,
 0.5778325055926861,0.6646577907448487,0.5034294383259156,0.7402059604546257,
 0.9514647570593323,0.5399120959628162,0.36869437439864194,
 0.22643431480022225,0.6473987554562664,0.6968404631519218,
 0.025816430596818263,0.3375011547228691,0.6238295569411935,
 0.10071820332046144,0.24200060608469542,0.3867941467573214,
 0.18405704948096235,0.3224788183803129,0.2271175736433948,
 0.25959494596524624,0.3937227369228723,0.5560068132152283,
 0.22202448039833844,0.9457126331135629,0.11308864430713783,
 0.8673578581068409,0.6968217857828287,0.7700086492872944,0.9483722019135546,
 0.803554628778455,0.707552573751551,0.19060888080125848,0.06461043600905192,
 0.31405091311437683,0.5083942505291406,0.5292227247750163,0.8136880080878814,
 0.4210078915744677,0.4081022356837567,0.4668200186706716,0.7712737856360071,
 0.9634891541168196,0.149846072958302,0.1743297333582493,0.23387515383623136,
 0.05098990005177084,0.8333490268251668,0.44079356189701324,
 0.6628834185766737,0.12292541560939951,0.6686558661547722,
 0.14142324205304846,0.42925712836391394,0.25634600504980865,
 0.9283840314540667,0.6706213172548003,0.1199580627169432,0.7844647882863506,
 0.07480395690184483,0.630250266057273,0.648725672571333,0.04116833730971303,
 0.185502841787746,0.6466906707720897,0.34517137476682735,0.21549985201998534,
 0.7183325882759124,0.8010689225575158,0.45648432314210796,0.8955071496195721,
 0.30295050943486945,0.49992975142298857,0.730662584962561,
 0.0013263302033339475,0.2600716785618806,0.1397847042641075,
 0.2875965808636348,0.7684982750315731,0.5266152920388125,0.2180957872505891,
 0.19747210554009997,0.7716121262023574,0.08331023159699091,
 0.8563241375530415,0.16377215448128102,0.6357106755424508,0.4290776751743558,
 0.11951251670584806,0.4437086597115718,0.5110621824215971,0.9626428278416554,
 0.9660915772267286,0.3524356830687112,0.25247795023708763,0.629242588781185,
 0.2637132957565467,0.17583571602193926,0.2661858103445872,0.5515167670255361,
 0.44160892051952927,0.5325094171716513,0.16598611446631495,
 0.6891625198746737,0.9839366937316625,0.28676122106483937,0.9597223345012559,
 0.335522567124947,0.8076236350843079,0.5163572944137823,0.3454038706259962,
 0.06321449106114907,0.30112765701895494,0.09743008898906269,
 0.31576965488401354,0.2977946862337374,0.20562681851818376,
 0.1811142734308655,0.7175703434682859,0.3422100165704739,0.6015049440193929,
 0.7468142740824641,0.7541898377899707,0.9960863574823868,0.6887338775534771,
 0.12843636260332103,0.5623562892789014,0.7502212399719447,0.767628390803784,
 0.19942229502736541,0.580680096188744,0.08126284839904885,
 0.09132438726028957,0.24243117772335598,0.6281151789612933,
 0.6493339603163916,0.7235998114983453,0.44721134002412655,
 0.24303837311597598,0.6841400294984646,0.7439109379452121,0.3610053092862533,
 0.6464170756577912,0.5840024024233085,0.4732183783253312,0.43856160223790697,
 0.9518028929827003,0.0064136842696517204,0.5344103547170572,
 0.5990412114551681,0.17509825171714677,0.17737084696142813,
 0.5018100665012643,0.23452393269325178,0.15459902588367536,
 0.4778154410658467,0.24475542746924317,0.28866457257322176,0.568932359211094,
 0.11134949510227787,0.9709807926154259,4.837105954424992e-5,
 0.6870304126575897,0.7843309266574608,0.6930638391788297,0.9579703114647943,
 0.5668313237010872,0.7959419024243527,0.5940565361075815,0.9554619282144078,
 0.20577159175805837,0.7194402462186205,0.41839210321152465,
 0.35598328012180525,0.6334780482244374,0.6303487284380352,
 0.044321789977941006,0.803167483947379,0.28448831002240993,
 0.3046862609472074,0.12501264765983022,0.06901730004023121,
 0.39572410729899654,0.3936256247209423,0.21228150938114276,
 0.33449173900159423,0.08862692977841069,0.12253062844591156,
 0.8601086485187892,0.6393596984074608,0.21851935390414312,0.4205116886676993,
 0.49899052477243955,0.3201787676952905,0.8654108856927795,
 0.35271298536708695,0.6868468281115682,0.8676529678981505,0.2876217384389963,
 0.41247277647428837,0.18985834446687455,0.05271154968179781,
 0.5297841367958219,0.11003035953256335,0.7707878200497607,0.1496451479813513,
 0.09608366992815309,0.47458821710418697,0.548919946017652,0.7293110552694704,
 0.6461601864765047,0.43312529366967156,0.9746043031699909,0.962422791033667,
 0.227138237022115,0.6456580216543554,0.7876949888722269,0.14276791404273437,
 0.7781700145279182,0.13508611603421683,0.23711928084165546,
 0.8758495554090951,0.7365910769060564,0.12238861046087512,0.5873540325754872,
 0.8731763373586435,0.9303759298245216,0.8049569324856931,0.7865495271762223,
 0.14351700125321698,0.9041909968270071,0.20953370272556326,
 0.9674835379226665,0.6153792225804646,0.6825016072500626,0.47733375842490844,
 0.19110757599455508,0.7858203143673033,0.5124177047997708,
 0.39491925919513937,0.896941183528593,0.6140695586162949,0.7665955864779941,
 0.15564766624291448,0.22262290690134534,0.39884228141954836,
 0.4059236870631341,0.28860532371273984,0.05375973660144806,
 0.4967139345095788,0.8437391897319845,0.888668290298685,0.18166302400926704,
 0.5366842082112927,0.6985143074891389,0.6379269387471033,0.9562380805261594,
 0.22931200243619587,0.2335436362255311,0.05808891098681701,
 0.06339563964832573,0.07408663139674787,0.3033551263154295,
 0.5226185699865254,0.5456766062410255,0.876389766899851,0.5114070964976896,
 0.22781160518144739,0.017174823785457427,0.4717510962570026,
 0.32390745101411267,0.990291808107668,0.8690634386995194,0.27764380509173103,
 0.4378350234029562,0.996615317287897,0.5488588840960223,0.5372908937653987,
 0.8754075441481929,0.7993874767409109,0.6333471704523005,0.7170499947389251,
 0.9268769894946421,0.1491039680770656,0.7816784297942674,0.5941928486423149,
 0.9975058006249807,0.26393066243548713,0.007247420426047424,
 0.33103999070392787,0.31527538031948343,0.8191818574285789,
 0.47484623125041414,0.5809500292817332,0.6382983969830087,0.7680644453284137,
 0.2074413792941151,0.7141809603848651,0.19784229643520557,
 0.48007087053243735,0.32354593562293077,0.7955937187669346,
 0.8605143107727452,0.7643040459018663,0.03927324570145563,0.7818772737465197,
 0.557877487983975,0.8208940030974584,0.9414120068961939,0.4854885972160906,
 0.20553823190450082,0.6758748318756371,0.813583094289921,0.17163587957672277,
 0.2912545072454328,0.389973161037253,0.9044947914961623,0.6715436992552265,
 0.4716119578965453,0.15624895278966588,0.044477268343662724,
 0.05747224556865105,0.2683879726038044,0.46090491429272673,
 0.1900103001482072,0.32887691445521416,0.7669629632055587,0.1774144748983285,
 0.4255255023574105,0.18941954269774774,0.693533391527647,0.22085155180194826,
 0.34312861644873605,0.5182182076128534,0.28448450085618904,
 0.26451982451871225,0.3145888937081638,0.38836985843265914,
 0.4427240734991569,0.4071655783878585,0.8520603234856216,0.04589179762872364,
 0.7186950371848683,0.3353914509425584,0.4889912212984464,0.01229501352236495,
 0.4705331574856806,0.5937524540221598,0.20677302169209255,0.7319661405335316,
 0.6805246712655532,0.7632500821364543,0.6140257040707543,0.17309327075478131,
 0.5427070447973146,0.06804312272462831,0.8228857690950777,0.3130862224136358,
 0.5132969729161275,0.10119691835282141,0.6793756827939633,
 0.036219849689334005,0.9160172795204671,0.3935699102481074,0.682763858930647,
 0.40294496914466793,0.20089407181945085,0.41999971231453115,
 0.11593283839726565,0.09925544861806612,0.1576736175243153,
 0.10853622666164908,0.2823316504791059,0.5858665897195098,0.724658116709738,
 0.6836459900608048,0.04074018026605086,0.6222315006742356,0.5169685585175222,
 0.2509830085117275,0.07824366156493812,0.6437210009539938,
 0.07958689039507916,0.15411141182906707,0.6340107235568253,
 0.09194351147211488,0.7110571915824269,0.5615232671554343,
 0.07477010457783095,0.4343182435275963,0.07330758273026816,
 0.09746233499338153,0.9088936457837375,0.7206502333438973,0.2025431852037617,
 0.3198268880520646,0.10075134441135292,0.7442336347245545,0.8627506702940039,
 0.9062731516915685,0.55060441767946,0.08822829488383643,0.11059488756344593,
 0.5411298514025678,0.6609883943861274,0.6898872402893363,0.6567539039154205,
 0.48511887501260464,0.3554913545214061,0.4174901224676504,0.9944703595350669,
 0.7550256325646993,0.3086896340095594,0.6166281604892265,0.19956810546163384,
 0.1504495044809896,0.16897644339598783,0.6488195349289643,0.8218339023437446,
 0.9329441436029854,0.8496662811187082,0.8443352174259983,0.9260927297374183,
 0.13473092454113367,0.7388894593479878,0.6162784976561307,0.5516800542089835,
 0.3546042132513406,0.3103549396567089,0.2679522807063568,0.8069080098889798,
 0.7044698625613914,0.6293304481997127,0.4845147557875923,0.24885077935172628,
 0.9135634087038742,0.5282098950733565,0.7655841205095049,0.18970542426952786,
 0.4301310823036171,0.5354452993926025,0.676692158957247,0.6932837746966978,
 0.11488871903142117,0.4517284546822946,0.15322823936333796,
 0.32228476730894484,0.9841405864934557,0.0471517431508135,0.8389160211121875,
 0.8154174446538256,0.5990872033180588,0.3160265472466851,0.9915433876708227,
 0.47255882402934146,0.1808731934496013,0.8942495299357743,
 0.18685568896761162,0.3426875714609082,0.5287648932588668,0.2474716331361848,
 0.8499307481249921,0.881042921800726,0.29807271290810655,
 0.049376793688132525,0.6931416128699005,0.019490577370660622,
 0.31156088514231295,0.42528472182843435,0.8641195212940308,0.412053891640456,
 0.37686297303750027,0.9683781859326316,0.2273951691961298,0.2632669355852133,
 0.12083534197781853,0.11301685948280771,0.17566615952094922,
 0.9217025960852698,0.8385899117597412,0.8559779284139433,0.8747556756437577,
 0.16640551972909412,0.45009964416912307,0.42226429180509817,
 0.664449294221958,0.6188988154773487,0.24710672550413582,0.8636078438500695,
 0.48017974780425743,0.26020301302688664,0.18124007919119933,
 0.15867117699339117,0.44695466465966416,0.9785423111892728,
 0.5737328369059643,0.27198635734478727,0.5138216834697557,0.7949745443378561,
 0.12849067334280684,0.5744314912631194,0.953445354900996,0.56222356239985,
 0.15278637643174675,0.7613938483672232,0.34713015136263736,
 0.6308921565628038,0.22294399204241833,0.3502373545515829,
 0.00515710316392437,0.42567901055034874,0.6292821888017437,
 0.07575326367733037,0.31959902735413925,0.25245183431759244,
 0.27513754430462267,0.5952364671653723,0.32478151994626736,
 0.6107764017583638,0.7898760765039945,0.7833890925689461,0.23487237456623655,
 0.8099411998289135,0.8176600723278237,0.48758257933725235,0.5834699515868609,
 0.21923476631905703,0.741259772446483,0.498880168918039,0.4689903854021248,
 0.09834030224638102,0.39808796201112395,0.2330543491457977,
 0.8961486603867699,0.681562676336761,0.7844139055121201,0.31514311163133635,
 0.6419660832404132,0.794417580646707,0.19338968753626695,
 0.054162700090146454,0.9529244397343551,0.537466127679212,
 0.16338769911271644,0.4631527340759509,0.5453909085613434,
 0.35071257840606773,0.18967183526942932,0.11985479328347559,
 0.28080383956427557,0.23256319428655992,0.3158117316645478,0.961013621171197,
 0.4595924813226663,0.9404934794395113,0.13968077089095043,
 0.44451797529454873,0.7422889270617142,0.37184859844951923,
 0.42736600698788674,0.5219598250077485,0.9093777835859336,0.6185292536149554,
 0.7737397615505242,0.6978038649080577,0.8512170425486818,0.9411901800208646,
 0.23372291071860596,0.8697146296669753,0.05760320927557938,
 0.6323614722099209,0.6767188130180529,0.4133390334040137,0.5184554800483617,
 0.9315044003068819,0.48680890406491373,0.8988247800091758,0.298477359773752,
 0.25960566984887057,0.029623979253819877,0.6901237051345901,
 0.796635282427804,0.48953592156268755,0.8165735212755835,0.29560756468019334,
 0.03395741906830885,0.24016839635251452,0.8019307978123554,
 0.6932069595815877,0.11977716362972124,0.6977824641762763,0.2945494768032325,
 0.980419263784881,0.9025237220539444,0.8323983280860061,0.1430423317935967,
 0.7662873716059239,0.9990728964584388,0.13505349086955076,0.9311385372570089,
 0.8974890012428712,0.7383054518430682,0.6527799493714936,0.9835967129621175,
 0.4408819191578297,0.05052851455714269,0.6284272627471079,
 0.036450535338227175,0.5098177254296111,0.7498557258114271,0.911715250275098,
 0.988107922768104,0.4975848777548708,0.9475108665526907,0.5352825195327405,
 0.660899609610888,0.8560473231673508,0.09655720553388603,0.5449469123337636,
 0.6559041907738778,0.018179332942651727,0.6048637045434265,
 0.7883195128551623,0.17177767630072016,0.056732144086134806,
 0.5708078940953475,0.35541901685542365,0.7681045037393932,0.8459097035760317,
 0.3176416793437632,0.8208817436856493,0.5421878048738128,0.29974155447931494,
 0.6922646004503922,0.6394864432241082,0.8765151125537363,0.5770786056951217,
 0.5321699984458353,0.49690465814812756,0.31138486502249585,
 0.5476959837336699,0.956521850764043,0.9999246146520528,0.8204886026789399,
 0.4885005571771842,0.8916426728472066,0.7571457825089818,0.11304025976238552,
 0.6822506485828349,0.0591681840180458,0.0606955473552776,
 0.007741020484130501,0.35447795814572514,0.39503366994702305,
 0.43713675599758806,0.19945736933184455,0.0038395845871987966,
 0.011160518227269156,0.1759416506437652,0.81292906608748,0.37499670722307854,
 0.35940944333106783,0.08838544538804738,0.12518944171036395,
 0.600174340449054,0.20135666798173846,0.9239860154179815,0.6848620489061885,
 0.012331327530186176,0.9582978686625679,0.5678257495122887,
 0.47376007044696933,0.967936489450509,0.7974083641903404,0.45319248921692346,
 0.5832276496885687,0.33096298743194397,0.6010005005601821,
 0.16693850939172072,0.6991303460624845,0.4601176178039794,0.5572044011381939,
 0.8007006324665238,0.635323882173381,0.5277076072806561,0.11298286346507269,
 0.7139819983701072,0.11103023493853947,0.44872918021481367,
 0.4917228331420149,0.17364536318019985,0.911241554501078,0.6281290609178043,
 0.6398044925102088,0.7424403874398174,0.2432273884792191,0.35151865297574114,
 0.7143142043443655,0.13725072560794305,0.5238501195272225,0.9311146376308547,
 0.6607006960759142,0.1617255842062647,0.23765202369059601,
 0.46448627536575926,0.44064806424521563,0.9602170884243393,
 0.6188367542814653,0.30420372684544483,0.9940830840061816,0.3973004368605143,
 0.8795613975459813,0.969374702097949,0.8284788485605417,0.12353515940139737,
 0.9905937320302342,0.10621232610136278,0.6058371724161237,0.9534740607178893,
 0.9215849588028519,0.4420522586821196,0.31082855903704987,
 0.37413017394918746,0.7716499380978323,0.9945453881679464,0.583647115879595,
 0.987255514237526,0.5411742425735756,0.7976778926952152,0.7114155839280187,
 0.9407002023049437,0.4110135807938805,0.45929621735132775,0.9119837829192992,
 0.617226188840488,0.5075219273440422,0.5022159594524317,0.8764712115514772,
 0.363065698546679,0.8843671128878763,0.41890242616324436,0.05191123189800917,
 0.08309464167702885,0.0639677152333139,0.5195894058298416,
 0.062284864958620956,0.580738569515922,0.9210837182967428,
 0.14374128485054083,0.6738121013231564,0.5419534829747595,
 0.06288354417248643,0.6741524902201771,0.745100392780007,0.5017885219530956,
 0.043192518854178075,0.16232442462756014,0.6541603325165408,
 0.9072565099365699,0.28671934684878475,0.853899389492113,0.9646447030379671,
 0.7334181473078317,0.6155627566528752,0.8908816263428195,0.5867868574341393,
 0.14135813724683222,0.29976410260751996,0.6580547497011031,
 0.17018182158536344,0.623458992098005,0.8705694222414525,0.9129290419103472,
 0.4593951364178325,0.5061882432397511,0.6094301216677348,0.6222474259271611,
 0.2557589351476861,0.6339416677950502,0.7895773996085758,0.3318657041109775,
 0.7328212325882851,0.0331409246958293,0.049355059426942294,
 0.09756119346799685,0.37822062224626096,0.17059940357425862,
 0.3406287417326751,0.8236111123669331,0.07355458565500039,0.7902174897746174,
 0.6028949286553664,0.6834884026956122,0.23592125431536481,
 0.09555253168797417,0.40381582818848993,0.4806987187685007,
 0.4884343299357148,0.4662398814368466,0.6499048924059968,0.4580479681090758,
 0.8094262880686907,0.531714444955031,0.8127033674255718,0.5310349911066065,
 0.2749681436691349,0.004959293622293748,0.24186562962454905,
 0.799659787321309,0.7001998188817301,0.9571951786920613,0.5737013372260988,
 0.6362426700887531,0.32886285586727215,0.2687121816900361,
 0.08889472297071987,0.2939490256788473,0.01504819882122177,
 0.8649566599705915,0.9049531492055786,0.6551599793619725,0.28762119956937815,
 0.4529542018282129,0.10047216644628498,0.4700820853720834,0.8591837800285741,
 0.6487943576238219,0.15290607103654508,0.8541947316895747,0.8856367292352959,
 0.26205525416617825,0.4452641592817591,0.13701089482434547,
 0.6355649830543741,0.379514670638324,0.917186631905267,0.3350764189802259,
 0.799298795415451,0.7636937573512907,0.35286127032664905,0.01703066262049957,
 0.8677652692847574,0.7275615944687992,0.6974608559906008,0.43179530594527404,
 0.3329043286766402,0.5852067790617774,0.616799865110652,0.2840700749669588,
 0.7713078783681443,0.5036560631213163,0.770407722600684,0.681967817668166,
 0.49766208144584867,0.3503275870639063,0.16239300160736792,
 0.6942571209123098,0.7498611813674843,0.6952401463891861,0.8348097914563186,
 0.9341815379704783,0.9459378609624346,0.13831094716528192,0.7371120840884638,
 0.6575832846531022,0.49978518595417,0.254372515368091,0.6896701052811675,
 0.48253221642582833,0.9756706765870261,0.16919778648758,0.3976209432178215,
 0.5565761287029639,0.667907348497292,0.7566220722725134,0.7840525565964099,
 0.244352651631349,0.7587577813809027,0.8497188841415505,0.8634251858092639,
 0.21163436265273416,0.9663882017814797,0.18083963989369034,
 0.5357928431847067,0.7469476856367034,0.8161626135651026,0.5692999897925844,
 0.8904008575556118,0.30912831503414706,0.46532383601762695,
 0.08252158102736051,0.03494442378583984,0.8152643580008021,
 0.7518688650630472,0.5307044124476576,0.5351920991842872,0.01343421234056752,
 0.44182813123443276,0.8414306433529231,0.4318348782597139,0.7284910684301693,
 0.29042413050297533,0.9837336938510701,0.2804261600542959,0.9497292740065257,
 0.04401156509842152,0.1695756812228193,0.746709004783612,0.1836579756826976,
 0.39925874801943884,0.09480314406132728,0.10928216978960115,
 0.1729668465768479,0.17847920291848307,0.6869334958558605,
 0.39536370447493585,0.5694899649697327,0.9979040094370111,0.9348924815817212,
 0.40909991399658563,0.6822083503316357,0.10705582115870305,
 0.7039957374155502,0.34231649040843637,0.7680588857292697,0.9766074401441645,
 0.9670808910887446,0.46301576500466135,0.5216076259712037,
 0.07373788365749223,0.14906242706966033,0.5017281521249369,
 0.9608205240210091,0.9280760283384835,0.2892017950680309,0.22132673761836508,
 0.3122593020017944,0.7889234083053331,0.2386263253283355,0.08158769713864067,
 0.5133521129215366,0.28436707507830816,0.4755558144312718,
 0.07421965542318287,0.18889051967896597,0.24345924627897875,
 0.8284081285964173,0.8411369481460902,0.012744557942028,0.7443792615506194,
 0.06777431122820854,0.6788740294844907,0.012770381312840362,
 0.4473239913686825,0.727872465521178,0.9405642826898196,0.011530398158674182,
 0.46390142196791184,0.9249308475285981,0.7029423788496441,0.6682283587753004,
 0.3540300363827671,0.8270985867973649,0.5871432245636174,0.34265613899965575,
 0.919035855274121,0.022129795398810548,0.8833011462811295,0.1785399778383251,
 0.5538080958336337,0.38122803420724294,0.5657791727079093,0.8315330678789771,
 0.28385250824919017,0.6310708848822243,0.9743971172052962,
 0.14499182725172877,0.7619503387573161,0.10658729451995796,
 0.1694490991926294,0.22971430067917809,0.06805414520701347,
 0.9116306782532309,0.39625517225950946,0.009047853748513579,
 0.7038150150551968,0.3622813679600637,0.030213450919290286,0.839893704332622,
 0.6165772823163308,0.3086616712184108,0.7550220299618312,0.6217046439027787,
 0.9801876433953942,0.32853796304096483,0.9670581519532592,0.735009378683336,
 0.478526039491062,0.5024973407512628,0.7451971047785244,0.5313816394068711,
 0.43880427642617137,0.5638208099355435,0.41613569528504857,
 0.4543309800903731,0.3240724802764855,0.653413458354899,0.1273977830018791,
 0.24805318929336584,0.6225466191611599,0.5860865897223578,
 0.039966533766178536,0.8853544500762326,0.018050638968304278,
 0.21548953280093375,0.6839413847831498,0.6077928006039967,0.4509573096078774,
 0.5534277514475676,0.42871807573481047,0.3346521824376003,
 0.37601219148637044,0.12405677913919477,0.668186553891291,
 0.08864269548613068,0.982119580627101,0.9646115243236331,0.9500803423966169,
 0.6914229519311104,0.3146945693713118,0.4531087384481992,0.8145340451680124,
 0.8677797802875682,0.10968173111996682,0.6148697858637255,
 0.08509422367938724,0.2623538256724094,0.9694294013083211,0.7624702986595956,
 0.14367432885090115,0.14885325057298504,0.4441918726424332,
 0.2909872620239906,0.8977055553668665,0.30991988186174924,0.6364556964518929,
 0.21588609432411576,0.18487202973734196,0.10845795162696359,
 0.7712153324028391,0.34814724901916005,0.6572169718336263,0.3969317224240413,
 0.7555479820430278,0.4298542412396351,0.9135147476786357,0.6108064969385296,
 0.3319977105000166,0.7327029912300542,0.8734410295470905,0.5661794321212587,
 0.5024625814869068,0.35916494681487354,0.7650690422637673,0.4101302720866369,
 0.7706372659624003,0.5429086634212009,0.41785821462705686,0.5398858173084002,
 0.17694122328557005,0.4763820908971764,0.636356905959017,0.9039214380186625,
 0.12825969704699614,0.9165520181364314,0.13666026287460487,
 0.015299047338974026,0.023254431527999775,0.9719140048408108,
 0.6766263841272067,0.7912858219593151,0.8117898390181304,0.6594941594433419,
 0.014334071644982904,0.48715641677776467,0.4326541519568513,
 0.4378559148064337,0.93377975098397,0.3592921918855192,0.6324519205955715,
 0.1887941687297482,0.9137824417483299,0.20972616888946993,
 0.47346073665504707,0.07473783388001431,0.8611427625284696,
 0.5087372782386246,0.06997918694679406,0.4170885177560292,0.5513686055306661,
 0.2617232839790953,0.254756979492164,0.19674816408561002,0.6164505116271123,
 0.8447955822028923,0.3602310504223575,0.970176412003415,0.38566599333799223,
 0.37247530064409595,0.14850170472580793,0.9135955450554478,
 0.5693227204686586,0.6251079300937287,0.9758785986756654,0.7190401705937882,
 0.4181920750447444,0.5214452929788371,0.8322334047196692,0.7414066249732237,
 0.719090590438221,0.7409600182578774,0.8262643028799794,0.8493701773025545,
 0.5058040987378767,0.40118667973344213,0.7742416703343432,
 0.22940358056505294,0.01460381129179067,0.03191824603727622,
 0.601292724787126,0.38928454142080005,0.08817683683778221,
 0.31167537719057126,0.7935729194741193,0.1436751803480839,
 0.09754012267921242,0.4969470281760676,0.8730059626561482,0.7030797261796474,
 0.49641150162517234,0.5576156069968633,0.16672895594940196,
 0.6227963661723827,0.23005462430188683,0.5934602040623091,
 0.23231024996928218,0.9374474614266779,0.7811651287078508,0.6142411263168763,
 0.2835380370376862,0.26960449490750515,0.47097717253794447,
 0.7925517099110739,0.6485359940394513,0.7434291043334389,0.6072739392112929,
 0.2838635934712336,0.5435316575224671,0.4813153401789282,0.9787681743218986,
 0.3085435268134824,0.6903334090944666,0.091219339281701,0.8336233107020617,
 0.3396480797677067,0.7560789504065228,0.10223491418996267,
 0.14773354563570984,0.21063372755420717,0.30509749890088145,
 0.19094817624273108,0.4789389464683229,0.24100523340526314,
 0.20015567282328506,0.2755691856272052,0.3097615263657515,
 0.32141609767816026,0.7711067328737718,0.6613243631700995,0.7602412222727432,
 0.7077109337806516,0.4155839166908142,0.37550453323646105,
 0.22810866433395582,0.4980123516631684,0.23959902537148148,
 0.06774829516787728,0.20084261372819556,0.682366013730052,0.9578094991640583,
 0.3335853177793544,0.7162662882931555,0.0747586034256339,0.39771760832602443,
 0.9005910130209853,0.9710418433077048,0.9584907606725106,0.3485395515383267,
 0.9500393731889129,0.431212920136942,0.18491258051155013,0.4069010538040858,
 0.839360537084499,0.8792380345824042,0.9269280708163871,0.8728648449983567,
 0.1329979977263398,0.827663246385155,0.8284112473314055,0.3688348685161319,
 0.21120378184126276,0.4966372678060976,0.54358466666693,0.9621457384347571,
 0.9485091394201778,0.646482896399996,0.26985951333488745,0.1460582540315798,
 0.9456576752741833,0.4061125758035925,0.18051300260816905,0.8020737115804731,
 0.5179958564482811,0.25494903725779516,0.2971699595599042,0.3904513078710995,
 0.5241870124618028,0.7365898816728103,0.9859800808444077,0.9750649717657753,
 0.9237345953252797,0.8248689926941896,0.23760027000321216,0.8363875956796177,
 0.36678081532258966,0.5335999943154794,0.24418964611974014,
 0.5921084519250144,0.6998654889074107,0.9287866131209919,0.44345644795764194,
 0.16288981281513493,0.3306687954455235,0.4220382267873257,0.1753237854431019,
 0.733459534161988,0.6165666372051203,0.780636234443648,0.034166666863712836,
 0.09818341893462634,0.08738094675497343,0.3910783942436532,
 0.3329639900975865,0.9212538417285359,0.7894540345896276,0.3367184416346194,
 0.46171967698636507,0.12963796597371546,0.370994980169709,0.6837477320698515,
 0.9840427936463634,0.4284666610492023,0.5449003924861967,0.47626468922776255,
 0.3783233658722185,0.005366507275658683,0.6028200564484797,
 0.14001254527907603,0.2855587571815351,0.06674377704082335,
 0.4048036701805866,0.6594687847606094,0.9021445973822486,0.2733103628668734,
 0.34971447200546,0.025307834214289837,0.07926123875997781,0.7240220314584148,
 0.9085736294004944,0.47247316662971806,0.9852821420278617,0.8764926581430457,
 0.4563755706440913,0.49106140132002896,0.409537492613144,0.5026781992267308,
 0.7368654039504452,0.8679227271109016,0.3808994217522198,0.5519148574699325,
 0.3730758056366661,0.6050569045924397,0.2678014050831504,0.2511256566710858,
 0.785561189886662,0.8038956880832335,0.6748555146774573,0.07563403834743898,
 0.6188595624364976,0.38608696697816647,0.5895844841081387,0.9721291464118935,
 0.6796922160407904,0.20005274589822064,0.8661613417787267,0.3020435340274271,
 0.6718562447891764,0.20196343218079238,0.19880152314026012,
 0.27061924935665316,0.1831915750418881,0.46567903035305247,
 0.007006776155757732,0.08908527565727686,0.6599911859831679,
 0.11215366316245612,0.329211348837394,0.7569236573172846,0.79974760513249,
 0.027957181386181418,0.5722428384637378,0.9355616084648506,
 0.2478540581072246,0.029819167868967766,0.29045146094855134,
 0.8986793043989518,0.5634190076402236,0.5621311268940001,0.9590536720633147,
 0.9707878014659911,0.5934279220147229,0.2866509682671565,0.4225718886887412,
 0.5461341996924644,0.10592405232757451,0.29369723062211217,
 0.24255149934848208,0.9242459145754345,0.07837755672968494,
 0.3902089704521704,0.6760835161053038,0.6050878715868057,0.08707848336791413,
 0.5839483980950866,0.7697668979020448,0.7620788491857351,
 0.053684105047425934,0.39690456887098613,0.9761315513721982,
 0.9663590340708306,0.8042428758625653,0.5033079076285476,0.24680751471724494,
 0.48353248603337795,0.3293458427516234,0.9816114456977874,0.5511173664732134,
 0.20478394142352463,0.8636892220772819,0.7415687867168204,0.468408944126685,
 0.1463016940384818,0.9541648544909025,0.6324713626137913,0.10402237744763554,
 0.11245042417498408,0.31199849095138343,0.25570453732998344,
 0.5918085211266312,0.9830161284781505,0.32092022800606956,0.316434407284909,
 0.8685060353670865,0.3196043514295501,0.7198958023988082,0.28409577391321983,
 0.19219876013074255,0.6372506603342488,0.7734798448981124,0.7618117996494096,
 0.7978298335546228,0.06637194821703163,0.05798391014112847,
 0.9996052502930269,0.5518228987613556,0.3934933282173545,0.09595464299904444,
 0.5631372118626387,0.6439914213180378,0.3754903934211409,0.8730962083403465,
 0.9967136741569604,0.12328792564763069,0.3920135415144682,0.1890920388746089,
 0.036192109810543105,0.8611479790971234,0.7862635295845053,
 0.9680996346969433,0.16286419186308432,0.5730691854427076,0.5147679299418272,
 0.12413305930647445,0.046043815087062345,0.70564526197932,0.6312518188795235,
 0.21956066409459907,0.25457876915712574,0.681488983141975,
 0.02274586264783318,0.8330774188771171,0.31785001264618074,
 0.6330419057909964,0.36706351844932017,0.9053421483843594,
 0.03840342256359919,0.6455492250964086,0.5764211402934819,0.5490801566267106,
 0.11538730909661998,0.03203026164235079,0.8037188030081401,
 0.9716653189337803,0.19266579254028626,0.8852357757536116,0.5082993932268642,
 0.5187863313987736,0.9216744901406427,0.6075554574789392,0.19298752710430056,
 0.7655712444923869,0.5901447788717982,0.9062984934586961,0.20933117518668287,
 0.4814714404161552,0.6591986008082451,0.4589331260552266,0.5416644550318109,
 0.8457301606544956,0.5537158363432626,0.16480161168332308,0.365883973214996,
 0.5862527816105942,0.752628084465404,0.19360721444312,0.8121028626709131,
 0.8053164657296592,0.08084524148769812,0.2260699559138314,0.9586907155087875,
 0.03280588541969853,0.10488356305690627,0.5177967835142233,
 0.06336643653175456,0.8501793104210512,0.2110024969082197,0.2880296885562674,
 0.24415455499243943,0.4737147622759038,0.39627373324074844,
 0.6116837854035082,0.13135186950401323,0.6730623650031636,0.8712805813171729,
 0.36642520105136844,0.6285868400043788,0.14250970403308527,
 0.8190415354272235,0.3724714959155525,0.8484316420068234,0.9106163968491846,
 0.40313426377571504,0.06536803964648363,0.34480319665896086,
 0.2642342527884687,0.9752823383239981,0.41711497636714745,0.8192108313483389,
 0.8754049728786211,0.9577226938746147,0.25231262549842914,
 0.19004663984842896,0.7342913609707192,0.8321167838835731,0.5979225651767195,
 0.7578723375836823,0.4038898721176797,0.04704039454606446,0.4016870488102081,
 0.23064012026726433,0.4119452940581455,0.7788101107726583,
 0.07249855767361213,0.6502783579920882,0.020332412391869892,
 0.38657096018907144,0.35041524740340846,0.39779875662401165,
 0.8692062437724689,0.4094260666497923,0.957095962216502,0.4057900994175633,
 0.5667726410213819,0.44668930012755004,0.02553333581797812,
 0.6967765606684632,0.41201822345009254,0.056362009509467015,
 0.9006510660135904,0.3897555123604812,0.12376097152848642,0.4604161850688149,
 0.41242886940857193,0.5703113430012506,0.9609972252939483,0.7769931713928252,
 0.3405475538826016,0.027617557428537776,0.25318122654491004,
 0.12664981191901692,0.2938466338518314,0.10117638463177347,
 0.8527565540562749,0.020723280757240503,0.9102528850890184,
 0.8093485970760813,0.7378709410725303,0.2635842401373898,0.8855437375519594,
 0.3436476360532785,0.1085931062537614,0.5735683278253083,0.6650140524695954,
 0.2923714865087059,0.4525189431830109,0.9512608567566737,0.1184369055513036,
 0.7049849069683543,0.7947794900010573,0.04304465448019512,
 0.36100242754349365,0.5826989831889455,0.9759701460779591,0.9721149015635216,
 0.037325432172509276,0.0110560627709958,0.21622348544872183,
 0.9619872978964945,0.19500722452603936,0.06731158961565686,
 0.4356293178172078,0.5761223144706156,0.062181784880629976,
 0.48609394723170885,0.5271939958217244,0.14720258207861003,
 0.08735762305505002,0.0968473985478312,0.4712299109160587,0.9412044750890317,
 0.4985643876309721,0.6732476177186665,0.7629328034963178,0.36224733564218237,
 0.977846348630323,0.9976933080904979,0.619952450907464,0.5369292160131667,
 0.20549788873030028,0.264690096460355,0.5371184443802431,0.16868593582495195,
 0.718915021723229,0.6103642104577016,0.6647315676428669,0.9872563834186483,
 0.03336788159641557,0.3243248335802308,0.19277203672676402,0.989688855153537,
 0.9111918565852652,0.7750801039924955,0.5410642213423921,0.5650230487136997,
 0.21708360270966065,0.2076514298354616,0.22957664283203028,
 0.8206565430472764,0.12150993353983675,0.7020231766102599,0.9621240924030411,
 0.3762181574519108,0.3912731614735089,0.13145729061959965,0.5383912167649334,
 0.7694323838997725,0.9253046981551432,0.9822401018780312,0.7668600622851942,
 0.8525339783340162,0.5265949564462469,0.9244668230916853,0.7584018914195592,
 0.5316742005804621,0.0990989868044192,0.38017300765447803,0.1496300756908201,
 0.0827124583622264,0.9105362202871146,0.213496678481655,0.5343730471583605,
 0.9705032035809028,0.9100695365454533,0.0853812943383252,0.9598478450299992,
 0.6839467198373864,0.2993069049697261,0.7543872874256868,0.9455567507006184,
 0.6932762170959939,0.9542732911637158,0.32170239421040037,
 0.09646018320198024,0.5843047739434635,0.9777780725057329,0.2142358468767036,
 0.7776746517194381,0.41994636141810726,0.7162670070923578,0.7353018497090317,
 0.5027219187848566,0.5303348202382616,0.06526112548524399,0.2389639718263069,
 0.05824847204713979,0.20676754734657832,0.02841282629099262,
 0.33849667010489537,0.533099062253869,0.10720652585885382,0.4247088935255696,
 0.19098115859726472,0.5218158496773149,0.367016573411757,0.4253075168481043,
 0.6133231914556161,0.5088705722053415,0.8871511635401148,0.13641283215803224,
 0.19289576881160353,0.27905962357425085,0.7358866980202892,
 0.7418732682885698,0.7969268724949263,0.04219883227338217,0.7821452387777935,
 0.19565302404232254,0.4406172384530467,0.8846754722608264,0.9188318674775874,
 0.7219691998493827,0.48267677151535937,0.9646728208376362,0.889200871539051,
 0.9999721833424353,0.9217374028748501,0.08340261136773475,
 0.19185657741040507,0.8724880428931032,0.8038455019053334,0.7634680185707792,
 0.19011546468510276,0.9971375074023394,0.7728235199436089,0.681561281438845,
 0.550889213246597,0.7680317705627059,0.5122719699545522,0.12329454238941184,
 0.9049211518861754,0.48351074191632737,0.5027409849112554,0.6840027015172239,
 0.3324971990384096,0.7339868061425651,0.0553470542065011,0.8216036761272046,
 0.7957425650208348,0.2939030368573943,0.8292167317406041,0.1553606420459308,
 0.3679708960545638,0.29828730245269885,0.3532007534296667,0.6129793752199091,
 0.9500009845780193,0.22949066104938642,0.6819983517459823,0.6517218064151092,
 0.054702616499311185,0.1041857681260141,0.5695836831328345,
 0.5902238305945562,0.8703384731115832,0.7679905341341489,0.4601400643816844,
 0.8783910057908026,0.6005873268632943,0.700143738575391,0.8603783955203889,
 0.10227038998359494,0.09925045664458398,0.3620841861357912,
 0.43019964224338425,0.08143229638564642,0.2768249217317469,
 0.49912400953060687,0.8025453911053453,0.9217382834090875,0.2836780020453158,
 0.16190993497006956,0.7803715386463888,0.6958036524839168,
 0.19669085348179027,0.8103820381544828,0.9905958825140876,0.4691505143938597,
 0.33749508126556405,0.8758268509806629,0.9478515401606766,
 0.44819660025110053,0.8193972336339158,0.9127490188692835,
 0.12693229099949832,0.10096021636767771,0.21211700996871585,
 0.21286699719690505,0.7213348371018992,0.13812386416974165,
 0.2099581597905129,0.44429887645542165,0.25757555754611505,
 0.5870267586584449,0.6279396653337346,0.7781481648680986,0.8416911902495974,
 0.7848152283241825,0.589375083316822,0.6632642494646017,0.4698394728673059,
 0.4530220289297847,0.9315948894435754,0.550537483530337,0.23181376737909365,
 0.6764625536407644,0.4702747619859051,0.8752491286775482,0.8725457385127138,
 0.22810621802575803,0.5384211358420044,0.9600530019343025,0.3534725104645784,
 0.09890273990315646,0.5985884104171968,0.7934024257042089,0.7336986247580429,
 0.84493871945644,0.2713788425949417,0.4658570460037388,0.4673990554700017,
 0.3764857639248794,0.47809992208814367,0.09367807809033935,
 0.7241407680928446,0.12792281764343916,0.06577131847590689,
 0.27895271646641195,0.91596632258808,0.04970271127660342,0.5886000124733197,
 0.13860709949340455,0.2742456617183693,0.4045050442490665,0.8946730657887101,
 0.4336544037337552,0.5559847946785632,0.9131527941570758,0.409071512319947,
 0.6860082287110185,0.7094810359097098,0.861690575044068,0.3739116642479884,
 0.16988729860590623,0.02903303260801271,0.1267165765215844,
 0.8349804228606367,0.8832788864374204,0.1773987916422386,0.9003997336915365,
 0.8723641990253961,0.2594180965659534,0.5635736947889414,0.03989900116459055,
 0.7877381515695278,0.17725327568590887,0.493069265503597,0.6103061161328649,
 0.6354945744897206,0.24875887264112484,0.5527171070434432,
 0.23500417145935393,0.402287950011328,0.6629891544267597,0.8598321272524833,
 0.1944433316958809,0.10015139662766803,0.32177971990430076,
 0.7351412356291539,0.5170642039773298,5.030112401325226e-4,0.575487452507078,
 0.11784160160791612,0.727053994432554,0.22608133577385203,0.5114368903296652,
 0.4497647924711127,0.13803848872572244,0.6472025689295543,0.739231379087242,
 0.5604031777430452,0.4259971381800187,0.37341266526598516,0.5203204283125413,
 0.4060035602071478,0.388920011970768,0.4372166435281587,0.18875784968422704,
 0.782573967143686,0.3834362918863714,0.8893254218533726,0.9281109998900109,
 0.19024702115443037,0.9624936222912188,0.7294953682474333,0.7906036142191843,
 0.729991466203095,0.4956276912200849,0.4759157720536882,0.4467903959095193,
 0.9132719015421317,0.8344572269861913,0.9353769878262215,0.8633321748650183,
 0.4835919667030778,0.12754761245618174,0.5289420887343635,0.2147878569088425,
 0.5415761406882185,0.06073850799944169,0.1320711513218289,0.9740883695677583,
 0.05270808223294998,0.014281140121078284,0.9459707277574572,
 0.12642866417001752,0.910708951557161,0.8934435222931737,0.22293698470647538,
 0.7761293072350605,0.42924959208869695,0.182865716213777,0.13575313576683445,
 0.028973618963035075,0.6888000437604289,0.6097825877698098,
 0.24870119910327015,0.050906141086597745,0.500324017533343,
 0.8729855924454061,0.32696502976539366,0.560733813421088,0.5780638297366395,
 0.37706197150790133,0.29522649307370996,0.6425558823887114,
 0.8452569922980481,0.6488840444011114,0.3916730874016928,0.14766444178521454,
 0.10039008379194347,0.9716359381222268,0.19250485709089693,
 0.7453857235350709,0.23226570245184108,0.7639939876804229,
 0.39410164618437427,0.8823421815081701,0.013130171455341233,
 0.1407724281430256,0.0582218154554387,0.08597285886361439,0.6517432127096596,
 0.1615564809291382,0.4796188131607313,0.178572932872747,0.48067567838291925,
 0.7867402178707459,0.44670003171529005,0.7116493857675774,0.9733778981450844,
 0.780375052143102,0.7084749837473114,0.2816938684044902,0.2471423515816269,
 0.7510261631973771,0.37684543848674057,0.3061036174328161,
 0.08111870776288721,0.27673386087420726,0.781643975729769,0.5571829603152831,
 0.9647356908409886,0.21775992759385254,0.3874636346838959,0.6002619499314301,
 0.9064097164996275,0.5673869179131765,0.008563338953305855,
 0.33267285956112513,0.7603410619269684,0.25763325511990287,
 0.5307605749183676,0.9411295775683732,0.6706721117734278,0.1252889171990439,
 0.9619809642859737,0.4541196813985472,0.5945016270645459,0.2052457170621016,
 0.3213628398263897,0.859600939529421,0.6250967155816092,0.9718808239054032,
 0.7547941589766372,0.0852513242207017,0.10372004587875605,0.5466368782652197,
 0.22478911141444402,0.8297460115689101,0.9376284479705924,
 0.0060100100647960675,0.5859490130480987,0.7719382124988358,
 0.8099054067192399,0.7437486335442964,0.9351391348272362,0.6378186061780127,
 0.7637694600490048,0.12550886421399354,0.8368583582827871,
 0.22431659207886212,0.5771799726044011,0.5803564309982137,
 0.05776793894621335,0.909195153247403,0.5717322022551126,0.7239215306255213,
 0.31805587170626226,0.9746538430272306,0.37085432667967133,
 0.38354622634781566,0.509936826400266,0.31463304769769307,0.5745888108519087,
 0.7383258468897033,0.0171263327067519,0.2865970440117397,0.1677548495559945,
 0.26570808411500735,0.5697223648014929,0.9779590526758218,
 0.22935777361778187,0.31609492867284117,0.869858132308879,
 0.19884660121951792,0.11465957594952203,0.9617843332078183,
 0.12091482490849725,0.057226891886073106,0.5913264358024028,
 0.1247346483048315,0.7769371128222777,0.8394846564290772,0.37800152182432756,
 0.5759144890253053,0.9302802899081543,0.16285396522160367,0.3195734234723363,
 0.7086102967378075,0.6662090857164698,0.780509194349207,0.7838944660176338,
 0.7175249349174098,0.8875635886167232,0.3550110352758544,0.6636367921942875,
 0.44895919171686827,0.49855362894495403,0.9153583830560589,
 0.49974051782056883,0.945129369116248,0.7693039309372399,0.08886612354258794,
 0.5376977482761853,0.7828653779247304,0.4677141387970405,0.362296160012076,
 0.39090178988901825,0.9902725516071657,0.7153797396634887,
 0.28682619943544985,0.4813921872928638,0.8249181150615098,
 0.43918378612251674,0.33317466368098614,0.1818007029341323,
 0.8289371156428181,0.9395628548073405,0.14830603939319575,
 0.42692776422106704,0.4052328534184768,0.4371329880857382,
 0.07337070155870207,0.9962043896492652,0.3897600870961868,0.9160703372396231,
 0.4949659482065766,0.020496507415150855,0.4406652332704307,
 0.4060563836304174,0.2247453032907697,0.09984172114220724,0.8651143157043394,
 0.4455911867386563,0.5265772037609268,0.08232743184495639,0.6608947990190224,
 0.0011050860090786596,0.8040688159274061,0.8618325093179555,
 0.8853685057193891,0.7151280739984426,0.4437940302292035,0.11407015749185168,
 0.9407875910472242,0.9411764387809356,0.18908528586727524,0.9766323487055217,
 0.027689048897338475,0.37146423286874486,0.22088601699183463,
 0.09464468631021383,0.2707468067987434,0.6880455751099274,
 0.010846479627017014,0.4983968287054662,0.2189912613854329,
 0.25622331856672087,0.5039158390438219,0.12306918669442402,
 0.7720895654613034,0.8706298698639359,0.8274330481443121,0.4088609427039377,
 0.3827456569860449,0.9588757874910615,0.1758464593344894,0.6783783565560956,
 0.5563268954626219,0.785188349554822,0.47455041942039733,0.39254172800031517,
 0.9074299991364029,0.9994907008032037,0.12821880963651477,0.6799543317370945,
 0.4031282389618873,0.09161778150363387,0.9541297654499474,0.4995242337477256,
 0.7686305320821841,0.11770883770300267,0.06358285182084833,
 0.4108363344767204,0.8703815076588295,0.8558196173171575,0.4472388598999095,
 0.3025182407764021,0.7168153145624016,0.2836409884989669,0.5641861652356216,
 0.12476752633819088,0.6269762624320894,0.6268804168446857,
 0.22655552635921272,0.6731108157784982,0.7557702762383105,0.2404624428649058,
 0.7303139753813189,0.030889443714439024,0.430547262612371,
 0.014362843213670762,0.5614720560919926,0.001062834442556282,
 0.09901687698255812,0.276275747534385,0.7231255099663203,0.5741640554501418,
 0.23741613828537178,0.8814272663886403,0.15546866534871795,
 0.015950318801525643,0.8085573165392699,0.7192834139024384,
 0.41794643318802227,0.7359363139177219,0.9566734885913467,
 0.22880792795820137,0.0555342486687892,0.19216776555599757,
 0.8614792357634403,0.24606683435442456,0.7940331222216332,0.966394318080694,
 0.7302219635152016,0.9185837846198001,0.4674375122203337,0.6671600626787803,
 0.6504223487956827,0.0693842211916238,0.8211508987245284,0.3907890222735614,
 0.645670546235134,0.12368961028592129,0.7551844240193081,0.9518476230090279,
 0.8969214692714088,0.6463303773508161,0.9621885255471725,0.15775110132018788,
 0.6282774664325882,0.6752988799797186,0.1173122186928196,0.1676190064572496,
 0.08547822062390686,0.09189604524007533,0.7280371444056242,
 0.4631787064158309,0.348771014744919,0.46528057912139864,0.9397434631461348,
 0.894932074659245,0.7998005718857023,0.20637263706315734,0.7521169176424699,
 0.4888841767113937,0.10956721575616934,0.9268718173099743,
 0.02527643230107446,0.27865086215370716,0.6526018858350222,
 0.5807413206136567,0.5970361559277801,0.18910355627875053,0.5394224663711167,
 0.7822084081922099,0.817760636890943,0.6061995569915403,0.35180720220441275,
 0.3135923618899116,0.12823589592040607,0.2995718351824008,
 0.039434128013016156,0.47280135798779743,0.9618103497618558,
 0.6083381012794105,0.7467197632341394,0.8487835442898433,0.09088798312122714,
 0.5784533426557982,0.7912654166465789,0.6046935255398977,0.8977348376307874,
 0.4616001790370682,0.1853111169353182,0.9831286049728735,0.0647515317066204,
 0.41945117755653905,0.6253021615229019,0.33756458939665135,
 0.3289668558275619,0.9922302828670962,0.8378222936606115,0.5731114943470704,
 0.9762324254729842,0.5785657753655461,0.9116294685951761,0.28442082141042824,
 0.21711560046855538,0.656818712952678,0.4218117945949009,0.7561811938185145,
 0.6138046587042771,0.14922902168436392,0.7201562116047474,
 0.35993550515393014,0.36271645146774667,0.8898966086918202,
 0.18451887445825332,0.8794022111780159,0.488735947337279,0.9363872069430612,
 0.1522594970345308,0.0136773447148556,0.46592581010472944,0.7686202074676387,
 0.4623718760914155,0.41882736680983257,0.6761603252058547,
 0.41608533286191396,0.46023006779701237,0.08930907348961836,
 0.11110649693787666,0.8288808432106192,0.6980886088724322,0.5555050547208905,
 0.8853125435872251,0.22462724475553197,0.6419200543645545,0.7265747855379867,
 0.2762146207168733,0.5502020191767036,0.8126658384155954,0.5253459883111541,
 0.6887492746446763,0.6005665191489113,0.701505109159006,0.23132163095257297,
 0.8384832669760529,0.3890249233674583,0.3905614937420663,0.1275421928448719,
 0.12817976303583845,0.9239903564717953,0.5429922150509783,0.744902659051817,
 0.4230207879233755,0.5316149255806916,0.328218094693732,0.1436017738399301,
 0.9196817936236762,0.09159845061315974,0.09276035060670207,
 0.5425390907462264,0.4814331348365478,0.18775358840783163,
 0.48760657606153435,0.05871054172184942,0.6758706302209437,
 0.3115893915194319,0.5355054737225783,0.19837507691480183,
 0.33566001177010296,0.9114656266337096,0.5478824764496841,0.8567567843713803,
 0.8082039821532483,0.18925884139780347,0.1608325010869539,
 0.23737304073128362,0.5001505741335662,0.4878268757513783,0.510051301172631,
 0.2977335675082532,0.592917336016046,0.268463999258863,0.23965995240232185,
 0.19732083176084436,0.5442189415909364,0.21151686158748062,
 0.2807929105393895,0.9570242198160526,0.7578101206001907,0.8159381617375793,
 0.6511431863026219,0.7618592150379143,0.9154602385347596,0.18158092125810898,
 0.034572228971034846,0.8571581894588155,0.3371794773030006,0.377638463105225,
 0.9791280993288086,0.8411864679892086,0.6788267916863564,0.3195087836047561,
 0.8280456009892677,0.7477576038897904,0.8176398980703233,0.16197447187034153,
 0.05803262570636203,0.6488920488122429,0.03999391039286637,
 0.4068129191528277,0.7768982634475079,0.2677974052570259,0.535924072933678,
 0.015228669872889955,0.6564006970935882,0.3726389623199975,
 0.5407753214291977,0.4919962984904892,0.4825704807729778,0.32767286698133025,
 0.021415581478552825,0.41050824104316264,0.26784396592531834,
 0.7361674990977486,0.4094528034133488,0.39026027488350934,0.7149920192254918,
 0.9398642513275067,0.06154097912335443,0.6777255986580626,
 0.22165861715775803,0.1269900362174845,0.9486545060153199,
 0.12750802796494842,0.3473079315999251,0.023123962593376235,
 0.8315038732239343,0.5663471436391395,0.4442807206970367,0.91477489758632,
 0.532233720104794,0.1363172827931146,0.12791127613569198,0.19202281481459038,
 0.38636441920166265,0.29509141801818883,0.8796114604150692,
 0.3519060611197179,0.6342636091447892,0.4563800449030845,0.9453876869129851,
 0.7101995752629305,0.5085263793860846,0.4859545358666082,0.1786031183954736,
 0.7693850959014142,0.4479768854281796,0.3283393482701371,0.2532147845751058,
 0.5402142549804374,0.4767033949130898,0.7603739732425596,
 0.038656836164586594,0.31149118671860276,0.6018515897005664,
 0.5141674874824301,0.5942603036296262,0.9252675434399258,0.08624908659352071,
 0.5509841318315761,0.45592864625015417,0.08503650331531043,
 0.6694188688020467,0.4384530611346501,0.6200040628197074,0.736998456565348,
 0.5047973202432454,0.7219004240736343,0.14649296763364128,0.4233107075940067,
 0.020422564713146296,0.8553780346070216,0.4291374452236518,
 0.9657784275029079,0.6278601741021027,0.0585854821039139,0.07790271008975713,
 0.5972649087083307,0.9108229297453154,0.1306551221530844,0.6119814556099595,
 0.7728677185487672,0.10453599304604555,0.12339545723493517,
 0.8827230305778954,0.5848055236794673,0.21129825561972082,
 0.42641277105061137,0.3261746742252647,0.4625928176268348,0.2637952562845072,
 0.9099905720211451,0.12099375053195338,0.23465361412238128,
 0.7557665486576803,0.6807967628140761,0.6791444758002846,0.883040662813932,
 0.8877761417319256,0.19287556931270544,0.23852210724751488,
 0.027129728065552805,0.3871675435010561,0.62960887060147,0.23775146312346862,
 0.8791636899419437,0.45407163238148207,0.13488014763822187,
 0.9692125315705735,0.020866281206176396,0.779983490705722,0.872743870509613,
 0.9557705390461275,0.997557800517017,0.1104550991233556,0.2540245785478318,
 0.2206322767914347,0.5320631101799408,0.10347166922553797,
 0.42851453079264423,0.8148670910659541,0.7096992303294234,0.6195357298523699,
 0.12397427557091012,0.008327228675403697,0.6742783863386215,
 0.6571074020569649,0.6960342503060889,0.49389597370459437,0.8713616645553892,
 0.014695076912514371,0.1031066576619527,0.37696308947750434,
 0.7101993065966163,0.012391471866995407,0.9153883290547296,
 0.02791157522098886,0.47216074677210496,0.8949171357860891,
 0.15555760741137303,0.3138415838580553,0.808564462015293,0.5870219787868698,
 0.6340837011623072,0.8861876637030693,0.15777150201355195,
 0.34497688619699485,0.7424842624791098,0.9813750172499185,
 0.16148796614805772,0.4690427711893299,0.9764941553000192,
 0.12488963828570865,0.3740959319428403,0.1591959038491032,
 0.48300011994836245,0.5519986996865004,0.28101448224524017,
 0.49528622342550954,0.5852518875733235,0.6897030948340802,0.5192347636967064,
 0.8355910226130622,0.38467789753481163,0.1768703073309169,0.4406040426952915,
 0.4208615438410035,0.7938394136355591,0.4695265512799336,0.877164218317477,
 0.08270548517234033,0.6065230133507399,0.13260818241765115,
 0.12750427810990184,0.559430562309201,0.6816673719818898,0.28880622074285256,
 0.9160611248201598,0.4461496730883717,0.09986521408597604,0.5910732414229266,
 0.964346399725389,0.7784301249582665,0.25149823033218277,0.3913779190198877,
 0.1982652117508461,0.09520887369724107,0.06243941434292721,
 0.1876710499609825,0.6524194013859201,0.7934734727165468,0.31817723666695863,
 0.5192344568708513,0.8037955351877978,0.8333020463712705,0.8947530983071146,
 0.4178194359193499,0.2867831468172698,0.6373481776020805,0.33767758015459703,
 0.6468821745259513,0.8998364645242065,0.8516378933074432,0.3950760692198889,
 0.17672343936685508,0.25667319533245847,0.08856488309695665,0.64462948837291,
 0.8894281934769699,0.4355855024616796,0.6792218184416922,0.19923500335832178,
 0.5774045820304182,0.3618403262956599,0.9959842712917099,0.2707907681814907,
 0.3627524635859183,0.8873442414752738,0.5331026819818271,0.23151046557746413,
 0.8934416667640644,0.05526476578619688,0.34919805509144863,
 0.7231654867834932,0.6414523634095166,0.2736026033152237,0.8003273517367687,
 0.09271294349301962,0.7482082555136128,0.1446941933584669,
 0.16874445558024986,0.37827921438069945,0.7544224394210399,
 0.24536343027101726,0.8708000660071332,0.8527004871689084,0.9253199293778587,
 0.045590659091913066,0.4349088624079762,0.33208938388954057,
 0.9114510277029726,0.15306464616318594,0.28747605133150667,
 0.1869613889759889,0.32729451554447897,0.25758945204493466,
 0.25514232195478836,0.6775098617421599,0.10078692042723625,
 0.8272736922038689,0.3448622668091975,0.6025262281178262,0.6249480709652313,
 0.4674950186396569,0.602349440696693,0.8747236364280294,0.6261256999409059,
 0.8518541099194327,0.0082283801777846,0.11084491523828854,
 0.15036162573344547,0.8705048784380021,0.6233420780631023,0.8048985767770701,
 0.5307586402901309,0.8303640952143477,0.22660050589443703,
 0.33730531227866667,0.9267077069277844,0.032396278973126136,
 0.48212609276085994,0.6001491921504641,0.6332063156811893,0.2645978974123028,
 0.8781476624815573,0.3767989869148365,0.5241896579757727,0.813962002013593,
 0.5253938385905581,0.39267595504070496,0.6803118795264935,0.2683504782464281,
 0.9871540762061289,0.21425728027864177,0.08493755366655131,
 0.5185823405989568,0.3094564964020363,0.7360179575495187,0.10981519703098896,
 0.9042224344751476,0.6909473599768964,0.33960302595882397,
 0.18695741601612792,0.870177188045265,0.5534792713703308,
 0.008888974400723226,0.5153432610390035,0.842344138015154,0.7816782452499419,
 0.6939252284420468,0.17173432597515503,0.48261939643818075,
 0.42662635658343806,0.9957334355076826,0.9246579893892245,
 0.22214941519187814,0.8463359884468294,0.33732738269544216,
 0.20486360034958961,0.01989278112254378,0.05563647931850235,
 0.14003640776397708,0.9305088358706722,0.9349106293715547,0.7690434142773636,
 0.858545838366634,0.8722143054728774,0.12034435568808965,0.7306636343329733,
 0.16196474804759664,0.4604311653003883,0.5937242507703008,
 0.10930818638208495,0.24111949171338343,0.07933285172745486,
 0.5221258917108597,0.721810691541493,0.3241678411812918,0.7561301362135988,
 0.43431513240817365,0.4583164662672229,0.9123325237780654,0.826605826593186,
 0.747386917916943,0.6769374424587669,0.6672295465089656,0.8895694049572969,
 0.3818150715479053,0.8741829307190068,0.4285247874062107,0.3252040547656706,
 0.2682205386483957,0.47875702052288127,0.23165365613752242,
 0.2286295117783378,0.9797195636218174,0.3836147740257142,0.9469675196549954,
 0.30167735458555467,0.8651647544745753,0.7798672219963791,0.9998476507374051,
 0.6251959830943276,0.9543716883691026,0.3434062899098925,0.8819153857000194,
 0.49535171383522936,0.4805134012763871,0.372914893569378,0.17657784375171948,
 0.14383053584059213,0.24911882638781024,0.05073377724138495,
 0.8487914283938209,0.05074961038464265,0.5832416724524296,
 0.09693554192669618,0.7864343337620955,0.3328292011416576,
 0.014127514624698856,0.9320902510250595,0.3772473536959159,
 0.07636424598913916,0.43397389674151954,0.7757496694992437,
 0.6843853638629487,0.7776883333622904,0.6819193765926441,0.3873320953858934,
 0.31738617885063414,0.21174579260699478,0.37818702667104165,
 1.5598907861780198e-4,0.8688417181751418,0.1035083594761399,
 0.1277710057354544,0.9628328704402302,0.9751690086065994,0.5705715273938079,
 0.18770826115721473,0.8243640208070087,0.9128942878332514,0.1845961483174121,
 0.018672398561949022,0.44468459827996687,0.7346781369331422,
 0.7994737356353449,0.5043101742516781,0.8604072266532763,0.2898856693833606,
 0.7471357136711642,0.15212466692413007,0.1339382031082229,0.6136140737506864,
 0.6500318997542551,0.2468118044758354,0.16488438283502083,0.9388172407161747,
 0.13760111854100043,0.7550648313442411,0.4340533533377784,0.6308217713418927,
 0.6055420625194664,0.03607091192758327,0.676249703962118,0.2611612124550501,
 0.5276316618639845,0.9402406203599758,0.8322998466928795,0.7268846481644768,
 0.8317260485940164,0.2929910121157089,0.8563507747902543,0.8634151912347368,
 0.2639065795035441,0.4762185237930331,0.6808275161921874,0.25344416635545297,
 0.16912731406480885,0.7072494612519052,0.48576686233612154,
 0.6317298943490612,0.19676506136202654,0.592004194150607,0.10928965210831354,
 0.8427040611896642,0.5212167458850344,0.709903329224089,0.49359093023612255,
 0.2269724464673495,0.43795490646174573,0.6412888752100006,0.946273915955972,
 0.06969886263955016,0.6718331794747211,0.35079848254219403,
 0.3690252333024723,0.03545984595525886,0.41893169383867723,
 0.6798024909381627,0.7616190990338048,0.7454320755596684,0.4718378083455985,
 0.051795678348072194,0.02851737386705766,0.3200899093692632,
 0.32849869134462684,0.572743301843447,0.5720033298526206,0.3283555794189613,
 0.8628482175133673,0.19413158108806738,0.6958264875141587,0.2550037986707201,
 0.8639566060531039,0.633222698872336,0.837339451643773,0.680924179171925,
 0.09523346485595285,0.35078982551335947,0.9296968728967437,
 0.6711552329551278,0.9042413039995597,0.8591062728583295,0.5973759169350622,
 0.7111667371530683,0.9946006429969652,0.9704766293675379,0.877963734927621,
 0.11701915481070091,0.1568583157664435,0.9077740905006325,0.4882554680809298,
 0.1756971060392939,0.8202558282210738,0.3036334209591084,0.24801758110502536,
 0.685362539727529,0.6604190510483168,0.23963074070273582,0.05632827733543089,
 0.0026662836879749463,0.4876096393707914,0.5830386668545846,
 0.4733251517045671,0.39832195984555163,0.09317868569161147,
 0.38948569728949356,0.08349560260142741,0.9425176181221997,
 0.8286783338525008,0.28340014577091155,0.37456704754163384,0.198168526741496,
 0.049528695875229234,0.5865819453675598,0.573035121500528,
 0.011806689863043007,0.9049965004094094,0.2027920278245472,0.741667953954486,
 0.6467861571455088,0.47257329614204857,0.15483033834258164,
 0.11237360335169344,0.8118650635226997,0.8112591315448879,0.3997652000741566,
 0.1394734657502853,0.9205573687948887,0.5627953392895901,0.4438840652450389,
 0.7477054794840373,0.2799573434564081,0.03707118029306311,0.7376702180526638,
 0.791059167273561,0.6870169351796772,0.8676444695784364,0.8710012059080945,
 0.7128270349211531,0.1396627477618243,0.511225504260088,0.7888562213253589,
 0.6295308020661898,0.4584208508678955,0.7988938027711151,0.9879162363324557,
 0.5218873301648675,0.13622849812359683,0.08760946217484389,
 0.4478316412847463,0.408189811108871,0.8383730715007509,0.5571086113542285,
 0.27200981636874866,0.9502312476990433,0.6433810364886945,0.6186189233914302,
 0.8091735409366041,0.2115543776412484,0.05751236910083235,0.5939252021201707,
 0.16697687326955213,0.4992787538976473,0.05610909594564817,
 0.7214098401696843,0.18445879908224472,0.7596369985394571,0.2460386935353145,
 0.07167949982822885,0.8563971335278807,0.43167608810959146,
 0.026884566356344508,0.33076606506714146,0.9514675709364778,
 0.8334246211243106,0.6764422585214735,0.5137375882849203,0.20382140498038992,
 0.4243510849837596,0.6674707582753487,0.350946904046749,0.866217736741042,
 0.7408207903670746,0.6659280901540182,0.062195618378217965,
 0.7368536898509701,0.6089583932634077,0.8862275819990697,0.28532718432079696,
 0.9128466520656896,0.11844909509228396,0.7137273528194022,0.6305731742603274,
 0.5244319504382056,0.12346668783590464,0.06575386492799518,
 0.16225795663262366,0.5993773501614739,0.10825451227626337,0.332850901675776,
 0.6151012163246528,0.7280165256605731,0.6538746153899673,0.38644146596741635,
 0.138841292775717,0.16689618170489529,0.7314453419240434,0.7308047790611012,
 0.45657567213908623,0.5750723499774627,0.3581401297630118,0.9612101858514329,
 0.5902892916151184,0.09695168819759425,0.3235206506017817,0.1487358082620479,
 0.5244202498078505,0.33740049694992313,0.22474339299638046,
 0.5008167478143407,0.7267659486533355,0.6992085200275386,0.8411744182878619,
 0.8937161301397353,0.15427303498996392,0.12451398563641702,0.258155205273517,
 0.4123230842601684,0.597799429755245,0.14795853077898846,0.7678775513696797,
 0.6928018230591314,0.46260146284961456,0.6324683253062244,
 0.40365850157200156,0.08865341592042231,0.30271353616038577,
 0.6490693325123046,0.6696049640990881,0.7442178476528036,0.2495711984238128,
 0.9586323366333991,0.8395976042321961,0.05952972948472379,0.8919100196919386,
 0.11725685548550346,0.1346678711282141,0.45295361958237756,
 0.042499996829154285,0.1557747376399512,0.82892711379803,0.44141816145353563,
 0.2968191564752851,0.343768223485591,0.15660887351518038,0.33201634267070534,
 0.2874056265853687,0.3982304020012104,0.8662679543761922,0.9498270509192603,
 0.0567582559564902,0.6328244306967294,0.5263563225079884,0.6841818123590551,
 0.9819307065676381,0.650835464167867,0.7388856119037581,0.9818441651699594,
 0.09237708135114397,0.4063618474746209,0.4720788592098316,0.9905566728287262,
 0.4859707699768576,0.3672001091150783,0.8929159368415516,0.9607800380322247,
 0.13170766816271462,0.3887009834799181,0.5707234338591062,0.3733894031592806,
 0.8071457141628963,0.2847713109754103,0.09902386500244453,
 0.40849466208100405,0.2565400395480484,0.2900255924685464,0.5091371403424099,
 0.22449346829512418,0.26213632056469405,0.8913953976294696,
 0.6562630323156627,0.28790668171709777,0.8081475976399641,0.6276156143256775,
 0.2455459287057209,0.4652932540114053,0.9645806870495132,0.2705403282656589,
 0.9073548200124274,0.1775887545007241,0.46116341394287186,0.7352987081989919,
 0.23857838564470235,0.16356517564863082,0.9675569205002205,
 0.1343010123516879,0.1984683782471618,0.842607725666205,0.29351119825407834,
 0.09712218030614095,0.3120649052764347,0.02508815130759967,0.945366345063089,
 0.8496720997438173,0.7678029760758808,0.0894745352630203,0.8219368457599273,
 0.451055900235271,0.3464533585863494,0.7687137086458289,0.31144320818219806,
 0.602315999169769,0.12412795143589228,0.7382577823162092,0.5364449861738952,
 0.1558813529732953,0.16196468364451688,0.6127449345993534,
 0.16224534701956927,0.40350502389000553,0.3618698967241123,0.894047733298169,
 0.1905287417955953,0.27029087754062253,0.5049751910347484,0.6110688518295821,
 0.4683788758758902,0.36957726528343315,0.6180230084387508,
 0.33299310302515206,0.4615508121753238,0.3800046551792694,0.1776200471495223,
 0.037091197302637546,0.47373223176303725,0.25253541219868425,
 0.2338571078416446,0.6659612865383336,0.3841620008831642,0.28567897246148544,
 0.4846324034227212,0.9502971387776644,0.8331999170869634,0.22840554346926323,
 0.9318438540809589,0.16840138287225803,0.2063855267312923,0.8213584034173878,
 0.6914445033939605,0.3380411102869222,0.13541037748281393,0.9633054633000159,
 0.3533887349158844,0.38764851844077763,0.3185247965603464,
 0.13500807518980407,0.025989876578235593,0.9005905943743453,
 0.8711282902815611,0.33450267086331176,0.38361668942417393,
 0.32956878267433376,0.836543668479181,0.6833964407065816,0.900346942714306,
 0.3837586438819478,0.9325558366975336,0.5673411973373057,0.393241539322028,
 0.07964679922000395,0.7344135621166893,0.832773794554724,0.546797643018328,
 0.7928520919013659,0.5743033854142051,0.35142954903304546,0.137855462138319,
 0.9732051319513965,0.9136711615075838,0.452512197639644,0.17105012399209407,
 0.21138765482607513,0.586985474729075,0.11698803296387905,0.131857181781467,
 0.7395184953752182,0.9865818709822242,0.7070353481643192,0.5893250169641284,
 0.5373684348989054,0.03692115881229152,0.5300172132194947,0.9938065406430242,
 0.23583380973750467,0.5798305210704855,0.9933330859550966,0.7876786755509193,
 0.6988074951061187,0.9426449948752236,0.3455491633463289,0.28509868931398763,
 0.9473138241862828,0.7444048267679262,0.106486534850593,0.15118113893096585,
 0.10814108891438334,0.637172443450656,0.3434935405147985,0.9012425871826509,
 0.6215322999141945,0.2299100489893211,0.8854169197770003,0.6965883398977814,
 0.19510408540715496,0.8475388791139401,0.8874121148824541,
 0.007632605634369738,0.8033128781485732,0.16719393867503196,
 0.5176376041201267,0.8405142240720842,0.9290248568071193,0.5686233674013114,
 0.01339491788352698,0.04485993397969812,0.780040502796006,
 0.31271660188355566,0.19915215686581644,0.3859751586314014,
 0.7058288275125747,0.4723850423324675,0.0011746239334410324,
 0.5922203436917115,0.38091961186101053,0.036540691710240325,
 0.9007367466872128,0.7852584166801202,0.9447748580701879,0.7723673993861853,
 0.31663650703739155,0.3520470054817959,0.2536341916247431,
 0.17672512219575776,0.3911849675185892,0.8261103100175088,0.8400673458340221,
 0.42343995645825183,0.3394835198922539,0.138726634871566,0.6695695711670462,
 0.9152385169079672,0.44381183498156584,0.7262610659596304,
 0.39085493351416734,0.601737799544645,0.12802851152565842,
 0.20395684127558267,0.3170129253735139,0.192802797911251,0.8276483744639133,
 0.4450573414397193,0.6796662006261976,0.9038306587316178,0.20103662856574755,
 0.02239426891418894,0.7796358272437386,0.1411081143970776,0.602859001373041,
 0.574769823930302,0.5175323929071595,0.38242853416936184,0.05977993816803295,
 0.8886751873317379,0.6620133439136463,0.6652301969225822,0.11775378944366521,
 0.01522841567403932,0.7913023441303044,0.20945051815189752,
 0.23045231246077236,0.005796700265180865,0.8265445601423662,
 0.6575541286217658,0.7573139706927579,0.08774544500616166,0.6358972052191559,
 0.22819592838478098,0.01840175056632254,0.7170372168329464,
 0.8491005734891603,0.8554082527326619,0.11571517776838913,0.1354159845527917,
 0.7137046713776601,0.13889211769375032,0.4319396880774128,
 0.33600080470706906,0.9721309304740753,0.6802919116664438,0.5339973287687702,
 0.3197829573664278,0.9102573812806128,0.11826690962529451,0.5363083658402805,
 0.17062287820389743,0.6737301984948466,0.13523045044191184,
 0.1803439648941917,0.24778429697165505,0.5792794750944736,0.401696702037055,
 0.3037520534694308,0.7179561246203621,0.38851381247233596,
 0.23752025517956665,0.08114294341219841,0.27030030139647454,
 0.022724555396975377,0.8716779382050397,0.5012359939055512,
 0.45766029879822356,0.03672044895097659,0.397073311508937,0.9218758170552439,
 0.8502035819948123,0.946693727922076,0.6625177754164524,0.6302455715102061,
 0.9339206842956758,0.046177873715046336,0.0988233285415987,
 0.0726119396011361,0.8599325392550733,0.13421392399708776,0.9753933815586148,
 0.9923154465786191,0.500102855855842,0.2142786549685488,0.08615001851397808,
 0.156795279607151,0.09644639917311548,0.24730495991594448,0.7475880817364696,
 0.06593601374894653,0.3221640270387822,0.02386400305102354,
 0.5208396498465209,0.8909762194275558,0.6747858150835382,0.924989807013,
 0.33599189163210086,0.36006970825860607,0.4491607223274412,0.386409423074618,
 0.560040807939111,0.033655980632935734,0.05976066073384567,
 0.41353700519692815,0.8701699748498675,0.6357763713684613,0.1478471321074759,
 0.5438412499006154,0.5740547529699972,0.11640989204082364,0.7910802062363058,
 0.14055129236692465,0.7661826293002899,0.25051501417977307,
 0.6508692266753613,0.028254810307904665,0.9648105901983237,0.865440988161347,
 0.8000702049689736,0.47853025482268763,0.4738904405961556,
 0.09479524597400846,0.26656295162254917,0.34085697398627135,
 0.43906628779837853,0.08865326921943473,0.7941124529138557,
 0.9281683171643584,0.09157847732925128,0.5624704971992265,0.5412782214828203,
 0.04833373933072227,0.2657181662968171,0.42548914681638983,
 0.7232928591374195,0.9034153959034166,0.4909254661125432,0.9468350952412605,
 0.4116399307986761,0.03706179455769232,0.20914002881271587,
 0.8923959435420402,0.2203316299549023,0.15830378999861727,0.2794644440147209,
 0.005471067047748135,0.9556987129590945,0.9198813096800357,
 0.5547907606504336,0.25376417157992615,0.9228402658598838,0.6454052922381001,
 0.35226868909335973,0.9726197367915679,0.1899945059548499,
 0.043299544900671116,0.9688842238479592,0.6230422332040291,
 0.4108147828322922,0.23325191433409354,0.8004847698517917,0.4131745246858821,
 0.7259906775949557,0.10058390911701942,0.3893503181833906,0.7995099020706286,
 0.9825561237410053,0.06568372307593529,0.29400798056252553,0.662789318944175,
 0.609030408071939,0.787997924907963,0.696974225792519,0.03644391634538502,
 0.9346226236229902,0.39945005478490625,0.23564876924600664,
 0.5498438165542633,0.4245485961247356,0.8498184351395887,0.6665318173369662,
 0.3906233247118601,0.5062885645214088,0.20624227303480525,0.1763578612067258,
 0.21218525447729297,0.3659546676031733,0.38051660450760627,0.997574993994964,
 0.7168630273748964,0.579919664382848,0.26075858267319707,0.19533980624691516,
 0.21586307755476342,0.8330318147786279,0.8091016687776501,0.816473119444641,
 0.22782232219805854,0.07463637970077386,0.23088085580060036,
 0.3638550598308382,0.7421538968771754,0.5437863802141474,0.8051348020951854,
 0.21020533393031005,0.24493908730474523,0.2892455402916633,
 0.41326916254066226,0.46851324138773154,0.4205639258308286,0.720014175042442,
 0.13685790190292957,0.6546038779455204,0.27959220522896366,
 0.8208713024134878,0.6556148008488882,0.8998230799734017,0.04650046883528802,
 0.9705545152057871,0.09587306396833362,0.9827869085566715,0.8185870935122523,
 0.41598924519383496,0.8285635832524183,0.41732413811910274,
 0.8834288515062576,0.6304118348856247,0.5113749401595625,0.6508638674233458,
 0.8826985154784845,0.33630032706705393,0.3414208684797705,
 0.09335822303927699,0.4365918249392582,0.018380074732303076,
 0.6481194208487071,0.15266419020360933,0.20084563538644862,
 0.3925030629042636,0.48125951181560134,0.8749261837532495,0.7617462533844611,
 0.556898962653521,0.8132694986886269,0.0231941697498097,0.953941715086755,
 0.9385439418513974,0.7244171435949169,0.9810078156712839,0.26825428255755146,
 0.9416763792845951,0.7335428583479227,0.6394560318585872,0.3241225634375874,
 0.48380809650682677,0.5406371250712178,0.2628259440570101,
 0.18631685394662778,0.12109882940473626,0.3626492040920666,
 0.7491925444855702,0.5544764724438433,0.04459491095434265,0.1728232839964008,
 0.3400909174761477,0.21667619686721384,0.6051725310761656,0.2743931932788821,
 0.41558825018858425,0.11093487602347318,0.7206054300700131,
 0.16117849777347537,0.8816577668763409,0.07373858764826191,
 0.21487298317623316,0.22759760918204697,0.38081771626746985,
 0.4895243566733497,0.14917596700896163,0.5811367234285939,0.3930277247367425,
 0.9794408782783729,0.9152500325546025,0.5813489899502929,0.8018049692181499,
 0.3910140572520827,0.6003742514010398,0.2302220631512364,0.6527096366994638,
 0.034159563382257785,0.8577689866034576,0.9170213804569407,
 0.0806258723929284,0.23537638602317723,0.8033253181248421,
 0.14255022483929924,0.12737293938600447,0.13004401981902058,
 0.15476530655799536,0.2884274389021375,0.21103642227925523,
 0.5222644815419472,0.46253622197004873,0.06207763361237717,
 0.9787310417113161,0.6891554505991967,0.7196571468179288,0.49287552682510605,
 0.6858566383877721,0.5700057107612,0.23343634513684453,0.1839878402882934,
 0.15255081433039896,0.7930595194247003,0.5941431577892913,0.831352054094387,
 0.4598272467851169,0.40840218102155623,0.025773464947868252,
 0.5605344857870818,0.15851070444046456,0.07149123453858286,
 0.2983800078150167,0.506987403304115,0.4169030321797508,0.16032562043977516,
 0.7705436950172099,0.5639251097869771,0.631332767003365,0.10045586333383327,
 0.5459030482232499,0.6444155077585199,0.9219120932899002,0.709609064321309,
 0.48801938775306475,0.0036288347938775445,0.9842796652935976,
 0.5448945706019994,0.032537137223395685,0.63048642556987,0.7816258881260074,
 0.3729219814141501,0.514827954062951,0.22851656373423557,0.47202031075280837,
 0.19568231007113235,0.5276016501878307,0.020402018619369855,
 0.8880650937175799,0.20726106663383148,0.607722203108897,0.6174727754673965,
 0.1070022618554497,0.6989369775579493,0.39951590795976366,
 0.28681318917766085,0.49373887369442937,0.6540072555786068,
 0.5853606269514007,0.08569474498498897,0.06659794389030782,
 0.3911557640969414,0.8783242347429455,0.356647712150555,0.5663618478463656,
 0.7960801686792442,0.9951354330803935,0.5033738098311815,0.1838228654337708,
 0.752224061602186,0.6486704417706154,0.11633387074097645,0.7892182018062364,
 0.34698220887934306,0.055043087707640326,0.46312295140800386,
 0.30813818052339825,0.9685793442626274,0.7785685838455252,
 0.20346899051436618,0.45068653965185435,0.422371801669831,
 0.004871865208999404,0.8243677604628046,0.5697362906249129,0.588571175282254,
 0.48910458522072575,0.42282896433907546,0.6728534033813319,
 0.8230020301511314,0.5293858116821145,0.7788407079479887,0.49062425105339336,
 0.971635850514294,0.6980750114363071,0.710307660038224,0.21901718521766866,
 0.4289923909593809,0.9418065331481822,0.9405179703156579,0.284791568813617,
 0.7113356908945847,0.006698811668827354,0.9824501374515978,
 0.2628389382936306,0.7831590104806775,0.8047173423637634,0.6687058200983614,
 0.8115275370240987,0.31002926432003575,0.01407546291667705,
 0.3764718210886866,0.9706610985039462,0.29002199643858173,0.999374016174121,
 0.1977742175055136,0.7915386304787894,0.7817985108890664,0.34305075666394935,
 0.374723067855549,0.3905559190469704,0.9763393071415128,0.7187488117278686,
 0.8748409567440728,0.9520689757062266,0.5189518212161257,0.9121955227612528,
 0.301657148968281,0.8974391244058757,0.45699491659143066,0.2961157077990254,
 0.36319710993993326,0.05760720800928709,0.27033664484847053,
 0.8775022269430907,0.3000006121041169,0.5088713978166932,
 0.007146239639507668,0.4328379868950838,0.5436076237326953,
 0.6550782696769808,0.4242743071324365,0.8613438549199819,0.8610745728442563,
 0.9845760163108722,0.3204061005457942,0.09510005579432712,0.8317745446894164,
 0.6125733574141029,0.9882688730229602,0.8079367009969536,0.3683737412408131,
 0.5137778354340916,0.8884461524964006,0.35825099978170116,
 0.38192740428717065,0.090478104067097,0.17326436625568942,0.8554976383597521,
 0.20116801100833426,0.48323532071964126,0.8616884549825304,0.789871153469034,
 0.3117683681794723,0.3215460488191859,0.5131575252843703,0.7852998546424381,
 0.4582629590457342,0.8804895017988869,0.44326187527428573,
 0.43076158086125194,0.8476438188133272,0.642675706965679,0.3755101977092695,
 0.45919601828142187,0.7075922328880142,0.6900886638512387,0.1585883805693573,
 0.9461057730505822,0.8767769568617302,0.6205621975613352,0.4408101343253116,
 0.23897961818183733,0.49242507959031023,0.68990384015987,0.8129448746281036,
 0.48537011443366007,0.08672629698595535,0.4951294927839339,
 0.15622919305893734,0.1397911035433873,0.534055708748758,0.6489455733034184,
 0.7926207873411608,0.42181174176145964,0.011059454637576738,
 0.600081517140792,0.1848805120826017,0.8189663370431558,0.9177090202166837,
 0.5853450265439955,0.3503038944695389,0.4559115568864667,0.04164941970861191,
 0.43604326122724657,0.6553461589557524,0.5607917440736241,
 0.47635790954109447,0.8290754480082946,0.8394402491128308,0.8225900336633748,
 0.31131278385103944,0.5184960880173098,0.9188100556887271,0.8894064826630355,
 0.3341005642499558,0.4232842355239086,0.3698537548543649,0.2838629466576237,
 0.6194816514217869,0.7282901510346818,0.2128028422359065,0.7041447182830792,
 0.662138080630171,0.2831447274017619,0.5395908099437923,0.21447231244506793,
 0.1857242041109971,0.4510414361970263,0.1958592706019685,0.49271402061471403,
 0.7033195038162198,0.5536250304852119,0.31411912621332216,
 0.44367234438544656,0.3626151091885714,0.8001172216251436,0.3843808819330914,
 0.015023229146711059,0.30757769383337563,0.3429394627048029,
 0.3410617212047563,0.1903571445571688,0.4002077767006492,0.25526537894055823,
 0.15921962618025898,0.8198446099299093,0.5076223191122701,
 0.13127615991543495,0.607837536215636,0.7770688202472442,0.09810377062451403,
 0.7057162178429865,0.8801576199896433,0.06714730466983232,
 0.20619210712252356,0.6413346950435734,0.2615998764567976,
 0.32024832670051184,0.1383229968692835,0.03914047107131602,
 0.17268413836184826,0.49799020564577956,0.4196801261506041,
 0.45211804653213583,0.8827765505686834,0.9914578626893216,0.6320057665483462,
 0.9946537657334571,0.2677287522300359,0.8690896876044997,0.8761085345453324,
 0.8408338656186629,0.52269782074701,0.7519238993471692,0.22827650371046915,
 0.8013401321055154,0.2289141103701512,0.4880647882566913,0.904617115194561,
 0.3838401031656886,0.9219592852295599,0.7292562740723056,0.5734381162396236,
 0.6356338261047152,0.16231372964227442,0.24953683257117654,
 0.22732988702739687,0.4397418403426826,0.3491509546202024,
 0.26830501015392727,0.9448720726583517,0.6932141820552544,0.8419934683993511,
 0.7309808750555754,0.5038210291858123,0.08262069955535134,0.5969466413904859,
 0.7753508830534365,0.5371005251536811,0.7891465371706574,0.03893638126712806,
 0.9282870111394919,0.09813465225032847,0.36200113455675376,
 0.21237341683980793,0.2925637469979919,0.3081691899424863,
 0.42244199324479914,0.9969221853529769,0.942724099700407,0.31894878171970986,
 0.6007307840985642,0.12923244022330005,0.6665216603101182,0.9549455905781116,
 0.24853260858805926,0.05786609603536774,0.508836858661045,0.1430284823161565,
 0.044003424246787604,0.3783085439009122,0.2790364890765784,
 0.3736031410635344,0.026574506937285802,0.9581353409296576,
 0.6038879858774713,0.606768665216683,0.3491056681844462,0.55309502719467,
 0.9025874009716675,0.768360322293354,0.5592282139377669,0.5651554640906465,
 0.43731550384765294,0.5518182183837806,0.4153890577355961,0.6667691414757421,
 0.28933402581962686,0.4261800497470787,0.2867123437611683,0.9950917686609079,
 0.1541888654487774,0.4804980585997941,0.5206593204540253,0.9435569366618112,
 0.3092458877468207,0.6107289192262038,0.540161546826914,0.6852369302385855,
 0.6965377726921151,0.7823076087580032,0.6221954868922995,0.48937464802588837,
 0.9653761215037683,0.8971413807792054,0.6191799366345324,0.8743583021355215,
 0.1884361979425745,0.07650578390309215,0.41773868879537657,
 0.35328863173006353,0.19129789609226044,0.6337201900051911,
 0.5964986824276224,0.33772514586392055,0.2647402644958926,0.9860384675711831,
 0.6470236220921645,0.20807041906199397,0.9552424784218384,
 0.31770209179074016,0.9654613401665285,0.043226842779583774,
 0.08886735166925541,0.9115503372033216,0.69840206014431,0.22277750686971576,
 0.9904349591028656,0.9771340001194289,0.46764222386166787,0.7654564707487955,
 0.06510184290194476,0.07703150570918504,0.9965812122741874,0.547095792648716,
 0.10921015100921294,0.9421545868623282,0.3080841709219512,
 0.41969737090304016,0.8678130987139427,0.12205972577095292,
 0.2904094157921383,0.09180881946085606,0.9472463741238137,0.7853518219716953,
 0.7424703810404817,0.1380328175771579,0.39615941912645336,0.7815081592394992,
 0.8448397222343125,0.06229179757724479,0.3578345548114895,0.9125240206329882,
 0.8076017756225905,0.14408993108373813,0.06699717413556006,
 0.3943806833177045,0.402536351082332,0.9754152874486578,0.48896426964210893,
 0.037709030572755764,0.9437255277412486,0.1717498553198864,
 0.9635335613264735,0.8966403819283836,0.8191537960008536,0.8899049557145381,
 0.3621618546711761,0.1441062534815637,0.35813307120563653,0.7383842206065601,
 0.23304865477589587,0.3661731965251821,0.9382746148333868,0.6878191352202572,
 0.5458470343270847,0.9134536949359772,0.8441899786925282,0.7359532579196176,
 0.7524760681316501,0.4165503125169845,0.34059150464772103,0.2324960033187371,
 0.26215151474330656,0.3923474682749136,0.3010761331955448,0.7866506393840664,
 0.19527483764434272,0.8356013493561386,0.3305335551130534,
 0.034511725288767114,0.48969144454546476,0.6224013052188254,
 0.7269438019354342,0.5204835616611867,0.20566268541168542,0.5819972306184656,
 0.016990313871408924,0.0832421098938475,0.28154155677682446,
 0.5152033771975827,0.705314805233385,0.9390405246156411,0.16418646341584797,
 0.8153731178391169,0.8546818345813386,0.06490510708213193,0.9250860356889914,
 0.7589816039380161,0.47394240425233314,0.9849840482048782,0.7761595160659204,
 0.46163388549992324,0.669976826482954,0.2912119995854988,0.086029310199908,
 0.24207355582678725,0.018799793459777336,0.3017642763306786,
 0.9776402694883108,0.8585696781619535,0.7527932535741657,0.06825973016545617,
 0.3533621890957105,0.7834155924461975,0.6206048502442312,
 6.525714150419359e-4,0.8072716747168132,0.11185478368791213,
 0.8944290981276026,0.8960855236916536,0.29782640111873504,
 0.02646559000908244,0.3742236688226187,0.9367863357337535,0.7163269373502322,
 0.6146548782089951,0.784654769328473,0.020021070054378587,0.7283427765360386,
 0.5049544794077075,0.8105577988454239,0.8345729112416019,0.1410363754690911,
 0.8677368340854708,0.24930223332447676,0.41733695922516767,
 0.4245820873240226,0.8942346889931537,0.8833737618470159,0.6672088880930437,
 0.24946047131605642,0.9357004725931709,0.5445498527420926,
 0.030855290756231946,0.24882822878301858,0.7706435820208327,
 0.26762694562311873,0.8687609460546238,0.822166156627096,0.5089271068444017,
 0.3724292502832366,0.5909328207235575,0.6157869407581429,0.12075203099591769,
 0.14936378272688766,0.12470684349032535,0.13586287582100365,
 0.23532699825678838,0.4402414717509079,0.6646692865844328,0.6680470507337993,
 0.8649449683001486,0.3997671700753991,0.8755416982299862,0.3975336614171887,
 0.6651736180117123,0.7070763885893087,0.754554733280654,0.6092793607512973,
 0.5963497460980383,0.7445898281515628,0.29740900397157466,0.5841821215691227,
 0.9514008021237689,0.7098907630608957,0.879166614311373,0.2737747180907675,
 0.05436147660401747,0.17013358610478124,0.5831704012892699,
 0.8315503800562197,0.8118108164752784,0.1608002637058844,0.995671155511984,
 0.7228859821913228,0.5148258276568835,0.08840288146393416,
 0.29694137174212587,0.8607318034117817,0.5840781349591819,
 0.35268573746759513,0.5329219831741864,0.06926961335961701,
 0.26159059892904524,0.4207850548570836,0.6881893974360098,0.8005698050412428,
 0.004156524513410864,0.8716539224982552,0.9946866361697472,
 0.5578141215037153,0.27236498284437793,0.8842790439780086,0.3447252070053093,
 0.10884901368075006,0.9985482513475743,0.6246720455784489,
 0.19892120460144458,0.8193050630911803,0.7658661056847822,0.6799080007845244,
 0.8579481386816707,0.699090969273896,0.18078864628658675,0.13461262306078803,
 0.2970003180215328,0.07473552172943532,0.6147722274275226,
 0.15270202774611397,0.3716752495071868,0.05956297981731773,
 0.5897847426718049,0.021518008361321672,0.5594958299909332,
 0.9064139307318277,0.4545126244379589,0.6016509181103985,0.23914105321222912,
 0.839106662356156,0.4914843035616905,0.5870517188312753,0.7631249623605244,
 0.6193111622968787,0.0641053533388819,0.6079669082489891,0.9169730609525932,
 0.4321861366625027,0.3090885570904195,0.19974098953351493,0.4616890088050358,
 0.7667836163411145,0.14305776765381117,0.00799163502692124,
 0.31966851164482735,0.669269122722402,0.3091617935260834,0.4969034854593467,
 0.9746091372213666,0.49280446464072414,0.5280317790167167,
 0.29007564270532193,0.3172362070178989,0.14992754790315133,0.89862460587648,
 0.31930778180403363,0.8499855249165539,0.08928616017242308,
 0.09724512126053786,0.9973238527438142,0.8315474274790983,0.5064709603735197,
 0.9839373969934782,0.46359025394632447,0.2547829391341919,0.8841074732365455,
 0.8673814397709377,0.7062142849367592,0.9760797292838642,0.26580898815779364,
 0.16731417292091977,0.8495161485052695,0.3670648285113651,0.2778802209783403,
 0.9951293966958721,0.4579936883883703,0.9944178164847691,0.9282012609222761,
 0.9918834217426675,0.8010023238260938,0.7948728365043025,0.4420441958581284,
 0.3111013441277617,0.13700081263946606,0.732489395003166,0.9624243335258218,
 0.5058103637706662,0.6285976420346194,0.3451085613782737,0.18887447609920116,
 0.5981336493504756,0.025220268951977376,0.5012223208047063,
 0.32039111693943545,0.9692278875621707,0.2601497488200317,0.6315372121488801,
 0.9854788143632521,0.7290096141766979,0.5918941318424403,0.912737990732521,
 0.09703412598104644,0.16408807443148854,0.980623686782344,0.2425588991010077,
 0.2702763314902924,0.5659803280076283,0.4039195059267494,0.4688309734319497,
 0.37345712377361673,0.7758129018085462,0.24607982185474042,
 0.8969736852108242,0.8440379069588159,0.3396865830650134,0.46027820735406066,
 0.9137307835274447,0.11887777076744199,0.705917673440162,0.5009823270546104,
 0.38626781945731914,0.6640396825807652,0.1823376910620208,0.909682591293069,
 0.7124863453954983,0.7282489482667065,0.36648930625701337,0.7344209540906794,
 0.18035592938875666,0.4306509748992431,0.19456108456517918,
 0.7759696636108545,0.5675979605249354,0.3512956142572945,0.46535721140273356,
 0.2127277759296644,0.19588026042824558,0.045811283620955434,
 0.7899788638162005,0.07646256171058408,0.73852432550177,0.16221920127044098,
 0.08571507935402911,0.26368258277807977,0.46788723197606563,
 0.7035067082819917,0.4871961518719137,0.2176132969240273,0.6955644643185581,
 0.16986067055584264,0.6510990848800575,0.04993135562016826,
 0.0015267217958249146,0.3033945946264678,0.8317877836296693,
 0.4000884720291473,0.6054790928236944,0.2219282172380601,0.40425147933494654,
 0.4307419026026328,0.10595851724810956,0.17957375931161623,
 0.7449380653532381,0.4133173027679746,0.2370546573640151,0.9848506558925217,
 0.22457193777497364,0.9161970161243504,0.5783460019068525,0.3260390693234557,
 0.4163132439814127,0.2353256007436112,0.6704729554369778,0.28928827101719057,
 0.987723697405248,0.11264623029686249,0.39663250845603437,0.4154923727543225,
 0.2622941163662049,0.2995650395121665,0.7722632241513994,0.05966427356902848,
 0.9413812563834438,0.13531859990775663,0.5062947501578423,0.9095880634264603,
 0.20318736323441589,0.06446536224972788,0.634118846874367,0.7849963451709543,
 0.7105511188405585,0.21230736558723462,0.902916419208559,0.6113089884853858,
 0.44301889516630144,0.4390996580298472,0.7990674326902146,
 0.16876083791804009,0.4102983763872883,0.18592488959412978,
 0.6146045706518154,0.48870513715118125,0.6631556926858075,0.2273856161355543,
 0.3610771111975928,0.9011301204370956,0.021051665144558807,
 0.13796022187896806,0.7418767120108366,0.9643567286962617,0.6815650808979606,
 0.6815915387709494,0.7131813761706803,0.46765569065277357,0.5140271338849401,
 0.8876522747713668,0.12682494626313445,0.8737126636765781,0.9810423322454233,
 0.48188522348834884,0.6681142706992427,0.35704258794036214,
 0.9256402577858045,0.2470307555155481,0.7639154457149372,0.0626523775047313,
 0.017188614392545376,0.4350401091028685,0.20696806932769696,
 0.43584758592585326,0.1943651672482124,0.045642308283263655,
 0.085466834194594,0.08581451303861332,0.06782816737775799,0.5534242518667939,
 0.007348504073914963,0.4316625047661673,0.9401666601082951,
 0.7760065382731354,0.5389782666526852,0.48588364845763954,
 0.10478380274858168,0.7208327453078649,0.1376202905011823,0.5943658618466043,
 0.5128979774921477,0.07555325881200714,0.12539158059832523,
 0.4782213421476792,0.04737503711614721,0.7342532621145956,
 0.07012897188979439,0.108478794611655,0.8365414343350661,0.44243473546832246,
 0.016096112058947898,0.561296352814595,0.4323399468165603,0.915542990644031,
 0.05150011107171493,0.1192481754676139,0.6927147124632484,0.8693861797815916,
 0.7313126400442674,0.5825373512444583,0.6459300478403146,0.3784938931465702,
 0.82132135489593,0.4311954173139638,0.8775081458737451,0.5100130702936636,
 0.387745991992291,0.15734543811328072,0.9461279315554851,0.798573940148198,
 0.5516500258367397,0.837227247097503,0.9393657848758914,0.13576668264351865,
 0.8397920383287524,0.6548040248626341,0.04827763848392108,0.8859716416181821,
 0.2611256050538069,0.7996870898372131,0.8582960740368486,0.6457985443440644,
 0.5818708889810976,0.3579176629823875,0.6966495809181095,0.7427467228060645,
 0.41714788965784166,0.30851145712381767,0.1342724838931194,
 0.14159483880083568,0.39517723914578795,0.9641051842070104,
 0.04397045245170972,0.2565603099474274,0.7582276793280351,
 0.48228137972811924,0.06585673012393123,0.8421923607626636,
 0.5555552796189931,0.48641812276505547,0.3129802620929394,0.2994186633797684,
 0.3462146468291214,0.30581222799535757,0.4150263803748967,0.6210612024659927,
 0.8497417246892666,0.8644139604376322,0.8179254388237934,0.7754204704139113,
 0.7313244682073512,0.9648725138248843,0.5423516815844462,0.0184484058842157,
 0.22944515840411717,0.021342413212268596,0.4430782382637162,
 0.8513313118793224,0.5086285185541939,0.6762582112714741,0.7238852737499518,
 0.777684086349672,0.10096755164304327,0.8813236668010681,0.0506300762332873,
 0.9071825244084315,0.26943271683316794,0.8007137657767744,
 0.07513255759355619,0.7562966317287106,0.38899493024184817,
 0.0908924059183186,0.2091630889430861,0.26739556221866323,0.5024319393462955,
 0.08035473623640654,0.3082566793316358,0.4023687007836958,0.776142573592723,
 0.20769437047793204,0.4882200030911267,0.9849074601212818,0.3242541961178096,
 0.26361484103798194,0.6140548159471162,0.23454282460229248,
 0.46754658004036376,0.3258881748981932,0.11632394717035821,
 0.24127571121959968,0.7062171561080114,0.7803381128694744,0.8542457994682416,
 0.7848944105180317,0.4793307210542478,0.9264546853457263,0.9095600656594567,
 0.19399981132256638,0.08825902949943687,0.9535826424178443,
 0.7702700169041782,0.2290482384773853,0.36870926100790036,
 0.40041978734047234,0.2313369295538492,0.36931712675186046,
 0.6413523815770654,0.3965154376962323,0.4533321072690101,0.6192531488391954,
 0.05636902753464479,0.39279868137019835,0.6373053394604409,0.267852383597467,
 0.8939593316774171,0.23999718646137336,0.11959433824183319,
 0.19753055503103067,0.6701033954523438,0.45607121497890457,
 0.6273857267562377,0.7472172932262101,0.9038368385949281,0.0545936111874844,
 0.9879937018158443,0.9271064604695521,0.0672588856122831,0.1602827575753415,
 0.5880394609747759,0.2358084525985974,0.8119026472179351,0.46450021333878766,
 0.11546724061522784,0.5234149158561158,0.6796794710805083,
 0.40149597218884914,0.988048437829791,0.8634110250273438,0.812739125555329,
 0.8612715052879499,0.8150639905497383,0.42935533871962583,0.6596397109809052,
 0.30219495356052795,0.6907795607446217,0.7220116912675807,0.9691687354145545,
 0.34983757462722453,0.012920785895237374,0.05468080023846733,
 0.6316247474856174,0.5169002671105373,0.8529426291854495,0.7488851966987699,
 0.49916152596121954,0.7441319639632865,0.757871512308475,0.5270899596497961,
 0.7711535369336703,0.6765161768624087,0.03329382692733107,0.5536456116987325,
 0.8391561249403562,0.34253876227457736,0.41887018861419056,
 0.1348944432382312,0.7829353963607167,0.8326709282194152,0.3938472301612391,
 0.20542556502053744,0.30745395785273766,0.2183188053644346,0.856691879600507,
 0.5324540724000072,0.5468956598050687,0.29112048234982446,0.5190880278082463,
 0.9815441936598664,0.24948765014981178,0.77328263067721,0.2406603698712554,
 0.9030110077123912,0.8906971940108895,0.6051817020424568,0.8508537163482252,
 0.3192517643112315,0.6974826572261359,0.9326851267414515,0.6498870560602173,
 0.11535657989940473,0.5863548919891224,0.5960304087272159,
 0.12572457687051286,0.4930483897253264,0.32645057640556563,
 0.3906786666307096,0.808473000861743,0.1637842391188996,0.499521611382442,
 0.26394238264335435,0.3758472711852444,0.011440315141141078,
 0.25148815327657836,0.45545688742382806,0.07871692624575766,
 0.3610117429871008,0.7353433732960263,0.4484605358640601,0.14650976494383317,
 0.7560097807164607,0.6822726976754053,0.1912417386410732,0.6615435220966189,
 0.5058577104222656,0.2398455521877284,0.6485379601920048,0.46476865570805714,
 0.6133967854581419,0.3683970945934224,0.38389297628574115,
 0.11449691394801109,0.17483477163140482,0.6925558535153981,
 0.15575367375088056,0.4291908855602249,0.9831410339149372,0.8679102815830331,
 0.4850677215987633,0.4091855220645796,0.3889926481314715,0.6058443232418861,
 0.844852333934319,0.501626557624499,0.08119295616309508,0.1181891382037088,
 0.5999353440048326,0.5824470529536118,0.8086659656163279,0.5898272679136457,
 0.1350484332486156,0.712285162016087,0.35739277460071206,0.16024654381861536,
 0.1808099628334332,0.8731124745515049,0.01782460530623997,0.809054360924517,
 0.2091390362438561,0.3526597334482239,0.4306325583113517,0.9755653420301242,
 0.11146599231499743,0.6176984989017054,0.5208712863432825,0.9171918612358181,
 0.8592125196332137,0.7990646571993602,0.19985178450313723,0.5481504425549855,
 0.08525474240524489,0.961449182327609,0.2540567337566376,0.45804032489422186,
 0.23604851467252186,0.9350472609466174,0.8710393203697906,0.9571776672707043,
 0.966174743967335,0.9046662877094968,0.30175982017077363,0.42992811591511115,
 0.9577539608058826,0.39826577952038733,0.7310991127721638,0.3535476905279501,
 0.035726954186136806,0.6988281467252416,0.8712203732230005,
 0.05845371861860704,0.3718196022250839,0.3781233667178826,0.2913133344653175,
 0.8798487819755764,0.2822426390292604,0.9349889722000031,0.4633228368199034,
 0.868156090898512,0.2855571439265162,0.7170381830114216,0.3235118226436623,
 0.546495775426345,0.28227770691298726,0.16391445385388537,0.6278668004727208,
 0.1428862278495342,0.25005362265373976,0.9306626897008634,0.8267866495528522,
 0.2564687510279058,0.47137063160694204,0.7265322387832913,0.585200000821879,
 0.4621304436888508,0.4454761353192973,0.1545724538692187,0.08453929079765587,
 0.9812837559572946,0.7311471617363878,0.6315975275203066,0.5989322814381726,
 0.2493930816739165,0.13002518311251998,0.22909554002484978,
 0.9345329925177075,0.020423899933726064,0.7863431543556215,
 0.35184031535408467,0.4961332638625018,0.7642532161669701,0.9888692337560144,
 0.5040593221544862,0.6541467105979301,0.45863480526561407,
 0.06970715565736463,0.6562207931557915,0.7634967216900092,0.1476087865325325,
 0.8055003287106406,0.8689997874205296,0.18481750068010983,0.2423628354933849,
 0.17891613102619874,0.31321297271374715,0.6999188177208219,
 0.7004298646579867,0.8774948855415842,0.4247410937629803,0.20991192972537043,
 0.6138890316470974,0.8786630217469716,0.23936311424176837,0.36248941634526,
 0.014232848600352122,0.06859554974170057,0.5348741915111306,
 0.7620389521103146,0.39659114744870505,0.5517092641646333,
 0.007756119307873233,0.04765769505847495,0.41123480390582967,
 0.008700860508602526,0.6988806240938216,0.6853652674761015,
 0.2437221138779576,0.047901400650996906,0.33286458879659087,
 0.38860289190744024,0.40531897411579954,0.561632607743218,0.7668818785077468,
 0.11579839561573835,0.3645455205096626,0.09961757619465184,
 0.8128784318201544,0.12695824659518296,0.42591330662131577,
 0.9971405356436644,0.13838266795656762,0.07635788866995774,0.789262129710597,
 0.3754404544739811,0.36886729813900243,0.9897837407374539,0.876795463907162,
 0.6354478661783517,0.4233321393074767,0.7526768613012286,0.13765338528327675,
 0.04124458961217381,0.5176155836559085,0.264868685341924,0.9520570905708592,
 0.6103790268221579,0.15099363308180902,0.7459570915144648,0.8406133022829413,
 0.9957083547736859,0.08617603332207935,0.5717872482908877,0.9054544586684005,
 0.6605613904185524,0.31876289303186045,0.014842090957980725,
 0.9121506439467513,0.18666184573682,0.02085889421252185,0.48560372580400935,
 0.016417715970587166,0.07671387436276988,0.28076514730149693,
 0.23374792392104704,0.9702320991808618,0.21631950470532635,0.253884536846784,
 0.7167005984445152,0.5357134648928885,0.46183466339903034,0.9690595772752705,
 0.45221287063666715,0.9954781962478227,0.5354483754869412,0.6742224618080093,
 0.11936636027429426,0.24185391324921446,0.33997639208715547,
 0.052191303970429925,0.1739999222940325,0.251056263080185,
 0.43270483604198573,0.7372794962167242,0.24018950867686994,
 0.6280331753529063,0.4193344921147112,0.3837444261201912,0.2525869874506026,
 0.8939274743560174,0.1131597397213353,0.45078629121011304,
 0.43673485417904934,0.35873464741195893,0.7862691485199472,
 0.011986528027951526,0.6798093425243311,0.6191506926550645,
 0.5808330987193184,0.4048171589255628,0.2774514148536723,0.33325770997431226,
 0.6445851794087171,0.3139755850151388,0.69389842515973,0.2532786559516692,
 0.6472902126724895,0.3613777116665241,0.5892136488051959,0.47591888750789235,
 0.9839177909558734,0.727505015846926,0.9449755591060779,0.773849091104249,
 0.6864256477583228,0.9418364069504022,0.2081165339712685,0.9588277880267467,
 0.17743647104207216,0.025977248447905166,0.0744905150987133,
 0.7531497362914074,0.8567608781854383,0.5689692300351091,0.8084458918541618,
 0.21594105422123555,0.032965206683765014,0.7221222696588601,
 0.7480160240463876,0.226057652089946,0.37715341542573144,
 0.039424952261266966,0.7297820617141688,0.7931930261078944,
 0.5359835473595032,0.8230121008762824,0.8105575445002917,0.8792578239615696,
 0.18139700140369253,0.37467810762026044,0.03180360457197806,
 0.37239816483748023,0.850680266760418,0.303677402585509,0.7183606540432641,
 0.3363876265231651,0.3279087852387682,0.8221944185338395,0.3256372745972336,
 0.2882532224729153,0.9671042562622884,0.4431736343468604,0.6842516572284786,
 0.6559598126527619,0.20703218718620897,0.7259188924676848,
 0.15484232512061813,0.1697291653109645,0.3952324130264079,0.7464174658665272,
 0.034717883798766835,0.43571791734461984,0.3846136990991035,
 0.9416469349675262,0.392204170545448,0.5458614015985823,0.7114547093439038,
 0.2064613214450396,0.44344306570961567,0.01645710262516442,
 0.19448214199689629,0.7732111789960908,0.8432299165421977,0.4126790339043046,
 0.387940561197227,0.09361660105621261,0.7445420003037126,0.08549356669466646,
 0.5104100705216829,0.8595778824887439,0.5848487768392148,0.7338921840408793,
 0.9192673386611581,0.09798443548362223,0.6861088951615509,0.7552686057569208,
 0.6289070661003607,0.3462227743056726,0.43681352230078674,0.8237445643641372,
 0.6359631031151545,0.7659557159133665,0.002741058172246791,
 0.6562791911652407,0.21323961368129118,0.7454465115128687,0.9668635121788673,
 0.08378611585292539,0.8773755003549405,0.3254619871893101,0.4471166079533777,
 0.5372218336330438,0.5799089012686601,0.08462434389514495,
 0.05015827432345088,0.8217988373144741,0.40769152738092984,
 0.4201455744884779,0.38767375411711413,0.8310229578990244,
 0.26472017004277515,0.23716323382635363,0.20228478835796437,
 0.31236204770249376,0.11115291328651455,0.559903242914079,0.64516509445847,
 0.608248959450124,0.2507501932292371,0.22166211712903705,0.37999248679612063,
 0.9348221485320852,0.9319925248631766,0.6916359241100547,0.1774743834800866,
 0.49686429477145055,0.9165244512307605,0.11772727477348632,
 0.4003791631467932,0.4678838580986169,0.13274497098977422,0.7148353784452284,
 0.43732426345555997,0.9788016648528155,0.7430335522196521,0.9544200469089464,
 0.07208102624932744,9.904616581770664e-4,0.2645789127628838,
 0.22058330297527573,0.5363189247360857,0.24290627986626245,
 0.7280890669476895,0.8022367950292768,0.4746952084696505,0.14473565939662902,
 0.16675710232351582,0.5102860936363338,0.3901421451544155,0.943484748171451,
 0.6868027167341811,0.4582540543651533,0.5254226879997125,0.2799094821788999,
 0.45839130670477,0.16094018853436742,0.3557726628322444,0.8227257283510536,
 0.0715641974376594,0.27583451492773736,0.9564142495051964,0.3421348847217428,
 0.8667281515437307,0.6528368192324436,0.5509686016142121,0.5336845577814439,
 0.40511951177662286,0.18200040437221232,0.07757908967093974,
 0.7499440945280295,0.35486802277484697,0.9720398558360986,0.2976150330712162,
 0.7534610647417603,0.6849855219024628,0.7264639557659831,0.5861627949566484,
 0.9937320512516985,0.38364488086364956,0.8447212644125526,0.7501074503088536,
 0.3520290792148747,0.692779166162393,0.5212692771807887,0.8803816125216112,
 0.21033949553209452,0.6116101906195895,0.23363523713043477,
 0.9597693717265046,0.08431597402713964,0.1297777982408237,0.2730860160538374,
 0.6380824909464682,0.6403154610933995,0.8788965548171043,0.5851684308998113,
 0.5428225426455152,0.6450595933456359,0.3603903781834452,0.255560711326957,
 0.15061077217880692,0.8239746347900943,0.7497566246733682,0.3333946627414016,
 0.770940515478536,0.48603067723145954,0.3210697217428954,0.09639286106981337,
 0.5541472066943062,0.10620271083071764,0.21197167629586622,
 0.42356694238685355,0.6473755210773308,0.6108634210457665,0.9147784475133339,
 0.7590948273505085,0.860027311302816,0.9535636623063877,0.14196571093066623,
 0.8337944237565863,0.8600184932397775,0.3255712645116079,0.8623788329060336,
 0.5766992977161536,0.5436177148801389,0.6450466067442991,0.09789592593846241,
 0.4915302775032925,0.1933336243508348,0.1204795756391257,0.8653082399552177,
 0.7047411144195425,0.927737862099162,0.8547372027274126,0.5358475306894407,
 0.5367882403360609,0.7161303864500751,0.6383312343027341,0.70054528834985,
 0.7159199772617879,0.4142991049109215,0.9704852177637402,0.7614601841303643,
 0.910770705212196,0.39506402969010235,0.7316899522351463,0.10654740942758023,
 0.9940685944031471,0.29399581691969334,0.19414974401619844,
 0.8984415550383962,0.9740727625964494,0.0891327314904744,0.7317391928232766,
 0.2750256580926095,0.9966139126799242,0.8656188818960839,0.6127365655055579,
 0.0038412715144510168,0.06318179124077983,0.5347336798748599,
 0.17856489964220135,0.9220394738226184,0.48471682217757794,
 0.9013431144260223,0.26141548365496814,0.9763324052892446,0.9803357302701116,
 0.697519206846577,0.7834641967095931,0.08828850792940668,0.5154105840309959,
 0.4626731459519,0.2974926342359856,0.3877089485693528,0.6819834656265383,
 0.3925120824384958,0.6461470190857243,0.24810302576355825,
 0.45689852322139746,0.31525073899387535,0.4376975810063183,0.34930189821735,
 0.941866322420436,0.8721370444818806,0.5703378497273682,0.4128471874096845,
 0.42858717090334986,0.4423482590101909,0.9891496695264392,0.8731762906667202,
 0.5573020371839914,0.31217775974221695,0.625226762902309,0.43720751747448217,
 0.8030972429588771,0.7932236558987488,0.0308408353633447,0.8431465897850651,
 0.8014638902964083,0.8462533835098044,0.21299760495553,0.8334226198636822,
 0.8676944864719438,0.3669269639961955,0.5758407699710633,0.9712592522889061,
 0.022740794438847245,0.7820245048196222,0.26913844320604197,
 0.13378236862105064,0.2402808010785733,0.23826244318031609,
 0.17544397454364746,0.45424400757270056,0.9725782300381006,
 0.43669985370251674,0.28526933570755986,0.7331249903449927,
 0.38961006720168057,0.08253999105224841,0.11983555244987987,
 0.3109379397738178,0.5287422999138041,0.3150565346231924,0.8533475358240235,
 0.3628671408244153,0.735942042659131,0.5712186972015066,0.3454195636357055,
 0.903487411185605,0.9269005093200294,0.23503440552989352,0.26484623045665523,
 0.9249108111140567,0.08815610926166106,0.10827176338026145,
 0.2499299973867971,0.2475630717453715,0.327950555823204,0.1311220181061219,
 0.31507558426804827,0.9901154368264945,0.37251563377394237,
 0.5510676371779937,0.06988137123945903,0.8299568708377797,0.3677625429626824,
 0.7292815629559399,0.9331440989308845,0.6614142081247494,0.9759231306106191,
 0.22111887304962774,0.12878973536762228,0.2519866818729778,
 0.8922759153486333,0.624600131957073,0.2301342611541506,0.3625229117057408,
 0.24137245384920192,0.3004822781777895,0.22662580659824305,
 0.010517276004360143,0.8438712527872048,0.5445436646952682,
 0.8209833375794233,0.05295446337528886,0.3784173582662549,
 0.12860186776510707,0.22972862276357742,0.3882716790748011,0.269451475120168,
 0.6161564947579045,0.31675540066690966,0.9057848132921045,0.8989740762010341,
 0.24359843322104868,3.7270242990572466e-4,0.1841778229291834,
 0.21878189454811525,0.5811837405688155,0.006923976660281106,
 0.1166499953454383,0.5365492674803578,0.0881560501986336,0.9284154673024726,
 0.8199236701961296,0.8131760658848056,0.9239246489601701,0.26798289658717744,
 0.6481209064715745,0.8524229155036428,0.5636817750035336,0.616722639555493,
 0.23103562611039719,0.32419581932952346,0.10884346630112407,
 0.8100757541008409,0.7859200999629297,0.2519130434539876,0.6438804316951001,
 0.7201292029909152,0.002033965650125702,0.6224912898251213,
 0.8905975204247214,0.3715586158102161,0.42714515912479323,
 0.24811697847884917,0.1542183994044245,0.31040559056066175,
 0.21835272133363137,0.8415038278190996,0.6916578523956498,0.8678937277046751,
 0.125109462933775,0.4041483717206925,0.1147317928386109,0.20750669525422094,
 0.16864334024807004,0.9077027113417756,0.2773334490819932,
 0.17779545101614547,0.12825011587225854,0.15694436240876575,
 0.12222041437564168,0.6555630432696009,0.2554182555674517,
 0.47584179475576127,0.6918071069541083,0.5174343195786627,
 0.20546822233794226,0.4402514844372488,0.018854025605895663,
 0.2094471344386991,0.6911429678887009,0.12871702281845998,0.9469660800332536,
 0.5356069693852916,0.8966540020249198,0.31850993332915845,
 0.48015459914319836,0.13979259532766375,0.9211794594963832,0.501918590371471,
 0.5111376174867085,0.1118433781941599,0.7319014692688874,0.4786865467535286,
 0.09269985511066814,0.8819529615998231,0.172321011048591,0.910358109309471,
 0.23266004490563685,0.30037956900237206,0.8427141465133405,
 0.7714699526367368,0.4541422133309092,0.3850913373635697,0.9864907486419559,
 0.3305897312650653,0.5135556222578277,0.7059224158698616,0.9828110095518219,
 0.18150139463417525,0.8129969846982632,0.33599636420150225,
 0.4225898258044466,0.4710099284118283,0.19519405225446596,
 0.04822847916534723,0.6309537258382908,0.1495234027542649,
 0.47357130818135784,0.5299992637089103,0.08855147982671352,
 0.8989377015528306,0.37247020733268954,0.5991461686150761,
 0.33898274494418934,0.6950431372706622,0.13237552088649374,
 0.05793371814227788,0.7617098419224178,0.9423146918811826,0.0687505479459144,
 0.5476855280936551,0.45375350547537985,0.04267674200528582,
 0.3457512019082384,0.2292672379706282,0.4532992400962227,0.8417220611985621,
 0.08400461910356138,0.8045767892425121,0.15691707697596402,
 0.20410266586353643,0.7034033130510321,0.5077041595876677,0.519489749414053,
 0.9669163429014299,0.8339416984636371,0.5061944872639332,0.7772833774827572,
 0.2501969902471586,0.9407182254856408,0.34318916939347444,
 0.41361265114607626,0.34132834301852033,0.7134091414247532,
 0.039920232006711576,0.6930314861833377,0.25833487237666564,
 0.3943727932597323,0.828043404393649,0.26778480497015,0.3367384785902099,
 0.5407288748089998,0.08101925176996816,0.9407679199452363,0.4499032593635679,
 0.144998228089107,0.2694691147125783,0.44595492792748403,0.11183156250846205,
 0.767271350033796,0.5192697626058187,0.8682202583405332,0.19698086308819673,
 0.5718624356581872,0.5044233485403283,0.04245421408769767,0.6250116497243966,
 0.5929062149026191,0.22277480753144419,0.8116920772016901,0.6705814147364425,
 0.4876216015507716,0.6984758944858193,0.93240459481106,0.4491680757175359,
 0.579755922901513,0.22717178800756244,0.5516289166622247,0.11556164488763676,
 0.14688247067804794,0.039415537074847906,0.5442760982744267,
 0.9827200215534725,0.3822520522827437,0.5838726238641979,0.2731069552851756,
 0.12560782769218792,0.656689752047619,0.45943872125027696,0.8223440504642543,
 0.6310412792407513,0.6165238775591018,0.8101022210523482,0.01465948698326558,
 0.14434474651055829,0.5057070928145989,0.42200283782026515,
 0.6745727794960362,0.41751129283849386,0.683039577683684,0.6952313281964903,
 0.10012940087679467,0.8010220299987922,0.31236292326400006,
 0.5186557042941684,0.03584005144505753,0.37931534430618064,
 0.6752947941197043,0.5608541464320365,0.5412565570525215,0.20634485277158832,
 0.9822722603010262,0.1903056589763199,0.2369211939722008,0.4022278666126077,
 0.5579231401400166,0.8528567782799062,0.6491775644273331,0.841062286401196,
 0.9826391544952197,0.1553348368135098,0.2676031556427091,0.8631618481421384,
 0.6817216046001353,0.5945270798310086,0.28549298231173303,0.5033410471506237,
 0.9790484045183468,0.2959963664927229,0.4299326232465275,0.22057862065374445,
 0.8233977340835201,0.8009395628234187,0.8326938989427405,0.9554599666560823,
 0.410577768913585,0.2507192324928613,0.16656945960717406,0.7966238913875363,
 0.775629025652848,0.6634574811018834,0.30666105357706797,0.5371315029920966,
 0.3574093081640528,0.03609206372617013,0.18033152838677158,
 0.01732243791647048,0.12354028033046893,0.15445527033817152,
 0.35250441061191706,0.738082975333545,0.5442765046752703,0.13726945606630847,
 0.8331778560482719,0.3332701914609806,0.8712451976133068,0.8208972874012525,
 0.8369701486966801,0.9264704791973064,0.3011261978372032,0.3023680693380314,
 0.013470635651241847,0.20166384917318192,0.5565358341235536,
 0.4031596070098217,0.5191567946206622,0.4950746687209816,0.7376262028780698,
 0.9555589256088244,0.6938141969417915,0.9861721368492234,0.6819309159801603,
 0.6855897557030364,0.11790630151904669,0.4129513693711615,0.6524193583123594,
 0.5884065228897954,0.33985228343748486,0.3768628274739131,
 0.49604476767524175,0.16760898219046105,0.096485237766756,0.5279807216468738,
 0.29488097093231924,0.8853778437952391,0.18012822824406105,
 0.5614039170590486,0.9052387920750381,0.8849908807476567,0.6969293971782349,
 0.08585564545276281,0.701461328357635,0.952716715108595,0.15241783261406416,
 0.421054162616445,0.9168989647812014,0.8720509878805102,0.4790432106289211,
 0.9140396283864282,0.21206688635290272,0.7324535647576309,0.5491693326383852,
 0.7458964690614643,0.5419576198218287,0.1344314339232353,0.8979765085132325,
 0.5790187406354088,0.9868829542096313,0.8888704571073635,0.6287307981797902,
 0.3547522236781212,0.2560735481710965,0.38091463312363316,0.9467661584278304,
 0.6469732534080032,0.7280877471821048,0.7666508421980708,0.15647863335400636,
 0.5584744196599359,0.08970967591096035,0.5375193762173128,0.9670969259305481,
 0.44443424320283564,0.9900221718299127,0.7956782299210259,0.5434274524001621,
 0.4942411648676953,0.6416636715645831,0.10720689799515659,0.6727064916770487,
 0.3555957153563676,0.9384053162371484,0.38117800265752,0.6108566450493672,
 0.7186667837412021,0.25307084201685326,0.8807654998589047,0.8822829037283549,
 0.8608315601067785,0.45626848356611394,0.612224478294362,0.3134646663254652,
 0.5951969743117019,0.2094386005534261,0.6536112173138905,0.8112911680546678,
 0.35638010823159383,0.6116926160235295,0.7566831582828021,0.8898273369531847,
 0.7920005686718794,0.6903382265718633,0.022100010104494694,
 0.8324558454040284,0.10732626767774489,0.25325218798652904,
 0.4633717032560225,0.3282355214682635,0.7591340650919037,0.23079967949368352,
 0.5074593820610173,0.3968381804265222,0.5003660016939999,0.44580349589035406,
 0.3073525219210821,0.27966786546442435,0.5207584614052773,0.5092531359063707,
 0.4912176135170788,0.0890562892074257,0.1803420849787415,0.6937940526066374,
 0.013330061795096548,0.5484841225570297,0.7335591280024758,
 0.8881441993125149,0.4169098210581652,0.8084001224793536,0.4197715638138103,
 0.8876547089245898,0.15173547136227328,0.798668172960522,0.5439473671641939,
 0.6515587449765807,0.3330580550734392,0.6080546107909426,0.5572082683550788,
 0.7086836135052335,0.47442425949671485,0.15734957117479564,0.501764843620978,
 0.13977908556687524,0.060535950954957674,0.4448817432505473,
 0.9592512028791801,0.1427887586059413,0.21581943158040395,0.7792056328601779,
 0.8580590168451889,0.19992896817715378,0.48278264536360016,
 0.3831370387167694,0.1800375350700294,0.11048610780628276,
 0.41370977841968315,0.9161821173522022,0.11850869787602081,
 0.6071936235253315,0.3382379460359527,0.38527936160802223,0.7748295061623901,
 0.7198565769486905,0.10005339368678812,0.023257542275464393,
 0.813231903705814,0.43841894477671584,0.9224084119084722,0.4225710739663243,
 0.3001587337632553,0.35366434617800935,0.5742235725596953,
 0.19999141277904808,0.16360369075238967,0.5214457828727876,
 0.2945403229156959,0.6039445486300247,0.8163350234570981,0.4107763654978833,
 0.03531300332867626,0.5244355463837382,0.17966823667570864,
 0.11898868220611813,0.8789859819232433,0.8695658304264499,0.5195840592377253,
 0.045204020709059156,0.3095654409800278,0.22540713389963685,
 0.3302135465660888,0.3600007829749138,0.03836638507232859,
 0.14243303792638706,0.5329848482736903,0.02575362892952604,
 0.05756690143360932,0.3291901156507172,0.7059879678590584,
 0.20143274080393936,0.11927862267484768,0.26907129933630825,
 0.7963642418047069,0.7445376988010142,0.6595399129919923,0.05966431287011276,
 0.6271630683786273,0.2722114856522184,0.17647074338803292,
 0.06968223117087557,0.9332156386618937,0.3577851563172897,
 0.0012355409584322574,0.3428949935958834,0.773547995901247,
 0.9177053071957008,0.7275030704892116,0.7062071925653933,0.5967753815906716,
 0.7576966349298622,0.0834337909921683,0.9231258134474164,0.8612028880868495,
 0.27412842716701635,0.4893119762947781,0.3399921402875581,0.1587778223371935,
 0.628907927055851,0.0029745865044848996,0.052032043662767924,
 0.18544571262868326,0.8589959426689355,0.8981340712699963,0.7760294739758107,
 0.29678317853498837,0.15761741883236102,0.5352048167394551,
 0.6988863382142234,0.1305383315683447,0.26238298018247364,0.5151817518168222,
 0.5368164263182444,0.014261380504782206,0.18705417459621476,
 0.8694998463858872,0.07820829934131868,0.7138725057665479,
 0.48962666983591796,0.892179676877547,0.2576896381218243,0.48084328926805764,
 0.298263862037096,0.1935773859753216,0.00852319060353346,0.7813185351760046,
 0.3422468761300175,0.40883033206532227,0.333278932543452,0.3663959794871401,
 0.9523709149946116,0.12455730274213306,0.7757941432570272,0.3893639870285116,
 0.034810382901307,0.7097709699269801,0.13484336807642072,0.6477604045777611,
 0.853893297668324,0.5889049677242941,0.37390956174025836,0.1999296306657673,
 0.9812267790405361,0.9436179127300127,0.7473605772897037,0.15587532781373048,
 0.14865335419745596,0.9283253838525168,0.9654192470143791,0.9764950222881754,
 0.5091654480182238,0.7494183148878126,0.3321162026762279,0.5008482283834633,
 0.38424769008692805,0.5973108773891338,0.17667143104445082,
 0.42478399208145834,0.5642628708653334,0.7378354183791904,0.3770180969182868,
 0.4663989712683624,0.20555486269656864,0.6769156960344327,
 0.06822233388952248,0.06999085419622686,0.7520496930429557,
 0.34829192661630703,0.7173300535896541,0.04795576870702456,
 0.47331066554459666,0.36764302636855295,0.7117094813269463,
 0.17704933449169435,0.18592428539733696,0.9526578947133245,
 0.8904289171925962,0.5635776150551911,0.5349032157306631,0.3362535679775377,
 0.6356495421721426,0.8164485264330796,0.41421425333284656,0.655755746980927,
 0.8921976198053757,0.9367252761291893,0.2281562089536472,0.9768930466881565,
 0.4659457896249969,0.17230248413505278,0.600552244339097,0.546657523678648,
 0.3519181384746901,0.06831710728995721,0.003344519615690378,
 0.18502319593134886,0.13595573518118853,0.7080942287108801,
 0.1140881272513381,0.6863647615540489,0.5441412651712154,0.582521181350413,
 0.9884046179501508,0.6165678849537515,0.3670924477747399,0.7046010273343505,
 0.009399310775631498,0.7158616959042253,0.07228588319356777,
 0.9812005838627059,0.7543533622994544,0.019215561899108335,
 0.03400709911608557,0.8401935315429045,0.9392324406722038,
 0.15803912277902177,0.38782737404577217,0.11444047569439864,
 0.36064611921878087,0.959503321144113,0.321443780770865,0.08697556438484266,
 0.7689085570642427,0.5707353924593673,0.30120645915948063,
 0.19580255874430152,0.61325324276705,0.7368612735484328,0.57505082595671,
 0.670361385179564,0.25657697809260394,0.37977649122678836,
 0.07594717624496206,0.8730235763384315,0.5316230715636346,0.2922354149075851,
 0.7567836533479714,0.06528376545540404,0.18656201754545154,
 0.6024724220227007,0.09940110607774777,0.3499923134708023,
 0.38776959133124667,0.1475636271580093,0.13052972002487562,
 0.18472817015254728,0.651898552358435,0.1299387557927395,0.35739361102204636,
 0.9271543350457888,0.7759300718626269,0.640793361551766,0.40755169562510596,
 0.5167218055818211,0.9410288884188528,0.522010674750706,0.7262601415497281,
 0.9387050806844129,0.9348630007947903,0.01042275867881326,0.8607727960563949,
 0.8409313753999602,0.10058435133303034,0.43535280973592194,0.44410065241132,
 0.9412929647122061,0.3316952557674917,0.022153064871898165,
 0.7325505011095294,0.5051860087769777,0.884440293133827,0.4703873290281295,
 0.13894403970059532,0.49027604777403944,0.43163764534831583,
 0.969043090109262,0.9061589853285521,0.38784707213258574,0.8897834983573287,
 0.3938379842433576,0.8693607412269093,0.28086310745359944,0.8531034465094177,
 0.8961315712275264,0.7758486001053472,0.5457577084146397,0.5626057030033277,
 0.3928265890155975,0.8707554719154372,0.9237007633040583,0.49879515637107774,
 0.18344495026633645,0.8349940245077453,0.2636115744663462,0.4762456574630919,
 0.11879303953338682,0.5412925171173002,0.9376196177977456,0.9311725085736999,
 0.22546514546811436,0.5970972108332826,0.046956365309357206,
 0.6125238304370403,0.15235640469599154,0.7461591282876033,0.2367011378392404,
 0.9964254965690121,0.44859823486117645,0.438996169222467,0.0977518661906075,
 0.8384572338126637,0.8124840241183136,0.543557330607238,0.9730433904131488,
 0.7901910882073065,0.7464883455985265,0.9056740712128976,0.3349301620839801,
 0.17562256035354795,0.982508018574382,0.24838186001710905,
 0.15851238428507997,0.3110619440178738,0.35191555888142223,
 0.19484511905728918,0.267941389606823,0.5848573450624573,0.6266210362699682,
 0.4174655959494773,0.8595157417309787,0.7665260137757892,0.45638935978383277,
 0.3900848583959595,0.766701854073153,0.6057705157616867,0.01821929915597531,
 0.8483051450520291,0.4160744965780293,0.5248218211684842,0.42287196169413377,
 0.29891085555402486,0.2348755601153748,0.5148921910406002,0.1704044244040479,
 0.44437231270655675,0.5126827881030871,0.029126575946753164,
 0.42560500822957414,0.9871468190406365,0.18636907421084492,0.687468526405846,
 0.1173579551193652,0.8032328898017078,0.2792013093169534,0.25741518605083313,
 0.4895900775996639,0.2411562653562772,0.9324883680227095,0.8188576529267414,
 0.876479760016485,0.09544078996190664,0.14490766639470354,0.0954851759267703,
 0.6232071474966197,0.3156436979175238,0.576108444504035,0.8265214939286494,
 0.49088236222774695,0.45161144946019116,0.38457022773717797,
 0.5166586689443307,0.6316515605359146,0.7927814371791161,0.08100025552511836,
 0.2061441137917368,0.102698765490423,0.016771627815362455,0.4492112601523214,
 0.684323531856571,0.2794634306459376,0.6080356020763439,0.037226164111600846,
 0.4716976822543505,0.2579972119622014,0.3094348689468347,0.6688292186898055,
 0.6069571633254047,0.3449590649435632,0.5997364786329211,0.27094641692738397,
 0.6660550623220867,0.6274158247996061,0.9166792268925306,
 0.004979348243633996,0.8789677110839715,0.657411934272052,
 0.02638146639200789,0.9167309557080132,0.2315803888069481,0.5228479557679714,
 0.7550541674932971,0.1112291370223576,0.3530388177643786,0.976569281262763,
 0.24367097644032665,0.10284832109116804,0.035913883751343856,
 0.9048402461912881,0.6217006976203351,0.4675714812998851,0.12158168603745201,
 0.06467912156967959,0.2685604045632237,0.46385000386484243,
 0.8824868171173335,0.2827460503703485,0.21274514124181798,0.6366624575037205,
 0.5236583583126654,0.48465697458310164,0.872875226285056,0.34062638339967766,
 0.9071326127014735,0.6701442376756598,0.6600435024313511,0.33497343358579834,
 0.7395410520123121,0.7180023155004543,0.8136931263526785,0.2805884531273757,
 0.6892397410846708,0.39539335374175866,0.5417045711739268,0.2135246533127655,
 0.12153769421239213,0.8369458898534125,0.9080398927405032,
 0.27596503015588314,0.31071075895158273,0.5252291556739428,
 0.8583063454581429,0.34330974026824923,0.37259020325177383,
 0.8834614572374827,0.020349843936490974,0.41953336968536314,
 0.14966922789377723,0.911146287796,0.11945051097600234,0.270971231611954,
 0.8387933104503431,0.9350708260670679,0.899718525089415,0.4562671445369527,
 0.6667449835392002,0.9317459919472599,0.32135982250162853,0.3428529357008697,
 0.8149236428238067,0.9803406202089059,0.7138197827634284,0.10503167516787659,
 0.6453311610804662,0.34762514501750097,0.16984870456862022,
 0.19824053638529693,0.8694577119746213,0.4818855694747399,0.7042630991894812,
 0.20049311890346722,0.09271233247609756,0.12589786236227396,
 0.2795917316999389,0.3773992018962392,0.30116598469485345,0.4062442528234409,
 0.625673111350755,0.9916105517808778,0.7934233005519703,0.3344977911096291,
 0.9725282945441475,0.9344555298888382,0.5361900642977326,0.6294342767080133,
 0.035435065362682816,0.4875245594711898,0.24665963869508878,
 0.0034060145725047786,0.023061746313393416,0.08943181805145206,
 0.9227338471595959,0.27606014208099006,0.8814870296418622,0.6300251030341237,
 0.6962981012150244,0.17037366874478177,0.09182666479978607,
 0.35566939165035794,0.21337324369603616,0.6479711120063796,
 0.32007835310680177,0.9171265224873144,0.7480504362028206,
 0.24082360186462892,0.44386374529459083,0.5590442190727587,
 0.6361147811952066,0.918535249435299,0.6549448731706489,0.7078660569906541,
 0.6138987146673465,0.27084658225721847,0.8823588248833336,0.7417014146018504,
 0.5315165881849309,0.13538647215343902,0.5405752106994056,0.6211871114591697,
 0.9972095117045079,0.004131274357214054,0.9531515264628448,
 0.8291936010309482,0.7098955985843877,0.04918482537215485,
 0.08928874560324163,0.1661921322970309,0.6626418855873886,0.8699989205605757,
 0.3128550461581995,0.4706689848934079,0.6360683236739515,0.3848508509193437,
 0.7967887395303137,0.5252775792457802,0.35400021323931186,
 0.30697033904744053,0.8269115266303827,0.06273948025066378,
 0.13921473399160011,0.21494009499986766,0.2652776953395649,
 0.005012117051839682,0.1267257873108707,0.1750995626067452,
 0.29830686180113763,0.6535552807296474,0.9207822816884792,
 0.46796065434409373,0.9261620945784409,0.736652724016804,0.856473624410238,
 0.4239394786530576,0.9843342231418875,0.37731752510007444,0.4048392420927355,
 0.34081668638390916,0.8567378643453848,0.8830290750006488,0.7615032654193866,
 0.05640642201620216,0.27602940238688933,0.05192912249791459,
 0.11380772456876043,0.7407771912176714,0.9747352649017414,0.8169859357908117,
 0.7071013169280994,0.40919384732842456,0.5854347615700396,0.7516949198009151,
 0.754306579547612,0.06790887358859543,0.686170766919207,0.8960871538124289,
 0.5302264065761274,0.7590738681723286,0.1721866985694862,0.152998942160526,
 0.678121826414459,0.5974867001604509,0.4709586987921537,0.4748415922301782,
 0.20547187179421422,0.05374209557785947,0.24312430222266213,
 0.8161816280720046,0.20806943493526364,0.2980395931556291,
 0.06311650622157738,0.6088268825021971,0.3384273264122435,0.7227211539575115,
 0.36457684874131957,0.6381709519169422,0.06156508873656019,
 0.7902694771311382,0.5853542568994871,0.02955175937096277,0.74893101476348,
 0.5435699972985935,0.14875166039977805,0.9432473390432143,0.5288952495061335,
 0.8610315328847478,0.13806136851477924,0.33730082084169444,0.882380603297385,
 0.5249967819821443,0.058983819053651465,0.1796598226987174,
 0.8334941112645844,0.5644959438652067,0.03693634946894864,0.3291670110272139,
 0.7704879339395447,0.19347506279886595,0.09098461945330738,
 0.40075955875845704,0.05906484378098131,0.19693716037176545,
 0.21880874212606405,0.4591988264930287,0.2609428423567166,
 0.046832328618010344,0.4947802321956184,0.08019078099316823,
 0.15474445067582288,0.6607474299508292,0.5406475554776101,0.8665034022165032,
 0.19264927178087443,0.2573525084340079,0.589891596248835,0.9747968137630277,
 0.2406787444640428,0.43107236989653486,0.6467149155684881,
 0.20926229746944125,0.29762274349132645,0.25681636728849133,
 0.8461332081999289,0.006870173035787186,0.8973506120713094,
 0.38978319163458164,0.31044562386454444,0.09709043953369312,
 0.3187682258754462,0.6237490508030811,0.8073847518669159,0.6351797028781921,
 0.1173756073968738,0.14266409839653482,0.7602746236324206,
 0.12596452519013024,0.20794730376705872,0.3560534968517173,
 0.9018991404385689,0.4793090397319326,0.026271676360015594,
 0.4954314222581093,0.19361961239183834,0.09915939010457808,
 0.41258985571241724,0.9785181889282373,0.7275321340377389,0.3368204452873623,
 0.7616401620058544,0.3904811889272044,0.978176900418018,0.5363855284688945,
 0.38942795288148513,0.2718639115959328,0.612646924954844,0.19518042304608152,
 0.6358733533966012,0.802283714608321,0.4900779924748211,0.31630227693696356,
 0.2845544741038746,0.984816226518363,0.6198107233919005,0.5128303774195239,
 0.840494364374588,0.9034058963722262,0.5387658037426368,0.40842962855690956,
 0.427635962339874,0.9223059819423596,0.06796452079151782,0.8823556949645173,
 0.12042929047830342,0.9325371537556031,0.804111156841435,0.22482909311044752,
 0.1630072878301333,0.23281045867925998,0.7223282814265155,
 0.18500579702122233,0.4053267851452813,0.8348480595655549,0.6794293561194079,
 0.5098744592312454,0.8150038208702242,0.14032527733479505,0.5302361197822341,
 0.5703985446163009,0.7605323973872311,0.0293964582054973,0.3186070355977749,
 0.1520328831374823,0.36706552265106074,0.04769006128520614,
 0.4325895634757796,0.8360816477934747,0.5436144144188287,0.15369134706014664,
 0.6471768298736464,0.32342376078819546,0.4571556447369006,0.577859753344206,
 0.708354996904169,0.4541354862403295,0.3126563591379411,0.6693701829388649,
 0.44351947969934136,0.3918525248973319,0.2537558113389353,0.7715218402439024,
 0.713711351450311,0.3854718450308464,0.6283195146487968,0.7130586774213535,
 0.6512404304018049,0.22636112888800397,0.9801554817289441,
 0.20981890569707873,0.9790998943066493,0.4007822035645997,0.4323426835389471,
 0.09304496949671026,0.3969406280288921,0.023738236253564415,
 0.08320889934760911,0.347311047003261,0.10371134291113338,0.6455270868828576,
 0.27525682240824,0.8018897640589188,0.9644090502886303,0.7414320098306179,
 0.6745840162583843,0.257273671108349,0.4107843899377447,0.393977300264434,
 0.05836050232488976,0.803056472316021,0.7768067737379872,0.6330316661375792,
 0.5784987551373212,0.4316541271804242,0.8163981975498354,0.9544015124656825,
 0.38380413227837806,0.3018575087252652,0.3941732347809894,
 0.17382603064141833,0.8922909376593259,0.6589314541290625,0.5009226708481144,
 0.13076627317132017,0.8042408938576718,0.11467359027551072,0.82835106824692,
 0.5756563604858621,0.34163326582537856,0.498937037192449,0.06898355419932367,
 0.8462787967104608,0.6519837438691193,0.8859512401666512,0.4391276155854378,
 0.18762496738697143,0.802731912942675,0.2870572422982228,0.9439774182289946,
 0.291143025451452,0.652426915916364,0.24122493079750174,0.5219111369410836,
 0.9039965184739869,0.5348093487703468,0.5928661669184392,0.10781385870036264,
 0.24775589128689757,0.04976281796911097,0.21086484547136375,
 0.3302880483232419,0.22367059287788538,0.42186935811184856,
 0.8964986754454662,0.3480303564593781,0.940087331812632,0.6076469024789561,
 0.25198518138059744,0.08632346012172298,0.8568510477558996,
 0.6228884022606472,0.8482244680153216,0.7166383215771847,0.7460603279822936,
 0.9492453031234873,0.3640834039779164,0.32468332894398644,0.5998843724756894,
 0.9641410958846385,0.10121164737384003,0.4041459870495163,
 0.06793285826022166,0.45176123075395225,0.052197232075863687,
 0.27651093992495945,0.5538366311601752,0.7904111410868482,
 0.22845842065066468,0.08887973077087652,0.2885823012333141,
 0.0630304835196761,0.22559128188598132,0.516250006688618,0.7292072968448547,
 0.5339399844953138,0.19139000524378413,0.6616750732903102,0.9175117741368806,
 0.8748216662651431,0.05501603730621962,0.010209280372723284,
 0.8268400197501051,0.22839146916228192,0.4822535258676012,0.7855977964240266,
 0.6518191661649642,0.36813653147079795,0.4014534706989119,0.7784497540385286,
 0.24344092450469068,0.13963139923479473,0.6785660782873102,
 0.8282461332992916,0.04161524583178533,3.614760687226104e-4,
 0.020956291754772427,0.6447962013024373,0.4075488442395555,
 0.8456961001028569,0.12063978646615836,0.280986130522322,0.771292986040791,
 0.6228896593774321,0.21205373268091843,0.919692604281417,0.5485763075409161,
 0.34324769141229106,0.26209948015198004,0.35184864932838644,
 0.9328570615720159,0.3268478136056705,0.5647364675336862,0.9766157994889852,
 0.7545122729750339,0.460002783695727,0.3574714247990689,0.6454913971571133,
 0.8302824165464658,0.30968558448205874,0.07953410029616115,
 0.3567442356675954,0.6189541794298243,0.1371199579864103,0.34633240345919647,
 0.5771730431255424,0.6692131173244132,0.897446356037428,0.8108063626903951,
 0.5345691292944851,0.5418446286859503,0.24182849199741463,0.9473805605007006,
 0.10102594696905809,0.3099507172311179,0.3263786963896762,0.8422131911205852,
 0.5941827057318706,0.9134994715218708,0.09281669546386839,0.1434229597726232,
 0.49725748384389823,0.8909429620496323,0.6691177374139068,0.9906177515540967,
 0.5351178313447201,0.6634147775906997,0.06790122642930785,
 0.22695715483605383,0.16221000017973863,0.13090824203042573,
 0.4539209024814831,0.34694373787846766,0.7650571750635057,0.9138690866365288,
 0.6711494770843337,0.6973215421906087,0.42353689687474794,0.8502919134203201,
 0.01442677296037731,0.774246262343981,0.8042528831183205,0.30050906314173254,
 0.9844983627224292,0.9752420021877596,0.25957050540453486,0.4301281419456142,
 0.6931904328195645,0.45997780482351835,0.11282398144234074,
 0.43231248382470855,0.7515640888936935,0.9469248458068293,
 0.014200180417231811,0.2710160562429924,0.7414329481295937,
 0.8466251899379461,0.3535489201438786,0.6753097517403294,0.8891697347413874,
 0.03971239320502806,0.2776286770347449,0.7571220735610944,0.8335840250395907,
 0.9783660088267114,0.9083256869516108,0.6041368577748543,0.8582479730265556,
 0.14181197526776612,0.9570157018729748,0.7423949631314041,0.7044389618251284,
 0.7430911590127752,0.04764372370026959,0.06176319493574778,
 0.9728320854141369,0.7829753957360528,0.9268960675707487,0.15363241356454505,
 0.5725400845545381,0.7065769859326146,0.8334120324323967,0.18952321103575032,
 0.20443757260891393,0.27826152957854466,0.44292335204874245,
 0.18342580429781818,0.350775007800341,0.28803260604206093,0.3584680274667922,
 0.3324512156341257,0.27000462225468347,0.545618463492322,0.8397995190392264,
 0.01762417207082989,0.8599633235598599,0.183761615022502,0.6819171791301367,
 0.3148453622127694,0.2619171741725662,0.7995716337911475,0.11339207866832979,
 0.18405111374937688,0.12524135146148085,0.35260251889163163,
 0.6561576422608666,0.15088383603044198,0.2072412563544742,
 0.03659954182139513,0.3911359901578614,0.4325347344644135,0.6274981348362556,
 0.886083593303883,0.9525249196967965,0.30885793723608534,0.7745389149828232,
 0.24640360792992422,0.8690384313678521,0.9224576522043353,0.9783027828006046,
 0.10089336449884445,0.8349046623320038,0.5274931815464219,0.4473832026949148,
 0.26884013066384516,0.5813924661198692,0.47802173842203666,
 0.9281378768280084,0.9632164908834965,0.9079034733229523,
 0.020666306470061513,0.3224739891780074,0.2704635047335905,
 0.7865130465997353,0.0016808209682821484,0.7770573713175375,
 0.9358175209043349,0.8612881455899126,0.7393976076518364,0.4058693335618867,
 0.012854206071689966,0.25794908739902567,0.8427455183764975,
 0.6735041970584598,0.4136916058653296,0.5537937444948647,0.8411300393822363,
 0.9204730941258455,0.3335405811835289,0.2551858569531411,0.6517623101181999,
 0.7279268135443536,0.348999265505481,0.10955474575781676,0.03128031582262725,
 0.2503069819892102,0.6239610477273144,0.7938160849320665,0.7973060658385072,
 0.8240653492625766,0.7711119019733845,0.44862611753603465,
 0.026655573979197156,0.5963058754299181,0.07361444299203379,
 0.8289229852867313,0.3556469450678089,0.2570049666939047,0.7852409367542306,
 0.20835179384288227,0.43060912035013255,0.8260025762066304,
 0.32240761041305754,0.02979860581264515,0.9852984322829699,
 0.5299863036990036,0.28633115225769346,0.22209367106591993,
 0.9837921918031974,0.13091934384268467,0.9792278909557401,0.9423444117386007,
 0.06430356843927665,0.42041292613858905,0.8688222771307604,
 0.10674789989859801,0.707068194942797,0.5859009044850789,0.6117062652203767,
 0.8271480043760908,0.1775195504081799,0.7557275735249295,0.5754809800809224,
 0.6747646453556058,0.829521426888405,0.8037943922929792,0.786529732620166,
 0.5311231057506469,0.09439056270477442,0.9972632587210591,0.5887371809305146,
 0.042485161112183145,0.038714723542017215,0.3774314949105032,
 0.8800427997536409,0.8159414835316593,0.44692946864459115,0.2718077848840834,
 0.9995850153781317,0.46944628545529676,0.41972968166004276,
 0.15794681054786697,0.5130132041372266,0.7046194048603679,0.9592037358711177,
 0.01339575358887024,0.4404597710799524,0.8545553307517676,0.3498501538734923,
 0.7299289733358444,0.8467147633583958,0.2364905063509355,0.38676304120531546,
 0.7072929087807892,0.9176543873634058,0.5602054854657323,0.680712438844174,
 0.46765466869194483,0.8760993918162135,0.8347741433331392,
 0.11623619605710955,0.7146339975685659,0.23867610389906668,
 0.13150726746093722,0.023753834809416396,0.21198422712893206,
 0.12228756052221312,0.7301496173662684,0.7087449327723321,0.7393186204018559,
 0.9043215514983617,0.2656259048583044,0.4757368085319177,0.16985487623475048,
 0.9359301875668705,0.7176102881595215,0.017882929176034068,
 0.7676993178700471,0.07834550404087581,0.40674082315058224,
 0.8722686810944554,0.12812370398283746,0.3156718504320023,0.6132761314611851,
 0.11006789223197622,0.6651143978976142,0.7821072974772583,
 0.038963974466157625,0.1296386636979775,0.24008587916981494,
 0.49233757850067594,0.6255582218022429,0.10759836140836587,
 0.6974514993455235,0.9757224068374255,0.8026178828721302,0.42781335587735736,
 0.7798938165642529,0.6738064696140805,0.20233939017879543,
 0.10151494997828248,0.3907032585184791,0.9289136903852337,0.3179824610736921,
 0.6220378937641275,0.8170273589898676,0.2000571903932028,0.17120943260443022,
 0.7528931804916206,0.8778269670883077,0.6034536408219795,0.744325243918683,
 0.6702096388109449,0.2628098187887703,0.6286819073218298,0.385388378853461,
 0.3085032157717459,0.4030759685978026,0.7352399071230885,0.17085627913965684,
 0.8319835616857648,0.5385273786280345,0.8716354452390079,0.6989078582812511,
 0.2539596242566252,0.40012335568979485,0.7755908656517786,0.7388951060786675,
 0.27697698507630664,0.9424213812474109,0.9619604432601976,
 0.05308119980673798,0.5687752948757806,0.7780642425341929,0.396446473339639,
 0.07024925038142571,0.14988333551568467,0.6940285642923321,
 0.4397397657502973,0.2517661453162615,0.7983245050651442,
 0.008380384592322265,0.5673444355172125,0.7519553213087911,
 0.6910286427290384,0.11317286017337047,0.3607690751858246,0.8323242177561694,
 0.29455525598493637,0.12912261735037855,0.7517228111742514,
 0.1429848253891498,0.4037340332767454,0.11774755661255165,
 0.052201361165215454,0.14125579223496176,0.07742156385869847,
 0.1894181084963048,0.12820653738471155,0.6825336218105208,0.8700773105338396,
 0.4777717826890183,0.09664443078387108,0.22635356949874308,
 0.4986789010301125,0.607150599547933,0.8336018315579716,0.35527718712580125,
 0.04498349301806104,0.6610144154876303,0.30135325645188105,
 0.06458765886701945,0.3381532430348885,7.083002678336392e-4,
 0.6989318050888342,0.8084595119641684,0.8411171809860796,0.14313503732841515,
 0.03684749437236212,0.8375771833380583,0.7579532408309618,
 0.42770109085724695,0.6415383489920435,0.8874662703410371,0.9794534200802064,
 0.7046322209817952,0.6638136372907947,0.14367000270846908,0.1768566080179087,
 0.780662636415615,0.4387566816949313,0.3932868647954182,0.6893968498980145,
 0.7194271547323431,0.9154905372376755,0.3425737941220016,0.0211884849301055,
 0.08113687606929132,0.09829200582961972,0.7831185046302033,
 0.8801087261123195,0.4982738542561864,0.8766617884660988,0.33025356490668956,
 0.5344351114565443,0.7833890560759391,0.4936821856128485,0.6331974244238776,
 0.3587382008471057,0.5209836997495494,0.4367649123231495,0.9926302504206715,
 0.18323274971977344,0.45881514787680555,0.44698690871763924,
 0.859516181755417,0.6665633899691452,0.43272126301892777,0.5269081243599055,
 0.4330200008428372,0.9326139034632098,0.03046611705327351,
 0.11314823236350413,0.7654211433792136,0.21856230431265722,
 0.8729026134521759,0.2701833946578027,0.07830306572770518,0.8803227561108244,
 0.5391564753825983,0.787262771105349,0.7518610832671921,0.3084470008633847,
 0.53359481934685,0.1848426452177402,0.975018406765447,0.5827724330357034,
 0.16340565217219805,0.1060717258534628,0.1864825886808975,
 0.38234132980850266,0.5547281267576003,0.3351422197069307,0.9372827722328858,
 0.6976816625005217,0.4813312337994018,0.382243536269952,0.31629843341004993,
 0.7195451415411639,0.09035746339648654,0.41732888071843877,
 0.7394602516757983,0.092602385051339,0.6376871954401505,0.7584790162857231,
 0.2525688394479947,0.33173067217500196,0.45899080815338467,
 0.5729688434394947,0.4062897787969064,0.2011821172766851,0.3386708136825012,
 0.9982984944994007,0.3682466550771102,0.8333066804416009,0.03360692304071755,
 0.003136945794028567,0.24991060711453716,0.5357019269980127,
 0.7108840946067806,0.8004862292565883,0.8296116230505084,0.8478987644930432,
 0.33843280412619303,0.3332014484651288,0.3572447357934372,
 0.27531520284015987,0.5557451586056462,0.09422153890722851,
 0.39219784907211697,0.6528187021923146,0.9227614756560284,0.6358548558898651,
 0.9793169759475654,0.026609567284551883,0.8132065003263711,
 0.8437687703210112,0.10540325660757888,0.23898494868914955,
 0.8573277586337733,0.6733681044009807,0.8522239419277234,0.2991625584163813,
 0.6572333396745617,0.07378981610248103,0.655788799227866,0.10497651799085583,
 0.1299125092049207,0.9689307292346201,0.39353071155933705,
 0.41665288979424553,0.20743919760641016,0.8605927820637652,
 0.2065599881272936,0.6641359697635072,0.08349469155549372,
 0.38745019459617036,0.16993284700687716,0.2115122315362591,
 0.8353706443803273,0.0024991941520695615,0.43991837212213003,
 0.7874951056718671,0.40835977777367505,0.3543955589957063,
 0.17196229693277287,0.308046965728316,0.2713201837215251,0.09973368194765964,
 0.27923129104816835,0.47762667788086965,0.20192846043650436,
 0.6564509432152923,0.7241552513551905,0.02774893103078857,0.6065013948376237,
 0.7029176279597281,0.7183499970700455,0.35459830175901286,
 0.07494507332470646,0.13383119972729451,0.4905040150036579,
 0.7748496212576971,0.6461385104434869,0.3009147824066264,0.23685233290958607,
 0.5701126546875482,0.5234301312531637,0.9351020606260045,0.7447407370738123,
 0.44726581362491313,0.5589392526099863,0.4884055748045467,0.1763125749307508,
 0.7637749006313954,0.8831356932977422,0.06932714499019565,0.5102333583965376,
 0.9874075659358159,0.4635181049194367,0.357426325582493,0.5123316274102939,
 0.6887745768728907,0.49716629301147874,0.684180276215719,0.42394138070350895,
 0.9460158925314774,0.4546029579947023,0.5246456520307,0.8893634243580881,
 0.34000189207077824,0.23181573938136057,0.3362318920307551,
 0.7180968386384657,0.5647118165537508,0.1322569369240738,0.32396349668953583,
 0.1139490507851717,0.23258620435377209,0.4049289978009877,0.835956529312619,
 0.7687952087613724,0.496371855917754,0.999710667304322,0.06401331336875504,
 0.2843166350516537,0.03663673704839476,0.9398762242525815,0.8911771908467877,
 0.615959655367552,0.5797932264143877,0.2956333307964496,0.5642381929928236,
 0.6940600396210277,0.189309709096235,0.5867460616837726,0.17587284399507563,
 0.5401202061158007,0.02864155107223698,0.9334317778243237,0.6482064218509862,
 0.8982184511465879,0.5581260744644719,0.8133653118454638,0.8431080500219734,
 0.6570860970536558,0.17592226770554809,0.3797855771926697,0.7537965924699485,
 0.9224455637176786,0.014411535064102976,0.5093472036749194,
 0.8393286978205816,0.8483377535166813,0.6237730342716253,0.8424798109486217,
 0.5088513426130523,0.5150338449749375,0.30983163275578507,0.6584839724379291,
 0.44178378163391996,0.5784472562491629,0.5534713579410192,0.8630145719103549,
 0.47667152488337117,0.36699422341980914,0.862280892791046,
 0.31615773967779237,0.45977886050579186,0.005750539956681511,
 0.3835607227067721,0.17920742620821273,0.6019102270161828,0.8536462844565155,
 0.41487380127382234,0.23933826834544258,0.8486205662765162,
 0.6783438535026923,0.8864969368923344,0.82913987194856,0.5654093219535739,
 0.5459735764694383,0.5134821544484717,0.5297024953179462,0.375794488039069,
 0.4990679504854082,0.6755794799807318,0.4486536531610449,0.16456027855160804,
 0.3255034777700876,0.9257125018052875,0.3739623260526369,0.03180849229236893,
 0.0549054258549696,0.47621634546072333,0.7225013824051736,
 0.09649823581732919,0.8110607732170136,0.9784412286859394,0.8399218370261414,
 0.9034799698417046,0.0636757766953906,0.657551520510558,0.5257678301484449,
 0.9951493523721485,0.9855796704593346,0.09328518956389276,0.7277803703709614,
 0.5391391313290643,0.6464091351545133,0.31675470564182084,0.9890687944633855,
 0.055552083830019194,0.6012581538133616,0.5549310458846072,
 0.03672738540769371,0.6944082613416809,0.7848785502539192,
 0.008440729222982224,0.6785657910781555,0.9565133377385577,
 0.9270279898119025,0.4423192010478444,0.8234374523045273,0.30311830834359743,
 0.6230510763776607,0.5574613911792786,0.24913162733554595,0.7064497568192519,
 0.9251062417834085,0.21495279534449285,0.07389030905203064,
 0.3705864559714911,0.887611726494946,0.4471653071338032,0.9376978615729278,
 0.4059942339940952,0.4961863836373561,0.1691180060282773,0.6427788089473564,
 0.405090759048389,0.6829612963187576,0.7198480764970535,0.48816852320379234,
 0.7130058714914328,0.44291733372234343,0.832212396402476,0.7279608393356816,
 0.6920945636486896,0.8858333589833768,0.9168227498588482,0.4184427394848951,
 0.714356411156317,0.3898899957578885,0.18718807929681425,0.7073393882493091,
 0.9593113446407507,0.46764290557203547,0.06355856117064229,
 0.23751220124152284,0.5002003799003236,0.22633868485528352,
 0.7305971480889719,0.3117034803433275,0.8047570691223264,0.11342936946839832,
 0.0074507169563970965,0.16675545928418,0.19909579860300508,
 0.21433408669814358,0.6832829801900896,0.42129114039317583,
 0.0656940851312322,0.40115116119122074,0.059441723082467424,
 0.6010440762963924,0.06763944744948791,0.8905962028723919,
 0.08361261341344361,0.7036924665756601,0.569373275544504,0.2616725353649664,
 0.9190239712703844,0.05294940658319369,0.4148587556372514,
 0.04343427841304781,0.5297392627057745,0.721144057944084,0.8455765440404882,
 0.27858199026047514,0.7387891356868632,0.4706287807350328,0.5432736670132841,
 0.6809172829436831,0.2922025056510069,0.16545716891923723,0.9282852286078427,
 0.19609897868548132,0.5029466625685791,0.8584097064518265,0.7719961797597865,
 0.6258701719867533,0.8752778135449243,0.7744059713072515,0.6889548241006348,
 0.3698430997539144,0.8644733259169504,0.6635695029828695,0.11641830108332529,
 0.7933589178094248,0.05546523490950228,0.09450508129045487,
 0.8703445463475883,0.9543161446797237,0.35424279365859135,0.6969064598843377,
 0.24250179435251362,0.4411803442021752,0.7608283674190487,
 0.46308182205314075,0.33097149855963715,0.7744086295844785,
 0.13023815542178063,0.3713867729851402,0.24917025835439222,
 0.4204167315910209,0.37767593609904926,0.38542326444694475,
 0.9070495942105401,0.9388278385235787,0.6819033967256191,0.7454752217778617,
 0.9077866530129726,0.1742713379992482,0.9228400629406882,0.6439858062908932,
 0.8243499810323208,0.2191432547844271,0.9066730992752688,0.5871284229098507,
 0.16874676998625315,0.07650195166202844,0.693467916320227,0.9941606346151819,
 0.3011074255268551,0.6906967821041315,0.5921563289197624,0.6220911545617067,
 0.9026270003203,0.6073159851625662,0.002435060892971408,0.5281851491092744,
 0.2432088177733185,0.4672902324996333,0.6259842028767467,0.8733652941956855,
 0.335331625755255,0.32025549015016197,0.1822815422358376,0.5266359939995005,
 0.9547385754979603,0.2377376737153506,0.9229149682519413,0.22787684329897395,
 0.4075713710364389,0.5846497678803582,0.9609751540950633,0.05418412211496304,
 0.5863798084046985,0.9789895017185738,0.5580385563373884,0.3097425311280899,
 0.38915223978810864,0.5318046376954098,0.8105778948755865,
 0.035669874298173365,0.39747277464830255,0.05420330922898409,
 0.8428087097798709,0.0948546430476489,0.06313359695427967,
 0.33860714329725805,0.22491027017387222,0.10271004309243226,
 0.15580351904848477,0.6764486981655754,0.2897785299293929,0.3861752997003083,
 0.20128956375668994,0.32240599815947046,0.5615193592260698,
 0.5131028725282647,0.7949036059060414,0.9704687468766491,0.5646386641461079,
 0.2497280734114854,0.15669588972084558,0.8117670087297165,0.437232848277581,
 0.6776094741107243,0.3002683294322164,0.7658523369222271,0.9531454211937707,
 0.5434815711117353,0.7199521844884835,0.17958594780644266,0.9246863301448208,
 0.03885719628358131,0.9125592445496274,0.5036138082787202,0.5246907880197375,
 0.7129635593637677,0.2602850197358618,0.7184649083284321,0.8934763495038756,
 0.20496600931093167,0.8516277655517396,0.11814997655059623,
 0.7872674451566358,0.550552142206458,0.7865000046626793,0.7953939441869643,
 0.08271933030120493,0.08289702759277795,0.6362582470335125,
 0.35577929442399403,0.754949281823472,0.9221336389406498,0.768659572644436,
 0.9275048894177393,0.36827904619128726,0.8720761669296733,0.9117135443325003,
 0.5885102142961443,0.7057592966336003,0.5747870552226443,0.6992310528711229,
 0.9370774168219534,0.39237124369869936,0.8378855606030362,0.9798065008941614,
 0.05092286997374429,0.5932634785834409,0.2751240669666598,
 0.38995586893308076,0.25048789403523664,0.5552991475461765,
 0.0467579266499506,0.28373845911276807,0.32482521174965073,
 0.7172880382046605,0.5361164260279705,0.24084194979224116,0.7055917414294879,
 0.21625664549620727,0.7170320392476582,0.24709812749744708,
 0.9847306751531621,0.6394510394322791,0.1622614006654598,0.5531113977946834,
 0.9769647733957965,0.8717139647604827,0.1608334565351891,0.7089905755254207,
 0.20005442374643123,0.08218154523565502,0.799202392477533,0.5652578932227909,
 0.8310930601953886,0.2730928146515973,0.6556917403273758,0.633210899860814,
 0.2273877233336643,0.6330411705932132,0.16320496121151407,0.3926126292811809,
 0.4332508334861451,0.5817329329078493,0.03334969936818932,0.8884701454757317,
 0.6763504463854861,0.4878939075704396,0.25362814976487014,
 0.19431389427913104,0.4938411022293848,0.4549454245151644,
 0.43447581272439284,0.07816849947001914,0.19107683085076366,
 0.7678285072206418,0.6726711530021765,0.40357094282808537,0.7285768818917236,
 0.859427905649011,0.1810680957000802,0.3992789232352869,0.2928781566093529,
 0.8495297278486429,0.9725551610068189,0.06216599279440127,0.7747687130596688,
 0.2381104288096455,0.06627022146942618,0.7057783461548646,0.3017104195716892,
 0.6403485582840931,0.13524429174749775,0.5990666648878936,
 0.45452878361282845,0.14925458603322184,6.072887655274029e-4,
 0.5883185629005951,0.9830440448072989,0.7487664122458444,0.9295561717139399,
 0.8221116962160416,0.17596108379123088,0.7974379210025969,
 0.37672657149617406,0.17376911668831008,0.6733071957071668,
 0.6624877944799865,0.15351652362559676,0.2810569065912769,0.2955738396073435,
 0.3757957853043309,0.10482269088929574,0.6386533436068076,
 0.10258740053496707,0.21155084310840688,0.08215970572099929,
 0.911395551123679,0.5966962180758623,0.41869532872236903,0.6872894834504519,
 0.6999087775908037,0.8682804793042628,0.42009177753155147,0.6920939330867875,
 0.3451049267290236,0.2995291534383413,0.07503395803431046,0.9675064801254316,
 0.2818960154573299,0.08977336166294081,0.04444007194888555,
 0.4167660850606606,0.16361009346468092,0.889364455758257,0.1953296850790458,
 0.20949960654076316,0.6170236600964931,0.0013019202271481456,
 0.014145289621664281,0.9653893783958698,0.05476675449377033,
 0.15251587550620205,0.9938910833715529,0.03387878121300136,
 0.8955671522809793,0.01743663934445039,0.5940498931467729,0.5919169507498976,
 0.30339548332294053,0.8129642890458612,0.21836868433483847,
 0.6301294927453492,0.021010957290584553,0.4667191420218042,
 0.6769844158693838,0.7708718973455329,0.45191610918888203,0.7759811807740234,
 0.764903963920583,0.3203255093163637,0.928478788939932,0.3106690375078617,
 0.651561136151953,0.8008679509904642,0.8922678842568451,0.28933238737160705,
 0.13170837641904454,0.5288487172517773,0.9678398499692883,0.3138735793769676,
 0.8348691232050613,0.333027254874965,0.8510534401497296,0.910293105341475,
 0.30622554942442604,0.6759263213527584,0.3223165107346355,0.5161417305745545,
 0.9921265074771729,0.7562809728385603,0.6700356266347831,0.19170616841822075,
 0.8187699091305028,0.5607709382375274,0.6732211402946728,0.5770406661471222,
 0.044108461134380494,0.321842422468278,0.05068984620692618,
 0.5336316411003454,0.33848919282178935,0.06488481145867897,0.770627400666411,
 0.4244573673344796,0.16012957003520467,0.8130115997747949,0.7542367669656124,
 0.1738540489258712,0.720070363497135,0.7641663245564282,0.5904961212243628,
 0.6297000978022619,0.47792307424443115,0.22632752216807683,
 0.5479918683386011,0.023642031277811615,0.29666222322732616,
 0.991253398800723,0.796670653063678,0.8219592912068555,0.8346151144740566,
 0.0665268950682637,0.14945706966806493,0.8828884237377491,0.8020781154748569,
 0.7400995304025095,0.7145030682161109,0.017966449640726356,
 0.0015311104585087332,0.8256028656799963,0.008706585811462553,
 0.753353633703278,0.12787506581098496,0.27196665305288414,0.7024440225141692,
 0.8449520333074384,0.6253815315249183,0.13178001862896893,
 0.08949807324141634,0.5294344422470902,0.0854094258315291,0.9587721706021888,
 0.14756122655667603,0.1747769871199526,0.90498731041332,0.9471558638181429,
 0.8190958176721895,0.8903299357784593,0.1213598255712266,0.6287601701280576,
 0.9342453063417082,0.11544220938053373,0.4092236824645952,0.8127155978884196,
 0.13779006650043069,0.9651820229683894,0.7034553856981175,0.7943057646766558,
 0.07865386151777143,0.5924285980400137,0.24218892486841037,
 0.6722012434948975,0.6465364999604942,0.8853399489244675,
 0.010327740005730335,0.4410256136359547,0.8113535078554203,
 0.4809899980168332,0.5175456953342006,0.8641092314637013,0.2219570679787093,
 0.9750747783869552,0.5821410300685651,4.924762847625663e-4,
 0.6399944402526636,0.08356579138631892,0.9785792161460859,0.8579862543354511,
 0.32186331357896536,0.7415527512805105,0.6325748013379069,0.3496361601963959,
 0.09630459977312744,0.3658050517729693,0.28335123336912826,
 0.6737694987399684,0.14381521619134396,0.6080712395004484,
 0.01696771904113381,0.40305094308402756,0.5489706179290323,0.528867316746999,
 0.48100211998398823,0.08368442188027225,0.017314040589861857,
 0.5577131301721737,0.2915815669145023,0.9997889085940421,0.05733567087666336,
 0.3722005061227288,0.8629715405921828,0.3491533509438063,0.2845725809419831,
 0.7781509803781294,0.13689225579095599,0.6973331978897669,0.3975677607784269,
 0.6359145913977527,0.7483690360989476,0.26025024956103004,0.7194646421257862,
 0.2640139765934396,0.6301713311188598,0.15923072570108765,
 0.17421777643117653,0.5125523239707939,0.39884434941417757,
 0.8303073059216745,0.5326412436650863,0.8230958536368356,0.2985887047291156,
 0.03253364869427555,0.05159210081067633,0.23257453759758018,
 0.8831574980491216,0.9509680720554557,0.038740228774886476,
 0.46464530675436055,0.8476473257424963,0.1527492888516201,0.8645350559573618,
 0.2726879187191815,0.7905164017387006,0.6894835802411009,0.7261729488876666,
 0.3566161766667856,0.6126540841781242,0.882576429500948,0.8509391154581812,
 0.4977079611548202,0.9549903255304986,0.32219657277834834,
 0.28759146456667495,0.2646083272429446,0.8949444655757192,
 0.11582799705600266,0.7218698835278017,0.6617513190862321,0.3591563545588107,
 0.6803989442722005,0.8749473583251166,0.23978632897127383,0.8571951072209527,
 0.3085561082678342,0.44885484334099157,0.05122796567845023,
 0.9844304568667287,0.0420561862404889,0.6995087116162125,0.9813516781879953,
 0.4008202536408213,0.5888058442783324,0.8634133360770184,0.12565686666981424,
 0.21938597715611274,0.7253182880733672,0.6669364053472033,
 0.44958539941847775,0.8671554627631706,0.16809282055525432,
 0.9854150263964101,0.9954868719840166,0.7400645116959909,0.09692225827840961,
 0.3836477083205707,0.11419362834371793,0.25253532541972645,
 0.20529390496531463,0.5075286371379405,0.036465201213430665,
 0.4702796060202994,0.2220433279411017,0.6059861088221518,0.36220701932811183,
 0.07979725534734872,0.4867050127229865,0.9301116683307378,0.2477059459943899,
 0.044850233728026434,0.7371408885140601,0.10206631209965655,
 0.8120242320617729,0.8901957872241415,0.5067546624591994,0.6840243683644024,
 0.40285532851031336,0.06741386068801114,0.5469569187067913,
 0.39589280977791386,0.210150402717588,0.4085149398874225,0.6557231551831538,
 0.7712094693452795,0.3607916030587791,0.5491720966476147,0.5998345809900018,
 0.9748099514271622,0.21298343491314242,0.15415278003890454,
 0.4453322407009206,0.9449110188106449,0.3606684012594443,0.4503806579854528,
 0.08927132257570669,0.4585065980326143,0.5129515828456469,0.7057142957992402,
 0.4746804475135159,0.9483067373367605,0.9960057878139547,0.13688668083104372,
 0.8348271489462272,0.2808317156804877,0.209072232275886,0.29192362193047217,
 0.6255014298525846,0.18910609113055143,0.34044634925533723,
 0.19609149894365252,0.7438461124550019,0.2091917583088866,
 0.06820330534095465,0.3136252212767665,0.34814216547781285,
 0.49808926850284074,0.5368256144630551,0.3029239727233852,0.9721720044390174,
 0.7779360822132786,0.19990352628877983,0.19934613017158553,
 0.23241689232961915,0.5664562266956422,0.734803079398689,0.5953196857656721,
 0.8656938229549617,0.4405209670048209,0.3127879450937179,0.19777785924637648,
 0.2754306727320259,0.8899887520729766,0.8701716553997254,0.2568462651543033,
 0.5683696734135476,0.9462495236373493,0.9631932794720178,0.2333589171574315,
 0.6149253540300242,0.2724185583940796,0.5309298989903438,0.15671821845242173,
 0.04902512559419969,0.86231674897953,0.18949478905759576,0.5534954525229522,
 0.20786080283315622,0.5859165209766739,0.20853824144205502,
 0.9480808992523357,0.7155581929151562,0.9304926612800642,0.41534624805436227,
 0.7772185664943886,0.9780696199741209,0.0680498135384302,0.2351085316726812,
 0.8138554172785792,0.030142698600720164,0.28152303625069786,
 0.3369332615465216,0.06615730913450679,0.6806595200827987,0.8053865212129959,
 0.5507083388753361,0.8903526015683164,0.9648240013174503,0.9447352450575738,
 0.18864676542860948,0.22527117773295524,0.2918536052926707,
 0.06390633513767241,0.4996684686741325,0.605949440935587,0.3021757849115022,
 0.5709545928919089,0.5725503845244,0.4173986246040804,0.6580912610774673,
 0.5559817764719139,0.021342002074037025,0.4888297041404812,
 0.8760083824703444,0.4193964585127634,0.8601698020318744,0.16305473056762987,
 0.2769784829235914,0.43441528267482343,0.33856969696356753,
 0.3090458181392747,0.6402941094872427,0.8140002249887911,0.05121031185205571,
 0.7325419076230926,0.3136001680939682,0.8049079533061384,0.4460729176840381,
 0.10602134775414285,0.4715002999736899,0.12559387431477909,
 0.8643076315591486,0.25902145180944924,0.775705537442918,0.38674657563025105,
 0.3462608567378388,0.22370392150294205,0.1846861753628718,0.5359074080186385,
 0.1713743031984083,0.9517114778663681,0.9178065570359554,0.7848019680484954,
 0.19610825879463545,0.8333351393492726,0.240139260653621,0.5090241136054395,
 0.5394636653220328,0.19399234306987523,0.6753807168128743,
 0.27541277372752704,0.7461681768113764,0.12175818729939658,
 0.8089455252956973,0.3819414278962705,0.7167295497478933,0.9060932231660422,
 0.22779172023769179,0.6032813427792392,0.37258384680422996,
 0.33407049924421095,0.9253341602202672,0.3256796764751756,0.6830504070982646,
 0.5490155244586634,0.8866819983146056,0.5712032358289424,0.8369431215794625,
 0.15755505726613805,0.6963429716112081,0.06331535691141488,0.660606889124748,
 0.9031905894929821,0.9503778859368958,0.421246940560257,0.05927374311982481,
 0.8644908035945112,0.867390910772084,0.8129582208036612,0.18205091887589153,
 0.11295534509228267,0.39157588729526416,0.6267885288497855,
 0.5315055133344011,0.038178067612051225,0.5547013658994612,
 0.07393361949987642,0.5407400183778691,0.3264629323258259,
 0.09796232187485843,0.08614989604664369,0.40977251390096725,0.63367915560058,
 0.8816873296278809,0.16269887020652063,0.04415295849019496,
 0.4931028492062405,0.9124559174338696,0.42586958852027945,0.8951104348633476,
 0.8364732159622631,0.3638342851325256,0.22315899299263153,0.38960376690764,
 0.9688734401746644,0.6224557764682938,0.9781754308474125,0.39483341644392356,
 6.732444309731633e-5,0.6570646961374889,0.3001051782522479,
 0.8049630591556237,0.93609866538643,0.8572625949661311,0.24373987054000146,
 0.09923265072247578,0.6550805182417875,0.784964626076402,0.6133992976787912,
 0.0212194011105723,0.38943521709123796,0.5828423643421963,
 0.34480354778543254,0.902033540063812,0.33009245010211663,0.4531828852596633,
 0.7470405794395651,0.7573020086159316,0.4075610002594414,0.5808214206976351,
 0.6995604815411736,0.7632222878950131,0.42313953907586854,0.4817005350478416,
 0.2700091506513309,0.8742825917204126,0.4623560765366316,0.5058039981033599,
 0.035851634286888645,0.11404662972245871,0.9981381224779302,
 0.5047913031615785,0.8108031256939153,0.5081543891286426,0.21803675880305795,
 0.5947483398592472,0.5756474899786356,0.8320808824975708,0.6693897325880496,
 0.26628048640114843,0.7042697818413315,0.508658106917128,0.43309118894952425,
 0.5751257301114486,0.21824243592669335,0.4522201504714072,0.3747765605362079,
 0.543977749839808,0.037613167612189136,0.39786227846315925,
 0.004103221728108397,0.8133832034677658,0.02562938057761477,
 0.17812092398171575,0.8292647829851322,0.2929197404734023,
 0.09093877328466227,0.43622217116565243,0.5186823442103926,
 0.2653006600318829,0.06472542312808682,0.28377132828771245,
 0.7841237369279055,0.5792275015841708,0.6850779472109587,0.6404027824343759,
 0.7623978015028392,0.16805704006821143,0.1471083190890753,
 0.11796435673424344,0.981969124848146,0.2681438793192854,0.9333044020113199,
 0.23756359543578576,0.4289385147222138,0.4430307141319997,0.9066430034638785,
 0.6869872451500146,0.7035088254786434,0.048741751632556674,
 0.4871906475885105,0.33315618807367564,0.3251527950288844,0.7932383102312756,
 0.5187458146322773,0.6717468760986316,0.36563313529351427,0.9605487136818474,
 0.7749120239021335,0.05775658857704835,0.41264023154252927,
 0.47919200987868393,0.02790990703590779,0.2571124466650241,
 0.9315347630405281,0.27267273173950934,0.9557404034662005,0.9304425018528852,
 0.028745028296722208,0.7725474078852743,0.20050975403175153,
 0.3608430226032343,0.08756463386507052,0.5176354273247658,0.9673910598821053,
 0.7977059435127919,0.29450251798720606,0.6179044433210203,0.7596380265046417,
 0.670537818790159,0.8147278639514254,0.059725023047647485,0.6915106652539648,
 0.18808459300642222,0.07562196304264701,0.5131646881914721,
 0.5908932097639673,0.3457649315754774,0.8993745646049556,0.327544726163866,
 0.4352273311763084,0.026934876867344304,0.3566979563410524,
 0.24932819696622388,0.22933705814920735,0.364149252111994,0.978022902129274,
 0.8356737719659313,0.5799319234489482,0.027222746410895726,
 0.6607569485683307,0.5215057957357333,0.37401875086584646,0.4830225941461198,
 0.9610679570063754,0.009206498690536469,0.2392503542001826,
 0.6318237814843755,0.958264145368903,0.6747724709453253,0.33292768437012965,
 0.34819731560706046,0.7224702263743982,0.4952178806441959,0.8116391076444831,
 0.2623694895281551,0.013284446988745868,0.5315027322471736,
 0.48154647729812283,0.4141979742982185,0.6251766009074463,
 0.30780785135436484,0.5953782371940473,0.42407926545272756,
 0.5335931053630908,0.3271958439943241,0.7819435521142744,0.18276016379954585,
 0.2198397632634549,0.2643490614658829,0.1024964799736906,0.5391108863511194,
 0.6799154868675459,0.16970777403446546,0.8427178952309902,0.8390383981589016,
 0.33000501643309665,0.24126455494548316,0.8525325601753178,
 0.26471695189872946,0.29562862825904324,0.5915304483953896,
 0.0844145259073168,0.02914396241285544,0.647919534068934,0.36478811180169923,
 0.5236451163803403,0.9865189827176488,0.103921340140396,0.7331115518498277,
 0.5924847059584673,0.34493082449278756,0.39211114672288516,
 0.6495283816621349,0.025129434509113757,0.7233817462736452,
 0.9664282590055239,0.27092367073474877,0.8493682532152955,0.8391830402493983,
 0.21344774281581924,0.9936999079494042,0.19833664760362568,
 0.9043750639048199,0.956124863765754,0.17660085421678662,0.10262795918936007,
 0.5301446550438207,0.6600824338180513,0.0015399019601001207,
 0.6838980239860264,0.4921518928778571,0.7915423718163026,0.44103695514585584,
 0.3902887584392849,0.2146040966888666,0.30971163662039725,
 0.046634795418953545,0.6198353822986151,0.28685041936968747,
 0.32590099608412515,0.6600757285929719,0.18540734175804374,
 0.7874740968847082,0.4034575907550222,0.41312509211759785,0.7330919925375585,
 0.12494587547970071,0.32540209212948235,0.2854300515712249,
 0.5412482633712652,0.6484008303152449,0.9520403359536065,0.9005887381642025,
 0.9580980636203167,0.64176170086443,0.1412968466682637,0.416926802239845,
 0.12395557654308098,0.4171109648823216,0.13645474526539625,
 0.47742958813609293,0.5656898680028453,0.937348964339282,0.9298085087106939,
 0.7159969255270411,0.6155240372329455,0.40582767675057674,0.5961408029568924,
 0.2640056068825779,0.7300966105016146,0.9495339984153137,
 0.047447488515265035,0.3764763449175207,0.9147279467031115,
 0.9423373424800104,0.14806347422027422,0.9290196811885796,
 0.44321819390016737,0.2684836672087938,0.6084677942307156,0.5530914629566613,
 0.6765466199172168,0.7830284342439189,0.5109054271171595,0.7065220746847997,
 0.22260059055647274,0.9600905090324531,0.8959498992833187,0.9027420328282987,
 0.8447019630921002,0.21618543777242427,0.7614655168799376,0.8042076167265388,
 0.8920425301731809,0.16051239391289918,0.4898924694391995,0.4211504720940977,
 0.1626217569304622,0.5709035941708515,0.471711154982392,0.669131193441464,
 0.587340888532495,0.983116583742256,0.007293424659469583,0.22642397811446813,
 0.26473442274490644,0.48119660788133956,0.30268495249488914,
 0.7064038633965872,0.5080262813203625,0.6990524473986572,0.49280088041486036,
 0.8326809250708371,0.10682739400139496,0.06112306328699202,
 0.5359965627635819,0.08434940445488259,0.6141096507926613,
 0.04926955656278087,0.3314982289507108,0.6224809116266098,0.1813102873649307,
 0.9445295257051742,0.2502479310224124,0.42514178007185766,0.9183321565088627,
 0.9669717531195238,0.20735310093902148,0.5630158401321531,0.7406319795856245,
 0.3199261684420136,0.16670335584957363,0.2671191256219644,0.6355109322855358,
 0.6406399866633453,0.29357232473234984,0.46277111800908066,
 0.3325681293290472,0.7203031257751401,0.5653332178932455,0.22337080063612358,
 0.44620374164787,0.6027753673717579,0.532794567101491,0.6835229074193949,
 0.49747545645812596,0.36303331504128267,0.49382270164053654,
 0.3201284394918167,0.9863774254197434,0.363219122428976,0.6383793925764749,
 0.13067442203044777,0.11137838277871082,0.17042164926658676,0.76973565314081,
 0.2559901591216148,0.8241680623254553,0.33905060274134524,
 0.13693635893294687,0.7326332451787728,0.4218173540385368,0.4208604513343337,
 0.8413424644677021,0.7986331659977283,0.32148773321016355,
 0.38262851447240176,0.8566201661298994,0.9438325425320518,0.8547990096913665,
 0.8767316035138967,0.32663070555399754,0.5163266133240324,0.1735588896511867,
 4.4535943069412554e-5,0.3316492731506969,0.35786238867725917,
 0.67108055883324,0.2904560427608526,0.5281887115739382,0.7464201166419059,
 0.08667590423167293,0.14759394135481263,0.6024802843612453,
 0.6407658549358195,0.8581962707969903,0.9319519025096749,0.6981251920347317,
 0.726636588213527,0.4354145969122297,0.9908399305136717,0.6986230253549858,
 0.4243971586979173,0.7523568279380193,0.15610636434843195,0.9986396974676597,
 0.7803592351787105,0.32078529742281714,0.9413682581282444,0.9174644806156861,
 0.04954676528980306,0.727003472916903,0.033939681818161094,0.519669301903423,
 0.23623935136766616,0.5677862693863611,0.901693177145771,0.444730816103895,
 0.7302237916086618,0.10911055366051525,0.4873230890984763,0.8577030631944891,
 0.06810431915840576,0.8218387133573841,0.8908471603514753,0.802811465378718,
 0.3216610039799076,0.21141262755801704,0.794844368711394,0.7635028122838652,
 0.9711206568668361,0.8453568528136856,0.19798809311472987,0.4367944235258905,
 0.41654509295844766,0.22234257214879571,0.20157996662911737,
 0.41839665140386684,0.09414269232733374,0.5766527531193326,
 0.23326740480890717,0.9137141419658003,0.6492928757764068,0.707825377715593,
 0.6988627757934048,0.3466896397092035,0.3733662510229634,
 0.017989805574081363,0.11269130450624898,0.9710617024592985,
 0.014101196271910643,0.15901274355535588,0.2807686523717572,
 0.9732995024976626,0.697225758499593,0.43039887073783323,0.12022712176515193,
 0.9616092672977268,0.8325289358445984,0.7427465423099067,0.6011696186135437,
 0.16712253981489178,0.6412351459991101,0.20818232786516133,
 3.853128224449165e-4,0.32224976582815545,0.8598765203926321,
 0.5159059889464446,0.046315870590328156,0.019517220039930505,
 0.8729140564093476,0.6384839737672205,0.3388997998502653,0.8557170530582874,
 0.06338710372786316,0.8528490703787068,0.14615234495114005,
 0.31308567009601895,9.978348046308927e-4,0.10302148701188141,
 0.0408334514496691,0.6531667845963961,0.9151172441817396,0.17463919420508933,
 0.5497103049698492,0.010284686193697379,0.04402988719335876,
 0.14844169217792647,0.07221048834789101,0.3454896975269057,
 0.09609828041650126,0.5331363452903611,0.1423599416580026,0.7124400174122159,
 0.49184567529728385,0.43246842961770726,0.2941782184614864,
 0.13130888278056219,0.737967219998365,0.9186825123589712,0.38170193364695315,
 0.20893722696038708,0.9774021806346759,0.6143627548975499,
 0.10382119703315373,0.9305204722999469,0.4271773852291152,0.9400276749977092,
 0.5888110229496368,0.1634903619874547,0.4256607787838931,0.5031084163470725,
 0.3626512105233991,0.6312907664774495,0.7225586188841245,0.32346350172317284,
 0.9737491662703154,0.6603734290174094,0.8432266111601137,0.33582683791494206,
 0.9536174304232958,0.8655349876714179,0.7206353461581633,0.3852791304062889,
 0.8945680714596146,0.4582663084709043,0.5372659341029442,0.9567947622063537,
 0.19001620292775478,0.21797045871049836,0.7850234699038344,
 0.7749170273313717,0.7098083257595011,0.6784698174711952,0.21396886322056385,
 0.18275103869395082,0.29619969437846616,0.19183005435601175,
 0.9207296457141244,0.2517829563998827,0.054235174260039876,0.881229923280794,
 0.2676780330868892,0.2528902066645007,0.25789732507880614,0.6623433263866562,
 0.3485917909890228,0.9598284585686562,0.6488150352091857,0.1493289043355891,
 0.4098220045737494,0.4411266108277614,0.8310579449865,0.9310579242511318,
 0.49512109379825664,0.2712040883952227,0.44365175129385304,
 0.2789575213637856,0.6179544012985114,0.4009585521633554,0.8178613514173645,
 0.5829792550272431,0.9104376161985174,0.37715253401630433,0.6280670125274399,
 0.029817221400584404,0.10838663406693094,0.05834259882971815,
 0.2999082668587899,0.44522367781338945,0.6624400584127875,0.5296322407633409,
 0.6455139259461363,0.6184933117862477,0.48537573806563017,0.5860475173189253,
 0.1509036938416496,0.3050885395841413,0.1762754118137947,0.4953121453417759,
 0.5049271556642524,0.5994625456323719,0.415763469645518,0.6941715359604171,
 0.574440244876222,0.15390494748457628,0.4885713275132455,0.6175674461236182,
 0.6207524027438395,0.02572129448709959,0.7255812266058769,
 0.009741672804154833,0.45768481485080614,0.16404494211629728,
 0.6250327619382211,0.9168283124503198,0.1430810100155413,0.43822686618720263,
 0.39460039965434146,0.4395646873663773,0.8477198844781386,0.6553464147582199,
 0.26385067360824643,0.31706097819853074,0.5580255383072134,
 0.6747508840327383,0.43361028562915704,0.3664443709426852,
 0.12462025813119626,0.8344829680345209,0.8185871373099577,
 0.23379522346416437,0.9564774744804788,0.7716141439687109,0.8386934778983095,
 0.019940192399389267,0.2645142667994257,0.31433386912175765,
 0.6276692391629637,0.6740004486626917,0.5315731073357455,0.8150907006926467,
 0.3406536554240003,0.22954040667189157,0.05652062050835582,
 0.6575237171616186,0.8150878346076763,0.6591605518359086,0.23555609909944752,
 0.1444409301440719,0.8497454061577124,0.859689542844969,0.3848827174888617,
 0.7253951674392487,0.9478825453029271,0.6172601262104553,0.3540189472616032,
 0.06905943286393823,0.6061965123867208,0.2143721118814973,0.3746259348314499,
 0.6658750715597599,0.6777853053878953,0.07396254592364027,
 0.35023273457318127,0.5774407995920303,0.8212009308769246,0.908290094714026,
 0.7059288236421892,0.36291966794089714,0.3306491524130417,0.6632733000134688,
 0.6902798855633482,0.2445962073289628,0.5466585889141612,0.9557411184658466,
 0.8317641821195392,0.6542928750260036,0.4973464175319161,0.09460377211375037,
 0.985678199310721,0.5598794431364172,0.23116030725302328,0.5768822930147592,
 0.7867651729679949,0.1118727096143517,0.04075393431199237,0.6010368736336458,
 0.8523661241001523,0.9929483872119516,0.7152880792334393,0.12758507349536796,
 0.28003062258546596,0.44948674481023776,0.1821029066889669,
 0.40108660974924515,0.18763541287357932,9.083167027931216e-4,
 0.8562747456300971,0.6382452502069142,0.5565496514062204,0.35214996394592724,
 0.3119484858898597,0.288407094959227,0.7105785550185862,0.07968311085804702,
 0.6584121442625298,0.4500227109324728,0.03926458195023064,0.8750719908997746,
 0.20388767531026109,0.11110462725075854,0.3362142640799958,
 0.5241598735985554,0.9506763861186008,0.5651980218111265,0.2555925536974917,
 0.4114825441916792,0.08927863079962517,0.6582444796572775,0.88868568192908,
 0.36552399573198235,0.29531989724612984,0.7597404449803662,
 0.40591438901522636,0.11953858254356975,0.03014078824558919,
 0.9846318419519824,0.625891943481958,0.1356026154223634,0.7665206712184424,
 0.11422604098854128,0.11043456414379571,0.5905062574144061,
 0.4900293967404147,0.7340274145656323,0.5597547835543616,0.12977723459992718,
 0.8828572171319667,0.4944532512943528,0.557948942177548,0.4362674565439292,
 0.7932209729326263,0.007697859026178477,0.6241081944018592,
 0.44128094106606264,0.6178244901291647,0.17623693958434916,
 0.40209835906967284,0.6818917921777594,0.6305048844331743,0.4510733574881238,
 0.12830291153119777,0.9702347939429696,0.07477983898428775,
 0.4911391109760178,0.2256763099127318,0.40831026163580997,0.5957454021624027,
 0.9675058340777251,0.9352130253859653,0.9927082279666427,0.9988442680070739,
 0.9864670756796131,0.9130959329675914,0.8552459135310604,0.22338680007756506,
 0.5041681850309075,0.011908535800561615,0.3452147282725705,
 0.39105344833682953,0.2428697384726245,0.2390144122666381,
 0.35416968415269745,0.8130277070656182,0.6387227192543697,
 0.07966656543698103,0.6570523245679785,0.6848837561234417,0.6125164734889373,
 0.4453237800380948,0.7750716424119776,0.24850556413751557,0.7388679270465562,
 0.07415415180509244,0.4369977989273799,0.7269052195775487,0.8888584527794249,
 0.8949196253071475,0.8811538600306622,0.5140177223111094,0.7441740523124485,
 0.31463469087981144,0.3320652768448594,0.6623283494890528,0.8739861006215063,
 0.840430754760345,0.4383489163337022,0.8798438959611269,0.4438761160751219,
 0.48056925812292695,0.17693996528578293,0.5881489279786608,
 0.8937349748863334,0.5094685605100562,0.3756173698708861,0.9984545038122317,
 0.26341605033931703,0.13725743083240516,0.9750226298197061,
 0.23063522358681066,0.6292049681394883,0.9037859019755381,0.8463043005495,
 0.7332101459188242,0.9091212075626269,0.8874121953645198,0.33840517355781785,
 0.5222651896668556,0.17659772351870828,0.2107264233150551,
 0.20160698782467812,0.5409815545886799,0.399538122208453,0.34802844094015883,
 0.9403077071439103,0.34630378644654036,0.23649814256935786,
 0.3096653181461627,0.9719456090239449,0.36710976799524675,0.6131482675866811,
 0.11382958879393579,0.43251369293075037,0.08766783168007841,
 0.39751864167592765,0.15301649626099045,0.2899643226686436,
 0.13983348687140895,0.6912931122076001,0.3623071859792505,0.8593422249376967,
 0.44018802516002575,0.2385734381055653,0.6945973351732313,
 0.05288722361087861,0.6344003761565142,0.21513854204807292,
 0.2789388026451223,0.9106073076753389,0.23000724989215204,
 0.05585766303778972,0.3499035788152206,0.7284265896432369,
 0.17377431244133001,0.7050852554074288,0.8109009456809231,0.557782218532044,
 0.5134452589626758,0.28506574078209346,0.3006939817309232,0.2369216829766212,
 0.7000435591492374,0.7986057190184994,0.45143802522834975,0.9992729964040516,
 0.9855438839971499,0.3675489661742396,0.5787289159426572,0.6471385380783562,
 0.6419158090917578,0.09387980157987008,0.9249797082522228,0.6542474532886078,
 0.14731889814346655,0.25445217314971647,0.5163784083616363,
 0.5322786992112608,0.07918470332674821,0.8782816659320949,0.3222301297489255,
 0.5576626117215093,0.5754291419229608,0.7632721004531362,0.6988265745713403,
 0.21510633393621437,0.15857013486408522,0.750166030566635,0.8359899995231126,
 0.8047990294251665,0.3823303710540319,0.09121577015726412,0.6277772795646228,
 0.6639673955227685,0.2758998324199273,0.02043538035726722,0.4161440348069636,
 0.34766378785236585,0.7383604274891598,0.8964189765783774,0.9900918327252988,
 0.6849014400135084,0.28475069552161003,0.8423969718478558,0.5343513528931854,
 0.6204998721102011,0.7427393671071064,0.8758557662464274,0.12764354859314864,
 0.4248889325922962,0.2820881452302805,0.4524221783524254,0.42406413348233707,
 0.7274848288697132,0.4438569406163769,0.6674653546871726,0.9366893121991829,
 0.9911796140845286,0.11535497354885538,0.09114079572858813,
 0.6514577941217009,0.004184406823410258,0.8477580160782124,
 0.2753062466561472,0.12496079152961581,0.4911310112774252,0.480275144288848,
 0.03814986636542972,0.7226247058079586,0.5338653364361755,
 0.27151307765987476,0.4573547257471625,0.7520991186353391,0.7306007772066997,
 0.9261559034484472,0.019653486506641893,0.9655640403291073,0.819329244533577,
 0.7825281214322847,0.9424499803139126,0.8170632447098415,0.9489528047874977,
 0.3435674997568634,0.1986537211443351,0.43698251521978915,0.2358251196233151,
 0.9488142550758416,0.5751752616013406,0.7216912196864749,0.7941710731122764,
 0.17657641031733096,0.20859328892729923,0.8748535127891248,
 0.9914184260484378,0.43704906596755266,0.7246845149814861,
 0.35211737027116097,0.807785705544263,0.6813468608195452,0.15894204005673207,
 0.9609419058017311,0.5381945930908537,0.04953146661473018,
 0.46039054304437865,0.1738634066656347,0.8474345657904812,0.4170426258731157,
 0.12031775180281257,0.725004244998642,0.36351827155597227,0.8083912078945416,
 0.6925710987025383,0.9454405359913487,0.6349415939124406,0.22738698481715414,
 0.8608084258061227,0.4277096085536545,0.9691252730946034,0.8318825410350628,
 0.8744217287016338,0.6283554874236711,0.8820556022770675,0.9316777969894756,
 0.09341041265038397,0.73858337490874,0.9507071328290744,0.19669997416168694,
 0.033873961301710054,0.6376494051601448,0.3449153159076048,
 0.9541056559448835,0.8604052116771275,0.3345404506660099,0.10875979228760002,
 0.686044650722626,0.017953544346391737,0.6307296123917501,
 0.005168303012279218,0.06797813404526887,0.39086179890693096,
 0.9156928399240761,0.6496211242735186,0.3895710255760143,0.2044241696206699,
 0.6375737503055631,0.08476241101727977,0.5946651426538155,
 0.49654486782293683,0.5064190564688816,0.16640577428232328,
 0.6945912697365546,0.291888886077147,0.3054351433372722,0.9040779894022501,
 0.31657780306904637,0.7861214311560182,0.46255097019688607,
 0.42751408959894555,0.9006761231348418,0.4735160344046677,0.9875074275515274,
 0.16476182635315806,0.8542339216677367,0.19124458228998953,
 0.35557214309067464,0.13053208548365847,0.8065594834399445,
 0.055678565486145914,0.5480433832452225,0.9847546095716646,
 0.7623613016458348,0.9859935763642315,0.45945686992430357,0.4170163060662204,
 0.9299716345123467,0.2603585427969999,0.04277205339660939,
 0.25352011137628017,0.8149234111676305,0.009997541468648952,
 0.5073203708002365,0.8546382801162441,0.6288834212123637,0.18924542379069387,
 0.7672766751539618,0.6854898452368797,0.018555571282568017,
 0.8272366083248018,0.48907783016831363,0.8873326236917322,0.9927417627766129,
 0.11792103338840398,0.8524060527298569,0.8893339717522402,
 0.014399928994503375,0.040160110214732314,0.25818251673326564,
 0.3925272596062229,0.6605195742484338,0.49802966919657965,0.9130837985280845,
 0.14014553779029826,0.5079410684771736,0.7536536347034123,0.4559472162322975,
 0.024306778548346508,0.5099535713879192,0.7872648814631819,
 0.04422519057704999,0.3601613909536222,0.31756872815872006,
 0.5769222128136722,0.22505971040081074,0.7425482182142777,
 0.39649810697246235,0.7665913648555996,0.696651030315454,0.45480745387503385,
 0.4997435850074826,0.36554123912851155,0.2528249123067686,0.509586400893924,
 0.37605912122592333,0.602021210904425,0.8754403344437358,0.3888515605995515,
 0.4209235235555866,0.7775815874017696,0.46442500267399756,0.6813221993246308,
 0.16084549225394107,0.9272415883707339,0.7387422866871508,0.657968285892124,
 0.19495307417895402,0.1841246010845301,0.8269043531285536,0.8920519066462559,
 0.05977048036539068,0.7773278233471731,0.7590637039506763,0.3035965902481963,
 0.17681787280988415,0.1440994673180721,0.19486041997431247,
 0.3291939860434776,0.9451024751500917,0.9099581579346586,0.6247095078306175,
 0.16300145150588063,0.14628464043997536,0.40974448843849054,
 0.6690322522754119,0.24492427015344376,0.1893624246875465,0.117521406478729,
 0.8248311026351149,0.9290322092208555,0.6600758247285688,0.8296180839212679,
 0.05879111821019878,0.642800539553947,0.43795036893857686,0.4829808537401492,
 0.07918450030902924,0.9829228661231442,0.19073537241285687,
 0.9698174428481982,0.723663444204005,0.40974242450556453,0.23485146939492407,
 0.655305359132299,0.6162818106336878,0.13878205706510516,0.8995338268489594,
 0.7053220162403872,0.048571345751449835,0.543868275353483,0.8814151681976227,
 0.07143201243315922,0.6891107582338412,0.3800453454445132,
 0.08073526542679277,0.24293938949887584,0.0880858471326269,
 0.7920494891801073,0.7634823775321804,0.6196041538055347,0.9374863538297921,
 0.7152389952895756,0.9907201494589422,0.8274028457766696,0.13865944350375004,
 0.36872798426445197,0.6744628491665084,0.3362891632727296,0.6734259770868642,
 0.7885804836413224,0.008679510834685877,0.043309671557024476,
 0.36863942978340036,0.7406621608658799,0.8332381836716505,0.8225297146715921,
 0.1334738483221518,0.39106507865016826,0.8636207057360316,
 0.48103113821186905,0.751766947127398,0.22643723239871116,
 0.18471160571621015,0.3572628480625236,0.04853627991129239,
 0.27631227085004406,0.765667986057921,0.6295079930774882,0.9206016818950467,
 0.22584212139203663,0.6954200049352787,0.06643967352555213,0.88262484752811,
 0.7831711499260755,0.5587619338949222,0.22757797090452925,0.8405058116314164,
 0.6192634902506963,0.33675477328331205,0.038545076796137834,
 0.7314518856018312,0.39348767836174536,0.5002680739157064,0.7210257527739831,
 0.848261217233643,0.20830328230428463,0.5364615620616127,0.6968394640190652,
 0.12510457634294325,0.08621087600510946,0.2799896613369084,
 0.5845097212035298,0.8957562133855231,0.3214513572447957,0.8331887047999232,
 0.5573663767013901,0.7983213594519315,0.38172889750846906,0.9167133372692404,
 0.5194289529815125,0.58337474331295,0.5855014306285755,0.6453308869447703,
 0.8242479582561224,0.07382527081784795,0.29476505203807,0.761873004357886,
 0.13125500868718987,0.025672741499337848,0.9790008682263145,
 0.1209740325757297,0.08799964106567781,0.34536790775696047,
 0.6078299129566339,0.007169528633304889,0.9610737290174818,
 0.5304684762896039,0.33377118583936716,0.08424557022875567,
 0.4096633246299626,0.5334966323183814,0.1450849112424787,0.6749689200674268,
 0.24624649728800685,0.0718898034871488,0.6821628008697459,0.9814851686724467,
 0.7106142783234395,0.5144024502313899,0.7563988689598804,0.7668972831150892,
 0.5798335255138422,0.9891970976408855,0.6841015048832227,0.9710417448139839,
 0.5409471712742335,0.07051834075296015,0.27736412682575007,
 0.49787946908863456,0.8741241013824886,0.9126929992675425,0.3627007956107142,
 0.5445757554303031,0.8134478941916271,0.9814612169768514,0.33900505242407053,
 0.5345976831899231,0.2195398103632228,0.5807108135145065,0.03258728182605719,
 0.7242230814517538,0.33847968907560455,0.9677881877056009,0.87699267654751,
 0.2342438186539788,0.8753697135105776,0.6414951474282276,0.08710078241638453,
 0.907806913608323,0.2900803889822091,0.4047635553234107,0.08627397533589809,
 0.5257236335402985,0.7762495561288429,0.7493476990183423,0.7347057629976227,
 0.9575304666075867,0.9635307440448443,0.5789177785925649,0.49609202889401216,
 0.40213973321736374,0.1317131951374022,0.609451060483728,
 0.056051357756942255,0.7635167058428416,0.26764903305194343,
 0.08620640677034996,0.9455751987808874,0.40804647877919376,
 0.43973207445015783,0.8205401544296554,0.3186411992373973,0.8101016176927408,
 0.694246025441913,0.5721294572558755,0.10047139259560289,
 0.022868247454564394,0.8244314623930247,0.6021398493801714,
 0.006441078798124056,0.6205249508452407,0.3431420622463258,
 0.05966131827380683,0.49497225513172316,0.8799849682130021,
 0.6069342347172798,0.05394625153666632,0.36927649849944233,
 0.3895357096408807,0.9800265225013023,0.3988571094222588,0.8945469645408199,
 0.41624351153882155,0.659230907387552,0.6905130380654414,0.669778654755744,
 0.7657066108671757,0.06002554775094593,0.283162009682132,0.48994615613165116,
 0.9629713757887159,0.16205885798308017,0.3655425510208462,0.1387034752736953,
 0.4280983804381221,0.3575817059029598,0.33077443773717263,0.923072402048091,
 0.9457818117649804,0.014815338667588662,0.052914456940533316,
 0.6052682794144317,0.17973193619490213,0.9106677370719206,0.7526142182524138,
 0.5082401097374951,0.3325296122609088,0.40519120813521114,0.8542728682175119,
 0.5203244811847691,0.636815547632943,0.29816446099597216,0.4148062615781537,
 0.1556290313876888,0.28298279731529197,0.5561049808724546,0.6755229258263638,
 0.9781719067541864,0.9348752576057054,0.583517946734366,0.2110897236770365,
 0.8243664545881282,0.9253957415957376,0.08737247121885239,0.1284565862460887,
 0.505760649292317,0.30837392527857754,0.5861851622853816,0.400008201377026,
 0.46696410692184775,0.4917617045906095,0.3503065470976573,0.7248791242759356,
 0.5624396547043247,0.28962415174591327,0.2913573271793941,0.924251912339581,
 0.4483785348697351,0.3642409719371259,0.5266442649443298,0.8477770859220425,
 0.5606698734824247,0.6134961963228368,0.5708060596216444,
 0.035952135421912446,0.06958162843143356,0.24129562470097798,
 0.003469703368034871,0.8345929331381854,0.16069197577291994,
 0.6456424906449364,0.5581938722332747,0.7041099647121655,0.8879395763719983,
 0.9598525117623607,0.36580762319562365,0.7994407283011907,0.8359612623577704,
 0.39155508083907553,0.036549989112477976,0.8358868858339981,
 0.7643084324333145,0.3996500261850323,0.20076232460688348,0.3143336595196695,
 0.9563463062002683,0.8617568083463873,0.2747908673651164,0.45093132454663465,
 0.21351723623174168,0.20286100843378663,0.8609025910894039,
 0.24457027635635298,0.46287357653179495,0.5961711384000058,
 0.1430671393311047,0.6206607440933576,0.7151007228074233,0.48767013081328325,
 0.794176628896166,0.8286314417908942,0.5037159181680696,0.27170079334573316,
 0.20604553640479994,0.7566132543708501,0.9373534709021911,0.6440150459932403,
 0.3063892138086033,0.9212854897766564,0.6537266791908756,
 0.017069381051131427,0.06297303300909385,0.6736702575429849,0.07770704354239,
 0.6534855316572052,0.8791588672971551,0.4082909770407823,0.4216921218161107,
 0.45047278775116795,0.2082302358959165,0.6569071261264592,0.1296512646954081,
 0.08037314151770403,0.8331068074844138,0.760125724303499,0.4379924248830416,
 0.36691485013216396,0.05024437382449343,0.2769718501890823,
 0.7697976435600512,0.31315783408731623,0.1977594772060438,
 0.01284649395983184,0.20456097125287143,0.958193180728857,0.0937601819804712,
 0.36326236343544616,0.06551151264848176,0.09961270716962456,
 0.19497547652120517,0.720102961001582,0.8220608752580726,0.45490525394883097,
 0.1830528244280516,0.5098595626841201,0.32298897348491973,
 0.45635798276761996,0.2386231597825147,0.18233423589833953,
 0.4533067907449251,0.6515475073778305,0.4331131863403954,0.3087342176729738,
 0.8235110600114501,0.26361277691069607,0.08970322204334003,
 0.8920928609977702,0.6003102160171186,0.7888834685238165,0.17044289026276505,
 0.19897245254873497,0.22140356222642155,0.6892618468085815,
 0.6932753770310834,0.6633691241322284,0.6371508624302669,0.8780578613020238,
 0.34669224572752055,0.5846514036130641,0.05921371243716189,
 0.4825127178645845,0.1135528502261628,0.5729662027022423,0.23397231671305152,
 0.2738817921880916,0.45561892843080976,0.9091457653200293,
 0.32011244466344935,0.8088634105264311,0.34406161780324807,
 0.5011521001756324,0.23520436897138208,0.33990754053347894,
 0.11961710111050239,0.6027243309873462,0.09335100941555308,
 0.4859938142300091,0.2602683521484156,0.864023404215276,0.7836932859733412,
 0.40403529839317176,0.4080094611688819,0.5018940480366266,0.6953184293209776,
 0.019219083417265304,0.35214840118314206,0.23183202578886497,
 0.4164575605954284,0.9431613269096489,0.40099325598742364,0.6525002097319529,
 0.866959555418505,0.9205458388886447,0.8389607042970945,0.43361020016558527,
 0.4993792268267131,0.9059542732451275,0.6430520378230014,0.7927114756110847,
 0.8493094958321138,0.42226929181803063,0.6833254529232804,
 0.06314005121880384,0.34957431449666987,0.8025878300015378,
 0.9352969056056079,0.7157792849491397,0.6327876515665078,0.8572272937061687,
 0.5179110696087044,0.8862144784118785,0.5335543125368758,0.6812126417212506,
 0.7524002742915938,0.5896290164236767,0.4346191145332672,0.6299115293818968,
 0.661873177073719,0.1109515635277657,0.9044536112028695,0.8877242922096082,
 0.42500603257398106,0.20893228994599777,0.12217234590999271,
 0.627683395782983,0.19105107689561562,0.0195742199076262,0.7758934110800735,
 0.9550619356051605,0.7043049535816944,0.5137111091398505,
 0.051066690284570115,0.8255110291026123,0.9680918867861511,0.276574404441992,
 0.6554440693242037,0.8352525436558234,0.8645676985788108,0.07120987589906436,
 0.4132006959161494,0.9327094977218622,0.15186911312330187,0.6751589721337479,
 0.5729822618430863,0.7967357651472905,0.16376048067639593,0.9284152653353875,
 0.39294982736221773,0.037734672250599455,0.6147347798500321,
 0.8359065035660268,0.8257915069007665,0.16336557227396908,0.57793339641276,
 0.26580965071228435,0.8609914057250692,0.3501147554119174,0.2331480479906184,
 0.7176895777299808,0.73200435415968,0.4687861473305439,0.8053930554682863,
 0.5741811011412764,0.627921363073031,0.5976606249320889,0.975679222748871,
 0.6319621863028007,0.19716776952392567,0.09493796694194823,
 0.2079206161908531,0.8925878092467033,0.7185027037968865,0.39639655275009433,
 0.11498433793223062,0.3344455116915126,0.5107054134582097,0.9981924179207405,
 0.9462394424226905,0.8968673072211268,0.3150293001038821,0.5201284819757303,
 0.9309028612774469,0.9536567341285169,0.45337361746087834,0.6754374297550476,
 0.43430416015942397,0.5261246738207782,0.8446443950852333,
 0.34973225215474457,0.2241789079725971,0.8317366205813449,
 0.10620895798145202,0.8627510012856854,0.26494067636511387,
 0.007887967828043463,0.2553994494675199,0.7831659759665278,
 0.020507826315588562,0.12471001324027564,0.0423539609675847,
 0.6106874478757159,0.5076119272331724,0.23925593010980917,0.9632289470765343,
 0.7464272440592619,0.2768473999983696,0.034050154260397036,
 0.4637081970474506,0.5238152425637833,0.2649529466885888,0.37465584934508145,
 0.006774482148936167,0.7069934217588241,0.06977511695720118,
 0.7996197684711398,0.6279557174973797,0.0785987593331623,0.6479099555476623,
 0.8570300987165193,0.016162260929973526,0.858901256408095,0.6004685452216497,
 0.5728790058375072,0.8344816683280094,0.6150101717258839,0.6892821815232161,
 0.12415200754418154,0.20030235135617402,0.1356327197043613,
 0.2577323238466177,0.3199725000808905,0.7842597275714229,0.7253462486633182,
 0.6334764129550252,0.7428532731856432,0.46450967015901723,
 0.08331957801616119,0.07181222116109187,0.7928759817108897,
 0.44465748864873955,0.49584307294738195,0.3652561477840197,
 0.15548353694999673,0.4441858388005163,0.686184573043592,0.0881730891034117,
 0.399089441983158,0.03934475666511961,0.6111528001100041,0.5680062675343884,
 0.18241548588603262,0.42531042862509316,0.9160807128326336,
 0.7188899322869057,0.30609496584588014,0.43705490948680115,
 0.7031219344359725,0.8556382556012961,0.18566111512529204,0.9843830216174517,
 0.384016948026208,0.9353349681274903,0.7390171510743213,0.7080278823645759,
 0.03338584374942977,0.4839104951374201,0.6423274987885118,0.7860602907259778,
 0.46489173748735824,0.8286215794579475,0.5533725981288061,0.9181532361502688,
 0.33025524879737245,0.7284702341087846,0.8006555693786255,
 0.19597547915963598,0.07302727395207409,0.8368130839228032,
 0.001970817762486754,0.07689580206038671,0.3881115794743234,
 0.40239521835781245,0.8490389674589243,0.5476827376281234,0.1155665816246505,
 0.9617094813671493,0.05558502756072914,0.0030014080035472013,
 0.07087343838161875,0.999792914308044,0.014148225796078329,
 0.4611132111652837,0.1815563033643962,0.41885269408360704,
 0.04509116486426756,0.4774828366114885,0.4892382353790299,
 0.06147357103417772,0.17200838996079326,0.1287093932511717,
 0.9129377841065347,0.38984453204035774,0.9598074916049142,0.3817677734079947,
 0.7805600755165907,0.08982089537601745,0.9110972959576038,0.8919986090571717,
 0.40049431701054794,0.8250010461063774,0.5514138268148165,0.7199942793099341,
 0.3411576511958567,0.5930083051298525,0.8151887395982221,0.6455612005822167,
 0.5352548659981741,0.029017875601774157,0.6834782859421534,
 0.8287283755988633,0.09323245881299802,0.07201622527427498,
 0.5493533254561853,0.48439908237325857,0.6140278180047892,0.8889235844435199,
 0.4437529945443396,0.4144766056648028,0.1768447569799585,0.41295116674120536,
 0.1825038466486486,0.7358820386223119,0.22071132715088748,0.9913628957672331,
 0.6138299156912836,0.2164636483278054,0.06286057158835578,0.902966521429175,
 0.331760814993878,0.8207841722450026,0.3406185242458708,0.5638779283055769,
 0.27140242688801897,0.748342532537261,0.6650829413027572,0.927795720814944,
 0.016827833741418496,0.9033997737534598,0.6244258224709122,
 0.9106979945042792,0.8768549208194539,0.06885147289401561,0.8375307580050935,
 0.9046454636198474,0.8428246086231147,0.7078250719921853,0.45126302464782775,
 0.24736383761801706,0.1150361146883514,0.050758608342595934,
 0.5867295110187334,0.6150984363001354,0.7378089779104301,0.7051882225036433,
 0.5004651145491418,0.9782641831245847,0.16699202160394466,
 0.34144616623493385,0.0369390549799532,0.671587780160003,0.5090664092107791,
 0.4965436063837555,0.03827872368017993,0.061135185099819855,
 0.02122288654723603,0.016038943794308258,0.22866097579055344,
 0.6547311137031736,0.5225084760065767,0.6911987727028088,0.7579697768795076,
 0.08255598348591986,0.8818368352645324,0.06888138132048516,
 0.4371105001502976,0.35544748174820184,0.5537130385518658,0.4124467866659002,
 0.9690244039933341,0.005009038079144101,0.09220112150625204,
 0.262582395920106,0.6713163437856992,0.39615063163740094,0.6276411855041986,
 0.44952216254056276,0.05359867766642157,0.7621464017964994,
 0.23863678783528663,0.0781414052806586,0.2662769232185286,0.9241708215440751,
 0.9387902406834375,0.9196011274898359,0.26899328457093774,
 0.022086712945402875,0.8666680122236387,0.6208768584230417,
 0.030938828642208982,0.045500146411142306,0.4002391567795268,
 0.5227273374206427,0.7940972825255135,0.7147610474673934,0.8239759805559942,
 0.21380802365986062,0.006999576938795715,0.2839496285049895,
 0.27690471812314843,0.14685184370214666,0.7320366916482718,
 0.06946607393261395,0.8331218313803621,0.6433031388864452,0.1920279269659949,
 0.04958498288748281,0.43467128175781555,0.5677286063399373,
 0.38321065927178677,0.9635423674948906,0.9577626746535864,0.7787590145269782,
 0.7195795180794949,0.0373521167864862,0.16307314137519258,
 0.14081712366402865,0.017907538553501245,0.5143579687153422,
 0.08708212801202642,0.3400955805762934,0.4074517625532186,0.404446219383251,
 0.7739518569417148,0.536230406134816,0.5842815919566493,0.15848080147608323,
 0.43129716278286256,0.3414496885724144,0.7584252628935276,
 0.23134991605262423,0.6473265548111519,0.08110086238692538,
 0.6072487406714207,0.9668445509484761,0.037339594358959505,
 0.06423017775283568,0.2028746547602106,0.8392938105673005,
 0.05431256759491876,0.8243381508655108,0.48737627253536187,
 0.8991765828906578,0.5364470718241698,0.8946591695187446,0.08608068391207668,
 0.7672396780520627,0.07552319946357033,0.2583622813267663,
 0.06430178224327399,0.648191446245855,0.7950570763657283,0.15884730488598808,
 0.4063635010619364,0.8466024653364855,0.6069837056924017,0.3965611243058804,
 0.7343960588646584,0.4896351043934569,0.518915799767127,0.5899218298582731,
 0.9064017657550874,0.6804067227305514,0.8520519195639806,0.40916157504537654,
 0.961551950650359,0.6075519282221749,0.71378244731403,0.21914870242339657,
 0.7127434326592212,0.9317064299714207,0.7307008509338239,0.4099014546421994,
 0.6676920407469372,0.5179049784618722,0.7993866162188724,0.09604776812918525,
 0.3384282713733273,0.6877206572468038,0.563064766485436,0.5902809456502903,
 0.5720177840645738,0.2528020560514479,0.9508306892036256,0.82072426693985,
 0.29188815678971836,0.5266056486980035,0.8541625524582248,0.7277074316481756,
 0.6038942708929259,0.03166790024549315,0.9324508264300302,0.6231106788960324,
 0.08670295146981966,0.8989665361820245,0.17276933120427307,
 0.3614100410294465,0.6386983741668053,0.8375038757369639,0.8050687792576688,
 0.0759167567859057,0.77579897891336,0.42756708299900814,0.69704242196128,
 0.7112375823368681,0.9440626192212407,0.0694356248950887,0.7900572222321881,
 0.23710582717836226,0.28558649058565955,0.6489874856991081,
 0.7741124294136736,0.7192952684961557,0.5431721205774093,0.10304416138984207,
 0.3751249260467686,0.9343753239750096,0.5398704175934445,0.3038013581226473,
 0.27298569206400525,0.7572005509117914,0.49028881139960023,
 0.03428095254057706,0.8134210582810062,0.9217484257090152,0.9586463984412797,
 0.5039246503915706,0.30309912661288185,0.16105170662691637,
 0.6029253494255891,0.2629504393116836,0.5120967488735888,0.5255495811551889,
 0.8883969227497356,0.5149553509898516,0.028597398910435445,
 0.3291882349994646,0.27811317160787474,0.7412736579664959,
 0.12278901309043311,0.08262997365985847,0.7446101667592265,
 0.1568648841019098,0.08026228006137703,0.9033442670532517,
 0.49420920123278234,0.7896649928866564,0.9629235601358785,0.9088546449398431,
 0.2976653923602868,0.1372859695504669,0.1753489327497948,0.7308899253532096,
 0.4008249713190942,0.20487075474785088,0.7037092641241706,0.3132692815727155,
 0.23892731759762498,0.3674967773669511,0.444770360924273,0.3331487402825788,
 0.15824192989538544,0.504977060227258,0.2224236530849395,0.8073947002606138,
 0.2052573237356905,0.23664777924565905,0.6791695818415563,0.9685785444108811,
 0.23997973387557647,0.0011690836174851338,0.3294546383939453,
 0.1967655061966631,0.31398735492924956,0.17742267857930727,
 0.05585782862633892,0.009499777324016634,0.8851193048535596,
 0.24894642564643765,0.2856674097534997,0.5401096533907295,0.6495689249054443,
 0.8852370014564588,0.5466309324294305,0.22613597617403147,0.9639414200853847,
 0.44752104589249764,0.3974736637682854,0.5596433279443532,0.7502894270677843,
 0.9601412163149754,0.9628754899341602,0.4974861099569243,0.7198072608343646,
 0.5122500850029581,0.9799899531893081,0.22125559681008888,0.7465294087571861,
 0.6749071120093526,0.17636479799816063,0.008275989259931671,
 0.4568236503371377,0.18212713833718686,0.4560119467931296,0.6248365381402389,
 0.783920104673044,0.5872624173219596,0.5494685339668998,0.7766827180466391,
 0.5053164152978784,0.8531631854256082,0.8356320906149834,0.18275831015271515,
 0.2604158261745284,0.5236771737696955,0.03142052900948711,0.714093953419087,
 0.8774644405180945,0.5674633780492343,0.9860616038331552,0.36587519825467596,
 0.7108072330510398,0.9081560048694315,0.48232107527489454,0.5953025006480479,
 0.28870412400469225,0.5027804273577456,0.2602913209275868,
 0.47596670893641024,0.43943162312324824,0.9419872916300223,
 0.6976043568764537,0.7541203219224061,0.15062014506734056,
 0.15836150627831957,0.029491268561695483,0.9425973181306524,
 0.1633042118092729,0.9895476440942985,0.44198370188934966,
 0.10263139757928719,0.5259041998258179,0.014916498227165342,
 0.951620871605507,0.5226315878262943,0.5609149002298166,0.7294391032998793,
 0.9505016738053941,0.6129272379699716,0.37231549347234427,0.2086598927445319,
 0.2609318508769223,0.7983126244184205,0.3541371450266798,0.6717533785433772,
 0.36379576028264693,0.4229655667767639,0.308758013338958,0.6010100504087775,
 0.16276430264628217,0.7300113313250229,0.8314344169204829,0.847636434005863,
 0.33628323084899236,0.06906138665781969,0.43596090140841837,
 0.4027325233594098,0.7606805780644348,0.9814099693331795,0.08229465884872933,
 0.6676746346274649,0.22409469265615467,0.17507735568856075,
 0.3659832680489864,0.933696328792228,0.4676860721501772,0.006340810361820992,
 0.4711790468345969,0.9019001036019185,0.8615236379141055,0.6784608848617374,
 0.0049851319775532366,0.8862621589433508,0.6954310474417722,
 0.41292216021839434,0.5208095739843348,0.7897473729521455,0.5156214121450335,
 0.36707121047646496,0.6013403271820096,0.12330220299117789,
 0.6578963523740766,0.9134656859301502,0.2673521925610267,0.9760435244589013,
 0.5879472936518673,0.10723010753945872,0.12281353039548148,
 0.3552737884255054,0.4983610712490584,0.8888355589064388,
 0.005651345607868619,0.9434912593246222,0.9502828206339275,
 0.8761814246213576,0.2658323767449108,0.8922921628791426,0.555095582406692,
 0.03769464677313039,0.045696682467420535,0.8418489163912866,
 0.7285000291762785,0.20196494523535136,0.45952593374009054,
 0.016111032647799672,0.11269903873535347,0.9372805624400111,
 0.2765599960223841,0.7211776967779148,0.236652552825391,0.7656492227109768,
 0.7453109485394246,0.787887507936721,0.4283274210772495,0.9911739679076101,
 0.41665227905971713,0.040189135932165776,0.14060373060799236,
 0.8188219250488329,0.03860972992044931,0.37422500382899326,
 0.8761123720035346,0.9866592357471794,0.527232929708646,0.14196268405585144,
 0.11650425853392954,0.5257953897864415,0.3996915177437156,0.7919027364596145,
 0.7621102395649042,0.3292008768905247,0.7190416301620365,0.0662143681300249,
 0.8815052194902406,0.8930874920489759,0.8259731222098206,0.9273707592894597,
 0.3357603740776819,0.13127341453173613,0.6363844654709292,0.9414266892485696,
 0.09374064243558222,0.07415949637969699,0.514025843576799,0.2752075547273276,
 0.3317954992046549,0.1260247379703534,0.8184427672243544,0.6293710594449052,
 0.8339119336096299,0.6343592543686238,0.8859630958644285,0.7703134953407077,
 0.5252505888039746,0.2551534930931776,0.4438565906346781,
 0.019871422039619335,0.547015142076615,0.6360883599419346,0.606821648483608,
 0.8316129971010678,0.11476268671744982,0.4122199171662453,0.8229427219983589,
 0.16530773505659968,0.23336869624689183,0.44009944433249704,
 0.7931280578365172,0.45467248564702223,0.36512314340408025,
 0.7632860927949476,0.9878087676666346,0.62848409026727,0.022854137736228797,
 0.7868907104412911,0.1039679854715464,0.6708099446760536,0.2580603450701421,
 0.20679883929740595,0.8790573742286828,0.87799348699945,0.4551515755269717,
 0.7690261672713444,0.2608064098932449,0.510590280002782,0.33051747277957055,
 0.026209914997394,0.36867848554138016,0.8677260830148703,0.6304512298650503,
 0.8720953431498842,0.9913233563388154,0.5853073817550428,0.7657984539970828,
 0.8308674934164522,0.7545198641666278,0.85893661929574,0.30429538654452915,
 0.7807353048410204,0.966430893097034,0.254098370654542,0.7230750341696234,
 0.2592144865948607,0.3861052247476129,0.3823582852328328,0.5134707650087447,
 0.24802630140432658,0.2254438985815891,0.8113972412352493,
 0.37347948885076687,0.5124207826944664,0.9791127864012407,
 0.29953551514198273,0.7615094103783675,0.018655678970852185,
 0.4209316743916208,0.3976577169221913,0.38129079021021817,
 0.034764099364341994,0.9901470218759034,0.12661129864272969,
 0.6502655576679857,0.12928863806993274,0.9906840133590176,0.768527401186596,
 0.6865403657782853,0.14664652768587494,0.6117659579373064,0.9315774297312361,
 0.15841819177837602,0.5490395194342932,0.5220892961906505,
 0.11857065671908629,0.2031196535096298,0.6671916737560926,
 0.047899656354952524,0.2595384694296383,0.982789750314491,
 0.30429855142631257,0.4362169506028296,0.645075300775113,0.4375970653769623,
 0.9502763203985604,0.6971524826078649,0.24298696992018398,0.5659868948513195,
 0.5902186974894595,0.44740812717756184,0.18362441752966752,
 0.9630826519694589,0.9199580763099998,0.7445230932307902,0.9082766846100727,
 0.8168815887879329,0.9361747678451227,0.19287569077133448,0.5859924280893716,
 0.9364369005887786,0.270015074104238,0.42002456589947423,0.8335246489608583,
 0.3118880899502806,0.04625974146259404,0.256225875813333,0.3046403467006754,
 0.5202474418289246,0.47717899775785155,0.641470493421947,0.8889529851418161,
 0.4310361607614517,0.36126893749323286,0.1807663417982608,
 0.12203824829288723,0.9700476223785617,0.4651110143057974,0.4239749133415569,
 0.9830947921845684,0.6459338777592643,0.3323436971478049,0.943916993299293,
 0.4540233134388735,0.21850253030877087,0.24186524619617067,
 0.34681979131905405,0.2764476907862289,0.4692973751392939,
 0.19680012631015575,0.3892299376556897,0.032427719437823965,
 0.8619815823218534,0.40248264612285045,0.8246100554160104,0.5579244931421943,
 0.321498684443716,0.6154362440817907,0.11497043239372007,0.5128692045523313,
 0.5193123773369969,0.5855209242603824,0.6939157457536882,0.2862219890439718,
 0.8848510176666741,0.7354715758816674,0.8365108475611535,0.3719565382128879,
 0.05367649726620094,0.41348228285039523,0.13672350248444454,
 0.09029945449417665,0.8529827742579789,0.7828309689876191,0.1401325857617418,
 0.9696297862127559,0.09030290044782274,0.19430902321427768,
 0.5105239046593583,0.591887782345751,0.28672602132951464,0.15540722575062793,
 0.9084025276343091,0.25606722988194375,0.9404471773425823,
 0.22306720100804756,0.23063104678099888,0.38825801874783006,
 0.5024594815551244,0.377643122750409,0.02090371148214165,0.9003038885449658,
 0.9809191298534286,0.3379633703697773,0.782167911320436,0.5736293724857976,
 0.7867601820260681,0.4209764194667759,0.48422054784572033,0.3289517801680324,
 0.25721303954854025,0.5422290958909266,0.8155795857965923,
 0.20613722475018748,0.3733777746417677,0.6824671003201056,
 0.07288565844270689,0.6475149300244412,0.6028735144196258,0.8924819441787946,
 0.06554011395164583,0.7064783427352697,0.3201367570189203,
 0.45257390623209504,0.23943120796284745,0.35457882957271525,
 0.41041764988141805,0.7722557230192846,0.20696272045836073,
 0.12189418965467147,0.06801380996800044,0.048364833236941296,
 0.09103962459788817,0.6977207297165704,0.6215162421599179,0.214090467453245,
 0.2554615504228168,0.39567434946618074,0.485509462331442,0.41799605198317935,
 0.21423899662449264,0.879179358054884,0.4092792542018503,0.9871439512857811,
 0.9223165343325127,0.275648791518869,0.5129192261423231,0.24509753322929628,
 0.9932176724464925,0.29993921932214096,0.6638112263630893,0.9331678332880443,
 0.8194754121074536,0.8193951923964359,0.17192936357233113,0.590444099455652,
 0.5391882193806656,0.7064401354711785,0.291805798240108,0.9039919227442861,
 0.8523451093030502,0.1374230256103527,0.13031782102823786,
 0.35215250234151263,0.21824091342982943,0.2551015438165515,
 0.33033965982280566,0.911697137703446,0.40638268541378786,0.8739855442293881,
 0.8868140136135295,0.5321786406475629,0.6898281314712882,0.3721825888553405,
 0.20203377330440042,0.2566982275524152,0.4230705394809873,
 0.32962089832338604,0.8367386052255988,0.8053794458449016,0.5527345746111948,
 0.0012422962697035134,0.9699623751989608,0.1999167200073002,
 0.33372550989617233,0.9512597236710554,0.5402471414806683,
 0.03859509963860641,0.4096579920695176,0.9327157596793881,
 0.47130190649801484,0.5536483720075189,0.2002744873212915,0.7277352858061893,
 0.14684439147384887,0.37459304246071967,0.9611430673156672,
 0.4298206193248748,0.7602956901467561,0.46009342048484525,0.9158834801591421,
 0.3858556508461589,0.2301710685776156,0.33835952562048444,
 0.017961550542437754,0.5893389800890568,0.606806580320545,0.4104517985925701,
 0.9300829313646674,0.5197472221216266,0.7144109760104654,0.04465448463778998,
 0.4771469677464564,0.10820991948481973,0.13215918240820146,
 0.2902140943801961,0.9190253862993779,0.3159510968203536,0.6462781637134061,
 0.6912663489247964,0.5636089914200771,0.031314559823148014,
 0.6859241725694905,0.6142928023727432,0.3302333023245111,0.3883007556210085,
 0.9156734938924759,0.36397325897811017,0.0021700044724723977,
 0.2916663705992313,0.5634878141513503,0.5202523325040291,0.7691857716759767,
 0.8713842044927735,0.9138764752981958,0.1215000199423365,0.40987853776814187,
 0.8571843640282157,0.021209848681535082,0.22868319870435327,
 0.042648986661592,0.5910198921354355,0.9215506125818044,0.5824573152338967,
 0.44216609661000994,0.13106591963519987,0.14724561939077652,
 0.5715791023109448,0.14901801892590827,0.8319299663587242,0.8372642012076257,
 0.7212732954349077,0.8030915611146321,0.29163846364563684,0.7625734361230484,
 0.002885680280413805,0.9072206937180054,0.5191672272514918,
 0.3980244168179873,8.913251667709288e-4,0.013032928549937206,
 0.9310367988141132,0.11665369983208107,0.5091853855453512,
 0.45450049023687533,0.721238218724653,0.5446197370898355,0.4495851226698462,
 0.9896026461273916,0.2149006384283313,0.8788325488377563,0.14569171942456072,
 0.07281651188332683,0.6360615540383464,0.1760158809192348,0.8773329260584769,
 0.27791688541084436,0.8752355733755215,0.7630570970389964,0.6626474426981697,
 0.13128753839140384,0.9811648191339322,0.5457607835670787,0.8603605118741849,
 0.8164320924569847,0.6288072277020166,0.05967133062716701,
 0.13497407079984058,0.5953503518826073,0.9504976756292388,0.6715642850231333,
 0.008938553134498872,0.8720075738658224,0.734660261976783,
 0.029074872466034685,0.8372226386384098,0.3247595993644411,
 0.40609975455784975,0.7970920881697353,0.7109351662730781,
 0.09719669201150138,0.30219391994274947,0.6650403712277047,
 0.9124299430790568,0.3069692250950272,0.35006545787010335,0.4548854402449757,
 0.9657227318043955,0.40932490837686564,0.6026570609906746,
 0.32797305139884203,0.28980365118515494,0.7135156104629441,
 0.7873343960989637,0.9462691225425572,0.3586830933123444,
 0.008051018540161103,0.7134528960094269,0.6416656595746351,
 0.5546841570788666,0.05831832442607099,0.7132245240208955,0.3516907236935968,
 0.9075321594106306,0.7087178437261453,0.63081766565008,0.2549101963481937,
 0.6477117648026284,0.345708004036132,0.699511898281266,0.5109165008094205,
 0.003708004371915319,0.3950337489649445,0.0033590107231465938,
 0.34894997825716956,0.5959952742295581,0.9635222430858201,
 0.00743485761284135,0.9730902232277372,0.8944085221048184,0.7962439221524953,
 0.32614392908269696,0.5515736730251076,0.7409087733693777,0.9830090336303209,
 0.039213957588088766,0.44935676050749807,0.03845877101629225,
 0.6174884963546021,0.21490303329737825,0.011211577230553438,
 0.9195827203289211,0.2529332577565192,0.5656995336549037,0.6748049004966652,
 0.2811581542674825,0.6208157880959284,0.12064851907770091,0.718452258612099,
 0.16167875446515845,0.7362145395719951,0.17035419704328691,
 0.6480218648009363,0.9511078181198358,0.28702195941165687,
 0.45801206595740485,0.140768072533496,0.4464367459102463,0.5383850473654331,
 0.2166212266980816,0.21542753699456207,0.33755716116867696,
 0.8272112408372909,0.6476984031683215,0.07255738368969478,
 0.08616775022792478,0.7569750651098622,0.532960440244441,0.11682633283758892,
 0.295524814816302,0.715354317088986,0.7665833896644672,0.12415522035181414,
 0.6533796004851583,0.1989388249352232,0.7257784854583883,0.19369348041681467,
 0.7948199397821264,0.3334777411438964,0.9915426467886358,0.7866869028547083,
 0.5807024017510606,0.8003060142666846,0.24382382683593906,0.6294586406218576,
 0.33776544530078867,0.23678494668540684,0.6920315777453893,
 0.8412859130583477,0.023142399011776504,0.36241145216996495,
 0.32081955759153924,0.935437815062452,0.17757705185137396,0.7611426537488839,
 0.4651174588065994,0.45339891249432174,0.087453021762252,0.573790068154162,
 0.1367084364266543,0.5663778258771738,0.5786579940201892,0.8422076933373059,
 0.9377524445819586,0.05802545433329853,0.7080535621537065,
 0.22601000832353768,0.8368689830342622,0.42688234696057015,
 0.9373357675185149,0.9877193602451257,0.4226163769173674,0.9360052848087199,
 0.7248214669674954,0.2759050644053713,0.16478947385312803,0.9916687261299175,
 0.8775986353516264,0.42803132871823046,0.9258385188845895,0.8131533050605074,
 0.34255084747644904,0.040002530099333744,0.994314927642743,0.169433602013905,
 0.23963628122453706,0.7932594438690698,0.6374757941935252,0.6183876794835386,
 0.4208090101448335,0.1414980388958299,0.30947628884997413,0.2699525407765764,
 0.13854876856643017,0.5227054598948339,0.7586820867996582,
 0.47401604795302843,0.4368453361872362,0.6073755083880434,0.5763842949283497,
 0.7794453110019927,0.5643512833174434,0.5936691276338313,0.8376879111767718,
 0.902166397323712,0.8314083603023279,0.2605896654103178,0.9546369073196143,
 0.40058013753540844,0.35783713973698084,0.685773902652581,
 0.22166817220915025,0.5636263238801693,0.16238159697271504,
 0.8831416085422263,0.9100104763676841,0.08914856082958167,0.3274002854150351,
 0.9484141410462803,0.10123925754856666,0.1545207001031056,0.8797948724333744,
 0.9743643878560396,0.8629315944045842,0.48756540958515504,0.7574508049730817,
 0.12976751180920376,0.6341656295601306,0.7367757154078018,0.9599573980023716,
 0.787982564819526,0.9663185757600101,0.058672299817600915,
 0.05957078682913963,0.024645186934085683,0.599566122062037,
 0.5180906608673497,0.45319615690183,0.20344101456775443,0.7376348259323787,
 0.9302930694599453,0.5191271329478695,0.3774971769474402,0.097099579090834,
 0.9526613645224393,0.8822114169957344,0.39647340003159504,0.9340528077891725,
 0.6012504294058915,0.6731483903171515,0.7696476737127294,0.8562348573636598,
 0.27789118854125916,0.358972337379005,0.4534890842260323,0.51346358835237,
 0.6776067059550632,0.7212863463086402,0.8910070572369508,0.6539458377854261,
 0.4175140633090728,0.8503170833154126,0.37162484330353607,0.517884100906041,
 0.5058844057461113,0.41997175467738723,0.5240565918179793,0.6879641782096296,
 0.9048403218996585,0.18397100131425592,0.4602628927148472,
 0.06910363958045684,0.1181699414763987,0.867312433068423,0.5742444204415336,
 0.7376131559161804,0.3710288061430296,1.576437880618231e-4,
 0.8138992152750896,0.5580352618840918,0.5143568439737783,0.5008829886380887,
 0.41790687942088245,0.3143528172163872,0.43925163182447424,
 0.5101321658962125,0.6819794462961503,0.8129216451793151,0.04500619187008881,
 0.3002738546312842,0.7627069945798177,0.2949718241170315,0.775828179275724,
 0.0062883098233139645,0.47681718350645264,0.12791597114484154,
 0.43576528002984616,0.37127825000669623,0.41662730802682935,
 0.38311317626824104,0.10367971906988249,0.2511317317420929,
 0.7379532356373248,0.7379987244350696,0.6810787736224028,0.3325143772898802,
 0.3854144278038728,0.019408015723069938,0.05340026847889878,
 0.006512727407779639,0.2394981059621133,0.055976297661420316,
 0.8102399832308719,0.30508489490684854,0.3978002164863167,
 0.07936702871225454,0.9740948879218336,0.6144431252463131,0.7439472886018231,
 0.23374268591685765,0.2166903740562337,0.42632370564457733,
 0.13366047675897352,0.1997321063136278,0.9929508376482208,0.0478828927556485,
 0.523818035064574,0.04781435087032326,0.9235782737141432,0.7379724752564986,
 0.4703377239740665,0.7493227783584581,0.2259302866549917,0.2400328880141891,
 0.7237461495097793,0.5995545797427645,0.9455116708321852,0.6817756675533486,
 0.08245839328839033,0.3444141895917284,0.7988787938682913,0.6433818214092853,
 0.36612143560423827,0.7980969463024201,0.21038121477185523,0.318065336754378,
 0.9651711391422437,0.6812588756408702,0.06477009631620001,0.2883040549459328,
 0.17988958575986813,0.17661739513604469,0.9590098678027629,
 0.4824049643241396,0.5162868854897831,0.7761865500769453,0.20904198720231648,
 0.6771577280797038,0.5549362778676078,0.34857264587863246,0.8535731103901222,
 0.16473685748725442,0.2975083774136579,0.18489262561012132,
 0.29428174403125995,0.7264762501785971,0.12723983107998638,0.641595888599872,
 0.45721520492047185,0.07571525630208098,0.017972903423463138,
 0.4467002337245197,0.9444869458586913,0.044371777649357846,
 0.5177256048507347,0.863501324269505,0.6278589694238892,0.09973923659198036,
 0.45598627046898255,0.695999732430756,0.16347270342978126,
 0.43149130948307435,0.5113640768844867,0.8222608720736453,0.7093475438638657,
 0.8182863082455508,0.6353221361791147,0.5730400324765081,0.30441161511997317,
 0.4425428777301963,0.053450324414258454,0.9525394528017522,
 0.5945260070139502,0.07136392975587369,0.6719776466136683,0.6079487031894735,
 0.9653806335416129,0.128218570172851,0.6366545308805399,0.7440151639314623,
 0.2498445796948603,0.8564210498502908,0.3595994957212022,0.736971643626949,
 0.09334390286629757,0.6910053436293183,0.6670273391875277,
 0.030622401676822175,0.11749220235132496,0.9588527793490855,
 0.5773242677204437,0.5046175490931377,0.8877381103840116,
 0.028287196583975494,0.924598797251793,0.10368811840204772,
 0.7895354319054384,0.42504156174865193,0.6784922569797269,0.8630366168550752,
 0.2630517498437258,0.7144507074045712,0.1719932826236975,0.5672426139158532,
 0.6806544333158474,0.5111040657014001,0.6592520233605762,0.6916014968540138,
 0.15194307834338894,0.7596696392073118,0.45584052667352115,
 0.4070779572588419,0.9028582085222819,0.7068749936899853,0.1961937104898478,
 0.6786400783914226,0.6496401482532906,0.5859412269638766,0.17177030226710688,
 0.3369215867233414,0.5719685291202317,0.3953971930801171,0.08910775574888019,
 0.26105285739090645,0.03241539669232696,0.41823862195247685,
 0.8681265675165988,0.976648149390463,0.3577754616228783,0.043362997624386446,
 0.19166684957468194,0.8829710410650388,0.07766606822642841,
 0.14120464946556965,0.27951893943782247,0.7757507184895898,
 0.6365697036293843,0.620471036875795,0.4313133893806915,0.4319221388954425,
 0.9712735931909378,0.3502180326268556,0.16472713494879287,
 0.30329811105702575,0.8861336577015534,0.6104013122169694,0.7749910948163014,
 0.4839869333330544,0.9602123583280606,0.7836716955234684,0.5504257456569261,
 0.6265271894569145,0.2521441549646375,0.6597627298792162,0.42074704403406715,
 0.41335539564678736,0.9789317027363666,0.39099140122108056,0.543476632665752,
 0.26567997541480737,0.7914150263312305,0.7133885814232481,0.9730219402212184,
 0.6740565402587327,0.9544909056830847,0.5402147598086104,0.13839144702139472,
 0.2233402645283198,0.38646507189719503,0.941813407385824,
 0.031319076967126214,0.3388909087520363,0.9825378040379604,
 0.22628499626842158,0.24311856714176072,0.014243979911838789,
 0.1700267261030235,0.2823892490258779,0.9447780328030443,0.35126859687502765,
 0.7521288239308033,0.008736942609118614,0.7725779182342025,
 0.6578100031598147,0.4038883067991891,0.7323122008098191,0.6270353862124237,
 0.08567328264306195,0.30102715631330346,0.2581035595056663,
 0.48369021315757244,0.2895965929882841,0.9286360082246853,0.8140736144155302,
 0.9739619807295153,0.948047615784231,0.7449261542667126,0.675162025990649,
 0.7625201975333255,0.6203808146065903,0.9136979629704842,0.5329875068703601,
 0.10133970889207744,0.460221656743947,0.7068108481593585,0.9669986008571821,
 0.602887452959625,0.8518665005223359,0.7013714115925993,0.5078550744631548,
 0.7267902455165979,0.871702846448155,0.15886451856988737,0.37969143130905314,
 0.5664351866283256,0.3268567242141608,0.529356137271064,0.8911215680058314,
 0.7356272102952053,0.1729084066776745,0.7665885147215424,0.9788268609982269,
 0.9371844209332311,0.4113307410947534,0.8536638230576404,0.03687206154361289,
 0.20075504534016375,0.2232150375486337,0.4985219103539591,
 0.11782554658508093,0.004260032797135493,0.6913844467359969,
 0.9233877268197945,0.11101688708626578,0.024764703811059663,
 0.4742804100919731,0.019779201100961874,0.3412518229332544,
 0.5539663553855041,0.42115003532218476,0.5929806809955042,
 0.13235671985029474,0.8193789553218388,0.628444882648819,0.09880802006621181,
 0.19230167607288962,0.6943704162260637,0.7841517283712551,
 0.30660911017521986,0.6341658025233805,0.5041745522938507,
 0.03669388057180922,0.10075527793901906,0.6675980986750694,
 0.7613322236546947,0.6046730429509035,0.11533103416748525,0.6376351570029317,
 0.018550291905473637,0.41510529597084767,0.6374165980755747,
 0.9472976235615709,0.7971142589777659,0.03205557088807598,
 0.23722715610434508,0.2794291655850707,0.12748470115370492,
 0.7679610561074219,0.14632802664538447,0.5422237688652602,0.2607371416627471,
 0.1659288520463893,0.8231371004071606,0.9639628359389506,0.1625599902883419,
 0.8301145301182425,0.021962315539447275,0.48691549161871606,
 0.1865161178100706,0.019550596789724928,0.2012832711252397,0.882470704110808,
 0.2257177193593426,0.5411292030247329,0.16145116671154613,
 0.19185281709570945,0.9158798024833121,0.718143573799676,0.32439799111792333,
 0.6709638691132193,0.609348231923779,0.8204919690604329,0.9580142503342094,
 0.05047868266044098,0.6440364969097051,0.5884163167953387,0.8741535604038921,
 0.5038679500714369,0.32526670262209423,0.283403633794985,0.2528023167699943,
 0.14264925487741187,0.6405211763721415,0.9671102312268963,0.6095330808212168,
 0.36308699934980926,0.8654655857532404,0.38941517427202266,
 0.8834221074303575,0.2698232423107727,0.5580600374090967,0.6445884133996028,
 0.7033573330726052,0.017773879944286364,0.3807368591567921,
 0.3020926646194201,0.2362023471147735,0.5649948699003384,0.8373592718704437,
 0.16520408487536797,0.9678126026619779,0.8640898614503013,0.9835514312349821,
 0.3299959807575832,0.8104558311494063,0.31350328994770194,0.2215617971196867,
 0.09888842047604451,0.6211780025278305,0.46444303417710187,
 0.8234357684109536,0.4770969041107246,0.8875633076538284,
 0.054716427566206094,0.48525779548602166,0.3461275369114132,
 0.2643853407981155,0.78617740392758,0.8804381384707036,0.9318510226141112,
 0.13259078879580655,0.8472464073042671,0.3430199476751188,0.5558525090341104,
 0.725050129380681,0.19039818257875474,0.6852476572668739,
 0.015758362931400427,0.3495783166328501,0.6710747900297326,
 0.8243307495371615,0.5355078793924452,0.1974353731203946,0.47492988514867185,
 0.5998553455519008,0.9978917465060618,0.7681248706864514,0.6481408300159429,
 0.8184113298848901,0.33575514788232974,0.5742615352558843,0.5956917724522639,
 0.3823579199473862,0.13408426208218893,0.21504406147630306,
 0.5216509516154272,0.5912747818255963,0.8648700470879865,0.4170984606500353,
 0.18408800371891154,0.02192249174279648,0.4616872913820139,
 0.12225898135439817,0.4067422709396542,0.5686567550392992,0.7871600744207012,
 0.012445882385363372,0.15726891108288255,0.3631247852426026,
 0.7217203226879066,0.3451053658694829,0.7021118930292416,0.9106077465042449,
 0.8846358084082696,0.27641314975733344,0.6494025892446909,
 0.005054200242801898,0.810003638078674,0.4998208612092071,0.2620531438683913,
 0.3247973910272629,0.623637981985168,0.8537033929444439,0.47842999901656436,
 0.7867948109881305,0.268033636890895,0.3515843733281785,0.11208453334902735,
 0.2307775243630873,0.8438951347475094,0.896596067954491,0.022151770063517207,
 0.4417431695301097,0.30351240273103663,0.47571425322824357,
 0.5642887886877426,0.6541996515278885,0.24368368367834758,
 0.33251011357693816,0.44132949123603005,0.7263866858462041,
 0.5169433967668366,0.35035227423782267,0.3373491384652908,0.8915196123005613,
 0.21093586767725092,0.5625636497398753,0.13420806689706444,0.400209592777818,
 0.5759972857634116,0.3051194651113246,0.44651892859208675,
 0.006562468326789417,0.7058234996961028,0.49809158726334046,
 0.6798564045168193,0.0021743015799667864,0.7114149069893001,
 0.7838491384786868,0.6602855908318189,0.7140987732965008,
 0.042288660718140525,0.8471916534736534,0.08324433927625086,
 0.23055989567469448,0.07726451760868858,0.7956947427006366,
 0.23016715563857926,0.8922114492062304,0.611627278515998,0.49224569801804996,
 0.9046859257235987,0.0063559067864651375,0.05874847554422202,
 0.7055308787072669,0.4860089408081103,0.13983982674047646,0.2261195110513612,
 0.5495526583639163,0.8480687727050086,0.1667688661170449,0.558122348260852,
 0.7419858775009663,0.5675812512982019,0.462828067856377,0.18956902523950703,
 0.7947983216820161,0.3660516220134572,0.9164790219122798,0.8768665043933963,
 0.19289614165857494,0.6092485834773428,0.3826154461776996,
 0.11910489157140991,0.6743434086361078,0.32519491996031447,
 0.6302263592696326,0.7506947446599028,0.9228575371459418,0.4676022428033755,
 0.8991091730745765,0.5308352643018873,0.2187403177360231,0.2688111162678246,
 0.8596382538227069,0.11642736892784811,0.667099469404132,0.3359269242602009,
 0.6870053540309038,0.43921175733437345,0.7592746560593762,
 0.45600067677094425,0.8754753925538047,0.9631573394164296,0.8592519796048708,
 0.46836046173646695,0.766832996377146,0.4005961728272375,0.9633024628697752,
 0.9347621994993647,0.6681237961154971,0.41907703418502174,
 0.28670255523471433,0.6102170534598391,0.41377938863753994,
 0.14329311095565744,0.2956713653233709,0.2735545228167122,0.9280783114275667,
 0.36636689747476503,0.12910443517892434,0.7170445519354789,
 0.2512962110886797,0.9075393573090756,0.1272459071653207,0.8936284883473918,
 0.4138931557724458,0.14012470303032487,0.07476640255898137,
 0.6595466987237916,0.29366067221183245,0.22328833678697935,
 0.6999677234348424,0.22703560411204715,0.30238769662916387,
 0.6124513304306439,0.28168007811014417,0.9968246129241058,0.3962869372161153,
 0.7364053381514903,0.22916334720254838,0.888142835002653,0.9372632438815064,
 0.9529171834793274,0.20960434999243482,0.2654234152802608,0.7663473509657365,
 0.22343338219247033,0.9567115662319223,0.6849874506102666,
 0.07839780977423261,0.7886735851828863,0.23072835667611524,
 0.18767498690447937,0.12491626060787153,0.023213355101979083,
 0.9598396465678245,0.2457468321008438,0.1074821362987507,0.4639652023662588,
 0.9327837231657841,0.9259001022985519,0.03534135759175727,0.984155449866109,
 0.8780066694788393,0.9387186713579825,0.8352510830483104,0.07104058238223665,
 0.8552421053126049,0.6205613801524758,0.47640534100227594,0.5465020382413875,
 0.2599981119597423,0.5108343384890566,0.15858920278632227,0.6949334990805153,
 0.755084821357658,0.4401196346762196,0.574405185079321,0.37370024969097326,
 0.4710696782303323,0.9444969075336999,0.8990462185370081,0.385973942751506,
 0.687457387876738,0.4776530607387208,0.2790441453397251,0.7428180429940678,
 0.36474101291711536,0.5673721368655378,0.4057630872156154,0.562426557676029,
 0.16711382120837048,0.08471913849847501,0.4652490444289794,
 0.7388364828628721,0.3536206837660866,0.1840228497598303,0.26053628045885924,
 0.8222960636175309,0.974302747662381,0.8516501214478011,0.8793738453644663,
 0.21030638032945304,0.7510292281881888,0.8208608484709328,
 0.30154266211237746,0.9754127887617557,0.015723875850936597,
 0.02330738794666254,0.22251784559956167,0.12718860987046043,
 0.1590704718468503,0.2552855178795922,0.8643518387067763,0.6945276070517465,
 0.5926690469786786,0.27584560483333465,0.6212552582091624,
 0.18982571128114567,0.9108612256335853,0.5463668248409556,
 0.23994461822273783,0.621599511916413,0.2442162237840898,0.3096882508445613,
 0.05309020976916312,0.7936719200766278,0.5119040752749526,
 0.01271254355492768,0.9857577496664938,0.6520991070225288,
 0.08444511345940364,0.4825022494946891,0.9381026982307548,
 0.29149368471902226,0.8882276487137583,0.20036405303678917,
 0.18432233185391722,0.9254827062683598,0.2421307272618427,
 0.43778668621370986,0.7692789484476285,0.6784217531142358,0.6495829650144689,
 0.07888052911412158,0.8465240915554159,0.3027881076141371,0.754207341691588,
 0.5271108454650755,0.38155640229221355,0.13833314897992177,
 0.8797371585862764,0.20586466610261145,0.23032037092979585,0.84490069617691,
 0.38599862738951907,0.10047199446142252,0.8788324390837402,0.838368629131589,
 0.7194201835007167,0.38835133952429135,0.33930297788753494,
 0.8532657657372255,0.22498731780618608,0.2626435917726976,0.0280441881426583,
 0.06150354742341757,0.06806197972684536,0.20120991827592916,
 0.02721128919602034,0.8195118609717437,0.07230601827763361,
 0.002375009393352112,0.39792158374952913,0.48848176443887736,
 0.24012514080605607,0.09988002931776185,0.1731966768492217,
 0.17077760976357037,0.7956162750020374,0.345585604671495,0.2282088440631307,
 0.39205950666719647,0.28856740376471823,0.07270036210720132,
 0.6434446480965016,0.693181559018007,0.42753451003695075,0.8026182011575179,
 0.8875653784630475,0.5215486437863934,0.2547957669184513,0.4972749105629273,
 0.9127103049895254,0.7688485939523848,0.5451529430903016,0.4624870126968321,
 0.615610777737329,0.2726238664350361,0.6671641222884462,0.28430339554023165,
 0.2172001322918813,0.5991018997795082,0.8434499308558535,0.6535222901364243,
 0.25829685704116256,0.7210037313375486,0.7705205107159305,0.7458690710589408,
 0.5610346976634538,0.1688129473143407,0.4110954094916114,0.194050917863961,
 0.22555867390213646,0.7383511365402701,0.015184901296312336,
 0.4666912282116329,0.9118448876349283,0.8470765246211837,0.2184252737109308,
 0.2425702905383934,0.04599797151670848,0.6218115755259915,0.9975161558341156,
 0.4950106684902923,0.34564684663041956,0.3650635871449801,
 0.45125371518325286,0.34387372625409185,0.2333628569493007,
 0.43157582874144995,0.7040730107390749,0.642330210824298,0.36018529829893187,
 0.6311232784883012,0.3657973682046284,0.6779061922551656,0.8070452007648738,
 0.4113907119378927,0.604760022204518,0.6244734474686816,0.9692003913616722,
 0.6527007135226522,0.6361489020662191,0.1292159076808127,0.42683666008311005,
 0.16887012859104011,0.6702039445735734,0.06788634858286158,
 0.5961541241870171,0.5702753072174936,0.6589801231505961,0.3584460390712373,
 0.7383210854141486,0.41130841586978895,0.6169884731188398,0.4906954399477834,
 0.2950693216664986,0.7927092569257226,0.525667863701694,0.6140966610919125,
 0.23633862919230064,0.28535975159906646,0.15616486758986958,
 0.9554815062298249,0.4120654279274394,0.4075863794233099,0.9334338575672139,
 0.29022667765382015,0.9731473357038888,0.9781117157725313,
 0.16037380237247412,0.9366209278203056,0.4344813501958975,0.6706618783653555,
 0.6956681869124259,0.28754290360617396,0.857803187989949,0.23679221031526088,
 0.07278689500843649,0.6317923928899971,0.4039925847395486,0.500677545675992,
 0.23554188344901172,0.8409181627722252,0.9064728941264127,
 0.08550318825171077,0.4572617086693135,0.9003527737137129,0.5224430975606625,
 0.26187454114960074,0.6918932300690211,0.6759051346883567,0.8034197630534122,
 0.47894539374117384,0.06099663105945541,0.0636746608654859,
 0.9960950495193365,0.12162560713200421,0.1720809170282983,0.6812622228694546,
 0.43385526992189705,0.5329506029962459,0.37096318593431654,
 0.40118612647024987,0.5898660329016229,0.42433901396261775,
 0.025848500778273498,0.9821130117669566,0.752302221925658,0.5725310475558386,
 0.23075311696955192,0.39378842217212007,0.7659067326980206,0.28530033087757,
 0.7506424713976952,0.6044787719298814,0.0297577100420211,0.8912400311941391,
 0.8274309737369309,0.3894613300723848,0.34310721810286104,0.7151737781801383,
 0.9298949435208955,0.8919106613618226,0.28861627433142445,0.9453666580398536,
 0.7924365723667309,0.8429285055667942,0.5653933146360405,0.6115724248507796,
 0.8289633831748447,0.5138720496472619,0.1991101658693768,0.2640075872472851,
 0.8818783290547012,0.3351438195209109,0.9612192056065089,0.7672883375587654,
 0.7641437232665689,0.4698257196169089,0.12424312766823631,
 0.23521525941669175,0.8128996680393203,0.4705337690525907,0.5732651277943052,
 0.5607957040389511,0.6399247579124289,0.4691456847799904,0.8255354165360159,
 0.6153658923489992,0.8775124694678031,0.8891005926394489,0.4348014661253291,
 0.6699736073426417,0.25574439412214534,0.2540002731674028,0.5147757965836599,
 0.988793135083745,0.4784939906317741,0.09189869210489532,0.43160461162879915,
 0.8818505256688529,0.32959069513714767,0.5662382135900851,0.6052550374107372,
 0.41334583688977833,0.8569454753316887,0.14942508837750523,
 0.24655627299146698,0.2601229201105715,0.9023321774014712,0.964676498784591,
 0.3667903368807389,0.8583892854373341,0.02863609220449126,
 0.030208851660248004,0.6108576789764533,0.964687514932312,0.6239840378678381,
 0.6432589461813466,0.1612007635889039,0.21943723789948388,0.8526301659337405,
 0.879558792952796,0.6682371526702171,0.2509462010755993,0.0699895302949024,
 0.3774678574292458,0.8458527149503623,0.38736844734344233,0.7419627246781658,
 0.17872934014259778,0.5131538498375343,0.6447366271573897,0.5880334003698564,
 0.779818787587498,0.18607111169560409,0.48629457027591416,0.5824216776858039,
 0.612767375378232,0.3761995567995643,0.9629643763653221,0.2483605054460637,
 0.2572535279589806,0.23035190448032206,0.06291456201169643,
 0.2396053370786544,0.34854555969801815,0.04383514809558442,0.675698492980464,
 0.16482154861626408,0.2744207121642057,0.5967778749535723,
 0.06811433699192937,0.4115759689158083,0.19443174173944744,0.798066317323029,
 0.017834546468992407,0.4731008043723349,0.43451652108530625,
 0.1982363984826896,0.032022893788450246,0.8750211071724472,
 0.22362124410121054,0.2223437041354459,0.4218687881462687,0.5262273124464573,
 0.675335510444489,0.5137957073487922,0.7005643560268735,0.34268831171741976,
 0.13235138349404485,0.5708399757653021,0.17082703015576883,
 0.11179884496329706,0.9112652772769076,0.6194069760646395,
 0.43526809292493085,0.9189762932127534,0.9654551049938117,0.9609646431854061,
 0.292568932156676,0.8687377350506172,0.287520122516713,0.7865508309666446,
 0.5394909303515163,0.5686946887751442,0.18338402014338206,
 0.10497479742258509,0.003144274172165784,0.5685413320502934,
 0.38628144638005213,0.23328024084680143,0.9808235011599997,
 0.7109941286369819,0.9263045055260581,0.935690045668427,0.20573561074734614,
 0.011269824968861863,0.55790956733073,0.824952158630812,0.720272334671904,
 0.7531476471455077,0.08006392603079504,0.32202725755145145,
 0.12454363564099258,0.6172733229005303,0.06856159613496005,
 0.018913299661815985,0.024642104505079643,0.26329895564026917,
 0.6451880838626268,0.4180184280156206,0.4254008455594728,0.8949898150351554,
 0.12652784693428287,0.4215585610941345,0.42718616191300185,
 0.1250887938464036,0.08288290617473604,0.42138399003126537,
 0.2182263779598218,0.19357066567601955,0.29847592124521793,
 0.9603661428776672,0.005554978503058083,0.0915968557259238,
 0.5726594749318115,0.23655368461788429,0.6890171776563646,0.2211405403785187,
 0.3221780618063984,0.9449438589821872,0.9833404558949849,0.06657686458544632,
 0.9360773734962853,0.04417379211704375,0.9017786764178437,
 0.22680315289175212,0.8540066415218844,0.9382921808320844,0.9572305307393812,
 0.4779598652008872,0.0675003290396834,0.29126670314356706,
 0.04753946268484721,0.6885423031935209,0.3503584756161836,0.415600259276382,
 0.5353234274704819,0.5017299248305002,0.05255008336441669,
 0.061597969227082605,0.9564402640488043,0.44529725952558064,
 0.5535737445725497,0.7656483252657441,0.37430578723759145,
 0.39430177154066137,0.030659361652114137,0.8751831702536825,
 0.07016050985477784,0.801249464183455,0.4385261055238,0.04620717055071266,
 0.5299942522262524,0.5250940926829052,0.7744447741604799,0.6826419488603269,
 0.844244598562284,0.2848853703152445,0.1829971546274567,0.7529955397058185,
 0.42572844490652884,0.5701184244113144,0.18434226256628383,
 0.38743874344749274,0.9188813186431627,0.9196188445426928,
 0.19005060841216714,0.4410435653333178,0.41673367997022526,
 0.15403023760018175,0.883489829575431,0.14606768871089992,
 0.018507609120739676,0.4484470434060689,0.9072560987691249,
 0.16638893744762584,0.7175396209198663,0.30375649676002636,
 0.5455005336290516,0.05652940627899716,0.5936629542677797,
 0.11110710197621987,0.5380328948133053,0.7127014472919699,0.8821234614909769,
 0.2013604254059561,0.8350751355000661,0.7527981375349179,0.7081174354172295,
 0.9411274046750039,0.8889668848072763,0.8775985273667894,0.21845244293966748,
 0.1994607839148037,0.8538352503234444,0.3068861925243891,0.11467115647618475,
 0.7175921025490527,0.9194488080888846,0.497857611886355,0.5025447762061087,
 0.8219697280998827,0.9880779571839292,0.2996932994364364,0.38296709517320804,
 0.8802858748013176,0.1735808368231634,0.1295041538405579,0.2953742397097684,
 0.4380666010699792,0.3601211248326941,0.1618783146146997,0.03032268325365417,
 0.06729951257427524,0.6578672965201446,0.1954128156426872,0.9514724001779306,
 0.43183133470374035,0.8977363518163939,0.40179603250556284,
 0.5381274148570611,0.6239344977237915,0.7277889695621576,0.4345392125868067,
 0.12230053793490847,0.6099543937763733,0.3304888910463014,0.7054095278092818,
 0.2571928975326966,0.09172420027513983,0.3358429794254514,0.6241536131121778,
 0.3997029105257982,0.33307288780984834,0.10350654128927861,
 0.8647146118880201,0.49819354519650316,0.5953653735871345,0.658560873096893,
 0.37659296535533837,0.26609456245357266,0.3975631509704419,
 0.23678351108255913,0.8521771304987213,0.31241907263129176,
 0.6974400817815533,0.7759577452785265,0.0685457339297767,0.3967755774657671,
 0.6183515839790312,0.7353611236310347,0.7174482680704052,0.8693134641773806,
 0.9398493371086822,0.2845339828887429,0.5097595530567777,0.913736783400227,
 0.9036480933563925,0.6636825111157414,0.9425632722184283,0.7803627200442131,
 0.5830246505571158,0.7581299856341026,0.9007437091326119,0.3055781979941836,
 0.4766063717570085,0.31879451938347503,0.38698096416098204,
 0.7087440934239398,0.9669293812969787,0.29012180263754483,0.702194988577099,
 0.9264049102536912,0.6462987999094069,0.867305247446935,0.5575909340290923,
 0.5296846491894428,0.2404314026583273,0.6874575432446621,0.9456101452272618,
 0.7236920455341223,0.1935258506266071,0.4096673528878142,0.8367628960203785,
 0.33064157229071445,0.8559117598791649,0.9032288961821465,
 0.08931809586611694,0.2596812298747072,0.034221971382223604,
 0.2905599398019838,0.4640305750375009,0.0968042204310261,0.8893956346847186,
 0.8421792761440122,0.28905514534832377,0.3809229011160701,
 0.20878920731071882,0.6175888048849244,0.20554694184431743,
 0.8466839074659095,0.02628242525072455,0.6902780356188932,0.7640924144327502,
 0.7755911031367861,0.6591928932891417,0.5575907396734568,0.8848528646528219,
 0.5298745729445187,0.09829201693634315,0.12175892400819144,
 0.7952727956028991,0.978100093624999,0.7852463667012635,0.9912074453878967,
 0.9217182762861729,0.9548139344011051,0.3327155740044129,0.5787076791532693,
 0.660214154850334,0.9095035219179874,0.23557781869410965,0.27027993023693864,
 0.6791425770208857,0.8227971159978675,0.30602212537097995,0.5329836620312907,
 0.7594674150042817,0.7558403197486379,0.8177250015567108,0.27250431486616844,
 0.2973977406378776,0.41867823179149544,0.9616562380468349,0.7083491832742642,
 0.1700636917743732,0.015941927689007707,0.3758256218506687,
 0.7960125209599417,0.8239094195114035,0.8256456987510108,0.2646889603628845,
 0.6451791555909071,0.622637182858754,0.22819608435363803,0.9587478639384746,
 0.7835891645495159,0.749004842745205,0.39875828065492025,0.9947998962852527,
 0.2559417162709735,0.8847788786080147,0.8483457569518565,0.6549545755789372,
 0.22311397982509074,0.47722917301925893,0.08410011504479054,
 0.24693062435240898,0.03901128765462636,0.1268483141051242,0.809012931915309,
 0.3894110004591539,0.9979041074740493,0.5535631015655893,0.09855411730882813,
 0.7669540061706875,0.914913526965124,0.029670084179305167,0.8751105049008479,
 0.1741665332085236,0.6519128803974227,0.9464615060171845,0.5705078443048475,
 0.47181078010614197,0.575711475901943,0.4870861680475761,0.03062323970544023,
 0.7343530301106347,0.8835843836474717,0.01586159555739286,0.8086361529234263,
 0.6764331895442123,0.6176407078995337,0.4854902755912517,0.08754724528507796,
 0.7066827712422558,0.6579118997702005,0.8012540984961648,0.4160762284272548,
 0.6585608440109575,0.4351184666510216,0.7053251533286762,0.7419493657170346,
 0.6201825996514672,0.08868595259390655,0.5018399030604221,0.5509010143094045,
 0.08641301167881543,0.7637189446594124,0.17549620222476425,
 0.5091309136116375,0.4477588542928833,0.1342468302163755,0.887898445485892,
 0.9609139732866876,0.9981968450607817,0.2596765875647111,0.20827803261626454,
 0.5345515411346409,0.4106964266501203,0.6430066262746659,
 0.049384311726558994,0.1616575709765229,0.5054986187440618,
 0.5268203647827376,0.43090951470304006,0.1305162237648524,
 0.46659696761689806,0.8044231799134252,0.38606143682389293,
 0.8069885824216889,0.5601795245370434,0.17997805552258883,0.9971485515737558,
 0.22720398879493908,0.07879874803376297,0.5553671061478667,
 0.2924481712668858,0.656379821371308,0.76588965341899,0.9610185466838561,
 0.16782275872680197,0.8866926038983238,0.7933387598209399,
 0.12277010452815285,0.6440957785498531,0.77386540785455,0.7167531096576764,
 0.32236771481507864,0.5446671607395484,0.08257189373450113,
 0.7190194353425675,0.22042501576981466,0.0556068831087857,0.9712468438030202,
 0.6648638702981768,0.471322611381919,0.5144426180634393,0.2500145833377677,
 0.6689242913847961,0.4485696832108469,0.8129314183333619,0.4217413425591068,
 0.4628543445538885,0.5175883196412757,0.5093438134880778,0.5872005753711976,
 0.14109643850929487,0.9186670576588348,0.2633533025603253,
 0.08781754191382396,0.36356631133254413,0.4422849908564407,
 0.26867916425646365,0.6278992867966966,0.73650539782793,0.38066910469880755,
 0.012181587674899741,0.728591416012185,0.8947193487206839,0.581084902739826,
 0.9725479071591577,0.5987192867201256,0.9174485010629834,0.5470826060360928,
 0.744317592367331,0.9300043704188405,0.7821320621337084,0.9766491325242513,
 0.4949394939479984,0.5466360592801618,0.34016667911390674,0.8339713085736844,
 0.7678996356376945,0.9126738942585852,0.033949008513878165,
 0.9180850020789962,0.4068914313697656,0.9486215321680216,0.276310056747088,
 0.18878736590305334,0.47430229833841786,0.6259098066081163,
 0.8340364521610479,0.3118440302355634,0.28275072505529053,0.6789430803496426,
 0.513109753011379,0.6846667844008103,0.6210369134319739,0.8393374160226674,
 0.4471714248889711,0.5718754543182214,0.9893428509448653,0.8244657695669815,
 0.6774943921273398,0.7206009362866671,0.0776782326038854,0.3463260864178448,
 0.23788527571515095,0.8261641712151003,0.044355974237058726,
 0.6559076412051571,0.15591193266154102,0.8115480187312042,0.6888323203913396,
 0.6799939686174797,0.5621404788081819,0.4971743965354345,0.3013893460006092,
 0.4735872460581445,0.7862198391793096,0.40016878847773263,0.3411015559426068,
 0.10805162505696209,0.6705894836589394,0.6789419401679563,0.651524460232048,
 0.06440596319523295,0.017908820900858213,0.18398136781682795,
 0.030975171421294134,0.47296675764791907,0.6422514002745232,
 0.8953413957406625,0.5380227236992303,0.049469837985340215,
 0.9945013542838794,0.18233253229406088,0.6455878508675066,0.1465670050131357,
 0.24299494147027656,0.7033989047655821,0.050038391457084996,
 0.10539432736559118,0.429269864879003,0.23134388582653087,0.9762901873824129,
 0.9561388935941268,0.6495751367937397,0.06983490877056153,
 0.062268863630622895,0.14463132342545648,0.1795118922182508,
 0.6034093250735039,0.840136577321283,0.05048155192002146,0.13440599391008023,
 0.10504390680943998,0.09768510488174198,0.578278434354675,0.9046111215203865,
 0.6762529555090726,0.7329276269983412,0.36012221035059744,
 0.36770153716105014,0.8002548178791007,0.7660275154547536,
 0.29407609477278307,0.4381845708552595,0.38775626819485154,
 0.8491146333236768,0.879023454704221,0.5428522519114685,0.6128041708885634,
 0.6493117676272571,0.046181506213920254,0.03354905594702262,
 0.8365933907021539,0.6145322938434885,0.7193225998142546,0.5917225438838853,
 0.28805932919427146,0.5304070201147122,0.006901614295793945,
 0.40464877482869865,0.2824102692529503,0.4007777671831627,0.7582867466714006,
 0.618374998761522,0.9484027790744679,0.6051762297655816,0.1357225117473011,
 0.2702148229374616,0.04247180031150555,0.12106853095975312,
 0.2344720015941082,0.8705636160570798,0.8008936521779608,0.9764243023586939,
 0.9284101432537033,0.7391175685556213,0.8065827326759016,0.30071252593003917,
 0.02627509493565583,0.3520701530304424,0.5838506425633475,0.3153912643218748,
 0.042833203113040175,0.008333721468661626,0.36352264936205847,
 0.018156817629037203,0.6039535674013252,0.15938848514956983,
 0.35959732635256836,0.5482181488764667,0.8082674784677497,0.8893284413042813,
 0.9213352950859336,0.4477787218708864,0.8486981392167252,0.3302646989888415,
 0.32500950572223164,0.18823788045188616,0.4087185787500336,
 0.2437317973080908,0.7299140124307473,0.005878562186362957,
 0.11003411855556744,0.6788689078663472,0.6174334798367039,0.7627618535122761,
 0.09336294076640117,0.8224710290500926,0.10783525180857212,
 0.6619510080309533,0.8307025986627968,0.9955788242977077,0.7725316956890812,
 0.4289283318594773,0.2172395762665379,0.2131454484804577,0.20072672989753948,
 0.2857553445844784,0.3065983281316148,0.8466656978753198,0.43436744803581506,
 0.13238674176957266,0.019160989815410256,0.86197856209685,0.9478454941998358,
 0.02696365514933219,0.29717517246746394,0.15450029699572598,
 0.0035273837537757036,0.7499674581959141,0.6584255122801471,
 0.9668743189523953,0.8998879853901812,0.10653846967750225,0.6039768989670826,
 0.31231539961126464,0.5574955080397827,0.18211891889086607,
 0.9743115971326084,0.800065070384597,0.5726772560567076,0.9002675067470619,
 0.1592321194760441,0.30428243812459665,0.18581615135743834,0.619506870614745,
 0.4723230702575285,0.30895851583272627,0.39558826483326526,
 0.016077090658114246,0.36411144156428366,0.12701882247395813,
 0.5955195608835171,0.9362380931208413,0.15912402519003754,0.5154497881031871,
 0.8827742086336579,0.3932040598059323,0.9824503010185253,0.6586817819211107,
 0.12815144814205492,0.7948926668449913,0.297660664364387,0.05384915995226758,
 0.3309009720665934,0.26690048166969094,0.8122151979424506,0.7800453715828819,
 0.9835719652452839,0.7714710171982642,0.5170385694347001,0.38343452186784355,
 0.33793684531861046,0.7521688593316022,0.08219513755288088,
 0.22069051064729617,0.5203331499122335,0.6167858197936098,0.8511399156067255,
 0.8894413430208273,0.8641226721861296,0.6238225312042214,0.15345788503062752,
 0.6974313660689778,0.20075914177013365,0.1440566072446975,
 0.13758281853444854,0.8059171468645567,0.4534027844372158,
 0.005233136550215289,0.6594720665122951,0.6350557676710868,
 0.8333295912289991,0.4705785374821474,0.5867482274081685,0.09244271117469771,
 0.19766659419110977,0.9362664519515131,0.43130465781945737,
 0.8398849924310666,0.21595671592871724,0.36019210203720875,
 0.7078103043740799,0.6350888777116163,0.3074784238235636,0.9365131974039512,
 0.8704036219107693,0.04134729176567775,0.23493537910433027,
 0.7121926551583693,0.8293674115339777,0.08802172527811503,0.5646979209556767,
 0.7843470201025663,0.9849787107236722,0.7595180990931464,0.8551949041264787,
 0.4651610944122271,0.21803053347563917,0.8949734751799315,0.5491664649608644,
 0.04472007091720809,0.42548417946332107,0.5377510834066604,
 0.7930444098939923,0.4450984800124075,0.7890651438275866,0.1902498813065402,
 0.6736284614762964,0.9440364927378301,0.2832316096585621,0.6422427113526656,
 0.05287205080040358,0.410566000032753,0.5278169286125898,0.38683204743736943,
 0.9200952889040227,0.21589176176266622,0.8352545355590437,
 0.19880287011446707,0.7440450026261042,0.2836424226741908,0.685059851768147,
 0.6453858177061542,0.4335079325822727,0.17213818018591787,0.1325585508312499,
 0.5846165607529525,0.08225753784295442,0.5915374192966169,0.7747797535087115,
 0.5309900820576471,0.8814191074304807,0.8283741228749376,
 0.020018280403477973,0.8455454395654056,0.2836248934801038,
 0.15785567256282784,0.5037377396775123,0.5842407285780412,
 0.39232944571189543,0.6126976325525546,0.7600459595349235,
 0.24618860750558613,0.7749127453507028,0.7834139243948841,0.6655767213309021,
 0.9412412374745865,0.4164839961212984,0.3618171030094095,0.20238653862445444,
 0.4428287732268893,0.8712305321698095,0.6944103288619958,0.39914817856778273,
 0.3247508856590522,0.526607184538646,0.9031069927900275,0.35427227396471994,
 0.22346615104559064,0.5243882420016365,0.9459622290813682,0.9796434194321376,
 0.46528545660947607,0.12740298496254465,0.4322209126100073,
 0.3928932383431142,0.8434397930471338,0.4685614624160781,0.6359137741963988,
 0.10341218532377894,0.7626342914015887,0.23224578327643208,
 0.2764508293921747,0.2420673685639857,0.12998238438554632,
 0.14903057437397715,0.3891943936309883,0.6333808392169612,0.6604439599275473,
 0.4414034477647677,0.9764949692675,0.6713703741171231,0.6351196252287085,
 0.642896814246491,0.7734113152077537,0.2939255474798319,0.6409970385217251,
 0.1457518774849872,0.09847127415886492,0.6352615163649371,0.6272808446155513,
 0.10658605360050943,0.2383063695869595,0.256322877617448,0.80315964467609,
 0.8023902349311098,0.5585979921697447,0.27320606942864634,
 0.14335810587968134,0.4591090835788857,0.6241230066178958,
 0.36845134549367575,0.13646830691582623,0.05865609582134779,
 0.30765454479395926,0.1258435038278517,0.9216931465412785,0.3348180956503535,
 0.03447361855538866,0.03205651196955417,0.40104167201854934,
 0.4588665215506488,0.7510856592566406,0.33334283855486035,
 0.04785441964012371,0.7674668805605666,0.9908456178170091,0.188199804129485,
 0.765533779093178,0.19207528733901869,0.1477015627778343,0.6082512439871973,
 0.0500024448011368,0.5952813577410403,0.38695315616951487,
 0.07299316139465217,0.21308248271528007,0.18825414274992636,
 0.1349687597104171,0.8026662303153344,0.10562076778110352,0.615554921729879,
 0.209955054482875,0.352565130589496,0.8163696105431433,0.046389442080060794,
 0.35420161682620654,0.4457799619197863,0.8725099541749495,0.6987122231916787,
 0.315225308781213,0.6784247089583462,0.5466604663127688,0.7634122858030411,
 0.025736889555860953,0.8392656384323209,0.17168351728789688,
 0.33920156206366237,0.05911611584827392,0.700922972255823,0.5077568266036145,
 0.19135827312109754,0.004544904476295741,0.5307150581851621,
 0.474966067859655,0.9905861932095907,0.5202547520336088,0.7156042201371975,
 0.045434185751095924,0.42436063587150874,0.8738651101386847,
 0.8019978451267339,0.44618316556625637,0.8293384537884715,
 0.36369015102652913,0.3286420540542685,0.17617191638912122,
 0.18718152556719847,0.3777097830114655,0.9426770182414321,
 0.41665370244119093,0.15130859978358835,0.4492883821839597,
 0.7997092681087492,0.4026742597367584,0.014807758416617411,
 0.16640613777489877,0.684875647609945,0.24627669340650005,
 0.044789641746803444,0.9410254086463204,0.5584621603924892,
 0.7466501535992927,0.22089414804174667,0.006416702835915777,
 0.44514102925507126,0.6070049074145896,0.21665229489871374,
 0.7445595292375594,0.9082167273881465,0.1825726117927824,0.7880528053810975,
 0.6969543570522886,0.1284841550951744,0.3809453876166853,0.911566000688463,
 0.37375075396600155,0.7085464772146202,0.46835006844798077,
 0.7028014936035482,0.1695901090845191,0.3250623331178696,0.4363300915456353,
 0.8359101337510945,0.29807992269952865,0.524679206399694,0.7034059299452937,
 0.7375218843313301,0.34898903679219695,0.10950483457463567,
 0.07170338438175994,0.6473726391268844,0.42596103193930934,
 0.5445768525958167,0.705662858805137,0.6949740664627333,0.3466462228891818,
 0.09925364997699049,0.7728124049472122,0.2249684254587161,
 0.18149673034388147,0.15525144837980942,0.6627757135896412,
 0.6246048632950099,0.08432634041462439,0.028222228908431735,
 0.8571434812580462,0.41196859206497016,0.21020792237635932,
 0.11764854351839282,0.5779560903060064,0.9219995722235332,0.5936003548977463,
 0.5119364903926337,0.2846655996812598,0.6126211267796196,0.541015733385539,
 0.580646028906664,0.4212505721039218,0.2204561914282931,0.8021901627164343,
 0.6867247365053086,0.3433801352708823,0.5642880364409568,0.2489703565244955,
 0.38296900760948793,0.585312748268245,0.5722592558220541,0.6792946167602674,
 0.639635086324787,0.3142720208101766,0.5147962723846207,0.6659916237143664,
 0.6633152495167753,0.37082688746751546,0.15020454927618987,
 0.7072575769761402,0.34928460436601205,0.5356155385207236,0.7738312895443892,
 0.4778341270491052,0.5565445781841959,0.5088973793700259,0.44156561398594874,
 0.9921312392762227,0.6929114443471631,0.4681514171641651,0.07744032391605471,
 0.5174041469739017,0.2852907931309472,0.6353486342210135,0.4836742047125454,
 0.31972936467450164,0.03264924675017766,0.16025633148427498,
 0.06301800086379061,0.11289053646432201,0.8744287649079051,
 0.42861862387506355,0.8985396293114318,0.16194542004786605,
 0.46913262943101397,0.2323616473735528,0.08214689818730658,
 0.5411271256887387,0.6186749942023994,0.1998991300487376,0.6104848214295358,
 0.31005944031349975,0.37925508275767617,0.4591151690258636,
 0.3004611085138764,0.043279559761118636,0.47246689275608567,
 0.33387682552596354,0.5875860379905957,0.49879920997291094,
 0.6615852737310036,0.20546539681489384,0.821266078028231,0.4012732556363541,
 0.5818512475898104,0.4305227016172588,0.5793894061164548,0.5317337863393057,
 0.07816637312995489,0.6398669331076795,0.7495398827694694,0.6603523996338475,
 0.5614885923405977,0.5773601474135797,0.08295459622826595,0.1777162417226018,
 0.48477219737612065,0.6486837081530086,0.5383970774652961,0.4286273008914321,
 0.7846655141518291,0.19007021642685684,0.5601820559488744,0.802232745992106,
 0.15601837573503807,0.8790388048519023,0.4824305774605908,0.7373533266782271,
 0.2325065420232144,0.9920345087217942,0.4462315484705368,0.01478927475671954,
 0.6832606538938814,0.5942826744673304,0.7116646921308767,0.3173311465338602,
 0.40516569145965287,0.7218362225964116,0.40422387682852257,
 0.3531207649328869,0.8495498845140438,0.02624778839005182,
 0.027152714928190824,0.03399367769028949,0.5791310014324212,
 0.08390400575822876,0.4797201579987206,0.8330161787701463,0.6309822911883054,
 0.6962795119276726,0.6727540078777597,0.004422963758213605,
 0.8751799322970836,0.7392380913103008,0.546686354600339,0.2407154514290195,
 0.424110547657252,0.6519436313822524,0.3033962890087061,0.30176805222318537,
 0.09682758124971813,0.656512423437681,0.8547858433835905,
 0.052975673394288214,0.09975338602852679,0.24142297091524434,
 0.35621323099827273,0.42615540750276015,0.7799355818699254,
 0.9065181620272136,0.03857943448424883,0.10700554615146918,
 0.9720149032972536,0.6489603886615379,0.6942264353729797,0.8331766536660741,
 0.5286719784586328,0.5921545179029785,0.41692569382919253,0.500860212974287,
 0.9861314238939016,0.24550033915295844,0.9494930691473303,0.9565699744447689,
 0.5164987428821581,0.42355584750164454,0.07804821883827129,
 0.7569786062280552,0.38747935703582725,0.36836360072510566,
 0.25394386527227164,0.6538786327480339,0.14065697633782048,
 0.8911910509760345,0.003159640191673918,0.04009050184086893,
 0.5679014072710161,0.766899610884504,0.9307516003440894,0.5379309332308063,
 0.6568462594246796,0.27552718297827883,0.8001738198373655,0.5473046729823815,
 0.713894404974203,0.36092689469169303,0.9165620928291519,0.6313281783455046,
 0.7693279554328835,0.5376350847899786,0.3294412455789135,0.13111728372534948,
 0.5898187843095354,0.9335366287138323,0.6612912761138484,0.16304353971093732,
 0.7223486699906322,0.6203719700908195,0.15911708815119607,
 0.29397506809598917,0.44182324855332566,0.4421190429576659,
 0.5022641892149451,0.10811967088790864,0.2742825226406336,0.5751934490664883,
 0.08231702375863426,0.09528640231270791,0.014845839586749232,
 0.2956861831217953,0.3375179936880922,0.3557427813730174,0.3719532627349553,
 0.47065040305534356,0.13903441269782302,0.528763539883564,
 0.23264832714312345,0.3629298052320602,0.5661828608055169,0.8853666013273309,
 0.7024150949324378,0.44367328626083824,0.4293731621989475,
 0.45446565408893025,0.6476745411553647,0.6418782473109156,
 0.09710258860934362,0.888101646033782,0.4927052205055401,0.09511269136195277,
 0.5462592010262526,0.6391302301859928,0.05932901394841916,
 0.36218659856540825,0.6902295602412875,0.017283412454785485,
 0.44334475705766396,0.4313077332495866,0.012439063555290897,
 0.3963182611384517,0.48121268841926934,0.92968650150434,0.3569423360013789,
 0.6803753993274749,0.04397041170379434,0.08844660287586381,
 0.27233622429861826,0.7973762780669191,0.05267972170270108,
 0.8334864851776999,0.7279967136328909,0.50609518477317,0.4112869754819698,
 0.022789540962062205,0.6650734101175956,0.6584571396521951,
 0.8424933896540981,0.3768667774838539,0.6264551318086732,
 0.0013854319958133132,0.7872695697269585,0.16883983944386294,
 0.8105227280766993,0.23842818522774067,0.8572105962466645,0.9119206304557,
 0.06959823448015034,0.23918966288818555,0.09462877733536967,
 0.7847385542254652,0.8849690448234426,0.21805997177854297,
 0.07667730751849622,0.43404017855178667,0.039848912855120155,
 0.9167564035838052,0.029848936895377265,0.7733796433791763,
 0.34911102974020625,0.9742012540378864,0.7567493352565665,0.5299860730518009,
 0.14687529335139982,0.8583060830926134,0.6330380867911172,0.5378996928679196,
 0.32184344637094386,0.30464330279784213,0.04797450264609615,
 0.7321688027635326,0.7524342924407127,0.013996347653004393,
 0.7391819902653122,0.9508014722824364,0.17715449995650268,
 0.46628625684192626,0.21650742875870743,0.7322514368154524,
 0.8172050794440148,0.5553016831766632,0.6081780364911111,0.408338742508797,
 0.4345115767482568,0.09068362854646594,0.9082039360140459,0.183502385724163,
 0.3437360561122552,0.19045418855532814,0.6163302651228243,0.6998636026450168,
 0.5984580396077804,0.23645181823158068,0.7143879260310966,0.5559048056765521,
 0.5060593741446797,0.23765094370812956,0.5073913238103839,
 0.37463541965914393,0.4325620336265684,0.11115127474954123,
 0.8328551065642431,0.2999103600470734,0.1327283551903481,0.27580335026604197,
 0.7121806468426122,0.8798232220845649,0.7357220346432407,0.3517766310218504,
 0.9716963621307952,0.8293177040279318,0.1727204455364445,0.8061718442312514,
 0.37843384855632056,0.1440160513155826,0.9859624237910021,0.7948829605013584,
 0.27556646742069013,0.9741685097525725,0.5461590700233593,0.3176223703890608,
 0.5128912618087286,0.39406350258528045,0.9121486013258486,
 0.47503490073459076,0.03156709384237821,0.07024485336183695,
 0.15913375358352944,0.9670641562950577,0.8473488277818408,0.7315443107439084,
 0.1579642064870237,0.7278169666130441,0.07703636250810142,
 0.07418224099252624,0.8498849658650556,0.29509929865163453,0.926689735173398,
 0.6958959048290447,0.8413915917157676,0.7828631186542898,0.12583281588308526,
 0.8765064589282262,0.4618648141220427,0.507068531973864,0.2612111476177996,
 0.5505415295640559,0.03648172939774985,0.4125528439492897,0.8105801046559732,
 0.5300098749349568,0.9732106896485399,0.02581142880887055,0.8899749282365781,
 0.06692659585388228,0.06957763578216647,0.2144873664828978,
 0.19600760383600502,0.4082678696088161,0.3046889257521599,0.7026971255570518,
 0.11568663697663384,0.1522382045536188,0.21054866550151707,
 0.9175438052398212,0.9232153085758628,0.9882217056140101,0.22996070798457824,
 0.6671479069696893,0.7817474545402338,0.4827262316982053,0.8901864254454451,
 0.21804715631651672,0.316829921028549,0.2599688884182134,0.6821246411253384,
 0.7917489162713528,0.011377657884424242,0.10630498558398582,
 0.10971725539057198,0.3304694475588913,0.1975834902771334,0.6394480946020101,
 0.14823584802220224,0.28975933625858996,0.06861094223959163,
 0.6134342173121415,0.9700582942490636,0.24461315063104372,0.9927524057003141,
 0.5228846607635476,0.5068571765023416,0.7142586314198233,0.3918018604106126,
 0.453873367140157,0.03543390666864188,0.7500157246973463,0.6238215230337391,
 0.6593564522493023,0.6535910008302209,0.03642618763656047,0.8866003727210037,
 0.6911764598700167,0.2694895240814221,0.5031031232900854,0.2827090971459808,
 0.18579335359338744,0.630369648987977,0.8136460329683727,0.6670076966342809,
 0.8212569138935691,0.8086627210139642,0.28295691643792154,
 0.34097642591823785,0.862105522668496,0.1793321720072566,0.9534912827313989,
 0.7738566377368166,0.8474499388502456,0.9631092451454428,0.1215909116251298,
 0.7703494153766403,0.1634477951563923,0.4363110477283866,0.6501907004073504,
 0.5272470020877534,0.5845500699916529,0.9895895870028468,0.17875622014171855,
 0.7556869556786947,0.1566907508739548,0.9982205753244499,0.35478658918230976,
 0.08600842277836418,0.062384948810401686,0.6362601602416138,
 0.4740889889330724,0.4848612707940967,0.7019938379612879,0.6159711155440581,
 0.7065597106754002,0.9067618583116339,0.8517697483701281,0.5183235425808973,
 0.559717255528434,0.4493899505525686,0.4270611583587274,0.1382781657521962,
 0.8552682171405346,0.020121369612554063,0.3517785803951423,
 0.5232152025825627,0.4591346253294134,0.16873908937567228,
 0.01340657774714804,0.9469212570790041,0.9635465904954648,0.9226153410110872,
 0.907409916672064,0.49472539308884345,0.8954122293491317,0.9111453624712678,
 0.8209059803166591,0.7849730447308662,0.1745289428024116,0.49146563835895163,
 0.26868648163651954,0.7073170915313604,0.8372450226646575,0.9857943142941491,
 0.9654435373830993,0.8199286779860862,0.729357468480162,0.7490598877765665,
 0.8980465934175449,0.0847771460345691,0.05062017977057831,0.574750384562035,
 0.34165035303815317,0.21533201633554155,0.7149312106026017,
 0.8581645066265203,0.009118379498958884,0.7869794119017668,
 0.45393182667767473,0.9629908843986288,0.07626795060398728,0.334331935006148,
 0.168211044222338,0.026850717025544202,0.9582403660745822,
 0.26603369180482783,0.8991576964166201,0.06982050581516697,
 0.5958851109564895,0.62030432392385,0.7680881581998461,0.6010473814211806,
 0.8005083996379112,0.5181365776757867,0.5738449869201582,0.42519159402884055,
 0.8337750447045038,0.07265101697875198,0.5390201860993618,0.2372479399797558,
 0.06814290161582381,0.47902252300551174,0.5951720787119393,
 0.8669334514553838,0.8447495282335137,0.7493932798259824,0.11029263385724553,
 0.01631858108888684,0.7215318479907032,0.48237483284467775,
 0.21591351249584556,0.8302643538688462,0.24180614826997793,
 0.11056899608184378,0.5335627388234343,0.5984771474065405,0.5792004602665136,
 0.38312288734334754,0.6152767549783467,0.030704815355474202,
 0.08510833992031352,0.3704764372271486,0.9299835348549735,0.8302245985911665,
 0.04110669470494355,0.988434214178396,0.2819579878397247,0.17383055725730756,
 0.13233396629834038,0.36017528275981736,0.609004548082013,0.4004929804200781,
 0.18345995580844143,0.07693869225671701,0.8986162164361255,
 0.4148779571685126,0.3380521872513883,0.00845350574943482,
 0.08629940776681189,0.6015919350991508,0.46032436552177125,
 0.00613242299685357,0.0426946438858713,0.35583089396875833,
 0.6602577323606598,0.17849014850645195,0.3251469526006907,0.6793213142088179,
 0.2527579904398487,0.014728623304461785,0.1276751868936179,
 0.4972868310271188,0.41126523269902127,0.5876253989008101,0.7978651976905973,
 0.9534826630976734,0.4591467542431017,0.9134291728739993,0.800809805191268,
 0.4496093891650872,0.017064170144816315,0.6710671581028907,
 0.9776742770964413,0.36036768839432487,0.5626181057460976,
 0.27119364072125196,0.6039293708288022,0.5472333251979752,0.5715760469672496,
 0.929619911343177,0.8295018194538983,0.24922876163969143,0.4334872440511117,
 0.6020002037657014,0.17603828647340147,0.5474579168138106,0.653480509934265,
 0.577810966173831,0.31793041832371605,0.15175330538314857,0.6930284347181164,
 7.548308888695177e-4,0.0528176436651242,0.06500187151366343,
 0.8958469828756179,0.09340978522569543,0.3412044089743843,0.3546440891624705,
 0.7602315782462572,0.4052999760692333,0.8279566284295186,0.639424942230703,
 0.2276849255270572,0.46004373751848304,0.6680529214782769,0.691763453464843,
 0.5003151949563647,0.1380342072224029,0.41068357351511753,0.4047632400225599,
 0.31958834914379064,0.5623109144950328,0.7593824381273068,
 0.014320064666008214,0.14093648665768332,0.5994048379345237,
 0.7805175804760379,0.9509188706924471,0.7448883196509339,0.7207015780686271,
 0.8182622100226045,0.8173597942242525,0.6602835756357341,0.41492404080095935,
 0.3900101485319297,0.33192588078402707,0.07515833162640195,
 0.4669679027444047,0.8926945020298872,0.06621216653613404,0.6631789663007999,
 0.6038877687138592,0.4153221528069452,0.14939976985263104,0.4786297684177956,
 0.670042449676275,0.012720218269252181,0.2597969716750005,0.8171362290048092,
 0.13160194637951927,0.8820011568475972,0.16614413971607755,
 0.9147825633640146,0.1762930138039347,0.1177127447656734,0.7306173072034228,
 0.6238593818187462,0.6140381009966654,0.10098691991810493,
 0.06359748749070171,0.6347378570761213,0.1060814486724394,0.5584414614384393,
 0.9573714627688814,0.6807625000208253,0.809569355363671,0.48190893256297684,
 0.13183998752334414,0.6510658326433008,0.8150983698799885,
 0.11775733033104818,0.5022400324719049,0.17913929033814413,
 0.0340870813913039,0.986463888744825,0.22958780126364275,0.06882512100062133,
 0.04799934526281324,0.09133152098025688,0.7092355329016157,
 0.8584371821213016,0.1336765677477425,0.11531443655481355,0.8396497604251197,
 0.34623997936283335,0.1033629754736004,0.2171647820198681,0.8180076103791617,
 0.34613205805418856,0.13463612098049493,0.6766630502518037,
 0.3463057540987672,0.02256025512143711,0.6929375579953029,0.8111532567884796,
 0.629982750518945,0.11675026189370863,0.8673199614705063,0.2676371083160781,
 0.8705611326943661,0.9884545293397918,0.8831531420664613,0.8241803942986762,
 0.28331331891839673,0.3243599762023076,0.02183617121474457,
 0.7256822491520116,0.30819184111170184,0.09666267269859619,
 0.1726979281369947,0.445335118402418,0.6009212777412335,0.914801440847522,
 0.9955877459546824,0.1519615377957746,0.19010439387142253,0.9438626966703365,
 0.33237477476854416,0.8493386593554078,0.825247911821837,0.7681080795823497,
 0.015896785783860667,0.3364400922002606,0.1361234044913764,
 0.48098871484597383,0.3008708917621983,0.40551141471136565,
 0.5107525770507845,0.45823898231715976,0.7916101275683076,
 0.12238629411029334,0.21429173946982394,0.3159668478160038,
 0.3673708639343811,0.32943178989326927,0.41476446822266855,
 0.23103980262337342,0.11337456873436624,0.512118546899272,0.2518845453303836,
 0.40991569408759143,0.09859086065259792,0.0765186158890725,
 0.46664530858121445,0.2378131504164126,0.5299969304122781,0.8593724714927702,
 0.29057531445555773,0.19860935566746174,0.2656964000880945,
 0.8655413676270868,0.9949307646828719,0.19392773785494977,0.5629049768231945,
 0.7655939011002507,0.29576934479916184,0.8017246083192219,0.783159458943595,
 0.34053239913805655,0.6037377750011749,0.8687395248083541,0.5782784606567795,
 0.2509895614680132,0.5670959070103766,0.4637797999796649,0.9094850161376652,
 0.3710098792552712,0.42643764689738806,0.22284086488075472,
 0.9150491016491726,0.6507287778612594,0.482457189324598,0.4214540420912084,
 0.7433402955205476,0.8404739826927872,0.3107687950582134,0.7307981882110786,
 0.7261449521241888,0.45855685954767844,0.9840594312001767,0.3731993971103378,
 0.3181777008709895,0.4724813125741498,0.9342002824800827,0.06258687592172618,
 0.10797317240304938,0.5342311930004502,0.4782275778998819,
 0.09247990724258881,0.6715807245701064,0.8725019368809288,
 0.07103286043253243,0.3848678562208123,0.08706210821656046,
 0.9007908053144135,0.7187302734617329,0.15846699369174755,
 0.33157133159930163,0.879305198594508,0.33787145468459934,0.8398464939741428,
 0.8449306305947069,0.8271396767813602,0.7603287274696549,0.8656419756841333,
 0.7911968779432077,0.37623696178104027,0.06585915457437408,
 0.9448765262326917,0.7021715607435327,0.026127063263214234,
 0.6315469327846946,0.42147136441372746,0.32237748788104803,
 0.04500401327465775,0.199269689862135,0.1888005796128398,0.8221805474747141,
 0.3265790205135386,0.592205017702695,0.9553868271851402,0.7970314574511665,
 0.6384537423828129,0.3192973828436897,0.18277464971913837,0.6652515879170364,
 0.7035430747281524,0.4000200410394231,0.6075177298070312,0.9505119834851606,
 0.7191482477828258,0.3886720086103823,0.7525792113824535,0.5288135535063112,
 0.9202751994100359,0.8383783441442536,0.7778487617405939,0.49100755816684216,
 0.036381048487413525,0.18256735314904549,0.6207261381774817,
 0.4268163698306786,0.7870167556027,0.9951330601019172,0.8833079768495169,
 0.0587637750865192,0.8063184309173044,0.099246002037647,0.8578735793833417,
 0.15029318856871565,0.2756528728514015,0.8010017519260846,0.635402515550209,
 0.20762921628807374,0.5178644726500654,0.6353880655404772,
 0.16589661239679654,0.11976145510949276,0.207672327315932,0.7128750941189914,
 0.2704280880025225,0.8958641171565358,0.003900331811293123,
 0.2539039270489315,0.440719479232886,0.65046994312146,0.17057539298397617,
 0.13357989736849052,0.46754800772962446,0.7522592857124543,
 0.23104191099619342,0.6069831482419045,0.14809227277512038,
 0.18554661910132597,0.2557174520663217,0.06273869975273572,
 0.8923317138004987,0.26705031924033873,0.01263017838521252,
 0.18088123532738054,0.5997694893142163,0.3940121836802827,0.8176179261405871,
 0.2281421668921284,0.741170695356944,0.8121370958178082,0.3153492505199095,
 0.00568522509977476,0.011641823711180166,0.10693531346031082,
 0.22608954866093,0.4664122842758621,0.21219942423925064,0.33660436347072,
 0.9873257829893474,0.10697865176383148,0.7647590873264332,0.2171343274034485,
 0.4070398050089453,0.6624003727400416,0.7514995857421506,0.7704795689621561,
 0.7744103472566259,0.23565826523918587,0.8377228156633438,0.5506382650755149,
 0.492987889584923,0.7996085624199499,0.617753012907999,0.9207425731552804,
 0.6890791919886067,0.601252604600309,0.624992596426777,0.20353150106047158,
 0.04998791266050506,0.9629535134330874,0.8129536786709604,
 0.22028953792963846,0.49343636612810304,0.913054918832459,
 0.49227737711399056,0.5184796430284767,0.5280111310044021,
 0.09353871065377961,0.46847885432406156,0.4213941636095089,
 0.9025397349432849,0.08930402316869768,0.7339574778578077,
 0.16867860532645118,0.9282882139346902,0.8084998847729759,
 0.38355709307086483,0.14685780906053303,0.6022544731927104,
 0.3849634081371278,0.0966174522134221,0.9525824965604702,0.8554896523593815,
 0.5140320780208528,0.40080392196459513,0.15229426607093932,
 0.22751208834696768,0.013112641642359035,0.20263771704134126,
 0.17347114996778357,0.10506705622603929,0.6196881097732746,
 0.9694522185272106,0.9764388556614754,0.30029098873445836,0.8819989532542831,
 0.3199771479458582,0.2655612452961926,0.24986834816903558,0.8926688128813449,
 0.7868561540754356,0.46442914646278566,0.8520610316985537,0.5564232000795535,
 0.2898215339271939,0.305933406661285,0.5686365585608092,0.5327909866850011,
 0.27873769040041885,0.8346887247915543,0.06769245971307003,
 0.12229303975868633,0.4872608439265279,0.9162376161088651,
 0.10909683889743682,0.8767212075538591,0.02730509146464899,
 0.8521097882979679,0.699043960516831,0.8229798758422773,0.33136714253756916,
 0.7301615480996639,0.58664595088534,0.640651860375545,0.15945499957269338,
 0.288264488247688,0.8107214551446764,0.9686598617103975,0.10802213086529477,
 0.27586008644304016,0.6725308712707738,0.3264266805766308,0.757057789198879,
 0.4128430783266098,0.25619291078173567,0.9549645213230324,0.7505699770735617,
 0.3620345237537904,0.9422309777467893,0.690617102667247,0.5177164874283708,
 0.4589466796774416,0.7552746256204659,0.5184300600747971,0.44245187384765294,
 0.04968562638632479,0.8964195295195647,0.4559904513450296,0.7491686861315001,
 0.5742680521931237,0.6108122858976605,0.9311303161400732,0.9908118577854241,
 0.42682133797190225,0.367099343982926,0.6249979052742701,0.363481238185553,
 0.8078899163656975,0.6678121096308273,0.7661884300197802,0.04613302295060773,
 0.5175502559109544,0.05901694210816444,0.18445938118656935,
 0.9342166747167605,0.41334164297321396,0.7011247006916997,0.1466158627616908,
 0.3575157405344189,0.8692161014394061,0.8438111397710889,0.1557489698664224,
 0.2661691642417967,0.2515113857667016,0.13370066685815307,
 0.37293356886081386,0.9706981999732529,0.056281685542048376,
 0.031182344366421755,0.6634607198529063,0.7687139666788593,
 0.12875014967842824,0.9927837315825836,0.45291315869043536,
 0.9947348344806183,0.521326828696547,0.38992582621941063,0.17598804635095844,
 0.8323693201592344,0.859691208189016,0.00908952066074531,0.22851048147711828,
 0.8358668806548505,0.12260711281422898,0.9918457092040887,
 0.30033999970970615,0.27517062663965586,0.3777532842811866,
 0.11282249606915629,0.3936357799551671,0.9308087485361034,0.2724912307893331,
 0.2720121504963722,0.7510872225988524,0.05607477707869957,0.688410264647348,
 0.02813603209867399,0.8007051768319484,0.7890660418105482,
 0.28152537398741395,0.4132859280618235,0.6771945596352127,0.2021181136017116,
 0.9584180000108358,0.4004848939123533,0.32623346469872183,
 0.25589800144896613,0.0977596967101878,0.023168313569012833,
 0.5592118181535854,0.6252900928360474,0.31751248663456744,0.9127308192844437,
 0.9614414044747419,0.6753714011208612,0.6450325065096916,0.7688377260169756,
 0.7556776502780765,0.6439874490500174,0.9072842672369062,0.36431273867228653,
 0.40262318303145184,0.0894396371283197,0.8848512968261527,
 0.005106698748553962,0.5779691053644105,0.32127215032260925,
 0.24863404694564684,0.8839166103823465,0.9108471267022235,
 0.08276213873410154,0.025026836284758436,0.7595860939124444,
 0.2198619291689421,0.694248006728563,0.6218714055605458,0.5724273082152332,
 0.03689767071844363,0.8677531299706026,0.7586349013756623,
 0.04443885525339586,0.6201545844501835,0.9136836984147931,0.9770511304717926,
 0.5694989017986826,0.4710752097884344,0.8463924680302729,0.36460085592676217,
 0.5543840768549495,0.12287870626673103,0.059187534235061845,
 0.10904254380821643,0.35632548253211715,0.41786830672616637,
 0.5454677337231874,0.7056129395143759,0.35138900279489127,
 0.19424522820544887,0.9851937493878352,0.9900733188946867,0.3106298888844232,
 0.737574271575209,0.6819929377217055,0.5958789487445502,0.43938216572679245,
 0.08299500416897199,0.8084583056696062,0.2361028452610724,0.5625667903198207,
 0.8791395685528123,0.43135437975385327,0.7208213326439755,0.123886715759154,
 0.4365671929748981,0.8040298109153229,0.3177335861656144,0.04903945741812199,
 0.3807077450105907,0.7685070327649359,0.9342333778109898,0.8465662285921669,
 0.5698004567009948,0.3448557957882069,0.514614912383963,0.15048955172410483,
 0.45515912332161645,0.3892348948571903,0.6083002521189015,0.6161596156387953,
 0.911618325358714,0.15015185774119422,0.3149737318658925,0.4206513675160528,
 0.5837642932635162,0.9791322860595211,0.16963641487093384,0.9723975205376884,
 0.5925296915959516,0.5589922442010987,0.8281395545594488,0.3794266893227106,
 0.9330394915365622,0.32080933007139834,0.05446104361069737,
 0.3519847336636499,0.1829540272237603,0.29244992685783067,0.3381229453892701,
 0.2338250843957267,0.2315934693766346,0.6110965377094149,0.8679062431450948,
 0.6164745337682467,0.8318211202711351,0.6071262714698624,0.428910606348721,
 0.8898175317647182,0.8404730696097665,0.24996501768127444,
 0.061560994488890186,0.21764159082970003,0.42378942731933733,
 0.006245502566252226,0.6517163599773577,0.8381336817220968,0.688298289906618,
 0.48768290027203387,0.1274573280219472,0.057780285812687016,
 0.5996872235843692,0.24703446678336405,0.13524315007248988,
 0.13586576750106638,0.34617197383264076,0.5716760250334351,0.466676550960332,
 0.87738716387453,0.3520661491109224,0.027297935366315085,0.9652477374120676,
 0.9093914286641462,0.04621004431732889,0.274802449866296,0.6569656137769206,
 0.47900224053583396,0.7719237820118623,0.1507415869155001,
 0.22937306557883363,0.948998908300287,0.21762856518939877,
 0.21003726393183308,0.6660318920612048,0.3770117671037231,
 0.04844787950576124,0.7597249830937239,0.48688393802044283,
 0.5937663314222545,0.3065454789851193,0.6573190106400139,0.6978433585621223,
 0.20357661902689927,0.36468925151565135,0.08032084248279392,
 0.7034082598326988,0.49429458263736625,0.1672375109013733,0.5451836110546261,
 0.8538641291884749,0.7644658122323524,0.23875000742781416,0.8008724016679579,
 0.8512519209990176,0.3369898479727651,0.8320332098138207,0.3012734198495305,
 0.07288374250399388,0.7430064240861136,0.5194069700485582,0.4546804046146219,
 0.3250773945582448,0.171521678917403,0.6908052267050948,0.23358755940459808,
 0.17188833691140226,0.2941604649520483,0.18695505660349399,
 0.7227959658308312,0.1843291732988841,0.49070579653546587,
 0.23349240283315975,0.4592185052714759,0.6150952572615741,
 0.12215604390671086,0.9477151899610923,0.4685162113683927,0.5395843891527156,
 0.5350275797546631,0.01988230546032055,0.9075185544953276,0.9638335155233863,
 0.7985604244000969,0.4346109075922606,0.953842402492042,0.32820255077290084,
 0.21937809885179038,0.666758519340449,0.3600245666324222,0.3774775565034586,
 0.6652308793232302,0.07668214084640479,0.6818977412135288,
 0.13900321500958102,0.7537997551611447,0.15347700588798452,
 0.5376289731670155,0.45670339095540435,0.5326605858784071,
 0.16769432059867329,0.5218894480505736,0.14094647273569816,
 0.4253978263031156,0.19187961382742502,0.6437329422049184,
 0.07580737190054898,0.28780707910954106,0.1438237735584088,
 0.7257346419343335,0.7508705086670233,0.9183619102328869,0.10503033791040395,
 0.655402940437432,0.7810420712077194,0.33077768752931813,0.7458287776431691,
 0.9914699693133525,0.44924479376180615,0.627170880954177,0.0871276152767465,
 0.372221247008787,0.8941379465845579,0.2648045408507781,0.3789301570828014,
 0.0820342433093122,0.14697072243826736,0.5922728108396379,0.8708394632611398,
 0.2782080452020612,0.4082282668372422,0.7469720344645863,0.13070164473065793,
 0.5017287732425164,0.09741364205344383,0.9661783327069201,0.2899855021701536,
 0.8842695562927569,0.09166567478599819,0.6909783335406564,0.9915677411477232,
 0.7205499721054753,0.27044637677605843,0.693006421889264,
 0.033576081847465855,0.9208496781479077,0.14291511015279468,
 0.800570953013457,0.5830353090536075,0.9210923821049636,0.15617788610878724,
 0.09520400273121865,0.29400761154097876,0.6711027571747363,
 0.4394165434864218,0.08837054166200253,0.9938858792816349,0.718166613463533,
 0.5711078799979463,0.6025584616435097,0.5289223864451137,0.27023946422930367,
 0.467506224101079,0.1979611711684066,0.3496020905404996,0.17207177433625231,
 0.40251512426737224,0.6791371822167268,0.9649885109081484,0.5928305061722303,
 0.3021809485178165,0.5723602386974211,0.24063295965801557,
 0.32718343747432543,0.6391244646653339,0.553926080447516,
 0.010274828209582476,0.9725803881879999,0.4427089266232809,
 0.25732407410597724,0.28857722755225057,0.5449274532402107,
 0.8792741943037435,0.08947869953981269,0.8430902934843988,
 0.25881543698377074,0.7473100755454667,0.22214463346998248,
 0.4049294395066291,0.39852450877125944,0.5912422109450505,
 0.023703773288660135,0.650418833894374,0.10832839224124402,
 0.9429308781495536,0.47602646265256265,0.9882747728027496,
 0.11246540987874709,0.9743573581595305,0.07459188784078252,0.938928937138609,
 0.6748441185648311,0.07800185152631978,0.9078327925001356,0.7357394870904581,
 0.057324948725224645,0.4861983520574533,0.847127594632666,0.5831628088721098,
 0.5300343812820938,0.5542706581723122,0.5852551221957977,0.5215909779767399,
 0.6784029629289102,0.8022235011942688,0.697536494739248,0.3110887637287959,
 0.3222736004853888,0.24202805339716893,0.6310554182388491,0.3322525643853589,
 0.254916456694377,0.6656201471072524,0.17012047698269872,0.26743096014057866,
 0.9800531233455536,0.6695962153984274,0.42620796901144686,
 0.42230930998776284,0.294279219830895,0.19482377524986283,
 0.40277976032524565,1.3453052589218473e-4,0.3050962230740276,
 0.8328465511271412,0.892806463076063,0.4405653650596508,0.8452971181842239,
 0.7820937430292454,0.9467046621967907,0.9976638448550221,0.5277372243037952,
 0.41134075252818236,0.9072762285047605,0.5771253472848146,0.925843832070872,
 0.6138012915205446,0.09111698388397604,0.1580803306038352,0.9076796791104591,
 0.7774225143833475,0.11914605938613543,0.3695627961197574,0.8712476023991277,
 0.7577620304293932,0.8170537752906235,0.49819587581045854,
 0.07583377667716462,0.40855341551831137,0.9107190073544784,
 0.6579592680421018,0.47339901872849866,0.9025288484141099,
 0.02332450433827138,0.0721248198161043,0.6553944863089817,0.0711896001892981,
 0.39285372805153773,0.4413397585909038,0.5429080805487743,0.6186417579081981,
 0.4130938922360018,0.36507665470357675,0.1334430174632778,0.7928271039425745,
 0.5695160886247564,0.7593986143259323,0.7518719870569988,0.8963211016381247,
 0.11929591851931154,0.7267695350205475,0.3785034905836766,0.8649250859622749,
 0.9506824041264553,0.03386506051689253,0.018056224722576286,
 0.8332286805593553,0.699459564212906,0.09699985296605151,0.14647132502163906,
 0.946468503684822,0.6699807319223883,0.10689901388266809,0.7712274079099386,
 0.46252237715940425,0.23487810928791986,0.0011890027439683815,
 0.3073523415924817,0.8231783612131363,0.4942809906310135,0.3285787016851226,
 0.9578828185778533,0.924808983853706,0.07034178059302598,0.23150087893814464,
 0.9181044835433165,0.8410830690815353,0.747523409306134,0.3184344019520632,
 0.6288608036402881,0.9959483032113344,0.8152349819248963,0.8793809430084919,
 0.08720944337925152,0.4870972692761868,0.0304114500827819,0.8676791293741677,
 0.6928088422538475,0.8306545775031919,0.9782683902762946,0.8498184298469637,
 0.25055013148293537,0.3826094527845676,0.94727019021941,0.6901977128393394,
 0.7895539513976855,0.21875650734025887,0.9184825232793337,
 0.33103802139728644,0.6687390447491308,0.5524164135243544,0.6251023620541206,
 0.0026946531957107167,0.05729430241237976,0.46998887954772717,
 0.7146171675281934,0.9767809344139193,0.12396632157529519,0.8630275250122882,
 0.9920610380009088,0.2956799398505223,0.6428889291300751,0.04571641159352269,
 0.34726103935119856,0.837403775691868,0.8714476642393221,0.393488620632895,
 0.45815019650986166,0.6364564019853627,0.11958007680740279,
 0.5701696788611638,0.8304179806748516,0.19115970283509798,0.610702491686184,
 0.4033683937491772,0.7153267210569327,0.7559042417232252,0.7363132679903963,
 0.3463045060153628,0.6052445212693192,0.30887743678501944,0.9228453309741002,
 0.4897931390266741,0.5124145654934951,0.7415745124496588,0.8192698888091595,
 0.5715699623737343,0.9406032570002401,0.3803041905857133,0.49564099720934807,
 0.7305670958980467,0.6801314185828334,0.7720196943623171,0.48151145037273424,
 0.5642729857630053,0.6376905686868684,0.01660645140624728,
 0.23017526718558257,0.41532330041710086,0.7455715405859201,
 0.23344827712881455,0.25060241774577807,0.10238638289254298,
 0.3197918537281583,0.8275643693975238,0.07561332722960556,0.4994980500295665,
 0.26320732808588476,0.9166345572954556,0.3234737870859179,0.2474672165099283,
 0.2820504209453881,0.2507041710919875,0.6294528534531519,0.32453631522258775,
 0.14318154929937987,0.7905872750942528,0.9825029290283724,0.5935005697662964,
 0.04535138661106164,0.9382838968100963,0.8591999588467909,
 0.46670631477191793,0.8004987725760891,0.9468375206528141,0.7840430194558925,
 0.13248899532003455,0.002548970889465263,0.39216257279440414,
 0.972066081823063,0.5363990400274551,0.14650990178105738,0.995458916659786,
 0.8477254436875137,0.7565183622881577,0.8636836373385192,0.5378441179956653,
 0.46789653631993655,0.7282158246443895,0.12984144654313834,
 0.9450643079438075,0.5928988719232187,0.33368233935477987,0.2539207722432646,
 0.6520944336306509,0.396657930121153,0.38255079965251526,0.10624090882045178,
 0.511590744426279,0.6410063241032903,0.25779360329054746,
 0.003213495345870565,0.41225940360013724,0.3774894655011858,
 0.7573792936545612,0.6246783215870354,0.6620481169746002,0.6474699202182858,
 0.2945125397958859,0.8660538163919445,0.7398956332579995,0.9798453386491008,
 0.7369162334554714,0.045416108935406596,0.4083125919877689,
 0.5484255981252771,0.5272932561183042,0.6634553690130485,0.05411138327304754,
 0.5261601272174203,0.22164257310526514,0.17194935481892815,
 0.9016379775012753,0.8908819053844154,0.8795451504333245,
 0.026885550155263283,0.6641855034859969,0.8046263602578052,
 0.7563511329803866,0.3707522461065744,0.3344574104633158,
 0.006923060642408885,0.47899694514467506,0.09910990363003547,
 0.06516929552799378,0.237923679998576,0.1308077829497526,0.7933002449631735,
 0.7379483237004882,0.014564190022376633,0.3802401437825206,
 0.4408671915234108,0.1443154820615392,0.796757574427351,0.4851997988031712,
 0.47502308319213793,0.5320735955758115,0.9149864056416714,0.5241488955311898,
 0.41004846898171354,0.9142754222307952,0.33978907979183914,
 0.27205584979443054,0.06423424363179253,0.18344905742131778,
 0.024592248972329367,0.17555476156937722,0.743185641516468,
 0.36861738451168724,0.19415116620254824,0.12414104584421515,
 0.3670846842480767,0.45213816054635236,0.7504500853081222,
 0.13921508889562972,0.5081591139039529,0.792392553287839,0.2711888637366694,
 0.4828213333027974,0.9487526405000065,0.7311107968496627,0.9703251539144747,
 0.9874075517809163,0.9873957953153947,0.451552575732143,0.36779756000345654,
 0.7298967563999224,0.8583381863720351,0.5872687532032068,0.5576090922608509,
 0.5861937597928591,0.1457546655363764,0.5549047157193048,0.7981969705642555,
 0.035604949510084194,0.12334618443011836,0.7370211913415847,
 0.23922762505556971,0.37375581393726076,0.9206088851107312,0.894368037189501,
 0.12898263700356727,0.9325837302003058,0.07071369418984397,
 0.21823539884045606,0.9684985018253798,0.7218526418496486,0.8824656379486473,
 0.3979650431760736,0.6269891621962372,0.16734891190528967,0.8176622290227574,
 0.5544511760270835,0.7868177082702463,0.6875686226313529,0.9713715353172626,
 0.6342226415629254,0.40138629806661275,0.3977436151889151,0.8617831106232391,
 0.09276624538541,0.048151177661237665,0.8534303963006667,0.7911331661527801,
 0.8044555486371074,0.16524968855819233,0.9517434020633452,0.6596410620381017,
 0.3223118077348007,0.07147846403282299,0.44166978560457604,
 0.8367176728501382,0.23307074756254653,0.049400853637520004,
 0.3005015242420941,0.34616166927150915,0.7551676652537329,0.7057187994604834,
 0.642990276572109,0.5519723477017263,0.8199064626118379,0.6335763418597655,
 0.8863224824201701,0.026331541771739353,0.41612581250066616,
 0.24424785013459993,0.8279553547761556,0.32607183891791125,
 0.2853176251704278,0.642713622428637,0.25038749902436075,0.0599333418089113,
 0.6776297142798194,0.8270347906008115,0.8325888847640095,0.7536113098381186,
 0.6506002266719086,0.17963609983453566,0.6454169672828194,
 0.30822115212630474,0.16039038379597859,0.4647723879891962,
 0.8503965074797318,0.6066592808094959,0.38898873556755553,0.977420840715076,
 0.17547776346155053,0.48530592042454757,0.9460057293610923,
 0.5303729500359093,0.9171311069892161,0.7418216615491844,0.12729949851731892,
 0.5464034377581117,0.6689854348798739,0.0020955814682283025,
 0.6928630558557871,0.3500765345157032,0.7013720578019865,0.8960547167503747,
 0.7398085084168301,0.37387746367385744,0.19647409292256923,
 0.08955055740899331,0.5768571302578008,0.2804401555189777,0.3407099329839244,
 0.632484201635517,0.2095437871325554,0.3683800939466033,0.28316730543731605,
 0.28774969405765005,0.1031810384438755,0.4136077307129694,0.6387268183575867,
 0.3324948161353478,0.8273384414328779,0.16793862867226372,0.4008757760010283,
 0.04294271551833917,0.05122133664527051,0.1667377655212373,
 0.08920861290150739,0.09625272508162941,0.13058876583147228,
 0.47110608707592994,0.825965314295096,0.602457529293469,0.3839049848932674,
 0.5266897754397067,0.06230297501767956,0.5560226282846152,0.820454037415403,
 0.7788350714011182,0.9301247342642409,0.19545174461034787,
 0.29726137221220905,0.5678113641664462,0.5164971110632302,0.5836944389998573,
 0.27462045022730674,0.11858127103175009,0.7168999999817444,
 0.34475108395509213,0.5100892164453915,0.5008670057726378,0.6660614333300634,
 0.7309891816231913,0.18513339014713903,0.16075990991462286,
 0.6451793170661216,0.45267968095595335,0.8419047235945958,
 0.16255363199198858,0.32386005115121763,0.023169359378422127,
 0.4282251557036125,0.543310242005015,0.07696887001921482,0.5652384483320942,
 0.37959591533716175,0.6654341849716884,0.6938742035362915,0.6552385539928518,
 0.825042434072596,0.7728826702293959,0.4100437243749723,0.12254077499769245,
 0.7474059712315174,0.9999652804011891,0.996353247571945,0.1154773988245591,
 0.7415139870532168,0.9437406089473444,0.8286826789812599,0.208428181466492,
 0.21777674075718534,0.1037723933027277,0.6971866533291333,
 0.11917925542983032,0.29633539026758604,0.7707319862852464,
 0.4344394544839407,0.16790425733541037,0.10851987399178487,
 0.8205674426529409,0.03400332110056381,0.30998961479091003,
 0.8072687406555186,0.491116632517113,0.8988333049541448,0.46618815656872337,
 0.15016367373945916,0.1324673636233632,0.082586669140163,0.32313183342899876,
 0.9887628693718443,0.3816426032169733,0.5542168313482416,0.9242840841998804,
 0.22253589076324765,0.7681568544997563,0.43418353425543366,
 0.06937517361610401,0.18927079805002145,0.4138174008558826,
 0.07037822153791407,0.4580134165375648,0.21768803981884155,
 0.5105776576179759,0.7655711946306886,0.891996584225265,0.9994135186879514,
 0.8803196236977785,0.47319687677867595,0.3440087100611513,
 0.35702867754686374,0.5067411601304657,0.04323828855702849,
 0.4312506945551653,0.37637362282762044,0.8245926248637571,0.5502031558905357,
 0.917050965853574,0.5548191985596628,0.8588109878155277,0.18648590166969414,
 0.6302005697770521,0.5318914305158519,0.7488299276992738,0.2642251161715786,
 0.5298613752421505,0.46274003751113457,0.5582074399797802,
 0.42603210179643825,0.32701893355495393,0.03443955618537886,
 0.08399780054892647,0.18584895078828867,0.7061607472978586,
 0.23563559319259098,0.25867588267680997,0.5509836502988149,
 0.2052146642836099,0.6321326507213336,0.4828233265141314,0.22502209980091237,
 0.7916578253516353,0.10851564428147276,0.007210451793125627,
 0.8809600175359685,0.8626700573846426,0.11896026452199471,
 0.09523776143256446,0.6096285368707587,0.9575142860514647,
 0.07049598155691394,0.5542042083892026,0.8267253117724483,
 0.27221456720289805,0.0037462135286423504,0.014828869386713905,
 0.8108160753034245,0.5602566990657265,0.8271249703579945,0.7244142193807939,
 0.6593064095785638,0.42343169632811506,0.13291212467212932,
 0.8886388689690118,0.6546103403520659,0.6907015316356124,0.7404888836751503,
 0.1260400182543917,0.7374114827697834,0.9743486656518869,0.31647056973833965,
 0.154278125281207,0.40136814134413046,0.8255807762270492,0.18980148347478132,
 5.690244786761278e-4,0.9056100280605597,0.2778724591248879,
 0.7828685716599314,0.9626458777829006,0.45189707230435094,0.6675042262855172,
 0.48076008223179945,0.46502703670266154,0.6926294263883734,0.557106429799767,
 0.0473217982480606,0.37716722826662485,0.8488644209341691,0.8923186433799823,
 0.5135206419155077,0.2026205453247647,0.36384032216929685,0.945497061238824,
 0.44120886741974386,0.5804600516187358,0.5580063294630917,
 1.918088636096993e-4,0.9700708871334803,0.9395368192640314,
 0.9339729403999982,0.10275500342686217,0.044309344984091226,
 0.6946619114220793,0.12226239331680255,0.2633720817933328,0.970861974117785,
 0.29067554045092825,0.4072634709208848,0.28736312076823367,
 0.7126754568546604,0.645882839035977,0.49858364918726517,0.9484869721065026,
 0.38343135098087233,0.5559897926951292,0.600687079730142,0.5593363118954692,
 0.9727455006401615,0.11919367995766117,0.16768024602459553,
 0.1280034010431843,0.02740074087753941,0.8519573027164694,0.5689866635213328,
 0.02172330729965377,0.20342111703194354,0.12047717398383329,
 0.8860045353925269,0.12669413750704006,0.250410994833255,0.7206773615924272,
 0.615825751607959,0.16082014405079137,0.10053869351622435,0.6442349467493939,
 0.7084844373722025,0.7681030833070034,0.9788131658384543,0.11411896141774791,
 0.013135152715079479,0.039095404007438606,0.6450478294815254,
 0.8525257642117902,0.19445990575746597,0.7260063750296709,0.1369715680206529,
 0.3521331421256696,0.6610006525022851,0.14518648980606363,0.8294800531331058,
 0.5380730914566648,0.6635470184494446,0.35254531157882685,0.1733055388221506,
 0.08179322294247415,0.6648926256243631,0.4958984082653155,
 0.20255397390083885,0.33782258903536677,0.6830168077195989,
 0.23445378005185646,0.15632569341938352,0.628363931831017,0.4976238235728464,
 0.23513767048452017,0.2504553651582788,0.6177000398072683,0.6541331487429182,
 0.40468254011123295,0.2921225581000334,0.6147908141074321,0.3342178037464568,
 0.09390503002752004,0.47513091211874836,0.6014792501463431,
 0.8894777304401178,0.06716546129662193,0.3212865256532992,
 0.21002852831211338,0.3055710273895226,0.29794897948727717,
 0.8313107740594056,0.6518980666901768,0.857822132420476,0.10905343504572418,
 0.8373562006500676,0.9063186153504278,0.22411724980078795,0.587687282938641,
 0.697743682381234,0.9203247664936234,0.6492222322586435,0.14647922498378207,
 0.5904849821795691,0.13905869995324882,0.8517191109923495,0.5074126058237242,
 0.35033993505351424,0.9917700218210901,0.7928019619327532,0.7865762769406349,
 0.6075928093559751,0.6120191943547942,0.7884260725750196,0.426295671082614,
 0.9649463889144971,0.6844941796417082,0.07622938536089885,
 0.47774690747008053,0.8551166525574747,0.9395586836863056,0.8714047576025004,
 0.07246026553792884,0.5354099231425272,0.6420599180354355,0.8568598802809448,
 0.05158406451930431,0.03702086297696061,0.06057155258908886,
 0.3819724533045663,0.39399300762116307,0.5573122826380437,
 8.269336684394446e-4,0.10099448459152349,0.16804784345491675,
 0.8908401144231639,0.25092169744964277,0.7062334576072636,0.4619650422420637,
 0.5420291823489131,0.007447833888818645,0.5798129542334923,
 0.7356105587006955,0.8773900870786898,0.9208517835535326,0.15620410008734298,
 0.9424702961293558,0.5869789371969998,0.8236039511666567,0.04706545614631463,
 0.727360380469552,0.6897048347801799,0.1859063546930827,0.28998208955930993,
 0.5207947347559108,0.7228286453480304,0.8645670958771257,0.1685680027106512,
 0.27962186306475467,0.9012080734413562,0.881704264045422,0.9615725692204384,
 0.8607682298942471,0.12882868105994527,0.7019793070423993,0.2891530240368093,
 0.32086790076140714,0.889394098379075,0.16272782184628864,0.5312152580533618,
 0.4499882855063655,0.0022282965027780044,0.8021389057093539,
 0.4604360059988378,0.34738740822902203,0.33474322033527804,
 0.8599419322887664,0.14845648066363326,0.3855618762325834,0.4870986678907995,
 0.8122951468806381,0.48713329819499174,0.9402693607125858,0.8513441133889739,
 0.8669119488773481,0.09459832522653899,0.6911729873409781,
 0.31111917673106426,0.4226601318847657,0.9396720661097406,0.8536744330467617,
 0.8061605014143065,0.8566265525611919,0.0881413547476545,0.900649005406799,
 0.9282939204880225,0.3196153552415225,0.2540677177656927,0.39170329181189567,
 0.6840440812519422,0.2703365454173,0.5551596310958918,0.25258405134328243,
 0.6889221514927314,0.6685801206668669,0.6284592535036196,0.8787689550556781,
 0.9157699769074628,0.18639282593587803,0.21048114742906687,
 0.1555916719259821,0.1430337084770723,0.08212457735720347,0.8654987813791553,
 0.9725508636690522,0.43790632389346096,0.9238805729165853,
 0.34653997598261754,0.7398981040073691,0.25811582145583734,
 0.4480481396016125,0.05696650737410014,0.5077927611760026,0.9851888429620794,
 0.40075654333840394,0.5113038071346184,0.9078049037378146,
 0.33645706997006886,0.7236392409522824,0.18819066070382806,
 0.2207793872429925,0.519862578603372,0.7047956328334679,0.15266128693851022,
 0.811887003650627,0.327948734904303,0.9288809623186086,0.4982268377256214,
 0.7479051566176537,0.5118812538959252,0.9454381843507645,0.3228955007514135,
 0.013098701574530747,0.904641987100001,0.7440691664061402,
 0.18216190471796945,0.6091301234726058,0.478375909509436,0.7654189827319852,
 0.2716562462865936,0.8076599195880637,0.3847268471609867,0.7341503118318219,
 0.11032919399236352,0.8887804954116784,0.30110028533221933,
 0.7368958693344914,0.41112161312476336,0.2315758024916923,0.2808688136259344,
 0.3683251246024115,0.5615978694935176,0.3244721956164436,0.335342947047615,
 0.474549121331056,0.8921010050724103,0.4560281319603996,0.6197451821135628,
 0.7644368957679094,0.4008767618685817,0.632617784263136,0.5950423052090476,
 0.8112024167080225,0.5613708284449009,0.977576980362273,0.3202522160894389,
 0.6062229388952506,0.3945138116882587,0.40717396417819995,0.9727654299666705,
 0.7059223310040468,0.15191401756655043,0.08253003289210936,
 0.6143619970930867,0.3718586723906262,0.13770286357093564,
 0.28663295863129346,0.385320721679681,0.8112497113190084,0.9667268777754132,
 0.8576507273234705,0.9588692958786201,0.0907497593872032,0.8909232061543509,
 0.6131246856581347,0.249938416337808,0.09755130256347277,0.7366526830689212,
 0.35156498515331025,0.6207513732021427,0.8871806457326366,0.8460509791009622,
 0.32621692575728356,0.13943107758659334,0.08410225168866681,
 0.6769713500985028,0.24280213068923706,0.812943402750604,0.39413993795984237,
 0.2410115926558708,0.8616630623181063,0.27774814064400466,
 0.42135533775517253,0.9438916559961056,0.7146231713616793,0.9473262043865627,
 0.6714782633459594,0.2886457214747552,0.47557648382721274,0.9240225020582269,
 0.16246375324695772,0.32462011650630596,0.17913005272880278,
 0.6446669868415243,0.1158350811522838,0.7573532017401785,0.5890842295442008,
 0.05590889133089938,0.9334408766956983,0.6356394731933608,
 0.10479483657318212,0.8793530797312782,0.4624732839925575,
 0.24157001077187312,0.2931942727020609,0.29573235811436915,
 0.5436916480009022,0.08344424111415583,0.6178292211709144,0.7998306128250454,
 0.5043370781472988,0.04817600001605571,0.9863865761830019,0.6512449682361776,
 0.8224541849942337,0.9927711787115712,0.6222624350202149,0.9528103011691772,
 0.5896671628848388,0.7409662486813472,0.3846753193063662,0.7299004981153805,
 0.8922048956281677,0.24206717328656402,0.6271186273479509,0.9955357352462428,
 0.032692246660387614,0.34964370092421193,0.5796082442736432,
 0.05969745340952559,0.46057795692364656,0.07451250012122168,
 0.2931839656292121,0.9274400008894556,0.942640589972061,0.7462138631526875,
 0.9985650077201205,0.9448192687526165,0.0393604994910538,0.21162352416001406,
 0.2145973946697416,0.9136909488968884,0.4995930116762973,0.15734763064888702,
 0.6031352257634098,0.361255726467687,0.04901439247437167,0.444613451728869,
 0.2313692723259343,0.03196711684222564,0.22003881408227655,0.78838502116706,
 0.19279285061134233,0.4866206314390461,0.3301417947072328,
 0.008751430395650628,0.366330429602568,0.08669483812040601,
 0.3904095578197193,0.7768850835443685,0.783096609654425,0.6806372582568128,
 0.9623535983083857,0.2952887478409165,0.699167938422639,0.9816961941978423,
 0.5176667343289818,0.042976188125411485,0.9420677722303885,0.899512628885825,
 0.6465097876235939,0.3636287186493583,0.9617904367226585,0.40900313046344916,
 0.36879426279546335,0.8412238610600139,0.8874763099327349,0.7797740554872151,
 0.7320010596265856,0.5040670438114738,0.6295146928827134,0.9698748009457285,
 0.1497411435276601,0.8559625975585714,0.25859843827192297,0.9203076078467464,
 0.628849307510755,0.20779290799099892,0.5203886794805109,0.7091812237636067,
 0.6309181757135244,0.16494431774220186,0.8176284799817525,0.8433833549494753,
 0.6837535157489336,0.7376594123888162,0.5161037274046117,0.3182585300786538,
 0.943021984904672,0.9182098711554632,0.5773352109626052,0.005160309064244584,
 0.4108101394873601,0.7017616322625634,0.010391059657105961,
 0.6236662369535099,0.3317120545453669,0.610167935264073,0.7594187502959758,
 0.2671159504806535,0.7400088669863041,0.9503484365707597,0.36782546301757135,
 0.09249656173707432,0.22450880454704059,0.8503338957478288,
 0.6406303577085517,0.5568402904878775,0.8233725242836345,0.9459328717757224,
 0.8558308393689987,0.591385638257597,0.11616547711765379,0.5367801010525398,
 0.8215818406788616,0.23606022304635277,0.536606846144382,0.4026248263654175,
 0.5818994812277148,0.24552379846988392,0.4061519613025698,0.995395473412465,
 0.02518758393673537,0.3382441031920945,0.05417066870343823,
 0.10431985178409542,0.6386154712332288,0.8070206822156659,0.9382652846310517,
 0.5875086012373646,0.05924060290669142,0.13029776671446525,
 0.47811913424604524,0.38279316297218324,0.7220257574068909,
 0.9444560937789686,0.5703961426619467,0.7441228271231345,0.16682468655487415,
 0.40402050090818986,0.8863118128403675,0.6470090927711302,0.4436465201654477,
 0.1354012222570774,0.07172476303821418,0.2842509554904168,
 0.04195054269926235,0.36165868097109943,0.792074161242227,0.8304245785289791,
 0.5777147918503391,0.3535488386356701,0.3918443094467813,0.41364691366432915,
 0.6342149682160543,0.09974805403434395,0.3861668892709189,
 0.39097891666430684,0.8902709410578205,0.8492985216588601,
 0.08512092278419614,0.6518292391127444,0.3385768695211048,
 0.45248742340025216,0.1960432392442878,0.4745374176386048,0.2880464222409531,
 0.31443525789865856,0.14218779269067305,0.477003150473435,0.3664748845871375,
 0.5352592212875085,0.245322999523792,0.9268667125175514,0.08590589226482503,
 0.3752575827070259,0.8399541738228142,0.5092508565138817,0.14838864487569736,
 0.31051426756053957,0.6734869876239927,0.37937663580188374,
 0.2056708124281198,0.9854965146770447,0.5159736687573766,0.43803744164835834,
 0.06652126276005199,0.7435818772245951,0.25225984510193267,
 0.6158103038300878,0.4232574168158081,0.43255458969242355,
 0.32640165359815565,0.6224164991710391,0.5432285695896267,0.722356558282572,
 0.9826473868706401,0.33079633898507066,0.48310218315672726,
 0.9014869643558882,0.9509587014305833,0.5187243887201329,0.5408640400699257,
 0.38614471903711367,0.6068369231524671,0.37631327200470144,0.513205614002441,
 0.09448085136768558,0.6325898714552958,0.891520862171576,0.7880633167320409,
 0.28824440633494197,0.9754648711524838,0.08607071588056692,
 0.48179182530830245,0.6748446355579826,0.9824148829896692,0.9436821019240046,
 0.4632430379867475,0.9537439020164489,0.4428728768973793,0.8243433621036572,
 0.0899509616333728,0.05111374114872824,0.4365854400027519,0.8981852248961599,
 0.38285767247637914,0.5374961178960976,0.2733477187369422,0.73462378639528,
 0.11306692753442649,0.2603947545837896,0.1460993423733462,
 0.010428914730768901,0.5135737868362729,0.6786827717595117,
 0.9170733626844798,0.43961840589216117,0.9170237557858703,
 0.43035673871077795,0.13462208096075967,0.4973906959270651,
 0.11343283217991385,0.4340496455043976,0.6703336891461655,0.4792564778766202,
 0.6563331698932174,0.003284786778680271,0.17246920264189358,
 0.11493457689260256,0.019779670157684337,0.17894603899564276,
 0.20425809972175613,0.8581016846596168,0.3928510436392697,0.7085038671986651,
 0.4411324596640841,0.12212534680502407,0.495071217897637,0.5512737096251773,
 0.6605406432123964,0.48755786292661174,0.05487129336603891,
 0.5747125639053514,0.024017504756453478,0.5064941498048546,
 0.11387869161729836,0.6793150354449814,0.19245080896474231,
 0.39488881602721027,0.865159686896589,0.7512485027389562,0.05471004079950337,
 0.9344881380756668,0.7800763981192508,0.6044881187751118,0.6396124706663828,
 0.2752718306544215,0.021463628465925755,0.4087416805232785,
 0.4004705845377945,0.7542880953085151,0.2049791936708797,0.261423563357263,
 0.23715464528779007,0.25425599474953753,0.9089268277979331,
 0.3164859237345199,0.9916136816442117,0.5229010571385933,0.01888399506011329,
 0.612715071707736,0.443460387498003,0.65879406184297,0.21769198130701195,
 0.7238679261390752,0.12611457955445549,0.2817489165899558,0.5847850556082596,
 0.553448351595687,0.5770072169891699,0.45077201521518184,0.2791883937404017,
 0.4999162327853245,0.9851833960580906,0.8411958684284164,0.5920473735527363,
 0.6033479247672954,0.011225570719848976,0.05916220055960253,
 0.30290322740570375,0.8447629171089306,0.008925913991192913,
 0.6590655528596001,0.4755763003268457,0.2260374586569407,0.8272830712809391,
 0.5648368105775791,0.23491289093195156,0.834919489340469,0.3760517775055898,
 0.14128847526842214,0.912734523987704,0.32683406945600313,0.5643771128922096,
 0.5787856261020287,0.7995175884402053,0.8829927217185445,0.39321108491730106,
 0.6730672755944426,0.6905262107098921,0.20027963037462249,0.4253256149340163,
 0.9428480306723466,0.8733739469331345,0.7395104686390028,0.1845041168620417,
 0.5483388735155084,0.7852420322729081,0.06223792231894909,0.7416119657026011,
 0.5815197891867973,0.7152990353909893,0.9211203582739667,0.5753346090963664,
 0.1714982885969738,0.22071704138064552,0.8207154732178361,0.4159349296367697,
 0.6827846209279405,0.05240812337584866,0.8026334401296441,0.6532818296754328,
 0.8729415077327818,0.7740687763719449,0.25913291194489907,
 0.07252478177409905,0.4760611167991997,0.7444792666051365,0.8462549824569763,
 0.20588131967731704,0.7820731268517366,0.8246165940385011,0.9562153417752278,
 0.5496635240665617,0.3945767291412742,0.07905670525727582,0.7321514480716754,
 0.9759927538246788,0.49300751121221154,0.7039385952830257,
 0.19118189330024316,0.9663042832786568,0.15016570204061053,
 0.2635636377405739,0.49451554632055106,0.5346966472666448,
 0.13497957617939527,0.09107534382253268,0.7896309896312069,
 0.5956990245130718,0.5754697657115889,0.2925416984395762,
 0.016735979849662908,0.9614370221545284,0.4072325498658522,
 0.9728635859318874,0.45557015737667783,0.0026780278138331948,
 0.5551835044429305,0.9274064214844938,0.7810264031199448,0.9720263892075861,
 0.020620306620007134,0.7125852401645822,0.6469946091432307,
 0.3416037836070198,0.49544095704953506,0.9560253364755342,0.7450964252071448,
 0.5190350774613642,0.0249263962245857,0.9281577435770824,0.9241278899483482,
 0.87839504082206,0.19508370341461267,0.46465087680608863,0.7745502553993724,
 0.14638410619711462,0.6516895640617048,0.8694672841268746,0.7747511595378674,
 0.6917866193232711,0.21640693330811156,0.3877089592894266,0.2061759936903359,
 0.2078311511555727,0.17256684932315947,0.21446351004883868,
 0.4852452107442987,0.9706163007044918,0.841815267849125,0.476380203031756,
 0.8954857894203048,0.5113499801755287,0.22447013556734818,0.7127601228603891,
 0.006354025261353313,0.3969942343376939,0.2784903260999956,
 0.14893556487192328,0.6174818870237747,0.345561162225666,0.6565722840032815,
 0.8002901547981653,0.8818465748963547,0.28649597327940424,
 0.14428861620702904,0.6786615670477023,0.015648184937000154,
 0.6318208668558911,0.7675748389334605,0.47672992587311724,0.9497507579181527,
 0.2657479495042374,0.6726792437021413,0.36507333504399153,0.8362573256048444,
 0.1736562141137301,0.8025290845831473,0.6807237290811345,0.8634541457695172,
 0.40209204650098684,0.9596135123380226,0.6178098740718349,0.8245515342886995,
 0.030357145265496266,0.2760751559074093,0.10796273552112279,
 0.5489038510859282,0.4910319958057181,0.9819564538848213,
 0.022580741815154348,0.6756282575742181,0.9801961358404723,
 0.5034871044022888,0.5603154909998581,0.2928574652186894,0.9173519458524003,
 0.4897254006161186,0.9332774492415752,0.751141047465965,0.3450990070092953,
 0.5407688647687938,0.19086114342205052,0.41688520766579995,
 0.7857617485386305,0.38119447997385003,0.1704139423563127,0.6245848968675995,
 0.8078238697450189,0.9484078095154457,0.9555620447876269,0.5003793660987473,
 0.6037942694567383,0.2857056282781676,0.508686617761489,0.389824467671284,
 0.604477675333736,0.9413976566977205,0.3785043907047274,0.9344690348610047,
 0.011428604537560116,0.5337020444953567,0.7566942055158089,
 0.4716449744560862,0.769989701926927,0.8459453000030381,0.28124545827333125,
 0.945879727102208,0.14672168728529333,0.06379878610084966,
 0.25931027469622425,0.24025898878192453,0.97445805628434,0.17309468416451657,
 0.25377361071293514,0.8539245820980561,0.7022312932658067,0.3572982047340716,
 0.9584134545413573,0.677053276524808,0.5930009196404971,0.6149295227331295,
 0.034028927662582475,0.2294989407628545,0.709490608247206,
 0.04968371899134749,0.16280231483952778,0.10232011162860444,
 0.20232056824372524,0.3498132719132595,0.3159884267327431,0.9179533691458799,
 0.33588216472614885,0.23398632235138617,0.49391585113430503,
 0.26696438186310156,0.5485186922378249,0.6453023734908906,0.2554767538157632,
 0.8213857985493875,0.9246721081498958,0.4929489957994705,0.948620998817425,
 0.5615218235397601,0.7050643250003293,0.39938498280568113,0.419720049930844,
 0.8717821293971357,0.1654658933806218,0.11961400644361919,0.2619510182528695,
 0.2774503293754196,0.3874473933998003,0.3601734979732176,0.7791874571892451,
 0.00473576703848122,0.894843336423424,0.7478642692018569,0.27308602870127896,
 0.09172525363409723,0.8676434496453372,0.931450795641274,0.35963276334865035,
 0.808539048805409,0.680841288040813,0.5930118692745009,0.6048128763802757,
 0.8561038493427445,0.30452122957884475,0.24826708535042874,
 0.16783997875206946,0.8503337396110404,0.10267936942736289,
 0.20568721051802252,0.10066746417933525,0.40659399483283876,
 0.39042521959616305,0.8220277013286186,0.32122537423918396,
 0.06476656606907738,0.5079715140098127,0.601791628599389,0.20470611165583175,
 0.795181439020352,0.5109054386245306,0.6962682031887895,0.12934996716095437,
 0.5777414163825041,0.462394685369908,0.7227161247115894,0.330848289522075,
 0.12078618187438872,0.4725484605737936,0.8880544890171268,0.8405402364184787,
 0.9266111008063055,0.5032835041073397,0.9786879048845429,0.3266379232030865,
 0.13336749174693385,0.40462831393808874,0.10562452928183663,
 0.6578409111496638,0.9700907730119442,0.389230258285458,0.44222770188239846,
 0.23115747524990216,0.202852862125528,0.25453180248924834,
 0.48306703361417425,0.8076699551997423,0.05577889934672464,
 0.5582204923971416,0.9555514001561927,0.8654073009644474,0.5510237103182754,
 0.08488251133892899,0.19246695290926685,0.988588203630659,0.8620483063809361,
 0.0724678554404623,0.5367088320071056,0.929724014243131,0.7390837937305177,
 0.90459238952706,0.20123542964537044,0.48933923252857137,0.39295789046023466,
 0.19803908912137191,0.29760143183300025,0.13457197425911283,
 0.43326277984478145,0.9818458088122318,0.15041721248682316,
 0.4197490146037163,0.3120463324564493,0.99237781747677,0.4261736966166759,
 0.1658078311629605,0.4355660220064559,0.4844568543629052,0.590698417336706,
 0.41581992239761667,0.6902421669195111,0.2118277704520436,0.4747753946225093,
 0.2858390280500388,0.2259622739724485,0.17682354631754205,
 0.28943816722171734,0.3569680110349748,0.12492119222421816,
 0.8130888314603189,0.9887076635064602,0.11035390533111356,0.654326706444356,
 0.7858135398353884,0.8685157969998102,0.41498360773458454,0.8935990391977302,
 0.7382429355731989,0.872447914698672,0.7397215928885065,0.4171029301950774,
 0.9201820612570344,0.8029446095495391,0.6542949660444105,0.7557360898217814,
 0.9955804778488837,0.43394152205566683,0.3924671881812015,0.3158922145104558,
 0.7167786053457957,0.24657529620137864,0.822118175510343,0.5618917293864784,
 0.3600691159894909,0.4059488733415607,0.6426959817901374,0.22275300570500622,
 0.07365649433955135,0.5201942116216092,0.22697389544077218,
 0.12737480120266043,0.16800084497299794,0.414734850033847,
 0.04523982452201325,0.8807545444871345,0.4364418438120813,0.988557591160297,
 0.26011077832941865,0.08128375599865267,0.8428393548289626,
 0.9483798995884234,0.8851733717656428,0.5072536527031948,0.3126227859454067,
 0.4996475666494362,0.062342094040446216,0.42939745638204174,
 0.9647327926450838,0.6569885036601062,0.6353726217915991,0.5331884599584722,
 0.3459997627905571,0.11012489248165858,0.6060756733987932,0.5663457890734305,
 0.11477441431940427,0.9559927116181657,0.8610527183270847,0.644058916595284,
 0.34701214153127424,0.1159121339166006,0.7985929441357955,0.684774458798555,
 0.0037059430157446727,0.35575991269886353,0.051505898038862186,
 0.2748633439873577,0.711420850071235,0.8620044850911587,0.7318709814131694,
 0.8146824449183699,0.30467153780383505,0.44090541493174396,
 0.32211716402773094,0.07011596883959414,0.22992034743953604,
 0.8716572994013715,0.40185818464531653,0.1928184655105002,0.9554501558170466,
 0.9192945973457047,0.9079232605660614,0.9428677254104033,0.8752466350082597,
 0.5359150584600443,0.2623298680888422,0.13165948641802327,0.6544005609639612,
 0.02958733212260889,0.6368487344326893,0.1898495368995211,
 0.045783146837081734,0.9646079551328295,0.9286318201245785,0.525448683581811,
 0.8448372256353964,0.407916867126199,0.0064868753093967335,
 0.8206279932234557,0.3891523794056194,0.9031416953152488,0.2203935429748407,
 0.4527193778121861,0.5081058664750282,0.271103839432413,0.922284176536182,
 0.7362659178164704,0.48516447666917406,0.7439512176395727,0.9628593476663325,
 0.8970309022704432,0.8256896477845707,0.3944093852902987,0.862211260497496,
 0.07798988991771005,0.7786343307745803,0.06329197504783302,
 0.7293403108414958,0.3012881923581415,0.235847489120657,0.3909277879109102,
 0.4230832520631629,0.9593009003774831,0.9935950763978855,0.19048674465553406,
 0.491908439696929,0.386265741399078,0.2604953318212433,0.9633916187748582,
 0.4174414449788003,0.8705740998058596,0.0024426399299593005,
 0.4410676043639148,0.14459606330215669,0.47821725771884804,
 0.30284062080095053,0.32793431230489156,0.8196874737921196,
 0.7383780770164157,0.3642877102765354,0.2660795709592644,0.9554109345687908,
 0.9581616938672293,0.021151432677505966,0.3472547650037491,0.599429527656756,
 0.127475226375107,0.9441158389739845,0.35257662162731984,0.5701129946709158,
 0.7325209540748766,0.7188487106515307,0.8549990631949442,0.8374284948208225,
 0.18439996061925545,0.8846150997185367,0.286181764029795,0.5942768065780388,
 0.8504675741781224,0.37021311423679837,0.4709308957838271,
 0.13304796793928997,0.03799542125759703,0.9788659266729289,0.93399380612114,
 0.27189147346358533,0.9282518583415311,0.9114727842263639,
 0.31493272880753476,0.4350767026308002,0.6356085294090187,0.5927619830229425,
 0.774457186789385,0.3598764590128407,0.6604804322604572,0.2462156153238394,
 0.15713118406350635,0.6118471326442603,0.39853709847603913,
 0.8970627053797225,0.7359381730492761,0.061541042126034395,
 0.38781361247450774,0.4444756277848301,0.5853468927577548,0.6799998247190798,
 0.13927687661336996,0.5588509068797167,0.22410595355810703,
 0.5317247432741666,0.08801959147232274,0.6058623849798019,
 0.35281246844966285,0.7378645117454871,0.06925887410155551,
 0.11663613308897347,0.13547211601925668,0.8274001676277921,
 0.24498773322130618,0.6768825890632482,0.98984643326599,0.02128101218731182,
 0.018756870597578246,0.7302480703091638,0.16954727290135208,
 0.339053310402224,0.19626934149267616,0.6599800813409012,0.28266699090683123,
 0.6596532084680133,0.294801633788802,0.738458899477857,0.6187601524105689,
 0.6268083659983199,0.7180316859626815,0.5649141032704275,0.6117081398113252,
 0.7438721740948511,0.8327911814567248,0.9601636248426605,0.9963741745895329,
 0.1843104736320621,0.8275028504018009,0.12396893901349315,0.1462207062555907,
 0.6322961172518188,0.9115298918085545,0.17658620745746578,0.2996894332342065,
 0.9746627835573886,0.16463912864544006,0.21324265062101322,
 0.07141528277338927,0.12916824787928816,0.10808458255210251,
 0.08619457996849655,0.024369511863531756,0.5962380393063546,
 0.19390780784609685,0.11910703876211626,0.025961183556525547,
 0.29247557977020633,0.8428621793598412,0.35753286983783283,
 0.2963407924539888,0.4663743982375511,0.521004216467789,0.35533550018498494,
 0.6825501188497828,0.5699570011385394,0.2633418070664445,0.5707385074495487,
 0.14519715624455876,0.8830959470075196,0.027381599893242403,
 0.06238817756429205,0.03665302210251775,0.9062828488054739,
 0.5899436097665015,0.2994591334285569,0.31875181519708773,0.2991673621938855,
 0.813335618050234,0.33865001686967067,0.7580192139662848,
 0.031521771962766065,0.3057982328511578,0.16421054795685142,
 0.3426547618330201,0.4759808870361527,0.8208438573020694,0.5575397817164351,
 0.36725076102964527,0.8833661788736581,0.4792842385467002,0.871621482836127,
 0.04491648854914021,0.10448738826919335,0.032671375020607796,
 0.9308435863044421,0.6904500881849533,0.3468486701280211,0.2807571531777475,
 0.17989128018394207,0.8678122876406161,0.49732965073580315,
 0.11975215895445479,0.5433408135044735,0.06871518580322278,
 0.8891556811157505,0.41829559568604646,0.10627143187605059,
 0.7574635632804744,0.418833705176076,0.9981558441403988,0.38082842013914553,
 0.38785735548288136,0.6098861878591476,0.4741738496609167,0.8260408809196322,
 0.9577081538405452,0.8848917198965899,0.5168209158064241,
 0.026533690868360438,0.9957599005126051,0.10868537014780011,
 0.3416935305116887,0.4672361711058064,0.5652893650964808,0.1877451791821494,
 0.49698732861623185,0.5549291614094058,0.20273144776986296,
 0.4599475861377941,0.982074229349536,0.0647603725700645,0.8560429589343485,
 0.6251079309282601,0.7195344221534894,0.10331518593038247,0.6017272614027267,
 0.2690861695277308,0.026250453207575797,0.6096235797304889,
 0.4898272194934302,0.20516333758980365,0.8305190453792665,0.3675798154469616,
 0.22369375884663145,0.491023196851042,0.25980203187709017,
 0.40442615062017184,0.15552486787770803,0.2691838982662548,
 0.3269171908552644,0.8965967744560949,0.10487553725212073,
 0.11416420579589404,0.3417659417945026,0.6394443989045395,0.9133856891261106,
 0.08371829440738188,0.8933553192979511,0.5635375169238355,0.7365056562703087,
 0.7052411387524726,0.49522734816674674,0.8739132587169379,0.9243474558482497,
 0.11612660971930244,0.22012168670352816,0.8959414568805174,
 0.41154673546640463,0.1300277614100508,0.4931481112272811,
 0.05415358743761689,0.5447841572652723,0.5879823349175529,0.8655570513826144,
 0.08327873008055442,0.7538849088683932,0.44880323321509275,
 0.7277316452185173,0.9377971456728287,0.7376686571132169,0.9624273459067414,
 0.8174086886570936,0.46525708376690056,0.8889032097643119,0.5925365723244874,
 0.7074611941749808,0.51328194198063,0.4079938577891382,0.9298627865485023,
 0.8724296256917897,0.8706870810900172,0.3871349664982203,0.07492092117890259,
 0.6102730764136599,0.8773275615010665,0.12008084504892491,
 0.15749692035408358,0.6090783766175617,0.6486449578992608,0.5775680643862183,
 0.6860501071192561,0.8648970048458479,0.9250545737446534,0.9792883100028994,
 0.17490347452654498,0.7815787118240728,0.4024767860818542,0.6477614268731886,
 0.7644725258116121,0.7376130288755042,0.12352488094355252,0.3137129344934648,
 0.6571082636312735,0.9643644537168392,0.4975151391127386,0.2312933722177719,
 0.09459088851679331,0.5882181786329796,0.8392188329301609,0.914103533976531,
 0.12626426350386188,0.06494906298349212,0.44711549840719844,
 0.9575546856953766,0.3627616363666515,0.18693402887512778,0.9258801593373692,
 0.09173631737527399,0.8102755616727261,0.961029910515055,0.6194140959987066,
 0.2602861729165661,0.20316650247209012,0.43641380599506296,
 0.9990211535571883,0.5455932935146686,0.0715589187696497,0.03369830560603815,
 0.9828132750335583,0.8773800810052174,0.14804385482056115,0.6003612311400781,
 0.27723147746501076,0.6806800106858694,0.9971730730357024,0.7421934675761565,
 0.31907352253934784,0.7871761976777931,0.07396383714303578,
 0.7542970457031881,0.3197401481519684,0.7861407486041263,0.28445455847694057,
 0.12356140268902394,0.6606315159366232,0.7684145966834447,
 0.49700198541811713,0.5405170871218887,0.4374184290336707,0.2398222016543332,
 0.3389712537933396,0.18910756476946033,0.6259200855682606,0.6091950861252476,
 0.7311159597169771,0.4833672336484681,0.523933864464957,0.8364143566904317,
 0.9087712935430189,0.6807048575686755,0.25638959157801533,0.5895269625277386,
 0.07909991022231389,0.5230445641142671,0.2999065839692526,0.9348279224046788,
 0.11464120930696353,0.3832882602425708,0.19124556877323362,
 0.7509305034056399,0.15624540775390017,0.9238022664299794,0.772279464303001,
 0.49898503094656566,0.6264050706393268,0.4102184807018554,0.6528289278366498,
 0.14424513144314055,0.12059537125523079,0.5822167713154653,
 0.47513144731005563,0.8415169468406831,0.21275533645625044,
 0.28949764833326064,0.5542711150984188,0.3075851441066766,
 0.41543887850480965,0.17713209414660924,0.9101569862912633,
 0.3861764350445682,0.6142637568621591,0.3927447094637794,
 0.030298457713386018,0.9835249553813415,0.9029965103658073,
 0.09336039130995555,0.33703701286332277,0.2346126128206113,
 0.6653758501167748,0.595664097050602,0.8614672141112878,0.8755954339251546,
 0.6390250106734187,0.790638420631298,0.6190016441091999,0.8918359465969287,
 0.29672421406620975,0.011954076806938607,0.7459601641256577,
 0.9676810669467111,0.1846497368686788,0.3993637238105703,
 0.017811066239094897,0.6899186851689103,0.32763732490548014,
 0.44583188951804964,0.3968619534995399,0.32441226386267974,
 0.6247502710494247,0.8728751890271432,0.20805597737304948,0.5321733532179056,
 0.8707213575206236,0.589478308244899,0.4791663357602298,0.012939983100147212,
 0.5814653345945818,0.1159618500907561,0.29543227660245597,
 0.26687843196119443,0.9951507204226369,0.07824116295257055,
 0.5631495021560828,0.2829546454808296,0.7125519620121081,0.14137616556902988,
 0.3841900232608869,0.36925526878621995,0.023996295559362446,
 0.5395657829330115,0.9404898295757524,0.9392402881803599,0.6058421449149141,
 0.30759503427614054,0.29990474329372574,0.5584992855218622,
 0.11519988838215622,0.02933015312917997,0.013102041057229785,
 0.5378783558206277,0.8244919084659049,0.2539665727445808,0.20663509233232547,
 0.7347058181693155,0.47767578745892625,0.16550123190007426,
 0.6020642938501051,0.5540515356270626,0.5929414354087492,0.44598137618494904,
 0.22385768964270447,0.015344431773590905,0.6931544470029909,
 0.4643095645009434,0.2672698558163612,0.7668370787942687,0.5193329517369076,
 0.5631047746238712,0.6233386196274574,0.5036898202205548,0.5390557287873359,
 0.07376093552973872,0.6732329139705023,0.42804711301012544,
 0.7438309888398822,0.9990483337299682,0.1465422253201326,0.7091547032342425,
 0.9670019241220776,0.17888409524082804,0.43801449404619486,
 0.6438919741368184,0.19230631942562337,0.42138965989049837,
 0.12756543011017052,0.2697542166228286,0.6606646508963441,0.3026790674353397,
 0.6608964533739491,0.9723832907607832,0.16954760112778444,
 0.31802694668546516,0.017766600337138505,0.8937489590633488,
 0.03707873718330812,0.03925653544937879,0.7987596677615915,
 0.9068467041398289,0.6086170360592598,0.6498010862646408,0.4555099567781521,
 0.5054460084851332,0.8048786391699807,0.10555238855153715,
 0.18451234098105262,0.16657644324227427,0.562778143582327,0.5926480675136495,
 0.8755568023458337,0.02961840686715722,0.3032964208616854,
 0.28143013801647887,0.7690699748938127,0.5976483034097846,0.8816243822654519,
 0.04304380435418176,0.8924946705959712,0.2849162341689041,
 0.20903557149718954,0.5152248951737788,0.8323297360241684,0.6063658702573805,
 0.4425987537220124,0.8197239455435417,0.6640044112916567,
 0.006075039721694431,0.062027609799674255,0.773633543327369,
 0.8093474810132417,0.24109923572866343,0.9968994022346942,
 0.19435584715085896,0.5828900083675876,0.20866480220526262,
 0.7504837796292364,0.6437338222973016,0.09367036589227962,
 0.29240402846242575,0.47733681567542785,0.44993258708561024,
 0.46236374074865194,0.6369715169749125,0.31167466778937825,
 0.9452307058730318,0.5307475887632219,0.11908154509845037,0.4486672651754877,
 0.37357120748498185,0.8678592484659132,0.3426109100589775,0.2595205203161671,
 0.6409500295907935,0.09790455800772746,0.8256703526242938,
 0.03828858722828665,0.5180050283132989,0.6161573436771064,0.2995358412103797,
 0.06970217059753292,0.3712010510802628,0.2734264240818751,
 0.27818259200147977,0.5629141067356093,0.7457592268843286,
 0.21073046741672685,0.14869877403619203,0.36594379618534256,
 0.36738460742094714,0.09707833081245065,0.9327637801479934,
 0.38914919993495956,0.5811949540198872,0.5714661069775286,
 0.15319101577518657,0.521414145142026,0.121201180728991,0.4099626695836147,
 0.9553792525440618,0.11402417050466285,0.7634812700574123,0.7059499505904357,
 0.7353479832904538,0.9360311399214748,0.9585108008939581,0.15469806799056562,
 0.23801984077648575,0.1974451982524531,0.6703542736998753,0.8034865225546025,
 0.2850944840622731,0.9782028811185638,0.6226850481809891,0.01665199722371291,
 0.12755795047260687,0.6354443621978919,0.5309469856069157,
 0.42198640312642965,0.4720707740646978,0.7560476980461114,0.8160408271623143,
 0.8607056291311049,0.39958041670314204,0.4837413650169189,
 0.007612671561772699,0.9082214728170163,0.5208108739137751,
 0.17076468828910774,0.9918296447119508,0.8157094252367175,
 0.28021895727766444,0.09588775661489435,0.6050392010634364,
 0.9521054922142044,0.2730412569424584,0.39223102579628033,
 0.45284929604929514,0.38583474903191584,0.8730296599728891,
 0.4801039825343615,0.3898207890781371,0.23874661035371336,
 0.45600243237350435,0.7503099970793948,0.10147026804026826,
 0.5236216258166677,0.3637343933931947,0.6789484841762564,0.44937017391938605,
 0.09141117881109562,0.9348292917355137,0.3095425604200972,0.065740856062244,
 0.5550666640727263,0.29516576912215897,0.19966520164347545,
 0.12592590197974352,0.29169199162413295,0.63023528772384,0.0985809969673409,
 0.8923319714695638,0.8301128008851574,0.4034269698791457,0.4192329465270861,
 0.5473575620448724,0.37638887916314245,0.9595105325506976,
 0.19524274509873485,0.8960441587103958,0.09877666262633711,
 0.9385596606113364,0.8159425405982358,0.06879746416303789,0.6788170414662299,
 0.7107850150966529,0.2009525651181805,0.10874730114576237,
 0.24173431214667185,0.39563976049162153,0.007779645365927457,
 0.8573653869346266,0.04815597509582459,0.3602688869040267,
 0.20252830378455888,0.7501666136277494,0.2751277644163256,0.4529987368950745,
 0.3705434073827524,0.673391428549489,0.8096097047279414,0.3443085002881815,
 0.1711423152797461,0.36445276237975965,0.7192704787597718,0.9167076258740833,
 0.47196089947949527,0.8343536031640376,0.4024452989680972,0.9868497877364025,
 0.43457963477754613,0.4775931524422373,0.871747598021226,0.387541420117673,
 0.614851897076199,0.844852843756011,0.2897103704613867,0.10937505333011177,
 0.5709838604446048,0.35411002014149484,0.3365951586249418,0.7445898059723175,
 0.2666255190488803,0.38774244949858516,0.28257119095358896,0.205906403612246,
 0.28583071367606727,0.5432404821757151,0.9966740087568424,0.543944363395105,
 0.9612741264039353,0.8374082258004707,0.8624750729217178,0.6865552172807935,
 0.6969211555877478,0.38279217762456597,0.4388827261543842,
 0.020699428258796337,0.8364471516563066,0.9290839894818154,
 0.4665424451387805,0.8584813715915194,0.6798191696646149,0.27709171165827373,
 0.7633439952159139,0.4861421748205178,0.3059169899914794,
 0.048308232986122746,0.005061529768396911,0.17090782691191253,
 0.4861088802317196,0.7832376975662169,0.38433398056012225,0.6607487455460586,
 0.0806100469093638,0.2950758905522447,0.7589348442103362,0.3810636271230393,
 0.10481038151300404,0.9830990713856349,0.3767637019305128,0.8304018814593227,
 0.6171529634683524,0.9351933490907688,0.8867444510657442,0.5535580339673407,
 0.0265922168705387,0.17026952393674144,0.05391771639906151,
 0.043135114681430387,0.6673756459649506,0.8286920973148946,
 0.6415857489191864,0.6060328237553457,0.3721178471810388,0.8102835770570527,
 0.17021918074431142,0.36578076671222093,0.8117936622356363,
 0.9129398604979025,0.5707377903777564,0.774091992291329,0.7808375971114226,
 0.8249273825939915,0.4173437482063217,0.9997432985200778,0.626317329245758,
 0.6656936555840719,0.45266178772891763,0.12803190287074784,
 0.0010670034684190366,0.0752043105673933,0.3621937150012977,
 0.6335969887809244,0.846761531616362,0.7337281079281182,0.6905215056906717,
 0.5965762541729155,0.07425427897146308,0.5335983771289553,0.9296190435879822,
 0.8228435537237365,0.35412240694557173,0.33575942322585506,
 0.5497344251913323,0.444312608075345,0.7828819013537865,0.27065954535350967,
 0.24531503500029073,0.3242570709023368,0.050281580794293235,
 0.20565514991120482,0.6936050303541986,0.6174716855998535,0.5308884537970239,
 0.6882488837917041,0.13321980251537358,0.9527821878069593,0.8192600422656356,
 0.10394995042874611,0.1498229445276955,0.02208851416848523,
 0.6940386609087216,0.05797794462662298,0.7479432898024925,
 0.10264949523307232,0.08302670122303446,0.3414588786092818,
 0.7745756942294361,0.5083090677039035,0.6790977334690435,0.7469044931118279,
 0.1343847471030094,0.10884204409643228,0.5472845826050778,0.7996804526021395,
 0.8659991230726648,0.52679043385032,0.5550674360724834,0.035240748148087975,
 0.15293978389632246,0.0628229877426697,0.6284412210477681,0.7507501146858117,
 0.1880144539464288,0.7117866689523906,0.568175948798251,
 0.0011162478745960769,0.514367472857632,0.09330370822567746,
 0.08131568092045804,0.2955000519654337,0.6867225280581639,0.1456717246868623,
 0.2159831035387324,0.6592883567296,0.6039006570815645,0.9654126843253668,
 0.9406913822237793,0.35341781762831803,0.2651707855317156,0.7907045809987778,
 0.9975810045500031,0.9309600400149926,0.35559954415570816,0.986713555546863,
 0.5630427431563492,0.08518136236074803,0.3293097114303819,
 0.02071983309901626,0.8842941926363556,0.21994707545026393,0.538183228177007,
 0.17819070490319888,0.8975015109051977,0.2249468935830179,0.7458448869226412,
 0.05531476669082591,0.8292672920715886,0.5058075362974572,
 0.42257097961321255,0.10203329317037257,0.3209676413237107,
 0.9011583090281622,0.0031080609336513465,0.6104940878653016,
 0.8952996058929464,0.16261089397967965,0.8011279880827493,
 0.31587549268241855,0.4202475323097118,0.2583385012708347,
 0.39567025886053686,0.3305522408888477,0.27577037383223524,
 0.3522212126780697,0.3791155373666263,0.7145795647078056,0.7390708970883189,
 0.5603235884110068,0.9567852592118188,0.5345451020162333,0.8865648612608688,
 0.22021223472149498,0.5432688985329254,0.12540602524824696,0.252588396533161,
 0.90217820301031,0.8808429807370324,0.2770522060973113,0.9045143314725511,
 0.4847924711081326,0.540937412149901,0.5501917606984806,0.4559709258313339,
 0.24981440284723533,0.8521004372546445,0.3681045059036999,0.5486125848337953,
 0.36547453965193466,0.6329285274709777,0.6568072236492436,
 0.33235614510730105,0.8988580613914636,0.05164978322451064,
 0.6732670615809419,0.7061430596977617,0.8421667684347381,0.1965819217532353,
 0.16345080796738987,0.06293271391065614,0.10322304371270052,
 0.3939359962926886,0.9604732910721113,0.2290892271522721,0.8675614593859985,
 0.11515972702221444,0.544485099776854,0.1431644609870918,0.9679925649685314,
 0.5906576183538434,0.5709526005372701,0.9935623953937154,0.22132650137461807,
 0.7777423042341413,0.9969814720675587,0.13168116405155794,0.6975954144542124,
 0.8842092024075052,0.7254224456036725,0.9354345748597138,0.3608586177992503,
 0.8577816998082071,0.47340845566858636,0.8546777972235188,
 0.11763153265605819,0.4552144610760198,0.28739088337768026,
 0.4419759699195507,0.45431798333278406,0.6328670046136077,0.5694897150706109,
 0.060704795313295756,0.40312594202515084,0.30908969117841445,
 0.22101482416712026,0.5937925725106026,0.26725061524688853,
 0.5370064672975459,0.8302067436974608,0.7357808105443703,0.4895018231859912,
 0.7777478452878654,0.9156083126623028,0.30001202246270564,
 0.42087064418332865,0.31272867218185085,0.9421648074367247,
 0.08805624541628909,0.3394804259256121,0.565515104721666,0.4278833548493627,
 0.5866788392596257,0.9732103241545219,0.42902212328699874,
 0.11263033610902906,0.8778967610974442,0.3815773735374748,0.8649204512256596,
 0.6299271615860923,0.7743681640406955,0.17814047756900575,0.570011238261135,
 0.7675331107412352,0.9106130656912114,0.09225780697193386,
 0.15511124813624988,0.00902968382595526,0.5750986490720612,
 0.10050191955795085,0.7773693660802617,0.1693482335376797,0.3993389531509268,
 0.7917272893179093,0.9972917463651595,0.8315743916045777,0.12087973592607892,
 0.04973195109214934,0.3373092266692781,0.9639965659851363,
 0.11517380263715216,0.43860701452319417,0.14208902823021285,
 0.7357360687700137,0.07859660567337001,0.9494628816896313,0.3621672800144784,
 0.6733777663279709,0.5764894445004736,0.9750198757458742,0.36894932888742915,
 0.39969622694309326,0.8987481356343837,0.5851550511348385,0.5752852797270628,
 0.360701553769724,0.5358081428996635,0.6576361882258206,0.6723377868190955,
 0.4932317284708987,0.7665323766408866,0.9544507173426146,0.7692620736015324,
 0.9265541103298938,0.8212659269813964,0.07993541069029053,0.1259912382778262,
 0.9954328825000662,0.455440936884006,0.8413576501591814,0.2749362970461876,
 0.05102088197599697,0.3790106609496622,0.6947648087076452,0.9024109683230783,
 0.8279209460504454,0.040459947653672446,0.18982394701636074,
 0.5441575203153035,0.9587606829997978,0.8842943761906539,0.48995235767190903,
 0.33664366993297923,0.8777040877592052,0.3709981920217247,0.471145680951082,
 0.8408126325825983,0.4756765407803796,0.6971902814840244,
 0.018140873565695825,0.7695869204606964,0.3395477975518526,
 0.9441365713133995,0.02904719591257321,0.5992137178080919,0.8243705613933272,
 0.885568441260229,0.29321664374226697,0.5245745550432465,0.12630594481945712,
 0.38066651898245407,0.7251660162043364,0.010686525978968353,
 0.04672555402688128,0.13926001040816738,0.22010139645426763,
 0.305822914905706,0.04991940490750213,0.7241155708969758,0.3984914310813198,
 0.4371679464526915,0.2841858256016686,0.07100672240993244,
 0.12226861444418813,0.15889130562435028,0.24580433569021343,
 0.6974890330148691,0.6105626630744306,0.19946177614532978,0.4400813910107979,
 0.4720117767971713,0.8312572995108146,0.4504968027501084,0.3035710447918528,
 0.5363153328740797,0.38193875428756896,0.5191293973418343,0.9302292248043144,
 0.3112660307508106,0.692200169723777,0.31199936102402204,0.899397501722785,
 0.4649930408863745,0.09405142330093053,0.9419723913681323,
 0.25467655524552957,0.35472850931040567,0.3161199639909851,
 0.6982910666347965,0.7636741877902147,0.6268989766141004,0.21839794964036263,
 0.19354131401295582,0.9598462838330116,0.06286086174738958,
 0.9271702271960522,0.23858574299583257,0.8345180941289427,0.8723233489416194,
 0.4170552719315832,0.23876667865233892,0.4252098440464249,0.5694451472833469,
 0.3359721040884923,0.23053826896085128,0.9617855975105475,0.7789376778736259,
 0.04297614240163372,0.9071397965151389,0.48857938590335537,0.357505549285995,
 0.0970183200360909,0.3898592316873205,0.07775096808714665,
 0.29657415776971563,0.2631418582823821,0.9005794179326052,0.3525299410704342,
 0.8625181359213971,0.15356997560851438,0.7074986870223675,0.7963506642546664,
 0.15156646791657846,0.20135976361725905,0.4727292103520806,
 0.7338297833898627,0.401661550455104,0.030609661244285213,0.5898023632496758,
 0.10713683947349273,0.9701567627515868,0.6888563864642524,0.9872177447540416,
 0.5198772507819956,0.13158684712419544,0.1723237307514519,
 0.13832561173693592,0.874264741974966,0.6202148829712395,0.6063999383586135,
 0.5514706078018352,0.6411714243343279,0.22824133339959973,0.7882988421195697,
 0.01887858536179854,0.0014459700284978894,0.0898866476528506,
 0.4243437058660341,0.7092406288103988,0.2191349847761651,0.46116059902927153,
 0.8116758018060313,0.6635049236394134,0.717014696702274,0.5542386243457658,
 0.45368794665592116,0.11630442116841455,0.869883313306947,0.9008674374826071,
 0.6419267903764567,0.7642386660508148,0.8787069824480941,0.3187379549649708,
 0.5506193466382877,0.4582704032590582,0.16918630571671955,
 0.08778997283368106,0.010309617136832538,0.8302573329133235,
 0.3135945399607446,0.7195710681825283,0.8690518366890345,0.4818470229988997,
 0.7258696520470939,0.9488051649887407,0.26444980945384833,0.587550889940044,
 0.9348556909038243,0.38213657391729083,0.042331625201544876,
 0.46728535840704455,0.8482966148576551,0.5682669049434275,0.6402601238986918,
 0.4292237895307097,0.798766942517218,0.4468785700288599,0.03404889504232411,
 0.6273601964457401,0.5849158797685995,0.4454551444365029,
 0.009824519997132075,0.18821349378594343,0.4868262095437227,
 0.38103556864104593,0.2635934877433197,0.9906884600482635,0.8469676449452374,
 0.25885832372059303,0.5920456720196817,0.019050784986197056,
 0.04961326691184542,0.7421222005760489,0.5847264397451267,0.9792109355523905,
 0.49908071468605264,0.6745479824392403,0.9941443394314813,0.4801205715592872,
 0.6127849096687769,0.027690426957467484,0.9350509013414262,0.575904482162314,
 0.7258551345198027,0.11167905113356047,0.6189151204688278,
 0.32480621113663966,0.4575499678677023,0.6046614616421647,0.559456679485025,
 0.902749241457807,0.3353199711231545,0.8170725283048522,0.09369835392731662,
 0.6086180503022395,0.9055492846112468,0.2945877343407769,0.7626625561244044,
 0.5535546516065246,0.7542347865026989,0.8631446983203224,0.3503698221972985,
 0.7845103794914875,0.4508654626124422,0.15648601517351823,
 0.037595886631436204,0.6196959877467857,0.284463606107233,0.7457909508527776,
 0.6471466780038658,0.305321137840768,0.9853015766902363,0.7604487366925744,
 0.39544879073925554,0.31120965832815983,0.45712328250863243,
 0.726901196140755,0.9058603577231775,0.38632777244021055,0.4403840820842798,
 0.1168225517778243,0.9023676840199053,0.2270535866486777,0.40909034372914,
 0.6405975728199277,0.34531019033588195,0.9064832168793049,0.8571772062756636,
 0.5529222020880294,0.3218578445952621,0.0227193544603137,0.7244964650565812,
 0.6198923881867211,0.1704274166348828,0.8418100336392225,0.9314236003297363,
 0.881358621713461,0.8846907661510611,0.17163118236405217,0.9154573810471051,
 0.8344198829413453,0.6794860057114605,0.6327269303450217,0.27453196519250356,
 0.11268352329755049,0.1425427106584478,0.9799142493143025,0.4246966503679943,
 0.7838653923628093,0.4841806293481634,0.02525531714800866,0.7944434860203912,
 0.6380170253339619,0.6950904076195121,0.7074830501027048,0.23643958258168973,
 0.8225577998096981,0.9996244121502038,0.8572889721210878,0.19601388536334352,
 0.7388419968661858,0.7325129725949228,0.9173806417165922,0.4934960258267841,
 0.5799959235938688,0.3208267191411145,0.0297466615584574,0.5285918772009273,
 0.6971631113364485,0.995707733390497,0.8945208486298101,0.27202768473278693,
 0.29593154627806584,0.24687943436836235,0.529564545829947,0.8366740480197196,
 0.8688041142275763,0.016098355768976558,0.3001670727621014,
 0.6008279159272565,0.5028514500968205,0.916148315730312,0.6835382272096819,
 0.824329075533151,0.7539582134372101,0.44723203998344596,0.5358881127736115,
 0.5743620426594446,0.4232546102728314,0.7349109893336603,0.34668431864474014,
 0.0012210038120640254,0.16791438587466112,0.4063470296515689,
 0.73736159392687,0.021824944679257885,0.7897182000576878,0.7367669817410426,
 0.6732813054406049,0.4429571430458592,0.3871565021587082,0.9094071704562767,
 0.8135974837368107,0.9221302475741792,0.3455999209414483,0.29135425453523234,
 0.6009195013241954,0.9154085765742671,0.40682397282525284,0.2965412303512105,
 0.3683420236863071,0.09482375380329167,0.8594501867195088,
 0.09737663079507641,0.8777452484417428,0.4854212415635768,0.7787033806132884,
 0.621062797501433,0.813356739336276,0.7953722702933386,0.5544513964201955,
 0.3032061645333143,0.6045959335482377,0.4526829042304159,0.5075667578601241,
 0.8368053824139202,0.4912335942795695,0.34995727321065695,
 0.16651248799240492,0.04187919119871919,0.005723218242469574,
 0.5366190327385068,0.7756047752863463,0.01808532366885296,0.8910291122048133,
 0.024559805004132795,0.5427441896643241,0.2721911260476595,
 0.9665308818865286,0.7405786854979444,0.5925252742240817,0.6359126589736689,
 0.8797312236528313,0.3929322110414827,0.5446043516423387,0.4573768137518368,
 0.5369114898580302,0.3583913605695881,0.2924771765900164,0.8007279839883238,
 0.45786417595651796,0.06923942279705542,0.5540215189787634,
 0.6549912557816733,0.29585673140892044,0.7774595212756127,0.5760245660993017,
 0.4160967948363127,0.05191939675335655,0.4388067509130553,
 0.06359005155980646,0.5629451624523845,0.7660251116872003,0.5613305197633413,
 0.9146074444347718,0.5262021128381718,0.8230660219198623,0.6189270641994667,
 0.9810435553189008,0.5356944864753347,0.03787002476060475,0.7182912250492889,
 0.25969463857299147,0.005951516204867668,0.18134600191947758,
 0.0011559531133910061,0.7404160731463446,0.8158763247604509,
 0.6193722144931518,0.6293014578320884,0.7667957646946497,0.9852408027020241,
 0.6398730216189636,0.590077071477318,0.14669464525791942,0.4034601346117248,
 0.7412148692588264,0.381984346049829,0.5383828397227087,0.31452661220424083,
 0.42149769899742073,0.6290248916134252,0.2852222952528596,
 0.36026993258284434,0.866220018385303,0.1364329168872337,0.38279466105935045,
 0.29488804347404374,0.06525392386796436,0.6759146866982324,
 0.9679799288144751,0.7707198773771022,0.6614779606709509,0.6863085317962694,
 0.028750040713204306,0.6176959922381652,0.38658288601053226,
 0.20163191976467276,0.41838445195860363,0.9230731046499259,
 0.5085575724222785,0.5346780814200569,0.8421985937404589,0.8486922023122703,
 0.9725777734138308,0.94406754835837,0.2228060929356177,0.5062493235978068,
 0.4700152484437301,0.8174449903820269,0.5246180200657955,0.6059241955848446,
 0.8849489215700346,0.8178737976391162,0.269377085217988,0.23356823983014896,
 0.9642813693096258,0.8843889152054235,0.21341491902505916,0.818958571489308,
 0.737664424470595,0.025220537755218264,0.08923948386906366,
 0.05608749471843977,0.457062500936938,0.7766774693736916,0.5437512776778527,
 0.12595823910193427,0.32011193692528617,0.46821753056560134,
 0.8795533435045467,0.2592617468607864,0.5613473042069612,0.9651696467168865,
 0.4565698772830814,0.555504627481105,0.4939717753693005,0.7395652694809032,
 0.7684331128302544,0.015123304738236087,0.4857347161604664,
 0.10119472775519678,0.2285761975955145,0.9312303881043753,0.6226502069141927,
 0.5672713102101727,0.9644825847534985,0.4885036224370849,0.8548512487771569,
 0.3027005835003278,0.9303978844111643,0.6013880276514753,0.4798528964891886,
 0.21262709367697608,0.14493574504841622,0.9276864592762835,
 0.6524623560784011,0.41045005827196734,0.7851622248944692,0.6716621135779984,
 0.08514414158467787,0.12312681948282067,0.7804935494272979,
 0.35459758904069516,0.7993965865427956,0.7158430163952404,0.9614200993315963,
 0.6120844624339972,0.49272106419568085,0.901378989756961,0.8138476163343036,
 0.9706617758414328,0.9185120270376165,0.37354400453473846,0.4712437561401279,
 0.6804567058231427,0.908256260920151,0.37891311231991986,0.47644994949179276,
 0.7599119547124958,0.8132752716989342,0.534816055535034,0.520831115342948,
 0.9414067980995575,0.6703273520094652,0.3555831762279211,0.08962388784261754,
 0.25023320213616296,0.94829367077255,0.029750983912098405,
 0.48794643801132864,0.46321386521752206,0.9866200854482771,
 0.3555550904782696,0.9854992457045579,0.25676456856616603,
 0.28592744935684367,0.2795912818690698,0.24367247386338742,
 0.8452792309761916,0.7636158174265851,0.14164778940185818,
 0.33439471522152886,0.678589199927693,0.615229040188063,0.17653550025155207,
 0.4016682960529714,0.6012309822518801,0.39212074898668814,
 0.19935030692828148,0.053780595326686664,0.17758598964813377,
 0.7141391677773901,0.8387038466170318,0.5341060674371697,0.833312726896813,
 0.9487031992589349,0.7413620847251134,0.37807606799011484,0.8735396340998892,
 0.47345265531418423,0.060367755345504315,0.4795221751581784,
 0.2243861590741646,0.42435517005966794,0.2533548015608348,0.6407088740955079,
 0.6576669113669735,0.1761839467754429,0.4038005459011902,0.06793674684990736,
 0.1466864550487238,0.2662381056529751,0.39016149920013854,0.7519437863432956,
 0.08822682309328211,0.708430411187649,0.0698399456395149,0.3852652648890259,
 0.3498891443381148,0.7322030818944568,0.9200310422598929,0.7156593603320182,
 0.7103957666230842,0.44549819497789445,0.5749525795277798,0.7174690323260853,
 0.392294622971862,0.7827867961445407,0.5886862075223153,0.35436865522285643,
 0.7305575318045701,0.977437658356553,0.28277730100571985,0.7441506357496779,
 0.4862841991028619,0.3453083539315259,0.7803367942308933,0.8058579675811596,
 0.3339835487313162,0.9193821468977948,0.48822942518874146,
 0.036130243763346614,0.7853935379042148,0.20199332272328774,
 0.241191664465494,0.42757449306258566,0.5985913685482364,0.4965068153082792,
 0.001635147727845232,0.9418905578164407,0.7029335993808301,
 0.3513845272523277,0.29048201593417644,0.3338260658066514,0.9487519541498568,
 0.27597199798906136,0.8842606576086437,0.4231944520711409,0.2117586559645065,
 0.6051818354764427,0.9429646739921778,0.8267064765339015,0.935351801712142,
 0.6014179772307554,0.22903466995671784,0.9307737897761298,
 0.24420876382493206,0.05979194403129651,0.5792338318730769,
 0.4907160928930022,0.6489553689375408,0.5487382885391823,0.5165745522100117,
 0.3915350676442034,0.15958469097408046,0.9845159071764513,0.3130549917737694,
 0.6042653463431757,0.3182126761915377,0.4588286866519089,0.26386956769422365,
 0.6003843128857942,0.4513710991882448,0.025233168621776714,
 0.5109228696888303,0.5353032317526645,0.9137593393467265,0.48235651273591285,
 0.6548263187423489,0.0995589450726988,0.38754872052169664,0.7507205802945347,
 0.6537131213115672,0.03200737687105826,0.3225427212272828,0.4614140643917817,
 0.5820819598827695,0.9301872925445398,0.896846385062306,0.1839346537207841,
 0.7893306541137634,0.9782256704286989,0.47820600601056307,0.7085851592403553,
 0.8474296947090078,0.9628120546268104,0.5074860331063638,0.4996495640429627,
 0.892972078717263,0.9591594395338675,0.7157166989658906,0.5568554623871923,
 0.3193026515358943,0.6866177932829473,0.021546969956666495,
 0.3538252053949389,0.6045389294635696,0.37616884125286265,0.6339708421777014,
 0.553697851009392,0.44194543711200296,0.23602093984511463,
 0.05919110059547705,0.8006927389384565,0.680066616269362,0.5978327291752221,
 0.48424173132368764,0.4200561063432222,0.9453561538664398,
 0.02524193770605454,0.3590661812791146,0.8348732604462747,
 0.15982111711872693,0.6425727573552744,0.41153973254536014,
 0.5142438931721384,0.7769572725616629,0.3891331946761586,0.6241706800705255,
 0.9429203509265084,0.5390315757912233,0.061933954749711306,
 0.8820871234612471,0.014145524601968051,0.3817570471707211,
 0.14536904132012196,0.8151094140710673,0.12402304376601836,
 0.5295198182131526,0.6574891818890336,0.23783867527041203,0.8485564680005306,
 0.8988665283578282,0.3796478478459774,0.04653867279711721,0.8370385911719418,
 0.7272900473023275,0.5353341144094559,0.66729487441313,0.9261548254268471,
 0.4534943237750848,0.20198622386010434,0.3769614644232916,
 0.11618622276296875,0.7004002553515573,0.04262096659230985,
 0.35564780690551423,0.7215690102201784,0.7883126134830453,0.6530210681367951,
 0.6833077076825292,0.08179693219707407,0.5582755184599483,
 0.48416544743986667,0.19620522010847874,0.8060144505316046,
 0.36047508241977977,0.9923351575120207,0.6517776254371311,0.7491589488329473,
 0.7916751939187281,0.4248750862551085,0.02369552535357522,0.7340927006801635,
 0.8434511883564614,0.2023158090093865,0.7661902417181269,0.38163245531485024,
 0.5992706239835708,0.15581494002726415,0.8478309991928399,
 0.38782997014292453,0.26640655760866117,0.679701630052023,
 0.09089532481839879,0.2741682741200645,0.9561027405886621,0.9791040877398886,
 0.30774123091424355,0.6170202369352147,0.6142046738033248,0.8829954093412425,
 0.23156747721222282,0.9615465541144171,0.7461789902039445,0.7670234160652328,
 0.94355060557203,0.8055441504514917,0.9699777988155416,0.40607593282722065,
 0.027347381011295635,0.9902682237315005,0.7790305470902922,
 0.47208260659401946,0.23816821420001721,0.5090262958178038,
 0.7192931102354249,0.08394266749548918,0.6542760791739268,
 0.22061055769261273,0.7656314462070055,0.7208054017394658,0.7938647602175101,
 0.2049263997141767,0.07669049927789096,0.12480262781690976,
 0.9529972231259064,0.7440593493987209,0.8610549764859808,0.19642482956945584,
 0.5178319641470099,0.5494478840987495,0.6996995145081166,0.7361677800436566,
 0.9894876404192883,0.037564756454350534,0.13993202582375108,
 0.4485412112873769,0.3479120406302958,0.6638214015981606,0.01753646795220265,
 0.5541071026901485,0.05507409127564655,0.5878040543492998,0.6586244402881241,
 0.03301296952169597,0.8176599996016275,0.7279159269329463,
 0.017246496557874025,0.5944233013422701,0.26806013509400484,
 0.5132640582320656,0.42232324146426103,0.46344780017268716,
 0.18034619793208628,0.8274336902232153,0.7351948359161735,
 0.21763160416424698,0.6862903853218986,0.790797008435854,0.8501861124512564,
 0.40322654140118286,0.6341334123458099,0.36691649303074036,
 0.5609565710146758,0.4422521355485213,0.0930999840059713,0.06639369705529696,
 0.5564637791320006,0.21167265881169717,0.6245266481475146,0.834844639358755,
 0.8046374038681444,0.6890142436131179,0.9217219592872876,0.27177951023394975,
 0.10068419489229763,0.8786795052945653,0.8149769493023794,0.5487629115816273,
 0.9565174073777155,0.18513994934511913,0.7719225250048689,0.3458207034121231,
 0.42176153495683844,0.31662555037070805,0.3708095912355538,
 0.43524584725555804,0.4812048015113668,0.3864859512996778,
 0.21911710369692272,0.6536962608992988,0.31045678422441736,
 0.37996635135237944,0.6310991168561464,0.16080321878216608,0.868956800340286,
 0.8233302618265294,0.9080569854033593,0.18552554686404343,
 0.09703842112030892,0.40086641516316057,0.5689756706693759,
 0.26099303730638745,0.7874014100088161,0.4999831754675552,
 0.23937500503046105,0.5428595180719565,0.46685028563532205,
 0.6279726941668876,0.5825903459173355,0.12467684641202781,0.7182404095605837,
 0.5272758533310162,0.19470510389509632,0.1683283971290913,0.3780522953675389,
 0.1149373071497144,0.3165057659753512,0.4065420062006684,0.8262941337523672,
 0.17765438988554483,0.11106496959762124,0.4991866560567555,
 0.07739581035990473,0.6250782667573969,0.9493145585006584,0.6026952190700623,
 0.45002067455889505,0.2325659856581418,0.8806289434259826,0.8139980764596189,
 0.36027403797205826,0.22593496021145087,0.6251343919857977,
 0.7656867906701836,0.9487972734562609,0.46376977273554376,0.596175345413906,
 0.13997712500619397,0.7422298233497486,0.9287213522891887,0.840155864596781,
 0.06703870749149987,0.746748193035812,0.8740697301625402,0.7203819314089396,
 0.41541876080265716,0.2970714773692259,0.7150181150248257,0.8499776870759898,
 0.9345832705925695,0.01213209949281957,0.06985266245835953,
 0.39832013650045517,0.41624736642526683,0.680788180319353,0.7759218299207915,
 0.09946665209712047,0.6991718739016388,0.9499892675364198,
 0.21121769226806297,0.17952383673564531,0.9783708764875229,
 0.14713383089817156,0.6882908031052853,0.41484592818341315,
 0.06495736582763922,0.15431765546676313,0.5944209333764909,
 0.25966778178100014,0.5223164699282532,0.40748232056323563,
 0.40300065467668156,0.8826585245336965,0.5294768137147847,
 0.35005931186031103,0.11326424796560897,0.22920739924542077,
 0.1643140242566392,0.6692213060854288,0.6625384786834154,0.4851920087281145,
 0.40975423997287574,0.27700857854371064,0.8068808102520492,
 0.7598877189674416,0.23889682763671127,0.7155088394841029,0.9470031669598356,
 0.2873661016447807,0.5121242878000658,0.5333900181800445,0.629753556265375,
 0.6838651499256848,0.9507059233042382,0.9598554307823857,0.8223518241102935,
 0.20392773518972762,0.3964323851339432,0.9845518622677462,0.6471347872959146,
 0.035185793028785106,0.4366796346244394,0.5927716653685449,
 0.027606605342003765,0.7916235152964872,0.5328726094128053,
 0.3987943849791198,0.10662589874439365,0.5252345100687664,0.8467316483264771,
 0.4900205262942491,0.8958611038603093,0.5226874942799965,0.6322308950014731,
 0.6774654533524507,0.25893413848167224,0.33875096780382863,
 0.19127357191790384,0.7573052082521958,0.427149975044002,0.5393864517873038,
 0.85517688838817,0.566953537100968,0.3976236031245507,0.13084235987783066,
 0.8955504518022321,0.298502572390306,0.6023280389308253,0.9756505956596426,
 0.6462650879443524,0.20086196483544128,0.46625966541581443,
 0.7479975315809378,0.6749680264336021,0.9852589472760765,0.9242917978610737,
 0.8476466333831841,0.4436475133432578,0.6860010116846598,0.34191120960390875,
 0.992616764650277,0.9619925532241197,0.7990050207248844,0.5697361149842693,
 0.6849376113402581,0.7316485418456244,0.8347631790914455,0.39061698470065853,
 0.5195833704137285,0.942482746855282,0.005662131297478723,0.2157366002691904,
 0.42541000819393915,0.5229662567452418,0.7935989397031789,0.2916378470682697,
 0.7823698647669088,0.416727668697114,0.6344517229140197,0.8048218912733607,
 0.7961632250547719,0.6888742012244357,0.5534194704718267,0.15208019584028343,
 0.8014006792978697,0.8669150893284867,0.008616360546882973,
 0.49392721955758134,0.5467185186274008,0.13893625116348607,
 0.3994620306813983,0.31155614520513775,0.18165418011911816,
 0.8862203539421525,0.9700756385618639,0.31933417412261644,
 0.03948313109361101,0.00476300214153258,0.7630588630443222,
 0.7124081200910497,0.5052954648972456,0.5825672010087534,0.4273443852265236,
 0.4902811717830754,0.3825534303197724,0.9443555998899182,0.8015197295946417,
 0.19259704911489695,0.5207320720993158,0.059986561810531525,
 0.7927847235962326,0.6557279698428518,0.9757210144987547,0.2914444138229646,
 0.9923987895658771,0.12671470625474412,0.6814683681279944,
 0.13957539508798333,0.06513297062647538,0.773800508000313,0.9045778414602522,
 0.06300369660485416,0.16773048984144812,0.9518596997848613,
 0.6657482560116872,0.7309013599236817,0.9172989493286565,0.10264567804961189,
 0.6987740101190687,0.12976387472855366,0.8296643864044753,0.5956300074375069,
 0.2933081717885646,0.7205782668502093,0.5350555451270946,0.19303103961750767,
 0.7091512567469952,0.7836374589394534,0.4368218314167248,0.4788674311713951,
 0.3238599869851079,0.532068955474254,0.9087223184514701,0.5872062385382679,
 0.15701539393607788,0.3965386907438315,0.30090517727577193,
 0.4613316740832675,0.9303260717927223,0.5233823855719166,0.9155764302420639,
 0.8273211738645302,0.9633126126208833,0.8535036309703448,0.8517419819069221,
 0.6855435983324476,0.6831703024149232,0.005647163859174964,
 0.5905189159065602,0.4237138490602316,0.308984850808082,0.30829653478063823,
 0.33847886838101193,0.434546110817172,0.8647360459950403,0.394263220996931,
 0.9382862033138422,0.5091708672906631,0.5550991511036684,0.43925650198463473,
 0.9564898236491127,0.7478555735999161,0.4419843109989158,0.5563376247033401,
 0.6422112481344932,0.8789290705900156,0.5484949404257607,0.7087947242102288,
 0.22586459253401858,0.6089854048590494,0.27027176676878273,
 0.9455091000718535,0.6760300155765908,0.7293543715855841,0.8900517655414875,
 0.4944394928865611,0.4487131595632108,0.22048972589080884,0.2677369481694545,
 0.7013211146121217,0.2401240638294979,0.2806053442910863,0.9959596078250919,
 0.8440801725277731,0.6495091840123225,0.7962332145415034,0.11833868051904317,
 0.145414116292925,0.5867062599126666,0.41382915562201295,0.6196375045118125,
 0.023767851841387633,0.21814607931263474,0.04591559748126295,
 0.5739852759532841,0.8991261469632674,0.1976380781627235,0.8724218227043534,
 0.37618694526399066,0.7998681548315942,0.9209211300793246,
 0.23252033164859454,0.7358143974209524,0.032506464277470726,
 0.8793950566844424,0.32138259977768857,0.7191701184386863,0.5674686265864579,
 0.11672903928543943,0.6466648007141783,0.931965666540207,0.4276812817088974,
 0.04580513631884442,0.7536712911966564,0.07466450187524343,
 0.5682756097634521,0.7010012614841871,0.17628825645810153,
 0.11695239172113614,0.5052536444991749,0.48090507648678193,
 0.2715722069262443,0.15945920651699785,0.9763480237324101,
 0.21521905774878125,0.5834725738346981,0.35905561041945966,
 0.6583010035139834,0.07426095641710928,0.5441735322233304,
 0.45123160285096686,0.9136611131225935,0.8569441182809133,0.1475345070400309,
 0.09283359681754089,0.02217894986611546,0.13234204569176544,
 0.412990729874664,0.4958417787768986,0.49731445048901535,0.22758563864736914,
 0.1592859180784868,0.632717455943207,0.047382309720710225,0.7924370157216372,
 0.423650227087605,0.2960872374195278,0.8671406733602756,0.9352151584039359,
 0.817036688939868,0.5016522320813338,0.12402550370366294,0.1799962089036371,
 0.0052024899049342155,0.5369048348027929,0.4005469749598364,
 0.900061651623016,0.8737641373382675,0.44963403595237483,0.9731687667611927,
 0.7539893491334759,0.24700145845811328,0.11148322803625532,
 0.19024422287635612,0.5817498418863254,0.7604939888469051,0.7383719218053532,
 0.4018842776956689,0.7732761804304927,0.4943648708190377,0.9605686950278101,
 0.2473093976851909,0.09872501715916593,0.07569591913727669,
 0.7192812542335645,0.732909651151609,0.3007298922732645,0.09695985704524579,
 0.7286460950360861,0.09675311459803304,0.9922348112595346,
 0.22369682902760168,0.6837896266619181,0.45921736899971466,
 0.4831221849649533,0.6066842099612729,0.6069552081664773,0.4532639954162567,
 0.16009412229963194,0.3430931717392117,0.6110255298435076,0.5902522725674844,
 0.8375043986180952,0.6116922586729447,0.44700322565529504,0.993571925418425,
 0.26818715154397754,0.1845661821153185,0.9850815524305064,0.6481498136878111,
 0.03547099214110683,0.9942019143748857,0.933669792556209,0.7600796620766677,
 0.2628901639631527,0.31340315349055103,0.09472923819870738,
 0.49298418336743133,0.013010919748018157,0.02271118011115647,
 0.6556920075763529,0.9915114700774025,0.7474692205025203,0.8363492088619057,
 0.7939195033258288,0.7293627501646621,0.40865958628467336,0.3785727491034402,
 0.32404637147755677,0.7087466018630825,0.2364155239790262,
 0.30521254261581277,0.2501291875936553,0.0537130705063138,0.2561223581495914,
 0.2879378373235981,0.26936317368393914,0.01836820222613622,
 0.9573822903318848,0.22640387505657933,0.7237043367987149,0.3714789294631783,
 0.7064572428255286,0.8590813981342834,0.020632470145703907,
 0.20544455933292594,0.33517437523495197,0.9633553016516826,
 0.005721491568753789,0.7000336356445402,0.582485749032757,0.7801910741350345,
 0.9417198521534532,0.29878090660078904,0.7109901052270435,
 0.40256436350632274,0.028930084350231144,0.09256753586025379,
 0.8012265907516528,0.8342987426930428,0.8244980823024537,0.38891079797630845,
 0.2640494419165412,0.9331663252190133,0.9461683398534273,0.537112024462929,
 0.9715627356069272,0.3430721136353564,0.3825223258523446,0.1710878193908285,
 0.9958648300816622,0.6153237089248939,0.8578570047757273,0.514622677316384,
 0.6703436718331113,0.5533543218931856,0.39876092918305694,
 0.048530235948293754,0.8377466678286997,0.05596122435714379,
 0.12084667738738242,0.5863224175870435,0.887162234577722,0.1922630763019635,
 0.8118568199238592,0.21919596331273294,0.3164363537734505,
 0.13301739286215253,0.03483174362745567,0.2879888669402929,
 0.3218988430859894,0.013923704078244392,0.22605634169276523,
 0.36103741123978417,0.33836082889750085,0.8927248185103243,
 0.19143182776038203,0.9506187711410852,0.2999254258108851,
 0.20905861846421625,0.9731404949969377,0.5195240952231889,
 0.31537666022856414,0.7142593615321204,0.7702606414337958,
 0.009837745018678845,0.028062507326069297,0.6224300594658284,
 0.12060051999121357,0.17670682846843433,0.5934537208285302,0.303460009090462,
 0.4965294543225611,0.17641706044780237,0.7007355636171941,0.1555515125078829,
 0.5869513930779773,0.973484872920195,0.3636085140205123,0.4260207132094562,
 0.6246028245994636,0.19590104233474026,0.2732455524394335,0.43168663465347,
 0.5225421598909233,0.875067544827564,0.07658741552313963,0.5127136124695623,
 0.576652185908877,0.13303139722746993,0.11633496713448466,0.8130586143729517,
 0.5466157630076329,0.055251974069344456,0.530654011976847,0.8886555553271851,
 0.9193795743536716,0.6271121262520666,0.11916667864711772,
 0.18364115761293542,0.6870853817990967,0.09333681608500999,
 0.8013047477722874,0.5603337917584416,0.8745256319336782,0.610181767662833,
 0.6016542045417064,0.2955707714629098,0.04850468303693456,
 0.37963027795974436,0.8595621212807427,0.2724826994574171,0.3112122248850089,
 0.5185952830159081,0.5971371217973521,0.1157314366264951,0.3466982682272395,
 0.772629051737186,0.922033311507256,0.716328507077922,0.6797217688190828,
 0.6395811626502643,0.2141562949325968,0.18023724796346152,0.6511718999337902,
 0.7493282492714808,0.4941196318776748,0.09176944517418673,0.6006601993052735,
 0.119177239691453,0.30635558390114936,0.3506494042389543,0.20186888188928098,
 0.40047430029882036,0.789658078900098,0.6913833725862968,0.4147962326435123,
 0.7493177904062357,0.0667202576144964,0.09255017812571553,0.9565906358125472,
 0.7838366001857262,0.39469127477694466,0.7655357865970965,0.6353930857306571,
 0.8356961377222994,0.910546884810126,0.6505478476472761,0.9580301833214289,
 0.1357969533115796,0.08488763868851246,0.0349876605761289,0.8815405858691604,
 0.01875387992780042,0.7548590128387929,0.6567987529990977,
 0.26574694555747247,0.1864411932771356,0.03306628812865442,0.687436577937082,
 0.1708790261417089,0.2075082384253366,0.2774235096612844,0.38600350721480514,
 0.5468427033341552,0.4431481565514459,0.5091746666556298,0.5089113083967822,
 0.9043108537109616,0.6325200114467363,0.8119584471958334,0.5586353692181382,
 0.5414612573601543,0.2188031229469879,0.9394410729796806,0.3930606742985767,
 0.873161904065503,0.6506987165768802,0.9433553212819038,0.28978244661492936,
 0.4960916846431349,0.4432255078894366,0.9242411426000922,0.992834246620289,
 0.5429567430112168,0.8756040104546416,0.5357987745297943,0.5822272902998571,
 0.5311267897228288,0.8852418493782602,0.3022405775151994,0.066854635418378,
 0.03236160760384654,0.06466232493959523,0.09660989861763536,
 0.8604419597286731,0.36605758667299537,0.6171825891342221,0.2209862012134507,
 0.34303344178276673,0.4812963962587927,0.498384080861239,0.25193989499540925,
 0.2430749450376457,0.28053752084854533,0.07704292302983795,
 0.7175618073798709,0.06215103186186832,0.6664689882821235,0.8994622384010167,
 0.0016820560988062283,0.8596391431345414,0.46206549984805234,
 0.9755433958834903,0.2970662926289638,0.4284802793711968,0.20283327404002494,
 0.23891811602112434,0.6555322843810206,0.8148583876071159,0.5469889384549315,
 0.9423682965396109,0.5602687174201235,0.924717081729842,0.3414218434450377,
 0.884315093475364,0.9888239258982018,0.5192657366924723,0.8362289548948477,
 0.4315480943573843,0.9855260537511282,0.24738893303605436,0.6496423467343176,
 0.5642361176651217,0.0935460560787944,0.5276335087271177,0.2865343006441302,
 0.8082352276929755,0.510527388125649,0.6595524091062739,0.7225462935606055,
 0.197503388231004,0.4845042099240061,0.7767183083539149,0.02778786686152468,
 0.24143659115123894,0.5719900242978209,0.008038813823966096,
 0.8217526059219109,0.4144857559941775,0.9175506763476583,0.1659292093238557,
 0.5411960639474217,0.6440014909056496,0.7467111105915596,0.8095922591261975,
 0.7095848200843481,0.3061326636264856,0.09786759405154566,0.7251378973542617,
 0.3835108545036311,0.26043140472525117,0.5696185138403859,0.9866118428584395,
 0.3975557174531641,0.7990012645610416,0.5857259847651105,0.5748120922185572,
 0.07701535905738499,0.7694486887095727,0.5353160414208048,0.9716504473966491,
 0.3669715125386128,0.9141589763732907,0.8286983003247207,0.4590568298236286,
 0.9735683049487337,0.48311551927088536,0.5782758661740898,
 0.12801026290770268,0.5155093984787235,0.1308355482162501,
 0.08833470809633481,0.8969978007448738,0.22968376696500625,
 0.27485978708092207,0.34866923382324855,0.038404458706063904,
 0.031149130464770525,0.8551391305554802,0.5664402643939397,
 0.5138745107816259,0.1830235178666394,0.46516126137175984,0.9746412609674011,
 0.4836853844473925,0.7275602365965912,0.858950395807817,0.14592937485960045,
 0.27743622963540493,0.058966805161586144,0.07800021005767388,
 0.6563456995466179,0.20069643761089806,0.5592743743024812,0.9812960970972493,
 0.04235353602281955,0.5514815125876864,0.5859104313134524,
 0.039281002351201244,0.2624305008703611,0.026779013458878453,
 0.7342460702889241,0.4426411971208838,0.7532274315897991,0.19895148248621264,
 0.15409575684648025,0.6000344043655405,0.5305696429591564,0.7461035289139013,
 0.667083211706073,0.5676429732102019,0.8773535968536401,0.4234650055090363,
 0.7769841674271242,0.6296423352531975,0.09414766476817793,0.674757022247414,
 0.17610492668933442,0.8179102985253685,0.4572486666890887,0.5919004567771905,
 0.20032542297543177,0.19597301257243271,0.28005895480006693,
 0.3825419169243751,0.8203932673966797,0.5483364866725998,0.3583800438270468,
 0.5748885428618827,0.124024760509441,0.31489859320422153,0.6222564805079327,
 0.004615213696312126,0.21605224179253946,0.4999060317290319,
 0.9238681086389154,0.4486629415177097,0.5281592170431314,0.7225957783906272,
 0.6469577568399683,0.2769848169272835,0.8514872162097263,0.22933035698603932,
 0.10248691085762329,0.23246944068344633,0.3675073378422371,
 0.6561515933402835,0.6879656023122064,0.4162738771783123,0.3382736071327752,
 0.10732409609128646,0.7628331242621678,0.23292255399053352,
 0.6145458419842728,0.3595966122428401,0.817026225195944,0.427834231175227,
 0.02110695449126665,0.512715197316932,0.6521613854817445,0.04467539627274031,
 0.9490424978662936,0.13774455277017328,0.9301183670398503,
 0.48322914968509767,0.7145450858613911,0.14167681161222556,
 0.30204335050692854,0.7824090222078327,0.6241110533631173,
 0.15017817989355353,0.8418627553628162,0.16403916514157924,
 0.027532719996970267,0.401239266776095,0.2009831184523645,0.733231566553768,
 0.202235701166263,0.34830508700858054,0.5183034063261769,0.9222458396593115,
 0.4016152621006357,0.6711876363779347,0.675007997682564,0.7859940715210522,
 0.8143141663454841,0.09803708358956709,0.48638563574934246,
 0.7438479403694298,0.8880753161642652,0.5368564955391069,0.8676110601001608,
 0.879320848848152,0.72428746893165,0.4966816040078599,0.39354547099022674,
 0.2860812765760897,0.5997818447075661,0.2521478102308893,0.4286310793601399,
 0.5214254406413223,0.4706746402571066,0.7352624092599095,0.12950626816141197,
 0.5421919305785948,0.44325249247628395,0.8626416658403109,0.5114051363466564,
 0.22893422299072996,0.7202253471559534,0.07016902088821997,
 0.8109661567544741,0.3228969942845892,0.3701735689737641,0.0499151595589542,
 0.21562035937529045,0.0671644764674697,0.37222872203197066,
 0.4305466228277909,0.928945393699528,0.5230608222664535,0.8543672168716113,
 0.8994496747237198,0.27487914296931126,0.1601464792077506,0.946315509200149,
 0.12924928348534548,0.9448673227009379,0.7620442925080341,
 0.22874045932317533,0.6048842799400261,0.7926723381083954,0.5992137376426979,
 0.7942233640271259,0.9340037261000675,0.7758813563819016,0.28950160035491146,
 0.6606903529310724,0.035358691332042724,0.05485258622754752,
 0.5557228771380106,0.6830425458042257,0.9413742841020276,0.29753598157513494,
 0.08241842547435507,0.37363116627196324,0.16101211378165703,
 0.7510488160066695,0.05000782007312521,0.04249906915953805,0.584047792766747,
 0.01716727602461754,0.7839964871690588,0.8921348695135561,0.7260442375140362,
 0.7725788159465301,0.6619507553907514,0.31740227551200384,
 0.11048733623850426,0.5177406996417615,0.8421132967917055,0.8209850353940168,
 0.49813373879375555,0.8760192297218228,0.8060216262295432,
 0.13524538834881816,0.4896688721501733,0.054227721492513314,
 0.6831490910807491,0.2140310543630789,0.5931246686406447,0.7318920724562215,
 0.04042628899128353,0.3560014236965019,0.8447469793802073,0.8968730382203197,
 0.6692905168863502,0.33295937468413006,0.11790585743954063,
 0.6306952288103045,0.8147747659993505,0.23567376554979302,0.5841514690566647,
 0.7170240316070218,0.27317024781257715,0.8198059729450599,0.4453646754741182,
 0.7704303521326655,0.8059029299428168,0.7921198813258336,0.24052131816764344,
 0.3960966410967537,0.3369656721539853,0.5728839066444846,0.2896638498324555,
 0.08581179424293772,0.8505859332534025,0.7805692772621885,0.5241445971069916,
 0.31784379994417344,0.12292133873338795,0.1075332187637632,
 0.4481380409599788,0.8167043197254495,0.5445881439238037,0.37488755447083844,
 0.6648578780465387,0.5668556369904946,0.7808453355532257,0.45997771115576414,
 0.26598900620782506,0.1204649502617271,0.2180282668130188,
 0.37978271792915774,0.6157779188570414,0.08759365863407109,
 0.24140202886041262,0.42399860298363434,0.20865605669129217,
 0.7160354580886726,0.735538646836956,0.45204813126580257,0.6193540249669552,
 0.46976736480687853,0.4245695938703049,0.00136346983782909,0.515893719878335,
 0.6702251482518338,0.04797315045779349,0.7131125591608888,
 0.17734513747577008,0.9539497933947181,0.026035386091807666,
 0.5800077077271975,0.090655030151096,0.4355503339886895,0.45461790244167843,
 0.14292165537413082,0.5166714497806595,0.9220018969433518,0.6740188633896103,
 0.14190242558755906,0.0697743718856152,0.05294025167338812,
 0.15048989878751196,0.3031437688337607,0.4384063818573084,
 0.006863178167424988,0.3500252226244003,0.798202906815916,0.8084710979190505,
 0.5216194315510185,0.70157544626476,0.8429387221162995,0.011931284081367455,
 0.6955344947290675,0.8256420828882748,0.1874408773527646,0.7047133826634434,
 0.9447555989410663,0.5543725242840931,0.4261593998294666,0.5580145388821969,
 0.3275780016004113,0.5462625340595233,0.3994386886866381,
 0.033756738941511544,0.7756208326308475,0.42052591489165436,
 0.6082138115702467,0.25109477257564516,0.19972949182704702,
 0.11369162097544827,0.5901953655819169,0.2946944613910667,0.8710017217568576,
 0.49583563035619616,0.9019996036795097,0.40782405306202485,
 0.6029352046064196,0.73920921922017,0.8525090346375195,0.573779380350954,
 0.010338013010146363,0.2104121296582777,0.11687773852432348,
 0.1269427958354904,0.3540360152475577,0.7486635239806354,0.43372999811549984,
 0.6283181905772196,0.2355118637693484,0.6613818440764607,0.21840230505477098,
 0.740498333737394,0.8174623425042358,0.3974866248270411,0.30797845379531974,
 0.0033799339368413284,0.8081633539561375,0.5569760631524876,
 0.7993525562260169,0.07743153348482368,0.8960284332369717,0.4243472425129582,
 0.9953367932372033,0.5126025016271959,0.22779189275262657,0.6286899171254612,
 0.7289577718531624,0.9053532757665204,0.04942433835661619,0.8830087915067693,
 0.3343454446853582,0.6256452560163923,0.7177515618154638,0.6277208272487859,
 0.9131821150527462,0.8039147903562991,0.45597616237587324,
 0.10644766723048771,0.8024488916084255,0.8733487232029921,0.5754503515292845,
 0.09787555302538486,0.04821047096309261,0.573041357895671,0.9576887908188613,
 0.23576516228957733,0.5996279838388768,0.1368561693937994,
 0.03887706643173905,0.43537141070211893,0.13916888052989929,
 0.6642305577017479,0.3987272408386492,0.2043043391494367,0.7456372408051666,
 0.9547573107891236,0.9331098147151893,0.9940350848420227,0.3943181142312495,
 0.8344244357326345,0.4908432899303926,0.619676188068947,0.019536464007166687,
 0.11357278014820438,0.9716941044289001,0.8230083340953398,0.7004807267544615,
 0.3015895833020549,0.05240509063084009,0.16390760705592788,
 0.6851301530802445,0.14229187652011488,0.02672204436229253,
 0.028250372635915788,0.7646202515884829,0.8393966130003564,
 0.7630848483529694,0.682012554025185,0.33870138180489795,0.4888499294122731,
 0.971099191803802,0.3044931687072203,0.5073778943853509,0.6326966153851884,
 0.4501363001999146,0.38309732971022115,0.17095857502226475,
 0.2903582884260474,0.06477735618876923,0.5886112584723252,0.8493932134061133,
 0.14080967476615514,0.5027552461194414,0.2927046208401257,0.5203926954774438,
 0.29690907711218173,0.30980543111253767,0.19958899033337174,
 0.3476956453138138,0.2636990912790691,0.20130155296735286,
 0.19793105291656066,0.8540637399945291,0.7238116812513421,
 0.05251806735804021,0.5955769645228127,0.1498296138556191,
 0.010077406457882687,0.2598042340134121,0.45413018001632255,
 0.46680899838268486,0.958922973552124,0.9887365199345096,0.8795492436104778,
 0.30487021212067733,0.11554692736938199,0.8638199127617631,
 0.1796716142051351,0.3121269662975058,0.29826500142697576,0.1855187606959222,
 0.8495144962184402,0.049637879761294634,0.35089646582538403,
 0.6997377603762868,0.6672552571315495,0.7395134675010161,0.6153850695028746,
 0.8364755941749351,0.13241553145111606,0.17640209069643442,
 0.19233571222411605,0.10098846179115406,0.8835015080939694,0.446239847537531,
 0.8387823253916101,0.494628025880507,0.22611773137233193,0.9715988553177338,
 0.30897219010412935,0.7791375106150766,0.414325399082036,0.5867061457678282,
 0.16573904494965275,0.9976771808795095,0.4672307566000362,0.9837226997388931,
 0.7868994681462549,0.787114614199244,0.8073496207514981,0.9914553150729625,
 0.3160546985246834,0.8336245235509279,0.05594857524436547,0.8659198217497527,
 0.40734854010212507,0.13151962982469045,0.5155798328379104,
 0.6081476584051169,0.5777834070839782,0.8131352079068096,0.8275705123178362,
 0.8216632600990714,0.23889467658941765,0.18554168834342832,
 0.03474081120756578,0.7233289591700472,0.20288756738955382,
 0.3957596745086719,0.5985405381196286,0.6929992541989203,0.3218431918085035,
 0.2922470505434631,0.5510672384199707,0.24327480354501052,
 0.41743428786965964,0.7594369980970965,0.5783406764404877,
 0.16872836962620696,0.2459123998582734,0.7676043302551194,0.8553608561321927,
 0.07126616573802502,0.31019795540425676,0.014727945486065597,
 0.3829755649228008,0.5495859265358312,0.8850905104763389,
 0.004140771789130193,0.843784773417904,0.7758558127792826,
 0.35190139884978067,0.09747885309438298,0.6965987023850454,
 0.7818008609520906,0.7888207158607657,0.38034913951036864,0.2572890547907726,
 0.11048061163737843,0.5638356521789379,0.9298586413077232,0.1407551000286189,
 0.8486286745944155,0.9972198935699277,0.7602190741652931,0.5714821979849423,
 0.60804495837086,0.2801859512277073,0.7098764854294991,0.6187823066752379,
 0.12120936326338205,0.36923490289636074,0.0036976707203333348,
 0.10085280237813232,0.7374251013827119,0.0810121870684769,0.6840641367945729,
 0.4929660989699358,0.8369444252862084,0.17243660408712924,0.9580854798740781,
 0.7230738676226969,0.41479167630167546,0.20983525559258642,
 0.09737167756176257,0.9602706750713094,0.9355370113778,0.22164326304451942,
 0.005418852915400563,0.6919326608140256,0.5874917361768338,
 0.6004787923409076,0.4056829434435145,0.3725182866046015,0.8128349043088089,
 0.014839169445953093,0.26121513084784276,0.840931650522211,
 0.05300063774181096,0.14671215111016644,0.7489675760421448,
 0.9076938543309049,0.21114038472038277,0.21955877100531007,
 0.5111437804429305,0.7695781228968003,0.35202348166441466,0.6051165210645347,
 0.40255770588585404,0.27387977263098984,0.9771328866325701,
 0.3239305212060385,0.689017627761861,0.22183180134287184,0.988185431188887,
 0.12904336402221772,0.874850087613551,0.6557837305726428,0.6791744208677071,
 0.4552268144111721,0.2992585757235019,0.7660491408617083,0.12880775473788353,
 0.10204330534805661,0.8442849348773597,0.4458929387697156,0.4604254781855689,
 0.9803466832758189,0.5929830905568024,0.16789618841998954,
 0.42304735365501756,0.28729395991950424,0.4843867740830097,
 0.2931814174806042,0.36561128612150884,0.6101260588664678,0.9002784448728983,
 0.4267603206479008,0.6088198136786115,0.18930084059610253,0.7091016979502255,
 0.7684895774992498,0.10372176800985233,0.23985673089232995,
 0.33382334170629957,0.8375703268396835,0.991275821361165,0.9873680829036402,
 0.19356215955835343,0.38286906631571505,0.6729700667949803,
 0.42193465364030125,0.41747946683387205,0.10246388546602225,
 0.7870800068467008,0.7500969126491005,0.0380360551237271,0.28832290291767326,
 0.1927503372474454,0.286725880921912,0.8237859651180334,0.8799005247007406,
 0.7199563035088038,0.7548744414591041,0.3388711889769266,0.6874606528984486,
 0.9442718860502577,0.9687701159657384,0.5367164638550238,0.2534684041521532,
 0.6716497097982281,0.4962557755037412,0.3101699224246278,0.5269330423425199,
 0.9702311243175282,0.05452551294922536,0.09518532417273251,
 0.9629846873122831,0.7381499587192151,0.27674247651846073,0.3354003581258008,
 0.6531734646504815,0.5053461956472486,0.9424939777332884,0.1597581754192724,
 0.39538196444307094,0.22395206738750872,0.8277630281095881,
 0.4491919131797548,0.6755999065140428,0.014590901198683226,
 0.3197210432177098,0.06459690767310466,0.05796028855953417,
 0.9569853195925913,0.6690867618994665,0.3040557628678555,0.6362578074600579,
 0.3458203231073289,0.6343438147915132,0.8045452295814026,0.05289161594805625,
 0.3313279119424365,0.8253919793026718,0.18490643156245667,0.8304819111469214,
 0.508851355129739,0.5389214649402184,0.18431299639029708,0.1424407114699362,
 0.12680797822623802,0.06438202297709594,0.5991481619929372,
 0.6758154842632543,0.42913022052721117,0.3730490305441643,
 0.37384413819244433,0.8245775826683177,0.7641751760947904,
 0.07264416971071397,0.6701445883745151,0.6970082965019309,
 0.48074417755163046,0.9340157379425036,0.012087260214444284,
 0.2659063392092538,0.5515848487945965,0.7810770465754112,0.34366125124604463,
 0.14973795874681817,0.9493502155073759,0.395573442918174,0.6810098617875948,
 0.8798173303916929,0.8449777897444618,0.3624512595993764,0.5074101509542273,
 0.552943023693526,0.4072742515012704,0.8148295973498616,0.6613118796936075,
 0.44255562460663844,0.45946103765930457,0.7471173583068215,
 0.6428672167859383,0.5699731251769227,0.2817589025813799,0.5518176316794126,
 0.534559845159408,0.8822959367236825,0.25954198141250595,0.2647418676029939,
 0.5443965791237955,0.2124957253132167,0.5941967557877866,0.4098437885582463,
 0.08688786508810764,0.76069141079521,0.9343448608632904,0.8850523104080243,
 0.04193130476097129,0.04547072800920615,0.2352831639522196,
 0.9184747875796075,0.8800697812028813,0.3978439146877313,0.41640732103051314,
 0.39784048816152595,0.3106600088079071,0.2667502886233868,
 0.14960262897319776,0.41255933811245293,0.14046806461578365,
 0.07422364883110166,0.5438762366302746,0.21711048324931725,
 0.8814294582620841,0.2999168873406788,0.9630362452560227,0.9544419941662561,
 0.19836884460488013,0.9888590997528535,0.747208557828135,0.38752259101361697,
 0.3121770948517608,0.19202322793012971,0.4389181289871993,0.5296531578412117,
 0.1875388388909569,0.047233566913640335,0.8154850093737938,
 0.8542903448046605,0.2973197702131224,0.6457371708926428,0.34680750604946176,
 0.064913046845674,0.25099758761395385,0.5553068022310207,0.6214298245216192,
 0.27917329106555955,0.6206205211846814,0.22563805853822272,
 0.39081098561883243,0.7108691646938609,0.25884369120078743,
 0.8190761192287309,0.3886564330568135,0.07712512805796257,
 0.24308161729984878,0.5407690410240586,0.002427456467671507,
 0.9548983701439441,0.2867023767739584,0.49321222112499286,0.7076053794760839,
 0.8176142391221272,0.3043610394265477,0.30058283755225434,
 0.30526494586069464,0.9970800797094154,0.8208002322243365,0.6990183184717075,
 0.5411156291734408,0.2576017289731729,0.7531675928193318,0.11485261287041404,
 0.9813334816054957,0.4636317732472097,0.5562407685672138,0.1767117620611225,
 0.0024656894168582077,0.10157601336269584,0.21894838888952395,
 0.43225794913547244,0.9348709310632344,0.43463631684793635,
 0.45840417824379365,0.7671530891408508,0.8749392904574975,
 0.09456289753593206,0.6109529776873571,0.7598355224770328,
 0.27259946947381664,0.5197925020041061,0.8280425830486043,
 0.32435096446039136,0.9018181447966359,0.6832090331957588,0.8517646610732796,
 0.7887681990024962,0.4186373941864793,0.413569176315319,0.678096225791936,
 0.2565876902869674,0.12785049759298084,0.4715725780126412,
 0.05130958675027131,0.17370034922981148,0.2657927841888674,
 0.1429051895375415,0.983275568894736,0.9921835862090985,0.5446097280409917,
 0.010698973444626447,0.20561197497147266,0.2674246154263942,
 0.11862872727498142,0.29011211493237055,0.7970169269013985,
 0.8179956671112878,0.8342602129734421,0.02142419865906431,0.589073559111051,
 0.9095195775960595,0.7307496749805048,0.4310949017331762,0.8317830581168154,
 0.28133965010093065,0.905018292486641,0.6826327112902425,0.900312856154503,
 0.7574327601147203,0.8233828447767133,0.8979796829254852,0.4606533913874502,
 0.7914697689688115,0.84870092211249,0.8321121762602487,0.8993101955106273,
 0.7748422931097125,0.2545331957928135,0.9339523443839112,0.675623182637554,
 0.5442247711330299,0.6126761371467276,0.4306541919635095,0.6976803213977046,
 0.5499627166604494,0.8078181556466685,0.3804214712075905,0.7822098274852282,
 0.07354692753055747,0.364638612827182,0.4373746280688092,0.7010731683370693,
 0.445124390152815,0.5271804843542027,0.7557588829223617,0.6240439963236063,
 0.6604980380971875,0.4431138114038322,0.7554225750313952,0.34433909389312023,
 0.2513272225825769,0.258632917709652,0.34434925001155314,0.40776725024956406,
 0.11877121065927054,0.39784724299426744,0.6460717805734046,
 0.26224304721034175,0.321529792339044,0.013589548725536305,
 0.011268774616923882,0.6022016633139957,0.8411215577594241,
 0.5822220892559063,0.39384743381349374,0.7020361908339571,0.9077080879546955,
 0.8103828025773562,0.32303498965978816,0.25596823597844853,
 0.7867756715284324,0.36886534584560116,0.867902534569875,0.17130966907282685,
 0.7492850158315215,0.1318211123502484,0.6434033068342586,0.38559040900958685,
 0.3731374838574071,0.3311446592331866,0.20006274735424778,0.9313301976971277,
 0.32801323339611543,0.971672605147304,0.5069943495773227,0.36042585958476714,
 0.09692975212452204,0.7470746521936678,0.15637297312240228,
 0.6264784270513557,0.7500866128461491,0.1591389135380178,0.36876163931036565,
 0.4580231487910461,0.6477627348186551,0.9352408702551478,0.29793233193226754,
 0.24344600058239707,0.7758737420757836,0.12676276879109427,
 0.03170403089579288,0.19260399781083593,0.610171459855463,
 0.26754591526796034,0.5320166015842537,0.3678160807198326,
 0.15759183406209598,0.13343718118049497,0.7349761595702389,
 0.5687183077910603,0.12000329306394597,0.4569275117288808,0.2189774114946459,
 0.6050266818363343,0.5146752445909071,0.3571034606281297,0.21022682068934617,
 0.3743144257376215,0.789655589633769,0.5631446908844108,0.1343541086712956,
 0.1732725310869051,0.4111570246255425,0.2775898437348764,0.14326021940389821,
 0.19824792348983267,0.2786504246623238,0.11716984419486753,
 0.9275364814943694,0.35991108377116054,0.3497360971474448,0.6940792308056171,
 0.9190350937774526,0.7711047923916172,0.9602015198626346,0.6222314025046621,
 0.11990570868684725,0.6764762035839333,0.621020498516854,0.47482454171549504,
 0.8860811514104223,0.30998062702272955,0.3946062062942157,0.24532124810957,
 0.052863823359518336,0.6674688259611473,0.30154625303556803,
 0.8979292178992219,0.0858525861265852,0.8528393223633458,0.5636638427640935,
 0.5202777652952452,0.19224460824614664,0.9540515142439122,
 0.12830922251583576,0.7688442126125568,0.7777360802981629,0.9072635907550703,
 0.4718728667064166,0.2208024911498946,0.7469559714597431,0.7674872156164427,
 0.8184260273196866,0.06640211542314711,0.644620010475966,0.9652821713527522,
 0.7184901060752917,0.12183925791465167,0.19363930159419085,
 0.41194833851084744,0.7497854350762679,0.561432108468044,0.31429924782240537,
 0.8750524764420726,0.3063459715428901,0.7638934277096465,0.9229109536995822,
 0.08066343632869422,0.46028255560140696,0.6235596336542035,
 0.40098785948056825,0.2972133762376683,0.10526951385955496,
 0.8666389366983375,0.8015852289759697,0.26185262862635206,0.9625808695150444,
 0.4141896810481147,0.3133068388315723,0.17060613933148516,0.7954069131022885,
 0.42903261853121377,0.8845930679096986,0.15457080360467979,
 0.7499290709595219,0.7969097726825758,0.8301026089015971,0.08982315786555517,
 0.9429197994157521,0.6114576641986339,0.643174912384464,0.10126414801889905,
 0.29885308266855737,0.3211560682935479,0.7211876667922281,0.6458684696069213,
 0.5009720747812073,0.5619205379421657,0.7187554020075198,0.7124194105779913,
 0.19895095145197506,0.3573565043077265,0.019730977191850352,
 0.4550698884087413,0.7426147774400055,0.24355190210265032,0.280626896380062,
 0.9023212199112077,0.09798774822674605,0.7508893207618151,0.8024801840440009,
 0.1227792868472467,0.34937335930758495,0.2544535915606305,
 0.13935536528926007,0.7851090556914055,0.06939682316943852,
 0.7256550896592368,0.6929118322114978,0.7900766460647813,0.3786885972475895,
 0.6457545909476152,0.9232300952985553,0.36173441845023613,0.7493237830028621,
 0.4454662179823655,0.2680821952492155,0.4020712760699755,0.11578104407728584,
 0.08196559222730204,0.21991333286123838,0.9782560772414698,
 0.0605257859746714,0.8979881996046265,0.49782394985127454,0.685613512379191,
 0.5393880503283629,0.4258051019482312,0.770848100963142,0.23048233397095108,
 0.8684625919146817,0.7073265347358351,0.3666036823736436,0.8112707077355049,
 0.17347835749450025,0.2723449575501935,0.9988727054767824,0.2300733629480579,
 0.813931038393194,0.4012827455241337,0.09636317626336344,0.6049268115793531,
 0.1301378070783198,0.33437494389785494,0.8734470508544909,0.801359505804061,
 0.1499989947930942,0.8521745786644049,0.37824460419990746,0.3118811362111299,
 0.9351855000760737,0.29483843049963254,0.8423378756712939,0.6366739371172275,
 0.3111074370082465,0.30530067905369274,0.48187619427768247,
 0.10648224783367222,0.3218428390703172,0.2598315824615667,0.8472205131198633,
 0.4019536615916778,0.6957241687353536,0.38232373421424193,
 0.012306369873142264,0.7893529547380826,0.23296184436523948,
 0.45258531600365326,0.6129639852814149,0.9527745932569275,0.9095832694284157,
 0.03886027057889618,0.08461064304094124,0.14653940135935617,
 0.6571971018479726,0.029508788419848275,0.2605478941768069,
 0.02261380104419397,0.9478222828261507,0.47868049944242685,
 0.18565095034798795,0.3733669381833805,0.20767753366428687,
 0.7189242357292212,0.670229712216199,0.4120622787695474,0.8102785738963855,
 0.3413715129147461,0.5238530978113574,0.7623056918529529,0.4081543207378433,
 0.12980900001867002,0.7359294133855487,0.8850120096866868,
 0.025244376134303526,0.19730152363764575,0.8214265473238792,
 0.36169409296766775,0.21032772712659564,0.7296190101818575,
 0.12191063754606768,0.8619497110394391,0.6645325633922098,0.889804534820276,
 0.15005598787336727,0.7506430230442745,0.588316032845927,0.4808434685477909,
 0.005066105905802476,0.007366342900391509,0.5434026969962853,
 0.025401114227758814,0.0368484349331728,0.38057687941791507,
 0.10372330320365197,0.4366115670785511,0.47174705843222275,
 0.2300842919502537,0.2556361203275984,0.07277099423294953,0.6016031522628045,
 0.5341730238713164,0.32010722451066764,0.1689335800535411,0.8480298434644553,
 0.8633663589629899,0.2692871848772542,0.9818639592599954,0.3477289939712973,
 0.695109862601894,0.6692293149352085,0.3006830600990864,0.6888713465112434,
 0.3326244399434344,0.13437133582854655,0.12760806967542615,
 0.1787511423765452,0.24802954478460504,0.7331261073505064,0.2634313977675308,
 0.4016104699042232,0.7846063697646081,0.9635779352163498,0.42940752911416347,
 0.07914886019330036,0.6239707280463503,0.2867888077023708,
 0.36556448664528773,0.6280830364213092,0.3523917819458796,0.8860645294378555,
 0.1389585114763181,0.7485594013021184,0.5755401195107611,0.9777507159153559,
 0.19682492981779698,0.6143733491854024,0.31978113856540735,
 0.6113423043596543,0.939190631155737,0.9495985949885933,0.8721670375704074,
 0.8443053396648184,0.4928646668288055,0.500285700451112,0.2812638461383804,
 0.9375339679222306,0.3855176898422299,0.500762492598079,0.1894104885269473,
 0.9258478129994305,0.5880926600363136,0.1452123871245583,0.3688972295922629,
 0.9775260147556766,0.053706491812579804,0.05505377627683161,
 0.42320802450799344,0.9484107682466711,0.9588895442549391,0.3383719421617768,
 0.7708251495981149,0.23142643338965263,0.14170013850446295,
 0.4860991729139972,0.8123271793518716,0.45589487765252645,0.9281204820062661,
 0.376033071593026,0.6458905347847395,0.6226298471258598,0.8328979009423865,
 0.10696420360572445,0.515757258991451,0.78901210422451,0.7335318850677701,
 0.1350054358526026,0.5224250762099638,0.006336349026957122,
 0.7340498584386557,0.6977839554100923,0.727684365052573,0.18183830149633218,
 0.8810430967208561,0.9006030034596323,0.3311193949672263,0.3224184888155446,
 0.30525301415920025,0.9401354376807103,0.0025017273041180177,
 0.5316615305481472,0.8515945655919743,0.3027476582532259,0.25207183820869317,
 0.14937242168890352,0.22931083522346551,0.4213166818711177,
 0.28195709664961455,0.25317426619422256,0.1823477846145658,
 0.9950362531861288,0.6065439895514105,0.14671521614907201,
 0.44611587855484036,0.022532204271181966,0.49775725470232735,
 0.17173023081897942,0.26937475173847014,0.4493598817577861,
 0.02504955067710868,0.02235752573391825,0.9962017293392104,
 0.5932408640854775,0.7718811386716052,0.19311874785768213,0.6993243763907718,
 0.9092104682341596,0.6429765417442876,0.7287345951073639,0.2398129889577394,
 0.23388927406364057,0.13876241075599782,0.4557935443193166,
 0.2995843945282157,0.5855796905155684,0.30679717745567225,
 0.37589496779910603,0.34528762984618255,0.552495119582878,
 0.48641096374028514,0.9552202455994305,0.3123488610669999,0.9264942244344677,
 0.711438740584374,0.8614147556933233,0.3604524124659053,0.3627315520174148,
 0.2937488760243858,0.9252090785832136,0.12815531323107432,0.5912051791475966,
 0.675369033435538,0.004402229732356822,0.9093136772322926,
 0.045492221048250236,0.8934550776599197,0.5281923257228777,
 0.6280482399507032,0.24510878809183667,0.5358013895418511,
 0.24306983474628419,0.7997899503791435,0.06876292374091142,
 0.8164086668868654,0.807595409241496,0.8520893531308801,0.8836260072568669,
 0.2734941521453862,0.370849911573804,0.2913402488643154,0.43044455431060624,
 0.28890127088878703,0.9224316210908791,0.08023737102717166,
 0.33492280134381314,0.2014634390617953,0.16010064883139352,
 0.7725821893130692,0.914447672342116,0.6125650449679793,0.4550599609602576,
 0.8686042613915269,0.39298476137899885,0.1836413599526332,0.5663164874012998,
 0.13509606255602424,0.29796805742603727,0.3210293752404343,
 0.5931151156075913,0.8523413148767561,0.6572151638043645,0.6070811650952513,
 0.1280483624395471,0.9198371948943633,0.503174077193483,0.06666913831019805,
 0.002952522085924336,0.12169529534573409,0.7567697444063712,
 0.606416905388214,0.12271570572034263,0.5789117191577661,0.9708163584177192,
 0.6506313539372791,0.3274091913212279,0.6542871249337996,0.08880780830954371,
 0.41375646682699485,0.6975185903401598,0.06219104403156139,
 0.15976266584410181,0.11460643368436241,0.2770945594200982,
 0.4767972481836924,0.9466091572002496,0.15084381901698674,
 0.35709951641548254,0.43199622023547335,0.5967862726679347,
 0.7182097816081296,0.8314548271996552,0.03748722307258334,0.53301072775897,
 0.049840019814114345,0.97152208790455,0.32563146924728825,0.6282667393950954,
 0.5046821075089541,0.9427361979434571,0.4077010677633707,0.3617123666719738,
 0.1764612896321287,0.5186164155032392,0.20247667925561208,0.8139210330785913,
 0.2725577580862415,0.7493859851549831,0.9698665046032781,0.518719202505543,
 0.20785618376269155,0.5131043173327063,0.4662045311811289,0.7035044709109247,
 0.5514936906161054,0.5948316008335451,0.3497847903786633,0.9826559959125192,
 0.16260105808897019,0.44761295801245093,0.44480498158358817,
 0.8132145068386097,0.9469918700426752,0.7216719906158051,0.2878255987745796,
 0.3161882299277686,0.6887358608491105,0.3692919050528257,0.9874018758084472,
 0.30799484638188235,0.46938267220149166,0.8620940057279417,
 0.8601013328962286,0.0963150599652115,0.09876175087183836,0.387460410555006,
 0.6825715926660589,0.06648702128589945,0.01611593042680748,
 0.37236166362353673,0.6141248219773354,0.1725440089446395,0.7445625418831495,
 0.8334230122122535,0.8320142415388735,0.33031342317539714,0.8532625640694813,
 0.9478867973820115,0.9404459613039697,0.7732402987419292,0.20523454818656606,
 0.3715862527720334,0.1161227876062152,0.2603316970426526,0.8013727619285108,
 0.26147496815574445,0.3361908567102476,0.9053925325575535,0.6469006371251442,
 0.7414534807672122,0.5604615530491677,0.25962501804009497,0.6730866413050809,
 0.6385738510123337,0.9056146054058688,0.37916390374709075,0.7659441931816582,
 0.538612357068818,0.5584445510779943,0.8581366881086454,0.6014628147444983,
 0.7325273246316588,0.15335528664416664,0.4910736044898032,0.6569241881095279,
 0.9842750268826858,0.8572969652413356,0.9678409806870075,0.1829824415934802,
 0.04407799697544168,0.7972356869240999,0.5345340674533338,
 0.05360905903972457,0.9512840334772558,0.520281357049642,0.08111981699641324,
 0.5109523417533809,0.0729452925641827,0.18355333177033528,0.861368462059651,
 0.39650592214676217,0.1010055306592873,0.04804202856813489,
 0.060162872485755914,0.36343544118872284,0.16904303989234748,
 0.542000269657148,0.9020586911783764,0.2231628728116012,0.48384741300400125,
 0.8942254394444317,0.23562168324372168,0.47316242394043484,
 0.5573715977290518,0.9759293330125112,0.48699452173791924,
 0.12342580653442947,0.6673594880075149,0.5178255436705752,0.3601811698250078,
 0.05798020121150882,0.6659110550822889,0.09817204522752598,
 0.8913279409970017,0.305999622432793,0.04045424657925867,
 0.0029695954770470046,0.20275818550010705,0.5563830615345888,
 0.28074325544984846,0.968479443226946,0.15160187073471898,0.865395249381653,
 0.5466579313780622,0.13037194665038954,0.5656950208134397,
 0.13813373898157022,0.47579242235005625,0.5532261427057861,
 0.012149380401313414,0.508473544746685,0.276715426095533,0.38733801969033843,
 0.23538037066368256,0.7443231557735198,0.21523089978559773,
 0.49909194814698754,0.635042030348053,0.5904399960336534,0.9527400462713111,
 0.8665662472168676,0.4197755139432764,0.5380168924269878,0.9486301452028715,
 0.5663158761909858,0.3398145603914593,0.9378806653453852,0.9650713020352549,
 0.6141641698358945,0.42377397183998544,0.14340341212225038,
 0.007300586794762531,0.7965401879758875,0.7720411195492126,
 0.21831821990981515,0.5812195454970406,0.964286567404578,0.775652221031272,
 0.514363536133677,0.1077688375086121,0.8829781394571523,0.4073283336573208,
 7.085290927397736e-4,0.9062612686843451,0.1398505706972304,
 0.44184392926422933,0.037465364822066194,0.6079646806112005,
 0.35229631180375354,0.597978986071773,0.5251224473486331,0.5781327243875245,
 0.12618859260509152,0.10354313374401214,0.4346893318087188,
 0.8971592299678927,0.99997273209283,0.3278575696453859,0.18943184533434265,
 0.9955154382159701,0.33975694892034847,0.059505843679248493,
 0.8444244308435737,0.6666825270838351,0.6035985628405333,0.3143566040061003,
 0.38085089810538886,0.5041736044651882,0.14437324789303574,
 0.13069612532423913,0.48003181438076226,0.9037214204611704,
 0.7767193503621556,0.2604590098783519,0.07277517864953009,0.681243512527954,
 0.5008357517308553,0.7030565559751267,0.006572524300726745,
 0.5327793473310661,0.8969410932036052,0.23947819731520992,0.9322403642204127,
 0.6373164801737993,0.20891435361549615,0.5590444575268345,0.7066538744526684,
 0.5483899584860396,0.0037097117694903092,0.2254119177058005,
 0.9696033940796892,0.48893482159964285,0.7938289886949247,
 0.43510688994208957,0.8587606258739778,0.5090359564476832,0.563391941109056,
 0.9462571146041168,0.4120338279932636,0.6763469816932643,0.8665306840018092,
 0.2541567779617163,0.4334369996049997,0.258208200397533,0.15688303460383346,
 0.547788912177025,0.6763212771835504,0.2640499747472361,0.41774364695868416,
 0.6783678564072454,0.09808552519619074,0.12296311660869663,
 0.7392906521774706,0.9234831703197891,0.6180131972583562,
 0.012422251250692407,0.2565830893518951,0.7071635454926267,
 0.07384227899151719,0.30051883856766637,0.7877394805456722,
 0.6414855607795472,0.44178894999211193,0.7903539325916675,0.9360356907854952,
 0.2826304866842172,0.23806608289592357,0.757510157573505,0.4200315811741441,
 0.6916382059762346,0.5014732295779272,0.39279328229184374,0.9545177760077898,
 0.8403859654529728,0.0371850424983553,0.7017702762127198,0.03573496779024654,
 0.6924530714092338,0.6616444916980203,0.5413689654537314,0.8750751296102245,
 0.9497840670333474,0.42157075417276346,0.9208356193821462,
 0.12094453335032818,0.9876730023157752,0.8189357041894852,0.9451713528502399,
 0.12798295541744864,0.9046443303345629,0.8613021865750771,0.7669601256474563,
 0.8413132710994687,0.696831813863171,0.0231391607413447,0.22922119777159455,
 0.8714512078558619,0.03348558392842205,0.16580263712143284,
 0.3090894475514244,0.06126625598998636,0.2823292275717707,0.9851671199415505,
 0.02298012105821645,0.6743107133832076,0.7688969191082312,0.7100835232934339,
 0.5222474972133222,0.5164970338218305,0.734464810212133,0.6814210465891306,
 0.5239603309210551,0.3816135371536066,0.5239520199509017,0.8574344252808541,
 0.5895435817183045,0.1525655548462187,0.9700999020421406,0.7926652541425142,
 0.9028716517300126,0.2211709595323248,0.37022938642584435,0.9538283013964012,
 0.45521726531299544,0.8584881391948979,0.8792458387180806,0.3859405386782513,
 0.15125692765171217,0.3432189340895452,0.6438766177427985,
 0.029449579744786125,0.2041483935971945,0.7267384068662823,
 0.8421430018536711,0.13540998276882,0.055026210545348175,0.15867762753130565,
 0.605937591910236,0.6394608578797358,0.5703942421238263,0.6205382058112509,
 0.9980107639850391,0.6343844798673322,0.373050693220368,0.1386946011478989,
 0.8274895455771866,0.06843661118709521,0.7573912217378066,
 0.27747693474885626,0.630876613593438,0.17860740238703499,0.4603300309256193,
 0.10797922767931611,0.05338366524138294,0.17949887562111932,
 0.9105747634863087,0.854040310798015,0.04166476214356518,0.288914426069411,
 0.3619321928631779,0.4235924880209646,0.4891438334126339,0.7833467110524021,
 0.2514314337982677,0.40781077856656267,0.14287927137632517,
 0.38149843075920753,0.8198115970479793,0.7535192525331224,0.8447356407115566,
 0.9260236986820971,0.9120411798562542,0.028157376668302092,
 0.13827555511548306,0.35386250495840776,0.15171922879843125,
 0.33733650268596915,0.3376627809601874,0.7414279426737046,0.3234370382553843,
 0.8712553860367337,0.5022131817615619,0.5258435426363679,0.9908038926962852,
 0.676760342256092,0.17808761913540372,0.35435756096162063,0.2214978333870964,
 0.5963362428748306,0.47226904816481685,0.05787281802288158,
 0.6071634731658382,0.4772369012297255,0.9103348669312123,0.4731972189959548,
 0.04000509183036782,0.37777906801791694,0.2002797904161282,
 0.4450018514934996,0.5747355817954783,0.5724640425491474,0.341247632116078,
 0.4607906521014664,0.188795831225536,0.4556066547893506,0.6962792568582524,
 0.3052926658369397,0.7870487537570535,0.5388488666806025,0.9953685807407974,
 0.25345703098181327,0.4229721983331963,0.5531200043627071,
 0.15072264074385233,0.17550079365951698,0.2390042433040359,
 0.07173747430329147,0.3675031862670963,0.7491385934872341,0.4929250241808869,
 0.8574978051669203,0.8818490245745908,0.12240633596094719,
 0.30268934452537366,0.10421847283198749,0.4092881445044352,
 0.7762383437799623,0.7964243196991384,0.2639921058905689,0.8332556250727713,
 0.6415522655014725,0.789233314412309,0.6551051273419866,0.07363529664885371,
 0.15914492277079095,0.43123761761123547,0.6741188002582307,
 0.9007982370927576,0.8011189696920241,0.9532579647124431,0.7746008332537355,
 0.10449794675381996,0.15267013386175354,0.24218797508298928,
 0.7261657358039598,0.426914620797756,0.14247457073088987,0.743100501555281,
 0.40215054549676843,0.055427949057272885,0.24560866885739352,
 0.8823302569507186,0.55195790618611,0.49587367394140425,0.018664659002017192,
 0.21337230057341816,0.25003256429074394,0.3673880684956271,
 0.7803243151297183,0.9889003036316611,0.771915176191922,0.9081373908955436,
 0.5041257159323016,0.8330245805819129,0.8228674297035663,0.8942389119100213,
 0.03516137449679968,0.24063161305391711,0.09190745751263618,
 0.8314260124310205,0.4326937785692432,0.662601232493685,0.5536735382716081,
 0.3251190619022566,0.023214967644647477,0.2135704865787933,0.480210931103368,
 0.8098624105138594,0.9350062475983293,0.24195574135067144,0.9188473234377024,
 0.46878212013633347,0.07648719328098785,0.2313238950110582,
 0.14953395591183827,0.8312673162362222,0.3205628495775499,0.7703242095190421,
 0.531519847026512,0.5549037196727805,0.8909945442887788,0.02155616044972941,
 0.13426974905063083,0.9778629052646206,0.46187592442262837,
 0.7056249758192095,0.442704040841858,0.2399007839067191,0.6025865767068659,
 0.47271979836354283,0.3642069667113579,0.30623744180774326,
 0.3817200869912004,0.20475508886694493,0.8573243893896791,0.1703865838537576,
 0.8366626735857194,0.1106986445639937,0.920136564682105,0.6877307034751357,
 0.6352135999341894,0.07183228663099417,0.3491037497250057,0.6393466419710852,
 0.38566812981098664,0.8511321211027751,0.9328364500888865,0.533831974303034,
 0.4563705135337206,0.10548254343872543,0.758068706162004,0.3279900306347585,
 0.04604748111728152,0.9892633177769313,0.41774303175692296,
 0.4106211415280395,0.7879004524555575,0.16548430414212445,0.579787444300691,
 0.7523697569311589,0.5876314220216662,0.06445361818577433,0.8017431834283802,
 0.37832266388570235,0.35914168207383557,0.3692887885817604,
 0.25904158014965273,0.18771695986878734,0.24966820319895666,
 0.47196815087428845,0.21262976910427633,0.9072031265370117,
 0.4321113684547643,0.5342209064482285,0.5130521325492065,0.1748089723579338,
 0.1780408693586666,0.05489280392143936,0.19423823001973528,
 0.2758715783394369,0.7223460520043519,0.16774659259968971,
 0.21890562833941096,0.22030232096783087,0.677629317963123,
 0.006092691907112813,0.9003477875735818,0.3839035930410505,
 0.8879188504052702,0.43105079784089395,0.6025719390109718,
 0.11669396772844909,0.7032327520332258,0.18205421418948164,
 0.9400354357357802,0.6669377902736793,0.06987000633711904,0.7235611304026469,
 0.2678363230625834,0.030436116695020554,0.18546956972509243,
 0.1835422820411321,0.8255802264756578,0.2589145439484275,0.2796820662613534,
 0.28779361137440873,0.008117958659199065,0.6480960810953439,
 0.3708962841388088,0.296851469962087,0.26521493804127105,0.32293311102019095,
 0.22996292428582232,0.7562206593776308,0.7748642311672116,
 0.14200341145353912,0.375364464593835,0.5815009976598032,0.8890106305765245,
 0.8134312664927307,0.5246818606194822,0.8164204854343514,0.6457478732067843,
 0.03798083731571189,0.2181478412982074,0.8772331630843899,0.8871628586894376,
 0.17244011730288367,0.14468388165882806,0.5905689718667128,
 0.5833179931969996,0.784641997079902,0.07562560617994452,0.453130601918745,
 0.6793424539501365,0.4887991105775875,0.13734355795214193,0.2912418917439933,
 0.9560222920860633,0.3165062625607693,0.1491845585647995,0.4078606537646243,
 0.35204895925912194,0.42913463653843475,0.08987897986750493,
 0.25825464303810464,0.6686055994793805,0.4609645389353453,
 0.45307092835093266,0.8259502900662781,0.7977374340487922,
 0.20734550260242968,0.09053714699164106,0.20905796040383895,
 0.6214667158037982,0.2815960954642843,0.721264631652407,0.12391243289743692,
 0.09214568513700017,0.48042564406826327,0.6499815779496088,
 0.28592385773807594,0.24766909833366224,0.16560722994045296,
 0.7218913575138248,0.3249417146026935,0.11302883641798311,0.424259164594767,
 0.7895044936970907,0.26264443195747356,0.38984652740880155,
 0.16544559992783348,0.5876092606251568,0.58095531379892,0.15855175146299527,
 0.19126508831774802,0.9254630362699605,0.7902335221328579,0.6886289925337399,
 0.3474762197451656,0.6127073315853597,0.09696858748734771,0.5513696824707014,
 0.8191703453395136,0.1270584793198214,0.31430248501260194,
 0.47968612428188284,0.8886070702406672,0.8426984528745616,0.1326529278813986,
 0.9249725660604069,0.9027715607514339,0.9600944705487179,0.2995188661801147,
 0.17678723456102508,0.04775306723765349,0.037221151909579975,
 0.23646878583383202,0.3972557617232366,0.22229255974320916,
 0.7809200088185856,0.22344754661161526,0.6710763130069698,0.7367713774278295,
 0.7382037363369957,0.6907794485039364,0.23941653368001048,0.629702614090981,
 0.3535487147764437,0.7158712811605713,0.5405280462544628,0.7837562685968377,
 0.25884415001692995,0.9584352230623302,0.6596486936391941,0.7780460153307085,
 0.8601586200691914,0.44897292892259144,0.4519921899133267,0.8859363047167167,
 0.36690796801181047,0.6582199998318895,0.7562463477549439,0.9322616704327664,
 0.11021369837060879,0.5291236136765507,0.2728981605114821,
 0.22764019147362735,0.1695440263427228,0.499452578176515,0.9814295056495559,
 0.5257568804248145,0.7116238142524524,0.533672639762023,0.057315205386549994,
 0.9649693151327305,0.316438749600078,0.1310993031635106,0.19824526033137324,
 0.4153511169545383,0.8365127785695186,0.5763986740759802,0.5043420424567981,
 0.3976519819821658,0.7350465919127955,0.02832251344542369,0.7050005680497202,
 0.6852457214379344,0.7232164514481831,0.6274008800485245,0.8973392982994627,
 0.8176977504538657,0.5357554477022402,0.8617637150496111,0.6717291862357958,
 0.6480446654694733,0.024318489674206356,0.0402501799948316,
 0.18625678618252373,0.7995936233923037,0.3960938498060853,
 0.39644955403696847,0.4022555474554885,0.0963427574572363,0.2048862598261223,
 0.8101650852935873,0.007410234298622309,0.3841093527567311,
 0.15739523385390985,0.5937050590627623,0.6274132893064153,0.8090377152996159,
 0.8855980558884073,0.7680959551686625,0.5838836073336334,0.16196147814421125,
 0.05794725274359014,0.5391615636752212,0.5528042555962698,0.6083960312399641,
 0.11635293277987435,0.13810511931216218,0.5199484116121741,
 0.31955902719218177,0.6451997525978812,0.19279808807994325,0.845873561860909,
 0.2620287625417719,0.16911807346647534,0.3416029119831582,0.4458005086961604,
 0.11432281953731671,0.5555227786038552,0.06885118037232596,
 0.7121700461487276,0.35999354290713714,0.559078140716117,0.4662652346468683,
 0.5167824095541578,0.9612541509793479,0.32784613083970415,0.7425917504277015,
 0.3438718841544324,0.028151243178320517,0.7335843589033612,
 0.18156243132540884,0.20615261865227863,0.13704758663071215,
 0.7617243265422056,0.4283809777777542,0.25431868937204305,
 0.34438035497846065,0.2870033508517985,0.11660043424538502,
 0.0834035616010309,0.5752210393156706,0.07307541672252804,
 0.16174832422013496,0.2602176584186716,0.09316516762069482,
 0.9289324594627602,0.9103025560572425,0.09844177797338993,0.9097761110278009,
 0.2610748377714943,0.4523976356563404,0.8900757727634766,0.7933893661580024,
 0.7072549917472066,0.6252324480468721,0.588790578786115,0.6552715000183016,
 0.2953718178048321,0.6811381415746106,0.9350799613566756,0.9098761255855925,
 0.078024952202569,0.9591211624572307,0.7843660836300266,0.09598888293000207,
 0.5858259617329904,0.34083576819493666,0.9898358614039645,0.8500701414065538,
 0.7521060083647679,0.9714711561285619,0.5573744574775863,0.43649215891313964,
 0.8258580463806877,0.5820523657705305,0.07346528044280243,0.881758028788868,
 0.8716375986758264,0.8661840480283909,0.7130971687915015,0.5644194019408817,
 0.9156345756348312,0.15352837739201752,0.3113333975729605,
 0.06316590661598587,0.4141663343000843,0.851859433114436,0.3402739737994944,
 0.6571035045090345,0.3706614979325883,0.8348569349360817,
 0.048899856693898625,0.683136556777397,0.3025464772325277,
 0.41852344760917937,0.10034328120101255,0.40412974192647366,
 0.8825468703277604,0.46237089612324866,0.15179639763144703,
 0.6164398791721774,0.7827499059646367,0.0317801121585759,0.8521865832894219,
 0.8514672540719578,0.6094090732408021,0.41510588173791807,0.9109991704980304,
 0.9134032224428562,0.6159394016533437,0.29418108066438,0.22002652377368792,
 0.0759862305463136,0.3135480543208624,0.7486536502456678,0.7720012662365942,
 0.29133645802156993,0.07041896931355873,0.10728438468200652,
 0.004638112571967001,0.5131422702487177,0.57461399886735,0.3096358479833943,
 0.9891840295844258,0.1358796749090072,0.8074824241482464,0.39263928423832917,
 0.32730133609476886,0.3209821784306487,0.3643291436531183,0.7118721198466995,
 0.1117676631877016,0.3768571168530527,0.49070602085295345,
 0.27549604608856737,0.31813424559636994,0.7178637744088986,
 0.37616751340066035,0.8306199269030081,0.22439139655734586,
 0.21384807451403387,0.23570136446606615,0.5655753587748893,
 0.22984128136193838,0.5018552388670938,0.11471013031039623,
 0.35899679382153715,0.2171517511187102,0.245374021558036,0.8052587264743845,
 0.9343327083576267,0.17511218714132937,0.49275014134410244,
 0.17069700093216145,0.11806824142672645,0.6826558403323215,
 0.39796785456905415,0.26191459618877444,0.6574596883112542,
 0.5957465820433437,0.31790012407046564,0.9692503346768314,0.5623570678927823,
 0.17425334822467675,0.12816933794095886,0.7766361889882148,
 0.42775633689799064,0.04417713748368712,0.8621397781183548,
 0.10208152748239308,0.4265966922483332,0.42760821406169647,
 0.5736644933749936,0.9564742778119232,0.11110942512074695,0.367101161312014,
 0.7031890735799776,0.4318937929223198,0.37359554984173315,
 0.32281303247920845,0.038888396762219446,0.11958601251805379,
 0.5511927853798226,0.4302044208629492,0.5406556587252478,0.5799765645022628,
 0.9472400873804792,0.2947836061483573,0.5278516818599505,0.02832618718838542,
 0.7775804102065595,0.8534546263567195,0.7060243979154194,0.11720196624405266,
 0.005493966749190782,0.814588168945253,0.2439212668459988,
 0.48335547207350216,0.9803548248127867,0.7134156997660984,0.918397817915837,
 0.11594305603359334,0.6137902242817558,0.20584712668726635,
 0.47227352141298906,0.29959704916392993,0.8256581681542816,
 0.5495330468329086,0.4501538791971904,0.6028143240320515,0.14389915246740792,
 0.8680563421094462,0.8959589026648322,0.6627245862669221,0.916624277603325,
 0.3464478998052566,0.6009650818758542,0.5189367330054686,0.19046706093306898,
 0.5927766599252449,0.5222747563875839,0.29181436186913756,0.6523369182889218,
 0.3228774074484976,0.9040805921557219,0.04012412989778247,
 0.048806977645986516,0.11196476012426892,0.7896751163858886,
 0.1333947890061974,0.7438135716501603,0.21707721180608552,0.4931692200249682,
 0.18389339701854068,0.1961564263484168,0.09365647148425593,
 0.18573040925540052,0.3455321847501237,0.2923664141243175,0.87304312970014,
 0.5961406496644279,0.5550323871579078,0.37364120332706086,
 0.09780324537660112,0.9790444076171207,0.14924166609954248,
 0.5582618465647636,0.2338244713289238,0.12220093057387516,0.5061549170430382,
 0.959339325454204,0.08422468003111816,0.5825343155948262,0.9248692855774868,
 0.1548396998645366,0.6525420295491801,0.6697551725906233,0.10711431229381918,
 0.3307112155147812,0.3904889719191976,0.1890016630411041,0.09275934888064896,
 0.4260745310048959,0.8971622744935737,0.314270942640403,0.8279586873768787,
 0.2798791017038519,0.033442677039199675,0.16415960265943919,
 0.6401847246121759,0.5333301609457448,0.7712231935483088,0.68700781727683,
 0.6772780507126901,0.25336073618671295,0.26500176070596093,
 0.2946460587069427,0.08549223871307554,0.3097275626365079,0.529374679178433,
 0.05892603497246507,0.9259191184127243,0.8474806507554195,0.0894702277866416,
 0.6885947234446139,0.27745256867302726,0.20462831830553052,
 0.36305421401830895,0.6956158268557033,0.21619433358513906,
 0.7689408006061634,0.8931716353569626,0.4293135529191011,0.08208304332777083,
 0.28918278246827567,0.2445030295992171,0.473209405444623,0.4116072195695907,
 0.5775384166653764,0.06303044934112179,0.6861328675582121,0.5375693008242498,
 0.2783256809763144,0.15141696280004036,0.4567736319597353,
 0.29366456238695493,0.13555089735431536,0.6091339144129682,
 0.03824875160264862,0.5917995097018105,0.522337254396934,0.1678420272724761,
 0.4344356849036719,0.8260738644182112,0.36832596139842244,0.9951951791182322,
 0.8612709361422319,0.6667759072463068,0.06753429401078637,
 0.14255252097150717,0.25915657973382145,0.5227718553003184,
 0.03895959183120279,0.03220895405025748,0.5406415606115864,
 0.5912966590573159,0.1935713945347397,0.6725685925985246,0.16192358009759222,
 0.45175510499347127,0.2626369103195526,0.2604325052736768,0.7609655700601352,
 0.35153843967614096,0.9832955370946732,0.11705990979444902,0.596535695885019,
 0.5145569357037748,0.6913341192969024,0.32433346723378564,0.8681964696865236,
 0.4117649762778255,0.341041411581012,0.6045743767575101,0.7530619176369898,
 0.8145481515445343,0.5983403193246085,0.45375202097479905,0.773115798777059,
 0.9254698277126459,0.7819969649555506,0.6843740625627502,0.33283883337986364,
 0.534741635908977,0.9250058939626233,0.6200488092661998,0.4212550966034001,
 0.34010439843415885,0.5982012031774295,0.4944762985008234,
 0.12175760441660943,0.753273950428347,0.2856414635958203,
 0.032981836597928305,0.464077123783764,0.5811899760027835,0.6634469451016496,
 0.3141618802314261,0.9288907179884656,0.6391232029261364,0.14215732755231736,
 0.9648650919769053,0.4003732444420506,0.764728293743422,0.37021253783283103,
 0.7131385662451767,0.10285762489384709,0.24770698939935099,
 0.9671853390096554,0.6041436527515479,0.4799637055251649,0.8196376575754364,
 0.6246576726097468,0.6611477149069387,0.04026188761062832,0.5403287972803364,
 0.562177649840469,0.4674390228259875,0.651470885126948,0.8060283141721271,
 0.6765170858135464,0.7715709352208272,0.18574845274039953,
 0.07590832978729123,0.2028857684152695,0.48149326617299215,
 0.40929361936608166,0.44452428640357533,0.9900686888138792,
 0.18216906797657328,0.46644879609476875,0.35458682642926465,
 0.9671982607469212,0.1079967397680901,0.5618077246329705,0.1773545073028555,
 0.9044847762037618,0.9066973591359352,0.8266570184428854,0.13284772260610367,
 0.9774865392669319,0.7755079098462063,0.861394511437059,0.306620983346946,
 0.3357504503102777,0.7977853747157746,0.7731430876183347,0.13169846599431945,
 0.5749184343964335,0.24062348428638713,0.1548484477587604,0.7331191965693367,
 0.28463033336410826,0.8858120779781641,0.5077735813519175,0.9372399230367255,
 0.9855655001495042,0.40365940895910424,0.6280587284869438,
 0.46681818811256226,0.5070876209159948,0.9110157720451394,0.91287335846019,
 0.06671934207023145,0.0101637758876435,0.30524292865100233,
 0.7043692035407407,0.959100728888683,0.3367643718342336,0.2228047898941088,
 0.7320641337743011,0.7197824650516979,0.5569810395213287,0.17153599152098198,
 0.34437595459479764,0.8980257706285385,0.06344674478015966,
 0.47792208966632477,0.17996312449906693,0.584992193672483,0.3302288597883438,
 0.7741989469336199,0.13692197987037302,0.0227426908367504,0.2452083199009638,
 0.86596440710972,0.6339724896040708,0.5969696547597579,0.519337431740788,
 0.6797651974969277,0.9496476454419694,0.4896640723031531,0.38521522156699084,
 0.6533820186162482,0.9668247737411062,0.5239709639792972,0.5836933490174226,
 0.8067239042006946,0.8016125346954027,0.12752062818719467,0.5150619832309575,
 0.47011830668502996,0.2934741792466572,0.7687144722174393,0.6373479792421157,
 0.1935759138321239,0.49027743289720493,0.8293946213735518,0.9199230632417874,
 0.2549344586348352,0.12423314565605381,0.424113745855497,0.5337029859017671,
 0.5858324389459778,0.8907555200511014,0.9449577013641384,0.21537185880053078,
 0.8922426437674901,0.8509505457062206,0.7694768997493494,0.767091927292994,
 0.9669251672232803,0.471214084566005,0.9597868332696595,
 0.0023873611233462544,0.607337306735344,0.04418557759064512,
 0.5821670835290813,0.3704365433360588,0.23279524605087543,
 0.07446730602720486,0.27902198780336995,0.8141695185160308,
 0.9324201073678992,0.46217764103896986,0.5364776803498567,0.8857712044386575,
 0.7225755407418133,0.3074107019810244,0.9715417445590193,0.7476630175061155,
 0.6857670888968849,0.8597240916199534,0.327221364782882,0.825020680355171,
 0.8927041505492456,0.9125541593752436,0.11466418910346843,
 0.049292697577073774,0.757202001453856,0.818206331900566,0.07297420108226937,
 0.5302559529056652,0.6110599885550444,0.7140884517451884,0.967926606020023,
 0.8192629766957585,0.6292940922595771,0.6859313882431276,0.7296652434147147,
 0.22347680989151675,0.05357728223453584,0.04269834015022056,
 0.23383155009140277,0.16570026088369916,0.8421488534254977,
 0.2925877198300916,0.6448949604205494,0.5950947662108341,0.2763891536970766,
 0.09856111175720728,0.924132947903027,0.9360129859424547,0.18075625318350552,
 0.851134015689017,0.6246842649197605,0.5273387813121471,0.9410345020247632,
 0.1750982714857532,0.7977256706943925,0.0763395367009112,0.2522179666960268,
 0.2446808494635856,0.4311443185763676,0.40201879051681244,0.3531559632038773,
 0.9169247200878405,0.618825512096842,0.7666958582996594,0.5620355178877745,
 0.385394896620311,0.6165167272738973,0.6862627200096035,0.01654469976918016,
 0.7772109169514093,0.5892773854429402,0.6326127995333823,0.38644593674947336,
 0.34500440203968474,0.08602382285434362,0.616088594755129,
 0.31828510170770496,0.7294053556705604,0.7938837478697897,
 0.34495838380688426,0.3524184429361794,0.31699246069479736,
 0.31507036531178956,0.8033245678080266,0.8046380503339468,
 0.08754217043452805,0.8248453343116128,0.17359707734452412,
 0.7719605779318608,0.2795421460015012,0.08474299181114453,0.7327658248124692,
 0.2147541700569724,0.6586248975468216,0.9276174550602578,0.2084750726163841,
 0.9021440342684393,0.5871262494591922,0.5199133379126616,
 0.006615520510488415,0.3150571492845643,0.5223868887147003,
 0.027324952253034884,0.997301621627103,0.47572448369137266,
 0.7311213989579738,0.4818860072598322,0.4339956372468111,0.7446281778363594,
 0.39040204955517843,0.7120670916868919,0.4990248815529955,
 0.39789914187077347,0.7909612952760825,0.2291182894297754,0.6713410447315229,
 0.9223235366885764,0.16458054587411697,0.8690651751578993,0.6350773442499148,
 0.7093818818603337,0.9326997673730979,0.8583582217690777,0.1089739429831682,
 0.9366442056987874,0.06014698106853156,0.1989002493447063,0.2329655770207546,
 0.2068328074215635,0.22961986643815147,0.03831464862990874,
 0.15777934784953063,0.3966503337066901,0.9665972355128262,0.8541684557328134,
 0.06647152728647165,0.10523140149733956,0.6720897018829919,0.670438145509059,
 0.908164289842623,0.8836227908203417,0.8373775562189956,0.797194545488676,
 0.7782710753135293,0.014627524384251522,0.2069854780658652,
 0.30603084013594717,0.9552661329673717,0.7211419438902291,
 0.44152334482872346,0.1758870307562137,0.9436158922221222,
 0.47591553000145426,0.39308345073885875,0.04324211279362866,
 0.109049557066636,0.9945745689862213,0.2832027994553469,0.9939425288501667,
 0.21425369078105994,0.8694572727697948,0.49159111187512705,0.83464835919692,
 0.7019692777690748,0.02276676665719779,0.2599760236013595,
 0.40635616507004313,0.6509402779532355,0.47927939714838963,
 0.25360006142373825,0.9579909893727706,0.2316052579435791,0.144842323084601,
 0.4146117927532066,0.49863250489145516,0.3961387000398855,0.3185800155385584,
 0.7011360444859829,0.6249411998419534,0.9632621869524395,0.24342972281823838,
 0.37621745562126596,0.28010006864765047,0.6303025906366663,
 0.8065769227423578,0.05285445055890092,0.9268602621610131,0.9939742796421218,
 0.34746549320683817,0.3686548356504399,0.7062154743587261,0.3869253991340712,
 0.8167107782520208,0.7543190982435903,0.5631357420250273,0.10493691265475025,
 0.6919213743502818,0.8337435775853074,0.13689883280854953,0.4061015994715885,
 0.3657933313037125,0.008682845156183117,0.4374778310722742,
 0.5991363722247798,0.9346311021087291,0.4416193913230323,0.21328849937473815,
 0.7024401785786586,0.12421833011991734,0.9283608005258013,0.737525054154618,
 0.5752035421896584,0.8992958845454422,0.7065322723714995,0.18225723307968855,
 0.5907105559988484,0.007410202028279245,0.006632925514060473,
 0.48038103534717824,0.44958727948483945,0.2946851238644528,
 0.14560882779439732,0.06895642772975119,0.964681592096369,0.4709253297537346,
 0.49649528293090683,0.4692835595252651,0.35067949140437893,
 0.19002451125539255,0.7485353249619694,0.8213631269858085,0.6382621326147417,
 0.8869905382341102,0.8095092993818684,0.835196436245975,0.7278045950377079,
 0.5283029561439374,0.4578561133090311,0.761955546640149,0.8080112805365228,
 0.18092364159616525,0.18782724103216897,0.6594688598106059,
 0.9815755703939306,0.6681788026191167,0.02089438262273935,
 0.14993412644206583,0.8010559944540999,0.24603141873296075,
 0.9229628110077268,0.2328490348516945,0.9084235667715834,0.5448889836761475,
 0.4060714371221543,0.4380203605073859,0.17334366653922284,
 0.41567373259533424,0.3637233985273116,0.5514539352720914,
 0.39019021356799344,0.3626518477148428,0.5505525017526165,0.6354814798047496,
 0.595047953387262,0.7553393318883374,0.43597212961804555,0.07794242573081434,
 0.9592485654277586,0.6316610015814121,0.36921629014308466,0.8608042888602052,
 0.3512405434470842,0.5924045635636266,0.48124690646776747,0.8505586343582182,
 0.006022337478830075,0.4318681185269512,0.9729331329986824,
 0.8512667057390022,0.5404913757758936,0.7520014951465107,0.8897202398605828,
 0.8474837836564655,0.5520486583569458,0.07663470070406253,0.4455444834733384,
 0.23019834073873624,0.43332737328675863,0.4053791473202788,
 0.8898040373143186,0.4455786060028535,0.5163199118222168,0.6515784069895999,
 0.6537916803072343,0.7270952309997003,0.3179438352395848,0.8055689235482071,
 0.21602358228635243,0.9340701523924687,0.7938334037763372,0.5110017068504659,
 0.6908392252153357,0.19685416470594452,0.15872949917703394,
 0.32400108723293897,0.38629790253174046,0.9321421308303057,
 0.03329532587498263,0.7204124512547531,0.15766994909758725,
 0.4213864192713449,0.013021604178367951,0.2422731390778522,
 0.9079201117418084,0.7426924826004906,0.47030825858622527,0.793423289895135,
 0.9158916541091575,0.4790958376306067,0.25351221607272223,0.8236487759999499,
 0.029020949510727756,0.9703901388056376,0.20310839023853022,
 0.5578471990214885,0.045178987255560576,0.12969242090066901,
 0.5298606612149663,0.09054595397285359,0.8969894480684516,
 0.10241457130919596,0.6267794350233895,0.02453966725382939,
 0.014944603449961846,0.2167274192826374,0.4591854576993415,
 0.7896825782224187,0.6076207236870602,0.20378437644677838,0.9181458105429086,
 0.34500080512695375,0.47826334703241136,0.5101085284615635,
 0.24968188979673644,0.7143222856837311,0.4593468856746993,0.8756484813654938,
 0.7679988786381081,0.5594859088949902,0.09510319599385864,0.8402499914401311,
 0.33555668865835453,0.6168568564005806,0.781939453729307,0.1500159557011722,
 0.4531044500559436,0.8214586939082175,0.8564763034573447,0.3750198499146782,
 0.2532305000461945,0.7759875662159451,0.5003311935293687,0.08882093856826123,
 0.3106333142811126,0.323815119278842,0.3021990256595378,0.8090880439222304,
 0.3410854363616841,0.39610319526549753,0.11893963423485354,
 0.3097286967540731,0.6616855727994041,0.9226765807800642,0.7176921492962254,
 0.7287172306121806,0.9824623031844013,0.8437995513655174,0.5887916358134847,
 0.8655489283019757,0.29904415499812853,0.5381162398451196,0.890350260454714,
 0.8825028916554255,0.8046687830543341,0.8613027966019956,0.5619305808699651,
 0.9903789305732102,0.22380573773800783,0.04290307898294754,
 0.10392936435052125,0.035671979498108186,0.761817053610761,
 0.22418929425024914,0.9226976534696509,0.5890583239016797,0.2910843991283488,
 0.7052883745126209,0.12577648007032693,0.5688374722045616,0.4152274929441012,
 0.6497371137784511,0.9496020013244837,0.5140935756802093,0.8827491378562526,
 0.5392428823819096,0.6375169613725629,0.6580087275600491,0.05186102664166781,
 0.21674508364859613,0.7114266465994845,0.9551481014315982,0.8885182883015094,
 0.11326748215269744,0.7211872896734017,0.3664112236796171,0.6625376080997079,
 0.2562553958749695,0.501336305469094,0.019269795197791373,0.975708131770551,
 0.4026509310159262,0.30465736191481374,0.07306132998525405,
 0.1666417794948406,0.4364040846471464,0.7219698799423199,0.12479098013306167,
 0.548200902436355,0.25051811811433033,0.7720550012716573,0.21022056495800745,
 0.8256431756494154,0.7380305587673038,0.015237776468698727,0.438010894505762,
 0.4841363823429422,0.389701158945847,0.26223421755144294,0.14960065346700246,
 0.2590355836411544,0.19365979075315454,0.8410189056563437,0.5049828898702521,
 0.5055992407555768,0.8300552170003015,0.4063017685170631,0.1874981452304545,
 0.7749548881501093,0.7284644006827602,0.8441460250668209,0.44991634572041106,
 0.40736039013064274,0.9544928252840077,0.550735784011166,0.9711288971405895,
 0.888246272879476,0.47108482530106255,0.15155198650030977,0.3522766686640133,
 0.8197636682767758,0.8260856758759414,0.9669035474321019,0.06815775774345334,
 0.7362018856194849,0.41566609653971864,0.927396480404157,0.61613388137216,
 0.5085727735219593,0.6485458611242071,0.8940055082907814,0.3522001498186772,
 0.9187376804202007,0.6083798304697934,0.8410622745846829,0.3542791390708464,
 0.43751466150367535,0.29837156670679177,0.5518377414265943,
 0.46865809778860756,0.5531497492724156,0.41504333847950725,
 0.06258525100959533,0.45394189872004487,0.22436734867490882,
 0.6408347624603121,0.19748139343020032,0.7239507329537923,0.4006018248157708,
 0.03334883519903864,0.6832986203827847,0.7747312270383947,0.5955746825181343,
 0.7937787952404254,0.424731261876686,0.038045766167504436,
 0.20399347929454037,0.9772090586896066,0.22134731636136673,
 0.3141467317257314,0.17046980762015895,0.3874075132245428,
 0.39946513741694567,0.8537659180162062,0.966653624100867,0.20212307916926942,
 0.5901087268100519,0.7784718258574858,0.7114142528109811,0.7808223965500807,
 0.24545357488010067,0.947517995290354,0.41987396399097054,0.9263570506218524,
 0.4262233640863693,0.8630796503659122,0.945791777990907,0.911194860298289,
 0.2525784675772127,0.43122729584039265,0.06117747963775766,
 0.5875638475915265,0.4166448465434984,0.4631730788693148,0.45312666635078036,
 0.7088223807207115,0.5471958163272468,0.7448248403134372,
 0.044642509264107844,0.6689869714252842,0.7761735561671018,
 0.5483959682248306,0.6052722589098739,0.9457341672605105,0.8337533212798576,
 0.09949521728515665,0.4814498717677794,0.9221297265733511,0.5288533022204248,
 0.5737001601940577,0.42829156583231043,0.6823027609529597,0.5255960305080392,
 0.34370450075054426,0.6715025326496863,0.8627781113559677,0.877935898776119,
 0.8508786252320519,0.004150407335884565,0.8229229364569353,
 0.7388501558296601,0.7553261746690876,0.6331781700785596,0.3843506841133505,
 0.14687695762470665,0.2454274970199577,0.23003666855614746,
 0.037560608211357316,0.9597220885377961,0.7886653280964516,
 0.6281630714550933,0.6390476994361565,0.8470894613091631,0.7722363976627595,
 0.681966905238649,0.5558093085704633,0.5606429444147517,
 0.0027300798571083806,0.31188015508602285,0.8797808563648041,
 0.3326623998265116,0.33753921581037283,0.0031170123819718176,
 0.26857303049973014,0.6200440331272893,0.4211922202827001,0.7091443308796161,
 0.9319585272613593,0.21368269064677625,0.6416268162192145,0.5621696252223756,
 0.8323062415109811,0.3065920980523861,0.013844691544952159,
 0.9693447017541462,0.9586050583806243,0.9202240443814392,0.5296646640800646,
 0.8872704008287906,0.14316495008003605,0.9173641540191422,0.6691019980495373,
 0.7531069400257965,0.23958953400027172,0.7145808127165597,0.1868208161015581,
 0.27153790288104984,0.1700671744932709,0.775280084326957,0.5807428919086073,
 0.345506742400873,0.5134883141378289,0.7310833101091309,0.5902205167443512,
 0.10249313498787338,0.8936004879891034,0.6773460207312748,
 0.40294917212574466,0.84002141169773,0.7230412488426431,0.5128842968942253,
 0.15079517727138003,0.47129870668986507,0.18472752329490105,
 0.8453054724753347,0.7491870876430635,0.5044638077327437,0.22075599707419624,
 0.3556157500157022,0.27207787670704786,0.7117041862309392,
 0.33231228576008376,0.3702400338740286,0.2627827748988456,0.5076626405887292,
 0.7014855417892943,0.13852675013929439,0.5597666677671906,
 0.07194035657973263,0.49763633245541183,0.21011820890377175,
 0.6765460917122199,0.5365518328974221,0.6848756469076875,0.31296208587406504,
 0.6021056249603375,0.006222320990309282,0.38152801165153316,
 0.19863407647501197,0.0950802927391522,0.16635225321044222,
 0.6025357312734453,0.29901285522308685,0.7279447350922823,0.6526872645857192,
 0.5147702104323814,0.6342606783056906,0.011391245022224883,
 0.7029683293892439,0.16871579871168485,0.3917182623109081,
 0.15236300810743764,0.004661936259095056,0.6453178968422276,
 0.7091529731484345,0.8071634056115392,0.07101640799688325,
 0.37961120840209617,0.44703200567180434,0.6316267160264368,
 0.6129625786219273,0.4714302524728061,0.22396981689286244,0.3569588123993971,
 0.639031108891737,0.5417639294197527,0.6794574806499357,0.2598737997544567,
 0.7543053981693508,0.20557551449047018,0.5414079944400962,
 0.42546243596056277,0.1727862116941955,0.12579024868972655,
 0.3611355843644569,0.238723425143745,0.8859526017374302,0.6607997367274582,
 0.3983150523335608,0.9041703037471025,0.43957863160652355,0.8970560309506126,
 0.7870241045088983,0.5607869682253614,0.5188864498509754,0.9902468096066457,
 0.3479166968686297,0.5612928303276703,0.5424012936784046,0.16315222709941113,
 0.9996117665936767,0.6420796269296399,0.9987728012078101,0.1538982451010673,
 0.2027393439357288,0.9839870389894299,0.38411633286577784,0.7901169589121025,
 0.5246158059995597,0.36186673377035816,0.3394440031944701,0.7371417369079487,
 0.8592131090864512,0.8820917097243485,0.965614800170993,0.8265830132698467,
 0.7750671364899554,0.8700123057358554,0.6496896819985714,0.1485890577988448,
 0.40040656848813994,0.039825119489038,0.47547298688548234,0.7714341479733747,
 0.9987547087752153,0.3750593969620897,0.2797754969587255,0.8176221651400076,
 0.8819324319677916,0.8229288761699127,0.9743117273334005,0.3971580972854476,
 0.9458703789907988,0.969034786092421,0.9923559623573406,0.6540618377014235,
 0.6744819860387823,0.7578324246256597,0.03436589706865456,0.6307372008273031,
 0.11785837527864074,0.11847724698987638,0.05622574485071985,
 0.642720815592122,0.1575711005715127,0.5848456663093871,0.9179758240467366,
 0.669337997685824,0.06101566451648888,0.6651400586319908,0.5169010912260802,
 0.3798155289333983,0.09987837589948623,0.8734441729329864,0.9017223822648452,
 0.6273371684171162,0.4264090392872011,0.09155235122961747,0.3601061049692479,
 0.8770619901880603,0.7507927309457773,0.24988084349237805,0.5360032170400609,
 0.6789700986781445,0.4904451567543626,0.475293625950186,0.48370567375528906,
 0.04668782448239028,0.8145097130429969,0.11055435991345164,
 0.7656327220063912,0.18262826402992893,0.5845788538084342,0.2688183731553285,
 0.8399413885730863,0.6437408494472903,0.815156507345325,0.3200828858793566,
 0.9380305686104016,0.4608728241323635,0.19299884006122425,0.9127081200028023,
 0.3457489225818834,0.23648658688327584,0.24776029128596133,
 0.7032817853328088,0.6701164332539492,0.42170372402818157,
 0.23431387139042503,0.10565622999792035,0.2054525529172878,
 0.32244503580494344,0.7765096973047344,0.18963648319767756,
 0.8663969293115915,0.5837641716536569,0.04141453204940582,0.2812186892057388,
 0.2922543227051573,0.11248114008254861,0.6363211664620307,0.3002989988855541,
 0.2866196994642902,0.8503844521385174,0.3048823503745488,0.7610726261539184,
 0.8796758277502854,0.9981920583597588,0.6691203507218624,0.16381565461544678,
 0.01994440240179285,0.5232854741379569,0.6041327101740093,0.9031054643202947,
 0.3396204485261417,0.13396100300506686,0.11673510487773586,
 0.3219252625340685,0.6348480789022221,0.012618377297608618,0.539317499152808,
 0.48455603941729253,0.15294337835717953,0.023316395600981865,
 0.3828662634292799,0.15207659449301691,0.10598884763101368,
 0.3485783484838334,0.3544777968753122,0.3153917369035518,0.9999150544512853,
 0.6242598520507866,0.5927605901628634,0.45351040788583197,0.5699417263838706,
 0.5123557503427681,0.17426008222453426,0.4465039620104133,0.8903001998804211,
 0.9460088354008284,0.9153079396110676,0.9116824906089183,0.15062385733574357,
 0.48779351859323294,0.20024996889408908,0.47043561879810347,
 0.9295998135282237,0.35730147687199165,0.12119760555034842,
 0.3948133720527799,0.8634454158713321,0.017589899831732758,
 0.7124342353796185,0.024215613326186602,0.7246719811999185,
 0.30298759586333934,0.0019620829098048187,0.46184472000129717,
 0.03278655310035716,0.8456532317619618,0.707845711080174,0.4177064908872026,
 0.6497287008613382,0.2625329306247085,0.045685273412276683,
 0.3247403993757262,0.6691486631678273,0.9273661036046502,0.9593291811871381,
 0.649196213422213,0.16982461449160557,0.6562993921214068,0.25535823260339774,
 0.034768392833854245,0.39325132451571987,0.5657377068942676,
 0.45113321844505083,0.5669200662995434,0.8914937605333682,0.8904010741802377,
 0.7050344601828925,0.6661571906100957,0.5973942193688097,0.8249918917194489,
 0.6085262366207423,0.17582321562349823,0.7165096841563636,
 0.22887629107683205,0.9695079828491896,0.9835191435132677,
 0.15033122902058815,0.9467347476761308,0.6908437408891168,0.7831514863710629,
 0.08666918642162014,0.988718723793093,0.009836063768485226,
 0.07786978870752792,0.6661174201990845,0.8779079933963756,
 0.46484761400055785,0.23283759279942629,0.35855210259236114,
 0.4455344110996222,0.6216798975672567,0.15058118438585377,
 0.16894624230975774,0.3519094520736006,0.0202924823595233,
 0.33432145975378247,0.7827746871390879,0.23949781803214887,
 0.7746442966050087,0.47614788596468094,0.43541206004154276,
 0.04500900485838355,0.9610145625527831,0.4946801754747634,0.8749031155082688,
 0.25020551072319697,0.5907388717269342,0.7698259565046485,
 0.46054778783449735,0.1835256702952185,0.6903699100491485,0.2354302675922736,
 0.9882365066485927,0.5879355966330143,0.3220588992122463,0.28933658742970103,
 0.4904856904769971,0.34071969313299777,0.615042909899446,0.6924683011545779,
 0.6525466968760469,0.42025627803195986,0.851985767842859,0.2211825984802438,
 0.1793851237034283,0.4701178565992986,0.3343044813080418,0.08621377374922745,
 0.8943146879733351,0.600291604007932,0.8587943809652384,0.48106712262123275,
 0.6376724126572414,0.5022273211760512,0.30833419073718904,0.6644617741726742,
 0.8784987541873829,0.579950681669937,0.4940186119018559,0.06484797740613213,
 0.15444055157281944,0.3592641218346221,0.0025475946256067195,
 0.9477692661023477,0.21517095066177905,0.7974965843342561,0.1575261118005873,
 0.8141194949795846,0.8530668917952855,0.7701416327558637,0.8299619699059801,
 0.09434499091838666,0.7622435952626828,0.8855073084625029,0.6312459683633013,
 0.1836811985621094,0.6990376996700788,0.7649421740020795,0.35698684904269684,
 0.7560234194194911,0.26064585937404416,0.1635896025846295,
 0.22777743119581706,0.3620857480062717,0.0777068239722758,0.7556868122755848,
 0.7388747102339244,0.1720937425219432,0.7710517470727917,0.9891883647170159,
 0.6802314232719955,0.8897439416580871,0.029187761749163488,
 0.4981246755552575,0.29340686297476537,0.08415515576652344,
 0.005208814862696642,0.8429413622482441,0.10235004102791168,
 0.5507084652766474,0.06077839757049208,0.19533117705485536,
 0.061766182584629514,0.15109529697934143,0.5434123219075128,
 0.7666217095985971,0.2836992231073363,0.812688962091486,0.48268046923454433,
 0.415981295858038,0.6706630359479876,0.1283507779051345,0.4090422128378737,
 0.5873925172333408,0.10621211770454009,0.9487178702855309,0.7731947443249513,
 0.022245229334762806,0.7679826851913221,0.9580239627798531,
 0.5330112953190329,0.3908926006550668,0.29200579032460994,0.7214026228519579,
 0.5581071887860829,0.8367390142162312,0.5327690344529283,0.4809136682747319,
 0.5162878055776843,0.6346054238351172,0.5721559887453057,0.12425081706974839,
 0.7722800822589416,0.554245692888315,0.7054393634975632,0.6038761219953894,
 0.7174393527559592,0.6661118599592756,0.8119627293959573,0.8676433597133467,
 0.16393576305837032,0.9819287809834777,0.6302775530662728,0.7092369120728901,
 0.9698014647332738,0.37658887700709653,0.5693398493989797,0.7252604856862724,
 0.8762043508006457,0.5486248609257968,0.2581962910965161,0.5823784782651689,
 0.34506057305975113,0.7850957056423608,0.43762098570156727,
 0.5877909090977287,0.5693276956785683,0.986836327520852,0.5766205664087964,
 0.6409529856873046,0.41102425075621885,0.1277180351141306,
 0.04533487393005553,0.8359032678268681,0.2011721385567573,
 0.03285995810068576,0.4418569006896522,0.8478695798117255,0.2193368323241669,
 0.8575142483872172,0.5995155952506758,0.46420828530949065,0.5783910614333446,
 0.2494438133423309,0.772784295274645,0.708761765374366,0.05833331007610654,
 0.5957251991521483,0.8827973772732902,0.9657396036152318,0.9361881970060658,
 0.45816543260366116,0.8817751258672809,0.8537271937405275,
 0.056798070890355046,0.8530761899158381,0.1553494767775292,
 0.7264999229409131,0.22661367135133004,0.7800391586254891,
 0.28762979165539504,0.5265085399231922,0.7258902456896924,
 0.014039665264280465,0.7193724128965036,0.40349533669071524,
 0.21775230317701755,0.5240741234048328,0.6606699820426778,
 0.17809026552127738,0.26122864243012545,0.21145977822537687,
 0.2984316246337242,0.2619787464775284,0.41989012102295076,
 0.08430075274000592,0.4776262295310827,0.5319055049077261,0.5302773539204104,
 0.38190084318430784,0.7611157509425439,0.6916381837077538,0.4736888463334755,
 0.32790073830288735,0.35370300946968747,0.7946764361942977,
 0.11547286580390026,0.07346671945745598,0.1890177048607743,
 0.0032989923372592767,0.6573089019166181,0.24084351346635136,
 0.2877802067101981,0.03709753955275685,0.7624994625562503,0.6546512583031854,
 0.449432868606073,0.7942876556900488,0.28165249463910325,0.46293602955672686,
 0.13109549325883973,0.5567644607226734,0.44636721139830204,
 0.44700328034769277,0.651395470641969,0.8577336976492762,0.2978044770248821,
 0.389934309080082,0.8272248317147579,0.41316666973457294,0.6678366888323718,
 0.5137866074404993,0.31074276337517703,0.5406090615451633,
 0.22810492124962212,0.079383501851644,0.08866838849694636,0.5672026687788844,
 0.5338354928359377,0.8690623707519103,0.7337546338764189,0.15203724181208333,
 0.10330683983433919,0.6532736291928687,0.7184619865564326,0.5467051244572203,
 0.9329715984457082,0.4605501602440596,0.8840430205586975,0.30206289967552413,
 0.34336603256617393,0.7373585534041963,0.9531768278136991,0.9192350278091823,
 0.8109291716662538,0.5365028099518412,0.7542254025147301,0.08177129340384715,
 0.24873163743464954,0.6958633623268019,0.5232677102226971,0.511360750030433,
 0.15524705945098893,0.02857592878118087,0.8653773570709222,
 0.6173983128925816,0.20640731207164076,0.06934760626236658,
 0.1956933802535964,0.582298246181889,0.14429347370203638,0.768681053423481,
 0.17408059315353974,0.660719726569432,0.20884726997502245,0.5882569348325531,
 0.9786837488803736,0.5272786879931034,0.6979229712927417,0.8909479205743454,
 0.29457570639710107,0.7925117960152518,0.6120657825116648,0.2948970769525824,
 0.674809417508557,0.7545225493802397,0.022090192294644817,
 0.10043656013492619,0.8460477908506007,0.6347433409324026,0.9402342812613729,
 0.12511250031156662,0.9697146654545503,0.345828851421928,0.45346842538128995,
 0.18968036380830677,0.3254869836287375,0.2171238975735782,0.7287946359518893,
 0.01641411664048864,0.74096124613732,0.5685624310692108,0.3977922262343333,
 0.7282616549474804,0.7103516748138786,0.23622773813965048,
 0.17981271559428846,0.5690237160810334,0.676039967192474,0.8333024573703349,
 0.6905401747619443,0.3057780167894575,0.3003415921178184,0.6770817570579485,
 0.9765182776903738,0.9459017988574938,0.015556351583621453,
 0.34187172953800404,0.17707837065422483,0.19256144553209686,
 0.4187475805943205,0.6132557338831746,0.5122579590416084,0.15943391188951173,
 0.5583032541301792,0.3456992096372019,0.5883187458562514,0.42002788094000676,
 0.19245401108642202,0.013424696122387836,0.15564806997546343,
 0.6693102531191566,0.7096954878467437,0.6367082876757811,0.8639619433257179,
 0.9816488650275246,0.1455148419043789,0.04373378056037969,
 0.06605464004008932,0.9296849943170878,0.03672204165858761,
 0.3336392545283646,0.9757199507634757,0.5342408763263327,0.11029465633231528,
 0.7367437767986411,0.5047116063145638,0.800235050006637,0.3181197848964614,
 0.9688378128778181,0.3534491765211546,0.6101161991988268,0.6455130428550563,
 0.028681416192062326,0.9861225521202636,0.9039782722014433,
 0.6039733324897628,0.6118256289282334,0.2580745038874218,0.838782752168474,
 0.92232841547931,0.0937676477934345,0.01350988481683546,0.3419006048411152,
 0.0883307365751026,0.7541784428552495,0.9646355504994597,0.18913880618953927,
 0.8230860457343274,0.5204749757608563,0.479922621455523,0.12024317059674872,
 0.02975066624192857,0.9104456840238976,0.555294954015715,0.1448869863336879,
 0.7631959173073535,0.33290775853624166,0.46150929930530094,
 0.46018730945357433,0.16612387460921962,0.7387121315508575,
 0.9300093823271409,0.12657455481765,0.5084456143965422,0.9431712212630838,
 0.7623529002794747,0.5138593112434441,0.5471906613332446,0.4155451815011092,
 0.7958770727924591,0.678372801738522,0.8893092344900733,0.5815753310689032,
 0.9060407692193699,0.6175850606843256,0.4411156011527757,0.9771197997432124,
 0.62732821635325,0.1347437732626292,0.6709682636218708,0.5133308986274658,
 0.9713491463929279,0.6244573680874003,0.7954658019475439,0.12378758351212193,
 0.4541374376365045,0.42442483333329106,0.6844526931662317,0.9131056117777138,
 0.7845273066778389,0.1390141563074363,0.5246557816441009,0.880259520211326,
 0.41375513050806245,0.9483241360877896,0.7222000003079967,
 0.18111790223284896,0.1914873783043738,0.7978530329029602,0.9167497972606379,
 0.3685006509462114,0.7801448317795557,0.6195329188726489,0.27267187675408444,
 0.25986709033665667,0.6203758346974532,0.361321382532547,
 0.054951779591104866,0.9029515670911343,0.1670219449421052,
 0.08737258549946902,0.859339229157119,0.8529830610312034,0.2106284730409016,
 0.8276727820947132,0.692932990927538,0.3618328877904413,0.1785260191032514,
 0.5969934461304357,0.04890401364533081,0.02674658654512363,
 0.6477165751242318,0.16486859313232438,0.8989509459168427,0.7132734917151345,
 0.2097175952748339,0.1891659456996614,0.9553117528531292,0.5555535607492342,
 0.590269199897619,0.1592651688778447,0.9749333928815196,0.6484258771918686,
 0.9693901956373494,0.6090609399102488,0.9083219312667553,0.8350558803071007,
 0.7689657106020941,0.1376932052382862,0.3615374366755124,0.5404276234791813,
 0.3991355883158876,0.2229717667493437,0.627045379597745,0.8658164215830955,
 0.8350406663014884,0.056939898623581287,0.7845103986141817,
 0.8095828416382647,0.10726207025154888,0.4601579101630867,
 0.37194100512663764,0.6994994066913471,0.1652589315242755,0.6143774614509531,
 0.9398108549586277,0.9742425972513251,0.7548985937118821,0.38353749138606263,
 0.13928200407843638,0.8409059113213792,0.7286372960025169,0.5539841899028656,
 0.12416110074527242,0.2194547031722287,0.9846687737407637,0.9275388970602347,
 0.35814784038159597,0.791412162467361,0.9474016530760151,0.13460453009733897,
 0.7843665954616488,0.6142954366406185,0.5221887740720778,0.05764490239803177,
 0.8251309744140756,0.7448816986611058,0.48166973424961657,0.1346744363789677,
 0.1542739219483339,0.006198737159575418,0.7134120002108465,
 0.7180439890497967,0.5048028369843598,0.20788798575292944,0.6235667251752524,
 0.6810436564595369,0.3091218621933042,0.917133101066413,0.47490284469973654,
 0.7674741118489615,0.13233141973588491,0.3042502583238256,0.7916772306609611,
 0.1607766542123794,0.1011644284308334,0.4715632554897766,0.5496235299264736,
 0.10261643495045703,0.7702477087605978,0.8241121016995184,0.3952160034362292,
 0.7179544793526552,0.04203670573278351,0.5418770399889135,
 0.07620548513466108,0.7860226560623128,0.4449147077690021,
 0.08031986118742873,0.3336486886765696,0.3715614867069279,
 0.48622428159096653,0.26283906290293757,0.5387648082534354,0.788824095429518,
 0.8770603746602774,0.8350764519830728,0.17347472424559585,
 0.25916828310000095,0.09011911805057837,0.4541850437083583,
 0.7297677981678388,0.26393134503483096,0.11746221813459734,
 0.4203860717756831,0.40126976345229814,0.15424323935255946,
 0.31690641036664424,0.35233188563623596,0.4983906443815016,0.366010409909475,
 0.44782159550190914,0.9310251465571696,0.14455770734391626,
 0.07741597978358039,0.1937786175637174,0.0842463626884341,0.2657678140581937,
 0.9614577339831976,0.4993043536927946,0.9748968931461968,0.5891088798975527,
 0.175716906645366,0.2537320576824489,0.4313757360352979,0.4915706898742678,
 0.21348033265601962,0.14546932843409033,0.6486963262464216,
 0.10251233029438678,0.902366509435202,0.9540074755799846,0.337930856768741,
 0.9625154600181864,0.13595309250618415,0.2656461708031782,0.5653730876745886,
 0.21329507495570266,0.27457159888802707,0.916771603761943,
 0.025727816176242757,0.7147820504856586,0.5720852878806731,
 0.5602368671983131,0.8065316365429321,0.7388825111321032,0.22739389849101466,
 0.43906723796388725,0.09998615076714312,0.25706708620351004,
 0.10940048644667588,0.8491212359134216,0.47396263980724807,
 0.9650369262800949,0.7491569260119569,0.3164931997503677,0.611566999422574,
 0.6124165396207608,0.3489304635263011,0.2768075904040713,0.6006899794980122,
 0.167915156413305,0.02551163676013235,0.6316022370158341,0.2318175383129173,
 0.24811575919616835,0.2735915056030147,0.49828081876513675,
 0.6881082075963217,0.4634407178600589,0.6472450177878162,0.8873444163909004,
 0.5878611088680065,0.8334636582191131,0.5980500629605984,
 0.062280489818803186,0.06945428259541231,0.9723908709379385,
 0.7004437765008714,0.0804552246887058,0.8641680883815828,0.46264607166026916,
 0.14586003792397007,0.17284231308205633,0.03784762275682385,
 0.9348301135013287,0.5040323622772052,0.28004396531279274,0.4397151649110834,
 0.060028987776079235,0.4545912466192483,0.666425676760799,0.819019407514707,
 0.8886646890646807,0.1708395172044609,0.422905986859066,0.6015139802627669,
 0.6652158861132534,0.7958168750088758,0.5037733397824926,0.8319370021441567,
 0.22672264682277843,0.9861744514384411,0.3093647058627227,0.3556317321971625,
 0.013271307138562505,0.7045661341431427,0.49619058940612404,
 0.057900991056970885,0.8068472333808531,0.18366075768117662,
 0.6209276020169138,0.3187357575395279,0.9391557320145941,0.7931557754530941,
 0.9252614489890741,0.2188719876007782,0.7109272881697225,0.6039480600001269,
 0.36251873648422483,0.35213288757450156,0.1719233366242865,
 0.0634874419358219,0.032784075801681634,0.6293931217717266,
 0.9981870608334861,0.6095121650858862,0.6266306317607482,0.09641261062810902,
 0.4869016572355731,0.2796629137747949,0.5272610899583733,0.8215953691864386,
 0.10836833086970687,0.18206368824828212,0.6665936215014315,
 0.04489470306216403,0.8457501604094629,0.5178239520018294,0.8936721767896625,
 0.27938212569151544,0.9232458283051233,0.669690530560421,0.5521710203425032,
 0.14185784750860597,0.23155573849828104,0.7158898131420253,
 0.6115604135002701,0.476983759625384,0.27975928172353826,0.6445559637565305,
 0.9848472916394772,0.5214760019056206,0.5190830880261498,0.9205413912745553,
 0.707718071199904,0.3679362076693704,0.21734592638878547,0.8876138748654846,
 0.5753544551886091,0.7942456539789772,0.7208960082595046,0.16194938060843012,
 0.9775976294873495,0.5090027622550857,0.33874158928684517,0.4232922963766751,
 0.20150985095816987,0.1563797398112402,0.16847431754040565,
 0.9690640710128894,0.6332359340224526,0.7241077250152494,0.0276936497823409,
 0.0720451705031585,0.6020938096010608,0.298101753817689,0.35465399532269093,
 0.3815261414749205,0.8070577914269976,0.8219812006099517,0.3345978106638221,
 0.9765251296783151,0.35456239441797455,0.11851985681783639,
 0.4761346019174574,0.1517491229610739,0.0341586832622186,0.34175920823000683,
 0.2904820183112208,0.12941874514863805,0.7404569285905762,0.6491176716247585,
 0.2549525695338115,0.08301885198033099,0.26168778925035086,
 0.8653616814805309,0.5370723160341592,0.8640145107152152,0.9399509052631841,
 0.05976264392762598,0.3313373533943538,0.2156576528890708,0.6356185065823173,
 0.8035578202337386,0.31913612931720237,0.10340588032386899,
 0.8698421761278173,0.4251095956004949,0.8615582851270913,0.5255790949798654,
 0.8418498044050741,0.10889976331284645,0.42485006656081215,0.832822256332012,
 0.2501622793446352,0.1018059226933935,0.9063198815100213,0.6950911776508177,
 0.23743387682616235,0.5985040621570417,0.38148024019990756,
 0.4078211261942656,0.4599640340079362,0.6192296534664877,0.4230855605546897,
 0.8838090530417638,0.089876677970406,0.40070710840855917,0.1153726057012685,
 0.3282056370216143,0.09889468951451497,0.9615344923376715,0.4140020320258222,
 0.25851686690551823,0.7705250753115652,0.3231746002349547,0.8410481149726603,
 0.5742840099833445,0.5790452270945741,0.8724761621673737,0.6458025409034311,
 0.6454323465597195,0.8209332600389455,0.2221121020500194,0.44446867497688236,
 0.9116374929048713,0.08565706208139523,0.48133674835193274,
 0.5531849032055923,0.42881466726319206,0.47298247027656004,
 0.10971535524815734,0.7589203684990046,0.6470314873638444,0.2222225026854867,
 0.33750449990673714,0.11298120945348489,0.6933228170362806,
 0.4500550752386163,0.3566648861585816,0.03583738155696581,0.3623040805385038,
 0.6572066423064532,0.7620230002204483,0.8421308622147254,0.6978870464184406,
 0.40060977649447016,0.9253727279745717,0.5120233751828176,
 0.39830435059727387,0.8277380146405036,0.14374895828684175,
 0.7758522506678354,0.0055676458984746935,0.1351110398390455,
 0.9886893552299774,0.9622783917364938,0.44558953966313175,0.5516774403449389,
 0.6645626259968206,0.869960588408197,0.3077410824156568,0.485525585872741,
 0.22576521052752996,0.48609600450628554,0.4869409878466997,
 0.7530162934833171,0.4363029691729543,0.3549885000999833,0.7837533443820359,
 0.6432424502405645,0.4637461783825919,0.6756897990000229,0.4277495914090831,
 0.7271743575644198,0.8842472409045196,0.9424602414470158,0.19388113992950573,
 0.277476977750589,0.22009136492288894,0.10196113529250939,0.2982002731653901,
 0.1738057426744276,0.28852815857286607,0.21269701045098088,
 0.23989616461573227,0.7021221509591262,7.089599267353641e-4,
 0.1137322800709335,0.30516517552773925,0.34680660487605597,0.537773615841953,
 0.5852217879412902,0.8253940067704704,0.07080608165839553,
 0.07126982842895968,0.6143015962638386,0.5036498037244507,
 0.26591139471045633,0.09076952701766683,0.6170387943595058,
 0.3329051931699737,0.6907206785405239,0.8164899301789311,0.7425844107878667,
 0.6557698534647665,0.056443986887076414,0.057638341279796945,
 0.27723164149679913,0.06548791927099784,0.6446810043815866,
 0.30065733379198567,0.4662574271946154,0.4534604780465331,0.5355422047874212,
 0.6527551815416515,0.9889524926250471,0.026374907911954737,
 0.056294037476643854,0.9917785248587206,0.48174465070815153,
 0.6488770023467656,0.2820729422837106,0.6247427860668293,0.0704060042610467,
 0.43099208125833266,0.3801720546905383,0.25621028576184257,
 0.3702478371101178,0.6710205447402199,0.6683340475672976,
 0.046139757142508286,0.27032196980925927,0.28238985859965005,
 0.5000440981451966,0.44426376494368547,0.7608684844628337,0.5141733096330107,
 0.14061734305733964,0.4034770937097589,0.447924889125761,0.3784121309171393,
 0.8924650242278859,0.8073162257865227,0.6873725745017725,0.6441452942348103,
 0.3967799794247848,0.261200380893569,0.6303905903817859,0.8647644317488977,
 0.9050390806966748,0.3662460970867649,0.3313481068698234,0.9203124610901585,
 0.6297940334329831,0.41814786579147456,0.6339463721991176,0.8034792484059785,
 0.5494710158812318,0.7012482776516931,0.44303212543035153,0.2786352866665158,
 0.09849301591414417,0.3284619053674065,0.848148136755609,0.9908730743908065,
 0.27329748376670515,0.2957638072177474,0.45316653178518573,
 0.3561811835696369,0.1541661180790933,0.45857021300655987,0.2571042512425311,
 0.876846213917497,0.06526400217911621,0.281921312283845,0.6417661243751799,
 0.5603240128342064,0.1423525899176249,0.5635332186167923,0.6558870754087209,
 0.3793002854893266,0.09669380154699381,0.7682563034785154,0.6405453362555733,
 0.7582136342952113,0.016815118433494813,0.3249646379936305,
 0.9887599092776279,0.3707572123082423,0.4836940001399178,0.8841334239454829,
 0.9796316340558345,0.7598285862471488,0.014633460304075419,
 0.050120131102122656,0.33161006403969884,0.03525754814149451,
 0.2819335137015495,0.4623367970651835,0.41062737327115684,0.6590011597277085,
 0.1422039140340592,0.5774410219720085,0.4050303677620405,0.24969968250466623,
 0.8801043939370253,0.2554691671429986,0.4203291538614221,0.8065057840515726,
 0.49514527944876496,0.34310434156957925,0.652728952823082,
 0.19266590335953326,0.9286727191781114,0.9204975003129714,0.9579626589157474,
 0.9900154511004127,0.5109983989470179,0.20226305508915898,
 0.08375171917231494,0.25331588680585204,0.3562716427713556,
 0.1549752754238396,0.3414244450503079,0.39886849042731864,0.3333498078259768,
 0.6205327338340657,0.5279752635138582,0.3673340608807235,0.18199373555443643,
 0.2611845408874611,0.2557083443663253,0.42393838733066846,
 0.17185813133518313,0.39703893362065346,0.8883245490128329,
 0.8843813340822796,0.9545093001493261,0.4061226523693966,0.11525711697608387,
 0.8893905521615748,0.5579265751790189,0.9934205265389796,0.45956061407613014,
 0.19738899019034306,0.5806223690259743,0.027625470304958877,
 0.5453630051551324,0.5906219767574576,0.6600205210278016,0.7261248932473281,
 0.2868494455166093,0.772749168645588,0.32797245549764165,0.7853639520762046,
 0.8141739854952216,0.18534442177242294,0.5037818786165384,0.9554399391110115,
 0.2264219645393345,0.11668124580131611,0.8590287877880055,0.7014635765037048,
 0.022951979491936947,0.06587047719469485,0.8251268760760907,
 0.3150464218682275,0.763181221960457,0.08058967786809468,0.1172964342547762,
 0.34680532141612175,0.31426609702903496,0.889251612644535,
 0.23250970404432092,0.5099317067087371,0.9994309303415614,
 0.16953026899294688,0.8003535625228462,0.4720175548266382,0.286061641562526,
 0.6971305936500254,0.761117652085633,0.9922305033441499,0.4275430265192889,
 0.7599207893625948,0.40309717467014705,0.9588960831845423,0.223708523397545,
 0.7066245161998752,0.30930765759805356,0.00411367915625227,
 0.7974209293816013,0.9571411956488964,0.7527483737612808,0.6319637933911667,
 0.5642839728769449,0.6131753344906741,0.6722820917206203,0.586222792745366,
 0.9555750467192758,0.943515911652742,0.5369495310445991,0.9757754171434259,
 0.3521347501072083,0.4841430409277414,0.16829498491688855,0.3267767437863611,
 0.37590497298532854,0.4626711921446982,0.4717170395330116,
 0.25580307736322394,0.8533500038908035,0.3636360354794126,0.7716345693715826,
 0.9006619668878293,0.10662655869569182,0.6318848524342918,
 0.003812065461869407,0.9384168312185183,0.15301456871547015,
 0.32717791747917335,0.11568345471330943,0.7154969156514659,
 0.4928598800834868,0.9563315536593098,0.6406656665330691,0.9353341674927285,
 0.14529280812610612,0.6767801219124193,0.7485759537265396,0.1020287527385958,
 0.6617000392397944,0.6916263684712738,0.04805883714137882,0.5453437571077974,
 0.20888420025517462,0.10593297652343703,0.5867207462445096,
 0.3564062542813303,0.6911937390719282,0.5020671539753171,0.20227003351660067,
 0.8012207026937175,0.019156848781363145,0.6450752953995418,
 0.7033846752353831,0.7174483629498551,0.8541877059376923,0.5901536701870133,
 0.9962353876561411,0.4892100228949964,0.2514161422693697,0.20852806667771806,
 0.042428617783109024,0.5889602198601054,0.6624035393624287,
 0.5421537675841266,0.02007230738583654,0.9215114193982153,0.2994328830787223,
 0.9698277333976889,0.647258852064153,0.8651423222770166,0.8708747056028601,
 0.9060453933619536,0.015087790620997077,0.6947052465672403,
 0.5799680442919716,0.9646532323064527,0.7415683266954995,0.6608700939946247,
 0.5719026455834159,0.039295679955212104,0.45369050799670185,
 0.3718157400796941,0.8483444456999784,0.4005747123817679,0.2192537213975394,
 0.8463895182258037,0.8679933640066582,0.5216618956272794,0.24517882628515797,
 0.7786236641420196,0.31567439137863507,0.8392169173873214,
 0.48399419115463027,0.9218581968411579,0.3193216158408311,0.5051843063713207,
 0.6726555276236955,0.3683024674611146,0.6447338906794112,0.763131198451731,
 0.12562445059249494,0.5988423187919274,0.11421030817368594,
 0.31491951312654864,0.8931575667710974,0.8365703794510071,0.3104387207977274,
 0.6986802151294926,0.9575530448208097,0.3510086821028794,0.5123568200965033,
 0.13259446484187576,0.8689101075075422,0.15672998593686294,
 0.30460514782455406,0.28064131058217057,0.9854682366226425,
 0.6218771587604539,0.5947614913281946,0.3912190921853852,0.7494556081070467,
 0.8106977964973604,0.020734263748718673,0.36328417410794533,
 0.7224478108462389,0.07743305972872072,0.8546873230819448,0.8262752043322665,
 0.2713583865260245,0.5266394242865219,0.44943620254546146,
 0.06325303841045515,0.7434743145223911,0.9446121165012568,0.2834067060774734,
 0.730550193719685,0.4326135929481304,0.8462636756207836,0.10327224587315842,
 0.12135156597324692,0.46130535891855107,0.11844109065380803,
 0.5648136454687666,0.8813197669411043,0.5637085728249761,0.2963769940677905,
 0.6967639430495781,0.33750299400264183,0.13289995718915382,
 0.3757712180580376,0.049326772054736345,0.6201535349478728,
 0.9120431627299806,0.4242349542019399,0.05213677062584776,0.3575887559438662,
 0.03943018757167749,0.9259438469351963,0.7560217743952302,
 0.23840363916977392,0.5866170714238198,0.5615663338466965,0.8771325696183443,
 0.8977237724283411,0.8200655824045984,0.5927702022057466,0.7787810206489814,
 0.8366821967186326,0.9458602809603572,0.44448138272143944,
 0.12154410794805748,0.7801736792389585,0.6883674710348983,
 0.05039171778809004,0.9906581441351447,0.7733440738728272,0.7351104486027159,
 0.34755380878117126,0.6406358330920423,0.8530318814445285,0.8666972057409602,
 0.8968169981445341,0.11918876563221259,0.8672379491307846,0.8087373078644692,
 0.08717322715122844,0.4370654098529092,0.6502250269698581,0.9134730478578188,
 0.9807963736477029,0.8757053066130499,0.019435778245496804,
 0.04326461709007534,0.33277390563987186,0.818047840916276,
 0.06074617576660002,0.5845075570775278,0.9939157334502651,0.8364423423407095,
 0.5886503838388307,0.7251403347424905,0.026551546876013354,
 0.35318597785683525,0.7517474023092844,0.2666089938618381,0.8595026813567181,
 0.020732244581632142,0.7843754349225044,0.4003760623858251,
 0.8227117926176108,0.13289644529792632,0.9292257222247517,0.3375452793546463,
 0.7583113995977843,0.21866501880139566,0.35768717287604845,0.779023878718198,
 0.12476734183808069,0.026646228198257238,0.4514777857072483,
 0.8148729701798244,0.8648832113826386,0.6320933181096186,
 5.672770681175443e-4,0.8523008977298866,0.4925342908772944,
 0.4979156433921711,0.4116048233126143,0.9990506844657583,0.07287003130246805,
 0.5622974274983926,0.21545123511358688,0.31357807695357853,
 0.9584652690393497,0.17175443649455402,0.5191214331090791,
 0.36628266658443065,0.389666541924769,0.6163345089696896,0.8340704427207453,
 0.9805391978564038,0.4605031311043647,0.09896121385540924,
 0.34709933615029154,0.492921237637139,0.7597100572055112,0.3674612817385119,
 0.16181019901376226,0.3616011239627207,0.45608257920857054,
 0.39474911619118447,0.28056417269038236,0.5901635789769146,
 0.3744923138688382,0.5840501001354603,0.853713629550279,0.6922267746427079,
 0.4763035950042275,0.006705584284913302,0.9438138049652973,
 0.6824028123441352,0.6655097557928497,0.7960328032910825,0.05817508478151889,
 0.7399546575795469,0.7756582388293997,0.1532410762232853,0.5771138402953874,
 0.620404245944326,0.8199607870655493,0.6033668489669406,0.15910832454180857,
 0.2530447175942818,0.0038847445088822097,0.8347003901714561,
 0.801347279062857,0.48175249677441223,0.12753300652400656,
 0.38626251549476986,0.9756326499547922,0.8224062828131181,0.6213042634698092,
 0.3692301274786831,0.07362820021176697,0.40662696915906116,
 0.46305627010470796,0.8207840991529884,0.5162726328734857,0.9386744113443475,
 0.560044274048784,0.851670472713402,0.9425341153596082,0.7836798054750139,
 0.3034333102286352,0.29988424885667053,0.9004093352576661,0.441348518137069,
 0.6222534847251336,0.9942589660364123,0.5328939559215805,0.1607126549999386,
 0.09153912607722725,0.3215302045180414,0.8752367952414892,
 0.02190225437091531,0.6299306445200257,0.8563460088628021,0.1564797161917092,
 0.04321833186542989,0.03674920936595927,0.7096099801058708,
 0.5838232494509702,0.9029618826075178,0.7742696023461848,0.4345462737990212,
 0.13262688653732313,0.6138812887822708,0.35785135956917025,0.493388194264199,
 0.9968016809712293,0.31537082962022034,0.8580399010820309,
 0.30142358625828314,0.30905473690962015,0.974272766053303,
 0.44198921250006673,0.462138039716254,0.980129112800718,0.12483612847115988,
 0.33724679239871347,0.0037557410877608888,0.17345937768069464,
 0.7239009137870781,0.3586178316588593,0.7336658632405518,0.4473862237492441,
 0.20657915283694983,0.7219855481792988,0.09027923096120671,
 0.9658782955567992,0.7796339805005459,0.7919230533967253,0.08541150348782411,
 0.05884004400134013,0.8411725171425315,0.8607011520220591,0.8891286330587452,
 0.8636273343965947,0.4563231037977059,0.9050044642334195,0.4904797363192134,
 0.15569876426286577,0.4924906399033018,0.1935027087206539,0.9611632151371565,
 0.9389151076924054,0.0860918079877997,0.9878581418349731,0.3886833551110549,
 0.024686904162043333,0.26357513011744793,0.9338863055120443,
 0.5200414045972631,0.25382594378300805,0.1878714542941352,0.971274250383247,
 0.9255781357323541,0.1675156859692254,0.03840820320393179,0.683752272675707,
 0.4973790852626685,0.10748381538921226,0.7618670947590058,0.6048746487232305,
 0.13979614708809918,0.01648960843546221,0.16076260770344108,
 0.3191886184994358,0.7309347093854164,0.6559012140933984,0.6282013369714313,
 0.561407971317234,0.8378442423945579,0.30697733237360236,0.7456697761351303,
 0.8621993493401733,0.8268602187050569,0.40222301574608976,0.8526350730690166,
 0.6768086294125073,0.7939839950463039,0.6208766122153404,0.7570940040662045,
 0.2040709184469811,0.21757146378771763,0.5970153451871985,0.6207193461419813,
 0.12420133348234352,0.9892869601570312,0.9763796144420731,0.8895966269044504,
 0.009655732632178782,0.9308981717865585,0.4234354336269568,
 0.6585843027229042,0.2173350838808049,0.10619697232564662,
 0.10040662138737422,0.031929244353725794,0.2097540432544165,
 0.5104234140869928,0.7760109210408747,0.28172849313801407,0.537028799462498,
 0.9329058389322363,0.714945489016722,0.5761676770117005,0.3077987882253469,
 0.11601071605945945,0.8201219497271592,0.4989815328666203,0.7128818928716302,
 0.3905684064227367,0.763750023889964,0.06547865607929515,0.3415793630508802,
 0.54335177761049,0.9355288034665835,0.4172951859606492,0.9093828227510207,
 0.6733675584953082,0.39257277392591317,0.7154181865433451,0.7363130043439843,
 0.8738571250915754,0.9009115861473319,0.0931201961878696,0.35341885440168264,
 0.9320267222790717,0.5166628652955704,0.16422631088225703,0.4376312406759889,
 0.32449002664843707,0.7186309734006746,0.09596580450672587,
 0.6876109922517575,0.03756062306849994,0.7631602908177011,
 0.10293017334012067,0.8198182068731369,0.5809488788049458,0.6658790872420182,
 0.8390831289441865,0.046994835701675064,0.3679478292949977,
 0.1628406051924628,0.21852482402249396,0.263054191811568,0.9499167268499874,
 0.39156006476935357,0.4616302290771207,0.14426038375515637,
 0.010097664617637059,0.44436757804608196,0.2792897846046656,
 0.3597249304517116,0.6989934535947345,0.6081649487012967,0.9106447588466633,
 0.9665604265175276,0.2637262678492476,0.08032726822828284,0.4227784590060417,
 0.4399273230327412,0.7251822417134608,0.30885775472446797,
 0.12000742494188654,0.8148593499833001,0.15508523779709948,
 0.2294933290008515,0.7518028566090625,0.8243790215933401,0.3266502273716868,
 0.5579489058560729,0.9801293755275838,0.10650457523005075,0.8325274278303814,
 0.32789999546784854,0.20934038541747635,0.3012352534088335,
 0.059435357543014655,0.8555099896047933,0.7339830001994799,
 0.15388908902147502,0.32842423743491445,0.40190899681421766,
 0.028130355276016705,0.3487243792847069,0.9846200459137555,
 0.6224452766988402,0.06150785393401359,0.3199291193557764,0.7484975864617175,
 0.9728286089307224,0.7078898502502368,0.5594506338278351,0.12535919152548003,
 0.3626023025735302,0.46828853225141975,0.10692430276728204,
 0.12225098255148958,0.36944592840202506,0.11431915391863867,
 0.47236150036624935,0.30366775967073023,0.40344496887131176,
 0.5288466398942855,0.44720692379775984,0.2743704357640173,
 0.25654792878156396,0.4991539883249311,0.28961637649950234,
 0.09150875585327256,0.21130183554205684,0.27567787908861463,
 0.5027509086131372,0.15661723724944876,0.3933082941871162,0.2634850186519535,
 0.337834743212315,0.7191756996649658,0.3142009721182841,0.3254407526407438,
 0.5432934347780323,0.8201503464894042,0.7224836372907744,0.2484890690588475,
 0.32503738177734354,0.2897327213902199,0.4053021582489523,0.908159132933617,
 0.34423474188560044,0.022386567221350462,0.941035705102343,
 0.7105180692556101,0.15095757949757627,0.9544863441272864,0.2618167257109164,
 0.40345717132743664,0.2872088058621013,0.9071733135372082,0.1363996131083599,
 0.5148420618786023,0.39797438756718995,0.34724987207876734,
 0.6277273778216809,0.3731416578751817,0.5511823668832546,0.2508189119467137,
 0.19495969170018101,0.8986497841642456,0.9819840571933108,0.9438467238444682,
 0.4786046426533118,0.6168294167097733,0.3469503275676422,0.20707825692345727,
 0.3852435857834067,0.12599681497809256,0.8773336538031112,0.9291999846276391,
 0.0546849323182067,0.9408605671039105,0.5318499148492127,0.9166367849677721,
 0.25902793056102025,0.6708309344381733,0.2572228894858424,0.7765125667301492,
 0.35563386328214563,0.1437096597536761,0.5811354828170366,
 0.29373675713455893,0.4790765409493535,0.12613489401638045,
 0.11664582178358163,0.7444370287275764,0.409956969708697,0.42623808127679896,
 0.8100875146867779,0.2715485410524844,0.2453666541223466,0.08421724509845319,
 0.3104062500638747,0.14507416424597586,0.16754184771679992,
 0.04788065295195465,0.5540544153633242,0.8951152380378037,0.8012903036329153,
 0.47070391558255653,0.4112994992449177,0.1809187686111744,0.9775434224031166,
 0.7112298549203366,0.12646527033936872,0.4492437425752037,
 0.22357116489001494,0.6081807946329755,0.3010273638180855,0.7794729539279067,
 0.2657473876081303,0.6801034881178356,0.9859791062504842,0.5441625334613996,
 0.963037960316483,0.12502336132150016,0.32600705153321907,0.6794266734577715,
 0.3000974649808934,0.6569136111212911,0.01042187794498084,0.5992120322698531,
 0.20765200199622535,0.9970599229740177,0.005859070188784288,
 0.09226123295946098,0.5351186663865857,0.16830562223731094,
 0.22498384482076728,0.8311932373778499,0.9009767897435936,0.3015363230178594,
 0.5482225104709124,0.15315541816704203,0.9591233005641195,
 0.11476110076154433,0.2845272423438976,0.5703088001918996,
 0.19819721473120744,0.9531074305054371,0.062137758495110496,
 0.6447814860191516,0.5160556948789269,0.8342088616032413,0.6969829454402727,
 0.4339461946342127,0.32046128589515044,0.9984699354296732,0.3735191115959329,
 0.03519805273878562,0.04625561736835837,0.2767682637188217,
 0.7278097356975303,0.20027777915424216,0.9377685755864597,0.9563576048133402,
 0.031034828664669112,0.40670730985220593,0.8842376734069497,
 0.9028775498962416,0.775215923725296,0.08991379746176265,0.4436705761556753,
 0.7956399236910692,0.17336807160650847,0.7095934776641469,0.8500585165812611,
 0.2694428851458144,0.6315092619343862,0.5894909142694809,0.2893667157545359,
 0.8789892345943388,0.060459125170608696,0.7151318680568542,
 0.3219743464339626,0.4534294155828782,0.16535433136154087,0.6199807707680978,
 0.4574248939469314,0.4570705134233751,0.03032190120760414,0.692019559304375,
 0.7098870511779105,0.5773652927291014,0.5230602522612439,0.550854997367563,
 0.5215581738893834,0.1669960330250413,0.21013946799005967,
 0.14548903893206233,0.36760352795117435,0.5057439335943328,
 0.7252030993669489,0.07174750667421648,0.8710725659950552,0.6190275656671823,
 0.13052005817444368,0.49995435585046044,0.5007574226079049,
 0.26658532586978123,0.9160861097508861,0.7201872364263416,0.7902037133630674,
 0.0040595047117132355,0.016794369650731666,0.41790045695150047,
 0.9142919531082103,0.9674871182299962,0.7014775706471529,0.9749665893090451,
 0.5862957945072558,0.947148729610463,0.2270396082307684,0.8983979417325645,
 0.24181243957715903,0.030359690381034476,0.029969624503040637,
 0.3287701890167435,0.9555701102211371,0.8406942072558587,0.5317663776557051,
 0.5963890673229206,0.9773037407764262,0.9144274878844799,0.9346406301717357,
 0.34312835129365293,0.9979948309352262,0.8008311696578143,0.7090678424387851,
 0.5461201665982905,0.49492867001359864,0.20314309765779393,0.44954314915563,
 0.6952530844487195,0.5072250017361396,0.3815163570208394,0.992932093036094,
 0.8887744231646008,0.4501270949034032,0.01974849358259656,0.2626252998890223,
 0.9331554454396307,0.8684727128500933,0.9484710550829056,0.6655969617463063,
 0.9869598658904414,0.2596314385810887,0.4074281587719202,0.8690368769942551,
 0.5651164080475997,0.2933357357783716,0.23368675841797892,0.4030549060015688,
 0.31601460044763396,0.38097196097819397,0.48086621714134326,
 0.09551334369753495,0.9940085300846037,0.7960934826404191,0.8720257598409057,
 0.4696388663289762,0.795039985996358,0.7981285928265021,0.791668107787577,
 0.793002795649178,0.7216344091128641,0.2498896251053938,0.2010726830484787,
 0.8908270804390043,0.42515276062607066,0.9426876585539327,0.7788394718582347,
 0.8724171561783914,0.7914128955093825,0.011895830976487276,
 0.7448328339761138,0.44146394204752626,0.5345398594152345,
 0.06470589186331743,0.8926518356577149,0.10688442421625699,
 0.7571987581693552,0.8146248094669677,0.26644169813720286,0.3958346794103885,
 0.43256387046026523,0.35712124658511435,0.90313200542101,0.4701922911842894,
 0.8833826712443787,0.7409249656310214,0.597762904032127,0.5663766353990102,
 0.4306998380195678,0.8874046890161468,0.1059130395438802,0.12144165652720929,
 0.36995935726759077,0.4756545898580873,0.548168060188243,0.4186514055131061,
 0.0020153824125989095,0.562763672368612,0.8423509934578786,
 0.11426937063297316,0.34510221262011365,0.49934570500883313,
 0.3531689105757111,0.5743455639872918,0.6993928995507834,0.7772904401993973,
 0.707424099018067,0.8958998736221603,0.33725247264804614,0.13540267633840553,
 0.8394964550893405,0.9817342708988631,0.3013716582922029,0.45089694513644396,
 0.9485424838145115,0.9789169550484041,0.4515904452494093,0.8005081501685871,
 0.722443750511311,0.805641230763564,0.14153690012629183,0.27242611922174087,
 0.2902021612160829,0.2769498396139327,0.8316945509611119,0.3583661027229542,
 0.009886752385227293,0.29926716982348833,0.6185645109463505,
 0.982896027492075,0.9329753852065189,0.9486288250837382,0.07153165453676025,
 0.21178098793057054,0.17351944656312046,0.03269188499141196,
 0.17826110371931758,0.2257728817206044,0.6101436679717601,0.997526377526192,
 0.3225591055815785,0.6319155029965831,0.010792169364550963,0.711435905448748,
 0.8927676816667666,0.3691662113483384,0.389707694478845,0.18094926492137087,
 0.9268822969641696,0.8744976271163526,0.9265514246327451,0.635570309388037,
 0.09069272094561631,0.9171916340346218,0.08131385389463208,
 0.9873795549543364,0.08246914717266096,0.5777351111698129,0.3542360029008026,
 0.5271016612370706,0.9444939332205176,0.07663650433012681,0.5286782705004687,
 0.9696494772107427,0.17062097010947075,0.7780949167807069,
 0.16522481717176385,0.26000510035351554,0.5423359889047158,
 0.23904929706764866,0.7093314839128578,0.16368744411376646,
 0.7990797863591138,0.2235099740299814,0.4395712709168953,0.9784568075988851,
 0.5676533836317053,0.03828265149752352,0.6650714705112931,0.5588621968803321,
 0.8698808253491974,0.7751114758061589,0.1510724824427898,0.6725707216356279,
 0.4695765985477798,0.6993902009269022,0.27959525238846006,
 0.24981917854093105,0.9678719717768637,0.44076808350621755,
 0.5126300773756323,0.6568222928614218,0.9191033314915543,0.8316094838045667,
 0.17497250740470216,0.9111914144845776,0.6330928774053214,
 0.39994124316938473,0.6438208306553436,0.1798205285606611,0.1498126825883075,
 0.23712612892010232,0.7784060577632372,0.4908725184879751,0.7808757547055183,
 0.020296091885573908,0.14728026099257885,0.4098596101620847,
 0.6824025507209474,0.8535919928474326,0.325126966614533,0.005996727999710849,
 0.03622004593797756,0.14942879899925066,0.12799511194366553,
 0.4597221388624313,0.2609216132554102,0.5347222390197033,0.41271608752865796,
 0.09922969588469277,0.5109424909392638,0.34477923349790873,
 0.4109788680935994,0.040536078256955266,0.14618434743346198,
 0.959409924050571,0.04507153190225954,0.49861681768948163,0.7735047692859742,
 0.4654433335237357,0.5891194589808195,0.9647141415591226,0.48594172521453705,
 0.06488465177260816,0.9138759295196424,0.6712134560252976,0.867221607189056,
 0.580624797150509,0.29347029390615376,0.5369840235743032,0.8522152621807568,
 0.5683335742761415,0.7208769126965218,0.6125734885833571,0.3599958609056132,
 0.5931844258955423,0.2771728220612002,0.14890955455034427,
 0.15723995722692863,0.1945246759074063,0.95157323725917,0.9106228385110514,
 0.0704458737789927,0.7976738678495047,0.8310822616274991,0.2533911653148584,
 0.6465072523248313,0.37706649982199625,0.2274749182717304,0.3940971078112383,
 0.06826268717660477,0.16890846259086123,0.032998004753199583,
 0.9584294517334638,0.27563999980660125,0.8947274835853074,0.9648609540301407,
 0.18865683309157422,0.9535744384811135,0.5428198555610717,0.3141754394852634,
 0.5479700299125247,0.08574600279606959,0.7848712437539781,0.7457094354445839,
 0.18831105322243147,0.45614494352285695,0.2093811125292362,
 0.1680103979941182,0.1403899888827631,0.3996407269678032,0.20915762117340764,
 0.056478324711825234,0.1066984261676509,0.14389660885771993,
 0.8298778262486718,0.18357815919054088,0.8012673926208638,0.2898885487880266,
 0.15559892752810112,0.8451896248311269,0.7406357278952257,0.6928687301746642,
 0.43704514100295755,0.12922217201662922,0.8385931435436128,0.758369153501162,
 0.629366736987377,0.11989635696596102,0.45325101476893404,
 0.014508998528309891,0.15722349338710906,0.5925870722145712,
 0.3956901515022134,0.9517012096605433,0.5237942729756225,0.8096842775747465,
 0.8484921366394437,0.09061303001602727,0.9904623250613737,0.626209910392516,
 0.37480172685891433,0.6125314006247117,0.8604474698705364,0.9018639014049771,
 0.8054788234227439,0.774648017223824,0.3387425767322061,0.8559716801267574,
 0.3245898642388655,0.8520905087277452,0.9531160462782475,0.5897885916302574,
 0.29049950092343557,0.7306681763298926,0.3274877303280276,
 0.09133146439372541,0.1791923523432234,0.03774622636293523,0.785334754784761,
 0.6437023607091219,0.9699615823617407,0.2981771699891034,0.8307265409963253,
 0.3775517341885607,0.681931960210139,0.7706830518446751,0.8733475448223558,
 0.941627606987239,0.7132724453335966,0.12060520624021343,0.7164560464936812,
 0.7025578203787947,0.370539872729752,0.6967776499911209,0.3098848136223771,
 0.5247947855117423,0.21979113055523702,0.8905862539444882,
 0.05536388057427255,0.9495260992329463,0.8297419400708895,0.2933052866322038,
 0.6529319076904847,0.41449964587153254,0.6608648304829717,0.4394264484633228,
 0.15048768009344027,0.615279023778964,0.9515437680438555,0.32171123125294265,
 0.8824007313114923,0.11706087136189902,0.7861603543556865,0.2027567749889111,
 0.5995716751567948,0.3738431625044749,0.5697287706123324,0.9304707367228241,
 0.8764497510836372,0.6819238843947831,0.9707257543054578,0.5642160591227763,
 0.7220096995445902,0.07208595869676437,0.4740951046871139,0.9331867559163067,
 0.2943871748814716,0.04128581472376469,0.6961086630576606,0.8650585229920942,
 0.7745123213227895,0.6014060795258508,0.7282999738035094,0.5177990245133032,
 0.3822650398964147,0.2092462895367958,0.30116713060928124,
 0.03174992257730558,0.6548362771342519,0.09878815996082368,
 0.8564785466719517,0.7236292621127619,0.26869910361117544,0.6729692745142578,
 0.34401588771980873,0.20114684467637356,0.790568741456139,
 0.10167187647857734,0.281306475548355,0.3965654757771808,0.3706572945637726,
 0.23733830861605254,0.5708266485074559,0.4189889451488018,0.6531613683581995,
 0.49883452687595664,0.008267627364286434,0.286001943766487,
 0.3026577279674336,0.215430184289512,0.08583181820957686,0.7660307008574562,
 0.6569802902848081,0.7711716208463809,0.05620937063912956,
 0.08791638630110887,0.24872613345583083,0.3551695116830188,
 0.9728552211958311,0.5140630283262795,0.8111661364538093,0.2826232889206328,
 0.635918165044904,0.1587190763640427,0.833944511127884,0.47295801616079114,
 0.8685707287485678,0.7884099415472614,0.9595754190726276,
 0.029664935787986124,0.9660194970002397,0.5034733585359872,
 0.5409992274192719,0.8480894973900384,0.9368728756179852,0.9430937028801128,
 0.17643566881895223,0.07457395364824915,0.3104846690567338,
 0.2532164130422373,0.28592930551398177,0.3219172365126546,0.6324208487432947,
 0.9181482340086456,0.3537899542587626,0.6694853398844849,0.597034554510134,
 0.2250080869965404,0.7486610202181504,0.6584040816597891,0.39119350202291947,
 0.029581948260321678,0.27860620100998024,0.08388649574503304,
 0.9414035734311004,0.13973855814168634,0.39746201811011483,
 0.4718442982502121,0.17719470880949506,0.5296054695239975,0.142687025434582,
 0.5861073256731072,0.4103586641054757,0.5607323532227325,0.05248308728327744,
 0.7005618826314686,0.1972533283716612,0.6166528465493433,0.9544479644751374,
 0.05595097701773244,0.25073832440201316,0.2063700987369147,
 0.12586003792723488,0.5896409638279652,0.29672463911592173,
 0.13662509205173134,0.7378414602159862,0.5783966105159274,0.5587649314743464,
 0.21196017060387173,0.34227333748390976,0.3889809847378797,
 0.6005042925476549,0.16837703132298465,0.5505426991397938,
 0.06033547179007359,0.8805664968325682,0.8672459221267885,
 0.03856466929349822,0.41866060322022447,0.14497508613668564,
 0.8394212411745291,0.3948756028034517,0.833905746502157,0.888428373043824,
 0.6735891388640065,0.549388774303612,0.06775093668721088,0.1305780651349555,
 0.09007040100334707,0.207719463275573,0.6732701531051223,0.24780116143555853,
 0.22005372629915043,0.9007356193302614,0.19121538329058618,
 0.8171447287183584,0.11910176104030423,0.35134848031354315,
 0.9996498035416868,0.32401312649068403,0.2853112968192235,0.6100997462150405,
 0.4186442452818262,0.5978566743067403,0.9992317688436964,0.6707472537751678,
 0.8621596369216474,0.16639931214271975,0.6930239734888674,0.6621611291977766,
 0.16101922896545906,0.47469067528556785,0.618518723920615,0.7576830524046524,
 0.05547343879955946,0.684158838340946,0.7665752158390547,0.6187227968525071,
 0.6848243743132686,0.8583615288143412,0.43161621126627814,
 0.027195557156540096,0.057716210766201814,0.5764659932676289,
 0.21253960111473746,0.8955266964776049,0.003483607040162817,
 0.9853130359961944,0.42859364125898813,0.8487559065370263,
 0.10669133467269121,0.6892621204113163,0.6133002373149599,
 0.04382878834614268,0.13763873581510266,0.2130084903015227,
 0.9351182343934477,0.2562669315863657,0.11555226953401088,0.3101804335929933,
 0.6765031629842547,0.3169134249147022,0.5454870024465035,0.21727615669838773,
 0.6496657313748248,0.8192979249018074,0.7292756213207298,0.5666424593976948,
 0.5827822575229682,0.8361938826873002,0.09623562833096666,0.9119695847179959,
 0.7252737730430977,0.9640364066747138,0.4556107142458775,0.7416883962699294,
 0.507123194530539,0.05720574216783114,0.45887047337510734,
 0.35282682008887456,0.5144706516280682,0.21453824738519256,
 0.6421387431575324,0.39035793261324003,0.21911512178873016,
 0.5441029160041034,0.7298661345669528,0.8452391690918495,0.09751782829245859,
 0.11384950581412934,0.43926386279297336,0.7992192574781025,
 0.6729308785391536,0.09423191084207516,0.4781989249628251,0.9746700900305629,
 0.5381876002542557,0.38909668523512775,0.4987061612633207,0.8669007338592686,
 0.9997026043491261,0.42850465934745574,0.828686336223031,0.2381289368941022,
 0.36999630116693594,0.45080874517030334,0.40746885607737426,
 0.38245187769613964,0.1361014263892455,0.3440688126828173,0.2628264599576102,
 0.39030415943093166,0.7181502851702026,0.5835725408185858,0.8953526026681496,
 0.0623648054152931,0.057429730936399914,0.42322170416804017,
 0.7514342614775054,0.7143284108087813,0.46535685647761993,0.9560642681909871,
 0.9201956091096214,0.8963090204963916,0.6271951809863393,0.2513342645788531,
 0.7447404605862351,0.1657068356190522,0.4581758560448618,0.5409905719230108,
 0.7566999993448903,0.5079216970167167,0.8326442006626005,0.18281500200563405,
 0.5741039306152107,0.16454624020112807,0.33447094538435274,
 0.4830737357341499,0.2713849123958527,0.16875598095402689,0.6133503214860362,
 0.127434651884479,0.4727869121847551,0.9056127729862407,0.22103397731854357,
 0.6682793455583246,0.8078591261407793,0.7786333631865655,0.0965246475009568,
 0.6390113557704629,0.8930244201595087,0.8698482837207283,0.4113587572082963,
 0.7841425999167955,0.746760361055247,0.8295032904546954,0.29391225326340686,
 0.006215605566992077,0.4879368353492304,0.9588039460605047,
 0.7244881054419172,0.8991544990282082,0.6185549842623879,0.04783488454918916,
 0.0872806605009353,0.4997937509544531,0.4853134384930575,0.8049587591245451,
 0.08748529430607332,0.5007256730413195,0.636650755012555,0.7884121249638714,
 0.27815774400216364,0.6676437789790214,0.18327630474290446,
 0.05915719085204019,0.9200594885096011,0.8896491504660428,0.7915870279409257,
 0.3553483148414778,0.3335822245336624,0.1388012010127595,0.1925042870519168,
 0.050824739348933945,0.30906709019116974,0.9557150030262399,
 0.7735744068328612,0.0838833788004355,0.33949413967318587,0.2932093674140751,
 0.5278364918513938,0.7599959889338046,0.8478579625526609,0.7918957937075781,
 0.542426828886334,0.6866169026355536,0.6884279886865549,0.16920731885488793,
 0.8555984370040878,0.05726927684007044,0.7270020504419796,
 0.07104688404233461,0.8180404974915013,0.4271868265161086,0.6297752889687244,
 0.28121036192872506,0.7166927782857357,0.870769744974511,0.04803268468257982,
 0.12433354118711737,0.09531102818358228,0.3339875956393674,
 0.5497938394504925,0.22433397278582112,0.40372179609102954,
 0.21769209645874565,0.5163951382437065,0.09383052006209003,0.715402883254329,
 0.7735326597471179,0.32536707036237467,0.9189060994878145,0.6671142126710212,
 0.6584951832003665,0.017563830672696068,0.9922505272177193,
 0.4973606250089071,0.18946106401411977,0.7023389332045915,0.7083290075976512,
 0.2540454461992764,0.9363033272531004,0.5775314517942729,0.21222041617383025,
 0.5782142484314088,0.4950570034768912,0.3999086468738712,0.9283137540851927,
 0.9532079716431475,0.7157245356888853,0.2607466161570986,0.6023219746886963,
 0.7129850995111648,0.11910769884345052,0.8453486692103436,0.4379585328680383,
 0.0865430846498858,0.36247602145064006,0.21341423981386576,
 0.14351147946332954,0.28570867781376164,0.653464385430594,0.5871798169307955,
 0.817814165791642,0.4901410144391063,0.33857705569516894,0.24480358352815024,
 0.14483512296430112,0.5947449464640656,0.39405322536012144,
 0.42901704255784057,0.4970217942348706,0.26135442486084226,0.83540685314814,
 0.5107177711534804,0.22771370750730635,0.8556190699965928,0.9925327695339611,
 0.7273379401735036,0.4079229285580206,0.8746299158156762,0.6825286014160687,
 0.12954912962186083,0.6259263637608978,0.7865785145307487,0.7413898966618804,
 0.7630786057738159,0.009482291368153373,0.30979162412297834,
 0.3329594030141254,0.6533523309701909,0.8337242334850942,0.22274541716183516,
 0.03979138445841224,0.3767877171099874,0.6163936454410919,0.531515164981227,
 0.6452575197136123,0.21001040468181176,0.5925791096032604,
 0.41868460344629055,0.28329974918263445,0.4339384468603721,
 0.14204205142098614,0.31097104276036114,0.07314088884540915,
 0.9459640359860985,0.7869115568683601,0.884215020450065,0.6607830670551377,
 0.8010739488891868,0.5008139770948543,0.11253696193956875,0.8227736030128958,
 0.5026621026881061,0.15960852558305372,0.9985026006102597,
 0.37064276257285433,0.6822900013026801,0.9607967786094543,
 0.24442942827949865,0.5832920278914283,0.3069981723501539,
 0.48527380026387945,0.16596232775603847,0.9408605842619799,
 0.3559099682187392,0.6081526224410063,0.28034573656943307,0.616403590611068,
 0.6670990938962783,0.8097728563299221,0.35136093912398936,0.5939916174447768,
 0.4341665868022023,0.002875601695436214,0.5011017415786084,
 0.9622555208381757,0.9754374099866355,0.19560733090651683,0.4517883504874879,
 0.050053862280219796,0.958544448800613,0.4769646598943287,
 0.12479538960606684,0.6519593655511994,0.21925505200405448,
 0.5136130743408951,0.5807942280838114,0.11526683728284182,0.9314226075164634,
 0.26366011587951754,0.6287407845636923,0.6868466142853725,
 0.15010952344766915,0.21950806358583752,0.3998344168231096,
 0.09205757156747152,0.11948249520180498,0.00747030387056602,
 0.44622001438704706,0.7438855753788329,0.7193229004398052,0.995512830083763,
 0.5253502574018624,0.7775100673399017,0.061274744116301516,
 0.7649322236126338,0.2490127558213001,0.7480505259731537,0.7868761157559802,
 0.9575351148675626,0.3624916379603318,0.7577602219461238,0.93786252099943,
 0.940134211474168,0.5720340382787135,0.6151506970306866,0.35423948917908843,
 0.5686875872264185,0.34327537468173897,0.1580941831863426,
 0.36470150094232745,0.5879159560338254,0.623803077221111,0.19175685942371723,
 0.14623520797263712,0.7009784266676384,0.9267711992134926,0.6926065436251776,
 0.6162320969872451,0.16094580465030806,0.7404109045359271,
 0.30304607620631296,0.7660212028736562,0.3240230432219372,
 0.16754231618736026,0.8689188628773858,0.6047871821730342,
 0.45932797314971674,0.2851951153164639,0.8044823271728836,0.3167456197526165,
 0.769132212939621,0.9242335012613132,0.4940135541767673,0.4840925190307659,
 0.34255317394193174,0.8313079969770145,0.755488778196506,0.7929345240302716,
 0.24933508401129145,0.8687429869098725,0.6018794800795308,
 0.07313939836063033,0.2560097773914478,0.3161238552643135,0.4853390358795401,
 0.15957981127254328,0.9620167144450483,0.15620977410425263,
 0.48850443356618334,0.29656918843329005,0.551580610261231,
 0.15483006584964987,0.3400425257058831,0.031388209590113236,
 0.5843437098601959,0.8905222362153138,0.22061781942056635,0.6799167599322135,
 0.07740812249601192,0.035321365372439795,0.12138696698665408,
 0.41697087993352977,0.5610771554123641,0.7373132101838917,0.8029381691057504,
 0.27094379603487884,0.12114740572439697,0.5968580441265349,
 0.47695334024283775,0.07336172386816164,0.9757122792238605,
 0.5585071782875508,0.5903340842163987,0.36144215544243996,
 0.03550875841920276,0.4068905384577213,0.5092666830860995,0.9238652669174924,
 0.4098252276379275,0.24541056952801743,0.34802018610462715,
 0.9593347456884622,0.9786785698097278,0.8734976476646998,0.2209883248678166,
 0.5421867069201235,0.6416770509979314,0.994568106541352,0.650326568475966,
 0.5248461663490918,0.9154612857744219,0.21015281242436856,0.2267561292023358,
 0.20389772841110176,0.6321383710336345,0.7207095428891499,0.9058305420904075,
 0.0019133046570433132,0.6730807488998567,0.5722277365702372,
 0.6185746272845802,0.7864695735952655,0.11441760122609124,0.3602971673735438,
 0.3323179162997929,0.39731207957354586,0.46163726274799743,0.393540613553199,
 0.7283510426457289,0.21623660689831214,0.06464973602347435,
 0.7406744127766363,0.825013080026132,0.02824213341329862,0.9864463674269215,
 0.17996281299056172,0.033858850622212056,0.5512875568160567,
 0.9078665222529418,0.5633852574989282,0.4497579644839379,0.7464654724372081,
 0.4723154419838539,0.3924091705499968,0.28846989464314376,
 0.42568469196143777,0.5558037964911204,0.9702578335645443,
 0.18221066244585937,0.764297537656085,0.674300692276779,0.6738581740365175,
 0.24752464496589466,0.9734574244412142,0.4533932816628652,0.2453976686828946,
 0.8520744754641885,0.49545280248266477,0.9763421545394478,
 0.15691386877226676,0.0846918121311302,0.1616703741346398,0.6756021009268757,
 0.46199146931884183,0.8102554682703733,0.8220201570807846,0.4968198766467785,
 0.1009635019981927,0.9502814104865077,0.6860586497186795,0.41108681824092874,
 0.9161188805398779,0.7723312559502313,0.7332788081040944,0.4872412750264655,
 0.008008400765094792,0.9364265860322045,0.45038360634624697,
 0.17570120843382142,0.25566666437194074,0.15603185471327152,
 0.4951867365976955,0.21499614323574456,0.4435703242377037,
 0.14493093451611994,0.9499095868665031,0.7520082716557825,0.3544009737299,
 0.71045554140092,0.8821449330418413,0.40532093488421794,0.11184510964039507,
 0.5018035509771165,0.6379277625219785,0.7163557106491408,0.5903540794030732,
 0.8315787943598236,0.029818898241129044,0.34458925114502015,
 0.11362741160484391,0.9852869105386903,0.22166774798071032,
 0.9326066287575513,0.7694045259004871,0.7850222848790215,0.4904041301892442,
 0.7094202922860733,0.031538666459216436,0.8731363223502207,
 0.9655506897325044,0.2393638115114899,0.9283592976904449,0.7966897159432695,
 0.3927084141138222,0.7991145516691337,0.8211835919368939,0.9406790005849978,
 0.7971803027437897,0.7843862769081686,0.5767242095217313,0.8678821358767529,
 0.7944334035148773,0.98451112277769,0.3332211773677287,0.7544862743407467,
 0.35735691244340073,0.512658434365872,0.37853727662188297,
 0.47127449802660937,0.6668468335374362,0.2170176637704604,0.8402029444050974,
 0.14461198035426115,0.8288408917723729,0.07908201324803121,
 0.7996078854649546,0.17232119011187041,0.1604476279918291,0.1245965296008209,
 0.2427689296720169,0.6271032260949563,0.7151256038627066,0.19858922123008393,
 0.5878560389372067,0.5526168100235316,0.5938911317903584,0.7731404460657325,
 0.11997680073225869,0.8916579808890072,0.796199958557684,0.9352788708822587,
 0.33208576449083205,0.4149226088097083,0.119678246140145,0.16835354089671561,
 0.35516817254104904,0.5742309883476697,0.6188928936439341,
 0.09461354111655285,0.2969350326958209,0.03875717015213387,
 0.31410717123879806,0.0987228387071466,0.7405150995751775,
 0.06959351985546591,0.7734591736662141,0.35239562430399485,
 0.6590018231192394,0.3325627050512904,0.6324636751839914,0.11186132526451287,
 0.32931361724784436,0.07690188028468614,0.5748037787997806,
 0.3459822658715348,0.35406849085683145,0.7337755753287867,
 0.32833391924304856,0.7309306929448715,0.2725722589807612,
 0.27898424985549386,0.22645426983196348,0.28904382648617133,
 0.5658915956381257,0.2227093917686147,0.5429294368038919,
 0.023046730484953093,0.9549562891674005,0.28553664653951727,
 0.9158817674613685,0.34953070159514965,0.11719290262308601,
 0.7657621103579633,0.8950406080176782,0.6916624322648361,0.9977087325445448,
 0.4318508559357783,0.6001666602515011,0.2993796704553098,0.25499857877884946,
 0.06850199159937354,0.3473972522399462,0.6997542591597627,0.5253385736931379,
 0.3634972597770222,0.9596110210730107,0.2821976886311104,0.8782627770034029,
 0.7060708447896653,0.8533371583471394,0.5268595871809457,0.5441877531295178,
 0.5905828335877318,0.19193517862390486,0.04440968916372112,
 0.44594138859286026,0.19257825093148995,0.1321237369702285,
 0.6373638705957053,0.9035226019234285,0.8538526922180226,0.40491377871094014,
 0.18832812685689326,0.045044358928918116,0.5656826188076614,
 0.5537337815329065,0.3342461661789622,0.4746115573004859,0.673064645824128,
 0.8739743736728955,0.5770991933335304,0.9138057712979085,0.6990964188512792,
 0.11563858198834132,0.42297460618914196,0.8788951197831599,
 0.8278073204127543,0.03200112636115893,0.9117063011527271,0.6689456969244771,
 0.8062806257152518,0.7090647822164784,0.8582104516152721,0.2456453796217214,
 0.8907573800649455,0.04220756992464514,0.4881149239582373,0.8255685129240875,
 0.9646697463731436,0.8727294291302373,0.10861145482111101,
 0.42187509266565626,0.15109105656320831,0.1570878127789036,
 0.2548825673294438,0.6658947813684679,0.1779398722844231,0.4976868933084603,
 0.5244374844480659,0.13027247864176195,0.14825777526001815,
 0.5621379127068518,0.6799424174343753,0.8523461398427947,0.46008787439909615,
 0.09111063029495681,0.9197769732807549,0.8247107278272129,0.2583554591660864,
 0.9598213518771381,0.8139084297181568,0.7164431794303268,0.6028293699652534,
 0.1798601846259048,0.2545658361719254,0.9988984994690674,0.3646812602671926,
 0.6303365662152743,0.15750671729756405,0.033795690283323854,
 0.600892742211721,0.2267143839783755,0.65998693070626,0.03195189049293867,
 0.21333866073711572,0.6504232324788704,0.6005869724652689,0.8385529137609371,
 0.46810718855041733,0.14724125632722607,0.6557753310452289,
 0.47404281751518107,0.03483790637514639,0.15647690375850432,
 0.6597856160992485,0.6255119228319154,0.5831051248160994,0.5172468437919236,
 0.960263318568627,0.33432281034394273,0.8741737507321204,0.32489729605801365,
 0.38684979486919013,0.3818236003208816,0.7163882045867083,0.7930394443982884,
 0.6290550737421076,0.29719567462224183,0.3926687213754968,0.5983790624404164,
 0.15739926364490275,0.5546386597295317,0.3778132151992738,0.753684957421988,
 0.10966365436792658,0.29649133302473263,0.025128570326591104,
 0.0442201029446152,0.8546469160499284,0.8211703863992448,0.6315202192283643,
 0.574341063982994,0.08520284055339322,0.584342776426907,0.7564630042768793,
 0.06898489610910752,0.12566936894989766,0.06452432614754666,
 0.9664423075522584,0.18692711549733376,0.9164751596844466,
 0.24961784526150022,0.6197204897584637,0.11562214712773022,
 0.20111022774170817,0.646756887168622,0.7393731767427552,
 0.0018822158267286504,0.1306627072922364,0.52944316158034,
 0.16570367047884782,0.5661397630005901,0.40955623066764335,
 0.1256282116157561,0.6080873741412106,0.4363467242188175,0.13356694042895156,
 0.3393617087597507,0.7996791223431475,0.8413253823488472,0.5626379293795699,
 0.08695282275556748,0.7534547777539011,0.5033571078636038,
 0.15366831292101524,0.8911364973413604,0.5140037387680667,
 0.34072731990652316,0.32958054186787733,0.9182871972152113,
 0.7907627666751692,0.8118122735094018,0.5005430032235694,0.21168598847155196,
 0.6253383362677556,0.6653462073021981,0.5990440840147027,0.6583397318222607,
 0.35909958280240306,0.3915507708731343,0.5992567262379039,0.0234056352846278,
 0.10775711334828686,0.4902372000039691,0.305359842630884,0.9122982319813157,
 0.33634996340522205,0.11581660154567486,0.8082615474021605,
 0.4965660393653025,0.5295932961237727,0.8756540333182068,0.17818704061664625,
 0.28391559226143404,0.61628239447944,0.319844025459486,0.92025675564737,
 0.7372972130736575,0.06328200774955928,0.4535703814999633,0.4215334405962078,
 0.5232982371860773,0.7336393555884805,0.10958779291816623,0.8418148936121718,
 0.022713177489211867,0.6470525164523029,0.046802896606235045,
 0.9680327394562963,0.9621824561566366,0.5420858242671315,0.6315794470342692,
 0.5122494281055696,0.37936159939031133,0.6920757180114405,
 0.38239267263634813,0.9435223500432953,0.16647827547710947,
 0.7818842800822745,0.11546330026732177,0.08912265269151609,
 0.24366648517344958,0.8496022839397604,0.9529692912805168,
 0.12738312113711858,0.2858495544242148,0.43227485631358886,
 0.8158474103806682,0.43722190789909865,0.6902212640781884,0.3219477669262423,
 0.34617648168351045,0.6190317611497688,0.4631441128330207,
 0.13399984437193146,0.7550371198666531,0.8718293568801248,0.5584719554357968,
 0.5501909373890954,0.8266684140668271,0.6032395045453426,0.3479441224139518,
 0.19732204713464396,0.5663554086142111,0.07963760556537103,
 0.6290482347714513,0.4246571747240182,0.35084767935357497,
 0.28491874205231726,0.07682369183142024,0.33270953879471377,
 0.5083647217276173,0.0832318739500133,0.8651826673593673,0.2968886114907351,
 0.892942503687985,0.3748846717247942,0.6921196152459823,0.24833706832899782,
 0.6461452712814618,0.6534761220357803,0.37151520763906687,0.7652859232798871,
 0.39802043252883845,0.24219730688346042,0.2766169853228543,
 0.23225383872268912,0.7574587910703914,0.5811395515039437,0.586399590964809,
 0.4740142622248814,0.29734866026769824,0.6684644181191609,0.8689498602076534,
 0.45003741335906067,0.889378380309162,0.971487326075031,0.9214947882929021,
 0.5309154964275286,0.2321671244899306,0.01193518940307181,0.5603429850378123,
 0.5444473871697196,0.4215353332054941,0.3600567089356336,0.23720056622715113,
 0.3012689647829424,0.5745015452365677,0.8861053885652945,0.8448479512020396,
 0.4205314232515681,0.49400723774291144,0.4856068254469541,0.8268563882235097,
 0.49325729758011694,0.47764586178337665,0.1234419305268124,
 0.7448112946773342,0.9051407458026208,0.4207987458237926,
 0.005529265663627547,0.9163019733870067,0.790593358477612,0.819590399142161,
 0.9327860656965673,0.09691809961096343,0.9389598775030207,0.8391770742181583,
 0.7358338040903097,0.7610381206640753,0.6512648816760445,0.5201804678536379,
 0.30652428751488414,0.7706190407078771,0.6893533587740273,
 0.42434359153349765,0.7076881032956346,0.16557696575609537,
 0.8508072099815958,0.373213664969395,0.2896644421822221,0.553241105554734,
 0.5355178249732624,0.2327184828264247,0.6323777595980119,0.9721375011598239,
 0.08944177851765334,0.14935452878169408,0.00610319573589857,
 0.4551040154314354,0.6720831246755655,0.9055738470325106,0.5935415667658002,
 0.2851865401981971,0.655286878492735,0.0811774814744387,
 0.0013864152267384355,0.30349125173590874,0.32435961906163613,
 0.26340340290012865,0.6185223602287653,0.2927977123595149,
 0.49013060474469416,0.9977881413570314,0.6281975398495595,
 0.031805666608710204,0.8669616812774216,0.4571230270648581,
 0.9835601006932949,0.12971296468830507,0.3632132934183213,0.6700195263535897,
 0.07409162981952178,0.6016064505690949,0.8371867857574031,0.9789484277676155,
 0.5865567727527612,0.4599072594258602,0.2612230161480278,0.5251671948371097,
 0.5323485260681865,0.01945379988578766,0.4682988093408068,
 0.12302939297517856,0.1346639248246938,0.8123244143448721,0.4718947937720649,
 0.793962891073236,0.06596311097720853,0.48412767044324656,0.799443878558341,
 0.3775808269453017,0.8024925705777628,0.6867270193445395,0.9745911979923056,
 0.6566483027587534,0.6557836576213713,0.46100438726436166,0.3565326562314062,
 0.9563905490175293,0.1436731061555001,0.530779832813382,0.5348132238581547,
 0.1815514514620178,0.1678050505209746,0.4489362657007917,0.7602615583321932,
 0.3322713327464041,0.3509201834254352,0.3041495772940236,0.3930051931573977,
 0.6715293983080877,0.6734982589446158,0.7703267744286109,0.5619452898212868,
 0.03556930515266632,0.9277942082453087,0.46494541192298566,
 0.5378330712128963,0.17580716005071828,0.5715350550111811,0.9529375911685205,
 0.8415269728883429,0.7852120377630214,0.09801145231505293,0.2731622347040098,
 0.936985454323434,0.40174447643593314,0.6058787159572228,0.5835045520968076,
 0.5704225051655892,0.2520675891064946,0.6818291702786047,0.4548844779150252,
 0.6052936366824183,0.9304083624311718,0.20457909518664186,0.1492775100155187,
 0.5317912397398528,0.48190892739225555,0.022892685488390407,
 0.384860753853971,0.11257065381001885,0.7300977288542779,0.7645419509360581,
 0.4465449537943812,0.8434466082833558,0.6720089678739117,0.4399982645253775,
 0.5491278491450277,0.29905804214179976,0.5265878042134245,
 0.27516121461413023,0.13331371844007978,0.5879886487751063,
 0.4212812712373325,0.3970719553671316,0.9880205676423744,0.6333425607230264,
 0.2308041574853844,0.26113237352468244,0.644353431953858,0.182755670252863,
 0.43580866829052545,0.250051127444213,0.8438089992197627,0.2573912860540395,
 0.9109113507896002,0.40837227150728383,0.09772769683802218,
 0.6808588717031167,0.0725255687990195,0.8120826694355902,0.16142904030485705,
 0.31303029638560076,0.13061614394240506,0.12149749213077699,
 0.16739719294369015,0.7909516777992713,0.1819884544184649,0.6779359129945203,
 0.1504232749338239,0.4337467907204332,0.6692962052296061,0.24793160374336654,
 0.653708130588476,0.30418220174884936,0.6123650084225888,0.4305679588901017,
 0.469569737709127,0.22099176747826543,0.39474033238886896,0.8505328186581246,
 0.6654360747164239,0.3564810435439313,0.7913618530936412,0.9127916712664728,
 0.6311348351083608,0.8163474900277362,0.17535599927124434,0.8213702445853912,
 0.4237035663546854,0.7283772523726578,0.07429004502295056,0.26155513726112,
 0.7608910018010733,0.19323760078494456,0.2390131709352441,0.5762108143121056,
 0.3648311822569046,0.1289964761396234,0.38193501889695125,
 0.037787007344560974,0.5041209386303797,0.08741164245318944,
 0.9311921357974586,0.4945384767442996,0.1354059295547415,0.16757912556318932,
 0.4768796153619237,0.7097661016054284,0.3643849234563393,0.7818875086272021,
 0.458110414923967,0.7226477223363388,0.22011928410105053,0.7763078181795993,
 0.8772677888950347,0.3364066273169548,0.7296171293657678,0.711517731950001,
 0.7710371660375381,0.30425059157129797,0.5324739846940628,
 0.08012640745000288,0.2476360732252788,0.37209378875393706,
 0.5277556091883496,0.9987635996332295,0.3271700316207179,0.9166928481397474,
 0.9044750759843526,0.3094733091578894,0.1286356027225266,0.7799679352979966,
 0.8577027724064814,0.48179827545906273,0.8316991631948965,0.5736354047478974,
 0.44390633747826647,0.6704973065571984,0.9465736818779135,
 0.23905961680227972,0.15931856342225426,0.4992423513475007,
 0.13154824237635276,0.3713817590141547,0.4276347473114991,
 0.09518994596749975,0.46445949233157224,0.2706142292967284,
 0.19942111992019818,0.13479024358445468,0.3927698184574795,
 0.8499209661994499,0.38551962918073035,0.735120563320029,0.8395692495821955,
 0.9239942148357391,0.8320165369692449,0.2468607716594714,0.0080640147917177,
 0.09168525325994747,0.5851516589281609,0.5469003704649631,0.542736864654005,
 0.5094186984217288,0.7714681910879932,0.01468959948468751,0.7745307281768207,
 0.3077024581644753,0.696649228949554,0.13287808874646434,0.990415638410566,
 0.3772040370557731,0.7635651513206395,0.7703533059991066,0.6626572259651602,
 0.6819259714245778,0.6220465376196033,0.9104043118429189,0.5241262266368079,
 0.7555293666320846,0.374258256321938,0.9161413273542494,0.20109529386301217,
 0.9644274936796288,0.5067050211891229,0.3149243386793027,0.7547108175019305,
 0.1090730915036302,0.19335750794351214,0.5727030951077251,
 0.23696850263203129,0.5465330201747192,0.8527460833258633,0.5795330909932164,
 0.07785409406186605,0.9807011352843414,0.4028011458308337,0.7911986047483762,
 0.07887084308496939,0.1487884615486984,0.3189781056158959,
 0.027756228401562333,0.7601490101153414,0.24620559465808012,
 0.2623358846817777,0.34350371180826633,0.27524406413046965,
 0.09735152975676742,0.39592750128507936,0.08768309855719114,
 0.4042320038078556,0.9398145037537224,0.4983757271999981,0.47789474081216843,
 0.10750872377315512,0.9472373682694997,0.801063890930693,0.9363514607078163,
 0.37172901660067714,0.6074753088859921,0.030163217839322853,
 0.4046362915926798,0.5926873975689372,0.9118746494265298,0.22627485561595506,
 0.6090051750480383,0.6958884806904471,0.09056259114061749,0.599418255281296,
 0.5419408279780707,0.9775126105724905,0.5687717799937703,0.340476174502074,
 0.7897503710889049,0.30226487271873215,0.23192147870879742,0.556692586689652,
 0.6691673237709518,0.45068969700979644,0.46089274114379275,
 0.22582275119052397,0.5328608698269847,0.039688704763765514,
 0.2130805338375259,0.6280200296254598,0.11403905040664264,0.5779784462354798,
 0.12399880133235819,0.3690672234896263,0.5572807572485227,
 0.13925357166418706,0.8132948836047583,0.21910682485015842,
 0.2707672332386658,0.9714746323016915,0.9702093009515348,0.2829723857533307,
 0.3783974697843129,0.6750186875868684,0.05464294292526961,0.658434201042001,
 0.03152778810522516,0.18544574983396833,0.7463005505782428,
 0.5152275902799112,0.3052881417222353,0.7619042018090678,0.7540803639456327,
 0.4226338970903355,0.5106515023893552,0.9373642879938967,0.2695596568614912,
 0.9934166808440891,0.24441398147901183,0.7890244749656826,
 0.16631626291992896,0.655485232434784,0.4653566845625311,0.9988291504069372,
 0.5264980514263082,0.5478782033285622,0.17209225547235907,0.6724368487436374,
 0.8892507857547828,0.9602807468845491,0.1733518980351738,0.29576927312101065,
 0.8915212962479878,0.8206457852782172,0.584608741216823,0.7058647229529056,
 0.5081194780123623,0.3202008990884143,0.5830603379696155,0.38188828667581065,
 0.007089691782481866,0.27223810842488105,0.8324491737602739,
 0.4565148558989378,0.4524926445569033,0.2618861792536249,0.47520292150888954,
 0.29470691014539085,0.14814602526831433,0.559952171288277,0.2729697936162524,
 0.5682929312880504,0.13178339767635994,0.5893213627587225,
 0.15926521759505385,0.7317912016104395,0.7144491447819823,0.0579241167764305,
 0.27392906996072275,0.6726844096489905,0.6764938237858347,0.3236784677213558,
 0.9882724960001441,0.9165714805927877,0.6090107506069219,0.6526975332233871,
 0.0572639640701923,0.460350740131823,0.8047071129119665,0.020690777091382895,
 0.6840323676750972,0.31685887177565397,0.102818299105367,0.9602714939856364,
 0.6460034299451705,0.02020092474396651,0.5638176677058484,0.7429772235177997,
 0.37329046720076153,0.640782744188424,0.880596961638427,0.1733636218452036,
 0.5109573498253103,0.9327799473390815,0.8306292053615684,0.7403323226642821,
 0.08876172923348824,0.671078275669925,0.2769223181496745,0.5218722134274904,
 0.9667593288631686,0.891706159215788,0.9082152994153166,0.09093301226558292,
 0.8011678527084947,0.7871154368485564,0.30505820088350466,0.935382874106189,
 0.485352811049503,0.5974827700726361,0.8436639477058356,0.785036456698539,
 0.8687390850560347,0.7844194491273349,0.6891390101223289,0.8388389299820428,
 0.0042682434644408795,0.5852769541415426,0.8724929339711285,
 0.7001132076506519,0.2636745223077469,0.5432065217181015,0.6692421233828092,
 0.06215677537378017,0.8376640223539537,0.992059821639986,0.2785645315137928,
 0.34340864286530315,0.9978492138555957,0.43380486151778286,
 0.8135878804623682,0.6289208240926737,0.331701980564577,0.46201476003205544,
 0.1434657893750041,0.5741089798046377,0.7095279574080642,0.32948190063126126,
 0.12017856947947059,0.9846152459344742,0.8576560991373002,
 0.24894459807996105,0.9424304549097375,0.8147467221250778,0.2458269094635579,
 0.2273648811274923,0.04453943519607528,0.5147940829273653,0.6558839672992988,
 0.7274143304892903,0.737373472841206,0.8065930863961964,0.21579937681003647,
 0.10192436009715533,0.3226401249833205,0.8365123866603874,0.2359842869549531,
 0.12456114406543328,0.8203160576870601,0.9213992942000153,0.652365233711064,
 0.28396455328343795,0.20192775593771728,0.5062539333022894,
 0.6318034967935358,0.025988437367009287,0.13959944617682662,
 0.34235632316370657,0.7772010299244285,0.0623397487366909,0.165522766607485,
 0.1181175967328707,0.950320391504277,0.2907196501713748,0.07152813573552208,
 0.07329703110510766,0.801655541484323,0.527292501593697,0.8206113157956866,
 0.8767783498758179,0.7169216392842481,0.6146246782921003,0.6520341523112893,
 0.6988939364773894,0.35783957016863543,0.04320972338449924,
 0.5668022027127803,0.683324450135572,0.005988099736378372,0.519714648104077,
 0.9906852380977176,0.8787237139733141,0.8799647005722705,0.8850311420445769,
 0.5726276377787759,0.45253035028076005,0.9565440174310524,
 0.43991323740049393,0.7931174625669689,0.9857183602478883,0.2741634031908925,
 0.6376431932443376,0.7626869199826951,0.48394954136515056,0.8623360731811778,
 0.8808626337035301,0.5064285664496276,0.3052921465718148,0.8123174982089164,
 0.9662083452396752,0.9980094216255232,0.7930234641257005,0.11097436728592325,
 0.1628142689331601,0.0296800044426605,0.26135590629944117,0.7412370867334135,
 0.7028811837372103,0.5752650851749418,0.5425355846518556,0.1278659146097041,
 0.28357775427430476,0.3290296727754345,0.5701053798597047,0.8973892460499397,
 0.8810542901132853,0.66047003650293,0.21077700651185327,0.359796109233982,
 0.25808784348354963,0.6415551294284889,0.5477802468931365,
 0.34624755709650357,0.6893795594570382,0.7038026100879906,
 0.20760632675255475,0.7490617322404283,0.025769024494335935,
 0.745915549556802,0.8529725882737288,0.46899722982760605,0.6928024095920603,
 0.6300698211450794,0.6678503974055838,0.6263517179125979,0.2886513277597006,
 0.04816549007073767,0.6418978260720942,0.7200894504021115,
 0.33474514855339244,0.16685581898573842,0.6172140780455319,
 0.3413412420437033,0.6078677098635439,0.012666601217195983,
 0.004244789988582465,0.5046198565245413,0.5950934129679708,
 0.7579936188434467,0.532404276576747,0.8100418967159874,0.96156753725599,
 0.8808125690136007,0.8631703221203025,0.2847617023781941,0.44790754334953675,
 0.9450956099951479,0.02966353087001594,0.24474437987908315,
 0.5461623693956926,0.8093390854046778,0.27714575332370994,0.6265320973206405,
 0.8887863388909765,0.24091782819482288,0.7393707049609852,0.2012338907501462,
 0.8478140662097146,0.7479135094551709,0.8796322340162198,0.20126442764674202,
 0.15991996147256038,0.6617438748657485,0.6172130946832762,
 0.14863224940095254,0.5394437911634347,0.33124812351169264,0.988646474211447,
 0.17215512902655183,0.37480902248093984,0.2889844046511456,
 0.07427925709437444,0.82914557591636,0.8504089888732574,0.9410737451685242,
 0.993668720960061,0.5758728652148767,0.02918117211375071,0.8678388130956973,
 0.755936749619975,0.29057997079404596,0.16408433108379195,
 0.09528676599525765,0.42638144960272173,0.788103327525839,0.4906712957411319,
 0.4895628709731795,0.8245291149736623,0.5771241330125583,0.962654207033982,
 0.6762506839113748,0.7658105949650627,0.04502211212616236,0.9372386621325461,
 0.2802076685946091,0.36080526366803856,0.1973543325438838,0.1790343141631825,
 0.3959503012915826,0.3393446628109803,0.7806675313497715,0.31093501090889075,
 0.3981679816131105,0.01420275884283273,0.2611696360358836,0.9212441261590361,
 0.2634376041934363,0.30581675040883427,0.7772198872332233,
 0.12071974119529427,0.4252937223480179,0.9725930113346583,0.7691653948185452,
 0.8415070689115886,0.9272247598460011,0.23413833529288286,0.8277973959275734,
 0.4002894145465137,0.18588480702170063,0.17682414807402658,
 0.7602874918965576,0.32509806980885525,0.21402151167628836,
 0.49973056127866866,0.714099374649691,0.8746492807774485,0.5042964228864505,
 0.2861164416927665,0.024207976724803126,0.7558193671997718,
 0.5370611972890658,0.17650073461872762,0.19169970158979088,
 0.03156088001830881,0.3190669722694388,0.18249620096052188,
 0.39908866360452766,0.40537751470736394,0.0371885646825465,
 0.18544322825318682,0.15102424511031742,0.8568338226239505,
 0.4063123940506047,0.8561469084023069,0.7457694792150542,0.2740454787587825,
 0.7384322857982831,0.7001888618860345,0.39160552543826077,
 0.26768100493811886,0.4209957412227783,0.7879742752913749,0.5300314184586452,
 0.45358709884341,0.6982115459033695,0.845473371560753,0.5568448998802927,
 0.1357827844287396,0.5139921100153025,0.09378096428107396,0.5867754603092632,
 0.46928779902994533,0.5860532670239994,0.40714805688766664,
 0.5316358225878515,0.8637629960491534,0.44595538874120044,0.5285054596779836,
 0.05519016962266232,0.09730007044972377,0.014413820918727005,
 0.035813356156810694,0.05564573994209143,0.40385318136876835,
 0.43036951299421544,0.7734366353607862,0.8982109118293591,0.9381743053392279,
 0.9587810361456645,0.577869512796401,0.8827686942832397,0.9606857127583366,
 0.014303299661124202,0.42909334730574833,0.8285511987827922,
 0.03154333732267556,0.2588301226319748,0.48311308286258836,
 0.5075022409528792,0.38701138028879967,0.20577551637181257,
 0.6546361530791036,0.8499826628192438,0.7625052092944875,0.6070693262145124,
 0.5839442708605527,0.7914108090422194,0.10638173347749047,0.9584094008768749,
 0.15952723233949107,0.17559065337697888,0.7593967181598662,0.723858252942626,
 0.5277106458065507,0.21878091524419196,0.8199027576884564,0.4625988151984558,
 0.10041573966507367,0.050366532060899805,0.6735809948169174,
 0.18868630733722835,0.30622335428487635,0.504137372645015,0.9453219292603904,
 0.04290536193077721,0.8579237465355891,0.424385018722353,0.902065914000185,
 0.08062264533554842,0.7642538370475873,0.8575929193267763,0.5634494436530961,
 0.4529128908157737,0.9560324981511759,0.9652896489359281,0.7121913406811289,
 0.889110628070532,0.753233887267807,0.8453838333452914,0.09959558694930037,
 0.3793810427063672,0.2938682738963454,0.13208599033633706,0.1907884004189464,
 0.4583007436319342,0.6497251194591351,0.7628147853553533,0.9629657100558598,
 0.24372332484890102,0.158871079054129,0.9571593875384089,0.19313127835404065,
 0.4698329562313791,0.667319313948076,0.5394646921079012,0.7467613949435352,
 0.04395473035483455,0.25139542591859076,0.19989639954691474,
 0.3078523818432104,0.12519670269954553,0.6231994302540611,0.9799697005013451,
 0.17826578421644868,0.15753357377245114,0.48187935261335646,
 0.8847556411190781,0.3057428337035899,0.2764669559899111,0.28270203533459504,
 0.04270490665460469,0.022167323666791527,0.5960580270940968,
 0.21821735820953492,0.6786932862199312,0.18236670125488186,
 0.28106808575743814,0.7436076057408982,0.7662973888098134,0.8188849077499467,
 0.15328998189141718,0.3536672465552868,0.8359345480219182,0.5774254778003997,
 0.9215501072389654,0.3798232739051699,0.19200084202557044,0.3267614891778682,
 0.9667666211717423,0.41354337020311305,0.07821583163052959,
 0.16937658339273787,0.9270076662140085,0.8094603520604219,0.8283771145708549,
 0.7762121264460977,0.1797902254497653,0.7332527450425047,0.13100404303563717,
 0.6669015524915023,0.24284533294220945,0.6816516407989023,
 0.45617843413069037,0.15720374766780665,0.2676808010063437,
 0.7250871738338269,0.504426907027299,0.7785644488202651,0.7178392149778198,
 0.8396117955576061,0.48767346639179054,0.8866347972190496,0.5280191999051784,
 0.6211449158592368,0.2820353880639008,0.8455279332854309,0.6770676228363239,
 0.634200063470369,0.0237979516850666,0.45214288579935324,0.5728006967207868,
 0.5819506956117481,0.26084769708297195,0.3911873814854395,0.259010961871846,
 0.4921873893998724,0.8811763027029841,0.3101116497931815,0.5671326407150106,
 0.11288186249572152,0.05481629994982329,0.9188545159494415,
 0.7891854548913944,0.6980655236504135,0.4338751321772625,0.6326893879079115,
 0.4629512891893818,0.3232481627755489,0.20486844860475684,0.49164019413477,
 0.9575110967276934,0.9343244827112057,0.3712626887981437,0.7768947394904409,
 0.21933251304549373,0.44816457722957015,0.3101054041138577,
 0.5667243566133173,0.08347962317274749,0.8078072473642426,0.5631195636709554,
 0.24217052371466097,0.5129804832736725,0.4193079919998213,0.2525291606852238,
 0.19264105721734381,0.8041188606197586,0.8360005588228029,0.4419408088401391,
 0.5840058100544525,0.6061105052945689,0.5228478915101692,0.7555016750105316,
 0.7091556320954844,0.9633866144454867,0.29893499489528064,0.8642873252637228,
 0.5506078524904787,0.8122076310218301,0.8305150808927139,0.27474778988562243,
 0.3847998663737975,0.8659209458921281,0.10713102241654227,0.6072395263370448,
 0.847582633559934,0.37794485632276587,0.7921947739920012,0.5187624840327711,
 0.560978278890318,0.9591081898339271,0.008025206857952938,0.5747977080474207,
 0.5784860271093359,0.4278909192853533,0.6870320564737644,0.9377556261296202,
 0.886891435039374,0.40996698000054854,0.7786498929926948,0.6488232415735461,
 0.9088189594248497,0.5543653592509726,0.0966735357109002,0.9309225737062302,
 0.9994543933113498,0.9003857378000215,0.531361943688707,0.051834141794553035,
 0.6726442824659412,0.4448307047575951,0.4758104697221148,0.0771018206529952,
 0.8221922907144906,0.17479640185778245,0.47402595019367644,
 0.05556314658993089,0.5043119392741695,0.9278484667449104,
 0.055776816749059877,0.05619638431470375,0.14481555438575633,
 0.9620325871528532,0.9043682784757459,0.46218256121315004,0.5797815185338336,
 0.2701148524061444,0.9005121432032374,0.8895605102046396,0.803725758151248,
 0.9714112083370523,0.7899830074340115,0.336067143440296,0.12132193126270507,
 0.42271312589531074,0.8712716371923814,0.2632870837068453,0.8124087333185069,
 0.9912083103816757,0.9457765667901559,0.860591518058683,0.33504148243462717,
 0.8988055030439713,0.9046654303958647,0.7128070977221427,0.41905817660748756,
 0.44909298609787873,0.17623506647551979,0.1790320658416742,
 0.5764249962021462,0.7992974287772213,0.5020796252247828,0.29573877207181076,
 0.3214760073525935,0.06967827795502768,0.4127584490141337,0.5781119973983512,
 0.5228437824890034,0.9024940115411922,0.38635078796797684,0.7648950698385917,
 0.29816654956346533,0.941617022346738,0.9060169846627366,0.49306897286399876,
 0.6290341974507041,0.6964643120588268,0.7225132415151512,0.5960794733828331,
 0.8910460433294135,0.29997945582071006,0.15982319595047645,
 0.15216113260013975,0.6179030123691883,0.5753869477283639,
 0.33366030299400595,0.4043888727743299,0.6820004063661408,0.8542812596299404,
 0.15014640877002816,0.16627937291402817,0.09811356665216475,
 0.8835524574337853,0.15886270447367812,0.22943931818339358,
 0.6971987739441952,0.5699145286889464,0.9995808403174862,0.33617771441469946,
 0.9816425600667654,0.7512690200275145,0.6540575231469105,0.11862389641220117,
 0.48457383770577156,0.4197971267261601,0.7048061750813178,0.4081131501661369,
 0.960782796357562,0.42850744065607793,0.024940435150640483,
 0.8440104607268186,0.9593185738528309,0.26219268828092734,
 0.28348063560285597,0.6918326998830676,0.41360714969493784,
 0.07351940505098375,0.9933640014114371,0.7105301409769856,0.2307269952619213,
 0.9692781497932461,0.1936825286961239,0.5761778196016857,0.4381785351203501,
 0.23246539181578618,0.9648063678819448,0.2684071596187094,0.4185720894677186,
 0.6840787911162788,0.1678793356329369,0.664306317005322,0.9467560049593391,
 0.4577794070410003,0.024942167502202017,0.5776260477755328,
 0.9426726127002694,0.017779507195177735,0.0840027941604744,
 0.2976011768776162,0.2286632706348819,0.3862581487727571,0.6901861620956489,
 0.9071678313387769,0.009899284900866778,0.5683744544015673,
 0.44431330471383346,0.7631986765543287,0.4550424873581491,0.3259908495248506,
 0.7454932236808354,0.36083623480378024,0.9886065717738305,
 0.18862288967963659,0.6684894442470279,0.7697882058992349,0.5801497206391352,
 0.3792528916589474,0.3386985683706504,0.6932888213622471,0.7472068510841128,
 0.9932442406423289,0.6711372077012604,0.43113711451396064,0.8870685943967732,
 0.015301084427448153,0.2785473371968651,0.26439190339747654,
 0.9287349794257523,0.602921915738412,0.7494617344924216,0.9425791523637947,
 0.15957445488790356,0.8683814123958785,0.32493534485755826,
 0.34647671481251274,0.22875113036133976,0.5838699933384239,
 0.4567781409595181,0.724586872456719,0.6021159797174703,0.4040103852940631,
 0.15594823534827995,0.8166195436757585,0.6140906713166532,
 0.33223581437477767,0.7893626262122021,0.8725718570367897,0.7179585661992134,
 0.8709079340223397,0.9214343565662722,0.09144618737257837,
 0.06375899808887897,0.8789985971945128,0.8445439507900467,0.5674088806366915,
 0.4481264147336146,0.8704692491684325,0.08091632425602535,
 0.014389497952279062,0.5215406929741834,0.978455192765505,0.7393680497288854,
 0.6101357711076647,0.26674853520291286,0.8812574609515011,0.8344643760401769,
 0.5269418892486026,0.7395948641065514,0.37401852472302577,0.6988383165520726,
 0.4244790149000115,0.3289602947202296,0.594885695597916,0.8278030965674029,
 0.820968844522975,0.8982117054145963,0.2448611476599697,0.40237595311504526,
 0.8728104147362722,0.33059047562671506,0.7128731265333904,
 0.49279039261570345,0.4220263276033658,0.7168457566986645,0.1957529917715929,
 0.12366956410803032,0.9280767278604717,0.718335556333645,0.7977034890522188,
 0.7531693884470303,0.8743937548988073,0.32364715554634294,0.8190772640990546,
 0.027321474889105635,0.976470522845951,0.8907592989742775,0.5557942951024207,
 0.8779243445703657,0.20618601295601258,0.6451554585399675,0.9066669673572008,
 0.18752122556770612,0.34711268309899634,0.7500282943676189,
 0.3167638662469142,0.9528960283087354,0.8572770627305607,0.5403418050597534,
 0.8164867973804365,0.29905900958206477,0.2636925753093985,0.7228741628986329,
 0.3243632578145047,0.09220785414851174,0.8224213551952709,
 0.05767296531424826,0.5697351172109734,0.09013352392007068,
 0.019142950256951008,0.6872461052910224,0.310773966201065,0.6895558967739686,
 0.9504797841347002,0.2251167684675941,0.3260759378976026,0.15585400918054204,
 0.5075786143571708,0.4806255272375798,0.175366854744054,0.6354930434582189,
 0.7414484576512884,0.4715475775390575,0.5260463605582286,0.8003516019991757,
 0.8837524846178617,0.9515358075011856,0.24920752998690998,0.9653099335669449,
 0.2868020121730055,0.8351395456651851,0.11063203080527217,
 0.45083605320053044,0.3830906374091565,0.39228738399099095,
 0.11752736304319378,0.5150593647517474,0.05588438989164268,
 0.5134095662544808,0.3495360816654933,0.3931740002719688,0.46587461018118403,
 0.23633450650198784,0.1778183045521683,0.17361246365079985,
 0.4956486226234522,0.1971168224769707,0.16536514625494803,0.5153503323817207,
 0.5390670487425576,0.54313431703481,0.44198831414628403,0.10379549446527825,
 0.4834811413771771,0.9400847095559917,0.7239527565919662,0.9270503777107324,
 3.4907378202797545e-4,0.9218066508196664,0.25890212378327904,
 0.1489586775796189,0.745732338091898,0.6327881064147162,0.793280702645384,
 0.5112753570192601,0.48537605959857166,0.940069586936746,0.6186559547067362,
 0.5234621420861846,0.8346602147076257,0.1300388303207718,0.10539965653468042,
 0.11851638129151632,0.3620567245204511,0.5849513813283933,0.4391647934891322,
 0.9886313327542678,0.5765045234531213,0.10157692519884409,
 0.041191389150258106,0.9169460285860414,0.9391841026336912,
 0.6151204855947073,0.500797710332981,0.011943052634941553,0.4139190305607271,
 0.6159828619501455,0.2013671182879353,0.06650465335104894,0.7925259313344745,
 0.9259852023554448,0.8694593566568005,0.9599383178278719,0.15667798029258195,
 0.08344002830702013,0.7211208275922165,0.5800978970577486,
 0.11969167071062958,0.19651668956787527,0.7055326371405243,
 0.014361691453522396,0.9048468409285287,0.21296076934146135,
 0.9827793461546175,0.07629136739642728,0.6893887001083513,
 0.24481379562328465,0.6829244787437881,0.4532840100005058,0.6162808836828867,
 0.8499582759952271,0.22894343218464186,0.8544669105874441,0.7391931553899679,
 0.7465727707997355,0.697577759929587,0.45059024142244,0.9484271149001012,
 0.9095686582089308,0.22977640868156612,0.3965504696146045,0.5666385328987681,
 0.26664407702562265,0.0010102414630280832,0.8693269690805396,
 0.8376046083602788,0.7801322188757114,0.22033091482143985,0.6001297139137833,
 0.9826340123176782,0.1400257937024929,0.6411711279047787,0.24316960948580213,
 0.6453999738828925,0.9293428525194846,0.2590795934289628,0.1450992285436412,
 0.3325193824539129,0.6075105164478275,0.721107803170667,0.2547498909935222,
 0.14563599177568187,0.6698956927369532,0.8219118360200901,0.8623472316686193,
 0.8355894855953294,0.6757673159692532,0.4871029336153688,0.9329478779232642,
 0.720998434644205,0.32661016023453693,0.3404574605408841,0.4509941760499756,
 0.4499842457290997,0.4848097993583511,0.048539309231807826,
 0.6323581452596492,0.4127837784137336,0.26301677361706666,0.5874875697664764,
 0.7019813465647854,0.9889307853592431,0.9423898208362921,0.5871556251788379,
 0.050312273915170014,0.4275613979274495,0.9074240003361549,
 0.06085136318173645,0.21051035495413528,0.3093970385846553,
 0.1324667766011315,0.021858647355388894,0.27880527565406743,
 0.051340502513166575,0.9841166358570086,0.6975888755209487,
 0.4365442581558183,0.689021191863075,0.5769035463548744,0.13453556988444726,
 0.45378736260271146,0.8470027041990297,0.08219647758380388,
 0.3390617946428882,0.8475553414444834,0.6212604134736706,0.45216848140606103,
 0.7600627695249297,0.7003457982269851,0.19038874529129968,0.8522977179593967,
 0.5104389687899871,0.8285419264921102,0.15518285411948252,0.9091839888555023,
 0.08050534107605745,0.09538409755432675,0.35980489429364404,
 0.9393636821129712,0.18908128814216385,2.9928819922275807e-4,
 0.7796826231598818,0.5593737424270151,0.9130728146820118,
 0.052306676914521955,0.6157605820643556,0.8891888092527681,
 0.31884593476554357,0.9983400105859498,0.024908925027116502,
 0.4924567783683649,0.15846434074654725,0.9786844898070188,0.865607699246039,
 0.9161628241908892,0.4968447268800158,0.30012232242132086,0.9453338637096501,
 0.36672020495822855,0.12764884107471863,0.2956107521210487,
 0.6484776567662548,0.7302460977585804,0.931176700114483,0.11996021682663305,
 0.8654061680235968,0.45678402492408454,0.32722881516289637,
 0.7871217545148564,0.8444892677795937,0.15235870884778557,0.2673622579842124,
 0.13234603428249003,0.5046905612870873,0.00858902633601355,
 0.7993762837666496,0.9093407415648985,0.9212159148630272,0.36491631566211036,
 0.8923010168022624,0.6513114221895102,0.5191690618852738,0.4947401683246747,
 0.5332586154359817,0.6808987469721373,0.08570113743517394,0.7083591173963921,
 0.09831802496988795,0.04616412165790351,0.6640850841791074,
 0.8459553662435038,0.680385376036741,0.4372298139922889,0.9388439827553471,
 0.2805728967911195,0.010787232682117343,0.8031950075698625,
 0.8910036990340767,0.5288529246995671,0.8198755786042646,0.7881335244349642,
 0.8473764636233319,0.05807734884554294,0.8037038480170501,0.642300560345596,
 0.2154356947475533,0.7152888028258652,0.581928950218888,0.9160719944199045,
 0.7163578883528443,0.2770434220011426,0.8185998103810096,
 0.011253362458888638,0.7147990847011992,0.4517321063081329,
 0.44324280003143934,0.946058851474268,0.2804375342442743,0.4213872348331108,
 0.3850741609840692,0.24591510447838139,0.8256331298924103,
 0.42387596577804165,0.215457485057589,0.3654272162498645,0.6198355609928093,
 0.8070090081297299,0.5015278595829058,0.5754510774122819,0.9132688524435991,
 0.5323446959343335,0.8515241683780054,0.34243755282369215,0.7608872519763308,
 0.03586833632478725,0.4535418176082818,0.5916605616250796,0.6562405256575603,
 0.25185541704950243,0.5213223644553924,0.47631071523801516,
 0.9372974830737784,0.6004888101620901,0.14921430250000045,
 0.07493979449111932,0.7914996817940859,0.8408529337485875,0.7725802939423838,
 0.2129495209463399,0.27399630409328213,0.9676771310872986,0.5328923698947423,
 0.811838719915416,0.6715150589071884,0.9009947066652222,0.8674988713866202,
 0.5692996622864605,0.34575748440644116,0.20985188012078726,
 0.20968337355943834,0.9260465077627528,0.7679007620180512,
 0.014753465484280581,0.8985145446500347,0.8434253988862084,
 0.34636665874882167,0.2126243302426426,0.862730900165593,0.4914621120494498,
 0.5955384210430719,0.020082231134309847,0.849121173130502,
 0.08709202135853267,0.030043959383230345,0.47691731674892957,
 0.837869868251015,0.7159407867988451,0.1988555807657466,0.312738333082176,
 0.5715070859839493,0.673782937553655,0.5103847853657991,0.45578110204001976,
 0.5027131630271767,0.6556407688370394,0.746605448639131,0.9180300629504852,
 0.4190547937742193,0.12875587603647287,0.9636763426732742,0.7717838871742563,
 0.07891442102667923,0.2773274548212572,0.6785190117092207,
 0.26355903037003825,0.8748600730894822,0.6663973937455485,
 0.22145170345058318,0.06250418579908235,0.915710540866655,0.8995949499514077,
 0.41836339833677094,0.09077296376833266,0.8785906398622984,
 0.022379384628504084,0.36971045408729464,0.119325224381154,
 0.7222537514116155,0.4632308869517612,0.46535020076536315,0.6353117280651794,
 0.5851495365872331,0.5900376268003166,0.7914018719496377,0.4524615537577986,
 0.47115004314493836,0.5424554541453137,0.8996499532022693,0.0825696879334602,
 0.7347305534852291,0.6013757148467249,0.6530715717641483,
 0.011707166404325231,0.2085604631806065,0.26879869128996936,
 0.5116385824947685,0.6798259386041051,0.15171391630604925,0.6082892188171912,
 0.5247353455842908,0.7004399786946534,0.2136542070853511,0.8858878222640625,
 0.18946222702919024,0.8527941126030878,0.6497351839552777,0.3437579715379022,
 0.8558653226529157,0.11425261187729574,0.3324889574746446,
 0.23334592678340838,0.9672410237445195,0.1960742533738371,
 0.04020361207687506,0.6100387504063556,0.17900593033370849,
 0.20562936953825184,0.6083537531924097,0.19090697813323732,
 0.8575227287060856,0.5744704015899481,0.43291333382038155,
 0.40184837243963767,0.015284797149589857,0.6362111166589266,
 0.1573199555234186,0.7746603285304017,0.7656461635854842,
 0.0026359570747148675,0.7536064630082959,0.3000399886162878,
 0.7450127245835756,0.6491298306746074,0.3974625390304434,0.9984911501332703,
 0.7618767126816248,0.03545187808196959,0.6792274661763752,
 0.38920122410000546,0.9211047718437128,0.9751605397959808,
 0.18099138696595451,0.9753639694458712,0.004368682057521167,
 0.4734792976995801,0.5925544252907234,0.8889210214564922,0.33423539026300975,
 0.5187766953921755,0.17691834966292685,0.49545935640286953,
 0.2761397896442169,0.9375800761161887,0.04776955962888019,0.578091208379527,
 0.5668260252236089,0.543730188294088,0.20960911837229323,0.17181867647973637,
 0.4110316574131,0.7566193795013813,0.2190096033623653,0.48105472053647313,
 0.3812644147223847,0.4071960157669591,0.7555492858446512,0.977456755601013,
 0.3650594107954723,0.7138750530480735,0.4828396168705118,0.8070083001598005,
 0.4833505210180653,0.36810246389256696,0.37705218213735137,
 0.039425117737670545,0.3151191143079285,0.5619388729483622,0.871296768190865,
 0.5897608532265538,0.4705726262162191,0.5262488595518995,0.5789279982043916,
 0.05728508876683991,0.607032859707444,0.34868770966250273,0.8160378459209269,
 0.41191511678569515,0.6857261516498803,0.7615495216558086,0.5102600052997005,
 0.08006871282220862,0.5679778069361577,0.9593118315063636,0.8337917705449657,
 0.5369149268399422,0.5943996289084426,0.920705178908979,0.25290072492887716,
 0.7096038762224592,0.20948227840092104,0.8550252987774627,0.6810609757823944,
 0.005674285048669403,0.28518098427124006,0.9838210486287233,
 0.023469398569102795,0.30957793244661735,0.3296156259844001,
 0.824536537765968,0.37477963326082064,0.5866514530497362,0.6488965224788664,
 0.5577957461607185,0.6237802262640749,0.188568987665363,0.09567996916316512,
 0.7551963707719935,0.35405844284301735,0.41942770943598207,
 0.6945959823593689,0.2800183114292857,0.5861580136611536,0.7370769645889974,
 0.7613011692487047,0.5735945443704858,0.36163610534438584,0.7843687616641893,
 0.12259679515321054,0.6025633931805883,0.7654346348749606,0.6751433793660019,
 0.2947933019491966,0.6651589305740722,0.2852282291843845,0.98700288851356,
 0.8839829843333946,0.5996530394802216,0.06841735012442207,0.9653301323288944,
 0.48825605796817595,0.997641076643811,0.18743875639187446,0.8847150525341362,
 0.33006681073730826,0.213147338184138,0.8464498729730537,0.9924219457462422,
 0.9016129094504357,0.7566818921513165,0.6545252900460624,0.4319258688960782,
 0.2166186885713972,0.32305926874511237,0.12566446103001894,
 0.2183706063416807,0.23804979473608856,0.11091984399850907,0.728613762767657,
 0.9593577159720317,0.153294636089221,0.6148252981077142,0.26076868343118376,
 0.9356715810146388,0.28415918043045396,0.7044372580693823,0.6016867203644566,
 0.5912191585056165,0.9879919408880511,0.5349205889300545,0.5534240743002738,
 0.6373062147445386,0.5632475482977229,0.04883311159305914,0.9097955748383594,
 0.944877177178584,0.48247969766377297,0.2776699087021769,0.9560389851102853,
 0.17438513860031934,0.0858908071366451,0.31742271361179886,
 0.4655376815893757,0.5339301608226664,0.8913610276569893,0.8139649953422748,
 0.6719406763237794,0.012045168071717183,0.7006109645204794,
 0.5071096127546457,0.9023083640785722,0.035538975293982866,
 0.7628555275564541,0.08077776249561741,0.45696828335684336,
 0.7022511372239644,0.6485438741060814,0.784450045497393,
 0.0064262194792560345,0.37135100844511926,0.3643728933744963,
 0.8369678793574736,0.6622424453083742,0.5326604695049173,0.7653193918559911,
 0.8996307901533848,0.20571134097939692,0.7966506108981909,0.6581207341188671,
 0.06410116980845759,0.6985214081920493,0.710511644908308,0.43833096173496566,
 0.016612556088415054,0.7349907286385567,0.1707164508658494,
 0.8235300686213982,0.6401697764592673,0.5111351913250575,0.12919350242278937,
 0.8870188288060373,0.5035914205505582,0.43904284014272554,
 0.09297472354157321,0.006078924595902868,0.2692083723930908,
 0.16964379273208585,0.9528007263983088,0.029748193980862925,
 0.8566933423403147,0.948036256264475,0.05551168836460918,0.5042530472296437,
 0.6293068124944331,0.8308284120047681,0.26500335171210154,0.4529806499140795,
 0.3414064603213903,0.9149802684053152,0.6394040784466413,0.9322316046578378,
 0.21014063472839295,0.8941238581282047,0.03365601677628749,
 0.10281583804723149,0.13938436166730162,0.6424018679824344,
 0.046511226949219275,0.324688687117373,0.6732215798357025,0.7148269015725681,
 0.5114628579176453,0.17642054892347026,0.5893560576449031,0.8457614026197698,
 0.5828932717704856,0.32567617679963945,0.1395189461022982,0.569929603352327,
 0.6018403091324568,0.2630682852937567,0.4637153044872008,0.1697229524164755,
 0.6664715711743138,0.31288388467448924,0.6232482137050659,0.8217314729701053,
 0.411400274709836,0.9011295855784166,0.7335925384176093,0.4702413073473787,
 0.296150098841803,0.19808620746085426,0.5861435991825442,0.9158803769134941,
 0.6650108392549855,0.4213585520101748,0.5836880647057094,0.4061407782054206,
 0.42442343281875683,0.9507178716012132,0.8704637606696276,0.6857703393517377,
 0.5915999225131523,0.9681242756677263,0.30688676058712494,0.819905569843953,
 0.887605390536565,0.44970240248406457,0.2907580470180742,0.3090499721375617,
 0.8483265555670285,0.02070901575572237,0.5482473709970864,0.5204982828483485,
 0.9822855173555846,0.07284891062742649,0.40994374250597476,
 0.12687350247776186,0.08660176783967954,0.6533235556120998,
 0.1610251845221038,0.6205333241106378,0.7723313860617177,0.29130891711929896,
 0.5517499317208138,0.6882295907311052,0.5165267126465867,0.7319856082125926,
 0.5171590699651717,0.04950338205583593,0.43477340007223775,
 0.3719697316734619,0.25590671007011423,0.20711843250322504,
 0.6395180358591961,0.3297578731039845,0.16037798331826947,0.8002302719115613,
 0.4507924758066779,0.10973892421445564,0.3865803743868519,
 0.33526632404877954,0.7637142369031501,0.013454896821345486,
 0.6489869417547629,0.053960339190305406,0.7681216336430368,
 0.9254544447741039,0.42890099891859534,0.13693186026436432,
 0.5657260839716651,0.6581675746072781,0.4004216557445692,0.7009376836843131,
 0.5823909805034575,0.46121965363433093,0.3549198758987552,0.6410849811302843,
 0.5866998725474806,0.5514889325521547,0.06849330399977305,
 0.22918476780268116,0.9360409722809018,0.7741972572157285,0.229655758953642,
 0.6757686270134561,0.9709036674811357,0.034846028296560805,
 0.25296685134294794,0.5834502349428923,0.7033543408890196,0.5486459199725783,
 0.43605578879657136,0.3220263063289107,0.7239598951277965,
 0.46223961375354006,0.7348328639200827,0.5706471901930454,0.8532149391648928,
 0.6379292671156517,0.6679978472354386,0.5132102012367112,0.0824356867924565,
 0.3366613707764734,0.8334230694837192,0.8125883262203054,0.40702107877024973,
 0.06772122399784253,0.6208890118410237,0.536144821531758,0.6488239345224598,
 0.3649792994222837,0.9808062956812852,0.40133884485353866,0.9627162985824077,
 0.948226152431074,0.15476182365039082,0.7142966255422655,0.45226969911428305,
 0.6886013360867282,0.990341980354989,0.37216854098088137,0.8230553857946815,
 0.47881680414060823,0.79638349192874,0.5493359468073535,0.6045859482355417,
 0.47397790845486976,0.23835916986976957,0.9444234091854062,
 0.5442090932803566,0.21399427630035006,0.6008906024933512,
 0.18522827652241092,0.2506202092319185,0.1745173092732515,0.588000960736067,
 0.24776166602728977,0.3486099459774932,0.9841023160957018,0.761611202298408,
 0.8441943981369123,0.6746660196020431,0.0736513475834566,0.21413742801825297,
 0.09314233083020207,0.370965597769751,0.11290092929520501,0.5152611933107222,
 0.7756789226128095,0.4980863742370333,0.594990094081659,0.8787098166680852,
 0.22131212526080457,0.8694076917303091,0.8669932987371407,0.7499720768808679,
 0.21247078322612956,0.4796360820239738,0.5984299998932048,
 0.040464575981956186,0.6695146080450257,0.6513483469483967,
 0.27632912892211736,0.2338349953152865,0.22520645706038334,
 0.7111084089516112,0.2753409041473365,0.7828908054549275,0.4246303113233332,
 0.013406202519027133,0.001799970004663587,0.24478837511340856,
 0.745826497172058,0.40934269261340417,0.8604062943062157,0.16057061305582498,
 0.25230358320615087,0.27248294882428137,0.7994735168183961,
 0.09802292528838219,0.3661254864424218,0.37325100316418713,
 0.7323940942194773,0.368632669556038,0.7417876937040199,0.9070513443584826,
 0.9784523075970871,0.9756964704430874,0.9404436556097935,0.8775338132563022,
 0.8880353516949594,0.8993696595966784,0.8589514795356994,0.752441195287168,
 0.549910544956534,0.026401537023982735,0.678458709655771,0.4149651921044284,
 0.987525053627966,0.6323407215157406,0.2896152559707039,0.01732464993548799,
 0.2443285580332818,0.3444160626089112,0.31954646366265327,
 0.41713356728074014,0.10495500233858623,0.47244804159865317,
 0.965878431860452,0.0689942970353915,0.3138298803001456,0.09354529900849984,
 0.5051136555727138,0.41244895035161033,0.2244469757274894,0.9968525467586933,
 0.25579186823309286,0.8776526550486051,0.5990632991451643,0.6303193604412992,
 0.6697965326998537,0.036727416885792064,0.41794837774214866,
 0.6114936572300765,0.7284963248318262,0.37239497015215184,0.8264465534737738,
 0.8192417843467595,0.7869451718027548,0.7140068649567926,0.10090658112682827,
 0.09385060297560321,0.12095956280608189,0.34554217917051844,
 0.10695282057755229,0.2717980665474956,0.22086678679399707,
 0.03886004772811724,0.5405861819586929,0.8583499735632348,0.3323130570352575,
 0.2216188030560884,0.6809660475289503,0.22192171760199397,0.8876976271052737,
 0.21647903343373087,0.42536992061996093,0.7828076843542308,
 0.9170575143704861,0.14137052867014732,0.31282860795328427,
 0.25490926355485855,0.6259059393245927,0.18579605017506495,
 0.8879463471040473,0.7325848815848716,0.35285518667628635,0.3449198930899364,
 0.6598848221371757,0.7239200397853811,0.4775482707504072,0.7621007301748149,
 0.4978755009904402,0.6445399332919413,0.28057686320558195,0.9394000995511399,
 0.23331641841040995,0.6724888632319022,0.14683895078465314,
 0.6523423372494679,0.4380652710573618,0.1891808252678454,0.6719095378170142,
 0.9889942650673392,0.18686412498697458,0.4765437564581776,
 0.09798848123768455,0.4608102613768045,0.04476063376906603,
 0.7450489630245531,0.7220858110693505,0.34102583681207144,0.7472071199085211,
 0.5763581709473945,0.120060966819307,0.9339949005088081,0.17849988361096825,
 0.8316327727087657,0.9948303402473414,0.24310394881084885,0.5161828741030801,
 0.6136880289215146,0.9517135608467534,0.0605667607003072,0.7863551897189373,
 0.4050259558547207,0.17277182044446582,0.326702243818082,0.47192960614038426,
 0.25897275134410447,0.1695785981622644,0.5224738302652397,
 0.11192799322199398,0.368388029500767,0.858229261439682,0.12575475369260625,
 0.6826556540224308,0.19973422733700386,0.1521235142413747,0.7690617266572068,
 0.8449727027159414,0.20132033400860694,0.08377948307175587,
 0.23871624073141584,0.9091707185164736,0.7049547666708698,0.2536701718227985,
 0.39592899601904685,0.9527204716458344,0.1171124934311425,0.8422990730409597,
 0.43148632945869525,0.8379968232249155,0.6555661368352582,0.7248426268746485,
 0.3154472304636994,0.9543427489198975,0.8932159613831047,0.6476975150028988,
 0.834073416438865,0.6757898198609035,0.7234611784920928,0.643281404125854,
 0.8029699739333452,0.034877197148066585,0.06770937982011305,
 0.11765887287151644,0.9796935037274257,0.6257008106745192,0.7310770983366568,
 0.48897504447197626,0.45199716874224005,0.4933324382597034,
 0.6460801565247326,0.6783447412832502,0.16075951585723658,0.3995001898221232,
 0.2783601310168874,0.6946044800259241,0.06251033887572699,0.7256384398428352,
 0.05168755463103336,0.2939302186208399,0.3713945380891561,0.806233668101011,
 0.022811828842023152,0.7179169207295114,0.5738562287075856,
 0.21606952960950065,0.001542295015975803,0.46585546448921356,
 0.40171267132702637,0.6509728727025971,0.24327544427074121,
 0.9597967059350625,0.1489462714874804,0.8371593071216265,0.41163619838136034,
 0.10123275050443836,0.0482091634952575,0.24191235618402862,
 0.40505425421336105,0.26706419278218696,0.6451869276341968,
 0.12367363041001833,0.8358229730276583,0.9514616531668666,0.7726022675625338,
 0.9849448802770748,0.6820285479812276,0.679327723129308,0.3505821575067376,
 0.07427012881577033,0.5208291634399699,0.8061353981722603,0.9756044278672437,
 0.22188658323663502,0.21159532213788212,0.8381321265931946,
 0.8463474599792031,0.17428712385597533,0.07702850705766584,
 0.34834333317802924,0.09794998365246199,0.07120406011493852,
 0.2695492292700544,0.6085430640176229,0.7910507422221501,0.1988492805797153,
 0.640511711908427,0.7127451073183662,0.8485658599337031,0.21470988957101933,
 0.04410219541204097,0.9200579389659375,0.929963530375436,0.199080978729961,
 0.7421732819775746,0.3930048198702163,0.6464374470265579,0.2488935047250299,
 0.07252966898693525,0.5388976799577848,0.3460481836593772,
 0.02219806758562859,0.5535918519927825,0.9133771196892679,0.7068563483207733,
 0.34203624477350636,0.9324500303218539,0.6691326750218544,
 0.037824146373671486,0.5262358271043782,0.8472974861048865,0.63806164434297,
 0.15504167268256785,0.5349511678004166,0.5647160529212228,0.5577136608074912,
 0.03197535569316412,0.45569961834305905,0.9771664287952895,
 0.24956010523852168,0.4054782976974739,0.4186473313648367,
 0.43027355059733674,0.4962423779895555,0.9173890949548037,0.246935156744717,
 0.19171664623645568,0.8413337152107893,0.7461549847135553,0.5371976688398947,
 0.09110881474810906,0.4047279935813429,0.4183408593329295,
 0.18958778614064953,0.36805405712505523,0.8908949242720363,
 0.20922203932579786,0.951359686643946,0.4819117414611811,0.30515944018063845,
 0.4334343724533628,0.5852254439595637,0.7439140319017833,0.6515412768042464,
 0.26127371650657927,0.612940321525735,0.8552935594148894,0.6576557702201303,
 0.8905447412594719,0.90523089146766,0.3588805261588127,0.14004564018970767,
 0.36528383495840455,0.4741498105698666,0.21914882309968797,
 0.5623135315659522,0.9066198796625471,0.08208827377294259,0.7957091579902628,
 0.5756079387057423,0.5717939868303094,0.821028181199579,0.9923389662885808,
 0.6308875980357173,0.8175684144838408,0.2249043555885626,0.7981150159736393,
 0.044407029697262956,0.04301378679952439,0.5997386161702235,
 0.5592055123898368,0.20940927604101867,0.7367721560634675,0.5127236885450613,
 0.6794830805193139,0.4077757448401992,0.9769889052764572,0.501745276314454,
 0.9430953861098952,0.20351969848173868,0.6130878373783486,
 0.07202524270468658,0.8839665798206742,0.23374991474029794,
 0.9402617769847103,0.636570034506479,0.41567932855884426,0.8793901458483877,
 0.9285304042709476,0.2863925830726052,0.035926774114339755,
 0.9611803345165517,0.034471830852691565,0.0720475436079191,
 0.9379213942314435,0.5556853385570084,0.4558071253424576,0.31056609699599214,
 0.8180259250074104,0.051925974991464674,0.3737457247890591,
 0.9931174694591967,0.5533026525684409,0.7056216008920884,0.34057713497896935,
 0.08815796454455815,0.7622816591769684,0.29700049635138914,
 0.6020840758983267,0.9111893745171279,0.8083009476661472,0.7963652445774979,
 0.030526993079952436,0.2059699716922595,0.029950724761155123,
 0.45669489610089353,0.4045800709602042,0.8605819167848883,0.990804551130577,
 0.42606855476654315,0.9067946148925383,0.27821375052815944,
 0.26238522468716186,0.14783152733069183,0.7254047911256166,
 0.04340147216874568,0.1355644227603473,0.8047104363487347,
 0.17198756902963408,0.7196023962592665,0.9394431546943611,0.4334473471202397,
 0.21123447918084848,0.9568163737086304,0.1989377060234453,
 0.08404186921041235,0.3449959438090924,0.14863566753986046,
 0.4533320568338426,0.4481882690843062,0.58379598314335,0.701375053352429,
 0.66265794293145,0.7593093714944912,0.5750497090652034,0.055473254371624314,
 0.5553864573380352,0.1965941721379384,0.2389991354731147,0.6539675799490381,
 0.1513930711240722,0.8744266159642895,0.22335496871267946,
 0.23560497462431895,0.6300557402189191,0.6384855294197851,
 0.24164288757287644,0.9054827682257055,0.2608956853410406,0.838593164647492,
 0.9542836772109707,0.9807832069554087,0.7840129132441712,0.7759958677630937,
 0.7998097087138539,0.5798277059828003,0.054548237117012954,
 0.36227237240075627,0.848959849992193,0.06538116367739355,0.7206711113528793,
 0.8645844138237182,0.6137427283048522,0.022092991122048877,
 0.08352063723015868,0.8964047033946967,0.906703376118235,0.6926495588751209,
 0.8125879701964478,0.02503778192160877,0.4270204387256814,0.3546015864404507,
 0.5469627678331127,0.8602549929083124,0.6964392493705308,0.3548672206111184,
 0.5763687297038207,0.07056443033955584,0.6315038404334669,
 0.48760472891243123,0.18948011153094624,0.1546735097041343,
 0.9249927706483511,0.13382222347019346,0.733549722258158,0.25738839868891783,
 0.6973292722919169,0.6827135700496763,0.9204202702984091,0.7859446570847588,
 0.7299639496481848,0.003495962155392718,0.7831974016653969,
 0.5078629493026329,0.6124264466807187,0.24963505892202043,0.941847030213477,
 0.10797800856910922,0.7276244321683134,0.7823804935585739,0.6331303974818064,
 0.8971443872869503,0.36367413768181145,0.833190428138219,0.30560335631064184,
 0.9047528120121118,0.057181460037360354,0.4530125720067474,
 0.4805332362628082,0.41557192058831294,0.07954114047330996,
 0.5397775738157786,0.8098105198413403,0.2439985097574664,0.2448947904478691,
 0.5557593368604389,0.7193799386909393,0.8078462666570335,0.26104243958209683,
 0.3991322164456004,0.2968764851531078,0.47645086777066425,
 0.39712227439899805,0.3680133342151606,0.2566348462864707,0.2463209852597873,
 0.023174574286204064,0.6471647527387492,0.179149837334033,0.8835077929217204,
 0.5345151399383381,0.838158284309164,0.6851432098269535,0.7415839777825805,
 0.32706276166088766,0.816225368909031,0.3985098132034238,0.43729512350313016,
 0.5888250871359058,0.8387542931341709,0.4703075047219206,0.24321643586054398,
 0.8361729103072469,0.8201811380934481,0.8270062316970341,0.01841002465038677,
 0.2911772935563496,0.09779821690929613,0.9093787241285667,
 0.38176037419064046,0.9441543265712159,0.7021403266414984,0.9371599377401352,
 0.5695353079952786,0.7458877714443017,0.7805690101406968,0.44568883309514695,
 0.26468839222053575,0.5142782050306953,0.7827312564895973,0.8759555338957081,
 0.6464100824989255,0.3135640727269222,0.798770788585829,0.3394737402942142,
 0.9681278326204751,0.5842142191670574,0.6102876420136778,0.7924518598262871,
 0.9220852387702277,0.7112380280185816,0.6278228510080848,0.2697146723676842,
 0.15255804109400228,0.47788518105936184,0.022041978971680365,
 0.4337652982871891,0.7117370041055973,0.46257526142347205,
 0.014480445273518994,0.6869650000136497,0.9329688266594282,
 0.8311073655847114,0.9741893024200909,0.3751606027161679,0.668311162882829,
 0.3083599944069001,0.0614619178754523,0.29065269278401784,0.8271698319510543,
 0.013224593673484941,0.22163432252337234,0.5526382100363448,
 0.6369338776480308,0.8325537094314369,0.5565492785730674,0.09122954525902305,
 0.6945052238844622,0.10053116689272579,0.6505781654923679,0.9794340139026995,
 0.003133455335933677,0.1946460336092355,0.5169625809559514,
 0.3918389202176855,0.763398410918914,0.07895596835831986,0.5596993166097995,
 0.8770255056249804,0.7042950248345834,0.018589344708598965,
 0.24032507157352945,0.21353107347190992,0.8487866078147394,
 0.8384647799866555,0.6545356567896589,0.675096920655735,0.8143674003949813,
 0.9032776971358872,0.02895084468314102,0.7714255830361992,0.9429753888808454,
 0.28497156466840634,0.9968588363772355,0.03252071445548543,
 0.14446027038738873,0.20146760363836325,0.2575308963730297,
 0.4679050107770295,0.020698070019716575,0.18694387681107738,
 0.30765959702205525,0.805433148944751,0.8246922990834771,0.9555634908277463,
 0.13658865926682884,0.15914999125048868,0.35011139175309547,
 0.609990134865895,0.07819741861348772,0.504272281167178,0.06374690522201076,
 0.775505212360464,0.007399527274639661,0.3446925847469199,0.5460411534192892,
 0.792858050936481,0.37565067849520994,0.12379904684944698,0.2752989356338176,
 0.4776260315478251,0.27093434940467087,0.6345079574095487,0.537597722727093,
 0.7828830749977753,0.6083248126366988,0.1345027116165346,0.28249506381759093,
 0.46132782460137567,0.06454748943940236,0.021660154910860263,
 0.47941853840465765,0.36460853413159666,0.7828645604704985,
 0.5221107679385765,0.05584294261521827,0.6751853316394056,0.8753844962101874,
 0.40953459886251087,0.3554321737519124,0.39995387277205574,
 0.16101057295776655,0.16841901242542645,0.2600685023534821,
 0.6307902756527148,0.7200717825495382,0.5584717397559334,0.9477609090856348,
 0.49459172338404833,0.17227811425458472,0.826618503547427,0.5943727276030988,
 0.7879329022881353,0.9234165469481823,0.2228731582330461,0.5555866803006471,
 0.7823415032273132,0.7188029344853862,0.8046050160162116,0.6831925516257112,
 0.8573539660575916,0.8006139366750177,0.4907155154777062,0.04178850481850638,
 0.1039392661073818,0.5085089275737983,0.5489222829849038,0.2896879843329663,
 0.5984774522483842,0.8504132604898212,0.7799671756297072,0.27442712088295473,
 0.5885786864535949,0.9848538334573619,0.19950227319842617,
 0.45466979051697276,0.25838502426059307,0.6848111427189185,
 0.5880980716919323,0.16719578434375426,0.3330806544608794,0.8378727864728962,
 0.13423936014592008,0.09916621651093471,0.1662693521198545,
 0.7312076414146338,0.48536192111536114,0.6197128078278874,0.986726264268917,
 0.7785477721836693,0.22012927290779716,0.007699228368888811,
 0.9771164067928736,0.17211192233636974,0.2850801840971512,0.6309660309266845,
 0.17247819677339526,0.7280054309371152,0.23637531441647686,
 0.7389995073258702,0.4409684060639164,0.6503828238393946,0.5247538729391995,
 0.06256136851019589,0.9624331119084508,0.46592787053738516,0.977224106042762,
 0.947761793567921,0.8218887017127428,0.47964099022637263,0.5478051800782642,
 0.6272792963692146,0.22859135228881844,0.8539472417427751,
 0.013478503295125788,0.4047027430152941,0.6559004999041781,
 0.9681623581250185,0.6099041559721374,0.08969359364577134,
 0.20803756031415732,0.47184496760684724,0.9562987802207934,
 0.5346343822988677,0.8561955115458288,0.6041292126910173,0.7052878398010218,
 0.36440288866270465,0.036225130620075596,0.1192226378240473,
 0.6439814283389397,0.5157177258614506,0.6281454160366203,0.6538791648353769,
 0.06426067180253203,0.1956566031701772,0.4198819735217656,0.9840672636748379,
 0.491080639109298,0.3776752073909331,0.37802082314058816,0.6320870634325986,
 0.717322853510093,0.609311337904984,0.014327145283677067,0.27690238308923987,
 0.06905516200862949,0.9435830944498474,0.4606001074371525,0.9463727023325715,
 0.030247192799511935,0.7225746317904109,0.25370518212683624,
 0.4715522482896941,0.7130488431411556,0.45322558846021366,0.6473203089075916,
 0.6296466339023142,0.13495151063924016,0.7126934189583572,0.91339508647746,
 0.2531401927834731,0.6605623170567038,0.6672726265320732,0.5069916231032405,
 0.42412538569715597,0.20768112242826942,0.894526061809523,0.9597145278124893,
 0.6188676356377321,0.30476518954262033,0.11591975876541694,
 0.1349258164251148,0.745710910734759,0.17118483518766558,0.990950802632362,
 0.328898352870457,0.43774520285643637,0.04007579792955607,0.4564041463192776,
 0.6856048374839626,0.796205156743403,0.9383045853391485,0.6665096096593704,
 0.6952215970832656,0.9524825578123215,0.6595460549173306,0.9973228798244588,
 0.19001253232254978,0.32154123315370065,0.8623491067635957,
 0.15557433374830765,0.807081689122707,0.9127719507804025,0.8097301171422888,
 0.3442039203082542,0.2679022325194863,0.6817304899605479,0.35090198681679374,
 0.5336745356114773,0.3310383811560732,0.7054563475332442,0.4874136650879335,
 0.4069241785648259,0.35858985499778884,0.8673727002918119,
 0.10353443715780208,0.7774624485878671,0.5366832822421941,0.1642263128687934,
 0.2727631888358678,0.37717407997768215,0.7650235937090412,0.8809204842753716,
 0.0031765771588100344,0.6729530023120921,0.9647108779313139,
 0.03490066956461613,0.21090890572823506,0.4248309798531711,
 0.7053613580799893,0.11704074142761989,0.5177612452639996,0.614180537389786,
 0.372846616767735,0.39879184199765716,0.5975031400452959,0.2735565489331314,
 0.9312700753841227,0.29277931481456765,0.5382936995309091,
 0.08869361356821936,0.10084035890837695,0.31889202485401746,
 0.8230012135005276,0.554363836130223,0.6901809881291988,0.15824020180709553,
 0.04941713441933182,0.8977808597822344,0.9280019605301022,0.5959124846282396,
 0.2589283093226147,0.6641714372257316,0.08630394376500372,
 0.18703394496343184,0.9098890005749339,0.5593624190448198,
 0.40787960824472413,0.46135101623974084,0.679024633143485,0.905315536294389,
 0.30773248079563054,0.1039033264091449,0.8919140774916896,
 0.11558149925572736,0.6036029775517393,0.5087718717745557,0.5646094010613207,
 0.13564698710227296,0.9014977112053016,0.8102757834952498,0.2513950518853475,
 0.22186258327432917,0.11921135964139717,0.6188979579172814,
 0.32600949116901934,0.13753948752345102,0.8417834595425178,
 0.1686565770531454,0.2198939883783686,0.16434257894016446,0.8154677005874336,
 0.34250302555874024,0.8046052631728955,0.7378831720329737,0.8724187251118629,
 0.12704695569385738,0.5539273856713575,0.9915400689560785,0.7322386027824207,
 0.8771959285578884,0.4368141021708586,0.7699821600245532,0.5209262295721726,
 0.32115013776561563,0.2365561748994085,0.9158542088969042,0.9038761258763165,
 0.20035770710772427,0.26977966975524836,0.9037013501733239,
 0.07759957113350968,0.7443342101708814,0.7763811138302684,0.9589014884402096,
 0.08610478242940511,0.2702717619850432,0.7526790801699607,0.2805614669730452,
 0.0318939727903923,0.42299571990171636,0.2492412007014957,0.9979670721993801,
 0.8376504330987395,0.0956672166608592,0.08578237214145101,0.6411214650909572,
 0.8542717680549248,0.16023257472806385,0.8669054698642134,0.2522704535338368,
 0.7829495082493942,0.13069693752578837,0.08431251725768385,
 0.06216917290453294,0.21299009351614961,0.5441748087270012,0.851003047910594,
 0.3230710131223512,0.4883282250350529,0.5210352348508877,
 0.003211204231012732,0.058364516369980635,0.2738515370994209,
 0.2808840853502117,0.9799969460778771,0.28314482807357155,0.9174073604301543,
 0.9779282811119991,0.5324109430459971,0.6047593785999615,0.5652778827172389,
 0.769478214848248,0.9367508989965749,0.6834542853931265,0.9072103731163537,
 0.0628526554116754,0.9741222903950073,0.6684646620156403,0.06951053716413291,
 0.6923540810980905,0.6938662816044482,0.8930763178018489,0.1326059607232113,
 0.5221682375907277,0.05226376321973525,0.6660139864330139,0.6309357658371264,
 0.301042547947159,0.9929120535103027,0.6708619441806666,0.7862817053900879,
 0.38408214078991776,0.9069065772835867,0.9542949541764936,0.1995695195650652,
 0.7713498688654051,0.25921426924786695,0.06873188982422007,
 0.7785657680678121,0.9354075783795972,0.830662001230527,0.8144172038001877,
 0.5248348514304594,0.7086266087545198,0.9294055828075928,0.28647955354862153,
 0.3740416754333732,0.41554720644231846,0.18111449439764793,
 0.25703688932048074,0.21856422312470059,0.1831326742703977,
 0.07511165872558379,0.42972243558439904,0.28547478030659135,
 0.27380435588468555,0.45591501669449186,0.7611411823986204,
 0.9531332834292977,0.5258977323772859,0.21314110863399105,
 0.015182263345080771,0.5194538488488148,0.2754488405064157,
 0.9444325882724317,0.05677580647148506,0.5986692580359367,0.6634784513718227,
 0.9780921801620792,0.8924638627836351,0.5836714956079165,0.2947679241574257,
 0.9927587778236429,0.8947864561977384,0.6185749673481432,0.7791764788340034,
 0.8308432114535154,0.8398651129485885,0.799988212147848,0.49564493561853074,
 0.8056727978383129,0.5137055888480277,0.5774301502811604,0.6306998979541828,
 0.6842200906627744,0.3383633372214534,0.8176464926342741,0.6670839138259947,
 0.44632020673293904,0.6584310498340485,0.4379800455638786,0.7280017887106125,
 0.6677093524946219,0.10780226067215903,0.40459218310699463,
 0.3993956508392812,0.38536666434130673,0.39746277024994914,
 0.8085342255691202,0.08206895594859653,0.41743059402000293,
 0.7899659118535783,0.45994128540316936,0.30363740877343415,
 0.1648140595853389,0.8666235573061576,0.00544849873945219,0.2861232188273377,
 0.25720463193641663,0.4367866332022525,0.8448284983383794,0.3370697463985872,
 0.9046871431911406,0.628814088313441,0.37984210498127624,0.7038147847230863,
 0.6340974954941707,0.3605557832153514,0.36586015501757674,0.8602472736706296,
 0.8293922031696382,0.5780405169668237,0.6094059323428092,
 0.032094112105658934,0.32017126289776543,0.8510290140143322,
 0.4856955773034949,0.8866977764922551,0.6353633453992148,0.01297878366753702,
 0.2169770026447948,0.029869018410430104,0.18141342056123289,
 0.4505044340503961,0.9865700647936786,0.7809273508327517,0.7501209408653191,
 0.6523770162034814,0.1408120230013432,0.5322191087089845,0.5247982452854337,
 0.5079933162943189,0.9287854871946222,0.12046617946462468,
 0.15797971485333073,0.38736652307257735,0.4022331462129144,0.984864049716293,
 0.7918018740306014,0.09422318558116838,0.36546267839320157,
 0.47785499601146764,0.9637866185308004,0.13376753646572082,
 0.9794488829403533,0.49163254475216456,0.3354403268131756,
 0.30665690445837523,0.8196361624257352,0.4599943194801128,
 0.44759396327856316,0.1634440375558669,0.8489580135103872,0.9809255099544582,
 0.028584536786920967,0.7658007531590127,0.8500650598315569,
 0.7932205941138792,0.8213445084169773,0.8299259847533289,0.25931905229573216,
 0.6979885167501273,0.6439525222842435,0.3514194234364709,0.7364523386552672,
 0.840918410966652,0.2548652196708119,0.37951107414445073,0.9570823468579481,
 0.4497930570592289,0.2182886977345616,0.6559549430265977,0.3633805629134933,
 0.41103309792629894,0.7331327752164607,0.08875513300060178,
 0.26878905942287945,0.894249861264302,0.11885159015067115,
 0.014532593904868385,0.2648795435765865,0.24949395408339337,
 0.8322310384571165,0.3888875817686378,0.6684860920059978,
 2.7588063304162525e-4,0.13872191556291213,0.3850421148183313,
 0.4916649208428472,0.6438229591226754,0.03526157801400276,0.7669633599178578,
 0.04533321875286633,0.25957731884765534,0.6070108117640574,
 0.7550143763010693,0.15377201208203028,0.7835871228873913,0.8207364417019778,
 0.9487552089303012,0.6888269290962776,0.6582008549014744,0.20528484163210736,
 0.6436672192613568,0.6547930731523747,0.3452643401383899,0.890553886132732,
 0.6049764857303068,0.9407042198635991,0.1298866357534162,0.5783503281446529,
 0.9574335804555912,0.24198466889412196,0.0520733510183522,0.6821829399346069,
 0.4435773329176096,0.5934512976162252,0.02178399350969451,0.7029881695564474,
 0.1780631864234733,0.9554394626980823,0.12398791532455633,0.3662067822326166,
 0.6527843824858574,0.5605825218493264,0.8222463031836578,0.37011459151464676,
 0.06754949344341366,0.1417868774433434,0.3420240858875935,
 0.38489595511309527,0.09919607453793444,0.27919794317994406,
 0.5615154545648079,0.6334314966697893,0.7721195469194538,0.37787589950617473,
 0.8449731639875633,0.862749290632384,0.2197528387408102,0.7047378839677909,
 0.329238145475374,0.6691774600195864,0.838195393721055,0.34922282689112316,
 0.36819585772358265,0.4967900309741363,0.6176654764095305,0.9813236254237532,
 0.10367793968179306,0.21605413359076264,0.6031830506630602,0.898988855127984,
 0.8908569683793004,0.7232911484769562,0.4748285736961264,0.4545463406277116,
 0.6151890242069294,0.1415705770234077,0.527412566092666,0.19636242009943272,
 0.6083194551005794,0.40645624101324274,0.4508897895850611,
 0.24708656257060635,0.5569035774753983,0.2976104587747307,0.6983585665322734,
 0.5765301104179025,0.5257021787311095,0.4836213045169778,0.11576844427588961,
 0.3519134768086325,0.9054768293410428,0.7364890337418902,0.6942834425034897,
 0.9613902032722562,0.2525646074389496,0.1813031848412212,0.2333026341378912,
 0.8134857306809715,0.29965337702191874,0.6418534323107995,0.9290188854796771,
 0.6214113257558478,0.45117317978682847,0.008824372269628622,
 0.2781621493013242,0.4462286441254679,0.9156704899774105,0.6907097097803998,
 0.5676699622519277,0.7599998225148539,0.1480760987548938,0.6093315100892945,
 0.690691192240734,0.4788863076736627,0.20267809030646988,0.9205697065966325,
 0.5981522733871543,0.43981546139425565,0.7286042172511271,
 0.32653110998815393,0.3894269649654821,0.9683960539143628,
 0.26849504019686643,0.8860282550544452,0.10481459980604535,
 0.06567761357405588,0.6807478316611894,0.857295306328348,0.2685082152156011,
 0.907444700033629,0.9000885758980461,0.43903054742551106,0.7290098817798939,
 0.6862647714573653,0.8947928643270703,0.14267911987938764,0.2653138423954099,
 0.3590036456931168,0.5909566663464013,0.39514160926912933,0.7137450737948067,
 0.44981005876666325,0.366270972706075,0.7358329610122288,0.971335925756077,
 0.7217097222298973,0.2576915204866995,0.4917696879609189,0.48349349831303423,
 0.2414375388303509,0.5570454014201593,0.8269577265879609,0.8753163344313257,
 0.8252324847272853,0.4873144820500551,0.940794780926506,0.5483763218477254,
 0.030864823633164895,0.09924868008013266,0.6796139470078247,
 0.4120911839351903,0.06582470571721821,0.9771652323855863,0.7608235959639826,
 0.6807179423959101,0.9210168714157697,0.7189837778677728,0.5323748568824153,
 0.03669985559898026,0.37192716816082305,0.1152100249349699,
 0.4201109705499245,0.9022130019189526,0.6375362425208017,0.41835207462727453,
 0.731273793966374,0.4795678332853305,0.0031760142441661365,
 0.3080003716617742,0.21143831278784087,0.21783956119332681,0.27018256577612,
 0.5678942069682769,0.15303884264553047,0.6888410116783991,0.982649840396638,
 0.6724456339522914,0.7605227539029025,0.48407059210961834,0.5659019096162965,
 0.0640451382611692,0.08118531117685779,0.5105751289800101,0.4403025305459798,
 0.9288413084238294,0.001094552869836729,0.5591661478057963,
 0.5519616086332257,0.4864319411149626,0.15883677812966146,
 0.04208886233810172,0.8667536753816314,0.7240129028241877,0.9829291139230234,
 0.10369961003344375,0.06884201524268574,0.14725172431493794,
 0.05339526488825774,0.5907760317424258,0.4556080152330574,
 0.10269817145053084,0.1923362446697655,0.6441158819127345,0.3028334176504285,
 0.31556075878606094,0.6240122544191469,0.47105837865723177,0.930211029055095,
 0.6100820400870501,0.17890017339248965,0.4610297712475263,
 0.20531828907540361,0.18352249988606517,0.4687770465036998,0.327623429276092,
 0.5139302007597251,0.06427417293311288,0.8651877635395083,0.5514820363754891,
 0.920323025622651,0.9130894316880224,0.24013645584167698,0.5951707341249438,
 0.9779782768174977,0.16980203721354759,0.9225102845024714,
 0.15908747037394444,0.11584871163895172,0.8423995193800338,
 0.3586265777884978,0.4578497194101292,0.9475807712580163,0.613302711191695,
 0.10139414805235136,0.9382907749052526,0.8937924013532139,0.7454657416072727,
 0.0378481422768282,0.07769013868679586,0.6977889310692258,0.4808472116306214,
 0.07982047891471322,0.6622436107242364,0.21989989714613878,
 0.26065511613789394,0.838063814237246,0.5885237167426062,0.724658523552587,
 0.4421525007210587,0.8969911887562578,0.16225327951010948,
 0.31607991275583613,0.938327943075906,0.3550264244955872,0.070590961361089,
 0.9071324205009156,0.6720286980024178,0.7952774167721304,0.24775781924690032,
 0.26144731727704495,0.5344344348944277,0.619360186872357,0.8537427095204456,
 0.6738759011640139,0.7448167890715718,0.33671903032832995,
 0.25313555327809345,0.06069033998222362,0.8630981114165608,
 0.15499548091239324,0.5917832234717569,0.317962329431687,0.7473977039611231,
 0.23053682163081857,0.7743225360282504,0.4694676467507167,
 0.10827067163007253,0.20577589357283455,0.5330190996147115,
 0.4907296237395362,0.3593877909432873,0.5742479239109435,0.9957883084719156,
 0.5136109351040922,0.8323641720600305,0.827923494180202,0.21359840645158645,
 0.7926152324285046,0.4283528937160379,0.4738123014783858,0.36266873465873894,
 0.13263311667155941,0.5007482744199968,0.0640541910344884,
 0.10813364422425376,0.24011316661267823,0.2539327536333459,
 0.29071623192032037,0.4149978689409357,0.4747645203947629,0.5131937843358245,
 0.8337930863525967,0.0206485234164,0.9649632726871171,0.4774611675476428,
 0.93289648004972,0.2733688425266525,0.2593422759226206,0.7412551239639352,
 0.8239077391940751,0.12188755709612242,0.5053335920950525,0.9999259378022605,
 0.20468475039940925,0.19887740169396417,0.594661253550099,
 0.12588662666647388,0.39020667581030477,0.8338259528508887,
 0.23681088513785764,0.6965944921561696,0.07577026188623326,
 0.8039932556805025,0.3813668817383884,0.3427009720231734,0.39866115422034976,
 0.4328808937216875,0.27865053980556154,0.9647654491764173,0.8978289661226784,
 0.8060325620001624,0.9332188548281909,0.15934100698123554,
 0.20902815702209177,0.8819718003426379,0.3440495360663962,0.5633018611265093,
 0.5511861563639027,0.48407823416346063,0.7725601601998003,
 0.16320765010300087,0.9980713496094198,0.2731739199835498,0.8896609262178521,
 0.2568625744456142,0.01584409088345995,0.3388758446916189,0.6683747616163678,
 0.5794790328215281,0.9387972368706441,0.30652165823651323,
 0.18966931169899992,0.8437741986621085,0.08927215873451888,
 0.4870631964234893,0.6040306124346727,0.7628413467935781,0.21057232479901766,
 0.6733115568871758,0.46549168520476125,0.3257203809853474,0.887329498560434,
 0.4034366088442556,0.5357598441415278,0.8012534483002993,0.5450400017379513,
 0.2465723012587353,0.15973829967475506,0.5589411093931506,
 0.18759323776153225,0.6272776937759669,0.4957169786059885,
 0.14802330731322055,0.43094426427652766,0.19546850218578182,
 0.4348972606832654,0.9554542088290249,0.5868682213936256,0.2823316283465266,
 0.3824213639616467,0.32077927335712575,0.31073061580192496,
 0.05373199770413417,0.2677755083801657,0.6825951778139885,0.5059919048368327,
 0.2523695821655324,0.3996184034103788,0.15498725903871802,0.373225249991338,
 0.5887558677757229,0.8377833682345804,0.5091719110593689,0.20202207096110913,
 0.31200318049123366,0.31423885297871856,0.1540583180483508,
 0.018501138468185685,0.23092742757285534,0.2680460423536287,
 0.5459145808227664,0.25259775502507087,0.3327751549573774,
 0.14574591538742465,0.8796634944780217,0.4226286443023418,
 0.48675599818758297,0.6791998751455617,0.07102718855525525,
 0.5769422524174231,0.9806939346422524,0.9050461943936727,0.13628518976594906,
 0.9618155480584629,0.4143448858655956,0.7110838839564849,0.4874672163035114,
 0.4862415933460248,0.10042676653552562,0.42739775916018385,
 0.5475270374155812,0.44126030521880555,0.3173026032881503,
 0.07768530341612423,0.7631385946275684,0.7231840711157799,0.7763665269939519,
 0.4615361926265825,0.8595287738373565,0.8955336691381878,0.36243568632585355,
 0.048476249332421006,0.10456301778856802,0.3735714350137104,
 0.1704369186827257,0.45222675858464967,0.1154114980046066,0.7526015870006035,
 0.987307080708558,0.7101816219327803,0.500339732340281,0.594798982722077,
 0.8241022737635274,0.7558458128981532,0.5501714699681468,0.24427585759787185,
 0.6433927836717201,0.944727869439522,0.12012090446359114,0.5804512929800756,
 0.4514652201716972,0.2916016852465071,0.7674469854461696,0.5176158319688164,
 0.7041151396495935,0.20771878202165117,0.3216404715708535,
 0.34137517988656507,0.30386622806351227,0.10434647926741714,
 0.34693423439787086,0.8093712710562118,0.39639304448847845,
 0.9960622367509729,0.9032241809849233,0.36163947603372415,0.6137277860259993,
 0.21806553280684104,0.5242990320347042,0.29142222537081486,
 0.06495936756796394,0.385117406629211,0.03142426952777799,0.5517460771023903,
 0.4790287170075166,0.3413455072774534,0.9526835379959608,0.14793358103943155,
 0.4771829865938235,0.8343636326353011,0.8232960967139422,0.45277651521171935,
 0.4150037321483545,0.7101633727110461,0.02258411178323938,0.8492600250279244,
 0.8946017229702089,0.46795670188460503,0.6575545682125193,
 0.18054184531183237,0.375398038537052,0.9989068005839545,0.04710303797941684,
 0.276124404112572,0.9019290364962248,0.41675413238991055,0.6334079043858941,
 0.8903907180247944,0.22026441785507855,0.34012368807827453,
 0.24209358585603857,0.8490270528946128,0.6485830567666259,0.7002346644243792,
 0.24986449665714494,0.8364167906324819,0.8034684848289665,0.6253973664939579,
 0.22049767786932495,0.25564906547069444,0.15456584737057122,
 0.692642274329385,0.3385272912717975,0.9671450472846093,0.6689085799218876,
 0.9268626707308394,0.26081108474607084,0.6706108601912033,
 0.03281900685484973,0.9815030197367122,0.6478207984153435,0.9848831749422796,
 0.30024028207159237,0.4939207170941673,0.3260817987248512,0.4166094084698405,
 0.3882126701372445,0.012481408357182255,0.10991608168758404,
 0.9304126747794439,0.37779771244738714,0.8968065133351766,0.5748989502188249,
 0.03058687579152397,0.950826933130494,0.4576266673704733,0.2792671506588419,
 0.02170694897530534,0.7934011154889933,0.25286740263435437,
 0.5302237495729187,0.30397711678675166,0.5385782907891891,
 0.16842603566984926,0.1824433970949817,0.18086570642556887,0.760423916730121,
 0.28803561814511014,0.31844065020527523,0.041036687705339525,
 0.44422909808268607,0.49414541175191473,0.6519471446028059,
 0.09846068225528071,0.8241433151135769,0.5319573238310573,0.8698208630969976,
 0.9947019153328955,0.3715077667084824,0.43658187311167185,
 0.11248302148013667,0.8970933659745992,0.3170248122268333,0.3609269431540678,
 0.7149350478913463,0.0034983404668672557,0.9763182277729183,
 0.9990791030887556,0.64772592911361,0.2124030250309895,0.5033789640082909,
 0.8541909895690306,0.5297987560515841,0.049311220928831,0.0652282522787423,
 0.8303487136133233,0.06521750036105489,0.554611617429221,0.663458472975169,
 0.8421721228372827,0.7392977428189544,0.6430113538827021,0.7128569731070734,
 0.8212228448143435,0.8814892593113001,0.7376841816785004,0.1653529605244929,
 0.12524704482687443,0.2878405530283418,0.9921367115775488,0.9321348270231679,
 0.585174801759753,0.6746553575410469,0.7046764569762738,0.9502087185190616,
 0.31142261273469773,0.15646418414465046,0.9819586111848364,
 0.09795439279494889,0.21746024727186652,0.5024340110740916,
 0.8559840678054208,0.23877072690284407,0.566891670837457,0.3123243720745076,
 0.8582617516916474,0.3227815510117218,0.7978518855536955,0.33478988458957304,
 0.403329919626063,0.0912065952336596,0.06136157682961252,0.7427349660708212,
 0.4560944529915818,0.602152132532322,0.1321457661836165,0.17457727939822887,
 0.17826337294263217,0.21633189311344914,0.6198841058235954,
 0.2859598511467166,0.5178605609098501,0.6388878119828242,0.8345737018712732,
 0.5300960870854039,0.10613825953205802,0.5868409844914104,0.6934744023883761,
 0.8608710738064622,0.9656223946519846,0.5614683099831633,0.9939606190374046,
 0.4437871606211621,0.3226691865237781,0.8317010130258148,0.9915384490618124,
 0.3105575047984359,0.23748496645921457,0.9250549714577957,
 0.02497936541815915,0.7207141235890695,0.4211235368458409,0.7983202462380575,
 0.19835303417628314,0.20424337778199653,0.024593232044668456,
 0.212443993330107,0.33132533659318064,0.034606083560359324,
 0.8453834836914288,0.6201045755757678,0.06257149557494823,0.6958857844846218,
 0.006293394471776348,0.8993400978966576,0.08128404138190182,
 0.9637665339513042,0.8765723050379802,0.23955267100825017,0.5695125634828935,
 0.7256269290655718,0.8594920613528918,0.5553345033813799,0.3578518171565024,
 0.11318456230416996,0.6751991144514129,0.027108059945613716,
 0.8254924527390174,0.4093811542427934,0.5455423286457401,0.09419837934515496,
 0.7771297058134662,0.09742906503845916,0.3623010480101999,0.3265656141545479,
 0.2082082982069018,0.44543887148190975,0.14039766836641698,
 0.8297152600157319,0.449955179889594,0.48473295187853205,0.2326412725909318,
 0.6363933061832113,0.7024725008340704,0.1614451436594011,0.41272070420213636,
 0.7327800468572224,0.3759785023407969,0.13429283860570895,0.8872764002366597,
 0.8235129199167234,0.016831780312170586,0.03596508426268841,
 0.07828965711605629,0.4320604860331234,0.807340865563946,0.4237297490478108,
 0.39294118893473196,0.42206948591767957,0.03286333963810539,
 0.07712692165503887,0.39415948233109543,0.18468388624311538,
 0.9621130211192394,0.054736251369870925,0.11966323705343485,
 0.28616540581923144,0.975460997759609,0.93189611283607,0.29417983684447235,
 0.013673072683972665,0.5996829989258299,0.514054506898092,0.9063373928991715,
 0.18821521530545227,0.9878645296277991,0.951611852792804,0.8579522186384635,
 0.38153265383762824,0.02704638203513121,0.4173876891251256,
 0.036118391909331304,0.5901787811466733,0.06477315039882592,
 0.0465764617125749,0.44970654068842897,0.5855986500668118,0.7362706298919425,
 0.7971637397608906,0.2692126723169712,0.09572502733174737,
 0.019352015740838113,0.28119665914352865,0.4395221392046522,
 0.06282042688246126,0.07608430170336722,0.8221225997038758,
 0.5004720377881322,0.2750005823458126,0.1169767330885022,0.20606061603224513,
 0.5689118293323963,0.15704787058794434,0.021166668008079648,
 0.5303501819495278,0.6328751945675175,0.8675283692097852,0.22426073140258473,
 0.1669196797202861,0.38891335777979286,0.5843935559090255,0.926193894022282,
 0.4520344572028554,0.9980986049495183,0.3589627317919737,0.9815701683220824,
 0.986549661008776,0.9358129513976091,0.6880770080629405,0.1579088984396051,
 0.8635482819145918,0.04264439122890451,0.021589261045289576,
 0.01993873689586323,0.09914331667471876,0.8724007865950643,
 0.4832954482879015,0.7812394748084309,0.2983285683368885,0.22651353336911928,
 0.4452226102553293,0.8436492139435129,0.3976801421940649,
 0.011803766232306612,0.7666803060860409,0.5339682167630145,
 0.15675736998251655,0.15538592202247015,0.8519204182479495,
 0.23204130943487133,0.1212624116047677,0.9605567484416724,
 0.36009131956861373,0.5748028281969342,0.6520629539367783,0.0895611626537901,
 0.4872368796690587,0.027959843905002835,0.03982003463474215,
 0.9632368291026678,0.47610980438681294,0.041980338783464854,
 0.19222693957318338,0.745472935511188,0.18812876788794997,0.9057829745788882,
 0.11873673894341408,0.514172359413786,0.7345185814097173,0.48032284923390756,
 0.5934439810586201,0.7724410751375594,0.9679589119782327,0.8555628951183141,
 0.713218614536851,0.3862311061322701,0.9236309548954841,0.4111511540188324,
 0.2155503611969306,0.6405524998423079,0.23921758283192585,
 0.17104152181760857,0.3198905533957458,0.3518684821920578,
 0.36277573546651953,0.17559417301298696,0.3584512407797704,
 0.6929363175246673,0.2734237848198302,0.6059563509862471,0.07266217022327393,
 0.5420732717521092,0.32548809640229714,0.3326636871666123,0.9786326412033894,
 0.4103656108574667,0.6056998054735249,0.685374846779081,0.35041830320725,
 0.02841502285589803,0.5165631720348365,0.7169010810747565,0.3164838854696814,
 0.5622160479590789,0.777743467762451,0.19115513644906312,0.9866251101274459,
 0.8236681821245879,0.5905150467257022,0.24881257075218266,0.8824036815477181,
 0.8932113047522355,0.7637942281118746,0.2189537388787719,0.05087825089301867,
 0.7952206445321058,0.5881350420791933,0.37368959904692867,0.8115541311472382,
 0.31009762202652913,0.7814772595608089,0.9060236392993976,0.67928641942787,
 0.2768945708991877,0.38236501404443135,0.4206660434686522,0.8630823988893283,
 0.7540398569545563,0.24627076138269544,0.8078304862849432,0.9087386044882021,
 0.9268452206747028,0.08483259487496408,0.29580995421437195,
 0.05060036109725197,0.3899204607155302,0.6524697047485182,0.9870498710038187,
 0.656208357168628,0.6323967900231706,0.6387301710585591,0.2923639820864147,
 0.8922862210728119,0.6606446017605395,0.3092175871766021,0.30134271147441316,
 0.5797880580650104,0.517408502995102,0.38274090578440134,0.768035535391856,
 0.6386214440777725,0.9809011187780688,0.08735548538207216,0.7736279699348284,
 0.16085008897567388,0.40900950503735023,0.2905504120788449,
 0.14205710911424352,0.9045368692243845,0.9000868879326838,0.4155476781197647,
 0.6164967103227575,0.26564855478213345,0.49694288759393135,
 0.0049626890864182815,0.1358938957715331,0.8723146031126771,
 0.5729276284902016,0.6347953189453793,0.03256761280206977,0.964852395729581,
 0.7466203703451236,0.9812361411351821,0.5345973733802264,0.1385429481575291,
 0.33832211129114453,0.09749821044340357,0.6888727495033904,
 0.3729189352779473,0.8696345872970233,0.956006003098807,0.33894305745291875,
 0.9718695249579623,0.23424653767355164,0.6668106148370141,
 0.21092420940585588,0.753578652045651,0.28059725867302787,0.9989755188043121,
 0.3624847495238581,0.2252324642319914,0.01574474285072891,0.5608085839671227,
 0.373716954439815,0.9779193410481261,0.7727439057922446,0.681598379542398,
 0.05137663435010187,0.0342908525948461,0.8731747847500501,0.56102332105991,
 0.5851355405644411,0.3859887138115118,0.7166857216051897,0.849909746800106,
 0.6882836863283257,0.9094160372903661,0.5907999817095192,0.6922276611060317,
 0.8922083094478277,0.4170955518498841,0.34180323942684643,0.690511370425122,
 0.7094396792193293,0.7596629694905238,0.20936503243863502,0.9420145511646034,
 0.6973900247741774,0.6538263413789582,0.9246618081025438,0.18308155398140558,
 0.4051514612642325,0.2588963836814947,0.2088863224969365,0.6022938607406501,
 0.3147335262434644,0.595542129842891,0.5242240934182226,0.5670820103844975,
 0.9221483741289044,0.9109262331357767,0.04081954889180828,0.8490622482146134,
 0.7501350356061535,0.741461744986562,0.5949376150108809,0.7668755829254766,
 0.4863005070082711,0.7927562549686064,0.1978891552724047,0.3725438915563538,
 0.6453412859885059,0.8395676595217434,0.49787172556744197,0.6295018736799058,
 0.42819698374065496,0.06188599999689037,0.9823716350868047,
 0.2898244001788728,0.9747138611031747,0.8719362340241081,0.9742764363176714,
 0.2719510324141765,0.8987531354091026,0.5597864554499108,0.8814595522598132,
 0.577290615767984,0.6302641256311368,0.5775188898310856,0.757477322089257,
 0.16151715635062136,0.9468861956569787,0.3209909597751902,
 0.41773412189069237,0.43670356693019974,0.9350077176739431,
 0.48666496037846074,0.5807737110097689,0.4027999237708275,
 0.10452954503306566,0.5276579754397581,0.4033023018587205,0.2306137921194229,
 0.9849243285798434,0.598450607714099,0.2940853493363986,0.0446951784750097,
 0.8413831980628339,0.10601795942354153,0.8629432782473009,
 0.45632305043035437,0.8326845639683882,0.4265025419375432,0.3871069076068243,
 0.8745199369035015,0.6611700350871592,0.7275813203081305,0.5541539324227279,
 0.11275648827832174,0.8432214747118963,0.8179319885141587,
 0.34603752246054786,0.09148152286909761,0.19828659503887558,
 0.332176727654777,0.4997472683203863,0.9811198296030866,0.1945389999372591,
 0.8644107758998472,0.4967306473477333,0.034093086825907726,
 0.09261436156179403,0.5379526944784127,0.25931382381722057,
 0.10392734918064916,0.5784824158941015,0.9126272264245558,0.6639108323242683,
 0.3797036758761718,0.8437066782037363,0.48738105607981197,0.9672486172742123,
 0.7721151093326163,0.5875963034000303,0.08961125633672995,
 0.32799467918684255,0.9941272584750199,0.1880410678889861,0.7292079390007719,
 0.8728801391901814,0.13055560229645213,0.7626241277496318,0.7622884340482419,
 0.07561414536580602,0.41122255165130517,0.5744403821485514,0.269921326995198,
 0.3066940557321456,0.9061742879981224,0.4214251255584177,0.5635847205925638,
 0.6176477028264342,0.6930496376170454,0.2503949596339806,0.7672615808562838,
 0.24925399766004308,0.615144561478357,0.008337016573833456,
 0.9205461183921907,0.7963072421054296,0.5265716393705799,0.9670062361272184,
 0.45940369793159697,0.11054057247371951,0.03480045338380933,
 0.2970467136368887,0.9745277377037642,0.9383448429330191,0.2747067697229646,
 0.1922330630394783,0.3562045173457419,0.9933911895425132,0.07468940186643103,
 0.5415087695292031,0.2746684776180388,0.769825301860641,0.007505141141582405,
 0.7185627092947452,0.14687852484525032,0.6321338371873976,0.7588767578189781,
 0.5185192073000332,0.0538953476843701,0.10899942562921794,
 3.0950865997647625e-4,0.2116721819178904,0.525820516680235,0.844717548654934,
 0.4243309515982201,0.2564198061942967,0.7826895986031456,0.818949766160964,
 0.38608866033441935,0.7994602098354032,0.2660452034410309,
 0.03586413775756658,0.9671505247039145,0.6320772029244847,0.6557881569708319,
 0.5240855044371554,0.47037767330807567,0.8702237861349325,0.391188005502042,
 0.23244546388155896,0.5631765066284582,0.4309537231878494,
 0.11385674718455485,0.24227341892592202,0.1702384627270812,0.95887739624836,
 0.37434026139307097,0.03895082173756581,0.09089690467363742,0.83416263764419,
 0.8216259877392699,0.3182575125309359,0.06709645979088541,0.2743597374425708,
 0.24224226724155085,0.5360633505884796,0.9597167031479106,
 0.47711961223341914,0.49754324247183634,0.40787770406827994,
 0.021262744847026838,0.08137277432683432,0.05925950566038929,
 0.7477145569375666,0.5207129065304297,0.5685412871610769,0.16714275575216597,
 0.5082521675689037,0.34610876407338464,0.4311459835176088,
 0.23982265426439886,0.26367307898833603,0.19342733314257632,
 0.12348326125515796,0.0019119304423390604,0.5768270391437319,
 0.13800476196335143,0.23352812006130486,0.7234546140431282,
 0.5513293285495475,0.7774034938143286,0.739385632616697,0.5931703068064772,
 0.03386935770249788,0.2338960443643331,0.377437281309299,0.09858763174342167,
 0.9904026980722564,0.7741170582594407,0.7504488149466849,0.1216931813002331,
 0.5654243116812014,0.6086919354547394,0.042659742645101306,
 0.27814474515751497,0.44688258732884445,0.7899396937653221,
 0.3782236495988046,0.9774402810763936,0.8423509170844153,0.4140346606504476,
 0.46901589591581627,0.7245730245963731,0.4698117850422703,0.6907421308778025,
 0.44365556001459894,0.15284182822120718,0.43867747871543694,
 0.45070094949847783,0.6524478060554744,0.9109493079710221,0.5968854060086772,
 0.7363971761931883,0.43047872742905047,0.19399530390794195,
 0.7275233111462908,0.525054658447499,0.26340317922934686,0.7394415790960799,
 0.058445864678402404,0.2101950046856037,0.5304195315719179,
 0.9457561379003883,0.11624510173148161,0.04821315641509316,
 0.6812497131749866,0.8247472805614134,0.8777207398543935,0.45652813700849504,
 0.9639949730614529,0.1242860335481707,0.9524140098007434,0.9682476246188113,
 0.8274214325473094,0.5031384952425437,0.06041469617842876,0.7365874486115532,
 0.3472739273225718,0.3087083845765015,0.9947595846484475,0.42636284978748695,
 0.15920015942794197,0.48099271998412296,8.483837197600375e-4,
 0.39079562330947126,0.6710636203970126,0.5316193483813137,0.7837467582159175,
 0.6359387416009672,0.633808545563102,0.7126367189174853,0.17885367814153008,
 0.7278568918952222,0.09624164821539327,0.19853484181359127,
 0.7940802827774806,0.9393169944380485,0.11495845266765758,0.8543100151048442,
 0.9006649198122167,0.10116698106632915,0.5694495434042863,0.8100342392142345,
 0.3294690515532803,0.8180398134845211,0.812337421420466,0.8570450124692017,
 0.3573014336609944,0.29543449160787727,0.3131985421746085,
 0.05618108253979759,0.582936722246094,0.9204836115314187,0.10523583256087965,
 0.03722462784692082,0.24696188322562984,0.09172721654061244,
 0.4532093456835351,0.12987117215393562,0.9271392297197341,0.816862999148724,
 0.17815217606168865,0.9761094895564781,0.9796585538939898,0.2226726901560847,
 0.7961491325592142,0.7464811009103823,0.9378335824610511,0.18490976893041533,
 0.25871067044261165,0.689776722841456,0.9514691848215375,0.2961199975822828,
 0.49634816708670515,0.9410176794362369,0.7395069072196502,
 0.12227163196925828,0.03609246520230425,0.40065828238033896,
 0.7519500543544188,0.8352750083482717,0.24643053044294727,0.7299363094318719,
 0.6342118054198115,0.5255989204784162,0.8627884126629264,0.14689944933011023,
 0.6952719028213069,0.5599560460500003,0.9131536202708281,0.6901057184972572,
 0.5146351178467664,0.7717014332144473,0.9919527535874026,0.5578094015389873,
 0.33517726154929606,0.5659580966670371,0.8468549782719156,0.6435163316063965,
 0.4713398518373201,0.3383309320258816,0.8592293230455565,0.8566742551046111,
 0.025659997042776217,0.2139098422603628,0.36064364827509654,
 0.3413438727599567,0.8292090203086436,0.8952662221642322,0.9102130299366683,
 0.21232399296169369,0.911871550682997,0.7814697138637967,0.7060547340954418,
 0.6927700588767711,0.01774164550547519,0.6141058739706241,
 0.43612315392517875,0.9710871216624151,0.5127172649776259,0.6285764822189873,
 0.679101416541051,0.7779220490651859,0.01688781397016126,0.7962375411163198,
 0.2675775353222105,0.36042280286797423,0.24229337233641257,
 0.5357100080044634,0.33226274537439293,0.46251810389803716,
 0.4153600659538059,0.42696775167096457,0.6992282324973713,
 0.34020700814863136,0.9954848363027622,0.22007122825714032,
 0.8795162763206477,0.5815160384417971,0.06463072454763985,0.2767556604154896,
 0.6411193827836476,0.8159210653902846,0.4423196562247813,0.4891589875757735,
 0.6292142277583915,0.12827325191078587,0.39240343028354674,
 0.02012226325877786,0.9132665086115784,0.049483687943053134,
 0.6434478615542542,0.6068729384940662,0.38940057909788794,0.7567448546739964,
 0.09118871958960717,0.32871361022739254,0.29942279578008,0.6848041525349756,
 0.07031600580852704,0.8733135596988775,0.847353709555895,0.07775584739433528,
 0.13475781034359524,0.5209333180354145,0.30256826590999997,
 0.1282611540647831,0.03336452267948509,0.8993876228029667,0.8695828406953158,
 0.040897178667001044,0.9783972634182152,0.2648873177444966,0.216649477467747,
 0.41127984020940034,0.2014565481339694,0.25242435250704964,
 0.2571075467539513,0.6932538351891255,0.08247991351558082,0.5165384576858632,
 0.7855129262214359,0.34695757652697956,0.3073514058834118,0.888526804501992,
 0.8967627714819216,0.8200484431700411,0.8578369416299186,0.7783754773226905,
 0.9216955143982498,0.7184915771771517,0.1450707023800787,0.9713501628888069,
 0.563971650603061,0.8054802103310308,0.8970009280053556,0.5735555983082787,
 0.7697536740054511,0.9123403623605252,0.4286888681659029,0.6404313862314737,
 0.9893523758732791,0.3045487203905771,0.677931604628925,0.6004926231995679,
 0.683306290749795,0.06746615860247684,0.8277718286696317,0.9961294695808787,
 0.08490758286808875,0.5599226794888581,0.574264643721778,0.3238019291924459,
 0.9474057591033064,0.7751216771196868,0.9019245517411039,0.26980474545224664,
 0.14476713993100987,0.044333526619530494,0.9504035863057667,
 0.35361229705122177,0.6525921116499862,0.08053954051397827,
 0.2095709654482153,0.38502849108341664,0.3954812428763609,0.2532958191604556,
 0.40670439929477575,0.4417094409392627,0.4133600630258074,0.7182875690601045,
 0.7740375693043734,0.4351790905862163,0.0442489673525811,
 0.035668097920224295,0.24150181443125218,0.8082253278560065,
 0.7043262621847581,0.8329934972672856,0.7849926577680694,0.5584940984119552,
 0.06451028260932956,0.2286573958076774,0.021387363068217424,
 0.29670573450078064,0.47275013901870355,0.8163136970587256,
 0.6040716471936557,0.8431837732853541,0.04776501748788897,0.7827639186444206,
 0.7013757435608357,0.8329757871151794,0.64868506579584,0.22907198773074589,
 0.11010768066427179,0.012141196561793466,0.9017233477176081,
 0.7781370029134025,0.8826733717864873,0.2170492347799604,0.13382653267679412,
 0.28559912046006,0.24282041881916683,0.020397038435473358,0.9711280875399798,
 0.07178002185708976,0.7307832578643807,0.30638596310500255,
 0.3851004649486929,0.9631635390207202,0.2950684473783102,0.6821431900732958,
 0.02266428742139348,0.6748789646794694,0.2836018066036028,
 0.03393485816999875,0.6459149276118729,0.20724690285490344,
 0.24890993078572854,0.7496881674597471,0.571942880769754,0.4427194834488639,
 0.30647289506887576,0.8440904166696442,0.8698168255783308,
 0.037334854374279414,0.8505719389826841,0.9282501837002779,0.482939389545586,
 0.750502482817691,0.2608626240451821,0.3283205433696307,0.20399563018397537,
 0.11133870142426572,0.1937856166929195,0.4829253812182749,0.5997261635706929,
 0.5937479307660236,0.1466436198568276,0.44098744792549116,0.035834959831333,
 0.232922059996445,0.4401641362161044,0.18822515224350345,
 0.030513115889728493,0.6666114601113812,0.04212518959013467,
 0.13697972706735007,0.7483244270164416,0.5388491332286398,0.7638909542226536,
 0.5479083022203622,0.5171542533038522,0.5463685237915661,0.8783855947115224,
 0.6780161258356162,0.39257443732089725,0.4179792718648875,0.7914058321031897,
 0.9911393617261907,0.5986288372198724,0.004777488738623714,
 0.4359071290832619,0.35177907255643937,0.3263001545896359,
 0.05596617065362275,0.9597045703101486,0.45845752451804733,
 0.7028350383957656,0.9398899038816408,0.6179150061666637,0.9741666921354527,
 0.6453671475100241,0.7532299637429745,0.8443328596384054,0.42637498769865434,
 0.10879797459764107,0.1220009574854759,0.7267235022186185,0.4200787362749042,
 0.7878326114404682,0.7544244225910023,0.46840704950845524,0.8388187599676024,
 0.1216693460599566,0.09691145784622301,0.15605570444755978,
 0.7981412718518067,0.6741203489465692,0.9259354996410503,0.04763509340438321,
 0.9546859736184775,0.19695966875377768,0.2695367164131236,
 0.31477351116461205,0.04961367097371838,0.5278250459849879,
 0.052307121572286897,0.5232280240588034,0.5669108553052916,
 0.6882159067238692,0.3281650836034895,0.27423718736214076,0.3989004522129781,
 0.7141284208353141,0.9965930899407434,0.450351781157838,0.3522177064779727,
 0.6034408874517795,0.6788558507334391,0.2970645824691286,0.25340777812264403,
 0.7568383817643523,0.5469450130066913,0.4121166883197861,0.5604506993938267,
 0.305798634196371,0.6660347910381996,0.3852818015701962,0.19192952451098078,
 0.044410709103672175,0.5593241758320222,0.8742064339681114,
 0.13374609015290595,0.0788170512615628,0.6709608317247342,0.3550258484911163,
 0.3134865963597121,0.20027274727464495,0.9922682006682074,
 0.10682055005380453,0.7666669957699298,0.7025387398553641,
 0.05155637229119525,0.501400967005952,0.11638512831636738,0.7202758180987503,
 0.4814030589457621,0.9872487163248336,0.7472680771822705,0.13522403831603114,
 0.4816456109541193,0.34050126583951323,0.8141480784322955,0.8376325346245815,
 0.2517925276982047,0.3644349546091238,0.4278199309760755,0.8059358541821717,
 0.931381559453255,0.9197425050747235,0.9206485500464199,0.07745228420759909,
 0.44924412886250276,0.8365810857529108,0.5177645393774691,0.6947385902413356,
 0.588664557627607,0.5610553824478016,0.19385686730342544,0.2451110915270558,
 0.23079682929176937,0.1648292496665169,0.3228308012169674,0.601591641228358,
 0.907499464850148,0.684049680786813,0.07997890226489845,0.6980956500745148,
 0.7749703105310479,0.28564778712923344,0.7638283529266687,0.5225842837695703,
 0.04382526369767015,0.691620273030786,0.7907468526350042,0.4847230972698854,
 0.910377989555157,0.43371652284099904,0.683394935595402,0.21725238001498148,
 0.18752696702327742,0.00309444799637415,0.13473615507666015,
 0.23158225251423226,0.09518362852121265,0.26001765325067694,
 0.8336759832456607,0.29181053212276187,0.5055150409400508,
 0.29169692614517095,0.9966064609350608,0.2428362144209728,
 0.24581792395222235,0.6187254647955094,0.49747099882004875,
 0.6845679104312288,0.5751699870933068,0.9349512624351469,0.3473938116150781,
 0.44448997800147283,0.18369162023508956,0.38322380437487447,
 0.7543807131280764,0.4048053824333875,0.8805708840306262,0.10252136555268954,
 0.7785444606129281,0.22781165190342,0.8830231646478002,0.16289217579400195,
 0.2089602512135058,0.14749868494744822,0.269897792990594,0.4363870055233636,
 0.634388038651569,0.795631709392275,0.48602048120374175,0.7655647000355628,
 0.22261362338284518,0.8704272535218094,0.21199340081774098,
 0.6074187512422031,0.29512003415805577,0.7617952194560127,0.9367869089913509,
 0.8732128877980532,0.9569596999761548,0.38930919543973674,
 0.02895153963201702,0.042813387131833025,0.39864971787617876,
 0.4735896611349253,0.3108193534263184,0.37537284837419993,0.6282543228069902,
 0.9424092976815359,0.4880528410668914,0.0669201786872683,0.4869696276584483,
 0.9162553034993399,0.06778967279230219,0.4928295169192579,
 0.12289376773690985,0.3920968553205577,0.20074139289310866,
 0.03546219582285881,0.9141918483420587,0.0943121235574027,0.6496277584783592,
 0.7580390286984874,0.3735227265041524,0.9117104301270447,0.30420819865017334,
 0.8342205944281944,0.09692510974216617,0.17796344813554554,
 0.06482584846675621,0.5418707071864546,0.7328704467769585,
 0.003601979196000249,0.9126959318869737,0.25214897206330456,
 0.5255403142128805,0.5190866547383254,0.463193746706801,0.589712801771352,
 0.08821866588024174,0.869795307842967,0.42430612586511696,
 0.055869582855725075,0.4229565972731941,0.601491422449155,
 0.12802099035368175,0.7887661190356956,0.06733878737538568,
 0.8283571320037963,0.8573866241577145,0.8184222751623907,0.9245097643352127,
 0.015313413049955815,0.43239476704491353,0.34883888734667157,
 0.6958142470818615,0.4008858092103398,0.5925596848129966,0.4163437536808514,
 0.8421093778582129,0.9715046353392935,0.3520279115168864,0.7312625441909797,
 0.7896392912302144,0.8823427434722102,0.7454669955316667,0.6517348602896762,
 0.12488924532238466,0.5849257944302232,0.7534451204494133,0.7725536266374229,
 0.1327484743744537,0.3600706528455695,0.6294348901781859,0.40748735847198025,
 0.4677976936450601,0.5759989758453701,0.7007213713923607,0.20516803596798405,
 0.3126696652222186,0.22711177634361446,0.4022014189186882,0.8072623185557409,
 0.48297261486592413,0.6236836341373003,0.12477218452732086,
 0.03154467896842361,0.4683957314787529,0.8561691324001098,0.8294687480881492,
 0.2598192143411835,0.3680887314021789,0.16666992727503227,0.4134264924613521,
 0.13402765277291528,0.5749669782883153,0.6108936104829342,
 0.07413954514583465,0.6269821156081565,0.3014798549157286,0.30176987804543,
 0.1446366474769728,0.16767531000350167,0.48864465857412775,0.63290785652544,
 0.5845163013079402,0.30147942052289145,0.5399630100000076,0.125174457028859,
 0.9132710497056808,0.5539678207425132,0.7087433048117481,0.14116478674344712,
 0.3433748162812372,0.8719120140534747,0.6023196705410996,0.6783020171572577,
 0.31659892624166497,0.0022672246190995615,0.5188429635404956,
 0.958695724782104,0.01210243716836834,0.5010094700085629,0.6475785722666731,
 0.169766663670037,0.9860161547530203,0.8022049742727414,0.3058185461331373,
 0.5064944557193305,0.8207782384721939,0.0774181072390222,0.2665787936277635,
 0.24323163201291242,0.5518902458030094,0.5642364796410223,0.3288782712195447,
 0.018519355013906225,0.15200574177971193,0.5625571732674612,
 0.8461440747831788,0.3397270340623917,0.22004693530241437,0.5034576750566524,
 0.7197564495714113,0.5810727675167858,0.8197141878406565,0.73640546033396,
 0.008811505627258076,0.07446581903890637,0.0301984618488228,
 0.46794291088416085,0.7116458601962299,0.8075412020838274,0.9811540578064626,
 0.7262424155580381,0.810172669968142,0.08622193744186502,0.5638838063109975,
 0.19337319729849067,0.5707958479819463,0.008283306423334302,
 0.8785376633198442,0.2865159264252809,0.4815238501571836,0.39562524503312635,
 0.04570670308277558,0.915016770219319,0.5241260087854503,0.8048089461824915,
 0.5856614296834647,0.7213251947170819,0.12058225839568537,0.6272003876901571,
 0.1672082938855487,0.49710426283777476,0.5640750855376409,0.4709469805879474,
 0.6021712250628025,0.7476331692668313,0.7443650296545797,0.07337861497857479,
 0.02799355722109298,0.314577847631864,0.49084374858024005,0.839971508619512,
 0.8535427771925346,0.48104575058276433,0.3571259951386424,0.0633122217327925,
 0.9388448054930242,0.7576069382526691,0.9569505575190288,0.574233822251809,
 0.707453765895635,0.09392402744739892,0.5285957948702263,0.6851389989045876,
 0.8757777124899233,0.417810637521684,0.3998904312800827,0.20322941689766516,
 0.3798782132660954,0.7628956240060983,0.9979851100264963,0.30950077571395007,
 0.13866463380018068,0.1966707708055596,0.24714755696226853,
 0.2778072072524488,0.9169911595422223,0.41035260140969754,0.7317574204446972,
 0.7344956325639292,0.7291218220142807,0.88914507571722,0.7369118345954477,
 0.528464129163178,0.7162305137471375,0.22608690855425595,0.2543022441391991,
 0.1465021183877534,0.9080290420670007,0.7941769251580112,
 0.004325714938464897,0.09569160756594297,0.3769124006130873,
 0.22207464689452394,0.15189660920482184,0.07941445307145678,
 0.5982952305818469,0.23177311066596495,0.0678184516086604,0.8540059938715036,
 0.8234518815469156,0.9950476839763346,0.8199191060140707,0.0764064971501498,
 0.3157453916638665,0.6192217095414728,0.8625992317504722,0.6922907690036532,
 0.541717505368006,0.40193444620339713,0.5680925516891617,0.9290431037705809,
 0.9480496197665643,0.08477099839519842,0.09387596570091561,
 0.7436308864034005,0.6529636589793838,0.05761968798065209,0.9251492155979443,
 0.7533582884093539,0.5569067569907215,0.617637796442098,0.35908870769069234,
 0.9409017642301881,0.08516962624173763,0.872035208021533,0.31087152577378196,
 0.7360210339985307,0.9067140900234992,0.6692696336993209,0.825911619430006,
 0.5447573527767325,0.8015133156723424,0.6448058150080146,0.5073452028613002,
 0.7673147413542178,0.09406193339947855,0.44188317754206785,
 0.9805534151217796,0.9222335386915543,0.5524034984173165,0.4065015136383072,
 0.596818956035494,0.14757983024384758,0.8399232586846377,0.7876750039842818,
 0.055552727467446106,0.38307858837869824,0.39969118432328543,
 0.4439976359776876,0.7410323883216827,0.7882431642364318,0.6249465124331837,
 0.5731137334865922,0.12267219388847139,0.9892372151343797,0.7621919586498779,
 0.7414009146270522,0.07111957015963266,0.8043374690502246,0.9478278423215084,
 0.7772788869449085,0.9089229002155736,0.8725773067976444,0.6131714890789239,
 0.48258363334997356,0.29667970498651364,0.5797255494435978,
 0.6003273645466379,0.06966842722430089,0.6116070677236743,
 0.14630559959842682,0.6478373473618065,0.24287956470100625,
 0.45806383239572923,0.8193987785426197,0.7631049055321965,0.1312874498266512,
 0.2592740002473508,0.46703463244733834,0.7404134047493164,0.487037506669344,
 0.39244703061928,0.7993024468397726,0.06128514371082461,0.8390138563525152,
 0.5362758446229959,0.47802356670911905,0.5343802010824419,0.2518153267435389,
 0.24408201762176662,0.2515613116848596,0.2532831431558332,0.4555335422156592,
 0.7008171373937622,0.8715120018798331,0.5898176192527143,0.2462409072555537,
 0.6827667780269927,0.4958644264220069,0.7844548445011222,0.28232083755925397,
 0.6494465763503051,0.853997324021894,0.9478310315366281,0.03071712624319678,
 0.844698273242432,0.47422780581300883,0.7364855654173176,0.6200553596275666,
 0.3716463382495735,0.39039309763694396,0.04040436847653828,
 0.8008205353058178,0.21539413718531486,0.5459378307449531,
 0.04711798304759762,0.4406563308693988,0.10100900375124489,
 0.2725443670986246,0.9122119147974497,0.012007031215371278,
 0.9614047176147851,0.6718208729784926,0.7228814864689981,0.6146709352154708,
 9.893742298705543e-5,0.5410068363937575,0.15864577724278062,
 0.8985956290992092,0.020403447760089732,0.5672780481404658,
 0.8786578230864719,0.33514363789701596,0.8772675347402645,0.9169386489446811,
 0.39447933996720763,0.6891254612601043,0.6686484287649912,0.5306218521739442,
 0.2356195091129899,0.5490577180867001,0.5545844996805359,0.34910722422007534,
 0.31693893701722986,0.4708026701772644,0.9180058832666721,0.902405523576198,
 0.9050080631851903,0.5670580341999105,0.7953006111109457,0.4831933034569367,
 0.5703224674392071,0.7636440261712399,0.5308259468227149,0.16849232000264003,
 0.0013667921063450494,0.8439798050842243,0.4033718661874719,
 0.3791202279765993,0.8336624990650854,0.592760054380202,0.38680117349287946,
 0.4215231243556169,0.3455414119288984,0.22103451867939616,0.1388814540101594,
 0.5599200889510986,0.21516435839764414,0.29545606060878227,
 0.9542463474699343,0.6079414172403641,0.21616698018346192,0.6758063346443799,
 0.11797241297090943,0.4230116572758926,0.6082513979028056,
 0.15840695334734012,0.03674279084474863,0.3793315847610671,
 0.7959823477707088,0.5585673284599358,0.6240716991606354,0.7454566183482549,
 0.13472414032508662,0.33702293707453035,0.38065008951688617,
 0.1790552949110894,0.7825201343821324,0.9720939324494507,0.07506337660944817,
 0.07766552732220999,0.2988855062216892,0.30321339292598803,
 0.8191895036144087,0.5715319501382364,0.8268361589702478,0.265062670627506,
 0.2895209743823848,0.05740828341128046,0.9598777558570325,0.9224003432867207,
 0.6741126702807482,0.24206302313527217,0.2101338741304959,
 0.08443816302971996,0.2546565939622062,0.7185052772317521,
 0.41960495897432426,0.1307597308084486,0.49225346284470295,
 0.6674650532212066,0.06632995936727748,0.18264344309888414,
 0.1988072055159691,0.10122318832427513,0.003921896750559317,
 0.7523626400598018,0.26322777675266207,0.9533225098970389,0.4298139156184052,
 0.32447184497060355,0.35546353447649603,0.11334809447092786,
 0.45288780857437017,0.33121093722559247,0.6825659372599974,
 0.7283262245629674,0.4913814701992756,0.05495113852423905,0.6584841476009389,
 0.800292214516418,0.06566296860480147,0.5892390746676001,0.24717864760062913,
 0.06449551246344609,0.7948049952594849,0.9030325510534855,0.7608811512534528,
 0.7237428773654705,0.5854584990952525,0.38713365720782944,0.8238522347159742,
 0.3390690521607118,0.3368391631287696,0.6699239516645269,0.29091752658193337,
 0.6699246902302588,0.001433169516197097,0.5521075279383331,
 0.9046948653879252,0.48412606626553634,0.3675676351841197,0.5674879290039031,
 0.1342660843410899,0.33156399543181836,0.474444985161229,0.8496233916279945,
 0.46640129646465156,0.38124156191409375,0.573113707132535,0.4907873407725847,
 0.03108539999016302,0.8255978007508145,0.1056725200998695,0.1229034703448203,
 0.5087574993931607,0.5891612024669554,0.1462748730049117,0.19211843866289247,
 0.48182668672123785,0.7335357944064477,0.22793423009158365,
 0.7613408507841155,0.003270308913622011,0.543096589489801,0.5664208547003702,
 0.949904155580867,0.10276620188757923,0.37513207039502094,
 0.20291342554629888,0.41853366008560333,0.9462541824389394,
 0.9978658886792267,0.6751946870264374,0.6046434328678859,0.6691069910328853,
 0.2819884461151476,0.22353903018822807,0.9549827611211227,
 0.009814712467373865,0.9434277040996194,0.34538803365504955,
 0.009935493331630463,0.7779253196388,0.1656134073891623,0.6256063516562435,
 0.9161440167432876,0.12507132110942776,0.5343434049107907,0.3179080510451744,
 0.10040621665347582,0.8604734358691756,0.9893352194687606,
 0.13615183786458807,0.3383984236080816,0.23008577398218555,
 0.21678200410270698,0.023887893443490182,0.9888519505389927,
 0.03242123892536275,0.4161284046821343,0.4678608916300485,0.8613417653461444,
 0.8698681929022217,0.12039432132516503,0.03593702898864404,
 0.10245094512294894,0.9194400319176602,0.8941784661356973,0.4074605115600457,
 0.7012461977075958,0.5639564585354194,0.4013057088244586,0.5947474416288321,
 0.007828660314886493,0.5088460926687917,0.8250362099453538,
 0.2151625560812791,0.36373781330555655,0.35051815675392395,
 0.8479905273395505,0.8532900037865422,0.9588307775416848,0.8687826619076873,
 0.8321431492974831,0.9441783394138531,0.15090359510829454,0.9375844469238077,
 0.3445555050491338,0.3718555062654485,0.028700968431281004,
 0.12969296102270667,0.7140943545682926,0.6926076196277159,0.8801169735102607,
 0.7196958488768741,0.8955921697275292,0.34043283917128364,0.5746539212211862,
 0.5111044069304703,0.9800932223102368,0.4328702571029073,0.6845306601506763,
 0.0930417420180033,0.7929789814495378,0.5073249493476949,0.6581664492924709,
 0.6523438309045491,0.14187821923725086,0.30757122927384317,
 0.9987460250493769,0.1640814731371485,0.9646486483930953,0.4217624165788145,
 0.029140629472290125,0.8926389753150046,0.5336040757263254,
 0.6670041022284543,0.3427264710658615,0.542001341992313,0.2932952195258931,
 0.34869903227930354,0.9768780447990397,0.31242643936645564,0.195202958565593,
 0.7404518908191425,0.6163871732535207,0.06866433811541472,
 0.40176783704852115,0.11480804828891145,0.5784533331867673,
 0.5906020735890506,0.8280073393543014,0.2726262775968742,0.36821475875253806,
 0.35979334255811446,0.02259833718491,0.08893586403950304,0.27873823534983266,
 0.13673065553963737,0.7347201941965367,0.48273952310952495,
 0.6260698281719891,0.3572869298376238,0.7338847789903089,0.7883820365062034,
 0.5837715537563749,0.8073981541209838,0.7037258066861666,0.857651088120725,
 0.7790745548224356,0.15664483158385722,0.8495505601965061,0.9862893193053869,
 0.8504235247927711,0.4560555904181972,0.4391310386235422,0.8768666402909426,
 0.020934814397207346,0.7573940434649173,0.5241423277026274,
 0.6120848508586023,0.5820423153772675,0.37755879368638645,0.6003520622201232,
 0.7567888038346492,0.8980006191399594,0.5783828951951536,0.4417408560934786,
 0.03015894716694545,0.17563723167780743,0.5607033765410346,
 0.2662862337612546,0.17185142585688318,0.7070029230330751,
 0.41975020302936816,0.9658826844200245,0.05151348756289398,
 0.6593107105139832,0.24143385005408569,0.9186922926380284,0.7133977757819032,
 0.3370335385226947,0.29605205589123007,0.739419038546328,0.5162048500005504,
 0.3594036203398353,0.32892774113222256,0.8793269148307243,
 0.22564376955452192,0.25270824921802093,0.4671661389084043,
 0.9565402430334718,0.8170894969938944,0.7926596759925949,0.41014833377399773,
 0.10190054810974736,0.7192001192221131,0.48097627867335635,
 0.2245790232856607,0.8078977614502965,0.8194806267025454,0.3205311273371787,
 0.6789466610766159,0.24965129409490938,0.3184935688660351,0.4760909371358377,
 0.7766355803467869,0.572131129632337,0.6274833475477738,0.351557784346182,
 0.377729353282229,0.9718646804500318,0.5717742399024305,0.6267962400310162,
 0.4920526351451243,0.6456205270223725,0.255034251835748,0.2073991761485635,
 0.11116812950797217,0.04640130219458871,0.9357997267199619,0.628713990222299,
 0.08482929730874644,0.25317689524750253,0.39960330873331884,
 0.8207007824293202,0.057990914141088146,0.1888340365211194,
 0.21787298745159012,0.8902748418204177,0.2694126236410387,
 0.010431151889921164,0.2609252004832462,0.77579243008424,0.37142580039793327,
 0.3229446601665572,0.040046598811343426,0.9299157519688043,
 0.2667895144126493,0.6334492076695397,0.09285748352375145,0.5489168535966903,
 0.3606433269079198,0.9841884935745663,0.7291389352602115,0.8466378855684823,
 0.1695457202447701,0.7463857791166437,0.9672494349063994,0.6906836044818072,
 0.9949889183881341,0.1316239474867481,0.4168429325058607,0.32654968320831845,
 0.511026311071986,0.5980371653782227,0.27769846632776873,0.5950827264849832,
 0.9560749314974946,0.8231360374834799,0.9118287214319494,0.21293835116340842,
 0.8331006352602629,0.022551930370666162,0.5133819028644806,
 0.5230148409488161,0.40347929287896733,0.43790657344264194,
 0.6017289480310742,0.6029517723015411,0.2279738085126668,0.287057163002158,
 0.09484356357545354,0.46512625809190833,0.45831367493348873,
 0.11678216943986908,0.5820201396468373,0.8191128096020552,0.632916567959786,
 0.14501171541703795,0.9663817712673753,0.7446375643792238,0.5721712598542551,
 0.06923496295343667,0.5910767908377316,0.22747345371106054,
 0.22444154320973886,0.67026526270501,0.5901480942503091,0.7091711610016563,
 0.08894237871319588,0.4310138920862234,0.5731397133922636,0.7747534658110558,
 0.9552685222875313,0.72160711621424,0.9560903115874105,0.02860259267878673,
 0.043468734719108504,0.09053315901902392,0.7896971855813212,
 0.8816059482616475,0.06313819705686252,0.3862883870770063,0.6054384178698062,
 0.5255542633136009,0.999420748377263,0.6701097985900574,0.521554680123337,
 0.8881062694575139,0.5948431452960942,0.15443229124328206,0.4972568460101525,
 0.3254798635823837,0.9472933767712243,0.9467967932656576,0.5063807535386952,
 0.31521155104838205,0.0818929401480285,0.14578949635387273,
 0.10570727888416231,0.9247465753764215,0.8560920536296257,
 0.48946067501448376,0.0067838940074866905,0.43739074948000467,
 0.034360412797756146,0.04597613533899614,0.4880196320794614,
 0.9636464828448845,0.6568886712821592,0.09382655270359996,
 0.09682769903752031,0.07260227462307267,0.1340133691018499,
 0.9296150999874563,0.28846813613769096,0.9705295006188402,
 0.10634866247197472,0.5405895586579582,0.5238503276038766,
 0.05926546204530725,0.22258895637363563,0.04136300565203899,
 0.5803815908527866,0.986140953236714,0.35708225783494973,0.3126853298917932,
 0.20453694294351976,0.6493652442742404,0.7293492802572575,0.1034234252591455,
 0.9074690925521678,0.3496100246502939,0.4281522870187259,0.676738546215003,
 0.9106910525622764,0.5624539952748483,0.7120722432149353,0.7249037213568897,
 0.3077778997030901,0.30766560879179206,0.5145648595816763,0.6627516834019409,
 0.1809010522653074,0.17968826671138538,0.047796163123527746,
 0.04810348147045884,0.6179063407000891,0.964021169211172,0.8437502615712033,
 0.2326127859857795,0.09867212242613144,0.43833177909234167,
 0.4128228021047212,0.27462103793034964,0.3725828758262866,0.75444637572842,
 0.9301880371303866,0.9561345860846096,0.8261291826234077,0.08480544373600174,
 0.8011877144214502,0.28585228648942695,0.6979135279756057,0.7560199097150114,
 0.6481999527924754,0.33306941903845355,0.3006364256818336,0.8243605555771724,
 0.5764641658495947,0.6955303057155197,0.4444680905722981,0.6126249283458807,
 0.5546909109681274,0.2748142614895951,0.1856428135498175,0.5699316219997127,
 0.3567027523234033,0.20400730468255768,0.04901836289913308,
 0.17094296870210335,0.6363302318378878,0.7296009239694523,
 0.21903327368668912,0.01607780404429837,0.8044788736346304,
 0.6051274433411177,0.41952950039577686,0.8484862941811996,0.9257614071785144,
 0.6668206265832194,0.1473514479553263,0.5319737694491845,0.8893632252401131,
 0.32613948910522117,0.6570359188770947,0.7604605743609721,0.1910312490263335,
 0.3386078753995363,0.22203623161717678,0.9543256125331022,0.5959294521471854,
 0.6309007250054045,0.31354286980867907,0.5642443787082582,0.3665156239830667,
 0.2691113944565343,0.7874050578290471,0.8908199703959433,0.49344223770829354,
 0.28369444838124136,0.5276484288525602,0.06455828079057568,
 0.42913822489259246,0.8027112501747484,0.4818152560624803,0.5180552931140694,
 0.9479662821026511,0.5372709602364798,0.15081573842532658,
 0.15144309517934007,0.5754018310274864,0.20630307178410212,0.743596803886446,
 0.7788654004573213,0.790690169250235,0.37201390340278795,0.8656371115950814,
 0.029751676763677293,0.4131499198101404,0.9630975586748536,
 0.7028470751290101,0.42045598292503095,0.9396814393265412,0.6102651835473936,
 0.3149010567578442,0.4330517744775354,0.07605727180636812,
 0.43071256043750417,0.6364949010048714,0.38433354403311937,
 0.41865352173330517,0.3776570764094068,0.6953443406527982,0.8339827250866473,
 0.6468914557109151,0.6530433280525906,0.41082220521906965,
 0.25243230958540797,0.775675125723211,0.8606835099118474,0.8255467958240604,
 0.8955542453101605,0.15929025171452194,0.8502742718852871,0.6881772223519107,
 0.2647506545669023,0.3165499942168879,0.9758245336960982,0.8837045345612496,
 0.7011536954518748,0.49456767205424323,0.8118310710721377,0.4267437678820617,
 0.5389843584922591,0.8440131581720975,0.8626764044945555,0.9839961614578007,
 0.43299734974897064,0.20593443546764756,0.37664616966922415,
 0.555929429032578,0.9111900572186774,0.34012522231541964,
 0.026730933693793424,0.300151722741431,0.7134829193460386,0.5865171919291712,
 0.8607717372129435,0.40664788590704115,0.7724091125780118,
 0.13716076800608523,0.7700476141053926,0.5812866591127062,0.7209165581452898,
 0.45141324367272506,0.5413679201118128,0.539209074836522,0.9393885415991104,
 0.18977926129240952,0.736353138280991,0.5685285620912105,0.8858991050699376,
 0.09531265778933473,0.5305275831747737,0.05037520530198225,
 0.8266628522875461,0.03190049672756734,0.1422291336554189,
 0.22582827727819432,0.30187874833297446,0.7187440493413869,
 0.6712525794577121,0.2660785918472881,0.5972355262324144,
 0.025364147764588174,0.30270094553067195,0.5691297693871142,
 0.5192157378814665,0.32137993955669375,0.48269576434908135,
 0.3885707644137655,0.4979158131091206,0.7523128527575526,0.582482839857833,
 0.18419253922456214,0.4547901836637245,0.1990837253332678,
 0.03398057881906014,0.7984226908761883,0.3330980242599253,
 0.11028664862569637,0.9552123448048888,0.9360421344782274,
 0.08400130460452415,0.43313296683055935,0.2962370861992537,
 0.5512835097677161,0.30773377856644224,0.7089857660127828,0.9348108066823342,
 0.9081134240530807,0.036017960166773166,0.3901044909658641,0.545815673279845,
 0.21340801388135733,0.6035114824663252,0.7180029424958243,0.8392496712615838,
 0.8746424721881789,0.8943757924603475,0.9911030191561409,0.8094685099629544,
 0.7112893487039328,0.6401394080129865,0.9726020441005865,0.593737682857989,
 0.9407007134265966,0.5228557849044057,0.15828588641278463,0.2399573568330241,
 0.8603453861090739,0.2557401515717169,0.4536386144218527,0.9128206478345994,
 0.26619432315411806,0.6169367559965403,0.03644212766992505,
 0.4211976156245353,0.1683033605086912,0.43712475288003294,0.8873251547817104,
 0.902247881925473,0.40230419336035583,0.7397503228809867,0.07231983164113576,
 0.4967355033480356,0.6736507235635982,0.136471717387873,0.08315537563607112,
 0.2751197335099864,0.5630686089631627,0.18995711269118742,
 0.46956887496854804,0.04496173836003259,0.5595494205052886,
 0.2559948560176146,0.1970133446171587,0.5015754507346253,0.5993819922908092,
 0.0991533622987486,0.756269388123448,0.7350914861922552,0.45011294473622243,
 0.4497519629904597,0.0871372435935387,0.9312681475657677,0.45342649442169913,
 0.9734237788421302,0.7513964963003041,0.2423372822767006,0.19807432979818962,
 0.843042243872602,0.24742829531843202,0.2798403815004886,0.08809253905684522,
 0.225619686014782,0.4305218833919171,0.3213915951618216,0.011870606069542156,
 0.41401166424184765,0.0795813134274237,0.927215209433604,0.1468441688715878,
 0.06362199634467336,0.4535929987942925,0.29776421322288804,
 0.3966665856681294,0.48966198367092595,0.1354720837673753,0.886837180000589,
 0.31928753124104425,0.7850376307150787,0.13052150017844633,
 0.36858722395366417,0.9315296573268214,0.9452396374181159,0.9584356449343462,
 0.622447142017594,0.1899999276546922,0.2523052260368449,0.6825574114766246,
 0.5813460673718878,0.8287066807939452,0.897865487311681,0.9775765876522609,
 0.43944751757653766,0.6361396334863391,0.12935721079817575,
 0.9389786728701757,0.9488166682961434,0.892327584829732,0.16795793149138094,
 0.4881195913395039,0.06861282965795856,0.33447436959014765,
 0.6989347787377156,0.004158992389928229,0.24637378653181718,
 0.9122204568035055,0.6150809412493865,0.7769947314311272,0.5571136555890386,
 0.6301124403324984,0.20379970802851,0.36749110205318625,0.015114307651983405,
 0.7561497297821409,0.9335576648121678,0.10580547781589611,0.6323360493395719,
 0.6627245064060301,0.8310496369605432,0.9021640464572575,0.6999565768793652,
 0.5349902499399974,0.4259200355434777,0.14046669165421288,0.565419934656172,
 0.4201889542205758,0.9043509989276791,0.5650288009833075,0.34507891991583883,
 0.9784463606863901,0.6973236432547012,0.16450382664023933,
 0.37779724831721073,0.4652781304783412,0.40483760080760955,
 0.49842950135008657,0.033441363217724285,0.3371629152565979,
 0.08630289732274377,0.16879117601055338,0.8079154279512064,
 0.7747615330504466,0.11815113171257574,0.294975875190326,0.5045531384040293,
 0.3816584040205421,0.42005768779676234,0.6191847225477008,0.4428026590389913,
 0.051993133758234,0.7696774008976336,0.7942346644015768,0.05966067866886171,
 0.5082713800726663,0.06612323977706347,0.050590962673304984,
 0.05978241497103143,0.24361561652450914,0.8596770503914478,
 0.5498805691781572,0.6888332113737228,0.9631994426698113,0.9758222687388126,
 0.505136997611766,0.10877283827464757,0.7671321611877846,0.6811120345833273,
 0.6497049070122733,0.8490374629380059,0.5610648091664291,0.6667585356194681,
 0.1577233676011882,0.21912392180377427,0.4415652558829287,
 0.10041774837002893,0.3929651894207177,0.6759791791075738,
 0.061922318519335606,0.21995333649917248,0.6995676069693346,
 0.8275022235581838,0.0038565134953159053,0.3558363641238108,
 0.8319521925361296,0.2774086888699653,0.16575298646529357,0.6310330274419753,
 0.6278765081051324,0.08131836043016016,0.5348838432626373,
 0.25985228078546374,0.9911317890430702,0.3228485380426874,0.5726623740267319,
 0.17991532365917018,0.7206432271474723,0.9240945524363567,
 0.31301422082709796,0.8438427765303381,0.98028153694436,0.41325342916692354,
 0.2716659010446282,0.39308253776405844,0.3136347878030792,
 0.09399274889623022,0.6073823862732649,0.4868514769598391,0.2802362993691193,
 0.28844468014484625,0.04967154059996193,0.3294521205406723,
 0.49639865268485983,0.26610448746965365,0.26657108686676856,
 0.11636475357600595,0.8617282564339359,0.05302644611097185,
 0.34663597387772094,0.16781859784348907,0.0739866102325023,
 0.7863379573805984,0.3524891833999715,0.28343457670712446,
 0.19794762221548312,0.165889522602246,0.5799372407444693,0.25207511329531196,
 0.006851150645002635,0.5806768681000853,0.27107058770088477,
 0.9225733570120592,0.25688134306732713,0.6510319167926114,0.7226754238354441,
 0.17871164501402326,0.38803663899262986,0.22670004118233,0.1692688370113573,
 0.6457132503503958,0.2469748220096295,0.9975957270464093,
 0.029155015478846913,0.3307553764243396,0.6854598819109323,
 0.3349337431829309,0.9197510713570392,0.3147002449042009,0.36117352209816067,
 0.09583626462153005,0.09462566549342744,0.3500320408429487,
 0.5905133596002754,0.607311571135542,0.5142844453852254,0.4275849525482799,
 0.25720611091578327,0.20114033197871417,0.18085117610233137,
 0.562021514733858,0.12169102434633083,0.7788679893472591,0.8137886270264347,
 0.5675459702305601,0.5449097356964143,0.2333108143683279,0.09514073069220641,
 0.5163940179107326,0.36352358882302027,0.2772197968490442,
 0.32658878029381677,0.9726539550148914,0.5608395435074541,
 0.33326329980792446,0.8600585802756945,0.6235988017966316,0.8390054377830306,
 0.9564187712760791,0.791328966156559,0.30940479575635094,0.06545425233597624,
 0.9673867534385853,0.16931890877170974,0.39884113321924486,
 0.11726117142184111,0.19157934424976975,0.0515558095048948,
 0.03451474117634279,0.05339745267311702,0.6559008798678919,
 0.6355816575233374,0.19050085509874748,0.9427854346000962,
 0.33528726000037556,0.4623882719397183,0.22319060017067283,
 0.15610922595985732,0.3393419434171887,0.4569689608881552,0.1911191551714434,
 0.05174152576556812,0.8550531493857356,0.6607765439718045,0.6505956467565076,
 0.9635588215307263,0.5148656737492169,0.3505908048834987,0.6395484351950282,
 0.04723849361427401,0.8306139435913937,0.5295073121817363,
 0.32671382955491235,0.26914633198895754,0.27893786500661544,
 0.8185739954145334,0.4434317952604562,0.446273446999081,0.6171357729036179,
 0.2954300306476644,0.66608681981704,0.6655807247083673,0.274055616928893,
 0.8653614725392357,0.4967583060052849,0.6087500054482792,0.724742958251075,
 0.6511709261462695,0.7392766856391253,0.8949917317794625,0.18667545019489873,
 0.48781868266412987,0.15198411516128862,0.5235005702207135,
 0.08080806247863725,0.9117210330246792,0.891565076013753,
 0.020029852985206098,0.8009859563107997,0.34434042735654025,
 0.8425810719057515,0.7065452301314675,0.260577702803239,0.6994239819880839,
 0.550956729421978,0.21741679585940932,0.8148006911483944,0.6166416763259907,
 0.06620098950805031,0.5902437325978714,0.23682331700509418,0.689808563786898,
 0.08556029917447128,0.9131869980304466,0.9661767325362609,0.5447367973950018,
 0.08924629769034087,0.860184209383333,0.9995747501745517,0.2361055094819946,
 0.8613935666924929,0.3713908535248688,0.6167761751320406,0.3022468380152694,
 0.3460847886538754,0.07170902932355538,0.7881753929536677,0.8032623631921404,
 0.6865155745627148,0.8720564349280672,0.648305749440065,0.623240960857933,
 0.19055819356789427,0.9905067088604645,0.7486308223771736,
 0.07324640909809566,0.21685224498621392,0.10874003466786764,
 0.15758883063326057,0.8383632986485745,0.9075417829561303,0.8894366174268132,
 0.17825552531261046,0.1396733150615136,0.41516784753045677,
 0.8247640026990065,0.5639301833752303,0.6514786335061835,0.3911479028184618,
 0.7965541451361141,0.3934109388577849,0.10982497284739523,
 0.07117601280367669,0.38317182029059094,0.8255660863871982,
 0.6419497801857149,0.6391028414080479,0.6140759899389712,0.35673875141811373,
 0.8118062605270162,0.22761241273478205,0.8708259321015628,0.6655567749736933,
 0.5089424318883897,0.525760845229571,0.2079836040411066,0.5470678070804961,
 0.7692169314269109,0.3156804658223029,0.2161434719093669,0.8961255454913359,
 0.8717936959423694,0.650475853434884,0.24266643585375602,0.8868031034966507,
 0.1884279911277983,0.33883001254790746,0.0961061043203233,0.4723160549261418,
 0.4847722596823114,0.7391007961689786,0.07469678026865967,0.5210806759490861,
 0.9921060472718674,0.521588168395395,0.5204657550272141,0.7631085072552598,
 0.10569455869660127,0.8200773771123063,0.22886026819305405,
 0.03508346268250628,0.9698216214166874,0.5494112607058257,0.2738853967090281,
 0.27526645908932623,0.2062279421432518,0.6067674203453574,0.868391985696312,
 0.24845226917652563,0.8085318475342876,0.2968053343526914,0.9429674968817181,
 0.9825705636360749,0.22654049762804873,6.055292789691746e-4,
 0.4510774351533431,0.36312089554604465,0.46188031728827317,
 0.5649681124818764,0.4478320093500603,0.4199385217606142,0.13478531575175567,
 0.7067528769420679,0.2867332345324459,0.04175105085000941,0.3945367385541384,
 0.09811047296373765,0.6588873331433802,0.4483243098824121,
 0.14957798374601428,0.4848600506997983,0.2727418617033507,
 0.21098116924257326,0.12805581348495143,0.04928042646339015,
 0.01463644840465883,0.7016453633618135,0.5597299893889401,0.709887120277347,
 0.8347930351736454,0.14957344113700488,0.5425654104615368,0.3671737572688588,
 0.013472080467599512,0.20463320551676933,0.13570382848379436,
 0.918849498782969,0.7883195688756213,0.21578993708972916,0.0507687379260271,
 0.5198369985000667,0.6828686186846671,0.7338431976305403,0.5418967519935641,
 0.9354652317738904,0.44106437854713254,0.01850957159926936,
 0.3981569201242042,0.9883019088061948,0.2697061239317604,0.34113565169166193,
 0.25987406773437394,0.5105377697886816,0.6327418609577418,0.65579692460737,
 0.14586837390745777,0.22127335838430684,0.35653790294893983,
 0.5721953969026313,0.08892092957212117,0.6711350299061924,0.6069588354546834,
 0.3507649636285679,0.7590700629136694,0.9685949311529192,0.491307330269104,
 0.2781717842269329,0.06591182184430111,0.5261110785490288,0.9164111838977881,
 0.18727066881238363,0.6947078867727545,0.22233206003089823,
 0.9397706956013341,0.25502483838688594,0.6800542276259717,0.2376145042233715,
 0.8643673875008652,0.9413366369177617,0.2073120113761293,0.9953138197542213,
 0.5632115107280088,0.24981811067488435,0.4730509647421781,0.4381957539837742,
 0.2855420480898406,0.3423461054956701,0.9121960022309825,0.8676402405896633,
 0.5355035311207262,0.009201758530004045,0.795785369618502,
 0.11246162458564801,0.9118858415871488,0.01708304330189847,
 0.7179312561565909,0.27497163523621815,0.12026382742841069,
 0.4944397586117403,0.6443524983972799,0.911415517562899,0.5546979197199245,
 0.5984169876496808,0.8921992625378239,0.27873275683028564,0.5846321302829236,
 0.12863317716742506,0.8175473909510877,0.016608436039479813,
 0.10258804689182299,0.5689048645338055,0.7590108022677385,
 0.09083276487598624,0.5947048585498403,0.9293241565410801,0.8740635823653828,
 0.066110459123637,0.3712671903770248,0.045744566642020534,0.8960530054821328,
 0.561202207144929,0.995546223565393,0.6921950688413205,0.14394560054495198,
 0.43441072647700374,0.5025039890655999,0.06105462246795257,
 0.6504411992210128,0.17612531070567694,0.6256029614529736,0.4028797418929626,
 0.3649919361088193,0.9043167450537344,0.5514264001228634,0.4481516296121033,
 0.4144023889016588,0.3360664350934034,0.19498904791008242,0.739537684642212,
 0.646078206996005,0.6672672113948874,0.6375964705355386,0.4236256237788232,
 0.35667887542197985,0.9807616790629446,0.7258082681860031,0.809380342050956,
 0.3134240406177643,0.2671028059151218,0.39626578067254503,
 0.004287649088822176,0.3842270275777423,0.2296998107614071,
 0.4556945814682889,0.5882192198173125,0.8469249916168956,0.3118305553187106,
 0.2103574324936638,0.2150301222753861,0.12250295398741584,0.8749585839108291,
 0.47193689606506806,0.8105024521024036,0.14403765323536344,
 0.31454262370269204,0.5507674660855234,0.9459239022037199,0.6945421463225523,
 0.1185060443304563,0.731605956895581,0.032059286293848244,
 0.18719283239480444,0.7633168599926032,0.36395805437909945,
 0.5681431756388752,0.574605764170562,0.5150294391576361,0.8357100744101872,
 0.3339029429067414,0.6537399890216741,0.9554921165828104,0.221261806585485,
 0.39272412076397445,0.32090062450898027,0.5918944443266816,
 8.484618243853603e-4,0.6022154624904175,0.5085794960063644,
 0.4501268066112375,0.20899282343297731,0.13007666606241275,
 0.19364868998113105,0.22621461364803086,0.6056302316261625,
 0.8855994772521377,0.7039898992108967,0.7220190593250737,0.6166345365280067,
 0.8019791423624271,0.9806783809533443,0.7348989566814128,0.5856428542665313,
 0.40586404389352737,0.025382019691836955,0.4400067623805465,
 0.48129792468878085,0.6217701382828362,0.4418839807608469,
 0.22265432738860513,0.14045997077606598,0.6375271151619311,
 0.7929921449817393,0.8358190558709276,0.7950044636639965,0.23910817703201248,
 0.10570074803044793,0.4591252286076005,0.9157388350424034,0.7009319355982818,
 0.6600616628077499,0.6940120511442598,0.69131220112482,0.30117815349858446,
 0.13804690712728018,0.872709893540164,0.7389763594474363,0.45009345514626276,
 0.6021388754738516,0.8165868329899091,0.1581489860674996,0.186323327100415,
 0.36331138016638675,0.9255762652480215,0.7011228299375639,
 0.19741878781704647,0.31615588236378545,0.1744735531362983,
 0.40465255065923145,0.11294140081194537,0.9420336800514905,
 0.46917092153757833,0.4002831564784356,0.35643410238696216,
 0.9620713716818531,0.6064722203077294,0.7048621361129426,
 0.041947179808330626,0.8715044203869866,0.4216372880782415,
 0.7288537292070215,0.6635371344177804,0.5883444435630303,0.20576739251060283,
 0.21128612119055834,0.9238141346637583,0.9314625910960792,0.5645901075131919,
 0.2738496351669645,0.2966083984972022,0.37430782013926456,
 0.36257491145391796,0.09516907506481986,0.9670792984854483,
 0.32397277007852765,0.026405758131563095,0.5494159526400103,
 0.5767066079689445,0.7773549895318514,0.29820655845735067,0.6943344633319009,
 0.30294617540655133,0.4781085766778842,0.7758310406886177,0.5465610740106899,
 0.5010634417696374,0.0972707035221545,0.054113095123037414,
 0.3944807127317175,0.00392101119924837,0.8519837892789485,
 0.04271830631830542,0.3311612967296663,0.3931989734090414,0.9614575849825585,
 0.09446207864418876,0.8618381483724753,0.7865387919399993,0.1485065086801216,
 0.46386437900076594,0.016159242718907896,0.5086560951271741,
 0.09892658122935871,0.42567063094682334,0.3196465959472574,
 0.04945257779105339,0.042752586071924115,0.08305123306031847,
 0.4905451380354705,0.899026287981282,0.9802470001226543,0.7286401960320223,
 0.12040976658628866,0.45528110864572935,0.3883915236087908,0.505494396532096,
 0.45508231219710027,0.04516336325169479,0.7921712798687579,
 0.6096320634533521,0.6280075009028199,0.5990224546437733,0.23725666641735388,
 0.11062375113915757,0.48147454561502623,0.22969457636693436,
 0.034786551092442664,0.1412326687198764,0.8256162070426414,
 0.5617884689077808,0.03453789744108127,0.7057861773932239,0.4688276627673803,
 0.636491929455299,0.2397910192748124,0.10484870712642758,0.3635468408739081,
 0.5092830044658088,0.5074394782718251,0.26901112739129396,
 0.39928410739053966,0.39263271234691144,0.12973109323181564,
 0.49809363268739937,0.2462677436412849,0.9196427536235524,
 0.12776882854659655,0.5747735864239195,0.19910439268815294,
 0.46941206593724416,0.05011695146768613,0.8517945064915984,0.437587040513847,
 0.793884964681853,0.054489360673883325,0.16255636696353126,
 0.7660491166231927,0.1126930280008619,0.27318518274453063,0.471937531324341,
 0.2758385355552868,0.34829473857679516,0.8344162567023256,
 0.14076890497155903,0.23490138930904814,0.923339039971445,0.9817210859457622,
 0.29314716273770225,0.9981993443061126,0.6990096564767109,0.3696175480688234,
 0.15532545586509394,0.5672574478445503,0.34500694839544377,
 0.8956983177385137,0.045951982807848624,0.4501453522361002,0.79443016427255,
 0.6883840668856607,0.9755144111009713,0.9357713836576175,0.34276103586053386,
 0.8419273613164094,0.27596233505280165,0.1915908566670108,0.4355833885660412,
 0.7615638278677093,0.3259554373651298,0.5203563622896781,0.4969924686778856,
 0.29292337539019886,0.208944166779646,7.079671687182376e-5,
 0.8683989575113902,0.9512630519442968,0.5250568384711757,0.8987698441158325,
 0.09760381181545352,0.535452075683094,0.914528245372494,0.4709504300519023,
 0.6875931837641365,0.6182444969412372,0.9587202912394066,0.0179028529488644,
 0.4635887375489567,0.15812744644238153,0.8843877642281945,0.6394561475439533,
 0.11641970458891615,0.13435941460305667,0.5270297104776205,
 0.16430967056474965,0.5326958618928301,0.5285624364014091,0.9547949579395042,
 0.5586875038600577,0.583243836839863,0.7343615899195863,0.8334395287851194,
 0.33546478487220477,0.5514651837428926,0.3603910155767176,
 0.07293108975810225,0.4707033061630339,0.043036535405574616,
 0.09698258880134125,0.3707393477331339,0.7595668498822762,
 0.18923528484405105,0.46344631288084237,0.012802589647947338,
 0.4454934130352616,0.6741183840586155,0.8699543221823646,0.5601545209002002,
 0.2674068980532779,0.9728116143438855,0.7312892185672966,0.5910919697475927,
 0.6140175009452347,0.32001885926422624,0.343377393672334,0.7716918245311358,
 0.44809966946614477,0.07638774047166741,0.001465038239283678,
 0.5013017440803555,0.7851763592587435,0.035139966432659486,
 0.9944422862861831,0.8954041934453016,0.7736039262596306,0.2359707805418294,
 0.5516188970151792,0.40568193969127697,0.7863033380803991,0.932345875213705,
 0.31000176420624825,0.985425660488533,0.3182469989941624,0.36227062936271925,
 0.997453728069352,0.21183782958263353,0.46737978233654254,
 0.03394252462779068,0.5809600277142245,0.582422076188507,0.2174049174792192,
 0.21233517547463554,0.5960058088539296,0.46456305584412305,
 0.6281340445635598,0.9625686996873488,0.29926391708798317,
 0.32826674115840015,0.27468178570152857,0.5163008479875723,
 0.9656264638360742,0.3616858820141218,0.7714169383781146,0.5331146387965993,
 0.940655647388737,0.46813763590230417,0.030491398030670136,
 0.34603226188940195,0.9318536163495337,0.7750381129735043,
 0.15812407567649367,0.1495083738018711,0.6883754469889867,
 0.09850442938189297,0.6697424608214905,0.3339821423507754,0.2226700561031918,
 0.22391331866679853,0.5086718780713094,0.43732303888935353,
 0.27764247410659826,0.0494893503774656,0.9334434585768465,0.7820698478760769,
 0.7744650389524212,0.5145476076061603,0.6782629646632234,0.23414384607256622,
 0.6975820943423788,0.29972672668778433,0.6925157997093705,
 0.35825443247282995,0.15757701401316615,0.29191269198134184,
 0.24520165728953758,0.7593996440992378,0.4504230434523031,0.6925585517473892,
 0.9481354709032687,0.2639814569730281,0.4283943128208757,0.45249396979118894,
 0.4161633018029789,0.8508263476102174,0.8047983352464758,0.700737758913643,
 0.7637412451607304,0.1806692794359389,0.026502757736943106,
 0.7087035794559099,0.8662887369624054,0.4908074643375625,0.24413047234619745,
 0.5017000228287202,0.0045570457705228495,0.8015493047541324,
 0.016818036822525806,0.4693945471237,0.605389794634579,0.4024593289593221,
 0.08788189026657589,0.2503153539588119,0.8385479844814496,
 0.09726110440372848,0.9732770023700097,0.7917673463064822,0.2971419250675278,
 0.5048413123543231,0.771210316675036,0.5893100360463287,0.14283629988656288,
 0.020412602691253556,0.6086607570066248,0.06782016780473177,
 0.9471948315169026,0.8471094626210909,0.34333126020383076,0.7137099692261516,
 0.6371145811564801,0.8768645194618642,0.25468011272159274,
 0.46270580633730973,0.4338070722402809,0.1584847700683112,0.7486196558956011,
 0.6852739840007838,0.7453297795716489,0.4677073267738081,0.6269806156781186,
 0.23474414538961197,0.470281427026529,0.1859263513296251,0.695241992544048,
 0.20597785621743747,0.08458942784778511,0.6676983664710034,
 0.14334621815222426,0.4261969257493321,0.37540467387146037,
 0.31582729803836984,0.4905914340635975,0.32758912159984344,
 0.004944605495747378,0.9134912938732944,0.2904425806865053,
 0.1861952777821172,0.40949840244372715,0.5070438676438871,0.472308173819425,
 0.5996316487072904,0.012770140006185593,0.8602508714718845,
 0.8854352591610926,0.0975589207375025,0.516958302440545,0.4018728605359251,
 0.22058604153797567,0.8977920886406854,0.6484740129819119,
 0.07586795432393201,0.42002400569689335,0.5133627350008397,
 0.8721069708460998,0.7156570713885659,0.7352383640207099,
 0.026593914150672582,0.4171186777533705,0.11182589969520573,
 0.0023682151592835377,0.3483566017295202,0.4984178105391218,
 0.6791610922695555,0.06683099432930195,0.03334698786257351,
 0.6436488065192955,0.7459881556533535,0.2090467534999143,0.7409286997595814,
 0.16741670606031256,0.33135299228881243,0.3929682363208835,
 0.7083232168304685,0.12194638030140936,0.15936985123091274,
 0.03804169240101063,0.9283428288036826,0.6911794239627066,0.6251604156342498,
 0.23552149269130385,0.9097979524365146,0.9060710329824905,0.1453515481617944,
 0.4586773571459404,0.7114597392194898,0.0780717567976742,0.45561105315563133,
 0.37386201038659006,0.8102262654501294,0.4057022947166299,0.8673586626636195,
 0.5840729276700911,0.6151961887746341,0.300640578103355,0.5435329442575433,
 0.7633057299340097,0.03692806351389788,0.44519931001895063,
 0.24455991910231167,0.7555497011612047,0.17736528168509025,
 0.9230075861737621,0.633429199439212,0.5513549075639117,0.5380197922169941,
 0.4733719954398805,0.03938573222246511,0.5069916476507155,0.9060663170944776,
 0.12142399658653136,0.4030491410441768,0.4534263609831447,0.7641073137716494,
 0.8992712860427104,0.44431544282360225,0.9696445007125708,0.1422217187477115,
 0.10133782301835503,0.0400274345295798,0.028769795981194335,
 0.3346280908536755,0.765944752812169,0.40902229201111884,0.8041952427134579,
 0.6161409574015825,0.4375540155798341,0.07098083838623659,0.8894039489733999,
 0.8523116255616986,0.2804143166061802,0.7748941141994725,0.15347764449197943,
 0.6769521112723037,0.1760796261896056,0.4373065099484603,0.496632884510088,
 0.3114194230287457,0.0767641555873263,0.2299374031268795,0.28111508686068276,
 0.7313800198205419,0.347098499833682,0.2935451464849409,0.7184115232088166,
 0.6710120316621021,0.4937331607722951,0.9203301917718576,0.6364141736504048,
 0.6846560786728906,0.14324719876968617,0.41398581965844694,
 0.6041256634761542,0.40476170838476033,0.16298287155891233,
 0.08220518285735745,0.12902964169042308,0.532935082820307,0.919916509322516,
 0.739855595409986,0.24159278810488294,0.11722560704468787,0.4828450199234685,
 0.976407393809153,0.7798407606773903,0.9060812161412932,0.16072003190598438,
 0.6542702600402815,0.026429398300994777,0.1016708968886676,
 0.6566997725871969,0.6523197617629064,0.673578659558897,0.6160839552011537,
 0.6129546911215182,0.22240958293191593,0.7692925517486414,
 0.051780196838074666,0.9562677498824801,0.7279509577125579,
 0.0817419283559081,0.9061246122094017,0.09075557200100659,0.7288555456391747,
 0.7478146203115793,0.9503690115666567,0.20645258212414908,
 0.08467602114004502,0.6328692298447133,0.921949145636613,0.10294521528327727,
 0.38600900227520085,0.720762240376567,0.2666793572869438,0.6281591362883755,
 0.7850541661928355,0.8364102392709828,0.8045188387086396,0.13696129508553678,
 0.5065743440496511,0.679275565325031,0.9797210128016859,0.7368899973726114,
 0.250016560585966,0.5974710270804003,0.24078601191136473,0.20931948881364337,
 0.3360684697843376,0.6938724717979932,0.3532404784446642,0.2415268264960173,
 0.9422101365692997,0.6832622300134774,0.6108328130250215,0.9513089122563136,
 0.8971976620332764,0.014831430874250318,0.047830762846887254,
 0.1870189177148054,0.002537451849372707,0.9775370758494141,
 0.5706673267265309,0.9007864962942919,0.36445061921553346,0.8497852015686205,
 0.7255896948101791,0.24473176684682874,0.7728116183288001,0.4471574928866564,
 0.5100045946987861,0.16503718461551986,0.32268034692723946,
 0.3224625040598129,0.5546990251558392,0.017782545045233022,
 0.5649077573018315,0.7004156992556684,0.20567038573653196,0.4969608238769785,
 0.08490386074607004,0.989192592715538,0.6434184466604622,0.8165015339658732,
 0.9741554408680144,0.8772367317182611,0.03954111265772364,
 0.40776730695728824,0.4285096316416038,0.36363143054339697,
 0.27073032617627235,0.9139883615238981,0.415693230590243,0.6305358625207331,
 0.7353397184177213,0.6044732693990778,0.001444027379791435,
 0.4780481051767048,0.3089558021589338,0.903681608567313,0.27597307180779906,
 0.508668149339376,0.5513815112267119,0.8761836875100522,0.6477919092772004,
 0.017007764216577792,0.5203398391823985,0.005449894944669076,
 0.7680526654390004,0.005735590464341356,0.6571360883114714,
 0.004215115961792804,0.25216866279154837,0.6825804818793716,
 0.1735217535417709,0.1026233288173457,0.8850079650809951,0.736085354863327,
 0.3862613421386828,0.48520295845092787,0.19444644852968307,
 0.3010297226922578,0.912694831359497,0.3756219261106596,0.7751302874357374,
 0.23838527216866923,0.7784012694756427,0.7997601668905553,0.8666240268526562,
 0.8182685715719263,0.8068511142186409,0.7153561026686187,0.5807147548949754,
 0.4206240955802204,0.7222675600841617,0.7332816339416368,0.07094416295552897,
 0.5390040206169391,0.32845860606304855,0.7107267965060087,0.0693615836313991,
 0.6325183921093469,0.9429879636145528,0.6069453122299739,0.6531123523091702,
 0.33782820856268003,0.2914215942983439,0.9227059023288141,0.7856755303591294,
 0.46390362631658544,0.7767569377601633,0.2317558856857893,0.1682323606099887,
 0.6058024350479027,0.3985455266623287,0.44575201231125827,0.8673331050403242,
 0.9066047539869508,0.5477045600610053,0.7583892279820227,0.32326962297399475,
 0.6593575246052831,0.3288294274772595,0.9276325239973756,0.9920009983014639,
 0.7990485677279614,0.8352342860073965,0.04967878738611631,0.3375019933326584,
 0.12098117562710387,0.911772640075864,0.01651038478620903,0.7887108603722375,
 0.21825908935732008,0.062364039704676855,0.7238275597708965,
 0.5969490653294255,0.1849608951927627,0.9887132164236502,0.7192018139611362,
 0.08923349187743779,0.544069548428892,0.5768051674099373,0.9264856426536238,
 0.3601477577104042,0.06580963885321556,0.5513047786098821,0.8810959363104394,
 0.33451588481231764,0.6620760133628371,0.1848656017543957,
 0.14060368763897724,0.32161127761452235,0.7905598983728823,
 0.6001146315511983,0.08100036438085023,0.7142638749186746,0.6089514474064649,
 0.13131139813377146,0.04480184773006746,0.06687016474127438,
 0.8063521603910455,0.1970280456575445,0.7445345157263432,0.6006514578588809,
 0.25954345427817493,0.281103550863424,0.9574157404657423,0.44307497316680844,
 0.9120512180655089,0.38413643286374466,0.584178064859141,0.21947706426160307,
 0.5602954590324654,0.5107990834595775,0.04649209913274133,0.9435121015126418,
 0.36160846775144706,0.6821019333310916,0.28157433165696577,
 0.1585281424804702,0.4636909272347183,0.7552504562671267,0.14172379035589122,
 0.15767475087327454,0.7643885885885212,0.2847257293175415,0.7605547956800369,
 0.7623247317323509,0.11235041645481325,0.25655717658560606,
 0.30855328899124523,0.004669609998404267,0.9781779788326853,
 0.46415513408113607,0.36236418364864365,0.001611609114311241,
 0.7059381288053506,0.5122258703096207,0.8821963992291504,0.13059273984112263,
 0.40464210317697424,0.41731372115542975,0.28633214372419313,
 0.40152965007868036,0.6803439806416375,0.872636168386623,
 0.034827094398352854,0.5140627604696342,0.4528093445100231,
 0.7283773457586429,0.5682615190143433,0.03289205091242264,
 0.04958472250252666,0.19366829623788806,0.060151295984592235,
 0.47119938701095343,0.2433459135596585,0.9024210222988236,0.3232216938888317,
 0.6559012282686569,0.31523783798670013,0.8665984528071514,
 0.32262482758725264,0.3877550374552404,0.646022461953773,0.19723448270368138,
 0.7163206546164773,0.5022679386296195,0.12701492127687408,
 0.25709310095634286,0.5489034915067723,0.7688653569282476,0.6331564308488793,
 0.4940187767389771,0.2570992338119349,0.8229025638815579,0.986289769571983,
 0.16653092516748114,0.5632714875261334,0.48032952219564407,0.392523693386577,
 0.0012345217907534511,0.4734282106456471,0.3099227996830446,
 0.8141291271493101,0.4471447657755687,0.15239706506950268,
 0.46035564846415367,0.8486787137061327,0.012063682168695467,
 0.49655999277696417,0.5132015056296141,0.7680786353199096,0.9786539831580708,
 0.6211183600682831,0.3551500806081892,0.09091145913402243,
 0.42672789673899325,0.8104950478765108,0.9149327894360861,0.7773452348886524,
 0.583369758276814,0.8040788765528161,0.5228229020059608,0.9167341007710202,
 0.2050767275703359,0.8186038030545706,0.1226658839336805,0.9503307774118026,
 0.2879517187051802,0.6509792997448464,0.7456359740728506,0.6668972984674598,
 0.7953102286005072,0.3694432578859257,0.978575464558251,0.8916868766921421,
 0.6383766931328205,0.8201973764138545,0.17126464377844886,0.5654672166262198,
 0.7191813869557917,0.2864677135651561,0.2943195130403228,0.2973374170744876,
 0.14947574029386446,0.5981062212130092,0.2716660799052667,
 0.29022542546741315,0.7342217884184352,0.5709697052661113,0.6365928424810734,
 0.9497536554830908,0.3099523942910565,0.5290873252903493,0.8660959816834014,
 0.35868685099014896,0.24795626829044493,0.9342730706350378,
 0.1752348646555434,0.20823449516075432,0.010517887230889311,
 0.5631244089331872,0.7996223616591736,0.054691194703892165,
 0.8693461509334619,0.9549139583620707,0.3140134895701243,0.855681327632782,
 0.6066398968954129,0.5315429690284462,0.8210975573929793,0.20252337852547686,
 0.38060610419893837,0.7410805305467363,0.2667713352602906,0.0398118613008317,
 0.5695097058018577,0.84603746554877,0.6515526895336697,0.378456704450705,
 0.01903384760524529,0.8756734135241298,0.9141760992678798,0.8483362756338675,
 0.8483194756701704,0.9296471734968836,3.3294093539637526e-4,
 0.09724177916091192,0.92519007355883,0.08664546566372366,0.912740343798364,
 0.7987965798297719,0.5213609176629931,0.92915093699086,0.6274850852988574,
 0.7937763771687492,0.9482670780382435,0.9132390100293625,0.772835641501404,
 0.21961143157554663,0.6528338389146211,0.8573062051716048,0.9723426364194974,
 0.9847317279014217,0.783643468349572,0.060968231839315656,0.9128788623089481,
 0.9223993232993398,0.6184863378439736,0.17064118410445217,0.4884123924687449,
 0.4816268530202218,0.36922041267432226,0.8181880077948022,
 0.24768928543986535,0.985274675657425,0.27392392057885506,0.6377795268181817,
 0.1731996975209894,0.598666413635885,0.4049677218610874,0.24675658861726024,
 0.3133157809602979,0.01943254442054143,0.9491276772600646,
 0.42114126011228015,0.7791532892987609,0.07254337983466475,
 0.46918629239895915,0.4865253314704929,0.1181292331229643,0.5154788604353849,
 0.8803955555553628,0.9972876122682549,0.30945309061807047,0.824956134032804,
 0.41491574794945674,0.5964299852587622,0.31262146389662837,
 0.6686774500265863,0.9235752745605749,0.6981523821739725,0.29943148233541705,
 0.6908228605966795,0.2815669459484825,0.4910733644282761,0.9951649578860675,
 0.5526696416107966,0.3408858530450397,0.3871624025828605,0.4951086378223819,
 0.5411786910513252,0.23414708298205067,0.2530981886019986,0.814719548162895,
 0.2823171079249428,0.9806978148391884,0.5187044790635653,0.5706766986783769,
 0.7283487202284606,0.6147984411781783,0.9648985325911975,0.9615859888105843,
 0.035101784151761684,0.30772620123291905,0.3685688493197714,
 0.22636072010302222,0.9942366648718682,0.6842673162098143,
 0.43873415501682267,0.4114041285192328,0.7396136845722783,
 0.052817722402180745,0.9574032270273815,0.02772483730421993,
 0.08613881476254537,0.9511767335993587,0.46233067661547167,
 0.22720781446947846,0.18703871601264188,0.423328284734668,
 0.029480950160127972,0.6256263748364033,0.6784688393741062,
 0.11194208775105663,0.0711556571867894,0.16905149279840548,0.67535822760073,
 0.19354992912905356,0.4936446477978812,0.3181303986489272,
 0.03523076553302962,0.7810568931756883,0.9283437964924036,
 0.004290347620016011,0.5645435313491951,0.9003718007479801,
 0.4316941682244848,0.5993769927172911,0.6697805124353414,0.5038041728240782,
 0.7391754328390637,0.6582143549405582,0.2139773918073502,0.8488226150993632,
 0.8458734712292916,0.7560178907911071,0.46308602809919075,0.0436208297152324,
 0.5322452044730073,0.31525900225084635,0.7027788648229023,0.4594724366485722,
 0.9407548955506037,0.7250225557996769,0.5336294731251261,0.621701106593804,
 0.9308566930425436,0.45873068156097074,0.9123226189326075,
 0.26663598919365983,0.14581370414195682,0.6591253213359975,
 0.09786801309739401,0.5308132313801318,0.28743030097723576,
 0.6455429182906302,0.7171468815462207,0.811915105195955,0.2828873291751145,
 0.9247907187900393,0.8722698266319815,0.2552918660251048,0.262520431442537,
 0.24531295329922687,0.680086065542304,0.20234811836583333,0.1969581285534191,
 0.5287295039500388,0.5463942276262933,0.26972478600859895,
 0.29031644080876884,0.7543116582382183,0.16269532599696412,
 0.4960897763233407,0.5753530171997565,0.6933195041827607,0.29453805669485644,
 0.9086607971465938,0.6282530427973516,0.8806294282422613,0.49726755869255823,
 0.9445967803304365,0.7948166915527146,0.6976990932087294,0.8760752784313502,
 0.12592336956981898,0.8229367803446285,0.8916677291569782,0.3525484963156438,
 0.39262975833686664,0.8199261289217843,0.31709153596160156,
 0.9771513532758048,0.4398807298660681,0.12882405942316466,0.5636958225488578,
 0.47771111938769506,0.38494689616991407,0.1850094677763593,
 0.7972826299446196,0.5290132032545335,0.25515844839782986,
 0.18182998721471788,0.2992884381192773,0.9660875975355331,0.6892731381999231,
 0.24598418722810855,0.2536070829257834,0.014938639961731948,
 0.6889055275108814,0.9410082636091689,0.14840945179032017,
 0.20984646635124782,0.5470458940133184,0.3794242287196875,0.3093536788833251,
 0.7620915590978303,0.7627863392985524,0.8109659690713876,0.5080004656980011,
 0.07855033276958756,0.011979043440612958,0.1474335273463463,
 0.13087100040441146,0.7799612820532489,0.9334455612037929,
 0.08617999538374244,0.45592344437695853,0.921589247438581,
 0.30436597161470647,0.2626821904714325,0.09351097248771179,
 0.28626627988423403,0.9042537255105227,0.39984578120501935,
 0.33775660366684235,0.012685986907694069,0.349105556114502,
 0.4272899972860279,0.49114251362712824,0.48815935392098775,
 0.6351098931356285,0.9897487648370449,0.4531441764052513,0.8014603987295986,
 0.15024674091034407,0.926271920229783,0.856674646618494,0.13350515603794177,
 0.7480406182545073,0.14861841749518212,0.8736349121573815,0.5902945965526999,
 0.031195041965130454,0.6826533866718278,0.7142994794797447,
 0.6733806612318438,0.29696879510925656,0.1330846309061624,0.508714157634171,
 0.8637802034928959,0.6409813525392644,0.40740689396177365,
 0.36528055816908667,0.7364850742427029,0.9670244125034947,0.6526145969600119,
 0.545815654869401,0.039813866349746485,0.7892403791208084,
 0.12117853974400727,0.7574648250709081,0.9555277028175225,
 0.06123686001581874,0.22462468863261176,0.15806995297845505,
 0.8434491188273588,0.05391772028249875,0.7141498708111552,0.5465056326847231,
 0.3690814168126709,0.9089894708953272,0.9654871631701006,0.9712133187485978,
 0.7985085704437891,0.4568549789298749,0.7591836347579707,0.7301847196974249,
 0.6763206846492885,0.3111888708161621,0.2637256105288176,0.6786766518666381,
 0.42855275707778673,0.6494019648189757,0.991974203822365,0.5344188574908071,
 0.7993927283902701,0.2655252567153039,0.6866557624906473,0.8762493665104812,
 0.4182924504102543,0.1337325389691697,0.07631085453317399,0.7124347687082224,
 0.7271714998959289,0.8073303558700455,0.24248424724123863,0.4246732608388917,
 0.41658251005283353,0.4446925025565097,0.4964007306818241,0.7328892592403686,
 0.9584647764081495,0.5927629465142983,0.017566789820268047,
 0.13606779495988486,0.7676278880346642,0.6178966514064871,0.9060822365404149,
 0.7687234976043722,0.38926925209038177,0.9338002466258142,0.3281340035116658,
 0.36091544247959495,0.49223584955233923,0.5773062231806152,
 0.8813743898323114,0.5553388512893419,0.8368459244487445,0.6423417604394387,
 0.20191301365073622,0.08587488487236572,0.23292176985019908,
 0.8819354207672818,0.09394916431854361,0.2906545747250209,0.6404367531867188,
 0.3291457898525765,0.24672942207244675,0.8674074299033451,
 0.17745895729920214,0.3864065072734695,0.6325673561540363,0.6592693822233898,
 0.33745414002955365,0.9284874572758329,0.5967595334124395,0.8360812793848896,
 0.6323829046858332,0.5498468803725313,0.9869454587451811,0.998188745831531,
 0.5711674767446483,0.5759631113447856,0.27102810183618553,0.1482756448148982,
 0.4190486663904659,0.2930914836806955,0.8489243416686282,0.638830708499837,
 0.7089198827271873,0.09679991475144134,0.3556268136105766,
 0.20385321742990514,0.5437765137125122,0.7932008216656683,
 0.19570283113767772,0.43322896369962005,0.5582651909603094,
 0.8906588715636068,0.9980728612998442,0.43608432760769844,
 0.18288916894571616,0.7967654015148689,0.49430952722724186,
 0.9491874843324785,0.024628624708397773,0.1122813058992711,
 0.32448599860946026,0.7339332319080326,0.658900879649444,0.5524872981585605,
 0.39464394097306066,0.31994822339267326,0.3094966349887851,0.165081438589206,
 0.8491825946578787,0.3621044817872925,0.7559787404337229,0.7251098846940003,
 0.9115626714615547,0.9889217470717403,0.9416032305671911,0.17922131294057753,
 0.11506314984449884,0.6005314766248799,0.2986644050167766,0.4441580897693377,
 0.6489275851456136,0.6035723807868594,0.3441268809766683,0.2726651420057833,
 0.9151792564239393,0.8253211646292213,0.008729227950729057,0.594907642987179,
 0.4891842245009218,0.7963976983483461,0.3484397918266121,0.5391021136419343,
 0.525223662129338,0.8346729777140497,0.7392132876650671,0.5507102344147436,
 0.9481442129878791,0.07154460996316803,0.4076646633698662,0.9007902419292891,
 0.7113694482396697,0.9365889959301515,0.6248223914092944,0.13274300537493144,
 0.9820936206671974,0.6546149677847118,0.8402512198461503,0.20672965629493978,
 0.931065015121426,0.442082098291969,0.40614799595053175,0.8099398945952644,
 0.44326609577770393,0.14776063734947087,0.5318884636006026,
 0.9114865918126396,0.5966429097870554,0.6168164394660826,0.22390571432902706,
 0.7663125270899325,0.7979409283998864,0.6262394012010865,0.04233608726273452,
 0.2605512239269785,0.4880475792092367,0.7966597809189462,0.24600701202530517,
 0.884961201291192,0.2762041196687408,0.5285668854139962,0.49512637832054485,
 0.8521855809375106,0.4632399252466075,0.42907307543611006,0.8794718480910478,
 0.03125096123984539,0.15753102204144565,0.3210409337875123,
 0.37703973686279313,0.2032751052296482,0.02607343731236078,
 0.5037312972690691,0.45449434662407007,0.5097077544417355,
 0.028622726592472114,0.27819282867941675,0.21074664970997148,
 0.7333367717108986,0.6415548909748442,0.8509161289642044,0.05432981186554198,
 0.6947674826980634,0.16216585444024922,0.942342486957971,0.4384076577276943,
 0.22443516566114496,0.91341727614145,0.24386401782113565,0.3859297818865305,
 0.28068205479125985,0.8970389848208287,0.9398271030091906,
 0.48330320917497316,0.7026596665924426,0.2773167097214736,0.6598897360441908,
 0.22217921660131346,0.7970702267292102,0.7326499295951602,0.79988291002042,
 0.9453822270560835,0.2817869522984775,0.4244502706096115,0.7609495193378052,
 0.7626406938456111,0.8691128511195793,0.6719706506161949,0.9972765760467593,
 0.14323271532200166,0.31922687765738056,0.7915126576236613,
 0.05749113429914798,0.5313870041541692,0.3186586102891322,0.665011059977946,
 0.8169566641552097,0.8054118402171866,0.3464557632858076,0.4922326514194303,
 0.10999265011016002,0.8736694570568715,0.36200961250245967,
 0.2524943915978858,0.25183532916448415,0.7478788554906715,0.3012870645089023,
 0.5701503010769149,0.769487991982842,0.15291905789042792,0.01258633149322741,
 0.6851231349723015,0.8495338028786562,0.7303193110596883,0.9704350575631684,
 0.33196334248433135,0.103427950056589,0.7315083221384213,0.45524822099015194,
 0.8283864551891396,0.3331830673125995,0.7788856185034103,0.3900125766646868,
 0.8004787490008172,0.04994650843462656,0.21267784644523646,
 0.43630568932304037,0.28267016602294315,0.5231778559384445,
 0.08767277468585966,0.41861384313164096,0.9848297787186944,
 0.9930772223965729,0.8626226571830911,0.2639371887600783,0.24888043183727582,
 0.44269662142904076,0.3544684629022504,0.45253189896828294,
 0.33587060062209484,0.5558876790178711,0.24029298449409808,0.693915969227584,
 0.47970509767892044,0.20016833789662836,0.6286676698243572,
 0.7296745696307743,0.8583635234638718,0.4948166666756417,0.8472696230935837,
 0.322295767539603,0.22244752484504982,0.04393835196462137,0.3576533346765446,
 0.5127213675458027,0.28724527096915886,0.9535296806097757,0.9977931725349645,
 0.768324425135362,0.6358850625866072,0.4870985316838945,0.6123059153088669,
 0.3360813275425607,0.9002472530166102,0.26399742250536684,0.8123162798802901,
 0.8648067996828654,0.28803409089070375,0.20764000944812466,
 0.7565979243975054,0.9022636422095093,0.6860383637850408,0.03333562728390593,
 0.7722730726815382,0.1341449300760854,0.8552132345140476,0.30658858533008315,
 0.6089082876504153,0.2541022131328511,0.5897764888597294,0.2355269035870149,
 0.19652811945264437,0.9247527199337072,0.6621126765962275,0.8390111222846518,
 0.21386895073739765,0.8763889429083418,0.506783263987469,0.15938617673568445,
 0.5355788896050405,0.36918522792540354,0.6041568590914741,
 0.26051336452922125,0.9758803109556511,0.811452030302621,0.703603143542114,
 0.7046408568948165,0.9965837429994852,0.018694023396863102,
 0.7694190810164714,0.1650161508364203,0.05089777863533007,0.0219794096459629,
 0.0600992819944296,0.9703398101107348,0.977739549778946,0.7014764287389451,
 0.5056580721227076,0.8447709455227009,0.2816073899611461,0.504088665119015,
 0.05738051807751332,0.37061153700157445,0.4859102216779674,
 0.5564806223820802,0.973569142196943,0.8814981132828895,0.33412995450783034,
 0.3252074024591357,0.9256934605534756,0.8791168762518696,0.3095528202805069,
 0.13918257326288408,0.5416193499798222,0.6079339865174234,0.2138432947062402,
 0.6969883220291601,0.11216896566882328,0.8424786721880068,0.9459851404131342,
 0.6211451584111666,0.8561058890964974,0.889268931658814,0.11046750476518019,
 0.3667236651024086,0.6802607033575024,0.8974074525922128,0.40950321644232013,
 0.3355628075979481,0.3603367698134238,0.7311501817260283,0.9699399904744241,
 0.5655560083007867,0.4333295623138018,0.23609332033915698,
 0.14435716857962044,0.46455026494536866,0.19606837484843054,
 0.32640520205580326,0.6241823295151026,0.05984687285080603,
 0.8280190174547452,0.3736976385026658,0.23252330726170378,0.5384232319819067,
 0.29126467273832135,0.24271477495989857,0.7586921891326102,
 0.19445097847385395,0.4832550766299417,0.8039736626147174,0.6999146527238695,
 0.6851697376821354,0.45078668442957426,0.9007319418099184,0.3867544721088011,
 0.7389459764310493,0.6714661710714801,0.023629937536772783,0.62342954628201,
 0.5214076318378968,0.22385212276337274,0.48434140444179885,
 0.10330237046905599,0.31630090387881793,0.8226583709379836,
 0.2704711475362864,0.4493584120345441,0.5646099234903614,0.26885304816673095,
 0.023975516269168495,0.14287499372607992,0.07579535824323025,
 0.9354865063770713,0.7395795854625528,0.1206658354646114,0.04490659169401867,
 0.7354778770260805,0.41008228463184193,0.7010873571227689,0.386341734468741,
 0.13763622245196172,0.7636896448940467,0.7938131730237984,
 0.45262534890858286,0.04184153222913023,0.5895482831638513,
 0.8232297381481168,0.3038820397221559,0.7039439252342511,0.5440670423086367,
 0.6750642678156015,0.3367807902817255,0.5881591280658932,0.7184555260723542,
 0.8213942766394313,0.5069107652477866,0.6455585244301058,0.5999484335260775,
 0.10505517188865277,0.49487280336754924,0.6001288105482934,
 0.6740324596546058,0.14997661006685858,0.26757513822969925,
 0.9709729675926625,0.9102347269015485,0.8248357771483708,
 0.0038369053133900932,0.09320871821960697,0.349631573776779,
 0.35654779942058723,0.20069642925079312,0.4291966699032115,
 0.05796972262973499,0.38551516188433277,0.45580566443015724,
 0.25795968158445826,0.8393861439977733,0.40618055930695296,
 0.6116921771729507,0.5669435777047118,0.2664150080812777,0.5923807897365507,
 0.6155195817976842,0.800518673799253,0.14440043943488834,0.7223837582843963,
 0.4198932620053635,0.20004212887802042,0.2690111176706983,0.6589345750040648,
 0.05274530054903592,0.20821076004311112,0.939896689700289,0.2715200646680702,
 0.8630521791195975,0.8644985412427681,0.04045911791238643,0.8708974067161273,
 0.8177156511335666,0.050294980761053765,0.803993681421337,0.5155193786374808,
 0.08112027649124465,0.2374953453958617,0.847457195955463,0.8185011026912716,
 0.7214942401619937,0.7694286191392422,0.08308772307753531,
 0.47209265728364924,0.4973287248923538,0.3127622172386305,0.9721544527224214,
 0.787173101153458,0.8362614827128839,0.42765027155345603,0.8674171610144354,
 0.588604402191351,0.49333219231301195,0.12277677527814124,
 0.49526612477433885,0.10916151652548667,0.5135324281278749,
 0.6083173587932891,0.9578398306749054,0.5028962048600227,0.06976919090325984,
 0.34818950687511585,5.866383346347925e-4,0.5837415668184794,
 0.19809429847018678,0.06232163406876956,0.6707694237852293,
 0.6564005480822019,0.02297658655082513,0.9140039365566013,
 0.07918078684378727,0.9240889048327722,0.1312987365489745,
 0.18898255031492694,0.2531386945786873,0.6846368852536071,0.5197356623932842,
 0.06914006341048773,0.44188343670967856,0.6625037092968786,
 0.8317167697382885,0.7523338127337186,0.7962787471113877,0.4738716773082784,
 0.3006653194692548,0.16181344045124024,0.34997722731757863,
 0.059970891268712245,0.4601452399573154,0.9347914972912328,
 0.4590607841134893,0.274277349016117,0.5978499289434529,0.4587133918692001,
 0.9972962553974739,0.4041908408267696,0.987805872919004,0.9823526098732059,
 0.11675372792962602,0.8728054238848258,0.970685319198669,0.3620660757007367,
 0.2801742026373978,0.7589012095805256,0.7396606794398346,0.6633299097249337,
 0.6496547894398113,0.36202647998735893,0.8130874976539615,0.2839724091183543,
 0.8817894343683977,0.4959785923739649,0.052795808788517085,
 0.5516721903494703,0.9282381085127636,0.913177822735047,0.7052861855347844,
 0.0682798169839216,0.024931093213194866,0.9295791835172946,
 0.7060397014543941,0.38742054506914547,0.7573461992697293,
 0.31227535186881705,0.468879258334037,0.8946662978973637,0.9173463047497634,
 0.7426590531061276,0.997798050439348,0.4544809137194271,0.29037856676771623,
 0.769115752631435,0.6657325850184902,0.0789691732847766,0.05337539712889683,
 0.525792898931992,0.43722903052436013,0.6493195232502431,0.8104111074638177,
 0.8130769199586043,0.9766598472611671,0.12023985832679085,0.9781605598207243,
 0.8199853319054616,0.8535295820144768,0.9020024825479429,0.1360926855151151,
 0.4329264447401291,0.10627686338503983,0.8827884710809144,
 0.04198397242054486,0.6751617203664324,0.27210342869247073,
 0.35196405306384715,0.5356546396084882,0.5418676909510174,0.6233601661198159,
 0.7361970994392562,0.005263292196323044,0.07895900910821407,
 0.1348039857533132,0.9007699145815107,0.3645700468631936,0.24109860291626906,
 0.061260731368530386,0.7329080252712576,0.6931334220831155,
 0.0927257883999103,0.5206679932040916,0.44682743473778397,0.9145396131096153,
 0.7716364432944991,0.046165886984408266,0.5783614438457853,
 0.6091857543762242,0.7386319669465469,0.025847855419558696,
 0.026310718223813923,0.42088774755458924,0.05981017971659863,
 0.7720225244563893,0.9490168599997875,0.19770498938293557,0.6530610951044871,
 0.5926500836839297,0.40148308091221296,0.4699393090459216,0.9054711821528589,
 0.3928578666093241,0.429214774642775,0.7149358311231002,0.4097671925832619,
 0.49762442019900543,0.7877473542321374,0.10929482974196936,
 0.36529333681562326,0.3153489002071871,0.06422857280409555,
 0.3312896206360877,0.021318064551127835,0.2160423114675214,
 0.9548627967178539,0.3964670686824828,0.13506013170914077,0.9012295623685996,
 0.30334290519001983,0.8303417126359628,0.20018328293703946,0.810329657569119,
 0.4383638301930016,0.14184906627264005,0.4192995390204943,
 0.09563184295580618,0.9789734559568665,0.24078608488759012,
 0.9159093715149027,0.8728667062473219,0.6532243945533653,0.13087302764340414,
 0.16550137600681225,0.31129965676921567,0.6840381456482616,
 0.28506486789942675,0.04286723604655607,0.8955660357608883,
 0.6788051717793423,0.8106426280624066,0.5619606272717518,0.9784142277380548,
 0.058795751525348994,0.9408777615824099,0.4936607335152331,
 0.1720502914230273,0.6069848319097231,0.33592975041346784,0.5526011441560741,
 0.6856061629794089,0.6630589309948965,0.2645733532671193,0.39165701898291105,
 0.8609403785923965,0.031509486761931105,0.588912876523107,
 0.03749931784812133,0.6093054910722915,0.8858212721842259,0.6004123718584702,
 0.253618137046155,0.9842615703504299,0.35848180473321556,0.2548403466925836,
 0.6940318710248923,0.08635772906740602,0.7340702836876308,0.6824038572276935,
 0.270319238503809,0.07408172462244877,0.005913031598500984,
 0.8753823371994097,0.4561097560253353,0.31834666524285804,0.6607914408963983,
 0.5461257918462277,0.6428830656202611,0.4936823177084697,0.24006651922948585,
 0.05011433921121011,0.6960006713491994,0.9312246061600494,
 0.11755935000229667,0.20503822726119458,0.25670901270452806,
 0.06359769497190282,0.8012318825663712,0.5402604890763479,0.6844222107154639,
 0.730110582786716,0.0061019089475196076,0.0444512150034756,
 0.9784728270937441,0.661577012968574,0.9993785461229233,0.6897293109295441,
 0.7157790853664493,0.34039529926330436,0.7404339548247513,
 0.011333524564495656,0.8249352311365028,0.08626550990931325,
 0.5688255760913606,0.35473482908969023,0.8059525422056844,0.7113978997542239,
 0.35717470869367784,0.9410886403426598,0.47005632771436856,
 0.7954193846616974,0.36134379434721886,0.6807188695582294,
 0.028671048645473762,0.20631856886737177,0.8297403022630866,
 0.8047514659248244,0.9742383858442472,0.8575891677976641,0.43631215846133464,
 0.23404106529471758,0.7081794612198219,0.890844080664211,0.4438959525111641,
 0.7953682328140786,0.9185988440673837,0.1856254046407131,0.7065325313642755,
 0.12919383173675592,0.49614819119920917,0.5290616053869194,
 0.27417745796267645,0.51348974492584,0.1450834205181799,0.04307232364595648,
 0.5114173523370383,0.8117379241441677,0.7865762944486671,0.9163218477391738,
 0.8813166726504446,0.9011441783288092,0.20397753495803683,
 1.6954972897487153e-4,0.49615329562875965,0.735366442507148,
 0.31562215391474935,0.28259674222178977,0.07645899814454582,
 0.011085147023199848,0.678629949294112,0.09120831288764342,
 0.5482397113685941,0.6176503562925989,0.6489544326894772,0.08423579409837312,
 0.39807390362903217,0.39488670721007324,0.4685016990544667,
 0.4750249174322436,0.43224475966728115,0.8722606298729961,0.8740811812535564,
 0.5637391175956977,0.3501396237167036,0.7026666053351021,0.19140318866329586,
 0.18293286266369724,0.04509563790838833,0.9923579776121898,
 0.8989895056978052,0.24479086992936583,0.5931210814509811,0.8866438810412224,
 0.8286453736119285,0.7278336741610318,0.5382673212696553,0.2887410773953458,
 0.5774883665020222,0.17925212548077918,0.5989781941135475,0.6856812582711636,
 0.39210630082368636,0.8053232433804375,0.15217848954859114,0.744478601323262,
 0.42881651883265004,0.2961843341857161,0.042788803023839805,
 0.13541484846580187,0.671373308061308,0.9147151139089001,0.2734784605436895,
 0.6071456449715497,0.38528531450460424,0.17411960092119882,
 0.5496951246668642,0.8035669990094648,0.5460606303244706,0.27576745724048235,
 0.9792971386288388,0.8340152253391694,0.870675206632365,0.7872424705262946,
 0.5402454432170056,0.020168495490043625,0.33124850886787305,
 0.7362475881229059,0.3124993913865842,0.7736270396507053,0.9783358732756541,
 0.14171492763167215,0.3983686881447308,0.660303726471094,0.5574308733861384,
 0.906329495745452,0.316716949852502,0.4343287536345921,0.7981917136499684,
 0.8517001240737425,0.6188888250831431,0.3021545497085134,0.8536619914003872,
 0.9822787766297498,0.59184255789365,0.5854308220222277,0.05681928582113771,
 0.1342693913481917,0.34740433816978944,0.8925654388831975,
 0.45854013597532284,0.28226722366030255,0.9106660849413313,
 0.015606364077544677,0.19325924599011235,0.0680912923410193,
 0.12612264648936755,0.9511158999964148,0.9429110555064939,0.3762115957508956,
 0.3842115399612226,0.36939093390515365,0.75062284649061,0.2709281170271216,
 0.6866943076165145,0.21378153902872488,0.13367777202476705,0.410940578874203,
 0.5694191436165498,0.3873262761528604,0.22799788403577104,0.8542694033694583,
 0.9726496812622606,0.6727435279595851,0.571635226998951,0.600855694130844,
 0.195617042404457,0.7246466174865424,0.7887404105994122,0.44544749824378405,
 0.19772107032029895,0.7335566092594058,0.833720739109426,0.09932474829550764,
 0.6751668293859461,0.7165576891743969,0.46020626871293946,0.3660008452105923,
 0.2715557719349354,0.5801060649058639,0.32972893419375027,0.9568792167610567,
 0.8452324695654988,0.584521456666803,0.9319432702641085,0.5542071627129164,
 0.2981024142807712,0.3623463123586351,0.03908409990521067,0.3320498725753197,
 0.3983955478885576,0.48226859760976193,0.841339878305271,0.35022428601018984,
 0.962038823933779,0.4638534587257296,0.45468402614740067,0.2949513780200068,
 0.5371794601193546,0.5701766302451927,0.796654950963526,0.14166720288613333,
 0.2088519821090291,0.6757480186444164,0.26957330387298606,0.9818267298624883,
 0.8512425581388661,0.5153141304177418,0.6656863271997404,0.6160341665200819,
 0.927367461414144,0.22452167853734217,0.9038807930381249,0.0891413837363515,
 0.5059193741626334,0.38057205511422754,0.8030150729629827,0.7303000192042601,
 0.8183130606719384,0.44773019416502313,0.6853816806896224,0.5335797428652225,
 0.5693088279709914,0.5268331482540575,0.8904827702012074,0.8729732813564068,
 0.05067656453931546,0.6965414460174596,0.8814522376317635,0.6469783289483559,
 0.7897186217060396,0.680201702678464,0.8935305098234879,0.8103214347673475,
 0.46208990502614833,0.25532694796450117,0.8336724529837499,
 0.6090946672032292,0.1141577568339176,0.5693232726529006,0.469694108193657,
 0.7544023084518392,0.5119511062507536,0.6898176021088744,0.220833014864244,
 0.19738580693416652,0.9471384987132774,0.16178857424618331,
 0.7019803442762454,0.6817759316256895,0.4405778132863649,0.8527980078899614,
 0.5749516804231144,0.2692158670360457,0.7994971582670147,0.22854159534275276,
 0.3592353920898693,0.8341581042678998,0.9412716121641848,0.6289065167384241,
 0.5935602254066092,0.4696957942944552,0.18513750170819732,
 0.004503980968326826,0.4905150035218916,0.06824143548250428,
 0.33867601417561555,0.11167186942193497,0.053844303086610146,
 0.5322283570247197,0.8432837406034843,0.5950297804392365,0.7360346638811093,
 0.4394013262188482,0.9887541937137067,0.02118832996597919,0.3469014249854059,
 0.8384469213699386,0.35698284690404536,0.6031783289092886,0.7024506785289262,
 0.20689685577314487,0.8763398309184202,0.08021062868009488,
 0.5223292279502559,0.8492354671603902,0.4980651864761453,0.5290349677352598,
 0.24726071644333114,0.9956331234293283,0.460058964499108,0.16568587930119372,
 0.7624385794635178,0.21653875471197204,0.2526688732555582,0.9694622099562107,
 0.35023540414678533,0.7494752601380029,0.726675471370254,0.8550530557314817,
 0.2225067925691363,0.23061217340651607,0.6197352847339509,
 0.08918003576420352,0.10140618062821327,0.17117975679168573,
 2.607611815226285e-4,0.4666309283825545,0.7818043151267658,
 0.23415923098325908,0.12351250141238379,0.4005529911561013,
 0.08456995968458403,0.2531029708399183,0.8986904140986108,0.3702523605776915,
 0.08441715424023066,0.5858822467369575,0.8735637913434625,0.8638516435046648,
 0.11800387997376127,0.9754718159008481,0.6926131222797108,0.4394605672653996,
 0.7311538168968831,0.6212180620574315,0.8429882220167388,0.3977370659156829,
 0.37221239027075537,0.09451247281690833,0.45379140365019077,
 0.22161490651868376,0.9275237340664741,0.3267297045807307,
 0.07345591331330947,0.8026624063431126,0.3577303051634031,0.468378752227362,
 0.9371840154812958,0.11823172363054901,0.22087359639016957,
 0.4337711703451277,0.40156597073454325,0.7184192671246918,
 0.19216516637494263,0.3373235279240344,0.8484696092141615,0.8783951110504861,
 0.24862980499688647,0.394673951106812,0.3940416841737616,0.630480870090846,
 0.17436895244392414,0.5495841288924571,0.6726571169033625,0.8634498482822349,
 0.8035196255964616,0.20493083937239365,0.5110222563212442,
 0.28882927373016065,0.626061992829733,0.7273034221542205,0.508334360230677,
 0.07588209540860091,0.5014577998138223,0.11775644487517112,
 0.8472465943481988,0.9890704822815026,0.8246043369229382,0.128254100751668,
 0.9739566694735382,0.07327425040624647,0.8850155917610203,0.2180202778484691,
 0.008054311374880885,0.4370107745573951,0.4627042166252685,
 0.042304232996055346,0.5841053266853333,0.16473970389555126,0.79036436726811,
 0.711755454545944,0.5089043552001302,0.6367610992748157,0.7719045734947465,
 0.08773995408668189,0.6419427850869156,0.3566900341003024,0.5193552653667787,
 0.4646182661433519,0.402341630918551,0.31282119880722625,0.0643453911970423,
 0.06870412611611154,0.46641261902882647,0.9537477720855708,
 0.03826442920772033,0.07339023316983084,0.27247970661458176,
 0.6674354119103914,0.3350078660678961,0.7503232662015121,0.651235610962524,
 0.9699827000198575,0.2405802282645657,0.5754024835300381,0.5889235451030735,
 0.476499901869185,0.4413174380697048,0.11815339006071279,0.5726522638675823,
 0.044038237135948344,0.6121043792426117,0.6344800553379224,
 0.8952219085712854,0.07338111795453339,0.9368317219049861,0.6865560976673257,
 0.9516080931979374,0.42855645801267384,0.04722733231621279,
 0.24606848753835853,0.54537118061883,0.617746973457602,0.05934707183518184,
 0.7204319509286439,0.6890544736897748,0.22385642165997366,0.3865858820643808,
 0.8865184195011325,0.29849048374250187,0.8893745168161432,0.8294524060989684,
 0.7055826748600967,0.6939344982364843,0.8084785536379634,0.41136361864326887,
 0.1607005814557968,0.25146852792703245,0.44905381330220884,
 0.4069899049420409,0.8515159723467748,0.32913027318565224,0.6028472577962822,
 0.6114413633848101,0.9549718447366732,0.8584121008526533,0.433746921492997,
 0.23681853446878698,0.9243813923608633,0.4282930645026919,
 0.32458768590857345,0.903091070454407,0.7148800990305968,0.47255268972943254,
 0.7840088506663467,0.2481822885860816,0.30404112247827764,0.6898591792354702,
 0.6000867092795563,0.5465876143017129,0.43260818565203385,0.3619701336071609,
 0.672555886006555,0.2351079942717944,0.8082937103918393,0.514634033238295,
 0.08599347279445879,0.9779726363108789,0.8063666194092742,0.9749405630702545,
 0.8555193872529658,0.07115114447028414,0.6689336522196353,
 0.20380568237745333,0.8218405523025721,0.20870395533195218,
 0.21427591428766168,0.6500024037257284,0.42371480763770575,
 0.6433220414930187,0.3560749023056504,0.38876167713066156,0.5212609884012454,
 0.876140382376184,0.8345704668865583,0.7154898794818709,0.6060877258592836,
 0.5148064378413401,0.44252007462051124,0.8899497759102959,0.5931375244285418,
 0.2077820037577699,0.3616076486122535,0.9306143273623749,0.8428561131740803,
 0.9136688221796001,0.7175228919725997,0.24287533176574316,0.9221707697019127,
 0.9541942312767638,0.23394409495154633,0.2959166923055847,0.6379804993009873,
 0.1291927522534152,0.6145476728210446,0.06620856533639563,
 0.014731882615467984,0.8094375092451175,0.3587015443579843,
 0.6447883960620693,0.31361080567627253,0.19517252092498594,
 0.21413444690356964,0.0054097641064475566,0.4349643340081243,
 0.590854198172184,0.13454130954481291,0.6941634340381557,0.9139338823717515,
 0.1326628631661939,0.6532770184976797,0.6843844409535625,0.6579573829130557,
 0.8602857050799024,0.5503461627552722,0.2189685440547332,0.6290493677610447,
 0.7422600116349924,0.9324662514186994,0.024866100536850233,
 0.9499731164251786,0.06340937835064908,0.9545707179127274,
 0.00502812613937631,0.30425728389729556,0.07329844398169083,
 0.09198903096268762,0.7917677060838109,0.8230716707790864,
 0.24558545372910345,0.3111632801088302,0.8992585091295221,0.6385964046123782,
 0.2729931965990168,0.33851638788827665,0.8615064999935406,0.3780563451782145,
 0.47093682130506276,0.6933079462568118,0.687783177479203,0.7518766148978442,
 0.14244074775881854,0.20951718462266788,0.1879870243593812,
 0.8818135019369275,0.7307939530700831,0.36923493783608674,0.5321492832865916,
 0.6289151992196002,0.02240774006846391,0.8747310324095833,
 0.016723791021964862,0.3611030418719605,0.6256280472672401,
 0.4129472050044094,0.2971783149550359,0.6747129647453198,0.2278050739185279,
 0.6887618087259404,0.5505535066752405,0.7406925260394575,0.11040564177629719,
 0.8683428703148346,0.13468095168372063,0.9375977815014116,
 0.05238347883524952,0.5005792125173865,0.517325604703386,0.15507425044115253,
 0.8626492585038138,0.5110970012248981,0.9615050984326913,0.5552797736834275,
 0.7100838724651249,0.9127831798229967,0.37401844223334946,
 0.10078650713218573,0.21182559752551178,0.528652889328823,0.7934531008126762,
 0.494923746301291,0.15943670782117692,0.8764344022447774,0.6012091073431529,
 0.8825574161956172,0.8831840819259487,0.9617594019260549,0.3086822115903547,
 0.7852805264652657,0.7308050155060271,0.48037764009532047,
 0.22970239768523049,0.30695215774922413,0.4828577217785833,
 0.4519424793927084,0.026793593000389748,0.344769259323888,
 0.16651345376268367,0.6092995949936888,0.7695934658225444,
 0.11492138154645681,0.879219301035905,0.6680913536423757,0.9235476174857651,
 0.6721859530338081,0.2968389009193124,0.722053379600071,0.014521982331230476,
 0.5307571254151953,0.8986636408464859,0.15307430011449052,0.5724173097735441,
 0.8067560302620566,0.058337784801623105,0.9114855605216781,
 0.3794215325126197,0.010732462276851984,0.3731613503757194,
 0.7625129933491301,0.05640300413634547,0.4574539179873801,0.8512358762667642,
 0.015994339120578238,0.46273829731941507,0.023162705748034806,
 0.9688164065661167,0.09259754696058653,0.8679727183422673,
 0.42619275191741846,0.05907301464872816,0.032468373252856564,
 0.5728465766071189,0.6662857097120397,0.17927466779432066,0.3384253800841899,
 0.455566636046286,0.2745211924995863,0.06565815026909705,0.19521240370501136,
 0.9679969542271004,0.16920104690682936,0.7461911335424223,
 0.35125250629189675,0.3156382272262984,0.48082929646975614,
 0.47258415285584454,0.2501105903902935,0.8493563840969616,0.2771469206809347,
 0.4408947629122969,0.6490549644116282,0.9912955909234801,0.6650032990377227,
 0.6157220506653163,0.6745929321136512,0.4973794937848427,0.2724421622616351,
 0.6527559369886368,0.8539061314592178,0.05399238171609888,0.9420093217671104,
 0.2526825932476441,0.9577790763396744,0.8355602327749336,0.5889510224580432,
 0.8848079089270137,0.6834022496507524,0.33398639851289624,0.7590838531257096,
 0.735930848700184,0.3806252317968437,0.023025481431764194,0.728557241836927,
 0.17334107922817632,0.4283550401310262,0.16307732862279856,
 0.9264617381097779,0.24618900050851855,0.3544864414717228,0.6963659931810594,
 0.7161732026147258,0.9436683070087274,0.39760110284262506,0.5085296608239334,
 0.46639976992634247,0.9972015546031653,0.5095019843130203,
 0.49193687981637046,0.4371288008883166,0.030383361673657294,
 0.7809616247106147,0.3844864805788003,0.7190010272225557,0.8063680647686868,
 0.2537430505785847,0.8553522019416243,0.8753852070047143,0.16827570003223946,
 0.6457705097485842,0.6705383488594566,0.027885905419924972,
 0.12701870480313737,0.45406997516268455,0.1609526025026521,
 0.4560689293884781,0.2620250760260616,0.33971870106318214,0.8054385229427418,
 0.8558033836157898,0.4131204243214567,0.07080177769820128,
 0.34307277815500936,0.6724057267059825,0.7413615262728623,
 0.010259868547726203,0.9828824025049976,0.3839182310235456,
 0.022608254139087403,0.8693729604313755,0.5318252934737425,
 0.35597059422112654,0.9925882540068058,0.5725317209508395,0.7139403404304361,
 0.9377703768719098,0.374738437712528,0.2083958624396801,0.010102217312773942,
 0.2987865788205558,0.3506762190305611,0.08095114105782508,
 0.14545070521218562,0.03132311575633462,0.8375213242305557,
 0.8600754154617187,0.8643206698991749,0.4126747370745901,0.8769110429381687,
 0.3841894134804784,0.6248137812758047,0.4445455577353248,0.8846628116218683,
 0.8955014099772154,0.8413091203248921,0.027010917209910557,
 0.8311548739725129,0.6228515254375658,0.2598317722632373,0.14928052150616078,
 0.49278003823152217,0.5282631215969087,0.33912255053068463,
 0.3174551662497018,0.5398001002018409,0.3127239055063731,0.24686450827429463,
 0.24316370936250797,0.5204072217402647,0.9415818465468254,0.4020995720921201,
 0.24328919826983264,0.46768948905088914,0.4090557363067311,0.929324783102783,
 0.23668740268074606,0.552809056439419,0.030545590386305665,
 0.6109976921203372,0.8143777253945131,0.37685149273982355,0.9123674918737064,
 0.5224448411347882,0.6418869817719561,0.6601615409381792,0.06329021936632495,
 0.6820393856232944,0.9855476214071937,0.5686359067423343,0.4985319622016451,
 0.3103104435727062,0.02527257892429735,0.9238082663450264,0.9844626861783162,
 0.806844855704292,0.8498723922639707,0.8286971587663796,0.23022430553679685,
 0.505253992799062,0.8656128981462178,0.4318789003846093,0.6484048021213271,
 0.52016358975056,0.6879378987624539,0.42274981977405,0.8890734793031362,
 0.08556435699558174,0.3595336443736209,0.6334055463264021,0.8599438837610007,
 0.8795607444852575,0.05775782575102528,0.1020397788915327,0.7067519937521741,
 0.2536926026471028,0.9920766577400695,0.8989842484922264,0.9117943802258806,
 0.6027360626517143,0.2617890786009952,0.5988913937660851,0.38252643164630795,
 0.6285218031360322,0.940436109493426,0.46576182006747324,0.05906158310595577,
 0.8370721040465662,0.9547870249154016,0.8402235509916595,0.9919312304352529,
 0.2727007350497388,0.6633835671053547,0.21580159675990263,
 0.30591660696767864,0.6490069033869802,0.4775283845400349,0.3727418231531423,
 0.762526533608306,0.5484454425129904,0.3692884657157146,0.41744623450471274,
 0.14036771914811363,0.8562738514345218,0.9803550810271965,0.6283663634217145,
 0.17669439064736614,0.5529971989474078,0.6240740059186416,0.6958701567648526,
 0.028595249566790715,0.1826192777724681,0.6344127507766877,
 0.31358880239311215,0.7638416539038453,0.5153742205240796,
 0.12467184270724052,0.8454092991227989,0.7789395955865208,0.4609568606399381,
 0.35193689231150993,0.5900146292728516,0.4593051522974102,0.3108469901790045,
 0.5294836265737989,0.6467217230126969,0.46505512899039003,0.9351421822664682,
 0.4662077541091394,0.19330906074737686,0.991317560052218,0.2816372607500056,
 0.7097431220124413,0.734137686547143,0.6819267348441395,0.9170148317766925,
 0.5853490418927639,0.44549889220407524,0.07674533968147423,0.819759015261801,
 0.40723717195302156,0.2936912576623586,0.5751106439455683,0.4505986074749677,
 0.3212469515443024,0.7972675279789954,0.2686083241398671,0.42634776990946255,
 0.47057224415924026,0.7509568585233961,0.03939556068846395,
 0.7443629774781833,0.7164110340674775,0.29773574166703387,0.310680399086242,
 0.964063025961428,0.6899866310638544,0.3596667923768253,0.31903246804522334,
 0.9748450968562957,0.9174929024697038,0.34782965717404934,0.5337983386757734,
 0.3447871642326069,0.9143096803539011,0.8029369153143021,0.34370553592348163,
 0.8815404212216545,0.09368954225415238,0.1432250906165785,
 0.44495141625570667,0.17089354748664054,0.32644379756763275,
 0.03929001576234592,0.5100854770510598,0.7035936282213943,0.9748839774324126,
 0.1344940742407631,0.33497072314714993,0.9609715140627427,0.5560855602503602,
 0.5808035115251831,0.3911912310179474,0.4831337833873883,0.8822893410163606,
 0.12503037174126563,0.2218626629070506,0.6200427375856411,
 0.30751687989331944,0.8990430911032566,0.9909467606567733,0.9252640350131152,
 0.5925052320712367,0.3628451322660369,0.011365550280337922,
 0.021340552091957954,0.521431453967227,0.7439704049001356,0.6035952461538127,
 0.6108688433205685,0.1370706093649564,0.7470877280813266,0.496879219490611,
 0.3263889519106451,0.19084400552581526,0.11554393272973307,
 0.10318726566215775,0.9441173752720555,0.9267513285545549,
 0.31436643766138705,0.4530110651683794,0.29751909276006494,
 0.14765328407374723,0.20618914692634638,0.3651643688084707,
 0.5187972749280362,0.20602202767949662,0.7199496793829901,0.9986216433295462,
 0.8453933699731225,0.5703019038427886,0.5335807358929315,0.31695042955429376,
 0.1441723519326582,0.7609601798891203,0.4068563646502993,0.27458874264923827,
 0.1730395480420036,0.39268289619526386,0.10735231773769205,
 0.36964206081057427,0.5115102063328824,0.4585321590419521,0.7405111508609431,
 0.42085001305781244,0.6570179037225882,0.2540217112576473,
 0.06518856184295674,0.7906048895908732,0.09208202269017607,
 0.09658804690389122,0.07406063710085542,0.701245882238961,
 0.11303300400698635,0.3204452124037891,0.0861045308877364,0.7481475739928812,
 0.988584335048054,0.7596757930218474,0.26077670669325226,0.7147490975589912,
 0.6700064792646054,0.9368733802257858,0.4522111974694534,0.2438212619213851,
 0.6750146064704563,0.49215206629076125,0.4319473062461262,0.498414149723718,
 0.44088741992775315,0.48648789251972446,0.03384583993933499,
 0.8091835507482515,0.43082226237761523,0.22947132374605772,
 0.6102443039157118,0.4994171786052225,0.8206046298497061,0.4584698008812106,
 0.6936507273267263,0.7963514575627348,0.9701889091563829,0.35600057819188513,
 0.29909978647819047,0.6190624840201022,0.9831861142435949,0.4758476370848369,
 0.9327167172090725,0.8998199153612344,0.8215001357207383,0.2354503116694755,
 0.23354672507312502,0.0755162197030449,0.2473980427874355,0.7725027004071768,
 0.9350749919865424,0.25747144639015274,0.03554755562851086,
 0.3096005106494272,0.4013874135372063,0.08215782312303255,
 0.13485195547375972,0.3535627771829386,0.8201773528459964,0.2262704734394888,
 0.7049977373968439,0.2163798041245809,0.3756501791634368,0.3617058474724024,
 0.703792795377201,0.25478484008167446,0.5856462917948322,0.3202913820470763,
 0.5611149952155327,0.3249172708193312,0.5877654504612893,0.37047176252822644,
 0.738497855751874,0.7267991537297315,0.7240105223709088,0.34191384680811376,
 0.5214484212232432,0.39818188845263036,0.9709119150669349,0.5646020787588633,
 0.7875879529056141,0.8725970097078631,0.579480418494329,0.5565144649472902,
 0.5406760805366428,0.794223422427126,0.13265274583668252,0.06586678172451477,
 0.26656764019876833,0.4996421151077117,0.5423928365964389,
 0.48314141640417074,0.7502348719862967,0.039077963834111995,
 0.7466055385132911,0.20075937992497195,0.35124886865181126,
 0.09027719546259738,0.47042551168818103,0.5380810336732369,
 0.048092155073741094,0.5746098842076215,0.9751741674439086,
 0.6293140184978655,0.794511754969768,0.8692167197170071,0.015054385741032239,
 0.9850815952791829,0.6922310056851554,0.1460235939418315,0.7932137105247925,
 0.022212734384087107,0.6952352838908459,0.44672920291101126,
 0.3972637895475529,0.22565148610461383,0.5931147703893134,
 0.44333814560072016,0.8222105418330936,0.6071727516547029,0.5676813465062979,
 0.1691105481547983,0.4552586791198433,0.6982859918974894,0.6965345457259445,
 0.9159806461909281,0.5914547437274209,0.8688139232664686,0.6547669508720595,
 0.19678386368485212,0.9269980762986102,0.01466359662158212,
 0.17034025257722873,0.008346399388613368,0.7837242461536761,
 0.03873580721863579,0.3547000926587891,0.407905780375841,0.39731619500419746,
 0.29169776843217243,0.9019815822858925,0.7221713823934671,0.8688509581032163,
 0.6793759385352356,0.6607002574137825,0.7674053974802425,0.27546558388457043,
 0.46301684179569436,0.06603301503466219,0.08650036512079151,
 0.2462083058422634,0.8279243101860903,0.8609952534293286,0.8277125116534759,
 0.24264382618352642,0.2039060460147254,0.29852813068583317,
 0.21588105515058964,0.3246213571781078,0.49218483927098156,
 0.8177220754790139,0.7389986668170566,0.7929753462460085,0.11052052101892762,
 0.8609712636917096,0.8621406489533642,0.1742429250091242,0.6837732570432291,
 0.748017342630287,0.5877597550206778,0.1339569120079992,0.6474535991558136,
 0.19651335141877135,0.8029572790530304,0.7466725975811754,0.7243412466004858,
 0.597502470476809,0.8534038639885119,0.47533340444386396,
 0.020258080515529797,0.27587872421125736,0.5486338411551461,
 0.9296294553787678,0.6254766852464485,0.5810387862252613,0.9188405940541566,
 0.3623466020156807,0.016162645805456566,0.20359541275786208,
 0.6395774792381299,0.7696108865624092,0.4876412254622605,0.9216749162029245,
 0.09338142302324859,0.5536426512782373,0.34004486847379223,
 0.3131493867040772,0.6128252283561376,0.7652721749609722,0.39272219610231285,
 0.5814747074841294,0.4201708302964875,0.4243753829418493,0.7661634419401223,
 0.3975849677996234,0.5133907710921412,0.19532469533154306,0.6134037742364739,
 0.8695859706202577,0.19537362027720406,0.602054273291394,0.29882332823806634,
 0.3221889854657216,0.7995936011388579,0.5853062584513368,0.2587424844061419,
 0.15368176644507847,0.7578878436236662,0.058990085911104995,
 0.2603973557700132,0.7854108974634892,0.7775792780391526,0.5153919841783047,
 0.05122492223478381,0.5234383254670069,0.0038154262267806605,
 0.9260236992478856,0.35372833830924566,0.5584323895868347,
 0.30442761251611283,0.7932940353189638,0.4734590359180486,0.6857984093370071,
 0.5949494359473723,0.9539625008622727,0.9575464427751413,0.47000651915289393,
 0.30625158070444936,0.44037885804105237,0.7244817890610901,
 0.5595730980773143,0.3927068661174603,0.45997442897212515,
 0.25027986104123345,0.4178421658205975,0.3468723671246866,0.9801726243454983,
 0.5015084146303764,0.07237065165216006,0.12359142374557075,
 0.4014857754128721,0.24011848742956265,0.7340943882642075,
 0.08183338905897808,0.4406056628298748,0.09323605190605899,
 0.8706038085696967,0.47736530562045854,0.743057368828992,0.04961976763053444,
 0.6782887955538286,0.4924132158177761,0.4830503619547145,0.268591451530071,
 0.06672425740078436,0.5663192861633524,0.910222203635154,0.3914293912491448,
 0.658413862117261,0.9856898794603092,0.17251337953334644,0.2778126882672075,
 0.12491155085442607,0.7162592213120735,0.7316790311740325,0.9300883382268621,
 0.9991114005790024,0.6864713128906189,0.42454893314579867,
 0.46112979363218554,0.08287911314599328,0.5316331878258479,
 0.7959277400365318,0.19702559766631367,0.04408982407146689,
 0.9763952600473043,0.9760599369304224,0.38165978413251833,
 0.47159390483883257,0.6322840006196405,0.8505577430854414,0.8673534144764897,
 0.4641602617010774,0.7712840609200858,0.9205018986097426,0.5931700702795116,
 0.6995706152071398,0.6995546650081027,0.4812337976839268,0.6187107397289442,
 0.49042417953219664,0.27129694422231987,0.477090405040691,0.6070193699286128,
 0.21934316663785536,0.5799509377665886,0.7953105263269111,0.9205125377213143,
 0.9018624477089632,0.6447589826130775,0.2108909597717772,0.7565309693371673,
 0.6597685311294982,0.8109588503269067,0.9432908713099266,
 0.018383978550432944,0.6058369514004183,0.6785877364900528,
 0.08461609689954275,0.02487747681531638,0.5065634126066457,
 0.1682724610709968,0.6530878125239904,0.7332180662192294,0.5719574184146792,
 0.10828996286095394,0.3877643338468021,0.890866356192172,0.4098666800366084,
 0.08769463230248598,0.22430294087841307,0.634827357555923,0.3136208378982017,
 0.10407048862488344,0.5491525268817576,0.6368692135240861,0.6588873146868743,
 0.4577263401247056,0.0034507023369492185,0.7496848769314001,
 0.9411243795311811,0.34442391605370626,0.9675196283015197,0.6489432427233258,
 0.029566280084460494,0.7321966023247583,0.9732500932401712,
 0.2610464491866912,0.105539151069312,0.4022056597866598,0.8462113722809808,
 0.908453080512189,0.6502542561821194,0.7070311778399887,0.9620252344746993,
 0.4021388379549832,0.22184167413362177,0.06478518750396789,0.917645623142131,
 0.4695976018653748,0.6806742889145088,0.320030950532297,0.039761461504208895,
 0.4759559162860335,0.9862458076341735,0.6563372432907224,0.5094800396623342,
 0.7158351551832177,0.7936595747701588,0.5717291661124249,0.48545449339382973,
 0.6539013987925157,0.8213536042820891,0.24645919121614435,0.9885536883941738,
 0.9035345064598534,0.16096651244147198,0.8095076849839655,
 0.38660997521298124,0.06419790544854664,0.46940684251767917,
 0.9319967455562682,0.7047368879880523,0.8690894746615311,0.04480075069709988,
 0.76203586057491,0.5315384859819445,0.13599773488722877,0.15344347940460779,
 0.5055768218639518,0.40757041326792365,0.45306705049882,0.5553772602519585,
 0.415180840355292,0.9217977565387758,0.25044153512574063,0.9520986940702232,
 0.4783556044848709,0.6821184999063581,0.10224401752463486,0.5668560932656325,
 0.5170644030413637,0.6685399517998534,0.14543477512463832,
 0.15422890746094742,0.7756781813932355,0.5861604175058357,0.9356221625405243,
 0.040846067306527845,0.33428946900845635,0.9538439932457901,
 0.12450842533521111,0.637402022588154,0.19797899448626932,0.351905200732874,
 0.2794741763196005,0.21634990812614918,0.9796871133384668,
 0.28307017989756167,0.6792285213538611,0.14609289890173804,
 0.4243047417319086,0.9350833871411279,0.0720791211063595,0.12474158870117424,
 0.5178533288446913,0.6453975240205154,0.08692676455904758,0.5300759232388671,
 0.9868534543880305,0.40263078563609833,0.8470954472089502,
 0.37024335871640457,0.04100558805777199,0.03933661753679618,
 0.5867520098939145,0.7242636130213712,0.4508474874771403,0.1797529482489012,
 0.4611955064872907,0.05703862875491472,0.6301003088555844,0.520086494568555,
 0.4644500956332306,0.9786016762212425,0.20517352873686323,0.7305460448742633,
 0.06387678968404542,0.5709453594387273,0.853622964530318,0.1359825581439007,
 0.28031075401377925,0.2445931394825389,0.7756490279094668,
 0.18234175601704106,0.42681983004998936,0.8258789667874583,
 0.7516824697022313,0.7973133635718785,0.50622985519944,0.7841145098023363,
 0.39733018805711384,0.6160738447032863,0.5824727653997536,0.8821049620347571,
 0.7953059985690007,0.4926548545212164,0.3053922411880844,0.8901700620647238,
 0.9418718052541833,0.4116149455866157,0.00624176582785152,0.6949672456475436,
 0.2858097369482665,0.8950255906172672,0.8809830386617402,0.2367046510408034,
 0.6973736401797026,0.39555458603858074,0.714689954145739,0.376050873913113,
 0.10023572037360717,0.245369554935558,0.6301219545965245,0.4485655539631829,
 0.02598583391120779,0.7307256495109098,0.20423072518622254,
 0.9868576485144991,0.0606968726351087,0.19867443507828442,0.5992649410119693,
 0.7232797836303381,0.03732854309816247,0.47070940994953037,
 0.5916358982887525,0.9502643830535871,0.13139027250938773,0.7468909153592751,
 0.8177754499684231,0.5802459275125844,0.45369644631943296,0.7985125268318023,
 0.7175237298955782,0.06712772906783648,0.9754989367434034,0.829454303551844,
 0.06315943432486393,0.2717728421918286,0.8463893273825335,0.5019075356855179,
 0.18312904037655997,0.03766074669751773,0.8311998010936033,
 0.5265590882190229,0.30895902093457595,0.5268375706048662,
 0.14793053446608695,0.9640731412597653,0.1997406568783291,0.7613127116114162,
 0.8738787996248047,0.5252954308929966,0.3990939667755198,0.9981944685949945,
 0.9461322679532669,0.9685464153235855,0.6941105635947981,0.28820244663310945,
 0.7706895542081803,0.9894054504404629,0.5047877651815775,0.5600612290080003,
 0.1483893754360196,0.15243200027095782,0.7150548778155078,
 0.24156764396281827,0.6029560172328653,0.4007048858758615,0.3902071390104733,
 0.07158543764718206,0.5613084733263843,0.45953552154874744,
 0.6578867537481619,0.8206030442909835,0.8639494115716129,0.8323255319006417,
 0.3123467053116321,0.12180162427323504,0.6240582669590168,
 0.041279255936494064,0.5152989430781271,0.14217290273632432,
 0.6690256446889958,0.13558506116762903,0.6981827201708793,0.5341010550837327,
 0.2459442710610119,0.7849753282382433,0.834900447659549,0.3283343714519834,
 0.061081463655753176,0.22288518730519757,0.2644462946911168,
 0.19812195701331248,0.6944643336855308,0.331807247127332,0.6534901106255304,
 0.05429014362707396,0.8393783126917354,0.41363442797504235,
 0.17951603051857412,0.1430892860399462,0.5883289267605859,
 0.025397768117402242,0.3758731548779729,0.3773953938208775,
 0.4120071821003032,0.7062413879824476,0.16089697102743905,0.0467922999543936,
 0.7856219515345686,0.06928419770447514,0.507376262492477,0.9684764075787309,
 0.8189727463621922,0.6179127724106076,0.7010282504922105,0.07906116178828515,
 0.4029151332030433,0.8516317753823973,0.8682769145853921,0.8877031353389762,
 0.9284109547234601,0.10259303360299732,0.872566696229368,0.1598653142500146,
 0.6445173568691503,0.8517224419961971,0.12312943548261701,
 0.28668712101165195,0.4707652262654556,0.4910475542622632,
 0.11563457538509103,0.7565477479170685,0.3458452600571108,
 0.40773037681269964,0.8333942254936642,0.5706210769619673,0.7486053807150986,
 0.2360083874147011,0.8479069030244266,0.590679335341124,0.6013864687417441,
 0.12184532372120216,0.1532438762280216,0.751941424039247,0.5435405131659901,
 0.794048084545776,0.9674964724408002,0.44522942342430627,0.3435079818031106,
 0.22112081007641782,0.7147812413838377,0.27253891635225935,
 0.4455150835057894,0.43692476660299595,0.7188880553621932,
 0.05411977309772342,0.3171502060952721,0.11641174569988022,
 0.9573456913364244,0.1998474748499236,0.08242597961764342,0.7648556401017851,
 0.9467227464289689,0.609339160105228,0.8126329534526601,0.5197862772636629,
 0.7311733739827443,0.8375554315200727,0.6453637178603333,0.4422563888431238,
 0.8913128925008398,0.7469710101293742,0.8867868244767402,0.3234180438011124,
 0.1500273710244735,0.8015739286027327,0.6215377647213831,0.9178509270874605,
 0.11825895407616716,0.04186249698546629,0.02413861692500907,
 0.7433722178708235,0.22769055052414378,0.21096059394862954,
 0.9713621742291249,0.47199917410489,0.008247431260351656,0.24455591951824085,
 0.2993735455116676,0.05611633829017659,0.24661688014160277,
 0.3202990514598061,0.2405490603318049,0.012597112940118227,0.651248727103334,
 0.0484024059968885,0.4793207869215669,0.6886058581504462,0.3210248935797536,
 0.29027157349877364,0.16348040934075703,0.23754391269452368,
 0.044961521293598095,0.6036101361135778,0.2802172971412934,
 0.5397282973849908,0.8790833301063979,0.07301724045469515,
 0.032631302335636714,0.1898573043154217,0.4399179776814959,
 0.6405396399637929,0.5077183592282812,0.1292753012009935,0.918190272369983,
 0.40123876609917764,0.6161343125204327,0.4606157767885134,0.8238547675744415,
 0.9646671191875685,0.017092570768228232,0.3455302967971703,0.173541919753056,
 0.7435459575745195,0.7225408032222941,0.4477549323408162,0.5052128597310136,
 0.558373284643025,0.7890616504515396,0.6768641511361049,0.7234701246318422,
 0.37288502365823706,0.7081836632803341,0.8947482668556618,0.5605300172229252,
 0.2720741626450254,0.03532220299055089,0.2877749142035342,0.8096306840146547,
 0.30272329738497716,0.31229255698281555,0.0054034546569105,
 0.09945359833424772,0.6446318205894741,0.6254356059615221,
 0.41367735198887734,0.17766581928553826,0.17076120408925577,
 0.9097641152271572,0.25324817078499007,0.4725695990303549,0.2897393330056547,
 0.7708018120103637,0.9441586835744826,0.7178167450165397,0.5748590881085011,
 0.19169666042211242,0.23771173885520058,0.2941341252925266,
 0.6789715561391807,0.48531501817579237,0.9041779733192903,0.6894011181488443,
 0.32147385696390884,0.6830523754784679,0.29085628977359124,
 0.7444307334068998,0.3770597343594292,0.3926427559583048,0.7545789187647626,
 0.7625364416094702,0.9981317390577111,0.958226575544265,0.09110834157672987,
 0.8177714925720659,0.22100122775014452,0.1418088128028656,0.6721362453074596,
 0.5223403018032539,0.33732763283528977,0.13906615445680148,
 0.3168138615657188,0.9955820899406782,0.9810182587686592,0.4366780107357019,
 0.6135848475530037,0.4915966002373109,0.12051900443783992,0.7538238138894676,
 0.15311238236363167,0.8798884668951998,0.8531682735137178,0.9572987031760084,
 0.1603044023190907,0.8320556419184372,0.03852240559227038,
 0.48395829196018836,0.843755378940347,0.28042967125633755,0.827927845522437,
 0.5602859131953555,0.7846075050539087,0.7961650243433106,0.6064364531752626,
 0.5465498587018025,0.5739676819832312,0.1391252287985073,0.3737402471211563,
 0.6855066815729816,0.07535489757038993,0.5840752919193418,0.7614788665419032,
 0.9732741583109399,0.6992433434909205,0.6137537547749007,0.2805702872469257,
 0.8130104645611576,0.6822393654722756,0.27206695852046114,
 0.27181708469282206,0.4214708881784326,0.6572403178596902,0.4351310296119244,
 0.656162031008782,0.7262128203155653,0.4527660358345289,0.3704769853158806,
 0.5397495190935808,0.38795024999430605,0.7278105917184514,
 0.24306003525346453,0.06284959372166887,0.7581251622079558,
 0.7653504180913584,0.9234574649748034,0.6006285593987638,0.555803994513725,
 0.6650479736873283,0.6810476181560267,0.19869579910954338,
 0.07388660978492512,0.737022197380475,0.873809960741607,0.3963275846410612,
 0.5350132386925683,0.023277235759349035,0.48201551468767967,
 0.05349951411071008,0.2325569898201829,0.5871005744672275,0.5583620076685095,
 0.9821364559619408,0.6033231483025229,0.03512590554975248,
 0.09850604748775604,0.10073993693338823,0.3528727958145679,
 0.6091266712514602,0.7240479428900936,0.4395314287942278,0.8415577047132723,
 0.9095806986597855,0.4324720533883314,0.35376685955765164,
 0.09940562247743534,0.5167896465806003,0.3280148216592391,0.9081885084472998,
 0.2120090175084125,0.9920964048625956,0.4055984040438778,0.2024409803146704,
 0.8188899916923417,0.8062595427111982,0.29761132065567286,
 0.03816238800526961,0.11746882578450046,0.310868774451863,0.4221112486610904,
 0.10621751617447728,0.375742248726474,0.2931167045029701,
 0.028356248614574642,0.9519849534383371,0.7889363670535327,
 0.6446926476005349,0.6450945290959686,0.7801296090768208,0.5136826347398701,
 0.10309393711665071,0.7132322783736442,0.4768963165188709,0.57328179953323,
 0.5854755686054409,0.9495329722518224,0.48177003219677217,0.7472594690097735,
 0.19187902981411797,0.7774461872085566,0.5051019178330338,0.2471663214739631,
 0.8364192526129635,0.32364276875846776,0.8272609606982473,
 0.45644734967093625,0.23816443472770077,0.9755268175189952,
 0.8303582573747655,0.6034617561994413,0.7293041044971941,0.9120180828173591,
 0.5879926549587707,0.14221660756331833,0.6538605225155775,0.9237322784902321,
 0.31510191656109365,0.20499472218430026,0.09066857889478741,
 0.4695045069614704,0.28181118113330517,0.6858888300849554,0.5515958979275827,
 0.44656988104202333,0.3532399412760192,0.5812737328748646,
 0.28973057674390745,0.8507484779943671,0.7432670354561591,0.6688078129096022,
 1.576871289271642e-4,0.17304339510823652,0.6770267161670364,
 0.9026419625650907,0.14175458883862957,0.20988482418595444,
 0.7581309831419105,0.24459703072536132,0.959590660513835,0.5902181186457935,
 0.3184807966292107,0.23474592823789575,0.7115056214914496,0.3708867150159102,
 0.9585107219741374,0.2320494616365253,0.9688833507953257,0.7026699974353386,
 0.3307304383326626,0.14386162151284332,0.19390355274277982,
 0.5389676747124703,0.7389695509555546,0.11052404823467266,0.9447146649697263,
 0.7920063238129315,0.2633627356427619,0.18947385931424354,0.6294910035755142,
 0.6191092261962585,0.5423539267570303,0.16266729169220562,0.7154285946458492,
 0.35204922536388905,0.7036610780912771,0.3936826904914036,
 0.007934198879467003,0.552377837764629,0.15115168520142008,
 0.7610005870316127,0.9963799451455142,0.570258212642636,0.14149047386412206,
 0.7758027279458893,0.06539570259164662,0.9033264688921763,0.8869885538520785,
 0.4405541195222067,0.7374034848308748,0.003231388242829203,
 0.31665153185319217,0.6513680661904897,0.04201905477852774,0.818742140465968,
 0.843135363225338,0.3213465106947645,0.7480953227248085,0.9933946318811387,
 0.17996722079474559,0.24553999858607045,0.9169436896028341,
 0.10329068072131875,0.9590998821256774,0.9955299247798092,0.4268492433664697,
 0.43710475154004724,0.6316010116155174,0.00642799085350898,
 0.9835964703329865,0.2621661895221138,0.8426359497020144,0.10324440831549164,
 0.019321896946463357,0.2824879519531055,0.7341696999572781,
 0.24698588517517672,0.038386117304674894,0.08434223140277719,
 0.7692100276332744,0.8805731664241159,0.06170777052592691,0.5295811816090111,
 0.3147014120853485,0.1581008203405393,0.7511237648540516,0.33222539977993204,
 0.9502615905130206,0.7132425762137782,0.684021197418601,0.5355801470435083,
 0.8432674183895116,0.48530235391433596,0.0947939852666182,0.7837906693503696,
 0.8325246636798664,0.2552459333393712,0.8484189285825293,0.9887572731638463,
 0.9204421961282766,0.03085482235980852,0.3153763893232233,0.6620581857809358,
 0.16706407676378454,0.4017275498517021,0.2779429129767812,0.8240728526004565,
 0.4713899169881839,0.29769859163677115,0.7824391738086558,0.2911340465058244,
 0.8184816937909695,0.4645483568586961,0.5247450223764937,0.5724283279360696,
 0.0176689974451405,0.07908495145054795,0.7611572927476928,
 0.06729101145910343,0.18486505019475685,0.029940946755879994,
 0.3817439788331103,0.4652786049112234,0.9261519999631658,0.8986742447610623,
 0.6238859123911971,0.18472010313061515,0.7121321790912853,0.1310897935453469,
 0.7321005206874005,0.23299546596774645,0.8581324981493359,0.9804928763099346,
 0.6461980337421737,0.09269487276099841,0.6090682003339875,0.5167096684342949,
 0.9798472022853081,0.5600164538851441,0.7704858211921607,0.9114457186914225,
 0.09056011420178078,0.8205681446858075,0.3718394923446706,0.8943316499622869,
 0.3119207886712503,0.9899043146839899,0.6303998024208624,
 0.026533145466239416,0.4210303571596343,0.09698568742526859,
 0.3000826669615523,0.08838722044497804,0.6645627623748077,0.4091272004507519,
 0.4879591574935713,0.32707146024686495,0.550624637195245,0.584795424680916,
 0.26243324306190174,0.1397202616456762,0.586389721264646,0.2543064303249638,
 0.3503355222562592,0.3873024809564384,0.022985405735009845,
 0.09044634798808071,0.38541681835283315,0.34374856043175106,
 0.27223061364954027,0.14523336805203602,0.33546522682389834,
 0.5976566278542857,0.4168585813183422,0.7956680558947341,0.22611505002815036,
 0.3543519197710596,0.32605032155800373,0.9160275562712424,0.9417743751190115,
 0.1689657656378446,0.9399536876838336,0.5681580400294182,0.4151804143572547,
 0.644197027975318,0.7140725368579163,0.13109035498098764,0.7358289532973721,
 0.6203585577564337,0.12742732476265317,0.08806293258460784,
 0.1982424020616388,0.8171997654959169,0.2596031863075716,0.18980257291262292,
 0.9177357107845625,0.3967934754131026,0.531558769734324,0.47093245702069697,
 0.6315380482875084,0.8427932011725636,0.894907408263675,0.7276834689582352,
 0.428776207994201,0.14542210963101676,0.657313409508882,0.6072724554571475,
 0.6045111068194237,0.3279382744912216,0.18905901149732507,
 0.04033523961818197,0.6763719381751939,0.5159197953355503,
 0.21908300990293839,0.2870379736049663,0.7892505931023176,0.5792794283188742,
 0.11361612544869826,0.6958952863976808,0.1571715748065904,
 0.09776401623386355,0.22148041883059166,0.4976006802446661,
 0.3179212026373981,0.5293348518286652,0.2961527867995709,0.5535048883587985,
 0.15505111678199945,0.5868444942609805,0.7372412493766731,0.8210469439562496,
 0.42657603346973394,0.8104357815580934,0.20784566570403418,
 0.5717529510130775,0.7974809560379664,0.29181731581475623,0.8434126814332097,
 0.28279489703158156,0.487209800136414,0.41651598248880506,0.6646466907871595,
 0.9089272704127541,0.7081589865667175,0.9302009812381961,0.3303175892928447,
 0.11869532486453804,0.30157084388508637,0.546099571940346,0.983634982176954,
 0.005599583442286837,0.7494181223566113,0.7181884394251518,
 0.7372938089555984,0.600059014789454,0.5314380842305098,0.1153112647559219,
 0.5400773991369755,0.2179346291938703,0.9204495443343798,0.04621295251310609,
 0.3816852822529613,0.8230913410930648,0.36640429953156384,0.6222128668634913,
 0.4837601911894748,0.2094110577690408,0.07481176504605058,0.8583852932503051,
 0.05420816952503238,0.9151057577395377,0.009532328338099583,
 0.17099935373367736,0.48399411732652176,0.9129408944057356,
 0.7952168138363965,0.3725102916958779,0.847908607296429,0.37440922258081644,
 0.9238606268227121,0.7039209803105022,0.798440761351067,0.42909222734350716,
 0.2500703806436164,0.7862907114622754,0.3511430254548489,0.0518024249681146,
 0.4991102905204008,0.5357841023776776,0.5820579700610161,0.5712253214111516,
 0.09613669836181415,0.022029591029429696,0.9548754763560816,
 0.20187729299360913,0.24908255139761726,0.7119938175467012,
 0.6782465958636695,0.6296931990566935,0.5879210555512685,0.8141300858048734,
 0.24802245243151055,0.7958376076772775,0.22253688179516107,0.557050724701606,
 0.3903111520315573,0.6912342910621819,0.47982404667058043,0.7787061244678988,
 0.8268359171811727,0.7963887066163009,0.4945449400853844,0.854780206644504,
 0.4590620641630767,0.7428150743685569,0.5528908578865872,0.7990407743289797,
 0.6878361865851538,0.9118542674139944,0.14198227775889527,0.4846371342780149,
 0.3154819653496146,0.6962344779419617,0.7475321576855093,0.7425568419387152,
 0.3554830265038543,0.6464715519616693,0.23192909229311565,0.2798673557867678,
 0.46276840170135813,0.8096550778737717,0.58038194735441,0.0232697175922022,
 0.675098032134892,0.5333889980360186,0.7257978723243114,0.5631033098548613,
 0.7639062373245208,0.3278745163590582,0.27655222424160275,
 0.29247455048730564,0.8530778443227879,0.6226936298995813,0.9045804278304487,
 0.8330785233576824,0.7907098645921035,0.6721773000302125,0.2670873209149385,
 0.2290205857417419,0.2270103226194975,0.8321070755739973,0.09556023336505026,
 0.16586483290578002,0.3831896099921498,0.47673153644218524,
 0.47721039542080734,0.9800836949765149,0.12707903669490195,
 0.27992816998880565,0.8305791702126771,0.4709589773626016,
 0.08189519214150609,0.7560750479428816,0.9746415666559906,0.7035900695151723,
 0.023520008759214117,0.6403044207831978,0.1157951260592831,
 0.7435060841737657,0.751784292981991,0.7050605777172047,0.6141505580821418,
 0.3617263034991326,0.40171045030235597,0.2440599634662013,0.5347828865689064,
 0.16870566639018603,0.4269066223017959,0.36818890207416977,
 0.8179670465998665,0.19299291845585276,0.6259667140445652,0.5948610683309029,
 0.3599436913931132,0.209088974355905,0.7544667437261211,0.45646095396909936,
 0.6845604717038473,0.4168538785411009,0.26284846977060794,0.7070852844255572,
 0.3211039828322707,0.335308718066367,0.9163907838789643,0.2805083706224789,
 0.042505178892918716,0.18031578032316709,0.10725180056143369,
 0.4081219792858003,0.8265515047713876,0.15155936122919944,0.5984135180636414,
 0.4733470365494511,0.926242553274995,0.3738209574447162,0.12803596990413013,
 0.3062206518671726,0.4887120350920935,0.4132440698846004,0.3384810883474141,
 0.9293120948464986,0.1587890569646735,0.6455272060302877,0.5935712868324008,
 0.4009188503173807,0.4603433169459762,0.1900700029882707,0.944027483580823,
 0.8401801537715818,0.40839661381179326,0.7366443119916855,0.7265693574933548,
 0.7556532266268484,0.9127327949464541,0.5588328111217676,0.7892948848372215,
 0.3132312758603193,0.8135565087158501,0.8023532273637375,0.9423704619898368,
 0.7172031749319495,0.21401308854863677,0.05535811064268292,
 0.7148636095144572,0.6577018107866189,0.17226436314489224,0.836060180951616,
 0.47828957211542344,0.16425437373983054,0.995708894144745,0.412357410538573,
 0.30672234550533506,0.9484997122216546,0.7075756100620101,0.5770865129353882,
 0.5243704245056192,0.5088555937747774,0.27046324807203326,0.7986763189360686,
 0.3641201317569478,0.7288557648366158,0.6273771720691446,0.6546631138201282,
 0.7099318394654145,0.35191913101626726,0.6515581677532437,
 0.27463571924790553,0.14168282383451436,0.8083396278002423,
 0.7726475327493909,0.23395836887556765,0.16985858389825492,
 0.3752984402024536,0.22410942445739013,0.9984427388981786,0.5940812536607443,
 0.36235755618692644,0.5947566283160785,0.0850519816616283,0.3528898188151752,
 0.23909973345550495,0.7767514026815739,0.5235994529641121,0.7427570263561645,
 0.4747427884591354,0.3535149987771089,0.9210341093168205,0.9020314824500495,
 0.26677985277786753,0.5216604605812687,0.6943912826310313,0.9048132512005528,
 0.3091268926581048,0.42020486657523237,0.13932443367627834,
 0.19862306850113287,0.286767162298708,0.3747681523035812,0.6925604786888232,
 0.294339642142194,0.06205523331349616,0.821482441057463,0.8171977366962672,
 0.961959805865749,0.7472267981167794,0.8722888384572919,0.47907290224037735,
 0.010985742944736532,0.8159893394696371,0.283827797322031,0.8024910213278135,
 0.989818886845229,0.2305656208861332,0.03460386705534935,0.08043127022940677,
 0.9216258270807692,0.5004530104941464,0.4289724253987548,0.847885739732021,
 0.9074276175973179,0.8321109284607401,0.48502200340164736,
 0.27530258642531313,0.4731479703311364,0.5178580827213584,0.4367674072400334,
 0.17074894172561628,0.3456064967740933,0.7455044257913697,0.6293003601517894,
 0.9840751169221935,0.8100824779728908,0.5889023840511811,0.0578513265164996,
 0.8817868490993124,0.43069923279496547,0.8150486011809963,0.8435267248849655,
 0.9178613870634571,0.05134997080390369,0.6628478568089234,0.8579601288668287,
 0.39042903196217815,0.7599515914065926,0.9217609172630166,0.2701871392074926,
 0.33668704019849927,0.8821638589517399,0.06567093074662489,
 0.20370290522616097,0.22028780105036994,0.8491637258653619,
 0.6695296248859164,0.5005925218676823,0.4193094011600398,0.9647411327345782,
 0.7350995514487049,0.5859458479778239,0.38844054664324856,
 0.33571145140631986,0.5419427550818732,0.7342005373127167,
 0.14621961088467628,0.7159163057207112,0.5585026880588362,0.7311561605331564,
 0.3380471889695059,0.21071998229241662,0.15981534433613287,
 0.0064076078857535634,0.16755360851343848,0.02890455465184094,
 0.8275734739826227,0.1918357562394215,0.5873660743403969,0.25213939044511136,
 0.9843194188745357,0.8895658233663993,0.39741650012312446,0.9984376565063872,
 0.9316375055140933,0.09150296268814051,0.9379053300545488,0.6229960068057213,
 0.9860456971791415,0.6338450149631624,0.9346725234317348,0.9609285188411776,
 0.8733759831436932,0.6201078919014132,0.7245394757633019,0.5860317118533498,
 0.2125601748215551,0.03884167128647731,0.8883903564431888,0.609795323885243,
 0.9036071889444419,0.05580598903971523,0.4745925723812249,
 0.32794968635035393,0.2034865599774015,0.6584243489263755,0.9310044472687573,
 0.7559518385614835,0.20475407262191725,0.913394050279679,0.493225835033937,
 0.8701697125877125,0.914506106864462,0.6208914837279894,0.6085311533536968,
 0.6984229269702729,0.5324654569633342,0.33976329986259435,
 0.17603229002164056,0.3804064762182927,0.8888065883600229,
 0.11108968153332177,0.12181172179943989,0.07209337779066216,
 0.6260843834186598,0.2524979622433078,0.9883810776312966,0.5371623533378471,
 0.4966584428331773,0.6254716925540368,0.7984962699563019,0.48845155389223116,
 0.6308999538706651,0.37863526110527956,0.7620418496816724,0.8108352384351408,
 0.004994296046859081,0.5076797383102971,0.8491610195555102,
 0.6663369589322762,0.6785641463484018,0.19899751326257054,
 0.37859611364913937,0.7512622499369698,0.21323757662361964,
 0.07161067643301411,0.575091466313614,0.5755654873901471,0.929994495164395,
 0.40826662619964105,0.7048260984103681,0.019624045519310074,
 0.007700415563113405,0.014029119751791441,0.37198172870078416,
 0.5407391408731526,0.38232636705881984,0.3432844400502302,
 0.37435006272204707,0.7614011390725401,0.6084205715685657,0.5182789644918125,
 0.703730127107162,0.7979391992978994,0.21638913998255593,0.8333376746256321,
 0.288938734271737,0.178420046257053,0.31145998310665557,0.12181196626835444,
 0.2738793728844282,0.8143218899694161,0.8689725736820859,0.6766333134328241,
 0.48867399879343176,0.06031406303935116,0.31203627891375607,
 0.6218658187652735,0.9021106151582016,0.5173123600248082,0.2722389093592732,
 0.6245912759645005,0.4639154842704899,0.6362364706763928,0.6170600739489053,
 0.4976441693170903,0.0607822008454002,0.047084819981759285,
 0.5260725327559361,0.6292944449764036,0.6683285827817851,0.7735967238293702,
 0.7662886685842311,0.8903820773325617,0.9525970636064484,0.64253581381381,
 0.9512033401315287,0.37431835584245665,0.9271865670088085,0.8100276829428177,
 0.8903436474455751,0.89860790221125,0.6775134318980433,0.6706195341887536,
 0.37151701871391496,0.872985895431126,0.8984812837190203,0.3141367361366516,
 0.7661342959696625,0.8043173838097891,0.49708589417793525,
 0.07080679388268074,0.5987647003225672,0.07809277950256721,
 0.4205201775845514,0.31574832044208245,0.7206067688012764,0.262188435991245,
 0.7697527745785396,0.913579631501247,0.8939297377521456,0.8825700643431909,
 0.19629474013153736,0.44783169037519555,0.8675743918427309,
 0.5968211727370517,0.6434070721501819,0.6733710967614526,0.43513410785024764,
 0.6271673269534561,0.6683698557249922,0.6177336485071999,0.35946408021513254,
 0.00750792216143803,0.5093167810149862,0.414664086731819,0.1959136543407759,
 0.544616961166013,0.9991128211118748,0.7759519395390582,0.865341544712819,
 0.7032588666633065,0.931995849920205,0.8693102037492815,0.9310711214270307,
 0.5877117863956387,0.9619759211938458,0.29676938320836144,
 0.09736346586264033,0.8562074648396872,0.8064726878872381,0.5798339212233778,
 0.27969210694528135,0.1319692819997884,0.8322837876630246,0.8665104311107875,
 0.13754673559011943,0.8722981864892698,0.1571636385327021,
 0.029849684721485303,0.6669496087400717,0.9362184126132735,
 0.3933823367403011,0.7222356497132931,0.025989347115555117,
 0.08425996294649918,0.6190633033137489,0.2435633487399913,0.7750574542714558,
 0.6882678133801705,0.09524655274407001,0.3455707589544391,0.4108681306692299,
 0.3520646563006322,0.6199129171730934,0.5799787375930371,0.0419920311598414,
 0.4265725722466611,0.008125102190103428,0.44737412546215416,
 0.2718803556267959,0.5575825113692708,0.7674046664286319,0.04536024027696084,
 0.5481998440803176,0.6887912093919222,0.603503438154103,0.467771418416025,
 0.9278785762825238,0.8388145466293124,0.24755508778991786,0.1837575410618707,
 0.6716474037514365,0.8697375482913635,0.7329154188108793,0.44203017254353305,
 0.4845296927232392,0.0345613389324565,0.9047556033197036,0.6429880776170648,
 0.7194554998005558,0.21421011730180173,0.07133296792211707,
 0.8273035029009962,0.2515937012075177,0.5901616026639229,0.26742785448448114,
 0.8278195377733624,0.8863594370071625,0.9597322685948506,0.26637009830691893,
 0.4110006764232349,0.03540917177011593,0.21564317497755459,
 0.6030277002978492,0.8764562935645306,0.03422044022159909,0.9471736916406165,
 0.77566470928662,0.7421088293409528,0.7843755485668875,0.857545204713141,
 0.6707817658430557,0.882119335414845,0.16860608973436886,0.6352098834636036,
 0.18759129126368235,0.14471052905549686,0.19104045821204796,
 0.5158451369060482,0.6021869116769284,0.24355090982619476,
 0.24390110511078444,0.39684887103191147,0.12662969292217485,
 0.8224861067154654,0.7593481937013283,0.6079223293968169,0.08342638240507572,
 0.07446616172644405,0.907788339738502,0.7366858691008114,0.4003538595823527,
 0.734292800061759,0.5034645347612727,0.9539050651016847,0.5299755630952457,
 0.01177684860149808,0.6838177168379909,0.3773203261032595,0.9799605869817214,
 0.7563057376892974,0.9627578552009286,0.7154110912120608,0.4678277477090521,
 0.8918681508817546,0.38531054458153,0.012589147794952127,0.8531220497585309,
 0.06086715356071981,0.32350565867202663,0.7874042075203553,
 0.8253195051571592,0.8940341806970838,0.3805704899546608,0.32984682905178564,
 0.015368736366380586,0.5822295739566125,0.26898079155972177,
 0.41550190980178436,0.08414051324426008,0.33510474921051236,
 0.062229270807985126,0.7853707480392664,0.10476647356180635,
 0.41812388413353774,0.06753994233011329,0.7252909152915384,
 0.8363357506521762,0.1230792841067081,0.5501600036285632,0.9987644548052916,
 0.36983415295645694,0.9484915982079769,0.6901129209488753,0.9977379106913432,
 0.7522661558787982,0.628330573343862,0.22422101646278614,0.24109770079542536,
 0.37699005525810947,0.5613011297336667,0.07380336098144258,
 0.8753191697689564,0.34621957754283605,0.7410817087637344,0.7895917370543039,
 0.7674581574216545,0.023347932250350367,0.5298415091055438,
 0.024083628903307175,0.43661436493019046,0.11463163239198149,
 0.47734523096179204,0.5278196705862995,0.7045137507117952,
 0.49735282752338095,0.7059090385472785,0.11729354560070682,
 0.36316659160378684,0.4182379761307955,0.012278475528766486,
 0.5006102131524063,0.38331722911847743,0.4436268392404483,0.7422482715253811,
 0.5307552536070804,0.8932274587234326,0.8824065472880814,0.18968409443412304,
 0.9965043352525458,0.7704856848165446,0.48137397370044654,0.6419718201628706,
 0.667100941099432,0.8417976501307538,0.720114492220909,0.8707975805519159,
 0.3483289719823066,0.7816388400989144,0.5726167558367817,0.9547388412943811,
 0.9678916036134189,0.639689103748126,0.014173498345619095,0.3354838814495069,
 0.3747015305443617,0.9473559029005181,0.25011328401685484,0.7340002289848012,
 0.8366535370801278,0.7730413528876396,0.908164942211303,0.10673037856351886,
 0.2766178701826043,0.5844726467431652,0.19906119866682082,0.8158359366112105,
 0.18079336525352507,0.39345268085531626,0.06851697177061811,
 0.5734532833949518,0.7225797757200471,0.8857032683153135,0.8743059520295069,
 0.4244638262631649,0.9632575162043563,0.45485034699378724,0.7521807471828259,
 0.2047664794094347,0.7377565780615066,0.13016062991530253,0.6347378989458151,
 0.29414524903738337,0.23437111443753744,0.7784613746825711,
 0.42320574551479817,0.2846192828930584,0.4122577901849115,0.544513756189029,
 0.5073692933842453,0.7964597818780301,0.024711099912279266,
 0.46271789438824507,0.6072870374563065,0.732158665080214,0.19006697658515026,
 0.23470193293029817,0.6747379957213167,0.08693280741384124,
 0.09691334509587846,0.5450851017815915,0.10097360507254727,
 0.8847654098841724,0.07944902094822559,0.926909824032658,0.08931996169778977,
 0.2577270731513871,0.8008315117255539,0.9401747791444764,0.946965683099782,
 0.7280510429585564,0.9596141376631496,0.8881241147489289,
 0.043075570065289726,0.5768360249392976,0.3906632868110579,
 0.7108691584376354,0.231017562841207,0.9150323236265531,0.6141699402290662,
 0.6421903486885363,0.9717009017741224,0.7927365514410973,0.5623571102485724,
 0.35672838093374326,0.7637196026031028,0.8691167189142729,0.9527128816744053,
 0.407665527196718,0.3332869388182317,0.7124923028264852,0.9890906441768735,
 0.2786854679216635,0.7085883774658012,0.4460642109056735,0.6832939971216343,
 0.37404493716471576,0.644526123357206,0.9457181522735946,0.5668698945664254,
 0.7464258851565763,0.9864137184110514,0.9172932480846305,
 0.060817910877172254,0.11228364272561242,0.9978821468786635,
 0.9657309387408946,0.4908171430833649,0.9538017995403645,0.7073132282800074,
 0.9999476396898861,0.37384346155785064,0.22014346644388127,
 0.9651703989180321,0.10942944936205512,0.8425948227347105,0.4528662839775184,
 0.5018619568333479,0.15045424457083056,0.4745559795690917,
 0.33006859401202837,0.13264376907258824,0.8884605468728891,
 0.8871223657726982,0.31114297392225854,0.19428578858947676,
 0.5694864576808472,0.3005737313792285,0.2827172024258753,0.4336296579251623,
 0.2839138087978321,0.7403565369050247,0.06558340812409669,0.9930845265799408,
 0.7343495332470485,0.5337489172524076,0.2598543520717582,0.6977095208492223,
 0.38464634098516925,0.6701777566943055,0.7742953965040801,
 0.09548877517239263,0.4329416657657672,0.8970891638274966,0.8141505491549644,
 0.23539525378679893,0.32660392439795927,0.1252699140008806,
 0.1431308941053704,0.7312214329556251,0.7056600840522194,0.36155566094634917,
 0.17387757942236926,0.7787873725561445,0.03504617652698527,
 0.4244155601262948,0.7642909415640373,0.7971974229340754,0.4952415417241466,
 0.04928032359701062,0.03355823736670782,0.5637716342868208,
 0.26954113012916747,0.6014130255067361,0.03726383061099814,
 0.4466818004293499,0.9696159188502385,0.3027488284060127,
 0.005557175191469699,0.472166058587147,0.16414972252315962,
 0.11869737616464149,0.9206465134998378,0.24935202034726234,
 0.3816838047842057,0.05844924381417427,0.7099977492677203,0.9585531906053117,
 0.11680115888816278,0.6691424776690139,0.11636954963618007,
 0.8097906320312778,0.42044177324231746,0.5452418331182476,0.9426027670599012,
 0.21180675390029524,0.7568975938744208,0.36843968189762477,
 0.34946903327916246,0.07690271613433908,0.12904408906489961,
 0.03986926543010594,0.09543846833441483,0.5431018693108935,
 0.7774042137479529,0.5126588487972731,0.754084124328425,0.5384314794609333,
 0.22488574856245025,0.5370000981482327,0.2294828670290442,0.2431760269235269,
 0.8598582965510019,0.8479453803404452,0.4441624720976086,0.5063118074065631,
 0.9539457661282559,0.5451266328923865,0.26412854633738614,
 0.01593767133882351,0.7979815228015018,0.4227218961177004,0.9061762787796137,
 0.6406332265160963,0.09837076402536182,0.6937878035156981,0.7150136803809338,
 0.5047788451590799,0.7528617421547901,0.8853044331105873,0.24470743886075275,
 0.8817611844940859,0.26650636713870834,0.1516937415379137,0.4225018969870158,
 0.842589863345909,0.9497301807583154,0.48844731684404874,0.9444274658209527,
 0.328898338905679,0.9939120150500098,0.8325489268980655,0.27242198209156343,
 0.42837651062918314,0.3810027802280128,0.01651985304235304,
 0.9935257154264233,0.6814479881914551,0.9263565488273657,0.5753103951540508,
 0.20423796033849528,0.5406096687683847,0.31293225020127546,
 0.5152707204577626,0.8763931430794822,0.02604697369020459,0.9677880638095773,
 0.2595283182788548,0.8320999274274342,0.6006365004161234,0.701197936102266,
 0.17458511294123558,0.6411363901812687,0.4901417698931648,
 0.41753978380402124,0.46631429770129773,0.7093963768234609,
 0.7806357295465292,0.8756023868923911,0.4371857242112821,0.04720799886416083,
 0.7182876840435171,0.9527505432083618,0.16169262865860778,0.3615298429834003,
 0.4891176771017184,0.22047902011752918,0.8174568311487564,0.9492133993495774,
 0.46169232631991786,0.7156196574867402,0.5359031986627132,0.8481375671151006,
 0.3589863220364998,0.6442482631902482,0.05578841041722249,0.164019955856461,
 0.9192302613275393,0.9542046154261192,0.0488330475841936,0.6957606398602049,
 0.6778380281350747,0.27928183261091977,0.3093621185447962,
 0.24504358134129467,0.11411102625884284,0.8526919355180943,
 0.5461483520593541,0.08960280243635732,0.009905091139756461,
 0.5988366061957324,0.5145650678902153,0.5721048036659871,0.3491843895069887,
 0.4880462183425568,0.5108418173604025,0.3549227689967984,0.8472926487803571,
 0.7236845734807115,0.20868974603745927,0.9015690598376683,
 0.27182994622077516,0.788552391890619,0.6660256944009149,0.33441029169529035,
 0.15196150735096048,0.1029924862259114,0.10903060662676134,
 0.020837903715983153,0.7921604685126068,0.914465771252327,0.4950464786741263,
 0.804180805989546,0.8825987379682566,0.12175577377629254,0.7384339707937922,
 0.7434778764064138,0.1504996088841728,0.5352177981472312,0.5748680223148289,
 0.5122968208183275,0.3143948489213392,0.7423717075582473,0.5332597526696687,
 0.11041719680013407,0.5226727242340107,0.6146289306685577,0.7290924161467373,
 0.7149055569920115,0.09563706921878135,0.5160856241611876,
 0.36482165466500727,0.11867156891877961,0.2238019471490481,
 0.34446488124277896,0.12979912034773966,0.28355090663615756,
 0.5664359232614645,0.9775337481465046,0.49589202729772586,0.8277065525227429,
 0.6113011854230875,0.32094503384717954,0.8780158966366163,0.5400818420449546,
 0.37893039992950117,0.9678650644550276,0.3308918310434593,
 0.44520848104371447,0.4793737885455792,0.9452001076608209,0.2927446186531607,
 0.8979328504081563,0.3590856397267116,0.2054960770033668,0.9498695580154308,
 0.9158464982253215,0.9682227370362877,0.06652332435886044,0.5845612534426771,
 0.9962389578179718,0.3079963100714066,0.5326130250867239,0.9037788636728384,
 0.32555377580040246,0.05048123829926965,0.5164939321498014,
 0.5787588585464153,0.2255061237551429,0.3305721451331104,0.3560559299366278,
 0.11910381531706737,0.121400486892697,0.2150178793408576,0.18983485370148245,
 0.20824095734712234,0.3889325026674929,0.3179933896581112,0.5595508264669593,
 0.9325014476180147,0.9365804405824896,0.1867361637818018,0.34251442775510255,
 0.9755778759970489,0.5007682573155952,0.7856332970806617,0.49877568661324556,
 0.36122478482920406,0.09220640803804803,0.16322665347878795,
 0.6183172919180606,0.9872712167950605,0.39356168111372136,
 0.39478841989828395,0.44018568094223276,0.5934173452465322,
 0.8327064574598001,0.9469024873343362,0.7718051188873429,0.1160249709639316,
 0.8828512403182522,0.29345057129834273,0.9154684724051488,0.4806169531169009,
 0.13136088430591886,0.20940729196133612,0.14176850526414597,
 0.10119200491508351,0.6613374667797406,0.7264334441740323,0.4809142155852323,
 0.32528126670515056,0.3716493522755646,0.10194325190391029,
 0.6998713640970475,0.6918942608190636,0.7598638863733679,0.33093901666636016,
 0.5957793811552271,0.8571557537257535,0.6116279127046137,0.13253482255405802,
 0.021553817317160218,0.8658699454580338,0.8517062935095481,
 0.16788793594446927,0.2698277897365047,0.04954676187893803,
 0.7258022049386647,0.3053190624200266,0.22459616326660303,0.9895884347078573,
 0.2791571493896836,0.8463283955042209,0.9053820031787987,0.9329581156678268,
 0.34070474891019,0.26986002320602753,0.6452157184892979,0.6994631317266146,
 0.3578612141425682,0.9344510044953525,0.9669869621052645,0.9594039236409446,
 0.29650348883682454,0.9443243531987998,0.952141588189439,0.47346620170569,
 0.9681490340184244,0.765715843045068,0.8192095577864895,0.6951773223276321,
 0.6402479316571962,0.8470673651855277,0.35190188021745983,
 0.06319941087220349,0.8451850868713858,0.9379840092982407,0.9771742086049203,
 0.04549778470084476,0.2415401900784777,0.16139984727443835,
 0.6498557311140093,0.12748696326906084,0.07780866729230917,
 0.15376907236373416,0.409606274231923,0.9569527000343289,0.06640911480792888,
 0.9144467458535498,0.20543067260861436,0.988069595190741,0.8773955353804117,
 0.35751109520294644,0.9840152986674703,0.9058923638744897,
 0.04323056182533769,0.6751082113032335,0.4414417786312307,0.6882136097743372,
 0.5048055278230771,0.9926364459121029,0.906268099303596,0.920256693024865,
 0.9450917571827333,0.19283089184034263,0.055359353103715314,
 0.3192060666497193,0.25136135250633374,0.8990852522074254,
 0.17829229649179312,0.6798849203353348,0.7375003143202332,0.1322867432954824,
 0.8927133339837416,0.3038693243580012,0.9858411577716786,0.1049369522468675,
 0.38697475575402773,0.29554888710863536,0.028848710456680804,
 0.9350450750198999,0.03671886836220428,0.06202888530498507,
 0.8705831061668495,0.24414161576752258,0.33437090980864803,
 0.1573780302036173,0.5588867031516397,0.31738176424762676,0.6030727834841975,
 0.224198179707792,0.07541832175783914,0.9052338786731359,0.2090050851691947,
 0.23717967711183618,0.489647930321765,0.7206700544543283,0.31293538061872717,
 0.23996886007640816,0.1566187926737671,0.48783779494615154,
 0.2751976645914579,0.4614178192134929,0.46101345303214747,0.8039098048068285,
 0.4716206445084079,0.8133005932952729,0.9878692960169148,0.12908703428112644,
 0.5141260225166606,0.5871674673918885,0.02514890545298465,0.8864512166873674,
 0.13369855045683188,0.6542180010055022,0.20145789306368178,
 0.07958950346499438,0.3210616185547709,0.8282805410532206,0.9086491376326901,
 0.7149407830915048,0.5924038559944029,0.35546862268876855,
 0.48959763834068326,0.3991113307654919,0.5487001811758142,
 0.09901395086717757,0.07546742513849625,0.47443878592611244,
 0.6813912713781749,0.20113218313428383,0.4834420011132927,
 0.10795946396064693,0.8743159605851434,0.8723738641957335,
 0.38721608920854367,0.48910128677113807,0.6308371887856165,
 0.42546567414472414,0.7356215957249734,0.2445726277147412,
 0.03417603002923486,0.8813037217908076,0.23457557980366772,
 0.7778568347159625,0.19132422039224606,0.5411526096879457,
 0.43933975556199556,0.3945033387587825,0.061318313600921104,
 0.297565748784517,0.8826066849301697,0.9535200441275491,0.9215873712753494,
 0.043764405218226,0.1848475821251392,0.9304059800358724,0.960779752104837,
 0.03553154336806641,0.9794686458055782,0.07329182247105992,
 0.4514582692310366,0.029439033051222196,0.1718488656196252,0.849306820506298,
 0.5763703524295827,0.9229727676325333,0.09013752332591374,
 0.47875774309038466,0.07067019752468429,0.07343025516194679,
 0.5614569884302955,0.3174415131538906,0.07799081689692955,0.7410627101274821,
 0.9617147494823863,0.4746145091682723,0.198103887299801,0.706837434375499,
 0.6348219137477362,0.688563566226051,0.10334734737205742,0.546037538197897,
 0.47501330401341624,0.33397655553512373,0.22277777838799162,
 0.5740467629477931,0.5016529279348968,0.16623390197018695,0.6521539385618809,
 0.34431622600891043,0.1319284653764704,0.576487469181632,0.5458572876602631,
 0.7314722006746741,0.05999376020177882,0.5825447495517159,0.8189737582987966,
 0.7497963981532716,0.6436241380936626,0.152981588820173,0.4298526261636475,
 0.1972383807305541,0.3109552077460256,0.357991162446036,0.6707153568757453,
 0.6418096956419882,0.11795543229231054,0.20910576780162315,
 0.9181106992215052,0.04986152430398827,0.1333386738177209,0.593523856144987,
 0.32507541927855466,0.5350607482887457,0.6646826983896101,
 0.21052277215189974,0.033599271635077255,0.8612986540683735,
 0.4947094410877271,0.8999859318722272,0.460686916244455,0.8666345858234068,
 0.38989896193295903,0.8185882109578609,0.6189978077370019,0.7101917590693124,
 0.5970927235597431,0.24397964419190554,0.8565065146909576,0.7751085140676881,
 0.3432520883307385,0.6299437091368144,0.9947800063649381,0.40392728748411166,
 0.87995537690762,0.8099406097137113,0.35091559748030365,0.3052700700896991,
 0.058565371372291564,0.5169064074888248,0.4154661603177424,
 0.30539689830738326,0.8610314813167632,0.11452890313924147,0.647408198686473,
 0.017571206603134937,0.19106054001927986,0.6782336982764835,
 0.9749981753705453,0.19221121417506903,0.9743833484207876,
 0.25965846100866063,0.04835259852273477,0.4671475478723609,
 0.4604122393010045,0.6437792527956936,0.20893438650942664,0.353549479608103,
 0.18659711996996808,0.6007943068768754,0.6556180839068967,
 0.11132616681222463,0.08137680407128334,0.9854872856527177,
 0.35353881078560756,0.8115424383333391,0.8654402032224313,0.4933417243118102,
 0.18524381518212496,0.4187397003720179,0.8092392510328293,
 0.43674314579342244,0.8138390580704135,0.9203684503041749,0.2857393325037012,
 0.8333389764576002,0.26247298612112724,0.4792534059974345,0.4889492571372318,
 0.2688150961366217,0.6011116908049283,0.7169586922897242,0.4103459830277447,
 0.41082106573574817,0.10322364313954702,0.7606330188717988,
 0.8544383268304997,0.7482724460244781,0.5947115780027283,0.6178173276019693,
 0.857678031859695,0.06195760233840897,0.38058293230324947,0.7409247710103295,
 0.23082982426884813,0.6927434180385138,0.5327991327754207,0.510268448049385,
 0.5524417493106106,0.6018417293068647,1.6880389670514973e-5,
 0.7479142050467841,0.9929201360885956,0.1238060015149064,0.12050353836293493,
 0.8730037527632156,0.39393041951361063,0.3233804142697232,0.8509600767200107,
 0.5196986207989667,0.2913873241769923,0.8563080683205538,0.6840197412727721,
 0.40445544802156963,0.8000667803653683,0.37788446426355293,
 0.8910271405612018,0.5098820513830901,0.11174769759444625,0.8818253155107989,
 0.2936306499342207,0.27126698688690143,0.03517772295419319,
 0.4351723727025315,0.06516855266842003,0.2582287419534589,0.5081802718216532,
 0.6061188247511365,0.9313349545847596,0.7725836047655381,0.6913722990803928,
 0.8517239737961916,0.5772918318460468,0.5358184597488941,0.26141618126882404,
 0.540360746576451,0.34119053223946527,0.5967228233130412,0.7799229576524781,
 0.3004341269508971,0.8756628161777706,0.7140747566149257,0.35483461139998906,
 0.9333364926033296,0.3071315560231861,0.7937701174286809,0.7455333844778483,
 0.030303683761900513,0.49584737245668153,0.12633865996100457,
 0.9776309716036304,0.5952131659657233,0.7369184128886118,0.3230766722754289,
 0.2416135956641532,0.9625299908728983,0.8251254518094151,0.2951724476478106,
 0.5299262745799124,0.7896734702718007,0.9732995297932044,
 0.036830656298211695,0.055927837291788296,0.7095031693969451,
 0.04145424757523186,0.5519156942090068,0.5629047524431446,0.7588397280456998,
 0.3839498148791587,0.1099165185948946,0.8470252230756344,0.14760720423016682,
 0.398161259616158,0.38187955468554535,0.3544015557946194,0.9957626460957628,
 0.4276723971270804,0.16880614701379992,0.16523183216415183,
 0.6705896582689013,0.5536612023462857,0.4545776015414359,0.5637216158743156,
 0.5339655137241923,0.7966719224503045,0.3246941743211075,0.6674205199510883,
 0.27533426001777694,0.6012323130233855,0.6951177958770444,0.9810364531493263,
 0.521951162115351,0.6526529220411721,0.17864213737032075,0.47016872069725024,
 0.11563820610351117,0.466476675360875,0.9612295241840529,0.9648305486173405,
 0.23058943723159897,0.66838706047582,0.8138460726055395,0.5609369618288074,
 0.7193542884743193,0.8497571730921982,0.27596748949822036,0.8000407671397927,
 0.8179214010945419,0.32345483291280386,0.6011277540756488,0.5473624869302023,
 0.0668521611302283,0.3120811819080723,0.6356887602143716,0.6013439678539076,
 0.22763524812438496,0.2866786436851142,0.23320258701713537,
 0.6176340587967583,0.12358128925468259,0.3692323774901749,0.412701944303585,
 0.3132083723935857,0.12101629087110055,0.5542330558252072,0.4417343781533041,
 0.4052875866802268,0.07946059699020203,0.47262448827259573,
 0.5247088140377647,0.08662835440469363,0.9109175887528284,0.6743191238528228,
 0.0236584152764322,0.35745256892585286,0.9562692710998463,0.799182281457121,
 0.8962412201889162,0.4880144902402832,0.33763965919651395,0.802911571185847,
 0.9966471815613848,0.6946315707349676,0.6340268515929813,0.8383007994588214,
 0.10714500405610894,0.5311257072845009,0.45780739973262163,
 0.24216107744683157,0.4158119919878851,0.5135618340845414,0.9746231186563353,
 0.17777397641513626,0.6552630439321724,0.9967587392007999,0.7649071513350881,
 0.3514048755054293,0.36839539941715826,0.08193663720936306,
 0.5570977751369988,0.13210192578295277,0.15188923968098234,
 0.7673837618423571,0.9195532946569038,0.054647098134459116,
 0.9390954480303412,0.39033965436778983,0.32212195461696536,0.722369622233829,
 0.9813996075576071,0.15053306850142611,0.2453294721947521,0.5009228473896873,
 0.9891833395255722,0.48642008343759835,0.33512973782178057,
 0.8378511098982147,0.5593588937082996,0.2262521201926012,0.11545168163982822,
 0.7525063559578246,0.5116960020642086,0.890305543298323,0.8114310040388798,
 0.18773737529550882,0.6862276154040301,0.09380188465392869,
 0.5422255189580076,0.5109116623482995,0.39723370055270213,0.9061515203083469,
 0.9409286791512078,0.12952005790865084,0.20839129014382518,
 0.01472786479383792,0.40141393065624886,0.0425940808491313,
 0.33954567688144865,0.017968710724200476,0.9745133295516611,
 0.3203222755746481,0.7802521664406857,0.9177784765289534,0.323800830156024,
 0.18773415749394506,0.5646232432633103,0.6750186024293622,
 0.45373492643153324,0.5495475573435087,0.003728179932643516,
 0.8565117898097288,0.39916874968108007,0.3947875393254796,
 0.16092553259552678,0.666150072804272,0.8537396644882582,0.31460949950648887,
 0.6077017566395746,0.28815854328886825,0.8713968231600846,0.9049259984342468,
 0.8491422960142093,0.9050828780719588,0.56719082909827,0.6304350961520376,
 0.2804322602542698,0.8035762428734202,0.373581314115927,0.4241440351646125,
 0.8356428604197714,0.42650310179316464,0.5193260178963547,0.6581551673107314,
 0.19807605940624495,0.10952110063966702,0.8964527202098019,
 0.5966482922409657,0.9698134972685506,0.9818078359134715,0.3161822887244189,
 0.8235571612767923,0.5863516628802093,0.33081790790558296,
 0.08858376854939765,0.9612427320596284,0.6943243541223718,0.9898559307319247,
 0.33270630477425733,0.73770007733073,0.35537392832814374,0.11718371342821887,
 0.874872388348219,0.08716204389741711,0.06924465976241223,0.5136676978926735,
 0.02493406758547112,0.3148590620534528,0.770900716834696,0.9035339619461723,
 0.7670242899779152,0.020542901306565997,0.3674428511218174,
 0.2173651983037248,0.5565787208893443,0.057510912887160814,
 0.6293062349947829,0.7015561197773375,0.376138959560102,0.18663420740054215,
 0.7085584775876792,0.36365630013845684,0.31727802897607604,0.671689023040781,
 0.05954846020212701,0.4253654735513783,0.07392709566850975,
 0.06993944108977501,0.3721049657593576,0.4239461888823757,0.8239383938497895,
 0.41158065392727217,0.31525826929401024,0.11717167267840323,
 0.7188054634482791,0.8652352724479248,0.6982138027701891,0.7268962897913251,
 0.6027090285940895,0.8624701135587586,0.22755250430235874,
 0.17519842290987397,0.39053610723279386,0.7519459205776137,
 0.6738862797988574,0.6638836905723322,0.192536719790652,0.4576382381472859,
 0.6893212232877196,0.804474827184498,0.5840804412134449,0.7586153573155974,
 0.2609020376585015,0.575729305180666,0.5632802465470753,0.6157397095505212,
 0.7391022831440924,0.48504326886786664,0.20035239698258234,0.635365753893489,
 0.7813123692651518,0.07991838209596325,0.31396120449118003,
 0.6680641634428088,0.1034921580273972,0.2607767574948322,
 0.008001018610495958,0.6222539432946915,0.5322758420108045,
 0.28072364803910277,0.5142711081551854,0.9019206074705922,0.6045420315753626,
 0.004797086820273355,0.37339297372611635,0.3446931413799066,
 0.5514059374573824,0.46477542902914526,0.6232266524944109,0.8525727352017543,
 0.8937382422492663,0.00868464988962847,0.6341425093466244,0.5184996962952478,
 0.21196239839578102,0.6103393293119892,0.6874386404233515,0.1690409920776632,
 0.28163027473908975,0.44941182477583086,0.5629304929679377,
 0.8368010514133174,0.0289343411683326,0.11823606926749719,0.9888989821371468,
 0.3328878495347093,0.5652107864066468,0.6258948649537209,0.28831180346385976,
 0.21236822091892282,0.5796313361610147,0.08272180012652341,
 0.7362393263856154,0.9791826258364459,0.5044627039275349,0.4012604704489674,
 0.1708331667984504,0.9350792622491326,0.26987068784792145,0.1166498459358084,
 0.3141040429713997,0.4955822822239506,0.9846827591187992,0.9309516150038382,
 0.18170371459293277,0.8248734608339896,0.2034154719456558,0.8022588831204552,
 0.7889797402321027,0.058153284934543525,0.3961211692040253,
 0.41439238841992077,0.27881878843587027,0.17415750812269049,
 0.8522303635230069,0.12716310257724794,0.8987116954475625,
 0.23972942934683272,0.30745795281843236,0.11477225894173648,
 0.11177629770989173,0.3991348979893967,0.16683958326638737,
 0.8325159242022693,0.47458804210695504,0.9635185794463993,0.4362483778336026,
 0.31537135927907567,0.7776887098922387,0.11850811439961029,
 0.33191510427749316,0.6976090342013367,0.5528176790500035,0.2359825424630725,
 0.054291379664271444,0.5920935560007894,0.9261291869718925,
 0.7024977458101903,0.4280925116875196,0.7659679583891107,0.2550398476643304,
 0.4041461680401143,0.10448870923882625,0.4902193618863052,
 0.002823663944481525,0.20251524264189247,0.9517980359020202,
 0.9185967068236058,0.9198199629339805,0.36585637460824627,0.108884786447907,
 0.4366210815098419,0.6716601617770458,0.027669426735786873,
 0.8145875937935649,0.23102374408720938,0.7479854630495524,0.7316674808396922,
 0.7953101014755399,0.5187270050842862,0.8316488333132941,0.5733622844717682,
 0.4084372982743447,0.756552854636255,0.01333643893292591,0.803924958903685,
 0.34044994831531505,0.7097345899956038,0.24240639992441682,
 0.5303080861333289,0.059094056438735665,0.7967874080486194,
 0.9536848231279043,0.45501013269644075,0.44727276187942033,
 0.5651760196922352,0.7517830052094552,0.14207501509661036,
 0.12304025706097788,0.3858771218975823,0.3175036130375136,
 0.08472413762701714,0.5154651850462982,0.13352338947798936,
 0.7711058721253398,0.47833156767569085,0.48927646204293607,
 0.8127563609397923,0.8283300730142767,0.4641005381429447,0.05637069424343155,
 0.4219755192933382,0.9625691760763709,0.20732682588685147,0.5889452142251742,
 0.7617088692479324,0.7018975578302381,0.9134316718191458,0.8908061459631348,
 0.41277627156844565,0.817794553351453,0.22398845260686928,0.9740878880886514,
 0.5238288391133108,0.05499510464253676,0.18694807649046574,
 0.5746379981101127,0.26958688226699645,0.764141607299867,0.02158259657721706,
 0.1310403722196728,0.8011426140014222,0.4625693633095245,0.2883813392913891,
 0.5716416967989344,0.4385014083543073,0.9389613459709308,0.8873850132816746,
 0.33075572720648894,0.7220017992562263,0.8942670149208217,0.8478253355536479,
 0.9283302590362323,0.3114386394974513,0.3200607733756503,0.8889518522970153,
 0.3369328015617824,0.45076633736950655,0.35430829322776425,
 0.5608737818642392,0.5704860479430938,0.7838465633002142,0.39063082503833824,
 0.7311882754223794,0.9274476591627778,0.6678786929157325,0.3951374481073071,
 0.3396543421631082,0.47032933383709863,0.12397978826979228,
 0.7041237053360282,0.45836090614997627,0.9842604998125147,0.2211030603561167,
 0.904603927615983,0.40652903772620397,0.3841138495393116,0.15604145574827877,
 0.43740363056423304,0.9255387689724911,0.0063828333087015965,
 0.753173312091675,0.3131555094015728,0.4888402575538522,0.8437996820434074,
 0.16591705916524924,0.2682634614220687,0.6938317687854068,
 0.09317662819348793,0.571978838803006,0.3113946970697904,0.6700888994195411,
 0.4845672433252053,0.3168157383404179,0.6727412866944714,0.5303515639498985,
 0.2677864203294157,0.3245391036924824,0.03238404147188123,0.7159927874218458,
 0.5719077244927014,0.6802738196489173,0.9260830465317755,0.8611273502694566,
 0.49553159769625976,0.3293661231622511,0.01714449832774445,
 0.15554668471733657,0.8523640105611343,0.6301046144375317,0.7937193693202097,
 0.20892683428670467,0.4447100852984742,0.5730477507693291,0.8161950079358181,
 0.845319515791027,0.9225241458463446,0.15660740415327656,0.22175252497291464,
 0.4279938266591333,0.27157288870706886,0.4233199418586765,0.4485079691815368,
 0.635548068575669,0.1395118825388757,0.4169811205894024,0.1699471238039565,
 0.30735816070010136,0.6886480759047569,0.3923136023025877,
 0.009656970424222311,0.575070653582839,0.8930016843081908,0.2737653633182684,
 0.1453665660450325,0.6592549947117237,0.871996131618554,0.9524548937683629,
 0.22366351101341064,0.4902124031030153,0.8626187598821164,
 0.24920718103295836,0.48638384162392934,0.419627086729932,0.2125979803567355,
 0.5921004568514575,0.8075320428166888,0.34512475612072346,
 0.058585482923070975,0.4522317318905993,0.5815372910277511,0.778758887174418,
 0.8568903990700568,0.046036108267389464,0.9811675267008809,
 0.11076065718352501,0.11385411845628812,0.6288549650908526,
 0.07985313638485325,0.4198304404630353,0.19621789670620104,
 0.6398085211830336,0.1742322011340921,0.11198694627954031,0.7970438408556398,
 0.5277383178924135,0.3625088148108556,0.3634978506860317,
 0.006131432348897903,0.6202008594100037,0.6994378503906887,
 0.9317800936084729,0.3680306713519669,0.20320231819215817,
 0.48161545219760804,0.0999005083897685,0.8055759363572668,0.7085914762424727,
 0.07026138368168489,0.9290888878349434,0.5826049071029044,0.414284816794394,
 0.6931302965806898,0.6210232125425414,0.8612400637773462,
 0.002997630930676287,0.32863250847236924,0.5832551274254446,
 0.1498031401425557,0.9758257816263536,0.8902355574280651,0.9594366807641921,
 0.7231907356121259,0.7617833366708046,0.9424247009865476,0.31725542696868614,
 0.2421584763022787,0.7180773929783275,0.6223882541311931,0.7620430340580531,
 0.5650940962455356,0.505025939464216,0.618777280243701,0.6497071547336708,
 0.7898375009624286,0.6643575109139227,0.628842085080434,0.5167526070592764,
 0.10739804990496349,0.08380153634774468,0.6704093336969386,
 0.28308666828982765,0.6155725826152832,0.26338212747872014,
 0.06827444078106681,0.4084477844621641,0.23506858223199656,
 0.41082263268639274,0.8648618413873375,0.1360965028592228,0.3533106264218189,
 0.8937605625491168,0.2186452410620816,0.9859710146013076,0.6664010300176889,
 0.2584442693874036,0.408113816835717,0.9331192777823336,0.5097048976784319,
 0.11538518045666746,0.20699297119588034,0.16524723691158827,
 0.48572107228667016,0.5099344034362584,0.6403369459697863,
 0.08660660398190134,0.4551083730417739,0.1757790401155408,0.6518685987641383,
 0.9430162795431082,0.403301704524478,0.6837002624480479,0.1584101729735573,
 0.5190138405772937,0.5867662416172835,0.6819215659210692,0.21418432035844148,
 0.8342907674821574,0.5841591156329659,0.37647727019265975,
 0.22030136100218212,0.3150698974388305,0.5125540223026297,0.9366368980064027,
 0.23632780386159458,0.8840602699780928,0.0025767361004719325,
 0.14917707427246296,0.7899860950477231,0.025063593047987,0.2032730837669816,
 0.3652122921799362,0.5872070414542672,0.8408287760983136,0.12739016098763534,
 0.27262144423159984,0.8466727284266391,0.2243926449048953,
 0.12618813634392023,0.22953683260080637,0.15879813640596585,
 0.7359298552551574,0.04484202165726148,0.3607428510759193,
 0.18264461086020767,0.5719097526613771,0.38306365895213357,
 0.9034226637545566,0.2394696526249156,0.1247822142466819,0.7826905964165519,
 0.8441593696133463,0.38851924625148115,0.8034198016680888,0.758742269405859,
 0.28138819457885667,0.5401429229654794,0.10419029791015155,
 0.6342169703352541,0.48585330885539957,0.9738812576349609,0.2987398411327772,
 0.6670165001178895,0.17008866624584063,0.3186967268556177,0.6809801771749807,
 0.11554493065719018,0.4134891000198404,0.009095938691355276,
 0.8030016397767762,0.668288228245802,0.7905023340326438,0.449869524409659,
 0.874952644948867,0.347021883654408,0.3891725079106588,0.026068348755021398,
 0.6736010210696327,0.3026712435028056,0.1763080209032799,0.3467012807220343,
 0.29105552618816743,0.828893256638321,0.4255347075075413,0.7205971721104716,
 0.6411483071977753,0.3648917284729861,0.8030282953092724,0.3546032183652924,
 0.37711833027962693,0.006461480154331456,0.2973974550585199,
 0.8876459504892116,0.6445193774828315,0.8592917255361303,
 0.019974457677758073,0.8401392991954425,0.3297568941011998,
 0.2822944410924201,0.39803218435089616,0.9357953454691449,
 0.06730429981440067,0.5017130360462682,0.10302071752796871,
 0.33200677046403526,0.7764273149077896,0.331532557696232,0.7214289076400112,
 0.6140755713842228,0.36194120916635786,0.9046392192778395,0.9545328660978157,
 0.4168643112366891,0.3852727057076789,0.5983338832438169,0.5823303945491466,
 0.4817164057451121,0.12405985569715428,0.7322421617062843,
 0.14736933865413993,0.42671653702672985,0.8033430994940025,
 0.29224140667386267,0.8227506950659339,0.12249714831347791,
 0.4765356436035709,0.830637936644226,0.4299956726699615,0.3786914334752942,
 0.661222585182496,0.6200265084793385,0.16099526401067976,0.01513747527129461,
 0.5826992665906482,0.6555508516610815,0.6248612758941253,0.5617614485636007,
 0.8234642569328248,0.4354621386331806,0.9076131013032831,0.36504683331618015,
 0.1813150546911395,0.9981006676365075,0.16411221297781453,0.908834528115342,
 0.19754711800362224,0.03637463479924219,0.35122676137412956,
 0.5823471127454524,0.6846396752615568,0.5882469765356587,0.06023578554754394,
 0.265758197587488,0.8702394160704048,0.8471766386607318,0.7528071359183498,
 0.9965297911419353,0.0018445537758357,0.20969751504684575,
 0.07146897656648599,0.5247800457299336,0.7082551263982453,0.9961732298126782,
 0.5587382861149662,0.8078324644092606,0.5743095366031289,0.2924647275796569,
 0.38916803702595726,0.1743410716483078,0.9785411178389891,
 0.12011292713691213,0.5759572451891442,0.25076362458540624,
 0.9440747402850265,0.34388848287090823,0.3503280633105543,0.9661395188678815,
 0.3563457807648077,0.36697101390048226,0.42342608944191473,
 0.0018326629089792101,0.13056158314947164,0.8308415977403483,
 0.3012905206369305,0.5118483747570695,0.661312059242133,0.01112909182412769,
 0.9937883070103806,0.6944307847722927,0.8904004207914308,0.19534762392927418,
 0.7600075862515365,0.8443812644755444,0.8342849430502709,0.937916566189203,
 0.8134625271764759,0.21013321189592193,0.9095937138368144,0.1676357384111825,
 0.639277671507477,0.05319105057279361,0.44792853070939087,0.2399694074999904,
 0.485817419436006,0.3283533161342873,0.5533682086723676,0.12945573649591346,
 0.9207019336995146,0.8744909809011955,0.5525536501447756,0.11901090500044043,
 0.21496064977229223,0.12103077903255932,0.8996514350114417,
 0.30730915915060963,0.27421111058555303,0.4725025005512905,
 0.6471717750241849,0.8754514653183929,0.5810450029479376,0.0371778716066643,
 0.7355605991162958,0.6571189348596672,0.6828009250564028,0.20607830700610175,
 0.9060628012826836,0.6143605085160221,0.7883655354291436,
 0.020841855608604187,0.7473841756786107,0.4821060717817066,0.800244913740084,
 0.10612403327208142,0.01548319735479653,0.8090118709709494,
 0.19075992815395681,0.3428892247240545,0.0973094523857383,0.2561652577524246,
 0.7821550698234894,0.7385508159952362,0.44254836488219074,0.4553280448227309,
 0.8962132895729893,0.5577903277820946,0.1310241446246796,0.9807186263539835,
 0.26050140531469135,0.7375360336514791,0.7012934695345919,0.2901749746296176,
 0.8006536830258083,0.8965921898523254,0.4258210581624957,0.38173018721031615,
 0.16247543120077623,0.4183437334677883,0.06487191289347427,
 0.8713884116487882,0.5572477842621072,0.1632278657523788,0.6501082605409165,
 0.014681630258128187,0.3804595145498654,0.38023180513013266,
 0.5339398575188776,0.7295963031439171,0.898097875774744,0.6734239435264068,
 0.21956568615735683,0.23780825967767105,0.7550932730051408,
 0.9621171544779615,0.7434101964308438,0.7304954910327578,0.3729963028318882,
 0.08984553542902018,0.7620665050032807,0.04517926695702257,
 0.4285636772407162,0.20094075335413375,0.23870288081451962,
 0.7505476053071813,0.7705745375851671,0.7559861350980777,0.4074671117913352,
 0.09965964866253241,0.9960625906912784,0.33271199498667536,
 0.7262933369781314,0.021434292109102637,0.36935973595065785,
 0.4154362847564885,0.9384708011379566,0.2165417779062605,0.8580910278173729,
 0.9120517296078199,0.8033498078591708,0.6837459988374076,0.7128655270200814,
 0.7133558172232521,0.4324983783152875,0.576665839673364,0.0284786129140423,
 0.07498293811357215,0.28578325465746246,0.04180192406955796,
 0.8331961569738955,0.5885146651400134,0.4197858784965738,0.5973412626453308,
 0.22037059239175882,0.12156172055238712,0.3795984234463823,
 0.12152123422516348,0.8634465429071302,0.5360684903016439,0.4430409774100732,
 0.8352098550055563,0.4738806552485302,0.37010311006194063,0.8809998346521412,
 0.8390067434922955,0.7617930653408157,0.1004793822203297,0.9311828439808834,
 0.8219124079896101,0.4345209550175052,0.7311134750964009,0.8433543220558849,
 0.9100092500068538,0.12427246461356078,0.8576806393040713,0.8665744102687476,
 0.5022838242330779,0.2766825768289497,0.7085429420657791,0.4405880660957402,
 0.6580325012103203,0.6362334624064719,0.48569005947116295,0.883671801617903,
 0.31279623646111154,0.8293035225784601,0.4533853206493116,0.7485747634579778,
 0.11476293801248827,0.669747972059752,0.08281208943417384,
 0.03069560653181369,0.09874042174493591,0.43324167378792566,
 2.8111973383415645e-4,0.5468925398327887,0.5988615339456975,
 0.4184126704068275,0.7548466082829891,0.6331880457764892,0.7132245447485387,
 0.3801987206705887,0.8854190905131426,0.7959254016460646,0.906843101164675,
 0.8989915092308266,0.25686914262013694,0.5975298074930446,0.9031267146891317,
 0.8653451353704048,0.2527681911719699,0.09593228512401053,
 0.06777684603035061,0.6806644420822181,0.11866745988395855,
 0.7656769857433273,0.34943846225117925,0.12638328356806627,
 0.23232380490228044,0.7534489124100567,0.9584032800180993,0.9037498764891195,
 0.08011070171849077,0.10364547712518346,0.5035837367384185,0.569318972749864,
 0.08334896355013277,0.46828100901497216,0.28462237127627554,
 0.8550261246963257,0.8421265637797731,0.6297210387735102,0.6212308354998018,
 0.21898358701938114,0.19125254274618553,0.3255266807455438,
 0.25766429863391893,0.4135299500831243,0.7532086610581732,0.9651085129960483,
 0.7344811924818375,0.5402837636071575,0.05016002172683098,0.5561156516169438,
 0.9546505450850425,0.8672398266696092,0.7675748455718245,0.5879560182568411,
 0.8662984770091257,0.6433626840835194,0.5145394586610381,0.4609037783278469,
 0.6381504007182253,0.748469319633427,0.47171901454530274,0.37701854236785304,
 0.90446129253232,0.24186061273478,0.6438190241343398,0.0422196170450756,
 0.5760103976028628,0.13833526227767368,0.3055579737492323,0.4994520237240274,
 0.21329184521844646,0.4854325301789163,0.31223116244176463,
 0.7609647752586333,0.4525381541956278,0.9237923153764697,0.7124844711025075,
 0.05724565493539291,0.9602948356776784,0.21238391400653045,0.153186748003449,
 0.9694444777445257,0.23373517511128228,0.9274997792992397,0.3812737749948236,
 0.452423547424764,0.20219452141091887,0.16507977435823817,0.7754391371323556,
 0.5384791016593111,0.46000426788804294,0.5275435148886445,0.5659237784099047,
 0.07711632164373028,0.1277925248465306,0.966289173047469,0.7112422857787751,
 0.4499935193133644,0.1523999089293956,0.7914772562413589,0.1768169664101753,
 0.603434969736292,0.2372745155294398,0.3005412274983902,0.10517672893823349,
 0.48792719993125744,0.5100947496174206,0.1925983467636455,0.9834462663406147,
 0.43800654486312396,0.9549446634467081,0.18894999809979773,
 0.4676090795207003,0.35468736294297576,0.21109301499613975,
 0.6403590340686266,0.8296736280580643,0.9016320387897441,0.0889641108668664,
 0.2381287368371866,0.47722890502948356,0.34500007029404034,
 0.6983560868088662,0.9648386765751822,0.619802241033822,0.9456731346064053,
 0.8323953635509633,0.9436511607783904,0.36517300314596,0.1445526974599266,
 0.0355843423908222,0.24026982358403803,0.34602997276647574,
 0.5872471049148246,0.8117037324030807,0.7279145525972393,0.0721583928049696,
 0.6006665940990001,0.8582494239782347,0.07962977392746184,
 0.42711565731008516,0.356479840606281,0.5392892708147108,0.914071438057732,
 0.3255733421975965,0.2918218813991651,0.6207171023892738,0.7880856211517967,
 0.20498082545455387,0.793366095599281,0.7094878592233672,0.04188420610375565,
 0.0037231669939757905,0.06484615165504803,0.17728390465224575,
 0.08620011396171356,0.31945029141252756,0.4247088482142616,
 0.18688889514372864,0.054594889843180616,0.19973362020572905,
 0.11975410998089142,0.2445407935275402,0.6215289163046219,
 0.23885695192052891,0.2550920385748836,0.7517178002094596,
 0.32380877925131457,0.022297609700552057,0.12732782803332576,
 0.32620868700356864,0.18934011869491352,0.5493766517630574,
 0.8676479676361182,0.12360535570017339,0.4769115809205464,
 0.37531123889324136,0.09538602158043197,0.7794556700880169,
 0.0016898604186756927,0.4876745260488845,0.3346715827886668,
 0.3980692029765086,0.6344859444039793,0.8766178592275076,0.09301087460408008,
 0.9085574752087627,0.33213765603033074,0.7045541482689345,0.120766149930156,
 0.8319154030977954,0.997171062780459,0.5684332819867937,0.3265624240861711,
 0.582706092301077,0.598509696919568,0.4155457571974329,0.4998881321269141,
 0.34542152831616535,0.4759095409316826,0.18028843705119357,
 0.6802350015287512,0.3674704010830697,0.34834958660895654,0.6905615880944179,
 0.2917916820287555,0.14281240051888522,0.12360469596261847,
 0.0957222322370711,0.24630181038888688,0.3136169781607512,0.5846705213822693,
 0.15728799787024705,0.2689515087786065,0.4577548704440879,0.5694117523105477,
 0.14913860551785985,0.3029470127914813,0.6715319163700955,
 0.10163461780995642,0.5546410136863276,0.40983017267432353,
 0.38674016102794395,0.39528864949212217,0.13059412375431156,
 0.5162743406860042,0.7573873943526783,0.17056603744592524,
 0.29970410094303335,0.5364802302166088,0.5179493490387128,
 0.15505508263620982,0.16109430346648934,0.2170250348884133,0.616779396097258,
 0.21609937532998325,0.6252700928629078,0.8120525313507945,
 0.028410549865068968,0.7390441891946397,0.24524008440128564,
 0.2561549102910399,0.017497540806069245,0.9746053876908835,
 0.2423981205908572,0.9881528138799426,0.49711813003678884,0.7428904372482256,
 0.6086619290225725,0.4580052752253826,0.7381500198968639,0.8760170520298792,
 0.7886694900987276,0.4338045029261879,0.9073456680822962,0.2951409220055251,
 0.9536642881122553,0.9392970130814969,0.6502873384394591,0.5520548120653485,
 0.04715099922678001,0.7100399152861919,0.7224413805189316,0.9509468755114351,
 0.40842527358970426,0.3384450689312254,0.7640424671235574,0.5426314005746098,
 0.3247693646335591,0.5492080632104127,0.44472775461170777,
 0.25925306561351424,0.29643558832286554,0.06662764136847399,
 0.6066826123196454,0.41942611039428,0.9939321338849738,0.4179169922889865,
 0.19553480117214883,0.3846856550698242,0.9602322307342694,
 0.04769558797990858,0.024686331803306927,0.4661194319116768,
 0.14255630452438317,0.007362095939987934,0.44701720838701353,
 0.6337200607293779,0.6846484847905917,0.7751927189046784,0.8541543195261437,
 0.19612209915772638,0.1142469907642989,0.28268564908025784,
 0.02292554443987027,0.06405674787431459,0.7677133521945226,
 0.4911646357207541,0.5944079808057732,0.7848976886470292,0.3197139324599283,
 0.7585788322392764,0.4642868916497511,0.08458636326576352,0.4359459139219459,
 0.4771099040746204,0.3252808288929835,0.4291502789191811,0.6493968197225393,
 0.2635743547328119,0.678363485126674,0.731099972428995,0.09654799236502787,
 0.33156653113391243,0.1909553371716638,0.5240280998040697,
 0.018274983108487963,0.6745407800319448,0.24948997107381587,
 0.03642754981241003,0.812413823320918,0.8421272109730081,0.8985409061158098,
 0.6221889410899882,0.22562980817458456,0.05108404091470331,0.447489299602821,
 0.32373195196786964,0.9623871377657108,0.7211408908982576,
 0.16841529423333268,0.836934701061712,0.7235868912895367,0.3107142543318113,
 0.7449425088828777,0.7264370090416277,0.9584431919602258,0.4759299787965253,
 0.4925545029335687,0.38367142250727027,0.4518724124808777,0.5156716836598209,
 0.56859449859062,0.431859506254055,0.2517782308198755,0.843976142417247,
 0.045467770596678614,0.17050721789765855,0.2529531620144212,
 0.678200841208652,0.08740133612760503,0.2197516899475923,0.7061186203553601,
 0.4173028983076179,0.9844514107026178,0.19704525849151466,0.5021490960144015,
 0.2938129955375709,0.4950578145167348,0.04726796317564952,0.5654352364582064,
 0.053215087639890024,0.7918437167962883,0.24938106361248857,
 0.3849276685151959,0.7352639133387239,0.35814843315962963,
 0.20422812927564915,0.5811623064728765,0.8905952642562814,0.9230085657867226,
 0.9837169507404964,0.9178239058756279,0.9435271256098412,0.6032355459409763,
 0.5344256724787256,0.9884264201699781,0.8401597823454847,0.15398240058955504,
 0.2926073042872751,0.047919820985303896,0.6517141141760144,
 0.024882869996108337,0.48704504334992793,0.5213501255625573,
 0.6496647122106131,0.9751556309837773,0.046709496756701996,
 0.7074800532931214,0.08744915402013387,0.6880180517822039,0.3073515276098976,
 0.5687190903271538,4.319161641662683e-4,0.3137776176488224,
 0.4628415777439777,0.30181985553167023,0.2243446162148014,
 0.17574913587875618,0.3519449980845035,0.032592516443977426,
 0.3450856176294117,0.9737779621271019,0.01827249370653583,0.9019768780059837,
 0.7684662721024832,0.11767535039663457,0.3784767434439942,
 0.11181669394709304,0.7892217675588606,0.5871944938872421,
 0.39645024045946575,0.7693010098867025,0.7357730365912231,
 0.12198758118132602,0.7057507338487411,0.37723940278030976,
 0.12515707927930708,0.8504703079924979,0.39255632948122676,
 0.43486108119202427,0.4923387376949169,0.24423241331701584,
 0.1133920404459079,0.39811032570762184,0.16523286229917344,
 0.5993158736382952,0.8693647474552728,0.07088488457439024,0.7214772418290015,
 0.21304633412764407,0.9649536077311915,0.9559786000690828,0.9097624097190653,
 0.5633065856660144,0.4629897603396007,0.5965654431895723,0.6103221724309589,
 0.9558768113217033,0.9166971641719246,0.9978089473905299,0.6556131562432729,
 0.9959724240819535,0.040517528332052466,0.924035028219642,0.9185746168112877,
 0.9472194939682838,0.299789237471612,0.481405230852858,0.491918311323705,
 0.36876026364109626,0.7249819459282034,0.9478413994137707,0.1356664233894671,
 0.13398816697351923,0.6641533227510663,0.2858369724129065,0.3399914879332877,
 0.9411852290122834,0.7077551091634787,0.48081237932577536,0.7712263414263142,
 0.5864111400036776,0.1778280063836285,0.33790810453042164,0.6030092857823661,
 0.4806065916113862,0.326374571027785,0.07845014695222441,0.6935889063749088,
 0.8270493724395233,0.46016089724549647,0.4787148512134778,
 0.021759075554367402,0.1284897206541442,0.6115981776393355,
 0.02883647717732918,0.4934846325825015,0.13509651022466385,
 0.25652765559643853,0.7887199767142922,0.9349343938392579,0.0823184974349429,
 0.02781638425080235,0.07066036267384057,0.6589905030328105,
 0.5390368670408233,0.09966383017922409,0.03610506170222949,
 0.8497583061956896,0.6151847579254637,0.7369939085046902,
 0.027351383511929228,0.2276087527596048,0.42938614128786645,
 0.7298602181602202,0.7395730046231784,0.7717025042655428,0.15958040456475664,
 0.5703644168891645,0.8040930784117679,0.42126344412662464,0.6460433225205234,
 0.16241608144543862,0.0014374068008217167,0.5625152010218042,
 0.704780108664277,0.33968356808499,0.22556236241479077,0.3915543312878763,
 0.4671506278215958,0.3653913650926993,0.42601153277317205,0.8312699630378215,
 0.0532352443720866,0.7844860981271116,0.6113460962392301,0.36064238072915644,
 0.5952869445693094,0.596020687281992,0.444636794738016,0.062256280332851355,
 0.8813291329856583,0.9521548959448688,0.9531433017554096,0.973407562187934,
 0.1242023190932503,0.20752929360135375,0.9111912229984813,0.2788255773867948,
 0.7277210301902591,0.25171072042421816,0.029015692859626907,
 0.007558545976773856,0.969098691214078,0.17142135545158788,
 0.43711221966939395,0.29975843088380594,0.43765547208325317,
 0.5070016207778525,0.17915276855734297,0.018218104396117818,
 0.30906450447300937,0.4229287850149781,0.011757767816952591,
 0.8794421753888394,0.9374933387828028,0.2283629163686025,0.16013200443861042,
 0.14504294594200462,0.4011672084377459,0.6794563893949354,0.2176593011466792,
 0.9194660660163382,0.6657944351663939,0.011508615087780116,
 0.6234414225649683,0.8025526579853477,0.9384386590598759,0.34266256184087984,
 0.014774853012093137,0.07186350995201218,0.37294518930071086,
 0.4753035942568101,0.47055085776732086,0.75044406548216,0.9030678742677175,
 0.34877217689819207,0.4043769267679762,0.07734782272280938,
 0.9362594772508206,0.9004316130861612,0.9717741632540213,0.72059160393253,
 0.32903657198784564,0.4842984403792874,0.9433837064979347,0.3570554475663672,
 0.8931364679508372,0.5951186933019643,0.23802588743646902,
 0.042528481238462645,0.11828171193352954,0.9558476980408293,
 0.8357126161752625,0.3623606227385038,0.270412496430425,0.06045429998132523,
 0.014458237485134195,0.47632656677090834,0.8206106246320458,
 0.3016979316620322,0.8150960934293456,0.8268482373771691,0.3611966035053549,
 0.7330941023760765,0.28126884803496854,0.7197195891858734,0.9940396985830846,
 0.8916307636635894,0.8328149716325619,0.4940657259686029,0.6838449173574569,
 0.2910174801447999,0.15832648445107378,0.2038023359854063,0.4867733972335294,
 0.007822686177350602,0.7793433934238667,0.4026912608624962,
 0.8149186545313789,0.9052478515833156,0.6071083576958569,0.8090746415784389,
 0.7112104550147471,0.06012701790476693,0.9868458876127383,
 0.40197120203118875,0.13997260230102704,0.7757939222924866,
 0.4055010054851417,0.40636808429783444,0.735702559061903,0.5479476929327413,
 0.025243046354367138,0.22283929342211495,0.45789024463559036,
 0.6227670427798467,0.24446509764843993,0.2124461690706697,0.7432230578047893,
 0.286109983435923,0.7391559425266985,0.7700073925506032,0.7956160493180987,
 0.3267689343744423,0.7585703065948376,0.5583129362207401,0.7403704194682867,
 0.6134239953245403,0.6674185357173652,0.5953317057595235,0.36932908696502476,
 0.15166887790169903,0.38960491253173346,0.5845188699990259,
 0.6758221647505851,0.34313213034686796,0.8019850229489565,0.8997838818409138,
 0.18110273621189654,0.1736069426475012,0.1947366540053841,0.681109692145545,
 0.5101879597729466,0.8601910345532028,0.15548093123683293,0.2530266426151254,
 0.7111710339107606,0.4529451677098343,0.3829503550847264,
 0.055660696122403985,0.8578935056846758,0.6360942119943478,
 0.005941872796526693,0.6503777028775903,0.612016570074129,
 0.33920897873886924,0.8475833501762604,0.10621576263302834,
 0.4712366137446409,0.3074515908157125,0.9351752854058996,0.6092554496876179,
 0.8217187746362601,0.5606245109894779,0.6376136442111203,
 0.054748193227288966,0.9803228892426313,0.9768277890115324,0.455041469573534,
 0.4812049179372472,0.7079726674937236,0.1943021739102777,0.3577961657474906,
 0.22384854699093615,0.910862979067993,0.7561363365058309,0.7359160194820332,
 0.9148050826673938,0.5195743583083988,0.46641601950900513,
 0.14657175619179275,0.7159299159127993,0.6029010945798211,
 0.18484039484304482,0.37394224897351847,0.9801847938639496,
 0.5288949301839764,0.922206118712286,0.437849308053729,0.6525654054907822,
 0.979680671364622,0.8533436836119438,0.6500851232357009,0.9588969307951968,
 0.33197327926431375,0.25009101810604406,0.4732340523670995,
 0.38145134124270774,0.3560810872248288,0.6566585574710992,0.1577241524319033,
 0.5204037647960488,0.8517051198274469,0.302472989947288,0.7806300232463391,
 0.5429832014050378,0.9710452463239021,0.6506541948622934,0.5069337515474993,
 0.7561512962888739,0.2446599457955062,0.2542421108958536,0.6962226859560244,
 0.39070093201143774,0.047041079609453695,0.9691934425631756,
 0.16657026163401933,0.8198327594924795,0.7803785721568499,0.8995678448497982,
 0.3204711649384111,0.0698506438551929,0.6989767697107206,0.38264593220426524,
 0.7458011078820151,0.3454529534408839,0.6891060803049682,0.8848182501564632,
 0.6437492537662147,0.3127898076443424,0.41104379170225647,0.7882144115134724,
 0.2041137145712998,0.0393675804308542,0.6667353293206645,0.25804537455264764,
 0.20883634063309997,0.050219254666107216,0.6814673185030493,
 0.4767715829278688,0.6574641817610704,0.5170257320094479,0.1348644405078333,
 0.32821499644281726,0.6256357759111898,0.11399678889067522,
 0.8301262860628549,0.5815673771671455,0.0780363470098161,0.40393559835504367,
 0.8811674639653896,0.641927861373393,0.8131860312959549,0.5697630486815961,
 0.5280157056635615,0.09136385869835828,0.5882964206579183,0.4616261510327514,
 0.37747915258590325,0.40665849402309595,0.45896561115705636,
 0.2924549187198304,0.7508155157674259,0.030951862166863076,0.831263766793854,
 0.7797926144568318,0.5750235135246742,0.05099768253388215,0.7256225787091948,
 0.36310972125455454,0.32254651534528533,0.030887784545820085,
 0.9715004821341151,0.18985816503206543,0.567360591568508,0.7600310224666893,
 0.9820722182825492,0.4414139929458719,0.10888063851060037,0.5328963590322947,
 0.4710227587316681,0.8578559594321584,0.7616079801110677,0.5896515122441492,
 0.5454280649170481,0.15913224903330447,0.3142503026976824,0.4743307315577321,
 0.8909149558391292,0.9209157577798563,0.3043258703934789,0.6725582972373693,
 0.6475071398473892,0.6138365628851231,0.8378409464059183,0.05192724414771799,
 0.12593670307537275,0.743824373033387,0.016736607341543697,
 0.13544376828463367,0.6904735849883233,0.24041921042027814,0.518434303016967,
 0.9336160658910476,0.22315767284205312,0.6820381536864165,0.7498576672922129,
 0.7187344804246465,0.43514463512971846,0.3565895961465997,0.7057031021607546,
 0.602366966045444,0.16186579546869184,0.3296546776502811,0.22144917804591535,
 0.9063915814132727,0.31574507277698216,0.7672505769297647,0.8581562274218443,
 0.7782892961208464,0.724769357986901,0.45264587059257866,0.7858748901274989,
 0.3427395204503202,0.03526072991397233,0.24601711015153271,
 0.8080398958027948,0.6268861650056432,0.9158950935463607,0.9673331774225057,
 0.6774760969697698,0.052507205588070094,0.8457997926704436,
 0.7726641213421169,0.563916909433918,0.18683827446603885,0.7803961867022667,
 0.07477937165320103,0.0625637267704271,0.8743586318916386,0.5561842850662418,
 0.5442281285182382,0.3448537290006063,0.12495325790305112,0.6213331922325835,
 0.32811303313754214,0.472734835575478,0.2865311516832958,0.4006918208605096,
 0.8536327318134411,0.8016699887543149,0.4029551158262795,0.10904404842891369,
 0.09791815858218444,0.998894449504526,0.07321636218182415,0.7657228302735738,
 0.33921955538666226,0.7715315523469398,0.2619027586506014,
 0.14029927788682017,0.08759916303494397,0.7572099890530789,
 0.7758980543185764,0.34468919217318816,0.9806569087270324,
 0.31787787592829353,0.19651634515626126,0.22300716936623932,
 0.8345638295422038,0.719064763383738,0.4132570588479816,0.3480283622132334,
 0.09369002258569514,0.8515849012487007,0.8479445878357433,
 0.23310942514941102,0.367542306955439,0.0026368593684735764,
 0.751523732393881,0.7745146960203217,0.0808342610898305,0.7957582420119816,
 0.08603555860973744,0.5370962038072048,0.9224298103044112,0.5855420269260753,
 0.3075830372454551,0.36068910462617354,0.28975535092877647,
 0.19844149424501167,0.9287562802129894,0.9069978342001894,
 0.06499099152181786,0.7334068534438807,0.5003294832255426,
 0.028058133111797945,0.7238936274642136,0.9663947515033865,
 0.040279639181795446,0.32998491692218845,0.21422706480275777,
 0.7093362487935325,0.6741347678405569,0.1339809657441987,0.2213567036427092,
 0.19882670344960107,0.672329873393896,0.17695448318800477,
 0.19178973337468286,0.3951780623970944,0.9493068087879777,
 0.005805088308681983,0.1379707127516553,0.5428041457788062,
 0.040589640509843594,0.6702294418837598,0.693598821326475,0.5128006318104453,
 0.5370463885052038,0.1165014236077554,0.7408842429641,0.7268674303096619,
 0.7953541637949113,0.409440863293663,0.9585481428049744,0.08234667131058782,
 0.3764993987461064,0.46909415647329245,0.9187948604065626,
 0.04771466039892769,0.8214526033389546,0.6120142447215524,0.9672964500373722,
 0.1826317049584706,0.24507546454451423,0.733753253418818,0.4857804541947979,
 0.2127748298244433,0.7591719642853889,0.15927528986849981,
 0.11162904353732472,0.5431349063413056,0.4251633416703551,
 0.11868306484400415,0.7577780024101736,0.5381339805480975,
 0.03063989682801682,0.24523504109033512,0.4373025783635016,
 0.19311564010945226,0.6311565921930345,0.14778472766173234,
 0.6801523695833881,0.6452211478146768,0.529588126675634,0.8794721118040729,
 0.06467247776746854,0.7400213371555145,0.7612685629806362,0.8048422153851754,
 0.8424328886532167,0.29834237872593294,0.22966231571228934,
 0.3548906678221282,0.973367504187457,0.6907716376345102,0.37462339204738004,
 0.512903777975701,0.15198182736636745,0.3737930414621363,
 0.037862824220922686,0.2025544206806138,0.8983565726410286,
 0.6468545812969457,0.7070778145381189,0.0033961250668022025,
 0.729013413369353,0.7534373699231495,0.14298343562111804,0.15263417295067017,
 0.584244940190009,0.8375092105831238,0.5452386928573424,0.987456689877349,
 0.16896992300744706,0.25249354788537637,0.21305965142131245,
 0.14823385314822246,0.4013958421199253,0.5027870545897097,0.9872673686873031,
 0.17464370533329987,0.7877353581905809,0.5246671322656499,
 0.49616634959213046,0.8381837643758048,0.8144914062636859,0.5081749168315179,
 0.11087747272270776,0.1544324499507599,0.8289843250058149,
 0.11002152287935951,0.25186587627704327,0.8445894693975128,
 0.6660387802618009,0.08110204385460518,0.3399288987626935,0.3891325065801876,
 0.5283542202096971,0.9357845497976521,0.34536076244109837,0.2683527563186636,
 0.2222539754240711,0.1118264547026997,0.9895185265317589,0.3170378623797372,
 0.66769492429089,0.792285596319144,0.18043660787329818,0.5858489003124014,
 0.8080961915976359,0.0217791849515524,0.3751121176257479,0.8231214895691644,
 0.1093453135542668,0.6795766376310705,0.22184809918985104,0.3690463665896042,
 0.3640738127906903,0.7797416599675423,0.7933480063060735,0.9849362604419637,
 0.2721129300605636,0.25386903577254305,0.37085838642138336,
 0.03408142311758511,0.7534961178713958,0.6848946506077781,0.7485426365503358,
 0.30134975472511194,0.0572266687875137,0.4164476627898851,0.5255018378945149,
 0.9500377673030204,0.6061999225340192,0.3930019670414794,0.5111711157090871,
 0.09478170684900045,0.8522694226787445,0.7375739293609956,0.5087890860517867,
 0.39201248209163886,0.7676486034556725,0.828940241640052,0.8708261428092838,
 0.6878792089586018,0.389812569408302,0.28352151784003543,0.10031260274909037,
 0.02318993627060717,0.7567802689047773,0.3958208716367795,0.9047550537905852,
 0.610160299945439,0.05504622878437171,0.649504679699237,0.04835913489745447,
 0.9507799648419166,0.8138765636075382,0.010649817630431713,
 0.10139025004753655,0.8094473921184483,0.7058455554276473,
 0.24595608568403315,0.18740545389104657,0.7162163619562529,
 0.24376392957721205,0.6898901742110685,0.07298274881176936,
 0.6133791664083249,0.05292670789502463,0.6670372355548807,0.6336949718344357,
 0.27253953896030525,0.09471214300056263,0.263669232360233,
 0.08952952653006896,0.27306072384866276,0.5912913615648888,
 0.5001509107917844,0.9072125254913251,0.4589378848575483,0.5350856835420496,
 0.3853582655948967,0.8882401053188248,0.14326966848117817,0.4837052699809844,
 0.39874974269101415,0.7885193715876084,0.42588408534851885,
 0.3351125318843049,0.2814532977110368,0.7258350522873492,0.7528612661553492,
 0.42396254350426377,0.6831318539551633,0.4630416475358603,
 0.08637466693502716,0.05995654679824525,0.26311475156748265,
 0.19660788110182725,0.1767191335160938,0.319069298989489,0.3595574246769522,
 0.824024233105394,0.8993542466016976,0.21453857115650632,0.8623566039955586,
 0.435555999044279,0.4237188355440151,0.854784719424252,0.9228375410230905,
 0.7783084182893937,0.8236675473498379,0.8073124531707674,0.7298166085128704,
 0.7869942798323831,0.022461232137517384,0.041773684922829846,
 0.2638715796246234,0.7504426097049502,0.49595052203962764,
 0.16586716304386573,0.003414156643873789,0.5640619375689008,
 0.33276472303269355,0.1583470549625975,0.38830343094109376,
 0.7196192381551929,0.2802661467067601,0.5716276872771722,0.18485772231467967,
 0.13443319331660009,0.56672191390035,0.7633692510067688,0.9173162830395887,
 0.01866979141368008,0.571478294375601,0.3681553361594758,0.8978281305498387,
 0.3276495599183765,0.2816395770432413,0.1303257339252628,0.7658956337459775,
 0.1875685576344972,0.5163817294356977,0.9719671002502358,0.833013576625034,
 0.9720429452523253,0.9208011796415106,0.06181953539754015,0.2198913764931092,
 0.28012647668675184,0.14861153370426783,0.7871799569378226,
 0.3559524595320148,0.86855648463862,0.20290120956212265,0.6667868315448371,
 0.012309024315055539,0.7396163993534306,0.7280784336244828,0.792486769966117,
 0.5681200787576738,0.20427723951212706,0.6729675866048489,0.5274838583851188,
 0.5739352657414688,0.8874389526678197,0.9105363817170687,0.615111517748642,
 0.6669437280912163,0.47611135093111623,0.7752116843782371,0.8445750019437698,
 0.04617944742807034,0.7780322862779419,0.6288924998848159,0.9963000486654626,
 0.3668719226103523,0.3722865978331359,0.9202573399531289,0.16797594970070073,
 0.7326759324940183,0.7811201540474694,0.4619408187261904,0.8194335156523089,
 0.6397909719942922,0.2069749195121754,0.7143862128806511,0.4792021355532593,
 0.4976602615187913,0.02926500047608449,0.013929119950758295,
 0.5828749516052248,0.14758002700713513,0.8504148432326768,0.0322469424477303,
 0.009195019633581936,0.3456005289343115,0.7060608150337738,
 0.9649817596207337,0.6528676637610866,0.9842377695015349,0.1440401615340562,
 0.7713993362628931,0.47288336744210724,0.7059675610210974,0.8771678103721486,
 0.7462981955964385,0.9824567180162198,0.2093360856215174,0.2881782062871773,
 0.22985718719425852,0.5346051845287623,0.8431011579607324,0.7668157684897776,
 0.34855003247923955,0.04420140061403044,0.8112379138663658,
 0.4364609860112648,0.39181654284897616,0.3873141687048587,
 0.24775440640219515,0.29628446094431604,0.3393579087276276,
 0.8901853170985988,0.90414218441557,0.19762541709839265,0.380243315258298,
 0.46274342994346207,0.41938598118884673,0.38831471219709734,
 0.1412589173273162,0.4216646736323033,0.9170371680462975,0.5288185216671666,
 0.1648826714863605,0.2403071110649082,0.07275272231635521,0.6286536357360557,
 0.1988738440702027,0.8060910629451913,0.7046569962527682,0.2601589668177873,
 0.9610583252720927,0.17388833500922274,0.7198263904934011,0.8902071091179695,
 0.024342000590564017,0.634568421223406,0.81416507409955,0.04393750905689564,
 0.2017997626236936,0.5103599690174776,0.7290578723892049,0.427165698186206,
 0.6175371815870867,0.5440662818545793,0.2563609013676126,0.8932769434825509,
 0.11788346955331541,0.46932102920978713,0.4422413108382842,
 0.1500376637921762,0.002476607240937345,0.6338787050645204,
 0.7095784125860075,0.4213527027409639,0.7223538003605791,0.1600958620844427,
 0.9140259932234969,0.4009204916528596,0.7650165023588794,0.8985588468147769,
 0.5719238382416236,0.610393780571855,0.562060437003729,0.1909430837950583,
 0.8600260142027731,0.2326205886736561,0.9101729539691513,0.12487294613631383,
 0.8952688348929853,0.7634556906642861,0.15525998878066127,0.5782778505216124,
 0.3052738434278114,0.19098655510183005,0.01422756455459473,
 0.25299316529582405,0.5147579541308077,0.30405236343498676,
 0.17195785359778704,0.22471548639916883,0.22521357133572162,
 0.6335743377027611,0.8014566586175194,0.4983089840210141,0.9381885908958546,
 0.9877927308461966,0.7223296472192082,0.7150720788021424,0.46917479385019156,
 0.38419602123811936,0.764058420919341,0.9328031065364134,0.891666806983892,
 0.39274146131357035,0.29681720548943313,0.549819089866307,0.5761830100231927,
 0.5283669185059497,0.9704919320861524,0.1142471401441707,0.3926888176535217,
 0.648564319396265,0.24213514894421728,0.9125616534031039,0.9637742559357881,
 0.6543845894585244,0.0013962908179927513,0.31882997161103577,
 0.7202481630868872,0.7546115532229486,0.4166455553786065,0.5981759529385832,
 0.7052993757697118,0.07679241587271224,0.48236973928165316,
 0.9485539288577118,0.8477166225481304,0.6532444558504018,0.35218025613535975,
 0.7794176587637701,0.28125868028873846,0.8677811795745487,
 0.15445222337509346,0.1801060196832046,0.5118799580870244,0.568424540878877,
 0.050418064138483265,0.8832283709932003,0.9428840425437586,
 0.4857289759190535,0.01665057190261876,0.4532787603149212,
 0.10576084736486668,0.8744254720669338,0.7549643166772265,0.5396033560192708,
 0.13434495230403776,0.7947150751109052,0.13530161838173582,
 0.6538749134825094,0.21130207236170873,0.29352574516977226,
 0.22413136398371936,0.2418674422374778,0.802138420164629,0.5362250293704783,
 0.5306415247704356,0.705954274768177,0.14377803526899324,
 0.0029893305408603688,0.5602401525409363,0.4019035187516993,
 0.19513281680168815,0.904465983343844,0.7452359801853893,0.7582175679719579,
 0.6020425462592419,0.2846095128937912,0.11602163092104245,0.718408709362477,
 0.2850438583007885,0.8076560300716968,0.6238258413801654,0.04584914629535286,
 0.3178920078700901,0.7553454999675812,0.17994637253678825,
 0.32767918303413135,0.12367598433501859,0.9055842413851762,
 0.35212041597508137,0.7548328991752133,0.40231578087214026,0.999560919227413,
 0.18160375866165634,0.6411230616713455,0.8727458143088426,0.7964379549906505,
 0.8588896946995396,0.27918025358381027,0.11069655151560887,
 0.6259464461326534,0.682938860260928,0.8530172541329462,0.24769620709432483,
 0.7138679491900929,0.5210898833675629,0.7198679333223664,0.09849702308025965,
 0.7750802474652029,0.9462990929660026,0.8626628340102928,0.6773166068276575,
 0.441516408741369,0.5249451951778352,0.06594920999216691,0.980496100333709,
 0.428564005647474,0.9572313649500355,0.7665360591525117,0.30504082945774924,
 0.5727069931238665,0.20267806397014154,0.988872930680947,0.705091469710113,
 0.6911559205365256,0.45249716148152896,0.1474700742734083,0.6451521101467107,
 0.8888946690459841,0.7314894966994314,0.8694510064242589,0.14236469143072772,
 0.537370912068152,0.9255731019358444,0.5076336031107196,0.017816831673687195,
 0.20962966097476055,0.602291999738537,0.9366110866817486,0.6322777436134512,
 0.9869816014407885,0.17484782655159314,0.29616201497428485,
 0.30698818944179374,0.7704480516028057,0.6919436954007939,0.2140849683138648,
 0.7353053967482627,0.24011388746352047,0.8398693374795366,0.2599582264539554,
 0.42616084649645325,0.5561503275523709,0.8434176945792996,0.790812887304268,
 0.6082092979021851,0.8615148373393198,0.8945633127686009,0.24561765267278435,
 0.8585825627572506,0.17291946620114607,0.0921026131296967,0.9110448360698473,
 0.615383432562436,0.455552409232969,0.6183535184569031,0.4408894339934465,
 0.09137533510193063,0.28355989335382803,0.30838255053257824,
 0.5182903865189745,0.7515180439228253,0.9198003611284428,0.8971490471165499,
 0.19381422077333132,0.3503811020380715,0.6789031735477559,0.8893612529146853,
 0.5693048913717456,0.08134256185071709,0.23948709214408992,
 0.08501017423164237,0.7523188738376125,0.8516277416720998,0.7051746064625117,
 0.9153909727502698,0.5464595850500202,0.8398657135822218,0.03333514259475279,
 0.7480101979037989,0.07232804641127977,0.07476403542259324,
 0.19221288443664641,0.0035036451305020533,0.9219599673012547,
 0.27520236104464413,0.6041756412642657,0.26834935457879283,
 0.47548247115035913,0.7555536673363398,0.7270113124456252,0.6493159762809416,
 0.45197382277450504,0.4712645175673147,0.5755763737879447,0.6575753592075654,
 0.4764849583837367,0.17204554406723893,0.15972718194621316,
 0.9533542737640484,0.08831875924289379,0.12713124540400078,
 0.9883723049467529,0.17786499876804485,0.5730702157504849,0.3666063183442595,
 0.15521791129042595,0.5654270650070673,0.4893257806470773,0.5019816678828505,
 0.7513588805809144,0.5712645898228682,0.5169239356076871,0.2662025104895318,
 0.20839020628824145,0.8467783243479192,0.25732722764737004,
 0.6447966019108102,0.5511024168048775,0.4584409933683977,0.5650910842137861,
 0.4289485392498865,0.2573309991914916,0.3860365871573692,0.5446077622000817,
 0.6105689944304377,0.02114936192085426,0.5010391902310287,0.0080749461340196,
 0.7471318803817475,0.5372821571659436,0.6027999631271685,0.3176014787264452,
 0.7329304168425731,0.48199591288797394,0.20238403184833442,
 0.9062827057947922,0.01591078138974522,0.7183540416939633,
 0.20597883414201312,0.2960125734900243,0.35146459393488827,
 0.12141118285304087,0.2865663952688793,0.06818774654246407,
 0.43198124898361134,0.6068055710145687,0.574123239302406,0.2887649367832268,
 0.00473178446732625,0.5950138650490431,0.81308075149455,0.8171682351743724,
 0.6792236735119271,0.769867141798649,0.9616682710649125,0.9471632629060958,
 0.6426494747044417,0.6191537292185545,0.5968617646863393,0.5559316320733833,
 0.4065029429873922,0.19371468963020289,0.8218293986105315,
 0.06133948235577524,0.7527664255542932,0.9027871328933134,0.1361465252094206,
 0.5215236893662772,0.49284326582130267,0.4660259758291303,
 0.45190157082971694,0.1743985098712456,0.42799329492790783,
 0.8740068914495331,0.4605051144490888,0.646526124545948,0.20098618621983166,
 0.9741039547682501,0.20463938087631095,0.2267338534082719,
 0.057181109792333595,0.620770928356292,0.3412256015521097,
 0.014861305344672782,0.3789949192915878,0.7917702685018667,
 0.6353717897769888,0.8057867615186126,0.15848297732220085,
 0.10989057825508752,0.7736282885628435,0.48995825028988615,
 0.8289466571439974,0.9038704113626597,0.014139580397407148,
 0.10793803437724392,0.3199011050685674,0.9241935879165093,0.7746237915327706,
 0.23822557795868504,0.3168014766562034,0.6093521324054382,
 0.11139598953818597,0.0948891355071928,0.3327823133185662,0.9007605337545543,
 0.25364281534077604,0.0013675644397846964,0.5121631490542674,
 0.5554730594879762,0.7013146286607367,0.5171022743732374,0.18537812901079664,
 0.1011773295173446,0.715940534736317,0.8870709752331318,0.23085181818881428,
 0.7448897234380805,0.897643299863266,0.5370491541258705,0.5835199949816755,
 0.0720200543940992,0.9614515831909535,0.19126781060673526,0.9026654870112987,
 0.9608702179549379,0.6778571475999359,0.4063688646709811,0.2242515045135317,
 0.5551752927708173,0.9908067194773007,0.8859984436030498,0.6896177069691046,
 0.0963669710836802,0.8398633728856992,0.4684181692497327,0.993924469361615,
 0.615662319827408,0.04009385156669654,0.448550413637432,0.2741315640874815,
 0.8780135692380159,0.7718925489485765,0.9780189846308133,0.5348830010384199,
 0.8622263950346993,0.860465439621335,0.03497610136010265,0.4499873785349867,
 0.15810062554075266,0.7850370308182714,0.33250289953345613,
 0.6244140638169872,0.6813824807234853,0.9282283765426165,0.26777947253697765,
 0.617440008095276,0.8850860321324028,0.412563026962648,0.5170345458944618,
 0.26481900055023,0.014797415620113274,0.06793899567727646,
 0.46625915824592246,0.25522800765081965,0.22493489497805574,
 0.4557875363251659,0.7879372437558643,0.8830057582805078,0.5394705043410837,
 0.9988558272481162,0.7539052684849302,0.45847452347130124,
 0.16371601785814205,0.9112318192001925,0.6769452019273959,
 0.49138087234177885,0.8160006145326932,0.3578243542151073,0.6868072780230505,
 0.9856554816163906,0.9977810661117297,0.6100021883329043,0.04314941099254199,
 0.7734083608844002,0.8123319205381215,0.858014254530626,0.48677755008505785,
 0.6512776969733286,0.0659584303541404,0.6426517223264548,0.4870538577728888,
 0.13545218484619836,0.3203218945290377,0.22607717977101072,
 0.9749273466839067,0.03644038532236027,0.5311538380677928,0.1218771469741482,
 0.8505576579766577,0.9223684875483925,0.3019508415863982,0.833189589019285,
 0.6088290631962012,0.38280255797804563,0.9172459457377719,
 0.07910877878009825,0.5972572774126028,0.9387009508762234,0.8117618863784913,
 0.3345583120014174,0.4023629896876918,0.5704998108538069,0.9901582541462015,
 0.4666325915484577,0.7028594668175677,0.6813354367745458,0.5626281417784251,
 0.4781349062321194,0.9229289920621108,0.3342247325126856,0.6918361908569215,
 0.4991206271590214,0.40106973331011164,0.8922113706765836,0.7166331338040017,
 0.040628091526671684,0.9319365282349437,0.24618589620115516,
 0.10586954517035962,0.4203007710320573,0.9261843505583031,0.7329746812549309,
 0.922715586872564,0.5410353690916687,0.9867033845919674,0.31629060676536147,
 0.25612201060049766,0.43252877486491037,0.43563546407089093,
 0.06541974787854898,0.1777944496236079,0.01136822086448408,
 0.39980405384497764,0.5819649169488736,0.11578172787167518,
 0.17476488782651645,0.732567014300957,0.3479398250554262,0.3714431224146265,
 0.18083438454572442,0.12025333500933688,0.6871467080784652,
 0.12776538379817604,0.7071818196200566,0.30547597163315043,
 0.2480488197375681,0.3148683407040652,0.9315971446446664,0.3893036878449406,
 0.20527880386977027,0.24713197966065015,0.9085247919935647,
 0.33032684200413476,0.10786947511964107,0.3342410119364668,
 0.7370751233819104,0.23487622128265062,0.7053325909514687,
 0.08410872165609806,0.11064374154485834,0.9544670551947181,
 0.5204542457328608,0.6235084102830936,0.48430169088605657,0.5119960886935551,
 0.6825164146875118,0.7545650091853069,0.850501637009646,0.699543919459911,
 0.5658561205904247,0.6185307028813823,0.1564930363123328,0.7851008248768004,
 0.255149288251759,0.511949647288553,0.8201109135349932,0.6166796450851276,
 0.5801304961055807,0.3447611101991487,0.39741858376910677,
 0.26479883989133624,0.5816684182773612,0.40094839673727134,
 0.8056703120547867,0.11595101603822577,0.25829881601503757,
 0.5683349994485675,0.566796021173081,0.4052917808068818,0.9871035296804462,
 0.8852703278489896,0.37093789531640486,0.8454740281280385,0.7307068278856288,
 0.9995546626057755,0.5352247935288564,0.1724072385287786,0.9903648740519606,
 0.0902769925272523,0.5413528657285597,0.6044836143261411,
 0.005612873480930647,0.10400024493488957,0.5623022344116796,
 0.6014321907222497,0.6857952108595553,0.10125296907709924,0.2941511618578361,
 0.3820789734168798,0.2723660117818313,0.38452308829866033,
 0.16399089066512162,0.5224159363272144,0.3503149932897983,0.537927098608515,
 0.6056307014356551,0.34384913457912414,0.32075727340512583,
 0.5189553367384858,0.9659496065181465,0.9315845287799689,0.5068303007773928,
 0.7790484388458064,0.6464971918469581,0.14757253613554644,
 0.14665537013747765,0.279747152603261,0.7975997453674681,0.5261194188964265,
 0.37705471819808367,0.220425371412999,0.7219830717740101,0.02970024420213635,
 0.6777261697887386,0.2149928230660869,0.25120975987913063,0.8392843873253618,
 0.17644640115076693,0.7803341066011316,0.44281694628162804,
 0.23775265984566174,0.18075125023073257,0.9338398239158485,
 0.6654675476788936,0.04366414989717271,0.48291806020180117,
 0.42481933624815815,0.6781518328644152,0.6616303792585232,0.7399165971635219,
 0.005041455936304673,0.16993048165915647,0.6897568288118375,
 0.42636280170221985,0.947658630366815,0.4153139627448821,0.6919217508554226,
 0.07864592851143815,0.6491046212923759,0.4323784984222264,0.2747870764304221,
 0.5288312069274742,0.21280977904242326,0.8208204002447435,0.9326413417863525,
 0.09492704090765369,0.340225198605016,0.9515946395569357,0.928150107609199,
 0.6160967349063498,0.7569687084516207,0.9250501086500265,0.989710229616451,
 0.14040579768884287,0.38444162045383246,0.8931875782990013,
 0.6983469874023132,0.48816390194484127,0.2569799566861719,
 0.44086470893074237,0.09875310522211533,0.7006123482362522,
 0.1636611526612426,0.41599005672288614,0.7204456121929919,
 0.28696282505380943,0.9082441436884149,0.5095912616443291,0.8185750127497509,
 0.05007487456910187,0.9986707815863377,0.6544498306962592,0.9499021680354688,
 0.895189812621764,0.7068607901551912,0.7153072465400928,0.13685553168383613,
 0.19037788471825468,0.6513231569499343,0.11494586481714164,
 0.8784219719401669,0.29578937588221854,0.46093890148308503,
 0.01720349759436457,0.6436520016739925,0.9027384589262487,0.786533042011771,
 0.2764807951847701,0.06590040393876162,0.1293414319426004,0.6705978193634372,
 0.9717981067574475,0.02756672089112755,0.5159948214078952,0.1499331171127546,
 0.01698334228606946,0.05756094783003218,0.7710417887627218,
 0.2553775717468486,0.08288773393109461,0.5169928770358934,0.8788496008022864,
 0.42383836004818354,0.29767223844613266,0.2734284112456006,
 0.20212277588091132,0.0505678105464864,0.935159904076037,0.07772764558185685,
 0.6685191315812131,0.9709202042512055,0.6286855003416478,0.12393251188116028,
 0.5072892521453161,0.9332598605388027,0.11889304909330911,0.405462326390961,
 0.2904745385278874,0.3667135993822658,0.7757851469932453,0.9498046087283649,
 0.9814045365921289,0.03990975511847861,0.8090571989492686,
 0.028605566572039387,0.6079796535872991,0.5880332263034567,
 0.7817170052573365,0.87794184846293,0.09864130058206899,0.47284602094835304,
 0.040056662547365396,0.30441714638686257,0.03663227475951414,
 0.6986161002713129,0.3566762492589146,0.4613071042442721,0.35269718943125905,
 0.08627457780650949,0.16880852463188378,0.37675475644774636,
 0.5266443748123756,0.1500503996722633,0.10149075748464842,0.9450007998665825,
 0.06909557326864801,0.32224915060675774,0.42117613156579825,
 0.8818350733123622,0.8393962414345538,0.8114370247449454,0.6840285886103874,
 0.9158924691160111,0.20141908827442256,0.20561005331968563,
 0.9723333929915661,0.39091906069990434,0.5142331728761989,0.5297592831013929,
 0.9476800558129068,0.31893172586537333,0.25803449469857465,
 0.37963241977092477,0.0896800543616616,0.147580699274233,0.6866644579171826,
 0.9131997608745553,0.3941413717934663,0.7042495308992187,0.13272313316849216,
 0.5917767768945256,0.31114024034479526,0.6906850837343406,
 0.06982939017131651,0.38232223824032985,0.6955036352219497,
 0.3272811392878604,0.98240547509523,0.5057484782455213,0.22902920287926132,
 0.9729821298671032,0.15533807588221082,0.1899682809091301,0.8487661540713163,
 0.922258140733606,0.30996961924714417,0.9916992462063782,0.9737383943883462,
 0.7493255552371618,0.9526096417976653,0.9330235867653119,0.296159157559998,
 0.865171473933567,0.03651402509300472,0.45439889840604175,0.6950272799210802,
 0.7450912594900314,0.7005142950604692,0.7146652414470369,0.4153594871161602,
 0.05932909399314501,0.8896662600276057,0.8970889567013156,
 0.24543941755521403,0.2684981935577052,0.19314979725478165,0.893678564262992,
 0.9619812899358948,0.8090062231398646,0.610428480604092,0.6572164998590939,
 0.7129259420840595,0.31377896402423666,0.06868466131824003,0.865266785680602,
 0.6731583709316478,0.8216344450018485,0.2691181348494843,0.03406349568921241,
 0.7616811543016,0.023889236340290387,0.40721428261460346,0.4096078405942335,
 0.412033015275194,0.8592213250959052,0.9542806856029759,0.9030935758270796,
 0.1555818505664115,0.3136421737696167,0.3551505423703244,0.6594795119840006,
 0.7080944030637881,0.7745827214092257,0.30341279647637276,0.9253347490566505,
 0.5138413961105355,0.39542797692791165,0.7631951770700066,
 0.22561524284791235,0.03816772773742283,0.7385736947983724,
 0.1356835458658634,0.4128865455419767,0.9338130113832183,0.5326259668196085,
 0.7425688024399416,0.1158229098274367,0.3769398182830269,0.7460078556864571,
 0.7664421436489566,0.21498734148869536,0.6479978919746829,
 0.04409810596454755,0.155831679699727,0.7739004476768286,0.4080590111311434,
 0.61277603952255,0.5106634875178604,0.9025993152230175,0.7062550975995464,
 0.5783940828082907,0.1939120771580113,0.5175479511645751,0.15544647261433509,
 0.41260291715078334,0.33160323001079195,0.9019540911366422,
 0.35002807148211057,0.4170687524728277,0.545832844799261,0.291814806542373,
 0.10123581537609139,0.8523293799656633,0.4601693019174302,0.5437125715194097,
 0.3374758925626612,0.3240643221374119,0.15400301398394112,
 0.19295749228538672,0.14119731685945647,0.41041641587659217,
 0.6436723811558697,0.9897045229632414,0.9139272054905299,0.11768158747551449,
 0.6740592445822203,0.029471294332075737,0.7313583742487242,
 0.6140579213506369,0.4835136004070347,0.0010283687261370966,
 0.7267470752160698,0.9691463715929578,0.03621565839417118,0.6784395506660693,
 0.7991776545659052,0.11990179316597371,0.7453615102979986,0.8073715759290518,
 0.5169355716063038,0.5884129525758097,0.44942379672178245,0.3073399347548682,
 0.9820343377078242,0.7080982606520796,0.04594712918972177,0.3310449275732411,
 0.47840832658198906,0.10518689622032162,0.3714523293743608,
 0.06055336204445072,0.7620087246006559,0.14163032184559188,0.725341439782412,
 0.48756872835085574,0.03900666814998632,0.18772578142350094,
 0.3702953416441669,0.9257250542734788,0.03138856581446847,
 0.04968884984647548,0.029741413579607245,0.8691691854184483,
 0.9120347762072043,0.6174697891441502,0.6015582141279734,0.3897233329164477,
 0.7761687473084592,0.7468272808300593,0.48446860922646917,0.6123502207642195,
 0.8209204627801913,0.28055575468616845,0.6732702346929412,
 0.13115558420567208,0.5277943073099333,0.39321101413606463,
 0.05541411339557667,0.47040678803228464,0.06524900756201368,
 0.45756034101987664,0.9650891150287196,0.7467130160665677,
 0.32062418165641804,0.747848524642905,0.9122776547661678,0.8587008938100922,
 0.23630214307038944,0.030101217819073822,0.381377757718401,
 0.3358063072377623,0.15671120814874728,0.06411764744528381,
 0.9782627918250484,0.09474696486763556,0.6139906277096291,0.5006851187619078,
 0.9319293363958703,0.5941787797710885,0.412516371679172,0.7834827510913532,
 0.7496192928211725,0.9178083717293969,0.5531438806759243,0.9701260202649391,
 0.05295704887158181,0.22098860648547303,0.7937473214683076,0.85935680813998,
 0.4946468326648389,0.8083477610746816,0.1545654556166528,0.5797940760268999,
 0.2671410765582035,0.8085478482274852,0.847094855258969,0.89583618933582,
 0.24061809421173286,0.8846077858686983,0.6140423264396452,0.6742425675354101,
 0.6269521265788491,0.14238539129926764,0.9189463992788285,0.9997645251320723,
 0.05981215174700216,0.3164280936677686,0.1643459330718352,0.3548902456137467,
 0.5650637965910649,0.31784358893849785,0.251452704976841,0.07854677555726708,
 0.6913253735684236,0.2247329694198441,0.5089087872888234,0.49390806840825985,
 0.8649909535919493,0.8785421463898616,0.26459701485071463,0.4867508491569187,
 0.013816347084818936,0.797813645553352,0.9997978155403529,0.847729746193532,
 0.002261518231231885,0.9511723562659596,0.2997411366474537,0.987247396760044,
 0.8147944317348343,0.03797765677189602,0.2567861306093967,0.6431720747281464,
 0.28411730004236635,0.4738135901450665,0.8615725655408099,0.732415762553814,
 0.7111472757358209,0.7513624457362904,0.12203064029295807,
 0.12274773517061754,0.3966301575243232,0.8438904113341921,
 0.15119542355568993,0.8528020568343724,0.020929740490823412,
 0.17642975254119486,0.13204537911134495,0.08690648467247564,
 0.7146626036731505,0.8073514351671373,0.9393632756007064,0.8625727356320112,
 0.3730693663926371,0.5766547431880465,0.22925144614656434,
 0.21624117323788394,0.9698062295292669,0.7117784328759786,
 0.22414128014406987,0.18999880592121543,0.7524008855916557,
 0.8845957163981618,0.5275399430369705,0.6506591357096871,0.24524464348865027,
 0.6962357301070385,0.10276872934406922,0.7833250217960933,
 0.46205919053197686,0.24436668744426215,0.283248802513533,0.4511634689059464,
 0.8760612780161741,0.5667307662510835,0.804360010173332,0.22621216880574635,
 0.3412530992137002,0.5103166124311052,0.7879097388319735,0.636151129587357,
 0.1172923025105218,0.791772837743822,0.9251354210286161,0.48990520893267453,
 0.29703032333299917,0.23930554315218144,0.8891360228566321,
 0.9387915621632712,0.10288382113575811,0.5452416095969751,
 0.09641483919634164,0.28070051729073275,0.4853942798743833,
 0.9154815903381518,0.7230644449800622,0.5349903692672735,0.24272511852826872,
 0.8125350392048991,0.4308178304545075,0.6553235303880971,0.2755659895348944,
 0.6737267620122596,0.6422182580880649,0.8175305666558479,0.5466627255706651,
 0.3355248965649076,0.5052305869046994,0.03762365511301036,0.3686154422628046,
 0.5531830942769029,0.37009407013466034,0.16274026348345394,
 0.6047940089263817,0.534099555445519,0.36096117577476206,0.8703535528001921,
 0.002086072211787693,0.5801153705008191,0.6554837680226864,
 0.8649533158678627,0.21111879951007018,0.8772901063383441,0.5736756423353375,
 0.3117169909935944,0.10085918062230559,0.6655632917883945,0.9607898005264243,
 0.6761301280857099,0.39083739195551526,0.6723866632183404,0.7326662444916907,
 0.9818872233884532,0.9990709876303763,0.6030693181240834,0.7646457776983621,
 0.4665423688101563,0.15258865255703802,0.6370700468284201,0.2313065585260119,
 0.7577961289482047,0.8683605468648987,0.6765330185199884,0.17446848744099708,
 0.7061115647894272,0.28792096529379496,0.8215266320621049,0.4612994421760538,
 0.4701322924628527,0.3041920972375243,0.5468757702986713,0.7641692359080615,
 0.7209293348569717,0.43436172376084947,0.014583316526124324,
 0.7925602589276961,0.02411204411581913,0.510872211528029,0.27920145628316084,
 0.4630967571887291,0.6215255514768976,0.7615539721128074,0.7785249293123723,
 0.058002104197198534,0.5584960061328583,0.3407111310248514,
 0.7793378231868968,0.060822667667096986,0.4533841276428926,
 0.4774074104082663,0.16922381916812423,0.8066689800304202,
 0.39907751016112614,0.04404696055900592,0.625194366821558,
 0.21587237957977135,0.7490651398466399,0.9401638095713252,0.8513278260946385,
 0.82525945740617,0.038073877180165416,0.997898876968953,0.9817720892267153,
 0.6310525916973959,0.8608244370139274,0.25853386709439774,
 0.21042486757420553,0.05209584705455983,0.2941234193394774,0.835372001289183,
 0.7476902254550092,0.19058039547552819,0.8586002761220919,0.7342448590796796,
 0.21920797879543308,0.47018392877260085,0.7996913078656424,0.567358377176449,
 0.17957395685818545,0.03767188751406614,0.43051442192449263,
 0.36326000276409876,0.01646214339447749,0.918595971592829,0.728100524928942,
 0.5086440275282101,0.32180129256104184,0.5162162494808917,0.720234798610534,
 0.262980969229759,0.6660106464588035,0.5882704846390919,0.3571946569246919,
 0.8893691581454454,0.8328856517732802,0.7148418468055812,0.353909041359853,
 0.27718663229872886,0.49759270109002207,0.2507606645311742,
 0.05511742265054598,0.9628858505630098,0.6437747783850342,
 0.24005112358739977,0.12675438193003186,0.30513327176423455,0.9952027912409,
 0.49220786966245034,0.2880921980493304,0.6236273983746811,0.6932332877447375,
 0.8973376259839133,0.2332461793725512,0.4829901728155499,0.4856043058876002,
 0.0015475914851605932,0.24681778306910374,0.10222175508749409,
 0.14506088545595375,0.10009689625830498,0.6878196034052124,
 0.5996769457447271,0.14058192279382853,0.2202068578191274,0.6113483936398363,
 0.4863150869608639,0.7358847836582215,0.2746719362633725,0.570735375679957,
 0.5917026604755972,0.6956066087278463,0.8897889619996824,
 0.0050252690350632845,0.5531228571039489,0.9873657096268398,
 0.19444312879581604,0.6617810958555601,0.23188407864797,0.6565426894447024,
 0.47290706101141233,0.9315846350158151,0.42979720239613817,
 0.9313856246856198,0.4389576094852883,0.672219963282134,0.4676669011028993,
 0.09288604529766831,0.4431503909838961,0.27731020218031555,
 0.7110042203827432,0.09500646373929256,0.36009988069292587,
 0.18206621696507885,0.1971391989566148,0.39085626291626996,0.416133841889869,
 0.010094065143319453,0.2318469740891722,0.17344031944806293,
 0.908327737412643,0.4839629196988726,0.3511065989733516,0.9529168581095703,
 0.7883410778502249,0.35536013505793396,0.7960073273455165,0.9922648386516484,
 0.44831492989561106,0.9056132121437003,0.7451872470144776,0.7320817737137614,
 0.3307370334862011,0.4436462916953663,0.49253515996128505,0.7380720920696642,
 0.6167824029661734,0.4267803543601367,0.8088640948903593,0.3173495883811973,
 0.07294923832440059,0.7016091653057184,0.258302303367013,0.38495943257478915,
 0.017884751154309986,0.30465168552382726,0.2084209794263131,
 0.30098333000003297,0.5287110462820135,0.4935829988998133,
 0.21559128690583385,0.3888677752184013,0.737758816956377,0.13193604176518847,
 0.8798294631332805,0.5983174080484319,0.8297563912154816,0.9789144425127285,
 0.12397846892722952,0.6875933100837311,0.33317651012590666,0.796424731221958,
 0.38123382209714096,0.964999855317521,0.4929499520484216,0.5202489941086192,
 0.2244129850831954,0.5828570440339653,0.1167038037803827,0.9294221054413823,
 0.9352598393192214,0.7833936417820352,0.42812147243991405,0.6870250031804386,
 0.34595525038867403,0.8295289100624285,0.8943340229906216,
 0.24089774893107296,0.40661473831121053,0.8819512971796418,
 0.9096758084200757,0.5239747098819016,0.23201046323212438,0.5097941008412132,
 0.5420240993027707,0.15003025496340572,0.4159963816438663,0.8616666033727175,
 0.570911585138181,0.7108474715983347,0.34860422278027303,0.10753100931097503,
 0.4798642406913376,0.5959362999103853,0.7313472453286833,0.28609890150505624,
 0.010897140766835567,0.6031114524242074,0.5510065762456917,
 0.2236410634838465,0.07716329357272322,0.3394399374751602,0.1458558332959491,
 0.2837152749655304,0.8997379532614055,0.050438521640093215,
 0.9768633491950967,0.6008423945921018,0.9464774133564469,0.21958020296340475,
 0.047174615396354036,0.9314738472860092,0.7926188838271557,0.652949797209604,
 0.06704115540590938,0.8592586012838869,0.8026541518969976,0.3360009360507207,
 0.4637118239042547,0.15893093704431394,0.8425960260141215,0.510020903411664,
 0.6514798995081152,0.09999320515488008,0.2673171393571161,0.8900945588077946,
 0.2904495551283214,0.37070720099493093,0.6250204969427591,0.4635354828340851,
 0.2871950002580881,0.0023661488597657154,0.16775032335568285,
 0.36774255725264304,0.8252609688294068,0.9424392901565299,0.8175472613565071,
 0.9282658398615999,0.19107704494372424,0.14376853667968592,
 0.3628888183965592,0.3323714669778032,0.8315438673781522,0.9869338260708933,
 0.7757715707419796,0.2344708804418546,0.6879944006781483,0.11456998523378592,
 0.23040499212037668,0.4368188856654086,0.34370876198555844,
 0.17499572571659605,0.3275974111077766,0.48182715137363863,
 0.8850203104196263,0.14144347230745236,0.5035293252433108,
 0.11745375793423429,0.029077048850380427,0.09549053183020817,
 0.16388669689182056,0.4583966743767257,0.8223611522713571,0.3370604432804776,
 0.5478595417397436,0.10263304949068608,0.9782414597211889,0.5025791819808944,
 0.1611977923169733,0.8776265973231667,0.8697784631378703,0.664437255400657,
 0.7490854271159069,0.30218781737573264,0.6767137710218843,0.8315828391299973,
 0.578002611236437,0.9068017210791346,0.5509312578018084,0.307277109303455,
 0.2656219294930946,0.6182189654851551,0.023762213134393195,
 0.6734000650735155,0.2793068167448738,0.6839739446840801,0.6812945977525457,
 0.07784366308473234,0.7782310115116147,0.9199574542740148,
 0.31029464128809703,0.9626216061509953,0.10062944738207624,0.895635639710578,
 0.07420954342761321,0.6720748729138986,0.5626425384533463,0.6130335195630703,
 0.23164006362344536,0.005420332221413915,0.6167542825863542,
 0.18344556899298736,0.6432640720850394,0.7223074223412659,0.9339436811416431,
 0.895788829897107,0.7452422706876658,0.9101420558118393,0.8768206341261949,
 0.014031975118068973,0.35687552575271186,0.5809172806496269,
 0.7639722313164344,0.8783963571773487,0.35927764002136386,0.9955508744092338,
 0.05298086244597411,0.5864408505096417,0.7633886636229715,
 0.38450715363251775,0.6105003378818363,0.9088645501360962,0.7662848235839115,
 0.1330489490312523,0.2870124635732759,0.8643820727778863,0.19962235806413575,
 0.5439601596952658,0.9079823480836702,0.19001297367482162,0.5367131213706918,
 0.579220309020723,0.5994673748390136,0.3774960446324065,0.6379304371437515,
 0.6399977606519335,0.14316986148574606,0.6308629598904221,0.0494895095975908,
 0.25246412092215686,0.3775823906753287,0.42850838883671516,
 0.9566169051028073,0.12516088808551817,0.08977901218253515,
 0.7755818339119165,0.7533941559953862,0.4794338674332833,0.9378134888946058,
 0.40412644161429023,0.55127992971368,0.6607677684650013,0.8929844871285036,
 0.665012288455735,0.622769846380873,0.7337111583171652,0.1702286731178566,
 0.4443674556179237,0.9648563146985614,0.8443564259203208,0.09831146424546278,
 0.6634759311497009,0.857976660801893,0.7212263479574765,0.2249196332308604,
 0.8924907143370936,0.786192945430677,0.6837745793457721,0.1208591115772959,
 0.631398048887067,0.9577775737057218,0.3322673244173553,0.9781116071866499,
 0.20130328293351685,0.9375470496204575,0.7179712027833702,
 0.37984308661078325,0.3323812867167313,0.8333695443810165,0.9424771139577707,
 0.416470361897167,0.8189119753110189,0.9098347011110537,0.7443984568574681,
 0.975526722087195,0.45332644960633217,0.5010004039943476,0.7785826011882324,
 0.09269285619364498,0.8553217529839615,0.40889121769344594,
 0.2211613570288702,0.20021395628840666,0.67696891171562,0.7350961389206313,
 0.14733913761494544,0.2176563850306954,0.5135190330955657,0.5921067840092131,
 0.7304911849719997,0.10294246658173167,0.21963763643366785,
 0.7627950609415256,0.18746517232007665,0.651845010697875,0.28533853689247435,
 0.6401286058689714,0.8507112674861994,0.3878494364110381,0.8615339465410062,
 0.2509440913186083,0.12450758752163205,0.02305062166471039,0.524723078582862,
 0.8261561771878398,0.933818999096865,0.46961326465995823,0.43196232747437713,
 0.9606287213573879,0.3275720451064973,0.68077791982686,0.7131641407431109,
 0.2340701440279872,0.10613656840447327,0.6910246452165416,0.4716027675476284,
 0.17564487968350684,0.6527625198046003,0.28962889964118366,
 0.9462843624102023,0.7754927523814883,0.6102081792174562,0.3998787642932762,
 0.8407086845198893,0.262111633775282,0.11982667740833186,0.43905299443873347,
 0.45567226284679707,0.9712643269993788,0.379614156567447,0.7936363407977642,
 0.3024660922127589,0.8910313292417223,0.020072618906675888,
 0.12946041298286293,0.4133278951454332,0.9313977672129239,0.2972707257728888,
 0.2412784954863485,0.8114352169005672,0.579425335261045,0.029323010262860882,
 0.7744660141043788,0.25545135050291023,0.19175315686886207,
 0.3358122100369473,0.27134152953964097,0.03870564793807252,
 0.09062152255270295,0.19789339396107042,0.5563232767852352,
 0.05488056604311531,0.7398305047755918,0.9021937405595951,0.9347992182856031,
 0.399453309532143,0.17012086458065312,0.3444352182319791,0.5328091854125957,
 0.6239172898276527,0.4607082680893171,0.35497616767322115,
 0.11177871752431812,0.2915957319062278,0.25891108304063937,
 0.8025757540936703,0.8368288297134804,0.7985192902099507,
 0.020121898531103022,0.26385664903372263,0.8042316044970127,
 0.6688327955410439,0.06427426047602303,0.2576786062191142,
 0.12199517778551305,0.10077459743230155,0.9848504039258343,0.761658422315327,
 0.27817128810744485,0.08843654897784892,0.5501187781869645,
 0.9324118127980777,0.43817799954875414,0.9062877039230935,
 0.33226869450225216,0.15556769033645002,0.3860206298876442,
 0.8836001162140604,0.26187907632434926,0.5178685517607152,0.137278208590898,
 0.8337096743257266,0.4613689281553922,0.8394221536127471,0.9787369704522151,
 0.7612294175740814,0.1009105898289323,0.49350255977050583,0.7563681732050013,
 0.7446882284684366,0.36076615358780584,0.9220856942161949,0.7370379993902053,
 0.21098975814865317,0.04287095164496724,0.20076736198050626,
 0.3162806936744733,0.9701926124941924,0.9811877817135313,0.3268018774451086,
 0.41236588717280953,0.3174809125431748,0.8723881818762089,0.548682069760447,
 0.2707914887604058,0.03355013051985467,0.2921088530563485,
 0.33565071716186345,0.5261883168232615,0.13330355558767393,
 0.8367990128530758,0.8222442495804291,0.4346483118898946,0.994317015381944,
 0.03169672461385642,0.38035823134004654,0.9795931907342811,
 0.5875214325095217,0.5493501324138255,0.8829910397227219,0.1411444909337748,
 0.6761446270039837,0.8041009913294591,0.16338530734684187,0.9658223701364916,
 0.053884609392590765,0.6153689094897408,0.3085733452300675,0.849059256069817,
 0.7513562064456795,0.029925738778629696,0.8431189736238698,
 0.1461063296844951,0.1599690643479018,0.6129186772579219,0.6536482856929466,
 0.6432296674041607,0.7363490305315329,0.3659835601199791,0.8019028336706489,
 0.5912587547229798,0.45238829523216795,0.8114934842762015,
 0.04832354252565263,0.515049949288291,0.6766793585918172,0.15409047103441598,
 0.0782561082404114,0.9443187371652427,0.4334369126260341,0.7716440002267159,
 0.45941137824692757,0.6315864126475984,0.2445346042309513,0.9041786686358201,
 0.9044902206557105,0.5380710444782381,0.04685096049974513,0.6934561466967036,
 0.33820935442073086,0.4771747214136638,0.16912606775282624,0.259360979264299,
 0.9515650899946445,0.6968503151984929,0.4781412133961198,0.3160878053638443,
 0.9993089703403638,0.15909744382889124,0.6471890889537524,0.8726682997492169,
 0.4135223194280996,0.6391454830514124,0.9988245313640557,0.5588140245712531,
 0.9698314750287309,0.12900625466885962,0.9826854509548651,0.4265131785718509,
 0.5359568840414339,0.8161165462264284,0.39106470950947825,0.8482990839953654,
 0.925084790100706,0.8261571099433768,0.0049217393580717195,
 0.2609423884068802,0.6061032172327314,0.09520897231521874,0.561260448531336,
 0.2761711991499699,0.31410608742269397,0.7704824005066755,
 0.00841560669263186,0.1933581447104321,0.2146938988951369,0.6150979199219796,
 0.4324484028635288,0.08807884532745736,0.4979015328275721,0.8494662445849069,
 0.3801230069633321,0.42552156582295664,0.9413690557422443,0.9913013282889094,
 0.7930053786724838,0.09538869618926371,0.7379757113161894,0.671792869747634,
 0.2285687925191553,0.21916557355715183,0.7571672126651725,
 0.16197492093927668,0.6130958186081198,0.896507619695968,0.7247774286420716,
 0.5498902758133157,0.9835358444749444,0.27965556212029374,0.7781268010232357,
 0.2716861145767645,0.2203495361862815,0.48132456560479253,0.9412687719578593,
 0.07822920038555348,0.9752551385512488,0.5318758791514233,0.8784078838977503,
 0.43920425226145976,0.4143906588057006,0.9786369176821639,0.8694547810419513,
 0.4278812082843737,0.5704490623261104,0.43682717191095255,0.9073023030354972,
 0.2870766652103218,0.3548972085399168,0.1786566850017579,0.4153926387321729,
 0.5100769499883429,0.3183410971754125,0.4257963514363964,0.9273870962336495,
 0.03349620868444525,0.658399917218604,0.19721171343070198,0.7507461341037248,
 0.9557813829549215,0.37207399947998987,0.05316451803634237,
 0.8204865394171441,0.9269312134072971,0.6215123993849785,0.2310060188967219,
 0.03837007596754427,0.09812752002034564,0.36918724016059523,
 0.5397246316953122,0.7136786637050334,0.028306794407767955,
 0.3673919540854781,0.7147067492197211,0.9131834200812428,
 0.007916786531962372,0.689634707557816,0.632210416588573,0.05109595367301656,
 0.40741328130757304,0.22045741294346055,0.8881819631712597,0.519433898994625,
 0.7451901715595347,0.36545021282420576,0.8417022955820476,
 0.026114392180373258,0.9540292542335345,0.6887996854609173,
 0.4743692963050271,0.653171401077332,0.5874223091104732,0.48576710843506177,
 0.11832201109586649,0.8050514683881541,0.8862379283105654,0.6999417809492556,
 0.131059641904142,0.03985909753260819,0.8311074807668797,0.5332685418574503,
 0.879982609378183,0.9078866341328263,0.5617911383174701,0.3694553419833265,
 0.20593993603481886,0.0845036172094511,0.717500270231113,0.2453420623221656,
 0.9274341419115608,0.6385180702122801,0.6355280445271728,0.9317673284934345,
 0.4908861798307427,0.7992069290038902,0.7656427407357742,0.3984220650177315,
 0.15857692712538463,0.3796479598243041,0.8995104036518092,0.1265690856610151,
 0.44022284153116664,0.5607199794906021,0.12328669343370857,
 0.8116460130069124,0.41574710079396404,0.060255053816102855,
 0.6766344752747919,0.3344363990131394,0.650130544160288,0.10847250036132994,
 0.6376114940764533,0.48305579740853055,0.8880046080601878,0.7551356967999052,
 0.2533690878765249,0.9747614414474187,0.503144907291128,0.3265769409355219,
 0.14394564222979733,0.99008268135191,0.8984849762206432,0.45504603711478864,
 0.9864103300040997,0.48463148683581725,0.32313544210923106,0.726440670205548,
 0.5789163926232224,0.5150436673431826,0.7150519950366254,0.4284897720254932,
 0.39976096346411316,0.23858470092092854,0.07262901813173361,
 0.631502711859058,0.42024615987394875,0.4185748361667355,
 0.042941513265063125,0.21442215947572707,0.5145180183655016,
 0.1092823575654257,0.8739707202995958,0.3332725287433068,0.7031168017441422,
 0.376664425978383,0.6233452545333468,0.3682803844711363,0.432923512055049,
 0.42460843011632077,0.7151187212420742,0.5312033826313349,
 0.029893878286786246,0.1918230340058671,0.8983102576528771,
 0.2060715287099697,0.44187349295483525,0.14162247567109054,
 0.43230030028482624,0.9329158802350563,0.3963186137257534,0.7147433736614593,
 0.6065812261315301,0.5276935344398284,0.28263076077845717,
 0.48850955570748067,0.44336337483417,0.7882472235654743,0.14403676617197714,
 0.1537980070478311,0.5639490418804771,0.6110566226120623,0.39693209138707775,
 0.34992260527323393,0.94695813949757,0.96084290267421,0.6647835262335307,
 0.6977532491573872,0.008066030461828078,0.5579869960156786,
 0.024178545121430695,0.8338133563377429,0.8574894073638013,
 0.23770885233201977,0.1500191993595842,0.6197208265736635,0.7675736247666067,
 0.5405051175083893,0.3400166204721021,0.2862156058019383,0.26559519207389415,
 0.08565798138293335,0.5590158542241958,0.6806704624714055,0.9054042212546899,
 0.010337167969942551,0.36202189817733843,0.5529443342316834,
 0.6737169975390903,0.5135713429051987,0.3548020055935347,0.8693567557620487,
 0.8521055192948623,0.9920285382237555,0.5095320628196328,0.0790801121275696,
 0.4605254334376738,0.830059177613598,0.009506470026925729,0.8937207396885638,
 0.5438685321478575,0.5904315143805524,0.6097641430478891,0.8442823711639503,
 0.8668212724255678,0.1981745984799972,0.332430535631717,0.5066085507054476,
 0.9965690723100565,0.11066571448647489,0.3016907739454815,
 0.024040925096971755,0.5516657649934863,0.9589942491371153,
 0.48063052902935754,0.4280712028572522,0.7605192923346249,0.5201697911163915,
 0.10742080413196553,0.23829475418986634,0.8983204992222731,
 0.38988737647731775,0.24369360296497988,0.4217217181310231,
 0.8973154510584629,0.7374193056915835,0.276003002646672,0.8006160473230789,
 0.17646819133176395,0.5894670075894944,0.6957264116075559,
 0.33886119674132575,0.8711287229886959,0.6755757542216709,0.6638849428661957,
 0.1628458506004874,0.5257472236889275,0.40409056088782724,
 0.056598352179008216,0.35252547413566415,0.9918858945297764,
 0.7622035454590894,0.6429500217197839,0.14479656633862548,0.652481963252801,
 0.9950679299696333,0.8840287898628656,0.754599975834827,0.32451007768211493,
 0.964021726665584,0.134827614687542,0.6853048237329534,0.9473233154969709,
 0.6551107745746316,0.8473357690058807,0.8277372023648644,0.7801813399048172,
 0.7255839159381923,0.20741044265496,0.774413315470781,0.6811705924499147,
 0.3905883237805704,0.38066960070594014,0.7110795034927775,0.3733839345571831,
 0.6701019819133269,0.06797638633759939,0.6129392808884528,
 0.02934650958755336,0.2758935729446731,0.38711572172372744,
 0.2651374565303559,0.012905942562804196,0.9981446548794679,
 0.6707276736931378,0.6158471872820694,0.0014624706793684439,
 0.36306022316585584,0.03020945592053592,0.5522247240172459,
 0.28150182020633463,0.5281110607229796,0.7596778282125535,0.6211747651930684,
 0.6175630641026185,0.12997130910696653,0.39442120209213116,
 0.1942317928095152,0.05491228156001571,0.3344850565296067,0.5244686232336674,
 0.1817745032163227,0.2799874647121644,0.09769850174871997,0.8021619633497935,
 0.25538925920140043,0.7393179963317129,0.5872201316147415,0.7238696121015544,
 0.2723194358913543,0.4804216300366573,0.40406916051996133,0.6137098584199856,
 0.09486810414304792,0.6771283274670866,0.4769814233815768,
 0.11692010421074428,0.8097243380966488,0.349477046988933,0.5710212503872927,
 0.36107290806328407,0.009881607405991422,0.18722963333592446,
 0.8790370217835894,0.9400871843582238,0.7140700384693155,0.5209721890177452,
 0.17831421201110864,0.8017478943392828,0.850750843317186,0.6323250211720838,
 0.516529925136859,0.06757443779975048,0.5658775082933126,0.9219241715036574,
 0.6964606134144099,0.11396570359759783,0.06876919417069105,
 0.3112997344585332,0.37660628663304685,0.3634218658998448,
 0.23037356830779576,0.22331113893142562,0.23659856764793297,
 0.5008341694706184,0.35524176973191546,0.1496965196822414,0.2634987501923358,
 0.594533246844064,0.8989831198805521,0.1321090928909443,0.4032854259862584,
 0.8868964961805508,0.3895416512651605,0.03707452142608958,0.7380489903782307,
 0.1719146981279982,0.10117603192252855,0.5538430643975429,0.7037366837363905,
 0.6269983970856409,0.37606808110009426,0.09964838023130818,
 0.23139027431916648,0.07188389879001111,0.5943033018459103,
 0.08635918775091234,0.5067749860660623,0.7947842398053411,0.4706721138852532,
 0.14265696813841688,0.3960137169355771,0.2955095729626024,0.973529884370692,
 0.44723740606575546,0.3024141880924043,0.44712937725945046,
 0.6106951635252056,0.5897203275497241,0.887776883114819,0.06340713581370239,
 0.6686331186248369,0.9662442976318132,0.788922277363701,0.8672996302646127,
 0.5400666847358988,0.7126764608225952,0.14984018012081513,0.6858775354415453,
 0.6778979753500546,0.18959898582718993,0.20474120601058643,
 0.6776176806747545,0.3926422734985904,0.6991039121531514,0.4184743058495771,
 0.4596898337783776,0.5386949983649025,0.18915463437295177,0.8298807547434098,
 0.633418995588538,0.5493896942065554,0.4166002296092468,0.24138950608757181,
 0.4684298576864032,0.4163848262468274,0.29594163750645497,0.7097118667561304,
 0.6610120739989109,0.08267021228334204,0.9332593048979656,0.4832690336218243,
 0.5321682012630133,0.18403564496753777,0.34940144038124554,
 0.5393155765853188,0.4751277343929772,0.48364678696728836,0.8455029835757151,
 0.9954840872071254,0.13824943213904128,0.9033772442760168,0.4231198572250776,
 0.9210675402024278,0.008619261643424747,0.4166843128336567,
 0.2324668007222166,0.29722852867290894,0.9003585490455556,0.5496831981208353,
 0.5799784185513996,0.29955287165282607,0.11757554510425805,
 0.015830830630930715,0.09189994530351431,0.6204596829402005,
 0.8941755712768364,0.9329035900183115,0.6045512429958998,0.19371045826261812,
 0.6171094034716077,0.3780770368169737,0.6952652373072935,0.20881239331779133,
 0.7972086450256415,0.37548915159072493,0.1723576626858262,
 0.47118708673283183,0.03467923999206335,0.6758182067891503,
 0.6937586562905911,0.44774028226858353,0.2910789519587986,
 0.02190551087056669,0.2035021013879642,0.7210403310479743,0.9324432253961548,
 0.8438618711314222,0.8587236287602401,0.7708037904430771,0.6967633232095914,
 0.533311568536334,0.08157778254454673,0.3637301722705686,0.6198680834998128,
 0.6230673010395329,0.5626248380789319,0.5150108103924237,0.41774151108593016,
 0.5995029010594151,0.6332032442426316,0.9564246272895259,0.498317908064279,
 0.35589569878939675,0.46705972865816525,0.16085570302753194,
 0.9308337221068579,0.2282176950632191,0.6129964724874746,0.07419035800319884,
 1.1183432357497747e-4,0.9435964893492904,0.8356127933415397,
 0.1994503178772583,0.8751100883768506,0.07012601407992536,
 0.20244911975897684,0.12199028752047802,0.7311584728102983,
 0.29541202522204535,0.06073713160185579,0.82545264897701,0.29776043900019256,
 0.9312511783264922,0.06686041235087847,0.053616982518540524,
 0.8915568221594857,0.4138778786519798,0.4245775376478106,0.7422797773769175,
 0.18297888491475078,0.6635908855913778,0.10924558989939426,
 0.13215548451117243,0.6851282559386085,0.5732853314153378,0.4325460205217956,
 0.05718843883632074,0.5119458447708235,0.05649843283228628,
 0.17438359481870247,0.04711364027295917,0.6777508524464795,
 0.8711635409107262,0.012840955277106936,0.5870595412757379,
 0.07220061398274547,0.9582009582150521,0.47269180713581327,
 0.4582500478002044,0.31478754186320435,0.23405470085561608,
 0.47578175835960723,0.689482134926295,0.30149718801693315,0.745061244486283,
 0.3021716618915704,0.30472643529133847,0.48714550887917063,
 0.4308219706066029,0.8333221932291787,0.40937927945236374,0.9687594604686839,
 0.5808133665477351,0.35086690326517633,0.268052092166309,0.45174000394108416,
 0.5860893524625956,0.41724363153832156,0.7906891998640307,
 0.05669390770123961,0.7097937182826879,0.8093768854268042,0.8580301601945669,
 0.40881824662033917,0.8478104998685891,0.7314312175372446,0.9847603786949549,
 0.7077652703531341,0.8119427047253424,0.024832155035199754,
 0.08990146895956685,0.3465196184483855,0.39833733740438493,
 0.3074202576098659,0.06323208647554024,0.8905470994677365,
 0.026424475738361597,0.19159974807691438,0.17907479076171373,
 0.935985934859952,0.773959206956346,0.5152144851247026,0.6927285991337154,
 0.7038643941662623,0.5283114811665631,0.3902508936385244,0.31800089047192837,
 0.18157964856592568,0.3070986072705909,0.05976736446267539,
 0.48538647682752467,0.6522312929740445,0.5647789686569026,0.7292482766068314,
 0.8875330202621936,0.5962914033723241,0.13606677784692134,0.6568720156737475,
 0.9170509998201948,0.9870727714418556,0.2446838627499539,0.3296320021853364,
 0.6852251980376082,0.9826897707299235,0.23179551264526843,0.5828082644469286,
 0.3901655877024527,0.7305103599062005,0.05454274068341325,0.1301985474639138,
 0.34886764312824536,0.1852488367712577,0.9172452426943145,0.9922934959227995,
 0.38995727390117496,0.6628460223438698,0.3983656890570171,
 0.14807125798765453,0.4883296624802781,0.7538460541005481,0.5533238286286056,
 0.20664467841946998,0.21653105675981243,0.8245319540141359,
 0.7104598877775664,0.5300021626228024,0.11190971822874529,0.7151755010583568,
 0.46490077925525575,0.2495813980190753,0.31372184368632805,
 0.9090725694869572,0.15320293587998002,0.9891345101331721,0.9068973114742229,
 0.12924808105625285,0.5157244522429837,0.08851213723411211,
 0.4580817530417537,0.5627673219556625,0.04342702922883701,
 0.42854227484867835,0.2482884157102626,0.7466604130187602,
 0.40366061672981624,0.7190543290713933,0.3598660581941715,
 0.27645483898311785,0.1004837905732997,0.8343912296863131,0.4276242087001145,
 0.7885279250353378,0.7638423807438189,0.9648548374853224,0.49635588372166195,
 0.41662617021146,0.9164738973416653,0.16583051429938767,0.8973467397293184,
 0.8350548321118931,0.5718737456462469,0.20899600878725078,0.7545668436742154,
 0.24973426038997393,0.18261629161249104,0.9132291403840088,
 0.5563080716981379,0.8782061283795878,0.06546740675294593,
 0.16046362929689184,0.9492454049180251,0.39111402363085035,
 0.4907473761394009,0.9553645523357727,0.24352607613751165,0.795421289205441,
 0.8441617794134763,0.6015612431351604,0.4402135823188569,0.5922669482033351,
 0.2493512003659718,0.009103840170917143,0.27452924405617696,
 0.20328789767697386,0.5881971623266788,0.399461627791347,0.7429651925114052,
 0.6842359377599506,0.9764077317348103,0.9780828048119209,0.18986754936193917,
 0.5018538020065201,0.9402883921790092,0.25619617901939584,0.7164241522786027,
 0.6099515211795663,0.4956092388659191,0.41891240560212184,
 0.06837997122778983,0.48605738945055976,0.46839328517807355,
 0.3738445716303599,0.08040967627204054,0.7739392304556303,0.2605375068895929,
 0.6625493783169645,0.5093327941031205,0.699481303243224,0.6985402864219449,
 0.11397568562605442,0.09029826636429905,0.15797872712725258,
 0.29917024067286624,0.9114478586880664,0.12330672432177003,
 0.5822214374434361,0.6631788916999924,0.9280405677338404,0.8322775398076498,
 0.3572475604544745,0.181022436691459,0.9598540808439837,0.4915732949684899,
 0.45813098671260866,0.43738362808933895,0.2576557548552244,
 0.3847145332350319,0.8550735693917281,0.007806905427654609,
 0.3198233002433316,0.5911769668125948,0.9541764097220721,0.941860522267611,
 0.48555094314786273,0.158757765931842,0.3048134099137293,0.25645918033267445,
 0.0927625842887504,0.3758869911328685,0.9353342114833706,0.2997936243858216,
 0.499111458135854,0.39580381101954376,0.9344756963370457,0.9631164592585414,
 0.47541472304305465,0.008231359219492806,0.5927780788232709,
 0.05676571550405474,0.5915194781622886,0.8048920400870885,0.9889328720793559,
 0.8341914095731807,0.14163154606407635,0.3896704234190079,0.8985563327439424,
 0.2358533940160289,0.4476263037718038,0.7841932195355268,0.13808775926112793,
 0.9006275365384817,0.2776408346132777,0.5038922604898174,0.7744267232055172,
 0.70539633285021,0.8366437283771668,0.9646398975810732,0.5767763343341507,
 0.2906385827245974,0.7473471787375816,0.12353200192222524,
 0.04543476817922143,0.11990045345246858,0.16826525064379272,
 0.1642716299069259,0.9630303910713317,0.9895878552180517,0.5986416029325545,
 0.932609031412673,0.6703352990009381,0.32725399100652897,0.9795398248840178,
 0.3879984286861323,0.8997240530345656,0.09460001685911767,
 0.023560698443704897,0.8541930913657024,0.43916087259930703,
 0.49797718869723484,0.21342350099141305,0.5024864888549013,
 0.1005376391397661,0.06177622938123095,0.5384816848576676,0.861981825850064,
 0.5921008445699885,0.34431204279336436,0.9831590117802871,0.4713606690104437,
 0.9324901147908293,0.715434513769569,0.11493965758140068,0.2593584274404409,
 0.7434832584039305,0.3678646492091583,0.3198800594859177,0.43040780327366945,
 0.6470920301886645,0.4433660369683634,0.42774616172304103,0.9718953602301381,
 0.04639143704704651,0.17216231981233587,0.8206859903907143,
 0.2408276919471718,0.9027162634499126,0.025459259252634925,
 0.2919697094797571,0.9753075416983323,0.46646262001192285,0.8434765650147377,
 0.6806410485050346,0.5637256255890326,0.4040901582526515,0.8112550144603972,
 0.6870372186968989,0.8616642917619614,0.49069245908250747,0.1716734623145787,
 0.7958169512223631,0.7297844716748151,0.12597599614184873,0.788485942257828,
 0.3314537362117308,0.5004752350514264,0.46647403034195567,0.4096654865310184,
 0.9338158920980117,0.18473812751274754,0.18352189943576347,
 0.09429244009943583,0.38800064180853333,0.9181666087620939,
 0.7143509770606838,0.2631193721745301,0.39958511998278545,
 0.14968172832594484,0.055588188059384924,0.6924755032615173,
 0.44317139871579836,0.9803938499267079,0.33574198097282437,
 0.7566317441518704,0.135592062124132,0.0922078437088909,0.5275804596383119,
 0.15870788377275902,0.09681377015186043,0.8364038203104384,
 0.7094142137192551,0.6416207768457264,0.25009993329052715,0.6306788014549943,
 0.03545847934961677,0.24293235148392767,0.6740305173610213,
 0.6502177686383308,0.331231453914857,0.8502233295903242,0.7404648202925508,
 0.03845575323568706,0.10924834280724438,0.13195116319835098,
 0.6746960188795971,0.9092248481510017,0.1400954599335409,0.8266335945149653,
 0.41976007877261234,0.9455260691620773,0.3076345658667792,0.297672820730423,
 0.5496464902094707,0.28642743410937666,0.7025774217379562,0.9293936293984761,
 0.5147914032732994,0.6514939107214507,0.11499837724585005,0.9930040096133281,
 0.17057795446438218,0.13777525392185608,0.964112085593258,
 0.16281770157503717,0.7855332435969161,0.389596068994459,0.1895378529234818,
 0.802025505877424,0.6226994979711526,0.5414106702069418,0.8510904698388654,
 0.5037023032012732,0.6141750518222242,0.22608225671236115,
 0.24255293223305574,0.4635492359541111,0.3504218915923467,0.3607607095350879,
 0.04143132603223776,0.4833723601748686,0.5782405838251539,
 0.13905437815946453,0.3897269338873981,0.02772985981539422,
 0.8173178434105471,0.9470734856447689,0.9153516739882276,0.7371795205094023,
 0.519469095312894,0.22714526566352444,0.5813459011564576,0.5554680945079247,
 0.8748945396089683,0.6054306366942988,0.28761207895234,0.014641897865948383,
 0.4682703528815847,0.7665601629810287,0.31197306373183786,0.1739647855609796,
 0.2753512685190679,0.09654041467224206,0.14352136146467198,
 0.6366246125679093,0.9961258671213142,0.9232116962097381,
 0.045881559753804924,0.7837159280832927,0.03378238154636792,0.52128313553125,
 0.29244580480416904,0.5242296469823352,0.7450082214765096,
 0.35047411832919984,0.5432749737099589,0.5200147303793871,0.7277583145690762,
 0.6812185187602535,0.5542287934970491,0.4543207764739977,0.8057895606705242,
 0.8952078284478789,0.3689098910062263,0.9971161503440078,0.6550047391763967,
 0.725805152095378,0.1735372462767546,0.460036110781224,0.5007593750806218,
 0.2910634224134947,0.44661000660789285,0.7341435752714117,
 0.05852399377577355,0.6134209183172594,0.6500085203558079,0.6686009572887166,
 0.6853140830960013,0.9330259645491137,0.9723578252158755,0.8754493221557127,
 0.4933536968286091,0.822341121597573,0.8366162233547842,0.35509279737590016,
 0.5200834515617837,0.8656974370316343,0.5975930121716221,0.10454783846940141,
 0.8296579893368459,0.44363480101690644,0.3868225610814918,
 0.18334227154509597,0.440858981579917,0.44321477280467314,0.9234576065068852,
 0.6534797215245298,0.4073323623706543,0.24460528294604267,
 0.22710820865326398,0.7489202486829655,0.60847632627324,0.9884016448077689,
 0.680443473382184,0.6941350635511029,0.0714501408383067,0.17843379512383895,
 0.2865229977456212,0.16699683322935532,0.1226975323609959,
 0.058079725112694214,0.6518502295559427,0.48750005586864653,
 0.4429818243731882,0.31653318165601096,0.3565467952523381,0.7333988866443895,
 0.09797763666559378,0.9506246237549263,0.932696138464785,0.44158201227085536,
 0.12541843872323444,0.9228309074883547,0.5189323434959779,0.8887529836079854,
 0.7904229969366584,0.7790298388423159,0.28050133743944516,0.6810297912729586,
 0.2114294784061429,0.5211531340477865,0.5757502857880036,0.8492597020324806,
 0.08401444904597699,0.9306690488522745,0.8844708450787567,0.819967221854976,
 0.2568950615961574,0.3115546936312852,0.44731783358001775,
 0.08794787668281012,0.943347189672792,0.9945853440405572,0.6663418130005396,
 0.5421047953437286,0.32705782230200753,0.14278085293384213,
 0.05388968857152521,0.6502325555037423,0.17943955174008341,
 0.6119557112579005,0.24499239548488128,0.35176630731870406,
 0.8758750515071854,0.010475281721639917,0.9992857940674192,
 0.8149638832184538,0.4873216934650335,0.4207482386523388,0.20830772053948132,
 0.7797839784031175,0.01942255314116781,0.8576255559586059,0.9807931676943888,
 0.6025658698966125,0.36363771224111463,0.10603235460787985,
 0.5268814313084058,0.7213488380564689,0.27368787378572135,
 0.28854755940599885,0.8226725259535532,0.4436493690986125,0.2270074460600704,
 0.09198484976417876,0.036436463845949674,0.8797665754774263,
 0.3568063276839408,0.15395641811555205,0.1436596297050795,0.2899333066034572,
 0.3544264279642715,0.516400328540955,0.08391762737561592,0.5472175245982818,
 0.36814312824957507,0.45214159377942276,0.9409619116255579,
 0.3946411105811921,0.5322091563775865,0.09724659055059637,
 0.19879512949836298,0.6606981812251336,0.18812940472535833,
 0.9408346933193676,0.4839323693778441,0.7552851602400905,0.1761999114481827,
 0.3919027571287904,0.14335564753441155,0.5788949797860767,0.7175817610864161,
 0.13938609304038396,0.14015145402674478,0.7567096632209316,
 0.5349582836335677,0.4770161451518563,0.4805648155513317,0.8593270195295407,
 0.23102220528688344,0.8447347802850603,0.4569785421834597,0.396993853376132,
 0.44363287858326805,0.8691209617074425,0.04776008137479071,
 0.8891319583774325,0.10441968950608238,0.13920961057228498,
 0.2542638040923163,0.26091360551562925,0.7129328059946728,0.7985905174248034,
 0.6525119631029931,0.6589446587683448,0.6332140824409879,0.5300188340217338,
 0.8950569049220252,0.2745860132568291,0.7246976429849746,0.8412944564787207,
 0.795249032066051,0.5346758935370368,0.7168425495141401,0.6546453503670421,
 0.006338835346658467,0.41925730679097073,0.5505602983747925,
 0.7319543353458617,0.1312384006994488,0.6597062024214464,0.7027978433983563,
 0.48959317931382473,0.9976970867337304,0.18994228654597134,
 0.2582180655392148,0.8145903647525197,0.5097849465014065,
 5.080212657027072e-5,0.16821917924672214,0.48596410212060603,
 0.15222221711081496,0.008262767151605233,0.0818641871751119,
 0.7793625033766324,0.8547247679943731,0.29810062026390016,
 0.18152118874983592,0.8630645439888907,0.6173050412370717,
 0.32720673964964175,0.994504632695056,0.6767905274788537,0.03697109507381502,
 0.9394632170047217,0.983722045158951,0.12803080901348318,0.34435666644914176,
 0.2604888134811344,0.21237982717101844,0.9962012456316377,0.7938253996759217,
 0.2344813151011853,0.6412529779633459,0.3017680226490934,0.37533832938091083,
 0.8738280956555152,0.20028031261080415,0.1886988481806795,
 0.056987508728691294,0.49864586297429825,0.7549310518705442,
 0.03265174577345287,0.8638891725219844,0.3106963509883709,
 0.011245929746552474,0.30188420077693656,0.7176412460822583,
 0.8808810757067697,0.18020277464643686,0.9431435928577238,0.4015748170119754,
 0.7259209158207127,0.26857668828996994,0.4118347727937123,
 0.12959696368306584,0.6882690218604914,0.7207380951159573,0.6032217739790512,
 0.4567609926876016,0.4938055443008308,0.4306808986215158,0.3104159191770942,
 0.7383186888200445,0.4715987561064271,0.671317434276283,0.469488015289104,
 0.5787582807993287,0.7622001351657315,0.5977498860550204,0.1799523142955325,
 0.46785820561257163,0.30972871787187617,0.8131705251035957,
 0.7033035340538252,0.6002666302917979,0.011146379544749174,
 0.46750617038566755,0.8321903247791791,0.7380947400533922,0.9064575713539506,
 0.9990306667610002,0.6471395621390874,0.08661463841054995,
 0.02540988422947188,0.7372989106826435,0.14097051140704708,
 0.6014504364400574,0.3114114424953006,0.005786974356036811,
 0.9182360983558226,0.8425697515607884,0.2588169022215413,
 0.015435437505073035,0.8391934699399546,0.4887868196453681,
 0.22371558209809606,0.5206809553140526,0.3543563537047242,0.6567800320906548,
 0.24241299437817942,0.6984554952140153,0.8340711229003392,
 0.25482987024157666,0.7010824027465979,0.032809127250114756,
 0.4164855460873599,0.5070354332318863,0.8772619446398299,0.5982004351226022,
 0.8265498489239689,0.06592359530346592,0.017551714010321517,
 0.887150945070361,0.06322466138982308,0.9519371982177242,0.9029591213237774,
 0.6913558631652131,0.14724341998324397,0.20988008364381494,
 0.5308060682092965,0.04209635661162947,0.8248672694872932,0.8413987772918818,
 0.7036821878062246,0.9775068686322833,0.8372173438116671,0.292028290659883,
 0.8681911503182083,0.9901724759156196,0.8052423917868654,0.4554911940791442,
 0.23321635757486114,0.13488891278808934,0.5429513071044473,
 0.9049425593858672,0.33575986104055355,0.7615712712330518,
 0.46922163816234697,0.5073440140740103,0.8098446164743703,0.6848311890306971,
 0.5133905285689586,0.3272326742121563,0.7000133960086251,0.28324142049435386,
 0.8370413934044701,0.48935363023010214,0.3146730879116755,0.7631526922969688,
 0.6601295286528804,0.790889334184409,0.40649173888277024,0.6993706408537593,
 0.2763081506558691,0.4316773673491907,0.6569892399518859,0.5795727898191532,
 0.5587147168650651,0.37272183093449274,0.7412166104493981,
 0.32277424326095394,0.16223077447293788,0.3152530442819774,0.652000125650203,
 0.8200907272761148,0.1663551331608647,0.7984220470882888,0.3639445123718261,
 0.7673665592459674,0.9336415735672975,0.7162631128910314,0.588252523848746,
 0.4039702927712291,0.03956888042608564,0.07045847346189826,
 0.27125334622440966,0.0761135828659395,0.21605424177261956,
 0.12206320757924072,0.6017780656151487,0.9355768875584314,0.8060942888576252,
 0.031021851450943605,0.7789541550882898,0.02890283040629349,
 0.8285076667506854,0.40812102615803525,0.41941242646900645,
 0.28934663349433554,0.5882494963074532,0.19679425218376156,
 0.2879500263149938,0.6066686449148101,0.8986665016446599,0.5900520987039396,
 0.6540510871262223,0.5106682839003773,0.9476203841120384,0.2735728710226355,
 0.8901608401731572,0.7556449281592241,0.3864562015232911,0.18579514577080858,
 0.30258127951610414,0.6282102159965506,0.8586888032372051,0.6340152032675678,
 0.7970512674017693,0.638078312729531,0.3694974683066855,0.09809209158141796,
 0.11886920311153437,0.07015196182339956,0.08158072915573444,
 0.6347014434953709,0.8053041537780107,0.21239907063989338,0.8096923040664761,
 0.6240782794888394,0.4520382661065476,0.03504627804299243,0.8382518017488545,
 0.7963045118120915,0.5092191690897199,0.7706256607307866,0.7595682920948369,
 0.532983559765125,0.9552363448708795,0.7508619602017693,0.5052881472431127,
 0.06808284999294034,0.440362587084728,0.15790576472686235,0.274343971445725,
 0.3601998562883679,0.9766370951009561,0.7486806641904473,0.3937880755477594,
 0.47021012860115763,0.27332827777242397,0.6469673583164632,
 0.05805763782376838,0.01532714710119576,0.40174514064673206,
 0.15608175762451493,0.05132028098755792,0.6374422964533024,
 0.6662051520363004,0.07288443105303655,0.3223475286437375,0.7782310598147674,
 0.36850432214884554,0.7613886189167625,0.19672325695954362,
 0.12051564819148297,0.6533371759642392,0.6247006091046965,0.2456803462437951,
 0.15746292474006207,0.47061480001104683,0.6585966968193885,
 0.9963246916911465,0.922280714853816,0.1737851794779146,0.5581857202406407,
 0.014107669622283603,0.21921027848920305,0.13405800850542082,
 0.41114945798659797,0.07372898755473856,0.8722478292261675,
 0.4292435798378277,0.6301620549021894,0.2026792373735996,0.5770471868398489,
 0.8735606259472127,0.7643123279839229,0.09156660850384146,
 0.19098135040564712,0.6279261917498046,0.40217105113351603,
 0.2853377834465748,0.255117023801265,0.8476060194380935,0.31993160665065346,
 0.8985378602838002,0.8499303942572352,0.7931934882888322,0.5692972173971161,
 0.19948145164312636,0.23649895564412582,0.6496274375522733,
 0.7358958680923056,0.07139978753611453,0.6217505034240727,0.4506919810986093,
 0.414083351288562,0.3018023910574381,0.017613866395540634,0.6569854010187787,
 0.9527476937794477,0.12553631285095979,0.19246181825357678,
 0.3881817839934387,0.024791752086203434,0.5540410898360286,0.710245521721707,
 0.267865833968145,0.2396143827748991,0.4380440429172421,0.9444260291449884,
 0.41801241289026214,0.6016305597156625,0.984105819037694,0.5763572742154155,
 0.0076311862022698795,0.5824340603680496,0.48145134095992237,
 0.9439764525896005,0.6451392427935387,0.24098419063343857,0.8840549977678076,
 0.12382591628559947,0.7622449710943953,0.2934329735578325,0.9619389047040475,
 0.47176905119489665,0.7380762474446788,0.7576106149290951,0.9281726190061145,
 0.584437906486297,0.8292928419183287,0.6374950523272618,0.5743696625129362,
 0.13017601478934704,0.9312276521938807,0.3611604191412606,
 0.13223182664375355,0.3597776068612566,0.11362024059200537,
 0.6975864056043253,0.4857041804607328,0.8991629502181341,0.9213224344933072,
 0.4505521524533671,0.5943923636486068,0.5652950607446663,0.5273167812915446,
 0.7512072680476929,0.4851058790043479,0.4682772173806694,0.9563039396613476,
 0.25087444316299456,0.26299988020510234,0.3837341899278749,
 0.4381069903685414,0.37371065947639936,0.5250462086559431,
 0.43289477537939725,0.43981869001406415,0.7627993379558818,
 0.029504771581626787,0.06841445242804645,0.936902788638205,
 0.6841395156143978,0.7770590200885015,0.7704999297006696,0.09841317476556144,
 0.08325728767312546,0.5166529047606198,0.5281135562771707,
 0.46673574855209354,0.9385501591906443,0.5024785073642415,
 0.17083950415229088,0.23143300863055183,0.1353675839307561,
 0.32456171408948864,0.3241560506807686,0.4330034458266705,0.4052200104643855,
 0.8587288919572669,0.9941871842684172,0.7863723026944887,0.09809886701717718,
 0.7780286460320612,0.816628276704968,0.9239207295432235,0.242846596126975,
 0.17932831319164433,0.2895150204003244,0.9824890077472097,0.265698558627813,
 0.14392656958191197,0.8601913433972467,0.5427597889165461,
 0.057687076732288234,0.46508036378512707,0.8710019168739314,
 0.08401189681224097,0.37804700688507487,0.27488857277117096,
 0.8215669522976421,0.6204284510220116,0.6009868519480126,0.18300051289369346,
 0.6073741492863494,0.8518144587196051,0.6006775180234422,0.46364491233783767,
 0.4093384089456428,0.8837959238560018,0.4759023349232575,0.8729133883213435,
 0.29897589369310684,0.6607532655174015,0.5945405912775162,0.0469421070852738,
 0.41859729122262956,0.6325820631293934,0.9893567875386067,0.8164097347176067,
 0.2661595904015902,0.9238840720752166,0.8709055429498902,0.15875394326291625,
 0.5791144108790657,0.20945729240620203,0.4927260801032791,0.8514550957093527,
 0.29283769528741843,0.304201886890196,0.9018847972454589,0.188324091448536,
 0.4046364394729449,0.48187798508846524,0.31658909083066833,
 0.26557594425527475,0.27026452574980175,0.5128537039610072,
 0.47870432818242303,0.19862411996635254,0.12535872480754073,
 0.24678550613783545,0.8635989731259603,0.8324447061028711,0.523967306934748,
 0.21541961351581485,0.5663164518039064,0.13938577622314663,
 0.07099838150103266,0.2401018015954326,0.9694110213765323,0.9669418022581932,
 0.4820730653234526,0.9961514426977318,0.18152009333772046,0.1593951148867645,
 0.2991870119340242,0.7450530144357179,0.4362060990005856,0.27213823175875607,
 0.6142734919360844,0.22219219285222513,0.46764477180072617,0.843976770347085,
 0.448252465273244,0.7472614012629483,0.46184888060527174,0.053365857463884,
 0.6173138943716996,0.7969219102995021,0.6277196776722873,0.2933374407055417,
 0.256909879514726,0.9655055876215174,0.9095197031579595,0.4850720028398565,
 0.9478147096849262,0.33428833915337786,0.7361705436901561,0.4475380008811074,
 0.021718288216186044,0.4275673401123766,0.4812301704074413,
 0.6437147628728329,0.11883538557853401,0.38028166141345765,
 0.11492418253449521,0.9142411005835497,0.44103597433517894,
 0.8024766156758788,0.44166625922327607,0.8463279239710835,0.8263376120946458,
 0.9949331732019473,0.6433673998508037,0.6834507363228872,0.9532236497494964,
 0.7993064146469951,0.5510921697141202,0.5624818172422626,0.13073594469266514,
 0.6886076721974711,0.7360756669052863,0.9317711424416484,0.46113631864669113,
 0.41093336811811065,0.7755670459158261,0.319150834285854,0.1868813517011494,
 0.5891814241156073,0.6197386494797451,0.9122751474332197,0.8020055284096875,
 0.11313877337684508,0.11506460320238032,0.47873298839783485,
 0.6936851770968737,0.5976374160758476,0.5285846146774776,0.18904259784246613,
 0.8319402976123175,0.8515374148047389,0.030950144669739665,
 0.8752018353422197,0.2758642066306628,0.6461778903032914,0.4463091405317843,
 0.42187111910005193,0.9712615394210916,0.11614975602351274,
 0.028200186397529192,0.018177593205930376,0.12036351444894655,
 0.1710889292769071,0.7257564585961421,0.490997649045933,0.28394856169087956,
 0.31760780771151265,0.8248688185089282,0.6642487112803763,0.7993736221330688,
 0.5260741422725607,0.7282362557365472,0.4119004770606851,0.2584604830853564,
 0.3960147672135239,0.9053826278794121,0.9893120990647841,0.5543939401049529,
 0.05543177015152023,0.15812530796762303,0.6365793991861151,
 0.5669213672045617,0.6624123462315626,0.5892073662163818,0.09188596153907258,
 0.6580199193305089,0.9188235057988813,0.25749054617097966,0.8678848315425558,
 0.29847281818920435,0.732870892647346,0.48541213276666406,0.5909977213082576,
 0.33001402678425595,0.6159838205893495,0.9047754273445234,0.3785112569635566,
 0.13793059552014064,0.3556653284335539,0.506131185431228,0.8364513778092642,
 0.7082627929254371,0.3501955714081052,0.8080789799719568,0.6605563091632681,
 0.5951855036830968,0.7448578058972918,0.7535759639294334,0.9350406169905764,
 0.48070152315302717,0.18899836876612075,0.6688250409157958,
 0.3994763773224729,0.1294480840881962,0.7833448254093976,0.01598507624986556,
 0.6149513575729154,0.5927428125480437,0.07512143570923446,0.411153788924973,
 0.9732366601545528,0.15508385394573132,0.7763740323007937,0.4831709643558133,
 0.5867357609630955,0.4580991044381808,0.8376463600443321,0.2947659219967982,
 0.5003662903634373,0.15324882622082148,0.3819306974026986,
 0.0041036472772036925,0.5155218095466779,0.4233567160893721,
 0.08209352506192436,0.8951085358017778,0.2249883115644844,
 0.12580248712256747,0.5956570893992741,0.9883935554304227,
 0.08496017978447934,0.15489172325689404,0.5527779099786801,0.619974925743268,
 0.38045501844060226,0.018910505260867794,0.19809950593627956,
 0.354713123158128,0.5601254907297633,0.4630452148523452,0.5046243510844123,
 0.26201845508073285,0.6461338820425941,0.7972908081302698,0.4520667376832427,
 0.06188624919390151,0.9110847623878616,0.15348340112882064,
 0.9771232277257474,0.6195675892242637,0.6139683734172552,0.11696647632491641,
 0.6295685656579539,0.4566501534362881,0.7045023377055781,0.22302934880798586,
 0.047637663577724365,0.4359749407771164,0.3099986365397769,
 0.02869955652299927,0.13890546253185243,0.6413169021822871,
 0.8769718288723762,0.7593834742647405,0.9335372796462779,0.880683940031606,
 0.05539894846532489,0.6920219901913753,0.7632577441845484,
 0.28353304459861617,0.07329609815988225,0.9298786101819719,
 0.1885210859592671,0.6567889293647884,0.978922047446488,0.20600343215312145,
 0.6979339496025586,0.030441135411618858,0.9697452827071998,
 0.9493286070665972,0.5608953728541798,0.3341611758333024,0.8419254646830108,
 0.5390460591288694,0.7273268620672604,0.2521042029681041,0.21349686557391956,
 0.35594209642401875,0.9027336277487968,0.44328753606566584,
 0.27583260072451066,0.1618183196302192,0.2885604666122098,0.4683632267842194,
 0.9481122055946302,0.3444992990955381,0.23454454900434052,0.9353547386252804,
 0.574049227187635,0.4600327770682333,0.44227170377541425,0.21272900918795068,
 0.07572477330314875,0.9884857943798278,0.6481218082408596,0.445611374154694,
 0.7886200178650913,0.4185837991718603,0.5711999030140456,0.7822635873438727,
 0.3925669658969789,0.02599637494352114,0.6240731454896812,
 0.056394083370987835,0.7013621364277565,0.01263282206171934,
 0.7978795218025125,0.9372656687776241,0.5164619284652732,0.715410951091354,
 0.7886936994012942,0.4024704611556751,0.7427801880053597,0.11896525133537639,
 0.8864655583266605,0.9086418398719374,0.849247318280482,0.33137946307861077,
 0.8715969393303954,0.8313747629256357,0.014530245538288966,
 0.5534496790191423,0.765967021336541,0.19983958671912405,0.44680761986490747,
 0.25051745998641184,0.2291522770401938,0.728991503168143,0.25326661266343786,
 0.9165153748089498,0.7094639631707319,0.268992284258912,0.833443431677543,
 0.3997267679358667,0.6887103063847909,0.43665063926152203,0.4375419924826972,
 0.6781337911922071,0.23792801918466966,0.9556175750849139,0.3522622546082871,
 0.9285668822846271,0.036647177520725105,0.9194678930147357,0.49781499084838,
 0.9521319166302346,0.13035061529968273,0.9756684897712727,0.9403767949496659,
 0.7479924448758115,0.7159725172026766,0.7403556816145168,0.9392291410383093,
 0.7884657864288629,0.8448077504564158,0.8527782293535014,0.39667456536573276,
 0.39774833042096047,0.15715163782640706,0.4985257947331574,
 0.8768523902775975,0.1637810814256131,0.36293183475271096,0.7739078094567197,
 0.7107642657276719,0.5974387170400657,0.028063344200690454,0.698668624624852,
 0.11723977703694233,0.39913783032051786,0.09999643802922864,
 0.5032031997127749,0.6378470125690605,0.4741834295663241,0.7801578220458815,
 0.36058948168469607,0.2977885027429148,0.1813506825861224,
 0.04688644666276931,0.8129846393013,0.7584894168652121,0.30497980879644093,
 0.01677546181411904,0.05381020972750539,0.5619228138308093,
 0.6743269534024605,0.15061297114968175,0.21956625082650227,
 0.1730864117472518,0.08867074273326148,0.4728532566856829,0.8248583011760035,
 0.7423451763622371,0.8445367958552036,0.5790534037781473,0.21028768349229066,
 0.7562687186776744,0.6742252897770515,0.6572757436011685,0.19595436738135874,
 0.7646535438606217,0.8818538189241596,0.062302648500677904,
 0.8049984744661891,0.11024191853570242,0.7515050192260643,0.2837965567181371,
 0.48175969479799585,0.5424841512914235,0.5679959269951451,
 0.28082573114331066,0.5906689540634471,0.8582338974472952,
 0.09192994102148122,0.28285107815804267,0.43079025751483924,
 0.8993020962120785,0.4761336482260776,0.15407511388688022,0.7693434047335073,
 0.45001156939365117,0.5766509927612743,0.8611997720685884,
 0.35623170380454516,0.5232698577030974,0.5425402157227159,
 0.035306817116765754,0.5352315888297084,0.6026089079475982,
 0.3257187018751716,0.5313383002238107,0.4313726460006465,0.43241920256464295,
 0.2910953000390263,0.0897113983327642,0.6145802750111725,0.39796744626438707,
 0.5638993364621301,0.5591510862713863,0.25356321686985306,0.6588465290328962,
 0.8791746741350382,0.6268688347712951,0.02969047565997207,
 0.07633213278291684,0.9259380588399229,0.659979191594018,0.6278462608291435,
 0.899470067278291,0.3681619662478349,0.08869137382142078,0.11053031252016832,
 0.9568834702008667,0.04967693984818011,0.49678252997363204,
 0.6775263084892741,0.21754060615815796,0.8663905336006475,
 0.01736685295999174,0.10307198985344057,0.8362312311096626,
 0.4115725552902183,0.9758522850886753,0.14392651100847498,0.5348305212995753,
 0.16048234100830305,0.3747461233946997,0.16148449590472402,
 0.5106447633796488,0.8137353037715597,0.6565678627683037,0.06039362660059544,
 0.01930788823938312,0.6206493948248668,0.7639899202477963,0.9973836163811522,
 0.19832734223067994,0.9653175871608157,0.9857614501686676,
 0.38449664147916784,0.42422229792558386,0.39451595069434964,
 0.8014039867949513,0.7660829530785096,0.3184625178645345,0.5318230673953871,
 0.5131553832158653,0.5167992174409848,0.8815926489226084,0.7679725097321873,
 0.14626146701722964,0.20343740265320087,0.21554971218828978,
 0.11881470596426091,0.3569384189330258,0.8492982209374508,
 0.41900261848324194,0.002186893501921139,0.20373406287991014,
 0.12572735032338656,0.46432468418621575,0.6029890854072655,
 0.7719536244015435,0.1827839826488502,0.7160207842298548,0.24895420085203412,
 0.07071463890766982,0.003921739210815622,0.9421381091656817,
 0.18273244409235745,0.736203085279679,0.38238639484760206,0.0798865156281171,
 0.48514059668970533,0.29535713228278726,0.5047008240875784,0.619950903213781,
 0.44263032522021895,0.04487800674543074,0.7151740259159999,
 0.5484799427579294,0.23454003680855762,0.14017230584825457,
 0.8226963408446878,0.871216319676274,0.35677126379063484,0.7806963724020614,
 0.5861562896387933,0.6070340279658314,0.2989157632704347,0.7988499248678373,
 0.4952645402702678,0.5315277148588955,0.5419434648456943,0.6545512176909484,
 0.4931940048501233,0.7330512962199235,0.42999049183900573,
 0.13660796435302758,0.2683938032452783,0.5834429841850268,0.5967904857532266,
 0.7480755050881333,0.7076208583475788,0.3876564132581529,0.40423970716383173,
 0.8069066254124008,0.4671634102281853,0.40650627451255394,0.8078718292944951,
 0.8350571138162985,0.9349361427484151,0.1086291928987222,0.24514179157235372,
 0.07085939423992726,0.14340164469118488,0.8699697739924306,
 0.5471817128291673,0.39788110618461636,0.630295191051583,0.921694789171706,
 0.1449333778347398,0.12172656253849301,0.5860323257219,0.6679537574940115,
 0.09812403828300487,0.6281349692750545,0.5648876785420779,0.690115920912902,
 0.9797720567766858,0.9064698767159399,0.3488208096333495,0.28668179444800845,
 0.6672075571837975,0.36190009390208744,0.1876351371060362,0.1800643631119041,
 0.7572143824055982,0.6137883106329286,0.7074234073986022,0.6996605974970322,
 0.8931889495460255,0.592530225674325,0.6945680289964512,0.26409210688153006,
 0.42750232538818533,0.6213361270808956,0.3586114020072537,0.6761985093866482,
 0.6377433349529071,0.6259958805337791,0.21076619351601544,0.7210977002083925,
 0.9739947989963008,0.24990158798387763,0.6524496748193813,0.2869860177791177,
 0.9156686226105474,0.6486023731895908,0.6128061015839306,0.29853016220484385,
 0.2041751773364716,0.9966004288167857,0.07359878955503907,0.3806289013530384,
 0.9883815180888419,0.15416544992141745,0.7439981013278355,
 0.46097884047510984,0.39127023735871824,0.6551782251917261,
 0.5202036396096401,0.0460846512073817,0.5273249743330714,0.2467957133664591,
 0.5809312948886216,0.083173109014405,0.8533408142445014,0.7046081079228891,
 0.057139577970747646,0.6489583539928475,0.9039872339535373,
 0.043037056019834985,0.5545634151663389,0.3462149741213395,
 0.2329096935520124,0.4478661692956085,0.06096262542483466,0.703091237051677,
 0.21950623843595252,0.9949002762614005,0.1963339805167621,
 0.30986887673755814,0.2001946387324136,0.06498542729889589,
 0.9373339536978428,0.3909828939747423,0.898860150329958,0.6084327542895264,
 0.6760762978405261,0.8907629461018849,0.5796133189701528,0.9431713938855867,
 0.3745029695156886,0.9934651866755281,0.5383015503411747,
 0.008552523867293882,0.8064530511168578,0.98631843017111,0.25414810652804176,
 0.39974994755198345,0.9872621284925621,0.40538547787025614,
 0.11064056762779928,0.5665422238593858,0.2542078285309113,0.8234633507334325,
 0.014139143867710757,0.4901508759334745,0.5560698631844924,
 0.33039928594253787,0.781514163108398,0.3590802085902568,0.18554046758054465,
 0.23630134096316813,0.8910906008208783,0.11804637902909376,
 0.08543385358709021,0.19536887795131375,0.5320054245795188,
 0.8344803962971791,0.028252398979152216,0.33623941609579533,
 0.5392686444664936,0.46476906271560214,0.4965380172162981,0.673405784007945,
 0.35167110635307985,0.39641155248673676,0.6627730575582416,
 0.3102893433986168,0.2789248877575875,0.6084771463353025,0.69270712318049,
 0.7032051858730111,0.8389074010262768,0.641956402287788,0.22332444985248245,
 0.17458431448187361,0.48367585238410427,0.3397854654581439,
 0.1689966385724001,0.27106433883613346,0.7761019140910654,0.5996204926252346,
 0.9302231439278702,0.2665159770258556,0.27924631558550317,0.6610085810782048,
 0.6835346483408017,0.8051763048511342,0.1587461415903999,0.8326369129827229,
 0.9892369320322207,0.13912966149120443,0.3629572446303473,0.392902185086733,
 0.9029508267127196,0.5659685528062529,0.35528574943767255,0.5024412306564175,
 0.9022806427964217,0.32035828694552004,0.00988735463492929,0.584438516333853,
 0.6970198586730484,0.24634263434208203,0.2646915798328504,
 0.06852175903219437,0.7948354339247239,0.8314599951318813,0.867213287412334,
 0.016682513889420103,0.18527902612672087,0.5269434993593234,
 0.18323821558570663,0.020169435726813623,0.5460295198920274,
 0.4428748223355662,0.5660521704951186,0.7841731650026345,0.7386812456971323,
 0.4048693190121144,0.5524959417050556,0.8335180081618812,0.2667601305093762,
 0.25039034217551936,0.5842719021258614,0.5220777564084528,
 0.11087741211945323,0.8660262223743541,0.3640647701582218,
 0.42874586595476927,0.6741358791685691,0.8425989996763649,0.564702618762471,
 0.7150740031998668,0.1344387268947823,0.26051824791454903,0.8832025181245338,
 0.5033478456525198,0.8642400746248667,0.835991650525594,0.9713147250431587,
 0.953224522103832,0.8466377247332483,0.7037463498324958,0.6023696421816463,
 0.784360574122189,0.7952343521241436,0.9398653995391437,0.7874278917625692,
 0.30026769379691254,0.5473719989558024,0.02599180544815538,
 0.005416196943197726,0.547095324400586,0.42558356233166905,
 0.8406883722261077,0.5305310464840541,0.19243637561847393,0.8194026052943095,
 0.05358699945339351,0.36916126477207545,0.5391538818870665,
 0.5177498910173098,0.34328510775216037,0.17765347405908094,
 0.5705798165957796,0.8116312703932653,0.9161794146846126,0.15206171598694063,
 0.8032093212081234,0.7256893529062894,0.9840954938098764,0.6529659910230441,
 0.6113389541029478,0.801254067852788,0.6190864041167551,0.827570500023096,
 0.7249983190721471,0.19248141584147727,0.34488996286639106,0.614546989271363,
 0.32758361361562716,0.8377935879033032,0.34610151942465184,
 0.7137732208712372,0.664164323227916,0.27441422529492987,0.5261081191453046,
 0.09737841648795753,0.7830747258062676,0.2051176689886608,0.914356569651492,
 0.7729098632763882,0.48442216189353104,0.2905391444362537,
 0.35326638165489366,0.35173302105054416,0.32963928973942547,
 0.6867383155432046,0.12074902672737964,0.5967639193602933,0.2916989290965083,
 0.9450562166168207,0.015588947599565417,0.591265303978653,
 0.37844359128103067,0.6706164021218819,0.19003437131478385,
 0.6100783182528139,0.831675493610647,0.09244055749165814,0.49203453516963314,
 0.6618744990548239,0.2736942705497838,0.8071839308863934,0.9848648721364651,
 0.13565966394208617,0.22097339760106582,0.39229123035849733,
 0.5494288842529443,0.37309503700897184,0.01880690339746227,
 0.7972405794690294,0.6468085197105939,0.5439461961181781,0.17069727746474528,
 0.5125176181631369,0.9982391923058582,0.7414390207937485,0.6156075126517246,
 0.5174355255928761,0.06890939434152599,0.6357578989704166,0.7297842841212482,
 0.8870861063589607,0.20730457112156753,0.8161278820292394,0.8494017464299234,
 0.8549121004863851,0.8207198317125077,0.516827311140632,0.09096041553093037,
 0.8429029086334054,0.26646449839058683,0.6188990517330868,0.2339069439869269,
 0.18832622516306452,0.3821643528678885,0.00587310319056733,
 0.48115482525128517,0.3428855649017196,0.2715145397298846,0.6292572431230399,
 0.7663442992666957,0.8886720634160966,0.3410843733017004,0.8316946845426624,
 0.6020137171033906,0.33048467634276324,0.3964830371611421,0.8415397175186398,
 0.31580072634493084,0.8035049955252423,0.007262954931892107,
 0.510686849805046,0.3888393211308402,0.8721635860257477,0.43598633711516455,
 0.006163271786719604,0.2924612843349842,0.5948696143638181,
 0.27322911402127514,0.2873865289447797,0.3869886934749647,
 0.37171827704837745,0.6517773590635183,0.00879857763348657,
 0.8219166382196208,0.0735593755082582,0.28866885592338554,0.2687781802385262,
 0.030845493868988383,0.1598680560802328,0.9432727547273708,
 0.20970244709474728,0.40598901577157964,0.21532991377520228,
 0.41811899607067016,0.7060002641660411,0.7761800110226208,0.2682589497250021,
 0.02612961101627642,0.706814729123022,0.2606323830446423,0.7502598115610551,
 0.8144634602985,0.8852651373878953,0.6016960956198198,0.6892883678271416,
 0.04953611976951622,0.9913326192410444,0.018208061168909162,
 0.496748846632875,0.5197863751510717,0.6661642621921701,0.29310783158535647,
 0.044003735611195416,0.6972793147778695,0.8798269073712276,
 0.3441242601632568,0.7717064078871836,0.022075931023511908,
 0.1173287621068757,0.33527304933081425,0.09370061899108839,0.400884881613373,
 0.9910565250984308,0.47356133250933097,0.4570138545458926,0.6978326161578606,
 0.6778775761813132,0.33908158347727957,0.8321021827608038,0.6251732283285668,
 0.5028181240305724,0.7061741276451042,0.9164127546898505,0.756933572127417,
 0.34246380484277283,0.10732299370839227,0.4360739885227717,
 0.6586794756938987,0.5877627120692758,0.5172194829645709,0.23649319747328168,
 0.34478868780072847,0.24097602763898085,0.43948776563250747,
 0.6152041210521989,0.7905096632025419,0.673065008507758,0.9615796495511155,
 0.2888027672401634,0.03138717954205095,0.627476263972058,0.5324628953377847,
 0.007746907028313599,0.3724648324331604,0.5878126042443629,
 0.30510143870379114,0.7022170614973426,0.35421888883830643,0.993156059685565,
 0.3931923424789632,0.3324220860268542,0.26798790800202266,0.9376133092257295,
 0.46319397380655236,0.819233028853298,0.6500380286647599,0.1259883667490438,
 0.5347451105481452,0.5447032081610672,7.232665878309508e-5,
 0.9969366678953439,0.668831266693261,0.92503289270019,0.1363839937409711,
 0.5672837161445993,0.7936389621353106,0.6457598995906709,0.7832462137467506,
 0.9021591820697972,0.7100305678435144,0.7016439479714769,0.6914719078335309,
 0.5249213573400334,0.46245530216202047,0.6616370136824483,
 0.04406660480941038,0.06301273119552109,0.4141315934806332,
 0.5581286442078682,0.28399647376683745,0.13547906338860383,
 0.40829219623532353,0.06983858499638873,0.043839802126035554,
 0.06270833563138833,0.8471816062214713,0.24108983207810686,
 0.03723348649035829,0.7793453506267605,0.1592169598062717,0.2507898088072561,
 0.7622107208769695,0.059726744942713905,0.5738559650153873,
 0.8207155304159527,0.6148174985100615,0.5891911336976939,0.7400439607240908,
 0.36829282727802526,0.047812611335475874,0.9154928945351656,
 0.7429690180176792,0.42767505069075684,0.7719832876448757,
 0.35123131067560787,0.29814144971218015,0.8254011258650612,
 0.06057212130796097,0.4306422373589961,0.4489319629162687,
 0.35768135518123256,0.7565409512552224,0.7684628108121266,
 0.19002161684916652,0.8150449894869998,0.5943063873726754,
 0.37407803766164294,0.5950239311707693,0.12912051634763122,0.275311239822867,
 0.7353309621651702,0.8450046512840496,0.27452082835165914,0.7513272635166921,
 0.7748588064710784,0.5941967144583994,0.2390140087455852,0.27740943510132887,
 0.6770423707801302,0.6327996931226993,0.3998612487029196,0.8213304662213059,
 0.6861066986749794,0.6072329325198607,0.7053025565112924,0.13015549347323252,
 0.6559015639747993,0.12247627742645273,0.8572723860496437,
 0.03332296241446775,0.9575098615724466,0.5753918603332194,0.770892384909275,
 0.10257561305185492,0.35687349957488435,0.70266408840923,0.6125620446852336,
 0.7239925794996434,0.8735148745631482,0.6222940817133199,0.41303084484696495,
 0.977735662753947,0.21994769423754623,0.3219771519915934,0.808837846331626,
 0.09620283618613146,0.8193582231456258,0.9648195568018318,
 0.16104831542926268,0.5125617452033046,0.7740060797467857,
 0.017937456042282918,0.9049380103596352,0.6505162249140558,
 0.3518184397322849,0.8205103704389772,0.5335109506860652,0.11640919014441264,
 0.40663633427668955,0.47577472704995205,0.5089202886160407,
 0.1015458573558573,0.5877612388148047,0.8144664147967166,0.19532596191151874,
 0.7316153978318246,0.3766644050891229,0.6299880407859136,0.31516312666024626,
 0.06641639516823428,0.662470106246395,0.9338691063907634,0.3385960499767152,
 0.978949458658958,0.17728959545083078,0.3519964872078669,0.983324587472783,
 0.8958137273083641,0.6383907830410817,0.4170523297986608,0.9165731553858202,
 0.0376724075248247,0.5435057319499523,0.37137417470078804,0.8234749736076692,
 0.8710456724050367,0.9503018249336295,0.008792228083968512,
 0.2575943359926889,0.9399881259152286,0.06004504848735703,
 0.07231208820129098,0.46477909055941746,0.9874985363872818,
 0.1098275206175684,0.33914662170120125,0.9064206447096751,
 0.32658215062762463,0.30159710482558666,0.14091110769522186,
 0.893759673190704,0.1254472097718785,0.5122612785807703,0.5396174469204967,
 0.964327538880284,0.02071725844145922,0.6662679402002862,0.29127337872710624,
 0.6700481087081475,0.36651664883565493,0.1864383381922663,0.5944249943090489,
 0.03234947735502269,0.9029877220870949,0.39384101190343024,
 0.9307046310942595,0.8170781823989862,0.5862129548639627,0.7973156404932521,
 0.2656634369621307,0.3546134447437166,0.1724002678623453,0.7918581324974889,
 0.8161621262053148,0.08351152548501395,0.11990890464217174,
 0.4823817063569783,0.11630608644621332,0.6058451731214293,0.5090543733631584,
 0.14459422641182096,0.9163445449068155,0.13649964536849968,
 0.7737614232735739,0.6674138692898132,0.0762766681653968,0.11988020069102912,
 0.255004120721704,0.982103488220337,0.5995941346134962,0.25915293461206,
 0.020538520548255668,0.2709467900045759,0.7500482678124553,0.235641216538503,
 0.2727978508641764,0.1528303069413781,0.09892988132604064,0.9492708240909825,
 0.06170160894484056,0.6328619020450457,0.8813235586666587,0.830263951656371,
 0.7767630535253147,0.29213080400783864,0.1417246031508579,
 0.036633432117904485,0.3328404747397251,0.3899760289777846,
 0.06471253842004554,0.3387080842137892,0.9244305067660267,0.7298174789849736,
 0.0868093458812047,0.7719697146037006,0.6350483225912198,0.8518054751644716,
 0.12473120451417186,0.22240911859281465,0.38376042191202786,
 0.7139075413243086,0.6557297075986488,0.702242060481742,0.5844402363254547,
 0.32172028763001914,0.6002228631080325,0.48336091013441584,
 0.1284739050645617,0.18543369044890912,0.4440361991478228,
 0.047262513078452306,0.0841842819399432,0.5744360830896121,
 0.1038139229290268,0.5297136387317228,0.6002763854365655,0.91243695457001,
 0.9594757358954136,0.036813514002696834,0.4525578411694724,
 0.2249521800866081,0.6184631909199173,0.27289925819884053,
 0.28324399343380413,0.48390457692093825,0.11852604218793661,
 0.20036213563501504,0.37034854156953456,0.5783480362665522,
 0.06769052464272951,0.7717000075705097,0.82994024459896,0.1421975777684139,
 0.9846319748197184,0.09398423617223672,0.8505895306771604,0.8086733961565669,
 0.36514057346241013,0.47536599235263877,0.6657208646981289,
 0.10839589879468958,0.16404251478447585,0.17464311714315373,
 0.6630365459839384,0.45228753775477726,0.28031803084247775,
 0.46306304819787436,0.05037370538925068,0.5142441312199161,
 0.8837829043417312,0.4078589899734686,0.9326093673107642,0.5584646776917332,
 0.7216762705666243,0.0880143134470065,0.376989916532948,0.7735404795810588,
 0.001498574687138432,0.29756039201405204,0.9751199190703008,
 0.6141667623161428,0.12497431319305186,0.8940701211703819,0.5705000108937751,
 0.3949799160455908,0.8530368189064165,0.10027200819420123,0.6328184280440277,
 0.29664632050461637,0.41438378084126803,0.5987883772410029,
 0.5623896355012177,0.6642952234084195,0.3855389416659857,0.49533107144074284,
 0.736691130765367,0.5412188848729884,0.660100469203112,0.8300422391061919,
 0.5769974442637831,0.12074159236773374,0.04523847949815085,
 0.8108885570570123,0.7112530901793696,0.8149159316341286,0.22871308275253788,
 0.2633128549899125,0.4164344092897375,0.010663235044315211,
 0.33833600371061845,0.005492113526277653,0.0027352968886635853,
 0.1379810711254844,0.3448849525920301,0.4996135948340661,0.5891012033937777,
 0.08951808410359263,0.8215845973460578,0.0329082430536356,0.7264605665568217,
 0.45596972531020397,0.2646783921702679,0.20727356152940835,0.682164828048771,
 0.6804458858779957,0.02103702003452601,0.4772049904004412,
 0.42427882873929423,0.3689678787627543,0.8640254305871086,
 0.39594483618742937,0.4681482445595495,0.6248669578956746,0.9691311551419612,
 0.45944377536748904,0.5874466796413871,0.7708273225703276,0.7124436804858424,
 0.3383769969186995,0.8667650534332637,0.43031063027143324,0.5733605385099174,
 0.39440362677569996,0.39761201254854406,0.962122934853895,
 0.14596485951915128,0.6043235530723283,0.10610043714745032,
 0.0015256207883468376,0.4983218859749382,0.5008844696966197,
 0.12684375496200595,0.49425748395259916,0.5704750935807215,
 0.8344898453648438,0.4731087042144104,0.19994983185806747,
 0.16180930131779647,0.42828465734743004,0.3063472378173939,
 0.7620855037170193,0.8576423826433194,0.11755389818170348,
 0.05783971129753063,0.9260929898237107,0.7246384944329004,0.4365623075588305,
 0.052846721762444204,0.6264337921161365,0.6910085614749392,
 0.9775092546421359,0.37407261048895624,0.06948453852664349,
 0.19997494295797968,0.9688306579265256,0.20334110852466325,
 0.8531482870138464,0.2871280840485465,0.6986048173329562,0.6215587205383647,
 0.6120446035769449,0.9268551799056743,0.26218912004102346,0.6302697573455311,
 0.6230570441158556,0.6644660379387088,0.6876657041997801,0.17360203723670287,
 0.620833989385449,0.8446926377733587,0.813555675066187,0.5655929831971988,
 0.022097092512568306,0.5937187539172932,0.5632719564752812,
 0.12461417820297227,0.6839503144848226,0.6567525461061139,0.9366082311617189,
 0.10174489859367686,0.055866606396168805,0.8791167417483168,
 0.1997366467050513,0.7141736315747078,0.1253294492245267,0.5669615785308565,
 0.7249437011011423,0.7382535334303696,0.5498093717035967,0.5656800151218606,
 0.12367201214922341,0.2089520988942164,0.6197644637964792,
 0.49062512667013036,0.803686961609701,0.28350238928242466,
 0.15063910366752709,0.6117441978945989,0.05587065577820727,0.474931846507438,
 0.2802796822704887,0.4694922127704413,0.9291053916146227,0.6929440447070427,
 0.4674171016231129,0.7092151332261475,0.4469439075639611,0.42512094506253084,
 0.6068945018876665,0.5294020303432162,0.21294717324504087,
 0.45588094374034993,0.51438940200787,0.15199637519950315,0.8129766608629696,
 0.7607758317002743,0.6194257950879043,0.989326942399851,0.9851437743925189,
 0.7800051227020509,0.9619233181517497,0.49256025590106023,0.7649533850441289,
 0.015047818404866353,0.42701475265087263,0.6636312916718897,
 0.9563250358664527,0.4056326094788182,0.5880206960081713,0.8727472586025028,
 0.1622308732321086,0.4540399816131174,0.9230106031644596,0.7787596117550508,
 0.7318416454388331,0.40463170249903724,0.7784627751587312,
 0.01187566169259846,0.47865309858913097,0.5265802796513653,
 0.20736913411270486,0.38795376839271434,0.8035815220732534,
 0.6298010824547536,0.5241150973423508,0.4285169616929819,0.7395831232829613,
 0.8125442944479332,0.49803028852401854,0.7531356735449365,0.2713702015609827,
 0.8505794696784923,0.6494237878877026,0.859973923934897,0.4069835132219838,
 0.5021725843659635,0.7403125772091232,0.13948001229327833,0.9359300524417776,
 0.520488945579733,0.18724168384636009,0.2887026181110205,0.786419664186717,
 0.955067210545241,0.4787964747908815,0.9333937475952815,0.6957164174629653,
 0.8378910844739667,0.9883876711707433,0.6675273904018139,0.2898907281758468,
 0.6562627011173368,0.003980196046646678,0.0069843704000964735,
 0.460111271700664,0.3465707181642932,0.7499684034367204,0.24133078126319307,
 0.3091982108213429,0.30377066968275324,0.21008342351419962,
 0.3766149985955324,0.49654710710316663,0.5186076615879023,0.7526412556341854,
 0.07953026299617139,0.4120187715718958,0.3592151297449515,0.3655807223209142,
 0.9938865825288039,0.7285174825336356,0.8113721335513795,0.5952559399747446,
 0.3135233536823475,0.2337316111595512,0.6133027506639521,0.16489751035057998,
 0.05233931673267844,0.6965267429311749,0.395983733527022,0.1563104268584965,
 0.4511611824757965,0.6105151667904014,0.6385166234947518,0.3467686394703702,
 0.06578061776062194,0.6054150568777554,0.1369758860865118,0.5776145275715969,
 0.7851202302757951,0.10225789782698769,0.9770647869764102,
 0.12415494059480192,0.7979632389471605,0.30020760226999266,
 0.9959157913359903,0.9087465671322673,0.07598523021985826,
 0.09288251874116316,0.4705602917004653,0.3026632733202901,
 0.07546881180257414,0.5731008648163799,0.2755224485811012,0.4747337096223442,
 0.518168001734788,0.10579194571425765,0.8227079696194338,0.4073538883983232,
 0.8477543317713755,0.9722568383435798,0.9650473966407833,0.7440933312995783,
 0.2621305243848969,0.009551789981654069,0.42999890070569147,
 0.8401489835646837,0.09920754431298828,0.6053590285428592,0.5788112246500229,
 0.6829144559156033,0.13593906574402947,0.1610117049989307,0.8159433748420272,
 0.6876672096598618,0.8207312718981071,0.4589181188127601,0.9252590488387363,
 0.6168761629309741,0.6539744607644332,0.9380906869618916,0.4443388459817198,
 0.14091179268474308,0.758501509904319,0.09883344925283166,
 0.05814523345704159,0.25189273595144257,0.022021897603866858,
 0.5279623885293783,0.3991891608310095,0.6411288093521699,0.5539434741092552,
 0.5406960519220486,0.9753640917100896,0.27218438394804734,
 0.16083482442821162,0.8633634158301028,0.5149101365542675,0.4202601375219682,
 0.5508690930122782,0.05593033913487355,0.5689921358255874,0.934586181845181,
 0.044912860081030126,0.025742521549425024,0.6226380608121879,
 0.18125978355247552,0.8367380762485367,0.9124548604780739,0.6194171602400499,
 0.4640429808018316,0.25928050528424773,0.28342968533409485,
 0.22678017056351996,0.15249281193212594,0.2611708060826684,
 0.6604301030479942,0.1960675926717388,0.9878800828302571,0.06083219502875259,
 0.9515409166708493,0.21627837814988193,0.3508403221645079,0.7309752351910288,
 0.1064015282931362,0.9491542496407003,0.845157442191346,0.5313977167336841,
 0.6798441767924968,0.2841517967221354,0.03648181084657731,
 0.27443227128389847,0.09865552199347372,0.061510156728184495,
 0.27726286038503134,0.7042090324635716,0.5681514001118747,0.6357920156531385,
 0.42681697187058143,0.616207979340349,0.4222605438934296,
 0.0024715568538972166,0.23064135383626594,0.07254729650484448,
 0.7101620613582175,0.0020026092530652795,0.010444915526216847,
 0.7883981483675044,0.9900827565851418,0.9215009008624261,0.47710881053249465,
 0.20302982608057185,0.17127594914715716,0.39766034924548543,
 0.7926953357941436,0.9748940497538444,0.2872020387085038,0.6478434340345762,
 0.620806171395439,0.4352480226791092,0.16653072152431325,0.7987747208490361,
 0.977712156783824,0.956376386977455,0.678012941406155,0.4326198497267746,
 0.00655055354376588,0.6482124408956477,0.08570494785197502,
 0.47379488618687327,0.11661760557506973,0.17337618284696044,
 0.7349028140543253,0.6635648687004804,0.1605438303691329,0.9985662761633526,
 0.29458699976459646,0.33610900984328296,0.14987993930823096,
 0.3665567180247159,0.25351049725007124,0.07355852886503855,
 0.5920160834215581,0.28533970145697274,0.4834843464417462,
 0.12306464540358963,0.1411811207798469,0.7396554535252048,
 0.030943399684842166,0.3255359747201807,0.2701471361297293,
 0.26721416544158505,0.2604335082169983,0.9462651762759992,0.4404208051731434,
 0.9670334767558928,0.22574367239162552,0.49975299944137785,
 0.9230550907390824,0.7572454513512952,0.3063550269740687,0.7295944100604992,
 0.11879875803019235,0.2898806968298262,0.32112212510382776,
 0.42929294049685884,0.6754427784629315,0.9687976697633448,
 0.010742337578157235,0.7759618110931815,0.6286138516918423,
 0.8362427247007518,0.3070440305189641,0.6942782850915972,0.6774151798819517,
 0.688709754629203,0.9713394071114326,0.5330240557331131,0.3608478872916382,
 0.6211576523425414,0.8450251688475473,0.060550853838935215,
 0.8856534165708388,0.8480512213223274,0.7084864257139877,0.26311808462971953,
 0.4662493149357234,0.760053799330945,0.12416672750564772,0.20368945749010559,
 0.8553963535610976,0.35428493257681526,0.6564654175467227,0.7843672921438534,
 0.7010582053591816,0.6669564450793786,0.4435873554204244,0.11328623985483599,
 0.36407885363096815,0.4986645536944114,0.2361808746183327,0.9362226464984099,
 0.7792629040016655,0.47513451249795,0.4688026205551874,0.6394111935516661,
 0.8796289739315151,0.595941333553011,0.3938366826036236,0.41994179727151626,
 0.94844490859726,0.8370852757049285,0.3004706765349914,0.055531329150260045,
 0.5225083616022332,0.9656430599098405,0.16880547124311152,
 0.02527200699247023,0.4092488441886447,0.008576699804421883,
 0.28927283239657164,0.14933599058493374,0.24555601850146846,
 0.48355444252097246,0.4258180476110436,0.2651663201925345,
 0.08237105032101733,0.48918588060539625,0.1511984519694095,
 0.3793295136526661,0.7471616043403239,0.9448451182818868,0.24472105774186637,
 0.28554072813125775,0.896465970505644,0.2887857171589766,0.8887565769276332,
 0.8640227403100366,0.14085733097355513,0.7345012810838492,
 0.06574988337873067,0.25104985532325497,0.30137100912613124,
 0.19120041883811756,0.7565507672021636,0.5867600977319831,
 0.11887096918586437,0.6161980843041173,0.9106734012558289,0.8304674462601296,
 0.7967975636627834,0.2873905145195186,0.7343010293157091,
 0.013024044686562708,0.7924535922276753,0.2679045475748131,
 0.6219384285911781,0.05433374839184968,0.4159815247162476,
 0.47153656404749955,0.7893491162839381,0.8804146478296245,
 0.13334657861802024,0.00906648887943362,0.10404669124594523,
 0.19817102376796547,0.22103603893296775,0.35627680651014615,
 0.586748222032211,0.11999381069270565,0.016734015005203662,
 0.20239971944413093,0.3716897464297163,0.8020364240815936,0.5499733946525924,
 0.569262991043676,0.07021462071382156,0.11387618534655997,0.6478719168810465,
 0.6157930233495164,0.32113343536946315,0.8553454480792416,
 0.01250095142699481,0.3716216800077281,0.030706916434701137,
 0.2563085523439568,0.12410671682213326,0.06741791044233292,
 0.8711046018329136,0.7746845789641469,0.5643204865504295,0.5975142666310611,
 0.5157815329125837,0.6094805211769042,0.5236440968330366,0.7771441791993255,
 0.8860885320709581,0.4237349027716,0.011792330598689449,0.5569783435137614,
 0.43786366480419403,0.8031579301466646,0.948429116749293,0.687797958663033,
 0.18647530872981177,0.26435934270258543,0.214389235731008,
 0.46567846294582926,0.8841610561009846,0.530562550730509,0.1763224350252931,
 0.42056071524658334,0.6197445777201878,0.5563674388835594,0.2015629869521146,
 0.022458967024198695,0.8352342517629605,0.8104194539618856,
 0.23116646833066945,0.4558053614683505,0.38220417648981597,
 0.8547816669928365,0.3654933984577755,0.6532880030317507,0.20427790855204309,
 0.004096000109074627,0.23060528051524865,0.38568479983301174,
 0.8513292483480454,0.4279892248696222,0.24677699818917886,
 0.01913902901722697,0.23923299588048286,0.8045023779452973,
 0.36572550350595323,0.5208311618071447,0.8267143689782377,0.6349810580047708,
 0.7665883826337367,0.3518985610548532,0.6517900778493684,0.2908976140045375,
 0.67856591928147,0.6703671178502054,0.023976739082047338,0.62562702565476,
 0.050210779391758886,0.9833755546812981,0.3312774339288379,
 0.4365956299375259,0.9168800839536048,0.5587010174372798,0.915767451566482,
 0.21862617466902456,0.08800968728576508,0.7538585866329919,
 0.7100742969864383,0.65949399395002,0.9876323549065197,0.5544216749798312,
 0.9219850281933405,0.17832617949076635,0.5103290711806805,0.7210652094270806,
 0.05351545797115442,0.7256503486329731,0.2999496835893689,0.7731273078283597,
 0.4119465447506653,0.2713179898069513,0.9542854663133608,0.5299006330037164,
 0.04647108858878313,0.31101693265964914,0.37369920771070675,
 0.9668817766548957,0.575309132396847,0.5593348767315688,0.9137918657237094,
 0.8797780804826111,0.9840786311584284,0.668908311422218,0.9173903641134651,
 0.37837388401097716,0.4163141267807523,0.7864801971426538,0.752531933183306,
 0.47705610598802284,0.13507330783169957,0.706473692371284,0.2758755267760815,
 0.526137345758975,0.5543638967005273,0.13422414587811649,0.33955747657297375,
 0.48002294518979727,0.5948845477602598,0.3330048575935667,0.6264378830364546,
 0.9733429966749528,0.5184652493301587,0.10801329396495418,0.8200172045988721,
 0.258857019816065,0.5832963665360166,0.32196701719291565,0.9683771343695573,
 0.7560637666717556,0.5654141059929128,0.2848413738259339,0.2748099720130389,
 0.19158930843122968,0.5993838278376395,0.5164001521804447,0.4201551639369151,
 0.6192099998069039,0.638508032030007,0.35992980217903503,0.15162508197816993,
 0.6399222765568235,0.8998893234234288,0.10173763882693332,0.7426865802807899,
 0.12647870323293764,0.17339198290456348,0.7006691488203205,
 0.33940123418786516,0.3976710267131355,0.6037528981151646,
 0.36859400495852346,0.4198125486125258,0.4483532402514969,
 0.09375119214860694,0.18927303666952278,0.8806570994893226,
 0.3086148178107334,0.6527404581757492,0.48954769448907753,0.3317123985532786,
 0.8720447402335324,0.6288079136137599,0.39104551356635464,0.9513558464385368,
 0.5659713066999813,0.030940744545132737,0.5953921081270181,
 0.6106420961047353,0.3534443311374844,0.07760521253692532,0.9042655150722028,
 0.17509165432732665,0.33849236176372555,0.9809292884793052,
 0.8631813610845582,0.6586252102095765,0.8211383305689498,0.9651703748378102,
 0.1018999156890048,0.8101185575944032,0.14659954945236509,0.6579067917332434,
 0.8149836796963935,0.4859248777852161,0.2836999306440642,0.08727651839012984,
 0.4018428981274125,0.607025376293608,0.5010373754045638,0.7179855299406644,
 0.6781652644495954,0.726390366505995,0.43719558287833726,0.5713580135791632,
 0.957881760343769,0.3714213291032473,0.5370520988148013,0.23570835570227133,
 0.4250921812092132,0.978220977918323,0.9135033316251561,0.7678790204959522,
 0.5162509109853033,0.8589213557019498,0.628887167096679,0.5610583405489508,
 0.4492943464326753,0.3834311213237913,0.605913107572291,0.6171940392729709,
 0.41689738393385134,0.13556406341615146,0.7436182598670592,
 0.8019863983535994,0.8461230681607059,0.45306342753884077,0.6674215091632939,
 0.24768478756360046,0.18510496169140334,0.5254568615992217,
 0.2455499230062419,0.8181288527984355,0.9493136700513259,0.7714556861620127,
 0.1524821017059641,0.44113636089008834,0.5091198689971335,
 0.20633331710371605,0.35151335959880015,0.5544331055331068,
 0.46880229523589856,0.24664579792727392,0.5712260805351512,
 0.8528082619449242,0.11734104428939518,0.544190453050601,0.5598728270998398,
 0.26977283247342143,0.985928471971214,0.4842014260997194,0.8644624486494009,
 0.10321016369674563,0.27674954175972,0.047243511228382884,0.6210050324682466,
 0.35856660011333075,0.5690958066772539,0.21877731571048653,
 0.36756139384475983,0.42156692518899486,0.11225763197473526,
 0.017969213887470636,0.7124030918618042,0.46140363133281426,
 0.34519289380785534,0.39600571941217666,0.04448454241694011,
 0.2041818728591117,0.3245447503849638,0.780794359874064,0.17320694242363333,
 0.11718156142973624,0.021931329485190276,0.022941476927199944,
 0.9725252887692721,0.6001085261190721,0.14477031983960154,0.8317561694263692,
 0.2882869325290979,0.7042424274090795,0.6443456118330694,0.3456895459942245,
 0.978892519287926,0.13078198646231087,0.6916968977143536,0.44847814863061397,
 0.41953970273202257,0.7699873569847612,0.036688651652253323,
 0.5125803787757671,0.06899544279132108,0.6030235374750263,
 0.35233123700685176,0.9729538572146477,0.42707442935220263,
 0.6383310455260671,0.30088572917525425,0.12083276547217736,
 0.6423165017825361,0.14473161233052734,0.8239913935019806,0.8674270799351806,
 0.21940229913765374,0.24244234241711693,0.09709895487819309,
 0.22235946404070173,0.11913241952957554,0.004413949224378655,
 0.595670603182705,0.06758989870342769,0.9905604845318436,0.4010679645887074,
 0.36544211000465743,0.30475012576892413,0.11051046579260637,
 0.8942661213713904,0.33426883848788,0.6041609464014959,0.7062849387365824,
 0.2255107550037576,0.25272262037027804,0.01741280626230024,
 0.9910020016296749,0.9258341251282878,0.33857023915534545,0.6013480852995173,
 0.664246223227183,0.3703426785421563,0.5473345306921636,0.22719796131900794,
 0.5328599804015958,0.3587421232961625,0.5818693018653164,0.9088843840491305,
 0.08794035871294859,0.4188337231651983,0.42556810985896265,
 0.6133037928844266,0.5193736091295503,0.5144044061333325,0.4403255497974252,
 0.9619507666942225,0.41030093199719375,0.8060820241646456,
 0.27578172237867216,0.1941921609610774,0.7571632338783708,0.8349902279606694,
 0.9038093563616897,0.6071410225860693,0.17397826609246902,
 0.42697377504648126,0.3406146739609369,0.6488354625387979,0.9683652650021072,
 0.3547180448721955,0.7534822396131261,0.9803670629638245,0.6821601059556011,
 0.7524320240051335,0.030738124310817905,0.99730790675248,0.7285757918542175,
 0.9844295637088264,0.7569416110448249,0.8138893268884527,0.17747996011701694,
 0.5839273782537541,0.3923891421308673,0.26686070965258046,0.4975342397233211,
 0.24852422114476458,0.8763922962700582,0.03755576727527721,
 0.6503115865541775,0.9160102688081724,0.5300740473047894,0.9429864721394394,
 0.4039653336175931,0.6667993561882,0.8172485806068587,0.7965116509033766,
 0.42626902205519845,0.22102048836586952,0.07192895736763583,
 0.2078427786161018,0.8204609864778379,0.12584486166311204,
 0.24913981136375374,0.24914071398491544,0.29051764036856886,
 0.40309075154564944,0.15134908004686676,0.795662002558834,0.5886912546466796,
 0.3126764248872308,0.18739638505557488,0.8137556213290511,0.0829527148263397,
 0.4549821915403425,0.558995604362218,0.5539123053838663,0.5222606000511143,
 0.9805190193285109,0.5687720245334492,0.7616733319372881,0.13909035958167748,
 0.097133299135154,0.9613290555011367,0.49147917325585755,0.4367887700047175,
 0.03964581115515198,0.7913322996422805,0.626417153116785,0.19111406662200414,
 0.4808739269836094,0.8731577497678785,0.21251357535506576,
 0.17215931990091082,0.9474996037629926,0.668118387635422,0.33750592673440505,
 0.5875930863687264,0.9775035459611595,0.1942169902694778,0.514288095915687,
 0.7318995295378542,0.5026329810423492,0.2213512555932643,0.3234988909463439,
 0.21339119108078697,0.5405677639314241,0.7463279374055284,0.2184875181587883,
 0.47238950467867824,0.12113782031311304,0.281952052470823,
 0.41236378996269984,0.3564436186239668,0.8381296721788812,0.6785175699341509,
 0.09408314775156834,0.7008875363317335,0.8621653252617454,0.3147502414012078,
 0.13628256910933711,0.8783560524889461,0.6543097790045901,0.5354154134034204,
 0.4388883097179568,0.036005609622764,0.790495997087746,0.46151316208919324,
 0.03126925682022497,0.021574358087302348,0.2372810679207683,
 0.25352793368525384,0.1947203901281771,0.7210059447960699,0.8176218013737231,
 0.053928139717288714,0.7822902798275679,0.3316203563759208,0.722764020984531,
 0.401166090149227,0.9050383924546007,0.24989277052215553,0.3501711768350161,
 0.9273019810337311,0.3631269890064769,0.6219012222428448,0.9042129503911049,
 0.05696310536800353,0.977129410780555,0.30382022845108814,0.6983952551603616,
 0.22257427590531143,0.8157253193113876,0.5176237563402414,
 0.23693832072411158,0.5576094137020118,0.12056268209510668,0.555182745697202,
 0.1030561493762635,0.6477109084212249,0.47463191956357487,0.9980057228002752,
 0.353168217527117,0.6256267020270819,0.549096137170811,0.10034889863260954,
 0.9962737007276286,0.06018635920648785,0.13122414729929688,
 0.5949726798345639,0.20700275589308037,0.712600547753073,0.5562276477573098,
 0.27655729777491544,0.3005118230946977,0.6781237556169744,0.4938178183345068,
 0.1510379078428865,0.7260728977692689,0.3338569380899795,0.980891651588214,
 0.8008736953732584,0.19558377250061398,0.6189163757738435,0.3724684665482857,
 0.9208005884062911,0.22177242108938144,0.4014180871991566,0.7803134223343493,
 0.0414688771185977,0.27741904809469253,0.225957085235151,0.25748306938255083,
 0.0880553507253432,0.6809602939663904,0.9482096765118605,
 0.0032543877820202516,0.036500757785845495,0.5450543982477625,
 0.900380784091446,0.4968760383779909,0.3783016765325826,0.39264206521615963,
 0.8951178010852723,0.9826547274580982,0.034513134421669034,
 0.08323387653418868,0.021468006662852868,0.15877497852519618,
 0.35842284823237847,0.4364928536584217,0.8088343445346697,0.9404453555250433,
 0.3244593304742547,0.15593757922827256,0.5757545330072359,0.2575005102303537,
 0.37133701628498894,0.3732767239722278,0.07017758742119828,
 0.2786514442613852,0.5484174880646664,0.9251981182722553,0.7730498220425598,
 0.45380150126144647,0.5223252478246152,0.1037177212551309,0.9050395792883893,
 0.31585157961585797,0.3295383424392939,0.8368150640544528,0.4557368504437843,
 0.4876843612378423,0.6935989653436617,0.559078326948069,0.8370586331344563,
 0.6827211808749859,0.5575348375531285,0.15394484414417453,0.5324908375947155,
 0.519342519255999,0.5875667323234359,0.9605933001101954,0.33583376329518133,
 0.13284586165050938,0.6735888850666469,0.9013260488990714,0.4072693790849552,
 0.47929373912080864,0.6247710620261633,0.7279820779134639,
 0.04439871750452007,0.6002467733537262,0.9368124467427672,0.4979203044077092,
 0.24146378695290238,0.06523537036565585,0.013401463204621988,
 0.38236579677312876,0.2877719127835604,0.23421002303473415,
 0.48275362183032644,0.33240009459329356,0.856141914225434,
 0.33942358117705607,0.7370205329936365,0.4870008938032935,
 0.35498129882874374,0.4316417437185798,0.9101299535702583,0.8374310036102782,
 0.9277069742646197,0.4325836031038971,0.8321165895553431,0.17427528482097232,
 0.09698211982399829,0.8700818674779381,0.12036184733835387,
 0.6604132636069338,0.12954053768071538,0.5417516054935034,
 0.13171617935918678,0.9812288753975265,0.975748947575121,0.26080865337229897,
 0.6114134943053559,0.4884726628272422,0.5624901217909726,0.6718712674755677,
 0.6086266278120864,0.5368216174643008,0.5807242013400128,0.4572673690909964,
 0.6698462515925762,0.32809908253415265,0.6918087088407,0.263684369605118,
 0.2911090863060092,0.09639515650015129,0.9344746934936455,0.9878921202284666,
 0.05865204429428006,0.13676860933933543,0.11394917716108022,
 0.6804705502035133,0.06902382913452199,0.689184608365909,0.9804852112760251,
 0.8547522530645403,0.9814429678042355,0.27553227279244674,0.8427478725574591,
 0.39360750809332967,0.9750486436186567,0.09678469741825624,
 0.5323007499469169,0.6121565942391335,0.9157783856575559,0.4893563052499815,
 0.9296412976800961,0.2320760711399864,0.9336164291632612,0.6104223858098794,
 0.34872972180703876,0.03290780203301136,0.5218955820083139,
 0.3627188437469102,0.48453142981381436,0.4163860542551696,
 0.01904867477943084,0.9700290741301243,0.8695923979823182,0.893102392404463,
 0.049377932382363876,0.3148504251342079,0.36685354714938556,
 0.3974441660251722,0.942647163419463,0.8307088883043309,0.29930324229900296,
 0.42660424641757677,0.5084308429313728,0.6781029821893061,
 0.07475391705382828,0.32860702504696415,0.16564224628642754,
 0.49271304666180626,0.23256368058801202,0.356010680178527,0.9355936705117549,
 0.7662845321366937,0.3406645607209762,0.8214780223336645,0.4063411210154284,
 0.9197473696856883,0.7480092371206255,0.40416166990824565,
 0.32498933783938755,0.3251904702158007,0.9381241891471903,0.5292868427170794,
 0.7613611815389002,0.06893861425017811,0.7146828189286634,0.9978321280913952,
 0.48662810920238697,0.5689868516308432,0.2792497326579578,0.6199528859707453,
 0.5923176600811031,0.4523098934934032,0.48306415076214326,
 0.15503531595786857,0.5949842223183553,0.9735041419872227,
 0.21770135309924143,0.43172523078337577,0.8712364984411024,
 0.4890393240218749,0.42595292726273637,0.8995619172541528,0.6496119005437194,
 0.623686093513011,0.12203633742660591,0.6076126285408,0.6758731256545502,
 0.9550906905303689,0.3618806598535088,0.26926092662124956,0.7297879231340392,
 0.5785123900130406,0.5309384709063266,0.4129504865876201,0.9429213078699166,
 0.7916090994586272,0.16324139048914077,0.23289208241175108,
 0.23062891874154778,0.7104069803447133,0.8260647357980742,
 0.04611132956126185,0.5930968055315,0.2589313626808545,0.48871022520555385,
 0.5513187375422908,0.3480731664223925,0.47305270763611196,0.4216834354649397,
 0.32381020188416787,0.6442113729627623,0.23822520271709635,
 0.43531209391226544,0.5828703907345846,0.5207174945486458,0.6718951280555735,
 0.32197739494823696,0.2197985991471949,0.381832146861968,0.3007422159054536,
 0.2869411614754733,0.24170798483325828,0.6125428715165891,0.5343274263641813,
 0.5820675417547732,0.9341030501968721,0.6481989448191883,0.10007213996662068,
 0.020839823732647478,0.7525864006518288,0.7378244073687562,
 0.7285532080582336,0.3196856027872559,0.697218762062155,0.3188341328668056,
 0.00706085375486154,0.18690613727327487,0.25023353307279483,
 0.9119548279351963,0.48607183642285956,0.8653530076970648,
 0.29494027403163603,0.919847266986908,0.7790862812283216,0.3944867641769202,
 0.25165181736806863,0.3038018550464485,0.9823144073633451,0.6029458124898854,
 0.2679588539312504,0.6744000787864329,0.34808767959833287,0.4487132136488846,
 0.18137363914946658,0.7289069080224925,0.34229520433899696,0.77566961374181,
 0.9405507350231368,0.4740113407542095,0.4080637911351477,0.6163067445239625,
 0.6445396615286407,0.7836796079922712,0.13712962005226048,0.530104816101674,
 0.19662746576200119,0.039395278265700884,0.6737207400131401,
 0.6496185403458958,0.6714143154067209,0.09179809535302308,0.1404553007386311,
 0.3272351080323984,0.3087213858763075,0.6447144213573162,0.12983703222411425,
 0.1312522473930562,0.2002160904074266,0.9326257411748108,0.9113534862802364,
 0.23331289495217733,0.1933724805438205,0.8382017062363724,
 0.08113246200562185,0.06588017527201695,0.14750194773456604,
 0.7144147148586237,0.3640274984225135,0.556596175388753,0.740337454250513,
 0.4665111815368901,0.7401594335309017,0.5699007077794311,0.9905579107875766,
 0.3705745789931333,0.5461884887491245,0.5255566591819696,0.4655975125253706,
 0.8507747578606485,0.24754960536764448,0.659399323737492,0.812236572244225,
 0.00415547741840272,0.18914957001010035,0.0485896451828991,
 0.9246247846769183,0.616128219575197,0.3051893499787194,0.9808221500290161,
 0.41869423802056116,0.6611094741222856,0.8217040025840165,0.2012049815210556,
 0.10777456321524548,0.958609567840602,0.35809568450440576,
 0.19361267378108105,0.6462398977370752,0.037229551974822694,
 0.35309412772114446,0.9407140441467342,0.6303793511531248,0.7084990666966818,
 0.4978153028802911,0.33449818262899933,0.13555847298852508,
 0.2956740883545962,0.6435458690112075,0.22767039404592726,0.2845538586861528,
 0.9933255717277655,0.5952276053418051,0.942592931924293,0.86517066354308,
 0.1069537217271963,0.9727290171303584,0.9346754033828847,0.2303754693206641,
 0.19789987373409956,0.2683374392566682,0.6370713702981702,
 0.23550273255033527,0.8496771464846782,0.5498706988238288,0.26657066191043,
 0.37910836996934993,0.8345369972745921,0.23600996283599573,
 0.46171078562514456,0.5326249998072141,0.9523331681436753,0.8630444124045327,
 0.23747222824073233,0.7502938119305967,0.8119320670875001,0.5194768849676842,
 0.22174787889701242,0.3106223239947705,0.4485624346006989,0.050973105571125,
 0.46722255955313885,0.9628204369080974,0.12855267884574972,
 0.45806804891606334,0.8725845863833567,0.6943594837875755,
 0.18638986165583693,0.3578109991260885,0.17386554297512047,
 0.45032693620082853,0.2099357410121474,0.4228347905234413,
 0.01797100525976203,0.11628449594005197,0.37375617799876903,
 0.25565557437530173,0.18025163269094646,0.04568572346948696,
 0.7854871735466816,0.8380806400830344,0.4977173049384802,0.8032956502313943,
 0.8608186732431017,0.5321145977377478,0.9685155596397983,0.5939860681494817,
 0.8093538320066564,0.7074393179859858,0.5784955910531023,0.2523042567484597,
 0.4713295453943782,0.07819096444229723,0.24088839988071697,0.049552619639603,
 0.4730696134944016,0.48955712959915754,0.35382897139352154,
 0.9460247522038543,0.9349924847247671,0.45859285203097444,0.3903678484940084,
 0.22814380088831732,0.8569369531459826,0.6396186025458495,
 0.40503690373462164,0.3621762694130608,0.22519192623656503,
 0.6887250195366911,0.15824041674422618,0.29190370805656196,
 0.6073264333875603,0.8960041210859857,0.3127656156862325,0.938065093004298,
 0.21786308260998688,0.450950830789028,0.07421408794370965,
 0.34254706554052994,0.9744117599668359,0.18170724255545423,
 0.06889021808244855,0.6227477440484324,0.6072697885424848,0.8567504114828564,
 0.11881387167211033,0.6270995749963669,0.42655046828538656,
 0.4521781305475081,0.9265019799357205,0.5343420922143958,0.9267079573288648,
 0.23130932787103753,0.3926559869862951,0.31382155569407133,
 0.9951215013101777,0.030110606198242018,0.07131249155242936,
 0.7608484027135676,0.14905951649293514,0.5367908948661598,0.8044869587428873,
 0.9524875740803818,0.9741647777816657,0.9966129871735336,0.19725877633613242,
 0.7411657654678997,0.2946583338397787,0.3848332798892137,0.16333614161100907,
 0.7690110534496464,0.3636186646973908,0.15002073393581528,
 0.061192009357646904,0.11082540495066551,0.08937928184532551,
 0.6417181405032598,0.9335214664606951,0.27208316371437613,0.9829870637786504,
 0.07296470982863168,0.6696081584489842,0.09948899516465803,
 0.15591273076955603,0.4569016244648002,0.9693958150583674,0.9825052213477655,
 0.4291141206063953,0.4052099260955301,0.3931035905244865,0.9190306982270271,
 0.1322174847888261,0.4495027529883219,0.27181558666013106,0.9811170291623721,
 0.7478256201908781,0.2880220108425605,0.4592463486707855,0.8356189180089626,
 0.0674842556172155,0.6620245418244955,0.5310960503742367,0.9503739705834401,
 0.5697475970784946,0.8197228503930398,0.6451098410773907,0.6162545582692287,
 0.16433010365347323,0.1385170249887535,0.9675282176360268,
 0.38962132836087116,0.23750687502938594,0.13504615953949184,
 0.7909828234107745,0.4796175923031024,0.588227065503713,0.31753747313356584,
 0.6736290726051545,0.5948985166972511,0.6890588447015343,
 0.0028237908706572057,0.35042971066310835,0.9853860681364071,
 0.08457218289459845,0.20682623103382858,0.41916775412984264,
 0.7152104356134025,0.26181333489232295,0.5555974120946472,0.3234580641093696,
 0.48398523040102015,0.9720151562407494,0.37782316007087646,
 0.13687322895788312,0.8221192272163333,0.9770792634357127,0.3039565707853624,
 0.8192936164989492,0.9811231156533456,0.595850770307119,0.4880510816095168,
 0.45127097190112064,0.481400502925596,0.9080997336927622,0.6829493219914082,
 0.9387088195270653,0.8920233530820957,0.39661195200622024,0.8619747536512619,
 0.011569305647709194,0.24535697722072258,0.2255009370708323,
 0.23989037858778484,0.9107138187773764,0.40757141771199457,
 0.07950697844243015,0.3994988195545025,0.08113623332000197,
 0.3010397758101385,0.7799855511271047,0.26962726159105765,
 0.06747757807768207,0.10385253243700776,0.6316975671661563,
 0.5114022923890797,0.6231129960478473,0.9874071738114081,0.7720807511468177,
 0.7939998616257751,0.21214943588060597,0.3350336195790099,0.6642723937173864,
 0.03533732832830838,0.4086325665002849,0.41032425306740183,
 0.6709127370466736,0.649440894926579,0.7137184069455054,0.15582455501335823,
 0.22763746540223284,0.9246901788552574,0.9461668159246289,0.8733074811558524,
 0.5115699858854311,0.32521408572221655,0.5937760863314279,0.8333889604268383,
 0.40837540420094387,0.40276674185643624,0.3575343907175914,
 0.29637449803486615,0.6416212665723143,0.7833565290125669,
 0.43853066218025494,0.26243254887188416,0.01854984099098722,
 0.06222300775210621,0.839628121955232,0.37804914692326674,0.6996923136910873,
 0.5347209254163143,0.5243770401105587,0.6888897648088933,0.24236394258167593,
 0.19918035137088597,0.8476168637777668,0.33171056733014537,
 0.5497825891178151,0.7255212931028147,0.9867572468218728,0.08796201042661056,
 0.09760513797650117,0.7992067598850192,0.1910884200455777,0.9461954169249458,
 0.6545910805095899,0.3095564089063114,0.8981520945510015,0.6146501518340697,
 0.592614819924683,0.9649242789467308,0.2253931373850463,0.061265575162567254,
 0.43657190525628486,0.31051947436277827,0.5459654252823813,
 0.9966333478601447,0.06243434740687337,0.17603047023875917,
 0.11468839101367523,0.36619624737080536,0.3167101469029667,
 0.6297536479002442,0.0553160218164499,0.45973761475963126,0.2439155416947455,
 0.39274975847772453,0.370117217116702,0.6041475555802936,0.21074102824204244,
 0.884282370313568,0.9557640384530085,0.9529069883880222,0.3448650198203951,
 0.8622823835902798,0.8377164654518884,0.3707348537055656,0.5923766001686838,
 0.570277375757036,0.9740413364728809,0.6934219801932054,0.3839086314568476,
 0.19329849433328117,0.6577574403921514,0.40111195930438126,
 0.1463648089634415,0.14759420549316438,0.7892797198124333,
 0.46420844918906834,0.06264257083144009,0.6544864709412705,
 0.3166818328102138,0.8208677426759042,0.26203902727058537,
 0.09830590681363471,0.49722057730498226,0.17209347888051751,
 0.7691040223525624,0.4298159576570282,0.5976383488004319,0.7722408006597641,
 0.6484189666995804,0.6165278694235077,0.41219319424642564,0.5430695553156335,
 0.2708779002350613,0.10004266342404367,0.7048484772706407,0.1766663272093868,
 0.03943980215521248,0.45566647942342753,0.7248277421512171,
 0.16194522454542282,0.9265008487452906,0.6590207828062319,0.5108321946137918,
 0.8784450583224018,0.6436180378587067,0.8674256560465822,0.31353920399575075,
 0.1872062491416362,0.5136444434249848,0.2952390917625346,0.12870910575494543,
 0.44512357921050205,0.8408608532012061,0.7591527043015266,0.7260854751024004,
 0.1773375307089935,0.37901744349049793,0.4246844585189271,0.5402937950717376,
 0.0770385169359501,0.19983117427441943,0.9577183394046122,
 0.10544133533760747,0.7893137758624998,0.17203573683704088,
 0.8287956408152175,0.07865672999668649,0.9034887849894353,0.1760266387510555,
 0.8132048985477398,0.8753098232026351,0.9755510835502439,0.3473773203614283,
 0.0764196966873587,0.6990041989577104,0.8951265563651636,0.42782183276457453,
 0.23476998751381817,0.9967351888477702,0.37657456503088294,
 0.011764084075003622,0.965337466874345,0.06216142782775369,
 0.06118548186914896,0.16054541850728543,0.9260966403444665,
 0.1541650106910637,0.8564983366174503,0.15070904086595172,0.7075373025607663,
 0.6373028063484499,0.5677063793157036,0.23366644752785704,0.8819647313892235,
 0.3485356755706598,0.7558471234961667,0.6373814069255106,0.29495806707892924,
 0.5900870511577823,0.5452554793407571,0.01702271632135688,0.6914590841967171,
 0.6573203081445977,0.7509619833895671,0.4067315889513722,0.5745023682804059,
 0.5842286109320782,0.6085845401773949,0.4815319175654751,0.6569926843344019,
 0.5908838004042174,0.48371102956214607,0.22051797158055364,
 0.0690531160225154,0.11784908041611952,0.6620398774999018,0.5163425121779709,
 0.8948398220691001,0.48124694428868664,0.08248110468880487,
 0.6142507944981144,0.08338443912093718,0.5858406010226531,
 0.07772854309451138,0.885007990394406,0.33163870914988336,0.6931178710486351,
 0.2690675600042749,0.3078195259392399,0.30903402447560435,0.7619817886754076,
 0.5281725601032177,0.27762039222162127,0.2926592446188637,0.2844452474676614,
 0.7132971462329799,0.13019531729572686,0.7454519426557282,0.8510815486865737,
 0.4367602956678961,0.9908775204339513,0.024074464552131913,
 0.05261322182968964,0.05042945446196778,0.3518020022067775,
 0.6916908755025069,0.7967182541368851,0.5163676124959625,0.1709487779748562,
 0.04624941149159634,0.8977960422503146,0.44658029701297064,
 0.4921741840588649,0.37487990242711644,0.9701003090864719,0.6596372254624113,
 0.22970422062739815,0.9452311032018642,0.8376684551305684,0.5027506382773231,
 0.2669902255688521,0.7374154841458234,0.08339973905417442,0.8861292986148458,
 0.7089317868582145,0.2736791628633838,0.9881460053528095,
 0.056194042884084916,0.6732552217048092,0.6645617092366503,0.794412941622661,
 0.27574304323746546,0.869307689979858,0.8437807836383012,0.6657932442251312,
 0.8247907536483318,0.13688720995572146,0.10946566659696622,
 0.15152008023345287,0.5512691462305588,0.5041659299738973,
 0.004356248624982184,0.6669034906051339,0.0895868371118409,
 0.4115147630240219,0.6807943620415721,0.8980677924324653,0.2895415525736317,
 0.2815278476336668,0.5814067682683903,5.05158355007751e-4,0.8648623124654304,
 0.21846517774145569,0.5005102711907288,0.8787954174240439,0.6173903338202342,
 0.42178368033552205,0.7569399740606448,0.6948849429690497,0.2772702194955492,
 0.22668280624055093,0.8192312711303109,0.5468502204557928,
 0.35948143634280294,0.44047646677011654,0.6702206473696968,
 0.14100074513403538,0.6544326782561979,0.9123325485450325,
 0.14994685491825266,0.3855369787076902,0.26533697371654785,
 0.42659864258199387,0.5886889080013068,0.6369757864395597,0.2246801954094496,
 0.7518752669074003,0.2483386065575175,0.24796790134867475,
 0.21783483887793675,0.5637621981095413,0.4243242459405576,0.784879360969743,
 0.7257920569078827,0.9110419713055481,0.001105416787226421,0.831198674711167,
 0.31061983836035034,0.9677220535430601,0.48314630006089776,
 0.6702160486066155,0.5604552232113126,0.04920977979740859,0.3282940541171586,
 0.06910757882219276,0.926896283943307,0.4564534172512099,0.10164592375833659,
 0.4719484741014902,0.023263856710077024,0.0023831035655381694,
 0.7100654813670987,0.3795813854615542,0.10559252057910251,
 0.42948615298034476,0.542397931508287,0.562950827428869,0.0878515273110071,
 0.36545783435523427,0.5714066690851127,0.6049577556577043,0.6413817772089219,
 0.08596677748379533,0.1244620393677377,0.37469667442932986,
 0.7712601740901429,0.24157810535767277,0.39307863212641725,
 0.13319612677730508,0.41040191753397426,0.15619878310248192,
 0.597242158137577,0.0029470469967483037,0.8044333540129325,
 0.1704619572693653,0.7821952488393751,0.23129303313677396,0.1780471565697398,
 0.7515581285283357,0.5268445503564441,0.1441746117289774,0.862465724387528,
 0.22455930606490082,0.800641573496018,0.6544724799844398,0.48174662396982837,
 0.31639586107627693,0.5711596517955363,0.945260748893376,0.31458961534906205,
 0.8061682652802273,0.14427451538679126,0.25282729352498234,0.258839205588138,
 0.6936396916598454,0.10570252067584307,0.9757604478713755,0.3331637640369477,
 0.762771574009196,0.3536150726327788,0.4791545915737412,0.9402135824525744,
 0.4996035835441006,0.310856881822389,0.3796909416411882,0.20479479981452842,
 0.30516338864668346,0.8224715689074742,0.5839821381039545,
 0.39858610694716323,0.9185096876307612,0.6703733746638665,0.2461284462345389,
 0.5535773148892215,0.4757590739129136,1.6221107143044033e-4,
 0.7968493357383116,0.15906176068851685,0.6603609924512523,0.3427214979625415,
 0.06819391050644497,0.8935751072443686,0.6569602501062352,0.8549666530769215,
 0.39192626131383035,0.2983765888183253,0.652893460474107,0.34566079810501116,
 0.5677275709891415,0.08525159655345926,0.43583756170165333,
 0.7868593008781432,0.7290990320243885,0.2642879745597888,0.9943342270581511,
 0.24501747051270129,0.976721829877141,0.10969301062538117,
 0.06763783649432964,0.5430108995873584,0.7646437474431999,
 0.35978659434427773,0.47361316213566684,0.8099048932353771,
 0.5242042338413163,0.21276856264758048,0.7344735728279532,
 0.14313970192608272,0.8022944798117382,0.8881463096706472,0.4035112760821732,
 0.48048169979063504,0.5642838129039521,0.8224849007914182,0.6229503450768897,
 0.7153605906422649,0.3305912649763749,0.6974857414302627,0.17572640058760614,
 0.25285440403723025,0.1939334117715461,0.9739624398024967,
 0.28077090627329704,0.8784815786625237,0.6115588515215659,0.273084365912285,
 0.7891249525334073,0.478307357209364,0.009542015814531069,
 0.18982864962078128,0.47224219691000124,0.4300178585546294,
 0.10128198282943146,0.9834780548936698,0.9032976380153307,0.5015756781811457,
 0.04974217390447011,0.151596291363844,0.7464565948698852,0.46981964885813743,
 0.36790627906055096,0.26743735751055264,0.1957517585939712,
 0.7841180510756933,0.5399374031430069,0.2875539379294223,0.881518994060123,
 0.9380202851410175,0.8135049609018524,0.8172740386994529,0.22446658657179952,
 0.8026562634397202,0.6392309212713636,0.20258148102073803,0.9790981818813422,
 0.3082738678305038,0.5846455848266034,0.5409036128608464,0.43040238938943554,
 0.7619122729410134,0.5054403863727659,0.1718277846162065,0.742611418772619,
 0.9309915022420524,0.5664090480406431,0.6888936404676719,0.5707763443997005,
 0.23272249042754212,0.6551361133415237,0.6251866723843978,0.9529939954581605,
 0.917037078228783,0.4228178990163439,0.3342488873365045,0.9835477338850879,
 0.31633710417505,0.4481332916517108,0.6146126576488998,0.42103742024332813,
 0.8244710998778862,0.305611194072996,0.7441456807167761,0.5398987964723231,
 0.4462739749489153,0.5830823591208031,0.5224016068994544,0.5762364962355921,
 0.37028438140890074,0.06381678663174173,0.2411395805743085,
 0.5210749762515157,0.6228806770819708,0.9249384554345743,0.45792827520668866,
 0.3881308717148997,0.5472396076393811,0.6916348084967352,0.9577088862877354,
 0.5467204017736131,0.8796466441897607,0.2840303399818779,0.7390257804387881,
 0.5934214361611173,0.15637080125948666,0.936699247913325,0.11929886065644868,
 0.13776101137127683,0.3523323286133718,0.49196611009760016,
 0.02683491703272689,0.23576317958992488,0.9325378651230557,
 0.3413785408969541,0.9653161345409965,0.9570257968785277,0.02712761042567924,
 0.7452304173765393,0.9577213977354609,0.5557439631167729,0.9440462421822813,
 0.840841868220005,0.7606774100982724,0.8982734977267212,0.08262878265207241,
 0.8133824906858935,0.7925732669325434,0.8455572362344588,0.6861354652310935,
 0.8916690950666355,0.0903284430902308,0.47519059322547175,
 0.46912296040720236,0.28640053681514954,0.9265459388630339,
 0.05533431511300391,0.7413286394175471,0.3201560748032043,
 0.28683237076029755,0.22926246977739695,0.7054237563979066,
 0.6767030585132013,0.39888630918546897,0.636622998611199,0.01931181289093764,
 0.06060611186569198,0.3493254839632597,0.6632517193710556,
 0.08802219370392561,0.9534270415512723,0.4486698650428522,0.9331514594099137,
 0.9060391635568087,0.206956954636365,0.6464927873403336,0.03952163257387942,
 0.39189953452296233,0.7572409106734089,0.5170699720239772,
 0.018040212309381576,0.3364181588636245,0.689629899399698,
 0.020931547275237605,0.35781907523661227,0.6527391265391824,
 0.49784569036062487,0.592716528596195,0.6217270478515784,0.7477758110151733,
 0.08808862889280467,0.8138194957974417,0.3776491486642358,0.8955324186806487,
 0.004923115523738475,0.6125071180771243,0.6272916293065018,
 0.2597926003702584,0.18413703119558777,0.879527181627318,
 0.017187136058045338,0.10972140775109196,0.36222688653687296,
 0.36692301046811027,0.8315214448493141,0.5301738035225306,
 0.09001988992988741,0.9301510329105319,0.876240290846746,0.5218285793136519,
 0.1763527160405043,0.3902074232685572,0.3540902047138492,
 0.007218144714332131,0.018090845109948517,0.45189677595706723,
 0.7333276597650923,0.8686018917295241,0.18507369901615334,
 0.04768639230483551,0.5024713887431954,0.7704885603024924,0.3595488155030625,
 0.05386209816723686,0.3150378738892813,0.5807551906655479,0.7423127973623109,
 0.08825017417573344,0.9661179600124473,0.05753012788723344,
 0.3702118927310274,0.8400004926361455,0.6755388572580453,0.24570354074017275,
 0.4346546357912382,0.14221729162213892,0.5230964904371295,0.862183620793636,
 0.06437970849880015,0.43905280939879077,0.03995540590765578,
 0.7315874364907083,0.6904589186890653,0.3869457445177281,0.17640809438970162,
 0.5339514713800843,0.4141204871218036,0.7756636679630305,0.6709582694314823,
 0.2830677197394875,0.9354376475197845,0.3769333171529824,0.17758997436205481,
 0.9305094261778502,0.27041107556878496,0.05532389545619987,
 0.6487868009398045,0.860929622243886,0.1350952838293592,0.6819473082369298,
 0.6160747206164596,0.6611438647158411,0.5150295505339659,0.19877974516076768,
 0.6567749734061399,0.3824718792471915,0.8066016428063719,0.34747547294900627,
 0.9691194571300543,0.8082870525836857,0.5406310936001019,0.7843708165817226,
 0.3540746824420907,0.3465235142266243,0.059839823002490466,
 0.38550439174530426,0.33334959640528006,0.28106520561534387,
 0.15542911027753992,0.8476381007113927,0.811790098295667,0.37350209655580624,
 0.8859253774132688,0.29992360611343727,0.3424174044104513,
 0.09247514296220294,0.011876751985089529,0.48205822452378244,
 0.4950391202268588,0.9311142892223004,0.5306491728700592,0.7919960695867313,
 0.2250995177298195,0.7591858905085255,0.5800866235266595,0.04098848980935199,
 0.9294175394810453,0.8894310699657225,0.050780741237761884,
 0.6386691812318983,0.6559149101742209,0.4767920311283438,0.0733459514193252,
 0.3593790205106483,0.15466873146848548,0.07646579222337657,
 0.4459426652344014,0.6045605737782582,0.4807015776851631,0.7118520029833815,
 0.6020849615624895,0.7725601598509242,0.6197223081657957,0.9181299102648627,
 0.9388703949826276,0.33618983286382476,0.6396554168340807,0.9907700938885671,
 0.8846392237443776,0.3791731631452613,0.5754470651341478,0.37147705910154294,
 0.9537728553268984,0.98955243674988,0.8934298595307592,0.5668255487089362,
 0.06418765057259979,0.7508255147885451,0.9086025306949742,0.8953990848646844,
 0.11454989761365608,0.06908159520813595,0.11333272477347978,
 0.3783199209710999,0.6054882310462777,0.7873314060666656,0.24481535028484847,
 0.6428208963080271,0.5003505740232118,0.7892515853425744,0.4595266626165375,
 0.47968946569605964,0.528738444672614,0.01820898185057873,0.4402906151401891,
 0.6184929402266324,0.8703584423592787,0.44238573855449825,0.8128250146204332,
 0.7599302917191847,0.45631919240375296,0.015553084778974546,
 0.6838153660157228,0.6904245446805969,0.6269855791879297,0.10131767381377699,
 0.2583079888582841,0.10750856270456155,0.010245401596342574,
 0.3386051965509904,0.21170194682322552,0.04705722835054971,
 0.26765422075903533,0.3063330469718769,0.8208286710823766,0.8852109357767339,
 0.09566575370882358,0.0992469734288729,0.4399650624121024,
 0.43656216986742136,0.3082349493291342,0.17249695397463582,0.941510010670251,
 0.6069489274983577,0.24153686771002536,0.5446045608823155,
 0.04439788984495174,0.1177688509062833,0.1979261523340926,
 0.40752146823510593,0.2130724984253246,0.24567548639110393,
 0.3690700807178545,0.833255175217112,0.3573809996070906,0.6566070441996378,
 0.04944673355614204,0.33251299169829385,0.310551705291765,0.5356366885514067,
 0.5272562231863708,0.023488062538420396,0.9098824841758808,
 0.9194345912955016,0.21461467085337893,0.7318112266833239,0.6187847605446457,
 0.6088937348356347,0.39172668895764995,0.10851385738674735,
 0.19084701154824968,0.47662772958016486,0.3629685256233406,
 0.6855124538444253,0.09424768086493795,0.580475537502454,0.4427234838226919,
 0.4638753754291023,0.7248440073358071,0.5211218944705718,0.9160123640961615,
 0.8291243217318491,0.33101018872863264,0.08807622828345196,
 0.7945416035575272,0.5965888052112112,0.002080912361001719,
 0.017075939686763908,0.21285159419291833,0.3431771837203227,
 0.6814183213723044,0.13852398290943257,0.33142763073252024,
 0.6351397659203525,0.789977853839845,0.20487863261766814,0.3141693216349688,
 0.8243697920045348,0.13332111959628534,0.31333170881626515,
 0.6129830496409832,0.8205745873589,0.3834017369479559,0.9421964341757179,
 0.21096968206735267,0.7089433067158794,0.7753809758256041,
 0.40440884553531276,0.46844177491256356,0.03911564202079287,
 0.575850783507988,0.921366513839477,0.048428884186003884,0.6576980510444667,
 0.044831697439027374,0.5421433558889666,0.3815129221619572,
 0.6371881192743686,0.7853915853680852,0.6218553554198777,0.15145028432509489,
 0.7631434382933338,0.31174894793071617,0.03210284991363421,
 0.2513630621544136,0.9726795702666748,0.08092699834410433,0.8283924923591864,
 0.8571274586514972,0.37912349766218456,0.41470735637875866,
 0.45098866583616926,0.4168714612041797,0.3932803608322455,
 0.33166860055813563,0.43562773225322504,0.950374789510953,0.2675122653562517,
 0.8777562133802618,0.8463856291533474,0.07545605736031369,0.2601997504203569,
 0.13298474916264613,0.17668849700448264,0.6280238674419557,
 0.3048801950857206,0.2604542302844899,0.25682119711840334,0.9070432479871561,
 0.7074204405016058,0.9126791473464596,0.8058055650324568,0.576288327796941,
 0.010908957301095867,0.29520291174656144,0.9235748792780232,
 0.09480162791783353,0.9820095035765463,0.5774906076863755,0.4549209292568197,
 0.3279027010647242,0.8541118503034568,0.26227431006867663,
 0.46431748453527777,0.8273686746466139,0.23455562261155605,
 0.5563919349622684,0.9011964340556188,0.742605639680368,0.612865193295849,
 0.11355004313163986,0.1387463580400613,0.3485713713802965,0.7723674938466925,
 0.002110503219145854,0.14150600314555273,0.08583341523566501,
 0.41150542280736724,0.8289942284662513,0.45194816887134304,
 0.6624203351023947,0.48312193993757124,0.9213263770206376,
 0.07414991314654029,0.45638043440763076,0.590307651822007,0.6248756877229849,
 0.6641821767734688,0.14136928409299843,0.49976959915655894,
 0.2651365914992413,0.019625233125051045,0.2255426203291987,
 0.3535474480216986,0.07353537794085774,0.34376381259670286,
 0.7428405044243458,0.7290032783634001,0.7751066593114108,0.14426729782409542,
 0.6552684824112763,0.12001942765804097,0.12279186226926353,
 0.43770870251847205,0.711450594711323,0.818178159225634,0.7474458789267224,
 0.8924784167606157,0.3259883278650306,0.6353305166955647,0.38503828319731426,
 0.2009890196574744,0.46311928275316216,0.7828512195112018,0.9858366564463377,
 0.4531665138434049,0.5064489672784313,0.6084249729952012,0.754364508783966,
 0.2874411030567692,0.18291411954275072,0.5618215394966909,
 0.45111785096945434,0.25714436388482464,0.09972779174406787,
 0.419198823731059,0.28101472124729876,0.3870202779567926,0.6716435134147009,
 0.06945144109985779,0.4405283428597351,0.28836360153412344,
 0.025450270278262632,0.9782114307946402,0.9576613896737367,
 0.4141273664420114,0.5248353871655784,2.133322050146269e-4,
 0.5095278938857658,0.519341018893812,0.061040355862825524,
 0.20508968527065397,0.3132926258347172,0.7088871759582612,
 0.03681610449665074,0.6475306404655088,0.24770480032869824,
 0.2879621817347675,0.3525932990985077,0.7729788897524329,0.09356479120917915,
 0.7881807134192567,0.5421189603009511,0.9101469866279718,0.5543865921929751,
 0.5246946109516945,0.2718695957922602,0.6491243181948934,0.3140086580122371,
 0.3981396752993174,0.37601072228998,0.7021080867920823,0.8338167210854284,
 0.8484751919333963,0.09087541007412525,0.05608881656975673,
 0.09995487799595315,0.8533315575863614,0.5578040697787393,0.9632110570367581,
 0.7343834096260953,0.9459306770591842,0.46431944799250513,0.2581381874035876,
 0.6303521132630557,0.08155813455836614,0.7335860719799886,0.7216377640956065,
 0.6846379435536786,0.5086972388821317,0.243451700808882,0.22527186764536566,
 0.40103318590712966,0.4567867805404232,0.8058349915543879,0.4355700859137903,
 0.10082168543357672,0.7810466641725109,0.9529520146850504,0.3600470765689294,
 0.1947587082768486,0.5126935798317978,0.07113227295163704,0.6864682330448791,
 0.5476646040279083,0.9617263107077485,0.35025972563977703,0.0861830142091966,
 0.266284715564888,0.09552858874018322,0.8789618659101559,0.5931921884993637,
 0.3435434927033557,0.3594915239537626,0.051367464033997834,
 0.6549104845123834,0.2753038308431176,0.3375044113414051,0.7879203043333781,
 0.8807730154482026,0.6525234377696917,0.3359515327416859,0.3856679390203729,
 0.2532164111937939,0.7744000779469372,0.5473990172978822,0.08595918565940674,
 0.658117426667362,0.15461592992774742,0.11073744878679836,0.9875673634058275,
 0.9206372779368758,0.6661002970301052,0.06037657454125911,0.5372888544746964,
 0.8661590599876815,0.375702846948871,0.815543410967852,0.3396720215060304,
 0.5348917636492615,0.6600681584346691,0.523441822253365,0.3430546883612745,
 0.16509412294954084,0.29184126724899273,0.22036353954281873,
 0.2767243058706824,0.4745171918567379,0.8376844403746653,0.4118764338129218,
 0.6358917224560244,0.16507056414657117,0.06517882989909718,
 0.3715784162262904,0.8430327715738011,0.7025553763367932,0.7772499300813231,
 0.92466026505817,0.8646487407352479,0.7186178073532463,0.6473810687606233,
 0.2894996366001985,0.4394141665895077,0.5981865070918309,0.792936648906169,
 0.48543072967773926,0.27307164335233236,0.9583201719283475,
 0.4794228602565238,0.15520611093195424,0.910666921046762,0.4576205195126596,
 0.5038856877445845,0.09370045264976601,0.5028072712778953,0.9918771228805163,
 0.06761906759893632,0.43255292363242936,0.6002534374178978,
 0.08349032587580096,0.2083878037561261,0.3784602309503613,0.7468567973035228,
 0.247054710607988,0.2688555519847735,0.6767768050779801,0.20862295914502094,
 0.5250615751392238,0.9697105029883061,0.2227295935494411,0.30296447514899927,
 0.09799038716077213,0.43795126241856797,0.8878183188078971,
 0.7786459712379985,0.10043213939339424,0.7800002608736902,0.6326522601480719,
 0.8482825945717782,0.7155123254164154,0.5512011881192486,0.14618786468370415,
 0.999959182868925,0.7469390000562837,0.7958003689264816,0.5046131026214379,
 0.2982170392967667,0.09790098683458892,0.08273534253298176,
 0.012420556771525093,0.8586519725534549,0.023173680727334056,
 0.4336787485039074,0.5330999516471014,0.16936449626898886,0.2074350914254005,
 0.3704318396579318,0.12136997360766999,0.9604785254855968,0.8728699837465133,
 0.8537315658397892,0.04296223784045272,0.940674776210281,0.3045861762910299,
 0.12957321190068616,0.7915942399538943,0.4085353189362365,0.653688732548554,
 0.103003200989636,0.9134369679844823,0.1792929881819194,0.7458283147146134,
 0.4172672648681588,0.9775616016891978,0.05310806768813925,0.9766695272167508,
 0.7328945764663811,0.48625735086627375,0.6754285295985347,0.9613799079429635,
 0.9542624233765734,0.45458404806732544,0.8476915071849447,
 0.045216229179250944,0.515062934995087,0.9599947330661666,0.9472501797375499,
 0.8502903970466269,0.6126769830151199,0.6582891297058051,0.4690992781997321,
 0.8697074800586864,0.24003287733720813,0.23566488516606,0.6251032435246774,
 0.036611363074332104,0.2612021175430419,0.855024654007921,0.4538197488317388,
 0.5084362928508873,0.7777282283064053,0.4682560002862818,0.4726966281935522,
 0.9346281468826982,0.9735336277239439,0.42054366076400496,0.5571267450332718,
 0.023310635429900084,0.29201356937268597,0.5451157453683171,
 0.16569280692876376,0.024018132806740478,0.37544224464651577,
 0.0829118309427519,0.7347158065880943,0.2753548890802511,0.49889876362664765,
 0.5652079045881465,0.47247569621689023,0.8529474281911131,0.7990532820180808,
 0.26965001977847625,0.4426682107653894,0.21599184842736674,
 0.9850691616465348,0.17348671368363677,0.7883645872407967,0.806294474880372,
 0.6409681838753524,0.7478885156541223,0.6459535503230075,0.07589069658028014,
 0.17375370248559907,0.1035550214722083,0.17552993515651172,
 0.35446871066319785,0.1984939965559379,0.7667005140978549,
 0.03973416157742282,0.5380912842163905,0.7649115373670989,0.6500476239001908,
 0.5807737538655515,0.2189789933929842,0.8529668537203362,0.9713375887084024,
 0.9593393956637585,0.6179765506866512,0.36680923639494245,
 0.46740018170908626,0.20881154231478127,0.053059862763236376,
 0.4580277957847101,0.6069211632822001,0.524622614286153,0.3919537500702268,
 0.020524631646789193,0.041486925839560485,0.8536501714392873,
 0.948832659976638,0.8044867098905624,0.8171719186429491,0.3827326647602626,
 0.3721220096369643,0.4141898432470804,0.8858807920584157,0.39546624867213076,
 0.7505205333952338,0.29492470122320014,0.8491017303627169,
 0.43158789637217143,0.8626334865891884,0.9539496143051789,0.8397597978021177,
 0.28783221065387876,0.6853549034852102,0.6724604790347397,0.4884742955678888,
 0.9859241127838358,0.43806056681363137,0.9538021259847629,0.798744107170335,
 0.30978252052263044,0.297393440990799,0.5755272249685428,0.11005147413705296,
 0.9588747170208142,0.8408480178214833,0.7144654733258944,0.760968742205746,
 0.47235526048445986,0.6402245408622416,0.7151601738096861,
 0.13201996886193035,0.9542844947634769,0.43482706056676235,
 0.9151959739887384,0.9535594977437557,0.3473985915205119,0.4341086633294188,
 0.5813026924671448,0.25974864959584343,0.5790007126088894,0.5245255311245135,
 0.47501672025565,0.43581999574915364,0.6169089734276265,0.26598309117787833,
 0.587050693205999,0.936917731784465,0.4288980098699643,0.7418856126132186,
 0.037363672863643727,0.9966085197013786,0.30295214418086414,
 0.676585617372465,0.7527307756420903,0.4953567030417625,0.7824132627577973,
 0.6398615044602828,0.4265381016713953,0.5937256888022377,0.5307001422849504,
 0.6224168525164884,0.7445671742322817,0.9334955168837269,0.23299455741969588,
 0.5163188649447665,0.8288729541647931,0.36180217194228104,0.9146840898889588,
 0.0029674786301819855,0.22447095406958162,0.1267589277759389,
 0.20478402060258616,0.07084846950958013,0.9338935560738706,
 0.16697389168503818,0.5091431836950499,0.40203884901156506,
 0.5158901738282439,0.019016874699912534,0.3112704775153624,
 0.5023435832935484,0.5648939445546307,0.8721522327862262,0.7228919339127173,
 0.7660941134504143,0.3956634673794338,0.5165273298457362,0.14026390249820953,
 0.977425225131355,0.4694829593425156,0.3962360592063505,0.8383924053289833,
 0.423934934809576,0.5862575493544695,0.4545302264320733,0.8306327166883722,
 0.9306739551336475,0.7106158847718358,0.24322509280601967,0.2352381166984281,
 0.17095318326634457,0.2048050321564303,0.5056141598588408,
 0.44452705925312597,0.5237223953863652,0.6113017854943598,0.922387137364826,
 0.5274040428129224,0.6512673493246947,0.2381989402258824,0.8405014020622428,
 0.11776290611594653,0.026096988675248567,0.16341390769153363,
 0.8050864530044439,0.6259751168406295,0.17940614270922584,
 0.37344958831537656,0.1894983404234899,0.7772078336881478,
 0.48191032177947324,0.35207064664774335,0.2694948242127201,
 0.9682293054848988,0.7391896214167332,0.16380345455730638,0.4811701096817833,
 0.8852900412336916,0.24329827209777288,0.991591300057773,
 0.040722747367213596,0.23074743186017366,0.314993624377525,
 0.1457756190982431,0.6042584409299511,0.9349097017538088,0.454236144922916,
 0.5274240240472734,0.9353465714276376,0.4909792787732469,0.9791045807866676,
 0.012831423772646566,0.6544993049334908,0.8324501382235554,
 0.26206710023448554,0.993036541692232,0.055870624520573786,
 0.9599541507638895,0.7257954155856317,0.5345016479232333,0.29123636603053293,
 0.03870487494476116,0.4323405106830055,0.4569521765418788,0.94322205120582,
 0.517814187697259,0.441122178798117,0.06970239630330033,0.8668098862335599,
 0.7367495978256674,0.04581650446281749,0.9790843551251271,
 0.009950585048195126,0.3650691144250554,0.07740450782886799,
 0.4063045122793304,0.6806806536493849,0.4644478532809727,0.2982862511389216,
 0.3423452729016012,0.7257308372429062,0.34993853781882245,
 0.03452329485514549,0.9387791756013368,0.6302047703699972,
 0.031959708262391584,0.043363669428207796,0.04988430496791052,
 0.12683658772084438,0.5881783707828044,0.6669494697084948,0.7022073465326953,
 0.08580398452613669,0.45250177850956563,0.27661747832536177,
 0.3808893860870226,0.9068253856512639,0.9727675635066039,0.7718324116314392,
 0.5260256065327095,0.2209519630403055,0.5335772315729453,0.8841313886795665,
 0.7322629609559456,0.6429580760837381,0.6937913857198309,0.41915488684533253,
 0.6807101223868062,0.9136202424793317,0.08697263836367586,
 0.14959269466620229,0.13914455779093537,0.5969866649551276,
 0.5177503619876205,0.9342624807474659,0.4156040443836596,0.23450765203243074,
 0.029350232665837916,0.4611614416863006,0.34748499055364823,
 0.3476160826578204,0.9137683516457602,0.9097394638713575,0.3544484182717508,
 0.5548109047483792,0.8953669776783035,0.6889012042078265,0.2576427336365831,
 0.7620534433072622,0.41759520477734746,0.0063912346433684375,
 0.9895965549168646,0.3372466998444301,0.07987339117211212,0.5109108582041906,
 0.17413875696738024,0.3162662517668,0.79241268398863,0.017184372912974255,
 0.5346203468166743,0.21342313978366012,0.8210646397942359,0.6481417936875593,
 0.04122546909737823,0.8472981702632186,0.11641619453673013,
 0.7956993293159531,0.5600978626454842,0.8116990969041484,0.03504213458537819,
 0.19497243721376456,0.27456853735947706,0.11762962495997364,
 0.2664763079319068,0.9514810191485138,0.6195838868610958,0.25063757753123717,
 0.1649513953087567,0.7925263720975446,0.2587472546502122,0.30017621593717814,
 0.042455558752075295,0.26095101177034985,0.9282716677985314,
 0.9359113852644454,0.5205165614136115,0.5075256050206457,0.3430105714730137,
 0.903291262219762,0.6234398528311056,0.7093160351338021,0.5169331435486432,
 0.8810280704947908,0.673865241394392,0.5745325624977089,0.1565166990075063,
 0.8126006638994031,0.739045583296081,0.31916953784800894,0.5346591461767447,
 0.8464061218488459,0.4070262061332154,0.36776454360412547,
 0.055376306554623445,0.25015025642078836,0.2338690024257335,
 0.8827733066223181,0.8789712232254836,0.6530631187126543,0.7283077597189316,
 0.7072446984024128,0.13672858313356828,0.8407793165385643,0.8448302117044406,
 0.480033310507889,0.5027609983851705,0.9186341052843711,0.6201080336991627,
 0.2256008951602152,0.3413198262864071,0.8195929881279156,0.503313582746395,
 0.6353936826080389,0.026132822589850502,0.8985511387677977,
 0.6943751289043846,0.8245040470521473,0.17530099489161577,0.6125161628000286,
 0.9969772198422401,0.38731465831264733,0.08523363764674607,
 0.2398135255421454,0.2957484834649603,0.3844871321528526,0.8932965923973721,
 0.626009632336529,0.2561726522576244,0.8335836923056473,0.18567103353803693,
 0.4566587113766182,0.5702045817585949,0.34368727171269575,
 0.09932322156081519,0.6344335664076258,0.6028674809266353,0.9196770773849653,
 0.46907785903440014,0.12863230731726016,0.05868614613231504,
 0.29682824786467443,0.20446118508533462,0.6996547661890268,0.530119812480119,
 0.7586126083931308,0.22548569249004535,0.9395704942031217,0.8142915800179773,
 0.4957510564908476,0.43001045924468717,0.8979057822482375,0.7908150919050051,
 0.5424226826889986,0.2864274419158297,0.529437654858494,0.6786750480856235,
 0.18372723356949605,0.43048541664831397,0.804710688041754,
 0.40807788524765365,0.48433686078500304,0.7359149823586503,
 0.8848777418516997,0.5193540081263118,0.3310106745838379,0.48827350564272454,
 0.8565707591317863,0.9804768015111645,0.19364558625631856,0.526267380248261,
 0.952333483246055,0.9684655405394301,0.8281320433653377,0.42026567249835,
 0.49329856443459696,0.9469856772200779,0.0701071503606443,0.3798712149561191,
 0.9158140675838073,0.8823715917170054,0.15232747423185078,
 0.13096625782758464,0.593597149442108,0.9912099526877013,0.13159206630126996,
 0.35005150730013257,0.0024227477531899577,0.9661128142063251,
 0.055379098764964496,0.879826386196528,0.32674185942455436,
 0.05630355538740939,0.5304588527858054,0.035222620530089066,
 0.07550898929398087,0.7586966396496424,0.39664925628727643,
 0.3371573067239424,0.016457528835086888,0.4690708305753363,0.253735334631482,
 0.7799943138915315,0.02243043153249802,0.12614171501461258,
 0.9258748417678047,0.43804659012849534,0.8653774368112251,0.7370464822247574,
 0.11999304321635762,0.9591064671088729,0.47110162429322766,
 0.1032150321773575,0.9685103751005113,0.012381637769186549,
 0.9204829416810975,4.1323752636790445e-4,0.6357417499464774,
 0.9090197245807103,0.052349405002002136,0.3991941021271971,
 0.1435988094452424,0.2167118593239803,0.6757880797482512,0.43159660055220883,
 0.244914807367846,0.4177614855971664,0.6192619531762775,0.5843204365595209,
 0.512003606351089,0.9781283906658014,0.7954353459978235,0.3512624139844468,
 0.41629608696347786,0.5837855879126244,0.509297566752524,0.8312286728794893,
 0.9561586653641443,0.7630723791370748,0.7166907228925073,0.8112855803681812,
 0.07909847417999716,0.6088361301706278,0.9112637385789761,0.2930249335746752,
 0.6619957397443832,0.8730799347620438,0.9378017644837826,0.5376630962652517,
 0.5325562879129564,0.6837294553485922,0.7139154986572418,0.7681510895355634,
 0.45459945314102046,0.14111575966496448,0.6086766763084945,
 0.7078177038077833,0.5031573475805357,0.9055864580386421,0.6294388227438046,
 0.48230871825512756,0.6096565584282069,0.6821309700656208,0.3168112344371308,
 0.3076602552809372,0.7588182105142284,0.5079672659001749,0.3918875066558094,
 0.04067923850253785,0.6333420718296742,0.18655858550878646,
 0.19926628395921664,0.16838147827761318,0.6378799671228267,
 0.44093146004315364,0.6051865306302052,0.5298487843604311,0.5556242872834054,
 0.4989697971601521,0.28794189024467687,0.9549771800069872,0.6721706809936318,
 0.6402548784748877,0.7140570908069298,0.10239976287638186,0.5159187500298162,
 0.6080240945519699,0.4532438596901507,0.9111131506449079,0.8503103965932203,
 0.2743118002374695,0.7520566398450911,0.5319812286319301,0.5133709654779037,
 0.8575813717944797,0.6966547445842736,0.6799869490523013,0.9577524180829982,
 0.5983537742311522,0.2573176378668811,0.022739514605153555,
 0.10604687723937734,0.11484508312054131,0.43457555511462187,
 0.3409793246546581,0.4490957791021466,0.5537479725118528,0.760219275668,
 0.3486180850207864,0.8053946154073044,0.6631053437981803,0.7256887780028853,
 0.9807162697878303,0.7502039826299154,0.2868986286216151,0.1348693085554703,
 0.5518942559164698,0.06746365312584568,0.8285632410377075,0.254826500125512,
 0.5910772734410265,0.6002943243054384,0.41943998112978986,
 0.28972002284544607,0.32227778349428504,0.19071590031596675,
 0.2162966604794172,0.6949861308924333,0.7808295590981277,0.5938603794072682,
 0.1620645618920138,0.29383607816971136,0.06688276693514772,
 0.5354421562242074,0.9809174811722405,0.2005918665789307,0.949585653795271,
 0.26623846856715805,0.4672191761285148,0.9944415007575372,0.5085484761310196,
 0.11706953387839547,0.684355366301585,0.2596432032767537,0.34824038101200816,
 0.6705338106154481,0.9956562777799355,0.5208451586065433,0.7815599203516609,
 0.99288584689577,0.4334173640773523,0.44564788757930607,0.0975765254235168,
 0.20585908556584676,0.8399048657361354,0.7791413473836533,
 0.03157587144633178,0.2701425173007709,0.5775288380199202,0.4078255393734789,
 0.729537081715539,0.9320373684272444,0.021589635394115292,
 0.26962573391538935,0.6412192921102815,0.35531697412338115,
 0.14375408013287216,0.2027729223112511,0.3930703740086591,0.779007498136172,
 0.4643900723033555,0.03504358909064054,0.44478747489615944,
 0.7726145404265745,0.13623620659947733,0.8483062198020804,0.7701946249745306,
 0.712106258809818,0.36730371471360823,0.525157594436597,0.7774593050730622,
 0.8046525168980686,0.7259235397023124,0.48449379602433174,
 0.40027788516568064,0.0067798463202553005,0.06497464485461191,
 0.9664022059779562,0.574702594279429,0.4219842498973023,0.2943494410024018,
 0.8112921045418875,0.18798585161468717,0.8471760717262916,0.6155165657907374,
 0.011419240290336474,0.383537872039154,0.8777262133276249,
 0.018088109674488817,0.9679801819086478,0.030162850392266294,
 0.06719940094278076,0.02106217249412201,0.18021415311357736,
 0.6700614331953086,0.4272087544737587,0.5891481166874071,0.6268842221405796,
 0.8168239011890084,0.019571760472735233,0.19512373833235563,
 0.7755189657444072,0.0508707255559987,0.1861883456559451,0.7840596054753118,
 0.2948266233250928,0.8104823629022169,0.12288073077332451,0.8507962763748051,
 0.1719212410448403,0.5429701119204505,0.522820249448872,0.7060859244404258,
 0.19877337699949116,0.993643391102365,0.2331209388337281,0.27085159393045577,
 0.9196000309696454,0.36280263065370344,0.6736832478027812,0.269330693981207,
 0.37466228301710824,0.8635825030348799,0.09072114971420697,
 0.2549051602393213,0.8253854305948554,0.7737012072556884,0.9556860310956681,
 0.7047625301254584,0.8585716780321904,0.04544692607280809,0.8586200749599272,
 0.3578871588194581,0.06460776133422341,0.2731681922961243,0.3825800799622989,
 0.45603853431420693,0.4983389216283821,0.36813978471981157,
 0.4543655312467181,0.554680712600241,0.4703958963728676,0.9963247618229046,
 0.3481885338042521,0.17566448319558892,0.21109472613001412,
 0.39520491175452555,0.8923731127296799,0.9064237153804618,0.5855172431339928,
 0.6058052594289652,0.24117257979978313,0.2941247370510084,0.6154776684200954,
 0.2602911431855127,0.2629799610476701,0.5965176233660391,0.9090252418194003,
 0.9021456282902292,0.44702634884360426,0.30089130135677966,
 0.9756884036152776,0.9413707258654285,0.9112481590168765,0.6832750405441955,
 0.3721788520928646,0.5822192015822917,0.20130855745757903,0.8809547954715948,
 0.2967591566007697,0.08078789846404821,0.387660392781758,0.07472824978862747,
 0.3919216432224484,0.9023093426406497,0.463467608987825,0.4247207908102867,
 0.29818078814824134,0.4192991898876338,0.49914801754870375,
 0.1417090985872249,0.886361104764429,0.02453013492087841,0.15701019519309067,
 0.3243172675503102,0.23629305215488472,0.461099361936139,0.3199108219737482,
 0.35278384134267393,0.3713828284049099,0.2058637472701026,
 0.051677619611061765,0.4270847963250324,0.42134973054895364,
 0.7966493398955501,0.4011712490929933,0.6582995036597812,0.774061301351448,
 0.26861658049113424,0.979871823467755,0.5396837378679015,0.2920818750328137,
 0.5347387580352972,0.9784609317133385,0.19548746600218458,
 0.19414157951310662,0.29508703036453254,0.9839394059489002,
 0.21289184343171086,0.9757384841674998,0.10325955885830351,
 0.40544349069017094,0.3893767559176621,0.02970759972411996,
 0.3900739696601272,0.26792364865299034,0.8671278018685796,
 0.41054387942196013,0.8611362895427653,0.1666772737031803,0.4604956917488121,
 0.7647071637428728,0.8114781078357722,0.7701663503059015,0.7873732280080185,
 0.2568366417593606,0.6051760077294999,0.8913827807417342,0.11268748465109069,
 0.5639101416637988,0.9711973357600441,0.5095795606639316,0.5676386073935163,
 0.49787047323507405,0.5118683612969341,0.5371755258730269,
 0.19558675309009033,0.4871824491356812,0.663192915099295,0.9400497724574768,
 0.26995120061586664,0.7597610355400352,0.7195064956933329,0.421244860491384,
 0.6287511786242627,0.7838368299616586,0.7751209974012023,0.5205924808587422,
 0.3774822558106089,0.8092057235535985,0.48619376120070723,
 0.15732784850232107,0.781491826242711,0.9289636628063355,0.26728328623068576,
 0.5970890716399553,0.6188482969311763,0.1172962029228074,0.8599060686762808,
 0.3502798933523191,0.7300724779131911,0.7427946399975227,0.1082823517384558,
 0.5039880893468406,0.49879756200234726,0.30394257553307513,
 0.7737016260889142,0.9376679939074344,0.8080502511767428,0.3067110365435326,
 0.9023169546645506,0.9835472025953703,0.9756894205983253,0.341837272494243,
 0.8096801790582405,0.4570689879609462,0.839496028305723,0.49016466597782093,
 0.224684806347481,0.286465395537729,0.9263103541583693,0.633011399806847,
 0.07713557657373771,0.663022832045134,0.2252562556805323,0.3877993092382971,
 0.15998218670546027,0.29099669691405294,0.9924544951549269,
 0.5261822297661743,0.5502791013592851,0.9775534261635024,0.21056300097405212,
 0.2188068973732793,0.21599116844280908,0.44492196019424,0.7022208808150868,
 0.4583445078819953,0.9818482689822332,0.05987167837951124,0.9890350035045332,
 0.6304426006154998,0.07791315047173308,0.4650613452275256,0.5772001177725657,
 0.3958035663602947,0.817790263210191,0.988082827054919,0.17164676039070037,
 0.9147962224658055,0.9272706108526749,0.08659107589133663,
 0.31625237832980424,0.05390150927065862,0.8661336389454442,0.81260200685564,
 0.025846230984658325,0.4632829613239249,0.028917630678247264,
 0.35531643718190553,0.4704648756852744,0.8742545325486919,0.8675106853179643,
 0.560722656541492,0.4011738674602632,0.2794106639767703,0.6738773810823203,
 0.5181210733467588,0.37533949492390695,0.5372711916026126,0.9678403310366686,
 0.06983144173933609,0.6795374967469273,0.7154432532845653,0.6024133902213586,
 0.9810934350947063,0.0532455452656464,0.1489220289275045,0.22458632352550534,
 0.6338297924856082,0.0459723991716291,0.21889251954371058,0.7008789079536356,
 0.6602048102023662,0.4793295160446398,0.3393586912026896,0.3421618482555362,
 0.34345645206088893,0.4944941563340852,0.7098631920632077,
 0.44247937952241245,0.46093356841803246,0.14837153931977487,
 0.09658569679036977,0.31230441910885354,0.9412214519799087,
 0.5981391656644839,0.7628097705947621,0.14511399306597672,
 0.055935129818478146,0.6815041592942495,0.8319072119118313,0.922709316392958,
 0.1896201857738311,0.6361078879402573,0.4696684863862185,0.01544113650204082,
 0.12608903365999824,0.8074924012831384,0.06753924897101182,
 0.6029526314350755,0.9215804804197854,0.31932096564498114,0.9461530637342398,
 0.5065360099123365,0.01284526334114322,0.8269402197491308,
 0.31215050260516586,0.14204371256849835,0.13987270848746713,
 0.9271789480682396,0.6726565728992051,0.9062503755605139,0.9918248183196707,
 0.7681139853742652,0.6273702489257955,0.15146132510473487,0.9331949832049949,
 0.22962884496903357,0.37948789447644127,0.868885023135814,0.8752485765659581,
 0.5808231198550954,0.7673623630488219,0.9208644841528848,0.8197209505030363,
 0.965087481360631,0.2299459219240022,0.4502069306737504,0.6757277518008115,
 0.024804548677580307,0.7222068746168493,0.14832288550572925,
 0.6261515277689546,0.5917849398575581,0.687663754117151,0.42877132986622124,
 0.3847085195864791,0.8302477764972802,0.47731005320433606,0.726159371131953,
 0.16110809046809682,0.9218789437044697,0.6556124872082147,0.65704112686123,
 0.92514189583799,0.5221016266192893,0.5449976451380584,0.6798591304449374,
 0.03367969845369889,0.1713421227246561,0.7677336540839562,0.7104061959841697,
 0.5941970092686079,0.9668726990382365,0.675017937581585,0.027273986333630873,
 0.5133695389113576,3.37754157094472e-4,0.4742774902557467,0.9474953330482546,
 0.031073476794209465,0.662818777902998,0.6534970803664749,0.865910318164703,
 0.036740138949399315,0.08952881871101437,0.00201914310004192,
 0.6682061105002537,0.5470653111460126,0.9128254215622473,0.07655463414883235,
 0.29899331051805067,0.282432401199447,0.8137816920985734,0.2876133857610641,
 0.290040419454229,0.7309469139127698,0.8139707250201912,0.43720538989273583,
 0.44486534074131734,0.785773008978707,0.9634569581232821,0.8046746763573333,
 0.9921963959566757,0.6794172084647822,0.6476706359068087,0.7390968480932271,
 0.5824888035073976,0.4801116596937657,0.869247381869497,0.8231184172328808,
 0.9699583891595244,0.1856936263986979,0.2169388985723627,0.5338144728936312,
 0.10929864226909869,0.23577266877882408,0.4992532154661635,
 0.6611388764308628,0.6797853806014335,0.6148854687228096,0.7921866745724401,
 0.6014033476766014,0.3926461301088535,0.7490646557853027,0.19419075629594096,
 0.491244133517448,0.555426427413534,0.7391593573110887,0.36068026914556617,
 0.8779256537346957,0.43922663034114806,0.4252226894215657,0.7545912681198738,
 0.06655066046188163,0.17892598109875002,0.6173269475815477,
 0.5259298453404601,0.09153276495042528,0.8064134034050842,0.9654770432492056,
 0.11869355081528754,0.3759714539471949,0.40097550467781296,
 0.9786404130273597,0.953238617457461,0.3281678878779468,0.6245735904194356,
 0.08596270891542312,0.45123037009914835,0.14299693481640485,
 0.4358549479032381,0.5546041090484045,0.6175673121517006,0.7650264936035086,
 0.371151224024738,0.549770683185456,0.8294986860776783,0.21148084753501406,
 0.1813347956419249,0.9860673634297814,0.80316181124635,0.003929636250101565,
 0.7575668572013043,0.9351493116590865,0.9444738122645352,0.9195849318941844,
 0.6368631390877868,0.1417769801964699,0.6311740190399369,0.9347532728821393,
 0.7950137775562097,0.7269457441184958,0.22353769215143016,0.569316466749892,
 0.35706142673287855,0.3832627773325957,0.09712406375398397,
 0.2967092000853341,0.757049042675599,0.18564415870929252,0.72657440902971,
 0.7298425981618972,0.12777423185006898,0.9434420941815297,0.9919582278257116,
 0.40373616935056833,0.40315964639208524,0.1575769429920053,
 0.7398873664836472,0.5633397110852029,0.46888989733950226,0.9883464494614642,
 0.9840096140019431,0.7116572447407242,0.08171396471183445,0.5813387846350021,
 0.03327026485927387,0.13713379823379657,0.3715325377843126,
 0.10042433630296443,0.0710818817796951,0.11616899110871726,
 0.5272556839348739,0.05513836959139362,0.34393489365306573,
 0.10952004184389907,0.0921712924470216,0.7296866107076636,
 0.23915844617218107,0.6455866905878214,0.6881051044451857,
 0.20381658973415429,0.789487410868633,0.9599840732071795,0.898725050703897,
 0.7358152265627812,0.44959018704360554,0.08342165155259484,
 0.08196398699576479,0.9923455944501108,0.7668070899005739,
 0.34041544264036616,0.532043831701422,0.6419462538406703,0.544621065489965,
 0.4411650129731215,0.57722023950302,0.7220667307221006,0.4120865724069247,
 0.23996356859156587,0.23773745636088062,0.7695914698053941,
 0.7860364685686969,0.5878452530807144,0.14118881088810853,0.6174616562191139,
 0.5444751966527743,0.7802782384105621,0.6309347679489087,0.5943770958039034,
 0.8643025925748258,0.9527610901516387,0.11358889545738426,0.6185105570645952,
 0.2193833072884399,0.6047425726550507,0.8437851233823287,0.7446968204433185,
 0.929574885941767,0.9019530019271912,0.7577457385745804,0.19741033533800856,
 0.15886684955718045,0.48312786582785305,0.7627735772661507,
 0.06575745396754529,0.8742780790905642,0.35332204061341865,
 0.8616228203076561,0.192538135473734,0.6864040484538824,0.6688470060125619,
 0.7533460225986113,0.9853877561884207,0.013268168776336187,
 0.08265534152726417,0.5154240550443306,0.8957271663213777,
 0.39072354644798934,0.7708537326496476,0.004468565979115491,
 0.947289268907056,0.24579715543310732,0.4835862745460008,0.2595348305111307,
 0.5683935799235492,0.3795213712100761,0.989312577312807,0.46523115638675705,
 0.5146342304402033,0.7305199807669012,0.8499176005046536,0.231154323647499,
 0.014980042984827024,0.38343872155643544,0.592031948439851,
 0.7422301247822117,0.37879653915686795,0.4739489861172195,
 0.11493175528935706,0.0856952080819251,0.8296399626941684,
 0.25811451167191246,0.5887140464078096,0.6941195391201209,
 0.006207634408202489,0.027126933007599052,0.2770803232817235,
 0.8201040269661611,0.4040630291804229,0.7085934456047268,0.27167986338880223,
 0.6397504282489512,0.9284123536465552,0.9755294587730151,0.5210947972568178,
 0.19747379586762115,0.44676388952778434,0.5122541227923469,0.572538379845995,
 0.6832095914155952,0.058471855907242465,0.20622056990087834,
 0.832260807966167,0.14818884653916886,0.3732563852522741,0.529096257518571,
 0.570414719690673,0.22052001850360814,0.526465006621965,0.02077914435435646,
 0.6378325849272938,0.21584653469982562,0.7831511336016268,
 0.016694587116165893,0.5736761781814577,0.3187242176153634,
 0.39226777074250385,0.6592279431055653,0.6979511760905786,0.7215493019266603,
 0.945864547797673,0.9860577304169615,0.12066356752479845,0.6952753329509024,
 0.7655222617448267,0.42077081401616645,0.9975799901299702,0.1543794681159809,
 0.13084283819180553,0.9518653779936688,0.008018175746422718,
 0.3515035067175717,0.7907130461777603,0.6068937303108077,0.032990406548007,
 0.8841205625329702,0.6907810839999332,0.7254461877753211,0.7992766788721051,
 0.7053222551285758,0.4356704670367071,0.6908879023844969,0.8462496807355513,
 0.36889538087693285,0.39970801727503147,0.6122119530982559,
 0.3075846815821772,0.31454370899951056,0.8553984151021062,0.108436871167926,
 0.5675454508597818,0.8326355884886468,0.3000364721131006,0.311393654555128,
 0.9843594083215923,0.8771508989112723,0.5644126741007586,0.8331751896655768,
 0.9269381880989171,0.9951358586981018,0.3664221613963471,0.3182132303501952,
 0.9832987526679385,0.5267170270866872,0.23160446846492488,0.801132105946241,
 0.5949285604709392,0.9948623783574894,0.1937859755524266,0.9738198622959441,
 0.16212009225578028,0.7801898701616219,0.2203146815473177,0.6377859908748819,
 0.36357632126673367,0.17604872612284206,0.5049062811503201,
 0.5863297845758667,0.21048300544678888,0.3287449276036568,
 0.07507260127938986,0.0941891642488728,0.4139140056488616,0.4106256636665435,
 0.6343968039349203,0.47375367552212355,0.25701556744666076,
 0.24186844216513936,0.5899733941288143,0.4555141738556482,
 0.19832334177068245,0.42107926995654565,0.23674368815230995,
 0.7279716087715686,0.9698726199497998,0.4931646176327492,0.7126301983676656,
 0.4239681423806879,0.13037390903730717,0.2746236594344591,
 0.16919494081055353,0.8578714607411166,0.6337873647070175,
 0.07399850459049484,0.5953803908635746,0.34946976665173846,
 0.7346573585385182,0.2118603041045618,0.3211575406649766,0.12352863556349314,
 0.09903336136358631,0.7142385895598337,0.32246096855860606,
 0.8818622095761918,0.5314374247994411,0.7850226382966041,0.6515623395671268,
 0.6219731282896671,0.08727961768376957,0.967230708575773,0.25701387624820726,
 0.1286892292679631,0.014489021580055383,0.14911934768015955,
 0.6662676041703484,0.494118727995986,0.865345149054956,0.9331273356526768,
 0.34520740159887353,0.012248225551385716,0.5756564879827419,
 0.1766056114089375,0.30384312174995776,0.7188861660281902,0.9899257235353396,
 0.23943057079687524,0.7630902232197295,0.7290717753342841,0.9547429409251174,
 0.2814627693496553,0.38346704148162025,0.209093037048518,0.875940428451397,
 0.45093521843602236,0.44414956281477924,0.019064627635699827,
 0.4152986496780039,0.07602001847902062,0.3745415774614106,0.8064775001212314,
 0.22795633929442083,0.44936484932213083,0.37638890418972226,
 0.46064635925140485,0.73549338465557,0.5682087270394897,0.05282019509902636,
 0.7728759771125591,0.8032366603351101,0.47266972399723683,0.3505554245428171,
 0.9649443610370027,0.6272823690013535,0.10954517880970127,0.1868812213308385,
 0.387459308198873,0.4653418976603261,0.012584955622253036,0.8414212338993756,
 0.8163803483661564,0.21633838902770908,0.6213415709930913,0.9180157626364275,
 0.9748055033067525,0.9395698205091971,0.13200361792843363,0.9815958436404509,
 0.35527050685112527,0.6901030647181423,0.6553183167543907,0.146060047540562,
 0.7911440463677666,0.18397425433547443,0.5597339627486586,0.5066659212913216,
 0.5536202355818424,0.8504757142493012,0.4403034007662273,0.5161770262114121,
 0.9118989771858675,0.21639393911921423,0.935967904259825,0.3327218650182019,
 0.09935460573523436,0.8029907977810573,0.24719678999139993,
 0.1824668551128219,0.5357964336445047,0.8623507903425459,0.2524265850732067,
 0.7760313009708492,0.10997864407350522,0.6790335893552218,0.6629607051441074,
 0.5914084784246396,0.21167712085937596,0.4199820469676482,0.8685450654617338,
 0.807482947255044,0.6287751436342308,0.026311748607658947,0.283913452822697,
 0.20387861161805976,0.9586622890876718,0.8582489487118181,0.7511482482202811,
 0.40705475820274006,0.9134654254641678,0.13768855128937552,
 0.8538795280714763,0.2719506545017576,0.11859769678278786,0.6161114373446638,
 0.037650609417987924,0.5831131518362241,0.8151813809346824,
 0.9464232906008612,0.2800892976459364,0.09490289883332148,0.4410173839758824,
 0.6211492028174261,0.7863983298322209,0.26433727762728054,
 0.27792022724631427,0.694163682481664,0.22897453542301827,0.3478126446069463,
 0.670075068857516,0.6384079015432803,0.8862317784718273,0.12172576150667136,
 0.9486045338943895,0.14966836784449122,0.3806271027964888,
 0.18143199056801063,0.1686834827182765,0.05895260173444383,
 0.3754209778229636,0.5690205127912298,0.22501995977986844,0.7782765882939967,
 0.3336244802668856,0.08718999829629914,0.18179723355979827,
 0.24014311260141663,0.4505784608907428,0.2717038267739148,0.9248387857001785,
 0.6829384250724015,0.16198439913200668,0.8486729722839119,
 0.08565482173647443,0.33159112013666825,0.9817698404819237,
 0.5993605591987836,0.12248978243935649,0.3370943544636169,0.2615166116363572,
 0.2940026382935428,0.21926078437933327,0.20281314300723463,
 0.42015836929503114,0.34044268691597235,0.7695901074222574,
 0.33935148480197197,0.5354437021913294,0.16752793895378593,
 0.08050903735634518,0.40990294355595025,0.2473839090492651,
 0.6856645049797524,0.8628311708258679,0.7195132131736265,0.7967566015086947,
 0.6606554268265235,0.9083455923393459,0.20379179643258738,
 0.20762878309470012,0.21787554185955005,0.5885382076157948,
 0.42112920280089416,0.09128472381962505,0.7944240861621599,
 0.7385908891043331,0.7331851494582866,0.9494588790005777,0.990068916874804,
 0.2366194548700461,0.45792244764206036,0.9521776873249974,0.705417578911097,
 0.6092361485002151,0.8707739171716252,0.322353531829478,0.41384353483710035,
 0.5842450098315082,0.17393209592686976,0.5591106091039284,
 0.21131992013841527,0.07572862138587,0.5546250594297651,0.7293836193428606,
 0.7267291618598699,0.8798400452547361,0.26921133951889875,
 0.09476783662375232,0.254533058733189,0.8726140665253358,0.21467613848881661,
 0.42539343294027565,0.7695740867481414,0.6127159580289371,0.5956310780750642,
 0.603708048021149,0.4962848036446881,0.34575635769558866,0.48137812838752514,
 0.0082669618042579,0.9325716666733281,0.09049084598750934,0.6657334401960778,
 0.31140778981372874,0.19600094952053515,0.043773881457539354,
 0.472046505893428,0.2969211337056261,0.1586936860541407,0.1965689643789167,
 0.38028919920554993,0.5338641097479547,0.4787618102998892,
 0.12002627952900624,0.2882898239091948,0.38052931707149407,
 0.8995415001478461,0.5638175843287709,0.07269602366100358,0.6027352620177752,
 0.5787544238204128,0.3864898555541767,0.9080516158579522,0.20296704091821638,
 0.20902979429619306,0.33755526799345503,0.35570223711992033,
 0.4480585608393297,0.35164842453292633,0.10070630715478601,
 0.028687524748108983,0.5579371325348483,0.3377675766886561,
 0.47696301147330133,0.3718292880024491,0.3983774811106321,
 0.43383708379869856,0.29982498253309575,0.05129279069381887,
 0.4895229405522614,0.8122278696830459,0.6718565084339269,0.9215255050600101,
 0.02818661944124512,0.10575007995036123,0.5804312299330402,
 0.7932894536502145,0.9470491027527269,0.118614475288214,0.9516210937879328,
 0.667954264829089,0.09412993528844682,0.8526689032765423,0.3744165917606538,
 0.32038111569507877,0.09565347804243451,0.6834849502986897,
 0.9309362089557286,0.5353019157619867,0.1382581602537074,0.16616383348613628,
 0.4417805051189624,0.003041741797323416,0.5748804143329066,
 0.01655019725825979,0.2552222695597868,0.532701614459504,0.10824598694851817,
 0.6555568843805315,0.0520447008289755,0.21941231790163052,
 0.30743691244748217,0.7395671672844063,0.3977370215970537,0.9579816768024166,
 0.6470798929811934,0.8908827798582749,0.04049133765304358,0.3271154342044069,
 0.18450634214658135,0.6479627348980114,0.6715964327577477,0.6240129711816144,
 0.9663286450997698,0.2553849385616761,0.9629147860602407,0.4995542887078088,
 0.39358085292538614,0.9853700923822737,0.6128955330309591,0.4467688669562472,
 0.8884903217657664,0.9490870324900569,0.34188056644403664,0.1543708786025031,
 0.5163425791584892,0.7038502135821104,0.8834515471779485,
 0.005210168495620837,0.05089905113757154,0.44486797459811944,
 0.6496638598986834,0.8886390805415075,0.3196492222542041,0.1084282611631795,
 0.9897770867082195,0.5583084973223336,0.6268603905790434,0.15992261257246387,
 0.36573521902963824,0.2806370839901037,0.058271434691246515,
 0.8870385598375835,0.21032715989605077,0.4406575561417855,0.9152851260395781,
 0.22894298348826658,0.24394024696524097,0.3049161633815337,
 0.43520008697188195,0.04859999964835693,0.7817402889818149,
 0.4984750162599919,0.6033203822883075,0.4357477073861613,0.23600584039178385,
 0.79315350831758,0.26504107116942754,0.3337648580375411,0.38887835912119295,
 0.4271895924779512,0.8143922659338402,0.5441209787184741,0.06438854041856157,
 0.7996805490790793,0.07007496091871102,0.07588584993905967,
 0.11549005652080557,0.06049572761062205,0.6094919187272502,
 0.4881067705173051,0.05644922875812264,0.7209912440121707,
 0.019730059851527426,0.13563972631794496,0.756396342944666,
 0.4421700543501851,0.3140506791792387,0.23263274715923066,
 0.26577980911183663,0.7228283437406053,0.30110485785510177,
 0.9878468411458108,0.37754626564030647,0.5610673696368041,0.3949265175131411,
 0.39482522854585067,0.30308542634605296,0.9800203814550423,
 0.07377815142621769,0.6537177071413004,0.5370534633017927,0.3034675870045108,
 0.23464490181824504,0.9647095779214736,0.15339302867396087,
 0.09798133946596732,0.2260666368031351,0.10152594731559872,
 0.9514775826797214,0.5194008605493778,0.5626090623453855,0.7008192256215943,
 0.9243928034066842,0.3790253628148228,0.0330369712325308,0.9019922097700712,
 0.05166655936627251,0.3287966621247128,0.8697478787678992,0.9272065821272555,
 0.08832302158533833,0.213710106963374,0.10602877858523063,0.374669792096731,
 0.9237311262824266,0.4528429394977487,0.5154379544117078,0.946897715155657,
 0.6748900169755478,0.6059465787318667,0.040677479448087306,
 0.20257730850654498,0.9711892536235514,0.030398184297728004,
 0.5094899554778803,0.4185492748757793,0.2730755706976985,0.08183943665131277,
 0.7467079940581104,0.804561816515038,0.3561668263666695,0.37605821104297366,
 0.6727757820805828,0.501104067605928,0.6248993517525203,0.2917712436647649,
 0.8089763802703864,0.5050989644519561,0.8632161226373523,0.9878310913321565,
 0.3465078849972574,0.9905248119946503,0.8967933556786207,0.22114862901538335,
 0.2838565093373876,0.011829764833206857,0.17935839074207727,
 0.2095727953968961,0.9594107525435107,0.17527828409300827,
 0.21652559781633773,0.08180981144357824,0.1457600974578821,
 0.41112340539157977,0.6397967862690477,0.9568528257500478,
 0.09268398647258058,0.8296095862099411,0.4408980280866446,0.7040477708908902,
 0.24271375510729143,0.4920955043118851,0.08852680912105226,0.657291787762885,
 0.2149406215420422,0.578329706693977,0.2649983115570552,0.9491604616086861,
 0.25083534902730265,0.36263121115295993,0.9001283453120044,
 0.37633030871317563,0.9223530996977694,0.5815235060121696,0.6863781729489723,
 0.06912211092448683,0.618841075336003,0.11889095147938922,0.5593250684467682,
 0.7290831991751238,0.5206803765682829,0.1931077536009409,0.14244142890886724,
 0.11939242799069394,0.20400502235815643,0.7919273909275621,
 0.9258778135592306,0.5522246943448204,0.7499981256476487,0.9583190375040203,
 0.4516031798652099,0.5264982787694609,0.6810227522847383,0.7953821614612745,
 0.11745581875438404,0.5503849447926144,0.6272297112402605,0.9825389166672661,
 0.43609156536888494,0.8426046672258836,0.32497710906319033,
 0.7056486045057656,0.5537699459696399,0.13101007947278553,0.4370879474914958,
 0.8110873019814647,0.30829839934217484,0.09569528233369695,
 0.9770996905634259,0.9707890618740976,0.18690375447020657,0.826033453624574,
 0.030964680096633668,0.34518670365549464,0.9551287412983074,
 0.7159654507093971,0.560127887411283,0.11560196972209058,0.8425480493626861,
 0.5380405466889048,0.9480686974372163,0.08275478316474993,0.6773562353245589,
 0.25959865108768876,0.27958322610089426,0.7721041637380933,
 0.4519521401429739,0.8550628349615194,0.8865924694929036,
 0.027591669390857332,0.9983507784005925,0.8861732022688428,
 0.31500160560599977,0.5141618505399234,0.9017294926392069,0.0629915651478502,
 0.25002983007995316,0.7750435993085792,0.696007478312372,0.5248567407000615,
 0.4278855221754667,0.8782021227960417,0.20657841143276634,0.6480385768089884,
 0.928745954916849,0.40291054972384,0.4982986328317085,0.02772681372793584,
 0.46285828531548934,0.5911613461452829,0.067957344069976,0.6922758294843294,
 0.8267008076798378,0.6289785657708559,0.6440585403337662,0.2397694001854077,
 0.5702226348181388,0.6542057269878271,0.06378210570594,0.7408498800569369,
 0.43254475423133665,0.8780002517093805,0.7681399833771093,0.6401647895998698,
 0.38341556281344313,0.6486897363805976,0.6010659764847858,
 0.23245545425689584,0.7112209269751365,0.5273408602682828,
 0.11902765198424914,0.9650418901898556,0.2773409521980734,
 0.007313767998715415,0.6777778232995402,0.4464493786852478,
 0.8400075142430056,0.12511849714723366,0.3885988762700412,0.5833144369009483,
 0.1182005701125377,0.2629396781069876,0.7808152825113389,0.23443243954176152,
 0.2175800799622859,0.5330727818340828,0.16553562065987737,0.1556205459128801,
 0.48933524794402794,0.5920451060852621,0.6988089497372523,0.9770858612425956,
 0.9960545483566894,0.49885813957138525,0.9025144082503603,0.8827921253178409,
 0.7743068297713902,0.45695901601784883,0.09438322758458084,
 0.9898148953807911,0.7263776252278167,0.5377496138376275,0.4302173816825309,
 0.9158055427206333,0.5684127519481691,0.6915647013004798,0.2639577189441882,
 0.45488666071686334,0.997168511528291,0.5391057489224655,0.4136211675791207,
 0.3747328924159703,0.5055816887073512,0.6927868684554839,0.3461594893676523,
 0.5148101944164737,0.9225001947774423,0.479738551869178,0.7231786063448853,
 0.9414419700533191,0.5060842069883864,0.23245077644864032,0.6943176376108945,
 0.13643893598853496,0.2337552035487429,0.559338199892708,0.05976064852522889,
 0.17020057352548534,0.6025811367263121,0.0013946948181939867,
 0.9940975253117782,0.5430925139312616,0.4613881289917693,0.8184509463946988,
 0.7194917937495322,0.7489522379218337,0.1855773016034915,0.38227523231017024,
 0.6060678366925234,0.4699483228889155,0.10889224800595365,0.6635604647284427,
 0.6473179871186656,0.7508877623797493,0.38965663609083445,0.9558119980165196,
 0.9970738244807408,0.457288069490302,0.5962371431360907,0.29840488609945937,
 0.013192327673976156,0.3178649323133358,0.5625365345157736,
 0.8134652045416862,0.6827150430231606,0.19207292031239687,0.9723951607982309,
 0.0738494727298693,0.6794781301120102,0.7376476805315888,0.25533462096302606,
 0.9977895726766679,0.8401750830945368,0.7391095072298656,0.6302702324304866,
 0.10713783566686164,0.4762502657310167,0.36811943572151196,
 0.37716184145863707,0.8615581798860259,0.1836746212054038,
 0.32475579591515824,0.8531895324029923,0.8235741062346537,0.7785581604237329,
 0.15565101262918335,0.7833342020213666,0.9224027733997965,
 0.42042790565168153,0.24886086073368718,0.10863054470316802,
 0.7453426792036518,0.002810681647008706,0.18132312540284867,
 0.5751271618814906,0.6931412195244594,0.058513816037992994,
 0.47075934878973347,0.6625274669994847,0.005945860852032303,
 0.41534102377271576,0.5702105391819676,0.5310156701626216,
 0.05344675380642083,0.35439344225637137,0.08794508324647321,
 0.9875623097839751,0.7260452335458821,0.21177007367556522,0.6689636226580432,
 0.7804589075318087,0.3005797471407252,0.1620103618468931,0.2956939886149358,
 0.24413980735422747,0.9844246747937523,0.8834591019347149,0.8589575787804972,
 0.6768112260585535,0.2467525817339823,0.43911920018657946,
 0.13692544863925638,0.6823182403382548,0.4043076745563824,0.5707236899892131,
 0.7459283004831336,0.03810240362810602,0.817202804232081,0.9224637165626256,
 0.9472390895859646,0.8797886878262683,0.24563574408752997,
 0.15712526905495205,0.8871530154057764,0.2213125655164806,0.9324339827151376,
 0.2902629754754952,0.3623353029445847,0.3310231829985588,0.7226647732268444,
 0.7379625016903905,0.8819644393936944,0.5445930409099,0.8374804707320144,
 0.44460289219072946,0.31020844623111554,0.9838580785593296,
 0.9049971435621638,0.3679353285984417,0.28055057166077146,
 0.47882535576549223,0.8347457063387467,0.21151769416214738,
 0.19092773989187273,0.6178374713229198,0.5474477577834663,
 0.005865223171855005,0.7396103448241117,0.4129987348791887,
 0.006266511288578913,0.34470698984261083,0.6643262316968465,
 0.13986866498814254,0.04118832893977398,0.5420829304451482,
 0.7840641913765141,0.7405129972572866,0.6319629171821435,0.17434126684896079,
 0.8975148080658376,0.8585982548529667,0.8537728895656042,0.3310696079878219,
 0.118162017081479,0.47983882420334634,0.8829697040166401,0.8346760024980072,
 0.9760264616584701,0.3779444683975425,0.692850331063844,0.2517769660083753,
 0.872324311720782,0.5706832274236263,0.1538444948046801,0.4628554035999075,
 0.669091241221347,0.6923506732991088,0.4754887818175948,0.9200572165185665,
 0.37349938359439916,0.8904781407656941,0.3222962426570499,
 0.004717658899872412,0.9361004778587121,0.06752129931418993,
 0.5895168415470571,0.7020716740211705,0.021353005228943567,
 0.10358763054125997,0.5740173870015388,0.8887354903930607,
 0.46773293414948947,0.37040067889088735,0.8079170461681082,
 0.6589990957059426,0.646498146222299,0.9348866823697736,0.7903321839733706,
 0.7728566730298746,0.18455854747868938,0.16528944744467622,
 0.10806485137573485,0.0633960418250572,0.05785917861916512,
 0.3513540069994142,0.4949258358168128,0.7726954322243811,0.5387231641597476,
 0.11103362585393117,0.5440031795163316,0.03910486930925605,0.752486564575682,
 0.09387926609804098,0.3614671147586891,0.5591276332516749,0.5888174648179747,
 0.7312987959700896,0.8371313297983853,0.2722912676198802,0.6619737614471763,
 0.8597147436650916,0.6581238679373284,0.7992760915231231,0.4181916028320012,
 0.2309111887945423,0.4338790547455774,0.9287916945978774,0.36847221001958574,
 0.9618591212715826,0.9423434984022744,0.44404569161819474,0.2864150202857051,
 0.71093788096453,0.9810833466876933,0.8116737195287071,0.24877744136157032,
 0.31066091840899934,0.346994420270348,0.31453222812337467,
 0.15259258852954172,0.697635330867385,0.15159813999623128,0.2713713064872729,
 0.16058511494990602,0.09643912236284768,0.3899003125936602,
 0.5023694911077559,0.7283623128647074,0.531495005756376,0.6601272541848942,
 0.7503686058938591,0.6943210012595944,0.9550325759846039,0.38384835237635584,
 0.6604883757074976,0.7804902187501985,0.258800315951474,0.3460358592668046,
 0.7348929549428611,0.383155471833477,0.35757492564277227,0.5374814532928527,
 0.8403729007904744,0.0916915049618835,0.6984271154345052,0.610393306018169,
 0.6079674971484793,0.8641118438689432,0.10447811546975916,
 0.22900158248783253,0.025223926594019508,0.5511386249285186,
 0.7547132398069838,0.6109069556353742,0.05679886495772146,0.5203519839713621,
 0.36069485606928586,0.6805160534206497,0.8677503616340311,0.930110946731537,
 0.3021686513326489,0.632347579740162,0.5293840678555188,0.7739869206021711,
 0.933291180709729,0.34346863879486134,0.7299838310963233,0.7984585768681295,
 0.607242190239456,0.2935854803406225,0.30043487407330693,0.6473374093203841,
 0.34556316065098314,0.07707118143732139,0.8423250961928003,
 0.40994201106116246,0.4233021435293616,0.3185095929144033,
 0.18449167751236317,0.2038337026626721,0.10866507060885888,
 0.8317914123376474,0.377325179617631,0.9042384150300035,0.385095694174452,
 0.9005093520267929,0.9256931067681071,0.5309703654203523,0.8799752748422982,
 0.2680471561814609,0.7843020556778977,0.7117849359104482,0.09199369040429461,
 0.409798214020614,0.7563930672360144,0.8382947928482163,0.2165441189651105,
 0.7420700922293743,0.7968437548453765,0.3952778380060593,0.11306999469032808,
 0.24388961377666496,0.2130679251734745,0.7281498310844612,0.6694154290757699,
 0.8365661676036463,0.012635630453826319,0.16980455359463287,
 0.8352601943610735,0.3029300898180807,0.1352144290490388,0.4786796959147567,
 0.3631960444773449,0.44133732425702366,0.016441549425072144,
 0.08903889386771358,0.701856445434417,0.6347827927193156,0.6704271404588557,
 0.515438070031596,0.38450992612170853,0.3071103303217162,0.41304113955487864,
 0.20204187175434019,0.005531227951131035,0.4076336564123205,
 0.5168801811684396,0.3138687427462592,0.5060478443496281,0.4865866271569248,
 0.721639343024885,0.9530369453973954,0.9338467987843743,0.12685384437204417,
 0.5628127011348775,0.6867944361454587,0.3581450234088036,0.5730907248723488,
 0.1609488039660424,0.9393237213131259,0.7934109781368762,0.1023010093886948,
 0.4837670879260847,0.7931194112534151,0.898655320817454,0.8490909697723041,
 0.2329466195230595,0.24253198268601606,0.41305938444141865,
 0.6944693002279807,0.40171902584623387,0.6523007341123419,0.6357635061915355,
 0.9901950738349206,0.955840216433643,0.8618909612662662,0.6430313777521163,
 0.4108957428508506,0.971902243936664,0.3204577452052657,0.5173929837046817,
 0.4569948638971246,0.47637994551974194,0.0896934130405016,0.6906527744508107,
 0.17028947456024213,0.5366693288244692,0.6982250544952313,
 0.022505354762297758,0.0895762833813959,0.5235514844115762,
 0.1753158724762267,0.25516522531636987,0.23416073287144246,
 0.0024563581138551394,0.18991825240637994,0.078152091607543,
 0.649611887438016,0.0928544803876723,0.2543842595938719,0.6625770968706874,
 0.5383006870549859,0.6067257223337725,0.7462368258304681,0.2571842094862056,
 0.465089698832336,0.9615627902303119,0.04231838785284392,0.7728690087220853,
 0.8640342233769918,0.38657095856898294,0.16181358668570855,
 0.26493218887837133,0.9104595525875918,0.6156233445001676,0.9912734759369748,
 0.22638221819847038,0.5374722977215418,0.8518901412810481,
 0.16034581084158162,0.22425602889025364,0.5975602545207547,
 0.1819521270088471,0.7818584075955813,0.9700744664824464,0.7683103849878933,
 0.4691024847139118,0.32171891444509826,0.15207278529348067,
 0.8691310571927016,0.6678396367175636,0.24801227809819149,0.2705210944322819,
 0.2707679614710725,0.5560746739711642,0.08698908638623581,0.3140677320903822,
 0.6789927568548886,0.6381948867817032,0.08114586333849216,
 0.15190107602820768,0.49562227220049915,0.16107959973383967,
 0.39013114052527254,0.41042705491185716,0.46314386109691685,
 0.17759692356734735,0.6281226520764022,0.3504889285733457,0.2942957394599208,
 0.9060200355358258,0.8279696680148174,0.9254637927086679,0.4689497456157865,
 0.6876644291195624,0.8491255152505477,0.8294515148084521,0.9792422781630606,
 0.23458319577908082,0.8720430328331253,0.20985466475805015,
 0.7153059243139931,0.4362073194830993,0.7614058953066468,0.3500236993109387,
 0.11814153540461414,0.8501743689065098,0.94152604427506,0.551214543695221,
 0.7755437806547147,0.3811442148391557,0.0980498002886695,0.8297517143115264,
 0.07269816426788034,0.745759099995441,0.9142027671053351,0.9020077868419842,
 0.521484777032327,0.2255195339217262,0.6050756500285072,0.18040147419269126,
 0.08029693069077615,0.9248685618009969,0.9966588250275261,0.5558665601303926,
 0.8882710270002209,0.42433028237744663,0.7240892805142776,
 0.34708704367678234,0.9705368592559931,0.8671900105539384,
 0.27362192138700026,0.35861924431509484,0.9925240266371548,
 0.26614869505344196,0.8476936059658186,0.8879926694442842,0.7642450236253377,
 0.4557762846244602,0.18128450003105612,0.044895257141624456,
 0.7678840897075474,0.5905542252829112,0.7171073401848415,0.6446684942610673,
 0.027892851028162746,0.4224507401064279,0.24474962759835384,
 0.08189153268595739,0.903063258510145,1.3365751710237017e-4,
 0.30591713176427215,0.6361519564071031,0.9963848774168986,0.6451172364346665,
 0.4075737966177254,0.11258876285587394,0.13999491699609423,
 0.9723426398236883,0.9035444659873709,0.3265896917548119,0.2573998156188466,
 0.780281556887945,0.07807162566773318,0.8064309988695155,0.14793943264411646,
 0.4719348301150399,0.6451064929997432,0.7132083032450522,0.20090638350909718,
 0.639209051789251,0.2105791273631059,0.8430522945694476,0.5095515874370398,
 0.036507977820419146,0.7897455774390043,0.9891348176571343,
 0.14498246030809958,0.6010812583227192,0.28965896814276837,0.508479159538046,
 0.7394748169484007,0.1052244491830654,0.4031926450788188,0.1794413705219332,
 0.5883768905581223,0.4659065561366005,0.9375423291561807,0.03240972708203038,
 0.6946987910041985,0.21080677992711283,0.759933626233307,0.3295439358803076,
 0.5056841278362073,0.9534620691505906,0.7029988819015547,0.874201882572375,
 0.8832976620112903,0.47587639404657944,0.5589856342246872,0.324584422845015,
 0.8328193141519349,0.38144417586563784,0.569296930597828,0.092743513560413,
 0.6345881997722662,0.21776733746723043,0.42199789599813176,
 0.16085399544023737,0.44361485119631916,0.7803412000751397,
 0.6177560243982811,0.37491439984636143,0.4181541034500331,0.8125767020617634,
 0.8042933509621661,0.007333164917992074,0.031813050558515354,
 0.23107011089295626,0.5103228104993652,0.8886343860047395,0.559334478524349,
 0.7972209797777572,0.21589231247043705,0.023969251814961234,
 0.6665299153331539,0.8529242591020593,0.016502417408116976,
 0.29576402852943695,0.5228932610724669,0.34425645812319006,
 0.9102197716699969,0.020055921420554124,0.8539227753039549,
 0.7737938257767628,0.030689669870704983,0.8024911133140523,
 0.9250239905028944,0.21061276103955728,0.5191951171316294,
 0.29456835806118364,0.06866894990262695,0.8892553784086485,
 0.34319351996683567,0.44041140702559967,0.2983063307897591,
 0.3776771943155117,0.29504667479345803,0.669871937159999,0.960576963654708,
 0.6976803318168159,0.12759628780389765,0.09195889372271071,
 0.07396234196137097,0.8698416747469767,0.20478584463877625,
 0.23449913245163345,0.8693361687412104,0.08025676455533504,
 0.18427129961157662,0.8237713731080906,0.8667856894147075,0.3521168862539109,
 0.35040170363233314,0.9461222785627829,0.41727757116305864,
 0.9024238782308776,0.49953479854938676,0.5691398039784686,
 0.20577905752380454,0.14997782789995984,0.5775557021993467,
 0.7233316059623479,0.7915051787189642,0.015444243861455154,
 0.3603256291506175,0.2233774505743633,0.6816274336225817,0.42358788358019483,
 0.1369167753356213,0.1085457308733434,0.1494698582405869,0.36190734029332317,
 0.9256398123344949,0.555618901055362,0.8252610531986054,0.6151494034877559,
 0.47143892482124505,0.18070321174681425,0.3575361760579261,
 0.4481625741562745,0.9655373057993315,0.39172527293815673,0.9256929815259556,
 0.8147203734691473,0.6266967849728219,0.34265518656207905,0.9098412488740375,
 0.2952603702679545,0.38339288933499593,0.9365530173166988,0.9553648177475554,
 0.621673044598121,0.6240388421829215,0.5214874454278436,0.5096407902443595,
 0.36659208478625993,0.7132472711658147,4.011598420077256e-4,
 0.7726185608636472,0.7146036062061207,0.8572765508951583,0.1376713588611549,
 0.7358212914448192,0.23262507051685344,0.025200935558222448,
 0.3764921642188619,0.7332182760014249,0.11893776026297109,0.3965180104422279,
 0.8901294731748464,0.8775536436051938,0.8181190572916788,0.38299034771864326,
 0.26528095291480014,0.6579586874216444,0.028834715550930534,
 0.8740671663975798,0.06211076041267316,0.3700163554488226,0.6225394458203379,
 0.09817344604457923,0.26715982542285277,0.47964401391515943,
 0.34337508808428874,0.22925860707253054,0.45121143143113607,
 0.6453705341707007,0.770735090199741,0.9695338496735137,0.9233522005584345,
 0.5793052456262262,0.017363444845675646,0.7263289117402787,
 0.1464344877218945,0.4917877769728989,0.6140967850381559,0.9413441473316156,
 0.10077729747073438,0.2222489177120478,0.21670117170142622,
 0.4191156253445484,0.3298519995410949,0.42973219969642673,0.8449175312305911,
 0.5336896750191011,0.42287586495156204,0.6225208323261345,
 0.046470023514876835,0.6264310804701302,0.13072900167337642,
 0.4425478714620329,0.7881716655784979,0.3542481568302357,0.4296659100695489,
 0.5390058850778321,0.0749767800897767,0.6155516449147104,0.956374270599283,
 0.4605862675245168,0.4366315139602681,0.7561725257476073,0.18335328798880413,
 0.6219974451783,0.6845043337892005,0.6872155618346083,0.9991412458390893,
 0.42948076794745726,0.29178448293669224,0.5977855176538927,
 0.46905950488735093,0.30362089457224417,0.4779979308598027,
 0.9664222731405546,0.3591679142573856,0.9748562053059391,0.6499051207263317,
 0.2699728147243726,0.6609753185446047,0.9574412391824936,0.21713633633243867,
 0.5054079974145262,0.2671647728606411,0.7346119675003044,0.261140111162545,
 0.6926272579073132,0.12212141490643669,0.7647122305872414,0.9458370267196335,
 0.1410877291799626,0.3609034523684801,0.9256146293661305,0.6005499992375194,
 0.5144326848205635,0.42238198090400625,0.3838385716640349,0.7667434457773905,
 0.4941060186319951,0.470421440403362,0.5043384091141582,0.025016933686819787,
 0.7258956016051488,0.8496190423903388,0.8631804852399215,0.9694858823143312,
 0.7495893864863327,0.46290473971447865,0.4517829562744121,0.8585238664876126,
 0.16413760336144623,0.608360169855035,0.1832937603957242,0.2458532078496255,
 0.6968897335707969,0.7866933692008437,0.3433940549895992,0.7419469058477453,
 0.9921864746058944,0.5099960882785707,0.5211959104622593,0.03157257230960908,
 0.921007724267002,0.015522255165735799,0.5409930097128147,0.1481336938231842,
 0.779909060211378,0.6407031402781992,0.37051980419966657,0.49559888993248424,
 0.4180707855047223,0.6488962042529923,0.4464141520704169,0.8891044290059125,
 0.9114937879953381,0.8418200950435589,0.5766994852852586,0.4559816001305844,
 0.3837077816733313,0.14499258342472798,0.44492967664512384,
 0.21518683466519528,0.3561145489037154,0.5585206512764459,0.3174730553165961,
 0.2016787348170127,0.3736247494236079,0.9855607706368502,0.9727763004619664,
 0.0030457925857894086,0.6726675788858535,0.5815235674955883,
 0.02495524232714952,0.2744439363961825,0.35329927017467033,
 0.5691401454609237,0.2671189568220753,0.6474372401146468,0.49588728155144823,
 0.7052218590881305,0.5251553064485416,0.5242389712148281,0.572775829793887,
 0.8340751634207892,0.3488293098747045,0.8389923694094996,0.8923830495835936,
 0.6063359449144771,0.9060693855579913,0.8333610097977352,0.5198090740527258,
 0.035403470022802086,0.32524362766770576,0.2585224278181426,
 0.7255150387781691,0.20874512944348844,0.9132989371946145,0.1305743530855218,
 0.3217841053562388,0.9503124103821932,0.49012357042673105,0.8512346343923639,
 0.2906641170386146,0.7087819139525477,0.44247492968451596,0.7775964164305684,
 0.7027879031193486,0.7231289524587697,0.7791411862495247,0.7791867712822091,
 0.5898726167440258,0.3677876163591194,0.961530920479101,0.410676022361383,
 0.806848956112904,0.9518398870547012,0.7190329567292378,0.696424186154591,
 0.3833756940643336,0.08320450469138041,0.34766333735568,0.31582934197751644,
 0.8432940870454668,0.30340601879562035,0.6122110928078446,0.6287841250665072,
 0.33616045775519376,0.7674189580419173,0.1157895479590596,0.813548681978876,
 0.014724770021163791,0.2707924661285892,0.9394974671664803,
 0.10025941967562557,0.3049062140734027,0.9851177724995988,
 0.15561044135442903,0.4775958302477554,0.057077626674478354,
 0.435988204153115,0.5461686953100312,0.6412188524047023,0.66947371747034,
 0.33584189228150385,0.5325759411553254,0.40292150442676433,
 0.8776266161218476,0.4561720872312037,0.22725292994865254,0.934712702364783,
 0.568141905248448,0.6490956512890325,0.44022399469065443,0.34748695772951965,
 0.478684778924934,0.36175215682628736,0.9832938551705026,0.111895166825151,
 0.31775974366177073,0.6423918550647674,0.8397345167566557,0.7206605566800792,
 0.05383801227765994,0.6108913520117918,0.3556819141041727,
 0.059969943425381755,0.8139151961315029,0.2875842340233714,
 0.5421066907488533,0.536282882641326,0.26475617761541637,0.3331203796788105,
 0.8792869564293868,0.21744128781623284,0.8872519640997824,0.1976570946956795,
 0.9824054629434813,0.46661435787517147,0.26021148628112956,
 0.4224867449259705,0.9366881515564199,0.5934226916213058,0.16923781543646388,
 0.5420244844118216,0.8965738669511196,0.8133846357670603,0.7254483298101511,
 0.5156400174691601,0.44205772680810373,0.8391174003408177,0.6707970907665717,
 0.5223484710018522,0.4959693651343837,0.892554198240488,0.03749297556354503,
 0.11833294028481167,0.8625227704228564,0.32071290335377733,
 0.7425779657789022,0.6054512762133142,0.4079788051014157,0.6977859103737266,
 0.789661878674279,0.8614812142309595,0.2995563099732026,0.16377091542795796,
 0.6013524164952913,0.7148764853088073,0.7001581568488449,0.08952159273652793,
 0.5634446938472909,0.2667812218462139,0.31042822181133123,0.9290972379348443,
 0.5170078636027969,0.4310231304209786,0.0991765553699624,0.3708942604879567,
 0.6255057017026358,0.15763010357932572,0.17071074605550873,
 0.9582863976082344,0.12304534872585005,0.7604467810173546,
 0.15318995983705375,0.9999486812579851,0.027511339297789927,
 0.5405094986488101,0.682267639769237,0.3917671996248133,0.7203090805368514,
 0.4623457127972187,0.47599351017190317,0.8696078758308072,0.9861835749746158,
 0.19088702036410188,0.6706675957167771,0.014560077520229031,
 0.16366904897350354,0.18767023031967,0.20048174223099702,0.4393859791541206,
 0.6976415357084627,0.01939202943659646,0.2439170888907648,0.9019220789114955,
 0.8117182622177985,0.44542511551698194,0.969158273537496,0.5988933204996139,
 0.4967842410058261,0.6253130561133218,0.5130757536338854,0.7413273442247701,
 0.08952213422755662,0.5417158887642578,0.880972255536444,0.8969015948830873,
 0.23279163162853955,0.01838749569161857,0.573084157537717,0.5066682999844095,
 0.1091319234107937,0.5674504723532077,0.8793792258681526,0.4166335225045278,
 0.7190514687040452,0.007052860702071229,0.462515993839256,0.6049424964342018,
 0.653515754069541,0.7451055993436161,0.20994392900143,0.7584621652266159,
 0.2771498629963084,0.5290437441022727,0.3888784420214165,0.4261177769753861,
 0.32268138737548313,0.06846180904046661,0.9956032488354993,
 0.16840764332020342,0.6344099376686841,0.6933797282410225,0.5088536388310065,
 0.4210150629925099,0.06873339399781175,0.6426451741286763,
 0.20707037655941773,0.16884335593781885,0.6364809328569687,
 0.7288232943387297,0.9644365762931708,0.9164671758204747,0.4792690396436099,
 0.6945741552825733,0.3293871205867933,0.39529570375054923,
 0.02819423566336643,0.18543200700737084,0.7386622762848488,
 0.6680583600152066,0.05273095991144283,0.16927225939105106,
 0.5496608309306018,0.05963196348089861,0.38597349716251506,
 0.8657032256411559,0.43392446302355636,0.6929635427082724,
 0.45678489668127276,0.8960050185456443,0.6809720421843959,0.4756210529882431,
 0.4397052221319271,0.3525851353933439,0.7810951382766919,0.3558509640938573,
 0.865672867474242,0.16373949636262009,0.6151219292464842,0.4050710956997924,
 0.5623834616684877,0.10571340133525886,0.512332119992189,0.8964264768453063,
 0.5288284978379707,0.8054853776162719,0.11458921433782898,0.4769388758720494,
 0.5500817026537598,0.9923824882884533,0.585921871882534,0.0810778238385299,
 0.614243433299981,0.9395942683939106,0.23177341079903446,0.08644891121670495,
 0.26648688089322126,0.09701206513125982,0.1644864862909322,
 0.8968731929805008,0.6384986604213903,0.8255748111454089,0.6972823260607087,
 0.6177037350603521,0.2335650231602181,0.8785500398939255,0.24690520396023719,
 0.4851566742069142,0.28416226860474436,0.2715491598868657,0.9432572139472214,
 0.742068279985028,0.05447178238125694,0.2084007988654979,0.08022145955061799,
 0.07505807713591905,0.9482551191913775,0.13183605597410564,
 0.1458867909572351,0.4192468889105935,0.43202148398552775,0.6664594856481302,
 0.8283203302522343,0.6355489126245507,0.11151860570503924,0.4849253122125087,
 0.0695933341215641,0.12230218724556674,0.6631068473054291,0.6518650942403611,
 0.5018601197733039,0.06861365098619421,0.8075815985449989,0.8295398676643013,
 0.8782600705147179,0.007050073484054842,0.43834379121301204,
 0.06687255036694384,0.9479500691650107,0.17940347472908413,
 0.5972562473799004,0.1420457300559652,0.4488565938465312,0.9014987778392177,
 0.01523897306615396,0.6238458336567257,0.20311688938599126,
 0.30485532781939073,0.9574847758502072,0.16623728081624434,
 0.4600069733677661,0.5209476436591138,0.05002641598571633,0.5961938481378616,
 0.11548769275093629,0.16084581510148677,0.5540108979753776,
 0.5448303911704256,0.636060495029925,0.8814143761907152,0.48587720806928536,
 0.34906163517861477,0.08705353964855933,0.9253931189294928,
 0.8584869802983902,0.7861388945643644,0.8468576785415642,0.9981375558877025,
 0.13493823722154918,0.41549081375796615,0.9407151258947937,
 0.3772279236806334,0.5961524760817994,0.2713893731273048,
 0.005093619570973429,0.8107683396385206,0.8544435117450415,
 0.1822473162131446,0.3528573985189797,0.13645839218802958,0.2220429041403561,
 0.8987644369135301,0.15355692690137024,0.6797910735631412,0.6842670152964293,
 0.6854533754853307,0.18171458388268125,0.2268913546275042,0.8037432698076227,
 0.5742869797442839,0.7266395203465199,0.2471140219085357,
 0.035484848247059775,0.7041010917758818,0.03278465252615026,
 0.9228503338102276,0.47346483106951387,0.31567664391617334,
 0.10037085673052804,0.2293280076526657,0.8080526263712993,0.4189623898817788,
 0.4894771616627729,0.9167181520305638,0.16423794511377765,0.7276607529939768,
 0.5086623143882101,0.5652805954515467,0.06276083086941975,
 0.10785054627525215,0.3015289226672716,0.4507542452487915,
 0.45808979433473795,0.388570890187071,0.45730609388970045,0.8059111430765608,
 0.5714380549874338,0.85653575005719,0.03117091275098094,0.3210121309674896,
 0.05055504100149688,0.548001595576936,0.5986799435867142,0.16877051013623512,
 0.5435635215319631,0.7118913547970946,0.6865142325796328,0.301211660729392,
 0.8738038041618603,0.007440616602809147,0.4498991736455683,
 0.7179344370186499,0.022477434405688168,0.16009429966534405,
 0.7663436018225828,0.004067615065181052,0.8398161735805318,
 0.3678657855207722,0.7874984846154085,0.6277539474416786,0.14591661617805474,
 0.32078019244439426,0.8454471518155758,0.16731587391623937,
 0.8416968909311081,0.6918120230037663,0.3304887880493669,0.14727913634280576,
 0.031746176093601086,0.421406932225603,0.7961730392139366,
 0.46810589065251806,0.9872463176715295,0.8806962676268141,0.6165838621388818,
 0.559220456572123,0.5296168418150003,0.19251688845130333,0.8406325173910303,
 0.09455880143249473,0.2612205144794417,0.04657319795493298,
 0.13229760783237454,0.6059975643705221,0.589068896983904,0.45632248566450584,
 0.2512724871936617,0.8829346158965127,0.8987392320773243,0.08911168604763253,
 0.02859348171308962,0.09670856464898125,0.8777380458938153,
 0.44264180911470097,0.046074923274603385,0.5248457257013602,
 0.6942144644636026,0.17358204136517963,0.25932816969035033,
 0.6368656153072698,0.051003032862859565,0.17752943496182005,
 0.6830972057747762,0.9761389114441773,0.031907731245296445,
 0.4108537198048565,0.5595411793323144,0.5641657550127683,
 0.023170015317893422,0.13355756024081955,0.3526283585767451,
 0.1574635895934099,0.26489602539734325,0.9637180631058909,
 0.20603485362871166,0.7926709378274458,0.5390624207262017,0.3727576930877079,
 0.8846007778121423,0.012190336665808044,0.7932786986148032,
 0.10733216508755172,0.9669776347742619,0.280366948819312,0.09408229192089324,
 0.7037129771082962,0.872547908969959,0.4424497432699863,0.4752042092814651,
 0.7198257722936439,0.6684811942066812,0.16790687562819417,0.8995335116932998,
 0.8513484720081513,0.24331099506146014,0.7294350006303669,
 0.15439196874848804,0.8866547019783453,0.18149329226803979,0.48078969195584,
 0.1656234477239915,0.7339210942907028,0.70218267130951,0.7988824447023729,
 0.19948065532509185,0.2957323155069015,0.8893518036721801,
 0.42710025670298113,0.6355627615460416,0.32022700057227727,
 0.4206962662062971,0.3506785831508843,0.2915253251147236,0.11904388650555942,
 0.042135504249961264,0.9462077321760479,0.6074395948200167,
 0.9984505941911903,0.5884815944042016,0.34545626727123424,
 0.25683825253175907,0.6464937164091975,0.04065600748388687,0.48974363418433,
 0.058966620666204284,0.48851490487377347,0.6655287057585909,
 0.22862820740780565,0.2630053088988753,0.19640644758858317,
 0.10167222925260122,0.9150633163771473,0.3097818354777684,
 0.10722086903247408,0.18838037738598512,0.605440028936829,0.7009414745253942,
 0.9806198159094709,0.024008798307830426,0.8231171054255482,
 0.5176636007069093,0.9480804545432829,0.3687563406855109,0.11216453246072022,
 0.7487057493973729,0.4358504396421371,0.05153546879865589,0.5283471327131135,
 0.8469718570057237,0.7227610943461477,0.34256040449477665,0.1558864762102654,
 0.06103358786396107,0.4492567267619251,0.7031036960429565,0.6900882332615611,
 0.48116758715373664,0.3538190797010098,0.5036961942439876,0.2750514150440986,
 0.07753011686231748,0.17654177434074025,0.061968275646248405,
 0.03509738928187056,0.6034552109618417,0.07943135535883816,
 0.35292384916444697,0.963458644502294,0.8887419172930777,0.1084089172853252,
 0.27789532631549607,0.6951570522693011,0.06058611581277307,
 0.7757876619127458,0.7684280528597931,0.6197625719545177,0.5644720012863824,
 0.6787660524011484,0.7479227047816487,0.11738834912030316,0.6492702508988629,
 0.949306760714286,0.5090524675120937,0.9945593090550153,0.807490854992449,
 0.006510982080825678,0.21764902384724294,0.8973462297882995,
 0.9297458890019086,0.12375451415480054,0.5214608944478362,0.7869920296278465,
 0.5757848495365184,0.497427580730616,0.3095105641419188,0.08252640828535185,
 0.7513981365717692,0.46929940904108214,0.4109851224338682,0.6088143617906036,
 0.9833696093334955,0.9511106424202823,0.8459464862058003,0.4904990272844725,
 0.2404678705326575,0.47338972186662576,0.9021666528732614,0.5776615069111612,
 0.3604632773654062,0.20387874625041236,0.3133272695870112,
 0.25410755050586653,0.5089451523251242,0.14981201191226734,
 0.6159445677827907,0.04531669593077736,0.8286311654542601,0.7235372529955207,
 0.0715551183150488,0.1052649123982935,0.0937157433998842,0.41427027043087794,
 0.36744551248939483,0.6647654855590285,0.43241153901997453,
 0.3804368497261531,0.45629773679430485,0.048145912554633075,
 0.17449240959919243,0.1877353938377333,0.01967993871716822,
 0.8714904432378364,0.7617812459918927,0.44951136297077376,
 0.22927450835824137,0.9050064055635231,0.23990924928911883,
 0.06633389707305115,0.1805268028375413,0.7718519129227592,0.2873625454945722,
 0.2467659791227813,0.8326332244399762,0.025738840555720954,
 0.16943293983242003,0.9791735305295509,0.5926944919398495,
 0.24485019406712083,0.8771057378523613,0.3677922519067174,0.5425875753597666,
 0.3029624724342772,0.3961253283837006,0.6876301918923333,0.34295142894725217,
 0.03921165769237156,0.9884308047452873,0.9814701239696062,0.9574685677557457,
 0.46103549517582787,0.3550648723544714,0.43249212920518576,0.63450638082487,
 0.7099202895330373,0.5052927335811134,0.5476105522346999,0.07957613456004153,
 0.17055959565488543,0.8774501880734682,0.22451462281782142,
 0.5746532392000094,0.5942684131681218,0.6437915561628208,0.7702629238793454,
 0.5466398379618612,0.9321949591895152,0.7815589620347824,0.2278635262753952,
 0.3189475404134765,0.9048399897848165,0.9579936350973045,0.454840540196511,
 0.6037688678390369,0.7432121500413371,0.3035822837758516,0.6295843031009749,
 0.8016943115729455,0.5517477340651089,0.7132190100076666,0.1945947344753326,
 0.9026545490499418,0.45359081586390537,0.25768317772570315,0.978866246158842,
 0.2774957913899846,0.10065723676939009,0.5184670300262924,0.689030681883556,
 0.304665039097891,0.9092151251287248,0.6907991189152897,0.8709651234719272,
 0.1044321542776252,0.9650705029078349,0.630894213897616,0.043047359364923765,
 0.45028499931703947,0.39068639964435325,0.5784449348450739,
 0.9485667779873714,0.3243434260185627,0.1047655157409011,0.8717580153912317,
 0.5346099574393232,0.054882789550327216,0.5502039216053138,
 0.6408988650534738,0.21414328582791187,0.38518606579899406,
 0.5183888683004233,0.8603806374026769,0.8348646945851602,0.12070380253154522,
 0.36717836601769827,0.6862020106853246,0.20813938592172276,
 0.26944368020225207,0.9432544064641227,0.7152544644821153,0.8520863671568109,
 0.8445213560842952,0.5103618620024966,0.34363145954724905,0.9323004406927276,
 0.34339871058229576,0.39448467440667767,0.05472150076269,0.3410170687042453,
 0.6278420704299559,0.4791029999284092,0.295495681973437,0.48682407995510446,
 0.5326017669288013,0.5934945046045814,2.3310715843061214e-4,
 0.8543241866450475,0.2988280353998608,0.2781288397226638,0.7005653272089807,
 0.540465707928042,0.6987161674346005,0.9097126876286182,0.15423630601396865,
 0.26165526883004464,0.6173894626676296,0.25687987652755573,
 0.9045719713406565,0.0341501463591416,0.6416472485396199,0.6165820913679294,
 0.17623591952049567,0.06587760230975614,0.20401484036938822,
 0.4206468089206574,0.6955151728590044,0.8832151036525654,0.06481764105460874,
 0.20059762490057087,0.5567547129964701,0.09253661999890217,
 0.49336969629438787,0.6946084791706105,0.7681129227055269,0.8556064504707442,
 0.6665905753559412,0.32649485366585196,0.7184512553604057,0.6864891196553794,
 0.5215086835087397,0.1974755280612862,0.2336407697245958,0.4979015051130127,
 0.08023839316184589,0.025902158497029704,0.3159534926422596,
 0.48635336101640014,0.3155723295822569,0.7448635102865895,0.7704475346429371,
 0.24792116280049803,0.1510655257681921,0.8088407434276955,
 0.15182029680407716,0.3334816394150495,0.9817398939458079,
 0.22302325362530173,0.33009680305589384,0.9747179696146813,
 0.6324287738951152,0.515719094293987,0.6396666605222555,0.11217286867603471,
 0.6376816103135622,0.8618827628655804,0.8911897431168704,0.8678474808562665,
 0.9489178169524514,0.898237223062521,0.9643529549983099,0.4058385970559132,
 0.10108643935576267,0.5210221509141751,0.30928246159580863,
 0.06713208749052424,0.6535269839669218,0.2601658407424864,
 0.08629684984338326,0.5512000632171861,0.08708424936416381,0.914226155603037,
 0.12483344314214462,0.6303514998810127,0.3257384394784696,0.3901294432882477,
 0.07103358969516194,0.6944572910326393,0.2613684517109556,0.7842669638427633,
 0.6616712298589275,0.19296165892196415,0.023598577649318386,
 0.562698586867169,0.13741529237965755,0.36605877378153007,
 0.48751417456874657,0.9156479576719109,0.6962590651430025,0.5732451079596942,
 0.7528643280420402,0.033526761509770964,0.0909234385254225,
 0.5178405289818753,0.630316384087454,0.851140963708719,0.8543386521196541,
 0.8488610396142736,0.8491111491051468,0.2541940199015785,0.7318376925283174,
 0.48973906211255724,0.7655717525338311,0.15211222136555713,
 0.9447217840249365,0.8940435417700716,0.2325558329972266,0.7997183058431384,
 0.27807301679454355,0.879374500557477,0.0013743476414332934,
 0.3827297563960561,0.2790649974823963,0.3483420147659553,0.4412756021252725,
 0.587567535911071,0.24027732163940585,0.5224498042632514,0.6017064211923099,
 0.5650781747002963,0.22523298123465807,0.09743747198993902,
 0.2830916961723705,0.49766174578406286,0.17827755858733807,
 0.9539993553928239,0.5421095789636265,0.43182305843530877,0.7964056402802333,
 0.8476611883765116,0.7775654919110083,0.610172193720633,0.5974086561028722,
 0.5395279909637867,0.7691346656411261,0.5597126344566687,0.03714056436841051,
 0.6018291929923076,0.24560506266323812,0.7802974873411425,0.8144992518924643,
 0.14259730368160206,0.6345935484685529,0.015757808732314138,
 0.187711989930061,0.42905292758313873,0.25136460225914903,0.8648161135550229,
 0.04426937632558481,0.6540029780639814,0.7248699856404546,
 0.10759563395244143,0.6172826065149601,0.7566593478730856,0.3727221026358769,
 0.012675874896798556,0.30306777849834,0.8462636707028889,0.1851913900903197,
 0.6710894574681178,0.3410948875234362,0.3351497384376734,0.23363498715410647,
 0.13247965132000727,0.4799529718127782,0.697999046376209,0.46438744035804036,
 0.2956632913447518,0.5600444081258275,0.6631382792270064,0.2532953209191423,
 0.6928589263007348,0.8924790276699527,0.48971301957173785,0.2190403744576117,
 0.7647166883100052,0.048010105719876295,0.5744744592893023,
 0.6801509584336646,0.10397341422409179,0.905580169835587,0.7287840291198298,
 0.9367308566194275,0.45506248078146827,0.12986992767837324,
 0.6752919682517079,0.6384153402882549,0.9671325663420112,0.5445640989767598,
 0.25725076084069565,0.45606812666735785,0.8854347539135426,
 0.9325997861160751,0.42993169705595746,0.6251582728616829,0.9136710251699153,
 0.4367161653099084,0.5819099013798702,0.348826244408851,0.23130424776884728,
 0.2095620907939164,0.31673835474982814,0.9809097501937369,0.7341754018470092,
 0.09622396399697142,0.42701953504238066,0.49512659232502354,
 0.6891281785973863,0.7035967749520254,0.5923726101877971,0.20512415362705738,
 0.6786799616949046,0.7356740081662347,0.6222432689508066,0.8139773374724999,
 0.7508155217566373,0.6864291805635909,0.29976226544966833,
 0.10888825095639898,0.21242969230528408,0.873176516951111,0.375741743857964,
 0.8755014387419218,0.14593811928732203,0.9793559368299672,0.6512087260422033,
 0.34671475513667505,0.8824985810609826,0.20989613173539923,
 0.47752756396460894,0.1663110063534209,0.9912932957124904,0.5865393647615997,
 0.3239945570483369,0.28517953407930907,0.6030351042134081,
 0.030295255669758547,0.8174047279365585,0.8587404336948629,
 0.2232921859457505,0.7228789071633995,0.7660646958217379,0.26195462967357186,
 0.6734738630845857,0.5960434121910143,0.09068495334850468,
 0.07386588957484241,0.4235232264329558,0.43612837645796776,
 0.023063840934935298,0.9010216831351173,0.26183691519385066,
 0.028727690517445614,0.17337229169802792,0.5966969354892978,
 0.5487280216630607,0.24556543469105563,0.16366824370654776,
 0.0338897471322952,0.13431280026756365,0.20659335213298047,
 0.6173728176694919,0.38536749534096226,0.5940503192398385,0.6489430249045872,
 0.21455609179165636,0.8472221085181633,0.5284171010991415,0.4058502361019024,
 0.5119017615979671,0.6512686750538476,0.8948538249225049,0.6568967316654656,
 0.46010728201147866,0.6812996387319922,0.30542883758212513,
 0.11396584693690093,0.30732615006552244,0.20749645516641213,
 0.9140810391051877,0.60898138972044,0.961228158890222,0.8687734112417618,
 0.3707352722123294,0.6762336232564666,0.833351343047388,0.5520497945225723,
 0.6727248389911672,0.7742983186173359,0.8660111049983698,0.3605926568698087,
 0.16208707596420557,0.7767995919528121,0.7078654508399861,0.858125164945013,
 0.6678089988414477,0.7490795781771628,0.37435612114784067,0.4389356339520859,
 0.611310923528444,0.8849867062696778,0.11089171355980587,0.3812229061871363,
 0.957685934153507,0.23611987130065726,0.10921896533087116,
 0.20072838992286135,0.4232041104023502,0.42151334101684945,
 0.16157114645513382,0.19905633033348757,0.05397717168987237,
 0.7051362790229274,0.781087129620678,0.379293519896459,0.9668431414844536,
 0.1538254629377035,0.6346917989927906,0.2062837639628532,0.6147794766731254,
 0.947192787432165,0.5968756123693961,0.0022712529541643045,
 0.9760106373179394,0.24336488996952688,0.3387517599132088,0.8808028735581591,
 0.528185244930552,0.02848539806469208,0.3002315564164375,0.1213222928093648,
 0.3876916030626728,0.18101129922120873,0.5334966136152653,
 0.14105384129797016,0.2737177703453616,0.11326525543166788,
 0.1906898677112514,0.4406412577616152,0.17913150551536772,0.6324669121182075,
 0.2400303575077516,0.0969847386450089,0.8446582068882932,0.7057381540104302,
 0.7386025935929678,0.2148884954852326,0.7364494623554182,0.528831502674548,
 0.08490333480949563,0.5454080181813963,4.095842502183089e-5,
 0.02908016602905772,0.2762989982087548,0.6735699658683335,0.8077081787917452,
 0.40922389469911713,0.946790466058628,0.3451966813291387,0.4760007558556447,
 0.16907562801519915,0.9162078357782557,0.11056150365687512,
 0.13412327366339472,0.8532502507720466,0.26710607468770964,
 0.8117822318082818,0.24756280535431918,0.030277751152139154,
 0.48790971093546553,0.9833555730370084,0.9327949212771047,0.7309732995886367,
 0.04915595623946367,0.784668131182991,0.4065368957326507,0.9794199880042747,
 0.1887178375664017,0.2626483451028455,0.14997570696421658,
 0.11269157822334919,0.6359320202000905,0.008519236489209159,
 0.5288175632311052,0.1449538766594134,0.462522483025256,0.9919445937218352,
 0.5230871126850709,0.34995248543649815,0.8599279508696721,0.562628052922281,
 0.7296224138914116,0.44671533340369973,0.9174118183391621,0.483813726098993,
 0.589922743914992,0.17956901473776543,0.8627574506177644,0.3384132839425841,
 0.3125446286100412,0.48252510935057014,0.7874467451368082,0.1914990479977916,
 0.5854286021759136,0.4445507051742551,0.3185420048414628,0.12020468273880813,
 0.41021775914140246,0.41367487974197614,0.8389375122657369,
 0.10898763558818414,0.4418526694523368,0.36468677672099303,
 0.48854524439748703,0.5357990247912052,0.6038126560570417,0.3909061440377749,
 0.45566287363197366,0.2963755623678135,0.5411774809387011,0.5140634613483402,
 0.8866676763037398,0.6156411544118379,0.5613876720418093,0.6881353555381257,
 0.7118189903689399,0.8396770023681959,0.3750743234657652,0.597401331691487,
 0.8141400252186796,0.8417533929066692,0.6985086442455378,0.09683319253572009,
 0.8688186230152558,0.7597317340611985,0.2661802096310405,0.38033072842066706,
 0.20209504512086462,0.20826987711947242,0.2861824584871856,
 0.7017771125124579,0.21561709685963226,0.19550520649255665,
 0.028907085759663165,0.692683227630844,0.4549599999304259,
 0.36278819160498443,0.009638112105341823,0.5022538696172282,
 0.1082059252464771,0.4130008448039776,0.3571898048519091,0.6875983166905536,
 0.5095148228673728,0.6067740943118984,0.3107133860595882,0.2664641784234397,
 0.6505574398975582,0.9646325596699745,0.11496944503329554,
 0.12353610159710493,0.742950958830634,0.7257997142628521,0.9789741651552964,
 0.6287321473280503,0.5857032548068654,0.15044531459087251,0.560914284734815,
 0.38408395734140477,0.6438844225461805,0.4308481397524635,
 0.09368849032560533,0.7778260683476118,0.6380459760290172,
 0.17724126950225838,0.39226015007668824,0.04306089041311567,
 0.19749479477808862,0.33633498436857034,0.05797043085544262,
 0.803416816921716,0.8775828812135947,0.37128951698736434,0.3669001546452133,
 0.19612186910918117,0.6170169370819827,0.24160340585062523,
 0.7202120215946091,0.27191117527100483,0.42524750073785855,
 0.3796114064645384,0.23791167883701425,0.13465854169570401,
 0.9067637715240952,0.30951667078619194,0.049664908586278145,
 0.9530871535968577,0.9998563168865676,0.718555074730957,0.06554119435832262,
 0.8361448790011048,0.33330872169070913,0.0561842148781736,0.854442722755302,
 0.503475566489941,0.5213993152420919,0.28706693751142864,0.9950884554135989,
 0.5498805266271847,0.5929120046775294,0.103096207106666,0.7420220069145168,
 0.6355653237807399,0.24514815603071738,0.7603719433519203,0.9978320420296397,
 0.10614748861383716,0.32634393393275296,0.878385779382395,0.6194374171055084,
 0.44455064659502663,0.22470911431312646,0.4540844025231763,
 0.9899602483679927,0.061852265957209496,0.302273298228926,0.3632706623706181,
 0.4485958815395502,0.9118285093023769,0.1820267467110912,0.8059284723430432,
 0.6066665262804071,0.28828746793607185,0.8047725297212368,0.8610823302449438,
 0.42003808813745536,0.8038296450572296,0.995995654059658,0.2059168432194023,
 0.060697173517441194,0.18327789222202473,0.05679681855363483,
 0.8987051776933956,0.5449581643560838,0.8672226670011175,0.25218867018278723,
 0.6068919579113001,0.5268203277820163,0.24374808450266439,0.4872835940641146,
 0.7268836462849203,0.9183863012696375,0.10735019775130505,0.5509694419351167,
 0.6614951184705596,0.8217776787413917,0.9007788658278488,0.08873774469233142,
 0.4689334656134314,0.5509668861797644,0.6410021363656634,0.2024268791457806,
 0.3604552211772448,0.11012372504517909,0.18240350497259694,
 0.32916862201938824,0.46849412249334044,0.3128257320741228,0.50323077517472,
 0.714771678145903,0.5438728639487089,0.8857716480282228,0.9905642896854153,
 0.704421770311983,0.16627164004410666,0.4907213468634051,0.06384983744159789,
 0.8384268501682999,0.24494808020640713,0.5869786389576936,0.9388229357157852,
 0.1283504565776119,0.07989001133546714,0.6390620672585844,
 0.22726506762014576,0.48515806960334207,0.9986120403709318,
 0.23082625218435282,0.538856199793356,0.6926118101151741,0.6092938401340731,
 0.4368411635200502,0.9076704812277107,0.8886814576493738,0.7370763513016724,
 0.6642443407763168,0.604896169128482,0.9649529863977725,0.7481019292615574,
 0.3719340320176545,0.9419253472902821,0.590690486205998,0.5143869885502346,
 0.8913015219035263,0.5345212350399617,0.7736485655619838,0.23091377218805464,
 0.8029450672319516,0.9985589626619009,0.31204433368153517,0.963811365561373,
 0.2426082392885638,0.8393803516921241,0.06903110175593608,0.7820045707662094,
 0.9189939382253683,0.8574007801940439,0.7056239924935934,0.516148385566557,
 0.6513521243944423,0.39849154851184887,0.8660934948021523,
 0.13518748310738948,0.3131446411977076,0.4667785806725826,0.3032943328608517,
 0.6807087025543977,0.02666451700729766,0.9158807705527479,0.6394089669811562,
 0.28577996773296577,0.4110362380548056,0.7756747864596755,0.8988997708509503,
 0.9789775641838212,0.6985286587645397,0.9144914082098794,0.9733089693104631,
 0.37454585532214524,0.6056663205382078,0.8247671516784371,0.4614118072030392,
 0.1793398508931734,0.8530866016552738,0.16881974808853595,
 0.050413442806596875,0.6771703207592569,0.4358934324310706,
 0.40456744146763823,0.007118733234505403,0.4870155942391413,
 0.19609083012954887,0.8386355791199149,0.3732718250274192,0.984258909385823,
 0.2506248037147579,0.44376748391979226,0.031241668939240852,
 0.22165846874455297,0.7308288198878566,0.551141194211814,0.03205576716582592,
 0.21838644948151553,0.5499025188228317,0.2370541349304569,0.4322429779840008,
 0.622328195751163,0.7251088567066493,0.07380209851908748,0.0466228656841039,
 0.479807167199106,0.04830221049187933,0.12872320733458775,
 0.11630943632640478,0.9358790032675088,0.025936984712264466,
 0.8530565072103934,0.3783888146335982,0.5765772816749475,0.3388372034750573,
 0.33425347325586074,0.4416193507531009,0.4571221484284105,0.3926486733876561,
 0.6895903603876831,0.09913027987316658,0.4286086015301346,0.844776745411926,
 0.8315107153181907,0.41026413230121617,0.24393993846080897,
 0.6598553559461746,0.9486294819779987,0.022951726714040888,
 0.7067259292810031,0.21450538204655112,0.11165878769112947,
 0.026916514169726402,0.06922588950791997,0.8512203853520135,
 0.369223496186716,0.39136085680452937,0.5318005044472054,0.5495210881317455,
 0.5853453379188419,0.2975310476059545,0.366419132673506,0.822938458846212,
 0.6959423147178133,0.7229607748821337,0.06791531347927693,0.9193349609020772,
 0.8633149855437378,0.5312884086508259,0.8309730643642852,0.3620084102084812,
 0.8873219551760205,0.5963629544420056,0.5492571053729413,0.35210030156017114,
 0.3257377261713388,0.430977702638357,0.36036451752158394,0.14172625941449069,
 0.5852924349240312,0.7749158085776122,0.2012287542499862,0.8760096496944952,
 0.14711152100592195,0.7272916154578011,0.14762453059220304,
 0.9527437872408443,0.8280316696339896,0.5008838561101828,0.8531613732863459,
 0.6104642455155268,0.8832351860939315,0.37595741777367453,0.7417759331244802,
 0.6382049733079831,0.3725083589520324,0.2962459037158046,0.5222101620733026,
 0.8137468807862449,0.06140280676694765,0.38240519008524587,
 0.6082473675933076,0.8385266438079049,0.3834905339475648,
 0.021546839462150302,0.8893421401286814,0.2633659506739674,
 0.3066544413814385,0.6884917206283967,0.8524082533885499,0.39785853200809407,
 0.4968091023865818,0.06127393653506452,0.18931650118373333,
 0.32359343841328303,0.6975392561419633,0.176067715422769,0.8403233825405445,
 0.5731334373504178,0.4688123384257975,0.9434822843825622,0.5668087269503311,
 0.7765609255741713,0.25214140211568925,0.4472529286945709,
 0.20784230106507307,0.9203495240817813,0.9412934654883709,0.9766291195873389,
 0.7740059758179167,0.21702491375528232,0.16929941096522605,
 0.4823668707423032,0.013426007782796345,0.827575489255939,0.60976860698144,
 0.12478856126992122,0.6324818236888621,0.5062331533622506,0.5323065651508885,
 0.1959132544768264,0.528388772196664,0.7592197722202432,0.7036540358057667,
 0.9172938544760384,0.7062733671882737,0.8430396530547838,0.41835492168432487,
 0.662497917300328,0.7098075078234849,0.5178905680550477,0.36572837892691645,
 0.6750266560634809,0.7766091740389888,0.12219418445139829,0.1854005481904727,
 0.03151782457324748,0.7379482753687329,0.46765391608608864,
 0.5316018122130156,0.36676102930558785,0.8538759668297987,0.7288520923358214,
 0.05537166566210505,0.6593491195314684,0.27545458381543186,
 0.4562874276204357,0.42206471717988847,0.805448197014146,0.8521327909901748,
 0.30955866682551936,0.7027546377862028,0.2590549639797004,0.8168762266181476,
 0.8180625962914385,0.30827013102640544,0.7045256257431232,0.0668257319519967,
 0.15443117876429613,0.5571030633795871,0.019338016464797936,
 0.9423245350636003,0.5819780148879004,0.5351234283094534,0.5171357219921167,
 0.7687407953410194,0.10192606910055191,0.6559184752718527,0.7527434544658629,
 0.9599319040023193,0.05756080044474954,0.5355245833599582,0.9429848027456476,
 0.7989440535253498,0.2669606386731327,0.7984444183215114,0.8625282348210991,
 0.18816646870828746,0.9436468492766638,0.5881905322666874,0.9720452613052081,
 0.5862329147354269,0.819205728378506,0.41532424774294574,0.5151207633253132,
 0.40867200015509064,0.11592875955682713,0.5999305435443705,
 0.3457574491170795,0.4898760786873396,0.452359881687717,0.012564082388266895,
 0.9541879132273322,0.23392972190703132,0.7237400073935989,0.8175708341864786,
 0.8543536151552992,0.8423025408870227,0.507598603992195,0.20599330129736282,
 0.8186989673413017,0.866315529442958,0.19200028638072242,0.3302881161741015,
 0.2710928251779193,0.5309787954092041,0.4277582325469387,0.9204815892439293,
 0.2329923852003869,0.922217849748955,0.9895813032049415,0.1708791303455054,
 0.24162320150207972,0.31098389735530807,0.8194100346367051,
 0.6783272309870675,0.8062224717202835,0.9265722331297139,0.39592074530908095,
 0.7750801565168272,0.14837295073614254,0.5797354632228009,0.4588713507714597,
 0.9943823495845302,0.0680559713397948,0.36478469421389526,0.785364954376351,
 0.18429648034813007,0.5924162979365374,0.40805770513839934,0.531973181802722,
 0.38570106162850026,0.5881407340691858,0.6563385358947247,0.4109719838580087,
 0.14785389569768748,0.9272987672674152,0.7886807784793033,0.2080652887861505,
 0.23116425124738083,0.5780772134041454,0.136612358997098,0.18459095493808153,
 0.08958324411693908,0.7945264592747959,0.20557528671379977,
 0.4936847122711957,0.8037017555390185,0.7293545241494693,0.761336731910433,
 0.03761465733160618,0.909900353740391,0.1907510401061724,0.2516739567574703,
 0.9066599170538207,0.6834411678641938,0.14477699163491253,0.519115158211946,
 0.9624886464998252,0.5262645608868206,0.6315482322623314,0.39552718495993455,
 0.7035297343737542,0.2582598492683249,0.7616209325676138,0.9770934126127453,
 0.9967830646463155,0.8519788141575327,0.9385436627289705,0.24280868867548755,
 0.7644073802531683,0.6116670390628691,0.1906469172340619,0.30848558319964625,
 0.8819308942556778,0.17849963708884353,0.04955868695647836,
 0.2992395630533313,0.4207329130361306,0.0756024195686269,0.8085639552813838,
 0.5110760801271834,0.100945368559889,0.9627018700433033,0.4281475598271465,
 0.8554607037272534,0.6450984040239063,0.5192798841113984,0.9262149668243885,
 0.959781054468849,0.9206474860916879,0.4408909177944149,0.03545041325521514,
 0.4261404036836711,0.24009862337074184,0.0060873329250854535,
 0.7106188527458162,0.45987457423038514,0.0016711253301243678,
 0.321636441634211,0.17177301389176203,0.740442229753733,0.3099098910730945,
 0.1727288725784749,0.8377880595131033,0.5367929420878975,0.5992638564825497,
 0.514606312117189,0.10626292665074655,0.6688337001587742,0.9028504571808892,
 0.9691643006980951,0.9899531085394331,0.3407765913466403,0.15788166639297685,
 0.7808828902316475,0.18750536499226622,0.041053512466276754,
 0.19687900729839092,0.9515905508813417,0.49866838218238985,
 0.4397404219256559,0.8280006080090521,0.7490494518462226,0.10819363934519488,
 0.6279417155895843,0.8038837385578493,0.5385574547009536,0.8059761815599742,
 0.02204222200050443,0.8495219297288776,0.7662461864629327,
 0.08819742623117377,0.40451091998649935,0.9117983709687293,
 0.9849391856191511,0.025981352196347673,0.5432171911558631,
 0.3996102520726763,0.593511373526882,0.09957999442864429,0.5575373429069023,
 0.5554349431423062,0.2187691915566763,0.8785990999455446,0.990176589336301,
 0.4566877419192774,0.662741136460501,0.17634460625342785,0.09869819291520532,
 0.8900515757147858,0.941908164127291,0.0652394868291225,0.2292297004648708,
 0.7812298884845559,0.15866050110130725,0.7880820649388208,0.3028244105361291,
 0.2795201156861089,0.9606003185494261,0.29608352585315934,
 0.11827395665784046,0.11626410904615236,0.04268851450984623,
 0.4605674677145427,0.12473523423645971,0.9972962959177009,
 0.15133929069556962,0.101257140766291,0.9678702552247999,0.11574795733158394,
 0.18495684765775156,0.7048829193030113,0.8416082671438758,0.4124765037079573,
 0.6411191253094184,0.185601662121191,0.6892556828749433,0.49328642691536206,
 0.3121769019158729,0.5704815902057974,0.34685762623672967,
 0.03053863067215168,0.7810107438153476,0.2548918623192009,0.1960221157922799,
 0.327803914457083,0.153978204614296,0.7375812478311333,0.234051197516436,
 0.39114291249902144,0.6511511111624376,0.7664228955626244,0.4944118407974326,
 0.15501644376197,0.4619932962895157,0.10232972392277284,0.5993327949020277,
 0.7212617833576493,0.2184327322653562,0.6226197067042167,0.4947483113229163,
 0.2135245729918217,0.9573797768584987,0.24290561805916933,
 0.40395485169261847,0.5218748083550695,0.3836400668330644,0.47498653059873,
 0.9866291083969333,0.5883291943257993,0.6040577011712298,0.7227139960306661,
 0.8078181615191218,0.798181906327825,0.5731014575097445,0.7978118138373261,
 0.909927579320077,0.5789586026456524,0.7991729313664466,0.4778414952203699,
 0.25456725645311684,0.8302580221531097,0.8275594430213977,0.899408952658982,
 0.52329238773974,0.9448932516050098,0.32280892787260373,0.7590646233077899,
 0.9998213866085118,0.5917646586038071,0.2719215048885275,0.3358397918759194,
 0.5266406848598625,0.7797209162746509,0.6728286376232016,0.4515449826776431,
 0.8999820465064663,0.973026466564139,0.8945165877004623,0.38650677762797525,
 0.522723611376741,0.49200463440416875,0.5616177143145309,0.9460724084038769,
 0.7000086794865515,0.3176840758137496,0.6862629887788639,0.6590136939555749,
 0.8033481025787814,0.9260445253457686,0.27790746540840994,0.6620197722632593,
 0.9425366897616195,0.0012942104004035615,0.48299131873948453,
 0.43941718605984903,0.8638543792463984,0.368844251465007,0.6601309304022352,
 0.8198259292138593,0.8994774665167888,0.8202365851785602,0.33324142079578656,
 0.06782443772805113,0.9401942816792929,0.8697766147711588,0.9292016256275413,
 0.3765601084395327,0.2446383789942077,0.580401891916412,0.4614836802491893,
 0.7537460461101217,0.48042191762800956,0.9847373813086125,
 0.43747075632348975,0.8663395245721437,0.09321398352662125,
 0.4027878712419325,0.6674724299085637,0.1150469928716823,0.778297013150438,
 0.14151117142011782,0.24473292354666165,0.7918727034397001,
 0.5548872068296478,0.13030909612579333,0.648600890077226,0.4849384296866409,
 0.917725984835164,0.23984699686370714,0.30840031438426263,0.8482366337704218,
 0.33484243100050226,0.7838706725069082,0.016591215567175466,
 0.17343822367992257,0.5111921284560177,0.6000563431959022,
 0.23263331065689963,0.9753950070992857,0.16143100218215878,
 0.6435395469107611,0.77793273364006,0.30180979461142987,0.13328390868330076,
 0.32739581949298413,0.07843599240921173,0.7690452015195429,
 0.4221259007469873,0.6481321084164979,0.5477825379975304,
 0.013858329361491117,0.5535127430746898,0.32449415491367617,
 0.3189737074165553,0.31650267150746514,0.8195786374269218,0.3204325539563414,
 0.041834870252475564,0.15381149361608515,0.7677524227863957,
 0.9239439889926628,0.5147114122913672,0.5811782778366698,0.13573574539370514,
 0.668258626389485,0.7583844981854408,0.6394269430834146,0.9245777164669071,
 0.3550049400421057,0.20672474363830773,0.10587322464521098,
 0.02442923809208375,0.09412709639902794,0.11256513577419058,
 0.2698433456631357,0.5811753335115815,0.7134909608824893,0.0550875336175235,
 0.5887411972266631,0.30739547064849704,0.45568331554164576,
 0.08344769796707219,0.16249548908684808,0.5707703750821763,
 0.5122666534473544,0.9206436320975747,0.0067070585952137,0.37844081255569006,
 0.7474818836123608,0.17957191069980605,0.6036284757343244,0.644153955060301,
 0.9301651587707133,0.1798252888336227,0.3341616593539406,0.9418868623780503,
 0.7881738437461565,0.4789345622311164,0.33824531993491724,
 0.04907650241011352,0.6773270294218281,0.33781677647854147,
 0.37635314516879004,0.36513304563341664,0.7328141832569045,
 0.18814946800931476,0.7745535850743916,0.7760868040228592,
 0.21046178762003975,0.19371096732318827,0.3110316857236757,
 0.1232141999429961,0.6956380693576645,0.2953127870678697,0.8630249310071947,
 0.4038556471232472,0.39081041298023333,0.9506934100491308,0.5865953489713179,
 0.3666353273071812,0.11970084052076713,0.032080198268419124,
 0.7697452726325306,0.3901894682377567,0.39495189875398506,0.358421890340195,
 0.4433716153370115,0.45112482289818856,0.59626885170524,0.28769023972711294,
 0.3359575154177872,0.6444661962310774,0.9319032442837634,0.8522655036550256,
 0.5801067321958745,0.9810290351692714,0.7360515656793821,0.8967339090279154,
 0.3581853696552695,0.047026435910792364,0.0805692110334909,
 0.9054556140322597,0.5833302491785988,0.44726799108336146,
 0.41304338791001827,0.6819754991869121,0.9010058692777444,0.7255235691822409,
 0.90320996300209,0.9352860989098096,0.9865939137055926,0.5020789346953378,
 0.18939312191926083,0.22335728055525228,0.7921545951089205,
 0.9798403166901868,0.861431664994698,0.9127994914827764,0.3481770358472219,
 0.45793992516185655,0.8799124830115052,0.26833322725809206,
 0.9500891185436395,0.04983746206257378,0.8659815728594984,
 0.058272755763959516,0.143121606351311,0.08415354417457244,
 0.5313134773048357,0.4754738129113143,0.5343521749193768,0.6645647681080712,
 0.9134270743908974,0.9941131703024901,0.3142366381117536,0.4650597134385387,
 0.2548624267453212,0.973989346328488,0.1718933840527722,0.10955554889847008,
 0.328174495609871,0.5336278868657365,0.6944497333089721,0.11013619480146164,
 0.3359984845742475,0.1343359447172191,0.422713933425644,0.9504913213695853,
 0.7457758295690929,0.14277780727397638,0.8315515773396251,0.4248775532665332,
 0.03666055606097085,0.4443306631061348,0.3853012622126133,
 0.004991225192278792,0.8863001575269523,0.08521045526845017,
 0.4672051530619742,0.07255877870121819,0.5218410933025222,0.5379609985094822,
 0.43192192678248054,0.8020236132387735,0.6331164007849609,0.9102230138054417,
 0.9355303856207978,0.22816908688886106,0.7975216973755928,0.9585820209156112,
 0.18119674899403693,0.1909190059257042,0.4733189578543189,
 0.004205246739121371,0.8247125008164676,0.5710042339800362,
 0.02203032955984774,0.8828026076859831,0.9533125170959195,0.9839885268517484,
 0.736218340062514,0.9845081136089323,0.09475639559967575,0.08471245634840296,
 0.5164652688150941,0.15367543351691038,0.04223020432787172,
 0.8681110101645949,0.574901854213768,0.23360494332539594,0.5805907817484676,
 0.31451518842429516,0.9585841665333226,0.5981535717117256,0.553541226518177,
 0.47798818814869304,0.4847164593543053,0.6786345653645552,0.8911856942688187,
 0.02951072779914221,0.1770702294633144,0.646249471084341,0.9337560557196078,
 0.31056101664451974,0.379904782056672,0.5072720464907601,0.9205402036039114,
 0.8646180775847546,0.388749765044897,0.5797543206024606,0.43555988193755724,
 0.46511221126394353,0.8289868793571576,0.7010314776815821,0.8335948039275525,
 0.7453654647141165,0.31116170615291616,0.1926988828377636,0.5466465667157308,
 0.5011966541743684,0.6252252053088562,0.7635958339519269,0.4121669133707674,
 0.07415149768660578,0.40848734701416034,0.1590583417449093,
 0.6464572740711533,0.7392653352430092,0.8395256583357313,0.633166281585237,
 0.643235829559251,0.3791680699963794,0.31419630892969497,0.376624742133165,
 0.6653687701360673,0.8360841217003223,0.06938027918738066,0.6754425556429663,
 0.3882097371445714,0.7758038144212636,0.410356495913636,0.6345001365852918,
 0.9338386886114641,0.9417452679215722,0.592524892149298,0.42507457866631915,
 0.5712094166806518,0.1163008672818936,0.5169604968715182,0.40107150982798334,
 0.11678522262963609,0.47178115313744917,0.0560255958117577,
 0.9086515572497571,0.7021333449096477,0.43510649594567896,
 0.17690148296233077,0.8552801227190917,0.789031251106955,0.6249492646022659,
 0.7026601622333869,0.8454298424198231,0.49618805680169764,0.9777746172966492,
 0.6641083252251351,0.31133061783259475,7.324405165308745e-4,
 0.8107563985072959,0.45917875978328704,0.9740160253808876,0.9327087145742224,
 0.19256437315168118,0.8898975855551678,0.7248178312545717,0.7988321759227914,
 0.7031158388549585,0.2076138900009601,0.4544285932595692,0.5374233397630813,
 0.8697212909299176,0.15847441159397302,0.5457711409893934,0.595900394702994,
 0.7070494134461176,0.15944301273666772,0.016403301070381904,
 0.0030268538760868102,0.3999867690142591,0.36496766600565833,
 0.857015915520072,0.42277960229947553,0.6282522792038199,0.29078463547204936,
 0.5225424407483203,0.9757465972394546,0.33613625163389227,
 0.26893008783761907,0.11574961766124892,0.08525280792869938,
 0.09455881595623128,0.3930156112924892,0.19343249275966012,
 0.5592591542099319,0.5639407007888946,0.6452236085247066,0.517522750951762,
 0.5740460512243023,0.7212291850685995,0.5349653766876871,0.14363990053510256,
 0.8907646604186581,0.1593752532339613,0.6707727291792992,0.8191328668538745,
 0.8659292998283052,0.42378130621733434,0.9066505643838085,0.6056203853912078,
 0.3278856083531084,0.3074750204643053,0.8628910154755168,0.47202847633452094,
 0.0322548653139898,0.6467500413745072,0.8894274451428471,0.1476623716868788,
 0.9332375046506877,0.44068542449351633,0.5718080065629061,0.8612735161422262,
 0.24348730829746945,0.5307104951426977,0.8424936023596928,0.6605046991156245,
 0.26987241347709956,0.5287564803316389,0.054039005127695106,
 0.767677437096358,0.9937543500952476,0.48140959930247207,0.5803061700234583,
 0.5710416860008202,0.39588727939539803,0.01873824830714199,
 0.9675491089250109,0.0715617993661235,0.5908026395497148,0.04613177128187629,
 0.9129716571930797,0.3679854131655722,0.7642290284756654,0.8271492834508724,
 0.2854042441129183,0.7214930429830971,0.5984189866246254,0.7325275846977115,
 0.5181339974974754,0.4359525021971824,0.5112815181104874,0.2663982430970485,
 0.8630871435598614,0.9734422416550073,0.5438603402580957,0.9723313829002275,
 0.6697437656017338,0.22870794226671742,0.31891075430395566,
 0.7353425078751626,0.09712096633196021,0.527822685405358,0.5178809502487046,
 0.18404679148840353,0.5547615899259224,0.6518705305677122,0.1885760526424779,
 0.5119410167331447,0.02171794856049536,0.4731423817147107,
 0.31577618139736097,0.8599968220020701,0.9865476390770866,0.7761270365381681,
 0.3428849127714245,0.29902595046664227,0.02497840067923629,
 0.49667711921710433,0.2181864738526843,0.8860497446898868,0.7980555769699591,
 0.31280641134548237,0.740428520356981,0.8088500798573227,0.6969671346782924,
 0.0530889410434926,0.5166337525023715,0.20866479381797431,0.4680959313361127,
 0.6896844166190735,0.9400280373071224,0.3419082138882751,0.8477635358768237,
 0.2480613961375817,0.09886975995561176,0.21345369835665517,
 0.9373313471768914,0.856956096940286,0.8215518592528323,0.1095285828382605,
 0.28566742795717415,0.053790027941791,0.9429644847974213,0.8712233930183054,
 0.670925357733881,0.09622889680708169,0.3414563796103659,0.15560031825619536,
 0.01531229081648422,0.7803426468623108,0.4029359441875261,0.370643029726318,
 0.14008145486133705,0.29338913934572597,0.95557746349974,0.8351783160952315,
 0.46891302055592254,0.6595057713549651,0.23125869316567593,
 0.4463222367696029,0.21662156564354418,0.26856955462473464,
 0.02503009378118881,0.5044295115543094,0.9322627253750057,0.8591029564989632,
 0.009158607563902854,0.0376985399776022,0.2289280051316761,
 0.09786171051175174,0.517069446233273,0.6208775844998223,0.9122761530369194,
 0.36511376785229666,0.004639627776138688,0.8866630871771344,
 0.5173417042997773,0.36411139829473427,0.5566171305859604,0.4915269902259929,
 0.12188313383710225,0.9641728216676811,0.8059548034475449,0.7961065628871569,
 0.9956656940982147,0.05320819893298989,0.566137260516991,0.1162063267519644,
 0.07384728011735042,0.19614383855741901,0.7565864350165551,
 0.22446527934005334,0.17663732076226846,0.04686469486329292,
 0.951994676826817,0.7598224660604332,0.6862297664184995,0.41285977180043,
 0.04743300784728677,0.07324507456509544,0.8365916730925773,
 0.6550650636040901,0.4355769480743672,0.3452978142213229,0.9588771633551362,
 0.670733049928504,0.6382366561023666,0.9164882986324722,0.5846557437507993,
 0.3154445110262758,0.5938474703233887,0.18039993476051908,0.4110556703420597,
 0.21282408253704255,0.3889480422519157,0.8251885854577442,
 0.21917805807522295,0.16391730533305993,0.6026498410411052,
 0.36953228567410734,0.08849458795516696,0.5847165405482013,
 0.9390669376736382,0.3518759171773532,0.37839926161918624,
 0.16983615799333174,0.6155710588532105,0.5332852682916386,0.9801175291798114,
 0.6906923757373693,0.8151078186132623,0.8904859393874999,0.22336975702289308,
 0.7915357671951949,0.887705063344336,0.8752757741273824,0.7995713103198933,
 0.037912285180883076,0.3484836182991541,0.5200508625687755,
 0.05849836073379322,0.3487638558508551,0.23835908229754788,
 0.9108553949023006,0.1863615881591031,0.7299342621499572,0.7437155775600957,
 0.38551957201523357,0.3885966758991559,0.1507871639158883,0.998386036279461,
 0.6334517386690779,0.8751852812804486,0.5387147580450748,0.5776429881142484,
 0.3145391363591785,0.5591810833008649,0.7986820758100512,0.48431130198655126,
 0.09958179623484986,0.003308271922475736,0.6362356329545419,
 0.6219173847220313,0.7958610226746451,0.3780750731465875,0.3924032614181976,
 0.15418999062503969,0.6152236474421874,0.11692559426955673,
 0.3445504632240175,0.11178245142109078,0.07391269587427318,
 0.8338496547931098,0.6616447700331246,0.11741038041319019,
 0.36795158855284527,0.3114841997039284,0.8885631429123593,
 0.14295827549789575,0.16606933888687303,0.3178897751544403,
 0.3788495924616533,0.5390698985575509,0.47748346637865424,
 0.47235281551116837,0.7186795858657693,0.9012255004211783,
 0.44997606269414614,0.2128378091770441,0.1442387326485819,0.4363276621875979,
 0.9837778102420557,0.5248002381081279,0.3959596069489867,0.196020663012612,
 4.866043305913337e-4,0.7438536402493516,0.8392304610076897,
 0.04539942080650383,0.4172772921481108,0.4175331509195628,
 0.07061218882324194,0.705363511735117,0.47412653663372895,0.2782727860673061,
 0.3960397224546993,0.10978830690247432,0.32339373263700155,
 0.6680121794545018,0.9887004794996321,0.12645703183212686,0.4842719495515394,
 0.07906767866940478,0.3078944514559041,0.9165313164890384,0.5769285986448052,
 0.7189536453514227,0.6886943993340239,0.21282998794325647,0.9143394469669557,
 0.10233096707286526,0.7620182808564571,0.9373036025640908,0.9774537021105265,
 0.44195910216572565,0.68055692000197,0.3824681510352148,0.23464813338591362,
 0.9056563955144072,0.718548028181854,0.05050096915713331,0.3120987735181452,
 0.8603777699351692,0.9059020825328938,0.39667622247913537,0.66909463381579,
 0.7904214086440026,0.9330350758865964,0.9802398180296699,0.7495992257141622,
 0.36178788084873326,0.11914103663556252,0.1247839370228242,
 0.6952213162912705,0.4617670377343649,0.3120040400509855,0.32134640065967623,
 0.3611276418217827,0.5597144760281121,0.6189199018938467,0.97155382177269,
 0.4180945324196007,0.4169032846555454,0.5986949108975435,0.3249014515582308,
 0.20921803783665072,0.24730774899000385,0.8443577840437126,
 0.5372207889338046,0.560977962772702,0.04520512711850588,0.7833225826472537,
 0.6151807116377819,0.04215694482391852,0.14641203878632225,
 0.7574207750266099,0.6867567005723049,0.3906475064530778,0.6019606033175744,
 0.9151920514252073,0.755700453645763,0.24618297440148407,0.15224297076572935,
 0.12663847047093835,0.5846395958134298,0.20469520737119873,
 0.18493019262865096,0.13197236894357756,0.9376643822348453,
 0.5171370816550426,0.7738712176378937,0.04931129215482799,0.2993970611920834,
 0.13040695785848155,0.09544579552446297,0.3405084134220234,
 0.9210301495527804,0.27436301318802625,0.7254249754058212,0.9899951983906068,
 0.7801016553305593,0.694911210590524,0.6526826544556784,0.24006662011274105,
 0.9890191792987353,0.31822515506820526,0.45682936226173154,
 0.6695153843281947,0.46249399984858963,0.19308459628681285,
 0.3254053429348239,0.9805907251301005,0.7021271360622501,0.754739064301598,
 0.13150717428811354,0.6068300916951267,0.07270826857142498,
 0.7251675645056161,0.46880759248052783,0.7082696964507886,0.2426590799672217,
 0.7293823600988835,0.9366556106091244,0.4303252920422177,
 0.050767972706584286,0.4093625805669452,0.9073929713359504,
 0.3040468301361078,0.7180268241383854,0.06657867934699879,0.9412868055154763,
 0.6970203165814921,0.005364383165111453,0.4305716463490661,
 0.1850994438653999,0.9458238473552161,0.055140324164440924,
 0.7361907186986805,0.9129424957010799,8.003934880970176e-4,
 0.050770970134639826,0.6049089143443017,0.273328643943502,0.5690679505945021,
 0.3038083169884279,0.4671379092636829,0.6671854710974164,0.41618464472795,
 0.3997751542500644,0.9063067907729623,0.7423945268503805,0.52964313585114,
 0.5053808670757953,0.3829458242579563,0.6056780678087759,0.18912111240814833,
 0.4214619673299813,0.08551126335861736,0.11387295834684819,
 0.12798624883420862,0.9705688568627356,0.08821077179255865,
 0.9385505606711941,0.4275172525834152,0.11790237577004403,0.1520142899795932,
 0.8896343382461301,0.8971256033429883,0.9670181154269352,0.9437603511865419,
 0.6918676525971467,0.8661287786361843,0.786711739053031,0.9097453585721509,
 0.9838696033077005,0.05933033547102651,0.5069166956667591,0.690809512244183,
 0.8951707187192152,0.009414529247409487,0.8668457983328873,
 0.6591996222593046,0.020715542257771946,0.9918310223314475,
 0.6615225631103963,0.7196138548131761,0.2069708656358108,0.3879955269659262,
 0.5494880600064612,0.19545944040043706,0.8378808582320874,0.1876713943280669,
 0.3505377301268533,0.8620453615579143,0.0040051111928829375,
 0.24956141537174836,0.5366210676955427,0.3228960523369716,0.6057013974632562,
 0.38271205166690303,0.8084897450246159,0.42042101273705745,
 0.8472913665488153,0.5511695539549103,0.6635787093262242,0.8909087077721705,
 0.794419126153006,0.338264507775633,0.6045114997978145,0.8751372572200015,
 0.8383997753041657,0.5065126060827666,0.14825840421977732,
 0.11983635765852196,0.5892367303485464,0.6429446924781691,0.9316943665492485,
 0.8360670853179543,0.5696509176347884,0.5416176900014511,0.18678934962326643,
 0.9562317887748839,0.39068755786670195,0.5031721304987411,0.3507608761742216,
 0.8628797547889624,0.9434326629220676,0.5432219030801292,0.8342517339657592,
 0.011930414740448314,0.005735953046693498,0.4004180852043484,
 0.6728287062035844,0.4799777138237047,0.697859272279865,0.4011164474230301,
 0.2126606552798027,0.7622704690768275,0.16196538751067613,0.3282936320397998,
 0.4001109855945799,0.02660699946022449,0.04288427419845153,
 0.7029212568107326,0.5270937278773549,0.572376664021926,0.7982851494620626,
 0.3372564697164575,0.07870557940981282,0.6886519349381522,
 0.04016365193136007,0.8651059383280361,0.40871853133877756,
 0.48794320843585093,0.556265259636931,0.1259904287089436,0.9701226819676949,
 0.9501627820337171,0.004520647178587578,0.3278168993479842,
 0.47858866113941123,0.7232983704027477,0.5225634031403513,0.1974090570333248,
 0.17678802261681548,0.8363982652353483,0.5408833718496736,0.4161750253221509,
 0.6847963546238164,0.4534981431777464,0.4307228775994072,0.378997896495906,
 0.3621366373998782,0.16950615280220283,0.8186396364026282,0.8001510547781664,
 0.44938029028990023,0.5629453791872041,0.003605789152773431,
 0.9455167499060139,0.7212891550864738,0.8711799062320905,0.7232660798613696,
 0.33245959206340325,0.5690309066366726,0.5871959547179244,0.9753677081874077,
 0.599490643994635,0.4710971145581312,0.1573588474437091,0.9871776304735497,
 0.1301183748674094,0.20870760612770545,0.6315999145833385,0.202958953015556,
 0.293363499924687,0.21469939040007846,0.7014060457756975,0.16171266942742435,
 0.535481109966686,0.24754171758685006,0.5145499658104671,0.6838996761803698,
 0.4902458348432348,0.8815289918412205,0.6813139465595546,
 0.0031072829460085938,0.6657093556123256,0.330368454000588,
 0.2379575938881915,0.9817928787796099,0.2687265502088665,0.8054544542551089,
 0.005273498339791183,0.679604329213859,0.08449873044432565,
 0.5567964590396752,0.6167148637984115,0.3583841894879247,0.15551888189769258,
 0.6606920220086929,0.3245513749643619,0.10381181665558259,0.8914233457975247,
 0.805586258775951,0.930781130875593,0.9389640441848023,0.6609551258501403,
 0.35515283859485103,0.6888505141822658,0.3034946557112276,
 0.07307351382392557,0.469872318030649,0.9377534407574266,0.2349999587599516,
 0.43396993900713565,0.9946845992666128,0.399907508391498,0.9544416862478905,
 0.35378770496559997,0.37359938199352194,0.6083007065816953,
 0.8839212592643669,0.7642060839747407,0.08357349416949744,
 0.18169747752808107,0.1865993131071516,0.5517161325405862,0.683228404247465,
 0.35262520002958264,0.31771508155890704,0.8256315371531163,
 0.1771614580990174,0.5686469452708409,0.48809817454613214,0.5516997812237869,
 0.6672523168576951,0.9137741825441859,0.3751361941204361,0.9322667619335506,
 0.942226943413851,0.22933933651297345,0.2344262838579031,0.5891569191366801,
 0.44739371211692835,0.012838304120921795,0.6253588055034746,
 0.05295309057797415,0.20317213182203853,0.9577155786529518,
 0.1098028584816012,0.7540618183005396,0.3040340730378551,0.9632182284800418,
 0.2419000130092085,0.3599794448406789,0.8186297730653546,0.7680941024843706,
 0.4565549027822947,0.16640274935864396,0.7321294708179841,
 0.41846829203753233,0.892555936049144,0.13599376542488428,0.8459164162385227,
 0.30004052011299276,0.7934680584242301,0.29456517408503735,
 0.17535249800210773,0.18575242573249717,0.6250751146858774,
 0.06319487456045536,0.05475168472059633,0.5453368932753636,
 0.026817462705947315,0.9730920842338977,0.08833328651097994,
 0.2404532195852217,0.33355438028064865,0.5064094884663342,
 0.21414762703196544,0.5561260031062706,0.9954240146002106,
 0.41296500283814475,0.2916368488586757,0.9888690168391447,0.5642119391833507,
 0.8206805799856397,0.16172964658295308,0.25345882939845965,
 0.8797046286142074,0.7872895985834582,0.1763566553296051,0.8301922262397673,
 0.429040726464776,0.21720303336542934,0.3072725865106166,0.3093682551909299,
 0.8177155872626821,0.540651891205708,0.4419605796454287,0.1793733276956493,
 0.9363699210729436,0.6703064805781852,0.39293673755398695,
 0.43889454397665106,0.31584424833886315,0.05808817039395442,
 0.5508111596334845,0.829717945125792,0.271831261809545,0.7817674812766819,
 0.9037408250169335,0.783262813854787,0.26940683711868607,0.08553961970991508,
 0.04010315018405264,0.3297227922672106,0.9780567577814868,0.7954561628829344,
 0.5011152860058669,0.6224141541005288,0.2906430665380412,0.32048386209050583,
 0.8709239076274905,0.16925784700930824,0.02664422032859315,
 0.08788747795498719,0.3926748917212589,0.5080562695127189,
 0.16724564510939877,0.008327531673519073,0.28795857543336845,
 0.5146412590784245,0.0282078691622506,0.08557384898103393,0.7972077294863397,
 0.8495490162002183,0.7179135274074135,0.3415836181507623,
 0.021193049700107713,0.22561922603830675,0.0843771338347048,
 0.6920970873314702,0.3370682779798144,0.25353694922761716,0.4706099678832599,
 0.0094856671676804,0.14003921006649456,0.5206704098772019,0.4078959553879995,
 0.025350290954960886,0.9906086045462644,0.2152006043736826,
 0.12980481169072322,0.35360592872382324,0.505681371626777,
 0.03892274951421293,0.8676014854616969,0.7515711616279335,0.6543267219367437,
 0.18218597510234125,0.5323334172498304,0.6917210400340741,0.9562938165064052,
 0.1333877153164671,0.0494946540424801,0.8883655763534734,0.3741061285836659,
 0.6878345709131949,0.1487169649846658,0.8968190273643737,0.12540525764367738,
 0.39623181637519267,0.8923956396118755,0.6874281303826023,0.6676201862707835,
 0.7111789999709474,0.11832603621169346,0.3483707756939155,
 0.41251649046046257,0.793311970458659,0.7989890702054386,0.10284915018690977,
 0.17529341953077493,0.24322195052823936,0.00456046268555077,
 0.6613281141006001,0.857686850644636,0.5552074269968312,0.12351118032394226,
 0.7336870305698692,0.47334942000347613,0.2935049438003029,0.9552554650528332,
 0.9745419561272207,0.9710034769141407,0.11682856628568272,0.306529721744513,
 0.1474247979236165,0.2868462894741377,0.3474954311104901,0.6681252114759948,
 0.8575909510936809,0.4242473967674666,0.8599326267328616,
 0.052429683806279326,0.40540957886446194,0.5606579616943924,
 0.9495414046490346,0.2775419334965711,0.3593972401824227,0.7139002961938027,
 0.27274547863764365,0.7811802265656069,0.0937824720360424,0.555296046992962,
 0.94806721584793,0.05361479295426108,0.7709758665415037,0.588105658495808,
 0.015362729875509573,0.14068342703362058,0.7709908858267108,
 0.5319907373261933,0.5497785005035721,0.209736768815874,0.6129181709099647,
 0.21053511786245604,0.18990394197699356,0.4766794129981726,
 0.9104467419316179,0.9593725864554542,0.32278099688422124,0.925276076253087,
 0.7588371582869764,0.377375063651882,0.4585825334644701,0.2826598177637062,
 0.15698863360681214,0.5099006233797091,0.4797441616478174,0.7849707877491503,
 0.8703889216667445,0.29405834904061723,0.5736382619389835,0.6478445349887993,
 0.1966472071991533,0.21087893889501053,0.02494402534799034,0.412997414283902,
 0.6753882529074556,0.07949526864353407,0.3185718683971056,
 0.26961245691989455,0.5966475614453846,0.46292129865685017,
 0.45563660919345506,0.1207934470450559,0.957559605912275,0.5476731045056747,
 0.39391970404994314,0.37053256762734077,0.07910361811749977,
 0.260716371093497,0.947746779480383,0.46130223951213944,0.24249522186663153,
 0.8071616530504568,0.24767148564931374,0.43852145923021235,
 0.7955050497560381,0.5970454802984827,0.421218148211455,0.8282393611262088,
 0.105935262356464,0.7032545009620375,0.4895170679666566,0.7100067514878143,
 0.6136345142577174,0.4817304041461131,0.297889190047376,0.37652167691893357,
 0.9484156238784455,0.9472501498272189,0.19107730323320826,
 0.01109939672478999,0.4686213692345853,0.22096030898218832,
 0.2835442297101175,0.5378106820257584,0.253434237825951,0.6368136780707389,
 0.43989989345371183,0.4079343572812929,0.8425390388580668,
 0.010550200994389636,0.6671605507870071,0.04040448377001049,
 0.7748390228050697,0.2749754017432551,0.10667608743024493,0.9257406512114897,
 0.39365873223166337,0.9471349415451833,0.3754666592619673,0.901215503448454,
 0.09466975966388946,0.06921724965439147,0.005440220402221363,
 0.748743294994507,0.2345587307332604,0.32123285407756696,0.10922116631035372,
 0.2422478237363591,0.7758752903566897,0.4667421240460641,0.6608831569005916,
 0.5986313435572802,0.33817031525532426,0.11679318639518721,
 0.16439962344313164,0.1444152617139799,0.12915822737468885,
 0.5497770736542466,0.09682701643952041,0.6361248399124614,0.8310395606731086,
 0.9936670619294132,0.4253085463383821,0.1461707538980226,0.9767736281003658,
 0.9390276509704639,0.4086042981498107,0.36140004527746505,
 0.24389657601104542,0.2836378436506851,0.9701128282555326,
 0.23465525507257856,0.4409474675696796,0.25107765689030836,
 0.47049519227714187,0.9306134330934422,0.3039627299104455,0.6062569169285144,
 0.4310644540727273,0.44174281680891025,0.4910537288130251,0.628964866171658,
 0.7230627874275004,0.49887504366324253,0.2298473436935743,0.5592648995652069,
 0.7155527814876209,0.03227966076285327,0.09383462694006606,
 0.9302079385014276,0.8832870151614927,0.6142152811722495,0.4559866428953093,
 0.42541981353514446,0.3325056015877652,0.9566314163314898,
 0.08742368987066307,0.879676986973871,0.4149299648597856,0.12222454528442506,
 0.1213182452890923,0.3736782374758476,0.26914715085829133,
 0.21063693645273696,0.11887337641429674,0.5403364796974173,
 0.6401449229911014,0.8905051538218162,0.22966966298793445,
 0.41039984606519087,0.6446720517439232,0.7890970923999105,
 0.18147943162412117,0.9154399364933655,0.8122525143994398,0.6105574655132215,
 0.6559316277166437,0.0911804011674433,0.1759251541142346,0.0622708004471878,
 0.016249768186802127,0.179892548756742,0.01068959958425364,
 0.7006338666895939,0.8105757907841729,0.10957470874981035,0.4278593293437549,
 0.6756166181733139,0.5573537334736074,0.5636073631449326,0.3066962220295527,
 0.7547806501990819,0.6088190716542745,0.24280562318483934,0.878204981083003,
 0.4638864603624532,0.8919503502908382,0.6277919886806704,0.35371526149058974,
 0.15478429374161118,0.4759871135500171,0.9311855756459457,0.8877300288903065,
 0.962048309198622,0.4934245971718053,0.6756932347556119,0.27327724157087374,
 0.6709627321991891,0.38489409288218845,0.5397720771144552,
 0.04788375522739563,0.837320077567832,0.02393298628396623,0.5781289485528348,
 0.668339210956474,0.7676276634925334,0.41517836951490306,0.7122165504073856,
 0.01743507442040726,0.700329980397354,0.8999881709064875,0.5005618191742296,
 0.32946328588952845,0.045160403688336404,0.5678639773793712,
 0.0029292480005403965,0.3662861770033805,0.11854330288891357,
 0.8734727268208698,0.593136495103625,0.585477961852136,0.6030077121209888,
 0.5178249011030371,0.13612494960986055,0.973245487079424,0.5138097961048695,
 0.6380017327283511,0.055678246626433725,0.6824271801734343,
 0.3022626566581974,0.9547680347426355,0.33094619907827727,0.9704000342855679,
 0.40738446250303995,0.8447879936227358,0.42538014686998826,
 0.5613738949217324,0.5633154808456667,0.5986406029966043,0.29514480035927737,
 0.654061744397508,0.0013032830860096878,0.7672760094595551,
 0.18125762121844846,0.30226027282098955,0.41707299116565877,
 0.7702116629246112,0.8531557813019446,0.5376285026197245,0.8253618318529256,
 0.15623000913990281,0.8973083152226242,0.9188879966442017,
 0.22165768175630585,0.8682611359959806,0.39715824249079046,
 0.5763490080417872,0.1365856896352633,0.0866545328921472,0.4807716536603519,
 0.39582259752532745,0.34638427039485153,0.733645937145592,
 0.01948614925793657,0.3195179249672696,0.4476397787469346,0.7445746854793969,
 0.16704148950618292,0.37905925062520873,0.5705837208129239,
 0.7084001525074908,0.800259098698007,0.8498568029099082,0.9191684411713356,
 0.19109843214172928,0.5918428220842201,0.9869684889284881,0.7849145761776635,
 0.2743128776473007,0.23028423153975153,0.4343995008855146,0.9048052827707656,
 0.4619158188823892,0.7337966440864642,0.31307742162715435,0.7076596845540092,
 0.029022673064914795,0.6104886948468603,0.14111761899716324,
 0.7024423026582668,0.11934146692440839,0.3431043352132682,0.6496074706166735,
 0.556903063768857,0.4328708888733419,0.6779289431140502,0.9119822694241986,
 0.47593386842651386,0.5941670632713592,0.23248428732662008,
 0.7770917691344292,0.6907641825417586,0.38782417869657904,0.5518990484210684,
 0.9499042469185409,0.25611402231054403,0.17317339265982556,
 0.3604053536251042,0.5639759599660451,0.6236489919028766,
 0.014037901291629185,0.6418869267889297,0.07634302434379614,
 0.5822077422442387,0.08209873885452745,0.25674426202687683,
 0.06933838620190624,0.3489243696378419,0.08608693830723357,
 0.1973245079832111,0.7390607901981873,0.6400683757281223,0.14520465961447426,
 0.7160399151421862,0.40391897749926886,0.0050272966081139625,
 0.8660866926516657,0.2514837786232682,0.11435986373910556,0.9891633257632331,
 0.6357104406263544,0.25753250194488264,0.4907561523046443,
 0.033156629700349494,0.09901006944745105,0.9779212610883586,
 0.43787043448879825,0.8065750960328734,0.9262313560588039,
 0.43394740928418085,0.31906037459389336,0.5191170520330805,
 0.5572268403267715,0.7185402922604205,0.701928585529508,0.8025958016076499,
 0.8041591790736898,0.2603765073400762,0.445515941575257,0.9065813238479734,
 0.781227289679644,0.7797006436064416,0.557397475270914,0.5957312247995868,
 0.4552490464813014,0.5486582490785478,0.4641877717308237,0.6879857409808985,
 0.8310219486890092,0.3953479623350966,0.32950822999785456,0.4137262502703569,
 0.8018421483656792,0.9303993808319552,0.5383082754710038,0.45892756830399717,
 0.9252378709939484,0.22507857552747534,0.2960854515143979,
 0.43528701737905107,0.9637025934926219,0.18266787993948852,
 0.33806938346701365,0.5226979448282955,0.7079134378146041,0.1412412555598621,
 0.26187817805586655,0.9029410708813128,0.6924315905406112,
 0.09845046257245327,0.7062097400291896,0.008502210169782698,
 0.6409626173021178,0.9458610101828633,0.07741960290990468,0.5954941696611817,
 0.9096659816410486,0.35055870892514196,0.2678027922081597,0.2634972581850208,
 0.8282137334666329,0.5839464763292932,0.022717570080720062,
 0.6217480479584473,0.4046976699924607,0.8164707488305116,0.2728319870066721,
 0.4792500968821941,0.9519986244346896,0.28101514125396654,0.8779526891349928,
 0.9857169541682393,0.6987430114265195,0.5595500022690019,0.4762840658928812,
 0.6450389232086615,0.6014313526661667,0.7401670197744951,0.9367487688056967,
 0.1986998274388413,0.4096999367541202,0.8597700075800438,0.363497892363874,
 0.5634605629925373,0.08754417300969985,0.5937248717028314,
 0.12010577898968655,0.4906687121165494,0.03471337933238583,
 0.7429464547409108,0.1782241343976385,0.7006478164017629,0.7877240612940583,
 0.7148122230415827,0.7600522958359328,0.9819083767104504,0.9603612686064195,
 0.7572479599771965,0.011693946484663598,0.41071738619753995,
 0.9199465933027865,0.435794716065952,0.33585862179452886,0.15457114106595782,
 0.8704776522304344,0.27617149779971406,0.8233603752668168,0.4123239814042824,
 0.048810181546914855,0.7076913554739043,0.12081159768566452,
 0.29995005183980217,0.48693035763648396,0.5616323298301253,
 0.04046236853753349,0.5561426942926326,0.3452011169315261,0.6096102513683308,
 0.6796176415874838,0.2070652980174844,0.5790027658990147,0.8465019566818688,
 0.028768467400493325,0.05851505094909348,0.158227888065247,0.941981448659883,
 0.849503562329275,0.07248612077351424,0.9713824102596745,0.582274520277254,
 0.01988901062662396,0.27363896526061227,0.1615688171811327,
 0.11188271147041962,0.491031173724488,0.6744995091164412,0.7214998669855863,
 0.6811687574905845,0.28111944542567346,0.40698259966420414,
 0.8080926156548223,0.11676958549936067,0.049383950332251625,
 0.09196031471555433,0.27302760831662043,0.6579719258599771,
 0.47372795889163266,0.3598055687312901,0.46076759956100055,
 0.5796464012388853,0.7237072461815836,0.8241506708132771,0.14482012737403271,
 0.055129913029626554,0.34892469572822316,0.6082492600687583,
 0.2655225349776392,0.9411779501422752,0.8843906180051162,0.11083442644371155,
 0.6330888266246435,0.9106074831804546,0.46455712332373467,0.4232506847429057,
 0.4238398074156833,0.2799668214001245,0.9648895206581144,0.332984872918644,
 0.6486115629857734,0.7213299075381668,0.0037410950455559755,
 0.6535539084363607,0.6214292480542374,0.1188029725124341,0.6543906937973323,
 0.1373739805363654,0.518826125281318,0.24776198523592163,0.29451506516908876,
 0.37503237519095267,0.9795974171460611,0.6906732371161238,0.1805981259080749,
 0.4436928003612324,0.7342453241575625,0.03264692401190172,
 0.21464169462886262,0.6184821382536454,0.5217734278846576,0.5600787161649723,
 0.8824415873462929,0.8729481865542682,0.6579707597262551,0.29638774198911255,
 0.9546598928513025,0.26743466354625134,0.42474910785749964,
 0.21245788385480258,0.6981371830234878,0.3918160738178086,0.552958911795703,
 0.8781731422908462,0.10362912114741174,0.5951068609916127,0.8525932227321468,
 0.8129944142314474,0.4751055206563345,0.8446249328662807,0.1420864527391008,
 0.575773886882066,0.6477113403623271,0.6403498045232763,0.4482584171531754,
 0.7054690334948975,0.3693951828720877,0.9304770418374192,0.47087694913369116,
 0.7245759310398114,0.227032310795941,0.7416490388349598,0.5015270311725203,
 0.12803479621671865,0.47090044269455,0.4990579517316196,0.9850705036945198,
 0.2759791562181053,0.20291738767685286,0.8415190284727323,0.5141015044589053,
 0.9805925660393774,0.09118414463454472,0.4094454095970218,0.1516608897935079,
 0.9080321785500044,0.20559450994210382,0.09474992270533367,
 0.32559939595752185,0.5485282438494896,0.891351539304923,0.7269546102175599,
 0.24939425325179498,0.32681657812296594,0.40158530392866565,
 0.44150167858954537,0.4858273295889092,0.671311095707779,0.3201648734412592,
 0.7726841337775701,0.024092060610503174,0.32831740773437734,
 0.33393858009380784,0.502534005692099,0.23445951198972192,0.1288525076343775,
 0.5074705711343299,0.021904307138492563,0.8028276544921513,
 0.4859961500302774,0.48049524959219647,0.151200452064158,0.23377522847126253,
 0.5684364380795146,0.997869605327476,0.8811281617378389,0.06969542869652645,
 0.28223493666833266,0.36695619245481215,0.25045800258771433,
 0.46826196395317554,0.06195559893526437,0.987157459603535,0.5785848705686989,
 0.8731894275067488,0.7324034906795351,0.41095415724198014,0.8334397584559217,
 0.702386362881938,0.9560391458156574,0.778647043340581,0.8035819893022619,
 0.5370320340471468,0.5907255638461432,0.19520035767865662,
 0.038188333147707754,0.510583897423182,0.5671961293251422,0.0799795146519715,
 0.5817155589356466,0.11683858734415198,0.8430502245004745,
 0.24998063340264298,0.016510579383394286,0.03601507631820733,
 0.9895820154247289,0.1799020666849217,0.2502815185922458,0.09037382958399665,
 0.673498359541605,0.6513713151103304,0.24367022592205623,0.9198940975091158,
 0.31457031149947745,0.27419862628231095,0.8846966011697653,
 0.6863055468661307,0.7335047294206807,0.7149941277185357,0.2633879843240542,
 0.6486670912714116,0.19464664904499474,0.9505794370925514,
 0.22045879194412388,0.032602413273843056,0.46827047507657393,
 0.19295981216435565,0.5708768578598202,0.23048412767647242,
 0.6187727680321259,0.8593643451111633,0.31670550545989906,0.5372769048014473,
 0.5688177421319474,0.5234167238623054,0.9404371962436371,0.9290434597912425,
 0.31739198574013017,0.6342505116076281,0.10734087005278758,
 0.2280803286193548,0.5923405905875768,0.6348261109380628,0.47789207693247426,
 0.09789917297469519,0.9786652833850372,0.9543173596969234,0.5537429458977262,
 0.34261164533168964,0.587791757178718,0.9447988634904698,0.7758579449702188,
 0.9706647855670589,0.6111634897875842,0.43201621253738787,0.704193728420717,
 0.40719395321987994,0.5265301282696053,0.9212761033321549,0.6611144351251155,
 0.1669236827313728,0.24364993503046206,0.7760749226537582,0.8366670738967432,
 0.2721040434257713,0.345014402990826,0.9219129154493622,0.08905513791808617,
 0.7526809572735587,0.39379349593856783,0.80872754426305,0.2153417012297527,
 0.7252262711276398,0.5907841303285688,0.14416019843035888,
 0.21696807690041053,0.34355582856881883,0.2528328128341728,
 0.6095290034460679,0.5785164257629801,0.67386194970179,0.19319151411555657,
 0.7675282215277692,0.09290793074698678,0.5304047076275311,0.7181744569879555,
 0.2976157236862689,0.5111260012037101,0.1483376481054114,0.05562858762244394,
 0.09176407571487033,0.9537345544617861,0.8154072735116913,0.8141793963078161,
 0.23643266754518566,0.4418691286825459,0.04177468732010681,
 0.5906767382530548,0.6257156421683083,0.7637424634028013,0.4163649978017169,
 0.1593614996459587,0.3439957707835939,0.8226752152422406,0.5704860218839681,
 0.517435080804054,0.4524520490000641,0.016157207247035066,0.5684778899620408,
 0.6907309551555133,0.819088101918884,0.011891992705216237,0.5861788847016338,
 0.09292625073078575,0.021635397659742983,0.8068299906368196,
 0.5726240576768942,0.02339836943983753,0.7309916837398922,0.0756953079812861,
 0.5430762252838386,0.22197504972544924,0.7349783993047048,0.9143296273577001,
 0.6804688160337413,0.8221409604957056,0.27591948031732405,
 0.45511986203410726,0.23012141036842804,0.9426936309113433,
 0.5679389898954622,0.7690803408803875,0.46729054463193476,0.882776777855772,
 0.863639777492782,0.5650443757647235,0.7096400126381044,0.7291545449278354,
 0.27855993187714334,0.9335331319532103,0.23338512646090992,
 0.4787582847601902,0.6664675988788497,0.8749332747586457,0.39915006782083695,
 0.53187129124984,0.3713636999058886,0.46396988763966196,0.21993381935656553,
 0.9904528918600427,0.20501449058406374,0.27212000782977075,
 0.8595401687278889,0.2925825517103684,0.9629844897167177,0.30007105164136794,
 0.7408473001276097,0.6261530102762238,0.3478663189120389,0.4614658026566226,
 0.40836047719882185,0.21766505958341853,0.7349181610636774,
 0.2320230505286548,0.35762936278544055,0.9764332793885981,
 0.10955718028819161,0.07537169552943224,0.5660916857496562,
 0.4683914530430232,0.3288391001493033,0.8423009919642832,0.5380351809187521,
 0.3045115644600339,0.12192769672297699,0.777663572757314,0.3701055351939422,
 0.9923637272631439,0.41675091678336995,0.7982369970990435,0.2221765626039125,
 0.7355161759337834,0.7668297555664199,0.3101630313871939,
 0.029691193410289363,0.5563531337489505,0.8864779095654266,
 0.5783801534367272,0.1681824608776682,0.4573551575839918,0.12754589944881922,
 0.929396369026975,0.48235686322361737,0.34917732561254633,0.297549810328118,
 0.7528846769211128,0.8430178188081084,0.11022511122264511,0.283399842907182,
 0.18644665222819778,0.8325890041242576,0.9311235775558737,0.8683709729175794,
 0.8720992365507619,0.4891659271337714,0.11402465795517536,0.5230357194883816,
 0.5623236903050921,0.7602737318676118,0.1364503902958849,0.9573929738244819,
 0.9723644772364807,0.6517927981936781,0.5888012920154488,0.4897597977419862,
 0.7463730249320467,0.8400277026245713,0.23499461902182306,0.7094729449132029,
 0.37259281222458773,0.04945506083245188,0.055960102769016196,
 0.39538484511335215,0.4595973816045372,0.6567859270845329,0.2889124468652139,
 0.44173946227420446,0.5493525939249447,0.7749159590307005,
 0.14111280443092145,0.4700785068707506,0.6083755374106293,0.9652232332382618,
 0.055761460490665216,0.4533512169837366,0.8757396017993428,
 0.6980870788048883,0.34856211545286875,0.6051279325799803,
 0.05143392476257169,0.2153291438039009,0.6642627257423727,0.7372105431917999,
 0.015751400934200133,0.5729886864392739,0.37075606502394387,
 0.8108340961553041,0.4124624827013048,0.15973892820764868,0.0653628302363809,
 0.37048584914077587,0.1802697774060219,0.792964035984889,0.8024453441490574,
 0.5255625604204716,0.5134159172924616,0.03037368049526501,
 0.04581770840637389,0.8062525938856274,0.9425919053601248,0.8679840334406357,
 0.4437768629752723,0.16770436032222136,0.872165715840853,0.8136695469357762,
 0.1280294025318396,0.7489132110000762,0.6357169201791042,0.6433986869908441,
 0.43618361026222463,0.6650529627425422,0.04325524852377827,
 0.1690955417401825,0.19568719695899583,0.08395161032271303,
 0.2921985213334807,0.7904654762238871,0.11651705434133375,0.701123497512702,
 0.20871766493638877,0.26745455762077375,0.6734893833361665,
 0.31857224420701236,0.8999469051568634,0.2246194628996906,0.4323562145427219,
 0.7357216191903255,0.901558318641188,0.6411508999124395,0.09513975859290857,
 0.8434395865544602,0.07080985987741295,0.5402408375818327,0.4907587837577525,
 0.5866328077482845,0.8685192098171085,0.29529625916067326,0.6484135142844407,
 0.7893059573452877,0.1979147533779414,0.7527872490159822,0.5157513893142099,
 0.9492418489308383,0.043996241890571275,0.801386190522297,0.3367295158012136,
 0.8054169457168129,0.2682983315867601,0.9967415921734631,0.3165887300414174,
 0.8810459266706498,0.6128370784746909,0.09491830949134261,0.4009265484871237,
 0.14590632767317446,0.9852815626856727,0.7882329865684415,0.8941675098196769,
 0.7786053047387531,0.6640917549179874,0.7860284630432351,0.35811795816329195,
 0.6588922202688993,0.1568704044207898,0.8053483763258616,0.4188520810852828,
 0.3764147531170541,0.32231956133050643,0.8384071462793978,0.6922263131000577,
 0.963612361559914,0.579931084158573,0.3786878806074696,0.2242340602746501,
 0.013450905216545905,0.6554507853182973,0.8205117731119936,
 0.3834543690943484,0.5803582734546188,0.07987005381136503,0.7100956712584704,
 0.6689914233020205,0.5624332316783459,0.13404872112008515,0.7404267663666896,
 0.15464222301116504,0.028439172002439284,0.9326288016254106,
 0.05942829025717866,0.1301212266455224,0.5910105645022439,
 0.19782522676145287,0.8194204052791966,0.8682667478831098,0.7135348774268759,
 0.09610346789516144,0.014402227267201861,0.7827409712274586,
 0.2803302489633467,0.5938165686965822,0.17391235722344445,0.5603534448616228,
 0.3637068586509259,0.47977342485350816,0.20255620480217384,
 0.4264875441085837,0.6846756380990612,0.2210201670582247,0.9296981377511747,
 0.4523948895077601,0.7891838856696346,0.36423962220311246,0.8161598634363929,
 0.17395681172166355,0.24424785127693371,0.6380555360212243,
 0.2858720066321009,0.39969644895018086,0.6507121177226295,0.9030456761674129,
 0.9272389081684295,0.913419444423517,0.7715229271335182,0.32193335924146627,
 0.4563743854835033,0.4655875572171533,0.8849434579287403,0.7636283915830092,
 0.49161158035842856,0.21495949256743874,0.3271724279356666,
 0.18454416091637305,0.9503084740015232,0.3910101164508508,0.3831788474580212,
 0.4433626239175822,0.5656316531316656,0.8515328224685226,0.8327586134750601,
 0.7953556366036326,0.12672443030484637,0.46491370488720796,
 0.3799188394064402,0.8898299607867336,0.9643149082804624,0.3394563134205104,
 0.3025583709289451,0.25189598118335155,0.8137735364898014,0.5745916920162387,
 0.3779779276958729,0.2349195667098788,0.9552289487777019,0.22514065554738683,
 0.8993832122220046,0.5165917566069381,0.21762235127747376,0.5731538687586206,
 0.06446031931870189,0.7950595033212235,0.8076241509410145,0.8163138171700147,
 0.5638354516417099,0.7765701741280915,0.3995958220627438,0.9352818765883258,
 0.14754404885579553,0.7755482353565436,0.47146616387555607,
 0.7675132348800787,0.10786206527026743,0.09910563640772008,
 0.5497658883521588,0.9246951536895817,0.6341486845136107,0.22698562495668106,
 0.36312088653495467,0.9121800763416642,0.07331581182907265,
 0.2844194854677319,0.5599245824207362,0.40170949349266405,0.2587211746371034,
 0.1511441380398223,0.3264930764281063,0.20847854209785277,0.943247038407516,
 0.45161164537671905,0.7269671836206666,0.23508980702210147,0.413763454812352,
 0.5264643343497507,0.9231869470806221,0.9177109124365457,
 0.0045196702246628195,0.7430518199746844,0.9149938510845212,
 0.03609399806547109,0.6075928430815069,0.3591211706156284,
 0.22492357362102566,0.3205407612437715,0.7353048324798227,0.7548423855748613,
 0.21217281023392676,0.3469734388461886,0.4550944363599426,
 0.09981893375415474,0.8053760400143104,0.3801538372320583,0.9814889709496344,
 0.9372847741225996,0.5856064907376715,0.7468591942025559,0.7055776614951744,
 0.976765179619972,0.18735302823951816,0.1414215541139099,0.23886570933685825,
 0.5794751162469475,0.43057075173404236,0.25488429759191056,
 0.3529399012147436,0.8498602218316444,0.9472572638423462,0.32670854369391666,
 0.6852716521037082,0.17601139056999549,0.8708387824707939,0.6834573659777256,
 0.4861592820499857,0.6849685322964784,0.2850901692913901,0.3077573922088427,
 0.9518770703131888,0.6450412275091323,0.03798585256443088,0.3772569866176394,
 0.03219235625452055,0.04788485677306609,0.6205518541619273,
 0.5658667626013297,0.1334085253955687,0.1629472886810034,0.349324032531657,
 0.16738388048839126,0.642235791925085,0.45819026907932414,0.6404641190953281,
 0.1503873998117915,0.13895889080081825,0.34016712141391336,
 0.6758087577591364,0.5267592548035808,0.7110602615360226,0.811277001198348,
 0.26515450738880386,0.7152286830720284,0.7516065425946912,
 0.43019351558114105,0.8905150902878561,0.6881691200436609,0.9291343758969589,
 0.7754289937939026,0.10997705151199666,0.5023206489466594,
 0.10992299896356295,0.5312446459812465,0.5714523708571301,0.3515097802970839,
 0.467920702169498,0.3630532072824101,0.717781163966457,0.22734848315014833,
 0.16768820628352088,0.7519139109916642,0.6410914600380238,
 0.47334764892891845,0.7345146157845714,0.3163243231988505,
 0.0065696948134505995,0.4561210639183644,0.5286090458660178,
 0.6842373519962668,0.9920769113980883,0.9144309073720187,0.6190242425483924,
 0.7099991893227608,0.31350407288500237,0.1967702096120555,0.6229990393248885,
 0.9821939986604326,0.2221974578224718,0.8613606677522313,0.13890047166009578,
 0.5101178545835154,0.33173270828896106,0.34294017972409774,
 0.20321862992111728,0.589168536398294,0.27973809705292163,
 0.19221806392261087,0.39594209607388064,0.2453672259249251,
 0.8699442576963532,0.32010628808968544,0.9649995185127606,
 0.27102374291011955,0.5089237689199505,0.18216692995144068,
 0.16705123764912233,0.3781241919002817,0.891939332307934,0.4068688586926478,
 0.2786626915808331,0.35489389785965186,0.06325649110858478,
 0.5210019347481903,0.7744597336004124,0.991051108683552,0.05964442093101652,
 0.6469665450538711,0.2916533031010362,0.23193178551048121,
 0.15113955479938512,0.02277043977638793,0.2520397309214345,
 0.9532279306953234,0.8143730162402157,0.10330051241478155,
 0.21543294090756626,0.4221160969558426,0.8033214247801242,0.3958243047442773,
 0.21546799612729084,0.9944967916342468,0.4506644566558724,0.772601388683338,
 0.45950555047902686,0.7782638282831266,0.5365724365211546,0.7010594926180661,
 0.14592054533043397,0.12936283859550934,0.011169898929092148,
 0.5262992685337196,0.6836108367797437,0.7902646908935684,0.927387299266349,
 0.43006867955338857,0.13159296315142077,0.669984226403272,0.8668041186852841,
 0.3148274099014422,0.610829457768576,0.32243820970645076,0.42745549859762755,
 0.30359564081958945,0.2462389826477881,0.7548015470679172,0.8893707229430782,
 0.9741902698812835,0.06463520011036505,0.12277907212262751,
 0.9699629277208982,0.19339226612818505,0.09376654876862645,
 0.05049823167607581,0.30107622584257543,0.9940660456694564,
 0.24767410585894523,0.9407625275057979,0.5224885537539775,
 0.48868551573933827,0.4450548154434253,0.27038441252089984,
 0.8853479084815273,0.5470920050498099,0.8600273566524737,0.3301818707231782,
 0.47736358863055695,0.38767031045788314,0.35182964837263375,
 0.8168240821155431,0.02509721053098668,0.8510030824475056,0.880716494204861,
 0.4504171779864221,0.2920220479482537,0.4638076360086907,0.8061406625136625,
 0.6050416803168408,0.06168166999865787,0.48223756519806915,
 0.02050418793611919,0.33348349848501313,0.9228396483520624,
 0.7847134726872899,0.8436912404196945,0.3071695423389187,0.9347331850303613,
 0.4161531842606001,0.1205079687201498,0.35526025239154513,0.1079185289751694,
 0.3996556868681733,0.9513063000103287,0.3807476753500999,0.6266719993474273,
 0.6250807020013606,0.06411985634797834,0.039332306053172506,
 0.42635924443594164,0.19858806291963882,0.3239867412020634,
 0.8913786940105347,0.04212782642294166,0.8293634329372451,
 0.060375338609321205,0.20060296066214178,0.34159357163749504,
 0.5133236990280392,0.8349635085903455,0.2299176949134859,0.7856846880404509,
 0.3596927075486601,0.555081305838098,0.9439699291006023,0.5912056265294746,
 0.1350392803762993,0.047784967364277064,0.878011537037965,
 0.19860493894725983,0.4400659559955573,0.3054150129130152,0.7969816837512649,
 0.6571671874766636,0.4989934761655751,0.4495423780601979,0.5514128016736125,
 0.9672696808746284,0.10790008231574066,0.0489119542189016,0.9864620838703823,
 0.9141050426779104,0.7700382365218544,0.6263900154981634,0.32309769220127893,
 0.6291989998690637,0.9498363311089737,0.5486831907913466,0.37882670924858164,
 0.06020228026670704,0.9834137525676121,0.7166935530902869,0.7848736805623442,
 0.8716326979019577,0.7488574304365032,0.8824900920698712,0.19602722885025003,
 0.3506312779279721,0.6947997851972483,0.3693472625245918,0.8329474515991219,
 0.3593441138885073,0.9027847945325523,0.7858780266969603,0.7233867061116979,
 0.6271067601983695,0.06832937411256657,0.7377353817406603,0.5787739755958347,
 0.05542553606779399,0.37139595402862846,0.9909570758777896,
 0.9912925005289928,0.0961104499904396,0.3834279389083698,0.5684627476851548,
 0.8738619119273376,0.1912288894373051,0.39415068623492355,
 0.36990678182655023,0.2057515890464906,0.7383105229572354,
 0.21885081550621222,0.7437924695507407,0.5412379519524059,
 0.39128784808785777,0.7202087210173392,0.8032368194527858,
 0.07581373954690063,0.07301409277368875,0.7160708991607405,
 0.7666808465753351,0.029423765820024372,0.21001759585828672,
 0.22000230951559774,0.1892959046705307,0.9008987560277497,0.6777043348571925,
 0.7118660931340706,0.3533013734319568,0.6415982628604706,0.7398610562362018,
 0.8040302301776074,0.22162530261238456,0.33980879255597396,
 0.8403794300530062,0.0262835272366424,0.7530131298409344,0.6622992934293388,
 0.5858294110389434,0.9521089270278075,0.07350221306419469,
 0.28239304399839793,0.11149586607829953,0.3188608901384494,
 0.4787388145643058,0.7026014572187795,0.8023886214157245,0.5060493835069172,
 0.5323401558997499,0.709766089280006,0.275404474324382,0.18802413939025253,
 0.08810993563815361,0.6912679787558295,0.29053191289992,0.8615984550050777,
 0.2556369281891533,0.06233822476304307,0.9525473436049451,0.5826105009346676,
 0.7604754905574644,0.300681720630258,0.5598634084515588,0.885312655513648,
 0.3596912741898195,0.621080388802781,0.7065648385547761,0.11359700958899166,
 0.32333236818028044,0.5364995913409053,0.9236634851343565,0.9004678913678571,
 0.4104937617773853,0.837957271814243,0.6844954917717184,0.5044272757215833,
 0.7708464005871829,0.09640391023637851,0.5467812873331911,0.3908313796452696,
 0.9924250225495,0.9489109698558087,0.8126408877143424,0.954959373850465,
 0.6113764588955463,0.8155360326866959,0.6949305132819817,0.45669243873797305,
 0.3504616023342835,0.4418791197969709,0.08518410316710401,0.4228610984066964,
 0.1691824182303433,0.2982528674344792,0.9091141775935505,0.6204786192365398,
 0.0696209668178962,0.6077856137195284,0.16318767263446976,
 0.42539832117068865,0.7805682268378407,0.14008432305390095,0.158762302349422,
 0.16714090613569266,0.106143984451601,0.1913764910584107,0.11454629340187406,
 0.3244947257200399,0.47120409140238995,0.25095185516714047,
 0.6582059409438072,0.7666748088645474,0.6848862355070945,0.5197658383322935,
 0.8109109020777028,0.9327965710619321,0.782346724471556,0.34555211683000864,
 0.9318169634448688,0.7877163084890825,0.7805074267967154,0.364074905081584,
 0.05270974628941838,0.05570937112520058,0.23365369236305777,
 0.16471864476028386,0.23736695971073662,0.2709176521467328,
 0.8876231985638073,0.29751936187332606,0.6564426118737967,
 0.28437091768352996,0.050390689987408144,0.8424381639907285,
 0.9625773012397019,0.253293759332197,0.8502568098510007,0.46867020282207406,
 0.36953072044011237,0.28335598068075396,0.009130979813438023,
 0.6030745935688829,0.22981356821651078,0.7200954441800946,0.9621279995440232,
 0.7335095759686815,0.7966875731852615,0.518738533219839,0.8226586111407582,
 0.6658496304922653,0.8113968433716177,0.5546391833946397,0.6685120504505903,
 0.8602694830033243,0.7267944321265285,0.32407076174218075,0.801638834631061,
 0.906847330458074,0.9461947590118344,0.14760039993577412,0.7327679297328561,
 0.8130911219353829,0.8051152487332603,0.7779680037067224,0.862887586224876,
 0.10696560957474666,0.12675429670493166,0.24185961536764145,
 0.3172716308674025,0.16070429789977336,0.8462581758732666,
 0.28519465091567875,0.6062375030549825,0.11636202853362654,
 0.47663076115784475,0.45029334564253043,0.251768566296654,
 0.18284367424437398,0.8906685803143619,0.4174469922856532,0.2170727086898785,
 0.12261859076929049,0.631169125135079,0.8519365079545708,0.10663417130113007,
 0.2849125367170813,0.957401549873477,0.36529796946677673,0.04149625139417468,
 0.11816581376900859,0.11039006585967359,0.6693664798882379,
 0.4972932330115363,0.37231323972609287,0.8116499006715769,
 0.07151925385406854,0.46959031433596155,0.6171793996371814,
 0.6251744267447219,0.2816606429671765,0.8784643567431893,0.11521881721022009,
 0.6172535810346512,0.7273687095246066,0.8125111555026052,0.41875479925631576,
 0.37606456579666725,0.14394022284399133,0.23387304262546682,
 0.8254109409476551,0.8076772910442812,0.4029400644653469,0.7041544437071902,
 0.3439741078885131,0.4084882251346469,0.07440730191526079,
 0.29462364177553724,0.7998444100795621,0.6579933627322486,0.3303380467584297,
 0.016046350409549937,0.4308091081852534,0.9362828759175634,
 0.05908676128313717,0.3676228541037183,0.35184424761470234,
 0.9542100228750077,0.1749680357641259,0.17842601208462838,0.7436409467449878,
 0.06249562352294302,0.7731120586694193,0.4760685593557662,0.2116752741994169,
 0.07928742882658912,0.6035005926874326,0.2143188536806645,0.8708083144807723,
 0.10867190301858998,0.27713123787186666,0.214444252752519,0.7393532933543676,
 0.23425017550134142,0.17137040762204847,0.18500940886550465,
 0.057829785525028266,0.1153405964637062,0.3487549725985515,
 0.4335290308576153,0.5947345084521991,0.44553176646918147,0.8743554964796659,
 0.1419738881570335,0.2460322801857575,0.8602180794980958,0.8978469620281685,
 0.23002962759977885,0.9929196456631573,0.614059621216239,0.5245785679354108,
 0.4765154260820632,0.9740910158876436,0.649334710299659,0.8770137090461732,
 0.41363431558099206,0.01831635640005691,0.9986872099689896,
 0.08564879015079807,0.3701670580302584,0.18686163622610263,
 0.8525567605819336,0.24107644240702997,0.547462215786953,0.25651448780539465,
 0.6487618746421632,0.49010795429002885,0.34671164838171675,
 0.43258749881057124,0.7606054631155613,0.590167112527849,0.8195459460830974,
 0.022626706575933353,0.9162676693409793,0.04601424104835772,
 0.7019770703422427,0.32365506096409435,0.9996499913129132,0.1054515669834003,
 0.13880056603504232,0.4308763659719378,0.2486129427748598,0.7719371224034371,
 0.5254078010300598,0.502866317627628,0.7942937960647083,0.7138241353313978,
 0.8094924880923416,0.5742734671738303,0.5604530274244583,0.5029139015027969,
 0.09769269690864502,0.5844717217101735,0.7436474251818354,0.7939067945568129,
 0.2799823853149853,0.6181862593263416,0.8132377187867855,0.3228024644336601,
 0.527230932103063,0.6623373594791754,0.9199871346352776,0.7007611480280255,
 0.3478105106117676,0.08058327472061544,0.16186851645145672,
 0.1937544007445835,0.2778433128905593,0.32452219706007646,0.629467191609855,
 0.28254545911206397,0.07644840021526067,0.7143463405051999,
 0.5271695540331925,0.34774204310935,0.9863838414019381,0.03763272598710532,
 0.3034777097520285,0.32183843369710263,0.47495861556849905,
 0.5051289981173978,0.866186262802662,0.4126906729927592,0.5870772910885368,
 0.1540832315183951,0.44815962261295106,0.9877360773388008,0.8927943001866041,
 0.5379254670393487,0.7662424984216861,0.42739190404575134,
 0.005579764901503359,0.09222837789969984,0.1045518409990761,
 0.05707889343837502,0.1313412904208462,0.21061187904441336,
 0.6533923378862916,0.6512075278505415,0.17010188108573654,
 0.25075269141732026,0.23884332542093933,0.24427097271583087,
 0.5354128892548039,0.41652431673850454,0.5687488572878036,
 0.018597454759367915,0.6628055878080913,0.5487674811352383,
 0.05823521570387702,0.8766302274422803,0.06526124104881381,
 0.6469079115559042,0.21053468287313204,0.04586553798948789,
 0.28961971171789846,0.7407103957548169,0.7974620688362075,
 0.16392437403350324,0.8208532100854858,0.5253602964439068,0.2288159449515722,
 0.07809600372091985,0.4778644967779956,0.0913781311288232,0.620415152942633,
 0.624928774745617,0.8788652530488246,0.9120174133280896,0.9409962099281656,
 0.3373791121136147,0.9117715525075455,0.3855461756865082,0.6336978648483893,
 0.5154124178014192,0.5550188540533642,0.87337695827761,0.7799785108153754,
 0.1934319039582132,0.9885338066265872,0.14959466644111796,0.2247382294055411,
 0.9289906367177215,0.981520823409238,0.7786389122349314,0.5514647198782053,
 0.46396533055921196,0.22231162396824788,0.8833357920123319,
 0.38668223880088115,0.646342740285198,0.38191563076573687,
 0.06119942043714244,0.4846951988385303,0.18257962915931614,
 0.46962097650676227,0.16386851943982772,0.05330106121406342,
 0.6377369743908317,0.797051475411206,0.17802348321674188,0.1334730867951297,
 0.9297896277396123,0.723859296865399,0.14444736273720227,0.2662619276679805,
 0.38389798133306985,0.9295762573229194,0.0207031994337038,0.2908839170944816,
 0.7236288465376233,0.5095138582371976,0.19867894769492167,0.4146222232992388,
 0.5243563020080727,0.819386831838278,0.5542870189788859,0.48538600895170825,
 0.726914094759896,0.5691411608220872,0.3236865965038349,0.4545872924635438,
 0.3361497841615112,0.5516070969412131,0.5562176924019197,0.9002406862878047,
 0.7931522559740489,0.01259980122260973,0.6266432537586808,
 0.13571687074371153,0.7949103811210353,0.05859713867925054,
 0.9481099531148902,0.37497229737005155,0.26111159326770783,
 0.8516545624557134,0.6684819730146279,0.36199886418248006,
 0.06370158131288972,0.3690857006385342,0.38495503624736127,
 0.3492972899785969,0.037236188862799,0.5570880087243091,0.3727631006710115,
 0.9384445114403133,0.5594872772240976,0.4972642181513911,0.7256233940058919,
 0.450742830653694,0.8050148188079769,0.16575342292965467,0.329564488342847,
 0.8587631338876792,0.15946183948903692,0.13980884560139584,
 0.7071942846966927,0.1284868449650124,0.8294576221254579,0.9656551534870783,
 0.7359420915580145,0.4194907798809596,0.8133937367384523,0.3938170156782833,
 0.026547402577390633,0.8840737154245069,0.13053926411623784,
 0.26613128530991337,0.9061059100230862,0.7232773561611562,0.9594887995949853,
 0.7816075181218386,0.385654151713703,0.5116022182818737,0.2317103704224437,
 0.8910945158694429,0.20746106713483792,0.09245052920871277,
 0.7577691204971356,0.9281834819083483,0.6646944731542437,0.12211323823277831,
 0.08261546463838798,0.3150903266328323,0.42510159111692525,
 0.6822588665815945,0.18803996790849065,0.1864399248289439,0.5887885399828523,
 0.8080000659869956,0.3042698126642793,0.856806516193835,0.4405937157255032,
 0.022784800559219387,0.7520892889877115,0.20140285119522028,
 0.6834876592179709,0.8433288998661488,0.43195637044997415,0.8956565462219048,
 0.5751921383867724,0.8086031468684367,0.6597191322544494,0.34163489829012705,
 0.8571139028403759,0.9366310773261391,0.6617897949476952,0.08672327119296241,
 0.17166283273125094,0.6976174918060711,0.7951870758625396,
 0.18726203644392914,0.6369009831632593,0.5458567931237936,0.9287031977974797,
 0.7192375292077118,0.23924946195059704,0.4394623814066615,
 0.24209013471966223,0.3946150900987838,0.10269288294194312,
 0.8442223026954226,0.9616901526535759,0.6851496681603543,0.9332898014344807,
 0.8519510909147745,0.007520175388629702,0.9192692190536798,
 0.43507054785793786,0.7320587447612921,0.854948012354558,0.10681233507230259,
 0.6812795980844406,0.43503673980773994,0.5844961004620945,
 0.15507071384851956,0.59786939412632,0.35970070127317666,0.46790413606167214,
 0.5948495448221591,0.4078561386248293,0.4824207764273105,0.3198576988592978,
 0.19146930100432547,0.4149427179256827,0.05978101066308361,
 0.6805432368327948,0.4643207019256175,0.4201956063935184,0.38329291345103955,
 0.1029864070305927,0.5249668007803465,0.46425056649356056,0.611589049480889,
 0.5751213356824937,0.3318663953320984,0.845461428598125,0.1138534254654473,
 0.47411839188507465,0.22556092401339145,0.4466053142496734,
 0.10706497630687828,0.4624070647635373,0.020170887523416114,
 0.3118783811685535,0.2058228559917874,0.89585470717224,0.3096567778559699,
 0.7061108619153583,0.4119552058138616,0.7916002355769597,0.8011247488416253,
 0.8847352388956017,0.19942556875449136,0.9962306762274171,
 0.21442494318131844,0.8710272263747221,0.505309557737414,0.11265761551081355,
 0.6665189726567026,0.9922146845136407,0.517519792741021,0.2953168782537653,
 0.3162878093029886,0.1623819386057791,0.18899657766044553,0.4365032540404029,
 0.7491505579103991,0.2709907769175176,0.53167233635857,0.18604331798276197,
 0.5118342673749525,0.08987565289465205,0.45091974908910293,
 0.0608247676473036,0.6454820026455738,0.039598954249352,0.639133249150409,
 0.3696433426417901,0.7342236580882623,0.23011634345462806,0.665721257619544,
 0.18237098706672927,0.8497940318246566,0.5419199686720018,0.5425215205443629,
 0.9914287567418859,0.1278691573513301,0.7993984028067603,0.5605076491441011,
 0.6018689629251615,0.6754075786145871,0.190168110311289,0.6142327103327555,
 0.5149626460085382,0.2833386607823535,0.5211360940579918,0.38936758355256196,
 0.3899647628967484,0.7649570450979405,0.49034688037400576,
 0.020392698219631078,0.32882371918662145,0.03016954923006021,
 0.30858402843980304,0.2676842617524261,0.11821853907079605,
 0.7628483315325931,0.009004584933550275,0.42280541743652467,
 0.5008836951803866,0.4363497080834552,0.08857300320064732,
 0.16061587539310773,0.0132399781957111,0.7772688778596958,0.6881233639141994,
 0.5363523830801302,0.13804756546984254,0.7533505178365069,0.5929853100396966,
 0.08347773529046754,0.6172915785320561,0.7738291212612662,
 0.15582182848822412,0.7255191180787463,0.8763094176261046,0.2754464439637987,
 0.5692708479935965,0.9163846352111453,0.17805550202296727,0.7957179035707578,
 0.23875984126287675,0.5902326952283922,0.8628356550800423,0.6741164345250515,
 0.4808778763310695,0.4878095474338162,0.3446722142843265,0.8609365421797059,
 0.7293340565545965,0.2691463041400045,0.25110416318104445,0.2881480288271838,
 0.6211775973282581,0.037110674960010304,0.6122468064098119,
 0.2089884434466871,0.13482067208564885,0.186605476672562,0.4530163019907356,
 0.6590333763333568,0.10371733043769615,0.38682915555151864,
 0.03055855226979476,0.4538107315218394,0.04635227039023937,
 0.6560414715601227,0.31003595875154744,0.19225417977994563,
 0.5725033945131501,0.6350831913973286,0.8046173927410688,0.6826822268203747,
 0.9138935553853409,0.11689020059914401,0.20295687802449702,
 0.5992589346329166,0.0811309664077614,0.1846861286539654,0.38132883167251064,
 0.45150413326521066,0.3685113781131367,0.2583304931218112,0.5579093073780106,
 0.5347879147602228,0.5061599719671259,0.19680190083043303,0.0689472722613258,
 0.39942250221758824,0.015293032180443089,0.07277666084117052,
 0.44416172476816085,0.21560874861065504,0.6913326886806929,
 0.29451302319824935,0.7133312984056727,0.3140890872899931,0.753029773597166,
 0.46129070831315155,0.7679514328684954,0.27865814619956186,
 0.022107170137479648,0.603281432412962,0.9963860297409785,0.5721490094739451,
 0.28160223801727136,0.5958074182337991,0.1213974848114292,0.4239081131224042,
 0.7052084722503664,0.02825564965604399,0.7506605819005951,0.5791728486398462,
 0.6992579742238488,0.30598366827382084,0.8021020598446643,0.2125043618065282,
 0.8532459993678649,0.787242214421215,0.11328480555338605,0.22433775981361137,
 0.534037995400333,0.07363309988247313,0.012642794340079422,
 0.1555611657510757,0.32076088972346906,0.2163567459599699,
 0.05994423462447185,0.583215547160823,0.08622396288249878,0.4109163987748944,
 0.7869236758243284,0.6040011070515561,0.6833464860662932,0.3387340444486402,
 0.40252164882666497,0.26430537745935456,0.1518852538143397,
 0.8277100725124109,0.7804023551567897,0.8999461628821108,0.34613759155277946,
 0.9928118602765899,0.2588173778212234,0.5548235983909633,0.9304407159195961,
 0.4618013905485956,0.6756010552325131,0.23652757868547358,
 0.19642793291875382,0.3476448157173351,0.01743484353720992,
 0.8375787803858078,0.38042984756636733,0.365336133192109,0.7121177519354099,
 0.6681338259992227,0.1209298767662017,0.08193433276719353,0.6083639178297429,
 0.8447601066692731,0.4801086648196532,0.5189201033745966,0.4394718406907644,
 0.08004253345775747,0.543866478191035,0.7219703168782082,0.126454353713486,
 0.9078882033425059,0.1465245027804286,0.4533710037606675,0.2893670471877019,
 0.7432512155082724,0.8848987013316567,0.33175379973094854,0.6987362032604902,
 0.7397803981546851,0.44767146147833226,0.13200964230337597,
 0.6719250662789761,0.21865233001996043,0.5921764096173234,0.8127196358701525,
 0.6955638202651182,0.766915360096612,0.44362132406120147,0.49393534750123946,
 0.7641164651595935,0.4771230220094639,0.0010800929622289956,
 0.5770115297461311,0.5614227649303855,0.6084596697893583,0.9605699997530123,
 0.4464828650632876,0.9965493791516987,0.026644398731841767,
 0.5882924032568598,0.22915570269530816,0.8090432706648985,0.3841679007152158,
 0.8718913749915856,0.013542976342825197,0.3125815263736672,
 0.05242518485136105,0.06287080262785893,0.20317011027516674,
 0.32551709570623044,0.5712062087772644,0.46316065753895685,
 0.3739090020760537,0.5034213102398342,0.07042095417246019,
 0.25527991338570166,0.19029771448014343,0.9517778976814734,
 0.38393501706839117,0.003338655193858231,0.6669806223400806,
 0.7464585724306297,0.5703144677120383,0.913730391650172,0.23636423990827782,
 0.021713213737455495,0.2933085000106721,0.524375614319918,0.9816649846487829,
 0.1449142345993576,0.55581861863582,0.7095834184898375,0.776134937453589,
 0.9878414908942353,0.3245981823329094,0.8630186011406901,0.7306437232977696,
 0.2265936364969401,0.7074420056368742,0.642273572011919,0.39131076190407277,
 0.6151730173210587,0.2881253146860453,0.2632906958883181,0.6908714269591146,
 0.22772392575834832,0.8142181595228434,0.9950762036997777,0.7618262723918808,
 0.372947040839981,0.5720965092427657,0.2528444308329474,0.7537591264042496,
 0.08210573944653188,0.37556151447246044,0.0960384998850301,
 0.7734065899369519,0.5209438862109284,0.8509629754247108,0.4474883793281529,
 0.7816572920392723,0.3234865568732258,0.2641619235463496,0.8173595380293489,
 0.5804745933229661,0.6876390307705739,0.7851972133492741,0.8662311820637905,
 0.6751231591141661,0.949290159866292,0.8135509102596505,0.727471461464107,
 0.42609438975322167,0.2203453804297021,0.13238602792511023,
 0.08280091487479635,0.6241562957273524,0.9755834732863234,0.5067985665761292,
 0.7094193723540932,0.6684461152910007,0.37334186727073426,
 0.12250735235566845,0.2435866033654417,0.9568687026699401,
 0.08287136716863985,0.9863838463723271,0.2958545345802075,0.9998420687594771,
 0.6684234823128785,0.10514374359950729,0.8461042364902429,
 0.32911631745940584,0.6346609167578964,0.37087272593648435,
 0.3356013776273653,0.162239598022772,0.5748567451791117,0.23503554587053632,
 0.00793858538050174,0.43191759886520764,0.41137970318069883,
 0.5644691219557424,0.5527640668309268,0.1777053261295929,0.2931926668938397,
 0.6129319596421432,0.3656987230909219,0.11793171439756733,0.5032419838293681,
 0.5229647024633182,0.4989109515004805,0.35040478446916523,0.7479834515853062,
 0.2358089019614007,0.22704167142010734,0.5373609430884542,0.7494371010840629,
 0.06687549748018418,0.2119239090343592,0.5579969388155004,0.8280029853783439,
 0.055080748910526944,0.8011181964791366,0.7049356434616451,
 0.4449463856551792,0.9168859672707687,0.7044173866054628,0.13012765272698057,
 0.9396660638591797,0.016498785062179944,0.34071771442270704,
 0.11226483619020966,0.45698217631607574,0.8443163746559931,
 0.9469285073769662,0.014498189024866548,0.41542874849299816,
 0.5450241602209533,0.3277833610450077,0.4104954039496246,0.8984400245008524,
 0.18653333132934835,0.28065090499213463,0.6167930117104569,
 0.6374196865300625,0.025902650275316796,0.6574038209575357,
 0.4987385638027728,0.4118854899747876,0.5510148281662699,0.5956401216924283,
 0.7768781651294349,0.33410586538162523,0.5004285819154144,0.9887755955645767,
 0.8528298888878926,0.46399984765179425,0.5182044204181574,
 0.28079388069075284,0.0899346112197692,0.43828903503371464,
 0.8636818399974954,0.27930188976114,0.7020610177758534,0.3351679341175431,
 0.7931998669406746,0.3765830968883854,0.14770965819735427,0.9591219558663429,
 0.27594921691237884,0.12138904118656435,0.8285854595064895,
 0.9872298681668263,0.14011923862903652,0.03925606784198088,
 0.32578070505837486,0.9393752963920472,0.23261185263356687,0.564679124157115,
 0.44229161561230834,0.7980696342133572,0.463527035303434,0.6517583156438014,
 0.4968748585698741,0.8569489898877971,0.2662879673679439,0.23242130326530774,
 0.29645394118705126,0.7605819194666363,0.39758641506369485,
 0.5230047616173034,0.1365877330151236,0.8414779635329666,
 0.045803725712235455,0.09285051969129654,0.7485448168625335,
 2.8175295062959904e-4,0.11213591120299404,0.09947158098990594,
 0.8071284078359657,0.9548434804275115,0.9169798021866865,0.6008233638888972,
 0.07862333647220976,0.4710860197579596,0.26021811017217666,
 0.36900059340626545,0.621150597777708,0.27717166214811007,0.529120577900254,
 0.06241160106867272,0.005794203592945912,0.604494288601568,
 0.10337201458686773,0.10576715955397165,0.27584056456630746,
 0.555318248026188,0.39576525456330625,0.6233219199861137,0.5340020707650545,
 0.620144054984802,0.4480467454671717,0.9349089575297787,0.9910994680144736,
 0.6294005017041487,0.43851539496320147,0.344476824906108,0.7460802586142979,
 0.5660644353831531,0.29712146555644103,0.7126385362382391,0.9748699882101772,
 0.22704797316043268,0.2610476943309158,0.43220748767732275,
 0.2483441622270598,0.8591950917138964,0.2038914442039721,0.362773565904293,
 0.7289247651545994,0.040448436036731206,0.3528377814653262,
 0.01928239637259921,0.02283283950988424,0.6099054956263816,
 0.6964170936790541,0.9228977293018699,0.5147389446329252,0.5984703125904023,
 0.4520016893383809,0.9828362153120025,0.2614632107390942,0.2115896999959691,
 0.2086320097787927,0.8037228952741604,0.8102174585351999,0.09343835274902323,
 0.9307145893804292,0.4191538170946175,0.7891259263002257,0.9853809758174767,
 0.3620578920442682,0.08410762090234536,0.14182641823480524,
 0.6360815521170454,0.6314737969125632,0.4782731964806246,0.584498754234689,
 0.13168038352576228,0.8144915618699571,0.5762233006394566,0.485664690547585,
 0.16774523266284613,0.91573569989835,0.4103236481033081,0.44832379598426997,
 0.6768233405112545,0.7127482722047132,0.9742450924818512,0.8632583395108758,
 0.8996604698373974,0.6893953609703752,0.10451202264088322,0.728497031155625,
 0.2979198380454343,0.2104730795662907,0.5182856322982797,0.44304302299914045,
 0.8130022760399731,0.8517280318216722,0.5866544954096196,0.591785616515485,
 0.7389852738860252,0.7316070504827519,0.2837921989133042,0.8928494035223487,
 0.15859417649177543,0.3908428139360304,0.8796650265708575,0.9419719156854819,
 0.48778299240001866,0.9126399556999709,0.020674220661056175,
 0.7513200593764437,0.8178797771143301,0.4536686779747646,0.7106566465740548,
 0.1472935266979376,0.3822343476058411,0.5257688812732466,0.9699175805140229,
 0.5712572714740093,0.1201016378392128,0.6974986056057326,0.16463957923061168,
 0.17946492775173384,0.5506088445249786,0.6063393030624104,0.6960260074324377,
 0.056775917170630574,0.11264592747102675,0.5885609592997018,
 0.5184454478170437,0.5408760938966144,0.9551754004349156,0.6353401631106262,
 0.7013201510854266,0.8772757484309426,0.9329891714931685,0.6444360601754056,
 0.832175883431987,0.38051578174610945,0.9978445825398253,0.16445412518258895,
 0.40407543763802756,0.9495234340562917,0.27483623341993624,
 0.9643117616735383,0.8068333373139017,0.39534784932644396,
 0.28523111437953963,0.7632049301319503,0.6129063915808256,
 0.43344540418738986,0.07444712129433151,0.8374260917049055,
 0.16055590503360828,0.6681145312745304,0.04078856166148315,0.720287131396541,
 0.5669693849400487,0.6590007270232421,0.22781775683605998,0.9165282816180889,
 0.013005556313632316,0.5575761892869089,0.46901855002182946,
 0.43468006290074757,0.9307482226547777,0.8810466442418192,
 0.019065898501356204,0.14301826462209588,0.875996086956074,
 0.5674932766724224,0.8572548081952202,0.6289937175498229,0.5957654250883023,
 0.5359661571357903,0.8617965826790285,0.26856541620772667,0.604718843243915,
 0.37135590874848445,0.38431903622430597,0.7643650985624542,
 0.9028547076697567,0.07686734450345245,0.42480428908133894,
 0.1928375963062655,0.8292229434416895,0.21569524415829933,0.6174394612271783,
 0.09115989945969438,0.8130632273667752,0.33422658722632914,
 0.7029270025542687,0.09630185259534785,0.5878400209511002,
 0.04238116669259939,0.6277956779364752,0.2853181142360143,0.8546652893568213,
 0.6815666305623901,0.723082399013219,0.12270075355371146,0.18186972025668136,
 0.3597176342355193,0.005866705882438206,0.23227892328607758,
 0.27441371864733843,0.9589281487350196,0.6597230634718046,0.9703491179710211,
 0.4483154489585517,0.8815777302186527,0.783861814011801,0.7160002781678265,
 0.31350836386877623,0.500056394728238,0.867680166584844,0.3945189180305785,
 0.9810347491255015,0.2160710418361821,0.16927579137808335,
 0.40474839630131176,0.3055862402423676,0.09254236005358674,
 0.6137371700183217,0.8466383058185811,0.6414810284569206,0.5064744432858219,
 0.22947174249896185,0.5142145297090167,0.14832095644851861,0.815525410817588,
 0.13003573427595294,0.9068433361240218,0.10399908587907358,
 0.9486177632734936,0.07224636316989752,0.7716422993031108,0.5636746658479324,
 0.030832997098654502,0.7199014026588595,0.30539543229161203,
 0.9627739277115637,0.8074766429996592,0.596444765503746,0.22842464093446857,
 0.0816937729242192,0.08387638493189453,0.12517622112769267,
 0.12250449466131541,0.5020079543483108,0.5203911493381712,
 0.33195525208487264,0.7649264244673784,0.8527209481366569,
 0.28428858551644787,0.02701110077077795,0.5141818442986934,
 0.6389398251892132,0.42173215103660444,0.37276732858251227,
 0.055356397326442264,0.21225541163343442,0.04278614152348503,
 0.3070296401023621,0.8214560824186462,0.6061973881096738,0.8764381532147361,
 0.9507235469005116,0.9216166886534614,0.061084595786680906,
 0.6246926142800556,0.07774345133984428,0.8418659420351832,0.4323797380921167,
 0.004862211165231445,0.7602426181139786,0.760724154041035,
 0.36635376082879134,0.1318165692066986,0.3635226506565634,
 0.048046606748595955,0.1929479708199362,0.13645206933786003,
 0.04724871454342172,0.4218486407576493,0.8809403662592286,0.5997572103323189,
 0.25643118342734894,0.20363160519773138,0.3705253600411863,
 0.35796563685334515,0.01519996873220486,0.38385612694026705,
 0.37600950825346235,0.28298737306112354,0.6676850993854636,
 0.5783428805206574,0.9149359070712582,0.6337395627243774,0.23761626254541424,
 0.8831749450791158,0.7972708165557866,0.657122031862515,0.12246664327074297,
 0.5624000889312939,0.15193343093980505,0.3859604395284071,
 0.06412515059103374,0.41463583061434695,0.07859199660146798,
 0.38440824047647126,0.26213632417623467,0.30916718162498946,
 0.735132529215718,0.180491570921531,0.042952029809173564,
 0.009107738395263576,0.8326570284474376,0.9789030655723314,
 0.4103053047330526,0.7236033205087696,0.5509172104956404,0.5378188804011088,
 0.13828571684423352,0.8764839518634199,0.347510922036939,0.3411999973993244,
 0.819867818334312,0.5197721327549832,0.12123846786679471,0.8780845726999741,
 0.6776813327372382,0.9095145365776404,0.8520554222134963,0.19204207978554044,
 0.7997298317040767,0.20313482146197137,0.797834495242484,0.9923352313665185,
 0.22911290108238125,0.017808759542595753,0.6763384277147201,
 0.9198762965468761,0.7670887538054907,0.6527092875870721,0.12985573007329132,
 0.8741444451824605,0.30288682124877153,0.20830025623292647,
 0.6823930113064355,0.46761669360885105,0.11202574327872594,
 0.2569565971094543,0.7316152122537959,0.7893649472459059,0.8057527464744214,
 0.9099170222464015,0.16535589859754574,0.7281936684862635,
 0.28093954943592625,0.6483387776659195,0.4684285564852343,0.7766728998224044,
 0.24164676300498034,0.12404629953845514,0.8613440877589665,
 0.3034624662948059,0.10851547698950936,0.5869112864900986,0.3073293206399409,
 0.7675499363776618,0.6797484559581612,0.4861450496701044,0.28609254437840326,
 0.8469838382459636,0.7728737330486192,0.9846764873827214,0.20482940437677077,
 0.03430995820655791,0.01779791071583725,0.5989008356761616,
 0.8300339912466681,0.7263205338886453,0.4008252418405389,0.7628522677477156,
 0.4916675679325079,0.05071285988746166,0.5083969544413106,0.9249450460332646,
 0.02305174415132949,0.29193843433676725,0.3688202112622706,
 0.6012497979307362,0.12352321353061557,0.4566568078963468,0.8333494360821034,
 0.11816394581992762,0.8723069478610096,0.13475286718485502,
 0.8552622045427327,0.5840107374811572,0.8415203545526428,0.8155308988570981,
 0.09372453851641704,0.15405204035605125,0.13655905085138687,
 0.6976217160354388,0.9808868345514425,0.6046149799898277,0.40118680538116813,
 0.8108818882268773,0.8824524782150848,0.779573146880321,0.6870309539431328,
 0.6271188862888062,0.5241438145936941,0.345627273916134,0.5085173957970149,
 0.8700362394729331,0.21645686049925605,0.3989856061243887,
 0.19735113740217358,0.5144693565036834,0.4432709429960018,
 0.38195312429995254,0.8569844482096315,0.5564671393738875,
 0.17580068664752468,0.7452601956962199,0.18721149054228003,
 0.9968418034947586,0.06166388122059541,0.47761863081168054,
 0.8854856466917431,0.7621226899632122,0.383472888277269,0.7602953789888082,
 0.8264851313481025,0.4995408241577519,0.9421769425596125,0.6815032674721629,
 0.6490076997536017,0.3033793339606723,0.6120739785901598,0.8156266820605275,
 0.12227273974271646,0.2660780997457113,0.06735698537841495,
 0.3972399538677429,0.9412672410099584,0.7766792107784275,0.08191936591269444,
 0.5773416605281215,0.4411437422317035,0.34634511714240535,
 0.45628412215675396,0.9341306239674472,0.44637583690171256,
 0.6755179765649552,0.2672731241624633,0.9907297702841759,0.4999375528678056,
 0.4796812474909653,0.024986977744347327,0.5388171293452352,
 0.28771572505931675,0.38098272085163365,0.9873222176882976,
 0.16349452562249134,0.5543103519531757,0.2929355332544308,0.8424299648567997,
 0.9229107533136671,0.9820120906282054,0.075882021634183,0.21506040381114622,
 0.5978576467746817,0.7896870951774475,0.45529918433024386,0.8845107843586529,
 0.9764813477324251,0.7121399239180759,0.5108613222870846,0.1631272790383217,
 0.8889793603667727,0.161736030181434,0.6814383180923419,0.2820896516154304,
 0.9383265513382404,0.42244302313860027,0.5379088097532644,0.9525726403234487,
 0.197639787758789,0.5417650972351831,0.3970514095646107,0.49096994682607453,
 0.37034572033363766,0.323464542948548,0.39433222958198166,
 0.03240357740058708,0.8826797433797838,0.3805437955284957,
 0.03860697485964726,0.20229201038973332,0.44156873695752497,
 0.9740889984763721,0.8991569465259909,0.16932583037531457,0.2632626663336267,
 0.18267447995127117,0.7154194798365853,0.5849160959602046,
 0.35444617200883266,0.92041446835918,0.8080665596746218,0.05393819957651402,
 0.564955116738336,0.05575967769818657,0.6933144726958526,0.892851658955389,
 0.9626006424768416,0.9231457656388976,0.33975089454108276,0.298901889054767,
 0.466481177925927,0.4723836935815442,0.693549242089103,0.7367410020437385,
 0.6854811756172294,0.2661241283413791,0.7953938735860333,0.5947739031427982,
 0.3870094889770731,0.6091164097304373,0.3343895167676778,0.7374859281422174,
 0.12802407903980384,0.6666054816110627,0.9501595186495719,0.8282839558102593,
 0.7086349847459757,0.8743352729376805,0.3599928981072136,0.6267852169702128,
 0.07436606673729274,0.30425783573855136,0.6622301676581456,
 0.20039862879484704,0.6074314103512476,0.6118682089938736,0.7997564400597793,
 0.14367397379438918,0.828449401282344,0.03732124935576919,
 0.27395778268694937,0.3851296276922205,0.4334290029131421,0.3464164542170667,
 0.06708539311520623,0.8003092369342073,0.573806317693003,0.11365542305611331,
 0.44203097951045955,0.46329919702531497,0.8166969534208415,
 0.2675711961746834,0.19838917245995313,0.7611074336993986,
 0.03830789247790767,0.6225806425182455,0.8257165599830039,0.7788100387209875,
 0.8115834228933538,0.41684161210726034,0.209490689409762,0.25071409403184464,
 0.9571604089094355,0.2213341234812748,0.6009552120683704,0.52181958287123,
 0.9579997425226626,0.5077287971800806,0.5074910284424954,0.930743621929086,
 0.16572413846675071,0.8202895403075131,0.17585237111895646,
 0.4860318131221212,0.7431425952335603,0.14034578709104695,
 0.35283047578079174,0.5936523545815691,0.8152871020234145,
 0.41268112382634936,0.8687885278620141,0.172841634989101,0.1800465377573519,
 0.49162206029812394,0.42516320509151395,0.026447187495301616,
 0.4049600361216794,0.8003878905132092,0.8188745209672502,0.46950985366005216,
 0.05410092504487429,0.7600604322946123,0.10602969800131135,
 0.7545928057819838,0.06171648034146582,0.977034477419909,0.8915699552729663,
 0.905666963318455,0.6483726187842087,0.3214313557677495,0.34720834313303117,
 0.009043076851530496,0.9825159443837171,0.2282886456754376,
 0.0690801278563965,0.047025712585324064,0.9004396985371903,
 0.8577661556311656,0.6697345760306745,0.038590581048426165,
 0.06300547309440452,0.6078695323560468,0.6394424888972806,0.7371001116803784,
 0.11807714720364171,0.7965044452636958,0.5844178995563163,0.7706839795977778,
 0.6404296078427606,0.9923425439217592,0.24684948911364746,
 0.10414818157021644,0.9538509687305955,0.86758324951982,0.5610491106570356,
 0.020561196160333695,0.014688736622312626,0.16706785653972955,
 0.2878962138264708,0.19867415346620687,0.9485491047722572,0.920550601688475,
 0.82495067656977,0.7306860286620311,0.7233003176877645,0.17297530925596227,
 0.8450757378094848,0.19931805554093085,0.24808090356849433,
 0.5089333517252996,0.3523602284407007,0.20050411956968162,0.2840524906862424,
 0.7666769162928628,0.131391883167027,0.9668855038278039,0.16991608825537352,
 0.36174185775124645,0.5772268753592696,0.6650351445914424,0.8220496579247978,
 0.5255662596376962,0.20054303603402834,0.7183689066621103,0.7426007827638106,
 0.7264819434762527,0.49677825305965595,0.9145920779191955,0.5458115230501183,
 0.7808138689711303,0.505041281159619,0.7239199057521744,0.9643786168381853,
 0.8161837960924807,0.7157762852670658,0.6346023443435759,0.3575720989876342,
 0.4564787035978922,0.2788078640822884,0.5326853978785973,0.8811108408020571,
 0.7543054203802904,0.2033976662500424,0.5626551025092418,0.5500589250369718,
 0.46259137797351535,0.8861883151943492,0.5566202889897167,0.8458839272297152,
 0.13766691619088445,0.28315472981052603,0.845393094133547,0.6399083147324915,
 0.26718085465219066,0.08490743376430232,0.3966174142181641,
 0.5919255988567034,0.28697294898756043,0.7425534122002276,0.8566615888710453,
 0.4167569733359835,0.9644766990405449,0.014599354936535303,
 0.4495024860019097,0.9107543272694267,0.1695334173090669,0.7599693494150453,
 0.6846372380159577,0.08087802690345636,0.6595359134138482,0.3590655209224317,
 0.8780207592074933,0.07201080855716402,0.1423789429539184,0.8810501904562584,
 0.350163214186734,0.952960313827856,0.5194141528917304,0.9905753032412336,
 0.5977816276626822,0.1038829609031322,0.9549109011592394,0.3075876210104611,
 0.3847641156017334,0.8925474408626636,0.9770464355769086,0.7991971337195696,
 0.5580398100092212,0.35020529122739796,0.6665194917211008,
 0.13124566391688453,0.18624103912631818,0.05687387674049016,
 0.8554878184254618,0.8394325905535118,0.4934552911456703,0.8119731420318544,
 0.08607916527567094,0.5821294361969784,0.591371306103936,0.6398595329855736,
 0.9717995668276489,0.702801014454079,0.6645399198914824,0.7405992765427705,
 0.5370127537342348,0.5271111663393083,0.5930633159420157,0.2471714578582751,
 0.45567549559963794,0.5007849352978287,0.18683047457889757,
 0.4642461721968212,0.08780435675744358,0.020901647218914743,
 0.9449673147959962,0.2661831973384744,0.6798561856563775,0.5664180016477807,
 0.598959016059093,0.1888183891308931,0.010248571198559153,0.5233746713476868,
 0.9748971290404124,0.385753688643785,0.029636834121810507,
 0.34901429763031544,0.8586802451547499,0.730503138426088,0.8339573304104013,
 0.564445584936561,0.7174375319757316,0.9675551819893298,0.8254724159238342,
 0.688474404490394,0.9305461358622348,0.9045536986940803,0.5341882189647024,
 0.8865071864148062,0.48645429291316866,0.04768803396720678,
 0.23476933495110297,0.7822983797210467,0.43239155484474745,
 0.22173889553681936,0.33707687541155307,0.6559886249818797,
 0.6500166234059737,0.43418072876220515,0.9892109286052174,0.5932608126768434,
 0.3812791292811589,0.9772510014276715,0.003044668897911662,
 0.8788066823091367,0.08274207979858617,0.07188037994643981,
 0.4126723717041796,0.7892813573569366,0.8627145235985689,0.25446596132538835,
 0.9004811118328817,0.1391849531775181,0.8547646101408595,0.7194745218696247,
 0.679558615625994,0.08544350831110537,0.4010933348230301,0.6110012668150783,
 0.5074136108073901,0.2810524304684605,0.3588343544931979,0.8520942198187189,
 0.15878943226626852,0.5558075777183012,0.9228556947745379,
 0.33238874711476496,0.9103387098977915,0.886758128264405,0.4337554675343436,
 0.3230982412623692,0.12389953135123122,0.704565300541326,0.7567610305548627,
 0.3148719840793217,0.04112663567197339,0.94059227161807,0.017087064681718167,
 0.22558420895049514,0.5238029817424084,0.7145379276903031,0.9828168553212121,
 0.2724617533578423,0.5058215850319754,0.2505347350510677,0.6201609058158305,
 0.8723835239456452,0.7799474423612527,0.32167809956966487,0.6870223810641759,
 0.8089603177619259,0.6206252410262473,0.04300676982801688,0.3775946366695354,
 0.17701617515197643,0.8642061512228961,0.2597156872487546,0.4238985391158928,
 0.6687136324934918,0.15732942984356957,0.26772842946783737,0.66266616934566,
 0.39006275691346204,0.7488460779409182,0.7463401638991299,
 0.23344389381224595,0.8265770040210824,0.9524439579565953,0.1465016509112953,
 0.6841578112697457,0.8459295294100164,0.8902098855265852,0.31014756663663,
 0.431144084367873,0.2629550495413495,0.3023244950023043,0.8214969652086648,
 0.05526474165104933,0.644555962982924,0.3706238855649602,0.7379409451120109,
 0.8479698622393463,0.46077472186906965,0.7595104073055571,0.301989437978499,
 0.5863695098504546,0.9017169252011388,0.5298925021315238,0.3458693897946824,
 0.6557350210893512,0.9771217409066312,0.7425190121938077,0.04491255388203319,
 0.40801667335725555,0.864239665992019,0.555451915953517,0.8587727244009118,
 0.3233352624967438,0.7729667257207252,0.357851960208454,0.4700684901705604,
 0.17393785174777265,0.20407582840779992,0.7406033003230521,
 0.8590983262169789,0.043942659354286984,0.38490134561488554,
 0.7494962356895627,0.5333990501483834,0.1931244666135875,
 0.0022851003816101167,0.8939969531033947,0.6825517265306349,
 0.9074061138066065,0.9430213345485667,0.4003146697666765,0.06581437421584557,
 0.37975469373892345,0.7143519919712576,0.5348771533317551,0.5207987210896429,
 0.7748236526668233,0.06639670055925395,0.37136415696278924,
 0.08943443458129707,0.3624464667402053,0.8585175072272169,
 0.07491447745352664,0.11668714176146022,0.034242000720115295,
 0.6183709633373303,0.8543872314838815,0.31770975975050997,0.6502917094514818,
 0.5656977254257732,0.808302915593761,0.6225356080344405,0.3016778632876307,
 0.9757125785658748,0.8854630422464245,0.24849022919032238,0.9744874084727922,
 0.9634091377021727,0.5674796173838287,0.9614318289113786,0.9461177507138392,
 0.7570147993753007,0.6644055380880063,0.5952741254521305,0.5968174094965772,
 0.1739353659440419,0.2327291458435582,0.042918171343285905,
 0.2604304386301165,0.25900157629069975,0.03266117348590336,
 0.32458490045494615,0.26474808226865454,0.40115310148034067,
 0.9828395326671576,0.19945153231970902,0.9934730903422485,0.9497420162644136,
 0.9927289080276536,0.4537453726023497,0.18707116465254525,
 0.40751337592300363,0.7667933096568731,0.4739296493738898,
 0.32507290765473085,0.6887434993138022,0.10599221153430449,
 0.7217360252705268,0.06103549574984257,0.7029174809477297,0.8936054837039403,
 0.052113536187715415,0.6555524528524396,0.15461658096582875,
 0.22139252854136615,0.7268331982969628,0.5771700497098561,0.9545011593992061,
 0.32655066049537496,0.10037332347014016,0.7389446459782179,
 0.6731227208896708,0.7244658700661079,0.5508118430754034,0.10640247662389923,
 0.9286638315916903,0.9913822342009136,0.9856058972486318,0.6138075188339788,
 0.369195086789184,0.4113885357983096,0.24428518568220137,0.2955263875128199,
 0.13390118377571236,0.03496622166999175,0.12706668578177527,
 0.28337104191342977,0.7868679540325649,0.4723569507301273,
 0.09221322668338894,0.5306249934720358,0.6628723924087527,0.8243089785021651,
 0.4519881846485775,0.19870640952788965,0.011039868555273613,
 0.4214230932078996,0.7504819295397437,0.771780616088999,0.12518284263603374,
 0.5995213508524463,0.2569727790907883,0.7195639839861183,0.650850147118321,
 0.1638329955486416,0.3927170363773115,0.17346109583480063,0.549429526546841,
 0.7069130124012598,0.4515757695339627,0.5706951014010011,0.6328936848427212,
 0.5723384461938827,0.9604670999491695,0.36795982770365987,
 0.12314814249867045,0.7906171064229709,0.6156368188722144,
 0.44661949579176063,0.4647127601494253,0.5100590575091487,
 0.22796247409720694,0.13298024914354423,0.1932803283226604,
 0.20592525643578274,0.17279152611540205,0.8588301063377567,
 0.9790306039596286,0.2110615381298746,0.8720750506467398,0.9998664465062137,
 0.3345192328738993,0.18313672458404884,0.8407575219219551,0.8873698463287407,
 0.8808132381782925,0.2902623712390786,0.8864426382881476,0.08540986892359992,
 0.15122218048429692,0.728445598609854,0.263680106276147,0.7337469615837344,
 0.14903276412407251,0.1196976492097076,0.6408585398203129,0.6691406819566037,
 0.474088338783236,0.5783907630661411,0.14001680433192765,0.339483116289267,
 0.5675806329240457,0.7936648030780874,0.7244490637371905,0.3044352301001477,
 0.20318807657611238,0.2065049196473474,0.4757163361510256,0.4249736656240257,
 0.9344960204163594,0.9670043946990419,0.1502815286835545,0.24797424790502665,
 0.1634423275597816,0.22463363958266547,0.6100953639399893,0.4982614133291431,
 0.43332179665497206,0.16490759377692943,0.3798089161011844,
 0.9797368961407482,0.6622649850839459,0.47765742714034476,
 0.35333735481959705,0.9019446364981148,0.022341319824833783,
 0.08038820902501864,0.5603559076003557,0.05898115453175812,
 0.015885948088288693,0.42327970910764956,0.29862578830204367,
 0.9185478986868497,0.3980319031353893,0.9149502452513243,0.5868957825085515,
 0.34924735867461676,0.7969898603863562,0.8374843051819792,
 0.027410158385188677,0.930499169872528,0.6892771701037218,0.3091770537285462,
 0.7743404099008722,0.0513191222342525,0.13497090843405657,0.5221034631263453,
 0.9698685593681133,0.5513437978622906,0.019075100743879525,
 0.7762673479263222,0.3443323219094151,0.20449591628543218,0.1326458494295527,
 0.6363770311060983,0.6884473004099941,0.014422730861307964,
 0.2901268254462688,0.7630000221895523,0.910440344974745,0.10599178734686987,
 0.6182016455878053,0.9927456026381526,0.27891785129154445,
 0.36805215256481305,0.6447374656663016,0.3596876467488255,0.6904553391154791,
 0.9117892144713334,0.7232169436869201,0.67579173385852,0.8546581199685467,
 0.6099813687659926,0.048978163336555136,0.5728651148718267,
 0.7446873482357231,0.20004403514329328,0.6632479425485609,
 0.44768327065876024,0.5492053237091524,0.20475017211625446,
 0.6498375584643044,0.7742701074889858,0.7288087351775473,0.7938156955434651,
 0.22195183015126996,0.8393668068705517,0.03397757573983451,
 0.27558015655778734,0.7976209132302668,0.33110309656282766,
 0.8349535941472759,0.9653386746634676,0.8767787450121988,0.8606290889662397,
 0.6692524788094403,0.4814352590340757,0.9036386114534384,0.8301411301722618,
 0.4586929651920125,0.9466597416140847,0.5771547286812514,0.27013799715794373,
 0.06779412640524907,0.7613508272335056,0.44315380141535277,
 0.31296012832586484,0.9184365362817121,0.9534086156134562,
 0.47385422741462513,0.4351620392728487,0.07235744973246372,
 0.8438748226816232,0.7651801333017763,0.9521543804514201,0.05816928466501092,
 0.2778404150403798,0.5396998136781818,0.27315294538814405,0.5530192132591225,
 0.21254711094933632,0.6899303144465061,0.9760770087061957,0.9759252303067353,
 0.620543343423115,0.0119615037080264,0.011833116478295658,0.6097105420188286,
 0.6505659044728014,0.5239458070879346,0.43169738197992924,
 0.11608302669661263,0.13411941248137693,0.284634376010673,
 0.16384246468172614,0.05805353547055192,0.7587692847859622,
 0.05566446812935666,0.46508239674394114,0.3884503231684612,
 0.9559717395092159,0.7048467894823646,0.6745620217310749,0.566049161552196,
 0.8244100715515021,0.04893238778181441,0.6588866933838208,
 0.06557911440529829,0.7809063176356473,0.4511861255558932,0.5571984821345678,
 0.9085747602247085,0.38816644032765724,0.9673475694895651,0.6002749942300757,
 0.5897541360352322,0.9049812484511559,0.08109942671590054,0.3198161886875055,
 0.3700205265261156,0.02206053268871777,0.6657708574876273,0.9861182678512749,
 0.4061780530605308,0.2867597124390052,0.12924787153536565,0.6448507402766506,
 0.632487049905595,0.40821025895669083,0.14955452668706437,0.7591445647356381,
 0.21812931536278435,0.5601799259793308,0.33103948299100516,
 0.47324272577119864,0.514567613913761,0.3759126902946449,0.4105758472523411,
 0.21402936731299504,0.9348694869610459,0.6369461963823676,0.8359329303689323,
 0.7809749171786582,0.8664211881779093,0.27221069402560316,
 0.35692587133931086,0.6109095875711024,0.8702433587161499,
 0.33111457329852656,0.7607894871308137,0.4077930441734967,0.9032940485314074,
 0.8786239264999313,0.32067518029267283,0.2601815071442183,0.4641147593341094,
 0.8302454596147063,0.44516499179910085,0.9365411470308884,0.6988637493123854,
 0.43776795901215126,0.5952248842196203,0.6121989139052229,0.6079596471354393,
 0.3811138342710453,0.8470809144672911,0.0700129918640815,0.48713584834623724,
 0.8206237158117563,0.44732324641593335,0.5850340842560561,0.8238902059586937,
 0.9289263578027374,0.831739101629689,0.4906973695730834,0.1237295443091746,
 0.21316225812582645,0.5755181520647384,0.49046026404034515,
 0.8611168322836859,0.929063298441072,0.08968601079509253,0.9384885886227091,
 0.2869446307892035,0.9911046573732438,0.30794025170706407,0.3648909741260036,
 0.6022775391911966,0.5580318564531497,0.5832662230654094,0.2887771979944953,
 0.0414064905693986,0.601306250991678,0.1585608071939073,0.6947872307145412,
 0.5785489934954191,0.005162896767799352,0.3140394879114191,0.841550922544099,
 0.9111498727791364,0.635026146179946,0.6946374771049775,0.2072057973788054,
 0.29532985206432283,0.45882161759444073,0.8534654507778763,
 0.45561133210862825,0.8177797672645192,0.7032949592009358,0.6221907186167173,
 0.7518202879421477,0.550460699703419,0.5357409437452579,0.6743413465821464,
 0.6994959313182698,0.6812726357270655,0.3770629280029656,0.8000503205960198,
 0.999286210682176,0.6775006490341092,0.13348476672536824,0.7423472152941427,
 0.9383578752876276,0.6683793933414561,0.07928847446707665,0.9056924311571816,
 0.767111503493266,0.6956617965812514,0.2037538558502251,0.5052320567392196,
 0.06578797802620251,0.3900536493771257,0.004898772426315756,
 0.12938602563443669,0.8050202289695053,0.17819753286465456,
 0.006481496910121232,0.4578458068498982,0.5180176352059125,
 0.7964321899928238,0.9097050473309052,0.46201653958363165,
 0.24018734171376221,0.07235538182358647,0.4513981641107373,
 0.7805703695361611,0.7363703026994899,0.3610867455562371,0.8680370588869603,
 0.259566991699799,0.22867993984448365,0.19085513920521546,0.6505108788732893,
 0.8644372209789116,0.43199113900675634,0.0626383429855627,
 0.06641288182494087,0.1490699872983844,0.6544671760022934,
 0.19926003794660418,0.44434158382229905,0.883483784480785,0.9947925240457857,
 0.9092658536622857,0.755437989453365,0.7107093354769869,0.10674620763428988,
 0.23383372661959922,0.5626615570498503,0.1940276224775297,0.5321343927768861,
 0.04106614498572814,0.0072683829700476155,0.6056046115603837,
 0.5732472025087448,0.6194441333215194,0.8112351317151032,0.8212820581610326,
 0.8259760755907621,0.7331256847973643,0.35123489199816804,0.3159462431487925,
 0.8631042809478353,0.870533588563909,0.06855706956827501,0.5928493902974716,
 0.08878402021896648,0.6868221473256162,0.7010037609811708,
 0.17579717693423735,0.15320634560660396,0.6243805410501923,
 0.6514327621396778,0.7001172425865981,0.08964828025490634,
 0.29781476987432587,0.44431711976877997,0.7985796108375567,
 0.37759756305404046,0.940367964636499,0.7607525465216121,0.20567191662823048,
 0.4187201774439503,0.32203602722046276,0.2500979521601411,0.5598570240275016,
 0.8558125728534901,0.5355165837799197,0.7005967493070658,0.21376040636282234,
 0.5567610674511236,0.2270379364895151,0.10130641689337982,0.8537196653436006,
 0.07449106594225574,0.629957871475868,0.5791277227680599,0.6217150402012239,
 0.14840071498708518,0.27327331536664734,0.6057957383481222,0.505304403538715,
 0.6385258544336345,0.38919997759254565,0.24256587695115273,
 0.23021811851688423,0.1773645250500122,0.08214777147008956,0.271809590020055,
 0.7648120967239772,0.5321414082250068,0.3312078832847605,0.8321708281336335,
 0.44874912936736167,0.8183111831389263,0.3606498592923061,
 0.053139761401015284,0.6628129804658306,0.44637655300834655,
 0.6918061758816177,0.5126334760385879,0.043158457486744406,
 0.6348379091558662,0.3794388882777098,0.5388480550586161,0.41759478645676773,
 0.26904253010227985,0.5975175731544868,0.6993551073293492,0.6303123394806702,
 0.6245105916332752,0.7487024768014283,0.9728414879077596,0.6876959434749097,
 0.5026932025902248,0.262746482162605,0.1116381139238446,0.17707503781788092,
 0.1266704144955143,0.6117548271832783,0.8587707886800195,0.09024927848656816,
 0.5104645982509263,0.5712276823988276,0.0722453080188249,0.03172809112492847,
 0.21850550737051821,0.09824820598047168,0.782326186359142,
 0.28203233444710407,0.7568552345214309,0.09570009747466113,
 0.5317358934285353,0.8898255161579828,0.9551057819124549,0.8184423074840033,
 0.3624267065985374,0.04184417679600849,0.04208660419020971,
 0.8116552106286059,0.969850549984869,0.44357972796212464,0.19420534210248286,
 0.8616674423048021,0.26394039846274486,0.635249599964107,0.8549252984785666,
 0.047788744531969796,0.6310810017600492,0.8577470620847762,
 0.7953393694004853,0.046467326298259004,0.5171283258077068,
 0.6704307006749998,0.15809783102050956,0.6067852113427594,0.7186290687842213,
 0.40861209453657377,0.3473369753316381,0.05174843442068722,
 0.5482383943570472,0.7374182086027378,0.9791794978893518,0.31058179123085816,
 0.9054089225626061,0.4755975378116769,0.557186907885468,0.6081393619082177,
 0.14113731728608203,0.4977508929990787,0.8081826283595966,0.4146159414535793,
 0.8888359908355816,0.7576372201625076,0.07258000242294183,
 0.24255183561608373,0.9795611575108043,0.12672903474687935,
 0.22095493829600077,0.30099997099900133,0.4691584603214082,
 0.02564831640044407,0.3645587816892233,0.6632975983258267,
 0.17254722471526518,0.5185391045770535,0.3558712724405635,
 0.07746512201628863,0.14968265709507156,0.8002512052335196,0.336689314477838,
 0.8528435251106734,0.3133604403536139,0.3385490034984373,0.9867867853960334,
 0.33666986227614026,0.38504486382753156,0.328998949906517,
 0.049651052948994634,0.9202025392985284,0.5277918247649553,
 0.1843272775345468,0.006620148639680856,0.3998992958093922,
 0.8339552662453089,0.9872780683586806,0.7143917572939477,0.6785440810408001,
 0.05711716169707859,0.7755329924094989,0.7796823813779865,0.8652411648009593,
 0.2836465485600804,0.5487711056131499,0.908976941786468,0.8213777650630992,
 0.5702564826339203,0.06551218274503301,0.6002331710071549,
 0.27930464633100593,0.16930982393625094,0.12434708735021371,
 0.03407245148181304,0.8329467753624976,0.6978167624656547,0.1280885799896112,
 0.4290621639698464,0.5421452463263305,0.6889858946770718,0.4768169844391841,
 0.8931768682660439,0.24715937958165368,0.2614891708725783,0.955828382217948,
 0.06307083777851263,0.6563041210878147,0.4049620050413769,0.8959441410005289,
 0.40330237068241526,0.1455908292779634,0.3131794766449152,0.3865785754794052,
 0.7947742034434279,0.5756959576828294,0.8200158121481609,0.5641207802678225,
 0.5239054802209678,0.8299160642603541,0.5533795911974538,0.8696342263234693,
 0.18264729894547038,0.9851861773141429,0.8751306479181137,0.5810576276214122,
 0.3846685122113629,0.280233073459374,0.5683890193818455,0.8795422372806716,
 0.7932573420290537,0.531454996539014,0.049964914179084996,0.7856229403376687,
 0.7374233687214612,0.08785455316941893,0.4929328057147826,0.4885158811165171,
 0.8500830524945219,0.540857134655213,0.7088312687172201,0.6507434908133928,
 0.5849964941209982,0.659179293886094,0.09963172457566527,0.9806783180685344,
 0.8545426225974935,0.526418946220215,0.2702189177414184,0.11236540494404135,
 0.5828759656296122,0.36158964102838487,0.7027473635287796,0.8181998334897918,
 0.3263347829965946,0.6440121110875676,0.0457262511307146,0.47762085096551765,
 0.2669095562976064,0.6222335879798796,0.3649793717109606,0.03281583036668545,
 0.7058438763826822,0.4963131933475462,0.011310559706713241,
 0.6420069831437313,0.9029959344111801,0.6102701136185351,0.0982440586234401,
 0.08469388628148744,0.30967355038276045,0.8731884248344447,
 0.6016646064085213,0.6902386483836687,0.6823844215266303,0.3588165743507622,
 0.9649764066398128,0.6523888398264442,0.4130162460159832,
 0.041661334305149844,0.711964330331685,0.010744343258336175,
 0.38629435923139654,0.2706581657242928,0.2559529847541181,0.7115656334781042,
 0.4622606597371983,0.42814778408650545,0.7103792868494158,
 0.44877331992718494,0.47794201406573134,0.7207361510493225,
 0.3427758923646922,0.41840238751075465,0.8353927544370519,
 0.21694322718716086,0.7595755312861933,0.48647248315948033,
 0.4032294241766746,0.798375517745067,0.268716916916401,0.4009752310489817,
 0.40179195124187084,0.2036499714487,0.5655137894809636,0.6202737177984198,
 0.31947814473960545,0.8534594147350439,0.39942299852297236,
 0.46916417960207324,0.11452496307825533,0.8010832933980545,
 0.18226747630793405,0.04199061976957952,0.2669212574206543,
 0.016089656962715693,0.8813106488342969,0.29680872822538096,
 0.2759690789210151,0.4370710699491167,0.18367613267838767,0.9369999565330855,
 0.1500739547873382,0.9278771488111002,0.6363607554969499,0.6165462736865678,
 0.14283363102980084,0.16210724813285438,0.26445844513857,0.5895225930175637,
 0.006754261919080462,0.9856975739013685,0.2369172712167258,
 0.16131402943727102,0.5311755805252684,0.29584340159879874,
 0.15728249990172471,0.05786456447388888,0.29361230950781714,
 0.6119315132323476,0.5749342170765789,0.5844096279022916,0.29061631628945395,
 0.2953298310628376,0.95503070220898,0.9191698714262658,0.3807774165154383,
 0.9183866505461878,0.2748394996642537,0.5031556804309462,0.6509283685282811,
 0.5577086593350828,0.6747603103777323,0.34581645165420083,0.3679936024309197,
 0.030813474556724674,0.4475410825300219,0.5115449193191698,
 0.5287569155466592,0.23481733949646755,0.3896906422012456,0.5657692254872865,
 0.8339254896172876,0.9972370936830078,0.6821579720187618,0.4751256419498039,
 0.7601181153162153,0.5634579563491389,0.3301367701570196,0.7166459933187017,
 0.4354409953059387,0.9495873765043985,0.7748729429228759,0.9323695491953568,
 0.9683845488779276,0.5982095851622109,0.24623096540441736,
 0.05757633956563746,0.3698123835227771,0.38745949105083477,
 0.8393467554149935,0.960153438322778,0.5796685048609694,0.37225635735072904,
 0.8906338040169015,0.15657719077028132,0.8032333572844549,
 0.09278621059850223,0.032050908414760326,0.4801418219735323,
 0.46773692951971935,0.8645706481882333,0.6873226708787,0.185184731160986,
 0.3812558129628729,0.3502078082784785,0.4839609326289833,0.12541582424204367,
 0.8989154446042105,0.997782394477329,0.6531101983680926,0.8138888833866381,
 0.9976231941460257,0.9676097118449499,0.4805362505619887,0.10862924159337173,
 0.7630750568917856,0.9201828795237227,0.5861575880714845,0.3453423472077706,
 0.3237335260274018,0.31001924510431755,0.5887360811368285,0.8248487858229339,
 0.5548254251343073,0.9067584902889094,0.9227164082951039,0.25884388435898553,
 0.8948366266979868,0.18612142245993857,0.4765632918942856,
 0.24779703508829176,0.6919437407782859,0.7685523689699991,0.707138410605484,
 0.5403099897079767,0.05083890905752331,0.5279897702694946,
 0.011264097759006275,0.24298754918439247,0.9990489877884453,
 0.23682140482727998,0.8552899544853211,0.4609498702995525,0.4901964626011063,
 0.05521192059596125,0.932183259224395,0.7012311180121985,0.9499263575709805,
 0.3052872232946844,0.6888396008945159,0.8733375121320378,0.7598407783853717,
 0.12878453806248802,0.7344908883792658,0.021743824274007117,
 0.22533684317559322,0.7720941997639842,0.4240189825314289,
 0.40752970366446406,0.6032146046774612,0.22870980265432472,
 0.27972376126424414,0.12232694309771308,0.7738091784639903,
 0.37789137565619646,0.41302292304960087,0.5629432147141219,
 0.6853410960728182,0.5471769788938743,0.05407969143795255,
 0.03516578247156854,0.06917396891988503,0.1543357073030318,
 0.6656586441619613,0.3324638126910766,0.8551610875429779,0.47319118680100414,
 0.7225152298113271,0.024101344681691006,0.8063538022330132,
 0.29041154779029277,0.7141785950303098,0.9192141199609908,0.3958889463404833,
 0.7155173887518793,0.3383469423674089,0.9503519389512906,0.13266617910871414,
 0.5220046189290649,0.7380799382677788,0.23667073511945325,0.3417406691181507,
 0.7238490979237976,0.49087605504170184,0.7927704677604557,0.1662882218868833,
 0.16827560435097666,0.6548298855072502,0.9245533132079747,0.7899018287882729,
 0.37151793809985,0.2211199704611687,0.9424591232734798,0.9817642570909407,
 0.18911132926801333,0.7608451302222788,0.564150847736701,0.6089786706848317,
 0.3078424117532431,0.5480596217753295,0.008542882239790828,
 0.5149732073588404,0.7334549875912998,0.2337205577951713,0.8309220291389573,
 0.9200185032133844,0.9951091306156298,0.772839812962233,0.7176602832524087,
 0.9165948886780376,0.5246394339408543,0.27646478210407166,0.6280051798462618,
 0.4713684111879266,0.6800229283445796,0.743262900765945,0.7756691236109589,
 0.06980902018691604,0.1635886020675138,0.19631979899973195,
 0.44945202007243046,0.3835005866739988,0.7251454557057861,
 0.27940915755379436,0.019886644879486992,0.14031542358964888,
 0.2846801122013989,0.6251392614168914,0.6748102867694898,0.23976685205603654,
 0.08442125818360091,0.426049524857825,0.2542722945633823,0.8679710400878855,
 0.4376551765096125,0.43075735186687547,0.761161566465446,0.5902445012760842,
 0.5603246803753771,0.3161219432334703,0.6642906411095112,0.07587899561046785,
 0.7706448037888473,0.7910625515862971,0.8464668377840886,0.732738734778258,
 0.8713600186487042,0.09632014342444051,0.048666332297124915,
 0.059440448857959405,0.09062766428777669,0.016478907325473413,
 0.43650121618573623,0.7202922817624074,0.719954344628083,0.135520789554461,
 0.17525041109301598,0.3056121333304592,0.2182634402557605,
 0.23709624066590274,0.06821065613990929,0.5343720328277873,
 0.8456288671917692,0.2894696687675299,0.04971431969650719,0.3344386278179253,
 0.8419293678890527,0.8506470587908156,0.6559594155395648,0.414637559920486,
 0.5063767143630937,0.9312301006894425,0.11085171557308937,
 0.14145833585118162,0.4756205887441022,0.6674790495640476,0.5267569406592589,
 0.8108053796697047,0.7377173512391791,0.22121569650241568,0.3512292095593579,
 0.0012403065597610397,0.5593245153090385,0.4933718948729501,
 0.6447562699257562,0.5055240925692203,0.6617424627462992,
 0.015525616020569566,0.19411946686813242,0.6069828445283517,
 0.8228092181917773,0.8112085255094467,0.12535978270055248,
 0.039986376210106256,0.828443438513772,0.6742918342024737,
 0.42662095147817325,0.902877516606192,0.31966812078231954,
 0.027656918421541277,0.04024644791795873,0.3979596023835019,
 0.9737940731734346,0.7235542450974994,0.7383422975541847,0.2685606947401653,
 0.9554743090870513,0.7378071733709292,0.5448581413389307,0.4152775425947106,
 0.8157345634394881,0.6511960758060327,0.06862394789176152,0.6236674730874453,
 0.9024443676194498,0.16925577571730255,0.6937199555638814,0.6411871974281896,
 0.7120048523931833,0.46877386678985167,0.07534096061253814,
 0.24597682395067255,0.6276209392169304,0.7688362218203393,0.8256431362814673,
 0.6299141941997191,0.5110224513673377,0.3620289093304232,0.994531367694078,
 0.2385367590682963,0.17208739264369743,0.8471001448311327,0.5404786304317375,
 0.5739597005379566,0.7763740554844687,0.02230076611074916,0.4686333363754811,
 0.4331645328858432,0.8500877759639759,0.9814272172581671,0.48333170564689787,
 0.5063830303012279,0.3518274643540915,0.7552924050913892,0.5034008191136304,
 0.39057400911948736,0.41292950316204713,0.4111432157836594,
 0.31480406861701304,0.9683872220111233,0.6676648878151654,
 0.05664284183216339,0.9401212033306104,0.8641641100658964,
 0.18163233752991192,0.8241794070256799,0.6111403202007905,0.1813994069132593,
 0.040892148926515826,0.5580008226956622,0.33042041023929425,
 0.37712381074254864,0.28432724954025324,0.45536458241130573,
 0.543921976554215,0.04103497749261442,0.3747511258473939,0.3358922310617919,
 0.5706636488054733,0.4300501117984573,0.3544451159625098,0.19739108669635552,
 0.512541016252937,0.7525901704106874,0.12624749985766337,0.3974246761386515,
 0.6221906169916257,0.06138883249747651,0.025738262890598995,
 0.9389061313259198,0.707050001218931,0.1487461142655898,0.15398895735337925,
 0.4684285652731105,0.3856924822387041,0.5036991926855078,0.8060028807713328,
 0.41922234535787706,0.3597647977877758,0.5284422094990421,0.9034604686672723,
 0.72216175164343,0.1250028306396349,0.045349772382413556,0.7017412544880988,
 0.22970276081929608,0.8073222891045654,0.5493098140881624,0.810295910841421,
 0.6529235554731784,0.7390738326645596,0.7775211699893545,0.00701612063706325,
 0.3124564447074154,0.679422504398044,0.9116629627394838,0.49613775088025425,
 0.30067236912844386,0.8115139088435485,0.9135632047721778,0.6481349923228249,
 0.17308118688675578,0.8775004780886094,0.33634556200767896,
 0.9679159801437477,0.8448420605349966,0.2908647727011726,0.6957472140248354,
 0.8141230196339962,0.04393201841803218,0.6339790646857254,
 0.006675290580564042,0.43718434962472996,0.25208721734129147,
 0.6862484261642542,0.7024744318282425,0.6852811159792033,0.4825238700193295,
 0.7748981604887282,0.3013717525708167,0.7333039688217793,0.1548821949027911,
 0.7475586069891322,0.9546655356469732,0.4170968121528791,0.28305162786073645,
 0.3079037980846825,0.04931263836183275,0.4035103482202591,0.6691858834192674,
 0.31625854269603737,0.3532873721377954,0.1268245883047936,0.6042987036479723,
 0.45385810770982316,0.8023531734278425,0.5853684552515811,
 0.29766057441536253,0.5475137801605855,0.3774199269378322,0.3345792712080745,
 0.2817145609938013,0.0739056595517098,0.013755445091947927,
 0.7313067096973523,0.3917917857310519,0.1717228870382409,0.862327277063661,
 0.9799359086632294,0.4703875684974237,0.7355908309724443,0.9963195830048459,
 0.09863396005914493,0.8334526343973597,0.02357711568098353,
 0.9779729122299038,0.5817811082655768,0.22237905150792792,
 0.08322823486223041,0.6671201129427791,0.285770801379548,0.22591233978717207,
 0.9910211614777522,0.8276124447684234,0.39089383593622773,
 0.15727336119880575,0.496611487966801,0.7131771806597316,0.9936640478400184,
 0.7145875366993718,0.9154261147929811,0.6738022734438122,0.16666176580994296,
 0.6496500298247583,0.9125760585235556,0.8610265989614422,0.6421670841072195,
 0.2337591208605997,0.6525536981656687,0.8684493862072099,0.36486836797217803,
 0.9775730694944514,0.12929600595087865,0.8576957619949177,0.4348101690982489,
 0.39458081980893533,0.5290342918981209,0.052000166324368435,
 0.8134113957352435,0.601734551739053,0.04301116700759067,0.8713457716624096,
 0.5288213517329853,0.17683276570746687,0.7679241982244225,0.666415365700051,
 0.3630536922799149,0.2780225948022247,0.4569233934703687,0.4871975294086267,
 0.27073411807641046,0.676830134265109,0.7702291424566284,0.4646569705424164,
 0.28877406743374534,0.30694657252939894,0.585479030885009,
 0.09812953430664673,0.3666463006787173,0.9049008040330387,0.4015603318752594,
 0.7725233908688598,0.6630127127163609,0.8674099063024971,0.6634166321939066,
 0.2668493264063442,0.8368631575548906,0.2951909259343807,0.27570084861308575,
 0.7340248431778498,0.30365629295508245,0.9320597111086568,0.8895045795629913,
 0.6799171531685363,0.15550816446712,0.8042584235526046,0.5523042603203683,
 0.32979696274159553,0.6667287919593398,0.4487439710598641,0.8861434722674909,
 0.9557304110476447,0.5089078149624547,0.9740427242533675,0.7957079801313846,
 0.9265015776435441,0.7866151079061474,0.5960353687024909,0.4475477557890746,
 0.6194687839954952,0.05444435681096138,0.7259718322613666,0.8251079531044592,
 0.4745917170820465,0.29350482337408734,0.393619461325982,0.4457308740806405,
 0.6887856173022839,0.11660002483015619,0.95092596708642,0.46411599212738097,
 0.4266950059442225,0.08748414058226461,0.014216834149120339,
 0.026493859252101792,0.4020093583733785,0.3606072175845497,
 0.9388387007540273,0.12448433476664422,0.6824929495902613,0.6393083431735743,
 0.5545505159759538,0.888515289128659,0.7556090210591103,0.6267631371636962,
 0.775677277658194,0.292102316340097,0.16253253114428035,0.6109345424130819,
 0.5542737193261668,0.9943613524011404,0.7517854268643249,0.8862359088350648,
 0.758899493707324,0.3984843310115729,0.03591248324959195,0.9990048793072552,
 0.32317520082087603,0.8980070909077633,0.4273796868661678,0.7187352274076744,
 0.65054938522809,0.9054400801487024,0.5258338586725214,0.9596055904677071,
 0.38591612681522913,0.4905866945716808,0.25940540875185936,
 0.6640833632341108,0.5272845248759711,0.5233681015250483,0.779977834609277,
 0.23934293152880137,0.4296596616439332,0.3290681789247547,
 0.09845975100720761,0.9346129574351463,0.4009225099984275,0.6483615615724742,
 0.47498767034556544,0.9220327755469,0.2400925801701329,0.8887926595710863,
 0.15568999819122842,0.08284742647671517,0.7476619766581806,
 0.7206118600041183,0.8989638073003201,0.6099415571035174,0.7324069576027361,
 0.43914390633435696,0.9406108390164365,0.9991434020957903,0.5962949001777935,
 0.7733018788590256,0.237510273009917,0.8746242588746539,0.030965848327189516,
 0.3605305084888648,0.32283200554323466,0.37483917242666764,
 0.38868432037477485,0.4416220161290171,0.8141227855353065,
 0.40380846636047685,0.4124887104617313,0.4356621943877209,0.8573793680561246,
 0.08881182175568203,0.04080606348460103,0.41013279559401294,
 0.043220822132078784,0.23929248709249962,0.5713506986941661,
 0.5104671703723623,0.16647483945259633,0.7084071325782839,0.7800390610825804,
 0.3290095999220599,0.6115582743857477,0.9302005651883638,0.43616752879803755,
 0.6421035146059768,0.2414927946927241,0.5638375201540575,0.31163922548111245,
 0.8518082454811278,0.8384702182213122,0.2427117556645284,0.8799344858637806,
 0.19365819949651886,0.8106738747275997,0.4026183431275039,0.2819800792812037,
 0.10873183881918036,0.4626174388291393,0.15826654384922678,
 0.7784204897000518,0.018672949387009585,0.37251072031785193,
 0.41271492597918347,0.5258348095633724,0.37028416908076955,
 0.3668833899628787,0.6512297283290331,0.875240818837163,0.45088974134184334,
 0.7536402066146978,0.4910798555224791,0.8077023596682567,0.7930861110580009,
 0.31642195950945917,0.8503817764434123,0.9607917803430654,0.5016038763572713,
 0.5073965846361642,0.9978190253661289,0.11649120953096515,
 0.33034584466427286,0.13295384935359122,0.7441894085337115,
 0.9701451473404128,0.0581336539132179,0.0018970955712882542,
 0.10374316585836307,0.7787837957602642,0.1793927495955252,0.6689879643789667,
 0.9370478108284742,0.4444062476897892,0.6777893308546454,0.5891568987920247,
 0.6103186446796456,0.8582234185784012,0.4189100614050054,0.4506574418013798,
 0.6377579808492905,0.5079868768356484,0.9591624316748165,0.7002109254455964,
 0.5116715019791406,0.383125103866387,0.3729749616695357,0.5356545942198648,
 0.15027806244207087,0.7965414497975422,0.015755438443798564,
 0.05527512708265714,0.5736163371904095,0.9442895582374738,
 0.48578166198973005,0.627084653433022,0.3634878065250019,0.5525574392597598,
 0.03152384248060636,0.08108783742891357,0.7536587393257045,
 0.06996480639377478,0.6592481747332491,0.5335488991104576,
 0.23287631439207546,0.41712239136338436,0.9759403358137384,
 0.9553687738516291,0.8638050245627773,0.44019537402771325,0.843287766991659,
 0.2969116906113092,0.3025759489518758,0.7262996308810821,0.2259942708928715,
 0.8171549438806066,0.25292567598512183,0.7700956538397208,
 0.21361214404767936,0.7976290579199886,0.5099619734348798,
 0.10342622620642206,0.58319314594981,0.9169817303792315,0.22370255169734987,
 0.8418788024777228,0.31156361721103043,0.497236817952349,0.5671736750968068,
 0.21326021235852277,0.4254109509852837,0.9815193364406803,0.9760239545316879,
 0.9634306725565354,0.5199459647193764,0.6305683576766115,0.7811876921910518,
 0.7557817687010978,0.43347307255041767,0.20019198992292997,
 0.5651436261743725,0.7009649272378302,0.285341480196014,0.7351030570709308,
 0.35344086640798467,0.42409672939107523,0.004184773856241941,
 0.68542054790508,0.1600198993332842,0.6187528246071154,0.1771034444320595,
 0.5513423771930553,0.7049393146466327,0.8390062304857624,0.9094708128742062,
 0.6910003293973147,0.3890056424948346,0.7058358946089045,0.7793876621346242,
 0.4217049303740059,0.8793490749183416,0.786436239780065,0.46311646273230167,
 0.3384039467400586,0.409547248847709,0.85470283408135,0.14664317584256836,
 0.3308990512979082,0.2480388116524218,0.7758882290427787,0.4017626645207256,
 0.27106317906922595,0.5186694678406578,0.3276983307609965,0.8679120730061177,
 0.9381509974222082,0.26530113632541463,0.40692560993086,0.8067262886700125,
 0.24687892907630582,0.38165515250717463,0.3268875434967403,0.943014269144633,
 0.9398071391154501,0.5096453457709524,0.8763477113247543,0.6931750598272903,
 0.21852163711171846,0.9141470644040801,0.8066966392636898,0.7659965802462362,
 0.1413846350325284,0.4858135589606398,0.9910758563499984,0.9509591603349956,
 0.37093322701996045,0.7821712029697865,0.6267784003474359,0.5039823353293813,
 0.1592726554957924,0.5939907954183247,0.055868811118902184,
 0.16375229691014404,0.8092430695528059,0.8617510134064782,0.7221984149650656,
 0.6790212368460513,0.4522311658051007,0.5979338244485272,0.3675380656840349,
 0.535811462304935,0.8450344366156034,0.23946952920143894,0.12798846883275705,
 0.6728168826883973,0.7347479293254775,0.7595561621114009,0.8567036166632626,
 0.1870037135705721,0.5247609668901553,0.6265009363022467,0.03616656507241711,
 0.06280253881743625,0.9237486147320503,0.6207954698234526,
 0.21900664213839882,0.4423729786204659,0.6353245978673073,0.9447047887079866,
 0.5949907942763778,0.4795278001841212,0.763072702877124,0.7833123402536699,
 0.06758187495455781,0.3682149946415548,0.20424706430404127,
 0.24973375147300536,0.648507196731559,0.6629033251191426,0.47453825906091074,
 0.30766338441281316,0.1466940190312982,0.35849469729531436,
 0.21805129271227752,0.4889886687672106,0.559555365698889,0.4172072093640704,
 0.9858118089821024,0.6846933902969337,0.18679002070869755,0.459437940462458,
 0.9660384301870047,0.12296346761924826,0.4834718375037941,0.8677446578685292,
 0.292094953220966,0.6387443301291175,0.8630270962357531,0.6284349569468214,
 0.8881955956342692,0.5692777862934907,0.21397290215190823,0.6272129226550726,
 0.8543980202134288,0.15484437953192565,0.9306191442948859,0.6297273814572076,
 0.823958731188149,0.04861806258409662,0.8832717023441642,0.47624802300242586,
 0.29153897097087655,0.046088795054861076,0.6048475194609507,
 0.5106432730932454,0.5088111722490182,0.7392304855746753,0.378978805416759,
 0.5688435540731481,0.40114482723612144,0.6789640804678403,0.5655451828751107,
 0.8662427922509112,0.4701167674407283,0.9528084561810904,0.4217960217216091,
 0.9119678793055713,0.14119801578055302,0.40372461565557627,
 0.2361802203379647,0.6797682397739032,0.1837651833529632,0.4653426293048205,
 0.83635520566472,0.6031030004676444,0.44921033484133366,0.7903986156812577,
 0.5116666629926695,0.4597031189385471,0.27781051818193936,0.6022604844123627,
 0.2256361521883703,0.2700244551929858,0.2740049074354246,0.38995016622676404,
 0.9326235555841046,0.2442025562667327,0.3938708684793707,0.825003340129447,
 0.9868204167207622,0.7338871503920631,0.7921466245140835,0.6866177428906287,
 0.6763307550628265,0.5432334679167284,0.4670914487743283,0.5038149787793761,
 0.8504748978023493,0.31835424865053596,0.832744930873833,0.9154091033243176,
 0.7217983773925521,0.2927797711627488,0.7559820210287891,0.07346702344510458,
 0.7405176475610825,0.25163608018184536,0.410129612055007,0.7877797387295814,
 0.8331053890774972,0.5547023737462401,0.686042667823618,0.6680982606805382,
 0.1827671797023308,0.24381131518641141,0.6721724522317306,
 0.017053434151034996,0.5843331119342929,0.4726203406297854,
 0.5266802609705915,0.5794606941085902,0.03827815395365908,
 0.05124112059653141,0.14352373596440682,0.9603409694229034,
 0.3910550307074988,0.6508613325251028,0.5480353099705009,0.6029872546710383,
 0.5680976026531906,0.5191218731891128,0.7722109265789703,0.48456565895165127,
 0.5314781803539048,0.667379376850077,0.85290308660138,0.1997240798363018,
 0.1930673164151666,0.9924158182489925,0.7624217697941079,0.6853857425773147,
 0.7273186161600117,0.704909818394031,0.8379162041923209,0.0525487960132478,
 0.6411827902972096,0.48519829743462883,0.7203427870325951,0.8356981961287053,
 0.21534058985869542,0.4158379886873348,0.10631158682639708,
 0.8398875413076249,0.121462878716871,0.18527983926452185,0.36303412276823055,
 0.6106693327654689,0.7838103420178454,0.7238880570415218,0.11868350674426653,
 0.10236618497932937,0.3131119814016349,0.21627980230591903,
 0.7715486004195019,0.3248813295701416,0.5362359800636814,0.8405354921888001,
 0.6039148012824332,0.5862087917643831,0.19396201183051476,0.3937276885933233,
 0.9455007891397114,0.9175527494759991,0.6422677269547267,0.36654805561990633,
 0.7699873572811807,0.22924019427460873,0.08236856186468267,
 0.5458256068349583,0.09320136456749517,0.4583938967438893,0.6311614372623515,
 0.046793186405074794,0.8879743166884578,0.2634640751770857,
 0.7707026068091186,0.5199861395119363,0.08742940872477789,0.5096825731542044,
 0.024678967684760744,0.6879644201670949,0.8622122343948058,
 0.9930734504691474,0.6414205165411315,0.23981090567423013,0.6166625445807407,
 0.07615396948558972,0.263190627579015,0.8502090007431916,0.7705880502967982,
 0.6154996609159351,0.4532656341976733,0.04006989275938,0.7028377603038872,
 0.8027537652856218,0.5282492500753405,0.10253962021212026,
 0.38106037455380115,0.10706823872128568,0.7850025007804209,
 0.32471662325380457,0.5187681047525446,0.7022733342901882,0.2523167861236789,
 0.21997391037969827,0.035984060398365214,0.2657016261528494,
 0.5675213280514024,0.8426691265235243,0.6421486685320995,0.11039325566066405,
 0.9129320164131383,0.9980329403749078,0.9965079514791413,0.22874811878262657,
 0.40718276586362656,0.8042520563144587,0.09576129286123736,
 0.06454513701291753,0.05703393146223934,0.8908091076625712,
 0.8926060840097179,0.7452122792055704,0.8572231246565358,0.27848668014052846,
 0.18862347408391067,0.6422801933886362,0.9490087843419419,
 0.02803935328486562,0.6366998279701217,0.03316883290419881,
 0.8256780139251124,0.7399508412860657,0.42437555577204156,
 0.17757426570067658,0.22140890254144763,0.403171029098765,
 0.16557370564084228,0.21820316860186462,0.17311306185268405,
 0.43459832454615765,0.2832506182938207,0.15061399050185087,
 0.7941526865388203,0.851083251785421,0.7495410107742695,0.08290683915482766,
 0.91307225636665,0.1612153747205648,0.8555979680262865,0.9737521496260441,
 0.5645777590444321,0.1786992303811451,0.3144612369309928,0.12008410138438075,
 0.08630226029057142,0.41600691154398795,0.9792799199185063,
 0.9933509038587491,0.9932333722698872,0.8854135963540879,0.15973761285164523,
 0.8724212282859215,0.26856368419877086,0.6402276484791249,
 0.28420351673285793,0.6790918143531867,0.516360936519035,0.23406962026568723,
 0.835081927293912,0.917173849484284,0.3263464938185604,0.031556477249473414,
 0.2338608001361031,0.9032944412902244,0.9835580773561923,0.1750980463967301,
 0.14558394224156201,0.6449419455440119,0.6688940419140377,
 0.09814915323195617,0.4795416010136502,0.25641800935083603,
 0.4715462013675079,0.1343540434725402,0.8843685200265514,0.2708491761693089,
 0.9115647090049334,0.3079287721745122,0.4931456369395407,
 0.056637356824404184,0.9143716251886351,0.3143571333300301,
 0.8233488696070674,0.9656023855141358,0.4579689494739093,0.1493788973685305,
 0.22144066239316995,0.4986974856670654,0.11534839501910688,
 0.6408669518880196,0.5569197740870999,0.6910136175272652,0.5514780530459488,
 0.7980109231898773,0.8702539602445707,0.7609443745549681,0.7010576944405619,
 0.49228593537887366,0.06021180706787155,0.018141745990493896,
 0.6790512745816417,0.07020960113869601,0.8406075937400079,0.9215724340493892,
 0.10537324055321506,0.40023156106781865,0.40024927438522284,
 0.7828474479415317,0.04502418791497187,0.4515354869683246,
 0.08257837202581697,0.5689214729972144,0.7307884040689073,0.6085015803547169,
 0.7115920399249169,0.38377654448727583,0.21220300627891509,
 0.8797287895821367,0.7869440707102263,0.25968079365366636,0.7596273217945174,
 0.04771679164195841,0.5393077364977225,0.5345712921538395,0.5404428788364894,
 0.7292891353257193,0.6502721704233868,0.7540866781309524,0.28696205722365165,
 0.1167806843460496,0.12026418267477124,0.43201760209679896,
 0.5949306272477647,0.9192202994729111,0.5465593706040338,0.12579620417946036,
 0.8765386725363006,0.2968722604336623,0.5413102157354835,0.12990811995523,
 0.5287480966332343,0.8988371368773347,0.8560417644708181,0.07998699038883172,
 0.19790825174687943,0.39760728278318314,0.11155777452544768,
 0.4476694029786228,0.09177292410859861,0.06121831493169749,
 0.2327523129225636,0.09189800363680012,0.6207377245770119,0.7166870256051753,
 0.9856451654149656,0.29761709012949855,0.3061524582433406,0.65088879553875,
 0.11782558097130014,0.9953418333498322,0.05644240806173173,
 0.44718621222469357,0.05158742856808818,0.4446132336423205,
 0.3250628426128198,0.73391550787618,0.05516802691921474,0.556867123182557,
 0.7129518149454792,0.3060945040371612,0.7951228621677353,0.7057116088854262,
 0.3719960645500673,0.2963822537406139,0.9229417191149135,0.6195136664264586,
 0.3460239862238862,0.6932789400903252,0.40341202816716504,0.7874378227240731,
 0.20795180506242383,0.38119742412084745,0.6280891313708713,
 0.10056650755550955,0.016705460638996805,0.961738553099281,
 0.7568806479049828,0.20490357598567377,0.5525915911900211,0.7953284331443142,
 0.22594012866500934,0.13129187089911087,0.2544185047892089,
 0.7271131123819818,0.18062454021099883,0.3607938072131668,0.4297184619673058,
 0.5660180709138106,0.5546759913663983,0.1146550979443538,0.5442802479370383,
 0.7057594031785663,0.3894421107393359,0.8336035658964267,0.7066945410609293,
 0.21607486959483657,0.00771944325901075,0.8025903577573467,0.50463019555754,
 0.909170160584933,0.9160043479568055,0.5864042791389948,0.008375401941934535,
 0.44953779165935914,0.1657931063811654,0.5169602823884863,0.4697832640890751,
 0.43190943847223806,0.5489710070848611,0.173568008635822,0.3097490098198966,
 0.906713334148865,0.16466948422042393,0.32449513539041475,0.9480605403720708,
 0.6967442657280795,0.6006881979539211,0.7638600634357312,0.8584575164317456,
 0.6823414423540766,0.2077805210016913,0.09836846939118404,
 0.19680946172604452,0.3026016742440114,0.5251384275909017,0.9973776983350069,
 0.5364752139557954,0.8102310267591909,0.7422447478346763,0.6418672703630806,
 0.6127018112590717,0.8076302841447387,0.10422441255657588,0.6356448075584564,
 0.35365507121284223,0.40676677456566446,0.4047262726294113,
 0.9033271289273022,0.5459635824943991,0.7615516678079023,0.42506925105745974,
 0.41498225169848413,0.8051241984769101,0.8244574673788443,
 0.06539248746567816,0.9998868797895575,0.6298535855913807,0.6083586820250537,
 0.061777095840943264,0.5875356626037973,0.639996966136489,0.734451513777521,
 0.11936480579501219,0.5072755254986533,0.3232347545477745,
 0.42819695230866617,0.9084056677792997,0.013638442100219583,
 0.31611079725328506,0.768029612974963,0.06251696058028078,0.8447830778800712,
 0.6248644036251063,0.2077641971693427,0.18723721033774632,0.6536800813277966,
 0.5915518845897674,0.6268263387692159,0.3355547112566587,0.48020673761736554,
 0.5553686329163745,0.30041512662974024,0.6240398770344335,0.8912314272568544,
 0.4235401105446803,0.39044672511200496,0.3557004975816873,
 0.06458870703967245,0.5670270884991969,0.534922799831777,0.6914603533478677,
 0.12641101949028954,0.5124574614324313,0.2799460607675853,
 0.04701501799519514,0.3643444509611219,0.5422901240069052,
 0.13520506968296664,0.6317201919423101,0.27984746274333017,
 0.4738106929214757,0.6747919438969294,0.27557147971163787,0.941095334410538,
 0.8209839657040499,0.5806836125388731,0.20993861861289964,
 0.04730322076333204,0.9063149349815766,0.34472848077726415,0.456166364990512,
 0.3035151547869641,0.342874948257178,0.17823094599378342,0.14550538026407,
 0.45069511212945057,0.6816929994325552,0.6736211793246797,0.455602213622311,
 0.5903473393211924,0.6765406491840693,0.3785366523123923,0.6263287511300417,
 0.842333770033186,0.2098510460189229,0.9554781752321104,0.5620536658253703,
 0.20723883122964482,0.5532205118896945,0.661948717854835,0.9281390441787981,
 0.26360807932088814,0.3255348489608425,0.5285166345714909,0.2867893859988727,
 0.053561085738454306,0.5742074145107527,0.5993872392258838,
 0.9865372475870949,0.8289620922845589,0.42424566515175766,
 0.15663224373524287,0.7973972746667137,0.8213574525039021,0.4868374115914905,
 0.7297067299075046,0.8715361816737275,0.01975305832860219,
 0.18800932873809684,0.06889953724252562,0.9158694934659548,
 0.46672534184667935,0.665487039660654,0.3388119589153399,0.5764277296273418,
 0.32422263097214954,0.7920813771265248,0.6570429303811686,0.2913737302370021,
 0.3142463989218265,0.8435171652399559,0.04585970969256137,
 0.33580597852518435,0.12847477086941783,0.09747559572661602,
 0.4988500072151203,0.9735032328654423,0.5600177004449264,0.7149258854398479,
 0.47527462475123416,0.9738900657913911,0.1309740293052386,
 0.05383235149334398,0.7194894146052936,0.4577861646438346,
 0.12689031092890968,0.29730044038190284,0.03853291641725087,
 0.6728520530805602,0.6914803496128089,0.6163039523101324,0.07677725626516274,
 0.4654199104867647,0.6599746934680516,0.4757431258631194,0.3329334714899177,
 0.04749164702700692,0.26832103199219237,0.9212230140565683,
 0.42217435491156263,0.5646609142244553,0.8918066739438012,0.7391653164647037,
 0.9292638606373722,0.08825501163736627,0.44409523767878345,
 0.5312030719012568,0.578273302540241,0.25983318440733305,0.43029923063340636,
 0.07214919688945887,0.7266883655558054,0.44633914140612196,
 0.9650712801228788,0.7558956418843882,0.45771045070670036,
 0.10258024106638253,0.5887392543217748,0.10253146928937185,
 0.5452217080775101,0.22542474906916488,0.00489317937670386,
 0.7931754134844942,0.34678337174545104,0.36396375823835436,
 0.5807161237377118,0.7988288042078209,0.18111726509105908,0.7378871743528694,
 0.6002377771446984,0.9475514811070693,0.9570442178442463,0.29807129317844727,
 0.2495625679370802,0.27361667761829433,0.8039860540628904,
 0.18253017601509847,0.15230559496007434,0.1443224555468381,
 0.5915680346878912,0.6305344364874493,0.8425113304612677,0.3963402211406264,
 0.4164533231420193,0.08578765834398416,0.9097987628177202,0.2278120688557761,
 0.4502417285969965,0.5508726292832968,0.02926813505181658,
 0.055540711219409955,0.977686299622415,0.07108846199871188,
 0.7204046646964187,0.9897778400239372,0.9938253720363731,0.22648012298927955,
 0.1966962037081874,0.43128368381831705,0.1652307602811912,0.908161199519067,
 0.33503934661307055,0.37791272956885125,0.4514244314177156,
 0.12284173751445893,0.3470780462595332,0.8908610593043587,
 0.41520478387225435,0.8649122096856975,0.4649551298648754,0.8921007127231841,
 0.05756825139683319,0.6307656436527223,0.6187984093997402,0.7828657419132461,
 0.20038477841762248,0.34835423165265844,0.292404633694756,0.4936154171618985,
 0.43650065477342603,0.6197130295760862,0.6797202565362742,
 0.26255102452138257,0.7302658504179532,0.2121724680405387,
 0.029939223052213193,0.1852513397712745,0.3296266766426642,
 0.7834927001439458,0.3831486827125278,0.36970596802814615,0.8029892799183671,
 0.10509093104113965,0.14937671149863896,0.6333315314859068,
 0.03861906021053629,0.9461643220352485,0.6344186318434798,0.9311151663487371,
 0.6802780564968142,0.6521374158677595,0.72582427950965,0.46888935497398265,
 0.24450708672509402,0.27890669461191975,0.5194227053459936,
 0.44851125132556924,0.769892498182371,0.3560665376714188,0.35961987763523395,
 0.3214236510797104,0.14470020578628473,0.24043186967381447,
 0.30299188570255264,0.33050614647725396,0.035810070675674766,
 0.6964134276319042,0.6444415492979854,0.26137042515847786,0.5749568269931523,
 0.008463354762049513,0.5976173462528221,0.4746201579084035,
 0.9963715365314547,0.13960586699021693,0.7434250477436652,0.8464524936852714,
 0.28077082440560197,0.12383005531473257,0.8385363669767727,
 0.40392620133279256,0.39365324802685264,0.2260940822567482,
 0.6924331389044407,0.43818133607828236,0.9110569026983305,0.9232658770463725,
 0.8142089282391126,0.1405606729616925,0.7737997328632431,0.8744502727105113,
 0.5645207957977499,0.0019090883304064832,0.9255268825108024,
 0.7655783186461741,0.1416987936857026,0.4313461595550766,0.9022408076432162,
 0.7592272652624906,0.02198055633176177,0.8704539457886263,0.4062550634148424,
 0.252666772743392,0.15033276432268494,0.6580107380035509,0.4137265706955916,
 0.5947203883510855,0.5200237964306815,0.663817842634013,0.8161248029081211,
 0.8635851778574868,0.9163133383930799,0.8961871071142573,0.3168030382597543,
 0.9971767775830159,0.9518395837234068,0.8548339489699045,0.6318867691435811,
 0.6450005478197006,0.7443383647766135,0.6392714217030063,0.9427083362501449,
 0.8080642720883923,0.061623730295192414,0.3047246263935759,
 0.17091022578567172,0.7755622508199587,0.01877151222772322,
 0.04960241959414735,0.7871458070195413,0.31489070323747004,
 0.2119831178454762,0.3167428615533474,0.9677503971312513,0.14640995039145832,
 0.11134729142561162,0.8197991026746614,0.35195878084349763,
 0.5601159761656168,0.5602519431826516,0.02370102671203911,0.745704814691138,
 0.8431117963270275,0.30163177944173314,0.6598741483466135,0.5649281491203038,
 0.9161670100979381,0.40037816843107765,0.10537144486362937,
 0.2210679462414008,0.21315479887627375,0.13128294164226317,
 0.4746302275084976,0.4319104487314668,0.5410531944628068,0.673537164509902,
 0.7002364254996184,0.2773050524669475,0.2155480971547693,0.9101239663508821,
 0.4188371847221034,0.23063213533499238,0.3331313441357634,
 0.46413034721297297,0.02217775400927846,0.36054827570631365,
 0.12824765362385238,0.8623705000492634,0.06563045916006227,
 0.5304766339875289,0.08730180287254163,0.4125837998310644,
 0.16599802352089588,0.8660492211850388,0.5630001198607797,
 0.33510104524363693,0.16793303380947033,0.7253658115170012,0.931491089880892,
 0.6622476993684602,0.9703750763583889,0.43606212791413745,
 0.14750918395611679,0.6660077366591564,0.20765991479824564,
 0.7307852833057371,0.23488881191356203,0.8428273749672041,0.3689282708797199,
 0.47731305540929414,0.523174779763599,0.4136386272785999,0.31730323184608733,
 0.3170825719536918,0.12334622354848945,0.41958030801449886,
 0.16070608633021977,0.0064778428096272656,0.3957745644646826,
 0.4006761871078972,0.8140668461632895,0.6412945998475834,0.332590219276533,
 0.07911459820214428,0.800431964355361,0.1356069514731436,0.22129237295145054,
 0.45168977527407506,0.5449119336802474,0.16055136549488647,
 0.20027441921691547,0.98227203177311,0.011328145303058368,0.2632996133282265,
 0.026937535083524367,0.9443294665822829,0.9592383255413877,
 0.15660556520517954,0.5078373690589936,0.7362271483807633,0.5836251511484345,
 0.7706300508577607,0.3781892040326805,0.45444877875860945,0.4857715291883027,
 0.08440369987613583,0.0018615284791573261,0.19016777309638666,
 0.6788612706307524,0.4640165398304247,0.6989356845434057,0.5560515290642258,
 0.0995104884318595,0.20060197620693299,0.19029440663689579,
 0.12425160746462749,0.12251645389754007,0.6604093330022955,
 0.31993520404159015,0.5882683738791751,0.14301126395281094,
 0.23632025102760434,0.014390473539967252,0.1341594191768276,
 0.09146364412481889,0.37999205364273003,0.4440938385448452,
 0.06342467792584883,0.8279703293652738,0.590934874594097,0.11124631466719159,
 0.8075769557867928,0.3911962948791168,0.10536964476856192,0.102459654476863,
 0.8664241402033024,0.6735907384309526,0.5380906683009724,0.34927720724875766,
 0.9162414820085277,0.32102436999654405,0.5157573731309874,0.9351836783113837,
 0.07182522190710094,0.38100973267920524,0.9233269795769375,
 0.8155048128389132,0.42630853144216796,0.5893657831453314,0.8400658013802845,
 0.07799718531689948,0.7484583380179749,0.6500719935095702,0.8357739852859087,
 0.07343554910490202,0.09598669502882651,0.16576638496263552,
 0.10035676877873789,0.7730357718491899,0.01744335168732214,
 0.4678353882730162,0.24152020097122207,0.19947619885545143,
 0.8030623050358208,0.99358812070203,0.7419111875282226,0.7730799861873452,
 0.8788408868474673,0.7631982645174333,0.77600311773086,0.3830990221953552,
 0.2079832201856867,0.21491131243065642,0.6228642331893093,0.8271735814350303,
 0.2104067721524494,0.018684569683486307,0.09315741113382336,
 0.5781522898738851,0.14272168034110322,0.8125354945950386,0.9408897713769413,
 0.578468559354344,0.06885803373653654,0.15927942372299586,
 0.37065374474090584,0.16032544801930493,0.02982480620704997,
 0.7880147434270948,0.3410609368135693,0.9516206202854471,0.8884732687999213,
 0.09158191014997796,0.5355106833296551,0.9096659857569153,0.7739473532793584,
 0.8224120467776077,0.3267527157507011,0.1752827909911483,0.6851148200782246,
 0.4890361845983109,0.36181908094546444,0.8627325153796243,0.9938721676018263,
 0.2615758143790802,0.8210935425289508,0.10215474084474718,
 0.45844102755612137,0.35430310809843857,0.19141892561541474,
 0.545239204382426,0.292005385464999,0.32636058701559556,0.12892898793681884,
 0.9822309718190718,0.9813409805317106,0.21207332626705255,
 0.11133793325741803,0.6614090879616427,0.8822724935381924,0.4169235966575402,
 0.7981413683206467,0.8127928463043242,0.4727736734057777,0.3407408248708349,
 0.4699299451728923,0.7063264630414248,0.2317182756049292,0.6978079030554843,
 0.7491869490492432,0.524345915898383,0.6831535367664956,0.759840054511394,
 0.011713973341229922,0.8043110474262354,0.19334712845080682,
 0.7505312608840427,0.1893573409900634,0.9954851338147769,0.39660419182448414,
 0.3375272769132083,0.7773741241197648,0.4374285388681589,0.4135364864391037,
 0.9522037870774794,0.21615364622330246,0.6960068158881059,0.7757579316057018,
 0.854783153988319,0.8935304205763099,0.9497495865812883,0.40084977655360365,
 0.8921639163523903,0.7335595220443855,0.5754044757663385,0.22133715790456332,
 0.20876748138960965,0.060943489493416325,0.6014102774605571,
 0.5963008829336454,0.7976750595285728,0.0524992230860265,0.20757207010467738,
 0.9334598616783142,0.14522804884319562,0.24663621084536003,
 0.21602254725712566,0.13676572647437024,0.17792381196342133,
 0.2708848044112582,0.2775374348137465,0.5829191596577866,
 0.044227760086953394,0.5263490661002715,0.1929132528123575,
 0.16793239103434354,0.7196507718377203,0.7497566418226821,
 0.43665831251699927,0.01012366142819543,0.963095518305953,
 0.19531620408129613,0.5900793109683469,0.29015505150763554,
 0.28817655629157857,0.5601679365129248,0.1811599591011701,0.5258838447924413,
 0.015850142147652346,0.07481628075226288,0.5212004116601924,
 0.8764189857347214,0.5567162409944827,0.7619630802895405,0.8055821438834577,
 0.7921248594541475,0.7984488156112459,0.4086499456451379,0.16330967216626535,
 0.11409276687197745,0.7992852040628277,0.1516153026503565,0.6897164013284393,
 0.2538899934915205,0.9624161111380023,0.18129815876978672,0.6324316196870831,
 0.15141225084598997,0.28180978674068935,0.8488203247919065,
 0.7456080829727433,0.4722086167423498,0.46513241219185164,
 0.10223291550134306,0.6915943324662529,0.8269991273729841,
 0.21648407504756884,0.658431671034808,0.12309080472004674,0.8560258399794107,
 0.4774460957461736,0.8429011225562976,0.2339296038034666,0.294310227497974,
 0.9223959567284716,0.3066942843509304,0.3026485608717313,0.9356742368351575,
 0.0014846244202538772,0.19843322245908046,0.17739756077502822,
 0.21337805711460656,0.5796665274108006,0.8145154827277132,
 0.28536824163151187,0.2550223169715341,0.23134459286611853,
 0.7768207070851241,0.28462426530299556,0.12281564133001022,
 0.4358376013066795,0.6410853736699421,0.4585934595257828,0.8577336524250545,
 0.4682025922802372,0.5151860410597925,0.9274651252418377,0.9793422812135728,
 0.05457945753773363,0.2295415717381757,0.47328537685875793,
 0.9673677758657662,0.803137940002306,0.28626226647774855,0.31415202937093845,
 0.5817015855178923,0.542994574449837,0.9108979475388828,0.2646495488263815,
 0.8500264925916579,0.7334708309548035,0.160238358771568,0.5716450951169761,
 0.29665609397097503,0.6475886090471477,0.9284605158096084,0.3965496007282493,
 0.06483314080595182,0.5827287323963756,0.5864436421137337,0.8934626710000595,
 0.11813127377960497,0.7232126501231638,0.5675391176738387,0.867147091847955,
 0.4756866582965458,0.6602629004316016,0.12939906589183248,0.9771042198054504,
 0.3005904361400641,0.8701326263167947,0.9697122365507953,0.862922461099881,
 0.5449740195419839,0.6394494123281523,0.9028372077644271,0.5242595883151839,
 0.5094926122893451,0.5144556771453207,0.010337258912607616,
 0.4746457277736005,0.5315577934079208,0.6135638959817294,0.225361924140154,
 0.6943091354398846,0.8249242986246619,0.18054256480377007,0.5211213884301468,
 0.6433955637507403,0.4707628547315156,0.15038156302564964,0.3253997295407589,
 0.21912317892610644,0.08990843794035208,0.8260589183803819,
 0.25362937168406263,0.4186103310370132,0.5128381777950007,0.9921721862656179,
 0.1447481994788371,0.5225014361204355,0.04308269039506718,0.8996760566178601,
 0.5348243266600176,0.9280348846662871,0.8603651731914733,0.4430670937111899,
 0.3814616358936266,0.9623209128807722,0.6956525721835408,0.7514706530107839,
 0.26978981049456874,0.5533362914943203,0.5968428494317074,0.681983185156466,
 0.047660773719595984,0.6889727435669696,0.9368275191439172,
 0.3549013373633688,0.13766932730201864,0.51492246947898,0.06990257567391955,
 0.3331152270379517,0.6220003330093691,0.385212353760386,0.808502470008843,
 0.5570269901764933,0.34094507980305,0.46295847191835127,0.344378566892379,
 0.4145553475249757,0.5531472931529733,0.7576714065179493,0.10798178569105699,
 0.9580204828306813,0.2811703735467216,0.040070837442483076,0.3960005575937,
 0.3267129586045865,0.17985824635662728,0.2399121546944576,0.7388331934118391,
 0.0913770276608159,0.5569219439079243,0.1185922408179153,0.24177461070313644,
 0.6680549884167492,0.7009960333449158,0.6012783443981949,0.16785221302944608,
 0.6065049546638496,0.520036570232827,0.1843731137827973,0.5148132909794293,
 0.5762214253722941,0.47489050706021174,0.9782334655288437,0.8204951619555836,
 0.9223125555928828,0.8457539745753028,0.30298211245638884,0.210759311822602,
 0.7332750043121925,0.25378684858762257,0.6867078045826479,
 0.16135200869331223,0.34659016186363456,0.8380766550290784,
 0.38305694832902926,0.9939583211346062,0.4915784398905998,0.7549596727683453,
 0.6365525229118425,0.8435877553580564,0.4306074592945651,0.21520204138601784,
 0.8255193890461286,0.5942972703504653,0.8474234094904731,0.7219216179892467,
 0.7079890652121927,0.4555464266197258,0.05223739431760033,0.882096509409358,
 0.42209520785952326,0.7223633867641915,0.18267735724531908,
 0.28883010227893935,0.7659064459255299,0.8187461400811369,0.7461022090314405,
 0.9751507034286272,0.583904436456851,0.30980356539759557,0.9906644694467854,
 0.21157679532250484,0.2226965853932541,0.7442318031995909,0.2624798170652124,
 0.5120569580499522,0.6106335378852213,0.08488028874454567,0.1928265451302802,
 0.33463565348540547,0.7296673154178183,0.5831623249219264,
 0.27086958829445595,0.12020268620016927,0.6962511415015145,
 0.5326735988396821,0.03821983708585319,0.7372543784345218,
 0.45610037887594235,0.6831154018937359,0.7442333729143185,0.9988203567632183,
 0.8599908813573982,0.003358682333250096,0.6784015929122907,0.547223779393248,
 0.6555367346500882,0.22733785723597977,0.20271899044109898,
 0.6959549253072113,0.9401481607398129,0.795074864825243,0.0809999309842897,
 0.7474405208646369,0.573457281199132,0.610809782027496,0.39106011491654735,
 0.6484870912507652,0.2632029722709226,0.4898712227986739,0.14318356555318634,
 0.4328640016100619,0.2162477344780196,0.6912509668188742,0.8527526198918778,
 0.5341636517830016,0.6059127766028626,0.17059857223538522,0.7235171239667493,
 0.23030070455444596,0.04327387862120857,0.7293612238687042,
 0.7890276586139048,0.14933388081688204,0.807926874253411,0.44644940231303976,
 0.09454303296451683,0.0791158696156683,0.750437189644432,0.4129486892280526,
 0.9889382021646097,0.5446173892665566,0.8990081531900335,0.31480050583085917,
 0.6136868601726935,0.12070569946743448,0.6712388349923964,0.8196169909365282,
 0.8938620526630948,0.6931131518221251,0.6567852317802791,0.6565163420734609,
 0.36764081995835096,0.6049686888808133,0.7400420895852884,0.7361423712354924,
 0.5460690709137015,0.7734854471251907,0.626962339611436,0.46157767365435876,
 0.19735334508256663,0.07767812985905032,0.061096429355902315,
 0.052294654675195784,0.2103667250848681,0.357910937008574,
 0.08295346321031416,0.8353834382300438,0.2463456082124249,
 0.45837606204129855,0.9358904229398186,0.48373210118781396,
 0.15448752099443042,0.12016584560403343,0.9484736130181853,
 0.6743503781409301,0.9477115424087859,0.8670913794523067,0.08784545773394548,
 0.33374045003145814,0.11245598548647262,0.6212170940560441,
 0.28689073286607913,0.7062648145490013,0.18458135826258415,
 0.1701080198152598,0.8767263049455762,0.9749263442831614,0.1178503819687281,
 0.6406098943443318,0.17705188291785723,0.04897277960009783,
 0.32090865339293095,0.1835837385673732,0.2451265829127498,0.9491422365292966,
 0.9201293767566229,0.8531146793920608,0.5261893812868094,0.36129153871016384,
 0.7317163568991508,0.8113575617059858,0.41001727034355284,
 0.005043942310722027,0.409338081577178,0.08608953453948853,
 0.9247525434722179,0.09465141030635293,0.17497223936098794,
 0.9333962473509934,0.45633447024915974,0.75945949726854,0.28386351909277163,
 0.7066458593707162,0.7218844526158716,0.286535968263431,0.8933603710805424,
 0.20582518087678492,0.8190887873161109,0.7820987861920035,0.6847051359542523,
 0.4615720109604018,0.8616586606625758,0.5691881706458995,0.8380341372402454,
 0.7863502609422547,0.4657517583373457,0.7817988518008628,0.23524966350680343,
 0.48105125555056816,0.3323021765981291,0.3994448979296995,
 0.18329912191323017,0.33194634503036413,0.5708098800875115,
 0.6780479298820673,0.0506073603135857,0.18216436775146205,
 0.21684044109099465,0.1416275970598142,0.7839472944745927,
 0.15256351606141283,0.2930619218953665,0.8739886088001151,
 0.22804109405427964,0.7247048714520182,0.8890094411930388,
 0.39488685132457424,0.39587457804026516,0.9759093475819631,
 0.4753732624327309,0.8318076664647703,0.5575742921557414,0.9686038752211843,
 0.1491072933330142,0.5086434260165619,0.4257681743786136,0.7648788832886153,
 0.7070202723665487,0.15335216524754203,0.5153259981776168,
 0.09478302767786695,0.4254612742208347,0.5115867672564446,0.4563253210918643,
 0.9004917773314041,0.16597812360391095,0.9425384210518445,0.7562764134018427,
 0.8781353443500466,0.8027835653505994,0.8077089607030348,0.7555662705865293,
 0.6058604806361018,0.2749044210733288,0.7605850415876966,0.2122732294799613,
 0.7623134480198736,0.047234207853670496,0.12753411673535997,
 0.5568700458315011,0.46962831958219786,0.4298265918508244,0.4370846073712361,
 0.33717419938245374,0.289195943361351,0.2484491912261123,0.45875484016846246,
 0.547893585227895,0.4040155103053158,0.7009678033744766,0.1331862884540933,
 0.7100655134461666,0.7690647976858885,0.7504029458654298,0.6848509164488128,
 0.03828566391117427,0.8302075425391486,0.7036994107091504,0.2852421833845262,
 0.8310332130625477,0.9345433439198906,0.2689015707646839,0.6814594403642631,
 0.6680671310054102,0.8172080596370804,0.33958169102400493,
 0.13003666588472196,0.33736113247003396,0.09249703428909428,
 0.8926695207530666,0.7111165338325328,0.9909668761706617,0.29650041599607013,
 0.7399185344649879,0.7490917068503362,0.3615825597308938,0.10695960025568663,
 0.014399245365854796,0.19658888337493197,0.7330235542367664,
 0.08605998881153898,0.9333321044597753,0.5554839317796489,0.3488951274037,
 0.8601952303713191,0.22904780507196643,0.7513891904025759,0.4896850130659094,
 0.3362478485181195,0.008722805217733498,0.13916010359985365,
 0.18737870409309754,0.8015549755602768,0.2908512130681309,0.702464139648463,
 0.6653065069211471,0.7559231546132648,0.8187948622025414,0.30225484226968546,
 0.22228786345714058,0.03274996597720825,0.2108063831304331,
 0.23744241020967005,0.0508351567961548,0.6504903019224756,0.9085517436648466,
 0.39619000746973076,0.35499472884372507,0.8285909308552517,
 0.5551317913123871,0.689143073917956,0.31782443613446265,0.7205793268532756,
 0.9738819808948612,0.6490984370304825,0.019109491953159585,0.5635732454194,
 0.79010676781025,0.14377164322080738,0.8427946627589789,0.7865061799050598,
 0.7953729081756401,0.9879055131983051,0.154758550706378,0.8130021765359305,
 0.20467231693327126,0.43880141869550704,0.16919796320474245,
 0.767322153346221,0.7398012081470926,0.9148242755344462,0.1184679737742006,
 0.0607266831516069,0.2991310557724802,0.7991934557735114,0.7601863364305437,
 0.8894338849675337,0.9355272480005798,0.982939810882132,0.15039347548050408,
 0.37935768305901996,0.3903199040473694,0.699899966279616,0.9188958807794609,
 0.17155051837682156,0.9757563766840521,0.3087457581658657,0.8084985227800876,
 0.8095229976143727,0.008786627085603627,0.49622738255041987,
 0.01911282543052928,0.3858684337429592,0.3407288316039121,0.7594427964531881,
 0.8728977942935698,0.4302774568696387,0.4208640701687689,0.34680917897466357,
 0.5879237233055361,0.38083157918053223,0.7895449698710499,0.8295199581922119,
 0.9198422339595436,0.3907686964715634,0.9603316049447355,0.8978771660310304,
 0.26043233744565186,0.5822579758974624,0.8988682204647075,0.7264336438830419,
 0.34200190896834126,0.06373707584515753,0.30168496844787007,
 0.11347449684477084,0.529095322603414,0.7721751106810791,0.9248800125796113,
 0.5295231001893793,0.8324781677768753,0.09877902276061212,0.4583719344373972,
 0.6768542803587451,0.9591108085824712,0.5218062350677777,0.6186463499188242,
 0.8444232207105321,0.6370870259968056,0.6029473404866744,0.9600384511397367,
 0.28364756050182993,0.997878533218224,0.8284294981909335,0.7762136419660861,
 0.993035750543675,0.192046099732476,0.7403614539367263,0.032288206103859296,
 0.5569908120567417,0.6105466376847746,0.7313531733541099,0.6760379943668562,
 0.5067477401757507,0.7271127682226424,0.1669535891504279,0.5791833357920241,
 0.8946341742477862,0.4630892276933096,0.970506645546001,0.8027285963352139,
 0.6626261380462719,0.8811820844015595,0.5369796185013898,
 0.054564949425488396,0.8031826515766695,0.9920597932556997,
 0.31183873456391487,0.9751705223281236,0.47162818960324515,
 0.8166338734840118,0.5128369081824288,0.48958678138262224,
 0.40776204856514187,0.6171907346193339,0.01425911037094818,
 0.2076358203189761,0.22338659803727579,0.7443105496314232,0.4496753558610429,
 0.4555409587722372,0.615761759934062,0.5253967401251014,0.42014888981444065,
 0.21295427196169958,0.4282602184919577,0.9794378094934704,0.5629747798691105,
 0.7194682406060253,0.9231806698318185,0.1466310720135402,0.5537778404479525,
 0.6974229706118411,0.7906588736019019,0.5038584172595131,0.20730743806309204,
 0.9960996725478413,0.018688703590515776,0.2726241113613769,
 0.34803340883340506,0.5488530908122391,0.4189974997343796,
 0.25695742702631996,0.38283272242789046,0.36787278060975204,
 0.07339276956468499,0.015160723271672172,0.11560547479936656,
 0.4978549884718837,0.11923432871272244,0.4048870510262017,0.1909351333953505,
 0.5471841420141779,0.17473593498516082,0.5432644227694762,0.9895691129745185,
 0.0015961188024379291,0.13132553704815608,0.6091201609369591,
 0.46759428175293627,0.10506178952879142,0.22231787695105354,
 0.49461373172638945,0.5687706944716491,0.6073940491896668,0.6444037050183433,
 0.7102425926553676,0.42910413919742385,0.797295831460232,0.9624428254170617,
 0.7981187234542618,0.694126558825878,0.49974976034333163,0.898975164579378,
 0.05929757896128596,0.5489867869280973,0.31447499240410903,
 0.6377788977507006,0.15535906188250492,0.10573110574099642,
 0.6377877015185305,0.7153857511493422,0.13867220557255444,0.397557930866802,
 0.9426183263684295,0.863636387136659,0.05874660041495705,0.7844781451967415,
 0.3872599575642651,0.8304193970091114,0.2801136939286285,0.8904894540590331,
 0.0485237418404425,0.9703397275488074,0.8344480137683974,0.37124818141298943,
 0.044193260305700144,0.5023338824328178,0.7108324468353506,
 0.9661808182911363,0.9127847737287323,0.2638308204376897,0.33183623448650845,
 0.8378604188294997,0.3962401326265902,0.20076401455701198,
 0.026408154211179702,0.6072873759957147,0.09525694715104316,
 0.37324076853955623,0.40336015696262506,0.8280692485383109,
 0.25490124994157826,0.23113228241153394,0.2143220166617562,
 0.06854346568434655,0.1626289435865207,0.8800404290065815,
 0.22876497121514727,0.6743670681029381,0.138871025421896,0.19319304008361105,
 0.0010686851506652717,0.27383031654990986,0.5707695486284002,
 0.6676189633463507,0.37085733975243973,0.16559386906509088,
 0.5195318015367154,0.6867128803799838,0.5572998128841951,0.13564475873219606,
 0.4718222086004006,0.8444100845186187,0.8510632726049808,0.35498652445409506,
 0.8119670393305203,0.7838387626758744,0.83526642883929,0.5692419885018641,
 0.9091249028734172,0.3202707425075959,0.9742201263494112,0.5417472850946058,
 0.8012570375187779,0.6729303818216829,0.6163883577247496,0.5196286547883952,
 0.5540925723386892,0.44665108928871455,0.8143237664781579,0.9607834945673343,
 0.14978093500615264,0.7636061026845997,0.6568426879818626,0.7476958826231163,
 0.4169864079980776,0.19014236377172156,0.21051897427224908,
 0.04925521667091848,0.643066738699146,0.8717885771286629,0.1679912669631236,
 0.5166722173291546,0.5725701528682987,0.5565660518521104,0.16126278478557265,
 0.8048720898723926,0.03406408075862344,0.5461276953050713,0.8285697616316878,
 0.4035207934767414,0.9871075003174534,0.3461663007704273,0.2910987639419279,
 0.2875758017349792,0.625044637417988,0.8298544771841323,0.32493475548648754,
 0.11307554284205468,0.9084107556761514,0.6109326441578088,
 0.27845775380243354,0.8075586982662925,0.853766095474549,0.15472359214057052,
 0.419559418690315,0.543283810623469,0.6401798285560081,0.1349548254501144,
 0.5478142858320434,0.4583581085443257,0.7599049297878727,0.6864326567962205,
 0.9129298396987832,0.6638670523865116,0.3273065086045148,0.31124287138433115,
 0.8535570887045894,0.2907899696437142,0.7824555980687932,
 0.046398429751581705,0.2667812777406834,0.1048716871907781,
 0.6219522075775159,0.8799610655100826,0.5432232936444199,0.06564860041929754,
 0.010538688672220409,0.5994608759929242,0.2635417608899987,
 0.06634294598758883,0.6728164943802569,0.7619289193123788,0.9293330122370042,
 0.519059809863698,0.25967874861069096,0.8582297966609289,0.9997802656905087,
 0.8540526988446636,0.46395637022835,0.6491309743521902,0.18605779631354213,
 0.2657601026983457,0.8353973561492766,0.3164961841059303,0.49348587077204076,
 0.8448676668236618,0.4233007926186846,0.4404554288295187,0.9411655020693038,
 0.15030498902695488,0.4639782451973804,0.8023991275822673,0.5312417672395651,
 0.5785378247690455,0.5576773081764528,0.9533879647759322,0.8776098954480908,
 0.14801718125720398,0.8698201608313604,0.8059127038073538,
 0.12139253261245853,0.6327286281151974,0.8560425018718236,0.1910904674358852,
 0.7527630228298898,0.2510549671410168,0.9431570215011562,0.3553280079928137,
 0.504401533358071,0.52777551384354,0.7117397052991925,0.11767167222950636,
 0.49218608759138416,0.3170303153843468,0.6296434434833549,0.8541303725182001,
 0.4124644869727758,0.9177820587097109,0.5169320089025785,0.470300178824254,
 0.7819398422104151,0.9439039831061543,0.8830261039208319,0.4550489544498467,
 0.3259403342790316,0.003451378313045761,0.25344281800020874,
 0.8882008003141557,0.40252220490348134,0.27046428613209494,
 0.6340692861609302,0.3780842004779239,0.2437914143813582,0.5855310225347375,
 0.15237417323935143,0.17627014331335644,0.031409853703234014,
 0.9533266016478932,0.40401216789729744,0.6543356012003896,
 0.16429472459760086,0.956264102218576,0.9905336829690427,0.4685695770897562,
 0.019681467700109034,0.5481944731246671,0.01631403208669069,
 0.8736035824700382,0.20296881898409358,0.8357570707105393,0.7804347929926823,
 0.4427505750606373,0.024722385838113814,0.0714053256565812,
 0.1477512145437203,0.7485886412599588,0.4046737058899995,0.5407452777238704,
 0.3012886483196523,0.9535932093028963,0.0954704209677738,0.5786560763208345,
 0.9736002224970465,0.46041772303732476,0.7221575005818291,
 0.14137496879636158,0.7633542789475851,0.735560949850857,0.1880746990782567,
 0.5466723274044524,0.16831498459108696,0.7963197486529341,0.7131789358331765,
 0.5657146506975175,0.023535721667901188,0.4058552414418334,
 0.7403726457429075,0.7971970272916955,0.48082572157834036,0.7258233137704005,
 0.7610240566091612,0.5436970604323702,0.4371577681796782,0.34297296247258047,
 0.710462338134272,0.5971807676374405,0.6026162149496358,0.41967168072144057,
 0.5533466755279025,0.9427280855118426,0.2763748358330377,0.9746852795128933,
 0.49119402890271036,0.3929542479970861,0.5999985076436806,
 0.36585048305939105,0.4124491010008803,0.38345843094164367,
 0.6413680130651616,0.7588435722995901,0.2947467285583155,0.906657054042895,
 0.08936503688947339,0.28716954557471663,0.5637421703972113,
 0.6387287830113882,0.3026311388103412,0.7386006453622529,0.22181219020375653,
 0.79080919085749,0.453104487887655,0.586103037396829,0.7860850882927258,
 0.9681383940488941,0.9578032829424796,0.8546055284075988,0.9452384368237324,
 0.6555923350789382,0.40168588122275717,0.8022857258497588,0.6908562249554404,
 0.11900303488296937,0.26672417648092317,0.4520459025359085,
 0.39272928878039126,0.2564945042268496,0.6054615484373397,0.5158880269708661,
 0.5102444890915193,0.31969533197455124,0.46531490328410474,0.615029662092957,
 0.9808554229246992,0.7160762318428834,0.03540898396997638,0.5801623655070399,
 0.006867936641498318,0.7623250117917159,0.26638833305788046,
 0.5823107267694306,0.28616558008461757,0.5448589440068284,
 0.11148803322158651,0.8716642955065664,0.16636810151748538,
 0.38586174185322886,0.09984018382860516,0.25124088586137305,
 0.2590062412356021,0.2749567987607353,0.6056351124886128,0.24789609404674118,
 0.6171765496886231,0.4057676518778086,0.1737993424520876,0.34525862845832245,
 0.8918693837351174,0.1422797105643745,0.6598236219845558,0.3569492027469209,
 0.4658422622894145,0.9962801656689354,0.19185033858874379,
 0.07783423926238542,0.6256102817324823,0.09540894231033592,
 0.2491919311055818,0.22571623208435898,0.29273375158570225,
 0.6204893689903961,0.8599400636601299,0.1272625091820272,0.8134754940491843,
 0.08995259636752972,0.8087376226792833,0.9729483129490197,0.4219323145870634,
 0.062675467819434,0.8246302072887862,0.045021269440552736,0.8739156817555003,
 0.06399173714303641,0.5775580232547591,0.9399192273718953,
 0.37090738704539783,0.6192427683030344,0.8272582494679294,0.9885148539876041,
 0.8510561607964028,0.9308857585939867,0.024343781200541703,
 0.9808590226424504,0.5333036895118858,0.9853966355802612,
 0.010715398482307181,0.07713215306753896,0.017177157753709316,
 0.44440622277188757,0.910088382333764,0.8093346476600737,0.3094961457410935,
 0.6713309057924826,0.639196157616256,0.9396722311164637,0.8916522541810689,
 0.02906815140249741,0.02107956002919331,0.9070816333858684,
 0.9018428240491031,0.38260796494280613,0.5482951278231661,0.9767591709011445,
 0.9542835310147274,0.7180987034503562,0.04065354127771681,
 0.009092470109517414,0.5046594619897752,0.6366916180858053,
 0.3086819294158786,0.5924517184598936,0.44376055772727285,
 0.28265904116995366,0.6543698997147697,0.4866806845516488,0.7669815543968136,
 0.7556876666207788,0.376745020820649,0.7840224423796811,0.8910524341998673,
 0.6081284224549423,0.0038106034698272007,0.28202683782976423,
 0.003496790393321514,0.17210224157873255,0.18530832621335358,
 0.5300416866710111,0.26007223794650125,0.6662752086322653,0.2817877825454861,
 0.07977349669909328,0.006126269563314524,0.794489378336633,0.327045068431532,
 0.9109514701035958,0.959498017638322,0.07206086571331327,0.2175236528677208,
 0.377009234001335,0.5209903023324072,0.49761880534971636,0.6699400389023106,
 0.25099919399255577,0.23446732769170897,0.30608733566509694,
 0.3222224371719854,0.49235365836202827,0.9387181329728367,0.761727462876857,
 0.2870618273124109,0.27035324353488854,0.40764850909961015,
 0.8189312331106892,0.19607339422632908,0.07176730892175776,
 0.7454250501913882,0.2621329452364122,0.6197260807596144,0.9484607412061018,
 0.5200078357569632,0.3542547854652113,0.45204793966711554,0.9840242721571468,
 0.9777684740041195,0.7460744848430095,0.42357143073062564,0.6578155927100607,
 0.4175021212697494,0.10719241615118724,0.4555714459693767,0.9802917765321303,
 0.2627049255345283,0.42313156789167095,0.039253404681970805,
 0.5028587873654409,0.4664403781750046,0.7159592350828835,0.172368081467431,
 0.32457560328168467,0.4195944412138397,0.22096684781403797,
 0.32171129674446475,0.5510247713073785,0.7379495425742223,0.5721780026651812,
 0.38948993267458165,0.25231084796987513,0.23427538676243065,
 0.3395345023476106,0.9104803602800984,0.6287973798681472,0.07956109033756087,
 0.9056702834483137,0.2406762389968674,0.06647001818673781,0.6880695253249486,
 0.820496067412833,0.6749072199747845,0.8617963046630884,0.1936976940417363,
 0.34886985388516334,0.8952358958800833,0.9996805314610951,0.1767786437672758,
 0.7884833826993956,0.837700421057406,0.14856661859588682,0.4686649922149919,
 0.30274353161486933,0.8063360139706243,0.6175392267348064,
 0.17497167359020416,0.2627951016700928,0.14738671622046096,
 0.9956542031115835,0.7539162059859882,0.20616347801929813,0.4718510324053976,
 0.7543084823796432,0.7394686539568093,0.08120276591646869,0.6963566043683594,
 0.5119752175815331,0.1454226469470098,0.46685580348661426,0.7679044156443254,
 0.9233094965097177,0.4686209176945808,0.40644305475955056,0.1877992457442531,
 0.04070224106275655,0.9539539714030912,0.5949413045250038,0.9394004063436772,
 0.3073131840842974,0.8206068359285447,0.593762316990371,0.59875369149819,
 0.2703708918793185,0.6145650140027813,0.7597320932663352,0.20487499237649753,
 0.20775303690080393,0.09708301345830361,0.14465063363977637,
 0.9809945312257902,0.1647702370549856,0.9789854895965789,
 0.042633171975851814,0.9041864351995488,0.8549057027245706,
 0.6300695611963063,0.48231815554667956,0.14430247792799022,
 0.7765471797412644,0.3957534952211126,0.02407939891035682,0.5937474539716566,
 0.3346174886931239,0.00395769992607975,0.05662822051856364,
 0.05484240961048059,0.3067558741034375,0.004325034552436158,
 0.5943559994526324,0.8856810879048255,0.35722395814919516,0.9549424744692309,
 0.2837218757827703,0.6983848999974825,0.49937053641088613,
 0.16560734526554743,0.9907259600207612,0.3344037878075734,
 0.013982122444873402,0.41558436646899766,0.9753614029298586,
 0.5151351760830626,0.7474774808371821,0.01666992793569355,0.6374206481659338,
 0.9349670774339699,0.3859877402057914,0.8071731833187755,0.8673464721369706,
 0.6405241491630981,0.42157439024402854,0.20451318740353985,
 0.7527760130253437,0.8164803474163407,0.7255175378618473,0.4370822998225734,
 0.4594955653784085,0.06124295449401029,0.9196575005487995,
 0.38702705043693275,0.6553224830072257,0.10292088283365874,
 0.009951294296355995,0.3344369209808258,0.7518760352865446,
 0.7316340703414269,0.6762494275214012,0.6458685040761667,
 0.018693894931553268,0.06076164417566188,0.4628554185886714,
 0.18907923691239648,0.5671092969783921,0.7226587217363881,0.5998471619561602,
 0.20186106599860787,0.393275659453023,0.9402665219602191,0.17594437244947314,
 0.5293741202650625,0.3640260532942281,0.6953531691491652,0.064981635057352,
 0.21454832377241595,0.9706083332278098,0.019878552396108362,
 0.022419283579301896,0.9132481308418402,0.9049633767709925,
 0.5577891747468284,0.5052971077571515,0.6839133680747287,0.8212521047992323,
 0.7903664506128271,0.43394014715988516,0.6368149000326315,0.5583392431836407,
 0.3806751793631409,0.338001193610283,0.39962452490339717,0.04336423817161361,
 0.9529240233161234,0.23035778162231646,0.8213946124489195,0.933049428894252,
 0.4806787394348938,0.8847090112488404,0.6404733799924907,0.24132513878940098,
 0.4078673497934908,0.11355886078588318,0.381612466630155,0.13008751185076806,
 0.3012665364757221,0.007831001995704745,0.2418974098999095,
 0.9067038978462874,0.6490458809321712,0.8200753618977975,0.9002698604415482,
 0.4275206211299589,0.21155387211341947,0.4500698490800893,
 0.17510392643064865,0.4148116063156667,0.5809376433531719,0.1630448386607205,
 0.33648090271054076,0.2846756301191772,0.359897920943131,0.8460406386212075,
 0.9742830452628269,0.061262762597785514,0.9222574873013059,
 0.8276523644126828,0.9925320833934999,0.20356074026201543,
 0.043298862366960233,0.4077300181224167,0.48867213835478696,
 0.3411229889286509,0.8641818533917965,0.33888820376901674,
 0.15810134553541366,0.9574194610585851,0.8383953154358331,
 0.23030898709401126,0.6246506076909939,0.20213466928559842,
 0.06684633312658406,0.7393367951682004,0.8400036229710263,0.2916500243485344,
 0.0203791870094151,0.9431177183797148,0.23667675696450396,
 0.42371441049606273,0.24216709289213212,0.27766380674734004,
 0.7058122497187176,0.11726482836725194,0.635816369331618,0.7400658094568896,
 0.5495032064467262,0.4553547300262347,0.7155106519957746,0.3775822038144372,
 0.9653971265515748,0.38819964540805896,0.46008051413697715,
 0.8072638694133824,0.9309463156753592,0.7598866337578976,0.6658770113472895,
 0.3894496464679885,0.09829295432385221,0.9605683308740163,
 0.23243754507724768,0.5753325948805037,0.12815067716684037,
 0.6226891961462976,0.3094137378059021,0.9897769834546611,0.92766354448182,
 0.9990396433172108,0.17180364503340917,0.8217654189593312,0.5019669761397186,
 0.7793407143452022,0.7540224844797464,0.441580191351178,0.1734256967936525,
 0.7129434250324947,0.8796834213995439,0.7216992417944381,0.8208675229563861,
 0.4617073893493959,0.8584306640955595,0.2463057120281158,0.3508475918811521,
 0.9899903152253804,0.5580183428442722,0.4870090805563908,0.5752368249127219,
 0.46934483845313746,0.13074425868411077,0.1741615729946654,
 0.2621926434658166,0.2381588974719807,0.8816805942003462,0.3777065660276673,
 0.025809839194104733,0.24606563297471817,0.3669852166039699,
 0.6216302526389047,0.5714286394604855,0.44324018190186554,
 0.09981612174081156,0.35957960926590615,0.3823425117725693,
 0.8745510107207906,0.8656115050833929,0.6603219802606054,0.8160135392370114,
 0.5626111923532785,0.30835006295752554,0.7568098647532182,
 0.09294717303841471,0.3182228191723757,0.8933245845891524,0.64626145793226,
 0.39216889783291764,0.6480160163992704,0.43088425276291864,
 0.3597894046637634,0.6074632156759355,0.5774695488946059,0.4709412357335788,
 0.367919716830276,0.9180953986757077,0.1252825269167921,0.29617538078404837,
 0.4313025749590613,0.6404585507362742,0.5040193747226035,0.6910643261164893,
 0.035060203592517714,0.25069839632033375,0.560662040140903,
 0.6469274899043493,0.6849378110853032,0.2611960764229442,0.8801551122332261,
 0.8519879900355607,0.03827684473821069,0.23360604666877904,
 0.3180430348793982,0.22977895720578834,0.06187325842895364,
 0.16257208983860183,0.2665474327168599,0.5282324624361278,0.4202290102801829,
 0.03672271653929693,0.5210645402520853,0.8079961376408382,0.9459618148643882,
 0.49532519291102417,0.5990154257548801,0.772265623028549,0.4177126767097743,
 0.8345066015594502,0.8757722253155177,0.7051877312448627,0.3912928584957065,
 0.7071518640390418,0.4880733321533206,0.027242748460792332,
 0.22857136598692585,0.7606190640653921,0.3011469363131498,0.7569462512706892,
 0.40485107743304805,0.9509553544487364,0.8642840763769647,0.5203786148074779,
 0.9022255025541837,0.6420672878354922,0.18314809489491435,0.7260512949219675,
 0.7786410934211012,0.06045244851895648,0.6382075706010146,0.8292601581243504,
 0.9005706262699779,0.33060115607499463,0.5718455703227597,0.5524448641253474,
 0.1964542502961325,0.1096009990491168,0.8881776061316236,0.6487596319816501,
 0.8820515787351214,0.04206403707203987,0.014048667197458897,
 0.4479428631835425,0.1187962751032372,0.23655239398415928,
 0.27255177763054395,0.024902879868145766,0.004731772905312459,
 0.8582023620311575,0.11201728873145866,0.11758604024255104,
 0.3825123995048775,0.0384416946259134,0.4652732238763623,0.4277827672490858,
 0.8373505711903656,0.8248470233078526,0.8800447594694994,0.5558471249819001,
 0.1302423143460718,0.4714884710532007,0.8258658013793192,0.2604940587215393,
 0.7245913270086435,0.045569186172344955,0.13800848778705344,
 0.5005852224982826,0.015247394678573611,0.3572725201433159,
 0.7641601541761806,0.5113295038832565,0.03917609661999322,0.8045113180466616,
 0.5151537274081128,0.6820434471371756,0.0030192062802143482,
 0.328547168595582,0.5959508879290148,0.861905721064541,0.27397036653667217,
 0.13631777446826665,0.46308999592306677,0.11885668140895478,
 0.9352419140084028,0.09504939147044056,0.3509695083263633,0.0866192556851656,
 0.8530253093506333,0.5060617456845086,0.06783943653567981,
 0.49219630795703795,0.8917553676789087,0.8567903858575275,
 0.14199933870238612,0.8639401330861649,0.8189589237902941,0.9933972555902455,
 0.33213410352392825,0.8938203079272349,0.613115325079421,0.682809225497493,
 0.7901082101764852,0.9305151246597109,0.9173295817593905,0.9791663126128175,
 0.1429367696717263,0.27846810407433653,0.841047993950022,0.35431280134848986,
 0.9809225059687067,0.9890645149596837,0.31562349205777807,0.5149201501083607,
 0.955165841938571,0.054318531042607976,0.49420849545061474,
 0.5574002093451103,0.3518652912445793,0.9190152295161753,0.8732082434531324,
 0.36913696446907796,0.9080625069613132,0.5415184932578494,
 0.05125345755512334,0.70140775194313,0.503840313429574,0.9746570712506538,
 0.5456464773793915,0.39740742440247945,0.6966935656131754,0.7792971453572974,
 0.5527269199086733,0.3381830897374749,0.4173802126914171,0.749552646083487,
 0.6426290904757099,0.5510571713771486,0.850820350450313,0.7796962441454284,
 0.20416756529916458,0.07935866789378443,0.1071266050230314,
 0.9891463521685094,0.10337643760103332,0.6802852042591819,0.6501204124399852,
 0.3724924757801579,0.15623990507847196,0.5236277392249209,0.5017920439686855,
 0.5263119244748573,0.4080595711204389,0.00572378571255272,
 0.04030211544836326,0.39666470038282986,0.06464759892497884,
 0.5935895597807398,0.561177814018186,0.35532165437449226,0.9627395074414261,
 0.6947040084994651,0.674376369769906,0.9445850420384605,0.11987919148718862,
 0.5803708660798614,0.32659138757588024,0.8365458022579878,0.8350266808176167,
 0.1267780553267559,0.9424645692236902,0.9227703980398467,0.6777913283079625,
 0.6803540241558429,0.5082932960642706,0.18194464293006762,0.5143976212131307,
 0.04669423966021302,0.16299249348342082,0.5597606458280474,
 0.03601381820704952,0.837030408792788,0.8497034606028586,0.870282599349519,
 0.12005083943725081,0.8059923822790046,0.9217521846407283,
 0.07956138957504022,0.9140709430654695,0.3923619147711469,
 0.11806539545627559,0.8107159937601426,0.22089317140154252,
 0.6452449220320884,0.9328054264890471,0.38600916444876865,
 0.26423020833312094,0.38176459768472115,0.43946090967637264,
 0.04359063613024139,0.33159751148716843,0.8267594085862341,
 0.19242824265842606,0.040219604939541465,0.9900408146590011,
 0.2619153787612861,0.012033450509864818,0.4746082750866484,
 0.9494040950151253,0.6674902176837593,0.7276915720651085,0.9934685405618541,
 0.6272777443062099,0.9278549223352697,0.14311158111696942,0.4553307854057834,
 0.5112196335164689,0.9927048534022547,0.6402244648105504,0.37773648027454043,
 0.3743778552310577,0.7600920170308337,0.3079594561635457,0.5474900710377286,
 0.5860186359086533,0.718636169768703,0.07882387797860146,0.6639571151216708,
 0.7396386458242257,0.209069755319095,0.08610246549874812,0.1376379014363378,
 0.052097762013870286,0.9964448109447942,0.5851323282541272,
 0.5441844085109359,0.9551640065313688,0.18192653950812543,
 0.09533288047241206,0.6241723835683718,0.8153450111498451,0.685875680567075,
 0.972676211120007,0.6604016584942786,0.8831957965123554,0.13120649475379975,
 0.6358078978738925,0.36373072957322927,3.6762817043478613e-4,
 0.19794699039851504,0.778191386268597,0.09449785839696123,0.2472803563587902,
 0.9292077132476776,0.936423498625966,0.9949757418796296,0.5232963192831707,
 0.007948577031356874,0.09109558987510025,0.46840384430468895,
 0.4330378228803703,0.1832388387894751,0.4149100453222986,0.03642235245450509,
 0.670916846625452,0.5827399427281197,0.1661682401457818,0.908489302117214,
 0.015588478188503174,0.9548588605056376,0.5183625150610336,
 0.27740946480575845,0.38114444453470686,0.2889611631434472,
 0.7342138610234524,0.6837682871611017,0.498087726149838,0.4528989106774385,
 0.6916649993850293,0.19102561852299216,0.37687443711961355,
 0.0867410433644713,0.07893883136689706,0.3062294744158305,0.8921080705406813,
 0.08680530556038302,0.8217247055207939,0.3403001781781837,0.6685956762223938,
 0.7810592560908518,0.6904401139699209,0.791099365673744,0.7043724037231813,
 0.3645225231048822,0.9391629837150863,0.3365998812486023,0.8465720908861314,
 0.0897965477193059,0.7570877446158869,0.110124515749627,0.7357121636271364,
 0.6877564786132635,0.9943806434632999,0.31125330879381985,0.2960858090264199,
 0.2300081876477551,0.7987509128361887,0.5914063400914296,0.8144903135425702,
 0.06556072281678826,0.5876424749412789,0.14432305506856957,
 0.14861224225608216,0.9731975430253376,0.23392776625640022,
 0.26499904073699565,0.6464238857004151,0.24159640870407673,
 0.28698518296764164,0.36354688256072853,0.010954538618207499,
 0.2911890227587649,0.4446877067770758,0.25851850953577815,
 0.13957486220312854,0.33969722333384655,0.05792830165242979,
 0.616667939163832,0.72710845005346,0.5586690712387382,0.7942706960792617,
 0.2699201558284645,0.5119855395831341,0.5894281779594659,0.549482513131809,
 0.3587411362981009,0.06205861506386101,0.35146690119409185,
 0.5311494050501304,0.7540406064562083,0.3806316534818043,0.18964540296836907,
 0.647463914502753,0.19213617822104712,0.19606083304868926,0.8462847916438505,
 0.49494112721018846,0.9091195934829945,0.9855984653488603,0.7000018684565577,
 0.9322050024714734,0.8603558493086367,0.7069773247937372,0.7552760969840421,
 0.31586483221814854,0.032672329083264895,0.3330369465430789,
 0.5367268953834377,0.43104276890991033,0.38531430104221753,
 0.8149877336718965,0.5409385360217351,0.48126631643606965,
 0.44950789630938903,0.567852256757846,0.36285592874503514,0.4295696204481345,
 0.4141835764108198,0.617160284415748,0.0876407498099977,0.3893328552553621,
 0.19200624945946843,0.3545307537735294,0.45741779224998247,
 0.5752909739898261,0.03743773184251431,0.28166395994581683,
 0.052843576401757186,0.8592979501016188,0.08869309013862336,
 0.13129859921239284,0.24250117477005584,0.18522979644376636,
 0.18257692528376168,0.48037291838944596,0.9166135246007903,
 0.6506835131828692,0.8017043888806276,0.1313080688098297,0.3742738283969119,
 0.4418184428343209,0.24109313029628876,0.47280666141394023,
 0.4797965714825745,0.432041229862993,0.015648001277702717,
 0.16256065144782272,0.37733301803359875,0.07982841574705613,
 0.7078227305797409,0.8061475764129582,0.35497487087737456,0.8499614643531312,
 0.0883174941231768,0.9586665013529476,0.25104842311279707,
 0.28397540513494246,0.7116586917889468,0.4850747119829386,0.5683462200493183,
 0.2815609457417808,0.16034087042053424,0.24974294506553885,
 0.08726338490379704,0.05222732806305497,0.7153202992512694,
 0.8595743243390752,0.055790706888833363,0.997331370884942,0.8754295046739689,
 0.6343709621513592,0.2494593245180754,0.27328763216682517,0.8989345142766234,
 0.9403712224541234,0.19808482176636166,0.4593057561219501,0.9424711386795575,
 0.8184984021300099,0.1321994419685586,0.822106026204166,0.28015860306268325,
 0.07643797945523156,0.5331121557857025,0.313607194445193,0.8325400728078133,
 0.27133530552461727,0.6649182445395168,0.28548848439638597,
 0.7396045313723446,0.1402462671862571,0.7692012055070929,0.7753456919207138,
 0.894746009552363,0.666043493346913,0.7095340675503937,0.2655713908518179,
 0.32447697666383934,0.3093439206746744,0.6621277922351406,0.5751493023189348,
 0.757360072366443,0.6760076223562561,0.584428327069149,0.2440215391744267,
 0.3505748807822433,0.17111862612077589,0.2761717964817181,0.9374753145713887,
 0.8733319147260759,0.2734678878194615,0.055044593752744064,
 0.4928586888109867,0.8500089822253519,0.8195771524694558,0.8882434087933352,
 0.6672864506018448,0.31115276916797663,0.7595850873967145,0.1603537396853374,
 0.1523963059140132,0.5062751282684351,0.2602144718671904,0.7061109340248267,
 0.4153089227426012,0.244737195564948,0.5488836335262595,0.5175756829567122,
 0.08212773494608261,0.8389557374633648,0.13489030549086056,
 0.5616600029520253,0.5929331111216499,0.7200447928795962,0.2582216228695181,
 0.8681968590073565,0.7343095042766467,0.148666483194426,0.5830727005681137,
 0.7257958734350528,0.15523001144039328,0.7691242211719719,0.8359772916862577,
 0.28065755567211537,0.01364045465511532,0.7560199482240139,
 0.6049040240910772,0.04517237116914952,0.5666799402304474,0.8483850032228076,
 0.4562619394568327,0.551628811264252,0.4087476545040686,0.40344164105983427,
 0.6028397212416824,0.8645899887082322,0.9892764069070041,0.09425952991265174,
 0.008851048235015369,0.13454248422298143,0.9036923812346207,
 0.284851385177498,0.8946482998480311,0.3840363485058702,0.5827900344681617,
 0.48848779050284385,0.03474320569388589,0.9581132638139612,
 0.7698618610659602,0.2016333529878962,0.9105437870131377,
 0.012348809937128391,0.6222145560077258,0.22102175118465917,
 0.6869076581661318,0.3537496581535948,0.5584593384291889,0.19195062258368867,
 0.9015068374767812,0.9060186421181262,0.9051099986049288,0.41221569088299737,
 0.8350378725878711,0.6700369805213311,0.3521484165144376,0.5686620811456524,
 0.8032529245231105,0.3159870783615375,0.31323259121691793,
 0.023909512487136797,0.7737411501782718,0.723530772231479,
 0.011400014697984795,0.06649656949509852,0.027747138009118766,
 0.9670174862825853,0.8163094580183481,0.37264607357559854,
 0.011107174938540632,0.20939770869398322,0.7745392507251702,
 0.6183176939642505,0.34376062853334977,0.7994051081726344,0.0071115464844973,
 0.04888905074151895,0.8714407541345914,0.02854994281528256,
 0.5204001979168648,0.7148175681697512,0.40449089107391867,0.2251516289359745,
 0.9148091780543611,0.37903162681014557,0.8442773022470779,
 0.36156448151980647,0.9433411076526839,0.6657841089374457,0.1404793522028862,
 0.2663585469918244,0.5076120451988536,0.5484517689781467,0.19639051094971305,
 0.7885466196036552,0.8367771411612445,0.8459156805503993,0.17084350471907472,
 0.45525375750075736,0.6772794992895266,0.440521910735293,0.5158352286097524,
 0.7991439724766507,0.9266269008018093,0.9257456558818555,0.8510044382319567,
 0.5288092100044444,0.631108774078423,0.5611522245871273,0.06743913826235493,
 0.2836485522758212,0.6534341162104702,0.14267621367842898,0.4123366754845331,
 0.2957409193688496,0.9166912352224759,0.8458690067526998,0.8967885474620748,
 0.8850448354868944,0.7472973682708801,0.7082486744756304,0.14338846476875555,
 0.981839893573967,0.26683993394072714,0.2392696105363593,0.8403926080195958,
 0.6202036855245006,0.0816688568804429,0.6573477347231219,0.9374338785717096,
 0.6532652571307063,0.4439068526434995,0.38660325187172395,
 0.14895680199723704,0.6383077103562792,0.15666793221545727,
 0.25505229500971294,0.618621577692336,0.618280258470718,0.6697201721841586,
 0.28141050589645067,0.016463331215287758,0.4311926614955577,
 0.512204855358615,0.94445731175726,0.2892244986047243,0.9752560507522866,
 0.11180918414358965,0.6233564949277826,0.045363981033462286,
 0.6031969039286443,0.7989378008520902,0.6892077683022482,0.43415143655233523,
 0.2905143512359126,0.9676210149829616,0.5069798794067626,0.08689404486678709,
 0.4037484569633394,0.1935484339906225,0.2691307372530155,0.5595883229293174,
 0.6021315965519035,0.8601545858138374,0.9211939166928556,0.5880675272611606,
 0.5520168310145142,0.4617185102555891,0.7577132185878233,0.7573771935500524,
 0.9093136073081914,0.16221658730870425,0.3157909611918853,
 0.13667782634549952,0.27104166255221973,0.997079121299343,
 0.04742466653122701,0.7538746045248275,0.7615016252942398,
 0.21547823185793702,0.9765615238657436,0.2392630731758557,0.7838357482292592,
 0.7747289606273662,0.6052406794589175,0.7765419630450631,0.23695173779442835,
 0.4051210531724321,0.7951141594333253,0.7054598747305484,0.3183992042416084,
 0.2856640762523317,0.24900831361001963,0.79781669839128,0.4513368879528492,
 0.4191752125484868,0.6869828870715002,0.5807199292225198,0.8197798391068408,
 0.378035820285722,0.1721283653533019,0.8871831606486251,0.609861853545422,
 0.11686086311374178,0.19634597744345572,0.5443491246141998,
 0.05916057460596158,0.9219333598707857,0.3052461041000043,0.1869658002189838,
 0.16366197940706662,0.4035262922621624,0.9576918936896012,0.6162422165786829,
 0.7585101957031175,0.7031789018893593,0.09077171169380427,0.5197332120380128,
 0.6085790555914161,0.11117011092327289,0.5909231309968266,0.1149683775040764,
 0.002034076208737412,0.44319478265302914,0.4746885553469655,
 0.09115501199798226,0.780744125560724,0.856554493697013,0.41494660868457267,
 0.7402064110634679,0.3805163385483272,0.1659500354969109,0.1314124545455292,
 0.2808355363424535,0.8129148969690696,0.30503437549269197,0.5660983084736274,
 0.8777197361694655,0.5177018243210736,0.6961195987813381,0.1338292450872699,
 0.3996514789210416,0.9394367196776428,0.8133528328079442,0.6326267775709372,
 0.4256112637585818,0.7049515693780233,0.9794262725519346,0.18733115285297408,
 0.6143302825521794,0.2574664974288321,0.5659828095302655,0.7493593663892542,
 0.9749900153921263,0.820739054944381,0.46899140783753634,0.49802117585755856,
 0.3351129182788525,0.1886501001422044,0.7170122461031632,0.9473748451480336,
 0.06189225881064153,0.22494530082124586,0.4668753255343132,
 0.7630928815750166,0.9411085338777709,0.7382678078329726,0.8928132650228525,
 0.9546653716448666,0.43411946228807197,0.4530442195286408,0.5646350969973368,
 0.46514687302457647,0.9085920492702148,0.6692984717685603,
 0.40733452742233134,0.5981012257080734,0.480270927451527,0.8191538840789947,
 0.4435077150252955,0.17467045665360237,0.3753596920933747,0.183935404120252,
 0.28399807214744066,0.8812050951144614,0.11131555464069798,
 0.5186015396804551,0.36649075189182934,0.16067909162833993,
 0.9369531145535746,0.9815223267693226,0.7119082214392767,0.6472329694153263,
 0.6980352688813423,0.6827152660653711,0.6059098686623928,0.9705022829845632,
 0.8963977000609017,0.3909505054482564,0.6274152860834187,0.3623879998019396,
 0.7168205795120859,0.4907056208649445,0.3006301526741981,0.9939030537907728,
 0.10627784653148964,0.3166308501989916,0.9029008339843194,0.455293981671814,
 0.9917301247142272,0.5656544774100419,0.9102996768628622,0.6547435877225903,
 0.13179105747777686,0.182643088005485,0.423270341289375,0.4177208511600744,
 0.9107259844210143,0.6108217164181318,0.7211551972029696,0.8372228033999225,
 0.9483968110178289,0.018936891957376245,0.8116006624196248,0.728816205168846,
 0.14163481886210305,0.7610247899595952,0.8881400304058605,0.6726655762600362,
 0.7214600375766729,0.04730603125760302,0.2088828056923815,0.6999542301409034,
 0.06154108331562824,0.020200879718119125,0.818723943959766,
 0.6652760549833249,0.8308738482317966,0.990416709901962,0.8718618208650492,
 0.631158840264168,0.6053863428537767,0.921617545366259,0.4883699188815651,
 0.0340388703860286,0.6731168933354933,0.746041216959296,0.3495004815906716,
 0.4889543886263713,0.8164442082160677,0.3433328044853845,0.8825711335957428,
 0.30647149026799614,0.8426437626835996,0.628575944084033,0.9986843327973957,
 0.6468309571954831,0.9652104541879519,0.7308110877459294,0.6719395497095457,
 0.13652280109941006,0.12384718333043954,0.6662695285782309,
 0.016810788444805436,0.2723931932570246,0.9690763852199449,0.593339754532842,
 0.04286693974368927,0.13162197291281807,0.3778263041015618,
 0.8862544768504473,0.46856374715820925,0.21446575520807254,
 0.14010254502809127,0.5564120022354047,0.6764411097474567,0.5946975906210963,
 0.8456054172476267,0.6383911920881398,0.1887643185011051,0.372047640696484,
 0.0971923721516349,0.665872438123458,0.7662156400912442,0.26374272307735336,
 0.5659960089857663,0.9066200901938664,0.85330786727867,0.21463356306297243,
 0.1044501100070504,0.5408311580140562,0.9470098949446897,0.5539925665851488,
 0.2839234520386008,0.5766162972638426,0.1095202148407054,0.25796049831195755,
 0.35356248282873903,0.6138747319168014,0.8526381735726059,
 0.09842555000147168,0.564989670597046,0.26511836510557696,0.7330236569713302,
 0.7986862777509511,0.3477796255411195,0.5536912639727513,0.530641849104838,
 0.16133057965032094,0.8652604403240672,0.4917103552618485,
 0.03481355607557213,0.7861452912742692,0.7311580344062631,0.61312991919443,
 0.6015873549541146,0.7442311075966891,0.7055571512931754,0.45976255811605304,
 0.6068109835302993,0.3214921056882263,0.6101518993257768,0.584632479529122,
 0.011277137443687857,0.36237590433924116,0.27730846616256066,
 0.17484918490317736,0.2936063997589341,0.3704597052145969,
 0.13715759297985275,0.245915775329261,0.01878281197417131,
 0.31607945502017176,0.9585172573458915,0.6890901653606402,
 0.45666019226509613,0.4420493787315367,0.5203942455246755,0.9899907439920244,
 0.2583770805344935,0.6882402117380453,0.340340849005225,0.09022167156196548,
 0.6111551622247117,0.018638593042856755,0.39344320726281856,
 0.9975578005179256,0.9020683164042503,0.2470355681506623,0.7143663171552773,
 0.5663738764941675,0.8832967031116337,0.20457152817979396,0.6792498288201327,
 0.2525908403616546,0.009062507691003696,0.3003787729919386,
 0.8062602697354017,0.6610449206418462,0.5859102168044614,0.5027670223472409,
 0.7672564726604886,0.3718534240238768,0.4422875911866686,0.4534570287522197,
 0.9561494992832779,0.8256403908701604,0.157441314589014,0.17958659751836248,
 0.7934570484493845,0.7183630520692756,0.5524459106563373,0.6615013014590668,
 0.7876975769026632,0.5114601355300897,0.09450775281107082,0.109042225057558,
 0.15497645185686526,0.3997256324494426,0.8767616370160036,0.7887806438134819,
 0.347369187759785,0.24916161584262275,0.6865820741962148,0.3742413303368022,
 0.1564614636990046,0.3850676776965185,0.5983171941754893,0.2175800388704463,
 0.4567556257923252,0.48838673784530995,0.815271473577914,0.8454391458535873,
 0.31291040454578856,0.928598457469961,0.25518382011537877,0.6449792066761739,
 0.788649760611158,0.4692675298281185,0.27352448020625275,0.04407579276609763,
 0.13896674121864694,0.3792447531665154,0.37266505935182925,
 0.9378346267384738,0.4049617290439512,0.0217101853111612,0.6218753101776675,
 0.44644492661793156,0.149774874292952,0.6567977171972976,0.6622373055289709,
 0.6885916426434454,0.6908044080295326,0.8818935891944367,0.09970181987669902,
 0.36636285169861615,0.5719149363955579,0.5502648331328088,0.3420997367678248,
 0.044892053806441456,0.45362726630068784,0.436831787718559,
 0.20893185566210404,0.7183289640222037,0.8096018112041375,0.9875771411481096,
 0.00809578508369968,0.49894860849950384,0.3726030849245109,
 0.16550853958609835,0.6068346833235289,0.39761841660954245,
 0.5447547967763499,0.6013818325862328,0.6048858255803979,0.12075003453964817,
 0.10552275602399286,0.8055064154534388,0.6827026220470858,0.7711415760984076,
 0.5241230232162155,0.8608562721731147,0.346117180542618,0.5625216095049843,
 0.671357427273664,0.37722423425332996,0.08036975381212319,0.2257063493612873,
 0.17165446239132176,0.7488124575916723,0.012675910576234073,
 0.09064584895285277,0.675213059260468,0.5062103748467988,0.21942241175595967,
 0.08005912058709752,0.33998976090955735,0.739650119795947,0.669204236944805,
 0.31977990239162113,0.2048479900927589,0.5767613091568556,0.9143818371052095,
 0.7149658798136066,0.3811456465401375,0.7190365429649441,0.24585238968896783,
 0.30787191119286417,0.07441440749762551,0.0535765993339713,
 0.5641482862698063,0.23195128948178867,0.03794436495257236,
 0.7095803602032124,0.5205991145944444,0.9582680483817873,0.4765443233306721,
 0.14355423063729578,0.22630306657680876,0.601953892720563,0.9576002695443024,
 0.8221353176521601,0.6967672295500359,0.8405306342880039,0.5571200897674762,
 0.999093235020608,0.5824214981201892,0.16718757100925763,0.2248705346891039,
 0.9552321177356999,0.5563059632455486,0.6587101401030029,0.41328128660086927,
 0.10831859719865111,0.08076815062781972,0.7851485245607592,
 0.24898953638333643,0.8978560164325315,0.9312388913686005,0.4071892034058723,
 0.40482459052043185,0.8255139876059455,0.7246660636341209,0.2608026396688452,
 0.6722435659068966,0.4996377634109622,0.9383504373165892,0.3884629944874861,
 0.2054629886237852,0.4247780165646251,0.7031849482868001,0.542126611419021,
 0.224591231830084,0.054026561172282994,0.9608248759422789,0.8629984311691414,
 0.2696470045321878,0.6174892197117497,0.08455633642694615,0.7770462409753442,
 0.3914922212790166,0.7285763234001272,0.31924401771956157,0.7279431941604617,
 0.7191926990047046,0.11225753900725943,0.5298008102124464,0.2274029151793887,
 0.7483320119732336,0.4957768264550495,0.9368201458571851,0.5018401242661418,
 0.08961136865177644,0.6322495992613524,0.10454352626155017,
 0.41188683962115924,0.5419968120009829,0.7355717471865071,
 0.40697577981549393,0.8486792318973465,0.9362345299633029,
 0.08554715075910369,0.5517492162248326,0.6770394026955872,0.8315904451640296,
 0.5246334311240807,0.5697253988292964,0.8072147965539814,0.22869594322917586,
 0.5335794472099278,0.2721889146442744,0.5224191606114134,0.6748293931842702,
 0.2947499587889394,0.49556123798896135,0.9106716636125489,0.6614835019257226,
 0.3324931005703785,0.5865292208480148,0.0046807984393479884,
 0.7364469425281018,0.01943879472675869,0.8756477047357677,0.5604525055784702,
 0.28681751924366783,0.3554470773344701,0.39203065418786787,
 0.19964196855689964,0.9669654654866812,0.6449404886237468,0.5136296357211615,
 0.9576056500790395,0.44607256119949823,0.3283512926572212,
 0.14419865022092881,0.46541642028000396,0.8784670536710657,
 0.7843734460797301,0.702783437036011,0.7540582548440351,0.5910489616409716,
 0.8942357211840907,0.9814247788309297,0.08552655933788988,
 0.11299005279549901,0.962939009081492,0.15591908798068332,0.7364859492796476,
 0.2715264042967286,0.6087666119946264,0.9888295759173419,0.5460803201839278,
 0.883834698412675,0.21495219236001772,0.8766314247932918,
 0.041102011810598804,0.35273579830074153,0.6797186800777426,
 0.8619209326369568,0.5088447310761423,0.005819592943969609,
 0.5341001034865562,0.7681663996451097,0.6707431112078139,0.9614004693442867,
 0.21319281579774096,0.3099167684984415,0.2027208524901951,
 0.41669405805861015,0.6799175929620745,0.8591136633609098,0.3546606665983908,
 0.40814414410798194,0.25333159301279873,0.1962322246119742,
 0.48124388934243345,0.959884551145791,0.5101941869711725,0.4517195422084105,
 0.08814921163181766,0.41037771053353644,0.3704265654662884,
 0.9238657568062594,0.6343242089936805,0.5094234392048459,0.3629048759951393,
 0.302849061575188,0.5595133620466456,0.056492140181583994,0.5128492527409456,
 0.7911001665249703,0.8483036376835313,0.7395198172319405,0.9514229279887291,
 0.851034388116911,0.01978862472473686,0.5998094794279334,0.6351061744589158,
 0.24683805805478498,0.5465688404567016,0.4760367458581435,
 0.09239208316819858,0.030348463803139047,0.9138678979806539,
 0.2841751309394487,0.0033753869378521095,0.6664695444054407,
 0.5022360999136747,0.2860972545710596,0.9426460822624195,0.9306616524901274,
 0.13720663576015146,0.5221721857848184,0.029305093304547447,
 0.060868080992218854,0.35242201745585233,0.2636150193273257,
 0.5088584039626715,0.2923556307327788,0.24580546843555318,0.8121670753069689,
 0.7678867594758301,0.0869611990859801,0.20971358671908202,
 0.14006963348226242,0.5656243346510748,0.9711493146270448,0.5373959291603935,
 0.6026718110121481,0.17394913373934018,0.7362316505104041,
 0.19121570870212679,0.7070345539703036,0.39634920990898703,
 0.6520822362624502,0.9347100689653574,0.17712277216807337,0.4119011635593093,
 0.1573840246771655,0.8044081686552988,0.5160814243753409,0.03776282774415951,
 0.34975654969423564,0.10951725629630427,0.04589491707993254,
 0.23337822351325277,0.6722445368590637,0.8233449886536285,0.125650608146184,
 0.18246535581780465,0.38426135782241466,0.4398804724527815,
 0.9713647547645197,0.8113180784679802,0.3685711531294391,0.474150033102514,
 0.22535045102848605,0.44251911911740804,0.9527562032141766,
 0.01587139158047235,0.4466900327840573,0.504128493059128,0.7684562078595015,
 0.990361645513529,0.9217152311532052,0.5895486832811253,0.36593018015354095,
 0.782712580641393,0.43356142895357896,0.21752853249371495,0.5537823677944979,
 0.3284761879812401,0.011972435034865114,0.9159797430298554,
 0.051816507318992455,0.09246169466037746,0.9611433074427884,
 0.6264520874972824,0.04272286287040916,0.20264030577949388,
 0.6386720497335076,0.10627575330027994,0.4800005255682259,
 0.41305775895264507,0.606521755712002,0.6650298599021086,0.5458417165647187,
 0.09831398100899824,0.12204720927886548,0.34848203330073135,
 0.48136837626805473,0.5301025454917745,0.1814573344282746,0.5593753316727749,
 0.792697716875496,0.18388598074964702,0.3418043862267409,
 0.0016718592335460691,0.476653175637912,0.8977793773345213,
 0.9182218381700538,0.132955043225788,0.37978626468477883,0.7851623792166438,
 0.5312096319731423,0.5990416054686354,0.25462852594263663,0.4899681499028742,
 0.10419234572047109,0.3408908681627153,0.8919915597013544,0.5081666841962952,
 0.17652734353117572,0.4244092973777903,0.9997077188156583,0.7082548833745484,
 0.8751753289035968,0.6221784229369574,0.29661577936248584,0.6933167794007712,
 0.03809387186630708,0.30968283942785535,0.6978601793570806,
 0.18876673084644513,0.7171680321983442,0.32644195503439755,0.309015375830359,
 0.672540535927492,0.510491146331869,0.7344419824589431,0.14576594663780396,
 0.26830206033260307,0.9945638834950647,0.3254826826426147,0.1929324999083134,
 0.40834182141726616,0.6004046635555018,0.05165929381703582,
 0.31447752577478594,0.8520325129697821,0.08735915593375232,
 0.6577391617848508,0.6269184532004275,0.19261666809988853,
 0.40031105096790376,0.10232556134987014,0.37744225841846746,
 0.5614530224160977,0.04837212296599813,0.04032353074421968,
 0.3076723013915905,0.3118883530303891,0.9248955143305758,0.3982656882043206,
 0.2147948955814054,0.9863959625546556,0.7788920695334923,0.6007369661862751,
 0.8117442215422193,0.6758682595884934,0.8650311419428031,0.3172574025906494,
 0.8803238819901075,0.8270649074736567,0.8070422079870705,0.7896004370248471,
 0.1531207247194013,0.7707889988458493,0.8128697522793539,0.4141227528193595,
 0.7118976785547626,0.9964823286143941,0.8539850888715647,0.8485740920077445,
 0.5123193913820824,0.7179692584855104,0.0774412514846663,0.5402602745069408,
 0.37572994458412523,0.37794172030121986,0.09085412623836608,
 0.9976477233292671,0.5613222256551326,0.5099487849943063,0.42515227469926276,
 0.8078537473070075,0.5840635313709351,0.47242276176652953,0.6666554605797982,
 0.5253791146645325,0.7007689954357675,0.24790773386672948,0.821894333539788,
 0.8499669360314859,0.6076644897392971,0.7220412840636555,0.9681842635162431,
 0.4683019720966195,0.8625767650958108,0.87427139810489,0.8251502621063985,
 0.39168695476704996,0.6308828050384826,0.7572774256718942,0.5475049039465262,
 0.6444967409262018,0.08734116223224586,0.05276502437054775,
 0.2831878215262166,0.9719960735102678,0.832660364860012,0.6284171672601058,
 0.3385573171731202,0.13910495438095627,0.13830961739655834,
 0.4452436367479744,0.004429654467174693,0.2879964442049041,
 0.6268742804025442,0.45448256827889943,0.5461280742427462,
 0.15118182174004335,0.057874912015326685,0.6584919987376392,
 0.898873279792809,0.4836616683987629,0.467432475563492,0.9971154015106154,
 0.937098473229064,0.9992474702631474,0.17366767454677878,0.39007727516801116,
 0.7391870691084209,0.991205193081218,0.28834456302133693,0.5698693338621383,
 0.26988724215042625,0.7527111989376368,0.7786575781002928,0.9577557603280213,
 0.09037948399972395,0.9128264797536778,0.22770134059821634,
 0.7144415100994996,0.47690303002463486,0.930939731179261,0.6873706473851092,
 0.9485761335664609,0.34701011036257323,0.10703007625334926,
 0.5944766534434317,0.007926812924990312,0.11078728957311779,
 0.06244646821541866,0.4183809844991814,0.176394367201401,0.6499712263576964,
 0.31109559323446356,0.9647738360043361,0.46838340094150843,0.644864579359998,
 0.04848221951506759,0.32116280702113775,0.48372657550530196,
 0.1722949792846944,0.2107548642980629,0.8921141739635965,0.8109103044848155,
 0.2649884567673535,0.22636073341526952,0.9689033300016718,0.2260979473248088,
 0.3589292558680446,0.39789875249839657,0.6668220030229838,0.5622960955115435,
 0.0679087420415947,0.7237358537615488,0.1565952807329316,0.6250462162820513,
 0.18662230126992285,0.9105021714537329,0.7793482845419566,
 0.14783708348356117,0.09634211531714398,0.5815493906519356,
 0.11961234155342226,0.014385550355511034,0.12282756109852377,
 0.8566223387158637,0.9848867001694098,0.0012773716106710964,
 0.4155528105358568,0.556795026726792,0.6579742641153395,
 0.0030448385322637828,0.1676725927805226,0.22825638148199745,
 0.9806384723535204,0.3550032516786694,0.12023815102074664,
 0.18530741370884984,0.8839559530702551,0.2590479600340472,0.2692091892054398,
 0.8871276230846605,0.2407998666275173,0.8511172226962312,0.4823446061849783,
 0.41514194306090524,0.07427380628306446,0.41612607046558825,
 0.7727442704253951,0.8752318309297897,0.22583776201385464,0.5167457733264328,
 0.8538138718060935,0.8439087405914678,0.562709343214693,0.6509289736196815,
 0.09436054088000168,0.314397949315651,0.6410851794910737,0.3159613260392228,
 0.8939346048865751,0.06360448741764302,0.8282596239108817,0.814849989408483,
 0.857140300519246,0.031152850887106065,0.5103383954623394,0.9530925268443546,
 0.6638737814646247,0.6341867551107622,0.5459880471303482,0.42310686763422123,
 0.7051725966659588,0.9509457351910205,0.7183744530411781,0.5023910739119407,
 0.645618721168108,0.9261569086089951,0.2832025879427552,0.7097670934688741,
 0.5754114032440079,0.9145643148403291,0.8001663372817753,0.5768171973749104,
 0.3806574738454621,0.906909614489268,0.3910947963905267,0.8733130101269584,
 0.8599110141979087,0.3040057818901175,0.3243350134839422,0.06385416541136857,
 0.40031989198163564,0.919205701461266,0.35635139479274847,0.6919463505567687,
 0.5162297829969644,0.3301271440536796,0.2072246320908464,0.2794388214398932,
 0.3659079690986179,0.9707094006751732,0.18722704158167724,
 0.029052314518192124,0.9172784887893364,0.6297219095102636,
 0.3808091727739964,0.25757541052902166,0.8001144954832142,
 0.36271687391404894,0.3314419133985891,0.4173023761281244,0.8193230844169161,
 0.07705650899684058,0.6568321077382908,0.8340180456418236,
 0.09472889452823663,0.16293235632744107,0.6069615146589665,
 0.24903679049461835,0.8588194036106456,0.47697566208426767,0.381854500926466,
 0.6096359427821374,0.5264374329222066,0.2994536849590722,0.01606025577886183,
 0.189710681249452,0.8492660871027757,0.5556353134474629,0.07348963410002451,
 0.8309081010421739,0.5679926476879341,0.23085781305512953,0.8143997292544825,
 0.19645043539976237,0.8983111644413445,0.4477075246440718,0.276069884622531,
 0.0030573034924865627,0.08122413228601766,0.7610255113613446,
 0.5494063359232538,0.2684920453996642,0.8577681054168653,0.538425662003277,
 0.6860448599156098,0.1336196592009432,0.46371001952346036,0.5386655820572235,
 0.22265260127625486,0.19292091164736913,0.9350420947754629,
 0.10825768378830392,0.26570930187346964,0.7223367794820484,
 0.27054417340489945,0.7390044443003921,0.9257472386303777,0.6275438365111375,
 0.959160843387273,0.6818020695912688,0.9322082428025856,0.2468238654670336,
 0.3674237964637832,0.18155365309101668,0.45277972136661593,
 0.8445365061482413,0.01260451460183487,6.956090387880698e-4,
 0.2593248329639686,0.4751397393102179,0.8670467328143793,0.6185757115483086,
 0.183987248085826,0.792466805682455,0.9562599371306871,0.1595330444685351,
 0.06747308069118885,0.552117406895198,0.17915937770931234,0.7418509585660986,
 0.2850307343816493,0.9968942145514643,0.11937185335273559,0.5172170017274184,
 0.7041215173895574,0.7152836881748841,0.37475297845737654,
 0.32944165923227486,0.4748980213589614,0.6046060077063696,0.4462913432583614,
 0.0838774155033053,0.8714229558575999,0.7927886868385566,0.9283084570088581,
 0.9606626664026937,0.206319137667049,0.8831051542634636,0.7490585045438586,
 0.8722483018635291,0.8191014191977871,0.21255416852548636,
 0.47437283066603175,0.553355115020001,0.13332089369102007,0.511174087260453,
 0.16245523942076245,0.6408156575874374,0.12066891694814652,
 0.08815041356187514,0.03155544641756336,0.22694626175432842,
 0.4469899508588623,0.26129294269683734,0.46493550359403324,
 0.13463778949016092,0.5034503141742749,0.7183445547555581,0.6881238730101308,
 0.9468319586525031,0.8997093354959266,0.2902824537279697,0.393133065384081,
 0.21302554028399545,0.5169375365724815,0.7162858882153191,0.1876638756800284,
 0.43614484732483516,0.7615330747048794,0.5139322432039802,
 0.46275935123801826,0.5909717290926438,0.21058372609095788,0.875895531166998,
 0.5924292254509325,0.929407138893254,0.8053169106844337,0.29191102794451695,
 0.3723286143503213,0.17339330473721015,0.8214695680473951,0.3794585291955893,
 0.5749616758353955,0.740220702301045,0.9147877752774538,0.8136163591984797,
 0.4768371790188364,0.44086486881807485,0.045681679236545314,
 0.9213474993978502,0.029390718843353758,0.9320660342533891,
 0.44319901080194535,0.0918757211185176,0.6737459333453826,
 0.22016131335430333,0.5754928179774029,0.7085595690092377,0.715022650665912,
 0.9830157661176497,0.8277440767317276,0.14909832950697155,
 0.23064402769444814,0.5943748691431039,0.7027756141185586,
 0.11851619498899946,0.8154069602467457,0.1457051848211124,0.6105193829990758,
 0.9198654404880467,0.3232252349960234,0.8992691596857584,0.3818958159802024,
 0.5288202663105163,0.1564604811888911,0.47685557058354155,0.7040573572015816,
 0.43854260350538854,0.9025381077949675,0.54391556169083,0.6251263660909011,
 0.8048487277706202,0.890315259334294,0.5359863289096367,0.869145962440729,
 0.9492382439463578,0.01562208244699781,0.6175149268400851,
 0.06108652463231323,0.23153171518093585,0.5389113920079351,
 0.8552470146535027,0.9239809860024607,0.14547732190128526,
 0.38374522184752025,0.8875295487189373,0.33975720257956254,
 0.2844253643126474,0.6474081180839499,0.33364423064121385,0.85606122955211,
 0.5675872595966277,0.8864916874680003,0.42979929177641585,0.6569218330383162,
 0.6123358913370494,0.493630164353423,0.4286058673438532,0.900549150393322,
 0.7641507942972675,0.2502535258626343,0.7993387481827157,0.47999833582894136,
 0.023135959144405582,0.22182356381065238,0.44558547516543423,
 0.40667723364571506,0.35818270753658876,0.948095528880125,0.8844264313530897,
 0.5945190614542001,0.8765241265980259,0.5336578368958547,0.2527404614470926,
 0.718390295263856,0.6234660111040603,0.06600164070677594,0.6263999104849499,
 0.9626877794677585,0.6374598420254356,0.8313628668838479,0.18973136027688253,
 0.7173616124919778,0.2953199920825089,0.8887565234166795,0.11661633977419594,
 0.8428972263181269,0.8608435553962326,0.6546317883633554,0.6782848834277282,
 0.8960280907726927,0.19520941173242423,0.2578308688415517,0.5960835991298586,
 0.2985941053730298,0.2870959485995247,0.8551930457662642,0.0831674921732043,
 0.5114992711871112,0.9138980498845295,0.8496896026028258,0.4604722295515017,
 0.14395310201083356,0.6063359331970803,0.2087834328508057,
 0.09578857332386659,0.9966079921876954,0.6283494242268189,0.9043278787780235,
 0.8670100693362716,0.24242026035909858,0.9210314830337587,0.7299858834818101,
 0.3727965827090225,0.23285466880906125,0.8689725031318509,0.7920389417929095,
 0.2911671698332944,0.5671971235043101,0.6842939386069717,0.7323382575876313,
 0.3431439130389039,0.407254292439705,0.5557875219544236,0.9480573166463387,
 0.3220724374905748,0.10181892890035482,0.6590650115301815,0.9668017182949135,
 0.6106746627982573,0.08682447155064155,0.24421782854464946,
 0.10835407376829376,0.4335276306634086,0.48085451079634545,
 0.7675770640764501,0.7794588969694224,0.463452294864643,0.46110449976713963,
 0.4590197947763006,0.4598593776428902,0.8926012938402182,0.3896735101635642,
 0.5898936684522976,0.5492136955664281,0.7728282900238773,0.4242155905313836,
 0.18961103854448302,0.42047528955013513,0.3664302782955363,
 0.8935683075998225,0.7119327897470541,0.19473523444420682,
 0.17854760669071046,0.03013384559760468,0.5143352572604988,
 0.7388163076199935,0.7076351797087317,0.37247772649746114,
 0.10944155972277525,0.9343317941517635,0.6222767860612929,0.3065392048597171,
 0.629050198562657,0.9580166270558325,0.031334310802998755,0.8788217124557677,
 0.12682044930936698,0.2943705977326696,0.9650729163143422,0.6842167523446769,
 0.5159744691569896,0.7486935346720438,0.7002007590598289,0.990906596128766,
 0.45976817998432207,0.9222782851973296,0.9296625021242824,0.9379164513000594,
 0.9695592247146373,0.6332372891042046,0.8221439818200856,0.7681488400508013,
 0.0258339059852567,0.5631280691779084,0.9102495804158438,0.4656491233674688,
 0.84568468738509,0.5022083291703869,0.48481166176433466,0.9539402674457237,
 0.462698868213651,0.8543115643676877,0.16321278674408435,0.38465445556527533,
 0.8694823098616999,0.10081010458031181,0.7847541136084386,0.2551337417768328,
 0.7808857842653288,0.18878876621827723,0.12785892867894344,
 0.4765062416016814,0.5193945392689979,0.03032090965615941,0.8299391451358802,
 0.6713777624171008,0.3402723772514248,0.2860435076743981,0.34460089565347807,
 0.42096166884361375,0.8907084539100555,0.11704558526864439,
 0.0014197334371339743,0.3557620381874843,0.5937529703851937,
 0.08365827588789176,0.12572097583756048,0.9094576768427268,
 0.7695979930440074,0.05279493118507539,0.31104862115427234,
 0.11789356738542545,0.1569738057553387,0.48182265579333927,
 0.6193031774513078,0.7877950337815184,0.7422722010674585,0.5623943923728103,
 0.7128615681194712,0.02907207571755155,0.6719396696536055,
 0.32090994694200714,0.4378321402362024,0.22660192823964054,
 0.15247001948190242,0.5322998153006689,0.4047648772533229,
 0.04640459033378197,0.4776240946104692,0.9320391871268562,0.1406085033789216,
 0.5963905750806956,0.14807750333686598,0.80024201785902,0.8831285521455047,
 0.5704035743643929,0.9223184042858297,0.45193747019491315,
 0.18311528786424947,0.9883064536717964,0.5425924416192192,
 5.439633365342811e-4,0.7503070634366703,0.13998614700308087,
 0.4576615687654133,0.7722226243972056,0.3845020994545568,0.06518357001740827,
 0.31080701861621673,0.5122584187679152,0.9842162677775377,0.5279435673375842,
 0.3087199282972216,0.6446177866448728,0.36147816910505426,
 0.42532475773636147,0.5486676488771832,0.5600903173321724,
 0.07715735651717581,0.5939502119304868,0.7011489275077001,0.3054735484206965,
 0.048735780358260095,0.4221025157261221,0.7142173849109315,
 0.07268417281247708,0.9631445639265398,0.9781402820304392,0.5505623815835968,
 0.16185197362132364,0.9614885481613992,0.12615794789911816,
 0.9159184289139547,0.12038593020393684,0.19170247754558245,
 0.29039483331288396,0.6737562491201576,0.6565640473986916,0.4867290697087546,
 0.6047130327075008,0.45350397578458457,0.883479661200752,0.8855971154334877,
 0.020863814531348113,0.9338488872988956,0.9661109293237102,
 0.9349518142494336,0.6617943367399404,0.3702862426538538,0.14950747189206592,
 0.7291829865104015,0.15880090820341897,0.8130844972658843,0.51492593434412,
 0.10044858243609589,0.33826190986503,0.5116736208301189,0.7199016924317604,
 0.08343545154705911,0.9423930184199311,0.4449608152043746,0.7897824513321624,
 0.6411537521879396,0.5992566143261884,0.8043873658716929,0.09449256743616519,
 0.7858068917084291,0.8007643502447197,0.5082439801468286,0.8863263652742728,
 0.55666246312844,0.8048379930368345,0.4381231624050509,0.6840178356059315,
 0.4019781286287817,0.5196600153351407,0.2906978212338771,0.06991182764369264,
 0.769448672317239,0.7636325898010562,0.81357656719618,0.7915519430841468,
 0.4411398193574162,0.6850358277231816,0.9472182792841339,0.1599336822653088,
 0.5082995823299,0.23377791904910916,0.44679324810781385,0.6340104141573618,
 0.04523206771527244,0.48969247754661716,0.7159327449713778,
 0.046144470536914284,0.3788348990816902,0.020127121350148713,
 0.44277443131784644,0.6075662074946309,0.022632842532513187,
 0.576682974142398,0.44707162559689717,0.265476932643371,0.8296318423970312,
 0.5965533904775309,0.8438606642954609,0.6367731735411019,
 0.0031585733029368956,0.2597752726505723,0.9902938882992688,
 0.4324656829209409,0.828245327779726,0.9100548450116299,0.8684627581484746,
 0.04011326355046385,0.8307621223932989,0.644087875831915,0.9855237336618443,
 0.08185049186224624,0.9497857278255756,0.9469781805388372,0.9541324951440052,
 0.832918272713641,0.5444117102175934,0.6895488817507776,0.10603970420517683,
 0.49163878775672953,0.5001898117647227,0.06460315452569343,
 0.5047081307505261,0.10517242568197127,0.7757446919148524,0.1342758113952528,
 0.04285893672309149,0.9028788071818217,0.12247811330677238,
 0.7085306424098683,0.08003890082408749,0.8145371154175446,
 0.35652854704962944,0.9874634778352414,0.6516990540081399,0.9088658857174198,
 0.4700913594198065,0.3602072016401302,0.6672101250562922,0.7471963924174603,
 0.3102509541541213,0.5468597205448036,0.34924220475423906,
 0.07795912636818736,0.587025623211249,0.4204359789525377,0.9572105857003825,
 0.5757500033054884,0.9759960919005155,0.7641032726820202,0.3429634199464051,
 0.12278078559186412,0.38805444895749586,0.10042691101996715,
 0.7240292820635371,0.20540918592028645,0.3210966261454403,0.800833921445216,
 0.6388144004739664,0.03357470244806082,0.5837993911903504,0.3154360260553748,
 0.9204126427746026,0.37495955722763497,0.2190176883749677,
 0.11711747681327389,0.29184181624458894,0.20052309100147792,
 0.24945633242477006,0.6583865290411138,0.9211487599558064,0.5248122334035188,
 0.6293239554765191,0.25602061121637865,0.7400882603447159,0.5540183480017651,
 0.42661030459860416,0.7786661819922942,0.8329653178040943,0.9054732746681238,
 0.9813037294769902,0.4624059408070612,0.0904980738757607,0.482111480026583,
 0.22729810869792944,0.8712906437802016,0.4952175235482337,0.3562763002130751,
 0.6740442230442858,0.7091252804450586,0.5086719508873009,0.17032283338578091,
 0.44446392277355506,0.380077206950017,0.5502921719729124,0.9240584126966507,
 0.48239884877045325,0.4614628823723086,0.31667270820307825,
 0.4790021843700999,0.18297107095304466,0.9686446125518018,0.5692495258981545,
 0.12953681919042315,0.4712681920534817,0.24516358618002354,
 0.5154803664214829,0.5983270029998766,0.34934613565187567,
 0.09098173403748411,0.16967211647673253,0.9711592883383291,
 0.08246967257633997,0.4859363764636414,0.5997688545216386,0.1818660653608568,
 0.2907602059691503,0.6158555960808195,0.3292941195625583,0.22603689389559856,
 0.3920457045539666,0.906473128615494,0.20317171202660989,0.9333433629819388,
 0.7715615138278251,0.6361108435998973,0.8094964825742592,0.8145428683614212,
 0.3585095216958222,0.16364708233601277,0.9628830185604245,
 0.14869030315550014,0.914866988896663,0.49605239867274364,
 0.22764100062065917,0.1955019401081225,0.7083561762037887,0.9805677267582177,
 0.38843887495725193,0.6447584779431438,0.18621274764408635,
 0.5748376046982434,0.3742805688453642,0.21257410953956368,0.3857613854161316,
 0.8091455340079178,0.2558208214339087,0.8243516643568888,0.04581971095822934,
 0.8687016834207422,0.16940512872722913,0.8065802851150835,0.9753026248875706,
 0.6042379472155391,0.6964826136261933,0.6169667945613005,0.47164803395009847,
 0.8802134874331085,0.9564747303490289,0.026808569059843412,
 0.5529772038500866,0.15194155503607754,0.5904815791558218,0.9921139523823264,
 0.08858002913579055,0.41940272249997146,0.8432127315262138,
 0.6358321775040795,0.3768413068887001,0.5250860936839787,0.5391673782690591,
 0.7806492162185796,0.9951555531694678,0.5786621022527465,0.5379323185523871,
 0.5093180185500356,0.4736445916431039,0.8472866245775432,0.5141205279769115,
 0.4821950378609434,0.69529992998756,0.46551381253809654,0.22207317089500955,
 0.7436958534947711,0.828984842064544,0.41134099453608464,0.4674034861035776,
 0.7367466019879113,0.14462024564248321,0.00902342400394096,
 0.8196571966911128,0.7525171113949474,0.5971316953845588,0.15151528121934588,
 0.4238033900565399,0.028799733048940124,0.39435917675180987,
 0.15795762011868042,0.3570917181002875,0.029105337625936478,
 0.9277452377670556,0.8931970053237532,0.25516741725674796,0.868260243733225,
 0.40019439483961117,0.2349033606940063,0.8423917101044706,
 0.10237083930010593,0.6275290648821463,0.39747568075100626,
 0.1664430856531095,0.9338712783692005,0.11335046080316813,0.0876435552541488,
 0.21187379495497594,0.3915874762799437,0.05531694590767833,
 0.41959852692120925,0.5919003723338283,0.5800285024440596,
 0.07400699697627477,0.8301978584774183,0.37092625719927375,
 0.5563302268301551,0.7788941545575286,0.36972561756604394,0.8938396181272437,
 0.5519790640495766,0.7420426405860947,0.17056934581722794,0.9692508334132977,
 0.22661438591065242,0.6408501280180412,0.05710358788105652,
 0.7942495022969159,0.2651750182545548,0.40856195772055415,
 0.29583013043515316,0.3962046426751623,0.7835478461808332,0.8672827225676129,
 0.9687768075501324,0.8603628832844256,0.8002759028131737,0.6663259935299312,
 0.8819709054213236,0.7815451684134013,0.4708095084971706,0.978522233467552,
 0.7475241908630853,0.023089154131831546,0.7735007545033578,
 0.7413595266189996,0.8741358806515183,0.14856134015747535,0.7572904299924115,
 0.8027183221581714,0.47491378709598386,0.01918141822648456,
 0.3428997214450955,0.5991782674279187,0.6344199816490308,0.557035599183607,
 0.7607335341239099,0.2609908086141294,0.42784546564776127,0.3084786452066296,
 0.0962509982404568,0.589411823327778,0.6553969866407952,0.7278643320028707,
 0.6541708867832704,0.7157407083606384,0.965421205333556,0.5753650534232957,
 0.8586558583833683,0.7806386109264861,0.5222116106764985,0.6291045310020292,
 0.8032196009978316,0.8984828535489843,0.4199728214134608,0.946950711105626,
 0.26370369529611504,0.7114734589053567,0.44765659075268893,
 0.043142917258422675,0.621787406746127,0.8137075004864949,0.873531696327719,
 0.3874797255939092,0.6988010711996367,0.265119192966339,0.5877463186804098,
 0.6191760981507826,0.6495615111342988,0.05537428509658471,
 0.22352270047735578,0.1550701602339235,0.19765039193234757,
 0.5927645252851155,0.882252675375019,0.7386980441593746,0.5861244016023275,
 0.005128878223428002,0.49659417676481865,0.21992230362279164,
 0.39464402202977245,0.3547196158763235,0.3097154585275199,0.5939155369975012,
 0.25700170409983036,0.16616963361588122,0.9282664352339879,
 0.2960493214726334,0.41920685284259296,0.1523609723712387,0.8894887282761219,
 0.12912905032113142,0.8215399083358609,0.9746615048862424,0.2036791006923988,
 0.8223259263058567,0.29461922006110963,0.8227432112993567,
 0.009537756901646044,0.2717354920431262,0.9512095366689447,
 0.9472362785782765,0.031228503889928705,0.1465423399122091,
 0.6663884200122647,0.8568961347438979,0.46186353739477903,
 0.43357342227106455,0.11218626982175128,0.548069881177261,0.5129521947630747,
 0.9426433001416766,0.5249596553926963,0.21594148369362476,0.8063668864846302,
 0.057010264824773116,0.28931360927889604,0.9235028848685681,
 0.11968131587530051,0.15698273893958914,0.6772067587519822,
 0.8561793758480764,0.4334997786500424,0.10350805024623877,0.5332819111065422,
 0.3365948571255537,0.7683573340860919,0.6940449053196865,0.34796079990386763,
 0.7824901885437432,0.11356463395888983,0.5595557612007475,0.2664777953115316,
 0.7889393375698992,0.7610575627224806,0.1152640221665926,0.7207222192003041,
 0.6835932525462134,0.4952787270174326,0.6492837874844576,0.231271861154012,
 0.2931636023199158,0.08225237938508068,0.2109519291734374,
 0.31029095498554626,0.01242809236413378,0.9856748862908512,
 0.09580074725106669,0.052278998215135486,0.951575621348289,
 0.6545520608224076,0.714806380216363,0.9343272171341533,0.9836463632299792,
 0.4218338575000887,0.23765170906448974,0.16283124438235286,
 0.15651229991513127,0.9669088482115606,0.15912695119109443,
 0.29332031792491864,0.4235849674509857,0.9227131420222908,
 0.04340646760615119,0.43138556000593997,0.06742719217819382,
 0.841901946290586,0.9435543246163844,0.14357800193834025,0.34807814197961473,
 0.3768166032419946,0.352478121974954,0.03349786514808473,0.9994177910156639,
 0.8419867219421209,0.7760226092721291,0.1425896739185729,0.7506751664364327,
 0.0815105228492694,0.5560372636016114,0.6814833013526614,0.6910889984242581,
 0.026966379097221216,0.004240304557239338,0.1410852985082851,
 0.8195270599290887,0.9440907853303093,0.3896302843932672,0.5696103316238285,
 0.6776937088445528,0.27990507790179375,0.8503167066135017,0.4369601543441577,
 0.03857877126575193,0.21354655094088204,0.07379824971304494,
 0.6816044576212266,0.7929758262667709,0.43918244693672404,0.7905107465139469,
 0.11131168574310923,0.4550741973243507,0.6218272545713327,
 0.45984801865256353,0.20477421065987766,0.09627375369391666,
 0.6152331650272552,0.49537283372230767,0.3655703487408911,
 0.38180950294325444,0.5838087892435601,0.19711575045749397,
 0.6289246222569386,0.8824458058081648,0.3606192647713039,0.4567223611970699,
 0.1317469332626776,0.2048436915428561,0.20532152576411644,
 0.02317215134181516,0.47842564852318104,0.168054181750937,
 0.37948405922834527,0.8578241309883096,0.49629546295740146,
 0.5892407842886623,0.8287032119365684,0.4184300145212487,0.31090839954932314,
 0.8682453949998573,0.32478336575370215,0.7257270808571257,0.391913762187708,
 0.1830654190075679,0.8253797070898192,0.13770543884319164,
 0.050835994167429566,0.7924213530236763,0.1376798134432593,
 0.32280160756829224,0.5824552194680247,0.6393094331319297,0.5706141936837564,
 0.3510204534809578,0.2206812353475074,0.23294414667729746,0.545841652996895,
 0.6200997432609665,0.5149838337478422,0.075446249129283,0.26729928953632887,
 0.8766471559849592,0.596067775793448,0.7615939540187789,0.019435051982125695,
 0.22970596508597851,0.2354199737862679,0.19653557943020017,0.642189788461238,
 0.8897599099260086,0.3802520298116747,0.8461614557728798,0.539419853283535,
 0.2856340301065107,0.2837835861595803,0.2326552710933495,0.24228970174930975,
 0.5293426764306822,0.8955530496942721,0.37178103991723266,0.1902651825191196,
 0.9716408065751415,0.33152598807585354,0.994119937024252,0.4304950285074409,
 0.5521834277269516,0.69257583472993,0.017686769746981246,0.48955042603980947,
 0.30231681500567775,0.08365552257200337,0.09966184074000506,
 0.15674450455133182,0.9402895612418145,0.42776767760394496,
 0.32174806487545926,0.43188684853094517,0.1452536901757736,
 0.07619398530248833,0.31244114502645837,0.8104219915369717,
 0.2605670577133915,0.2758242907211128,0.14971653326126344,0.9590912519879494,
 0.38591949019264093,0.2034885117427916,0.6470531337848341,0.6411830003526371,
 0.004573715131288347,0.6459794036139443,0.8765753475904793,
 0.8511104481053398,0.7287528900106278,0.46430733101502086,0.8475946145112665,
 0.4024312012321034,0.3698302355652531,0.49812288683178474,0.6916163300420447,
 0.5641792767778936,0.832711277525278,0.4787637529756339,0.857513277205519,
 0.9521181168109825,0.7224480577870757,0.5493955432417864,0.6949118398963374,
 0.12448432540572907,0.37516384820265447,0.8518833829417108,
 0.24388941525579444,0.176345937945537,0.321572128611599,0.07060918850681708,
 0.2728200173065045,0.07748142886217202,0.07222074130867673,
 0.04387263184896639,0.3749052641521671,0.07763540960170456,
 0.14271009481453134,0.9167285093078242,0.05981752405163332,
 0.24760922531297236,0.15006995701989811,0.7547101245233976,
 0.010397270144227866,0.275213309824905,0.18367555279016556,
 0.2800926718126874,0.5314234180918586,0.5700266759681977,0.48259042966964194,
 0.7874806640351256,0.15097500522418972,0.7442345720564806,0.7035501478521028,
 0.6712950988756754,0.4145077003829105,0.9520204274028466,0.9353863478859371,
 0.7841231916905078,0.9394117347585603,0.5096898257979414,0.19348020128585186,
 0.7243454550904901,0.27523890237604065,0.7083172657357756,0.7251522054561094,
 0.5333076636713459,0.1150799496579118,0.52593678362908,0.43726578718363296,
 0.7938408314690747,0.3377118051235318,0.1390744510916827,0.5155026706912986,
 0.9166479769257082,0.6386888591181756,0.1192173903456768,0.10495954230203541,
 0.7913531769468873,0.22521002578277327,0.9844237140212848,0.9181165097601288,
 0.004130339136364181,0.4880581916693121,0.30187316134552433,
 0.4789041288009689,0.32622404434999663,0.2021659507868684,
 0.19002263155501697,0.8993336730018157,0.48184636676117454,
 0.7988449634981448,0.6967617609226525,0.4871823176757214,0.6831474442510679,
 0.6929929015469004,0.4203165409903584,0.15401649675160567,0.5865015967327694,
 0.28812330075839854,0.08927585980355279,0.5491977096907747,
 0.23128432946257205,0.9128923030662693,0.36871076123657753,
 0.8274162304146431,0.7587371933001514,0.9426375365653429,0.7819448229165252,
 0.44099465762024326,0.9482221130163011,0.7580776531990221,0.592124809678284,
 0.5716198571658465,0.5257403138261325,0.9414080010388338,0.18576672066794542,
 0.8367270511131248,0.7614237125126159,0.6263546269060664,0.2148519291693508,
 0.4762875574178157,0.34511817901649466,0.8747235445008871,0.9326141530620636,
 0.962047379738912,0.4297981761399359,0.49410066551797394,0.865124478235933,
 0.3790565153183576,0.5217221632405059,0.04561617758854375,0.4027757672133361,
 0.37303515573338586,0.15138765760724204,0.9618073676609686,
 0.4282686057614772,0.04122467288211207,0.6868939198616291,0.8295624637539138,
 0.7927419133327027,0.6896353588869314,0.3243649588327645,0.36226764567212166,
 0.7769720518112327,0.1514593778715756,0.8442010292851433,0.17346427442507872,
 0.06496924177477792,0.2803669207229108,0.7328251949569019,0.9767213112971794,
 0.663350250097456,0.7278529025489521,0.9779406084962441,0.07684342518883336,
 0.33719707913843266,0.5860652450026336,0.6986739140035424,0.8357798999250762,
 0.04421777769930679,0.6143802704434892,0.9345557827275338,0.5308155413841312,
 0.13509264070629523,0.7879444008255053,0.016412488528485758,
 0.5198418529351577,0.8008707041022982,0.7680627015979772,0.7790545034203843,
 0.5931574783543412,0.5436457521823705,0.4079624566050116,0.12493657316758455,
 0.3203650782285611,0.28837197782399615,0.3179695780976336,0.8591948421533838,
 0.22932277308878934,0.5552320423907139,0.45548940896862733,
 0.35627181155366405,0.78425569387774,0.6334641666656982,0.9473203802847637,
 0.9206145596525032,0.3936530841258141,0.624233991558627,0.003128634789483109,
 0.07381742495041477,0.8105360096811504,0.9852795278449057,0.511821875221822,
 0.8229276911593654,0.2009428573400175,0.08549853947936548,0.7822534093401833,
 0.767667239336605,0.5801779931681674,0.7548976941669744,0.11688455614520321,
 0.25762481052479524,0.213448803336862,0.045955399013475895,
 0.26737845070575206,0.24325574677781336,0.45563179839769186,
 0.031864852034434454,0.22830714979528444,0.90533312506306,0.4269490968446503,
 0.07856006544095917,0.2647102369705847,0.9664829341114237,0.9282821350014989,
 0.6032550698061355,0.7982052850383559,0.8839675260844692,0.6551581969383886,
 0.744170169037644,0.7603780522467737,0.6169268912204628,0.512509363522459,
 0.5435562074347619,0.9164898553228438,0.9972331356050604,0.4347725048929155,
 0.4133068315562868,0.2617172798952697,0.5585401305869866,0.9257457027971958,
 0.9472844465433657,0.24404528994319563,0.3785064269348297,
 0.13119662502259044,0.8249121464322671,0.9116999667910948,0.4122953677089659,
 0.3865948121142818,0.06474060192616127,0.8370775141746992,0.9054898018389403,
 0.36439484681063616,0.77424684610122,0.5289180819288141,0.03122123849453229,
 0.4461000865185164,0.148200009373608,0.2276459732900762,0.9366238989510397,
 0.6888980977344878,0.9306865077982667,0.2038682128862036,0.8723734141051332,
 0.3884524608417239,0.6635730268100266,0.9894125162083629,0.991646041920592,
 0.25189786384685575,0.8673903207781103,0.7310370104016756,
 0.47263967590922984,0.29650175172349624,0.8837082078720521,
 0.5392209380965207,0.1890082871504497,0.7034373281061644,0.3630958109547745,
 0.05830091156997674,0.40349640414841526,0.13296784622063917,
 0.1577062963961725,0.7659728064369236,0.07277979950007285,0.9129084179791622,
 0.9142963113127082,0.01640984970218584,0.8694603042226317,
 0.21245087354443515,0.9294424953020328,0.42640876677375306,
 0.9884956364603031,0.15365410176224392,0.43383890513930257,
 0.07244757777632083,0.9960287292772713,0.4074261940876224,0.835268416201427,
 0.6600119196650965,0.4930294991833346,0.5929731641602431,0.20454516271477874,
 0.4978953713117129,0.9309138098540741,0.9710291729101,0.5696364813366361,
 0.09444184632857877,0.519429284601165,0.6312136811792669,0.6278831733079929,
 0.2872459579870701,0.32025776405446715,0.8117996706081698,0.8909412008686453,
 0.5744703828176541,0.08399537558777947,0.6572306860508093,0.516848548860615,
 0.5549347841725626,0.4124589817326424,0.797370243779179,0.6073759316188531,
 0.6918810682013473,0.6103643469664379,0.48830002361791824,0.8628307933716207,
 0.33627192141620876,0.7255564268458348,0.14962754036383208,
 0.6777092634240094,0.5198129443821081,0.33564220786917154,
 0.43516328185149233,0.08666535364241879,0.05092985939723127,
 0.38288176556439546,0.539647062207948,0.9208045550611943,
 0.004762359626792501,0.9767895126944103,0.82603967891676,0.7114279099331853,
 0.6638811761069274,0.09163474164842234,0.3974719413187803,0.6363463841876044,
 0.10095591890982702,0.9978004344686463,0.6883187574229501,0.5782694640996642,
 0.18362385275097837,0.030108042373924282,0.7799737675095215,
 0.5329214072149059,0.377210547198336,0.6637535433197905,0.42586339140805496,
 0.19693184782669682,0.8576427721882222,0.48044781717436524,
 0.04644144009578677,0.3520232827465415,0.6089847242035751,
 0.10040837472699615,0.5061735013483086,0.3485469628331527,
 0.47032965657058345,0.7859627679668161,0.847749015993247,0.9638142308423687,
 0.695347628396425,0.5255885473140215,0.6389355331118837,0.8869405323519739,
 0.7138661733183946,0.9937257854222443,0.3052718564754924,0.6420420120252394,
 0.648221929727146,0.3410371749558093,0.07847548097127133,0.22025744890621957,
 0.2823411426758271,0.19237118085779326,0.2827215636382875,0.928400739451147,
 0.4015066778089148,0.7817754383281462,0.003604371023953168,0.742267346849701,
 0.4334891364467457,0.06986734925112947,0.13420543624805903,
 0.3245139352026841,0.7199816227369746,0.9257438740638825,0.6510779011771219,
 0.8834313792338518,0.6584948804020301,0.6300678370281337,0.19234828651075875,
 0.19056203775265468,0.2310724850464565,0.6018253174462463,
 0.03135631072236178,0.979652013378651,0.2896277064594456,0.4669682659184019,
 0.3975786759876194,0.6511131036493845,0.6450636914821216,0.4574813501594389,
 0.7447041512816046,0.17980219110920115,0.814580389225695,0.19540431071610986,
 0.8476376020603757,0.009274920033958911,0.2946436797587195,
 0.8920580184613519,0.40803943864638104,0.700127491011213,0.7836172914140103,
 0.8180051598311068,0.16811486519975216,0.8548597611555716,0.9373940739457547,
 0.6715009362966853,0.522850094428763,0.12269504848183366,0.24570573608552793,
 0.7483695732553357,0.6181448638351056,0.4103336351975866,0.21759375541606074,
 0.718794853292991,0.7760293665057991,0.4101354791219981,0.42077222242452006,
 0.262530038168331,0.5450737637432774,0.20058166930247323,0.6338725969857658,
 0.8978271484717187,0.393699155388891,0.09736376822523174,0.2450564785052044,
 0.6444142721402508,0.8632287078789396,0.8887988551375438,0.639830841859478,
 0.47716430953730604,0.17748705517740038,0.426655711591568,0.0388790306787169,
 0.9232251897174765,0.2073243108896805,0.12168297285373086,0.3840950200966735,
 0.8000252070737379,0.6018415179591386,0.08478622516270384,0.6804118419942308,
 0.41355197529786303,0.6372555137005618,0.8408995729989762,
 0.014831199239871418,0.9999078531508344,0.3662290096280473,
 0.0817314212019733,0.6316880076590777,0.1334025077761568,0.1322606393145458,
 0.6383848263985246,0.0435314399023049,0.2735279895331838,0.6730271106913942,
 0.784773169782514,0.30068493764223725,0.19760385948633152,
 0.08865231919977457,0.9723859815865972,0.6144590985326772,
 0.08009515275225665,0.28962884983200543,0.8056047248033861,
 0.03691359491455459,0.327497952158068,0.29047139267384603,0.493838739329601,
 0.7978280162649914,0.936446552229755,0.07149453004864803,0.31955662997459056,
 0.4750875755797046,0.5178173622747969,0.5609504260096977,0.9265415608083712,
 0.32605964449443126,0.4115628386306732,0.21231443119077875,
 0.31356805702090373,0.5745688711735504,0.8386952865693205,
 0.43191830264390985,0.39153036912231787,0.4854358035467782,
 0.03038968674672171,0.058261545004392956,0.2793645798010499,
 0.9200842321230382,0.4122359369740999,0.6031492762659024,0.07318476822906472,
 0.8706265615489021,0.6976763845620654,0.594654331850096,0.5245406262731899,
 0.625669520993882,0.21396877892949429,0.6032156158483173,0.1962361789918905,
 0.842096951023058,0.063961654139847,0.6040795415395338,0.11276259381596954,
 0.5694283880358598,0.8362873449536784,0.7359119830389996,0.9068830878979612,
 0.30289044584782354,0.05224307039073206,0.25475078969589504,
 0.8623385753839254,0.8146690784218376,0.38633642808016755,0.3897748653979278,
 0.40752172753685134,0.5387164421829587,0.25009080589601373,
 0.006608339302635491,0.3949233145959443,0.4814574631445595,
 0.9081793021531442,0.5813632901261414,0.07920254830051121,0.8361094649089214,
 0.04100560485058169,0.5137027784450952,0.5996084739757593,
 0.22899614816251912,0.8110458881035216,0.014161688990317223,
 0.8367024247831271,0.5727175097582405,0.4905327892810152,0.30408926486905774,
 0.8874560444651509,0.23474808994042307,0.6188789853752718,0.8540305411374041,
 0.7740725954718654,0.062454254970094936,0.7235001725345246,
 0.18653700903685289,0.3936060135906718,0.7460426442474485,
 0.08011856664827821,0.5765373369881768,0.07858355382080373,
 0.3454210379993632,0.6430289443829851,0.06645259957900096,
 0.34353097034514635,0.08648035776477336,0.27333213815408774,
 0.8235559809076575,0.4470866502901826,0.12632419597994704,0.788165087548274,
 0.026843896568437775,0.2780316375783669,0.23327199796882603,
 0.8125225771307102,0.8917979172537255,0.20377587983887568,
 0.06783996764593092,0.6206506921174944,0.8064696005697385,0.9082518261064121,
 0.019870008544894735,0.8323979207359485,0.7305013166866333,
 0.15308813713964586,0.1620708731754602,0.17629182013591316,
 0.6907655723709389,0.7055294429878233,0.25273837758966433,0.1156270470739994,
 0.04892302813572402,0.12422654147278556,0.8225796664008767,
 0.5715427596031519,0.36546503172597156,0.9032122047941876,
 0.23875471521833802,0.33165332760575983,0.5193241412993201,
 0.32146205176318143,0.1802618083538794,0.6431917359610987,0.6262174845681576,
 0.630090276602765,0.15248807309136503,0.7227127002393868,0.6151218298549898,
 0.9658632960100532,0.1486469558173248,0.5944006912838395,0.9634605726893413,
 0.43550332416749427,0.06048019011174521,0.22413560683293432,
 0.40152779304270736,0.7506329077734661,0.2808418991350523,0.46149143643377,
 0.4781073266544408,0.21712759036431184,0.2996308907322358,0.595335242275873,
 0.15906398024791135,0.9081787065621545,0.04934429619127201,
 0.7631289357242236,0.5545294537110326,0.11323976412256032,
 0.48737940004006053,0.12016152792151813,0.6702769969708046,
 0.7166042380625666,0.016765437187389542,0.8388642009163,0.2914512352904144,
 0.2449024488874474,0.7782495661156085,0.15098896716910992,0.3629276317675407,
 0.2538667073288017,0.7686155527252801,0.28778960872074755,0.4778423969749439,
 0.6255476604535579,0.39896697873107845,0.9880405246557958,
 0.07097833127002373,0.9553717604899074,0.8279317342275216,0.1368239363627205,
 0.9312839611183796,0.7124664690386328,0.760187163280205,0.47031644524957594,
 0.9581022369319886,0.27449233033030507,0.773799704347837,0.6782248035611261,
 0.7349798486199665,0.9955411965415993,0.15638940186778794,
 0.13349878813207605,0.4311748090100471,0.2480695502469734,0.9995029505991189,
 0.9939898938142441,0.08513889112434936,0.3780496800983796,0.2558719639919994,
 0.2257591374971436,0.5251273608762574,0.2359708219851966,0.7587258806562598,
 0.29351636762997424,0.8829288598599698,0.48446501015857124,
 0.9119737886874929,0.7497168742107555,0.5098741773780997,0.06024311879512001,
 0.7615760612856636,0.028661992411830983,0.7407517533851863,
 0.017126800628865846,0.7618571425315028,0.649109873809449,
 0.06275530660007078,0.6781975327329,0.9364612627270966,0.9562038234061245,
 0.7820699693476701,0.004639880419015641,0.3604301939358262,
 0.7555884122566656,0.6599509253530274,0.354045857005979,0.12193050464892474,
 0.42013684595680467,0.341582161917964,0.26013063586715246,0.7985313061139508,
 0.2805417375413841,0.853086214632087,0.26237883053334965,0.894101422190758,
 0.8702612972281646,0.8192130442140683,0.9106357482119475,0.6511620581569598,
 0.5069978740058094,0.09755647348702856,0.38939753446540526,
 0.5212589111190383,0.8665332945712156,0.36207939521958,0.27457215546785896,
 0.27438481086312316,0.3220743564845888,0.35014818043104345,
 0.7726643063550448,0.8657859128370802,0.514057163251005,0.638401888739252,
 0.9552418506666175,0.4801691234022726,0.6175951484568886,0.49649703971534,
 0.2792240220157538,0.26393071934336254,0.4981957119967729,
 0.16907605414090487,0.45113688288702425,0.33330962505183104,
 0.9338917631980173,0.3132870349149679,0.7312962961684969,0.07982832552202324,
 0.31888795586666707,0.7100718653088164,0.4382966516389759,0.5369719239829486,
 0.10077040506702761,0.8924656209966174,0.05440500411948934,
 0.7525460634095958,0.6498755487016071,0.5738541376860632,0.07596670306699838,
 0.2657588527119302,0.9797877713320067,0.5099151334444908,0.6199439905920724,
 0.333647142205621,0.35885895995855244,0.8009450043883848,0.8567355457684537,
 0.1924434672156542,0.29225514569598876,0.4527393709730877,0.8702926067423509,
 0.8363391080583574,0.39714928565628227,0.7273972646410831,0.9014377969937284,
 0.2787565953360214,0.568493503490806,0.1542755008066884,0.20700743365078755,
 0.36878232431265756,0.27176073878605733,0.4354999592907284,
 0.28605842354778044,0.19606713475607618,0.680173874268343,0.5798347782521877,
 0.159398297505138,0.4555962042815982,0.20694996921475584,
 0.003683711835610559,0.1808544668977714,0.9697370793747154,
 0.8703998540739022,0.6411763408442556,0.1810347827805825,0.42614896877963904,
 0.6612391367762066,0.6026212434709881,0.49359167685906524,
 0.17340145341270574,0.5241688453236392,0.17517908841690033,
 0.8531929562266101,0.9937136318830343,0.2503972025720972,0.7285449428777614,
 0.5105784671835687,0.4602009178799348,0.6531256329073016,0.2971127748684803,
 0.4114467765660079,0.6544549725157015,0.6561373788316682,0.7289716997676502,
 0.7372083342706186,0.49691946735282666,0.2983445171005942,0.8176053246077502,
 0.8491217274126903,0.08690763612992525,0.5115272662007315,
 0.42159784297656533,0.6968766165322913,0.3631661244567357,0.6687492094682926,
 0.8893286993069275,0.7465660235339954,0.7293114956022425,0.32275494642500946,
 0.1216772357699708,0.4182854589757584,0.8478371612439997,0.9264396567353631,
 0.31487985534813256,0.5718660907455326,0.7163690348332974,
 0.46836936583885613,0.049027092545948725,0.7245840816330914,
 0.39684182389652367,0.3480822055091126,0.752633508429921,0.2557869600931384,
 0.6611637715734983,0.06084620945822439,0.3707498817563244,
 0.23358066702115332,0.7603913589348181,0.7469421860526,0.8806690882147352,
 0.6626685109459818,0.990591351874838,0.8185699833284784,0.404849819769008,
 0.057319349632108385,0.3121493009318248,0.4098764155812554,
 0.3468662202745949,0.02104717579994375,0.8370454605446076,0.5787198921396388,
 0.6844909270989048,0.9262551763788336,0.9811327818515031,0.39170008642850107,
 0.8903805738573091,0.3560856870002749,0.4174387537508333,0.625331191669424,
 0.7219191205916469,0.2494243740587353,0.6977290003290503,0.36969821677456993,
 0.6437044078085565,0.48088311289315466,0.17994992954142153,
 0.2879013108448655,0.38471606085314425,0.3682619353649046,0.5831318764978947,
 0.45257833291841565,0.9662565010073021,0.6722807453876684,0.5013252602366995,
 0.13375894627535645,0.1443986158443218,0.12535369775062055,
 0.23953150967535042,0.45112066129184536,0.4506262745553723,
 0.7820031439166972,0.15675303225202286,0.5936160584362058,
 0.48758454950854324,0.4693251123060169,0.4429815374973005,
 0.09643219750215948,0.8921750977779804,0.054182678865533296,
 0.013182532759826837,0.23598242880586118,0.09393954273604443,
 0.8282500692732161,0.7272840186174222,0.5037194529630791,0.5856569527000649,
 0.6770642929668497,0.49906154619309884,0.12522239463831508,
 0.8288906218068076,0.059688288195714256,0.9671599470556214,
 0.30821416070686525,0.547370677926827,0.637751185536073,0.6227369018341034,
 0.8490357097487791,0.289356023388324,0.01136897325424091,0.12036469731313015,
 0.12849919085386552,0.43713193607975465,0.12408677171835858,
 0.9641297361529109,0.018321426147378084,0.9406473837778067,
 0.04401770699379093,0.2067609698022278,0.18802931192062344,
 0.4195566443514209,0.3777778000726377,0.2937032475376855,0.1870520628315656,
 0.8919429608866807,0.7069093111368298,0.3692519329760837,0.46137337027493097,
 0.3447646966379523,0.6751888975446262,0.49473247758786953,0.1746091266758073,
 0.4507021666036627,0.7320882664137512,0.6167056230528364,0.6130975399641893,
 0.808461028450894,0.7525208811125561,0.7793173858906787,0.42246583724478537,
 0.4031636803528642,0.9951289521773059,0.23196731893300992,0.6977097481422803,
 0.6642228629972483,0.3148333096511584,0.07268289934108596,0.4583487492162873,
 0.1239504722618473,0.9891101079870595,0.9699500907020172,0.19818136069159809,
 0.4578638006716341,0.7790616485222391,0.6472244310766588,0.2114037065111879,
 0.17087926637282735,0.9037351662203006,0.5198881693195183,
 0.07581462239632897,0.3295694881079084,0.862324603366707,0.7328899097954806,
 0.34965245467302675,0.47347104033383736,0.7761897876065782,
 0.5708114795882574,0.34614942835852947,0.09089577929627923,
 0.04385262559889869,0.5374679862296368,0.3310463617905761,
 0.09074703798503703,0.2790972777487295,0.545894424296505,0.16762129534913162,
 0.44662570716847394,0.7371122975279628,0.21885196891223757,
 0.8759175607614367,0.6774391221630397,0.6350408662065613,0.6825218656779815,
 0.7335452128677984,0.32495908096324366,0.46588560912593224,
 0.1299039493856261,0.7964010464242337,0.841145292042531,0.3641655066954047,
 0.8512411952684329,0.09428666951463993,0.04182782910338345,
 0.1472355239070594,0.47172488144397584,0.8325125255504401,0.6995306531152317,
 0.8184345472078918,0.4003858888403594,0.6968937072729239,
 0.008104482675813318,0.5112422855095526,0.15447656065631898,
 0.11079853462988354,0.9823523014143979,0.4228623079791237,0.5383259377482374,
 0.5055174387701662,0.561314496806659,0.11387852582298208,0.17786429483462318,
 0.4106613381500013,0.9060300512514949,0.8877596827815029,0.4086579344379049,
 0.1868511725156765,0.9911141534142753,0.20212134529665804,
 0.08695399149451855,0.3445769777739621,0.8657628355010387,
 0.08637272458922318,0.12276221083145211,0.7795598123384798,0.917434710737905,
 0.8155006738929456,0.45717951421731895,0.2570627017036663,
 0.33356093106997375,0.4283697465561811,0.34446674798885735,
 0.08682008364312266,0.3200038280546339,0.12055218499771803,
 0.9308599948336742,0.6132999771237804,0.2856315611495719,0.6384181586051341,
 0.8047280694013882,0.22955229565359214,0.48570190121409096,
 0.7792161150831884,0.462387140541895,0.5803275116265111,0.46062242072944437,
 0.6959499696330907,0.037416693639029,0.685036900905367,0.44386524580676434,
 0.9411315714927668,0.03543185240888669,0.2633048056822257,
 0.41177687797483375,0.022887418861405218,0.7005331370292084,
 0.5461290236902887,0.38074567587902164,0.09651341188438312,
 0.39318744719806153,0.5880532894718002,0.7223172081548247,
 0.09991728963221325,0.7766390411716882,0.192957457074319,0.17669582131668027,
 0.5242782362197344,0.9989401175653618,0.792823216969026,0.7252004957495988,
 0.6301871472305757,0.59222315327583,0.2306913547863274,0.15927289951025536,
 0.007129374856531201,0.028080205682360027,0.7574737131097808,
 0.9510275724635239,0.7562080549197988,0.43921335854835875,0.9028076679837002,
 0.5044260472217301,0.7581727075422002,0.10854550689567077,
 0.046260516883054326,0.647171347230951,0.7921120969950246,0.9143411551507677,
 0.7144665243072771,0.4742063051707842,0.2249165438353501,0.08540398419609352,
 0.7786899105652955,0.19307319423574232,0.20460981114516752,
 0.16343101483102473,0.7276917432971264,0.01702908389869695,
 0.8295799458580748,0.7198926831246725,0.43486897357208065,0.8508194430912691,
 0.6167648714664966,0.2684307124788814,0.18521278682738873,0.819104512124359,
 0.6850492804438733,0.028777078527133737,0.809009523627305,0.8238527698487843,
 0.3167387561139071,0.5642697451864535,0.8416700527546034,0.5315963360345128,
 0.8061692522820822,0.03067590589153224,0.9357901095048858,0.895717828638013,
 0.518726220189267,0.1327581904535149,0.8862809722855238,0.3346535795967016,
 0.054219071398535434,0.4003097354074975,0.6382986212527552,
 0.3973185850648767,0.08995539467992797,0.014359067605459686,
 0.27159297061609544,0.2569040440310443,0.10775023679814744,0.670583822247312,
 0.3427804406111279,0.9628277961995158,0.7439202938459707,0.14865644422263524,
 0.868576504083598,0.9479871548278793,0.9009696468977358,0.968692106896384,
 0.5753266628715604,0.3452375639015067,0.6249911209537736,0.736574149613343,
 0.34529935000989365,0.29802905764472465,0.972019604360485,0.7223618348273425,
 0.7058313909761039,0.7275407907813651,0.3479231327246717,0.01638010079537222,
 0.7675471259543262,0.2516951593587725,0.5468601366937632,0.3519979072488716,
 0.3707910915189825,0.9541792687927254,0.5015449162956739,0.3058178367782396,
 0.2593247122112615,0.21858624572326435,0.25665655866236503,
 0.26410068118485064,0.14687763370068618,0.0709856672720548,
 0.9758040669402597,0.9246150968141635,0.9701543488720484,0.05129200070642426,
 0.8009931812156175,0.40926477504112535,0.8920207427353453,0.8576578949055567,
 0.8437430763446037,0.4111135108981808,0.6591867089080303,0.45911973151322116,
 0.8122176534199514,0.30015235303292576,0.3989102651349248,0.3914064074441217,
 0.3384101554681642,0.24138450364340958,0.1740211376484273,0.8064237543537052,
 0.05341067749450357,0.2857001535635768,0.28883558283929733,
 0.6616141698020098,0.7141672217978267,0.5058272957500244,0.3052310068582297,
 0.8508944959668152,0.061734653366603665,0.2798508032602347,
 0.4846324855256823,0.4136146499642708,0.6340072831408359,0.2563441127695768,
 0.7456462615537978,0.6722684416317473,0.06944249311411055,0.7397490396938853,
 0.5026586030697983,0.4650497539130739,0.04178426342864061,0.4650937159359422,
 0.9636825614121718,0.44028897997460126,0.39389021068440533,
 0.9506850821349758,0.7951707754498293,0.6815673517088525,0.14878218836708157,
 0.804406125327655,0.19170062441806845,0.11672164929667095,0.207290731945746,
 0.20557189646025076,0.1264574231885528,0.48162167126710653,
 0.8132478413227296,0.028701658248859285,0.38777025010803234,
 0.9277406462159139,0.8444563855734106,0.9090786592598686,0.6888014994577352,
 0.6008876144715875,0.9817905405384867,0.7415702617957591,0.2336854076548358,
 0.26148222177083624,0.06908869140362095,0.9109925806354331,
 0.08129937758497108,0.8849807563852828,0.05740676047335869,
 0.8933108644433384,0.5528990295962802,0.9276898951339199,0.2541556701735651,
 0.23444483799397042,0.001339474346773839,0.27124566835312913,
 0.6829798012181868,0.08049641867519919,0.850957311416831,0.8098569340649887,
 0.5791710921362729,0.15921946017519284,0.22204435787403953,
 0.3688328135345692,0.7977923607427632,0.5229893701322235,0.31007482498835603,
 0.5652571778630936,0.8307640640426284,0.8320274975440829,
 0.038942407212416597,0.4353321674979064,0.6347298540402965,
 0.3575804189227254,0.6777582223654162,0.7610913649716331,0.6618098833851919,
 0.36753941657598377,0.8064148435525095,0.16788414787422223,
 0.4405384321115454,0.6649835189387358,0.5636570217419264,0.5794153478999104,
 0.6783751837754698,0.018146280677266202,0.24597174566244417,
 0.04917258377587552,0.08885232226964568,0.5164077936917134,
 0.09222109887273922,0.760039204966773,0.5020070117163346,0.5230334709263119,
 0.35292788291244914,0.19099202013797323,0.7705832475217836,
 0.04244203358347243,0.38808962796373925,0.16284453898254103,
 0.3154524673960497,0.9382988805638955,0.8093611760032287,0.6239983416369316,
 0.5110631553103345,0.5411707650133398,0.4192622733457534,0.28474641349251784,
 0.44572815434137425,0.22257056624095717,0.6539820542328576,
 0.4948874982990584,0.23651764306305356,0.021631128344603234,
 0.21272511933951277,0.4802985545328624,0.5434834601010062,0.6736996071886124,
 0.13591308818563075,0.22912073479822015,0.34155003830049036,
 0.24204688987871426,0.8404603794214321,0.13596571644811173,
 0.6496346698614041,0.7707629698363565,0.761721163755918,0.43988132892528153,
 0.33794265515183963,0.3657006926757158,0.7248377804185859,0.5360367615140385,
 0.5831338393031388,0.9961786204215469,0.1850765340008359,0.3934910377091635,
 0.09069678050558838,0.6510487748389794,0.5579379437508613,0.4650471835451514,
 0.362228804460871,0.34025463389712174,0.029966121575839333,
 0.07274832794947361,0.7622937605036454,0.09888970613653136,
 0.036568481791710905,0.014266938596247614,0.3867180352802331,
 0.8633561287209726,0.4269138255173689,0.664553298461924,0.1039914745120678,
 0.16743592099257054,0.49222247882454184,0.2843858150126236,
 0.5321515542199373,0.831846448613496,0.11712992010055345,0.7721180536418761,
 0.9875636608089413,0.21099565892877148,0.38075973400300855,
 0.09318185146786284,0.5603872003264776,0.4770084089951715,0.6133601987973092,
 0.1068303625828404,0.3817468683852807,0.6787465106535675,0.373697687702802,
 0.62786004310962,0.044847351282656045,0.2835193072541302,0.8701186444565026,
 0.22038330799827177,0.056719210434483025,0.715445789874391,
 0.5433751323213302,0.8567355109816841,0.7867904564803653,0.916612227516916,
 0.7979790195908203,0.12469905687601845,0.0995135668365168,
 0.07977431430457849,0.5953511688633939,0.32734302609363597,
 0.5662380514330496,0.36143921381857536,0.18632288414968212,0.765441689553541,
 0.7272949893189343,0.42699518308227313,0.2729704486222997,0.7920798595713501,
 0.9539670518798654,0.19209179721240577,0.8552757493887547,
 0.26553955644196015,0.3906793816635633,0.6793683021224184,0.4350131100911825,
 0.42091115984072325,0.3104174151402955,0.7431330720622902,
 0.21608399009002033,0.2451393960725975,0.3577226276699603,
 0.41680851433917354,0.9388868878894676,0.558742642800398,0.27900286045431355,
 0.8878095502474335,0.3383777597204982,0.47597545657780116,
 0.08337083085919117,0.775196666924422,0.287578862053369,0.07257477965751158,
 0.1627372625833876,0.884440205654216,0.4625087147528214,0.7299450171686906,
 0.2786935901894029,0.9976789228593066,0.6783088376162475,0.4596470615963969,
 0.05853150527871642,0.1321922557736971,0.6061954787307361,0.731843851692093,
 0.0034362432529277953,0.4146711628560432,0.40627136327043434,
 0.3003913268574534,0.8815033295258556,0.41178245492836596,0.8363763245569805,
 0.7310229542278919,0.714478624823208,0.6541065386264383,0.9959881205658392,
 0.3745626041404416,0.16213091293107007,0.7869212327400399,
 0.46417690768663056,0.21402095551566158,0.9929178952010688,
 0.9167921623032578,0.18476553202850565,0.8821535520333332,0.6011948805736704,
 0.39419031795312875,0.1424230883893317,0.39975460283149666,
 0.5937263745786849,0.06726011513490482,0.9197424151154467,0.5793158248957336,
 0.08180183545473607,0.566356114141464,0.09128753032569215,0.4212473110274836,
 0.6870815398495427,0.5822345858474377,0.8721900310499299,0.4699369015175565,
 0.9069617968366706,0.5448079712708885,0.24720994337929159,
 0.030659510394437373,0.1061812162568585,0.46856798014501166,
 0.5460899100108318,0.27702180637198714,0.18520768636217633,0.402525151854747,
 0.07821414038020935,0.11337253980976236,0.6477440644131872,
 0.3903469634714677,0.4011803465357078,0.11714525518407082,0.8043748171244799,
 0.7509699746432645,0.6886374838612848,0.27389667175571164,
 0.44602906139822784,0.17925893381876468,0.49827981319122927,
 0.8232373966931138,0.9084866290474491,0.4914721853719701,0.639913749716397,
 0.9572918249477431,0.47386623173658715,0.7376113615102751,
 0.07922212758281255,0.14317586261609017,0.10712019157063435,
 0.7524146453798232,0.16528243465564973,0.2730112033510296,
 0.40036514354250985,0.5486917367035317,0.08702527889149969,
 0.10264200677146651,0.9669385612824306,0.8009698098199345,0.5854821947407818,
 0.9852520766607555,0.40429759403639687,0.5073450176403972,0.5218282276241151,
 0.03839608110467052,0.5808949279035783,0.049879927291542114,
 0.023486510588078602,0.09834321490897191,0.6766406091484918,
 0.12627536426311747,0.8235353379457206,0.24469415690094076,
 0.11640317846170922,0.7565576422339119,0.04233316901692352,
 0.20157588607534815,0.7613658447768626,0.7725429927429202,
 0.24656913976439176,0.5735074077327111,0.5196290186081712,0.9478895361558176,
 0.26061733441276436,0.8261910775913907,0.08895613891671683,
 0.9605828520116998,0.6129393674311505,0.7610521765604793,0.5799752977970276,
 0.546367176247064,0.8456969135958826,0.4020443129335093,0.034525122545637665,
 0.981804023979159,0.40880025588624,0.12786198088277556,0.5831604156176448,
 0.8669916772447471,0.9638803356771103,0.11415349573381606,0.8195166075267217,
 0.40342011808524214,0.9991345461782852,0.11974211022493231,
 0.5240565975660068,0.21651449111165832,0.2103682810582429,0.748355741581307,
 0.06042894707215707,0.4540546805326757,0.09755570305845995,
 0.7089455902553281,0.1783092927015082,0.0943562555612456,0.32654506393850835,
 0.1871306201923586,0.905559797407109,0.4368409099281355,0.9707835457688238,
 0.5405691744527227,0.8043299669280949,0.44350991283411023,0.8261946980521256,
 0.7653204450259494,0.6004280193967413,0.2228367604007193,0.4989430730705213,
 0.18666587387394984,0.10481892807008619,0.5744260190504015,
 0.030175682805666848,0.5253963626147815,0.9375348271996191,
 0.5808863345935238,0.30915381181785284,0.18736016264276312,
 0.9611550862781435,0.14060895353935665,0.9750449569468463,0.8097660599568808,
 0.23485441891999348,0.8821613184680774,0.8172206664480832,0.6069311522550801,
 0.855687315948666,0.9325517943860251,0.6238077741414577,0.8679740063919203,
 0.26309260052619554,0.5274423123183402,0.1013430106777502,
 0.15071771049572258,0.900731449054973,0.14975362034023154,0.3193977922679415,
 0.5314747783800959,0.4793717691397452,0.5970935771306679,0.7992793721479248,
 0.5769881957355083,0.5893485275885728,0.20730468270426639,0.3098089129482794,
 0.48712116627486957,0.18632622384477715,0.38286684471285537,
 0.880720080235418,0.2282857845976619,0.500228115565313,0.28933227017405516,
 0.12399722292418747,0.882790769710666,0.2181485850032362,0.10703626043173153,
 0.23012844825370316,0.09189770973474798,0.8531053865808266,
 0.7390576904116586,0.7086670218730438,0.12456623905755093,0.8546478880304036,
 0.2237127415933845,0.2694837245930204,0.6719444195885419,0.9320037898564003,
 0.9322255983183756,0.1059587051408174,0.6446059007118554,0.6617310138861505,
 0.029218995104147494,0.9103826071266332,0.8499925837333673,
 0.5245266983263317,0.4814129333631979,0.43021954715556343,
 0.29769438976388907,0.6034937876533284,0.698867205311297,0.7442524699367057,
 0.9832714858584434,0.814076693450154,0.8255842152238106,0.5690030561318251,
 0.9626546366408072,0.6970200321174117,0.8849911695607592,0.7208226730502816,
 0.22205527654713153,0.2365559954924945,0.7306487667114936,0.348489848203845,
 0.41818188401874157,0.4715046186556021,0.42890208743688807,
 0.30358889463370153,0.7991839567208496,0.15550297227285215,
 0.10584049085408065,0.9981883885724933,0.10785594504189822,
 0.8333640484226629,0.7113739974593359,0.40080506715535424,
 0.35138974574495285,0.4386819764422578,0.44519934013743956,
 0.03929473863255928,0.17275188811562048,0.4049922148496855,
 0.3464542805724594,0.2251905580734993,0.8578061835112007,0.2174319836045513,
 0.03439992724481744,0.8911361478356163,0.5117673524305822,
 0.24897131311865872,0.7849731834924166,0.5747744312081872,0.9393833986306218,
 0.5789569022190314,0.5703035716700489,0.44398888967452643,0.8819758348315372,
 0.32754502786986184,0.8119102919133339,0.12021406549539271,
 0.3777710786743951,0.1046556589531612,0.7657456166998194,0.21774275867212567,
 0.3554669253928233,0.5216045675315643,0.8430790072185872,0.03882256615176838,
 0.8827619394596282,0.22876875392117402,0.49391322677048377,
 0.44262740994250005,0.5691949860240121,0.13832247074986814,
 0.7376791930743428,0.4940139597945662,0.9002360398602431,0.30555062690247237,
 0.5388908198898372,0.056692375943563134,0.7518204177178629,
 0.3860405075359591,0.47222383807271817,0.7721187798640214,
 0.37061287136035226,0.38574897823700194,0.447115411333483,0.9271843773962274,
 0.09240275815887722,0.04891810060031776,0.21318462173207497,
 0.5085995173927417,0.966136148560451,0.47197635109242697,0.23355686509739837,
 0.8199303880531716,0.4316723722713315,0.11239468142963593,0.6839539606288993,
 0.16705002532641622,0.8393324542486525,0.470075167798497,0.09517387656672627,
 0.418796581763176,0.16124617710695066,0.9386200452857332,0.9527478916442913,
 0.42384131337597075,0.80657004842274,0.5384157642996911,0.87397885913712,
 0.6732955655178288,0.05235880190845765,0.5844155679357158,0.7316256956852781,
 0.4289834022110044,0.039716392450713345,0.7876041217493657,
 0.0992190682035029,0.8918049015422672,0.05089389752216944,0.5626900307559733,
 0.415970733421027,0.1836379900285816,0.5785283066803264,0.03702884691904618,
 0.08681645869842602,0.9277415852821321,0.6476580988375238,0.3049784400144935,
 0.340250812461371,0.7971023512548014,0.5880440494927377,0.7000172925821109,
 0.12378525112083905,0.7428042866744917,0.5411332635296174,0.9652485643602966,
 0.21580230885890594,0.8028499392216704,0.039637397509265315,
 0.1499051661789319,0.9039678020661659,0.8684138962052033,0.32630200193006653,
 0.06619914734038002,0.7984974478173644,0.9894051856860494,
 0.06694284958607732,0.4912748942370755,0.3106562750942403,0.989386125429458,
 0.3104040140694263,0.064258996741966,0.2681157708831178,0.30123702102059136,
 0.28062063463929166,0.31415890071333186,0.9034155424018598,
 0.12395548304507709,0.584132911605285,0.2528361326964036,0.5332418491498659,
 0.028873059804972123,0.828399835151048,0.9254008762937749,0.411687622318479,
 0.2919454840644613,0.3236829786665195,0.07185658330295397,
 0.39811639489143336,0.45557527561395894,0.0780557857899743,
 0.6868274738125418,0.18246088604640498,0.49617676502303365,
 0.9365953225857041,0.15774470429537413,0.7203287421868948,0.2921665825395321,
 0.3819528611788052,0.6471755663959633,0.9225747259207085,0.8361296490943332,
 0.314780827207797,0.52335397305163,0.8242591269796271,0.14987351089756307,
 0.41794370335425235,0.8920492863400753,0.6080739789087066,
 0.011179576119742074,0.9812491913258244,0.6461228134729661,
 0.005473704472087504,0.9736536060570607,0.8400391963519115,
 0.8815907862950022,0.5401293574263821,0.3319482549293232,0.8320144249244057,
 0.6597967064496686,0.08141803569436645,0.6976762104883455,0.529451717906179,
 0.9629970348922269,0.09470569336041701,0.9625852509618973,0.5804421240535331,
 0.8858945426194796,0.900325121548163,0.5402196472653514,0.7383061370454672,
 0.13441441811914356,0.16221566280441635,0.6729031017679423,
 0.24889163060532749,0.8931495764980057,0.9253940428417838,0.1804368348077826,
 0.968200573196338,0.36460950443686324,0.2852593707169412,0.9519652281673505,
 0.5957674337788287,0.5525456418023278,0.6824046617565052,0.27811208089576356,
 0.6410399960641416,0.4547741782217348,0.8219034374284837,0.4106403694950578,
 0.8836918474571702,0.14044330616350575,0.21552074453863457,
 0.2839355890564842,0.5394666962944549,0.7894957398416625,0.6540896815934852,
 0.8848834802828404,0.8501565135914034,0.8251577277848905,0.7654527547279605,
 0.06752328167026866,0.00954436797704905,0.42106614631560935,
 0.3375664895672955,0.020087863072022927,0.6328590333598654,
 0.9900054675229526,0.3057925850282731,0.8827455569198284,0.3313011891735329,
 0.2674266267951608,0.36933750127437404,0.2891063363664066,0.334112493722159,
 0.655563856528038,0.025187280380144283,0.15168142311861776,
 0.08854490961683215,0.5237790124923961,0.34853185315857804,
 0.014142318187252978,0.5752401602721938,0.8470402742452023,
 0.08592707657825183,0.8137085769752137,0.8105625123398091,0.2470932706422284,
 0.5990741459906601,0.2726095692785174,0.6805083786418982,0.27059175505934363,
 0.48049414550736513,0.2607553028412637,0.9559684811831597,0.5830622226413423,
 0.14463619938869232,0.22850362980375172,0.7985895430039761,
 0.31807049638197815,0.14957829416253898,0.09015224648502518,
 0.3410812399650949,0.5100687370857464,0.37723260218389476,
 0.11594318296806427,0.20604447258215675,0.36971672403535427,
 0.06936345248961906,0.4067098353475209,0.30549294318754583,
 0.9602117026447232,0.9281149360269946,0.9121600801252286,0.2665972302861167,
 0.049984690417195554,0.4649142938034898,0.822799548571283,
 0.10426996923057141,0.4312605780640135,0.13539280538075482,
 0.4187215945037931,0.7074885685445328,0.8617422324893469,0.40230834964159023,
 0.3838359012823602,0.8464862688788544,0.729396281610997,0.09919312531269964,
 0.8435122320898192,0.5311874571159696,0.0054519037493740985,
 0.31828922841738416,0.25290595363950585,0.6217981628662158,
 0.8791997637338144,0.08467557964985017,0.9925535682539852,
 0.01991672815992518,0.014916769528701468,0.8855586161256981,
 0.671294560825892,0.7766794143385595,0.06557024004455092,0.8061951798279156,
 0.8042364908195484,0.03761069219127233,0.876996462331739,0.2630499171702677,
 0.29847535952886883,0.995532126135985,0.9238408504975895,0.04744805345424552,
 0.40021026832013096,0.7178479995944091,0.39231531846946965,
 0.8211667474382782,0.5839481778398155,0.010633407657032579,
 0.2494576530896193,0.16367474830834472,0.8609996923788408,
 0.23208805989039027,0.3129536140550284,0.5230098822126465,0.9054082210477397,
 0.6868805596033728,0.9822345481668902,0.7790422897311434,0.44346094401619607,
 0.015383150467300766,0.9022910560095991,0.9935887742666349,
 0.39109029232598824,0.5080394406374749,0.5279505268187675,0.4453562926392357,
 0.7800245280301097,0.6661593926559256,0.3422137816760269,0.9748663486271241,
 0.10039089367039078,0.3621694092075547,0.7916576974525359,0.8571704880988377,
 0.5273549864961726,0.38288430897856174,0.8738442836682487,0.3729821273782269,
 0.9307293166814921,0.59259960984019,0.8022011529924767,0.3811244103647564,
 0.5580881406361444,0.4166495864973091,0.5273771726004339,0.577132314331168,
 0.950360375154499,0.18275448498599134,0.8689024570347521,0.07150327290403491,
 0.8605043486383879,0.11922131925986468,0.6150902588585989,
 0.07613044806739477,0.9437085438597035,0.7538914801918433,
 0.017047980626019643,0.7542377554877292,0.329766641448603,0.83160128731093,
 0.1362680440199281,0.7289833096468948,0.15749382508924226,0.9660044534821504,
 0.14335232684282384,0.03952044107861674,0.3221542220625716,
 0.07959254524403903,0.10644784368935822,0.10600286341082876,
 0.323094531044243,0.17539649251965617,0.7239122511674001,0.02175030816781931,
 0.22762163826806625,0.08567858180787724,0.9755680770967816,
 0.35731800412059245,0.1965347811600745,0.7444441226929477,
 0.17893181292884364,0.6794281294445498,0.6430446935142022,0.5685503045647927,
 0.24677493742215106,0.9351653906729717,0.9547979491895945,
 0.28216866635908155,0.4328987831825295,0.9467591755548346,
 0.19095483509895805,0.958562382765063,0.8679606199294495,0.12680456105975535,
 0.21489630684127714,0.37128379914392795,0.3661347473796702,
 0.041190387642549786,0.9916435360262846,0.6910080640635246,
 0.37349441352662294,0.7623626254650301,0.4014181578902484,0.9786404237004888,
 0.9553428226202981,0.9019685472539586,0.7337852586262708,0.7917439312681444,
 0.32622937673114927,0.7536084140145274,0.8013399700364519,
 0.20731046732277036,0.5573516715614951,0.9657563508036529,
 0.08922868782306947,0.2686197215389827,0.01101903630836587,0.923659578110687,
 0.8328732564141634,0.8838077914269374,0.9866455306163577,0.6028741141172629,
 0.17111244173469708,0.447218970988127,0.16235698218784456,0.9530853090612466,
 0.26430559793056374,0.18315065103532957,0.5269325044821914,
 0.40852222483678213,0.5586211305035136,0.25127838788707635,
 0.24115157410394328,0.5857137620674661,0.9613049979873414,0.9067333399984837,
 0.7189287745010545,0.5422377084149961,0.5910798267784478,0.1573135276925791,
 0.3506769936532106,0.3097165926702904,0.8638896586552465,0.4849733288132688,
 0.022801684321779758,0.2822604541733922,0.615730231110836,0.8062453349046166,
 0.3127762150187565,0.4699706793006926,0.5964436798447935,0.0700899665085345,
 0.026695708144135577,0.6815148925927119,0.3570355732044842,
 0.9994979231881856,0.45517043293931236,0.8147790605878941,0.4688519113638916,
 0.33578898390555656,0.025250897565277763,0.2704452063325261,
 0.6238805508599852,0.2696734923198293,0.4219973004680585,0.7476488727480123,
 0.20959138256113596,0.48154887622536113,0.7951494456496909,
 0.0030904632293686785,0.8003376889628014,0.4564913347858519,
 0.7085340744878718,0.022380831580697036,0.23565226247685211,
 0.8101457614601983,0.811807809508103,0.7564617479230714,0.11755243560336037,
 0.4535991142685839,0.8286207720549161,0.18324716478174874,0.6724014553487453,
 0.8041945462457716,0.7770169011211123,0.8833767220935017,0.8768048371133162,
 0.5452957849872896,0.41477697171615324,0.20232919664409277,
 0.36775781109443273,0.6116430300898341,0.309435530335608,0.8241176813063144,
 0.5221627616137058,0.3555744346835674,0.4460628269453838,
 0.022991359844794057,0.13472708015920543,0.4147789667806545,
 0.47017414010957737,0.34304934988212177,0.9067891973057619,
 0.4207184559842728,0.3271712648778322,0.07180306580483553,0.5368277361939744,
 0.48946208846454997,0.6793214522872484,0.5132963163890218,0.5134418709301645,
 0.0589186684877423,0.4908839080326056,0.4275152522971658,
 0.007498665692317608,0.8715046641703043,0.6842098681490365,
 0.7098432411904405,0.7822319947958888,0.9962943224109158,0.20141210483148775,
 0.9223052872073276,0.5854463705154904,0.2640703475132935,0.35592489565437424,
 0.46977205896102125,0.39248971042379754,0.7868202356951368,
 0.20365765949243186,0.023987482940519,0.25450907982797266,0.6281544913617622,
 0.18377319696119232,0.45367778264224323,0.7164050572104297,
 0.4390812863771547,0.3564300008432829,0.7118242444728399,0.7583890762895517,
 0.015532736111340095,0.09892209584724532,0.7057867299186202,
 0.3450010400448853,0.7814813096398131,0.23696323901821736,0.9880429833250859,
 0.5878631149514869,0.8038658622301386,0.3857151486936937,0.6720355351615731,
 0.8233048178337566,0.02756017108580444,0.13659032903510548,
 0.6648013489773066,0.6526045732640474,0.3035364206349217,0.2998292001267989,
 0.961269155343721,0.29182140462085526,0.0562601136366816,0.420900007802142,
 0.4237374802072905,0.9515584957930359,0.17599586905323472,0.2627375393443888,
 0.6589038403429439,0.8432782171907156,0.6566068682136653,0.7877128150077976,
 0.4576904620426163,0.3572872091674181,0.9258176269408646,0.4644402660285484,
 0.24399937424647788,0.6301689547579715,0.7434567999007601,0.4299234383782211,
 0.5269137897744016,0.25357521490178947,0.9512882946121354,0.7346421128814273,
 0.45631874280433116,0.11176495456652069,0.10777445665067742,
 0.12200504375376264,0.8037549263751025,0.5735360756743156,
 0.044981530061007136,0.8662359976649435,0.46332705622981163,
 0.9057345869669811,0.3035196096084226,0.056417487205621786,
 0.7672968188003738,0.6531856817204843,0.5622241967783399,0.38235373705416675,
 0.24553199081498644,0.30189552902170635,0.05580101954881822,
 0.37260845732948067,0.6481185178468761,0.9405474386646273,0.6829963434372458,
 0.11484753021469807,0.2624426243614757,0.4375613323921709,
 0.12497733780323439,0.8569341940977044,0.44811618776472206,
 0.8369922145543734,0.007409667594676761,0.8286426563465216,
 0.7141355803551057,0.9156491931963194,0.39696337697638484,0.9488078094485948,
 0.29843363513021404,0.6030015482196175,0.04709790128075142,
 0.5127981719515365,0.484175195981983,0.879499886614069,0.4076636862210248,
 0.9142101069291688,0.23692596185194503,0.06155914837381138,
 0.7439144742395518,0.34040492548924306,0.7522811470937054,
 0.03345643771358853,0.8829141579179767,0.7565710048243423,0.4860784232279869,
 0.11350277806405629,0.7794715443795106,0.3207407376506949,0.4971023698676069,
 0.6063872065055343,0.655687191317775,0.6142199715448717,0.7926665024758129,
 0.5658407147885494,0.5246437219550403,0.8418639633114386,0.4527194660130499,
 0.35113960061417426,0.2588203813354196,0.5537096803357758,0.7618840232292587,
 0.6172376892702007,0.7639621698038468,0.24171751715915202,0.3549346423161691,
 0.24454452743557642,0.23936279268327787,0.4338257743382058,0.44354120041956,
 0.9063188762426868,0.6892211480990885,0.8992909022027338,0.4032495926863884,
 0.2297675314207891,0.16137158748682667,0.42267198731715017,
 0.8861491967047889,0.516601545682866,0.24943843568643675,0.0994606879972576,
 0.5084831607860274,0.6438745633279823,0.6891369351932155,0.2547001525340338,
 0.35522986749347507,0.7994069924791294,0.6548297721925074,0.6469107622740004,
 0.6093543679787056,0.7687048691481159,0.40152328775207924,0.4878842605831364,
 0.3517984255330503,0.11257050810529556,0.9326258401277697,0.8049155457014863,
 0.4400608055493316,0.2962796019928726,0.9533450155209237,0.9397402494819688,
 0.5453353667154408,0.859731788374075,0.9691856020873173,0.7400220473836427,
 0.16822520100590177,0.7791797625434758,0.15626540786507603,
 0.5610180524474842,0.39923665400441566,0.9848334573469895,0.654949370789792,
 0.8303192067961497,0.17704769676374066,0.927373113476932,0.9930417648709092,
 0.4128039222781871,0.49864628738543615,0.133204545950098,0.11224545076449166,
 0.6614747536624832,0.8158501432428353,0.6776320132459546,0.31664766093505425,
 0.8369655229503484,0.2671712021783774,0.5678781592595183,0.582720225233503,
 0.1636357540721729,0.6387109172512787,0.6113360712992995,0.7145365258916424,
 0.5698884011218427,0.18948340387687512,0.26498429126214285,
 0.7342060260642562,0.0373078910099518,0.7906938845804815,0.1030866320672521,
 0.34099638064027615,0.35658622031714315,0.9386020047560534,
 0.8955640015207758,0.7462450481731419,0.9196027125816212,0.448288855668616,
 0.5875165916426012,0.22195835649947449,0.8899427403086113,0.5858873612535391,
 0.012034388607276791,0.28479266076499554,0.6544787517323305,
 0.7302368641270308,0.5285528796199849,0.1750410248885842,0.9782890458663093,
 0.7972333940807783,0.9237755056955356,0.822417334309073,0.14542668619474786,
 0.033039056544575396,0.5788829239270934,0.2611368015869081,
 0.16315662215315951,0.5068277992635,0.08170594924905028,0.14077847715133407,
 0.3385717912517817,0.20152960572631862,0.5032953621106441,0.7396102461392683,
 0.9812051130510479,0.7318260423382067,0.5706794402052094,0.11153579893468701,
 0.8296802420060578,0.7877449189603644,0.6975678901646821,0.691526163649301,
 0.9748312286082692,0.24763233865792844,0.24106663906976733,
 0.16829042430397845,0.9876725298605177,0.3912581348880432,
 0.12618829988734548,0.13189970735346745,0.8321815591452437,
 0.24803806665543893,0.742332394238831,0.18288085834091472,0.5148707179773339,
 0.8767138945429941,0.018507164676771493,0.03750619451053594,
 0.06114826504788362,0.4609771360352275,0.6001198964338238,0.5181364223026877,
 0.8042866189392619,0.36757744072118326,0.5229479229889599,
 0.48573902203104535,0.6706689077702163,0.7974569772703867,0.9004635682306493,
 0.8653264893292862,0.5695392086372126,0.10423523460644679,0.7546800634937242,
 0.5242942839973506,0.5758228439035764,0.1800788436612133,
 0.061484511910794626,0.3438048307796949,0.9292691794023002,
 0.3804035497732593,0.5007857398642479,0.0583441259100721,0.9755320666002325,
 0.4699908702287261,0.7277838664003933,0.5464826201524929,0.9426318202443008,
 0.8812998079293631,0.435548065410438,0.8513737250075559,0.1991555350020151,
 0.2995996193557785,0.6754644970770771,0.22099026210482542,0.8734707257416447,
 0.47092311710294965,0.1805920653616453,0.7786722404807263,0.916706484798791,
 0.6094566720819511,0.7720951098973805,0.5102297399840756,0.7568752641920564,
 0.5618280309246747,0.042893298218546816,0.5117982112182112,
 0.5776638729711746,0.10936495039946259,0.14545912008440187,
 0.6714671712597668,0.6532401339555436,0.12145552901919143,
 0.008662713096163399,0.6415345516825903,0.7356647290309805,
 0.10848625725720962,0.7315656114729053,0.6324761004976818,0.8922320833648139,
 0.662434238585957,0.24102062742389885,0.47787838857964005,0.9812962338786236,
 4.1569955924880553e-4,0.18093640805129863,0.4626707895618104,
 0.28263575465408997,0.33356244840675675,0.7670821319036131,
 0.6747519199773077,0.009311476726664769,0.6199984849183378,
 0.8359245462307988,0.3731640119482478,0.5595484237152497,0.06088432241194719,
 0.35094633546600806,0.21712921704916122,0.12405144651881816,
 0.5796889299522725,0.87224356064261,0.4642018118137865,0.7415618393360238,
 0.42393149953461606,0.31182889374304623,0.7463714706558514,
 0.13548697706002055,0.8559405232285409,0.5512134587249916,
 0.40547268731424313,0.22130948018260022,0.0723653808671475,
 0.8689709851948139,0.36404204186182954,0.3164984470971869,0.3176350620634195,
 0.9112185741842308,0.09554876720175287,0.6958307224405396,0.713440790656408,
 0.983522325777562,0.9709038964244655,0.1744760229566007,0.31135567734916625,
 0.43679092638748296,0.4434708907082259,0.976384710315011,0.6012196096778272,
 0.3145218509565374,0.96519345283826,0.40599100655400544,0.6629227209304688,
 0.13435896450317963,0.5777150506729175,0.8971713404253865,0.9045759475579482,
 0.8360288039818544,0.8035861386036595,0.868335600400584,0.35950868669949787,
 0.6478665224441783,0.41478546117081516,0.9877211035982779,0.7570884241401442,
 0.19627073748424972,0.8514379316438118,0.5836686231643763,0.3238425706458026,
 0.4255534054038519,0.4258412432913128,0.9449657470898692,0.9892352689290284,
 0.5405632949221114,0.358384365023123,0.9048894906615338,0.5747513892473203,
 0.28695180080839255,0.2482298020624617,0.8628771955736086,
 0.08949327070899626,0.13274126745836856,0.42410169137338793,
 0.6774259982436734,0.5561937205456265,0.5647223638271318,0.9433235634651373,
 0.12142309341770696,0.5600366199352954,0.9039779403731851,
 0.037608281628122464,0.35678238182097743,0.6798656786188204,
 0.40478251269162446,0.39735030566436647,0.4864241340561737,
 0.5364812481892669,0.8632629171555178,0.8355438300355101,0.3646845896580233,
 0.1983357703847718,0.42703205477236295,0.388407009183631,0.7854672737613903,
 0.3515874831405489,0.9736972437593606,0.3484600139854793,0.11808706804075286,
 0.7150618518518193,0.5420971480482756,0.7479707008070776,0.5438366951543971,
 0.7683410582255219,0.4368532234355611,0.3951131836554471,0.1403159497782459,
 0.5193927575914841,0.6530186883791514,0.04212190857597253,0.735177187535909,
 0.17250462431271452,0.7396014513133082,0.591910236049872,0.653073345861831,
 0.22522101009904105,0.09666953509676679,0.3023399571883475,
 0.6115567791015774,0.018270624912952838,0.985792168333792,
 0.05485147099046728,0.8234271425584402,0.14940097080570514,
 0.5806197957204516,0.19962130292961933,0.8010240465293423,
 0.06325620448029279,0.2333210425948854,0.6488586669472877,0.593741192080099,
 0.6230358260608417,0.3213576522097288,0.7881960381255912,0.6383169181443519,
 0.8009521397837011,0.684097782148153,0.3408117323982074,0.16505174566430703,
 0.428847136501766,0.7504516839608601,0.9519144115400704,0.7022451931567046,
 0.29705526893783474,0.007947413729258224,0.103572439887746,
 0.03427922380956594,0.23341814750224477,0.781451316124929,
 0.19742234908059364,0.13012287546361034,0.04938677039874528,
 0.01362016624915341,0.4340843101074501,0.5886470283630328,0.4226400031116331,
 0.9322426666662185,0.8416517713541142,0.3197890613419655,0.3963421276432857,
 0.04435727734064443,0.46830440307025123,0.17241479272715443,
 0.9791403655589944,0.21247358998954258,0.33396330341923974,
 0.9812368328358415,0.7229713748834463,0.063238914025048,0.03613678452909763,
 0.7973249811130378,0.21222303969851597,0.17826501499628156,
 0.8749783023509171,0.26062866001804375,0.29022897564126104,
 0.5576556552663122,0.01700616539169686,0.6940627000139781,0.1693944816670696,
 0.9470833452111873,0.8239082663838874,0.5996411177958407,0.07318707891131515,
 0.5087918289005775,0.8345103665730988,0.28996268022750493,0.8613040773440066,
 0.9167679225727016,0.41813493295208537,0.19530485541981046,
 0.41851487626506456,0.9280414845892525,0.446619674271338,0.9828564134496063,
 0.9512100986813149,0.8129832681561979,0.8196055258143067,0.5472641932560633,
 0.7119321329337709,0.7748699610170185,0.09095645767085847,0.8196013346797365,
 0.21491071598661982,0.6432916395514259,0.031005436350254545,
 0.5916536570578509,0.09385665435355461,0.410212537184516,0.0781978762894755,
 0.9700214926216729,0.3243785737282038,0.6556276402846958,0.19720349081143773,
 0.17763385578359725,0.6006699549853656,0.7140880067224925,0.5301818553486,
 0.5185338934907306,0.5250764302060302,0.894466921543542,0.2722441253888338,
 0.692474866966128,0.6075594231113666,0.19833240695416499,0.8605102011009376,
 0.26042363810070634,0.614870094237499,0.8375770594576127,0.9671920129042376,
 0.32688196425025373,0.37864497319890256,0.673158516464064,0.1333232769539604,
 0.8253704000935087,0.33424458219167597,0.9713842325583507,0.4336891410957261,
 0.8575364954222462,0.3439388694320721,0.4467189733417528,0.8729350988568328,
 0.9105877348899665,0.12411915077123725,0.22796620308525617,
 0.8440969335481711,0.7935137820034126,0.5960453601617246,0.7280893233028631,
 0.2925145961288832,0.8645244251531069,0.08422306138454971,
 0.10017566331436578,0.8302928088779424,0.4099871239172188,0.9030518813905948,
 0.9151663945153001,0.783921003957238,0.6538428243792734,0.7850780989225534,
 0.35438528887228593,0.5398220680734389,0.5239301562445036,0.765601836235929,
 0.17460567602695776,0.3820654956767572,0.947161411700549,0.07579685841964323,
 0.8993924575726736,0.3751158471298548,0.8770105649715522,0.43164790529160413,
 0.039319948281511974,0.7904963694724052,0.35779569962386637,
 0.747680543818973,0.28000094849374135,0.9906300722569298,0.44466513362412374,
 0.1817290523049122,0.4993184787504509,0.5417454570628568,0.5105351692111443,
 0.5925650606963558,0.07943023471907162,0.5319487330309466,0.999332240937429,
 0.10522800110081831,0.6416055722893915,0.14774041727381437,
 0.8740823940086295,0.6886135905957586,0.8922435842493842,0.6229213350860197,
 0.026742171604930642,0.3675620345281615,0.8276768543365368,
 0.5820065055321839,0.465788860842658,0.407865890554883,0.14877035520834303,
 0.28125476407040484,0.2706457258953462,0.4069915641486578,
 0.16964515685426185,0.43656380630925473,0.6119998283199316,
 0.7423732253106956,0.10451451953516766,0.7957154399371874,0.3689587803284292,
 0.5702189908588083,0.05614808810979954,0.4482159105070086,0.865636076202653,
 0.3115912117307946,0.8526324413925066,0.11386822377980055,
 0.23972337993435355,0.612624831874091,0.631459191528129,0.7104791821315924,
 0.8598192091088698,0.46780206181267636,0.9420972368641012,0.8873185200298148,
 0.45716041872400526,0.6092887698861158,0.46682381407172213,
 0.9684441460141422,0.3020365271896934,0.022684074383277242,0.171635251239609,
 0.7388917206462057,0.33476849644320894,0.7516255829857003,0.904722582627769,
 0.3220915657953609,0.4046069005657864,0.05385300067839227,0.7920519785820475,
 0.6951039678314729,0.5451897298712479,0.4148396572104882,0.308160715880349,
 0.0684119898816028,0.5292924419849437,0.662400294872402,0.5125046879749375,
 0.47300644068994013,0.5091358451508331,0.423349534218614,0.06954205198237784,
 0.9164704036270462,0.6654701454175833,0.5327117869225526,0.5840330913464068,
 0.7513200528782268,0.5031802284112226,0.19158933430269787,0.1456320746116585,
 0.9496276036943888,0.560411832033308,0.6770463478829317,0.604429790522955,
 0.17348106125128004,0.28074648950122605,0.14032282627846493,
 0.06597579775820861,0.19821017009303188,0.19953845878998722,
 0.2591485741753009,0.459220627523665,0.7176743345672938,0.5055233900504936,
 0.9703469433916785,0.34182644446482213,0.9868734286690324,0.7768117436113335,
 0.769888493077106,0.6570961650119517,0.7605869872963055,0.26065084438849184,
 0.7667701118283063,0.06990740679688323,0.8693108210636148,
 0.45925311079853015,0.9024256666239641,0.5446007980431363,0.8949379573741116,
 0.5879725885962364,0.5139986191278068,0.16390016632732718,0.6599192989616058,
 0.6730429005954867,0.9518257820324488,0.7305147488048812,0.8598529896087678,
 0.1986076910437481,0.32492983018438504,0.09662910096674127,
 0.6541508066667857,0.025876287038429813,0.21003064849283415,
 0.45177538061582745,0.4196404735854322,0.4141859793261381,0.4091267734164882,
 0.5899317600592736,0.8243592950684617,0.05924548353986803,0.5889984436938777,
 0.7818064484581578,0.023753877891953934,0.0417110294695584,
 0.33623891452769594,0.5031277270270116,0.4729185841452975,0.8931401555862929,
 0.12407540462890909,0.4882737993345553,0.15165063886232855,
 0.8203115158804718,0.8333897090802161,0.9928788563446785,0.4083652855914667,
 0.557095533443936,0.7134008056133654,0.49381465997346696,0.5298882207734589,
 0.5301568862118248,0.9074257397066741,0.08528059426994417,
 0.058097086288195876,0.6513072564217701,0.790029178942609,0.7461338281993658,
 0.4962013232026018,0.6678889902095912,0.669945031620622,0.8045756242830603,
 0.13140324391949176,0.540410288737257,0.10424125123606875,0.7373078286983789,
 0.5770345063050893,0.6774251148218449,0.8035254485500222,0.4149036592411969,
 0.8879586789707083,0.3572176107862616,0.2985665692575513,0.49218240763284304,
 0.07231318224253658,0.7697174401283478,0.4383327207790181,0.8003942585620402,
 0.1960110955708506,0.33979051422146656,0.1749223809640702,0.5461078735181655,
 0.001197176630427843,0.014689988037564161,0.7099152555012844,
 0.9971949164963448,0.4280687181204774,0.9929184569060717,0.6508540391090751,
 0.7279183104866893,0.2079735252187015,0.481056725778393,0.29170485559171944,
 0.16857250778408517,0.8666019881515447,0.8486766883636943,
 0.05983488965337902,0.4629735733073922,0.7624342708401601,0.6528159787142668,
 0.47902652531521395,0.9379035574193566,0.028188073001823688,
 0.6593717345193466,0.4577600721610797,0.4801272757321947,0.574121225627426,
 0.4621287587362719,0.85173633833248,0.11942286849105122,0.7616727348138648,
 0.3404595280200413,0.08820475492541469,0.15136494219369157,
 0.15817404640888888,0.9784850156767951,0.741418213003487,0.9549654058917336,
 0.9181739591573524,0.6323987091426538,0.27315683144333636,0.8176156920879442,
 0.8639225143941301,0.7742460359736227,0.02438920170632619,0.8916064517941721,
 0.277584699521764,0.15508182570269868,0.7480376331989141,0.5211808211127018,
 0.4221771859955139,0.5649032730241106,0.21601329047386397,0.4671896533800146,
 0.6487030168687561,0.19360820538812695,0.2880770674975083,0.5253321690454046,
 0.7727259733079064,0.17213988313527429,0.11041273173753952,
 0.09150854525669772,0.8233872171006256,0.9631802109292837,0.2537569799863717,
 0.003087143331001019,0.08896793439621542,0.09852762068487797,
 0.07372298740685568,0.6387667256859093,0.9325343690666139,0.1917685848431906,
 0.7540112304313398,0.671320935183946,0.2994783957039795,0.02485469600487422,
 0.3261145660246071,0.887464162215083,0.4392483457078179,0.4204003808539646,
 0.7738788257378594,0.9025129168895505,0.15056370542748387,0.8564715470417934,
 0.8974996020214928,0.9795539217558037,0.05806752283843242,0.5399246987403767,
 0.6546837451448307,0.7978053153874284,0.6361191887144763,0.5973985134001645,
 0.567516775549079,0.5669506551688089,0.6952422196147189,0.23854957299523172,
 0.03544513882565958,0.5417806432003774,0.19748224878219012,
 0.38497115158887485,0.6852370785487507,0.2365000280440357,0.4621832618195296,
 0.19558587859487087,0.37468196842157364,0.3226130033102882,
 0.8492778850342879,0.3954641122728193,0.12334750943605954,0.7097087833210317,
 0.2923909095742644,0.19513374239156545,0.5915506282306069,0.6933395487098761,
 0.5751518111657941,0.4111978072894761,0.386817597479191,0.33886954087931487,
 0.7934122022926856,0.15363873596740396,0.6841653768340478,0.7568236305366876,
 0.26955433030301623,0.7474648557100483,0.065431049906159,0.2809550187835098,
 0.8637356831370501,0.7617188000871146,0.4501210560695921,0.6554846827258408,
 0.27463178189319903,0.5936448627276164,0.18018631566555998,
 0.9048592162016567,0.7591009057862839,0.22700759954282823,
 0.34024693069344203,0.4997166181591024,0.8417440852172909,
 0.006831359270850879,0.04295766442408655,0.1557350299563368,
 0.21082009163237336,0.32388855544906947,0.11442659244736142,
 0.21929593137447057,0.026051093121086777,0.2722643948028681,
 0.5825485238988236,0.6463854095444795,0.339741494975092,0.41909982008557134,
 0.02206982355728937,0.9786007296700773,0.05760964722836859,
 0.8673571921825934,0.6090908780484656,0.22278069418254176,
 0.09996950317895792,0.2171753092645078,0.411566550963948,0.9796400514749485,
 0.8683260369883654,0.8884805308258658,0.2019919684550261,0.3060235405418934,
 0.49024410587633205,0.5979724266087407,0.10169445946967959,
 0.10705032412854387,0.09520509967791857,0.36229107512180425,
 0.7590643182909718,0.550865106500962,0.4634041085123709,0.8932348821740738,
 0.4934425621413564,0.03637726513211603,0.05437377514615538,
 0.9227215613159784,0.03613761764282819,0.21327187969843808,
 0.3608109159799351,0.6722194003201393,0.2513876864193638,0.01268566045568198,
 0.5329203056647809,0.8062550644173436,0.4240619931676266,0.8872030087508308,
 0.11482643888762245,0.9387427386000766,0.47704951504229953,
 0.25025991184243623,0.44518198070305504,0.45898575824475574,
 0.018366644341511695,0.7755248805520129,0.4075212801269992,
 0.8216856324196009,0.4445908466496129,0.1502308298530014,0.8169055387738814,
 0.011633471004634077,0.57107718465979,0.9481264730570736,0.9191646076719698,
 0.17163381521653942,0.2012734045061454,0.8792455036907934,0.6797462385533765,
 0.7435100198511415,0.3384241649065757,0.47628195473094104,0.5838954833838242,
 0.7768062944036279,0.18539068139048165,0.2846449838895262,0.1466089638918059,
 0.7213467247765529,0.24865785434785836,0.11288674111537755,0.4731425052557,
 0.5530533315271347,0.3931278040451678,0.9863978709423806,0.23381635549980606,
 0.8529065505339842,0.6153028090307275,0.3679029501773048,0.9899765371340076,
 0.5237439993473061,0.73087244691413,0.4508496463426642,0.2604882093528609,
 0.2342466822159941,0.5977213128604499,0.5558804115101276,0.5692724941065422,
 0.245596842703373,0.6020962424969706,0.4792340116926772,0.5985975651319665,
 0.7119951462201433,0.5069948325551648,0.7146766672886417,0.8436661052985316,
 0.9203122705992536,0.9666547414191476,0.15649089628284962,0.0417182971694019,
 0.9194161162772576,0.678888003311675,0.3717287115167057,0.4593384426357008,
 0.043110495437708285,0.4022708882163546,0.4784791138572988,
 0.35581434347062324,0.006208047637111469,0.1877564222597775,
 0.46602506666091215,0.10918681598915436,0.7305637810789973,
 0.9456929794332888,0.8476016316328756,0.22263767746820484,
 0.20950182502656733,0.14500046632302022,0.7239304774200322,
 0.4777466298174792,0.47530503704144855,0.7576458693772107,0.5626540029401359,
 0.3353242011369785,0.5230079653807175,0.6444620432306546,0.31273642487686326,
 0.9823905486721701,0.29745128647351093,0.6291281697460005,0.6868170834444036,
 0.5703873595628999,0.13218363970013247,0.9741875697232206,0.9289667880707408,
 0.2258255385587622,0.7681381595173357,0.2858475804560068,0.3456256128407562,
 0.40288519316186633,0.42778729649206515,0.48954235642766775,
 0.5427784745948426,0.80756352702361,0.7832348035524376,0.6510697719116628,
 0.26368249065870286,0.9619680273742303,0.08953097546394717,
 0.8652086043624662,0.8683292193593138,0.26308177364496177,
 0.14973150364098053,0.9140622414170703,0.4900478399485142,0.5705321004622137,
 0.11984386131992208,0.06117240490788278,0.3144957396625817,
 0.1821019715007619,0.6083196004260896,0.9455667547700247,0.647023384370675,
 0.9124569983495538,0.6315002728371407,0.888359934819908,0.4504148312064542,
 0.9253139461655431,0.7873079637207023,0.6393972660235705,0.06831112479505219,
 0.0013373760628057063,0.43000121014118564,0.37482056105157546,
 0.676212203524059,0.09365119659341181,0.4397893137430602,0.4989052259762088,
 0.31718240424513733,0.9042299999860152,0.4176267208850897,
 0.21981561164676466,0.6808130162354423,0.7978637622795942,
 0.17643801568317408,0.3707169462329888,0.5029791785336644,
 0.28584553516287037,0.5647454489737549,0.9818664098336268,
 0.32363447970793957,0.2156811191561705,0.1427659907616231,0.4797793726036924,
 0.5368418512717434,0.8862256410950625,0.7657342291867872,
 0.004246393568454576,0.24454247724131195,0.34246544383295185,
 0.5055213045352228,1.252298150722364e-4,0.8380764204832325,
 0.37967654626977687,0.9182696994787573,0.6925582699546694,
 0.27635022260175257,0.5543563809085056,0.3541067551960364,0.7596004726525288,
 0.9408844657410453,0.6367186438877209,0.5945296198684874,0.7570328276230546,
 0.6607607899955602,0.8412483059344896,0.2064248502087207,0.5852740499923345,
 0.8615610842820874,0.6236303643594202,0.47768157512511356,0.1136463331536669,
 0.6508448825253896,0.07696161205735297,0.22705657874362872,
 0.5158522791775388,0.0958024655086609,0.7808239584144463,0.45476295589849647,
 0.28696432937719496,0.8202104630368882,0.23086510205734279,
 0.9682804758429666,0.6258284757358503,0.026738935527785057,
 0.09558092204612245,0.9506259924723736,0.13521894689841585,
 0.6180492632228477,0.35920317601087726,0.1472932947005906,0.2870565014706191,
 0.8073452863074837,0.5343579454035478,0.6359226073130385,0.7848993498734346,
 0.9473975470647841,0.5570295129381234,0.11137420147875376,0.690574440264802,
 0.713963813805611,0.7735056208279775,0.5714962657530896,0.02966443380991446,
 0.11641071099465006,0.6534536658762523,0.9228946608378328,
 0.06431587228218205,0.15236221778614256,0.2272118825089272,
 0.03147454098611202,0.16436054038094927,0.37028543092892696,
 0.7247032512500476,0.9208404568754502,0.08178312999384574,
 0.10977218674019285,0.6907620128502684,0.7444793897286179,0.9284745578931412,
 0.027523775507797943,0.6228764607944332,0.6769548296665416,
 0.9122019427915152,0.16747946512556033,0.9881104154331659,0.8780535525149581,
 0.6625724636389052,0.9449617074305701,0.7444493687177068,0.6733809054633489,
 0.052202846512946266,0.34431678265870447,0.30274158060148815,
 0.9639592459017501,0.8763348782409768,0.7244395196572753,0.34818443911571284,
 0.9197626678836558,0.23115881303488062,0.2716608102203958,
 0.20239437363695445,0.7894373643534826,0.133401623526056,0.6531041979151371,
 0.8304362469838901,0.39912004794963574,0.6132745735920552,
 0.11928381912878327,0.20834382973171905,0.7840694302618836,
 0.7409092911005185,0.9971207215290286,0.7504497599692483,0.24896721065702354,
 0.7775209653540113,0.1919190980269102,0.45481738662425175,
 0.006303278556841985,0.6564575412154057,0.19635325952975513,
 0.2730054067795906,0.4629516232101064,0.2013961934077162,0.11509883425158007,
 0.8104295014161657,0.2166426836722639,0.29448513898801687,0.4196944211200069,
 0.08265965909321993,0.7389267834120427,0.8002884343884787,
 0.30979276960637714,0.5749731430401472,0.6370936041992734,
 0.41975774804586274,0.0953387103228236,0.9439852025971742,
 0.45528301266190896,0.9424580562773489,0.9956709375359399,0.790582753922737,
 0.3336045408057242,0.4416293159733594,0.5653789202854299,0.3747927875984932,
 0.8017075398703966,0.4370501896548017,0.625444037912559,0.4357213645779383,
 0.667976394581182,0.2503120069401348,0.08020091584371247,0.8431940274782657,
 0.3844720921407818,0.019628031358167486,0.4503054358644971,
 0.14616598791869517,0.3530335535341388,0.2946431403254307,0.9230269339429125,
 0.7091139763898522,0.6798828415824866,0.02579358919221819,0.8158529161505419,
 0.6902087983350761,0.30864669133820644,0.10640614920976021,
 0.4091764370335734,0.006929407141905819,0.6190038433561057,
 0.6668733319468455,0.9744507265781734,0.3430723201667786,0.7612635719842555,
 0.95078953019766,0.9278041163340467,0.12085871491891453,0.10773435308095047,
 0.44848550313301394,0.34536411384946286,0.4344594394455057,
 0.7158703148315481,0.7658558050460815,0.055100056504078454,
 0.4165652605315847,0.24582625231780753,0.2994999539774288,0.5752967412832523,
 0.7216609144597609,0.09169527497166219,0.05671823885235616,
 0.6263683381006279,0.6792055129714213,0.16920202767265624,0.1421325347885286,
 0.516271720441895,0.3705057186675227,0.8797924212676662,0.9400484425466188,
 0.704342131831385,0.5562072628456748,0.272594381352421,0.15669246336428122,
 0.8312067856621626,0.450366745171633,0.6124812531155926,0.1702727927326826,
 0.07161088542425298,0.5551555964148797,0.23520072192348807,
 0.4842954844591584,0.8486833492394639,0.7438753188295655,0.128455254967345,
 0.8324244730562353,0.819422226162715,0.5720563258420116,0.4911431763363123,
 0.6148240127280064,0.6050112898229,0.07340010744608871,0.2948284469740844,
 0.0334074729666437,0.9550538276287566,0.34366770033301164,0.4351831904119391,
 0.8400836927350818,0.3602380834493727,0.8344641378802599,0.964407487958144,
 0.9057070013732501,0.0038675685762907808,0.35312940188725395,
 0.5419119496334264,0.12264073325940072,0.8650995125518529,
 0.37918179452227185,0.7669461073985896,0.3679491936680115,0.7172676386593335,
 0.5561677002059258,0.6685037075317924,0.8014127971665573,0.27760575524566455,
 0.7012098941672752,0.6504767073672966,0.5809219476323072,0.48284614515676816,
 0.3023063509081233,0.39171090916718154,0.3938673202400468,0.8801432589286741,
 0.4152323619489334,0.7839427188631873,0.767412083072429,0.43945326767380144,
 0.3494071090571117,0.04146920816412536,0.9013978915304927,
 0.21864695763819275,0.6376239221168717,0.8587647219702439,
 0.36846562101789215,0.6900907796894079,0.6747842916389252,0.3095267571364635,
 0.9678662844710575,0.8320595617849418,0.02061124828840799,0.5821076418441756,
 0.8860421527800674,0.18994526010621426,0.20615779613935226,
 0.009210552363487334,0.9900550922041383,0.49759102277567,0.8757289001890896,
 0.5142712362674763,0.12715107817144666,0.42277549050062047,
 0.14545924689328238,0.4910951895804866,0.3084930660681171,0.3606945983951595,
 0.3153913351922053,0.5131639984878922,0.370074867942613,0.7695948798680748,
 0.7614003970751908,0.7555949779861529,0.08629351960011156,0.9089575300877548,
 0.5820722809742097,0.28457042924258924,0.5799640266026533,0.6540174514762925,
 0.19559614879246934,0.3116016213528169,0.12963281042018382,
 0.6803487921076283,0.06406762580987846,0.1953202991378059,0.5236200723015405,
 0.6645145665927142,0.2983700981595113,0.3960066268893838,0.436375433880568,
 0.9254741705041409,0.9706836929619573,0.11378431864967353,0.7048615586025302,
 0.7100623420244544,0.2567230448446143,0.17286228639764178,0.7718279441059315,
 0.28035846598277914,0.1750150854240925,0.4909592738189711,
 0.10562579640235603,0.44257563169502456,0.34019487019292116,
 0.6062604281208978,0.4634685891117831,0.26224117310550554,0.3731469287176483,
 0.04446210190080835,0.5662233366006003,0.0010106926570873087,
 0.048103604537175326,0.47714168273639623,0.9861441230942525,
 0.15523646369858335,0.028939951277999643,0.026223291201925014,
 0.5164126686696877,0.29400999467255684,0.2879597821627905,0.469563630215319,
 0.19803319733859215,0.3724487256970831,0.307089579404071,0.596819065396543,
 0.41637828178314407,0.5603237712159674,0.1482357406932252,
 0.38607152315364346,0.5689113212877351,0.884715304630988,0.31136002964141773,
 0.01117270932537151,0.10493178746394205,0.23412457428269107,
 0.7485914511584743,0.3560391799381636,0.46985910836600253,0.6582135672784126,
 0.8856549589375126,0.7505533096584085,0.2739735830177046,0.4662128929556406,
 0.9645506476285394,0.05987399526554127,0.31997594918430794,
 0.8396234585414444,0.6105429079315965,0.12129694213104314,
 0.45937092755003006,0.11923343012681864,0.6282802150393242,0.572445166074748,
 0.41216186992597037,0.3307160143954858,0.8119681346841919,
 0.042698518696390186,0.19117224329488314,0.7687217382208241,
 0.06227096591635506,0.7533907307427664,0.434197624223032,0.2447854933923821,
 0.4994041836348262,0.5772146759211612,0.5807853688332187,0.503680850455663,
 0.6918153171206367,0.06602841416227012,0.9904398068377475,0.3716917343142834,
 0.9403519966476536,0.391276035838344,0.33575553302398464,0.13484519912342163,
 0.947632537012064,0.4972594196631379,0.9628248109685006,0.403775826566371,
 0.9651364086374183,0.6013081156895791,0.11606982215783523,0.8281245563944363,
 0.8623587547818787,0.2570058737528369,0.6298529483749482,0.3667863206119366,
 0.08383607879370436,0.7583518951813873,0.5159298215970632,0.4038033205834668,
 0.47969704244963773,0.9597068001392789,0.2168226999832149,0.2743329084104973,
 0.5509848383279435,0.16339478114645778,0.5536040841095942,0.8754666645211154,
 0.8224163061900908,0.3107625891054082,0.8142084848830364,0.9437001017311226,
 0.5114266549725222,0.5117978170935211,0.6918901493180563,0.21269875527588056,
 0.7902304005507508,0.6240315409243079,0.5943707096802738,0.325275670166295,
 0.5949242691315645,0.2010721260428736,0.7576838584969405,0.25281807180045934,
 0.23624587614279058,0.09186250317505784,0.19550067137213856,
 0.37624965778107644,0.47944410621271627,0.07159610374206071,
 0.8870945398183706,0.7941094466107353,0.93573632926451,0.6368121855588806,
 0.5390683564619968,0.6510403121729559,0.7655804920085004,0.734552067224065,
 0.14974733636759108,0.027285595801894935,0.7984790140862045,
 0.1603397001023772,0.3597718741547976,0.3198162520315918,0.8602528923056181,
 0.21029164665721078,0.5905288586678623,0.8425181198723947,0.6038788420171526,
 0.10765451739143683,0.27099049968759115,0.7392177621381216,
 0.9640351233147646,0.9171579749324194,0.9109930513382096,0.4673943977397309,
 0.43956637392649633,0.00993297933198467,0.9170720791770062,
 0.2963828483079298,0.4664791626046847,0.37227534631985293,0.7027738416519452,
 0.7052075002631366,0.7074787073164951,0.07892467420608007,
 0.08327174086365974,0.6363153436438238,0.7132204897174331,
 0.015192415596424036,0.7070132955966333,0.3348152707960972,
 0.13956046044828596,0.7083240196929445,0.18855299104307477,
 0.5057676419779922,0.6948268839693432,0.1597719332524432,0.6046088047714673,
 0.40430656204310056,0.8413292394201861,0.5960099919860672,0.2897191420933447,
 0.9607919252078114,0.21952430200247897,0.873979054826069,0.7964780032730977,
 0.7999049945711677,0.12138715668227862,0.3452624704564514,0.149820398355773,
 0.6981804043238262,0.8709071934879368,0.7363683192257032,0.9448379462861356,
 0.2616627440442555,0.7771895591510959,0.43073490292355043,0.6911598791242421,
 0.9760260976646346,0.9677698652077642,0.17060677612926378,
 0.08986870944156622,0.4288472682806175,0.1465530909943934,
 0.21600829964918444,0.42401943209594073,0.7943168657269986,
 0.7161401053950873,0.14391625090187865,0.8869266083989429,0.9253277982432475,
 0.17732949735149472,0.010314517391984523,0.5230493358027148,
 0.8865833849537575,0.6363761618181495,0.3028733254095375,0.8795539775521094,
 0.03457693225158798,0.5845768045215456,0.9232315931064504,
 0.46634014653186917,0.602500433011936,0.3116258499171519,0.6660129311105184,
 0.974837579778121,0.9677615137425208,0.7815802941149532,0.1183903461343443,
 0.4321315283924083,0.08640660602353312,0.12017103944954588,
 0.026066239788994805,0.9605225566179431,0.31437698345947696,
 0.5147403860477245,0.25853443561184,0.9598357222528496,0.9052482748362578,
 0.5014363125554565,0.04472754997439432,0.5575994285078687,0.7536486962471647,
 0.8292443425414914,0.6596924333736088,0.15831084696222608,0.5687128900063894,
 0.7559915807041138,0.1417370725665701,0.15334868527374035,
 0.30668719138767253,0.26872859405272786,0.29347162374194447,
 0.44282876433483365,0.4894006565512866,0.44439713303876194,
 0.9642327780294142,0.15925820201257757,0.7003611231457725,0.4537956886686756,
 0.11438056382285067,0.6023268497163564,0.5687869709552457,0.9915063000730991,
 0.8367960197398043,0.11813899161686814,0.5806262508141685,
 0.18143923874888102,0.17033005341328988,0.5190329571973299,
 0.37554587903523273,0.2560023682537075,0.4807836258786171,0.8350320771989603,
 0.32288981501898784,0.483647640069912,0.8411518740756033,0.13813471927721743,
 0.7813450880785471,0.2529601431063748,0.032716649128095576,
 0.7704627567751569,0.08985248605485419,0.4267071694426513,0.66530074462788,
 0.822159366752556,0.18527994840398232,0.43163200692468884,0.8976572616023685,
 0.928915719856421,0.48551819164512466,0.7419632252516468,0.7788408698828484,
 0.16710509370629678,0.8704563133464462,0.6042153736134759,0.7010566087242713,
 0.7543262840268917,0.1962093983597034,0.2670123269432697,0.7329410801248337,
 0.6653390444955689,0.36657940526600574,0.7945193981610161,
 0.031761563457477715,0.9743575090631734,0.5764787020980314,0.505276405049768,
 0.8225636236147077,0.3219250042539398,0.9676907198263055,0.5515018944346173,
 0.2639370743689482,0.7858287294084104,0.7608240174672956,0.4991408962430971,
 0.16710001053577528,0.1802769662800927,0.7297560559608971,0.745846412939698,
 0.743393207865868,0.2742401521112374,0.922549062049593,0.21658247698644528,
 0.5639856612138147,0.7759589891892335,0.6528740311715477,0.8886769460583989,
 0.33492838068273556,0.5368491566446583,0.03309945394586544,0.510135936004167,
 0.519600348161241,0.08695145556477946,0.8581947678502228,0.5074368512234932,
 0.9121720795540512,0.974945509114975,0.15410541284765633,0.17662704634748294,
 0.5012672649854318,0.8105702829355471,0.597233594163961,0.5878223319623544,
 0.962397780509472,0.25510571663222137,0.7990168164300326,0.9401765546627368,
 0.045475780329926496,0.8762113053018052,0.7336984633183621,
 0.6420028863641792,0.2646019267627926,0.06371764824360349,0.902437911054227,
 0.6810729489285314,0.8692515186198406,0.27720109518233904,0.595084597496175,
 0.6873543790337079,0.5227498690424349,0.9300807184207325,0.6975882030448127,
 0.36961270527843304,0.8448255223815234,0.2718163513342714,0.3933849176294676,
 0.8298223059437223,0.5920610507796158,0.777298290368783,0.4891376654726701,
 0.19546940513975264,0.7483526247985538,0.7944156073625394,0.3758971189428538,
 0.8188627288377679,0.8226829847230108,0.037315870462834355,
 0.049963436966291686,0.9087753765235951,0.29537438198433985,
 0.32371200140103296,0.9457497367417111,0.43775254304815536,
 0.48616263845222574,0.760985381812482,0.5791156330871593,0.12366091599811568,
 0.5408561039354162,0.20529723501163266,0.405988183852967,0.69684391303402,
 0.4151718897923242,0.4902070128993814,0.527934627258761,0.7442220000609077,
 0.2859791135014551,0.10512768995981292,0.6408074114245953,0.9988349113614865,
 0.959389171399792,0.9610935233606734,0.6863832007859928,0.7537629311391253,
 0.34156433306057954,0.5057381259654044,0.4887406839840831,0.868520265332879,
 0.966505293848545,0.9408884433987598,0.214921392513584,0.7347721209557054,
 0.4645102595141235,0.23577550038781303,0.8730652942117569,0.8984195641692042,
 0.74822385434537,0.4490904241815634,0.9798497052072167,0.98852496328441,
 0.7894139008169563,0.4179512972550873,0.46858573569002904,0.7735423051422335,
 0.6780333614127757,0.37879514506050316,0.8918405638362932,0.1124774351526423,
 0.8182858085925832,0.9651634480500865,0.5763259199321126,0.825342593103858,
 0.3841334096376151,0.3988631532316407,0.14823618228958335,0.5891744815520394,
 0.0915986677046865,0.7347706784538036,0.780836703212531,0.07907358216958937,
 0.3182581810174745,0.8564993623609806,0.8787064173761073,0.22836899496534835,
 0.4774826969785819,0.8865728096670188,0.03266119128365497,0.2980260809955154,
 0.22374215920142582,0.2162893705547596,0.33367901387937215,
 0.9056538030103208,0.5787802415311789,0.21584589483666883,0.5893185862499828,
 0.0709932183127967,0.6821002016541046,0.36506815955770444,0.3366727878175676,
 0.1890888570971987,0.31036302657123116,0.5643357606495771,0.1355753317361328,
 0.23284517199316446,0.48038940706808764,0.707843020654068,0.3678743919572548,
 0.22095978946587758,0.22798498606171014,0.11254662372021051,
 0.8068211558341793,0.9239194987016711,0.8954122010850469,0.5963637722262786,
 0.6136952605223632,0.2701619335539641,0.34288716784871165,
 0.35459561062785705,0.8324496640394808,0.8339172824173133,
 0.38696097006918795,0.021784373894972675,0.29398293130294173,
 0.7139363323614569,0.6897087053329293,0.5967053596902836,0.2731239483283725,
 0.5680425658235625,0.19001649340482396,0.2506436908638665,0.9792599858544879,
 0.8896027583599887,0.4348473129482875,0.41221764849540177,0.5693707750403859,
 0.607859043613768,0.6117261818897739,0.3259026936334315,0.21321275085392055,
 0.43066965018483094,0.9258312250945981,0.8318599185515436,
 0.16970378749816306,0.6101129604208715,0.8495360127361505,
 0.03592268034688273,0.4597368488018835,0.12174973767685282,
 0.2038478323945554,0.7785941525228643,0.13271011416261735,
 0.37658287778547594,0.9935633664929346,0.41861590047694397,
 0.7924161043973706,0.2602138696099642,0.7616133674421773,0.5557627206016031,
 0.25068470939184495,0.2563155250784276,0.36351373142585797,
 0.6016658889045176,0.23079520803350118,0.4257268715950092,0.2400328642832159,
 0.5312608259936257,0.9856721349105649,0.02372221973873323,
 0.23601650987534084,0.567633375856137,0.5166966228974897,0.41945064397237,
 0.6477990824781794,0.25171800893425833,0.6677642076215636,
 0.44364583257831314,0.6360491027352523,0.48740961232918356,
 0.9962988882760605,0.9147904408883313,0.4471346250946867,0.6531450900058411,
 0.21618150689933047,0.9435872078275148,0.04268647212667909,
 0.14081593446740848,0.3163024013192981,0.3672951631217132,0.5807256149750234,
 0.33571494468374885,0.14203138354491762,0.9449316425013058,
 0.8064835677703238,0.6319135853476799,0.14568268396799167,0.364863930559596,
 0.14909882574158873,0.4253405085931594,0.9179518523063338,0.9900366388960355,
 0.5919418423521302,0.3133280278860727,0.6772004633994056,0.19878616285752237,
 0.5064031076577951,0.007195933139919175,0.32557830898752804,
 0.022663557935794887,0.4892314271791577,0.45972672978600226,
 0.9763451055566558,0.03989881854250643,0.11116487010574827,
 0.9195046941914408,0.6601677898183597,0.2146394596911141,0.6688634091618894,
 0.6174303914912991,0.19803945999102057,0.21978228418100532,
 0.8189129748138023,0.9172331656335416,0.7595114423452075,0.14607913184718702,
 0.05609890981846366,0.10028707201534648,0.8652316726752931,
 0.4115798200326355,0.701109555332635,0.427646167476478,0.3693851637598191,
 0.85218773908051,0.011677841236160003,0.7672491316400788,0.632149399233473,
 0.38170933319719913,0.27783340108056587,0.63850132507805,0.260557042325837,
 0.6104264875949641,0.7921812276440845,0.21907949958139938,0.2209539048461534,
 0.7858906500608054,0.18715019699697266,0.6211367369688066,
 0.04440428955630471,0.6861940922046357,0.3939594524024873,0.6542859826959609,
 0.8146695533547477,0.8514730672075019,0.5874468538886836,0.3641478693989144,
 0.84459546775159,0.9674868341643159,0.4782981104296685,0.13081558308330288,
 0.5499479537189853,0.7847216479207333,0.9122509435495105,0.5360170274320795,
 0.1906788253064594,0.6069627488620102,0.5199519593658921,0.4594299231224398,
 0.3695695088702162,0.409179981561793,0.6306976499905932,0.05274094362107595,
 0.7816205190485881,0.9948750071291568,0.6666152723046253,0.34023824681248316,
 0.49456952945662125,0.20323425400563655,0.3918760062735991,
 0.31698303824648155,0.9776401061029948,0.8970192405863615,0.7322848797420005,
 0.21581075543766326,0.9447637049055372,0.3319348037675416,0.9936410375975551,
 0.7408066231637374,0.029420280792036735,0.2551604449285637,
 0.003624617914269823,0.3748342853379246,0.16969740457227667,
 0.9675925213936988,0.0900906274195652,0.044826683268984535,
 0.1903060987791868,0.9495842640554457,0.4476810913809507,0.25123885703004656,
 0.25837784419829746,0.6400507162317095,0.7044941958525505,0.7042230470895003,
 0.29498132916598563,0.3783068274469593,0.3995457845227012,0.3567761446721964,
 0.0626789056728787,0.2865161764485429,0.3073777544567614,0.7914240304054636,
 0.1382796839399676,0.7167681102553768,0.951370360724614,6.928913191012187e-4,
 0.1271075357432928,0.1594332705782946,0.7048721081257767,0.7084541957323838,
 0.5509610303015453,0.6522551249183638,0.017405555297826947,
 0.07728579946632919,0.1316447216396086,0.4193323975487958,0.8166779722101104,
 0.9508449562852891,0.6051356142999696,0.10180715815690511,0.3851056592441592,
 0.9735170244272121,0.989393391934188,0.8169414922513859,0.13702566011935258,
 0.7465282114624929,0.27400157918206514,0.1053796300021288,0.1293270511736314,
 0.7341751982466727,0.3383372733377745,0.4950588028463426,0.3574537951169946,
 0.07549367086645964,0.23847587317775454,0.6658297634156468,
 0.6113389595390353,0.2623653255862015,0.06340156132238584,
 0.35687052951018305,0.1942954974088752,0.9369793711119625,0.7415131413991521,
 0.2588247559937271,0.29231644419361935,0.41749655199466407,
 0.7442710826568724,0.7667198941809252,0.8176182507827789,0.501385176273364,
 0.598045403592591,0.0569817299308677,0.3800127942153383,0.7527934324949009,
 0.8186340971032794,0.42837723127699323,0.7000884947582169,0.5916367710946108,
 0.47653251130435004,0.2933457871029592,0.4796776479522583,
 0.014205675816272079,0.6653201758301022,0.8019297536428946,
 0.20895479962197017,0.09455414242221716,0.9640398976414165,
 0.5399871526180204,0.28045220101362556,0.6796334953699232,0.7227436807252856,
 0.4137147597919951,0.2517323398831899,0.0198915597455821,0.21807831099126018,
 0.4344517501236088,0.6065858904547485,0.3446558022954106,0.0326095431117871,
 0.5024804218248604,0.3240086402617367,0.17888428047180904,
 0.32229200067713226,0.44172420154216874,0.13012626446732556,
 0.42603583071256046,0.22488313494640855,0.9079290808818582,
 0.003615285410502822,0.39493124838802207,0.6937494658726605,
 0.9304893456220538,0.8928067715045652,0.8621206516404467,0.06330020429808991,
 0.8662614185388867,0.949443410279246,0.5284622956873875,0.2945831181423808,
 0.3101620137970187,0.47086823952744483,0.552557691232503,0.966895614527175,
 0.25415931837507855,0.32054012794824494,0.887719823780045,0.282318756627573,
 0.4812836252988262,0.9646946390335059,0.5239111434239556,0.13187736926323013,
 0.2927346960402748,0.5744128432520277,0.5474831924786565,0.10150088997792717,
 0.8255720226181382,0.5175488532244544,0.09712315439962416,0.4579677206466869,
 0.25401921941052963,0.131575526125835,0.6316866173717016,0.12637186025535496,
 0.6679473514019754,0.7815759439100404,0.37211711170690165,0.3415366035920633,
 0.3170345254469543,0.2778574527503864,0.19397092845386943,0.5313495842083686,
 0.6027514894066692,0.8491340302372199,0.9452526187983249,0.789272406369425,
 0.9284138097836164,0.8162237636945264,0.376242707959869,0.9787670647799891,
 0.43794659807721825,0.29293588304000684,0.4143922021573492,0.86187744924724,
 0.7531199153205395,0.19617669728163256,0.45280047052465344,
 0.2829694849009726,0.6811264509880707,0.00946405998300337,
 0.20760592878715212,0.5821735711366925,0.28930575015432314,
 0.36232359273771275,0.5480990228897185,0.8912384596547794,0.4187575787380001,
 0.684819749797017,0.053372999482587447,0.20600060833074185,
 0.31339194771583845,0.48617180904894486,0.5729658246551214,
 0.05964893763532242,0.514112642710383,0.8909662554021784,0.2889921435849796,
 0.004881468768240027,0.9932560736238035,0.38297149447605905,
 0.34454613499342135,0.6559608476732269,0.6285275786417915,0.5756618106272537,
 0.9230332880525948,0.4847703320016379,0.04379407413200198,0.6608275202685571,
 0.17666237400105822,0.30716997942267954,0.9329776966572716,
 0.4549388362310176,0.4568853620778973,0.06673976202077991,0.8827613535646434,
 0.8788113516258861,0.20620524041271548,0.4331088063154459,
 0.27617580920188123,0.4014421370468556,0.044478932601444976,
 0.10208793485764911,0.016358918147452828,0.9220598169217957,
 0.9400310511987753,0.504687904524711,0.11759281018218715,0.8202504838402948,
 0.8350604853724877,0.21609712437545758,0.9655941722473221,
 0.25486385995578353,0.4694961051455856,0.5608601955699473,0.9781742008139369,
 0.1380546623502208,0.33009586831118565,0.21257602182991875,
 0.06363245653030558,0.33733853865692864,0.1155387345633998,
 0.3684160408634741,0.7319483088800666,0.7401787808697237,
 0.012315967041765408,0.563603965514163,0.8766485496459537,0.360637064402322,
 0.5985174208080427,0.28012009210659206,0.1099293892718296,0.5144925468671558,
 0.35745493270535755,0.7774835806554495,0.7686133809970859,0.5100640027061311,
 0.24926502227379954,0.8921869121266494,0.4465464614367668,
 0.14070579027663166,0.7060303419092648,0.35951272027681724,
 0.7582671633178257,0.773593487693097,0.02400639315999198,0.8694036490302646,
 0.7489308642762714,0.23058680018199396,0.9716771110965186,0.6244044989513277,
 0.9453156985082803,0.11827018866242989,0.9972367953333207,
 0.10198661093544337,0.4310078046487392,0.28287402836185194,0.66234421204919,
 0.2403212544864768,0.6607826812631745,0.15125776828199256,
 0.19978321340049576,0.06504238993255895,0.4170187626591918,
 0.6080262738008404,0.7160143565445569,0.0010436398086656196,
 0.28282239163496337,0.27062202036535365,0.5911481960416601,
 0.8310614830618135,0.5895432284054491,0.2351130018989489,0.918343723402016,
 0.2040882935349424,0.4498598081352403,0.3508470954073748,0.9788859010971761,
 0.5917983802154874,0.7741878170439799,0.6778935947086963,0.25538554576376205,
 0.5097716864092238,0.5204501491709271,0.187655589986796,0.16556085640895446,
 0.13773398137605009,0.8556293713928218,0.05275328927119283,
 0.8835431024233172,0.7840563328023542,0.8323498469180741,0.9198458589209367,
 0.9441079052547182,0.9998996631282723,0.2865692606751571,0.08945166891303802,
 0.7341123880271389,0.8224015671340346,0.39877862226740723,0.6202055143947104,
 0.33707436825606996,0.4996319082721262,0.2878964472939949,0.5939055393666551,
 0.12398801621355493,0.8003961891162148,0.056220762720723294,
 4.854606570674491e-4,0.13026464722615316,0.391205024692604,
 0.3446165748046637,0.4746725800518571,0.060433265521949386,
 0.42289924923478606,0.6870190530124922,0.6893718093167492,0.6374161429303162,
 0.5616049878003119,0.7597889416604655,0.12117351449215574,0.4944999381753894,
 0.7802484613517703,0.6635350703280378,0.5041102574628468,0.10251318263203013,
 0.0820003842381909,0.4687238465707573,0.27715105198667667,0.4973209181715982,
 0.337114128875485,0.3012565647153762,0.3537315338698731,0.9254606644275519,
 0.34141968621241336,0.5571826017961305,0.5269966495280822,
 0.09729007465801154,0.40810643263625046,0.2274471988702681,
 0.6214660590917374,0.22101550646294899,0.23594472770912578,
 0.9092258203311469,0.1060404188844174,0.9097581122386317,0.20827860136298537,
 0.13929751132248291,0.4663716403093655,0.9008461385338586,0.222529571471762,
 0.21582223322183702,0.2962907449007737,0.30123366354949777,
 0.6813218753364401,0.47006752409859354,0.540710896018571,0.10815814340869423,
 0.8353127231717533,0.06162245363760088,0.6038759781357342,0.2308804712588962,
 0.2447462874391233,0.8507911603929115,0.3460423483149013,0.258007767054825,
 0.15736841272416657,0.27057492341868605,0.496774330483289,0.5541747074786809,
 0.43704910581765455,0.4994808793032277,0.7334792636743755,
 0.30240980598250056,0.9163480293954923,0.04425327152170189,
 0.5212989427712659,0.4971229150161208,0.48404296792376966,
 0.18177450392475625,0.4545477684164124,0.6289881595365594,0.5812431397602422,
 0.6722011304260146,0.4150805088990035,0.5510859628622766,0.22239291583763154,
 0.21740371754189858,0.9496863906307031,0.3695729170803955,0.8764423536947195,
 0.3951795307388414,0.22690634856110348,0.2302940815897686,0.4273280506085594,
 0.08852695275558875,0.4184893794264045,0.27344274355657006,
 0.34843760638825416,0.46359110072230325,0.6908239139496829,
 0.6328337005003453,0.817932962039321,0.7881127603709757,0.9937233706277848,
 0.3905666169314651,0.5536826122319336,0.7840285603001172,0.32134075469202905,
 0.9621250037579924,0.7567658741773926,0.5367185913264977,0.2503506887952398,
 0.02921172100911451,0.33438799023540455,0.9946826543146017,
 0.6524765032777693,0.2205743627052431,0.5394866527985933,0.16624992061058497,
 0.882047870447528,0.060146696092927665,0.5091666875673028,0.673701488364369,
 0.2539037587723516,0.7129871350618329,0.25245755038137385,0.5152564385339994,
 0.5295333835907021,0.6074607484879972,0.512044411106343,0.6771112028243111,
 0.6757927701056402,0.8475705004272774,0.7874390813644307,0.1298876328686519,
 0.49699168289253726,0.15460202895362607,0.8191566920630265,
 0.1397834587368919,0.6359791200983843,0.41928972593915037,0.8028112630358902,
 0.8487190625638698,0.6136272051247866,0.30860190380995,0.9723299223886375,
 0.43610209376245435,0.4183580157889123,0.9062588880589327,0.8936638803880833,
 0.6664426337802652,0.5524235437324534,0.7285812453954524,0.6498720492919787,
 0.04565047407020295,0.40524871600625234,0.12077730871720127,
 0.7551434181377107,0.7069247768706802,0.9003450029423984,0.8180457958312838,
 0.09299262562191846,0.39213081948261497,0.3443330476684987,
 0.9714717090132943,0.9312373210178145,0.4033641382433548,0.2444700209371582,
 0.4031366637147078,0.410698375413343,0.7063924576915336,0.844983981413427,
 0.157229413332027,0.8777999665954332,0.5038593676703899,0.38874235478970964,
 0.7618627633347005,0.42805302253291144,0.22640252206067835,
 0.2921357580769107,0.8590854344968114,0.8734007077240467,0.4645358163732459,
 0.4117204276403096,0.331506270449224,0.36374686467368644,0.10181761744442774,
 0.8726988420054759,0.6202799521127063,0.3395947527862204,0.6978660663344227,
 0.5437634008475078,0.11017236691736332,0.04174253334400502,
 0.9383793118034592,0.1548514572898071,0.5916163687285589,0.1255003286025983,
 0.6040372684425029,0.5390535259198572,0.9562272671480356,0.08509188067842388,
 0.825573610841377,0.8614943451544926,0.7809375360883305,0.15157066926481777,
 0.9513232801472499,0.2702366136678246,0.2797338052377273,0.49203641400952935,
 0.31835784202921424,0.9360284346943526,0.13545817162541918,
 0.9173990733231251,0.49694814847864643,0.6671840574152241,0.8327337454495122,
 0.22809308492262315,0.3840448749130254,0.9404805241344696,
 0.34916372619427327,0.40262737961555317,0.9361776034394969,
 0.24364400509078554,0.3348321945527327,0.8957519493020994,
 0.32566546267314755,0.9337751757434178,0.16416995108058607,
 0.7081885160478592,0.32445362949508716,0.2687972560658147,
 0.22418634519536895,0.8393065311723154,0.5104432502154826,0.9759726144832154,
 0.9203190225665749,0.06764912886340446,0.019425347981481766,
 0.033146754037001736,0.052805332095542945,0.42973264291908697,
 0.479397357207214,0.19389477130465427,0.5365307234281292,0.22690450914668636,
 0.7388258813564357,0.6714535449749846,0.7523341215639656,0.38503263008709576,
 0.10747885402760893,0.6559773613749238,0.6151096641375534,0.4694528285063868,
 0.03326929496102926,0.6780975631846566,0.7658171667902308,
 0.06297174885241996,0.9691295158000782,0.6622631868504995,0.953592988525707,
 0.31004641441566805,0.3145996389417768,0.4221937655656459,
 0.42437846162861315,0.7210523133523635,0.13890008650535568,
 0.5569725671356826,0.6542552285542473,0.5100825016127412,0.311536332289289,
 0.011336787289686123,0.31874224510821947,0.3891624031863624,
 0.5651802452202223,0.6631311617061347,5.992687892584225e-4,
 0.6618950246636522,0.8295281347480509,0.6666396670649054,0.2588603197910736,
 0.42589900943516845,0.7447327226410384,0.9969533510821202,
 0.45049302647976996,0.3590410593299771,0.6122238152253832,
 0.04435547106719351,0.6556743542338443,0.3922624502552019,
 0.31545001411407336,0.9684360403799301,0.7884017441290332,
 0.47074739440436886,0.23061718778007856,0.5761545796258848,
 0.23920922583777304,0.14127190829815173,0.36781804946613805,
 0.20787145956383946,0.6167781536704686,0.05527138507223894,
 0.7988827198985229,0.14264420818613854,0.6014547981672271,
 0.34996366290865333,0.9147507704822994,0.7668551965406507,0.58904219819104,
 0.98833207422009,0.961926552164924,0.14891776716180039,0.3954123192096719,
 0.7129264212726755,0.8389451206041981,0.4208074782722314,0.21566066769527847,
 0.35251903198658274,0.6419714736625728,0.2783894818892153,0.7557321339848954,
 0.29577653368322343,0.7647001768605766,0.09051896342749088,
 0.9628666510279845,0.09751058374326849,0.31733273875855317,
 0.43970432997461195,0.44040106116688893,0.897459154607467,0.1591248459251553,
 0.4402903923401579,0.39137684326677424,0.21681889680414357,
 0.6632104429749952,0.6880698314081617,0.16737116778575878,
 0.11851962108617065,0.08383305098363569,0.13557664747967624,
 0.13054887488588374,0.8497052554884656,0.22578878760022536,
 0.8584556268771198,0.5065943440451883,0.23668596829236865,0.8115149554831881,
 0.42069541259845344,0.734019474708899,0.966706401547945,0.7014370521843365,
 0.7350451437657962,0.5549582731668942,0.8414856528500152,0.997139889494841,
 0.2877417419268816,0.4835273985697175,0.4320090801000195,0.44748581513906605,
 0.2703745309138659,0.29597217026285183,0.1820654108070937,0.9402881107562883,
 0.30699580640352364,0.648824665024167,0.5586184120148058,0.6596317851372706,
 0.5363150468223528,0.9754231158953123,0.38552277546902536,0.6947222919994012,
 0.47399644248706607,0.3769116497780457,0.2038815502427893,0.8648850448982733,
 0.1479001295037603,0.4128869729395038,0.047322829118916054,
 0.3545515433279721,0.21182901437843527,0.7643882319774694,0.5113748801003979,
 0.998221939890549,0.06930231207453919,0.25008275066339314,0.6787998351963728,
 0.5678467397426954,0.6619565272639976,0.41484223638570084,0.7289341299454051,
 0.3872174067708266,0.1836851381854807,0.5870829544188013,0.9790650814329563,
 0.05006049466116025,0.7521291312105989,0.28993678011042945,
 0.39741669553903247,0.2940291175196731,0.2670968644531099,
 0.17190982880526484,0.8210569028213902,0.7575390468941854,0.7917960841097504,
 0.7544435836435142,0.2745057510151776,0.0602711195184098,0.757048634947551,
 0.6177705392589387,0.03494327754038762,0.057973218139083516,
 0.7803403466471112,0.5672015196618416,0.5598811325245197,0.3703417600726451,
 0.9621697334085872,0.40510409367275546,0.8260471107681105,0.5037908179805419,
 0.1632881383054503,0.6833443520570504,0.6268438644993767,0.8980558039631227,
 0.7108204518306869,0.0028095958017789746,0.9894563589525699,
 0.5556237739071381,0.9596311568475476,0.6327711546514763,0.5161166824544969,
 0.6080113126950346,0.6495754717906873,0.663642837254149,0.30565002002887065,
 0.7278567228784169,0.982350278688123,0.8856655730405859,0.029911974141435316,
 0.4325152681538732,0.6635779650205947,0.5991992386086056,0.49603268316047344,
 0.6838880848205736,0.35647129402150224,0.8752047216098128,0.4536463188918868,
 0.8596007530423513,0.4985608727237454,0.9579609419704381,0.8526250659807597,
 0.10326886639938648,0.007678381546915025,0.3940162485420905,
 0.19112357181160577,0.02340601907690587,0.6633713233119137,
 0.1252175511796293,0.7968716158358043,0.10335391009456885,
 0.33995714853788905,0.06369453425953764,0.21644559945615538,
 0.5360476920500618,0.8876980646377228,0.06215390585911229,0.3411721706664237,
 0.6814939045780551,0.6464201098489786,0.7825453459707609,0.1852965444711846,
 0.05538451330594618,0.4293730803318181,0.2888062439504534,0.928656025063139,
 0.8402614142045725,0.45241202088539423,0.9530576414223579,0.8570021137614854,
 0.6461842511388943,0.7118999041027804,0.1602395418019732,0.7759135182339402,
 0.6527805115635211,0.5470219838367806,0.20551969150362026,
 0.13260565036030236,0.6588202297422351,0.866351527605741,0.27000475000327173,
 0.9214568666998391,0.10982607389435506,0.8458125774062235,0.7271361645226174,
 0.07000278321061104,0.41757234819935485,0.03753258454387054,
 0.17576607861211124,0.8233153131192916,0.004612835028207307,
 0.782676185594618,0.802211037736982,0.19798712125355045,0.901254108544733,
 0.7557650996313634,0.1690988814212726,0.38266227593155444,
 0.18926764310773925,0.1452074029567506,0.5276040791206643,0.8715060378301509,
 0.031670632047839486,0.3977874161482817,0.7213484647986228,0.804096188662589,
 0.9862497244296602,0.6940277482198007,0.881697675777412,0.986501640265267,
 0.026382415853427932,0.5732019517118765,0.5931088642001263,
 0.6723686579747277,0.20080982777628498,0.2861478598261513,0.7613412792833242,
 0.5493438522680382,0.0804230740203058,0.19537643554300643,0.9749626343968211,
 0.027342057837415612,0.5492091206345182,0.21233149172344734,
 0.2700429590052207,0.6504177812185947,0.9498290692536695,0.8728641077794612,
 0.2711090985925997,0.3309816511279233,0.14356373335819128,
 0.23010945974257835,0.18526350535084002,0.8148196518225207,
 0.8255831105764649,0.7876127119012648,0.8637041083665484,0.09425478351042338,
 0.5493256195741878,0.41716241035121837,0.12228391982639786,
 0.1876286935392688,0.37471125300051145,0.951610493254771,0.8326928702670923,
 0.7970399736073326,0.33629296323950286,0.5933907289915937,
 0.22923696216567846,0.14145846185044175,0.8413038455593285,
 0.25639515912333777,0.8549386857057257,0.10466684689685468,
 0.6666924311849375,0.1043455281504686,0.5793344172572948,0.01995544949946393,
 0.35078427554026004,0.3996002653845643,0.4363588531898587,0.454426717902113,
 0.7700222360423137,0.5760493732172175,0.7601071839104305,0.7481885709255798,
 0.4999154664316984,0.6971903606984153,0.07078073003386676,
 0.29861585886580533,0.3444106100514778,0.5069731640324376,0.5255045689657976,
 0.6849694794868617,0.33889257355369184,0.42043091945682265,0.825777305707544,
 0.2681721886694445,0.01180576852531201,0.6299839128009868,0.9501820070056013,
 0.6023827502240302,0.5632862828277201,0.6201966359307479,0.04075844972914777,
 0.27523092783600156,0.18810217599083423,0.050225394225203224,
 0.06947117723217211,0.05262525445520816,0.9856074193232667,
 0.8372441005913598,0.8531526538004873,0.8603923406670175,0.6458835231297098,
 0.01490038678524852,0.8555120922645744,0.7152545977261712,0.8821017734558232,
 0.4152547028089181,0.4383533805577262,0.4025820931171973,0.313874862236004,
 0.2879175610368857,0.38463846880357533,0.3628419212898102,0.8243179398489214,
 0.45511058816537286,0.14697401870786908,0.22494713241745523,
 0.40482381687516156,0.8640978136998156,0.4813313071026103,0.5881859165308094,
 0.59370203872868,0.5154458774686855,0.8481524022115988,0.40248541732310206,
 0.9814870015342657,0.18814228255587762,0.6577528515042983,0.8385702444435381,
 0.5609350335170485,0.01806888264532036,0.05875285987088785,
 0.18011674999135796,0.13963543668834622,0.23784734740043334,
 0.4056448699319013,0.9366098929857034,0.03783879620341313,0.6824543592522416,
 0.43065230153225764,0.26115036692361027,0.7672878234058309,
 0.5120273181255687,0.5532001383034969,0.9262027813131399,0.6598808315600632,
 0.9287863487876705,0.07985880165464654,0.18130265006314994,
 0.5280938509912494,0.2187036786299782,0.7300993127340699,0.4375244533103735,
 0.603072096376204,0.5465903245881496,0.6195279558364729,0.1585517905220032,
 0.34974589063197303,0.3101621414915545,0.8708716579071798,0.6871537325450814,
 0.3571936680479316,0.8351683937343982,0.25479273101106525,0.4244526780409057,
 0.6304670700840721,0.08636076089445421,0.6497480197819097,
 0.05837977808080774,0.25323469870286686,0.6039017460572574,
 0.3408229457023023,0.9489029351859788,0.253907547359687,0.3705782549393184,
 0.6164179714455403,0.47156747835136326,0.621685513338833,0.793126005310555,
 0.7289228446681785,0.7002598314014571,0.31176073376014457,0.9802011871698214,
 0.6619078173853438,0.11981694229902995,0.3566256976938237,0.2843395877291317,
 0.23644504597308358,0.43454096435220946,0.741630331881825,
 0.03678485635730344,0.6291621343349583,0.10112880586930917,
 0.29313751909509145,0.4897970761517584,0.48588505799634385,
 0.21533012577567567,0.9457243713796264,0.7234498556827444,0.0798487305882607,
 0.6730065298093202,0.35646874348134316,0.7403650217349371,0.5176615710793596,
 0.2630301693416768,0.7614624845684622,0.13990206080271617,
 0.03225574910569684,0.049694567382891575,0.5714711797113109,
 0.8117142103224785,0.2801090561860936,0.874530987036929,0.6104057138853629,
 0.4736485861690388,0.41563231505647247,0.6801069017195315,
 0.10523703524555394,0.6958583234117207,0.8937715165115214,0.332961380543335,
 0.9016546935663349,0.5438448799095242,0.7492216044878096,0.7227298876737359,
 0.02084754013988499,0.5261886253803374,0.5504580886498698,
 0.39395099597192795,0.2116573921476418,0.8017040653864639,
 0.36621580625016714,0.3550758290460062,0.16335953233669875,
 0.8813908995549643,0.9171015543762443,0.6183114752286311,0.9088856865692865,
 0.3438824415728974,0.9127780886504987,0.09969736458249534,0.4523986352642895,
 0.280587985890732,0.015760881420758377,0.9925042901900952,0.8766221186669844,
 0.14349564087187783,0.38419005111561966,0.5311100753227651,
 0.026638285760858582,0.09318572911435341,0.7207148010564596,
 0.4112408096163601,0.059250724395263465,0.6847232138946945,
 0.04573903188279471,0.21068790429782136,0.21253624293577866,
 0.8426319437298808,0.6091372212620942,0.49669330174015613,
 0.39873453845891693,0.8802373740779177,0.7015951895410252,
 0.23878709009301902,0.13067825389259968,0.022624107809458338,
 0.1332922089050642,0.3076749415289801,0.09967117328334762,0.5737233180230801,
 0.5961177802026215,0.8958887206149194,0.13063574384278342,0.8317537762017503,
 0.9456110941119704,0.10862403015668476,0.35584053134261173,
 0.7006168050072414,0.8820890318903998,0.1507946797421692,0.5508959580776139,
 0.5826429649330425,0.6634214349067724,0.17414689122766225,0.2799020350669402,
 0.06042878839157917,0.16648516911702427,0.93535442577602,0.4230978508666924,
 0.4436511045880829,0.8593288022935154,0.4031230193505648,0.9687043680566144,
 0.4201463799438685,0.6122142320951918,0.09035409131785543,0.3124842121728605,
 0.02626497925069493,0.623166854109131,0.28571730924675,0.9314456287544439,
 0.5672936644132406,0.07516493229483945,0.0355884634517285,0.5345371442636107,
 0.2553571269964612,0.9927111821935547,0.26370394423912147,0.2767659899981656,
 0.7649307888400605,0.9397471238146891,0.4231531502924478,0.34534061464134613,
 0.8139439061134811,0.15599741033537506,0.6812394678286587,
 0.37949972051410763,0.9242380016958273,0.8527952741095639,0.9443887513245139,
 0.3458374352869813,0.08997339517050462,0.971665061229207,0.46856291837217634,
 0.7225410818163457,0.09181509925907583,0.6038249550154575,0.1386030722079835,
 0.8964874607115202,0.22943265003292201,0.03013348686619277,
 0.30236355855055486,0.18054595579495347,0.16635447910017198,
 0.17064904933059177,0.31957844336431407,0.8665165885700881,
 0.6696421666493416,0.23301822648816461,0.7722483252005093,
 0.023402632459794903,0.9617464084241997,0.3008254588013285,
 0.11719556716803325,0.9459632500426155,0.867760739220615,0.22824076660298154,
 0.4894507444516577,0.9607616930921156,0.8180294798124453,0.4832537405937666,
 0.9294452961664026,0.6451222432473691,0.5421894757881327,0.3555354137801382,
 0.7494283641942973,0.8830648729537057,0.44726793866126924,0.6644044824707966,
 0.32625415514754486,0.795536994852777,0.7663234923887172,0.31262907227597503,
 0.8928404560844024,0.3673619360195748,0.20511975414958483,0.8907615371048774,
 0.8998316453773638,0.5126247660016462,0.2986296274098327,0.14802469921045824,
 0.7153608501664178,0.7917458470059681,0.4851482335065057,0.8376429771151253,
 0.34298419281863024,0.4228927827871862,0.15498111369724665,0.274346790880438,
 0.61314588495064,0.19515207227346376,0.013214877078246201,0.9131046367755515,
 0.645033193823142,0.5967676264409238,0.7557588929529064,0.6820444668327765,
 0.13669032615967946,0.39589041933229707,0.7370033989545296,0.929356466801133,
 0.7566166833127582,0.3362228586953564,0.37552040812429455,
 0.008289213175073695,0.7249613467497319,0.32331120442391137,
 0.4349045499679922,0.6850019160805345,0.6425810473338215,0.5369936673111851,
 0.6586367020826904,0.3077647184513538,0.960546297594124,0.363843503066841,
 0.8186858646890747,0.6909431882000128,0.7354941405892419,0.6810539349360822,
 0.35235489399418973,0.37502748918473805,0.08110303654541584,
 0.15505168930150426,0.8851849503781325,0.7544777686201875,0.7740913440042294,
 0.2188361245376621,0.9634600393904661,0.7061612634740927,0.5197158588062211,
 0.1459996316788773,0.9626349310259333,0.5963633184871779,
 0.053148429842458045,0.7957999658825946,0.060268700604295455,
 0.11100898581037122,0.08938653920182493,0.5141253528425332,0.905498837004545,
 0.9958041247922975,0.47052954296472305,0.6821987228462824,0.8750741122400092,
 0.15235802280386967,0.9687670306368023,0.30138879750127456,
 0.3772735937249583,0.29861454846543267,0.358145206170404,0.3975489357144507,
 0.6889724734530458,0.6831574757680265,0.2368404782906809,0.4175166596568418,
 0.1094685384387365,0.26829589677435706,0.5034626338565488,0.9037710946775536,
 0.9287265050761195,0.5671807147214328,0.055699700648523476,0.949249951700699,
 0.9032800076535283,0.28170649809913295,0.8098422015444506,
 0.26882539505151926,0.6332329893927695,0.48325643448897493,0.438622311072579,
 0.3164466669144088,0.4448745217112555,0.12212116036895215,0.6741068487429073,
 0.13123995078989215,0.7988363467715223,0.6419262731635877,
 0.05221214070879576,0.9472688136808295,0.7149465540015538,0.6708907403545943,
 0.8068695777658607,0.2135756457697615,0.8860086440421334,0.3405204317391699,
 0.29026999937662545,0.3813756836592581,0.5709858446958623,
 0.43438330524595414,0.23033715978836744,0.29150392085612875,
 0.6832307656810391,0.20572740261401445,0.1267172710643898,0.4924122443059159,
 0.41099611301587025,0.4639348151138619,0.2176541135104677,0.8361163665020349,
 0.8249909530379298,0.18202870297415874,0.6741881663701228,0.9732435829652932,
 0.2730796691012204,0.03876284353088266,0.032451230377400275,
 0.8297164082091316,0.12417795949292809,0.5060525244452312,0.2252814776074642,
 0.5413670596812803,0.8530003795615191,0.6948553459551037,0.768526157787545,
 0.9385432123690716,0.6678133022046109,0.43490416043357616,0.9922037506451705,
 0.7406205826888781,0.11817824162718582,0.8465815582906304,
 0.04257495527242827,0.024749881461620493,0.1493350159642176,
 0.2219029905171066,0.4900352875133338,0.3694773352745414,0.5515793480045215,
 0.4860900077236804,0.9907397876907713,0.5717065745970891,0.2411835344404949,
 0.9697996384609138,0.4693728046297734,0.398710873031473,0.5076975371438083,
 0.01660022089191371,0.0395699291402245,0.3126419528685276,0.4880581019711149,
 0.3105794142598559,0.8482605815096192,0.3046657517207543,0.13925839393850936,
 0.0345633888090664,0.8307071970839303,0.5616406932801168,0.26781282831726017,
 0.27424803847303914,0.8972830719361584,0.06596169366274895,
 0.4241895115057315,0.7716848391951828,0.9878280434783949,0.4337718585681374,
 0.2100926211102836,0.5322266943084567,0.9205291033742788,0.9603478648478472,
 0.6039700524039111,0.6075787759951268,0.6719542761250541,0.96477575576594,
 0.008826371149566525,0.790139525166316,0.2482718501366168,0.758186080435843,
 0.12182216166496584,0.18613738582786676,0.3134033428205598,
 0.5259879894134201,0.8910128707660336,0.8374712590807625,0.9034539241882545,
 0.7354321501555136,0.12676325168389413,0.6753806498230078,0.6892049324909788,
 0.9631507807831001,0.9344700185680909,0.9530842440328886,0.22181730026629287,
 0.42271522337904044,0.7079290577446479,0.12185750413380492,
 0.6066222381364047,0.550043393856431,0.806663251266443,0.14490476049628376,
 0.6087432938659436,0.2576447442863308,0.47068812405300553,
 0.13715408971053067,0.16731809137913467,0.018534363318935432,
 0.23902430791909612,0.29424419203561303,0.26932494442312804,
 0.8357567813325595,0.7209758068605391,0.4176975751675913,0.9228535115246189,
 0.5270232667079651,0.7033824427214324,0.9265935561370803,0.9790169404431196,
 0.4900910128416809,0.44989778903517197,0.8285972810324854,0.8424646086788509,
 0.799942889754042,0.4427064991441396,0.4136831633553687,0.6807845742060696,
 0.6268529151193653,0.6724143849152258,0.1314021746128904,
 0.016769736021093662,0.6359967914728306,0.06929729215097546,
 0.1439279451615708,0.43263466867317857,0.44910171227624207,
 0.04712900245460283,0.9887151189434216,0.5196978277079264,
 0.21567583303574278,0.36968259830537975,0.40618447076174036,
 0.8580328255811487,0.8857343003560101,0.4660524048250968,0.9349754001864403,
 0.15800554297249725,0.6875857539305775,0.7924398585803526,
 0.10565102771287826,0.8591616527732802,0.7020941487300028,0.6074159796215957,
 0.11667522573328337,0.3881122621429006,0.5413881588597721,0.9661427755391183,
 0.6927003033260634,0.41232656511105126,0.98925462712872,0.5338081456506433,
 0.8671470425035879,0.8727336144822542,0.3944161496226892,0.7576968953516062,
 0.8629291833147231,0.7344065552118083,0.39382027349096593,0.4402010236807945,
 0.9730967632159937,0.12042730930186496,0.39768352576811594,
 0.7842757936137157,0.7655349944854402,0.23269154755957178,
 0.21217563917006477,0.7476181291537412,0.39770386830157856,
 0.1697754281766971,0.04680254456616062,0.2961798076139237,0.7794664325072976,
 0.20395505503632194,0.8098936785535202,0.8153822621449062,0.4883291967648955,
 0.9378362086404497,0.937888869574105,0.7562092237284457,0.32687540819274874,
 0.6219789714599726,0.027706643164627787,0.11481663838146139,
 0.320302993705208,0.7530264021631783,0.7140774707600053,0.09622243126497576,
 0.07359001820870148,0.030665504012470812,0.19147052985797341,
 0.9653830419012278,0.6912260213622462,0.4354544947798191,0.8842565824159594,
 0.3723131760841639,0.41916704678692074,0.11207763277614202,
 0.9109667821633404,0.6685613395471577,0.24435770246592659,0.3737209513725167,
 0.11397984539274586,0.9038834130369677,0.5451892425783597,0.9079489185054009,
 0.41266576742398375,0.7289264149551146,0.9779296411986366,0.2965421886718983,
 0.8976667730725143,0.13910024337492255,0.13706205176703445,
 0.3290355455459517,0.7898893095126982,0.04703210629041821,0.6143187518408121,
 0.8159400137071605,0.5851501798315929,0.2093356316756264,0.39340541714329025,
 0.014913479996676027,0.024804642561150114,0.40039543022494595,
 0.23627930352491455,0.04787047991055304,0.2626405799630118,
 0.7099452775744882,0.600017283256853,0.5043973898212374,0.8373431415468288,
 0.311627442834873,0.36968124168375405,0.04130677308969888,0.5990885746065165,
 0.7208540030343521,0.7486029790566623,0.9030420866372598,0.7564446725202699,
 0.3302410672453421,0.27519830912288545,0.5099327564110074,0.9331663905413996,
 0.7059330064825664,0.583168197528919,0.18731301518504617,0.967229635621888,
 0.3767036353686336,0.7729124646379981,0.945243633888397,0.5432444830541948,
 0.03753539370915304,0.009663049138979307,0.010416471389254278,
 0.3218075831592683,0.5791649404649064,0.4032272458214178,0.3701065306912419,
 0.5036634627974066,0.004820734280955974,0.7789524163253159,
 0.6479663167527246,0.349300467446672,0.8223538475317381,0.684160576755835,
 0.5385045007289073,0.9290515585356756,0.728307861676797,0.5731781931296317,
 0.8838278715710853,0.061798842593322245,0.9767959230108726,
 0.6642871005744837,0.3544344039181002,0.4460293241545835,
 0.035778752458493046,0.06812570517094763,0.7486804488478367,
 0.363415819141761,0.19216069883029985,0.4671035963197745,0.475910910476409,
 0.6862131993831363,0.8451806977256936,0.4100523772817626,0.8963953690874681,
 0.5985885419569813,0.3406077525170824,0.7295692109643771,0.14173628448925846,
 0.6518630304350193,0.7933946740970407,0.284696926904501,0.842778281984452,
 0.3013745804842074,0.5758479532302816,0.35776772827569303,0.6666628103956347,
 0.2596899274174145,0.4427817338368387,0.9226219574779291,0.22846802180387504,
 0.933684016156795,0.032206021238592064,0.723612964473118,0.5682457318451758,
 0.7151579646218384,0.27167600443011075,0.8767084705047805,0.5657702663110759,
 0.7766216802365874,0.009568640867608558,0.2633369674116205,
 0.9809655969929261,0.957991613986307,0.4164266202516146,0.9751117549097391,
 0.035287494563862154,0.7110563708082156,0.31538274843377034,
 0.20670483894814695,0.8317239615784588,0.5641434081834138,0.5898904772409161,
 0.47325277331141724,0.3592074670325932,0.735476904343376,0.08736000120597831,
 0.1932934992601132,0.14529741383963024,0.7034451382362104,
 0.08457724555128265,0.11288127981362728,0.6273078979322725,
 0.9626191013080001,0.9897024987478809,0.704648168034905,0.7826939486708064,
 0.6966383279087667,0.31853092586880327,0.3263270662870257,0.6931947733623666,
 0.27691378777896125,0.8880495516714154,0.1379828229167196,
 0.03764609528670382,0.20388742908228707,0.8140912753597567,
 0.8640760303342209,0.6524389774056375,0.8482986899138291,0.24194878514684315,
 0.5362647475090137,0.8398241780067852,0.5396376869602281,0.974126442504088,
 0.21096111757430447,0.9200168758816373,0.0676511396942816,0.9181872672888385,
 0.6648807309986073,0.9658025268917656,0.5197692589987508,0.35162615449879453,
 0.24904867032486822,0.34579817609703845,0.83181854855875,0.2725001607749693,
 0.15941437749216136,0.9655871995124568,0.7340818418550056,0.5883635813977856,
 0.8269988005694614,0.3640408499565029,0.9771973915855184,0.8044885467612721,
 0.5859978268056313,0.8263977177733723,0.5400383565666237,0.08792631966653286,
 0.8015548824875369,0.3921027312156037,0.10189186558999064,0.7588902433118703,
 0.13657243716512513,0.5220043347937933,0.738001270547167,0.46417716920239027,
 0.17112606716087508,0.26869009772891383,0.9294691725787854,
 0.18259702023164526,0.5086945472452999,0.7253179979950775,
 0.15436143577064676,0.7608185298960074,0.03784294405267874,
 0.6338035080761545,0.3664168785282842,0.5163268417672247,0.03610679808843531,
 0.3696722317133373,0.5127738620067053,0.25867268258688403,0.54521183365193,
 0.7945082333466439,0.8161197408790603,0.13812443818293874,0.6444326410432096,
 0.5650503822897535,0.134992157457401,0.1449530916570535,0.5360561616575124,
 0.8823382418503025,0.05541208503674189,0.010368945780076189,
 0.4470302889561798,0.4415717450701102,0.35123619883976054,0.4269763673250504,
 0.4746291009971031,0.36248715314613267,0.843366244320563,0.178660947411311,
 0.47919668003273586,0.1681905891576313,0.8190117732144206,0.8118759156768524,
 0.16460037801248706,0.6912950271200392,0.035328091671656736,
 0.15000920119245864,0.894437191977923,0.2273911633915574,0.3323891918850159,
 0.6312851423431529,0.6145276987752353,0.4400736444027621,0.9925939717337603,
 0.10908759960396852,0.8050801819197807,0.8905403807330463,
 0.31497700530635453,0.5684050012830646,0.7534562820544445,0.5844931143820618,
 0.2491199752047646,0.11557950230627079,0.5249425760328466,
 3.490817309211858e-4,0.9714747670135523,0.46409420594162587,
 0.020246033823233978,0.5518412756743943,0.4133090531605704,
 0.36469367460586233,0.3043447823312825,0.46669323375474725,
 0.2202959236737273,0.35232471170056134,0.9192208464272991,0.819023573795458,
 0.505775127632563,0.5456894559932635,0.3138588556668678,0.450452679679735,
 0.7558723569724483,0.9577555130500022,0.3313551170924941,0.7591354890773873,
 0.48412577288947967,0.03509181420893692,0.13974825157020443,
 0.8954356922729614,0.9379122049485997,0.5137102802714093,0.6080543847876573,
 0.9980654120001451,0.33495401330525976,0.017512841207207974,
 0.7826783935906013,0.06364617687366958,0.8058102396400662,0.7721462762386002,
 0.043708439467744276,0.05562260037840672,0.4073879695407714,
 0.10359354322981307,0.2108382495762645,0.01932909721694087,
 0.5789957707880541,0.7960504996252422,0.8385559741489547,0.5841411767035405,
 0.8998965525049215,0.5866492431975701,0.08625235656452246,0.5509166119312642,
 0.3319900831811321,0.9423900489265062,0.5885547233943859,0.1753953928123425,
 0.3062324295811655,0.45156223827965747,0.13743171682179134,
 0.33661505501300715,0.0701825198618099,0.34429282805623795,
 0.016965879003407736,0.1981296210475263,0.1306560642587864,0.517315877191699,
 0.526427966162197,0.911095396684687,0.7470001004499625,0.6577509514323038,
 0.8109665915220654,0.46863210788575704,0.029251205865174268,
 0.240930692228137,0.05932126048328734,0.9548261058273311,0.7473307254337622,
 0.3779860678383844,0.7851951318567243,0.7055457775699427,0.16158458640482753,
 0.24386476006044988,0.8163652507689341,0.5908547879899407,0.5424878287173636,
 0.7621258615757759,0.6308633029743355,0.06728466297167968,
 0.23922903619829894,0.5853820395504612,0.22581816626395568,
 0.14837715295902631,0.19796293220406802,0.2763500421950216,
 0.5018530108968171,0.5104355407305564,0.40694948122963925,0.6143529721467768,
 0.30777305702500823,0.7580038945931088,0.5876283265354973,
 0.026117385531948756,0.5541733978379365,0.659607848253562,
 0.42656912656710466,0.9153959351525092,0.5827590485196189,0.7838107355357327,
 0.8013071719726865,0.12950614474260647,0.21739891413600732,
 0.2674516227767666,0.21527966301271018,0.8613871292596144,0.5919861984069055,
 0.38805461175264877,0.1842523192952794,0.9961895672355725,0.4241886262272694,
 0.3319098773691085,0.5989761430349289,0.6960844189984539,0.5702670286828233,
 0.1425147446984254,0.32410272373945626,0.7771176447381153,0.3318055644049217,
 0.53173906109108,0.3700720584391123,0.010853127721552891,0.3081652455791397,
 0.7630700446209853,0.7868310102627485,0.7324620690032377,0.5682056737081418,
 0.6990858810741476,0.29158125309048266,0.4538402326673022,0.4188661779249304,
 0.811617592689675,0.2624792355630632,0.09591486375600333,0.5222185577699184,
 0.3775865606953731,0.49349957309951376,0.23907816619501676,
 0.033353527788501536,0.10828505493223683,0.009434886151846511,
 0.21147429899985948,0.7306729826736923,0.39368700025244086,
 0.3001981498839553,0.8232472032403038,0.015437604691319717,
 0.1811030092788274,0.9951041739792634,0.44848861997148004,0.7549889824879419,
 0.4085504846715964,0.016750569562431483,0.6019129103609508,
 0.3111984543590791,0.2789446004971925,0.2219113969748686,0.551037710838788,
 0.25901182851829807,0.2027550693006125,0.23183488164256794,
 0.9186845167734663,0.19682114240579585,0.16872270716639914,
 0.20793143011009674,0.45494101679099985,0.6269082156793554,
 0.0425318634952816,0.591961057818218,0.9266076862520998,0.5171872608256406,
 0.027838229995955976,0.797012565355743,0.8137165340880771,0.8695256368045213,
 0.7710445542795095,0.7750288389989088,0.061490745893002186,
 0.4044697307541172,0.58848095469964,0.6730627264873094,0.008610810357803422,
 0.016174353874936043,0.5074434899997257,0.9926303031329323,
 0.9535424588723053,0.2678990380997137,0.49344821742845,0.5136276802234385,
 0.17414279595120996,0.5069022517089855,0.19126079096700976,
 0.7381717239399784,0.9764746200748836,0.5122561469885316,0.7626676269157344,
 0.37627309090092165,0.20369214968952187,0.8555098476350713,
 0.46650514055871006,0.46107020785879105,0.8223382490743156,
 0.8726468747249401,0.15453345908360683,0.8783800256604394,0.5914048494151979,
 0.3990401966717483,0.23237025147139367,0.21230027925293105,
 0.7691251530383854,0.7201992246889949,0.6562812344557601,0.2712131586061013,
 0.183031636104783,0.9999908131770858,0.016668517240483127,
 0.23422628420469127,0.3493755555625673,0.3494805990661317,
 0.033098542763353334,0.8465822028168399,0.8704406837176554,
 0.9244033846497581,0.6142738811197218,0.18756763501987228,0.7104149742100581,
 0.40318856099963707,0.2839105567345501,0.9505268560562408,0.5808193828625543,
 0.9768847810218084,0.23516533307780496,0.40160569403415336,
 0.5626143434992111,0.8923875072997473,0.22313786075951048,0.9004264169789962,
 0.8747068163665438,0.5786903660181361,0.53668384713461,0.30468502399964437,
 0.6567830002050216,0.5463001225556167,0.08907205013076769,
 0.024707658029711976,0.47512742190933055,0.31070468015103725,
 0.48441846997168847,0.02438794868155103,0.6914493147404681,0.800403348622326,
 0.1980472025826897,0.11983686600621091,0.4342130728307988,0.3915885101296961,
 0.4221312332808189,0.6258405212151034,0.47092040683286784,0.8246262153653926,
 0.1635012275946065,0.13981646982341822,0.3263933418745748,
 0.041128697616762855,0.4086664350642776,0.2701973940731134,
 0.6210011648454421,0.6092527875027598,0.6714436977353327,0.15153421141999324,
 0.5721651136474963,0.4378681800513048,0.5916392525871034,0.6338073614536649,
 0.7452750347578523,0.2069413721234331,0.6554323233303618,0.37056521657780617,
 0.9704346436508478,0.4187936179288715,0.03431601730418432,
 0.49737573945292357,0.8789977440413461,0.5771061566154052,
 0.03997577222968807,0.559608175630491,0.11194558297407575,0.8013286757803594,
 0.0590182876981481,0.11091606512172103,0.09308770215569129,
 0.09701896087419859,0.08608933571146549,0.5833345127277965,
 0.06077880535529245,0.9888428014287212,0.03394773543195162,0.453192464032775,
 0.9481623436296465,0.12395234972476965,0.3930507740581154,0.7602228578093937,
 0.7046963060993663,0.37183973005858983,0.8794495993943235,
 0.06734408387853053,0.8430274262822911,0.012383444040317215,
 0.6014254986237594,0.3531345337776516,0.9107704019090066,0.5626390649191987,
 0.3669949968656736,0.8385672523273519,0.5561784064521328,0.088536735087274,
 0.14308069167974757,0.8070237841786917,0.5934349061086557,
 0.31349062329470734,0.07985212285474852,0.5258755643689867,
 0.34356058665875366,0.002527156026071564,0.4361785719338286,
 0.42089390301734797,0.13602082445223196,0.5198563708767258,
 0.07173507866268869,0.15839888290079662,0.3035072114243754,
 0.14918013027431676,0.7335652756214506,0.07023082133798142,
 0.7604341525149778,0.11332986740515583,0.2933175719912039,0.8630215785297571,
 0.012089174088252497,0.6796152076597148,0.7376544361406252,
 0.9480977158320196,0.5165795862943197,0.675015630597351,0.4131841428973815,
 0.7328386470763153,0.7932044681398959,0.44436372547441316,0.3513208790457273,
 0.578594636890109,0.7393831895286544,0.30987129323182194,0.16997220125687262,
 0.3111964640380457,0.914295319160713,0.20688751467888167,0.20224945730587085,
 0.7030251502347156,0.16320523200960402,0.8943914347958303,
 0.11226349355324261,0.8563257637212264,0.6194748696997168,
 0.10432950428278409,0.7333112329399881,0.45453645776294915,
 0.5040763226538002,0.4566994060850662,0.07497977212817586,
 0.35584145328998196,0.07040506580426997,0.8569626589453568,
 0.3164573000629216,0.46015562818673794,0.9366322372206244,0.7773589347934045,
 0.6542882303738214,0.19313030804267528,0.4929944147925611,0.155763079326397,
 0.7631207417161079,0.0035460232478005116,0.30018837198201576,
 0.15054249835109101,0.06089009007708246,0.5464047599973355,
 0.19004744230448856,0.5193431839083146,0.21654691714701668,
 0.14502455948764692,0.2415788811095929,0.15248290781730778,
 0.14344401756065916,0.41083107315198497,0.8052388928278267,
 0.2848721251646946,0.4796233857624187,0.1177132168964432,0.16393640338836757,
 0.9882070418293412,0.6405145455975212,0.7993374471119967,
 0.0018656485564848557,0.3077306067127137,0.3271611723765554,
 0.8518440403711891,0.9646455815483594,0.7785094803924753,0.29115298644046816,
 0.7116183133919152,0.1538770942306732,0.06550436257420267,
 0.05504234428916577,0.7998917113531167,0.493578114628933,0.5520490087528785,
 0.5365716196754137,0.8836242862746241,0.8850165333402801,0.5011448684378746,
 0.7975578130850983,0.7543704250082897,0.6557696690001246,0.8491905610710079,
 0.2210217297218603,0.27063491137128237,0.3513830267779007,
 0.07984529573959542,0.6130167554825678,0.37880696094955,0.46621032392391726,
 0.5296843775441381,0.8616864711699872,0.9393403630925048,0.20336941014900578,
 0.32261068807179083,0.98551285806434,0.28103694414384317,0.5657593048480442,
 0.528774367421295,0.711073172477389,0.1518625387054554,0.1800528402820809,
 0.4045242459990406,0.9887717495746176,0.5308537908154514,0.0255457529144133,
 0.9169684540327323,0.3187227188225348,0.39982457576365316,
 0.12620106237289552,0.9352235840106199,0.07009114579810027,
 0.8636371476238415,0.9806870113568661,0.06329394722087167,0.5573217694513022,
 0.3424571877609379,0.14940215574623494,0.013236647139546531,
 0.02583986263090221,0.44140130211956685,0.9427404336724081,
 0.9108664655445429,0.7938917890656667,0.41984347427773805,
 0.01890484050249941,0.6755992666068679,0.6750896329033219,0.6635889655641216,
 0.5391096520401554,0.1628837067807416,0.33984653783437513,0.6177078582254438,
 0.36070385695804386,0.4689869867168772,0.6390732236540533,
 0.18928886880107232,0.06129921725699172,0.9037464516393974,
 0.32261532218694144,0.9245793330711469,0.4463921693594476,0.081488370430018,
 0.17930579113098966,0.4589009288255387,0.1630210127925089,0.6386932097961029,
 0.03870687424901109,0.009411172636686116,0.29258655773536635,
 0.016973132225977072,0.6904843781556416,0.05491377922188734,
 0.8395108384236194,0.026889645083863267,0.21739275712936657,
 0.6180567985393848,0.10543003422335895,0.37698278860786616,
 0.5455650402346208,0.15636215786869456,0.018496931082034762,
 0.6506521929653687,0.2947779309208627,0.8406544094680516,0.8117242635381794,
 0.22991961027778118,0.7701720725397047,0.5424754266437717,0.6985974479142044,
 0.486705086589315,0.20150265912338916,0.5426102729556688,0.878485482671781,
 0.2893204102672301,0.16532121366403163,0.3497444842046188,0.7152213859147166,
 0.3957497568742596,0.1497288601909489,0.6969844613007675,0.3241367560697065,
 0.7566810814089361,0.4562071022904419,0.7265930918879479,0.6947333236187604,
 0.8716654410332652,0.5300400461570417,0.7159567095513066,0.3757323239396282,
 0.4214611335875058,0.41699300249907867,0.370505557190686,0.8650585043384332,
 0.28891289242164675,0.28953056789980824,0.2645512371398637,
 0.3118310257680734,0.8701837845705134,0.9676183532645667,0.6635833673498952,
 0.9417021401834642,0.6405839276224203,0.7733858283126889,0.5361172688836467,
 0.9324132824912135,0.15129393412630165,0.1408746664481494,0.7971303525775707,
 0.28226996111516534,0.7690842344414432,0.1572365693426394,0.8495164768712917,
 0.717057683938667,0.8748268891476182,0.20259539152268136,0.6619063408043634,
 0.5633770907591987,0.6895188811127586,0.5874603170628536,0.6583005298408229,
 0.182994481653165,0.8142325025914579,0.6422104856478894,0.44955305777043597,
 0.06355123276614527,0.07022879045483699,0.5430080746441434,
 0.02147128151019584,0.15227713620270478,0.09528230533691628,
 0.9686670439697278,0.8500316615982709,0.22508840769864435,0.8087296088156224,
 0.5532097580853621,0.5854038435216451,0.39762115345572213,0.3800462025102641,
 0.9930604534407418,0.8655695093183282,0.7557072485382086,0.15910891327246535,
 0.524620873436981,0.9225205385163922,0.7298941745856954,0.6430303226830378,
 0.6285338751483532,0.3154375678284894,0.01255680938460213,
 0.17640283173524107,0.7752254939487246,0.5666501498407834,0.8600410209893877,
 0.6326049433681853,0.4997766921957627,0.23618634799840144,
 0.32490615141395385,0.6579828054821888,0.07983005371155416,
 0.8137968310760315,0.6015167952518918,0.6242367990655084,0.7806084542789697,
 0.4839621110427842,0.786974733599851,0.8465259973393755,0.6177732648746737,
 0.5221197057589514,0.7867725593012309,0.7691638607269384,0.5639496846994622,
 0.988440002205772,0.8746100259194035,0.8311742805758863,0.57199243881345,
 0.46985482515171195,0.7208442935037062,0.16998743463775,0.23342480283963174,
 0.42474184075992716,0.32672032101868353,0.9313373181250734,
 0.15934825568926747,0.08512021123535762,0.33829188556377,0.5092124444955384,
 0.3119959839135549,0.5082696423830422,0.06958546185784409,
 0.01841389472105348,0.8887258285384517,0.597357777334602,0.34403753474919463,
 0.8069740922493288,0.8238937334261126,0.882125752186355,0.3720557909627553,
 0.223163627586628,0.27757186620887886,0.7408765715873002,0.15788576408195576,
 0.11892937440120654,0.8925390647149205,0.15325525221665104,
 0.5517424480477339,0.5246231211478263,0.2737224683212257,0.34696478574050293,
 0.11341838022265127,0.8126533031345484,0.7273081851136618,0.665898882582116,
 0.8958785070093385,0.9770972263790895,0.31810310214290405,0.942271351349645,
 0.40779045791987967,0.8988937838231152,0.35013052235125963,
 0.8416668154271768,0.005866736923434979,0.21973117334724424,
 0.9511963808142149,0.4737960797258486,0.7445239115549932,0.4524519675370071,
 0.8865890959805722,0.5637740034428737,0.1016923486606891,0.921157165288393,
 0.32724708124640367,0.9172505111021138,0.8931400352780527,0.8144657817576965,
 0.20397397096296077,0.020114103110638326,0.6284596022729269,
 0.9163997835122856,0.489219626009729,0.11710537946998423,0.1836210828567132,
 0.33047576521023025,0.9159563527066646,0.4055841160721203,0.773663839883908,
 0.8621800691537368,0.269897797262846,0.34649944708013714,0.22208741613989413,
 0.012323760585991494,0.9032468876854749,0.12058134376874918,
 0.6917717192737234,0.6984194743383333,0.45493732129329834,
 0.12883568501978837,0.16266940546567388,0.6323547830729868,
 0.5928670512727985,0.5419150250416273,0.7344808353805476,0.3150282766437976,
 0.9251357981660764,0.9075562167912731,0.6967136192420794,0.48777026160654324,
 0.03283537623231525,0.8526735610427267,0.8575840683173654,0.9834893636647222,
 0.9315153118565,0.46057747283109973,0.6641665199650949,0.36957795412793937,
 0.8059700620443456,0.13421957044607347,0.623812831807686,0.479467055706545,
 0.14311265354777902,0.6687176456616958,0.0941682351061226,
 0.04169497715133119,0.001294664857854122,0.0061888147480038125,
 0.13505160695221263,0.6634933713850866,0.04956350561950873,
 0.12633495046065202,0.29516724624055035,0.29928359917935543,
 0.2774356033235855,0.8944193378940314,0.7723301809986209,0.23424936837261645,
 0.8122986489884146,0.16502867247040198,0.7602469611991928,0.9812961255994314,
 0.6017950653459543,0.6357357118916931,0.21647389593555677,0.4225171629850262,
 0.675191302167722,0.21784750865430558,0.9028944795015431,0.7991697174111609,
 0.05875625179749733,0.8660592728281602,0.12733229648191424,0.867179961263742,
 0.2908033716130376,0.1441298258365259,0.03389465125932689,0.537000252886613,
 0.09151550115528351,0.6419971310763499,0.6358018371123023,0.6592857576078905,
 0.7622780112004434,0.6701597596364217,0.7935875431127539,0.19798802902642365,
 0.3993213255629672,0.729274670199923,0.6668848123638809,0.21685517199064142,
 0.6758408822014336,0.7933582814358026,0.17197471922855345,
 0.38087629941917367,0.07907566397834365,0.6127578869439939,
 0.31878277847233305,0.4802591521404638,0.4741356227763871,0.5329835297557964,
 0.8743108488764717,0.4260854259871293,0.4801662189227017,0.5315914566700222,
 0.9156905624923775,0.4188701754158335,0.9571460924983445,0.2366974041662847,
 0.7113997405730464,0.6424570424696767,0.9644659715037137,0.5584469888751028,
 0.6938732696340797,0.8972451976740556,0.9090315599130744,0.69644194966071,
 0.5148681461034993,0.23509110779191966,0.31460722522905393,0.80374127164507,
 0.800410218225094,0.9193586101971465,0.6849690590994657,0.09506601086466371,
 0.22261856794017154,0.16857583153932587,0.2944843054947329,
 0.43787110187004286,0.956021025965731,0.42963668381862674,0.3966430967751915,
 0.9291727862333681,0.637834108008072,0.2804283598190158,0.7411076286806699,
 0.38785352880578694,0.514520385535102,0.8763626242004888,0.8571327674522984,
 0.5787888228450423,0.6111021356231,0.838964353040897,0.2679763099682062,
 0.15762351834817423,0.5969824587023388,0.28286227243114903,
 0.3981070840927091,0.39774502446519544,0.032993947473715424,
 0.5663404791134472,0.003563370349037309,0.41240480951292846,
 0.9582952087163804,0.901362586221623,0.4669896319347824,0.397082567405023,
 0.30788587835055936,0.3765729538731104,0.8559607533742167,
 0.037937613190970665,0.77569035620206,0.49574673742281283,
 0.04580004238397817,0.21384770065468794,0.9238227814029157,
 0.3495200077731996,0.7983009282395047,0.16688469438012332,0.8769600032564416,
 0.4453686363771916,0.776718730956824,0.648941332166823,0.8527036966618121,
 0.849132434280816,0.22141620891287261,0.18841529197859186,0.8474804385397479,
 0.4117197977840765,0.8621903049108552,0.4345998253128801,0.37927047121587687,
 0.5798818132304995,0.5055857513088995,0.45121934517772366,0.788238465665265,
 0.5944437397515971,0.8841943563108524,0.25501924264553777,
 0.17912306245941445,0.7323556024728756,0.49698349516636764,
 0.9702332202969209,0.7067401617429991,0.973112851360476,0.6152301644980688,
 0.9364590218276855,0.7325736032018566,0.3488223744021257,0.21274735197034145,
 0.337557173534062,0.5979117979256096,0.6003649434581377,0.0530475009260718,
 0.2834675949614238,0.46042214632252354,0.6464008225267807,0.5058452831747697,
 0.5182796627935136,0.23180210294608738,0.19751280540897764,
 0.24157291563725058,0.3200866415179614,0.8649508807486581,
 0.05677699137290426,0.27339512919478093,0.47810219649473995,
 0.02059241321460359,0.964464289178938,0.61875289710176,0.26729968793954473,
 0.10589713505710618,0.13073384935003973,0.24207059894971394,
 0.17106229306175857,0.8992390759826803,0.8632998230828939,0.6619266607393626,
 0.34427875225739646,0.7342903779389743,0.7599684656221734,0.265945958156126,
 0.08377169120433237,0.3991285147750555,0.40455543699985175,
 0.0790583739740045,0.5820377656623972,0.2011426965776829,0.9953817976098583,
 0.71465913064562,0.8146666964718774,0.4419133547056894,0.896252640083965,
 0.11000889361351196,0.832095986058919,0.26169662550900563,
 0.35317068941340113,0.04488222438915923,0.8249534444885676,
 0.35997319469789824,0.37264394058098893,0.6431824931621242,0.970297435557684,
 0.8362586844248826,0.6111892928134788,0.09365954658027598,
 0.46964822304598663,0.7264457643862545,0.7935015291214528,0.8932616483318083,
 0.5681811983003078,0.9479467997816592,0.312377082339406,0.4891863669196499,
 0.6866001547799646,0.5453760511342599,0.6142384494702794,
 0.031012622072375207,0.47371879942348305,0.47513874119903343,
 0.3526819456439483,0.8162290632827759,0.23163700365297568,0.5936804485958758,
 0.06281075906141986,0.943529345484973,0.5249149623955223,0.6983527520162823,
 0.07132870440481454,0.8819519877215185,0.16562552128606145,
 0.7272436646323638,0.28938190557568133,0.593157079208118,0.1475830784485802,
 0.874134059457486,0.39743769157684083,0.9184471672018947,0.6090978504831539,
 0.25735387434846924,0.03737305903631716,0.2131264697608617,
 0.05181274781864609,0.36290151529662373,0.35409813508751564,
 0.45680763784262235,0.4851568732949496,0.9720152458647032,
 0.23913443873548246,0.42373812518244025,0.7869671788809024,
 0.2651218602352282,0.013102353766421437,0.7554355323530616,
 0.9389310197760107,0.4191823969470695,0.05402757801016933,0.7961693547503133,
 0.03977317045947604,0.06013228216946276,0.04193836159521047,
 0.7105947680332095,0.386046242958359,0.4422268265608572,0.4773291833564278,
 0.32084169792686257,0.2604478873887589,0.022187306791645378,
 0.5261632401361486,0.10831362198353989,0.7725858698371612,
 0.21911815411324176,0.9330242996511912,0.15754926031446637,
 0.030339016193522195,0.5120787231796771,0.7576995362904244,
 0.4618276840259712,0.14666930097314534,0.2642143923489124,
 0.035052727681310514,0.4983232808781741,0.8856456080563423,
 0.9922945317983155,0.837028962288604,0.6021155763480824,0.9566384298401538,
 0.23541625075612616,0.20186734811890739,0.10491849194955982,
 0.7307840400681106,0.31962175492180867,0.5789245661539817,
 0.18740720141468237,0.025409508629233968,0.05495588986111244,
 0.12841762682498659,0.5182686871409141,0.8526697380358547,
 0.023239431993482107,0.4171784891304514,0.3085072311479118,
 0.2142439567380685,0.8282482789684753,0.16181374938264392,
 0.47439546557452084,0.10231466791524724,0.30241120188653436,
 0.22957024030176998,0.2596731793399324,0.9120975083117261,
 0.13822191680206997,0.2781307400708113,0.09444524539578081,
 0.7602185248504445,0.2859620697400831,0.019033288402627746,0.462306555262204,
 0.0677215336195689,0.11098975807362443,0.6106425507908134,0.4825654119955015,
 0.6702965038785972,0.05375367193055747,0.6640962613023569,0.8028843249975217,
 0.5374664742659515,0.8524715137848449,0.3384195292036638,0.23103012714413762,
 0.8227266166103929,0.35550023818470244,0.5345705060634218,0.8108328878453601,
 0.9687286249070999,0.2805611168516453,0.38509166759301283,
 0.21176364074399923,0.9296994827684011,0.5356602954541145,0.7171161386312197,
 0.4158036741493679,0.5835278398553226,0.4931926597009907,0.06111596670420494,
 0.6416610140449287,0.5345571613878843,0.4766625082095871,0.18223559449002857,
 0.8864123481296563,0.7171089762822956,0.3098940590471324,0.1126095507280952,
 0.06703951252825213,0.770002172420311,0.5758604077719449,0.46319182308337936,
 0.42626980871089315,0.5029702791543169,0.9304207692341697,0.4393282927218565,
 0.08388656676636408,0.3857284818648796,0.16695912562076265,
 0.13171397588811684,0.45893678253142267,0.8725345156692174,
 0.7589914191803417,0.9611151411789621,0.11161202203303378,0.386740069794554,
 0.49193654305953094,0.32547063247825425,0.977134795752352,0.5970316489565078,
 0.4132810974267068,0.04814933835325386,0.0870215874926471,0.5621926316562108,
 0.16747684302451338,0.7604575551200883,0.4161153752870007,0.7763721320310767,
 0.3331638786210923,0.8902965610072879,0.8477633926208032,0.5223335123841834,
 0.5928568985611412,0.5425979201193009,0.9893861215557247,0.616030534222163,
 0.9934331634168958,0.9622992651601464,0.3904665338855299,0.4960563614183532,
 0.27669237290436577,0.9132032342583711,0.09841857066284299,
 0.05938536355983892,0.6170332874837152,0.11179505061596962,
 0.21652464428571327,0.30332125167484814,0.930006914637684,
 0.46504399925371387,0.30345159272105604,0.17012252107428694,
 0.943129599374357,0.40424775372621036,0.4753936996545841,0.8321231207548185,
 0.2715255431886284,0.0591617191273619,0.9584892639179545,0.48493599230961415,
 0.3617909419177697,0.0718114860078114,0.727817494812141,0.825794414990195,
 0.7243027761669897,0.3153291519505015,0.28204412079106644,
 0.25878339186911103,0.7742044468593742,0.5703389561796851,
 0.22366984827202507,0.8562147061355709,0.3002934130388951,0.1000530956061938,
 0.731615993279342,0.7452225647281625,0.475253916850775,0.999863219565368,
 0.53955521512652,0.10266690744959872,0.3638458220746795,0.9473085381041145,
 0.7282808566973198,0.5427519191968144,0.8193478374587143,0.9297886980283024,
 0.7554467773508071,0.027568107721233193,0.18001444154809254,
 0.8716275397510322,0.8785914081827852,0.868896508589474,0.5482615949578895,
 0.8866833703344569,0.4200983550021631,0.2532395739571368,0.8527512262951381,
 0.8671372760253161,0.5051221325707904,0.2823491128230743,0.2600903543599602,
 0.6778671093191816,0.011874136287143733,0.23252637394715403,
 0.10312083540741745,0.25837930157575006,0.8588779720847404,
 0.4574869656231767,0.933180167685029,0.8454181089993711,0.44597804265127594,
 0.43975546964917,0.8950950196609059,0.5722455456062002,0.6611917695929517,
 0.5945429337020015,0.30298560456619206,0.22406300803410784,
 0.46595573346024777,0.31873067004520117,0.14104707653362747,
 0.26668866391203194,0.4969687276428981,0.5686729732237608,
 0.08322597843469948,0.19108135416667804,0.4838020591778983,
 0.8378011419553153,0.11776550175697298,0.02494168721358392,
 0.7112147026529524,0.5036900173014385,0.2126805845315175,0.40978303421826345,
 0.4421254497115721,0.660713920793773,0.9025770077484715,0.03452319611295951,
 0.5158307408545981,0.5341553550998867,0.7653237501796968,0.874895477341854,
 0.13306436290524215,0.5916908251602445,0.9484535403039936,0.825152026489946,
 0.33261970211284997,0.8107562760412856,0.4390477249151773,0.8010130858246732,
 0.9713251608688239,0.2984617363581956,0.730651837484186,0.14063038123162264,
 0.049965384930468426,0.9078272549394066,0.5210276123307006,
 0.3911241235121904,0.2713786543668143,0.33669868054390917,0.40606329519787,
 0.9970448726307425,0.5765865777967473,0.612198566493817,0.09769080830327415,
 0.3306861716655415,0.9990559694225541,0.12410470907357851,0.1588783337781855,
 0.4106823411280268,0.7078412197605135,0.6625386216532101,0.730938875599289,
 0.10794084313757846,0.6822932833253005,0.7451381600672674,
 0.21982633013914565,0.7811124045774309,0.6793650597342616,0.6506533317080783,
 0.6277011379820046,0.09610343758781026,0.059145261022008455,
 0.18222430800417688,0.22506693752115292,0.8460175673171829,
 0.9170900772263029,0.46815697411888224,0.7370459661898473,
 0.37037496224877575,0.27411277881558194,0.688243853796918,
 0.06883266985602099,0.465993339903491,0.14003379510340408,0.9828635996570002,
 0.8240544632599213,0.11413339956518342,0.2905219466642245,
 0.19632923737211427,0.706679435804815,0.2320373381895986,0.8509472730366848,
 0.47959777875716947,0.5051549471935289,0.5562561078915443,
 0.37527474181402054,0.6959701279060551,0.41218843086937695,
 0.18249226674373042,0.0017887802297268385,0.12974392935785584,
 0.6442651833802965,0.93243290958807,0.5970929323483363,0.09815864307051847,
 0.6571831352456902,0.21105738582950906,0.283105819557843,0.3026155307854852,
 0.021384929489091764,0.6188258939553095,0.4431448594068442,
 0.7984597299051382,0.03721309343623369,0.25111402911936076,
 0.2964919054889752,0.15912422692041817,0.37738564703061805,
 0.5009867891026039,0.2516143690765321,0.7581635292795246,0.12971633127212379,
 0.36511384047165774,0.4397499347646022,0.9986470085681005,0.3146533637102663,
 0.9305712950274888,0.5326849163964344,0.5280863505663663,0.9919990079647659,
 0.9313248870119808,0.473419726324246,0.25654426378593365,0.29720921179775894,
 0.7005236525883826,0.07280935247046261,0.6843202546281326,
 0.49534043246153925,0.5828637289825044,0.6798897739299925,
 0.004636914400714454,0.7421435676725707,0.6223283769018881,
 0.07110386488424725,0.263665586460489,0.9435330974647481,0.6609625444223263,
 0.4438992954339869,0.3643664438882688,0.9101752687117398,0.5126570614697868,
 0.14510114041647226,0.9096689064070554,0.7093846220254263,0.5718508538435594,
 0.5950039530285638,0.5216609716762313,0.024191517679714614,
 0.9850619971374496,0.3974859240179981,0.997438560938571,0.025213477529208328,
 0.13601117029114984,0.8175185294442318,0.5907683400086939,0.6012979168711982,
 0.4740537761958593,0.4815648977782784,0.6236646051802095,0.1987600039394538,
 0.95810013653991,0.7985107487808855,0.2451293350069007,0.6554907498839965,
 0.5882459198238682,0.12739524550748071,0.7939938102086956,0.5320435700129031,
 0.2879548223582884,0.8922667156759753,0.6353005459391919,0.942468460585704,
 0.16481738550474634,0.7600171231131787,0.08632833859042088,
 0.46017201287092035,0.8626197121588262,0.132418420076982,0.23074209952307534,
 0.13861830407870535,0.8763274529194395,0.5643091492715799,
 0.14711240652759205,0.47670432394452533,0.3674890518623226,
 0.30821343288341974,0.5561737673502816,0.27155386651653335,
 0.19650102478190112,0.3833390483209844,0.6984347983880949,0.0525826398838638,
 0.14021236139864568,0.41736345502712946,0.5944269043302742,
 0.3100179012256822,0.2947610481303474,0.5803377331629245,0.37430674940017605,
 0.9824616772437287,0.5743856009275222,0.8148196296177214,0.7989318394871957,
 0.390177474646608,0.8438367200311869,0.8858238928601454,0.046047789019313945,
 0.6053285045332039,0.9501542264124486,0.9900840652983467,0.09053403087635359,
 0.30869823572038213,0.5824037698377065,0.22567573103586236,
 0.7019056491176121,0.8583913401081543,0.48247229178972884,0.9180722460544012,
 0.470061481146992,0.41649567851657787,0.35202167688875463,0.7350642983816252,
 0.9217565722637137,0.3599656981577699,0.9781203643253154,
 0.009182847580924802,0.04234880129382523,0.9853127091174015,
 0.539664024296546,0.4626784974807524,0.005621991415968974,0.5484088637671239,
 0.30598658411118385,0.9900770282893163,0.9599738704757291,0.9916589172926807,
 0.5228855599705599,0.41512058415583164,0.7512215773972317,0.8992387999158916,
 0.7377259182436394,0.35957290369574746,0.9190806049899989,0.3747915837718273,
 0.1176731925234058,0.6760109303987029,0.9463855208026799,0.1355677815929922,
 0.6926719113183659,0.0036018177054805145,0.8233403857669598,
 0.6244480364242028,0.6810081598145181,0.08484690075496593,0.7085699955381379,
 0.09416954713961045,0.780071073330942,0.7076841590324475,0.11629516854792032,
 0.5264683030924835,0.40297169677595746,0.9466867906556646,0.9918379682695275,
 0.8434741358125206,0.35296131726951674,0.9860576562664243,0.5576741479423344,
 0.8382521817336195,0.256048785634425,0.33052683771874536,
 0.040015538280228746,0.7066177810492388,0.16704012413168512,
 0.8046939363827779,0.1900590046548638,0.543785988717254,0.9084597762183629,
 0.48215162041292436,0.7288651811597447,0.3886895992239813,0.8192348018834859,
 0.569983074388254,0.302525017282944,0.6239800598098602,0.02422961425303849,
 0.3998949015149986,0.9384983097652101,0.6496612996470024,0.5473002328000414,
 0.2648233790623701,0.6717160264210369,0.6546302810136181,
 0.029797427538890142,0.8766756714480832,0.8869562631268785,
 0.8192297662394447,0.9059948208601412,0.39311295799007184,0.1615382680486983,
 0.6847900117861476,0.34607747455182025,0.6381010631821157,
 0.35682530553069003,0.3310771914191031,0.5161137530194975,0.3436683340998483,
 0.49105844079921757,0.3000075661491044,0.6457357755124816,
 0.38693331466284697,0.0513707595351206,0.09577938000799313,
 0.3978231636329006,0.9407428582902237,0.679430426174232,0.10797055482167772,
 0.07974482037804131,0.5175309964451481,0.7377964827096539,0.4262701214466422,
 0.03064954248042162,0.39970654452675836,0.45716266160013297,
 0.0626389321069013,0.6055356130990618,0.9193888159901962,0.04792981091832471,
 0.19446526413391763,0.6588163200816243,0.3035703325548713,0.7449791341110958,
 0.5501941978518521,0.32977099531270415,0.7193352495525371,0.9815963671352973,
 0.8599142413234223,0.45412760318012646,0.29990681782209616,0.810203256096904,
 0.9610141110503926,0.0703931495846346,0.6634018451683714,0.04361076421727095,
 0.6221596053581102,0.32802174218445845,0.41594787451541704,
 0.9808814977403256,0.330892211932011,0.2922648159932465,0.4585847395620052,
 0.12732120964407279,0.7494257724836911,0.900671077354671,
 0.042878588994750944,0.5341521092608418,0.9110138185884514,
 0.16728413793133368,0.5236956599422595,0.9440223425387969,
 0.17872150953718735,0.35321207754216843,0.08874590561978069,
 0.09030422503638746,0.39308081697150266,0.345815762598759,
 0.42687723821588186,0.9630279485442844,0.4519049222857551,0.7234930357985928,
 0.4925892015742357,0.55734684070819,0.8324469007901205,0.6972452193471272,
 0.6494984604868078,0.6225473181177686,0.7275696483482206,0.1732421236630176,
 0.39333956948412174,0.6337838271518679,0.9770047309456356,0.0864058705883517,
 0.9621049600223388,0.25052792263428136,0.4408295017591176,0.562326666150528,
 0.12152628946065835,0.132912154056384,0.11019669131536292,
 0.08713165119075916,0.9368095299962831,0.18067335755146485,
 0.9689342951951764,0.8811151891618736,0.4742190609622994,0.21474796699087684,
 0.6932688739512048,0.9820123610276814,0.6641072494138175,0.8949138232361634,
 0.9163076033095756,0.40171368043951516,0.9897364431434719,
 0.08903287492159051,0.7240946633511929,0.8071827841129998,0.416632989030981,
 0.4886873717427569,0.22660255229300663,0.07546133056828419,
 0.030024968261688656,0.622597422698205,0.7690641318723312,0.9993316147940271,
 0.21444822018510612,0.6608359361021552,0.24404695207261506,
 0.4143958264353631,0.16316142728212613,0.3033683054525014,0.5259836922464349,
 0.7461735683881142,0.5523134908588736,0.7955634354661951,0.33759004169150875,
 0.5534229601226593,0.23439029923328691,0.08270177396761713,
 0.9782748872223932,0.09599832248271323,0.5507422071861765,0.8965393649516272,
 0.39903247765825656,0.5823563575561351,0.4352009568481052,
 0.15467579629099426,0.19516202797318816,0.8421540279435052,
 0.6024484186462178,0.5256457751774198,0.8339050054619935,0.04664805338716205,
 0.8666552312451304,0.17506584107317458,0.18788942441322865,
 0.2583987131251748,0.873565916494649,0.33166599124555785,0.292617223116441,
 0.07659110940113112,0.3046348952419011,0.7109919148030079,0.6365783952546271,
 0.825309828962951,0.8889098652622349,0.4111176102718539,0.9720025828553758,
 0.023876383606163287,0.5084419308443457,0.4793789152761684,0.994009658653763,
 0.9076035388133712,0.14597806835287253,0.23838210160373274,
 0.38554848904774164,0.0889274143017289,0.8333177136722475,0.4288771800162584,
 0.9090843385148882,0.1386347480408242,0.1026216957082835,0.14278015132363808,
 0.3217399259160916,0.8256556029405432,0.1798634823342512,0.6056205685284739,
 0.8675274222085121,0.4669947289727864,0.6085227609269906,0.00305048222990445,
 0.7253699763637048,0.015311942040485249,0.12679724345390153,
 0.4038443927870454,0.9806384350411824,0.7051541446221085,0.366801348093758,
 0.4219449747899576,0.9046644194926288,0.9761354856042421,0.690865277838008,
 0.5350376941102801,0.7386144841503139,0.6931123010142669,0.4806879625765993,
 0.20244978289434556,0.8229767423306755,0.6133293195029915,0.9571205002408806,
 0.7038299748383784,0.43770512645004633,0.2535024604649937,0.5899561627183347,
 0.2773968149400834,0.4443014642492422,0.22205926630660477,
 0.15202570247614655,0.9187732276417693,0.15173784382961053,
 0.39521400769909576,0.9390661809487598,0.026566615523531678,
 0.7208205240046093,0.9381142615131561,0.19833642176632793,0.6568080297003829,
 0.11019640082242921,0.47211917837748674,0.5449510050916948,
 0.0036301217456874424,0.8664633290632119,0.9267307066587243,
 0.14720462565656434,0.5051421333850621,0.38664334472720485,0.774028203649126,
 0.7567202252150229,0.046104186365161226,0.03717756774463021,
 0.6218922405539826,0.5805438303323148,0.2705103906348314,0.540224581570671,
 0.3039926622951299,0.23870132464420168,0.5282188077391989,
 0.10845712552872933,0.7865920140323498,0.5824638119412369,0.9007643182969476,
 0.23498690997842708,0.08654772593615712,0.9150776140574254,
 0.7833522339823502,0.19375575827558955,0.07621251012715302,
 0.2668725702061434,0.10745246517755769,0.8110677729885968,0.5434551937658749,
 0.40095604892530723,0.16232177101964174,0.2561299827454766,
 0.30900127489128904,0.13193381091126455,0.30228243699856105,
 0.9764092505289013,0.4641769896903809,0.5210085471490452,0.835116204612296,
 0.28786948158315995,0.660787407734131,0.8395152727606839,0.8545251838158694,
 0.7575365470948315,0.8429173899135332,0.37332572948062337,0.8587520374797546,
 0.9584554885586835,0.6229702889258679,0.9575506439989063,0.46427451521746554,
 0.05345243755392082,0.1625431096563249,0.1525258112984209,0.9985087299336984,
 0.7545965949241272,0.6704424970841386,0.5483238494230606,0.31370793910640626,
 0.17518930489123974,0.06341046577935472,0.33637351762829215,
 0.3375410468926823,0.16397195411633836,0.8271347003900588,0.6128017739386739,
 0.8574333670648401,0.9081855548753694,0.7274730901020129,0.6567748564133175,
 0.13366195487893562,0.08424909554981641,0.0982072572383873,
 0.13207107982612265,0.5389081807298512,0.16882759655355262,
 0.45061234951870255,0.6796933281534917,0.5748526534557448,0.6163395328760071,
 0.1473235472072183,0.2621976078867353,0.1926207551849679,0.1356460319337618,
 0.20712947942927795,0.4514153352201149,0.177243757337453,0.8321680092413203,
 0.8595290640455828,0.9450264905069229,0.13660422207327005,0.9321332203888802,
 0.6961298587368158,0.1728871006222681,0.7366595795443489,0.23306949876444105,
 0.664138760575858,0.07789519159938652,0.1531214194600421,0.16191087744674604,
 0.3105365455663446,0.57358991474497,0.017042793449838833,0.8942915837091544,
 0.41003927664799045,0.8866605688253735,0.5998475845503805,0.7052621995258077,
 0.48484848224806787,0.23636618963546496,0.9555570643170096,
 0.38886424127275576,0.02566291701274004,0.12159480040916737,
 0.8976041049992838,0.39968076311649736,0.054529498770330864,
 0.0773732057901837,0.7270148101647584,0.5293059385017325,0.8770721755504438,
 0.6775348575177574,0.6290003169370667,0.180015740449746,0.40338189522499235,
 0.5274663746990733,0.8902004839080616,0.6728800305238458,0.21566392001273593,
 0.39686636641740414,0.7448791773786955,0.23255195675091067,
 0.2109607053125908,0.22256105397526582,0.6961390944844661,0.1665293537559761,
 0.09024627599275559,0.5699703367813521,0.8994959481683278,0.9382963221170849,
 0.2333747939495565,0.13222293748037006,0.7369386454580561,0.6741216463210667,
 0.8634069931013008,0.9229159242274251,0.8240147297484719,0.09538235218861146,
 0.594503870245734,0.5757141585188836,0.45238972089717056,0.05263455976443998,
 0.34349268155074986,0.20793094235220322,0.6347837315850934,
 0.17874580315342148,0.6862137628499524,0.43606926261236933,
 0.8673486260654638,0.796183744117752,0.3256433777016903,0.42708822100495225,
 0.24838769517682024,0.7400425364094221,0.44209515391068177,
 0.9962700865577779,0.9575323525833742,0.7325090020145891,0.43398874871800774,
 0.45300347488423465,0.3980793497267412,0.2804329862195031,
 0.48858866281575475,0.310594721279182,0.2536362606555326,0.04370159380090599,
 0.4830497009709661,0.8841483106037528,0.7479109491174919,0.8127158397870687,
 0.3764695108655911,0.8728175175078576,0.6300050471754147,0.9712032849211365,
 0.062121221290448614,0.651898160258695,0.07425147944545352,
 0.1113609342636609,0.8449922041307608,0.3340199000925864,0.5606442824586887,
 0.64325364218101,0.1860442728348355,0.5748095211210058,0.7853859226988789,
 0.1111130921626069,0.6448859894220784,0.3399607855175317,0.7827874502227905,
 0.007896975086576385,0.5041539325052268,0.2394956850009735,
 0.5095428666324071,0.050412323471067944,0.008939389345529514,
 0.3562661136052282,0.9976264473744612,0.9466060213957308,0.8808149280056576,
 0.5007045351530139,0.20337737897560926,0.5360938101301386,0.3015127490953312,
 0.8959368940387117,0.9509992371238466,0.4261280060938668,0.4562075674113206,
 0.0901164748294141,0.9493060689250729,0.9861728536982534,0.8092440888586573,
 0.7190880867126184,0.7025408042717889,0.32367115867696694,0.894408893226087,
 0.9152532293901886,0.3938247141701634,0.4306511604555877,0.7106848048001019,
 0.6184091887406594,0.6608219775051029,0.3535629291558754,0.11082810153755285,
 0.9147760583917379,0.8445719663320473,0.721832589467588,0.4144355844640323,
 0.9547933386143101,0.16661790427534273,0.16698714046325025,
 0.23386447669214783,0.8580056921137841,0.15328624353885023,
 0.7366724578833023,0.9146413883499301,0.12939529855525822,
 0.056109135267867094,0.030645192781417196,0.4337545489572774,
 0.8430852244720758,0.028907790587159976,0.41331588032744937,
 0.760032054762489,0.39167544520840547,0.3446082127615179,0.856693448344562,
 0.6292566644803947,0.2819384596814616,0.3485942928424932,0.8043202190536287,
 0.5208852640855186,0.35149354764818086,0.37010395512066385,
 0.20425308932157038,0.17926909626763454,0.5179613074692822,
 0.41415535599923703,0.4846283369692982,0.93211731138244,0.3098910122371662,
 0.5909101978782628,0.03556151368403537,0.12341777401188825,
 0.49045467810591026,0.19019737190707298,0.023850035558893423,
 0.4100942232286574,0.557824917062925,0.22100322044329945,0.1438957142200028,
 0.0559895362315298,0.49229691281685806,0.6156073296281944,0.4896678805933019,
 0.9118442384050143,0.5663726133481194,0.9136650596648793,0.23773051580053473,
 0.5762103919100529,0.7697944701454915,0.8251685224991238,0.61011843658629,
 0.1311877180942712,0.3311488633896953,0.17392915086879557,
 0.021392023728303733,0.3093121510422884,0.5284325838601951,
 0.8182139126186064,0.2411008796058719,0.17923634036362024,0.7696648973917619,
 0.6298284231629486,0.48144109841279925,0.6268282508788202,0.8976819137927855,
 0.3301464492618015,0.3615135287043473,0.533252130542191,0.4526655644406895,
 0.793043165721013,0.7079743646363353,0.36373235094723655,0.9466088286990271,
 0.15394015511371184,0.11200257743890718,0.9597590846965526,
 0.011655354563025133,0.9763233765685587,0.5192534380823779,
 0.7603387118124352,0.4999959432647738,0.3477054676698179,0.3864609298930759,
 0.2772922407722366,0.02603650643630051,0.9261668358757663,0.7389404462560434,
 0.5192697360374546,0.25617410496730264,0.8138838636971208,
 0.40759926078428477,0.3742054704103459,0.09081514760852627,
 0.8728786105687383,0.5661321826985064,0.43662367096428756,0.3889324614815094,
 0.1483208346274708,0.7650997980440429,0.4361389143832858,0.7248286310924629,
 0.9858001178334864,0.70266724768487,0.09500415750233804,0.9343875448535042,
 0.07397550517468066,0.9562060589476449,0.987320543939301,0.950813545348801,
 0.24993431356761142,0.35183309754340386,0.8988167593733579,
 0.6510324622781131,0.6688366297754763,0.7828820264842835,0.28187915023154453,
 0.5535412730591865,0.7996841981347336,0.8364912264137135,0.8130706935205921,
 0.8126020442721811,0.2750686707595955,0.8480173666657194,0.8988766360541994,
 0.4324914814717995,0.3572138161716757,0.717803156224244,0.3414363841143355,
 0.8068625222449752,0.1870384774232582,0.11005132953810293,0.706045343379071,
 0.051053883040978465,0.49903065727721296,0.5809927642165628,
 0.6145177173241945,0.3011333502140633,0.5847619028427898,0.9336657789752623,
 0.13599303875515611,0.641832662026962,0.6083468874043263,0.7514796181916192,
 0.5541164218335916,0.3442564292565674,0.5580499661386504,0.9973236681472731,
 0.6067709883842449,0.9314930298711533,0.793834760547226,0.8227200194290655,
 0.21145862221465317,0.9837852244887609,0.0034476170934091277,
 0.37250255833520274,0.12336002667490864,0.27877959125325125,
 0.9503302644949049,0.8890959447364569,0.3958092625258216,0.11100814643236179,
 0.2931019931316462,0.8285697618816475,0.4566726846829561,0.6457757239860807,
 0.02121979675905694,0.9667895943618897,0.6379194914485998,
 0.22682856952625574,0.566264475206146,0.9903882800438724,0.66293207468984,
 0.9965215077270378,0.914325604583192,0.8153022100852332,0.40649637004444816,
 0.22785937452170402,0.9320369272688164,0.1540552169279097,0.8325397271381529,
 0.5308797189341237,0.6059975284812474,0.6397336161541567,0.6124990379290378,
 0.5256238335665128,0.7408926098852753,0.4631564325389752,0.11907071752006138,
 0.2155886476232779,0.9217408657514862,0.06588457816958226,
 0.09434923557691421,0.033398315058262074,0.9528875416771547,
 0.5152906837276083,0.21398111447123014,0.4604039824494732,0.8496623775068084,
 0.8396054827509587,0.8774249809349864,0.5662612167781564,0.6750063032524026,
 0.25363489950204854,0.04557140628090839,0.8850418292604199,
 0.1538940610318047,0.08414585060463065,0.0776081065656894,
 0.40099274119303197,0.7853109435891731,0.6647593814352019,
 0.006197059239040348,0.5077941888409899,0.1810722240880127,
 0.6528262773249406,0.4254003379910687,0.30387605924001526,0.5205749999703055,
 0.9400569463327396,0.2800834682933364,0.007043378682889623,
 0.45630466845564266,0.3382275332883822,0.9850414200287798,0.2391445122821404,
 0.6851845587386847,0.15805739475303093,0.7365491125212512,0.94209859226875,
 0.6770759063743405,0.14880982391156727,0.7057863893616195,
 0.41036186160510246,0.838548041049886,0.5318766952435731,0.9110100267628097,
 0.8101011614885539,0.3142140734600378,0.7292233515900096,0.9735788797462525,
 0.7316366355541382,0.8329863557178643,0.5297865520037045,0.05914567396022574,
 0.5184497439630139,0.16947623086829688,0.3803783486336779,
 0.22282373666232647,0.9776009686302429,0.35727473462006487,
 0.5209212933836208,0.3863684804032502,0.15991962781648217,0.9760858740390592,
 0.1190090039184043,0.5476099090425637,0.15807378192216082,
 0.01328192973950948,0.3869280512639235,0.8970414107325735,
 0.47570134789364804,0.9044754450387517,0.18655589907912073,
 0.6871041070924728,0.19666494161689452,0.4433964124214831,
 0.39197974819772363,0.08436724111382965,0.4901066920545011,
 0.5109951511805761,0.2503111529939348,0.5059583973789276,0.11338611132078313,
 0.9010796226075984,0.12343900620179937,0.6697594478810974,
 0.025830488787698558,0.6179803129270266,0.24969572025220477,
 0.2792978961150545,0.08405349357253089,0.9368227453345005,
 0.08088283985367406,0.7707641825701719,0.29007556992648165,
 0.6945138793428615,0.7325905583517536,0.556834820773334,0.3752590868038955,
 0.002164218771741666,0.2674389440930728,0.40528096699971816,
 0.3785400604561273,0.8637745990193726,0.5654239853011183,0.7333282637249524,
 0.3082551657828817,0.8997760407459926,0.7441339404812362,0.3210571224525596,
 0.30213344912780893,0.5123390677705217,0.4676845000912978,
 0.44976255032337553,0.605116014469404,0.832096945938872,0.0313469406024256,
 0.37456599410102953,0.5241686377019978,0.8125054528637194,
 0.35093665616928127,0.48451926047742655,0.6983672550187395,
 0.9161161963775211,0.17411569439584085,0.5537956187991318,0.7485162938409776,
 0.5623678109229169,0.8054146609808368,0.003773781008827859,
 0.06495268693768641,0.724175238797019,0.9294890681702378,0.03479519047098445,
 0.5326398370778549,0.18376195665375272,0.9768790763287989,
 0.22426066896445396,0.5459086765167227,0.504143414386194,0.616579940882104,
 0.011210645279401299,0.252641431301891,0.9657058590451137,0.2896200238422699,
 0.8384705803861726,0.16224301889555526,0.6879440772517825,0.9208387220075757,
 0.35176312822179223,0.6573932603973165,0.7237068819019313,
 0.32500075792515537,0.42138760743481407,0.6467857797274184,
 0.02635615404662872,0.4576301496015591,0.6667576972884646,0.6683743933910901,
 0.31257576439391743,0.22257090185714912,0.26296474717973395,
 0.8595358556228252,0.6341992119530386,0.3082318086393083,0.9112249909194646,
 0.5874203235636569,0.7476890585532593,0.8431075624536974,0.5646617634480096,
 0.6999575944781377,0.6236414154111833,0.27816531480401063,0.7339230226778979,
 0.15202280558156256,0.43061815155548,0.2728912665252543,0.1542688203881588,
 0.7069907113400815,0.9159397728887727,0.47413595274032294,
 0.30049215105182703,0.26850498163143566,0.8465526472685748,0.730252145176482,
 0.1914826549563422,0.38938541206395083,0.020037021538167665,
 0.9604869869727709,0.9151563629929796,0.3536141629052716,0.5415322239220148,
 0.034500305629153405,0.8664127171288377,0.7036033976070327,
 0.11775936520815145,0.9033139818925466,0.7486248300653043,
 0.07417694572313871,0.7902298522470936,0.1988659046418857,0.9690095279530562,
 0.11959351905703353,0.578088738179491,0.8637915896171481,0.5617250220139819,
 0.05124885882337382,0.017244080241156978,0.4705108680710668,
 0.7592351925051496,0.5869789213425495,0.5756733484683573,0.2913994912015294,
 0.2829381740485951,0.7235218966022172,0.4749617596980862,0.08719501534358043,
 0.8381890011296422,0.47398004896819657,0.30873308267336763,
 0.6742576399193486,0.4694673518038097,0.6038219260920191,0.44180337093720423,
 0.02915915939188407,0.7910896138170799,0.2569261924666798,0.677058619871938,
 0.6985197166363981,0.7526222026069279,0.4482450285965145,0.9275035643145678,
 0.920897281297945,0.25033372783124586,0.9667075543626944,0.13157562744575635,
 0.7863859595896385,0.9616032533787198,0.027147748513643544,
 0.1477681188372404,0.3766035184384433,0.9585590063948453,0.9644803572824298,
 0.29796952767608753,0.07299259415383175,0.924556244488531,
 0.12433891782662942,0.9199423360958404,0.7811891235138289,0.9555572480483201,
 0.29534330235143746,0.12113546598474767,0.061162515753793834,
 0.9201208196023368,0.598051446802927,0.27953225709191587,0.1912271430241521,
 0.5227820109185236,0.5125777398285477,0.9088152224016236,0.24878361830386764,
 0.7887672678909997,0.5247665876290939,0.3459117906070527,0.9942410231712936,
 0.9009878692659462,0.7522174846368271,0.8408000419299578,0.4806443730833414,
 0.535800392828763,0.9111686478276362,0.2602020477400132,0.7208495141334461,
 0.7122091254586407,0.47260941145274993,0.7433623183187642,0.5768525048608611,
 0.652111951349662,0.8698600373890887,0.7441673835475099,0.9932585822469299,
 0.3789107845167976,0.9698519251551828,0.38716093928175865,0.6838848481553369,
 0.7465633429052819,0.6718784479601813,0.11399564175244703,0.6681851398651476,
 0.7766625103569249,0.38755306601023665,0.12399354345232672,
 0.6039121422990386,0.16706904917247634,0.008482005141321136,
 0.6160329419158314,0.5748267488818268,0.31329040988750034,0.5169284032965971,
 0.8999122554591543,0.002426295410738799,0.00702156861219827,
 0.23098956232491052,0.17488104650583725,0.9305571113069554,
 0.008618115542432636,0.6250103799102541,0.648056017255098,0.6390907922901975,
 0.17982536206383926,0.7074511445575531,0.7391464785385591,0.5978897836040001,
 0.7484588568621369,0.5631080783407614,0.4025447145588661,0.9650732620901348,
 0.4009600736849135,0.712937786384249,0.07958817624255721,0.8071534437523104,
 0.40383250541195537,0.5026114757815421,0.9430134009542902,0.3472418824486263,
 0.34453697829490404,0.7462977264630356,0.5974861799108592,0.8068387699903896,
 0.8233335769140976,0.8263007625782894,0.45626995173569695,0.753946898600318,
 0.23850454698251256,0.6658650604834695,0.3367095319155463,0.6397935723620091,
 0.17986479271693312,0.029473243857690012,0.6236608298848663,
 0.6813751462587532,0.5456099976072911,0.965952684019823,0.5701598087618729,
 0.1152486955828439,0.9774171341632644,0.08826198251977535,0.3958180007164285,
 0.423723202536317,0.876052523824679,0.4572224259285249,0.8697380128034816,
 0.5256510896225514,0.33887008198275925,0.5113435104193474,0.5372592461054148,
 0.9126827015974844,0.0474378944076006,0.2914377980694626,0.949601749071675,
 0.3195317512123518,0.4985399865291288,0.1157572008252662,0.4220619500662145,
 0.40166122097197443,0.6841488320885017,0.8651166495286601,0.6513823793918042,
 0.2476103970167195,0.9813381194113643,0.9131757472993947,0.14545583852353994,
 0.38081998734629763,0.9378186795862926,0.3514505243022943,
 0.14738724756465116,0.13496487215872555,0.07342881419390612,
 0.5671327058237087,0.43880700611470813,0.11972590052901533,
 0.6097774080431687,0.3572517050690075,0.6864340941579112,0.4945877987795866,
 0.03912251700682057,0.015475652481152569,0.062874985252931,
 0.8781284682504723,0.6871400283743391,0.39630609716000353,0.8390299320395387,
 0.22462064508760027,0.13155199687323849,0.06339219290833709,
 0.3883411797983042,0.15101070870035338,0.7720857810334081,0.7320394629691387,
 0.14371198628289328,0.2783600163659665,0.5711232511262746,
 0.021466278775904457,0.37446946994608554,0.763533410693683,
 0.47805699682460123,0.086492906974734,0.9621333391832072,0.5644490428490777,
 0.1454364197657919,0.5991041671399658,0.2252246595238545,0.03687813322745004,
 0.3413181882968943,0.42193906186705865,0.23652629424439187,
 0.19055623500729046,0.9476722584461095,0.34632501128634574,
 0.4482150758720428,0.12355160454993708,0.2698992070214915,0.6431080102321372,
 0.2608569691298823,0.7016395680314814,0.425114640574911,0.21471976200396958,
 0.1195152479197944,0.45750109944057515,0.24133115785713144,
 0.6116091855979666,0.8042350797280964,0.16723388511773574,
 0.07638694428975279,0.5838786233000447,0.4810532783862468,0.7720380951747642,
 0.6486358765095531,0.99660492526391,0.5674489119407287,0.8806989114995742,
 0.6402507822123638,0.45937194331454356,0.31176723036314047,
 0.7949879583775286,0.1521596926006218,0.7003957209051811,0.15779552102406091,
 0.9704413723384285,0.6475614814706403,0.023607063615263213,
 0.11334517670983313,0.2259517406519952,0.6085338731510459,
 0.44054359576333135,0.8680548886506645,0.40893569876323466,
 0.6323845089166699,0.6148645445086054,0.28924854262938826,
 0.49569217828303247,0.04371819841832181,0.6162572685345334,
 0.33833588851497853,0.7082761162274948,0.5262067334724327,
 0.30655112156799424,0.7189176396944228,0.9240625719305966,0.3113830860627035,
 0.06711610874458696,0.7774329798499584,0.036595073513019494,
 0.8009989741923189,0.40262471835155955,0.093184885085426,0.08872090503419805,
 0.30898461665087995,0.9759392850373408,0.1355818649711371,0.389041247120664,
 0.567912190361701,0.4738382928173688,0.7438829600384286,0.4035087813303607,
 0.9923993297677822,0.441317446125326,0.30241938919377775,0.4112220927496273,
 0.36663784419564716,0.9612246035642309,0.5825661478934602,
 0.14004191655049802,0.4284581896560685,0.7299066937032453,
 0.21165666992844479,0.20769200166431734,0.2502482459415044,
 0.40365357343761055,0.6741520266778231,0.4409898748858049,0.9949476268602538,
 0.3441114253768951,0.7006388021686129,0.571434216669222,0.6623517259657735,
 0.37063569040018407,0.5380059769082041,0.02666166287055516,
 0.4038162910843741,0.18717425537245236,0.4982439152359114,
 0.03828031477540561,0.42275757572855754,0.12662487366301844,
 0.9575068070568376,0.9265955459084867,0.7699320052015234,0.49396441357444576,
 0.8372078112819571,0.8629676313216741,0.42788417949067936,0.2129291715293109,
 0.6598502352361814,0.9590241878657251,0.9194982505199825,0.6620992301140802,
 0.7449910466119789,0.7200345796980229,0.011309131517889925,
 0.33776421357470443,0.7485235171892977,0.5100780581758263,0.780098064528418,
 0.16346698908773516,0.057045519635835595,0.35150001686911647,
 0.9243740117782774,0.164688690590722,0.640405155606678,0.20032984916108487,
 0.7926965658323621,0.5858082417841304,0.6542742722992467,0.9417900722869575,
 0.13050977277304276,0.09668016208210317,0.7539361900846254,
 0.5860338797039273,0.6978609937957311,0.055324986295137046,
 0.9323414012758477,0.6007433161308036,0.8580292714453245,0.18853381621433007,
 0.07297401599899922,0.24322798666002643,0.22733984110697447,
 0.6839788246097244,0.9164991557792422,0.3985065839022499,0.99356886742171,
 0.8333836568019416,0.26990996769367825,0.9943112873598813,
 0.24408372866143768,0.6841541259805455,0.9080130715671323,0.3875183703462268,
 0.5783182955698791,0.8954941082731962,0.9280349110176751,0.543392291042621,
 0.22799155676395522,0.3217039912969335,0.7317630961558601,
 0.23772511276265884,0.706180598593902,0.9121288213622134,0.5058265414584064,
 0.4095245078970233,0.7192272592455747,0.30959659404074635,0.6520810733418723,
 0.5326561330616275,0.14331012625842876,0.7043236871627148,0.2676095976976227,
 0.6055774008788792,0.6467418102894065,0.8840979569178568,0.6372017944064047,
 0.0672735605176199,0.3312875323234805,0.015558802133710414,0.991491843262377,
 0.9729179427792544,0.2575085364942271,0.9502128230156985,0.8467254716237568,
 0.18961335641566102,0.2615255209943237,0.36891092790025315,
 0.17849228391649896,0.10480107663738147,0.5526902159003885,
 0.7036436933263506,0.20032856686024436,0.671326870317761,0.5610149495921092,
 0.5613287009238966,0.7773751915603765,0.22073165698203745,0.4593733878919326,
 0.9458974881297089,0.41673596578054584,0.354196476176342,0.7498790388523275,
 0.6769740028887947,0.13387433540569338,0.02229943110961008,
 0.07132444544779959,0.4773981984176069,0.08249251162089022,
 0.6655815566285578,0.543759576904725,0.13335696211226122,0.705014180235089,
 0.025063114149047938,0.9739837871639718,0.8768418596909523,
 0.3760193073923025,0.866867516772427,0.8977527244558927,0.0919659094432611,
 0.6529762590664046,0.4260118693070858,0.32277120028980333,0.8691975923949506,
 0.32616072484605885,0.9437137018853886,0.5809271039880136,0.5933807580351339,
 0.4111385704563688,0.9497113590478985,0.5943650305254121,0.611870914669008,
 0.44211735967125365,0.2819167859137831,0.8956483015312775,
 0.053794124521479136,0.06391729639194565,0.5703111334109718,
 0.3117495440701493,0.5432876432996313,0.669900851091172,0.5194304201840707,
 0.9528401671042189,0.003692570482419333,0.2644481026679032,
 0.7060547583386553,0.1429590141196082,0.7109888930899246,0.8791286686646094,
 0.8374767440138288,0.9335544120623075,0.9992152852932094,0.8976130238441934,
 0.7399640480618291,0.929799523841047,0.06835081404804566,0.8390334653266758,
 0.5343101841557545,0.6540269657465686,0.8688566692890929,0.03959493346220333,
 0.5151199570213405,0.6827090154438958,0.7418746870546967,0.2515223299263142,
 0.23209384247004972,0.9923190967983523,0.7941050137066888,
 0.10988606889672559,0.1620448803585508,0.52035613015031,0.4997355628568323,
 0.7595212894475392,0.8722124752732884,0.9841530806489158,0.39399022232847514,
 0.3542871538807565,0.4436376232524646,0.6591325591626527,0.7523302271835884,
 0.8489283139513383,0.6482531281790622,0.36500266305726536,
 0.33983076938777323,0.5415300004490595,0.29899198766569035,
 0.5469031553334951,0.6303792298376143,0.520984760703752,0.8545741586487011,
 0.07693853870501888,0.7054339918742207,0.19019223187318257,
 0.44097595934182543,0.631991836924599,0.3726785826274803,0.776188573631299,
 0.09478478835315918,0.0024735722502396973,0.0414573506464615,
 0.011091991829440762,0.8042987888352806,0.289075676847101,0.8534926926819488,
 0.5499704206886576,0.7728284216346968,0.0627926390476603,0.6827315466360662,
 0.34990480829068105,0.21081385474936998,0.6113806821099207,
 0.6437588181660152,0.7844314083062914,0.9358266277147971,0.677872215908222,
 0.5384668152565866,0.7501196261407513,0.7688283021238425,0.8406400056798122,
 0.16022796662017225,0.3157628871332029,0.9941009785676806,0.8740905260830897,
 0.21054317264077282,0.04414451902784644,0.509625096458072,
 0.44371019221970365,0.91865588826711,0.49450587404182256,0.6084423938115591,
 0.18398141636542398,0.16484862815217982,0.3143130162641079,
 0.18572141536967846,0.2046728043624686,0.17909759858313623,
 0.021552833471678357,0.3644563725929344,0.32560654953037727,
 0.05128283694391167,0.8680601431487757,0.025003777742892863,
 0.5662918152674007,0.2026429593504454,0.40901977499296194,0.5205533083769641,
 0.9595114350038243,0.024170354510379077,0.9301703712220178,
 0.5820666433561097,0.589343739080357,0.9056237338470746,0.9250895965211458,
 0.9451004134026617,0.5801486188244677,0.6092998640589525,0.23315428984138664,
 0.27417144422553874,0.2015816044403812,0.7092124982126398,
 0.05548297883610509,0.2634413752559719,0.024171539855824764,
 0.6382962497956145,0.9466138330224867,0.6599195856575524,0.526105487546567,
 0.1962738946499396,0.6515417684361476,0.5956936148481523,0.9359707538447932,
 0.564529736827931,0.21505533923284958,0.7856572791534354,0.8194087325116012,
 0.6296119218453625,0.7469052156191924,0.41286576450798396,0.212811308242162,
 0.41165194160775165,0.03924174491450505,0.7090695385920076,
 0.7223195022244918,0.5712555538473472,0.6512479548545993,0.32646839813393047,
 0.4940275648271091,0.4060791480890229,0.8179905102713072,0.4666984952322025,
 0.3442936633579714,0.9030612950756357,0.2386912606904824,0.27827617132046634,
 0.43973716476463276,0.8947846757540826,0.4210783104481396,0.9605167711060143,
 0.29466161387318845,0.09415686363661913,0.06765916340107336,
 0.2951342672321563,0.3870527058410874,0.9407335636087687,0.3578133058359573,
 0.7607052315866094,0.019184906722285455,0.02776023078006007,
 0.3058386486420316,0.23484367002269935,0.7325692501323549,0.9833804238507233,
 0.0711004833985549,0.07859423054450587,0.3184264602330118,
 0.08949183105285097,0.8273494641006756,0.9468165711840681,0.9614767773419374,
 0.4907270987303969,0.4950913459088552,0.07142336383075765,0.8145407818833303,
 0.7702069237399124,0.35451074057514376,0.6062174286893094,
 0.43880290096820207,0.11008870384241509,0.18928170368584873,
 0.8362937265103321,0.3061960752089583,0.6783373014022297,0.4260757380817065,
 0.6579463774676532,0.9567364430169574,0.6851326893575245,0.86686731183242,
 0.849512056530421,0.5449478363394179,0.35195779409286676,0.2507429269933399,
 0.22529229951215302,0.7221557480124116,0.714535774243554,0.7297372981262338,
 0.026693057769065387,0.7660672258993642,0.08852709826956306,
 0.40800679748930224,0.06524043989496486,0.6290806264592155,
 0.8485754397632435,0.514092851043479,0.5754474215803717,0.6200209569418534,
 0.390518334939583,0.9940379148305154,0.07611525252987217,0.6954382241444939,
 0.05473903116064294,0.3006394262150267,0.09358167908373716,
 0.5602188378616005,0.39714643077825584,0.8403704865844193,0.8929701932837648,
 0.8081002210463434,0.34314139114037223,0.5034398921415114,0.162477034818862,
 0.9308353493844987,0.05783261619354718,0.7616090646453758,0.7896313745119851,
 0.24241423963794795,0.4221392988875041,0.4165385703645663,
 0.11505860329701723,0.9083377180947838,0.25434025149982364,
 0.6487209212683861,0.7274201751846572,0.3405387518798202,0.5756684834361722,
 0.28346791764309776,0.4432442266645983,0.5304690400548312,
 0.15396566323844973,0.6531988158681717,0.3870095458122629,0.5050176525466242,
 0.1407955734318782,0.3678974985873896,0.23988425606100272,0.292059421837329,
 0.4451662316084445,0.03580919337250499,0.3743633969786614,0.6962725207170531,
 0.8970903178579301,0.214741440148694,0.7024733534059185,0.5293240304609225,
 0.28740055585473256,0.19166692141963027,0.22281709038275233,
 0.7800502706670105,0.5362012285286644,0.7794082679901722,0.37029183364879314,
 0.07800372349970097,0.31689306090900016,0.29357874109011817,
 0.7325384030839398,0.8663757019855172,0.2003910718793278,0.49095626026823747,
 0.7489052980032895,0.8233932346854849,0.053575818014190424,
 0.6014325594759103,0.17669202466478995,0.08694939598746143,
 0.4639924356362083,0.19621593176231322,0.3327873773596618,0.2540626290688366,
 0.890350423631609,0.46947308097007223,0.2760376375332837,0.784635512724745,
 0.06529253270802349,0.7329512295473464,0.855376251087858,
 0.011949323219349162,0.4880497862705858,0.11358247116870124,
 0.29776288587450084,0.4058364727897177,0.6579143814988522,
 0.05051888610962374,0.4000758524417162,0.7682723302928892,0.3742487054396122,
 0.7748420592725087,0.28205432392544805,0.2620135568614185,
 0.23172861759968544,0.39821288299121105,0.8547072167371385,
 0.8591320036278358,0.28847486885089857,0.946133161708138,0.2946890757030315,
 0.5570531085711549,0.6000975214113605,0.008662749682584403,
 0.17320202545739782,0.06570298316407164,0.4322402041681932,
 0.8474549601363478,0.3064581568988116,0.13147103390663906,0.8606791717300168,
 0.8625285053669014,0.678397760029316,0.7822081124168132,0.31257786702699075,
 0.8337019081589954,0.6440015549407702,0.5829065304977923,0.5114780236117894,
 0.050966606370650314,0.3256596069158585,0.48393940218230647,
 0.17425368722864054,0.6829735122161068,0.9080721050285955,0.9777911245949026,
 0.16910690665988348,0.5988555178545463,0.0762306248717477,
 0.14041524640101344,0.01690571069014113,0.20623656640145438,
 0.996290313711476,0.3149958189938915,0.2940354447358208,0.4998130128036612,
 0.5793877212840917,0.04780458083518091,0.509626088053546,0.9894960907447933,
 0.3631587934041097,0.6431583554794852,0.6711328057071891,0.42649400480154653,
 0.1349827572193999,0.6324676078038,0.9508188051704518,0.47022947867544795,
 0.5688212615249034,0.22369353699131678,0.3520679191329331,
 0.34928440006577666,0.5727881632262632,0.8535653003198,0.2576967992807425,
 0.7044649040489598,0.6673279671627863,0.6222490988236029,0.505082126160952,
 0.6969182335505675,0.3606713188356937,0.036281480834131075,
 0.4967099937137347,0.25846403442224153,0.5040757446601406,0.57748445219032,
 0.26203403445856566,0.7100364547452989,0.7726165602286624,
 0.32521279594500685,0.04454184313317655,0.1406708679355596,
 0.5245525836648063,0.7710949271799467,0.5195945828142824,0.6654082144038681,
 0.8354828033464679,0.5988922675594974,0.10188729198885893,0.9875379273397515,
 0.03706986817840796,0.4972400625089185,0.08949644250296274,
 0.5726970223763318,0.602139447873802,0.11107837257169362,0.42905088588602613,
 0.3149643237652162,0.03897322073514198,0.4825534701847741,
 0.49029269376952267,0.781439268988388,0.46143249381017915,0.4105416262216556,
 0.21704845798982675,0.35282466764536646,0.05987254182488755,
 0.17211818009219715,0.5399441611841842,0.9900033555287758,0.7007565825118917,
 0.9083919323967962,0.21327641388303975,0.946772737295627,0.26520703180112726,
 0.5383254041300061,0.36912178823044606,0.796771441687831,0.4352733899097396,
 0.2793541518400574,0.9984020262369956,0.7988495765121114,0.8237283861462456,
 0.7078862443330571,0.17012416121139384,0.2820377544039633,0.8699810115280759,
 0.06220049176269493,0.6578456015424741,0.07984330457401978,
 0.010332237582107773,0.41140736346450957,0.9027711361144686,
 0.44528171444818065,0.7471085563253366,0.6180800706189423,0.6595477085292931,
 0.6482912933285662,0.09791011943933503,0.7865731014679773,0.5011561267933771,
 0.09616372984612898,0.09950929033999345,0.4315975011295704,
 0.07177342294300493,0.523227185943873,0.24835931676138556,0.6180645342916834,
 0.012430996984010845,0.36763882604866926,0.8716591168280625,
 0.48328808468619755,0.4289010643051635,0.44032769576438113,
 0.0996790874914355,0.982188262945671,0.14060627847238882,0.7520454838133124,
 0.31707616687486884,0.8124669965514469,0.6227881550478443,
 0.14891048911654337,0.24706577348339787,0.25863516161629674,
 0.5552248952373384,0.6583039674110843,0.8213312061304575,0.4445424188693844,
 0.023215904921683794,0.8386967978729246,0.9635095148204185,
 0.5374158147256913,0.9605268938569399,0.7382842687199405,
 0.007446956125825577,0.509324403116719,0.4701093073232875,0.1035086827003513,
 0.24749953536171965,0.1070701565127542,0.2788403085126624,0.9812425582595677,
 0.3903993297716999,0.6209919536504537,0.5469411139107057,0.9892572906281905,
 0.8028813297606219,0.44088398949144336,0.22151158048746467,
 0.7589773245977922,0.9497075382527342,0.04895519759592537,
 0.28561323780185055,0.21806145322569026,0.17319475529922912,
 0.9824728987792277,0.5088433270317807,0.7676504568459381,0.560349856778113,
 0.5388515103403494,0.16294159917588846,0.05126045431560422,
 0.9517733980773863,0.5628604561295664,0.9973513939612971,0.2960865553285361,
 0.46050641334398223,0.49829217553106253,0.8422656869986214,0.619052791179522,
 0.5403776337913341,4.131128014308594e-4,0.841994347859804,0.9168907314904627,
 0.10231010391901618,0.22990269560755805,0.41114663209649405,
 0.19068393791992977,0.9365984512741309,0.5803593254812008,0.3095182843773028,
 0.952196100613483,0.6150620142019425,0.2755977653051781,0.5240424495261414,
 0.8737031408146475,0.017833651125944816,0.18939110349590416,
 0.8372392510136445,0.1443217720381016,0.9158459304880464,0.5944184617012571,
 0.3454774543070249,0.7579370582043028,0.5620657055701063,0.922400422756447,
 0.6412474081967646,0.6681681372882143,0.5737392744677228,0.18082824010955756,
 0.3607446711612363,0.0747204570153076,0.5976301485953274,0.5397450603085977,
 0.4551390094374136,0.6168894372420505,0.19843397252823747,0.2950146247567266,
 0.45826646891878053,0.22858475364331798,0.4149739487485383,
 0.8458222938351723,0.7783560285081536,0.9906497907408598,0.8154876689252556,
 0.12930325040553925,0.8268065209553201,0.3870100823568414,0.8536148421881602,
 0.11308211732577544,0.45440056164459,0.019814016678921464,
 0.46716671457129755,0.228459675449302,0.7820503939837973,0.05675973133436396,
 0.8883905567133513,0.35180815282215117,0.457207734514147,0.3494042745656867,
 0.8402664605421644,0.7929957410038906,0.9754593382913869,0.48822132924867956,
 0.09117764211829804,0.02400064204851393,0.015260519621957735,
 0.2947743141683594,0.04124673303857396,0.3426562100979056,
 0.34053966155986826,0.16909392199097506,0.4414791523555551,
 0.30676567237295616,0.02081070112498562,0.4627615866436501,
 0.9129563276319869,0.8004565056139736,0.4717564258055902,0.20243196778982342,
 0.2561840173413066,0.824811585585356,0.02957935098347364,0.05862679080040234,
 0.86226956040754,0.36278264455535914,0.5719829504663035,0.6432519141326415,
 0.06582848275896391,0.5428294381774226,0.7157566162457134,0.4781684998915816,
 0.045225534491503794,0.22249984516454246,0.07402064719313606,
 0.25665793177601104,0.7860201584927652,0.7236211342930947,0.7420144143401963,
 0.7301367634337625,0.05460124544031353,0.7741585022651125,
 0.38450094686857295,0.15750757250083636,0.9288016600865718,
 0.1828465947112442,0.2694791606366437,0.10751766850521727,
 0.08779761343645609,0.2968834821233547,0.3200517156081166,
 0.10429839558573095,0.67415242337168,0.6753047681574326,0.7613111391463627,
 0.5213408410175028,0.2331603787872386,0.4553706055383483,0.6210641378909583,
 0.7208031852582021,0.794951968863771,0.7723036957287277,0.033788947902369415,
 0.2712825004581553,0.2825127821679515,0.9588562042133303,0.4353455293550509,
 0.47847739954686586,0.2527511408468659,0.799648902938832,0.3943860435208748,
 0.2003171115780319,0.6219737874103615,0.8180559767192246,0.603111389063112,
 0.228717224265227,0.09570054067408751,0.7106843410043719,0.8072087316416972,
 0.43888139316950736,0.3661379276374347,0.9606763623017041,0.2745399410746464,
 0.7419990181374599,0.6544656513947075,0.7538322377609297,0.35817873060202254,
 0.6453214709245951,0.11782366274572442,0.16467914200428857,
 0.13455775044648266,0.6109248317145805,0.06908748490199179,
 0.19346743224714114,0.7757850337169874,0.45098204488054283,
 0.5688306822026395,0.8828064980111536,0.8448751778461188,0.9592737131759711,
 0.7021366009095427,0.03796903352993708,0.8029148056490122,0.7058521652231725,
 0.46773420216610806,0.6793164476773081,0.1851144218759191,
 0.24963624749843405,0.6625961345610029,0.0043930495335486874,
 0.22268369735806703,0.267429900015668,0.6908413271376815,0.8442578503427329,
 0.4613594686332304,0.792004228999509,0.5085855001062273,0.25317381049172427,
 0.5979121821808259,0.14532633030039277,0.1701107125020801,0.5111839329491027,
 0.5312588697803149,0.35223204765107385,0.2783467467016202,0.6445885511192578,
 0.7545460790030802,0.07808056243518613,0.13776030547141116,
 0.10750804782458123,0.42727354842395715,0.9336087145145835,
 0.06513633711621425,0.5162494819807518,0.23332518333882968,
 0.2412908826670812,0.19930362130590584,0.28303634352493545,
 0.1292177568010171,0.5407463564918411,0.3805413505565215,0.1314106177216029,
 0.5174714416242207,0.21269056253240481,0.7426332467214222,0.258051176817904,
 0.05415521470009499,0.38265003762144645,0.04169069796272162,
 0.5015397956612249,0.4634051994367452,0.6664885244654046,0.2638882049434762,
 0.7883688556584583,0.9314968602381258,0.8358636508019494,0.47243175498613366,
 0.37136972527258405,0.8226252063969587,0.7519973092686014,0.4175065544416805,
 0.5147018555414986,0.8899866317032683,0.9299251569664275,0.3492198018957785,
 0.3079826749921927,0.29963630179622114,0.7023963896276986,0.6963019721336066,
 0.5152138633231589,0.735713336511189,0.9232273991468607,0.568782514490794,
 0.9208475510934762,0.699909211281171,0.9296347570093851,0.8765015349950223,
 0.0608340098522,0.05000700407860226,0.04670266567305914,0.19898846416574212,
 0.2084546545546201,0.33146233867926156,0.22884581515312674,
 0.9603668505240408,0.5761941643758496,0.9812010389866975,0.5835678256486678,
 0.3372013779996331,0.7870075956520755,0.45260687861196847,0.6038828092135454,
 0.6917198051788922,0.9114181473626138,0.11734021121004123,0.9158998847398049,
 0.8623019190673069,0.7978388272769251,0.2795403161271903,0.3996479850017195,
 0.2507920743319122,0.44486973615485126,0.23748698983477456,
 0.36614220292473876,0.29588932192696427,0.9645416485744425,
 0.2967753691286418,0.9082172892404163,0.07556790186673024,0.9232140044010397,
 0.2162650511389096,0.7899962181609369,0.4912975562839179,0.9758524142468481,
 0.4083566743735201,0.6752344269513813,0.7291078492722151,0.09347537353810542,
 0.34196843926999476,0.28217944080074553,0.37756006086585936,
 0.2257481241008641,0.17525149226163095,0.6386069031285371,0.5897328998409256,
 0.9272397576825977,0.5023076219177602,0.9206894554604561,0.5841165713311647,
 0.23450448962644832,0.04725190320944239,0.14135352063992956,
 0.2514043815363567,0.4304675695098029,0.005513386881287996,
 0.8291128514091163,0.9146873748310363,0.21248809228210863,0.6258010469127943,
 0.28468607100087173,0.32747836196009095,0.5405357989689121,
 0.16513342762700867,0.6273460189734859,0.22978249665269623,
 0.06813594315054372,0.0694900482394486,0.06424739575755412,
 0.32646726949360905,0.07332875197991762,0.14513957253955023,
 0.29030807583045815,0.14393082670811286,0.9450633607592802,
 0.9332987197860441,0.8268119814907042,0.5478283015041013,0.610065293241992,
 0.7767526320789251,0.061526241517840585,0.48321738929824953,
 0.6876865149023345,0.10546793739125349,0.42409141355730773,
 0.07450844936491785,0.5828357419006823,0.3796832902325238,0.7198870697739232,
 0.9155394026942624,0.884319521894704,0.18278701907672112,0.3004021663587476,
 0.5255947195197793,0.9885721550325142,0.21577346806237896,
 0.40119209000812994,0.8684155982529301,0.6187661353375794,0.7805962938343072,
 0.6747386672866467,0.6915993557965273,0.8821866063670868,0.2968995186447402,
 0.8451159255896015,0.8479228224653333,0.035472603009780834,
 0.9466520899215116,0.9081890796452214,0.28368853264791005,0.8705750694164252,
 0.20028184640043678,0.6375788971115763,0.14460013937349758,
 0.6033086234537621,0.6805658122375867,0.6730137558601531,0.6645088354464777,
 0.2765616598764823,0.7876083801245821,0.6505758300077266,0.28651184685877396,
 0.14134172559320923,0.6254891308364614,0.7998166053934462,
 0.32825840565880193,0.07714339419238359,0.11304850215056506,
 0.2674591675425777,0.8328175061664403,0.15534428715096193,0.6386617702738598,
 0.1516397059444211,0.786619336105677,0.31552110348881923,0.8352146438017067,
 0.9915490760224716,0.3011804202243923,0.32342987747279306,
 0.19125849449013288,0.06476089412382446,0.7502313862084555,
 0.38717974766268015,0.8020199941989905,0.3117655031322417,0.6286611228033521,
 0.7730200529399854,0.9660543056165356,0.1017746075964645,0.4436545875876706,
 0.12042022450378265,0.8538435291185692,0.579000499445157,0.5597832924603768,
 0.27121744714006857,0.2718645296324218,0.052080623587163144,
 0.8791161314278623,0.8291984148620546,0.8103603239637215,0.5451621309094685,
 0.49191784884412326,0.6978690723240789,0.783106893390881,0.487641740725397,
 0.8731937058113975,0.42457671227241145,0.7070059225053359,
 0.12662555749679494,0.5909368950644501,0.2990130822055589,
 0.30277347430381163,0.005759187471643834,0.19904072145340246,
 0.2703376560037515,0.36368572032579727,0.4461269338564069,0.5141325736617387,
 0.6202027691841778,0.6289016345974141,0.9510031815785542,0.9191276150391972,
 0.0743868683369362,0.887870536825849,0.2519450263977502,0.20393604952802702,
 0.22125384563350214,0.22415005624162554,0.36875568359680333,
 0.500222633055517,0.9069126155770316,0.8929276381157293,0.5108687819531613,
 0.2068428764415694,0.13053697687483112,0.04587711756978785,
 0.8537695748921453,0.20034665062162438,0.15547426867613656,
 0.1974088936036421,0.30165118367811694,0.7790351496923449,0.5786286176013447,
 0.47586654095647296,0.18216280134757767,0.5895128179048451,
 0.13904764910687983,0.20899699856593945,0.8412712895589838,
 0.1182710863809392,0.4838590022947159,0.7316589529144115,0.3443108770243828,
 0.6347292237946573,0.10074614543216032,0.25137913891964014,0.883660835994254,
 0.6635488193827118,0.4344169887583075,0.18602616756318735,
 0.37860002464172704,0.5717935063126561,0.7910870426921202,0.512208256042842,
 0.957815652241511,0.38491003670098045,0.5793342830377696,0.803964782497377,
 0.7933316004556167,0.5580375104921935,0.4694897788170863,0.10832437577177667,
 0.45904076772205604,0.6731546735859333,0.018480849319734594,
 0.9792492942115326,0.4140622613910516,0.7628858728314578,
 0.015059376352558607,0.44266155251171424,0.9705925706805804,
 0.6214133916195141,0.567823721558552,0.7503758334360539,0.7288528437516418,
 0.3644672102941051,0.2899511856169299,0.14220755915243644,0.6477292970546324,
 0.5748793637400356,0.04119741559651435,0.9081630149484949,0.8468892405323415,
 0.5611291545347784,0.43619845853515704,0.5513932356547849,
 0.01205648467533671,0.5188809047370295,0.6966052864086993,
 0.47707378779157505,0.07094445140747785,0.7069713845990844,
 0.6381588677918688,0.6396518722265736,0.7607729894781029,0.9071242359531622,
 0.8598570472167735,0.8505198705393984,0.7987540750006838,0.9771343361339107,
 0.398850776148152,0.7639051788722607,0.7742065659051132,0.19387589463578925,
 0.40004390032349635,0.21047247972027083,0.30390686329694205,
 0.3184159180395043,0.24702332707757568,0.8178171511704154,0.8586277458282338,
 0.683548760142875,0.174579160607637,0.922208500969246,0.8939338651960418,
 0.2260680764499019,0.16826184620728413,0.8355143835832735,0.7643808116137835,
 0.8608245375626865,0.2851847178654432,0.5789665777362389,0.7757573049142364,
 0.9137093855217386,0.5155864641109966,0.7309989577485486,0.2638802416321774,
 0.21817374748344598,0.3279625106934396,0.6553633099012455,0.7594458320351514,
 0.7446982762156671,0.7529557078363933,0.1778806958446395,0.731522220656672,
 0.24070967782968034,0.4990980178005988,0.18529573892396212,
 0.41307990696368824,0.8786154677255891,0.6105664366801026,0.1504123263066811,
 0.7464701372450382,0.9490321521092847,0.5035490136636127,0.49308073829600907,
 0.5042619454685041,0.5761629166824287,0.6238666617103454,0.09576509398163469,
 0.8051520038083252,0.7013587808753488,0.5515899277716473,0.5636807592493854,
 0.22779880706266675,0.7335635685651684,0.027981851013961445,
 0.7917055336885268,0.786449896497292,0.9878642443326425,0.8507777354463132,
 0.6502304564491026,0.020370554608480806,0.42693875568755923,
 0.0373801519826531,0.6004085332928252,0.6139504221504418,0.8695550651463327,
 0.9588891455793755,0.9285691630691697,0.6281455174883264,0.22881067729598314,
 0.8807059881081486,0.25657838410278067,0.04497270884727578,
 0.8150718642482828,0.9919688342383224,0.359983865105135,0.7092651371522019,
 0.2510901872608815,0.40061845785268146,0.5213205892597804,0.277132852510635,
 0.44508633318033586,0.6199607065211182,0.7541547846633162,0.8131330669668447,
 0.7145485379479977,0.45378755760351375,0.9220199763672152,0.4735136001872883,
 0.9094271822979995,0.8886175449960102,0.5099392725669768,0.6728455103136434,
 0.6862883949777532,0.7840928155981669,0.29098885988937684,0.1757531705294354,
 0.8382976166528513,0.8281672874559795,0.11244214817690101,
 0.37067148997519983,0.03623133550238189,0.6483857715695469,
 0.6735426651823821,0.6074934519219585,0.8395328375436399,0.585072960346919,
 0.7936079189872168,0.4613620036732414,0.7968193939341426,0.6507751479734224,
 0.4800301025554925,0.10184304485622342,0.9477048844326167,
 0.32055673802003604,0.05089833345102923,0.45677832848278344,
 0.4552085312456431,0.8463218561135586,0.6981328829042877,0.3700923342856255,
 0.5286071392248872,0.42199879316269673,0.859861540007545,
 0.029481299492622015,0.5077685190467974,0.6319313541291451,
 0.49846540145545615,0.9237257365498797,0.22784420119950388,
 0.0180887282907497,0.9353464601492651,0.4483910444064383,0.38842163608803393,
 0.05036821601414665,0.3682001412870257,0.03231596833743944,
 0.009704002014158553,0.3912915906735095,0.125248322420226,
 0.11818419186998941,0.499481479543851,0.924448946427375,0.993002561115593,
 0.5670368358098585,0.16416652175323532,0.407992225101967,0.8311846403130214,
 0.11774695099276733,0.15076680790289698,0.4905104773527422,
 0.18563971244891797,0.5331965965098958,0.6346878829982621,
 0.02854906659969192,0.4173897800082764,0.4518519791252984,
 0.09371211407531255,0.037396846864345634,0.2937572007493727,
 0.10031987234232576,0.3171014155374201,0.7158896584298801,
 0.051557287043945355,0.21267160436618215,0.20319443715793284,
 0.2857552540715548,0.6747163045138567,0.026117473709775907,
 0.8377467414834536,0.4272646498646794,0.4582680885476835,0.16387392809345913,
 0.9242724723765726,0.015740321124550705,0.5333534065253323,
 0.9254636383161816,0.6703588855542635,0.18808108187252437,0.5702168318230175,
 0.7014558534503447,0.7443101176709214,0.8242543491948378,0.6615008759195444,
 0.469991962300893,0.4117403760797721,0.25660841183625493,0.2809254010757455,
 0.7275477497769572,0.6455227178471871,0.0978678467108759,0.25551509508205705,
 0.7019582960701173,0.48126770693303345,0.08564940065962601,
 0.9132474241904355,0.22254489788221088,0.7788894648637126,0.8485958206521685,
 0.8935864823875128,0.9506854074134594,0.8483089062547111,0.5338797790037727,
 0.6929961711025281,0.7940012925806987,0.947171420760263,0.7777306724522052,
 0.016755270917066745,0.09210504176612655,0.23652357663380474,
 0.046089264049241185,0.7474717209618681,0.19556638232864354,
 0.22419286862736298,0.5247410826111275,0.15532928558842807,0.864830464387803,
 0.44892339433573825,0.8283245120162721,0.6666786370067065,0.6500499978223199,
 0.7500835310858773,0.8054201754047917,0.7923770775449664,0.5335193963574013,
 0.195512148538594,0.8651811738117807,0.38459003325298713,0.2779393435600447,
 0.09541502685630865,0.19661239280619058,0.7226677197419976,
 0.5722385752524506,0.45757659925167404,0.5542090820069057,0.3166886455330743,
 0.018873416996927794,0.6408502630500755,0.45370128627684014,
 0.7414343072423204,0.7459919918306563,0.8483861431499601,0.2953799536740285,
 0.9404418767209233,0.7793943528788442,0.7484247332077371,0.6237027730958776,
 0.6841604301933323,0.4622033649120503,0.3795768287312107,0.7430957685453661,
 0.32675397675780016,0.050509709124676005,0.4223767195343998,
 0.4228331786725502,0.07518958939831788,0.7992782902517795,0.2540280129981799,
 0.10564929304276993,0.5294240956524571,0.7364290154196341,
 0.30398038867858956,0.3185462552787245,0.466061699585971,0.05895601349374524,
 0.5214582559145324,0.05478986215996384,0.6365390878231384,0.8205302163077774,
 0.2867726670016959,0.8030738769601872,0.2253400591099719,0.16734228765026948,
 0.19618154447285685,0.13822888944602685,0.4018128443136316,
 0.7481414155482292,0.6703422599384291,0.2501897839923729,0.4026425862621754,
 0.17390021373761122,0.09662821348496609,0.701681891816413,0.628945107695254,
 0.5242096579398465,0.32384421690834286,0.8526498963377378,
 0.04062309971196576,0.963763486720547,0.36841926848295714,
 0.06716419202253221,0.7168733613658447,0.2747136629828856,0.535532465846569,
 0.7872586321123356,0.7027753785957255,0.29513359242963566,
 0.46199802834192605,0.07953683973847081,0.9169951825722925,
 0.8491412873399954,0.7284090198153388,0.7705143905705372,0.13019016397849725,
 0.9515100594664908,0.05685709029871011,0.1592078912943603,0.6865960272269063,
 0.002989791295864963,0.9575280486407508,0.02912056077473002,
 0.5100999298784155,0.5554006836616907,0.13880436674074426,0.6157373533373928,
 0.7757141361713186,0.39458886739698373,0.553976565641796,
 0.024907538399286278,0.4171623166069365,0.4586217020940144,
 0.1220317747966756,0.4223363818216521,0.3040291459709744,0.12867958479162467,
 0.6309473264715538,0.8603302407317761,0.9198459956558482,0.611193818626479,
 0.7142995389112193,0.25543237854657275,0.6193146089394325,0.5375197930150641,
 0.19242853842271446,0.4464410391180089,0.6449033457682973,0.7520770320623716,
 0.1864197866186018,0.8992772970219806,0.5636180669750642,0.3343357760673482,
 0.1789264044205573,0.8801110615867452,0.1546289800791545,0.6283786924289847,
 0.4207565277116738,0.06655205359563843,0.9496327332044654,0.9912335372096547,
 0.33740170221369414,0.9607011081444075,0.08907622935192805,
 0.6481795300955018,0.4248203471597829,0.4735794111908097,0.13827787550734227,
 0.4212399734010145,0.4025605338879894,0.8770739943788786,0.4066566960951764,
 0.7535896613427783,0.08154724023360549,0.9004971608941643,
 0.017261511008372477,0.09069778981830201,0.12611074936618238,
 0.7966702282413626,0.575280567948755,0.4192522602056473,0.9019319296661308,
 0.7739537112069583,0.522905305014378,0.1421893584995908,0.7994011568468546,
 0.5010243512958695,0.912274216451413,0.7770877047875746,0.6415274480336003,
 0.23094165239035735,0.9794926589845256,0.5660355872361873,
 0.38973839484188644,0.08294337935346341,0.5875085986520385,
 0.9337533523584547,0.4433372858394943,0.969772706579051,0.850900263466758,
 0.45327986067005266,0.585071465746736,0.0646275292150017,0.8390760137469724,
 0.8747088513841547,0.939203817767334,0.16730068248292662,0.27097068186933015,
 0.3770523631396565,0.9883288761966693,0.2248026080561114,0.7740741813199536,
 0.8975658991787447,0.33148248420860305,0.09061040862904424,
 0.2432194020241395,0.7550171109847823,0.41201408921102234,0.7750505618240682,
 0.4188910891954254,0.5017725598050741,0.6066713853116387,0.7284569115638588,
 0.841786902110907,0.6330620073198708,0.3656760244758581,0.7848593336193068,
 0.7448225210977354,0.7650184698419853,0.9059315459752502,0.815285162632468,
 0.5634071722102109,0.32247976230005393,0.18207534450378338,
 0.029543088539745854,0.74457856562048,0.4603359357249822,0.8824351982110198,
 0.1856342938448977,0.5527029103805465,0.5919255063660014,0.2580855018840005,
 0.2667380505773512,0.7426141773214304,0.5814179234670821,
 0.011824616073224381,0.8014816536928433,0.3436690151896733,
 0.6584605036078435,0.9458668920247814,0.7243564852799595,0.7702389443012887,
 0.23714737382361661,0.09241370971151441,0.9238399471458368,0.225095296170989,
 0.8447727233837505,0.22211528322784924,0.17692432655860457,
 0.04098384529685106,0.2953377093755357,0.9548344715551298,0.690103972062331,
 0.22032135011223197,0.617087598425549,0.822785373876453,0.6827614237675229,
 0.19480941807808405,0.3791980928873825,0.2604125777167181,
 0.22445828813758661,0.030109382149388852,0.7836468012301051,
 0.5715318366501777,0.5810623584960896,0.19272625519786657,0.7102160382462799,
 0.41557651264393924,0.4745189583067202,0.09816278406118939,
 0.007476871394960338,0.4550803882835981,0.11374511323006076,
 0.350526496612077,0.3180772154460343,0.21724245218120253,0.9276600513662824,
 0.36419568304568883,0.7798872572045531,0.6689249695590238,0.3466193178382767,
 0.44983247931872283,0.9464140303353962,0.08236261692008506,
 0.9556213566377336,0.4236361745966163,0.5818844673232971,0.9345477441662584,
 0.23010973225799547,0.9942144342404144,0.4772302107423735,
 0.35468168897093977,0.296726876967332,0.9204594044660801,0.9348604407141422,
 0.5741355778395324,0.3522948681361934,0.7053457707195293,0.38176115452150516,
 0.9685589334386965,0.18659298193504847,0.609387015298444,0.3964844489070505,
 0.774737088745451,0.5452602724481455,0.528932413760114,0.3855263815135721,
 0.07668087574468008,0.36726875489936284,0.3128257892471119,0.624829061758024,
 0.19282335667696648,0.6958593160278115,0.3794492854812459,0.8664116329716286,
 0.2403966213388219,0.3211936436879782,0.3916378786727446,0.28392919686721885,
 0.9153139626777056,0.21832787641970075,0.18754815981522333,
 0.39162588730066517,0.4205902316924688,0.5051570235005768,0.6921084321679963,
 0.5124537645423911,0.6688177740714564,0.45929401499150757,
 0.037030617198756555,0.24163948200928853,0.776493536527878,
 0.5918393101311428,0.35429866518492137,0.5371786451651681,0.1810155993836845,
 0.08848475366884123,0.9197809717067458,0.8924794396867362,0.98505498782752,
 0.7369706141237339,0.5860634405494918,0.17059421715995182,0.3316103181351603,
 0.17944324713260806,0.33171201595221744,0.5748069324118318,
 0.8685269844091056,0.7278009256377153,0.4902673497410115,0.2322760036633128,
 0.41372305591303504,0.020739928226995774,0.6441467217089281,
 0.5995224923945451,0.6898798326768035,0.636080826802805,0.04601769216130869,
 0.35475203611624095,0.21310655680176327,0.1939010334101836,
 0.08864005659039198,0.18488162487050708,0.7938037357526428,
 0.23401682021377335,0.208502392487401,0.4737960857538669,0.3595521756141813,
 0.7185401144573638,0.8943929099852789,0.525249924929854,0.4090149741866981,
 0.15682041438782335,0.48717707869849225,0.8310907925524981,
 0.8721576181662106,0.4933801648385928,0.9704078116187707,0.20570950813862499,
 0.03154049018906269,0.06943004199453917,0.731136569047926,0.7696034848918264,
 0.17535406287392852,0.028969802447891713,0.09586569554287061,
 0.46757690142549413,0.9802236020646125,0.5269554470041536,0.5442526650274585,
 0.2079420470834462,0.7071085071177535,0.6994786959228478,0.7093864122277175,
 0.47152196122674284,0.2304037686938708,0.13921247781577706,
 0.2322595545361158,0.024024434820095197,0.24751389437911808,
 0.9696986802274862,0.4201883758054925,0.16146470455341633,0.0785009637660895,
 0.5829393460951462,0.8530472238521993,0.34905466953326947,0.5483637288884526,
 0.9873932569261065,0.5505592500987855,0.3702472648972469,0.20690775831251784,
 0.10879057459708452,0.09223500658798467,0.27286159120568065,
 0.8367314489872946,0.016531688685488044,0.8362519380419646,
 0.42512803455860526,0.9411759430559202,0.8313758354169872,0.958706770338186,
 0.8669107608759642,0.8434684212606886,0.7543215309062478,0.18602188503150663,
 0.9208253015966479,0.4612040140132856,0.34703407541687736,0.6216834449483073,
 0.6773717414956256,0.9477285900746996,0.9249081124166474,0.36749545895622215,
 0.505581721046995,0.2562323354305389,0.7817489162604634,0.2000679624913626,
 0.02144468000278954,0.9043531789563263,0.5990679615131744,0.5156446689083677,
 0.060382016135097305,0.07856090638815494,0.4233377529556285,
 0.33844198798438496,0.8478159989903811,0.5597616329182001,
 0.16373816448940026,0.3902731608052393,0.891067579810572,0.08402019501028724,
 0.5350071622328912,0.0974531352881921,0.11515532234190673,0.4538993823508872,
 0.7836268964766699,0.2654049005519703,0.12046153569792051,
 0.014307454303404299,0.4968007697964195,0.31304871428533687,
 0.943184830917703,0.1123878542465061,0.25238057138661885,0.6352529733499411,
 0.31103398601466614,0.45505291178026663,0.6144985124510545,
 0.3061060657934217,0.12830080117435483,0.8208027050970242,
 0.18374314248260326,0.7846231363482471,0.26344017936707886,
 0.8505059727936377,0.025919425228831394,0.07717846855837063,
 0.6337711234841417,0.9660737984933593,0.4326636115965914,0.6795461765533742,
 0.8962085570090029,0.6512752025351526,0.5614196706117711,0.6159627822966018,
 0.17470885974883943,0.31189506560848024,0.39053580712502056,
 0.8624321116768631,0.6669197124400746,0.7849871767675923,0.624208081542107,
 0.09601226834862975,0.4165650755772836,0.18733609574484533,
 0.9709762510752462,0.8159945352639646,0.25191710528208466,0.7480687209137377,
 0.9276368658984165,0.007707863864592879,0.2143700205582807,
 0.0995358809932777,0.6663825915131517,0.7887674304113754,0.4343107539741221,
 0.7446076804804317,0.3728956151739863,0.44994479982215174,0.8339449969283098,
 0.2909607870880526,0.6361449822849377,0.5666269203862857,0.944691412789837,
 0.47577489585006927,0.3433116010461553,0.6800839631907246,0.57322059868478,
 0.5731953555099161,0.7716155989955885,0.7466500969160761,0.73883587883277,
 0.5938294686065929,0.1795804951947303,0.5183704879345697,0.19918366674467203,
 0.02103814252051739,0.24783416236375566,0.1919786303896739,
 0.27829637457563083,0.3340355011057593,0.5994269685995944,0.7604936771730424,
 0.8502800495539152,0.659965378687065,0.8104027620011992,0.9246719130745228,
 0.24200247912091355,0.033640157466802934,0.5001541110774302,
 0.8863666469547514,0.7028419692432007,0.6465974416937788,0.7432601445566888,
 0.0030112891244372664,0.9212158323142277,0.06834463026957593,
 0.6507612809287722,0.5577746431341172,0.43490093204554414,
 0.24209580834879674,0.0744963776397688,0.02056559609279751,
 0.8048934711216205,0.7831750010986205,0.5989342007944244,0.1598257962523062,
 0.8790023919516204,0.9306913730653253,0.5253334907742784,0.8205697236258598,
 0.08089385268664173,0.5547959509743497,0.6947779573002539,
 0.20620143212106123,0.37820925153590734,0.7026488714851689,
 0.9720453246935268,0.09904494075261006,0.426125597941558,0.9136113323962877,
 0.37657682424052397,0.5802695237274834,0.7789906363564402,0.1844039194193986,
 0.5500325781328339,0.5463298742719337,0.8515381725238627,0.5439134493823606,
 0.6129420510115278,0.6031091028267308,0.5659733378124464,0.07205487760980855,
 0.3693570784667739,0.2176838674242294,0.39986720349318683,0.328235670137652,
 0.06973333678280325,0.7995149867656848,0.3230798688129073,0.8533323153180727,
 0.7423357466290986,0.7188986885550355,0.7245639987033571,0.4234600299121458,
 0.5446159518244407,0.7093613019896088,0.7073060926775595,0.924608497699398,
 0.0848137247463443,0.010527682794379878,0.8600466345251732,
 0.020439914173052354,0.7400669426275758,0.7785241454463828,
 0.28913827997341324,0.43136929848918926,0.4532914037965168,
 0.03434034309232714,0.5691441789539666,0.8151433580992641,
 0.45011309796759347,0.8235968299601246,0.34936976071172,0.7227000379532941,
 0.4256228096433192,0.5474532521525682,0.3513601892815744,0.06054065899500072,
 0.6718470125277755,0.15120923589806856,0.571111405924531,
 0.007090853009057763,0.16818618698712529,0.1451076226924195,
 0.6428715095844997,0.7051425760573964,0.05546150966628405,0.3068124470676926,
 0.5556462617393465,0.1239840883478206,0.008438591127835737,
 0.8477177912682979,0.4155524321852043,0.7184049912030593,0.5588843493034642,
 0.4711553765875034,0.1433316795709736,0.8994330212730084,0.5225163725004807,
 0.9745967661514913,0.7976646274648037,0.5057048414755825,0.34696579300247166,
 0.26405385878568965,0.20439655025357517,0.3670531743681442,
 0.02286788360884473,0.3066777610656929,0.5554561925594425,0.5528534075798124,
 0.7241723009166473,0.6596095606702704,0.6244343382132104,0.7873797438049668,
 0.7149537679620008,0.7572786495769398,0.34359662703009475,0.3329723915239501,
 0.4015242253247081,0.07733505674155228,0.833651847553701,0.7775288876530512,
 0.8504001446830672,0.2957103075403662,0.9332481390765067,0.18076091538781558,
 0.4714267417302178,0.17610029051620946,0.6744182892173074,0.7488013926579274,
 0.3136188308860257,0.7482370645201775,0.897240434585588,0.07152148946278791,
 0.7697670970226218,0.21445991365157946,0.5995561417169518,0.2051979487293044,
 0.09071106755888125,0.8393473094633378,0.10943074272029929,
 0.36045441205172235,0.06352918383952866,0.8389861238644252,
 0.40116238242040114,0.9353816637041275,0.36149189995805975,
 0.25055126980364517,0.8805499186128478,0.5178300137267775,0.8097081066183295,
 0.6223341808571778,0.0942607635112962,0.7399906598786004,0.5911101495425721,
 0.9014945740845567,0.28626928408668684,0.14177912363050316,
 0.07642091868800405,0.563274247747622,0.3528991042373022,0.9340956000500598,
 0.22919151059370158,0.6565516783767124,0.9154351596019684,
 0.14036029517558035,0.5227402315503112,0.9045914256976668,0.206276033149843,
 0.42739000577399755,0.9180940559577105,0.6518322599070959,0.6231583693512395,
 0.8230596939598619,0.25173074683047036,0.735469835729617,0.07135224586060385,
 0.05228907914887093,0.37242599147962485,0.09131373566928402,0.84677976379613,
 0.31912475206525415,0.3215457627555107,0.3244221958174792,0.6256438179236379,
 0.14861190339460273,0.12540110619897882,0.3111791486868084,
 0.9117388200491972,0.6568717881208286,0.29331184516412434,0.6582765468652683,
 0.38475799103515596,0.05116903119142069,0.2830690155224149,0.538659903923254,
 0.897196151217092,0.09586819197906205,0.5657471784095174,0.23964734552200062,
 0.7621933981796111,0.688165782830985,0.8358328189623873,0.032264323486323354,
 0.661610502424517,0.4840502851914177,0.6636214602083373,0.6515480508761717,
 0.23766552354537873,0.1301717945923605,0.04865607073334122,
 0.3196499670200581,0.3120555607732981,0.3243366217628487,0.6717894339040166,
 0.29804480547829626,0.29525252239656674,0.40790852022474033,
 0.43128071726628003,0.4172910417553106,0.6642801535432757,0.3961433060704609,
 0.9524900978269384,0.8403742816314104,0.4967006955304597,0.10401250807677498,
 0.45903261857298183,0.9778373955792382,0.2494471390180748,
 0.06989019247406181,0.0649911412017905,0.057171389555783714,
 0.48876759872321984,0.4101780665065373,0.18086572746047214,
 0.29230263707001736,0.18610249444703875,0.1945098299504885,0.800937966103935,
 0.0953343898307143,0.3144740057734421,0.4237115333421594,0.7673854026698149,
 0.4790470390997891,0.8541749685295124,0.016439765188652755,
 0.6307562081529147,0.6915397288481914,0.9046266291856779,0.8578933126772923,
 0.6410876221882341,0.4320827405269506,0.9434084931035263,
 0.014883996201879746,0.6045720614126981,0.9439706096737258,
 0.2635824601822062,0.05656877185089215,0.44558790490798783,
 0.4874779729271703,0.6946897853212637,0.7843151918771931,0.04789941995576752,
 0.7643711916175687,0.09745864174766616,0.7993111860992496,0.9768016698263953,
 0.3443152104382221,0.08691137565733498,0.5180516806305708,
 0.25943143212186703,0.5172729936661999,0.03312385073282409,
 0.27909836020452994,0.48588847953974956,0.26356833464366347,
 0.10113669750826904,0.9822718683032163,0.7270779566767822,0.5348918083570817,
 0.11534983155987122,0.7444557751661479,0.4829564194056172,0.4300025447714738,
 0.843718796894636,0.45165087924696967,0.7487244171863127,0.1782473900665471,
 0.08036661149811442,0.7838541695733461,0.8607240344008692,0.791425999181086,
 0.41884305036930636,0.18549149561903444,0.1606108690849769,
 0.5106197695422139,0.26077353163015016,0.03294069785799669,
 0.10006420788279369,0.7814078797624232,0.2529875916855172,0.931657620167941,
 0.8943828133809779,0.2165262034068931,0.3072501897711185,0.38614427807068763,
 0.8898337365236828,0.06117145273757851,0.653905924743572,0.5487717663160475,
 0.38405849885814514,0.09836126008358126,0.5932859840243858,
 0.9134188777684614,0.024786734655233622,0.02240460022292068,
 0.4569756445935692,0.7840714766473739,0.03231553577090296,0.6462208047396173,
 0.3095525501264944,0.10178670765165243,0.17619919985074706,
 0.8909237471273934,0.7152250814065536,0.2407327989701633,0.9725496102842945,
 0.935875740326361,0.8283780756634416,0.6906156273381812,0.49275996537312095,
 0.10497964668059168,0.2296557838753358,0.39112566791845393,
 0.1378462267267062,0.598873876907517,0.6327040747394853,0.2246647381203537,
 0.4919850296456906,0.5293973941211053,0.14357240967532892,
 0.13652265589586288,0.23582441675510046,0.2648592585811883,
 0.9682187043846013,0.8846168170629398,0.6563162628674059,0.5085911147555489,
 0.9563738654801033,0.19293747075827383,0.7094035424348071,0.5648801425833379,
 0.550468786365113,0.6327220458316071,0.4677407985766129,0.35301843059351723,
 0.17165111565562008,0.22961655764805144,0.09770697228560421,
 0.8964332952356473,0.5214286795730642,0.46086899996932507,0.9004307022326461,
 0.343311172216868,0.29363865078947815,0.022768371831272116,
 0.8594677293826095,0.3786304070849753,0.7723686235278883,0.5736796714253335,
 0.27970754007905063,0.19369896960306998,0.10488999190382564,
 0.09473680715130639,0.7926147215975059,0.41846493315274835,
 0.7043187712367298,0.8763970664098936,0.35677442432651396,0.8214803615603132,
 0.25637112277381446,0.9967974319670021,0.2364071771591092,0.4052391632890948,
 0.21124711252286665,0.23954509673420743,0.8853151941303071,
 0.37386307246697437,0.11284508926702508,0.9369169428177642,
 0.6519483914388676,0.8229113944321563,0.6093308757238934,0.9151111092423739,
 0.5651274140998939,0.15280927224285734,0.8360012151206078,0.5852457948161038,
 0.8170618076682553,0.1621633385947293,0.24250524062648615,0.9554849144367621,
 0.2961423280580221,0.5062834392146216,0.6703879392876606,0.3996320308973871,
 0.6026342489939082,0.1676610897387656,0.3204795118463222,0.33888889166219416,
 0.5260661395003514,0.7358578495676841,0.6391906554860991,0.16778188226147628,
 0.3726829550681934,0.184836482443956,0.7474140972500753,0.4929716270759853,
 0.12746144408897475,0.7773743592197669,0.8320945569327609,0.8960732659725176,
 0.2944102187571147,0.9814052846562302,0.32044555283539855,0.9303837838739972,
 0.08244384107628411,0.5715717525793405,0.8450186968871112,0.3152127449509531,
 0.5886159079483085,0.017602887237219256,9.326572001034927e-4,
 0.7047064204580777,0.07746149931668322,0.889187898376459,0.3490681612115317,
 0.7194097173996513,0.305309105105278,0.3729727539868781,0.2812093583038331,
 0.29507117679029393,0.8659729823078396,0.09679133763262393,
 0.7824481842929584,0.9183818040819931,0.677772537722793,0.6581938872056347,
 0.24173355933963248,0.7411279521613376,0.12925809581435554,
 0.45931062697183855,0.6926983851480163,0.0840473014817874,0.9012410484797313,
 0.9558299882129071,0.38820229038103615,0.589426010923472,0.20819782523343877,
 0.5018612309471309,0.9799807778197364,0.8905229970225375,0.7465747044234724,
 0.11143140936496443,0.96451722935609,0.7224602025051785,0.6169766371659424,
 0.28955997389291566,0.2821858434765182,0.272024551091239,0.25365822102248414,
 0.15765397387096136,0.35141627462486247,0.43901632382437206,
 0.16211322217941315,0.6805259200339139,0.7664465892671048,
 0.10555906805210602,0.3995963977022031,0.07672630717879314,
 0.5640601418566377,0.31274564489491463,0.018964567120142717,
 0.0983758125507691,0.7827260938827931,0.7574486432404844,0.9898585129473829,
 0.9245898943137731,0.47189638224966224,0.791448155567233,0.3342515114456326,
 0.8432367339961294,0.24910146971054403,0.15860239695040879,
 0.7124384955688781,0.48518495961508024,0.9668521151284813,
 0.24068126657093503,0.13226960067849403,0.054627551321844714,
 0.890003656988632,0.34277403455860034,0.9663987469307229,0.6378607088309207,
 0.19126219991356963,0.7785345249737511,0.866205329472011,0.5400130995686336,
 0.797164242074894,0.4782103710619311,0.9495006531020455,0.5230460452678292,
 0.5312877548276764,0.4623470916233088,0.8120758196983272,0.821662976138118,
 0.2170772024080294,0.1830635582691258,0.3865726055254759,0.6957335585772,
 0.9791413569435543,0.9931909463880325,0.3223027793303006,0.37255284465593674,
 0.4034136890421909,0.44122910162314977,0.5940195652905377,
 0.44910832628641506,0.2218954877017847,0.34573231038214847,
 0.6373218877498424,0.969401757860485,0.7270462475285743,0.40010921782933073,
 0.09646135294887759,0.7155907258775965,0.12532223226576134,
 0.7491076416069311,0.9787270286122066,0.4714112168536667,0.25092279497552905,
 0.6032214228720512,0.6817836386947248,0.882887587835361,0.15044403751131374,
 0.31246426231940616,0.7907663575176821,0.16241369152811713,
 0.34455729716952177,0.5903086666556328,0.6949875453898885,0.5420907357773365,
 0.6613287426045145,0.5313752840317572,0.5518761555925219,0.25519333974977887,
 0.2687471758917005,0.9158312640333106,0.26587991897343843,0.7470017433134444,
 0.3710510328254222,0.4848484940430371,0.7869663293606344,0.23690869530883618,
 0.6190678699485331,0.5887610698692781,0.14101613097117038,
 0.13282630996282652,0.812196048425033,0.43720607439396864,0.8775305742902919,
 0.05256896094632402,0.4465758162933625,0.5287836957173481,0.9036470822293414,
 0.8127322215672449,0.14703312085455367,0.29205263810833015,
 0.39113276856582413,0.516547183561707,0.41267917613461147,0.999895799396828,
 0.3471039584757172,0.5358990239441944,0.4562066383189277,0.17859972560011328,
 0.4696945374134226,0.9594454494663789,0.44369693036746016,0.9578752104509014,
 0.8696542257012531,0.08502080093925701,0.5772126289153846,
 0.07689904183721241,0.7243537094172634,0.24823828270895332,
 0.07557017021895296,0.0667188380106315,0.8583874756571072,0.6961772833434987,
 0.7773421275679088,0.9731537111691685,0.11126924121981796,
 0.18359088602823836,0.7071507755628925,0.5187510751940086,0.3942071209375375,
 0.9437245393688054,0.2581822954208737,0.9287692690626808,0.2252018644321464,
 0.004567787583541483,0.05712646919025066,0.42437425851569544,
 0.6920456537376443,0.9664720505665774,0.5429601387388748,0.7534972739931899,
 0.008544823090457343,0.4041282825793221,0.7607073023113842,
 0.6625153735514907,0.5827217290757783,0.5552336159697324,0.7183452359002035,
 0.26899680811111915,0.15848693026448601,0.38239807369699375,
 0.6804087427703515,0.6926873994833415,0.22669357964022063,0.9858910308660237,
 0.08934091864627447,0.790021955010315,0.8986549134163458,0.3921037044198934,
 0.8478295004092868,0.1969035735694722,0.9834812384469551,0.21037928588213373,
 0.8012240834254227,0.2878993009400649,0.7970008818089942,0.27806284557426864,
 0.9546061658685167,0.006107626328596072,0.4813838947674255,
 0.5098844405790391,0.5916259736813182,0.11836442407655334,0.959292439485491,
 0.7432972596092955,0.32623965121549414,0.9666341820807424,0.6406746074533956,
 0.7048652698537837,0.007572237998346387,0.8973429260129032,
 0.2288887980238875,0.41919019383555,0.6859556351831191,0.7167135442951802,
 0.2987388444517143,0.5039816341225548,0.2612702924296936,0.07584738824112625,
 0.2736568659663515,0.8401531871438099,0.7654244189275174,0.7867567434818863,
 0.8762116816868298,0.2590842201152619,0.9455904899524826,0.6179596629002377,
 0.38376631235234626,0.4975879695513947,0.32190547883667386,
 0.7812740342330058,0.46679195057969936,0.459638507651584,0.20213366628350438,
 0.08741991869440513,0.8152892791357137,0.5526580465398794,0.9667103111390182,
 0.7349293311828519,0.16464761648532544,0.034535660231248655,
 0.9891330626319843,0.518546010312959,0.58689723535305,0.570711269268044,
 0.1631977256350342,0.6483184423325379,0.860794933807054,0.2252666051220268,
 0.5711919548722894,0.6954103591607396,0.78611214937879,0.06525527002074605,
 0.9669433226587575,0.19374314966129336,0.9703326176018704,0.9783090112349564,
 0.8214355250720832,0.35429300103126127,0.9736607214509663,
 0.16290283607859313,0.5620971530912657,0.8070438799666468,0.1777280718546307,
 0.22319489494064748,0.4137772776873845,0.6186705171936056,
 0.20252933912510784,0.8422386938019137,0.8029442235381307,0.8677356490917283,
 0.40598811228552856,0.9200804970525832,0.6234445301052471,0.0753387862333681,
 0.9921123005263187,0.9408896667497608,0.44328821011651787,
 0.06664623542481307,0.5580180445329241,0.7270675445592371,0.9461583474309093,
 0.4194897161312523,0.5738031885333369,0.20667585078396344,
 0.37687895645446623,0.9661927592066936,0.6143938025114087,
 0.24837001797375857,0.6854497771706302,0.14824953794908602,
 0.9905092258595285,0.813611601119616,0.9477669076141648,0.3319885658416173,
 0.9744739495428697,0.7911267570758053,0.8594455451737901,0.7513248523793506,
 0.7353592180525088,0.40034667787020917,0.03972883655916992,
 0.48199382488064213,0.2566186374422039,0.9914461866461641,0.426725267995141,
 0.8357900502053234,0.48022205489762415,0.7320102469169422,0.5273124095241151,
 0.924173692045689,0.7489131070637189,0.752970563155531,0.9312553693474214,
 0.9768367185308481,0.18483502948819364,0.9410407630534161,0.6143313052624815,
 0.5065975214607282,0.48068052320622545,0.38818402554761855,0.937332455361755,
 0.6710117644595429,0.4120567536323688,0.9173194035251847,0.23567662259969357,
 0.5331215838525611,0.4215818384933505,0.520401350850409,0.7627305324503595,
 0.4813842829730657,0.5235161549850719,0.5400089771902683,0.1059223962300061,
 0.9145398207529665,0.058437495053579536,0.6870289428666918,
 0.09701736933777727,0.28705982697556853,0.3153657416528589,
 0.6805546415741285,0.30124176235219446,0.24642536656570257,
 0.21329425597417961,0.6686872051109968,0.4657581540645441,
 0.14289640022537287,0.4444852239995052,0.5943655294159885,0.7777246437880291,
 0.6696756379101209,0.03631614041586739,0.7217808060443357,0.8404117894037189,
 0.5048937446516429,0.4680612330572983,0.32867171624871494,
 0.004697715757028376,0.907970394813566,0.9093143442497632,0.5412006841058248,
 0.07940973468541157,0.33835262464264926,0.19883063178608662,
 0.38321109331135417,0.5406777403477437,0.27299610547857,0.7448615463912616,
 0.656928684659395,0.04713466504525332,0.2651794737392992,0.8134914901943799,
 0.8452981554275033,0.9409993365017657,0.511498571815323,0.7017357114151768,
 0.44622748696451675,0.5458765107590402,0.49981286062650665,
 0.3462554531912244,0.12275627634546449,0.858814984203691,0.7162767323087446,
 0.9368047641651164,0.8517149388252289,0.1502413876324319,0.07709142493243026,
 0.6881177648514988,0.24268315685153075,0.2633089509039549,0.628088891035344,
 0.9714190529697535,0.5482697420507815,0.9210516698859126,0.846496308335394,
 0.2855741211911377,0.3329346848909527,0.34439037689065777,0.9555037862196218,
 0.3269200819698157,0.48849544000268985,0.8483304172882197,
 0.37496410392117663,0.45539699645908505,0.6711900345838089,
 0.7436114835521759,0.9882378134088009,0.09925217144043752,0.1672955052860433,
 0.7935551819699961,0.29749208620475076,0.2484251793833413,0.9161903622730299,
 0.396337633777494,0.002472106977891353,0.4011864849288228,0.9855099620705718,
 0.3599387627631262,0.5834578196692017,0.9457443558388481,0.9333721781488266,
 0.45123301616370404,0.1991308480797378,0.03043708081317531,
 0.6937443065417869,0.6973991133282667,0.19804482193471296,0.6298946512732387,
 0.9059279143058748,0.15510874565570587,0.37584441700273885,
 0.22127677760208941,0.23614075413975943,0.14865597606902303,
 0.5240781169859107,0.7628876966850201,0.39457319043595573,
 0.11493519156054133,0.3674242616622234,0.9722676200213367,
 0.40026929968971836,0.9790323257439247,0.8252257409515479,0.7760738817373801,
 0.986626144294669,0.620248429558136,0.8158944098146816,0.15336529635386142,
 0.3874276543247369,0.2979342679239718,0.9893224248553468,0.24706927098811293,
 0.23460983730300222,0.3435882304640202,0.7607530138770403,0.8420380562542831,
 0.07394087839521579,0.7232207422839293,0.014523374022817781,
 0.0022355613932761864,0.08954413443266207,0.31448301780891996,
 0.07879234596945128,0.7441963223372042,0.7554064256465289,
 0.02996745055477712,0.756539530451864,0.9718868107958647,0.4859191279603329,
 0.17729346375120414,0.7934690344183649,0.8106301700070784,
 0.17126636878279222,0.6407420370246447,0.6404938863003933,0.4754264096461922,
 0.4929634422889324,0.507850712450531,0.8988468920326381,0.8842166468716465,
 0.06024791224630843,0.35128034642346084,0.7424595228893681,
 0.20439564608117655,0.04257863556856145,0.9908003160680422,
 0.4705396793900317,0.9362303706017456,0.9301980924858212,0.7271860656514629,
 0.44539615571740465,0.5345459729330511,0.704678310997463,0.5595769030089229,
 0.5209712634701593,0.4291702012469629,0.4607093383480374,0.6350943934152304,
 0.1681334126232381,0.6048608881991357,0.6172525741898555,0.6642548456708528,
 0.5441923967338548,0.1263885112441827,0.48721172625927267,
 0.49709597953311113,0.6596517037789724,0.5805434070418451,0.6697383587067514,
 0.4229634503520098,0.42075466165729347,0.6454279576473749,
 0.11730845617641816,0.9714049174777111,0.07358121214138036,
 0.3975388140433541,0.11804237970111597,0.4847423680183761,0.9279539620650472,
 0.9123821418550387,0.2809411917152961,0.40603589942554874,0.93209883498292,
 0.07305082293745713,0.2902902451093501,0.8507635628965259,0.267627291816498,
 0.7651634620529788,0.011575595873394517,0.29169364611288173,
 0.4161742155677556,0.20916612019514025,0.2010416247619612,0.4300169366216742,
 0.6398456624471114,0.15191844890053174,0.9157055213491541,0.9457510272222824,
 0.3154795404854773,0.8723321822889535,0.5629118204125432,0.1446922372553141,
 0.9379395835537401,0.036638907101030904,0.8096821526304323,
 0.6692483012603558,0.30556849981453227,0.2870607289539123,
 0.06511653449993149,0.1604994238121603,0.6194136972877671,0.5576063199583222,
 0.375577090928389,0.9145644037017377,0.5418769725119501,0.9368943339918506,
 0.2772144098522341,0.5230178408169134,0.7545008652506409,0.44705182728030135,
 0.1914640498962762,0.14697157999212995,0.3525172917847156,
 0.02278190917409928,0.3614159968518821,0.24857983298696384,
 0.09252356545854923,0.9314039324504325,0.34171326532241475,
 0.4027791882593259,0.7930466267051014,0.6841433138757288,
 0.026143077790810043,0.621636548740737,0.29376982002787855,
 0.6456492893021418,0.814508722746758,0.22074831372831838,0.6009577854321545,
 0.2153365190169556,0.5492806070168025,0.2575185431592112,0.7247999078924288,
 0.2964855698690907,0.6291467096913642,0.294409150625082,0.925009860856171,
 0.007208286788176976,0.9827878034578206,0.10312416811591352,
 0.9567614181958428,0.13512308673777618,0.6352202752496537,0.5640059409890283,
 0.04280277664450527,0.6521645022025199,0.02910673865629454,
 0.6921760770089435,0.5720459082593627,0.03597765290987198,
 0.18951823038973792,0.2960333289733642,0.7518298461405232,0.6340855946635229,
 0.07402095811688003,0.74749732726945,0.05412460826282817,0.4074935099628001,
 0.20172119121755505,0.9345073316235392,0.026594723693421862,
 0.1768331208198144,0.774023172145897,0.09893164756329054,0.23083948603645488,
 0.8238972289662265,0.7295056506986338,0.5222180659087154,0.8085396992800085,
 0.6128264292559359,0.7927985564525262,0.10316651726692949,
 0.09185054665130321,0.04757457459157888,0.1381513616738378,
 0.4348827310518424,0.688607677937545,0.7922264497426583,0.5731832799131994,
 0.6144246312434625,0.5571295580868093,0.8562809956817659,0.10278894824856788,
 0.36039744937938545,0.01791029479356565,0.5543653296502299,
 0.05899787823721114,0.14294639312497992,0.04850480727310513,
 0.4816228714658113,0.8799973316500072,0.32504584456616037,0.7992372684648739,
 0.9019238367297525,0.4258193075800213,0.9057325103369464,0.3109174400632383,
 0.9425691537026184,0.8141984205764475,0.7920658146428887,0.18332586135421225,
 0.7097041832503339,0.39865623436840214,0.19608737918616947,
 0.9455980453595455,0.4744217018864163,0.41238028101607793,0.8135713241651744,
 0.7484662768549868,0.6131046033092115,0.34742969393225753,
 0.25881323440982584,0.44395263538564966,0.08667881383793496,
 0.2510277585335371,0.7847849896659758,0.33643325558286175,0.6716823311173253,
 0.14617826573119164,0.42897584685643486,0.4840254813194266,
 0.05849663547961648,0.9261401644672369,0.04895704825300884,
 0.08646635803286551,0.2236611806153379,0.09921570559942783,
 0.22926998408604105,0.3168100255008174,0.733837595880904,0.8157443076993597,
 0.6902877939203975,0.8069822191214754,0.7684407663531089,0.5339046573663175,
 0.6459763378253659,0.5451844190536198,0.2888109276446732,0.1405235342325375,
 0.09095898261889468,0.4030401547571867,0.058072601563939896,
 0.5496717683140026,0.8953296703799989,0.1621817086405577,0.709126882201989,
 0.2611287471438005,0.7361417855267379,0.8524682815390695,0.2616722231780889,
 0.7038174357874522,0.9124224547348129,0.18140317106352688,
 0.14670271309852945,0.14681050973313658,0.20528208409940385,
 0.9745969358841122,0.5659084942837879,0.7017899392841668,0.5401192415174807,
 0.9685413449087067,0.029827377897172158,0.8708365878051769,
 0.9636810426405165,0.19947850624807362,0.6983205215731754,0.7320667512042695,
 0.8111751891939872,0.2722946692956856,0.38156402248398624,0.7132950975025568,
 0.5411958067484225,0.27307947150737544,0.40042275363674973,
 0.6797064391178234,0.6902563627214926,0.2065245924733029,0.25996850098772495,
 0.5797700223986277,0.11581822375825512,0.8958123476018685,0.5565795168665737,
 0.37517680366507133,0.7241205723002556,0.3173062521260527,0.873237532525649,
 0.7761860754648935,0.5307186356296959,0.8137996349657142,0.8679852684624599,
 0.47381345811641484,0.5569062188582254,0.21881354631784755,
 0.24484089684346988,0.6903541911666743,0.7946791082449761,0.7924976359508323,
 0.06013635246937532,0.020017125358572252,0.6231689004939507,
 0.7906509266564047,0.2308658307935023,0.9549543952557497,0.09490307978626944,
 0.6938777229198099,0.5682727461696987,0.6315846609444861,0.36851347566703085,
 0.13984345569359624,0.5269236627945755,0.17077822268437304,0.982069491475687,
 0.3826621741234456,0.080125467263366,0.6037518881241182,0.3222220590610654,
 0.6436328790341915,0.6338311199676453,0.32478757853828466,0.8328026997180896,
 0.16733283678885003,0.6914608596862517,0.58206934576876,0.22833066101039645,
 0.9842421593055245,0.683633060789383,0.2757362972110293,0.4107001322741901,
 0.035418235940937626,0.2021973956584867,0.062052624530676614,
 0.2496510958216941,0.4156093850006809,0.5872999912187699,0.839617014834509,
 0.884072511021029,0.22359925664106728,0.4381731122454956,0.4151676823911187,
 0.6997449065584267,0.4676857035280245,0.033911512230100915,
 0.9527589915552208,0.5020060253803234,0.25020911806502155,
 0.22100010506098944,0.15084769996606162,0.18802816550760115,
 0.38270110455560347,0.2612816352721822,0.44558943054109346,
 0.8528001285316638,0.2990373560474344,0.23370758370654243,0.957353002437948,
 0.9159013597256405,0.7958213989849756,0.13058102331636312,0.5749870290655144,
 0.17834547735914752,0.9777219271882175,0.6771727897220011,
 0.37137880856604677,0.5196797673881081,0.16229247435029626,
 0.7270101713108402,0.025850911528627063,0.6703478247608897,
 0.8884362026499241,0.9368943938040777,0.6809906608781355,0.7450433723254205,
 0.9808879619869009,0.6722286365606102,0.7454569735787221,
 0.036646503016278476,0.7379425169451208,0.06105456237909512,
 0.07577273662134765,0.6247970452389604,0.9365591931502254,
 0.015301898397169333,0.4407937871006755,0.4064195073324347,
 0.9930856776569271,0.1447108004613884,0.8538086614169086,0.09242848571624451,
 0.7102229471739645,0.9299049025945773,0.089321214830129,0.9213615976544212,
 0.17047477546384626,0.691369621384789,0.4991266962894457,0.18753411902917827,
 0.46546060995083105,0.5161206952256436,0.8067859906674126,0.6206581761283658,
 0.9513252211051525,0.8941847192157922,0.07023847774070024,0.8749205418372396,
 0.051181689492746596,0.7747735987618063,0.2619186340704731,
 0.7287072622415888,0.4050371479259357,0.2223230826484941,0.3062562521593464,
 0.14164461500807202,0.6886847709213232,0.60412629467772,0.48271861796087867,
 0.24537913411747492,0.342677318150691,0.5389414782969159,0.46527678812977447,
 0.9233939984646151,0.4766701867781693,0.47234906385274544,0.6340604883822777,
 0.6590549257799341,0.4264293586908344,0.922402893063762,0.3728419693172327,
 0.09051513208105089,0.16994812708956675,0.16372461510101166,
 0.6831212475658621,0.40776812510095484,0.3376061949363316,0.6158464113665737,
 0.32850693353246196,0.43155157708409575,0.5311962921135206,
 0.14794930339103984,0.4885168859617054,0.8454747769909122,
 0.35144700354067415,0.7246169249481369,0.8464125472320433,0.8747801419037438,
 0.1815787432897762,0.3541140002414159,0.1814348438966571,0.6096348863803209,
 0.9561519304032849,0.9756523479908416,0.8495112079937771,0.5479033135014588,
 0.03405118559959952,0.4153643600813366,0.7043494459440532,0.3299051484342471,
 0.045212293183076624,0.09230172555691252,0.3389342918738629,
 0.5581341367778918,0.41095840069438105,0.772537987694635,0.682263038369278,
 0.07230011309893003,0.4822245271649316,0.5451743554292185,0.2579543370212142,
 0.9835231934785694,0.31712150428729635,0.882889000359679,0.1829424879968552,
 0.6908538472105648,0.43965237493975706,0.5156878357620762,
 0.15335447842267314,0.2005987066365834,0.970315599683376,0.25019605700096104,
 0.9772637817841997,0.8645170846277639,0.5526990119636909,0.12937722588638834,
 0.6272390768786348,0.008510481866379593,0.19971629536186974,
 0.8823847828037245,0.061753434519968065,0.06409954999118694,
 0.8553452097778212,0.7783220960521303,0.1971312695713746,0.41523966999309136,
 0.9477452021260004,0.720639647504298,0.5644961688505852,0.33657409591140974,
 0.44689854678062213,0.2748301712850877,0.24867100804517406,
 0.9961531866579836,0.17727912805419,0.7838229625935328,0.3930622369426935,
 0.7824750961021659,0.7271373066755494,0.3448575076868726,0.7538156837713472,
 0.13541705073882204,0.935420601292214,0.17459963434655534,
 0.49110723253672517,0.6227621907527702,0.39239104040914197,
 0.7196443362481887,0.5617615715260551,0.34025465393491594,
 0.21463675281985095,0.39379314324980275,0.0626212222195327,0.615170369105894,
 0.1868111615919118,0.4396434533707174,0.9293174706514072,0.7801547600909837,
 0.9801647906759844,0.5242438842775743,0.21823068533391965,0.4977316205158393,
 0.49599841530455846,0.951286627176637,0.6503958609902973,0.27082105802823997,
 0.905751628388168,0.6350567474297666,0.15741310855066915,0.03740179867140969,
 0.7500123410203479,0.514397519838368,0.6770272578941338,0.16754530050064942,
 0.5771301033179984,0.3361625078331094,0.4895696189608314,0.7165248204491534,
 0.16465922349779472,0.8435617040518422,0.36724726808483654,0.968185091156955,
 0.7057419613378826,0.720636530381745,0.5474163174106585,0.22566791593388336,
 0.27949146330408037,0.2877098564216517,0.15361973784532545,
 0.5653575129120452,0.0033012106624370485,0.8308210482272639,
 0.9726677740909669,0.5403248620840162,0.785356531263769,0.006316836393878478,
 0.09193073063617085,0.6353624645641899,0.8044837699468621,
 0.10688581190595725,0.6188601002477433,0.5774659653326636,
 0.04978965823260462,0.14775237720002865,0.3210431847429067,
 0.8220627171265605,0.2591690874174317,0.9503612723488862,0.7172248927840323,
 0.31176622626723993,0.2191642246526465,0.5305038321816844,0.48405488637271,
 0.3729460275454274,0.13396026167478503,0.7962696776356847,0.9310130498186557,
 0.3050172253628267,0.7548703096761403,0.8846502742918895,0.04050349140374632,
 0.015145689170384058,0.7059053052619516,0.20920126781711823,
 0.3105638355894116,0.21231960818511464,0.45930173771223437,
 0.25653210223830814,0.1009122376104915,0.5120040181982946,0.8980953579573024,
 0.9172635270235441,0.45069305960852135,0.15402502126490858,
 0.9374873071057475,0.11411050114988719,0.40439279735901923,
 0.03489423537048575,0.5860856133775163,0.27672328458535,0.5196024694919426,
 0.33213930695775773,0.8358812456699617,0.13921906887172752,0.742667845908731,
 0.7309656110089868,0.6821922432607052,0.28452179059201566,0.2570873791133883,
 0.30734672002394847,0.29659825574819076,0.048474358746698254,
 0.35117061883838885,0.11355474332122728,0.20955336926770363,
 0.8965806348217024,0.3075744039041275,0.01659770095669577,0.6664858714557925,
 0.2943589745435523,0.14500470668299148,0.6472746342609459,0.8235930380775067,
 0.6642170989188995,0.5508558863669858,0.6162117021243043,0.8010093523412839,
 0.7323104100042315,0.16423945284078956,0.6588777589654213,
 0.21395564402272882,0.9040007829038479,0.9298177741396133,
 0.04281122988289421,0.3410394415866025,0.43909077198133784,0.797629290572941,
 0.9282589971614204,0.9237237838229675,0.6166190257507225,0.5901319164916484,
 0.5145164391963013,0.055050722275966746,0.6899049251311622,
 0.6584729103881769,0.32382925924014083,0.8673925666200396,0.2593871966926057,
 0.209252061963627,0.5699579208003056,0.5095326961490321,0.04742931546382756,
 0.7871785202917184,0.5691570708646383,0.926952805460965,0.3220154580370138,
 0.9153519651635251,0.23404058172104292,0.13046230850257168,
 0.09303312190445079,0.4557888377762983,0.9180669323036936,0.8354294165644903,
 0.3223659892251143,0.4582868607311288,0.9247326441643061,0.23890422222225338,
 0.4912824519962319,0.6793820416925627,0.34669263397519856,0.6668806161749525,
 0.71836228024436,0.255525585269866,0.3028252987463166,0.8173245266902978,
 0.6732432558251021,0.9729914163569602,0.8561344189845251,0.6560006037864508,
 0.9467628486950918,0.31509419496176494,0.5522203372864974,
 0.14842135851755411,0.6871847583259341,0.970170330723834,0.7535035025938116,
 0.3419006050782385,0.027464123462012324,0.3411187308136816,
 0.3933860405502474,0.4476328752629287,0.15801280288529596,0.307120787686635,
 0.17379986262673452,0.13846026247354315,0.16056770241169938,
 0.49515222426393657,0.7404775569566956,0.7132825828852465,
 0.35072537488559685,0.273789224601074,0.5271307103562668,
 0.058506692132131666,0.3878615216305058,0.8154272178523219,
 0.7362140589188169,0.5744061208310708,0.9536434952972532,0.3173499836237752,
 0.3135081782421437,0.5046721244417387,0.1728084962210299,
 0.055398136484511085,0.48827611351792044,0.13899490345401422,
 0.6454714489639587,0.972261813959894,0.24432161762335247,0.4368451178495125,
 0.3160134787701603,0.18041963958123186,0.6318498056739372,0.5158011485654528,
 0.5564030302174665,0.369330826940403,0.06267398666322543,0.4949604907989139,
 0.2839284552779632,0.004908422978413651,0.836275571849266,
 0.49908210062683633,0.17688238230447417,0.33961387303276946,
 0.711197369414652,0.7268758302010814,0.9082478206793927,0.5444980833578592,
 0.3033776450152137,0.5882174213516879,0.12916308178213232,0.1945337277139657,
 0.6417395395817395,0.9006162727041793,0.4780032550631328,0.8129248253753976,
 0.8033662884065831,0.19860637182993668,0.39385336715400654,
 0.8249805585115553,0.922694946381077,0.029308031559502457,0.5069666692073328,
 0.9035837610088451,0.7397632605952696,0.016379921304370826,
 0.4080401220698807,0.6751283931255686,0.8670835379234308,0.3634742138512572,
 0.8777503201631001,0.5905509904656048,0.4016548855368255,
 0.039028501543228455,0.1616435133693208,0.35613989908612287,
 0.9918745178275012,0.2882276397382094,0.22193943117313242,0.2503793097703475,
 0.9793765364105615,0.5600012050188281,0.969978381334354,0.2546345008220997,
 0.9272293838299338,0.7094306428320019,0.07887331429854405,
 0.10505711733482603,0.4951141944359091,0.20317627665866478,
 0.16753182308595949,0.03853779406333946,0.6896690218088881,0.896313206666222,
 0.5411956042866501,0.016352063858091515,0.32281441416032275,
 0.7250832055944416,0.8749065425476756,0.9448012699315255,0.9034491217576448,
 0.2820824934089582,0.1102682141306568,0.353197520165068,0.24151174899498684,
 0.5462533123718262,0.31544565370976874,0.3694460049920383,
 0.21068580068785403,0.5631636683092952,0.7868142731171738,0.5308295969272073,
 0.20022888504036884,0.7681482263107936,0.37921351583829455,
 0.06387673656212478,0.5977011594574678,0.2389978837720652,0.4063255877186547,
 0.8118486011976267,0.015780835991223374,0.5519826289752175,
 0.2474625530097775,0.6025827474396165,0.6295609522582375,0.5835943800316225,
 0.00834817574002189,0.7428029215905126,0.9983410986444083,0.9562559691175109,
 0.7908791688131671,0.4469771932284674,0.7420996185584996,0.29430285468295647,
 0.4438186552677065,0.026510849079563248,0.6901465329337082,
 0.05615823941671483,0.28201615787744483,0.5361638908502948,
 0.4833773808824019,0.29195722553311354,0.5292457599665776,0.7371738210000929,
 0.45591097014213755,0.6245664341307755,0.5938320542518323,0.5503409649422696,
 0.3883648429164337,0.9209544102688494,0.8995318739501903,0.4513980724354142,
 0.4609088125227033,0.3324842807498619,0.9063919051937286,
 0.047415573644402964,0.3985451250556218,0.8648538399007618,
 0.3793538568332253,0.3595849397941351,0.6210499120993334,0.8334463662328684,
 0.5850436342276657,0.021228327055079044,0.9679279127527577,
 0.8469579550635078,0.7280412885799707,0.35390882873926366,0.6746615175968548,
 0.7389793303905158,0.6698645694500516,0.312832475632851,0.6113285538212697,
 0.32855809867861263,0.4646419066368238,0.6731449443241401,0.1989699619559786,
 0.3454094395659778,0.5765976839222472,0.16215249866084125,0.858123598235181,
 0.8650758376957617,0.5197241608298473,0.33868167052834675,
 0.22118442859820076,0.17436310476923567,0.32973166902897777,
 0.036667256042626106,0.5210645719793527,0.528353937477538,0.3299604560537759,
 0.9742010260535413,0.9062449596337392,0.5138605453723649,0.9048234642274583,
 0.5981293753631098,0.420593972203238,0.8536217513284116,0.9728293493092067,
 0.7735099650353979,0.46390333169178255,0.7405799173662267,0.6096225370108791,
 0.5516896865683699,0.7592534923173151,0.8962450046185093,0.9911042696348022,
 0.00980538892195959,0.9849585506323731,0.14538685607635993,
 0.1668413013608312,0.4045740388067891,0.835690948609608,0.4640874642864583,
 0.44763419718050634,0.48674937019558073,0.29500648264876594,
 0.2706833760366012,0.31917940594292193,0.5307569082026988,0.9689148029686078,
 0.37207406576747093,0.42286408471980264,0.7158361299317073,
 0.19813605543286272,0.589361404106923,0.8089518193520943,0.468135450803685,
 0.8660820763682456,0.48076033992815204,0.3095472583126734,
 0.44322163985119034,0.17628632245175058,0.9806129065615642,
 0.1501917043619163,0.11670639775621727,0.3472150523040638,0.7334491601737091,
 0.23745148350017953,0.9243569627836177,0.5691052036092051,0.6722144313266158,
 0.8265256323088194,0.8570277275585579,0.24762436836150248,
 0.21083527891735077,0.3539739252899906,0.052466249066968174,
 0.004862287862226422,0.7654979698592688,0.144818161674802,0.9860760025767266,
 0.6311388955954895,0.8639337279943385,0.3932443854348088,0.5197848922803135,
 0.3121184576041063,0.30316519938608455,0.11100318592428748,
 0.6130921120742677,0.91271881970551,0.8693896911627713,0.19570798842393644,
 0.7792572959300105,0.09370382850490933,0.28681037866794556,
 0.08160139447418469,0.46434065563801086,0.9447599583019763,
 0.4691474425303742,0.1669896156851548,0.16014269403668646,0.5001100492381954,
 0.4663025804054185,0.31055071309677396,0.09843016196457433,
 0.15788190530811752,0.6052892614362083,0.018412605186700293,
 0.08758336641555575,0.7954898706903983,0.9444013041592535,0.9909862826535754,
 0.13149852707192355,0.6776318993822729,0.13866213911588332,
 0.6685457595166655,0.2275983430183668,0.3692353320451458,0.6556952560952026,
 0.5035159345779614,0.6847760590369815,0.5914202265919382,0.851772892844702,
 0.7352250003122531,0.09482416155739959,0.8958386431469323,0.9239134963000257,
 0.11361680900911653,0.7598227455011078,0.23577843396057263,
 0.31908555286393525,0.9639871527550696,0.26604474259588395,
 0.39628296426402176,0.38128066304263686,0.09941737520916194,
 0.2810935174347352,0.6855095892618422,0.22242702535009018,0.9275307646082896,
 0.8636260680322901,0.5785143155290181,0.34580911870958597,0.6648964664574588,
 0.020619363279222203,0.26332617424800653,0.4684976722029096,
 8.0566977589136e-4,0.44863092736492993,0.9667879077056462,0.1321933586669254,
 0.9226844836821183,0.45069992068183107,0.30081447439375997,
 0.5863294027417871,0.44154655665104137,0.516087033013013,0.2921731138291369,
 0.4523715987701281,0.09768368911100556,0.27290898120471974,
 0.7994221507099866,0.43769534690031164,0.2412272944243229,
 0.18596976418044553,0.24547788016764804,0.8367573505296435,
 0.8162043126088632,0.42562026341674464,0.9097898552065625,0.9090534721778646,
 0.4425379666885889,0.4995896953744132,0.5143920282454715,0.19519428697437724,
 0.8336399503216394,0.75258849079102,0.9286728474087168,0.5908532107675871,
 0.23074242289695424,0.846867728414779,0.019991256672220237,
 0.5924277430039501,0.54298314973765,0.9900531617438431,0.5787788256137149,
 0.4599639929544733,0.6162171615516648,0.8376442136167797,0.5354188338895536,
 0.5451364383275843,0.9996027532667069,0.43673925994107465,
 0.49817100511168844,0.13626444455184328,0.4537365611359231,
 0.8219767583253966,0.6777731796464123,0.3748829599881994,0.4894207112326693,
 0.4553331140477628,0.7708562389338227,0.46207427056821093,
 0.29405006468052886,0.27068596787327437,0.3741601898675264,
 0.0051798498956239225,0.9400408221265177,0.9363936220392141,
 0.7009291698416903,0.11884271034614957,0.477800359445146,0.7611659861335761,
 0.919827531381813,0.033127756901478844,0.6930837679754264,0.6825700510494077,
 0.9654804298688993,0.7978478965735346,0.6137775219229061,0.2214083213510606,
 0.316049366034692,0.5486569152551393,0.5677672278946515,0.2641378812458923,
 0.8633948263852079,0.41369711328361636,0.31185557840472833,
 0.5215837588066848,0.4706391192006988,0.43397219544468746,0.3120721726094756,
 0.9700116374833527,0.23586274271115382,0.9330046951001595,0.642024287611885,
 0.37886423067012975,0.4378408285636707,0.9102574799954373,0.8601856013297278,
 0.7332838617399811,0.8474254559219419,0.049020192261562734,
 0.2973436606953671,0.1508957328185465,0.9960141007257951,0.6789774103226163,
 0.8886464345430201,0.6052798694911145,0.8137698140971279,0.7207602077089623,
 0.1251210420162825,0.12410080827623471,0.7813169601316613,
 0.019970617045133143,0.03210344978903901,0.6328609913643384,
 0.06784420455195261,0.003847061427492915,0.7012842964348507,
 0.4539890472843684,0.7589539169237591,0.32167442528670775,0.553581861455568,
 0.7604393759885543,0.2617707991967886,0.2841701259938122,0.7676042363602553,
 0.025542885040551133,0.8261109109077492,0.31438779015260176,
 0.6154291494136492,0.9280449690354489,0.9001989470240891,
 0.034085601107215036,0.45979443438614853,0.01792951918212471,
 0.22886102769029426,0.24605154527322637,0.5622094030094993,
 0.6480587699418133,0.9143141384277123,0.2029089407240532,0.5016026097970192,
 0.5366310686386868,0.39493226285906113,0.5186803583479979,0.1683702298723262,
 0.9921644090193404,0.9922321445845921,0.7620631819676509,0.7041652145313593,
 0.08378327177118827,0.4737864086512703,0.6722413475609847,
 0.03354688840055242,0.4046656673694671,0.377680845721182,0.32356047417109834,
 0.6382802339142375,0.46153392528111536,0.46770723266799175,
 0.08238694180760264,0.2615448046904262,0.6814098873418202,0.3594472204080358,
 0.10863411703131154,0.7199910101502743,0.8670790283833439,0.594631817750905,
 0.07842146557085383,0.5113526567015884,0.6067474375021022,0.6701532441407944,
 0.46919112380727923,0.1138581530174313,0.5763067506893036,0.3142372427468493,
 0.7764288086279326,0.6835507983159075,0.6034430690490135,0.8206050112714448,
 0.7579155028434774,0.514713817925264,0.401655538929651,0.8707546057167316,
 0.09591249713024985,0.8259435281886732,0.9065847079308796,0.5388267184450466,
 0.9384317971261579,0.15539310122522365,0.11978843532410743,
 0.8205522949738207,0.5594205020806733,0.6786863893752786,0.19467351480018602,
 0.8838300017081001,0.7736592269070779,0.663890471829487,0.08787616001611287,
 0.6523026094727746,0.4529209883030616,0.9667412342545894,
 0.013108798452991177,0.46519686557409556,0.9450573806101477,
 0.7257169660915423,0.6289352211512959,0.7495354251937313,0.6280041963426262,
 0.22824587844383593,0.43771053035194074,0.9480975065338713,
 0.9883438716845657,0.4111187864641642,0.9112657327902098,0.8041050154356093,
 0.6969820197368621,0.09849833708244027,0.08766918158790493,
 0.9265602798634516,0.07005109280390853,0.7132773149738618,
 0.21196200048999725,0.33676758183791244,0.31115977102040404,
 0.6723243942554074,0.792966253120809,0.26226947033772785,0.35213929471028227,
 0.6889134154317721,0.02647064593303383,0.08333404095934815,
 0.7064779318433204,0.5430249276783888,0.22056999109262765,0.4481827137492701,
 0.8623261439512127,0.7921915575917166,0.8897132941594769,0.7160525396181674,
 0.6356448119056325,0.12203671375028302,0.8419624482700555,
 0.46763537810426026,0.6501596570017824,0.8951902865015842,0.8368730787832305,
 0.7172256367085411,0.37318484940154495,0.711821648385751,0.44737577098142056,
 0.22992351010490508,0.14716662500968025,0.08435255461497071,
 0.8203211785887735,0.19526164007084257,0.22052303696824493,
 0.24635216168328522,0.8948256929114315,0.1421941736413922,0.8037676733298557,
 0.13412155770369494,0.7546300025171755,0.5641501848831277,0.3208623045379845,
 0.5991749494756582,0.37282203964706884,0.8590472230507461,
 0.19301508014641233,0.4694271294586453,0.46996494594758453,
 0.7332407440889289,0.7534922915156184,0.4133689918601676,
 0.062438651928494626,0.2779281787642026,0.37164044045507627,
 0.9087789963943124,0.9790167173902382,0.9484642121983926,
 0.009212734198839279,0.44794460184013063,0.29146517600173827,
 0.42678641339318235,0.971234722563299,0.8936363728700851,0.528283160040572,
 0.8218606605772172,0.37041014347503876,0.8154829687237802,0.1206555201827545,
 0.9150116718975798,0.7519036986773769,0.672925434403474,0.8805702149971784,
 0.8026684933265777,0.13516333941531644,0.7343021226500107,0.7295066216222716,
 0.4245663465181816,0.0469596456478365,0.5401364283423371,0.6474800869615581,
 0.5522192774240502,0.9838033145323994,0.087917271186578,0.9559669376476619,
 0.12012053691005287,0.07984596992850945,0.35299153660403715,
 0.715488210692183,0.6077599123460129,0.8851013541414543,0.44804660543228025,
 0.26695011699349247,0.5319812446043105,0.2600664676198161,0.5187933780151872,
 0.0037673787748100374,0.750469733020025,0.9783915741543828,0.157633373208248,
 0.21453555230069044,0.01813510415099595,0.3846164784014996,
 0.11933048596993168,0.36279196297611815,0.6685441387281789,
 0.34846770009004024,0.25198897888912464,0.7529904426773112,
 0.7251812934097022,0.1993846915585633,0.6001442072442165,0.10664409802243202,
 0.2819641115436792,0.9688199738593135,0.5557961810074636,0.3643309978853333,
 0.5463089012265823,0.8216182854826561,0.37646217207067223,0.7872997932183533,
 0.8937263900660497,0.2603203689686593,0.3083741105011064,0.2043115856676836,
 0.003965654633735216,0.6149190923122805,0.19350438234293588,
 0.667467485091027,0.21400717283805282,0.21747087610020088,0.8936742453790796,
 0.8004481887399947,0.7974775384220368,0.7879583206874469,0.9902784485022325,
 0.009340240067152505,0.024115181529852703,0.3467307907470737,
 0.6454267552910732,0.2683047181110907,0.673517098597951,0.6049371724019935,
 0.6221793229886537,0.7726444510213388,0.9759954274748568,0.7509606206265124,
 0.5435685334831397,0.3275665213773459,0.2183130019099504,0.5727046719889127,
 0.5520108126248118,0.23312051011914348,0.720798705297051,0.679581860267195,
 0.9912491694232609,0.04297226302712531,0.5685209110841584,0.9355220296590984,
 0.8684561898918524,0.965289190638085,0.3378815929886283,0.7484431785584486,
 0.5221048665227273,0.502153954930943,0.39579603631622196,0.20272507646060078,
 0.9815679139993453,0.419095548980819,0.20679546301055973,0.9863420372591218,
 0.23696054967390978,0.36210580966222605,0.35432748017192517,
 0.6848773614445614,0.5385786399850746,0.5914254831994364,0.47878089293910286,
 0.022262028658808997,0.9131862808161955,0.6571818124251233,
 0.0413023835323062,0.861442780846917,0.062169582990218686,0.5553038240773123,
 0.9774414131243838,0.27110936736337743,0.4459183071122528,
 0.11670042572173489,0.23831716309904305,0.09837010990188289,
 0.9607360766287933,0.9940852603773477,0.631776808990248,0.006469646219178005,
 0.7210479490330808,0.8266697780150223,0.13033180488220975,0.5514409372422614,
 0.2771317784310128,0.3397358270391586,0.7848279681728303,0.9598678326255486,
 0.6683615078693601,0.9752442055492337,0.7660286487170188,0.04432435303776738,
 0.514476790062721,0.6748474087626978,0.6379475830668824,0.3686035483023311,
 0.2792244951042634,0.49555407000486296,0.5041256169075007,
 0.058005499005692296,0.5226205613575128,0.9811989362346977,0.340932216102678,
 0.9514730831349869,0.44378520431238644,0.44072981438410996,
 0.3666154820487163,0.9423775263455559,0.12062789809759111,
 0.17671132216955598,0.9788927679451221,0.9946024082106075,0.7033893177014023,
 0.3549721606758274,0.9552448311505125,0.07497495726958714,
 0.036698709731786905,0.2364135911046139,0.3358392601006599,
 0.8852063415734227,6.160927264894367e-4,0.5873446448598514,
 0.4975883221640085,0.32089699885337464,0.6179259113902825,
 0.21737713302209027,0.3998883433432072,0.06845152214078554,
 0.7417922028038642,0.30856628032331734,0.17151559524017912,
 0.14130559918786711,0.6154350532780809,0.4345932473976012,0.8781869041600426,
 0.19670155096323705,0.3455294175649287,0.9973319003630838,0.403508367137114,
 0.7160165966044143,0.19794464769442444,0.12863908432647853,
 0.10197828853379398,0.3392133277660948,0.7630131464442598,0.783474946919505,
 0.44686378493019596,0.4557385159919124,0.3333711383718133,0.5119570690768367,
 0.11710217577665061,0.149956753769115,0.2719460627594572,0.38479810144678694,
 0.3094030648202617,0.3960050616627877,0.32003209830377766,
 0.09964311609711374,0.08777868723589599,0.09621186952966976,
 0.3029340443860906,0.7724311897685237,0.08719594812508558,0.973543099894262,
 0.34956167254383774,0.6884774452585918,0.9708056419401009,0.8122528286579379,
 0.6696653899779859,0.48632274401969156,0.8369137944203647,0.8084517587773418,
 0.0015688383446158527,0.04573875308925446,0.661361515677168,
 0.7793748966741078,0.8386188540969456,0.61863321015048,0.6648677502591568,
 0.2282371736660206,0.24807031662196632,0.20092568471069328,
 0.3094712101561117,0.45545543901374574,0.2044171658102969,0.5498225861835763,
 0.022875656275684775,0.627989496330838,0.04065373832447616,
 0.2240708370951433,0.2835346146507439,0.6986950797637839,0.6043120537565801,
 0.11069342390331938,0.5519591247790091,0.03889276399100594,
 0.33310534828059857,0.1288235586589177,0.5191251604613976,
 0.14405640505155914,0.32143591023264084,0.7568905165825027,
 0.6756744941430388,0.774087604239773,0.9978816611292187,0.7599011114510472,
 0.49574925246813406,0.43057676029481173,0.7755811371298791,
 0.15000863270454867,0.6393995693172417,0.31209355073982104,
 0.3197603101584585,0.38941890144481217,0.5687954936116346,0.9514269685731843,
 0.0909453148822793,0.3102653680777029,0.6269257629883466,0.25844399147200037,
 0.07228857624270213,0.349916528242079,0.7562162910887523,0.9166318483097541,
 0.4920701757965533,0.32261615651448494,0.5148284040197095,0.7847296439356511,
 0.6354139767965972,0.17804111273389578,0.9324402966791854,0.9336446045050129,
 0.9615916933585579,0.31808684761519157,0.4325204353695602,
 0.12081706424739047,0.921106945743451,0.6464963337705252,0.37067879077463417,
 0.5191230846576603,0.9942982868770345,0.9391511259151953,0.3590853569230503,
 0.025333271061701468,0.9690105310525078,0.4148106476983039,
 0.10917353839813326,0.31150725600797624,0.8279628449948178,
 0.37558572162767556,0.03315610332962593,0.9093925909583047,
 0.4466111613359458,0.8215194394077878,0.7169698186220084,0.4229259750551203,
 0.2959428008663333,0.7439764239039147,0.7653890854326126,0.9768207933353747,
 0.6349458105585188,0.20782809564890203,0.18498619685890083,
 0.9603733550675314,0.4842367797869822,0.24102901770718455,
 0.10795166804614664,0.14552955301818693,0.6360533547435031,
 0.9073156125882614,0.35143782134331236,0.01589486048420541,
 0.4508893030841663,0.23357404344457633,0.5145864950666426,
 0.28130558514128867,0.31690126005920327,0.7969871094655002,
 0.4643597842773105,0.7822576631969014,0.045877092485448134,0.837396374023389,
 0.41130099728675984,0.9188145248062208,0.34540008629941144,
 0.004027270462483523,0.03758898346014572,0.12847258727472866,
 0.9103841372797269,0.211688330853658,0.4097411582626269,0.21123906891181865,
 0.0010154615436291925,0.8325610823313363,0.04782704532629656,
 0.5441925936338812,0.5397779665603992,0.3711376547874641,0.19608004720901295,
 0.6671142497743588,0.33397038906901455,0.14018615046310934,
 0.6424171992857619,0.0077955207242873525,0.6396189852166421,
 0.9527308646833058,0.8749457217651648,0.401545548976616,0.7581548358026482,
 0.6236961230986791,0.1712825383126697,0.6147763410682519,0.9092571604554941,
 0.3129424896664824,0.5270823056369812,0.7461111229773597,0.04165602660213419,
 0.5226253491722123,0.7571725385937275,0.8076647772782806,0.5139732665392162,
 0.5182841382154966,0.825337350287206,0.3962226323771868,0.1085188973418647,
 0.20561624543377977,0.155087716756089,0.635014563142203,0.8824391377238807,
 0.611314974915871,0.2837478719149197,0.5805214637253255,0.45739876225387777,
 0.9626068771231948,0.8645129545509634,0.8944484815456367,0.2822056513395005,
 0.5575987402982847,0.1740955456726302,0.6093420518904719,0.198453274468939,
 0.5281853751933381,0.5860278277479094,0.42491712672283743,0.821255763875033,
 0.04942289003339273,0.9257348185484127,0.8153811155106659,0.2962126018473348,
 0.43930475364149457,0.5091596048591013,0.6527904594138678,
 0.38362652118749474,0.24248373902223908,0.02455809002908027,
 0.8497499634774398,0.5965101249599566,0.4561741374224816,
 4.978712522486495e-4,0.7117594247956724,0.3099728552432862,
 0.8746143334417139,0.8413178510142778,0.43224328796214584,0.7642784397745379,
 0.9421372254180916,0.599819006656824,0.5301803243505021,0.9754594297805206,
 0.22608525129567691,0.781173696887841,0.19154805734629043,0.509746064631144,
 0.07948560204239752,0.36171180076432063,0.4910363580510064,
 0.08131574721386514,0.9255656527660051,0.9258201884411286,
 0.054527757889742734,0.07350768868857649,0.30756322970600225,
 0.10666532115742311,0.8455247630631807,0.016297829286945675,
 0.6736546972555887,0.0026641919954543125,0.2839375688005108,
 0.10812478035288053,0.23973918492929824,0.858727638860805,
 0.23244744835524944,0.9172804086034428,0.44677891635449407,
 0.7234036519762193,0.004699029707503799,0.2647418674428256,
 0.4908151670007693,0.4304985945889439,0.08313150492551369,0.5996209904642941,
 0.4555174670924056,0.45152610119614034,0.4623082932858855,0.8442219906782961,
 0.6224620982854429,0.7028873141210858,0.91100503580373,0.1973275956666778,
 0.8803363853960509,0.882720659556622,0.9853481802911007,0.2223814051840064,
 0.025124215973152286,0.708946192729284,0.2719569829153312,0.0861996946369552,
 0.9747934631834386,0.5441129778307134,0.7189858483561787,0.5376486268744005,
 0.5582820792045261,0.17933321213954845,0.37019642437921096,
 0.18578982176035808,0.7566422700584684,0.9479037992357273,0.1661695469665102,
 0.7027836049493941,0.2913346022091523,0.4919402307272175,0.4127724574427212,
 0.788553129911838,0.22960352525395156,0.3402289597176189,0.9615762973135453,
 0.3216401328301932,0.5829881627457714,0.1222199364903086,0.884166133842978,
 0.8610744826018604,0.19418751997543615,0.3030788244316125,
 0.41969213359101987,0.2435351691771691,0.8922671668720911,0.8926880221113836,
 0.34880259394994595,0.26205383646347613,0.2741243243326513,
 0.5876427333835496,0.21147843947311795,0.2380742035372564,0.790769350741663,
 0.8063741735955612,0.9368481698247216,0.5734503353519254,0.6696912873127112,
 0.3485741855495431,0.7782321636134052,0.6406437874137965,0.2053375280160813,
 0.3186679794073106,0.39795185822627366,0.04999713015414986,
 0.3402103167723023,0.6801903526510444,0.6561125455860203,0.22212984195371255,
 0.12224944834955387,0.558578725439175,0.14443835447404252,0.1348174106914335,
 0.6716537653267327,0.6042539900116751,0.9070927370425113,0.8470719931231254,
 0.1756993523004673,0.6506730811912537,0.6424177494105214,0.6970703741014251,
 0.05376496793290819,0.25970562114684315,0.07808515159366503,
 0.8931276837239007,0.4407337708118364,0.17709865233919142,0.7354527850235837,
 0.020026236873672554,0.45621986592648767,0.9374701235362662,
 0.07727876042213855,0.8098418407711905,0.469393918034814,0.7977025056844652,
 0.6096681535312413,0.7152725267632152,0.5169997674603211,0.5848836238195799,
 0.22080597433588545,0.313781102513915,0.6075691730491458,0.40572027169681146,
 0.952283133961388,0.9693932340682544,0.9290191620389364,0.5932391041303617,
 0.35998853378707263,0.7729754697494498,0.0960162173197524,0.5943280195807302,
 0.8468500956777582,0.6448865601676873,0.6379963516348219,0.27161582004353424,
 0.2665314995649234,0.5345699521841598,0.03179249002325579,0.9122584251999765,
 0.8379669415265729,0.6912846179443318,0.3645035311506045,0.4179645258043073,
 0.15548980738065754,0.21218698096892497,0.5019360288590066,
 0.10906836272016918,0.9932680994143854,0.7011400893748774,
 0.003984512302741106,0.9050565554036692,0.4488195631594938,
 0.3496940949257564,0.758782868062147,0.27102046238122324,0.13961380677634927,
 0.4332573875011715,0.041971932147474855,0.6314297286870949,
 0.8116139426834559,0.2485631621371367,0.9659732774272572,0.17961492806760893,
 0.2455577215352246,0.8528881819468818,0.6438687655370593,0.5273491096709267,
 0.03313889651921098,0.14561025544288275,0.5966602310449127,
 0.6287949541491826,0.04056709977713746,0.14932569513996707,0.495491418149022,
 0.7531741879528848,0.9728050745475376,0.8176446879140592,0.32249073680763973,
 0.39479133321307636,0.9731155624421931,0.22039785489664265,
 0.7661409007353104,0.82429264822667,0.4203833809047006,0.4026087582368246,
 0.9373169912100285,0.9722616032481866,0.16309178984361916,0.5715624086927901,
 0.3875002547535249,0.14833844533992457,0.5280471528368766,0.6195705106645663,
 0.5840318333009707,0.2465054336037269,0.789271419583693,0.6908453913694488,
 0.50939978778295,0.17095296274267913,0.0894740042551474,0.3954789813525954,
 0.29500586097554615,0.6127965012648969,0.6456860087604921,0.3183178070094985,
 0.9283152093055467,0.4302906740622441,0.016418429411701463,
 0.8458146781315913,0.08600717156526949,0.2509193121330944,0.609499662115309,
 0.4791307543213489,0.8294559580477225,0.3670467285321426,0.442372433982789,
 0.4745912820748647,0.6921850495180388,0.041705101584942006,
 0.27315056162499185,0.4476479722701492,0.4130636348674722,
 0.016681510471194705,0.8660944895058029,0.28346602044061886,
 0.49251308708063546,0.018837509723312884,0.06879995396220773,
 0.7523477824088444,0.3379123134174934,0.9847345702989447,0.4533324867771723,
 0.44320778813199735,0.03833151170124416,0.6073256060544417,
 0.7320287643389769,0.5848652353275582,0.8795554706393679,0.7473542788362431,
 0.57407643470514,0.4113442931332203,0.011724014369794644,0.9457595825103436,
 0.11572823116012365,0.821868077850553,0.9890158065203513,0.39799685021562836,
 0.8771989237589315,0.21445544629433622,0.8059887598780172,
 0.21758978215868108,0.7160026532813143,0.7901415419605383,0.9319401101586039,
 0.6882313171703334,0.10022663537764798,0.33265770600571987,
 0.2701779889673718,0.6187150813003429,0.28236830642140975,0.280384278887409,
 0.3323765891868091,0.8464959919686453,0.0883286032410433,0.0368548774795423,
 0.8891437945985943,0.34888555227975704,0.6049607726770644,0.2090267563725985,
 0.05812779808056501,0.3998746925674025,0.6278453272926325,0.7472890677410081,
 0.43663473883655135,0.11777927483517625,0.03072655758335774,
 0.9224188722855993,0.7981906193118437,0.8407079958126066,0.25199234902151035,
 0.976559839080738,0.19351463576310568,0.34443047002905436,
 0.49446173146893413,0.08268113518615672,0.848765842877872,0.853500038619109,
 0.661007103943725,0.05651264648081389,0.6874013954268028,0.8874626758440174,
 0.4395066742707966,0.45213345364649626,0.26272857784535986,
 0.9617841529085231,0.8281141143086029,0.5207815054052554,0.9327341394029403,
 0.7202747046404893,0.30474390491639514,0.48860017299174563,
 0.6625488242255302,0.1680660305504189,0.12807459911093289,0.8426674793949793,
 0.5408504405954334,0.4995986571700496,0.2438233278092523,0.8981185749018737,
 0.6331340338026434,0.37247391720220224,0.12291119163224384,
 0.8933107523860849,0.580030646572082,0.8032126283803884,0.2599497021442859,
 0.07116692542659042,0.325537858912869,0.797614137646737,0.15726284895713127,
 0.1633444372497227,0.7259085661285247,0.3552947907061259,0.6965795286458215,
 0.7478872358241874,0.7389072711829358,0.484239989826603,0.1077020973740348,
 0.39347261291456026,0.3243427841095067,0.8194304351004468,0.9493294137829591,
 0.9159658509286044,0.3905168344513912,0.7373573133634198,0.2148964170887735,
 0.10219346613995273,0.13964041352718115,0.2651508713680212,
 0.1945701462969307,0.1560706431647304,0.9950985861910019,0.8225460486234828,
 0.38474106717140055,0.29210208471125876,0.5376975674991272,
 0.029360480365731134,0.7086977301864842,0.4861598536064127,
 0.26191641204714255,0.43651502917849505,0.3331947761455746,
 0.4098874410531542,0.8501902335858729,0.8831872364148567,0.6373599315505846,
 0.8749374964625874,0.17134972094451784,0.1103146308342482,0.8571391146809207,
 0.8142954325928206,0.9332688845060834,0.982998076583397,0.7588464125616798,
 0.6586214651848531,0.12022521199844804,0.12818853605450986,
 0.39559487829519924,0.956399944473159,0.9433927051767096,0.8166835991386499,
 0.35149342130335093,0.37931433416465044,0.9738375745317585,
 0.5026301645872033,0.6970683498650664,0.7021402727438326,0.1487050651371361,
 0.6761780601713463,0.774639736562965,0.7719236752341709,0.3022840531768207,
 0.7285249194137602,0.6935257429727386,0.8992290561440973,0.13583905059465617,
 0.1467497155245021,0.10265984174214693,0.4187582288817848,0.6454779257223655,
 0.20091009943656624,0.06617273727740447,0.42319168475525915,
 0.6275128740215661,0.5847083968931526,0.6911272643811235,0.7277507791500205,
 0.33355206700409334,0.08158710894147503,0.7899625896771771,
 0.3895704411716657,0.6742053162456122,0.9665555133829606,0.6312313431199768,
 0.6407385306772401,0.8140380100309716,0.30904931006996705,
 0.06629364826755035,0.040143547544281066,0.6111746851210457,
 0.919629979546151,0.7142393499283195,0.05606163620089277,0.17089902879206065,
 0.7986377154246538,0.36143597527937876,0.6785702177379073,0.8148459870664039,
 0.3296495801475178,0.4446491507703261,0.07144583642906777,0.672748011682549,
 0.12217197990274498,0.47372507723945545,0.19824484030986356,
 0.3538242791945009,0.8550697936017521,0.5629457358286262,0.17711189090718693,
 0.725907715658532,0.32662814840298515,0.4041807726201424,0.09014744499261007,
 0.9964547300836393,0.004788611719475533,0.7476352807039891,
 0.13743839019090598,0.7120812719372108,0.518065814816791,0.7982033414428159,
 0.3345330981026411,0.28292919224657354,0.9809278592733398,0.621921372137439,
 0.5417566941345302,0.9982270718926782,0.2328409689836094,0.9675400726500989,
 0.4717320213966809,0.604372738732635,0.15258173934143093,0.9713275572992182,
 0.7018395012012617,0.6068134962020312,0.18437469971615894,0.7985208302625457,
 0.45526063114093385,0.7457869710796796,0.554206661275263,0.3584781369812797,
 0.2031597516011031,0.9319475929804438,0.6670977888155171,0.8498294747276173,
 0.9042526406125841,0.014555071275795028,0.5935807367402025,
 0.03981244544712037,0.08970766761718496,0.4898969243208132,
 0.8831632742697406,0.9332122599954509,0.243675306634889,0.23229862634461773,
 0.8245562141263455,0.4062190250609158,0.34900201050462865,0.4508532096087491,
 0.04564138818223129,0.17662336533174372,0.9670495492529084,
 0.8187092072224517,0.39848134597481366,0.368744367416581,0.42318339554900686,
 0.42961906266973293,0.038085713270922805,0.7426792095697792,
 0.2693432679115808,0.7035122869204573,0.2897993840134012,0.09381866527207972,
 0.17568951556569312,0.5496356651809997,0.09774488855685382,
 0.8878292409667559,0.35325372532659716,0.4015601736192572,
 0.31307942312950077,0.3929268085073738,0.9336599799718434,0.7008090605555918,
 0.9735279001357323,0.041142690357062084,0.3931958582291081,
 0.3487958222043124,0.3285791317821204,0.800901376178681,0.7120800252098463,
 0.460118552624885,0.6294200676651948,0.978960083701711,0.5383913271276306,
 0.9741387805806171,0.7417238091852204,0.6320465388316471,0.45356906637190253,
 0.9260853639810539,0.6781957589804531,0.589382176817611,0.16372236778770566,
 0.7974749103230993,0.9179523724686026,0.7880708889351057,0.8461376980224572,
 0.8118307714799566,0.93848574329474,0.4239852976470122,0.48798431288584154,
 0.6142355548924867,0.02533152522648152,0.9559621942950683,0.2658152988500181,
 0.9298875728086735,0.11791554219759315,0.6202081325145065,0.9926815293830823,
 0.7216066038139802,0.5243669789325008,0.8588752537761632,0.6273613017281856,
 0.46408189870115957,0.9032927200871199,0.5513498016442387,
 0.40295960488872884,0.05460738823534539,0.6432278714302347,
 0.9999693350936498,0.562053129776431,0.7106596576977702,0.7583095392681473,
 0.5277337772285334,0.7916241831885734,0.9783685661479791,0.5109446671677741,
 0.7266182747793043,0.6863386873086045,0.564577307843432,0.30188053137598403,
 0.2893390075052238,0.5207494439203769,0.5762903242796377,0.5006163168790969,
 0.7300744408736753,0.6253019351613116,0.32852837137832147,0.5611392124013918,
 0.9086145722530488,0.646048082824927,0.9933009779725381,0.3088610739718556,
 0.8535185440697987,0.22855755135896128,0.09955774714039278,
 0.5462807712992696,0.7536604012063168,0.27858420240186454,0.8674689342643165,
 0.3658894319500763,0.37960236945344006,0.19665698075077775,
 0.4928087990257509,0.8296935210884435,0.1869467830499928,0.4205912745946029,
 0.47867847056432267,0.1540804986417892,0.6834892278267904,0.8706904267535597,
 0.7058913769834233,0.3935599762902847,0.8738131898879008,0.37581793224188653,
 0.7499553280475605,0.2740189366167528,0.7330888913318293,0.6648117173679462,
 0.2501881302048844,0.2448429171826514,0.4702716934841613,0.08092948887248641,
 0.5829120787807294,0.020913772130820907,0.30620550016367976,
 0.2527262513642886,0.9668700248458946,0.2924498458128908,0.7625420657559103,
 0.24819164776156233,0.1732979390865087,0.7564380194538798,0.6378033759060597,
 0.9176820692696698,0.30832145599346306,0.710941453186857,0.14642536851332077,
 0.9188982373597961,0.5490252071099753,0.24826309858548867,0.7806006074230597,
 0.426323298738308,0.7934352622426727,0.45890047446501314,0.5794938360765581,
 0.28477428601877897,0.5413850246125141,0.8969729428719111,
 0.40672893595277104,0.3490024744255591,0.7057849352003526,0.840504663228336,
 0.2637699410273733,0.13798513073650498,0.22032790648267342,
 0.6876483015069038,0.8964739960016953,0.16119183594159747,0.5699121365196894,
 0.5947693981413188,0.652156209058614,0.730278177698035,0.26259591509395364,
 0.28243150445220555,0.3801546512387979,0.6447006539727091,0.0297275108506454,
 0.008709277624859513,0.5953379902467948,0.919771565369201,
 0.28872993249886636,0.3709215414843675,0.23773034342854338,
 0.8529549576863672,0.6231137949107742,0.7373713372291819,
 0.002506834591635032,0.8702333763741618,0.9348330601373269,
 0.9035973512172099,0.3729087435239292,0.010555525528495702,
 0.5349983548998588,0.3964478243400553,0.25471453968468827,
 0.38434389844932604,0.34185264412198346,0.7344708812676813,
 0.020218544788208503,0.3124660835629415,0.27373174464618744,
 0.5288579899104231,0.16646056183084046,0.9895443726524792,0.7111284753386684,
 0.17360588654485254,0.499606296833742,0.1779156092567249,0.12455807009017894,
 0.543345962459366,0.7088752245488779,0.12644091094968413,0.1903858162342873,
 0.6721972657144154,0.8551938445186261,0.3621341045455414,0.30981324501644436,
 0.2490841273113903,0.3199859921620297,0.6474284113713115,0.9092670385369075,
 0.0017414736848408996,0.24409239149276607,0.9661884836613189,
 0.478323477525172,0.8639515331847498,0.3444955063290741,0.3500415861072851,
 0.8530981426138379,0.9457601793672266,0.9232289890946777,0.2510773665992099,
 0.22138241113963286,0.22612245608260984,0.683917592795854,0.710967700241309,
 0.01198307579465463,0.5122383105019224,0.9570933569420511,0.6205057679399594,
 0.39751757276687105,0.19586351750019582,0.5969516457178042,
 0.5206394033174317,0.1097056640826437,0.4181364202858071,0.868594009289075,
 0.2192320376792698,0.7696620975689724,0.33090119753771396,
 0.12022028072444024,0.14764243351894524,0.6026144448019249,
 0.20268816796410294,0.6244949401249535,0.29265010671801417,0.363552717535219,
 0.1989599534867652,0.01582470990948992,0.31680426487831537,
 0.1618537073976425,0.5784992457516362,0.8983028704877621,0.6158506948641947,
 0.8809885659057121,0.11353720191361616,0.5404316011304807,
 0.44045343388715696,0.2845111958198898,0.0578689072359061,0.6176966834128509,
 0.20011137097180898,0.14798510213625926,0.24393892414924978,
 0.22945791401049598,0.6797602816621365,0.7842725265282439,
 0.09632973584783044,0.5303458250379759,0.6548720177472963,0.9607404390828095,
 0.6609484266247021,0.5157509856446599,0.15915580054869327,
 0.38050179316711163,0.7065305896000252,0.5085468421891083,0.4896233249468308,
 0.3570711154027416,0.4020482695546205,0.02664123125468676,
 0.33061615321113325,0.9435504696090952,0.4985963391261543,0.6727221099936169,
 0.058549574174662666,0.4739248357620065,0.039140884072716764,
 0.13328927636297705,0.4155166099292107,0.7209812117409351,
 0.32075326162756546,0.47266464905683536,0.2348967877069823,
 0.4243561580385178,0.8683595084720199,0.8610804169754764,0.13949603694207668,
 0.07520011294605422,0.6123214211264174,0.4248995830265836,0.9511052673960955,
 0.809240982058849,0.9315859388758647,0.6296954297964089,0.4034191083816383,
 0.8703023757907746,0.11550183366369682,0.008274406276249091,0.66039062641582,
 0.6828831098727848,0.6900652749681894,0.9936536550832329,0.21176703131033048,
 0.40277694226782823,0.54032018674147,0.34232849630589735,0.5269229680550568,
 0.2856571119205643,0.04654172411417257,0.11360018910822833,
 0.25554273222231416,0.4641460723960308,0.3548430390851057,0.9868300049161008,
 0.5878521149587836,0.08964345882864577,0.3995682204564899,
 0.04673233656625575,0.7851239819182534,0.37170207670309485,
 0.10511591928719966,0.1864935871898019,0.3463009216332342,0.9504208172784799,
 0.3990890512898161,0.8694655130454861,0.5353492404802452,0.4475410910247587,
 0.45566104821578235,0.8575078113609661,0.4508352992862068,0.9816816407913103,
 0.6810778903864829,0.8067526846646831,0.10264683493170823,0.5321814316678184,
 0.7891223280803745,0.5790215372290056,0.7887374943516321,0.7929963059887475,
 0.022009910910622432,0.755826576148287,0.4588441096396102,0.3607046354612533,
 0.2938196405794071,0.42542854628147864,0.7448827861203626,
 0.15328938471748754,0.586557047423377,0.8255573727300836,0.3062361268707199,
 0.3979795528529141,0.19238584031876949,0.23766252539412103,
 0.5260239442762085,0.16341298813030825,0.11004771982934636,
 0.9026582175398083,0.2388316569017721,0.8052311147026364,0.06561262351587849,
 0.009353955686823956,0.1519240616707347,0.8329656982040161,
 0.45056378934683505,0.02258321102048555,0.5572582241748724,
 0.8612880734281533,0.5098538128660762,0.8812720760488184,0.5243035489505014,
 0.32472237705087226,0.1794505308450728,0.6186957840778444,
 0.49541470076816496,0.7072057246947437,0.12675815469096807,
 0.4223735597671837,0.4864141222173938,0.24967640604772412,0.8721217784851205,
 0.11961736047630234,0.8312022419573651,0.9978170355244819,0.8012073950057367,
 0.7052945702339513,0.6083168792378737,0.41508504473732966,0.9666864951654526,
 0.648415893022394,0.6382331041413313,0.16714837980161557,0.32660996469402104,
 0.18171831368114577,0.8711575787663615,0.31607229453364816,
 0.30015654680042203,0.028317251013990075,0.5627056209742413,
 0.5084179078575024,0.8052195780735143,0.4095773593926799,0.5731971796914348,
 0.47314885314664346,0.2955789789758505,0.081879010948961,0.05602818036420587,
 0.194409230554343,0.23442214737595413,0.835584485118243,0.5488475350410928,
 0.7946971392264893,0.33293845569414093,0.3221853553316468,0.838051463104156,
 0.9353637962776713,0.02208555327792394,0.4269655292390757,
 0.36501319670636234,0.5921605870025891,0.6722176378082343,
 0.35447450827878935,0.8265427184796013,0.3432112757190734,0.692383294634499,
 0.9195307431142827,0.1752690661225167,0.9536038831408129,0.4651645934662453,
 0.48462515519869576,0.0420108789579382,0.908584910354069,0.09756263689584077,
 0.6629775365045872,0.1333715297877277,0.34218006694932146,0.4550779686819191,
 0.8098915657641953,0.6037792866730458,0.3511178090841146,0.8076321669164406,
 0.014174150802278351,0.06367414161076801,0.6381726695727945,
 0.17524997547859167,0.7953807485886691,0.1730984589926089,0.2691924886929653,
 0.6703137465769852,0.3155044988904673,0.9799335733014991,0.7109798447324209,
 0.24835417511451507,0.3561088071787367,0.4813041730082317,0.93840784844625,
 0.439315267664793,0.8017172235845215,0.2703103315570067,0.5793872552633347,
 0.49510703722713767,0.942618242397251,0.8859634662865499,0.8406704902213881,
 0.06660129938120607,0.1997026437705237,0.8204046050795674,
 0.25131767802772564,0.36645719955160494,0.4976263770808068,
 0.3009493492444303,0.06750740820924905,0.31586706246082374,
 0.8706230725358679,0.8437002105674637,0.4183566120106331,0.11251884051160965,
 0.32193742056349905,0.24069158361729548,0.6770422009647753,
 0.6852799213140417,0.7620353033367736,0.4854040451854368,0.16462301964351767,
 0.5229506492223944,0.9606167385318067,0.9394167705293129,0.9900666439009455,
 0.28370127876404294,0.5028898944773044,0.7607950430555879,
 0.49027548595493275,0.7428098773978493,0.06285046272069983,
 0.10980625465862837,0.19131340869596802,0.1352767198119026,
 0.35771537968013556,0.4219126480198294,0.7179225041851954,0.5071458157434563,
 0.41107336135155637,0.23323444781207547,0.17622609121995236,
 0.8845574973918993,0.7561905634226945,0.5747645338788258,0.7861970758842388,
 0.5038050214526998,0.9432531471584651,0.29948298581988086,
 0.07158838899826292,0.6336514604945755,0.23458871730619435,
 0.5996587228647641,0.8034258216674061,0.9884652395185821,0.28131662084115583,
 0.5155792715069655,0.8185626357962501,0.4268083904702341,0.625803898159971,
 0.6908732918509926,0.40223813428143196,0.28874556998463774,
 0.39427621426444126,0.4166761842729805,0.804600648153229,0.14477216360860923,
 0.3788195833384085,0.31812462273558706,0.7441996048577538,0.5614884106879245,
 0.589292328466812,0.25005809401010803,0.38619551882150227,0.5766050063878042,
 0.642734634059591,0.4601151687587376,0.784744907676414,0.774423428176277,
 0.03930417989207757,0.30035689015207556,0.4353271177123963,
 0.9142808623566616,0.03486371631702878,0.8872148248839483,0.3192902929709891,
 7.588544573267075e-4,0.4441829601166374,0.8736979645583479,
 0.8051041253623753,0.46526855080916585,0.9544399624794957,0.7081327611892266,
 0.35518588023699116,0.8627172985872921,0.3168377045543167,
 0.32282634092068285,0.8433355912282803,0.5820769971275966,0.8915907392402008,
 0.7592132801364064,0.7207925119246439,0.6131640918673342,0.5109534298895049,
 0.44794457874310445,0.12628103848579963,0.13366107481041167,
 0.4754498671587034,0.4090333335505363,0.2115387124101482,0.9314209015339343,
 0.8269686110413054,0.7285790250865849,0.081942106389292,0.4811666876508285,
 0.506788334668219,0.07707990079769589,0.40443558187862383,
 0.26246184595499855,0.9040197554568602,0.3408249099184414,
 0.09652018330287526,0.18184942067508303,0.21360679005223682,
 0.8138390183566253,0.2894019863698042,0.2458365148642464,0.9346340827226374,
 0.5451836282000765,0.4123751608186933,0.46111002153267744,
 0.34860716377657175,0.7140617832606476,0.8047451885027713,0.6949556315534282,
 0.11199426063077722,0.5213889054474208,0.18366871218005332,0.447289510651096,
 0.7262965084294124,0.8867924151355238,0.27107204399253615,
 0.26264130643747674,0.675976548528568,0.07781404545615223,0.5248904722755008,
 0.07618418913691405,0.26910633994528266,0.20416148366374454,
 0.9489784453453906,0.8861815022440322,0.0656031836245835,0.4660766532359606,
 0.5409049710553694,0.015172477669925688,0.6958392673495788,
 0.43279594777084784,0.9107805031733291,0.016282457359586777,
 0.3666469120533332,0.11329911312597385,0.39685657947495645,
 0.1260190981078877,0.4974815617811602,0.8400068356369147,0.9616337196396629,
 0.2346942467138099,0.8799209130913394,0.9163127327304352,0.6507102419191951,
 0.18454497398941683,0.6168123231176077,0.1956629922716564,0.9656606493134794,
 0.043120035608201324,0.6468965503979675,0.7037407795740661,
 0.7599997773028995,0.02689310786479271,0.8254425163699619,
 0.012465072364097218,0.11120300449146636,0.06744366088428289,
 0.12849924426419457,0.42337858776222037,0.5163497920021756,
 0.3226711748993386,0.8009994972941971,0.678705530104047,0.9687319954440818,
 0.6535824807190058,0.6797898569481248,0.8824275325574028,0.6828183318697169,
 0.8342874543920619,0.729889425424205,0.2038213292022325,0.7825359442205274,
 0.6180676846150814,0.004974581011716439,0.6614873936935575,
 0.4115687695503132,0.3180864527962446,0.9288174432724479,0.6747553690011573,
 0.07310067969190559,0.6668260235551818,0.1552562390658344,0.6767943965600348,
 0.3552027012216865,0.039900939894446985,0.38213276794431705,
 0.008044262335585639,0.8864316792068847,0.2755575135825711,
 0.10160317711027778,0.004339875051079134,0.5769333074295533,
 0.09976227785264691,0.2929034036483231,0.8346270131544297,
 0.14858354983032118,0.832449373707329,0.2289656393460766,0.46840534575347736,
 0.3054979015112024,0.614460786121323,0.26701199156140776,0.11314643993736995,
 0.6988898155233645,0.010595737912335479,0.7652078728983633,
 0.10406647533277147,0.3226219735354878,0.9752377052687065,
 0.0013744015282549649,0.1544356390623124,0.931278177974586,0.648942227857541,
 0.18319767974330703,0.5064663075120465,0.07808680859468808,
 0.9330717111736642,0.3900335559641862,0.0662498816492244,0.44799879201641324,
 0.3185554612855864,0.0346953122664615,0.8443726708143781,0.9188122690099422,
 0.3034817497758897,0.6280858864811458,0.25453208739088207,0.5421209593291183,
 0.5306781229220525,0.18582736385464005,0.7420007788183146,0.8515474490567453,
 0.3169135748732841,0.38772458252950237,0.6472312794408538,0.9672751003371802,
 0.633394117923902,0.009782123851486912,0.04631647708290254,
 0.6109091516834622,0.8835820756140141,0.19524467034759274,0.2943564529788567,
 0.34612731114703,0.1264374535715822,0.5942391180085006,0.17661793235080236,
 0.3819151441554359,0.6253697584611054,0.7057948165354435,0.37807199925287205,
 0.9045001071800876,0.9799728855770737,0.33672726902146344,0.6028761877924018,
 0.891341216819587,0.9466624456880207,0.44320992138174875,0.6438843849212968,
 0.723810293321451,0.2067712361799663,0.6499430922896282,0.715370540610535,
 0.22800080044155202,0.5486803884383825,0.5807901477211983,
 0.13420931802223302,0.6034563186370996,0.3484684535270418,
 0.16228235931368784,0.623706193701146,0.9792356042434149,0.567593933796237,
 0.9772042972012279,0.14598588706387683,0.14451372392706507,
 0.27350197015401456,0.37159651997238785,0.7038991052193547,
 0.2412458132685903,0.8387256470557246,0.33215525327302464,0.6660111425025921,
 0.9709546514107897,0.3538254816337226,0.4386865248939441,0.5854921269462477,
 0.3103436303399604,0.3679756751277825,0.27376153063294095,
 0.20499322386615393,0.8891427741248393,0.16119869480425597,
 0.12320765717198867,0.15002417393296485,0.3601042494290879,
 0.10536490068303561,0.3578884179524714,0.6110538758048301,0.2789881593125533,
 0.9359203128380771,0.7941379191475648,0.677439420799924,0.9278878432541823,
 0.2648525915641686,0.022185090958225073,0.6677300767303119,
 0.12080622190334178,0.7652466125102318,0.4909963348993144,0.917096362387795,
 0.7532574796511441,0.6776767372726548,0.325888039905896,0.06935822051496776,
 0.9683219124112289,0.05076764524144506,0.8202183021668799,0.4668483111674556,
 0.5084983642264812,0.8614951609931303,0.8184664791616812,0.9784568736105599,
 0.9247056975282297,0.3350692791348244,0.11949439183884569,0.644304223435097,
 0.7870540682074851,0.630177450596191,0.861964063268978,0.5940708362759956,
 0.45297532971404253,0.43196342356029027,0.05004046628740699,
 0.2511526459564164,0.7208397835691595,0.2858183451381453,0.9820710746847764,
 0.28015481454747837,0.38621003926637465,0.0088047926536029,
 0.5228240682610891,0.8649719529175802,0.6708850714386722,0.31520249088728547,
 0.03610716695538818,0.46407479496134996,0.003853539906132797,
 0.12006893704055543,0.2997727712256437,0.7572457828971424,0.2296571803350469,
 0.19058814556983905,0.39834794289096587,0.6347709917979538,
 0.12784937776119154,0.8875114563989889,0.16296323072798513,
 0.6301071596108052,0.7105130033531881,0.34326472976318245,0.9978425352132259,
 0.999924895765235,0.37692473922848413,0.6337879458558571,0.3735855215616416,
 0.6505992611450673,0.3418530604299682,0.7097895747094225,0.7430535875884436,
 0.1198970545749577,0.9731273745443011,0.1852262291343041,0.15456283795376569,
 0.1986060291997207,0.8057594634472032,0.17246398530816143,0.1692201456401241,
 0.2104301285442517,0.4490412741804177,0.7438860231374356,0.4430558343235199,
 0.5619308213139826,0.5467473814037835,0.37346358226179155,0.97887022553191,
 0.7977793688877007,0.9330882924303384,0.9133113631925469,0.9447612737886573,
 0.33225410517064036,0.081611984682052,0.7562009715065477,0.7703079339176909,
 0.8863725469902606,0.02396525366937785,0.03058854772954689,
 0.7291854582622486,0.1693095175308913,0.010521209589588354,
 0.07048340283588006,0.15711581362895122,0.9200606414515564,
 0.9346000947701812,0.7045159587947353,0.611043566963112,0.7702701165279945,
 0.16349383072303236,0.48651375336964164,0.097114919383816,0.918040652006901,
 0.728332191091496,0.366519235261034,0.18461365213841696,0.8778909270765478,
 0.8935252927816286,0.49221386786272203,0.27444636754826346,
 0.2545979538189066,0.5277158753117073,0.1649172619210355,0.9639823507785734,
 0.4738680466614086,0.3443821559708409,0.08126136130746187,0.9208784518534239,
 0.28169463356594304,0.10562517796463233,0.09296494074394623,
 0.6963889681083142,0.6155831834577229,0.4421360600315042,0.9592335996883437,
 0.2888000388018199,0.5573728413365711,0.9698303963750281,0.7526059476307901,
 0.9502988676241767,0.870181613597212,0.5045397257494157,0.5469817482670271,
 0.46884763462122403,0.10213437801935299,0.964988963993567,0.9935402339729961,
 0.21078729628376935,0.21163418182760496,0.33088484178978683,
 0.051947398343140216,0.7118020489534558,0.6508160041866471,
 0.4989748949487235,0.7159380695240907,0.22141033345769356,0.3592456092434885,
 0.8442359861535772,0.2410039435999548,0.3087812407171586,0.16641152365776135,
 0.34126388397732066,0.5088165693114902,0.6468508818942105,0.6320320751419128,
 0.8501971809223333,0.43895226607064264,0.18593126254687997,
 0.8756610461167255,0.5474354574008475,0.10962831870043954,0.8028343093119418,
 0.7338741000922573,0.9961993182677386,0.6385663989376665,0.22268042125503607,
 0.10381610041103251,0.2202575158260751,0.6995794508831771,
 0.16110613395942575,0.7467081860742698,0.5710254760352473,
 0.021499362668739308,0.14633089793300813,0.6119460574688063,
 0.7058399374695901,0.8947538912425651,0.409771408491025,0.14246728802451192,
 0.7946255450845837,0.02903202728570209,0.7170574262401448,0.0993442331240566,
 0.704803960055547,0.8031269112896686,0.17915548302496132,0.08939968064773651,
 0.6099505560265465,0.6838022864261613,0.508095891194689,0.20949625605503863,
 0.5121320789269012,0.03255645675343488,0.9612047118638265,
 0.18461853533759776,0.3750565186468252,0.21025397090846054,
 0.6948904138534437,0.08474592770136091,0.5924158415794366,
 0.059833314153910355,0.8871244301180551,0.1894268364341417,
 0.08976766849876838,0.6367367041075899,0.949991420649183,0.9615331098224963,
 0.7516603631302516,0.9491969841485175,0.6682493271392141,0.9210234654701149,
 0.47125009125575734,0.03693055286856339,0.6611763863122061,
 0.3029782439603187,0.9207330300247127,0.5462566053268985,0.908482751010496,
 0.9824053268062539,0.15456147971394973,0.13131659535108886,
 0.5672171402873099,0.11611025859277968,0.7142897607507692,
 0.46436714814849067,0.7302937191951868,0.1718477695946522,
 0.028202023450597524,0.8557066739827027,0.7520834955573483,
 0.47001359170349244,0.2195263486371264,0.5221460159059582,0.7764068340852569,
 0.5990946019709265,0.07454965419887971,0.196611712522746,0.971220260425183,
 0.22897624048361542,0.08586287618871913,0.5660937664237514,
 0.4681735563863785,0.3377394250548831,0.516487153448572,0.3354734839147371,
 0.44780383478897967,0.7598034090045228,0.9627523942406545,0.3804200601042239,
 0.378256455767238,0.60319260038473,0.24574249316616592,0.24178401283608564,
 0.015418337654600878,0.4527850160416176,0.20798335155514802,
 0.04465875173560718,0.7548982250024014,0.04943528546341314,
 0.6043436462175589,0.3193085868319533,0.27813721886912,0.2780060701659568,
 0.18132030622791317,0.5114960287136355,0.8124807721764096,0.214182456196785,
 0.7574658439540268,0.5623150118193692,0.5020073243838571,0.5496949276411578,
 0.699126920692639,0.8630846887659543,0.1329359413664457,0.7716038377043327,
 0.2672869911379627,0.49047607688702943,0.5461672961864008,
 0.021930869640419237,0.2958558053343222,0.9750088583546619,
 0.029366676322853502,0.46798871294520406,0.46795640643031866,
 0.18745878126406101,0.22698467390141408,0.5266805885033441,
 0.7664478635456521,0.5772174782981141,0.5390161518527459,0.3317607821128393,
 0.33308704242317433,0.903484654286717,0.34235670931811146,0.8306734590570042,
 0.7263740432142141,0.13718716630658478,0.2853942193183492,
 0.47727469014604085,0.45880018380785814,0.41005979069424436,
 0.9810488951538081,0.5656990331834447,0.944618715185349,0.9908175454213015,
 0.5503192281242223,0.13847317476205645,0.7284697669848131,
 0.13194617948028642,0.9296564681891563,0.5884526217423071,0.3818850250164082,
 0.7448038332261835,0.9718472109479076,0.05267190949078382,0.831975615169484,
 0.9436640459746239,0.5539426043664377,0.7713038036362103,0.7912402694090775,
 0.8401034326204624,0.43442499646345634,0.1545102361721994,0.3977076432084604,
 0.3279534079038876,0.026008802552757837,0.8203643549023646,
 0.7903736295899351,0.9500722877754154,0.9406565561530316,
 0.025156345319465978,0.37618338107318783,0.6139746261522707,
 0.5459805735043413,0.5999517797704823,0.5969236333915048,0.4742212582294899,
 0.29523057637082006,0.30842733087964425,0.15697244268275234,
 0.5172172470679587,0.7073303572938655,0.06455420376619325,0.9157248304542077,
 0.5815399507428638,0.24763604965205477,0.17864794605933665,
 0.8557433198170132,0.585377578378698,0.09158643404051581,0.7864363836897015,
 0.6610731569780686,0.7918050989409365,0.9559708146392069,0.1324420120710872,
 0.30444937698438246,0.02354011362476438,0.3571689054022832,
 0.9054488249146677,0.36987892202284256,0.8373309087077407,
 0.18348164257440425,0.48974576394443314,0.5403073636409113,
 0.9084345882350466,0.5095670362016739,0.9506550703209393,0.6891718328579294,
 0.5576158483921656,0.47738371412439295,0.7770506302450964,
 0.20481023776051077,0.5469496242860041,0.15667540927202273,
 0.2794773097410027,0.8214400531186716,0.4713895956315972,0.14543505998241713,
 0.07408403538868624,0.619962869399853,0.6926979031840997,0.841178083671668,
 0.6705340948263818,0.13578206467247855,0.2041581565493953,0.6693734876678133,
 0.8562136763055256,0.8990518841390059,0.5636551310566162,0.35209166806177616,
 0.21163854783927571,0.5186454290620686,0.1376920293919307,
 0.12756338647110743,0.5246523148604325,0.5996516173702064,0.9040266880626033,
 0.051931133015012176,0.8974070052680023,0.26547657818716475,
 0.3014657961279853,0.26614684118767096,0.09626521877205685,
 0.7841224145833428,0.5213360169993838,0.09754814303142256,
 0.00334345659592028,0.5350993208627086,7.350100298868512e-4,
 0.8644701856311034,0.7388473650123297,0.7298450476737446,0.5966095860981199,
 0.9927843099815225,0.7216735015713934,0.030618364894221317,
 0.2454575957150742,0.5236843213736779,0.7129907307938301,0.5012074694829574,
 0.7351554764284012,0.7500949030547385,0.1251359536632174,0.6719415390929798,
 0.4481860221327376,0.6442797227522049,0.2997116537483995,0.12933568646687466,
 0.8400545560634811,0.7462424278933892,0.26290024024539393,0.1356985877164848,
 0.06756610332690682,0.8747715969108845,0.6707416008123965,
 0.31765249606908685,0.3409392252614598,0.7501748734615781,0.7998232052517742,
 0.7529928092276643,0.7940787762801309,0.551002455639511,0.33493559737137024,
 0.0026546956899321517,0.29197042943747487,0.45232228884127623,
 0.5773457408072925,0.4079443185298438,0.9659633469049647,
 0.049679534171247375,0.08213115622901002,0.34453341915550195,
 0.05970909584818618,0.23914462034987027,0.03440059704886811,
 0.6013031483639576,0.6743906195452366,0.5914619066751586,0.783734375339793,
 0.41791370778469006,0.2955669686777208,0.6746843164581288,
 0.26195101093803463,0.34201348417151023,0.06296908456388806,
 0.6090664507938063,0.09092611355413349,0.5518307127450714,
 0.45038641486506126,0.5052116344210889,0.569347218321333,0.04246672652070893,
 0.7410217196118742,0.39143000682998064,0.513793824792897,0.7809017868949835,
 0.8623835262751853,0.35041753124872976,0.6789280108143204,
 0.06790056210256756,0.12398430915328307,0.4609945215691278,
 0.7003669544380674,0.760479204855787,0.7477136438382059,0.6765596184969941,
 0.305890932299162,0.8935664796258952,0.22207841113012106,0.8264448332519404,
 0.6634403346109784,0.533686755313481,0.7660693549517256,0.27838448278019,
 0.5236168561018945,0.9808092707721366,0.033397923207671654,
 0.6013635237540855,0.4263148885619835,0.6577451242411568,0.583900875506782,
 0.9326970254934067,0.5650632811703238,0.8230312440717984,0.10972572666574898,
 0.7054013565389017,0.877002399671862,0.2933612318534019,0.0844238916302883,
 0.5822454268422586,0.1118717528187162,0.5127160411982962,0.07337404522178104,
 0.811773886738075,0.9919090219404221,0.3999129424611161,0.3395976602064298,
 0.42474864812749613,0.10262623688575045,0.1830959336249982,
 0.9833643592458825,0.6938590593918729,0.8984638981957449,0.08989388263102738,
 0.8186529914836175,0.24637612396774766,0.5000603828466669,
 0.43310768094730145,0.1208510170985555,0.7285511156931838,0.5453015327884371,
 0.7848300778231986,0.36757282790151624,0.24796264805951718,
 0.09540931131002883,0.5528633081946769,0.13647138509861967,
 0.5501637580938169,0.9489071359192897,0.6887390891433717,0.29657268927428826,
 0.7793752291913241,0.323980272403517,0.4202672481550773,0.1409117293929344,
 0.9143430089624437,0.664005960109869,0.49628690950795495,0.6707100620979485,
 0.1675777712823623,0.293011970709833,0.44274337626341886,0.891307331865899,
 0.9776050685882771,0.31965082763745034,0.31503926869258647,
 0.8648431587855818,0.7793163238228552,0.389784665516117,0.6766737225582502,
 0.5308706969367505,0.304337999732222,0.2753033704837232,0.04848634720435563,
 0.2691067735071415,0.27284159103483885,0.4824711470277465,
 0.45634415447012566,0.03107202475630888,0.059703650681010356,
 0.5666307589160187,0.8754535363245702,0.639223027174706,0.9079765581550896,
 0.6302446875426669,0.9998127887571687,0.8252026264158601,0.7865474487130479,
 0.3571621190043519,0.2628612365036591,0.8508634898519835,0.29988374969254505,
 0.2775060962530398,0.5258049354003366,0.011858364516416864,
 0.039772948024355714,0.7084996720830841,0.9500126342199288,0.8006005866727,
 0.9868935811825899,0.4863942239499405,0.2994064504128172,0.9101314485557912,
 0.4089425665513471,0.2970537876760284,0.8266601402819466,0.2809129167367469,
 0.15772197899203877,0.5306957568241496,0.6655287582928462,
 0.41458613257188826,0.83181340479189,0.6907277487681157,0.4925687277635493,
 0.9133957819841576,0.2078426455459288,0.17586115279561743,0.3407813629807114,
 0.8439239819899373,0.8620518911466293,0.0025330519712337685,
 0.5215509214153914,0.43089319795409653,0.16171357867255998,
 0.1749110397621031,0.9117620316417202,0.8611161514836161,0.3691473910331584,
 0.2877755350645024,0.09088956021804562,0.9900444510744361,0.5572094205571021,
 0.6416650716149221,0.9452368047524722,0.9979977966995631,0.1671812897657221,
 0.10979635296868473,0.2973488349643857,0.2330971614538575,0.1739066019411828,
 0.8627042323757903,0.010483726285491435,0.7000367294800893,
 0.7387211000360937,0.7379401945229392,0.18950569048558974,0.6510135224039586,
 0.09829434991131547,0.3949537835028729,0.3220819779765426,0.250905337547784,
 0.5590480202267187,0.27111091601464865,0.1507717350709743,0.8148138446658368,
 0.5326828462423697,0.41471675616938974,0.07244687200507882,
 0.4796857253872092,0.46544871172903135,0.12808073641618967,
 0.7247457995716965,0.14730140201337538,0.606786580915096,0.2853345589236783,
 0.19979579463771857,0.08423883703510815,0.12381161612140767,
 0.5556918576308436,0.3286386058507924,0.7489309019999899,0.7400132892603765,
 0.33875233231277957,0.5581921573424149,0.5034223001077844,0.6934373025418243,
 0.13483467429747065,0.4624422400006615,0.08946161926853902,
 0.9559663862244092,0.5093786541197695,0.5614867152760854,0.5718692271940377,
 0.9439421448353874,0.2087062043903133,0.20852194901030652,0.4499242374484609,
 0.7400151479335632,0.18217985325311625,0.2050120088002948,0.4815482133368014,
 0.3142037151745899,0.5001907042724032,0.34304652869745167,0.5811382273972904,
 0.7601865526290623,0.8214513680365908,0.9630946900998022,0.42979865808392803,
 0.524489704625185,0.6612791336498076,0.5905167542896215,0.9463143395104461,
 0.886973758452207,0.8957302507628576,0.5422057099310686,0.7715496907440348,
 0.9630283079421963,0.36221775234744913,0.8874444910202792,0.6370922209113229,
 0.960414793334357,0.7268004245311046,0.2539438987279663,0.01100224172915476,
 0.24032701600521922,0.7865375275995473,0.2287330509421004,0.2670577170433007,
 0.14453836870288062,0.9557588893696605,5.979959225822506e-4,
 0.3993195290797036,0.594255310998204,0.881194218852924,0.6393516373892574,
 0.3355113560643558,0.7176186904897603,0.4769561204489996,0.09866225427759856,
 0.4683069418452003,0.39468306015684507,0.9059741905279944,
 0.48405132700603126,0.3571334630927764,0.184361724415826,0.4921271181613057,
 0.897667110262318,0.107595451912749,0.18916581894952222,0.2479801689032083,
 0.8974446040780913,0.7436917430530239,0.4471977514722981,0.33522735350821165,
 0.2809577645826934,0.011767847054558356,0.9392121784304912,
 0.20917303876942062,0.4555770983467097,0.5453980867399777,
 0.16589588850293224,0.30134706118461974,0.6863560402722803,
 0.2536473429042674,0.31954628774952487,0.22131419748254721,0.901619584289192,
 0.6769495729063222,0.8505905037885505,0.3591269760187493,0.607090714283197,
 0.7034351010756589,0.3034387712503921,0.42917804997482334,0.6894700885109054,
 0.9632333324558086,0.37828401001541223,0.1828466379882614,0.5091368487824552,
 0.24112138530400684,0.21851997467683804,0.3146045669993065,
 0.1506437944521798,0.9963230097536654,0.7083858289552563,0.3377083661270369,
 0.62980054079643,0.07686278914472011,0.27228655765268983,0.6857640684377194,
 0.21562606007868568,0.04148123404648296,0.9092778448514249,
 0.07355243961519053,0.9453522684179645,0.9948920740234096,
 0.18990396126922338,0.9363049283017925,0.8539043276506929,0.7910944628435514,
 0.022177334375191426,0.3328524679180377,0.3061749445611699,
 0.5014228699214967,0.28543588517621754,0.00817075560040914,
 0.4103549076687214,0.4489774710353279,0.3291837584524151,0.8770862527015428,
 0.029140880623131027,0.3365850376430205,0.5369409935812763,0.387381756289487,
 0.5065112001544593,0.7731114998939058,0.013833293199408736,
 0.24747345016672806,0.5123362875995492,0.8444034257143815,
 0.21643473407692304,0.724484555068758,0.011052624704858238,
 0.5893750163584166,0.8000110113594159,0.21118716640826474,0.8875067667987502,
 0.40541474723214976,0.08524259851912008,0.18478675582079018,
 0.8147767434317463,0.7889239469443157,0.5986588006504174,0.16570539037620902,
 0.5563920728504635,0.12209704495669405,0.03984299038750416,
 0.5062076931542515,0.9485620905444477,0.7089178819760299,0.3226853641394949,
 0.8875422202055097,0.6337729560359909,0.628508574525741,0.4019671738046283,
 0.7206091859388682,0.22325155199357472,0.8797295356197246,0.9112262504315125,
 0.7249847115609699,0.714913808116005,0.9078518458304417,0.22340694187948396,
 0.8266612162365676,0.7621691213164135,0.9222951889847731,0.8296477691557345,
 0.8499118134165721,0.7034213107588083,0.483937264658298,0.27434396079511847,
 0.18093919313010087,0.7419955095329609,0.7488752502350675,
 0.47657201557802087,0.337174043150714,0.9917628941809162,0.6490296215109155,
 0.5036186685591055,0.031235491043911545,0.9347308822246619,
 0.07015314415533991,0.7956972220175945,0.788183806822262,0.6962611950992517,
 0.2150630417157371,0.17201982250788106,0.7314689167033311,0.9047023528709811,
 0.1628618681304338,0.5518603727670742,0.6442582114907536,0.3673209029111686,
 0.6344836726295634,0.9580951222982442,0.2576245540283152,0.5829994767923221,
 0.19971592328125887,0.3220609236650682,0.5271582235390343,0.8624018548581661,
 0.7523103981574997,0.49621835443923235,0.16077131090002028,
 0.9270218144148135,0.9987597044650985,0.26996347147227384,
 0.13508057802783446,0.7234469397262375,0.483733884313006,0.00783070272478592,
 0.8524327383495169,0.39337403798141446,0.4824707071754045,
 0.46177408572916656,0.4419677965948292,0.6025737583162325,0.499629961341317,
 0.6151082431612938,0.8545340480417594,0.5421930314805622,0.5459138812369981,
 0.8143190251702128,0.02712590026302153,0.24070683645498425,
 0.09347852317895977,0.9869070416016034,0.6386795592477719,0.7695666647565685,
 0.6612863014319237,0.7960176584401217,0.5651148385644871,0.6466824980008004,
 0.92664953404096,0.09842611787614741,0.5625657635959629,0.542565536922552,
 0.462790327370576,0.5986974652920121,0.5080061802474912,0.015729903590323224,
 0.6571487734117669,0.6609982243534064,0.3452569504952484,0.17635897716748694,
 0.20391862545406336,0.9929331006355173,0.4403800891302899,0.9189834283108405,
 0.7372394339073899,0.4469483635511915,0.8950153285522832,0.45816984243027437,
 0.9197648814237692,0.38899521131716497,0.4972629059412733,0.4075145387364383,
 0.6050896936944583,0.784553800483486,0.9087342662123845,0.9199548920458648,
 0.17943665281976928,0.12288930943630427,0.16887639971617996,
 0.8294952416435785,0.46543323557565064,0.421995181500154,0.3610435243852449,
 0.009977089633075198,0.18382745428891145,0.09207264051802266,
 0.6002301701940768,0.7396669473399673,0.31623640961438026,
 0.41381244064954814,0.4718857520199432,0.5223591173666342,
 0.36737080160939384,0.7505817048711826,0.009015104047590539,
 0.5731712476631223,0.6800572728853606,0.1434618430531851,0.8327830614576527,
 0.31817510292161155,0.21955860256732196,0.16873808928319756,
 0.6574053186829288,0.2372791205871969,0.020517597105049856,
 0.3420736962022698,0.22387654842602212,0.006504008250657667,
 0.5162357292668655,0.9728723413308659,0.8366722949438842,0.18776230786887294,
 0.18401664886103308,0.04360099139741391,0.9944284888024606,
 0.1614148655425135,0.1724074395129993,0.6021717432025421,0.16333633868656594,
 0.1600833363570251,0.6327021181757946,0.4982909123412572,0.5436980655231448,
 0.5197321449320226,0.7039865901560107,0.5170011038014793,0.8722698878712603,
 0.09919652249354016,0.005282661304077618,0.8445063032065727,
 0.8578945758145501,0.8512550082306672,0.9547350450432674,0.23405561371815808,
 0.06189400198108563,0.5456867060656061,0.5454592826435709,0.9653434735355199,
 0.7909957311404645,0.009650073240598744,0.31410666338824156,
 0.6004519477465797,0.20003965376747512,0.8158052038294727,0.7151824089217831,
 0.7309711456657904,0.2545884163435771,0.39030675889728705,0.7724104930859363,
 0.8120302185304193,0.16126818400156617,0.5341428076496554,0.317085859751269,
 0.33126924252390266,0.7774348656178712,0.13314237676671326,
 0.7574419295360317,0.36494823503909135,0.03813834281414685,
 0.8941099242035098,0.35460599886534816,0.6157555794877123,0.2621088158122864,
 0.1717572859278822,0.3131521241451468,0.6906650296177105,0.2471191757191279,
 0.8335418267466006,0.9832809552195384,0.672269083361014,0.18341260438120555,
 0.929866055982869,0.37307249447032564,0.9378990744954636,0.1758250467380419,
 0.8200086335615083,0.41861546571394637,0.8460801541397388,0.7199998413667298,
 0.39967616844589204,0.43382230187978965,0.21984035776176225,
 0.06026800873645244,0.6340623967581016,0.8836707814116088,
 0.39719582797114195,0.5028989149650619,0.5596913994711892,0.620015093637552,
 0.7238151661529777,0.9218075737285886,0.6994857752553537,0.8377341585518283,
 0.24068812916204718,0.11061910062755897,0.23679969685456337,
 0.011943463311572144,0.15448150264703953,0.8550982133796167,
 0.07568215809103418,0.8885264537034748,0.9892704098824403,
 0.03419936683618929,0.13898155138410462,0.4515045600370332,
 0.13254154861446987,0.7312362582177526,0.2557700946863215,
 0.08123764832443381,0.7174363953095131,0.9676754969698255,
 0.38514109056510537,0.9553870585647513,0.6045440561221187,0.2968575245584214,
 0.07957957159163509,0.6040164600917561,0.029014306324409156,
 0.3766716725575906,0.4783533083392334,0.6347205319472433,0.44304043252644765,
 0.9067204086549312,0.5708180157054815,0.411073823876247,0.4795793413419718,
 0.9227937552469334,0.19657844769688726,0.3226533358771131,
 0.07872106749031793,0.2796183871878575,0.475561587700456,0.30349937598963217,
 0.9497524161589557,0.26482033657377635,0.569114201725636,0.3190728782589407,
 0.9941900588257496,0.07334874699192073,0.9231728265217027,0.5287894566816239,
 0.9017678656339164,0.18917194400570614,0.84847699452983,0.433003219531904,
 0.9829087065997193,0.2420118681838711,0.4402879555240624,0.4661466639217995,
 0.8814552000846123,0.25963506598590413,0.35848695427713284,
 0.04651448412633008,0.41219783621414574,0.17319731542127892,
 0.569101781467871,0.490776503494299,0.8571597476470242,0.6395663879459758,
 0.7262941951550607,0.7884831791795863,0.39034792052807477,
 0.48163672223270737,0.5405095719076354,0.3179533903565822,
 0.23758039915488238,0.27295687006722513,0.05949077837775374,
 0.827908311200489,0.11397227874621974,0.0015048636880096167,
 0.803176244402834,0.7422502662340124,0.5336314722764428,0.4411972825819893,
 0.8638154091465812,0.244673324703233,0.9552065694266582,0.5793127640379161,
 0.3630397537446125,0.7881251985199251,0.108117479048362,0.7523406463318119,
 0.5615297995572059,0.13755234348283085,0.009607333152095943,
 0.20893204947688726,0.6227523809668982,0.848374942766769,0.3943338075679508,
 0.8884631555539322,0.2468682486541648,0.06929731726376609,
 0.16672279086993136,0.21039950262670282,0.10819019877665592,
 0.5182899146678703,0.5337128179346974,0.9604684601981339,0.3616404065129827,
 0.47187943580042335,0.9626656788686137,0.3007448718263642,0.7089730442308888,
 0.16798360903820508,0.5078424774829161,0.0024466734456876083,
 0.6277960697641106,0.11307936904485238,0.7802682895839225,
 0.08465036581239205,0.5556037302242514,0.7975062726597278,0.2974928200583712,
 0.8474278354782986,0.4037268232463158,0.8142121193016723,0.6316151948519465,
 0.6039531568471912,0.5668526385178234,0.9915649306725317,0.8020828761362208,
 0.6328601811271798,0.8783840505839093,0.47886915558944565,
 0.44708679510432847,0.9421282889219785,0.5074620610825255,0.6720580041559053,
 0.8448313675625072,0.5823892123338871,0.1927355781940694,0.938597652843842,
 0.735150227082866,0.7192206078613872,0.0685245115134181,0.03768515376241366,
 0.3441788311559405,0.7203241752985399,0.19796072033070378,0.888998140350004,
 0.036679431344740054,0.9814908366927239,0.46789661519863546,
 0.8781118375617313,0.3646445641760434,0.05810939560342021,0.8591952646398203,
 0.9681467125607068,0.34389112497767116,0.2792168636829657,0.9374509510004385,
 0.9374086818514622,0.7958928416315647,0.9562300248185274,0.3980906168447803,
 0.6095702167970931,0.7753272115157025,0.22008699053594216,0.826371580451033,
 0.16104072313994744,0.6498339445036704,0.22162936920792697,
 0.8849523603185492,0.023498236913140347,0.10947277756422802,
 0.336497635824925,0.8794026366310657,0.10274798163071819,0.4418921124953018,
 0.006040808698267286,0.11163828934706188,0.2850940338626622,
 0.37840186549626875,0.8733482212674699,0.86002097733556,0.6655248731231012,
 0.4450810682088846,0.0800676301658293,0.4173225779886176,0.7747674400369671,
 0.7133754282357309,0.007044309549943395,0.8486066092217102,
 0.2101967007815232,0.5195739359714883,0.5180093027171393,0.6438793569321253,
 0.8762650390152796,0.6161130642240612,0.9041875941252145,0.7512265812660601,
 0.10780835723163862,0.01476252262576283,0.8760310021591339,0.580062953582186,
 0.8669341500816915,0.8673181094127297,0.8009965200601518,0.1215394358843519,
 0.37895918543272356,0.8649378516348151,0.22838009333159404,0.751318531139041,
 0.8203043482248473,0.2605130335709639,0.39599901571268115,0.3694821300655122,
 0.39439505553990084,0.9793305522658181,0.2548275199368977,
 0.23059736567886258,0.07497822022735656,0.7351092954050953,
 0.2328570667180534,0.4314347957467649,0.7177397969475129,0.441269277293986,
 0.9934346299008024,0.5137796014177674,0.9174142163600189,0.785833300211753,
 0.5468791851971473,0.05120804342094909,0.0703983026893763,0.4710441221232179,
 0.31139482512002814,0.026196225087734448,0.8926519709599011,
 0.8013242162803021,0.8070366065798386,0.9106854500977778,0.5221685789405554,
 0.4966283042867924,0.8645448975049683,0.8692307206782736,0.89621831626231,
 0.044096848073697625,0.31374060556413796,0.4082874259403777,
 0.7298510376913313,0.6495771645140107,0.5030781590034992,0.7713905378136887,
 0.6647756691915557,0.22252693161072956,0.39990347412924143,
 0.32292369691805023,0.7566757271597271,0.817155710239063,0.0830343047494353,
 0.7815820028374845,0.7591788356342196,0.2235192061836353,0.3429537668932203,
 0.19154859913901578,0.5238019563030104,0.091197603066827,0.9090176199693704,
 0.3429034470215988,0.5655289009477492,0.22193545055217168,0.9916097603293103,
 0.9923169836301456,0.19425343522206462,0.6739980763139873,0.6569500645410894,
 0.48088638447962473,0.9712427448809842,0.9713624396339098,
 0.16634918534308574,0.9646411637370673,0.11398372203202045,
 0.47689094609711835,0.9104729297274589,0.38407204924114124,
 0.3924425930913841,0.4854488113878911,0.9315066782569758,
 0.008671513778803708,0.15791522326655671,0.14516623668723394,
 0.7611879850053496,0.7101467579983078,0.13477220459719153,
 0.41461055509282296,0.44983194012418826,0.5240211668356282,
 0.6119637186975031,0.957814197921555,0.4395130443657532,0.24013538481744934,
 0.939344415130707,0.2314431029412808,0.49944220583871257,0.13721754111851137,
 0.09686503933315538,0.8927779985264396,0.018464244513206385,
 0.8994487614581146,0.11316921654817447,0.2965664179424644,0.5959523001719346,
 0.04615859781800413,0.7949921470769528,0.024742958411069438,
 0.05255191278124005,0.18794850874347224,0.4239860298264747,
 0.8722935279109105,0.7978509012729917,0.10122430643999547,
 0.19836968987957249,0.46067388764441763,0.38820928112887987,
 0.9971288035301473,0.701992702558024,0.7449826693935757,0.3832195699203945,
 0.15076163651954555,0.8617296987953998,0.4825020773609562,0.3771469846582113,
 0.9939491333939774,0.35213399415839053,0.41181735515358986,
 0.7324101299265828,0.8266110946699884,0.2377485611448087,0.8960008254052771,
 0.28800487963413657,0.40441161507676915,0.6078717189464029,
 0.1171430529865719,0.6406431116451429,0.2913074405860253,0.5565560918901993,
 0.2677808628160978,0.666821388708345,0.7708259998186926,0.16422963475776453,
 0.6219309852358523,0.7149387040169116,0.3970169375902767,0.36578385239153,
 0.8726240826352998,0.3192475856429319,0.9713735661403666,0.13973757755630478,
 0.5751530589018237,0.917752213927497,0.6427438044671476,0.3886305266799388,
 0.5536320444661267,0.595920991802323,0.47346345026976633,0.7836359887904556,
 0.9287330943827298,0.6853532694036293,0.489259744097887,0.20848671468438995,
 0.8866209802769486,0.20432390166877112,0.39052530166658883,
 0.4786438128522207,0.9702187401781202,0.6713012665768945,0.16605444982841955,
 0.6433841996902271,0.8732327027755677,0.8226043868513019,0.8844031509942867,
 0.5861109674652399,0.060747806608993815,0.8329777029144916,
 0.6388704326932202,0.06242860909641412,0.5160755193264698,0.8778785737745352,
 0.07985916583018171,0.710722376776855,0.768466982897427,0.8213995397424951,
 0.813404708496777,0.9614231130989811,0.7250803480056597,0.5542985634061416,
 0.18666539549773198,0.6165574147337406,0.8378625150376116,
 0.03187333030969575,0.07391974104188592,0.4397974688211421,
 0.4015905218218554,0.7237884849036886,0.5573801250535895,0.8566868097668364,
 0.103871425916511,0.07426969003644768,0.5114353838153305,0.6797345415229933,
 0.09998166161887712,0.29526079803571037,0.11544130177413359,
 0.7349283400968729,0.5055431099574723,0.550641684974686,0.21189627497151986,
 0.021289341241498727,0.9906683744739726,0.34870116372682514,
 0.2920492273332155,0.11989454281179412,0.6664479860275948,
 0.41882202229552434,0.40333845912535415,0.9209055644227827,
 0.3863637518452786,0.2384276519782288,0.9536018672118183,0.6690739144748056,
 0.44283775610819476,0.6288077560012306,0.9154043977509556,0.928043110270713,
 0.5471570839763475,0.0926235966764053,0.2048345810652321,0.30856839873346653,
 0.7358007376680171,0.8299776866511833,0.09195242758655975,
 0.34497714939771684,0.8714231429702401,0.5032309571918718,0.2828301234045748,
 0.554334115657177,0.4902763695707292,0.6618660309149852,0.8771873195900121,
 0.2681822599717548,0.950908613636098,0.39098177214678353,0.9438720815330767,
 0.8432036427445112,0.8341943276290271,0.8384159899507875,0.7632220867472956,
 0.1771358658096771,0.5632141156312958,0.8916226468486301,0.39656778927733627,
 0.14722282668609576,0.38873761008433805,0.4486286435279587,
 0.47407537639607245,0.5543134058089242,0.16832096545821518,
 0.027410677894899593,0.5385074855008775,0.0667932602934822,
 0.6237499165273375,0.18231490918581061,0.024136386391407694,
 0.7319888203502063,0.23900603437821866,0.13381695291278495,
 0.7272933208143665,0.7621502238361477,0.23479580563710833,0.988697166129923,
 0.465272356942791,0.9718144120143505,0.8856452103117248,0.18187942176345406,
 0.9605183993194949,0.5354177750190331,0.6096976550938862,0.1642913670751569,
 0.6997373114750655,0.2813068397149159,0.740562087888348,0.4784544901901936,
 0.9780233164919273,0.5097494249054304,0.1975416133084762,0.852172449509851,
 0.9005257580041869,0.5064220867726368,0.7659765529085196,0.23446549852053122,
 0.9093830214296459,0.8844213042338389,0.015547307179712422,
 0.0851020659897398,0.4618647866731589,0.6246959106224034,0.985466989748842,
 0.3078434215144509,0.4980309464028393,0.5238327808677885,0.1229878665548032,
 0.537849954237839,0.5462072982555667,0.7188582853867848,0.6574543178821397,
 0.157323504334286,0.500687720458658,0.6919457166715749,0.46420797721472895,
 0.5000381558810123,0.4160902495634853,0.6574306812875025,0.8627682401212909,
 0.9801877961964349,0.9797661921482611,0.42347080161729134,
 0.08037542695752631,0.4024514650800358,0.05300630068576684,
 0.7912401966899958,0.5918567845914029,0.7401742603966682,0.7028109461691888,
 0.792110444470092,0.8326278915160673,0.06699521880043069,0.3994430215087945,
 0.6629945475017122,0.407764678181357,0.32774132677229806,0.6522208035464856,
 0.8782238740083526,0.5664663655079129,0.7713736088068175,0.26105237356826905,
 0.5800914703393227,0.1278044540037947,0.7704214910276669,0.9722700020614136,
 0.0886237108238842,0.7514507981294275,0.886907984749894,0.8265979900612374,
 0.6352739925665923,0.4695364657288148,0.4977967614798501,0.7008777902839741,
 0.07292071013281731,0.8168753934582593,0.369453172446078,0.939348193957788,
 0.5633460738501183,0.581174763748175,0.17476713354675077,0.1709511164302947,
 0.6303074727549814,0.6666299236852734,0.5054168190399081,0.7391206334694799,
 0.423829028232776,0.2930373901429748,0.04309095121678286,0.5758214420872071,
 0.2512131573051085,0.3837794253510999,0.03194236931430339,
 0.35210226378744647,0.860877589151266,0.4005840369398991,0.6301937961829213,
 0.9007622605488652,0.3378165441374399,0.9200319599774637,0.48792763478165324,
 0.35608233249128196,0.17056495189415055,0.7346556357781022,
 0.6237612902076409,0.485390928194643,0.6856214925069836,0.7295011415714824,
 0.2335760021788219,0.25337780621683015,0.45184070821169997,
 0.7425131041103014,0.5598134049688439,0.5975149057168683,0.6731384347654991,
 0.6181344596257611,0.03937457200916916,0.04968855715899789,
 0.9156526589153802,0.730912702020205,0.059567435732231555,0.5895745772187245,
 0.6883375613686126,0.07460794146105754,0.8273058791630502,
 0.36664871398051924,0.6972183277836498,0.25614254637877054,
 0.43377460205509133,0.1693149038485694,0.026886242424275508,
 0.6621930541573939,0.7406438337595866,0.2695539755837153,0.9124426565654731,
 0.38645400180641576,0.990519391371385,0.8064978449387319,0.5977761136855765,
 0.4145126576706688,0.7347456701122442,0.3383465845244492,0.5927864022275705,
 0.8835003517142779,0.5214301471915879,0.165592565180405,0.6145164078069085,
 0.8568228101498858,0.8049865302157926,0.17097150036822306,0.9497834002213358,
 0.24673975957667138,0.30511460915741906,0.43220118923996986,
 0.12628016238985373,0.29489538293274353,0.37481426652497196,
 0.6942531637107512,0.537255054983033,0.5481295698898858,0.7641203572532868,
 0.31886689637894605,0.03665369232140281,0.8759909284708356,
 0.6364151794815244,0.966833561002349,0.6088354376332008,0.7141105333106663,
 0.9108896204477838,0.971827240468454,0.04901313139650765,0.7523011293870878,
 0.6555480293652971,0.037726153589003886,0.6711974097668615,0.804772130798492,
 0.44104818340284957,0.511137565446099,0.010742705006618336,
 0.4738480174132432,0.35286296181333865,8.292552947550824e-4,
 0.9380478182641353,0.9165483448472587,0.8971849053499058,0.3268552827444078,
 0.6339459201350405,0.2833994188158667,0.8117912728032792,0.6252898443734756,
 0.1505078686640644,0.2201571925818564,0.5264758486092497,0.7420070311204665,
 0.671829562085528,0.6824243038991865,0.5748708506620775,0.5633914067552075,
 0.17693647549192515,0.06759726422179402,0.05217670464519808,
 0.6808714033036813,0.9144451053173039,0.8328895396375865,0.7817256670240759,
 0.12731946421468832,0.9546447935365441,0.2893638832630956,
 0.21676511500093731,0.34079621993627107,0.03540335133814143,
 0.0496923551835492,0.5587496141989373,0.004695042875889288,
 0.7105081167174713,0.4920935492535816,0.766546000675677,0.7367654679930301,
 0.8530135479034546,0.500943947215184,0.8032327737704428,0.32791340287491977,
 0.8593965161471809,0.7056362956092097,0.3253361158929522,0.8780883606616624,
 0.9455146736939031,0.8199016569622108,0.8519310485161885,0.5989964802635547,
 0.48106411599452636,0.9073272920197863,0.9116605259982393,
 0.18515554897520814,0.028795919982141815,0.5979395313293225,
 0.6261327983795014,0.05285871529881747,0.44144404932244996,
 0.22194450476051375,0.3775642559467547,0.9574869925349283,
 0.29087695199006514,0.8444995352636901,0.4059722099782953,
 0.19374632363124433,0.7492319293041211,0.47581582546895973,
 0.6059783883024632,0.1473565875859394,0.12150740597710452,0.7706176891573039,
 0.09562787522268468,0.917561567999104,0.4252409136061137,0.3764220269420129,
 0.03185469545977193,0.11916318768692269,0.8835145141868683,
 0.7677632595576661,0.2862158424136315,0.9951612485680286,0.27757006537510653,
 0.6464208388761845,0.40333849969491675,0.48360170031152727,
 0.46002101066655954,0.7125223998698065,0.5878910629001919,0.6388063644032427,
 0.9283875927046003,0.267916587342625,0.396669461579489,0.16456884890512147,
 0.23152576476906506,0.18753988194610327,0.7554286276331817,
 0.3036864164860459,0.8657080679595763,0.46434094501817436,0.6638218897771657,
 0.9123583308465429,0.3674720870745627,0.5395096797211217,0.28038177837841427,
 0.6874368063521115,0.9357253046849626,0.6302508046300213,0.47715670681963984,
 0.8430295166995921,0.6760497603283798,0.2607216793360203,0.8540691544213023,
 0.1748286200929564,0.8753202070213658,0.7476716417423664,0.8188493213464795,
 0.5787168967167823,0.1904575741085569,0.36641265315452465,0.9463280435448658,
 0.8621013367783078,0.3441866807325695,0.5808876750780825,0.2564454569586043,
 0.8121962801283991,0.2390166769819373,0.17121168383045926,
 0.45500857480586165,0.3372022595704477,0.4050482225037979,0.8689345553783101,
 0.6042316242026219,0.30999419210421464,0.7127957432827465,0.2779002062041648,
 0.3292030823624035,0.3843385423047596,0.6108571934869266,0.22552061807107981,
 0.029469153381440294,0.25047260444278374,0.0871518924678939,
 0.9387782639962945,0.9542984215621593,0.6123439851641619,0.33561435053492794,
 0.09245242996949099,0.035838305679157245,0.08344373851114562,
 0.8846396181977089,0.5043452774962238,0.45475149756269173,0.8692265911094017,
 0.2979900763465979,0.5407950028809553,0.5710729008655162,0.7661001138692213,
 0.16870349369442605,0.7661951592855168,0.5378611266028303,
 0.38894889461699167,0.7119318273247072,0.3720513347937887,0.2972985532987379,
 0.5065891420340664,0.3441106022891175,0.7692835755669234,0.2902856847235159,
 0.018889802955569968,0.7398642412451937,0.4475028792946809,
 0.5560478452796649,0.9626477149865674,0.5206089026078311,0.34158901200515934,
 0.8104281977949448,0.2089119906109015,0.07398435653989399,
 0.14094851955406662,0.5587244538120489,0.8943075034323733,0.3622856523556701,
 0.09436530033507928,0.24630118105165832,0.2377359619682311,
 0.9962146703409337,0.9238136287727049,0.09449270964150036,0.8749133671579717,
 0.3581864130612513,0.221800434255484,0.4855120397644934,0.5209259659281094,
 0.033984623024566174,0.5237815291030109,0.4997607455536832,
 0.8827512300252062,0.05166144451024013,0.4414454468576757,
 0.42364654999157425,0.34742619577994593,0.7592027309138943,
 0.3992674051436962,0.11899814930015329,0.8987688338367362,0.2329822741497375,
 0.6847377018564078,0.659897469414666,0.06166741912363605,0.7746243773413255,
 0.3421461964054362,0.1272683941164855,0.7037647984575901,0.7229585891279204,
 0.30001665530019284,0.5116609909210976,0.9021766806056845,0.5260028913357151,
 0.31511945679473274,0.3505113432015917,0.24049243422921907,
 0.7354803985508935,0.350871716420895,0.7264687331700597,0.5453834214673862,
 0.6261611338705353,0.8103605645110081,0.024006457197208175,
 0.9624390501740806,0.33862689887782693,0.19234258696302997,
 0.9407290315463053,0.5003838160145595,0.3899954797021601,0.6558121118126361,
 0.345458678316818,0.3449778045179328,0.7497290075306101,0.6961353756829504,
 0.2277069637525304,0.8456793426602576,0.6937813474900473,0.11054867616561714,
 0.8245513568620217,0.8890801001218093,0.5681548942849354,0.979578732835422,
 0.8769213639254054,0.7872375410267285,0.25016633490467777,
 0.14502803036213552,0.9584023820008728,0.5942161971830193,0.9122789842513815,
 0.8517899206982296,0.9310590050767102,0.028978987589230454,
 0.6622985289134256,0.9906918298660703,0.7814382468310849,0.937172779042147,
 0.6352964186803324,0.4273597773414214,0.2944777498156873,0.11135950907784753,
 0.6709759442132716,0.6010206387309429,0.9034920754096286,0.6341417579036954,
 0.03330769767340136,0.5619467670468297,0.9301996497365577,0.2779010669601166,
 0.012728879217361322,0.9945518422415902,0.8383122092476414,
 0.07317844413964203,0.7745971601745386,0.20092114123235483,
 0.9311243804513193,0.5548623593646993,0.4105677486933985,0.8803832914488501,
 0.6667712503553451,0.4591903626398114,0.24777270589018052,0.0834249379052513,
 0.2949771367622762,0.924158820794513,0.7333081686514469,0.7259745258349847,
 0.30495677558396983,0.025421552544842463,0.8907076362307044,
 0.5437664909990917,0.8795047513326752,0.22855150176576344,
 0.09557382517464152,0.47644864817615284,0.061373672286246905,
 0.5850613486639662,0.646739511245405,0.25055216511971234,0.3371007037913658,
 0.16873257971819444,0.005996017301127754,0.27484328721794316,
 0.5822041869660767,0.24507913433893103,0.5667827720006756,0.4803092330102102,
 0.9284268424472049,0.25998451882735285,0.7067915833392795,
 0.08339555181563141,0.9759960415582046,0.476088857703616,0.17261857208053621,
 0.12205701984566752,0.7542694946577463,0.2681307394564426,
 0.16855050100224966,0.3672606602669748,0.3041698850599519,0.8325382364954823,
 0.5841926292475829,0.2067516468999574,0.9177284874707928,0.22596969795572597,
 0.09272613231286353,0.2255716475504611,0.609465711284249,0.4560349466402601,
 0.26913011819897104,0.5480351184899631,0.5405754598196908,0.8318927737031823,
 0.5001866874526267,0.9426736030218934,0.6840551980585984,0.6376611186823633,
 0.2997425274625899,0.024553273056613256,0.9016870937368904,
 0.3378973604929746,0.10202260791080675,0.07604999410296343,
 0.06628615443478436,0.15373143850173254,0.956019952382364,0.4144451272837406,
 0.26262514078328447,0.19754192809993776,0.9070160266941701,0.544183602178102,
 0.556803635198149,0.4956687445457496,0.9196730055765122,0.06373463008426017,
 0.6599136397552212,0.3308334255652301,0.3560991381079266,0.6078440827487288,
 0.345371571384673,0.6776781179300341,0.007435113933677795,0.508343673018427,
 0.7642864541057057,0.9870150904509597,0.17575158564641868,0.7425568974791513,
 0.7648192208130353,0.11799562895910498,0.12616790315957638,
 0.7305576834012012,0.9004311721436736,0.48685532986517566,0.2723673553379382,
 0.9620957229102404,0.8160986337439328,0.587563820997697,0.5950208547514033,
 0.6854944162439117,0.21029095934607056,0.2368442795773299,
 0.13742236621631376,0.9662828187089608,0.1729905080418085,
 0.23774339074905293,0.34310217261350096,0.5880809600722048,0.490437672950488,
 0.23710719400534608,0.6834155823821656,0.6433580142564436,0.242466123499389,
 0.5789598804645644,0.45545366946288235,0.07569204890683323,
 0.35194021481175697,0.7485112530579802,0.6151571693335897,0.941219541070368,
 0.27250661771815265,0.5348998618232192,0.7042890841579225,0.4230450466544803,
 0.42943519410229547,0.8363682898861013,0.19734595640923747,
 0.9225562788073421,0.1496616235154704,0.35962037467715324,0.6414413027846254,
 0.32039579024541376,0.08115401008090128,0.19234307899190717,
 0.29076392200177437,0.010385343686548643,0.6201213283908508,
 0.18046483773954503,0.9867316155374786,0.3448677124350885,0.5080982573850005,
 0.5266562626633733,0.9014870146912028,0.4270754909748077,0.20756631806392167,
 0.31904399381258974,0.27695380155318705,0.41134910042760797,
 0.2490791955011028,0.7920205767238024,0.4554588192106078,0.35135196030811,
 0.4736830758620011,0.13636395465638695,0.9763851467857716,0.1845261378438896,
 0.26786647099188965,0.5756567229750285,0.37058066038042026,
 0.8679612106020212,0.7700935459833155,0.8532352867681444,0.8808298001780053,
 0.8382854314102196,0.6749416608566995,0.2705085590968628,0.4710943890972109,
 0.6409460955858085,0.372305192468651,0.9122410942755476,0.7449117801996298,
 0.8750516492906478,0.030332150053547835,0.29407145214799413,
 0.22658882950738646,0.5855214552040545,0.7785314346443896,
 0.04898824589602102,0.8541470270977045,0.523458362058122,0.7904961447238344,
 0.20999518325718403,0.3909935095426207,0.22737734121011677,0.189913090091113,
 0.5148381643067989,0.7044505290617636,0.39093605006949694,0.7155385116360611,
 0.4106343723933221,0.7412851246416466,0.607714559910485,0.12272854479468676,
 0.8913691855003769,0.6566529936815174,0.22346461631389114,
 0.12297325058197983,0.424567652248909,0.32632405292568323,0.8556564029322443,
 0.6164775496662974,0.030083981733192378,0.5257159901940623,
 0.8873441134176134,0.25445460917568863,0.6886368620343356,0.9136427491460607,
 0.909816345723211,0.5169003821934579,0.5788832319772034,0.0517837462446743,
 0.7259317141287198,0.3997028330282595,0.11854815069662894,0.136017106527906,
 0.4546610751478225,0.791133852440156,0.5102697241211683,0.08154254581416476,
 0.20457659084699842,0.18723650721635332,0.39944356385070756,
 0.02808721971412298,0.7727613452403679,0.42538231210304933,
 0.32919958533837057,0.8777429948767709,0.6066652421923181,0.8710276489156575,
 0.7789770604988534,0.38051305118919176,0.40534858973987287,
 0.6133347099080464,0.3480380682896469,0.6635192460455476,0.5338896049018579,
 0.726689088831325,0.3896652514401132,0.6812853854732328,0.8785135507989994,
 0.8490912965480348,0.6856523715148564,0.18506928726476146,0.5979946109586118,
 0.8768209080326713,0.6732075462949447,0.7789755158222685,0.7815075543841413,
 0.6130939943719623,0.6400572531067024,0.5793342205733016,0.35172558559913025,
 0.579913272884984,0.7144678483050043,0.9633098760665423,0.6277487303089957,
 0.5628412065020169,0.1282276178674001,0.1854796212973887,0.30241916573862604,
 0.759550078312487,0.31692274359414374,0.2422735502948078,0.467433715002484,
 0.5432830394652546,0.6357228170455995,0.25350168258616046,0.7899445957540883,
 0.5615592241364413,0.22716797843928593,0.6854913489747384,0.0650804221399921,
 0.8136474047252635,0.9902875426059994,0.43906059902439454,
 0.13361692943365966,0.4135693338419405,0.7526287810079846,0.6908639816331877,
 0.6660322262145942,0.5408960227593759,0.4388856837490114,0.5275096690545783,
 0.5471656332856606,0.2854875251007538,0.49321017602209005,0.4870477709786363,
 0.9367193777106338,0.7713924797550371,0.8112797679094544,0.6398294967327586,
 0.7519586355313312,0.9034843005710719,0.9340852328721414,0.28788574770049635,
 0.7110097598209821,0.6934808536952073,0.5420923993198584,0.6042799543141323,
 0.7504504194438931,0.3842338167411258,0.5037396558984403,0.578908864585613,
 0.8414304900277633,0.007493834423957901,0.6886466151212889,
 0.8942315264959698,0.5580079077349925,0.49866502185108597,0.7574369331556621,
 0.2527775060447012,0.2934757271433902,0.041169866068145655,
 0.011648571435336819,0.7747384849003413,0.5867879336914884,
 0.9934741241696637,0.5879978543155825,0.6731920380095625,0.07192236411990038,
 0.5729122716677395,0.3736537166271052,0.1243242832262964,0.7989232211918647,
 0.3242436760378723,0.9642370949945013,0.21268672710233272,0.6930139820305928,
 0.3757589736975716,0.8890626981856591,0.2676082942927813,0.8274839265338939,
 0.7867365155528601,0.4216093885003034,0.7697256274133394,0.12646860008624083,
 0.5098214111695044,0.294904215148293,0.7812864854272571,0.17888287163373773,
 0.5396955746205638,0.7902821126060822,0.2631975251770997,0.9037995662639717,
 0.540573830222812,0.2441372831862747,0.4908965698539235,0.15445347092489703,
 0.7577579909315403,0.4274171439289197,0.13396259516695652,0.6033071756936443,
 0.2547555535511363,0.10227424356773474,0.386840144801167,0.8937875512257423,
 0.35238584551661867,0.5408154196338228,0.10931241974657124,
 0.5608818660476823,0.18456577563304344,0.647021049572969,0.3014143534355591,
 0.7371635091156291,0.5008547591558241,0.6114718014446192,0.31091317319051903,
 0.4096102052500582,0.1316270383887823,0.9741058431660033,0.42808190832992865,
 0.30993639068133627,0.7463235816896743,0.17045151603895936,
 0.21783376173165103,0.06931975382449096,0.5392850540326527,
 0.9764859213980208,0.6915294617297315,0.6058259664323467,0.7790574439681722,
 0.5737277837035847,0.8135700526269697,0.04785388117391176,0.1574642743074517,
 0.9775197559157588,0.5018042329798427,0.789422122298361,0.5213064967801415,
 0.30410540339971626,0.5992819883465555,0.21686687966911466,
 0.9723694603073726,0.11230210372630234,0.9262326195107292,0.3437492548898029,
 0.17360106968276523,0.6125048282116846,0.4077563228386477,
 0.42024121136828774,0.3117891523051929,0.5909152624229494,0.6286490347985904,
 0.4413974315574912,0.13496515567906808,0.20383271746503073,
 0.06181584354525016,0.7112795220131041,0.6472875317730509,0.4990278630991286,
 0.8948003309064944,0.821961233058452,0.20334978004723592,0.8451932979303728,
 0.2006591791277974,0.17644388921366128,0.9517473975780381,0.7728070718811844,
 0.8815735391877074,0.11009048324060788,0.02044163903581997,
 0.9508438059882138,0.5607754268445405,0.8754785428232458,0.46819670265458224,
 0.3619736730960771,0.1454642633679729,0.9225483644770525,0.5559007268825944,
 0.8335762513828451,0.07467892981931801,0.48187928648149436,
 0.30908793238854737,0.9699979486387187,0.6935811397825303,0.9856824872471471,
 0.7038380896984483,0.8628981005526076,0.10028235763336435,0.8609180251361077,
 0.28663406920753953,0.49655607946893654,0.2611536420440831,
 0.6209627459661299,0.49971727459775495,0.5995261609014705,0.651878782362459,
 0.34883627997651256,0.04867020202854977,0.4926897250512807,0.945360834052191,
 0.24389806019813542,0.7527821362408978,0.6295027734371524,0.8544177827084566,
 0.9427620442879111,0.04542529185263977,0.08450038897112255,
 0.44668065890590036,0.040037328306320896,0.42089050599581157,
 0.9005868572065141,0.3018171581214224,0.6614016716779288,0.6746575570058232,
 0.7224805194029384,0.5336183716858287,0.18397800014761068,
 0.038087659232009186,0.9689134537983549,0.787556216455423,
 0.04005624836066923,0.23530344041171836,0.7931004593545243,
 0.3254050960999333,0.06933423002050276,0.8709020755465026,0.8482450479027005,
 0.1947224765170391,0.7454984229277137,0.43173565023986116,0.8651675352759975,
 0.9722349646497588,0.5439786678083033,0.5334813171857244,0.1448455714443163,
 0.029714137799366047,0.2129272100824371,0.9440745385823769,
 0.2620224293743386,0.42384321076455644,0.40279119466698676,
 0.9629474034631196,0.8203277904004234,0.958170903636839,0.5468951007998194,
 0.5042280839740504,0.49511628423832255,0.6625289951929338,0.8398614101434787,
 0.3547238995035933,0.12310880579212646,0.954675692880637,0.3241271976172375,
 0.06504095706642232,0.7713729969514159,0.5145346865382324,
 0.49115084481626303,0.6502508404051531,0.7623365237050748,0.971006402107031,
 0.2871035330999421,0.24547431286337462,0.17338815330478818,
 0.6984165421521604,0.005589767983429028,0.649528587965989,
 0.12264364285961493,0.16224528798672866,0.001031975616828329,
 0.1207255568821145,0.5221985836503628,0.8011211860034484,0.8914853866023319,
 0.10411242187922254,0.8007927605040283,0.9189086144088278,0.8701333558129396,
 0.34622229022663886,0.17732462025293527,0.9885229879109261,
 0.32736576456060196,0.7941344897294891,0.2807117320884528,0.4467090188291818,
 0.7568202722858092,0.7985571307069584,0.09222568860348757,
 0.22710027434750346,0.19256932133913507,0.48647411768039583,
 0.12171653700347096,0.27474520311931616,0.6309207910548492,
 0.9089401096320963,0.7946045043336396,0.01869462279255185,0.9184341153802191,
 0.9903131939769756,0.5166481357373862,0.5338809994927006,0.9415040346283102,
 0.31161272701693365,0.6086349265545669,0.08642999331649259,
 0.015615090193484615,0.1930698430197293,0.767555318600338,0.5031554502394175,
 0.2581478037242124,0.7046149009088724,0.9370230061957427,0.9369590390380245,
 0.9478052720608274,0.07313415626751474,0.5812889051453208,0.7366435394453804,
 0.4612014879153411,0.932486438185808,0.04401144666202905,0.07067509700223418,
 0.8449158455749131,0.7818597926905001,0.5167536180080493,0.5896439384624356,
 0.4999037830913303,0.9199840742469059,0.6072037469362357,0.9369677934376013,
 0.05186966450903119,0.24820512015480278,0.43315597657201366,
 0.3636575366462783,0.3178589523225974,0.460391771641081,0.8295776164752912,
 0.9227365763580624,0.220588568366891,0.5706527980473012,0.3708022561022355,
 0.3659305709731998,0.09830082928434714,0.7301449051897605,
 0.38503748433593354,0.9391698968477812,0.7595181849179574,0.9416694198625809,
 0.5070068628707133,0.14726737304887982,0.3819833184003788,0.6117484480562002,
 0.7708155983996298,0.6519680692580829,0.26535093665948906,0.5528276516777662,
 0.2977143089270373,0.1270912310663571,0.12135205330729237,0.473854036677531,
 0.7393918236906085,0.654675571511751,0.2958297324886333,0.1814032211551474,
 0.6850105821915917,0.5481106842132734,0.8673253980133393,0.8807089669617275,
 0.8680595067743051,0.7417499957522316,0.6756051985083933,0.2918014515652545,
 0.2817977015335196,0.1543064443553125,0.6447060130027669,0.06931137581281299,
 0.4229767275175216,0.7986217102016029,0.8835713521444628,0.5807211973723652,
 0.37182929261156483,0.7574203946921333,0.1849216814324408,0.9490463850655234,
 0.5040210556392832,0.9647858993019831,0.86558589620321,0.12069552514198045,
 0.37755912646015966,0.8719855691777754,0.6176109625222124,
 0.21617371552575482,0.1715579600955338,0.3472026678228708,0.5276900750584436,
 0.7077946312760451,0.8497697478312096,0.4630031492518837,0.17961074407339617,
 0.44279436860881294,0.4423681237701451,0.3049927415940399,
 0.16175478639740992,0.41407086329511456,0.7858637595979525,
 0.9354122842079352,0.6879057416613328,0.42543912893143343,0.2334222468832574,
 0.7830359017992806,0.16062388908358327,0.1734974300610852,
 0.12433844777883074,0.9705918397402117,0.367228876488801,0.6140337647432227,
 0.2682462931136945,0.6699558991315387,0.5402712303791052,0.44129343008830735,
 0.29837521636074915,0.6478326668524432,0.4225942096962193,0.8073931431770889,
 0.2895120154908911,0.8330757608456345,0.530096880257235,0.6872788116673902,
 0.11699754468692025,0.2740363844407856,0.21120357992969263,
 0.9063293703859964,0.9307068700430783,0.5485451062794632,0.9493914138984293,
 0.6754702183877064,0.4490477118534525,0.8187196200623753,0.6908451100275542,
 0.9631403149945287,0.8479171092226262,0.4640136882525543,0.3963826155846881,
 0.17764661650277236,0.41013466609551774,0.5714851742252862,
 0.6020618334868297,0.7350910199149593,0.4727692076825377,0.2589528063269446,
 0.8229889932481392,0.2383363538654758,0.6227544980410469,0.5846387117645004,
 0.9562722333412081,0.9075985334177562,0.4707624791390953,0.9134060834357501,
 0.14645594579588828,0.38489322438524753,0.783285470870424,0.8908312649748317,
 0.7192171585203592,0.47567265255516344,0.4482207412927458,
 0.20922947810399883,0.7738813911055205,0.5664987878609239,
 0.48533055378379375,0.1796070062264079,0.19309460065569195,
 0.14055330416620104,0.5662498755515943,0.22767357954420975,
 0.47070771638269937,0.14446238211797668,0.9071954602291795,
 0.5575384284037368,0.6327781810154097,0.2010004643527611,0.9861940064698772,
 0.7091120744679238,0.4706964653404684,0.8937403718361219,0.955333069986824,
 0.6206226256584616,0.4354540749801834,0.4700390264474652,0.8510278398415765,
 0.9013999815450662,0.46449043693192693,0.2601108300845837,0.6771896517274374,
 0.8674887391150893,0.13785964363107484,0.23276670355034157,
 0.2829375713482335,0.6163135776611345,0.7690917627760573,0.20937336558538877,
 0.8121514234798987,0.15427439476619942,0.007977867786386317,
 0.11484010405194756,0.27412923872873174,0.23869376450924829,
 0.6298349919850245,0.8087677274801863,0.6552381698177393,0.06120115286384065,
 0.059978075063745195,0.6768920501567136,0.3130767016758498,
 0.27048017318987794,0.28963387320743783,0.16789955529418166,
 0.6098523066182582,0.30157007131526714,0.43521773865786073,
 0.7291694869275089,0.07940277561771503,0.11937461586142073,
 0.4125102976648104,0.2268147342982475,0.26265833550007867,
 0.41454310537134165,0.45970052823924157,0.7796079859654087,
 0.8196696109688635,0.8798152435512151,0.7441145586580452,
 0.012384586985098811,0.5474254809657284,0.3335899852667832,
 0.9398053857514588,0.24445311568181904,0.05599404187572321,
 0.2675840755486453,0.12327780803314281,0.04285596499992472,
 0.2564701346930649,0.04669738035953319,0.543544916326654,0.9027510989053987,
 0.8221387648234305,0.021294712122749027,0.698467181191207,
 0.04187500795319099,0.7470522824400253,0.6735218973457504,0.8180411014953158,
 0.7897906212477095,0.3869083586629216,0.9899907923404468,0.8955748892787146,
 0.3963767712462837,0.179317872451891,0.3440605374335173,0.524277899707944,
 0.769952648768611,0.40772695567155004,0.1934915013327858,0.2572589685387524,
 0.2418460395446811,0.7556034338829326,0.2924195800239595,0.29906892752455494,
 0.24418495316914834,0.3736868436175549,0.9657905079194395,0.7983540088627421,
 0.9691860595862344,0.06417414513976283,0.4060813025063046,
 0.26690424202338936,0.21420324835522564,0.10125319583207781,
 0.45646150885740766,0.9424989579531716,0.7093650248452478,0.3107723010232547,
 0.18384588479037223,0.5835512622111394,0.6864949760339826,
 0.41514096991497573,0.8544124297429285,0.8649553069014603,0.5284299856737803,
 0.7154133787156256,0.7858390536666544,0.1927966990740143,0.9900263696444012,
 0.08627070899312705,0.2657726211339112,0.9044248930698577,0.5919518445408773,
 0.5279696889127037,0.059570509372114167,0.912246929412053,0.9307738466531159,
 0.2151076566942247,0.6702889632017043,0.4548530907790077,0.7585242712600404,
 0.8253250643099665,0.7766345259019253,0.8794657964894403,0.4311042374874551,
 0.09035095134397975,0.9153384761705913,0.9175982005693535,
 0.49008247778457337,0.1391160927182693,0.8327636842465852,
 0.48899454969369394,0.3385827847543378,0.695155223400577,0.8011305689295642,
 0.05506918942653849,0.09603112722579177,0.6957795834211564,
 0.8270312072450411,0.8903257039790186,0.18237062022765316,0.6238797906903786,
 0.1975788518414281,0.3293524500366938,0.44202916929394975,0.5287729682603208,
 0.44635505025847333,0.9354898121878377,0.8328845704047172,0.4238520886752122,
 0.3642831935365827,0.7766702229096921,0.5846419726180158,0.1991847794120607,
 0.7959342051736714,0.9647253655024338,0.8223227543975727,0.2678578699084353,
 0.107617676077568,0.2786047714082508,0.38778386618257277,0.8284824754328544,
 0.5706868079562133,0.08290579628722361,0.717235087865111,0.37917908521444155,
 0.5264810244800714,0.5009781537196037,0.7196957576319724,0.6976858049015833,
 0.3481542621547318,0.601481523518765,0.7508535669080044,0.9678774103919386,
 0.5027494728872357,0.7492890030342687,0.33968090045144583,
 0.23184314830585584,0.5140808251078415,0.8226156898500904,0.5022292164152891,
 0.49911173984236523,0.4334700954202815,0.990104344149975,0.2434990511273658,
 0.5805024781222266,0.6186653582539662,0.7631057552591748,0.37944972186083736,
 0.25568002891529074,0.23313313286371173,0.0067321912179334475,
 0.3547500274803945,0.7389863726400652,0.18659357742473914,
 0.27176676549827394,0.06494620926886252,8.754707979898857e-4,
 0.1497258729774419,0.8929761241663359,0.9803225110772528,0.7265048398901395,
 0.5476919442604012,0.4077725213539306,0.9532773272948025,0.6504610433023831,
 0.68177553729869,0.4844717642745743,0.30802282900041633,0.517079649316473,
 0.4518102803129982,0.14581227206346958,0.7309342934261761,0.8120888274973382,
 0.14728408292537565,0.34242717478604956,0.9189128238616991,
 0.6013808056975317,0.9410309425842285,0.16686771633055975,
 0.05309644028016236,0.488514296056926,0.8535000261859949,0.7813127468169061,
 0.5900413453943474,0.027358856704335688,0.20764248846736466,
 0.4162568469276352,0.1409563989078827,0.029892489886773288,
 0.8658798589755514,0.4621884574698193,0.09165753074313021,0.9201497551968464,
 0.8232959981011779,0.5966769292575892,0.38336846186144025,0.3804783941352824,
 0.43340520525621684,0.9204314115976292,0.16646272770118598,
 0.6751328859248229,0.7096684596365268,0.7942343706207022,0.0995677140755844,
 0.1252803615753968,0.8596989861082038,0.6344413265781707,0.9961343436164982,
 0.9112179560626614,0.6238037121934863,0.010155631265542753,
 0.4787212520006414,0.2894485008963418,0.04636277423222335,0.7504096864709568,
 0.4098231878166796,0.11866335199910494,0.21042636403409376,
 0.8650659746759228,0.29238048342430345,0.21829652690707824,
 0.15918267528904761,0.8425336791657266,0.8409912878696321,0.2179276221973917,
 0.2670473357979335,0.10797462619999487,0.05455648454282358,
 0.7909491961699343,0.02443876281321744,0.8725758375819032,
 0.12026957750471012,0.9263010057866464,0.863673512070731,0.6305086410832447,
 0.14935130007875064,0.7907453976818093,0.3089031653568787,0.5282800670196259,
 0.46948634125750177,0.8145020997508767,0.7727340702495883,0.4749731848514389,
 0.7356599213015249,0.7693547058142127,0.9129309225872129,0.9074318771327796,
 0.1033628442732124,0.2755038505828231,0.6011067013428627,0.12124930961583158,
 0.354751823058003,0.9421053446239963,0.7072419983476845,0.9491959039019255,
 0.2662387212431929,0.7972315011103429,0.11651262650570049,0.8625693827437313,
 0.4018514514636212,0.3845706828311869,0.9498185404998964,0.150993735770662,
 0.8167029362853455,0.7350005268351308,0.12398685444276836,0.3385311849439281,
 0.39136951566199474,0.8411614047207995,0.46371586218130534,
 0.7634266857288976,0.6054486131160352,0.4834373481737758,0.537506119968921,
 0.49386617167925984,0.170389461323589,0.5847018914077832,0.9522548401953953,
 0.15609030903094068,0.7216859343903712,0.9001477616318471,0.2740961089748507,
 0.827965519061877,0.9991207412242841,0.7834976904178714,0.08871134742636999,
 0.9934722443449031,0.7742855311779087,0.15602182183359703,
 0.08984945569784397,0.4116756439721362,0.6611232498941861,
 0.18054313334462324,0.7705751813776153,0.5332960942285041,0.5447238418595342,
 0.4585761462568727,0.9081742461266283,0.634873173180459,0.4521620659016733,
 0.46212458626229624,0.780698548603642,0.5936103473243719,0.5488491393630933,
 0.42098024704796433,0.9272686352834377,0.1613996136716448,
 0.11686627712484343,0.1639450734989515,0.1315618364916159,0.6485570337503673,
 0.6521571497379335,0.24530822026180887,0.0931819154929886,0.7813753330557616,
 0.12911124377749905,0.44196649082865913,0.8594413247593589,
 0.22217538272629833,0.739690510979793,0.30359349631983823,0.9835236946495084,
 0.49583928305883074,0.39419901651904865,0.8415875752188665,
 0.46231281825173287,0.8835997559374746,0.036255861919293086,
 0.6783493813280886,0.4473031306730276,0.20442918939003574,0.2424048085144198,
 0.5360852593076064,0.7696064588796707,0.3038761446692785,0.15706968541813948,
 0.41279579739456385,0.6781799219862504,0.8072823635676398,0.4533383736767139,
 0.15254266979313846,0.3621067476220705,0.7317023128169124,0.8311239604616278,
 0.3750702889844889,0.5322099724870636,0.9035529671457379,0.5823893940746211,
 0.6597317323365917,0.7745029404153556,0.754451314376206,0.012288993211220456,
 0.9078020275415888,0.03028449110755571,0.14792624594873738,
 0.7110161843436902,0.1309860579469736,0.630803061624035,0.2380984106619558,
 0.5298708397019368,0.20932743770613527,0.5739120434973491,0.2773184194784519,
 0.5353091751482048,0.4312236861552572,0.5011952978496949,0.7172915328271404,
 0.4501960166242458,0.22505081996999654,0.6965025598667521,0.2297200929037525,
 0.5824418467879358,0.4407764795777882,0.02058089033393795,0.3015489365111844,
 0.7504878011227425,0.5223441069771367,0.02641669428320792,0.9486786490288545,
 0.8226742733729041,0.02805759056960777,0.46428608667456883,
 0.37786754868656447,0.9419420885714049,0.045510578302709326,
 0.6453625604990832,0.6191321322934255,0.8297338073266509,0.3431934389060439,
 0.3932299908120641,0.7461319378253034,0.5987483997466904,0.5345940553718069,
 0.9417051271896694,0.17079620369280424,0.2816596289572244,
 0.047969684681414826,0.3137302017791682,0.02324635218189197,
 0.8675616490154981,0.9477269549660949,0.560086388415734,0.6247469911527352,
 0.9114760166181008,0.5369673315476077,0.8234644860512298,0.564512852248289,
 0.478466541812006,0.6052230728311899,0.8379062512625224,0.8799641854464293,
 0.37557298075182566,0.5928813127096878,0.289739945701039,0.3611672211868361,
 0.1862212270307978,0.5943449740440447,0.47797000769857545,0.921658733586018,
 0.5336376349128276,0.07690093187670877,0.7362169761669091,0.4964335367500273,
 0.703068363511434,0.8483045976910629,0.04526174327917709,0.4912204772702191,
 0.3896391883313485,0.9559488677191684,0.9631520817935949,0.4490106030940899,
 0.2242707561523425,0.5380932306685938,0.2891087072844696,0.2737396219195377,
 0.6485950239507889,0.15519381027389356,0.18305642693942092,
 0.8685607785540219,0.21527967858285268,0.73989583770343,0.9571231054128004,
 0.43567781411372075,0.29238526628996075,0.5296070434711889,
 0.9226980226911385,0.36930457080122203,0.4607608998628657,
 0.31892980131176485,0.950333879106551,0.7787996175543132,0.6162511300859118,
 0.7363578222503937,0.6058611968945675,0.818393303596083,0.7171153264484172,
 0.25707350262315054,0.4214112713949074,0.828430148993526,0.34729725743013795,
 0.18406320479802585,0.7270460828333222,0.21032051135200258,
 0.9764195723674085,0.0650759441988652,0.6716833501686907,0.6494475390282629,
 0.6580819657344625,0.9478168771152117,0.29044677980519096,
 0.20388104655543537,0.5426577925375892,0.16087191723199334,
 0.8170704175448957,0.9497529689585045,0.5455793468291964,0.8220521272105,
 0.2629401122562903,0.23062884416994567,0.7688204097148592,0.4115550388458479,
 0.024256944684860948,0.5543013594994243,0.08674994158397775,
 0.968302335069303,0.4035737818484554,0.14904703844602618,0.6808661660534384,
 0.696108477677064,0.9240745398009761,0.5256727956677504,0.8502596935947058,
 0.17294544019815083,0.8711443016276296,0.8669739543144853,0.6802793410050005,
 0.852128116691819,0.593690961010701,0.040980354277147746,0.0090079323168204,
 0.6323378010833188,0.4805649775787093,0.26148916086269236,0.5777783098725775,
 0.49728626312042135,0.04780073279382313,0.7026784917187863,
 0.20676589402494394,0.8413094155564386,0.8198908820092843,0.5391958029588932,
 0.03752821166509812,0.650394012279926,0.4810215368426335,0.24612964962650286,
 0.26089719801749656,0.3903341948720861,0.726754998680076,0.2698279428379382,
 0.8438640766139516,0.4721325719152648,0.9467782275055455,0.09936822536613177,
 0.21695183159995113,0.061927107265101866,0.3991856161831221,
 0.3797165564139472,0.37811008417220404,0.9472598192498976,0.5468889026352753,
 0.3053363236438287,0.551483172357024,0.33643227410491827,0.36888462527950383,
 0.6315433239679031,0.6228485550370799,0.14631046846705564,0.8351408189937994,
 0.8377806736166569,0.683107384614504,0.009196886461303677,0.8928496260309008,
 0.8258595132681492,0.06481457125306911,0.13586259914166632,
 0.22423495114108616,0.5785128020330425,0.7105148183748967,0.7399597705495644,
 0.34258024133753107,0.8997093185414281,0.21821081040611245,
 0.6258893110449838,0.2502615514755536,0.5045487513599217,0.17135587173027766,
 0.3776840920636553,0.5082131980191091,0.752839709232801,0.21422015258008298,
 0.6707230393218626,0.36855533646033156,0.2819031599096149,0.9893792258080742,
 0.67285278633089,0.0794950653341262,0.970462703338403,0.8615460916925636,
 0.7066884191492336,0.9782112563468811,0.4368143763186362,0.5112386421769085,
 0.14578989820453936,0.24638637561295296,0.10180157135205103,
 0.15922342212228702,0.4775473172785524,0.3793796812808311,
 0.39240394433712045,0.9710847449563917,0.6324240187219022,0.6511378935655437,
 0.03723236031625232,0.5981314349522248,0.0273348979923973,0.3776292765415965,
 0.23914118380037352,0.5452971784727444,0.07508261790418935,
 0.31942055257118973,0.12192452512531149,0.10754138378773492,
 0.3795725100088265,0.4401092730690572,0.4642759019135654,0.1856850630309268,
 0.8004253686185452,0.8109715592014818,0.9371857368696289,0.09096658347815723,
 0.667924010367726,0.5513885198589318,0.20221660612519798,0.11705777904063275,
 0.856883974394546,0.6389535777453806,0.9531266075055705,0.6165615989790526,
 0.3142163961715796,0.9450509450545117,0.19754570190046283,0.3976513256036639,
 0.36515255071085573,0.599886177220209,0.6787211537576452,0.5717525248179024,
 0.4957173653044583,0.556131154861262,0.6651641619278617,0.5739072357541847,
 0.3540673434985403,0.257787546405408,0.14682661733368185,0.26031272949516093,
 0.45909085696593566,0.27151470793938803,0.148559509031017,0.7516083324377542,
 0.3746724853197674,0.8417027469828601,0.8358795670203911,0.7612489528635594,
 0.09684534063180117,0.8803395751714869,0.8503573228166,0.4565641994275008,
 0.36922558072844824,0.2119482775302296,0.15803792814674256,
 0.009950221184820252,0.11855630185683763,0.08385311988722577,
 0.15782009833493205,0.654155582436714,0.684744609572976,0.5927593125057561,
 0.6453803199280963,0.464865478021143,0.2551271659681794,0.0607012942654499,
 0.8330151671463579,0.5698508373429187,0.46003738453036125,
 0.16859151721205246,0.7953850170565516,0.18200026409369252,
 0.6544826940072421,0.32469695183052216,0.7211614454295585,0.2165929906945505,
 0.4333163969738576,0.3287289772568819,0.18490674106699123,
 0.08297038380929711,0.4213388928305265,0.6367540240187309,0.9375572939727631,
 0.8204787812913827,0.028473779526743126,0.9607629029103804,
 0.4315759852190968,0.6113594086375674,0.24808192494413217,
 0.23354150643810867,0.7859935007448323,0.3786158060199938,0.8903681989237388,
 0.8542385526093041,0.4994250839469182,0.5143571734802871,0.4732621716071378,
 0.04640069991045859,0.5708825320797977,0.7070577307546433,0.8045030890926494,
 0.16442583539682587,0.2623543061207969,0.1170880402418748,0.551263654101262,
 0.670964121094884,0.10703517814698416,0.21804280684266486,0.4059521480332029,
 0.1874179841605551,0.29233351757323933,0.9759633373118217,
 0.06967295380017258,0.586500598733866,0.9726576798781803,0.4421351110084091,
 0.5961434093813104,0.7562299132511866,0.1260943712425867,0.6909611868008899,
 0.38655337085719,0.6453653189228945,0.27928472039722974,0.03861244847996825,
 0.17936405918565368,0.7602351431729842,0.38881081316070243,
 0.04207205335453634,0.2672487220568497,0.2880493153498428,0.6429985075180767,
 0.740042677294982,0.11404335575668112,0.6864956061420472,0.11526449936374583,
 0.12868583334717776,0.7744404016654834,0.4306757258915146,0.7278403646418624,
 0.7230734264428722,0.6471767639620605,0.7661940372131887,0.0942346126924355,
 0.421880291764155,0.5025981906492449,0.8006029781475144,0.2374692889975516,
 0.08501918664712294,0.719873349318573,0.21981525979471284,0.2645921154145796,
 0.32377266753141454,0.6341819341154235,0.02591515109112852,
 0.5869623141253765,0.7951414033648978,0.9370144228356028,0.4274785731462847,
 0.5123107635982697,0.48070795041596726,0.44714694055544046,
 0.12486415908250825,0.7912202110386597,0.889757999929239,0.16888433849820106,
 0.36428275020203194,0.7525519407051467,0.8001549968464742,0.917002134203489,
 0.2628825958681905,0.6039427906950431,0.6441295349748867,0.0774595703133526,
 0.48810730757175136,0.3394196954391119,0.7407301005934502,0.1534619239438565,
 0.2611679512987497,0.8035156136575403,0.6217163373120442,0.5863410830551201,
 0.7722287405158533,0.44794413134151423,0.02360575200274795,
 0.47257777212141283,0.26555436610481764,0.21126508717331705,
 0.9183102414258593,0.4112229731884688,0.11763212567595549,0.3038967415002596,
 0.3108047000366947,0.6480934922697222,0.23596318157612217,0.595847704491421,
 0.4487433417802186,0.0882777811890082,0.8485978854644796,0.669283137814332,
 0.09980420400203405,0.2701082710550472,0.44370994231920735,
 0.10783714240016073,0.8482973515106323,0.285327240242626,0.4140469146021072,
 0.9213099046911498,0.2801345546832009,0.12932307169092006,
 0.04790552578778973,0.42760137562312384,0.3602193975827017,0.870610375396438,
 0.15843490776912594,0.3721164198652491,0.09072184345615009,
 0.5408624095439214,0.4164330783809792,0.4403499211111628,0.7141723446726043,
 0.17107010584605375,0.7351006311320462,0.5088141104368537,
 0.41625765989352137,0.2512202104191975,0.22965804335255247,
 0.5559855532456113,0.3058213618825235,0.6310925178234947,0.7375851522614271,
 0.2022194150951253,0.8645650600134196,0.24612325457019713,
 0.44122653449561433,0.9093840565170628,0.9056661526041458,0.7007496672730172,
 0.11239564321668016,0.4889967388793902,0.14585733429619263,0.39857754616022,
 0.22398764216017997,0.022391227941987668,0.27756979025391726,
 0.25799130174942675,0.18339349424120632,0.4496636821261859,0.960651056025104,
 0.3346152233073406,0.9860185003021837,0.16274001181389275,0.8938063984920817,
 0.6857318076099703,0.9196862349484658,0.7017261503957815,0.3260004930828043,
 0.7038492458247122,0.28648686887641706,0.21781122339331505,
 0.6865052399667109,0.991587749157175,0.399020971412705,0.8688056505633791,
 0.5017673247294223,0.8333780708810513,0.8743589471287518,0.4443559722233812,
 0.8500599239654272,0.5448061314707535,0.6608013127370568,0.29391571648675174,
 0.01870844905340041,0.2535065133203621,0.07864230654000359,
 0.11668490340067228,0.8975782369075914,0.6994508030040657,0.3729796502135889,
 0.5110791767298216,0.7496036540327911,0.8300580162063476,0.6640686290978488,
 0.1350278505437491,0.7793242642787187,0.9725707913539822,0.5399806353903366,
 0.7636568283637026,0.8379002257391357,0.6557350377151603,
 0.029869149464581635,0.14313386225408986,0.6610596131401242,
 0.24758922986691612,0.5859758264979958,0.8640391145000605,
 0.21485093252080678,0.8072567736310643,0.2641434919844631,0.7459970557566665,
 0.08913431234072955,0.25969579557417144,0.6681641965523847,
 0.7109160902663569,0.035159988717779256,0.9222114594144676,
 0.8304409958653203,0.16742331010278644,0.237468042800292,0.4983126399471446,
 0.6863759466390139,0.616399185605481,0.7226866058749672,0.3895657055044818,
 0.8459460289136262,0.6648251279178664,0.6926918109307066,0.516720213992271,
 0.3074303966898605,0.2576601181991184,0.7511721626615471,0.04279167934305228,
 0.48789427357139403,0.7284284035856877,0.7865817916772019,0.7134901762763196,
 0.7829148751118662,0.05721645139511422,0.8645164015136254,0.232761508809192,
 0.7217000699692094,0.8517039182224511,0.007725438819123576,0.625429942703671,
 0.39369511321664596,0.8053812973777289,0.15351379203519466,
 0.8359093038084059,0.9925066980609495,0.6620266672933446,0.27310302691360766,
 0.42972537169151537,0.9107059868417524,0.643259078687485,0.8703420026837684,
 0.058210425493163354,0.996715664983169,0.28396738870917326,
 0.7834722388400993,0.2738750798501014,0.23182681768084545,0.660275664858992,
 0.6883555240531397,0.15028753096250824,0.6978705129027488,
 0.21747204241532292,0.37910168583614956,0.3182312047765531,
 0.03638774855257121,0.0504690768144872,0.2601754489179171,0.8745415138740713,
 0.25154275462521725,0.9792493120779152,0.6344765882082223,0.9087886147847408,
 0.861269100703973,0.23072273710226532,0.21682649094875994,0.4251451969043668,
 0.888762280566482,0.23113677207453154,0.15968264305296287,0.0950466283842375,
 0.4783726020384912,0.21443879591252402,0.46669648146421083,
 0.04572420388210119,0.22234160573224537,0.18326156784261993,
 0.44950603182473536,0.986746745815456,0.8698349585732547,
 0.008608869438254563,0.8406121787273433,0.08760551201579236,
 0.07177820327007978,0.7138248187438554,0.6940858449666699,0.644316074777093,
 0.16523016051992778,0.03624726286410496,0.9854302077298578,
 0.7123050443886291,0.2395058783270264,0.937654566140217,0.17104842087390748,
 0.7449643563300246,0.25236414050170464,0.5901838672895738,0.934277274559871,
 0.5003698949981251,0.6484486992888195,0.7921760428139987,0.13986714590316818,
 0.012232203338346626,0.42958524881977733,0.5485008100781686,
 0.5732576392701789,0.6845778768660548,0.06645988470745057,0.7911287505042923,
 0.1769688932604312,0.8788375632644081,0.8521189750596342,0.5469490326828654,
 0.28264849072132614,0.8100770023440995,0.15248194470751675,
 0.7608882368769059,0.22262398683894558,0.8329755283387545,0.6337083231993353,
 0.36147541984748,0.16400846485987586,0.6406818045681449,0.3264848618717322,
 0.9787210535416635,0.9490769541277466,0.11798023150279902,
 0.011835391502816983,0.14198656764589568,0.30089002328561343,
 0.7936512537057694,0.6535314853724941,0.5465395734212356,0.8718414403047275,
 0.16344699247850225,0.9801009251586156,0.16152977623103193,
 0.23672835926472213,0.36053751214527163,0.684048978002264,
 0.44966854647634347,0.8596674879213567,0.182450813431979,0.09752444833631313,
 0.40064266422134975,0.488809644309671,0.049321127062149084,
 0.28599534840512464,0.9992557521831371,0.20164684045609838,
 0.2878893038536545,0.7143990630675102,0.7664277679690813,0.9964591049833296,
 0.9029678365450129,0.9590138485814031,0.9716624723545298,0.3702728446133351,
 0.657378224872338,0.5780592836704259,0.13106174883700683,0.4062169996059798,
 0.7537451015921977,0.30105944501085313,0.929418729040372,0.20393209714173,
 0.3862953197731326,0.5258761659005434,0.07941426249517836,0.4503576567322075,
 0.8336102447793644,0.3943733083913601,0.033877647350197826,
 0.41566513517073034,0.15329873235201696,0.5882330591294631,
 0.3042318844880575,0.4025740518952148,0.47189672172075314,
 0.35979469056106084,0.4076134840493303,0.20022352600490945,
 0.05615602492852845,0.5722947430250496,0.781660465550526,0.46571858676217104,
 0.2570264596977082,0.1175731762775647,0.9688016938237715,0.4305965247374345,
 0.6682349650960768,0.1840131111562926,0.19665149489671174,
 0.04791278877641436,0.41049574409382705,0.26868412637498973,
 0.34304070706582523,0.8900900118391162,0.6756165391565866,0.629064704243239,
 0.9142901454042472,0.46398011737660494,0.7793891433243627,0.9429724917553043,
 0.8354696992289142,0.42066288838397725,0.23532599657920183,
 0.7536064759832453,0.19613368546467536,0.7643228701354166,
 0.07956507375199029,0.9125730951273291,0.9278836311366025,0.3763212035042227,
 0.8695940427452197,0.02224289228475862,0.4482561506515904,
 0.34036170423407186,0.4679364442087175,0.5571888586110989,0.5949788552589386,
 0.94351866001121,0.22301841835138103,0.6424834715789162,0.09616611860682533,
 0.8048562129393761,0.07644340308900799,0.12129033129320677,
 0.6981934548847478,0.9319583186157282,0.4349992544364636,0.39616835138727324,
 0.5877540107561409,0.9668207382268846,0.5764710252375204,0.7163511588470789,
 0.015071095788539268,0.5041710575138815,0.04802139442747577,
 0.6538552816989107,0.47884783261522457,0.8681884601888189,
 0.19377406423546284,0.09839729941593567,0.24145754605540393,
 0.09333459326025328,0.289826770353503,0.5522225024862886,0.45498057351254084,
 0.8967888719088841,0.4384035704812651,0.8484324670430944,0.4886407434655221,
 0.21710920241232345,0.6543567218299627,0.7956526498996088,0.8684228829778652,
 0.7632711202817718,0.9749889220640435,0.3063511390221215,0.529713715120224,
 0.7848231071165118,0.8691373174000007,0.4737252331429106,0.38905182616302125,
 0.5157706972784679,0.9222292251612769,0.8151000594715833,0.11270671215719241,
 0.1268575101028646,0.1340325515553732,0.8057792024451301,0.2789674292371117,
 0.8437656137870435,0.6952991338605182,0.3096675805323287,0.7474543627854272,
 0.4764372292667831,0.9193964719148913,0.6833494634426033,0.425268283592543,
 0.10404646869809708,0.1616305346278304,0.06569202961457887,
 0.9033335731012284,0.9646680866262134,0.9312092359643914,0.08971581368201542,
 0.20224173086074526,0.4177772288534993,0.5534340091716419,0.7500490239989203,
 0.10136597711065665,0.41479544450404005,0.33585275224778355,
 0.3946767925409521,0.4503155202105529,0.6163620156741223,0.11263191153979524,
 0.04779799814904029,0.5032048909010425,0.6395151552470668,0.3060052287332391,
 0.6613422365837003,0.5267837389916825,0.9214938505643115,0.9232118977622387,
 0.0786201452885602,0.7754979097946789,0.8384548289428485,0.5183062985205428,
 0.9702533671945994,0.11654654969229361,0.2685496322329348,0.7935506993515643,
 0.9163741209234959,0.8661636903041918,0.6071522073364509,0.8208870952666409,
 0.9220904212968964,0.03691038519461909,0.1764575000034898,0.7070602940298477,
 0.47623188054651044,0.40713117879919314,0.9139335670082135,0.363513932111335,
 0.3046716514812874,0.9450815504649939,0.5093848719236478,0.390503989232748,
 0.9113791102793473,0.03096028662581063,0.33398967406527635,0.686469717058479,
 0.7550243921004552,0.08855817362984231,0.12847209936219173,
 0.7361164526451669,0.28420958074409663,0.2864467911190691,0.9433123911177361,
 0.5035531384903406,0.63918097794793,0.5021959490319389,0.7560779336117736,
 0.060740617600905544,0.8093357956740305,0.3722322576429017,
 0.5630521930916866,0.40353325752181046,0.3446148302666302,
 0.22320358390372852,0.009791276894399292,0.5821186335686633,
 0.6746431366631652,0.2724924176474811,0.3561908495092223,0.5776033600022881,
 0.7103873781548881,0.5373424661729913,0.6943764139359492,0.09659361999449745,
 0.41664563285777,0.2252073012518927,0.39489981665832596,0.7992900658476715,
 0.8707387315379537,0.013512433224530995,0.18726803582628704,
 0.8007229200359552,0.8525500589649975,0.49618276707961506,
 0.010093680956586693,0.9650634455962279,0.3480753029884376,
 0.7140823175679555,0.20261849128474874,0.5056464674337001,0.9414348732410851,
 0.28698129935442074,0.40210295038441934,0.47677281599702537,
 0.4531001213288176,0.3056037017593627,0.7339357874724799,0.27880095338981714,
 0.6845999685316827,0.019051681708310397,0.3094169535132901,
 0.8564739460128431,0.9013925031548344,0.16411376896524033,0.5980753246850142,
 0.3722238293534056,0.6414651377875993,0.40635384377813055,0.3817447028858648,
 0.49827417687617703,0.5528183410608507,0.9204085522927838,0.7194078338204364,
 0.6895004498216436,0.8766373531692947,0.018618269104908713,
 0.9090554131271226,0.23439777894724834,0.07906713668389187,
 0.009480347458858729,0.9619748799712824,0.6391946036523725,
 0.4178847678487022,0.5777518495099141,0.5086233266818175,0.6579568122265449,
 0.705456861204189,0.3771143333373633,0.05016170779539575,0.6766229764138143,
 0.5065922569395583,0.8700264494290176,0.95375654943074,0.36949234079887194,
 0.5706103294322863,0.4994375977391421,0.5441868912080302,0.29910410179400426,
 0.5101048915817527,0.6406864983572149,0.005213791317215,0.8232168782260448,
 0.6819215906838467,0.8759736372825727,0.08194719512530768,
 0.011155554284670077,0.7133099349351024,0.7614434324572217,
 0.43338467821707005,0.041991209650618866,0.9264428954986412,
 0.7213681786498846,0.6302236814760493,0.7219786048834222,0.19070579337181692,
 0.2269054030887615,0.6032168431493269,0.4299299271553134,0.971551076937725,
 0.15135648348756747,0.4221173257480093,0.05691805214592627,0.810037405279598,
 0.8071194561879202,0.7772757004427856,0.6265506512502927,0.16617937012925965,
 0.4100345536250186,0.271611416175618,0.7033119668734285,0.4246272028263357,
 0.21351528629485217,0.8580244494710475,0.6256774525240517,0.8622035843077669,
 0.9011680593317415,0.44508361901040394,0.6242302749926608,0.9701555710347819,
 0.3778293833974571,0.33388987089460365,0.24721681714823662,
 0.4675798551025414,0.6507890236837401,0.5447771665464202,0.8069641736217802,
 0.6083041366690298,0.7445888355990871,0.03836540540467115,
 0.47864812785710054,0.24320018348959072,0.5087948630210128,
 0.6226728659824381,0.530224662784201,0.0744761861734784,0.7487066648976342,
 0.3318383174724806,0.1694452074892624,0.25427405595145336,0.8380173454397275,
 0.44433563321899305,0.22733264056431235,0.07611735430535793,
 0.9374392051432796,0.5742011506077438,0.9286560603838796,0.532630473595328,
 0.30518121931348363,0.85837709230527,0.46667326847687984,0.5613440202179768,
 0.7103951277368534,0.22881118240708398,0.5787345511860459,0.6813664637764018,
 0.9085182727351172,0.45766263588216727,0.4786277378833885,
 0.35191754654768803,0.5464152211574649,0.9645758670528385,
 0.21996156852448678,0.9491031541777968,0.11394985232783528,
 0.49837078965205694,0.24834466814864098,0.047460070097616325,
 0.6907920708273522,0.06308329226426179,0.48710448463892086,
 0.39870757684670555,0.1203749037797649,0.8254111811813633,
 0.03177727827444454,0.001733795690347928,0.505693681302178,
 0.9559170416454267,0.34190659269376844,0.5256701959998273,0.5163933436365884,
 0.14391428508337878,0.36591769102526017,0.06120040759292633,
 0.6652420906438585,0.5327197886828298,0.3851125157579698,0.7317377889783613,
 0.08371612118233573,0.26126727694076,0.32047602790321505,0.16649401947904174,
 0.8242489211290731,0.35138310252144367,0.01730862841024705,
 0.24361352194560992,0.06363318388192274,0.6177263493100589,
 0.16971608248447212,0.39438851562369215,0.4361321793545303,
 0.8534072268507151,0.5182084368024011,0.08222897174548705,0.8312934979312301,
 0.39415059823928467,0.8252456537296938,0.16740047645453016,
 0.7439864912033707,0.5347797393020038,0.18512394256770337,0.5077054905571565,
 0.40214968945910323,0.07038212683669154,0.9239426953353428,
 0.7118932167172463,0.614155814701484,0.5933292828595038,0.9112808848200356,
 0.3081542747516559,0.07041560540169145,0.057040506130318525,
 0.08679892565387848,0.7973303662550797,0.38455919267473615,
 0.9749101609062546,0.22493431133736885,0.46609897231473907,
 0.08023278181226268,0.8074417320025543,0.14255629352470878,
 0.07674174052562233,0.03988494563610134,0.3207016432913333,
 0.2960758573176695,0.7311083854974523,0.31437337406261356,0.3097789065217693,
 0.18792823895678745,0.8142465993894781,0.8265274321302685,
 0.31231165910514824,0.6109728062975325,0.08323638352748208,
 0.8153091936912881,0.016112386807004886,0.553564080593759,0.5202147131380511,
 0.5781519409913468,0.3613651762960941,0.550485150640552,0.3618308622807689,
 0.20973865977153483,0.47949532328975464,0.576467569748094,
 0.30018539710238323,0.9204690401565935,0.3814292928092644,0.8473032610767997,
 0.3234199854697891,0.9850937219982174,0.35560047097160496,0.7565823695356086,
 0.8036023048604082,0.6046806002083913,0.8861224901741179,0.2258964068964926,
 0.5431437916904769,0.8281768244431061,0.6672980034747606,0.2502150430712372,
 0.14733544810200827,0.7068721677663775,0.5859894179339302,
 0.006423171590332566,0.7545255034118293,0.11842698334660862,
 0.38485148878849285,0.42193365901036883,0.4849012825460254,
 0.6711399086079117,0.1907931970707356,0.6971543502705582,
 0.005779948598973794,0.6715487602081809,0.2306475345869654,
 0.06449798506321591,0.17742486147217884,0.10758418614342358,
 0.8966529653565566,0.6032905959494638,0.03368517663656101,0.7194595200469445,
 0.8306816917227928,0.7471755428315977,0.30125824256792966,0.821911732454811,
 0.7475308008885154,0.0686110562858876,0.47515507276797675,0.5284786951130321,
 0.23812109032005735,0.2861993547361714,0.3086553514399476,0.1836301586881851,
 0.46692343103432554,0.5081085811124427,0.3516634367707627,
 0.02932650634003464,0.9533092105187413,0.43304555174214576,
 0.010245448837708726,0.6411453766941949,0.978568977357928,0.6407619199666034,
 0.8394459335934339,0.6453361657460114,0.021207581535145015,
 0.30965376420183033,0.7783451607199813,0.1247661807632382,0.5198315106272816,
 0.7693603800668403,0.6929531624920481,0.7827585238882638,
 0.032080339621159415,0.3781990474231912,0.3620016718658371,
 0.020535939402350656,0.6918329551110082,0.1607656500269281,
 0.10260312280915473,0.7485098863615661,0.2692095181335923,
 0.45375856363159417,0.014454877638992336,0.5595034594900281,
 0.0993673868152618,0.35745367472770284,0.07017567376612899,
 0.9970672693774328,0.7558191713982182,0.8852809596505304,0.8329933595935268,
 0.3408266294729003,0.2925669480884592,0.39419159405687754,0.7299774858523286,
 0.6455855448655119,0.7179357150368905,0.28599750527488554,0.7347761356633057,
 0.8419697765071839,0.9877123122162922,0.2824773419060883,0.16038895184963597,
 0.05666642047488368,0.1479464329042829,0.002209517020576146,
 0.883326511771759,0.4349927605245366,0.22844274015874533,0.8620252472224739,
 0.46596178755222395,0.9663071660497997,0.9495028777816015,
 0.10178516537586513,0.4292075719007571,0.41976571283079367,
 0.7065540732070633,0.10748326800539942,0.5003664913209701,
 0.07422265338376854,0.9453942502278786,0.8956862367181019,0.7822224582886959,
 0.8904010505286285,0.8821389043898102,0.8386988434384474,0.22537816483947504,
 0.15230286787579161,0.49223169050892734,0.07752424027858362,
 0.20879528404559689,0.627007803666927,0.0960486206661757,0.42588594657674983,
 0.5950132630633069,0.23891050859035445,0.6749837280817331,0.6941314955732745,
 0.21606918931954933,0.9090478401556186,0.9254912329994696,0.8723507126915202,
 0.6037387760591606,0.6572761505856705,0.30268662554575254,0.2751194578110322,
 0.6122007435865827,0.025974906109130447,0.38885135903880297,
 0.3897611310899428,0.11205461855939014,0.6579286988418362,
 0.02422260058015968,0.27596729288526267,0.2275142687702807,
 0.20887591800635297,0.7946010133294066,0.05980753183387122,
 0.12555801366755004,0.6714080024779918,0.1301034763181127,0.5921332546712456,
 0.04342475248230515,0.9061476764853281,0.5946748303703013,0.4972303243390982,
 0.8526112236526525,0.09819357062967704,0.8267255988362602,0.5820828053780991,
 0.12818908773125348,0.048427147162594886,0.7448497928712496,
 0.9773576046473008,0.5987300045959824,0.7783454900136739,0.38131079264757695,
 0.09530372978943324,0.24061061043151055,0.7369804521293751,
 0.09611314702111273,0.7424098467684694,0.7517532391915501,
 0.25398361211334386,0.532569237502798,0.13114504842599184,0.2763664761235043,
 0.559448917973203,0.44836693866681854,0.0647009689522311,0.9344425094168023,
 0.025037385289717218,0.036861267847564716,0.40580854959370605,
 0.6964285119138768,0.5354993787985863,0.9623407161517443,0.5009130795541178,
 0.09703058687333999,0.8948970945965176,0.11788950278692423,
 0.5990876154025351,0.4834336191471722,0.8089578052950832,0.09351161857302914,
 0.6517197980122007,0.2631435663992109,0.9281562396699192,0.7679784930851089,
 0.8316565590154698,0.9958127222887853,0.2377706244024128,0.28947031863014294,
 0.631178061571532,0.37809133856768107,0.055879841943499264,
 0.9308669697394828,0.5302143094096363,0.24859677680277137,
 0.10279879516079005,0.3185362077637385,0.7361299939170004,0.5095876164274846,
 0.43186303083108113,0.7315654437073661,0.3737327995098566,0.6581399003877899,
 0.5486059899463355,0.5110164102778687,0.5037177321788362,0.5209910656836879,
 0.2854413340493548,0.6830067057819109,0.6526882843347926,0.25700281804434555,
 0.35582712864367283,0.4067498181909256,0.7903998021975055,0.4490274835556429,
 0.90268590232762,0.7444648430805784,0.11449126295125622,0.7808979093819832,
 0.6251017455053044,0.9114766376665324,0.19672889483238298,0.6662542618775389,
 0.38503753321098,0.7943999509427003,0.7663437569718947,0.3352668976257416,
 0.8294632370397304,0.5000080480050527,0.3514569259223139,0.642501014756091,
 0.3302854763544665,0.6866969258510346,0.8142388521757598,0.9034151404436892,
 0.23681939186191803,0.023777385981244126,0.28580050236854093,
 0.5300480294185937,0.5509866119289126,0.1842279256347712,0.842017060670982,
 0.3339670476893266,0.49206800070810675,0.8260686330098023,
 0.43910110405593017,0.5998440804929887,0.5553632064948376,0.6352068361691247,
 0.4661687533406714,0.3602084881553739,0.7123011287506283,0.23387672066008447,
 0.9020003588960397,0.8396802599732303,0.9884108086920488,0.9443226984467112,
 0.8777774162876899,0.940902644778568,0.5688580940722334,0.6308224503809211,
 0.8774585555348244,0.11836838490796153,0.1338049838324853,0.3433294235275334,
 0.37036700958244395,0.34700119509192995,0.7244368744604919,
 0.25722026456345093,0.6813184501644522,0.4968707180916493,0.8177478796140203,
 0.07349453468696854,0.08682510141830901,0.8164493940656788,
 0.7720567286352581,0.5547190538078195,0.348494449739814,0.8633683320737857,
 0.8692611451939682,0.492316377764777,0.6448619580999849,0.09565297377570114,
 0.25051382690056856,0.02575279597773006,0.5244850861966798,
 0.4058863227029321,0.11960996421119552,0.9118537240818219,0.65458745869696,
 0.4002131270359488,0.4191089371347525,0.5324042920767578,0.12427086622293892,
 0.21387807289025684,0.4861454451509166,0.4703725396522599,0.57018019320376,
 0.70573025128178,0.46365793894512897,0.3803773492587482,0.901932374093084,
 0.42516401231024525,0.17781999835324946,0.31125393585881234,
 0.20477018293018312,0.5760045546695837,0.7328939244725768,0.6807576129005011,
 0.9173208769334714,0.345359103687659,0.5654795236695637,0.6915178895963259,
 0.3275843964273598,0.9068956903047385,0.0482548262500897,0.2721463333223303,
 0.4474661173252029,0.35610071500686424,0.7501714116116804,
 0.38373527107610794,0.3062208895431954,0.9300986649184929,0.5957076850668,
 0.2186359112690195,0.8475646015491629,0.01605256754221096,
 0.26796192707664124,0.7332110250539611,0.4680044468980601,0.1314511437579572,
 0.3756995493243981,0.4982725941933125,0.3820555244751642,0.1293582169286871,
 0.3874913769825602,0.10392392228195191,0.3959780681732976,0.9123705474811784,
 0.6046282451147417,0.09221452346325842,0.9441323016203752,0.3386782559313658,
 0.6960151712414019,0.4553442828706483,0.3576874856223341,0.26732120097646217,
 0.44110553355619775,0.32402497920501927,0.39016533609282433,
 0.5055519781555415,0.2151842003083556,0.2403487200167962,0.15979854950043382,
 0.8713575938146321,0.8870719087082842,0.1537067390235628,0.737206623912499,
 0.015539614282759628,0.11695490748851156,0.7344843218418519,
 0.6271162803632433,0.8815489793873159,0.25859871881319074,0.3499632402512176,
 0.6640972523652917,0.719514869329265,0.6171132862302698,0.00923106887940528,
 0.42090545871022933,0.26525477128157704,0.171836462398001,0.5424463143896511,
 0.06775580298974071,0.40833561143762487,0.5564213197686494,
 0.5296336860384037,0.7726850380067111,0.24002096761476355,0.3735362457087571,
 0.046475440404240564,0.5536824659568044,0.8135796279069142,
 0.5065461376799241,0.16184617584748962,0.37801630607321624,
 0.021234556783981673,0.7316946043817487,0.08343046767879025,
 0.7626382424291093,0.5896552446299663,0.03488793797665635,0.6810099087543549,
 0.03377755845481789,0.24698522844570636,0.43191779798118146,
 0.2362861939142007,0.8541780919575312,0.15894573179839355,0.3954064699316975,
 0.007931642696368257,0.8244521767641463,0.5697040232522373,
 0.11439095870427762,0.6746400264962615,0.28526218274434934,
 0.12066833923152454,1.4535404458115853e-4,0.140517017392376,
 0.5369829340187993,0.2833743227560135,0.08349080880174808,0.6657587477096208,
 0.26580828316478355,0.2007872060930389,0.10043390610667446,
 0.07317537060438384,0.45688716655130424,0.9387061198766982,
 0.020265940273642147,0.9751424086849767,0.6716786753528459,
 0.8629076858838945,0.36597869876110856,0.10487081199799841,
 0.47332448193646615,0.2037418510317931,0.5453454596946437,0.6664059657140995,
 0.9643384534459123,0.8841152531171795,0.25217221051431704,0.4004277820671718,
 0.9491141926708924,0.581534889480606,0.051964941371918316,
 0.0028978437582531935,0.023896196561060234,0.8874487042366883,
 0.5548791152719432,0.6943572264799622,0.08267047235387404,
 0.15890901528149248,0.017256943769624167,0.9994188972895734,
 0.170929237523475,0.8415066442674963,0.24838481474439766,0.03119361736112536,
 0.8624266667261674,0.6922996899035258,0.5156063471816171,0.7874673660508411,
 0.1771928584388619,0.2800058842328297,0.8254598791181639,0.25616454671694355,
 0.6795053340892183,0.5152839736395519,0.07575166738748873,0.6555734054441367,
 0.6637818774069172,0.466201154631561,0.8752209404988625,0.17873484936410033,
 0.3505487033279556,0.35223585454136785,0.5914292334143672,0.6585220736601639,
 0.48279064651136594,0.9992726494412956,0.22001821572791314,
 0.11076205450329135,0.07002091344553318,0.7457442439149616,
 0.39707991310599555,0.19030687226865117,0.2968871956614656,
 0.13472117157160723,0.3657158720679775,0.440659198065254,0.7441349672726223,
 0.30934129808375344,0.42497385572517044,0.08360686676232643,
 0.4799049324569321,0.013108110344155488,0.46028664906684924,
 0.2994585577876667,0.2701226602149067,0.4466511131223241,0.2180195557025889,
 0.6560364264243346,0.20852991146778954,0.9843828025912337,
 0.15253635856231385,0.07836683769419361,0.44325464417013183,
 0.25835449398609256,0.05336910176852405,0.3942144677171956,
 0.20149961915127368,0.320851703745611,0.27889415672280615,0.9734631707558102,
 0.6161018621234133,0.04714488621458801,0.13568173407038198,
 0.1532084066391085,0.05114028495302381,0.40288510988294735,
 0.37776201454754554,0.7987310023237442,0.24080438553635863,
 0.3348496106351412,0.4883236206099223,0.4766782427355787,0.8161800946438799,
 0.7329239980968039,0.08566239145411858,0.5993396233664098,0.8301431079482309,
 0.45659821275199575,0.5652450890785139,0.2129609016124142,0.9888202709457424,
 0.9265868126309279,0.8901621896224471,0.2791738328046103,0.5270113015850714,
 0.40771878344887336,0.922269275132417,0.7132296014458352,0.12388974376943451,
 0.1995972764410754,0.6326281374150542,0.8362715101144724,0.417213581669002,
 0.6192732731690905,0.6157466381418965,0.8621414286649804,0.9216501695171387,
 0.10884534463244389,0.2283938560692924,0.20001694379357582,
 0.6168626924407739,0.2561340348414821,0.8764531275514676,
 0.009260866849542748,0.08111595778217162,0.5153841306705571,
 0.18796858040328646,0.7305667055957771,0.5837909216218125,0.7883608878047937,
 0.7066576441808882,0.6242819647341513,0.6013761350931224,0.9702865977083962,
 0.7676140133344507,0.37779471359289685,0.9097088449730976,0.3971578768466105,
 0.4726036367949591,0.1758530100685367,0.6739728140324565,0.5232919658324926,
 0.9204845124569455,0.6936057499496948,0.776615749191448,0.18559195863637534,
 0.40466026883089157,0.8077461839647198,0.5587121314174456,0.7219028171725335,
 0.9933041728073845,0.017890751021279683,0.2390511033398941,0.951569657759508,
 0.2709896067972716,0.6067835207447961,0.16858114463829377,0.6022173924461437,
 0.5154341637476152,0.7951024072631796,0.9802516899592976,0.29728262102723735,
 0.7682724806333519,0.2177680970580924,0.792689970099696,0.19991507263088326,
 0.9110966900290249,0.8850295121047713,0.6400384912690765,0.17337618638730545,
 0.1017395838787647,0.19955732178535157,0.6855749283134867,0.4762055807077805,
 0.3135215860067061,0.9410182403195991,0.1755976280031748,0.18131907354774424,
 0.6517229061648254,0.5271340868634061,0.21106374864909405,0.3444636959800862,
 0.6558221657228548,0.6757081775746504,0.4118714604227667,
 2.2062180366944073e-4,0.3237683102604426,0.00428612353061264,
 0.3592338338801897,0.6220544198791181,0.5954096258682768,0.403850098880852,
 0.7082004461824647,0.816631306793645,0.7644616053164841,0.5050146612678937,
 0.3982036952298018,0.7212062577639244,0.6709731313595599,0.04643246632058118,
 0.8534237007546194,0.08772057534055522,0.026335811825281552,
 0.31017847903916484,0.3150649274689865,0.6508133119122209,0.8141401353914354,
 0.5570354053023114,0.9501865722311853,0.19845279745201239,0.278061235944818,
 0.2685295767817001,0.1818892275501901,0.4325742980140105,0.4323668926782206,
 0.018929033391355765,0.32604714389842737,0.054089042490209294,
 0.03691467958172523,0.1899366751548064,0.23447840745418147,
 0.3741170499161146,0.7313448285388766,0.12721700374973188,
 0.42429865794216115,0.8368903768265405,0.15289513736548066,
 0.16493185773459007,0.9988691976751888,0.0874223700131661,
 0.36749885311236463,0.9297105505622129,0.6199932167996594,0.6699087999052866,
 0.19809217707478977,0.02726432251914146,0.5721573279993328,
 0.49080756194231134,0.5434768570137434,0.56066142263176,0.8179524512289467,
 0.4076285384609716,0.9561174865720219,0.5245016478077937,0.3138183363318082,
 0.9464539136993236,0.27811950207985325,0.07802655072176068,
 0.1981001793594085,0.614157919027844,0.3537140636895735,0.9558688621294806,
 0.5377657474364604,0.5549049569188017,0.5871073569758102,0.0114500740957314,
 0.15282500902281837,0.5359200822394549,0.5105490670492095,0.6888423701701082,
 0.33044326688354997,0.6897053706058424,0.0013974114584450748,
 0.5148742160396426,0.23362865743324202,0.8361954823253201,
 0.24246204833049356,0.28962220538557815,0.6418598616822847,
 0.9552052207234619,0.3040887064230303,0.7500017472270644,0.2915440318313198,
 0.3928380113140175,0.5692689228349963,0.787091370063657,0.34294130959228486,
 0.9977678305241662,0.040707240051212956,0.7294669766128116,
 0.20246657306601346,0.7940048143770934,0.6331892418060796,0.7521192200916551,
 0.9393061086867162,0.7694385573004685,0.3403347068259026,0.2692095429555269,
 0.9205510537366867,0.8619483953439196,0.19509109582600004,0.6884133877800583,
 0.5667844381577785,0.3813078513444069,0.6880912511177081,0.7665031307864858,
 0.441747100977014,0.109581322647466,0.86390702645488,0.2779791806756339,
 0.7338801018727168,0.33862793013451087,0.7439548869016624,
 0.44252308793247286,0.08875863823229746,0.7597680251115028,0.918392829158431,
 0.12313180707905225,0.17814887372172472,0.48432485741246756,
 0.7169797099912183,0.13304987642875787,0.29784630473374807,
 0.9876350021439488,0.5135482021440086,0.9335625437928272,0.42218210384362753,
 0.4951878524001979,0.5332363431039105,0.27638713191129216,0.1389375772226793,
 0.12612881472715187,0.23221817369419884,0.565234524283249,
 0.15278645672694757,0.52626778834865,0.8518815944945404,0.18323324975089372,
 0.35238641830810247,0.1342241570492393,0.19353400810521937,
 0.7896407402994833,0.7889800260896063,0.829559332903481,0.09521262940885467,
 0.5412816475746106,0.515577683545618,0.4772255494582781,0.9643213976778257,
 0.5889622217458986,0.19005275211395667,0.014806049999735227,
 0.29684931810465176,0.9425328962532602,0.6162184170842778,
 0.31671126507354275,0.8402661717444052,0.210427820738335,0.9306810119504997,
 0.6011951624998824,0.8990909159201411,0.39758129910616447,0.8942822167577491,
 0.9919040901786967,0.7534251606619818,0.24382258053120032,0.7131813232317308,
 0.47505775000563055,0.9559435510298403,0.9686734518466794,0.3716876646815709,
 0.8174793039948701,0.5535178555814211,0.28115680637330476,
 0.08797878344180599,0.9335982679625704,0.2289821453685248,0.4052228649587969,
 0.6700422089866364,0.7741243696631234,0.751911600922283,0.9887793735162286,
 0.7031417996315853,0.04948848735823985,0.03061896218704152,
 0.3096683079452638,0.20171664334425088,0.9554859666303009,0.1748921036328991,
 0.7247938938281475,0.93125629690736,0.14525581112739117,0.7836123995704516,
 0.5863388381103759,0.016673135823930818,0.5381594905584666,0.691215045636527,
 0.9601254841739835,0.30659148552289317,0.8907203553128702,
 0.08056038621701367,0.020317541492758773,0.36435328894980334,
 0.0013517844018244851,0.33575424209889226,0.12045915850757694,
 0.1719190716901421,0.7408667187108245,0.985976754963012,0.7138411969086405,
 0.7068498069888646,0.13954331248800167,0.7972607801954804,0.2757932962289128,
 0.23433141838425353,0.91596385490365,0.1390321517394344,0.6762886176775696,
 0.47834046335494995,0.6847558090621326,0.08304112153721921,
 0.3551616582706265,0.7825904902237374,0.5711216539139868,0.4958177779681263,
 0.011639142148388015,0.0709797739059157,0.991583237227458,0.5287772884369992,
 0.592929159256567,0.08340588063580778,0.7149835719343208,0.5486317007561116,
 0.7531900071444099,0.2677056027552861,0.4555345792817229,0.5194566285976429,
 0.5271233491182704,0.2754775223625564,0.2794443249301757,0.6072917640579102,
 0.8582521203204336,0.893355102913216,0.7574116475565542,0.5409622385410795,
 0.8233055510680312,0.5237338944365967,0.0826071736900762,0.6106720001679053,
 0.3039178503259946,0.6991475303213757,0.9877282940111349,
 0.014934202849864353,0.5018703153494426,0.28371465187854505,
 0.4009256811500723,0.9850658024487897,0.8780063103520245,0.7355637218458112,
 0.5477100145401581,0.9753923539577287,0.04297005749713689,0.6381068232797447,
 0.878554016485427,0.6679599915546245,0.5465566177900063,0.7563226727276207,
 0.2529128300504434,0.18483494728058691,0.528117856697263,0.12124344161358214,
 0.8266559192346092,0.2567340306481386,0.7798960513568575,0.31648925581540355,
 0.50227562334922,0.7788732419889315,0.6849263480839347,0.6422470107449183,
 0.22299511480866974,0.5493183196836008,0.005938101790655792,
 0.666311248454087,0.15458341308974988,0.9436784061653519,0.7803009159976592,
 0.04900896029651525,0.38969257705494165,0.8969083545250012,
 0.38462411538983476,0.057342630277040674,0.8702212085939522,
 0.3089062590559427,0.6132613751852902,0.751414057542605,0.8556200306830647,
 0.6642878950946822,0.615728292402018,0.3861057138511055,0.030322125845539394,
 0.4592008186959329,0.8663138525246277,0.2536241685358235,0.19996453198995745,
 0.6772482014264485,0.5572646306371447,0.3279850641575105,0.6874244065715306,
 0.859520423072263,0.915948753035023,0.6842519532286566,0.7662320643289788,
 0.8039522972665585,0.08761409018753819,0.7215986401789328,0.5116393468933234,
 0.9408184933177052,0.734639565462985,0.6961043675182946,0.8859336036010775,
 0.8179449480863764,0.6796223216608616,0.5065012807327085,0.39287520397673104,
 0.5866581935760983,0.7581328705509975,0.7324324277945435,0.5037528857175368,
 0.5352539918550383,0.7799845851016343,0.5789029061146355,0.8852115842773318,
 0.9758760745754044,0.19672980980945398,0.245027700852968,0.04932923105143383,
 0.3045865439311546,0.06273252435525134,0.25911164337617965,
 0.49796417143184657,0.43344588924165184,0.6722858455489996,
 0.29047781102085224,0.11250075555622241,0.20152753512443433,
 0.0696041191993374,0.2927459555056956,0.8383755360855843,0.8814864815357237,
 0.06020350634569227,0.6099725733315703,0.940201381849666,0.2730280633141132,
 0.985278161022816,0.32450121185326497,0.9176754977956987,0.4726942900221507,
 0.1470426572286887,0.5066043103941907,0.24313257213866646,0.4397255209444296,
 0.3514077253315737,0.08189991544739506,0.6754110259516882,0.9568739411893339,
 0.8365438353164113,0.12549783688088723,0.394475344817267,0.4526405822478332,
 0.4667548327236881,0.6974130969991759,0.08404793767573915,
 0.49271233504526013,0.7654271978949363,0.6040751722408482,
 0.09664300821793792,0.13815768448857835,0.3577894856962047,
 0.8799832138755646,0.35923143438049066,0.030023170164421464,
 0.4291226431887052,0.6959410915858275,0.7318788246441826,0.7765363917131409,
 0.4618681921949648,0.14595027360120205,0.5639243864448621,
 0.15108330647903945,0.6551957208321542,0.6700449823645628,0.92067050643462,
 0.22289868736650698,0.7246924009148497,0.7144730488131471,0.6789353409545907,
 0.6456075524322781,0.8155058620006796,0.9529652038038673,0.8505759685451932,
 0.3122922483745312,0.9187574177138924,0.7335607143323861,0.5257635254614955,
 0.7393540118023308,0.994857157279614,0.01338907289348834,0.15120069921967405,
 0.7037146795950183,0.43969172458500216,0.1272785155467775,0.8829784676006236,
 0.322760683315381,0.4448450020644582,0.056083781473352246,0.3097576771169567,
 0.5973504672021219,0.2635257878051218,0.37627415295567423,0.9483566740067488,
 0.27893929987033383,0.4118546457076405,0.7643006498603403,
 0.48426072949074206,0.7522240448306198,0.01171409129572365,0.750813301569461,
 0.891501937383649,0.7798406483192025,0.8976205105107354,0.6848246800452292,
 0.8530033889974117,0.7729376440993444,0.7235921518137951,0.6880372838152414,
 0.8317481613578669,0.9891087903101221,0.687588803212257,0.06842343305393861,
 0.1327466176614066,0.052490604206064884,0.9456970239781423,
 0.09495337222254774,0.6959471750149255,0.29460902637832787,
 0.7970064208710098,0.5363004534995842,0.6120704562326289,0.7340209392584683,
 0.20822019797261282,0.64772344800392,0.3287465788174304,0.7709331255869671,
 0.8214201219822412,0.251436298866081,0.8732158944648589,0.31761477705815255,
 0.546748490660581,0.6578312733945726,0.3720010404995071,0.22118113370831438,
 0.11382277544999642,0.23360570253146806,0.3115938175052707,
 0.18024585940899573,0.6402280286104066,0.7420981016474646,0.9007173697284467,
 0.13304137289297369,0.8345922706641439,0.4705567585489342,0.9199231240389324,
 0.4870030973947522,0.13232521314365664,0.13228234664283423,0.919435806026719,
 0.3967419620877467,0.11021168975954565,0.1506197810120119,0.6235774483504066,
 0.8040342773673976,0.5531977905150037,0.6741567141237239,0.2915301030747691,
 0.9966826940376038,0.7269924211625312,0.6099931905157465,0.3838507618472191,
 0.7010767044797918,0.806755634060689,0.48058661209433684,0.4911800250253273,
 0.5772169870889752,0.6864047581986638,0.11091318230195069,0.9372008786740962,
 0.7173693964268079,0.05458278644437575,0.8132379083745456,
 0.24150675045704273,0.10118109702911382,0.5529315476563516,0.960010299085622,
 0.5711580792855803,0.7723032080506366,0.8768789070574485,0.20617273450846751,
 0.09695290190957917,0.05088250504915304,0.6956154822843225,
 0.32295947498173894,0.6014749241949218,0.8419666889894405,
 0.16829875629345525,0.01835156846060304,0.8106159484548101,
 0.1475913409259243,0.05960258200099444,0.7213870451335944,0.5190262875545896,
 0.010644764939533702,0.19106677946010364,0.311983953428305,
 0.8089290550975498,0.331117719114963,0.27798630981137595,
 0.020972277055611777,0.09079502247419458,0.7025094081279298,
 0.11733703435111187,0.9113539753823955,0.4492335540507104,0.810268994673382,
 0.860784402037156,0.9639207551537394,0.012619013646235855,
 0.28308607514709283,0.7937200809944056,0.06995410911146605,
 0.7106395160906762,0.3276934677568626,0.7151340571144829,0.49981137062065895,
 0.22535690200953928,0.9757380894589225,0.8441096957791318,0.1552833036111041,
 0.02693802803257961,0.15038866938091555,0.03266855210461339,
 0.9843834456228973,0.34231647957013933,0.6732984004797079,0.7202069969968428,
 0.510988676560972,0.9235803901029832,0.25940970737102564,0.29118091102071575,
 0.4231509855094939,0.3364146896457182,0.6596723243819582,0.7827349336561923,
 0.7766122601893775,0.2667532158066793,0.9725852368501786,0.7465377575033871,
 0.06945787686626625,0.404858669712569,0.15943787042713398,0.5763946125563825,
 0.9459644203188258,0.5440214695972443,0.21808449935996155,0.8143097999240922,
 0.9506133893256246,0.6841314676898179,0.5950827738809475,0.9139008505375156,
 0.2833545558828068,0.6591368026162918,0.4169704056919763,0.2581964664955544,
 0.2890530660855354,0.6945618839777193,0.24054538984189466,
 0.16207901213141984,0.5596332928119288,0.9530109395351508,
 0.35325714140650033,0.2203959622484306,0.8290821128066856,0.2585203363131806,
 0.6941149330299058,0.5036710121713139,0.20644179516435324,0.3623104852499254,
 0.6330400753777754,0.4678691658469104,0.5497165185675482,0.9759695526967441,
 0.1547349043312063,0.6241042856002491,0.11097472170694644,0.8939948447952291,
 0.28082787878498927,0.7437044656876246,0.6130971599512415,0.5534071953439912,
 0.19428657810813488,0.972090798728719,0.24301762337918975,
 0.24543850881948903,0.31232371713780394,0.9077384404418383,
 0.26942193202203857,0.8552871264216584,0.4502338159473054,0.7181390044669551,
 0.7436169353770066,0.6718030451375051,0.8292037958280772,0.7766694311901026,
 0.4052275127745295,0.11231413329819784,0.7215965769556034,0.4957638170213783,
 0.2915577902301092,0.2690072814188673,0.9091622717172018,0.38497276526192226,
 0.8301279209864558,0.5378498764223737,0.12646712026475904,0.947212418382476,
 0.3081260485831191,0.8482547295427789,0.9580190662257203,0.7010108764252938,
 0.05235157432052018,0.052269297009009286,0.8035764910072013,
 0.6523199074037531,0.8213762355412688,0.6367367236850491,0.995631681977317,
 0.5869014695254122,0.9916899716661084,0.7384850572947486,0.7492010673478169,
 0.5234717897424854,0.4550016822724907,0.7005233727619553,0.09033865889096282,
 0.7319659704540765,0.322522776676458,0.2516035510432586,0.927632114827168,
 0.19497316517014918,0.606887816446279,0.38230553307123016,0.259689323601036,
 0.3505841647663944,0.596537721226902,0.6827255666053079,0.4166633595296614,
 0.5597579698371303,0.13620222998911868,0.7804112126035363,0.9888740868129435,
 0.8068501760913209,0.20488435482713963,0.18180416299528235,
 0.8090501309273019,0.48192397839792356,0.4394822364159058,
 0.09562245009216042,0.6649883500317499,0.0020936563800840746,
 0.3057333221176086,0.6379788896211117,0.6354136805490884,
 0.014288420764091647,0.16488502982255515,0.8864535416925925,
 0.22432042392391516,0.4583855480219279,0.7873710135647192,0.5301154350584765,
 0.8281620148313026,0.3254601341577107,0.5509788836308549,0.6953123752747468,
 0.3241606820982199,0.8479222807799128,0.2372963191315598,0.5105241587613962,
 0.41082210095957805,0.740593740677338,0.1179805434535054,0.9701524184771775,
 0.5017389020164325,0.39687575160847555,0.08789418013587535,
 0.9037016270808925,0.5239096414296645,0.40833698761218884,
 0.10299683668130988,0.666156689831046,0.5667564042235382,0.4556561865962906,
 0.4266483363665443,0.23803343370316632,0.272208993293788,0.35349832068646403,
 0.2559150522452689,0.5629825378246806,0.672449948323214,0.7823589362535907,
 0.11299060486065848,0.07692875122286669,0.3280233069140993,
 0.01921501168033113,0.7053930065709395,0.9624187199344245,0.5128886589332488,
 0.88728373994718,0.9506125177359638,0.7289534395709153,0.8564470122591558,
 0.2066488552361393,0.010330163936572156,0.4487465494286037,
 0.2963490615555959,0.04845015575614353,0.5042464291914419,
 0.20767054835474952,0.5421084253797577,0.8373787342361148,
 0.27614510731589925,0.9243353485856308,0.6763264650124162,0.599314666976266,
 0.3592222327886394,0.36346466025743407,0.5964933489522205,0.5794861058320563,
 0.07773327116409323,0.9254555860610936,0.15567231696566453,
 0.017690719252844156,0.20363135760835316,0.8628415170488436,
 0.4315089332606733,0.41622839396057865,0.2814733058404262,0.5644822387931703,
 0.5881028313290949,0.06745551879206924,0.3523758490242005,
 0.23760058557133945,0.6555364278058412,0.7712891491107164,0.7441724213647966,
 0.11767535719943734,0.5133046003690535,0.8991088663826037,0.5872928203350741,
 0.45084652918614243,0.3263610869328126,0.008285558776085278,
 0.8333224308868932,0.015415023652710302,0.9354789336332797,0.501574091120016,
 0.5066012469153874,0.48167194803044056,0.5844200858642825,
 0.22574617188279222,0.2206443470133922,0.24101193419662492,
 0.7884729156161164,0.09987745150059069,0.1368095854211384,
 0.33796891477824276,0.5306094015463038,0.9125455406070621,
 0.44392955267347367,0.3794872562492697,0.9670916956889993,
 0.31339802930832084,0.1865661997085205,0.9195027671883543,0.1127980881550934,
 0.5340087499283324,0.8751480726643612,0.28905801207871473,0.286219649869322,
 0.9376585903658432,0.5825177525051656,0.8037628706848725,0.141749829926614,
 0.23005113812162747,0.8404665511256151,0.3000433957033669,0.9312930490616863,
 0.11162360846501207,0.022581242730467688,0.80193099094883,0.2755902484583961,
 0.4092913070623153,0.12161054288425222,0.3794910579715405,
 0.14757344310551312,0.629532227262311,0.6492804590222799,0.2885291047251085,
 0.3012296473128926,0.5020769978745079,0.29581428817615996,0.7594691656822509,
 0.8697577206984798,0.9008494544404498,0.8278662527389281,0.8541456820326624,
 0.4307619500226706,0.050370401770305606,0.33564131888305093,
 0.5453100487011033,0.9723098287320932,0.16663309099098567,0.9955606776899686,
 0.906814685304274,0.20392761552199612,0.38420531660355584,
 0.49621377742721473,0.8551874500397101,0.5130033948080882,
 0.10067297457619584,0.07775762588031998,0.8453139329921906,
 0.8693683413336472,0.6081267006760831,0.7140508007554405,0.02584566745718364,
 0.6458945511730183,0.05863462555216459,0.20648770674717254,
 0.25158307821803527,0.37030684996215435,0.9394933732891766,
 0.8568906344223509,0.3156564154022913,0.4295862604929054,0.680732906655676,
 0.14014816541130848,0.40645152393554673,0.8017730345298032,0.609234432758506,
 0.17707370986078497,0.16065058466708093,0.3597824649523911,
 0.38585210420905025,0.5461773124507553,0.6709169529260183,0.0260895917293843,
 0.5443017964483107,0.4152639269364268,0.39729175613843404,0.6496570530234459,
 0.6145424248637357,0.17078661031616638,0.9376521352997466,0.5674035326421221,
 0.3355709344148702,0.3309016422473178,0.276151193577464,0.12586374481265794,
 0.8060037287060354,0.13395548510101185,0.9164153889011973,0.5902139893151506,
 0.258565151315181,0.5881441456005793,0.30261278641665446,0.8367748025177226,
 0.5136613465754214,0.9755116986228719,0.7155903250423907,0.0394604381263689,
 0.6515225658616424,0.3436155696085288,0.9191826269715103,0.3620922194027387,
 0.7957019934059569,0.10991215177735203,0.18449871518021033,
 0.5318799273357605,0.23573094405435047,0.6670653158509191,0.5129628312860908,
 0.6598670246867722,0.48533858217207804,0.07846963076037505,
 0.3574284059158769,0.5044359562989889,0.867426992308189,0.010659196914889368,
 0.9122908889640887,0.9945376587388722,0.9666639423876268,0.8962337327621787,
 0.7816138136992015,0.6916304638466523,0.9886054382568426,0.604374853002232,
 0.1450548177330525,0.2633802342298438,0.7753879031915377,0.6386132674396342,
 0.8003814941131512,0.27699306661504663,0.45284887601326496,0.97293465092626,
 0.04517129707190948,0.4486854683567283,0.34730596082876597,
 0.6935193549870924,0.321607765099017,0.584414289652469,0.9404064226980938,
 0.2798080615596188,0.16517828688282488,0.7555874664552151,0.6593573427774698,
 0.9842678873162917,0.7482992718201906,0.7516796693464116,0.3108732018085396,
 0.7949153641713748,0.8600455630797035,0.1710571741010345,0.15297317015631995,
 0.4141174687898357,0.06488077659109082,0.08808374499843785,
 0.2666458589842279,0.3609010932338629,0.4876348831553524,
 0.055646182998215554,0.18626572467950586,0.5316936554190361,
 0.061065600713810064,0.32280651955312645,0.6267884535538422,
 0.7095122037748643,0.8924445147289527,0.7287676403589876,0.38715385460222895,
 0.408268410747465,0.7347838604832212,0.639088419700557,0.15288680405226163,
 0.8695054260631605,0.07116411927364719,0.9383822535834658,0.7177564338070961,
 0.6110469631629367,0.3192592353445911,0.5412546347822038,0.992329892348691,
 0.9757465942397138,0.3538696948258312,0.10312835279466992,0.8227928132447899,
 0.7958508642288304,0.4641368802969351,0.12206690353602823,
 0.07090099867820465,0.2990721826909237,0.4719816705558203,
 0.14717396505305813,0.2466343637578684,0.053388426497814834,
 0.2583079766725077,0.5816249435701952,0.7611038380133355,0.9722048718998678,
 0.17364577635444567,0.8359073619655016,0.3647091623684634,
 0.46569375198863194,0.15855329696305542,0.2802965912278669,
 0.6826525837523038,0.8366078206263872,0.17575836590660998,0.4702330302127692,
 0.5965099767859744,0.8731535026185943,0.5750757573941926,0.9924016564920208,
 0.0564374925109109,0.21077829063888487,0.5934344700515276,0.3352363869664067,
 0.6898341333985104,0.7911302586476339,0.6944694583728214,0.30415350853791645,
 0.33129969603729137,0.09444184933132715,0.8496301826988313,
 0.2681107260952642,0.2096718552823943,0.024109291122029863,
 0.6922102896315616,0.30680900461546123,0.8087632835014283,0.8817596496905029,
 0.5582036335030364,0.34598585680902194,0.426839546895301,0.67773408638538,
 0.16003414192608012,0.17442541345681373,0.4171825309056151,
 0.18471715683705314,0.7323619311669438,0.5865926526359734,0.8827338402348844,
 0.732798392005774,0.49586674678996967,0.42704984692986636,0.9934792110537587,
 0.42201158959911833,0.08105248695668787,0.6181888926075492,
 0.1244925435267582,0.9222505410115354,0.24330505086959253,
 0.10564411350163982,0.4374398526933907,0.24089638184332385,
 0.9827599451218142,0.7322517050077892,0.5650609028883592,0.1719260057416938,
 0.49933105761799257,0.16075024318794062,0.555564670419814,0.6999295047199676,
 0.30250211627532775,0.03295009443371366,0.3360410256941062,
 0.22393872808636928,0.0246503988089275,0.06499179068981753,
 0.8457657668395746,0.5358874794260476,0.6176570643987834,0.20117944323492254,
 0.39118333641733627,0.47863920285018924,0.2240956437129289,
 0.6259024226814686,0.8145285389417546,0.26369242548278526,0.4461730271662703,
 0.7111401649538616,0.5955255128394326,0.7025627202458913,0.7658497110966702,
 0.7437966784438034,0.06242727776993984,0.2515131563446008,0.3759093195543486,
 0.2753065521198904,0.43252750241862303,0.477330795665866,0.6343757547292113,
 0.8254868840281185,0.9731913244959386,0.575060518250895,0.2717542594316975,
 0.30906202422101914,0.1524563515306342,0.7022199798239797,0.7151582753887429,
 0.8683444624649932,0.5917463239921263,0.7414192483966197,0.01760284592080885,
 0.3555890003902613,0.4046215180909246,0.5829853873974835,0.07072514012540743,
 0.4641545737221783,0.042619497140470974,0.10591359699017922,
 0.43653580896094235,0.8555305447349939,0.5027415131406628,0.1996835488201948,
 0.20200859131502913,0.7712911517370638,0.7032021172427438,
 0.22280704816206331,0.4527810656101686,0.8276711907008379,0.7656078834063084,
 0.3212005712007303,0.037381622383446,0.687803848290613,0.9529989954175286,
 0.8656493792072706,0.22697596212448679,0.5467174162582814,0.3586667842452701,
 0.32316552638576335,0.48037296169502897,0.49659844467937786,
 0.6098708548537815,0.1512857507711065,0.18393253239822027,0.5676439358588723,
 0.4404028890626489,0.22371685095015648,0.04784481725444867,
 0.2457665472153786,0.32328942421456586,0.18043161918327266,
 0.6863549535890343,0.9852505751473017,0.011015344778823155,
 0.6410095401509298,0.36835425490722973,0.9084779221201499,0.6389306716888097,
 0.22921518436789334,0.29341304058748907,0.967072555366805,0.704177101934181,
 0.24219976693038325,0.9130856438278829,0.21307737184130515,
 0.23101683793449235,0.8609463105882311,0.5521723434381064,0.2985001244805239,
 0.786479514401243,0.7063766738153467,0.33613994708863315,0.29337273739990144,
 0.7030847806661241,0.521766140985821,0.7239302673738554,0.5782976256004859,
 0.7365120966345101,0.005162166018996772,0.7368669887996963,
 0.8758912377108361,0.6326921439359026,0.2992615819955139,0.20571011849714416,
 0.4499309142317347,0.5756969925261562,0.38756036285821716,
 0.22586404946755945,0.7413515626221482,0.26289703280836807,
 0.7352916504975268,0.3366779532958686,0.4936362969721563,0.11710737249786352,
 0.9408953213076707,0.257138522434214,0.44927741336665605,0.8417945875303993,
 0.8204374192502758,0.257380255665539,0.6199248155566675,0.18835137217260445,
 0.030451229698360227,0.42392109401837386,0.3617225323860327,
 0.21718998873721473,0.9001633630850211,0.9879144839233767,
 0.06663976560418994,0.893529592175643,0.024296413378763204,
 0.5291150578510277,0.6689898946468397,0.29399400023062805,
 0.37162442299787446,0.9564333068572427,0.46592619584464756,
 0.8427437824198182,0.22489071607011346,0.10353721560212681,
 0.48142752419797097,0.437407211457831,0.0016350187760437729,
 0.6205675167579611,0.8724436303773273,0.9153491183254567,0.28383057370673126,
 0.866200520819391,0.07247246321592626,0.031005771716893782,
 0.6646031107171096,0.845042856888949,0.9339322759516655,0.1635555383983366,
 0.07002216162726471,0.8453110690576543,0.15520325672758106,
 0.5093868693293916,0.6042435825740214,0.31094754159781457,0.6261213935911529,
 0.16992492108103618,0.1290214892012359,0.9302268706904375,0.5520556043629209,
 0.3872076425014176,0.8926435148119758,0.9240890339573908,0.6526243499684351,
 0.7206673544904187,0.32744887857101934,0.11063320182102021,
 0.1648037459818028,0.3551457930294455,0.1489830089278874,0.2285293517567364,
 0.5041979205740383,0.8684065466649666,0.06384644736135414,0.6802461033900117,
 0.23253976587758263,0.8928146531563099,0.8012567648470743,0.8703912005433555,
 0.39913681236147014,0.7669221853201399,0.9431633300539016,0.8586792480657824,
 0.006468215584915726,0.3677791711002515,0.9257451044258527,
 0.8316947661620121,0.11926145534387267,0.4481481280559453,
 0.25843186867256907,0.7094904864124435,0.06159982056762536,0.550506486038441,
 0.10286268314209168,0.5465757277317314,0.5054444534034641,
 0.055632537881001864,0.1279090184621704,0.062034731721186454,
 0.7251780275133711,0.4135769206482215,0.4489771346657917,0.2830285430013827,
 0.203243644765561,0.44241363765730335,0.3311669980850547,0.18683316476607348,
 0.5679017646715688,0.8264396111300427,0.4521454436127589,0.22324513545510472,
 0.05698232104880363,0.9849244200909645,0.23361608381344834,
 0.06751208139244158,0.7108231127439724,0.7831773262216857,0.7622597116664819,
 0.1309836502531312,0.40288601881948793,0.017636306442082805,
 0.32768986259353505,0.28027439434082935,0.47696050849273053,
 0.21095894853746266,0.21901579580712272,0.7982770962351596,
 0.07694131718755026,0.06719423746431619,0.5788768932600931,
 0.7162295955645578,0.33511209312987167,0.24298252171523071,
 0.5275534490408057,0.7586130191960723,0.65112969075087,0.5993702454933789,
 0.4188517981868012,0.8000064585034111,0.24283174801569796,0.4967616123058868,
 0.9706955068012563,0.4142421573506735,0.4144372975955204,0.7279664845663446,
 0.35677436841069776,0.565325374565991,0.27581863182776933,0.9669856030856142,
 0.31449524575437526,8.2371487613353e-4,0.23590687514329867,
 0.6480526913427412,0.7219368235775937,0.34996885106618625,
 0.36999608646441995,0.26780390947553623,0.3319200879367392,
 0.4774386858227372,0.5217797279397427,0.12027292296167869,0.30626405653681,
 0.3671640926140818,0.3075469943747742,0.19334533586324887,0.6093014713470645,
 0.8963560376728108,0.8965675558817353,0.293457496509667,0.7780633890252433,
 0.3155825957229602,0.7057086097682077,0.5376418377508179,0.35814773699960256,
 0.923151676379414,0.19431299597363394,0.5888802163451069,0.7876283868390892,
 0.9642646133638965,0.1421964492230421,0.3023731002234876,0.07019060787310605,
 0.32219749672946185,0.9019625738412242,0.16360384051810084,
 0.41306041638850644,0.3123425110693159,0.7467917265729853,
 0.006453655111321233,0.5458337368930334,0.08640778055569054,
 0.4925293926976807,0.6672048456580376,0.023064945961169148,
 0.6378255433614867,0.997098716620762,0.47035104714111153,0.1608037555256584,
 0.7501494839760495,0.7475389869294226,0.9921975401364969,
 0.006001645354159479,0.47678432543713056,0.3521138114057456,
 0.9729959505770949,0.5146489814979814,0.6228095322817937,0.6520329260343184,
 0.952948309331311,0.6498725774098665,0.03630535235498633,0.6028573063342262,
 0.2791161885025112,0.23667619974304377,0.34292375406730147,
 0.041712755482953945,0.7567199324165796,0.30351007749470515,
 0.5501172916885577,0.384026843042801,0.8752271564289578,0.6181298198905267,
 0.8209533218039596,0.4695615199776627,0.7123740990451702,0.6528629027565963,
 0.9782129686645726,0.1549650485473293,0.42741367717220713,
 0.21809572612225547,0.5487308416487834,0.933033307586835,0.5024417102391205,
 0.46168055662964325,0.974256141278561,0.9051768633730337,
 0.024450762698059902,0.49113416533449994,0.4617552672955112,
 0.04799886019704991,0.7881536128950328,0.3480440450435175,0.3235047378435395,
 0.8017687258366926,0.09248817068041348,0.278833533260874,0.6202135057625812,
 0.1378338268125373,0.1254096167121742,0.36841176417648447,0.6368954456041813,
 0.8452702067298647,0.970313734715876,0.2482667513843514,0.46493832865189477,
 0.24471302056569644,0.12495978008493713,0.7853272098039453,
 0.44590568731232993,0.4616174861851047,0.7724637557724818,0.9726706662761502,
 0.9836977216888121,0.7613897452982179,0.9778750669985763,0.9439004012224151,
 0.3604326556031121,0.5653494539229902,0.25240583884018664,
 0.052062284277532456,0.6047985355855636,0.34051088759643944,
 0.4128944490109039,0.43447875303757555,0.2901905148835805,
 0.45828041260817065,0.7928752836156444,0.13636993444803291,
 0.28436729269847494,0.689650820652161,0.0473441160311594,0.37367214905478274,
 0.4804259348931089,0.8035778782099524,0.5493184923534707,0.5811639992162917,
 0.594286696764545,0.0137959628436779,0.8753048739789983,0.6323478212756581,
 0.6050140397708396,0.9594300128403047,0.7335670540694906,0.9049108891219785,
 0.677125771657987,0.4261709693788496,0.31059481745462536,0.06385707848238542,
 0.2782940386398779,0.8445576486253813,0.5520875374841225,0.45102395522905236,
 0.11166484272857946,0.499514801740669,0.30300394495238536,
 0.20217697255146683,0.9418695942590971,0.46138071651262014,
 0.8579625391806811,0.8068345086872715,0.32444229802658964,0.7520208467536008,
 0.3503896052407207,0.22278177205832927,0.16308130803163123,
 0.6346418756520921,0.4221668272259519,0.35734581876966776,0.5816009535096812,
 0.09409980495455317,0.8680500554437476,0.7601515713380077,0.5287963747429713,
 0.6601537142633455,0.5595253377135954,0.9210478507325691,0.8346789589869303,
 0.8755847411593911,0.6998595633617749,0.9593915189510402,0.9877477697217262,
 0.2141041891275688,0.9643035428309144,0.024691070702742346,
 0.9218968100340855,0.4021174074460777,0.4908771305922395,0.13776086476465244,
 0.4144453216616396,0.9812610572653063,0.968001181807873,0.8967884042517836,
 0.45211496384572225,0.5092307779050329,0.5310733608373672,0.6761922138556331,
 0.3391547306298969,0.1311827440464144,0.19194010607375822,
 0.16632639088765477,0.8141862012660402,0.4338567617035325,0.3985486460884956,
 0.9388739195359841,0.6316627834488203,0.3974429311352624,0.3700349534868549,
 0.8323696238199519,0.27800814775430827,0.6393058910743465,0.3554593572909349,
 0.4505358888789819,0.9794665571201605,0.677578293499476,0.6315546967893574,
 0.37602934055493875,0.36000321432873517,0.9095254402863784,0.619802015394932,
 0.4621938217212057,0.808372310445077,0.6217154021734012,0.5758486849469341,
 0.4394296221171201,0.9010861037146698,0.18037889209898383,0.7564441175910875,
 0.5399803282121023,0.5236399421876395,0.6019461822039188,0.11248467143874674,
 0.0642967375345137,0.9179332644951751,0.14820232608562744,0.7635983351838734,
 0.4096436825172438,0.852192779039851,0.4101320478354755,0.26449466910221586,
 0.7912556137619443,0.6526447965405748,0.7291198807903883,0.432296114250196,
 0.6987168461315798,0.6724422259027097,0.5139337049552378,0.41029277613420767,
 0.32887981400959687,0.8643648908856959,0.16689915489350338,
 0.6877366756480038,0.8574735987595554,0.1716604031173533,0.9458392337129228,
 0.5822799281338672,0.7247066930251098,0.3577231622749728,0.05297306680347735,
 0.27211883848764806,0.12972046989521835,0.6309468244386276,
 0.9527628531928367,0.5627480606926433,0.1657256022470841,0.5600795743545138,
 0.7869261806639268,0.16919217907090756,0.9926819112052934,0.9452365607634563,
 0.8298535883729778,0.06785136507707312,0.4862360030681202,0.810416591553199,
 0.646909703729904,0.9138184585988014,0.5780423806931616,0.32271618747140796,
 0.3312514764473393,0.5133743789352234,0.7872477545071885,0.4091255847174903,
 0.32564826975823347,0.24469935816539223,0.6956837505672011,
 0.5318395290911289,0.3836120200827936,0.8698528848541143,0.5165928034157427,
 0.36777004956539605,0.3421247710641877,0.1548448287120241,0.6005426735837319,
 0.7031244173833328,0.47950643073699306,0.3541150949981462,0.4426737013000659,
 0.2490617816106544,0.6411119425186339,0.63406287344655,0.5199091897996629,
 0.3678805199275108,0.9594861134599078,0.0774236513959885,0.07912135026634493,
 0.6224164964508312,0.08013331133149526,0.2778953582838123,0.9139759561537429,
 0.07533179047689176,0.38488382626692685,0.4776818599333371,
 0.07521861544938657,0.41830008421554543,0.9658608425270735,
 0.42700959987517795,0.8628121754025158,0.014134398647727142,
 0.4429936940975253,0.08739229193262721,0.4631995176683722,0.8223696420930158,
 0.6532945513782855,0.27295240299021795,0.319513179328033,0.17565196879761402,
 0.5484016406020154,0.31209212930285424,0.2691453460962403,0.8688034931977837,
 0.5211688834025613,0.30178697496532836,0.9940441983974883,
 0.06761732935029385,0.17897176083764899,0.4103053487496635,
 0.036572012027568435,0.9375768548750726,0.8209076416587381,
 0.7872289723560151,0.6109880717487688,0.1522972809901577,0.6041039699792156,
 0.49556280013210574,0.48621409462292486,0.8030521875198736,
 0.4051237108837661,0.07495763592120663,0.881355468467144,0.8874610246060585,
 0.8357108334446928,0.6331440339894855,0.7543460578452814,0.2955398848515909,
 0.8083461466714237,0.8301124618769061,0.4354436962786833,0.8695276266662062,
 0.2780533019343764,0.8738656840340876,0.00849038531405888,0.5216309829286563,
 0.7487508328136051,0.899465555936686,0.46564510711402896,0.12544154926544038,
 0.6736999512329044,0.0065500389824900696,0.46463640955791463,
 0.832756307215246,0.08372293514773765,0.12469848774020797,0.8353042753337553,
 0.35019101084674487,0.6288521188174734,0.4353968609944249,0.6151306659237598,
 0.10925397018000282,0.5576408177480051,0.5342834417253027,0.4386162513601405,
 0.40380429031523335,0.146414666939908,0.032643584346856436,
 0.3552203743348098,0.258279903542993,0.19598928935575133,0.9783987304142745,
 0.2439948278683418,0.8185858908891823,0.8374756696885278,0.26855859884731914,
 0.2707418753370002,0.4033388087584596,0.745720801910492,0.3686396918155558,
 0.03301434494189681,0.42087508668506846,0.6656333996346158,
 0.39995565127813215,0.7597101077544869,0.8505349221539145,
 0.22828797690210145,0.6373005347786673,0.5764012394949413,
 0.05598616284893276,0.0965502418443751,0.16493923047405512,
 0.14544090200111937,0.6858213367332748,0.4268255839855827,
 0.20241509530323099,0.5614483269794347,0.3190919481163632,0.9089409031651684,
 0.30257463992052647,0.5278927848520963,0.9194676789358581,0.8178598954632316,
 0.3174296836824211,0.7362282258343612,0.9178972029470516,0.4840362218583425,
 0.1454443400397829,0.45027204674897847,0.037359079915314175,
 0.7495647752547216,0.9899358443902786,0.14080011879512866,
 0.36258738158568027,0.6109971603760537,0.3626944970129927,
 0.0077065427704774425,0.7136973271053266,0.704814131230158,
 0.22193895398550512,0.014393776091078836,0.59321129934148,
 0.47443077950024104,0.7114904817857031,0.8435606664261883,0.620452074440769,
 0.33685755296272846,0.7467317854309309,0.9848517777523311,0.5088250722218509,
 0.22131423982156484,0.36607788788352746,0.033345923523988685,
 0.02650200257983626,0.373462458359179,0.9515776666117803,0.14365785738280945,
 0.7824497274382555,0.30283830479969664,0.28298740356989227,
 0.2914277966144502,0.6337844272345211,0.028248038843775913,
 0.7052681692627314,0.08353036912575074,0.6174231540109567,0.8486655515265021,
 0.9666054064548628,0.7272236934766619,0.6924711755223427,0.5002744645478853,
 0.6113990627757558,0.3379820365148878,0.8173845337912949,0.12191707515991512,
 0.1613394666836654,0.019666398247515482,0.4368768728388507,
 0.9198522059381273,0.6926373548789347,0.4966637062475179,0.496845211966254,
 0.558085822427651,0.0795198965622822,0.07497409960376678,0.8172320633339137,
 0.07254134566874726,0.1938420195881213,0.658010287675039,0.30530551668975026,
 0.4808243353115357,0.9589135862465864,0.5048441041964598,0.6890699782276876,
 0.8316493983296379,0.18890219507743655,0.9443811049534754,
 0.00838926021577524,0.04897015457880094,0.5362709566599846,
 0.15851250206237233,0.08880152276241293,0.7056632930962424,
 0.7211558826468544,0.73653279330917,0.26220270645972465,0.14407972790521806,
 0.7949545553111834,0.6138644519949991,0.8188360382260305,0.8316786632371176,
 0.33854590712591626,0.16626939702017585,0.4269139986360415,0.138650036461742,
 0.445504121053207,0.9340437146170284,0.20804724515367778,0.09133832700990485,
 0.343881411931525,0.6959991534637937,0.9719059645829906,0.9080852863239378,
 0.7514106940003786,0.421884438269595,0.6362496545501746,0.45931803466066656,
 0.4785472827842261,0.940060061115474,0.1815384290708112,0.8215035679772754,
 0.34071468758319856,0.28224533039871846,0.5251671465596552,
 0.1518087902754508,0.03125369979846726,0.631670322195277,0.7186654547973772,
 0.6897402085106513,0.6793706535009554,0.5740167524646284,0.24894290194726032,
 0.812031607894736,0.7816079746326022,0.12909655293666578,0.6481166900715332,
 0.15062414493361354,0.43388208271364415,0.6158333942952183,
 0.6623920102081776,0.038217899322198656,0.062241617178919295,
 0.7966064966350833,0.8424503084424689,0.4129642012149489,0.3450341687784674,
 0.771259343371415,0.15967696966753253,0.24184296255670035,0.7504125925413648,
 0.08774256701693262,0.19684991958806286,0.17149426060623418,
 0.14324002214959775,0.8241513915497615,0.1414181690066969,
 0.13772182447738357,0.7480491366277091,0.024042415565906694,
 0.46367314707241125,0.19617164191661518,0.2897923671211833,
 0.1883956480417962,0.5348739495213154,0.4658237613009776,0.2056368578482657,
 0.6450575665579004,0.6175180580405764,0.5207839118572513,0.08705188885553972,
 0.9125583741773694,0.3480782686263104,0.5516115125482143,0.7161797837130979,
 0.9004636031161273,0.053471663762828414,0.9872894973230036,
 0.6438345437247945,0.212087324582509,0.9399457423773504,0.5381285426773849,
 0.6167569687794873,0.238331798344125,0.6062353140003806,0.7340424173084735,
 0.7055703747220182,0.2668315691336878,0.8621170612810266,0.19356163627782386,
 0.5119681056686368,0.7589759045752345,0.2187512091015561,0.9753507819854061,
 0.8706032403870239,0.2927338735546062,0.905664786268755,0.29346845199724014,
 0.1302729675313684,0.6984153603368214,0.18939276522191661,0.5242713818017712,
 0.93368870806311,0.5181322818906505,0.47866305260246467,0.9428312652124214,
 0.8986411335517519,0.8836214987747734,0.7791344202500242,0.4236872121893427,
 0.9292431671861178,0.873773205296167,0.9058870828980544,0.9169370433083879,
 0.07403329666218272,0.10144429075506267,0.3811451422632215,
 0.8296232696065895,0.4790745167529231,0.9494950674542816,0.28318918442313923,
 0.8584830739476699,0.5888233752887777,0.2592921190501506,0.32270951527037495,
 0.6910600206965816,0.09496369012591899,0.09618727667935034,
 0.5739297947363383,0.6553354634369852,0.631025374495513,0.6988638363649635,
 0.9621360249701676,0.6278950836692808,8.398967744055463e-4,0.820305922546876,
 0.526367921562327,0.3333306552341222,0.9658352625387293,0.7511039483197282,
 0.15083062802028324,0.23392144227211276,0.8947510241183302,
 0.11568319224953894,0.31638641976217763,0.15252675117355163,
 0.4908803436591752,0.9278894820832744,0.3044853087048037,0.9301463159678277,
 0.18389502982682793,0.11724207527601871,0.6748395341275967,
 0.4360852809088628,0.6242007588122963,0.21111618608626648,
 0.14939680583528236,0.2921910974025822,0.7304761407495248,
 0.37121683217167334,0.345822981481959,0.8224639712096254,0.9125819147132296,
 0.6969915247455502,0.4756083304932386,0.418963201131084,0.8247559519420607,
 0.4057799727675694,0.10991571248590948,0.6659981095307482,0.119836881929971,
 0.583907522795938,0.29157150958170786,0.1741169231850499,0.16770657150441926,
 0.8133221002673957,0.45589957977998397,0.5240712223423543,
 0.37733764006098025,0.9262486188147858,0.6284864896955917,0.7465781517476952,
 0.26533853384621886,0.11114378776627365,0.6019801288386277,
 0.6543018701821318,0.9688771572020989,0.6783173636782175,0.379030659245278,
 0.02764913015233117,0.34768553856760676,0.7370401683547114,
 0.9187658407371878,0.2646192113309962,0.7656391708368981,0.10235349153025863,
 0.7875512416341705,0.852784175199793,0.9173348374529705,0.9677728899924389,
 0.05985700049244458,0.1067989828077297,0.9493811830661563,0.495659903884462,
 0.7591992952922779,0.02099831216606829,0.47766333345917533,
 0.3240267080161702,0.4454738004589812,0.5429441486844024,0.3328482468622851,
 0.008751654793764163,0.1284459627524812,0.36279872718475514,
 0.7411104405158274,0.02442971551692652,0.5204418992647306,0.5377716278045088,
 0.6341448281543004,0.5460161833851539,0.3828977723364221,0.8758074058002109,
 0.8310313428684218,0.6750296070197651,0.28157861516900007,
 0.09520769413618169,0.6349171322189247,0.530518928185446,0.582047993700994,
 0.949665026503909,0.46349468352167755,0.47209201567542314,0.8267321180566803,
 0.2797640963493412,0.5334981951379839,0.8833592015561602,0.6382991660035499,
 0.591751448218381,0.5774292356849055,0.17498970761177202,0.0820473673831511,
 0.1907974621358769,0.4189139828356325,0.2697444323170254,0.35851908862767656,
 0.9992658448692427,0.7239668043915762,0.23896884966147536,0.8680523559895025,
 0.7807700816290738,0.4845418806160322,0.4871395149106357,0.21729084057020254,
 0.8947369133203525,0.004142813362847719,0.25714110573407833,
 0.8880176114046582,0.21401599826179918,0.34270495874076634,0.770214081655325,
 0.8893347952848971,0.2502293428609862,0.9116574936676493,0.7776124426524278,
 0.15198135795482348,0.4356739380787802,0.2108617748324384,0.5414048694168512,
 0.44187729795885655,0.8295734488986348,0.9786975928450358,
 0.005466459183873895,0.9166932559327386,0.6418172315188125,
 0.4780095312544572,0.32726797110134653,0.6165535426705321,
 0.40590272065960187,0.79795835969105,0.0897846789951402,0.9271407566401086,
 0.157705455072795,0.11622305168294045,0.5381051148585325,0.6134952768085962,
 0.6532009948815567,0.151209929887009,0.19162944619069966,0.6878463491108523,
 0.9121765526438451,0.37032677886676146,0.10990146914555732,
 0.6608591684812485,0.7154192420246823,0.19901611096868732,0.36126787668544,
 0.6539249435532108,0.84975378291243,0.2116118289790121,0.050140152382909364,
 0.6169145674365295,0.7728930229604715,0.2428381458455563,
 0.0070580915609139705,0.9391625688530388,0.5014403155462704,
 0.35993920422058756,0.9803261148920263,0.7338491382195234,0.6998519233267521,
 0.549227477297794,0.6746087851613106,0.6381798932812387,0.4175131912697355,
 0.14352043084737698,0.12231916117601727,0.7653179434290838,
 0.9597089899586154,0.5989983700212899,0.015088664719753564,
 0.7930740471881091,0.9683635884139005,0.480368065623372,0.7337715528771787,
 0.5393625097143693,0.46235142164921084,0.20871051605179636,
 0.42901860091413535,0.12527813254825249,0.3629654217808821,
 0.015367422519565688,0.22383460611179007,0.357397625097437,
 0.8398616447270234,0.36060356805624283,0.1424435801464109,0.5920887822718848,
 0.4819700810617703,0.08800176657478587,0.2856688876911456,0.9073553299540493,
 0.5105233837293163,0.8131736048185795,0.49414823689358234,
 0.44808065343430925,0.16156278739016405,0.47288481728705156,0.18189074068818,
 0.06920500662974338,0.6400375822822935,0.491178395510653,0.9519268258887196,
 0.6206035451008628,0.17799608577311954,0.6484814695980888,
 0.09107685470237903,0.44894263778843724,0.9692460141181941,
 0.5906030734337903,0.912729054554861,0.3641737278953673,0.1746656295254031,
 0.1854734633788554,0.882589611879095,0.7302350332358669,0.15459316727292605,
 0.1915384662622036,0.9930810872712253,0.6863693457495526,0.5957477455356956,
 0.8549863723727881,0.8332083302561109,0.4463140031571432,0.5509992658016813,
 0.34872548633872946,0.3813191787602591,0.6804585668403296,
 0.24656477235702512,0.161328025062086,0.6028370443425428,0.8996937952114974,
 0.7776392704747962,0.5340967391692618,0.5160638483485911,0.1311390040580782,
 0.27353338915683834,0.2982644145358778,0.41799792263003244,
 0.2767295541188862,0.21945535076730605,0.374031075798015,0.4930332626149173,
 0.05334937721143518,0.7950413963821629,0.08827730734226691,
 0.29064938648541916,0.1948198508104726,0.7318944316749324,
 0.29725282455036783,0.4753206535325101,0.5655385714881461,
 0.09638245676692281,0.010556209644551062,0.7647632379620375,
 0.7601921231953057,0.03206323249274423,0.4230295721415731,0.6274416691428883,
 0.9102404955168969,0.8468676363636627,0.43577776244827493,
 0.15204104489961812,0.1911522430870145,0.18297167574709416,0.982505099241981,
 0.6283886932297641,0.3680477757820195,0.703167842472947,0.711011405685061,
 0.8569492858358363,0.8450410063014262,0.4202562008368439,
 0.032365928923454024,0.2522190518126699,0.6847965092934998,0.759991923677787,
 0.12189755467468111,0.5529571430220556,0.891125029378737,0.8861142365222472,
 0.726249698113859,0.4455842156128681,0.8549651011379744,0.6449811801331273,
 0.31779315180265544,0.9542980406908567,0.8068205708565024,0.7545519455181438,
 0.5877419800723783,0.17201163669395103,0.41177825560597314,
 0.6547012781246182,0.5042335636812291,0.362407660603955,0.8628157665973291,
 0.9220680874431959,0.7521413921158082,0.886549605501847,0.49942198943750316,
 0.0360821741866193,0.9630646407778287,0.41152267961811195,0.6271652739003154,
 0.7339849081231736,0.4840088865920146,0.5150541757776032,0.8083667627597455,
 0.013886168470505389,0.21888761824554226,0.0787534753936785,
 0.6456003122913632,0.47636244805522787,0.8911206950930077,0.2041937107120808,
 0.09483343793312593,0.6707512282918983,0.46913951492280226,
 0.12686230076935512,0.7977205971940138,0.3984873443354504,0.543647732642878,
 0.27109169786979737,0.22406155580971676,0.23304551504943938,
 0.32166795836737117,0.5814490259468149,0.6802757328606742,0.8999920676768016,
 0.4662120253443004,0.9253632514556415,0.22543619840487783,0.9972516455233827,
 0.844368414133756,0.2802253601821111,0.3820387706392211,0.2130095212276617,
 0.967841710406892,0.6934073281055444,0.4900961210356949,0.8678923542905913,
 0.40439750217888737,0.631429486092195,0.3608073260063289,0.40453917558534114,
 0.6618000936570335,0.5805095247453107,0.7315997245431435,0.46900336440623736,
 0.3545692174787093,0.49257855812091766,0.8975276759507315,0.5107690736223086,
 0.41037095433489734,0.6711146176455319,0.9510563031094146,0.4166350511737661,
 0.5504956601144804,0.9169566372183332,0.42342901061120775,0.7875244676668627,
 0.30008149059030165,0.13834591201159907,0.5933348338128928,
 0.3773690561913239,0.8233689714422644,0.778615253166658,0.474307236448015,
 0.6805033242620092,0.22027470685972572,0.40616274014274023,0.772177853568556,
 0.6265202868618485,0.919454599647834,0.6172750111801907,0.9034921295129245,
 0.43208857870811135,0.9633368866031999,0.2180321862531437,0.7010698319721093,
 0.018801641688034665,0.626774385619464,0.9764572232542141,0.5796964421964937,
 0.36388957022586765,0.2560303542393967,0.919297667999166,0.09217797940438399,
 0.25636948713628416,0.5421516087359954,0.45834323882208616,
 0.08683012386677325,0.47826770539507024,0.2349887128551459,
 0.6404295399208505,0.3739175115815485,0.07558507075014487,0.844008289152614,
 0.7674669551938536,0.76301868770486,0.09503564024161715,0.8234306462622288,
 0.7330306896127922,0.4120057340450155,0.7728424017444918,0.6081158654950332,
 0.4643420942468153,0.6286996671596271,0.8595559987735177,
 0.021663489388544566,0.4146216838548139,0.13542633778027435,
 0.15922102180995812,0.4628694825677089,0.6179857291915722,0.9017350120878324,
 0.8279614490025987,0.7372791318656766,0.5282941052453304,0.95453519304064,
 0.295774388278482,0.8690210346482654,0.9811124624471836,0.7650401001694043,
 0.8822263000817692,0.8409187626577934,0.9206107542290798,0.09515896007733748,
 0.8758076217792498,0.17515701213277102,0.1686188668420603,0.4385240086962203,
 0.6711910297344498,0.664219627656854,0.8821589141595245,0.8336315822134297,
 0.2910357085795806,0.8609800244476759,0.7952787828042289,0.03478936259990817,
 0.21547587243487198,0.5177351452216943,0.0699903114641226,0.831793059770573,
 0.9757094096783083,0.8059878768475317,0.4533934635745609,0.20097264422799732,
 0.6943488510317556,0.021778281255027787,0.2930210434837274,
 0.5103497913958216,0.041467621127838195,0.8274114629325013,
 0.19264315328531056,0.6712302929977086,0.3964920512539717,0.6695154613477895,
 0.02887835297180219,0.9762893689233837,0.8573201573805009,0.3718274038727871,
 0.8339649221240673,0.7210836640478749,0.3670831342261198,0.6629340442442035,
 0.994366477678778,0.5002362052294566,0.4424033831048958,0.8337542720626931,
 0.21355500608188405,0.31555012132976157,0.030843182678851,
 0.48100613716405227,0.751671412593296,0.43249425485591186,
 0.14322368435720845,0.07723306388149265,0.9807381080979347,
 0.2885507900213702,0.35771446546913177,0.47948323623481315,0.617213486271019,
 0.22157777637675158,0.5232323063053732,0.00914047422356723,
 0.6597116351645242,0.5986173494034612,0.9904817971039986,0.24838005441303102,
 0.16395598930164645,0.4037706446207795,0.6246843965128795,0.9600773400665098,
 0.24185895898343435,0.029336190771640624,0.9417707742323815,
 0.8618838396333158,0.6848302926501054,0.3352257719750674,0.5689965485689562,
 0.3446745856695671,0.5775531976386702,0.7587803882519936,0.1511125160462895,
 0.6765260319573897,0.7156664629127054,0.011149127804266912,
 0.16642711566364643,0.055832382666737956,0.33562873038923446,
 0.4666815757999673,0.1216931804930137,0.3635300702627787,0.927472691426925,
 0.3060403772438527,0.6287187154145044,0.2314758187822472,0.9119601801921469,
 0.4837763880520427,0.032397836880489916,0.8370882173424432,
 0.8404649772399224,0.5493613969122293,0.5939131582274372,0.09479164504372362,
 0.4102799678754815,0.7730530686686327,0.022500107768683497,0.606929108333682,
 0.8999079869877379,0.22624337073639222,0.670745582301949,0.42840876776666537,
 0.6878623886531233,0.14100858870079636,0.03258347636342451,
 0.8647250945059218,0.9176444067898385,0.9472395694860057,0.26934581481483577,
 0.8288549366533187,0.2343543469568038,0.7468325608614498,0.7053631996966001,
 0.5665657029398484,0.7818852159526157,0.9868521909801498,0.15373307700086014,
 0.0349794352418632,0.560112370881356,0.6576074638945202,0.1292156480578659,
 0.687788928761391,0.601148027606574,0.7292207049929785,0.3880885434828256,
 0.7024181891724595,0.5498576796420181,0.4750408747620698,0.2941858120917833,
 0.32362998489173134,0.12704347527147464,0.9387110362146939,
 0.4985993378047231,0.44930592993776286,0.1649238941589518,
 0.46749532420357076,0.45697878662953395,0.18966895928165084,
 0.20839492678826788,0.2676940029405359,0.6877029894916208,0.6156011438218447,
 0.48164193084583684,0.7642857728409714,0.8765277141044776,0.6041121108523858,
 0.18575928041939194,0.7036891837434804,0.04355938233041168,
 0.7692547776923998,0.962505925729229,0.31474910448711657,0.15475984750135974,
 0.7922651670352805,0.37717117523384314,0.03552975741002573,
 0.2288836817611506,0.6985763960441961,0.9051634863099588,0.3556723863202521,
 0.0747180496537031,0.01191283842260793,0.9001068976156081,0.3727702925539461,
 0.25647255404880576,0.15739586736280187,0.06811909603237243,
 0.7498341958809885,0.06445813632233155,0.36847961413563113,
 0.5088984823379215,0.4349089525930885,0.44062750207305024,0.909025218649057,
 0.8300829718528079,0.13659620775502646,0.19504521074423853,
 0.6315083362329035,0.46721700813870115,0.5482602523591701,
 0.12242613544516023,0.5123673491522718,0.5889809813492958,0.529551229522172,
 0.8699262766019532,0.838009646911797,0.37524245899336395,0.3425331429425085,
 0.5889452102418888,0.16017783385362394,0.6319877298113759,
 0.019253014754870512,0.6240505539553933,0.30868860458800085,
 0.9673473591080719,0.7426582416331692,0.15084834425725246,0.1377997689466368,
 0.932843489474288,0.6945398292286377,0.20385559657894592,0.9280089558383889,
 0.4944219217763235,0.10956752356972721,0.25605658140413345,
 0.4296050129095026,0.5975724551058832,0.2994458472186301,0.32908970532688353,
 0.7947360514789295,0.9050137256765416,0.082566592982807,0.2441711507490708,
 0.02464845200062704,0.8188756995465716,0.35240047814598785,
 0.07170282928834282,0.8397853335224253,0.7290852893627702,
 0.18748837260378493,0.44239358554159636,0.5638936097902322,
 0.3726090645151211,0.8288373005875038,0.5605927265739965,0.1411736767622549,
 0.07432692479266934,0.8325012201902772,0.4357548352481555,
 0.41657427125750335,0.16179784566476152,0.17791462586268214,
 0.8028747608732659,0.2196519538422559,0.8390438202482184,0.46982547914781836,
 0.5582108604629191,0.09590725077839957,0.7363839058457183,0.8367927842638085,
 0.7708551867664659,0.11788382385511842,0.1529024914162519,0.4414572696196517,
 0.16594825834121352,0.2799986790977955,0.9979740116647491,
 0.014794174753960276,0.2827571716403955,0.7698867199422494,
 0.9944896521891853,0.8184953077445538,0.5347709074942575,0.06671777343299512,
 0.4134552945408715,0.09747772289513112,0.9234410131480978,0.2849874086306976,
 0.10243967267924203,0.08572712689673168,0.6917596037649792,
 0.35368593392613124,0.9050188682806698,0.37167826324609354,
 0.8639617866190792,0.2674532040385843,0.19457445651272642,0.9225415691337187,
 0.41418563945757725,0.8207126276242622,0.1967287037958393,0.8925670160892686,
 0.05048251698818329,0.4142169356096579,0.6956165270027888,
 0.38418721437233194,0.06764250283710216,0.2965041324895803,
 0.03243084605187019,0.9401760694459868,0.7722424110738041,0.4023984668996047,
 0.4552296751207937,0.14228531703462122,0.6011002626242514,
 0.07291836894372816,0.5614665348834764,0.8470620145990817,
 0.08525201699788909,0.6025364574637477,0.8107774733185471,
 0.24539785831240912,0.32396897383250967,0.059754889530922006,
 0.07948738826127522,0.6129433950631346,0.12191383331857764,
 0.2231392366557906,0.6553361113640018,0.9962074585351068,0.6918206873154837,
 0.29606613339819376,0.40401767239336217,0.01987108670681037,
 0.22416638195791172,0.2429750148276325,0.6582138850643365,0.9167658962776252,
 0.07632698523185699,0.7139158462223556,0.6799200603427832,0.4044281298200807,
 0.2794019292033749,0.42575297584778704,0.5902651745135025,
 0.31198899760601373,0.7042046833683966,0.8241958602682511,
 0.007713551596010282,0.30201922890702115,0.9912733224394751,
 0.9003270112036182,0.14363289993421624,0.021765115844089133,
 0.5198406609262043,0.7078987594870758,0.29324249715593886,
 0.06858565781382886,0.5669195276114718,0.7857534347001208,0.3003470427816607,
 0.1487751646565313,0.23642371785711236,0.8905972737151502,
 0.07529890671023509,0.5801296667810409,0.0909081355034943,0.9254294057565906,
 0.24894788259081813,0.6158739551572554,0.3741162240187411,0.7606513442060077,
 0.10745037766876908,0.9694873594089456,0.06732326957183588,
 0.6910284730614012,0.8946598277927954,0.8106165510089904,0.6097622907337322,
 0.15243487039952874,0.28941333017573867,0.7428830058372087,
 0.6686606651322992,0.5937774588471662,0.6414330479511889,0.3208773581471339,
 0.21994454470726843,0.7641469328071379,0.8883257477202625,
 0.30807777378629264,0.9966879303361694,0.06013451023613936,
 0.07349309240071134,0.2869151952530663,0.3354970676249832,0.2687625890077593,
 0.7706499302622458,0.8977623657418632,0.8461178920221053,0.1306553982720895,
 0.09148033048792581,0.9757626477197316,0.3206032887832341,
 0.47029616972143407,0.17121812594333663,0.5528104917597634,
 0.23663014696644313,0.8873314776207125,0.04617766875250595,
 0.5743420347910421,0.9318071373207208,0.08535496646067942,0.8595521660739471,
 0.27291405240119515,0.47716170996142093,0.7296381093864963,
 0.7359598677891859,0.2571068641301987,0.9346066921505112,0.49827152468651037,
 0.5021575562973102,0.44346828357981716,4.8095457404018216e-4,
 0.28226752512177133,0.5868658737982357,0.6870532076972817,0.3803342795499829,
 0.9906087088122828,0.045640312974830954,0.6587720690391163,0.857379830634194,
 0.39260123511065026,0.7532640336490541,0.17623604888306255,
 0.048972518628393136,0.8500923522576737,0.032356314921162754,
 0.804569452721038,0.15021461105584832,0.972768633715596,0.7948906061108539,
 0.7560521860466283,0.7753946434566239,0.6133357432529845,0.9931534008202271,
 0.977964315440717,0.9850805312772825,0.3674448469335504,0.4322889427456468,
 0.6964867743871945,0.16633962279530434,0.5112409687939012,0.5254408374398822,
 0.4683577694986151,0.5656396936538393,0.8180061809971715,0.39487404776243884,
 0.38675807484200586,0.7688507333631857,0.9096657685223021,
 0.19063866604455537,0.10518620224185471,0.6371800555143421,
 0.5594363783881714,0.22731863661084017,0.9069871595936758,
 0.47459537406175634,0.6554282088555412,0.005676994774141075,
 0.8854945597089505,0.6848371119888622,0.4694163930953713,0.2585548534379739,
 0.7337755442723439,0.3588470492017759,0.7522944267781643,0.6147908411322489,
 0.45027101995939445,0.8276834565454811,0.9123912861281362,0.8694639969019266,
 0.9392617281460098,0.7841432224351274,0.465162649511638,0.2836050113578221,
 0.42575864383366113,0.7124938404095469,0.5073876440982731,0.3812749237252119,
 0.9584744343756448,0.4379439163766057,0.6020068535360856,0.3111478614123323,
 0.7195797983990303,0.282601305257133,0.8173963942473889,0.252620002642977,
 0.14674808508075565,0.37568778322846064,0.8397509290191766,0.970864720410389,
 0.4670732092573785,0.7637141437613439,0.9849238055739615,0.7514263055338537,
 0.8072367346635543,0.8913796298026021,0.8935487955332875,0.43790755613871235,
 0.39879121925851435,0.8045376446053882,0.8165972487118965,0.8804064547869733,
 0.12356660597303093,0.29876444090442356,0.358173944298537,0.7534668691187428,
 0.8586011308099961,0.04997537682354258,0.952089468839871,0.39503915514752586,
 0.9518950362829861,0.2856172479727026,0.5495322008826312,0.3105807552103885,
 1.8391250058624742e-4,0.7803306154422655,0.49747114611314636,
 0.2574758724490418,0.723013507787123,0.10352093593535105,0.904457072869472,
 0.5090697960168294,0.9394899084334538,0.4742485605090254,0.4714136686892335,
 0.5035198191789161,0.07126788318812471,0.8054270391603989,0.9658623629887747,
 0.521615338569898,0.09176210565386922,0.6728340533977533,0.5273727670969562,
 0.8819197480899887,0.9267983253526824,0.9528960414598432,0.27891169583194886,
 0.6879424845596313,0.2910215060129001,0.6205750138786882,0.9125561091951115,
 0.24547494147616278,0.681898277475589,0.19132151295772581,
 0.27921320555098694,0.7993322026912587,0.41474457372043727,
 0.4938974032664507,0.8372528630000733,0.15057040022954926,0.5195440314100747,
 0.8864953247733882,0.062419160709601074,0.05456603775866564,
 0.9163237726240677,0.7662733240209245,0.0021566892195749032,
 0.4120341868542827,0.6406513627412305,0.3038358141441144,0.28069451651481225,
 0.0645574947216695,0.05781074481987136,0.18841197214819938,
 0.08902072483608514,0.5543423320642324,0.5004367642935841,
 0.19399645609917182,0.9667160593902103,0.9762824785440363,0.3521075672531683,
 0.3099474725514805,0.5423356682174304,0.7481580973925591,0.8345334425719267,
 0.7231443832388199,0.6049324368631436,0.7644902579983616,0.9401326397240033,
 0.2601570145675628,0.7811114678683767,0.07164806872449492,
 0.22146891447749573,0.9438226428932183,0.44562914213321614,
 0.6510791747621586,0.5922229534167006,0.8198007114587649,0.3210756580290416,
 0.834225313841942,0.8874291461267244,0.758856514052046,0.3936784679053592,
 0.8931709321539929,0.1694907750018162,0.8487447103236512,
 0.008289782264894519,0.5619601421457429,0.10000919154552135,
 0.16307781843975122,0.8427636366025589,0.7962227296507064,
 0.20863718150855193,0.6658800445181126,0.4766237537347575,0.3651242450785712,
 0.3709720998896724,0.6936130559768008,0.6172743058077125,0.6763854395904078,
 0.8697646995344468,0.7717683204998478,0.6893558947252152,0.848373854609713,
 0.7317353392208564,0.5050355169645658,0.8944732500753549,0.5909872292316369,
 0.7611291795646241,0.43923899542041633,0.6382978114854239,0.5324329269856686,
 0.5936489110478885,0.23681869462578997,0.02221711612652444,
 0.6935262211271858,0.3419937624114431,0.21527203518878446,0.8053436333737896,
 0.31795543100156287,0.610852123227192,0.4463608310386247,0.9533654004190821,
 0.441956819220126,0.26576189135804973,0.7652297955546243,0.36781905822633243,
 0.13463928652818113,0.7697511765519158,0.5254490929265426,0.4901293133834568,
 0.3849773512831556,0.4418235767591283,0.21580786377610162,0.8136187677169247,
 0.1525008509668183,0.07196758823980365,0.43978426762479517,
 0.17521807290364777,0.01582206920801299,0.19607925627613043,
 0.5459332149021305,0.6195738303809722,0.7839711959221781,0.12748674751571099,
 0.49263001017196884,0.597720467754913,0.04344314235114921,0.3871814058560218,
 0.363721131698044,0.8110912389544366,0.7043515773051726,0.26711132307212615,
 0.8728654415434818,0.3874498213418608,0.03775169846380977,0.5486601699177539,
 0.6335185036824649,0.21501625664774093,0.01625355430208686,
 0.28020688115121994,0.6563320385699334,0.12303388795843839,
 0.18972240825149156,0.9321776989238271,0.7639647064302535,0.9891236732295905,
 0.6221432679239617,0.5353364179064306,0.6562739528141269,0.6044549043994414,
 0.3944760465708289,0.5748992870959976,0.32303207697768954,
 0.08810902243708763,0.12175983771409638,0.5737876462943107,
 0.09487242849318167,0.3418265430723094,0.5770165919037691,0.3010247904224218,
 0.12888851458548722,0.9754159707064578,0.2576832422563847,0.918626223696144,
 0.02496690335251639,0.5648169020503416,0.22128109828081532,
 0.4436226879111502,0.6748807902181345,0.18111136446001563,0.7198661791311476,
 0.3849660766598858,0.7373142371874412,0.24398928994910396,
 0.45483412276804125,0.9967742219531583,0.8425216259065633,0.3536493958164001,
 0.7102346823136317,0.40220562676152816,0.13584576434415485,
 0.4444127750480905,0.1513030035718942,0.3761728770064383,0.2709588311597897,
 0.6114497306996018,0.21269624573486867,0.5442800838572275,0.8598836129275391,
 0.11993656476314762,0.047094336412322524,0.19362612551856373,
 0.8330302252132761,0.28728814608651887,0.7653527679213996,0.9890409603253152,
 0.7432942631666547,0.7185717687733936,0.9155625531113878,0.7718645256592531,
 0.6607508756345517,0.7567380861416999,0.6639862512000797,0.9238198587618496,
 0.7520885853438082,0.04337709162598691,0.09830314291691766,
 0.6679846544570427,0.4512237714951701,0.6636192293417403,0.6551371469433857,
 0.8772179578302955,0.6734750046829594,0.3380651150066646,0.3907234724069051,
 0.3497342368961599,0.9633224088366126,0.8329526574029729,0.26370719199981163,
 0.45657349670024394,0.1647043373575794,0.0940358358076977,0.7203013805035677,
 0.1642920227115343,0.7554944887974021,0.5184193915118743,0.04952512258415909,
 0.36667787440491295,0.4311172977112946,0.5358480997918371,0.5125547500668828,
 0.5167635144539721,0.5805001338681908,0.3349876403398624,0.5193633731776883,
 0.7800399520173463,0.23470188854952978,0.9774238963278847,0.6380083844411066,
 0.63295401005892,0.031304404550140275,0.341213995922795,0.3595055759897344,
 0.09940306472507365,0.61235846718916,0.285570569899373,0.029619888654287774,
 0.5175176486756178,0.10448918210808888,0.04066451090434797,
 0.16477713222664048,0.7538928368411006,0.5811656181562328,
 0.41900660435349013,0.4177410468302758,0.24699930759367772,
 0.26113741413097413,0.74556616672501,0.7957866738581472,0.8903529208824902,
 0.4529679731296524,0.39636458169221134,0.17086827629804924,
 0.5980301228935317,0.6908252959219929,0.586433325842682,0.7713717373535977,
 0.4469274423910563,0.5798708724886742,0.08218325874265664,
 0.43221624474757603,0.7127766942152841,0.5458308442389466,0.4740059142385076,
 0.17646161011895178,0.8298697607546595,0.9166543695796072,
 0.06578844357568525,0.7831937610676483,0.3892335454977842,0.6949860298544296,
 0.6239265590215173,0.32901756594578657,0.4285162380896612,0.5297772193981874,
 0.3019346253917333,0.24377903037158954,0.5802288430081116,0.8548385868269519,
 0.935742762353883,0.7251292864089371,0.9056545214681525,0.7288397051778631,
 0.5309877657406037,0.8459048110260813,0.43496534293332945,0.2920433641832628,
 0.6057845343589504,0.22795852913018855,0.27236961998979525,
 0.06986043910704509,0.14514742650243218,0.8152708509949388,
 0.030426943004469575,0.7234784824848953,0.6932832933622866,
 0.08080452849897579,0.6273920063384748,0.45514115717440184,
 0.4699536048052886,0.9464544947763259,0.8638005896610749,0.4533670058587065,
 0.7924403620052088,0.4344383120206524,0.2732987479001727,0.9508441344721901,
 0.8779604946407495,0.8409233896200224,0.38993846200519633,0.867868668427521,
 0.41279766353527614,0.7270557903618416,0.6572828333929323,0.9512452409980625,
 0.4783741710091918,0.6010655168759186,0.35561798482875984,
 0.026765420346862467,0.653568028432156,0.7058212224208655,0.66478046160217,
 0.9324198596621605,0.4702439913801054,0.921139274155305,0.23634707985641523,
 0.27240157691017175,0.9525090660559513,0.3378245678542344,
 0.47939594359894044,0.6856695789182412,0.9037230719965785,
 0.03443880994159043,0.5427755640581237,0.6527710708872944,0.1574315179838136,
 0.9506343021268301,0.1302028445899489,0.10534869097636301,0.9308866273758349,
 0.7142699874816724,0.03524053123783544,0.39719789158493013,
 0.8745188459347087,0.7826330456886617,0.8045206764156952,0.11192890820582646,
 0.869691316838818,0.9294446239865924,0.007914645310421853,0.7492261251215261,
 0.30100311834757654,0.23350777102749487,0.23042184906022845,
 0.46244455362497505,0.6472378591140376,0.4267972509850857,
 0.26946678994576456,0.4060499662505048,0.2278510440797349,0.3607278449518424,
 0.5659209325132363,0.8227042938924589,0.7401963520742988,0.7133063488892737,
 0.333658158909661,0.8869279968131377,0.7997698581417583,0.05933422163682289,
 0.7595917786535235,0.21359850638819633,0.11988861463840594,
 0.7288247767227114,0.5819285665717586,0.6743857394725679,0.189895139503936,
 0.19333246714498287,0.06495425499765883,0.41063034656936825,
 0.30100753529757174,0.900919039065591,0.4522768239734548,0.0795221263012652,
 0.8215073363124875,0.26482818561869315,0.4019220871325715,0.453056860333819,
 0.15378630570003793,0.12887083211356853,0.447728844908893,0.3541494500861062,
 0.6216197300318524,0.2992830091501083,0.26515053255283105,
 0.31321609151690044,0.6983629245522,0.254241392039657,0.8305513390108287,
 0.6149797364575743,0.6959214630760686,0.5807897507008428,0.29735334933612567,
 0.6525159900761343,0.34162337082000915,0.1692091092307496,
 0.21947653722127058,0.536867041932353,0.22577463368845418,0.8807009799550262,
 0.4560367626665822,0.27890492677235723,0.8616966137895392,
 0.16539675088599703,0.5125849510418594,0.09724988921110489,
 0.27884708506200107,0.32017124365814864,0.0501085810829367,
 0.46676728173347826,0.8845146506561549,0.9681794286482572,0.8619148530543925,
 0.8045323751131997,0.05844317835929447,0.921974484210981,0.749963490083865,
 0.35181190309500177,0.5173258069114933,0.9574295229905807,0.5873666315349989,
 0.25173499996566073,0.35564783886021867,0.8459053392707966,
 0.12822444645073072,0.8850327579128513,0.7914987497730128,
 0.26855457653996506,0.7686191926946433,0.23128414438256106,0.648420689363968,
 0.7292089763088323,0.19033404192151637,0.7532339448111548,
 0.45755857980363934,0.6829774747285712,0.7852705997787209,0.7957743906648681,
 0.028365227813793248,0.07862432552466914,0.5243257120342206,
 0.027502329046271146,0.1742011066716772,0.3580694876742122,
 0.0037177178831322433,0.2814516135385471,0.4975224776810473,
 0.8604126118661702,0.12069622730421492,0.8624422137485614,
 0.34696213330509384,0.9191182098855303,0.8406537373487775,0.6151068588781151,
 0.3403965688210052,0.27493217396611924,0.7640568466007304,
 0.11182213669937591,0.10483789239184338,0.33627774844547953,
 0.40764474311230703,0.15120270807049496,0.055784357513902205,
 0.8651171115493499,0.7312889182008182,0.005485217867168379,
 0.7648966137832325,0.5157888094792944,0.37767268250057306,0.7239289232077557,
 0.4934044658094422,0.43593291086373964,0.36820293432692086,
 0.4705966357319903,0.17186099888488804,0.7625761640063047,
 0.29659095416099357,0.055682950146529464,0.9168473971636537,
 0.30040600958196506,0.2397316121833032,0.4108043139893175,0.549467634717716,
 0.37854400737776894,0.5169640053916961,0.12468074574850552,
 0.8893993949277385,0.7996176028089952,0.1519839588782318,0.868814507451737,
 0.04911530849123724,0.29349604457402556,0.5806503493305974,
 0.9720355957497124,0.10844046573352017,0.5919972705952993,0.3001305787065617,
 0.765751976087089,0.7315459386209485,0.0580531476482985,0.05776058253675742,
 0.19880110865046652,0.5821528292783587,0.35627426381928085,
 0.7028274637980653,0.8111885000625443,0.3263867836716581,0.4711375898425596,
 0.7415358388378495,0.6061782721558872,0.016974680010495624,
 0.21108308221614258,0.013847189600464382,0.7573698170225451,
 0.7388184589777621,0.2831419139450233,0.16223497277041588,0.7937475287391571,
 0.6472718685216897,0.19548789206829853,0.5904218041333992,0.3184831837313348,
 0.6479841916623853,0.3968162222900746,0.6552066347306641,0.6825980589110979,
 0.4839075847252857,0.8879615511300479,0.5912536178051834,0.2121005393790112,
 0.6916331175944742,0.4492284066779182,0.4900174621214728,0.05297950044992639,
 0.004271566025646378,0.1850389895737301,0.7289356214271129,
 0.26065603622939615,0.09009160065128063,0.6477274961930463,
 0.3338537986185345,0.1251211497628022,0.4388501084936144,0.2578932498622477,
 0.605661002746011,0.8259346187751785,0.5825611693265558,0.2102611715200542,
 0.32116885713251864,0.7355430181426523,0.1479614156334459,0.6763115624574348,
 0.8252040572513814,0.5939690529294039,0.8862856211084772,0.04116223586654244,
 0.6094714249588745,0.9188960209600284,0.4695292842154314,0.472786180212181,
 0.7805117757149983,0.6162567905982008,0.2602807283394645,0.6317403670490617,
 0.6241514218736156,0.7961722758375064,0.8214096891425785,0.20941776363935427,
 0.7365653135795359,0.6816837081820959,0.6926543396257113,0.7752555561040261,
 0.21128183550110746,0.8549736198005,0.024574367431265842,0.38886629904227465,
 0.3138413097383522,0.6326520224254109,0.4780207931865116,0.12490664731145962,
 0.6347209981401521,0.053892948927154016,0.08111519135614598,
 0.9754710115503721,0.3664515356946083,0.6769465111248971,0.5090129164330508,
 0.4809364772372935,0.36904232930105807,0.9667886037335994,
 0.021160895426078796,0.12920598759127466,0.38816756765283944,
 0.4029461742544721,0.8879795206444555,0.09113465438119794,
 0.37030421235085587,0.4697729066556392,0.7653274508734684,0.7014015027535028,
 0.8898312895421425,0.46134432883035537,0.5476906473285759,0.8068263941040253,
 0.8403225963394242,0.6801076788314548,0.31042535827564155,0.5545799182688286,
 0.6424509592964553,0.5003352539592484,0.8599155007726955,0.31974703305851515,
 0.2759010624886493,0.8506148269722421,0.26691253915380897,0.6166066882306339,
 0.4172596459330046,0.4114637296073438,0.4815369054653341,0.37193500056325246,
 0.025924286289774723,0.342204995275628,0.7594173211028993,0.6747549144761513,
 0.4221252980769953,0.7987660472467656,0.2410268696202741,0.29145489836232263,
 0.08722637437644076,0.8729092819999114,0.4540784794791862,
 0.13931290112434003,0.541483334389147,0.7534183213946194,0.2637816364007699,
 0.09279275360484629,0.3184617222218802,0.46005200547529135,
 0.4155652045228053,0.4718905974351594,0.5804880080094754,0.6720590444939638,
 0.6494836731710181,0.8860986183762399,0.7867351461604972,0.2951298658364675,
 0.1935585278059765,0.4690781138986546,0.4729300750832446,0.45515357741192064,
 0.6733251247145878,0.10979873184787525,0.6794802108147393,0.5256346010904549,
 0.13408982062432195,0.8266716429499131,0.5730173020478914,
 0.08998513132333197,0.260991463977466,0.9655160741178754,0.9027449333851447,
 0.276002622976778,0.7553097098808098,0.5724023530004283,0.11477536315631909,
 0.8239137194941103,0.4034355416931623,0.07025398297056784,0.7636585680093644,
 0.11587914822427459,0.5323364648445792,0.5001943037893017,
 0.09739676099397276,0.8935770134185244,0.8265687006225567,0.681115698307888,
 0.983939831542344,0.12119989621725313,0.9573414863573754,0.7668507848590839,
 0.6663837952128425,0.8470160172409236,0.8298546746310385,0.5788070957163975,
 0.8414471914782473,0.3274382643843059,0.12768993758057468,0.9073116754781306,
 0.014580338750792676,0.10611471844906772,0.1773146167171572,
 0.2994898175210162,0.309622372467091,0.6825067500851361,0.26155305609910506,
 0.2899195753200863,0.9306811481552865,0.7101275949704933,0.5110628537631317,
 0.5726426281868573,0.19286625255163936,0.9474429915580984,0.0661013474282044,
 0.369449422106747,0.5813185113153959,0.21652892445368943,0.4905449237710817,
 0.9092768781585966,0.32027596633206634,0.7517358100583001,0.566539221876635,
 0.7181170701935702,0.34817562336242835,0.8625765241595288,
 0.11468131067810172,0.17810602955688537,0.3488467646885032,
 0.9790009052688098,0.6129977423241569,0.29502274189579747,
 0.33550715572280787,0.014481913078449704,0.6494238581956626,
 0.21395483002396432,0.4042041480494831,0.5828787396572842,0.6055854226373865,
 0.24165370537914344,0.8905264158511031,0.7187376790315154,
 0.004640445650326175,0.99830474414362,0.844597837876473,0.8752835318031204,
 0.16251461182348192,0.8258011004273601,0.10136578097330506,
 0.9776556250758587,0.08679291938271061,0.08727384365060109,
 0.06464190104940726,0.11522654651524111,0.21440747270180693,0.3124672273369,
 0.18616223573231772,0.44553816088876064,0.4021049103620352,
 0.8508171110827834,0.2778408888955213,0.5183097000147653,0.733048652426073,
 0.5760025068100143,0.42718480832050054,0.06746959790466589,
 0.5723725116051669,0.4997313872473004,0.9624976012352051,0.6781079796285074,
 0.4675461922870797,0.7835905225533897,0.289351657080931,0.7103365986159429,
 0.16795213633983008,0.14346159795613633,0.6135973995298524,
 0.9464241276648904,0.8698872022609078,0.5256388243817716,0.4136753069873804,
 0.2554248900775463,0.879654559451949,0.22670587223564465,0.321003898885278,
 0.44373945865853404,0.3577861919027322,0.6352560251944597,
 0.24097113421187044,0.8408706031642326,0.5740940069807156,0.807552755989911,
 0.617343294738986,0.8004351493606328,0.0859420482244716,0.2494430817093084,
 0.6788155730402686,0.02429549408135412,0.49444794615778187,
 0.017531151996271976,0.21859086052511534,0.803524092461499,
 0.6415068272791735,0.9239342480678775,0.15531464170239462,0.7612143232151157,
 0.84431170618444,0.9784410117950172,0.8892362780838504,0.5815509203703887,
 0.8859785522094888,0.47821278289494795,0.9826682143195163,0.9290114588581592,
 0.14619484145875095,0.053847349405408806,0.9542233292376355,
 0.018186459598275362,0.43593565006902835,0.20703740161191486,
 0.22643877882227348,0.15941332081611315,0.2339668752790579,
 0.3720011863056024,0.34618614144170157,0.9337759093775233,
 0.05694254888279848,0.2009392812643046,0.8219174221902508,0.7792568163115955,
 0.6242910635168326,0.1593992617398189,0.2335903933583412,0.731077669625265,
 0.0025422119990394787,0.32916296129413514,0.26028480649464303,
 0.7927645658860096,0.13069826341382274,0.2507128646613658,
 0.17136988675619969,0.757296898437135,0.24010732742954022,
 0.13632211480441248,0.5365464533705042,0.639346432636438,0.327579247398863,
 0.7223013679454867,0.6350018879286957,0.31732069373125205,
 0.03810879913913534,0.2827580108419464,0.5785889033340937,0.1353046947552773,
 0.016119629563029703,0.9783826271134203,0.40026059063063535,
 0.28115064295746695,0.8839874853458775,0.433108857676348,0.7619392054857496,
 0.6379330276149557,0.3234832969771173,0.5827771203215655,0.9517350730762518,
 0.2262634070691184,0.30122669586934514,0.15703577544401748,
 0.45748088869465064,0.3197590843053634,0.3034806320780643,
 0.060208602498477815,0.3912003227472646,0.49277306412201005,
 0.20793901374708634,0.26142962600285014,0.8457641005122667,
 0.4807886497008622,0.29707532277843196,0.2170702862977708,0.8413707813139361,
 0.43817829449144574,0.44200051678862695,0.7754471784310751,
 0.5525387903999199,0.6399328574619955,0.9857379589207114,0.7528997737786188,
 0.0970129381316448,0.4614966067603088,0.6836069308246715,0.32732788022595805,
 0.4444444485809519,0.4048340670499414,0.04898933661430227,0.8792492716487943,
 0.901398606953745,0.9699042465782366,0.599786046222223,0.6915860314843334,
 0.8657054695029511,0.8933941720687189,0.16808170936206646,0.8124477938644841,
 0.4665600500110715,0.8678462220146099,0.6401320118925644,0.6870637935837355,
 0.0325969182606175,0.599928580299044,0.5035091313261005,0.3675010170282901,
 0.1633176807568546,0.46378781001711433,0.6183952921102722,0.8636277862390666,
 0.7084084147010136,0.04794505196400667,0.3298083525534964,0.3193885302692776,
 0.19809652792807586,0.04399654742139758,0.7839534279639104,
 0.20231666639582457,0.35767400226586965,0.9268939768411186,
 0.3867872955507796,0.14316541914046255,0.21868953076865028,0.806101918975961,
 0.7997655401866225,0.836352785086485,0.24890764791949616,
 0.041233936999553134,0.47739279684937297,0.32933013572060743,
 0.5236034495846527,0.767931146309804,0.8596200415577294,0.5451088117491949,
 0.7267723308973303,0.8722008851192892,0.9263473556183354,0.07882844701883795,
 0.1726012738494408,0.8116929500872981,0.810579854448669,0.5986054520803283,
 0.9436077479839229,0.07506449659068859,0.7920849069996388,0.5606311888214287,
 0.504757014112997,0.29798220871363634,0.02340820634203189,0.1422114012601131,
 0.45664127268995325,0.18858230619968974,0.07030710102819537,
 0.9978301454529679,0.021544826452550625,0.35553435162363956,
 0.06521520654360558,0.36665547113304664,0.43975214777092164,
 0.1074290111412719,0.8099740531675914,0.5121396039687433,0.31974904554012473,
 0.9579813523067517,0.1012000782251492,0.806451196706254,0.27200663277558246,
 0.8549843259148282,0.2185172614256602,0.9346947099499802,0.3341262116124101,
 0.29293997877609834,0.2615349296422652,0.049226450660187604,
 0.8626810128561782,0.8530602657245859,0.8708083295316258,0.4429581420727404,
 0.45142907281979006,0.7954918712898187,0.026745574658585847,
 0.0756932924775282,0.7161387490483693,0.05220301742221023,
 0.16057714199161466,0.04384841915926452,0.15936908054115517,
 0.4963282376088817,0.32886026544010694,0.5539242093697483,
 0.07605859386879199,0.4652368074059524,0.367442366077696,0.66938492021545,
 0.29907104819759656,0.592787478996808,0.7591152160559274,0.1129357878045264,
 0.44551233938215273,0.44697633275087956,0.08961281222808526,
 0.6663543180577578,0.5799911316044848,0.7724176745551381,0.1865155718771352,
 0.2863935668369725,0.11885938332468104,0.02696851848342119,0.865639533105083,
 0.47304684113267703,0.12944069080621567,0.2600659733887648,0.154374463533426,
 0.16659026554929102,0.9594268447850127,0.90722125744664,0.4057907664402627,
 0.195358665835351,0.9315673405978588,0.6864143503869028,0.2185717043058637,
 0.5894735510910426,0.47133396898697866,0.5848253219657853,0.5054387837349218,
 0.07899648042220464,0.4741466434956777,0.9802250437248987,0.5225475255898476,
 0.17291189187906586,0.6030840139751495,0.13560900874298942,
 0.0030628128272219834,0.6826414077787003,0.02862888881950243,
 0.6690940946726842,0.16334554137603186,0.2761946845268849,
 0.057231389662075194,0.9060345567037948,0.13514444189124808,
 0.33477144837163797,0.5083363183070596,0.8176587228965654,0.9704578844918057,
 0.7452117192990434,0.8307065762712128,0.4167944784834007,0.28420793954833856,
 0.6519010644578045,0.5464556139575535,0.21858579215005458,0.9739397263799338,
 0.801046213296008,0.8049457746791231,0.9700106768984336,0.7804338750118432,
 0.5545045480171933,0.011221181968326666,0.3017315024682994,
 0.5429510186983728,0.8606302002503706,0.932050387645973,0.9828269822725706,
 0.781455175557896,0.35124202671946236,0.1962958595419868,0.9595087693766439,
 0.759081750059631,0.38182795952128745,0.2545796526886718,0.6129780904476748,
 0.8820588651425157,0.16319131499207296,0.8076900525003202,0.4581443909949634,
 0.24101031519688254,0.10051477977025203,0.6911884163314689,
 0.5423733101706865,0.8627776181559939,0.8451538176625293,0.04444728137254139,
 0.3478302064453218,0.506557980890448,0.3981469656335245,0.8764685494827228,
 0.5411990073423367,0.9702178337203855,0.9372753621275552,0.08315123615901077,
 0.5225129375829769,0.5067701961138503,0.9421054300398821,0.20633080895319944,
 0.1670125129501301,0.07994339472491963,0.6750339051832752,
 0.42725536764592764,0.6141772050201258,0.7710173653282206,0.7247941690810623,
 0.3203940257930108,0.06454866306585594,0.8141740280650662,0.8166463917777934,
 0.1884903174880349,0.6652711833467571,0.8829481323120264,0.47087551462740596,
 0.14350668565277913,0.08492695258157223,0.5733971601370299,
 0.1765028684021427,0.035823902437605915,0.24108999475496296,
 0.2919951407049064,0.9706337440898245,0.23211705933022353,0.8236028049791647,
 0.15578883706407243,0.9564815636091114,0.5231473065598677,0.6898762982078291,
 0.03820944692619743,0.7924865828857176,0.5640235034054019,0.5680036137640726,
 0.15756044134365654,0.6489608032579949,0.33216139380105214,
 0.14888492682229482,0.1895301893551462,0.5856692759217963,
 0.07978473156471888,0.25268139217885643,0.31099300512269945,
 0.4628788519137824,0.36292839372443886,0.8103329554318317,0.4338691390040762,
 0.23481824968798992,0.43725612225579025,0.038896272651096186,
 0.2179866002467451,0.12141501265145871,0.25406668781716013,
 0.8867605166437613,0.6210610820041966,0.5208001606379827,0.8012138348567786,
 0.5804472217752029,0.29242477458341165,0.04489835556416366,
 0.6213657843912028,0.37448032533689435,0.09866499194966394,
 0.04880253034274806,0.08176001421092127,0.5232890092437417,
 0.6419168752607252,0.3925570382155438,0.4157684423438075,0.9601325745604634,
 0.30480690636882835,0.9743090544378527,0.9658747806758653,
 0.17053721292246327,0.7350429909626972,0.4032752568923813,0.9205885455155846,
 0.6702995804004369,0.59174240707152,0.3471754538452626,0.49609996575972815,
 0.9713974134961006,0.2654696439603479,0.4682312130638785,0.28809870993167075,
 0.018931718868258396,0.37195923396925623,0.13035328664305768,
 0.321917976895551,0.41336080983078094,0.056560264875700206,0.551347108693747,
 0.13766044147732293,0.16010428509997998,0.30982626940137337,
 0.6601141950137018,0.8275713492149537,0.3464986141692956,0.4965646397136346,
 0.09367510779030497,0.7148093553951727,0.36274658228255163,
 0.8300751013349262,0.3265765396489272,0.6752197170940109,0.6543464710737383,
 0.6844666108800824,0.4914724498177011,0.9296392511180916,0.6471481890466415,
 0.6827201470099015,0.9197427950289194,0.9057040892970852,0.8337601759335989,
 0.14286330648745016,0.35673131253511836,0.10367503329585126,
 0.47180938358782254,0.4276139637308083,0.7496181108924197,0.3420930937041273,
 0.9308198846526745,0.5825547602032246,0.4161356735183279,0.3918313137964652,
 0.966719664204973,0.22599982163652288,0.15783965986956816,
 0.022136254732436966,0.22084560340867276,0.8374986458411563,
 0.7109365063073599,0.3394275338745779,0.7691857414381544,0.753339845822616,
 0.8320122747342087,0.7592564259680489,0.7840653554479288,0.2762855481565155,
 0.669806684913378,0.9899803465272341,0.19204710258667668,0.6957272263800647,
 0.5575040599943396,0.7664010072375724,0.2972302574394827,0.8400138918000394,
 0.7567221278360062,0.14960525963951865,0.5486211267898675,
 0.012093933859921413,0.4440968891334942,0.6307567328583359,
 0.19907108137076157,0.9385218585882151,0.07952890139135038,
 0.09303514102325694,0.44028119248665265,0.7248969645766494,0.641750124879607,
 0.4186342862497968,0.7822014559070158,0.5267668817790846,0.3451566485508426,
 0.559321435929703,0.8819008833808998,0.4103010598053394,0.6489124413572169,
 0.19735408848731884,0.6402369811990043,0.6613513914997248,0.5764670472662752,
 0.01635460007264955,0.7047110066862425,0.1624413853131238,0.7757517699916173,
 0.11598302154182905,0.7718165923506531,0.6717393045743445,0.9078943548210964,
 0.006776407760983427,0.7041877504569951,0.13007065867818957,
 0.4307770326790874,0.4284285646200071,0.8704462104089012,0.35922004669007246,
 0.11651002037396674,0.16209532499734847,0.24047470668770177,
 0.13623867690038172,0.09735915435128517,0.7011415066331914,
 0.10326699232697412,0.701474857887726,0.6667838034006941,0.3973675523143799,
 0.7089239727001334,0.7135634685144279,0.2913224277241454,0.5738780310197833,
 0.34997180733697175,0.9579120936709931,0.7499028840661592,0.5468908006528415,
 0.6189340449505825,0.32757750948896047,0.932645951131734,0.6106895906221024,
 0.434918308351365,0.9798142652119951,0.08946953610245367,0.46520314194259327,
 0.9997475581258306,0.14228098874848938,0.2642857097593939,0.6878196290327732,
 0.21366491528118914,0.4868335369439545,0.15831344067466635,
 0.06648386241214432,0.8972046652536437,0.8356878792391788,0.7049933115076046,
 0.20508204730048762,0.4332975187259861,0.5465209911605472,0.5402316950959787,
 0.21766597806053412,0.07915454633314589,0.5814563286721962,
 0.9237701770363834,0.5941601522696545,0.5376481063267694,0.8982339239029692,
 0.0722132530128774,0.5760778756747987,0.08680959008508737,0.5330695131963287,
 0.9801430221719805,0.665277985277787,0.4573005064898824,0.5256231240164182,
 0.33888774744877304,0.23870733416590673,0.23477232474217402,
 0.6884005500848799,0.6729610345089717,0.13673014709978037,0.333293284600107,
 0.8085567958824567,0.2642383545490419,0.06520000260194148,0.6073006236020269,
 0.4566614645390441,0.10489288273794928,0.021706645132569036,
 0.7595064250939144,0.22544712661142263,0.3475882657364746,
 0.056714257468438056,0.9817395983167823,0.6670082731182191,
 0.8890558116457841,0.5475478690590908,0.452541877659705,0.03700360202001407,
 0.2772381929752168,0.07854370691787482,0.14719216080818787,
 0.42560700413257413,0.8052558173350725,0.2907060127444272,0.8103699089009746,
 0.1544096654920566,0.8882617419028674,0.9571980456787889,0.4341430497869353,
 0.48063441309832144,0.29758092229017563,0.6926784271613253,
 0.9650689670062742,0.05444947017160551,0.010083205193062272,
 0.1859533664288443,0.8833132652503148,0.48488559577215595,
 0.17536335983285511,0.6705203012965438,0.39782537620849867,
 0.9047638424355552,0.18268978901195476,0.8927204636544714,0.9182978719222603,
 0.6196610294258355,0.13456213466771494,0.19554490986494533,
 0.39194714766152516,0.6974742988814302,0.6043598320378671,
 0.36083930688316257,0.5260407485543614,0.4865143504856333,0.6976431089597935,
 0.38014313001682776,0.861687567721733,0.689057875036936,0.2259220686173723,
 0.5720761740833229,0.005445465186950438,0.16550185618676172,
 0.5287356955142276,0.014831071506635651,0.5110727463590349,
 0.5193496792082539,0.9921671230217151,0.8529977315745384,0.5281103281414554,
 0.5957577975908387,0.7051421812309668,0.5985648931110468,0.8692895101871163,
 0.23478351563970234,0.9838827116541111,0.8189182073656518,
 0.34201704935206223,0.7595106840089537,0.5781180947628176,
 0.19942623741407361,0.6510283610539356,0.015955598128876813,
 0.873777606084525,0.12778599265599855,0.3456618383858584,0.5207153166084968,
 0.3768999672267678,0.19209711461781664,0.9990755290208518,
 0.34402296539412125,0.14095977991515918,0.415048621213168,
 0.47491359380678455,0.39424485584595403,0.33603097991181574,
 0.5620096024872357,0.458281795490479,0.8789198898172189,0.8522596591119839,
 0.7518483144067765,0.9987563963718826,0.37433816851685264,0.6590617235579073,
 0.9092185667098827,0.5391317185210125,0.051189703992227664,
 0.12092883624969641,0.8279144669134558,0.7437315228784392,0.9007031910458014,
 0.7387149262152546,0.5517832599316989,0.8058833939785689,0.5092539922529202,
 0.8814221058390838,0.7145834789998041,0.3697070202221504,0.5548490801963586,
 0.6917195516255097,0.4031345061695337,0.44507433890503223,0.1821946518640748,
 0.97316883351689,0.7452013954853327,0.8351456478132199,0.5893446309905137,
 0.1506933491780883,0.581080184602108,0.014910214675842526,0.2643860638678365,
 0.2831456372059924,0.9686493199105158,0.7728964051327273,0.5853133896220282,
 0.21724773923407892,0.8300099904737145,0.2516060906201213,0.4306442220119886,
 0.668311372498613,0.1528837069919884,0.5637665542994541,0.9326644899585708,
 0.2557189824609939,0.8903477739330649,0.3408926153263131,0.4515706724419465,
 0.11087351264742529,0.026663782253265378,0.28094383470997175,
 0.716753731106415,0.9940177120128435,0.7180360495631034,0.07672141301209201,
 0.8031220577470147,0.39922710508963044,0.5807489790560487,0.8335230153821303,
 0.9523725086189161,0.49300227776907446,0.48733191516941965,
 0.9089570194283312,0.6683721155686414,0.7989412725346965,0.16581203387951593,
 0.9755255493617854,0.7481363417410615,0.486453302752622,0.37654781846615903,
 0.2785271114006418,0.2675706351673479,0.8409686246554421,0.2690339877190099,
 0.02227417132143117,0.15233427705460223,0.16087918920158162,
 0.8640903757303706,0.7594990645532141,0.8593735249495318,0.04677986628614539,
 0.7306780890309845,0.09512173095040755,0.9007437886914118,0.9691401686131063,
 0.9141212614619225,0.7160605430676942,0.5937072944400116,0.1826358036186524,
 0.2600997236019891,0.7171508359350368,0.08803646877193294,
 0.023294260355608776,0.38858746107248787,0.09824105658977134,
 0.5141885635265901,0.43811333276009967,0.8315056461923489,0.2024674936440497,
 0.6109312317690999,0.2577123701239099,0.5740744595803932,0.6528162393194876,
 0.7821753816678917,0.552777756866371,0.34162745123953886,0.8334444332237579,
 0.8201281384471124,0.18745928028798475,0.6594292704228011,
 0.03280523574276961,0.8035553562233729,0.26412178557465316,
 0.7850003996897162,0.6933538562743534,0.5758254897258259,0.514909879466149,
 0.8030545861234132,0.5976538311965767,0.9698228100271972,0.6105108918946361,
 0.7963394562547162,0.2155196606168749,0.8917058544849383,0.5645354044335797,
 0.42528135233126607,0.5815549892522419,0.6024732358384259,0.329815287811953,
 0.9920307805233407,0.3449489329008647,0.7266025022345008,0.9284767901088288,
 0.38827662159780896,0.9354035118948463,0.7078336214741427,0.311020678278234,
 0.42188824297196903,0.11001324013419922,0.5977099362077875,
 0.08093789482457059,0.1267027326029646,0.490660380358811,0.7577300106537749,
 0.07042350609305392,0.07053601688543609,0.8379368658413587,
 0.5639125083062476,0.9949022492625138,0.29895681600950275,0.7626587132149407,
 0.05149809127426441,0.0801268732688063,0.91634551299282,0.3902431979590133,
 0.2656366571144816,0.8490929253843869,0.7225042843206535,0.24551661480176157,
 0.34567565061533156,0.32578300111844294,0.8156576716263466,
 0.9262142422915515,0.6748356674827545,0.7249895553128118,0.912629627052882,
 0.057196670533204985,0.18442852413288902,0.8866036824349633,
 0.8982418723765468,0.3799234229491727,0.2891041976183647,0.5484705880837297,
 0.6906760655680063,0.557328069915407,0.828720067985156,0.6431031953247527,
 0.9731355697315016,0.30022914589179195,0.9607100023616795,0.6219322330075178,
 0.7364769724849317,0.35237115493495186,0.027108538016322625,
 0.8955326157612101,0.3324796616801967,0.5056076462280737,0.4943168241942091,
 0.40660530307142284,0.2288881947911312,0.05677844877698024,
 0.5304828990610084,0.8999830591210236,0.8222303964723964,0.9943366557207753,
 0.1464758275826148,0.5867723851101049,0.7587100925908895,0.3406351567486364,
 0.5422612159881195,0.7937494591268118,0.4232575504785697,0.12139097712213809,
 0.7057519159142452,0.26756910729469996,0.22960353560670776,0.995126781352273,
 0.0643238228591767,0.215943884394322,0.9636852053762415,0.8283161966324094,
 0.17685975510156304,0.6777324707786722,0.5202891449183273,
 0.21004197387479262,0.3190955112716677,0.8104927127917694,0.481422099095403,
 0.3936733371124951,0.9402451649059606,0.20543259738424258,
 0.30071443830331257,0.2967884974087439,0.8473173368023793,0.7355972608200431,
 0.014866409106121692,0.67718949625772,0.46699197833991846,
 0.40467770036802875,0.5055120792778192,0.6309326898854157,0.50086488370472,
 0.04705767449653486,0.16339256622456033,0.1489383329000029,
 0.01074240218182998,0.6086201478630939,0.7982608847219015,0.8999819217712832,
 0.8049501391819796,0.8774288627170236,0.14630316788701703,0.6482125017906867,
 0.8801573671475235,0.6590847437367167,0.878123346960676,0.9340106798831946,
 0.3290383945760864,0.925987630398109,0.7616021926664134,0.6163729471932982,
 0.387927853567421,0.2087100537614215,0.05648866800403274,
 0.013935425733817675,0.8869700784602221,0.33367744961732526,
 0.4108443943465169,0.5599928592243627,0.9410526638448033,0.6031356447763127,
 0.615285345489721,0.06898822320079812,0.05188593561382626,
 0.36423395039031403,0.2966849705421394,0.6702801597015771,0.5589348909824096,
 0.6712638247507253,0.23805115099638985,0.47123100454562206,
 0.45739229406912885,0.7693942118176001,0.3834528488435843,0.9861215385029984,
 0.8592712622633456,0.6975621486692924,0.46631637121768255,0.7648270056402535,
 0.07559866581184982,0.7841995459194058,0.19008185285137402,
 0.3217426110550514,0.868130782098837,0.004145157843600145,0.5580000885131697,
 0.8148959851376089,0.49682297586026236,0.9492578894723736,0.5050338568709462,
 0.7384753399423326,0.4321180869757668,0.9993969113101117,0.972007290758793,
 0.6008595544382251,0.4777341361301214,0.9259380923415078,0.13621117372818625,
 0.19943970465718275,0.19461791335395606,0.2727417674092355,
 0.8133219458356099,0.15600057585227767,0.8629282706517541,
 0.49985923136066446,0.19119965477298173,0.8326469798551176,
 0.8663417031416288,0.9043789689999725,0.8490906508937079,0.8711067379732367,
 0.467617733790223,0.5810949437478505,0.3997590413662736,0.213234275007276,
 0.21012311687822904,0.35532863236456014,0.8505854613031824,
 0.39016240365528077,0.774985751551106,0.18849523124961032,0.6422411627706276,
 0.8486865828290954,0.8512695508692572,0.865878084206585,0.27303105713258324,
 0.5463222036552079,0.43119332370457386,0.7013227841222411,0.7101643445643293,
 0.3310191255299093,0.44959994458428243,0.46666880865984306,
 0.6741571656143401,0.8937066369675507,0.615873860196622,0.46291675531361165,
 0.6036724570977,0.32258797257165306,0.618535845831299,0.9782784408394443,
 0.7116457928927067,0.2839684284444929,0.6852412218829897,0.9439668001679313,
 0.055331520437749804,0.6129128779598123,0.5219266135657071,
 0.3241372779802817,0.22399598676635524,0.7508846209450395,0.3708016031860293,
 0.09404695321852463,0.5097362906813034,0.9665389427894614,
 0.20993133177310186,0.13018308034492942,0.5624416670113673,
 0.6417062835278722,0.587133970114953,0.8597647906900368,0.7980657515141074,
 0.5209748481115922,0.4245133577970266,0.4796245766987298,0.6104673199951381,
 0.9725382401861076,0.2301558759034794,0.12990392260713046,0.361840492923048,
 0.4213383579254959,0.5304511757558178,0.45872934920549513,
 0.33313506848319085,0.6701944434079726,0.6795238610227912,
 0.029302279661675867,0.8591116372025609,0.6607425671700744,
 0.47999430058853243,0.7490181918909703,0.7718995786109951,0.5067777903649012,
 0.7826647586871409,0.8921934481565947,0.3617817930812036,0.04145660935641404,
 0.805313235929891,0.909087271581495,0.5879319167742392,0.5078892639574089,
 0.7721461090023116,0.0476323833295339,0.03399290462496529,
 0.20503700725612883,0.9996304555814606,0.38512965724939074,
 0.6093360111811782,0.761889805713942,0.2811375848762345,0.7745664139529415,
 0.44994412633577574,0.19225871008417816,0.13041750216374348,
 0.2707882632537868,0.2864831283024514,0.2602323606122442,0.3530040913888567,
 0.815668384815461,0.31978957660980145,0.2538194446358122,0.3359461703399593,
 0.5082625432894092,0.43859919549226556,0.37903862026221435,
 0.1507746757755909,0.11413116018863922,0.28379914110899707,
 0.15643879535112204,0.6386697868959814,0.46043840504681177,
 0.8373364591054571,0.12972784167204587,0.6131924942803553,0.6598173451663133,
 0.6529473044917632,0.47165863378511474,0.1371864405208597,0.5004727493272525,
 0.6220079937444146,0.501628826355085,0.6443116751422713,0.3577238939516909,
 0.23220239554326838,0.08108900395548946,0.23441327562762904,
 0.6753668369828374,0.02009614920603009,0.8179741695803102,
 0.05995647995612219,0.11403011487592118,0.9970387067430176,
 0.7306964622453883,0.6941359786131122,0.20956765275793343,0.6466866631587671,
 0.7418722014091224,0.3310408310452323,0.10782303989447506,0.8680023847158167,
 0.2052771547571156,0.9928848047461566,0.5129822244142117,0.0571195800803499,
 0.06464372687442155,0.27363390637667784,0.3622521948569666,
 0.7208453789620736,0.8652616183180194,0.23100819614421852,
 0.15657868322746338,0.6104804092337022,0.09752502684866604,
 0.9881915995260709,0.9509757843490446,0.5333928664441282,0.8007593388702203,
 0.9918085621463869,0.8150674252856532,0.8433535354279135,0.8414072532185188,
 0.1568784467376978,0.7388076562070063,0.12872418653285267,0.8235762644116547,
 0.9752962967878238,0.6109213708450811,0.21257218596872085,0.1451710690150131,
 0.9787824871850833,0.35869878504814745,0.5179672227529184,0.6657412748970539,
 0.0227306736666536,0.7497831107735137,0.6609926573950753,0.2943464951455753,
 0.2272239502239226,0.3371700578016785,0.7902727811911614,0.8166588945712603,
 0.026987089324135516,0.9661681420318969,0.7471371102978283,
 0.6463328553771238,0.6797262479031537,0.38800973197020516,
 0.08116931606135525,0.984903278410972,0.8932701399558746,0.4518985738579664,
 0.8466130463609683,0.6122318069170815,0.3884741562793206,0.4664935731029709,
 0.27738526794092133,0.6544595412160206,0.6593453383326925,
 0.41707811797315963,0.18968312968978362,0.39802403648810447,
 0.5133645516136023,0.33934430420417894,0.41732927723437196,
 0.41391279149067517,0.5013549158979247,0.8511286131039026,0.4569145489147428,
 0.8873973818687059,0.8803673669985126,0.984967735811443,0.18187570994670132,
 0.3116284274900444,0.6600963161923785,0.5292175718260649,0.4872818285256745,
 0.8800308262782882,0.2511532496585104,0.041878766813919666,
 0.8541323990420726,0.7005876313564268,0.584653222861824,0.4651270515371083,
 0.3444065935619005,0.2587388964636006,0.9386689758427723,0.2907921618650281,
 0.8808298323675477,0.31435907038302224,0.9705097225576563,0.4485801410819942,
 0.9578517287999391,0.445790810139966,0.243915733914695,0.06471414078741222,
 0.354232865998625,0.9047983606397306,0.6941957530218404,0.7802957427500392,
 0.2604356754776974,0.14057948684476418,0.19628770042310983,
 0.9498545218872343,0.9576413536508294,0.32341896637116874,
 0.04594303418694479,0.4044757758026256,0.8842919961014103,0.8370941843543924,
 0.9723167991900624,0.8134336712115028,0.44856268894230256,0.7333843335751344,
 0.5327475604126545,0.09482634920231325,0.6411884185310872,
 0.07090156151275484,0.8736903736933724,8.017529794627221e-4,
 0.07158764079022506,0.5255204039377632,0.6693432559269188,0.5809842076815777,
 0.27902604156603195,0.7561901186291143,0.32633254449428784,
 0.9611204493867892,0.9135448013128145,0.09153390369400316,
 0.04611210575636382,0.9820180934440509,0.051617591442183586,
 0.1643580564807131,0.8233388741692443,0.8720548232228564,0.4370324084664676,
 0.9264142193482082,0.09118897110584412,0.19924781099260658,
 0.9923598970636819,0.04467687098875728,0.41214653214949537,0.41011535913719,
 0.5305616903727987,0.9134907530061742,0.6600155398245025,0.45950628391069215,
 0.03270498225573082,0.7284319825577713,0.7887261726690238,0.9615716800735505,
 0.7522854302102134,0.5349356599311502,0.23068471051878758,0.9695944436338126,
 0.8273230286877978,0.007500949835872128,0.3047210378864077,
 0.5576996280550462,0.10750889389611307,0.17632424652081113,
 0.6682653720377258,0.41922621511198443,0.9056921306940506,
 0.06313827170064368,0.02797188557529906,0.9726935326852835,
 0.6344993739982919,0.785139096025574,0.7896652932014345,0.5792797246702407,
 0.5295433284877502,0.7491874970126342,0.19837273001438327,0.8731692700713417,
 0.08531332867256602,0.6156176789487736,0.5541016206919308,0.4552661182215827,
 0.24723991666368017,0.00347982059485219,0.6021388051252803,
 0.3095039054707289,0.4020552802319435,0.9199684848455473,0.5812579473667405,
 0.4381369379701924,0.6502280939052839,0.8990021736390443,0.4944642404489652,
 0.218377081264286,0.12932053445058211,0.5973013930418071,0.01725467366879152,
 0.28219882759020753,0.7860964933216779,0.519136300032821,
 0.013518003223217723,0.938401782997569,0.4237648609989568,0.4755572852748936,
 0.18038728393590886,0.9383608396436901,0.07935114385226893,
 0.6744547909343823,0.45382765148020465,0.9931066316105732,0.5776263386379737,
 0.9256286354181964,0.0013253938837305412,0.8844483033315046,
 0.7823270398305376,0.13120925286752605,0.2193995391540523,
 0.43981709706997785,0.22374922775878003,0.5015951752710407,
 0.027627537586177642,0.46198904294860565,0.2817240540816406,
 0.886388022090786,0.893133867552506,0.4503171762596895,0.5177893140307178,
 0.9981883257030492,0.3260710502940478,0.10551190361354013,0.788750503884325,
 0.8472137528886898,0.36513373529110094,0.7957648301650999,
 0.47829605261647734,0.464915356114294,0.1744097105860437,0.5255295476204397,
 0.6386225448102696,0.8370230774564684,0.8245568158655141,0.4403939852080614,
 0.7138794280872517,0.3539268383479449,0.0900655240563829,0.5861868183923907,
 0.6710359494304633,0.7717406451130884,0.7838164334280625,0.6576547540940523,
 0.7533182933370548,0.7564416124856421,0.6642322636333324,0.9529081193976842,
 0.9047501630005377,0.2082977982878128,0.6416971806294006,
 0.021720602045012027,0.9787048573045259,0.33217328013916025,
 0.6735813908817087,0.7172629251775392,0.8896633572680505,0.6626366800342959,
 0.20809281528013912,0.3984836084532314,0.5496944478765805,0.8362602290300777,
 0.8736658600517319,0.02030913690394942,0.2666916149965557,
 0.07050974936456156,0.9322822327019297,0.29448941562547515,
 0.02168595861253142,0.36455492168324555,0.41236817891797906,
 0.6863494586131944,0.08557068804478674,0.08551375300475517,
 0.09357521370323052,0.06168544311766899,0.560709790767112,0.8265903142400277,
 0.29865314187459535,0.3902270746097243,0.21755134889665706,
 0.21271932606624477,0.7896466403364708,0.9321212625536048,0.6474041748278517,
 0.6681865179952146,0.7731184670147847,0.019476080977070542,
 0.6006538726702788,0.6057260117563157,0.6493190525465805,0.361823846813237,
 0.7935398483475674,0.9249169927702947,0.7387364080103638,0.9678043968264076,
 0.7844764770261973,0.19693640404171942,0.012826762052488117,
 0.5709779931549351,0.8436850974614186,0.7159853858875719,0.07387977505148235,
 0.4074923897735542,0.4545493581889043,0.0755635213847825,0.591816370362126,
 0.7696837518870094,0.715322394793752,0.38870333575562177,0.3189265683161502,
 0.6016683024303675,0.1379532225400606,0.08384758407622639,0.602267687942109,
 0.5725559662794341,0.06686290687652185,0.25005537896004804,
 0.08054426659064329,0.5372197885641549,0.8631984056418451,0.6561715576222734,
 0.3952166599113771,0.9456610263149674,0.5401277493530853,0.7405708455414959,
 0.3848438362315144,0.6779025903269579,0.6841791055312857,0.16616266164256666,
 0.9449752395205238,0.24487849644156268,0.011585456924246063,
 0.31739415710752206,0.019286869964005415,0.211982062474954,
 0.4188001180390415,0.5623901828037802,0.6222869559483923,0.6882994856787961,
 0.7133603460651089,0.5724992440315546,0.8680993622659433,0.8283087774274185,
 0.06603423865088998,0.4230397034288613,0.4413046664266739,
 0.004253614588682453,0.06973859314940434,0.7647032035084135,
 0.07994425852573461,0.05461861308811744,0.16706398144725676,
 0.05817113700382737,0.21928759763222183,0.07278922612524308,
 0.7453348209455954,0.32865819876541225,0.6176413296714706,0.7943236820754178,
 0.24327755877297652,0.06055006143628894,0.052181938705838427,
 0.5644973091957061,0.21880387566636672,0.28881203943775335,
 0.8545219448353103,0.47211354063575883,0.19958088214503367,
 0.8902895346983465,0.053598322206257154,0.42299173362180253,
 0.06635272435005124,0.5344130808939577,0.34020366113501643,
 0.3555430910728756,0.08071830312954564,0.11671506446557123,
 0.5406399163047118,0.8063058130615178,0.9046149227746181,
 0.013920990065616312,0.7830228235097313,0.19232314150692975,
 0.05381397336566496,0.08276720816731697,0.8047057775238831,
 0.7659880790227461,0.7137424328462383,0.030802711348347223,
 0.35018094959409984,0.48348849761413426,0.45868735221087586,
 0.011297398378003032,0.5644728632293757,0.11946698469319728,
 0.8232264296947109,0.49870786004344725,0.048206605639581235,
 0.5672739466884493,0.5706824410428455,0.10933100372253635,0.7282839948616411,
 0.727849976733039,0.2598087274108535,0.14440660520195603,0.7289496996370788,
 0.5888739098563676,0.9459066554942982,0.8326212438629821,0.42018118343001154,
 0.1554908984213702,0.5386379252422208,0.9062401650979027,0.5143148575187224,
 0.4698477583741696,0.4705655860711989,0.4863164317727767,0.9383856243213297,
 0.3009792052969483,0.7160225152544348,0.33932901495296064,
 0.36631300591163873,0.16076719500119363,0.1225794908287896,
 0.05459459282999135,0.38838711837504647,0.5387693651528067,
 0.7052981750015204,0.43616059573169297,0.5786862201053882,0.5110653058678231,
 0.8923819747579322,0.9224013761245666,0.10167815814024761,0.9403383818853946,
 0.37772442368817716,0.11577982107319562,0.20796735710748582,
 0.4106318888497976,0.7421375951190157,0.6606616845952183,0.8638424891890745,
 0.006971105279244161,0.37646402372609455,0.9350961590526229,
 0.5238265334657223,0.2621932990478082,0.6854275823681533,0.6464970276414508,
 0.8989599475176279,0.8766015561232117,0.39719146042784714,0.6900018646902231,
 0.6340205865575984,0.01395819008044974,0.8222157711671704,0.5513934334890352,
 0.8823035137122781,0.8808145264233251,0.56481843804873,0.46265125663162754,
 0.217228456534043,0.6356326294828637,0.7410938192780229,0.7217590748095103,
 0.2244958681018926,0.22014958152614394,0.7883201458028011,0.8910525661903158,
 0.8041816297337429,0.6809556566287173,0.39831171641958807,0.096076885208906,
 0.03535838835261307,0.23197432976817356,0.5302642254205702,
 0.8362167785618171,0.00447186673087363,0.29269463264368945,0.296462124115245,
 0.28729373937195035,0.8346348380580126,0.24161019601786082,0.653895074114879,
 0.7306244125654057,0.21228484708827766,0.37230695686870374,
 0.7503723177613268,0.46643045995733634,0.6460320712939777,
 0.30899022669842624,0.08855650350009092,0.20161277587479132,
 0.21473323127076172,0.4712995417734185,0.2523343002400744,0.857015318769581,
 0.8479692189364944,0.874949375173711,0.6458964824716948,0.7361664305832737,
 0.20051593904333498,0.15051645746697895,0.28284288263103596,
 0.8094456869900226,0.4462259740767084,0.8684670960369116,0.47953927246990413,
 0.39365107113159004,0.1963282639366456,0.45577334524037094,
 0.6456337503898284,0.12413496589661122,0.9677537754932395,0.4815277099180939,
 0.38007316497232135,0.6139127718542462,0.9826067618110268,0.8633428733805316,
 0.913037225629403,0.5202977031833214,0.9962394879978286,0.32826974485816707,
 0.8701220878930568,0.21933044096322774,0.13117617525365,0.4741851209225656,
 0.7474190678742731,0.1695239350064185,0.9043963991125109,0.8571869839675111,
 0.34292349653266463,0.5628541123606698,0.049861366120225625,
 0.3680203343818046,0.02352152134341534,0.11246740396644195,
 0.01967844634794902,0.10027417993747556,0.7398967626724903,
 0.7796257449203854,0.07917962365535436,0.27959267194983894,
 0.8070570178939186,0.28665570805927854,0.45309862813632984,
 0.09047278571793482,0.1859490860872891,0.7143831480586754,0.4722534214442119,
 0.12432885473755251,0.5360005302117347,0.13208048483995394,
 0.18072081481599123,0.6610550229785606,0.3142690872902405,0.5975455943756985,
 0.22631246073293976,0.6744942437638246,0.9602203336665587,
 0.12234100066566256,0.5390369326359288,0.46056025183715654,
 0.7102813375457409,0.6300200746988567,0.43749694452309096,0.1609038322786529,
 0.0555414429744705,0.25787954869106533,0.45743571527505267,
 0.5126155842868889,0.668382994276792,0.3074518713616643,0.876720749298895,
 0.59101619499431,0.6213864737850289,0.9981114977834047,0.5285968845091268,
 0.043174837284377876,0.42329526360924885,0.3757936287051723,
 0.12145871966588129,0.6034421390545633,0.9990884860583962,0.3236011356683096,
 0.3777137231158485,0.541715233646554,0.8752921700346525,0.7498748954201327,
 0.041867968768718566,0.14761746316202784,0.93320746830411,0.9046569850247118,
 0.14689075940842333,0.7469668417580955,0.6710809397899522,0.821836797238024,
 0.517241250857172,0.5622653152247735,0.7150970856415197,0.11910365509817356,
 0.4043162997149876,0.15331787599048097,0.06333326047454291,
 0.24037839812784445,0.4962120085022663,0.20727031040131172,
 0.09715186831393552,0.2418157116542924,0.34685624239888,0.007538169559696439,
 0.017919028915380775,0.19392695245545422,0.939441412709865,
 0.4840419029132208,0.6814977372262482,0.8470611210222232,0.03682394451277626,
 0.8415343085362302,0.09009701872100773,0.9076857297548054,0.8788555425269633,
 0.21198934442766082,0.6469017332674412,0.444440248953047,0.33191827744076596,
 0.24226029883040634,0.9438115222033563,0.9075840315559386,0.4442811390567418,
 0.4647372177091398,0.09709957674445746,0.7713657177151467,
 0.032760231509952775,0.6381624308479322,0.7968258380671351,
 0.4993701713166463,0.1846878235709435,0.7417493453363582,0.8176441292001488,
 0.2644426813821885,0.8569906316669132,0.29765847743777507,0.5917995222017036,
 0.06837106512442293,0.7790219999564176,0.877431932653364,0.4459413645582503,
 0.42862341199840126,0.578677396287565,0.018740533299549167,
 0.9983253399357587,0.9700578332425196,0.9430208784114182,0.2341167282073927,
 0.6462947026159752,0.5196306297767304,0.7227933911551023,0.8630556719765644,
 0.49768874721124823,0.29989491633464094,0.08605728872700191,
 0.6131222377195299,0.7540503009683394,0.780615728544099,0.7850676553353653,
 0.35684630051531985,0.37568171328627586,0.1871225008630083,
 0.9231693977802451,0.9910410741682625,0.3678233963951335,0.07479430410231047,
 0.3797029133362676,0.9463224873807913,0.6845594046780975,0.4982466379277608,
 0.7977318897220333,0.05728201051867021,0.5355356136664687,
 0.42869429210711063,0.7646798894409333,0.43495192063593224,
 0.19629509197627693,0.7299865496597433,0.9288313968389341,
 0.24901923352780808,0.22668888916395435,0.5485758678068094,
 0.4001600838191224,0.07541916434474905,0.2637399694556496,
 0.21514406815705178,0.12263401992915768,0.39178634272844504,
 0.2407617780522181,0.9142878455085844,0.46291368755625617,0.5857481436908961,
 0.02677056137267886,0.9808808756524627,0.07253359613415977,
 0.6831260612058003,0.16499211735164165,0.31257166632754263,
 0.8330124689528079,0.8927459518687063,0.9219432331134487,0.23388886665520658,
 0.5730820721405586,0.022415316922974937,0.9012967374665176,
 0.8730894493067359,0.2807239834464982,0.49708087425876624,
 0.19144054700255064,0.9288478134421586,0.9445856364726157,0.7184486850714403,
 0.32621226589418184,0.48791271119567203,0.9275961640744511,
 0.040998684021683274,0.37731793693992144,0.7229914129454839,
 0.4130941022580188,0.07367478645354508,0.7021579395853847,
 0.021802729627894535,0.2282095836505631,0.2992123082793027,
 0.26513371844116174,0.8996921850663613,0.08314293876410572,
 0.3849580937748933,0.6223109498169298,0.2351800673075345,0.9817609558084086,
 0.8677717247273355,0.6672470496120667,0.6203016940081573,0.7562658186621565,
 0.7307479974362076,0.24868171428901698,0.023731042584878792,
 0.4681599431754584,0.5314606312808355,0.9837708717841052,0.45281307389732084,
 0.2112337313118725,0.817979556838297,0.179621642372914,0.09395595459627304,
 0.6921603677600127,0.5082537442129528,0.461491202739463,0.6927984347754824,
 0.13929702301458113,0.7784021403268643,0.4156820092550332,0.7641380674281082,
 0.45894295439083044,0.3458796320210854,0.19512137634080484,
 0.4015765541359638,0.3295777891283651,0.5120767689243091,0.8518333387551515,
 0.31927759299121705,0.6136444549430647,0.5337293646584618,0.9367692923956786,
 0.8049944318296995,0.8888993333118778,0.5419927158345417,0.514117314493592,
 0.5763291699064804,0.21268698568447242,0.16037891096086043,
 0.4118687338814946,0.7557092377734851,0.4909150366559041,0.08631516220846291,
 0.08483704201002606,0.9031743501364923,0.793431042090762,0.24280682740468862,
 0.6408168153953311,0.5673698177489309,0.41711106857636615,0.6378449140610345,
 0.5554113742490803,0.2857686714974663,0.5010589149188222,0.8544304137978774,
 0.42261663172509667,0.4601967533884308,0.8712016625951571,0.5989475184548003,
 0.8474619262785662,0.36887601402108616,0.7258470972307195,0.6621970681957168,
 0.16462633137173854,0.9664482145995512,0.6735633491290641,0.8035133954457235,
 0.2979291396774929,0.36722035919152085,0.43771967893965036,
 0.9902391042626911,0.5551483167913002,0.8900593729630517,0.9467206654400373,
 0.40325750192314336,0.20608928152261152,0.7000384765199638,
 0.4052158627718292,0.1958130603464714,0.8500184753397806,0.8054574578069683,
 0.14640513818267442,0.9539906845299129,0.6123172503965087,0.5032423382730749,
 0.26180875678118565,0.39645158327312857,0.765084357035672,0.7344653470530335,
 0.09732757233719247,0.45736017030062737,0.5021064428898074,
 0.14661470357781836,0.3671520237428645,0.4324767868140863,0.4687520994539609,
 0.5972609752203292,0.017860172046631484,0.4506001781545619,
 0.8227232932060837,0.777561910234859,0.4981507234976943,0.38316396858973434,
 0.48397320577597036,0.2626705792783033,0.4497341283144016,
 0.37523866124845684,0.2487726579735945,0.5189998623503072,
 0.38969285866932957,0.1197706731731889,0.8506984123774067,
 0.10669391775259329,0.98745507853335,0.3748834445771143,0.04302062818998997,
 0.25701011510376714,0.7913307086546506,0.14802627908588717,
 0.7898501073422337,0.746980501255115,0.27326924164435806,0.350659739478991,
 0.13388492223789938,0.9836111577000775,0.5622779715524536,
 0.027720719682293615,0.24396719504881748,0.9763666481408785,
 0.18406787646452583,0.6994235970025187,0.025855773796217507,
 0.3857518824612778,0.04706866935549592,0.02846230965059582,
 0.4654227075769187,0.8860692461359357,0.7123375810278781,0.1294348677603756,
 0.6075960533808216,0.2339954317038293,0.7547926200449482,0.41973921947556203,
 0.765599366370983,0.28456293982703607,0.4573594732130585,0.05390626258370801,
 0.7407704492457398,0.10423115191105725,0.828140147000683,0.5857155372971751,
 0.246453484538945,0.907638635492845,0.052892127347281215,0.7141575028171735,
 0.8819971075492511,0.27696768706637864,0.2435804642845414,0.2450077357983016,
 0.9731752995721366,0.6690906819684578,0.7608636677105675,0.5491534045221981,
 0.14960361654449916,0.05338506665918674,0.3991908601436166,
 0.6605981537357202,0.44144455963568874,0.5224039047041047,0.8697565918472928,
 0.26016554550980797,0.20977862446644957,0.9144937309691491,
 0.3509567486298355,0.3677250792872374,0.6841582512458404,0.18236084652785833,
 0.904393039647983,0.4121603884909921,0.277831742282318,0.07725626622775328,
 0.5427947207471274,0.6947032813372512,0.40204001527113475,0.97302721300767,
 0.9838209757508656,0.8743558334803497,0.08471564377371288,
 0.11110207188587662,0.9548223558030612,0.3938242269964296,0.8723617991674039,
 0.09719031858013483,0.16066132728742166,0.7489388378788688,
 0.7766349548647018,0.09108902553387421,0.6932295147755383,0.562284549106102,
 0.5638518462607804,0.5217361308104674,0.7996030485021788,0.6891213835365415,
 0.4444734584076252,0.6697009807692103,0.9114624044101629,0.7496948201628877,
 0.20056728611133756,0.4475250677119019,0.407056051036099,0.7300797022590488,
 0.5632438618204982,0.6029101782810661,0.40989420269230203,
 0.006410052974753944,0.3326567948427691,0.18060268534359203,
 0.3682903960198449,0.4560468793722384,0.38003419363692403,0.5731397846294145,
 0.9930005142619747,0.11156613406450144,0.9304689568956099,0.926113649640291,
 0.8023941933753306,0.8653543248982221,0.16752138722799914,0.9910680412485362,
 0.21432420748716452,0.9538519746035735,0.12258372629791048,0.655026161039034,
 0.2998654948228344,0.6913571675786798,0.9198508326637351,0.6041336706365995,
 0.7015318011509877,0.8135359906630321,0.36317996481483816,
 0.46739725152841904,0.883043689513989,0.3085058116860907,0.7599806985870042,
 0.8596586254716262,0.7391830095407358,0.9360262911793327,0.29933113742303186,
 0.9812968953260559,0.4696390178256995,0.33223886176789086,0.5676963726377203,
 0.7219062561757862,0.6936621826570336,0.37121522139050445,0.7167410712016458,
 0.5678180829668961,0.20593665904997216,0.44356249384462576,
 0.8875349515133114,0.7352547471944122,0.8921179682402411,0.4343623593241338,
 0.9945901586871096,0.003383805663846551,0.4937083197912888,
 0.5959475680633581,0.5000341703592687,0.6601026912534106,0.32695079073835487,
 0.52746938763472,0.016206596042921406,0.14566899184674287,0.835587440371468,
 0.12294427578946554,0.23855299972783928,0.621194626840802,0.6542662477379126,
 0.4834792394726344,0.8147681925185943,0.7620273435181834,0.5970587951647616,
 0.18191665266154167,0.08561962395669398,0.9580794861002818,
 0.7878790428794865,0.7558237797872606,0.5549050189231463,0.30933150724415726,
 0.8060321173770791,0.81091593656725,0.6252602367537959,0.40810384764576335,
 0.18478455746913447,0.47420041303817595,0.03577715460032871,
 0.263521308572064,0.6860672643276853,0.07495563351485424,0.7072610654773408,
 0.009565763576237618,0.4389978667718014,0.2295983345586542,
 0.05019498919595089,0.9860192557421358,0.7876841543784361,
 0.07288305380050941,0.0040441201593122145,0.3032100726954887,
 0.9686498533632555,0.5874490793354534,0.7837840455986229,0.4181338155833816,
 0.7159536795124122,0.5853764701160538,0.4414319152195961,0.26340502599981264,
 0.30666826247905,0.7755387528476119,0.45051886723131385,0.10363424669158983,
 0.08402533391111844,0.15161118174587707,0.010870857046542004,
 0.7873752954261677,0.7174403487458425,0.8325726115017457,0.08594921943269462,
 0.5848559265464368,0.2793573702960902,0.5596752485150942,0.6527340722964065,
 0.6371174859766545,0.9961637998922862,0.3806629562956645,0.8040503567164176,
 0.22668572965499278,0.2967580684452068,0.5359407276904972,0.741922004208646,
 0.7769448615228575,0.49696685804559915,0.21920771157290453,
 0.9014437891582542,0.7734564213532734,0.6577831232780913,0.6182900514130122,
 0.797150525743599,0.11920497904631644,0.6223506234848049,0.6292152764393794,
 0.6993831063985934,0.815868849424078,0.2929683349476294,0.13330147231541356,
 0.9108847893731475,0.26109725523177185,0.976034317489873,0.2941415546578391,
 0.8873819586649525,0.2659951913644616,0.7231539999837072,0.28830865627911084,
 0.6524495379401324,0.8840199678503524,0.25219731703528325,0.9821479448639118,
 0.14480950926395886,0.024429970940008094,0.07580131334780627,
 0.6829977257829193,0.4589563242085746,0.8256792451494845,0.9709219093383277,
 0.6743695896996306,0.8234525146718479,0.2865360833182369,0.9758826749620875,
 0.8134575624371796,0.01828145327959929,0.07740608858656184,
 0.18735183294838653,0.3864481549167721,0.3879601558862452,
 0.30994413547018707,0.22881008698031347,0.15715023711487508,
 0.7358065153006311,0.3486524265531842,0.20892161665590814,0.4040585763850172,
 0.7450448028197544,0.3036817578431894,0.7258480810568817,0.7530333009534311,
 0.40391319646625046,0.9776156192118247,0.7737783734197597,
 0.40007847217019354,0.5016607139462979,0.993408597989293,0.29202922617167015,
 0.6415843699022493,0.6578301736458995,0.13765131447930867,0.3665550783615614,
 0.348947020615896,0.8311660303259186,0.3483080721229793,0.18517283217136182,
 0.6697326825811015,0.4430019897906682,0.37534596615821836,0.8083746499042421,
 0.2285048958683349,0.32994775705198864,0.9672142902385638,
 0.32478240929259805,0.45665297776881053,0.7465038465665207,
 0.28919203215968103,0.388820908934033,0.030934988670675567,
 0.22171804765909997,0.17399687857724366,0.023493559853095647,
 0.9813770805573844,0.49495253286140994,0.7861206273123253,0.4326033315115766,
 0.23593050286721362,0.8987163227115497,0.7163259704330763,
 0.09114207087242154,0.2216779652297447,0.12101440544940567,
 0.7000819222451179,0.9866723325644036,0.19381288196662094,
 0.08732417451193186,0.5195470901664615,0.31596389962639126,
 0.02351932159261072,0.08534740139318442,0.969662458332648,0.8649661497051542,
 0.2871475726658984,0.05801418481928333,0.16604669622029689,
 0.3597856447047547,0.08976283852356004,0.7796560564363106,0.8317278838256484,
 0.8155597109346187,0.30159120967730524,0.05805992348605682,
 0.4154622072423324,0.22155744246277642,0.16583512791683785,
 0.48299681442670506,0.298711736946632,0.33729868694015097,
 0.32349082713178323,0.16049673700612654,0.2468471630804775,
 0.8173533979237166,0.6211325932152215,0.1670786099514504,0.13088374371474198,
 0.725533441108266,0.3063273984842456,0.1419760590143977,0.724047728879993,
 0.026072456091602025,0.0028377870422359175,0.05660572329218427,
 0.1458183796777336,0.8907965824846114,0.7314919936819085,0.1816893318427113,
 0.7130871528968474,0.5201320794484092,0.2796591457951155,0.05572771165621504,
 0.9394151379004417,0.8041050790018799,0.4694090519060474,0.8072417263472806,
 0.1950911799449606,0.5367073689904852,0.5228465452131738,0.9396597501954096,
 0.34034308505261945,0.4427155463187271,0.22108231770663989,
 0.6167671085500913,0.9513578524217737,0.5249429360426099,
 5.923914064309699e-4,0.2903930939264997,0.6898249775440158,
 0.019936845877997156,0.2826325846930806,0.5225061111848414,
 0.3161831179215693,0.5588164240227008,0.021726898469721978,
 0.9594370189396904,0.6570174810735117,0.45564877854116326,0.7361685804188446,
 0.3080569206160201,0.5544674999490243,0.9117057595334876,0.6397940357801277,
 0.2756304916343987,0.5400563949023439,0.11800041672342543,
 0.37681625561861354,0.4400035178523958,0.5351668487586912,0.7534452156062251,
 0.8720707714218917,0.4913833350411223,0.3155972200419187,0.6654689850900297,
 0.22631864701314153,0.22210129805175316,0.9547284463761148,
 0.5287916287505271,0.43573727665913997,0.3249219293184463,
 0.49576641159161217,0.5875854396249178,0.3163167717862535,0.7202620270364473,
 0.19836911571408689,0.427577685570772,0.4679470010631508,0.9780316735552611,
 0.922292825822973,0.5000929453389535,0.14208038550722624,0.5380451084631496,
 0.07142874483399864,0.47056473303392543,0.994598432346989,0.7794868683784408,
 0.977668878802359,0.051545785165443725,0.16416851970799196,
 0.8753643861092918,0.9889938888443733,7.166142095754724e-4,
 0.09570584921152958,0.7511073437663371,0.00275586521954585,0.508130049291596,
 0.04336765775354279,0.4104003091611781,0.02807961608731946,
 0.7226463971300662,0.14737677822131523,0.4842577799465537,0.5379363063912189,
 0.4410621485181818,0.5503484289019989,0.8819886201431668,0.9974398835607116,
 0.13315881286831044,0.39894527065085394,0.3323676113611128,
 0.20723600605755288,0.4994812052030282,0.12877836439210055,
 0.3998165986351645,0.1277089230144559,0.7480125376986482,0.632941967018248,
 0.5348824831429125,0.7724565511604364,0.8040867582189982,0.5721711663692993,
 0.3120541026252468,0.22465693211106785,0.16942445937172212,
 0.37522506332332706,0.6012849594913553,0.9183509363969365,0.7525165263754907,
 0.8815862315360213,0.9969362336452102,0.12060702049942784,0.828733000438085,
 0.9999340701382237,0.4185803076592086,0.035643796225059754,
 0.3171064743229115,0.45127692015734155,0.36911715860000005,
 0.7638613704884587,0.43708830583909253,0.8155364469739205,0.6994266113308145,
 0.8869573352851122,0.773482189062753,0.6091220104567309,0.9249963740904864,
 0.8394240269521699,0.9916178867483472,0.0797284079100068,0.5433072094399887,
 0.35446824151425416,0.7403677319532926,0.9219437150496187,
 0.26877931270084054,0.5512392897746722,0.9516869931925905,0.1549356153998156,
 0.07460622215195112,0.5977302091161364,0.061831516466187786,
 0.10324601861074467,0.30347674030273963,0.6178478627292382,
 0.5874731245835861,0.4907797097457135,0.207101543072016,0.42186079826089806,
 0.9539359740142346,0.41733325031429036,0.6509302609979958,0.3896883647643128,
 0.9306682476293487,0.9961333587630072,0.34404684320823853,0.2924171792621931,
 0.8877850469517702,0.15050729264902796,0.04953387432178791,
 0.8177237387728499,0.006187852392471216,0.6551972588222617,
 0.7540977679805695,0.79118105621667,0.3047643500056002,0.19416134546449115,
 0.03482889912540432,0.30129228348921966,0.1525700371649028,
 0.8145281175663016,0.09900610114319408,0.7678908097217394,0.6580078799156991,
 0.915947805700386,0.6944492288460621,0.7248755952939449,0.979619601068765,
 0.6980569841299179,0.3429391668988635,0.4064047070380149,0.765234899601776,
 0.6008528728383317,0.21050691486821194,0.2556681174497144,
 0.33073736353376637,0.013798883656114258,0.19109732691579362,
 0.4016120003411716,0.4596803607412918,0.7239264261408392,0.10757144550195685,
 0.42899274192294234,0.23267878517859508,0.5420425025881099,
 0.8445628289891326,0.794602433763869,0.162045976537201,0.6079587520932048,
 0.9295407544822678,0.7361681619392993,0.49509734078828915,
 0.07670437237698935,0.3384640388615443,0.33608303674261664,
 0.8513650812858893,0.458249032253072,0.6461932018454309,0.25731609083863394,
 0.5855087039793888,0.18524285404267415,0.13338283097813197,
 0.9034545717403646,0.38241126019178395,0.8324154640428146,0.5260165812769076,
 0.22147561448105213,0.8710685985616443,0.5909414371664634,0.5122539891874125,
 0.7243497868485278,0.5445464627141934,0.2300521739384621,0.6672294218365249,
 0.6687376045220643,0.5011808199499375,0.1254934082735799,0.6600570500375439,
 0.7126557803237265,0.7740917473724382,0.7988583493917427,0.4359780570576901,
 0.8262242844031792,0.6067193010532256,0.8198346954861597,
 3.109238088132571e-4,0.160074795758406,0.02260473512190453,
 0.35574981643789805,0.7122498934627508,0.9839558608073158,0.2430561659488084,
 0.08086275658559239,0.9428694723061916,0.005021666646953582,
 0.7938741496156734,0.5762425285975289,0.9450011664731347,0.981637334178924,
 0.508557517870934,0.9816587717651275,0.8680862684411119,0.12817206817493942,
 0.35267879535856783,0.4148518307075161,0.17969224237813242,
 0.4108834635649702,0.37460107848136937,0.18792573050001038,0.662170694180067,
 0.43900184182288127,0.4033345602229689,0.2089007510732025,0.6781007642341078,
 0.4460540889465209,0.3432137165363818,0.33480373434641386,
 0.41703230105016276,0.8151445962401668,0.433764840281328,0.4629696612493507,
 0.48954399992768693,0.5175366793524555,0.561427620292945,0.6557894127498324,
 0.8655289721481508,0.3704019233072158,0.43868536941644654,0.8175884415624218,
 0.044169992015373305,0.48935298696480445,0.9009101331381186,
 0.16704969131985825,0.6357447998863075,0.12809662128960975,
 0.9722707896709029,0.7867126467485582,0.0805406411849966,0.31373441042093264,
 0.48680636166964664,0.39496354996366234,0.19209479708463317,
 0.09930661906803373,0.08193019711957317,0.8778609312382561,
 0.21599882676333282,0.6605401011147353,0.4293807257777337,0.2549970274195026,
 0.11305678617666737,0.5599650982346233,0.6330980434555197,0.9289544237994977,
 0.7961166838009484,0.4786820443721165,0.17461753418247827,0.5080857481306176,
 0.40702050443841775,0.7132281427573606,0.4943287378737836,0.9981635272394939,
 0.03198340298860025,0.5391145044509199,0.10591248788354812,0.527174575378974,
 0.14341158942862342,0.7231319692370587,0.8111958083827491,0.4631622331446189,
 0.8555206008316825,0.1486421779294923,0.770037852635204,0.11991666518693445,
 0.6682200970338219,0.7620076474256114,0.09993927244400469,
 0.40807429334357814,0.44610673614096596,0.24517241665305978,
 0.18415461350919737,0.2883611783086395,0.9602233123216307,0.2636560827175731,
 0.39285071680232453,0.4675916020075168,0.6050479795139481,0.8102743182958577,
 0.05131098986054916,0.42263870549413807,0.5597721157165567,
 0.9677131640855927,0.7574580895187399,0.41154700330764904,
 0.40179949062261877,0.9141468542281229,0.16945917321126436,
 0.6559051607469297,0.9981062510948792,0.5999326028575069,0.10844503131402716,
 0.8186673593713405,0.4181856945393403,0.39671934154098387,
 0.27693383924174886,0.022153840676034986,0.8678097044431388,
 0.37793041854437903,0.15015743866424747,0.6857040701687894,
 0.5601884165609949,0.5951087069511044,0.2758027846278357,
 0.040081893992619966,0.23558697801722772,0.5775413908091136,
 0.6387016326176241,0.14556714193560327,0.12438045397712705,
 0.6876991305896611,0.2574202940678104,0.25777503407611724,
 0.08327163569387419,0.8234487523174282,0.43560429295492153,
 0.30666950505795587,0.08487189910477866,0.4706776470611759,
 0.2966694168510273,0.046717644715948525,0.12023577006662856,
 0.37112017585319457,0.12840414039940284,0.3858109282689308,0.540950316392354,
 0.6127050667861766,0.5353524408939616,0.6059475112428204,0.9194406558257394,
 0.9430152945118119,0.16842530206867956,0.6569119710487701,
 0.32893543628722066,0.026183832381957095,0.6945493510414892,
 0.7459457758894006,0.6042667846026526,0.5087415350900655,0.8671466974570587,
 0.31585568526808383,0.5304426020559851,0.670296800382222,0.5909846736505469,
 0.9600505998883374,0.5448562193793935,0.9917626874430595,0.08290649171570641,
 0.40576946482351406,0.07343870600394642,0.7165728290611882,0.903673181148883,
 0.3702791753790704,0.6414181541352265,0.7056762165323366,0.11263331430734569,
 0.8936396582316163,0.19287199126479082,0.273309265650287,0.28662010720809217,
 0.2614182789251098,0.7625893149383632,0.1786810861097713,0.5896835392172696,
 0.379446369757909,0.3827167220572305,0.39612490790874333,0.7525311258290909,
 0.858067547208771,0.21125205423445592,0.677508790160317,0.7806803087126359,
 0.8794020088378351,0.32061077471348465,0.3544028011269924,
 0.014650409322964575,0.4055525464887648,0.07134186712913493,
 0.8400336143520544,0.2658719459875335,0.3772441521296893,0.42863885821548464,
 0.5227836752619678,0.0691367173968177,0.5180286966688024,0.11349587362319158,
 0.2797670963832486,0.53237136339972,0.268969386601545,0.41203860130547,
 0.26338565390731195,0.8949724535036888,0.12386098886671792,
 0.9031712083915827,0.9737202282617289,0.18422842164711373,0.7467580917275126,
 0.8311836684801625,0.7134803667417378,0.0347690338828327,0.01072983854177545,
 0.0245943932575744,0.9198464185286538,0.7687229244091452,0.5638324911022792,
 0.0386554984961579,0.14800172502224618,0.8833913252968691,0.4249211479450632,
 0.7933522331579738,0.3383009933154433,0.5664217700732174,0.06929734853714475,
 0.20063878993894368,0.33775003778119184,0.5085217292212911,
 0.4093653808072313,0.06332156549442569,0.4009527136541349,
 0.21681517159535735,0.41834694960365726,0.020174244624699744,
 0.3452278164139295,0.016170646703422387,0.21796665474063104,
 0.140611810597003,0.32526766470836743,0.6081449777098052,0.32132897836383467,
 0.769119492399472,0.4858459344101228,0.5593248687606274,0.17604064989519763,
 0.08314932186081891,0.652308517135235,0.7070649779327022,0.5607979882243155,
 0.3788451095216858,0.2973209209739627,0.658749240167636,0.14069539354256488,
 0.8680249542213437,0.752847784163968,0.6022485516192523,0.014235158507703072,
 0.6068806619678969,0.5206472467239961,0.8486840753426111,0.28986044744471207,
 0.31146250413077015,0.7037758324284966,0.20392202585348518,
 0.30746772379207177,0.6876898372552406,0.9014083083384631,0.6992927337397823,
 0.0750772456269756,0.15146688594140112,0.028103137531522604,
 5.79809376370477e-4,0.24130839664303483,0.9301066100055795,
 0.42176750963570775,0.7934775919294748,0.06988731346966637,
 0.8319927296428876,0.19892660814317098,0.013246863643784246,
 0.7066654429654975,0.7006055028247731,0.14119582961325594,0.0185423799982124,
 0.8068417729599222,0.8532825584777469,0.8441985493881929,0.4013378713640834,
 0.8802801080894102,0.04650549772357904,0.9244703005442848,0.6695285386891162,
 0.17205264621400052,0.18511532939976338,0.25305647712924906,
 0.9594722543289729,0.06199337982170949,0.6986457282210586,0.1959606938230648,
 0.8022757518421276,0.1620402764192609,0.800767317023307,0.1781756868695773,
 0.5533937670463234,0.339313377291367,0.1882169937157292,0.8689767635668199,
 0.7949224089334269,0.2167504043768439,0.014367933212722117,
 0.27676479055531467,0.13531660553235403,0.5044716720629567,
 0.36937638036322973,0.8785471334987964,0.23947383125017907,
 0.3600873854107244,0.6730091096378582,0.7612520545961431,0.05228281593158368,
 0.5381995745945343,0.8548286682229562,0.6999802841739666,0.4105054198657778,
 0.6152223625898131,0.3703417665905169,0.7866688579498746,0.19240940990787958,
 0.829869460852086,0.5327402826513223,0.673045563027702,0.5680318357397077,
 0.8523941271156389,0.5363443985070565,0.7416911824280338,0.3173487935828855,
 0.27163502077577395,0.9842398025987741,0.46992368821632313,
 0.5768089511050982,0.40861947123793485,0.018419600838617822,
 0.1441221228290528,0.5700208429302013,0.42326048092259305,
 0.21225089605062097,0.48192354061105713,0.9230857213648253,
 0.4600864861063557,0.07976743619999094,0.8572269466493849,0.9043348617747322,
 0.10254505120028712,0.23313521213608335,0.8231230765263493,
 0.7673198742134849,0.7729308483512375,0.2251739092879993,0.49080832549281317,
 0.22861554661534045,0.4839775118417108,0.2839433083305979,0.466091679772731,
 0.07795731997182276,0.8087535478996092,0.24840018481240955,
 0.12516784677295667,0.3677871280055993,0.6210911944336581,0.8237570906558218,
 0.14210147064641265,0.6241728927540805,0.2827000317351094,0.9532845200090072,
 0.78958688648206,0.8531234879230355,0.5926966926395407,0.4086264768713168,
 0.6263593233807928,0.9027223571212257,0.9940708795091951,
 0.006676490852258854,0.6698927560708323,0.7663054378951343,
 0.9902275304869236,0.4500834254514817,0.8406180423063819,0.7708762703718216,
 0.2286345936529588,0.9777009565672021,0.7770178972379336,0.5176468096357023,
 0.4825163851051233,0.6082184986521881,0.10335006819322334,0.8993226410194932,
 0.15763358275060269,0.47525410623077546,0.7053595932289712,
 0.7477400208134838,0.20552196320629068,0.18132543120242717,
 0.9752847562325333,0.8545192132504056,0.3087494745025062,0.08929213203618602,
 0.22849087172136073,0.9499571296713486,0.8057409873198322,0.5125642979275262,
 0.21956154447529552,0.591069101343383,0.26509761841353596,
 0.11628458823850119,0.7851409552585026,0.7068218745866117,0.5000284940517945,
 0.34459068590238906,0.39661408226761397,0.7729258137006996,
 0.23469960921991073,0.10266440279311806,0.34972166822954964,
 0.5638212958076665,0.2885426385243163,0.372272240245899,0.6363297715252448,
 0.4221674408908779,0.6608286803185623,0.7417030053164728,0.5420469226405163,
 0.7061186379710985,0.19745981736893436,0.5221648169634118,0.7501127751938688,
 0.2657357517989257,0.7762882782675095,0.6301791471103522,0.6428554648807364,
 0.6509862736150542,0.9797270152498697,0.553240953938686,0.6690560135513066,
 0.20537507429362256,0.5922481557015123,0.6136135651806824,0.9643291761805466,
 0.5699066878210194,0.705519783663686,0.5946932114482022,0.5343582677073855,
 0.4929909106648065,0.07758003526031565,0.7789721866874475,0.1763062179792506,
 0.26950614465197664,0.5604298618120017,0.6250564234273389,0.9676058202791632,
 0.4732003958203642,0.060762545116771594,0.18968550453854705,
 0.8490338022119253,0.9714228160346299,0.20381814616795269,
 0.28733750162107274,0.5609476351994107,0.575914353982845,0.7209118030015261,
 0.3223447306693269,0.8128297423025543,0.33317682590121167,0.2431664730118841,
 0.6622802909243195,0.1320603538262598,0.5930308812158437,0.2518793480954229,
 0.44823190570458493,0.7373006290898341,0.6032841748608614,
 0.22196169208511285,0.2288947616014323,0.7762738700561937,0.588554531178707,
 0.7095027495403676,0.3295729857624562,0.4533634949883394,0.5263780056586835,
 0.5149818235385252,0.20708245328801267,0.9801000430613934,0.2346927120098874,
 0.7564186150290325,0.017394265060632796,0.47382939068610663,
 0.5422986679027348,0.15246533863876444,0.647719727883993,0.0632570708862048,
 0.5844176129620039,0.06758827625237673,0.4612251686580896,0.8779508598326512,
 0.13948320360825583,0.9203443733561417,0.7402362512054401,0.7724955032424226,
 0.8194799568911348,0.9557137286954679,0.43031412547735715,0.5880310136200413,
 0.7318483921134751,0.17741387007357368,0.8531028208090221,
 0.17002070016192417,0.5302629599786195,0.5746606871418812,0.5490045259283413,
 0.5002296848538611,0.6396231895325741,0.39604188120014105,
 0.06722952830688556,0.8907212833251795,0.08934456486669207,
 0.7133576598300887,0.038809203734785136,0.44562003627235214,
 0.15463864501154578,0.9893149858928849,0.585225520703404,0.7685502127458896,
 0.3199839003864867,0.016703963368241426,0.09408390321429405,
 0.060348337816486075,0.21575685076463957,0.4136682763468724,
 0.7417762791387703,0.470515185550049,0.5643554209576059,0.012096487886985141,
 0.7348958165407473,0.4063905827315605,0.5389450329321209,0.9294544597444923,
 0.5895529061256258,0.9301780273307829,0.7710303112311037,0.6350007580330749,
 0.9134558278455999,0.17918746631803018,0.1073669138843707,
 0.46566409877281556,0.04487906305741429,0.705618008176607,
 0.05052311340614524,0.4060082906138801,0.7093592253707701,
 0.03613576193646162,0.7405018925062793,0.733030042673936,0.37986655315974527,
 0.9717705648735521,0.21516204343170986,0.5615330466654054,0.8121492969184438,
 0.48731827293999486,0.11760844477116539,0.04723532313633272,
 0.5469321679885921,0.09591369810714612,0.45261527449575556,
 0.9601230318919309,0.1528307790432355,0.32321777773588634,0.7265764552717235,
 0.7508321668529989,0.15762320386443018,0.9298111216148945,
 0.21090258892858138,0.42594292245498355,0.5602785786183849,
 0.8044511492997256,0.39042013149190935,0.8379478624813158,0.5170539417856187,
 0.49554060157904967,0.47364058017394917,0.6968661360062853,
 0.49842117996556545,0.8584533352084646,0.5652458400109559,
 0.10425756546736098,0.42041580099464904,0.1880836067358448,
 0.9231268944047689,0.1329810221458042,0.037554933479913455,
 0.3832067935790021,0.6870627078945553,0.5325268245022697,0.35407488291075695,
 0.06109624655117496,0.8555564390741052,0.39776958004007956,
 0.3626515529798542,0.4632726243725267,0.40681678038679625,0.5376093620303894,
 0.9926615745881495,0.43034240262323153,0.6739344616182913,0.5742877820737432,
 0.6127666801760902,0.23233077901350452,0.08201057901891773,
 0.8668131974456439,0.15983956729050242,0.13286996929724926,
 0.24722776418851033,0.1744144615740918,0.6802671773126548,0.4847668002080959,
 0.8097103638437128,0.5311405695454489,0.8100374639006296,0.8036145156493839,
 0.749968466956703,0.021867905508154095,0.8136669985332456,0.9482643427193982,
 0.9726089733306095,0.7973235692967713,0.4213357283147108,0.6271722165750823,
 0.3712969776592059,0.4322758329107206,0.1183819979360844,0.7559224121105631,
 0.544247919154199,0.23014513678933557,0.23468939808312017,0.3538986042980885,
 0.9078907264796651,0.5982477156136815,0.7486951909450685,0.2527418064286776,
 0.28315115879318564,0.6327927596944946,0.8653957888340035,0.9689883479707687,
 0.9169359265841457,0.47026402818640856,0.3371301560264184,
 0.29939749385021086,0.26472618159383066,0.45643738446422977,
 0.48816688987821477,0.6707859839428668,0.2543254651077863,0.9194354500208174,
 0.8385221141525825,0.027281759675280126,0.19763242208442822,
 0.9030215755332679,0.6341962100991567,0.6481509452530979,0.36778258332129365,
 0.43521265044356616,0.09528062814398552,0.20113323553757068,
 0.7166060112154047,0.5520294362038792,0.13449534950320086,
 0.023711244450266178,0.34611022352293486,0.7008476349114028,
 0.6196328668196723,0.7060883875878502,0.14185022811616976,0.4228418545520032,
 0.313111456442392,0.6687039220752174,0.6179651081048143,0.27142065593235665,
 0.10218720414343407,0.9401515142424088,0.7251291495403129,
 0.36937376011218637,0.12400345959976489,0.21430635465523595,
 0.9832996025791964,0.5173835881135311,0.09963426164864841,
 0.44680072304120366,0.03738444672438945,0.49013897959896413,
 0.8197222207936721,0.17866378297892582,0.991186703348682,0.7204128517612398,
 0.02910359851941635,0.47820581767921055,0.19855034691235995,
 0.20906662809361032,0.5169907448257677,0.678662817113769,0.8535823591360759,
 0.28348982853656446,0.5500207403976802,0.814220376646662,0.16558614676852867,
 0.06465013472352565,0.22089446946052937,0.35362705331318256,
 0.8598540649288385,0.7489808691750761,0.6604270046294828,0.4627025681324739,
 0.332930692788467,0.48915601409300113,0.31393629341251916,0.6148094134674218,
 0.15294267028748842,0.6808602956881357,0.27041548048527353,
 0.3634934417704676,0.5936925889418733,0.16931986938414534,0.643309687177581,
 0.8234463262232159,0.1869767363290965,0.8016337850598483,0.7747740882212418,
 0.11242563010692386,0.49386116302266325,0.3956879831356579,0.686149038051766,
 0.6703224849860985,0.9819231686277731,0.7865345021497635,0.5877986733762123,
 0.12437045638690314,0.4126074785676218,0.7031696483842026,0.594186973501856,
 0.977341639176061,0.25682942166526634,0.8640920264919126,0.7116293712258931,
 0.6212929446936328,0.29096623323438164,0.8190250730426512,0.5399781335168773,
 0.6098112751606338,0.03410879532970679,0.5225101339668395,
 0.17362665831029556,0.4504026995433249,0.5957844301417146,0.6613190819564081,
 0.4879797175811099,0.3062552244690181,0.2758265194042415,0.5441618761875169,
 0.6406237903079821,0.8409274639482348,0.9809674291998894,0.632278895758365,
 0.4325674984582304,0.6165256678636685,0.029049383476827728,
 0.8363257414383966,0.3683337264107637,0.9808375331547455,0.4176875400762218,
 0.2509945101173977,0.1073242873338518,0.6610469493041101,0.20954566899939175,
 0.7707966347333125,0.47017956713398124,0.9241087508627028,
 0.09815449394312958,0.21208067012772114,0.5101561187723473,
 0.9987469321094338,0.14706598485076128,0.056725630166802765,
 0.5021782997543118,0.7309495950416826,0.06046304053928653,0.6589979354143621,
 0.5953786328063924,0.1304960399061369,0.1276810236453807,0.7538972069146753,
 0.6699701208853771,0.3341186864327481,0.37960007148230224,
 0.13861462955732862,0.1790009977762732,0.4381252539984203,0.2734267374341921,
 0.32708671912011333,0.4504183310423847,0.6463124126229457,
 0.40435688225675404,0.7865399368248402,0.19439654691754482,
 0.9799485416347516,0.22925901504048696,0.39532843573837595,
 0.17386389359700838,0.5838169264972746,0.30369187808851184,
 0.7748320346475761,0.8569020401064558,0.8930000117477216,
 0.0024591071514320495,0.4481471607091535,0.29322552176373295,
 0.9701639825340206,0.39125802275445165,0.9412160557443314,0.2715302495890347,
 0.689496963824536,0.9579890179981044,0.15591704095026948,0.8604165003207165,
 0.11213302988463092,0.29121696890779136,0.006309861083207191,
 0.7437041042058966,0.5020497633612125,0.3609161723575599,0.5208449915320275,
 0.11702919765138631,0.6481992419818929,0.43985162360383834,
 0.34778489793033185,0.2349870835017399,0.6908585839403042,0.8391172954687279,
 0.8063729110123036,0.25775345331610644,0.8636668744199195,0.2374451975509707,
 0.30622504272991946,0.03982550276996155,0.5856632190117605,
 0.034412434758677835,0.0658394846078908,0.4111485783965839,
 0.5425262673184669,0.8200346086320082,0.6860325632007671,0.7966798729275637,
 0.9285956367057397,0.10539380139184651,0.22011266022060505,0.689496736406594,
 0.3904282509790894,0.23598453356035765,0.7376224409292746,0.3509437819625125,
 0.4624724243029653,0.9816272792169356,0.25557017294814266,
 0.44005687639301905,0.5041671958059349,0.6598453539652425,
 0.049847158246733914,0.8231675826976383,0.9679669376190125,
 0.39308596330849144,0.719019730804801,0.8430711852951068,0.09373336413590772,
 0.06847122691672736,0.5268764032579071,0.6168316889275388,0.8848740008407487,
 0.39561414220395696,0.7116881331650287,0.5388397572442788,
 0.20214996873809543,0.13114937350845046,0.36573676496678,0.47471657808473555,
 0.10740386326913653,0.03590835086225996,0.8182591156486836,0.846137591271738,
 0.0776474491618766,0.43145880802682013,0.0563390028407279,0.6170813268650895,
 0.6308289912034363,0.20446785373756005,0.8010261014622297,0.7998887461716621,
 0.08729106288220545,0.003969283671549917,0.5801532881815212,
 0.4355784162179239,0.571359224875051,0.5733649310166451,0.635586006581143,
 0.33337541024968953,0.4576290344348507,0.8228339545962513,0.9181103836724166,
 0.47598927976412797,0.7670998792684556,0.3985692849506456,0.8844480140668171,
 0.6387124705625518,0.7726134377535978,0.4980760007280024,0.28271613959416775,
 0.6702094163724956,0.08122944551109523,0.24376218307462838,
 0.33344624794830935,0.5011871506276264,0.17907719803749966,
 0.002297699122432051,0.18648649261023242,0.6837810911700521,
 0.4105003049168199,0.9452864667186259,0.19831693468465084,
 0.41829985612419074,0.6826458504819432,0.8046104085739776,0.7355267805446145,
 0.6773572479781546,0.9447829420245686,0.757296197137404,0.7653280699990866,
 0.2322241952538162,0.23600022801673448,0.11391535667916186,
 0.2757331219331547,0.19533838962356576,0.8621639432959223,0.9426259903241923,
 0.19142293687736966,0.5688125990541223,0.8881550868358048,0.4311261657604193,
 0.8274069356682912,0.19598280669852808,0.7113614321284661,0.6966322960409305,
 0.5041770175396949,0.8048587544079225,0.205957167877459,0.8683556482008423,
 0.48150339847545387,0.21077249012751897,0.36162192882653355,
 0.04900549108354446,0.207099043913343,0.16683897701504813,0.4376101438202057,
 0.5532433612762778,0.44935396958168417,0.759650549939676,0.21813137169268015,
 0.25847824302947975,0.5841849862900813,0.7313807059586969,0.6955941709221793,
 0.28773660137606094,0.3661229620488142,0.05812739082635976,
 0.04635095000261957,0.9495011493819367,0.3507281287502245,0.6196706040939975,
 0.7128907749816229,0.6987439726107315,0.7088149111064426,0.26523195412746203,
 0.9369279759845033,0.4240013874208326,0.29392364970632123,0.448577728062609,
 0.07793796531574881,0.6052894597552374,0.7306164837676742,
 0.05834588449142275,0.9626774481156872,0.8146837838511862,0.6722584231984273,
 0.556149468394891,0.7680032162619057,0.7750394902742244,0.654632690589869,
 0.6736234486124827,0.78507630441971,0.3029348367059602,0.39431773162707595,
 0.8760215035640386,0.8321216185696342,0.9957569229891495,
 0.031845398965456795,0.15151176151733792,0.9394628006749385,
 0.11522185047991718,0.5297372700196878,0.5543213868610162,0.6709299271109446,
 0.7706104791795668,0.5844660192670135,0.22675031810593727,0.7961639594504442,
 0.4773684254297841,0.7694357664266042,0.6685325848692806,0.5195741529787468,
 0.4757946382615713,0.5599988228739833,0.15567786410176632,0.7769876398275295,
 0.149858474733502,0.42958132641864377,0.3926052108249428,0.995119528999465,
 0.19272674627035236,0.6025576267170399,0.21229769064512916,
 0.7973130069280479,0.5302399430064534,0.08416413615983398,
 0.22311320674623902,0.5151357755100506,0.9696847706667722,
 0.19655772722882126,0.705887829098833,0.16082933167364954,0.9300311391635863,
 0.3754097748044135,0.624489225673722,0.20338624230090407,0.3755869142560203,
 0.9610857107481119,0.6381830288968933,0.4069610151001255,0.6773077489859246,
 0.6676042430059925,0.4118845513035013,0.3062037742531998,0.058556122633176,
 0.502994430792452,0.9534887652429554,0.8364936675856708,0.5449889792748385,
 0.8327289374648714,0.8919930282689725,0.39719487684480237,
 0.22426180248393834,0.3603445911525722,0.46085261168333314,
 0.5983643561089069,0.4588238215205962,0.09808776039312161,
 0.36827393635576156,0.7856757445901689,0.2453311901531542,0.234133367993128,
 0.38879298559181485,0.03027098252430971,0.918696624326246,0.5153016295980741,
 0.8546662781630678,0.8132919024296666,0.14130044749580684,0.1865741105140054,
 0.6883664561167159,0.8702684941018443,0.4794326618576039,0.5507332608473396,
 0.7084102709852201,0.7107664302371806,0.6875084789325456,
 0.0040075830079085994,0.3136448696743692,0.5977482937316825,
 0.09870129926957028,0.37283885041384623,0.5381238762854436,
 0.020979065258719953,0.3146976508856013,0.5580778972756183,
 0.26243701843566214,0.6360629853111186,0.11238942691808418,
 0.20752323567166275,0.5139868231450583,0.20296686617125725,
 0.23548574668676459,0.7050891367494624,0.1751660515421707,
 0.04992141118270099,0.869842241887862,0.002921265027214459,
 0.5442137499219113,0.7897781073612403,0.9817174079836295,0.9072309032965378,
 0.667331955165978,0.14888509409264306,0.36935739542184864,0.1390717880923138,
 0.15341585616574205,0.9890479238389538,0.6012295991531301,
 0.028898643881274833,0.8870378907512121,0.46659823307506765,
 0.9034931335697945,0.09482998425552014,0.3815934704902938,0.7519462769665579,
 0.44799359618214296,0.38056841685540654,0.37380831584001617,
 0.830037530554873,0.758021654965838,0.41150090521690363,0.9014736445045264,
 0.19369734725533816,0.1387532975274075,0.04753575606714122,
 0.26317339807368734,0.2719766554044979,0.49666169264196747,
 0.8599601906303597,0.6379105526282547,0.016628219810803868,
 0.46044227713341335,0.010441454795737437,0.1282825563532859,
 0.9352684705646849,0.6890128024274247,0.3583805169537193,0.27508908601403614,
 0.13816070639814504,0.997848040211083,0.9612826474770959,0.03646897896238532,
 0.6187353500333782,0.635469683751156,0.3361848455125912,0.21172344218832495,
 0.057067058598913034,0.9355882325752858,0.7954657056300184,
 0.10991353758396893,0.26262525969861383,0.6048197683991907,
 0.019879897981636918,0.47527685152910437,0.6136911068293928,
 0.651974378771406,0.45618538777133066,0.39780843757911444,0.7561143637444657,
 0.2585356993667409,0.6241551013278017,0.2948810089526407,0.34193865311150196,
 0.27640550668450603,0.26754245195970194,0.31492980825376016,
 0.817162287280104,0.8956451752805472,0.3135619142765155,0.9141805912092138,
 0.8878368344666814,0.49394159916187097,0.8574434002366706,
 0.32144521959461736,0.24360140250418272,0.267712309236125,0.9566349202680524,
 0.6005224050208734,0.8733080303185968,0.02815700181284031,0.7008188849325858,
 0.06522220019393365,0.7410777471220024,0.902544191774606,
 0.036577277755821425,0.9682215904115303,0.17215025456425093,
 0.7028306215432608,0.4506191913989175,0.8893156676593295,0.22377202124493245,
 0.8298448323464269,0.053627574758050356,0.3749986228774612,
 0.8975224629291707,0.936975720772517,0.5092503609131284,0.09077947383781704,
 0.5253046433887354,0.35237639234523543,0.3264297304687327,
 0.06061069583960044,0.6820674587430268,0.07652622258347985,
 0.37750550235893643,0.4303670104837162,0.1450930535583257,0.9041364551801812,
 0.30867848656458774,0.2955553688640671,0.2535112338846618,0.7534259111776296,
 0.18488126408352212,0.5662908608539438,0.5099936640058172,0.5898822664350368,
 0.8725637926067005,0.09629533572529791,0.13358794420649722,
 0.5276967294467608,0.24275437938825029,0.991428372672851,0.6959761900784878,
 0.46652005038004085,0.8799780416262829,0.07976235225195694,
 0.43332272862496046,0.20824071433702296,0.691832275290972,0.8762882100457094,
 0.713918634228939,0.049766608837737536,0.35414414714260334,
 0.6346824560106257,0.1128959163685459,0.4801025617660467,0.5769827932881342,
 0.9466241188616781,0.5793066241584061,0.42255215137489377,0.9594835895112536,
 0.6042459253255491,0.3005514061665895,0.37124838529244675,0.7245600947617368,
 0.9816779383175582,0.2570501296176648,0.5048788529359917,0.11229603442129621,
 0.6194961191992892,0.7145157534035346,0.7222097500472907,0.9895036726086291,
 0.03682178493057209,0.33402977119802646,0.6462983514488929,
 0.08032915162565102,0.16450836375465983,0.05796663434257876,
 0.21972329008533287,0.2217056875679343,0.7141046331958402,0.9548726909196433,
 0.28779578203429934,0.1578764003387617,0.16924347580221832,
 0.44176433245785696,0.2999893926289172,0.062285633206468516,
 0.6739637338798318,0.43264879846888615,0.5073307320309142,
 0.19817423789535005,0.02619046846952955,0.9005642308523253,
 0.0019064345824086892,0.16594377130841242,0.9250292385998747,
 0.5911782861082282,0.2682333572363269,0.5460306251800848,0.29480054120041344,
 0.40310922727494847,0.32977971833475683,0.34839385889235,0.502818039126012,
 0.5632712304238807,0.664267039587804,0.6351285402483994,0.5541767700989321,
 0.7791603488484677,0.9011823927833278,0.8811029341002103,0.8028784667421277,
 0.621927514228816,0.34956746612846634,0.46942825378913355,0.7600793370623818,
 0.035491365643284145,0.28982319795811196,0.8890629684977418,
 0.287829080680184,0.44942712752289526,0.8855132341456403,0.7445653048989896,
 0.6189241975281472,0.799498499787533,0.9051948427802622,0.8200323164691544,
 0.46524755753362135,0.24020302834261142,0.9156219119614583,
 0.11748330201240453,0.6226377117988483,0.6297835696792111,0.7447233338001814,
 0.9314060340906656,0.5512583261572105,0.5876396834950203,0.49034020415730173,
 0.5325842646570268,0.178409504055014,0.548324357353333,0.6283835224556974,
 0.5217676189738473,0.9172403162375169,0.3128484608251033,0.2107503899536166,
 0.04729206027284416,0.5617687475231599,0.7656692880459666,0.9234176840900731,
 0.30970571947198344,0.9220692313084969,0.44069124182331065,
 0.7289629187985753,0.166835506449739,0.15997009067250878,0.3389056583387474,
 0.28829576571945725,0.9049707602623924,0.3919021012950963,
 0.018587843691888928,0.6050603377459518,0.6350786357614284,
 0.7458707353358054,0.5650509809032552,0.32838282246617523,0.7219904536043709,
 0.07712917117152629,0.9120216141245296,0.9441622775626087,0.4966093329084853,
 0.5896365164814249,0.18534719865145466,0.803239238659594,0.9529676839102669,
 0.5619568230157967,0.523253857310013,0.7238729676875398,0.9436192793879317,
 0.6903010806334785,0.32755964214413047,0.8857475596963253,0.8267009149862761,
 0.4343716025037354,0.05032249227772012,0.16312389384496795,
 0.8841771558266808,0.5550242251316504,0.6793682397088369,0.21376833102794301,
 0.26864316452821424,0.6365255333973312,0.28753668466747995,
 0.9842518064945371,0.05859785064133227,0.8030471643583643,0.3470426758117138,
 0.24521607862465444,0.4243876554780007,0.3384802028199212,
 0.20775715737815914,0.8786761755100132,0.22483339066093155,
 0.34368675130973403,0.9962052558974029,0.01397155856869936,
 0.9538207321602677,0.6908084867321203,0.8783014746992451,0.628583062581343,
 0.3009822392182494,0.9801113064568008,0.29973356611331803,0.8606095238887721,
 0.6543856735220163,0.09507932259835239,0.4067954989656668,0.5844681620400582,
 0.5383080437247784,0.42217056353103743,0.6598033630755774,0.515692321472794,
 0.026955692536981046,0.815830056212631,0.15709990514693295,
 0.6732578633810172,0.864721585925951,0.6470548279942502,0.8774086682953952,
 0.8705669772747058,0.6637317515152684,0.22145533472792134,0.6262924646692426,
 0.007316036776346624,0.8517327733032326,0.1036772112451746,
 0.09692955214579757,0.929983597878627,0.3840718885285199,0.20492265292149914,
 0.9966414517821353,0.4999742361100973,0.061892824589141804,
 0.3943541675741149,0.2646616359331283,0.6344402602986916,0.01047221048777125,
 0.7574733818418943,0.8182155855847354,0.8380738184211143,0.4337988829497519,
 0.6758538657591832,0.9033906247276662,0.7564320281606072,0.7982468320228713,
 0.6354208023454151,0.38599125888893226,0.05880850838018292,
 0.7858754337685946,0.8615647591280046,0.6040286630494596,0.28483604405646923,
 0.827141086776495,0.43010074440037316,0.1148710737114077,0.43720065080452164,
 0.9516841202848632,0.7149944550061154,0.3135095561965583,0.211878074777892,
 0.6210174849846832,0.1663189027623968,0.17059446760075325,0.3626118411021474,
 0.9509077308318077,0.27431164924481033,0.9215697373752278,0.5693650799501633,
 0.8284232916719724,0.5360113829334483,0.7296002580244789,0.2520141976801392,
 0.9719958529712027,0.2959178792159365,0.7405625244830558,0.2504917885517506,
 0.003619930957130424,0.6393012544164539,0.3020447967058134,
 0.6565500400380212,0.43506706023293484,0.6077006235990455,0.1670360903826934,
 0.7210658050631419,0.33330838021694265,0.19968864191968339,
 0.07473955587744119,0.3658453559062308,0.5828518241615592,0.9084306521781981,
 0.8203398188677307,0.8921268297061976,0.6053825259867447,0.8419453441443377,
 0.16397970993937228,0.26159938288682916,0.5796915722596054,
 0.9921067094632964,0.9470506005853901,0.04515051592257624,
 0.23092929131447648,0.6370914947503133,0.08200637323394833,
 0.6404328898675589,0.08816302664412767,0.3399693153798983,0.5858267662456141,
 0.2577355159025492,0.8664609486184006,0.1160308291294383,0.3815811020341935,
 0.24893615813114378,0.1952163208067973,0.7523363918756515,
 0.008511839900312457,0.31438512552136044,0.07613662066817073,
 0.010462890851537288,0.39601558880123466,0.16358510988834662,
 0.041035860030022575,0.25337862535257394,0.9255466812046531,
 0.9932405481524017,0.8628014057830272,0.6453931141915826,0.7061822170790788,
 0.3265579690817203,0.9571713379796336,0.6914747258185322,0.9106513380493804,
 0.8776310274211848,0.7999622328973433,0.3934670020398392,0.3816411665618582,
 0.4847574356363926,0.7317310069698688,0.06658462212123417,0.5112580067482465,
 0.2190004357227846,0.08487806971984058,0.8923942240321654,0.6050489551753135,
 0.5414876440562111,0.2796676664450385,0.2168570130156694,0.6984802928954879,
 0.823832887299258,0.5696309847012034,0.25629313172862245,0.17195643577743824,
 0.002211004918760029,0.6310270218023963,0.7481270901957066,
 0.40081792924749404,0.2603767095688756,0.8456712537787631,0.6664794288013313,
 0.04674891019222738,0.568669711017048,0.41463739671589117,0.5756988272642838,
 0.6116940363705261,0.0470246337814032,0.6902209723569189,0.4118046294744714,
 0.5436600622278633,0.09063270031602699,0.32480872596789157,
 0.03412363839463306,0.5495341269888762,0.9808822303743671,0.7354241508656164,
 0.4339891569390225,0.5330557093619865,0.46308322362815846,0.6582922571633439,
 0.36846713821298216,0.43448779440186924,0.024265425179286093,
 0.09445345841980668,0.21238408355226923,0.5435076451297046,
 0.4981669906186613,0.6163037828078008,0.909583264132691,0.056037101394745825,
 0.9759467660468001,0.48439819135573936,0.14729098817657538,
 0.035217576076531865,0.26632754151338867,0.7871465231527528,
 0.858771544519161,0.6015571801012322,0.8083720714955324,0.7941751924406856,
 0.5569192890631027,0.6826279958478498,0.9606109192043504,0.501481230447227,
 0.45517328503895205,0.7856591860054807,0.59821053254264,0.015287499930717985,
 0.38407133448553277,0.17230600743990487,0.40793253404918983,
 0.4057262421811215,0.8542598137874462,0.2843894587434055,0.8151598058253751,
 0.33155748768914806,0.34461454354974697,0.9991134380074,0.7699876854773837,
 0.929579428822763,0.4338745990317958,0.11354631644410385,0.5194930650124114,
 0.03152806483369963,0.7352227956345432,0.7528233243902688,
 0.24015807236980935,0.870084088437341,0.46221558206281343,0.7849401286223805,
 0.07609085088231282,0.8082541002284167,0.5166137665872531,0.6876661228317191,
 0.38404018953277164,0.1175792584392713,0.4295911851319931,
 0.17440906898491837,0.9409821629954738,0.08674897149216176,
 0.6013199169884764,0.6949160525233572,0.3345338081395076,0.5005623770989422,
 0.5253840378408757,0.7207885739038666,0.2864449689994717,0.5709494475575018,
 0.022828842595966736,0.4071086198762822,0.6736715438089222,
 0.20103401812451616,0.5023493948890629,0.9593024384369409,
 0.19054740420079508,0.2226676476272883,0.6463787968769298,0.5997063526145127,
 0.8224911993414857,0.6373362060566189,0.680785825130673,0.8253387482411303,
 0.8513942849035181,0.9529760335814219,0.056086704030709034,
 0.17311595769775479,0.2930214828211394,0.6423418704522382,0.1887919407659039,
 0.8324703681955612,0.3138460776120835,0.6196740788198457,0.4707389820201122,
 0.011607240761467796,0.6856898926185561,0.9609057257662889,
 0.3497992630917486,0.8809384483078011,0.9208253685531094,0.890907830412882,
 0.7641126982599129,0.8099064267613786,0.677280126384256,0.773870344519151,
 0.4663942771791103,0.6693149569595719,0.058383008633607125,
 0.01309174649924727,0.37944043126593063,0.9451524289184893,
 0.7667971589063141,0.9327299897419962,0.435655834383048,0.7032206147176607,
 0.21137296447669807,0.7031002421626807,0.16092126324348732,
 0.6858891870986072,0.3280648377237496,0.16504681655940945,0.915612973508243,
 0.8269006181646812,0.8860871508121632,0.05354759333342041,
 0.29639844798951276,0.8064705478468536,0.9214989016545874,0.6059640006635987,
 0.2928120081798675,0.33764949722684734,0.04116553922187516,
 0.9760676184853623,0.39467722883586653,0.5285006731975267,0.2536755968530615,
 0.7942872882835469,0.905354553050947,0.5566600366945982,0.0323230750706156,
 0.7806827578287656,0.4292616055969696,0.7392458930131794,0.49748547296307455,
 0.8943196794119865,0.7366032893683583,0.8389034496071737,0.0198419346755615,
 0.10801136919186138,0.06594607777828276,0.08706561824507386,
 0.8683791564147064,0.9115626581430036,0.4421589633419021,0.14630931638277944,
 0.8652139255125909,0.8865620298654957,0.644305863162142,0.9479910651929973,
 0.5849442419961758,0.13660364644437895,0.4258401286412098,
 0.06378827903312612,0.7687698732429421,0.09396353604091612,
 0.9860466589668434,0.4567262742893571,0.4345106466511437,0.5196327994020041,
 0.6172228280797644,0.48367034224968686,0.0920568964444084,
 0.12517800402576829,0.42342032838131904,0.8069177282751351,
 0.30827465587967207,0.34728629195611016,0.10505153635272768,
 0.3532340412428424,0.09827147488519139,0.28102120625919036,
 0.1773137111634816,0.07691691911890486,0.053646842919861326,
 0.9734481407452453,0.6368134798101974,0.1966949279928758,0.21043758953332492,
 0.9065859524412159,0.9614938429070221,0.7567087134421058,0.750942463366685,
 0.7497722569753633,0.07674752783916272,0.370653421144573,0.06786254350492305,
 0.2913602174471738,0.5279304787626274,0.3653746628078782,0.14491641422040524,
 0.7669632439554904,0.5184986803419984,0.7500578543397632,0.5584749501929729,
 0.9355258242297024,0.3709015879808173,0.8886791358486836,0.920141977218643,
 0.94064190633735,0.2640511026749841,0.25746938730707225,0.6697915814121951,
 0.7995336119325938,0.5940933441881701,0.4993984845680578,0.412761826186108,
 0.09254987062413644,0.4143989264375706,0.22222870372367287,
 0.015672714088259765,0.9176107408586957,0.4780537650997153,
 0.9592907125184115,0.19436868777175542,0.7140328578212433,0.6770984135183273,
 0.44655868052631564,0.9379972135272555,0.9585778311429062,0.7275024484539029,
 0.682300864802853,0.4664714529272974,0.6582551173142898,0.8579669227576163,
 0.9551532300849213,0.1457279507816216,0.4998893905273971,0.768877643468276,
 0.6520724131127389,0.7972644536167768,0.3624919582007977,0.10506274565235696,
 0.6354145513563928,0.13732678100171092,0.2876345340318224,0.8642850730243056,
 0.24756648240139234,0.0061140421049780125,0.32985552823832265,
 0.22823793713658835,0.09911977586600595,0.9327715232280602,
 0.9882310006530813,0.9253319506529181,0.6785440627385267,0.9070967420531956,
 0.22542296343138857,0.10656020799601451,0.4617582329773098,
 0.06319414443053051,0.6106256683340828,0.39398658893164984,
 0.7544550034188284,0.09079625801486191,0.23002608208626651,0.745942947350424,
 0.7691983738946916,0.4247933065113959,0.7741835378951125,0.04439369656051817,
 0.2571842794734135,0.5110508288149023,0.23927140579855877,
 0.02874022013635502,0.48896511160600054,0.303069067736232,0.7200563312022314,
 0.5731909994843883,0.21963613918741764,0.669299990895179,0.9764559065082166,
 0.9463441421595564,0.8839379711839902,0.9424502869428808,0.6582358534797195,
 0.11679710127509824,0.14236461022125468,0.6162433263022163,
 0.6053357029372282,0.15952590847061754,0.11873391738268235,
 0.26544337744172963,0.08004726694559816,0.7718422287883046,
 0.3911111113608202,0.3407777594817679,0.9103663734554376,0.20054079997746022,
 0.6295420224808583,0.22811984594404422,0.6681594637437128,0.9277171231098171,
 0.951948751497169,0.33462688317028044,0.4998312819070647,0.4366747977618147,
 0.19366095403443995,0.21061040743595605,0.7759802900200704,
 0.4524005226938419,0.8600045710759763,0.4783107428784641,0.587065353454442,
 0.4277611288235026,0.4995282623374355,0.7000583359806908,0.4377983535562858,
 0.1648967319918,0.9978345606041236,0.9512281249412564,0.07666375425770822,
 0.5696434787865297,0.4397621087262712,0.6075394496842126,0.8939806410752751,
 0.5871631182476817,0.9429528363576406,0.655974548002373,0.38128550051965104,
 0.8305781763852074,0.9059835759309173,0.31501101424902733,0.7360069272101049,
 0.21582837371213626,0.1312262354721263,0.9319282164706976,0.6157740104326365,
 0.49455455512632673,0.7872571606485756,0.18238131037377192,
 0.07966740044626419,0.2091658568886512,0.6098685581887925,
 0.059439393333929225,0.4027273291107475,0.26230279783231414,
 0.00413788306553553,0.5046280862866047,0.9822267926129542,
 0.026726115663759153,0.8010800030520762,0.25344229007535446,
 0.16823091183306982,0.23259026965819396,0.03633570642405137,
 0.7059096932542361,0.7919300680624423,0.016058039811124747,
 0.6312629625611076,0.8157032590191796,0.8245032599785056,0.3355999201642743,
 0.5566558701750561,0.5777955876904994,0.7875090156179595,0.6770425614142362,
 0.28870872837971395,0.2603129637163838,0.6050560673843903,0.8597608392290186,
 0.46336019304410936,0.3090990288779393,0.7765244258364062,0.7607451058951354,
 0.4202427984295102,0.8812582941500834,0.03205296760822607,0.0715113140756235,
 0.3669830339970076,0.7030754467505842,0.3322153441378195,0.4119137603769728,
 0.8962086449433391,0.5676999946982162,0.09198713801968605,0.1947544361243777,
 0.14516778444701572,0.0666460381670313,0.44320626170258726,0.963702972517656,
 0.4213267409428172,0.4308528777432856,0.5921915445165549,0.3203986340526629,
 0.5442597508681952,0.5493585822448269,0.4421563132051577,0.8145994005467124,
 0.4405746474863207,0.6413346177742028,0.4754933179443396,0.7437262655405157,
 0.5121258130092037,0.8839164508979108,0.050899650537770125,
 0.5920429321295605,0.9926011417231113,0.1746238194321167,0.30365558772692336,
 0.48805673398247806,0.8017707417708161,0.9313867813555037,
 0.06253699175697292,0.06917581113861138,0.2714579602124394,
 0.8224582580285646,0.11510564403708867,0.5803300834629832,0.6517697691259553,
 0.13254353369691418,0.6175798936673819,0.7982852740487554,0.9100763253959027,
 0.1611392956363521,0.9619445094440764,0.36419208614168996,0.8804309367026053,
 0.2528651095660337,0.3600094052807239,0.27640433577672363,0.6210806763375579,
 0.27250401068564867,0.15005919087794795,0.15847658156652955,
 0.7512180844922867,0.7829391861038906,0.325059312785003,0.010217049741219841,
 0.5920837909759601,0.4771498555486401,0.64108542882401,0.004536930011348694,
 0.027184747253143238,0.09872425517099537,0.17356162510398032,
 0.8457423865347021,0.012559792676893333,0.8528320714566904,0.269399400173262,
 0.28131990837164655,0.9649746116283341,0.4080264977597923,0.7024590195052615,
 0.14917620228569473,0.82588113033181,0.049885686144913954,
 0.09856232057260261,0.6171577163288997,0.6378798684417492,0.6463358797456529,
 0.2860891913744972,0.08342105376415021,0.11076436350706687,
 0.027408818368821697,0.9017729360268398,0.44968869846087856,
 0.2690768069113916,0.8684487262171667,0.5531830447116144,0.4263919594822976,
 0.8555706027256846,0.33921081471304704,0.38246426796645927,
 0.4842445787572819,0.6253062472773436,0.49143272354937806,0.2714911987071137,
 0.8806041481480106,0.7430152465761156,0.20450400629248444,
 0.24457746561302918,0.5346150832780996,0.4402687141693443,0.5527425593468419,
 0.26370076194307823,0.23998679501834697,0.7340264736883579,
 0.19275901188925115,0.9034120776256842,0.6847464862323004,0.7374938795857929,
 0.9706870992162755,0.07912172021570762,0.41905996324149464,
 0.7696649157600205,0.8662969857106307,0.7726303820167606,0.05579915329956742,
 0.3014330481887081,0.9001509771633031,0.36828097119880554,0.9876023961001306,
 0.8298458580302918,0.9638464688963021,0.3042597385795607,0.3552133970275273,
 0.4536049480354747,0.13459891667864043,0.21614258463535163,
 0.35434682918804494,0.03763574415945492,0.07770605996315783,
 0.9253632872916397,0.5348686377313507,0.895713059878314,0.836030694132067,
 0.8629395544733252,0.016419620835730897,0.037240293862317975,
 0.2672956294887886,0.77163943165216,0.27213142303773963,0.4709887591011638,
 0.7226877774616886,0.6403471925329252,0.6193360674415301,0.6673909489671466,
 0.7937813544283758,0.6754089737901776,0.4572753981361106,0.7922514487097291,
 0.446597452909376,0.5410814037384067,0.16139648612338287,0.03320956460010793,
 0.25202621673118586,0.2523106702777167,0.7939503165460449,0.8204285331073151,
 0.9947319825972765,0.9785352641629139,0.8765840931264203,0.17985066552790252,
 0.7351776159893447,0.2929385375625402,0.5359681738550707,0.9302806412279859,
 0.7729934310008675,0.4992376387834302,0.20571113805625751,
 0.07451916291240113,0.6593980791978683,0.5581929307352786,0.5217535375853085,
 0.42266298814281666,0.3578652018369799,0.9592093721067994,
 0.41311312510630616,0.9040144704856811,0.0118399701633074,
 0.0026062846526938577,0.5951998475117528,0.48105719295908955,
 0.012379807944168486,0.66205873249767,0.6398013458756064,0.25141324048785896,
 0.9624876639271605,0.38534797562631473,0.8149383250515505,0.9845777962991547,
 0.34874721576236,0.43872959908635156,0.9433406691655245,0.5364683342253811,
 0.46020786647583345,0.5830021822884949,0.10538438423948004,
 0.17984777124145934,0.007196349456648932,0.1525079802449455,
 0.19268124892775051,0.9344715206203986,0.03143053972132093,
 0.17221783690324255,0.5287872540205885,0.9149890860934703,0.3407259434429818,
 0.19685947420187155,0.3615806017414749,0.18841067331435246,
 0.37781805904040977,0.48295001968712126,0.7434165713150394,0.633842260724399,
 1.9077352633556366e-4,0.35951537298719183,0.29793384054711547,
 0.715479044563065,0.45227763324399584,0.9030791896199305,0.23830013190136912,
 0.38037518200681075,0.046826885052234424,0.08112547297058326,
 0.4843632840747478,0.5526087014831039,0.47660219665921233,0.9986518682073974,
 0.19092245685054143,0.9526569918358774,0.04870542658193078,
 0.004723681093467991,0.5240396281662217,0.3605083191116505,
 0.9744068066198094,0.6742145942835439,0.295414078742145,0.1591656159885353,
 0.9816330220317855,0.3838644156474623,0.21118992508226309,
 0.19605209509147126,0.536977313983459,0.6446565240323869,0.9378397325381278,
 0.048463840353459964,0.9144245655683392,0.5812600037603747,
 0.8052775608832906,0.8289103432328901,0.9748903832423617,0.6587832818004649,
 0.17747874465595181,0.847471307973975,0.700541882632053,0.6195726741911999,
 0.6082920046778849,0.8088785901631402,0.5281495615535671,0.05615606890394931,
 0.14586099265684482,0.9436872705770982,0.47868679519745894,
 0.03896171324641451,0.47996199486398594,0.7243215767097998,
 0.8454787398907334,0.6053771772276432,0.34921117138833124,0.5388582548451586,
 0.05203917024032323,0.24520613416124026,0.24090114235760873,
 0.5672611596377215,0.20059119344171839,0.5853302523839125,
 0.17669130239816033,0.5804740609050293,0.3579076361518243,0.7789725708641481,
 0.8025198128549473,0.270191034323206,0.5609670308959728,0.18328915050671024,
 0.19364919748246068,0.9822337103420581,0.5996086592212039,0.4447723845211756,
 0.8244434507293441,0.6669159151181642,0.630390942626667,0.3964828595382668,
 0.08060864256870148,0.9967746799813206,0.7536718803463228,0.7409276903237433,
 0.5566053133651846,0.5023583231163566,0.47867690679668684,0.4146813378745152,
 0.9301069785008863,0.7346150426050814,0.11110938148109362,
 0.043011002581352376,0.5219102701798801,0.32154735345395935,
 0.014149039638039485,0.4225579802058994,0.17727000695602912,
 0.8782669691157612,0.43975907329611585,0.45171967289516135,0.694482944194095,
 0.4291000606663884,0.18528589103086546,0.577233760752733,0.3514738117095464,
 0.6668061099261415,0.47871114942394977,0.22766045164804882,
 0.47573733258289763,0.7940454297732136,0.2471606660823169,
 0.49504844878650667,0.6664751765520689,0.8832960665549758,0.2658781538572079,
 0.36427823305533524,0.5510884597731994,0.2529620920995066,
 0.27711114728470854,0.969065628330132,0.9446107128032055,0.7424931068321418,
 0.21481327688595053,0.406758268672051,0.8468171742020515,
 0.029458187429138527,0.8676353961274259,0.6789498052303358,
 0.003060889536585454,0.8613875238796643,0.4193187306618559,
 0.9552382460244127,0.9262738536442636,0.09285237722702944,0.0644841319692595,
 0.4561715081548495,0.3114927048960532,0.07328983062121686,0.2975733161428099,
 0.5373727153513324,0.0227525004905611,0.955547813276512,0.21228213431974108,
 0.8671199564264876,0.7498835505053044,0.8875531406442511,0.29703709294352343,
 0.3289561148446598,0.883707757858706,0.8481296334740097,0.32461291966312655,
 0.15858086873322363,0.20080524895106044,0.45665036136417614,
 0.31975509851433115,0.7023815615042025,0.3702519451212105,0.4142308728852431,
 0.8745780438075536,0.7542163197642326,0.5146448946183886,0.5463532248605547,
 0.11349626753354058,0.5117660697065336,0.17639396269957253,
 0.6725437369324968,0.41457115049742965,0.9749289980592525,
 0.10162976410664648,0.942385524780739,0.7601714629310631,0.7510198807931969,
 0.6620761097424115,0.825552822373346,0.15293235790935444,0.3673842851872512,
 0.44985652791457265,0.9181264224810124,0.8961554334803665,0.4517875820048769,
 0.19471265768790247,0.7685964357530761,0.7444319017677825,
 0.02964788671415386,0.0015879429003875511,0.18376224410673558,
 0.452462218260155,0.6159377036998033,0.9937692179159774,0.6485021023536416,
 0.39726766241253186,0.5081709521201979,0.17552807056967779,
 0.8880141609883307,0.3381603785239594,0.42176331459203364,0.8989527248975624,
 0.2607469791070057,0.0812440611088111,0.3510075579071824,0.27372888127018125,
 0.5463717761173832,0.267739902794626,0.18022917480343925,0.2837535999395615,
 0.625795310645063,0.0745649856903865,0.32560448505408235,0.3957441410146616,
 0.31752569888075044,0.17246107468471017,0.74872129109599,0.737782052814859,
 0.4574089602678769,0.5881168876039072,0.5880088614899972,0.2688160464268965,
 0.56166628373042,0.03095559682917648,0.6596875662944305,0.9111855467296818,
 0.6986640909397018,0.12695768775567873,0.7440206518889914,0.1176645149514548,
 0.10730649259666603,0.8412850367382648,0.2124561878878084,0.9005542135853551,
 0.15022597038273455,0.33463601791956077,0.010934487876374632,
 0.5072639492609425,0.8623391605246917,0.8894786142703237,0.9958324292596936,
 0.7929545182294498,0.7419493796482888,0.6753289194573331,0.2797300062837377,
 0.0982745749947056,0.5050510326533206,0.9779236834609705,0.556168840202301,
 0.7703102643330034,0.731983249591791,0.8014571242024529,0.95292069636045,
 0.7917508514725209,0.7420378046312606,0.6327892090884023,0.7412906963524319,
 0.7302113728823887,0.6648364156144801,0.849216708922373,0.8132443173952544,
 0.9987323214549857,0.9609468654600576,0.6908251821216247,0.6310722199401928,
 0.8675372020485763,0.2973508065892847,0.2720287602705338,0.5460912385129427,
 0.9289982081252114,0.15250006783946257,0.36046239643933653,
 0.7005394581412626,0.06095993084912987,0.5864997771055485,
 0.11298726269877923,0.3382539465942105,0.21562316777585944,
 0.07735601834889616,0.9196111729005892,0.18370304793148384,
 0.6544288504436075,0.296052395781942,0.5354146977480566,0.3681278825822626,
 0.16049008278923482,0.22678003262004354,0.5389913699885726,
 0.08625484983806186,0.6571512136239052,0.20610935884476933,
 0.5899339103715028,0.5102654899598196,0.3701134965727838,0.13802622838314005,
 0.6817155294134948,0.3508839638189676,0.785221551477408,0.9528802960419678,
 0.5437371218663183,0.8365273210555704,0.1985955491792767,0.4207571839548344,
 0.8377661992512024,0.7043766751216514,0.33206697588813827,
 0.14655055607190115,0.46947828397639346,0.735697268153322,
 0.44489832743272206,0.09777386544846745,0.5261080275866309,
 0.8047533679113171,0.8206328462415581,0.1760495494950124,0.6403020513497163,
 0.3217628282126519,0.07604905486329394,0.9990969005733211,0.676434576290642,
 0.207183430883898,0.8142749732875032,0.20494565755655336,0.1475927925379601,
 0.2732453047451351,0.5702305728985244,0.82350426937667,0.6241011129139388,
 0.39208047054337036,0.4431268197476711,0.5645917649820349,0.4727993264867698,
 0.2318886048735901,0.03772629253042381,0.2726901825120962,0.5548921290612747,
 0.3054459025822345,0.07595532486061207,0.8235054730899142,0.3685221473254947,
 0.594475479891867,0.15850802768587557,0.5782048920296862,0.5692655944135812,
 0.17487494896894418,0.258110028435783,0.5475759973824602,0.2212719008730607,
 0.9852851581986238,0.23425828812263916,0.620891562489507,0.24982100957145936,
 0.7590148169023917,0.6406203808807827,0.021202496268278415,
 0.6464717570138042,0.6899389835020878,0.503693587434453,0.232261037852445,
 0.17081835293756542,0.7677642774681573,0.14093786919468254,0.952541768379531,
 0.9428520479935625,0.07289217093418887,0.25959237974951543,
 0.9458810948429933,0.5522869506734209,0.009939515903678253,
 0.14133227163523643,0.9897251627814343,0.3547836356965539,
 0.16872685788422492,0.9232882860567261,0.2411034910796942,0.3057430366134002,
 0.9065691549273319,0.24497724270669086,0.448436449970385,0.5160117829616615,
 0.5461266631072923,0.019769807112453464,0.46179083748687055,
 0.26733861778261514,0.39302209979891045,0.43186875482766385,
 0.549830883957818,0.19068870282650408,0.4571529424590687,0.5596006479609246,
 0.9072308735306592,0.6987215753555555,0.20661092742915277,0.5041857112274583,
 0.5907541724196957,0.16812930736538745,0.47520555182752167,
 0.9226437179082793,0.17634545634146315,0.3437980156302898,0.3811856455670637,
 0.07796909711109312,0.672453614428047,0.4389402047070259,0.9635641736146686,
 0.5738753954019019,0.14254001379451453,0.7524257315285444,0.3718375891684784,
 0.03801531474654418,0.441224078927535,0.8376884404055744,0.3960204360889258,
 0.3675778470118374,0.5457203150436549,0.638874765619729,0.8764932390349699,
 0.8966097990188477,0.53180430079183,0.07502267695023757,0.49519413986040817,
 0.21516329836414105,0.8336009034693629,0.5320794452091544,
 0.22645234493631117,0.28030238247704986,0.2677447536574171,
 0.7085864461927858,0.03210549839920851,0.6340633138301665,0.6751443073997352,
 0.3167444062375826,0.8524073957227896,0.8152783283615845,0.2675292058569363,
 0.3557700025345433,0.6051461189203226,0.05340336371951815,
 0.13432974101829542,0.4700729748917877,0.36351280317517654,
 0.32943563438164314,0.8547718578415539,0.08375552150590637,
 0.22296061329449723,0.6576251302723051,0.9530544160480535,0.6775300540472621,
 0.5815101769080627,0.8448992345306355,0.41258915507622973,0.5603843737678534,
 0.6463427645639349,0.26806472111062973,0.9346138167352773,0.6453710650087711,
 0.01688583573676883,0.9216331334811257,0.4768349027551372,0.607844804404855,
 0.5659801728979625,0.4698323980801201,0.17348194374945858,0.7357562663781474,
 0.47795899094984784,0.44062821558934895,0.9342247552620013,
 0.3197136793128773,0.5935147016261372,0.06785591422119042,0.894099419204306,
 0.39965302056713514,0.7727135417244295,0.7598784145199137,0.8288620573535737,
 0.8358310853441351,0.7184861179501378,0.6714188659015146,0.2480428460835784,
 0.8314091919904435,0.2738287028224312,0.16686395664376907,
 0.16720399947636577,0.744151070282944,0.4365104550578782,0.14066717295066988,
 0.9970880383127264,0.6580590450174093,0.6308141017060042,0.5739888302066845,
 0.21027700709604158,0.609138124738782,0.18962644473836632,0.7119017797466392,
 0.9654441023478031,0.0585013697185105,0.8434107399624884,0.7490047387884723,
 0.12918311872009514,0.3863721442178907,0.1396000294193941,0.6744090084036972,
 0.6657765570719455,0.24465108530497282,0.6138964296422483,
 0.44708861336996686,0.09174856601933923,0.7874013000790551,
 0.7549327122345505,0.8110940106582777,0.40952162812348136,0.8986390817858709,
 0.1869944391009859,0.5025751957523873,0.3237053616977641,0.2326821353959181,
 0.6259929406073694,0.9180179392711645,0.05978360081277445,0.1965883214784072,
 0.921741529593498,0.6724847381267457,0.9956080535428138,0.43546670925987196,
 0.5390535562618567,0.06686146629903267,0.2492139095969873,0.9654117635421954,
 0.935684318670721,0.9156038311772137,0.26736137360498247,0.7281321718767344,
 0.06778272368891303,0.20345003523266603,0.2542201675560143,
 0.10352042249939597,0.22991622662108258,0.4941585496446992,
 0.9492167439725511,0.39690452888071326,0.6562104401631021,0.7829440909702852,
 0.8467714472986806,0.36247047394069176,0.3730866820459333,0.8103873302815136,
 0.19613592312207317,0.259561671049633,0.5807560323200603,0.731635931987963,
 0.6099914883440609,0.4289986744043366,0.7890759028644987,0.03707158218327178,
 0.5875223488820401,0.2249121669957853,0.8566990366116077,0.6272306676238095,
 0.40529950192416153,0.7754830087001293,0.7042893574510654,
 0.05696565986347224,0.10971638351625312,0.21953417936448516,
 0.40499561336237355,0.4945159922745066,0.2095902574631382,
 0.13593015819903664,0.017632967708325897,0.801226055171075,
 0.8038923751195234,0.36697305679535097,0.010221993488576553,
 0.6770702102792023,0.9874427771507246,0.41600124584282583,
 0.33319956158693353,0.020086277620086412,0.1372878146935398,
 0.30947552320111527,0.40656027865710964,0.8537184433370382,
 0.5695351953492721,0.2734499835762708,0.05326472318631481,0.3978034081866201,
 0.6739930788307943,0.2718353270744934,0.21567343998520838,0.7851610143010876,
 0.5556472359472916,0.6896338946800682,0.2595023673720167,0.10792810346281478,
 0.8090248622984771,0.4379793263282846,0.22060687036081916,
 0.34263080271095026,0.057392695204799615,0.910060037426402,
 0.8750425499823205,0.21918100295497334,0.0934951115375916,0.9917731305464093,
 0.2239884889827911,0.10958114041822631,0.21528536261117337,
 0.6355336256186915,0.32530840358078206,0.7607812538372222,
 0.045869097001730474,0.5837698149461336,0.7635052418243428,
 0.6569014196285715,0.050574311765221536,0.8356836601676342,
 0.6737451865361255,0.8658947273376475,0.17351671199475316,
 0.17733926890596274,0.5373143400528754,0.404867359373723,0.1799074157489542,
 0.7884030408517296,0.3110841180379299,0.37443705775063696,
 0.23389580093909157,0.5619327051926076,0.20381304985994342,
 0.4387569943231838,0.11736635956304464,0.49862576260001423,
 0.42122518033192946,0.27422438985742625,0.660197740093839,0.7803084931743999,
 0.9176698412306208,0.680045295751396,0.26766437936026877,0.6140538377359721,
 0.3445554611877175,0.8796339520450372,0.3184405110023011,0.10800709414885534,
 0.8603289523147725,0.4852004309809832,0.3854709366565109,
 0.0011141581514614307,0.6763787152746076,0.46302948746447936,
 0.3275685805641755,0.029390430830792624,0.6607298801546492,
 0.2555238882420968,0.4453908405384752,0.7616701149751491,0.26293517445130354,
 0.3927738517452488,0.15616207219893952,0.5409041851127553,0.3169017242893778,
 0.46863094428299323,0.4033195863713259,0.3800009281349749,
 0.32281352520006423,0.9348013093012162,0.6513737036077063,
 0.06586797434462477,0.4320993813617654,0.4999790050191011,0.8378212749259224,
 0.930475708410075,0.6781177179613797,0.5087069490765063,0.9164710991231162,
 0.1259012501966199,0.0655036722634067,0.6310963654314942,0.6384898312832173,
 0.7504656764094667,0.37348518736700487,0.048749254806754405,
 0.979216658233717,0.0076937975787995105,0.7950265508535732,
 0.6782774063118536,0.7250149447495097,0.6976618110411381,0.1351638983268082,
 0.5603474552676496,0.022499653376640616,0.4295134302907252,
 0.041602105190676975,0.5934287585001085,0.420398008178963,
 0.49676900787090195,0.5200900531721054,0.80959110459636,0.21076457626840395,
 0.9110683392611049,0.5553553929929688,0.6098613623438056,0.9139228223798629,
 0.07937711196718711,0.2285300408562605,0.8965018374342788,0.4262137424842942,
 0.582497985564751,0.6986582982893512,0.05903898817829567,0.13488332016990012,
 0.25224533219980083,0.8172108240648781,0.7804006624744768,
 0.09493828492462975,0.6284207162084203,0.7000688916057446,0.4109036978394246,
 0.698248019785553,0.9896322451707364,0.8882689866372395,0.7332396839335834,
 0.774206475835447,0.279204631078392,0.11496086902596248,0.7221918104541094,
 0.4680198803981147,0.10503917037057842,0.838931042665116,0.48534344287505304,
 0.09439454278338244,0.9445829651284057,0.5820110422109925,0.5662289868672271,
 0.019050438130779557,0.18416931501103062,0.07704637523780422,
 0.41643938266821967,0.14082188443627608,0.44677049265317625,
 0.5912356937799383,0.8326040701270905,0.02128565382643066,
 0.13163980983115642,0.05270282507506585,0.6383298662299259,
 0.8153949588732494,0.10554420535197029,0.8268152131215782,
 0.21204260712372014,0.9910972420251967,0.6396126191702876,0.5007808139098608,
 0.26611607878349064,0.9705352008864189,0.5458210845801595,0.7648486045852156,
 0.6448661404605128,0.09027659209901362,0.38361172610735705,
 0.9342281325828901,0.9533739342014034,0.6057798746031399,0.7468644678414397,
 0.28486612493585173,0.4647032089006973,0.6094579292654339,
 0.30322280164001614,0.6589539839840483,0.6915993210081164,0.8303030631050939,
 0.1696909021173788,0.46886934315513684,0.8274399884106914,0.8245415754107471,
 0.48079784940355164,0.9890971627004367,0.5376578558411723,
 0.21462354926509175,0.9569361463674492,0.23365815756348374,0.991535322575623,
 0.6027719876599964,0.043466081139601154,0.6376791491220362,
 0.6628263285895154,0.9364425398074284,0.7602742753444196,0.953680089695127,
 0.6401876929424991,0.11466311352185299,0.9216849267432984,0.8542381699088454,
 0.9249407265286824,0.0780236086894307,0.10788519619305326,0.9317615690583428,
 0.21743839193197634,0.14587535478473646,0.16127468023933855,
 0.7764114546788999,0.9008399239215092,0.6679308953320042,0.4734700200238775,
 0.9324982135605182,0.19744204143176602,0.1287465622585554,0.6575525931055888,
 0.19904976688125842,0.7883611839868244,0.5882436355825764,0.5624071203413505,
 0.16518447469758335,0.8272606870113388,0.4985845572307327,
 0.20416835919125087,0.39426591425059243,0.3723725184328949,
 0.45913468162677473,0.5800759758442526,0.29028329221111127,
 0.7729174350040404,0.6031790352278856,0.09427918848729122,0.4355279467387978,
 0.5787804879518207,0.13610155514321598,0.8379621622700374,0.7410963573618343,
 0.6101143744480054,0.94176523895015,0.6567868466159243,0.02598671664890395,
 0.9204370223219196,0.9524132320595705,0.6713803599191129,0.9106974828629503,
 0.6159698880773611,0.8542797257226485,0.8967050250964244,0.22629902459611417,
 0.6590783237734474,0.8736607261740437,0.3262874662771911,0.3256199812696553,
 0.24863344321677538,0.031236852760188394,0.9997594796090662,
 0.3810211455907635,0.27237085416569584,0.8887692270401906,0.7114378858320053,
 0.554919029217909,0.2922962554810492,0.7787796886705405,0.7477881246448317,
 0.35705310837182325,0.30956082400837026,0.3763122458782405,
 0.7856038465867492,0.0745341711818156,0.5251586699644023,0.10983145991668108,
 0.2868938573542992,0.003736733709074569,0.15435111021573966,
 0.0193380034864159,0.8991426441007085,0.9647146494089137,
 0.048929437907643814,0.792470621246528,0.42753104124905184,
 0.20612299127833067,0.39197494669475086,0.43932839761426123,
 0.11800077197863801,0.3782652075134645,0.7920136659008653,0.4711417511031408,
 0.1985017909800758,0.7098460810297338,0.6658467482004149,0.8531760252875886,
 0.843280571575014,0.26077819197310326,0.3053739131321408,0.5450504571607249,
 0.7420017494832084,0.8141430170054011,0.9170060899595803,0.12323635789397935,
 0.7970715544181752,0.9390195876276042,0.1432484436321365,0.5688381990821533,
 0.9708818369477958,0.6100076089736711,0.9485291981745367,0.5265374215797096,
 0.47075064112262044,0.239464323067887,0.7814806577759027,
 0.0015702519000485582,0.7631517767630984,0.46659612273300555,
 0.15101580869462683,0.3379941132230778,0.7327851697421484,
 0.11125986500219276,0.7234277921937525,0.9854680392761513,0.620776564774575,
 0.2560524201557618,0.023736975820675243,0.91185857747297,0.9948061812850567,
 0.6005271718580987,0.8555372337930801,0.3202018422420927,0.6400263423673733,
 0.7888627677347081,0.11221550935892322,0.5576790109839179,0.553180401863973,
 0.1273427624316037,0.04283006680581569,0.040193995927035764,
 0.03508974901671469,0.7177829344945937,0.37846721041797393,
 0.053574998102570826,0.4033121539103819,0.5686049487102411,
 0.13619965018246094,0.6058077026899901,0.7996883332437714,0.3827329016993031,
 0.8046741761088249,0.31163917101145444,0.9545878235578346,0.4859599277400507,
 0.48258107294868513,0.07447212533157155,0.831610515052002,0.5548101252896256,
 0.9483858915991084,0.8256384612481276,0.19263963007201435,0.7165267967432654,
 0.010983566779117049,0.8401233790051218,0.7764134018510264,
 0.4373754069322561,0.22899114713881763,0.6809977348261922,0.4467415296084398,
 0.5629658286102666,0.1349596149314216,0.4663857997854355,0.38164844074882354,
 0.2383964508608254,0.5677526243061604,0.04049628721342502,
 0.23920963961368324,0.7833682038383586,0.34305027102852725,
 0.03636911277869537,0.3564516091409867,0.9581604620071877,0.6439490001054984,
 0.9047175202201174,0.17330258092937745,0.2376443082331987,
 0.28989763177112615,0.27727615423043916,0.6409096812592243,
 0.5645812925121015,0.21139079905738145,0.9079100606223502,
 0.07469450539609723,0.049225824988918476,0.05221550206857639,
 0.365059354826188,0.4756754450466121,0.6718251624059346,0.2836494192086585,
 0.8763449448654804,0.2566787375993905,0.4479845858218794,0.7080355343467255,
 0.5303109816992321,0.01852157840103108,0.2956073723934656,
 0.23647085933352263,0.7188697476747791,0.4067211991327393,
 0.30449685224907874,0.5170710513811101,0.3205070815746436,0.7716610671796031,
 0.3603656814548677,0.8644891636409967,0.4642266546986298,0.23267599321045718,
 0.2139593287753936,0.3579820648996306,0.7282525369465174,0.649518536361195,
 0.5311774647683098,0.7840643544767815,0.6174849776151521,0.8277183671036429,
 0.7747831192651431,0.9778530943957848,0.9860826783042346,0.9689906340095893,
 0.9176677269529233,0.526295984688131,0.850758045019412,0.833014184308649,
 0.9957751157317561,0.3001130261645003,0.9732155395672769,0.2171308686304877,
 0.03598545483101956,0.853385372813407,0.026725047213507658,
 0.2598405699544619,0.3963245240772698,0.29634722810349856,
 0.24528785478043857,0.17861818280540342,0.4809305533749454,
 0.7543720071490333,0.19503542585358025,0.14223842326588332,
 0.23165142480760104,0.841976228825248,0.06217940436983482,
 0.26045619201531145,0.8968345532183339,0.6868682099830961,0.9538246688569003,
 0.9122517342900213,0.6542641973844391,0.38838043305630043,0.8448339214284392,
 0.09915833971591625,0.9882012850919326,0.5006704652748005,0.1715928212754455,
 0.30764677309970523,0.10342804320212262,0.2870607309762012,
 0.44737994582407514,0.35732062605927106,0.06716752123115233,
 0.09453975166903328,0.40229239760416113,0.24151109724820152,
 0.43463674222922044,0.32720965531155044,0.8704761819952361,
 0.9234545306997792,0.7751149867596604,0.5545700364178024,0.10786444849831345,
 0.2476811916365096,0.9852972496674344,0.38170840252889715,0.6918243917046839,
 0.8671385671518845,0.2943664143848861,0.21302172194235813,
 0.10313749163299624,0.9491617109958036,0.9695323169697965,
 0.10293305397861607,0.7878395492838174,0.9118070365299649,0.8808884046889295,
 0.07842642951438328,0.8623747244582942,0.24198420208969662,0.908723378357775,
 0.2997743654148254,0.7896234382897329,0.21125849680936948,0.6750786147885559,
 0.2976140542286918,0.7685243990440048,0.1216858350183394,0.30942322652721943,
 0.49443958137620814,0.4265139941966791,0.7728290786632706,
 0.22241168985815196,0.0937139985281642,0.052015363148352534,
 0.5746593347562388,0.2969378908152197,0.5379643883577202,0.6054748328635344,
 0.13772818019774757,0.14969384408163133,0.4793869202987461,
 0.9301280791763658,0.27685200130769083,0.24947473728757874,
 0.7860947028174092,0.4409653089894433,0.25729044955637925,0.6778747327168556,
 0.9486400456071871,0.1500237186658635,0.6001264049785753,0.3738047878877977,
 0.6208351433347715,0.45223811752331367,0.606993400278132,0.9076639046063532,
 0.6178882269967894,0.09295854528610203,0.6568211575686913,0.633406968504235,
 0.22295649515642424,0.8763153935671762,0.7732123598376658,
 0.05054204387342198,0.43614763414797797,0.8941265849645285,
 0.5518011595398915,0.07087104870645644,0.022210916431533945,
 0.7725367944043386,0.7894565677675143,0.16764747942942426,0.755629704372119,
 0.4907825745354637,0.05526027906602948,0.7609830165478654,
 0.19352526400215908,0.0061533172691023985,0.09546441827076146,
 0.28139194172509063,0.5566562431664318,0.5056132069583944,0.7071865952610633,
 0.05641342286050943,0.8561095976048599,0.8969431381388779,0.5820156985492912,
 0.31963197200781746,0.5325439193898025,0.9598604967983498,0.353813585599122,
 0.597204861714058,0.3482074053198577,0.49511464308951847,
 0.021127887153611025,0.6804621848599625,0.7841562638489802,
 0.09313003191291724,0.819807668242024,0.28915808692613576,0.9871220859146532,
 0.1041622465412465,0.4944303078900718,0.31228272726578254,0.781484717493187,
 0.8633247034759104,0.3949547107162077,0.6752554992486623,0.2474151094424708,
 0.6474978530995019,0.8233798857152881,0.8485134567727433,0.5558592936390083,
 0.8192536251991616,0.7531302831466348,0.8385547811081536,0.24671661463578376,
 0.7965367660370808,0.8449194681434753,0.8636518666250449,0.01269256241977379,
 0.24106980017490587,0.3220071133310365,0.3805408338856493,0.4122411814027027,
 0.3438128956227321,0.4819661279175389,0.2264855544138794,0.8862913582991885,
 0.4271506246027289,0.2910910875355629,0.8521604854353859,0.327402899608614,
 0.1389203073879457,0.5563966630703489,0.9319746321152395,0.37418841754114196,
 0.9168364154546303,0.9578236153459202,0.13837500593976626,0.6958919299022677,
 0.49379320003673177,0.5318976064033601,0.7956830277502497,
 0.15101218773447456,0.47637938080352704,0.8000947443520695,
 0.20890398571548718,0.31809974326049995,0.3635177570146745,
 0.14718845938290237,0.7070358700667314,0.836817631354383,0.956259537024598,
 0.9486645142678732,0.3041968811282154,0.8533732681974157,0.4151372415754466,
 0.855251931718623,0.8170074321065826,0.2896401675586282,0.46386911996482105,
 0.575289477854607,0.8399255018146713,0.2039885413466005,0.3115795624169764,
 0.6189254012547278,0.16887344695096318,0.19409183844249722,0.506985472764961,
 0.32617100795676823,0.30127440456077914,0.4488802141245928,
 0.07594839944078491,0.8094800790472308,0.3156964128790757,
 0.33004403185806985,0.43549031111939385,0.2667226297854012,
 0.17077220676377658,0.1654703391661132,0.9899607639273459,0.9669861961286164,
 0.5223147536426224,0.8288539947957141,0.6991639238650662,0.12371309933522623,
 0.870308545200547,0.5967590228751137,0.19436935500234076,0.9130449276969671,
 0.16535128658928888,0.7472359328375744,0.21201716938854664,
 0.5905058202176461,0.8337891377241192,0.06182321911341382,0.7765183864482881,
 0.44133032288814367,0.6463088130356266,0.6179541178806425,
 0.49915275142494697,0.5922446052240097,0.6093564963258654,0.4704485252632833,
 0.8959768992197467,0.34136775347732884,0.7786578051139657,
 0.26473007650608604,0.45350347747255937,0.3705313633063454,
 0.3877967662035653,0.09670951371887881,0.028452574643471285,
 0.8245585842301753,0.17882576642617265,0.545586242782365,0.8600712540334489,
 0.6170883277072488,0.46142528843999586,0.5981764262462728,0.9658051941872415,
 0.5081309535242036,0.8459297831698211,0.18479429752367915,0.5980565574344889,
 0.7172471824924119,0.2796038725892793,0.7319184572414869,
 0.061437730925735345,0.31280940722078865,0.3417595737844136,
 0.27494728786934197,0.021421016519895253,0.7441346621040865,
 0.3368476747387197,0.9433712671451132,0.9756209858368023,0.4400468653057038,
 0.8606738250080163,0.12320195409042245,0.27841384134045744,
 0.11884380721440613,0.8654331896673945,0.40506752200157947,
 0.42286328979467525,0.24188073659505338,0.24681077939122054,
 0.29372399815474903,0.8530541194880158,0.7185750556268385,0.5089347169789065,
 0.14472275095343545,0.7488826696389127,0.9669551724304613,0.945793827299269,
 0.982026765248353,0.14809704033584659,0.8535021391693589,0.9962201961608085,
 0.13415914257531747,0.2679977712516465,0.17334413069402865,
 0.22721562142087304,0.6724199135461889,0.09618575623517522,
 0.34968147387211734,0.44944025833025736,0.16441222890142038,
 0.7603934030154771,0.9629137507022262,0.10874513859521662,0.7551655305861643,
 0.21660782376804677,0.9984137586904285,0.5499127752061259,0.8136901500334158,
 0.538071586517015,0.13105452348603175,0.6485392471628653,0.7795016441815789,
 0.5861149198026255,0.7417872094783263,0.061304618854051696,
 0.7794983818868672,0.3367905439979143,0.2171362468033604,0.7788097450933582,
 0.04593312295073659,0.5533417508748495,0.4849053311699968,
 0.10863412645151227,0.4189234011220683,0.46922279849429516,
 0.3152321134409585,0.7496576260637658,0.8735365112340376,0.856494595756951,
 0.7551722893995224,0.4582920365267563,0.3953615082250167,0.5502950457388412,
 0.14993477302150926,0.3069118223954722,0.4579866073667692,0.9682808828916464,
 0.14747911797924962,0.4698898171534941,0.765221193075705,0.26996377012313166,
 0.9808871117346087,0.46436204617381793,0.7042510288649835,0.529871766275302,
 0.8432616823313485,0.7330062252162673,0.936415344075193,0.38561466600063243,
 0.9326853258002809,0.6377850823651093,0.8852293121620526,0.4182919692015681,
 0.37467271455027984,0.7509207813279766,0.728796938465458,0.08970798489906262,
 0.42312299073516735,0.7414200240865975,0.8671990779427334,0.6038286556263174,
 0.16475026385440095,0.5943498563804788,0.8631586884687411,0.5723834424666178,
 0.4567893188864438,0.2192109684837138,0.05817476472503991,0.7549255102550061,
 0.32818476019506115,0.48712738661166644,0.6749863042476977,
 0.03246735131017975,0.9955249231936866,0.2205764708857586,0.9555768024787957,
 0.3266781683681318,0.6602624981567111,0.98935355650928,0.3570330610445509,
 0.7219127585342354,0.40620854145680996,0.2627012105249906,0.5929792246447269,
 0.18789038194509333,0.37144508274872123,0.8349820833500086,
 0.36584211120390986,0.6433114888335134,0.9156095113985802,0.7320995919621994,
 0.853747386864868,0.731252812328747,0.5970538812706422,0.13366291541286168,
 0.6250926689150466,0.8530477633267695,0.7621433283005423,0.7155430378581058,
 0.8708529833514111,0.7398057924604974,0.2937697842211059,0.45317127590072026,
 0.4059112080687738,0.941706608582143,0.48627221742311355,0.9661029751692884,
 0.09442783399144106,0.6550538983130259,0.6194355329130756,
 0.08692036702119144,0.6928717748091793,0.9160817491223545,0.5110149840073851,
 0.6331685620411872,0.11382953343678859,0.17985116310093707,
 0.9636785389021091,0.33278278483721646,0.5403770465542475,
 0.48549936670597593,0.7850157541865029,0.09938716361458988,
 0.5418654681827786,0.5326374016772265,0.3773206489326939,0.383356391700701,
 0.4126602516610548,0.7510546256251732,0.2411875561385579,0.39879213841749117,
 0.4898817207407382,0.8086286690882194,0.6235751192551449,0.5912009385571791,
 0.6714517785111254,0.7548838482007212,0.7127542466076883,0.6486141280227757,
 0.1247110363557945,0.4257844035831553,0.03631829854544366,0.7835368285312807,
 0.9158768237427166,0.6390024797373913,0.3618789521139051,0.7852703654434874,
 0.9685143629083407,0.07701407918612657,0.23437971732948237,
 0.12150288952090094,0.6908593467923326,0.7196751101527079,0.8179452178047496,
 0.5831464274983195,0.8358811095524167,0.8300427227107727,0.142362271452512,
 0.531895988667983,0.7419883991654741,0.1493435485120943,0.8097834896413861,
 0.48537892143824235,0.702942634413394,0.20625182598207015,0.5987811566572667,
 0.7664335015008001,0.18579198162937016,0.8429872006277832,0.8869115648632715,
 0.20325333964344872,0.6636622595027194,0.6096770646565239,0.836959809933497,
 0.18156672439425892,0.7569468479724347,0.7057744860169181,0.9343889444988124,
 0.2554583905952936,0.4597276977692347,0.7014668409514748,
 0.023826282245877017,0.19631626786328538,0.17298503500304852,
 0.4480095183371705,0.8984789569127442,0.4940390974836666,0.8912593576358573,
 0.020835936078472672,0.8179407449072524,0.7982733212591118,
 0.6914523570085362,0.646588116297629,0.4145243338591895,0.4177700006844185,
 0.7355475212736938,0.8673969951745817,0.5133331579312471,0.6544296324893903,
 0.8314486329702971,0.03544100995799315,0.8807655592142986,0.6982719542962127,
 0.9318822703028267,0.4186178450465049,0.8726112609149491,0.27811269684845075,
 0.7927956898658239,0.019142795354543996,0.3327511392880669,
 0.5813587286302065,0.9606255957835041,0.2522763823798949,0.6421131915502458,
 0.21327831551711318,0.33808688190582825,0.27781322560014976,
 0.6957992112963398,0.36201270336531566,0.11689032654814402,
 0.8838981801554507,0.3978074468815528,0.5762691565952758,0.2699712783036018,
 0.050026922649937355,0.6621166033227757,0.4066822435870281,
 0.9513459826206069,0.03324529722048619,0.7222711286167237,0.162168576716068,
 0.30917960909952014,0.5065638109751757,0.6113102452854262,0.0390108243796381,
 0.1011428247222198,0.21117478223760988,0.03167259222005747,
 0.1603662233954959,0.7770715696551657,0.9932579295776102,0.14626627457905794,
 0.3895232706364513,0.4431829464350362,0.7416475103067205,
 0.024552590606270464,0.40158489395193864,0.11409550102942134,
 0.8484428054568487,0.2846601742448136,0.5596952866641624,0.7910605820384796,
 0.9471156967941899,0.5269998303495176,0.3166317593445739,0.5541423948112623,
 0.2825002713022521,0.14267301392449594,0.6795621029402267,0.7282542754185403,
 0.015987567848920636,0.9191073797232298,0.8461088965905829,
 0.08196393918032774,0.12904283593684807,0.35616414376368977,
 0.39077894630788357,0.8526847611469769,0.15555874493897082,
 0.29474266914663594,0.7539348590257917,0.5438638254889357,
 0.31665003423924365,0.4589841814297535,0.36022700787885265,
 0.6136382744736959,0.1433898779056264,0.6437287821732709,0.04706217794223766,
 0.7950314791418449,0.38793049249784795,0.860895889770072,0.3231406820620004,
 0.8937270143751523,0.3947486510000878,0.5583292058771469,0.41747224406461014,
 0.17998992112881929,0.026122072803672958,0.3692052981225037,
 0.09632444686560726,0.7628000835643831,0.9987713266670636,0.9707373908412296,
 0.27985976737666873,0.620734463780148,0.8330093137337369,0.8878782897776508,
 0.05934231123424605,0.877928235593204,0.7287506060549671,0.71102720166414,
 0.8926565435773623,0.7809912206166716,0.32256275102609144,
 0.28032968276105885,0.2001854422401056,0.1690131793186479,0.9402259278687303,
 0.4992728459051108,0.05674294689364556,6.387455007393905e-4,
 0.02178421183784407,0.7995891765520955,0.42230801674536433,
 0.30878103535007084,0.33279014840531607,0.5386630534191476,
 0.40853822220780944,0.9117532367257124,0.24208501732315924,
 0.7600994584013907,0.7776567689604608,0.41035226723102336,0.5270402127209625,
 0.2218624592948678,0.95243548802008,0.9361046950289135,0.13468375556656376,
 0.7828280753140213,0.31277140190386543,0.8537018706271796,0.7180889662096412,
 0.9939799480097994,0.8639973665425228,0.3827305004664706,0.3253311686376684,
 0.5882351045375466,0.36821434183816915,0.8865395679170253,0.3546168013336697,
 0.5913976534214113,0.5704508238114977,0.357363347228148,0.060655003348501824,
 0.8400418585627378,0.7890252209330557,0.9381529979668555,0.6442435024193562,
 0.47933537076927624,0.8988923895550885,0.4079625150640481,0.6265146368045739,
 0.08225988287173902,0.7716090728724888,0.198856935246863,0.27622233009857455,
 0.45398486493643375,0.1371436086049328,0.7863984746500419,0.3716442454727864,
 0.11077458266985596,0.9091086224507584,0.044163950126113516,
 0.43954221547381456,0.26312773540615264,0.6475461821778129,
 0.0964706323792357,0.14232642365255144,0.34307088164489796,
 0.6587594240384905,0.285679829390307,0.21240343404611695,0.5301281956139491,
 0.5071289480688081,0.5975308140243999,0.008029294474795723,
 0.3407902572173601,0.5416462249094435,0.554007317207727,0.7393939576957794,
 0.6106759864839186,0.17722841034689873,0.797815276839096,0.2778252788929162,
 0.3055123815166201,0.34533991103325357,0.9763288347984969,0.8728840106492349,
 0.23107884891273378,0.05424477982146858,0.12262989109099687,
 0.4923375859168718,0.5248207608833249,0.5619723176140052,0.6134799486625692,
 0.6627874922461197,0.1946682767512058,0.2531943687336463,0.02799253496533427,
 0.221225477408871,0.902286427777199,0.33099842967720616,0.48575728383317285,
 0.16445054237706735,0.4447848344929741,0.6021073742066501,
 0.24268723956273552,0.6647682052957646,0.30314638251641857,
 0.018636214592334555,0.34558180947963746,0.4526347689216913,
 0.04508159334249606,0.4751243661175558,0.6118642827146835,0.4853626986954517,
 0.362465154332381,0.643501365527921,0.5925187017337377,0.2892661109944894,
 0.6522528931611024,0.648416780563274,0.021298255649137898,
 0.17217028817078728,0.34032198148764337,0.6383039778024514,
 0.5156462195274327,0.7246104895901309,0.25730836181002903,0.5746084560711329,
 0.7116130430576135,0.3709250322862393,0.8529076864217233,0.8174694099007753,
 0.8792330151557414,0.2113198355927013,0.8341681747405754,0.7884204500383926,
 0.9863419511104707,0.8947381723634679,0.652155689778949,0.9648081182998915,
 0.8946183012565119,0.842820539642994,0.5779909241690775,0.518747362210214,
 0.08276510099474266,0.4226185721061151,0.3735326837817403,
 0.15377750667903622,0.6925023581529065,0.8230770759504697,
 0.14971024375648434,0.13114182661807627,0.788733598277139,0.7377507003245337,
 0.4442905939443026,0.8137693388680728,0.9013082692640619,0.560447682012038,
 0.8472829063392029,0.2837967530284352,0.5661893087796418,0.44387638755565817,
 0.255423230814826,0.33721408786622575,0.858618794526571,0.242996561554353,
 0.5873369161320094,0.2777649168770028,0.4354978390371932,
 0.045222713850930574,0.45697439652948024,0.505636028785585,
 0.6728008879511026,0.3324159501522279,0.5144381626356904,0.599139116412569,
 0.2352507649108626,0.28232030063862645,0.41193166557149585,
 0.8636634983140576,0.19536807971639192,0.15099301843948942,
 0.44641631751989663,0.44377232347191486,0.08284315843699053,
 0.8429716163998352,0.2569044711976578,0.4645437665629034,0.7708777968607728,
 0.052600791181586226,0.13732659415838644,0.7298557814791584,
 0.33730736868156186,0.4466977623285826,0.80628561869176,0.07772017154264421,
 0.06112971815790924,0.5265931859654481,0.6599563332476202,0.8024068165593694,
 0.7518127658948792,0.17972439762609715,0.5696049204003211,0.7865040822872635,
 0.9535145830158843,0.14669767476191187,0.674264646304572,0.8760293741911663,
 0.5349578372440386,0.367702501531188,0.6265989031222465,0.43561256938946535,
 0.5209524745122671,0.782756844591787,0.9122670430299159,0.9624121014932716,
 0.7539743200266529,0.34908299544955745,0.6720649158380161,0.6081627068912288,
 0.6600770847996027,0.5586138168170653,0.7104937779098316,0.0776448205570851,
 0.24990882631161482,0.07311603273415923,0.04962660537654462,
 0.7459747283586216,0.1827962498208613,0.9852321565007741,0.8744403292210987,
 0.7186807209945504,0.5233610216770407,0.6947346448125049,0.8619389745290851,
 0.28786979725877704,0.14614436586335966,0.7292389094077825,
 0.30145277799522674,0.6013806952267043,0.946634704173459,0.6967352516342054,
 0.1701816758072522,0.37290633375714877,0.06564599999094523,0.600758315311068,
 0.6008734612176688,0.016278058068647128,0.5806715259564684,
 0.6505245725357325,0.917910849479454,0.24553569484346183,0.25255564259103436,
 0.11899336034140151,0.3997604663344526,0.18018150047513204,
 0.6565864640027526,0.14460090299130757,0.6083425145733881,0.7111806702995274,
 0.5672386346394126,0.28585175909638805,0.8643333835614122,0.3708522158058005,
 0.6880872181342416,0.3021355961852974,0.43363025434240443,0.7724203121442953,
 0.09662448352459019,0.03639137427382566,0.7235971527920466,
 0.1556165882379933,0.647597915516379,0.9761296372554291,0.21078629645667013,
 0.2543548839876084,0.17501247088494032,0.6314550817680078,
 0.34727136139517845,0.19906858195312616,0.5368316427083397,
 0.7286994229239383,0.556386066035007,0.9657999561350881,0.6834101786757644,
 0.6747897616241333,0.8393156671357526,0.8128916792251104,0.20143550731486626,
 0.26196266690172954,0.9057635702603756,0.5261506672566344,
 0.40814948835834985,0.9503851785517153,0.7144312101095528,
 0.31331321635135845,0.8646337502854935,0.7692644501433287,
 0.016938349158373223,0.5857338802311527,0.20949336420337616,
 0.6597003423055583,0.23667925498661546,0.09124273501808944,
 0.9581102362662396,0.42268469194910685,0.8503302288828644,0.6096799644179351,
 0.8091643526735679,0.2402485407130437,0.27314852099242437,0.598230237678107,
 0.06338811863877236,0.20428649247294695,0.27337340539051314,
 0.0525525232096028,0.829851895196728,0.6457593029051896,0.04615647036207493,
 0.552562193947665,0.16440996862724278,0.7318113192919773,0.9883115823133598,
 0.595565855755459,0.8698109187774434,0.061897358422075444,
 0.006726405681852121,0.34177955982449204,0.7001691679388103,
 0.20840336232472634,0.07029282927519964,0.477124980108546,0.1046541194812669,
 0.5635457287449892,0.5823679387237011,0.8749446571475421,0.7743980804763542,
 0.04835442777425081,0.7578743690753083,0.48142992052230216,
 0.3501268156241142,0.3567685537082488,0.3283301949508446,0.13703467194289531,
 0.23245108302154827,0.18141259998220416,0.12487844364216782,
 0.7567416073918187,0.05141077980483477,0.36242134350545674,
 0.5108281400155958,0.5649419427494643,0.7725058326195791,0.6580146586505338,
 0.9444313434854938,0.5824089337415264,0.3607922618064482,0.2119143465221388,
 0.6129330065866162,0.8877075857147513,0.21165969125985895,0.411816050985536,
 0.1560591798316746,0.17426405876535223,0.9895024087245351,
 0.18122512175278904,0.9877427968918185,0.37598863550709727,
 0.6214324798386723,0.08584140248014016,0.3852763949342096,0.5872636674069333,
 0.8736083361671768,0.25212335670113895,0.34332028627535516,
 0.34677866048552963,0.2568111531650653,0.9588365373384884,0.7377754248228933,
 0.29748971125875856,0.4774185731540158,0.5035046758517797,0.620715528485881,
 0.5328232731112723,0.15021094819034553,0.9356423989814668,0.5514640068268488,
 0.5616581361156845,0.22575969416853314,0.39538057577353325,
 0.20192592252627783,0.9132600910170723,0.5360267518906712,
 0.32547025461892176,0.908914059415752,0.34518024424703386,0.7944841611008262,
 0.0925985502105896,0.7972975846429671,0.5328331875225739,0.9892613944014038,
 0.10716354900793013,0.6697423091519431,0.40755099985405197,
 0.14474013550053388,0.813843403392975,0.9959080045786207,0.6395681313273582,
 0.3009327255679827,0.9471808986090537,0.7488177182876742,0.874833377956578,
 0.16933653586166753,0.9683338350308637,0.9415735161464411,
 0.43862332417852956,0.6374047745156483,0.3490300700721448,
 0.30168086406319505,0.30391120789429416,0.9791142867943373,
 0.2125555428412078,0.15069283155784607,0.15038501605115906,
 0.5887132246689503,0.5107865055136853,0.303107815974475,0.5084905620170216,
 0.9824035915847547,0.4547390990715737,0.3032513266737721,0.5733743819390962,
 0.539232016137613,0.5695337092129672,0.8148104160892796,0.4111236641609498,
 0.2879707740573544,0.9339129537279787,0.40855670652217424,0.5434098709733084,
 0.40470490478037335,0.1450741444170187,0.951938583428875,0.6204222209015006,
 0.9386750755417853,0.14559263749607343,0.3491857391681099,0.2929591023288024,
 0.2169184658518476,0.5758264309661029,0.23013514702118476,
 0.12440512576680851,0.9586820769726958,0.14734685619520638,
 0.40019773830501104,0.8214488170593882,0.4568408947660356,0.562790985120036,
 0.6849669298500718,0.6870149893533699,0.05582516650690439,0.385180065564238,
 0.1366060728185624,0.22423342145595404,0.5378732255605556,0.909469536753664,
 0.141755114951709,0.666763943919867,0.8592473763358824,0.42746576018938787,
 0.7339726161247027,0.10159307830431297,0.12537078709843996,0.998059955093039,
 0.3678353321011586,0.41212952457224716,0.3785137499179414,0.1925001858893567,
 0.4701376679564503,0.26916634906092707,0.2181512784459989,
 0.19651914516490632,0.1921224416808156,0.5917464426512912,
 0.37515459333782053,0.909851128844807,0.38899038176156864,0.9363024683767734,
 0.9267852140943914,0.44543920339883925,0.38628474131232493,
 0.9285862155138273,0.7114268077293818,0.30555592395942466,
 0.24264739557460135,0.6036761355016963,0.11785451436099836,
 0.9170027579082352,0.3197861317895999,0.27705409770475475,0.8004501937999641,
 0.6190042202113142,0.7299456211097793,0.7628841558516595,0.12653398908676317,
 0.37521085072566107,0.6161172888492729,0.49261100700273885,
 0.5161143501079684,0.4786177888401476,0.18369995931545258,0.9303011508731516,
 0.38958619223267743,0.7619678512086345,0.2583606457467059,
 0.05519341492519747,0.7670535848066172,0.3505838505313993,0.9279199622335258,
 0.3026027495155128,0.3268535576083711,0.6439473543566887,0.8498745359489056,
 0.44414814175319206,0.9072954757594486,0.3443292075957517,0.7142113187768087,
 0.6085631681315607,0.8544257307280164,0.5357178597489499,0.9978253068510841,
 0.7730389526559136,0.8549262606466962,0.735699313369009,0.16564663249751266,
 0.885273516795017,0.20812119538328266,0.1696175219169439,0.6582606590336679,
 0.055572485071143984,0.629736120223269,0.2815171035527311,
 0.04897263378051386,0.010740880224513027,0.5851755696533061,
 0.5444399289774132,0.6475044784182379,0.15842429983346895,0.502757736449825,
 0.870749356136545,0.09988113174745139,0.5934410888390641,0.6192778254213398,
 0.9927826568784398,0.3731353523270281,0.806564613994993,0.12168192614773488,
 0.796892449897337,0.8611057763796099,0.8782835640257277,0.22644637370927867,
 0.25114786677008716,0.11622944988475437,0.7162031724779994,
 0.001319894167390312,0.49391145781766366,0.12182811079290268,
 0.7180560104588019,0.0884020843846901,0.8302197403708717,0.6010020527104185,
 0.9844357186971519,0.279835621275342,0.5206721579690027,0.3608900848152383,
 0.7086370548289828,0.7086883489419817,0.8035326291752625,0.45786409421129726,
 0.6614514123569777,0.7235491602348069,0.0526129903092879,0.1104005076294291,
 0.0655556671921137,0.7938537417953474,0.7703512858048057,0.4361608597914034,
 0.849498093293329,0.6573074846613882,0.28011818263966226,0.5246165443138227,
 0.7952902266640104,0.52216060067525,0.26223842325493263,0.4653403965822671,
 0.639073584582034,0.9208027303447606,0.4462157527212778,0.9749008788202599,
 0.23317589810930683,0.3978474873609227,0.0714827443796302,
 0.11424511891698697,0.015271169286391384,0.29787981959525556,
 0.7633408729709193,0.23291048750395293,0.6508303485311078,0.3855067759656666,
 0.39370542073365467,0.01954617879656595,0.8335135896160165,
 0.9499196144232779,0.7048086232154248,0.8713798189070551,0.5584807699429554,
 0.14382371543116146,0.32187765218454,0.05077641818217027,0.5192064438043071,
 0.10230391608997547,0.9714750932379854,7.498695481878981e-4,
 0.6574582781201989,0.6038354550977636,0.46165174699313527,0.7615338174048112,
 0.612667630530188,0.5938043450414998,0.5418692511082039,0.9905788931106343,
 0.5600217993623853,0.4081565425317878,0.7259324125353136,0.06482730198598707,
 0.924732070262635,0.29140381766703727,0.4909238652171387,0.8997705727253101,
 0.5161633479136865,0.2503716684169306,0.631483882675131,0.1427525095220713,
 0.5413879662838522,0.039225856157552474,0.821339049344468,
 0.10807523177157952,0.11083855444312074,0.9492102339166238,
 0.8895342271528717,0.4915800019642427,0.9149836599893698,0.2738428316773076,
 0.325988707743747,0.4923456678170477,0.05795881249256474,0.6508015967418683,
 0.09342026694473549,0.06420132197428308,0.7872791848110545,
 0.6698123098244596,0.7525294679303923,0.7443813987180858,0.8133392001018221,
 0.9603119483379792,0.2322560643066286,0.4757174174308819,
 0.017168123894897414,0.12003472263641368,0.5919667929839697,
 0.5776087730857838,0.04350950778219154,0.1962644732766975,
 0.37245399340492025,0.5213154538587011,0.9977761997461738,0.9010963817160815,
 0.16427097226360865,0.415516080010636,0.23341147241585924,0.1104348433644593,
 0.5750460909952156,0.8020894394571756,0.46886205992732677,
 0.21453309798085496,0.7240071888377353,0.6143721998068861,
 0.13220542739967733,0.02470642731241046,0.9705867831385601,
 0.2592111987486667,0.7641065254384327,0.25805940819861994,0.9041849615878174,
 0.19267255232381642,0.27845784122551287,0.30746823619531793,
 0.4935001294711573,0.5766317236043081,0.3523921045402474,0.19719937569139523,
 0.5596098750186134,0.7201655352481183,0.6028866348504623,0.9971890789889181,
 0.3893003601745888,0.03302310892752702,0.9562176077089114,
 0.12939517112257004,0.9213072903988071,0.5948547713722544,
 0.29622551793757923,0.6624497396160478,0.6523290263941804,0.670026264984597,
 0.8475544605531964,0.4947568462070324,0.30229665788287163,
 0.14939992450594608,0.462665011218363,0.3362746356670874,0.3523916678708148,
 0.5279364308109545,0.30614671501916935,0.39463906324688214,
 0.2906327715236904,0.7856173415244062,0.10693482069013238,0.8216761761508877,
 0.5306616034292558,0.92576966368367,0.17195335226326858,0.9062690054640578,
 0.5461639639634852,0.8219551456604447,0.2951612654346448,0.9407590681125738,
 0.9593648195703091,0.4738427747127094,0.3625065066517109,0.07845204081664625,
 0.6396855664589478,0.23393563699654285,0.06614492908973169,
 0.7131007744312297,0.7099870514781954,0.004903527195295299,
 0.8776384159047953,0.71546567308529,0.30018505844584853,0.45328255230469816,
 0.6312508678203682,0.3831227187602193,0.9526242853785442,0.3446401456948581,
 0.7174395722499424,0.4575065389234202,0.24651339597759458,0.391552255017357,
 0.33942236287199246,0.7441669446806728,0.20046123988165399,
 0.9639353329079436,0.46100763092310837,0.6974344453027552,0.9705142576684258,
 0.7161864370298785,0.8141963102583893,0.7479955760909761,0.6862571968540156,
 0.6025565937133288,0.9343398840579933,0.479068756304341,0.4115847975388308,
 0.9642419037136457,0.9328613397864977,0.5412867731769855,0.6249143869932561,
 0.5910066132109565,0.3962455457062859,0.3540345366326749,0.6878810712978941,
 0.3544596719835812,0.25696942206439954,0.7919386137166436,
 0.08443651565349586,0.12452295036650851,0.015472774503094144,
 0.9616534660719276,0.9463018738898921,0.5333688735918533,0.7292132106504954,
 0.6345079713224329,0.11042007301811307,0.7557875260403973,0.5901814099066217,
 0.37145814412677236,0.9670825649939273,0.7758094534425163,
 0.024741072242594364,0.7957199497588509,0.15656437077237417,
 0.5257250817472277,0.7861794744013392,0.7169911995917716,0.9275603671032154,
 0.52931048452269,0.8586881274144288,0.3302725647224747,0.1387124550544765,
 0.01883714099051137,0.8408476795636887,0.3645709125700851,0.327922355096838,
 0.051406361499573894,0.4823445660808837,0.7475128091394938,
 0.8487090126279677,0.17483243869408438,0.053741633697044766,
 0.6395422078839532,0.6400243745197265,0.8115707823242101,0.5434941924892052,
 0.916937999839583,0.41899900109735955,0.231742233521402,0.4053756024966504,
 0.7101347217507356,0.8084585809817733,0.8518120163303158,0.8554359469209886,
 0.44755201197147887,0.06179169737412016,0.7706973042838193,
 0.5764585882862534,0.7124694666243044,0.01381425510248957,
 0.33664830892475295,0.7337967243874731,0.5706897251073431,0.9714965346382531,
 0.19796461185926806,0.5150480868034969,0.44219229929414716,
 0.9899624151730183,0.5455117746847892,0.14367058564424628,0.1139433398873606,
 0.15703473913334332,0.937951422514555,0.05938806099825211,
 0.01599616781065505,0.174899283394497,0.30501987089145255,0.7824371781451199,
 0.7303472854193289,0.4448806842562423,0.7105141936372557,0.5320892638516402,
 0.9288253763413267,0.2767165614150946,0.9829554178641019,0.35346074166454566,
 0.5194424544557421,0.11712719096075663,0.9703131543905786,
 0.049725959763151195,0.3928573098745274,0.18842828062343142,
 0.17319413198521239,0.3759173451459915,0.828149672449994,0.9183113392291237,
 0.4797761362908156,0.40455866096646254,0.17744193461575963,
 0.8516991045170703,0.34194033535690693,0.8597103479380793,
 0.13994617818673338,0.06875371586897439,0.6106186809315185,
 0.8726023438720033,0.9079674955124425,0.47902440210912567,0.720691882408743,
 0.43805588444788446,0.6518419699507736,0.13088014403501735,0.876591108543014,
 0.7518979924321243,0.9244024658024721,0.40395455862177565,0.5339083912374734,
 0.7237119956538377,0.3040737902949583,0.20971334300762456,
 0.11782959955616346,0.9820315543699537,0.6910278804370761,0.419078347319935,
 0.6464737089120974,0.40791266805793347,0.9924690513103467,0.4383101890931741,
 0.004045315396393412,0.41647014742145794,0.7852885625999224,
 0.3851072436737224,0.26116026873882425,0.0344960940377792,
 0.10487984271230699,0.23137038689853617,0.11744036992251472,
 0.4978446431810122,0.9222354544089439,0.297128167864268,0.8133196425922716,
 0.17009478413985601,0.7715596447422636,0.793866070523837,0.17490882320183088,
 0.8706991236205794,0.34231035745994376,0.9068066913211643,
 0.08433411395815027,0.20021921520103114,0.12775568153632433,
 0.2888620307358146,0.40988838858517673,0.46700106172756006,
 0.09007781850336383,0.4749920238425076,0.5850969795600702,
 0.14652494182927978,0.10098988206729342,0.6837349752285183,
 0.8262891514205456,0.4626714862067144,0.11764293875861687,0.9719058457531728,
 0.4106255170212808,0.8803360143558435,0.2544220910189896,0.4052233882036663,
 0.5990022057685958,0.24226415549360447,0.20779053255323487,
 0.2576807542176113,0.4080164071345812,0.861732804520033,0.2806451148641005,
 0.7100172434792602,0.070596843825666,0.2826430128019739,0.269087666085733,
 0.6860392770757657,0.8266492440222164,0.7910716423840043,0.9447993202497478,
 0.7551944030318447,0.1470341276610757,0.32620400709394604,0.6137515799648504,
 0.5583121350885425,0.6601523891790352,0.05160535405929201,
 0.36575919733698237,0.7515377686792505,0.9657149186443805,0.6560056233182319,
 0.9634289422320103,0.3712785696428339,0.639753318101687,0.6278885042398693,
 0.06761848559073702,0.21338062964528992,0.01048207692541181,
 0.7450445061853059,0.6427547117288236,0.2569330953178536,0.7537574908315133,
 0.8924517818130575,0.1661510180381689,0.2926260479767139,0.5980712171557543,
 0.4423494967504883,0.3533255133041774,0.6140233939541228,0.9051165881835811,
 0.02326617952624721,0.34492969863605694,0.26174350669090096,
 0.5678062960255315,0.4293442946237599,0.28150081701895613,
 0.19093073131414373,0.14422185178539915,0.3952703106636678,
 0.6765978199583793,0.8901216251217264,0.6786458769208188,0.8181285255196523,
 0.8700018115629019,0.8172608306443117,0.44290702299499374,
 0.038109673046065495,0.6822675472098343,0.3754322366932136,
 0.4627242655984217,0.8778086153764,0.9118339937010953,0.8769335513102375,
 0.3791626606029337,0.574249223566073,0.4098349913856685,0.03937507798000417,
 0.534752700247319,0.012129741566234586,0.49137951201080954,
 0.8179713174869327,0.3715202842418116,0.8537808821536605,0.3142568209971509,
 0.6492726747976418,0.6806640421605239,0.7778285019960061,0.37973762233602204,
 0.691792065751713,0.4931152452239489,0.9162619381998545,0.5543230850806576,
 0.4964831882021348,0.8351380535223273,0.38186601919949315,
 0.15939673637445995,0.8530821732879192,0.7033951194580821,
 0.38647583925842277,0.5577635680258994,0.8203801443599296,
 0.20335050300161028,0.29763949715584626,0.8983638809190695,
 0.26090638304761204,0.45051584350147733,0.7700407677483921,
 0.060771750675591596,0.8051252085635047,0.6207678619356222,
 0.41994411678285815,0.07681847276613163,0.8244573107805115,
 0.6890434457670783,0.9094289909156447,0.04566481277726564,0.7037341406413101,
 0.3170295377881297,0.7385997540382899,0.8679806609103198,0.4045916286723821,
 0.8507355388731268,0.8750870457202649,0.08022179338899316,0.7191084792218579,
 0.529831862177329,0.014428872602540221,0.572296244996616,0.8067348116729537,
 0.3359987989661033,0.8129437468584858,0.9988780458823775,0.9151399529696839,
 0.44747562616763537,0.02781881370076522,0.018220036958417007,
 0.35821120029969766,0.25293002812337595,0.8503510920844877,
 0.3132574502259855,0.5543266832349795,0.6543389426077385,0.7787458324502237,
 0.995824430791802,0.9162928237943514,0.3676007627553052,0.47725434415504,
 0.9682446308562334,0.7024035114621968,0.7508981514837098,0.2827747732429261,
 0.12005245807448739,0.7715308593730746,0.735098249920903,0.5105390485430443,
 0.8038836734991659,0.4895077365981185,0.9137705253096369,0.5422972210130319,
 0.4359155215658089,0.6628542641621207,0.591040502515704,0.6404750250154098,
 0.28482855621393177,0.654687318596606,0.23798761237981358,0.7659146110856209,
 0.8356721451096649,0.022833319416097675,0.9215763198248826,
 0.9194821528597952,0.6841775173528308,0.9940634970925679,0.5516773465386187,
 0.30151475044069564,0.9151321827979207,0.3053293454172552,0.8371175151117187,
 0.1715678812437742,0.8998911140817817,0.8050159979264595,0.6465334073854024,
 0.13167094198022367,0.17611757830127361,0.08732645201310663,
 0.8940368094942764,0.6607983272890129,0.6673546060971564,0.3501990674754598,
 0.49707615904122104,0.36460164772533665,0.6736883364592241,
 0.19248679407527536,0.3078598401311521,0.10877063896934813,
 0.9812824152416271,0.4817992179016728,0.7772177871089223,0.41625139340526673,
 0.9241318723872433,0.5803161793431972,0.17236025197548144,
 0.33953715312618105,0.7627643102139103,0.7150770387755043,0.7005220610123128,
 0.5210235171670294,0.4680400380509172,0.13659506405295518,
 0.41256882721793975,0.5167482963514579,0.6346941530969135,
 0.24613150334705192,0.01290762092497133,0.9957432317727936,
 0.009155737239935235,0.598092569849508,0.5276345618056816,
 0.48652674224049075,0.25304325659254434,0.48777799481350437,
 0.5407988012455242,0.7090485304783364,0.5251786974153079,0.43408710466665634,
 0.5775704823591108,0.8991145883965916,0.6760386549620784,0.9757492437184787,
 0.32995670238557806,0.06487189175666597,0.8988664565505328,
 0.2539225927167018,0.21136277959701,0.34890220724737875,0.19182500961828397,
 0.03667931259613033,0.5018179977374018,0.15037530291885026,
 0.08930961189498587,0.6802346857496853,0.4304981096671294,0.6453187772593094,
 0.6886118024746354,0.23632257459184125,0.3831186322312604,0.9442366408175096,
 0.27807818838687115,0.29016759612991283,0.7910960753533471,
 0.7596280552121121,0.1088905490401546,0.4267289555437416,0.05308079087671225,
 0.7461614150775583,0.8654259168841714,0.7988160766650557,
 0.010176493363078576,0.5287334703777157,0.3034177099783917,
 0.49411859396505564,0.883966990901672,0.6674365055630206,0.16870993459212946,
 0.5506811271970309,0.807207014216756,0.054804362798798856,0.6628164385954498,
 0.5045641282392024,0.6629863533648213,0.09269255037441004,0.9774045902439352,
 0.15597312312701206,0.12547597506287445,0.06087885238704882,
 0.06463723875537364,0.8341533585124239,0.5583657830154882,0.9822438761681133,
 0.49774162407188205,0.5161962395884387,0.42482747379017327,
 0.5018736992431233,0.7732341506318163,0.5526244754996025,0.3169593696257772,
 0.6442246806090199,0.2342415544043125,0.8318471824165857,0.8455255831803966,
 0.5722400345552153,0.6572972376920465,0.25878455473299167,0.9363136923103268,
 0.46008462415904083,0.5606399892912445,0.23316782866948849,
 0.7349855135499426,0.7196303547021976,0.3303615893401818,
 0.025531538208826587,0.10181611102704802,0.3420152125997492,
 0.9697792484838623,0.648641567760973,0.7952919036955617,0.806744938618051,
 0.9745670592646702,0.1426474946655547,0.9435989594214585,0.3050383121967054,
 0.8207692746654215,0.8881011587558549,0.4586908706619355,0.3959624672020736,
 0.6793521737409827,0.19694215488910793,0.9275777456996446,0.3697444904913034,
 0.5204491714844381,0.05906571032389396,0.1264480188949163,0.51811025244916,
 0.6642863066924787,0.12559837009989838,0.12258349220174625,
 0.9690058387755234,0.4104523975345765,0.2113037537760527,0.2524775045686357,
 0.4620430672282402,0.670967037899117,0.5775795576547791,0.27101400237272955,
 0.3897124364600437,0.4342078621578759,0.41890939971445584,
 0.20448758268359868,0.19168439655582148,0.8954315397826813,
 0.7540639141224411,0.6949746342171927,0.3324700476317546,0.9829107520626907,
 0.06587684453933496,0.7056362811761071,0.7572619408419455,0.6505822244252368,
 0.8325695232400034,0.36523861428693993,0.2633505031580342,0.5643773544493246,
 0.04367862589031768,0.7353229180549514,0.0012814924340749378,
 0.8615913264328025,0.6966519899904057,0.9716042613835333,0.5150439638022393,
 0.7336454282444728,0.9074791281225667,0.5225664360916551,0.7570301799463383,
 0.5835445590019211,0.6156835041400383,0.12090088127503895,
 0.43507851885211046,0.07160952154393774,0.9994230238682191,
 0.9522905917711195,0.4122666419410954,0.7995308375566329,0.6859222349176217,
 0.5613390876497536,0.6402195917436826,0.825233231516428,0.6283967603778433,
 0.24032059761118396,0.8897440804958119,0.5376925794220081,0.9465198545575152,
 0.35235164005931685,0.8614081921824475,0.6036027404086315,0.523466957960109,
 0.3711766077447155,0.4406543817953509,0.28117399654290787,0.758414337276273,
 0.9955453752793482,0.736109838387903,0.08927692732796344,0.8364250644516023,
 0.4815888751113905,0.9843969323479413,0.8181444877215388,0.32729261454219205,
 0.3336924137477226,0.5343373008202628,0.7291370691046941,0.25193660117717365,
 0.44658429820070555,0.88822306636738,0.3125911135570558,0.7978353990378152,
 0.3098907691117351,0.5724793811952942,0.6064981951725565,0.2938620503123379,
 0.6983969639184534,0.2516803182685211,0.5728814857407396,0.9770041464222669,
 0.4666095503735298,0.22574162221035665,0.054863977496366845,
 0.8438457276356219,0.8303234497888008,0.8006342763838746,0.9722739998785871,
 0.9634528152583203,0.003928909485963694,0.21612016169182158,
 0.18613177601981234,0.0237412707861564,0.37909121823466485,
 0.7642193479293743,0.12264262039220508,0.1128086832784293,0.7329994477096078,
 0.49217509154176864,0.45781420338086753,0.9189220969885946,
 0.7277655807037546,0.9239576991847902,0.07210202213231798,
 0.015601194206204871,0.356550628905535,0.9134984560290937,
 0.38513064469293845,0.6602390234649286,0.7927524640419673,
 0.10023202280351651,0.4998233558174675,0.774903364457303,0.5084437316269417,
 0.7504588879149026,0.9255370491075776,0.46449867778838005,0.5338765959549823,
 0.12223158267822887,0.4361302923653708,0.28963873473085044,
 0.9666538718190811,0.3645433067604543,0.8970078244628374,0.6077521252384046,
 0.7828220872787378,0.7696928177984929,0.5263241494316532,0.1171412131946521,
 0.32597200930413905,0.07653878498909605,0.7550398379497276,
 0.07255793256070309,0.37718144365147044,0.4857593951783836,
 0.3411737849939608,0.5200317711902075,0.6298933563546659,0.9251031466601809,
 0.8931393490167059,0.7591679125410404,0.8348666679685893,0.5671113826139232,
 0.43948223404233233,0.621814643168982,0.5573514247836189,0.5517284735276043,
 0.3461174538637807,0.3410571797206976,0.5392956379692517,0.8140675377520573,
 0.28989865278612426,0.9834329823131176,0.41413258377975093,
 0.5553252748877254,0.11917082762637987,0.8173066689294727,
 0.01027095897861785,0.031105987074041552,0.30980692717136504,
 0.21206691947394862,0.584347395145515,0.891107421336804,0.8055782858000546,
 0.5535856444967973,0.8724001081453614,0.4496474383600818,0.9369281581558838,
 0.554704728991407,0.15162513511227294,0.22805757856549558,0.3314998931457549,
 0.9262330720239651,0.3661051924715948,0.8646806374916558,0.7410284329853549,
 0.1120168256360482,0.7887346975854593,0.13939701362942292,0.8940667574507091,
 0.6270186029731021,0.5174827078995483,0.9288087808780842,0.9098694141417162,
 0.5375220971682376,0.8261775112240695,0.015777560306049376,
 0.9704107979911277,0.3274475416087359,0.36246269101680384,0.4219792930317423,
 0.004867746134828099,0.5301005049521598,0.4014960933860695,
 0.007791308755069615,0.3048658328369239,0.5931312037408181,
 0.5673906386174385,0.8622251063515535,0.8972674102953602,0.356183694611073,
 0.06849584111630114,0.979740786975517,0.4029711245003944,0.09605768412119453,
 0.021355256307987625,0.09069523739875118,0.6307217737634694,
 0.777409815390076,0.024208982044556904,0.244209236335291,0.1026632308215859,
 0.17101109829221317,0.3160384979592399,0.0628759128142321,
 0.09220434590572635,0.8516489045355541,0.3004245588280834,
 0.08820928144838347,0.2992202357432122,0.8084588715854645,
 0.19630990996199282,0.1441839943999984,0.6137823231502146,0.4790778901652374,
 0.49144401944929395,0.8010492168253515,0.9727603635421195,0.5954861202168884,
 0.7937489505532722,0.5679841271997752,0.674030388974992,0.25676106901285545,
 0.17121619421809153,0.5828379468836911,0.7239446625018219,0.312598321175119,
 0.9635462069597049,0.291015884295049,0.9900278794222309,0.7093282474141605,
 0.6652584174731843,0.7523596208342261,0.9386130912775776,0.42370666572042914,
 0.5978349932815209,0.5104483588660403,0.1331201808144692,0.5720998148662544,
 0.867864041641309,0.2583871965744775,0.4165108242708906,0.3888780262648456,
 0.6397906580314241,0.6527908097529789,0.17707325982698263,0.8675254336457231,
 0.3429112801678543,0.4689923174662669,0.516726810860552,0.2961518665749351,
 0.69836566521947,0.15296470641690407,0.03929603228301981,0.44185283921687524,
 0.8119840848445867,0.449577886722613,0.8055237013690765,0.5490024637553792,
 0.22588099807745587,0.5519644935670006,0.30579202967181396,
 0.44017585394078584,0.4855469609593537,0.6492721598665152,0.7221568702888722,
 0.11293631185017172,0.5978068915674417,0.49897131448843846,0.85384108944511,
 0.3853738679724955,0.032511991249620564,0.1387294956107925,
 0.7761782593372315,0.6998451105587845,0.20676831753319458,
 0.33787065683681594,0.26761921640721764,0.12115896134110427,
 0.9696136701173368,0.8335689156621892,0.04552049622197685,
 0.04074332650036716,0.9956470512731844,0.08405496996772999,
 0.35666120295617243,0.5505321883961038,0.7935282878052853,
 0.42177398781900277,0.7856210457913844,0.710153249544118,0.8354543063221834,
 0.7908930907371712,0.6617577939073627,0.03602828342142417,0.2514791131462063,
 0.7037890075110813,0.2950784796512531,0.9541035880972768,0.22462658347415165,
 0.9070246626159618,0.4658365587695348,0.5049046460280515,0.8021004122050207,
 0.7307926930391881,0.705860243369859,0.5925520763893435,0.5123262150991469,
 0.1099649132530146,0.7794058966629929,0.8320505868889819,0.11774361751869045,
 0.8830548378346088,0.9197008646242948,0.02077226237275842,0.5550460137263267,
 0.4942592941699683,0.6605828119045664,0.4626725923952948,0.32440272648737156,
 0.8390311670519021,0.3566217354641462,0.9535006312547689,0.3979199868657106,
 0.4758367589677994,0.27612798166384434,0.5118961860680761,0.5181550187938266,
 0.06628550195035343,0.4206067531559685,0.9660784708178707,0.7353828440441925,
 0.5655689173755659,0.31696422990799344,0.3915231166669002,0.991482488047649,
 0.11268780466874306,0.7143204395413322,0.49260474644993013,
 0.9853562693147951,0.9973810788111633,0.846674853876199,0.4877199229393562,
 0.3583528790072642,0.08478140944325174,0.050714627234150544,
 0.056770292326753014,0.6330196013813965,0.6476576881240118,
 0.5141928294150601,0.4845605561287434,0.16898918428923315,0.6060639311196282,
 0.6460938159868604,0.6571715057111851,0.9820176586641444,0.9502955264186337,
 0.34226367996334006,0.09762965900882303,0.6488210538574004,
 0.8956306519069861,0.05629289646797897,0.7451959855391179,0.6885780243299411,
 0.5323295712323597,0.6377206166745422,0.959999757235272,0.39042418888313424,
 0.2628785324066478,0.1152356123828574,0.7429423059974058,0.47390585752687164,
 0.28144140714642896,0.638224565651678,0.872028738379959,0.6286510810318914,
 0.7474847038381486,0.2978407740492999,0.7823789142701264,0.7544487254605223,
 0.5486085025710705,0.6751233955288408,0.6875404996208141,0.9621460372647977,
 0.3748241410070682,0.4809967839757291,0.7179580889316312,0.5115924955584106,
 0.4076750437597262,0.16699721014367253,0.2723120656912329,0.8388605727805235,
 0.5275920650014593,0.30527152083405007,0.47653093157716264,
 0.6631499225520942,0.40877540969594517,0.7866027961383549,0.8717291282650342,
 0.690034716333352,0.9533054167233164,0.47919200787936933,0.8885401504494325,
 0.05825713159737056,0.5738755410076228,0.6281453238816389,
 0.23253353168565627,0.9603358526395294,0.4015902125490447,0.2540866799634811,
 0.08129804350500858,0.7405358369103778,0.15716396544709155,
 0.17220765169509544,0.8463628704597423,0.206268926502025,0.3911220791123713,
 0.16032322832618617,0.7866878890145684,0.4556724284758421,0.8787110082209814,
 0.5468957294739323,0.44528293715356604,0.9156148067801165,
 0.05799844336944482,0.30390877261791127,0.9894688258224725,
 0.3823181383641481,0.4255819869128773,0.8605329101336084,0.9795607836815686,
 0.6086404789315193,0.781823781050726,0.2055512028399471,0.7048860163689228,
 0.5020451267922784,0.25069761657274114,0.8811616163886093,0.863667436751883,
 0.413591276057105,0.8210484495147125,0.8938612429008049,0.1716680257943256,
 0.35214581757658947,0.7428544047018316,0.47238842732800046,
 0.3495430617685966,0.9573184991001644,0.5074165915991781,0.9032207067522995,
 0.8477702042659723,0.5174336598628348,0.2230061651749946,0.6534386171617016,
 0.3787659113890066,0.4553658233804142,0.10237590487616677,0.9068780386458997,
 0.1833438403259009,0.3995835960128601,0.25654966937693524,0.9085033327619778,
 0.5934311521525742,0.0642691507441342,0.9543715081910754,
 0.004275229643073808,0.05700582605813609,0.19831261920318322,
 0.5526075482465997,0.8364478757365322,0.6965793755616345,0.39875593079914395,
 0.5693687259043042,0.1943658328223039,0.6301952304682952,0.619826209989408,
 0.9441384605606578,0.8275304745819658,0.5462573516663004,0.6307376502960312,
 0.23642735859867414,0.5719693514747809,0.6713511808503061,0.5610723879732397,
 0.7628118757573509,0.5460088859126593,0.11102049910806999,0.7926502539068206,
 0.2198926702415479,0.9142224131742639,0.40019107172092283,0.7952410064878689,
 0.1771411565635923,0.6272474882264548,0.9435971837042358,0.37170446762774967,
 0.0027982822109452465,0.40214533042810674,0.6482273693785049,
 0.11412164917025891,0.6250095537155732,0.9953783169480199,0.5078099552764165,
 0.5756340738605603,0.02807497537305892,0.3050481310649924,0.4398048971717815,
 0.3302779565357321,0.011249029890381124,0.5475672289917695,
 0.6143276687905811,0.8722244041500719,0.2209225106454321,0.10475609668645192,
 0.7783323310106849,0.6939595081621797,0.31560848645512674,
 0.36429813712655323,0.19407757267433912,0.10855195064499357,
 0.3968165297518218,0.9598581304677157,0.300693599198504,0.07594804694487411,
 0.19845252721360074,0.5734339059232452,0.6626647842645255,0.8005506607631238,
 0.06844312924173646,0.9300974814383335,0.9194361681062067,0.7726515203870526,
 0.564805628132926,0.8939884561494214,0.19670707116155584,0.24847320022828578,
 0.9245138162005109,0.11245625605726495,0.3576938779847525,0.942753159715321,
 0.33179632364435285,0.13659201288548073,0.2683575856666396,
 0.3535231900835949,0.38637181099560347,0.5424187378471667,0.2594853235316016,
 0.6407240513801186,0.9472067732510348,0.8135709698753081,0.18255524657073308,
 0.25589140962330403,0.7393639579696475,0.7912567235731185,0.8824620492287507,
 0.1646952820055455,0.03722598078880157,0.3223442674194634,
 0.19626629380303562,0.17772184005236447,0.9452601510071843,
 0.21024058772313914,0.5019060948836775,0.8793100834392199,
 0.013369588405311883,0.9867457905165534,0.29641058852811675,
 0.3425509667910468,0.1575210068493188,0.28644121885471197,0.7156629715963516,
 0.04714422602786539,0.6405611357469341,0.542238377853844,0.22645147894131,
 0.10429329714574098,0.6507069469588936,0.959809958740553,0.9126350836142663,
 0.6627822348051489,0.45802177012216816,0.8316288150080829,0.6403169896957237,
 0.33032368310022475,0.34003914156586623,0.7767453151888427,
 0.23432335819718342,0.44194142017406035,0.7857379361781536,
 0.7026689562701346,0.5620927657333665,0.6364026265593995,0.7311999057934242,
 0.30215919714446404,0.43887041737689503,0.3581250131722348,0.807608803405477,
 0.8567088917494247,0.9587880955375052,0.7785606194563963,0.19818223786524802,
 0.4306279148994081,0.47448722852613656,0.17076299008659768,
 0.45246533777960574,0.5805689958875033,0.4483919919671341,
 0.07978431912077266,0.08737797269599312,0.6984308348473051,
 0.04391966040947126,0.23371623479572146,0.3350279633964677,
 0.3780773157855708,0.48334627081659787,0.09856950604956616,
 0.8382132796050075,0.15795081241984943,0.03099079722665732,
 0.7535175179405826,0.9566081387452913,0.5808848320429847,0.04180277551168665,
 0.4113812648424888,0.9189153231325291,0.28747807724284513,0.8310687002035563,
 0.38532621412739787,0.1504931654865398,0.7102686038047411,0.8253518145181267,
 0.4660687081634671,0.15178175460628962,0.22699290734792066,
 0.014897655489427009,0.524622846931251,0.6024832410653326,
 0.42502703809451536,0.8877275618632602,0.6156398427040158,0.5281114957633373,
 0.46021580913160864,0.797069997268465,0.5483668482357605,0.4637756256412603,
 0.9916487552389367,0.05969496227180404,0.04898062251033797,
 0.33403124078602164,0.4248938263028271,0.27925234351304695,
 0.019670583236245598,0.27215615000778626,0.6467336795009081,
 0.664616637418098,0.07107517643521843,0.2626840079889702,0.39890897956553095,
 0.6994682004316504,0.4345755646013305,0.07935117095250277,0.2707652858065094,
 0.6695015963299967,0.08950661554320893,0.5606105129520748,0.3651297888826588,
 0.33350577674764736,0.12372550379711333,0.8253594438041552,
 0.4706593886593562,0.8007766400775995,0.9181133619204024,0.8122925780955442,
 0.4761800849858563,0.7428806005594224,0.5800948326890875,0.4287954785091278,
 0.6516524367162888,0.1702190477306671,0.32181669152086134,
 0.25601575542355925,0.2556389271435584,0.5034096465023528,
 0.12534557274046043,0.9768387889609279,0.872237766206866,0.7577326841857717,
 0.6684096816330013,0.9961901848765837,0.5608279984096666,0.14409307287823092,
 0.09229259231491493,0.4487402747836623,0.555259730583987,0.9980419467570056,
 0.646353166149379,0.40866316617889775,0.36667719968974555,0.8599663146273238,
 0.48864811714794054,0.7715124780177421,0.9054430108106991,
 0.22352272133920859,0.9543392228494115,0.6341790543673061,0.6547654843187024,
 0.9568005326270355,0.45141652314403535,0.019463789071181625,
 0.03981249622121952,0.6229538704488016,0.5102222537696456,0.779997637535909,
 0.428786532179693,0.9323319857838118,0.534804185451235,0.6016836515600701,
 0.9864780047787453,0.34766140986313576,0.008882793929888888,
 0.8749776637374259,0.5467280475332776,0.19363055365593262,0.7482371462125519,
 0.3217035745194107,0.5734363660600319,0.7892991139761222,0.27874529384702973,
 0.06069252569082484,0.318756715075436,0.5212983125079126,0.9507633246538988,
 0.4232531233247837,0.2708791947688828,0.7178460936513981,0.6383686189021304,
 0.34441274176141545,0.9345042516628921,0.9888990142362708,0.7610992442551827,
 0.3015793359787041,0.955685320510761,0.1820877318970785,0.4883658431871335,
 0.08769995410367049,0.16330553150898386,0.5191697329666645,
 0.9374177927860584,0.14885960982309265,0.5777484638388142,0.6772170202031311,
 0.28519187540055424,0.7803172732333168,0.24950169336909256,
 0.5917433602702392,0.9657675653005962,0.19747586924750993,
 0.14902910343213482,0.8389886497068967,0.8899448372134421,0.8711166397643497,
 0.2377500952635121,0.330485660960283,0.6590773036997881,0.20615242865787664,
 0.5666012091270888,0.9799664864719206,0.48282443168035727,0.7543787685958522,
 0.5506135775178035,0.8775062392698207,0.4496343130650977,0.8108026303445154,
 0.3874056806905426,0.5061416747785918,0.8250039025989745,0.9689485002351854,
 0.11552765919583496,0.8083937251034066,0.6612627062989374,
 0.16308218823412324,0.2644144279788013,0.5403165929162692,0.6387036760349616,
 0.020285499880139612,0.7606122821277993,0.6241627709224875,
 0.9833731667206433,0.30447788902935713,0.2866132323643561,0.9286454611872503,
 0.22830847232340679,0.3790940662763477,0.7550338526857717,0.656742924021248,
 0.05157810513396144,0.8782571849527202,0.426892060161118,0.3457344636798225,
 0.8240606471449803,0.40042819846001654,0.15014822858426247,
 0.3905498965171551,0.3045953235546792,0.9094883349155144,0.1581629241896233,
 0.9324734302662258,0.6243898805250522,0.15369213505319834,0.7263884375015776,
 0.24998184761342412,0.12107253327382117,0.8233675256830085,
 0.6100054348332441,0.5997659963010237,0.9467939212667515,0.46273133288403967,
 0.49697919265386437,0.47378726673568217,0.4198164633132072,
 0.043010630644900405,0.7899603959671277,0.6586451510755656,
 0.9564695548671688,0.5498701147897971,0.994874640750766,0.5213160356193013,
 0.2112822880821672,0.13646235114475158,0.19196473013240778,0.381398264234042,
 0.9994142287787589,0.746221006830836,0.27094773862842636,0.08568088201021595,
 0.2953021128011767,0.5412934437893365,0.05328149282291006,0.439039652131885,
 0.2517408776216843,0.6913549145655902,0.8992352652398867,0.0384008103703779,
 0.599031768822947,0.2544977673601301,0.1450523089288217,0.5080450341493997,
 0.8756222711629648,0.260140613149037,0.4965447270871597,0.19066106590135246,
 0.43972869542353366,0.9888840884328425,0.9383612715824144,0.1700576138479859,
 0.3582077191559,0.24027621252608367,0.516769585923822,0.7594458911027736,
 0.27246975088759684,0.7528199295048568,0.7774554973459465,0.3524561065524253,
 0.5165700289260817,0.36799863645953734,0.612503838980467,
 0.053866129358067005,0.5494774332222859,0.4291458066794751,
 0.9668338216123551,0.5754880324259427,0.7317034141690358,0.9731003849542801,
 0.383215576370183,0.6975136429840094,0.09924103328058642,0.7563973956960731,
 0.5204518461918022,0.10971708552244575,0.7664077444570134,0.8762743630914328,
 0.3634345450034233,0.3226912017682064,0.9659966824112406,
 0.007042183680922687,0.12400774771901424,0.71360887599429,
 0.17485752061556958,0.7382571794079471,0.18858011699813637,
 0.7710268596917884,0.15115788144274178,0.4384912994258563,
 0.37123777111837697,0.7481728481354101,0.10326002193831463,
 0.05800165082587139,0.027683083383303164,0.4891852936895844,
 0.6472119066177234,0.40097338308380126,0.93929838942609,0.9339703583127645,
 0.16465222232522692,0.9963828054756505,0.06453302220455759,
 0.1348104996354349,0.22591414260959974,0.17934038875779357,
 0.5004884665240642,0.2869090223073726,0.6326171296213453,0.9388821490773479,
 0.5973833933669859,0.28640108687467136,0.36445911478123505,
 0.05484079357531657,0.632012481354539,0.6050675493431672,0.632237250637131,
 0.08012171168876658,0.7259295348598167,0.7258986324886507,0.8334725853114633,
 0.36559798625245044,0.8058524811935385,0.9733221216271957,
 0.47262517286349237,0.7500534618927815,0.21266184038922853,
 0.3262247475712765,0.34067244898093296,0.7215938382946593,0.6278514201763642,
 0.9197989066035059,0.12729606768003343,0.44759318717289076,0.228778624838876,
 0.08830752111197482,0.007546639537582611,0.8999590877647059,
 0.014642807918297085,0.6150660705925243,0.5639780598962927,0.571170935690555,
 0.5935841136805977,0.599316251231305,0.6001664479495142,0.7147181937485526,
 0.7125365021072205,0.16748987061177123,0.08314810563194819,
 0.13680442329690512,0.42228104065530037,0.6563735122193033,
 0.9290732441505729,0.707315940064659,0.6292268125426275,0.6340945330938937,
 0.6818330752527345,0.39507296905680833,0.04798013250248767,
 0.7519961349783201,0.2508764005271541,0.5477176164959568,0.11692310120181115,
 0.37628376713006595,0.24711070652268052,0.5700014208371381,
 0.8797483226256934,0.1953877289032835,0.23273843225964952,0.6534183020229569,
 0.4224223912752749,0.26707679040077204,0.7395050590406638,0.4399629103279512,
 0.9144499240498449,0.7702362468902784,0.9653770079892677,0.7169306805121941,
 0.245866220526756,0.3436220128831796,0.292335872276847,0.9474084353548521,
 0.864283881352318,0.3042162182420114,0.689822935050316,0.02825121509818662,
 0.4312144859583631,0.6756683735215103,0.2559770887167041,0.9484905149325544,
 0.30340070617641934,0.2243504824935888,0.7793476008565111,0.3253716102974106,
 0.7967438053620409,0.5939310354242774,0.5238840747132725,0.4381286032089373,
 0.1962866859489043,0.22775850135000586,0.6572953371893694,
 0.03764581042699522,0.6297606722502478,0.6204390823767397,0.7394625637056889,
 0.07603310622750092,0.5878578268910182,0.7922169189877349,
 0.018987062974614166,0.7566341639497298,0.9748071626305936,
 0.1590873080922568,0.7394510772621627,0.796375665853005,0.569567908454759,
 0.31988119939256654,0.41758093940123253,0.9224329756932826,
 0.2864696101435351,0.48606422294117124,0.7917598361722691,0.6800765082052541,
 0.11075275447027955,0.8965929666904909,0.037639531271868476,
 0.6792883793335975,0.5530734309369975,0.7496877411228401,0.1820584200973997,
 0.9851034660171145,0.1792069661018425,0.8749347581334812,0.3008698694921461,
 0.23137226815688106,0.19189339486883328,0.5122484778065525,
 0.8554054806583337,0.2528918175163647,0.21167446010419555,
 0.47763741832869633,0.9044606285629178,0.527531153728102,0.619990941407138,
 0.5414169895843287,0.9842862730806495,0.44438702474419656,
 0.34884309541721614,0.659749908645054,0.5372981426366241,0.34918691654123535,
 0.9475521535314079,0.9866399370917235,0.11391779653291478,0.8276675164613613,
 0.4006139671073169,0.8708146847536196,0.12964560177094397,0.4377998078535532,
 0.7933651087515303,0.20017820249844032,0.13830231563382867,
 0.8864309818838808,0.044947041224361106,0.9937016921116647,
 0.9669571965282698,0.7363388467596007,0.40736769765634373,0.5301448882787091,
 0.21278667401372708,0.9103164694034411,0.3532183625400822,0.5367147222667709,
 0.8679988210260083,0.9868777831732748,0.1730574135962052,0.6511972871102575,
 0.05212307286668438,0.29807718979602016,0.7278256839245495,
 0.8227625244446138,0.9169739130119139,0.5293643048646938,0.2413545044066715,
 0.21699586933027604,0.6544527887421828,0.40989523183490595,0.541256871496334,
 0.5958270502590709,0.4133200851262415,0.5647020788905431,0.6134480360776141,
 0.8698302549256012,0.7160371267118391,0.941670648470354,0.6045215215504655,
 0.7569960988859603,0.13489604014026646,0.5899423637753922,0.856542170476181,
 0.41947843188376244,0.39645620628979605,0.5751325630377021,
 0.3090261909907839,0.2897051320067955,0.21575191229021773,0.5962011372829054,
 0.051769472879363065,0.5247381864102177,0.2685936341881976,
 0.4262736644833277,0.6494463116490751,0.11303268516183884,
 0.016482776916792674,0.26607750018052156,0.06981812440622426,
 0.6215208449230829,0.42759251738031245,0.733789869280858,0.5662302237629686,
 0.19824901176683574,0.9020450691489063,0.6312369092378745,0.8106338316601212,
 0.1429560909300499,0.07518314870409959,0.7011969368318959,0.3161911512469976,
 0.5733135109328489,0.7538943219651464,0.0022988662220425615,
 0.45108476909065975,0.3335560318867652,0.7232959629082157,0.9029291505052573,
 0.2831734237337382,0.0334130604023567,0.33670577052824935,0.535648772370392,
 0.8533126775584624,0.06255653081978796,0.20648096214509515,
 0.8574684593988285,0.6711672798844344,0.7312639144460791,0.9334919609944641,
 0.17450654943550548,0.6894729316647982,0.08178774861334248,
 0.9001737278661397,0.8898029978321523,0.9847680443669864,0.201364658630668,
 0.3508279894057037,0.4586845054576736,0.47767508298288763,0.6895207473133813,
 0.3188906793287284,0.4710571347651097,0.5133725890881402,0.44793233894326556,
 0.4696058055733634,0.42486034135269224,0.2808156843650512,0.4250757275724597,
 0.4501715762648879,0.40154360295306113,0.18475567614636645,
 0.20989184285636575,0.0777085772559466,0.6078726895285629,
 0.07580914721719456,0.1889904615908381,0.3906533454509037,
 0.38739228480299415,0.14828908621901882,0.2226388251174949,
 0.9710397366282648,0.8183991471764466,0.6750244833635087,0.9098487176633305,
 0.7667442952843571,0.1466584183662517,0.991843937843874,0.6905506854913166,
 0.1497817454087188,0.1599031549893427,0.15181703201468233,0.7154081798460293,
 0.2570689026421096,0.06744912850650187,0.2462991554447922,0.7805451233276386,
 0.7833586599115866,0.5195390880847113,0.654228012731762,0.8219593576944113,
 0.7231933446732572,0.9309484586657195,0.6178831979818434,0.6809261273281766,
 0.2669386237881226,0.8989986619513808,0.6570374245654511,0.5395821351924295,
 0.27582777090738064,0.9547889233840722,0.4420938925843657,0.5328860001848396,
 0.26300878546357154,0.19780263144233645,0.905741272419532,0.6958937649042696,
 0.42076537546945414,0.711534894104394,0.1889436091266261,0.8054533803875306,
 0.1676237051100723,0.7976249051874278,0.5018579763456222,0.7230569733098611,
 0.59730767448969,0.1862400056826976,0.1405870463949378,0.4820761652825304,
 0.6529988395182657,0.24535824795015027,0.5053369793364575,0.0686024678455075,
 0.7839660207664383,0.4495335872649088,0.14235693833767982,0.7543404174631503,
 0.1004230507906898,0.4091612539011221,0.8197352854559207,0.5723331957935595,
 0.411506951457838,0.8588360460551845,0.68503608845051,0.7898767491815727,
 0.04940386299108346,0.6373296468313503,0.5998862427988916,0.7068899666444246,
 0.5300648127224951,0.40914363267345044,0.7264110492515461,0.3658292176833724,
 0.8973530184277372,0.08332445660110355,0.6757402815695208,0.8772864725471676,
 0.25923491464730186,0.19091014354157032,0.8987141686214456,
 0.15161826090874808,0.06368301319175729,0.44818872377979346,
 0.5959585199249242,0.36295472327338807,0.555518600714283,0.23995237359752053,
 0.4187234061726245,0.6766954633447221,0.26041606626758607,0.6964786042174634,
 0.7053538263417551,0.6393722085658148,0.15497316923733295,0.7088750839180368,
 0.725310951646844,0.4642207032466136,0.7445770916858341,0.13763809044128428,
 0.09884728282009614,0.9417170587459706,0.0922933362779006,0.2889258214862598,
 0.8224502545527834,0.3721048914273225,0.5487769040012066,0.6758631142662953,
 0.47676987917069646,0.9268186959770761,0.7533329578405158,0.9156285835805792,
 0.48623907273497546,0.9585237556646452,0.08057390366736872,
 0.5966055500304581,0.016411759657702363,0.29724934430916605,
 0.010717303862657568,0.6665099706000265,0.35136197466293984,
 0.6272849069877854,0.01980825283355081,0.42036959991412104,
 0.4063822816130521,0.6563262072234474,0.2892958059041484,0.7949150275560316,
 0.8927631455530975,0.9841893678511439,0.2203518869023844,0.7929821330800347,
 0.9262826645796165,0.0741770172755567,0.09620298938098426,0.5040258960330779,
 0.9804863942570987,0.18814221918247087,0.05913135932892144,
 0.32273120400003685,0.2437644637623857,0.6715987627918932,0.6047318667998306,
 0.8906196562217562,0.7888349897297211,0.43911828867027625,
 0.20487752554375704,0.7049993185913552,0.1874177908022785,0.4240822986056849,
 0.05116409102783892,0.964808073233869,0.05675134057902165,
 0.44787452797044036,0.026550890054138354,0.5121863416220604,
 0.6487337841852683,0.6870831254167081,0.84323373075827,0.9191776529287439,
 0.4277430381682118,0.6321233902046173,0.855133593844953,0.060631757194698266,
 0.36629308783293957,0.7539198151412676,0.6772946371497642,0.6508781048094284,
 0.41847193415772477,0.44656525302230876,0.7753101101866796,
 0.4488383877852562,0.7158322401892695,0.3122264749174821,0.5661772738266022,
 0.27164594324617786,0.3185323637385833,0.2644005704747152,0.1312008345952087,
 0.18996003573048514,0.9048695367430593,0.3211075190425756,0.8016192260094246,
 0.22045002714851725,0.3329858272389208,0.7131579588763323,
 0.09465691040646729,0.08280517827980238,0.8361634889428211,
 0.8856956635498382,0.7948998166112751,0.5116138268901366,0.656277579755534,
 0.1620318334709273,0.011188414454067708,0.3959577607726822,
 0.5011143085416332,0.6458745758777987,0.5710095059470811,0.5669374623502114,
 0.3088101066229558,0.07655830478487191,0.28286042317939564,
 0.9035625659131259,0.634608596053017,0.9749632351707345,0.7404364015887691,
 0.2988528435191491,0.45323329212509667,0.4594995247677053,0.9828089712869263,
 0.9650003474172844,0.6363557088658661,0.7452303445076681,0.8865109396943802,
 0.08838628454150688,0.5134335595673952,0.7793327172093514,0.1823172786036954,
 0.3533283673098394,0.13806704403585035,0.09977018163432128,
 0.21513040532326255,0.15042182022930528,0.7246448217700364,
 0.32418134772330764,0.2776971112477792,0.8114057187080919,0.7647137933546772,
 0.04609325315004631,0.8859495725896467,0.15150360507098137,
 0.08437275221597507,0.48339530569828715,0.04376083190915914,
 0.8466202851763037,0.7739099329291623,0.08103792715639868,
 0.41071451277055493,0.10016313390811782,0.6354906895601345,
 0.9232352575507918,0.4429299875256001,0.5026368051799589,0.25755108781532987,
 0.07501608641505764,0.49029539777614806,0.060157267992084096,
 0.1855938487114327,0.34522476941063496,0.6327643129690268,0.9679407416049289,
 0.00899166093077608,0.1124415158468266,0.670711865952508,0.21138623141017898,
 0.6672694227341007,0.04343724378752578,0.8838727604568036,0.9770345383688465,
 0.09086142089671012,0.364896516452483,0.3122275464971779,0.8593203955881662,
 0.26965420823050334,0.7359968799249861,0.14581175777186062,
 0.3547994637590056,0.5604989217582468,0.569975757467359,0.3463841600742369,
 0.4849297783493616,0.01302273331110504,0.12069198356051336,
 0.5560362755630727,0.4312910293677711,0.3147177700307995,0.28002093428873254,
 0.2350773624941095,0.5886216315456791,0.5421987828050361,0.08740859012608859,
 0.34272575883295353,0.35332826153069097,0.37140358853868316,
 0.3961008255639946,0.16361482714869902,0.6482500727177662,
 0.43748275066610876,0.22578862710807635,0.4696917301775142,
 0.09704251482473836,0.9916619188709881,0.34787249704405465,
 0.9214442145018389,0.25123302557317406,0.5543591061237536,0.6309565290673113,
 0.5333976332957914,0.7489026907668611,0.3783005085332458,0.33251882275503775,
 0.7560999245173696,0.6713360064307858,0.1133531331461427,0.500268440564439,
 0.5014900263364643,0.31404092216499213,0.33075475665686715,
 0.12999791088930246,0.21328919993288675,0.8325389058849265,
 0.6883070326420689,0.20687466613967342,0.07392776580660365,
 0.9524868941850321,0.9662913349282192,0.08099551746326916,0.7791138133295672,
 0.8054150506841791,0.8527660638594008,0.523033080003009,0.12415086903522021,
 0.32746330058541984,0.9675646514165219,0.10254429910570195,
 0.12879793076676205,0.5257880373828163,0.7578916548281275,0.3268216416462668,
 0.1406230412151689,0.9869912297964247,0.6429828764101047,0.10187314228655964,
 0.298245339767395,0.9714317246375759,0.5843494878960287,0.8467883326070489,
 0.19383956522939882,0.6674895551073454,0.28006482460109106,0.669838948626601,
 0.08200708117739453,0.3347960117446427,0.7769785518974974,
 0.12698759180514796,0.44304561887642,0.7389890914563522,0.09654185063154086,
 0.9921854010986387,0.8397399357148823,0.6338520249445972,0.8768219780465469,
 0.8469331899485941,0.6983491561893893,0.1387422142062027,0.45483046570238517,
 0.3512571432054832,0.6362001723451495,0.5652688670980712,0.30742678230737863,
 0.031699222805362526,0.16759944712659725,0.809631484611945,
 0.45142134017751534,0.9102796607269354,0.23069533135236986,
 0.8033516822562317,0.2646702672690261,0.24243174735766915,0.9002234610638387,
 0.6382100132104623,0.7010443023766899,0.3656779458346632,0.5731334223246992,
 0.18046189437433424,0.05613349787364963,0.8424780658434788,
 0.9528255479808077,0.8393058499107481,0.003548985162310947,
 0.3815544764043476,0.34659613917011034,0.049902606366731606,
 0.03242107907493896,0.4467889493415804,0.10009148901468179,
 0.8325391928182005,0.11427210581018443,0.3060013578742481,0.7560798131315265,
 0.2912325048619191,0.20579436157403097,0.9513174521704425,0.2448598199348896,
 0.5854200604149236,0.5779146261754713,0.7750084527877894,0.6482334044032314,
 0.9133414942468521,0.4793959072147045,0.4510343149877014,0.6915914789971341,
 0.21492346703994047,0.801266802449786,0.4879624668597614,0.2104952361333099,
 0.6798885890208181,0.48078682990789234,0.9408608596245718,0.8957969674568832,
 0.8777314112988253,0.10668793297652301,0.03182350393132838,
 0.6159629106330677,0.45596902120158955,0.10413839729662155,
 0.9482550918200634,0.5443448572008877,0.9972078906560149,0.18288244806159804,
 0.8205043336951505,0.10584150913576051,0.23789303551264807,
 0.8889823866195449,0.2674955539748931,0.31930190235779243,0.9058207373353823,
 0.9478733515037032,0.12886776828682078,0.28560308938056944,
 0.19354661919912186,0.5328241345984174,0.17530572293800795,0.68806286298806,
 0.035889444677975946,0.06529379159027249,0.1100001071069392,
 0.34293387665487773,0.5842740634585349,0.7269351724491485,
 0.026706412762146026,0.204403338795333,0.7522809665546227,0.6553120608049671,
 0.028191571156208184,0.6906415803165783,0.999110335105862,0.8902110415242155,
 0.6112569074300561,0.14943021798212397,0.11200285314523839,
 0.8828700689179242,0.848205525239606,0.29705561349454157,0.8423805537096434,
 0.3236918349229003,0.5289096822448345,0.9689326144976931,0.9144039515488938,
 0.6885587096176693,0.23061337527696335,0.8575805640048805,
 0.45324710928427325,0.7118482696701078,0.4597380116417815,
 0.20800479833499574,0.8534774927604246,0.11184087726280278,
 0.9407348225411931,0.5602885454925,0.1936994205926863,0.5246479015754495,
 0.4170159179681254,0.7458749069269215,0.1011640777017474,0.8644350852688183,
 0.5805539365238777,0.34885651731938416,0.9872074887613588,0.9118288298060824,
 0.32687934051845347,0.7720627007079639,0.16092246311999714,
 0.15699753358029156,0.39587718042849673,0.5305081406912566,
 0.7355953062816329,0.37526794629211224,0.6060534812180745,0.7725503938967756,
 0.7396383471180203,0.4786786281264763,0.6361268825783833,0.7395613500107454,
 0.34719585838952516,0.5682142103216692,0.22396421241930542,
 0.30620965666298783,0.4136815995379065,0.8842171768157772,0.9143125893247196,
 0.7099498232753851,0.38686033893420435,0.1738549979621603,
 0.35456716641390873,0.781656467946312,0.8149910306331568,0.2517777691699221,
 0.37556907299593956,0.027781243118073062,0.8818592726973712,
 0.4037420275697192,0.5005831479636124,0.4687973935852249,0.08308487875636106,
 0.14200427241729074,0.2707678209687677,0.13471831116082755,
 0.13035921573947817,0.12067151912826135,0.08439444214743042,
 0.037274743974710645,0.2355873715653548,0.9601025869444293,
 0.7256228297335513,0.8063051613509657,0.39891011915736474,0.709839994435229,
 0.28494347545491705,0.9594847334391279,0.2345489763370604,
 0.46561641376609253,0.7360280351710757,0.40484066834487675,
 0.7379433765775427,0.18452107192161393,0.15996486020164857,
 0.3396624856554087,0.08091675814100596,0.3051383540897449,0.633521625140719,
 0.027696069600720463,0.16061949091319028,0.7149617475265179,
 0.7394799799559507,0.3058124815937193,0.1030877622373807,0.636587394521999,
 0.602925691452302,0.365224437691404,0.7711852489950434,0.20971538430649983,
 0.5104985704515473,0.020430388065888816,0.6354792738341258,
 0.19173229884397003,0.2855436537123175,0.20894858898452495,
 0.7124191064377827,0.5967159191821299,0.9369708024284781,0.5521109306397282,
 0.880697692789015,0.10063585061351887,0.5877038642252226,0.2725354701122693,
 0.9788876616002437,0.9539227339609208,0.7614681887854071,0.3901194701756292,
 0.3788647039296509,0.1327771030680619,0.47122594478080226,0.719129841882532,
 0.09082113411972137,4.5327772154513113e-4,0.24183220867609667,
 0.5204543892199054,0.8528433953751067,0.7363858111886922,0.8165032050036661,
 0.2507718230879249,0.31368485374657507,0.40934463169194035,
 0.6117793471774007,0.640998949681937,0.1995880284055499,0.5500742116999622,
 0.1817269688675568,0.8404547927988807,0.6357854771781105,0.2738707813116421,
 0.4133189416093467,0.15607018607585998,0.4924142656815561,
 0.013631489828226151,0.769101428455005,0.8841014240316087,0.8594518942579688,
 0.9766156469823188,0.2902223188352514,0.8726290048091587,0.6136166022884119,
 0.2333832338201146,0.6097824925814294,0.46867890225283637,
 0.23676214556037123,0.1276071603840332,0.5007404885178561,0.097149160409981,
 0.5554780451469837,0.7306938419050892,0.2856105424016764,0.23175356446053408,
 0.5507507797414651,0.6350551654142191,0.19771067346002702,0.1532141705527903,
 0.6106523632482105,0.6887705123623324,0.20385060178736414,0.3375162915811155,
 0.2565493949328074,0.8402653317185217,0.8073506830197622,0.8311092742610261,
 0.5599182399176503,0.45653185711521194,0.8648401479119572,0.839778087246523,
 0.12361168110126342,0.38549698673833255,0.27547831346148777,
 0.9584307609468409,0.3381645936506241,0.3656204851748942,0.8706754173386779,
 0.48696861809969605,0.16289862908913855,0.38588005920649204,
 0.7829474957277135,0.7038666605681517,0.30350919128790177,0.8623304085235151,
 0.6395527793895702,0.9838191040504533,0.9011062100494682,0.5812221528776805,
 0.9786611732810998,0.9360653938276579,0.45260164736898434,0.5308516187052961,
 0.7911094046189712,0.7330613075531968,0.05226396990462279,0.8028156617241071,
 0.4748474628048045,0.2986989336222876,0.7088084580697098,0.21719897057125792,
 0.7949368369266864,0.5949140253510172,0.18338263937847,0.9224374698283573,
 0.0884135719971414,0.08271228640250139,0.24521969689833079,
 0.8691043798197441,0.5015963347601327,0.5208547422507346,0.6562707822708791,
 0.6713321665280215,0.7653194552550887,0.5564362490012156,0.9388621631069929,
 0.6097981415173078,0.8135614415492547,0.01666645262920663,
 0.011085578361832438,0.9980694081797363,0.26569213097181366,
 0.012082525519850673,0.6886430167121775,0.4964363057905582,
 0.06312100080179639,0.022607164926000456,0.10567468511296185,
 0.5710014942267861,0.29507790467619943,0.31783683344876723,
 0.41681345269038306,0.7405479172695111,0.5752512194047272,
 0.21461465605197771,0.7106507950591775,0.7618466548206785,0.6324904073062215,
 0.1355751274500142,0.3656732339287837,0.6123524314549709,0.24557644390045164,
 0.5221562397487881,0.3376856497474594,0.5921094705050985,0.9875149487353793,
 0.8118859191731538,0.5049167152101067,0.9389752600625111,0.28298208144939174,
 0.035298006031872164,0.9672965567857219,0.5377236048273925,
 0.6016986599398859,0.7253979851906383,0.42301044389503084,0.64352804357329,
 0.7062692495122975,0.4234387563174582,0.7913984227337925,0.2620904707888413,
 0.5513376430257345,0.1477796569875386,0.47196326066476446,
 0.12135706571377369,0.37789977933311525,0.3530495761224225,
 0.6464585351105462,0.40959396453305974,0.3828016643200053,0.7594682693933451,
 0.0796205959674533,0.7669703135799659,0.1230705775047739,0.7028040084555685,
 0.8639344203087047,0.5452678354803242,0.6083485744511367,0.5843697160461577,
 0.8910090927796108,0.41597202915726716,0.5817790100982314,0.7820065638749595,
 0.3594731820127942,0.6864957100641876,0.9001901008073726,0.7372816750586366,
 0.05645326052100086,0.06042878024173892,0.3946077998232549,0.294820436359978,
 0.7170747346978745,0.6740371398313925,0.024834849068780862,
 0.16429229146071855,0.8204749331213163,0.09970430415869547,
 0.6885181704124043,0.4617571605514237,0.7529009350654023,0.6758204339909488,
 0.5246560891806673,0.5080644027997522,0.13195447127189563,0.7363406131906943,
 0.6352618673966172,0.7333120801913501,0.21901022518703428,0.7499610644446467,
 0.738252659242984,0.7297773052585356,0.713170605768629,0.7825305656868157,
 0.18494659718603557,0.5130291922988998,0.7230839843334044,0.5390324909352555,
 0.5950864996032592,0.18002649591721076,0.9927076621549838,
 0.021252870594885898,0.7383523791409948,0.5455255193129134,
 0.5653191970004934,0.21991396796182094,0.6653339234934134,0.7393133002764068,
 0.505122874913989,0.08664484146969642,0.08317371814198726,0.6196890674556425,
 0.8650491476485674,0.8474291437631721,0.8406528477686311,0.3770703130512262,
 0.6259050289285981,0.6403099277654257,0.5711901969918821,0.26250377405590863,
 0.44134010506474675,0.4558213745636852,0.6449149593828728,0.6296193244252654,
 0.1996389861497948,0.9505879055408588,0.3090351256650252,0.1899653679713872,
 0.8128696391201106,0.33293954267432657,0.6251481960746316,0.8679435844685698,
 0.18060135322837856,0.46556306479825793,0.283921587386325,0.674852820796529,
 0.6495142712481541,0.8890296195798321,0.9885012271322943,0.9789302144946069,
 0.08990013990159074,0.0061735374862724735,0.6093508656139986,
 0.006876961837718132,0.08770063072429102,0.44713015886964147,
 0.5697020587514068,0.7082309798064319,0.1597418978711752,0.48913673662306534,
 0.4962088247333942,0.6783123624201607,0.9088310800851405,0.4299565688440731,
 0.7167358577952665,0.09861531773676524,0.8032134344835506,0.8985840464113658,
 0.666563860022,0.5413104160809619,0.2588455836473301,0.7461387686978488,
 0.4366465230603276,0.13360409778458548,0.9280756226280386,
 0.46634625575044386,0.5702391525383363,0.9875135150660617,0.8058658241360158,
 0.7897411205860978,0.01572909109245957,0.11074098442711056,
 0.17260069874623196,0.5508060036369131,0.5430951584423653,
 0.02563309925624646,0.878144459269226,0.7078340775026637,0.9578761984494091,
 0.5876406487596298,0.34869441964424297,0.3506115216942517,0.782089230353832,
 0.745917849003907,0.5453428713011007,0.4053666483099818,0.6539209231381862,
 0.23405303328916216,0.1428838980680639,0.6308955896875043,0.8824109655256758,
 0.3714588855278963,0.9903879588750608,0.967879245888662,0.8924724782337863,
 0.43449333950625946,0.410923039335897,0.9284532630385302,0.6217227651739989,
 0.5180676587651503,0.39992328864067594,0.3770211296596794,
 0.13569657926888956,0.28075663759776737,0.9867015311233087,
 0.6025106254676914,0.12432139105291962,0.4189778910375639,0.6104153476234814,
 0.8255488463058608,0.420074277042529,0.9920052244086123,0.04513740033280378,
 0.09663863976281761,0.3189922089694579,0.06085786186750519,
 0.7461451741235587,0.35917330398054925,0.5335532157245321,0.1101301716406452,
 0.8855266186372761,0.8217275534741324,0.832304356927654,0.8371402833354351,
 0.7303842856452165,0.10552305688965524,0.021802349958387368,
 0.8840436316836129,0.9323190174405583,0.08698021350488527,
 0.48158430692529186,0.704909625889763,0.34717228049573634,
 0.19894610883012764,0.6718666120439095,0.9894121456567101,0.9805843258559449,
 0.586462743862425,0.41235138195968035,0.9062221934398206,0.7023385641413697,
 0.8448899370147933,0.4166119153742446,0.3433519567246466,0.22866469488328944,
 0.6676079076764947,0.019113413124923095,0.9618366316694692,
 0.6600628444062339,0.5820699288680496,0.46747262499624553,0.9589614058805733,
 0.489769796705387,0.19023302314622936,0.6950421305886076,0.1576983848013408,
 0.5922918849718354,0.04550997017457914,0.32535890666482503,0.324399432248347,
 0.6729171340170083,0.15922082702224016,0.9008274065655788,0.5599909352212424,
 0.6236721630807933,0.6487939288151272,0.4225472616346201,0.0498011054027383,
 0.016267702833440145,0.55374049229765,0.5528925164708028,0.6456189131295683,
 0.6389497967705242,0.6255824619467228,0.8387738553748497,0.3772229676303692,
 0.5938806956734878,0.8685597402480189,0.31413069072179256,
 0.20516792733403422,0.24927460111026845,0.8347312854602835,
 0.007244700336291432,0.08205633400792012,0.7959982796582707,
 0.2123321926541235,0.15893137564132964,0.7012787931281271,
 0.35430497069904776,0.6779409679441649,0.47456863181086617,
 0.2776971466840963,0.4187687464623483,0.07961356847852397,0.5839712011043041,
 0.26721245098803037,0.29143707020322496,0.044484799420599574,
 0.11672044674154047,0.6794152506996218,0.5683535800938961,0.9732207214448237,
 0.6334155002590874,0.512545143615031,0.9786232432695819,0.7379890956301207,
 0.9596638666379046,0.7133653408278129,0.7177997945509245,0.6783829777357945,
 0.3757401918515574,0.5572338540637422,0.5940225269705195,0.13831333537673185,
 0.7310847118407305,0.6034795552119944,0.47272869013001695,
 0.49821024283990545,0.24067239031260046,0.8896529592786608,
 0.0032056502088669836,0.629892424622101,0.4341591718941672,
 0.7424037032547555,0.18042107320561163,0.5777953539884929,0.955473383550565,
 0.8011504749684378,0.1686358734344393,0.7855144975105639,0.203338514889988,
 0.39082583267041937,0.8891473378010085,0.10547298895337875,
 0.06011366821782038,0.8989588140259948,0.15317221069526787,
 0.43499766650213323,0.21808661566616405,0.07354211321735138,
 0.6477512752489094,0.8733499981727254,0.6151586610303948,0.4357468219893529,
 0.3170182225932955,0.5107587252134149,0.4848920144886819,0.6993963423408047,
 0.599472264379671,0.050943542807228126,0.7546251455615077,0.6870904517882048,
 0.690296450415732,0.7937877407651819,0.09624430280851959,0.6300210011614723,
 0.21717884908125673,0.3905543443996784,0.429520226085933,0.6300276932146788,
 0.4083131306561011,0.4745749125546672,0.982148100050749,0.37781109820158465,
 0.5637830372087197,0.052291790196818844,0.41650934073591495,
 0.6004031292812548,0.9091855689541862,0.26918025780362,0.6679900762003496,
 0.09830607921063894,0.4089624894005903,0.31954738849094877,
 0.43774768684842036,0.32130662846726865,0.736725131182477,
 0.38698466818068045,0.13348345427781194,0.39445354375593167,
 0.8914215954826294,0.40576523154771993,0.7902029816216884,0.0599352018648579,
 0.7751814604864439,0.4493512775144237,0.9727776496890702,0.22046264818595596,
 0.8408355258858662,0.37139797028354893,0.03388055338928475,
 0.49293244705142947,0.5549252417065317,0.7092645879708112,
 0.02841215993937163,0.8584696068270347,0.20877027312328944,
 0.2837833034955376,0.14470701285565513,0.408868851476901,0.8823477868092909,
 0.6463822677160274,0.8114036580695786,0.7848878950990387,0.6935931912591361,
 0.8371732159688707,0.5372907320888337,0.7429705060123447,0.8206137919297853,
 0.673592411353187,0.9962138211954853,0.29996627254866814,0.4559529259145776,
 0.12862745366219397,0.060959773235717885,0.4627914069949397,
 0.2112381866507833,0.7510531330765091,0.026703648682018577,
 0.3080976669849925,0.5545176500358076,0.194967709803064,0.17709513777407204,
 0.6826503538509571,0.19335001735391466,0.09033608709247665,
 0.42705577669250305,0.6519698922255435,0.5236056875719666,0.8936206018159569,
 0.5373523744908522,0.4356047250282765,0.38124395389376564,0.9213082040646232,
 0.303473836675464,0.34595715780639513,0.7131834113221194,0.5662735998930247,
 0.2643352604818051,0.8540924062109595,0.47886423607724315,0.3943178434571307,
 0.8428820128954639,0.26495002760115915,0.1640430805198595,
 0.21274697103527618,0.4451645594913658,0.526942411568288,0.6066916334983148,
 0.647175432775862,0.6997957045941327,0.2691445071478905,0.7364723104781942,
 0.4216283382489934,0.11135232521135985,0.20280154960597785,
 0.2043731064046972,0.9902715751913445,0.9624608684458473,0.20415155226926074,
 0.45036709954639864,7.55285060679789e-4,0.9562134410814321,
 0.6412070557136399,0.05340108108745478,0.4818209910938904,0.3655036971259492,
 0.17841854395059198,0.10666095514361351,0.3509600306057499,
 0.10752243578488685,0.855568531819037,0.7328682693454424,0.7491423066878105,
 0.5595954657008885,0.6481542663917644,0.09712097418070387,0.6243729057625313,
 0.38569367823785283,0.22784770413136446,0.9583750471825373,
 0.4619231684889503,0.7866444068464007,0.5189664735819496,0.574534080710624,
 0.9755032469747176,0.5963288743395405,0.9797840065685188,0.8112279666168406,
 0.22975685900127862,0.29257062573470305,0.291524843554254,0.5364897528515299,
 0.24927927805255012,0.7624686176165223,0.2144773445301249,0.3650491549533865,
 0.5333830074627619,0.11877157470021016,0.6180859115905262,
 0.09128843510648854,0.686727135299186,0.9032057682242446,0.4258186143257081,
 0.4225053939210077,0.8506646355095849,0.6640687485042014,0.46215293440754646,
 0.8524508647594717,0.13410501833447208,0.6846246609333342,
 0.45553550470025617,0.76476300608274,0.6368249251951497,0.7477801399537385,
 0.4941806078068365,0.45872235444588016,0.1566149384711103,
 0.05302145796681956,0.2708244478857669,0.4242947885874291,0.4179724974192116,
 0.6215336666863865,0.5036745289190055,0.9825020775902388,0.9732178557385132,
 0.6102760467289283,0.8284830785558994,0.7306021114694249,0.7792039391516044,
 0.6229650320802146,0.1388478245789816,0.8198461573531305,0.9091707742638956,
 0.13078733470535675,0.9932466892284086,0.7548460338203057,
 0.16505641839079643,0.5693373040605053,0.5716916174669299,0.6594216165693783,
 0.10528783476911285,0.3215825645656051,0.2049784139238311,
 0.41375403905196084,0.6407936637357736,0.48254846612925595,
 0.42440245536325694,0.7794386614168406,0.9985512570989074,
 0.40953730286977497,0.21183208887828697,0.3440953097756444,
 0.8610124213375218,0.8122150021424662,0.30450744706352906,0.6098778697992014,
 0.41774217940466274,0.13798626091396915,0.09929454901189239,
 0.8868767683615604,0.05090843981371318,0.3150248733133204,0.2544337426272226,
 0.40238461723595464,0.8328914534760539,0.8457344695341088,0.1290869014306566,
 0.8143554409205633,0.3459438830424597,0.2649806105830609,0.4514829368567692,
 0.589714674440035,0.9036630933590463,0.9644835734255717,0.28100050770160423,
 0.6505103586692608,0.4092452739727006,0.885239673063609,0.06313158384135353,
 0.014519259157664943,0.6178941322635315,0.6126906081043026,
 0.7553047812276046,0.6158058162083572,0.1731693458168606,0.4121612585936164,
 0.9939283489209064,0.011556695475510859,0.018447702545185907,
 0.7316527374008591,0.729389574146114,0.9832448418394633,0.2103128267824197,
 0.13168309221450192,0.21538476123855455,0.3754570629465891,
 0.7196027949357562,0.41213330622495514,0.7907725504978517,
 0.24785376890275213,0.7118299206721783,0.7580615655980875,0.8547990354957943,
 0.0652823393153511,0.6743965779205209,0.048386661058476554,
 0.5048646975471024,0.8254002425865558,0.3294778253237497,0.2790900195874737,
 0.5335862300075669,0.35636786989049485,0.444281849132513,0.19402904934407927,
 0.9414540529993765,0.6545128441045198,0.17733122003388624,0.7975192731214042,
 0.14286489787065582,0.4722943554197798,0.36914187199367954,
 0.06214607224055568,0.016078844315041918,0.43122773024910743,
 0.7810007936931702,0.23318105347904605,0.8438248607078822,0.9601518366134256,
 0.2954910244115374,0.5775204073506384,0.6966732040002885,0.707865464755161,
 0.40296500075078634,0.7112118294988281,0.592310436703537,0.5822510814154115,
 0.20921357014584296,0.02221586411660048,0.05821846752992488,
 0.35166987435852504,0.23888854236106527,0.6958949777119787,
 0.7551874516389304,0.23152962691956236,0.21025254699559104,
 0.4970461012044721,0.010912989508216508,0.03940740727259784,
 0.1441047993236907,0.27494777949306703,0.659441284459201,0.15449477825268743,
 0.5324386238880129,0.6125327697670836,0.11332746193867371,0.7640232627441125,
 0.09068614940040276,0.9659051231149374,0.4742492401956073,
 0.34612067621363185,0.04362006808071339,0.41531585857767084,
 0.01232376629828913,0.05207332216166893,0.30597338447459554,
 0.029820670459721366,0.40600020347193644,0.01232096208803135,
 0.1336711899128591,0.9002580458875193,0.9085090928733177,0.6827687728586155,
 0.7280270269849605,0.5167088972268092,0.011528056812738607,
 0.46274137717091846,0.7341176830496966,0.35651841009018514,
 0.28279240143467776,0.11679147181591665,0.5079672462953413,
 0.09245125562832868,0.5920603138053577,0.14904403471673489,
 0.7216185980879131,0.003993000397370605,0.9616009393319672,
 0.22930591782648257,0.08053966596654683,0.405125599179216,0.1724977347030865,
 0.4401524782820201,0.27300647088040453,0.7284812392980228,0.9533027399383236,
 0.5705509599983118,0.4576545780872724,0.24407139359540264,
 0.44257631827627975,0.2674780550254433,0.2991029639627274,0.0421236628315681,
 0.03143202091139541,0.8880103384492646,0.7619663239602276,
 0.37917464373360865,0.42022275439441714,0.11739688636166368,
 0.4387465513762394,0.5212423524130916,0.3213464457879093,0.7571180332427508,
 0.018373501116025315,0.8478203544722032,0.3798404161124277,
 0.12960218692172598,0.6481387543179674,0.30677355337230783,0.213605108627399,
 0.008655299582201681,0.45643062346122176,0.17156973660119146,
 0.33956229692226336,0.6917623730638305,0.5062490979105212,0.4555228439360608,
 0.7821949514195314,0.39293647123192854,0.9979812801543427,
 0.06581341849663569,0.5215047771905479,0.5820900662237283,0.3232681647756075,
 0.6202958306505814,0.38345415302402386,0.2685608541241291,0.3042125631949577,
 0.6034394449448884,0.9333855752646116,0.4093367479844815,0.43995143852825724,
 0.8249937964679942,0.7739717055276493,0.1970483663099175,0.2842931528335081,
 0.48994004040084516,0.4631491919518962,0.49599073029535057,
 0.11748433156220517,0.2671671596631867,0.7020898936313069,
 0.07219010400536152,0.3599585155567,0.4579276025184227,0.8749020879726338,
 0.699953745441898,0.2426150658886952,0.2643742981996424,0.22053293313242295,
 0.9505942404522857,0.5445352198562007,0.7231219906708223,0.29386251397055885,
 0.5971210994658249,0.8882427308074426,0.8963260465084496,0.34274945737544815,
 0.10504192625107678,0.008668505800371262,0.6486723598195647,
 0.8332043044238875,0.03412668894838655,0.16399787399226295,
 0.6396888855721791,0.3457130470136942,0.31781297142288256,0.2926714688040415,
 0.48664298022401853,0.8514697921535039,0.15450641795878872,
 0.37848407236202497,0.4023346409707187,0.2555373722761335,0.2617180625924823,
 0.08808978842897985,0.9352119145700638,0.652313041343679,0.3988636587154476,
 0.2253290496753403,0.11238906796858272,0.6094803522516107,0.167455006714889,
 0.7474797714546532,0.16347605225735196,0.771133235650816,0.5496267848660868,
 0.30502363166881463,0.01814953010450482,0.6955912743212362,
 0.9954306697191384,0.8721601655249431,0.19775690313207017,
 0.21056105110364154,0.8872902762251462,0.0193753568293209,
 0.14840532566592346,0.64016447951837,0.43466295044743997,0.08669122358343717,
 0.04579422431577962,0.603559081982066,0.4636660307476069,0.6531396996783154,
 0.5253614191398291,0.8920573762108934,0.510591163196469,0.8243614329043412,
 0.6917268327324372,0.6003631570791751,0.7310512768141602,0.34896499784468726,
 0.29623480913969513,0.4083785132179597,0.8415860228263653,0.856892067162715,
 0.9498988172070744,0.7144776161670839,0.46180711546697784,
 0.23328515669027416,0.6706380018032633,0.030535312175658746,
 0.1555263598804012,0.3555118268595203,0.9273577204072029,0.8107891756187633,
 0.15056022205022412,0.7476685373733217,0.9333917793757154,0.3296619367646081,
 0.47268052251040105,0.4072363408937959,0.11799577132504102,0.507288357859149,
 0.20559437182954532,0.7001690721615849,0.8250433756704288,
 0.46398352534695375,0.5025694752739752,0.044713595652197125,
 0.15637225809967858,0.7122213049690412,0.995927886850926,0.7166556601836065,
 0.5526236201236526,0.3489991992565261,0.27127409541264225,0.6635630043827483,
 0.2766977138397684,0.865893464913533,0.38594656628619384,0.8174787632782353,
 0.9606214556901158,0.27327409655449353,0.6744339605509031,
 0.009002304719481935,0.913992603528431,0.6015713186136712,0.6577607335110643,
 0.18660193026286964,0.07506851141839199,0.6140843387382939,
 0.5351750131643093,0.7144507905412442,0.7781667026267537,0.32782249849838496,
 0.13639332089038902,0.5354148122804909,0.10454339600826834,
 0.5197761976300388,0.1551186439132598,0.3485746527284155,0.9145253697078279,
 0.8753841970912281,0.4178398579781628,0.2408754617338258,0.25969336035490564,
 0.7117428774593562,0.2955703783898801,0.7432248281625731,0.9948886149808527,
 0.4596057721451051,0.1316148938938123,0.682210329208054,0.9667852545846761,
 0.6976485263431693,0.2202327623215251,0.9739704563530133,0.26812735047126746,
 0.18445294900735143,0.19788494800569567,0.36790703468607455,
 0.8627432774226224,0.4581058781737253,0.44226601327320925,0.534661742958286,
 0.6881087115282427,0.5374829285900042,0.2009200395068217,0.4510195537558185,
 0.323303233730057,0.4978215364829046,0.7103124182003296,0.6482103672222624,
 0.08720562515849606,0.21873064789083152,0.25882069582760203,
 0.12514515470165366,0.08832526275410535,0.5363166734028695,0.769791235693809,
 0.1383713890555266,0.5254879301944655,0.019928345475340614,
 0.35936962056236443,0.8487507079904699,0.8253700403780801,0.5592229404269636,
 0.043219736266483455,0.2690463090756975,0.10502444562630786,
 0.9458842637732295,0.3023216524690615,0.5599338405428702,0.6340735411239481,
 0.8753559370723345,0.3976612113568321,0.7950438132688461,0.27092477376303015,
 0.850612699410783,0.9371650119285279,0.09342395439559859,0.445082765181395,
 0.3429588494104312,0.7051858105658616,0.13676027326772122,0.7173748114551506,
 0.6022432859717222,0.9888012791126145,0.9670620706979768,0.6910910705161275,
 0.7864005390547584,0.8823379719791651,0.12812839014018107,0.6259522628468696,
 0.27671493873704067,0.4683824329441827,0.6583096959394545,0.3540975009975016,
 0.06751394485947271,0.803477475133181,0.9955768651346687,0.7489745399837204,
 0.8611965753246709,0.3382570541644401,0.22386048877862097,0.8520270070116759,
 0.957274673282909,0.6667029069736509,0.8507939836199233,0.7593756613483345,
 0.5189701010331981,0.22808927567585768,0.22148381252995952,
 0.9917164021656208,0.4903349233101191,0.8806692335920376,0.6517251719130221,
 0.29152729301205504,0.2450696924084369,0.4611065664919911,0.4353838464176101,
 0.6801080146757461,0.33778850905294244,0.17616940909393963,
 0.8339617240872376,0.7835647932041413,0.11155900309813949,0.9597679116756853,
 0.974702675340566,0.5886509834012275,0.2886460672463247,0.4253497090771108,
 0.8899361001119969,0.23219732775238444,0.8452168275206563,0.8276603781663983,
 0.8432439535857995,0.5828480846126668,0.4320303724197134,0.5231094466922783,
 0.8147778871147288,0.16598254059093231,0.24785768383312934,
 0.8660236414243202,0.7141655978737036,0.7832603857414742,0.7499571813864294,
 0.4497922734113262,0.21968525758899216,0.3460922804396318,0.599945710090089,
 0.8239799072086034,0.4955479033549893,0.6144806996309816,0.8514835113791959,
 0.017747442793602852,0.04693739435245281,0.6943691201437923,
 0.37621305998052557,0.29032424871532303,0.5808260625810894,
 0.9143636844593059,0.01851183927650979,0.24396453200487578,
 0.07382053529978938,0.4647388612949873,0.33612150517341943,
 0.9412591211592374,0.3992326781644092,0.16698580344777758,0.6383630898738796,
 0.8818470409863806,0.9811095501498707,0.829649601991545,0.07763581033169731,
 0.9554511530301981,0.15123390030732042,0.2599250030546987,
 0.06656530044590048,0.9804691203287617,0.05444757550042467,
 0.034335876011665545,0.6998414167126424,0.9103772783765662,
 0.05415158128893105,0.03858924011155329,0.27165614219817535,
 0.8369325302745899,0.3999981268314229,0.7274064533872215,0.1951526024682949,
 0.7972393138969095,0.10824372913253388,0.4573549648513503,
 0.14264537004211153,0.3200321359066516,0.4844479871965952,0.6687094360972619,
 0.6234671246685942,0.05666118422020572,0.6916058279663304,
 0.23127886755445926,0.9174506255654772,0.3814618064453279,0.5884763502397926,
 0.25577897969380314,0.664284006652674,0.88898204120029,0.7661292239375035,
 0.014937144073603359,0.8844846021649918,0.048996941833586605,
 0.11989518287060197,0.28592362524123616,0.7710543731756103,
 0.4776145564565081,0.1436694126635133,0.5302288647998123,
 0.009269347809441553,0.6425480302083388,0.48628706580805137,
 0.3693138905559932,0.6491540599231367,0.9970769910084296,0.17049647734646534,
 0.25542979036364066,0.015617727381992408,0.5242489123784279,
 0.9186810123523073,0.6876748850812795,0.544203624366915,0.9937567599885628,
 0.6806794336072024,0.1600610133934326,0.5880713948496695,0.19117106263282901,
 0.6273893373126556,0.3085288798955038,0.36341345028284944,0.6479264961238184,
 0.9676815379518252,0.8862272666056983,0.9615034355784488,0.16830973864202659,
 0.003685905492807673,0.08066483921608447,0.9309258818894747,
 0.1937773815516357,0.731593162947085,0.8551809313087892,0.17511793106132834,
 0.6131895382985058,0.7723706551715529,0.32369070391985544,0.9376421706673133,
 0.6885992673250652,0.6368055202338808,0.6007425264354679,0.8249327787929452,
 0.5177112373315378,0.32043268167470385,0.010458116496409375,
 0.834137816508734,0.08938970985028494,0.8538133277726228,0.9102407189852536,
 0.2910971305833483,0.33631233639911096,0.8162758200545723,0.5368430947867953,
 0.35221881134580013,0.5883468517677555,0.5366859979204249,0.9313706703390598,
 0.3824257401245431,0.7557882093713197,0.04400709645964773,0.7952648034848809,
 0.6652349879283935,0.06782978601655554,0.8348950919921108,
 0.17599283072096872,0.9248814175096876,0.5888064788141691,
 0.40076205720489755,0.18220776800870964,0.4907127158540471,
 0.8288322590428092,0.7178560828062879,0.44133800477572216,
 0.21035282089191543,0.10217802009431887,0.7942024019219297,
 0.8192591101983996,0.047589361971077215,0.20567106723156536,
 0.9639384528234118,0.8379813953550851,0.2918669522532633,0.8103176931499236,
 0.7581959435415343,0.11647656442224719,0.1837098728886476,0.7183231045136566,
 0.6245417620344418,0.09757149309125668,0.9134091702489795,
 0.32649529281501555,0.977019884925309,0.10229896242335512,0.9885329717333755,
 0.7470276404814821,0.6373176956308767,0.5388097058081092,0.40043695258459355,
 0.8021777173502138,0.4621333360923462,0.3049021841968911,0.28553391789731875,
 0.1289396116060475,0.15945893710472347,0.4129999008894498,
 0.44668140517953625,0.6330577897510677,0.5126459765128217,
 0.49723356749212766,0.6077948615010198,0.26492856801695097,
 0.6802874440988477,9.014684427109954e-4,0.9729790634823439,
 0.4846776620889063,0.19925972124109892,0.8812859028673677,0.4725630712008426,
 0.01314548275534233,0.38405405831268846,0.4960311243602815,0.9542398344144,
 0.9577855374550915,0.37000915116603394,0.32559183090038013,
 0.7481058185000131,0.774294455103591,0.4602447608481305,0.8250739034203493,
 0.5033731625014031,0.7913328593550135,0.39158023974240863,0.8257897404026516,
 0.49422157198626737,0.8836726729828287,0.015448141794695891,
 0.7455033532170742,0.43150467108600477,0.2985762893422915,0.67696148880185,
 0.2529790047110193,0.2406576231541384,0.040512298313934325,
 0.6407828297393053,0.9021009083790853,0.18721934937265217,
 0.23558435223108187,0.49810132105522664,0.26563693332323446,
 0.6390733330518157,0.18543650478395501,0.31042526490423616,
 0.8626475493470225,0.49663299600845734,0.004729020874388645,
 0.6287141196940748,0.242853510140277,0.7113608927228497,0.1596071551560102,
 0.23754470151852158,0.5529189128311619,0.619429506566639,0.6142791679941733,
 0.4545969031632153,0.5580579157595198,0.05479967636129379,0.2489239375192699,
 0.4988309811399625,0.8422373079182999,0.33888468377504566,0.712786231131022,
 0.3935388268909149,0.3295492986046642,0.8743726723890568,0.4685286271298894,
 0.06491287103771892,0.6554645647173564,0.00323302907074019,
 0.5234818710251327,0.9086227332160728,0.1718191087866503,0.964500418287459,
 0.2527657338377186,0.6759499458991528,0.556131420602188,0.3114611708437225,
 0.2583650193058662,0.1271569096824825,0.9029762232530124,0.9788686351453262,
 0.5311073281691893,0.4024350471329825,0.608997993168463,0.3071029870675285,
 0.7573674433150386,0.3351033704692321,0.039236991844871616,0.902979805343269,
 0.7179629013454235,0.8914104652217538,0.9906210821438349,0.38571477930007037,
 0.25004939653497893,0.21680883782078442,0.3870852752618549,
 0.9933857568532667,0.6434052211999388,0.48228072123190024,0.8781949466262994,
 0.11955592348272526,0.9402182327802491,0.0623778660791553,
 0.09564443179878324,0.5624118179731372,0.850033217344349,0.2596570169717529,
 0.5906398939101264,0.9037508165035792,0.5554820600558567,0.8142266618029916,
 0.26582333270129677,0.44691203849120376,0.23418762593808073,
 0.7602163156782407,0.4387315724219172,0.22436506445007953,
 0.37112932192128434,0.6714610899852383,0.1290924384393488,
 0.23443563145947244,0.817902663738261,0.6650187298486991,0.5294114555272096,
 0.08485268625902587,0.6885305968515907,0.7312427685142426,
 0.38132168594348637,0.004731838068199101,0.06597034834839055,
 0.015800030105993845,0.21277611300809762,0.49258497059105577,
 0.9310574018405054,0.4363894891993022,0.01789070820289196,0.9467382847312743,
 0.5342033385104618,0.6044115197038781,0.19870566564191583,
 0.17465440716022862,0.15126666201648187,0.10040447365321525,
 0.7365387436190087,0.013570578301305591,0.12932190861419668,
 0.8094687072821642,0.8583686052721459,0.8962506123913303,0.936969903135645,
 0.12877877899293444,0.34834006232490733,0.4079321304439927,
 0.10042445096533092,0.417177375268282,0.47230882504381844,0.7503306911751337,
 0.6811665758943299,0.4524278016555976,0.15239887744322522,0.6756566061814614,
 0.9738055695511054,0.47422138356230026,0.9803153068152773,0.8404209472153864,
 0.7281380591190781,0.979043373952267,0.548635171212355,0.9570615157045195,
 0.003570325138821717,0.7289173098040014,0.5407805868707098,
 0.6173071918641839,0.996679698769281,0.7346448466576497,0.9173455349754137,
 0.18108867295805753,0.7971717058831955,0.03214550763720869,
 0.9909203754951973,0.5353084400153411,0.2243324528598819,0.6756941943277276,
 0.5465014446097036,0.5156150997155438,0.72622769569203,0.3946059821603982,
 0.4204314673717816,0.7305268013915467,0.17630665581047178,0.6120157819688709,
 0.934094091677912,0.30714110174336695,0.7382332197315888,0.40097499790143365,
 0.3189256074598722,0.7069346505341453,0.21091843690072543,
 0.33038661112711754,0.5799652376032619,0.15758282074724472,
 0.9719820080508371,0.2953588597533803,0.5306408394596502,0.4061133330845751,
 0.17941946208992754,0.3050882803703735,0.23189270926067795,
 0.38494231964798487,0.4658457571114014,0.8296802627281704,
 0.14784153427693825,0.5411324401744988,0.23097951810096773,
 0.49680503160210066,0.6300099716553803,0.41833766213504575,
 0.43433941050058433,0.13644609970196764,0.22557390759348228,
 0.19012184666888798,0.6987270291374251,0.9027216181276904,0.4326684269704698,
 0.5650941994534463,0.5403961301027154,0.8484477045200524,
 0.040154460482123455,0.5281752899321206,0.46753424037123237,
 0.2955734166611377,0.9905465640461087,0.14496857704334964,0.7315086407520314,
 0.40300079035166114,0.5663757266452228,0.21273311258015004,
 0.12647273515165602,0.35265168050425144,0.907500013423515,0.087879938392602,
 0.4719849439200925,0.8319089380286111,0.6530192393426528,0.8770100829919302,
 0.6066373518531198,0.7002057963011245,5.010685754278654e-4,
 0.5537364361496685,0.6389945792474292,0.08575236635095129,
 0.42187238911889324,0.9393051383225137,0.6213201717685648,
 0.15137777503467775,0.46495917117376084,0.08708588783586713,
 0.10368252592838423,0.40073995461150114,0.009459876702837988,
 0.09477035218364005,0.5690328677366284,0.8425494413306289,
 0.09656525532609816,0.3305806546449406,0.09261102296371182,0.505862060333067,
 0.6311322150255153,0.30135780257654876,0.21264565339262265,
 0.5550227359559973,0.47620545011033555,0.2652086312074845,0.7024898294302135,
 0.8524113356392862,0.7107135265088889,0.4209042490669428,0.9246875806228817,
 0.6859948789103814,0.6625915135054994,0.33688086773448567,0.6318467206384077,
 0.5806008133188211,0.6287003933261817,0.6014457748272686,0.5517237935494224,
 0.7836052765902453,0.6003811280755961,0.028989234905114847,
 0.45142380473327615,0.590688059214426,0.44707195850817216,0.0438273443442152,
 0.3754716832718099,0.9061421867999216,0.001977939582259114,
 0.8763118987359164,0.5986091866211514,0.5844869376935624,0.1263601116079971,
 0.4560352420904952,0.5031184262789421,0.7965118316066424,0.3326035702094525,
 0.12966961807185517,0.08860916255077889,0.36505199644576647,
 0.5079939440232893,0.04085695169879744,0.3151826325971212,
 0.029833610159831436,0.5719355312143863,0.8155107044253597,
 0.49337456780927724,0.4920236906005654,0.49246623234056774,
 0.4989754272278203,0.7949920875887573,0.747199121021412,0.15910737015862508,
 0.8115828270454124,0.5063136107119643,0.2518808808343812,0.3202946228983907,
 0.05271703145911144,0.6076807048127223,0.702534564217449,
 0.042260709153627674,0.01956077283598745,0.7446306744710591,
 0.46052579729177334,0.6998941073434711,0.787589104191606,0.15589149633406973,
 0.33962783983069755,0.29516388043659203,0.3854625839238016,
 0.6039116964339785,0.3698208218239797,0.9959525748230232,0.28139021236136774,
 0.1398297572971161,0.4675147055714597,0.1521372682680434,0.8229243001740262,
 0.11946797369169782,0.4882431709771916,0.7472040696579038,
 0.10799486492053267,0.19427815950152227,0.7937222574447256,
 0.38132063958412576,0.8076660271817238,0.45034397141870497,
 0.04663850098331157,0.10328417624348096,0.5446910479602057,
 0.8205113456869961,0.17588375264506184,0.5249425047325237,0.7232380250585272,
 0.7052921169921429,0.4269400655710449,0.5932686920542168,0.19553847561983417,
 0.5061769860205058,0.31607432228043153,0.48604982085892046,
 0.48632998693676854,0.048400073923510756,0.30159789693823214,
 0.17728405963807992,0.030316963819440046,0.28796364911711236,
 0.6351128025429624,0.9925514453386705,0.7854587939610063,0.9999813930075411,
 0.16165695675815317,0.29541405827741474,0.25891473284627387,
 0.7316568793623532,0.8224143809909731,0.0743896293665508,0.5537869741738056,
 0.6614718382927557,0.5577075826620637,0.9826431869770861,0.24513639187043224,
 0.43011417680396524,0.8636903044345821,0.9010409625418931,0.9715203956650977,
 0.2840363077542909,0.6578917829052097,0.6144791829329318,0.5613015409105353,
 0.3341006191374626,0.026772563977924713,0.05758456479086216,
 0.0738965192245351,0.24042229311905428,0.4289866678172076,
 0.32309529816823657,0.43147027179025,0.4997193532416865,0.30707416024983547,
 0.9845056343972217,0.05241356405054487,0.004663453965006781,
 0.4958806979718038,0.8040259382550695,0.2843100729481497,0.9003571553478347,
 0.37337252482867445,0.2771538316550606,0.6631735690360187,
 0.27579143067347445,0.6032548061282635,0.9125658192399434,0.0681326076982447,
 0.435256735644559,0.5124480933529378,0.945859192146288,0.20427011272991757,
 0.639499605652539,0.11453691176374947,0.29317037443381677,0.567641982293608,
 0.9561378290193886,0.2915089314747551,0.6595333623332139,0.8243810620373228,
 0.09948927924268958,0.10985515827734571,0.7593283398722608,
 0.05932025682967512,0.4142493912146098,0.3906018156547404,0.6849288566120403,
 0.44264045283291353,0.4463603431094074,0.6250983287859495,0.7387002156645694,
 0.12000271734846213,0.37575832182862323,0.5087895314919539,
 0.6208494918702334,0.7009761697191876,0.42356202591583725,0.8477010842618937,
 0.4706895156058277,0.4585792393974365,0.42578181611954613,0.547830027026749,
 0.3822032387508816,0.23624154427821986,0.6725341542429546,0.5986293679641171,
 0.9849295707760212,0.8660960092136928,0.3145028761328771,0.37204983092909927,
 0.40724702298159576,0.9154189646643225,0.1924157457406953,
 0.13900437692660872,0.5331232889703101,0.07496587224432727,
 0.17553940895978837,0.6518158583373377,0.11458399760777216,
 0.6059306553518049,0.9802105829208363,0.354580437610581,0.5697323674266962,
 0.05293416534999262,0.5361819893519028,0.4507171888882002,
 0.005245782993897441,0.9113216191862922,0.17898010323961555,
 0.561727310488869,0.08955967834050416,0.2177444042747133,0.46332420787215645,
 0.697283489769429,0.9015416902586662,0.982739512800291,0.8656102127463016,
 0.8823876549059358,0.5818064315955919,0.6019160338625147,0.0219737356271289,
 0.1469784803861145,0.360446187591848,0.21657677971397704,0.5458631904063901,
 0.8140116461812195,0.4403671189986964,0.07336742442460109,0.6107666172847814,
 0.5570654578510972,0.6009146045090291,0.3618978584400925,0.7103331442199436,
 0.07641352332032192,0.8966610679004179,0.31711726073144453,
 0.6277201414610569,0.30465475464745473,0.9024153178553848,0.6790775999956372,
 0.8406253409236744,0.14053495866200783,0.9514627561812581,
 0.18504226944181723,0.5245145043901398,0.09447756503644456,
 0.07103802565971251,0.06696931205644363,0.05870420804874599,
 0.4150603077514968,0.20948624885845923,0.28287905363582044,
 0.9067355564288851,0.8004778532491948,0.8901668535896116,0.11078468006900966,
 0.8931303916101648,0.3987195879870832,0.598647908486316,0.47618364236717836,
 0.45324539891797255,0.7877626228764514,0.9679679443370723,0.4177008481937615,
 0.5333584221834045,0.20609003690363414,0.5224570499897271,0.9082185125699062,
 0.6868923299615021,0.33161101623789335,0.3106292869154015,
 0.40501893913774956,0.6962625864940185,0.8060308516570711,0.9597142540900938,
 0.6563222557319252,0.8515110505695573,0.8689756261704873,0.18119254931027473,
 0.9031781746866757,0.9190216578368555,0.9402590514071186,0.4295733615448529,
 0.22634554869457812,0.024094955804293217,0.3470086363021335,
 0.1582345644159443,0.280134599324933,0.9900475243921889,0.14129354153871343,
 0.25022554524460894,0.2579201331262133,0.6067893805383615,0.935944919531928,
 0.9353588005645476,0.4519275621812052,0.425388774193865,0.7147651651611283,
 0.615510876906141,0.8835871649691919,0.03927291802297672,0.18770742783916405,
 0.007068030127367564,0.0013415066917217278,0.9308909057137722,
 0.44678899522384874,0.9218183900518382,0.8254136995339748,0.9810861590878242,
 0.6915000203006081,0.2776422399172096,0.5933943377915436,0.9770077040733159,
 0.9536828750052126,0.5516266493502313,0.05356431762993896,
 0.20656032227211352,0.41792109285282597,0.3117065195645997,
 0.6745868186519413,0.2873126279319257,0.03493385247079095,
 0.14619832754857198,0.18026190024889943,0.2742132419785195,
 0.5566637747388152,0.35392334969449424,0.03192369443756049,
 0.9542669524544349,0.7263053852952287,0.9440520152466538,0.21271930367939562,
 0.8691483189618111,0.9681744713953472,0.10477230011168759,
 0.22369688224210826,0.947542580017813,0.3072133884358663,0.30583107140243326,
 0.4574873162940595,0.37963749571697225,0.042758765167610235,
 0.39572554289111006,0.6746431930211371,0.6965355892270418,0.4182795886976639,
 0.3620896775185055,0.30390021284063573,0.025215926138899225,
 0.9624264831285696,0.5833691173035477,0.49915595237397714,
 0.04059461829029021,0.6439740476725281,0.8536743994272291,
 0.40989050522020776,0.8915903491022459,0.4766219095182246,0.4741583489308152,
 0.6632481627060155,0.7036768340112529,0.01810914674874231,0.1011913629092086,
 0.18361713770752908,0.10142670071920179,0.8883975581696654,
 0.4853582023542219,0.8963240616183578,0.7225605693957002,0.7903361753794579,
 0.37852339975671767,0.8162221809439816,0.4359291160637856,0.8554434033097437,
 0.16769490820681565,0.6204263910772264,0.14596618309937381,
 0.5037080700882534,0.4360725897530161,0.36424987095227157,0.6382424661719848,
 0.5856117872919189,0.32473042312571154,0.16735089203369202,
 0.03683889824496234,0.5520214731358308,0.7503349805492247,
 0.14500817401008215,0.7952081367450308,0.7751435069535819,0.5908479292947789,
 0.387979272935832,0.04312198421688307,0.14382183231024703,0.3270758861015032,
 0.5999316021843494,0.5595897485833649,0.005468800037853994,
 0.38602939564592575,0.8077967607254399,0.9786547892391009,0.2545386631323028,
 0.5299108847574739,0.06376987037841575,0.5643600268911761,0.5999914439444337,
 0.1842947888613793,0.3226826462309772,0.2718810712930121,0.644278749934542,
 0.9514352535424907,0.24824860324164244,0.795356575837237,0.11944266533619208,
 0.5879030672841228,0.8711281432318001,0.68494199439998,0.6062875031850848,
 0.8430554744183814,0.88374201294965,0.9527091225637305,0.5190060804477074,
 0.648023326436924,0.5730021269287033,0.22905521633182502,0.8100442847381896,
 0.619875666529857,0.8938675864204517,0.915644396593607,0.5442034828516066,
 0.8808007074449916,0.08515119792456094,0.45148166578504456,
 0.3594307069151179,0.9411340497292028,0.9555858361521464,0.8781666710747359,
 0.6553099377202036,0.5240575400767702,0.25638052538970957,
 0.04027963655864186,0.7067503081712853,0.023990038653494694,
 0.43704289520366224,0.34154038534573117,0.032375246022966264,
 0.9529235598948491,0.2540564517388805,0.4389428816226395,0.37231532271366474,
 0.7596980941925011,0.8573462136997068,0.1828643668515143,0.36885095415748814,
 0.9178682543028858,0.8075529580615711,0.6554775984357921,0.9741372876188789,
 0.19532401225942897,0.0667218885416423,0.9534281968734867,
 0.32003027441067433,0.9681934349470371,0.4165880998900733,0.7041346283472965,
 0.014869976348715341,0.7826979655591055,0.3151635751050563,
 0.26121768513267274,0.3718846260899302,0.5174708483035031,
 0.40935675695220197,0.3946715373255253,0.20859310816830712,
 0.06560799843144804,0.5415650669312924,0.0022525217737983105,
 0.9441136999174676,0.9701502352430232,0.9414521802219861,0.18134687891545398,
 0.41842568370012045,0.2687947320377386,0.9862302528651394,
 0.40196282614562273,0.8780407808518532,0.8272394506577573,0.8128267581004807,
 0.22652783300434076,0.7741567776058768,0.6196445526424562,
 0.38071925539490226,0.5573236454711291,0.09360277765472669,
 0.5631033725339128,0.9639421724819801,0.5602747942302481,0.6862581730047829,
 0.7803145970264643,0.6680963247323711,0.6651441175918517,0.5514081230090713,
 0.7026569420264978,0.26676712609139763,0.35727974041044086,
 0.10905972940579944,0.9696376853231837,0.6782930921207799,0.6293588256517961,
 0.7065517608717291,0.08546980305905993,0.40085794734628455,
 0.12250219440465238,0.23560188475788024,0.42819174584086395,
 0.37356054798565386,0.33224539665120656,0.7322425063352859,
 0.5394105170599388,0.9518945965931492,0.15726931350795148,0.4581583550817372,
 0.8543155281182354,0.012761155605004126,0.22185703301874993,
 0.7762784106812464,0.7349619540123542,0.36868851498904254,0.9266711843150233,
 0.7703721386597081,0.015426605720840647,0.7998435250512073,
 0.18434257217647354,0.4107112354743315,0.0987274772268647,0.7536144548782773,
 0.6113458629860012,0.06650443554598151,0.0262932913511853,0.1336119205321309,
 0.3086065745597789,0.5695673996810146,0.4058227535122818,0.5610711490219237,
 0.6384127873798362,0.9933240574639196,0.29725533508515023,
 0.35204786469304616,0.15212453994551178,0.9526820943811575,
 0.8510791171227675,0.951031370942355,0.8176478891349325,0.9046579457711571,
 0.5684852114988425,0.03205120071122136,0.7208794086896753,0.3746449336861961,
 0.63240141891698,0.14982837300110374,0.5363469688789634,0.22172326667603637,
 0.9716625268346564,0.2607151965848383,0.20827388899623833,
 0.03630103107423699,0.7067838676165091,0.9613933763009007,0.3552288385031006,
 0.6893667059802026,0.8814427768487713,0.8681086207770036,0.6485124571680563,
 0.27280779016934453,0.11714481204914706,0.7697047581858528,
 0.6799565634068574,0.8175650387830461,0.8536763938996672,0.9024013891844268,
 0.650732829328673,0.27560345472417724,0.4663035396732974,0.9094151935107173,
 0.5275598358232557,0.34038966624134337,0.21580424571929713,
 0.5269282973788532,0.8773559146642405,0.49477923128059265,
 0.13975473792802173,0.8509853804459531,0.1846708957828237,0.8839172831236417,
 0.8538158959173775,0.2183968415870896,0.07586490331150686,0.7046467250108898,
 0.7546246451409283,0.9132796935960911,0.07668200466776864,0.7468948314593881,
 0.5940103976862922,0.37057587749866183,0.11691301441080793,
 0.8641223575813655,0.6776556574203031,0.9271437093429431,0.09449515396174613,
 0.7382729557230098,0.942919201245486,0.334537962131072,0.2341324870747088,
 0.42339611449429726,0.037640206811844434,0.20317860884324157,
 0.010508367570024046,0.6356631930258655,0.38823898499448617,
 0.8388225910195084,0.15732665514497246,0.4718089681013363,0.9899470602734923,
 0.3023569258243376,0.6511958004640557,0.2933737853029149,0.22227634709857436,
 0.43387522204652906,0.4058222644386462,0.6048956091750481,
 0.44578140362240115,0.5473802737972882,0.45998737665087763,
 0.29819043283064983,0.9858201664120042,0.0769450622575889,
 0.39026993268124843,0.003374486885551753,0.37682675499114326,
 0.5098468880858051,0.5987827015591969,0.6789907337174119,0.35306536828337154,
 0.49757203739913214,0.2703501710031091,0.43018609507331973,0.933627892342668,
 0.28705269407134204,0.645350731878777,0.8696552175753843,
 0.005520806116078636,0.34471201343050495,0.38684446415169693,
 0.04923895588931604,0.2820169793820798,0.14877082678463005,
 0.8775179757356251,0.5460327897198434,0.6946765622477067,0.8257920374744204,
 0.8747317519275665,0.4820261958981913,0.7494496824431444,0.01530080444362325,
 0.4593876679513461,0.504252140265938,0.7052777639721053,0.026890221590347485,
 0.8094980661976275,0.7217567561016389,0.032716517543972645,
 0.6345651304075909,0.06588946462281375,0.43535266176584975,
 0.5246102172748671,0.027155010269486635,0.5487970491647699,
 0.36369792937427825,0.8717585752853699,0.1444808924109645,0.7199336484536403,
 0.3346935383172017,0.026633143585816588,0.3129574081839632,
 0.9800300462633705,0.13749161927068876,0.4994384456410501,
 0.11869551469547135,0.3215536815625958,0.41393401109068806,
 0.33242096480008543,0.47019283206989126,0.4818113880295438,
 0.3843032491573952,0.18148945175274733,0.14556821981030355,
 0.09647210001170103,0.013456607990927183,0.9019175911804257,
 0.21234870000074546,0.7368100863079938,0.6600648617236488,0.6603147100093006,
 0.04274890618765537,0.3816805612772898,0.2334488205043609,
 0.020641698247569895,0.819900605754581,0.6855606466674794,0.9678681595108886,
 8.223175599679777e-4,0.18693790478158345,0.01834483277474508,
 0.44882667035036716,0.7397728921323804,0.47293914997843123,
 0.5007907271685731,0.6301888844676721,0.8043058851264564,0.2551748956707409,
 0.2887116110808078,0.6944909732786512,0.2215820195076348,0.19756256809325135,
 0.4810369493599662,0.5392729094683213,0.07104720955316723,0.6391903293207448,
 0.2605291528990732,0.7704274621245113,0.1129450304744457,0.7147393567498883,
 0.7629768398911128,0.7613431872955805,0.8299203579846312,0.11322044757703076,
 0.48912794353947286,0.8761789309608176,0.04653209985611617,
 0.3571691825315736,0.7436576696265687,0.507211054656457,0.5362667592577827,
 0.3035976728943486,0.25007550457581995,0.39666019186016777,
 0.37956260807436815,0.8012870215829464,0.17801748703165043,
 0.5076430741547847,0.34930246647668217,0.8291197558174669,0.2051323113208301,
 0.8811384812035223,0.2006680175335558,0.6676633181530162,0.2891494166484435,
 0.5305539904680939,0.47439438724160454,0.41022698355987264,
 0.5040305293028838,0.4205009786457091,0.4800145632989469,0.7752569900661422,
 0.2277612359418254,0.3556674985633469,0.26644411962216563,0.4856075487962941,
 0.2709792142591504,0.41651437352060117,0.4455234160530901,0.7814010738988956,
 0.6688039957658355,0.9921325633045406,0.09305255040661864,0.664905738093511,
 0.8183077992093692,0.8984034438676535,0.009191514407529633,
 0.7550698190747575,0.8690745196432657,0.7222867693362769,0.9233336368401075,
 0.7081242307081739,0.6451439231246571,0.15318100020923187,0.8002267381888616,
 0.4150292576872485,0.8519410808473656,0.14137772677507987,0.4255592584382615,
 0.2581667170196037,0.9455817463646627,0.5684772347003354,0.5114580479026121,
 0.9117303137677091,0.8580984580109733,0.9658316112792831,0.2347868973131969,
 0.6484902154599292,0.37392001424913823,0.06535043321699496,0.378220452393329,
 0.8520000137793738,0.6969720729746369,0.6688964259515338,0.20444777643445267,
 0.5452309627352612,0.9679837479697098,0.5778888826349795,0.40645101038419895,
 0.8676524932549389,0.7854423175123364,0.2010903583683652,0.5509669648684404,
 0.651198421766898,0.13047159609089576,0.9318290888829236,0.7825840559497876,
 0.543050276449013,0.32969520373193617,0.217280109315652,0.7423654970861079,
 0.2851733788729369,0.6710006722667108,0.3061353530477744,0.2744978907518254,
 0.4352730596623342,0.6830912712743412,0.7352979118678117,0.9942562489034455,
 0.4911242459584444,0.6596013336329967,0.9700900467476213,0.19513668627951108,
 0.7664606909641756,0.13027972622412598,0.684554305753186,0.4406043120719648,
 0.11781313322314535,0.3695029272578304,0.6906206948156036,
 0.07727849078222693,0.5747085304112544,0.8283074411924913,0.9620587802085624,
 0.9872018156146226,0.6445689103741313,0.5474604850872867,0.47857680644055145,
 0.014307168864575837,0.2762981725708685,0.2033444694440334,
 0.1747700151969428,0.30285377475758,0.7156340655229927,0.05563768495077148,
 0.10796476091550089,0.5959539003008815,0.7044872005754832,0.6027154537870225,
 0.9105336857985418,0.9334765185819601,0.5464905522819543,0.9198376463736625,
 0.7829582755554213,0.18328208161986104,0.11686797315127218,
 0.4000719649277772,0.8772433939382747,0.22662878102295358,
 0.21481166597887136,0.686821841223335,0.09718616285581239,0.4801177067709482,
 0.09648269269759469,0.08384302762029094,0.17254142839691244,
 0.3314219444275064,0.8422779615273249,0.8732980495275332,0.3873654644103728,
 0.2518304896754009,0.17365918409196135,0.8965711146166483,0.7366928005535838,
 0.8888660468832159,0.8414531456816667,0.05154826882672692,0.3961784419838851,
 0.5653852943024908,0.26721669713528984,0.9958358192484082,
 0.08196099793733425,0.745296593103489,0.7238294727572209,0.7649470528947606,
 0.3380947246943108,0.11546902383659385,0.15907387505474346,
 0.00757349077584446,0.8913990483771547,0.7045491941210167,
 0.13367610455633716,0.3150430450876057,0.920927883331444,0.6818783061667286,
 0.4472843622416224,0.663944917215865,0.22420257378790232,0.6578413561291309,
 0.11939182133902293,0.38565005043865863,0.5386534601005071,
 0.4771758045583273,0.5089336625548456,0.886287496454097,0.7858865631600724,
 0.867008567992927,0.02032495285963043,0.5685459529804748,0.41982027974710723,
 0.3635715347765446,0.6945347436848343,0.651122282613616,0.8445817653384424,
 0.8597453021296291,0.3170145565239162,0.6003774590963356,0.7283499771728457,
 0.36451569648886994,0.4346725136765133,0.9753391497546013,0.5286663043727837,
 0.4272128810044954,0.38330189917583457,0.846584011241121,0.4366866687836547,
 0.20329709495081105,0.9200679207080904,0.10382289111412302,
 0.8896580082289647,0.009630545252480083,0.7643923287730809,
 0.14455681749735871,0.20113543245716503,0.3571213860679815,
 0.7793841479758271,0.07305889247844688,0.0746455385522039,
 0.06412662777459943,0.056314485906094,0.3910032902024839,0.4896503446881605,
 0.32850096453692545,0.2702736871908614,0.4970996719769658,
 0.01968830433644464,0.19286072758443518,0.004994835876496184,
 0.18254920572943356,0.8203480785282897,0.5706560125698517,0.5488255357410996,
 0.7858636338343251,0.3153105365599491,0.5410329188401648,0.871430485438025,
 0.03443693270639159,0.9400810673216735,0.8683486661100823,
 0.47827110184751853,0.2268448740093485,0.08978225705463216,
 0.06311604677314209,0.7934610000156262,0.28886954587072833,
 0.1254421785451385,0.9417650757431599,0.7455293340775546,0.7256234548059199,
 0.3243340757807247,0.14914542439220324,0.5190360963871369,0.9804944102928677,
 0.7166512963916762,0.40708614184357417,0.5455965273239036,0.6925003241412907,
 0.6784791030671128,0.2666241114621283,0.0903918208381802,0.10716167362870188,
 0.14192115663714977,0.0727983473308631,0.5630793582237226,0.7899092112967436,
 0.5705415577933957,0.607034666121562,0.001415138677933414,
 0.16373965108329736,0.15604328077383778,0.8609507666143389,
 0.7273519264545835,0.8426002505046348,0.9378234100083132,0.33968942320421136,
 0.5307076281797047,0.7643360900474749,0.8031446729545251,0.9636318054409496,
 0.6107018398802564,0.696794548641664,0.8403183875160986,0.7056865505736389,
 0.8038923642304854,0.8924228725673363,0.019108860154117502,
 0.19834346070910658,0.5275954853139256,0.5319393339399321,0.5579950021245746,
 0.545963606488806,0.26504223198791255,0.6455628459035964,0.950188546788351,
 0.9397278463375927,0.5549640120177368,0.7450168644534683,0.775191820975102,
 0.9623143634342975,0.38115630856425575,0.10061595971367099,0.68347528601311,
 0.6515274676955423,0.9466987884616743,0.5092288489368259,0.9495071287242522,
 0.243851469578246,0.5965941244297608,0.9427873157200811,0.46111401555213416,
 0.523226781007042,0.8156924019910774,0.15764238235689831,0.8653018817974016,
 0.2012042365130613,0.1567940845470348,0.16597550409320982,
 0.12479128395657968,0.7585624361470943,0.9135399876603603,0.385001838894542,
 0.2827901688330209,0.19399610233617015,0.6021025875980116,0.7787212939902602,
 0.8178107435251987,0.3237878808956145,0.44143730680863613,0.1621432651057647,
 0.6529290126560332,0.8502982744551604,0.6050033347079649,0.41497575568075895,
 0.412535997322389,0.8389289612982366,0.5994875707722143,0.41851429260207507,
 0.14592375816970216,0.9592625177204062,0.2283326417216025,0.8089419568062914,
 0.06536893173795333,0.7749572964161847,0.7225338594362917,0.6395147013832901,
 0.22877820685421213,0.7437233579740425,0.9984594515020809,0.6188241274620035,
 0.9992856003082675,0.1920289451354159,0.10789778183502874,0.5747589965001887,
 0.8486105342309328,0.2325400782612146,0.6511375196188491,0.5293915938716631,
 0.3012974188899925,0.6135138111391116,0.050360902074575664,
 0.014710336302240501,0.47882007317171826,0.4194998551074176,
 0.491345240477175,0.6990102916126535,0.08500411489821103,0.9345004483557924,
 0.46609703702190786,0.49512995878557653,0.32009516496461954,
 0.39279027366240504,0.20902599161829527,0.3991308483029239,
 0.0023888609700263297,0.6825786574701648,0.9890918444785838,
 0.7591463026723111,0.9995544359637847,0.7245164909993684,0.10435973479222349,
 0.6317681955515756,0.3275926237588279,0.9168430568740895,0.17855112982617205,
 0.5734162357014453,0.7554568269171353,0.8407007880775438,0.5813280308238206,
 0.3206084907571848,0.7890146345495248,0.41989217956806923,0.7693015267642364,
 0.34668386929496176,0.002267980905541589,0.8690068368093361,
 0.4227543489700629,0.22120221371069793,0.7022259678694532,0.7835931082073875,
 0.9747489092200793,0.7186466820663459,0.7600276219754134,0.5336983364472893,
 0.7802722960477886,0.02315316896381836,0.2387209977347614,0.4865053173733962,
 0.44960627570718337,0.7116472759865193,0.5839340645001745,0.1712299698562777,
 0.12103247416486373,0.8991358055182053,0.8007616037920041,0.9453027770631066,
 0.07425302767592679,0.17508367143743342,0.01833656887558499,
 0.5100152134743787,0.23537695853635165,0.8028193308921031,0.5444586693748734,
 0.38291547294686723,0.4729744674868942,0.7607675994557573,0.5222024360949702,
 0.5393220022600914,0.8684443684254676,0.8477133595060676,0.5214498475634064,
 0.26953169340400274,0.448377412109652,0.26987618293135385,
 0.30685179191490153,0.3767985572012962,0.7796940404507422,
 0.003537185311186186,0.9352381238394977,0.14980153919778916,
 0.12782178613822337,0.8038696010973225,0.024580084750316922,
 0.28490935249401705,0.8390958936769042,0.43638896984507836,
 0.25648883004069833,0.07369109596216061,0.4122304804084531,
 0.9356780902676407,0.021023231288980315,0.3961104730145315,
 0.23378280672144136,0.16276131317393272,0.849797117619543,
 0.40682980094051224,0.27203729872528737,0.940102984217707,
 0.14999774730628113,0.8541480050475498,0.4659893364431149,0.8966884855770012,
 0.3895100698354982,0.35384610540010497,0.9515091618694086,0.5802973901680928,
 0.1070198174367355,0.555362213065719,0.10526644507473992,0.5292319018559266,
 0.8061491716741386,0.10987367556437655,0.6735687055284801,
 0.0067154522493338975,0.031433583360273354,0.32726079452246903,
 0.43681825294184173,0.2838890188629828,0.45788798687965304,
 0.5271159262568731,0.173831936395609,0.5749866940721476,0.4194947923557164,
 0.7082067845977276,0.49387318466845265,0.7081194092828003,
 0.031936073705638224,0.6860245944896108,0.21177693068256165,
 0.9521680020527835,0.4176447062446905,0.43336616307148557,
 0.23908473869040459,0.503557469402822,0.1711108348622119,0.5759242599158266,
 0.5434893568634623,0.7691225894367864,0.8514528502082357,0.8034553603599791,
 0.1141416184116385,0.8013535853371726,0.4919612619671331,0.508289682122027,
 0.6288987611413481,0.6314145619303788,0.9033595335771061,0.7223085673493825,
 0.9924945019052178,0.21598180222065733,0.7658777286249112,0.5115261973938935,
 0.8475423877680506,0.2808726013823699,0.13392896377734853,0.8658391375608503,
 0.5543249814532248,0.5342095857874387,0.8948034768914591,0.4693068219674983,
 0.8776982511352148,0.43113419992922575,0.012936241695188588,
 0.33172898072043067,0.6203481692830312,0.2555158391872815,
 0.09922311792707428,0.2260860265513639,0.2476279003766263,0.7750018462965026,
 0.8099692211985429,0.06536623876156433,0.8980610174032996,0.5032799416147954,
 0.2583329182808566,0.75953587831992,0.41959757383456286,0.6156707164197113,
 0.7900784145100975,0.9079157984887337,0.27384084860652536,0.6133334567185025,
 0.21611148315069695,0.8543954414778605,0.20572509602097167,
 0.2654705563641222,0.37340672782035056,0.868961378614245,0.8284582083010782,
 0.8649195213336927,0.7968743819116682,0.6984798199649087,0.5253886844131875,
 0.020300805068215633,0.08668565533204531,0.31515376580402377,
 0.22884482975982146,0.27622169083899084,0.28732030917450313,
 0.4884334951044046,0.4492604593952042,0.04021241233896922,
 0.31697416292220415,0.4803510237408609,0.0866620519487854,
 0.17917812584054926,0.36698953511003496,0.40038511335816274,
 0.42796385325036135,0.6553109978287035,0.19497074145606108,
 0.03261050944778665,0.38094118461321613,0.6137405603718916,
 0.7592840057001743,0.9161742193067987,0.956785695793976,0.25871406959277443,
 0.5739937646359503,0.3992073979366536,0.7289754978700805,0.1812720267344674,
 0.779828865104547,0.8875968137768024,0.3844631795630642,0.6970064596593099,
 0.35227750554666526,0.3613657771920592,0.00518066951032492,0.244479208182248,
 0.370427958224927,0.6001416650423911,0.23276118441371818,0.2164073333638611,
 0.26023493762025685,0.08651656193060553,0.9780672988776661,
 0.4837063885797662,0.822587798807429,0.37190918642957105,0.17280847801253063,
 0.6463843932681801,0.5431837807196838,0.7856047511709994,0.8879290320713655,
 0.32436193100828425,0.30821631524282633,0.2548736556583012,
 0.24895747192302198,0.280195999447475,0.8459692397848301,0.5015654195524069,
 0.2563254676285155,0.4521048476026245,0.10550358214597166,
 0.23977851817371132,0.9143952390785983,0.7586517608015622,0.6170914709482421,
 0.7378601690968772,0.19466010642771536,0.39034242002707953,
 0.7901055292907276,0.5719663742897771,0.20151431355743732,0.9428859292883242,
 0.6046932050163029,0.2728933080469813,0.1293678403232681,0.9283099080192487,
 0.35779902622216475,0.21179485257829034,0.5626320766331191,0.630680708435322,
 0.7754206508938329,0.32007666247313404,0.8072205432762029,0.5145051076934285,
 0.053968802041734376,0.3280849487601688,0.3502807156251524,
 0.5102186423357589,0.4633948643768043,0.8857739334357272,0.32620661779381566,
 0.6192062103662052,0.45789131561041985,0.4803004646028274,
 0.05851009142060526,0.33954968281197173,0.2437587866757095,0.692153813372261,
 0.38823946150677335,0.2551252026861289,0.18377960080332068,
 0.02541161787868851,0.9105855907604068,0.896395346611144,0.1043193718416815,
 0.743678309574433,0.8754862236616547,0.12855181583948194,0.9983893515437577,
 0.039037052531551986,0.45317410500080235,0.9741816059331836,
 0.8556038943369509,0.8273554062318533,0.3249537184237712,0.34187333373519846,
 0.47725772279024814,0.5494733382138594,0.44706276631180175,
 0.3630525162262326,0.08931118074448663,0.11053543277122291,
 0.29704739375744815,0.9607269214957544,0.9959765137100816,0.6260864578136834,
 0.438628146240595,0.5828382798473543,0.2141170368549905,0.7107609077377579,
 0.284755377340991,0.6962624019210732,0.0986595861746673,0.6420404370030156,
 0.9933740874235739,0.21017949857805152,0.5999064401655733,
 0.16475366579498063,0.5703729312782015,0.765988309971046,0.26950359888429776,
 0.7191785998064182,0.8254009644337605,0.23986657607037187,
 0.0029777985784105176,0.24523085190501492,0.29672315983154796,
 0.5491105050646399,0.9858877031696026,0.8182519173625614,0.4833995374436405,
 0.6269167195898603,0.20961175054847114,0.48862322380734546,
 0.1838745790419435,0.9658851172909737,0.9434778888789256,0.35526499009432144,
 0.08680337020311124,0.7223014646309541,0.5859667302150443,0.829111078884217,
 0.6992382706302549,0.9068235235836063,0.8960119317723336,0.1995843995480423,
 0.31082869275049474,0.4620231260673684,0.03161514344706917,
 0.34786038288536314,0.7118123498160741,0.3774527810731223,
 0.19463919252005335,0.2165500272316926,0.07049432347257922,
 0.18144940634077866,1.6887480705363167e-4,0.6741961069803357,
 0.05980455471715307,0.2716086933948192,0.7061711834872846,0.4432906099816386,
 0.5125258410470532,0.012183406362770488,0.2790070266561677,
 0.7770514005139388,0.09305034426095293,0.4356382503146474,0.5167669323424245,
 0.597355950233154,0.24197578095155958,0.8042397107845882,0.13530459687887697,
 0.7054022312636616,0.7662235284926513,0.17480639526691344,0.8455954171256888,
 0.2148643999335038,0.5605871331004175,0.7212707554479731,0.7692224532357943,
 0.1024372880064981,0.4933183067277591,0.8167726144614306,0.8180272331932626,
 0.3154184140912125,0.11573067956299421,0.9482399758421044,0.712683506277281,
 0.5447860918502359,0.7895374955965884,0.3088173257940837,0.7682868933542328,
 0.2436796971577453,0.468143169702108,0.22737221476473335,
 0.007016367944318258,0.5290832852819752,0.5381647674629836,
 0.9978630860004003,0.498143845896984,0.8360072401415571,0.13886424713218992,
 0.546087373838223,0.7906045910956813,0.46071167605193186,0.8938650289157504,
 0.8026036475474632,0.2312631343295195,0.20316446765838492,0.2429275949317844,
 0.5430681741669183,0.7520349093696246,0.8446073573485485,0.6662825815951088,
 0.4201199914537893,0.5193720851652294,0.0808699719913244,0.2250552317403911,
 0.586075012955596,0.7437086883166465,0.981474351130474,0.02873719377931283,
 0.6127080890699054,0.6635711011823325,0.6052521462471621,0.1169001963033709,
 0.48846383338628263,0.5777438700588288,0.5866132152685478,0.8812064991798595,
 0.3645814675292708,0.044179260808859144,0.47233376829815565,
 0.32986305872827804,0.1341433387386184,0.06657706323824841,0.582196638869275,
 0.968700427227469,0.7895445804758368,0.2512644589187476,0.3909682422668772,
 0.8444221592742289,0.7417820647418919,0.04182474230257094,0.9143620097216556,
 0.9436712967688462,0.9240086617073161,0.15638823822307002,
 0.02772676140260666,0.7306108355407576,0.5933120732781362,0.3017711033148345,
 0.9687311294261635,0.4799373891793096,0.35315092455072294,0.7714292402253763,
 0.3961624188004914,0.6070882605069291,0.43481475174121786,0.7428901992680672,
 0.12225306716641371,0.5941596878795713,0.7347251430231101,0.5047583986341697,
 0.8277854073335423,0.7891290333184973,0.260060824948957,0.19971512485442966,
 0.5826236909609566,0.9989190149181891,0.8918879688681967,0.8233711167678591,
 0.859506201459527,0.118862017164585,0.6639766787516439,0.13864610913075182,
 0.805051964457901,0.14482748784781296,0.7606106000836483,0.7433510759525738,
 0.9054421361883286,0.3108400653506309,0.6176003160140646,0.18448278133041685,
 0.4845899515358594,0.6013455925982851,0.28645800092551066,0.6254750634851947,
 0.4657986601195335,0.8951454965122186,0.7121303955061998,0.7012370684907165,
 0.8418065653541383,0.4279322727872593,0.48840910092409695,
 0.004227795419919911,0.8028645504360367,0.5976500506210144,
 0.09735015202291586,0.24174845265636546,0.16443766958132178,
 0.4009345759505618,0.1634036796850119,0.6557871086799906,0.3656299106435148,
 0.9346908176086631,0.8074791759710801,0.6606441047716637,0.7860797616865848,
 0.46371037979469243,0.15694931288475622,0.941027707526771,0.729475570652911,
 0.705237447863795,0.6026737353046567,0.45429165558516316,0.6581975855026018,
 0.5370717343419918,0.7234959020239895,0.08272612547635316,0.9276869476816572,
 0.6408445882634637,0.5154241155814433,0.3366607282784526,0.7665622243864005,
 0.04368032665020105,0.5401642013272291,0.2213514105570269,0.4519738498146679,
 0.121113801638565,0.13092712093848935,0.6441714525549515,0.3884788419831213,
 0.5219663779025063,0.5425970913173794,0.5707948786809224,0.26056705126215696,
 0.08433923000116794,0.36220419912121826,0.07847009139630168,
 0.12686174164791508,0.7607681699323388,0.08720722689770455,
 0.5714032831325011,0.04428044417710164,0.6078808725215552,0.3075697067373515,
 0.9265607549238591,0.18890557925565188,0.9010324582997367,0.3385752031321704,
 0.13163630961656914,0.8188338248040372,0.6521323972950005,
 0.44208838828223884,0.7365453582235209,0.6363920169464186,
 0.0031226589653802206,0.8088153598644859,0.3990088944798167,
 0.771869591414002,0.6291399238489466,0.5604424538008026,0.18172534818880481,
 0.9770155639406831,0.5878719310575484,0.039301991343479115,
 0.5142554471677243,0.5446125483625984,0.2481957597153327,0.15495977023171503,
 0.48683208513698295,0.8036752160508852,0.22966376329108162,
 0.9381670236260301,0.716300766662547,0.4205157963065127,0.7062834788967328,
 0.5542784378426839,0.5232760847063178,0.06513219404396586,
 0.31772566520275813,0.48227009615606764,0.2975389323253832,
 0.8961729466398768,0.5072718606680233,0.7790348449800821,0.8461722057093738,
 0.6318730432748321,0.38721073910153725,0.877755845508818,0.5676874170314825,
 0.4848503053034573,0.9997585664318183,0.8293799971809649,0.783745193028776,
 0.12963872045449287,0.627034692635172,0.3446270377733428,0.25113426767918057,
 0.3525790684276494,0.16307722791511825,0.03976076112822269,
 0.49014302314918623,0.23120100093584883,0.2693220400133891,
 0.09652446606950071,0.5893018675196587,0.7363764000460895,0.7556191108572782,
 0.3986708963444522,0.9312189344456947,0.7065260418476949,0.49637465670656256,
 0.14058876819580102,0.9588509286668936,0.13035892775444957,0.375265911471539,
 0.5546076167421298,0.15925245869641202,0.644082142201749,0.3593935328521709,
 0.8099519574588606,0.04538944890820251,0.714982721882933,0.06735492809143262,
 0.6995349166673855,0.001044500549286953,0.3381458046785819,
 0.7712792635492756,0.7041246051049329,0.5178932987224882,0.871735534665844,
 0.018725567291261846,0.5897498440855107,0.2067213576758704,0.620205862970098,
 0.5986846710735292,0.13663788221226758,0.15727532027411562,
 0.2998492688194635,0.011850912653993184,0.8329027427150293,
 0.05205871292531128,0.9435536755307662,0.6313431424408833,0.6066649123228098,
 0.3552128033752827,0.9946798377162831,0.6761088442889369,0.629748155799195,
 0.5814636271363484,0.6726669362508626,0.4874189053147837,0.6902639678396859,
 0.924388144399058,0.7931484795040776,0.00807148623394438,0.9343944118016247,
 0.584193880078787,0.22497189954038443,0.12064039201300403,0.6534787412186736,
 0.9170623063464857,0.2531255263816472,0.5095022848468289,0.7480572890101456,
 0.6510138859320641,0.9363854545146052,0.7229496872695989,0.3111015950655913,
 0.32954134313969463,0.8944389154536497,0.11830963016204021,
 0.6434895444777695,0.8168693556396616,0.8621414026451972,0.9552751597664098,
 0.2740496522483433,0.11584553996419,0.8097547564102258,0.9438554063905976,
 0.10725339434689263,0.9799009123774917,0.13127463304216447,0.053401022717529,
 0.9825597452564686,0.3161386475063791,0.44641608593043614,
 0.31859212762658995,0.766634413196178,0.6989924589779467,0.7933303098726585,
 0.8141205959787954,0.841982741312158,0.38297157458965614,0.10491794990517811,
 0.6610338238807166,0.2989351493413609,0.7497160221608588,0.49390104759775977,
 0.7533949627475641,0.6962107070622595,0.16314568420771025,
 0.15844532698119662,0.41158410375489174,0.5374870373391715,
 0.4712178830376077,0.6188351808640797,0.7047605486507502,0.3734441451669708,
 0.044779102149113736,0.598994543770461,0.17950329868300452,
 0.5140284043363953,0.09843272349671905,0.9820913326846203,0.7229475302810564,
 0.6651842749447201,0.3276779226558817,0.8345476610621687,0.0770439848468788,
 0.9299393743055976,0.5122056850526794,0.48874970066241474,
 0.11948770025688882,0.7732867722991829,0.08672650886699518,
 0.22770934934645248,0.002189104172356404,0.9231643868794643,
 0.29958107278579915,0.6776267357246283,0.986044052095738,0.07748174026833765,
 0.5722517214923944,0.06017297894229012,0.6848496974636058,0.6869986221653608,
 0.8958201124667439,0.06902761391996681,0.6869109620279991,0.5564323193676749,
 0.4748379187489219,0.24234195996708507,0.8167040617967986,0.6548058116066445,
 0.7709435829323805,0.48981254901903093,0.5875470937225173,
 0.31401532319628456,0.32644499929204096,0.9566442240281394,
 0.8966789049855459,0.33462180492186544,0.33053001402216453,
 0.7787332161841456,7.614652462789584e-4,0.739199096473462,0.8064116553057794,
 0.004077552153454733,0.21850034276651398,0.4857591100331007,
 0.6176751308327383,0.24739247593566216,0.6780718992622552,0.8163509878495443,
 0.7204345428970926,0.48022567419882456,0.6517101824734465,
 0.10846805739651944,0.649997990874215,0.8616115557557403,0.7028713903555001,
 0.14855216433756135,0.5394575630048649,0.6392972181848533,0.4600568008192665,
 0.249964666824795,0.5934677477458243,0.7419171744313364,0.4589176704667869,
 0.009024046991470414,0.4164744122916222,0.4181607298640434,
 0.8253321920170023,0.13884848652285198,0.9292727657008136,0.6687937127920599,
 0.6805658535433976,0.6629544318354447,0.7773339489938071,0.14182082718483768,
 0.816147934382522,0.6285695085585189,0.954273921706005,0.43302843626143805,
 0.3759174914379675,0.40170756907267036,0.5728564869731398,
 0.30369581914581045,0.0760591527391915,0.586063349357367,0.381323011115441,
 0.15512539311586915,0.7386128700717091,0.40347993727106957,
 0.9523272092985445,0.8867219224513554,0.29589336315491266,0.8029037115596155,
 0.2501157675396305,0.6420656868401107,0.9827516968581668,0.3576657752238126,
 0.6194134910772348,0.6557977727777953,0.3332054811768801,0.2689816467409716,
 0.4143465138626874,0.9988166985105009,0.8454076942009323,0.9256031472747501,
 0.8731670507485519,0.13133622147828905,0.7126936108475762,0.5568905771847296,
 0.7123444425551907,0.552276542080696,0.2203739700524595,0.8330018683908501,
 0.9054830941645982,0.5722025549937205,0.35593489151563007,0.5760052059430831,
 0.05306876100280766,0.6593673053537292,0.01830450182907306,
 0.7568743059786122,0.62001407057931,0.32246596046302234,0.537373176509129,
 0.2082646081733145,0.41044541733177275,0.4748976219721429,
 0.39099039019403325,0.969052736269737,0.377387642776154,0.4580523555117434,
 0.9317453730617037,0.6972816582135023,0.1586771025833964,0.42349936082408246,
 0.8935436540272574,0.8248245551475335,0.10794688316514234,0.2645016609126376,
 0.1704362708495123,0.6988752182204475,0.48305937416040146,0.1816187115899145,
 0.09498338007066898,0.8413940186026427,0.5131717107727046,0.9060867479463869,
 0.8053831669292373,0.36186606229868257,0.9667099869873556,
 0.22571245017685482,0.7162180747145866,0.7409674508211785,0.5487647598262212,
 0.08893558319027395,0.05465271819612494,0.18600739450479387,
 0.6902753493017945,0.16976931254890038,0.6711138806502166,0.7478740715496994,
 0.7373745793061955,0.7561257457133871,0.3834546240754798,0.25501901849095876,
 0.1452076261108881,0.5896362912378569,0.18967611424593256,
 0.31442043289449895,0.5589266327740522,0.8993770936353895,0.939237765015158,
 0.46885801150023787,0.8664306910237942,0.35093897070176283,
 0.02231547751679308,0.8426263920099313,0.2383181671380421,0.016808626233073,
 0.12832650524186517,0.8528603067293568,0.8367509389856524,0.773416948749664,
 0.36077723808537776,0.32867412735570756,0.1294233392661387,
 0.7124668370885988,0.09206629010071377,0.4380817188022028,
 0.35945070476851737,0.7868742279023992,0.7373426851486785,0.8971037460778502,
 0.4632641594298619,0.7085750579788476,0.03610911875454259,0.4210354744334921,
 0.4863168011956498,0.41892790058590046,0.29510784631061815,
 0.4975403008136462,0.5443647435772447,0.9462468301812741,0.06431330363291632,
 0.8850545648016294,0.4603966489135347,0.46554853461073475,
 0.24640113073888226,0.8688685799653824,0.7109869862334046,0.7002524335371348,
 0.32154021065928107,0.3324067047372453,0.07683237333619264,
 0.2887105949757641,0.75652839211565,0.7569716739165762,0.5119911185206628,
 0.15901316281575129,0.4824151831220509,0.7702006903697468,0.1616889454944841,
 0.3148236371978027,0.3393978488066085,0.9761647398660406,0.8334834351845151,
 0.7128505763505708,0.8101429389975737,0.6011678160132825,0.10884202603614834,
 0.40637039789965945,0.9539667478574949,0.8696386875778673,
 0.03435768858027788,0.6754294405223142,0.030896105369859983,
 0.2319959117635546,0.9781552948339941,0.3944791201939295,0.831794274700751,
 0.6064615021963224,0.786118043589525,0.3091798322887228,0.3030498188810068,
 0.4318281650115323,0.4677413987375738,0.7338604474202249,0.44179635608077017,
 0.03768127877870486,0.7342648087112198,0.4633235420964895,0.1039653107322075,
 0.12398579031838619,0.44588476434129753,0.12204287533942249,
 0.5362076581765961,0.14362532323528476,0.8759197484306981,0.663086635882255,
 0.6364049416845383,0.8563752761460081,0.4941330185054815,0.9129962015540666,
 0.7074366605926415,0.7360332291193952,0.5842521979548411,0.05437796638033232,
 0.721340716801711,0.5183748894979493,0.9936501136584426,0.5968004002887494,
 0.2352976628794342,0.572722668080009,0.37140916941291424,0.4183648936678733,
 0.30416929733572073,0.012756080997834318,0.15769885126431493,
 0.0917673298567705,0.7004743375228126,0.7859548181304079,0.4670084630343744,
 0.8170314324025082,0.4997633374937476,0.03525628863337493,
 0.26448353937619185,0.6869828185845309,0.9265293149174125,0.6160457939867128,
 0.6151759432987297,0.7023223468178046,0.4602808033080117,0.22602901986614654,
 0.2795959560475807,0.6851235082982724,0.2821380046256754,0.6866245410639851,
 0.6320548074109195,0.7324442466262566,0.02495094658693764,0.5025251861091,
 0.5840597088534363,0.7473169178721617,0.6490858487486688,
 0.032642910538217884,0.6391565196002438,0.22099748791318108,
 0.9270590523614879,0.47290048038210997,0.6309553292227976,0.576999508955989,
 0.3266537261579846,0.7953016578491127,0.8004855757828911,0.5692652384666547,
 0.17522324061561367,0.5084038969130705,0.7247957190093324,0.80517897040699,
 0.5273845905759247,0.08494507251237349,0.46158896983223685,
 0.8998990231243408,0.6407913473135624,0.48131416850438,0.5226450899627617,
 0.3091128274787147,0.5618347697646557,0.6686564252340913,0.8775303493689427,
 0.7824506965330252,0.6233402824914998,0.9182295732742021,0.21984756312454934,
 0.9748302459776402,0.8771860256359458,0.09292894019568876,
 0.08709929539881378,0.6533997986219564,0.8980088422021012,
 1.7055381888952947e-4,0.4040828923223069,0.7512979949049369,
 0.2325903196419331,0.47453101402776277,0.5734859670040681,
 0.06417602636771191,0.1654045416522767,0.9493878643151089,
 0.08448941143584365,0.41047964363618294,0.8013836770382949,
 0.23033198510299502,0.48714225654940735,0.8015039150348626,
 0.24629042302138204,0.958138465375137,0.5302787604631607,0.42997653023566984,
 0.7303175831309989,0.28942398430047,0.2166638173526031,0.4286992150780319,
 0.2702710997807718,0.16199448694643392,0.33320821148561,0.5739073866658916,
 0.8087380700029004,0.20174881481952867,0.3625082522088531,0.7783526928311826,
 0.42015369326114016,0.8975377341542672,0.0033867236381387444,
 0.5824030571047636,0.3333213399316197,0.7704436541967458,0.2946359680156162,
 0.2388630256126375,0.7790025200662273,0.24429774043511632,0.9076399022405477,
 0.7004101883234005,0.7291034672712667,0.2639563177826425,0.28017979606302557,
 0.9276493430033474,0.6302131898361297,0.9641242436050279,0.13909518227548479,
 0.18557006234321216,0.574083773447883,0.3947360292309233,0.9013518648631836,
 0.22820491247758823,0.8344984726793924,0.7396694160908749,0.2651904636469362,
 0.24960058180568512,0.4010161652743057,0.4610499664866231,0.1503110764456631,
 0.3898077583763072,0.2753659925392319,0.3673718582510266,0.6950318861674337,
 0.08300478800148536,0.6258431136655008,0.6122228374449559,0.842526965145989,
 0.6211342492857425,0.02021016863339531,0.6219357111094918,
 0.19652622839902378,0.58172548357775,0.9586901738701523,0.4831884836395196,
 0.24323507475851514,0.0051033792551762325,0.45521951678656625,
 0.14556490141627154,0.9312096348384238,0.3004772025339113,0.9948604708091907,
 0.0652417460774598,0.6149950893742071,0.39972447830744784,
 0.11575286627965964,0.9552791656631853,0.26747856743394227,
 0.8659284183092918,0.19720536591413118,0.28077800340674874,
 0.7162431623894593,0.3906527553210053,0.5552617351965397,0.6861373177836799,
 0.9529581480378276,0.20774821389971843,0.58252593382184,0.7864191402023468,
 0.49051693879971736,0.14519106085180689,0.6670451296738104,
 0.5936226263440118,0.15420357622973857,0.14278549650842454,
 0.32718654583838913,0.8935809538859615,0.5599797728875711,
 0.45299539167713476,0.41740164290761717,0.3863004118483955,
 0.8531639891096474,0.8498739344529163,0.8372921658679122,0.41841792338815686,
 0.2885165597524061,0.9312264557998732,0.983774130652974,0.6099874174330002,
 0.3214522325436311,0.5378938805153787,0.03761622830573863,
 0.03290102708995157,0.7667581716805791,0.7476529755018193,
 0.031230826205625672,0.716649418652501,0.6389708563856865,0.7035809549769069,
 0.17033841140192463,0.26987208116858674,0.7775920385499079,
 0.8734486998497472,0.6853509930979176,0.5393017838641554,0.3797331333354159,
 0.23384308389243447,0.35869789346204783,0.7917523939195947,
 0.5449316650893337,0.13756577083940724,0.1280640940431994,
 0.31581180379708884,0.3723187102814862,0.3275612514635188,0.9904781312416472,
 0.7218818335780794,0.16121138792891254,0.01913745953599011,
 0.3067035676992155,0.869969315188898,0.4941620552727146,0.6058078356674631,
 0.1941622785707815,0.8423480405863591,0.7679639672079714,0.8128316079500434,
 0.9760034422741085,0.025089031651855387,0.3938046086670497,
 0.22421616064863803,0.4744597218013761,0.27935637268523683,
 0.8654163293658808,0.5957293062299159,0.44729736183325164,0.5274662412306016,
 0.9030538069959434,0.2690101725738516,0.5920773483201069,0.2784934410317379,
 0.15560233058862327,0.011800558465775768,0.4647275973308209,
 0.38230567401474214,0.4550473972534982,0.4758926914243852,0.8496617228710217,
 0.46969387761612524,0.8404171458705983,0.6194994903947031,0.8037875191839012,
 0.6469708176917404,0.9347980691542667,0.9515768739188205,0.2555318994147544,
 0.13497857742752495,0.4713182708873891,0.0926334697786586,
 0.32174336281272664,0.289581776792625,0.0022157548687045114,
 0.4870677883070227,0.2103881980344665,0.5254597419565663,0.05004407119572096,
 0.6022157109419682,0.9988415549015824,0.8755211913411292,0.507555943674342,
 0.3823775035063651,0.4915912862119771,0.15281394592986886,0.6013874574921036,
 0.2739827311041423,0.9570152177100988,0.538756658372738,0.544733460344294,
 0.238221954663146,0.48987078284647645,0.05437411367204281,
 0.05409676360501925,0.37616179302171304,0.09961549083757348,
 0.20413099664733703,0.9652098281082993,0.5626454413408143,
 0.058081444753086076,0.5331099605033192,0.463068189488819,0.9081436348201595,
 0.19488579354474345,0.5918637666798228,0.6505119034701053,0.287050567444081,
 0.7802790492870914,0.39010273743405177,0.41914451199367997,
 0.29778334355915814,0.7908105401065814,0.555729527556899,0.6370076969713687,
 0.6726344045200042,0.6908303691233748,0.5037161154568561,0.8452850886765612,
 0.6383310304571673,0.14446507680734066,0.7124652160229846,
 0.04389554648509497,0.4372316618122404,0.18552194652038878,
 0.9830236663705699,0.7886948623563763,0.6860436898541526,0.356596249527262,
 0.36518286856862836,0.8523871354149515,0.26032039362148973,
 0.45070011186103065,0.9101266482263672,0.203215860179069,0.3478304603723372,
 0.21203614088936218,0.7467304920449114,0.5290681667720852,
 0.07395178026989546,0.11339522144424463,0.9411703780515565,
 0.4277326764019376,0.17108349082652807,0.6617066096046383,0.7495191091142371,
 0.3799558168128455,0.27251216914259,0.3775599800050158,0.5463678169652856,
 0.19145531346660927,0.8862869984892432,0.6749851384217399,0.5229443983557341,
 0.2044044266246312,0.4573606449895794,0.35493296744351954,
 0.34242251174101246,0.503121155137098,0.7651863009231299,0.9968944040019154,
 0.28732593027979325,0.9830143601722853,0.19294450995400148,
 0.6461633231164288,0.5106968172950502,0.9739946393541205,0.2603634225387592,
 0.12317136149680274,0.7000827436218847,0.2218942273311475,0.8413548492831475,
 0.4037525194648539,0.4284768034566193,0.6327295146264819,0.5135358910530569,
 0.13740261100991047,0.6852461498640652,0.6624101272906552,0.4179106788443401,
 0.16965153971842817,0.7506408516601296,0.3080988669672001,0.535157004813208,
 0.6738320962875146,0.07763655848938555,0.8012443992089993,0.9299628962716282,
 0.8504755297349569,0.42233163102900795,0.16978125852299852,
 0.9412144148429938,0.5805923420705102,0.9615427199530929,0.8268382973245573,
 0.22293015299422092,0.6117479720421067,0.3196770218471936,
 0.48246911352805855,0.47681926029219535,0.12186913453067794,
 0.019780209949104233,0.1020761028536682,0.797324337725872,0.6694097919709706,
 0.4610165392268456,0.3797427797034245,0.4154564082452259,0.7863734938749687,
 0.4599940546298359,0.18632393313829487,0.9267846339605491,0.8353645439248157,
 0.9926096143058766,0.23438638835669634,0.2983510647313874,
 0.40595911426341513,0.9789204389400163,0.3082619376615494,0.8961627646758066,
 0.045394596821666355,0.7855205973548359,0.428845833432548,0.5359614725828589,
 0.11170986976417296,0.0929112445360416,0.33999605189153215,
 0.5022078223235279,0.6832338127626246,0.22021800500137123,
 0.43636232794801566,0.10312758997789817,0.8245763164903621,
 0.36145014877051196,0.7867245617209587,0.28766014920866356,
 0.17049136140801613,0.8695625819989864,0.6345738053845318,
 0.16432135754158994,0.7871098452530932,0.3485901340391371,
 0.40635089269451763,0.8592135045308503,0.19944065391980514,
 0.1377060362609842,0.7048549519450283,0.18068164978079548,0.459776891062485,
 0.6881164778259726,0.8073795831695829,0.4849789016309758,0.9792101797300855,
 0.18245178888379654,0.012313431846553313,0.6565445660004309,
 0.5559577807307239,0.3261000507521753,0.4410839286640309,0.46898896841752846,
 0.3963282261397182,0.4139951608187239,0.45805194156039797,0.6317574289009803,
 0.2091059930378507,0.31903792006363496,0.524250849255748,0.41122057917436683,
 0.9602801596835552,0.1924853351078908,0.8960131288729263,0.9084725635810112,
 0.7803891127021447,0.32651989656813885,0.8483120212751516,0.6477272450863956,
 0.7693845724223044,0.2882629251840966,0.3475051541497698,0.8681687003669715,
 0.09372681571915864,0.3011128137819994,0.1608796891103219,0.314354021128638,
 0.9340076645782049,0.6928111305009389,0.6453029227039798,0.23105943069705326,
 0.6886556147139117,0.9919805824618343,0.40233525347440524,
 0.10298274418190079,0.6947946498161371,0.007218874950124965,
 0.021413698949847015,0.8013631565613364,0.9769048926487955,
 0.08435208308393083,0.5819666000626872,0.3831122202861962,0.2360217066851188,
 0.6715973275694411,0.1665199941360509,0.20835148183993057,0.6811831244191706,
 0.738090219765643,0.873437562629153,0.4893658929633976,0.28087534959778016,
 0.9687157056848826,0.2941542845938928,0.3091214887402338,0.4597710649010627,
 0.6350531114513125,0.6424281500597753,0.20254137848734288,0.5269113995628847,
 0.2773617862281714,0.12196973072761896,0.23361067421177906,
 0.33571029506940886,0.5981765506814027,0.20738292177069573,
 0.18240100496563738,0.36865496482536986,0.13311487179220083,
 0.1530549947310773,0.6731880348437449,0.4876516347174238,0.8875007755128529,
 0.7174952200437179,0.5182068015702472,0.07884176068158488,0.7128915975895556,
 0.5366849360332333,0.002167688046395666,0.7702116659221051,
 0.2829215021638256,0.3625156271790484,0.9245944190563364,0.8749340200415533,
 0.532636947977815,0.5218561496965971,0.04620363803842087,0.7207017217962226,
 0.6382264524073079,0.2586436075703905,0.9104042318519759,0.4437761690209645,
 0.46087286663153404,0.9590372818363803,0.8847485464772296,0.8142745742580757,
 0.41695462981732956,0.2146355476835632,0.07580911248454258,
 0.8222161784115767,0.7612884634779767,0.5820623875252074,
 0.047036393469297555,0.5616271419865487,0.028228929199194952,
 0.8708351676199415,0.8612520976816151,0.8255985307772985,0.23551908339840688,
 0.9793342253506983,0.1484847835856251,0.5759568252900409,0.9045969713076568,
 0.20019568760108852,0.0838117875500568,0.36491761242485954,
 0.9817274727501993,0.8416870934498168,0.436108802856256,0.6420803877780534,
 0.6690692808084064,0.8413101942994974,0.0992870468706274,
 0.021789673494634898,0.12006379968404368,0.3815994378740786,
 0.7913602257104455,0.9062778136637534,0.24654686989150287,0.4629043534549645,
 0.7816466466584,0.04939867815438592,0.5133839112437232,0.5869839702303146,
 0.0692539935112616,0.5583319815424581,0.14768251988484216,0.6411280229323838,
 0.017686243899372234,0.3060687637719649,0.060794636529821866,
 0.2179725726420515,0.7311848808304406,0.7460503097521731,0.8133315188912283,
 0.11008582603138484,0.09118397896113573,0.6986235060740116,
 0.4592486773940374,0.8768941366903625,0.11470966048439968,
 0.38081922643242194,0.14059740280209498,0.7247510555776232,
 0.1841180588099599,0.6599348151514053,0.6095055638737414,0.2672450693928107,
 0.6489484181391592,0.5037203824648038,0.7819816997650455,0.7831264710730668,
 0.6576120216490718,0.45572836152580676,0.573136922021335,0.40729961282322025,
 0.7696719962940693,0.5991677430083678,0.3512847949802279,0.8903986856480063,
 0.9411547598513047,0.32315919455439523,0.7805293279524758,0.5087892015152907,
 0.20370189042680464,0.5241843622592687,0.06143862706542835,
 0.6226675683248803,0.9054434235261797,0.48304608840372,0.8644452262667603,
 0.7809333400796161,0.37903915323135273,0.18390987828386185,
 0.28346925348048324,0.8588875101495538,0.36029960035207553,
 0.45061551059365956,0.03872011110668405,0.9741639780908052,
 0.42038840580409753,0.42319747858621104,0.3895541254878975,
 0.3730011387944432,0.27256544762767687,0.8868381065201877,
 0.34605751970441534,0.5078819654153033,0.9796176172856478,0.9618020323499612,
 0.1055929556225702,0.953237178434243,0.7318558118796422,0.23178044272942722,
 0.18539940725981918,0.3675262219932901,0.8821516972337465,0.738308431256878,
 0.22097502012727266,0.9586511590929031,0.11826924144548612,
 0.6344375310784808,0.9557636649590615,0.8919228515496849,0.3920651215233857,
 0.6159125540613974,0.9284828739428689,0.2815080609144489,0.8788593433809094,
 0.5120752501028987,0.11772590082085332,0.13576868221554927,
 0.04554262471953219,0.5966449868333747,0.007060857958447686,
 0.7970072918163418,0.5692928298784544,0.9642742747572434,0.78932448538642,
 0.857621299637606,0.038697476362915095,0.9432290084849961,
 0.016388839990670645,0.6662488394262772,0.6805466618315557,
 0.8301885599986456,0.48808811458302104,0.4306043428115942,0.9140241395231254,
 0.8723992232746239,0.11602685251842082,0.2656273136822034,0.283024735469822,
 0.696625510014027,0.7102186848876418,0.8719401985975367,0.48305556536990324,
 0.8171516676973771,0.2573869056306468,0.22861322710775067,
 0.014161777807238374,0.34942067124214204,0.3753983658258211,
 0.5883708573018881,0.3589392330497758,0.9512143110124338,0.21182145177719092,
 0.6780650473398396,0.005994528516715336,0.6116500796878627,
 0.6576370657601892,0.9119079566245615,0.07376071558905706,
 0.01916378084197956,0.49545871024387644,0.8234473441633248,
 0.5483760816469326,0.8068794997951906,0.9642702567476699,0.2345167278796162,
 0.5823124402772673,0.9713813332861247,0.43436276688583064,
 0.44986274500384915,0.19302850127207094,0.9347582433794202,
 0.6538135000471147,0.6488477857975731,0.08666269143943817,0.0969531598065736,
 0.987645570349493,0.6128334081270893,0.1974686030773647,0.0657743196533499,
 0.9808431300083502,0.5847477436159745,0.23903886080121983,
 0.04520632780870237,0.6621432768419899,0.27869432649053105,
 0.31871481260974355,0.4263971444281749,0.8241683794206786,0.6641642218292186,
 0.15048783553174672,0.9183453675664858,0.17608678813997491,
 0.8937341931111028,0.7011627994127617,0.10218606967841737,
 0.06945655179603905,0.38750466848685705,0.2539400603438,0.1138596188839216,
 0.5512542985201411,0.31714854278019544,0.0843719956052893,
 0.09589731016558045,0.5505597226416388,0.11049721469874141,
 0.9948910633587085,0.6103897531815254,0.024496110633629553,
 0.7595173835163519,0.18978876325714122,0.685202053247793,0.31955500932667436,
 0.5591503284751925,0.107259207358955,0.38442888127960884,0.09999027826407336,
 0.06334707659605598,0.6363121952562961,0.5622288325397834,
 0.08974599590062649,0.5810728482314123,0.4478803669930812,
 0.025390282236173234,0.9745730765445325,0.19006511345375132,
 0.47122168283902965,0.831996758358456,0.40544904467810017,
 0.34869757039324245,0.16529939177632313,0.8885604430634385,
 0.14006612891745907,0.8076511072097897,0.38492309411843817,
 0.20731721673850523,0.6660296072405689,0.72603830567388,0.6884908637423,
 0.8855057355362151,0.03753518305631065,0.9199017610715472,
 0.12025990038465373,0.8386320836654562,0.6766416767370363,0.5244427229984615,
 0.5357608926415206,0.47667717726587855,0.5157841673242236,0.8710664591842913,
 0.3199528068477978,0.669552345147698,0.40259464393521405,0.1585267996103621,
 0.9189224722064582,0.12758971446016731,0.5243506833810095,
 0.06086122363127411,0.15868723679128627,0.11270069478827827,
 0.7757702047943588,0.5442620124297962,0.4762015923312014,0.4839561048059524,
 0.40244479758459184,0.4301628467643348,0.43116565008417485,
 0.4691647725939694,0.6346779799336254,0.7927876665771026,0.44825157207323774,
 0.3281251453834614,0.7985468491310077,0.07604750023884788,
 0.10390627886781134,0.746167229752751,0.8322604129684545,0.30942863098272677,
 0.806252889462159,0.004602384607512944,0.8259420129627908,0.8363032145193134,
 0.545103825138646,0.43916776432012994,0.19002246375008747,0.3021698271500516,
 0.08844452357110122,0.23595396787567968,0.9491077144724385,
 0.8906815333234566,0.10031508226149966,0.7764302244432574,
 0.47161990158748957,0.5585812713860847,0.4288317138943746,
 0.38331023012702925,0.24803146386845554,0.28659850853625435,
 0.8501102100412409,0.9496225704815807,0.6202066315578604,0.06192872133379912,
 0.15452968052182126,0.0835002365271238,0.12604922682918074,
 0.07471382654345804,0.2673753098160532,0.9529186343093436,
 0.35373937393669574,0.09314723718602869,0.7802200016622469,
 0.8935184834079875,0.9527377033993344,0.5051392874504157,0.10717835904684947,
 0.8018391620086305,0.07092307700699008,0.09579241559948304,
 0.4652335860809925,0.09590430798889626,0.35088356714120117,
 0.9609151200077038,0.02151438692602603,0.44139784704426965,
 0.28597740423015405,0.3469858481772716,0.34367508447605455,0.714617039650155,
 0.16120841004735087,0.04934901868592079,0.7859250324471553,
 0.24865529355076033,0.5211126422497703,0.44735701719207743,
 0.5627683127038214,0.259422151227497,0.5575828192969879,0.049904477092909016,
 0.9346282151487612,0.9788711958184844,0.2394117024548188,0.03833760849496681,
 0.29876379761762606,0.7413536115962992,0.8468235700122798,0.8298183901926134,
 0.8904484752927333,0.042183461506217435,0.36032737711488183,
 0.03355960487933374,0.8972197167738478,0.2953122655053846,0.4165848197560782,
 0.5825193664121178,0.13882453895265778,0.2882182208874894,0.8020092002639526,
 0.6509289232175672,0.6646680908074807,0.5444335477933434,0.09965394387861393,
 0.4651471941744364,0.5340893557537355,0.10894110622544995,0.5540331272600083,
 0.8537661193023618,0.4491669891870318,0.0885313312239876,0.05132992952810023,
 0.17045987760724113,0.12435943214299727,0.8202521642972855,
 0.17499087670665459,0.7042107118233258,0.370930130239015,0.23575042653268152,
 0.7984276864452051,0.6123334278135261,0.17211741162036376,0.8154924796477205,
 0.2892912174474531,0.6795097984084715,0.8891929490162459,0.4602301063683746,
 0.28114300162774275,0.5889232933703465,0.8557079684888675,0.5763169077002083,
 0.32883926066775016,0.36473433391870624,0.10542574500756896,
 0.1048647886373395,0.33307248915604926,0.6015374642725541,
 0.09880803206793665,0.7271222757970701,0.4256905513241088,0.5120333037881418,
 0.25886004468984547,0.32489011388336964,0.2513981533992724,
 0.3248931047542891,0.46173023906133037,0.5475528973301012,0.4036896501652891,
 0.4175195204790286,0.13096821904127176,0.36425924371464,0.02596869244609934,
 0.6236163067885335,0.1357535669412384,0.7950436640431785,0.8551401117951579,
 0.18405289325132534,0.828376255095796,0.6978583271979307,0.8679495266687384,
 0.24090729040537584,0.8156000356103107,0.9896331855725721,0.5041456202729301,
 0.23703663384235363,0.4384350279644377,0.01741019396174137,
 0.1347821600620731,0.2084287281539825,0.5153177011110192,0.9823628723297021,
 0.5157606061360044,0.2714078584413754,0.05273910231638301,0.3991308804327386,
 0.031487514162997,0.06524219047598578,0.2272600180791473,0.6110443968995047,
 0.9586291726786722,0.07583923385310054,0.5504010775908509,
 0.47125174006623305,0.8380889856075868,0.6012258472295758,0.6030132561366013,
 0.19268116162443483,0.8533143897454709,0.4169000339269531,
 0.39747209435802267,0.48829947289828446,0.16441119313719188,
 0.8361375218490761,0.7180667177521037,0.8758185965300518,0.07892094090688928,
 0.5028855186946687,0.08839228094940854,0.6732638851571531,
 0.09385128781534802,0.22422236932536144,0.3906828992417558,
 0.9113572180803531,0.8232021979623109,0.7604846643925702,0.7948773772339425,
 0.8928702420133064,0.9599249469337617,0.02226697630145469,0.8312748644597012,
 0.9619952574870605,0.48788367417922385,0.9215283384337255,
 0.11925657403472834,0.23851559966901126,0.0631827312132549,
 0.25761078787919445,0.22345678171668093,0.9461250783945274,
 0.27068745857966037,0.9263689212029791,0.2823471982141632,
 0.38514829668207107,0.9474550531627196,0.8631915916568517,0.8377383568656817,
 0.6715680007929608,0.772375345169604,0.33379428012202483,0.9141819924610763,
 0.9797270010136885,0.7175510919088949,0.5122161750186468,0.5800511072614606,
 0.5010176338292975,0.9945733120439223,0.038948377275674795,
 0.7017643648165847,0.31804470769015714,0.21356771509606454,
 0.6685704797899403,0.5882668461675714,0.5177340566956962,0.6932810748633224,
 0.3785459914408599,0.47837035355096236,0.0578716279799415,0.1321109439457998,
 0.5377758848673057,0.20822552814145323,0.7557879983671235,
 0.030350239422273995,0.5688099429043726,0.14459922128684488,
 0.3822853971920258,0.9558455092848849,0.7757548195437539,0.8589071190281754,
 0.9658573051788635,0.5908717820735391,0.8995791397306703,0.4504742160356776,
 0.20061346731719287,0.9512526564088332,0.11461922197590269,
 0.07166938570211812,0.6966419591451608,0.9790956296276241,
 0.45160509125893167,0.32119617026192093,0.5726027192297642,
 0.05157464073999474,0.8098686654642716,0.3732449670983988,0.9203904146439702,
 0.10187227733137649,0.906995403281891,0.9011060362378205,0.24693441033556907,
 0.09371785943177235,0.8585368246292995,0.15991043938263438,
 0.9130511151269273,0.9645209960894497,0.7807463671258086,0.19954958216008367,
 0.09137047967553658,0.532753623116981,0.4753892137419494,0.14787310559743616,
 0.7510877532868475,0.7524654827551636,0.5327562514024475,0.8119063169590587,
 0.4006878041831051,0.0556187313653993,0.023339330378858603,
 0.19642602540179077,0.34741244515056313,0.2739717406383625,
 0.22731585626635153,0.6631827644196149,0.2888774493282382,0.4685970290902187,
 0.3992605518178566,0.8396418894008187,0.8689315144271703,0.7440614332892643,
 0.31265908061486436,0.3501239003796003,0.10786534825254301,
 0.7805418603824245,0.2459541089604066,0.7231052455178222,0.5727135708869926,
 0.08289237648687764,0.9385514903063604,0.8739209233285586,
 0.25292154346368567,0.8018567342083042,0.5326223562331568,0.5853057483959346,
 0.2155041874113709,0.7000781609894973,0.16509977521072627,0.3022211260666895,
 0.6010869628262602,0.6629050019907815,0.6669667209701987,0.7464902901990899,
 0.15045062405134169,0.8720619007546015,0.5333512877753298,0.8879549737796895,
 0.3824051869720695,0.5885991008090049,0.53692038108608,0.6245305056920974,
 0.4151515577662961,0.13211763449959724,0.8778976867208275,0.522776098609932,
 0.3753284390587106,0.785272862425144,0.5372061640094921,0.6002526846427496,
 0.916720901091276,0.5108342742238678,0.3279429906874537,0.7189332341364201,
 0.31014046000325823,0.7047513284404238,0.3672081322221694,0.1723477855389025,
 0.09193002550227347,0.958092901286425,0.8981919595856578,0.6134512694321438,
 0.35821733294913305,0.2987137353017657,0.013141953017859143,
 0.4868242223116561,0.07848958736763778,0.08368532150278463,
 0.10652227640799383,0.16140223945279863,0.2354431434285189,
 0.5409789666037808,0.4342196871510148,0.4084946903936181,0.2638133123685308,
 0.39296788211950495,0.30656468697235506,0.38326881350275466,
 0.2691043802108878,0.2996252585654974,0.16116779096936518,
 0.41750551100026456,0.1264482758493719,0.5959317380406514,
 0.33455412943147467,0.3351552683513088,0.5428951081404939,0.9939351938737736,
 0.4393814211322523,0.539756233502359,0.896604063892917,0.6801818455219621,
 0.5339184028106594,0.783206284473149,0.0763677502020732,0.05982477220207072,
 0.5603337886106797,0.7387960612216672,0.9745372615774884,0.22687144341910426,
 0.45843162693497075,0.9949759811130025,0.41757822435247893,
 0.10678967523388272,0.6294601567798876,0.3002821847196233,
 0.010687255913255034,0.722312424286255,0.3569759870631243,0.6341937610462781,
 0.6739828908633199,0.9942077296886762,0.6885122224185365,0.18270426786645144,
 0.9386285043953596,0.22050114155800926,0.32834826405774176,
 0.6130595071812388,0.4829386941620065,0.357814965314579,0.5935052951983857,
 0.45593776448730194,0.7122628398430619,0.2422823043025002,
 0.13565374018994325,0.9507747149008794,0.79395473713868,0.7680940765751126,
 0.24122175974725868,0.17273738457984722,0.06381963186977435,
 0.6628362119427633,0.4636636906575017,0.8485996083272722,0.07710362484746869,
 0.05905250794154737,0.7097638710369716,0.26629443368792416,
 0.7476161642402483,0.06783199255153183,0.5107284341261695,0.5102638676743401,
 0.7201373838357794,0.4519352648397975,0.2126966234222687,0.6613772107524297,
 0.6762203501186091,0.16547194301633572,0.10525550747943324,
 0.8359286272695754,0.5808665620128695,0.317310384831232,0.3483879254105069,
 0.5315986522472345,0.81165097706568,0.07410140197975135,0.03775287166842467,
 0.39015570629441043,0.831526164017383,0.8164731593231455,0.5818268198905715,
 0.9809882871292207,0.256803638979174,0.7248098294421665,0.48588581545424037,
 0.44349938164943115,0.31420452705646573,0.2250090676604637,
 0.7372953759866402,0.0075443600730616955,0.84862214048619,0.6298347281943141,
 0.5461540419596318,0.6187904383811883,0.1876716139061564,0.6719610115315577,
 0.17206199179493098,0.008759192068900856,0.763794552702767,
 0.7260153831485877,0.5403580436466865,0.7196493620908803,0.3479094422626946,
 0.3265975469558582,0.22947681482391835,0.032520447615372094,
 0.9714299392617398,0.8744305710707168,0.11847374530508237,0.3064631473470726,
 0.26067680472552524,0.2516117447345748,0.3168756568884089,
 0.42563673569774474,0.5503900430214493,0.6756847761815301,0.5918015882126227,
 0.2959656366579696,0.5323695148410631,0.9936113439031136,0.18598791468618525,
 0.9447865538728579,0.583028143075615,0.35606425854609225,0.1527384035315944,
 0.1389317199715353,0.48214703242871015,0.45341779865681653,
 0.6627576113224015,0.018529080523883357,0.4383683400621965,
 0.27229752691060194,0.2240257888033368,0.4525669470608986,0.8521783872035191,
 0.14496394533663137,0.4205170390501056,0.4460002515911766,0.88274575590218,
 0.28189241034008383,0.9798087417052953,0.4494418006168076,
 0.013081764822625463,0.4589015791823591,0.6250836236902921,
 0.6208323391558515,0.8418362867478546,0.42323374367290667,0.1788323114706415,
 0.6443241610082532,0.4472330342745454,0.46669282337008633,0.553945035649162,
 0.3374789798548349,0.5960189766565458,0.20731233764096768,0.3701645670591207,
 0.11307258540460341,0.7053826229268063,0.22791644984727255,
 0.07920329566714768,0.20603268132839603,0.9729555702798568,
 0.5378487074568638,0.023060747347186994,0.7638488918468496,
 0.02114371972518936,0.5195166188814136,0.22520077031201535,
 0.30404355752897094,0.7774006690966235,0.33204560322380916,
 0.8261835558498941,0.6374512869773503,0.233113244121527,0.4003682957260575,
 0.41605743618107527,0.043487331789755945,0.9403022470084718,
 0.28024379264716914,0.04336957649947548,0.10492693755060356,
 0.9018031352492757,0.7708228705716322,0.059483672104788665,
 0.3574108212667897,0.06513303321811348,0.49069750048269445,
 0.1876897194209417,0.7833064541730138,0.3517098372366355,0.15118726689323003,
 0.3924335960158084,0.09198526671509122,0.5862877537549475,0.5327121545223482,
 0.9655187688964911,0.4249637836141854,0.21574069647000893,
 0.0023683208988548587,0.5687584263204498,0.18765337097412704,0.6065809763182,
 0.22346319241623236,0.17958166301069278,0.5149276750644157,
 0.49919101034797064,0.2680525588275424,0.15371492123373465,
 0.5715050484759236,0.07082929726528497,0.20019618357678526,
 0.22039430853490682,0.31279486074982277,0.0035162293319710436,
 0.13090539862220218,0.8411188159562858,0.1650155181770745,0.9172395385568843,
 0.751703627828917,0.09740813191064956,0.7845326737181691,0.48134202520837555,
 0.5182472762962597,0.03382562138137901,0.674740521044707,0.1627565970152548,
 0.6494197945223241,0.845104943024292,0.34288194869027333,0.2956418860813358,
 0.26760840807605246,0.1347751291370355,0.018932238984997185,
 0.4781443014406258,0.8849962874694416,0.29817694245835713,0.7972038167428587,
 0.16980608601741531,0.5653833873003936,0.7504492132875792,0.6226966667533728,
 0.4133760608313286,0.19683766388322532,0.5656025303490697,
 0.0024134645360659457,0.9466285725713203,0.17274705324389417,
 0.7502361085707256,0.34470314914609135,0.09572435043562033,
 0.4581051569412148,0.23065098292896835,0.7847378843933045,0.6898103520728527,
 0.7034374805941929,0.36587464013127435,0.9128215111052411,0.616452615232103,
 0.35851262114805893,0.36641449799538506,0.6703892961658577,
 0.20215919783324576,0.9010131600464837,0.3262288762511779,
 0.07627307612571776,0.5727736548769577,0.052835817084992964,
 0.26165104544529993,4.0218844151251876e-4,0.5538366543180345,
 0.9411611707647198,0.48915355954395345,0.6044551679796075,0.5898864115436784,
 0.24617606821076954,0.6785738673080168,0.1584700104387915,0.9355946919632803,
 0.842444159291777,0.16455589175777208,0.4239359005626082,
 0.031406965755545246,0.07811583905379293,0.8650671064600147,
 0.39354987563662736,0.11138279759574532,0.8924899968099688,
 0.14463050738748218,0.4631388103199211,0.8729954580280992,0.2337089041926086,
 0.6478666979205461,0.31519240959209616,0.023421151691810804,
 0.8032167258528804,0.6665383549003888,0.13375418032768094,0.825094278518612,
 0.9430119747113945,0.4572125218261447,0.799765193807181,0.31999358677286593,
 0.8789420091336938,0.3973567317825517,0.22396240722867422,
 0.19379066988907578,0.11977415348767662,0.8264396495515353,
 0.00782514780348642,0.07779366238258978,0.4988690416608137,
 0.5447715372717522,0.9472877256187812,0.3242995396310645,0.18382609419084595,
 0.29974928100615383,0.7543692200373531,0.9768990453357054,0.519798399453452,
 0.5834236835078784,0.8359199928294577,0.039012363375983794,
 0.9392706828144516,0.04949609002445654,0.038748898823825595,
 0.13075699640595784,0.5127739173814645,0.160020987241275,0.2757909701474901,
 0.3799488756867603,0.002471109069847599,0.35361971387008606,
 0.37570580977988555,0.30137906967703365,0.8431189603579432,
 0.040143310370419294,0.5799337476039654,0.07870830986742228,
 0.3933010862419407,0.8840732188363719,0.26459251776669623,0.9050061226064265,
 0.29345001642656665,0.4513733861073006,0.9983147390799657,
 0.21937712566270784,0.20321075462084937,0.053182428421396644,
 0.6067875218240975,0.35338012520267725,0.31500841259343537,
 0.6272214507514994,0.2425004228086648,0.44692019800301375,0.5920258337314913,
 0.0037561065401811033,0.6283662552855666,0.9236981837165037,
 0.5538718172929662,0.45546568680490285,0.4956448962450526,0.4689785259068593,
 0.7211707624361927,0.24204232071963916,0.8605350833857053,0.973899637992088,
 0.8849502815868302,0.6625914384293611,0.2890870344527805,0.8028610123843276,
 0.194071774836011,0.7694957954994337,0.9150388058657816,0.5105005800096769,
 0.36392609146259214,0.4952523899606083,0.14755951229813358,
 0.29146776132001473,0.9660467710520791,0.5395344619837432,0.999464841375733,
 0.5087483996104003,0.2884614332178349,0.03721556849854335,0.3408839115084987,
 0.33171266981904357,0.6452852370127711,0.19185057636349068,0.568274747072556,
 0.05573878530020571,0.8205104729804823,0.41147543501093586,
 0.7553377887508835,0.040993800130256663,0.11387264816547293,
 0.1402254485985156,0.4919135400915088,0.27811183529270145,0.9066218778604831,
 0.17920660188069282,0.8328784859147615,0.7239060520700047,0.258182134690086,
 0.8492275369910757,0.04491250612912756,0.8550963623069374,0.1348287463115565,
 0.9644100190028086,0.3801600929808907,0.8473640329204422,0.31853529707157113,
 0.7283733639427141,0.6123928852211007,0.9142118877275139,0.7221869410571145,
 0.8975206102553293,0.07564478434414768,0.6042056955150442,0.9580560089105772,
 0.9325491269115501,0.8786946438412108,0.5347794734034729,0.8320189454715287,
 0.22855088224522724,0.594449755948313,0.7042857776825766,0.7993886202833475,
 0.874394803170641,0.9278308528485025,0.4639102655506798,0.01034463454657697,
 0.4241893063554135,0.857724287125013,0.46606675335470216,0.04965243994190183,
 0.9201297576778088,0.28112027373289683,0.21082030351407766,
 0.47874331691497785,0.5982242866706474,0.4596814810376805,0.1402902564160805,
 0.9635211825211626,0.9851955118757538,0.7844964722265958,0.44270789167183255,
 0.2670084550994656,0.6260216214168697,0.7727040356446926,0.42583359712372104,
 0.05099230233275687,0.981179924197438,0.14821506967300324,0.9978381068096354,
 0.08636043980072317,0.34431694088348275,0.8438181324901736,
 0.44480942644102206,0.9568641076773495,0.4888034990014152,0.4256773178049583,
 0.45931553158433924,0.3867367097141948,0.9087953913571112,0.9863562051548261,
 0.9936823459945349,0.2589660662334783,0.5826944223845375,0.7965226147442926,
 0.5138217142098909,0.22465553510243585,0.5276519317331722,0.6454933950994902,
 0.042876635373098604,0.6964044544863873,0.6177374652723965,
 0.5045349215626423,0.5340086071859415,0.9941390833376905,0.5078499350896186,
 0.7053177722055712,0.07467375336906112,0.9078314803951514,
 0.29890254821560813,0.27804900273507294,0.5018531475835086,
 0.36255209806592825,0.9038453621601297,0.4289058311374204,0.5505695117472136,
 0.270805201878379,0.30633280227205506,0.9889466457214361,0.8462696044984798,
 0.22673231441968222,0.07265133466615403,0.22213492769762266,
 0.8627438780122322,0.7147397810146371,0.26918629415535966,0.8057340608808674,
 0.9763027416568366,0.3757568073501695,0.1938751996514333,0.4322427093859492,
 0.7535451120431078,0.40748247976833496,0.5997117463613112,0.7719394277245292,
 0.572736941554067,0.7301062504718271,0.36442290007057365,0.8704636756634403,
 0.6349253259932367,0.6735119447413016,0.5506306011446996,0.8901112930632905,
 0.7289901084776441,0.812334244205937,0.6726176580414003,0.18690487533178757,
 0.1986929704615772,0.2649714427775731,0.6647953379645916,0.472833733599507,
 0.866611406981719,0.8789119148597975,0.10304465126368623,0.15896642582195608,
 0.11892586492596613,0.8848176158009586,0.9033434397907796,0.8082342183790905,
 0.2483116878685866,0.47160214356532704,0.3849719522943902,0.3800373135024747,
 0.021808845062903037,0.4053392337799746,0.7617228210900144,
 0.7343659972241533,0.245437359919923,0.8518569013554432,0.8805565143233138,
 0.016505797594582372,0.9457957361153091,0.3363819402842627,
 0.4519156467250792,0.4274685419673063,0.9484166070612163,0.9455786033234678,
 0.8862762222463415,0.5398924180461279,0.8879901643397157,0.5470204454746237,
 0.5763618396157589,0.9409884485110375,0.7145218054835335,0.9219860230674034,
 0.23747638038662933,0.7718609228137968,0.9521515243665413,
 0.16555299784651534,0.7531113873384674,0.7373899539683443,0.6492770759518004,
 0.044348896190749776,0.12940399362540678,0.862287884566285,
 0.5549821101062247,0.38016618809959357,0.9898104327667441,
 0.06965528283871047,0.2515392970349942,0.02950913438781899,
 0.014016509542134181,0.49985791352739917,0.7816835469692075,
 0.5845642294921461,0.38631353347151764,0.575894514760675,
 0.036416105730523674,0.4299716381773423,0.5247007661072178,
 0.3861892469121505,0.3906795510596077,0.5593696994825171,0.5516076095139815,
 0.7391248548923058,0.6234668722868255,0.5183469079992672,0.06802055547623531,
 0.8497159261977218,0.9584950422248995,0.9615783987072519,0.9166840914689681,
 0.9297757312358396,0.12182745093927294,0.9790853956512536,0.1968558570454264,
 0.2957230721032754,0.1085867000681976,0.3801569924683015,0.4292828603711235,
 0.5592238737384234,0.49305989482125867,0.6902023126050564,
 0.06370305082366146,0.1377457345280041,0.7995454368109762,
 0.11527668032785787,0.9076867526506359,0.8527024377681377,0.2412888087601226,
 0.9654062846714693,0.06739322655818136,0.7828518972432106,0.5270557511247049,
 0.9291736807165657,0.16226231144531167,0.572213327944981,0.5681647046941091,
 0.13242651419235774,0.6359547692875389,0.7958983567640728,0.7842467923371828,
 0.3485972349756147,0.2718980240613179,0.10704546884070398,
 0.25120507663172187,0.5517444166274736,0.1463670916885732,0.758581263887251,
 0.23063488750383787,0.531197914788996,0.14949785083583722,0.9876638903502281,
 0.36586447920974163,0.10213677544075062,0.21207416220646713,
 0.36377039952346346,0.22610727519366702,0.24667695722926453,
 0.14392187155233305,0.5589889769224556,0.0697862473537082,0.8468283612877059,
 0.1507534408788308,0.5237052816186273,0.41756321564915866,0.6234023327339667,
 0.8730949677324594,0.04236802396326811,0.21385006000844797,
 0.9755115812848686,0.06358665977825606,0.6532539992046464,0.8903857416148426,
 0.4274211415421041,0.6239864842093616,0.7387042275456244,0.5328458226506589,
 0.8345198127239288,0.6422982446948508,0.9047165617088637,0.8520861113712731,
 0.8614312901121753,0.9852261253119318,0.21599343959333972,0.6995039743395782,
 0.4975017513621791,0.6980035348679186,0.5708590788544546,0.44687663326749927,
 0.27133828557971007,0.3323057571762159,0.14657457795402928,
 0.3673661554673291,0.04959823086743165,0.31369717525599483,
 0.7916988870442878,0.5959059606051874,0.9262019462071798,0.7113210819595491,
 0.2865786332261515,0.43492905716100483,0.5090999121725094,0.527234441421977,
 0.3479685348134339,0.6251208957821103,0.885634751177585,0.3269789276562256,
 0.0032797661191291683,0.46677993503723614,0.9623332905208221,
 0.2468907214908832,0.5343240058901626,0.6212046457486106,0.9031952290607416,
 0.7830132244328779,0.15163359754354755,0.4332603191872191,0.6039435747215167,
 0.9453650979975748,0.3260454727161758,0.6428126599577361,0.9826671951614426,
 0.24766302593828604,0.7605335044349748,0.29997515824367993,0.868584113241848,
 0.6198808008392866,0.5562527177227536,0.7711764445125612,0.39849454573118104,
 0.13602477422999137,0.5202852718441219,0.6132841927837509,0.8555250675468795,
 0.3731673653663028,0.2962787562108734,0.9345823335654387,0.4735526111029703,
 0.11923686243057019,0.008473107497241616,0.21435157346316225,
 0.999421864778999,0.2193425137225533,0.47865332975481323,0.7172543872251396,
 0.16281894163216215,0.3342516286727336,0.9138485635554194,0.8983901227675272,
 0.18342174452302684,0.31299511662813295,0.9465457318965688,
 0.7399983774344681,0.8288262103247129,0.8638983785911102,0.31642176114541765,
 0.841056147874235,0.34626816588154574,0.9642265102889628,0.18762042346854557,
 0.6463079987913714,0.0589227927165803,0.9087050629112577,0.9174674360659923,
 0.5197762436179394,0.4628740196188931,0.46950203664078083,0.9508629788880832,
 0.5148761823368887,0.846276861629509,0.6733625890584733,0.043637892266819245,
 0.530225368790643,0.45257800707605234,0.042087819552454,0.2742465364709812,
 0.8414162069245285,0.1909786535549569,0.4211173995528452,0.40772353893204216,
 0.2742310793009076,0.25415581166768975,0.7311943586117049,0.8143023317634048,
 0.26358595281299857,0.6131376026936798,0.8185124378959284,0.9312242323754979,
 0.917887995273439,0.2298444305339522,0.08896040912847503,0.3525949607561967,
 0.8010670946285128,0.2564494347992099,0.29223183701036926,0.6328615831461303,
 0.8414015609344228,0.6107768235854284,0.632790607551498,0.660059503026242,
 0.13399692869427526,0.27003018178764526,0.9250089268846958,
 0.3077077338436184,0.37824415373583875,0.20194362015243417,0.588531048747759,
 0.9635924867601443,0.3358447852638283,0.8491426686404216,0.5849937650290967,
 0.19508834129948593,0.5279765330285495,0.9438822228638947,0.5357283973251411,
 0.16491830123284545,0.36265119771825904,0.14293380457272598,
 0.2695453173041761,0.9217597770557464,0.638958417626305,0.5401799766484414,
 0.3176852884803726,0.3056695553508195,0.45046213473228736,0.3578250883317178,
 0.760390738228191,0.1335122930347804,0.7749992474148487,0.08476422739564748,
 0.06469456419799025,0.8009376273972825,0.6207426583918118,
 0.06645824850675985,0.21366973364086062,0.8411610868018137,
 0.2525296370549843,0.6887114512226348,0.12069255490016606,0.791992455392509,
 0.4557027696625541,0.38948425711238077,0.2865692888064061,0.9929382729520309,
 0.12892081432923308,0.9637471233157071,0.09321883948005827,
 0.40977695362339506,0.3070252137465941,0.00871784126865649,
 0.07107428072001543,0.7843103998018617,0.05647784997012795,
 0.5804898478884525,0.2717908347657052,0.8666069616197696,0.40152174552596065,
 0.1516454276636524,0.7338204060236949,0.9001614291043691,
 0.044970554456016276,0.19046117629146153,0.41494181756661896,
 0.6927390980326373,0.8027544460705488,0.9230446061297852,0.9846440597036913,
 0.0481094602749661,0.12102531159763363,0.8107703901254635,0.605094909123378,
 0.38797942229877214,0.27130902401957313,0.956324222987828,0.7171296295147863,
 0.3637676513549486,0.8549095072303585,0.7298324011413246,0.5066263436928076,
 0.26404654130340177,0.04952969050814804,0.9481480578301444,
 0.009010574558834805,0.7483976343797982,0.7927704135825445,
 0.4120345546668944,0.916168466312715,0.6548727116815889,0.5947656510297631,
 0.31474603060837547,0.31940157949291226,0.154628059396275,0.3933725106416808,
 0.7419496815106829,0.8986468166280999,0.2993755900493593,0.9099698888238196,
 0.549484422596824,0.8526971551113134,0.6157540456006199,0.4840702628071213,
 0.7203664737367776,0.3464296752413075,0.7897048809473503,0.5323267617872656,
 0.20755732384080083,0.2784270896172636,0.6957508423495261,0.9710062076418283,
 0.28932410840966094,0.5748868310924004,0.12321828703518156,
 0.12436944134361771,0.636312262933588,0.5277204127466081,0.4140786134743819,
 0.996800510191652,0.8978824571971081,0.18582209500257096,0.7579811936221633,
 0.291477523388858,0.5345639553100727,0.9371040468949708,0.0549552995807997,
 0.33108288182317125,0.09414963820308409,0.36766563810195496,
 0.2616228503400049,0.7301006325645176,0.3281533702503646,0.854373752973111,
 0.5127753955119893,0.3142591921680846,0.5586817274224996,0.9492355965057273,
 0.18288699775283523,0.003891450951518105,0.42929535181646516,
 0.464901385758383,0.3324064366644297,0.5603007351227574,0.4954227318362344,
 0.3139038669193214,0.25612973121206106,0.310204065686589,0.691707150597224,
 0.9487577731810595,0.06303980915323981,0.05224535088370896,
 0.9954773488671066,0.34943070861968717,0.9576625165932149,0.8130306040840081,
 0.9690547778019698,0.23909041879110282,0.97935587571987,0.6449281994186212,
 0.03101134804451744,0.24393600320078734,0.3753310693266634,
 0.4280781627607295,0.9327989894139764,0.34042170845233455,0.7822068520265634,
 0.1709614915697546,0.017142471092882028,0.6377177992428189,
 0.6679202310933194,0.8908192247650257,0.25400281309082373,0.4767174297539388,
 0.8497934563011901,0.15846112116971156,0.057148607670844176,
 0.5940669476559924,0.4060257413641175,0.8490565467550151,0.2049924222359325,
 0.4941534734868329,0.1770572855507152,0.12807802780212862,0.6413389242352499,
 0.05858722889552792,0.4658737741707286,0.38085401336755065,
 0.4394460995525127,0.13473375091971318,0.6915350144623837,0.5810704300999074,
 0.16366933007334838,0.13469633767176603,0.41472893454476456,
 0.9607917692380609,0.41293700766414665,0.19464526963244155,
 0.6402930399007132,0.4421446834656855,0.7916638166569033,0.9332845314382816,
 0.6606069067962074,0.7509664462903424,0.5835626964112711,0.6829212021701976,
 0.8594859684790441,0.09665945179574442,0.9001356283994092,0.5103487908587716,
 0.021983675507486256,0.15547360420871492,0.9451039899859011,
 0.9139043895307228,0.46635807642367477,0.8832376620629554,0.5329902534962317,
 0.42996762055409055,0.7438118790513306,0.3763197247257394,0.3211077212818979,
 0.3857738411014575,0.8385667764420327,0.7793771785147807,
 0.022125379001841083,0.03777869463275585,0.6407319904870444,
 0.18386504092004752,0.21290294302160728,0.8729628317140079,
 0.7042153214684334,0.9247744933337453,0.48348626486253166,0.7665692274987197,
 0.376207261803884,0.05148232038293876,0.21803004948178772,0.8893677340767174,
 0.8631081117807302,0.4368420472122825,0.33771528336258816,0.3615244091063877,
 0.8088206102304483,0.19661732415968214,0.304808581901749,0.8197100550911722,
 0.883727405622475,0.8130037483158264,0.2886335248746382,0.8745547120676576,
 0.33718503356626706,0.6250897772533633,0.6062467032516025,0.6480404787032797,
 0.9262089434791475,0.6317466841911267,0.673789099344085,0.6066364160531559,
 4.2476892009835066e-4,0.3907099226477827,0.7013186977682576,
 0.4767760018588353,0.6271477898483153,0.1193062904166986,0.02871725366726574,
 0.7040461994570125,0.36363363632350465,0.1307069008663414,0.594951950276771,
 0.00151413021979252,0.05798728553866972,0.6224343021510017,
 0.2875272804836603,0.4921623069542751,0.21023225685864055,0.7201219381989452,
 0.5150694556736587,0.47108988534438667,0.5466085914014868,
 0.44271921906379663,0.0077786743253595025,0.24529254533818134,
 0.4858827480017962,0.253551708056889,0.8865730110106879,0.46701170121689584,
 0.9260197507676242,0.49281090673695005,0.6891803100523202,0.6918269766201915,
 0.8819805800884238,0.586807911151167,0.9357390036085809,0.09886851206193292,
 0.6533752318556526,0.6252493543594185,0.3922810308338808,0.18649232925025483,
 0.3737744438251439,0.1846658903146381,0.995942810727436,0.5307034662825788,
 0.9964026255706401,0.3512628197039862,0.5858184762243296,0.4423783884582869,
 0.4439657133462773,0.9155292640438722,0.41551678298375483,0.3246292902997153,
 0.8741850218378338,0.05130230360503363,0.11275801475504088,
 0.6060305281496343,0.49045311974871586,0.9920785333677961,0.4718930896277449,
 0.6495530260771696,0.025745591943955537,0.5382116080186122,
 0.5907961487600615,0.8278566932634365,0.8410437326150353,0.20680547995654663,
 0.97050876869964,0.43129904100551997,0.45033961898740904,0.3483726900614974,
 0.4657821219231989,0.7690139394106071,3.443717518014644e-5,
 0.21333959790336676,0.6480769621928245,0.925383645873883,0.7802118460737653,
 0.9435689533133326,0.7032336107516819,0.68994416243048,0.1169599270511521,
 0.34391565268628965,0.3197740376650057,0.6441656545748601,
 0.47126531068409505,0.08785569850909025,0.6512008514511184,
 0.5155058909899197,0.025773551001960326,0.1737930711326081,
 0.9764296499961146,0.7048380099135029,0.6713575716041907,0.9626401379664419,
 0.5300741289501006,0.031046491919427432,0.7203591896133115,0.320754673478582,
 0.7667573808540661,0.7300334462721207,0.614054298809936,0.7714772737380207,
 0.09875957631975751,0.46624052122372406,0.1600856834373524,
 0.6632505904948384,0.8281690823549679,0.14265102602286817,
 0.23630583785856052,0.03682099313916132,0.6123679769490529,
 0.10855833125167591,0.3802404442111581,0.9509436048777865,0.8091353732494335,
 0.8010032352550515,0.2055502882889434,0.9529187403375752,0.4489381202093712,
 0.705113009052505,0.8832991332322455,0.5203611646156308,0.5717341334832259,
 0.6952283315340836,0.7721369962051241,0.44564829567353903,
 0.35688500540575196,0.6159063990715054,0.4576286833704094,0.529546215527257,
 0.9439291903656943,0.15680406236391908,0.29357379726764443,
 0.9067820664070502,0.4130129423524823,0.3245296457285177,0.7473146472303778,
 0.36511898352179095,0.7042286996891967,0.4286533061665557,0.4678813215495419,
 0.3514206352087814,0.7811496642603721,0.17527836173790345,0.3876811812659595,
 0.6767354348116607,0.8198325919782169,0.1104597292123426,0.8530439755814719,
 0.16124822275153028,0.977490132634594,0.36471272306876423,0.589616233207761,
 0.772990073943445,0.17495588163390485,0.5194835269700457,0.5560948687033049,
 0.45383274143905805,0.3683989600543881,0.8216071791741805,0.8906223988741576,
 0.5790133591806226,0.4573963912901722,0.12754755733788592,0.2639981457669782,
 0.005115456206889246,0.6044964602939694,0.5000409712375784,
 0.12229199620238473,0.14343997602067105,0.8018372356853174,
 0.9903290329616314,0.004196507902651581,0.8772334792269652,
 0.7340168482251467,0.9235700702418786,0.4430070742544264,0.4544335752144717,
 0.27321442771877447,0.02635117516680563,0.027286926630652908,
 0.756158223219708,0.9708564863716151,0.6308194570159521,0.6506429772624632,
 0.9303289757475115,0.2574351765432511,0.6420535988853903,0.327445305166685,
 0.5217370918498306,0.12436037978072845,0.7645813201953684,0.398989487767459,
 0.6228138470017441,0.6581385280496542,0.04705563665923551,0.2936171286203647,
 0.8607188530110126,0.11827694446419679,0.7062951339110158,
 0.30859968056015885,0.21925369298878217,0.5501523607128828,
 0.6829738810637062,0.7467338529542247,0.7252799870114235,0.9223151501581888,
 0.12451765816846727,0.31841955702862956,0.06831607115582372,
 0.5870568390726865,0.5595326763394328,0.9568481548622626,0.8520157439701872,
 0.5245559205699538,0.2635766016454242,0.2600926885218855,0.65539715297898,
 0.40363580090757767,0.9723130606343009,0.4567268218742391,
 0.10962377651977184,0.6824449320450944,0.9714999469722729,
 0.34633423390549045,0.6809451317988097,0.22289560107622042,
 0.8348000069038656,0.5202801142077236,0.781984832872071,0.2550045972279844,
 0.39100814787882743,0.9827117702836984,0.9404877585346774,0.9993921768809002,
 0.5769190632945476,0.9311250041268923,0.6539552167816675,0.10914783056694022,
 0.1479338062771277,0.2612059761883634,0.5803365345556201,0.2512167715554642,
 0.9609855838887869,0.01582436298069534,0.3160568566525015,0.1946280731937231,
 0.20614221028816848,0.7153874541660252,0.08904212751781937,
 0.09772939674550152,0.4492461092471117,0.3460687839574338,0.7245170605058041,
 0.564366418944394,0.07235928318496765,0.7900494062252053,0.3749877077233933,
 0.9158197336792897,0.7655387325703773,0.03286760230316077,0.640149379378426,
 0.8194282598311958,0.5905355783579893,0.10645753833171567,
 0.42539369489561074,0.6692232543170457,0.9630140190167613,0.4852677363840575,
 0.1317587387810738,0.9857842153410246,0.32011274702501835,0.3187133513895368,
 0.6585268960863804,0.908745176551779,0.9489606705696745,0.3925514382812695,
 0.9289681369910158,0.31558339347494635,0.903545196187582,0.4047062585590806,
 0.90932823404893,0.22786948180763844,0.5693941732283208,0.8988946933052169,
 0.03160625344856649,0.870997258025315,0.778098972802036,0.4076107843030352,
 0.9972250307051385,0.7900572944164385,0.1686878657155697,0.2044424186425311,
 0.607411120502128,0.45524950252453344,0.2416377315172089,0.8126562179468133,
 0.4457411443335627,0.49264261058292547,0.4732561543156616,
 0.15978273305555324,0.8798209138662522,0.8785987083874301,0.801171670901203,
 0.9642162314310123,0.7219632983144311,0.9606599119332815,0.30658344556908745,
 0.4736819422609915,0.5058692176207427,0.9231075237861044,
 0.028384943947411312,0.7569208817363345,0.03983584143810259,
 0.3985429735058241,0.8284413940076011,0.8858189068400877,0.7922876694442411,
 0.23123404908094325,0.28568833378044656,0.258058878977367,
 0.48814000475980635,0.6517256879739709,0.9845871168887886,0.6863600954798156,
 0.6307953078696426,0.7752751863613825,0.6874341893723125,0.8520212223938716,
 0.07250774688905737,0.2664349753984112,0.2970028595695835,0.5846484532672394,
 0.06154919004219617,0.20579419633693774,0.8281491734075659,
 0.6893557580420527,0.6262935694210472,0.7931750063296191,0.31731995111464373,
 0.2616352857197579,0.18790429752929427,0.8937460664964307,
 0.05055207120550809,0.6726990931068364,0.5508621052493219,
 0.0044312884827206744,0.3721366809902015,0.5210394685447284,
 0.762327461311298,0.45910946167218636,0.39625033588892644,0.2797768963151034,
 0.06710197944423979,0.11215617542898604,0.8252897488529369,0.556611595227915,
 0.07137015846345196,0.741106135299497,0.7922778427977013,0.538840888504122,
 0.26532893253990875,0.5023897385427164,0.06462459622771088,
 0.46904426036728686,0.17310165356631724,0.7691661441000607,
 0.5268627309184195,0.6579416455697353,0.09509734105238998,0.6912389625225618,
 0.516327456692983,0.34945396965980036,0.7709905527710835,0.2449154120821645,
 0.636422207062394,0.11448144249059544,0.5909321873758419,0.08968869943975288,
 0.0315745517351933,0.21589738001624015,0.04500526706073982,
 0.16846939299193064,0.6749862454986376,0.5522661191650404,0.6157790164468993,
 0.8973335126740051,0.6792289766495493,0.6912379205110587,0.27171931122069626,
 0.4206166593652514,0.30581390125036434,0.028521205278854045,
 0.705843146793506,0.6918766131643838,0.3802631323802147,0.9779107232425127,
 0.9608810307796756,0.34504061753830684,0.12205294213810403,
 0.9555581313498818,0.3862150608383086,0.39331370021447787,0.938146657726883,
 0.804346683839538,0.004453024507118708,0.5237442415901912,
 0.22659765626115125,0.7062994922024092,0.8371460525576405,
 0.35229993180890984,0.34296175718774236,0.40042534310487743,
 0.8979495975438828,0.4289775735270196,0.5699989667414066,0.7565506896884754,
 0.06510284111605913,0.9726611533815291,0.42312139962896844,
 0.5109464621022085,0.20461476372414267,0.5240390814693673,0.8900574265897014,
 0.6258832826059627,0.11073983783498997,0.7720263371415319,0.8033203550734559,
 0.6018731761249068,0.8879594872455525,0.9013279556397257,0.3899408768097341,
 0.09801875643378521,0.5040790628016691,0.8002912720567082,0.3915056626493538,
 0.4790443847004028,0.3895972840316222,0.23168079467116975,0.4949110441914315,
 0.12971476282831795,0.03658866146564821,0.5181265023526704,
 0.3485573125169372,0.6080088351747402,0.5762464040709251,0.596742539824206,
 0.2891999117353541,0.1851984050539247,0.2672857439214623,0.22781101117456404,
 0.25028463281674773,0.9168082115229721,0.043528917298579106,
 0.07125772407408615,0.15663827742719805,0.5379951589372263,0.950366367403456,
 0.7738365054832249,0.01979425864739659,0.46435757678671596,
 0.35757658393764147,0.0554786808357155,0.3523922322833699,0.95920368913152,
 0.9409843951026764,0.39925279428363314,0.8482539931162042,0.7770694910834619,
 0.8908480695524545,0.4617168233324228,0.5488283621639536,0.5910360245075295,
 0.149710085717772,0.980022838763906,0.04717932362677357,0.24312448712965806,
 0.3147582410832439,0.43624902039460733,0.8336467783086758,
 0.22707691433694532,0.6001663724003895,0.9228861627295493,
 0.14249299385552772,0.7176520647144019,0.4836205033674894,0.5769871620278684,
 0.6146596004223268,0.5882011398975164,0.6835316145036664,0.624083712639862,
 0.7860826071256096,0.05732080194454969,0.008103907853832193,
 0.07043591718940612,0.25441405879838463,0.36016640301850844,
 0.6470637420972678,0.9238913777826044,0.13025603964661892,0.5992057731364001,
 0.9148352269650831,0.3395670214289601,0.3384039709876062,0.5746495799706358,
 0.8714687972750502,0.8168411496609912,0.8124446623555199,0.309972900800787,
 0.6265481632831024,0.011501109033331014,0.23035382077542144,
 0.7864116274758169,0.9112333920183829,0.10439453781027885,0.9088477123116594,
 0.8091939434191333,0.5432491858582758,0.4673392997537992,0.34395899204936486,
 0.10900593612984943,0.31787433565664047,0.8713518613929528,0.655101893900923,
 0.5865136801544546,0.8278794336280162,0.7389326499977097,0.6739804769274055,
 0.5116232748869229,0.2232362047807328,0.5308437008998796,0.3145167415658102,
 0.4544162293589923,0.21625229074681218,0.9813178115250323,
 0.17328739373220858,0.5484425671284893,0.6206881857616482,0.5362588895610785,
 0.3748896974993412,0.11660248976548082,5.809225306563359e-5,
 0.6367252023602119,0.9557152255336602,0.8877114335530684,0.7033384896702808,
 0.6872226966217727,0.29472746660567994,0.348464393516694,0.9885041564105033,
 0.030921830852035037,0.40569419125999384,0.8238636230907556,
 0.9850098131341453,0.0669374957789668,0.6606223953372204,0.34569008941048773,
 0.06464636652918476,0.1640947450299427,0.9462542445846576,0.3284366787852532,
 0.24884347590269662,0.014614960572371438,0.255870744229339,
 0.2645995914484599,0.8560637443942891,0.6580171862915543,0.9272620035567024,
 0.6703879325542953,0.39652298650799944,0.20281198202914696,
 0.5788888321052136,0.5451122151273389,0.3493361828585809,0.6684694845827361,
 0.7216718224700839,0.5621617777057024,0.3298985073424371,0.12365697246781182,
 0.03102081479843488,0.6510839962913987,0.9176336356466143,
 0.12310390479597555,0.784661332589732,0.146995460650314,0.23146998921072925,
 0.06498427463467182,0.5336253515315533,0.8438378115903972,
 0.013100220944240015,0.6645566438113903,0.816434287085153,0.6496598139688635,
 0.3541150057955239,0.3857968241213132,0.9392894695104027,0.6466120054503612,
 0.1294792792132713,0.9943558962910715,0.6332678741631701,0.5248737239209079,
 0.6762131149476165,0.2901887538588901,0.7485299599020712,0.45374140485128256,
 0.5745496890883308,0.2857478777847249,0.8424471992326983,0.01893978294858767,
 0.17814771748755898,0.9944806409253485,0.25041855546649183,
 0.6187248051972456,0.5495038573469281,0.49890245184369797,0.9551686867966451,
 0.5211196279932496,0.18694083486679947,0.5303010107050788,0.9412428030610668,
 0.8834007891021087,0.06612767257971275,0.43602406147721817,
 0.8330026093781364,0.1475827599318893,0.8641467815654172,0.7466159041382382,
 0.894400340459683,0.6008056428077182,0.14328671714538488,0.9162232063632213,
 0.16529530874729614,0.33928557507399204,0.8068124595891975,
 0.3120008128128716,0.29464894251150464,0.20052364910538523,
 0.6789641904668134,0.30916932333657376,0.0670841226020793,0.9361135568935515,
 0.9090123512271375,0.24053514376331542,0.7674809196197121,0.7746725832315862,
 0.7959708638006581,0.9483720340935468,0.1784847108899884,0.5454944477052472,
 0.7234550959676053,0.2892047214049054,0.949354994962982,0.2086851012916895,
 0.8335698103806413,0.5727426725009498,0.25499066036895024,0.5810486364174281,
 0.09364332152388599,0.6686623715134562,0.032614470156910436,
 0.4421443765608746,0.9625842070225378,0.16138286862539153,
 0.29676682974716373,0.2443628609913171,0.4906919500671445,0.9963096142440069,
 0.8994415690620312,0.44418696954599723,0.33203690508119854,
 0.8099536134179485,0.9271024343695041,0.4029537141921825,0.39023579305558,
 0.6014099568440853,0.007572418701351835,0.19425388675412913,
 0.9578532330254145,0.35616775650429666,0.9295562920507148,0.7257366858484686,
 0.1503863925346196,0.5798881948892376,0.7213784085104281,0.812818029014385,
 0.6959368359311539,0.6331035742145101,0.2839193897142157,0.9224830577831233,
 0.6336594155491492,0.08094577940679148,0.31742383233313354,
 0.7736387904140226,0.39190434808121566,0.07651377449594099,
 0.5360088925719153,0.5921947709347893,0.5018221441955328,0.9987411970554202,
 0.12402249231495077,0.5997203256438399,0.5871647464350271,0.9793768552819345,
 0.6294313714483061,0.20532295779810217,0.9723870651365452,
 0.44427949825644875,0.19956877026257636,0.7906670204589334,
 0.25430065125503964,0.08590950530860397,0.8625485906782293,
 0.6880432199181765,0.642999762029383,0.6910705284641956,0.4397063059196278,
 0.5755245148078836,0.9630610987926072,0.24184841698265735,0.7119210567146117,
 0.04315227143116063,0.8514285552256199,0.3216096249218151,0.3255095982589744,
 0.4859836729746906,0.2748928904379965,0.7646593690152407,0.3809660635071318,
 0.5857875466034671,0.7730847818813033,0.5984571470536021,0.9146331193163174,
 0.9669691152994946,0.7615622678923478,0.5663701875790074,0.2505671387755375,
 0.5696477325553496,0.8526719776268906,0.7278967751998807,0.2751289841108452,
 0.6475976267809402,0.684381027376888,0.83450949043174,0.9544638593956133,
 0.18724839469431165,0.27406930709440847,0.11597725823897642,
 0.012054298826485121,0.10512808686658437,0.8685043881027092,
 0.2790148470801569,0.32343747831879655,0.4150617915820425,0.6984213682078129,
 0.49511331730979347,0.6263665671258487,0.2514978467753507,0.8354527830679316,
 0.6242848385634931,0.09832433492730208,0.33582662922060424,
 0.08328211240178118,0.6412122849710726,0.4062822911991908,0.8810280132123424,
 0.3383600679563028,0.5638841581709104,0.7922810057852537,
 0.010403813336276158,0.44554676772870394,0.29844220446467096,
 0.9719892732010028,0.5851129678071567,0.6535101183748329,0.45594609418144083,
 0.11882801302897184,0.9981732002166608,0.7382563084790074,
 0.14659272908106702,0.2557079263218519,0.02193114512463601,
 0.31214419877723154,0.9587801343730253,0.8588767013728259,0.463404820137445,
 0.782204937982823,0.6128872615082208,0.6074579160846173,0.9893364643993006,
 0.3409283618275346,0.10191117252211068,0.5843208134515216,
 0.18297936089576727,0.2328547139727668,0.18360115089533657,
 0.44334275560872016,0.8562583443204024,0.45635341989833844,
 0.6365959679172726,0.318454558904781,0.6200252286780141,0.6320073590527997,
 0.00214986173444065,0.1588396522007135,0.7675466264846323,
 0.08431412452506393,0.126882719982786,0.41568338527032567,
 0.47662581496761613,0.46738368455713064,0.21562435276514957,
 0.4603963857676331,0.05023040843746018,0.8967702199133943,0.7130197518563647,
 0.7288120623627957,0.9310639786244544,0.2125679596067197,0.7206402403871275,
 0.2397838949058434,0.817125298576399,0.3836135149227756,0.09667693455097448,
 0.39716708182738725,0.23999297246027362,0.3616742314604777,
 0.4096214100610269,0.3255984807695578,0.5347510333182716,0.9655407406829892,
 0.48831433358649934,0.8683476682940404,0.8914633562839441,0.4277898227938328,
 0.9762933593709742,0.3215893833180675,0.3668059433785329,0.49898206894444275,
 0.5094788039376671,0.8022077645652971,0.6978339719303296,0.6695180282052369,
 0.4296443441535016,0.7211294933893478,0.8682044955084463,0.25991883653873027,
 0.0030151717735389383,0.014877492375307222,0.12169984526974009,
 0.3574993188286044,0.9535380552130188,0.25317521557718525,
 0.19123594744390915,0.06643294425355561,0.29831534540368465,
 0.8782623974334562,0.2597538082639498,0.4932514137196239,
 0.041870685586102185,0.5390848409948157,0.9292442063180019,
 0.5000381553022144,0.4057437468490156,0.27220781319877196,0.9019844244571279,
 0.9263691561359289,0.5255593693607074,0.16337411791643808,0.5279033710167927,
 0.38740830515816216,0.35026360018699654,0.09459982794144273,
 0.7743622519673157,0.5034017767771894,0.45998810242598587,0.7076766619228577,
 0.3829099047785425,0.8608932644503063,0.8802395175563056,0.24809913272396722,
 0.27094041785711387,0.4850228619309632,0.759487158586376,0.43491882130654147,
 0.8670899015518526,0.7117186145509301,0.2991238008328062,0.6617672251629085,
 0.006748754266112766,0.7404193150666839,0.830099034185226,
 0.08976788251891155,0.04717117472757304,0.27357521239816307,
 0.3441070057303799,0.7095620965446126,0.9561890090482862,0.11127675553887273,
 0.9952255416638416,0.9330065300286491,0.3226798747953924,0.4902301326349552,
 0.12711609345591401,0.9783714531685438,0.1570901175610543,0.7047807983491805,
 0.4410250300406273,0.03735439383305228,0.19593954663996616,
 0.38063068230286545,0.7737502730077013,0.7099625970559228,0.9339117467317881,
 0.09224349688287492,0.1645645526014764,0.6785511936762714,0.6121629986928002,
 0.607301822697622,0.7701200149366164,0.23739587930876926,0.8102237192244226,
 0.49844859237592387,0.23629832696407427,0.9792199105645953,
 0.10308924833745126,0.5410547700873845,0.432831740358954,0.5499736801198417,
 0.2892055789408946,0.36911812996295656,0.33375983778136564,
 0.7487316558742574,0.7105799581018413,0.8729232522800213,0.65195131830262,
 0.7242533662836261,0.7200809217611498,0.05655575925005207,0.7215357851537921,
 0.6485268388131457,0.771150795662018,0.9228798551403807,0.12697852503615215,
 0.10011052771708961,0.12211822799125671,0.37379905076696474,
 0.599434718598983,0.5032474819640158,0.4796725793832224,0.3060505341279435,
 0.7929607942029833,0.7108079601350824,0.8883784840333814,0.9343632957287757,
 0.19853453018213119,0.6603449548067872,0.26625414180631846,
 0.32451176524471725,0.49855627892024745,0.08021364659933783,
 0.5138796434604753,0.9809193439590856,0.5648541339414832,0.08031857793286179,
 0.9355217000674373,0.17205695478846994,0.9667351112544509,
 0.049291459677819116,2.4799293241506426e-5,0.845314286020101,
 0.24398334312008407,0.04285353793722424,0.12313988445269763,0.90912802442053,
 0.8624911932233627,0.937773798750736,0.20480975478752228,0.3678071660966318,
 0.062427601959390344,0.027820645208220807,0.5303968310606894,
 0.9533366353398026,0.30461989821716307,0.19465798531570855,
 0.15357433792289188,0.19568005183187975,0.16055954791670013,
 0.9144391293285471,0.8297760238420439,0.8307642124965158,0.8111696099691842,
 0.6019052322209199,0.4824948577416023,0.9277799868001233,0.09781378302902177,
 0.7988167165572605,0.9033055405411372,0.8948914651274745,0.8049146331854915,
 0.4859586179249341,0.685195479422367,0.6193880997793534,0.6231980345712149,
 0.8960938125956023,0.37800185679993425,0.17914651256515546,
 0.4785273328825109,0.6964648240774732,0.4172815503325773,0.23788567485527234,
 0.2465851141186155,0.7571832703877155,0.4787988575919053,0.8883616769681615,
 0.9309361382061605,0.42831808703047414,0.6672859392762777,0.6244325885143945,
 0.2854976402541719,0.5994307201100969,0.8636457043766851,0.4389774834354927,
 0.9067661278135999,0.18023642783469906,0.5979698510129197,0.1961565131008316,
 0.378739236430917,0.04729840340210245,0.9907624811208147,0.12032785190431516,
 0.36426620589776504,0.06951815024171937,0.22087115404401125,
 0.1067593593771684,0.6764935952080174,0.06432160188576685,0.8961361306934161,
 0.5973521418647775,0.5123205497756783,0.1828061712480794,0.14344146990634066,
 0.40066926684203197,0.3241743337768124,0.8796023018950075,0.2416910152003915,
 0.5063470279538997,0.994692853913691,0.5247704214626381,0.24977271443236015,
 0.8537164175201312,0.04696608142853376,0.756596677731887,0.6773786597790034,
 0.3086826627216942,0.8205592133552831,0.4220971254902516,0.07028742476693894,
 0.770577932615839,0.3074254511453758,0.472608948284046,0.0796270126390094,
 0.8058150440794658,0.9529959177966922,0.03490227200157647,
 0.34728594442141036,0.024661408377939953,0.3349075573131325,
 0.7773526982986365,0.28946456730675263,0.030968823343078156,
 0.10412241704523528,0.18883188030033593,0.31696560055993295,
 0.5400576294263729,0.9800451776524959,0.6029334676512069,0.034943877256291,
 0.7027332495851757,0.9322664077980395,0.5714094142984908,0.01669562691768045,
 0.6294926806089062,0.8791243725266724,0.468661331811128,0.4310829179907759,
 0.13862698644378058,0.8148664993853587,0.043021679764737586,
 0.1423281945651479,0.04024236746694243,0.5941555978391255,
 0.20745894662180864,0.21983860539761857,0.9637966495844725,
 0.8955714272227691,0.11356284144757045,0.9899881305031715,0.8050942472569096,
 0.5345543751629214,0.19806600880378777,0.3255978779092633,
 0.39981961625189666,0.0651952269826569,0.13916085935816414,
 0.5286475886000026,0.8203212824290598,0.27224957030358987,
 0.17357849752023768,0.6589650549315714,0.6813816960459445,
 0.35222961004806097,0.6564282815591523,0.4282494098233649,0.499457294927093,
 0.8082440926555062,0.6667791827337585,0.635915093048029,0.5979364190335182,
 0.17748036725518557,0.9520504114553044,0.035605556839733654,
 0.4812506963051427,0.38544845430321584,0.2963137273477827,0.6390525570981072,
 0.501826447133648,0.8240463318657067,0.7859773684016718,0.6682674523046122,
 0.8411477007818476,0.9281696052867217,0.3181718512124163,0.8782100111753762,
 0.43469394660574034,0.686659269884589,0.6301060326523596,0.9902000169911926,
 0.13247106473845427,0.9187322502042697,0.443338703576567,0.09580622491039115,
 0.40380133578186195,0.24345109493761175,0.5027224276278133,
 0.41759019802166364,0.9796826144278942,0.006327424611850652,
 0.9005080693446003,0.7075797815083344,0.41644628522753224,
 0.024575165764481377,0.973365532844886,0.9884817375945848,0.5815511183775325,
 0.8312305439687434,0.6526396825519474,0.995462471036212,0.5176496470655602,
 0.6044894942029684,0.4837996425926717,0.7169594663414373,0.3590372108039407,
 0.07974921858277828,0.056628769348097574,0.542198043523617,
 0.7858348867106791,0.1326908926633823,0.4229668342865348,0.09541981674151234,
 0.8999681230878327,0.4405229509920805,0.600105360365751,0.5234705200473804,
 0.5314813640117005,0.7997885828585332,0.2739129979253394,0.6201763769379725,
 0.3966883081775634,0.08020643947099537,0.9441252225886612,0.6377968401058198,
 0.39736070306289173,0.3621167368498377,0.885005334381331,0.17695585233354938,
 0.4302216623546895,0.6522037783898738,0.1761903578705004,0.32700727326120005,
 0.8430481108020628,0.48349857286868325,0.6587181016952789,
 0.38725476514452084,0.11436698911502108,0.9144217842011103,
 0.2017481938479868,0.6187711527711705,0.2993776967261942,0.9407090394435709,
 0.46785294761129903,0.6721380062185919,0.6672132000598494,0.6832204415235277,
 0.5460681365268581,0.44819462583109315,0.24091125789213763,
 0.14236010810194688,0.7644883645503073,0.45552631066791704,
 0.19669566563498075,0.7064039544888365,0.7390404440340952,0.8421887994577151,
 0.4881867545391131,0.9102304745457107,0.8367697773663194,0.5160358693602654,
 0.4712843367616517,0.36922216614002945,0.8540627402546894,0.5055380135720356,
 0.9162204698918845,0.7142293656097136,0.8090185218677055,0.86551307337518,
 0.6226775507439766,0.11677710670095232,0.3033662800619831,
 0.13020339524417923,0.3132653026172848,0.6651040136053531,0.5529261545429552,
 0.5354204147108796,0.931223512963907,0.6445272772149246,0.47297447160260997,
 0.3202526643257426,0.04166163215558727,0.36373780664496125,
 0.5057876065910503,0.5319843811170104,0.2779037520776083,0.592249986608847,
 0.8360927329993495,0.23013881876055897,0.2636817135842481,0.1709470030111151,
 0.9557904768260329,0.4831667062630699,0.1419259987247744,0.11262032121396148,
 0.4960631147357156,0.4814231510845772,0.8412003628582615,0.4605318159872088,
 0.059952718407581984,0.8124494795207805,0.29873418378660566,
 0.776364635356335,0.41735304859459,0.43196965553992095,0.5988790173060479,
 0.3423930191314971,0.6416377026743916,0.8948181853777003,0.21831749732301942,
 0.5200974985243686,0.04881669655309895,0.9235311575652814,0.8167411278558109,
 0.0065604682090496524,0.25357736436975453,0.13104121612322706,
 0.012710141984932144,0.8603613932787935,0.9392443481829937,
 0.48394081894754315,0.017638030424414364,0.3359464070817826,
 0.9289583669841517,0.13028330230882634,0.2804534263724795,
 0.12187914581581993,0.1754958033794456,0.807933294820925,0.7885224912669891,
 0.833514155462687,0.2798338264691651,0.7394810795880896,0.3702177738155461,
 0.2893447540825764,0.6002032678271357,0.2036263824369191,0.06117195368544137,
 0.36388381651044677,0.039442613962202056,0.635242844882926,0.776767567648636,
 0.2262692161877703,0.6632360513245058,0.537401414612636,0.22212313933132355,
 0.8240014378018121,0.5702695519191574,0.6104563018283041,0.9759286875023079,
 0.35440847712967805,0.06705199944132467,0.5835654361182148,
 0.9782029838775349,0.6344829501833857,0.1746915859516992,0.6209929941787797,
 0.7926440412295548,0.5330666120742379,0.2368305380897049,0.6416020428165485,
 0.43520086084718657,0.02294810850453244,0.21656680504687797,
 0.6496265397149127,0.9255433730775882,0.8992260575632168,0.7091944255712822,
 0.8086477482433828,0.30960741088490096,0.6286195603695668,
 0.021974796305941013,0.020334758727319602,0.30962452456349066,
 0.9235886753910141,0.9039523067604034,0.27616395935270976,
 0.32729420880217774,0.4564569741997282,0.295753775740747,0.4445798029747191,
 0.20366177705782007,0.21187902415523685,0.5896208509833357,
 0.8170519449385378,0.6053569618236765,0.458842199078907,0.4992840079943981,
 0.7360946610594,0.14135818095884245,0.4212095241415448,0.16599999999007709,
 0.9184128424993395,0.5286454183784494,0.7185650976586829,0.5990038521828767,
 0.6771626565546527,0.2946984399502811,0.8439849363282758,0.1576103026302066,
 0.6593978596550092,0.17739474240053732,0.3204022400671036,
 0.03173476746532333,0.7381483034515767,0.3506623059532773,0.3130299072682169,
 0.6589026999134643,0.8364935996382511,0.42999464405573806,0.9129587745891339,
 0.13433888473099231,0.1277681904751009,0.3115432029469428,0.4743358881482044,
 0.29961661565849007,0.44006010388045036,0.06512684532867552,
 0.17889860209757535,0.9349227000176408,0.7404705511748404,0.9555920028456,
 0.28306477545104325,0.7106064900028171,0.9551315365223854,
 0.34364445087525286,0.16573343889056324,0.5835814543413375,
 0.34814890211663574,0.7101067470057767,0.22401074317209202,0.811923308018878,
 0.0487357888483424,0.9457103618325214,0.7438662462427258,0.5684282742298183,
 0.9016870169671731,0.8010808918818829,0.7438138678804043,0.2823954858215937,
 0.6176738107892307,0.16083532116346033,0.07715914888015607,
 0.31661648072530346,0.5582601854793191,0.4256321014655621,0.137551906280177,
 0.10374284456900773,0.9472269482505332,0.7247174380292557,0.2695338327666822,
 0.3550787720533424,0.7553362342427592,0.3949209313090689,0.6327500912393896,
 0.591333241896196,0.57644004013848,0.3414465621318311,0.2970209861689467,
 0.3915435450861282,0.5861047481890138,0.416674211244364,0.8331652577521675,
 0.5092954093434219,0.8340247293490252,0.9605881787879298,0.5261708676017915,
 0.7236770286838169,0.7688730129225313,0.06307443673439184,0.6872569456480719,
 0.4604610134307352,0.3524150373567815,0.5786249211454102,0.2705610057348874,
 0.06909713768291381,0.9336989927499137,0.5140410116721097,0.9125588804860014,
 0.7904177137993514,0.9514578442212575,0.015389838280305379,
 0.7748778476829171,0.9722653014144165,0.2142583582916605,0.40220995552749794,
 0.09307777525976502,0.008586271449485872,0.1643835801407345,
 0.8005338043043956,0.34258176191171774,0.6522533372416078,0.7443986912125287,
 0.1182353427124796,0.04506248277812919,0.208207633244474,0.10607971953714324,
 0.3920587631512348,0.5656510243234074,0.26307430920601527,
 0.45898096855810333,0.6442778632085967,0.007635653111418561,
 0.021290017298172592,0.9536436163096098,0.7345005149936864,0.672209969534503,
 0.6333006769716977,0.864976280471422,0.240349765131308,0.1504362960940938,
 0.9536436909500144,0.09711322846111092,0.9995467442571532,0.7237383545429295,
 0.1340335649712422,0.3581349490965302,0.7568720282485826,
 0.014861079931834498,0.39137878432594486,0.1397271937460327,
 0.20915966857318402,0.4689416838142556,0.6313371407057259,0.8493435384574304,
 0.9682945696774679,0.3645856207225978,0.35529564117771406,0.540631632097348,
 0.7930898435894602,0.16472062801369436,0.5497160885434328,0.5502207233214099,
 0.15078305713900708,0.6734714691364411,0.09111485564652666,0.619401242005204,
 0.1293406322977031,0.9339945652825137,0.6397782519464239,0.8122037982309463,
 0.364503119007046,0.179848762208199,0.9137480594414653,0.4811145668243164,
 0.6303782948432993,0.07224647300414355,0.5790764595612669,
 0.03269015395600583,0.23113747864203826,0.13721389574143905,
 0.6289593959889886,0.8979639595241358,0.20086671443652881,0.6681239889997493,
 0.017006814778718904,0.19885854913070722,0.20390230546602106,
 0.35094647997821005,0.8474847671491564,0.6526674083786995,0.8017725558125542,
 0.9635218867192266,0.12785129881203494,0.3226837546163178,0.9567325563546545,
 0.17585205694449646,0.5437289442204785,0.0627323858712554,
 0.30824088038500463,0.8932937830524402,0.7272856865404199,0.8774382550680868,
 0.1521438770784449,0.4473270321633046,0.33245625470347795,0.6996610229638768,
 0.21776996902621837,0.7311036301469424,0.9800721092007494,
 0.09861684786536662,0.5196447546443573,0.21206528241035472,
 0.8695933574255145,0.7582450145642722,0.7837393804526558,0.22442945075977394,
 0.4725319389548449,0.43602364356164836,0.15244557231906375,
 0.7365914397378075,0.896102177016715,0.6509672745845336,0.289466583379958,
 0.854089053265351,0.6222540427067931,0.5233885445885799,0.3380169573597631,
 0.22950015823650938,0.41450879826751996,0.030191489036261454,
 0.5915411262585456,0.3306798795231414,0.7001309677038388,0.6682456507852826,
 0.7280776637838814,0.09026404892029194,0.433785150114679,0.3692574313148008,
 0.6343194712869263,0.6126286071170346,0.9704474783065214,0.6415886459763585,
 0.6332869633535357,0.4558760115740006,0.6374121695773858,0.6673048205295035,
 0.09537729537897477,0.4925446612196267,0.654351231639993,0.34864006509150647,
 0.7755469345221719,0.3401954189447378,0.03936068602881437,0.6505549214186174,
 0.17237525650774166,0.14642437284348686,0.015964609926861772,
 0.7238644092985071,0.45177157286627523,0.7085747076731781,0.8775893611302954,
 0.22144977532732601,0.5741891446184032,0.8908897280036996,0.7586277801185661,
 0.4473982550863542,0.5763423424343259,0.8477555246411415,0.9681189045705085,
 0.5118341370801903,0.37139948799855793,0.9653142962910194,0.9779447924102879,
 0.4127161453128301,0.3735002542222099,0.4634988775250588,0.8756393125110946,
 0.6642598181726521,0.7494862622958188,0.7755875255155874,0.3922598218376928,
 0.8320935017875897,0.9662635678450652,0.032946179620764404,
 0.03659033847913962,0.007955674550036873,0.00829377209744464,
 0.06158010314639839,0.5803255922910546,0.108957612651152,0.04311483578896247,
 0.2237803021593443,0.7774117101767266,0.6472323595039742,0.3920639351443207,
 0.503898713935933,0.05629969393262613,0.6675346546342957,0.5122883110815724,
 0.567108339127132,0.5896560379380185,0.4012211315136125,0.31373122788663865,
 0.6376582855784357,0.16454592280497948,0.6433984438300372,0.6910051844010378,
 0.3326363375265794,0.8014315643812403,0.5068753330236714,0.7902347490886678,
 0.4778050723302184,0.44133522113243095,0.30680696970992327,
 0.4898047325609115,0.6025948483147899,0.604483953844268,0.8724587669034884,
 0.6032052487431476,0.27603652103317877,0.1572519034703148,0.6048101851055252,
 0.939622415909626,0.026946040152204698,0.715176655217593,0.9541058507680167,
 0.9199636269139547,0.995296309805686,0.7504359623050099,0.8727784268996175,
 0.47726156704192857,0.31847039776707176,0.023359558063960728,
 0.5485384507984125,0.5402006776997496,0.25050636378406166,
 0.41879356187814054,0.6230112425188528,0.058894458518206516,
 0.09250339167913757,0.2854665017906767,0.17566035432572835,
 0.07800696767709947,0.760486469227163,0.40794525911470847,
 0.15525141404067677,0.9959065684288702,0.010760460688483353,
 0.19266592621572098,0.2287756966817216,0.26160683678225205,
 0.5408281793715446,0.6375265380609318,0.7895256189515175,0.9978411150304366,
 0.04017959331336951,0.18628069159329064,0.2820634292555343,
 0.8086604872676032,0.7907054708445426,0.9995151061445899,0.585957303561892,
 0.15850806790125804,0.11202962177959275,0.6291363562338588,
 0.011016220288680989,0.287156870151703,0.9607121895347308,
 0.12847116907349776,0.39233281686464017,0.7713667831948336,
 0.4762457052914618,0.5189650267954833,0.11756244636376667,
 0.016284108874902503,0.8882800666887299,0.9511550211678518,
 0.22475665798171496,0.1106078818980254,0.1084185902178092,0.8982064028120941,
 0.7836632403890479,0.22318329718761143,0.9706561410920514,0.6686078938493865,
 0.1424485077135914,0.48409248457654064,0.6816773188714345,0.8009640292675533,
 0.5845603553961539,0.7776520707098163,0.46604529136057404,
 0.03248678904105329,0.6453433455711628,0.8528735736313066,0.9015803766463082,
 0.8986971793849462,0.8094217991797521,0.5258586766648958,0.9893840388478377,
 0.051999776544485954,0.28816270212965667,0.23448431781425383,
 0.8804328194262625,0.5419900870592065,0.2646198051547125,0.6049878036739191,
 0.27657589560687723,0.8888139812706861,0.8438973425350196,
 0.46111245233984866,0.7814757095396894,0.06404041595573173,0.570908168930599,
 0.17131032176919336,0.5459293962432921,0.6591751031015589,0.4355861809350897,
 0.9414927007970232,0.42087910930563066,0.073734331098812,0.49723375257432223,
 0.8673505714750677,0.28976586389241876,0.9274577065096028,0.8776611376734075,
 0.8024825564666602,0.8612637930786988,0.4314191537387667,0.5275868781816082,
 0.5860547360792517,0.969526079536038,0.8121301105483435,0.21513038540161922,
 0.8812198903107568,0.2946282488788241,0.8433296363521776,0.6423820079598124,
 0.9449613235957454,0.2870286703195658,0.5961322224387559,0.5138450547605984,
 0.931601774419937,0.16720232350759057,0.2408082722663526,0.551283462124604,
 0.1175288943584033,0.19707614507284044,0.532387007894096,0.62620834819382,
 0.5978483746056049,0.9255929151400387,0.6697044642108008,0.5488766331640106,
 0.16520216871422833,0.04291169050357102,0.3936415985417345,
 0.07205203430557194,0.687677198787035,0.6728063353576693,0.9105343579165354,
 0.9206281947909793,0.36710375468914413,0.17610352680670793,
 0.8966254586917104,0.19037216249847444,0.49244598175265264,
 0.07336168306544966,0.16507392844294455,0.34288446948606055,
 0.04542452718534051,0.22161432543180704,0.3668877272538873,
 0.3439786870652699,0.4425112408976374,0.6235389537871137,0.1289236090005209,
 0.6847089502705053,0.530298545265225,0.01507829374172287,0.6246541994133622,
 0.6665991163229176,0.37458891622973,0.5870316171597355,0.7522526257710795,
 0.14324492883084095,0.42590855093784774,0.784842634074874,
 0.007484767803884296,0.11874914736708075,0.8089116271495332,
 0.5213913105064923,0.36314874672068254,0.515412606690569,0.6092512728005659,
 0.9055829110940643,0.433506613882119,0.59919409492808,0.3776953657687864,
 0.28605179820947446,0.08290869985763505,0.5315281313635726,
 0.6249626991907311,0.5107638979358423,0.7328393951344678,0.1843452462064793,
 0.11778443816089945,0.26274945478333955,0.166200598851171,0.4398847540748011,
 0.015752278571532607,0.4012105199535062,0.47794295289823385,
 0.2453435767982416,0.4799165171066765,0.005321264559813965,
 0.3939851353469851,0.9517565090192962,0.5084129132093663,0.7808625113113274,
 0.14340932471057544,0.2105628223216368,0.7577912713171644,0.6627229286938711,
 0.4294082084816778,0.6261285965842922,0.36102941603966576,0.8392337291450028,
 0.9694285574264414,0.8065520930564416,0.004101061718398524,
 0.9617074768546038,0.34197306656644266,0.032438028114591844,
 0.008984934604450379,0.7842349086286772,0.28982978829224626,
 0.5167534566319508,0.5650481279309257,0.818311645963248,0.5994579440670696,
 0.6359750435786775,0.12746887509903826,0.2987369873752518,0.6969811573628456,
 0.11529960423130214,0.41091992093809204,0.9575943529635159,
 0.30240190571235837,0.6809720985875496,0.7043087492765948,0.5606548712442042,
 0.04066457993729056,0.6017203014057078,0.04316801910536616,
 0.05150700647245621,0.5668244975166306,0.8662013691710875,
 0.47758295314044463,0.9894719432259467,0.5906200807025153,
 0.053347860219538745,0.3222087378655173,0.8435276717577128,
 0.1798590499505982,0.9623378352415072,0.41560629162057605,0.4098737244849173,
 0.259512341186945,0.08724171262900238,0.17720589727003622,
 0.08501318003921787,0.4132057110219093,0.7710612988670466,0.1848853115311716,
 0.1504958735891564,0.07460691992458934,0.2347458131199125,
 0.011294748203419691,0.7431632075724941,0.7869645818367108,
 0.5181216646352255,0.1468206133030372,0.03731045811315714,0.3609696886025662,
 0.6267159461101113,0.7874445607831468,0.2429374927587049,0.5680572096308498,
 0.8314038266213365,0.4327782190027444,0.728756265198827,0.044976427642077654,
 0.6270229826628291,0.4941239630387473,0.15377143381647806,0.6458707625499855,
 0.314994055613678,0.8706879966136447,0.1522434532970186,0.7096499615842059,
 0.8754697587294412,0.09785282366874604,0.18567827481469334,
 0.8381108899070885,0.6600668419865778,0.7614770309122808,0.6407307380675313,
 0.3565326075679658,0.927307286413801,0.9906020539828054,0.3382696845236899,
 0.6855289224065224,0.6661082969051311,0.6948247029397517,0.9360169140858668,
 0.7866815087195553,0.97914384975466,0.7681256648536244,0.7443316206531778,
 0.7411163138251763,0.44904851944819413,0.45919530524292973,
 0.9613363192606943,0.7647402077325063,0.4849077525280896,0.8362815405753761,
 0.9939157408045945,0.6045448300613946,0.7429120209813452,0.36616548261159565,
 0.15694588342263316,0.46671733671184434,0.09653004748324656,
 0.1787284162153875,0.6741765354331076,0.9993973930096482,0.3247320889008575,
 0.6483305583337394,0.9389649464835736,0.5200201911672487,0.9360081982427989,
 0.2950456153820098,0.992536291394288,0.756154815594604,0.7574699329905961,
 0.20792626706623696,0.5563949535241624,0.6067206486700198,0.9637119836637561,
 0.27000366270754894,0.07195125997988294,0.26318277953389246,
 0.9041885987966825,0.27764182453194164,0.39020619632543485,
 0.10793125608324794,0.8110716642132159,0.27505501230009444,
 0.31178680541521364,0.956509167773917,0.9961080229243343,0.4741500131092271,
 0.3854633840368795,0.27467544075276595,0.31576147941082156,
 0.6135922511879954,0.7799235489638224,0.8226211102810516,0.770584856335035,
 0.26349263466280126,0.46245083310817403,0.7056399712392484,
 0.25263701511311165,0.9202532727399316,0.9644093368348313,0.566020081136039,
 0.13196475203986524,0.8179912516284439,0.23378720921016594,
 0.34881780377828253,0.8535101946855561,0.637570548933665,0.29825604198991085,
 0.33997747101338904,0.9335934864815546,0.6528466229890822,0.1386824482908866,
 0.8179421284204444,0.7159193799758338,0.30015260618020956,0.691473596282186,
 0.1691203304304877,0.7687055360341114,0.709469238174362,0.4079295117007793,
 0.37993106711182545,0.15004032721431804,0.9117275825178256,
 0.6827232879504668,0.378252741265519,0.3475095689535862,0.004706560310897223,
 0.9379236751506447,0.23534660565447652,0.20632415702427453,
 0.5268807291892362,0.3429012657396153,0.5796733213384688,0.9818553399869878,
 0.6425145958659663,0.6006467026425347,0.8505008518279965,0.7533752419711519,
 0.6734199825328807,0.6823973625034018,0.5636337395297831,0.7748880402184325,
 0.5368820044502074,0.12273518900761282,0.5686356236331598,0.5724674868551859,
 0.43793519966422,0.17758728660521483,0.07141269898084746,0.04639457973706895,
 0.20176128449952901,0.29152102649094536,9.121895480468378e-4,
 0.8305475408043297,0.8591538174608562,0.2958667389285906,0.06630626838948961,
 0.1153663701617692,0.940933487226078,0.4201428624394715,0.6416114185366529,
 0.09329199503742014,0.6053955607234366,0.22362890970656768,
 0.4354843449574197,0.10482883621645966,0.17198456565861042,
 0.9446966851087845,0.6169786304991369,0.7433276793185484,0.4890466379964029,
 0.25121230870047484,0.26843854241436904,0.9103992715762027,
 0.04452294765194276,0.13663661852942532,0.4083591555346998,
 0.1338430208749728,0.6925013118438891,0.1947239503904924,0.31321518755934175,
 0.5303493934789781,0.10657251785901423,0.03637984221308466,
 0.1005286412483275,0.42012588589553657,0.06299512552898634,
 0.5621224546751672,0.8344336110784925,0.9490650067628646,0.40467406496524216,
 0.7508031133928772,0.16244734571099262,0.91925303321284,0.318140486881792,
 0.08355739842326526,0.5426572156299384,0.9913270478069479,0.8898921625657619,
 0.3064189761543763,0.9220551462956275,0.6165471041803054,0.3237926236232319,
 0.7657208967281193,0.029207050552647607,0.8287459973214136,
 0.5523103785099481,0.8376676407813415,0.2425646302961082,0.7718602817888552,
 0.8107278404966163,0.48851477841982016,0.9416210272495388,0.5005652114674264,
 0.2436825678146467,0.941893818364227,0.6492912763871567,0.8557467426298022,
 0.7112742758275797,0.7734051461988888,0.46514677262990467,
 0.06004601653739161,0.8489867149144543,0.8149641160859178,0.6711195745382857,
 0.5849367003657425,0.2464984595476959,0.6061489800929675,0.8703832944801718,
 0.3911009376260035,0.424336794062944,0.7748780385764769,0.9908396982274659,
 0.27380939443641394,0.5066533558746905,0.9161513860852679,0.2595434210213867,
 0.47776222055572903,0.7734365892714615,0.6544788872816301,0.533091221174851,
 0.9064310682791663,0.35483523380343196,0.0591541248310975,0.5657182510041764,
 0.9849503129640401,0.778196411496477,0.657985014691504,0.7770213980701957,
 0.0821773549041237,0.8602790607947084,0.4219513105239213,0.3609304591908278,
 0.19589554715009172,0.24827619281137547,0.475179615282477,
 0.37810888332440806,0.5464595765560268,0.9221938469201483,
 0.09603169508520393,0.10158552497741291,0.4850974927973998,
 0.7055568952317018,0.21981084782282578,0.8758878612649464,0.6519355461249385,
 0.27248630365309523,0.8666637968392095,0.6593211038197755,0.5965026139967237,
 0.10728075392053271,0.24403132944439276,0.3255617774003947,
 0.020185776476914663,0.791112129658604,0.71631007268105,0.6739958114259404,
 0.2959447476586823,0.7030597696035888,0.4814434292429506,0.5138224210970409,
 0.19367785704927165,0.29517060510028603,0.9145347699556327,
 0.2514350439540487,0.13658585329226802,0.6438978811289897,0.3098400633386996,
 0.741358153062686,0.8340764398996253,0.8813649549476712,0.4620465240266811,
 0.21402327024674817,0.43042612981376116,0.6848780547565926,
 0.5171446025598149,0.17095959417209883,0.03503067713873109,
 0.3144353147549863,0.01172587670410652,0.714405485149852,0.24828850092792698,
 0.5845972893575484,0.61538505840123,0.04406156729888333,0.20448457168940315,
 0.4252619503500198,0.49102176384361285,0.5529639610158384,
 0.06924236149217744,0.021868803694992,0.44296500809706363,
 0.10868894061942991,0.7845145456600113,0.48086255988141824,
 0.013784046621322488,0.559882877406013,0.045644487379955634,
 0.745519599245574,0.6920991951741908,0.5924463901349932,0.33305233012069757,
 0.49128900822537713,0.7479911777012019,0.49345066257193493,
 0.5658764710087095,0.8676156614590935,0.2218476310745242,0.5754095551068877,
 0.5375630235797973,0.10671361345511066,0.7157652308540404,0.8744807489286877,
 0.23012642702302877,0.9765507854641379,0.02555930441904819,
 0.6460026120261992,0.8553946776793909,0.8723701143446848,0.9092902472808858,
 0.0520781963417819,0.5938032058179569,0.6261582716245033,0.43428025306581375,
 0.14883561862615768,0.7948544212728168,0.19540910584878546,0.752505911746787,
 0.4209580465693352,0.26946239418415474,0.561719394209493,0.8213842009589147,
 0.162313659625002,0.8585892977207654,0.5333047906127047,0.938450455257911,
 0.10130038279906728,0.5168912357700814,0.26983168835549476,
 0.8205273806205914,0.47700477290154475,0.8898476104508881,0.3561611665069535,
 0.7361373073855367,0.10891334749244064,0.6401054606575393,0.5620569340348306,
 0.7085665803611062,0.9912401431178158,0.44243386905492266,0.6971378459192045,
 0.08216383922208281,0.5042248162406042,0.06426749758955519,
 0.8394145955508873,0.19760529966783935,0.29562087365497514,
 0.4376404012435011,0.8566818647247683,0.5727830282586273,0.6377967949877855,
 0.44714528979423873,0.22354498242948684,0.125303128618108,0.3112368416855882,
 0.2765288670415563,0.5285946908912313,0.19843632480089635,0.591600534879643,
 0.6668799986400932,0.7216056491007851,0.491479508509729,0.48438398355811274,
 0.5696010517053497,0.31821130362379535,0.9844618481050854,0.7015007035520048,
 0.30456989758439457,0.8053157852883256,0.9612003741478827,
 0.42670640648451585,0.0598807404313183,0.5964883034483949,
 0.39941628232294635,0.7171322819072014,0.41174938864618227,
 0.036231000272470615,0.9244939906992587,0.7867216956582181,0.921898742700794,
 0.32853416865459484,0.8596820035166787,0.8201183748629408,0.5714051869628682,
 0.11426264280620713,0.8542083130837048,0.7865303902021846,
 0.07179917057873941,0.6640116237309375,0.7601884450727736,0.328716763496285,
 0.4387318971894102,0.8294938713185979,0.8248742405787456,0.7461313778637713,
 0.35924936261001217,0.39506662978949736,0.13419488767682353,
 0.12038057472508068,0.8823164203655974,0.3969340096689681,0.846851903906299,
 0.45839714005341703,0.9390699491224758,0.5772907140328999,
 0.17915028438197933,0.5858075755808377,0.8341733936319596,0.778525049715061,
 0.32067328288365105,0.7037528165497379,0.955983674149435,0.0835212539674004,
 0.751380258135349,0.27223248394498634,0.4591283464729462,0.28835082427819303,
 0.6730588793919878,0.12609557217531553,0.7966226351884356,0.4464562447003998,
 0.7496752807507423,0.8912052614664971,0.09329758545865163,0.9076966680606018,
 0.12437757367018443,0.12318776909158957,0.6535076070923304,
 0.5431125090311048,0.19703975774527438,0.37354078327780316,
 0.26129126559321425,0.9542472073720895,0.45516320998259596,
 0.7736521461473495,0.9526037908704961,0.6196606273372846,0.8236103745559213,
 0.8982849414723065,0.7439778692648874,0.7646688092275251,0.49841904617803834,
 0.6841321163763537,0.26060676964290097,0.8762631062845921,0.819294113966126,
 0.7670637381531461,0.9826892658755083,0.8840522866194723,0.14855107537819734,
 0.10397910107002861,0.6496745689454265,0.15776219253878632,
 0.9706297682940299,0.8937261015319903,0.28471642006791487,
 0.10070604059643762,0.6773699598797087,0.5916411226144911,0.3041747899374623,
 0.48773458259127456,0.17185376299535637,0.7913866732567516,
 0.7537732319735863,0.521245742107635,0.5618451225019496,0.3097340683581,
 0.9868695098468211,0.4348984096170765,0.07023113176368467,0.7755844335563675,
 0.7137856469091891,0.16654477981244964,0.19685227920814152,
 0.2503089141703647,0.7491148157685775,0.5681361724255335,0.5168940319752837,
 0.6016128144441497,0.40950584294439396,0.9516409204352143,0.4548794193896597,
 0.8423995019582263,0.8077025829558278,0.7744841688517965,0.734263960233617,
 0.32076339183476743,0.857610642537381,0.1254592899397935,0.7056976451471677,
 0.41884823891274836,0.4992815871472731,0.024693202965396566,
 0.7624880793792131,0.3183617872136968,0.6137087541482364,0.5113999087278261,
 0.19866254028826058,0.6299538222198142,0.6204475859804637,0.1309171423088612,
 0.5573411939013765,0.12532535692457059,0.3055647412037702,0.8608645549490999,
 0.768097472093759,0.9389854487725402,0.8160053939647551,0.5688769767496857,
 0.240339164682299,0.24251504302685378,0.6298743930619837,0.810803732649247,
 0.318959536257594,0.51966934946033,0.37405734272341307,0.3163721625694932,
 0.547002671078222,0.2676068867537231,0.7075020619352144,0.08689621022239447,
 0.49534866071348227,0.07450928317914718,0.7324930528314302,
 0.11050769390841064,0.49232266111608936,0.9304939491911859,
 0.4278436794298297,0.16604061911029833,0.2585407112143775,0.2299116319500888,
 0.18120338095537558,0.9099196738467601,0.9682297411789123,0.9437425012950565,
 0.7501509947672168,0.23514518460224798,0.3228132495903817,0.8435347787038445,
 0.6216831754048167,0.8867290839824388,0.5323167495349026,0.6739778327736892,
 0.9055376924596639,0.5048482395853625,0.8938682024718343,0.5672098216150953,
 0.6694871177895994,0.16496413938383148,0.36745925920917943,
 0.5492267664925098,0.16856721940627772,0.9120319349051073,0.9396730162144168,
 0.647859373607692,0.7403622309727745,0.0902839898761092,0.7551151042805145,
 0.5764661976669756,0.7859209214495965,0.025807043904216842,
 0.03884179799437726,0.5245224488161282,0.010885307784401888,
 0.46368319158276106,0.09899974438132453,0.6925161542646849,
 0.014582828427332384,0.12653655448695944,0.3013642331163331,
 0.08890941962380972,0.6209083762380238,0.5890327823004371,0.2404984423990707,
 0.743294413878892,0.4286711072632696,0.6166212132430497,0.6529589983744931,
 0.7160202461455104,0.8394943793674593,0.8159599521757546,0.36232668346416663,
 0.6590694348890568,0.10666365719081627,0.047472476536367436,
 0.4802560282881333,0.8928846837601999,0.10593250134385579,0.7811124697642449,
 0.6986596095659059,0.6221422835000845,0.8589433660553528,0.6293638546821685,
 0.6138595809085392,0.14746678399365598,0.2291237802114786,0.4708480627674765,
 0.6362710954012549,0.25828464825203923,0.5347260384827162,0.8961808627254988,
 0.6723382607339261,0.6785435873915446,0.4533171233060702,0.6344522195665399,
 0.8389521386042805,0.6056991919543936,0.09158046018988775,0.9177263629725008,
 0.6790163001299679,0.5131669062001959,0.6327041351481735,0.3505723006842756,
 0.8049399702594726,0.7797110937019868,0.9902829913728126,0.8717575250893512,
 0.4320882395405825,0.5086625353239296,0.9271603594376724,0.15007083900972873,
 0.6909957037059418,0.09652277101176376,0.7055817903088761,0.083997111287885,
 0.25342395702938747,0.4579098808466213,0.5110795733640705,0.0285770826104369,
 0.3450080792590354,0.2504067921342027,0.3570734330728903,0.06139995989756786,
 0.3605017735870356,0.8859245212401984,0.3304787737672227,0.42717423177304203,
 0.8980924458648127,0.30873033312251275,0.582896700392381,0.958236113028499,
 0.25757113605786375,0.18277476911301882,0.8215071203859597,
 0.2999546301670487,0.871951173048656,0.8974585197527096,0.3344669865875648,
 0.746775535774103,0.59066102612514,0.3023436875916996,0.39414554717110273,
 0.7186201238140885,0.28539057385740507,0.16044757336473225,
 0.28575590260256745,0.929875669485636,0.9008867922688841,0.2917639962265145,
 0.6450014707596236,0.03622901943444412,0.613910860292532,0.0651070875397225,
 0.532837185288583,0.19483929341543638,0.5900671580885176,0.15146092609216066,
 0.32104877840886503,0.9259227772249788,0.26738284339011964,0.798160230161697,
 0.8230247265741077,0.9032732221754225,0.8988156192162663,0.3741244734642978,
 0.14217451052794117,0.09274097756345134,0.14599798108233908,
 0.12992443251291463,0.48324006960812915,0.7812956855201661,
 0.08560392627814817,0.08528996439220815,0.28288622700106725,
 0.18960676011939626,0.3229704532926124,0.20044931106334252,
 0.026226518167277613,0.1165766253470083,0.5009336793562726,
 0.8401808146224992,0.3393575386590233,0.5547945222124961,
 0.024703109463089845,0.6639815461058454,0.32096892323171333,
 0.22882376248930436,0.02577571883677876,0.9525563145555807,
 0.07635665238442912,0.5463912709844866,0.697782236089408,0.13866880800690895,
 0.8300817645122672,0.7941358126023403,0.061975524181918096,
 0.7155844534685182,0.8944888221265749,0.8178735012213912,0.7634251741121654,
 0.6154872637548636,0.6329711308161294,0.6942724051652428,0.8284314533078893,
 0.4452650197827537,0.25429488712345993,0.27190424923627254,
 0.0480106250597081,0.3983366717694117,0.13941920066019797,0.9350516332027186,
 0.6766643387170153,0.9572824608489964,0.8907672238955486,0.913268728373104,
 0.8775691413193859,0.906852255410823,0.4315842753396929,0.7630571045214714,
 0.5454668178167491,0.24776910849586753,0.5493473676749433,
 0.14774970241448482,0.8944226813482941,0.9744121936552704,0.606030501400703,
 0.025576512397548212,0.399189650408035,0.9603341008682043,0.5371133770190055,
 0.9749012987374931,0.18050138146426775,0.46574055432032024,
 0.06529495474761426,0.22437636932228744,0.644864016612393,0.9268567527220173,
 0.2866744563816551,0.6178207353435369,0.460001067547955,0.4869347552900547,
 0.2501931873930052,0.12342497414392184,0.5537460786720457,
 0.42467721196124486,0.5562631683568384,0.3907135480637276,0.8256452473398775,
 0.0428932298456135,0.03851776518573968,0.97781800393743,0.8111075370934603,
 0.15636037556203775,0.6194491142426647,0.6683766815893927,0.7972926169089548,
 0.4698715806812269,0.29337171821712105,0.857843493679026,0.35509651433799994,
 0.6394663651760311,0.26302139021866866,0.07019795501512605,
 0.6978441583544484,0.12450956723220719,0.24549430310008902,
 0.7519381838946597,0.6373249976993224,0.3405795598953192,0.26393584162194306,
 0.7465234627255523,0.908279127900123,0.04803086473687268,0.7145568731130911,
 0.49658522013775974,0.04781638078346617,0.8832167889893341,
 0.4491557751672335,0.7185665536003563,0.30684601858323746,0.9602029234332264,
 0.8558321475621401,0.060797601784926214,0.19624651734765786,
 0.24233338011916838,0.4230941794894504,0.3726528120394419,0.300312606639211,
 0.8414232139864208,0.13190376308587515,0.8361496850596631,
 0.049989709880930544,0.35635649876835807,0.8828071399745704,
 0.02746815240119249,0.3033312783994616,0.2897759228182585,0.7688166775584458,
 0.3199558716154963,0.7947347650791858,0.6647800662007122,0.7549529464230091,
 0.9930524990603599,0.9571215480881949,0.22519277124565162,0.2997379624246411,
 0.05891451002082648,0.9012913952230466,0.7955150865118529,0.2840662233687601,
 0.47811854338154824,0.7998323014022604,0.9741574906300606,0.4556805756740674,
 0.2850878637409474,0.9978418362987083,0.6503842948762852,0.5598856532251407,
 8.207847159191761e-5,0.9143963804809406,0.7405511750450765,
 0.8063971802076738,0.954230613196464,0.019194605943433696,0.752952559814304,
 0.6446237382233826,0.31109165570843944,0.7502471421447852,
 0.04342025762807844,0.7474077438882514,0.4938379431146305,0.9267507455922228,
 0.9756426390256551,0.7253200488103321,0.02875823480621953,0.909208972456634,
 0.687555011450118,0.2653454068351555,0.08314269397120544,0.43463169221248166,
 0.7105994824571625,0.12299515585016896,0.2143366613218769,0.8938228921101206,
 0.9612238546304614,0.7874286686168055,0.9777694722677949,0.8647161104459091,
 0.816141529890389,0.7815862587121178,0.7437176079034961,0.946609554706438,
 0.9998479055659203,0.327649642128543,0.11971479549094233,0.3293981425314527,
 0.10301250961209063,0.23817803812989236,0.7423614898902291,
 0.7832819457020553,0.021211582214831215,0.5910744644821643,
 0.3349529207785106,0.21919861859419876,0.368977601626549,0.45697285895033635,
 0.5705826983627117,0.8553284383337385,0.03538598755705591,
 0.20707957177416758,0.8033216634446176,0.1992283362592513,
 0.42725982156284836,0.7793366749111129,0.05324829958159638,0.850737616379466,
 0.12171091108305121,0.4218165460727976,0.669695775056934,0.7577744338445942,
 0.6471442624385978,0.7273499052850746,0.751470367590605,0.6155920005337722,
 0.7847587890601727,0.2803347900968878,0.473141120738515,0.40288230816979365,
 0.9709384526284426,0.21400620316609553,0.26703842606641426,
 0.29632219564267037,0.5165787491868995,0.8780382792235316,0.3613237596938986,
 0.31527777786801114,0.2391207400025731,0.2253260273517811,
 0.08816965817012767,0.030521377375895664,0.6822584407843603,
 0.403447553428387,0.40560448093305546,0.1868992467814703,0.9393421796009848,
 0.17727180612123505,0.5162446413721228,0.5138039991614608,0.8291929567250013,
 0.9459498173896148,0.24044733589323508,0.5250482485723181,0.6243171560637317,
 0.39159461627788017,0.8975732890584963,0.04502305613211899,
 0.9244186690148116,0.00959096923599112,0.9552600410634222,0.8646940175869132,
 0.19028965513772866,0.8875306811802341,0.9480069876108275,0.5047710617006038,
 0.8776201042597384,0.48777226458685385,0.1950008445564525,0.822127160419076,
 0.7724791054683546,0.5685002079501026,0.11538224755814042,0.418178950762495,
 0.6278939654625441,0.9301108624875843,0.19057247037206415,0.1719823679962833,
 0.11631477464474249,0.2712176149430613,0.23530514941199598,
 0.9181725804661424,0.04987422272727571,0.8205087441024093,0.9639007935440694,
 0.44221327697284907,0.3648588155308228,0.928859146935799,0.311234287548958,
 0.6689064480366174,0.6368915732687893,0.462975290513022,0.7250308996193205,
 0.7928993352113195,0.2059900718593014,0.2858776342057664,0.7496496956618958,
 0.4369702878995353,0.9196939322324237,0.9578721337470432,
 0.024362563398312553,0.5335925095231382,0.3497627740054672,
 0.34668587097024894,0.33816824065430307,0.2540913144235879,
 0.9619756730986739,0.19420967078908158,0.17637309964573078,
 0.8512341652095199,0.49568594031942537,0.8582881500851771,
 0.036501773005187865,0.19522882145269227,0.7339038711600069,
 0.4963565965727701,0.8738680920500362,0.5528367090385718,0.3612846651462509,
 0.4080327566069406,0.9850741855565709,0.39327471689230176,0.5551949358625289,
 0.7965129873871448,0.5113300682529385,0.5948905721454739,0.7110243174704022,
 0.11620137465192659,0.712946787637179,0.3551445164682594,0.6519387096315861,
 0.435599529185268,0.18052162916130043,0.3731980135560351,0.7150807424678487,
 0.22875328760248115,0.8458100390422952,0.5930526033466481,0.961275590930044,
 0.7523893848463621,0.6670292400544299,0.8697303375650908,0.8869779010944837,
 0.297369662025024,0.32052128766068355,0.07297814259720736,0.8800732607785868,
 0.7732548454908804,0.022942299062629345,0.019646657836823844,
 0.024662003184118375,0.4870952299407382,0.9454639833513916,
 0.4120897173132363,0.11844066726070623,0.8486953653649749,
 0.15551046973324345,0.07160263601254868,0.0013086499406769203,
 0.3073476232304635,0.8899576037247652,0.737303046246209,0.26826203206984855,
 0.15158122701394872,0.4548223988390515,0.9133738040107537,
 0.02918691598744727,0.571551315472817,0.017268858260250042,
 0.4618843736271403,0.6895890127268527,0.7088599586175416,0.8171295337131745,
 0.2667791760456637,0.1585374490930741,0.4755075714187338,0.4299500904102834,
 0.8449626210334288,0.00900874296377796,0.41445875516337194,
 0.43236960880853015,0.4894783614482262,0.8093711780256088,
 0.03293667134536571,0.14058243757191147,0.947519145874469,0.6378373117411411,
 0.39159783419000327,0.36678493567478176,0.14591838185009376,
 0.8048374375641427,0.2915462626893722,0.24863452540324904,0.2765902580451318,
 0.19987298551268362,0.7790709743674439,0.2786622525022999,0.6166517097313247,
 0.6959842077331068,0.715372672737506,0.5040203745206171,0.8705046283186716,
 0.5201766415100242,0.5314993702094282,0.47285800481559936,0.961279215064335,
 0.18487968384508302,0.8295444136432185,0.5141947261534626,0.8046327146108494,
 0.26243803108507746,0.1652389201759118,0.15175686599438676,
 0.8402233027408875,0.11941516232055016,0.19959370482525163,0.544147731038077,
 0.15456218589269677,0.925069795218508,0.1735316004793369,0.8639955180382395,
 0.6575986435258974,0.26621225660249836,0.3646088620306125,0.5596866216921389,
 0.23285048599429947,0.8058283176199706,0.09290395786507133,
 0.4251946179626437,0.0824517609720965,0.06809332325947592,0.2875689243269892,
 0.44826855685307143,0.5099093574950977,0.6925456511193667,0.9781774029493409,
 0.7958480898344021,0.23382208342661248,0.9865261298404524,
 0.016407468165162253,0.7981491318328439,0.31330898837001464,
 0.2582407179124827,0.37605166288822056,0.4340061128728061,
 0.11369271005757586,0.720360048495909,0.3134653862337108,0.6660248105103304,
 0.2264166957079834,0.1475956889903921,0.1620351143830464,0.42245426358866034,
 0.7617325108413439,0.10711871296764186,0.7473824221732746,
 0.33715885572487836,0.6654272166158645,0.3417264483635166,0.7177455878821676,
 0.016155050324772824,0.9554130761246515,0.5854618361228848,
 0.016253942581571867,0.7273280845520762,0.42994185097538007,
 0.7712752821090976,0.16891867132686744,0.14575242967093427,
 0.9913348846253348,0.045289149136348406,0.0750394420703463,
 0.3472992889154525,0.5236076690736391,0.6907198059044168,0.5321004524246815,
 0.8478744079348246,0.8984991119038932,0.20243678682865762,0.2039716331149357,
 0.7060931469256124,0.9038680588718709,0.8083707460626111,0.23140433934738125,
 0.20018149009279962,0.7363977835078247,0.9095847100366724,0.6656328692109686,
 0.18224376130921005,0.7557234078599409,0.4598190339737327,0.4527769455942886,
 0.5276916939092189,0.8183529431814868,0.773588428065083,0.42718268737665244,
 0.7187595236495594,0.4228971864892421,0.46903609178017835,0.1404824438094734,
 0.10223442257068016,0.35231916648328165,0.6897736570282207,
 0.22631802329900363,0.9201659056013171,0.9151576931782142,0.9304893529842543,
 0.6420833897483685,0.8851917906420858,0.7524468973445506,0.34021568496587196,
 0.23639844688391476,0.9969808686908813,0.5393012687314535,0.6116671934526409,
 0.9918424493417738,0.653298060889873,0.9478007669928862,0.8789207732229989,
 0.6631264268293774,0.5148599028718422,0.7689000205932889,0.3168304631946174,
 0.20948513023054072,0.7047912309798045,0.8619346126659804,
 0.25940358946865893,0.8149485261287984,0.5117089389931888,0.5082413074658506,
 0.08045821306894907,0.9377583304092422,0.2185768118906808,0.5333327508141547,
 0.8059279438832285,0.8528531193526036,0.778101968079935,0.6789170245163122,
 0.043039318694423456,0.32799996191469616,0.4019229699988729,
 0.4486639397832436,0.6993214196739111,0.8593084522940646,0.5521314561871971,
 0.8371802614382787,0.8759853721144538,0.053177615705464865,
 0.3692099887770113,0.12074650634370754,0.17591314056051677,
 0.9423741866266877,0.17671340280553371,0.672215781785069,0.9603431306737231,
 0.9593002183884742,0.2899960835190847,0.2728027975048193,0.07318228192476595,
 0.31381424550062176,0.6175498452238417,0.6019721788883015,0.7054588749448415,
 0.489694235179961,0.10817855326416737,0.3346773417445469,0.4930127302592978,
 0.09121803277185281,0.2012392081421488,0.30509484219782557,
 0.5967321935995455,0.13715209120121274,0.751220145341432,0.08337906205984691,
 0.9324621955994701,0.6363471602039791,0.4708435453614208,0.3401264440711097,
 0.7240841995549069,0.9423200283293064,0.1683327429419118,0.33180874624735013,
 0.38841773390247925,0.6780743278234656,0.19859006150528857,
 0.033290145688709916,0.7518459068576587,0.18224393485966328,
 0.19936784824248843,0.18638681632930587,0.02044212463313866,
 0.677730567869755,0.3535966433403366,0.8738517227415087,0.5122191571343664,
 0.03946961741017263,0.4669945064589667,0.6301288772081428,
 0.013349078590579766,0.8451031707329222,0.9639538843737129,
 0.4301157025708082,0.0774893139145304,0.4267417015188545,0.47482219312941953,
 0.7017982181037122,0.1423252979300672,0.34946283375686094,0.5864942262772092,
 0.308554453374556,0.0245129775610341,0.38396274850691536,0.45986828609622377,
 0.41020788237149786,0.5048638097740457,0.6589305989087516,0.6224627067965036,
 0.14925424533076892,0.16325410652966799,0.516824650487784,0.3680271630872938,
 0.3044985723626886,0.811057015068099,0.9508779473690844,0.03363103287796798,
 0.6654270663683092,0.5401412964509286,0.3094103183643617,0.9442701939873345,
 0.2235360307403229,0.9708358352349998,0.47171747017257704,0.6399672549823522,
 0.8467479707702951,0.04437931543393958,0.9093454936496095,0.6710757191321971,
 0.31690789267813746,0.029293125370951967,0.7301050015747944,
 0.6007544744459412,0.3556551461176267,0.8372210572651152,0.02972607435095953,
 0.9375351383096338,0.16068891908783878,0.25406681790108765,
 0.01790994672048296,0.37877968858928523,0.7021756906925385,
 0.1691613476720636,0.2894714908092819,0.3592901932131547,0.43457758693641135,
 0.7803240337987306,0.11867551092090345,0.8356072042799122,0.4533715593864426,
 0.03975935297634203,0.010741665250412313,0.30080444025501196,
 0.02421611502430132,0.9744421381548124,0.0042103338376900945,
 0.739873784323107,0.9096472011632785,0.9151992868106594,0.42072757065139543,
 0.6945163813701536,0.43251574950915894,0.0419593875518095,0.7778942880427191,
 0.11290497469321359,0.8467559416666088,0.9200258383649265,0.6717035882914054,
 0.49109484780067747,0.4475861527942331,0.529443579642455,0.5995690929761046,
 0.620560889759994,0.9090163156582158,0.20097635983269602,0.6787488379776472,
 0.035397770717985666,0.40159793589527437,0.12892844627780664,
 0.9839364237626932,0.616600209630651,0.5930609412249455,0.7375337697342261,
 0.8628773725649126,0.9840461096813158,0.820838703117767,0.22759265986906507,
 0.8724560718595574,0.19363967604035326,0.5554564408734918,0.7285229370041987,
 0.09193829380331553,0.806936372846064,0.8316632840551696,0.26274790351641686,
 0.19726923941715546,0.513482857033213,0.16068040677550766,0.5069248794683274,
 0.47300429413665923,0.9981863677270871,0.8510140159185843,0.2682782221630554,
 0.7318809035380962,0.2321215119862402,0.9748314164276619,0.2882839465849265,
 0.25622595221348643,0.9164458434996838,0.036388431251433095,
 0.2852143672042474,0.5893682370940866,0.019909495461267634,
 0.8388528594365534,0.6434199366270942,0.12290901503018237,
 0.044817150748683865,0.6470567576723137,0.1507271011078536,
 0.19841857862369705,0.03997900286789302,0.4042929142307381,
 0.6688793782626042,0.3648821998954922,0.3564379529707904,0.49973895388610223,
 0.037067750808478794,0.0791721095987985,0.06545009320123973,
 0.2758724889251071,0.4933135971935382,0.06516028146818431,0.220014873781065,
 0.10877955429393371,0.5650370744349117,0.06822348060969108,
 0.025135557733264924,0.18203188145759752,0.7423033105791481,
 0.41254131675271943,0.33409205577886536,0.9340046573105,0.5600931952940066,
 0.7913939715284922,0.4425833747644541,0.9345034167452757,0.6182916290649088,
 0.7529051891836127,0.5561049051200537,0.6678430697576071,0.4992402715936489,
 0.16532727694414173,0.4737972578155345,0.20105457603079502,
 0.4089178744920162,0.3071774606091968,0.5685064966510078,0.8412797576114004,
 0.9306016127806318,0.5230516866432792,0.2041547343468625,0.9425879507569087,
 0.6328576451545567,0.12412189178882971,0.26452654141862597,
 0.8090552238517456,0.8067303695075297,0.4647706201744016,0.8083615148499029,
 0.8727532698849637,0.8687510670654353,0.7806063349757042,0.6126053902834613,
 0.8851442214244192,0.7720953246486955,0.5156661326562206,0.5358014685582951,
 0.17703807453803488,0.9994179283150189,0.34641075824931566,
 0.025533334016058395,0.35371904843521784,0.03640740890370331,
 0.6801926649135601,0.15875990215843416,0.317736665923516,0.7188223415733467,
 0.5813393447071058,0.9808519899770742,0.0848748425014908,0.9392411168371466,
 0.5375893704555358,0.8291967107359239,0.9386653501013483,0.8111349184108814,
 0.10770504579053974,0.7678826273547541,0.6697063109389142,
 0.05168211436661596,0.6497307333602437,0.7556953702111507,
 0.24183699772423017,0.9603235280869797,0.5179019774299514,0.5567688721606744,
 0.25849540076079225,0.8854128196761039,0.038756826600220395,
 0.1853840514945111,0.641038016634883,0.6064671442623362,0.3122538832884595,
 0.0710786202792052,0.7242754648766576,0.09421107232697035,0.6467147537055581,
 0.5873573225059296,0.2741414138326268,0.8816564224191082,0.8652531567237151,
 0.3858965189422351,0.7363591034304126,0.20518753241972365,0.1817512243611159,
 0.06365742445091782,0.05506836173765306,0.1565950099794471,
 0.6657660211202056,0.29316806043872257,0.7880303765501272,0.5475926046034519,
 0.10804682512481179,0.3363098503859232,0.020937137085460966,
 0.8210751653187338,0.6141800348913685,0.9495210057397689,0.697322465904858,
 0.06549106330748566,0.740033250701596,0.7831559477310649,0.9240769080494421,
 0.15812404970780347,0.0722723671393034,0.3261053458379669,
 0.17867385007532943,0.7214068627380196,0.17983868455941954,
 0.4038906183205353,0.6256528291636075,0.44564502123275906,0.4375869909173795,
 0.3813309404554249,0.2944507531099324,0.3670221295714886,0.3727300552926107,
 0.14027157596709117,0.3490372367346688,0.2932178068691791,0.7627970435410099,
 0.8516562305917791,0.20666694725613866,0.8681439691820846,0.4085884618910719,
 0.3482872120907352,0.6912087375026922,0.6292845460589599,0.2509179797927702,
 0.19185687432291265,0.6516296821929518,0.8701949349284047,0.9983914417453531,
 0.2713700342982891,0.2860856274151692,0.5294759583298873,0.4406450768401966,
 0.6805934604886663,0.5628288728802583,0.32640076425947584,0.5099632848087907,
 0.13462823115384215,0.9348633673168965,0.46689993983295364,
 0.10511340142379877,0.7024627295417512,0.1927603002898366,0.9440694068699323,
 0.7290554922785375,0.0837846021519324,0.755593350203355,0.9064319273280436,
 0.5393818929411233,0.9554138577674557,0.6387276483579952,0.17207058703356537,
 0.9869468753324827,0.13518700769374092,0.039606782966611465,
 0.593600826226992,0.5835389855556049,0.3604929349837206,0.20444025179345504,
 0.188858099592779,0.19759432370174168,0.433102920464985,0.8030884068179808,
 0.2333009045581662,0.8186021679880116,0.33886494176181425,0.0735158020264528,
 0.6174039469214414,0.8102717355226255,0.0945448278145754,0.5337337813539905,
 0.05710353429066861,0.2733290793320188,0.6352880322404784,
 0.17178028386086552,0.314257284176931,0.1528671346119327,0.2616078693263262,
 0.8763266849820867,0.5110076424960641,0.8700278942354686,0.08502196436645837,
 0.23234478864360608,0.3527159689017181,0.18066629118557198,
 0.5309336427956903,0.4225750708364058,0.8685102159339433,0.7239609494988395,
 0.47943676495489673,9.983895552623245e-5,0.5921391527872096,
 0.12224799381355411,0.655402343994482,0.13638327913798687,
 0.34298184564319545,0.36163474119839856,0.5517475790760695,
 0.14221871466192082,0.0390025463102065,0.6082053810356014,0.8805131081797487,
 0.49414656657094447,0.9359629969757158,0.9598452228946304,0.8099634666004811,
 0.7860469582739513,0.48558368645850514,0.09277806300553204,
 0.3401854286245015,0.17049054086098836,0.3462379263825661,0.8520358286977302,
 0.9220479831179405,0.48088918361269306,0.3715250637228813,0.6802720412602179,
 0.25535753490617297,0.9589202199059959,0.03883999140119765,
 0.5157052355344268,0.037292159602715036,0.18491172257406963,
 0.34978027933839295,0.3091180091311896,0.7162894671771632,0.3752055436204731,
 0.37356160709893893,0.6091816961236076,0.12764598556256435,
 0.5127047633197669,0.1554022732742233,0.681986681931164,0.8382717842946654,
 0.44294100107495693,0.02205647944113931,0.018743556618383783,
 0.1985839945533604,0.780665328750102,0.7602258126988259,0.19136987698852248,
 0.6729638485010693,0.07225561346255105,0.2614864248668187,0.3693416235191016,
 0.684914613737414,0.3352548926884641,0.48701966482585357,0.4777823621986218,
 0.9528268418168319,0.8925654879671969,0.8432501780900911,0.5061570467609217,
 0.012460323800017825,0.29103768352508264,0.2721556341867777,
 0.8943594596569865,0.2786315504000334,0.5836570007183154,0.14977020410727515,
 0.9787599612372067,0.14036422796339787,0.7153126165415634,
 0.03433150714994848,0.8715271085031746,0.20452855988387375,
 0.7806801590321122,0.29077355062919363,0.13618073979563527,
 0.26519508583166707,0.4322110228224647,0.4669065856532064,0.7793896791477584,
 0.23531398600755338,0.5561017275695523,0.6517530608412633,0.6679101026175108,
 0.4437077368007847,0.1562612728007301,0.5595077597264727,0.5143086926248852,
 0.0039160657982263425,0.306557397289279,0.8361915696179988,
 0.8275386174815021,0.6314496392138532,0.5248781420014622,0.10057984812630583,
 0.7753809503658143,0.36333537145996353,0.9107352675703002,0.6054004662814646,
 0.12818787089213435,0.44992705599412697,0.8426801564734658,
 0.2814304955238204,0.09924633568777175,0.24435626850024617,
 0.23754027512216325,0.8937728066144843,0.7091786582095896,0.6214137739727539,
 0.702519089814037,0.114636708464941,0.6369966004160773,0.2854798967950567,
 0.3506347052312876,0.5656517082578063,0.3475650889884061,0.16461384754370634,
 0.9338211721184349,0.5380395501568654,0.7854057606613942,0.7215344190187747,
 0.4485546932826555,0.05710980774935892,0.9190885468437284,0.5782761598354125,
 0.2857775272296802,0.9234669863406233,0.5765862385700046,0.07052967561914891,
 0.14737135183802674,0.9020255288369735,0.17944204343028747,
 0.7576786614774947,0.10942092022860972,0.4846283926577477,0.581916579609922,
 0.9285515049689249,0.4835673242041678,0.44128356952782044,0.3923563435515679,
 0.11630180221967301,0.8263536518644434,0.35543728330520685,
 0.3093131337328866,0.4233954030870034,0.34398976405171555,
 0.48661565365705395,0.6699507780280505,0.23546010998715705,
 0.18147593912489934,0.7338309085409976,0.14066011156543423,
 0.7475704294962918,0.4457152682607628,0.05394146481935114,
 0.24294206034479848,0.853751358127662,0.43727814257874553,
 0.15013326155367435,0.3635070301869362,0.36027358600401627,
 0.0765766695874549,0.5622433695175341,0.4042006085494211,0.9299881180625627,
 0.304121238558564,0.24202700266341348,0.13812893632339018,
 0.12486379461791564,0.7098504621635339,0.861059037051386,
 2.8790307801418447e-4,0.9763115083094884,0.6603153731931533,
 0.835514188767434,0.9452431823279782,0.10580940445223352,0.9595248296463956,
 0.8926135803390401,0.6316768374152448,0.8971644945007884,0.22265611949212671,
 0.23859910164550835,0.3005813689005101,0.36024184468511133,
 0.11159982539321722,0.6992050195160416,0.47791241094919046,
 0.3761959355015668,0.5863127233057143,0.7087637556792061,0.3118656980534438,
 0.13653882740151502,0.8417404114802551,0.3248198760315998,0.5738763971374655,
 0.5218397218569456,0.2389175025360052,0.8654884228124525,0.5039845200704905,
 0.07084581324455508,0.7593594622264837,0.19570512557653752,
 0.49155595512570505,0.8742635170547192,0.15294262982034812,
 0.47044150946139807,0.16040886389804587,0.8441992626263984,
 0.2525522802787603,0.23755012342610649,0.8779148631934159,0.4643314616450779,
 0.42369793672633915,0.4718494701816215,0.02103507199463217,
 0.3425961769531637,0.16824447943262566,0.22491781477664896,
 0.5111831202391816,0.040595662237020824,0.49379619663337593,
 0.9910177656252533,0.6061624227525041,0.981486646314755,0.9100359667370878,
 0.7376605563043802,0.07238981283523716,0.589642554784459,0.04896031756590136,
 0.2901315267967445,0.7340717285090967,0.16614804712666553,0.6499534811105074,
 0.5753050074893735,0.11227798539306633,0.5032184836842555,0.8372770551582797,
 0.6216212227722867,0.08165606325966035,0.29206368447787545,
 0.3157570403052897,0.5499052267762989,0.5330441795063118,0.7368126032867323,
 0.9423936191739037,0.031211780508726417,0.6640002871608199,
 0.12463516420299281,0.35417968421130486,0.5973127102977763,
 0.9582764729546488,0.70769032422963,0.23097256062210159,0.9427703421828979,
 0.4443371919558897,0.24234478500287016,0.2416876673839533,
 0.43128270171668304,0.8070824043804633,0.8738007007275929,0.445719298020371,
 0.3850750331468209,0.3857834872286101,0.09603094898097841,0.4730289343217243,
 0.8691383802520363,0.5255180968254816,0.08357617279030927,0.821657683178925,
 0.42104408165916807,0.41715357770900274,0.9873314642357376,
 0.6847897276752241,0.27084759471923214,0.4831648265869455,0.9554635413016469,
 0.07563455410571351,0.5051546421006955,0.7951917164100207,0.3958185052319515,
 0.03710067797082639,0.1618659844348611,0.39033589012387127,
 0.4683495662576934,0.5616482991152711,0.4201012384066558,0.8292860475952352,
 0.587286996635719,0.6333391515942894,0.12345126649598293,0.9851235104731393,
 0.3057327729382955,0.5595238462695608,0.7397958884976586,0.5879926477253569,
 0.8835816691238398,0.979144624796557,0.6623450535708839,0.15364616395511232,
 0.7985906154625769,0.7669098950655208,0.04752465441219855,
 0.17107992723634524,0.15329023000524866,0.09431206218386678,
 0.36625026291018004,0.07924642301180584,0.17807182992587722,
 0.9272800881784757,0.09050257301654552,0.2649572047923865,0.4338445184048959,
 0.502817351188804,0.6395951301090074,0.8653896326433396,0.5235499290008023,
 0.7848898348139701,0.49654358790200026,0.30250750952530414,
 0.7621672506954127,0.30802699968936786,0.6643811037064083,0.5707382944413854,
 0.836664220417509,0.17143497821384202,0.8115059398447853,0.06526378222636908,
 0.9824998288899732,0.4396176456735159,0.7498803743958544,0.2230496895046693,
 0.8365475664560891,0.862596037334693,0.32084867855459964,0.07923094566026301,
 0.9278376311793067,0.7603682132750944,0.38215665181169856,
 0.18146283377544925,0.353096170795002,0.21069232171427021,0.7402155068786702,
 0.6503433138042392,0.2718420128049496,0.7363048916222658,0.5074085285141052,
 0.48041508652741194,0.6116307830424588,0.9614953255253345,
 0.11295817968297528,0.08888584245790376,0.6997200397645382,
 0.43219759284773407,0.00822341912232638,0.9875482960226464,
 0.9643598328929549,0.23185530394170117,0.6430499172788832,0.7393404157734957,
 0.3593883621755437,0.8962876945358749,0.3754897762442504,0.8671057386472275,
 0.9003599420888687,0.8865430806911879,0.9519095791231938,0.6990925252789598,
 0.8753895273825549,0.9696951166951359,0.35778827441366856,0.9138171162577238,
 0.1284140512180173,0.6302400823299461,0.6763608867503138,0.5075376453292757,
 0.7514327521067797,0.8607751759548818,0.4877873553576919,
 0.005539808109604083,0.3759121226585497,0.8735411358233752,
 0.9310762000591353,0.8715107437600896,0.3506106506309997,0.671609424809161,
 0.49064342708537034,0.1666907199797617,0.8380186473237704,
 0.28174223181209745,0.21544583712526544,0.9753943458702002,
 0.5431330617854492,0.15181294355623998,0.09283902152271795,
 0.31514691503175596,0.6356933737442031,0.574849183799826,0.9548652257625511,
 0.3333061025861065,0.2692506035094948,0.5325736381103886,0.29890612354374324,
 0.6591361590787896,0.07099233703063113,0.5872456145837861,0.0893835011744506,
 0.9255193287505413,0.7910210968670937,0.024371585432588994,
 0.15529426949031822,0.5733433701190068,0.7869567200014527,0.9317544537552808,
 0.4600735171498833,0.7451571403832086,0.007424323917623488,
 0.2913988987132442,0.5419850446849876,0.24500970100602715,0.2347817845516792,
 0.6028284718798573,0.5483038595677872,0.01474913502517905,0.4806314687914369,
 0.1679175801786893,0.9706450073026863,0.7127763688799079,0.6542633683478006,
 0.8796592385693125,0.46052121572055527,0.8533428977380726,0.6962264367381552,
 0.357096792050248,0.7076371476666674,0.4762740133308492,0.9850713765325477,
 0.3811488404119733,0.497419043288266,0.7827462196248325,0.8607000192548915,
 0.8401190974355721,0.0425968936366683,0.3547560628769687,0.06647276548572767,
 0.6760170416980684,0.19008519484710173,0.25500659094715905,
 0.6070285030566659,0.05032389774740442,0.5740099060018591,0.9488650806858443,
 0.5078129451745678,0.6472802377747499,0.04131304062337293,
 0.08926294888224806,0.42184772007257,0.6890705535658527,0.5046069675611485,
 0.09558023048348274,0.15849272340817266,0.8650903555039193,
 0.6851293034865034,0.6942453131069533,0.7311613981340255,0.8118140763723836,
 0.9349432136718059,0.00961780638452292,0.21291364432071214,
 0.49156176562798704,0.7921728033017416,0.5104593516684487,
 0.18973684734446206,0.03371468850050596,0.36072131940086116,
 0.08883192950300778,0.4734858747303289,0.8832893116038002,0.6886464134397389,
 0.6619154615085022,0.9712409863618287,0.24882190568655727,
 0.18491255557624242,0.7795123523731167,0.29658768448565087,
 0.5851375004714032,0.9636179150847015,0.4956667280933872,0.833727751480615,
 0.21784689272287328,0.03199055403679263,0.3690232384505504,
 0.2370878097021597,0.8767443392115749,0.030600390002611166,
 0.18103355723617542,0.5316293331703689,0.7883347432973004,
 0.37967064159183406,0.30631350338159713,0.8602493302924953,
 0.03998766611368554,0.4731802028634061,0.9422194601768552,0.63065936510241,
 0.15125960143972528,0.8547223641347398,0.7478684439222492,
 0.47801998206069096,0.04256982516013974,0.0817537305451882,
 0.7398277308489185,0.10152726673779089,0.4912681215202256,0.9393481367006967,
 0.3907834949454504,0.8107039006486628,0.02107642712352642,0.5475039100806857,
 0.38119051798568493,0.44122891481991366,0.11982448789352484,
 0.23277549082360083,0.34499384557485124,0.3433594172215364,
 0.2088151467519328,0.23236587335018055,0.6538809798998857,
 0.18280803227416786,0.1442320096432933,0.28252972251347075,
 0.2946247365853585,0.43367479026684264,0.5470903690987741,0.8846552790257274,
 0.5236546355094563,0.2236898517560153,0.9524379207157017,0.40325393173360724,
 0.998616483083547,0.9036936765826669,0.8059235229025492,0.1773942444017823,
 0.14939769032668393,0.25627168952039736,0.31478724864066354,0.20630852895452,
 0.2847777393293144,0.9213930948535919,0.8403644426665884,0.6947314068249592,
 0.08224508984131829,0.4502314502756045,0.2601869669935777,0.6399292767917677,
 0.7508475123161839,0.9623268603057632,0.9975858974348428,0.6536257776225228,
 0.26448214535790004,0.7040001402625564,0.6777490824721024,0.6162081649870592,
 0.9451686512854761,0.1688764216394033,0.8437653564429053,0.41576242191948853,
 0.7287085660029264,0.4897111080759783,0.8107862963516838,0.9840941155410042,
 0.8279590042401387,0.8710254840050401,0.883888836104095,0.7250220149710008,
 0.1439007879760914,0.9224262387815781,0.7594561499431156,0.22611286322018276,
 0.7069746162822055,0.11088251505485047,0.5457992763962662,
 0.17830302713435042,0.8290120380333412,0.2587330845816139,0.1340469271043403,
 0.42199623959041144,0.44610657119245567,0.7380370833340262,
 0.20200994203309874,0.6125717995481519,0.6017164108507794,0.4417076470774012,
 0.022991605653516878,0.11220654724354928,0.5139786026830203,
 0.670930034597726,0.2749170864278946,0.8725291441659802,0.041480175821633436,
 0.506565875394481,0.127079082366378,0.0908166147834012,0.8844760218683588,
 0.8514455893914119,0.30257940314338994,0.648173315133515,0.7777574692388417,
 0.6797652126157526,0.786120205849861,0.10352859455995533,0.10563810281331376,
 0.47498969088252596,0.9993079990814379,0.16947586097395206,0.732832574578818,
 0.7441263320845392,0.48138943745697294,0.43629172402841343,
 0.26144543362969386,0.7241477570593208,0.594741905049269,0.8644220684738309,
 0.22040345918898374,0.7077297944272855,0.8619194262320244,
 0.13545609983461337,0.15133657483060226,0.4736968995650055,
 0.22605919145161302,0.9963611420879986,0.7855734927489334,0.4847348424266108,
 0.8886504503319353,0.624392592600473,0.6582976005823498,0.421399379360188,
 0.1901870734762653,0.7358784437083126,0.8046311246804968,0.4638775532591842,
 0.5766390073083532,0.43605261180785904,0.8348072304677849,
 0.05745116295613106,0.9329679680589684,0.757358849247672,0.8151044957526306,
 0.0652389540342716,0.9025379958163692,0.1494021032633982,0.3828384382684352,
 0.5055720798586185,0.3678456934279457,0.6265543220987779,0.7049225048919554,
 0.06617079270282344,0.8283453524906903,0.48715091048004355,
 0.3155778187483924,0.007273323048951785,0.7842425806065622,
 0.06657794914563653,0.8046915956525392,0.078683960594082,0.4512325212788998,
 0.725946708859849,0.676054179428286,0.8433729580039028,0.780139527759776,
 0.16037348056761824,0.858910246499301,0.9341805669052173,0.7526458094696071,
 0.10725979392897678,0.046573957482978656,0.006308541136852974,
 0.8912278870719703,0.9699463012573492,0.24366851324800376,
 0.17616181357813943,0.8364147213272256,0.7486838689704355,0.2733118847671212,
 0.4616751902387125,0.33030685938676285,0.6658299801459384,0.9971267085440444,
 0.7805080473163895,0.035668442792454735,0.580353477224013,0.3776893428294893,
 0.616701848455064,0.36880708440523435,0.039908580186727494,
 0.5220905038500657,0.7321091053626504,0.3432149795224734,0.26715638486371784,
 0.8698873622499514,0.5849558820800673,0.6010360850835603,0.10872144078079793,
 0.6046893467421179,0.21386963735427633,0.11274535967994215,
 0.1458141539919855,0.36141902949670324,0.12959286824861693,
 0.7377032188480416,0.8398518403530638,0.3963828709590974,0.32795151162251435,
 0.8099044472805502,0.9622482503897055,0.07882363505642354,0.7605999024543368,
 0.555569328663224,0.25840875534622687,0.9108741845885509,0.7869060017734224,
 0.5741039737294273,0.30532746510822084,0.5678335620954158,0.9490254416487995,
 0.8999872307761642,0.4144599222436839,0.783611751235219,0.32591422861442765,
 0.0017495129672134846,0.8163764630796547,0.5675533864209026,
 0.5645066832231634,0.507090168098101,0.3271770155965513,0.6240895739492698,
 0.3497670703043819,0.9000544139281305,0.3866836254828817,0.5334045198659388,
 0.704224994011568,0.862995099173827,0.23376806294686392,0.17853558729660235,
 0.014382333630564847,0.884865316770854,0.028576245301105252,
 0.47921971183202283,0.1332128478978496,0.8391594223472332,0.9632041979217285,
 0.664383500909156,0.7626776928655388,0.3364569601977202,0.1933493826413808,
 0.45997152389075735,0.7475856344006583,0.10691886173638465,
 0.9467758229683024,0.2467496443491064,0.8620113964397071,0.899078095995639,
 0.22706652836699603,0.9250984411990464,0.020192831588176197,
 0.9284281467202962,0.4607568772812337,0.8045107750987396,0.46500026575587383,
 0.1090632641530076,0.27523586277287626,0.7208005438691335,0.7477700662783267,
 0.05710616803933022,0.7430198800970912,0.771059262325822,0.5721115486309648,
 0.3244416818308249,0.28040367939916955,0.9762787605605407,0.9075330360511554,
 0.038810164985248696,0.5266083227109298,0.6172916778679678,
 0.4130053506063006,0.663557566845419,0.578905794139395,0.4965177139203668,
 0.5308424603106832,0.7882742184122974,0.9985166961546215,0.5460166668681111,
 0.4083659999682907,0.6393303476305234,0.968005719643139,0.7732264266315814,
 0.5284473657658905,0.4804145166788568,0.9254939184660445,0.7022695220243999,
 0.10976535543710231,0.19006891480442423,0.41876339350818703,
 0.028122131882229118,0.9780813240602895,0.03963856728336512,
 0.3949016931594471,0.6129044220805678,0.030364758063624486,
 0.6447498186578238,0.7956229090945656,0.4903836448264285,0.11351882563327065,
 0.3203651456443085,0.6673356698123005,0.23923315170546755,
 0.49328286815192524,0.9642571269358223,0.17965883429014362,
 0.7056184990134138,0.9509335033863232,0.5101459579757293,0.9736597120962286,
 0.5838583910715236,0.031788751804272986,0.2674189103986564,0.7020151566689,
 0.026270504996978694,0.7228386067751442,0.5777627940453182,
 0.5605039769589754,0.527104728925984,0.24355919015038396,0.43539438389171425,
 0.3065595751252784,0.42303371771045484,0.7149980009757462,0.5551753082230687,
 0.6709610845578043,0.7164328064960269,0.2024885323955733,
 0.004296463280723972,0.6872479619047567,0.7307538004634478,
 0.9881191429120206,0.9607252694375527,0.755721540302464,0.9104803285743485,
 0.6308696371034863,0.7299135947587877,0.763193366999652,0.2867749416150993,
 0.45875775312785483,0.9584415401677159,0.7440683934889656,0.5643601861279733,
 0.12550702175317519,0.9932931541272001,0.10289265268128323,
 0.12667362645769398,0.5394573145536304,0.6908702276242695,0.0325847614310828,
 0.11787936090881557,0.3505503342346208,0.8525316438424569,0.8900937939839745,
 0.9851856275010985,0.592010963254909,0.2461737507687387,0.9327293741688477,
 0.6158302001514503,0.39148639161312926,0.4400225055920344,0.7010210014045004,
 0.7505957838115276,0.15647004633820494,0.36586006301195473,
 0.2368563704041532,0.5837039476093906,0.05468160286984225,0.6246917161580375,
 0.6217414089199421,0.060818824603024235,0.2801060830082712,
 0.20279527392262497,0.9464869675492155,0.6550355367440096,
 0.20068422320290868,0.46729798146292967,0.5505861793676025,
 0.2775032148982166,0.2630254617579021,0.06930506227398625,0.7968597933520088,
 0.879718058122575,0.7307003013454049,0.7455768913029307,0.5180559645618407,
 0.6381808144088901,0.40969296216913886,0.049649100594717055,
 0.34208019415964996,0.5806238495671323,0.5336472491382992,0.3377692557368249,
 0.09434625667769503,0.9817262781557657,0.26154021078514744,
 0.1390457613231808,0.894709291207162,0.9201810671851,0.4602101336229881,
 0.3330164167022649,0.4903140799436564,0.8689747907470459,0.24663921717640025,
 0.7223317975538287,0.19875001296796468,0.9038926140507126,0.2553970636114159,
 0.1720322860230855,0.29596919840185854,0.7933488448469541,0.1557202937141403,
 0.6221447774457705,0.5399104058493345,0.7801409642132265,0.809561640060956,
 0.4638323166042003,0.01227058968082484,0.6737361232703827,
 0.48001134659939937,0.46817301098722497,0.35550994488096765,
 0.8561651218125763,0.0602082013507792,0.054828758429303126,0.735982000883058,
 0.050777454989252235,0.32800745589327984,0.875475856295936,
 0.9543757334773372,0.4385576732071549,0.3132916954214816,0.7547145594185302,
 0.10176159330232049,0.623994129103414,0.465300246851728,0.6817195935118422,
 0.15794105459498398,0.7766490948283048,0.7923247605992975,
 0.37074491270364873,0.8061586691133131,0.19926146668734912,
 0.43871446385230495,0.5829335358045692,0.9739985008434138,0.6177702537165114,
 0.592946979643307,0.10494309566026905,0.7481442071783262,0.7605013979617229,
 0.752784946029788,0.4736714119223808,0.25117762162035306,0.9072937440144612,
 0.221362199225823,0.6872943822349448,0.46714208297591375,0.7781139575608719,
 0.40597835926918646,0.5685793927672781,0.8145616683978244,0.8165472212255755,
 0.6489077446228776,0.4345325256266793,0.675725963913648,0.21090536580577468,
 0.09951706023929496,0.7027901636334885,0.043183495952288764,
 0.12052241038990474,0.9428641554637632,0.5512572112508627,
 0.10338296093870603,0.7803661498412809,0.152070310694152,0.27238948029107446,
 0.7642759733774565,0.7398648864477215,0.030483181718462093,
 0.29612821203224327,0.25932504397573564,0.41479804825916045,
 0.7108820759540474,0.7980711055609706,0.4871199812515571,0.3154960206225448,
 0.1356609676537266,0.8230536883167086,0.5929471222494417,0.8191654316540755,
 0.5951069221505422,0.3182936145472256,0.6141406462221585,0.07493251065853623,
 0.7865455973768768,0.2684420653358862,0.6191662804675293,0.2671776391379507,
 0.08233353712967284,0.12830923697342778,0.08876642376354249,
 0.16264070995829638,0.7327555166955106,0.34728812909393125,
 0.4605436017245571,0.5715686482682993,0.9565299018597244,0.00829722426233992,
 0.04972877927207586,0.06756384445773644,0.42791043407050877,
 0.37480413641123944,0.43729156483691733,0.11002613331871691,
 0.48788314952913003,0.02573058172787468,0.9074046350830289,
 0.8085027193229984,0.7639956695383334,0.8135463303269141,0.593687672952135,
 0.08126781512052217,0.48054003594424055,0.7944166608296649,
 0.21818199619635792,0.5709162422122923,0.093340764857339,0.6927512236245988,
 0.9444279553306018,0.2570117896523624,0.8124845221439555,0.36233801527068266,
 0.41322082253904013,0.7558005471560865,0.5955642751417629,0.9066191938108001,
 0.1445022537639864,0.39444252038366445,0.646764565375529,0.3442431517105944,
 0.5360713347493916,0.3172480434088738,0.4298823745762874,0.9634771625455155,
 0.7689873264182732,0.9233807727911286,0.7927009247766357,0.6792796153215168,
 0.20683182612891016,0.607659025018937,0.9633195803934552,0.5942198639015939,
 0.24283166246371335,0.7501041866788696,0.5351249803026842,
 0.021239417920965842,0.02594431750791537,0.8624557157911363,
 0.02176871572872119,0.5695342196416328,0.3868228460566353,0.3120864034942704,
 0.46859648612369686,0.8518589039976913,0.5920761503447519,
 0.08933582356949743,0.020842287534410642,0.49481932692038,0.7051752718242612,
 0.7157063554761269,0.5252350503911976,0.15511065416405068,0.8034597175473038,
 0.027172987917163605,0.6960128249914502,0.5283111972520815,
 0.7277496207304974,0.9228761084328164,0.8140953080349163,0.8164436091802456,
 0.20108870982022797,0.2711050965796018,0.1103005342181077,0.7461013802998896,
 0.3879490366925492,0.2787976209242746,0.6020893503259647,0.9939975208209845,
 0.30025253571554966,0.4371376168483234,0.6198081144742363,0.3119418769577553,
 0.5337670904827357,0.9186852193623664,0.7830914826053721,0.2923712318545032,
 0.1250159054670379,0.9096767342311067,0.3060415901614878,0.537025768659189,
 0.5194069849413987,0.8088974336329441,0.3077870522318905,0.8892511979702693,
 0.5743261417342542,0.5984941456176979,0.5479384324807025,0.33331317538355953,
 0.6970250853760006,0.07117517616815461,0.5807037615610121,0.4671954229998605,
 0.855489845536021,0.3013954044026298,0.20383486656255378,0.07558981990545355,
 0.3501338622588641,0.45595327647759876,0.7442924159366728,0.3825438021032559,
 0.34307558628413326,0.8232290508336488,0.14293718291462554,
 0.1850200442232588,0.3078327002456325,0.7462765654754774,0.6090576669191917,
 0.5513865701961836,0.40653920821532763,0.6824153596207223,0.6202106963065172,
 0.6836886406024754,0.9343259756392605,0.12912689743354766,0.8476026931238061,
 0.760993919336618,0.5124225868014585,0.1628409785711047,0.9239695176890174,
 0.4721707884254299,0.5257126779874385,0.19181706266167708,
 0.013486266202349495,0.7806043149564548,0.655292274492746,
 0.10068285842746438,0.27321807024115574,0.9497349776216499,
 0.08336556078381219,0.6528114675078754,0.5533572200879519,0.6767194209308535,
 0.20300051888060988,0.5488638573756337,0.7355419079735867,
 0.10173410659477378,0.9242023720091864,0.9856604116905484,
 0.46230412496780715,0.6885721549836291,0.5956834052142366,0.8348782932024578,
 0.9942228285863256,0.3206767882403009,0.4524533843555488,0.9505645706360233,
 0.08725586701620003,0.6255236518584354,0.14099847877966265,
 0.15173829091859892,0.9011586766169171,0.9448810506031542,
 0.27620199535450707,0.48996201460786926,0.639883363384269,0.4389308848007297,
 0.7623941209298896,0.010501070222323827,0.07062056778741366,
 0.7477594262161571,0.6569256269216878,0.6809412838865225,0.3500709478027929,
 0.8091945418039599,0.30116871519061894,0.09871511265549415,
 0.39091415393298934,0.5171641266280814,0.3238397933594981,0.6582010292775877,
 0.2576568415152387,0.7933590550855314,0.7958947269014758,0.4472656038473959,
 0.10101945541250545,0.5287424987509827,0.2674357684277967,0.5304057364878202,
 0.30761931928632436,0.5934516934539008,0.9369198758315419,
 0.07905946097643635,0.2365900098814797,0.3527792422059556,0.9647884679625273,
 0.01003446847195233,0.8325049906238827,0.3290253116302502,
 0.23497370098792258,0.4914292214333039,0.013869494150595796,
 0.5335021920314272,0.07101979219369225,0.385562718731302,0.5358580945475006,
 0.1049737329785505,0.8067262195142653,0.8161397622487712,0.4009994081503393,
 0.7232245894527102,0.5669062740869687,0.4096159438677275,0.9377931743200129,
 0.11200604153136218,0.9296972289218556,0.24812523197980552,
 0.49590695216205116,0.9645213805563527,0.5803914114567342,0.9729323768921053,
 0.03414987922781676,0.16610396321431542,0.07985220789182546,
 0.16335950323888226,0.6216897436367831,0.2155280126948539,0.7038320286396194,
 0.7662766431567417,0.41701334149691727,0.8561500230153323,0.8255878484486485,
 0.7466291079634939,0.29796059881081993,0.5915043305353429,0.6105978932330381,
 0.09578320658768036,0.08644739960206427,0.44697094773242585,
 0.794971563386613,0.4988031147964358,0.7139123310771724,0.6939589099399894,
 0.5649956731347464,0.30545254865081184,0.04172612341092852,
 0.3800762252334302,0.7199457095918644,0.7172351205799585,0.6104356307744482,
 0.1424801297462358,0.6915430544338931,0.5753847580951957,0.23165503009192645,
 0.2938334440292628,0.6066903996173928,0.4755425880733084,0.15433316906448913,
 0.17722454099594265,0.1708494586097915,0.9289757485374819,0.3267855818578781,
 0.2750456632915843,0.4988531748041871,0.615045051259659,0.20702359477660615,
 0.5982568462405915,0.8257995117625699,0.7274837561667798,0.19940355453068004,
 0.10704884336544795,0.26944839259649933,0.5207052280469382,0.395018811851661,
 0.5847178783328554,0.3266097592218614,0.31137813050198804,0.7129198670630783,
 0.03187340933103289,0.03304787537307041,0.5958932736164304,
 0.8609544182689841,0.5544828511742789,0.4835100334447636,0.8529662235002655,
 0.03055449014454037,0.52356796793556,0.15799908053030032,0.5889467268523549,
 0.11836643219327725,0.08724462378240705,0.9402673833747328,
 0.9435754969279544,0.8697116164410819,0.9281827188665344,0.13333929127861754,
 0.005560152074965408,0.9734545206411925,0.146397858053955,0.8368626901409854,
 0.9541266046256048,0.463974437347799,0.7411052625021572,0.6408961241108512,
 0.9955748685713082,0.879168268382616,0.4668303248662259,0.2087437348556075,
 0.15435900868101604,0.5614223941279683,0.6712621831678016,0.3097549043963448,
 0.06904726971273589,0.8579767305645223,0.5565202238569016,0.5567052306793472,
 0.9383081930485478,0.9152151436592013,0.2584974966982271,0.7646156022152844,
 0.41644025367804416,0.05083244290346278,0.1906100616080567,
 0.9129000113702334,0.22277706304498335,0.35590378256170707,
 0.15903400986221505,0.3424509535042608,0.10074709145730976,
 0.03734414638529504,0.5570726633569514,0.6170498852038055,
 0.22317054737620057,0.7618235919166156,0.8919392823734722,
 0.45011187338577463,0.4580176391347093,0.5212505815606825,0.6839521974343102,
 0.5537606351115026,0.2542077530421657,0.451911032765944,0.9746333496496915,
 0.5455273704693854,0.05605744228942011,0.6578379471133196,0.9105331715646623,
 0.5873808324388231,0.3449636066967853,0.3225390770116915,0.36735291694142025,
 0.755944963315182,0.873332458031196,0.32018951019946384,0.15428988815089506,
 0.8588221644292792,0.9042949749199143,0.1754980480418079,0.6312182489956232,
 0.8743314972206655,0.40670323024432353,0.6226643246527939,
 0.011999214001177538,0.7424279016191306,0.08250556790393326,
 0.973709138470607,0.7421070119410933,0.4389979511556428,0.8729702540520115,
 0.6295963738827998,0.29030288883083444,0.873784567306691,0.01459466081126215,
 0.2754865410023364,0.34512168783300945,0.14520196261128238,
 0.7368094242416955,0.6037008197963132,0.7285206221878898,0.7398980001133476,
 0.13244229376130345,0.29259576473116966,0.4208326419406617,
 0.5221585436434578,0.3491015810049316,0.658116854709612,0.29743947090574596,
 0.974896428217498,0.11184181559685613,0.6731844133286906,0.32742759340113137,
 0.8691413301359,0.8770173848821238,0.07310756569399113,0.9693972561511799,
 0.5901395475858483,0.6696946663027094,0.5253101234214395,0.9258445744274314,
 0.6352185524688471,0.7013586609232826,0.9525587134813819,0.359422783614881,
 0.3649036627547775,0.8717615444540022,0.6986881277457009,0.5457022469438504,
 0.019141178352271027,0.03931536152514714,0.9986882737656031,
 0.584349898102893,0.0653825269047481,0.1378065899486055,0.3992354846172066,
 0.40977584598157124,0.39781139291497114,0.48258857803450217,
 0.8870967291696564,0.20786745619323788,0.41145313923905946,
 0.19623377551582932,0.29418556485266323,0.8831289189450207,
 0.8426052993121559,0.9858781948713461,0.8881403211330731,0.3925568152967436,
 0.7546947528137378,0.4569416191037776,0.18742420631224355,0.9315726616382901,
 0.09153170832697877,0.9501285969648847,0.06071688530471342,
 0.2946707296951202,0.6580917698936464,0.1559107804172135,0.9280585421001847,
 0.32381725035585285,0.522046972992396,0.32405915913190264,0.6612452904902077,
 0.5063911879170733,0.5607222748787705,0.04779578510153082,0.9541974134138415,
 0.9517283938054194,0.7787631397526934,0.5921912849113393,0.768208908233115,
 0.6735699399590014,0.08944837795406213,0.25667175615583016,
 0.13409337536581367,0.3928690920424385,0.7101424580718273,0.7286372401367854,
 0.19395295332320306,0.8980657641662824,0.6377483061645949,0.4373394824551641,
 0.9244318180817194,0.872174560799392,0.9094996327112103,0.7073879513301501,
 0.5373144641942138,0.4355474409176373,0.2430303578982843,0.06629036763445884,
 0.3231353917821095,0.1895461317116519,0.7311700776694798,0.7674480055492239,
 0.42000539779053636,0.6628176419874434,0.14117998705199164,
 0.5145944527704231,0.7475826886195581,0.317600900963559,0.2316172161280694,
 0.7736630304352085,0.9249059530631276,0.09822807242122722,0.9083531675977573,
 0.8320289057565722,0.09363217152899705,0.7335691059180017,
 0.010221708340439895,0.832731280515027,0.16750469046403327,
 0.6814069627518556,0.735190487891736,0.05636272436860423,0.9629321200914414,
 0.6342875639647375,0.7723724323700022,0.8679009655715791,0.34657732955216136,
 0.027386260430864118,0.11311753626522125,0.23317146583298365,
 0.8714437709433577,0.4236407436671401,0.7077519316868223,0.30342310861246424,
 0.14779619802308896,0.7874245327662198,0.9923514606569359,
 0.43642549820144116,0.7564943485968718,0.6704211491961654,
 0.04526048203950739,0.5973843252915348,0.8643849775858466,0.674810462661109,
 0.15496248182853423,0.40441059077665,0.3994578889436058,0.9662849755185368,
 0.8214776002703598,0.23787560610487124,0.43276766100214425,
 0.39790177552674455,0.6962528867371118,0.2578073010491362,0.6581518147417481,
 0.6167739273693523,0.744437971175303,0.6041279805686178,0.19448664587819497,
 0.20743333131916308,0.525120033120039,0.007294558565583142,
 0.025555008895558284,0.6294080991531534,0.8155227984093733,
 0.6874031481612893,0.585311952278415,0.6126757448508118,0.5531958563528178,
 0.7083997992783938,0.8229666600343352,0.5471946173582021,0.7017233050377774,
 0.5841537650072379,0.33548419609998104,0.19018945722916447,
 0.5476451538854745,0.9120541229530784,0.21185219891467522,
 0.11666331907934069,0.5905299709046359,0.2146146162694489,0.8430429138485833,
 0.46417834073703723,0.5104290602494238,0.12010951003938042,
 0.7096352070023048,0.18377349776681962,0.5616583124094742,0.3794352435475564,
 0.005857018311630768,0.550959036963317,0.47490177093047237,
 0.39990637311790844,0.279022292597615,0.19593074887330686,0.633815843642146,
 0.8272642586744559,0.7990102160939684,0.9129549024077767,
 0.011840294877475177,0.3898070423158315,0.6826953528670373,
 0.23891791873450596,0.42680362968875096,0.497874108251779,
 0.37557118063556894,0.4107835809758481,0.6714288094594287,0.6762026801392523,
 0.570686510101636,0.9561063978478902,0.5476334253865377,0.24696142522822462,
 0.4880251619888044,0.8865743599076825,0.5572591176312925,0.28084619976348735,
 0.7735988234206416,0.9387796464305485,0.5784778998464055,0.02486938784174586,
 0.44981128024313854,0.5091580294004903,0.776559110178014,0.28250932908171833,
 0.7528542590739092,0.31355396936908764,0.2774281367636765,0.7718731778271937,
 0.8577665028515087,0.6438710769288184,0.3121513796263481,0.20778518523496514,
 0.618929004776477,0.8301831863582001,0.8326204106521666,0.5806832392450421,
 0.8119487940362398,0.4183549072963687,0.41700287449461537,0.4672744052592668,
 0.6848993630164806,0.08078107840664595,0.9238930538927458,
 0.21541065494392997,0.4280112589811721,0.4214292024695637,0.7053409242833178,
 0.10354668139745937,0.7566727357386319,0.8129405780653329,
 0.21259562876630633,0.9505918064268742,0.03531086950141937,
 0.8097785795637477,0.3422624298380642,0.4250272073886159,0.40989537219718053,
 0.24829376594657904,0.4752659450096498,0.7237347306239825,0.6738323540994846,
 0.7229976578438095,0.31855210701960435,0.5875888607977617,0.8556864911070523,
 0.9370983183325922,0.9371289631830945,0.5793279361709002,
 0.046491620627865826,0.06449981369869595,0.5569617231138408,
 0.5427912993015231,0.6626739970754925,0.44567543317922775,0.8832141514789094,
 0.004861899085378374,0.1542347915579444,0.9763097660321336,
 0.47579250806683593,0.3609479611184837,0.9788891223730661,0.2909690023232885,
 0.9132877216134644,0.9624572910878376,0.8018762795543214,0.1509140041406919,
 0.11049874106131241,0.5858923789671013,0.8963842639747397,0.9886435469600577,
 0.5056596482718291,0.2078057169685753,0.6562887880713074,0.3718169245295726,
 0.7057117173308302,0.024721505570047708,0.06602683887911776,
 0.9941629379335202,0.9984315112913931,0.5148088840291097,0.8606408014575267,
 0.7380718069933913,0.726889628698288,0.9597059606474092,0.4043861523655454,
 0.4770115619026871,0.837218093501997,0.9533029905263861,0.9911014827594283,
 0.43300942030485956,0.9421568386039092,0.7253276107007193,0.8783087692052391,
 0.5584631204155139,0.9323882591104646,0.16356977160116937,
 0.022367005000877516,0.6441439692071227,0.7719289970752117,
 0.4771430481839576,0.17842271404377374,0.3351790972954698,0.7512380488808436,
 0.21995645487037163,0.8988323697224263,0.2845900683687562,0.2950805323815451,
 0.3731005384383521,0.44121603545353416,0.9559571895773665,0.6204116102615247,
 0.9118035892001782,0.03246511160848253,0.3856046034840076,
 0.47136060049801076,0.011515978691047701,0.5714292989357912,
 0.6786851013799406,0.6694101273824328,0.3022253605865033,0.8554690052333053,
 0.024434936014574138,0.1612038605626528,0.532198884858534,
 0.32722636977657504,0.5850205940174961,0.22949763848221727,
 0.3462551782405283,0.66981814630358,0.4106666122804158,0.586863906817018,
 0.14234056769655123,0.7907981659056126,0.6977391287064989,
 0.20397865297641016,0.2830531562756038,0.2536802031293741,
 0.35186217025425315,0.39480443362510376,0.7174504122744708,
 0.6087902656819154,0.7318575450096101,0.6486928779493956,0.505018294813623,
 0.2674192622406264,0.6150053885762997,0.04482686924353785,0.2799520573293466,
 0.9835353438118678,0.24479872599071584,0.031076967018676016,
 0.8900088816059869,0.6992058984707571,0.2070979172487899,0.15491621633051378,
 0.8281549260250227,0.3621113197870497,0.8469703833018511,0.5706498743615567,
 0.6723300599179484,0.13088143676870134,0.21079006776697762,
 0.0285974460276468,0.08833820885440069,0.13615562557432237,
 0.8726934884206282,0.02731073350804125,0.08445184345485235,
 0.4080677692306607,0.5190185261798042,0.043587262177143926,
 0.6291745182625644,0.20153225797396845,0.271029871760199,0.10720292154024025,
 0.6552625521350531,0.9750055995579848,0.3915009307555043,0.5394062320018603,
 0.08483816905862107,0.7708197472766627,0.951621648756821,0.5380176865869526,
 0.4096974959741121,0.41255495656997443,0.31891783707004995,0.776887384877033,
 0.35556447373507505,0.6068648419353688,0.6524378305273053,0.6912140614327289,
 0.7318698460625763,0.7626944876573564,0.7272678846130552,0.576241312062988,
 0.7507617908403247,0.7619061286408916,0.5940084266852617,0.12685084089624366,
 0.02552214496246208,0.208087787212905,0.7217124539390489,0.9956188062606788,
 0.7995904263598983,0.6714627037357951,0.5529447872326398,0.19140188904435518,
 0.6982133515430932,0.800789074002106,0.1823117359108597,0.5239677946351229,
 0.9542904437526434,0.5177978611940595,0.230511391587774,0.02272335727035113,
 0.9184963867730628,0.5215248694351066,0.9647619443041284,0.8260315754033543,
 0.20279466430961213,0.1814685071314569,0.6526750916115808,0.4297853431227965,
 0.7355574623428297,0.30869480077920897,0.059826533457059905,
 0.477661899556723,0.9600235646581781,0.7963957318270906,0.20114102708221648,
 0.9672546207249544,0.7771552618811707,0.5732523811446615,0.06933222491465119,
 0.4687587410072256,0.537167230218759,0.7911143364359309,0.6998378900608094,
 0.943957421155601,0.8389752413497464,0.21269984504456263,0.5372766214355782,
 0.810467090789357,0.490959436395491,0.183813329281049,0.3731033360354121,
 0.6469903199430945,0.7845163511337929,0.9985115898295822,0.400941470785924,
 0.8158764786141579,0.9797020378524365,0.30867943703972867,0.8899602642247222,
 0.3191441451488862,0.31456880683300137,0.438261524053531,0.6586693363401839,
 0.9031354443241691,0.530910237631071,0.6072393624561501,0.3043804949623259,
 0.2850545091061778,0.7515965849493115,0.7803414056332179,0.7158393914995694,
 0.8219986067068241,0.028262648059227735,0.8025282209095573,
 0.4345369743655839,0.38220720453168744,0.7049959761050174,0.4804396716783703,
 0.6474004222413857,0.5498070032103234,0.9810176962910157,0.20224562468937046,
 0.9476377721262791,0.47534620709897324,0.5233443205853873,0.9743065127102982,
 0.9098846024616907,0.6263032796877075,0.6462343580896578,0.3585333620997756,
 0.6333738574515116,0.9273634390913834,0.9777920683206108,0.5108213953698761,
 0.08814484920099819,0.28159562759464896,0.0220740171721604,
 0.3278412641601398,0.9291124069817505,0.528571800511713,0.5096848147589453,
 0.12075408067195048,0.8639493217943695,0.07683703755921822,
 0.6650618090578878,0.2460402496547579,0.3063782138113623,0.9576097807101438,
 0.7670943633105507,0.675321105828203,0.8928153673108992,0.7875863373646619,
 0.41281304241158345,0.8988496276890229,0.9021885161621459,0.8293089140135037,
 0.9293771927622998,0.544218993266933,0.4500192909682109,0.7192359312330916,
 0.13030498664376,0.40246379086613193,0.22791761849594805,0.9428102783492817,
 0.9697382008141375,0.5040904315478266,0.7300912400622792,0.13232295355320012,
 0.31098350323539015,0.007172234591945936,0.8274231366642891,
 0.09902412552823847,0.35032370729191165,0.2315671780848204,
 0.9596705317302735,0.5996241695313145,0.6299589169641588,0.38763859124657307,
 0.4896134776933022,0.9321667147483097,0.6887092601640006,0.4327895725142429,
 0.6476429428180149,0.9745922932412665,0.9227665424264057,0.6956501381185903,
 0.5138863330188574,0.7010993851032523,0.37254523035128084,0.7144128651903401,
 0.2672457468496614,0.37813419618006383,0.9115227773742736,0.8956144654133593,
 0.5479956382197132,0.3913226539307155,0.9481212349963678,
 0.006547518565148902,0.6594887564619721,0.15312505193068382,
 0.09731145479147285,0.4478195238142454,0.8751179428089702,0.8023612824132689,
 0.8267757213069563,0.7519933038595724,0.8323877518345695,0.31914398617558226,
 0.7497368540795082,0.8973007437178373,0.8753672928446216,0.7996177473240607,
 0.645587916715626,0.46054494265632684,0.8523135069101945,0.7703512247653006,
 0.014433513331024828,0.40046993770092365,0.012740701348002426,
 0.86070577641745,0.43023028122434726,0.46705909861567163,0.9219427398209017,
 0.32080698995118173,0.029414590237502147,0.022191938249132792,
 0.8391588282782119,0.7582999444481715,0.49873075887229046,0.216793624761637,
 0.9206936777921327,0.04339723232357062,0.8185505532650694,0.9237365329472509,
 0.48983869870216146,0.7388391135982506,0.02784094596587361,
 0.44289950503479414,0.7929221897674379,0.27842494816865504,
 0.4848173651395479,0.5788454599295575,0.5895282130842445,0.7547906264684762,
 0.9985720230313335,0.26950795437217945,0.8115973037998843,0.6270452587096902,
 0.3934705937201922,0.47001760882626675,0.9932154644373054,
 0.02475697339900451,0.34339636681591845,0.7831305248434243,
 0.1766136330069219,0.9393645611402646,0.041519919882001544,
 0.13388819494039084,0.8077133834291816,0.8170396265575551,
 0.36578997272561686,0.2907665236931223,0.032217448152785044,
 0.9201237706443078,0.7954796899978739,0.6817093921917106,0.08919873492032515,
 0.6500069779795559,0.6774644249676935,0.9328216434515021,0.06719088877340229,
 0.19794285045405202,0.4294463395834408,0.19077162303501016,
 0.8861979392223993,0.5286983672314018,0.5597369862550907,0.5792689767944673,
 0.6166944378679581,0.9122701762396699,0.7391066014777375,0.22335231737235972,
 0.8125682081902981,0.4452956334049356,0.6830675106839983,0.6654031905000077,
 0.8895213549840214,0.9304418470558247,0.5872920891182836,0.99418600923815,
 0.3399542937943254,0.9649965221033641,0.8480117624923462,0.6797379439330681,
 0.3597528055649887,0.74241602074721,0.04306022991653313,0.3944722323321971,
 0.43991127636656124,0.7831402465948617,0.7239749905037577,0.9662838022699755,
 0.5956128832667834,0.6080707420537971,0.022473481562195197,
 0.44663598733645504,0.3773562584767669,0.5713482765507664,0.3082121075073835,
 0.9589883069489119,0.4952385587236686,0.00987971952333555,0.3710389242195381,
 0.9722802151316541,0.7410081418225142,0.6786583455658595,0.2800704960779189,
 0.25259318550713483,0.6672731991955648,0.7628139037557737,0.9345969803991423,
 0.0752668936967048,0.8842136619056875,0.28877925422552797,
 0.23277895896329204,0.23751059440048894,0.40404676204247714,
 0.7657123442689879,0.4474086758910111,0.2799397990537884,0.9624785130883177,
 0.931571978985112,0.6674634734312284,0.3980696264907232,0.6825760462191868,
 0.5791397087003186,0.4619854850669689,0.045783779455988216,
 0.44240413398763745,0.44255061815868313,0.5059709732804915,0.36760246478516,
 0.8646173686133009,0.4244417820981654,0.05074552674356525,0.8800052602124869,
 0.5646563278727417,0.31890066215390767,0.04020162307518538,
 0.9420386204013771,0.8576940941211444,0.9567800156960043,
 0.045937510893933164,0.387051331312806,0.38735169017062565,
 0.16659516975330801,0.39820986332940633,0.1450483265100706,
 0.7500714782357899,0.5170150811992794,0.060102490552465304,
 0.15957632265867572,0.7285475677866543,0.15876402146485202,
 0.4755296318330392,0.9520120585632511,0.8290977765784157,0.560393478679958,
 0.7087208748837736,0.6199113605689347,0.28297835226681833,0.6994926523232285,
 0.5467746649018382,0.5727554692138671,0.8653904914861813,0.243734172712188,
 0.9703795101817642,0.4621377477218308,0.9495455689320251,0.560943848816983,
 0.9150561256974499,0.4199788138255016,0.022718185052764128,
 0.39679632073686877,0.057144357337687746,0.2589334398294213,
 0.7219795286827375,0.36563271321115975,0.26228510471120337,
 0.8719309197655949,0.13599636212801047,0.8996423774567044,0.9924052452905004,
 0.8020783653520738,0.3687989705283994,0.10376931498052322,0.247841349325035,
 0.9730515698683824,0.4924788588602711,0.42923305154794467,0.5831985616742489,
 0.9962008135081465,0.7894348209015023,0.9682545355929756,0.4541855924516357,
 0.650591089118125,0.0834360259481105,0.7919111724448656,0.6232036313175613,
 0.9575892932331117,0.030633163472497338,0.8408808476371016,
 0.4693040207578012,0.2499506848841868,0.2603509124277513,0.5097493960741329,
 0.7827556857771927,0.9906022136022854,0.5066233971436592,0.4983654391511163,
 0.4805879847654416,0.7146741880823062,0.8978520478053146,0.06749132917629086,
 0.873063003289646,0.8993881044068873,0.0912534581606983,0.45234187378057056,
 0.9199595935014033,0.3071684617995126,0.3764534780584732,0.2126514634079114,
 0.22239816901021314,0.7739066265341373,0.8456212040683498,
 0.25430163695779273,0.8395410743000978,0.8315126728061234,
 0.09400198301957219,0.3355870544673675,0.366798691609183,0.41965027855123926,
 0.1038696323672259,0.5342207324092972,0.3120850195951319,0.35325855707216025,
 0.878569983293137,0.35890162970162987,0.9551491687772304,0.2626087516688591,
 0.04625967752689997,0.44834298730664757,0.18518867682387286,
 0.1323708036532627,0.5369142877196438,0.7640824179078056,0.6805008376734414,
 0.06627099602063413,0.022891432528278965,0.0389033032142061,
 0.581518992784915,0.651435925299836,0.8904718346725041,0.5394698840179426,
 0.5998440973912367,0.8640549575669498,0.45874400731718246,
 0.18568106849413757,0.19207423018241043,0.6509005351977217,
 0.6603955765994257,0.4718382817111323,0.8853038964309069,0.2865366215109242,
 0.25439343239180967,0.21364720636705825,0.43109583501541826,
 0.9008155153623942,0.520486004422374,0.32578010838718985,0.4704031143144055,
 0.5935004076938646,0.9212222304452223,0.07513434549001963,0.775412057314907,
 0.41452965786115326,0.5976395514568436,0.567847505744746,0.7296234579407473,
 0.020256823087519082,0.6182460944905681,0.8978334313037937,0.849604770508324,
 0.5194675940651001,0.8044702051536725,0.40777906752750126,0.8023696575494531,
 0.8158340455265646,0.7810355746723083,0.6696018905150778,0.8847976703760052,
 0.9102301149313793,0.3220183348824771,0.6863360048006784,0.8505329728344371,
 0.709756108848772,0.7675170607642519,0.1186106920488279,0.4126875430764223,
 0.014715746299793442,0.21201030953894417,0.9433968345489687,
 0.7835132232210724,0.3599505766048603,0.5315504276069211,0.848240628208143,
 0.3173291508101317,0.5601483599907594,0.013354167638011116,
 0.6447817205836164,0.0463828658934069,0.2225805075965095,0.1409624866234378,
 0.3980876279387461,0.27821724911388557,0.6960077185056399,0.7131187110212694,
 0.7622931514014858,0.588521745808363,0.8585398338761385,0.5707399520412979,
 0.28835873408982116,0.3516555891519513,0.20552920499462324,
 0.5573749317880103,0.5804297814395121,0.8567821213160695,0.9368440315134383,
 0.6672325720562827,0.7208924733838142,0.3968300429555802,0.9560045636440458,
 0.5413922920738314,0.09006946027252893,0.9600309794964704,0.8852748818251822,
 0.31564568183379305,0.3528376436903966,0.6529975149163623,
 0.004560582430103488,0.6295223839431707,0.7852869478985115,
 0.4441743601378899,0.9944056092861623,0.3316995470722277,0.4837319672473682,
 0.013520109140146008,0.0943949412925914,0.8458990019519815,
 0.34157367938775207,0.8622795974516576,0.2588742427012569,0.8391194141190216,
 0.7272113776016984,0.5997398773190368,0.9073942689011363,0.6750785657734991,
 0.6832794585494621,0.44050473628818176,0.5797373713156573,
 0.27631231002398615,0.4066954304078353,0.8242262577803878,
 0.25949037408742526,0.8796023153425959,0.8674502334904228,0.9442115365667524,
 0.021214171546401928,0.949707482078379,0.8089607912202722,0.6847733033614657,
 0.4887877277138125,0.5922205257480544,0.41205244475516845,0.5852083556249983,
 0.8138368455402756,0.3927395240030169,0.1930818303666405,0.20429740185127676,
 0.020810493720861967,0.016692321652261866,0.7229618833378529,
 0.6128449337579578,0.1210398858757622,0.3601640116599024,0.5865775074829029,
 0.7041280956668318,0.09309676676462264,0.38307159786647293,
 0.6083871809536727,0.1527541097303049,0.16223401846631402,0.0822846924442372,
 0.20929369154719812,0.2029150711192902,0.18641940683170566,
 0.04399202605925223,0.7916863514378605,0.0935044874941453,0.2875337418620043,
 0.5577669491928924,0.6953381910978084,0.17334177038947907,
 0.49525703668064736,0.4357321863206265,0.6481319554071248,0.8461500589585995,
 0.3521352675331929,0.0869798078044034,0.9489598121786302,0.7855629761869757,
 0.8146406009909496,0.6550082733313077,0.31473451716890233,0.5993734644357404,
 0.8158030043002551,0.5585994201580252,0.58945493251503,0.06267108653420905,
 0.7385391801064874,0.6381836480502346,0.8138874436338092,0.2191355775773446,
 0.6513006403866427,0.4623794801709834,0.6606593762136957,0.07457664257908603,
 0.87900731995825,0.7779332875625736,0.7326931936579981,0.4775524372578148,
 0.309055446020436,0.9420739184128414,0.5513324542716298,0.8685714339964458,
 0.6195285752425852,0.41228641717326253,0.43939948402936,0.40816665816867315,
 0.0040097921451642815,0.19153077884448289,0.860425552299418,
 0.20467690758840584,0.5815914166814583,0.39974557263143695,
 0.08610611381706446,0.46909437439060697,0.5485155778277085,
 0.2911519368051596,0.7205554724859559,0.9263158959705007,0.4265490430245299,
 0.07988087144243639,0.5070148400354201,0.8978456041731232,0.5758164316841615,
 0.9732741015857567,0.842533888146164,0.004936140278011258,0.0791331866001953,
 0.8787690229896151,0.9089695303516944,0.3415630167635616,0.11094211482346972,
 0.1112388705267281,0.7935766300757041,0.6001520201453079,0.08577919792336941,
 0.22463691531901187,0.3803945511551121,0.4582885357512725,0.7157745976264664,
 0.7877692919026885,0.8514605201832083,0.4064524955844915,0.10601806099730471,
 0.7146882022871013,0.9945652606043143,0.8086943308665339,0.22356364632120274,
 0.3995279544400313,0.03244365434136376,0.24044474347020506,
 0.8434150692620874,0.3191134272657119,0.12130911221217278,
 0.059428698239647915,0.22972564142991703,0.08874613328176462,
 0.06615808848274884,0.5565452190436496,0.20541396402731082,
 0.19941181885343806,0.8260059766518116,0.21679253485545624,0.308630528791739,
 0.0830470064602905,0.3995227399362924,0.21992900945077976,0.6310865093788667,
 0.18509808249485804,0.43603968545828475,0.5063323099098751,
 0.014420374119668344,0.966064310995326,0.47228800403550264,
 0.8729418273450024,0.23023425443603895,0.9728550645284649,
 0.35125282935707036,0.5407816012512701,0.939714291513553,0.4850345896533289,
 0.006608461275273481,0.25496237562538404,0.959473780385815,
 0.7080858373190653,0.3283286611367855,0.19947493840866315,0.7479430217362637,
 0.6226117788950385,0.39715126065683504,0.9302472533649796,0.9781405399012312,
 0.5170467744744158,0.5457429216973037,0.8763220988476951,0.36977229750598717,
 0.9051062435535883,0.7505332544500328,0.18331314063317394,0.9952197673822337,
 0.9000294424681035,0.9424056251397384,0.9965065228626768,0.06486867762298665,
 0.12733145847473926,0.1819704946792282,0.7340299772008937,0.8824998167095147,
 0.1687691239179001,0.38652031813418686,0.22605177486727523,
 0.5196392983686448,0.19980015862986,0.2781488895297003,0.6600462924175142,
 0.838701476304063,0.11033663997809706,0.2617496230566282,0.8664728047844779,
 0.6383729507078784,0.16546953474698778,0.028193989430345456,
 0.018080650737470116,0.005328041888073121,0.8705879298420224,
 0.3578893308832699,0.20940902697042307,0.4628821496560699,0.9425371966102261,
 0.8068365396340345,0.3088936431207323,0.7441504861715573,0.7393711463168918,
 0.24942883966860574,0.0589770068830936,0.8485807488771462,
 0.22117539060276958,0.24298449857671078,0.007249838644317497,
 0.3447361415138058,0.928974184072404,0.09489652623949452,0.62383096704748,
 0.21187803924591375,0.7213764870878734,0.4481692981805553,0.9544439772957335,
 0.001121568847804122,0.33314116907995894,0.3839091065211838,
 0.04047813037027015,0.1448776900379477,0.05959505955799271,
 0.6891824365186541,0.7675187310328851,0.27524552722987605,
 0.30942115979091955,0.015861084619534394,0.3947220064475433,
 0.1000978509524374,0.13408665410331255,0.6826697094710097,0.710480778491989,
 0.040285856090046646,0.35547141314292763,0.4832316849837529,
 0.7172195169011315,0.6736067453549508,0.5470619720234207,0.4807349981889827,
 0.490300861847608,0.9096611655513568,0.28304197723218816,0.9067911099652353,
 0.6914046548001376,0.2787874363234947,0.12338132949554248,0.6392280399811018,
 0.41607770159611646,0.455734010645809,0.37959485237299806,0.7873172475496759,
 0.7620745279422342,0.22401918965648515,0.7637473393663632,0.7621590263404734,
 0.795746639492857,0.76807075763703,0.7399023553460378,0.019114232526549313,
 0.7313764055886535,0.5724759064247015,0.03973341462408608,0.3070336953705808,
 0.1564103545000053,0.9403341654738316,0.8408224606257769,0.9190212907629426,
 0.5669758899016599,0.5821191480739897,0.4990238907819674,0.3716549630574214,
 0.9122355990970268,0.11831429658663051,0.20636943834572097,
 0.0130897612502473,0.2009907772827727,0.8609177062970169,0.6157019626814514,
 0.04335782290965473,0.1466582048759849,0.948500396197109,0.9808796657445122,
 0.18291909254331462,0.6670219306662377,0.5066595182414364,0.9032128475633131,
 0.016009627690816597,0.8248756833417302,0.633112952627052,
 0.11983487174429452,0.17323345219499842,0.9592954955666355,
 0.4989692646956355,0.8569553027285505,0.8688974525108274,0.187852215753869,
 0.2703375799905188,0.48547817711551,0.25812412626037184,0.8505210544486246,
 0.30985474788744705,0.25873009141384307,0.5986971822953139,
 0.19090274222418702,0.15890469405794705,0.18110461060580052,
 0.6394817062996591,0.644752687093569,0.9160297835619292,0.7199551096863114,
 0.5993179468285544,0.8320253111355363,0.2182373927115615,0.9874108341053789,
 0.8774727068589394,0.2301884148318133,0.9878199468477821,0.05127146090490253,
 0.6709897168717366,0.17793864448641106,0.14111842157264676,
 0.8834082535884721,0.37641006765323526,0.9377782835791566,0.8157674041333501,
 0.6039619663583647,0.9017649989749736,0.5058094600643466,0.433486373024613,
 0.04660142000400491,0.906925422443094,0.6216112443425986,0.12182288450684331,
 0.8180685161623794,0.7541385921318619,0.7985210450803591,0.39529276621515286,
 0.03266902972246333,0.7887431453538917,0.942485635727081,0.6265566719984371,
 0.4523625577449446,0.45691977042575505,0.8639642778564829,
 0.35579260652360656,0.18317705195874268,0.3462586739973793,
 0.7332734962410875,0.2756387783181543,0.5942967188171373,0.3640126646429931,
 0.10409487421103691,0.2298692965904915,0.3894780353031815,0.8852616253525416,
 0.5095899529050529,0.522152556781263,0.22345809986137866,0.8450962193137396,
 0.9438069870749543,0.07976789303528165,0.8389226741543582,0.8955850993605474,
 0.002718592473034853,0.4572240302419657,0.11163111237593515,
 0.4053463433297845,0.7952545192407288,0.6410844696307034,0.5778543040569519,
 0.6231637816398599,0.06308684894695615,0.15046030972095037,
 0.20465858477933208,0.2640325095318016,0.7187947393863945,0.1532129297953735,
 0.8591241127792546,0.7480727768557259,0.6012252267628435,0.7816897383922179,
 0.6376431127203561,0.7163472339404494,0.02052349613292237,
 0.03474353787036211,0.15636331626210043,0.3132446531713994,
 0.2878105374886236,0.9265820227785161,0.4387991943577835,0.8510527573090486,
 0.07212897842140553,0.8640927183347882,0.09204130981626246,
 0.9240881309909798,0.62268880201814,0.806419133806228,0.08321224885757983,
 0.6644458349082636,0.005762267625216477,0.8104616411459912,
 0.8357889643444785,0.4521867037195316,0.07935014757014014,0.7595960637770461,
 0.46099513896304156,0.40066419753201665,0.4555934830152931,
 0.20317692224029982,0.8920847432209801,0.2251054392308498,0.9370092619320095,
 0.03360979701204991,0.3049967950481951,0.7687736963337661,0.6978929888066822,
 0.0756077021099294,0.6717875753715532,0.9344115749363119,0.942321475127366,
 0.3970661819074446,0.3942325582937335,0.5712995008155124,0.21049878154629875,
 0.017179166670998502,0.010872689424246906,0.6653773734702749,
 0.13504714622128633,0.36008835193635147,0.37570455130845315,
 0.9514038079931174,0.21148839907099826,0.6286007401061128,0.1430433856902682,
 0.6129608828049761,0.3573548566722212,0.451570588603904,0.4849676408950979,
 0.8873337178809908,0.4523743718267812,0.3103538016395603,0.23716061356196838,
 0.7570424564753836,0.17902362161746666,0.9693952498517546,0.793989639112309,
 0.004505333304057335,0.5943523148060439,0.10120337002944779,
 0.8433443090636297,0.3296687709141961,0.14916179118282646,0.8938311913870378,
 0.3670383412925977,0.8077448590961435,0.2753157046785275,
 0.007388217114153384,0.7982661966840635,0.5958484956936781,
 0.4327704854265544,0.36853112660413934,0.31544332685174004,
 0.7199431317725862,0.282357298480447,0.11320249105257052,0.04854732358872427,
 0.1528640957329157,0.5643364569823768,0.15245606692769065,0.8408927951736822,
 0.3532166019079753,0.5174159649297032,0.6274766313296525,0.8495437577144738,
 0.7824481103880645,0.7475050243732899,0.7115292012427585,
 0.034149629371594936,0.7526935594816978,0.0063536024756791765,
 0.10689754299707854,0.6611276608753324,0.194997491270672,0.4399414626588627,
 0.9995145413512486,0.9193886700090861,0.9626296591629555,0.659915067615749,
 0.622956423286126,0.9783232548949703,0.06542528805931336,0.8731012559593693,
 0.3815353130224364,0.79049957043477,0.13311011550389007,0.7372236440777331,
 0.2504765611821549,0.4407333405308772,0.06899502846303118,0.7618443064074658,
 0.8891637588583563,0.34577842655327473,0.773169797470948,0.8143351962923278,
 0.13682358731399724,0.4294526874089706,0.7585302090491395,0.4768378736256398,
 0.1399397156558938,0.7018292885365558,0.056184145260196994,0.346880033188402,
 0.15216764792534399,0.750769389853009,0.5881208437350113,0.26834594077462137,
 0.21824222522997427,0.6027765113302779,0.3595105784956868,0.659095642791776,
 0.05864431936914272,0.5367585195891312,0.01446040713070007,
 0.8231029544340559,0.5520794717621335,0.6600667866343645,0.7635990281349372,
 0.6210231140800796,0.6815373848795357,0.5800850010933563,0.9530599100358204,
 0.5712990725488613,0.24476540901246435,0.6608311950815301,0.7743987728719096,
 0.19187582048582819,0.4650650437912287,0.6096616607493972,0.6853816303791151,
 0.886694878078692,0.30433891941880997,0.26160586208239034,
 0.24634213270665373,0.7024987315350782,0.6401925766117708,
 0.45655865643369375,0.13091355495331147,0.2185044760956436,
 0.3184606427860036,0.9046842346271188,0.7077970836595733,0.3052095077534055,
 0.18611835455717063,0.580805418189763,0.6883111382678317,0.08734146064363701,
 0.7947845265645681,0.39729150471675967,0.3713262251580809,0.6204953962008358,
 0.06659407790682403,0.22112816642004973,0.3991940783611192,
 0.024483489056722596,0.26349844241168574,0.8740849267468623,
 0.12804055760608268,0.39617059496275475,0.2036885291658148,
 0.2868384887831943,0.9803937339461134,0.2639848847653373,0.24193217590462346,
 0.0703573548770523,0.48251288451470886,0.13161315128258033,
 0.8419300925981292,0.8684476711297455,0.6631967662815224,0.47982912985535364,
 0.7658050635293993,0.7679358500227011,0.8297183047669158,0.8315123059209873,
 0.49320642800314785,0.5208346074733756,0.7943219385832287,0.639125539108212,
 0.3107879480298612,0.9673538858359256,0.32173594254322646,
 0.12044663433736869,0.7493372179598606,0.6562535327134816,0.9565596136266249,
 0.3152232241165759,0.4585508668357403,0.5383530383751188,0.1829087421549297,
 0.060008474805593015,0.005286254944348534,0.38980182127051544,
 0.7838505328560649,0.5574715167718682,0.7719664026622608,0.8188056871572282,
 0.48550713539560597,0.29834373018483484,0.15956716276430527,
 0.23022448700445886,0.8107978709478617,0.7167036040627995,0.8883635102430767,
 0.228193268502353,0.4020176898420428,0.5135218747047241,0.6475038847000276,
 0.44918810305744683,0.31147671846205893,0.16140862557755697,
 0.3261484925311542,0.4331883710098301,0.36032237250786003,0.5778018429317578,
 0.1718267941128866,0.053218063025652906,0.4564735800925289,
 0.12859230863274962,0.11913922103665653,0.308640599255927,0.4268035895614418,
 0.04067094505137181,0.8153705983441134,0.11018856173060132,
 0.6628214144319526,0.5970970546821216,0.19419252158761435,0.701073318033242,
 0.9751553888027975,0.7331959137312034,0.2436485917059693,0.3044868638112175,
 0.98251216612889,0.746628035295423,0.05635138110087756,0.11860296188149388,
 0.9411906346962905,0.08112221540206,0.7746991252024642,0.9809969330405333,
 0.5170782116906515,0.6179362929116042,0.748195510695535,0.5314520531317165,
 0.7903652784063646,0.6488057382101453,0.27914434094848817,0.8357921826677595,
 0.7197538394190726,0.5140071680393762,0.11192639443287655,0.4510100681877077,
 0.2881415930839718,0.7798306609126294,0.7081067028285708,0.19735947867309078,
 0.9178369297164202,0.543125571925174,0.5233592931920716,0.5540098992023764,
 0.5768082112810065,0.14969954792108564,0.8918916391858814,0.6026116731881181,
 0.8083296073182206,0.20095345857959712,0.804336931707656,0.9545020880968907,
 0.7935616146481766,0.8258588839339936,0.0882496847835349,0.21084423107816397,
 0.40657351169796785,0.008264605683546145,0.7678583387888234,
 0.34501977794660343,0.31605441518386923,0.31955148641193876,
 0.019025798577009145,0.8103372680269584,0.7070168932494969,
 0.8699115356349942,0.5910018230251116,0.8295284710826104,0.4564696605022227,
 0.8452532069767981,0.07592852341674172,0.5192890857721929,
 0.25912699012357776,0.33399616456582615,0.05937932399931167,
 0.22898407716372726,0.6607803812412386,0.8120097212065965,
 0.12293733950458696,0.8599358871615612,0.8313433077068001,0.8349327425757377,
 0.4604142212970769,0.35335906733118416,0.8322243300714086,0.5320341963476167,
 0.07829182595598161,0.37967323293036004,0.7580382492727764,
 0.8140298097689891,0.5691408513647132,0.6618902695786187,0.44775344174818876,
 0.4874111340704026,0.1169345124763379,0.43562291827900834,
 0.0018113933939500892,0.7183700624162035,0.5269027782390471,
 0.07075872035592479,0.9115804131630358,0.8308274098212615,0.9553300112053488,
 0.6017253741082794,0.589822286359332,0.979214782693011,0.5584929504581444,
 0.8869165305153972,0.34154933022084266,0.4771042499238224,
 0.22979496211418293,0.8609002348411209,0.29605666885586457,
 0.4645083958301959,0.05279330503470969,0.7425582973259288,0.8123746213392423,
 0.344021161353935,0.9981916353240701,0.11597580734881441,0.2551896832685192,
 0.8995116340005818,0.016576552764594132,0.138807100620567,
 0.15094689299335218,0.2504946732308653,0.5629852879495489,
 0.34739581692652743,0.34458816531882375,0.5525699692642584,
 0.7178277010290434,0.20045338145892466,0.7877288735253805,
 0.17172273480026645,0.4740379153426242,0.623260234791935,0.5849187094615307,
 0.7995964972143736,0.5244685360205326,0.9186609228145475,0.3187546522055329,
 0.48354857858996414,0.27410170662088007,0.9770088078352815,
 0.5907544247574268,0.13843240773750465,0.679425713255154,0.4706662437775506,
 0.710814645456564,0.8433044842962295,0.7504430461578373,0.49932945908106474,
 0.15380634449656982,0.41469372266759896,0.8388714019610305,
 0.4099064930736953,0.8865146914172827,0.21045382742497898,0.4200155942943081,
 0.6261113741163573,0.9437124064592521,0.2192884102045387,0.410120984002295,
 0.8775934538887387,0.019409408454386656,0.32158399280827876,
 0.5950115874653468,0.4065571512881677,0.5563406234979941,0.3929131527136258,
 0.4145979584458812,0.09405114664387093,0.2091484510460515,0.4029913030371691,
 0.24989716107383753,0.7965665195344214,0.6588487712958551,0.8214432845723446,
 0.466178403712147,0.04603823953414521,0.5645045079907725,0.4401682849016424,
 0.8578927090034241,0.4882615994404864,0.5128453694324363,0.39359519456688774,
 0.2086418662656373,0.8815625468864136,0.18318093894340226,0.9675642500280978,
 0.3331218760028506,0.6269034785863271,0.08109507190770437,0.8707259235456754,
 0.7756364072986475,0.4732245905284005,0.443492634414978,0.5428109345589521,
 0.7591499560921838,0.25313415241159376,0.5256452817020417,0.761948508492587,
 0.7072678659781285,0.5202355097561908,0.5552855031513192,0.3392208716565627,
 0.05612609947206526,0.45977155030845374,0.080726017640249,0.2762307931427861,
 0.2904651797130189,0.41665901428605945,0.7300076891686421,0.9878251992680099,
 0.18888952456951436,0.23392789803763425,0.38251564716993547,
 0.369121855526941,0.424145271446175,0.12334983011306311,0.8342632336819484,
 0.9526558287723681,0.13248870878319174,0.9404650347685591,0.9938175980138866,
 0.0054064116855050415,0.3358801987083142,0.41735001588369836,
 0.0011193077220349457,0.4124576666784572,0.8015320031074651,
 0.26434662494303274,0.48183814014178383,0.7029787503605176,
 0.5903742329049046,0.24124569374601768,0.6236801279262081,0.5238285120786397,
 0.09199174178539482,0.3784196632948207,0.5367328797633169,0.7167580091506268,
 0.26083280091533956,0.6458036330997868,0.1668425955175611,0.8969744889955173,
 0.048612640476933366,0.09801517801305792,0.7597210461320199,
 0.34101904100732616,0.959850230259564,0.7320937271862674,0.12440747006271136,
 0.8679515542660132,0.7241812012840765,0.10879253270893585,0.1266353349274647,
 0.0074200712136864855,0.9216097966211346,0.38761463524230955,
 0.19796228787805292,0.4106054180975405,0.2548409997659977,0.5776085104716695,
 0.23087098155858476,0.09991819175756722,0.6561888233113634,
 0.6400265438499285,0.5099569606362688,0.6529465823672054,0.707093638792767,
 0.7671379267179039,0.12229848340801197,0.7389403018420374,0.3812142317409253,
 0.775847328027098,0.6615374898161037,0.1499881751428126,0.4494029775207594,
 0.18330716254618462,0.20909955756251808,0.8736848509981299,
 0.005107777974119654,0.24510759054616849,0.926580060004743,
 0.051900053531417245,0.4468000371833809,0.347980745758419,
 0.03410721441177522,0.6267594346929962,0.9165094962871059,0.5988393220771737,
 0.42481437421230894,0.5242385792667226,0.5317927395828066,0.458346707713705,
 0.4591585906824228,0.8933231069383891,0.4517760457151532,0.10373395827485743,
 0.016964756383451296,0.28191270902671395,0.7040121148566263,
 0.11781012807133695,0.3691272739956405,0.04135578368168458,
 0.8466605802694921,0.4754350947876451,0.0764358963886046,0.06867538703522946,
 0.5989470206535632,0.7825860972333835,0.5925319932542856,0.06671338257254478,
 0.5041025054027233,0.27749918573739596,0.056729035842775444,
 0.017470862124667863,0.1494617369117981,0.681530200291243,0.3902582697588489,
 0.39250458820587697,0.8536541032455436,0.9958550499029201,
 0.42990924198578606,0.5676214721931422,0.9366068126212916,
 0.09381440374508421,0.8344856660446403,0.910975651434536,0.22037035069073196,
 0.4712384512264537,0.7558937993569551,0.8441369925758568,0.18627588838828757,
 0.23175764333032467,0.30136934222303735,0.4522839239888139,
 0.3967817130209941,0.14602720201698371,0.44701464276451264,
 0.8995136202544756,0.04188526811949844,0.2334925785206856,
 0.09127619126326458,0.21955177586843033,0.9631592220266182,
 0.7898676747193443,0.8789026511294531,0.7400565665682588,0.12847218292918694,
 0.5144782561303122,0.30506976467111924,0.374290024342793,0.21235134871126138,
 0.21706535217274503,0.6565848981229239,0.5587507767882823,0.5076898577522542,
 0.026315650409129754,0.8613708625104307,0.6633783598744302,0.677308144256019,
 0.9298265573775024,0.77783567777018,0.09613796212204184,0.23712165776452798,
 0.9738849662767024,0.05334348481370188,0.32800714673517983,
 0.23070523215791217,0.9660864455445491,0.9216773662180477,0.6902793805283729,
 0.5670553315953378,0.09017988756302986,0.1809964689121225,
 0.32344795183565245,0.3508056704187624,0.15598328922050408,
 0.1832312682774463,0.7443155489073032,0.18477075605599147,0.4987581267795397,
 0.13095101111456808,0.6007734090642367,0.08194327759591735,
 0.004602402926166427,0.5171178220640092,0.5687642768808682,
 0.00613057134343209,0.9199717534741687,0.15805270209078137,
 0.6541963561649662,0.3730989287691937,0.5438117165270601,0.47526635110866167,
 0.5769930420812968,0.1312043139875173,0.6900899866316679,0.6593488753356905,
 0.9268466750035068,0.4481019082200216,0.5839087028624446,0.20729640935342808,
 0.694266679787807,0.9850003574246776,0.8147573321379364,0.12118358421506936,
 0.6206304867708731,0.051739940427156794,0.8702942901603655,
 0.5419006973636837,0.7405255237450081,0.8146235829589309,0.7734932923607983,
 0.2481643875367222,0.6265242447192193,0.5721180799639829,0.0846649343190442,
 0.9936154518295701,0.4333936590148546,0.8679942637048226,0.11616940994996305,
 0.33716870801410204,0.9833852233032138,0.15002316154813566,
 0.5973571851686524,0.9041985479975386,0.05027007991893595,0.7259222153434999,
 0.16739752171566502,0.5636220278721743,0.019793997979653533,
 0.19766083137319024,0.8268330900816747,0.9814810524819912,
 0.20009242339675182,0.4762311897415078,0.992869662007829,0.8146110062468777,
 0.8866159093283937,0.31755158652139737,0.07008179634257927,
 0.10498451525360963,0.4265826842981837,0.008076054578019853,
 0.8487075003858429,0.7518004182019933,0.8616211329340253,0.6018086868339785,
 0.9673845796187169,0.31358655260789214,0.4026185469499386,0.3987213481459222,
 0.6444641298160971,0.4865692031525951,0.7973342299944373,0.46223940047981615,
 0.9232232948961764,0.7936761638749695,0.8483147333966473,0.8025884941045865,
 0.9173428082034494,0.04546546285311526,0.3160729964896173,
 0.05258671577536522,0.8179494737933476,0.9554256716455062,
 0.47974975974310397,0.24315440402780297,0.21658106415962686,
 0.22092999932340607,0.45921273713073807,0.34032408859157504,
 0.17929623978909348,0.0074127964475358965,0.3660309374392665,
 0.09007612883905547,0.3460760205143333,0.17317678679687243,
 0.5467557186657768,0.8276712199000983,0.16729752766803074,0.9252132269120773,
 0.7263892704674183,0.1377730252456616,0.6408489952631383,0.25792713396810396,
 0.5602187439025726,0.053155284787645085,0.6888860542866005,
 0.7921947819025796,0.7465809144052677,0.9625921923803146,0.777170869565841,
 0.3092838627365835,0.5801628805520526,0.22659963108827486,0.4398670475191844,
 0.24757744607957333,0.40636421378743803,0.719390202609866,0.5572622814440031,
 0.12839104765783116,0.3599581367124042,0.18544628268712826,
 0.3532508184423895,0.3623234727832375,0.053058323366239724,
 0.9616293137666824,0.8526206577301623,0.032306771646303356,
 0.28592216030868267,0.7733699569343474,0.5185013836416283,0.513876084162003,
 0.6134996159608714,0.7170885429716676,0.23853083316947854,
 0.18379976176827284,0.6228010316093513,0.3642657838877027,0.981775699883117,
 0.6487047231937187,0.7273483864376715,0.9545079376403601,0.08634478944535262,
 0.3502476369430504,0.5301602968535196,0.9366315592554891,0.3571840375296711,
 0.31786060619964995,0.7690524762591727,0.8972348051242358,0.2971072785860047,
 0.7975179768019671,0.007420253741074911,0.7722662283102409,
 0.9105018716398137,0.8177602366145339,0.38304574572564565,
 0.019665973169125328,0.35737984135397594,0.6572757757041462,
 0.6688689035958281,0.807232519120858,0.37392044058657903,0.1533534517182814,
 0.45499232581796845,0.8878786605988587,0.9190814696826893,
 0.44897433169763645,0.28058744684378056,0.20934979236761764,
 0.7059370500276118,0.4662336936108693,0.7910716422122702,0.901365915187405,
 0.8022465744321963,0.920151328531729,0.7830108977969216,0.22182691405206134,
 0.7687937811955532,0.7126498838251552,0.8536007656661246,0.9538846461598133,
 0.4058170273493291,0.5737686988596182,0.7841900686681047,0.6536075419487213,
 0.6179378235182477,0.9167079540495494,0.6576080543542616,0.9770503166630583,
 0.9114329768903752,0.2640510966385273,0.36862050066434027,0.595862985717646,
 0.11737852205389154,0.15768541815566228,0.138537417468527,0.4019442130167823,
 0.32705971681281687,0.5400127313954993,0.6125180717776421,0.7044893203648623,
 0.1044288337185566,0.37634453783150257,0.1985956892856966,0.9818912168754776,
 0.27431912626399857,0.7935749532335244,0.8660669596000935,
 0.38374888872447543,0.22960278418070534,0.9390846660897183,
 0.5589460858660031,0.4718388223825448,0.2121201907066923,0.8002290217225684,
 0.24949714637460474,0.17903173966621244,0.3165142075590296,
 0.9827992833314119,0.508867860906755,0.9123770370531269,0.7880746449958618,
 0.3534421347731972,0.3636006415927403,0.4106270186314652,0.797018971861329,
 0.4274897781368362,0.03044456607298529,0.7136230186729354,0.2821321767710815,
 0.9005213196655909,0.9478719371087055,0.7281518597974764,0.26851169909257433,
 0.11985845981992338,0.0446778453072082,0.3361213988042079,0.7535153820553471,
 0.0521332790530219,0.8184978334712975,0.8072234995436282,0.481146342804178,
 0.4619088662818568,0.40759922199631216,0.227210552171641,
 0.030828369929997046,0.012735843332766672,0.9441386853186756,
 0.8490209188932095,0.9134250611794665,0.9101787254879089,0.5565993599591014,
 0.8331204850644234,0.7173661291039002,0.6548036801290964,0.5303968825099621,
 0.7221449347841222,0.28831184380561303,0.8634159418983698,0.7288940690518059,
 0.9630146754746879,0.5669217494390462,0.19646705017603205,
 0.27189569941708513,0.37242396709366266,0.0016651957213312807,
 0.6993240312099834,0.19479126606446484,0.019538562756461708,
 0.18595197610744318,0.046493266043710624,0.9190774703033399,
 0.623326065638384,0.9042772489278297,0.034340743643252236,0.8867757386453553,
 0.32818580897119587,0.3128237252915371,0.9528373856871295,
 0.31606041187280365,0.4434994154877162,0.4176519512126943,0.3745008269534176,
 0.5912535254793779,0.268742518826226,0.6293961511157953,0.9338169915028025,
 0.4875051277511444,0.8387824262247509,0.9340451625844013,0.9641666635470063,
 0.37497619290195294,0.896350653415976,0.8534178770957237,0.22556692869938866,
 0.8568074369043752,0.6956333471691565,0.8867764597882455,0.246067464657155,
 0.8827844301536972,0.6063463700351617,0.9254898749276806,0.534774392941882,
 0.6130629763442919,0.012582197613193902,0.6292992141756183,
 0.14478511962014962,0.03848748864961182,0.6233816356467293,
 0.033880455724691894,0.6790483011918947,0.9744250627413051,
 0.29664077975746383,0.5137965602437328,0.8621105483280407,0.6608370340526973,
 0.6745238204812493,0.678151430572279,0.44486017086160834,0.6429452050806579,
 0.3454150586261836,0.922158211796323,0.36502382496164787,
 0.058849146141425335,0.9429518316391237,0.48829777011498265,
 0.16420663200172614,0.6690981546253867,0.8539379896236688,0.4842315422137118,
 0.13520546717267679,0.7106275926413113,0.7641506753404366,
 0.10041212460599691,0.9189344154171765,0.8313717550192303,0.7360826861296347,
 0.6186565362062227,0.4938016929976963,0.3767241533579242,0.95034910673168,
 0.12391430907787648,0.12603629948044026,0.8899334905372827,
 0.9024580839499067,0.7895074844980364,0.1861716725286513,0.32175747473274285,
 0.1997236287608959,0.5547632411070181,0.9214259406786725,0.434108237734697,
 0.49129762706052715,0.19069074923694795,0.028955869446738336,
 0.496827478484806,0.4143396369148328,0.9027695075551238,0.3282920013825972,
 0.6059325431582431,0.8673232831617018,0.3623312759648607,0.4631312425122185,
 0.5196113632962521,0.9265421060907452,0.7287798504597133,0.849031570063854,
 0.18382801510849311,0.7382745750827361,0.03983584923983863,
 0.44754696520293347,0.6299737834002801,0.008926578304438992,
 0.7538481083178723,0.6053181564793744,0.5649270022183677,0.17629525697601567,
 0.07347863551678724,0.958230501219858,0.5480050293477502,0.7843808229797359,
 0.8652800792442621,0.3297318595456773,0.2311743988225613,0.45398601596601806,
 0.2718077856621667,0.9091562275305292,0.9785236107432272,0.2932422986440679,
 0.6754374419665865,0.1685770541658007,0.7630346558305261,0.43425052802466846,
 0.8988729785942094,0.8215805155224224,0.7259291615202204,0.9680813725413845,
 0.7055764940312496,0.6903221129892387,0.10089468474864272,
 0.15676660467564607,0.15553526463537892,0.1081629434096405,
 0.6047913574355184,0.622937426279968,0.030677936286588436,
 0.22667370384221264,0.723869072775675,0.9399612230986948,0.5277840503733592,
 0.25749156249197125,0.6997786173972738,0.8895400298277191,
 0.45959570152624096,0.33195713276974115,0.5844532803317549,
 0.04662236991876301,0.2433421302048724,0.7559826460458495,0.0256572535477263,
 0.14823878669885915,0.5140553301618798,0.691344975817391,0.1464372391661939,
 0.4483522307899518,0.2727306574502968,0.27172050041842133,
 0.38323173649433084,0.6999188642453466,0.3485620999780832,0.6370659407817477,
 0.8198607640190623,0.4898140169961136,0.18835778310627826,
 0.02662370776473777,0.3415182867229196,0.819983854450115,0.9601653334700652,
 0.00944840876924069,0.591847383288538,0.4375115012055999,0.8389450395367278,
 0.4469302095990483,0.9461167303532088,0.28032469814264316,0.3867235523688384,
 0.35474096196172555,0.5969406625141612,0.586908393116982,0.840503946149621,
 0.33422871011004174,0.005166633387228892,0.4424783928578888,
 0.1711218646037196,0.6965450401404981,0.8778144311198512,0.19888504295765208,
 0.2024534408276415,0.4847407455762507,0.579203602804127,0.2672619644490447,
 0.6890930408474093,0.6481483112218044,0.7059297949027794,0.7394559865065138,
 0.3255075779152564,0.008146910636879223,0.2322027979269785,
 0.8159302599804169,0.37730355966277995,0.55966739439323,0.2550764181622399,
 0.13012770588855904,0.6033805109466497,0.26393162855043495,
 0.8946240850209913,0.3408457180415658,0.3769267153172403,0.9639161282030908,
 0.7638405895584753,0.1702509967380721,0.7453261703194058,0.8750273744649548,
 0.5949983859663159,0.8010364469840561,0.7394433996813138,0.3747721796334028,
 0.549686292854781,0.9809108744602804,0.13878207715502655,0.9976642838026349,
 0.5006685767751917,0.5085138511442098,0.06375229440476116,
 0.11172226798710405,0.590572490376356,0.9923990678649312,0.5442336486858935,
 0.5171010876431474,0.36398243667186614,0.23826453495753075,
 0.9447579673358825,0.4718019884268071,0.07190981977839428,
 0.19618187575407198,0.0771244083700453,0.014026352638598305,
 0.6645737043775962,0.020817754376278086,0.8614124292609864,
 0.09007420328257498,0.06543873854284066,0.0016202766628330423,
 0.7934426851999726,0.26201464212261905,0.653869303546899,0.6740530260551206,
 0.8100988850604266,0.5609709724863642,0.003331507143850443,
 0.3236537377047176,0.8568151183636851,0.3867377537820109,0.46579055237715206,
 0.449416122883429,0.48602994476044326,0.9578046026152184,0.3203833599774295,
 0.02441723411756469,0.8794290758390465,0.634516940832326,0.18339121680791926,
 0.5714929832504196,0.9372292085920576,0.3008783598299054,0.0709300621366139,
 0.5373534558650187,0.2982679491600545,0.29513019206665914,0.7346227768420643,
 0.703444520293634,0.13154942118617463,0.35846140830984174,0.7633679087360186,
 0.6811511008768902,0.9618085977572239,0.4201057530940233,0.9180541664527807,
 0.17360219162413826,0.5387817025937899,0.9399990802552387,0.25050679593349,
 0.5350954049397847,0.32886815046957507,0.2256758360100174,
 0.09187506378879706,0.26846480351990754,0.5196876598322836,
 0.06669383648788219,0.7836908499714907,0.30512134970047544,
 0.6661591411847182,0.26692667729309083,0.6417657477190586,
 0.020628174419777046,0.16785907130452404,0.32592586147612335,
 0.8278312588321479,0.0715121633410869,0.33801991303534373,0.7830848155546785,
 0.8427103405223326,0.4326774686301246,0.3046519409727889,0.6426454648982751,
 0.8706482689898748,0.46149362050946285,0.36967726324834194,0.376972854243832,
 0.4933213443489771,0.37173065197470334,0.5745606369770561,0.7733629608199499,
 0.9256964980066725,0.1732655469895491,0.5880469278555593,0.1725547703538125,
 0.36483573405262026,0.9615204171788352,0.9819362838012492,0.6620392621727578,
 0.21168664486465016,0.5202714755680883,0.2677814647631247,0.9904345153216191,
 0.5776220382895394,0.15077630318245772,0.6975599618442111,0.8915923403538236,
 0.21737306985750204,0.9874589679968595,0.4350146116946938,0.9549951490320724,
 0.8569271092525891,0.008062626706382914,0.24863806176352643,
 0.41812201716339104,0.6177054088540797,0.3041489770398973,0.5386636984280796,
 0.5571051096143171,0.6240980608356117,0.8645494815164806,0.9951133710049973,
 0.6539982936189371,0.9458520012116325,0.24352984687842893,
 0.044654375361339316,0.7535792934908315,0.9066281629765752,
 0.9565837614396473,0.8673104138315462,0.934692681202394,0.336468066802011,
 0.6072707258563538,0.2558439894402309,0.6823683154500922,0.7015902531989912,
 0.8213403566446591,0.8451923308760217,0.46290400731391634,0.8353295140006212,
 0.12596508782237936,0.05075454716710537,0.10088890873814016,
 0.6976832576436331,0.4968527207753982,0.27435443270016413,0.6426202566380202,
 0.8047082830642053,0.18974661638855495,1.4696599513264186e-4,
 0.8507766366096944,0.9769641671681734,0.9310512348761656,0.5977138804018495,
 0.7738014549198883,0.9415842209648516,0.9532878582648857,0.979447673848271,
 0.8596383823681327,0.24551720136581534,0.32077272920823696,
 0.19520725307383624,0.35997260731288394,0.05354683063819221,
 0.44079289366814844,0.991763051195301,0.3632649323449899,0.8039447272578151,
 0.8992631353544318,0.88220963984185,0.9870778483848801,0.25725097210284287,
 0.40440921042715605,0.6457071619568725,0.08277681360561584,
 0.13273855512979993,0.9834943788631929,0.906354496632549,0.23821014565552068,
 0.6477816678873902,0.5391511694635264,0.7029351285766017,0.5023553360034769,
 0.3341477500620825,0.45353927792856474,0.5270167467269614,0.6254245651350393,
 0.3523535153686146,0.49325562487505936,0.3298419868587136,
 0.10181493989215151,0.9459419623003225,0.965362510314412,0.04964879207452966,
 0.695953853319392,0.1552397411225208,0.7734038226172217,0.7873950875589986,
 0.12235628988777714,0.020892526900582076,0.37914631446475255,
 0.7953993249269293,0.41252945432648847,0.5885049835601986,
 0.09522600731094355,0.22635207379025846,0.6292639846862398,
 0.2524219665035784,0.6004368223963639,0.12740960407110402,0.6560097923039222,
 0.03886695996629419,0.6243544709631772,0.9265816759910522,
 0.05733080876530672,0.04583905235926322,0.9461863980165826,
 0.24631331824275027,0.668639843521231,0.3579018476847816,0.08402649527685924,
 0.17411995764938715,0.44597757252993664,0.24493924931570288,
 0.07661682161301009,0.6181487767244846,0.42971319789432616,
 0.5823509746677825,0.21549401332787033,0.8391350849693704,0.9809389059717994,
 0.30671998891809416,0.6279334771890094,0.6323150353078353,0.635981258079812,
 0.4925516394387708,0.4135578057153966,0.03911681909299236,0.9668063345341227,
 0.7686741058333768,0.9129050117593751,0.8362731574142035,0.6003578755980906,
 0.9548908656273829,0.23251302116435268,0.23265815932950817,
 0.8764160326330938,0.17620948472157594,0.48687968998521436,
 0.04186007374494827,0.8427901794295585,0.9369630801130074,0.9285373839529666,
 0.5141047303658327,0.4187888508150931,0.05977300883565395,0.7728171702869666,
 0.9555078106587355,0.3539151631806524,0.8352708470654207,0.3568768704117591,
 0.8582731256072565,0.9551268260193188,0.15436667386258218,0.9489033575234851,
 0.3437365565555297,0.06585983716129595,0.45058166774325537,
 0.5540559674585745,0.0037334689812367206,0.4842805978913215,
 0.05695110260745617,0.952370664380426,0.4204570647544035,0.17078931301159705,
 0.32361110877327404,0.6077343277584463,0.29682243765689686,
 0.25366220188682187,0.6290933431273072,0.3531752295696652,0.5379336597051322,
 0.7420148983237791,0.26807700845984206,0.2356227286848075,0.5087023984026542,
 0.4286571810134361,0.054283262921962505,0.9761223280518363,
 0.6786472875554728,0.15493768660869778,0.27974396428995973,
 0.19022193690100464,0.33712105839273976,0.15720475938544953,
 0.5114699641433712,0.4844676203765358,0.6695549531750751,0.7683751255340653,
 0.4647306479175938,0.6978316501901548,0.31376371923660606,0.5388562761145681,
 0.9115636209200552,0.02697925645963417,0.9104165752741887,0.1804450014520974,
 0.37453455985631434,0.034689860181083665,0.1387965942228343,
 0.5586297154725665,0.4620979716357074,0.014690685924579117,
 0.4757521952027336,0.536069375393165,0.2827864297632574,0.755981039186257,
 0.6342413436781781,0.8216227616508566,0.24113041315722983,0.6197443427345183,
 0.8648026730020532,0.12139603656574016,0.9935242353826911,0.3495972364304959,
 0.31792764880612956,0.024443886122056724,0.7997305337197465,
 0.5655535997165145,0.6455994097032961,0.11345158823209034,0.5788830395324628,
 0.9929333515346888,0.8845787592248022,0.22555637377004856,0.2971843546612274,
 0.20193793745957744,0.4068938331546412,0.47037165818946514,
 0.6333567390636674,0.4322999747364681,0.8977814826150498,0.23756615541767978,
 0.22297305789020172,0.7301845672739801,0.6798350296502613,
 0.47838024902158993,0.7279242163474072,0.35936721527755133,
 0.9518629283676399,0.06070068465231493,0.488258449644591,0.5514589169956041,
 0.7196721313007489,0.8965361361235755,0.5169186558378684,0.9183970613134369,
 0.4748558066969468,0.8199480743648297,0.7182505256071828,0.15258521486499232,
 0.5326475576553599,0.3455545725352118,0.874141302652791,0.11233835879381437,
 0.961944023722469,0.7003253678815403,0.9579766098428256,0.9402856397085884,
 0.9621291318955192,0.15203977467197416,0.11950265696322548,
 0.14780096125650133,0.019833945420259735,0.42517981304655195,
 0.2779030683666518,0.19666032321397708,0.7207048417350624,0.3376478856572872,
 0.528981842770472,0.41955894901139323,0.6135442725558823,0.1920628339936934,
 0.647047136414691,0.889958808147004,0.45504714416492165,0.2766907763428501,
 0.9197098189917194,0.7777745896276558,0.7705170029494192,0.35709349547159186,
 0.9963559301160525,0.679463383933884,0.004606225701973776,
 0.19167135760450627,0.16669707876227557,0.06505473911863358,
 0.42537234831137316,0.9960453175245578,0.0437437767855442,0.7344189984917446,
 0.11645051996425548,0.7358348878425733,0.358675786467497,0.02978681596052457,
 0.4683706323903578,0.9404731425112862,0.8238314458938948,0.7879497365160243,
 1.3820354156779402e-4,0.4472894009177204,0.741273643366258,
 0.8517096521996692,0.4787368368288818,0.049535457931205285,
 0.48180172216326356,0.7238809561297155,0.6426662142113697,
 0.30943030287260076,0.9616610280727265,0.5575443090406597,0.7261025797637439,
 0.8289131302189201,0.5174790532896469,0.12682771485328992,0.5956294816677923,
 0.6973982550502174,0.008343318528851396,0.42089316621074546,
 0.9523237810709965,0.635811397558345,0.4856730122978018,0.9779183085050701,
 0.011766182400317238,0.6259014786383441,0.5813673975854754,0.63391998626217,
 0.34913519343867405,0.600453863518573,0.3278600843385664,0.3525054404736868,
 0.8003324423893075,0.4404570534025901,0.6024970035568058,0.2610198012355529,
 0.6708985511158284,0.525998004917075,0.4735597987161845,0.7483108759747498,
 0.06871565849600303,0.5366608415842195,0.3209677631872667,0.8927907176063832,
 0.9734996694120162,0.70777472691599,0.9989086012208285,0.5448294097563965,
 0.5748942061559384,0.07338748060014577,0.9902724254737402,0.2675731044110381,
 0.7895669610025982,0.4606155301466588,0.7428390528142301,0.16348104385285556,
 0.34161383029924597,0.4003430052327621,0.11159947805319526,
 0.9797648778477076,0.0026468686342701053,0.4133435774859421,
 0.45023643847045414,0.10512407603245211,0.9662953632518907,
 0.5308694185970672,0.022106354645325066,0.43730255547503594,
 0.503296411185407,0.5390103126307624,0.2747676533702581,0.9388115996795751,
 0.5750933470022815,0.8975387704701239,0.441028462317377,0.7409188560032018,
 0.41179672465407635,0.16692261301234945,0.0775251064361242,
 0.8797893850195054,0.629324222132381,0.41320073016744185,0.8866836067428749,
 0.28902956564815385,0.18620921482386787,0.8707861650564475,
 0.9611736674463635,0.8434267359996309,0.3529355166481566,0.19029839397339432,
 0.1395453041363991,0.734183157087514,0.22574415373517565,0.9875073116748476,
 0.1150452620231448,0.36049467393344836,0.5584845982713411,0.408354853212229,
 0.7203534545427905,0.6938223912783422,0.9067721720974047,0.8809983919922107,
 0.7027985824460546,0.2715767696402145,0.38298551138529735,0.5412177723648715,
 0.6336810255927319,0.35684845980489976,0.08353459143398001,
 0.8769572108747645,0.1315122938617166,0.4482991224020977,
 0.027557111554650904,0.7022798018569177,0.9622783000412857,
 0.8960599542145459,0.6560395216727545,0.19496947506313,0.7211095721337577,
 0.515238085728854,0.540678886530658,0.8215703860878395,0.07740668442761311,
 0.9130319126523609,0.036109993694693365,0.5085922601724091,
 0.6972741000830233,0.15132114257114737,0.030678191796368548,
 0.3924191505843354,0.7933352799498133,0.405882162570827,0.5507995497034618,
 0.12042831963568468,0.6636528774817115,0.9066902124702894,
 0.04858702023669814,0.3481646217736375,0.8721063536132078,
 0.39946385656222716,0.9525372108566358,0.7782299658812503,0.5161557631704251,
 0.19256141286641526,0.23561402245964091,0.8860291948389553,
 0.16165579931994523,0.9446449179424824,0.10223731979511952,
 0.8965368113431136,0.7975251379064142,0.795444712479155,0.2437514184421421,
 0.5151783121367549,0.39346790476122995,0.09451835523912355,
 0.001018689040366283,0.01915658264814646,0.9210667996062044,
 0.8972688220018623,0.9483780924184879,0.9987309185895215,0.12874940040695215,
 0.8593016749616709,0.8034774637392403,0.058933525132642495,
 0.47928178209793015,0.4509658414382235,0.0495495713538866,
 0.48448043698037324,0.05035990366487186,0.3898221222248145,0.277702421940762,
 0.20406308337284385,0.9885179674493589,0.7266565434653102,0.6524086845119392,
 0.6570597600699849,0.15750672387279907,0.8370096971521963,0.841679760550935,
 0.0646249274908065,0.2787199247540817,0.4531408865319948,0.05029148820455631,
 0.36986677341626006,0.15043658432884477,0.8599404478873586,0.764898903707475,
 0.0062788466803817,0.8486996287327332,0.7241744226421802,0.9692547806441453,
 0.9718925606156248,0.33992059743709935,0.3082649585209416,0.8699754458651525,
 0.3133967545145646,0.1611054983467054,0.43452503231341255,
 0.034549468844378595,0.8634161068610768,0.3909021440197671,
 0.36418225685584915,0.16666021573088274,0.9286537414434598,
 0.5621337942330735,0.2996480401583764,0.9047228135527874,0.03026402486029356,
 0.49173836520104375,0.7426083748376875,0.8627692119350993,0.5704582476320964,
 0.7540904813477037,0.8185644975640826,0.8536698277273586,0.14534964019074204,
 0.978986072176378,0.09260989287895027,0.36650098482613847,
 0.08511663373082856,0.4382493818898616,0.877914543880549,0.4917558482875499,
 0.12324057543266842,0.1110257006349975,0.21869868413493032,
 0.5970882452204714,0.28300439749678763,0.45297371182020896,
 0.8131016127239888,0.3938651936607317,0.14723769441142065,0.9412056124734394,
 0.4807221534177074,0.19996562615598146,0.9033656148206699,0.5450432032521957,
 0.9995397472736044,0.7955321268885347,0.5243489581507299,0.9327567893598188,
 0.23419344637178563,0.42766542275203345,0.7278295432528327,
 0.9035259423589368,0.26786896733890464,0.640312091935856,0.5792740639781487,
 0.46412305690783695,0.1593392995904731,0.1458115536781298,
 0.35867480016569653,0.9530816297298792,0.6386141415223388,0.6998604051708899,
 0.5373712696869174,0.3653733124437294,0.39047146150532874,0.3190635361993681,
 0.33325312738842505,0.40808733918963114,0.27588000559288783,
 0.9511944992450779,0.7621197881028986,0.16970424847691468,0.0328086936883788,
 0.7760390527026694,0.17101870234309424,0.26221851717086486,
 0.49532484339496785,0.9770073056659785,0.5030945751126481,0.7127089262880908,
 0.22177294424571925,0.1603429287932796,0.7585536444067817,0.526938726081394,
 0.6389806380889401,0.7137549448593248,0.1545239572258813,0.35545844493652556,
 0.20899253163288423,0.8404899873922416,0.3969939353360925,0.4689754782971445,
 0.5552465606458366,0.8461976277268494,0.5757865298512063,0.27953648996888114,
 0.5710504865104854,0.016917121642227917,0.8907248749507657,
 0.5563177480803256,0.11635218826530402,0.986313106659346,0.16780861509354383,
 0.09355140466093392,0.9513867723931142,0.7176516625359451,0.8958355852449282,
 0.2447126087842263,0.7746300189367479,0.49947257065895967,
 0.41967114799582317,0.7173062602387006,0.08707347071321736,
 0.27283901938115895,0.8326965450178591,0.5746845576643075,0.8158499679181869,
 0.00404121881117836,0.5126872226981167,0.15981326647043315,
 0.4651542550312914,0.6034885158788353,0.1851536647799943,0.6358651868540707,
 0.6929302435148993,0.859793195621417,0.20446045392122147,0.12829838438477248,
 0.8629069086422438,0.4528943739060144,0.042366740219703436,
 0.9564587302608717,0.544061393846486,0.8198967247857735,0.8051616611629343,
 0.7522423775664906,0.4830229317406959,0.8860990827718529,0.6495209028670118,
 0.3760604652951306,0.4974649843681337,0.38650656519660764,0.623883944468035,
 0.01396542373446752,0.24750692967616794,0.8444799373289774,
 0.8527495577775194,0.29724805799601406,0.6467978387394211,0.7046620947814517,
 0.9180177846783157,0.9037195493908351,0.40957845655389324,
 0.027973816544054952,0.8302131664278515,0.9665120957679152,
 0.7839926108513432,0.48901352946712195,0.8068904150882109,0.621002342950819,
 0.3679338923138742,0.35008391041886966,0.5475141208639267,0.8054472169857722,
 0.901517801861478,0.9877276397196447,0.9303134024389552,0.7103880068998619,
 0.8811187872319497,0.24818141416093453,0.7091793348380444,
 0.43288273403781374,0.12639913711237183,0.6107661972828885,
 0.40667488246350936,0.20570733012143794,0.5444413822513149,
 0.19513630737108478,0.32357156413592725,0.7659469080674821,
 0.42970713897390245,0.005478363831792565,0.30947743474826206,
 0.27782468813831473,0.43619311902483116,0.15913118919978564,
 0.46778293496557277,0.5727464587233639,0.3850927848857274,
 0.0022437075908614545,0.0772963920750982,0.4014246184038198,
 0.19205570433651764,0.15841948112423077,0.9333614381655512,
 0.9531477677014413,0.9091928383475498,0.0014843491820004662,
 0.9886751656479715,0.995814990278104,0.08966714345226856,0.06437445549513632,
 0.45524681204299755,0.7390008540442685,0.33319791393563114,0.78949726608377,
 0.8068552833240672,0.7455178656534621,0.39981632158001457,0.5777814562452284,
 0.5768945945672335,0.8164086713299589,0.3795740454410197,0.707504526781762,
 0.742102763973912,0.05792353355237734,0.11310892521410587,0.1999876304025472,
 0.43591573714029896,0.7774882493403954,0.45455650006006787,
 0.7365126724028134,0.18965665284455957,0.9691316861332613,0.614258640288946,
 0.6368565175695491,0.3093361923368868,0.6564708937243421,0.40230146967946545,
 0.4987186542082994,0.31510737230214925,0.6314805848661799,
 0.43184585999352954,0.3173507052339156,0.9349375734496627,0.315108932161549,
 0.09461723246108034,0.9472908979468527,0.6494684731149347,0.1553034978685095,
 0.785669507218897,0.1437422634522072,0.15731928995254474,0.5917692851311213,
 0.8616361124809719,0.7906950851661894,0.6372177130671384,0.43146815569979935,
 0.49818188497510374,0.9501871134621289,0.899030609108093,0.35831691673103816,
 0.970431086821642,0.8452144681437846,0.38564248134861046,0.13679570216138115,
 0.9083251116370592,0.0866929291921198,0.09597764601702696,
 0.39723022140514397,0.4072689758361344,0.0899830592390033,0.8824594774695357,
 0.8901377096660594,0.38598507405525995,0.5538985806174233,0.6749619247734963,
 0.24821216928661494,0.38161926432691706,0.7268975842609106,
 0.8077989269960487,0.6543310467038788,0.8724501966088564,0.37909313226599606,
 0.19645358848191663,0.3981389294737515,0.6103816257649197,0.6975725383519407,
 0.25974515336285653,0.06147890641510467,0.6099465044329524,
 0.7295931758582805,0.9069291816459546,0.0602818695991042,0.11844012224507916,
 0.9657196290054995,0.34033275998644397,0.27213507204102205,
 0.6773328604111089,0.4810200996179279,0.3021359254137642,0.12357782525789185,
 0.9221320892983417,0.008110107041123027,0.7888068175412342,
 0.6090752370263807,0.35156991203562626,0.38803134339486844,
 0.49966448667768826,0.6074910605543384,0.8511668452522685,0.8314736459833706,
 0.3774064387790407,0.23251564488697662,0.4312095103334116,0.613678791578241,
 0.22052346748038398,0.16094112654323878,0.6454772486053307,
 0.7525332947688433,0.524493118922606,0.2037927100837319,0.5149334986550272,
 0.7496154481466978,0.09203287620755218,0.6239254756368531,0.6239227789492974,
 0.34205645665236395,0.16556957203088696,0.3043199159561353,
 0.9754116999128161,0.8233381147360409,0.8126940910478017,0.6962634729870817,
 0.3588346123579993,0.5964716995673587,0.717770268031511,0.6291968080862571,
 0.20307124566478174,0.11284898175396751,0.21511440898258316,
 0.9408908727361347,0.8140746767670587,0.11367968750151036,0.9114022434880302,
 0.6438852864612508,0.8335249676836175,0.5542390093013099,0.22388676413131692,
 0.9809422109186885,0.8341262687175856,0.6494177299107462,0.772026618338227,
 0.9966691560345089,0.33726075332626837,0.9331421867400455,
 0.10655910584016182,0.442375192576149,0.05377080121071531,0.3132088258028958,
 0.2520178761239298,0.32529053161466337,0.6228628752548228,
 0.40241627200412544,0.22580254089385354,0.389843170526444,
 0.11300089377139388,0.7152920923888337,0.8607130889024588,0.9913327750825932,
 0.570318697984421,0.12855674570294617,0.3362944728087144,
 0.0035975775835727086,0.5395971535091779,0.21324723004494095,
 0.8789143373750447,0.22440915773900472,0.978603326171678,0.6886475240408294,
 0.7344277974161452,0.7192321969564193,0.0769405918497501,
 0.056833831788102085,0.6202570634078642,0.7066186894853024,
 0.3603237064888155,0.2904503950909717,0.6811771770042023,0.6964211992613885,
 0.9340450526317137,0.8082402469999644,0.7468157096433654,0.7598085744817992,
 0.5108332364068783,0.5097552021484726,0.3803317563453096,0.5165296128438978,
 0.9463372863255386,0.49641798156377615,0.8587594734023071,0.8735030859142474,
 0.8023386734951954,0.29266337226356065,0.03393142173204222,
 0.6316265958585315,0.7930104974447648,0.08194099629590534,
 0.11111113404287687,0.44907177478822324,0.5441209807624537,
 0.6392910706139263,0.15907081123064148,0.8528574384024876,0.3924297057931333,
 0.016724522673592257,0.4174629387193416,0.16737349847677163,
 0.10541891316524477,0.13684740932597983,0.32433067471393684,
 0.21638662034410538,0.04607032877859363,0.6950726526042367,
 0.3092366371822518,0.323008621462357,0.6248247836181156,0.4440071928002839,
 0.13399100424464028,0.23591727221616832,0.5631868844764988,
 0.6454622846447834,0.2928260373612952,0.45260375924231533,
 0.29181988409611537,0.9609970540628835,0.89860658823972,0.8000029457791068,
 0.3244973031933349,0.6837157092330531,0.5457213959897599,0.5539985533557398,
 0.3741098891970328,0.04677416296682668,0.40655297779789656,
 0.4391438886188552,0.33641059925109296,0.6837750431734977,0.6350213890405142,
 0.6999191322249441,0.9970126951100877,0.23402166636889044,0.9933174043120604,
 0.9263079524378992,0.7598102541872858,0.0012859258745707702,
 0.17659205451385962,0.09004827931355508,0.38435581804610286,
 0.43387225174546107,0.18667956431142507,0.9333228962272593,0.783902668261487,
 0.7802363460480558,0.295960101521079,0.28432219168547435,0.18305546653432603,
 0.7724425586677774,0.15561872365825846,0.05181091519613057,
 0.6065670350037173,0.7911937994672327,0.13502702552829693,0.6843395280426604,
 0.16399044090138826,0.24778707614559037,0.19618213778611993,
 0.44964003119301665,0.8112825500092984,0.2759092482553529,
 0.18205179758055545,0.7697741806333436,0.47500964663825507,
 0.8616786149483148,0.7724708534009811,0.1869329697287071,0.8857547688999066,
 0.926892388120347,0.37791275623702625,0.4531167631095011,0.21150850528550313,
 0.764690782633934,0.269095643977032,0.32265725785321964,0.40116932531995664,
 0.3663443416837713,0.32416963538264254,0.7890813560037385,0.8335067093914701,
 0.9768015594518042,0.12567852622978137,0.46101528597950836,0.582775304156713,
 0.9931445644725831,0.23369093985264933,0.921242433551149,0.19332356441736576,
 0.10153768196697344,0.7414797432795565,0.43841441788822766,
 0.5705108845551965,0.8425564062372767,0.7038496811900644,0.02022435977084258,
 0.20042089473874503,0.8557100220195909,0.654450390187245,0.6028919826304293,
 0.8482736412966295,0.041851707968641794,0.6501932138718562,0.625340042433579,
 0.47533457757564734,0.13556611878748437,0.8383856468673876,
 0.5478124994934338,0.47701739923521247,0.9064961943732179,0.4551058916594293,
 0.20680449962952818,0.6222789573733751,0.17626861154348383,
 0.058053615591960206,0.036584816814811516,0.17837259251854554,
 0.967248708034389,0.3477811392390342,0.3623137183374663,0.27854099883618977,
 0.802593489401513,0.7668966289875382,0.051566437146444155,0.9527600894433699,
 0.1298615528996513,0.8827736202423884,0.775676955701423,0.2671023315283446,
 0.5786273027598487,0.2662212197139727,0.65040938529986,0.5841271727573873,
 0.6199021659268062,0.9508760798814863,0.3046215828830129,0.20610880468303194,
 0.4393205632303123,0.46090670392676236,0.15336430611037666,
 0.10184285005316362,0.7538422469749457,0.5780122672751881,
 0.02930127217803602,0.6357395195808478,0.9826675910182714,0.4595801206002751,
 0.609210071092963,0.315121961993009,0.39901163742369494,0.18946953431197255,
 0.4442244316042685,0.509263940046986,0.5644523137677474,0.4572100585351224,
 0.6688712105299759,0.6165912235295561,0.9791530453714715,0.7251060931815986,
 0.6754336416538614,0.7122361562731497,0.5458315396731925,0.17179215060052988,
 0.7855533744579771,0.6441319851080092,0.8172248933972397,
 0.011624641991915596,0.8581686366744613,0.7257202579268426,
 0.2641657152809582,0.1631694474994827,0.3440244788167507,0.40938754862167126,
 0.33012069206476624,0.37840066546435314,0.4447544886821433,
 0.8634018892415127,0.3003933295135359,0.5603457302519881,0.5135611370840034,
 0.036432703732301785,0.7745982274055474,0.6106338415977179,
 0.04786685820262804,0.7957407548326396,0.5356611016088744,
 0.39840219138247013,0.9161866919855649,0.27679531590620066,
 0.8858792393600904,0.45264169770534635,0.9947757016225968,0.7250594650705785,
 0.18501090701571155,0.49791988261308295,0.6594247632834627,
 0.2652567708336847,0.6388205694262069,0.4719030890866144,0.11047162278733191,
 0.6371899838415622,0.6359006356181192,0.8668624033055174,0.7469717579130597,
 0.6003101878822751,0.09569742000912251,0.7463244575018843,
 0.011346271360315296,0.5677562461876626,0.20004944616394071,
 0.8522361348078751,0.04499044179569833,0.9319163022986626,0.8984475480282411,
 0.6769914425250753,0.17255874257458792,0.43649921317993723,
 0.1392574938464488,0.7914510142573571,0.7140171241818131,0.3556632636267981,
 0.8981712640465914,0.9924457697230458,0.5028463249797772,0.8053445769188452,
 0.800048237525503,0.5518320648995886,0.48494373122490697,0.28199990579261214,
 0.5508279370350972,0.9400211431972663,0.15093225737013893,
 0.25945795736658805,0.25363260925464315,0.13382166036495635,
 0.5273972685754764,0.6543174406269727,0.416362373086854,0.3318712877973867,
 0.20043915008276958,0.2486795516043674,0.15632710624901058,
 0.053023321382604016,0.4076527382385152,0.09687387734549413,
 0.41355307544139386,0.3089444896307678,0.10306363542149677,
 0.27689489283779545,0.4709558354819029,0.4920565439374669,0.6174841687825704,
 0.9319586810898056,0.38745395865700794,0.5601674763474659,
 0.029417508588254315,0.8164274461124392,0.5603027676278248,
 0.5397946059793788,0.42918793570609504,0.9072457808201804,
 0.27505025546859374,0.7683061901326147,0.07791217438118447,
 0.9565290689192351,0.8066071143309791,0.48130295594930206,0.9887562845451302,
 0.9914843301003744,0.19742956330315165,0.695630609152812,0.21185950171608336,
 0.45291786552543445,0.21891514811924584,0.9845169389431587,
 0.025614910426561166,0.8276894870975701,0.5190227880878877,
 0.20052015522156819,0.8521750723955427,0.3622727942825056,0.8842009960950559,
 0.2595354694994245,0.6010289323877319,0.7329863017810847,0.12874754985747483,
 0.8655932020717467,0.9326868260391484,0.5216836345887697,
 0.040354722362858975,0.5201903227573313,0.058927168580522715,
 0.979802894254189,0.427812800509657,0.6381275282554979,0.5030335922963526,
 0.8151109024381029,0.7148475934394337,0.337990708368075,0.2524915380212024,
 0.09177024476896434,0.8155979757845573,0.09729482466657424,
 0.09889613144498421,0.09726231320447287,0.21138428415635813,
 0.21288944713876545,0.6768616043888058,0.23556444596503234,
 0.1214503154846589,0.31158977777662933,0.4579950421266332,0.7869425381535775,
 0.5130278346350475,0.991043443920335,0.09904764999142501,0.09734220998486387,
 0.8725153716645425,0.6626558041447796,0.38056410351889525,
 0.057578954932806514,0.5257370160470517,0.9751538000121007,
 0.2740644472221119,0.5440283406548552,0.024270683097061818,
 0.11883831561833735,0.46074787365759995,0.3216056875260672,
 0.18172318791946296,0.6166418529462355,0.8527455146077777,0.8284201195266562,
 0.7991550109136883,0.4710046946793004,0.5073484283294772,
 0.011979022721920862,0.667660397840587,0.9009423223248066,
 0.20362998200747395,0.04693837851421889,0.8750699700845314,
 0.47100434302114336,0.3328299534928131,0.355001860505335,0.16400692348461954,
 0.0823288939315735,0.12612687070211137,0.3752235802605952,0.6898901393243737,
 0.4463767309707085,0.6125105462169956,0.8948031615002663,0.22159051114943717,
 0.3172925387935329,0.4072671042732422,0.7381781929605497,0.01031208048492993,
 0.9765768146273626,0.26064292093667174,0.25527793846422475,
 0.6957697885313759,0.7923417674034579,0.13232547533267947,0.8755971798338377,
 0.36776329650427475,0.4618719825821471,0.4493315245902938,0.9816381489013353,
 0.5914422008420706,0.6584994022136385,0.3124933286162216,0.2662902324212023,
 0.3653681260896239,0.9045051848493493,0.39980288561239474,
 0.08109362170663381,0.03219372767198503,0.06987305181421677,
 0.18876181962436178,0.975561745066369,0.7590122608243232,0.9927987308786701,
 0.4670995471884759,0.7007283661457622,0.17932866875093112,
 0.07939575611125149,0.6565853725327201,0.10018185283106362,
 0.7470389453445943,0.5793010152029793,0.6615957385867457,0.2303560331279354,
 0.2249526883164784,0.9042499924630195,0.2037580338124111,0.7882325064914222,
 0.24358355689681377,0.1620413060619168,0.8027710163300402,
 0.09890425678447712,0.7370368292393304,0.41922993317034163,
 0.4111992840116033,0.3055551302670092,0.6975860905259808,0.6114196780860888,
 0.14429367665510628,0.8821378811788149,0.40121497567767517,
 0.7975862248236589,0.03572740794265872,0.7060739066922813,0.6806577146957814,
 0.4312664744952117,0.9932153054982868,0.4899744751314077,0.7828887674643205,
 0.4561724355617587,0.265835664989879,0.19425362722581763,0.8988898031949678,
 0.7290091048230547,0.37639412589641297,0.6543589260075576,
 0.19482474201273514,0.6346349139669213,0.48741227556864763,
 0.8178906556828156,0.19468111191738457,0.7833038636248432,
 0.28271743843356656,0.6402058406530156,0.2935164943916261,0.517091584683376,
 0.4084606052836597,0.14903657662266045,0.6022919100002806,
 0.35040786554417613,0.4210669422599306,0.32931779088977653,0.182738649529104,
 0.5930120219590236,0.4294419199753068,0.9993480685988843,0.5315031593564845,
 0.970085154449398,0.015340853226106232,0.7533641777038491,0.6015486466270782,
 0.557627918225624,0.9562324287288576,0.4081666015097427,0.21810290429750234,
 0.7808945905011291,0.5180596606769179,0.611562199251774,0.6043927244709206,
 0.45006913106502155,0.24077703300265207,0.5138087404502726,
 0.7534682410168088,0.8887749879977996,0.07026348463852894,0.6181301665461887,
 0.8157861860988698,0.523857940184199,0.32423725503383516,0.7787221865114,
 0.6238237715065276,0.602946197981188,0.9483076815315931,0.2334241908655148,
 0.6984609072149358,0.36953487883358815,0.5231615236529287,0.3287024927787965,
 0.256117479904609,0.9433438646391242,0.17716783720760731,0.08039031200798707,
 0.3675731150243805,0.5254254747797914,0.18440355835952515,0.7421967330504624,
 0.8778626431864336,0.9683966852520671,0.9079788817761908,0.7708365568858845,
 0.3090151590042105,0.4484767601008781,0.6378389411118678,0.41708443597703493,
 0.8656920911659798,0.6603608744556589,0.4497232805066914,0.585241608913786,
 0.8724918878168595,0.8277761496611481,0.19109806047922617,0.954524434415823,
 0.829293442929941,0.0017723931424225015,0.5259579668154734,
 0.6930729213865181,0.11513071067236946,0.29319985518051883,
 0.39459854027248953,0.7709842546752362,0.788079394171713,0.2989239033872244,
 0.33953312609249553,0.6952553645426863,0.9512074224041817,0.7701182064268316,
 0.03390668541294284,0.9023634396517295,0.32779204308423526,0.582405192936194,
 0.032282569230302816,0.9419372415645372,0.4498748888375673,
 0.11457456704669033,0.5662914342992018,0.27949971130974804,
 0.07762938266703379,0.0873065254833898,0.5524896866813453,0.6099487610977095,
 0.6083272446783846,0.8059481337657037,0.2341245129452787,0.8381074973145378,
 0.4086100003935075,0.4434175396428781,0.6204204957020845,0.7139847808263842,
 0.3485628909327644,0.7372411522688069,0.33562334106817604,
 0.25657434983486094,0.746303978202737,0.41062201429882506,0.7318570272593363,
 0.6026359343805014,0.9911814798999462,0.031659965577789384,
 0.38082444546382965,0.11122792668672232,0.40601890063051826,
 0.7699575476203135,0.7736979283803289,0.5507644285715624,0.2597819660111784,
 0.9010054353474757,0.9223485611688246,0.010584964624335225,
 0.6781102204529234,0.41007252897312896,0.19272681029915784,
 0.31833701144299364,0.7058675886224903,0.6805608177867689,
 0.40245092461832965,0.8471980475345315,0.2403733147968543,
 0.08014943805927244,0.9492908293604236,0.45203179010221317,
 0.9802426101407897,0.45289994107786746,0.9161353115226728,0.2657871617976105,
 0.753107192511067,0.24497697453867173,0.04199514872318,0.921120377903999,
 0.3450770922011964,0.25045380287116814,0.9183632057312935,
 0.08782231659321982,0.612761844483806,0.24897868809803891,0.9900170731421574,
 0.21499543336067495,0.7437847598651619,0.5895441883474101,0.8494278745651099,
 0.7019757006164271,0.018973633004974455,0.5402496500044245,
 0.8338457030837764,0.10003485864957995,0.2656010664755256,
 0.36865920157724863,0.29090573698676936,0.6449451934738353,
 0.5350009663323423,0.6994202946366953,0.7404946358642189,0.17837111819548523,
 0.6793628139805198,0.7485474972957076,0.6704489266681922,0.00833690523868813,
 0.8182329583537172,0.2590389105222817,0.18700958933191147,0.5213674333607421,
 0.09238773822890045,0.6083602699767926,0.02600300204664885,
 0.7949086276260667,0.4198558145912624,0.9828990597145543,0.8041247245318444,
 0.5559607394519872,0.3302863773344006,0.4421725889201863,0.5560956729343552,
 0.05628807053975682,0.31862768070455927,0.01074977092589624,
 0.5533939539668745,0.5785027212461186,0.2505191962670861,0.7536599398295886,
 0.42774737541964136,0.5291352834265224,0.00961621365963461,
 0.45133192201324956,0.044788656638341307,0.12747382185605616,
 0.5961194657424496,0.9536286032371948,0.11388615619403142,0.2025473688117193,
 0.5412201329350371,0.6424603602821324,0.20855656419640622,
 0.34125314722438194,0.30712734804640907,0.4689426599720735,
 0.1847403324035436,0.5319632472077537,0.02771484860929807,0.9581042401712418,
 0.7314895885803167,0.8284376835913135,0.8486693306656433,0.3623558631304844,
 0.830953581688178,0.08366077958720008,0.596740002889312,0.28043707712558197,
 0.6572055675471511,0.3472317172425854,0.28623408517774773,0.2819630472250829,
 0.5852872238110423,0.6745646235795996,0.40865897624417125,
 0.10362265699335949,0.0037408744157370055,0.09220387816872655,
 0.2736867972155611,0.37209207220232277,0.39248818527017904,
 0.6556727360784116,0.5884748515879555,0.453937048373515,0.28163053227942847,
 0.7300693822463574,0.3142941028749283,0.39303635234562273,0.2988404076462787,
 0.5997901446946395,0.7367195984454874,0.6745251629713873,0.40882558306625194,
 0.061544102696464154,0.9833238099695126,0.652272420075244,0.712020777239625,
 0.6394887109723608,0.2620921769345863,0.46760289187524795,0.4564536018633072,
 0.701888001020174,0.645287775668213,0.42552781461440325,0.25469673549000427,
 0.6364657531981359,0.8019689688945721,0.09776548556941267,0.4468034957844167,
 0.1725067425322212,0.2767082309563139,0.33464830532008605,0.6225008166371189,
 0.8622388562874651,0.552436024832594,0.9531645659779199,0.9468041280693809,
 0.3503719468194022,0.14084032196555796,0.060985116884654134,
 0.29381657411177464,0.22238048523493525,0.21068759991310515,
 0.0334223769316494,0.8057840845479768,0.19002552994800215,0.9954807185408444,
 0.9374242322304243,0.833700985565384,0.16283798534556393,0.2963946887161416,
 0.9443189464860638,0.11490100811314341,0.5937371487314077,0.4102424226843594,
 0.9446803259207499,0.9318757862145067,0.5287362495477849,0.665649594992228,
 0.434705416396007,0.4739496925480172,0.0228664565754213,0.21822413802303642,
 0.8831013749683974,0.7888163773516896,0.4916685111012572,0.6640766143627153,
 0.5477049467029468,0.3219606843821038,0.25597193345453284,0.949608996572501,
 0.8282668852929118,0.49899155652972627,0.1963605821844896,
 0.38426058548046593,0.12324129887363977,0.5784131166564913,
 0.08303773048731378,0.08524626974413985,0.7380188845143743,
 0.3305849722173868,0.3356707759387699,0.4770212535954734,0.3094598936992341,
 0.026483043157423336,0.8037507461943274,0.14528235203995254,
 0.3106979477966858,0.4664167486776811,0.3175357744088787,0.5572406119672495,
 0.21477308867482403,0.5083643479377236,0.00823607702960194,
 0.6511984713443466,0.4371151300134437,0.4112808245429841,0.1058327490360802,
 0.48238595535729756,0.3126738996506725,0.6037872652529893,0.3215157406013526,
 0.288919670586427,0.005460022263803155,0.09707186300722315,
 0.29266310910311644,0.6028687800671415,0.38059470669168305,
 0.8450632283930777,0.6758438161791278,0.9329273316363015,0.34302648388845947,
 0.9406370187042534,0.2468438402744636,0.8149805904554418,0.08186555415724128,
 0.03330720361171635,0.6259902229832321,0.008921800181530548,
 0.3495959511967992,0.5884601775157736,0.7316672726917628,0.05992144100729535,
 0.09272390567899202,0.40475307977742325,0.14083422338845186,
 0.06885461562765027,0.4218403221467302,0.15952564070209052,
 0.3682877209089923,0.5986945335254445,0.837839322869466,0.38444330460733556,
 0.9208173948595731,0.1621909349573285,0.2573108691941265,0.4009900653836608,
 0.4649780286480091,0.2868293772073759,0.24524324441858936,
 0.07935692369464675,0.8000351827102625,0.9849365143217078,
 0.03390408160859293,0.7223868104522752,0.9087364372232583,0.7629617785562175,
 0.678429712099254,0.9337292537347525,0.9465818604691398,0.7331433686293295,
 0.8799470090350917,0.5647943538801975,0.27843556793503144,
 0.40440277344812836,0.4023201965527714,0.6382735073356749,0.1116743334189132,
 0.682780423018479,0.2642472902609747,0.4165826295270715,0.009982959946634895,
 0.5158657397058459,0.4383206394194846,0.5645302097796624,0.96470314618208,
 0.16907319828940082,0.05200068574986161,0.8235858499703502,0.933253540429486,
 0.5362404638469733,0.22860363333371625,0.8374665267554047,0.801114916675192,
 0.8641269595393527,0.8661502687396417,0.54356995670803,0.8573702183701862,
 0.548189759640127,0.9588312127828809,0.07823733140565681,
 0.019661201484541202,0.8582792027476457,0.721788950671075,0.6046328596020863,
 0.33701581258995994,0.4644906700933825,0.8817085171331929,0.2509076021382114,
 0.3388900651628147,0.37394408988801375,0.013354741708404116,
 0.907283741563585,0.08870869080790533,0.8334264296777455,0.49979463022280834,
 0.10208968850585354,0.1488456031734371,0.9099952100950359,0.9819781712701479,
 0.8325553368243196,0.576348402000481,0.18428279273045778,0.9934961625287926,
 0.5792727576091233,0.09947940061484706,0.6533952470179969,0.7169410349211065,
 0.9464346098835817,0.5710459906445434,0.44500164462213787,0.8620442585740608,
 0.939243173397903,0.7154589362494073,0.887963537056221,0.47489200771057405,
 0.25246790548019205,0.3941300544008215,0.08063099408981467,
 0.15947610074668916,0.21474964238112182,0.07594448663423115,
 0.5762482073560733,0.8772983068735066,0.8029392868345925,0.03876832386824047,
 0.08396062417998607,0.5297222170677839,0.8116818843587761,0.7040640509861259,
 0.3285051318965263,0.5113928158488206,0.8887780075975993,0.26545724826475126,
 0.45609137570604996,0.9561455693543869,0.14020713882918678,
 0.2030156477427114,0.7682777672186841,0.6529354156338598,0.6980701446251292,
 0.7733329945378674,0.04259275298291909,0.4682788593434448,0.7588484365153514,
 0.16429779810160905,0.3505764745983505,0.7776583238047329,0.7025094152551448,
 0.030250187956236685,0.3196343681766859,0.802933970344459,
 0.25122671576298683,0.47222397037020514,0.3891364833300932,
 0.15045804295371645,0.6462349615998878,0.021907148685141986,
 0.3363719438881104,0.12648151024137844,0.6205159284993471,
 0.029829363313208512,0.5506337539230277,0.8546276569947527,
 0.2852919685259401,0.8833565199256181,0.10120122463119541,0.5725025804195842,
 0.7438550489020882,0.43727146526353067,0.273803768114549,0.6255961616533615,
 0.34473621189538894,0.827263684345799,0.2140452438802679,0.569209764827408,
 0.9914923667055582,0.7173215453707354,0.5807035055803366,0.7224595662755687,
 0.35064359285132474,0.4081791682143089,0.9123084952403248,0.8878006129202436,
 0.20607238708926423,0.6120521612978213,0.6534368075671216,
 0.29832348468735737,0.7368414570638331,0.11693216476289281,
 0.8294654389110909,0.2838222810503581,0.004309526723137447,0.555981132771115,
 0.4433561025903663,0.33271506786905647,0.46539059027269225,
 0.26682927430225667,0.2980258444945293,0.13945465656207823,
 0.9737264329490808,0.617720755395961,0.6582203065394588,0.4652966954916904,
 0.6306843906479543,0.5988600668745971,0.4482524732983444,0.17435233186294663,
 0.4684743035226302,0.8475378165403191,0.841910615460563,0.18134619047938394,
 0.012870120437410648,0.09385803407138771,0.40079230815491085,
 0.3329121926133799,0.21612553527229073,0.01740063162522243,
 0.2796114001931417,0.15723548523646969,0.9317614230123036,
 0.34413374509810546,0.3357766798532248,0.3160853197398896,
 0.41150576145500917,0.12147244384770872,0.5186779471447771,
 0.35879504686136166,0.7011116119772768,0.9946012861953631,
 0.18684432023150133,0.5325090098593779,0.6944152332054303,
 0.18257996227369822,0.3930336180055862,0.17944107840814472,
 0.5704466146040004,0.15142409192752082,0.8108376312973503,
 0.047579243580773145,0.5282695062769966,0.8406553577287047,
 0.8040410283711212,0.39985113864551103,0.35517354349144403,
 0.1567623836685268,0.24666871172930105,0.41817963792254065,
 0.4526567056858808,0.6806213805990847,0.8870059975958877,0.9300805798160481,
 0.6109621709872672,0.7438047211145189,0.13762496869419216,
 0.031088444763087475,0.10653262023583654,0.3947811711449438,
 0.04047972196385308,0.6534442792311125,0.7061552624249365,
 3.8720653705592856e-4,0.6038876932799593,0.3752649498003473,
 0.21771312375726004,0.6673108753974694,0.07372063046164001,
 0.5268557672504872,0.20427874267828872,0.884412248155423,0.982660187197509,
 0.6400421762756289,0.5640708673499071,0.02746047566850085,0.7907013005511333,
 0.4918200937855556,0.4563968566474004,0.3012813454263972,0.9905285860651705,
 0.717204685076195,0.49581736837725,0.23864246618235385,0.4424804646143803,
 0.8455168223809644,0.698092029707391,0.6771006986197045,0.3163881430390635,
 0.04052563974972867,0.6425131828897501,0.5230922209501077,0.4823048835902157,
 0.28354799951988063,0.5049925233409844,0.7831567858902166,0.7384109737421003,
 0.6514194039128628,0.5607831634538541,0.6214225875009437,0.18855739837806484,
 0.012020782607722436,0.3466587192348657,0.14369814839604433,
 0.12378364667988018,0.1225992734260114,0.3720433749310801,
 0.34429700860820767,0.017655309152701415,0.6036403205038893,
 0.05327725650822057,0.9408721072572449,0.7479454768586833,0.5086727553679281,
 0.02904423672219203,0.8788903708146769,0.642964009019073,0.296914027271883,
 0.527721910723507,0.6790301258030783,0.6091549910670918,0.8300431392961872,
 0.1353424666690246,0.3476870339410105,0.3798607770253506,0.7120561739399212,
 0.3188910767794446,0.6037071832282415,0.8786576696028275,0.5963984364705099,
 0.12199506671980775,0.5791372495720996,0.004547786135408405,
 0.2683411081230167,0.6360623384165053,0.8016931265089222,0.6633034014580195,
 0.22555928574890372,0.30479096001053574,0.7223881184215659,0.796868956638491,
 0.044191840976044894,0.14018860111694897,0.42889733910269523,
 0.48640151625601025,0.7924951845628201,0.5684404549301245,
 0.13890700103362508,0.7162124559275626,0.283342063104865,0.20842188832540332,
 0.4285030217823632,0.900313284779976,0.6721755200768587,0.6577629703328286,
 0.6959642541035792,0.09472095806720182,0.996101180706971,0.22562274022521334,
 0.5486886442920843,0.7661857619056173,0.39709059402430935,0.9250614758481018,
 0.182788511538704,0.51475213156763,0.11245030365224207,0.4983635043026571,
 0.5521269710058939,0.5863241396909947,0.23974831035973476,0.9131826816806803,
 0.977281461563253,0.6129222047134134,0.09644643482608006,0.504633751961688,
 0.06983446337193844,0.4270133521234034,0.07564208701716113,
 0.1975973558247478,0.688983406102309,0.941191985489916,0.6240745422285277,
 0.6633238792264812,0.9698738202514883,0.18851373140073613,
 0.45673369702806144,0.5412577611001995,0.17921208604440397,
 0.9223083735012293,0.015571245735607975,0.29898635477025726,
 0.9298146045407567,0.5502433699705498,0.9520688353104784,0.1548613513177428,
 0.09565789672862812,0.09059570981588028,0.09047050383595101,
 0.8272233840796767,0.517716812579102,0.2392385408622133,0.03821009566281175,
 0.11462921626280342,0.7905815021571275,0.19083704035661786,
 0.7528848432086441,0.018374095412125313,0.4297143036542711,
 0.7360580101005104,0.5984193217938399,0.44352075520136947,0.6484138542553192,
 0.5297434403375066,0.9892570669664746,0.19220253923893837,0.8324970817939157,
 0.5004533123743082,0.9958636389867477,0.8447983283459225,0.4539888985858548,
 0.27990244109311524,0.5854073116993247,0.7259185701177471,
 0.012147260757371603,0.8555498842218057,0.6797348188113639,
 0.6159155250608758,0.22585659404571623,0.5604925391751915,0.707842093455381,
 0.26879120058101524,0.43124408048588014,0.13364017017612495,
 0.9425389244567574,0.11327831794861531,0.6466087935078647,0.5952349288808045,
 0.43831499477415825,0.59772501729068,0.4200904884220177,0.30836332006797473,
 0.398999337209069,0.47614578456681644,0.5312237227962496,0.07993859420129257,
 0.43097954796383475,0.19654475289033935,0.5560092073210443,
 0.2083601045138549,0.3093761716417399,0.1136901224339566,0.8003163854367483,
 0.5448608843230682,0.4778532357476277,0.20905056558666135,
 0.40591911609144804,0.20266122396684938,0.6070559509365426,
 0.8837897731391221,0.2319414771061229,0.09390796624807407,0.5697945161926139,
 0.6560111293078049,0.5723567058692751,0.6223690554829963,0.1932490430456768,
 0.4108805068567901,0.25671290651117573,0.0778053910324914,0.905135754398497,
 0.8025122327463653,0.8856599930281963,0.3337204173068432,0.457446512893083,
 0.10221601058579788,0.6773574993618222,0.3638954156445543,0.2924328908385212,
 0.7557879236761578,0.11954494859983034,0.28367348104896883,
 0.8045814025583297,0.6731421520154194,0.6872170953993552,0.22070888361891539,
 0.4700960156930871,0.37526964948607,0.6347761634023399,0.6612448754564182,
 0.10967713090789166,0.14844062445558281,0.4505653319534454,
 0.9301954418185567,0.8098190141101891,0.3982208401750754,
 0.026942821403130646,0.9965630197175618,0.9995180767081974,
 0.3004927854252313,0.1899794560663679,0.6245009922579928,0.469301138234096,
 0.4278832983932036,0.21014317853736575,0.06525873746366928,
 0.9704294645313986,0.4398137599913948,0.007238050451357281,
 0.8793097370793417,0.12520956819679951,0.8816102885674454,0.8628124510721848,
 0.05508202954287622,0.16268218352599662,0.2543283477389734,
 0.12430870719539322,0.7973026220340588,0.4103164259657013,
 0.20757967505002117,0.7990021255693913,0.49604157181492714,
 0.27100560021419984,0.8240655037921834,0.325468104351478,0.177228255208025,
 0.8486292411950265,0.16783256038906524,0.9605645072905944,0.9207991465319609,
 0.3745336505319996,0.9155329621977497,0.7433355872512777,0.1220010822523604,
 0.8559045141741098,0.5015596394035499,0.5607118882893934,0.40226986819098887,
 0.8153471972377474,0.5236920448667802,0.7794889892078786,0.5762659106059153,
 0.755255970976233,0.4466635687465207,0.4770213473011007,0.698095201113843,
 0.8162092316031601,0.4657396913316241,0.013428747936228835,
 0.8818011609236327,0.9512518540290851,0.4413611947505781,0.6301025244987113,
 0.048175116036137,0.3328090109883638,0.67826766660328,0.5553682316712244,
 0.6286270220506529,0.7978317898028032,0.030729770469464013,
 0.9500084656743469,0.2502939988214171,0.8012049314951631,0.2868545499055125,
 0.7537644314113747,0.9090741937243898,0.6770179324474281,0.9363873506484787,
 0.1421579987222088,0.44737074706461244,0.23812422837357505,0.910468950561649,
 0.3979761162144648,0.8343739400595643,0.3026729346915096,0.5634144956753115,
 0.4087302924062002,0.6606461027841599,0.0018828703310452832,
 0.8252426612792135,0.5052987756438733,0.9558203751172957,
 0.024483782238321705,0.9956427881104032,0.5696872315270587,
 0.8026553239425877,0.15598963761128237,0.9176977745391188,0.7827470911607551,
 0.22973575230504373,0.22106794998532941,0.7020268138840691,0.528607611309769,
 0.8794817781352171,0.8403717197917154,0.10708456171933789,
 0.29677107801704894,0.6497510709599328,0.6851988663016094,
 0.49414802086868226,0.7800915332959512,0.03226304523377477,0.535363899568677,
 0.7044251142720697,0.5931110143009751,0.2118535761163869,0.3078571107441749,
 0.9452410083470502,0.3534618191972222,0.6727719340510658,0.9565641275695649,
 0.2550425391679144,0.03391955379225431,0.7676956431649304,0.7886952339641092,
 0.7483212848953544,0.6065608405978241,0.4528988014642802,0.12964210618248095,
 0.01672309271374195,0.42708254774420706,0.5660657744426137,
 0.6056944839787749,0.4509695386956841,0.6753332847347046,0.739682300012897,
 0.7799785525494861,0.6860055062275893,0.12263980305721733,0.9884128046230218,
 0.018098470054688787,0.9189625966533497,0.25653111581296595,
 0.8163119031364566,0.11583499581158374,0.9938043448813697,0.3797997982902265,
 0.3870549883811869,0.20993659017244726,0.7949659384729661,0.5675603439400361,
 0.8117457112361809,0.27989525463713405,0.6858842381478757,0.9712919404317638,
 0.7011239300577579,0.15249789374290856,0.5667824337736018,0.8769171941318881,
 0.9192997011221251,0.22564924585352353,0.8213854403984265,0.5475827886007579,
 0.7537614760345159,0.14284733557961016,0.23691758843801192,
 0.20578091444795787,0.2880142272466335,0.165439087857705,0.7725935772816885,
 0.29512573728969627,0.2739562619640922,0.48482906590797104,
 0.4574830264282005,0.20696364559121905,0.5226808994934528,0.5307978745235606,
 0.15426300499972645,0.019093769024779883,0.319589153596584,
 0.5179983593255137,0.9378392485430761,0.6724473523764519,0.7380944353912726,
 0.14972112737125076,0.1413968769113445,0.7181866343309196,
 0.39351625931391254,0.29596108112586117,0.437732336382991,
 0.03770179628931891,0.49077190729420483,0.05926413458443469,
 0.7286410107504082,0.017919131142082323,0.1597597309328479,
 0.9212878651086053,0.6827069167535637,0.2258624940850621,0.7494891391679588,
 0.1682527868645366,0.026144004379550334,0.38090191842121224,
 0.9268294459936262,0.8857259748012362,0.13714584928267237,0.5169710547442177,
 0.5787817904044361,0.7918539377295504,0.31415962551729015,0.5125494617643993,
 0.09898855430933062,0.7812579691304654,0.7680976076647794,0.9562677827984492,
 0.7278989183959303,0.5836791034355401,0.5036543668623029,
 0.057933447647658176,0.2299151803234808,0.04261596642647869,
 0.26512167524233543,0.8874282981710939,0.17228836252870128,
 0.7634674353448279,0.5201477917564008,0.5559297321719644,0.8521073428746316,
 0.9448911106186176,0.6533235727466276,0.3816945876154152,0.1364361467798001,
 0.7021883550127345,0.7330934350020946,0.3371069490580817,0.23677672083526347,
 0.9932836971755628,0.6896072887193717,0.6179448776357942,0.14277855691441288,
 0.8093144316402509,0.37486557250195285,0.3124673230716821,0.6507290829315266,
 0.06171739213272287,0.6528452105567575,0.665285900978653,0.8916158127022167,
 0.42485176435696725,0.16381603181926196,0.6924989669838983,
 0.4659280271139278,0.09280767597792661,0.9653246007108145,
 0.49880635184648514,0.7128803263011831,0.13113942531030076,
 0.5701360332863611,0.34842213582552495,0.016385996725911012,
 0.7225449161040409,0.23017092548818507,0.8765032086102463,0.6586600024692844,
 0.04988071318490428,0.6155267417073946,0.8764073868770296,
 0.22154581430468923,0.953116293503748,0.4435528520140233,0.13486238724339727,
 0.7896764966619096,0.9625593505120273,0.7159164283469479,0.43285244670087564,
 0.5074796739607735,0.5993920039747763,0.7579167094922783,0.3569104634661313,
 0.9447041889950003,0.05836346216134369,0.7840361965151567,0.797863875361362,
 0.19859137178713115,0.7318614764146842,0.9908810757706357,0.9308770864199415,
 0.8389235157624177,0.24210801481159183,0.41896149329059007,
 0.5656849441292329,0.8507070695221154,0.8557578790450732,0.350662413842896,
 0.5036993383599012,0.23723976930874702,0.8554727141852484,
 0.0015500317210073167,0.34752104856008825,0.760046690439196,
 0.7765209129392326,0.1664110488738335,0.015868342321998963,
 0.8455971419518704,0.1382757724016379,0.6961138163884123,0.7683660873875929,
 0.4619802701234389,0.20787616556656174,0.858347063602406,0.7551415552319982,
 0.007951193161952474,0.6508844296967082,0.9364209125547133,
 0.8402216111892932,0.160781454595882,0.08713858557787779,0.31159458497661807,
 0.9996565349197366,0.6662034269958408,0.45974460718197663,
 0.24014797935830945,0.9823318000448245,0.6083521616289802,
 0.21904825528597016,0.07429438574602953,0.2984838104239692,
 0.3487046401845293,0.7389840536385175,0.7704890150421954,0.07128425896981005,
 0.8958710886090459,0.35715869912043263,0.7311173723450695,0.6778128560628798,
 0.41990813360377,0.614352128559052,0.9173317622393038,0.6012087642665477,
 0.5741919959122688,0.23737886896302796,0.9244108485448636,0.7164672041731294,
 0.13447048127304329,0.7411025565368314,0.21638739698407816,
 0.8435471914538537,0.5534543316654145,0.6307756840437737,0.6074540367547528,
 0.08224130063696045,0.17701138633571833,0.0936139986497262,
 0.3603815028091022,0.1807009298512139,0.18174249247157914,
 0.02991116385748316,0.23781751481622437,0.9157163745120177,0.904306379355907,
 0.7040626146125839,0.6305169914651884,0.04634358247700032,
 0.35210797373919045,0.7479162149747935,0.5337223706827072,0.8537255184157898,
 0.9187826701900039,0.5156845940929736,0.09610215660186516,0.5388880991632178,
 0.04958057408515304,0.5798290522238402,0.9056056663803541,
 0.01712189129404873,0.6967492044899487,0.005164439770742768,
 0.9666783375788072,0.047733310730550316,0.7906338948331759,
 0.8359925019061636,0.05206584613975618,0.02221108393767013,
 0.17124386335735575,0.5148543069679368,0.772092109989992,0.12256424032830904,
 0.7909169583345999,0.14524663477173372,0.682955089713058,0.6142874777903446,
 0.4649936596422952,0.7021338795883467,0.4171121446916203,0.19861257619882966,
 0.4893554659391701,0.5311048306302824,0.4835173175690356,0.23077505488438887,
 0.46074330189561685,0.16696972853202052,0.1863603595780965,
 0.6931719133222818,0.25218702578343777,0.10104585940613864,
 0.5849251464667858,0.5602243449666308,0.4146591081327151,0.5113118752792173,
 0.9772685301161236,0.9930952804750756,0.7114756861102814,0.9699350375776055,
 0.053592403244226205,0.3098459729284909,0.2954214416021288,
 0.4819038648709013,0.04558557818040654,0.6116252838976046,0.1885134701866955,
 0.13965926068224555,0.016083699058997092,0.9955615914936228,
 0.8791797283029501,0.9820078029949144,0.09672812445765011,0.6299064824238033,
 0.4063656995300331,0.9535189337519536,0.34988623239360106,0.9139748255876314,
 0.497326415857432,0.2182658808984117,0.19647984977664057,0.8457091000094366,
 0.5194260199788248,0.9516565300078873,0.027577778672353004,
 0.9818251772178043,0.5649749108766439,0.6007149230060426,0.1416559294202555,
 0.5797109954280147,0.0290229997845749,0.8725851042027503,0.4702125633209878,
 0.27719112704266813,0.8665885221085541,0.1007147894767274,
 0.44935303305857877,0.48758913305549934,0.38673497385245537,
 0.642202886222593,0.8907388179385269,0.517508252999969,0.19808092842716685,
 0.8324970304477614,0.22315548483902192,0.8853833410421494,0.6589252216342923,
 0.3681602659324177,0.028271256992636884,0.53833989237668,0.0717265382798089,
 0.9901454919297437,0.49841615748962453,0.722038108246296,0.4644014842634324,
 0.12742409520123932,0.883121566934271,0.41578922392114215,0.5030608994027955,
 0.34978015554491027,0.5234324621484951,0.4965243187004823,0.8457881215568215,
 0.1330515080343031,0.7399985796859827,0.24258829608047416,0.6999841805692751,
 0.5861342018067351,0.4128054004189954,0.2443062066014524,0.2796639927133009,
 0.5052754505494844,0.8614619104689273,0.7774701514089334,0.8352188594974228,
 0.18229035861770615,0.2712516657323657,0.7915781663432603,0.6206495498018497,
 0.4163290281627251,0.18610523167902382,0.4334934340217679,0.7758844078051804,
 0.8160790197131529,0.7754507699021592,0.3942681943639226,0.8452486228717883,
 0.6811164848769299,0.38308626939722523,0.44408317715155887,
 0.10475209915644301,0.7314841583911593,0.6306374577651889,0.7530772611821938,
 0.333175468390537,0.9166298284457068,0.5081691799488344,0.9831025225657636,
 0.4050552887127611,0.8388560119612825,0.6045053569559696,0.2189630877159947,
 0.767736148679981,0.20676673904272902,0.2974569432127647,0.19969098005062746,
 0.01699777888025744,0.5201635430691197,0.17404869930507993,
 0.21631939675108747,0.07612717615024045,0.251653694584721,0.4246763511348838,
 0.6066308937108529,0.7668686394010047,0.19212063632859966,0.3459399884206077,
 0.48500764504378957,0.17418029936563428,0.3222333854920473,
 0.10886407248578489,0.815522322320051,0.2932275768156657,0.9661226403029518,
 0.581306549752952,0.09702825559794459,0.6844194142228217,0.543866129717043,
 0.54582880601482,0.9005764984346162,0.019698266464626135,0.07673892916578451,
 0.9907168454218239,0.5540026105934657,0.25594774784952457,0.0756594093457974,
 0.7731733883348734,0.36987033053057805,0.24438422226062206,
 0.4756984303269056,0.6573944511106264,0.21779097049242935,0.7982414871560927,
 0.7194318497634065,0.9921727312445678,0.22967971249192842,0.8478340707279018,
 0.3852647911359526,0.3282311201026217,0.16561496146191412,
 0.007098345998741262,0.7797452837204839,0.5194151154127209,
 0.5042537162556978,0.9965786128235463,0.27284012297109506,0.7462623660869059,
 0.3740578156994042,0.5400734349830758,0.8147803598721207,0.44312165092063927,
 0.4285612384305788,0.1278758754268713,0.2476320863504008,0.6254483309132081,
 0.03669768036294663,0.19303023508929051,0.15789181773280736,
 0.9672648140483022,0.3600148718037346,0.23413762357042323,0.7480941623922097,
 0.9429961988706138,0.05807556702636396,0.3500780568180748,0.5200470272186392,
 0.48841495531034707,0.345824847338443,0.9448531047248844,0.7019945868548566,
 0.22433451899355394,0.7595949895722877,0.5473842763998283,
 0.32130162973538035,0.560718806569138,0.2229487312556152,0.9896343779175938,
 0.03329789426255092,0.7025030436296527,0.787218995191464,0.11101615177316437,
 0.9492683683571879,0.48558314618141574,0.5729708960163385,
 0.01874442775488505,0.3869076445576952,0.15853107544744915,
 0.8539147670337773,0.5979253730542657,0.3523124364330078,0.3937793919466631,
 0.15763608518820094,0.21785915909077902,0.7288418246241068,
 0.14845584352076713,0.5205658478993366,0.21165033128746635,
 0.5452149390743227,0.7985607993956555,0.18246605068815347,0.0837793785515869,
 0.16356766941595613,0.7045952537440613,0.1877007961318239,
 0.17349158127994724,0.758334711290079,0.2579663067120209,0.04095114962644675,
 0.1249371809411618,0.3101629191316918,0.7430270326619058,0.04848123031076501,
 0.31800464555258057,0.27776120065530974,0.43803412228947347,
 0.563192735375126,0.8181864751767225,0.1431221113271135,0.7938390274217345,
 0.8206572051759291,0.6665091645482937,0.7895456990944729,0.5212248759859593,
 0.2123875672455453,0.37644548522272814,0.23973356403957147,
 0.7111486835219623,0.3129233618102091,0.0744892679947905,0.9557889259544399,
 0.017077648523345923,0.3413050872823721,0.31517100642897267,
 0.6385659956821381,0.3069873278740409,0.7811883621203703,0.9447425978015239,
 0.3236743565665703,0.9924159201753069,0.9850029303082843,0.205242610932986,
 0.6474664198303934,0.31495065167509295,0.7324226590848975,0.3234193813455528,
 0.6669475281096144,0.46143024087813933,0.44005524011928854,
 0.7076715466913036,0.046215639375906226,0.9191608449591384,
 0.9823809432032201,0.3779437906586155,0.5123500947737772,
 0.004447962600796274,0.25555837208398446,0.6680462989689838,
 0.3874463988428879,0.503281383008095,0.676487735003221,0.5380358014219717,
 0.4722513944876543,0.9552094202455493,0.3851759855622058,
 0.027453478793893593,0.7721331127571435,0.17386233183905342,
 0.5135765097464479,0.34534163674917273,0.8553962051326647,0.9881971322640163,
 0.8845779387326445,0.22245479493924902,0.5584204569039278,
 0.16082000065175706,0.22956352430031812,0.10713983133098748,
 0.23301897498797397,0.4209786521023191,0.5705911166284441,0.9215349421807134,
 0.833248892116923,0.20546383424163317,0.6383845323870213,0.6646336256446084,
 0.8028427119203321,0.5842471479745937,0.9050938671508358,0.3463810132342695,
 0.4310236777972969,0.6158134147156464,0.4196135962938141,0.5786497911629204,
 0.17196679800880188,0.4687675398687571,0.47197273803281437,
 0.4607650868207914,0.7480143816851205,0.19717489009344846,
 0.05474466768373465,0.9085430310165149,0.6868218865829017,
 0.16982522386013588,0.4098137495396088,0.06941866452335099,
 0.3328576824881654,0.7514265241961468,0.24146983721865678,0.6461283842940168,
 0.0861913462823537,0.5903830559348072,0.3352459763931852,0.7728558386416993,
 0.09067537344942067,0.41634683885968105,0.26104852712665516,
 0.08234589573872864,0.028010150915989462,0.6173046042719472,
 0.6720488124395532,0.16347046264844656,0.8254161726795817,0.8003323841879324,
 0.3336342593504661,0.07447003113025408,0.6947502137265884,0.7781950224620156,
 0.8856587996986408,0.8908665966615353,0.9139988644645491,0.862166131124877,
 0.510833140775363,0.10677892087496565,0.710832480316196,0.33601985310052984,
 0.5511509015230681,0.5793649043835152,0.8594375824441224,0.7097772178796462,
 0.49591979572357836,0.48500589410485795,0.8106601098987758,
 0.6338012000806177,0.9475009100411274,0.023725200084153575,
 0.7095107653279219,0.35493413338741553,0.10116854244003815,0.85505281380677,
 0.1395988296243048,0.2512371003519419,0.20934817319741994,0.884242892381967,
 0.7424038972639775,0.05341729228684955,0.7724796204704081,0.8959107248822339,
 0.9202392833977223,0.6608692451839641,0.056130010010864484,
 0.8584518339801932,0.3593382668125147,0.8192625865192825,0.345640041806794,
 0.008427603755740476,0.682692854996217,0.14755277742620976,
 0.7252699298378378,0.12946822382904122,0.43967398756363696,
 0.5560707285467944,0.05617273655393473,0.9550429702036216,
 0.12149787688954572,0.5936873083989997,0.6858292206745679,0.1587182051408692,
 0.7586550108134802,0.6162610562400843,0.5642537087971904,0.07558982520701807,
 0.8276085115296717,0.8361582590911765,0.9399661441390239,0.8399060086427342,
 0.3237441639369433,0.3893904061825355,0.054633176225438396,
 0.1890510731531132,0.20988779801521518,0.11288829623474184,
 0.6274803401716333,0.18432052274696986,0.47678092795628557,
 0.9600802271301109,0.21641524518245558,0.6604031096296449,
 0.33640842369074997,0.5348266498787227,0.022669993908047137,
 0.6926112679862743,0.3581920810280995,0.16206294744041538,0.9275547921919536,
 0.193953377115488,0.5099274262875297,0.7367357296282946,0.43517934524393775,
 0.6587035935770575,0.5058308268847069,0.8463528233682822,0.5677850167393221,
 0.34060370015724784,0.26782543430468286,0.7964249282163995,
 0.8378013927705015,0.022229114877215128,0.7525466369603077,
 0.6510854809734009,0.29259341922637727,0.30663766477853616,0.658124288010307,
 0.5703678810827466,0.9494443903934822,0.49175577335298515,0.8785134476453546,
 0.16401548795529408,0.3107095878065762,0.1914231802568782,0.853748782741508,
 0.002334840774202096,0.48245601946880357,0.10533405851217958,
 2.3572338962807304e-4,0.05840177559983839,0.19390464073032743,
 0.05840934358881589,0.3048451969896001,0.869668619705976,0.5528857855105116,
 0.8782261146726827,0.5449586665522882,0.5290146948692432,0.6870441546507207,
 0.10812908408995159,0.8587498964538148,0.1954386716161025,0.8182146437287612,
 0.8342791297775835,0.807652305257167,0.7916520358656319,0.8887140814390505,
 0.6026821919475408,0.3061885993935296,0.4240217377675357,0.4469843866081288,
 0.4453241188678453,0.7721670762002137,0.7495088404121596,0.32996740522225654,
 0.5558342298927063,0.5744450652548362,0.016968862253933237,
 0.6352862471767867,0.36021281883160494,0.5199085478695306,
 0.32050235095314505,0.22333058605626266,0.1892275724793192,
 0.7077652930123276,0.6778355287350982,0.6641229060935177,0.3979934764528251,
 0.46750085648241724,0.9072526719111835,0.08638413357261854,0.691616645999942,
 0.8764265328586903,0.08458920763582134,0.6258452372065919,0.5534006341183612,
 0.5023808803391125,0.021058593102242118,0.5430615690574927,
 0.5576297839801663,0.6479699115481846,0.6451583882198779,0.28437185237666807,
 0.6627866622435652,0.5628366317318907,0.6949396400788892,0.4377742205545303,
 0.487154105115033,0.42684241996263106,0.22425846960678175,0.6877236606162411,
 0.9441396361394094,0.9085783566739067,0.8960128274803536,0.8814776634139208,
 0.09491065654890352,0.5523493667140175,0.03354972089200492,
 0.6551139317000485,0.8029373566543101,0.17277918903393397,0.351102742364346,
 0.0024500970585126236,0.15199594297054808,0.04391860532948977,
 0.3291323284692913,0.47739496767949696,0.2639820503980336,0.7127505734841397,
 0.8472534401013921,0.6458069133777291,0.7105076531794224,0.2967150662357817,
 0.76918606841857,0.025241245639522458,0.35359795620488255,0.9221375818346452,
 0.34594348486584736,0.18865989300482489,0.63626872102876,0.3344621803430794,
 0.1861769577718867,0.712248013538759,0.17375128233630122,0.2739470107579779,
 0.919488169648696,0.5298466022656934,0.6623702038859339,0.8855588536936265,
 0.9819124549833814,0.21208144099021087,0.4421377494706993,0.5550807239493575,
 0.7191686069085367,0.7470796316566101,0.5808940576026433,0.9844887059981261,
 0.4010747217497226,0.360365581811372,0.5855501522359452,0.2015734008693455,
 0.18448404476731362,0.18641125119280155,0.8257108343041677,
 0.9386360990891618,0.9619681923031377,0.7128579024735225,0.8211349752274792,
 0.21021167635456128,0.9097763570638139,0.05807174784671554,
 0.8007967830208058,0.2334535127131282,0.688799450209399,0.6813910568197659,
 0.5996141036274542,0.015725357971647536,0.7424547548381677,
 0.6599211849099755,0.8104781494578384,0.06506141408359545,
 0.47232106141465713,0.6588730418571098,0.6197645348002007,0.6156104571495851,
 0.10017315942985117,0.05119866189355882,0.00286249362181612,
 0.9897449278215913,0.35275755740463266,0.5709646917554996,0.6300638760628774,
 0.5509050702412842,0.07641967425353613,0.28122854499476224,
 0.38170214021120963,0.8979016477514001,0.49090601452074334,
 0.16953873343091053,0.24247894001582637,0.7017298448789849,
 0.7629314965701918,0.1935180434712721,0.8211887654790655,0.8248119555221597,
 0.37058564595390264,0.07353027358818354,0.4370632250264671,
 0.16175953720481284,0.06276729454776675,0.3129781609763528,
 0.6489957397252172,0.6456794132889196,0.06602851740239413,0.5703319818906116,
 0.7959009927639205,0.9217139477842827,0.3851908508042493,0.23908909322384464,
 0.21899638256425025,0.97166796852903,0.8066759035092613,0.30445587739571145,
 0.11200653928838755,0.6487103640806073,0.2722633383998706,
 0.21803164443497414,0.618181810249852,0.6198094738361115,0.34381497089138424,
 0.36246275014157614,0.16289928952600674,0.615289643001072,
 0.15670154623696086,0.48259130657495763,0.6957227894601163,
 0.45690849335452444,0.791208503236085,0.705359155066169,0.3750219437290918,
 0.4186770454968677,0.18020587030842627,0.4899441564725515,0.8138372811746174,
 0.7542117349534678,0.43133709036674694,0.8378273569882987,
 0.21029499379301764,0.8639078550917206,0.5345372812713816,
 0.11404411495518474,0.9300471327459583,0.6405786641903098,0.9507420170292604,
 0.6740528902430594,0.628637220416926,0.834816145311811,0.9868239224117586,
 0.5150837325078088,0.147491987038589,0.7772739731309397,0.9385186697043529,
 0.23029742576724788,0.6100668495575184,0.6196023451167333,0.6922777883159993,
 0.4427502833368646,0.6685792168347091,0.7298273164088707,0.5257138173923879,
 0.46278199159191014,0.9174400687607998,0.6218321369214861,
 0.40630028310886523,0.4278957571218531,0.8266098777115624,0.973677714953485,
 0.6947159527921123,0.9754327940929722,0.12620144949846612,
 0.13442543033672816,0.8329989417803091,0.4894632174574107,0.6143635692107636,
 0.7866458882508779,0.47988904951478006,0.6735785704625338,0.7656610986138774,
 0.5706204021459567,0.9325544680971724,0.8032956376467218,0.723741334129049,
 0.584202725087863,0.5058471992851405,0.8280218927290949,0.8078110156017415,
 0.9499460459385065,0.28642006429447897,0.1723059652020441,0.8029178961974476,
 0.5090701509443037,0.5788752875267206,0.7358635901605626,0.4283948366125968,
 0.03363010214883788,0.3814250222321135,0.22629492800527784,0.847591658551185,
 0.8853118373806199,0.9529645242543888,0.2812777371127897,0.01763082057707943,
 0.5978749614212786,0.6138426079926004,0.7439030370698052,0.702785092747092,
 0.10207728950223449,0.15935820406584378,0.1634314895804636,
 0.0725780993079087,0.8631545930796551,0.3784368815112408,0.2643958452241316,
 0.8488189655380115,0.07807485585697782,0.11030433711785048,
 0.8611025834020236,0.1835197865035978,0.07677337085805858,0.9910197527209093,
 0.1689080874316855,0.8002300440422419,0.21183633720922446,0.5943424885026507,
 0.9080508384206427,0.1632016671593668,0.3262407263649745,0.8180428053653437,
 0.7475923007071522,0.13621051957092523,0.9091099043297567,0.8757333151615329,
 0.8647016216077775,0.6520758710997523,0.5403423815872891,0.04014589008262914,
 0.47006154280920165,0.7358602865946096,0.8341362737045696,0.1650019412029341,
 0.7716006080657574,0.45048042610778216,0.3264293198848611,
 0.20800950612021807,0.7812785415609608,0.7010326085622226,
 0.02669096202348098,0.2566534864968304,0.2954883619671196,0.5145336275784241,
 0.2491358901370322,0.3805855526730263,0.6404593440809858,0.0536059002144873,
 0.7773160965398046,0.025369936118224468,0.971373632269581,0.5879887455891634,
 0.3345296055797672,0.31802293636831924,0.001910428112672613,
 0.8565468461645388,0.8089840393748037,0.605771476168505,0.8603079486889117,
 0.21969016507752892,0.8713754999933381,0.285588207640576,0.9830763861529921,
 0.28524396174003064,0.5778991038646446,0.33486179117580506,
 0.8712535766780468,0.043665061034011576,0.6255303118848924,
 0.010450210187728515,0.8998803364242961,0.36492118410445085,
 0.5500128692032566,0.6180182492809189,0.8279511172145657,0.4578846826969549,
 0.7532610535718651,0.192027940286005,0.988313009358854,0.5631024159171578,
 0.5034097964898231,0.3054605754705726,0.3503949303178008,0.9783384833352587,
 0.519242057837644,0.853192769704874,0.056566118102452156,0.8174584250439777,
 0.696113063801424,0.00925469507673693,0.5543260901161791,0.1319788488621555,
 0.669654087720597,0.7901604657327763,0.9537979071953508,0.07884626269121076,
 0.999388547778605,0.7794465135409329,0.35249174912853143,0.2523107706385973,
 0.24458189207727277,0.8839297285243963,0.7880230989579702,0.6753711774267682,
 0.1951091475846003,0.7493158635802449,0.6938278022086283,0.22581834251168176,
 0.08602592080568783,0.7138909951899128,0.4544580131352771,0.7107688380054275,
 0.8115866019374516,0.3133633050772372,0.48639516251718906,0.9159936242020403,
 0.4265386228239523,0.18191082588810337,0.13771854994249033,
 0.7053385780604995,0.05464226394156935,0.40398399388097384,
 0.7295066954611114,0.11486286045840066,0.0468890166685485,
 0.19760917609028927,0.054990819563445203,0.10585379617737378,
 0.26807050573233815,0.914672776494631,0.645410370850905,0.34175357567929887,
 0.6331841205013186,0.2665017229446486,0.36228162039179657,0.7306529163596891,
 0.038372396794927544,0.6270165818097361,0.8854920206924203,
 0.8332181466833897,0.9502637966286783,0.3861071555747355,0.8371333545379481,
 0.4992553646215382,0.2094746713606842,0.5335915910917369,0.16340664589349307,
 0.3036203014510541,0.611910821404088,0.1601486394042524,0.8543492270377581,
 0.43052887769404347,0.8349465797312273,0.23397881464695658,
 0.05601512992613522,0.36698561127149565,0.01746555591989951,
 0.7152945412548614,0.6543513922237291,0.2461785524938387,0.30207471200806246,
 0.594172377441663,0.16458075888543033,0.9901247733941528,0.5379730937636675,
 0.2651306794813162,0.38976978294064013,0.2589261662086423,0.3589926189402288,
 0.7346536081601578,0.8881086321696656,0.8894621107530147,0.32617912820100947,
 0.25715468557654886,0.7779393484020813,0.41505004478653007,
 0.8287817389516362,0.2354399654776128,0.4695490512441285,0.3673813248554907,
 0.8420831849910693,0.3321602823962878,0.2720282464899735,0.0562627822979177,
 0.7185589423259909,0.05129376654056417,0.9345318545304628,0.8706517068522035,
 0.20612242896182675,0.36426886796264124,0.7893089551182638,
 0.36186941499449166,0.045223194197454686,0.9367418715306421,
 0.8742838269446449,0.3768412197842058,0.7063010314286283,0.26673695585796753,
 0.0389363520859588,0.5286992940607462,0.11460889881166336,0.6206940325975867,
 0.7365492135663634,0.13627375229801575,0.7970428849957804,
 0.09233215298978703,0.6586647116893044,0.65665301073773,0.22029531253517343,
 0.6527528817219769,0.34683554524871374,0.28148920816190426,
 0.007298960356578399,0.7294741752780648,0.5125764538684949,
 0.9108521045537414,0.9835747163830001,0.4432292916315663,0.6767575961890019,
 0.94645331407219,0.6650656620230545,0.28013813439727553,0.5129914118177671,
 0.9317534498970208,0.9853328310830828,0.31498918383323415,0.2585651927869643,
 0.23271921144562702,0.3465576275206802,0.9009612923226202,0.5508469299625159,
 0.12165996720631533,0.18737269244389376,0.5182661523727131,0.945355570852831,
 0.11841152367862606,0.8906125920676675,0.6223969854086812,0.8102796699700815,
 0.07659554755024944,0.2698737172332184,0.6164875169712167,0.8911556146661592,
 0.3374802674124715,0.07086767614140588,0.7091834911449927,0.0622574053745365,
 0.17124054213497752,0.3100220006219314,0.10844476209188092,0.685221717147569,
 0.256084526864816,0.8207262612777656,0.7660622878961179,0.9380161448301627,
 0.44249655331336435,0.5365762783670389,0.677439701543949,0.6203984456949725,
 0.28943267916870963,0.6085303955269339,0.13112123161055977,
 0.5581203614519137,0.9768635309289695,0.18561927965316594,0.363129389022372,
 0.004777923773490289,0.1784743701507091,0.09405352692396463,
 0.4376069412137321,0.07434845793010925,0.6662037732284604,0.9514045320226969,
 0.144831327759251,0.06553140005507907,0.0607199793636376,0.5109151818805767,
 0.4150259116919902,0.6532531702817465,0.6149300859350357,0.8997294138450834,
 0.5552139834392142,0.7386732782034802,0.051589051562494936,
 0.9129868047710122,0.7498441332124696,0.015018474659866832,0.770045770282117,
 0.10784101168186477,0.11212544132355662,0.25583622272332185,
 0.9899940754600131,0.8594534204903471,0.191773997205124,0.6704318137433232,
 0.2378034611092389,0.8431741978446137,0.8659268563570052,0.33602634983119584,
 0.3455184887772006,0.5512973928788278,0.8226375628187369,0.5460561265606187,
 0.30026807215868356,0.6612353962005643,0.05319908584929012,
 0.24569472343270937,0.5428780947178923,0.6685473239764697,0.4264553487755355,
 0.4005176312331402,0.021051750058569962,0.43755037098630245,0.73774268859926,
 0.7018972444949368,0.09924782851198333,0.6040850466694102,0.5795545624372065,
 0.39960695794053724,0.3499913812537063,0.3177012656805345,0.5958731718601467,
 0.25676532311905453,0.6762785911275444,0.7853046657389964,0.3385229892474715,
 0.35434976388804973,0.280537980635828,0.33972161749206276,0.8213616717761804,
 0.13685279313134635,0.20890429825543722,0.6452844658936268,
 0.10198775145595629,0.6048092811018423,0.5776026003942877,
 0.07989882455808317,0.840623390481719,0.3544081046126062,0.9194698327204262,
 0.34374768638726383,0.9573210036474235,0.9886131779314051,0.1323846443040957,
 0.28766097818738456,0.22476392888323027,0.3739125175654747,
 0.8099261037033032,0.2597397081958952,0.9018457804556672,0.3113856629154951,
 0.4824661273694242,0.4001113747205979,0.7530082055068426,0.1272078572171349,
 0.2609672087821948,0.8954026742921614,0.9962303353306384,
 0.012107748449169509,0.2459255646517784,0.5199905133013336,
 0.5981160687475042,0.5710717999012258,0.116682111173249,0.24757995153136236,
 0.606601620244574,0.5938747679090527,0.29045400900234586,0.38320369869179016,
 0.41468237138033337,0.8763534454414601,0.6533945648937483,0.495992810903661,
 0.3290824676263987,0.7778866223919536,0.5308405366567768,0.6320908864442185,
 0.26237585573173094,0.44652325535606385,0.9093558617251846,
 0.8174664424501809,0.5010083487193844,0.2419186063068447,0.21846669794446671,
 0.33466759473783925,0.3449322034450093,0.12834019674660158,
 0.07898785722478563,0.6454919381178634,0.7507014786085141,0.9767937427806297,
 0.035806567706298975,0.26697312589428734,0.9709716602351954,
 0.3835293119525415,0.9828073586049243,0.20795780146250809,0.6007887907094642,
 0.6548501370360901,0.08189864363534127,0.310902961852791,0.9253267898531253,
 0.0017061461527549948,0.18579097470790495,0.8492834730900954,
 0.31819462352705563,0.0579026499918307,0.4777521606286692,0.8230282274172652,
 0.8557893891499024,0.46158743432593696,0.728975711197989,0.569010498933788,
 0.7264715651833767,0.10589014069012548,0.001493490351885396,
 0.255351355438754,0.2913266228607899,0.18997694674452947,0.23270903785118857,
 0.4124260026890958,0.9223092583654866,0.492961136340196,0.31953173001189883,
 0.19226896949429073,0.6166951784541288,0.9932632461151741,0.8554763065723207,
 0.04249198840149604,0.04513778141412972,0.5621238100497513,
 0.8269922182422169,0.15444822308677164,0.6113684439542851,
 0.28767393986994605,0.2375565524004848,0.56251182029548,0.5232252496048626,
 0.3590316054361118,0.28049146195549357,0.1342603849473003,0.9686417372813011,
 0.7482509620847473,0.4782903576174511,0.4374114731881803,0.858529666055084,
 0.9959022040763567,0.6966517144996585,0.446471426525626,0.9158386606988439,
 0.6745857525817304,0.9779720262403426,0.253890152633898,0.026375228701814524,
 0.20906936311840985,0.9594881921092501,0.40013491001867274,
 0.9600737318108643,0.1374143705857377,0.09330818426374154,
 0.14609774572750134,0.24195025162268868,0.8126238936813699,
 0.4767372751223997,0.09026945874063452,0.501509824018395,0.17956050951227986,
 0.2818462945848852,0.617968915710169,0.5847409565471313,0.270177162917401,
 0.0025737803468611764,0.1692490357020593,0.01789222036385818,
 0.2540500660864027,0.8820214494241934,0.1737246368138754,0.6309465092671447,
 0.5680591844616355,0.43360550779276463,0.18069360257154765,
 0.1983477813777963,0.8130610110341943,0.4598058040256635,0.9757318689258725,
 0.31380837620462987,0.6788532268613684,0.5949467039948274,0.818940086600168,
 0.43842519733861585,0.8522879892753533,0.25349093767944486,
 0.26942858986317586,0.9934515053954471,0.39638530245877335,
 0.15305854882240055,0.5603481165492866,0.1114256522153364,
 0.009847298443493746,0.24943778117212945,0.8964320522292158,
 0.9403238347494335,0.6698606305797876,0.3670678894607191,0.6830038251081478,
 0.33874052361561424,0.6659731653956262,0.984315759782338,0.22352234962245998,
 0.03132633507900795,0.9625610564490565,0.11672969552041979,
 0.3432274372282609,0.7757789538137889,0.0040498301468394615,
 0.9250196800236057,0.6979997167922388,0.5927832111454264,0.1950265756046493,
 0.5666963515370584,0.3625511322116558,0.4564878005992714,0.7728121310933167,
 0.17840729566064428,0.06365984452839768,0.04198077912016496,
 0.7914063321580967,0.4291696709292565,0.8855878071844735,0.3809877543163477,
 0.775129044366884,0.21374997379989014,0.3326034267981568,0.8777015915648,
 0.673418655587068,0.8569663196469807,0.9165378639588627,0.11605360770538797,
 0.4167486263094481,0.5053947246273061,0.30761190334636535,0.971255982884649,
 0.7883206818738592,0.7217765576920635,0.6637404164010287,0.8904722809476247,
 0.18050467964158068,0.5493021311871762,0.8788196582691595,0.3701404229156814,
 0.965931153403651,0.5308145510126846,0.7471868622640456,0.8651501505059647,
 0.7104373014619368,0.4317587308332884,0.9574383939254048,0.7409926182674811,
 0.8594410202475675,0.35140553518775786,0.2957978678438081,0.3325480360597226,
 0.7849483885635925,0.45044143186601915,0.4776242742183646,0.9967002483199482,
 0.21039147598287922,0.4069429157580018,0.3205725035844982,0.8041273743286371,
 0.6212731615598841,0.30782303181471105,0.7747891575226845,
 0.12280567565009204,0.21449273893342602,0.9704603394778948,
 0.5855271937867279,0.8768349608600492,0.7430406996804159,0.7592475850786833,
 0.46423399612277394,0.5367764406350569,0.9895387226574035,0.9005747171104359,
 0.9706417608564654,0.7619353553266583,0.3902741997399808,0.13724021807379905,
 0.5254308569183457,0.36509454976672284,0.9244671603898603,0.7486759993314221,
 0.7793187828556962,0.5214006476365074,0.21053807003370362,0.0724130443335731,
 0.6988311289553231,0.17893828972716586,0.49315839418912477,
 0.14232560652316983,0.34364796432112577,0.1287620558734398,
 0.8939986868193266,0.1514286999836567,0.5311243005946131,0.734540353512449,
 0.2749753784048218,0.6685618790792037,0.15565108097852431,0.064644156357058,
 0.6924012920585606,0.8043504025743932,0.5305230704003905,0.9640892440660876,
 0.24457796525114606,0.5973471608816964,0.21103362851331753,
 0.7197441885313881,0.1979528098089649,0.8363338424145934,0.29870834765241006,
 0.062385055164150494,0.7943628503324351,0.6917975954011342,
 0.5563998002247721,0.9812816571242327,0.4456521776681621,0.9468578477340719,
 0.18028719396841264,0.5683260124265427,0.6237489939326827,0.5811990301469128,
 0.025777860352999227,0.06157105012860342,0.4046943462004645,
 0.3297231197338706,0.7732147585986763,0.2754995965007795,0.5530592547388764,
 0.35623751613807564,0.47746661795445744,0.7028331148147794,
 0.4209405896942786,0.9910964243825826,0.15324725415553042,
 0.35931183245081155,0.08390212029358257,0.8124236943832477,
 0.2774525490761476,0.4281747774612379,0.7208549607784934,0.12285399736954694,
 0.9402643078597086,0.7614764337007126,0.26580570294279837,0.7348782731643537,
 0.28564455425656454,0.3496286833704926,0.350768849851233,0.7874088434711601,
 0.5866140211426237,0.9724902245396816,0.19067896740518253,
 0.11366021807768145,0.7991571615017602,0.6106597891262471,
 0.15153642955067514,0.45840323359339086,0.6854138144486582,
 0.08141899764026483,0.8439094496031797,0.6831968293901343,0.7157017077304694,
 0.709563559854973,0.6545022609040585,0.43725855431685035,0.6739213166085704,
 0.43864550971642047,0.006122802061192645,0.1826426421009727,
 0.20776788251937783,0.011631008430298961,0.5936673636132491,
 0.42230750014202934,0.895562014860524,0.14484405218253948,
 0.05071064009999515,0.22290344213277613,0.9514328248407367,
 0.9712194972628704,0.6377947070863884,0.653036740400534,0.11065941377740274,
 0.5489353904206093,0.6907763399967215,0.0455960434958671,0.17750848914913342,
 0.004488560406710529,0.6886978913373072,0.11331227873143657,
 0.9876935583153132,0.6435031196605071,0.37175542313344623,
 0.24731196658979637,0.49332710253759027,0.21693877181229004,
 0.6216908471344381,0.554772088201487,0.11231362612647888,0.2898150315174173,
 0.00917818277870297,0.2988298901857134,0.10526096166394705,
 0.3903663797972927,0.2329225813842387,0.6525043420638624,0.5382632206553143,
 8.014691395119566e-4,0.8891814381496637,0.0699911352415361,
 0.2798656481437579,0.018377950493749684,0.4662914307203232,
 0.6251458548483584,0.7113395170113663,0.7043734512779577,0.7533853123072046,
 0.06955755841008204,0.8960677567141082,0.6740878888464685,
 0.21419491793872913,0.6915041566318629,0.24763871390948022,
 0.8553379726130048,0.6716511485733366,0.8087945673327924,0.13080933507425874,
 0.29075735702827965,0.36043062818818694,0.9825542717002742,
 0.8751485364448668,0.7521674109387824,0.1553500517929337,0.5193408441337884,
 0.6387015654016437,0.46171847967322543,0.545117755620349,0.760649356353085,
 0.16209231738599172,0.4659366120139925,0.8650672094081995,
 0.11232990927045494,0.028128225903762627,0.5638385238011311,
 0.8696444217723655,0.9360236255218078,0.5042607436167352,0.6428022003212572,
 0.9401955341421058,0.5116127216889533,0.7297328233183393,0.10440900262128172,
 0.7795431897719887,0.12616789005418316,0.8528083626431449,
 0.012681833347266469,0.5009990014782788,0.5580760126685832,
 0.7401072598604225,0.725282027308392,0.5045748331350888,0.047009991482634406,
 0.46581439007611625,0.0422116657874253,0.8878586993883574,0.9814045583250922,
 0.8152117347475963,0.3998220892212153,0.6732450192727669,0.2864195708993089,
 0.7648869661379242,0.8769131128224583,0.19307651580077123,0.9528031289756743,
 0.48860977903884584,0.3277425880290068,0.8296067389005971,0.5725050506931547,
 0.612212114075907,0.4902727400834056,0.4448161824135719,0.20692116063163102,
 0.9229803512722439,0.599004529915313,0.5174726861433525,0.41417148992495945,
 0.6292118513561038,0.0539502322308445,0.6629112246354663,0.7229935573229105,
 0.129097282454812,0.9731529845516542,0.9793983996661303,0.9986026972211968,
 0.01575090723813166,0.06667713532529596,0.6193483734428188,
 0.9802898822625526,0.8833584225648949,0.42159431699867234,0.7772371538405176,
 0.89377454152076,0.42273733466143393,0.3327109536881292,0.2101512788367863,
 0.402548913438769,0.5406401036042706,0.015406019075306476,0.3527614648328332,
 0.29905666347598125,0.4681445923689844,0.8845423122750329,0.3148953875401119,
 0.0772783715440104,0.8330440255106802,0.16644134836255786,0.5971766768941091,
 0.7955977042027162,0.6545087238796808,0.832591189974524,0.7605047031286468,
 0.6970963320770969,0.40813232571467917,0.6844698022296613,0.3631937469228125,
 0.6126760185576545,0.19037540581687173,0.9229396287308337,0.7441052239225916,
 0.9578332430288391,0.29434918056805326,0.07155291298901223,
 0.9156378021373319,0.49519478887215074,0.555347727281987,0.24498472500894108,
 0.17942513820787465,0.6165085895752392,0.222531134119219,0.35358834727753985,
 0.25691526250464025,0.37916400320187327,0.45820081481050323,
 0.7582928641886932,0.9689494992840803,0.8588280942174086,0.28818871423036574,
 0.23668837713628588,0.5795107891174186,0.8343101774893978,0.8977390880086242,
 0.8438901562009768,0.7579182088564292,0.5259097438313807,0.7302343733015598,
 0.497211820284289,0.4232807272167004,0.1317058152716648,0.7724758807703532,
 0.8515963470449966,0.615842776395184,0.39920333487608717,0.5701262253940601,
 0.851676572205483,0.8320438881671316,0.28268321482947967,0.24752596675272132,
 0.3857418227829722,0.18857619815513227,0.07903336543929118,
 0.15672279157015812,0.620642034158851,0.6678613152442998,0.47543273477640424,
 0.7880396277748876,0.8774395418430736,0.15231684965745684,0.8813191387774497,
 0.6114145367034839,0.06888637468793213,0.4012794693519647,
 0.04303870729413173,0.4801172872614389,0.4807709484274828,
 0.20403253958635303,0.31197522997818716,0.5677697021115862,
 0.4754253276775773,0.18373684771909882,0.04465435337156298,
 0.23006534856554173,0.2577440393755478,0.11822864212782547,
 0.8656819261897571,0.37562348831106296,0.6739199107539551,0.3326322807000046,
 0.10880147567982956,0.3798434558031025,0.6320470921583079,0.3595134601954868,
 0.841877390660469,0.319819046693808,0.9308880515434587,0.16192643510439086,
 0.6524089501825259,0.36720904706378843,0.855102244815795,0.3189234927090767,
 0.1191269861944515,0.572889204177744,0.028856994901997113,0.7759189974870255,
 0.4692687808620949,0.2723673182073476,0.7184742039985156,0.13834851134847492,
 0.14264425588524687,0.9112039744424003,0.6059175228637003,
 0.34594309566437986,0.16406171295667304,0.292940982861872,0.7332759034542038,
 0.08052339968572497,0.11780270414121707,0.9150076122779691,
 0.6400152451569656,0.6478066911447675,0.5098460271264932,0.707502321568387,
 0.3711788325546732,0.5687884917496392,0.3924970171327743,0.06381302460847027,
 0.47865764740098526,0.725668705637158,0.6199187896030292,0.6809038198213362,
 0.8800192635598282,0.8449289792954617,0.3605213865035659,0.9269515203903326,
 0.0022268914831383713,0.18160272350178286,0.36268083379634675,
 0.8288097655323617,0.3149480106207013,0.6864214220284164,0.05948942203782759,
 0.8659644960071113,0.9180249123259175,0.137104820028095,0.6404356773526846,
 0.2820301709603482,0.9565863780144558,0.3097153593832147,0.2823291137529469,
 0.630165574685689,0.10733137604114673,0.4872843447820068,0.4881284449450105,
 0.8710002252812121,0.39774170164778966,0.133968163678434,0.41105584321403155,
 0.6306266229468735,0.3378653441031315,0.8880371893852863,0.24038662333329353,
 0.7235565291810238,0.39054555261449597,0.8133188915005495,
 0.15593653016187858,0.042362514434510556,0.9351677053766727,
 0.4475058716244861,0.5111176772014541,0.7656251284080685,0.7869906182080826,
 0.36888510199841673,0.6683843061556661,0.7220671987251279,0.166691275466341,
 0.9181965974878236,0.4246636393723705,0.25156331994143577,0.8621579503729971,
 0.5468233540708409,0.2556531077138511,0.42377014182583617,0.8015441453392222,
 0.9521317351394587,0.7424285743298984,0.9434187093055015,0.05318265725627169,
 0.40689254998136615,0.0294316517074108,0.4772135467435331,0.8181691447435401,
 0.42357680895371974,0.939857152734524,0.5998193387355996,0.4845270337129899,
 0.8573382749821132,0.09730562584908742,0.6877284052429437,0.7993876022120951,
 0.9561986348391999,0.9060019744596206,0.4928978244098111,0.7467300522900056,
 0.02261003831724484,0.5634342617979927,0.45822275111763855,
 0.30385527605217455,0.429398976877009,0.09826806243484143,
 0.47582586874867694,0.5257756190425191,0.9987618989029857,0.4326847409619886,
 0.11550369831685825,0.4326111736220746,0.2858147006041758,
 0.10173119062488034,0.8448791141208772,0.660108503379961,
 0.002321908458428279,0.5942072677530899,0.7753808884954075,
 0.7039951422606258,0.13928377516419765,0.7078153331080316,0.7489609920325424,
 0.658317701384946,0.5299609656109672,0.8300387639141993,0.10823627328561403,
 0.9236314336007946,0.22085191077240163,0.8281508029435607,0.9285463634571346,
 0.4009365244228299,0.35638817069839845,0.569118450428872,0.388133939736504,
 0.36204660046963977,0.6282478496124992,0.9953386659139513,0.4166645783113737,
 0.8217860288084509,0.013153712872788437,0.4428057339664728,
 0.7664866729658921,0.0054271507917385,0.9726737949859521,0.7378645674738322,
 0.7912970350030565,6.423287918699838e-4,0.5461673796044961,
 0.5121050194095599,0.5142300777178211,0.528074085446668,0.15858194249592072,
 0.06279674947739289,0.2847434722699258,0.25931607688947333,
 0.3114189757257939,0.8208443519675638,0.25352920052038375,
 0.28810649601140226,0.4813181840523606,0.8822543513859226,0.1790340664961888,
 0.31713656330216433,0.1766664568842493,0.09108978438133497,
 0.3916976423610443,0.5911531120076479,0.47160000239224675,0.1876805823627078,
 0.7720580664193616,0.3311167013442373,0.959471922329996,0.9632670000607004,
 0.959622128543088,0.9510650111121216,0.5930817665975232,0.7112552053004887,
 0.5852227249504715,0.8838554287499335,0.8374577897067039,0.27871436103137814,
 0.15605513438388174,0.5547886774652789,0.9366837434449791,
 0.10185824303763147,0.5217234622093565,0.9632967749404885,
 0.16532248666274085,0.31522790683353263,0.0031515957056018618,
 0.544538426155795,0.595995128744925,0.7409413376021469,0.2063673588299757,
 0.3065598251620013,0.2850758944371313,0.4553841229856189,0.6582322361160963,
 0.0830224073995749,0.6315682399398969,0.9248151356672489,0.9129114705395969,
 0.18391221105979827,0.2599497556624254,0.13413071731475645,
 0.5814460702431213,0.3677976521306412,0.75830822449859,0.41542728880685864,
 0.20108887741752812,0.053007625406898295,0.9851817085169505,
 0.34951798791475785,0.13692971213096738,0.6186523869837616,
 0.37594321673543596,0.31294045973281737,0.5910594541356582,
 0.2104238947508298,0.33218118524996276,0.7177464704063511,
 0.48872203251577273,0.2295427956618018,0.7469758503757564,
 0.19745888405945822,0.09484773012972703,0.6226546707590173,
 0.18638816825935978,0.6281444761393653,0.49042578103748424,
 0.09105535565307332,0.49565223620600896,0.48148172730788175,
 0.18368849030395928,0.7405320153406835,0.25577177881505164,
 0.1445766688327833,0.004316505093829082,0.02356915229653389,
 0.2435283182336475,0.9177556205807893,0.6198659816206572,0.3621119973361143,
 0.3113116262747082,0.3335470205958978,0.8095991943479374,0.3692172482784847,
 0.5672594257382839,0.511549763966615,0.8104500184385034,0.0399217032469551,
 0.862101873668456,0.7088391609084241,0.3203303148570871,0.49785613668235784,
 0.48681490842275277,0.6947051994805004,0.8681043262929695,
 0.09418007525106431,0.18390628670971487,0.8047059725418095,
 0.43740843442648325,0.36759116798935754,0.30807076241256004,
 0.1774689334469195,0.7300892712892908,0.5339291373347316,0.4102483567335118,
 0.868029638179945,0.9360909374371376,0.10564144003179965,0.11740638057844077,
 0.7887085945470764,0.9261004179901522,0.5649639909352617,0.7656073152102625,
 0.7671805268409744,0.24655808677568736,0.9637163199900881,
 0.37621750054628267,0.9669544227824146,0.3141544730280269,0.5382806008687133,
 0.4366160463833699,0.4109169612838778,0.4190823491422242,
 0.030908316951881787,0.4851177090102793,0.22537846427073704,
 0.4885067748186851,0.7673771175850904,0.29963085476180384,
 0.019584487558833574,0.42481677551151875,0.32563970433207057,
 0.28633887989542295,0.7459300515520545,0.475157559521878,0.34683239544180766,
 0.3238304293952067,0.21201655616826132,0.08628353603693972,
 0.09405280404618976,0.4859909495301965,0.6321381292405199,
 0.26178183803961785,0.7884450278553062,0.053737490444950575,
 0.11676207763428437,0.4591265920066169,0.18302330987708182,
 0.7950773842510934,1.6684658171284283e-4,0.14598050559707598,
 0.1047469307442842,0.5007171675377862,0.22159630052659196,0.7524789256048309,
 0.6727019006805051,0.2612111653797696,0.9285960362808686,0.03825362714552305,
 0.18365914527126548,0.9409696621899261,0.523427562894423,0.37006804103717716,
 0.919874329245227,0.4127808947297572,0.45835337773132157,
 0.0037344549875212296,0.19400165234113653,0.918307251732509,
 0.3955032050080991,0.39454582245589187,0.7635028213612244,0.8806587354248654,
 0.11370056156061248,0.07806339845516508,0.2380196377279401,
 0.3352715583049153,0.5121430098154143,0.8281984272145001,
 0.008499904034899863,0.26822650930729386,0.8810748000322474,
 0.9026206227386202,0.3403430103554115,0.08467626736957012,0.9135839224624515,
 0.6773935990115884,0.8536014170044671,0.06329637516954256,0.7346713309942603,
 0.3288868218715836,0.36832885079084454,0.6385388932129991,0.7086457601285786,
 0.6535945442636775,0.6991138639681508,0.6967595166763424,0.8000972533337657,
 0.4518201251474061,0.26816364126124503,0.2730764002086685,0.110409900685558,
 0.29185436413289534,0.317406638604778,0.6979873742946122,0.9862899670279903,
 0.35238450158509316,0.1497734779528751,0.291159303967065,0.5826022234501057,
 0.5357563772081231,0.36138713721345084,0.39531835554409867,
 0.43143117821020494,0.4858687446289689,0.027158127073944494,
 0.6982600569270188,0.3848064258445224,0.36258615727143373,
 0.36951846908604813,0.6466983205769644,0.5630842639785113,0.868346163357806,
 0.5511028144524036,0.5419496188161844,0.8395085948009857,0.9144006161084175,
 0.3672538219293465,0.3030188418688483,0.10438780475821274,0.132917507051039,
 0.27807019342726325,0.07094731652465558,0.692932072687827,0.7293459349587329,
 0.745952811921778,0.34609399959872933,0.25896160003582247,
 0.18362171566479724,0.49245034717602254,0.9138799005086456,
 0.2445890274253435,0.7125875043451436,0.8401286254032299,0.19153294902737483,
 0.9490336215358645,0.5405583087608584,0.16857433660016474,0.8362523997533562,
 0.7569478962152144,0.5589713493114901,0.9808147907183732,0.49774357264009705,
 0.4510301749401804,0.8182606580446149,0.8870673689060954,0.38598926161669167,
 0.3105006640016085,0.19083526004109863,0.12586855571051503,
 0.3655388193532968,0.09771293345133603,0.6117418592118866,0.3081814762239645,
 0.8648634294184454,0.7066952996186513,0.4908929877040923,0.6689476587769367,
 0.9415765577552937,0.8433303390423521,0.2353549982427816,0.6141788606222216,
 0.04301353675536512,0.7529827218513864,0.30821058824293746,
 0.031550519337398386,0.016907105749677687,0.4439367259933147,
 0.5192026103392069,0.5378801839388323,0.2736800157682422,0.39108284473230914,
 0.8561117788905832,0.040492452981918725,0.7012365453014043,
 0.46983272885099625,0.6918576351946535,0.7376213716965412,0.0925467789448986,
 0.7300343598540107,0.7859524638215829,0.19526461570834108,0.1533669036559051,
 0.9749255510921846,0.5456414053971087,0.3687771857454202,0.8675365930898579,
 0.8547807148937847,0.5383563015492256,0.5389111776807809,0.1655713250331733,
 0.5521413120756319,0.12782429456423117,0.17700367639439085,
 0.2012971709729925,0.9084208792318034,0.2209453798387203,0.34500288260091894,
 0.08034696410419873,0.4862191348769419,0.9462888176070892,0.8634448159097561,
 0.4587995341204384,0.7393283216835368,0.33923672072243627,
 0.001791060719367299,0.022196320946710868,0.6783455519774017,
 0.8963286442326404,0.15686620910027338,0.27001530689970754,
 0.1787426458129271,0.9772461519453968,0.3228579694194882,
 0.007171765399321295,0.5260694531654726,0.14338584771059515,
 0.7623956160784946,0.6032718570197925,0.06633071084737041,0.8474739170222632,
 0.7883362204762707,0.7428052276652699,0.32464235820023724,
 0.06550390828065811,0.6580858582774668,0.6262177891706654,0.5263163295097526,
 0.8752883062024253,0.9201345972675952,0.5442127845148966,0.12172696820780327,
 0.6499518703293208,0.8152950127216338,0.5455521688678874,0.39194560954492386,
 0.5846665377161927,0.11726135298774176,0.48047935992572466,0.860033597021568,
 0.998311580506063,0.2528101038605115,0.4519723455571736,0.9353628861079635,
 0.6237874445604844,0.5679965876566107,0.6690985334190049,0.7607731507901881,
 0.4419288916856522,0.9495095763821071,0.7605129998010486,0.9702731441951999,
 0.5597174669480204,0.3580739009988858,0.29279504267189593,0.2583324534977568,
 0.2147889338020229,0.19394559242719822,0.4543297377982205,0.4526448672430803,
 0.6854121460119942,0.4895758651896236,0.35606916965914326,
 0.06974536198901293,0.889005795339947,0.46007271408841255,
 0.19095822479237412,0.7747195506293519,0.060553808203152704,
 0.08310132155640704,0.05856624991500825,0.035928258764751586,
 0.1993915276297954,0.46145654843934136,0.04218203043695401,
 0.39500206646837843,0.3341821970070592,0.5133339285886996,0.999478446818302,
 0.44966653402452283,0.9849010592792785,0.39882881159659034,
 0.2665028700750546,0.3120829224423277,0.7153749573278468,0.10804624962629739,
 0.28993283752274557,0.9282835575883657,0.04752964616272859,
 0.22055245935264778,0.5829723517491956,0.06981850264952871,0.448201352304022,
 0.5026211932134768,0.8778652731183121,0.6090819242464989,0.37447437962003505,
 0.01779371030992849,0.814042276768654,0.031750926641753496,
 0.3055246972713106,0.9884069432457112,0.08650996922542797,
 0.36831090387877785,0.11788236815944653,0.6066028274240405,
 0.5341631118447476,0.32236706187975817,0.4444937507221437,
 0.03647577245306799,0.09591249298548588,0.6144455219123091,
 0.4604685041900214,0.3953630977375444,0.6141326150082212,0.635492806604696,
 0.10633009381425607,0.9523935534388971,0.3306576387626159,0.2058271970472918,
 0.9302040170094136,0.7602853650427375,0.6300328483568471,0.13242703040481685,
 0.2915162004562749,0.23770024917690513,0.6684393236678998,0.8205235366202583,
 0.44627554556675053,0.9786754842806098,0.2936622474066759,0.7025357282643536,
 0.3081096605000341,0.8221338891859308,0.5795906778512068,0.16642368952820064,
 0.3671602925344478,0.7744572360830695,0.3112950751940734,0.6111544213570796,
 0.4592296885537045,0.2203537792970658,0.2757282027199949,0.7436063941401845,
 0.437365770710629,0.43230719385550387,0.8682801610940774,0.16854179673376546,
 0.5929100326607101,0.9514518108097985,0.4578028387436206,0.16824062846034804,
 0.9344450358309107,0.8683127936391059,0.42829734350985793,0.8962837958125475,
 0.37878874222025216,0.7395573122638862,0.34966684761976474,
 0.6325060867025165,0.20867248218903334,0.08120864390317539,
 0.7495141431723586,0.8477093686061278,0.9333079827970225,0.12461855921304932,
 0.7973512833380703,0.9567802251600449,3.431489920568964e-5,
 0.8545106579571141,0.9432754588107286,0.7901032452262942,0.7631579338803752,
 0.13732191288484485,0.5008280391451163,0.14655634259647154,
 0.42993909547302966,0.09875381235894043,0.8731477016572254,
 0.4920521522006708,0.4467243543878636,0.4241068172895338,0.036972569238093,
 0.7860687940815,0.05204863737782017,0.42499585765206915,0.06939265197209055,
 0.9782066282212865,0.5789546475703792,0.5820394482933005,0.35608374043762403,
 0.15106642368583956,0.9556039741061569,0.9603954220758709,0.6671800335639398,
 0.44846520016949465,0.6430933974176212,0.13059797577978172,0.66857532906793,
 0.5511877878997193,0.4692437122795895,0.3124844256805088,0.769889805866514,
 0.8726261725208969,0.23235150478000288,0.021204678562266266,
 0.11302330894679402,0.7628795852195287,0.14922665265996293,
 0.14807229957118762,0.5159117808166144,0.5550579774879295,
 0.13300055462826177,0.26202002230659793,0.09175149072306477,
 0.4676646493023222,0.18551870300147844,0.6318658541547504,0.6493831844655352,
 0.5741254484449047,0.8476356120851134,0.3785844067686043,0.5223139271652973,
 0.8450625728695693,0.5539291101951059,0.3937446353415738,0.4131341837841289,
 0.25519748503314443,0.08559335788136024,0.6158869887532041,
 0.36774294199528323,0.999298186027337,0.7591950597031694,0.11252086256982896,
 0.4616559909733763,0.8480738189680765,0.2067642462929271,0.08019569085589129,
 0.7473193621907075,0.09257498183408164,0.3518633729841587,
 0.35133128418546966,0.08867005282377982,0.7164369139564756,
 0.14693557466388363,0.8425622871847901,0.4873303166088252,0.6227864857374571,
 0.8227429416170818,0.04315550303178617,0.6356594042530851,0.6106736043389029,
 0.1595223508880057,0.12155416837003452,0.12938770474696726,
 0.3715398570842997,0.2983209982559636,0.27194369437605725,0.1871835523483114,
 0.07469343244960325,0.14883459640936758,0.5958354221708589,
 0.04214760471326662,0.6334664215632543,0.11957223904826175,
 0.12108242591294127,0.7159154551848291,0.802823134171531,0.7600595513701965,
 0.9970703222588744,0.6343718974255738,0.9063511774621522,0.9533894067616079,
 0.1766872736144287,0.112210992100088,0.566738526906226,0.7067321245899626,
 0.8264881868627105,0.05377378662838461,0.2547629066418712,0.7397969017158971,
 0.3055114963581915,0.9550705161436092,9.295595810874246e-4,0.567957469593537,
 0.3665240873467315,0.45180034482057296,0.839086587265563,0.3206784502193306,
 0.2025195755273077,0.9146738311484032,0.9049492204976709,0.222395129113237,
 0.12169820535861708,0.2116388015270223,0.5750416261975922,0.2691482417915835,
 0.6569224951872471,0.5362140930401782,0.3414713666895802,0.8214087291379027,
 0.6053091372570718,0.11603421249846924,0.05375809039279389,
 0.8477744491305066,0.5342285469543702,0.5874844459579733,0.213041621302117,
 0.4608431264919155,0.7121639750140538,0.5020746978778544,0.3943530990700287,
 0.5133438098759957,0.8580847512485363,0.6117820454106694,0.21184245086628417,
 0.3970458988500831,0.8619211228871142,0.6931548418047149,0.6660693735932625,
 0.9062062751882298,0.26845155107472385,0.00936598550339407,
 0.9936116229004852,0.04123709791405428,0.4906242650109073,0.7376588680302475,
 0.6623151395087207,0.7671422158271252,0.5819050162333808,0.4068415488660283,
 0.6492812280495388,0.9429399726715644,0.7009999762957713,0.5662700764981353,
 0.7417049390943435,0.43710404446726014,0.47863010868423317,
 0.7167134879437083,0.9219290112468762,0.025610392703765084,
 0.10098345757780702,0.07876206101207806,0.7956030012682217,
 0.3933540386317922,0.9817145786384394,0.3327433317892162,0.9416397272629871,
 0.4316398788716681,0.3393539093115926,0.4223478434376805,0.4862414677835952,
 0.8448345526874993,0.031612203444319364,0.45118357231565165,
 0.1575939076169055,0.0485032921679156,0.36071782531808994,0.7548844689680687,
 0.7624731980634227,0.8138528851735087,0.06481090349531016,
 0.39047442068400207,0.28642023049668097,0.016631843702103244,
 0.5961303069802126,0.37992223841264017,0.14238586664859854,
 0.14160821529592438,0.09674059974064109,0.9459987442949997,
 0.19982070530433582,0.7381248888085145,0.13357761554177627,
 0.7697781052868089,0.7112588524208519,0.5975521787832463,0.14833935999856318,
 0.7729542767117108,0.3555608491556018,0.022496753848681772,
 0.28962051674774025,0.4577103671427424,0.30054352012908625,
 0.8701199293978876,0.9027542104292253,0.6806396291443431,0.19108642811692245,
 0.2625248539890348,0.423035875907124,0.5143113458688805,0.4612474682054445,
 0.5959113839802749,0.806652502958379,0.06227321843437761,0.8765038495357235,
 0.6315891927199545,0.6423278439455837,0.4617669665265981,0.8063207258881269,
 0.6824710506151986,0.4180049828996506,0.2699757423766068,0.27317301676472594,
 0.12485391066628149,0.9314462826987342,0.6494857575740213,0.8074484532998532,
 0.9949525887081581,0.6429636539616017,0.3036650370706052,0.6152363759981148,
 0.41394866353157944,0.4416290593644009,0.15820872303158295,
 0.9674508043292173,0.6420785874623901,0.2628972225831726,0.3006365537393104,
 0.9095426665680942,0.9437275561611269,0.1989113988472786,0.01641999367682856,
 0.9732104451884269,0.22061258810769058,0.7179937380312786,0.4128103814128352,
 0.7469278676822905,0.7078256373037268,0.43704668211082065,0.9672469145801971,
 0.12025821191193842,0.9910694761403219,0.7207844598265,0.44822788572546857,
 0.10925379341692265,0.42810339532146713,0.24758971651973494,
 0.9441794637970317,0.841438242164336,0.9736615756236942,0.5226642236770576,
 0.29872440439730463,0.09867608570847386,0.8529556491981771,
 0.28536906651679605,0.960637148662255,0.4460911226469123,0.1737031772740406,
 0.9283177060741633,0.9007906106309946,0.2644862925698729,0.5154142284681067,
 0.7246616111422541,0.03599574657712168,0.5911794776792182,
 0.12411626524249542,0.7293263890490049,0.5318831260540127,0.3325395836797582,
 0.9819943673329795,0.08513356278239192,0.8908046042689601,
 0.04386572550092771,0.5534065403583754,0.40080223440751395,
 0.015941892473323227,0.2856046469369784,0.2722916122582235,0.819024674818376,
 0.8155747478502635,0.6032533977060169,0.6653722062233959,0.8668061013352419,
 0.10637402947461438,0.9772549992024805,0.007957617512400428,
 0.7891385505780822,0.9820258261126331,0.09970047639200319,
 0.12822555835000937,0.15923884300388957,0.9504975654405159,
 0.13572443189820882,0.6899600651800636,0.6540344630421177,0.5596848570224582,
 0.7424781903859824,0.4103070721962524,0.4272086335981883,0.1516652635658322,
 0.8710889340920479,0.8283362411759765,0.49367339386799314,0.9682614638794897,
 0.47657243933033056,0.12188340293544486,0.9680913094668299,
 0.5660176469728608,0.7634290822692137,0.29313223513141307,
 0.02627904584923635,0.024857834991678596,0.48972056680837517,
 0.17910328025498812,0.5947182293276247,0.9421173331625061,0.3313149544987837,
 0.7071726815220083,0.37406168567948783,0.5601894584796784,0.0635592252059225,
 0.5285395794960657,0.04655398345638406,0.4267809154417205,0.8238912289833527,
 0.09561769155246103,0.9764382143169101,0.6778207445606489,0.7347539944755391,
 0.9351283866111851,0.15609067217802441,0.18298072592379477,0.738366293318639,
 0.7714574749803902,0.6025739216437448,0.7971934825907789,0.6262507554111727,
 0.20167128309838667,0.11664032707663363,0.42979893665168567,
 0.12068032875906942,0.2690527612538125,0.22691063531162725,
 0.3828517029785724,0.6872280239867333,0.4236571145805803,0.24986513071805805,
 0.7894295674340438,0.5317015560533307,0.9067250159251709,0.2764008054587148,
 0.5268512076724403,0.6604427728879435,0.7954917712664114,0.45732478602237714,
 0.43507055474896195,0.17014728611230234,0.9157859055492709,
 0.9967905042380762,0.9801535298681484,0.46577173134042993,
 0.41991251481933733,0.9655537701783878,0.5816000635573947,
 0.31910723581644385,0.9390420872901514,0.853084958801272,0.22806769446194586,
 0.26697031687819606,0.2490102667223506,0.35316586381199977,
 0.14827845065533007,0.6933955439669937,0.20607900015231329,
 0.18136526973716804,0.1647662826122338,0.42002912261572967,
 0.5682772432019465,0.10803601703800259,0.48816312764968794,
 0.6131635837873509,0.5017946337671588,0.7390039196607908,0.927140705086391,
 0.40970980344513597,0.7646634828026233,0.825721783587925,0.9025180479450486,
 0.6570702595380439,0.7473398421303016,0.8435497849920165,0.27747339863847065,
 0.4773465050877561,0.48147596484369815,0.806831588170912,0.7403634580185301,
 0.6433697582814715,0.9768313012652877,0.3154911357068878,0.14596586016308633,
 0.12420477130108898,0.3676342033177843,0.5148465538080089,
 0.40005496266342777,0.8777604025762967,0.6003165129733421,0.9875983543313203,
 0.7567940435997602,0.8490149125445619,0.2931702450781224,0.4989747471984869,
 0.09736745794282387,0.744557043629908,0.2774521630347625,0.48294658403507973,
 0.9058506213226164,0.4973330097502008,0.09535112045750349,0.6815077490695489,
 0.5787493275711459,0.9336858992248429,0.9033277568237358,0.09344485472482478,
 0.35117229728921395,0.4070696294489117,0.14145021099531108,
 0.0931454703131872,0.7961282291993843,0.9902307905768891,0.1822115820164908,
 0.8868808264188414,0.8137474670151436,0.38576762454541014,
 0.26076711698967914,0.8843896416419325,0.047100539165079836,
 0.4752957188528044,0.42635143852170165,0.8182268590273611,0.9716916339353587,
 0.34880202463098664,0.8334313415672578,0.3618843109538231,0.6894896095352574,
 0.7529693455336003,0.5188841939376034,0.854495996427195,0.09601832165339874,
 0.4713467795638757,0.26851827731214295,0.2677622216457012,
 0.020783870753636302,0.6804167192885376,0.8134673861924041,
 0.6024733726823688,0.015957565035193966,0.8405707285097915,
 0.2695743855825484,0.6405305444711887,0.9969803013822655,0.10356437025601128,
 0.21120769811385243,0.6833429808172671,0.05907706234615773,
 0.8368158265860046,0.6631897011967984,0.20468868721711275,
 0.45984011548188153,0.9432319501326474,0.6298148626854823,0.6428141268770619,
 0.8734384148075155,0.22092738608942997,0.17377506326700254,
 0.7574444112450235,0.7162569862822826,0.8603412090478998,0.566839971085215,
 0.8274361734847739,0.4183901925098422,0.8298879982394705,0.07810635139181987,
 0.7034052141222334,0.9658415884832718,0.8033893697589423,0.7468957863778514,
 0.7250933342686285,0.7118402733996793,0.8328585045248218,0.42562063846717424,
 0.18534501594701003,0.8493613019720163,0.32138810507524684,0.530089777591274,
 0.046767731782088595,0.15275378496286895,0.06774833076047049,
 0.6329496946160598,0.08579788127544175,0.18548220595496234,
 0.6288848813306488,0.2090998984259934,0.10142477266742689,0.6627156909375495,
 0.5534282607467979,0.4744097969242391,0.9358188324869734,0.25141430292904043,
 0.015082368153566339,0.30824719330068706,0.10704093093637446,
 0.7241230775384111,0.4308919359370731,0.3269312421549141,
 0.042102257592630976,0.21647544756495773,0.4784912014791528,
 0.6339699092315625,1.8837052611786476e-4,0.7649332250422909,
 0.9325984191208311,0.4853986939388628,0.5858988423088055,0.8044231904027718,
 0.10892507746332991,0.5063876610280081,0.455472130970921,0.7794415430872783,
 0.49382438140733576,0.9939836399667767,0.4369596530295863,0.7769354168272403,
 0.7183419545385349,0.00988865499885605,0.44448388485903945,
 0.22282012211808644,0.43783631778688303,0.3385497800090761,0.879466624112989,
 0.11293905903508528,0.9975576003549982,0.13176250643760812,
 0.11237063135326308,0.4015146980147749,0.05225829242561708,
 0.6210774305364984,0.518742970293326,0.7147150354776171,0.17948005745561868,
 0.2934259484561986,0.8749932743683921,0.4251355565175521,0.5150055491230391,
 0.09392606049731944,0.6704472494658064,0.864067122823203,0.7895308533261746,
 0.12538257701793376,0.02197281569347853,0.14153089361090443,
 0.8232320778023433,0.8207793900139932,0.34440507637478524,
 0.23621806261973244,0.1569980424151166,0.92382863968158,0.7688473633312708,
 0.3767357047859283,0.002456306537551134,0.6049838070792428,
 0.17339143846158478,0.9462075330945264,0.7899949014730464,0.667001181616943,
 0.8180432006686345,0.10828011415767813,0.09145448452493266,
 0.8142339305837017,0.8697583306396378,0.9921400628895204,0.12703452649795222,
 0.504433785759905,0.7651847079299855,0.06174449723254671,0.7766033290361637,
 0.2688622506298808,0.506250312787711,0.8502593725296741,0.7901753801029265,
 0.014027680048831037,0.7453384405167252,0.5416839935202092,0.854329276367608,
 0.3198599726056379,0.3127874675925235,0.7455001227786162,0.21809993277032236,
 0.3590846694494878,0.624622850152987,0.005794175849927474,0.8478664218626891,
 0.7153497453743702,0.5691404686932567,0.09604770977661559,0.6511047652369538,
 0.9736385342499945,0.8040416811792392,0.961962002527115,0.2061075936387985,
 0.8473815174390917,0.39304354950124487,0.9206085493638576,0.5659910056595995,
 0.7515999087150624,0.7965277750245076,0.4538737253675713,0.5099757054598225,
 0.9016385023347158,0.24454665072288184,0.016066443774770955,
 0.5381014632382528,0.0752963768435434,0.251649878059678,0.01647633715989605,
 0.7404908464765235,0.8103739728780106,0.8529314991021758,0.6961797709500088,
 0.3459098029819653,0.2792571706874336,0.3787238617650396,0.2079247007352123,
 0.6873915626244747,0.4140865132346091,0.31166063031491087,0.9416737022787348,
 0.02608665953066902,0.5031596318246531,0.5091832644536588,0.9047686646045849,
 0.4032556724455638,0.45035493924280323,0.41434202476500814,
 0.13780318852501705,0.43457978415682263,0.09691925243012944,
 0.4380170589512562,0.25760341800442244,0.6106360168097767,0.5866382032967334,
 0.32598857233023704,0.13481156982026354,0.07521092986032996,
 0.4851080195964682,0.8058306703616255,0.6182124565917124,0.7576643494958377,
 0.5329820341592842,0.526164337716236,0.6919902275281012,0.9549676103646805,
 0.5962785643457079,0.4408671455355325,0.33125709672817627,0.514014427595119,
 0.6982422374561218,0.989525069559537,0.4576642101216687,0.9026144224584073,
 0.4816701709282908,0.08753404447362745,0.9942051001235658,0.4233134496552229,
 0.08629231059511067,0.4921821492130716,0.011443839290433133,
 0.42607708681194667,0.33317793001083806,0.1493534362711506,
 0.9006921246971831,0.7322241578939825,0.5619515630771553,0.9750988210152771,
 0.9406613639400037,0.48051309493060923,0.7737050480652251,
 0.33422656325606837,0.4734918546754048,0.30834546039141586,
 0.4622928660985536,0.1969740847963537,0.20134681623811046,0.2865911154010313,
 0.05779191005047257,0.30002159138022877,0.8607901637784088,0.665306441859834,
 0.6504720625891428,0.3637718746460108,0.5115688246094694,0.8019169797536291,
 0.38290391496687437,0.7670791576257053,0.32867306317832967,
 0.25606688726447946,0.569074683490216,0.3347824320667976,0.3417053132316594,
 0.8334081449688812,0.004725697134852713,0.35333263481026533,
 0.6863242877400394,0.3951140280425418,0.7736777891316297,0.6177009181370468,
 0.5995344560666284,0.8518757524055717,0.062084012059301474,
 0.8264237149274122,0.8052296023599277,0.17656051240848092,
 0.22145793291521498,0.5815369983854006,0.896042022308063,0.9997382143723811,
 0.026483756319479856,0.18584860849135199,0.46936567985765565,
 0.3468007365166307,0.4157633632090655,0.7777885840323874,0.4613429569981977,
 0.3129680222335135,0.9913830395575203,0.5792654561042787,0.6397548177679836,
 0.274948737770884,0.8119701045305018,0.29639649124594225,0.04610895576908991,
 0.9986077923373069,0.4582394079334754,0.528416350235155,0.3301285459043872,
 0.899611317435571,0.22384756186929156,0.6103745714491497,0.5470938880481767,
 0.5836960873816014,0.8277773646543939,0.1654776972747859,0.3691858254330562,
 0.42373389154240193,0.16381505681309172,0.5023418992201262,
 0.7397355905004064,0.5537901019111569,0.2796867268871611,0.9249520501706003,
 0.976601918734167,0.5588231797071929,0.17656369232309643,0.6302814368607508,
 0.4012351960235794,0.21483721845724013,0.4991747180538161,
 0.02170430986480454,0.05731416807253831,0.40252892477896374,
 0.42392998278461014,0.18080591990480455,0.7443779932445692,
 0.0499412108887306,0.48047130230401525,0.9700942811125858,0.7583963805594651,
 0.11036571829174024,0.4536192148693238,0.40871149249667693,
 0.4815126371154069,0.26806219527996944,0.8187058719545723,0.5576790478689546,
 0.03407651475690565,0.7159458722972882,0.49555014282193655,
 0.7753422916015421,0.01469998259380556,0.809729032424164,0.8193662732373214,
 0.28206474246235513,0.34677796374029934,0.5305054879543288,
 0.17495407842066113,0.954058307383868,0.8629773327394743,0.8491043189065073,
 0.25910727938251765,0.8530645662146793,0.608822177367797,0.46966578303802575,
 0.6106078072276009,0.44695715351692444,0.30904736572949465,
 0.15111608757677153,0.7432796017724708,0.3166906885950441,0.9510053765301141,
 0.7203380678047328,0.434342376828343,0.41261050011856604,0.49125887471058116,
 0.22073448289993458,0.9638611211484419,0.25126238663656053,
 0.9821722272320692,0.5634600696311693,0.7887923220691073,0.10885730139236915,
 0.7537736590960196,0.23552617219286975,0.34341047915142564,
 0.9326736177598937,0.17551968029866616,0.3209043452563123,0.9694720909069864,
 0.26622622000133966,0.46270017589502643,0.4475807854455365,
 0.3999028781064148,0.9301769396879683,0.8124537052024415,0.5262661949303608,
 0.7451992134184249,0.34793114333585784,0.9267655842316357,
 0.15483221323552343,0.16066310050422017,0.3332857704535701,
 0.003909434594662642,0.7289620392098591,0.4285271671986045,
 0.27038401463738204,0.29226584531805433,0.3872294442214793,
 0.14001048466583343,0.8114502108740297,0.3345054904968596,0.7101958475709957,
 0.5910481288447427,0.9220478666843012,0.15331443925299038,0.7145254166175099,
 0.25114019610830274,0.6999052919711509,0.7559740437319912,0.739304708459322,
 0.051811483378840784,0.6807254031133436,0.9277988842564486,
 0.026906550422659103,0.36197733079826144,0.42890594471508525,
 0.025669509765036014,0.8589032643990536,0.17259142741296707,
 0.669223338270846,0.8948540063118894,0.08394031203511787,
 0.059849209646133494,0.8740023443231879,0.010438760213903664,
 0.6453974453867032,0.8370098369944223,0.40077247202622923,
 0.21349449770144868,0.24104437074586038,0.14387642104916964,
 0.03526267722956311,0.1732735101594528,0.7636990600586168,0.6821378332563341,
 0.9401258833279855,0.1767574282042108,0.7635938803832223,0.3867922425104303,
 0.660839663699414,0.6664804898592932,0.33136970393261367,0.2914247078034491,
 0.554338298903744,0.18704444521072727,0.5563585029283769,
 0.005930246399188377,0.9649357705190136,0.2636100831380741,
 0.7300967254356612,0.40270541165332907,0.7876278255668459,0.467204493532427,
 0.19134714709113732,0.16326790545527414,0.056265912530380935,
 0.8857041379024829,0.42038906062934045,0.06064016546443107,
 0.7007173627888522,0.33446530755295767,0.3723124083580781,
 0.19169182482103087,0.02858427080616377,0.23792194613318352,
 0.15973539436293116,0.8713305953925986,0.956358302956387,0.19739527388995315,
 0.047535808620927544,0.11307539391413446,0.18275209893034106,
 0.4027198134229344,0.11726836550617803,0.7747573292183556,
 0.005667623183178816,0.7247829180059441,0.3439155478679662,
 0.7394136368973645,0.1427042720637165,0.588210853117654,0.9012628960841044,
 0.29561236277189995,0.7486705577246211,0.7915765329755172,0.5792930569056534,
 0.22969105779968046,0.14128277250301047,0.3076290037128838,0.912535419845028,
 0.829556203815771,0.21134199519768448,0.7820524374187267,0.9023026009947042,
 0.9414737886828107,0.8272547700011552,0.9205327662215536,0.11080630447134332,
 0.24370613369277472,0.8666715158981609,0.6523513957908484,0.5301406175878228,
 0.45739861741934484,0.6774968066429995,0.0972020594408316,
 0.002931634775200642,0.5587961157235245,0.43345525011182173,
 0.03168902955762176,0.1683974675505837,0.8593359014746716,0.5221909173075181,
 0.2347652890177938,0.5352792240122621,0.9856316090160849,0.40605280487653206,
 0.5227687262575694,0.8688996261690315,0.13977149377998044,0.6797905981847913,
 0.7169144174125625,0.8087763110036612,0.3464617980932726,0.764272591370312,
 0.923892401921832,0.18250650200650154,0.7632197375349155,0.99311768645177,
 0.5168772760134217,0.5360203660965781,0.13124817469650152,0.146975692290079,
 0.6904807509744839,0.9501405192535876,0.40582719245128895,
 0.33721411441291405,0.7756349774589485,0.6537907273188338,0.7633415973081153,
 0.6022374239524517,0.33757880930441786,0.5304366156279611,0.8869300770496567,
 0.38354893907802945,0.8346940976350107,0.5423972391852989,0.3268789436269329,
 0.034657698224506706,0.7107868882077292,0.8212785453488033,0.549371188978087,
 0.5828926341895609,0.7752106387932811,0.5559723856971145,0.3803041102999938,
 0.3770708339165705,0.287572712549371,0.5643339539088836,0.10244458392263234,
 0.996565975396531,0.5771281408422942,0.37582996251599754,0.9056453313770679,
 0.651086016021719,0.5102931314282984,0.9644690867475285,0.9651291250219189,
 0.727787204441362,0.35189274810685733,0.42831231444780804,0.3180949247469558,
 0.5379340265355719,0.1736808942418523,0.9328722921231076,0.14564944627687115,
 0.32383331957064165,0.13204779838147862,0.4839247848189462,
 0.27183701971133023,0.27746689971678284,0.6960578111858747,
 0.05634319736406879,0.7174242865343767,0.5482106358955563,0.8512876075017313,
 0.9343454481029445,0.9552776041021673,0.43032784848944106,0.738441359288433,
 0.8409287481638887,0.7643850903602428,0.3248015305887827,0.8630190903828148,
 0.07724392107454714,0.6870654380841692,0.26777115172609345,
 0.21293076169848824,0.37836437651480104,0.9007767502262812,
 0.5197276093589355,0.5083650671179745,0.924543930591285,0.8336645377336007,
 0.4103411658432041,0.7287238564751323,0.3023473841729246,0.12596225116362392,
 0.5085385339628683,0.45026955464157825,0.954273774266614,0.7416483800781123,
 0.8355931364445348,0.7187851986306154,0.19014597278086764,0.3831120010026964,
 7.929152855942201e-4,0.6331796011615929,0.7471723662220844,
 0.5202983436522133,0.6407047597978376,0.3379139953158713,0.43062400149243296,
 0.14319443194527404,0.6590275214517443,0.44396584851541865,
 0.5869036496802861,0.6520479788138688,0.28239088845931504,0.5456805768113183,
 0.3788072130496716,0.28618901712708855,0.9136875721415519,0.3142523971782494,
 0.5703705428618283,0.5414974205231621,0.36227706253551917,0.4129382860727573,
 0.7018501680835637,0.452131215722562,0.8469047925282573,0.5974130469285241,
 0.6911648811638625,0.7425768179935566,0.5012005061846303,0.9226879989954155,
 0.4877962620217975,0.009068156788063608,0.2039631697574693,
 0.5978496468856812,0.5946318738175655,0.15482987840736917,
 0.20871870278301763,0.45386612687562156,0.34836713671082564,
 0.676505137391139,0.7551247571433313,0.6225233530673788,0.3083299181914204,
 0.024222396445494954,0.09886095486044966,0.17815872129840327,
 0.3076830085411535,0.10099224594543543,0.3790002586797031,0.3984148442939133,
 0.7767019980139175,0.637009875745005,0.8047250722807943,0.5871551183478237,
 0.45302913797916056,0.6170284305575819,0.1836226539349307,0.6210736105963429,
 0.17479974565182943,0.031886998337880224,0.5985882610885443,
 0.875975015221627,0.20294474834054177,0.22931983233967945,
 0.32785420987563807,0.05376707509097434,0.99625080020921,0.28138565223177614,
 0.22253918503948844,0.488374317357125,0.6784306611654874,0.7232868967408248,
 0.5601842735713215,0.6651071229889008,0.38246744287371914,0.7586283537724706,
 0.5796142190063699,0.7577404674246454,0.9522860751526895,0.0111686533526143,
 0.8884017463772595,0.03307565455650918,0.08606263220179433,
 0.4847276364774732,0.748754632961689,0.813738428094448,0.2728388314757699,
 0.3289244761773794,0.6101082916893948,0.3957917676802253,0.6356021779975751,
 0.8529117350052242,0.74342142527267,0.4360010298420979,0.15851895367473656,
 0.09014232576499681,0.40182968377226236,0.10842210081062464,
 0.21975310289886918,0.4543281301520933,0.532653825776998,0.05628039774853377,
 0.19311965478485726,0.7881476317690267,0.21874112706320403,
 0.43746026875001265,0.29008718333401773,0.8259699970722448,
 0.47793087111181787,0.5700946973000746,0.8489274656271544,
 0.04934007265197693,0.30816590822567247,0.5486095388716775,
 0.026154768399407002,0.8811498901000836,0.7823977613722466,
 0.8633658936792541,0.941208703929582,0.9466408475700651,0.5756776715715167,
 0.7479955914787574,0.46254270021484767,0.29557148279581047,
 0.3622606077137843,0.15568459813123192,0.23860058538320783,
 0.12485864889017906,0.2667419887032686,0.07808100404384888,
 0.4800147505897504,0.651758929552259,0.422901384010247,0.2217439864119668,
 0.6993962677591553,0.3613126617694473,0.4015403379315827,0.30089142322600193,
 0.32864132016134406,0.15823395952070374,0.18020442534181536,
 0.9476855678959457,0.16970763848622772,0.49113945108392887,0.344801982433908,
 0.5950223470901101,0.5067728166486989,0.9957819038197082,0.4196570950020564,
 0.3341401466446562,0.8814807577419604,0.18437624924376117,
 0.12859812222727873,0.8268804149826997,0.4984463510840962,
 0.07657556524797238,0.3718829398604825,0.6376003760627881,
 0.014666079047615188,0.5700239884831441,0.5767616095115329,
 0.8409146174067704,0.32925506347463296,0.09239265905639915,
 0.3212487736303441,0.5222810534046344,0.9981029034103509,0.38781832604419986,
 0.7337715568832608,0.5537827309937914,0.6507951479448767,0.511606901979036,
 0.7676428166288641,0.7198979189563393,0.4787796234009287,0.9416392013935762,
 0.24535936715735385,0.7388287722147602,0.09423259738455758,
 0.20589709450072602,0.1286528758040374,0.6416961131800359,0.564729272988367,
 0.6852263391928956,0.11995868111607544,0.2955868670040199,0.8536469500109573,
 0.2839420787513066,0.3190724394188311,0.461713665175982,0.20045984010781137,
 0.9820487811872394,0.2795914881256478,0.9369279024836052,0.06026087886684972,
 0.9176592139604387,0.8182800880524685,0.720748166472134,0.6153316277993186,
 0.24741547985785328,0.9958794521799604,0.10879348719743931,0.516966252092766,
 0.9683683559923458,0.2935458304813531,0.4779193988556526,0.5644862799085422,
 0.5846718919032549,0.6257368139985636,0.9734771193999036,0.9714318747439564,
 0.23975097649360166,0.09355652014513549,0.7519836265037854,
 0.35254828424488927,0.2703994109694886,0.78029168948384,0.11354213119794743,
 0.34618268284896403,0.9420649916768468,0.8546370330054596,0.9238514507586623,
 0.6047248376220801,0.4080953740062072,0.15593114973374644,0.6036530239341655,
 0.3507571456320042,0.7353676852886619,0.837554183546633,0.7412504856987333,
 0.7792564865004684,0.5652381566848173,0.6492453858256264,0.7788751985457776,
 9.05304962354192e-4,0.44751802996914647,0.22362804670658587,
 0.35137609703736106,0.13843135886437952,0.1746766247759639,
 0.9315513732313896,0.3251191398865839,0.83862540340356,0.8797073142877585,
 0.06720791414899074,0.06531383605744745,0.7388704087684356,
 0.7272261760070995,0.1139948373650943,0.2415362003155649,0.9450790023119027,
 0.7259096806245953,0.19754195354920523,0.20574324392563403,
 0.6732173997246985,0.20602759440685048,0.9860992449363267,0.2738628346280969,
 0.043697270537894184,0.4300281866183886,0.09807762691138977,
 0.30787771392506225,0.4773044800604843,0.2125143863751663,
 0.16225067783870573,0.22696576426694604,0.7800866635709149,
 0.3825624521344222,0.2932205929978908,0.2759439916456288,0.05073080224397031,
 0.1746522813295609,0.5628710851635491,0.09470836915018965,
 0.23057610643943138,0.2583734770885572,0.012677614209607002,
 0.7965973099278325,0.9210761040337727,0.8458463487784793,0.8543134584964142,
 0.15452790812873263,0.9155246993651457,0.7470251800040126,0.7714083129591897,
 0.9503476136107328,0.13740712392889398,0.8887632374569083,0.8859681246325765,
 0.39712998990267023,0.5406310998219194,0.5964775844456424,0.9433987787475241,
 0.3869237172510498,0.3371867139510546,0.16615402598341278,0.2872435327817098,
 0.819893025131629,0.18697393912892257,0.2182289622674345,
 0.050830600537983495,0.5360385971051134,0.3139723926038852,
 0.035723964657980334,0.530241566732061,0.7806095919980023,
 0.034878362421334685,0.9766246849635416,0.06304746541774531,
 0.6295109811772133,0.6692414412656096,0.269646118001587,0.38205375274796005,
 0.053677669615612045,0.09407971749077226,0.5976094385352566,
 0.9668433796873493,0.1776016547049678,0.3222883445983826,0.7169216118864841,
 0.8244404840894876,0.2248510981798617,0.6239381025817894,0.6907767809605342,
 0.28053612249778004,0.004529590391210148,0.22037014242208097,
 0.6342796951461791,0.2796838294281786,0.6733025780183033,0.5585393967465739,
 0.32885676492427685,0.8008259148399846,0.4881961254589495,
 0.06409282688341822,0.43517686097488284,0.4742421717360755,
 0.9168939540212258,0.6944297545459046,0.4428712707506537,0.39471688824271967,
 0.23391234034462616,0.1683477609092262,0.1776264026874672,
 0.27458814026259093,0.027426203546192407,0.19795351989226495,
 0.6912873010987762,0.4199587360679997,0.7003277434299597,0.281009875072754,
 0.6629462886692151,0.10257656250606251,0.7089233846651564,
 0.07352781130657227,0.2330072266747344,0.33934624581126527,
 0.2368919775205256,0.6661200510579444,0.5847057757378211,0.9535578688585218,
 0.8201945793610178,0.6579047021065537,0.342608380221908,0.4103629629662152,
 0.42015111518248505,0.9672695756147821,0.24008859698867246,
 0.18739255316764314,0.8905780397997003,0.34887088203358896,
 0.0999936540849885,0.8322363190594606,0.4258624606330733,0.2513959881888521,
 0.8385075598387307,0.8567827596426718,0.037639573994302844,
 0.47535934980043815,0.5966317790481505,0.3532180681029148,
 0.31687363888406295,0.10678760982167523,0.15813405508585632,
 0.7583878328672637,0.588310231578022,0.30154559223855637,0.6594097745418726,
 0.29424992570408415,0.9454218164380824,0.14577673050942386,
 0.5684865509835376,0.663764532461033,0.006827325178103694,
 0.21214153429912763,0.5713623791188328,0.14913645744283255,
 0.5973876188213909,0.3358387989314302,0.8575240013730862,0.9994866408322713,
 0.3587106887850504,0.19026667916754914,0.5307232990973345,0.4896626231241574,
 0.6565747319054211,0.005713684906803906,0.02652162163712324,
 0.5984977091600576,0.026399132303394857,0.09634561419102916,
 0.7460218786648299,0.3811927262056255,0.03259340945118028,0.4605709515135725,
 0.8414516425965356,0.676942990467903,0.16868438000740993,0.49701312423743393,
 0.7033751135334273,0.9270777668870975,0.5856361616868434,0.9917899078674911,
 0.8956306143048584,0.3410827950856812,0.5105008747035291,0.18090561614915168,
 0.3833620321845074,0.18540017286954213,0.44427874656075983,
 0.8214459564131165,0.09274678652491042,0.7696189064124679,0.4457905431392931,
 0.16899870457388533,0.3363303022676226,0.27837684093297566,0.994462120886847,
 0.4338848118417368,0.7467640744642147,0.2563420237751137,0.15808339380831782,
 0.3621868615809132,0.3212406265959792,0.7643302716016327,0.8789546389835654,
 0.3785612908724437,0.5037511879827551,0.2920532301800831,
 9.942099160268336e-4,0.634702765774334,0.48221184097545844,
 0.43869547240361273,0.18713698142911472,0.1803569774816023,
 0.7178398674443208,0.3994680277019038,0.9285374105800948,0.1853080622476041,
 0.7138561752492326,0.07893345536041618,0.998294604796216,0.7893162991884891,
 0.7418495806515119,0.20598341531936293,0.3045725610653549,
 0.15912924292746866,0.2551286421287686,0.21676173439579027,0.17820351693255,
 0.5038992462013665,0.3463529293160922,0.6008940989739893,0.10821239732624421,
 0.049819263406968184,0.11608217685724909,0.9831374895704913,
 0.8377710163436622,0.5121175089372285,0.561387965311111,0.4413209202971642,
 0.15786094104530435,0.9432070902048174,0.4074699688428429,0.8796198081201533,
 0.03167796964749092,0.8467982308585736,0.9329272848381579,0.2593961543700878,
 0.7545639300677834,0.4563685678004362,0.1727257790080987,0.06568532130863225,
 0.6816478236870719,0.11044583644814399,0.38138050200608686,
 0.21136193945482218,0.6988848768543107,0.22886502225400096,0.657709156086557,
 0.2440017703673638,0.55139790284288,0.2754304371967792,0.6323674164910643,
 0.7197901248022663,0.686790158275951,0.25194523518933276,0.8258786468863346,
 0.22632554143108008,0.31979719424981323,0.310110378547229,0.092497187198751,
 0.31431768125056203,0.6424563652663011,0.8760622872710501,
 0.35667430686604384,0.6528102931032066,0.9758109474721995,0.568957879023951,
 0.09027586370599605,0.5564674087213023,0.7881185985483677,
 0.30565669707329435,0.6741950028309307,0.9395395892215469,0.5226133771244271,
 0.2559705001881264,0.3018588956768983,0.15622392305971033,0.7330586014463204,
 0.17738306349334443,0.4345548743275077,0.657968091894697,0.7607642228091932,
 0.08103704781837173,0.4059560547811717,0.7637626818811818,0.1731886472223334,
 0.09875612402723588,0.532440384292154,0.8147815063657898,0.9364965642148015,
 0.6854549396610967,0.3067512637974499,0.7045970184153085,0.8791149875728884,
 0.31014578834699014,0.5379112314634099,0.13107421585089896,
 0.5203969666173759,0.3428544225039155,0.44575489626275056,0.9248100514155251,
 0.6084164765675185,0.3391769392916649,0.7732667258811304,0.21621664652659367,
 0.23860321712986898,0.7333105634845904,0.6038697474195948,0.7973714152881131,
 0.2529344457253676,0.4641037088229528,0.019420454996658565,
 0.9263371395442315,0.186019541977556,0.5706928695957907,0.7820225551031132,
 0.4450219754497061,0.9686272767086109,0.14565620092933595,0.2710545883220221,
 0.11164285572086696,0.5480725299745732,0.3170854885736185,0.3302898696172937,
 0.1512364016413662,0.14763394438916755,0.4735613508855073,0.3525199933713835,
 0.5795447085144536,0.19508525645305264,0.9941303437478298,0.9092545747091172,
 0.6972749323443109,0.16381963714369818,0.727542100130123,0.859888223956532,
 0.7336256483740847,0.9477943226047838,0.6598890242933746,0.42473573974614753,
 0.2324058969673829,0.992694837036006,0.6488737456140815,0.44603288574055866,
 0.5194172051413452,0.5268219056825496,0.7479792200380659,0.10837555735821336,
 0.12704477859192886,0.704767622436048,0.9673087008918677,0.11179318855233844,
 0.1905401874265794,0.36397121790584885,0.56568330363426,0.031126300644353955,
 0.5000237599544738,0.7246516691789948,0.8861247536197951,
 0.028602648358531657,0.24071796104566623,0.5941899509753904,
 0.7231202171605283,0.736045778762169,0.987931456300427,0.7514828666233188,
 0.402944715812679,0.6137638828204802,0.4379029836470454,0.3669671072830992,
 0.8661357311924444,0.3144713398185569,0.5452922285193926,0.14009715226116315,
 0.6731537406924948,0.8543409818989399,0.09729833369935814,0.2298469418979725,
 0.19801857080072027,0.8176395616494305,0.040092945047225825,
 0.4418577603467646,0.2927648424907332,0.5593932046701573,0.09235526509618308,
 0.5349257706867794,0.947592130642384,0.6219095517909895,0.368317434025658,
 0.5738418304593481,0.47326086376744014,0.5482572851806115,0.0911677757279955,
 0.8598709812013922,0.5825216226061811,0.46850073947102777,0.9809091487919436,
 0.16397238835831252,0.7765415721825708,0.7625263348117346,0.6721308784664839,
 0.19760094952772433,0.6148864174836958,0.42271273993041814,
 0.6350416101208628,0.19694666351276724,0.9935263856737606,
 0.09314031218701757,0.5550704526685529,0.6551773185968179,0.7852545563132802,
 0.695809172845874,0.1892104236062373,0.8461184867097872,0.464113942296128,
 0.5441032721367234,0.8553745091974053,0.717861344079722,0.5634742667702024,
 0.3849211337254075,0.9763098572472458,0.18085698170597853,
 0.07525237962941445,0.8816046094603209,0.4044896980933783,0.92925312120775,
 0.18024317726928885,0.939949237820782,0.9050039846010248,
 0.018814969252081815,0.9602458510637696,0.19186306483587945,
 0.7051558383794914,0.27991002321062763,0.5955775814849291,0.8178100895401312,
 0.3073672986319014,0.4072420798453731,0.04411061775501557,0.345274852210264,
 0.24968353697140577,0.6716117763365737,0.23723214127157843,
 0.1059402861086236,0.4920213658726934,0.26564118586778296,0.2883066184630505,
 0.08777853951331982,0.2002799719909396,0.2595383546742067,0.9935447833545421,
 0.8660427999875281,0.12097457308260062,0.08091445920616891,
 0.7790190871249862,0.7519615216667602,0.6440648218700047,0.5064418661849597,
 0.2954278866452982,0.5366803073227604,0.19027559163768137,0.7082306212004578,
 0.6251890228720955,0.8796605654073991,0.9566485937226018,0.5720731414803839,
 0.8568481973255915,0.5266796387137372,0.4199461469145248,0.9235220261203541,
 0.05068873751529002,0.17131781669610602,0.16806721566915883,
 0.9978651380492769,0.010648567515334717,0.33492736270914514,
 0.35786658824630035,0.06960627132561115,0.6225874512420374,
 0.16392825449751192,0.6391476123926129,0.9652001615869015,0.5845997445603005,
 0.8235735955743736,0.7235903798094723,0.869900548368486,0.23100673523743054,
 0.7094806568956937,0.5062111157259979,0.8634358534369295,0.10116182112854277,
 0.06576655234968365,0.11142361682474378,0.5100224390196991,
 0.5466587448852809,0.5027254459799545,0.17889371620674,0.9013832034644446,
 0.08930517750448408,0.3044324473897938,0.12756007669398362,
 0.8115347232365872,0.24388359510372448,0.12681035520126172,
 0.0738301829960556,0.4145924440000903,0.7257028505900307,0.04382690042145443,
 0.15133075004645669,0.30074718568056935,0.10650453407675164,
 0.15121785353491823,0.7740082588995556,0.5616299120078694,0.2570850345732103,
 0.30018248791159374,0.31699720837164647,0.8744628972157912,
 0.11437416556453273,0.16676200956673704,0.1916198427152569,
 0.33940419204306327,0.6509648472526876,0.4728935959948377,0.7703770637608675,
 0.225571659306533,0.45426179961492474,0.5106059080777235,0.7495377454461982,
 0.7993478680122478,0.1843414864986408,0.6230547845801799,0.40661793049416906,
 0.6041886769111557,0.7675809255989292,0.9793547887242424,0.1485307936369532,
 0.41683983018655335,0.33606838551112517,0.32984625733083306,
 0.9500917120096383,0.6657986951679582,0.5845632198831241,0.8208826221279952,
 0.8613574625338485,0.8310915847700225,0.42335074999211886,0.2760546444156341,
 0.08874349612639387,0.7811681554020641,0.7783522187335595,0.7642882998738949,
 0.9768298214270712,0.41024870198582564,0.26028790239423316,
 0.024138028768189046,0.03636491286879906,0.5104209329204135,
 0.06041043507667654,0.458121921067555,0.9048619638366018,0.520157503719383,
 0.8488085681938173,0.9780936464870743,0.9579582479818353,
 0.008769718450927666,0.5869813591640399,0.5190025742532721,
 0.9364384166565367,0.9542164379884879,0.9003114383910766,0.237385131602181,
 0.4824370574163892,0.3672835963746468,0.11365173882366697,0.9966931058280378,
 0.9144044448714421,0.013984788083768795,0.905073952785006,0.9565704272740753,
 0.5695474032266921,0.5209145055695975,0.1501606284838949,0.9109465757464842,
 0.5119938651950191,0.8930838312402402,0.46075244605306065,
 0.26332189818122687,0.6790977220581678,0.8287109969498982,0.5127219704909833,
 0.9950534631432344,0.9334372014321785,0.9982848378026796,0.5725039476249363,
 0.7522801409244874,0.11349829599709527,0.9215878611435546,
 0.28248592711079523,0.5360654265246843,0.6746384405067434,0.8110181159536048,
 0.8361180744909383,0.6025313575580177,0.30590412745471474,0.374886040941682,
 0.2721982735852,0.4599893406597112,0.8686730141697205,0.6093167029006231,
 0.5149476606305207,0.7533344329848025,3.063126494008417e-4,
 0.3287154368856222,0.8265684713018178,0.7049887757594653,0.8740799293046049,
 0.7264334413957574,0.8507697160277672,0.654248662455253,0.6065742750555985,
 0.34540012796144504,0.8949377582188783,0.6375018418693831,0.57545709008653,
 0.23775507908056914,0.3121520808915832,0.38783859485419514,
 0.9750333478853663,0.09279533410556606,0.3947774659354836,
 0.34548481673142317,0.09431260990264956,0.04777940204257147,
 0.33389178541658393,0.41087597944646026,0.3511685480373625,
 0.6933471590969759,0.5377070834893225,0.790829482593715,0.25824078281078944,
 0.20856248416038514,0.35875058516878644,0.7627673332637779,
 0.7770321686783408,0.15106326928641378,0.4789484798619399,0.6687211441928637,
 0.6771765606196091,0.9374216702766833,0.6995900113877231,0.5630269327892508,
 0.9429313526703039,0.5833198150523289,0.3353136688575319,
 0.057194383243861946,0.8816345915902313,0.8189499789602598,
 0.3041644618413557,0.6966701204288561,0.5043831633494078,0.4755962170738074,
 0.3687710116905397,0.39325493164780223,0.12055358766958424,
 0.6851176685423088,0.8572431218605356,0.5646390054286822,0.26152610508007545,
 0.5791973909991077,0.683651681972781,0.6591158039189949,0.5448737684059404,
 0.8984552098539185,0.42373924070893576,0.8908399171221385,
 0.21341369839713809,0.99376989699102,0.12774015169208341,0.4924251342841869,
 0.5086423968835686,0.3062020226187794,0.21632494926510826,
 0.025828623224083413,0.01947449419131153,0.9446225743558982,
 0.622623998448602,0.9859120108074102,0.30964533959848806,0.06994936652608552,
 0.845636111336329,0.7107411714591966,0.46063669902463356,0.7513061248468234,
 0.41491102822448456,0.22772308755851967,0.9880633910172123,
 0.7726992872118242,0.9968676032610533,0.3519457329253308,0.10466061575954033,
 0.4613449183043752,0.0951051662414486,0.9982709673524615,0.17829135617607517,
 0.5919960380827017,0.6259397072486299,0.3021255329312189,0.09425756247151618,
 0.2085013668384449,0.7551840645368053,0.9679874339153346,0.4549790618010583,
 0.9812807326753661,0.37886778286040235,0.21566764399290272,
 0.7523676911858129,0.13713890426461295,0.4973749978909403,0.6517040863388431,
 0.29459238297495827,0.3444122560735381,0.2959676516941623,
 0.15326615886419004,0.6280460518144823,0.6686058599743672,0.3245465803285389,
 0.2772448639595687,0.007070676687462751,0.64423078772959,0.6401241383693249,
 0.985392067095433,0.5598131329682019,0.9607074779236274,0.9993981807742565,
 0.23002413747588613,0.5197298999411496,0.4048137456198764,0.445683932338818,
 0.9512020772981529,0.23334754152162895,0.22488560172653393,
 0.45784553786674353,0.8022900688430414,0.7617446566336168,0.2943647982694636,
 0.8170835907444346,0.45543876650174553,0.5692489969395615,0.857506830399504,
 0.6913734217264788,0.9285802766867218,0.530983627146442,0.9030963998944216,
 0.8818507273154799,0.7721096350975568,0.4194787663581866,0.5909553789991022,
 0.6998200903180903,0.28626484150753395,0.56098555749791,0.2287870691824181,
 0.9218487768843939,0.715900222420542,0.46334503085040457,0.08455839018818634,
 0.8056926304343152,0.8279618821018986,0.36629713947642195,0.7012414835605907,
 0.6489816232803585,0.9475043090060545,0.2188156046101163,0.48679695807873813,
 0.688284389028956,0.40030171007711435,0.7678378594872054,0.1386094832241126,
 0.20920461663149992,0.23914734315592456,0.12437955504954024,
 0.9676754890373994,0.8111427889902092,0.05852671801770537,0.4432171197166035,
 0.5634393684092766,0.19086619477396982,0.1357784246534096,0.8971205079623521,
 0.880787122659828,0.6610047878056057,0.7984821847591771,0.11234247849925028,
 0.4932046455795095,0.0414764974596431,0.11666760730164782,0.2080818139072279,
 0.0769475820701141,0.9538051627596518,0.5241671475843174,0.9562520337792811,
 0.11939274638261854,0.44207816037455205,0.13304658372008837,
 0.9639336249332775,0.7549793583210463,0.6341080216588627,0.11502031806043089,
 0.8338465107861409,0.6890592434924641,0.914845393066681,0.9303022699738586,
 0.9693453760245221,0.47327362566057785,0.22786219481010328,
 0.9239809610354291,0.9017999311275009,0.45686434630305417,
 0.12744520214515886,0.39708619633964626,0.05581822499593514,
 0.6618551763483305,0.4763425611500487,0.484471728225111,0.5144111362285688,
 0.7932732174805575,0.06934942232045405,0.7714616469204999,0.8676550099345866,
 0.7657298914128622,0.06968277037779302,0.12835862544638632,
 0.16815618166390311,0.28440649781318417,0.6584907752066628,
 0.7736328702123184,0.6993818654643105,0.04155472029983365,0.918668572580613,
 0.8228780803921648,0.9119010610348377,0.7746748706920044,0.8586910829655073,
 0.15878251862457538,0.802645578138661,0.2740509415824236,0.9087427902519798,
 0.6136865525722571,0.9505158295388038,0.2010282988126031,0.7360117102289184,
 0.27785138430481326,0.02670222721919946,0.14473169654438223,
 0.7129988652107517,0.8470317042500111,0.7769163460869165,0.02944297784288963,
 0.6599364372907868,0.8801957061176797,8.458124136245804e-5,
 0.5269533188842608,0.04333950000394293,0.10128186170003328,
 0.6371731684989166,0.09902895545050394,0.2734199916059955,0.715803393695038,
 0.05739539941675553,0.5893222489896943,0.04804972847664646,
 0.7998879974282745,0.5107839021227025,0.7724411610545538,0.7940014606176627,
 0.16734118101505613,0.4276501021988499,0.03622914115833464,
 0.7325829231707448,0.9812532001508208,0.077507115595149,0.059301296298637785,
 0.18939970775361303,0.0772887462216385,0.37885747218732246,
 0.7193512235761899,0.78847850111977,0.5070254960971206,0.6887779054879042,
 0.359885376851131,0.42755725739283934,0.7651036161727808,0.4472060696198399,
 0.9137805648022083,0.09854152677513672,0.24838669927115065,
 0.32361079310380825,0.8948130894328585,0.08821365955809646,0.639929161993209,
 0.9399957450778061,0.9591807926358781,0.01363151283276709,0.9736585994948884,
 0.7184406043845231,0.9205527762077804,0.09674564542527764,
 0.30494397801847783,0.10727753219573533,0.24967876649667275,
 0.6397709079888223,0.5754295121159153,0.8806765005927271,0.5062168939716593,
 0.5152386974026195,0.7107991806231974,0.6182946626829372,0.7997526246749184,
 0.9299776871533771,0.5976901407378831,0.7414515872291282,0.8226771789405519,
 0.14496451240069275,0.003553711890310729,0.8592620398296387,
 0.5109773651374778,0.22641508308307035,0.6973752176688717,
 0.28604973562122793,0.333951440619935,0.5770369402909282,0.918357113053105,
 0.521565427782624,0.6599660752717145,0.46171091843394885,0.7146671716006954,
 0.2988000686126986,0.6928940120916738,0.8331370747953659,0.8866888796057126,
 0.32328218397270025,0.1917789427808374,0.503037709215672,0.13596246073457396,
 0.42471273392788533,0.37800661410100767,0.7187157974241918,
 0.31675168321267844,0.7931018967102078,0.5155114113434877,
 0.07349997717718948,0.6659635438332597,0.23513691840375528,
 0.1563108583981624,0.7023994864430451,0.22088232843553202,0.387513709313685,
 0.7292275782902217,0.6950511624484543,0.5742901307399076,0.6452589950224638,
 0.6009114903278323,0.9253816483785087,0.6271144123794056,0.4363593054148893,
 0.47800067917686173,0.6628524339715378,0.8792696601374238,
 0.03351472837934677,0.809779799535471,0.08245869384837923,
 0.39315515071657126,0.5782699101858789,0.06683564677601916,
 0.2238070490807219,0.21151663530002807,0.006689900719692221,
 0.0028157802807815546,0.7288562082201357,0.44814033616641535,
 0.27605901569075775,0.1288336545171621,0.5045388354319674,0.506516560311319,
 0.2772636111445047,0.8340726941177495,0.8541341449021301,0.8382291982766784,
 0.5249376474493123,0.4876536261555704,0.7243042234067253,0.32956848096084446,
 0.9388369345852482,0.7173083736524042,0.583706209896083,0.044013795580930326,
 0.23576900545833723,0.19741714856557202,0.7344784540236868,
 0.36198659421973844,0.22072108931808887,0.7506234060698671,
 0.7714454914928452,0.8502395413392256,0.90960816955359,0.26132501468830094,
 0.4691337738358714,0.10602555238524525,0.6802400073916481,
 0.26994404119734206,0.22600499711411248,0.8398064016212682,
 0.9980926196774411,0.3644992972926242,0.2203723517262517,0.33931639308548356,
 0.46914774694137196,0.1629790534579788,0.6825541897025623,0.8206169110759292,
 0.39558880920460426,0.5799661938829164,0.22049951933431777,
 0.5249468909017256,0.4164209793965997,0.6181693989469699,0.08118099442837445,
 0.41026201551799624,0.8935585536656001,0.6982466625632702,
 0.34032960434539905,0.028156675653733587,0.8549992438275864,
 0.5255284197147656,0.8492925082218378,0.16204288760095198,0.1580993272578879,
 0.7690140975754624,0.0689932228425687,0.33031041890616974,
 0.053833818403297595,0.8079682493964315,0.7235285071570912,
 0.35961052698389273,0.8443496002212002,0.18572649680810993,
 0.9356075636725691,0.8505900078874702,0.3289980155105404,0.8116412257538289,
 0.8781591171669155,0.4715950589165161,0.3055709983343764,0.6799894397170464,
 0.960069315416202,0.6533533824877454,0.5016112040531145,0.9804140176220301,
 0.4241986094466522,0.9382942794161327,0.6573120054364969,0.8385643007425363,
 0.9407933303022121,0.4173181749123597,0.19915183231960298,0.8847090964226634,
 0.625214287427375,0.9558410103680471,0.655367106038551,0.22935006018481863,
 0.9904175769522895,0.32718853251726676,0.16367583329675217,
 0.21774145170143755,0.5102124973883547,0.06971238188293516,
 0.5351118884930797,0.8750681944425084,0.21134520246315724,0.1899252041743683,
 0.255285231544352,0.5228340897888527,0.4262501857068318,0.9395932657436791,
 0.3356483598799658,0.02158326559164636,0.29610018925229853,
 0.1275058822559645,0.901488630320042,0.49529206197420084,0.38920163421198184,
 0.49656441479252966,0.9086493358684542,0.05689276675399302,
 0.33284101706112856,0.6528156097692147,0.4131477578777253,
 0.015749292749118182,0.450691251113414,0.006919015332138012,
 0.6112545153878762,0.6676487351959097,0.7135436624695439,0.08145783069349566,
 0.5168699199548015,0.20207771096825744,0.33624988501176656,
 0.5723258839416907,0.09660598634591244,0.5571529020662151,0.752985902926582,
 0.4779375079146645,0.33144409857349555,0.2326960521397894,
 0.10291862786292283,0.5937872391464417,0.9928184902634812,
 0.34084140802949203,0.38420550658307995,0.8797077373519097,
 0.6902893172563622,0.9430889807370624,0.5093037605371188,0.08578862046057034,
 0.6950973726525017,0.8391760697307737,0.7720067517091086,0.8286337237064405,
 0.12067802915020187,0.6994156856950379,0.7397991558427529,0.5996968054597653,
 0.4261557802305955,0.6295669597904779,0.11120454907996946,0.9979555995194269,
 0.176344744019868,0.7350841371286028,0.4603008432695943,0.32029746327871644,
 0.7539010789832357,0.11907124698284022,0.399935677375598,0.18002005115919983,
 0.3172832811453019,0.056718122158169515,0.5571690857503414,
 0.8066274907897847,0.7317073352426087,0.43947379411608944,0.2893743705512115,
 0.6586248512618693,0.8736653641675928,0.9378514436125277,0.3275880571242936,
 0.06249519919531754,0.5313293340682785,0.5971169376112534,0.4164589616723927,
 0.9207508710955743,0.641103952931924,0.079781757081684,0.6515785804850907,
 0.28671125211698634,0.1270077826392063,0.1841756415505813,
 0.23960151122010953,0.660833483616705,0.3783420757847501,0.7380607992032616,
 0.6813138673656086,0.4178187722325475,0.26641569857284675,
 0.23244994550797504,0.9169267075602671,0.2703865796950271,0.7133113431379683,
 0.11667174775416067,0.5652383285965016,0.3160078625402659,
 0.26538099155204653,0.8529615565405527,0.14739912157412116,
 0.013355925838571814,0.055801721371323665,0.7936471782977227,
 0.010658573061566479,0.4475391575711424,0.1646273991592384,
 0.10594412377612938,0.6486772148753206,0.44858377832011775,
 0.05703850620094619,0.5759078420680931,0.7553337817809967,0.5068763221168398,
 0.9026822143251655,0.14732960839099296,0.38221078578277323,
 0.7034933286347199,0.6196355250127483,0.34527658651107795,0.3260064124438996,
 0.8410779151928839,0.7504916936802618,0.1067588905200817,0.7706291001893109,
 0.712171846498571,0.08776219828276532,0.8662258120096091,0.8685962311530914,
 0.4652142673544909,0.46854465472854,0.3251709740617359,0.5254660138110158,
 0.7061383676537611,0.10291932787257274,0.4886690560153607,0.890062050790759,
 0.13227692126144153,0.6908173664303172,0.19134930797165373,
 0.7215347526175827,0.6173402493900998,0.5798640328216204,0.35461150083477977,
 0.5186585036935693,0.611409200095515,0.21143218008642561,0.45966167859730445,
 0.4833804284835336,0.8902906482683699,0.46908786557825677,0.4305317756839758,
 0.6533373442944809,0.10714376251790592,0.8284321100022828,
 0.39810329271005873,0.9831071840218573,0.9480919871874501,0.4731469373307472,
 0.3791513611654681,0.8054902114775006,0.4254305908160756,0.29271754359986424,
 0.52713717985076,0.8624238678544265,0.3790186397284705,0.15964827258386938,
 0.26493169575501374,0.9063220369833167,0.9923304087671873,
 0.49507397181102397,0.14112730148268682,0.35045181632687294,
 0.340090279286257,0.6629374117884902,0.5082120025909076,0.8482248859947511,
 0.27370298910865487,0.20004088066233572,0.7485271102046372,
 0.9012051445889008,0.28495471164282826,0.1492545662322009,0.7688338275658937,
 0.7434537734280641,0.14922241780610246,0.2999464113513196,
 0.13714559350867017,0.22356151273954017,0.03731059717477825,
 0.5312615770405545,0.1421841936901309,0.3929473923686305,0.9239621867232378,
 0.7096669655480738,0.8799015538344177,0.8791049803748543,0.4969258235033117,
 0.8657928584442469,0.43227347852310327,0.5564456724638572,
 0.30660577868121663,0.14669928456583092,0.12265383951582776,
 0.6560738350622768,0.6638816708221907,0.9710069281142354,0.8091393794025072,
 0.7325747590057107,0.780689770421989,0.6370683437566219,0.9988238996058472,
 0.40274923353910663,0.16255100027063374,0.6114255181209779,0.707280714883503,
 0.6862108905418078,0.14968481177820536,0.7735908256140896,0.8917939246283864,
 0.6994828982193269,0.6999119897685661,0.11853912828153401,0.3705323624301875,
 0.5415481608088645,0.44514218238849745,0.4815298628947827,0.9334743193338,
 0.5962951061726448,0.14900382211636654,0.8347869050547405,0.8045049691876431,
 0.39501678457785017,0.610209545862367,0.3481058192377685,0.8882019986639949,
 0.7761463961113184,0.5534023668850845,0.6824950501045153,0.5979981912023264,
 0.01086774614111341,0.04154626141877826,0.1783243770337547,
 0.2882428288052674,0.01038448206141196,0.05639270564232235,
 0.017194632673512955,0.25621805193650216,0.046337184024717204,
 0.8273237719287375,0.08732725958908372,0.8846336696675879,0.5218228975961294,
 0.3683244850110171,0.8459936218150624,0.573834466569413,0.052169662091467894,
 0.5611909209920702,0.26782709631067736,0.7204246014462595,
 0.12470947794261955,0.3353500644644217,0.8171841297617215,0.7953893713078679,
 0.6883056223692916,0.8845570421196732,0.8640770441235962,0.7527889926065522,
 0.6582185442689026,0.43102701003097044,0.05238221948320443,
 0.9207900532109704,0.5689056352801625,0.9788127729874977,0.48792059374393093,
 0.6615069465637405,0.36316606903404647,0.19402935431729207,
 0.9384585467334225,0.0794057484028281,0.8783315651901619,0.16283683210412925,
 0.8504385063039566,0.6122740301479114,0.07391144809997585,0.7980248659996475,
 0.9879185179092174,0.9854745754025331,0.6505817038469608,0.955556277357647,
 0.8341440382591911,0.7608586562931265,0.8916227724827637,0.6281415583230876,
 0.3466259260396659,0.033919975319477325,0.14751792097857108,
 0.1600824100553312,0.6929278892078289,0.6165860452209725,0.23534300570292033,
 0.7829575599601806,0.6142189349005502,0.9894454063007593,0.6079823687519506,
 0.06562767053234353,0.7173892212773284,0.40789764104016346,
 0.13923221090147153,0.47757376921426997,0.9180088270259064,
 0.2153256654057818,0.7798719241779052,0.18106055594293924,
 0.19723579432578742,0.8922825507655899,0.3160136975059272,0.936902409054489,
 0.7755566936431484,0.7660281690936178,0.4469524008088688,0.03508231255310135,
 0.7698811370196282,0.037368899182969595,0.780363598463653,
 0.47474920730199466,0.5321115402740689,0.4696724788704476,
 0.29501753491534854,0.9387080050230274,0.3061555947316881,
 0.43372013252814745,0.6462040232062939,0.11778860749195319,
 0.7657281618572358,0.8733114188960153,0.2840649410190682,0.4703280502182028,
 0.6336223289742868,0.678126378240828,0.8577994644908405,0.6994854168805084,
 0.4847882437611952,0.4804716668358764,0.843007905337404,0.9369061856845893,
 0.07599808183880818,0.3488213249515565,0.11825301585841619,
 0.11960162002860741,0.2134475242458196,0.008598851584585843,
 0.8652125677240973,0.6967462830724049,0.5447386846041794,0.7074546485867187,
 0.8343109866432811,0.8208086922075116,0.3073609205679776,0.38504540859447367,
 0.8092799347180516,0.3731462560619183,0.9020656832898281,0.9921057101533329,
 0.07419591700695394,0.6069615515434631,0.16649305962334027,
 0.1443537621533868,0.595131493165895,0.033126503914337024,
 0.12235794144535639,0.818777835526836,0.7792167741464199,0.9580878993936467,
 0.7700591419464033,0.21442531749841454,0.4812397451572179,
 0.08612977138869315,0.7379749334279079,0.6787822026324135,0.6095644143444092,
 0.030668437145484173,0.6468538185627518,0.7874480345225047,
 0.9345551135388452,0.8197967255470784,0.9740292802089626,0.6292618226992833,
 0.39274318187343127,0.28175966218196113,0.41375322833523587,
 0.2780066008001476,0.5771048147236849,0.9640997812946379,0.35140925628734077,
 0.7201143013103414,0.6341514235833421,0.8013941552581124,0.1818656581196616,
 0.28999936820298267,0.7616864511781851,0.1750402843008747,0.1531876605417566,
 0.785456845060627,0.9101968549942431,0.9360459853639355,0.4292114421345403,
 0.33097304660491644,0.9740401179020001,0.6333203489633159,0.9918692403824573,
 0.8629858064656022,0.11005866086203064,0.7959158961469861,0.215437637315662,
 3.829251982881976e-4,0.0881115891769787,0.9567723083812444,
 0.7878248481880444,0.04538164950886292,0.8689313155839575,0.3179358134274085,
 0.9904046322680066,0.18126385071378115,0.6977740105790902,
 0.005395669895947974,0.15840008229026048,0.6938690732083339,
 0.8797274353102585,0.8762546448083854,0.9586042070717357,0.40854027919152336,
 0.23022989064177402,9.892213364994795e-4,0.601832687401145,
 0.9516996267437681,0.04023358974085711,0.09575079808069131,
 0.4768261723249624,0.9003245288358787,0.8280083642833626,0.5900925306604701,
 0.5534779744846142,0.301340012562481,0.0625359753927095,0.6303643583988786,
 0.016759972213493013,0.36432575101812814,0.3873846816091592,
 0.3931620805014501,0.40865738774717464,0.3679711427255936,
 0.45184021416406317,0.0392495505306385,0.5949726022074399,0.5526051019510414,
 0.36043908339902375,0.43319991262541957,0.42492407349353156,
 0.5754342840743342,0.8276134744280365,0.1967591511724216,0.7034791358354902,
 0.8530838135597526,0.7675926006868992,0.12004772633431293,0.5257069463015409,
 0.5974085675138034,0.8726978090447413,0.3436863993839274,0.35824403590670173,
 0.7128740044508637,0.9102656520727885,0.5571288508260603,0.9784404087859376,
 0.1009613032133514,0.5774708677128313,0.8057046560507177,0.9854432079116567,
 0.13354075286693856,0.56301851034899,0.014001639331870042,0.8593799705350299,
 0.46127357555196313,0.03637562376050962,0.0014746939576421036,
 0.5710963678461298,0.4261850368377138,0.09015591189791994,
 0.28191946256818756,0.41958834838414427,0.663402290476629,0.107693869070141,
 0.7200658903622619,0.9081451118397947,0.36773190603561023,0.3937831415394104,
 0.5495478488586596,0.833927545508865,0.8741360581334675,0.22442178427903758,
 0.16208494966246167,0.7500228025956575,0.7095859096116786,0.593933471450948,
 0.4163618158801897,0.9675575275154559,0.22872334141728756,0.7449165967631819,
 0.022303556618631593,0.1716307848905395,0.7553139921268025,
 0.22446360434983148,0.163898630876265,0.5195719980835299,0.23055815720415018,
 0.08810259544432719,0.6431045010649157,0.4516983212083371,
 0.36881604698895365,0.4296971467209185,0.39482322818652515,
 0.4402252522546568,0.03570797988980423,0.306861899397435,0.7950576477397678,
 0.5345351257597681,0.23710392811886116,0.8959427833228094,0.9685241182510131,
 0.06805732339384585,0.23702912423995037,0.7728405129274338,
 0.19251366294632555,0.7641596138364365,0.13031460938458916,
 0.17331939337048652,0.12958314916590075,0.5356931039816138,
 0.8435884261987229,0.8312817254459445,0.9177152952396757,0.29178968870927136,
 0.07458604775162803,0.591376104105509,0.40687573931812515,
 0.14678528155913784,0.31888808963388593,0.5466537914444662,
 0.033819746780521354,0.3563752376716418,0.9986018107111688,
 0.7580948117919585,0.04463653883888585,0.3418319547296703,0.6286301234609152,
 0.7857190409327298,0.10278981028783873,0.47299043188583867,
 0.25328943247214863,0.005300561670891102,0.6499114706090553,
 0.12097776100456648,0.9691410867160754,0.29105649441089354,
 0.059089938305292744,0.6617672168303613,0.2977569324246856,
 0.2209615443733869,0.27287314059276646,0.39075301431756104,
 0.36785726671436547,0.9739475759146127,0.8183421569066568,0.7474064881081436,
 0.13523785712293657,0.15861843707216983,0.2366519778707652,
 0.8377655623086963,0.3965262781740988,0.1978104107322648,0.8055133003751146,
 0.38213508016048303,0.14022285057389983,0.5037372222737976,
 0.08685284395480397,0.18180114897110955,0.1314670673315964,
 0.1277488276599006,0.15385432355527462,0.032436517141263055,
 0.6857000980001647,0.30347243560031023,0.4201046761431326,0.9565325625017944,
 0.8144510129605519,0.7835712068704221,0.9339359258180352,0.2807825070833876,
 0.8715117905504238,0.8278885920207728,0.19108235709440324,0.5727832018088539,
 0.13125002135928432,0.7061914888140881,0.7633587245663039,0.2483326327173071,
 0.2778095662850585,0.5553503956760134,0.15892231181074512,0.6115807898289454,
 0.962473044078219,0.6284651123725434,0.5970499609545934,0.8384970913747923,
 0.999252286547473,0.2175579505317644,0.7896712184814961,0.19070103651255355,
 0.14369221512893737,0.2569058133233385,0.6553865118743704,
 0.011676406223531721,0.31834729806945317,0.3144974899719567,
 0.7190951949572697,0.8293076249288873,0.14609058900466299,0.6939009718356903,
 0.7709323404636368,0.5759475019816275,0.5989070601982076,0.8840780722483172,
 0.8718403228948052,0.1323914917546153,0.20937011832632602,0.7060329372147552,
 0.9531866137036735,0.2701574708443628,0.815029457851511,0.8564013232268312,
 0.39323041488739796,0.6571137186327012,0.2917797217314394,
 0.39813862495434293,0.13354719014812932,0.9546290646747523,
 0.8131159711710721,0.7355860435022,0.21181786348249687,0.01572733447843455,
 0.21459742330127318,0.11427141422181986,0.5051607579942431,
 0.8162044427028116,0.675535167173321,0.6728459435542912,0.17205245693421423,
 0.15014466583241692,0.3412647297658068,0.7219015081566177,0.475915572842063,
 0.9635804262669557,0.04387967347229238,0.8445558827179706,0.2979782958452396,
 0.9866632093325979,0.011973789349341346,0.04222609265477162,
 0.35205282017580897,0.3317034486846412,0.02522481614944405,
 0.7686990803977392,0.5929384350286225,0.5091458918411117,0.7933073986490757,
 0.789534705145279,0.03776292256563274,0.2706438412783537,0.17276871535371963,
 0.3018805713333502,0.9701714113198903,0.6151757260942897,0.9781322780953173,
 0.2566606272919991,0.5380459180141801,0.23157981089567548,
 0.03216984475274076,0.6479190132849326,0.7179552361579592,
 0.15464780581893522,0.14326403143929367,0.40839329510845235,
 0.5739854604200465,0.7526032528180496,0.38577492847851247,
 0.05368830682081205,0.905136740834457,0.33682627207381666,0.5254939657695313,
 0.13905102320266283,0.5766625270178943,0.800760662466346,0.8504018344188741,
 0.14155366266335534,0.9592514970253316,0.23609580896558913,0.438916733746861,
 0.4272179601051562,0.8777999276248719,0.5022560880280924,0.8728402456840679,
 0.6096689815008492,0.013894385780150786,0.8177343530109842,
 0.8390012045496789,0.5316977072291482,0.4299326313595875,
 0.0021787297479712864,0.8260724433456776,0.41714693789148183,
 0.6672836256170885,0.008338679840810381,0.24417597611275554,
 0.7686261455662101,0.786056365881896,0.8502543505069082,0.6364766295517075,
 0.7241440288097657,0.4459973281691445,0.5944514024579864,0.47325576795109237,
 0.6394441556760517,0.8322715212864137,0.3633103458520329,0.3508568550456487,
 0.6142206303161817,0.8948006502420206,0.1356190173150552,
 0.029623741816056537,0.09630379922198284,0.1260738461136396,
 0.9109029633430619,0.705177339837163,0.6144412320257495,0.7941363042490844,
 0.7329142589329868,0.25212366412068754,0.42070479772731895,
 0.29122042333234865,0.7469551926921745,0.2736309445338716,
 0.41089359837902395,0.04393121838669012,0.7246361844245789,
 0.12130500270891542,0.42255379884024247,0.2261284912648417,
 3.2229944334871696e-4,0.7635130203922709,0.4561584224551225,
 0.603791914289274,0.2981188539000329,0.03763792399001653,0.3003072011401221,
 0.9277443465935421,0.14839314967099315,0.13822698320280946,
 0.5637812549158004,0.45643607814962495,0.24565051861296905,0.600788669113487,
 0.5770174939694632,0.11358405450512576,0.737959106742672,0.22907807361109211,
 0.24325043320368478,0.204561823148512,0.23687574211664175,0.7265894033058263,
 0.8236287131889396,0.8211884021045771,0.30235311007878574,0.8838504198646925,
 0.1034352552418808,0.6429901412662132,0.9252777198029207,0.974080345913021,
 0.2885152186978318,0.9058933382386144,0.9040358060303971,0.8100192432476748,
 0.380512531475912,0.6441562567631218,0.9566658887375012,0.5618197580464558,
 0.7336274108773989,0.4953686796755322,0.8453293718268547,0.4098804100652388,
 0.3823708494556277,0.8606602163087229,0.6603873634768231,0.9130161634420164,
 0.1871261339641006,0.7255093299647539,0.9293871886512413,0.08682528308747273,
 0.5881215481737451,0.8405201547476362,0.11994160149181499,
 0.09511538884205872,0.9434382364171975,0.8974411997016357,0.8793768404501866,
 0.8441471928223604,0.03873724762712483,0.2839819782136216,0.8860469211085945,
 0.4635277167777717,0.2448205217120667,0.5144530856545857,0.21038318487944396,
 0.7886315613822358,0.12801606000574506,0.3463959115090476,0.9642183637471543,
 0.9347807649199699,0.45611030706830047,0.7016423459412109,0.8320152408468157,
 0.523208234319901,0.8242587780779994,0.6898417516317751,0.38396258157201635,
 0.971334948148295,0.05257447906783508,0.5979994656495571,
 0.023346602875176092,0.15308446044642787,0.8715925735386586,
 0.7520619241868479,0.9484768366299566,0.20349223343848621,0.5397806613885137,
 0.21052704643165643,0.46260989878487646,0.28676185318902303,
 0.48826099188160865,0.841235607075562,0.27011593704608083,0.3959079144714628,
 0.4196836804206713,0.01742535425247016,0.407308826232747,0.5131068414863482,
 0.9983908326156312,0.6456468660314629,0.16726956083630895,
 0.31910610117162386,0.2596244492662323,0.6379986187882477,0.6028659825597961,
 0.12496303597017444,0.8131839536515227,0.7335891684953649,0.6304331563702048,
 0.7146331481633073,0.2813026682166433,0.42965637721053473,
 0.18798967583439585,0.5759436617695417,0.6660231856401259,0.6922673964261348,
 0.11390139153964152,0.6312457735202939,0.8056579503850925,
 0.29615461602064297,0.8391880124573873,0.3917298479372453,0.9597656758533478,
 0.403980775140682,0.8346251439347178,0.15813131819436577,0.05444593486923621,
 0.3568949187308319,0.36428314914144866,0.8630181087534258,0.5947950528269812,
 0.9539238130927674,0.30360393598359015,0.4223271694579235,0.8532421833338532,
 0.8484738236418742,0.40212251136264243,0.05597438772681529,
 0.7677215963171554,0.3016844254051776,0.8917123839775061,0.44537882612744695,
 0.9760457956296763,0.21031675607795952,0.9265356276704271,0.9943863999836497,
 0.48464244620191455,0.5030085625225565,0.6772653260621417,0.1579201371559179,
 0.16352754772884548,0.46371423879047,0.09384015134307644,0.10154581246101602,
 0.2251740060218883,0.06591250014312589,0.1755904541184563,0.7002097760732119,
 0.6822414893404695,0.7579116906230691,0.9191013145699043,0.27115323707205585,
 0.627098910082962,0.8238276544578662,0.7676349480545437,0.43309517496214556,
 0.1470547851388233,0.7230285696689581,0.8896580151810446,0.2025423161792339,
 0.13188794601313925,0.5199641134376987,0.11018395041863016,
 0.6531925107993991,0.5382534547249824,0.36248297680777974,0.5750605409757671,
 0.6484985451852686,0.052190719920732875,0.820946684406787,
 0.41551026303944505,0.9529974811034022,0.07875995509102829,
 0.9640805142258834,0.47239340302181565,0.9955537841061995,
 0.11637170097635441,0.9282980396409191,0.666827263906596,0.5071555197843978,
 0.40028541611183677,0.4229162240159441,0.8794240838633187,0.4831929146561813,
 0.5681897102160223,0.896562812596234,0.9905940643700396,0.24220896693915228,
 0.5656359326410865,0.16652551966990847,0.5364108908626428,
 0.39890717647090335,0.6797250353983109,0.975792284096293,0.2165609647073139,
 0.10495015710624678,0.19407399609029652,0.05606825257697012,
 0.16176290690119932,0.10389256489510956,0.28195030366130125,
 0.4643320484026582,0.11745421563774472,0.4831429194725855,
 0.03232304256348928,0.24132409500697793,0.7540387390089704,
 0.5236335408684661,0.7025919033584234,0.08824071139167167,
 0.17042658851364623,0.2940944728541157,0.785670378318405,0.37047056567122194,
 0.4487878149587359,0.9196260595194797,0.8955903485504332,0.43414108779652394,
 0.5408801574507562,0.35696750484005213,0.9579548523760254,0.2313912181800971,
 0.1636873559097871,0.5965028261017786,0.3953914734184637,0.9366405267705621,
 0.4485482735302968,0.7334791307569352,0.9418985172938735,0.3660074143192934,
 0.760714956249358,0.1939313083693851,0.49992136206733206,0.41837422421931114,
 0.7128538982123239,0.37155887645345387,0.7756223703236362,0.5672054991574786,
 0.2755909850529118,0.26006802021011577,0.7806260009828526,
 0.25887622520661047,0.28529463670632205,0.21747714846510546,
 0.5974680148224579,0.6007609916920146,0.34420664355674835,
 0.24309188780691737,0.278770632115957,0.4545184804830781,0.13983184386799807,
 0.204404397322133,0.9246319602037687,0.2985759278898362,0.7989837396780664,
 0.11873815900384643,0.6710731054051708,0.9496499447413711,0.6347644757487079,
 0.4556239196568437,0.4215609571677036,0.29446715863467643,0.4544372649936578,
 0.6664066887852163,0.7627900957123533,0.6755606050386186,0.7907121125854777,
 0.6843316012088563,0.4905044498712272,0.5232202926025423,0.8094394755020112,
 0.5739975655043236,0.7672811455647698,0.056304006355391345,
 0.15911240039785568,0.37279543636730916,0.40554255518681037,
 0.43669254908799804,0.14113866275099163,0.6590344672705104,
 0.3081177291289702,0.006166887580483893,0.9981429184382952,0.46005971532404,
 0.12594109100959006,0.058805510474548295,0.653061025010051,
 0.26265980094761876,0.9270419292209259,0.036297491120848235,
 0.7329149925882538,0.6540568720323202,0.7148650312121344,0.6609854084814001,
 0.33905968569511247,0.9128810021523696,0.8846798742535529,0.8387761580135304,
 0.33707045406269,0.4245124675919514,0.5003847287685508,0.32427332623528116,
 0.4768093406180618,0.5630325386043451,0.6640018673739209,0.5659312066992418,
 0.9764481948986728,0.1548250082712147,0.42129711568757744,0.9978112142267028,
 0.45092361224428945,0.026206828922414993,0.7890013232227133,
 0.5733234975125018,0.5877048130968213,0.10914293694275246,0.6184005021471686,
 0.0383605334900462,0.6303048208601576,0.862063617864375,0.9086415472663166,
 0.9618209872912159,0.5703081996734237,0.18612797051540575,0.7846809939918867,
 0.7530508396539908,0.014760448520620417,0.23681786127018312,
 0.9409538569045012,0.6055971617468191,0.35441586087461907,0.7478728252079118,
 0.6855536592328683,0.4397078083571959,0.24926549938947296,0.8947047928743777,
 0.8568748865908983,0.005614568908062223,0.8055803135717202,0.03695403558999,
 0.022283291340605915,0.6355647461360384,0.7094557226915802,
 0.3394188416479973,0.09847966756994153,0.7454892648750553,0.6749738501717634,
 0.20144139116637283,0.44493521438823824,0.10542229153683413,
 0.016454550539973356,0.7713080950205408,0.41936289635352275,
 0.32156940507339893,0.7293984523472077,0.4151302257362808,0.823570009152977,
 0.8387151084357148,0.17119811712320931,0.6608694041368763,0.7788388812079741,
 0.19109767755753926,0.2690933031859275,0.6881580495817803,
 0.14776375272077735,0.9583047370447572,0.246219012853782,0.10004916369741867,
 0.32581595496105153,0.03958703677192943,0.23815613912579747,
 0.09497231025710817,0.048460509596375045,0.029355991613470134,
 0.06075376499457208,0.3336227319508581,0.4356321675058604,0.9274898001403915,
 0.2443798996440627,0.06986308430238075,0.04916097861524882,
 0.5123042713958437,0.9065190048373143,0.613820866183428,0.6030706491462362,
 0.8582970347610528,0.45846821572653196,0.16322273538651033,0.620329041537428,
 0.5441637078709143,0.8422793202183282,0.9718460107706149,0.28465240580237505,
 0.08892904398641233,0.8965485890208827,0.2877246051119984,
 0.42628515567247527,0.5137776330891601,0.9293195155208245,0.2598325942290356,
 0.6263433309685851,0.3849630551173001,0.6441206501014685,0.3599632213294689,
 0.9219109665694168,0.8437527040091058,0.856313817692691,0.9096549352038472,
 0.13821477983941055,0.8976167491107866,0.642675624166056,0.0463610700640783,
 0.3097508009619364,0.2671807880498882,0.656403873185146,0.583636128060898,
 0.7847400451527147,0.3233158252725232,0.26667963816439066,
 0.29036977509415385,0.05993335585609716,0.6832865165671018,
 0.25136119030473036,0.10144331314660837,0.27790433244410795,
 0.03206931706288385,0.6690343794446345,0.7375669656367047,
 0.24110604064965413,0.21638257766842772,0.45668649768484115,
 0.27365714036902977,0.7991122427773951,0.049973875049217575,
 0.8630217250719316,0.54693144585057,0.5465436278258922,0.5557694253267045,
 0.12015952144121111,0.7162718626107681,0.3305183653066255,0.15468453781949,
 0.5622084779649721,0.5582265772038455,0.31518068434223234,0.0729860249833979,
 0.2918076254891232,0.8072204731254263,0.041909866939700446,0.699362229705716,
 0.8307396898840428,0.11355044255034019,0.6914607286736221,0.3749751684457504,
 0.2798341051895772,0.2954635261488757,0.2624748923131577,0.4864258880324531,
 0.5979448154216371,0.6385253803183257,0.649580349233538,0.4227884550058708,
 0.8614476076209964,0.1280200880645691,0.015084601458659419,
 0.6771660669705712,0.5914796205685628,0.1067626239894186,0.8040781111166045,
 0.6321235853098166,0.19614921573916233,0.16234328087884242,
 0.9292786567432344,0.5097245459679304,0.2944146796012299,0.7259282917638562,
 0.003188732809285888,0.7241069136920917,0.15479049544475165,
 0.28345599140452804,0.617350800819939,0.5809537712021002,0.5911974900730557,
 0.8357926522868991,0.5166550686869102,0.8048005573549356,0.3651511355419138,
 0.2749964401165551,0.7279071658622591,0.5957634087976355,0.1630940037525329,
 0.20162698981590588,0.3121454242583803,0.4743890666352132,
 0.08002254572255074,0.6927561374204851,0.16287503381164892,
 0.23650315951047085,0.9613132504791906,0.9601380503394175,0.4883970284211979,
 0.001306534723020225,0.8655052264980122,0.6082187721986658,0.637555356299944,
 0.0209858727764638,0.3330724083600295,0.574971951950476,0.15587530916177672,
 0.051000836029253205,0.32797573143017966,0.3223975402364442,
 0.7565165759689628,0.8747750008011975,0.6142593864168904,0.6944583656504925,
 0.9757512961904268,0.9160110762465332,0.25988901927457064,0.6720120220431435,
 0.5728964430159204,0.6602236733937774,0.30325025439316444,0.4899173268149598,
 0.19860277743718502,0.06288730706665413,0.5394115675046423,
 0.37966961644393094,0.6565630342494072,0.31041572149615837,
 0.7601398455850467,0.1465029609060572,0.1556824295612892,0.31414484858239544,
 0.3033933886382657,0.8129868260522588,0.872885263264091,0.6822473868449067,
 0.09176675199887763,0.051592760799762605,0.5636541807903238,
 0.5423902297135579,0.8262657726452208,0.737062337591042,0.5903689246482194,
 0.4208749272132032,0.8214414587133705,0.4752733958426498,0.8543320444796805,
 0.0540674640185862,0.3595546027093115,0.7381160564984327,0.7900628898545576,
 0.013234084033806659,0.6400250403375143,0.46672200426886745,
 0.8782121528558738,0.5098839113747428,0.38694069489221705,0.3847482106702579,
 0.3331001666745209,0.47366633196597074,0.5857908657690182,0.859571621522392,
 0.27652124224295227,0.8767452020521073,0.12236504994846364,
 0.14880102178365473,0.1495285722581442,0.049454072777054825,
 0.5870418780275531,0.38255347389012795,0.5815231331248263,0.6732048839292175,
 0.9828265252828465,0.30676304084993267,0.6405976433401555,0.7316159554739926,
 0.43627731687439264,0.5089814934910148,0.9800290140254188,
 0.05372269894735593,0.5662513231741435,0.3221429666258955,0.7375494207215437,
 0.7433612325772658,0.16302779801533718,0.3546617832884932,0.5500255268312587,
 0.2682486905078739,0.9973885401259298,0.5640767455107787,0.5507156074442194,
 0.5431633869942276,0.26277901261751835,0.41636688143227296,
 0.5485249705831097,0.06571238737379403,0.11628720722091423,
 0.7966090149319588,0.23534942910161605,0.6360406071462987,0.667586189696353,
 0.10205169475852272,0.20839086640213678,0.41109357637089994,
 0.36607807365373735,0.6851583699093967,0.08875334966529347,
 0.29934520585653057,0.1794061284359716,0.4728929099561284,0.5505098292061346,
 0.49012751883552574,0.7306535398733423,0.5369017090666046,0.6942488098402154,
 0.71861908678733,0.2475098766187832,0.1832999669558708,0.7209759905232482,
 0.1523751637285835,0.25541848653770427,0.7672191889807642,0.2090568547912921,
 0.9372714707495909,0.11848774617543134,0.246812384483578,0.09224763965929039,
 0.6694528401791958,0.15080026920480838,0.8586688484037042,0.6189835815942567,
 0.18727742146913584,4.968316123338257e-4,0.6980363667822064,
 0.8641559749762164,0.5657558919335762,0.24614235661587935,
 0.06596661024165507,0.7988761788262367,0.6058331314574891,0.3177881621442714,
 0.9142128639814643,0.10117131455959949,0.8489926554267797,0.7814088036259148,
 0.3236071515637511,0.04892445800027123,0.3047415146267074,0.9321866056656954,
 0.17946760546815088,0.819585685354127,0.6824836172544034,0.9725042378645471,
 0.8893846287885193,0.925110148438222,0.9073863984303587,0.8266375989061467,
 0.30075567359073263,0.8278175225603611,0.7088938952342803,
 0.23050478189545542,0.8283423443096889,0.5899224936340136,
 0.19977330563856055,0.2987117997612071,0.7033783314079808,
 0.36216271585928994,0.5955174199544264,0.46333352827274255,
 0.6892000259836194,0.9201772705049586,0.9084277831467474,0.20511157691851167,
 0.016178555161901587,0.8303518805951272,0.24768141615323458,
 0.20555924669590453,0.9936499183061464,0.015224692711225951,
 0.7536936163161988,0.2532693526178782,0.44287633874086685,0.8215495658140783,
 0.2494398261926396,0.2347937269616882,0.6008548387088585,0.8741676011618266,
 0.780639776341263,0.8069390110991248,0.26134664930734086,0.7373923015246654,
 0.03228732483218988,0.9864203187830679,0.025421214267133507,
 0.026955953677878508,0.8407654927744503,0.6888366871045366,
 0.9365021303417893,0.8318585196313177,0.3802085011454006,0.8473675685988259,
 0.016116539858707424,0.10746666793983861,0.464952978707997,
 0.7137304864858393,0.7049789630039748,0.7149168940602262,0.07213642723225189,
 0.3772348016514889,0.24009104803628023,0.49681470461257904,
 0.08685824100470141,0.008644095134262564,0.09737105653474865,
 0.025636709881636244,0.28170535692391063,0.2497993038831362,
 0.16561075797130131,0.8371632973997059,0.8604820905383931,0.9284527721047818,
 0.04541755142204251,0.560507525186224,0.18422429251500394,0.5450107187715523,
 0.7954995206856869,0.7753985762549581,0.3092426452300501,0.6512470328707712,
 0.2685343460316264,0.6198398664517608,0.5077235279994107,0.08549835118285964,
 0.9249779004531161,0.4494237141813078,0.3698713730029769,0.5075059101056923,
 0.8613161277455562,0.24771699458979213,0.07834605648172654,
 0.45874827660832884,0.16227853492749345,0.04516393680567632,
 0.16869900069863142,0.14044619526267554,0.8782290537168999,
 0.0705954995432061,0.9772639575033043,0.307144120487569,0.5301447900573447,
 0.19972193656213388,0.5679985543276341,0.7095206893083044,0.4686370704599414,
 0.9971941265270152,0.03619709422318329,0.05444007925030814,
 0.4795247728831612,0.9847743154417019,0.37629192430207414,0.3049149439112281,
 0.275841871632737,0.5897595509272033,0.9361260662335623,0.008216534209936288,
 0.43269313264441256,0.928718521814663,0.8257394693763266,0.17351938770971143,
 0.20500948863880497,0.6107247631520496,0.12916523582092443,
 0.1983521403869386,0.345423798033071,0.08309313677310004,0.17396357305567578,
 0.4903967722226993,0.14182796633207773,0.6687031799998399,0.73818198706058,
 0.4623442747775762,0.8272152510620255,0.9434156169863065,0.35633819860561644,
 0.37570393827730797,0.0693558673268897,0.23409693352674488,
 0.8153645242367866,0.6327006619061214,0.23799134899428764,0.7145162296185219,
 0.14386614029792,0.20742692576195132,0.6384676392561094,0.3389785129971853,
 0.2594556054050454,0.28474722553828324,0.728545545424893,0.8063196094033268,
 0.3152920674046087,0.959654931540413,0.41566663350320066,0.5725743966751216,
 0.1026406887744209,0.04196328651757797,0.07178514823353122,
 0.9570986093026475,0.557519936894456,0.9748000261662327,0.7277128630188083,
 0.9563180972227591,0.6318679478205337,0.3802388767404985,0.8350115532000214,
 0.07118184312177289,0.3755324898541045,0.11006623453399722,
 0.6023751864327174,0.27701866631873884,0.9788805567623439,0.7394465643833037,
 0.22344238312754827,0.8147073808591372,0.22270907142150176,
 0.5683049610544372,0.18974650701083107,0.8615689236567511,
 0.22779756224322845,0.6331855876574787,0.2520263558672753,0.9431610156320697,
 0.6003723086700392,0.21351370373059075,0.7982639935435745,0.3203610539268783,
 0.9895249632825079,0.6944730683361952,0.740040745357458,0.2466179917441652,
 0.040615155569398986,0.09943526174787987,0.25812510863932114,
 0.06598898151512034,0.9534540950384571,0.8792360693459667,
 0.04844022899757561,0.00351939016756464,0.556692674767471,0.5920886371464791,
 0.12154769010102906,0.8721031476740773,0.21890656140862952,
 0.8367244569570379,0.9821071350565685,0.04895590909486436,0.887807681823008,
 0.09185061218715784,0.7295678434418215,0.8749754061840641,0.7095652386344526,
 0.18481789174053176,0.3765950920657791,0.621562323505386,0.37279966515604623,
 0.4746820646962918,0.6237997789306147,0.6630584035508407,0.6802828532926244,
 0.2130427442298769,0.5618569587232338,0.7981513289360155,0.10801649351400744,
 0.9423012756111889,0.8482524301964344,0.010360928670689717,
 0.5092587632839514,0.5043652564138896,0.9634046881541046,0.38993023657932213,
 0.6678828991617977,0.7417296842778819,0.9541534128989109,0.7935132021628186,
 0.8337872055658212,0.7340339647040389,0.8181614364563057,0.8078392510398098,
 0.975474907397508,0.13832156265000028,0.690402802458898,0.34770686830451836,
 0.9671341737194331,0.24770196301378256,0.07516256970916269,
 0.5002659792719575,0.5894812943465565,0.5745586610482057,0.6473381271693858,
 0.9955611194019648,0.6148988271403595,0.9620307981692711,0.0645048115294552,
 0.3436060791473554,0.24886208121139997,0.8573983280269107,0.5108715701519396,
 0.6100256753660954,0.6955669294101097,0.8803057969311823,0.8963882739224484,
 0.7354639786117864,0.25670489449420564,0.30821895807906574,
 0.7024448206792958,0.7245269370804127,0.3615501507118999,0.4063062908334839,
 0.8018834532113213,0.1107929579669591,0.9383593340347374,0.7472248312295661,
 0.4131898495197621,0.3783929099162341,0.5891687439141545,0.6801966544005231,
 0.3104217974561042,0.7997441740787409,0.2016310857147322,0.17030708187743504,
 0.9407018835675679,0.5574568097125422,0.8162434927316495,0.592332106013123,
 0.1536295319600336,0.6441539339226993,0.6656747322294857,0.35169139340670696,
 0.5359442933916754,0.27905995914237647,0.602287905122065,0.02366415194185789,
 0.26386215516849765,0.6602486738161468,0.267448834438869,0.9899694645166298,
 0.7100925588907328,0.12011191069852045,0.780055853443623,0.885021706248769,
 0.327770326340879,0.06768126688963327,0.10612223508826746,
 0.09529374917978195,0.1529282793421043,0.7333316874420704,
 0.24595619045742811,0.59016823690348,0.5732740913063371,0.5299094893803391,
 0.33662076175081346,0.42342284690198284,0.8030305534757174,
 0.5697612659112427,0.03684349066812609,0.7861159045239337,0.1528555813854725,
 0.9009571737421185,0.2802936154117779,0.2919565622490917,0.1334508730312075,
 0.1506038796634187,0.9804111917038248,0.767839242422159,0.1815308829237413,
 0.752096150856491,0.9413182059340395,0.9812525878584428,0.19535187340893323,
 0.9677953632204294,0.7215653646686953,0.24380694116196766,0.2776519193586341,
 0.13994261986532863,0.09298151739373695,0.713316685205569,0.245092858503292,
 0.15870855592860966,0.9801584449891039,0.2880032237595146,0.6258221357479314,
 0.9858518932428698,0.48654290740885286,0.867095547020206,0.6613869548368813,
 0.7847096144133987,0.14641064130306214,0.21103601909261682,
 0.8080772677992875,0.8824474964163365,0.9749376504001767,0.9622218637231859,
 0.372732396584737,0.8532248908424094,0.8197874251115089,0.402064031117746,
 0.5389045637694101,0.08854152347886879,0.10947562293884117,
 0.19509309295452615,0.284773501391289,0.18838876648798075,0.952742072188961,
 0.6375211703132628,0.44912858359472096,0.5617054397855007,
 0.047397215789611424,0.15382054958953661,0.042912828193086394,
 0.3583864843254905,0.5261687230432658,0.3401231196723985,0.9254303431095997,
 0.8857073488718609,0.2445313431740952,0.7500690987407655,0.8654118800119655,
 0.2241580790837996,0.05764625635996623,0.02145162458869132,
 0.7819040878454264,0.6693016593968609,0.1233720815070628,0.7472080121475396,
 0.7272939556390238,0.277765915267705,0.6915844502413635,0.6429742842720255,
 0.3495482694494654,0.7447926389500151,0.10118091887153258,
 0.07761991119557954,0.6890157987763933,0.47602840879021213,
 0.7216894599070549,0.6830588447137457,0.600328989219167,0.1802788282130865,
 0.4417758895763899,0.1867117644718097,0.3320683477306615,0.7248695052098351,
 0.6421157452143887,0.3318411062508255,0.10390190441878255,0.9342854152967848,
 0.7590152382959743,0.9322880239736498,0.634455251221466,0.22828004918465927,
 0.925933640370785,0.9674708207697849,0.1712475277100387,0.18551616885746702,
 0.10068324698424436,0.18072263990799076,0.9226628387433896,
 0.0772343627477331,0.832985388078112,0.7210985631298739,0.4441919506351617,
 0.8843119529596978,0.8488130258410727,0.4637562144426206,0.8050408796331335,
 0.8988675586200056,0.6812714647397368,0.8309431189084168,0.8074006089708725,
 0.08290611254591784,0.8560276518927713,0.32863842140596244,
 0.5754581254465371,0.6640324659399183,0.9433478012596667,0.850843447868157,
 0.8188669860688104,0.9405692911103511,0.9451121429168331,0.8418883349322369,
 0.020246170143994546,0.3365498805078757,0.6060020416323857,
 0.3255917756448762,0.7380910354463731,0.5526508646529933,0.3634237663064873,
 0.940374189458919,0.3767703701005851,0.4384181644364433,0.9094933835063734,
 0.4269354771151812,0.7753461654911464,0.14203858672311687,0.646513075634151,
 0.22408571333891214,0.06559737113317365,0.24519602564360798,
 0.7911825135272995,0.2189741899675316,0.2696987533923697,0.5437930401360256,
 0.6117562514092947,0.3267955138138464,0.5464473308234276,0.5398780122908907,
 0.5868212889206008,0.03248323508515982,0.2162170126416343,0.6362942294331705,
 0.7749929916564358,0.1849652814013708,0.17133648824673042,
 0.36225965666999005,0.5266978127488358,0.32963721557087733,
 0.04334550886756705,0.9831384538674331,0.6621832783026528,0.9554557200924767,
 0.3500042233129783,0.4257124026304754,0.6630846882959348,0.9003502330015459,
 0.625858235737076,0.5417981423654444,0.32264711116694533,
 0.045622437201664146,0.5993774856994789,0.2955762814027425,
 0.3674411773995486,0.9551581529160262,0.6030667204352622,0.08390340087193948,
 0.39882648889798256,0.05509460338065619,0.19105941155166417,
 0.19046270558079526,0.8124388267229317,0.41091550093985596,
 0.6318509087980816,0.8571477555589954,0.8270135555617263,0.34361692646561626,
 0.5642450687219664,0.1582409100349933,0.6672706307207921,0.24260832239717312,
 0.09817326946416571,0.5193424638222864,0.8957124019275061,0.9848287734249502,
 0.2792724236715013,0.01580526094579826,0.8558510658914961,0.7122064452622059,
 0.5235280404855952,0.754647574961265,0.15274240015574758,0.1592556538616331,
 0.6615268190601021,0.39684558916868085,0.22754183963341446,
 0.21177427331321852,0.9733138332749363,0.46131752136520365,
 0.5325638848187175,0.25901268943866307,0.27537579452508354,
 0.9273127557643464,0.4515508166220442,0.4315460786390639,0.7607674741008152,
 0.5949092201798492,0.024866505459197885,0.646333034060086,
 0.09338434269445373,0.7175600647306476,0.3017645688321917,
 0.017301160533617854,0.7538202032431092,0.0726720405750696,
 0.5335827379697251,0.026786478515917533,0.2832768365234293,
 0.16689476543738824,0.44555045084443656,0.9310361481471972,
 0.5952984630880638,0.3376956009461788,0.4726581890048175,0.07351411419299303,
 0.6843420696845801,0.20304474109513182,0.843633760885967,0.8048619931949005,
 0.45653118794676817,0.10839905473565958,0.8122112841714731,
 0.057635440965884355,0.49368282622214443,0.42816759637419255,
 0.6866982610381331,0.008713207229251063,0.3910649346748536,0.285190502608043,
 0.7481490063646893,0.6654739418435677,0.6592133003493496,0.9410363270052988,
 0.6387227967256798,0.8984027846554468,0.14792675608074124,
 0.060348620464746316,0.7964692743513774,0.6500497944500924,
 0.3800666332490702,0.05120762108796917,0.6994043024267242,0.9555313666897804,
 0.5871004665177716,0.8781524179868626,0.10686879952468131,0.3737490067525866,
 0.5264512601364241,0.6730277780763826,0.5432680776020516,0.8247236477666369,
 0.9835840819826296,0.42544338764289147,0.7376427829713378,0.4839350895909753,
 0.7657147031656002,0.4464362850276907,0.49407306506928383,0.5468384151581467,
 0.18874035557668767,0.4205107039780601,0.8305033809742546,0.6577265081309691,
 0.9111953261666345,0.7943724287287754,0.267806710602812,0.3217508758798754,
 0.503579185963086,0.1582977189965219,0.2338180337837099,0.9644195049722265,
 0.7302837522084304,0.6541480304762334,0.9612430401568137,0.5161809375575779,
 0.5065130422283681,0.16862920433370454,0.022286599441907806,
 0.7620903205223856,0.3290535095120054,0.26320146015070733,0.7068094468778292,
 0.9287072014936574,0.12851597972105522,0.047390231772937885,
 0.41987587964247564,0.5810624630980521,0.014987113477870184,
 0.12035802589430444,0.0867461951889863,0.22992989870843716,
 0.7884681038255407,0.9956620885105646,0.8337242757212655,0.4191633715800922,
 0.5665436958671418,0.6517906727878101,0.8421675661420585,0.8020410902304802,
 0.7287848424212671,0.5356316522435886,0.39037787308882455,0.6356428164995719,
 0.23804160853255818,0.72323641980752,0.398322616839331,0.3926445586458269,
 0.2700473417445338,0.6304035901272371,0.3498814982109739,0.2354327640229913,
 0.302179781606584,0.9391643995889981,0.4995415107363508,0.7721545278744205,
 0.704694551240348,0.6529905633075266,0.6007709481895112,0.4943078349443426,
 0.4526470585687623,0.723427501903773,0.6642556731011217,0.88075323528181,
 0.7423534608799064,0.03455813515766104,0.8527075405588544,0.5194055630835535,
 0.5663332933214769,0.8655692143924344,0.6486090834703276,0.08397715603860856,
 0.1411562284284048,0.3540325777031531,0.5359256642706381,0.62509660344859,
 0.16444603112375222,0.17873254945132888,0.5092332147680665,
 0.3802200493848925,0.9957827790434598,0.22980284826830266,0.1687019573970857,
 0.07075139633917371,0.7764121558054312,0.9162761293554171,0.2842580252225656,
 0.4184184941984824,0.824014651186901,0.4434571483424671,0.030541691255999592,
 0.5219193744595908,0.017160970844323886,0.2205024333138923,
 0.5636505704070996,0.07297774302307558,0.7106335662398974,
 0.14738605396192195,0.24646842311563644,0.2085226405844759,
 0.004780716894572046,0.8578608467950102,0.08497547871531153,
 0.4829283311370388,0.6825570781777757,0.32565532781358764,0.3882470887440356,
 0.09509809765263244,0.26276602407856364,0.9011778214366157,0.745323448928665,
 0.7977089951338607,0.645163724618139,0.9732239193703912,0.5392413420742908,
 0.5626644580145629,0.3869088741506833,0.414753485866204,0.32170834150086125,
 0.24411840115766603,0.3372294764963599,0.7769773084473193,
 0.18084467329744325,0.9434168060250353,0.854399268082908,0.5923899783224537,
 0.7563378730241511,0.36106535162053754,0.03269885816388263,
 0.06093262789794518,0.6039112657862029,0.4522646217375158,0.7214408563368495,
 0.3820634911380909,0.07473551357871633,0.13935312950587986,
 0.42294165468002376,0.8671398355022545,0.4335105716232721,0.7446608289665667,
 0.13178239367522449,0.568121627638276,0.8502296596677716,0.8818675123828622,
 0.08771447680676092,0.9573952725517815,0.3869493360001012,0.7054511435109471,
 0.5802204129834482,0.6966399668010281,0.48087361506953163,
 0.23116169370442008,0.11973754038207562,0.3592704857315693,
 0.4516566961669344,0.6985073286153737,0.8949735287606899,0.9134488561564932,
 0.37220683036028457,0.7096334721865073,0.6296524829051989,0.9444929990195666,
 0.8474334829827508,0.5561728552432613,0.8967161506647315,0.9434916075427617,
 0.012245383445670721,0.9002798112741901,0.2657108460338683,
 0.2757602718343801,0.10481428866774789,0.38593689885053184,
 0.5327796718689248,0.4301743386892516,0.8938722901860818,0.8111147327909446,
 0.7512756544038872,0.04793776105621406,0.1068016772688739,0.7026243869452311,
 0.5693341535269221,0.14607634505013,0.5124968821913806,0.3482655903668481,
 0.6956073300825242,0.12911440271162655,0.46803619389034967,
 0.20991326059454496,0.2451076026149771,0.6187342162680767,0.5707517182641708,
 0.8269913640919698,0.5067397530884133,0.15077408939732495,
 0.046135907331854464,0.7986523418351272,0.25716803736904625,
 0.6312041990366646,0.35032523084917766,0.8094243682711119,0.3319889586481547,
 0.9147834689633683,0.4984174653747597,0.6968842679282078,0.3771276355251989,
 0.2109623516959327,0.9240429381818506,0.5625391624138449,0.5809286109275269,
 0.846731526140121,0.6462433070853849,0.03602372618880556,0.33290849543514656,
 0.16387829965710032,0.6472541771168474,0.0883490469124194,0.6829785192059706,
 0.49139016340585606,0.06125696909525924,0.8921045363060203,
 0.15301763961137915,0.22094114975680057,0.49888057071656366,
 0.8101216826106408,0.4175705172331269,0.11299302914914877,0.7554974923229747,
 0.46464446393889436,0.6075971895515073,0.16147229163244892,
 0.3208360613850061,0.07571833863038513,0.2772262023233295,
 0.26415421443911513,0.8452292104541518,0.5221825799810614,0.9788629241625142,
 0.4345572742084405,0.18547921265118117,0.8895326113198254,
 0.49587984771008087,0.8114495823546489,0.21237306726052296,
 0.7676638932178942,0.9847709772650707,0.8033269103660369,0.6543085479234073,
 0.2358714045744561,0.8841020276801889,0.14686751641279028,
 0.20485181867608904,0.9316293272574685,0.10629694157815561,
 0.24383228249878686,0.6620110625307255,0.6466163200256921,
 0.14350007993979896,0.5611887280214221,0.9316854040549266,
 0.26719889173426536,0.7481626899526048,0.4833305853599975,
 0.17630395471096372,0.95436383585399,0.9496896633412344,0.17499500943233925,
 0.16520560535774953,0.03142852193939205,0.5272955396138228,
 0.19806974419273904,0.08966232411051878,0.9728596971925784,
 0.7192015707433036,0.6768840113030823,0.16668200103780972,
 0.10432589990925278,0.26847227262429985,0.6600711596155648,
 0.7786043042218934,0.6849709430440766,0.7268190257317559,0.5794164427330567,
 0.9610913203033038,0.7597032143303336,0.04398578848434531,
 0.24260145152668544,0.968220862334323,0.7821328952300507,0.8139852149671826,
 0.08092144199869544,0.8073621197997083,0.7073961293643546,0.8273192703260591,
 0.437203567794043,0.20238516990500255,0.5315709928550424,0.29613684988226596,
 0.12987381829402278,0.3123912237051284,0.1900351799250144,0.5963012782345578,
 0.9484655693634452,0.7194307301940213,0.5388548786703625,0.7021754093621594,
 0.6029730530688123,0.7753107019026964,0.8104624579502164,0.15035145528686766,
 0.32596962783524885,0.3554825267677849,0.936384064448944,0.22128382051498807,
 0.6875795373323911,0.7428037851839193,0.2953412512241207,0.9042889157454719,
 0.938244722955132,0.7183466358193273,0.10541551543683358,0.9379843351577978,
 0.9988944306086427,0.1977316676330213,0.18320708502952798,0.9514942958207415,
 0.30903670855835874,0.7805977558203925,0.056726189210026035,
 0.06645730265133143,0.181747523779616,0.6788597696377098,0.7482865450197307,
 0.04576761172975785,0.2818341855813141,0.7341503239465582,0.4134682730782421,
 0.6882513332718732,0.025345167645834166,0.5758130690232995,
 0.6415401044373902,0.555056935893753,0.3268290642788486,0.4887478206920196,
 0.728818437251451,0.8176666458001853,0.13139018471839803,0.6722197820379752,
 0.5432390660920483,0.23659995805010736,0.6079607647055819,0.3021170029463328,
 0.4377845390772015,0.7388373377253231,0.6978667594158467,0.17705805237045902,
 0.6938310760108544,0.7164566854462834,0.7344851123105828,0.5638072829450419,
 0.19853327515028174,0.5745369739825544,0.5978639437464349,0.5250008898537656,
 0.28016887346657726,0.026880446747349795,0.5788450649023238,
 0.8605311702738178,0.8826899939636492,0.917320838840989,0.3464192338306791,
 0.06672358066821371,0.5316759898216261,0.03976716369122857,
 0.7424213948642504,0.5287700806599095,0.4590886175606088,0.8457680610206927,
 0.34097535164965487,0.8761314089100968,0.7748799263329009,0.7856544608343066,
 0.31382523677801333,0.7177971915525229,0.9021898789077168,0.6377337873774881,
 0.29475301811654675,0.9963975752225478,0.550170266028893,0.4873894975535673,
 0.09261788701623208,0.4943648746343696,0.5890473470579594,0.8526967920942365,
 0.9542056263594186,0.5365964090188894,0.4344015642301423,0.7131339808935607,
 0.2781697119751584,0.24505147510809688,8.771972204035805e-5,
 0.012048648250904237,0.27042944903171695,0.10235930320064857,
 0.7879161605101979,0.9353207172724848,0.7689369487467793,0.466595270780392,
 0.25421267407292203,0.18175019904861267,0.4351203772214417,
 0.41142306908410986,0.890044556505573,0.405389040076777,0.557737706961838,
 0.11082595920583371,0.26744878957641205,0.9173112609082286,
 0.6768812790560044,0.19721385540711633,0.8332001372004143,0.6736821782242933,
 0.49605038469592555,0.8485479057902984,0.2934386086060182,
 0.09582405306602726,0.8371410593525032,0.4573075403632413,0.5347046954653187,
 0.07285255548848557,0.7401436541822793,0.36551873784885114,
 0.05796823773194959,0.10585128505852981,0.7147774470998701,0.579909885139756,
 0.3489541611019815,0.03173189387054365,0.5186782573931589,0.9639750953144474,
 0.13940206060412708,0.1220293817847059,0.3793438986258274,
 0.21473926761706763,0.7328979351273162,0.8734571254520501,0.3063115507675237,
 0.019049022048770126,0.6098305601060428,0.7593312154127886,
 0.052235454498638756,0.6336266999497217,0.7694383538406226,0.399368274176533,
 0.29522256108645883,0.7129364056907946,0.1510666920376097,0.6132376515910636,
 0.9427181103836981,0.2520659746583547,0.7476165574361814,0.4183403625825842,
 0.08882562889420376,0.27349162768293334,0.16177935417711964,
 0.25452333234232405,0.14365204762428807,0.5438372293516324,
 0.7576658826486372,0.3136539937507874,0.039699957040531864,
 0.4870547000063168,0.5543124733087035,0.6600781722552745,0.7714965448314637,
 0.5559596269085813,0.3841455152178216,0.3855614247978326,0.5279110081744391,
 0.9242047623581073,0.06437540869211356,0.23967292912865668,
 0.1835104385167241,0.7319731805391305,0.5501366978821814,0.0834365252589061,
 0.5950157344704202,0.9381182955286786,0.18749226408289932,0.6111805940744802,
 0.6116559948799348,0.5651395615093594,0.03868989032253545,0.9675781719969677,
 0.8727068126433764,0.015550154051125031,0.03887172852446452,
 0.8190537775835204,0.19713227260964516,0.16061360543275716,
 0.02300679212975254,0.5939830003212296,0.5341717532119302,0.9389657566758717,
 0.7271361216054069,0.7596522928771409,0.15002550720295182,
 0.19882340583284486,0.8355282423552186,0.29586110292193535,
 0.49810689456021096,0.9030238511735275,0.33477876744468915,
 0.2625958817625448,0.25486765438943126,0.12341239745786825,
 0.05320839888957962,0.11201642665138722,0.812302919415384,0.9142867060922847,
 0.023619996333395066,0.9809169544997974,0.40771947695706945,
 0.033992936922373485,0.2137930975118083,0.12127311240694016,
 0.42880957058220615,0.22580172194834858,0.7060005203232774,
 0.2351007056848513,0.3837835436009708,0.3458800736536891,0.2636781929567059,
 0.036044510929272744,0.10781559223502601,0.42583686177340907,
 0.15028823010532888,0.05931104622329819,0.40829497596760733,
 0.29961050507768194,0.10628701913429028,0.8559673893049459,
 0.6548198106619906,0.08310894488578169,0.7778146702165247,
 0.16129352939013797,0.27993264117614247,0.31728280976263645,
 0.1428658211102083,0.4804813061875084,0.34358199578982473,0.8479270531128691,
 0.5952527093788196,0.8888018994430903,0.9446514741615766,0.8360237906216208,
 0.39145840799818066,0.17087740621315572,0.7019594193681815,
 0.9948673654753308,0.17310767433740593,0.6478096326246507,
 4.687615898311126e-4,0.919195226931552,0.1996230334574638,0.7037514659765619,
 0.7875446438414133,0.278641268477383,0.8095940755053241,0.6066182888283912,
 0.8936054281398622,0.7112316960511915,0.15026752104589491,0.5846277850571552,
 0.6398991829407745,0.7364437971733548,0.33779453651911207,0.6647976492270345,
 0.05449230895278134,0.059722261409720145,0.3592811445825783,
 0.7267529132265491,0.302597166338919,0.31244514277620083,0.5353444683459789,
 0.9164182815926981,0.13570322244852417,0.6084549679456647,
 0.07340055283657176,0.389642124891063,0.6509224130937615,0.49048868033503934,
 0.2201565276078754,0.18591184264086424,0.3864359728410751,
 0.14390573588724442,0.9810705034181391,0.6909683979218664,
 0.33861744054208565,0.06741979609471416,0.7764374813669419,
 0.34689433712817064,0.24863190991267725,0.7485680238679177,
 0.9615105303516964,0.09490445369935785,0.03827224265541351,
 0.35483054445031526,0.1439889446996413,0.4716270252732313,0.8020367222278273,
 0.734042095646377,0.3336857806506972,0.45603507280369937,0.5291402355629727,
 0.9088545079053547,0.49148441153357636,0.6577173983334939,0.5502462648951382,
 0.6035044731170645,0.5184774343459367,0.14316181460261956,0.2726592560378376,
 0.25586348465568376,0.6491850791263939,0.8911800106569459,
 0.39241849854112787,0.1478979922890864,0.3537709910444651,
 0.23222546835311375,0.6888195005404733,0.5254933816068144,0.8763103785717823,
 0.1369489052156111,0.32325006797221645,0.10660223503069144,
 0.9694186826147061,0.618746278955817,0.3360566732498039,0.8784665481886209,
 0.8637836203183179,0.06861328306865289,0.9066196818898811,0.9926507748619415,
 0.6556462655276721,0.4706817653805584,0.5715967552363003,0.20382375094310412,
 0.9498136378402071,0.015232736670818037,0.3505669761608663,
 0.6948797108989575,0.3980396579885056,0.4186782232902989,0.4247966453604246,
 0.7214965565388749,0.17881101503353403,0.6054383551042481,0.4158089137968356,
 0.2912905456538748,0.5033105263248873,0.6950037161029816,0.6247026032571759,
 0.6204223633699842,0.8862696664534924,0.6156593669983127,0.7686869030228682,
 0.04662317003645389,0.1594803889500338,0.41799992717293755,
 0.7907929191685228,0.8568237494157123,0.05933497345437788,0.6886861189230153,
 0.4617523882194061,0.6975015414367026,0.24548970968632922,
 0.34977205910684206,0.21345531118679884,0.5459722437709036,
 0.8054353625492489,0.9289953688059823,0.7459701959835358,0.03626061868325525,
 0.2470448876735838,0.6043141612201084,0.565349057840272,0.06523503553865628,
 0.46193638774786316,0.2612527890073727,0.5896991102251548,0.962379355473106,
 0.5227165450950666,0.04069663351740682,0.06118575050828301,
 0.7994352249461337,0.7660461422182088,0.42213379295666886,0.3281474394330174,
 0.43892743534975587,0.5851863151636802,0.29786497492194675,
 0.36555504137564254,0.8452912600148067,0.7571748744717088,0.9155427070652982,
 0.64148457394323,0.8624999346465012,0.8193769931663653,0.519704228485024,
 0.3432987208128486,0.38695842472022157,0.2238251638931431,0.620993950949112,
 0.9137896588517517,0.27622022583907113,0.8379307577851225,0.6141889557881749,
 0.7808108666081832,0.08945341494439096,0.8274633656665178,
 0.18800986928640662,0.5956898134145371,0.9967258318641581,0.8878622835670251,
 0.8367776468041023,0.0619395242674613,0.5916403557640078,0.5944316387590879,
 0.46401768183368497,0.913170980518361,0.27120992103670394,0.5893439970911805,
 0.2723725857218995,0.23691622239498533,0.7134167028645193,0.6105794130684453,
 0.6553543603526286,0.17252542997234543,0.43329585743586374,
 0.46852528338409294,0.6355706584968247,0.05702217976150625,
 0.7068058466613997,0.13426638848524497,0.9085620045683096,0.1190224983088175,
 0.6605513574613746,0.11373692365411603,0.991761235569306,0.7887914382390281,
 0.33665782100644004,0.28489278577457744,0.28068553849495037,
 0.18113861959555166,0.2591684504427004,0.47150594844386395,
 0.008757718146169324,0.7143633671084205,0.7265558816694689,
 0.8881868589537509,0.36662073902223125,0.5872181559702109,0.8235862061701272,
 0.6430632820487895,0.20719202271678927,0.623939654278692,0.1712271111208884,
 0.46357476735345826,0.8157121842319612,0.7276206668573081,
 0.37790641422353477,0.8273324285521982,0.8161022779711834,0.683483825923786,
 0.821283596650035,0.1929999645144298,0.5592529383124698,0.43181529711631705,
 0.7913256483548367,0.12947186858448623,0.4134940144711313,0.6278187240763715,
 0.27445924318534487,0.9993598154083021,0.22135689919702906,
 0.5312182976290569,0.762976337416986,0.7770861656996753,0.1198013452156278,
 0.2700720246705507,0.4375635790461735,0.1534837676717532,0.9610466766954046,
 0.2884393959464857,0.15920711744817584,0.23530419163856198,
 0.4202646005764066,0.554128272241698,0.623850807400262,0.2716598384990273,
 0.4742413519342664,0.7644184757467528,0.3511291248050331,0.19448025390718038,
 0.5329587651871174,0.26032550539501587,0.6905575130442452,0.4068230871980403,
 0.9470955449179544,0.06670193385438816,0.3570735526283254,0.4987563101998782,
 0.32013212103198607,0.9529660235106256,0.3586257169055421,0.8638971212850302,
 0.5689826772877101,0.21592622701217823,0.22122622142780823,
 0.4454343730647583,0.8523903811766448,0.2904106591135833,0.9380797430788004,
 0.21547642087021046,0.7222099019970213,0.46455302149951894,
 0.3634104865255673,0.49419625734901795,0.1305800700646309,0.647415820090364,
 0.7558821853396405,0.671215844864298,0.9263979250069648,0.0669902256223801,
 0.2820818739899461,0.6846192248517127,0.05179778034357019,
 0.16203741041443742,0.26618913688878176,0.9870303699237686,
 0.6284080555209455,0.7385266880559044,0.5488783383380245,0.3035521700446322,
 0.609527165866047,0.8312277081730942,0.7929371407424861,0.7417979365505624,
 0.1577202286681466,0.7669456397282924,0.3171801320471723,0.388435974676014,
 0.9001724864683317,0.6920476511381504,0.31689932569044543,
 0.18177376829384162,0.09926323546086391,0.837715687570417,0.2976401847024812,
 0.911082532827775,0.845230585628013,0.32953037384776085,0.6750039821411419,
 0.14296312931945887,0.41138539082267245,0.5025680748744199,
 0.24660096173452573,0.8729556504832525,0.9037697007526748,0.8650234571342748,
 0.48325985403214033,0.9732264786601917,0.504012201263388,0.1375067038768747,
 0.40366419919086316,0.4389334694425602,0.12239800142762303,
 0.8063603913397177,0.32376816969481925,0.6787433689939458,0.5234991253266877,
 0.3896465619236874,0.3891642593562259,0.5299521787482037,0.18787668707745342,
 0.15270869764538564,0.8355108639921983,0.69479892283511,0.9600642309664208,
 0.26750738044428624,0.6909077338953686,0.04573799186651095,
 0.6709760234153513,0.485176237756064,0.3850243845609459,0.35218256295696215,
 0.01217145887997595,0.21628212957025728,0.5025793140446483,
 0.46088099605543476,0.07620964243982997,0.8340719989687206,
 0.5409561205186324,0.8221751573271382,0.9419757689368614,
 0.031344069472382885,0.6093481459360973,0.030740563733903103,
 0.5430464848097993,0.463028857477646,0.04438406138422768,0.0567290387260887,
 0.4217852939218666,0.5706772581504892,0.6929419455193246,0.3745448146004048,
 0.21989793730542906,0.8043264927796657,0.08308984775436312,
 0.10687824764048093,0.6543274964403272,0.4562143373510116,0.6743469011603255,
 0.20891739799240105,0.8411262270309547,0.3196260948091295,0.8331269664758543,
 0.7015765013041524,0.9419238772036238,0.844104723156033,0.30631440980150326,
 0.1887763799868616,0.46920750993436633,0.11510347416109545,
 0.13941393174124395,0.6244669887090482,0.10536972295739688,
 0.9444673086490976,0.021409414253694803,0.87944460953737,0.04048055113520721,
 0.939218667351617,0.7633596904385601,0.8077931559584757,0.1856414693473123,
 0.2504027948030121,0.09222704830463235,0.6690567435192684,
 0.18410623204921372,0.5236797512819131,0.28500646804414687,
 0.10689658661068424,0.007760473722983385,0.46417548027932953,
 0.8633798706587318,0.805577393306552,0.06859150847777318,0.583427796509605,
 0.1460681638667154,0.7921606320422134,0.25668839097905705,
 0.07391978499895857,0.9721308289917565,0.15276993209104317,
 0.7013932452676866,0.3369076376074618,0.7015004879174165,0.5156212124281757,
 0.9293693350377104,0.6876196226536319,0.43805881111737077,
 0.27299067820068834,0.5837292164229602,0.48247865724832595,
 0.4271575199441259,0.12982332181764267,0.24115647664055673,
 0.4060962083664522,0.12437888062864089,0.9037708171818502,0.7888588342923114,
 0.09800293907362811,0.4184964454450727,0.21616809959927752,
 0.6666507824631478,0.9339628005169329,0.7834075161194977,0.9927435586051726,
 0.5903581728771026,0.7028266927231249,0.5354546184506188,0.5490403417588339,
 0.12727203081359473,0.8191442439205087,0.08913572672316494,
 0.2122087215902353,0.041924476095119845,0.732272056001272,0.8072851478056141,
 0.5404217026270609,0.3789675980521454,0.8409719444983126,
 0.023452113539388697,0.16159840592141617,0.845698482176844,
 0.21698722190394015,0.9227233585195397,0.9398044225304567,0.0172718383824908,
 0.601823658548701,0.17637349626132537,0.8958127512625669,0.20643626887992672,
 0.7128424652583695,0.203065223838219,0.6442849275964522,0.15736417492658983,
 0.8251916655465181,0.022653668919591197,0.02574401728662834,
 0.8649503361327764,0.8548289238552457,0.9499550309928012,0.10674413844927522,
 0.7623722438601338,0.16840106141304068,0.4016482861716453,
 0.14445342987265275,0.9671360934515228,0.06619037367345448,
 0.44118055766399356,0.4291274784138869,0.7600155526270792,0.8553104525691976,
 0.40581735815335895,0.037421750911675655,0.8710012900980109,
 0.8528409260379951,0.5167381810751547,0.878221929048363,0.7267379746159874,
 0.645046787553458,0.3207023523107331,0.9981287706149921,0.06901599484875975,
 0.032935896207357485,0.1364379011885929,0.0064404345408786146,
 0.3040841848459862,0.526452429937204,0.9954538066214729,0.31194405855040197,
 0.8814509941513257,0.1100993348707977,0.6622746810400071,0.7716627663938996,
 0.0709115965536613,0.3657623921707379,0.7181690335452402,0.40398027490546906,
 0.32502355617180667,0.8419988282477635,0.8412232260228603,0.3308194086772317,
 0.1657714742349664,0.3501043125084189,0.4139039438120601,0.825276457799925,
 0.8869799913995432,0.18866852640916532,0.6639689287389812,0.3849842203163776,
 0.04794908298240519,0.570187483815332,0.9414886600827784,0.3702217144142922,
 0.5544936641888107,0.3374619370969473,0.8660540266782265,0.2751857836853202,
 0.8718768189004084,0.24486673452890162,0.8435020019124844,0.5057527198401537,
 0.6234247591364704,0.3964790126617349,0.8136474339895989,0.8626557340864446,
 0.7522341503916659,0.612482988724464,0.4408524536940335,0.008349398441708678,
 0.24769206838338187,0.7490202985947395,0.01840461984588393,
 0.4903515763137518,0.8838437732566633,0.4190793485158619,0.6678499368062476,
 0.6656701163846865,0.17416278375327132,0.09986167610485897,
 0.14541909847345735,0.24251201439285242,0.9211083855986063,
 0.3226836301008402,0.6229476650824332,0.9103240119427586,0.8333951854524162,
 0.03257453780530828,0.37690950494165054,0.46007386485092605,
 0.18218524583318496,0.7509246227266607,0.9651162166417931,0.9701935328760548,
 0.42619369823645126,9.016398845531626e-4,0.2862161335026646,
 0.7367673626362176,0.21121129374253,0.22931282293193878,0.9432292827432829,
 0.9910032042119885,0.29025068955842825,0.42754078506536153,
 0.18416654727216808,0.6579195202590384,0.827272208441153,0.5801980894238453,
 0.3734194429326436,0.5138042047745779,0.24732053588341796,
 0.37891239392079745,0.5204661141929277,0.5863117170452786,0.5551122168701265,
 0.7240679224885457,0.8211229836282588,0.7453770905931707,0.13368497418582792,
 0.9341122655327103,0.07407490294532915,0.2337081976564106,0.4018289116160011,
 0.08251114594941467,0.9162763828306374,0.6818085335898698,0.9586085030971584,
 0.479555053355026,0.36378025711482664,0.6120826117145024,0.35215105365035015,
 0.901552108741799,0.9997123139473154,0.9092878495213643,0.5091192311658673,
 0.08245446870290929,0.942131459952108,0.3051567616382276,0.1338066061605987,
 0.8030809469211507,0.0026398393310792523,0.6238807626177351,
 0.23937292726076143,0.6952960068013195,0.6929708032413877,
 0.29803146087635124,0.6791467173792767,0.8277622270850545,0.8094896333607322,
 0.6822851514728882,0.7054302657514008,0.9641996414748919,0.8843789266857214,
 0.2898609885772244,0.9063282290276831,0.2307121347829646,0.01612385354883794,
 0.561161144245199,0.7211859063211863,0.7271966483123931,0.9538684218331266,
 0.05967727790146227,0.7362636533851572,0.0820581086146045,
 0.11852376747156934,0.9873890599063716,0.5106495898637889,0.8546585333829297,
 0.37357583736885935,0.6451246398770931,0.2363006908041969,0.3459946668873908,
 0.9943069979242977,0.4435357462304015,0.6754392418763163,0.4905173359859375,
 0.40141560608455096,0.33226204068563703,0.16180938282019663,
 0.667849847784054,0.42120264682758324,0.6723524295761629,0.12672087269600696,
 0.7145450923157767,0.7237832305749571,0.021953597222246768,0.473482356409455,
 0.7881653299588501,0.9947658922674623,0.5542025410975805,0.6646729172418175,
 0.6734796499246251,0.3060327948373771,0.058749764619632217,
 0.3200604931573746,0.1667781145846554,0.4811105085720574,
 0.028151714924570914,0.12066234616324445,0.44040214832394653,
 0.02322090589839132,0.4331331637331107,0.19466935118936968,
 0.20388605329324427,0.8404172922848526,0.2611277743884014,
 0.33255480936359816,0.40054288801247795,0.1590453664388375,
 0.1270567083788967,0.6390909180480112,0.9179896015200443,0.9365206770654897,
 0.4183048141377934,0.04061037853349314,0.5401000323912833,0.3767739811382411,
 0.41840044112107544,0.7960993715319116,0.18926224801191438,
 0.3498953845127125,0.00796378903447681,0.4227922150378224,
 0.06760496989983245,0.5745283641576728,0.6147566072291382,
 0.046325249351477416,0.9375338661877661,0.6619342685797704,
 0.5807647362856156,0.5464503519773866,0.19679482475523957,0.8560900884413836,
 0.22736778135305058,0.5484369647350128,0.08457209320429926,
 0.8123737877968897,0.601994630411301,0.5308708535135037,0.9403824818169929,
 0.27236597917542826,0.8029542080578835,0.41874332017110794,
 0.05319881585516617,0.5502068750714809,0.3039069097824051,0.4479940654510437,
 0.23374291826178673,0.32530046805018153,0.2455353363683873,
 0.19429407268037036,0.8030655804606175,0.15120330312402708,
 0.7376326591302503,0.7380142801151797,0.5610221166971808,0.8509113501392931,
 0.5790848768373363,0.1055274473358786,0.7974814483515147,0.42201131894665644,
 0.27867076628036946,0.45284672255765357,0.06606828064699,0.8292493761690343,
 0.9280131755345475,0.03557479431836408,0.9941925202381359,0.6612768881271801,
 0.19845193992430177,0.8083243518215141,0.07302120724054462,
 0.6955419940600106,0.8942446581063903,0.37068877281115764,0.658593348482561,
 0.9722092947403518,0.8783924186206947,0.3810632645426554,0.7681269421317161,
 0.9565462727036226,0.9837302371307836,0.27315223921414944,
 0.18264574476647732,0.14450644960966952,0.43987596268221607,
 0.03469112425214271,0.3874488092180204,0.3168668785886878,0.7047743625517016,
 0.3202262293320366,0.7431932845425129,0.4205273052138455,0.55296120275843,
 0.617717151246815,0.9690923046085246,0.20347216871851337,0.03225890601006842,
 0.6548796318123868,0.846762131339243,0.61237057022787,0.5363299435744018,
 0.21376224351977524,0.8425602517290541,0.09107838135025337,
 0.6902419505171609,0.32939726328312036,0.9421013877722855,0.9607484759823028,
 0.08643244152532314,0.6301651113045041,0.40812950122182645,
 0.26272651345966824,0.8840493967696105,0.6227718850144875,0.5118933132628453,
 0.07964847423741217,0.1975644113620928,0.031647186418827356,
 0.8119338281656304,0.4988149245825182,0.2766863346211065,0.13973460957792327,
 0.8522197526355045,0.8234583022640835,0.9416170613533315,
 0.030266842431577956,0.16370996365276946,0.33753940677008576,
 0.9020506570239172,0.7717735706387127,0.7116704793738983,0.640914726316974,
 0.9105190913664035,0.14906447952438817,0.39284407924847464,
 0.6827970345364186,0.5832910499230599,0.13488300849328672,0.5311994007112603,
 0.947749449347012,0.06885544762181295,0.6864618388968111,0.10880496765997749,
 0.15880638353930376,0.13935008600919818,0.4287042119327442,
 0.9022442315504633,0.9095254952816912,0.004053708018700974,
 0.3744759257624898,0.005016623744885096,0.5754249963087784,
 0.43035267979949743,0.8684617459036149,0.2447201440822574,0.2540931273065089,
 0.7866439618017926,0.9486977558182232,0.822865978598657,0.10269702811896708,
 0.4964097569388868,0.8291297398060431,0.6620646576727542,0.9106451760220672,
 0.3355295319470599,0.0375252110163351,0.1424387330104293,
 0.011209034449811917,0.8724070166309938,0.8847672202541205,
 0.8042890569451017,0.34103941110179437,0.07247225288033776,
 0.14595635968309795,0.33405800502089367,0.3956723243797976,
 0.3157441760339511,0.3038578010209856,0.4387000003986826,0.07810872153452841,
 0.15732732381327585,0.3318142704596142,0.7298658250504917,0.6373140138529851,
 0.8676040712113866,0.3405997668673383,0.7552553764906684,0.08807958023445461,
 0.4195124712011441,0.31676573711821754,0.5797594360115484,0.8331120769518123,
 0.7538926342420453,0.6745296991224219,0.27670598423890236,
 0.11512192695250534,0.9354911754630741,0.45932967742879616,
 0.5494804200076081,0.5100627609896642,0.015366779654850893,
 0.0816328729640331,0.39077079118986613,0.6600230769735492,
 0.02319407975169918,0.9052874981244377,0.6492954765272833,0.1559798287615345,
 0.8589600479025373,0.6213023402376645,0.592195667304823,0.3949093912929996,
 0.5933332208047393,0.1309695010432107,0.49001730861939685,0.9662087298103936,
 0.2534261109900646,0.026665604827827982,0.4465285403432401,
 0.17534162642581153,0.8775717552679348,0.3827456681279471,
 0.49865328666815123,0.14773897165361338,0.054780446532850746,
 0.8963829945058694,0.1607477873236166,0.04878532615695219,0.3791620417101492,
 0.9663765425794755,0.7656917274558944,0.21670325554265846,0.5383301088551198,
 0.7927960495187737,0.3265784078777474,0.4188660052966727,0.20439034084539442,
 0.17347323152933392,0.7085840719690347,0.06668941025722808,
 0.10209684286605725,0.6766551536239951,0.8047063970689502,0.4730280212076008,
 0.4566145745506651,0.8565823542139898,0.10887584447502652,
 0.03298572166129221,0.041387513426346834,0.7981527816741665,
 0.31968608106870533,0.27963463069097994,0.4543850598036978,
 0.7637997791545358,0.7021449838631495,0.8833572298765081,0.15289705484658345,
 0.9210902876886529,0.46290869391178435,0.3151923901731888,0.3880734840271225,
 0.14342338093781304,0.5140640518227778,0.6153695270273988,
 0.08434101890552792,0.8663332641977632,0.9835961172146538,
 0.14169923245898142,0.9009671698435497,0.3848778401724685,
 0.034069348820771395,0.37167121173435813,0.17135436854395558,
 0.5166639718860406,0.044549958510039955,0.14778082178607976,
 0.766266095805422,0.3032970015978863,0.43799302355878056,0.6226333403144566,
 0.46925491578896206,0.9533823516267588,0.15654191054009892,
 0.7627142210300187,0.2616689720313148,0.7867610023870708,0.22338079158558577,
 0.49606850118071777,0.4839131665748504,0.5115066258622857,0.655184680447118,
 0.49735392367794695,0.391525699496142,0.1640586123645369,0.6796469049375078,
 0.8268913997482117,0.540395980589268,0.1130835851984211,0.33954307186072963,
 0.6837911105250492,0.6503360285649732,0.14181210276758383,0.8341347026762862,
 0.09327593413479762,0.12811810528677514,0.18697858355885189,
 0.4109972395016742,0.7945894661485615,0.642254844319832,0.8522140125953159,
 0.21681357888848973,0.4774484733341948,0.5607625959891772,0.4572539474791102,
 0.9899267916212006,0.30221099823116515,0.17540751889341044,
 0.08715572374724867,0.5322422508117888,0.054761229093284625,
 0.6086326145927533,0.13266608744821196,0.6644404023389876,0.5816766975180327,
 0.8931059859557965,0.21914012893879298,0.317072780311774,
 0.040635183739322844,0.9872818786649117,0.887112627008955,
 0.025902585162853264,0.5814809783590457,0.5521881391102561,
 0.3342743483621182,0.41302475779325176,0.15262661750272521,
 0.7656931854451542,0.38110150681447785,0.34833310999301503,
 0.9936429625363749,0.5793181860634781,0.40087714079014314,0.5809263659601837,
 0.9583719878406907,0.4933769545655442,0.11370654990192675,0.8183968175316575,
 0.7241021632933643,0.452917204321018,0.04362892133968277,0.5634269665238607,
 0.36004931208603685,0.8117960401877478,0.9294117011803827,0.9703058164558903,
 0.8829673372490734,0.7675932916194369,0.060520810184191864,
 0.2347383385294024,0.900314030946051,0.657923469212773,0.6856624575788884,
 0.6062059984983934,0.08948165390276297,0.04239984197388047,
 0.015315884762313603,0.12549105339310995,0.9626047336293034,
 0.3466832839325299,0.4869060806648401,0.3031439436683685,0.9351174999062808,
 0.7926466954172017,0.42996449109225143,0.3523951278994354,
 0.20370200772319935,0.7576356833217648,0.7395701985120008,0.8494437086895065,
 0.8067995830839005,0.7414747024221914,0.43663145021037386,
 0.42030796085321465,0.9139893006217992,0.19314371243068995,
 0.7245282058681184,0.6231149446597216,0.8807223232996269,0.09472565354235352,
 0.5383771478486575,0.9861336069491066,0.4930951552095305,0.4491188838320511,
 0.35447876803320266,0.1368032626461766,0.4019855567009807,0.8335495681292845,
 0.29739441728277505,0.23128525637931596,0.8525608778483987,
 0.3188597332740565,0.46908419536477,0.49014118066682344,0.214990763529882,
 0.9266163001346617,0.2149732413035954,0.014944191550560149,
 0.34666925868079523,0.5270611844134743,0.19775957194504867,
 0.4375172270469069,0.19377603149709,0.1624471919741075,0.8180918492555769,
 0.6445625404662995,0.3789236771463804,0.15136160787069564,0.5722255478199156,
 0.7333049516740278,0.7203387724783987,0.1116131890341745,0.8264585930599831,
 0.549997132266797,0.600735854975112,0.5133850216305379,0.6924800775629554,
 0.14864537873108274,0.7372927294392312,0.40850975255608246,
 0.26465323053056555,0.5499584346711596,0.5140925678918864,
 0.11374013307723319,0.6232556999205827,0.8503397389703561,0.4684714755826871,
 0.7886153703903986,0.686262601627073,0.22187757054034563,0.42473729296716944,
 0.8957452658888303,0.7603339456987809,0.7112606895548916,0.9858983060325168,
 0.815849820694103,0.08561598017828553,0.27421242807827184,0.7846062115868797,
 0.4174974325591341,0.18427702822080372,0.9764453069542194,0.5326602146983042,
 0.43348487242768374,0.3628881468015428,0.3463255946078563,0.5797678217418216,
 0.8443651321722746,0.4490168818654955,0.5129708843784435,0.26882146577004706,
 0.04228588951842582,0.47404592458824224,0.6672176843839628,
 0.8477353527662366,0.21377801412142394,0.8489673112841112,0.8357525100689609,
 0.6228432787989501,0.6786837680536802,0.30772507149391604,
 0.08286281042459975,0.5460304062316661,0.9732100847442667,
 0.025012360307181347,0.0975591403141679,0.9455423446725899,
 0.5747793196064224,0.6606480482159027,0.060165591042212085,
 0.18487029117141096,0.5332219490067811,0.24181559777178796,
 0.22721100660623672,0.5094066566953862,0.9069894735652899,
 0.30663742929476956,0.8793267672730801,0.688534197144171,0.5289714466404526,
 0.7539262040231652,0.027750511550426515,0.8857272233179524,
 0.4460891496727827,0.0838410362098223,0.5573048590933218,0.8430528010171431,
 0.5683440012608855,0.13020745674568424,0.4661830051729621,0.4994099095727499,
 0.8100829741646689,0.042307902401469866,0.9791646841552194,
 0.6509818193208764,0.22427574828893626,0.18212656468804245,
 0.8344152599642126,0.3313694023092425,0.029356612669200866,
 0.14850812698986093,0.657288404413905,0.08172579840419048,0.2624508893975487,
 0.11529001645678216,0.7695029974609809,0.17407146992783284,
 0.8712537718526298,0.4026724590467513,0.767577704183169,0.9752901265823779,
 0.2833127325454108,0.8900558483607824,0.17763887837282288,0.2556607464513838,
 0.2173356650896373,0.5697546008102042,0.6351318425893675,0.8294231779707116,
 0.6596990668794921,0.8412744185416967,0.05134309428058037,
 0.22456942470691144,0.839857657194104,0.9036163532559662,0.9595602486513478,
 0.83685258072493,0.020073385599611027,0.9367233798824351,0.7768050692728352,
 0.7323617775461778,0.6895124222165483,0.09624827475084519,0.4419016087608618,
 0.053142817889138105,0.651211793436758,0.6120151870764324,0.4745309975398325,
 0.5572384183217858,0.050930317314325735,0.7217686802741466,
 0.8552716753638584,0.0624808618926993,0.2933850192285571,0.2706604376993844,
 0.08163073699379852,0.7016699935658204,0.44522159252109683,
 0.8976253357380445,0.11486115919929585,0.4235081057145136,
 0.49777887087007067,0.6958470731209386,0.7894146725152515,0.3673324635091758,
 0.6378688391231915,0.25024204193553135,0.8914171231300069,0.6746278825645233,
 0.37212265124597277,0.441928996788182,0.9310333198076984,0.15295714913669234,
 0.4881458225195514,0.14508520598676278,0.6250005299254935,0.9933186881835464,
 0.15813566808037693,0.9748631591082713,0.8769648737746836,
 0.13654330590135266,0.8535820466991952,0.9960273216712048,0.6008409783883121,
 0.8888362593191198,0.8637826319715596,0.4800295178354078,0.9657538863677114,
 0.9401015511445525,0.9977836692606447,0.48896979718386724,0.944597572437299,
 0.4018064499564685,0.5760099235461694,0.2335640524057898,0.844881071575296,
 0.3436658223407014,0.426809166048541,0.5073748608984868,0.41078248407418305,
 0.17058646173229375,0.06665510390690077,0.7546015023422208,
 0.3866111238994824,0.7558461685152704,0.7658058879630443,0.32793791235129044,
 0.494703613833705,0.6907130122827224,0.5769101753205635,0.6092104348232158,
 0.12301670304363199,0.7268263479577004,0.20192437118023032,
 0.32512033060880496,0.7588861370485538,0.6853000010965072,
 0.15679867874997833,0.1579156557793897,0.3199354128337808,
 0.30170009694799693,0.22251101168163745,0.948949267490031,
 0.007440011151689574,0.12910332231434452,0.25989284604955554,
 0.3495625185458838,0.9705522564429886,0.5452020331756228,0.22388638749241163,
 0.8441684990729059,0.3441654764591475,0.4617382491158797,0.2950018317028169,
 0.6794591591984466,0.40067476908819355,0.15742476983627207,
 0.7146651009631259,0.1416118116012678,0.17044838162284792,0.6199290391130039,
 0.29278794691161647,0.3037999020105153,0.36305473764912577,
 0.7631493210833186,0.5799916514677327,0.9383490496025458,0.5710130827016484,
 0.27762814157352134,0.43575167524949165,0.7649299340243501,
 0.8810485195580524,0.7343877516060914,0.24989489083542538,0.8983282206030911,
 0.24523792139299572,0.9907844110101459,0.6245029068009552,0.2547476971692696,
 0.252188500779171,0.3660693737555393,0.5583742918601039,0.5349643594012263,
 0.5531154587149484,0.7200579724764714,0.44850552890490636,
 0.0065353804527068515,0.5084844473126447,0.05461386663308476,
 0.808437918431288,0.19645067604040423,0.37259867003405234,0.9610998399220241,
 0.7161328158510419,0.6006009785630999,0.8987710994159486,0.07856446069137446,
 0.4710727310590255,0.5075190467840702,0.02272313537680981,0.9335154505372143,
 0.8760446136296646,0.5654297025220313,0.3697067934489987,0.07182510643036788,
 0.3229127496818819,0.2434241671499504,0.6568175238379634,0.4762240217028092,
 0.5983492801934339,0.7686291571525509,0.8105188687561352,0.05579584123848347,
 0.28091975727702123,0.1350755311945625,0.9922147752186893,
 0.10316290639342618,0.5723130989822321,0.8156322595924204,0.7591392598407575,
 0.39204271980153327,0.15290494175654057,0.2493616620178739,
 0.4311263508799502,0.7423470757932868,0.38107952474743756,0.261038531547395,
 0.20435563797143474,0.3449258901538079,0.3390092500535976,0.5329636159351797,
 0.9646411788117983,0.12083158167659525,0.6008328014848973,0.5039423156023929,
 0.328583358411799,0.41714182355530205,0.23499236950665991,0.0637390729146291,
 0.17829720794530757,0.9761956061318973,0.5917270789915576,0.4253724755562527,
 0.2857353058931492,0.4859956482664368,0.26917405739114764,0.5460536437653124,
 0.43652908109499144,0.3233053818602323,0.11882094089748285,
 0.4934166652077483,0.8682532952120032,0.5098672334811204,0.5853865563670335,
 0.8255112536598297,0.06136718553842335,0.705924803112827,0.6413734146922087,
 0.5268698494328776,0.7349100265349389,0.5791113814487914,0.9756150230085674,
 0.4602086529101549,0.20079211850169199,0.03752780141330947,
 0.7047300991172712,0.17017497229738154,0.5868748958278671,0.5178271415196637,
 0.36371894474117805,0.7580988084638773,0.9908876053661041,0.6282819918312428,
 0.8902312867926374,0.8565763226188108,0.828800887500904,0.4546810348516612,
 0.6244607665434406,0.6487434497512724,0.6424130764689617,0.30169415831386726,
 0.41234821604852046,0.8858735921725185,0.8348527257365814,
 0.050749491554567316,0.5147564903890434,0.9540394188705689,
 0.7124168102775927,0.9020556085441239,0.8444884634155193,0.9752627473639972,
 0.17024956238716138,0.9352603547388642,0.7954523074354363,0.4944369354313579,
 0.2992050656052694,0.12519771530504653,0.48447290853406877,
 0.9509014523193818,0.5138895542288724,0.12040085224054642,0.7955895481364963,
 0.2945487985207128,0.48575374123786663,0.9806543322732919,0.7264992516169753,
 0.6841680700364433,0.4240358650103814,0.25435711453889764,
 0.22044137209171233,0.4390115891969446,0.45531433363920215,
 0.42075410679917247,0.3192138319544785,0.37464909419254855,
 0.030400100392557694,0.9795094038495132,0.3949064134845367,
 0.9383968460223803,0.5364146807535162,0.6072841633342118,0.3833847852160195,
 0.24424990049323758,0.3844031110093862,0.4089758455297736,
 0.14783692893264033,0.9745411644719506,0.1428171847630706,0.75262718404225,
 0.8126136922363911,0.439688856712002,0.05810527292841117,0.5723831502970398,
 0.7013603368681305,0.15583942323864186,0.9940382601048919,0.7282495439087693,
 0.01238563964780548,0.39836657547604115,0.3546525296178763,
 0.5466952306242032,0.9595913438846143,0.9823499777131849,
 0.041884141874868885,0.9903906395495085,0.8029890771179582,
 0.8841787638887901,0.23178888761065397,0.7084593054024512,0.2915923517942398,
 0.4976634430457819,0.6986850520301531,0.010555827858343525,0.721402021800185,
 0.17643936539702276,0.9184537427626606,0.6225343883322817,
 0.43346922474193306,0.17694813602662973,0.7029579587519391,
 0.45666838844648605,0.05101666626531842,0.7642411634051508,
 0.6477877812056326,0.9189746886911918,0.919437030587921,0.6838810759930496,
 0.6128287383632256,0.9468356929192603,0.8370060964201107,0.3229038215400283,
 0.7808265975753292,0.19093295645175912,0.976272811884375,0.04666822774072599,
 0.9904267868539489,0.6986177846321961,0.3417151814263516,0.5257389544096783,
 0.6305156849564321,0.5031995715317095,0.8922472431868913,0.29184827833506255,
 0.35699931678242136,0.03146772944162202,0.8916517101930268,
 0.8647134621258653,0.047480524289432724,0.6054345174855394,
 0.5246085956717906,0.568408294737275,0.935926752821977,0.12112556068110669,
 0.3481461686751186,0.534977275719151,0.22951483800929917,0.8904818662312037,
 0.9275295979502985,0.16217162910320826,0.3611966136038862,0.6305787166741766,
 0.2437554241113944,0.39872810776206624,0.034318660477399066,
 0.8481546646595276,0.4481400487623878,0.27804129569033864,
 0.45980987865493117,0.6112060462325398,0.5026591519730911,0.7141815103331766,
 0.5346261250016688,0.9572352990219735,0.1571495062838001,0.09285312454770578,
 0.3084539255338957,0.09024239926682776,0.3629580576876177,0.1973294038457145,
 0.3116566125999469,0.8256355218421827,0.8702141074238308,0.7795430922301843,
 0.44719184888424324,0.027110344445211765,0.40752114719736166,
 0.2907773598742769,0.17523796775888723,0.47371984605406137,
 0.47210243078635006,0.7347992182421629,0.673619941336211,0.31303836894900905,
 0.47377329391493883,0.8735714220852858,0.8907158876683163,0.35427258930925,
 0.06645599853804152,0.546411539460704,0.22952856762247043,0.7435625854819631,
 0.4863283309715275,0.004389439124280914,0.4884796593362958,
 0.4364173002408922,0.029984657018882532,0.3412993720032057,
 0.06554513265226947,0.9229937963635164,0.837797318926694,0.3695715108261092,
 0.2481803087715273,0.04032351928017064,0.264433577723139,0.9796630343499713,
 0.4151659112836522,0.31039586547526354,0.9360851640544957,
 0.05167180638096169,0.9802260091032357,0.6715385405351595,
 0.010872649234805687,0.4381647101276185,0.6634957184983941,
 0.1376419211455443,0.1774335425709348,0.46349147892745535,
 0.45309292133433976,0.24244194667010277,0.7998935164841781,
 0.41329183499537836,0.17239804302080852,0.05602335150077897,
 0.29469443047068655,0.3965587569161093,0.2385220240296978,
 0.49511866063204635,0.8673997199315874,0.9640181470688698,0.8298781208028452,
 0.4304808504096258,0.48302031943075774,0.0651808714815092,
 0.31236688246556743,0.6188710081524694,0.053744649135313405,
 0.20540983667406176,0.13250159858963007,0.1355549971089378,
 0.08757208406252415,0.4272956754101468,0.11397006217650918,
 0.5161033805192528,0.3199112192068342,0.8165196778778537,0.9917121691836733,
 0.607299258538487,0.8026855348974605,0.36718474680761537,0.7088057446226861,
 0.351440332129592,0.5130641734442097,0.9572877955549196,0.7284852926661922,
 0.11219594645744058,0.28988739850548795,0.13421712055073065,
 0.9148600440460954,0.02983450517144659,0.7244989455728197,
 0.11467519317620645,0.17661704875221917,0.9086930730697125,
 0.21150902799746274,0.07105290301150868,0.19681705501131308,
 0.5913098260169908,0.806461878121326,0.8201466804387714,0.6043034435816202,
 0.07470510934322216,0.030653734818114287,0.5298653645202458,
 0.8549324275653563,0.5820095884088845,0.1513945825029429,0.8740612260854357,
 0.3583144039193268,0.03064331283020405,0.05459290187543042,
 0.02472664182836548,0.2464849707100657,0.9324836230438649,
 0.25488364759104876,0.6690014137892364,0.00814116546443866,
 0.6896469671198395,0.6245353293893889,0.19334082852942913,
 0.05312424478414657,0.11034503761084691,0.9711966848150102,
 0.43306110943595966,0.7864545574251496,0.6445042151216657,
 0.47691653139484547,0.9721788219225234,0.0019076349367984635,
 0.9874138207304812,0.03494541623607439,0.2893506768809466,
 0.06629749713282351,0.41932590525738367,0.08617765385856435,
 0.7283747878003209,0.5450576369864284,0.23508017372843915,0.7416210825561592,
 0.6482485678957461,0.9356725403587834,0.5051642971303759,0.43484568866883655,
 0.4199043406972299,0.08408445479097881,0.2925566134856912,0.4716555595927576,
 0.5251517331052443,0.8889094105458448,0.9569770321730798,0.46271535025307187,
 0.36173728915027925,0.9105898660446234,0.30619844462447776,
 0.8909259670679582,0.39415445279153294,0.5324207856821559,0.9212999481976413,
 0.5494869179436931,0.7324412075315272,0.9349458045559412,0.6270406804188128,
 0.8586929554138932,0.8132194718072802,0.5050973504383727,0.06558506802836261,
 0.4909936597583503,0.624039166140812,0.07613127470789938,0.8690197961460691,
 0.061464699724245175,0.19567190662036937,0.4890969201500298,
 0.6493472700197059,0.47004308312909937,0.049413309016409834,
 0.2243883732648959,0.03902553463693559,0.7556540831930825,
 0.18750876339098277,0.776580888962526,0.9556598621031102,0.1847956406256165,
 0.8817046029421488,0.18655354656171874,0.3166823573778178,0.8122044247021979,
 0.6682805693231579,0.06171024436039918,0.014787489874662496,
 0.253888887643941,0.775318368209169,0.6277143511436127,0.5894546293477879,
 0.48942659696288526,0.807652385495672,0.7311469395374581,0.27342755112862904,
 0.36799930057232866,0.6224489338437748,0.6428832706218134,0.7016177671091451,
 0.7079219023581057,0.048012155669451384,0.4465554053520513,
 0.10980740494040964,0.18414443428613114,0.42366458509348304,
 0.06751461321778707,0.08051854080316256,0.5357886039643426,
 0.25213258368175695,0.7565471300227972,0.3750827487230989,
 0.17281552880977902,0.019166107420433676,0.24790232620959152,
 0.6307966185107102,0.9686095923650961,0.9922952934182011,0.30747803863614664,
 0.11554012982127515,0.31360499158321586,0.6438301950066296,
 0.8220856893357047,0.7211942482057938,0.8792051249884059,0.04601252155903812,
 0.5257073160504244,0.28522241003120663,0.2136556149082175,0.7824054647021436,
 0.379008775943418,0.9378905879852386,0.5469717221645635,0.7720611626691957,
 0.1658891344574932,0.7138178053646077,0.7806097104822474,0.7315183324269365,
 0.448159520177808,0.5382867898444799,0.4790323195435162,0.7177511522539405,
 0.37489946859015677,0.6684979830731783,0.3841958431644533,0.7352111770965758,
 0.7766844612266594,0.4360324091943346,0.49077812844783164,0.5218964014683193,
 0.2902261308551781,0.9083070732528071,0.2919003355776516,0.5713302693803393,
 0.1887619337515556,0.21025020867338384,0.7716628816842238,0.6982876459244169,
 0.3526976110725568,0.46341819512772475,0.9477626419987456,0.672562261122533,
 0.9002849445046947,0.36704570481131205,0.586022657707939,0.0219424690135136,
 0.06826374664178814,0.019438103042007437,0.02919880660740304,
 0.9734391678128392,0.988462777185708,0.3200879566419794,0.5061344108450312,
 0.47085779253071336,0.9566726912195939,0.039308917560369916,
 0.9496334716014184,0.7349460146228095,0.32389030700865973,
 0.07652411695427441,0.7516293181596196,0.8212651225659896,
 0.37907245573090254,0.6702053265189228,0.13213013310014154,
 0.3957815936620136,0.1829737757004073,0.037516317443829106,
 0.7942404972458154,0.5273324344620716,0.5119701906688948,0.09056560287972404,
 0.5179866013594246,0.8401639713358324,0.14473264832141686,0.3879298741223445,
 0.9482158245279808,0.271240991275702,0.853344267630766,0.7920921320278749,
 0.07377465216077073,0.4207640542564345,0.45379616617722507,
 0.29735688210549405,0.6517923396151827,0.15222034665386386,
 0.7070091317084585,0.8736535032472628,0.8700577387514369,0.9275930852069129,
 0.1742326212830887,0.566735913925166,0.4869680011283052,0.8467833130961855,
 0.9989943182700187,0.04746445600431759,0.600994600472894,0.8686619069575189,
 0.11306957394895933,0.6825959505783531,0.44927163189531005,
 0.37390121497459416,0.5725962069101747,0.6508816243277242,0.9461010744556029,
 0.40339284349016846,0.8166420113310366,0.2388140396968842,0.3783947142647599,
 0.5203140901164092,0.8933188449457773,0.9086295699314172,0.09786179018275676,
 0.9535265985735614,0.09823903578055682,0.15085261624332458,
 0.9154350163534678,0.6110064699076337,0.278318427866434,0.2074335678206598,
 0.185633386040699,0.05038570268809539,0.8885245664974128,0.42339249107225285,
 0.5687775609998896,0.4401643562295616,0.6654628842846831,
 0.019545040102062372,0.29994067655413137,0.9902096913227407,
 0.6232714738715008,0.458194493535351,0.06456851928967255,0.5401706072486522,
 0.48008472599321617,0.22497970723448568,0.9696870491079542,
 0.13432937406159873,0.8005798109582428,0.6728336217256696,0.6333220727545823,
 0.8410940325069152,0.6679440928662057,0.1095952959626535,0.6595764537189751,
 0.4127793329214279,0.34608485333907424,0.9828079912731901,0.5250432942437342,
 0.3769927079102199,0.9581133616129076,0.9362306159835734,0.3300090034664118,
 0.1801193099079712,0.3813864061326957,0.3433121836410463,0.2736227519012091,
 0.9490892163598613,0.7156783683396549,0.8877895608520738,0.7485211422649174,
 0.787102214975779,0.044048081145529716,0.4162724342948807,
 0.16038870512563674,0.1915583340306477,0.16175433285111795,
 0.8037340320010924,0.05135692773549527,0.8810216498730453,
 0.42774564092744527,0.09444933241482778,0.9936321295183835,
 0.10484855316763508,0.6516108972515898,0.7110624959147669,
 0.33624212113955265,0.2746014824221741,0.34053629299398236,
 0.3448454353649917,0.20242314078779755,0.47228790974824764,
 0.9533314908742708,0.24625075435397403,0.20859869660746388,
 0.3944687141865456,0.8680304872933549,0.5016657828973965,0.23493215561980874,
 0.8150038458107478,0.06530549243288153,0.9876361444518041,0.7085751070799148,
 0.8629507637633687,0.4002103741632015,0.26804002214930034,0.4935700199539452,
 0.8447362864766742,0.8525877223640352,0.8769939451594182,0.3368781861451704,
 0.7435916058272193,0.7583844426293009,0.4090763157607391,0.13999327541076279,
 0.5626377655445605,0.9978607969988841,0.6777774220720529,0.5150752149119254,
 0.11060868921965228,0.7256638868706113,0.30050051067814865,
 0.33313461686065393,0.2356185223316306,0.43457804750724416,
 0.9787084373811158,0.567778650479014,0.16443035261159367,0.6682632827739228,
 0.13271747003574585,0.2785544774175138,0.5552062921169502,0.8520635876211982,
 0.656716157444837,0.40489897898510485,0.41992627316809517,0.3049221265977968,
 0.3246320795289722,0.2027230019723757,0.12659385891093133,0.9939726823880172,
 0.3522109388238438,0.48166859098885473,0.2648383313329903,0.7990774001376176,
 0.6583698878363546,0.280731393702783,0.5328135646480127,0.33043458429322126,
 0.39013897397129804,0.7735221653729172,0.7317527186216221,0.6120302689731086,
 0.813397248151669,0.051135458798843514,0.824993659607762,0.8517904562303592,
 0.8315226871505595,0.25906515317003054,0.8686457561482104,0.6493510380661415,
 0.5644914354801415,0.7670874731707521,0.9232673977740218,0.5415597345310927,
 0.4841641056949113,0.5785068660484314,0.5089787817189025,0.12538318536105975,
 0.5856450862012568,0.526875429552638,0.8686911924330312,0.8129237339324441,
 0.9538543805891373,0.2683142961140722,0.8630407803598948,0.8214795948348732,
 0.24107460636986788,0.08040158285209209,0.19360654573908487,
 0.5515028387031433,0.8845467336932636,0.939658060495114,0.522260620121242,
 0.8421085092020125,0.5071972810536604,0.8679538563939238,0.3957700770428836,
 0.8270601596276039,0.27032523927268437,0.6393452031763819,0.7770585354673536,
 0.6741865830911301,0.08735381625629057,0.508756577366178,0.42362163744863046,
 0.5114732825030044,0.8185612423917719,0.6045564719810633,0.3827423229748339,
 0.6696305854736943,0.9303142647339862,0.5934583494637808,0.1040696605258713,
 0.47446805467265685,0.8147986891118697,0.5248959291176525,0.8834312366378945,
 0.7999896597557768,0.5053669199731586,0.9039730615841992,0.4508674430106453,
 0.6075432920548444,0.6930872454304949,0.37657803191251593,0.6776173188035599,
 0.07343600098574832,0.3244653448672091,0.06430261315965113,
 0.30811518781723124,0.819959053494073,0.368390793091497,0.9597110358239678,
 0.8814477640688686,0.04357784839585355,0.3186045589459725,
 0.25689306857847183,0.8252913757325142,0.9489726824899833,
 0.49533141556497995,0.5117478481084125,0.18957134630399786,
 0.6919316497251436,0.9060251708649605,0.14662077825980013,
 0.24229777352954152,0.09408469140659548,0.15380113576168464,
 0.6789003969388521,0.003068157707814323,0.6164895342774227,0.605702725684154,
 0.03199738966748178,0.12452671035083585,0.6740365353876004,
 0.26118896179439177,0.6221570916805459,0.27166433740803486,
 0.6317973292222516,0.820106646640816,0.9945777995884781,0.23925460032390367,
 0.21201645938529978,0.3544825623358261,0.20285034780332034,
 0.16002464839943764,0.8878247512290235,0.20081771850900665,
 0.27587644217660434,0.4624574345054179,0.8299637476492343,0.9719407614507172,
 0.15993803002698892,0.22027547807213554,0.5688281306460793,
 0.9069559089713393,0.7956257055746166,0.3637734316935983,0.6267915709630926,
 0.6247898089652483,0.33380894810552986,0.014999246173825531,
 0.5008707204497543,0.16473256622205146,0.4367252401497055,
 0.19133863216718405,0.832460616928583,0.11415097586696876,0.9001934890340311,
 0.7208501446069818,0.9934757339613478,0.07977847339413835,0.5269046361733675,
 0.4163640225281696,0.12895166634953847,0.19765451202651496,
 0.5447633061551667,0.286513526353546,0.5909809385256282,0.33296449770329273,
 0.2866614058310675,0.3671319321958564,0.0782171613910696,0.8047554347353524,
 0.4026389618129578,0.9372634331762101,0.317113670369635,0.5424642513965017,
 0.9995200900443061,0.6384426574639721,0.43161817369320443,0.4244894573123831,
 0.6281540920799018,0.7520004945818065,0.512760859692636,0.5303248416999204,
 0.8297575612033733,0.9778307309025092,0.01811597861695302,
 0.21086778635033487,0.41675588633878613,0.19932142292555233,
 0.45948692562947635,0.6075187827843448,0.3930947942195151,0.5017270475031218,
 0.7610180968273523,0.42439751840136475,0.7626424659925095,
 0.12329570757111119,0.1562271984012037,0.013403443381887148,
 0.2696046496779938,0.9326108982731219,0.45004321039850015,0.7284055063935614,
 0.03981533100643686,0.853785822461274,0.32976161324039044,0.7651130211381462,
 0.3871823496498281,0.7141946956630962,0.8005676212573567,0.978145660426904,
 0.5176237287801392,0.9235981049720582,0.11047032896137343,0.7733111957594249,
 0.03715296203212648,0.10424371785051056,0.40288007346653565,
 0.12378000609479456,0.6736324173102741,0.9015902257152931,0.8049455961416748,
 0.966977970243631,0.41679197174709137,0.21386120740620396,0.9153786022942332,
 0.7837058587043547,0.13843727005282525,0.17334800701583997,
 0.4460115750196789,0.4734555132865059,0.38358800024023565,
 0.45754978466074436,0.6867671143429029,0.20815950505414305,
 0.9184716805681701,0.292765492774697,0.4854442238426724,0.5978891876477759,
 0.886370582227388,0.8307587030715876,0.9526476616197257,0.10875350942115203,
 0.036253129781628424,0.47620381516945776,0.08752926404046535,
 0.7324102100553145,0.5166217384943035,0.6778293760862424,0.718779682502995,
 0.6752675776126376,0.7671827150707318,0.71794083097493,0.9552677524926939,
 0.40895185762027775,0.38297712388998717,0.2643655941536476,
 0.9227226551081124,0.7479967178592298,0.640426032067773,0.9872264925756622,
 0.7583918649162026,0.7236899947132195,0.8516325546929183,0.885067386199426,
 0.6373152332148272,0.5759765586291478,0.39920584106691304,0.5019178624516455,
 0.11246314767549137,0.8786113502133984,0.9314734997786038,0.9463406901601188,
 0.12313096866700712,0.6668250270400512,0.9608699848331833,0.9613477124611943,
 0.6947551220698,0.12254945602518219,0.6695687573003315,0.027494733529100368,
 0.7978589558312719,0.6586298548596822,0.9522195271298178,0.8023444842911219,
 0.33041103182197207,0.9572952751175592,0.9507969243752338,0.5288656729434973,
 0.5128819213919239,0.862446941497042,0.7574477510979697,0.2905515835176815,
 0.2195389629501785,0.781060592186451,0.39754068931783215,0.5550494610986161,
 0.7387348211237507,0.6947057412760957,0.1018746734980337,0.4110359506824113,
 0.5960875924249098,0.7938444874033224,0.2575021915408495,0.7502696996062699,
 0.9899320417347952,0.52099801619038,0.40128863565248263,0.40288121210784467,
 0.3745404280500776,0.33929778804972877,0.7908261708991298,0.8041899363372762,
 0.273612130252052,0.41208312686781523,0.08598472729930462,
 0.41430881116155527,0.4587647218840274,0.74030886190154,0.04465838136621736,
 0.3131269554032603,0.9375067008130794,0.6427967675791819,0.8236195336061047,
 0.30096114091337156,0.6576971618640821,0.06076081193236704,
 0.46951719099148725,0.8999144548463007,0.7523968509305695,0.2089486750903391,
 0.07034563560395157,0.5825160730635386,0.4716616644758521,0.992275803839182,
 0.8829893589087144,0.4069107091818359,0.3666009927812909,0.659129173834327,
 0.7844561152445255,0.7667131710616835,0.9810934495073047,0.05294564685870917,
 0.6181609593104382,0.8570676458982315,0.2348725623367809,0.38434365168457696,
 0.31764596126735944,0.4187307217389926,0.8511760641490147,0.8700818371598579,
 0.7726277240271991,0.4659644506114208,0.19205081254767242,0.9624630638688282,
 0.9122191322183293,0.3921739626659434,0.057705314049091516,
 0.7700815901900527,0.2553691832009436,0.43720008389279075,
 0.25485827595679467,0.27995015263268774,0.2581068183895451,
 0.3094413757280312,0.7558458195149517,0.9755299158816888,0.47169199123311956,
 0.9466432845722532,0.8958586030430316,0.6923406861148549,0.1038186993968715,
 0.09577496210740954,0.7104169797790842,0.18893834883978944,
 0.4303990012756054,0.03430924032919547,0.23923554231314448,
 0.3222270011136773,0.6831947247258938,0.002725574097998784,
 0.8656787352434594,0.6025627010209049,0.42995021334250194,0.3412506811866023,
 0.3539603522038439,0.2775847642576572,0.61961214028835,0.5345289179503797,
 0.8989255750769394,0.54340485836509,0.705647641996095,0.1169134537989005,
 0.38410032995887755,0.7850184945314145,0.8732838263479429,0.9314848620925175,
 0.852347154259534,0.5206718094475116,0.2293264495388262,0.4279321284757788,
 0.27720045523763515,0.4232830890592214,0.5581272841445936,0.6067403529756017,
 0.8908978392650119,0.9080507560587912,0.003990587012202673,
 0.5971693463271357,0.4083040355327413,0.6617574655547266,0.46602872931373507,
 0.4238874590750574,0.9120490001598262,0.625386468709517,0.9808822128671935,
 0.16034543146981384,0.30410298838215755,0.6040375501176091,
 0.14088683301530203,0.1984883473888832,0.5645469000650953,0.7543910304758299,
 0.21942369270294793,0.5030047698237275,0.13024800792549307,
 0.7070897222914553,0.05624716157304088,0.9483085432462899,0.8615191632757261,
 0.4769915077104151,0.9122997033732507,0.4881415998919806,0.7941428322988794,
 0.49421833070607857,0.8121106587761321,0.4448755231379383,
 0.13030452731661446,0.13537354063522888,0.04001267377314066,
 0.23276965757106116,0.24706722093041544,0.4895239845467394,
 0.7584588351439446,0.5739682999472526,0.9295614787724695,0.3193867344929111,
 0.1911027566211041,0.30649024336271213,0.7464168816874962,0.2850176402562088,
 0.5539110792553866,0.1619695351614353,0.03709673686859394,0.8271924473921136,
 0.19091502065913324,0.0336661866017528,0.35682146505639056,0.448620185195891,
 0.028622217166914266,0.5117043036159337,0.08244899110117765,
 0.6256491361344227,0.5268241455170655,0.9782637638130742,0.685513193253061,
 0.08480578316227283,0.5519636629989226,0.8512555666183886,0.3590465062248641,
 0.8312666236651356,0.4437290464987713,0.39720900198563336,
 0.47337506921820915,0.10197510020022138,0.9216319139949541,
 0.26282478961773215,0.33067916366137673,0.7201321665512078,
 0.2731337665422189,0.2275374461969073,0.20739295521811607,0.5933396483590471,
 0.03726003106294029,0.3872026954236758,0.27119156253165866,
 0.7270301123736198,0.7883207595770593,0.33222218574057827,0.3676186129034461,
 0.21016312369324253,0.01675546356650237,0.7398088284000367,
 0.0866340136761291,0.9837149688915513,0.6320191410013382,0.5222178553287999,
 0.7577749365913886,0.3436765228799795,0.3253043455231711,0.6397003208858261,
 0.7470004018525269,0.3843643118910629,0.6494511970747551,0.6605693291311983,
 0.08538427255019909,0.7194661316941053,0.08362806248337318,
 0.6154784889032495,0.5518317578598606,0.9289589902846924,0.561687574872919,
 0.5222394342364683,0.4155861859835859,0.9977158597656427,0.5524718992687857,
 0.8008175157778425,0.5413315157858399,0.31608356063973575,0.6979294190651935,
 0.4591181462428996,0.7134603879023862,0.32682161131091625,
 0.03333768181688945,0.48655147563215795,0.5020251611344041,
 0.9806345765530273,0.4292568291426546,0.49216116093627194,0.8304860547123344,
 0.01450635271626699,0.03589275498506839,0.1996463776859979,
 0.3703529146412785,0.4298122496793477,0.10585312656575663,0.5461860990999815,
 0.14949891207481592,0.39648489369368345,0.7948074002893133,
 0.7892271586644136,0.305729306300071,0.21869010536724287,0.2052409062109637,
 0.22875155327801033,0.6805540898293909,0.6892941542405793,0.9229999820508543,
 0.683320646053759,0.7134926792626164,0.8462580648218295,0.41340004063611235,
 0.7031657858582696,0.27488446607742567,0.6806945428617324,
 0.07923004880691153,0.18635996064926363,0.9962782418063935,
 0.1930777054182975,0.4936131943085885,0.2365449333284544,
 0.014636760162630269,0.2714023957946432,0.3179974749128449,
 0.7943926725664454,0.1424486576594195,0.7953318285700083,0.18961888333696753,
 0.19140203150415047,0.606169568435773,0.11403987390492887,0.2471177667837149,
 0.4510121511207338,0.6869468245072333,0.11018243482576995,0.1441349026484975,
 0.324520674646121,0.361700945625714,0.31245498533107885,0.7936628812010972,
 0.7463610987818099,0.6090798951725104,0.29930250549617643,
 0.48746964337317766,0.9682444105706794,0.11331402004809388,
 0.7048594514212863,0.8827724814870553,0.9100187311912493,0.23423007701899357,
 0.8142564836581255,0.8816082978897672,0.07102619700918744,
 0.24186460487398576,0.2877048414800648,0.18016694530374844,0.57731078439877,
 0.9658944016486827,0.6678445256568171,0.5733106223424902,0.589376118467391,
 0.6671573748402381,0.9150400421120378,0.2757409423764441,0.11523787347628867,
 0.8137030809636828,0.8011532559340893,0.9003728055145092,0.6480824185909895,
 0.6945792271519987,0.8366012696496719,0.3355201424786568,0.9382110742473988,
 0.18521777532164285,0.6738621902920914,0.11576213227443133,
 0.2794494160431893,0.37795078928710113,0.3142966096841989,0.7405790640664587,
 0.4365910925592823,0.8221057866101547,0.26750002282243224,0.1736129055974137,
 0.2677429671846717,0.7266519789789929,0.20849713215775878,0.5455418644624928,
 0.1946258462822087,0.32412895472871717,0.4213889826775924,0.557868986663915,
 0.6327696743754029,0.17564650351446875,0.024700093407879153,
 0.3041749594007028,0.7216660604425832,0.9375971761037826,
 0.012927447004484338,0.44062507551830954,0.5486607912712393,
 0.6364095738083542,0.8206432989264842,0.7741335373759338,0.9664755378768957,
 0.5291114682211949,0.7618215992409881,0.35218928475685496,0.7996415131459875,
 0.6887792863530985,0.36400634315467806,0.7351898174379577,0.3651950528239273,
 0.6644373885909136,0.5585905789190947,0.19569351665589585,
 0.07907324507776137,0.538242180414341,0.6679358126155767,0.556089851283073,
 0.5641356897794687,0.8579414850461718,0.17750327922962594,
 0.12509948092420853,0.6307986079854431,0.5185231343085821,0.8526548251318039,
 0.6350318089151489,0.085093992729809,0.6296525478962678,0.9278961763522497,
 0.6290164747185435,0.908067223836451,0.19680649189429655,0.7264878693002093,
 0.11672659779492944,0.37345610019395015,0.2805338529022201,
 0.2636056303728256,0.3201029921316789,0.7059948868695588,0.820569537079639,
 0.33830854093863694,0.13137431303577585,0.8944084379416981,
 0.03560489315908655,0.5559937102426354,0.8284840259535915,0.4503050475434559,
 0.3425619467860166,0.6003974627120694,0.889921493055857,0.5263067442582557,
 0.8653583332391995,0.69489436875732,0.07691610579805797,0.11192560248909489,
 0.21489671399489463,0.6471340125269484,0.22806994156242943,0.712934466232274,
 0.9018131364135472,0.17961503054334005,0.924427765913894,0.5351868395950488,
 0.7530166169400039,0.22709236059945104,0.3325033647972999,
 0.42250477718836255,0.061622880341124286,0.2908155076734634,
 0.3403656616649352,0.9305683529769023,0.3505879464376649,0.12137513659485566,
 0.7048003751246581,0.6524385264085169,0.4188337959741727,0.7254966989402393,
 0.9932022785056848,0.8827379888618978,0.08727020365796478,0.7678174215978767,
 0.5535201799935866,0.9474531319423583,0.4948478946145851,0.9566795785199854,
 0.891833578842739,0.8362807142856074,0.5051378937394267,0.9350083183188823,
 0.8858651438417389,0.6777770734959833,0.3650202250036516,0.5928586426235425,
 0.6038089832493556,0.09767860812504292,0.028616263743263204,
 0.3575215348980446,0.47530613694126456,0.5275452577930774,0.5577661753184222,
 0.8221503537322653,0.2609633840662311,0.7526607030209833,0.26062057588889065,
 0.09799972065935025,0.23990146546031188,0.3393202356010726,
 0.7865611408733977,0.994497164699772,0.5399977773632438,0.37165941496134725,
 0.6155065773690881,0.06800920744166994,0.1530838711924869,0.5934606429792086,
 0.6059554638422049,0.14247088251001305,0.5158680079205444,0.9060675761133951,
 0.9848286281155814,0.48921830980069136,0.6228862872799449,
 0.13132338676599387,0.6881366398517905,0.012032865492119793,
 0.4528973798067176,0.6279793821677925,0.13998561676050425,
 0.04941880179700642,0.38868048064762206,0.21215882758794924,
 0.012419055199449636,0.2823508919320037,0.06527038818066133,
 0.43134766359019816,0.5355321959965393,0.2758962490068012,0.8376102750779323,
 0.011062445210993643,0.18332903145163937,0.44240222058328516,
 0.31225039249632125,0.32247519624827137,0.4758360377729526,
 0.02593071220503962,0.40586842075497453,0.6513791748322693,
 0.6935896753190746,0.8746655649099627,0.4176348842228189,0.16682584914761767,
 0.7935572740613405,0.30175609444065254,0.14209738396162552,
 0.9829872563769351,0.08546013919721784,0.48318621152735874,
 0.25026928050494146,0.6621477432998896,0.18661401786623366,
 0.38671109125859804,0.034907338471200866,0.35397228645266965,
 0.3190632548916089,0.6094278358309725,0.02921303312233592,
 0.09841908857798753,0.6386803861370971,0.5764260575538039,0.8650912664895549,
 0.9193313419018462,0.19186620403653287,0.6994606362115006,
 0.35434704977659115,0.9221437751807544,0.11576219285964173,
 0.25773688715205334,0.3235923205580171,0.30561398150284425,
 0.038816035416532446,0.48996143993804253,0.029659484704552153,
 0.6803790562689574,0.5868873405854794,0.18973768813268788,0.8250537756165742,
 0.0987132096206248,0.9607371496973985,0.7724535806016211,0.7879388366579484,
 0.5326714553290719,0.9379117909994521,0.4039160621943616,0.8573685519987334,
 0.40856901843553595,0.40510016888207745,0.5790970939625296,
 0.3652196805385364,0.16730987691873644,0.9436234391729783,0.9767894943960512,
 0.04509871562739387,0.5699280035718463,0.9033578550150149,0.550117616918748,
 0.19496205550029944,0.9548028254325986,0.062487461988545934,
 0.3075342961146903,0.08694649949299549,0.44219195855448024,
 0.8386407301592491,0.07358661168045078,0.054109428275330496,
 0.5823214373458453,0.23560712392317607,0.1399525519533218,
 0.10469748736707984,0.5633903634314362,0.6621924255399818,0.7656059392994224,
 0.4700043561589713,0.3019019470361476,0.01134471048589647,0.6781825457255194,
 0.08965014093992751,0.044587660512394156,0.3753475976914876,
 0.5660931816372204,0.6917922955271544,0.8639616781239491,0.4640308759830758,
 0.9852475482103997,0.7602774135406909,0.6727594238813793,0.48350485493850537,
 0.6509485224857371,0.6295913320763808,0.8132400787430427,0.7272514125060787,
 0.7425509726916407,0.7670271446837506,0.1992556352631354,0.3638294262003168,
 0.3138923813261715,0.9624525483646144,0.3929772095626052,0.8549521015295751,
 0.31875955316620863,0.07437042519004333,0.6003268388648482,
 0.1802681974730298,0.8069931998512484,0.5927964201273047,0.172032141448071,
 0.9513435517431106,0.8578046191484525,0.7448918559079429,0.8654403410522655,
 0.4106772820564837,0.41282172877371837,0.19174246917637117,
 0.8585511142401904,0.5081380837040175,0.14883299794466043,
 0.16008423983343223,0.9784780172091656,0.7376247669433528,0.147678199437428,
 0.4759555803626708,0.7539739359660511,0.7948762329742679,0.4461021232972734,
 0.8038715506348024,0.8000826166671388,0.9526607285995489,0.6757712762805359,
 0.16052345526653744,0.8548077149525428,0.1379378832324143,
 0.44967395357240836,0.23851214421411915,0.2157745112992988,
 0.6518605004990808,0.12101540447013204,0.9349228288218688,0.5594504689510665,
 0.14675703688565522,0.43448804800734675,0.9819453751599087,
 0.8337812476183566,0.40023096213708864,0.18737726972591418,
 0.06418339661327355,0.21587469909356438,0.7411844240286699,
 0.8755218003593164,0.4225132694541922,0.9799436389803007,0.7587836961940746,
 0.418542183283511,0.3062813980200897,0.29700303627548985,0.15984357496580048,
 0.43605468460476515,0.5234903153470367,0.2289183752633076,
 0.46454415054528286,0.9654669732604558,0.7241698256845988,0.6664745144181018,
 0.21773336587130332,0.6640483629167017,0.15252274464336057,
 0.21439933605209405,0.32477115959906544,0.39477576353340704,
 0.08118664786020846,0.07424108475488667,0.6608235080221978,
 0.13821791595952582,0.9508478641203387,0.33116367910591615,0.383288202075018,
 0.7746410904313483,0.9554498708799366,0.6057112198986574,0.5411254157079657,
 0.26581488882382553,0.9761681901333478,0.6735705579495934,0.9332692665856651,
 0.986163461244052,0.41146285472854627,0.9422216498714571,0.7797414009714754,
 0.909902117197765,0.45337597691935105,0.753940296187304,0.33335959135350424,
 0.32512965923398407,0.9703183428834734,0.3840702971224278,
 0.02796324258438232,0.38237807250859124,0.5230894995933744,
 0.6925050153021219,0.6987071853598984,0.21959892997983643,
 0.23365353625053842,0.3841826614881393,0.21025640675330126,
 0.4020539164679382,0.5291325909555202,0.47972828772101717,0.8878703710848376,
 0.13988508451659099,0.716205544741343,0.32482966339508745,0.7246802219453714,
 0.3497597621874108,0.4193405342364753,0.978524916569279,0.41586447484363376,
 0.0435738883118425,0.8449042714627397,0.8206905427873785,0.6537062778450172,
 0.6942253746324216,0.12482291836231929,0.48793782816148734,
 0.5932601023900738,0.9730162422625384,0.25771146252239463,0.7909837197338816,
 0.3477086594116362,0.12073371861617399,0.150449365774301,0.374818267530889,
 0.8324199612892526,0.08241959283986722,0.8578654451361077,0.7336097120323186,
 0.0320640570453633,0.9871470981192891,0.7020122979281219,0.18014155168292045,
 0.4877749585255845,0.04555821806485061,0.17636874580394413,
 0.7239286291157406,0.2880645064385572,0.5917326312171215,0.944295326038794,
 0.02453283960596242,0.9354943841233762,0.7595566786447303,0.9616563677994352,
 0.38904328586915293,0.674422731346386,0.10614034110214277,0.821337480890794,
 0.9934092431622061,0.426112753469704,0.892471979273789,0.5327069567543924,
 0.9971930076415997,0.1457325414129509,0.6876905310539062,0.10742276137059381,
 0.6051768496563423,0.4943351846678964,0.3836241441251602,0.11539460075941754,
 0.3130379364513163,0.8469188671338468,0.7709097743942187,0.4291654776200241,
 0.7502051206691779,0.0021026758346136543,0.01848991061827776,
 0.6316113262295091,0.4754341254070269,0.5540503755820024,0.9030825842718833,
 0.8099207822326572,0.11649265202005599,0.22138164182111175,
 0.19156179406745188,0.26758280082904196,0.1788637233863789,
 0.7479458272217169,0.5950875104129765,0.624459486024229,0.35040866536116233,
 0.3168048051045491,0.22340712179316935,0.20616983945442724,
 0.9476955645092406,0.7475003808406435,0.43597651261664916,0.3362459605680226,
 0.6357685635725129,0.2674618816944978,0.18870037728584044,0.2615211171399824,
 0.5398314515552258,0.08800426244339854,0.7328224385878106,
 0.07598341350714055,0.7304285396891751,0.024563154419987354,
 0.026236540570353295,0.1663497257525628,0.2407884192495059,
 0.16455907737202313,0.9042704564554264,0.6866102013028376,0.5936602094990754,
 0.03181510977677282,0.35654260087324685,0.18341748763969945,
 0.569978646200812,0.8949034365688882,0.6263866895020263,0.08223544518612202,
 0.4080191177290672,0.8005743005878142,0.8463020401570287,0.3164131086072347,
 0.46786137286957596,0.9380586335427273,0.847748578089197,0.4776595813462532,
 0.7264044607943086,0.8709721232076191,0.7775938643204725,0.18841068521268667,
 0.666866166490382,0.8648762415573188,0.6689608315236029,0.14112611643425022,
 0.8820622813717106,0.44684325255029045,0.042915428210951956,
 0.9430257201135801,0.038639250128169045,0.3540280702935783,
 0.17668275851208226,0.1727296915668357,0.4382630783351721,
 0.35305645533419105,0.07107905994743546,0.2073201983081947,
 0.0031019591026308913,0.34513877207169563,0.6164797779364256,
 0.5024943774976222,0.9602773761940386,0.6256363540922605,0.47402281882004427,
 0.573431576266469,0.38208766062622423,0.3077185430520104,0.3396720313014797,
 0.2826044921141414,0.5609034189755499,0.4602385095329464,0.1113212585643002,
 0.6175183446141372,0.8365079820360133,0.3674140452585085,0.22606672118914184,
 0.021108015379098233,0.4132286897192258,0.07327596060334085,
 0.9493088113414808,0.7914131155815426,0.18263990107316075,0.4068184175095664,
 0.1436256873399726,0.82411828935569,0.7650207326795406,0.2365009168882638,
 0.08255757666522523,0.6794583955462301,0.4317558621649118,0.5482114669027118,
 0.02207805559242404,0.9861282306247149,0.09123412523079644,
 0.7604492083747301,0.9131680029414113,0.2602473959054269,0.7526668217199497,
 0.9372612161552427,0.8011866273219135,0.9313623548440699,0.33974296943972637,
 0.4735783435081725,0.6083911627358469,0.7271109963299193,0.22918208008949625,
 0.056939054612115125,0.048393365076593975,0.5759181232990808,
 0.7521607428125915,0.5182934928553227,0.46693584630352025,0.7809908519593587,
 0.8313948442452183,0.982317722535222,0.8509447666871246,0.8303141628458096,
 0.31008190326394824,0.26365126082120327,0.4338498846171479,
 0.2692264903601177,0.3975759511943525,0.7087978648859392,0.6665969505634217,
 0.10772245953230641,0.6785356058089499,0.9322850689092205,0.598851052669149,
 0.3694774360458657,0.8258586983219947,0.3054485085693397,0.852797515035335,
 0.9055003030099618,0.9061138564113227,0.8761834641589072,0.47330656288040407,
 0.6445089620452636,0.4356375215888699,0.8567545203085022,0.9563278165583902,
 0.17905348846686475,0.5648920829894107,0.2607871238694983,0.9680407965397146,
 0.045653062500704555,0.5952850035865478,0.712760765679206,0.3132716827835694,
 0.20518079087624652,0.23616803106316075,0.5911333692597459,
 0.3175440666829443,0.7561620898457541,0.959349640755964,0.37003621793019914,
 0.8575289543234004,0.04261063485306971,0.6720664618287158,0.3940340791609336,
 0.8699378395339622,0.46829316877199234,0.6067371924042546,0.4032792563510218,
 0.6431370995126633,0.4601816786421512,0.7109782657383557,0.5923266970584864,
 0.6053150466318163,0.06158140548903934,0.005349243074511012,
 0.20644917707497334,0.2016197597444056,0.2577811365235487,0.322987628731103,
 0.9319253855486078,0.05373319948486499,0.9000305526511612,
 0.03908194110024521,0.18922618865574237,0.8103889730271192,
 0.09420531005315425,0.8376407870876051,0.3957701303481841,
 0.11776782538548736,0.07772207154696209,0.8940899332769329,
 0.7179156862103817,0.49114930185010375,0.20321651552002185,
 0.7622389374614529,0.18638103958448793,0.19178482386539342,0.43270777243728,
 0.634304089876053,0.830433802496693,0.8854439681061345,0.3610926452112231,
 0.10876247321998511,0.1397128823647933,0.8852966692141565,0.8607829598349546,
 0.19467068182479508,0.053095461424875645,0.6142556130178523,
 0.6072283918082055,0.43219250074584015,0.6116309529198334,0.6173762751006132,
 0.43789628648800216,0.29173696585102715,0.45946236964300424,
 0.47572996052675487,0.4178765999010404,0.8617464518776973,0.2708190092278068,
 0.6749025255030552,0.28488659969411745,0.4715811511599939,0.3090607563329574,
 0.9355768403129062,0.24298757503476331,0.06340242148887199,
 0.16082975998058746,0.3320945648354544,0.29560273750162436,
 0.7509871389489424,0.05082999571415592,0.07927456959051338,
 0.32700966535054077,0.5958277366822745,0.9321195537739017,0.3642221700913628,
 0.678929850400193,0.07586028280898682,0.8030549261573182,0.27022884754216425,
 0.24932689446472078,0.6983610625679257,0.3267873748058834,0.5987973251515422,
 0.19977044521638698,0.7087515940415978,0.08304136099216652,0.445049597191311,
 0.21211276289727055,0.43414311258913174,0.05588100138879548,
 0.10502583781328412,0.9175127614978025,0.16101342540083363,
 0.3588288756222585,0.7876217440228013,0.23609158949990905,0.4855781847456494,
 0.75785676107612,0.37309037750571306,0.30047792375975013,0.7277539541384324,
 0.7696728794625212,0.5929795429925628,0.4153295171716602,0.5677998877251612,
 0.3629300272358702,0.9171401951418388,0.3721356787324178,0.12363757548119869,
 0.29642324199143577,0.9715284396363251,0.9220590892497484,0.7148000761405338,
 0.8053727075417186,0.6418097828633269,0.8824753170790822,0.2916480291566238,
 0.6092676175459752,0.5085419793047112,0.9611003033406705,0.593910031603052,
 0.6175930450450575,0.9541929136569631,0.4877887638034327,0.10070445247906168,
 0.9153305087757938,0.025864146544018585,0.14030921763423865,
 0.2015043191172935,0.99643322656368,0.979302379809918,0.9031387999849644,
 0.4374449767480404,0.6910187040030641,0.30504790379768776,0.779016267807065,
 0.5619597720193019,0.5095381207110874,0.5696882022664201,0.4185323435399164,
 0.5672428433757457,0.08415607984932183,0.07502807831835168,
 0.2790343147419516,0.16776208294670647,0.0914819193471984,
 0.026127290705062922,0.9498591549938531,0.5273811107697455,
 0.4240713695757069,0.7949430706401277,0.9760558990681063,0.6172941831362477,
 0.5754618764368167,0.9871103713360153,0.29249061764061435,0.9145574884170886,
 0.7506385231064987,0.2373278656114528,0.15521862535416686,0.6071538320182569,
 0.7052950141497778,0.451713347929099,0.13295891397639026,0.48043823369208905,
 0.40215663294220005,0.27136617044626654,0.2853939441040192,
 0.6388146288699815,0.18463378425653587,0.2946893458308666,0.7352550076654242,
 0.041594606166480697,0.6588228706664816,0.5254906233651051,
 0.45306581239588695,0.9967292349839006,0.3853300381758099,0.4951739662223059,
 0.9499095155364135,0.08812656842410216,0.9662245133330116,
 0.24893560494924105,0.4646457173523235,0.16727270098119507,0.570903507845771,
 0.4055308636900442,0.6720853711804934,0.07402499702609033,0.5272140551422809,
 0.6476305427636702,0.1190624328824107,0.5133464221844048,0.744026284894816,
 0.5342936631454726,0.1996100350312372,0.5572559153293778,0.6787398176378783,
 0.7179120159289077,0.7822488513435583,0.22947998742861864,
 0.27134094216780136,0.9089187290323956,0.4486122881870581,0.3464044682043088,
 0.39834759047599533,0.9310294308683997,0.23016487516841067,
 0.16058794327827774,0.9397540855606719,0.8864658635124731,0.849476184141953,
 0.6310682097660257,0.8151203174038892,0.9296489417740035,0.40487172756830025,
 0.40996041401538286,0.8298601956562435,0.3106054230391715,0.930692837839212,
 0.4763447284145419,0.479114872191596,0.5232510028641416,0.15621106579775468,
 0.8233366466798286,0.8181662675975987,0.05841464719623346,0.9211663852265841,
 0.9941407456804789,0.7250386719913523,0.15504403045182813,
 0.35502656285603096,0.8086131130054149,0.40266470882612393,
 0.3547675664718235,0.3590796135561434,0.23831359694469356,0.6523515358733765,
 0.4397449222505403,0.7004180812181452,0.257451654680959,0.39483639747755594,
 0.17275848309137465,0.24481454232121147,0.025528953962774992,
 0.5493873569891209,0.18489030958612063,0.285531825163464,0.6974565276702106,
 0.45245638689483214,0.1571979679163158,0.7641066892725207,0.7491462519340869,
 0.40739451737041565,0.5265407377714603,0.4591513546654369,0.6649065022573605,
 0.0949845161431575,0.2971810912398537,0.1979312866653299,0.23133942114070427,
 0.2539864605015387,0.7633771658091625,0.5061982575345215,0.7651757978005834,
 0.1485531100794999,0.14850976383596703,0.834189104824137,0.5837252633536588,
 0.8407107716823358,0.1461460112747872,0.4455577011513844,0.5298770180113106,
 0.7930523143661569,0.8164497354684619,0.9706570106950928,0.31875526255517417,
 0.6781837846402099,0.9854983201942498,0.1932681570036907,0.15964579274823854,
 0.5624710838365892,0.7595715469461521,0.824266932959698,0.6696151669696162,
 0.3438081614731171,0.34141219431615466,0.8210040206141671,0.9245488238986156,
 0.46951197505618225,0.8703323222217765,0.9774279401447917,
 0.12347567421542116,0.6500629059345618,0.820794646372378,0.5971785073888515,
 0.34314808479998826,0.0473077449808289,0.9419643939787562,
 0.47088262626582156,0.3178733148804942,0.8013786722491273,
 0.13244689345249783,0.289595409918866,0.8287925328257073,0.9116562049176712,
 0.23187523625058215,0.08590191341273312,0.6534652090710159,
 0.03996704043852528,0.8879303853360966,0.742127698332816,0.24068942457350206,
 0.8205941248659095,0.9867988069288752,0.8030805226338468,0.6646535802515533,
 0.7688242793578607,0.3473765398628349,0.6927187724644256,0.4701509562583007,
 0.7551882509792739,0.8078739787441228,0.9546636776959336,0.5339830253794525,
 0.5662004672036568,0.3067297509277206,0.9023863342823687,0.2750727307868447,
 0.7649105924123776,0.4438150726572405,0.6596018934566893,0.4186412351080472,
 0.6821289596927325,0.13160132908071953,0.4446099810344183,0.5927121381609116,
 0.3621824739451326,0.7169316867267262,0.9701057838218179,0.4650157303609874,
 0.25550436349356564,0.5227654540530782,0.7238477184649164,0.7425697927505588,
 0.9748428298405929,0.27987784820793504,0.4528295101276332,0.7546346947808299,
 0.9651967727076415,0.2051379924214326,0.1291467723609494,0.18310402062306352,
 0.8263141706590319,0.24274253422119085,0.8354279200842081,0.5339474991787306,
 0.41521166228989503,0.7654825355212777,0.836715104951922,0.07983230233852734,
 0.08754935805043862,0.8437898690357855,0.8242205080978608,0.1421570390706437,
 0.5533090865282158,0.031259313512607134,0.6967502578143989,
 0.48390953519145974,0.7857560480539078,0.3402386300708823,0.0988731249412188,
 0.019646086635649596,0.09529428424290676,0.5069842303309539,
 0.9639506373327209,0.43832397415245616,0.5731710427874684,0.831744145867179,
 0.9479517149479544,0.6269563552148901,0.4223693838823267,0.18918419378897133,
 0.7208257662886093,0.16548127895120113,0.6873529772614266,
 0.0015523129209458197,0.7070427334945802,0.21398320328177678,
 0.6810588529725806,0.6776487582419691,0.09165110459928583,
 0.18138909457059882,0.7354683914989782,0.41520670371635615,
 0.8592456558918957,0.9821419972694649,0.31351133089930516,0.7850292365416004,
 0.9244980205133085,0.17117456246664553,0.4526776868094802,
 0.13081829687511237,0.08383883281226612,0.18939298282639416,
 0.4058733174451157,0.42032234266129787,0.027264672075385965,
 0.6155935111533747,0.7789557697592061,0.6752935357886412,0.7450828266431883,
 0.6589243029936191,0.4252117577906971,0.7534877834531364,0.49564938715618745,
 0.042269264864363665,0.8740394858533057,0.9342614729766301,
 0.4755659405328253,0.16333506949272747,0.04498767802933623,
 0.47917848558097476,0.5373067908974334,0.9622201038321919,0.9113041340597127,
 0.9939652935483807,0.6043637333215633,0.24334552846205226,0.2265106646784203,
 0.34290634220617267,0.9449886029132153,0.5908958220434583,0.6174842013551493,
 0.2811855287805679,0.7666730785218809,0.8689257708882097,0.41069866254625753,
 0.2792401249985361,0.5871224950629246,0.2185076300307508,0.698724921290585,
 0.14364058909156407,0.24905180016683137,0.2740089322378566,
 0.9520042134642542,0.37385246480826007,0.5168062849864256,0.1409893740771212,
 0.370777568957133,0.27898837731384596,0.1620368105258737,0.4920879686166034,
 0.4719517192828595,0.40429695107579056,0.5513659952643746,0.7220365214860043,
 0.761639545181224,0.19627009210002822,0.5092258101320158,0.29788958251408537,
 0.14924305170958385,0.9102456153760948,0.6436501711782108,0.7500328123174753,
 0.6780902333290664,0.5026938650840279,0.08857186337762368,0.8619925671940831,
 0.09244589540092696,0.9624701942850513,0.6461391737559508,
 0.30428010212906775,0.6357411009164595,0.035616813576939466,
 0.879549602702723,0.6553424122365843,0.9454093902555607,0.21090030210603983,
 0.4091014234750652,0.4147469355399902,0.3747235959603906,0.3637674842682762,
 0.8581565619581966,0.8127030998148947,0.8680031604803,0.1293693507351138,
 0.8547275378240166,0.32284701180920605,0.9111235195331536,0.9536782227981875,
 0.3563150960317961,0.5894261159716672,0.3686318547899913,0.2084889177731879,
 0.3632639454408676,0.932001796400623,0.04517759967570645,0.07382089905040445,
 0.5723698805817417,0.26141392035750344,0.7449608753230257,
 0.24878400286417623,0.46440915911625125,0.7345236650249127,
 0.6520611692541699,0.9312187131488621,0.5564458068285375,0.7533833888017089,
 0.7576506669851591,0.019923092985427604,0.5896676713696003,
 0.5600485407829978,0.8886256039491713,0.5744348939326764,0.1398050213133868,
 0.5665692703703273,0.6142543734402728,0.32482066150074795,0.7000247424075413,
 0.030107063812280455,0.9605869584428833,0.21729337939124416,
 0.6669937698605586,0.7247885841126596,0.7147072473683426,0.9350484660880822,
 0.6627014927275874,0.22504959233159083,0.7927648760230228,0.3958448391695585,
 0.46008121227549614,0.3642087719521395,0.7453716161286348,
 0.46424325354702556,0.9229488549308777,0.1887528973158772,0.7091733090876081,
 0.15585908777728663,0.6569613762664565,0.798400166526988,0.7828222871770988,
 0.20619161038147682,0.38881935669977485,0.8652183481824778,
 0.09133941796191214,0.2898408220653751,0.30249298625106014,
 0.7478598645110808,0.5020196711088545,0.20721969700688814,0.6384809048821308,
 0.04200422131494885,0.7248395305288493,0.6064110051514219,
 0.19137427672452312,0.22325589457358275,0.8129424219533273,
 0.26279462389966435,0.10680275394338568,0.44586547216593575,
 0.5836085840347169,0.9738186956963102,0.7713705674804645,0.07327113535336549,
 0.7547609686970017,0.010925521519472237,0.11749027519740896,
 0.3755832358316099,0.16395434838525758,0.693782872080336,0.8303562954467401,
 0.4451003702398939,0.5648842744479236,0.8289288079594108,0.41942352084499457,
 0.9374430808209999,0.12482093032529751,0.46313809186112664,
 0.017590699630843365,0.04131243913793481,0.628256034316171,
 0.5503667717514613,0.10557190546234485,0.006179068271846222,
 0.08802455975468981,0.6377318061748811,0.9806859409022639,0.1793003331207632,
 0.5704801695045737,0.46670804867055704,0.5109748742003543,0.8834392434099286,
 0.6537097000330069,0.22081967516875645,0.977869292468316,0.5664657056425758,
 0.30132806420255664,0.3778182319347052,0.8151750906121774,
 0.0023889331677411363,0.8199473832033568,0.5877829705039678,
 0.2708218242639562,0.25025629669716054,0.9059107859840942,0.7418806798373394,
 0.4544412813184243,0.7129119324961154,0.08105469132584198,0.1526669216553278,
 0.4655943500022782,0.4727596492662387,0.38396859441018083,0.8727559734441089,
 0.2485751420440303,0.69923227249664,0.9186834190491613,0.12338613855213498,
 0.33891388002828104,0.6894513353063936,0.18122268070989123,
 0.3237117982141572,0.31436760643467043,0.9157151322756674,
 0.32820849934987495,0.7885435077068964,0.08722399432572736,
 0.7223684140228126,0.8371387819944551,0.9538666086384657,0.14183009585440154,
 0.8954557574788932,0.2752222597207804,0.5611090491629832,0.26111908781234217,
 0.05326667093961157,0.06491104368874712,0.49444359914654845,
 0.11707733740240533,0.16241964485129312,0.6012808231827352,
 0.03590006327969486,0.21098463942969026,0.38305254209005823,
 0.4579030653246382,0.01318847261529743,0.09723531881404535,
 0.7610321497528122,0.8541029689777242,0.27555858645975595,
 0.29008986679797655,0.8419576734997071,0.5229257214146963,
 0.16604042236729388,0.054603007598157394,0.0886714770659095,
 0.42383383538199904,0.6788863472598149,0.3537852407201486,0.6406888521849928,
 0.5422778138089313,0.15084721578617333,0.5920380900649742,0.5609061417946224,
 0.5136284114311714,0.41569912050027047,0.21617602425092097,
 0.9331909061625767,0.4812482537678283,0.8536572316743826,0.842934405247595,
 0.5326496681535953,0.9912983391467847,0.6911127617388576,0.38851948807291226,
 0.28472078320626226,0.2793266429574869,0.6838233826618942,0.874573372562939,
 0.6742087916594014,0.8807120250894839,0.7920763724407338,0.09327284652248158,
 0.6074675489055054,0.3342370492432917,0.45360697049317134,
 0.47125293379852373,0.9128864660463967,0.953053868099841,0.49310223828616095,
 0.010115855569357346,0.935710791132752,0.6098231184907514,0.5168788710595741,
 0.5159325788984284,0.13687688944927634,0.5946515586859893,0.8063270447766887,
 0.9165563543404913,0.42826021010633497,0.6738421974675672,0.884270532139159,
 0.791487093402405,0.7259794898534191,0.8051328843737158,0.8549673785848894,
 0.2368417368728748,0.02351160451257328,0.1284850856202705,0.8667236554632038,
 0.3141066176050987,0.34831603055976035,0.5922334666396464,
 0.42884255703898955,0.26124258128500366,0.37102635561678976,
 0.6882358192607148,0.6763971268100644,0.19396313829388412,0.9690991132247481,
 0.7333637224640398,0.3228706650139228,0.4521818994860216,0.8637427724709589,
 0.46022218685597305,0.12527279832105387,0.17818481010657095,
 0.10562678156625194,0.9370394231147022,0.9497023706260835,0.7363023734785302,
 0.029550308464850294,0.14189213904969034,0.636939687448898,
 0.8086905973076678,0.7542304931768518,0.161702399451907,0.4144192057519209,
 0.5650050263587011,0.08099048339072268,0.050849514664133455,
 0.19755854810991347,0.007686073033992269,0.37402392332434165,
 0.6231328459410264,0.1832641260321337,0.23492735141936627,0.2560065581091894,
 0.7887568233086834,0.379878794874292,0.3926664753352366,0.6135445042958673,
 0.4989789921313379,0.29537731025250025,0.585887474291973,0.20938528726625916,
 0.9687029683117437,0.2296417973294983,0.4259311111851991,0.3169830413961111,
 0.7573896902629426,0.13774077644549654,0.20471192424940687,
 0.9108573720089086,0.7105311218618215,0.43285967510611334,0.3767128406062521,
 0.7004852449198189,0.12963689333175332,0.5611664561531857,0.593813698178353,
 0.12531825084572534,0.12579842832907384,0.15942678215378148,
 0.16051129296472488,0.54841260160311,0.41113191550766615,0.17235821514856187,
 0.7549596218956451,0.9224455265585764,0.596229876165046,0.7318810031971603,
 0.09860501170561753,0.5100043884733869,0.19913266193985746,
 0.16188691732899307,0.45416142609292154,0.12416845687709666,
 0.09016198575065615,0.8923589850212725,0.4251099455168872,0.727119917995255,
 0.8282223143307516,0.8083654395564979,0.2370367742748012,0.8120970245394962,
 0.5987756395082788,0.22970756645335888,0.12341213616500779,
 0.9204525716817221,0.18722745339994973,0.954932736716641,0.5130521553322691,
 0.5394028841250902,0.011510123262440164,0.36669901152667317,
 0.7051329832481079,0.5869023649944584,0.28716208775423135,0.9985263915333888,
 0.5803682148872892,0.5807355533393207,0.4212441880923896,0.7505994644612312,
 0.9794308749516123,0.6079863067271226,0.5780463528033787,0.7456119873525189,
 0.03816788034671126,0.8883862233691582,0.38547479216348024,
 0.8958552091746701,0.23546346575614652,0.8337571094840724,0.706714193301641,
 0.7338087686371086,0.8581096656655053,0.8565896834560711,0.2892469826169586,
 0.06800943327978948,0.6802159155193883,0.9700996465540793,
 0.026141143371359865,0.6349719309660478,0.3040689714083874,
 0.44192326030995854,0.49820328391909374,0.5874849869361733,
 0.6503795375641858,0.7642973330155423,0.7398254069656459,0.7082048087090704,
 0.7026242329011456,0.16123320745019354,0.39459644030989005,
 0.1139582299195594,0.32369797717741344,0.8094111331675456,0.107455433120893,
 0.354089763424674,0.5542312514151546,0.6637640144960787,0.9815586400842222,
 0.808884474579957,0.6125003859864473,0.6264096660197492,0.7947734786072417,
 0.8852949770718227,0.45384037007955025,0.657097008793325,0.7736862528937567,
 0.5589916157561483,0.2585111289981634,0.4402485720115349,0.3124926912129077,
 0.49666983395345654,0.030379537405288204,0.8312992296420613,
 0.5522244895724561,0.7765157773495731,0.14265932959729843,0.5648634254392959,
 0.6777532196240742,0.22702776128378677,0.09138779765787042,
 0.8768253649856793,0.16962294871448746,0.5858471687554155,0.6862360935026836,
 0.1523310757517311,0.6993980090218103,0.12419078522623217,0.8134028483416211,
 0.9669770608972801,0.2618567147948432,0.9228783989603825,0.08477417896742012,
 0.23168906067917794,0.27423939238749606,0.8236059207010735,
 0.8097408388883025,0.32751630237889773,0.3069658303030185,
 0.034278096956498416,0.03552359558119156,0.5878571619776826,
 0.3644330585582448,0.18834667324350018,0.6432690930202364,
 0.03688853052211405,0.31337775556968706,0.9599995387884257,
 0.03824948051971522,0.6951741506998065,0.07157448722510484,
 0.4293454805782141,0.48139767573763603,0.21637074981636972,
 0.9180936569070925,0.36035458658094,0.5137124315182393,0.9466081448681326,
 0.9112342738222392,0.3240320056821018,0.5604575615813889,0.1994113523635338,
 0.2311084586646981,0.9284883557705091,0.26400338257420164,0.5617285293062814,
 0.07939573669230748,0.3265021921259509,0.1932310617221772,0.3951734870587339,
 0.30542612803506775,0.26626295409373146,0.9798311587584146,
 0.8854601106723105,0.07507183201232093,0.4098619027148961,
 0.08523002826203452,0.4484497262346455,0.11810742178865519,0.941880022550168,
 0.14543631556907166,0.9282451098796882,0.12742599195683424,
 0.23436494952213505,0.19462321585842446,0.675809829245152,0.5860674745601518,
 0.740720366110645,0.32454797868735796,0.44081656646271883,
 0.020357783893157233,0.1982264387574011,0.9469047833819124,
 0.1924471766280491,0.4966659516288986,0.7905670069880996,0.37134094695491815,
 0.28925929725312904,0.5719333009353785,0.32113321729514843,
 0.4107818747815313,0.9654666675123278,0.6638475188058033,0.146786194183131,
 0.11460045334744684,0.5121006379938654,0.6031489619001735,
 0.01771989826848097,0.10210607699384722,0.044826533482107744,
 0.24979462310229272,0.6087439008365672,0.6981340430633429,0.0653317439549087,
 0.48244299089337683,0.15794693716033592,0.3861447952184073,
 0.5961780448483914,0.3836305487529671,0.2326897892976807,0.47559305583559297,
 0.07251556984363638,0.023592742949083645,0.1967128432692271,
 0.8774251776011657,0.2823611417355524,0.6668396575291567,0.6857478953495146,
 0.33546716671036403,0.9973392848388214,0.28527470096619034,
 0.35322591154681016,0.32621513145978887,0.39307063482710614,
 0.07534009075674963,0.401079275026169,0.42692060293529566,0.4163806161424817,
 0.7946270139470216,0.1176588243005613,0.9687650540933654,0.0972809868028246,
 0.21343068001725596,0.9654656294412006,0.15603854387982197,
 0.7742738173128759,0.3509298583866658,0.5031545389457646,0.8539358256451158,
 0.3764211490566248,0.18510326271036548,0.22275898619816537,
 0.33921944596891296,0.5085345169337584,0.8801278888201438,0.8576246600674137,
 0.5041341149059299,0.7938256407380739,0.402573544931793,0.48667798423020947,
 0.35240284283266954,0.1291431980442881,0.5274821207166804,0.9813773941553772,
 0.3970720812759565,0.7259376229847423,0.8720865048860136,0.8440035807835435,
 0.8889854707777314,0.7751192500694638,0.6106826638714726,0.42404318609248126,
 0.6628868894842166,0.652381771655402,0.4462528892001587,0.47751679000957936,
 0.4943850574152321,0.4255661066635791,0.7630940912431141,0.04571162330583978,
 0.45950878028027187,0.8023516133126092,0.7375736704817272,0.0531719060799708,
 0.09168927885286815,0.7704118538304575,0.6279827957094437,
 0.028177487958441327,0.017938858785518375,0.17508659316638997,
 0.39311945082111543,0.2461475177102117,0.6542077403497166,0.6104477392615532,
 0.055601156317503286,0.06114411346038173,0.11073541558246558,
 0.38331479115992273,0.8671692597618759,0.39939374084959145,
 0.8498306947752984,0.20681460102424798,0.47431470437328216,
 0.9746842300161611,0.7504799989985522,0.26952645373005746,0.5640090464906045,
 0.3283475772816027,0.6891416886118447,0.32875934497384174,0.843439792863298,
 0.4310015871333043,0.6428260524323732,0.5178509245855949,0.558748639421584,
 0.618119905978383,0.04501423799597992,0.7405788529304248,0.2791658555080262,
 0.9413258356709421,0.12797469316190913,0.7315488124565064,0.69763722730684,
 0.7481482497301374,0.20287223230886842,0.08431956308065891,
 0.5265785216213981,0.05971420767235469,0.45865495955779056,
 0.48412931206561116,0.2182200144641262,0.37281195565225367,
 0.9262381290814974,0.6191284203849446,0.9220994548509699,0.8358122453230006,
 0.4263191558334998,0.07104945805747398,0.28178400355863065,
 0.9016787535849278,0.8698339675578024,0.8506529702507998,0.4907719313175458,
 0.32396321207019796,0.8296630381881366,0.028277951596176965,
 0.5683114116687925,0.2143406813645542,0.6062502727170673,0.5649947901292269,
 0.30101127764684,0.9716698466398622,0.9653092979628292,0.060130113839239274,
 0.4505641913751044,0.7034830682070339,0.27033797903563483,0.8259233903517273,
 0.21882075258677425,0.2735689573529584,0.08399452546904007,
 0.7572262413567287,0.7439399890251039,0.9889037357137351,0.630007216155864,
 0.4903078319679819,0.299358060632568,0.9498567266436246,0.5583793134641328,
 0.8364119621321877,0.40175107817791633,0.6693381324787923,0.8148219961598075,
 0.5376757806560155,0.9924672660474738,0.6833826546258353,0.9675783818334935,
 0.08651827911098453,0.9278286651884045,0.2936912100725467,0.9219503449535517,
 0.0797724971555307,0.7811597495990182,0.6263009244866897,0.17677674626036932,
 0.6927790532185862,0.9911627022711816,0.10379732778391437,0.6265503149188241,
 0.6735062023239972,0.681886038839661,0.33479174117863064,0.517012473086554,
 0.684217915517765,0.6093628561967309,0.8768530412083501,0.7946718487881157,
 0.7106595237326597,0.09461460941050603,0.85077785688628,0.2462788588196716,
 0.4843122852499494,0.2828271820546393,0.4849542217371371,0.41656939476166044,
 0.9051520398191005,0.48448500064674016,0.5523425393965388,
 0.00911199055295342,0.7070667489724446,0.7881773154530596,0.7014259598587901,
 0.1884518036795947,0.059870901167354273,0.07508792531732711,
 0.4832668932637858,0.9911139269787242,0.9851415089080754,0.8753530585352671,
 0.5523898583609641,0.9278174811942937,0.5644734398985428,0.566810702829079,
 0.9776862119935938,0.8501899944724886,0.9814319453090786,0.6221846354706445,
 0.6634856236554636,0.7719305780166644,0.4205904450763197,0.40781051905599097,
 0.48620788446425156,0.19573384738376787,0.3558951642609246,
 0.9691924490937813,0.6774545272939168,0.39753105529126886,
 0.11202475465597161,0.41439427186659583,0.721719609354512,
 0.43107089467908677,0.029764407247290015,0.34619431661700173,
 0.7563877546712832,0.7846996757099625,0.08464734583972222,0.7337950685520126,
 0.6026430942332465,0.5987523523446246,0.5632085012162328,0.07246880733686412,
 0.2759266426396719,0.21169753915672862,0.3567904806901997,0.0372043418551804,
 0.41834995254508256,0.23686649152872474,0.06793909638710305,
 0.6401588027160777,0.036550901534902036,0.05881815637510279,
 0.35841676243377707,0.8699486952637955,0.6147824084756216,0.9554374511160054,
 0.19442273207092264,0.9648274834108705,0.5776615275766717,0.6435133047772968,
 0.9809719047498671,0.8682860531411187,0.7938007035521176,0.396959385803787,
 0.3236339659383958,0.3057214052817444,0.014609809559208542,
 0.41947104952343617,0.11776225280760089,0.7819378879739709,
 0.7167440803865203,0.5829191683725496,0.23786024073863876,0.7153204257145629,
 0.45643388347397496,0.8131495340935526,0.658856328341693,0.861839692090808,
 0.8227758402015488,0.7360306140733581,0.8672914573083347,0.7515683961281667,
 0.6736502378999353,0.6266932074287858,0.8483618317092044,0.8566212602508547,
 0.9883328087589375,0.9782910397888691,0.4001920979873408,0.6070791526625474,
 0.8579153610773835,0.09801558916652764,0.05482056581924066,
 0.43562872835072886,0.5286864317287131,0.3606593237522814,
 0.22612954716902767,0.43899153955282677,0.06499085073848032,
 0.06637326740002347,0.27635584527765955,0.808546096888993,
 0.39015247494938066,0.27719869741669956,0.6862196670729055,0.581397087378236,
 0.10636638524317843,0.8368569939192856,0.19879976018437773,
 0.8736932459718362,0.875095585950892,0.028414860127661923,0.5513678837186747,
 0.5663742987234247,0.9793572388477091,0.9103520360410098,0.09028142167682407,
 0.23817712363906074,0.19206213372125092,0.7706891099993005,
 0.5281183681956515,0.624220163184517,0.5055429214319598,0.03807604817647925,
 0.7148206008550003,0.6404127521564666,0.7606821515620779,0.7805787294042343,
 0.6131234827599574,0.38027440240681765,0.9593277167532486,0.8856282207322894,
 0.42685152103680535,0.32418491856158416,0.509751378871842,0.4660903461844971,
 0.12270705767144985,0.39001644267091473,0.42130479116084196,
 0.508673394514721,0.8341911547419947,0.6921742127275645,0.5270811596502897,
 0.224842317884424,0.5694942757970339,0.11060649613488782,0.45330265724855356,
 0.8807405553535298,0.43825004066419304,0.6956491600552756,
 0.31607458875462213,0.2556370980908286,0.003329177839317299,
 0.2422710956715315,0.5139577418391738,0.4764110670424344,0.8366323898563517,
 0.4323751792857329,0.38660685130111583,0.772355868083213,0.2529043591537783,
 0.28713824315889114,0.8880052278698992,0.416732032319302,0.6612005511116931,
 0.4893726333406274,0.9636810006065004,0.2383479638256305,0.13728228202571358,
 0.4401192133721905,0.5228186419092782,0.4496055282502036,0.13147842361466733,
 0.21501122618334867,0.08383035658354665,0.24886034432481074,
 0.6661903627531381,0.7919300089179627,0.27029707221550414,0.32784228204622,
 0.07008983190939932,0.7164358895906948,0.2482830370759519,0.3245671013976229,
 0.8775420321059954,0.4901815551263591,0.46996490982702555,0.8011623523341774,
 0.6741769473127881,0.08881960368295927,0.9352179584573928,0.9838732859624639,
 0.09365359042398846,0.42277524463053695,0.5468762017729064,
 0.6236340172038702,0.7621134492544053,0.48295217573932137,
 0.11905543663269447,0.9430410787905052,0.2471312360837402,0.3460845274647213,
 0.23623470319972528,0.01262691187620868,0.3680008052458743,
 0.3735521214718084,0.47566590183240254,0.5049110557410098,0.744077532050634,
 0.668641732520539,0.8381902218386315,0.19087520022414872,0.6230903360831571,
 0.4633779397721163,0.03329032434296453,0.15955929716014383,
 0.7276111830187789,0.14939076534819573,0.5191906125549388,
 0.40299927185648865,0.16878748293838797,0.4743604020101262,
 0.1363164112305173,0.6431661863507502,0.4146391149188059,
 0.020792186771882437,0.13694734177224155,0.5504070153047059,
 0.09384339198778469,0.8885924245966682,0.7678388116640439,0.2866464286063758,
 0.04386034814996387,0.11019031973303894,0.6657256768794928,0.85242460587907,
 0.3293506758351553,0.8392388230014874,0.11813646299281655,0.6379697688032817,
 0.9659541596771225,0.060275322200910586,0.23331779053779744,
 0.5930405274922409,0.05759229002980937,0.07232664256587784,
 0.15153125684426938,0.8718414267918609,0.48901698705577,0.4754387994647816,
 0.898224089281264,0.8477809585264677,0.5750346278800855,0.7755735734205931,
 0.8391142019690861,0.9796003198852337,0.6378440072450207,0.3796934913740484,
 0.09346809770728204,0.23720367172551404,0.951416317476472,0.6363955880212208,
 0.03056729751817666,0.11685694308039207,0.41979075096532703,
 0.17415736820720285,0.9668348511057203,0.24342505458631092,
 0.9591580440990365,0.12420733953292684,0.9005526556683883,
 0.31797388468500287,0.39351608651049075,0.030677306683165262,
 0.1307925559219837,0.9958972305529448,0.8228633067689727,0.8768930321718876,
 0.394762439352775,0.2660675441907856,0.18551936129638846,0.01706785059284832,
 0.6359705930447727,0.620398980726814,0.9824047262396962,0.9237674882169673,
 0.9555233224862397,0.4093306245023911,0.19008642771637152,0.7407029211373487,
 0.33558610643330344,0.16473186584743493,0.9636684513117766,0.799176054678949,
 0.3714488402547502,0.28589172804714746,0.08490538041959794,
 0.4049628106648472,0.15051678873746843,0.29530544042135487,
 0.8068479260991812,0.8679331201228676,0.751521884463457,0.6673840346923434,
 0.3121571028275629,0.944607330006171,0.4446474695990553,0.5524064845954584,
 0.820979884321527,0.6822171330713462,0.2364318935904085,0.08451212198788716,
 0.8853412434127327,0.39729846069081765,0.3916668116027531,
 0.25058257127752726,0.3744710371725384,0.9169965369308583,0.5007324012663572,
 0.632244654407193,0.5923823709398324,0.28680171836318613,0.6752600490685413,
 0.22298368153311254,0.6848507091842695,0.32349975769216865,
 0.08348304758270386,0.05506185908089256,0.2828972397376813,
 0.31067408368459337,0.8997393716705876,0.6385817546149367,0.9344833266330224,
 0.06165696287254163,0.7821538834721703,0.0873248876395265,0.6674338083212559,
 0.8101020346097737,0.6151457569186095,0.4691253896332368,0.6008633786184321,
 0.9715746530407414,0.3097406328315295,0.30089102860357453,0.8339470415138787,
 0.41226417384844594,0.5865763212544566,0.34381161737640187,
 0.22441111329088403,0.2626833928705572,0.20326287073520055,
 0.9286606452034107,0.9024466089913031,0.1462340415954022,0.2822512533611382,
 0.0656827384939368,0.38914666226417804,0.9754296659872417,0.8469719190532348,
 0.9342017145340223,0.13671292226331955,0.6458132500118828,
 0.32888151517842945,0.9890711255533431,0.9897833460737977,0.7749274552342765,
 0.696684556870865,0.24794576532625845,0.4257193085330554,0.04742209332125191,
 0.15942693809316977,0.8174536196996272,0.2929512794759116,
 0.14721514817704073,0.8441059182794055,0.6562518988213627,0.7556465381503708,
 0.9760365999759183,0.8520403782896081,0.07310090600841657,0.8050811911962209,
 0.3455200704813829,0.1518641782805753,0.7138657342087149,
 0.002390714294341234,0.5702044036561839,0.814170407392302,0.7768476094565373,
 0.576810451487335,0.42759809827532347,0.7751214942775627,0.9538816998733969,
 0.8318856777884522,0.6803907756204902,0.992140695484583,0.5796773073196356,
 0.4888417972115885,0.4033076755411792,0.5069735542041733,0.09187403334961308,
 0.07558396741370244,0.04411944913367882,0.7854920984992679,0.066289300299601,
 0.012428591889199536,0.0037911828598305997,0.4130095989721575,
 0.7841279654134745,0.19292941904491911,0.9609803184873098,0.1331752637148027,
 0.8426788519758478,0.0798550767561903,0.9771587792642586,0.29190668524404406,
 0.240386078840645,0.18274750936190787,0.06489425744347976,
 0.17555613456145935,0.24079788224972043,0.765080004283978,0.3747364349053375,
 0.3642425166064145,0.9812679648970956,0.9912179672598502,
 0.024296456841258807,0.0750297119213248,0.9665957292275441,
 0.8727697780115966,0.8310280994891461,0.8735133009233254,0.5765802680780594,
 0.027731737532451395,0.07894147800396434,0.6152679088993043,
 0.21819717060525745,0.946721838166278,0.833089844490413,0.1339908016271456,
 0.9655060100046101,0.10175536025869136,0.3231562880646339,0.4607184339653534,
 0.6783449329709402,0.4581814116014634,0.9278882776651692,0.155117483293427,
 0.14672153866415638,0.25894754046699564,0.2878418852926975,
 0.30716562438674366,0.5225439536960816,0.6664849015956074,0.7232429493680506,
 0.09361850601615307,0.5509990913398557,0.2825110107364729,0.3911295673381818,
 0.8578875833934316,0.6709236080751416,0.2616978133178698,0.7958739010796895,
 0.13667854142275804,0.8435162225463597,0.5208852767646235,0.9179497043542683,
 0.01975635348305793,0.8549051676237761,0.14623666859817563,
 0.43051031224178793,0.5637750597705259,0.9228267838310172,
 0.33611382191069406,0.34211726469121884,0.020184364577115144,
 0.04288262266734921,0.8829687618968912,0.6070904251399466,0.8201363984496958,
 0.7244717501192538,0.4857635425422312,0.2669248810441345,0.308324392191486,
 0.7909960632782419,0.7459791430159262,0.8359098173957061,0.7348803841567788,
 0.6703029938302909,0.46026885486239033,0.46232034267813416,
 0.45597413164824574,0.30041545412727155,0.6607481911176742,
 0.27390109058302403,0.5669457086998482,0.734777998289553,0.6931189354512458,
 0.1397730219205251,0.8519982034326692,0.38867188649953166,0.9370281380137997,
 0.7879097897027962,0.30839547573670245,0.9967926506214739,
 0.018372310555931803,0.37328602557937474,0.887900205385887,
 0.5227438341356859,0.3499631416994492,0.5567692346260299,0.12246713374083429,
 0.9267169987964905,0.8002084078082237,0.782662812731732,0.1081613454960102,
 0.5181849820097811,0.6927535186144461,0.15889457813906305,0.2283484543002119,
 0.22767374434090648,0.8050426292238019,0.16490327463461507,
 0.7159686526059147,0.7767713616539217,0.2925153460374367,0.0233529342123997,
 0.5102415779622547,0.28456586578546217,0.8544373872980645,0.5189364501711872,
 0.7145368869291523,0.19698357149173273,0.3335614398534106,0.3876682167239439,
 0.7602188817644258,0.7530342223670153,0.9676524193541078,0.3827745975017841,
 0.4465568322498157,0.7577189311851708,0.48164231142624925,0.943403222594887,
 0.992865770795135,0.8781720836895367,0.6306651299214208,0.35623111425485576,
 0.6230782371990473,0.1757593655155696,0.6857032772018291,0.4724549258432055,
 0.7446409885291916,0.9772960746546999,0.029793788748145866,0.934587975569746,
 0.5350716560180016,0.4369801045826036,0.589063896825968,0.8799411632029217,
 0.7269271468494852,0.24127786717919308,0.8468790539044259,
 0.08445250934810544,0.20272739122123273,0.3593047928506745,
 0.2890173062171698,0.5256293161678928,0.24339097019068978,0.7482194899503214,
 0.8748949634914203,0.698765743488214,0.5443712517173587,0.3277923840591017,
 0.025125849617565255,0.17863951542116008,0.5126137027587478,
 0.9770779397215155,0.6056666330485806,0.09924233953772843,0.7058539407073023,
 0.8228256748052297,0.18606654038922932,0.07019789265901677,
 0.2828635347657156,0.9307864880512126,0.9180478314661229,0.6273637950107038,
 0.13526620560669644,0.18973967305126183,0.3319225341578349,
 0.8354695456465944,0.21994878119876293,0.3539618065028636,0.3686886985122465,
 0.9103705855146849,0.651962226615017,0.6976965133639877,0.09439341872272611,
 0.8446597711449538,0.3096771997404244,0.5986121712761163,0.5629435114351796,
 0.5198350032758958,0.21906335550944211,0.5214340578483619,
 0.42470726088874333,0.010397657822980522,0.7467506720868017,
 0.6545657222599456,0.7784734190518632,0.46850500817202545,0.2827020161387015,
 0.15346276110214208,0.7964361022865437,0.951289884391767,0.11528739848919456,
 0.3194441695916589,0.19164818262798888,0.8932804600958744,0.292519700322063,
 0.9723380113582776,0.5189895235688463,0.5552078193900135,0.14814603015592953,
 0.5482251870487694,0.22196931829392796,0.005022028574694581,
 0.19837942096727856,0.32075153593846006,0.02071183543551791,
 0.010416061461033244,0.13187774867923285,0.16237240116610296,
 0.1744139405078018,0.2550816182217499,0.4638567603906322,0.5030143605217401,
 0.5497010635372137,0.23357702561985294,0.4740560479177601,0.8697727059819141,
 0.2761143849920229,0.4589012469502608,0.7864379526782168,0.6220461407305374,
 0.17817623841380092,0.22971235648313781,0.9848610233300074,
 0.35448303430809314,0.4582333998590765,0.3421142269036632,0.8602405130057232,
 0.04707911896367478,0.07008625683396708,0.11538822200459398,
 0.6252660551008594,0.09971488295091724,0.9595745454970537,0.783503646079622,
 0.7576075468510723,0.8552731556470358,0.15055854897294352,0.5196417519677261,
 0.5525854657330361,0.05649787850241783,0.6667153936967797,0.7705510618262544,
 0.5673025132218833,0.9118781436957273,0.0305192567328767,
 0.019357683807695647,0.885590946777003,0.1265671791333337,
 0.24179238414987836,0.8365680852438346,0.49454765341462625,
 0.47381494564095605,0.191005139939746,0.582978717959124,
 0.0040500539596795004,0.39066647013201206,0.22850258645762656,
 0.19814539776504858,0.8728909963060352,0.8167790494660503,
 0.16771366784351927,0.480238320386029,0.21763507629485035,
 0.36801373182289565,0.8834731277169497,0.7173529409030158,0.7616179350258486,
 0.7985478067955726,0.4363645761078453,0.8098850820033052,0.715906974522959,
 0.01139243794308653,0.585468849043541,0.6473604296374177,0.3862947381663221,
 0.5413578741665566,0.842238866516466,0.4299979273600454,0.010162296918971525,
 0.8483783840130159,0.05792845422636239,0.10875548929795165,
 0.3443830709557234,0.7880826888507702,0.00906177944158526,0.3665680717597133,
 0.22088553986365123,0.6740597006568214,0.6120223616336868,
 0.07578702532500214,0.19651842224905314,0.11004870729833105,
 0.46843023872124034,0.41293551752369284,0.6215511319806741,
 0.6436017221402633,0.2698443445652515,0.5953810490243826,0.5866207650474539,
 0.2441586661591486,0.9350574141128217,0.719274243520158,0.6318557973536754,
 0.14473490889214813,0.5837868503732321,0.7841033071786714,0.724601140868976,
 0.5517314863772316,0.6337014716841686,0.8054054224435291,0.9712686544366008,
 0.9505512404430195,0.12657170713536625,0.4412834134814104,
 0.08462924473778088,0.3977627490196728,0.33001548597231045,
 0.32854513853028466,0.44524546651631725,0.5102918042619998,
 0.9620610449256396,0.3370039215019248,0.4401772873705738,0.6033640321788326,
 0.007291252907528412,0.5784629420973311,0.5287390482181346,
 0.0430437736802024,0.2881651855285424,0.1558432352421848,0.24772289603949837,
 0.3211809596639088,0.4743236566324407,0.9085996994497086,0.2749100900872026,
 0.9649962873971794,0.13812631862431646,0.8392419679614193,0.4703892941748322,
 0.026351733639102592,0.7727084762601633,0.24618026434691853,
 0.8446325658463826,0.170373606489298,0.44413753229587916,0.6889495417828809,
 0.7516334569312308,0.23313087695188062,0.9918569994467505,0.8479760161086826,
 0.8128985201411533,0.03610962671641493,0.7705503385168961,
 0.011028228738498957,0.03585204816930565,0.5923291894444507,
 0.6219665326440784,0.2888527991996189,0.017358425418812606,0.577394082905775,
 0.6963346620091728,0.39269223237193707,0.9463624300958043,
 0.19218109149740104,0.6729794274025737,0.6387203221952804,0.4867592824892951,
 0.9865277653452591,0.25537046652210593,0.4608448282221853,
 0.13422628727044106,0.6469624412713582,0.07735103737371696,
 0.5881733262310014,0.6051485062738738,0.5487414853237212,0.05504311703648268,
 0.6010736404207428,0.25898816682164016,0.30995593040549274,
 0.46126222178957643,0.8879718844850567,0.8019951677686926,
 0.08691987795283684,0.9456994045443973,0.7832180396616022,0.9446273835803066,
 0.04514085462639694,0.9008567805720771,0.3925719557341292,
 0.05832582852820367,0.5518525867112791,0.05842753938720324,
 0.32583465931894684,0.38508508120780593,0.848202148723514,
 0.17334289606591913,0.7119520907550315,0.3625226238711815,
 0.48887811104415435,0.16449199695811156,0.4431800560844592,
 0.8928888709798217,0.6161395308575603,0.35957497881837464,
 0.023997230161831973,0.047454825151603486,0.28104202997429384,
 0.2646355286228398,0.6408243407115815,0.20439546981582524,0.5115762644640737,
 0.6832778208880049,0.6349115142019285,0.8628233881588937,0.2412312452715023,
 0.8741510581327039,0.6037304740175838,0.6670955193893409,0.6286057647569586,
 0.024303599530345554,0.9550601049360625,0.0292055212991279,
 0.9397608636394525,0.7259327716033335,0.6688730202688403,0.4402288024458506,
 0.9872298842634412,0.7859218626020503,0.9988696249053549,
 0.004492065924839306,0.45138237964331707,0.8103905178461298,
 0.3565285704909794,0.8839930580561536,0.20375391861476744,0.4022832593508455,
 0.5706919417603863,0.7402934841450302,0.18099925676662953,
 0.03843457577928355,0.1483763686338453,0.8976265688165666,
 0.19546530104241322,0.7029572501568612,0.9922493716125879,0.8133581726069913,
 0.10148967697792177,0.2266092296740767,0.23611143979448979,
 0.5514061828441885,0.2790131400309157,0.9732981071495208,0.8777308776904627,
 0.5163420707388328,0.45444240711224926,0.23627054819359306,
 0.2982945474976745,0.2942042130725355,0.17834232907121472,0.6452322560084435,
 0.27396921792889706,0.33890037549079643,0.5227776121117889,
 0.9180277430829652,0.9864282426209753,0.06147395367991004,0.5603254300572519,
 0.576419774527082,0.28811282704682084,0.3746094555489915,0.6365104885794395,
 0.7450333742750476,0.5721856200656867,0.2649749329246882,0.818997315789183,
 0.29128495757927675,0.5856040997390323,0.09341728972984009,
 0.5820386556993417,0.4174246890211687,0.9889698830629738,0.6988351806960033,
 0.8811704118510337,0.4994728345026266,0.5604891281874858,0.5394479776322993,
 0.8607953505185393,0.5676238902086814,0.02708239493737108,0.5907434142995609,
 0.7479896110730969,0.7951638435417846,0.8023242006901593,0.7818525068277322,
 0.5125261761645971,0.4917148580339117,0.25442012792801416,0.577639839901959,
 0.31732033898268575,0.9916599340212204,0.7214317688382584,0.8706063240607644,
 0.5048307131105986,0.616103496032596,0.6212445225075707,0.8865346436639455,
 0.32712081556875494,0.4700343519034472,0.8668220223111375,
 0.44852507062403013,0.5783480467325915,0.1021833748865969,
 0.24410142858861994,0.6984359142974736,0.1166751891877531,0.5368122123635635,
 0.539125100983312,0.5509172351993591,0.6870633578849549,0.43609949918179514,
 0.16229514002744494,0.2045901680768254,0.8848343290932147,
 0.26222604931060056,0.058865833371329224,0.29685251082411057,
 0.22430029790625428,0.15351269431002001,0.97998042129384,0.8219967196290238,
 0.27859056567369456,0.710834775551642,0.38376684508380243,0.2938424202028235,
 0.6577194209970788,0.5735553729431244,0.7483516918168244,0.766463234471706,
 0.8078895945255571,0.05307370694494795,0.024534951143456496,
 0.33822276025775977,0.09536412678515171,0.6337237581747124,
 0.7736751349155617,0.2668947778172749,0.31237279109759664,0.721506074072376,
 0.4900261776974063,0.5808781219450448,0.6723743200657678,0.5391216258119204,
 0.9586801069438569,0.1672373544439184,0.8262772792328347,0.24717085614544743,
 0.028961517525824765,0.8321628052748418,0.3873842139997745,
 0.12972393769126345,0.2674732622809747,0.6287511479265001,0.9280806726471598,
 0.6459698584230331,0.240316478670748,0.8376357022207771,0.21463482796226385,
 0.5717277401068249,0.17444859806091184,0.453677828943163,0.05232867659285567,
 0.1398527227525369,0.7712190120994961,0.10946765235895795,0.6004136712591291,
 0.8658117281693934,0.7984575961658616,0.7950220239791628,0.4237332685570061,
 0.0712268001695634,0.10106899336803998,0.06462585887690953,
 0.28868911169029443,0.22645611076724237,0.47765620533269637,
 0.19537245334240283,0.4860248775902619,0.14968659996263978,
 0.23221684782892715,0.4609582982617213,0.4836953162519989,0.6098410106416472,
 0.8569031557490501,0.6921343200438159,0.7488464333712763,0.35694195692467356,
 0.6080594970170985,0.8246664043705819,0.20322067559178758,0.3650944986428578,
 0.20477917433676485,0.18691305537132075,0.8947209033027654,
 0.06900742005745997,0.42865599132802634,0.6259415709824278,
 0.8293198431482117,0.7778970291082131,0.19686600376766972,
 0.26877515229369875,0.10760337658756725,0.1862504384731809,
 0.10375529258014549,0.93200835789792,0.24547747355433636,0.7735856970117133,
 0.9502118708512649,0.6921050898759886,0.23502031570779702,
 0.38396052008555537,0.5408237198887192,0.5107979377320605,
 0.21915629300631734,0.3899157418432623,0.0059659982457457295,
 0.31084317627897307,0.7727942067760727,0.9928432905058298,
 0.026028701570936286,0.5238010146728661,0.39578166945082993,
 0.6627162883376341,0.8833186526529279,0.29972997552926917,0.6206173007806262,
 0.4908130842726376,0.8072284027914092,0.7811944001065422,0.9686882839185263,
 0.7259884261987329,0.6353761765803727,0.7725827760040197,0.35446283758780894,
 0.8885556130476687,0.4589105164475029,0.2991683016157304,0.38836543118433176,
 0.7829642777142714,0.9920148856248022,0.2750522802265246,0.8336494345311001,
 0.040470123265397784,0.7991124264452647,0.13699432728114103,
 0.1372280660652766,0.6527493300240519,0.6486144786281721,0.9291692931556405,
 0.6970799519213821,0.6705800563195969,0.9846860607958874,0.6709873615968432,
 0.5345788460169943,0.2267919045243062,0.5494269041561236,0.5491406280954281,
 0.5270128536198456,0.24459322441796072,0.022597648952617178,
 0.8910685955002108,0.14581040899828135,0.739761224457956,
 0.015312286407286657,0.7818643045040617,0.2950008751364782,
 0.6593130518863141,0.005948612900111061,0.8793235667892783,
 0.5980873679326623,0.384330753486142,0.24762909284402956,0.538666171236155,
 0.14845045560956227,0.6584684023778797,0.3045076506717309,0.9131377553642548,
 0.41694436269575286,0.03576983980532833,0.8361888762232274,0.675350654879139,
 0.8788830045725725,0.1481426833400934,0.8240341219518548,0.6097382189556187,
 0.03836325166681309,0.3750445773087212,0.8950894545482986,
 0.047087914175043655,0.45224654297474665,0.3115775101330953,
 0.6406537335601676,0.8842614720143654,0.1323202058189804,0.3786381150190474,
 0.004435367146669589,0.6478948590578558,0.7223221176242006,
 0.7347080476408491,0.5988229850984984,0.39263076595059254,0.8709767141745909,
 0.060605511593624706,0.2260173386923482,0.584202483787916,0.599265349582688,
 0.29948668256764643,0.27665870910902324,0.5826011211928221,
 0.8099796497584918,0.7868265657327517,0.32682316151709256,0.9562848613785421,
 0.44140996463251014,0.9491556272223817,0.47644049041724534,
 0.9708594906365042,0.4700747910933468,0.24937401688036354,0.3526212319434241,
 0.4597853967666248,0.9451371602882416,0.18572181793924536,0.6872994425400045,
 0.6334339015713571,0.21281909393614207,7.886994027626049e-4,
 0.44218126869987306,0.07093383778407614,0.11657456013012957,
 0.017448952903530657,0.20015095547621153,0.9836326723847031,
 0.9505718815142625,0.5426282218372688,0.7790365434785342,0.5556313918789675,
 0.8255022805897128,0.40746175098795834,0.0025993600886143886,
 0.04363989087199038,0.8060837389577956,0.06834160516545473,
 0.2569262827176081,0.32399711663293673,0.5449077121019787,0.3680413821111257,
 0.729223370371618,0.8762879779846475,0.5099553972960815,0.872697500566896,
 0.6339260254084118,0.4417970300076013,0.22086855735242517,0.4459884559176778,
 0.32293110072142006,0.883586461284289,0.9209525275716633,0.43878643431241304,
 0.5847925318553975,0.19730359692492228,0.05989766382885131,
 0.8971106088730858,0.39973114042303826,0.21587977057969154,
 0.5964875622742305,0.18890099909097413,0.6719577335100269,
 0.20295049414048028,0.425998888936145,0.8402009011408066,0.495438505798105,
 0.8822372784671156,0.15924466987062869,0.8839969886293214,0.5735715900488713,
 0.3531900470674191,0.011303117694434706,0.14729321389612382,
 0.9886355703331026,0.5487877136620153,0.6195310136882893,0.14359565140465913,
 0.9037469539511627,0.502736919923645,0.9961983337225642,0.17852839488983174,
 0.6916255639236835,0.19824464791363072,0.2199835460496945,
 0.40289437438551534,0.7840495457829078,0.22371388543712645,
 0.5797155094683712,0.8826123380270526,0.3738272684281734,0.9915244281568357,
 0.8238709650040987,0.4839902052485682,0.5484191170452852,0.4871244593734131,
 0.6156107015385541,0.31101108423399815,0.2828899582705471,0.8322550880130976,
 0.2356013389087277,0.0373758495422426,0.7057661063918323,0.6175961844646973,
 0.5800874802124654,0.08306801228548533,0.7497763483235301,0.6267597404346974,
 0.8525027076587459,0.030338367484762374,0.38419678696570947,
 0.015029440158018326,0.9386841555108881,0.02729756457583743,
 0.07002043211367448,0.2384865681597964,0.6094193952054797,0.9668574374467058,
 0.639872688106438,0.5604446024932214,0.18107922609370064,0.20631615106839385,
 0.2673377079759959,0.6545106122441565,0.13073567333617586,
 0.42650696976527924,0.5834791850048332,0.16810006685756074,
 0.40194229819346505,0.43692119422657727,0.8142942295441418,
 0.6115251900722141,0.08603515599599132,0.46252036825354326,
 0.0363402260332325,0.8362758123661632,0.9353765095050735,0.8146635460971897,
 0.8904291295180972,0.6217767622003535,0.6881114403099085,0.3296563910518855,
 0.4063322264839122,0.5126633364783797,0.24041060459470043,0.8785180294225894,
 0.7659189291139432,0.5448747445938481,0.06510654771820246,0.649828114425275,
 0.9491512482665091,0.7742789318792077,0.8470321778083595,
 0.042114630442653955,0.931579278058013,0.3749385327093554,0.5441702657672793,
 0.2302006756098487,0.7111987711707146,0.16923120974430061,
 0.28180196222234566,0.4463914405236221,0.42827777148325186,0.842727915548187,
 0.8287464651208378,0.2520546508959055,0.7262647916962349,0.8430609937320285,
 0.6025684961133609,0.4044499385001271,0.8345497895900958,0.4717120122085716,
 0.6167041322723134,0.33372446966873126,0.057608540265103536,
 0.5476081009539413,0.36407699931594695,0.5605124759501059,
 0.09870008651450479,0.8356079325905125,0.2690435951254201,0.3710623781823136,
 0.7766071360706767,0.5177418777970996,0.6554718301183736,0.13050424061944754,
 0.15617281937077643,0.25760669086054555,0.7104593252714121,
 0.5372586841580463,0.36634972848819325,0.11264103232129408,
 0.19136539678617148,0.6840187678452038,0.05981378894252232,
 0.3270536224543573,0.5397641177693938,0.38659846367226514,0.6937543203229221,
 0.08414118718087094,0.47466218051893283,0.15860667554895358,
 0.34947074483048657,0.5644224874256601,0.3479382251690679,
 0.41701922227243227,0.24366338376918406,0.42514053691458764,
 0.39208922519551415,0.8062776822570086,0.4094606841774412,0.3934473373221632,
 0.1182152255538722,0.16268638721976225,0.7360904944123152,0.5222834575409286,
 0.08563747447998882,0.03298196324395297,0.1345336555844322,
 0.8902673261529689,0.7427889261577665,0.40998227063937354,
 0.021066114576816286,0.8164442654152204,0.3589090084361587,0.992118181707668,
 0.05300265368966972,0.19418600917737538,0.2914837949206941,
 0.7017048800509725,0.6334859005336451,0.6112609195259331,0.7925205295930332,
 0.16111193648451405,0.11701999868196167,0.32772238693062106,
 0.8383087820308187,0.3282523551878641,0.29359584117566484,
 0.10270602119228867,0.19065110510878447,0.464310905141194,0.6057288456884173,
 0.7391310134662744,0.7669255657071702,0.7761385128686726,0.8146667559705508,
 0.7517649715265667,0.7796553887743459,0.5605151409488274,0.7688488539535864,
 0.32438259380311185,0.6961282462861151,0.7775185905126906,0.8246138064273477,
 0.8139269187822757,0.8766156786800705,0.9590209785760043,0.8565541802786752,
 0.13184212668441397,0.7968495918555878,0.8704969194493334,
 0.07764144247025462,0.01367647724915988,0.817246901808017,0.1585349715846024,
 0.1775645275413319,0.2982157840705463,0.41798271895341266,0.6471622423042221,
 0.8762975849444666,0.8751493644359807,0.36010123783942993,
 0.035137746952262905,0.04354775165150937,0.6434275115446072,
 0.05179396203331921,0.007330277129144691,0.123872111186078,
 0.26162009530713126,0.6702383114143792,0.5392498373266472,
 0.22137334930172003,0.507642857606274,0.13360644409276434,0.9379797179593599,
 0.39650397226040957,0.04203992223230535,0.8679257799312076,
 0.1697178266665419,0.30303674141821324,0.2351037957097798,0.8473863847205282,
 0.6203448669745517,0.41571043903702254,0.10256365639479625,
 0.6535885324224583,0.18557985722785098,0.18992598788035775,
 0.6207321548704178,0.07324781833998262,0.2363301131735075,0.6285766365331547,
 0.9354215583859093,0.7984214066586577,0.8870576167974293,0.5186483093540296,
 0.44280844058119473,0.5387304590940181,0.1500638736199188,0.8308137726589926,
 0.7409036802198512,0.6855146649360306,0.5613449240445482,0.7216897701722228,
 0.2803646899111595,0.14877910808447126,0.774232703106702,0.5939596246986933,
 0.2098459527375297,0.5074165336939755,0.8184408262849566,0.3266060377183906,
 0.4487157531580471,0.1993947903472303,0.5907990307798562,0.35234445887556676,
 0.9982565402140563,0.8770885890415515,0.5651820542925522,0.8821368462400295,
 0.9584085526896388,0.7653339794935295,0.8131985709419282,
 0.002991745450983352,0.835386109630707,0.19813529277226583,
 0.20179624924483697,0.37113374441068936,0.10790087598279197,
 0.8691177817518634,0.6833581203319189,0.6457672424343814,0.9445612773259328,
 0.34060131896577295,0.2902621746890591,0.13997578090937202,
 0.9981097385324401,0.25242357822776185,0.1716135018086843,0.542738617716814,
 0.9808161475880556,0.9521866877754014,0.3224785077808394,
 0.029412065324509573,0.35599517902853717,0.9066231914745193,
 0.5233556062310876,0.9397467291135928,0.20258359164271478,
 0.32255776489301957,0.5449120265300804,0.6539030147825723,0.9406279199392226,
 0.8171070736829905,0.19956455821086072,0.9952260814346982,0.9203729463172651,
 0.7167748784443109,0.4980742704813048,0.07427169595906236,0.4375707552140988,
 0.4482185694497217,0.06373818793767294,0.6742805673322696,
 0.20257983470820806,0.7489393973414372,0.8689315233739788,
 0.044954930512425384,0.7222751971710162,0.5418436872874853,
 0.6936319672045431,0.43647423808670494,0.5563532294025919,
 0.034745168569246365,0.9276299249124048,0.464400247558612,
 0.32651108061540357,0.30704340560403254,0.035070484067282415,
 0.2366958836808507,0.04259293609008352,0.25440449476132254,
 0.7717468379612703,0.9226163531835048,0.3247795632739481,0.7764835944537842,
 0.13292423092284866,0.8834770638628779,0.16501643475387073,
 0.35490583663956354,0.9368971727031008,0.7379710086233393,
 0.24402355718645807,0.27806067748709007,0.21239270077725791,
 0.7158455458249128,0.32416973733585586,0.9534575099521061,
 0.12892108512348555,0.06454015728700035,0.6208153680449693,
 0.30143160626577825,0.3747637879479151,0.9493445645858938,0.8407359437949207,
 0.813258832582506,0.23724587150696397,0.3545746278448543,0.9180337289708277,
 0.22785512559138732,0.43393796507154203,0.5531750287673284,0.580988014247722,
 0.7571349427980478,0.5340520648951486,5.219655160404457e-5,
 0.9432492236339705,0.5028020824951623,0.8249997793965124,0.11688915424840163,
 0.6475386332864753,0.019653672792099952,0.6806957579503957,
 0.18413690422531492,0.9631789361577229,0.4496435125996897,
 0.17557522356326705,0.7115805015730093,0.5283748401538341,0.6961077124868374,
 0.4795965726368262,0.11697013354048746,0.21691271093068631,
 0.18227405386106743,0.32436234283167253,0.6769773093252598,
 0.9687506928622664,0.44927377588158546,0.8627884568645064,0.6267004038143525,
 0.12428081721970552,0.7478293755861181,0.5729735598164173,0.2558136563326999,
 0.7489062010300092,0.07317033644085857,0.9833246214168411,
 0.32481677891833327,0.7241774838165507,0.777316683954034,0.9674010059952421,
 0.00249850900082349,0.5278462613501345,0.4898518250830003,0.6194911696800436,
 0.7867090538748591,0.7401140756127563,0.29612423805331656,
 0.31481954457469696,0.32188353865596897,0.7318083902903642,
 0.06521724438748444,0.6851122496486183,0.5233258920504422,
 0.06421658168965216,0.5404757740945022,0.6628390093242285,
 0.022922748284760708,0.1889082894118277,0.43685918557822645,
 0.5980369442019694,0.2671501496469293,0.8641114888269998,0.2335858008496723,
 0.6459092995570689,0.14297208287016128,0.024737921833985155,
 0.4806650479347119,0.6495109864419789,0.9531375439386013,0.17951919571231312,
 0.9796901765277909,0.8321685471605894,0.16663155550265618,
 0.042685880685648714,0.901598460152383,0.9453418782689,0.8754365018991964,
 0.2107591041695125,0.620953489214737,0.8257737747569734,0.8889893273365718,
 0.9497562537678695,0.8413112030613705,0.3088304256629397,0.48286650170569034,
 0.8287273688486819,0.4227167121737181,0.7957118049998733,0.30619747173841216,
 0.079677475182925,0.4807674500600594,0.8423360953556058,0.4680783202078139,
 0.8323109319033435,0.8061471814616147,0.5264296327818914,0.04992112406272986,
 0.10961427178843142,0.28228584854795435,0.22434853776771868,
 0.06557511186448706,0.019383209146699443,0.8668093063278567,
 0.3462762353235609,0.06333157810338008,0.9120922569763645,0.1883177783665918,
 0.5187502097506487,0.7909811438170138,0.5469250799188813,0.6649077077196266,
 0.8270746935696807,0.6675411941350838,0.7069244628572023,0.6926773813223804,
 0.6013244424167367,0.43437959763528233,0.2493875218111009,0.9488138745804244,
 0.14273297443928523,0.9333319655281384,0.8442585101744249,
 0.07358382121818752,0.26304492768994436,0.9860474465561067,
 0.3053990547374944,0.589949788719629,0.8596848193432242,0.8047177246632404,
 0.8732886326494559,0.29839548445658104,0.18021697279908377,
 0.6512582840020167,0.13159615686554493,0.14492485306795277,
 0.03347620401667384,0.38568316158628724,0.4193409931292992,
 0.23263159919735132,0.9010806627654798,0.4081683645049221,
 0.016644701187203736,0.8708774017518995,0.8230117260924236,
 0.33666637254428233,0.8521366570529688,0.6768330213628602,0.7717435626924544,
 0.3176286358240591,0.37475215985860055,0.6291105186347468,0.6520200333009342,
 0.20591217903195358,0.6483876764922679,0.12117027735423591,
 0.5725471696373435,0.903898321740839,0.9054213164145715,0.18394366431363962,
 0.3528630706429704,0.43476086057484153,0.9411821640655023,0.8245311358836908,
 0.8129278731724504,0.8346804053927033,0.11783187311402366,0.9401021476155345,
 0.04669342602574478,0.914320887259241,0.7182397917428847,0.7668429577076921,
 0.014219592389046132,0.6675628755970039,0.6123705791836742,
 0.8475754523578034,0.01164377898839497,0.40069595329619057,
 0.37933792598834004,0.5908612466826342,0.9310443660289273,0.5033322002838795,
 0.6517084480557208,0.5599698514691507,0.17944935239395665,0.6040597891763078,
 0.4484890236654555,0.43832464474591293,0.4145164240841034,
 0.20706925174726032,0.3919488174546314,0.7862208645025918,0.3473530822634977,
 0.8037843549454241,0.5995351664653311,0.440610271075101,0.41533231762098655,
 0.9582840368343184,0.5964198811205487,0.16830694886684827,
 0.43343586145729607,0.5388735379798081,0.6868165865725855,
 0.029614528980346666,0.650180709850133,0.1876619352668224,
 0.38709928925448767,0.2174366250349441,0.418932733318627,0.8619296302236744,
 0.038464695214806,0.008277506340471197,0.22284634080357535,
 0.3184901406996469,0.31804484015389,0.328031091522199,0.44711341954564,
 0.8045943004885786,0.06084875813794022,0.42285587089311094,
 0.2881146647921974,0.8536349018659947,0.604722014363218,0.6244521126318332,
 0.0844235614231239,0.21670418559997728,0.27545676070301184,0.404379660884157,
 0.7645814124548815,0.9757325145354684,0.7853689799614819,0.43670166542058886,
 0.7393479556068645,0.06991707820044701,0.6277469246420392,0.8448343679319988,
 0.34001903300012815,0.7284972108976514,0.30449332135224105,
 0.26385384841833837,0.017406428511339,0.6148633188086086,0.6906683120768823,
 0.16773315357525242,0.8853493443864598,0.1504812142987546,0.5636254246588563,
 0.7063482576746227,0.5710801584888962,0.7043606522110064,0.8801219153226499,
 0.5273913754852614,0.4733975840465934,0.19330162533082296,0.9275518334391748,
 0.5395378094197161,0.5384444871779437,0.7498289165229628,0.6002755908822481,
 0.2837858307764165,0.3001020082991548,0.7473174414322064,0.33631491395209023,
 0.8703210930591565,0.6579126907988925,0.9287850244497222,0.831899148011496,
 0.785301253334942,0.04119008495062326,0.520519476944414,0.7027798695020823,
 0.2066877586171717,0.13219045040807242,0.9578133029025179,
 0.20648445935738258,0.8498995338979565,0.19362052747137837,
 0.6511953277779052,0.6438274976075321,0.8493551280149986,0.4886959099268191,
 0.3464261711743917,0.5475506637368774,0.45785810339279054,0.9820703726863206,
 0.03639461617869433,0.34927511848140746,0.23077641576980867,
 0.26465825059475057,0.6686120816966756,0.7278237922936144,
 0.49191331685775697,0.547765540919971,0.6353852182239519,0.6299541146452879,
 0.9225010200288943,0.22501929376468743,0.37188855707015744,
 0.6129124399450407,0.15067084400100816,0.8656597083788693,0.7484596994102595,
 0.40624318738782494,0.5033004860422975,0.4787195920774433,
 0.14966939334216456,0.4498373367761632,0.5714215853774602,0.9023060678590888,
 0.05812857727072418,0.36631634605809194,0.6059916290409056,
 0.8648522560377251,0.9421261849617818,0.5813077252174887,0.5414306465789847,
 0.25663937501116696,0.03093017897451933,0.21979719739930614,
 0.9885133597981043,0.4104609868285771,0.21513670001190766,0.2611429139248995,
 0.632043465220436,0.5462934082998523,0.05589881607241809,0.45480332296284554,
 0.8061109143977199,0.1782463167732532,0.24680196207297744,0.2876355651100223,
 0.7149529404405288,0.0527139712751199,0.12499878116094554,0.1538033430591328,
 0.9296269811748601,0.20675576750252778,0.8484725260856919,
 0.08746655669795567,0.467914005960937,0.5553951358783203,0.3093828821338853,
 0.2350633117858415,0.3551187000530627,0.8446957466485151,0.4624846048745337,
 0.4905341985307604,0.727034854362886,0.01779617810875711,0.8280710165168077,
 0.6595460758329659,0.22462752843431255,0.9941080891557592,0.4790853292812801,
 0.1766892519569797,0.4628552393943528,0.6148723260192289,0.5055846045260478,
 0.6073428308993862,0.9677506879593395,0.37959985305816957,0.5489632363581083,
 0.4906578725681888,0.16734156846406456,0.9298743867178153,0.9999215575271612,
 0.3874978358745751,0.192320168760167,0.6262294615936344,0.9530666573619921,
 0.31907068654701376,0.8099404748951099,0.34534115642240537,
 0.6512145056212871,0.12426941052344931,0.43986947612940064,
 0.10369356210266745,0.7132781640885575,0.13627660799912167,
 0.5102078402795494,0.7143688654817172,0.24453478665694062,0.3949121975623361,
 0.08174232944612647,0.7125298382758122,0.3731593513841901,0.7993987141962978,
 0.9696557018720497,0.2021943698111579,0.11258272642317224,0.8989023080651937,
 0.9546163428027691,0.3735993319685178,0.7608697735346887,0.16838053297139133,
 0.30868538229710696,0.18505634449715502,0.3884662942866842,
 0.5762420404683977,0.28938096772817046,0.8623650559897132,
 0.023540048094281896,0.8326136261125793,0.7676368760746297,
 0.7214650991032099,0.3883783782948751,0.7315965401978792,0.6853956704706546,
 0.2334397645048173,0.9779603796129464,0.5355804057468818,0.34165639659914604,
 0.8234320131038826,0.6945258219961941,0.697623806597617,0.7060217942871732,
 0.6054985743282698,0.6260920533352753,0.956818416504793,0.5473655905362657,
 0.33908645213832467,0.12450779310648952,0.9997664535260956,
 0.8056511838857271,0.2204195716865519,0.4943104824734422,0.712481046494277,
 0.08363044977865586,0.6277722788139993,0.7102781093240959,0.5418621055354009,
 0.7329746184311874,0.4347668680064549,0.6600341899212948,0.7638196487107634,
 0.24014795681661016,0.5383072767570776,0.9600310431914105,
 0.25222898693491125,0.6854809021598888,0.2898718310104955,0.8774397321644294,
 0.84248300601181,0.3197579586178595,0.6478191214428458,0.37138704800745226,
 0.9089227629321902,0.29406101815825325,0.8288925715594603,
 0.16186347602378282,0.1521186880146912,0.782246133812111,0.8792830865374688,
 0.4462156208917837,0.7033357894060593,0.6626258962273969,0.9245941419284482,
 0.043869839387536524,0.09551455261911179,0.7578045111003062,
 0.6924861346029505,0.31716890086673155,0.7186533051168326,
 0.34420369624716607,0.07319802596971448,0.2004013297789532,0.777000046668661,
 0.11571302785245108,0.21055431420687565,0.1017724483289063,
 0.20615528156212637,0.285005922031616,0.7590451838107691,0.9357855063783284,
 0.742563810556868,0.8922480195301612,0.24659461373232883,0.607149032979303,
 0.10924024927878173,0.7714083334080737,0.9067002200124862,0.9768232716325619,
 0.9197610839797881,0.37813014526085476,0.052030696709789104,
 0.929711041635641,0.208463261784541,0.6423498439921211,0.03939844302929274,
 0.6711990128448928,0.5071617519778041,0.9535149349990111,0.8288517833731736,
 0.10367820668010908,0.29692612078836633,0.25333312949805953,
 0.1759839186410287,0.4525976547116043,0.11028683210565249,0.8847072214299558,
 0.3412849394341716,0.15647793405566113,0.3075204413428194,0.3225774843130943,
 0.35204463196717084,0.33859267113693003,0.6246342831337266,
 0.7298205495925585,0.6604057575799611,0.37717529333281996,0.5656661773508369,
 0.7324650158965998,0.31827688581845415,0.027301641321060166,
 0.011971541325321411,0.09206724330722338,0.3017467994446117,
 0.415509671363982,0.9942911725580679,0.5544257742121211,0.7840473708483369,
 0.6164460396022426,0.4262194435143163,0.5979497658304674,0.19785168464832026,
 0.052659195775838996,0.36542856750913577,0.40187579602500634,
 0.7642577280961407,0.6962756003955891,0.7061578591220129,0.03581142524404479,
 0.03828897428420375,0.8822778958880971,0.3391700444135147,0.3812156660997619,
 0.7971709827082106,0.45820837695394856,0.4130147326076712,
 0.14440471061045068,0.050771313089883696,0.5261279110176442,
 0.5607894849811957,0.9357907352355381,0.5455348628863244,0.41925404833428925,
 0.6482357017142675,0.1474828049380077,0.9885783848848969,0.763387694883866,
 0.18067403510523194,0.42586684176335143,0.7229241860634982,
 0.5511622376987813,0.23816430593592997,0.2987512871416719,
 0.42451527814549783,0.26370826124416336,0.7589222366143075,
 0.8541022107606628,0.9547798370691369,0.0883858465273909,0.189860165275491,
 0.28794858468943585,0.10276281642124885,0.05630400986144812,
 0.9068152103317306,0.8659356282145383,0.47100735866293386,0.3670471358719114,
 0.29957685265427125,0.8212911963651793,0.6306637728345006,0.2564150122153367,
 0.7652957170080987,0.09192559163130753,0.8027129632153285,0.6161932513038678,
 0.13277500143490917,0.22680471733564045,0.7736553413651343,
 0.7554467735933683,0.7749726132957675,0.0886831604517625,0.5773968819032764,
 0.6455912142288739,0.22126135133889313,0.021315744092814892,
 0.8210210667266705,0.03297576916340017,0.30291029440475503,
 0.9494836679364083,0.7203037832608993,0.411801778564614,0.8124773636411553,
 0.5650699886793927,0.3569420320424095,0.1337907205179515,0.24904020016184203,
 0.3446441020514035,0.06931752943886715,0.020172421021379727,
 0.9947530032495787,0.27400894283479527,0.5069025579650704,0.7041462172819564,
 0.019734961605732848,0.6688227018831562,0.09790487815628024,
 0.9082974247950961,0.8120794182898289,0.06047828576644276,
 0.37528741524039155,0.8365156159256791,0.27023530957868647,
 0.7279029005133095,0.18020874402072173,0.7793937465126625,
 0.36707130989731307,0.06092456878203667,0.9250086002654938,
 0.3339208525990097,0.3816680062357949,0.9064419664558852,0.8433027661831841,
 0.6119805474160256,0.813373106728395,0.10187961468386297,0.8775084533255629,
 0.9026052305940546,0.9203147113558324,0.9366184307298914,0.35183011412569587,
 0.8473571244121272,0.4941448929518053,0.017409543662319393,
 0.0064806958003093,0.532205536148007,0.8674474912844886,0.1874954806477901,
 0.9545944594112132,0.684493848813019,0.822128292750446,0.44478129177586556,
 0.9977103524895613,0.4057356560757136,0.06761413794306859,
 0.13433378732476897,0.698895289630274,0.81911410107372,0.8361394578337539,
 0.6669358276445657,0.18844304651441468,0.08815925198798458,
 0.16353067529808551,0.03345146834201129,0.392689408126627,0.3184566889356062,
 0.8983052877937138,0.4081667182585287,0.8653286995587566,0.8797833782175084,
 0.9100491712025656,0.1305454375884605,0.14471473611322427,0.7623831511896499,
 0.7527500673487296,0.17988431808885996,0.6198300785807809,0.5536867589417752,
 0.054848127987623396,0.1199259170275192,0.809497893636646,0.673056566740114,
 0.037416013967327966,0.31652568653903335,0.9479195682835779,
 0.5316788206837642,0.5652339916035058,0.04659469521294057,
 0.049639912387416274,0.3647294992812773,0.7765540994673352,
 0.9623234956218242,0.1992279403131949,0.9868739264169156,0.04431066240507864,
 0.7617780322194176,0.6902744221099881,0.6617893257809121,0.27883816307216247,
 0.08848047033116013,0.42613448081826677,0.6311776098274009,
 0.05493648615923097,0.18395706719813532,0.649936634716862,0.9082128181733568,
 0.24478213835036522,0.5498040165594351,0.58017542477913,0.49016301227539405,
 0.07219339433991423,0.6206362061669866,0.30012655352078477,
 0.010620793308531273,0.7958630310812773,0.5504054531384439,
 0.4031586286754276,0.38513901935382155,0.2650169814456024,0.5144357972286016,
 0.82266730234527,0.5121720892038448,0.6959870994084715,0.46221466551081714,
 0.0832084756601097,0.3889556450613657,0.22418977445326316,
 0.019098115606556387,0.06264906918005997,0.3827904134902296,
 0.9076701672673836,0.1987501821228027,0.738585540912586,0.9113087427346885,
 0.6275616851352537,0.11453210870924224,0.9627386340695349,0.7405190873292101,
 0.13740105330738872,0.7827835542192375,0.9137107312010792,0.9901975411597197,
 0.40918724301526066,0.5713830707568636,0.7349125862733346,0.7556417152889031,
 0.4408268283557525,0.9683750002536673,0.4776075204159166,0.2766516175660203,
 0.12884976546601423,0.700225254726875,0.8661456619488549,0.8774283462771556,
 0.8587657348321234,0.05276881226249097,0.20111533268906212,
 0.19337018464544842,0.3267174301960536,0.5813603546580087,0.1995646086756635,
 0.16702334407058073,0.14407236685073255,0.6640911292782711,
 0.7829248582703521,0.7250472073987466,0.5795093835536002,0.18103418799732895,
 0.20279403679353591,0.24358810567964073,0.17764831090195432,
 0.15397434009617084,0.26395154347155014,0.26279450402799,0.9285116124871454,
 0.16182066670037087,0.6189222476928912,0.78443831352755,0.8276326357344785,
 0.07899268267331871,0.634995864419211,0.20376287929576953,0.8419472852811369,
 0.043368867521651255,0.7625276500673044,0.12339843745727408,
 0.5034328776410313,0.2097080021405191,0.16343337799720592,0.1618771964685377,
 0.8938536006422627,0.29289366195060007,0.3946299306136134,0.8546093444666137,
 0.23353327965477277,0.9223847699836933,0.3501902124797227,
 0.032891228218153934,0.8638111638996483,0.5706512037312753,
 0.2768619526339866,0.1332875564475169,0.5913287621386147,0.4661615126551658,
 0.3028173815022641,0.5018926749374902,0.09493857395257921,0.5552269148376772,
 0.4049113772555456,0.3573524679131038,0.8651793503350618,0.9166734301228081,
 0.017038088818756347,0.7474074477468186,0.010901044029525186,
 0.994031657278311,0.1827534737365475,0.8252209266532624,0.7981512261473993,
 0.1922907187663495,0.5321244788486177,0.9016531112942839,0.06919165508645275,
 0.9763304940266178,0.7011007125426285,0.4061281182377998,0.6685560375413507,
 0.23367915451138643,0.8682420710963906,0.8444485509149795,0.4916652067284587,
 0.0638081633006915,0.6936251473489303,0.3140754956328231,0.1349538445358972,
 0.7943920985921813,0.9289604787727802,0.3881837824648412,0.27346897329866604,
 0.45386129857333424,0.27911001675988456,0.11129022544958866,
 0.055711731160899114,0.27420039756939785,0.5309370022666787,
 0.885758628928115,0.9472055770130822,0.6536389851732276,0.6822854469670521,
 0.4215845846772125,0.40116853216631587,0.9668574336739046,0.529636163346089,
 0.3816697543120323,0.8191180736113538,0.882648366893477,0.22188002560553288,
 0.3773174241311844,0.2709537218005895,0.37551000220662956,0.5084420231040124,
 0.8995065526332824,0.4996102966616194,0.8230319925772474,0.38923698405471496,
 0.3159568462371992,0.2942547102366203,0.02774553606022212,0.8669745681978126,
 0.48655415744279307,0.6413839154070197,0.8144074540648719,0.3813505598896072,
 0.18171528435869067,0.4489559671458472,0.548489342145569,0.5822955959192467,
 0.1503183105040652,0.4584621557026759,0.011633421642101904,
 0.21276747643073313,0.9578638482276741,0.9152880345102528,0.7118042406946912,
 0.9462070939674985,0.864041637711895,0.13116718786075676,0.7967185533701996,
 0.7563504602230868,0.9977270679120377,0.5780355690458502,0.4322614411192257,
 0.4756762486612117,0.5385751616175198,0.6525672877483284,0.5674558006520523,
 0.12838004406469006,0.16693008372141616,0.30421782064644864,
 0.1389923178628828,0.7209649103037484,0.27200341835967756,0.4369801788842591,
 0.9185000653815146,0.07496558214438642,0.9595724221601062,0.6708419475236841,
 0.8704096461104579,0.13005511518036617,0.366766495723717,0.07124320124924899,
 0.5360741755817192,0.2757191189479036,0.6121609786634301,0.9618341464280165,
 0.8713803941416945,0.9205393601927311,0.04751407036537425,0.7925610889714797,
 0.4456734899011977,0.31333248893985444,0.40647199059130723,
 0.5193339760279407,0.9427982263111442,0.32934016637647057,
 0.28526813242220583,0.9988191658645156,0.2246785674525713,0.6239421109125372,
 0.3951508225835648,0.8798474674566267,0.4680232075522075,0.5331104804012692,
 0.2928623268886624,0.858011596056724,0.4938553281753131,0.5715267057602056,
 0.043066677920437724,0.598556281715805,0.6665865812279462,
 0.29661696039894236,0.8036694690520086,0.9969443621121041,0.2764244831108129,
 0.028412161668862512,0.35766152039389587,0.6493445981140077,
 0.8125870261709539,0.6502990215922521,0.907684800030895,0.5094100904540924,
 0.813844404482361,0.11107112472377756,0.20252379470166082,0.3229829558909977,
 0.8921876140084462,0.17209612745325065,0.7903757492371883,0.6950321656244993,
 0.09331874658513062,0.015102503697002057,0.7472716346249207,
 0.18688378169334052,0.5651390645116402,0.6790110760487038,
 0.48158500198779286,0.7196678946235415,0.016833139978224,0.08456708661771284,
 0.4397496671379483,0.31315813758759325,0.5577871077575132,
 0.18273533325938884,0.32621796464198993,0.18154860255640726,
 0.47396612178057895,0.8977603447396234,0.030295840308698252,
 0.7020031238858601,0.8118630402821565,0.6019856353204598,0.9872460010689171,
 0.2288685178093074,0.7739561947029376,0.9355399889358703,0.5899063044586306,
 0.33039171578895543,0.6546109011788515,0.5407534417133393,0.6855612264409461,
 0.42822726460805105,0.206266886276129,0.2699972926080473,0.708657859297873,
 0.10910688918688494,0.9204258534047342,0.28858216753554633,
 0.4672912936677597,0.7114505184618901,0.8646021111335862,0.8313329447128749,
 0.9232216995139665,0.09779934150483072,0.016198896157575637,
 0.18442452611136007,0.7303593025811306,0.23303278419351892,
 0.5342976656324504,0.9882301602047744,0.24707159583056237,0.8196433545979148,
 0.7916760694087482,0.4134757210053207,0.05544817882124964,
 0.12225324363202328,0.15719023283006306,0.309572778406232,0.4868121224520795,
 0.012753322026183622,0.2027972089665503,0.9225815485036227,
 0.09739432592068231,0.28713602295815055,0.5174731805854215,
 0.7270323042126103,0.4912851342918588,0.8807959710503996,0.9135686330421615,
 0.8251817265932218,0.9732271335796505,0.6775008971950081,0.2145113150871304,
 0.13405966194690788,0.46055259156235495,0.41913241360947595,
 0.4244011082493584,0.4888195508189319,0.7850620009955862,0.0639877127261399,
 0.1824491841411232,0.5867326536912852,0.884514066650792,0.14679637819261215,
 0.9786902248902675,0.7527305870976196,0.6947632872512303,0.3253518665535996,
 0.9527133215083329,0.05998329234654387,0.021229076587082507,
 0.05456978446607896,0.5558864529748272,0.30289023544545757,
 0.21648525551841213,0.2529714850091641,0.323208160792422,0.2827042760977404,
 0.14800045898216396,0.2820423988341678,0.7270923660215792,0.9187393526891465,
 0.613258147492345,0.040739064204107556,0.07292982251109392,
 0.08652613721077984,0.397064202118365,0.07918783165790633,0.5692320782046635,
 0.8415710982628547,0.7900966839355523,0.6888526301785517,0.7432795175116554,
 0.366235081985256,0.9991046725829635,0.656045189199171,0.4057895571497073,
 0.4010611652484133,0.25331999833431806,0.9763379792746719,
 0.055212922055315694,0.02037096890098078,0.7188247758017335,
 0.5853926722961164,0.7564198624124949,0.9049275882727883,0.5778692358185592,
 0.4903454157225938,0.14606817644209047,0.22379636814015058,
 0.6649732192592402,0.8481332799593991,0.9381524029853813,0.3742680463880853,
 0.1920624819492469,0.8228925547724144,0.4159608158977165,0.8487599582859584,
 0.5726857293856445,0.7653123905628112,0.7969711905145525,0.6910221613861282,
 0.8886030168709232,0.3566990170396349,0.5787809959455654,0.28759103496481875,
 0.6227357226747207,0.5871989703945204,0.4236001063478283,0.32156361484481555,
 0.3782519602190507,0.40558675299267744,0.2735844240490921,0.8886946127370365,
 0.19458159343131953,0.3499807148668008,0.6321647237536673,0.5804260454647693,
 0.3935576759074828,0.9639839178412893,0.42998938228859485,0.3002224276240424,
 0.8877983662366548,0.5832654320920287,0.44221163349206993,
 0.19201744600638448,0.6059164452886973,0.3836547018548584,0.8809219938647643,
 0.06252006990430792,0.7234533595738455,0.1386673166477086,0.7469616779716611,
 0.9552464219151926,0.6516999501711508,8.661202293643866e-4,
 0.14746063442968027,0.5708389626071102,0.2742067794995984,0.83056380211766,
 0.5065222882890599,0.6680271928831485,0.2889077687892805,0.8497934511120495,
 0.13121543689543613,0.2540655875301895,0.8490256625693868,0.957517768951219,
 0.40860157604503544,0.31017658312095375,0.7389822924888179,
 0.6065269221748607,0.8175544748465536,0.15978869946209984,0.8444239489384696,
 0.4148999250312905,0.813745302737713,0.20849326534031254,0.372302412540815,
 0.7560923619614212,0.2518342306281379,0.3174325236883405,0.533241037299903,
 0.314546167705835,0.7582693783457137,0.8564083016528614,0.977580336950627,
 0.8970968612723749,0.08385639162635683,0.9302381845179498,
 0.38672274995170397,0.4913521354178927,0.2580097280358733,0.7755434361148267,
 0.10410499864720568,0.24285908517047372,0.6067914559382273,
 0.7448723652799425,0.5257874216431633,0.06976295850222669,
 0.25599198919607735,0.1967940434249441,0.02088328392334593,
 0.09374694964284891,0.6774169536336458,0.029347178741857216,
 0.8030923841244063,0.8317339571488788,0.5601725166046897,0.6039705966912879,
 0.2075930185852467,0.45223171727113076,0.18052044467217743,
 0.2724386895539628,0.8649623014823815,0.8510025488922598,0.45808069667297724,
 0.13094551868962312,0.44473726431757665,0.08565607228664196,
 0.7009892393538021,0.7670064233212,0.09630116321982396,0.8819011046226676,
 0.8628030491990623,0.5567032757823528,0.1698651820914614,0.7757027062544379,
 0.4758432118963566,0.9159879813764967,0.4129796136210325,0.10909369680439684,
 0.3579617868528575,0.021819680804715746,0.5235509624476636,
 0.3896388859329498,0.4765702659960366,0.09685252892321683,0.5037466547404583,
 0.15608230546555213,0.09738296294812454,0.8639960503481025,
 0.5706465373706475,0.05650565095521587,0.5319635081804277,
 0.056332289153578796,0.2537683596682456,0.6152025655119594,
 0.11134540214711142,0.41264099838087087,0.1710939054197239,
 0.7705971015640553,0.26088655802862526,0.5081997415680581,
 0.38731900081074355,0.3553637626017022,0.5186872602511805,0.8338290010250309,
 0.8442941813757766,0.8571049460905851,0.13088797355777992,0.4468891714255865,
 0.6757409051650447,0.16235559567403368,0.4551170549960307,0.5636836462908171,
 0.5963722420961496,0.43993087273883336,0.5058149698545839,
 0.23178693742865875,0.9475820579022063,0.7008508509362382,0.9856908370435534,
 0.4547593206257736,0.1261978050658994,0.3359653296173807,0.5142688502913229,
 0.3524979405119122,0.9366608240000157,0.16979148704317115,
 0.11982325559028006,0.5112861090546517,0.47030792270400923,
 0.6911598971917853,0.5209149634734644,0.5844286618299136,0.10059907791840217,
 0.33889812924880824,0.18586715785872965,0.4755195010000469,
 0.7005348563300944,0.4599049111003086,0.533935471547791,0.45422538857624795,
 0.4130720942992384,0.13294853348964097,0.2714400796897024,0.9091175932165512,
 0.7156230431261968,0.8468224328897263,0.01927073606453633,0.5025858108957866,
 0.9213112491177152,0.5195322584113039,0.8946086560944471,0.6012288163902201,
 0.05817219362295323,0.0485580580483127,0.8860965563515338,0.9927936249519543,
 0.1432107800399416,0.8003623259779775,0.6115433219432245,0.1862045428686251,
 0.987035572113913,0.24588730641661327,0.7266583678310391,0.9206076954738627,
 0.2670613532024876,0.19719050994503706,0.46202984333126695,
 0.4854005305493959,0.2382903324652894,0.4122248548260511,0.3197334753793736,
 0.21138552994497972,0.04981420767203615,0.7558433210191002,
 0.7256438731212782,0.32169688836671884,0.15694522061581284,
 0.8347731935634625,0.038823023819915314,0.4113763334434162,
 0.8200332579543275,0.5418775245762654,0.9487526802157964,0.42378942549130927,
 0.8095614428023142,0.9835872201480338,0.1769754497884124,0.9236094029410405,
 0.9841213916245558,0.10935034146253575,0.5577093138220135,
 0.11705468702903643,0.8277667602237824,0.2830893113594384,
 0.33584687438629657,0.1764848940559991,0.09398638336779441,
 0.4083086346582049,0.96868962794463,0.644187803353173,0.00873981618090458,
 0.7533679114003802,0.4859160660006865,0.310914697996462,0.3834006620985151,
 0.805878056031294,0.5861855894478734,0.3393290333737019,0.5945202192265644,
 0.8058916995482696,0.29035640424378006,0.0523724164056979,
 0.39708540390777813,0.6896783092593077,0.20916136911877747,
 0.4005434978396385,0.48901390918186327,0.9145200916969142,0.6880720621728975,
 0.6986363656740375,0.013205098476284483,0.21994238912908848,
 0.5184673869420584,0.5230067785571726,0.46290282689166007,0.9291393759486226,
 0.6895540526066196,0.39645405404780876,0.03487574281021255,
 0.38758205627177267,0.6525029688789238,0.9298606658983458,0.9591831189348481,
 0.3436873237490776,0.1845964163938253,0.4102551371492602,0.2491174122761486,
 0.8498089993258984,0.09301059869001127,0.92559726593076,0.7742290997760634,
 0.6511799487251191,0.7358860735996838,0.286183954450143,0.21857307142957905,
 0.4008147134523049,0.3592926552340103,0.7968176649293639,0.8223036434247876,
 0.24071868618083825,0.8382835106824703,0.035242140749284,0.5407790349905125,
 0.6261675666785806,0.46507930661648866,0.2829086477734618,
 0.10859822870511004,0.5299814995899286,0.8635004768708903,0.9990921778532703,
 0.4910387747270115,0.6299096185291112,0.9142057866475051,0.910451124689418,
 0.8994240760631282,0.7983618506981225,0.11829547714439792,
 0.22991471902485305,0.7654746596706083,0.9000199637802037,
 0.008223710201529344,0.0013231449462626221,0.6130086206280618,
 0.6501282855441173,0.09314839151761978,0.05877314750578333,
 0.7401260420136443,0.029053047237042717,0.4243752879816399,
 0.2593176314431699,0.7430742077817549,0.2865754880198973,0.42868095512815685,
 0.7738964253532247,0.9245234045560953,0.35099816450910326,0.8672391149157442,
 0.3310996395859669,0.977057374349918,0.9461740764210416,0.2927524027584738,
 0.33874768210357276,0.3838093168209704,0.9284212308515729,0.2943676649784335,
 0.06596654682398251,0.15081656958347145,0.6085387621911981,
 0.5077363678934534,0.992280014885062,0.550231611859937,0.4128146676835389,
 0.7893085753969882,0.9458229461170984,0.7514371921961176,0.581863001961831,
 0.12298682108628722,0.6294575750221927,0.9636913810057135,0.7755828589925621,
 0.6474129562289895,0.3355809210355677,0.03651824281493099,0.9924391350513135,
 0.9088334295427307,0.6655000366605549,0.5192502163997492,
 0.025212844940362467,0.5016919633767598,0.6502533298235994,
 0.05173880392447694,0.01833530045532339,0.6733670456369977,
 0.8410639310028539,0.9994137130507308,0.7711951016156536,0.524580725917786,
 0.7298038342127776,0.1669718834135563,0.5375989003338228,0.1785508883653638,
 0.6510650486867484,0.887374281940627,0.2152397421972454,0.4082589888478779,
 0.0873680871348177,0.23088559448051826,0.026121920350530292,
 0.9674630506966435,0.38874445753122755,0.7691676571983659,
 0.39438849481981264,0.8886612219081129,0.8270598115989432,0.9805962612517558,
 0.401905246774398,0.6986568018514754,0.5198161483701003,0.591748801801063,
 0.6458138444019876,0.30426777629430934,0.9560970032772669,
 0.10078072415538553,0.6505717141233036,0.22324378763599984,
 0.24150129753883653,0.30410436382130734,0.7110242249187385,
 0.8732979310408633,0.18257705995109275,0.4285649736469198,
 0.12689632975350762,0.41985573316327185,0.3139921439279305,
 0.32706111362778567,0.40445583769954085,0.5879143807364631,
 0.33117357013885673,0.9272507515047977,0.1910148643171763,0.5651046199325033,
 0.8998615617694732,0.11567611737748795,0.4407716461675233,0.0732409788575954,
 0.6311595480269778,0.09079628094363945,0.4139607979066754,0.3532673854585274,
 0.30105778460695787,0.3898620399517324,0.6663815563115103,0.0567338495898424,
 0.5148575016428839,0.40321507756219044,0.5892725135662205,0.5023004187189739,
 0.7106014263664153,0.03501112022284947,0.2230653942318147,0.2830925425468648,
 0.1714443842135851,0.9608015102121505,0.7445879589884931,0.8235461675616079,
 0.03795593240623685,0.0841788408345614,0.6033117099127692,0.3383504032894048,
 0.9538975355313315,0.17746854626435526,0.31862883533636444,
 0.40681141346248495,0.02702103633153552,0.22682478050815702,
 0.5463352230183661,0.9007742885086669,0.366094270051048,0.873185367340657,
 0.6425135827787039,0.25156916340200985,0.13857304593295505,
 0.4533622079476586,0.09824936209622026,0.32959362636159273,
 0.9112813935468587,0.8106592143157315,0.42655892490632796,
 0.18271559907331536,0.23475732113695746,0.21369281704066445,
 0.2911015519838269,0.07740672433288154,0.47079171810716425,
 0.7884642601691516,0.010927779175239127,0.984534162244781,
 0.36758086196972517,0.7245423519053984,0.9189858012264025,
 0.022270612335636986,0.9238579893312828,0.7396662102865703,0.768089029773302,
 0.1383502177547592,0.8367864766735482,0.6497426778717308,0.20587531429647765,
 0.24691402009948304,0.2652396626968846,0.6328947780760997,0.8626017636082056,
 0.7301211840724625,0.9812694957492778,0.1217422885519599,0.09378878876111818,
 0.09503338818784046,0.8625735319328717,0.40574542963366256,
 0.7694322890717131,0.5954485845781567,0.9046865336367405,0.16276401812300212,
 0.9044749332388757,0.8248784357690572,0.44723844467069696,
 0.17839996966976857,0.593353241323553,0.6131486036451582,0.5030003012264475,
 0.5911299215350784,0.2374433842410285,0.8183511536603894,0.25055407965249665,
 0.23469612065014012,0.47812838429200166,0.8514209441126468,
 0.9051397989358377,0.831884447403743,0.22750502334631828,0.2908115116254948,
 0.958734479476284,0.2792909204552412,0.3490788348366881,0.2655565331415948,
 0.277260211073467,0.7002857085678391,0.41879897314793535,0.8957289108505514,
 0.9543792541849875,0.6647094957652433,0.8080313817206903,0.31319736856364144,
 0.8133747597109015,0.26158627401758194,0.5839671648166156,0.7142133758456621,
 0.803548518420122,0.37137115202011706,0.393126993188715,0.08688459114029068,
 0.5794099585944186,0.6576545725906129,0.7122498133053532,0.2772007364885499,
 0.44783164039225154,0.403679371896819,0.7847023544379403,0.33854455138542217,
 0.34073927833084194,0.3908942006207453,0.3594540321526012,0.6305515273060226,
 0.1862738974561361,0.13758613325764957,0.8493300083954658,0.9035356185278591,
 0.6706269357677836,0.2822917202516593,0.3324596688113741,0.6851943309595964,
 0.8330352193309982,0.4702992261632031,0.4941331740473033,0.4059187865909477,
 0.34665495347716635,0.4653813779043132,0.04476059265116561,
 0.17988647562537663,0.6659587612510149,0.8924970545330384,0.9884671189855041,
 0.6694753041464698,0.08486499689936555,0.3075085958049325,0.4018941327368034,
 0.9285355537035574,0.3285421808831084,0.26382643463823596,0.9192074954588311,
 0.35702013535454014,0.005686363792404836,0.35830759943411405,
 0.8964446342922264,0.1510988041740473,0.6597953538712911,0.4814253670982145,
 0.7191381912160375,0.5943236970902191,0.3908618919390757,0.4558623427531001,
 0.46577405878240585,0.13951948152525462,0.4337942394810438,
 0.5203851807696035,0.4824039466296245,0.8848321385485828,0.8159762804757009,
 0.9837569277189996,0.7860829697767272,0.09562220591136283,0.8910062285168944,
 0.6805327663518345,0.9192351033258792,0.8850781755546305,0.8152016485448648,
 0.7656509960049888,0.6915695925731453,0.5519639088331363,0.15440781261574266,
 0.453242072546865,0.7955439936226651,0.44531181818969623,0.7320259585785704,
 0.179941538176266,0.3671609170217659,0.04851277750823901,0.07416909127596527,
 0.23386772594120264,0.8597039067515538,0.9357866603456896,
 0.11116988514307236,0.2406732077664383,0.6859764668662953,
 0.46248181881427164,0.9383384854308952,0.10287548931185142,
 0.31078271626922305,0.5570475899916547,0.43377758989933046,
 0.6062453508150742,0.5105516932729876,0.8669324456149752,0.4471997460157269,
 0.03687966286158095,0.914073421522386,0.09025018661637141,0.9225236273671605,
 0.20617000520317807,0.835528841179088,0.35747362221925927,0.9066451623331473,
 0.5523973042147843,0.9547042690275002,0.7022027485362193,0.07256836087848095,
 0.7747861562089061,0.17480466317353704,0.22086662565024195,
 0.011478338683022726,0.3397176385496412,0.8711404100401698,
 0.1386679457851684,0.5773742342586127,0.1992231437843337,0.7683338764489478,
 0.6699689964158976,0.008690001182684703,0.2521406538359148,
 0.7012143776923876,0.6934860954084645,0.4057110721915743,0.9334460933296183,
 0.9585170314838354,0.1758207790326327,0.33120577179699895,
 0.18282167758682255,0.023505879061901824,0.1719674691348697,
 0.8048862909907117,0.13551368298790079,0.23183401259825132,
 0.9277807939025113,0.3811692270511786,0.35665116826034027,0.6949096492510126,
 0.32411669007586863,0.7281197217557908,0.9855126611425498,
 0.33649418836640654,0.07630027168691689,0.5291441166551676,
 0.8359679029914803,0.7524917709576635,0.28657901380743367,0.955528598685776,
 0.6364107610599493,0.2722146455340624,0.8235218596591097,0.6776404832412699,
 0.5790402066241941,0.18011313052011824,0.8823770463923573,0.1451247261524028,
 0.7672478140572482,0.0555258289650894,0.2859884978738513,0.3669477867118327,
 0.9540033970466094,0.09104770033601317,0.17656301995444412,
 0.5490629889969816,0.21882247616517714,0.9483582946518961,
 0.13019168843647144,0.020337707546605133,0.1514924937564519,
 0.11730906394025442,0.17392654414601572,0.14330742186214318,
 0.3652648624623247,0.40011554942303684,0.2529113290299213,0.4434325125766686,
 0.7812073141772907,0.050574443363511135,0.0934872529801748,
 0.8546051298932116,0.8268358043213537,0.14878687417191605,0.803322947266088,
 0.4881790572732718,0.26023892115001246,0.3070294738236701,0.8696718226654738,
 0.02207881851564697,0.7859512947164158,0.23552498345691875,
 0.9270437638859433,0.5973645611610136,0.09321403385709803,0.5908818137580739,
 0.13407743853719556,0.3725705726447571,0.8744666310560252,
 0.28574678931201414,0.9794749598240133,0.5120337244205949,0.4183501416011437,
 0.6657164170346039,0.925224963089225,0.33239006964920903,0.5750266128041326,
 0.5674045689682454,0.06046620791342805,0.20159329431915685,
 0.19950265246671983,0.012194494101630315,0.6104385288104476,
 0.3976577614696758,0.730861890389141,0.6164953833715898,0.6548005445896994,
 0.030233031453138715,0.874449100515755,0.9457277133315936,
 0.26143037693389626,0.22107175764563403,0.6049849809451869,
 0.36932227372695003,0.23335795078386679,0.07717422005875507,
 0.11003211312501493,0.6467573969376028,0.2198964490334263,0.8264146502594699,
 0.32740509552409447,0.8466206802270145,0.19527819542072056,
 0.6788246399096528,0.5004815393664034,0.19383749129223882,0.0571453719134245,
 0.0689524408786214,0.02261902119841608,0.15411993037014637,
 0.48726794974642107,0.3611443918692666,0.8603357678138216,
 0.10489534323550698,0.08981757103993326,0.8272439440883846,
 0.3343979091298511,0.16073204245816863,0.17068446962090889,
 0.08437941748263644,0.6872443009988491,0.3962902262482306,0.5780699825856539,
 0.5286876826184074,0.9472938169420465,0.7575276192514054,0.46280107879069476,
 0.357131970880908,0.5831383198032241,0.031149707976208307,0.6989754767851253,
 0.5862903412371673,0.5033662923023952,0.04378410322357651,
 0.10160948341505349,0.2763716496515426,0.4402249954376598,
 0.04542212513416399,0.9026396352365196,0.20730967527728195,
 0.9593698182695367,0.6843841108804688,0.09046121673013241,0.5358144647449201,
 0.052434433458003915,0.7922820111877606,0.4471844048695437,
 0.5453772651239703,0.7009024543526352,0.2110588292944734,0.27315062118495503,
 0.832704079485386,0.8570165874707082,0.5644720529681548,0.17154573273855578,
 0.040549354315167596,0.45720891715126566,0.9244888626115071,
 0.5152327637226827,0.6695851001141159,0.5788777159712759,0.24408312419510947,
 0.9772973248365606,0.9196744773544214,0.917230413556278,0.04548370653085976,
 0.7682267729129205,0.4138945112695509,0.9856295793433892,0.5322979917119038,
 0.10394943822066782,0.5780382835450342,0.6019877626974537,0.789365469312528,
 0.9127498764814211,0.38255129247400765,0.3553563951922436,0.5708542847381134,
 0.09006365015074302,0.2701749834090268,0.3175705306132215,0.4089235094383574,
 0.5870919650314191,0.10541473599545936,0.7374900500306242,0.6638716479122853,
 0.7504152330147085,0.9805193296968775,0.3369110417499418,0.44944665350220336,
 0.15316826631763591,0.5601324332599769,0.8966771826511302,
 0.48067575700012666,0.7402340266091763,0.46181785303644296,
 0.5178715123829117,0.3374794902300232,0.5407983109271437,0.6558990342242603,
 0.36303835575006815,0.6509480530820537,0.48214764482525374,
 0.1314708613615413,0.8497228201319511,0.5880609716746173,0.03700797889187735,
 0.6379277999181191,0.7676756257798149,0.9758303751631707,0.7961542638854748,
 0.9812295492771652,0.35395339879643684,0.9511042095330765,0.6616977944106768,
 0.03837037177330338,0.8019852032450019,0.07444259354207439,
 0.5064243337269274,0.03275241604223067,0.8736268077330003,
 0.28159929303994424,0.8272150344840564,0.3710855897197458,
 0.27668864117777325,0.16700785800568663,0.07555084514344346,
 0.07442136755055062,0.9244951233466409,0.2531889769465362,0.5642136276319073,
 0.5833786691410232,0.34478046911824234,0.9962018089045731,0.754557591708176,
 0.18441767784628094,0.9297324318332215,0.10263704962195175,
 0.20853637409179404,0.47147630862640444,0.276774153625194,
 0.14891701346392805,0.918859802460738,0.4701196126789302,0.6282918350270195,
 0.8193921043621724,0.32232745375952854,0.18358805303890924,
 0.09773187225322522,0.020533789606678843,0.9839753655830201,
 0.9839562425810973,0.43159436380699645,0.7079632239792485,0.9021616127010197,
 0.823285128895697,0.112428795011985,0.9188388245330147,0.6864124155655699,
 0.6872825666083359,0.6105928079572057,0.6502155953454264,0.5883624836503177,
 0.8138241131274817,0.5847790633539705,0.5528043460180705,0.29208507892302604,
 0.3899024148877209,0.9759344569883573,0.8499360391370255,
 0.023791419461131724,0.4077861080197147,0.9293946109627249,
 0.5830678840393346,0.7926580622651367,0.28827273691323274,0.3010629594645666,
 0.28647047513275703,0.1933077721508154,0.29947592600863593,
 0.9366737312765352,0.9739604931999168,0.08551589603638743,0.4307996365838179,
 0.46668134297915986,0.5728353249023177,0.9525292138995779,
 0.22441717471402245,0.9415315377804405,0.44703166156113405,
 0.39234893386818814,0.388820575903144,0.1599856928190646,0.07735628254492355,
 0.35405046875035306,0.17826578614336475,0.19722357891174747,
 0.01618879903481052,0.9416549446774832,0.4102863435735069,0.890558497507379,
 0.06904809638739284,0.7137232885080144,0.27781231987489075,
 0.0397219892579761,0.7568320160806323,0.8107255759914501,0.8689198392765762,
 0.17032561337906993,0.7186416679126184,0.8671174912046365,
 0.08994619433956641,0.03295021108584395,0.8306003334413246,0.745226969315586,
 0.3175405918210391,0.013106753572320828,0.466632535362975,0.8763766246598859,
 0.8012997371613068,0.05899441472478795,0.020888337439163518,
 0.38095547634759763,0.6169753231917847,0.850519661456421,0.0778323236668399,
 0.6483997546702138,0.6531174527200361,0.28962970851110326,0.8030468494014151,
 0.5606474313933691,0.5222633170632721,0.28390329744089216,0.7824565650034869,
 0.9799962542409192,0.5796792506008054,0.31484870418385014,0.5709314177135261,
 0.34961504514299313,0.6009609178372294,0.05808345642841084,
 0.4123865184680431,0.3481476883757335,0.3141657821564986,0.2932459775593719,
 0.3697973056632021,0.9175553406687053,0.3242193830387703,0.5967882399533742,
 0.4360118063073395,0.8468074550563431,0.4818965312815997,0.7499263621453907,
 0.6760464023720603,0.890913445368089,0.40441747795231564,0.9766211297405241,
 0.6158346667750616,0.9018974325031468,0.36536082542953274,0.0845432012423919,
 0.5124097536341602,0.2582932154821923,0.18618462944765535,
 0.08282246279224315,0.5019142682618327,0.4123311233942879,0.8095961651939636,
 0.5779255288840537,0.0345062795455634,0.24341105092535897,0.4802683311938568,
 0.16277395875495482,0.24766205392173435,0.7270364491154602,
 0.17035566145606384,0.7562945370419852,0.014912411168583106,
 0.21396228449605692,0.8692093203970404,0.09342231748152974,
 0.2789856932612821,0.4661129857992621,0.8092146749353607,0.8987551797993347,
 0.6802805655645381,0.2786748799461365,0.05450102273536728,0.9667967172639632,
 0.2167725200817956,0.21644927463110852,0.16107463498259356,
 0.7841310865484326,0.5077074165933391,0.13723946604769388,
 0.04753911392267929,0.7887685125045654,0.5530595762434471,
 0.11870499780025734,0.05476798234861646,0.9521160436283584,
 0.37029304082197056,0.3717081610864049,0.5927069102159719,0.453003730612464,
 0.8931785524803162,0.4589292670055064,0.6472326263555345,0.07509280294008536,
 0.19818321972377284,0.9539665236805701,0.38034270458944686,
 0.7715070476414307,0.45932751051961096,0.7276516886205784,
 0.25148931853381185,0.6297454919915493,0.4002212748846311,
 0.12456775396078423,0.09728419981114866,0.025145402567780528,
 0.013116957981245903,0.5901909149484187,0.4541056283908079,
 0.6734082277596679,0.14707488308373284,0.4854371295344395,0.9900672592449172,
 0.19846361514043143,0.7780449011784317,0.2901518787469195,0.8930871252159661,
 0.5240579401300582,0.5418721883664573,0.993522288460759,0.10880073749905672,
 0.5540883701103257,0.27161891558848017,0.5088045157317813,0.7368340226893229,
 0.6327920433304761,0.09083552828489638,0.6989790956216467,0.896612608561885,
 0.8766879662265854,0.2155975735569543,0.980597494213399,0.5294703142500293,
 0.1480694848165477,0.9309403899900709,0.5727681052730091,0.5560366027006624,
 0.15223945543432205,0.2733325625850168,0.8420251769241985,0.660752780090651,
 0.6334025981238496,0.5393019335566688,0.6003424784874126,
 0.009707344310987631,0.7060924897176396,0.20070012175537932,
 0.6155130975083954,0.9350628791615885,0.14663216293382064,0.791584708705317,
 0.40770436404494403,0.06070611328051967,0.8358048041584288,
 0.23331099066415806,0.5217099952556923,0.36406127891070794,
 0.0651534409118606,0.9312147632232138,0.12337623470467096,
 0.08739822204253778,0.0324211398316292,0.722630651584391,0.6193568641193871,
 0.6926653925320851,0.5592849567643158,0.5471734534207533,0.4818765031039617,
 0.6095283504705404,0.6834583470787581,0.8855360046996155,0.5440274386040255,
 0.5747364777941205,0.7264086622859629,0.6696628902724903,0.643536456976392,
 0.2744155509618944,0.5893948226335588,0.3381642489699329,0.8750635330694579,
 0.5681115554016551,0.728960000380043,0.981005436628688,0.795277257283095,
 0.8880749594178692,0.2432862812266331,0.7666137819242023,0.9810170450185836,
 0.12739934025850985,0.45154927446472426,0.13519482876829092,
 0.9715304571428474,0.3592129855957892,0.21283487819976155,0.7660526877987276,
 0.5416593856582047,0.036346727113555466,0.7764332261704543,
 0.8544590253893329,0.24459095516526064,0.057893556388202816,
 0.04580310488162287,0.7529285268006538,0.8085555805154994,0.5805183800779209,
 0.5607375115940522,0.690219121974006,0.7643471114595132,0.468050725060569,
 0.46294631932524355,0.024713157108419148,0.15920001453199684,
 0.8639211655065864,0.8414918616338287,0.4114508745013534,0.38580713080628726,
 0.8999988313946108,0.42531670140715006,0.7833617421048515,
 0.28440674876263916,0.6158857339941973,0.9923487119916962,0.7747307767141488,
 0.056615890947243064,0.2828477454213253,0.981268046023751,0.5198695684578667,
 0.2520261122831523,0.9711992814026809,0.14369731613643477,0.9884527576963681,
 0.5106350689969691,0.5284365951313117,0.817664136586902,0.44503641279859585,
 0.9444168490728914,0.9112046331273621,0.2651103542416806,0.35058638660390196,
 0.4995410107833964,0.017310667419178083,0.507595932876148,0.62793582152477,
 0.08052802262012737,0.5339086077987154,0.3328220689813708,0.9165013329537158,
 0.10883669895250059,0.4593319250383121,0.202732754584231,0.7872577339429928,
 0.5112533149234918,0.6329912431883418,0.6698521641716626,0.27344803030124565,
 0.5752928857606451,0.9558667012677656,0.10908097251339122,
 0.21354773664217708,0.02715981524231781,0.3040935064452325,
 0.8053990377344077,0.8003198019146109,0.43973886007244667,0.7764392486714462,
 0.19015495254874737,0.7608358838323477,0.3345730531167418,0.417959730513507,
 0.02030354520106592,0.9285920858049173,0.9011323796141137,0.5322352148850932,
 0.08966381901764209,0.6885525514955734,0.93296367840776,0.40095787471374533,
 0.5727811524127406,0.04825380884085961,0.1139788430083335,0.9425701352844266,
 0.15603609972576227,0.025213684131754732,0.4615043745745513,
 0.3895780090573283,0.7147479478569291,0.8045609286465942,0.2791820881129774,
 0.6893980673351716,0.45070974340892633,0.5478492554387152,0.8895154108426638,
 0.7069143545584249,0.885438264938316,0.35446563165979783,0.5883603335988586,
 0.7971387844269201,0.41278274481066113,0.10669477500282953,
 0.9096169762462183,0.3039779929432327,0.0072454847811883605,
 0.7306592467608887,0.8424560084329071,0.8413161551945443,
 0.017939426224329535,0.8395547536378078,0.5104591160925177,
 0.09372212941269287,0.055436807928067156,0.5789588648576922,
 0.701921609306317,0.541805253176454,0.6609307410211265,0.3806213554338521,
 0.34088271663189096,0.3526097804282827,0.8769387819459271,
 0.38686884762660334,0.3722652618924559,0.6924988553066448,0.2869446434403259,
 0.037637650739633655,0.8441769054145315,0.31487566982194637,
 0.7322310034216558,0.7023420573779771,0.878019431858953,0.7428978226205666,
 0.3012557793337932,0.695522285556739,0.5999061329092508,0.08343397484321069,
 0.8244588756811759,0.7752837983023507,0.8333372803315713,0.8030929333372723,
 0.7536905814645952,0.93398227267172,0.1548105845048584,0.03961482543829642,
 0.5840995320374573,0.9174995747976669,0.10684461472227236,0.9187626657787832,
 0.9044611608278749,0.48324855801399513,0.0642057642564614,
 0.47824835372387553,0.09962839078443697,0.20996644309708912,
 0.1945721706111817,0.2534040896804599,0.9666323238619337,0.7337956975623263,
 0.3013278999563771,0.5170811730994914,0.2576223844798109,0.2789377326100312,
 0.6731567621600467,0.8678774516717805,0.09968506888724049,0.6223774185777191,
 0.7724871414749308,0.506358052776349,0.8647473831154497,0.35289728501304296,
 0.7813638826956175,0.7724885730087021,0.469761850273992,0.8102711868497589,
 0.9474983674722588,0.8163296460169971,0.755721403382352,0.2724604068824005,
 0.48989673700349456,0.5098918225061942,0.052656971400653174,
 0.1586494823049962,0.6931062268024649,0.09949001906650645,
 0.47642576403267634,0.8013995877895691,0.4287882067621126,0.7430151977617228,
 0.4212398180541972,0.5362279923951012,0.09595702574629594,
 0.04194075616834281,0.4630275721675585,0.34922854206971743,
 0.9555977058714369,0.17604146856040126,0.8585181772130904,
 0.32869731681514724,0.3914562566795695,0.4938696196531489,0.6666430645721975,
 0.9966288190648664,0.0359158865980449,0.8458355143465043,0.9791584192998303,
 0.9180169874675438,0.8154096477124027,0.011074778639441174,
 0.1934943338648718,0.4772859662055279,0.34544484259983466,
 0.07418548302614225,0.9149808093116714,0.6661736089430574,0.2339450448963558,
 0.16191493029290904,0.8445385858634594,0.9061039810384153,
 0.23532422205979542,0.19729301167733215,0.02502435451638174,
 0.7924096924551495,0.20988312820163668,0.10483403438281769,
 0.6814392579272914,0.5668409338759188,0.8772601083022926,0.2324163183157666,
 0.9412645266650417,0.8729201106625472,0.12461936808237162,
 0.36349361566156424,0.7899184958685489,0.5642826084828817,0.9968435431728525,
 0.044644739254702914,0.2874620336342505,0.08410642151336034,
 0.39858546580522936,0.6633100788822384,0.004614211833867898,
 0.9793952772364444,0.3859983553928028,0.6408236238934727,0.7385476301690042,
 0.35329884729350036,0.7393693977874136,0.4919955004911577,0.7032502497155831,
 0.35935708117709486,0.6351964003268004,0.3081308954649117,0.7143267014953777,
 0.23483664047222308,0.30787056533040336,0.7330760998796061,
 0.06314559620178672,0.7634360590937509,0.894742660935999,
 0.018995646953116396,0.6511785386319796,0.33485097231491845,
 0.9852561362183522,0.4897170684367619,0.37602225204533113,0.3789616297207279,
 0.17522364885514952,0.6021845402800095,0.8342089596922201,0.5588763249401512,
 0.2986475021421964,0.811099623030803,0.5047874433021133,0.4566461372181093,
 0.5402774390125529,0.6797378832774348,0.5062909043306667,0.01001704911220902,
 0.5905824997111598,0.14168947138336618,0.2528252825682378,0.7563599494948018,
 0.4008565211778815,0.9861300534855809,0.9604604548902975,0.12270979932201909,
 0.5599596054910442,0.7865473061062618,0.9501873923336388,0.4710719946272941,
 0.3940810525931907,0.08084904905126522,0.0018557431240727729,
 0.23411831426779228,0.7029413755822613,0.2472956522163735,0.5937663434258641,
 0.042578934006837876,0.027964086750644435,0.587278442173836,
 0.6922206269661934,0.7092163792417072,0.5776716201036258,0.6542316369834797,
 0.6544377128863765,0.19820955290139963,0.6979684948725492,0.4955672813774922,
 0.25347206787386756,0.5468380504835707,0.8859258492080437,0.5667355382304451,
 0.9388699469115608,0.122833097101403,0.6564353031527604,0.36924616220601714,
 0.33732664594282735,0.7891286578871634,0.6387552143594326,
 0.19168369079976522,0.03025107207224853,0.3327305345153231,
 0.6068167898728182,0.23887662876281712,0.6592405985082843,0.8405680973861418,
 0.2925684910888984,0.6549795216868255,0.8661421549730113,0.27925214021964384,
 0.4352576987154765,0.31666715571773896,0.3473165720299901,
 0.47243546811280035,0.036514099124810184,0.825750267734873,
 0.9915562160647784,0.021007453923838626,0.025922056295437312,
 0.9373427822290563,0.14114997320945266,0.5144902437056101,0.5302014881354344,
 3.978609768405805e-4,0.030857034124102567,0.8814980747346368,
 0.8646254453732276,0.3557889256517718,0.9622769126450705,0.9860689203422611,
 0.4495872668750053,0.5583354910219688,0.16904934591305287,0.8630049736355894,
 0.24671811074427485,0.7439735303034762,0.4488017130710893,0.5010401637755652,
 0.30420465535491914,0.9511069211509251,0.9108361246982482,
 0.48393145954871974,0.325250557185249,0.554256407978058,0.9854261811524022,
 0.9973496090749214,0.5164237443177343,0.9120460655209013,0.38038679988123447,
 0.2882217111799208,0.11405270420323266,0.5007687735361824,
 0.004538031843598467,0.3983077984571367,0.5869967182420952,
 0.3045800352377944,0.5170277497072013,0.01550197023855282,
 0.19764826588617102,0.47337066225822977,0.9216137592793242,
 0.41457607121140905,0.023342864416237674,0.5769278155439904,
 0.20298236973543304,0.21135156945178912,0.8457945038880438,
 0.8291825136219556,0.841226734321453,0.20836062322478688,0.7989667012992926,
 0.6404744468719917,0.3252928432829941,0.16859065627029535,0.1904332963944373,
 0.2750605554250729,0.6199224797465341,0.1536025463352526,0.9811384251307587,
 0.11920148934794694,0.4894634307846264,0.5004739550242892,0.9536376886121334,
 0.37833194061512243,0.45878077621147695,0.3855846784048341,
 0.9079083972931808,0.954753904595681,0.7271636930412672,0.7827175765747649,
 0.9252126358530759,0.8946097830609049,0.9198112501932654,0.9538253460190546,
 0.540205064159569,0.3902161747674677,0.8516513440936488,0.655433036309733,
 0.4078604604103192,0.20149280528782565,0.4009044606253409,0.7119228780188479,
 0.9179136066649649,0.9618978568077293,0.2890523246556901,0.12468409075449116,
 0.759013581675629,0.5834998161186079,0.19860526816904256,0.6006299815647607,
 0.15630355440095012,0.14239963689599178,0.9249949526436075,
 0.5999431526789273,0.5213207013294485,0.5660743861965394,0.6578826739545456,
 0.68180310447245,0.03359843126015305,0.7428951721036073,0.1981662741892758,
 0.35045954378346,0.33399393380371833,0.3719265145787479,0.8869220084953852,
 0.6870480261160081,0.003368702933482126,0.06258680484408718,
 0.5655706249068919,0.20554723396051633,0.6759113514634483,0.4674615567712018,
 0.7608262563213498,0.06914134252094584,0.2526816402375941,0.1481127554405114,
 0.8648168480722579,0.23916568792054427,0.7487308474634519,0.5038700859449827,
 0.048681399733441766,0.5165920509009576,0.7708741395529436,
 0.01164724320430044,0.8572515326826596,0.2082004326429655,
 0.06252205717550163,0.5394991403529357,0.15437481009364984,
 0.5693035997943998,0.15994052976795492,0.6312792811250452,0.9310860998401728,
 0.2906791857506793,0.8975489663873937,0.7552310715163438,0.9876120147188413,
 0.8918563777583041,0.6116822381060354,0.4756278302392757,0.46636457234427253,
 0.6878503516497474,0.5285364132409466,0.5331213079069181,0.8716650744261176,
 0.19754218280536107,0.7288139659426156,0.005910464366232215,
 0.5625475150517796,0.5193689398046041,0.7168218687351184,0.9419762192733692,
 0.8760655870278172,0.9691242461919568,0.18424855852386623,0.5365774699821736,
 0.06955993259202031,0.9959323576302687,0.9019433127286897,0.6659678996549532,
 0.58977156675277,0.6204312377330278,0.9231341005314088,0.6264209487014756,
 0.04729354648991513,0.07153021149221517,0.6912279576770565,
 0.1356188661491904,0.5532517171988116,0.7887572963266748,0.7203701070484793,
 0.9454241997265718,0.2118686769266167,0.2968142440766649,0.6645974102526256,
 0.6337285615902513,0.9431999989168198,0.6486771795252987,0.4645449934308523,
 0.13824840244245284,0.3975540541730638,0.7630167015606425,0.7619248322634684,
 0.6591247866201101,0.25600013522748555,0.3257009698747787,
 0.20243935847439432,0.9584738439640287,0.9401109971143179,0.8654201918608663,
 0.5747803112795992,0.8761949877989029,0.732616939691441,0.2555576749434785,
 0.8674727529247214,0.21587915132330515,0.5916902988076916,0.7017654433110431,
 0.9541138263473752,0.3905864187763942,0.6270979925635766,0.07935987851644499,
 0.4258990623159056,0.33356909561023784,0.6112498077454454,0.728618561390901,
 0.5588322665178184,0.27459757194061774,0.7423637558796087,
 0.10626092351003491,0.96865895820503,0.20991544527551997,0.5925354627580451,
 0.9864103907219974,0.22389736400731197,0.5530013302663357,
 0.22399390886320847,0.7068432838322362,0.779313302785963,0.7239380530123499,
 0.804952449296837,0.9414374266192099,0.28420483498131954,0.3338889811638337,
 0.9708272043595969,0.6663446435526394,0.310295331004636,0.6546817986013429,
 0.2607038385957654,0.13739567752368742,0.8838507264909684,0.6986960737677554,
 0.5085008690950302,0.1614904923586742,0.5045770547204889,0.6156199148666599,
 0.8157551406555824,0.0973419474524706,0.776777533409768,0.6111144642008571,
 0.11796505714749517,0.8000656779518905,0.5998829155787703,0.5341503849993405,
 0.5591664790484819,0.46529113339204575,0.9283027828621736,0.954872359720174,
 0.436715551189911,0.11404552798751222,0.18530428860681192,0.6286934777211188,
 0.9753018203733267,0.6217383596120132,0.13457455085968473,0.5905799660586304,
 0.9098393516560425,0.6807396770686509,0.6754665657772488,0.5605602746201123,
 0.8750439859912791,0.38778884459928054,0.8624143173125703,0.9365125148885628,
 0.6201548020774503,0.8800414310626578,0.9903736935489642,0.28465872116634905,
 0.7060047238502882,0.581884351976084,0.433400674096308,0.8566582728328029,
 0.6547611838237573,0.418908296804989,0.421924443238188,0.2511439260208218,
 0.23363140788972858,0.541720864099526,0.8655747765437861,0.9003654717119045,
 0.9629566923476865,0.7730272706849838,0.2981093276941994,0.7768179106929488,
 0.5494198679871133,0.2692927620474933,0.06262708088824909,0.5735765845321361,
 0.5631578731417872,0.5542425963164813,0.1720142047287364,0.7511600147518932,
 0.5683449843636946,0.22513218073997665,0.23601961564961982,
 0.9897365447785454,0.14516187639821077,0.9576676591260107,
 0.004285527572256753,0.48219456005791317,0.5532900565168419,
 0.13194077350826916,0.16488034361079917,0.7918709935659932,
 0.8164302485064094,0.07032792556408285,0.4491683792489768,
 0.26982858950882593,0.09291991275939948,0.6011543899304685,
 0.4174936979411381,0.47340869775114447,0.9785084468556735,
 0.22767148084932753,0.8590176655226871,0.27126456984429825,
 0.7780570058902251,0.5957500856652743,0.8990400814374426,0.1267035493405947,
 0.09861679356241582,0.9104099301237665,0.07449879746047028,
 0.6096904265896547,0.782735988078317,0.8309610589743922,0.3523827100800474,
 0.1584483996856213,0.8740938433333316,0.9163591650520744,0.9869798555957943,
 0.4607992675803251,0.9363700083134728,0.23148376549052563,
 0.10317865795468406,0.6363589361545913,0.4564082125824632,0.7285291506510503,
 0.324113070079853,0.02238673793262247,0.07216429202435148,0.754283125972846,
 0.6698050648141557,0.4075914305212782,0.2672998962609081,0.8655664828151108,
 0.06759605245439626,0.7803270710064054,0.680016631259638,0.9732267925662059,
 0.17391541887026873,0.6540875335710336,0.34706203338506236,
 0.3042140604993304,0.5432664656854927,0.8090492689005749,0.09379785626581061,
 0.06994099682956245,0.463638790830352,0.587300403730476,0.7941969169098048,
 0.5188896574441354,0.4518238422083257,0.5534545454632086,0.8776513564541573,
 0.2599536943681192,0.022574743685916054,0.30878302015958814,
 0.291280716209793,0.4999926084156123,0.4628022926243214,0.982362626128328,
 0.5783163186174267,0.26840733439005704,0.8156202549821353,0.6532305330509353,
 0.319112138946138,0.27508511327653196,0.11617432117785942,0.2442430165017938,
 0.5723367005817828,0.14299919877625922,0.22720081051934526,
 0.15555218832686246,0.8156792254175164,0.3726410073382348,0.6786982491213928,
 0.46381645974638785,0.7510511812893732,0.3778351133364355,0.5794499807703335,
 0.07735563714208649,0.02090029325717291,0.5183309255398799,
 0.023315974844006937,0.15966148227591503,0.3937080599773074,
 0.9368762910846365,0.553071370664705,0.695898844771669,0.6240516132998234,
 0.502969262010367,0.5435260764681312,0.18284006822858712,0.5666457996328753,
 0.6645590627136505,0.880191492887066,0.23793673527220105,0.8280069355292393,
 0.36641758427769644,0.8729612837020474,0.29860189231790724,
 0.8424424483092593,0.2500695235072553,0.3390892710049118,0.804802942048676,
 0.9629101343478634,0.5263882259849656,0.8302225394500163,0.21771119586472276,
 0.5372547810308899,0.5196972630502765,0.04800103165096381,
 0.41041870771714917,0.9135154574277008,0.09815185463511644,
 0.9448737472116032,0.8572699918998383,0.8962478785298922,
 0.052215936209485614,0.5477250700854185,0.2564701808016757,
 0.42586114850086787,0.1275727573299783,0.4499296893514454,0.9169367446766058,
 0.8781876802187885,0.06199808189181388,0.36673918302572717,
 0.4897189027264033,0.0632539492430716,0.31111948235303755,
 0.15136515681051732,0.015404264853947192,0.4347570944487894,
 0.7184831593875537,0.30472445050314,0.023191652343734215,0.9594920577716078,
 0.07641347842964119,0.6495426999832075,0.09474116832850432,0.67137338613144,
 0.42889776017306014,0.20448801528719374,0.9853267618484116,
 0.18400200692598434,0.9828117959299519,0.6963670735792649,0.9796311776109029,
 0.7824604765704827,0.3652573996719267,0.7694782729574884,
 0.032326700684804366,0.7477645263399724,0.6152160379937435,
 0.004673677737629278,0.30810704469800043,0.682546728700969,
 0.29255921313524136,0.03514685760123637,0.1380583723986295,
 0.3396924525357695,0.7800333072612599,0.7450695352774912,0.6453196766467847,
 0.6470352347962797,0.34291618685641245,0.9296306157020331,0.7083738584688082,
 0.5830742715601013,0.06265717476257726,0.17969674996328067,
 0.9784846565059612,0.24770513699986385,0.5064642763894411,0.5581929178218087,
 0.40126420043880384,0.9491486078520596,0.18685906842264854,0.855998933790722,
 0.024298707769549455,0.27899077091583124,0.6851594517835712,
 0.3851059801341541,0.9938860313604716,0.2622626044648608,0.5268786864739706,
 0.6150529990189809,0.5879941910099958,0.479301301020878,0.9903578233565773,
 0.16278563082321784,0.3669179635196246,0.24160213253538154,
 0.45818761823546217,0.8579425671861116,0.6661513976533953,
 0.09521415069068984,0.38863896190728897,0.39848015489239486,
 0.6539848593699075,0.62276487919018,0.3571372490628262,0.17392137804599828,
 0.5821353049051743,0.09757991557054402,6.499038829723691e-4,
 0.5170672284938436,0.9459296018434348,0.8296405337235074,0.11762371686347639,
 0.5079257883973185,0.38207254187511464,0.5121036519812012,0.8706479361095293,
 0.8737593246031302,0.7824891196147366,0.6396449203631802,0.43911196860622403,
 0.6690088763719323,0.9665308905581452,0.317612869564702,0.8930140982032818,
 0.5287827669932094,0.8733627463167318,0.3839810825492931,0.7844558623523258,
 0.8836978972346946,0.14146786317112625,0.5431568173881506,0.7814621730298832,
 0.21667174089623165,0.6601160048253931,0.6649063353146651,0.9090709073325839,
 0.2335155880055073,0.9785644477365667,0.5680825752719596,0.9764487995932164,
 0.12695978750420267,0.7976104108888324,0.3521710915558891,
 0.27705216594177506,0.4268844419358164,0.5184043818963681,0.66679918100708,
 0.16428431926219045,0.04607916674386847,0.04315952096538089,0.52711071874081,
 0.42538843248106095,0.41416379367888834,0.10194537719144436,
 0.9210383363190853,0.28293208259437574,0.5275848012070901,0.3037327700530902,
 0.2374114636224013,0.4550001564353674,0.5508147153941071,0.8946407862280095,
 0.5139786725399078,0.6237523550661817,0.7484644853911,0.18102278822059248,
 0.017313175778174372,0.9416130493225582,0.7118504402079439,
 0.24271861525245497,0.41300581664459723,0.6932305744947237,
 0.051655107957683355,0.013366640985330913,0.2579176429953117,
 0.20586634086960887,0.9778442690640333,0.5708740658523538,0.3666658337246054,
 0.4593596073230354,0.6355283091519527,0.19855417202502468,
 0.10297621565094917,0.7362180666999981,0.9261140741621982,
 0.43773929983005866,0.13571021361983238,0.012808375807104255,
 0.6843717949299474,0.3086454649637467,0.22278782688289178,0.4647514012868158,
 0.859689220997325,0.20983390120278123,0.27437521329002046,0.6842402217613046,
 0.8998203514275,0.6605188947605518,0.3149752434188674,0.8775028418737463,
 0.4169196939116242,0.84530437399938,0.32719189778807645,0.3481660093040263,
 0.20299002160658752,0.3146578093971285,0.9866243298246524,
 0.07368458073116702,0.5819076207037511,0.4638420089655655,0.8746015241078305,
 0.7316426698522176,0.6223938766898507,0.8292123034690159,0.7142544043139823,
 0.5688032702003618,0.9571527622548257,0.5985460018794838,0.6375277491439725,
 0.594089726305504,0.6250542600677808,0.7362824914688334,0.2977368501746942,
 0.8543803383629033,0.028939650100150316,0.08219024680789444,
 0.7425680543668823,0.6935982552431799,0.9723714122796029,0.4208837555839001,
 0.6234663002076973,0.9481112478026115,0.9949886527734438,0.9074871308754042,
 0.9141756813503504,0.08450867440955723,0.8035527853962875,
 0.07189811833740622,0.4894692238388897,0.17824542675076815,
 0.6879983819805254,0.1502539954395774,0.10144620239511504,
 0.40573844164727213,0.03812570576545804,0.6926229460225163,
 0.15983489083618596,0.6679442791376933,0.5184408767917354,0.7630901380575316,
 0.09977653095767991,0.40746528357862677,0.029603197391727898,
 0.18057249349019244,0.5199382383070327,0.49704043799919684,
 0.4432982743815851,0.6665930901014109,0.04699701047486127,
 0.12934629930173924,0.10399833286865989,0.8827620749871603,
 0.32272292017888105,0.5400534781663981,0.8121418066669822,0.772065750257637,
 0.9549398470358906,0.27431307239292857,0.7516206158478562,0.0634339492841256,
 0.2462026021263959,0.4823644272907712,0.1137035823199779,0.5058584488043238,
 0.3575423819511695,0.40110176158472577,0.41735558455721045,
 0.6431435049387578,0.8794034827575137,0.00601715454448748,0.4589642007429714,
 0.4203157312613366,0.30741345534784903,0.8189803764232222,0.8237480760122294,
 0.27887982949557166,0.96502052323949,0.15696442427337431,0.15665654917221705,
 0.8699685121618675,0.6522833683516024,0.5467106130230346,0.6444294774311341,
 0.4159816550518656,0.02046899264724822,0.2764962087116738,
 0.41344943105043674,0.7346815622374734,0.05492248346424633,
 0.10330490807142101,0.3109236679528853,0.4059979531739347,0.6553258435937276,
 0.7864259004125594,0.42953721533077327,0.7683291225662077,0.7558956642535571,
 0.24209799304874446,0.062078080221775966,0.868551719681148,
 0.9230772241912668,0.36244442920403275,0.8473437224249484,0.5085652277169467,
 0.49686536652818103,0.3884887246367038,0.6209876427514615,
 0.19183836933760134,0.6101490252774444,0.26809562624469907,
 0.9004909408128904,0.8418331621496469,0.7438184200753274,0.4287587937009286,
 0.6033249236538887,0.9383316608004265,0.10156134894435442,0.6155552411666089,
 0.6030009587299702,0.6353451785691334,0.19786220634532836,0.7997236294841499,
 0.7405304992378079,0.33317289457679855,0.5172593923695589,0.1656535619937204,
 0.07865099536461206,0.9140021530777965,0.2658068934926199,0.4464002154533079,
 0.6530776246416317,0.9252189042776928,0.7106637814468789,0.5701733966052012,
 0.08383436356381502,0.9417695795677004,0.3753702000112601,0.9884588358579661,
 0.06417183883834399,0.9847307339389949,0.5715933783574055,
 0.12039668794029024,0.7548454346041449,0.9942469525080482,0.4895443313270035,
 0.2175115827892632,0.7007845452087927,0.32227420308617805,0.5449814707000857,
 0.4395188500464231,0.1908052422893387,0.6226602942028804,0.8577721891778822,
 0.37850736921031436,0.6178146573952215,0.796325062596151,0.4699543577097388,
 0.561718657657667,0.5395435896596104,0.825147516122755,0.7723537934151531,
 0.9985895498294831,0.3459026187012961,0.7629480628818912,0.806051546537236,
 0.1472356275105976,0.8584508970174549,0.20668799158813522,
 0.14432495482442653,0.80584660337006,0.4188432494194858,0.07343800632389408,
 8.603550659825876e-4,0.009623156071482453,0.7358460140726408,
 0.9893029876621701,0.6639330536170786,0.6263096596705834,0.4258113972666908,
 0.44835510690939195,0.7191583267402684,0.17394979863903537,
 0.8895482825969641,0.08667022280864234,0.4625903862288976,
 0.03375832933205869,0.061997469869594535,0.11151915895058173,
 0.9840565229658931,0.6831953824342892,0.10576325765029648,0.4363860686409281,
 0.3146900462577773,0.4083024997490773,0.8157533388278837,0.44586755886592255,
 0.5580723164403267,0.72505531668753,0.21719098337755827,0.4129176071799169,
 0.012583831963944525,0.09946512943334529,0.9286719989588571,
 0.16495492324024386,0.2747832055044145,0.868063794045275,0.5643174506945924,
 0.891717870810542,0.3462829919963871,0.6240199769746659,0.22342851791072016,
 0.03446062214322365,0.13073958224466586,0.9265674676247279,
 0.8714227701302643,0.8471397031036618,0.22390381772717405,0.7667423999299301,
 0.3122981617044587,0.02946831832258001,0.3326199379984959,0.6374508220847079,
 0.27004960726363725,0.49407836758191737,0.9226948306205918,
 0.13961156168478595,0.35902812883909807,0.5933790232201297,
 0.5680212444332945,0.0444191069084362,0.9863965395213705,0.2636939157699869,
 0.9105141193896409,0.8670590804126457,0.4305222517825912,0.13801590290275523,
 0.9564853587325332,0.09936244907110281,0.42949948637731183,
 0.8438516039635382,0.9640340407919149,0.2238828175976001,0.8947340956726536,
 0.9564533437887125,0.9046450243931635,0.8412377764889327,0.55199835510788,
 0.21185538230246415,0.5518517250870745,0.7424799969905389,0.2844083514134027,
 0.27181357537548423,0.8194401223963681,0.7629934988641823,0.7597199948220913,
 0.5012680126033202,0.5854080427296464,0.877024610672476,0.7592234112505545,
 0.5398352417618801,0.4227609559751808,0.04183511030384568,0.4312195007005022,
 0.9043653521758311,0.9847995101794251,0.8687931641511341,0.39135121345640334,
 0.14957632324754633,0.3172758020691927,0.6192423537046611,
 0.49442053973703837,0.5663724240360634,0.18433020212583273,
 0.8593033456339733,0.14865209757025077,0.29558371795441696,
 0.33103592704461193,0.5937923108063957,0.6525825168855888,0.7418630489822328,
 0.73450489619532,0.02438978508308831,0.8219061901130164,0.2374576345142707,
 0.9694081491780734,0.4025264356306464,0.8712321957102852,0.7345575962538484,
 0.010522315496312218,0.9551718224359139,0.7880256644826883,
 0.3848178724370608,0.4737226444223498,0.7749991627429257,0.24516158621204354,
 0.7078964776069742,0.03168232740331345,0.1545568872768076,0.3981799824373966,
 0.6178794218324425,0.019931654656261233,0.8187325714259176,0.78482738634471,
 0.28647582329613563,0.038020688190568475,0.5754570441279054,
 0.1771144248985047,0.9787648347900842,0.10357706093924823,0.6063033493628139,
 0.4827277269981062,0.8036254498393517,0.10047034676920152,
 0.19932026661909974,0.27032730643871183,0.33454656887494305,
 0.08806340186234018,0.05467958440341869,0.2743135913478093,
 0.2678609994585712,0.9423534745713653,0.06645141635808927,0.995233169441118,
 0.5409999996451741,0.18377743572042193,0.5755626674405693,
 0.19261792009873635,0.49810721455265994,0.6872126906106947,
 0.14822561268571632,0.12452241689474841,0.35212125856026155,
 0.00617860609484866,0.02469323946685953,0.5143768993731056,
 0.9852830602805236,0.29020174571669854,0.6897814740064946,0.1301937885487825,
 0.46321719918544635,0.7977422858233802,0.7574406085958412,
 0.23542980582318263,0.9384654767189764,0.3570183559023875,0.6592621265003115,
 0.3646364566219128,0.9523144431760534,0.554938325462843,0.921710344689968,
 0.7570797672143138,0.7379764652176228,0.49111742307531536,
 0.055840921156507406,0.8376613872958304,0.5191718777826717,
 0.3126622656728727,0.2030757861447937,0.8554682582346715,0.3845481935991524,
 0.35683865306441465,0.5231668626491786,0.43537166260039817,
 0.6615731102579027,0.6312691776022302,0.9419001555004403,0.4082668693373367,
 0.485552592991234,0.7249948288538033,0.29808977351637267,0.13848922776044004,
 0.08496781655916685,0.5003106758709114,0.5110827551513321,
 0.14813210316425518,0.7478379925280634,0.8107905892180269,0.8475296219229423,
 0.2206558807478156,0.4228065116556088,0.46460633385388783,0.6031053934767295,
 0.6438943469137084,0.6965033077317887,0.06003169788316065,0.5247873668371279,
 0.21042320464136188,0.6135052017804699,0.5922154889612419,
 0.24997383673175533,0.1845980992672348,0.6496982027339688,
 0.00801982799056622,0.3863867669218808,0.907016650709495,0.17364642341933179,
 0.7217462252347465,0.6264864280826031,0.364686695548081,0.6881621300870409,
 0.2835951174817123,0.8353787939451726,0.7804327611830613,0.1993698669908015,
 0.6591437111645436,0.8531917289949232,0.38154618801696294,0.7995533292587959,
 0.8860066368417412,0.5870741179706401,0.4384810114388594,0.28281972336722516,
 0.5263662881166199,0.21992292273542446,0.8748679711058548,0.3824590534374701,
 0.21612986122302513,0.7131072152569314,0.2828182940384073,0.9480431774228075,
 0.35150220488544814,0.9123641497683271,0.442776016922777,0.09012352692367143,
 0.9114878607082458,0.10498554860703335,0.7586075848284203,0.8071228106510984,
 0.7522644648754259,0.5375036249659779,0.17301860900499144,
 0.03953197384174656,0.7128391414841465,0.45490231052118724,
 0.22065514701198008,0.5036704570718431,0.4307794336019788,
 0.23533466950410586,0.05299375193378375,0.4727522074466334,
 0.006600563783474733,0.5931680808352796,0.3667992980937279,
 0.09278533058296201,0.9673653455816119,0.5305747552408074,
 0.13790734144451067,0.4736394677841417,0.07124748628009137,
 0.04073826319371299,0.03435977918119637,0.6751484649343353,
 0.8527909567469593,0.8612323072117561,0.73050224479558,0.8668690576036848,
 0.884049076379127,0.998675141679449,0.5677763614406933,0.27094053471784585,
 0.8881009129809421,0.7219484939655553,0.5435469852213594,0.17099779275727944,
 0.6631623223590316,0.9090689236738889,0.9010488264382677,0.6822425294550916,
 0.4294020467735342,0.3396220534740193,0.13297100654674898,0.6582252487166746,
 0.39250506367978144,0.5728943756467896,0.4950482174946539,0.3056200724297147,
 0.314642331868479,0.8052091556203467,0.5051705065711205,0.4083805750713442,
 0.0659930707713301,0.3143831016949442,0.8259263025621433,0.21931198815906516,
 0.4443584062009309,0.9054255361992747,0.5847739801552055,0.9374766801298794,
 0.1073922809339698,0.25499059780006905,0.34483338969428556,
 0.3504253887603327,0.8805244255279427,0.9211148372577735,0.3798043074650823,
 0.10420472479899068,0.26551774287487895,0.10312160293878614,
 0.008557406025171521,0.2808103945489121,0.22893996497368962,
 0.7235305640452949,0.4586804116740065,0.6398513301718854,0.181833933468736,
 0.904505987417167,0.5406382872096772,0.38616306819906465,0.743602361809128,
 0.4973954835224551,0.36507783715510733,0.1786674277028104,0.6892778316193032,
 0.8717656494663916,0.9904888134454448,0.9588639553723396,
 0.053480287902799795,0.6832505217460572,0.900596864917002,
 0.33152357208888783,0.8956161641482374,0.22803059675387238,
 0.9647695300542135,0.7547649543764907,0.5488586189049741,0.5714902601807242,
 0.4818757537715117,0.8912830245133222,0.5770186752898722,0.02508855568394497,
 0.3366129753198973,0.4504152091727319,0.7432136911329361,0.15454344212641025,
 0.5511498995355782,0.4592083006110166,0.17604845133614577,0.9019193865961377,
 0.5117579854365607,0.4131710539262561,0.3316749555036309,0.6351438725533867,
 0.9218852723063135,0.3364670920236009,0.8788533556758966,0.7586031162990856,
 0.7558438529366487,0.922382403678806,0.7171951552290045,0.9415019158204555,
 0.05423489799602299,0.34685855325916093,0.34530030370600817,
 0.559474016364278,0.9062535747502222,0.619133346698229,0.73964659160447,
 0.4047019251187898,0.25665252493612845,0.9893883391680647,0.5297895506834622,
 0.18576627971000526,0.7865400513741835,0.0010166455922331341,
 0.8966852658228898,0.6865166506286561,0.4623465996013927,0.9641108714435143,
 0.5397457373381856,0.4277990078508169,0.4680604854585113,0.10173005724567852,
 0.8374847935787051,0.2491910593196206,0.3068681836905044,0.32159408208910123,
 0.8782031527465664,0.6356567036391687,0.9479550920847545,0.19195594421708628,
 0.95969944379899,0.7687595282757138,0.2991317498754902,0.8850554583470983,
 0.4106243749380012,0.09579435588016683,0.9817654197254634,0.2519926550660656,
 0.2814394220055014,0.21127325793796514,0.10767792377438401,
 0.7295428017357897,0.3027332507259922,0.2714149862786067,0.2149706081725603,
 0.46107459991425126,0.2501209557014915,0.09487039839172418,
 0.36258953846242414,0.6568355738213488,0.28440142400741697,
 0.21370054416714734,0.9315018611897729,0.5262266888418554,0.7615305497564715,
 0.7338851748369957,0.11436858223201396,0.980462019796393,0.7741817814857098,
 0.5237455402166921,0.04038519030090304,0.5056335295379702,0.0845937416770347,
 0.7514734025276858,0.9595177615729055,0.2254953317894366,0.3899316088236211,
 0.49150833379373005,0.9447712195587114,0.7999780450902132,0.2878350244079694,
 0.22666169956204263,0.33973087036332317,0.4844687193933591,
 0.5397075346089273,0.6313100725355488,0.2530279696764146,0.9848642907661161,
 0.3208753768290121,0.7471415946409037,0.3171024932408597,0.5632097075847274,
 0.6016256110292793,0.7498373809081973,0.4043269922466144,0.4340445596685021,
 0.8695809729089402,0.8458124400154957,0.415809809460275,0.3594439506920203,
 0.6023282412355487,0.4782799461990689,0.38744809761548704,0.5817738309063223,
 0.8583858038170735,0.14648867819706146,0.021905078579128823,
 0.749073955485655,0.0310941172463699,0.15046097562154004,0.6577158010938331,
 0.8584846588554742,0.6348547636397751,0.17087240230392609,0.7324075332974448,
 0.05082212656141416,0.07247654339999854,0.15235625079560788,
 0.13663691170033065,0.4655851953770689,0.3565494660415385,0.863202076884189,
 0.7406147486092693,0.353304867093866,0.7512699375823165,0.40438482673815435,
 0.3988922032568254,0.6514096359392993,0.21873901116812466,0.5488016318338683,
 0.17665727071050286,0.04724698773437974,0.43993789804853123,
 0.2035530036408304,0.4600107744673497,0.4951509348571308,0.8972168173922203,
 0.051325113836143244,0.6125824473506745,0.08197044382290908,
 0.8918178985373274,0.5080468319046215,0.6998306777565791,0.5440427760108915,
 0.7709222676917405,0.9049335945411829,0.18592465999204366,0.8757129535172002,
 0.5703873193599686,0.8529095183284096,0.41470316376694494,
 0.049822782653396125,0.8856553470412781,0.2971257365374794,
 0.7633213807041843,0.9496601009391404,0.7363536965857856,0.23706200481742035,
 0.1450088432849408,0.3663801552315423,0.8521697957998591,0.574120176791425,
 0.9915701692189525,0.6449831680280674,0.7140974971093581,0.5754581573997892,
 0.5485493102504748,0.5092459831783879,0.2669489931053616,0.2624168944256975,
 0.6097655683687857,0.33100506247504136,0.13996333171334696,
 0.4302394943712585,0.5716161014393565,0.7268575378613755,0.8240816895020586,
 0.24543267727900164,0.4586623400095182,0.8007435392719753,0.3142553864048946,
 0.7958589337602757,0.23915029924856523,0.19815977059883083,
 0.9224398070471407,0.12597876669003105,0.3457220921866456,0.9155609644456657,
 0.9663285582242632,0.5372358254508838,0.453197862741339,0.7156233607828351,
 0.760248878952943,0.2924549585756341,0.3166380629026567,0.6821972681554223,
 0.9111303275307959,0.06273694849369882,0.042237406647445463,
 0.21286633658102738,0.03821286469863827,0.34273928468036774,
 0.46073724349199097,0.204887992216097,0.39145572141684604,0.4592274723270757,
 0.4538659963842925,0.2920871570690029,0.8319114792026987,0.23597892023905098,
 0.0709693288513693,0.9088010532704013,0.1531579942101754,0.978469452833972,
 0.9208664946158404,0.4708415728926758,0.5022705656611953,0.22176409269550712,
 0.9651324372724616,0.9358060552793768,0.36355582788592067,
 0.45560227019861554,0.23406858852248058,0.41159045943795236,
 0.40514364636344935,0.01418493995370973,0.73897551082608,0.8623193732144938,
 0.5319506162723023,0.7906973576101487,0.6709618013072239,0.6176189847097794,
 0.4004366759282455,0.9213695482431435,0.04535391976354097,
 0.12207994574596603,0.7382978727073058,0.5311158995370735,0.7912662342628011,
 0.8283057125666687,0.3609393012630734,0.29514760731068135,
 0.45991497126287584,0.2542384057924869,0.02801401902611811,
 0.8236723600531604,0.5546824192046063,0.4199103614694769,0.7268937075522561,
 0.13711736235705152,0.38819874156715795,0.8310326231300602,
 0.010403389761096116,0.8362566676772031,0.8880020698274981,
 0.29571464593535046,0.15807988879609347,0.37762234785636395,
 0.8977488348879334,0.423720869224805,0.8818994512073494,0.8988498649330273,
 0.3059894102843205,0.25247529089815646,0.4604771817638207,0.6267598064956346,
 0.38449710802635284,0.02932036445523667,0.4293378356870995,
 0.9294253899400958,0.3152776165710759,0.9384030832699247,0.990956171045646,
 0.5349781525225749,0.7400114130156449,0.07705728846739224,0.3214205850116736,
 0.9598497334746322,0.8748947119269541,0.3549753193687295,0.9891973892629977,
 0.052415114783309846,0.3958467459713446,0.6696799733719921,0.782170178310833,
 0.7601092322147693,0.8866924001516266,0.10242688546151768,0.7102685697940692,
 0.5023073121673165,0.5747511782985226,0.019621058178192485,
 0.03078590222645139,0.29852163541962473,0.07330326891107963,
 0.9585684659778869,0.5045600218389531,0.26915141930352593,0.7335791642638108,
 0.4617837373049255,0.4082386767504971,0.06182675809818805,0.5729324129652827,
 0.03668987372685695,0.6509324647390463,0.9301439156842572,0.6689075105596607,
 0.10698963742256795,0.7803034701364853,0.9365104157873478,
 0.46391020977583575,0.8767392306280245,0.7641553225990451,0.7445689593321504,
 0.5818960502987434,0.446759571050629,0.5107300093328625,0.7715613475282642,
 0.21957784524397816,0.3610764819829725,0.5317010125322218,0.901476735832086,
 0.9755605673758195,0.3108323876497078,0.8470727133877456,0.4173942363408122,
 0.15237826287592038,0.1269551026211384,0.8429888905625622,
 0.033284422365379984,0.012137873962191459,0.30630727111363054,
 0.8460499161163211,0.5136684197109235,0.9514677738236988,0.7268408189231579,
 0.513091475961821,0.17712255595801896,0.7645252754228589,0.40328873904006635,
 0.4600243489455913,0.823834523643656,0.08941078611398179,0.02321353988303443,
 0.15122443510040795,0.44928892767889417,0.5168091893388115,
 0.33841861294075903,0.09804202897354897,0.37190000672182344,
 0.7120320221037094,0.9543292255384868,0.8704611518525722,0.6257841526372737,
 0.9866575886301978,0.07005659314815571,0.4072424362880128,0.5983056540724204,
 0.9015355759131803,0.26802826984485795,0.1393748657038013,0.8944297847266491,
 0.5844376187591875,0.6522904520842521,0.8474967884338998,0.3320958822349034,
 0.04759091573056573,0.3514843517289319,0.6821427935252963,
 0.27868782249293333,0.7042784000536677,0.09146837663403562,
 0.9478386051485125,0.6908735924143998,0.08474677847469947,
 0.018755881576644984,0.2022795181023519,0.5002753384423926,
 0.3311577697164393,0.9547619546177445,0.44343001449021924,0.6537221560794723,
 0.36137877764471227,0.4576066075523927,0.5127002744804271,0.144336747350009,
 0.027667826987874067,0.5258139050277394,0.674856920057199,0.8549703136268653,
 0.7592646091326972,0.59252590828517,0.0845231782526149,0.5149082777898016,
 0.6396705916342446,0.792138201972986,0.9248037761890261,0.854617708859947,
 0.25534484238672106,0.8040567677026722,0.6828763970639913,0.7854178884004593,
 0.34347769027176733,0.8834251052261937,0.015101659130155931,
 0.04073338532054638,0.07003806546273139,0.8596348160780253,
 0.21804164893892086,0.94787793142425,0.9938098269381397,0.7889496043232048,
 0.45365498166471163,0.4884087113390915,0.7150330916268164,0.750421022671429,
 0.31866697595086413,0.47475805772195645,0.2902773196034135,
 0.5325587120491546,0.4051051843255592,0.12595882236843026,0.4441515234214555,
 0.7789553675132527,0.5057881297420037,0.15235554966130005,0.1143697190613916,
 0.604101433915534,0.3693469812222703,0.6019879850799611,0.4575529167747757,
 0.6187588101180945,0.7110658782988124,0.28551652736623245,
 0.05140899448682712,0.9761674265039398,0.862075849542995,0.3871746759720667,
 0.8766885747162187,0.217192317626678,0.5136833828726499,0.576566996688044,
 0.7268331168932356,0.9846469493046949,0.7218984948611981,0.3844182110494,
 0.48441511460578335,0.9409268979962526,0.39739413888421893,0.288308451621921,
 0.5445864963996985,0.6901345336395793,0.1515365103805313,0.2332348805106258,
 0.1920892906103392,0.12419736353625055,0.7900287296725299,0.409617332288966,
 0.40585297012085353,0.28598172707432057,0.8765659833338023,
 0.46272386329182424,0.046974120920716955,0.791009925874103,
 0.9552393646845635,0.4021175830142856,0.49899304391421717,0.4947119549924487,
 0.22083943609805545,0.5799528160841659,0.45891725709594233,
 0.5080048377618986,0.18413518127307704,0.9662159079420674,
 0.06295583485290701,0.9956544552366945,0.0066434593180250134,
 0.9712362467157791,0.7473384812257086,0.7416533584127091,
 0.014356524724385533,0.4375496864078865,0.9892583696004728,
 0.6946727847582831,0.6588512627814117,0.6358961566578326,0.2512129653804699,
 0.5886635944506817,0.026174901664893002,0.016119378039434862,
 0.8146226218280792,0.6767829307568709,0.9624210286979835,0.24805492111041005,
 0.3859447740095123,0.5079630417345113,0.7788960552583516,0.36626983110600075,
 0.7546301850007325,0.8671148497664573,0.7382131856096668,0.9468469206318794,
 0.16754752527640138,0.47541722212430326,0.5293958916044583,
 0.20597181121073538,0.7289242280265265,0.0429566963033754,0.8136213173335319,
 0.8303303214594757,0.3981814750968041,0.5299476302936844,0.36122867523949875,
 0.4702176992036541,0.745336356990552,0.3436190458487388,0.5539677051551655,
 0.6289174215398664,0.870418185758436,0.9744099041904065,0.206213370178138,
 0.4025870221515607,0.8976268283507965,0.7750998141374353,0.27161686815441743,
 0.4921422049433376,0.25585948250348367,0.35056758044499303,
 0.3931318103431838,0.10769968744348046,0.6422194477595594,0.5415722095842286,
 0.057826056322657915,0.08484870168936198,0.6172384699532288,
 0.19561534777578016,0.32008270722048104,0.6988374835160084,
 0.5286519253758543,0.2024838535217135,0.028584990643828245,
 0.34705033236579097,0.5708747256845255,0.8430212201223667,0.399485194926275,
 0.44672342506881957,0.6833214821598013,0.3289970007927949,0.6420030892428366,
 0.7229466442869011,0.6766666879483207,0.7529331148222895,0.605038468199288,
 0.870459090930257,0.13254880958361703,0.3363736547671885,0.09381357893708697,
 0.7303137982125208,0.3411082735805848,0.4940103769947215,0.4780033944101203,
 0.7685773797378593,0.7326824693861155,0.07352046730117223,
 0.017566540124740126,0.4521754258120275,0.9318635506199043,
 0.8190991799777688,0.3665083234716441,0.33989547541011644,0.7275616439547846,
 0.7948455403243575,0.7126256970001494,0.3239768270114193,0.9911066817840068,
 0.9457597673873257,0.34683124843939894,0.3464010330410119,
 0.13538238124123703,0.4864808998178105,0.5686753303553452,
 0.03506220393855031,0.06661303077418124,0.4132008738459152,
 0.01753453355047241,0.684660475277131,0.02219204846262468,0.5568870748509169,
 0.07097461658848014,0.13905670451964625,0.8279254419933905,
 0.2754274394187156,0.16725333838558343,0.9804053406173834,0.854248249585288,
 0.9541198534154376,0.5414511520226288,0.617644558057572,0.5875700737400291,
 0.31218759748021196,0.8476316046456497,0.9766856300266764,0.3748706248465292,
 0.42220126296541305,0.9143264358800879,0.9645340823625845,0.913547212331657,
 0.16403959989627337,0.948901916188529,0.2938730323841924,0.9211258635415934,
 0.38940344391220594,0.5579127588688589,0.15268015182198602,
 0.8776575298345499,0.6562048423931909,0.003559884582859607,
 0.43143175783773857,0.03900899031280436,0.4990197863795698,
 0.8048161571446373,0.8799745693968865,0.987408354786495,0.4898712811520053,
 0.3375539162546829,0.6888359262488448,0.4372226772022644,0.9837490739889615,
 0.6606921847386793,0.690252657553422,0.16449951191930368,0.7468521892189296,
 0.9651138000556059,0.8482790921252646,0.318965114753646,0.32245983758445373,
 0.3979894316777889,0.1424647348724397,0.27377337289991,0.10090422725870263,
 0.9194182157367126,0.2773651512787271,0.47511494450029124,
 0.30645164480584797,0.16273657560895505,0.8325490999670963,
 0.25191022958711917,0.7832148680202347,0.5552214930979218,0.3762896664165444,
 0.41734833691948925,0.23189640915286114,0.5145450712421578,
 0.9907842045343758,0.18299131464862117,0.29233348521742486,
 0.15296411367707607,0.839797890866388,0.7581349268524095,0.5888355880986698,
 0.9923077397098702,0.39364173647972855,0.11456060670546742,
 0.3178743573279421,0.9015052902609852,0.6280941562968332,0.9311177355354717,
 0.18226966628850927,0.3110862470767185,0.7896807135721404,0.9509715332897257,
 0.18328313438306842,0.3038556211209562,0.858915260176078,0.12261734180536077,
 0.7677465387242799,0.8120889334357524,0.5347218751632651,0.60115149588952,
 0.40712735674692224,0.8323062909829111,0.9993620203752379,
 0.03674506555877166,0.730310831685641,0.8552421654249789,0.4093677375430296,
 0.42374725201479835,0.014794140583172743,0.7921939826585002,
 0.5977525064587264,0.7776205156076336,0.4746201890964237,0.07310781440456016,
 0.83768404626765,0.0634645586577377,0.35810522072485984,0.43971870111671874,
 0.9810797487695764,0.7064840464544663,0.8988445426601173,0.6452759325481887,
 0.022832242859350593,0.38160835294573026,0.6806981778057941,
 0.46566277623865215,0.27397654750389366,0.6125759529276015,
 0.7970791873723182,0.38541894457128656,0.730059139353716,0.22441551639134338,
 0.642976113784126,0.48219440483752496,0.35044907328773733,0.265015728888093,
 0.08410107020706614,0.4399207256983957,0.5818988390936426,0.4952458490826265,
 0.457997548428556,0.0034616316334721287,0.9023356497673083,
 0.6777012432209935,0.5930109428689819,0.6854824751085458,0.4119156942932909,
 0.39275397190539096,0.5779237696536635,0.1916928057833932,0.6340411207364164,
 0.7261507707561273,0.8470448482616296,0.7414112020565391,0.39267406566636764,
 0.32305204784728525,0.9958209131254806,0.0828810003164624,
 0.14420066215303007,0.05179869509678403,0.04680225031056762,
 0.26104049624302916,0.9163706311952964,0.6848320716060995,
 0.04204231257820856,0.4571727746318195,0.5146285625314004,0.9121825903728288,
 0.13103487722010498,0.71908735506646,0.40034645204740305,0.06527884073474377,
 0.7616518075565042,0.3305161990308556,0.1437097544792667,0.5061559861899942,
 0.05906295135077144,0.5539866795797886,0.5518448499930843,0.2554285742897988,
 0.5528019878400998,0.282895996945121,0.39093362546729427,0.9214896582695725,
 0.4273394861924803,0.4729003650899578,0.6248366367933225,0.7072105164532226,
 0.7472073337722872,0.12697758138770465,0.1417969448978269,0.7410925055526812,
 0.9297937482884305,0.5839984382938138,0.38336119781479217,0.4979897583940327,
 0.3245470219632727,0.612312610635431,0.31630009455179275,0.7953440875316569,
 0.5469210954337783,0.5226128510933257,0.028179711449212497,
 0.3008096309732955,0.4636106211493116,0.39963720575898354,0.7842864433813127,
 0.8786009270517681,0.5712162955508248,0.06560067860149565,
 0.005296849366114964,0.5030312785360683,0.13613083764175127,
 0.16401002677341192,0.19777158253854255,0.5429946931117539,
 0.9465439135499019,0.1072238233139583,0.08864596229737964,0.5834044680346374,
 0.580367940667796,0.8416020589571406,0.09912148692101441,0.15343059765688372,
 0.6205638849395813,0.19892196128967343,0.06242105547040755,
 0.8315628654698667,0.8100631647106467,0.6472596891238477,0.19356940593834282,
 0.8835733066460201,0.2356296245956263,0.1974170877981022,0.29746392922285014,
 0.21693080501715867,0.6762214685259851,0.22848158251052142,
 0.5630341277402514,0.01656895123711244,0.39438010927389755,
 0.41328797203078393,0.8583724440146662,0.18115197655608073,
 0.07187713907316684,0.46593242030128534,0.5065478435974288,
 0.5327935406624648,0.5492366102166001,0.30222591148871525,0.9833368568578735,
 0.28509966120450103,0.6874398499677876,0.4343275259835254,0.7037524406350715,
 0.6268858489969128,0.8352961745124083,0.9446123762202983,0.6549596010024659,
 0.31106767707021443,0.071191351055656,0.4826284566577612,0.8496854170779288,
 0.9793086899738221,0.9334117086233444,0.3217671245361796,0.6459659891278239,
 0.6772556219865241,0.1246380255945232,0.00967435037361597,0.7548202038501349,
 0.02410999215380738,0.790210333930473,0.283619749287108,0.9894957185158819,
 0.006952883014276545,0.24752043996390105,0.05010173566004372,
 0.7926826847309397,0.9593406319809925,0.40583636085994346,
 0.039805976923552255,0.2972589685397684,0.9293592303588317,
 0.7698752942060038,0.27921711478188616,0.16337224202858813,
 0.8286179039139927,0.30731142865905337,0.3186894845655329,0.1949992037426309,
 0.55439959781483,0.16977315649765867,0.1695072071166639,0.9002418165510785,
 0.16146462116139537,0.215548550572915,0.49047659698479606,
 0.14732472327089974,0.32952960138183685,0.051916283589948486,
 0.6813740426959825,0.12559110483742053,0.757183141773777,0.16394494969785112,
 0.8619267260020775,0.964132855137512,0.3204353804887744,0.07948659110502365,
 0.9650682418532164,0.6481296376267977,0.6922795259369977,0.7192826088943054,
 0.8393138984198133,0.47526672705678263,0.5095257148467912,0.633678576368217,
 0.2004864856486538,0.4600754108620352,0.22582891257793802,0.2619722456599278,
 0.17253924566392653,0.6719760984811506,0.8436675569010447,0.640093724204126,
 0.5929118788835456,0.8939455263531594,0.024744534468951307,
 0.39943333450091434,0.2737042091836227,0.38669821260511916,
 0.4805040648317045,0.25875857131004165,0.5781579665357405,0.677920038541515,
 0.03778130020444026,0.7966545386154816,0.19860230322340744,
 0.0014860734362578398,0.8192586643642074,0.22982883248348573,
 0.49853272102272084,0.5815705877002826,0.27896051391375476,
 0.5796682353494406,0.5812660440473918,0.6889968162714051,0.2843023957668862,
 0.5725619298897412,0.09139050010287764,0.12212469517039082,
 0.7489515379353863,0.3790454646181426,0.8314693564722704,0.7399585189152681,
 0.6062324443380009,0.29825226615986533,0.4802848971215252,0.5976310697840349,
 0.23672890496057974,0.3208639829576232,0.7866152365282338,0.6397377251880588,
 0.4177877278339859,0.27191203698775235,0.5790716121711842,0.5271931124106968,
 0.9357588870669342,0.9875767213703579,0.2588524896969334,0.7129275776849308,
 0.821086091826607,0.7693536323642997,0.15378195437775688,0.684854189189069,
 0.4429094703320291,0.9150180446716749,0.5632095323915514,0.5409471114745654,
 0.6448837040102361,0.5531963389052916,0.605317893349097,0.9314434013168521,
 0.977768463594989,0.09301988158370555,0.5479818494848162,0.6151172968931132,
 0.4790603152235563,0.5581655217708656,0.7689241136400263,0.07423280157393575,
 0.205290215290431,0.7959725034809129,0.6651868519703494,0.9190385507050758,
 0.656813609268242,0.509766490614384,0.3423305312230269,0.3058150766866057,
 0.6392534217722223,0.11639113487871366,0.6495007043948743,0.864221226330118,
 0.7667868704501609,0.9155259392093298,0.45108816407271324,
 0.09142594224336031,0.20785618660260508,0.7890832034663467,0.596863331319832,
 0.7518185608949073,0.6164541510082261,0.8869132695353888,0.9362031333614542,
 0.5481180507555401,0.2794203965201364,0.7967812450933688,0.8269233223725908,
 0.35301627812930914,0.9882654931329854,0.5359898315057273,0.6945758196502866,
 0.6008144938813329,0.1661919787194045,0.7129694921822226,0.8503837794793023,
 0.006957108340089779,0.9181320754290898,0.013916180248553678,
 0.3844645553126751,0.020095756593785374,0.8377830796722129,
 0.024050350664489573,0.5268665657766787,0.4046896843352876,
 0.42275694097944305,0.7433619544597824,0.5426316115985153,
 0.47169693732955886,0.8037053967002674,0.34260325570476957,
 0.20858259639408172,0.6600214149495792,0.4437034627340367,0.8621662214096419,
 0.9491317193520302,0.680483841437334,0.7556334312206265,0.3816146700531584,
 0.8120156821461761,0.9555782141840787,0.36792539155589277,0.6422934607048056,
 0.6277861716733171,0.8152504320587843,0.7424012983685602,0.574647916144446,
 0.7004396292750968,0.35076520729805605,0.09438857355546526,
 0.20893483443501637,0.49197278372456543,0.5660689847566163,
 0.2686029970304299,0.9871079717853849,0.1800248397182811,0.4000343129910534,
 0.7811098769104852,0.8769466982570172,0.06388432914876396,0.8658316788394071,
 0.7976129345702738,0.7970368566070498,0.27403485675665495,0.538737894524231,
 0.21888510982545628,0.1850149060599875,0.770174293033355,0.13796347430381872,
 0.6982032605811663,0.24178165440290655,0.2916472903948132,0.7484521584045587,
 0.7836886219518132,0.7466970253150205,0.29254868413539015,0.9413798717556028,
 0.46082450678581,0.9177185543711035,0.3771066656283518,0.7875277857208851,
 0.7449113323276677,0.486818175893466,0.17898196556281898,0.18749743808794272,
 0.34915455013147023,0.202008616509756,0.9874740710047061,0.2553201150174965,
 0.3414475133186776,0.40760461090140054,0.41740578075769086,
 0.6723267416418118,0.3956480614989333,0.12887089002141472,
 0.08656913007494982,0.6589107220071683,0.7866528424205873,0.7832918975674751,
 0.5076624644451629,0.03883700952639113,0.8879827579730442,0.9980438884927374,
 0.08647845381977859,0.587576044887721,0.6185211860908582,0.4202390701297869,
 0.3558837523227738,0.7868034655033399,0.7021003482336945,0.5176558970745898,
 0.05044490918368216,0.3095573261366349,0.2179542120211817,0.8949865478047089,
 0.5737727575979082,0.9727936572465999,0.09378334276992217,0.6020780015758045,
 0.236437714854012,0.3664771068205064,0.8139696338347197,0.6730905409520865,
 0.5125917436515591,0.5428920439385579,0.8778390268191317,0.9856591560980096,
 0.7097063249689607,0.09058754893781784,0.13356898136726836,
 0.5288459622183965,0.7552316258418889,0.6140608671788859,0.9976763378842527,
 0.2632739475824636,0.038221772197999715,0.6445975639271576,
 0.4193145511191625,0.9912358074273276,0.8233173410621555,0.33920110171372675,
 0.752113706486042,0.7354206870878511,0.7812774407351373,0.418940131542499,
 0.492312886042772,0.27706348802086667,0.24278029410366597,0.700774118292604,
 0.9640315090051855,0.9187436415292505,0.9836001180980387,0.8718011251901181,
 0.6357652541070958,0.8165144546323732,0.7136586807224219,0.9748713476905542,
 0.5146392974519182,0.07949345792163198,0.45238812568897646,
 0.7631417603540022,0.18311796621352228,0.7376518942259941,0.6729960294585273,
 0.7830906188075969,0.8620494877615172,0.4882786909290291,0.7195844632924335,
 0.6180641047740351,0.589668515887776,0.5289142657408904,0.06898201733011455,
 0.011546054380587534,0.8941301901876118,0.40757811091598706,
 0.7084931332359793,0.5691429181206774,0.9010313730033194,0.6283708639526845,
 0.07068332854817772,0.8533186126794776,0.4778885690791944,
 0.20682807048420704,0.04488049970226493,0.06927352636734119,
 0.9535476378044604,0.3760417002068589,0.3107098339302825,0.7853682728724504,
 0.03988846246733768,0.11089835127721193,0.5130411434056487,
 0.16544807425469032,0.12127688511407264,0.29674529382557313,
 0.9857787623189389,0.7794923922390449,0.8546324629927984,0.6342517988860008,
 0.5729840218749064,0.23103658768187918,0.4060077200674519,0.7032438234622973,
 0.18218847571845231,0.879425620263647,0.8600259074438477,
 0.034195116990605845,0.7206659428893426,0.9100788312564759,
 0.0011632625998261314,0.6966915710461233,0.21143147181960897,
 0.8120285676803547,0.7150978540949872,0.8604741417443929,0.3487720499935161,
 0.9476418821443824,0.7863739883069515,0.06506483238576066,
 0.17094030939498406,0.5300687410551954,0.7530840170796281,0.509004905650366,
 0.7045214456974993,0.7323921088822979,0.6945697428290969,0.12058817427642676,
 0.35429942670205505,0.8069466540085234,0.043321755701787534,
 0.8834362690925629,0.9963710509815317,0.4439490860915918,0.41802711527326475,
 0.6435919941157348,0.9958751642204963,0.10829921247869512,0.3519046106086928,
 0.7570748434635345,0.10019022614207851,0.5157104304347027,0.8888399524501985,
 0.3962328694548394,0.008865249291808541,0.6128785911640389,
 0.18751367740808156,0.4651014968211733,0.15223587214159828,
 0.7256565441238465,0.06180165528735704,0.810013923342721,0.2645715895763139,
 0.008783379022385018,0.07507720771881832,0.04750103769989222,
 0.032048590242341635,0.7508099208153766,0.915626023015487,0.489487313781933,
 0.8020866454826218,0.5542409450844816,0.9851386588918932,0.33538981001966206,
 0.685896747333295,0.008235328512690265,0.394916529288575,0.03676393421491031,
 0.9145113100030835,0.5074135163897996,0.5373091365814144,0.7364029973104691,
 0.343325001890567,0.39792754494017757,0.9200471418732736,0.4320798277480159,
 0.7496240134369528,0.9747810397737269,0.9456034489850987,0.03080905781366683,
 0.14342538848615904,0.5863265753918051,0.3575915001254406,0.4793151840610762,
 0.5250856602310138,0.8560811124081978,0.20430249424479152,0.8382503801303124,
 0.04258717968653791,0.029660724410330186,0.2091190587727184,
 0.22866867935790336,0.16822253700201095,0.46203998456227935,
 0.14182934802263736,0.19800365162413502,0.25004038068294443,
 0.9399037626360948,0.23919040603209374,0.2499599439807727,
 0.35193788761726674,0.44319951472252805,0.5586992388165408,
 0.31976662124022837,0.7245681946043458,0.7921451906071161,0.9376157663660808,
 0.17561596124463763,0.9901083836340976,0.46963493048096117,
 0.3466510474755896,0.842895301367018,0.9691161443236663,0.9700355773724886,
 0.5236186767376052,0.758249882241659,0.3946791691608831,0.7410220111183654,
 0.4705751967022719,0.7154497795889615,0.48920024509504734,0.8237835133964911,
 0.3900752396140479,0.9159265492731762,0.7395976273006349,0.4051763490526936,
 0.8427240959600376,0.4779464522279472,0.1499246554951097,0.8517134929191313,
 0.08960253100383708,0.1971277051753103,0.7381140061266172,
 0.11437185207518552,0.036037573122586375,0.3399168856937955,
 0.16780396679726217,0.37503393358090253,0.35285943800130515,
 0.692953905276323,0.2538977172651594,0.20041597798134847,0.19329686551118108,
 0.6832303516418354,0.5026082977038453,0.1725422374719383,0.7961374031937207,
 0.7714999775789619,0.7182743731188931,0.9813677847040321,0.5066437466555678,
 0.4977260532736184,0.8413008716106247,0.3165498032669857,0.9020846497645976,
 0.8146267862827035,0.5112444225109878,0.9201776874734369,0.6904346615991098,
 0.5675210651803481,0.4933634320191478,0.8504446294908561,0.7021558650972834,
 0.6538866444509311,0.24730185739631683,0.48108507451564686,
 0.02347238376412819,0.5119644331551331,0.7294689636855771,
 0.040372615770902676,0.11120420534263054,0.19174205101432573,
 0.9423202187522901,0.5305069781011064,0.5965386911321255,0.20821783405088845,
 0.6234025477369814,0.9672462568867529,0.9359055491687158,0.7703670871470247,
 0.27305560067925294,0.5276097061951265,0.44481329238063994,
 0.40491955683400704,0.3127164475767007,0.7807438702867888,
 0.15429060149260776,0.21072297442423016,0.7657709301613362,
 0.32542067407422237,0.26511399209388464,0.766069202056657,
 0.00986580721260455,0.7429480496260813,0.6670192919869417,0.3293688821869817,
 0.47026003001893446,0.17063102188998625,0.5410615881731627,
 0.31413977126791726,0.45469133814776164,0.7565016080341073,
 0.30231736172441404,0.3877157137940942,0.018657684607040537,
 0.30680519775950643,0.8688276704928958,0.4404553214010173,0.5749300066272022,
 0.3558755339211285,0.3117019729777589,0.7055642569907792,0.8295385873588218,
 0.662211338834856,0.7416065627325074,0.31994659181084106,0.47549812371984723,
 0.9852375704207629,0.6779238772944162,0.2748976522677852,0.414911594833434,
 0.926223660352358,0.30564494133216125,0.8242079653206451,0.8003360578604645,
 0.17696481845154832,0.7995387028385222,0.8856582949678238,
 0.26809969200045425,0.49293118264848157,0.22545001174428314,
 0.8288063475975624,0.16389529737372444,0.26350520691838275,
 0.9031259196249745,0.5977134033914199,0.42517031218529644,0.1587938440328116,
 0.133448498641086,0.9319526814365211,0.5225662992548639,0.7359814935677148,
 0.007341607120115778,0.8565262181460871,0.02226248051700197,
 0.18669668374023163,0.19250737314609578,0.5829761092039395,
 0.12164352133336986,0.1591024164611764,0.817538673344387,0.46546377775880743,
 0.5875856286968247,0.33164045274736986,0.5277254151690319,
 0.36635063297149273,0.7933812958861068,0.29207836270262644,
 0.8037318953356216,0.667341526782938,0.454294011597587,0.9940288047159843,
 0.6059286711202019,0.6984738082541639,0.9916062818560032,0.49783712406253977,
 0.27766652321240226,0.3809804317659653,0.2564339003478553,0.4216601360556238,
 0.9903960270053764,0.8325432638585057,0.5923627916812371,0.6294362448141384,
 0.8335051114017921,0.3182200489901378,0.2775637447855792,0.6004917325465856,
 0.46059033131756866,0.13949703173799755,0.04155917553561461,
 0.593473915327095,0.4186322237379644,0.6652657883507346,0.4500782440196319,
 0.26367330504891673,0.7520973586743395,0.20135518259002838,0.919326950500152,
 0.9196634039511835,0.2601289822364413,0.6753827850034426,0.5973615541644316,
 0.4275164859196606,0.33812965981378373,0.9969588678597889,0.7152589084799105,
 0.0390377954297525,0.35310571544788494,0.6723339468128998,
 0.29233063220338873,0.8910378943517027,0.3990173796966613,0.3957574614237711,
 0.8799287995178212,0.3915690419923531,0.29706012362509504,
 0.13311687442961306,0.5184094318026315,0.5472047747088149,0.6143629355624143,
 0.03189290577635895,0.46634927224674483,0.06580458399033451,
 0.6331743222644978,0.7689861329860572,0.7624809409807964,0.16044773084509023,
 0.31053083241476853,0.8268018113650217,0.6354165855637689,0.705198225062529,
 0.49447906594621593,0.7243352554683853,0.7352541556917674,0.798481878564923,
 0.4539436643374739,0.2582521530305516,0.7734921604472954,0.39576396796713675,
 0.4585424793515326,0.17983884887639046,0.9538037853697545,0.2372425767945372,
 0.9148330802012223,0.4977307529234364,0.18899922636071154,0.7104286275214441,
 0.24546013577899395,0.38436768232188134,0.8138807573260378,
 0.19936726166605423,0.8324970234342458,0.045775990044617076,
 0.6136055169672276,0.9812840805217791,0.6203002422940396,0.8621774500363161,
 0.7709955952681624,0.9970483144110143,0.545349777844614,6.041836178052318e-4,
 0.22685444320764392,0.26735500345664054,0.09039568219868421,
 0.15721674769309113,0.13432204372914958,0.5206059000915454,
 0.06365850882419288,0.09956934877200796,0.7004292348372798,
 0.6636009961211129,0.5483895297062471,0.789702352059302,0.47124475593447523,
 0.7807273549774728,0.8797626620100039,0.5900040046656879,0.2643435013275556,
 0.39275711678085756,0.9402289891575358,0.99206115872422,0.3571581456787797,
 0.2690302196203611,0.23033768381473219,0.2795841909524279,
 0.11247360916457982,0.6206023485950395,0.5439157108254361,0.7981390129394585,
 0.0902797304098848,0.04493174436109426,0.39619579855182807,
 0.1629981403114137,0.09242919637791491,0.8515393154399207,0.3348170006787381,
 0.7643550314762091,0.08285587972021247,0.5337115623207543,0.8706081975276277,
 0.392498906651812,0.8474658943271859,0.6207602513365311,0.15772209083173816,
 0.28725919963347213,0.7943299510206921,0.5896458033025909,
 0.009031521502310902,0.9240658994227858,0.33562996345006324,
 0.09332086528070671,0.8847485517763882,0.3931022827815942,0.9192607937320071,
 0.631583946087017,0.1280378719285128,0.6919285609352421,0.486477642772294,
 0.929454022850009,0.953645247310841,0.8033021349070886,0.817562347415073,
 0.11233546069392664,0.07001042931199608,0.9191209906441272,
 0.24087421910700346,0.663035065786671,0.7726974023658562,0.6940794731436619,
 0.2998251977593016,0.7413255828133931,0.6700625472261922,0.02485191355019134,
 0.27697108481697064,0.4827708929257635,0.937310779100143,0.1466861308666776,
 0.6333193076889,0.6296187075022697,0.7945121641832089,0.8643825989794623,
 0.533365386039444,0.8614511713464391,0.892027624232413,0.35705657175789796,
 0.766868722040174,0.5111809476365122,0.4689998334068738,0.40446049577813614,
 0.7353658878669042,0.6750057849796269,0.34871246851712956,
 0.27517999485844236,0.06389290524318914,0.911605391072651,
 0.48842244737016927,0.10682316762704747,0.020069179816709526,
 0.9925234459276786,0.09601545385448762,0.8495873388888028,0.5644938919347955,
 0.04025029284718307,0.5558803824669821,0.4192656331789236,0.8174788403933747,
 0.32988815767073165,0.652309793919127,0.32528710172924835,0.9971056878235874,
 0.24897993419609943,0.9096313027640623,0.07578658267796401,
 0.0884941593120292,0.06985573047099114,0.7072271913017384,0.9379631275423466,
 0.8578758182554612,0.38424872625182416,0.9966259645279257,
 0.28281620231894666,0.9541241508877597,0.7458726329777574,0.8518508339742435,
 0.7955444445224334,0.8052268817874434,0.10824567192841505,0.6363786910625266,
 0.09560419226026973,0.5709810035705278,0.1593258990147457,0.9332097031236236,
 0.8002073482158829,0.1827820494229967,0.3062635613027609,0.7127803367931271,
 0.4766734425781458,0.7113230613031452,0.8273623174517072,0.2165188431447126,
 0.7514626255865752,0.11847980200056163,0.3054935470283251,
 0.17857085620438928,0.29916431456357295,0.4723133735694506,
 0.9975612265752729,0.31476801967992374,0.10546209942648832,
 0.5592627674723308,0.8057275192934661,0.5530153520992491,0.8164441963035832,
 0.6358700621949744,0.6086171477627582,0.5706397671676453,0.3176293155445923,
 0.42144753053876904,0.5916750515852625,0.4796383688485608,0.4373539777196229,
 0.05309423887176268,0.02567048026689489,0.5530063387943208,
 0.4844704476516023,0.9792946914881077,0.011999620760810181,0.463307083899548,
 0.9868731395551577,0.08465146991225148,0.9918636845332541,0.5984211953007175,
 0.6808604183573062,0.5638208800397678,0.4098346863834148,0.443078097426842,
 0.6226621251568442,0.9776748838300054,0.7085467172753626,0.6448994749507881,
 0.3865927137049251,0.9668528176202307,0.7877847933911941,
 0.036180373493398865,0.8762634672704452,0.007500009606562497,
 0.04591267995989923,0.19619162435341364,0.018848419559584162,
 0.16638350353753184,0.4035010553480358,0.6193610431545686,0.844154636679737,
 0.34466448752887724,0.517916600187272,0.1364124990176726,0.2598598333346279,
 0.48260102915652503,0.19016224406518356,0.39309908351301004,
 0.7939669498472929,0.35126676440797333,0.5750679012779277,0.5992745418572101,
 0.46671674507297467,0.8421767792212802,0.9611753044219126,0.4188491644320944,
 0.6815044159175586,0.6424989742510345,0.006776966447967214,
 0.7855497377636328,0.22969899847912767,0.9947473285358261,0.3340273448403025,
 0.9983403837443574,0.2784349448186536,0.7949881841207915,0.8122958918796,
 0.6313600958328475,0.08725357437583392,0.30868772554799473,
 0.29617217102465176,0.6788010558857617,0.5621380625901262,0.3840988410408155,
 0.135384261860801,0.8441162749450491,0.8125335112797429,0.9859803650923769,
 0.06957886195658824,0.3757557326857808,0.8083353396443109,
 0.10424018518260314,0.818258863257321,0.1617194387777785,0.28169066943280985,
 0.7667021095325927,0.7996691382751576,0.9889097188782339,0.6410865814564621,
 0.8674450663893608,0.25713827899624175,0.9180984762606391,
 0.49526439121562427,0.700539163115362,0.6851567056545215,0.858457030667625,
 0.9491916667221939,0.1443456151642618,0.08399096739736434,0.4030035777424176,
 0.7509005262940939,0.12713852072440268,0.11945853844172005,
 0.15862239284362567,0.2408535355277296,0.46741787426464254,
 0.009166377853050744,0.5280742846843154,0.5297412071157116,
 0.8363682412679923,0.9258258378427799,0.49848140641991223,
 0.15632797954636202,0.7280309094563834,0.6541493360783767,0.6996072299275408,
 0.9226350494470147,0.5543814778816997,0.05780662403262127,0.3314853063373798,
 0.8383907469104319,0.5551541612659023,0.6618228690632199,0.9585020902531519,
 0.7551273963844414,0.8087110392897731,0.11477296699972861,0.9586174029086123,
 0.6147360048264718,0.19817350342078432,0.8974021358588076,
 0.29241393670756366,0.2288584017316312,0.7250741102677873,
 0.43169606513927516,0.814577168656025,0.25477445497019013,0.717613713279247,
 0.20085136507838564,0.6960153214517556,0.8680042257119864,0.8533023705414317,
 0.7580351158507737,0.7303710966942327,0.4547676939582582,0.09891644268906519,
 0.7457859183140683,0.38816263954725294,0.7494518922033746,0.5159716902868278,
 0.8062717824055701,0.20243149916576098,0.6874457561605825,0.3919076649206994,
 0.6405670685447546,0.9484726582714342,0.8062815056102891,0.6226956860401223,
 0.8622038269268569,0.13278300124106768,0.0452942804800317,0.621670561239641,
 0.1487198931785032,0.010067387102228431,0.6572851463269668,
 0.025318796445856218,0.027591547665161475,0.8257011258858519,
 0.856990158595351,0.9400208023965917,0.26152644571813266,0.06951349976564858,
 0.04782961289292098,0.6856743025029867,0.6701136996110166,
 0.45029111659192833,0.24935489569874036,0.25851975302918395,
 0.6177382993568553,0.9713711968890716,0.24597549871410862,
 0.05157209312105282,0.6551882774118745,0.8249702481389395,0.8556131352732519,
 0.2751245081660766,0.5441268248358319,0.30038404406883235,
 0.04110132505550135,0.3814367850550727,0.5027133848302239,0.4361227283432686,
 0.5876235705824022,0.41076524264942316,0.30047843524144213,
 0.8810809945300614,0.01942271052526212,0.025192367121726345,
 0.6686089249176201,0.9792644494292857,0.3285452243462549,0.2538937021997486,
 0.37812997647013236,0.8316328707028647,0.6289270940451819,
 0.08971640581960194,0.37344019250759697,0.0481580461723734,
 0.43399422151466505,0.2978683420254373,0.35383526115622876,
 0.12910307396086262,0.31218842569004135,0.18947186735335764,
 0.8810129821996121,0.9845089547693295,0.7374165510946309,0.6553848134721824,
 0.15179808584549537,0.9144644769803387,0.11355242561694401,
 0.13342822002310195,0.21297031648475628,0.06340663655170886,
 0.05602647637276803,0.39337451659071043,0.2911871485981744,
 0.6449278834448008,0.01768673372022389,0.2100031642224408,0.9972481819901823,
 0.9857598637213563,0.6789555209406076,0.08126892326196533,0.4670017236248708,
 0.9723141273251245,0.11208201343354196,0.173477811780698,0.4775999894402213,
 0.043469663733500585,0.1907475598571029,0.978627310610284,0.3613594025439447,
 0.7907523708548856,0.5120358164190809,0.6511843018132721,0.5168933819548059,
 0.773092520752616,0.08944919581714827,0.6611038164273437,0.7211154998431688,
 0.3965704832011636,0.2630418094412921,0.02715594901618612,0.6155698829121671,
 0.5789261030394237,0.915828055109844,0.6387102193368515,0.719592458096273,
 0.5964332683293416,0.045814625428891365,0.5553012623494284,
 0.7679813371436647,0.3705790140789331,0.5812452627835587,0.19125589865348191,
 0.2853214832444041,0.8372190364914932,0.844060825254915,0.8862917713455266,
 0.7806296434928711,0.24821733317208305,0.5364447738828074,0.7150335065702805,
 0.848479586277728,0.7319756445531344,0.6822653873765855,0.7848481933794497,
 0.034052210937609684,0.7855419081065641,0.7338759255099722,
 0.5101115595383953,0.8323011937980951,0.25372035562383877,0.7868600736335947,
 0.5294882072919846,0.020932842137015406,0.9710706877234051,
 0.1763737723386174,0.9227686220845269,0.643538802949264,0.2531558449243778,
 0.6855305960530232,0.396968715532936,0.638757455466839,0.9510569907267062,
 0.4197550843339608,0.28101658739682844,0.2122556513948689,
 0.15103976113887552,0.12477850234614918,0.21118023449459478,
 0.1592921242145242,0.06374470937520371,0.8289294706871483,0.3750474474022655,
 0.6682144794402015,0.051036717731962944,0.8708213742391725,
 0.6497808810370375,0.3709788123234067,0.11642494225418976,0.6760773228671523,
 0.6825705148620075,0.2872785557093639,0.04793290356848312,
 0.03617903265876243,0.001571663653953781,0.27850217435295066,
 0.37485590008467295,0.8415419355293515,0.16201942563750793,
 0.6958654757657448,0.78570570362914,0.2943797530939444,0.2817143658971012,
 0.27054510921968455,0.3278511671220561,0.7621245323422384,0.8266882465571153,
 0.9395720273832299,0.4236720946931475,0.705992490343223,0.25446096737905954,
 0.4288023876540614,0.6428554278264041,0.44186762817614134,0.4663640216314293,
 0.042671965751044594,0.5090298596533364,0.8412675877357634,
 0.8420877334476711,0.23728809750370128,0.5457631651980044,0.8682777454712549,
 0.13694792709634374,0.42966600465756744,0.5593383458781651,
 0.9238694076538162,0.9977579747423241,0.0031322873676188356,
 0.035949561659765816,0.3871267169848275,0.8410317130539734,
 0.25361209451753686,0.951153440547086,0.5940193120445806,0.9686555402998869,
 0.6423274911984223,0.35936247440272706,0.028577758747796667,
 0.36756138280151474,0.8692550808230763,0.7979029424632684,
 0.18165275294624428,0.9225494105376538,0.15130163834817734,
 0.5094058195371831,0.23804201433701022,0.9569207159170184,0.9647531719141323,
 0.21549443032595128,0.5754208073805858,0.6163271796700379,
 0.34137028436433137,0.6466906722152583,0.5487113697523617,
 0.031196957582982132,0.3610549322638301,0.8449084337128949,
 0.14292799653347343,0.792363452798942,0.8118487165856341,0.9563276020409678,
 0.5088806105833833,0.09336860203290309,0.5044956783918383,0.5053058683302261,
 0.11763890030140944,0.46672908564256554,0.1916129004494992,
 0.013235809932956255,0.18225013131336132,0.2576654096321135,
 0.12194492124228307,0.09923154996125472,0.6529785135242306,
 0.8107280934694604,0.6332606092609943,0.033066043580023186,
 0.5464769855651895,0.9792108822005873,0.19873963411279782,
 0.31828458253427183,0.19115778544090722,0.661416056676052,0.3496142105403196,
 0.30187623846947353,0.14790874185264458,0.7531885219720752,
 0.19500644818511725,0.2574450400409334,0.950363881127006,0.18452995729578536,
 0.47702118684615546,0.08062332525775118,0.7559853160544977,
 0.9964759775528593,0.9710976822108762,0.5996210718639076,0.5238123797969595,
 0.8410985655470842,0.23073916643432757,0.008300030971402528,
 0.9797240929654837,0.6759194873624623,0.7956981610762761,0.5207439133788753,
 0.1929058041570001,0.009290107360010125,0.7469072006650593,
 0.3001710811426055,0.2060678715174109,0.28957979357002395,
 0.33124731990364176,0.7269869971326319,0.9008662675146903,0.3763846391471428,
 0.4432444395436659,0.8401829286184281,0.5812487566647676,0.8994908975398953,
 0.6101160728466124,0.7858581140084026,0.7509223629987203,0.02339715291428257,
 0.1908489339782541,0.6939032278949696,0.6553583629916061,0.8333379506027052,
 0.7169427935716808,0.385869556039917,0.4189753686693928,0.16812425317220647,
 0.6535357035622681,0.43313191201525103,0.5171325505414534,
 0.41341949101295816,0.7109611895174408,0.6098448770170983,0.2658230099695288,
 0.6795802238807858,0.7166891522386502,0.6501057664557606,0.9944438246264753,
 0.7789387044164615,0.24197294663826263,0.3878948268785626,0.9617894305341322,
 0.3589680118878067,0.7175956405489645,0.997204975923937,0.28210645815920765,
 0.4439648241827452,0.803315458624745,0.49186661025399203,0.8017392220165904,
 0.33416927548621445,0.48911425180515944,0.5581087521125319,
 0.015101108511996642,0.6826787881303217,0.6086321002317233,
 0.20199686160858932,0.05506110055940183,0.11614917219912602,
 0.6409817410452663,0.62718019633194,0.14769693864804234,0.33127545249551715,
 0.42144602545427723,0.4117859469614342,0.45139184393191256,
 0.8199945217749486,0.9919840387639407,0.6476956053109129,0.9399274065891539,
 0.6898786593351576,0.02979775707298582,0.17946991240877497,
 0.9808423122108246,0.8343147997371047,0.7913220665565336,0.3358546877023094,
 0.5007950174603992,0.4771083424526892,0.23560048445490467,0.344395161817593,
 0.006934594321456045,0.9309729915853847,0.1712616868968222,
 0.6861329777143214,0.242745159993624,0.7906887210056281,0.6752338841998243,
 0.2864673181184876,0.587957546479864,0.5676038357771929,0.2962759815779481,
 0.24887908027795558,0.6046347186348018,0.8056486370813974,
 0.08695925650139458,0.3462015247363557,0.32617794427124447,
 0.06855542496929512,0.29599652759097494,0.8396655830671085,
 0.9005076372690666,0.5260202210129211,0.8549609352179354,0.47541811027826664,
 0.23171026787960125,0.5325077761130076,0.01915862695176973,
 0.6251167199234633,0.7772007070067338,0.43303846210024166,
 0.16773569166200797,0.7507712645224522,0.014649859730006765,
 0.15954355866167846,0.8405798673836197,0.7810073048227235,0.6939771683179822,
 0.22590327319721615,0.5243092842157839,0.1284217737630431,
 0.19801845637814153,0.3571246243271291,0.07337409468298128,
 0.32553809674542955,0.16028139529502072,0.10293696670846675,
 0.744017797014336,0.6422738776798749,0.8779018587328684,0.11268793706000879,
 0.8070608600439544,0.35136942681976646,0.22793081134383153,
 0.38903927090659096,0.2772854905164185,0.9284376298372287,0.6283113983691403,
 0.17308216212563887,0.7252839638142694,0.2293166909286617,0.7555060498127625,
 0.06879302508534979,0.6824389663586047,0.6619055183974216,0.5000051370957167,
 0.3866929822450744,0.2444781085294092,0.17812598345030362,
 0.025902162888587932,0.08234945379339098,0.26025862113075804,
 0.4524653251348334,0.16953508027891473,0.12951769405195335,
 0.7431647252438482,0.19249997291886378,0.9263064229227091,
 0.06247523842107694,0.9423930335251346,0.25687795980097616,0.771130847915036,
 0.24429568196877882,0.8183493503130862,0.009565899167001568,
 0.9916667279992902,0.5239887176959042,0.08605514069602294,0.547213692033278,
 0.24859403517541945,0.607102765177986,0.4672832889043731,0.4981852632640603,
 0.5049838649398686,0.09241098412085147,0.2818568587063628,0.867596759464568,
 0.9117574363349015,0.5138794706664381,0.026420131020408744,
 0.03221823069905827,0.979172851990407,0.24991070197196552,
 0.19212851530640707,0.6628374435929805,0.9440964696658685,0.6335187465907883,
 0.11055136852998726,0.7005744159900489,0.34610489421403157,
 0.6007062028173564,0.4469424177072653,0.40626109848965986,0.6756445990988523,
 0.02870602371805764,0.01593416397141212,0.40774287292598455,
 0.3963419362656422,0.4644765806253376,0.5800823075271998,0.34021534622109617,
 0.3704531381357514,0.24038975271058216,0.07077654856529236,
 0.18468027677019772,0.5463893620973426,0.025065825258774432,
 0.9763418068599937,0.4673847106682364,0.5798219420030746,0.4820843084443607,
 0.6409698640825306,0.5867256790895294,0.7259597768920959,0.6484577375916412,
 0.7441485051569606,0.6915784109784733,0.01732398511118316,
 0.22904270116333536,0.009501818855439437,0.47969072587845707,
 0.9769210711142853,0.04335108747555094,0.03864761024337238,
 0.3062513244624485,0.6899982714803223,0.10502515042247551,0.6092875635453623,
 0.1540451499281027,0.7485621990480522,0.31056668101772755,0.4040010387547541,
 0.8733278900449639,0.028576713393017705,0.5576596766430878,
 0.47180827579547424,0.25990681169378405,0.2898931021889659,
 0.9228823836982806,0.36147686624134223,0.39382660731330354,
 0.030092796484820772,0.8390616484875641,0.06484547972887933,
 0.2938112146265395,0.7204390015839118,0.8178457126753875,0.6209631141427593,
 0.8450392135083169,0.5994241358600928,0.04524310669978182,
 0.34987770689560016,0.27725909612307087,0.7687689445084902,
 0.17749579877879762,0.14762641828980105,0.40818387770217834,
 0.834726511142998,0.09410096923986666,0.8892125444326959,0.8253918648348627,
 0.11205859188861889,0.6279154340604444,0.23320940757807573,
 0.4099235545529054,0.6652979518338863,0.6524512727538632,0.17655124844311176,
 0.2680923879882533,0.9747516502172183,0.42747364332465587,0.9142770770340461,
 0.07784107820502417,0.35289172989055706,0.8812048672583964,
 0.5576484735662154,0.5805546601418161,0.2290295029299153,0.3637036185461636,
 0.5530265123223561,0.5409407080973221,0.07121705699244929,0.8263505768153258,
 0.11983876313939235,0.8147675082192736,0.7811188662169853,
 0.31447131941360407,0.6480347535113428,0.30912251161270066,
 0.11983343186929496,0.549381433666652,0.08628910134084244,0.2695141309113833,
 0.48721756476133893,0.3952396185568763,0.8443293697478201,
 0.28979343945798364,0.76677885608091,0.7708624838511116,0.3462313875207116,
 0.47476160271179413,0.48165445510575455,0.3981694237470941,
 0.30274161100591135,0.6523249488473306,0.1390265786215864,0.854972043052155,
 0.9552899635578718,0.6950390510332448,0.3954660165703826,0.33851051217450334,
 0.037070553155959085,0.2895571996601901,0.012721129232602824,
 0.7886518733357581,0.13461699371901414,0.9039122771934224,0.609848691607341,
 0.7915392040341828,0.4120690594647535,0.358716428225452,0.42502325567172916,
 0.18331744770395808,0.030604139246547146,0.1063834901728764,
 0.3705368876806886,0.5148806733014261,0.7347273777894293,0.6320011869611957,
 0.47310496080750053,0.024935869281860312,0.3788174621523135,
 0.9693444909950204,0.3615641604165232,0.7698172684837563,0.8533841057996947,
 0.4096594811206684,0.9107529713501887,0.7569760825565855,0.4663699676213787,
 0.931719459060653,0.6044346573275442,0.6810335579359305,0.790251750996405,
 0.08646562111775635,0.16303543634548012,0.6870980018420123,
 0.4968558799977647,0.4032355085802114,0.24414509206117097,0.4709249371562543,
 0.4339382715982085,0.3596003364400281,0.12929614017361224,0.2479151673595652,
 0.563963556430475,0.20947311543437241,0.37800181162094426,0.8155980993531329,
 0.7707657467700825,0.8172810902460753,0.04194993897022081,0.5987276314355194,
 0.4638736095304238,0.3563438636172732,0.3645268477915825,0.5275667924800268,
 0.21219534291980358,0.9320504710613027,0.6963981803369503,0.9764072927828686,
 0.3914178511379728,0.0881456764584303,0.020541682882366574,
 0.04029105092421725,0.17284615842412,0.40441156697352276,0.7391279866315764,
 0.5140409151525324,0.507792902326739,0.45841129553192717,0.6950630165991644,
 0.8018329299566844,0.4597883173698841,0.8430501902574915,0.07753212029141698,
 0.07616253682037388,0.339448260594316,0.449859772211342,0.07418512772037,
 0.7090785413982836,0.44264500550006247,0.43115958045261404,
 0.5869501503494764,0.5476300259809213,0.9251555879819953,0.09664825635548036,
 0.664653465796881,0.4320447975295263,0.75354576772392,0.5151887328178603,
 0.10331359460998335,0.9214740160391414,0.9295274882538523,0.8587826736664029,
 0.29327998693205104,0.8889491168856877,0.22600398179576253,
 0.6039380100205278,0.9606746948801947,0.6013488254393652,0.8600739373589181,
 0.7614252609131038,0.01781624953233829,0.7471751806192899,0.8165459216747295,
 0.06482501862757983,0.12477744627047016,0.5468416017455859,
 0.9026710556126176,0.6038512835464128,0.09879724039574123,
 0.30406017446353517,0.9872811596114167,0.2765479640506261,0.1777354406126237,
 0.6039078612725343,0.8788670043133785,0.7024618693467258,0.33210963407931604,
 0.5569983480317374,0.44947406425921654,0.986736077156092,0.7338339983803527,
 0.8404333145666846,3.6699378317339715e-4,0.9800957725886431,
 0.022624040905391274,0.2415614032370701,0.5695686116601512,
 0.8979191433155359,0.7906202846572863,0.8932697602713643,0.3271850327092042,
 0.971894103628155,0.026638583425792617,0.6858981886814441,
 0.01946753081343955,0.46652687630794676,0.8203389210991929,
 0.06556673466877927,0.38878308468418354,0.6391444720428396,0.713499511278415,
 0.026539981501960686,0.9901849998838954,0.75509682203071,0.946426449306899,
 0.9963029565614486,0.802146860417769,0.7073618255954959,0.10592695261737928,
 0.5140765275224642,0.013698562495561895,0.6620449369194246,
 0.7606429005304275,0.1362813963981222,0.5180954370936278,0.24647362945839546,
 0.9031184048964739,0.7961808384362552,0.5380090858626518,0.4846138339562934,
 0.6096590884075956,0.43143925558082397,0.47844385427973846,
 0.6744319198757837,0.6953738195973121,0.6214425642973207,0.5404651495508587,
 0.8973555118639118,0.8168227868985465,0.0803051213667525,0.6078876061812688,
 0.41197945607210185,0.08479255105530414,0.39489733102147206,
 0.02234660508525166,0.37194202071338656,0.34953645213305184,
 0.0702276577605434,0.5267424552759157,0.12270588990009768,0.3821504769466889,
 0.372625435410769,0.06771718911738778,0.786557815032887,0.1577597659893527,
 0.5673786980375998,0.8202184687582961,0.526049167057755,0.5161643889321385,
 0.434837820007725,0.5851634334173224,0.8221490265869525,0.6679859632141244,
 0.391921945742996,0.5522216518680044,0.4939426384742379,0.7111019833995784,
 0.9636149206848216,0.2835662292382757,0.9573767302233263,0.7212561415877596,
 0.4315289214751886,0.88960767957039,0.14506526250815177,0.3762217897567365,
 0.20312392152264702,0.5620936060461147,0.1067323697402377,
 0.13775972444448914,0.7505829628828224,0.8332004475217087,
 0.39930755591978906,0.02855597970717383,0.029322700995993545,
 0.07567993182983668,0.9930879246593347,0.533071516088901,0.5820480890309645,
 0.332300719444387,0.983733173980724,0.5623519583942377,0.6271585922548061,
 0.6554552524118594,0.17611676280345778,0.025757329768271053,
 0.6414766894465461,0.9114986518075575,0.704556845104167,0.23128527966477785,
 0.43025861557918044,0.951647920889752,0.9073195738908776,0.8925131594771184,
 0.8710075984186788,0.08569813364284484,0.7352816659298641,0.71535758902765,
 0.7548778999676037,0.994238171971365,0.46366259351230066,0.5121341701493607,
 0.2325598233033731,0.5843039894888242,0.8608121792024178,0.19246746810649862,
 0.7317694799521774,0.8307168358240141,0.22679794548904963,0.4403941446108621,
 0.06147313367975504,0.8305099929148948,0.989031029664075,0.7309880666278813,
 0.13334816601618105,0.24204486376921153,0.3939327479915967,
 0.1602422913527426,0.5958957319502677,0.394495983706013,0.035025396810478115,
 0.28141490555038506,0.263872932591387,0.1357922189954981,0.7105827606187111,
 0.481076548392399,0.33894292265243164,0.5911684466067681,0.8523705709338394,
 0.8067317159103078,0.6014988943288118,0.2711330810295248,0.03782339805122703,
 0.9621501703565408,0.8480988206759745,0.671778759537244,0.32422606976979373,
 0.21473116671980297,0.27715630456523643,0.09009761003138195,
 0.1116088639908307,0.7531159271190615,0.8692228559543714,0.27920052014054064,
 0.2135534113693891,0.5065164962597357,0.08031569795441362,
 0.41557722698768484,0.6695385588346592,0.2018770010279628,0.4047298529464066,
 0.3055872702355108,0.6283874467945345,0.08111988950299032,
 0.05899060661251543,0.8023833267555345,0.4703949463364244,0.9043542186433863,
 0.13120914826104935,0.5829001653986041,0.38769777413497575,
 0.7185931880547466,0.692054234446084,0.20017040738845604,0.940825130710395,
 0.4372546884715507,0.09188051914261453,0.20413706269915932,
 0.1551464351300994,0.09247901056798313,0.5571899440223173,0.3287917799596707,
 0.075773348948179,0.006165581279763521,0.9816886117391691,
 0.49678710943247273,0.31976170235534807,0.3606177012086713,
 0.7003926393045475,0.9684755019282042,0.3831216012286859,0.2411964887396576,
 0.6792228587961704,0.07593799068853624,0.2500231622512211,0.4673758081340982,
 0.46079922543110174,0.5041351613612739,0.992962652777519,0.34811740227190346,
 0.8364146771349013,0.708718909812021,0.2288595545787857,0.027643030559111104,
 0.07214012583934926,0.9628859841869981,0.7898571719137764,0.8325167936308873,
 0.5783904729018943,0.5586451277933869,0.14072839743584553,0.8476075932175295,
 0.14254489832967654,0.4290916886105909,0.4865925748741272,
 0.025516615661191566,0.3807945502190465,0.9366172846106394,
 0.3698338923211256,0.9153223511323261,0.5980041310597078,0.6057613957322013,
 0.18462393681214107,0.6994530247580772,0.2724291326578985,
 0.11953096804381136,0.29309904875044546,0.525211952267146,0.5259026488675899,
 0.8905404844460184,0.4393507628884983,0.4273574076356007,0.4712001225087489,
 0.10454325390698382,0.8888715300420754,0.18654143629891906,
 0.12058119712282445,0.5596852135052127,0.9765600576611443,0.2765532936207379,
 0.2031767958726255,0.034779844884494704,0.5842697663039031,
 0.31072789741729456,0.35796172957298766,0.27118600901931067,
 0.2681542825170956,0.9596227024477012,0.4469531331189722,0.3726812445617004,
 0.9151893653563173,0.1715375140901959,0.8668393024540642,0.4128296199550858,
 0.7941073514822174,0.22454493587964064,0.39120774772464206,
 0.3595066875320734,0.9945530272848552,0.7912193529506716,0.4163208650716551,
 0.07981490136475244,0.9904014163963194,0.8771083074103048,0.443456116100883,
 0.07782276894984097,0.5275742054900319,0.38090555285675787,
 0.7319604792850528,0.33539506586189227,0.6389695778623472,0.5791144600045696,
 0.32699067965899753,0.30026602131725266,0.7585950700138481,
 0.6113530043446983,0.15092515945213936,0.7410194858204358,0.9851256949760255,
 0.830357507974978,0.11702916907819627,0.4953140236193383,0.16932132009577006,
 0.41746920140917976,0.11635261561277954,0.84555700491521,0.6582875548182184,
 0.9764410194304795,0.8219881256994269,0.790848324021464,0.7016353114569809,
 0.6486537753551151,0.3080962647520412,0.12248173893395187,
 0.11766082451649096,0.161047886695053,0.1846334499861677,0.1873246176037825,
 0.18292650367915142,0.4976877495069738,0.38558396275462814,
 0.8426062483228186,0.6525924581720168,0.9540322656462549,0.3974898737691853,
 0.6060583206689334,0.44870010231261326,0.5237624625622148,
 0.24390476129521943,0.4027556750091835,0.31850865014914065,
 0.7221911811765516,0.09274579669622829,0.6693738174793329,0.8749475481005966,
 0.1606202791764868,0.42403296298407844,0.7433453135657532,0.5757426711385072,
 0.6320826746627753,0.6217668865742729,0.6476587875946233,0.3130906664393467,
 0.33142583250030466,0.6017752336627181,0.3853374789840627,0.3382042424433792,
 0.4884411210207773,0.8026610967478447,0.3697110354312335,0.43504848794804707,
 0.22661896589946495,0.17191900114194514,0.2518652273830122,
 0.7415016139708871,0.38446275320236856,0.14613055069405556,
 0.4688160540586104,0.9919051716746127,0.6497245418233282,0.44055487563908646,
 0.30296818695959915,0.9526253195998984,0.04323891703329685,
 0.01352285771764139,0.3560925063162732,0.5233046682768121,0.8171649927840979,
 0.642841017447891,0.5256718674897763,0.011280025062838095,
 0.047272972003727376,0.0879230046550491,0.43023731944415433,
 0.41551617895266646,0.8974761530354863,0.2587957607653444,0.9307635084703579,
 0.027487791136039275,0.9777939504576091,0.626483961557837,0.7196982780493022,
 0.9361370072100826,0.2650663760623031,0.13816665586944232,0.6709145126959812,
 0.2918660851597672,0.09743491915567193,0.41043257022441404,0.896526143058703,
 0.0068646554130877835,0.827215792232993,0.9668315510795268,
 0.9841341082830347,0.7539905809815725,0.5712854417874104,0.6996456471140761,
 0.5567916852230839,0.190236309949801,0.46786032236052044,0.5259925585772298,
 0.7971881028234782,0.20065134768738868,0.3444646470420255,0.4891849510464985,
 0.0196200625312567,0.6634027362371704,0.9451906336254652,0.48270638848516123,
 0.8884517906775997,0.5266679861498045,0.21614165395451912,0.3906538059175446,
 0.35300296630873773,0.505115797029305,0.9252666354174908,0.5490989771637693,
 0.022414240585764023,0.08171659688617705,0.5118794735143841,
 0.2831464005648068,0.5543773787931849,0.015229619466372801,
 0.9496968932264872,0.7289206056824419,0.09478732311325344,
 0.11929592440948633,0.06270158516673652,0.9513719366370028,
 0.5735322077035837,0.2507218439652915,0.7876665574485707,0.9495425393885135,
 0.1990493601793797,0.08899346134271158,0.22299762806902257,
 0.17870892881273848,0.5306999677020726,0.9513561887920696,0.3305787061573304,
 0.5234863606195824,0.21218814919668794,0.0014031458938306152,
 0.879792384737415,0.17498099473934114,0.7879456488965534,0.8231685360330545,
 0.14995324258877474,0.9144614483742988,0.9046900324889583,
 0.06188641061377209,0.6286628701544426,0.3821042825409672,0.2676518249391703,
 0.09323565525054056,0.5471654918720844,0.8751042225279086,0.9954520799600969,
 0.7482604592045362,0.10944041160861218,0.2753791158897695,0.8556496875334205,
 0.38184684361995724,0.7539032557884151,0.5937124727910053,
 0.30629575572305934,0.8306078072947812,0.4673196355445096,0.6094047727508601,
 0.4074226565648216,0.9700608402374601,0.8673715220868357,0.6941469267074014,
 0.3229252422692187,0.5662071390343877,0.9353002119795756,0.3087832791202013,
 0.07615969923437083,0.8171418483668275,0.9725513594187976,0.4723559367961597,
 0.669066545426965,0.36402431624965836,0.6605824321879286,0.5908599768243445,
 0.835265545000802,0.8242075560892703,0.5276279701877976,0.9190183459643801,
 0.7059815919835756,0.8033546968857567,0.9694769438088257,0.5826851374925375,
 0.43217077401739024,0.20259774567011768,0.2492117196360637,
 0.8752019372422546,0.566040102886874,0.6179483621036378,0.7720061409062048,
 0.9625478696918677,0.12846557596805608,0.36220938076912257,
 0.9610688096177927,0.5155803011979231,0.9829235762583791,0.9434964941059156,
 0.02402345033167219,0.3493754722616782,0.4552731065160621,0.7672114279538235,
 0.67474373092703,0.5772468815929146,0.8262795019261541,0.4578882791669455,
 0.07250892698227518,0.1749309434743327,0.9327151394189432,0.1737831331198142,
 0.8882298661875315,0.31951133675611243,0.6115148220141802,0.3056931390530393,
 0.18136184663667287,0.6414952645684269,0.05240881077254256,
 0.6243272524901352,0.44652067575203713,0.06832713299876869,
 0.5247131205583171,0.8149171359574224,0.8177927690614787,0.1240268832414867,
 0.6623648961037537,0.13933267113147807,0.904581826418019,0.6036988431878443,
 0.6297047343775923,0.4928684420496894,0.3969870400236416,0.20010597398812113,
 0.0863733871660679,0.46275707060771587,0.9367117720200008,0.1321748844687376,
 0.14155512043345586,0.2881455940156561,0.9654849142353857,
 0.22228982452758128,0.8425278103876902,0.7546954335300111,
 0.14766793572623416,0.5804697812422666,0.539566671875287,0.22599052525755936,
 0.6119116295349994,0.13958259521325989,0.4994171200281533,0.3511389461307124,
 0.6952233891782545,0.6040310496454667,0.33802277705027184,
 0.21784715854792214,0.49652685996620816,0.41645209409688755,
 0.8765062774203934,0.3871870707494598,0.2448047516670836,0.7055953281506042,
 0.2776081699699243,0.0990978402492586,0.519212759694693,0.9541704216735396,
 0.35546597075376907,0.7456795808789292,0.641686591073988,0.4553032563955284,
 0.850402115758435,0.2969574250237661,0.6647283434963557,0.29047165545423115,
 0.9216250776466723,0.7447492497776734,0.24678732306407314,0.8753893141686268,
 0.14678440843940987,0.12964428319121546,0.16177080183886183,
 0.9055407988951522,0.8343940161706351,0.06599243029838064,
 0.01146855694992921,0.31564024302781835,0.8357898155601678,
 0.8031919785943402,0.4339551543293122,0.7746746651292353,0.9051107794871543,
 0.1269386038526077,0.7196023732547445,0.554845046128213,0.7927343766214988,
 0.7441631775582173,0.8441968743233178,0.18644237612039594,0.7602763065183907,
 0.9713874355480847,0.036829627596811854,0.29528352932583957,
 0.9237302579921055,0.40380370730313164,0.5613040302214034,0.5299301353483762,
 0.8688370256539979,0.24236055767764986,0.6013010326908773,
 0.20883998872618448,0.07710325530995787,0.11544606726673234,
 0.28980438994853197,0.9526023974519573,0.7349270415539424,
 0.04765291698601881,0.9622711911199832,0.623477656656877,0.5971964630917025,
 0.18086696199355357,0.3470047973319712,0.7645853809135897,0.5460823456349012,
 0.39154247361368766,0.15190836553057685,0.1379216447481867,
 0.3555892468095785,0.7351548267718668,0.42528315121169535,0.8251335325525194,
 0.9802294216805221,0.344624836231032,0.546877193794959,0.44636430032279273,
 0.605275617265223,0.8869583523756513,0.7247793454046971,0.9550617722817158,
 0.8025616845660734,0.28474733451653755,0.441748034275223,0.3118912249653162,
 0.18384506499766573,0.327330892513751,0.9107709826662641,0.3025287895115998,
 0.6766633966474221,0.6276656649983934,0.8541628199438007,0.05075922831676338,
 0.8900206769643076,0.16973109939049713,0.6325753072433082,0.7983378487193672,
 0.20459966780596595,0.13959764086657622,0.8188774962711317,
 0.01406663282246956,0.9478551342892075,0.8960348885132842,0.7676926001277585,
 0.22131320988811676,0.6842397118454095,0.8159288511274158,0.8057019028555982,
 0.14234249878208127,0.7655526148238267,0.4803657661888121,0.5303376721055637,
 0.5329853945810982,0.2976450737663128,0.9565605095195729,0.12430341215171903,
 0.23187303680652915,0.48198332277206124,0.3129282759707215,
 0.48899710112132133,0.7222612900569315,0.45856492227727585,0.494020345656653,
 0.9241029062238909,0.21174479673608482,0.15058574950650827,
 0.4701161525281379,0.0389985429277695,0.7979168173334831,0.7329921285787213,
 0.8200179566718178,0.9672239193671238,0.2311467661704799,0.2388475497011331,
 0.7395774791656502,0.4980916682581702,0.9277894033618264,0.9154700868705875,
 0.4080717434357194,0.45718239637810654,0.3022484875379412,
 0.32667410833256516,0.25234229469862146,0.7339517691322399,
 0.14312721297010644,0.1976454201324196,0.04848639282688305,0.617757280621593,
 0.5111615395802273,0.30487975827039693,0.1409992205716456,0.6635954305479456,
 0.9541547777852237,0.15380191581079194,0.19123636091402385,
 0.5722865971952902,0.3319180895396989,0.3481954925521821,0.6077535565585961,
 0.06904162051982732,0.6295412998503964,0.386886001140177,0.6581031578530607,
 0.6073742533012101,0.654148347022724,0.4992636950398579,0.9881256724895954,
 0.4228380610659259,0.9461088110425299,0.3398401276750267,0.04358869696428003,
 0.8235199954634936,0.2216087775071821,0.8282885780819884,0.6482288815092023,
 0.7729880175082242,0.6523389563510142,0.5680619647629581,0.8459653308316992,
 0.7107095860700496,0.6117787087418617,0.23219438979930285,0.9176105576288678,
 0.9680896796527576,0.015304346596300533,0.1797580619531245,
 0.4495920135130975,0.30083995508735595,0.1565291754844903,
 0.024226545133219757,0.007356088374242842,0.5707092013401426,
 0.6269162269674344,0.24192102778444924,0.9021584618259977,0.7960804974355127,
 0.13526451907959314,0.9299246030246336,0.8200986925504667,0.7426217841198427,
 0.4888323682854706,0.7133781487946274,0.4100015254332935,0.3999430300471857,
 0.6430996140183293,0.5863096081833021,0.43853847693797476,0.5613879491578354,
 0.02794668875982642,0.4359930031957088,0.4441018628505652,
 0.43762886800294865,0.500952360123207,0.6043809424166039,0.22312695285447715,
 0.02371617912277524,0.012429033539737344,0.03762164284280756,
 0.01562561591973677,0.4644703148526884,0.3537631649823153,
 0.11498875597037395,0.2635382652004873,0.4880744252383453,
 0.06267511825617378,0.5132838122881461,0.4447290947972746,0.3423018639263423,
 0.7200827296527463,0.7030150833138408,0.7801131641525922,0.5274985712194732,
 0.5305995840262188,0.9557530560512103,0.8009577790401718,0.7074269811740396,
 0.25377888213193134,0.8653547747179264,0.6111265346539195,
 0.36926069095338043,0.16771276644265587,0.38797074933259146,
 0.06698222086897432,0.6189964090679919,0.1921383424279961,0.8564136919705654,
 0.49179356530274143,0.6631916729479082,0.04135263444730808,
 0.4211040552118148,0.033187180050502474,0.8696257103573625,
 0.08273580359361454,0.8123714422711223,0.8417998768404017,0.3983646356555315,
 0.8756386625747036,0.5193095195566908,0.7847531444982551,0.2681835840984689,
 0.5870023808502594,0.6367074492692331,0.13314642476237792,
 0.04426190244199801,0.404652745596885,0.8106932284643129,0.8270347173873448,
 0.22293385165300272,0.9768162915443585,0.9993272812598634,
 0.32605851076801007,0.21886704956495984,0.04534182148498178,
 0.4583773827315515,0.6592306512916484,0.1681860184825552,0.6172942163051898,
 0.9567703273564003,0.9268149731148115,0.4722646927171059,0.9533177941132202,
 0.5153000329149385,0.8780452418687542,0.8271603274863323,0.605012032053718,
 0.3989849818725584,0.0803583738143886,0.45582629620834303,0.3384174893706474,
 0.3718241344269243,0.14346701789317318,0.8112784708264008,0.0904721073652246,
 0.5175011076177911,0.8274857438444321,0.9730607148653965,0.932522596705182,
 0.7444531842596049,0.9808633947592607,0.5595017740919904,0.6124895809415846,
 0.6917765857290686,0.30686847859495525,0.019736371578553236,
 0.8916127460106507,0.2028473867401871,0.5294328288698691,0.5041803060962352,
 0.33795542781439536,0.11743934413431956,0.5611868821426018,
 0.8182210503640416,0.15488431422958304,0.3985316523642113,0.7587670962396246,
 0.5298652904889275,0.7936468713009664,0.3083321546784604,0.7682417265565928,
 0.24212141642466034,0.04118765614865749,0.9677140638098036,
 0.7872906846138493,0.3706987887686508,0.716307260984189,0.038054013870248315,
 0.7053816248766728,0.5369497693176817,0.2086247796595796,0.9933213716480289,
 0.7978599156329633,0.15770004618958633,0.48599160280543086,
 0.4187958300746065,0.8634886384677661,0.3011959182953362,0.6719429342436566,
 0.3739041934880801,0.9473465284910517,0.03617576709533388,0.4865549611483534,
 0.5134529900306526,0.6952522065097457,0.6172932253003904,0.6774893602447659,
 0.36721032814052756,0.4855050039105081,0.08162564267018646,
 0.31184977733288843,0.30789142452630847,0.7723652077454507,
 0.8699125121633183,0.48167952709754547,0.41945227154374887,
 0.8101690778629442,0.3306685153044593,0.737756683267918,0.3985168127800035,
 0.8187697642981107,0.4770029413531488,0.04523827120362045,
 0.14328456930218836,0.6000971095327593,0.7768676931968638,0.8588844009131936,
 0.4006823223263778,0.4728646488425813,0.9113224953619027,0.4248304196838443,
 0.9567028443172683,0.9771402909831837,0.7186676617182218,0.8490122125709486,
 0.24113981984964417,0.1424675388081904,0.6101082953340247,0.3499738069326235,
 0.2170685424460943,0.24281423848900496,0.8535257033810117,0.8179529851133596,
 0.24019376425446626,0.2538491912514784,0.9630874817628923,0.8677976725854322,
 0.34196804474685305,0.6866494739473895,0.2444464470055967,0.1597536779117601,
 0.14343091163419952,0.15174427875530416,0.18686068942529477,
 0.6376091552391573,0.08652686336600979,0.3808810731187847,0.8801016294509241,
 0.29321142687929946,0.6928773146814073,0.5773767591340159,0.2674949801583345,
 0.7512070986914329,0.5871104666431621,0.9305896381819673,0.7120756746827057,
 0.3567057979040852,0.7454692299764724,0.7658513905295072,0.9834675207185224,
 0.21576636804024163,0.5861049190382466,0.7549486502206153,0.6004490423549202,
 0.7011171567525222,0.10139287594725865,0.24334253325184596,0.770989870357092,
 0.4656202899812416,0.48234102108221677,0.2722246194751343,
 0.22789127273046939,0.4241309296484703,0.9103520620511313,0.4179510746075382,
 0.957821990442213,0.6062736852651845,0.24893618589011557,0.15766849670718042,
 0.6414976399705945,0.6189529003406282,0.8480942188902326,0.09910494329359754,
 0.018222067841588308,0.1476566747276592,0.4996625474731201,
 0.7846851899174674,0.8127522494659775,0.5060424889621206,0.9753872337725977,
 0.1450682997572883,0.014031139507030277,0.386896834693529,0.8900208010018478,
 0.8003026779379488,0.1693657612224897,0.752555661457976,0.5309630634659404,
 0.5345284753565289,0.6017899327206189,0.317845394974438,0.23276322115361858,
 0.0970410620396086,0.1626262019224216,0.05790017292026639,0.5871277935432119,
 0.34694604331932244,0.26458714981196096,0.6798949875642356,
 0.9471492421830223,0.4960195563497448,0.6399964984604124,0.23589663269742844,
 0.6148305038488067,0.5346056632766798,0.7840988135507004,0.389075317804476,
 0.2721888288184364,0.5300520449525601,0.8052168508060018,0.982844489026706,
 0.2583008522714828,0.9581312247688685,0.05013888137115852,0.8491853449329498,
 0.015061431159542171,0.9160209573800613,0.7136334474105513,
 0.7194163791268947,0.698386673213642,0.24193274440606227,0.5232668075972942,
 0.88076377813615,0.9093352375469638,0.8499124254037697,0.8548410079949478,
 0.7299434513071396,0.42814118163842096,0.01937055439894042,
 0.7553624193694003,0.6504238714524003,0.3597328221992623,0.05384156958077402,
 0.3910525833664946,0.9421437909616861,0.5580505030308035,0.255463098762172,
 0.9225968854043357,0.9330657462709883,0.16535083509323556,
 0.36691522110801733,0.2533996808058512,0.8812466931604803,0.8559142949831625,
 0.5745837143469419,0.6423440485186196,0.8513371101693352,
 0.026954402100647812,0.01481319147132254,0.5110110491250502,
 0.5034392234523569,0.5046213206749673,0.7581015371327386,0.5499092824745834,
 0.7144583969415149,0.5195264718877873,0.8058455442416814,
 0.026644262384497686,0.4653121541102556,0.3079640726079994,
 0.5196547820523817,0.09168770621421984,0.96719847090124,0.7761010777472774,
 0.42041715809565283,0.7299501142835454,0.3974726637412732,0.3363332249733809,
 0.5909727131117484,0.9686941348687436,0.2824110279334776,0.7898141067104177,
 0.09504101541844057,0.6591374955268918,0.07997531281647063,
 0.3570228751415917,0.9475730781317597,0.11684517628695046,0.5716260155017211,
 0.32912472358221745,0.850399235658406,0.2193792782589019,0.05603874635074424,
 0.1767547548358629,0.9950777674905555,0.6386911066665552,0.08156678385544247,
 0.7904820557883323,0.9198238458522349,0.8273192995754298,0.31904038897105547,
 0.6389150920149645,0.4402794557367532,0.895297777880443,0.9634750709317372,
 0.748997047431156,0.26841642500340224,0.13895452392818053,0.8994026965516491,
 0.659130228377702,0.10257434407802435,0.08282575805054104,
 0.29510946675959815,0.6537327405806101,0.45469385266672824,
 0.13339033037432657,0.41568821004451895,0.04762427956723858,
 0.18736008577612207,0.3514647644372698,0.446257705872134,0.5529674177963005,
 0.9735249826303122,0.5922425805332854,0.43337243695441086,0.8384714827555129,
 0.7377253126009056,0.830684344382208,0.5965507457834363,0.051753671871102136,
 0.4675267579884933,0.753274063119595,0.0636523077370239,0.3329649447929707,
 0.2661141758666038,0.4746647387468186,0.7600793079158128,0.8473318516997937,
 0.8181573727148289,0.48633234143833126,0.017337362793330757,
 0.49870557189644227,0.8389990945519039,0.09013088262561031,
 0.2228984359461823,0.5049578808247449,0.13572270721797763,0.9087264166439144,
 0.20288610161859377,0.9520422228230003,0.7248899331562374,
 0.39463858188389267,0.21920704867643104,0.8505445699676,0.664012622939453,
 0.6667570603349661,0.43191161948840395,0.8927384831002427,
 0.23632877429937615,0.9507234498176823,0.7750130403507739,
 0.39522508969055314,0.03517688214019188,0.23982314972720808,
 0.4113723087398288,0.36020871056473747,0.9552909626265923,
 0.10304114894174254,0.9178789912467165,0.6303521746705969,0.4845726592184971,
 0.5302246254345959,0.3612785325929663,0.9025878849941604,0.25534862419681825,
 0.09850384185238636,0.2759535782950763,0.6098440306098334,0.6528503824537009,
 0.9386999550182062,0.45752254351338006,0.13695834675282836,
 0.2517307547952431,0.15122003163867537,0.486159446833066,
 0.006376921430051863,0.12243459807237989,0.5020312837115962,
 0.4942061441365573,0.5693527544857209,0.6046638713880546,0.1790259721563483,
 0.2813320663173017,0.27852694803800404,0.43533525423638897,
 0.6074313624333533,0.8131405221258271,0.8152363912407053,0.2738221934338628,
 0.1783094739075911,0.48191790960808645,0.4734349861994436,0.6388648046017041,
 0.9770472259326336,0.28850263186931147,0.9051166546284304,0.8353597629228098,
 0.6961115737191711,0.5121164859759957,0.023890558332842482,
 0.04337256657415001,0.8050237997058801,0.9709678042226584,0.7631137030739026,
 0.11069003287017387,0.027054388279763253,0.3383868884678025,
 0.3451365634044028,0.7825827360731678,0.7044474397294758,0.3723198251436364,
 0.8658830381990664,0.8860726108022937,0.1368558299315683,0.8701332548630396,
 0.07824967652430725,0.5815569764559008,0.6841926933689543,0.3333060473838566,
 0.7908536897022599,0.8019696005799466,0.0846160660439419,0.41293627553121537,
 0.9448348657892034,0.3677974493089079,0.22183884288025202,
 0.07037887909773877,0.9310297762953852,0.8254831435290121,0.2280408762302465,
 0.4160207296712555,0.029847205427436574,0.3615793094497797,
 0.39346399242062935,0.47450111536675543,0.9222061556432365,
 0.47604966932342363,0.8218768385250671,0.6814463690946672,0.3793130366212769,
 0.5059887962704135,0.3641630123900246,0.34222402636628013,0.2452475841475814,
 0.12832382318289803,0.5743777557939507,0.6296351962045171,0.7919724506971922,
 0.45200474186716655,0.8239668069479883,0.39411035006744743,
 0.32615646364414297,0.9328879861251657,0.3168958077960824,0.4894698740995733,
 0.65267508067846,0.7101892392060698,0.4712857685786136,0.7741304706261894,
 0.7254045626278889,0.5041672393494077,0.9938272883067836,0.4707992968360909,
 0.9251477097485585,0.8354320793438607,0.059100574198621136,
 0.8470493502252942,0.5722230751499897,0.37982523432636317,
 0.13722766749668547,0.5766341361806476,0.9354797723535289,
 0.33283773000587824,0.25445320854149145,0.43767369648979615,
 0.4079151006697508,0.3658508479134449,0.7244403839636383,0.9740698281745312,
 0.40581174047029545,0.8346157107258682,0.7230162242044241,
 0.36567676331458554,0.9959311869193479,0.8896390025019695,
 0.059236837791420305,0.59074196026095,0.4643512805517275,0.42053375050350417,
 0.39944931383287063,0.7767855203472834,0.4957640296854574,
 0.46028816290546926,0.3919163050585772,0.33559567310564353,
 0.8430631953341569,0.062429260487095295,0.8994267195149128,
 0.8439313992373987,0.10172689308940497,0.9181942415955606,
 0.42766716202988764,0.6493973181632129,0.17310243455215468,
 0.11172152012413916,0.6839250320707512,0.5908314540134443,0.1564999134210976,
 0.5899976821586058,0.576979835578941,0.9822102135885773,0.1369239966809288,
 0.41414293583356077,0.14708462380305443,0.937048099621543,
 0.35292405401923677,0.6646734395367442,0.8904829216993454,
 0.26494373702654506,0.7298651597936466,0.1162053767916239,0.1829407773460564,
 0.2941569959123931,0.8260850896260226,0.40682202788637956,
 0.33323295531733177,0.4911262420274579,0.8910428213404173,0.8678614374828384,
 0.8619025356953245,0.8462644195591218,0.979992314834477,0.0913147998421816,
 0.15933080947300227,0.8665245622672434,0.30501929084870416,
 0.6658472660822242,0.852917210364581,0.9038102019951421,0.7120145352730544,
 0.21229908951896004,0.7309483067667542,0.39026785200476644,
 0.5366784367602311,0.9787900178927614,0.2442278983052969,0.1338974364273119,
 0.6892852667561016,0.7410829380660945,0.5746554926186977,0.7664735104313455,
 0.24280252894288024,0.07323707908825394,0.8414428176301828,
 0.6229196489908512,0.322340604544721,0.5879289753868627,0.8219607260509344,
 0.10338726101039697,0.011866689139802822,0.8648665485681901,
 0.2608957688520007,0.0966261568132496,0.4107739361358411,0.9451789238906595,
 0.868882749901824,0.06775939228551897,0.8265449634887949,0.3511457843140532,
 0.8230675949472671,0.6379109491540712,0.4492729427699135,0.6534996558080604,
 0.43953902341280016,0.740681056155826,0.8342549638108278,0.41577401444343065,
 0.2694997212994469,0.0012201457312592812,0.05734191490816776,
 0.8007271832201538,0.31363129020460845,0.7455620919477192,0.9555989694509889,
 0.17285175068494074,0.571071043494603,0.5730578995951453,0.24074521723226439,
 0.34201371903375544,0.911233285379569,0.167898762974663,0.6834367753159624,
 0.865126396278439,0.9349009872481053,0.5951945628313359,0.2199976575733411,
 0.6271546245756865,0.2231450876147082,0.2498421319034022,0.9107542689923731,
 0.8770604644350092,0.24079800530656448,0.49590574920741415,
 0.11161767301940506,0.2335493079346288,0.6651148617459506,0.5499233657257275,
 0.8199073303658442,0.47799875442980144,0.8445527786900512,0.3105119704200108,
 0.8394603233151461,0.5124146238204799,0.6720160138875267,0.25441276134341206,
 0.864058451057158,0.6491427290649275,0.19768730896571662,0.4627258469713653,
 0.35523130667057157,0.7856871018960866,0.2091569245083782,0.7385959707361576,
 0.6524268176634457,0.8325498483471948,0.6738685027548155,0.3634112325611022,
 0.6140804928462299,0.34238918629334614,0.11725559532868224,
 0.6691518028332893,0.7965617947179782,0.05805713000131352,
 0.22879885599338312,0.10188387380598019,0.4572959952174268,
 0.7497546926334273,0.6710152175846058,0.801299477103261,0.706045221708919,
 0.7768333671891183,0.9602522815273055,0.9408355620995492,0.4991230866927121,
 0.27291209718289844,0.8836465918104317,0.7277530972324356,0.4801435422458751,
 0.6754567215353976,0.04228154841152565,0.771352437187961,0.293219581330426,
 0.17208532510123653,0.32576619141929475,0.8403821950610896,
 0.31513943952054657,0.0755952389943747,0.3199228929446183,0.2639776509290359,
 0.2507079858717951,0.7373241830721855,0.7524671308878339,0.7266473000460028,
 0.16231561068684164,0.026402635909923422,0.48623062103468784,
 0.994652258561055,0.7157703176202745,0.5248420123965218,0.5040271795288866,
 0.5594363997603296,0.9640977158145599,0.30554985559427716,0.9662576965179439,
 0.8442685678777815,0.4005601996109929,0.6102683382618695,0.5890772612902347,
 0.5869691959053934,0.25483420592894246,0.48156719332081954,
 0.5146111296612069,0.043224430444052775,0.7932414454860025,
 0.3022992800992922,0.45303518108902296,0.4439303807947673,
 0.18224412709794435,0.082725918395403,0.8764114300006182,0.7369509388905127,
 0.6529353305508502,0.5663681701963299,0.14501030032471163,0.8153777161313401,
 0.9649294775389695,0.7684742558981664,0.5159885883519713,0.12252854931956403,
 0.9381843461753516,0.9289558448758367,0.7290024488423705,0.13207178106905104,
 0.027319629126191236,0.9486922064291717,0.4531064569845129,
 0.6567096567879066,0.38698703569923976,0.1236699368296178,0.2617100055919819,
 0.49319687386624,0.13157508766737747,0.24898320774591498,0.4804262437945599,
 0.8153352684225349,0.35362534844318816,0.06115484788072001,
 0.6452114574299945,0.5142675055803541,0.4538614872205141,0.1344053104369356,
 0.3348715037506199,0.32145318087919894,0.39949500111629244,
 0.6695868992195618,0.11143822723532715,0.3757308596278026,
 0.10758324483146242,0.7709248833062983,0.9109445151448893,
 0.026126199778415904,0.5962424367866018,0.6204279132760877,
 0.9160463428464327,0.7715382200506542,0.8881569182481681,0.8328374564259736,
 0.8183162974123407,0.5391340728534479,0.9822506530572664,0.238803058901076,
 0.8764825281790678,0.3018307543860562,0.769036843618433,0.6079296954766478,
 0.45584104392342517,0.14542923101243077,0.7897455574953275,
 0.3624868154912433,0.8770680942626451,0.6478942155803235,0.7079138503641524,
 0.22898020933403973,0.28535441341836776,0.5235798607263811,
 0.5991988688570177,0.8237561509786647,0.4967524281456077,0.6415851519638387,
 0.7857544635596533,0.5797288052376223,0.3014263916077412,0.8408352903752364,
 0.3197697257757286,0.5375041325324168,0.024090237195404596,
 0.7397878190705496,0.38644675881887514,0.9751689194152598,
 0.36129025479937515,0.8180986089142482,0.5279837191545059,0.5508310626174814,
 0.2315946312131304,0.7930276846206538,0.6378908774902639,0.37143035445979067,
 0.5435313438370827,0.08605953274497824,0.41718515959765823,
 0.18573946395547736,0.20380568594503357,0.48812889705135465,
 0.3362046889405609,0.2134663612056653,0.17664303513592639,
 0.37736058110813375,0.6406515606627027,0.7155240808006806,
 0.30541312605021986,0.9039488692648175,0.22769944354895277,
 0.6770447503645844,0.7507828344750249,0.15495242767303474,0.9422835386174928,
 0.33541533247637656,0.25997160489156235,9.608527687803292e-4,
 0.5480168026800115,0.5451800090827705,0.595730873814036,0.35001096071572535,
 0.10139322384493532,0.6546553019004409,0.6709789320688996,0.3606912070097198,
 0.07755006710669332,0.43980858164124237,0.5440269580071297,
 0.4666846900687144,0.20123749106652922,0.10869340989540965,
 0.07550908283209246,0.3134195903731176,0.2045651028377331,
 0.26547523181163246,0.8126077739690041,0.9484908685387896,0.4731072124673895,
 0.5479897802201645,0.4141460020514264,0.5530595215153148,0.5505111846469876,
 0.9614047874324901,0.26102251282780187,0.37354278141400665,
 0.01764721302569927,0.6807562599601281,0.8821561684846867,0.8742668559219099,
 0.365873257191791,0.015854988235795897,0.44970842581030446,
 0.16092903339057907,0.005848662213958078,0.1895019038980429,
 0.44212088012190653,0.19534134911840795,0.7602688368410188,
 0.41152091791072043,0.42393105926271035,0.7799746558566267,
 0.2821201647838193,0.3180723121807634,0.6665405072245059,0.825963478579717,
 0.6620792330698756,0.9259109972899741,0.7033355312012273,0.9457301971017665,
 0.7067343554702598,0.8026067114260679,0.13239095284433033,0.4744300392742201,
 0.20746287871864144,0.9577644678575774,0.591381188627587,0.578707190861634,
 0.27930258735502833,0.6320581283201346,0.8280589059444801,
 0.38411987752042864,0.8693793349361629,8.120409633227688e-5,
 0.8538525070822743,0.1651097553250569,0.5986006884390622,0.7463814366084784,
 0.14908467751204213,0.9979253818469193,0.873660690786066,0.10061329168934818,
 0.8035825359414169,0.5347444796222248,0.6675132056136609,0.7566855210277142,
 0.6925384823501861,0.14061010386769313,0.8514110004526032,0.9878167666186171,
 0.5154658955112111,0.3363724981306826,0.7053025041435764,0.8013462219403331,
 0.5206955704223831,0.37623432033632187,0.23709849417993312,
 0.7715766370422409,0.7961069772989924,0.46683798248945985,
 0.20613280478614882,0.4676476891640773,0.12739416029133122,
 0.4234309675826067,0.1464867980855613,0.8954632284863383,0.532759107465925,
 0.945014545101979,0.5731763607192196,0.22698222046298577,0.3472917545840938,
 0.5272897972508772,0.053519358052762644,0.31097234536480345,
 0.12596020535165453,0.4042972753585542,0.6208150754730161,0.5598038380682739,
 0.46500039843540963,0.5884772254642691,0.728049157537876,0.5044350111269806,
 0.3325291791258579,0.8163619847886431,0.9387337876221932,0.06278733418480265,
 0.2631439755777748,0.9920419553539304,0.3761522159529084,0.7498670263444792,
 0.038678936823417165,0.8943709528452554,0.45378434319987826,
 0.7139187201519063,0.6379151052394627,0.5048197902024245,0.7293014628887495,
 0.4252570472182311,0.3571730248322642,0.9807164494825921,0.6147292758473101,
 0.21463581490822847,0.21780209634828407,0.5461396078239864,
 2.5013565954634664e-4,0.9591001878145663,0.9474780830324356,
 0.2143041542438483,0.0332297948358089,0.9717356336295457,0.6729066609970152,
 0.9710195531641261,0.9607601095263376,0.3303100749897965,0.00973880585828224,
 0.9864632903689045,0.9106914532167484,0.013512214014786084,
 0.3855791519690449,0.2445914988712934,0.21042520157761957,
 0.14548737576500925,0.17438087090865206,0.630981579877819,0.5215439362085197,
 0.6532525514645006,0.7057978327327797,0.7407144024166089,0.5181269557135508,
 0.0939113880035114,0.4097148448822716,0.16291531224153322,0.4109375200844424,
 0.4253817769220404,0.44832371125167647,0.6407008268768037,
 0.24280688909698778,0.8547724347620346,0.11160316289238315,
 0.8471230064221826,0.4004816659525642,0.46573843169103024,0.5736092750375056,
 0.1700727530456354,0.5961205441351118,0.5870465718419268,0.6508599889581164,
 0.8716893136965126,0.6542950194813617,0.10069302790772416,0.3126589122597382,
 0.39420764065314706,0.08092527927305992,0.4288375092195855,
 0.0707266818302128,0.9704503238641344,0.3043985949680045,0.1914629783229863,
 0.3736298361422329,0.16961247121986367,0.5743787902659605,0.6229189404268243,
 0.8053961257518393,0.6319348243441548,0.5304001967501807,0.25627554265739694,
 0.7667378154028769,0.3888499915025724,0.5721321946375182,0.826427716253666,
 0.31412317929535827,0.9611356836141276,0.7005667337099207,
 0.09496763532915942,0.7503470608212187,0.3967768439699074,0.946812068468059,
 0.19910318005725236,0.3970980293486165,0.45496488263314316,
 0.8256186863448888,0.34537790824088954,0.7477400559875397,0.8633427726289392,
 0.10967562259309993,0.3045294910246201,0.9255576972259961,
 0.07284474653667239,0.3423804426966073,0.43890401065244766,
 0.8035313282067325,0.2538228477210538,0.4271641273896475,0.3262655056795649,
 0.17769484191435647,0.7009305936479143,0.927215494673767,0.04464455012581414,
 0.6042343402917164,0.3046973303230541,0.38627913036285366,0.3616298587517266,
 0.07841278535432272,0.8724599479539522,0.9559592603873416,0.7200282659503422,
 0.03354979534911495,0.39801039131791827,0.347740999901775,
 0.14262472807611626,0.41264598528694696,0.7833050713814568,
 0.7062316273319337,0.18103836458538014,0.20885400437368518,
 0.42033794717765716,0.6549178144318909,0.22032170960433706,0.682372805210655,
 0.760726703170484,0.525606737960245,0.3400434126604768,0.0543831447331693,
 0.031059391908197953,0.4675864735819104,0.9408904498382822,
 0.3891792380887339,0.33203229536747436,0.22437526226694782,
 0.33562303851821085,0.44384966451295904,0.8601521965377684,
 0.2803542794937184,0.4462755532530389,0.8223767755229705,0.02216999588291968,
 0.2744739749739107,0.0931900796817684,0.7148204988149766,
 0.051769290366504794,0.5165915058154684,0.3411299228591269,
 0.2645874539892735,0.5722249794977654,0.3320368968089952,0.6226653854534651,
 0.836808838470396,0.9027477936820052,0.31055374988697104,0.8977446970990544,
 0.6259435203299198,0.08263475081304117,0.05384513083208742,
 0.05984180622649116,0.18201831748693065,0.16573452955494472,
 0.40289210061830805,0.40302716800900595,0.6151390437848778,
 0.10473313720556854,0.148743560233898,0.78246540219663,0.03410485750565029,
 0.016146525697739222,0.8521742286902125,0.8141151498142547,
 0.26144597670643066,0.40169877442677593,0.9052419623068327,
 0.9178424438003953,0.26328412497747544,0.6951797813706588,0.7177049251146383,
 0.2752769559878927,0.9663865084686426,0.5226133501551061,0.49409063322305535,
 0.5063836697544526,0.5757609903038767,0.8125237686414586,0.711024354721731,
 0.5750832489258139,0.16145133379142307,0.5936018741941831,0.6581843107465323,
 0.6207166624735051,0.060158247122021224,0.5374693128969472,
 0.6937032373364996,0.9762364507701,0.1822284757303143,0.08781978898661924,
 0.16305191292768928,0.9468943616579578,0.5968819852582324,
 0.24845497034117503,0.3904400934243457,0.39042291345577496,
 0.4097151095803654,0.19573502848124558,0.43137157136295257,
 0.8330720875147093,0.08176088747391264,0.2589393571190486,0.483578616138467,
 0.9349502339641587,0.05205841836486247,0.5121338594363258,0.5128620984073888,
 0.8249497501240081,0.7313554852815939,0.6366289428770766,0.5847370597965909,
 0.13496405181783055,0.03156644613833981,0.7912166243394695,
 0.7283562896257786,0.4080441057706595,0.6088831358559627,0.23276372435714066,
 0.12035791494132975,2.631147354489327e-4,0.2366214812948304,
 0.17389309124367225,0.8298101476052622,0.39235456726320406,
 0.9245960640033578,0.26326094377160414,0.5217611951010276,0.5055352320285738,
 0.0010316881040135506,0.6805679833542811,0.4735848883353655,
 0.9330934934039574,0.16307564313515144,0.44339850514892476,
 0.0570721511413973,0.3616471428188788,0.7023754763441759,0.5998025603597925,
 0.06511473603754248,0.9772473049320494,0.942856214609293,0.33766710782672993,
 0.9903554550181703,0.8870948254723761,0.18812868766313862,0.9298245848155215,
 0.15809465380410015,0.08565429242049716,0.004542479780908382,
 0.714108501185556,0.5185724928068703,0.9546148676532147,0.8786168984385494,
 0.685520839012843,0.14133892256722824,0.523766698281279,0.674056040208538,
 0.494589282803381,0.6516031332078576,0.5749210164821452,0.167250084651562,
 0.8771870404289765,0.18062474455608513,0.02908424262862197,
 0.2217562726572615,0.6113976142050336,0.03766865984633316,0.8601465443929673,
 0.724943208452769,0.19457383874804923,0.24270775739478534,0.4035106814934257,
 0.02630388921903104,0.5845671518765599,0.033528172861940764,
 0.43816299459531427,0.9577999062293039,0.6598822558498834,
 0.13245044559072616,0.3702674317566449,0.619414786384296,0.35089746914873965,
 0.5693646675422341,0.3373374105891217,0.4073605678766571,0.4027450042370587,
 0.9406884343665551,0.2725672968544012,0.8535230767504276,0.10435466129497895,
 0.7636535925364143,0.9517638284960916,0.24802341434557607,
 0.05220290780108616,0.4855558312071099,0.9793718652926142,0.7118634213995517,
 0.7203621303155472,0.7852842789193998,0.352036586983966,0.329217539889717,
 0.9961020939883506,0.228392653410179,0.5695993153768075,0.2769952463602864,
 0.3413269914536855,0.513484460903466,0.6375371062497831,0.6621202115198681,
 0.38140220962116467,0.025242723928187605,0.7594522539636882,
 0.17160503835799779,0.24904954951171632,0.810559065350651,0.6729861121769585,
 0.6616076390137532,0.1210071589094267,0.9790666874707122,0.0793451363625447,
 0.06151927289359749,0.48560310982728927,0.3564213752751064,
 0.5078830184270146,0.7142575289362331,0.9140524643675982,0.8153448968445434,
 0.8487805041945843,0.9290467097899637,0.13397827395953654,0.822609504533925,
 0.5054518245958111,0.16470572138604156,0.24675608538208194,
 0.7883931245930241,0.05560875836714385,0.5640254481767606,
 0.09952305791063487,0.8140689115562808,0.7555223507394674,0.605941225484098,
 0.6823031225277036,0.048150751813570425,0.7703518761883613,
 0.21224903667364226,0.18954702976947735,0.838047661056094,
 0.26247266515219214,0.7031759278137493,0.8982564829277085,0.4846120958418023,
 0.5141400849996813,0.918538972960475,0.773800296930433,0.7712779264409062,
 0.3942690785236067,0.46198559669565753,0.5530704910488994,0.2921792925747043,
 0.36717377890970493,0.39368707382119417,0.519665089602134,0.3742819196717132,
 0.5047939349311515,0.2949045196251552,0.3558833346375093,0.7652719632274735,
 0.8991050465026832,0.11839041504161585,0.001594217558498645,
 0.4301758224590564,0.018295701470918635,0.6905174654458854,
 0.9434743873911342,0.5722761147984513,0.3738579373683173,0.5721083019360155,
 0.7917742647164255,0.6734069031113522,0.32681242405579514,
 0.26751616223393404,0.9712549127811598,0.32750717498085624,
 0.4506127397290287,0.9651643698805039,0.03695161673997971,0.315397537162744,
 0.2050208235385711,0.15031793574020724,0.7649557826091825,0.6882357075419743,
 0.07922076417012625,0.4264400505980831,0.39839207783811126,
 0.2562500819332164,0.07277017174453104,0.11560366159554569,
 0.3039161581412724,0.7399299264297908,0.1932495812138031,0.5292430446364906,
 0.4872931402577716,0.38721551150767586,0.046306747304790785,
 0.6681140530654223,0.5011071181355112,0.11180855505408249,0.830880197006582,
 0.3220871028972271,0.39307400834944006,0.4940150620310676,0.3612216843371746,
 0.4014097827110149,0.3659448044378446,0.9679292130441467,
 0.027719513762876624,0.2513730472378348,0.797281279514177,
 0.19441999199890048,0.59095289817031,0.45221763357262035,0.5479543770164667,
 0.5379364233186927,0.4770652919194768,0.44687949696924834,0.3234803585597108,
 0.9493348848108631,0.07839016431170065,0.5621541221905639,
 0.07808781649462126,0.7790559156720368,0.008607885917351954,
 0.8947280462771244,0.042152388343837166,0.5615263483623297,
 0.20432332820119437,0.8713722375350974,0.19332260204351004,
 0.7780317228636496,0.8134451235647469,0.828378487975954,0.3855159935518869,
 0.3853792498415587,0.8863654621336233,0.24008816689229695,0.2671262704169485,
 0.33890225724848977,0.0202407382819203,0.36325903975024043,
 0.5316586140506039,0.7441203808561073,0.45601744439867353,0.8725826425492027,
 0.7331568768654212,0.5909996219558215,0.3261712658922118,0.15419866493026213,
 0.9784815775704954,0.9199754671729251,0.7726010282296786,0.81624380134516,
 0.6312634765421316,0.16560217416358913,0.9799289259173068,
 0.05280310756851747,0.615564889025545,0.686167268731125,0.3438221445515104,
 0.08202994857716517,0.9697233941310067,0.16959760192515128,
 0.47325690712348134,0.9827441380122323,0.21434353480413815,
 0.7494143180035828,0.22159465898061714,0.4400653001943663,
 0.35228599764942126,0.07381163734977447,0.8686617969229395,
 0.3588255027321626,0.4313966101669452,0.8879270746080493,0.43518986533692505,
 0.4682892286245983,0.9119965111819106,0.942054848702376,0.3745868652118872,
 0.5985433388247741,0.22476425608084416,0.7270392056839337,0.183839225182152,
 0.8773505560904793,0.15559851925784862,0.95066870155491,0.992492077763746,
 0.479206243751197,0.002110856952247997,0.8553235467839039,
 0.12017238584643042,0.7212814965977492,0.96232054239008,0.6251023754870503,
 0.6008082041616165,0.5850202086120438,0.41245157099371377,
 0.08500225705882036,0.31258135056415737,0.26383909896707913,
 0.6190232459212919,0.7436327103015745,0.3084369722831676,0.5563972246197924,
 0.5998416806685578,0.19185731213917767,0.19727331945271598,
 0.6735940941614447,0.8036135134773742,0.7610748314146449,0.8013529367098187,
 0.5781655327337532,0.998904381274436,0.7636397418309132,0.9408823293404083,
 0.277375344572953,0.8703963571077017,0.8654592727720474,0.005627606651734229,
 0.27145339655917355,0.2544637358011156,0.23532115881224513,
 0.8865807234686246,0.3743481904720013,6.228022086987028e-4,
 0.7156996838152375,0.05246125516865563,0.31259186137251316,
 0.6306851205969284,0.055825844723465745,0.44835203627240405,
 5.709437627348102e-4,0.5896037320564631,0.9383735383168031,
 0.46833740068892493,0.3664461646121374,0.41805173133143014,
 0.25653668926603423,0.2707003381652908,0.808530093094158,0.5491229432360851,
 0.8392136818503664,0.30862379760237524,0.770445852959875,0.9135292967154046,
 0.1859955924035085,0.5277831657329581,0.7020524606848274,0.0166845909515676,
 0.6614325048270671,0.7390906852702175,0.041388732206911416,
 0.6391735884437315,0.6886699931292359,0.07165929921505121,0.2851604631408764,
 0.5137242143868864,0.2221666575274015,0.8518794725220414,0.3718371472103644,
 0.469955233702088,0.12230156365233047,0.8180802859371445,0.13786651886858037,
 0.09278853876230764,0.815705208830183,0.7037573820502196,0.33608474454810155,
 0.5725395509663349,0.8985873220829985,0.1652735641689551,0.2795655451870932,
 0.26019581645246503,0.6355821444549419,0.7845297010447418,
 0.053464447027257056,0.10145058378363647,0.5769885700666915,
 0.5785530035756188,0.430634179299787,0.7485122997572726,0.03746183672838643,
 0.30554717710752577,0.3309694551918826,0.7874457733656519,0.2955800058660232,
 0.0580052655829445,0.16858679960062473,0.5129795394790579,0.220681412972888,
 0.07249037773569433,0.08048254992595283,0.41371530054728256,
 0.4678430828549959,0.02782870684320804,0.10095953766158827,
 0.9767145860036461,0.5590220634812634,0.5093242130691581,0.6002115075375166,
 0.6550889837612609,0.01579262228040723,0.6081601198082779,0.8481717177986283,
 0.16350481394241134,0.6802043605807004,0.38505438169588957,
 0.43588315459224025,0.5846378813285025,0.2790222098261893,0.7638977006775401,
 0.741025596464688,0.5433672694878595,0.7375935391621864,0.30556281934553553,
 0.5747350085191606,0.4750155119463384,0.4721863211336822,0.17483489990095868,
 8.197089157291337e-4,0.14263839165850578,0.984227778025378,
 0.7450771826326708,0.13701407496222895,0.39779500461483974,
 0.49565983816298653,0.7297845768920717,0.2679172031382383,0.390092465358557,
 0.7548799301759693,0.6653569243563839,0.02865803971847114,0.9702762479964445,
 0.605278156407546,0.22133042446995477,0.6510549858979784,0.11245930036894769,
 0.6463121753167668,0.8701285233671376,0.6036014240215708,0.45259930569793594,
 0.1771916420007762,0.5143660121764752,0.37266278903602024,0.3738958825798826,
 0.7100546568098147,0.13280625150872338,0.03946737968039238,
 0.45442914678309965,0.14721579573374588,0.7643755025714748,
 0.9580310877347158,0.7595847658517378,0.31468451938025177,0.7466886221835367,
 0.8614269739887932,0.7841439575820596,0.6452655582519632,0.7094804312843258,
 0.5023490732870131,0.16506926496346463,0.7297573489380286,
 0.15351425046274925,0.31729709664433614,0.29507439545250147,
 0.6000744999673483,0.09975277850114839,0.6164255456093657,0.5742081569129937,
 0.5735370911585426,0.07903054409009647,0.5489659417848355,
 0.36005695138056737,0.48217273073922384,0.9316596005368228,
 0.5812607979526225,0.38364735103914793,0.967646344235241,0.08646755323992161,
 0.48784647460538055,0.19928094061846857,0.4144888450145705,
 0.18138210456326753,0.40662212347241156,0.3622288942874098,
 0.7159548168380861,0.6640956693544378,0.7284377069484209,0.9090627693387787,
 0.09029049629609875,0.2691650520092984,0.42214980043595196,
 0.9648073220416106,0.4855114924291172,0.34057939750243416,0.7409617413475207,
 0.6865053572230897,0.9030241178903051,0.2274996604236812,0.04131069632266049,
 0.22018075328973064,0.4646679940291638,0.5507232034683619,0.3092739471789717,
 0.054636769432312104,0.36974701754606,0.38124538890942516,
 0.014076325545771429,0.7300426292627006,0.8406433497501211,
 0.06564772494233506,0.19166014101915652,0.23520863566523853,
 0.31366309686443317,0.96763256795779,0.8795245550310368,0.4116038710177986,
 0.4639338345239784,0.33965845523373395,0.5564943717271301,0.837578256003763,
 0.9888294864485704,0.591780540180536,0.14930657768858413,0.7050066169150105,
 0.06665438204810836,0.16409561697370412,0.11905194435297539,
 0.5526679390638618,0.5194301376570046,0.027812315512478425,
 0.7468428058676598,0.992634283647476,0.36439288805312897,0.796301428480757,
 0.5269031190173159,0.9080058913836673,0.19273327572937704,0.5563637640532754,
 0.6830169892043068,0.18071324613157902,0.5514033586893858,0.805226480199378,
 0.7690197880167915,0.22975532121082298,0.7902710717046559,0.6988374565952421,
 0.18993966531516449,0.7043211267154046,0.741840300054696,0.42977800290337187,
 0.9651090179102632,0.8028393219472607,0.7963685457143737,0.5369946965407448,
 0.7937704884669172,0.7887130946213852,0.6876202801238892,0.9413817234476046,
 0.36679168212267854,0.8335685083652511,0.21999956661696074,
 0.8289457538009772,0.5662727173045777,0.45320395187416473,0.4666842999957911,
 0.13605636598056414,0.8925579779736299,0.8314288318573189,0.4463303895156516,
 0.3182599743628297,0.4685177634526174,0.13675308846846868,0.5332693749622786,
 0.8294639275746072,0.6301976035983539,0.05762269129507669,0.249627805951809,
 0.7483954258171441,0.5018969811853711,0.4536367793568985,0.5793835719447377,
 0.5136108535032313,0.513280620206125,0.775806428702826,0.6150877265547317,
 0.06329685916456729,0.9106246752340158,0.30563973977660186,
 0.6679233629418198,0.9864676680478075,0.3544093685024454,0.7835047776708576,
 0.11891838170911062,0.5946037311101313,0.20026599523178656,
 0.5341280907997131,0.6314352165079431,0.07145772418090768,0.7956263195535405,
 0.03639922890853475,0.7230168452361715,0.8588954146613285,0.7021932814346815,
 0.7636999088616102,0.46280857239167816,0.9123829501061017,0.3435760578570535,
 0.12538541707040907,0.4585546598637853,0.6448824594145827,0.57139951592225,
 0.54462484192609,0.9070366246655107,0.5673618933666132,0.18993011341797805,
 0.37276839459902644,0.6867897824144176,0.06484934334578041,
 0.6783011317138505,0.38329558138852415,0.2850727911071085,0.2866461500067685,
 0.7698577857214746,0.8472367772543509,0.7202200616622773,0.38192962177522316,
 0.6156513707258792,0.8008534692332231,0.5890246478152176,0.05652877131549161,
 0.5827957936451594,0.6164013391097691,0.8354606679240482,0.31317431169924914,
 0.40025507155930173,0.9202376573367941,0.6514634518009133,
 0.006107543758543499,0.8252512698691576,0.5796366121723774,0.975586814782518,
 0.7805367938851384,0.13959611704384012,0.39490144993111376,
 0.4811646346323347,0.5151767887449649,0.14784094584631668,0.6109404380510902,
 0.7785601096352686,0.8346044885808824,0.868379237037684,0.6039076645536579,
 0.7262814499623564,0.9342898717690664,0.962502484534408,0.3747618213541478,
 0.7791987262018951,0.6656295695254628,0.8954406849990664,0.6452612489783772,
 0.6179726106955877,0.5526981491142965,0.2743882123364755,0.4381854473809117,
 0.5671158050013317,0.024523630333947777,0.3705790191329511,0.228722622467334,
 0.6568984453892874,0.2444982713626348,0.35922055421190957,0.6774971993536627,
 0.953075493817888,0.5188838044532099,0.8976726220032796,0.7567348118143135,
 0.6938202551029582,0.17876649690694435,0.936778134636577,0.6016506036793945,
 0.9683017555428907,0.20898342253665392,0.05492749529190544,
 0.2096965366784841,0.7483743243646365,0.37541125378619156,0.668053309865798,
 0.2496514167727597,0.9138161018850405,0.5606048423489703,0.6042940467307738,
 0.6941789256694666,0.13869874994236553,0.5415207694790433,0.7933227120680949,
 0.31343517218902084,0.6168771969547339,0.6829160698807477,0.3215872775172469,
 0.026522088194129934,0.8380976858995011,0.9510294593243487,
 0.6496100428498415,0.8976434820614717,0.09644141939069129,
 0.41983207154994684,0.7368022049408905,0.14919601766573398,
 0.6945911949547752,0.4692286643358201,0.4816692898520266,0.7510752106982337,
 0.6083899684830633,0.01751629414274558,0.16168827399722552,
 0.6719698831168601,0.9646624222685287,0.6528409114313305,0.8446192156722446,
 0.2547628151224689,0.535730679566484,0.11688496340904342,0.6863165621984819,
 0.686962994745683,0.1397315667759914,0.5865465629205244,0.9286043987653467,
 0.6025403594942426,0.6289801764441101,0.8757282191401516,0.34223165012231915,
 0.009073448194458766,0.4168590189986301,0.22504178170377098,
 0.6121692029539348,0.32775617789225975,0.8835491163162073,0.851275944965746,
 0.48656090864239465,0.5703590843869921,0.9288167141268812,
 0.23435179340413093,0.3269026730602902,0.5191995670857283,0.4449254658136039,
 0.304745245337246,0.9987795466469456,0.2767894311153688,0.43873445076017514,
 0.599525029621811,0.7033972634285899,0.44730279818700835,0.7057108599643753,
 0.4226226762583486,0.3708130901391534,0.3850290535251495,0.6787580839884346,
 0.10865303178946795,0.45834742539883266,0.051479113549887434,
 0.39427182749997847,0.1033216733437361,0.29918581627216934,
 9.503618526902624e-4,0.24704293385248044,0.47201203574549744,
 0.21631387141951053,0.1527165168136987,0.0629090383231391,0.7907891572921533,
 0.3804857512216614,0.21888573684223578,0.8415791536900225,0.4752596249126403,
 0.9213765974917981,0.17513524984735807,0.6914356005272052,0.2575919000446534,
 0.401051941216924,0.9922816348020302,0.9699842600054959,0.034695185180102506,
 0.5777232200849092,0.05175772297340475,0.8610587187846164,0.6813473038314152,
 0.2564132354544918,0.027444293036832157,0.9454649152033705,
 0.5120548063349208,0.9400687574228593,0.9257311672231457,0.4289259380562346,
 0.9825636203232829,0.8448699459886921,0.025189169132846745,
 0.5389784679079617,0.8606749658358712,0.9078010457436694,0.44731031024700585,
 0.8008138448515019,0.6515428483980795,0.995443828569471,0.5335446161277784,
 0.693040669366983,0.538870535000189,0.4552355076240253,0.12266738039036029,
 0.4688776192118953,0.9868721279842047,0.9060492509728899,0.9415663933653176,
 0.5563619470645434,0.3179391386128678,0.07983212483889557,0.4106495548862725,
 0.2708329057696752,0.009681322319413765,0.19250735135877883,
 0.7729876415771954,0.7019904457460048,0.044131691911414195,
 0.5297164319285472,0.03819674514866689,0.787422334871003,0.5907734842770448,
 0.05983625562579531,0.17700360768167356,0.46811425265420237,
 0.5572513111671948,0.8653431551082139,0.3078434829571124,0.6876135074797951,
 0.5109116589608275,0.9692396624679859,0.2454037482203658,0.6883663003294287,
 0.4977852350789015,0.768592049569991,0.08721810419581022,0.7224741263866222,
 0.5653933526947272,0.9584333407161548,0.5950362234070649,0.6670669248383877,
 0.8107224862366844,0.3739632466825029,0.4356796508723144,0.5424985502511027,
 0.3935983990166001,0.149022748321461,0.9004220278325268,0.9599006942733184,
 0.11618354884169657,0.867843485613852,0.2586576086596495,0.8204541103910806,
 0.5010021214205387,0.44945455558359826,0.2487235351527195,0.8547893629687973,
 0.5242045907887798,0.2500906968790997,0.22083830409337712,0.9769781193943688,
 0.7754715016513967,0.04395752539668618,0.9537140015470666,0.510114297624139,
 0.4990897678962134,0.16991741779849767,0.4986246246111572,0.7635265069403495,
 0.7639268414443664,0.4841386188900908,0.5502773414494019,0.7870804418234614,
 0.9315522822946993,0.5406111826050415,0.5348137765986997,0.43696838086385215,
 0.9125202310136852,0.5461990679617396,0.36135662543650016,
 0.03851955831987308,0.6533744185866399,0.8536017589993847,
 0.25503468435455634,0.12661036125174385,0.49973444867930183,
 0.17401362480424643,0.6698750299493337,0.477467254496034,0.5176984428709328,
 0.05154502706525832,0.14201254001455743,0.04995977417916042,
 0.5707834732312017,0.27858256354092714,0.06089825119102421,
 0.8350586899064636,0.11701272947702968,0.853490481711541,0.333889288555113,
 0.8846885116197516,0.8540332841138472,0.47908265858272114,
 0.39928254930035467,0.34697575116609836,0.5119168507899237,
 0.11479022051198484,0.8361480157160932,0.8331204957026678,0.5781442869412008,
 0.2620053839255563,0.5663697868082838,0.1984033922872741,0.04225689311521241,
 0.23872644506338936,0.9729233671560237,0.02935291344957691,
 0.2469795820157834,0.3909764454770047,0.5588277921862183,0.41896652860137473,
 0.5779654142469368,0.6866867618002438,0.6817261960190908,0.4811798676022687,
 0.32437040897638414,0.8094708562620684,0.2184262117313822,
 0.11978487115503422,0.48883879321377,0.308418631709088,0.9627030566417348,
 0.8513174335512661,0.20060534887063564,0.6599217399001123,
 0.34495244758968435,0.7380247863790814,0.5162260561065984,0.4337238634702759,
 0.47167597741884804,0.574920272900878,0.12469812296331684,0.9017192331000256,
 0.9063535170555214,0.598146832941338,0.07644888488574719,0.4868164203149745,
 0.6376063189167943,0.8439894312586853,0.03217247695357717,
 0.48029375899629057,0.28217283396638626,0.36664589291090655,
 0.250907294894954,0.782019673189639,0.9431152515734746,0.7617847629281127,
 0.1157703946683718,0.38213929569186567,0.7974126243975098,0.9750531726901288,
 0.17434022259287563,0.7147543873491766,3.477216800304994e-4,
 0.47022306503469546,0.7337382153684565,0.22015734237249462,
 0.6070322352040641,0.9731525569789853,0.6039010930240831,
 0.028873507294098877,0.39568329609790476,0.5785366897255014,
 0.3236953464689488,0.6673228999070041,0.08833585333774552,0.2837792931183001,
 0.1722898547124515,0.8287458149972645,0.6625554571399129,0.6179932458062025,
 0.02637211206457213,0.9627298614258227,0.2723246168718929,0.4332324116925568,
 0.7544531052121493,0.08574158928743036,0.5766650022687363,0.9072355377827175,
 0.2692541679764674,0.8724906493821226,0.11260927783834063,0.9768637911083933,
 0.7753114944612712,0.40446652394668303,0.09648202806246498,
 0.7388467614511153,0.4255179797079206,0.09096515681554984,
 0.32288165489871645,0.22794955879119094,0.9495478424956147,
 0.3605827366341736,0.8191641779380636,0.5200895358883577,0.05327813343522303,
 0.3872546931650219,0.09345262722114633,0.8683118134704113,0.7112315569600135,
 0.628671028580782,0.6476448888533657,0.7940069667247228,0.3173479905249641,
 0.28910268554891216,0.37642140204459373,0.0935764297548578,
 0.9924987287253247,0.9817924085606063,0.7329647545513249,0.8370249217954289,
 0.4572189480858584,0.15508701286771032,0.030716623945966748,
 0.8689710476490536,0.7006123111127371,0.3085020758843672,0.3292521351670734,
 0.9793850038611847,0.1521732920933876,0.3147862172823578,0.18724264382299638,
 0.393233206010407,0.859198091411329,0.9341449903418689,0.8965912995104125,
 0.2024694592662143,0.1651948221905971,0.4763140073984935,0.4540781974185527,
 0.7130442849841657,0.16525582814909023,0.9891262178592577,
 0.26103837074654224,0.14095626374583747,0.13052424208103353,
 0.18455520116725133,0.8076115635829137,0.028177066518218608,
 0.9784766695939499,0.7986031946988819,0.548997402992959,0.004683983072686582,
 0.438492113034462,0.25490798802057923,0.09368811061127091,
 0.17008030345701586,0.18718597511287538,0.515578647541947,
 0.09154934469732479,0.8750704898537431,0.8229340862463514,0.2864908025844184,
 0.4533830714522804,0.6491174833453014,0.9050763199158954,0.45239095945316343,
 0.8273705962429458,0.8336579585184379,0.9723570882577416,0.4900967905225224,
 0.746639803715543,0.48904842557462924,0.042670486073763225,
 0.41401585395723195,0.37196218348835475,0.3907729667933466,
 0.7738244452035825,0.8853212726753372,0.34038444019528347,0.8714203272377956,
 0.44205534005473734,0.4703151947596388,0.49333821319865945,
 0.0849008964737974,0.22851475662700516,0.4963783855725745,0.5940220866478949,
 0.7338530665068483,0.2292134297456938,0.7856673452377795,0.7750298242877776,
 0.338465645203073,0.035659206641905916,0.6163307367078085,0.6821419791950127,
 0.5194371294723686,0.5579443703861461,0.7272271757824506,0.32547791229636314,
 0.7762177124265307,0.7008485153720595,0.1131501835921136,0.7614940841071778,
 0.44442918944054066,0.8215819239244585,0.808107593660149,0.34122611226099764,
 0.8741459972248719,0.033811527540596575,0.8015966988128589,
 0.36689782086261724,0.0627930727816235,0.37811334341652625,
 0.9094556935817478,0.5607268914114041,0.3765190959819411,0.1451954888737338,
 0.8992102282222726,0.6788558702773374,0.8817316047564543,0.3905869316638789,
 0.9211720005131416,0.006647360424417381,0.8155206167123372,
 0.9728838141209687,0.31473108934736016,0.9448598531882345,
 0.12151216287217725,0.9398434574103569,0.7605419654155476,
 0.06034944986468482,0.7293632802233859,0.6978456018482025,
 0.05434944245982609,0.7999182729843984,0.7645833408257651,
 0.055668284260259715,0.36744536839967645,0.2860971597869989,
 0.8782087665756185,0.5255572761368508,0.19963698246304973,0.5946832221249616,
 0.19309725785786336,0.0363634132450712,0.9863893030989828,
 0.03453529491763563,0.29527017320922955,0.6239250694118273,
 0.9130720120725756,0.3421644849646007,0.37368487793794936,0.500995096685509,
 0.754649222371312,0.1843867806899956,0.43926522057374917,0.5898659782530539,
 0.8265599204732516,0.4345703546103947,0.06263690087234153,
 0.21714826249702213,0.48532709777902683,0.8262311515315091,0.937807905227092,
 0.0015037676119293852,0.8185505608767205,0.882615121988902,
 0.9125223108351384,0.6857306983295185,0.7453285257203018,0.21127491626914563,
 0.8767027252347592,0.015356510299406989,0.15330071514409704,
 0.7019497092974814,0.3088727983638774,0.5968611611843305,0.42653957020702515,
 0.8002100669280645,0.1903414642204877,0.6316185665647301,0.8139557866731666,
 0.22540336063292365,0.9633926730314533,7.357708812042452e-4,
 0.8360702142673393,0.49553091194683907,0.3911917568189517,0.5352971223158787,
 0.022872841762491603,4.2219387917219464e-4,0.0035169628041489265,
 0.4510365225154689,0.47942127375489907,0.6729684549951124,0.8955465479600662,
 0.6581813780716695,0.6004336588227055,0.9801041266370747,0.5002002397923672,
 0.5985583021830533,0.043337059373117603,0.605382464767968,0.8211047902161597,
 0.9434005946137903,0.16520968751312826,0.3675795092707501,
 0.14913799478393963,0.9983660669386946,0.8097233354002216,0.8382946472371839,
 0.3108520166389509,0.765009769402826,0.10991701473917526,0.6396780822567003,
 0.8933864330808771,0.6633970078201255,0.8015594256128007,0.7264490748413642,
 0.9931259453040393,0.23792305233403843,0.925035408279234,0.7560903479534651,
 0.1382588516424933,0.9949804367548545,0.8550226603009712,0.3126578466006851,
 0.9072005552201402,0.6536056297922428,0.6539234262312716,0.4272927403607596,
 0.9846340674291,0.18053298418367514,0.3280948222556149,0.17651244845836567,
 0.3334071986675571,0.9978447084929218,0.40176175171232276,0.7479011687011045,
 0.4776661852942423,0.6403030622552899,0.33080152728376233,0.5089959574962734,
 0.4298048574985224,0.6466672578730625,0.1665393877794027,0.7670873916280142,
 0.494540451345146,0.9446626227379328,0.38270887095968764,0.41760414920961075,
 0.5282713474360184,0.2676305197341403,0.8196227872235187,0.11053660011248712,
 0.7107161102821611,0.3626398293247757,0.13322347271472423,0.5735162787908279,
 0.73123335325937,0.49582359306250845,0.43278364011840453,0.2701111637309813,
 0.3259539394524811,0.25023761974814984,0.2969461956100087,0.8249556726514697,
 0.41111050802345084,0.7536583068167021,0.775088193463844,0.34457173652295725,
 0.1884801167460678,0.19801999714719076,0.037795487677082074,
 0.40523262366952706,0.2854933346295735,0.7262851042166152,
 0.02779136706468921,0.9431938468035453,0.5894327589657823,0.2169801834790318,
 0.7284025326439306,0.6019450204071158,0.6443803041549058,0.9972819623823219,
 0.8962760403612522,0.433332918211785,0.3585267593914565,0.42383410231263596,
 0.6659159540258093,0.8306032858291751,0.8518948408948767,
 0.025429779712576206,0.3422828514529208,0.5455754358713825,
 0.22881896513348454,0.23225078025489132,0.3302527466710218,
 0.6555945651795638,0.25137095382762475,0.1389758801408878,0.6724139202485458,
 0.9412147123389806,0.1808499997801668,0.7146293037767479,0.965640581690325,
 0.12178339026816265,0.9340109533899525,0.9319717321591872,
 0.032026688747528786,0.21217651200578125,0.24145753138000947,
 0.5037543231516552,0.5096971752894943,0.9042144054879667,0.19056342677532334,
 0.5012775854380486,0.4030875040867772,0.9143939047747728,0.3577990544424744,
 0.28160186034677837,0.7061362550909634,0.9271606403997169,0.7173952103143428,
 0.09185329566658074,0.15054092702104238,0.1101484431444878,
 0.9399587402116767,0.23536755355693184,0.5088005613913326,
 0.11577241634400648,0.30532701311827404,0.8815164636204319,
 0.6989828708282267,0.48139492488703073,0.6900710694194409,
 0.23448187121473374,0.14781755810231378,0.6688012558927784,
 0.5331635252132649,0.1654111310179054,0.5825657527312236,0.06334937359672199,
 0.8210964614486869,0.8838714990866318,0.6256535141276568,0.7411403212983523,
 0.30271207171182635,0.9931867577786351,0.04385653591591043,
 0.03181765272369652,0.1998406621637394,0.3599835959859058,0.7470616129790382,
 0.31232677596339387,0.4197647911332789,0.6492730245321662,0.4011678332578712,
 0.05029286650597409,0.35035435186545894,0.21390325435895596,
 0.21526938508060323,0.6499411175603796,0.9131544757881008,0.4377555304884949,
 0.48906225665074055,0.06491570335865637,0.16051387289046137,
 0.07388038599969837,0.3307363306097373,0.8868495963704326,
 0.17346727565810238,0.5892184117087359,0.7637144343827482,0.8636496224937542,
 0.8061215636045204,0.3951976477933865,0.30124643436410403,0.7140607974780689,
 0.6473492386363477,0.0962338621802814,0.3057884823469844,0.18348547971659257,
 0.6011314154710433,0.031919715152846084,0.5925397027717477,
 0.5575122365405825,0.7717496919770565,0.08729698211274095,
 0.043881310397457796,0.08915067078766326,0.25184376980134027,
 0.9272810645615502,0.1582262749215957,0.6043236293180542,0.22724527798623317,
 0.5204728297097427,0.5107945365812245,0.9568428593770142,0.6370083602062963,
 0.6738774969914161,0.655200367646525,0.20766907036039506,0.9881805698767147,
 0.03607008523020827,0.09616551894945136,0.5312131045249999,
 0.9680730245600502,0.8414237999025136,0.11732857246243011,
 0.31305325832873376,0.07927155991514623,0.5728468745197703,
 0.5886768381563535,0.5078922635088797,0.44946222631191235,0.7456627401613832,
 0.38437873178262427,0.6931405753696318,0.5823979525804327,0.7765135794248071,
 0.21807662139758666,0.3975907549804356,0.4822540963546348,
 0.0027162827339785034,0.007539442727752088,0.7585753304733879,
 0.9800522437160014,0.04459743365329061,0.35689849177420374,0.982913042767032,
 0.5337498174817416,0.643829344484064,0.6541983709624785,0.07366754720823354,
 0.7443744475952015,0.3539762329450802,0.6302291042195581,0.6913904997138388,
 0.45544982023539315,0.7089320099351603,0.3540302005198249,
 0.19104801573014218,0.44859898402457277,0.2982605516388178,
 0.44247231086701966,0.468691461304903,0.7317745103471105,
 0.006401653837284416,0.3248874918000436,0.44438363647924983,
 0.8224479125343453,0.3401102833845594,0.690071092747831,0.6221354227612593,
 0.32171131327919,0.1998465891700424,0.9268949016612635,0.11931291065661165,
 0.8014725128355754,0.20302345992084125,0.6204836643641657,0.7801387193221124,
 0.3271226417399563,0.9314063076735889,0.7720945970099775,0.41586203568079483,
 0.4358977222992263,0.8083381112441579,0.627853986118209,0.3858758858392968,
 0.434640314685389,0.7317668012678147,0.5497401192415824,0.6779303179506895,
 0.0521917237571804,0.9828011837082763,0.00961222758035718,0.6514999716998013,
 0.5139072863614563,0.9843411310260898,0.5502218327356352,0.506506337788926,
 0.969659101038729,0.014110978889885883,0.5062466064697027,0.3237203666467805,
 0.6291721068048338,0.3445358291616978,0.17789777998226952,0.9552336377702707,
 0.9163663396164855,0.30424224341848194,0.7179297438542711,0.3430789685674197,
 0.4411563551504988,0.2736874980269556,0.9599178323622847,0.8902227803480354,
 0.3242756694662502,0.2558667406920164,0.3574253946152829,0.8512229026587783,
 0.32663658766655035,0.14851283799571813,0.534814768443585,0.5015327266432652,
 0.2785911948393589,0.9606111791145974,0.6236861229120476,0.508816502262668,
 0.3925122288666719,0.915540565502672,0.5170797094023704,0.4558769584042395,
 0.04334771010230132,0.8772672073530914,0.8810122740387386,
 0.16932378462006814,0.4732853038107858,0.31884862518464385,
 0.24336864126417634,0.4996533443833375,0.6903203285343488,0.9469403887255802,
 0.013046273273328302,0.5935761510979628,0.16272585687701713,
 0.4194765667744397,0.2596007015610594,0.07711718943356138,0.8698791789969363,
 0.16988683890147238,0.22467413728714736,0.6110087604624274,
 0.47570445602407796,0.3574819110596239,0.7201850249035331,0.5669476556807074,
 0.434795650372668,0.4096911511818184,0.7225859716656702,0.007929463574644524,
 0.254400520130562,0.3125907867696043,0.5400042718306806,0.27120467765226586,
 0.4335996730144457,0.8011876761849478,0.17496727212055474,0.1319099804363506,
 0.45848367637346377,0.4483257932879653,0.3587041315102125,0.9440368286506348,
 0.596209145118214,0.20468760978142164,0.7876507955025481,0.8572165898836104,
 0.674227749657012,0.23056312398219414,0.020551293887280164,
 0.7563539928528571,7.071874907360609e-4,0.24782576227500697,
 0.838476856763579,0.26835268123352596,0.5561101257885086,0.13565586312889522,
 0.9350038605977816,0.09389047131994588,0.0890592267033048,0.6085483269137936,
 0.5409084966771356,0.5270153663013293,0.05664988470162835,0.7709002472279981,
 0.4903140328412555,0.7405581543350724,0.2785365919170709,0.9912986234005774,
 0.34108223433587503,0.3619857736075085,0.6327491287987217,0.681340713436016,
 0.31074402809020374,0.6931315584649503,0.6193177858946719,
 0.008388315476160035,0.6895764249783197,0.12637621287005352,
 0.36883779440042475,0.23182194623868146,0.9634753978688868,
 0.6208685497933585,0.556219091256868,0.4848642067017259,0.7475439503868944,
 0.8960157409869569,0.8979558738620952,0.8240581620684306,0.7551731161952691,
 0.41421081205248456,0.9424384859486814,0.5601642285878422,
 0.18147143736324434,0.3750816788884077,0.4516009068280149,0.9256440998355077,
 0.10712027995415663,0.37832804631917816,0.14190789064361287,
 0.778906747322634,0.03669393633691409,0.3682451810678661,0.7566996235009086,
 0.4058119992055379,0.45556090707448316,0.2647620810347383,0.2720171109256332,
 0.47381698532996175,0.38955205400615345,0.13102191690014742,
 0.7132534204303432,0.8188930615395364,0.6305929624337214,0.3680098224366414,
 0.5207914330360867,0.279530021717174,0.39476693654695716,0.8024473861316079,
 0.15588780304753314,0.6712389257751935,0.6169184914133914,0.3602117121841195,
 0.5048084661515259,0.07869167819565714,0.21697949909382774,
 0.1796460333040364,0.7330234056785836,0.12343048778057653,
 6.498297934264263e-4,0.41221525628937994,0.9189867828736314,0.87707533967279,
 0.4301835845798062,0.4029381125072251,0.5830128670463854,0.9297358694164672,
 0.1550306269018551,0.4729749322182969,0.9084462034744891,0.37848836781489625,
 0.3893880063503026,0.9399589773637508,0.8665620320598638,0.3705228148894506,
 0.1958791658182798,0.9471389477562661,0.3673045990748186,0.10555375279200163,
 0.23493626685616076,0.5607205004391007,0.03197892918787171,
 0.9106108904857789,0.27235837294926046,0.9665656104978536,
 0.20803271967932147,0.8487110576341516,0.26972720543579953,
 0.4864036099864758,0.8551167575800714,0.8719974286485472,0.18081834842455724,
 0.14016192808127914,0.6133785486183764,0.8838449212520917,0.3741422718039251,
 0.4663062347315804,0.41140927142292205,0.6952117459855865,0.9367445137506366,
 0.30705871408650776,0.8006419547279878,0.4491110396911613,
 0.13747092593016486,0.8886332088309173,0.22097346788666228,
 0.6927397695369396,0.35365268091389546,0.8087092185645234,0.4772236972319356,
 0.9966681306869261,0.10063962107817492,0.019068981118577022,
 0.1557359290121607,0.4196255767343029,0.06157048493241346,
 0.16777335068551025,0.9339854694813082,0.3166514447772584,0.2317046111182246,
 0.04181439192479919,0.7232745687092557,0.5632089532449345,0.4393346472077192,
 0.5404668376502793,0.32494857435264024,0.2750755552825096,
 0.45984092108501684,0.44300774775343366,0.44520642817321,0.9309084816217913,
 0.7688820616371062,0.21488573755209828,0.9517552118532768,0.7385763720160496,
 0.7701039454940241,0.6136049338956122,0.48445909316848346,0.4556865616391791,
 0.9708751010848241,0.6016161329772872,0.9398055059647356,0.7226702593149763,
 0.6702919445279604,0.2349836647404807,0.5324629683240523,0.9744421947221027,
 0.40877331068204137,0.2365048165211041,0.5404471035494817,0.8025560986289103,
 0.31281129860075696,0.38005674190173955,0.5384468367329697,
 0.8726606649623101,0.3902539265149958,0.044344912619602384,
 0.2416085092811988,0.8533492805004851,0.565341207813988,0.9324695840450812,
 0.7057727155356222,0.4439023525842474,0.26960262609608476,0.8196747515977485,
 0.8945534125242591,0.5454475612201526,0.8622895013141654,0.36028807659671125,
 0.4579126102195543,0.24160722817786784,0.5317472352364072,0.3747948344932417,
 0.7769978386368781,0.16035014808307857,0.13467207971364692,
 0.9721138118484293,0.7451634802576175,0.02405415944366529,
 0.02109892913448108,0.397376950022991,0.7755334481551941,0.30488945959271696,
 0.40694770576222294,0.9398682794451005,0.9913005752891539,0.9917312219610539,
 0.4234281133084413,0.9908609070788963,0.27202348771879725,0.6546643693103628,
 0.4099976333162715,0.07247895333011733,0.04864809064757081,
 0.8071597026623083,0.5710233505988493,0.978084540141122,5.433847772173817e-4,
 0.6408924475063961,0.21941948161218194,0.2844318303560349,0.4581995534702825,
 0.9368835292467448,0.19534276144410123,0.18269839241322328,
 0.2939565887176985,0.29177923605762146,0.47004724866812353,
 0.17801569901203695,0.641218375746364,0.9193813872962421,0.8944144845263251,
 0.7511476598777965,0.7638516609364141,0.36911809425062636,
 0.40633682988185593,0.8548409069737161,0.07500546784345641,
 0.39768185455144334,0.7466434070094685,0.7787498642778287,
 0.32454044408883786,0.9475974544338956,0.9936761306867283,0.6421677400255036,
 0.37828152874516574,0.6649971873195551,0.3348829171399854,
 0.35691967211149067,0.7849035563189762,0.44016852536088813,
 0.3491217570659687,0.7259305084274812,0.3893731798990473,0.5428783498578367,
 0.4961221997520977,0.5267187354875518,0.39423951078701736,
 0.06550686003866502,0.8017336056295244,0.41556462874069,0.282230723532939,
 0.94229127155971,0.9421239335497863,0.5528564036787766,0.698596309179251,
 0.5592307636334052,0.9540984742523044,0.9958686543771114,0.8368484157421645,
 0.30419561965016306,0.2865443511755599,0.24783659330013608,
 0.24540172700577945,0.24798150858927226,0.6516580172374064,
 0.2969843967547532,0.49231608220550627,0.430508026372174,0.2768172709743447,
 0.8606776848203171,0.7019825597268001,0.7340689776831297,0.8663006099690895,
 0.7625653808758941,0.14479235104973331,0.5156932801843925,0.563091178718013,
 0.6876653590800065,0.9225154399833962,0.6507411909474431,0.27399431042137457,
 0.5272875353380598,0.08926358083637098,0.5763148509053393,0.8540966745223132,
 0.13869344169655806,0.05066025577335764,0.11611669350046439,
 0.701483936866391,0.5453372763833523,0.5242220828654153,0.11470338856087403,
 0.5848742644743159,0.7026914852100579,0.38588698429342416,0.1826940864265445,
 0.7410432843517644,0.6637023067125154,0.23106574012351977,0.7065788322227954,
 0.7100039849634604,0.22563305360899222,0.3764395328915937,
 0.18245631803585594,0.015975272770356153,0.09934265768670913,
 0.2514786142245078,0.021183863715608764,0.923257555373725,0.8607739205226032,
 0.5567436998859389,0.029207310055501035,0.6293175210026954,
 0.31949229454108297,0.43215325050238695,0.7306938464162653,
 0.6776406538846634,0.9782385039602262,0.05857024563899116,
 0.42982994922708095,0.7078003507014705,0.3140017329517847,
 0.18051780172192566,0.09510811357005933,0.23504328866916935,
 0.4458477357985985,0.161387071496787,0.28089629299586627,0.9538962625603816,
 0.7530906149293383,0.07632521464877473,0.27272103614406784,
 0.8145164409339861,0.2902838583951459,0.08856754840599801,0.9566156313913182,
 0.3669711289231268,0.5662050897311817,0.9951986917688711,0.0672731641358878,
 0.7785474115978879,0.7235201070721897,0.15119687367097168,0.9833932954546194,
 0.7600036163944611,0.9493474249331927,0.17800347378128634,0.9533373340008179,
 0.8892280332759639,0.19526357237444292,0.917405858054829,0.8743682400365907,
 0.6852298130525605,0.7631800775862237,0.8970011625631849,0.31835240548023913,
 0.6662845824370583,0.6045122586119893,0.3859324107183919,0.8281503151856529,
 0.7388235192554672,0.9257572252052659,0.041550703079921925,0.71638753472422,
 0.46136343902928934,0.5341254139913254,0.3633221688075099,
 0.05803755212908335,0.6401157021643683,0.25392841946264355,
 0.6566108710649313,0.2886853172489292,0.8272819535373965,0.7473352494007013,
 0.9969311739569944,0.16362824326521086,0.07728418484148836,
 0.08278761429814441,0.555871169452135,0.6698663583579496,0.9808344040966147,
 0.13857949748912934,0.15874053742954264,0.050505174347769466,
 0.7021138135747356,0.20417335287266025,0.3713777006138136,
 0.44068606933947607,0.5859617683126452,0.18043641418267342,
 0.4642483379723137,0.959957116595093,0.24336998136713983,0.08358352768662347,
 0.5406051293423418,0.27092675977697667,0.16859325000526026,
 0.5621075857093234,0.520911642215207,0.9668389019168442,0.3318088449401265,
 0.830995935271216,0.6638617591136337,0.09211846808086166,0.5063530832947408,
 0.22950370393275066,0.7599128538762855,0.23239654224749073,
 0.0080678913134411,0.7202697718423337,0.8376542384856036,0.8145537792024422,
 0.24856792836443786,0.1568570504803285,0.7871952885351184,0.3131077467404131,
 0.3845908794393137,0.9433030788105827,0.1837585901516836,0.33195053682695586,
 0.15064664584992682,0.2629361377199504,0.23980052249567474,
 0.7935459662393304,0.07036279599355144,0.09217424893245096,
 0.8427354197224666,0.0014383271099024375,0.6594580037532253,
 0.10072228757262558,0.6157153358123529,0.8385485239613986,0.7583383585509043,
 0.35216105693425825,0.04547863732535762,0.857079924910557,0.7510057428772007,
 0.16290089000760677,0.6258850731362974,0.7918629647525776,0.6849198644667834,
 0.7596387074704305,0.8007792335716387,0.3740908153385417,0.6368464162928167,
 0.5411465109412404,0.4853363874859873,0.2575059873430866,0.8165892083155344,
 0.561614279085194,0.482284152844637,0.4118039037630108,0.5527103663488115,
 0.5128406697163674,0.6090645343454384,0.9926934071350901,0.7487194740957667,
 0.9774239050491177,0.6834604249381112,0.6246468398230623,
 0.023379803881007244,0.3545603949662274,0.14508975569857785,
 0.8621344374260076,0.08581313814474278,0.19793165794997125,
 0.06367004515785657,0.5419424523363021,0.24789523028391502,
 0.1508300514840265,0.7966546837865419,0.49391072164627414,0.7693055218329707,
 0.41177069367278263,0.4794396581688336,0.4308689105240029,0.6868111006674543,
 0.13502899653365596,0.06593350476814108,0.5081635036169196,
 0.6173215697613214,0.5148329627072752,0.9728602909315681,0.3527652312477674,
 0.7975400672458638,0.2019149635127122,0.24549524682530954,0.410539003493509,
 0.4474186288518751,0.12089438529892949,0.5484083165116937,0.9438377274515773,
 0.41092496233745457,0.0402439046842914,0.3713536967804144,0.6247410152802243,
 0.7801012490429657,0.3276411965764955,0.8032158182074584,0.44120534404677947,
 0.5887455194527009,0.23448281552071348,0.5592320913075233,0.5575890424556335,
 0.4718623326569875,0.3834339101723725,0.7533619427543037,0.3602547128411554,
 0.43295140788397335,0.7242114235972811,0.9708208824795909,
 0.12546385628589063,0.9265503123818578,0.29764889105163306,0.954506527582017,
 0.011316117976098328,0.6277335442788308,0.9509062574403631,
 0.1551369592067926,0.22723867133084252,0.31891735418772416,0.679390099307822,
 0.804067626743579,0.9233317744126454,0.5557360288706671,0.49751662864917634,
 0.37318131388444564,0.07959465251735298,0.8762115875221895,
 0.4405223960904412,0.6957597601789616,0.7400533548279727,0.7508442647482434,
 0.766510978573682,0.7622602853641842,0.2065898581726563,0.6416721216739644,
 0.09506237973094189,0.12926778878080047,0.15880744540521097,0.25347241155334,
 0.7027303622447801,0.6835984216968168,0.714543827545559,0.9030594408682748,
 0.9021252475687681,0.2366124218083262,0.10152345896353865,
 0.12142579373013285,0.38426165260889933,0.11651035727873016,
 0.045574234722698614,0.7157310273502626,0.615929702142676,
 0.39827562971181796,0.6910237772332117,0.060852494268155755,
 0.7448899722598225,0.6995273015422641,0.8928522888296534,0.5360682901259244,
 0.8719990125315659,0.6660021902133679,0.7541575357555333,0.959408813965374,
 0.45944175991542036,0.8228382936848457,0.4908295238909489,0.9967315557890584,
 0.7468344407618548,0.5916698476491925,0.12236201718015782,
 0.40320986640299783,0.5331630635126914,0.9128811699538092,
 0.26386891802281887,0.2566065667128583,0.8168123109214918,
 0.18352804676755385,0.9060924226369514,0.9441123138880403,0.3726997815725632,
 0.014385539604089281,0.5659916062058118,0.746696214905215,0.9970445161209059,
 0.934764516650312,0.6640461432362933,0.6752950713772811,0.9825531217344626,
 0.16992063708832783,0.7157967269115424,0.6234903474186964,0.8603239350073074,
 0.40195963682746216,0.5519583412417305,0.5715462251798393,0.1500679799508382,
 0.32905921221642975,0.9449090634647109,0.0627262131725268,0.8541975955580765,
 0.7145211256830234,0.2249265214432674,0.2930664260996859,0.6724828708157439,
 0.14950697811019475,0.014423504719821012,0.009695240545097183,
 0.9143138773196808,0.34772630376899527,0.9558025246638696,0.6107353915053769,
 0.6355956664904622,0.7875293809858459,0.8461104593414961,
 0.057720088390062374,0.3809673972729123,0.8452811246530818,
 0.7803412402873952,0.5419170347564948,0.4440607753182144,0.5865509485719854,
 0.18486119663365985,0.5661437314635533,0.7177719735792906,0.7718558831203801,
 0.737838053964909,0.7984892168518734,0.6528575063534491,0.3315361386831751,
 0.8305273488335624,0.7244022230088256,0.4588159357997236,0.9704599422599893,
 0.15664864809257584,0.7624050701310533,0.3359257896124449,0.3193416769561406,
 0.38753539760723177,0.15288171229021208,0.2881953825847343,
 0.0321642537338539,0.7458422957389078,0.6271490712373631,0.675689702589489,
 0.6213316421052917,0.4583813637457209,0.129131400935375,0.8815426391352709,
 0.4609313633808657,0.09586824100811853,0.0022879841673026924,
 0.22807436118125646,0.39347401277368743,0.25596249811553906,
 0.417843252342718,0.9330400363679453,0.5105347702993834,0.4072661780422784,
 0.6958062703551102,0.22069831312728305,0.32374650837831587,
 0.22641554371751038,0.5510726293335533,0.025411364778715084,
 0.16850134979825826,0.36411071159832264,0.5317622260939566,
 0.7996333150688166,0.001508380551706212,0.48438359143929255,
 0.5077304648771374,0.546760340627642,0.3105325035911013,0.7246273797353313,
 0.5344760065875299,0.12933153910966422,0.6413281958565897,0.8070640153462427,
 0.9483491740721124,0.6069012780266316,0.457200941388206,0.48475401935222273,
 0.38067064177238774,0.8213266723820709,0.8160817005071013,0.9398032765297988,
 0.11639209877706735,0.8335688352764868,0.23798075005281571,
 0.5602066770776026,0.8993292990426627,0.8901412792233876,0.6278560416120027,
 0.02089742883831991,0.8246065545396398,0.8550886644282389,
 0.23883863107044667,0.660969160610759,0.9196862813789428,0.32055998408719444,
 0.41431885099061383,0.5662832897895095,0.7930712929086452,
 0.00801547686355042,0.030537762920030964,0.11595714679840097,
 0.5443783436956018,0.4887322578324673,0.3298501433490363,0.7672104165951074,
 0.18954811430165197,0.1862191941317829,0.4539520756116785,0.2911278490464635,
 0.8447996567522663,0.29055655706536554,0.08376316195987399,
 0.9742982623434466,0.9058097480244711,0.9217228966071889,0.510010326937772,
 0.5055963485265907,0.734955810490056,0.020520224277272736,0.3504855385448793,
 0.5247588761236045,0.5692349449683527,0.382591519110683,0.9595106740203797,
 0.24401415498879364,0.730713098135827,0.6947147642835642,0.7205216326562572,
 0.4777388395219556,0.1209494493575276,0.7434769043542769,0.2683570219992034,
 0.6592441915216587,0.7217935520479408,0.5234931035917646,0.05097284374739264,
 0.47085321626013577,0.26906899014122554,0.20885640112981163,
 0.870995454436119,0.3258263074900778,0.3893944782755685,0.5682002077179626,
 0.45819920935255576,0.7495048092374611,0.5470922544713163,0.6382294875177488,
 0.478231816983649,0.9306238883820969,0.46732731086439117,0.34072512526283816,
 0.4312373503072445,0.2801246359560794,0.6424883533906699,0.19926636662741248,
 0.7034334270006944,0.24278524993975203,0.19669575265012929,
 0.6934385766497764,0.36608295582512884,0.11710726676954919,
 0.4389536901041402,0.19115345642360104,0.6082170440173876,0.5543492926750851,
 0.7230534180854865,0.20740578841561685,0.6425793661992928,0.8322151918484212,
 0.5368691029823719,0.42549763405526475,0.6280698121225778,0.2377187103211995,
 0.06789194153515055,0.3639495644425125,0.7700736306300319,0.1630596989123384,
 0.7920467675397865,0.9750696505818668,0.07531642703043528,0.7412972061540749,
 0.26605360598916505,0.8941660805808208,0.525707729478775,0.27583799383987273,
 0.7114672906142558,0.3351144753214712,0.32588248092959504,0.8055868466991115,
 0.020830627776134425,0.051809019115171995,0.7746451346630219,
 0.048872189567508606,0.8769147427694194,0.08144669122776382,
 0.18921277582781626,0.49144352574311223,0.7159667466491513,
 0.8181580822623948,0.5476054408552162,0.3705505187656385,0.6419824491759949,
 0.843262211692376,0.8421743216492549,0.5516261180218559,0.5402749259450872,
 0.06897847528841694,0.415451488128767,0.12163724745856663,0.6272308240214591,
 0.2832642238604809,0.9878959838319251,0.6157094738050367,0.01545213636404219,
 0.049824781903081705,0.05481270691257101,0.9848438344832007,
 0.8894708237694643,0.6693334345174026,0.9035805190452267,0.21954680528724557,
 0.7999258265495811,0.3496861718417851,0.11252542048023695,0.9894936180142526,
 0.48414239305144147,0.4332963745732933,0.8173715321312321,0.5691110046560809,
 0.01201977634861262,0.18383582821098443,0.26718745693649737,
 0.40764275562704744,0.5418702611203048,0.22129854281170436,
 0.8741365920955492,0.44124064277188446,0.17711611336838218,
 0.09409130654212206,0.5344778335260811,0.9709386437423998,0.504854985338769,
 0.6518511823421839,0.5007830174595109,0.26481677495271616,0.0283811757986121,
 0.2666057608891773,0.8349814684340667,0.9076755047788859,0.25873542911084746,
 0.32234336910128003,0.7541366669280012,0.3457378611572248,
 0.24556765691453952,0.5204922474161725,0.2803649360591488,
 0.13785944740544698,0.3599055155072528,0.5039396933024133,
 0.24120206522409748,0.5706643411459105,0.5380706737734413,
 0.022029531007040637,0.8722492842553904,0.20012058940341015,
 0.4639958382328162,0.24157417259794778,0.6249201517501236,0.6432399949875496,
 0.4006759451204994,0.7119750552719322,0.5955301479633709,0.07704883805518892,
 0.801894676371024,0.1479950860325312,0.5746695305847779,0.15578804376084776,
 0.05270131895209518,0.3382814436438739,0.7422031972605075,0.76794796605592,
 0.19037874828495915,0.7688147797923248,0.12081464332319547,
 0.7719513257151643,0.5668254460309526,0.4268021427799388,0.8618115545551329,
 0.500615229746729,0.8272144654280713,0.2035241068217294,0.7772363991007847,
 0.6174045138419928,0.19834013389302207,0.5684429424016475,0.4235930266235065,
 0.9583034125076731,0.4468637108339478,0.8678674126338055,0.7378453713119228,
 0.41980442030021337,0.9491662812423379,0.3486971507789005,0.6598606593835258,
 0.6341598221340956,0.8966750995257461,0.6607812166029688,0.10497411811260127,
 0.5675397270609621,0.8077690068465084,0.3993468338080939,0.7856881847109031,
 0.35052091890751647,0.6837313549886067,0.06494643336513506,
 0.7242947575720527,0.17397644051263617,0.3357339760383806,
 0.042285358381597526,0.2452781473327046,0.7044532399384676,
 0.8982195624827395,0.3481045544199617,0.9048553579400075,0.5854750800134648,
 0.012198206175848436,0.014879147823422922,0.12624453163843463,
 0.5472710348617723,0.7135137174900639,0.3820723836814114,0.12439271836083499,
 0.7603660023672741,0.48622141338432734,0.2687728267706144,0.2852397765225274,
 0.7982458294625605,0.8116567130365147,0.06899274408714384,
 0.058895243874206415,0.6394598997114828,0.5594653417754529,
 0.1785914805447888,0.22057848033156613,0.31838556000892504,
 0.17653025450887017,0.6975304004600564,0.10541593631039592,
 0.39803710917835566,0.696319122304223,0.637400874087467,0.4542731245420424,
 0.48059089849836845,0.9157266225600339,0.36620942096238274,
 0.14122168284945558,0.8998596727363668,0.6227655547096983,0.8304639097779264,
 0.3136212311295624,0.9710482840915309,0.26639707124649137,0.8068922457468323,
 0.09777890358996377,0.398406289455684,0.46211334134460225,0.5694022897071711,
 6.813613185870926e-4,0.61212652158205,0.9716284080003091,0.22040491004782536,
 0.21177380970994475,0.7813598883627598,0.2879811823182623,
 0.38669843618937405,0.7398855191556353,0.6104761105251894,
 0.37485525170778056,0.8714371949672995,0.8673096875850744,
 0.036551880164408246,0.5339564798907785,0.7671797970884631,
 0.20114123364421044,0.32925626890614557,0.19154641736540656,
 0.27312877781394884,0.33489082086680244,0.3345847725040304,
 0.8676241115894985,0.1590850676157417,0.2918389209087384,0.5553811531429413,
 0.8051036509514803,0.9098110696939474,0.7156447719625125,0.1713838595559085,
 0.35373516172891173,0.4718426897173379,0.2777691167310594,0.6179876631951982,
 0.6537911791975547,0.06243887723868902,0.1805559588236051,0.6826778797344527,
 0.9874983145691213,0.8848511654149575,0.9881665317939599,0.9777983191805026,
 0.7396223591457931,0.3155793712467847,0.18575157240074114,0.2661320050775927,
 0.4296221120265272,0.8105045453229748,0.9315312777158986,0.47421160906559834,
 0.34120843334950357,0.5104264961673945,0.6325271389410911,0.4214533887313068,
 0.3552002911668253,0.5031308418918247,0.6953050707877063,0.3797981340685356,
 0.8562554668405159,0.05636227776300595,0.9347280179289778,0.3108043700252011,
 0.9990949270306554,0.1708618756521798,0.44686816251608774,0.8194108331073763,
 0.7693999409575836,0.24811227621296106,0.3675324691147728,0.9176192603278843,
 0.4919926091043122,0.8888268978836827,0.6085213489209486,0.08443024425696366,
 5.191234517964993e-5,0.6577705512788349,0.7974411905323205,
 0.41978318420558336,0.6486377912405877,0.38853185691418324,
 0.33211733898070417,0.34197075990216486,0.6240266420717324,
 0.2632138681745687,0.3938011435910026,0.8997004725471873,0.19620798868433642,
 0.0038304705090584923,0.4898196168058224,0.41419477000974425,
 0.5341877463031108,0.6443643528543651,0.7167374133482625,0.7082611711609712,
 0.144563455093779,0.4526785367193392,0.880485341299468,0.7638386178379605,
 0.5269527670396694,0.18342436522917283,0.2486762695982594,
 0.39042435405474907,0.7337541057915224,0.9459982971146371,0.7582457777958422,
 0.6329962373155165,0.39118404649189353,0.0784263659321921,
 0.22694041119376496,0.7393101148443048,0.41546001817965494,
 0.3038021300873236,0.489618018024747,0.37328213984221836,0.5216003636781049,
 0.9384571811321013,0.9538874262318874,0.7465754151522718,0.566435153472826,
 0.09572337625721328,0.20686044257539016,0.6389730096846304,
 0.9349254204939865,0.6095159986954607,0.8809688971451105,
 0.034744946390332654,0.03812398471373635,0.9519446096782482,
 0.18102169223869624,0.44756707023396236,0.9928176889672358,
 0.8769024042913071,0.8469520795833855,0.022670165244631613,
 0.06934398045847667,0.5603783162075295,0.13896538544263248,
 0.9394623349723992,0.821279909597498,0.2848814814167471,0.2547742816321792,
 0.30858346326566144,0.8776901285165697,0.5588853705798571,
 0.14983748161788757,0.21357662100544872,0.34798503582830587,
 0.1814501712465355,0.037538246469502656,0.22848881409046595,
 0.5946342564005946,0.4154106207036047,0.9716616940857065,0.3915658263187004,
 0.7518979693189541,0.3527925980595471,0.46670051912948274,0.2734834398438609,
 0.36988657350858434,0.5693669436524564,0.19479559127321822,
 0.9114805031576771,0.5650889826137322,0.5755196208701386,0.2672084272658334,
 0.028408919877217498,0.5200736346507131,0.10284508285926885,
 0.25565313663894007,0.646220911636377,0.08708849385458028,
 0.46063320479004644,0.23324796796989855,0.004854189702208189,
 0.9944917817194905,0.011641811944513303,0.02147205776402583,
 0.9847068834665643,0.2643059812020868,0.7808432113377273,0.31135635671780204,
 0.6086814693615005,0.5946923018192434,0.5266277466010075,0.5563164914827504,
 0.7950609383807595,0.3328877498400554,0.30916832394903726,
 0.30809411638205364,0.05115287023786408,0.4236198788441524,
 0.14353107985260072,0.5444277715789337,0.7016966793279823,0.913358303259565,
 0.25071694066996264,0.983728637573987,0.7720927564905175,0.01104801549966894,
 0.436449921902604,0.5881577328045353,0.0892064183249538,0.43862635308761566,
 0.5481698432557195,0.148356640787077,0.48340850806578906,0.08496861929118349,
 0.7920879143974828,0.9516199168267044,0.7580286817837638,0.8991665876218999,
 0.7178198065857718,0.514982337276837,0.10389073027500806,0.7705139852750221,
 0.1433443539167324,0.8252769120377752,0.1775471349394696,0.48056107731703457,
 0.9444652701560017,0.34543883878119985,0.8631701086934663,0.6487580676024807,
 0.0010970797409590727,0.9993977103497815,0.933461532469584,0.51407785589662,
 0.16637899233753084,0.40739071038676833,0.6727283008551889,
 0.4999113472622273,0.27069855573231283,0.44641489833451964,
 0.3219211570688789,0.033266810524862245,0.6543572704941943,
 0.1753429448091942,0.37782717718004655,0.2856772953683826,0.9139356709685454,
 0.23698587957569806,0.3853387289635811,0.42176534221966655,
 0.7994520979368421,0.5750351857710408,0.25415179865919346,0.7142973417254869,
 0.5463563535870255,0.5879862732173433,0.015578185662639021,
 0.41625061371853755,0.19662651945755694,0.5083496224340137,
 0.32052233859060153,0.727550665420709,0.113838839580374,0.8955471860229908,
 0.3963443442521365,0.761319348462395,0.7551174029019304,0.8185830863394823,
 0.34004440136577707,0.20350514363682504,0.0865869331863941,
 0.002249047047936603,0.3980070490845202,0.23942273136556946,
 0.015376525177508649,0.011004497600414664,0.15707330455563684,
 0.7049852477110411,0.4356978284813199,0.9996969714960262,0.5562927212036716,
 0.9248640146925217,0.04038658415653473,0.32574644339378,0.6116723898415342,
 0.5362008773645287,0.8901561506196229,0.5342736432940709,0.988710426351052,
 0.3235260496158351,0.8790205388052439,0.905623206336518,0.5326920979065561,
 0.22113409286077068,0.852215726653828,0.6264317035263102,0.6052266160066123,
 0.4662043210265602,0.2937775477789313,0.5368806964177864,0.7136654572958263,
 0.6133352639648445,0.17880312802161002,0.9825187384160168,0.8024352993107073,
 0.6788698808322289,0.6576368648640025,0.4433603481318277,0.1309063489451242,
 0.9638373827353184,0.7088909732678753,0.4874459418893483,0.7903173652016009,
 0.6175548859449388,0.06646388696460559,0.7619142378881163,0.4969029042717915,
 0.2039976463735781,0.06899020159329072,0.7621440389481893,0.5225656661972599,
 0.6278914492575134,0.0267755974499454,0.7702389139214142,0.9617455706621376,
 0.07130497234283983,0.08863119321036872,0.6656564159661547,
 0.7078789141814444,0.6541535517561352,0.9300435277835548,0.18907303191990632,
 0.5671188159948061,0.8241934187070471,0.4896968474697301,0.874373693971826,
 0.1656916068271158,0.771967917267086,0.8630281197175629,0.13906147241862465,
 0.06798667376453427,0.6630126995403606,0.14957529101818334,
 0.14116169444809334,0.6607901265643268,0.5268483483223398,0.9819531100991994,
 0.6002064782702767,0.3823355400347044,0.9738828935053712,0.8480091244686754,
 0.6212834672057852,0.5835710481274042,0.3770736331023482,0.2240172967530487,
 0.8848258884289404,0.6290143824136248,0.23470404588511362,
 0.19938972719843695,0.7652717508879984,0.4005620023548351,
 0.22824967237927696,0.383219312576068,0.053656286883228654,
 0.6555230098191711,0.4967186171161331,0.4118909346683908,0.03389714177255987,
 0.41018120769683375,0.8088764172956784,0.08459656621556555,
 0.5454931609111753,0.5018818040550941,0.7704846143514205,0.47796512546544545,
 0.4541641577841481,0.6238788387912512,0.9495703586212316,0.6087382892250827,
 0.4166415889524038,0.28412672952115914,0.42582957035375124,
 0.3912273850470741,0.23945291194285845,0.4922795175997944,0.7310944456460535,
 0.6868524522599408,0.3357549449326602,0.46541705557597723,0.4653429718026878,
 0.9515755421039963,0.7627549387464192,0.5379068107278541,0.4843018626560798,
 0.29834831993870947,0.7129384120358634,0.19253371042664502,
 0.6711220948199984,0.7842550584260473,0.5407172183091665,0.6661281709026211,
 0.0809102086461514,0.0067327397573739844,0.5921136874849341,
 0.5780316137188629,0.8130940325063101,0.16220573682263684,
 0.017095039456425543,0.9814838831523016,0.5024366492471957,
 0.6198023091861744,0.6142895471571412,0.6715106936565894,0.35448948120228796,
 0.44593578461686567,0.015782177317459234,0.008700827751571327,
 0.8837568429646921,0.998829906906526,0.8204112581214645,0.7398374176899066,
 0.15875836651749897,0.1724877582897264,0.051215276816125144,
 0.6092316056089059,0.5696945914059726,0.3784772821797675,0.09846869351018683,
 0.6671285378004406,0.6317359042877375,0.8457263946875812,0.11545845492398488,
 0.7629229277392855,0.3574573774971348,0.2274721972077498,
 0.003438749935734009,0.2892290600221382,0.6918059259034381,
 0.8703534786278835,0.5770657830368877,0.5624789162789243,0.6651352751819284,
 0.17527903243319276,0.7119949791477178,0.4371959321830422,
 0.13485782898225873,0.2873385277346159,0.08213663328537879,
 0.6995498888255954,0.3045346059621489,0.3144834341339322,0.2665439808086866,
 0.20861230481347004,0.4978815411721026,0.6543958992604763,0.6859634509252204,
 0.09955478097205062,0.39312045078159474,0.1425032716432546,
 0.9813920445361439,0.9693335847599983,0.2743686000179649,0.7537420220630142,
 0.5420859739110367,0.9999734685476467,0.7307602765349634,0.13649332524821556,
 0.9900968836282837,0.6635387666688154,0.7812792060111591,0.5785506294935671,
 0.5169911521792724,0.1668591165965757,0.8801889615565428,0.4649428490575447,
 0.7430882885039805,0.31161613922095943,0.519672519944887,
 0.015234984857191547,0.04265491321563242,0.34629627594850887,
 0.9775024065240886,0.300032521755649,0.10601794569724232,0.17866380652938185,
 0.5761567671271968,0.20160150309306424,0.027322770063796686,
 0.2010464529426027,0.3886120512751088,0.3778426219807297,0.24458434760505632,
 0.3379444245308467,0.6906555306683165,0.595325444007906,0.18519968499785044,
 0.3648694315513513,0.19096566104965518,0.19052460667887616,
 0.06394030071455181,0.42052335118718753,0.42360994322600865,
 0.3200444277899738,0.4064269517466439,0.7164169849754745,0.8430846706706031,
 0.8027386695036794,0.1252282826007438,0.9177917313626005,0.16555806147328778,
 0.7752813152268776,0.7046091610675392,0.5361040671485631,0.23628098560212996,
 0.15366437851676085,0.10797109008969985,0.18171748215175576,
 0.7120682034398935,0.14989041504649148,0.4713869285008363,0.7906840450381507,
 0.1663055739104361,0.8409505181400537,0.34888785698444125,0.4893162775087192,
 0.4123704623214123,0.7950709918774524,0.10983867467854203,
 0.38225499183904366,0.8918797246273563,0.8920283143990556,0.8133888890360377,
 0.10574603257531778,0.7796870663475881,0.10298003881252005,
 0.02960794115598886,0.16670539464163314,0.17676567149431832,
 0.10204929643478355,0.5497727769220166,0.5771681893940305,0.9883091741358508,
 0.15842642684148978,0.8102927401991704,0.3079800474700751,
 0.39696734916784826,0.45064095202980714,0.34162632829741557,
 0.5544374573934625,0.3216947514101518,0.07061158067681106,0.2928144475871213,
 0.6118307064826577,0.9601349841394747,0.611600110324122,0.07919982349997012,
 0.983227052923152,0.8610587819518155,0.4415922791454473,0.8204588368707716,
 0.8069327415763632,0.09053301635836164,0.7391999173893897,0.755303818595527,
 0.07379181997559636,0.6971216872170319,0.1061747175557145,0.4191972864564656,
 0.08197616114890538,0.6527493480506767,0.6464897329313407,0.9359293143903806,
 0.30458157755266424,0.6027585915624636,0.2542439222338537,0.467217480750317,
 0.24232671553504082,0.027513783592357344,0.905626294306539,
 0.33051418539665867,0.6277697410396215,0.4586974703492891,0.7672811474207789,
 0.3552983508386234,0.8263016975323418,0.7553565245914776,0.5894298023794111,
 0.33650134947583477,0.21442207558671367,0.4024750758496244,
 0.10742923476024724,0.4106185791206284,0.503046863791605,0.5492888310383672,
 0.5156902282783319,0.5279906217043195,0.8695597303088298,0.8242028697795243,
 0.9855263022176683,0.944173096760213,0.4299872488588621,0.9192377600910207,
 0.06377461424892406,0.03394386772476832,0.9396140191966444,
 0.5963659585698267,0.38315393770449657,0.390445417914011,0.8325435286056009,
 0.49120422949603015,0.5860238036598585,0.3439952939193285,
 0.040561309234091736,0.6799940038929633,0.7525463624081163,
 0.8599920014880765,0.1405868979476611,0.9601269741825532,0.5135394870878431,
 0.9486175701954936,0.14094373781235725,0.7130530413245264,0.6417986045444101,
 0.2974580116793827,0.25307695720251644,0.41274497180736813,
 0.6161887342324459,0.987905811600807,0.004222222407595511,0.5728623711929346,
 0.34420921890976675,0.2737625689652554,0.5921086806718,0.011521283555840811,
 0.5373654165183472,0.9414704404530672,0.15179533954796165,0.7352184924048119,
 0.5524013236972918,0.5711034894782,0.3864855812224153,0.7814839905294163,
 0.8120198809386618,0.5811134544867596,0.734466524070237,0.0860792612166894,
 0.5888671282213803,0.8272806437984618,0.9231358030828569,0.963751377467365,
 0.9452322268924874,0.2513767175725734,0.4588780730505594,0.38533244061208716,
 0.894115703276995,0.2585821557713388,0.7680796507817315,0.8130080551524526,
 0.23587477829409342,0.7776851785016723,0.436508177049456,0.07001058670721705,
 0.9761492959595178,0.16012768444254943,0.7017425910433135,0.1284557651687589,
 0.38935938651275914,0.5895701861386072,0.5505349522505809,0.5427312686360038,
 0.24125129722449445,0.27958246127979225,0.5096870489816919,
 0.9958003615200748,0.5498369759630538,0.3221256633900187,0.9790998988608907,
 0.09842884530890472,0.8784151139229083,0.7767176403236467,0.8864886929287484,
 0.4248867541773691,0.920023820968658,0.6263627964115511,0.5696724105158355,
 0.7203098826421762,0.04970087076103147,0.7014273652472827,0.5895455687163706,
 0.39199131537121423,0.47869213525404875,0.8751632584006305,
 0.3836642797237738,0.2824390980106486,0.0062621044047808105,
 0.5221690109767065,0.4343689937900129,0.2830195894363704,0.46982755405603616,
 0.07136077191559709,0.6109800893587597,0.5401025222883298,0.872130879380542,
 0.8715155755493005,0.7632735016261045,0.41276631504333905,0.2671257367513984,
 0.5815783897804822,0.42745347961804725,0.6125569161108787,
 0.19832601368431857,0.6505480330185384,0.8389689257825625,
 0.22946048934566543,0.14379065009340364,0.3728755339695291,
 0.48202203209150907,0.7187535969522199,0.41794913416700297,
 0.1281738651022749,0.7631986526774504,0.824071120108603,0.9266852433186019,
 0.32464797877272145,0.2758440773696422,0.6599802713957753,
 0.32136935526065014,0.9429593944906086,0.852696421609104,0.23746686260361005,
 0.03894163837200537,0.7205116228370787,0.32966399789911394,
 0.13064826970448506,0.8479245952891542,0.972801669809783,0.5961325718139291,
 0.4180423055966377,0.7352808024730672,0.19625952008967584,0.7833045045677876,
 0.8420127783652189,0.49738207445465166,0.8137797670750132,
 0.18748916805052562,0.5286355147055579,0.9213742408838498,
 0.25082028843009807,0.10300607092813407,0.3885511575100744,
 0.13978298909927467,0.6706310365938609,0.8268555772248652,0.7999615948677963,
 0.5318143732452458,0.8203899866478146,0.9163103367254023,
 0.028767879497774373,0.5653059208280049,0.2208287042043604,0.456726806145646,
 0.6646114641464754,0.1653258463719045,0.14200525283518017,0.3403299427260277,
 0.18258193727284755,0.20909339710774222,0.06326506081392025,
 0.9460819559268473,0.6395281393378005,0.8755399272829063,0.8261072831928091,
 0.6708949593951011,0.2761064454995801,0.2654873707723229,0.8809808006019442,
 0.3315921532782298,0.08950834467922031,0.6106842189484535,0.7041393876902571,
 0.7125299524312456,0.2990448411912192,0.3454314579332123,0.47897408099446137,
 0.537619282101031,0.15384220944157512,0.2327456703112254,0.19294150696532686,
 0.4160174894173885,0.591471752732202,0.5306587598787788,0.46310116486313724,
 0.20172543177806235,0.4678449434912262,0.17943995039469396,
 0.6653521780385147,0.40728937060231873,0.48067342739719354,
 0.2862920381898917,0.8062342569686226,0.04040796654245693,
 0.26550152927487547,0.8498983158874374,0.6556176047144012,0.7602917798765352,
 0.09609507676875928,0.732199994125771,0.42372728007204197,0.4038688943695491,
 0.26030778729170667,0.8065396130674836,0.24404369717493557,
 0.6015014554009276,0.8788743134836129,0.9745389009446508,0.6134869348204866,
 0.11011471954687202,0.8144822665026789,0.6580935105839363,
 0.49603337349261356,0.9316677312466674,0.2414421606508499,
 0.21772623325954144,0.6903947825212322,0.37216146229547964,
 0.4365525940975008,0.2732278304587288,0.5118684327487919,0.37490752908410163,
 0.2218502158683261,0.3723628182778612,0.43384931512184366,0.9403009751571155,
 0.5045321888605272,0.3925827651714491,0.43204077553782216,0.7352873025540163,
 0.8629828056762641,0.2226630527304856,0.2400497170818332,0.8748447992093906,
 0.693080041744412,0.8151757376261459,0.7447178545125582,0.794958507357802,
 0.9514556795618334,0.13151225194438032,0.07438453798500966,
 0.43175124962268363,0.025404400094908297,0.49225605444719356,
 0.28716913448347947,0.011199579843982232,0.06270780946526888,
 0.982715398040305,0.3863897693279209,0.3278468316309616,0.6922189412279861,
 0.1632995179449278,0.31707601468000735,0.9861216537269824,0.2556907312115848,
 0.04343238035598307,0.5802565903851822,0.14969153673001567,
 0.5995153963438631,0.5512079880604159,0.30856001708774394,0.1461410856874642,
 0.36404443856157764,0.032988524745635406,0.17156495982976117,
 0.2550889586475883,0.7460521043036691,0.19970257011029025,0.2896622348019018,
 0.06889626473898591,0.3532555978881452,0.825904358131716,0.6822522738560521,
 0.4841635653222921,0.29783237519664174,0.889073938427894,0.3806423422724077,
 0.7740675524661387,0.7253883156271617,0.02382429704697342,0.9372625088291661,
 0.4637253207731318,0.4607960531925428,0.6314084068082996,
 0.020612661539053945,0.09433847173509491,0.6942615076431706,
 0.42079426672448417,0.625874628927596,0.7790392991825725,0.9773984281974544,
 0.5203020071440714,0.09079772967257171,0.8437696896667066,0.9797806577651181,
 0.9838616455667198,0.6244201448076683,0.46614404385314423,0.3207578544697064,
 0.3063247475381695,0.16515943412331557,0.9165818018667979,0.819895997040376,
 0.7287204993960363,0.07183491994049651,0.7522925322507653,0.8998151439823014,
 0.6156999884806864,0.9407470608249379,0.4952436305386376,0.7953969151359519,
 0.28790962294928346,0.7935799236514149,0.7790375918674103,
 0.17545413318677816,0.35225528748842716,0.4951069223742476,
 0.9962875112537406,0.4897973177296938,0.4396043663105317,0.8890175648936113,
 0.9393264353278149,0.49592279944756057,0.9141788822510813,0.6797388324915165,
 0.03611497264614072,0.5012061392882966,0.27956807912922677,0.554830964908037,
 0.06323567439330402,0.1841046720890379,0.1258706061871765,0.2745924118684162,
 0.1219360161177594,0.43442679807368434,0.6174750429501933,0.4930768424592379,
 0.014126782525372583,0.3423858614194114,0.9809146774179756,
 0.5634119157754213,0.7496773656620849,0.33536413717222624,0.3807785484180404,
 0.2756642634343103,0.40069367250579013,0.29590356368484594,
 0.05164094382943907,0.9263879856794153,0.3059655368704065,0.6658998555580193,
 0.9095601882235437,0.15039731450048255,0.1933955695612567,0.3015672453301652,
 0.28957052420794693,0.8057903587737378,0.09510519239791326,
 0.05004770084914656,0.30902748009642944,0.7950187214123677,
 0.8908546985039291,0.26704570833880636,0.7033663656805208,0.426170453788274,
 0.7701309575155162,0.14321637822773658,0.7429945042007257,0.8858410891129354,
 0.013697306333463888,0.4357976720294191,0.24152189232890375,
 0.24290661314781548,0.5296311836698502,0.6272347140235265,
 0.04147495935585355,0.7435208458343568,0.8235702149114588,0.7828544721102906,
 0.900002578456084,0.7858654086511138,0.60632944201867,0.7895327371147595,
 0.6658427677465952,0.9767096644688854,0.7315558348734905,0.7777370987127117,
 0.8700782512218197,0.7696732807647338,0.0011407376272781766,
 0.3923612466243911,0.2626003663493315,0.5099716841537115,0.29719399272349156,
 0.7240698819822393,0.09881426109990443,0.5994949918468518,0.9913790171836449,
 0.04025974592513337,0.12969313005999306,0.13493519705779422,
 0.7306650892471824,0.6165427680772216,0.9465282473923841,0.6854422713648051,
 0.13781178284859286,0.7881973092809125,0.15725311919410445,
 0.5807949723058623,0.039359263407527934,0.054401804105163265,
 0.8161108609263564,0.3857789610874218,0.11242098161138048,0.4287216754652865,
 0.19803161841806693,0.8518147959589067,0.44094327508197706,
 0.10494381916599127,0.8952551606353824,0.22128152881127405,
 0.7205149138897206,0.059675081292748566,0.9195243964917876,
 0.3237454332087095,0.5341966360273437,0.9805519636434438,0.16828887128546421,
 0.0603617752655754,0.4936536934661604,0.28628954134520934,0.8636220381480959,
 0.45956693402949134,0.9079773601051289,0.08068903016480566,
 0.36153073057270546,0.41281671872454584,0.9030824316769505,
 0.6865136055222146,0.07936165084323632,0.33721138219239377,
 0.2751853830827936,0.9096290293830605,0.1020387862951978,0.6417872745071507,
 0.5338970280344127,0.23868646926560566,0.4028074373230952,0.4687607527111112,
 0.035665571733625834,0.5426655451481192,0.48711046420937465,
 0.9454591874754943,0.050286772389874224,0.7869160194308925,
 0.44572956562838184,0.348038753574501,0.8925289492941044,0.7185416811478931,
 0.8514437120458909,0.046345720248703426,0.003687893619953031,
 0.17495071485675395,0.6060136302062811,0.18485994135656525,
 0.8413645598589582,0.06171544843715604,0.6246729707706551,
 0.22402821984885446,0.553784700794875,0.22350964510243354,0.2914149772160405,
 0.98113336572066,0.6025269371804229,0.15393296356393416,0.5819027885427976,
 0.04424247149836902,0.8882476076331591,0.625195550250913,0.91455640710578,
 0.6698870899315194,0.6217347732979359,0.6057854917248526,0.9039915851400611,
 0.7271778156905054,0.365646828545235,0.6334368204303282,0.9769458930164506,
 0.2029987269522936,0.05789493768531728,0.26135830966723717,
 0.5737283824991226,0.5874433262537653,0.3947414194007418,0.23617907758134193,
 0.3574447621004895,0.6168239278861612,0.47452900086256933,0.7547769005471394,
 0.2316438735267753,0.4898855186829526,0.7731288908650978,0.8640772371909454,
 0.8010195713383202,0.3422719725586113,0.7315665157917447,0.6823384808708142,
 0.23112726297897324,0.5920755097736019,0.37324530752984253,
 0.21868871655334488,0.5234566120081307,0.3451939212108678,0.5338139929950503,
 0.5446732291875814,0.5636765967496467,0.46435173245244143,0.7407809188174014,
 0.12461551100507051,0.5674943548183289,0.26214418531529615,
 0.31403971817094567,0.47476056179093384,0.712360858096285,
 0.10071903109311908,0.7222881763921075,0.38045879231520685,
 0.16704688303703674,0.9571829614221374,0.5039637210818416,
 0.35409102428686534,0.8844402713540332,0.6858157980086992,0.3144228196423673,
 0.4898884998965496,0.8614618948293135,0.39788588681277104,0.3313106308150524,
 0.08382900187525699,0.8340136145973754,0.6532714814105836,0.4439696672813259,
 0.865772034326415,0.12299922803616425,0.11169430899999089,0.1523335907057788,
 0.1290520191632698,0.24456181025797152,0.7105649484053521,0.7581227201042599,
 0.10095211516635483,0.1592935286081063,0.3382676881094939,0.3744143048646824,
 0.1288280235933742,0.675080599144275,0.6209795379718677,0.09764777266395142,
 0.7256717206319061,0.5469036772102289,0.6538216012304823,0.41970861260835435,
 0.6572354228549269,0.46950349842729366,0.12320760562935384,
 0.9043243321788491,0.11778786566545874,0.04381393408825629,0.916869309437427,
 0.28808618335368186,0.9414655352420519,0.7819283860024072,
 0.25587897964389605,0.46131837403454823,0.11164206260999476,
 0.9401811342729343,0.3875227357285026,0.12455797638085542,0.7123392362503358,
 0.297727004024291,0.25892903080029683,0.39701786275361006,0.670633401313007,
 0.25295624556334984,0.14465958897530395,0.13695445037104337,
 0.8073599630605363,0.6492839544303868,0.6615971573859902,0.212399956470542,
 0.9890569078240681,0.1365310294786234,0.29790299995167613,
 0.12290624576785425,0.5225445451570959,0.3796697392918986,0.8517504278073724,
 0.9082499704869378,0.36938706003141497,0.5404469942649711,
 0.30648561645928685,0.9026500630576978,0.7482424679602143,
 0.11962236488392164,0.28437023581443943,0.06085144807816634,
 0.8026376401233998,0.008808769824356522,0.7113645820763851,
 0.9960871823776439,0.7487075236591797,0.20041097537798191,
 0.027787101335688402,0.28786770369444215,0.5187814287938758,
 0.9583491023206523,0.4514521405183525,0.9326615454522867,0.7248875817562968,
 0.38542185114840255,0.4120312559740553,0.7309125969325435,
 0.42524038243914997,0.7407630306370607,0.16765675758027387,
 0.10767182337065495,0.3666744006778464,0.20551730803535684,
 0.5414616336172241,0.7879318382223466,0.8672365503526989,0.6432025029557217,
 0.969061649321502,0.09440945973689596,0.06353073367912587,0.6513090924444086,
 0.025245159645800652,0.6736446862504417,0.16014159644036619,
 0.9452093024534264,0.2522423297275187,0.6554995965532816,0.8684543696261765,
 0.4357698466436596,0.1559659051518325,0.9700258879904694,0.5245160200507043,
 0.970038512569479,0.8335456826150388,0.972283888191904,0.32312484784093165,
 0.9200007927265186,0.8158310952689968,0.9350829603712177,0.2960274299625232,
 0.25299646931478936,0.1581278121148405,0.31147808827443635,
 0.1149097838458355,0.10015660430004825,0.332933832656675,0.5337476059164332,
 0.2858697432021291,0.17907323541208953,0.5538466723301122,0.1882924281025138,
 0.7500324735659533,0.7911622432247398,0.8393964815440159,0.32140812257532025,
 0.6090692779129774,0.5673365195146438,0.3301881634035021,0.2203149632494733,
 0.4323590685227704,0.9193918213949538,0.7913111243362703,0.9643872513264676,
 0.43926970027399914,0.836173508870792,0.5123910863753214,0.4493081173308381,
 0.3252530318452673,0.08017025886035634,0.22783923742239454,
 0.10035350811721133,0.4104317958061833,0.3780543131414055,0.930121913695374,
 0.7047397330423896,0.482598789696563,0.9180246305621271,0.5346496463305651,
 0.6404962905749869,0.22338881328866067,0.890182455304588,0.13394838276947907,
 0.43020919337276486,0.6255711575003566,0.8326777288943616,
 0.45130691688338065,0.760768509191199,0.22521101722654246,0.5933970099767445,
 0.22708462658983597,0.8655554628994199,0.5549670824323205,0.4671480607861528,
 0.21848069005707882,0.4310347492533969,0.46522852079243915,
 0.7677130498360686,0.8703302132730892,0.4067102423041572,0.7916822071903618,
 0.07791698749549159,0.4984801060924209,0.16106019756866918,
 0.8184313303634738,0.7575505966368982,0.46988772699989134,
 0.19480967444879138,0.18899417414553432,0.6608060399355501,
 0.0028662036194935236,0.8261678003691239,0.25536308406870434,
 0.9864783317703345,0.22518594585363616,0.8895435753928814,
 0.39585173213778124,0.9801942627407443,0.34176957525646745,
 0.47354573944437384,0.7626658632911458,0.7471990017088868,0.3517268420578047,
 0.11085505909189419,0.8306400593460563,0.1691800646701933,0.5338712152136142,
 0.2648928026629508,0.5125380301593844,0.4704736325079374,0.1624390674390398,
 0.45903639181194333,0.8147457325584545,0.7110030921876472,0.6698663289544313,
 0.6681020222622533,0.5928539394350134,0.44560046652543317,0.5909392062258925,
 0.04376842927052882,0.5220741397500716,0.5475943585742854,0.4614396402191768,
 0.8829204233258855,0.40920110171898505,0.6062891907648555,0.8808134233267384,
 0.7865943198238541,0.48447104649569717,0.16433459964226727,
 0.1259690111255708,0.7637932143069446,0.34066549334099083,0.5372808593054185,
 0.10972968780763048,0.5323020122151286,0.23060661277672534,
 0.20518540228837745,0.4247449054721536,0.7914857777922419,
 0.45095561737847767,0.9696015150573933,0.5536027527350605,0.6114628190614406,
 0.879652889277643,0.8617732972245041,0.16990693751682884,0.7860883549111733,
 0.46669499937289627,0.1894505971476913,0.7416530210189387,0.865962736794629,
 0.8009177614297297,0.23213966599868363,0.29082136141488046,
 0.9217070619794585,0.45896391193675967,0.6325178561212014,0.7980544558845236,
 0.2563154586048424,0.7247856805581053,0.19496492612725524,
 0.39092228432678244,0.17531627836498787,0.7534736309229536,
 0.09962021030858315,0.36489178686949875,0.3845432658149227,0.055104650829779,
 0.10050639258983718,0.4813181081030745,0.7761235562816415,0.7989136127866088,
 0.32974392683116127,0.7621105886877481,0.32061365836210154,
 0.6505970444758357,0.2543025809254207,0.6872748045312298,0.6610643686225742,
 0.5828313982975446,0.9671510623032616,0.6959749647742167,0.2973391365872522,
 0.7551193972729584,0.7043210606901796,0.7611962413153197,0.7473525248754448,
 0.01670185434729976,0.12969706918225354,0.06092786179894838,
 0.34097100728906304,0.647879239905071,0.37459330385691936,0.7031248585432857,
 0.255986810474511,0.4415279959963192,0.3727872939338872,0.305764524126791,
 0.40441027123921613,0.9705731936610429,0.041728116034555174,
 0.5945995867043204,0.8956497098595898,0.3990468401794899,0.7573295548781861,
 0.6131353800843288,0.0930322838078359,0.9968623148382364,0.9406298449021355,
 0.5777818568381657,0.3925408174299658,0.6290421710656187,0.6995359690980911,
 0.7574391301968632,0.0565210139739275,0.39693958277178554,
 0.08479957849530462,0.3265155378224329,0.9705015736674941,0.8267192808159483,
 0.6562583768478777,0.617515698812772,0.3637859075717278,0.38510138324831744,
 0.9010910830557177,9.504546619431009e-4,0.126512434738273,
 0.44240824129231404,0.15112973162466814,0.6385379110300575,
 0.9404584873889128,0.8012808111045604,0.4453035112704913,0.8413489127868938,
 0.43635090983805924,0.4921493944479991,0.7584740945417576,
 0.25681810218096324,0.7158208788108869,0.539822308012222,0.681101695094192,
 0.9833378550087519,0.15388528381193667,0.15124892581291605,
 0.030868343765111272,0.29382334138414445,0.7114364321761378,
 0.8980000133243408,0.3938724273533307,0.8745543956423806,0.2315018352998316,
 0.7417759933575485,0.8059520259247933,0.8121075841604336,0.7516655693445419,
 0.24333869465089086,0.10123859921140532,0.47275903314230905,
 0.5900908059281923,0.6130252076933166,0.1571217530485154,
 0.003849561676233182,0.006645779638807303,0.013274359839235972,
 0.40615375672796283,0.803578180248558,0.6916406159172477,0.645052908824319,
 0.17591746052484225,0.652581349177009,0.2712954447020999,0.2226732341396921,
 0.42901602673739736,0.41361871055360944,0.3958321178053452,
 0.8426699361857207,0.15345712884529505,0.44229192712107634,
 0.3689238261033714,0.7695750332304342,0.8327741490796824,0.9469219015694789,
 0.29559012348796754,0.826960633652182,0.41487766221562994,
 0.15185218057514682,0.08042904647470384,0.033887167126345874,
 0.4530929788136149,0.39173968926679714,0.5622246254000196,0.8571452788105306,
 0.26662188248689056,0.33980880671542735,0.5258307573708881,
 0.48952418550182164,0.4546941027929805,0.8302438300847717,
 0.14331895032933983,0.43054234761539234,0.6622653331723907,
 0.1561456603985767,0.08799906848373651,0.9786511695850306,0.2571888164058923,
 0.8943284319493227,0.4371315349043432,0.17140697527396342,
 0.15893707602194918,0.5721306118430957,0.7392671275569398,0.3219480572444522,
 0.19484457499776675,0.9989989714247735,0.40451400343690935,
 0.6040326001084718,0.7155727763355906,0.045799973532847815,
 0.4042401334135284,0.6268590853962135,0.21406963158522774,0.7338191429810085,
 0.2572618757126973,0.9289807125118554,0.7342618290016847,0.5106695163669842,
 0.20577474210058577,0.3946399697772296,0.9661481257270843,0.7020851463185254,
 0.13655455709627973,0.09225531188375236,0.7663658639398027,
 0.7368535806003887,0.5509257415487114,0.17480223394747685,0.2799200305596933,
 0.06306757105808825,0.869701484140485,0.06905986757828464,
 0.15972413380752648,0.2562771399345001,0.6512983210352663,0.7618143228259872,
 0.24251079514233642,0.823111183158496,0.8747511399608504,0.4406333678362394,
 0.27297319921986885,0.7249575903536302,0.5886271411384872,0.4452296061836385,
 0.944293371348019,0.007532281500821081,0.21307444266439635,
 0.15290505132745413,0.026788637408323934,0.21999077351907959,
 0.4018147588584501,0.969085428697965,0.728681880858772,0.7587266832063534,
 0.6566228025076631,0.14066670850044927,0.7830582864850949,0.5966771954603902,
 0.4733153422757028,0.7973978140129375,0.6421817126047307,
 0.013344197643925804,0.5242238991315462,0.05052628593960984,
 0.28521256223591207,0.3368464765601461,0.9085855659431289,0.4909873682773902,
 0.6147364080870656,0.5131443836982124,0.5812837771822502,0.08354788376902578,
 0.0025470913688460106,0.5040595034112562,0.01442044591682512,
 0.7161553375535556,0.5287855256125646,0.37490309539582156,0.2206466459241947,
 0.3707123710816088,0.1256794181014167,0.27236834672952637,0.6440798896844441,
 0.5425792230880603,0.31090738949739993,0.239304219383203,0.8178178628555227,
 0.8409770618665363,0.9276344307754923,0.175383381819053,0.6190552983939636,
 0.37901919515208105,0.39777527412593094,0.4084786858520002,
 0.3933988495646965,0.9362264272685582,0.195656422810915,0.6526847611599003,
 0.9481813676920718,0.5446725147143018,0.4604081838865983,0.7825288266767527,
 0.08727101300461493,0.8704874456449211,0.9203843249159999,0.5985436394596705,
 0.7927708540171866,0.5640899324923317,0.16401108435218958,0.1918617037907664,
 0.7188666291821895,0.4331390464395741,0.22375181735124416,
 0.12741336484320842,0.5217868123334521,0.7587443579180376,0.7609108114079214,
 0.19531069515957977,0.7680537640152143,0.16759470501305518,
 0.38692029013886087,0.24423713509051836,0.6973731389125776,
 0.49532161976502775,0.6647259442120299,0.2546998742252351,
 0.13720045151966198,0.966079059323515,0.30274920138833283,0.570797684862525,
 0.9724617432507818,0.8672398735425774,0.4520989117523996,0.1996174645749994,
 0.6155584715182398,0.5741015309205786,0.7534841817551599,0.42159967070005466,
 0.3093333983199831,0.29800878380179574,0.043534617231898776,
 0.17663440278489984,0.5330078231084667,0.9127985901122563,0.6465219344529112,
 0.24472317068448923,0.09489668146534191,0.9355589590897843,
 0.7440112115094588,0.8929441255709033,0.008551338104439976,
 0.38092202335442304,0.9548235846890741,0.7240722877209187,0.8678974809171404,
 0.04168915535448525,0.6514494723191332,0.951217505019339,0.6127372371329874,
 0.1563306063764215,0.909361406187721,0.8901914239661005,0.9091730398317832,
 0.15647957032064652,0.5585506419976984,0.24128114275487877,
 0.5347743684392037,0.5897565262296262,0.23295885733333543,0.6033000868812917,
 0.2153971358639577,0.6654899467651214,0.021806741528648477,0.991303092341084,
 0.7904551214564739,0.24759029271591448,0.22876457793481864,0.420221923432163,
 0.22661922252753142,0.6565429053605176,0.9788845719981948,0.1604923290167929,
 0.9492471076534783,0.03389990324372938,0.06698197786908455,
 0.06106536129116802,0.21104582358281543,0.6432451961390871,
 0.31458440699094814,0.8166480513221868,0.7951057763960421,
 0.12536576959061674,0.41619128913066517,0.519618081843153,
 0.15094005133246824,0.4299333741716609,0.912036381127721,0.5374744612623369,
 0.12960588301621756,0.8304247149273501,0.48782637721872923,
 0.7594088451681115,0.7756234767655612,0.006486364193119987,
 0.5247086058676185,0.2646084315820568,0.09006404476925434,
 0.19732463597941918,0.08489901892826768,0.1854382474399977,
 0.17496349364574593,0.9179646737473681,0.5737939903795708,0.9105301511898516,
 0.12981880372682708,0.7211948588468783,0.9712692130538852,0.9815278835006032,
 0.8465500549490913,0.5667444389539976,0.3126397378909256,0.507196605396019,
 0.7193902225087969,0.18811600818737628,0.8633263808183175,
 0.12465113085568946,0.23416900594491152,0.4478714774726449,
 0.4894273508043323,0.28531428715810636,0.28966300175637194,
 0.07685854182775409,0.5919933303191972,0.4947004012984084,0.7571794668854455,
 0.9516940398859971,0.7201079330910343,0.1422893468642944,0.42586578356646465,
 0.6976659965384191,0.14790821930053089,0.27488957082897547,
 0.8136611976949231,0.5546510702624556,0.028234428381793686,0.691208813373545,
 0.9099399303964387,0.06403595353465441,0.10810643342630288,
 0.6961820467115493,0.624490073708902,0.6013485003396895,0.10808113592808288,
 0.18718454657158823,0.3766344647427099,0.8579830328758195,0.8320193802335807,
 0.6279633277275394,0.5017398407576426,0.7492765911047181,0.9887590891701938,
 0.2350331851692702,0.8383500287612132,0.03455052582276463,0.9513962905014739,
 0.5744040326693557,0.4954167145501599,0.9806447880364536,0.07097448524010386,
 0.7213889993181025,0.5036832701193599,0.8818787364153348,0.8459763052621283,
 0.2697115034172355,0.35706184310546696,0.7476168677467615,0.7435166334221617,
 0.7914643603504014,0.20350953598322774,0.6925703558422079,0.6375951468419674,
 0.36799881935034506,0.6583967632402415,0.5768726738605178,0.2771334779986365,
 0.24384790971150594,0.34121849695942685,0.4063245207754519,
 0.6802503758459791,0.011651936841452382,0.8857428710409381,
 0.5046225442021558,0.6529230412689789,0.7427157603279918,0.07398224751673566,
 0.09151277731294016,0.4055921232138272,0.6363853582085308,0.2686009041806636,
 0.30096097481061634,0.44408643498785694,0.04747867384021798,
 0.15343565145978877,0.4414399429444873,0.723327940278103,0.9267178682388223,
 0.38611256268190597,0.7447580591058748,0.46703346686937375,
 0.7151183875030559,0.8296720650685668,0.7507773695018344,0.46405682199659326,
 0.279834353628133,0.10442079801014037,0.3305427396606675,0.31415375171513926,
 0.5456266218940238,0.13166371055325854,0.3205926750899434,0.8938743274346695,
 0.6999527983662805,0.6042554231977085,0.6699306896901438,0.12276999559925461,
 0.6860559597325254,0.5047075508299441,0.8156060765889102,0.6542270513085265,
 0.0708948843722983,0.7094328641965449,0.04467970538390087,0.5596171015794473,
 0.7728001359480858,0.4581087187251657,0.3994559338853896,0.8153491746109371,
 0.9689075765222843,0.6402093654076934,0.88749487830675,0.6156798477071874,
 0.4804104942286451,0.1865103041815358,0.42921635098607624,0.9950513662485118,
 0.8438118252866353,0.9490220480711807,0.2047208183042155,0.47923270578161614,
 0.20636945729367528,0.6320843958728717,0.8179652263471092,
 0.31949107025304524,0.01956063734367053,0.7622077757281437,
 0.8527269739920167,0.8276669782012793,0.6660940311752616,
 0.022689896364386564,0.6467057008021148,0.5334418082580986,0.652389276821831,
 0.7682469664980462,0.22260507825425202,0.5074327588919226,
 0.18448644962985983,0.002830246154964633,0.06710140097132533,
 0.35997194850161984,0.5962530273312117,0.7215457229422337,0.9227709987631778,
 0.8236412479921953,0.44874595149269647,0.41958199193539814,
 0.5948733040186993,0.15155693749364152,0.25001309201192434,0.165648188015598,
 0.07829951216252762,0.5539343858736027,0.5307774419627938,0.9871178179410459,
 0.5604433074306019,0.5922264905407962,0.9159635457776248,
 0.016876078780032056,0.4626763576775822,0.6887253329765507,
 0.9284174483656201,0.4816283777219038,0.6204431024034041,
 0.006247992737253849,0.8439350618302354,0.9138077895174123,
 0.25020097242134287,0.7729921840121322,0.4384199901848901,0.6617522824893906,
 0.11065798499299084,0.44752791717241613,0.15058991699933544,
 0.7957071774752904,0.4510029864021846,0.4395808109290985,0.15212133676932493,
 0.8143013182101582,0.5372250749305335,0.5165788865538595,0.11419190347318542,
 0.02727122391366299,0.4588789628963823,0.9247855018506967,
 0.43500189023173785,0.7144903360882591,0.7247021038764885,0.5945571716101172,
 0.7055240893279038,0.2758088857926766,0.576076692055913,0.4661796195824338,
 0.3661771624857485,0.5415479977344706,0.6906397206596848,0.12952650581511638,
 0.18137358615354215,0.6455116957098177,0.04992368889159515,
 0.9126071608639109,0.3434253154199137,0.665694511280949,0.6831605861065108,
 0.6231957545216517,0.9035755788397233,0.40383886510816913,0.9637765620177028,
 0.7847188823796136,0.10232372167984471,0.21068426011408403,
 0.16301673319379584,0.7758627664657913,0.5320957169352679,
 0.07587391013639289,0.049624339241215454,0.6576418120731095,
 0.7069981926261036,0.6446003328087505,0.9027846327544138,0.31989225880070804,
 0.200695521521861,0.8228807146441357,0.9889767574204512,0.43015035580832583,
 0.17093887547797793,0.764665953060342,0.7017156064604405,0.9020222682210932,
 0.7825605841387929,0.7470620259757477,0.7337333896836529,0.8042930632565006,
 0.9018778427833747,0.8609447122800128,0.9329633767913349,0.7903523821757203,
 0.9278255734187038,0.04028003558569415,0.497672653356913,0.5902773081088771,
 0.10244054449671758,0.803390921692095,0.46118865362925476,0.341521191844646,
 0.3154614654711586,0.401239536497504,0.5625680829162029,0.97638538020488,
 0.8949120095558604,0.3739122614318581,0.06020637671726481,0.561292064932164,
 0.7177988137164591,0.3224422016049263,0.24241789539397296,0.527101638141987,
 0.9749794804097236,0.6909364519834443,0.07034753085290257,0.6096306073992168,
 0.39126702400153823,0.34845227611376406,0.24232363539478152,
 0.23975247597822313,0.9205443757147154,0.8739885738523513,0.4395130346154906,
 0.04779529311435049,0.7072048411511741,0.5111349353260183,0.8822005257949921,
 0.772799252959396,0.22872310818465813,0.3540511523598193,0.23916159926997482,
 0.43363607741195387,0.9140705972508324,0.36049462848557,0.2960533500655954,
 0.060158972476045736,0.3050866331125036,0.23423165970621163,
 0.2896348928959235,0.7594465998876759,0.49209612905906,0.8916929404693769,
 0.971464618712005,0.8712396358159585,0.6160294170378338,0.02601042474777082,
 0.9910738507449283,0.6128467074045759,0.9993600262418156,0.2861164749368007,
 0.16624446485749522,0.13048508664957414,0.38250395539065374,
 0.12948542846216948,0.8931539831848343,0.8406600904670607,0.3757728351675389,
 0.6923496787917875,0.5007647190446066,0.766634736805557,0.03105616288754276,
 0.5200978259934841,0.07115260920556243,0.8773600075899063,0.2728513107154654,
 0.7768029459276787,0.5703417461296845,0.9169767612286879,0.677082881656456,
 0.46043359315384247,0.28038393430734687,0.7907310795563388,
 0.21028592268723667,0.653706021899386,0.19625499134426005,0.6929093056966292,
 0.24924431918272938,0.6490545408472024,0.6635736418834582,
 0.46706661379347314,0.690479701677415,0.3435883479581071,0.46652292005175533,
 0.6254582587721986,0.010980137969510362,0.32017932978841124,
 0.03135478377690448,0.3456909008590565,0.7201457812114764,0.6715674976924897,
 0.619370735139014,0.08661037669855409,0.6299677204653762,0.5560881444558404,
 0.7486792986319478,0.017134534699369985,0.7912201984516278,
 0.4599948747234748,0.11930486523504213,0.44070574116784134,
 0.7367954398142204,0.520881440673806,0.39707268996026146,0.5606918242419838,
 0.7432793761157551,0.07790187440994012,0.0829553040225417,0.3815010157977835,
 0.16779078069801012,0.9833217271274899,0.39804338460650524,
 0.5104223714852704,0.09323045311958511,0.8519992740336689,0.7231574999581611,
 0.5770318154016033,0.38792115869689925,0.7843901768151574,0.7052095802185419,
 0.8653840307084094,0.32652647034450666,0.5760733242132398,0.5305149945839925,
 0.8141630508691752,0.7557317004876208,0.5881518751157334,0.6130409955876721,
 0.5149106518749814,0.7883747494526064,0.5213639744731108,0.43838034816085725,
 0.8172597934657264,0.15761398928944825,0.3836089335801528,0.7033459748595016,
 0.8478358897477596,0.5664462640900005,0.5843138921908055,0.5782600730685915,
 0.9902162758685322,0.772901407278449,0.054862987249653106,
 0.06170484651351915,0.9840145192815263,0.8337238889408096,0.5510525514500573,
 0.5132751148634862,0.4681343082142616,0.5360876936320915,0.9292206486284365,
 0.056172147473506406,0.5256866044345679,0.10303603860573385,
 0.3915134232213301,0.6012542682453995,0.2245655748022204,0.22236589020196817,
 0.5204842496096002,0.5339561716264571,0.8769212416446296,0.04592347276420594,
 0.3137400835592715,0.7273642041370446,0.46292381378343705,0.4165206296877496,
 0.823760899315048,0.7846460083788442,0.43305993229182216,0.4243180461771905,
 0.01890360181059636,0.5587522579227491,0.6352347117286914,0.7751593217944098,
 0.11253802667673407,0.4663405757765401,0.09545514481416817,
 0.7199533815567075,0.29511328119592417,0.644460411238134,0.9924534088918924,
 0.9788896643940536,0.9042945600253329,0.8114315156082511,0.8341550923327009,
 0.7245842034810797,0.9257284747127537,0.6668230438903591,0.4306254090684576,
 0.7642537968718865,0.3658173016682045,0.5972512012216876,0.10820482212744387,
 0.24949846374696494,0.13827686989327514,0.7176715068102334,0.130681275787499,
 0.6997160065026675,0.1260850631138748,0.7549269161153753,0.6999740957164976,
 0.13069564406913925,0.1141058862458334,0.7024037823572454,0.8117892476616946,
 0.6285542729275615,0.170008910620027,0.901918357592318,0.35748189554427,
 0.5943672699689159,0.16168838568898736,0.4335589078343507,0.4119490904872293,
 0.6989618144346084,0.33111407758570977,0.04667188050942628,
 0.48950624855575753,0.4522125613950473,0.32602805833629844,
 0.45012335135131687,0.23581180934392898,0.600904223275492,
 0.16248524417288368,0.6780386201760654,0.06675102874575123,
 0.3611357928926152,0.6122730089848804,0.8451747188678125,0.34635456644297136,
 0.9215713419146314,0.6943601622363907,0.5362655326548367,0.6748462866476284,
 0.9491114412520224,0.8559845943596367,0.9910309898732432,0.5030568096498466,
 0.7520033678847773,0.8220390125201914,8.542710357224426e-4,
 0.23407285314182025,0.0627682857492401,0.1976598715500416,0.5707483418741138,
 0.9823280184644164,0.12148895665778003,0.2106964676234646,
 0.41468885924101995,0.18741841099123846,0.6790879609661854,
 0.14214934523138423,0.28459696965159853,0.16851088168325012,
 0.3171106014617453,0.48695554846980615,0.8001896092260741,
 0.47290017779162974,0.9098831775492718,0.44445242407740493,
 0.8436710059286513,0.7221994711132609,0.5342369882355879,0.19147408861172344,
 0.285547087016074,0.6108905605389822,0.24361876465678844,0.16996694449166327,
 0.3026054639113501,0.7012635325305064,0.936977468303041,0.8043858504155351,
 0.9938771167207369,0.7895837661745739,0.4786328086402303,0.2909462452122219,
 0.32504592706249913,0.7685580204906262,0.15603045470527221,
 0.6860259473442286,0.5355963967819625,0.4633630094914273,0.7318602851929905,
 0.3828434224284889,0.09496571174197044,0.24547479141191086,
 0.08097620117249427,0.06258879511368987,0.19841858690839997,
 0.07930727413496064,0.043154850211172935,0.5021808663272571,
 0.3298035946515234,0.7927292853958141,0.9132324119614552,0.0374124789117013,
 0.7816073910729981,0.45249373155279926,0.8129369551849525,0.1822045786816323,
 0.3963553186805283,0.8617936677908082,0.9580246641577378,0.30084036140169046,
 0.9653258225191276,0.17035982823404794,0.9409860341506742,0.2701837594905516,
 0.19506412611590085,0.5684681594531508,0.41635112329112667,
 0.4705790155397285,0.10827616876776558,0.5165368264671191,0.3741369415641552,
 0.1786469683585301,0.6679252247695007,0.3937618606112844,0.07752730422218779,
 0.2758767459578052,0.09604193226146707,0.6142461955046941,0.3397010829015179,
 0.07942096903175833,0.42726571524166956,0.9359412919880122,
 0.11556345840222315,0.7515360949250142,0.7111312827573999,
 0.47152812005605804,0.912392867865623,0.7996823890985507,0.320251985315388,
 0.10886454403939705,0.3125166571465202,0.03305624496692516,
 0.41954233960359344,0.4503439918620282,0.8027318754826082,0.9989994694462524,
 0.5978633823293666,0.7312441123040789,0.9999654583302419,0.9263006951712083,
 0.7637497548302487,0.8132065345959109,0.36366053872589277,0.1221582653462352,
 0.6363639669801021,0.7252623992345553,0.18220950902632305,0.8356923163698524,
 0.292597398290498,0.5343924982533735,0.7549847796028988,0.7565453128248459,
 0.27786229762680714,0.9028960793675336,0.31695829304384904,
 0.5866478709431732,0.3474130912474255,0.5471634311591422,0.5212494049487947,
 0.6738865364155654,0.6841394482463652,0.29519500703622537,
 0.19060936757241853,0.7118758419591308,0.7835281766728069,
 0.08095824362136095,0.7886817988659038,0.2839852010706396,0.6370487790070575,
 0.5039201230935911,0.0012904159493777323,0.13732332745112485,
 0.3173562032708288,0.9931271938812453,0.9119516589275668,0.20691784406136338,
 0.2713084945871287,0.12471741461132724,0.974899650583953,0.6020739100483616,
 0.9033615578910934,0.6843158674415623,0.5299629125984401,0.3801745335370972,
 0.5115005572333851,0.15121847032343172,0.2681663429580544,
 0.27749351898966856,0.041300464160134887,0.7991824871333233,
 0.6990338335446364,0.27367154934032123,0.39539436049066523,
 0.9225623825988886,0.027336097378142354,0.8928505967073264,
 0.27856280136710343,0.6686365597891464,0.9746318522802063,0.6034527565575727,
 0.5341212154399549,0.036952456019542446,0.9690779172094602,
 0.15195032264521158,0.15536005136129272,0.7035160832827926,
 0.33745468457159755,0.8644401267745623,0.9497440512682108,0.4999555710079111,
 0.47841179543189294,0.23711640809918644,0.7657029200157373,
 0.2969830276237583,0.7683964427189371,0.41668260900675413,0.7916175715787604,
 0.5961800912189515,0.10854170455205725,0.03711613283534676,
 0.35293357770963074,0.01921181506665881,0.03016091820605371,
 0.044775915897951823,0.8829800979792287,0.4550572077060093,
 0.5152597252795079,0.5069864417757853,0.42764131886023027,0.8177946037205855,
 0.1492645629102728,0.4235119304695276,0.38341490349827145,
 0.02885688224507932,0.873342156249624,0.21494260283419342,0.5015990234998877,
 0.26838836399405774,0.6913674380703445,0.9033455417175349,0.3427809720552555,
 0.4227710786095392,0.7783835066463249,0.4795277139865386,0.8397155299444636,
 0.3496289777847782,0.5422189168586183,0.13533882825621746,0.338187335119304,
 0.952054674278735,0.3411820548067699,0.13562276096247294,0.04471331521629118,
 0.2913452562705836,0.771711381880181,0.888721472041926,0.9695013536440377,
 0.635961528428422,0.25709246057068147,0.9203812269518451,0.40194491668862564,
 0.8997245020575965,0.49243006382274823,0.9404833238778397,0.2582788192081943,
 0.6817500167673448,0.3188206273623836,0.4551333623823216,0.6882672312079512,
 0.726421982458069,0.9496994766588743,0.41610539060904217,0.5118760708209797,
 0.15446088664792312,0.24557354455941027,0.29481264111173344,
 0.6439548697541838,0.28224076346125115,0.7840854571414952,
 0.19537657824955224,0.28191353469318314,0.8856437292782473,
 0.4309869710929567,0.3525281825976706,0.6061634411266973,0.5407128594351426,
 0.9125119129746199,0.8442617013547279,0.15201729429803557,0.7956573577871247,
 0.7682752121361973,0.07805758995768008,0.43582612584759495,
 0.9287737063506074,0.0787209580031425,0.8600253042566874,0.8342185546765701,
 0.559217562388085,0.2853022302856115,0.5398519389278112,0.7038538003706571,
 0.9019524940246144,0.5687459975847051,0.355434503571055,0.17402740509148584,
 0.02326315850445604,0.9598347451376935,0.28793199891540244,
 0.5580862014474938,0.6418182158547925,0.05696669432408741,
 0.017628632908000874,0.8858906156453987,0.7026328829296531,
 0.5902652677866745,0.5628133005604172,0.8633084370236594,0.7475121793746946,
 0.9801834468069089,0.9344571173008642,0.5489522683543109,0.5328351936619641,
 0.9778881039501883,0.9658618270455152,0.8625979283377722,0.4970959710239974,
 0.5399415444396889,0.6127427379923112,0.4034518907830792,0.8863701081731558,
 0.06627290995313206,0.03874423231467983,0.2662678672993679,
 0.3709097904609998,0.757606351726997,0.5479807691951775,0.9578895214250119,
 0.314645816936317,0.5314152209477736,0.9652545596921037,0.5088681579345212,
 0.12513896293099414,0.6951761763660986,0.9198029107637293,0.5426715135275457,
 0.006930771945762304,0.9301556045917391,0.11942329399963947,
 0.8805475478572156,0.030692318319962664,0.8660032076491471,
 0.7577899148340455,0.6002927105581346,0.5945735708515889,0.6043139494997213,
 0.45178651972177764,0.5408072765987108,0.885498343818892,0.8004273084990028,
 0.921983047916147,0.9398699935918814,0.9868296079491614,0.505362293016426,
 0.32974309789405043,0.48115991489075216,0.9883703266921445,
 0.5034277415947281,0.9639433382361772,0.42873416945326215,
 0.20125322298638715,0.8804487606100838,0.8523807946735464,0.7840355580930269,
 0.18404879469416524,0.7301689872939077,0.2643525910823178,0.7612693059563326,
 0.4842099246201452,0.16580961634674696,0.34744629031236385,
 0.9618222056595611,0.24218862038546607,0.9881862489465538,0.4656678760802835,
 0.9260732535576908,0.9657037187295551,0.11706864413006912,0.5574789085472859,
 0.7373714314375807,0.48815129351423847,0.8761657916618638,
 0.12265258864794049,0.6129534411577815,0.3693966685437937,0.9630467407326272,
 0.5998181521579233,0.494283387053006,0.06653871334122508,0.8981824095063753,
 0.2304956188606203,0.6193685146152795,0.025683216955641686,
 0.8285771767331037,0.5569448302290313,0.34802773192999925,0.191126975124709,
 0.028336978307508143,0.752472281951385,0.7975471442623675,0.9808719758870457,
 0.8496124710443371,0.8370305593959715,0.3229898551488709,0.21474850348860264,
 0.5524489718981405,0.0011193816758098274,0.9429391893207455,
 0.5267650822321608,0.42121602947868864,0.7914715533893909,
 0.04825692064862219,0.41686899140632194,0.024026161106682986,
 0.4627377986776804,0.9062955882441401,0.7457917247527796,0.6022870106141188,
 0.7057530942897239,0.2738502509077615,0.464288389782876,0.5207536564168442,
 0.3455615449970446,0.425700530386409,0.6304563747834705,0.25351641446771733,
 0.9589617249707368,0.1669421590982354,0.2108476908389324,0.7390844938507045,
 0.49492681446336073,0.08257641581777841,0.05512952780355995,
 0.6040384010798513,0.4802713191629797,0.6723211654985349,0.7738430665165209,
 0.17521200284647243,0.8688046194086291,0.45888135320392875,
 0.07203251292936597,0.7001026819695293,0.7720406276242242,
 0.37384746080774955,0.9447289218818032,0.1208060736844142,
 0.37374422438997823,0.7735506631670288,0.23573419489794334,
 0.48511667511627443,0.9595539528894588,0.7301783892643984,
 0.26223853352504567,0.9389913284856565,0.6722387790126901,0.4145900214309979,
 0.11513130886405543,0.6716442121683205,0.0913762142384863,0.8602206071775566,
 0.706544996776999,0.261540499908404,0.19101328672196616,0.10121071316874264,
 0.6036938363326739,0.7361807982452242,0.03727528171658301,0.4606542434060097,
 0.6990541365429948,0.7894154836465936,0.753571051725126,0.518364664449902,
 0.5335617550232228,0.7096776565408187,0.7405604228715519,0.4000432347152204,
 0.8096004478557173,0.8935646813370439,0.28120131991465225,0.7507549282685962,
 0.32609460269619095,0.4719993967544035,0.5825786113023649,
 0.16397113304604338,0.3239187102567602,0.7421722825395283,0.5023719677544893,
 0.3901435990097297,0.4146742548781934,0.39668063650281715,0.582138147055771,
 0.2413616612647015,0.8684922849306955,0.4393663155997962,0.4329620711148042,
 0.5409531413174825,0.2335358709072839,0.08282006468911574,0.643513294623546,
 0.9538557116781208,0.12520966071284123,0.939155261924779,0.46983895990910896,
 0.5814150902921063,0.8902805708678606,0.20563779859365994,0.6293630509273025,
 0.7835479271778535,0.2674740045867139,0.27239873240352985,
 0.20945378470856835,0.0670570462063731,0.7157313551373197,
 0.26514559952180294,0.5916351103875102,0.6894065480198857,
 0.28055177593170466,0.608213898313995,0.5260713657170111,0.3557365303744282,
 0.24427458578242633,0.3782827669410227,0.326932474155219,0.3389521317837635,
 0.7306124008870858,0.8012239340176596,0.9392722271829858,0.515047632423046,
 0.169011964535686,0.22904725599913744,0.03515776138594229,0.7738800277673776,
 0.6070471298889314,0.7740504947531546,0.830325644809257,0.6577773269092078,
 0.3900167113859524,0.552128159673882,0.02339634776068189,0.7628506597753695,
 0.12721480360252257,0.8568928751133282,0.05042862128824788,
 0.3825096429119109,0.2667687504744407,0.7309617624498814,0.07557685176740003,
 0.4771614921007785,0.2887481355842887,0.7550725647065907,0.39780963337011677,
 0.8417203202629243,0.15971823351917358,0.26347343650103217,
 0.9266609762032232,0.16987421317179996,0.11465495394528324,
 0.8248041996396898,0.41818091903039323,0.5228766054398911,0.7582333381242151,
 0.42227282397491606,0.9099863144141067,0.9780937746276893,
 0.11082214522994926,0.6894579042132885,0.35924815297029544,
 0.11788837494079374,0.42835362239060304,0.7494730630759465,0.859256931469283,
 0.06868550112008387,0.6278679232702095,0.8756313410948112,0.6315311989272084,
 0.531448867625025,0.7458374881218274,0.7599661485867507,0.10484511921441408,
 0.19299780582782977,0.44759110035737415,0.7900258350074895,
 0.8620883972239312,0.05643961227317151,0.9160814276769612,
 0.08934332972995718,0.21399488575763415,0.4973977958794925,
 0.2057318339777373,0.7212440185028762,0.9262699475258591,0.2123058664440769,
 0.2617857007184897,0.3976313673827119,0.9844204025174749,0.9661341365096643,
 0.37685436954192975,0.9716687383434207,0.34956730251549306,
 0.23773544362561727,0.9551483080088198,0.3770103523319106,0.0648473078521169,
 0.7944226448362838,0.5232900029218701,0.08457029658764881,0.5357030588429182,
 0.44177961284602874,0.42563918748439267,0.5692278405032122,
 0.6610603270185985,0.5620894355878496,0.07282574925131546,0.5424673771210261,
 0.03320661675364722,0.6888188329227253,0.9274802423963421,
 0.11818979071601143,0.32867656640299625,0.24567949073455997,
 0.196543971670339,0.03182065833905501,0.5565564719139796,0.33519453900846463,
 0.9879804521408678,0.14832181684579537,0.36980562857142185,
 0.7241021435593434,0.09192227357604288,0.7717274347091718,
 0.33906014738263246,0.40024873502173297,0.49883866875549165,
 0.2924210936377786,0.15046251640419128,0.0033367782887797626,
 0.6438276690858606,0.885165770078489,0.4072716788572054,0.8568183267428618,
 0.37732101702609044,0.8505089288000643,0.06470837712624966,
 0.9474033216468387,0.8677648738500192,0.25384429046142276,
 0.13460112691204207,0.0020617929260422496,0.856699333387749,
 0.6933683279492201,0.6688968092840847,0.8259987664817282,0.5530269184431896,
 0.8835918154028842,0.9991024861583697,0.35104822651344314,
 0.11035319437503799,0.0799181286593843,0.5692386632255603,0.8120786438806136,
 0.7160104091545206,0.12844085756271073,0.5534750373169217,0.8125762012153979,
 0.9590519033458984,0.03162109987160033,0.8115622001953033,0.4561393798840797,
 0.7367708252139199,0.8777880639172048,0.5347336635160596,0.06023376300945649,
 0.06150061975074561,0.28763312387716944,0.304029583262157,0.8274120220021144,
 0.9503366739552684,0.7129724601720181,0.7524402348771237,0.5858994795543323,
 0.2431406130703342,0.32383101616499577,0.3699837135808305,0.1235214196632084,
 0.8504243946686773,0.7160201815064671,0.05932106827577999,0.5108608125849872,
 0.43345069847984585,0.3888873434192246,0.14220130144629284,
 0.4756365175721108,0.6934975825393328,0.5674867671794848,0.4023933609224116,
 0.14517211721789147,0.3554558946379617,0.073430604543953,0.7530121556801527,
 0.020863087388367507,0.4491748444419855,0.499601205147178,
 0.12678531573754226,0.5360261095797465,0.29828358557829027,
 0.3123196709241427,0.6527999987977393,0.5519087037003294,0.23242142159260892,
 0.20330252430508822,0.13128897957961594,0.41689602016427785,
 0.15322075566491278,0.5810091628739437,0.710560849601593,0.5362379150229839,
 0.6649633282226722,0.6403748406750515,0.10878772537819648,0.4704297021818584,
 0.4630378235327578,0.05666230111378856,0.058527937185342105,
 0.8144305133259697,0.3697519220176667,0.585091819619607,0.7554047589210591,
 0.8333168658975318,0.7102578791909897,0.7380294838810748,0.7059430385245263,
 0.28743192371630777,0.0186365967810197,0.3530710614730139,
 0.10314180164609998,0.7900833242312162,0.4366569023251642,
 0.31793453012590234,0.6726546915630159,0.9851986818505725,0.8705945964903427,
 0.7644973413302625,0.45180424726624224,0.9712085879102151,0.9037872715755252,
 0.8598284630688432,0.28935335455527667,0.2768019493856255,0.9401385886093326,
 0.0143919444790086,0.7580307556696866,0.8825841896422448,0.4045982344497585,
 0.7536915007478875,0.2028070966772565,0.8696670113018442,0.6481145173151427,
 0.19470597954787838,0.9123661899772183,0.5118186480162993,0.8272839738459213,
 0.8032142305915202,0.6805697617435337,0.7910386669900561,0.454191532570533,
 0.1669553071625266,0.6604368096315971,0.18900119876714205,0.1378237319982626,
 0.45523805555416363,0.6270421209943339,0.3826097163254052,
 0.38482582351886774,0.651348963523429,0.7906387126059116,0.19423003190384136,
 0.06382253960271544,0.7298846210435694,0.4092758024882366,0.3756863177016513,
 0.8810728686859104,0.10708570379634219,0.9898346264559051,0.3701462534858293,
 0.3184508805805991,0.34939906719375347,0.7603950837587027,
 0.41027617446570963,0.35873570882156414,0.6359046223870611,
 0.5593067687608763,0.340660022809259,0.8397941196289912,0.8319712988105599,
 0.3028521320064981,0.938932366675431,0.8716606244036629,0.24457685250951577,
 0.3190519724926294,0.03293269539798027,0.13974535415678002,
 0.7161239336583439,0.9428605037801754,0.07001168355398646,0.4862879963083463,
 0.41473581124481884,0.6929785697904098,0.35485387322652984,
 0.7283011228410242,0.2597059083272043,0.3482996183505839,0.7850907594077089,
 0.9333013700008504,0.3507959043746848,0.7238595747899603,0.7190139892761462,
 0.5557453424641264,0.6312006699727992,0.9701258225304187,0.44517285399089035,
 0.6308516110550899,0.004359353275256828,0.6272631068929677,
 0.3160985277246292,0.587086149893861,0.9345463907253291,0.1771922138277503,
 0.721540197603682,0.5291010386808855,0.4282573637371496,0.7099086162692413,
 0.859167365763735,0.07483757518270873,0.7446670675639282,0.9046522627788565,
 0.039452232037326285,0.742994841683673,0.8112063614210249,0.775509819368166,
 0.40456866601899866,0.6832024151466248,0.8589482511673372,0.9872659105839354,
 0.9847138655375993,0.11509906137535408,0.5474052143802931,
 0.13853726073217076,0.6510126811295947,0.011104675225689475,
 0.5505909448089499,0.22276035608835254,0.9142148156739128,0.8889904403288642,
 0.9514260866988749,0.15938068170141118,0.6032102079826838,
 0.028039677981922195,0.8789936823844595,0.8712933381257979,
 0.5334954314509143,0.4994710038069531,0.26277363751510796,0.8165034047556732,
 0.7951231851198888,0.1507440350086785,0.8587576919459378,0.8241446401626776,
 0.9455381082834033,0.6574969616190973,0.8888616318125606,0.9380482630467342,
 0.1522871036713267,0.08869565469081153,0.24421678414817183,
 0.6263639952275843,0.3583461185528085,0.021558497479107652,0.538026724988909,
 0.5658041590806293,0.9311879486088537,0.37414781426395005,
 0.15879006894864722,0.9528140017431248,0.019622225353403744,
 0.4558544867645917,0.7715683726141171,0.19100898990991444,
 0.11580684121377449,0.6024990502193933,0.15521798812139875,
 0.6422651192373422,0.2683958150947373,0.209694127661473,0.8545995932267899,
 0.792318626007126,0.7046179227695497,0.24909696166934214,0.31501723293835393,
 0.3914910369763027,0.13773730622360414,0.261452646847516,0.7423891023017533,
 0.6491406737004454,0.9614002563612944,0.04809203895842795,0.3934584141117504,
 0.9975241142857757,0.6453227584574953,0.35687571026519216,0.2971882933193599,
 0.4842213618674641,0.18539320538083925,0.9441506572609858,0.2669216227037249,
 2.405447611710798e-4,0.7487004972922117,0.4006224870213877,
 0.22767614102565692,0.13340077016081853,0.7282003964820405,
 0.8809463823679501,0.9133824080376912,0.3208545523901041,0.2717424281109617,
 0.9562660226397173,0.3282338270378774,0.5350328043424113,0.40606429366672814,
 0.8349808065947361,0.2059490441370947,0.6658210091130141,0.637721785706042,
 0.21744933846416237,0.04220299922573878,0.2782230823680868,
 0.42217975943548525,0.6058755962605581,0.25897265134925307,
 0.4635313975059807,0.7990588367384736,0.49048615120579253,0.9168710079970694,
 0.6267485276312975,0.7522079911032686,0.15656058577677123,
 0.00895257463217114,0.3861522962927876,0.6421195915111242,
 0.30001701574378503,0.22574400838828068,0.424037936152612,
 0.01975663793184934,0.461440373933772,0.038438683073952484,0.236582584031246,
 0.9228418003525447,0.1347273441772936,0.8584799955197469,0.12803309644594385,
 0.8598800221108752,0.05331632826896249,0.3990692078386936,0.6657474118033243,
 0.0116969419262265,0.8302822185495807,0.17235658170978785,
 0.22349090100239766,0.377230989413584,0.7834988432448267,0.559652501415912,
 0.5379756532342176,0.8707722543654248,0.14539202956145747,0.5800332121116835,
 0.05094778953063006,0.9834204903581987,0.7302793842621061,0.3312865684601872,
 0.5046749960720877,0.24277777510829157,0.5453922888027597,
 0.29432858618343016,0.16977228792804655,0.44758036372340537,
 0.6108794104871759,0.3490524216140293,0.1083885195623332,0.2148487155567449,
 0.3833927459571139,0.3235955963945223,0.7543417237085832,0.30837836453176026,
 0.3661212255725853,0.733557779617753,0.5181833873893202,0.6934886040914134,
 0.6783707432418961,0.5138634226667633,0.10361586549890545,0.9729259023959294,
 0.7194845351622509,0.287718409929119,0.09183221642789996,0.7666850302009579,
 0.1050715601514931,0.7917027803737281,0.9922298183894758,0.2902777042436001,
 0.4712742489335894,0.2501140850002548,0.585683292665849,0.08670809171092075,
 0.22492929000287376,0.4298768682963161,0.02851712160243347,
 0.8877623933521562,0.046227161310863796,0.37987617603129853,
 0.3604282761548242,0.6528991156555591,0.07144400464658573,0.2619529354905551,
 0.8565908704054819,0.6911228098342652,0.38690918731722423,0.5467895192161011,
 0.8269643025549346,0.9898570944052469,0.3188416237554281,0.16101973384372525,
 0.7048489766676652,0.9870283966876453,0.9245867838129784,0.6498184569508315,
 0.9349574751868615,0.8192745086373913,0.11598553666803779,0.9199176050251459,
 0.7450512174508669,0.5184855626255203,0.8949188853850767,0.1331324833131322,
 0.05953339963319704,0.35769485484351904,0.7277653843385975,
 0.5718359992018217,0.5097928531383228,0.6067202773705883,0.6972036684126248,
 0.4819575499679668,0.95799791057864,0.29769761578783416,0.5067407010745071,
 0.7520930355977811,0.8084926132493788,0.8723726372511499,0.44924273290608296,
 0.7500337483838314,0.45131724852495547,0.7328114006088038,0.7915065349396833,
 0.6105405417141461,0.8890679667175722,0.559186065909596,0.266068151693865,
 0.5178256528349274,0.10668328031354868,0.4244890839188874,0.731355326586421,
 0.6355807323642279,0.21878054147955284,0.5837972431746057,0.4735269572656595,
 0.3669443871823518,0.5614565452146425,0.7822472493311011,0.19928603834894565,
 0.8236769029727832,0.5235568545865287,0.2672659331433078,0.4140734806119145,
 0.037791410085691646,0.4706371290766086,0.01962638219933832,
 0.7439766699703876,0.8987222327051535,0.19927156632889031,0.6785875085737025,
 0.35253823018611485,0.49206150324596687,0.5131300069033031,
 0.8810745274291142,0.7381139256984157,0.8361214389762507,0.2568512184640016,
 0.20959890020417749,0.953952280267064,0.9659180377497442,0.12049560159110284,
 0.5930857533020721,0.9230828002352677,0.8345750628949906,0.12693126800868726,
 0.20622923172900087,0.9265244467886604,0.33649314320109347,
 0.9445906462832859,0.5126713811431277,0.3627113859836606,0.874180126120107,
 0.9283606311972589,0.9666265214124519,0.010612732089885779,
 0.8866030351336047,0.6736671332579792,0.050924396771526714,
 0.5719570062427332,0.5089580734117991,0.05323199243569543,0.4315266119430472,
 0.17118606586920726,0.7041366528770543,0.18978334923660667,
 0.34463410344871004,0.050127519748921934,0.4676356641823759,
 0.07552895122648529,0.5042889348471449,0.46009893935924795,
 0.6879795452918114,0.9057903507844909,0.5077727329534002,0.7075282150571143,
 0.8305703182257902,0.010364079111061297,0.06544256593869213,
 0.49738724236975185,0.16431209535733815,0.8074779524593463,
 0.3393502909238596,0.28552310474356535,0.7068893090295895,0.9994714390137923,
 0.2725525785085644,0.329376618936118,0.8025859059887828,0.15274407101046572,
 0.8222452183386337,0.9825215644380385,0.6167714233401653,0.83265679299577,
 0.24042916157439131,0.15839158571114698,0.8737549818575034,
 0.019256797660323732,0.32961136507308664,0.885666640170176,
 0.6632534267288258,0.10369724512474299,0.2065087068986735,
 0.43749729006758276,0.38971136162047326,0.6022554926991085,
 0.17560752874946206,0.41498143377768204,0.25144501108155237,
 0.13013803985577155,0.8434775525793927,0.9761519785210752,0.785073048598231,
 0.9920784719973722,0.1803106800301172,0.6070272027965357,0.37688360796112985,
 0.9807336803069522,0.7014939975411288,0.5937175382603237,
 0.024543939727498798,0.8640207080620965,0.9178629226194827,
 0.038701163661528315,0.41848477913285653,0.4304101397809783,
 0.30765565731028044,0.9505872577526636,0.03741468791650737,
 0.5442823654567833,0.7285317079575,0.2544835885665989,0.655830608807832,
 0.19120274263923953,0.35655059502180564,0.8001012821066319,
 0.6046531647637207,0.35775134701569955,0.28442870429020606,
 0.3152516809461591,0.6497923234076846,0.6685644119412804,0.9812146137727435,
 0.10575701638119384,0.5501236925086412,0.15903918403358608,
 0.08377261700562622,0.5448804116601542,0.7375735357352196,0.8622976577766598,
 0.0318464089943824,0.20984651996728232,0.40623478344746733,
 0.1480181448381397,0.8794130264416895,0.7004084925056998,0.758758884414067,
 0.36478850718048095,0.3742407027928174,0.42031487886132324,0.897920832100114,
 0.5783648619838653,0.00625865696660266,0.7036501405271014,0.661219002632444,
 0.4667955448224891,0.669290221450462,0.7098354846566725,
 0.0058979068530434775,0.9890940811811179,0.5620963860814389,
 0.06969256383371636,0.08377955059764775,0.8712173805966935,
 0.9900396159374671,0.19055612994932314,0.4875423504891292,0.6029185382103511,
 0.270772180828607,0.3751130248748358,0.11848125948837107,0.12133062527222715,
 0.35107897635290364,0.28329358519157943,0.8660427945465837,
 0.4430928923676968,0.6616829486300223,0.39652795023634535,0.8468602270080402,
 0.7519390461950964,0.08572475897613818,0.23931084541239844,
 0.7025135707616125,0.10146615553888894,0.9481147772570422,
 0.24427411735546734,0.432652682327744,0.6882139350615177,0.5042639325764066,
 0.7083736845127271,0.8804938538503774,0.8821184568502739,0.6907385293247883,
 0.5003219206314161,0.18488988372225468,0.34805804560043674,
 0.42741337276681557,0.5803519764023432,0.2707313844813769,
 0.038662419484615795,0.3049011472020583,0.9179921739409076,
 0.8694415823404514,0.7051544695519818,0.4647167683758191,0.9237885174356718,
 0.9540871219379701,0.8667714036809634,0.5336965947019652,0.4151267444766522,
 0.718399578273819,0.5920110854818127,0.31440739895703096,0.05200687413618443,
 0.7528741456168269,0.5561359980265843,0.041935771395814614,
 0.5290786656334755,0.20138409161517812,0.08281478008037546,
 0.5757304679424635,0.2085752599890749,0.4714543175932496,0.18095666342963912,
 0.8641722313380941,0.24152854000821433,0.04650572039134504,
 0.5465999311743454,0.8313548916041616,0.3503952327783809,0.6016953382777824,
 0.6372000988709676,0.15028359048787154,0.3999157421822819,0.6783582245472163,
 0.6151310482763162,0.31822332549099097,0.8829268467810097,0.6666478767728711,
 0.12359216445301213,0.5779903263741722,0.7281208116642128,0.7152977714029076,
 0.20930446331740282,0.03052645309580715,0.5438333588921526,
 0.7218055862531759,0.4436464255580126,0.8126318863899049,0.8933132496091227,
 0.2893046440704182,0.8065572635872361,0.3932073674098765,0.6281856874062152,
 0.7680208330825575,0.4350620280058177,0.7440840375923303,0.8475034856136138,
 0.9528604075948021,0.08891393234163081,0.4995863323727594,0.8702135563668383,
 0.3209375459952878,0.05592171313824945,0.6888845751764006,0.450276534322257,
 0.18549476353039152,0.8858303591459212,0.25780474264238007,
 0.6544728797850258,0.5001663376723418,0.30398373537571544,
 0.0036614063104611905,0.3863431617337494,0.5192587862655231,
 0.0822161971602251,0.9862382332135184,0.9090349359057806,0.8720078603742495,
 0.6569999467013748,0.006212608739829095,0.09729685545984412,
 0.3058987319742734,0.4096948050274366,0.2772601745583705,0.8351714536045881,
 0.969966058061621,0.28877035087342484,0.6871111226830875,0.9945730143987864,
 0.5445623953325209,0.16096328502354285,0.6074708974237213,0.3350470672111844,
 0.7817946123397789,0.971225386505752,0.08398219945176022,0.7486103188782289,
 0.004955606240293675,0.12851858174884756,0.7510578651996985,
 0.9066992683832231,0.9738112740143358,0.105047772294108,0.9909832428058308,
 0.1120033792131302,0.12637196521536165,0.5305258815249474,0.9473712476096433,
 0.9900907889491435,0.15752615307134787,0.1557946972788692,0.6267146775648895,
 0.3406193766143819,0.4772567862211681,0.8553464914306411,0.3307011419932886,
 0.8561733394964673,0.964079998286364,0.22193142056764448,0.7414203447064921,
 0.7214652594968847,0.985327371227153,0.5093239172985901,0.9055254391717572,
 0.4054988287428788,0.013792077687261206,0.8366232651979054,
 0.5909194426914266,0.6708088849160272,0.9577716023001203,0.6781139387701537,
 0.27238947847435613,0.46760211559924136,0.6887460060760736,
 0.9845768960980976,0.2927232407522108,0.17098017623941297,
 0.16686354188556374,0.9460020792523809,0.740125969450899,0.3680893363079387,
 0.4015181198832567,0.25632064094586915,0.6736325307850918,
 0.48770171165869547,0.3209273202377335,0.698797589080323,0.8664196916575007,
 0.09400602101534183,0.674707424927266,0.3510562269367178,0.07682134313188504,
 0.8980979946196854,0.16930443419737584,0.7411051191858404,0.2128229076037268,
 0.8735503968893739,0.2918997299253383,0.8488992467486247,0.16310425055008404,
 0.1303623648349953,0.7716923480206848,0.8633921875752909,0.21822275655682033,
 0.18692855972157152,0.4879424945377626,0.12801414178995996,
 0.6589278890252208,0.1005711359214746,0.22647561645127756,0.1936405150986632,
 0.9034530753915324,0.002157560989263785,0.9225551912906459,
 0.8598058658513802,0.7390923667128592,0.5313376083684083,0.1614215521683573,
 0.801418335585914,0.714713726107272,0.3855946301724785,0.9044138658330826,
 0.3712410816848367,0.4112961164035268,0.7532741023556563,
 1.400634395400946e-4,0.23697653172360322,0.7167674921789827,
 0.4957818959039437,0.4174430136249869,0.38337260658080374,
 0.32096172094284503,0.09678388732037135,0.4801939419261384,
 0.8297942702253688,0.9822715776457611,0.020545224422427788,
 0.9641380583673023,0.9606976301299888,0.4569858459015511,0.08245153513592796,
 0.1011560904522596,0.6979742267442166,0.5438610247214056,0.05119170374797266,
 0.4042265498690817,0.03281787840140926,0.8796651723191681,
 0.002754214011001177,0.5905289391996582,0.6736366656362743,
 0.9797100371848909,0.312443858192356,0.4543763712789185,0.8829330727789959,
 0.8355728462797807,0.7117798022931239,0.7726877465487932,0.3355407317804563,
 0.8693994316734127,0.4457392708193044,0.3856958259382106,0.4450363847118737,
 0.36686858996398086,0.3271558221962898,0.3928612497159063,0.4656077181963171,
 0.18791333729706972,0.04372233593411312,0.18795919467087263,
 0.2142116964579196,0.6974783301980088,0.7846166441684024,0.5724875390642745,
 0.1474013190124932,0.5469729344612996,0.5279934519219934,0.05508070005302912,
 0.4984119136923556,0.12653750340321013,0.7126170751092027,
 0.35693999228264073,0.43976917697562334,0.871559037441391,0.660162418373522,
 0.8036440602310455,0.7863117151117683,0.7805043515120198,0.9080541309763253,
 0.9962499378380125,0.24460428769027676,0.007332846630224266,
 0.0917982416794545,0.3248740936984281,0.752538411755815,0.6664439096123874,
 0.897137475071305,0.5933401875444889,0.07338281299092608,0.6318551244302262,
 0.48227498764393295,0.5908180542230197,0.09005145811441007,
 0.8642247651055813,0.6412187940638291,0.26322023289187557,0.3539691075390907,
 0.8240350722913179,0.08580058149568015,0.9902783004140119,0.5737840294857506,
 0.10629573327301,0.9389661597703335,0.49154521185657996,0.30004464040701173,
 0.007526998792633299,0.07192622627140566,0.9793979176995902,
 0.7792986130246539,0.8531567425128636,0.017296239697301763,
 0.16841887371132291,0.30312619071337243,0.5259826426700764,
 0.48352914126201263,0.054276678520731925,0.10015807886730299,
 0.07213734985067088,0.4918509392550193,0.02552354395439671,
 0.8774851426839895,0.5052833521570186,0.44219215510733734,0.2813285135640635,
 0.8503427710088066,0.6800395263979805,0.7790384442865768,0.1560369557103103,
 0.8206332011568145,0.8157628870240348,0.7689197073130496,0.7392718081747766,
 0.8396799241393965,0.8842571942666829,0.8109767327299938,0.7893064122659849,
 0.6881668967026677,0.7050819709545598,0.6443635913716198,0.1291146813652515,
 0.7503556207080547,0.01645960139773317,0.5978239359301714,
 0.021759236372751012,0.20163490462001588,0.842866576473821,
 0.9638220412489668,0.3768181911269022,0.9451633057681134,0.33230961803939263,
 0.1951816635942688,0.6104688095892936,0.5192916021409647,0.45613606652631844,
 0.5388645102324622,0.7781877772444746,0.7054805693488234,0.8331665242822809,
 0.5458536224008232,0.6150528227613089,0.22873127076883315,0.4105660425269999,
 0.1213562046064548,0.013942588106717224,0.16168338683888062,
 0.8913088171233312,0.06107658709773789,0.5548671744239598,
 0.03200503552164047,0.3484728311528862,0.44689565406485876,
 0.9413706562543124,0.17874103421789167,0.49327795966372323,
 0.8346495269940991,0.6565486587005501,0.856084922481253,0.4120492474844406,
 0.6920125208393095,0.8378534535856794,0.9683568409278943,0.8842784517767849,
 0.2744300126898118,0.02398056905166046,0.19929711824154406,
 0.6049066474250496,0.209690302103146,0.499658841325586,0.09167821503431917,
 0.29385651985156913,0.8745460178307228,0.3956288469767937,
 0.20354777504739885,0.20869725466301792,0.16181930635024988,
 0.5881236153603941,0.021538044703327763,0.5445216904458892,
 0.1526613636376386,0.5627110555660481,0.001781511857906537,
 0.9165147529211054,0.5625173979331246,0.5521440915718696,0.1423731876654889,
 0.216161928889262,0.6461091931996469,0.18696256754347407,0.8061402319801023,
 0.026412422465716068,0.9044982161460896,0.1866799368382105,
 0.30634064895371393,0.5586056963822603,0.6231490413464196,0.8216284836232599,
 0.24381888880435798,0.8795833767375392,0.2728262096015218,0.5877494370977734,
 0.18521877093131767,0.6084366861631133,0.04231468759910506,0.848512518618401,
 0.6926384288959757,0.787284672359001,0.20262576867940607,0.9343659070958721,
 0.8790730511063578,0.30131245407855134,0.6010434786988648,
 0.12409012907559702,0.7927866524530924,0.7641941211309777,0.7124634020676197,
 0.5813433282132133,0.8744720076018937,0.20497534357541247,
 0.055379554563858124,0.30644285906160484,0.38579187907548596,
 0.04003837502639096,0.8240528193727737,0.524702338887874,0.9913536173940429,
 0.45241963143875563,0.688856629244191,0.546943461352361,0.15176767238238043,
 0.44043001370034707,0.043923001284862484,0.9273316021091281,
 0.19657127815392983,0.7305526604935976,0.04291062348394026,
 0.04720553625975732,0.7484923842388891,5.799730532645153e-4,
 0.04495409097493097,0.13221488834331507,0.932092841244036,0.5417568030039797,
 0.8633793158688636,0.09547774536082354,0.3807101530146261,0.581913243921059,
 0.24797840208555189,0.7939771613316264,0.9564883773947722,
 0.12836765286686824,0.5931451332457992,0.9482100022374882,0.7045366634308919,
 0.8902572162128355,0.7316494036073741,0.39754441372469174,0.2247504391296724,
 0.8120860343315123,0.42102831541126895,0.5312280602187793,0.3460188379589434,
 0.47006005622733393,0.6466657217437981,0.5780592916959931,0.7108954097327597,
 0.31172390827971497,0.2637835541769581,0.4222035794333444,0.9578637460265854,
 0.5069909671259969,0.9141805110637563,0.9207401462482929,0.8652300810131751,
 0.9800938363260577,0.9424941126923704,0.7408085629693871,0.6839960412016319,
 0.31286779043518265,0.7943430964180594,0.03268587452722349,
 0.48735629153518556,0.18354557338560096,0.36562080660745866,
 0.3251043578359556,0.5994368438667455,0.028720036829946838,
 0.8079191222440604,0.3598000654652771,0.5251154230124342,0.4016486435479385,
 0.061579851926586326,0.38183843609763757,0.26825456566273775,
 0.45553906054662474,0.48926314796128034,0.9397608098195466,
 0.16164427901085276,0.0779244379659303,0.9560244200228075,
 0.062435332518684716,0.45317826563452446,0.8070859577062676,
 0.863594182384543,0.5060333926865953,0.8397915751052781,0.8619849658312702,
 0.6473403690684824,0.6130346890851207,0.10806037919152733,0.7502166544447129,
 0.0432156404531977,0.806261678799352,0.06105822708410413,0.2933357123932486,
 0.7543722141464103,0.5249355199130277,0.9797045736904013,0.6259188518480941,
 0.0444966535520358,0.3461469555954423,0.22651202703193185,
 0.03671986261090199,0.8262946680638823,0.12876790271533856,
 0.6244460331457815,0.3889468758924849,0.5745325661614582,0.7073698422181403,
 0.8883686379009668,0.23772094980358494,0.362439460232036,0.2867849483282867,
 0.07304953100270095,0.5895918522813399,0.6787389958187998,0.9386535532922103,
 0.24159154594387633,0.05752615646047898,0.5647466375718193,
 0.11008259653330787,0.2837767551919622,0.769428296966117,0.8466577782188193,
 0.42555985152325493,0.26630236759362014,0.7231847026254462,
 0.03887696761458226,0.6979402947981505,0.5949036592788767,0.8004611000405716,
 0.3631820953719336,0.019980256503970284,0.0017339290182813993,
 0.3125417098826684,0.4231581616243336,0.042477828929835426,
 0.8540615049028168,0.1677727133558511,0.2549837410481124,0.7077408559989924,
 0.6481361687286833,0.23042612433987342,0.7014685589843765,
 0.19385025724386162,0.954343675529089,0.8620861103416535,0.4869809607889395,
 0.5126325603448845,0.07541375055589461,0.6710064236174608,0.2100017409708833,
 0.3006791307406925,0.12858264542125342,0.6935305865186694,0.8799968680609404,
 0.680030719797808,0.47865656066040807,0.8970683380788506,0.6596464950047293,
 0.46489409091516565,0.5376604957270855,0.21559425370361995,
 0.6081416131780976,0.7789353257300171,0.47130111646635675,0.8106148885745429,
 0.446769170716801,0.8207196268802288,0.9320214101431697,0.0918649102461152,
 0.7432285535226832,0.45776720074842614,0.8518137854171004,0.9489140247502802,
 0.7710305607462006,0.2322870974136796,0.9953544050827773,0.9565069771954637,
 0.4234308864480093,0.5061231367802168,0.615486382775504,0.8868857769745434,
 0.43772189028736475,0.2232894454988159,0.3242325437479783,0.9087092303461057,
 0.7009527600778125,0.010338163841086878,0.5772818980192871,
 0.6000675554332812,0.5152645603066274,0.4953803667596588,0.2958153836916573,
 0.18885736085360671,0.7758167017642434,0.2209252201605587,
 0.08352351157319571,0.9596315275713652,0.5641213617325038,0.209885728636255,
 0.16663564299180844,0.39045275540267554,0.548702485017497,
 0.49306157004951834,0.6902926544459871,0.3996024429308115,0.5181181551744118,
 0.3103496840705311,0.9823698097044684,0.3798457593124994,0.7288853982249004,
 0.6215939020855415,0.8975025232208091,0.09524674485430706,0.9385545368590156,
 0.09486432653375454,0.18733307159197865,0.20166713824227345,
 0.5102039282706875,0.9910242180764183,0.07822619798564778,0.1955557615620317,
 0.7169471238368772,0.6616023845688027,0.20419781215853516,
 0.23224033402272315,0.9910508155831059,0.7098115221260799,
 0.029968665299944353,0.46278557812385235,0.3601805194892448,
 0.13806822836512744,0.1069262348760377,0.6127767788133871,0.8531899968099446,
 0.9723407456396843,0.6544852954748365,0.6545198347903513,0.5785778879090807,
 0.19992890875459213,0.937564540720133,0.3442093145626457,0.36258150227356634,
 0.3725629508125837,0.373570571877602,0.29254760536724655,0.11410608821465718,
 0.9228606366687165,0.016876895660459912,0.5702083363953785,
 0.6483610458027643,0.17594969110175251,0.6187210001970133,0.5144310518073,
 0.7228451675008498,0.27628340169433174,0.8773082858824197,
 0.40370366894826526,0.34416045154172537,0.852600313506683,0.8298568026149518,
 0.23220774702913005,0.20915154411784587,0.9914294606123726,
 0.21006240550237898,0.3063918332929746,0.04984825678893323,
 0.4662184844623448,0.6857688866896559,0.6163384917514054,0.5993958950483815,
 0.38238538350661067,0.4357696350255642,0.517052981777317,0.8471988296272168,
 0.6348876897201444,0.8652225673500991,0.629051206018353,0.7913908671266595,
 0.9337368689498657,0.7695874214614322,0.23590390534659678,0.0741776858725608,
 0.23924506519631938,0.048050137886610056,0.7677379490284559,
 0.056372922365040434,0.3255799119970282,0.28209484932572026,
 0.1649305879574856,0.46622679971927505,0.900444318244092,0.478781723335433,
 0.8331079649562552,0.3661280360947261,0.36867816578480883,
 0.18713588252318292,0.7812683556653327,0.4930168576905022,
 0.20480150855306378,0.47145479030383086,0.4439965848120203,
 0.6011047176953214,0.5190569866216556,0.09814055245122189,0.8047221851794669,
 0.6857713575837314,0.32205622287521185,0.020863631688650797,
 0.15836201486411428,0.8971475023410439,0.7987569524401701,
 0.09103750571683777,0.6569445100694382,0.031163384281447715,
 0.8245956945761276,0.9581593019359764,0.8122487510391622,0.2307391665233144,
 0.29742930021010006,0.6152287933321859,0.7787827095068407,0.7614953960625123,
 0.35007361261614767,0.33727021056849726,0.5962999345033632,
 0.0012793618961012276,0.216029745596947,0.8231197670790785,
 0.3820892433875368,0.29330638372164086,0.0026815096109572556,
 0.5097458270974964,0.6078587009347853,0.3613907213007419,0.5011073762176139,
 0.9322802920836738,0.3887465808993221,0.16523351010414966,0.1015227785306233,
 0.282482525039384,0.25011642917637966,0.6806362075997898,0.1126135862614317,
 0.3514779324522299,0.4925593463505592,0.4440977001031654,0.20117543757039913,
 0.47304416511661385,0.11026775041447823,0.2274021537305434,
 0.7923003956493675,0.11377849483881941,0.7692448600186822,0.5863591955661144,
 0.3341269955278813,0.16873053374608182,0.7830838654771632,0.8035118178152323,
 0.6612993860949158,0.12879796239943275,0.1163746222476838,0.7569490354605807,
 0.695210361038307,0.6543232613897468,0.8517197541468867,0.028878582070018255,
 0.5805793162008827,0.03192219520482564,0.7097784767841686,0.7350648526927268,
 0.19837779018141566,0.8934060944104913,0.25637470072771507,
 0.6146742212333184,0.7036792538046308,0.9801903014527176,0.1797160646315843,
 0.05443658543909724,0.47881970508615557,0.5940015437578213,
 0.15813494813249696,0.861736397763045,0.35559262588549856,0.7718021055800112,
 0.8545506394756692,0.2468101296437708,0.13719511602536127,
 0.33033283221584464,0.44398112841232773,0.6033186516625992,
 0.04829782807987493,0.6967340883066385,0.8038345226854688,0.2955692330602322,
 0.3609987808889741,0.5772327210090825,0.037336027159022844,
 0.7099336697936497,0.9647842365948569,0.48916016554102804,
 0.37570782688501214,0.9613983278028082,0.34687514653143725,
 0.35661311556176745,0.06695181668905503,0.9746381406514455,
 0.9370825343119252,0.5139617601214794,0.3709394851477297,0.13793683821387448,
 0.37316856111320795,0.9764592924620786,0.7105072750832527,
 0.48317096489092115,0.2930174064895912,0.29485635283826706,
 0.9507488050109584,0.19231495668343312,0.25329501985347125,
 0.4627104356866385,0.07878254176742217,0.9551755808855775,0.5941460874169702,
 0.3366717053528293,0.2974751875824224,0.6194682595515939,0.09785870549788334,
 0.7378714922295455,0.5896511000455964,0.4545492197229707,0.12613775036084607,
 0.1220660851683314,0.7916788792946647,0.7298929204841489,0.4078101949857884,
 0.9242732041910965,0.10668294697565006,0.9326167572798753,0.7734243158721026,
 0.8894424043946757,0.6763013397266016,0.43068787676621434,0.6174758916980719,
 0.8009616454245592,0.5735133119521068,0.1916238660027194,0.6218061435255814,
 0.260693071841383,0.49712948153384007,0.7907580971491419,0.09866734270263011,
 0.9548688153794518,0.8620724196073053,0.025327649637159833,
 0.5405917958559789,0.4634433585897789,0.7641833734625378,0.8156653415863064,
 0.13519728079010795,0.017548921276535512,0.38228927185174244,
 0.24214674434890282,0.8098074981467436,0.7430902455731224,0.7696711035614157,
 0.4399032413591666,0.07178599884305259,0.03871181807488089,
 0.08275385555221981,0.6999296544926508,0.9249735696003346,0.9568423026369683,
 0.8245328699655881,0.2650887493658006,0.05413756465352071,0.834064924886099,
 0.871005084722998,0.4078880674126738,0.8745378094588764,0.26944353414678124,
 0.3036195704788125,0.3362744897222688,0.5472847282891826,0.16737953711839582,
 0.8264790778118201,0.05469344358292416,0.5173858741232376,
 0.25587160784654683,0.1992376118472574,0.7310046350686654,0.5369467043056181,
 0.6097992038424924,0.982339363361232,0.6097894858938624,0.9624421805119043,
 0.16385049426659604,0.05433957960805946,0.5539311471468109,
 0.7879519971603461,0.3377188645725737,0.5321114209820749,0.15560775521294135,
 0.17506077644201867,0.18453922042497062,0.4562390937555205,
 0.6980132666463816,0.4820304200653861,0.7332769327340997,0.9281122634644475,
 0.4011306810560664,0.865218902296866,0.8424257333446179,0.5171032075747257,
 0.595012430211807,0.06364328633972194,0.8067475577657218,0.0799424857147103,
 0.7684336953345765,0.6505132303072028,0.062201418002956976,
 0.7803394259200639,0.8390975137836326,0.1553782881604837,0.9024435623686511,
 0.759676764460096,0.8935704289153777,0.8101872093374785,0.8098589068651417,
 0.42277600186264175,0.09178991009646165,0.20366583332298116,
 0.5120857887612915,0.7107792441562926,0.7903019147807147,0.4563205809518722,
 0.9396096143987559,0.1619457489200803,0.5521915702100862,0.4796296490066021,
 0.4534740467920946,0.33445217225112256,0.44518466489742137,
 0.4516398895700092,0.1631524901978113,0.25177782546251526,0.8184576745667131,
 0.5697953685635854,0.7118088043723148,0.48293793927909034,0.898207622276984,
 0.15292835426138096,0.5950408925790854,0.5621132708159702,
 0.37447195060576577,0.22824962970097362,0.7963688995517235,
 0.6927856883704252,0.28428491663872535,0.12050349877403577,
 0.8855755406623876,0.8850001305018768,0.020579769727451014,
 0.5774083058609828,0.0917895408136854,0.6486887312896124,0.7772087900219515,
 0.1818417831061303,0.09644315829019157,0.8582868775510156,0.8482733236824958,
 0.7913367222286343,0.8326625462549763,0.8938977087291023,0.17493251865115622,
 0.3304123705079055,0.1782249042689381,0.44418397550964395,0.9977674962840465,
 0.6948555890070661,0.04633626825142845,0.6822735475258043,0.1199478296662645,
 0.7280357693031495,0.42116913402582845,0.8053720711992707,
 0.19078102821881093,0.48404579348045185,0.8530057655779745,
 0.9821301893912725,0.08291260133595524,0.4494021003835681,0.4503080303973571,
 0.026058371184043416,0.520458931068347,0.27730426334437286,
 0.9625594302746797,0.29098792837680865,0.00857739276816638,
 0.6189417083422335,0.16165772125676148,0.933239585925867,0.7800205246625587,
 0.40145942243724864,0.1868036418827269,0.5316069659218432,0.2188236596383597,
 0.4874520932939055,0.5765792112082638,0.5877538068137993,0.7560918997064174,
 0.7738635367625242,0.5845734897544625,0.7144430755870825,0.79901311725506,
 0.1780660377668417,0.9629435898649388,0.5451416837702602,0.41288890650202026,
 0.17903983742349272,0.9082191898630443,0.5097307242068138,
 0.13482245811692617,0.8849077824266424,0.6068861946369489,0.9190052118962584,
 0.4752937724211064,0.9086350900322321,0.09989889143205966,0.7920328583556708,
 0.7917321769465037,0.990623073668313,0.941313069048447,0.4262985402468432,
 0.17621163704888276,0.6531861426017779,0.2110175171930262,0.7138696791524951,
 0.9350751211320374,0.04726009444699897,0.8629824358564758,0.9978403271662958,
 0.36846934017562705,0.8782891284579929,0.9846669367565323,
 0.009404026886832395,0.5366412698158027,0.8145714789201188,
 0.9519672442004151,0.2858049995217118,0.393976283793239,0.7960132220240668,
 0.6928090695189,0.8895064544213387,0.34289950566955973,0.9935465768661783,
 0.4445468901695059,0.47347193396173215,0.05533636437351175,
 0.6981878998863106,0.9221271760872785,0.5084203805215722,0.1709424011908357,
 0.32932016986132306,0.41259459732094783,0.7225694874175854,
 0.8156045692641277,0.44719734646127673,0.7320272963361109,0.7093285084367911,
 0.35265890929545995,0.12080197323118036,0.08859596699786576,
 0.04942913028198659,0.8614374170850116,0.2937540089829104,0.896295687070738,
 0.27335412843918816,0.3468416315976379,0.4402520010987536,
 0.41989402567076317,0.6733964103764605,0.7971906337947973,0.731805970064944,
 0.17189232457517456,0.4707031563357543,0.05731335585110453,
 0.24717903607722624,0.666456288806322,0.02715973227592383,
 0.33380032095821655,0.37888897430643487,0.7769699204694197,
 0.4011179509938946,0.055128832483616996,0.40387307714872145,
 0.4688844656590473,0.45420115480847667,0.9459504725981096,0.2329877760755722,
 0.327290044871134,0.9155925554257297,0.4842428610250703,0.8164273699315253,
 0.11986205569302422,0.4413113806263943,0.8850655139843586,
 0.28839436051532186,0.6827369843692407,0.6895616956659516,0.5362915060733051,
 0.8135304887095509,0.5534941855258408,0.564706270286552,0.07017952568481789,
 0.35246784796197483,0.7504205676304931,0.9473899355407271,0.4042327100607753,
 0.6375991301435108,0.022232660046595965,0.3397942750333478,
 0.9082559168379207,0.8801337890050052,0.10604293509203466,0.8863234649035824,
 0.8043314022889456,0.8035045178173835,0.8082436884730823,
 0.006089026556428845,0.14887767828454757,0.65268086571326,
 0.025077633347000083,0.7277810410054744,0.8555903696343701,
 0.7323585123776226,0.26326307051605613,0.895324669782252,0.5544067023226351,
 0.4401446566145787,0.6014755384516207,0.5656077001971329,0.8265331160060592,
 0.9330554924293555,0.6215975835883846,0.337318010097581,0.2471397384625742,
 0.9945097326566933,0.5250156651360872,0.3035104093298474,0.3551254229912959,
 0.7576406524702969,0.9346214488293823,0.96546432066735,0.11768686875815193,
 0.2681342665171892,0.07390678396001882,0.2746618527751957,
 0.14098972307756197,0.34855256555361236,0.39255042429142406,
 0.37851061091302185,0.827194078841031,0.33239809265586717,
 0.44398595281304865,0.952148707165182,0.32680993648306045,0.4710302363061163,
 0.9225949911536646,0.9774016750109198,0.542470888277436,0.3671009922428877,
 0.6529665015630154,0.07240461511096818,0.6641001801871911,
 0.09796052045774428,0.15518598797548,0.21745674517923852,0.9518281758077023,
 0.6596525814669464,0.7274181293197494,0.45404837222455696,0.5675209888777804,
 0.9957959015005877,0.26418107137014823,0.4504076869532234,0.2232000866715943,
 0.8129816702160642,0.4518687582118799,0.39484132295564434,0.6144195312194158,
 0.7422921495875459,0.7771424972684745,0.0030766670675718144,
 0.07080671624592705,0.023774478810862698,0.9286155856247728,
 0.9667858333505239,0.12017362297028822,0.9713359320915708,0.3959132564434169,
 0.4638600055769715,0.2198548728873665,0.9221290269699308,0.0807570482872978,
 0.7820543723920669,0.1407455941921092,0.6739073678181715,0.07878131615068384,
 0.6365831763638558,0.44706554176270086,0.7499648507763585,0.4299679617572081,
 0.5524154319770899,0.941940142965029,0.8118207569521247,0.20414429369644993,
 0.6361522663471434,0.08942647764752998,0.8919811500247949,0.8340640682743109,
 0.5406290734773707,0.07834159372216709,0.30371823242262164,
 0.3944573308562507,0.11341333152198829,0.6499549720287691,
 0.033156105190301544,0.6140898423547446,0.12596625988606613,
 0.0564787317947979,0.01563312953520002,0.36235777510041933,
 0.05710875868879817,0.5943949266694483,0.44506585866679116,
 0.23527151061461127,0.7425146923759025,0.1674464334851048,
 0.05885145054973684,0.3415879512829082,0.48262859801967284,
 0.06909091354249897,0.9851567959604296,0.13857761948969427,
 0.4130319342911054,0.4121406822992455,0.26704470459814866,0.6150385623658571,
 0.4974755356163959,0.7240007326779653,0.13534902597084897,
 0.27429718345938714,0.3095163599680586,0.581635785579754,0.19009142866397133,
 0.030396574568030688,0.5690225226273357,0.2270618891709537,
 0.7646351905668026,0.5185629500261861,0.542659329399879,0.5583951945206008,
 0.2990099573377234,0.18583456099628992,0.36276185022587726,
 0.45452673277995437,0.8774152676502819,0.27404111107681184,
 0.9900417816009556,0.9001001127573913,0.4106723886008853,0.8010666749661579,
 0.9762974886635104,0.8664091866264425,0.6743740809779938,0.2527561283818893,
 0.4192692139094525,0.1615921743664973,0.13359847838358063,
 0.09135948486132806,0.4448191717371607,0.9258020839625568,0.7314605353373103,
 0.394020368207471,0.5844886929102657,0.4134486499276566,0.7783616775472436,
 0.39747460405139434,0.6633586276809365,0.9876040877161708,
 0.24593554269929518,0.5906234256127285,0.11516659419347897,
 0.16915090368474806,0.5563102120681427,0.2976530754373833,0.7464723181967227,
 0.5434149504801572,0.7114936268722487,0.7145992968200537,0.11340539243564507,
 0.20259996734937014,0.6516633138370156,0.4292088824150265,0.4995912469531917,
 0.451503304465412,0.8327835554156973,0.7035012811156227,0.7091265819294474,
 0.04885113415742315,0.9046727938516684,0.7198417175814806,0.1738600772364287,
 0.18404752934332946,0.3988726582505422,0.8981397198726201,
 0.46383244026218884,0.7174152865677084,0.5815823414868005,0.6944550431350295,
 0.6071367763055371,0.8936002355466467,0.5891075712976791,0.6436900679292602,
 0.8581164633686492,0.029071053936013502,0.10650413301549944,
 0.7978662631928537,0.6849712871971938,0.9608493003922804,0.3419162002282614,
 0.49946941185744786,0.4743140366998364,0.9420844676061237,0.9271021701360288,
 0.37141601194262885,0.04504111312546166,0.9938153028713573,
 0.2235832121212873,0.41144162379087235,0.73296618280403,0.28069445816260963,
 0.6975030851407552,0.2302282008091343,0.11415971960592686,0.4004919720399448,
 0.5814179627997401,0.21720057717470775,0.06612889027545577,
 0.1951185849442708,0.46288778725720625,0.9738611746808469,0.5883746049252631,
 0.4740575376265693,0.8500338033599265,0.7079843583277399,0.32841817576020094,
 0.9383733383411922,0.05398567761149464,0.5442806466237584,0.2679551229022181,
 0.3826571757909649,0.4662473952038768,0.6899920985104099,0.7625777869049423,
 0.4980883469054158,0.6326634428464831,0.807924304756835,0.532119463843388,
 0.7733509095216826,0.6933885884776083,0.2756538213528198,0.3160727499939867,
 0.13899830854148398,0.8506098984067051,0.8738502180171267,0.262955189008227,
 0.31441653389737223,0.0884788896063442,0.41763949500261155,
 0.9354708602281607,0.26495535807926607,0.6900414105479444,
 0.19452383299712783,0.7136678732400862,0.036864252542500586,
 0.1485643052065161,0.6990762910330628,0.28928081746149603,0.7859773847345358,
 0.8040777117595336,0.7367186879473534,0.8339327719465154,0.9932241646013262,
 0.8539848712685463,0.8453836320792711,0.6558585888741542,0.6431955048542373,
 0.3670717141820943,0.4787268592233215,0.29566536859731785,0.7561223506891722,
 0.041411655624685295,0.5228825133321306,0.0060629003512909785,
 0.9919259566482455,0.3746664156796212,0.5647007133224224,0.895264011187939,
 0.6318020116804718,0.2815592319551635,0.059907468961634125,
 0.7816072901869906,0.11243906880761645,0.9235573595371449,0.3914028651733341,
 0.9905138803542025,0.2671808639747073,0.2824237102375148,0.6926569583793177,
 0.857752447564644,0.9092882024399447,0.4574106398170189,0.9790970349790841,
 0.9864367409060018,0.8434888112025379,0.8754893575732163,0.35482534402499255,
 0.13552516289460403,0.08820983716986774,0.0737099865149845,
 0.03990790611780082,0.8513962968084238,0.7140461937056041,
 0.20921445914519787,0.860773681887587,0.4971207271247,0.0012231103419477751,
 0.40374224794516156,0.37462749795854466,0.32001565409820054,
 0.7206250180712395,0.2644369144454949,0.8777543513178012,0.380221021442293,
 0.1992099054205947,0.5290037283658058,0.147651924055055,0.2551287620109427,
 0.9600375945326852,0.4189030033768615,0.5952363490707914,0.4501670203648298,
 0.5505292701118331,0.5782461658457418,0.9942270455299028,0.6433707649408325,
 0.06418887935297668,0.8080641173648758,0.6773672014615642,0.6719455720072911,
 0.4225632663663108,0.9607848192807859,0.5747613004905743,0.6931519534889724,
 0.3069697487642078,0.43637411871002596,0.16463333023142834,0.834540676527382,
 0.18608350934380669,0.9625273000933325,0.6758406745830814,0.1951004384075855,
 0.6700738536275574,0.11992649305953673,0.1599237255285798,
 0.20864703184835076,0.27302814436884615,0.8852920979949828,
 0.5677237432179453,0.2249873688720223,0.7529204421367847,0.6997304503667205,
 0.3906417204122018,0.09407616174027444,0.9108215891244238,0.8383550051937718,
 0.3088105091582709,0.6342808074889955,0.4985007008453072,0.2158070522565534,
 0.12017841544575791,0.3301685917642432,0.023119105567886655,
 0.6945551979906459,0.2960442870253328,0.4144758325896666,0.4681356014679616,
 0.5338506980200791,0.19075682384887627,0.5303140340259873,0.9611032274997806,
 0.5798037518806273,0.3615195382342903,0.12564647735766832,0.9065893848962404,
 0.5886957885179431,0.3022426976912721,0.6724900391345825,0.7447575205232035,
 0.5953199686324072,0.8612704942925021,0.0064951234139525305,
 0.4751551230371971,0.6357723809330426,0.35447004141969474,0.3156284164960256,
 0.6253674603064907,0.0577621933161343,0.26535343498435593,
 0.25896779867505526,0.6435624139361332,0.055600836360482186,
 0.8224071002782658,0.5356134495596856,0.15973648186335565,0.5138438564305121,
 0.49415731055166623,0.23571473895438788,0.6596256234232776,
 0.6204671870298901,0.8070917404400482,0.5603916260431869,0.21056031972749034,
 0.7267548697886662,0.8487541089040668,0.5290513434716344,0.6950520325911792,
 0.3502023667677675,0.31392304635797574,0.6483212869167277,0.2203283162758699,
 0.5337606602489403,0.37612375974762113,0.00741429659514381,
 0.30282582420593174,0.5406607953045192,0.5503937347792381,0.5027903703061829,
 0.8381812536656785,0.21049810434727634,0.898870455888405,0.6787231273356209,
 0.37629971499699144,0.3890732655621778,0.8915725709194527,0.4755552694569537,
 0.6179509132919522,0.3039409987016829,0.9103006129092678,0.41230455418965317,
 0.4505895489971651,0.1354963050021708,0.8048066136911395,0.441984514626117,
 0.9774245737123146,0.7971158760570498,0.027304863581628647,
 0.12230288439479198,0.4444147357736028,0.4139565536226325,0.8929441225464771,
 0.683891153702243,0.48186373968939566,0.1905723180605875,0.7146263736420101,
 0.8465140237590616,0.2743488543172239,0.12580520104947235,0.6635312530809229,
 0.5085839119647618,0.1873641516206126,0.0394346948201062,0.05674471211623755,
 0.04556995847196499,0.7054265382966406,0.3430764751826403,0.6101571581397323,
 0.10323705254678428,0.5023087988774192,0.5848425429494977,0.2384180120999122,
 0.3607489206789687,0.9043643388948442,0.136685529146379,0.9430917771649205,
 0.5112697928246593,0.04960765328464434,0.761690928844724,0.06379434587152666,
 0.786097618176893,0.9181822295867879,0.0030884287061923033,
 0.38108174842940457,0.6598111471150464,0.655603091733462,0.9394416919297023,
 0.4716635240711322,0.5253236846096223,0.5325507977352308,0.5505801211059228,
 0.25559741854439233,0.8947748721678479,0.48396450682815173,
 0.9062167343360574,0.6220276146882594,0.9444250860381646,0.20583661085038552,
 0.5582389047249406,0.6633556659179348,0.2625816313086826,0.7540339369058473,
 0.042551576684990655,0.5283090353268711,0.6904193465492466,
 0.19519209671874438,0.8855821759934515,0.595502043095051,0.9087738920412792,
 0.7829619748390799,0.5330000818912548,0.3605454372587198,0.08221239831043992,
 0.9927386729003025,0.9919657644975363,0.617167748421176,0.009448500944701066,
 0.331209465463549,0.1104656565366049,0.004250218381994353,0.3387921937447449,
 0.13337914835123954,0.9523732267703692,0.3860239006085163,0.9887120706564916,
 0.4729575077771442,0.030478674087771385,0.7809668040119407,
 0.1706383438344996,0.4158013670071874,0.8291797219829125,0.7753206612353829,
 0.5284450414059336,0.1375463350903593,0.01155423949331369,0.5378954929813682,
 0.9901952820638105,0.21255889917376014,0.32370052125809,0.16476890155596946,
 0.9872867601909975,0.7550279529768308,0.4452765773477577,0.9182763074199548,
 0.274124304412815,0.1709100572332517,0.18128188067504936,0.7325370132934129,
 0.2709022535675416,0.8495773805298404,0.7999531958706645,0.9244667682081325,
 0.6839800110011989,0.6393280618537457,0.7593698049195294,0.8979460351078872,
 0.04660359631018596,0.616909217803058,0.17191092718275536,0.9106546358354197,
 0.5327633309039209,0.6956254894195156,0.5218093751410605,0.526132381280882,
 0.955172983004524,0.7830306949337462,0.25518876302486104,0.7684977694937433,
 0.6351811986290702,0.6867853001668887,0.11641612375439914,0.2667413625018288,
 0.9526000054889322,0.5433529535421877,0.4537739581150362,0.20407842191779635,
 0.8424746388656955,0.1429335283294786,0.2676525988061551,0.7747313801701297,
 0.679955055009981,0.7225160285369915,0.17535713570379663,0.9538411672334846,
 0.3652392690175785,0.7618045176057514,0.6910185848381024,0.08763197962027691,
 0.4251002955028238,0.5880171252934503,0.2434359290111262,0.745011667025455,
 0.09586166389956308,0.0655645859304892,0.8320323875842391,
 0.04413034710899477,0.4285426273980497,0.5275217775750659,0.9905676978221442,
 0.6400359265443,0.8867795493862501,0.012429523031095457,0.10017031365467033,
 0.9513820223324103,0.8461465177803078,0.8718049511413208,0.2861905270484255,
 0.6920211052879719,0.6565934706723945,0.9108700825569579,0.8970767270240901,
 0.6251037458709114,0.37662184995461345,0.929553817787674,0.16471094348862347,
 0.5932156879623255,0.9874328064205578,0.13519771579336737,
 0.40673766105917464,0.6709736055573511,0.5914842613125426,0.2338618901532281,
 0.6450614981247346,0.7784939170176669,0.9621228920586622,0.3483789330304101,
 0.6634576649336222,0.45550286749228464,0.16696010682475482,
 0.06995344639408974,0.9930054332615479,0.9164717603293034,0.7240548746464828,
 0.025693509901066225,0.10460496986512624,0.5257888315425421,
 0.020083878569120084,0.5503689617279781,0.8293978610561471,
 0.5092771779553577,0.6918525050872577,0.725310088272856,0.7133618681936272,
 0.11981506937224051,3.506221111322372e-4,0.7942864289505199,
 0.3388688529303424,0.45253990669502187,0.7732388060947846,0.5352805426343942,
 0.02742298473966498,0.968721547010341,0.5723118169141138,0.9354891461625587,
 0.6946754285775473,0.25108257525827415,0.7167643029359163,0.7672761385984741,
 0.1085019476775394,0.7516606993540652,0.9232238829467857,0.5715303888522139,
 0.09646031157004054,0.3301214706783283,0.3132724822157913,
 0.08208932140287983,0.9639517599310012,0.8179464319311142,0.6303368861397353,
 0.038488849671341585,0.38558125126845244,0.04549812528160724,
 0.6457810615537357,0.8269587253206162,0.06448114870733557,
 0.029019988840098376,0.8238225675444122,0.028509830472225772,
 0.9360981768393078,0.6289029035870062,0.5552092207861409,0.8085080202839311,
 0.30828897156819624,0.8998178222199915,0.6185826703891646,0.9641965426689342,
 0.2251406561787288,0.9765936727540943,0.014379963177345467,
 0.3845242513394338,0.48915359424667515,0.9899711875787863,0.3543610751600841,
 0.28912839917759503,0.16328489870355534,0.4558032279385976,
 0.10924107338166877,0.7498537078674421,0.16480390776388842,
 0.3665771314223606,0.3115547318286144,0.9764059959129988,0.4261952315504074,
 0.2009811074528095,0.39528104756220095,0.6010951827195711,0.9131341620663689,
 0.908444872972073,0.7066576839783036,0.4851524763353542,0.6773444404616997,
 0.8567096351518823,0.9798627105917298,0.31331487870571983,0.4798671033063253,
 0.9916824373486527,0.9631208063809339,0.533460640285491,0.40456416982671595,
 0.4851179377595417,0.9783101062650418,0.34897412590321375,0.8864363909312057,
 0.6157164163808776,0.47371373628661995,0.5415792407015996,0.808221251313817,
 0.021350954993044735,0.5128530526787206,0.12385155377412138,
 0.9137475976837246,0.24273602914571224,0.22822396741256135,
 0.8972732122489736,0.9517536687838065,0.17232385349057955,0.5288229764551259,
 0.8663366896064263,0.1956956797876962,0.5809099415611031,0.7830937310067566,
 0.4039297456261446,0.3974194138243865,0.2729945116832767,0.5959455364797054,
 0.8911348273085995,0.04233662443930242,0.23878567460062217,0.755786681070881,
 0.317051209229973,0.08495499298821985,0.9199989435899207,0.8435528138536383,
 0.9116146036551412,0.6000739606161567,0.8208700091330506,0.3824697172615511,
 0.7800778462830305,0.5762225726403847,0.7888416879366621,0.9225574955802988,
 0.523056709552275,0.6919239648199,0.3303097457887586,0.36339768610632384,
 0.9744763621932879,0.8528236152067347,0.048823745652532446,
 0.35651217655580825,0.3556722251041463,0.8664691360302428,
 0.13838102444952594,0.21145214880691177,0.7815032648067054,0.998030947127554,
 0.029248480084958173,0.2237677557209209,0.5922832134760889,
 0.5556837000360142,0.6478191744209567,0.6910625136947112,0.5108492896681904,
 0.17044532466664586,0.908403961160066,0.3578181878609672,0.5577785588252695,
 0.7956747788642793,0.10680968452544448,0.31569688558147213,
 0.9288607150695074,0.3537171655757334,0.1372518810139327,0.5686312661195573,
 0.4004203348519325,0.15750994888797998,0.9667862874418744,0.5842337046700238,
 0.10272442437102125,0.44704547580601917,0.6933298506822294,
 0.3390652521239029,0.15150501571238173,0.7506131201477405,0.7914704236408654,
 0.9712136197869656,0.29114248028750256,0.9126102025608157,0.9655432069944245,
 0.08684359066964731,0.6015862841541364,0.5295162187193744,
 0.009336113678037217,0.2709815389587792,0.7421697657888877,
 0.1587593195877407,0.40942338197306094,0.35319810587721434,
 0.6060394353498129,0.6680715177638523,0.32715941399819315,0.4356503016284219,
 0.8796777122908619,0.47669097707790853,0.8120266601788739,
 0.31877770491950264,0.12104998019787738,0.8876165479590294,
 0.6482133113633116,0.5500041139656536,0.5879508966731428,0.084338282217716,
 0.3454265538746787,0.947980402183469,0.9811728295830543,0.6504072136643799,
 0.2268484614198426,0.7353322483702766,0.5955194923838618,0.3957846824911244,
 0.18689326775614212,0.5230031616521481,0.3184657823114142,0.7809748562970317,
 0.20903167928549604,0.7895030412510461,0.18680600240026835,0.91143054706672,
 0.5760861345220114,0.6187787613017008,0.46529348413447424,0.5032361501391087,
 0.013518708881787522,0.28783270150706564,0.18555915171955806,
 0.0476432043424877,0.4380006975621661,0.8075831131148535,0.34049323721879043,
 0.00132141948392539,0.17582240820547246,0.7112626696132799,
 0.1314458070560569,0.7792775868777154,0.7672766796828056,0.29137416838181585,
 0.9120506833397062,0.4535857215266925,0.29623458088661814,0.3596308995923274,
 0.47895884879830875,0.9717107093453112,0.9676495912762266,
 0.19210199916891435,0.19182257369081435,0.6714872203277031,
 0.07516491414918369,0.37980003891646963,0.24563823070756063,
 0.7700494913603915,0.48753998669236454,0.7691604012975043,0.2107188619601632,
 0.22676993978608628,0.9673430062157744,0.04341602651082155,
 0.5180772357220143,0.4322254044272804,0.20176130478369803,0.9852754528746592,
 0.7291473726117236,0.1165909061597511,0.5696062419517731,0.9253804563920387,
 0.057694374772268464,0.8227820224144635,0.1654295814576513,
 0.21792430735441648,0.7929984986234068,0.8552873765738063,
 0.19409318365251205,0.47549956232271706,0.7801609846390571,
 0.33870204575487484,0.7839190533519053,0.6896659908651119,0.8852250044025104,
 0.618816744497222,0.4788942030970915,0.27579607383806704,0.5437963654023295,
 0.3530622212290089,0.9934783859109024,0.23347978638179134,0.1763019741359153,
 0.9304883184540442,0.6075256378750409,0.1385839473726186,0.5943595908693768,
 0.32658334738218686,0.3854237983909843,0.22564828755533406,
 0.7691377208072315,0.4021884438147315,0.6023829538658272,0.32296005672948114,
 0.762378600288328,0.6134264352030058,0.6442016803526334,0.7672926284362651,
 0.4834874699124927,0.9506456851777793,0.4339019452988325,0.3556416240118526,
 0.3145107755714759,0.01992990284663898,0.23186849981629376,
 0.1972694502735317,0.6222414016725518,0.2598859539456718,0.23398283640959006,
 0.04206547148398965,0.046895852835352914,0.8811546302353214,
 0.3587469057917818,0.6666642503576996,0.5000757447792034,0.3195502443472761,
 0.06129286299739933,0.46967552859970263,0.6762980268341224,
 0.23404843061342584,0.8504785907912908,0.9998084891400705,0.8701210373089463,
 0.2751250078289411,0.6673860187149532,0.9682960133825025,0.4006735538632452,
 0.8797924911582563,0.9569290580051653,0.9550177024829248,0.6621688531708472,
 0.22103447355009642,0.5339147676102702,0.0762787270150116,0.860169265868211,
 0.08054457873149501,0.44007532525870163,0.8328541383101561,
 0.8507759808365244,0.7054801194874096,0.9604890089693866,0.17346378781265848,
 0.60684852427505,0.4079489708002956,0.650705500599969,0.8151781823187786,
 0.8501602418292998,0.3960503539589595,0.008284739221893767,
 0.2600696397610106,0.41504422943684527,0.8169647836743212,0.9669509475182596,
 0.9828563747798853,0.4154625743727658,0.966553086885692,0.2169890191322309,
 0.8135727872352889,0.801302941213824,0.32815439442342986,0.43244862198253275,
 0.7012452423861789,0.28748261374907613,0.25349606222102905,
 0.8265775844931145,0.051777999371880545,0.4659904958728389,0.938356163890817,
 0.9456619931865015,0.6772293288108243,0.9878050057708101,0.5680337145193822,
 0.38391164204073347,0.6533170896835564,0.7452993283088414,0.31155729156086,
 0.4067835411342229,0.287340050236943,0.09463213828074646,
 0.025831181794172986,0.7728995800612435,0.5091539197361111,
 0.05071043301210898,0.9507561602236037,0.6458060955204299,
 0.41962775296338906,0.2759271179227628,0.13292876016055322,
 0.32904218324982626,0.8841643008466712,0.9966243019189145,0.2563246542199683,
 0.5666276263254483,0.5643228142374701,0.8024969424883933,0.7373161700533273,
 0.45856136144581117,0.463887621797597,0.6318376557632978,0.7883188009466866,
 0.6229757453298899,0.7191575535747851,0.10410513480651784,0.8141142613952281,
 0.7305045508116395,0.9275304642164477,0.13833153542571697,0.3559868643007761,
 0.4417265373030457,0.8878511859850041,0.6084563066713145,0.27989636946704166,
 0.027169608409194912,0.3867605472648843,0.28141924063631363,
 0.9849146973102803,0.5173543032841458,0.36805119369372563,0.701468764671479,
 0.5399322028355508,0.8833572361983337,0.22299655578383892,0.7706209413441916,
 0.8935873046816992,0.8149146826242761,0.48914885246125483,0.245832809605963,
 0.5090031414968956,0.648228573134612,0.6662983675145817,0.366737864627384,
 0.9590038847886021,0.125174263404789,0.34162799261667653,0.5463527718122448,
 0.2043420887545394,0.5074716334425202,0.5739719788120722,0.30541009169974775,
 0.9072597975232618,0.30741670682785327,0.10650458826448328,0.925065092336601,
 0.22506606067081547,0.016855042778511464,0.3066428292698884,
 0.012152226967825142,0.317246083331418,0.6673493677856306,0.2483854827225671,
 0.7400152631666119,0.837871817814224,0.7051876938162147,0.7216391558719059,
 0.7514332994603001,0.24800311736109215,0.08766486430790454,
 0.7450801550110632,0.9119775352057099,0.19474565445158376,0.7407290601826537,
 0.8398681475179451,0.3896530483813442,0.9565679377948393,0.2639829841590594,
 0.01337128211509242,0.3007556175370607,0.31651616347965583,
 0.24050667054317798,0.9938546432533699,0.5495652253886744,0.7910153988271725,
 0.4978366835884621,0.5200913940773116,0.36143413946938563,
 0.23683154172505771,0.7526263450387524,0.01468523350661155,
 0.4206260568876484,0.7654182787303249,0.34131992130056854,0.8512574410397759,
 0.6252427759187373,0.11014421443534694,0.3670507919170177,0.3280410617018442,
 0.619951921163798,0.46704702254136,0.9755701510240595,0.056816487835368634,
 0.1333656999727123,0.0826049169980928,0.44015940324872127,0.7282710435726764,
 0.2882386824448373,0.6700187434851484,0.25650999687368403,
 0.14421007310940048,0.6580507157867591,0.6668001266025314,
 0.06808473865394205,0.3089718128091735,0.0843353202065995,0.1886226342595566,
 0.44472706355596847,0.15306506305071255,0.8156182234696652,
 0.39515857717114133,0.6343799263161028,0.8186276079755042,0.2546132695066503,
 0.3385611201141062,0.38460664604367334,0.4932249924833799,0.5382226726720967,
 0.7883655772507376,0.9566196425131513,0.38376898395084413,0.4396943094936562,
 0.6740736733478534,0.2582997407606604,0.5756019728813634,0.24601865276028145,
 0.6743237381911261,0.6912603735365589,0.09043214110814357,0.4500050040762281,
 0.7003093544965633,0.8265396060239162,0.7788621736231544,0.19009672829426894,
 0.7271138885919296,0.9670968103738562,0.03467991077694077,0.8583647556016805,
 0.05236717263356372,0.1210746108572518,0.14694453129698304,
 0.5532453778341111,0.09316093003291692,0.6057854018426705,0.9414412885264434,
 0.5250673237417187,0.22985403526698167,0.4152108626020161,0.2564446882478453,
 0.6639654762811079,0.8195851609065105,0.07915398077549185,
 0.28837385552014283,0.5838069744844758,0.23479992232140035,
 0.8651358922612951,0.7493527826991113,0.6187673421644221,0.19022054754464535,
 0.5123026465138886,0.6105063562681701,0.6308998789688925,0.9327776538664047,
 0.15923827942480306,0.7731762835920986,0.5314288103588137,0.7650651843401275,
 0.9643523309459298,0.3813317106021983,0.24465872541597677,
 0.12835404612948365,0.4084385953767038,0.5045323941820599,
 0.20913052275962385,0.7246366317862786,0.0694287459204509,
 0.10296376683099351,0.03718442600285454,0.510411104210676,0.6932828923003148,
 0.27581686359361823,0.11071018517339681,0.3584994063996739,
 0.7010381323068551,0.7992076616430908,0.1930178251451028,0.48507442459609607,
 0.4244408031090632,0.8201869992439632,0.7758351974948745,0.23889443265571297,
 0.1601214785651336,0.24478305608437712,0.8297378806368993,0.7768910640226268,
 0.7958748873970786,0.6846402279158397,0.7050077015545373,0.3706742664658069,
 0.4901350351995035,0.6291659600171859,0.03106327656967811,0.6283974765047946,
 0.25280017102386587,0.8734144339770632,0.26770818481510594,
 0.9527959302587501,0.3437966716980273,0.10669591897288566,
 0.41993320315399896,0.31149390714189373,0.34666275967709825,
 0.8149195859606017,0.9290547194411405,0.3033493902286001,0.18285292596680136,
 0.9309963419897721,0.4146409756365923,0.8381460535958744,0.9473333844307037,
 0.17380725098323102,0.3553263294350967,0.7730138251262596,0.7101243420124078,
 0.09561094088839972,0.12349517692107381,0.47733290402726647,
 0.21022999186866076,0.9637074198100561,0.34014091751761555,
 0.9647146990911871,0.4345990148220298,0.40172704808668147,
 0.24879952803555871,0.4049937197891684,0.03580842922462546,
 0.7047894281423659,0.28343429242573126,0.7118079413633364,0.2882162133170646,
 0.594604110968209,0.3662933418126093,0.34854433904665094,
 0.048490299147105675,0.1906393154504099,0.4295223742788964,
 0.8649222052138692,0.6090701771955266,0.6036083375969452,0.05922388400396783,
 0.5507639510083646,0.376311112218748,0.43177731739390357,0.46209447978953755,
 0.3839240262020249,0.4555255170492565,0.4252799131977618,0.14616397335339615,
 0.4636669350744942,0.6661045154648227,0.3716685640187549,0.878006448885234,
 0.5824630288009701,0.8354157966280568,0.025228951507734187,0.922407094566959,
 0.46471895636695704,0.8004240676816572,0.2293687466713924,
 0.33257280817904555,0.48760975274609764,0.8606825865023646,
 0.7469382904088857,0.032673243139817676,0.18560446851821466,
 0.5349924383225927,0.12969975393250432,0.3079468541872493,0.95364255084467,
 0.05869436085845059,0.9649992264686137,0.9188068467590551,0.6859140460920402,
 0.8957652865954765,0.030922234675007765,0.9164471222641433,
 0.3997308500325808,0.28293304078034365,0.40358542905564854,
 0.9969417386728923,0.7620551393109775,0.9050771466917299,
 0.002102246977338851,0.7217201618637522,0.3192739565584014,
 0.8236114084307689,0.5166449425607789,0.4707424861806163,0.7855882767646029,
 0.2030039477377914,0.0828139196668981,0.8954567484379901,0.4059982881065308,
 0.6207765275641383,0.44387334661721123,0.08499696810293822,
 0.45016970354223496,0.894845512594568,0.8069163622583568,0.7674601846355765,
 0.7682660298289086,0.9279865854247269,0.7155127972282511,0.74125072688556,
 0.6145072097839257,0.9689831769417669,0.5575199045449426,0.11176703451478542,
 0.33960700030676705,0.3244047152972518,0.44895790630965493,
 0.5966230719664456,0.08892675733779498,0.4345251530785257,0.0714517291044997,
 0.7969151108812302,0.39937180395329375,0.07252524314291997,
 0.8296148661586724,0.8909696148346115,0.9180110496593743,0.7660405853012998,
 0.5085113639468485,0.15964976884542137,0.894458623680185,0.8623559980516832,
 0.4220818791768547,0.0918273798669027,0.13133958426166714,
 0.47064023136178346,0.6212993372353949,0.047536287339326244,
 0.6907031596459174,0.5240559137835881,0.7635366589290153,0.5405471450948173,
 0.9704812678549999,0.4027161385436461,0.9483926913539371,0.3522301293258928,
 0.9876978219509571,0.9862646631127084,0.950418901727884,0.05749510007354042,
 0.23783554528521156,0.3109628163058713,0.6645852200058868,0.5372281933804924,
 0.57763922227687,0.1240975746505173,0.32279815663485467,0.47113687870860266,
 0.34666448095422264,0.7625162638312513,0.4848490603186817,0.6344494650387074,
 0.4632607910051768,0.19245783302226271,0.18819862827602118,
 0.8273318819805188,0.4105023967460525,0.9796195747474689,0.3751209584393682,
 0.8982930232349545,0.9050297778812904,0.45064789424223295,
 0.21093740121487092,0.7497735193346108,0.35212295740629096,
 0.4421658887520622,0.19758509490027876,0.4812269826287334,
 0.19394312696109128,0.19076750996861302,0.28366404981211324,
 0.338680341946609,0.1970150828101268,0.6306683063628972,0.3801569649679398,
 0.5199571798972246,0.3227414886132519,0.2484112669315407,0.4904515113691752,
 0.44218271231172146,0.08321727627658104,0.08440725122571835,
 0.18213282296606315,0.8904947023228574,0.8132821248762291,0.789079349359771,
 0.7508328519689988,0.40085666765378336,0.3080687152349567,0.7285745497607334,
 0.9977319082173364,0.22682959707378614,0.9371162520454843,0.565679762313344,
 0.6161786886789576,0.5939419190449637,0.5805932257570005,0.5244476833635415,
 0.5626883373108438,0.2584300017114237,0.44903882765506054,0.2742670114445761,
 0.5862896198911509,0.9369880046771467,0.3075342967779734,0.6997630617684109,
 0.010172929101335937,0.9256311773874923,0.07565296624489237,
 0.7392655389069617,0.9579985034760771,0.024680116889528758,
 0.7966318963268256,0.28106491951555834,0.7557003522467506,0.3911085386512996,
 0.7773383867079705,0.6001071058099957,0.3044253532394934,0.527359646855139,
 0.5892790691010632,0.2589928341562632,0.7169598003304336,0.4878697069363441,
 0.22352634026458607,0.9826294389019099,0.7277043253476199,0.670401021133104,
 0.5385725974725472,0.48548256614552077,0.08077893403510172,
 0.7029298889162885,0.22413193090668082,0.007383407252398055,
 0.013466547592080658,0.7378298559208339,0.4969279078140578,
 0.6553094688553746,0.4077932075364733,0.45092056451939255,
 0.09730820163087961,0.8967713135284384,0.9280163668438944,0.8974542748161697,
 0.07835319748780845,0.6503033065662681,0.45629459970328534,
 0.6702595064676159,0.9207522460766238,0.14163614231510546,0.4271845259655699,
 0.262561824620811,0.69156070394868,0.4799448564472669,0.9264295923659905,
 0.8784957937560761,0.22221870228873508,4.777665369282147e-4,
 0.038625721752471365,0.44088446376052315,0.6497061744853778,
 0.07362478380054238,0.8978923475627572,0.05179414278468564,
 0.8989119464436088,0.05375067911114817,0.9651444362513877,
 0.18361100530416885,0.4206029384990082,0.701793242288897,0.9159729371391739,
 0.14694616273974648,0.7442754674870374,0.9812437405912489,0.9744242542130914,
 0.6030551287217814,0.14855115906374494,0.5268885430433153,0.358848539814829,
 0.6625391563567096,0.21185933535338675,0.1673326540730674,0.938853661444668,
 0.4486083288891709,0.11519122961954531,0.17896296511761156,
 0.2599410253322426,0.8068796211560076,0.3572011997404987,0.8671450539872932,
 0.32493353684935533,0.5023405649277135,0.18186634983907168,
 0.5938925858699777,0.03161951206693059,0.8046370035307138,
 0.41210167838575296,0.2392246311938918,0.1316354459497705,
 0.09059701338111481,0.4843396366764767,0.2647502569511373,0.5720744179511321,
 0.40475767225461,0.4212883824238679,0.4272321834683537,0.7041114541194178,
 0.685383326306437,0.13629864696554916,0.10403521945897998,0.1402830036952727,
 0.9787310193350971,0.2859088178248743,0.5558335855622771,0.8494299625779063,
 0.8957355953749404,0.4275951495255791,0.21626290396029046,
 0.15250509175432625,0.15180140055022373,0.3097255368631662,
 0.39523273148535487,0.04091846733792914,0.5212100261849207,0.681927020565769,
 0.7971715708632978,0.2949248033478259,0.5253618953566501,0.5036888789410764,
 0.4425197757710019,0.6250142361641866,0.9162508155813412,0.5341345294245403,
 0.7876225960637786,0.6498230207631569,0.720076474539852,0.7373819979057649,
 0.5483533964067935,0.6987021381310088,0.7143306226719038,
 0.030608339727991596,0.31145428473686276,0.9669140814350107,
 0.920923163626653,0.23528141514324896,0.3156956383898615,0.2069407011874983,
 0.3227740154219273,0.5553046891332196,0.9631443683068178,0.5400953900783787,
 0.36438469044818267,0.17077979698595136,0.7904681711888143,
 0.4770790489003247,0.7126940200472464,0.9096764706745997,0.4121958220244343,
 0.4384121355494429,0.47020876961724,0.414103488735754,0.195868705223479,
 0.5500582371851092,0.2311036981439113,0.025641190927655266,
 0.6310005956704277,0.5153891443952539,0.28394169405016734,
 0.28764615333259136,0.9007345102689428,0.9790370883997784,0.5400476435147991,
 0.4652429262477904,0.6898385221385116,0.20528020647866363,0.8846803270764639,
 0.21321312610728171,0.1581311227167116,0.08632156982596029,
 0.13853302380416488,0.49935845500599163,0.19117302130023484,
 0.7420461049608302,0.27025188716244797,0.8740006436473912,0.6023677676852958,
 0.6929973277163544,0.4058317034990536,0.22608861645438805,
 0.18391165388565778,0.7646445582247184,0.2216080146414754,0.6887117981889836,
 0.5436869167982642,0.6562306475807215,0.0248428978038594,0.2156167695688206,
 0.9144533181541228,0.24413152499452406,0.09655834912104688,
 0.9784246471180652,0.2527560562817307,0.6352092230009851,0.340060289439992,
 0.8311440284023449,0.6057011458639974,0.11962449598029767,0.5886070230211081,
 0.18876186074510026,0.7266286866594162,0.3459747335317843,0.9201535865350983,
 0.5926776764342885,0.42035398872750285,0.17706447280084114,
 0.3288524706371686,0.2515499945887266,0.2095973109717728,0.7033875460250871,
 0.7084173693801308,0.8269383633743923,0.15170153642007278,0.2393937972193183,
 0.5856881149915298,0.5788156206785862,0.8778472187717973,0.496910177793252,
 0.5405691878222321,0.1651505904812527,0.2531972467812813,0.4911840856092471,
 0.9591641086828393,0.976069571332381,0.3124077724023988,0.08948469650430357,
 0.025050385354518867,0.998368691720364,0.013064675551493266,
 0.006199099480491355,0.33339734109058916,0.8655560634975521,
 0.1795721242618542,0.9931098746431799,0.1011074894896673,0.750525384659994,
 0.6709005092727631,0.13937801673264927,0.5071541514073862,0.8478875766782583,
 0.032019915194936854,0.2208556409910748,0.6499468189830458,
 0.4941888022571421,0.364707408617009,0.33239312571977975,0.2798973895549528,
 0.6669317351143571,0.4304470174741023,0.2451897961823144,0.6806781051786053,
 0.5121415829751218,0.9434565557718597,0.6778108560118105,0.33400819954974337,
 0.9600857191132756,0.7896277949517737,0.6622758764728569,0.7466724373537341,
 0.14617517560643123,0.15883189877230286,0.21561922916430276,
 0.38707804729159023,0.1684958590480502,0.24913701448288494,
 0.6076576805543645,0.38506173224795714,0.8458737190752903,0.3113633977614487,
 0.31045540163864127,0.7962357704287124,0.9474925128858961,
 0.34164434544628497,0.02573292759396062,0.4070760258377256,
 0.8309465180831874,0.13944671472217207,0.37531863372540597,
 0.12010538045813413,0.6446355385956112,0.15380372088022343,
 0.7095287331111468,0.6739084965241774,0.7470011047077304,0.4491565495238884,
 0.4807855460749687,0.7577906674233389,0.4801915747802442,0.8233480385788434,
 0.8705235300104744,0.6549187682648734,0.5874894753983833,0.46244131129395083,
 0.05604275050298391,0.8646832010320461,0.05657959705220139,
 0.8219464536600971,0.6581560463900561,0.14656547849084234,0.5050276206845815,
 0.4614423830620279,0.6945209656567874,0.3813747426550771,0.6974047401352591,
 0.41656472580042747,0.1762380483124354,0.9038605901478607,0.1514956513867527,
 0.5147661720957273,0.19719571566390426,0.7449043775108075,
 0.15673619147465367,0.03751625072707965,0.43094179407893507,
 0.7282304482120657,0.4972109230020958,0.620717959530594,0.29311796999544426,
 0.1434789246517122,0.18349038688013497,0.4190494804094539,
 0.10898134819145244,0.9735815032786638,0.9525709907818237,0.9105402793375584,
 0.3474816413508276,0.983781236231248,0.20111175065901943,0.5783566094938433,
 0.7583551826090702,0.6197137580324734,0.7105351971113002,0.18892166781768216,
 0.36598400405190157,0.20324339468784214,0.2538909001011598,
 0.8524831808698989,0.29545383179616924,0.18980553620068397,
 0.3668171796719458,0.9138432185451189,0.24325928626108628,0.5127366728451244,
 0.510666627268793,0.5083612763397738,0.9740097485471975,0.5331559296380372,
 0.9149225022232523,0.37565016158630204,0.3717994528117996,
 0.37510270409353086,0.019072928576692738,0.8713401158801238,
 0.37224845314128197,0.12585014986444099,0.20747834143086963,
 0.3842226007435019,0.8243671164245363,0.9697489014116192,0.33176998469753116,
 0.1367975382734513,0.5832486043908144,0.890566908247904,0.5390630961955216,
 0.2528981484399918,0.05947209901152162,0.04932129113207495,
 0.6834937817494883,0.032037517787817604,0.6217158976987505,
 0.044923812723559275,0.3635344121495959,0.8300814160773831,
 0.07286956227238195,0.34853196695506805,0.6351154510323087,
 0.5843731785417214,0.5520567387360479,0.15427482876809506,
 0.40806398222572204,0.2517907628145054,0.10557600384828203,
 0.9810818817724767,0.5805835952100092,0.6254509809128399,0.4052648512438297,
 0.4351066309884165,0.06091671130768861,0.36070048021535783,
 0.5119002581067618,0.9410685778783907,0.3534259728740471,0.9856377265880019,
 0.74199802309594,0.663662541151451,0.267487944262894,0.7816178937785493,
 0.5033891565823987,0.8285278421936716,0.9062577634524541,0.37837048551385166,
 0.024940367641903216,0.9246048606326529,0.5981314811407173,
 0.8943699874974065,0.10635077565427775,0.40927548145775217,
 0.8728752773978926,0.9381778686274759,0.30150956962389786,0.3204271856715497,
 0.08233415445288872,0.5718427741416771,0.8287082675572716,0.9094815348071954,
 0.9335234226862731,0.10514244364665903,0.47643515010966464,
 0.7742986463008921,0.33348905499335435,0.9785550639244079,0.7540398855585599,
 0.745505220108794,0.2883562011918419,0.4903162527626418,0.50354465290318,
 0.8509191642238455,0.811388361809095,0.23283724516101756,0.6857202577696094,
 0.680487779867358,0.5267435925668722,0.3943329781173758,0.4216390660688436,
 0.9625186769008683,0.3656870342560772,0.06843767680532187,
 0.13899581569537878,0.6956153317372885,0.7718378205841628,
 0.15771803906503645,0.0655177691075175,0.12406820084692416,
 0.27386479085885906,0.20096641156226436,0.8704175142128447,
 0.9277677231632664,0.4005786770504467,0.5346803212180946,0.6049859403571538,
 0.9120337891385476,0.8753663362290737,0.03693908456628958,0.3791139166605375,
 0.8207228101777688,0.4396060915392733,0.9611106998232701,0.8471133237340976,
 0.922558308454311,0.8571341534861622,0.8924731152280198,0.17548810788219427,
 0.2280936027896564,0.34937164449648683,0.14597482927097527,
 0.45227326183272554,0.9091920586648126,0.46868259378165644,
 0.10429793311519087,0.8717197609730472,0.1740378729058345,
 0.42223811762037133,0.46939320007926977,0.5493804463305969,
 0.9826762394967131,0.8939902538748373,0.1199914839203503,0.8770074677020142,
 0.21464971363702368,0.4289631234009593,0.11115134117049763,
 0.41046128233904045,0.720557266134131,0.5664655574636329,0.9694133604251436,
 0.07877571409329309,0.7867045533309894,0.713573538792963,0.0155080922412969,
 0.15618166162956326,0.9151646148062551,0.023697931813937312,
 0.11445788266094337,0.06267436241413438,0.9643193838722217,
 0.9087440717162637,0.7652553158924392,0.13072164034589706,0.439725760791117,
 0.4507926722174973,0.7452054136535985,0.989438216884914,0.27077488950180084,
 0.20737409925958605,0.5380756635996228,0.20161241614329994,
 0.40468444618132693,0.0558314826633286,0.523629173562999,0.9011659657385221,
 0.15961561782702816,0.16471280426129786,0.08071178144536606,
 0.4537848468771801,0.4192745216262763,0.496234321301532,0.6166977201714338,
 0.8830537524625062,0.812307042357029,0.08762508223036791,0.38954159518175346,
 0.48043067101347603,0.7001501466161484,0.30992295583027285,
 0.08661845595042872,0.2856907070621887,0.1155770207703184,0.7584037091178242,
 0.31967905374935235,0.8636979794300476,0.024441695154787113,
 0.191888329113032,0.6175248140857933,0.9190203698629525,0.5502288198554511,
 0.6712379047977234,0.47525610548051,0.06316318410866328,0.03691764281512677,
 0.21970143204305437,0.9215732856950559,0.31369261499587475,
 0.8415929876001432,0.9684124949871497,0.19542147025979506,
 0.49292533443347275,0.5922297416027187,0.6525477445825717,0.3087159612174022,
 0.5012770626339208,0.08300214453477717,0.14214166191805044,0.851280315524922,
 0.44732420819625274,0.0913101768375062,0.08671564816280664,
 0.6631880858573762,0.6587747747325268,0.6212338888236549,0.2284812026126558,
 0.1519680092652409,0.27600280300337576,0.053630261457342066,
 0.15460073540812325,0.42190923142665937,0.8169462136376393,
 0.8351255267886853,0.7174475936314757,0.16133590423190358,0.7908278910230037,
 0.4902162150294289,0.3840481025798248,0.9878589694881525,0.18135136385879091,
 0.8399245498977906,0.674148069216907,0.2564268178607837,0.09094272821327898,
 0.47415022719374034,0.4335590863455906,0.7427379142627558,0.306490293558059,
 0.04276644401991492,0.6363116456075535,0.18794146827223246,
 0.09562710268661467,0.9016578415444777,0.753693037260021,0.7747793079619163,
 0.05881050017526068,0.8472478155610877,0.7429056840175217,0.6388652770642035,
 0.18383969311835602,0.7086205087326636,0.6674340586080577,0.8462090888478905,
 0.4619025957458861,0.3287236621165631,0.9782891509276378,0.8912804422215763,
 0.18329547618445197,0.3520933625595146,0.517567975939453,0.8050914031436878,
 0.9043734983054845,0.697381891060904,0.18652658632954755,0.9062677243025847,
 0.5131788322192148,0.331810009081274,0.7811186462788042,0.32250011905134734,
 0.12851402047566285,0.11678913490441323,0.8870839658816672,
 0.06117319146851663,0.12430713403719862,0.9117594718754805,
 0.7629635122309533,0.370196842456071,0.014527128300234171,0.6741918356819838,
 0.04241684772725507,0.7916239585384777,0.5966366602718911,0.3265700577985843,
 0.1697655880810438,0.3753335194638523,0.31810215180119195,
 0.04067560731386366,0.8255673219285283,0.4866020014144866,0.8561663396088679,
 0.4630079487837556,0.09875601724156557,0.755533533041481,0.8189297151368526,
 0.854672718665614,0.9900077093309301,0.693208519650778,0.7581217552842001,
 0.769083940865382,0.8939149409291511,0.35574081584811845,0.4913147524014274,
 0.5492560032118002,0.09762960845300961,0.5409281436215211,0.7182276125622251,
 0.19777616184541935,0.5323612733364423,0.4143554078541465,
 0.005141873787289697,0.275546688095217,0.4208607806340623,0.7822343327296707,
 0.8344651942103012,0.033409934845459954,0.7149862332480856,
 0.06485116947502267,0.6421885951744283,0.5707534465572585,0.7986900350982012,
 0.4463414595655032,0.06438947614187529,0.7452079954656288,0.8657830420619993,
 0.6268262096166424,0.10054691674851624,0.26540528442575084,
 0.7622811744620832,0.7426022279339956,0.7743109932355957,0.8270582185664243,
 0.9002706136802228,0.6529780549073284,0.8479136396067487,0.2359867069321152,
 0.1396045792672086,0.029069098062784793,0.05186380083502151,
 0.9639762726602469,0.7528149715489739,0.23667498094217443,0.3683219062515276,
 0.9234160741064649,0.8660053107536467,0.010634803331004083,0.890189124504498,
 0.2711532513410476,0.6033219622472473,0.5708709910380575,0.08802701597618434,
 0.2760456146848336,0.7910752539474931,0.25380826879547036,0.1055449865564222,
 0.9499792316743488,0.053009654520618676,0.9002829871623982,
 0.24053717298280575,0.6062515155632295,0.8028651624786384,0.570568398316257,
 0.8359686129795483,0.035767504791175186,0.4268741350535725,
 0.8521771339645855,0.6967799025329446,0.9684077091261022,
 0.022587341077727974,0.4013812704451697,0.01697211371196161,
 0.3874218858788844,0.5774742805803658,0.6276195766259753,
 0.005314050652420077,0.6004987071917023,0.502598623228097,
 0.40086519671552523,0.599147330521121,0.7694828723302345,0.33990262408367267,
 0.8283866114914628,0.5850312632686565,0.9197498545652768,0.08645190352973275,
 0.7747486691437553,0.3268958776080445,0.24264019009099003,
 0.13338462838493614,0.769373226749856,0.9830047448330315,0.6259200101887871,
 0.3239118485329179,0.38354933927441404,0.8904708766945075,0.6329483683559026,
 0.2788422012840296,0.843288757472829,0.515260834197239,0.21563402977717694,
 0.13404057066022157,0.06495223641472048,0.5984893598486775,
 0.10800416531133428,0.26885004458955664,0.09978135378411834,
 0.8283704851141175,0.47251419862914135,0.8242318231944986,0.4011876588605734,
 0.5136187481012735,0.3266074344344386,0.30069710922921133,0.3312091386591828,
 0.39507296226664235,0.12116486165273865,0.3503198818137402,
 0.46517108570400334,0.9036155104615006,0.5302632942428005,0.513124785352302,
 0.7460559710050244,0.5599837143352956,0.4128666910677765,0.4136937844848091,
 0.25255982803927335,0.6135021795564862,0.5156460160995908,0.5899861911116892,
 0.7072032790500649,0.8427013680068747,0.23063932158917333,
 0.14884898500657318,0.645380912194896,0.4637026883227051,0.8959554540922131,
 0.9789012230911743,0.3224273908916191,0.5360827443154701,0.3513578449117648,
 0.1981877071441719,0.7830482727566213,0.7268149633572498,0.01286247951631836,
 0.41348393634704883,0.06984805439189545,0.054715794676778895,
 0.3133904570401618,0.029804582400570045,0.6175754182981061,
 0.5048398906417686,0.39368280901365493,0.7756306404113466,0.9786441765332321,
 0.7091675883284286,0.7573423224509447,0.8917497406617485,0.9336218160722295,
 0.09907967824701547,0.5556072306977208,0.5335093560668369,
 0.18260916217480583,0.6131852820894385,0.35453527173326793,
 0.4409001245244648,0.6296869231888085,0.6384249455617359,0.8750910230141548,
 0.9048678495669199,0.338308424313741,0.5827306977943947,0.6581353116442996,
 0.20750989920085716,0.42434125012266033,0.7803239954366357,
 0.10501887064145121,0.703658612186798,0.533611079277776,0.8789459730052162,
 0.575926674412534,0.08328202779883687,0.3378766971026921,
 0.061344846185134694,0.07426078774892797,0.7445488175761725,
 0.9539438073679319,0.6276474881237415,0.6248129406212763,0.8046979745259512,
 0.49854500172777505,0.6199472180766628,0.004787811009638987,
 0.47878670244341937,0.6929315284488627,0.8652392484762521,
 0.30338543097790416,0.163009687294286,0.08921556587830781,0.3302159412523973,
 0.34337160363157626,0.7641915005411242,0.8014374240616372,
 0.49435354384068464,0.42887646803017143,0.2801510157623309,
 0.6640498162025563,0.670660025545091,0.9086131041319313,0.6575019261587736,
 0.021092080726386686,0.5373773707168314,0.7246129280286008,
 0.18553257186288374,0.6422846110169469,0.2966447909706933,0.3228406008334561,
 0.24491208204190473,0.11106836449283553,0.21270557721237504,
 0.04678755256412637,0.17140193132178638,0.3570507803989107,
 0.4469313138250437,0.34173853597685755,0.4748118946581167,0.6007845506460093,
 0.5647168102604814,0.2843199169518169,0.9444367111448149,0.6548980013060401,
 0.35434983529493813,0.3667585479471177,0.5843967971024177,0.8704978616185748,
 0.13659202509168922,0.25930604530851353,0.13678916642377315,
 0.19261512428618566,0.7451656709769869,0.029054729243729316,
 0.8047665013427319,0.15391858115798862,0.7965307545785654,
 0.30546957280494946,0.5451141545390815,0.6911547808829619,0.4187626830032729,
 0.1036873423158231,0.16183663261388226,0.37278884786187394,
 0.5361378775558174,0.3272389630128,0.7994975581551145,0.8639814876120013,
 0.3031045220862596,0.460281589972814,0.9922051826253734,0.6450680158990718,
 0.9100349870502737,0.18730403538680784,0.7450657554221879,0.8087444896705849,
 0.8390436374228263,0.6245971232750923,0.23229752955841776,0.7784878415650187,
 0.979517262793536,0.4964780870981441,0.9720000518178192,0.11381049053797332,
 0.8110507422293634,0.5938818996674705,0.5419094301544808,0.9002063407312092,
 0.846109824718256,0.7632909070082322,0.09505050111157742,0.9104846217181227,
 0.5619061306351696,0.9588391839606171,0.4420003408336195,0.41271323167640295,
 0.38314810281151823,0.238463251002198,0.805573671185906,0.6206232297711871,
 0.6784146915116822,0.7408950683458582,0.9765103924143945,0.43005890314649486,
 0.06532743089791704,0.4575951255050006,0.6334835666035439,
 0.25556080832545014,0.5998545087216195,0.6054837155289797,0.5259210892503828,
 0.6010546991786968,0.6225277419108901,0.20290922656967414,0.6998743393130226,
 0.39817640993099446,0.4497318757522655,0.14775897137395755,
 0.06133446094011885,0.8942637933369707,0.09936685957236768,
 0.15496866980380386,0.9779434041124703,0.4249079914280649,0.47727612132105,
 0.1452353910371158,0.46102494204018385,0.9491912144890061,0.7873238121308055,
 0.2861382646684011,0.9853087204540372,0.12158531534062755,
 0.23565216636610142,0.38167120889625417,0.8907772776652749,
 0.7375385662422849,0.1260254597717141,0.9020673338342027,0.381976049929862,
 0.48690169280694395,0.5923033066124365,0.4744679893468555,
 0.15406075187601753,0.3197162266277592,0.003620814093439062,
 0.7684391012161068,0.2541920925074662,0.6064891777264221,0.7683623329430348,
 0.11920059649187098,0.3059926302498944,0.21138420446681194,
 0.6071418451601955,0.3909286397116457,0.964563540207269,0.8851231703922242,
 0.6103514994201694,0.2826380223909397,0.3100655884729818,0.3383974485086707,
 0.7436870807746451,0.9374272438779064,0.5114871146483942,0.05974198617798865,
 0.5684252842973392,0.057853828895974124,0.599975922118586,
 0.19459622390332632,0.9148013650445881,0.5016960951993409,0.3111569445866166,
 0.31369083468064207,0.3216227059452387,0.6991899066095252,0.1804912689990643,
 0.22213836396507436,0.6469064491945389,0.3069699114172072,
 0.11664932418185481,0.09512484065090732,0.3308502756320254,
 0.6302847501770553,0.43936811788694863,0.1710214874901287,0.4784164560850933,
 0.7642194652215281,0.0020878829874535487,0.631266275554311,
 0.7923904541145501,0.2952323720991399,0.9681670501765604,0.423364717726851,
 0.8600178587458862,0.25871331321635027,0.2567544052658012,0.8501177404540455,
 0.02645006810015771,0.3797720170516008,0.5337862723482825,0.9553816991889129,
 0.9543926974246791,0.7188230175922288,0.8388995282886981,0.6411527809336205,
 0.06412728752555164,0.9427478004796941,0.505597596578335,0.4928256721191864,
 0.46686738628651103,0.5778419647585278,0.13467210747093716,
 0.22856875824875134,0.27276588263245294,0.5497819715104459,
 0.8807674909115857,0.3666591299651365,0.611359214722101,0.7678788481368761,
 0.936128947738721,0.5356236517188776,0.9710520381451405,0.46193088166324714,
 0.5822947356326782,0.3004985259526678,0.8367128458161058,0.5060714025106627,
 0.9058648629569822,0.39228401355533693,0.7996005104115825,0.9444010509963559,
 0.710987471028575,0.44024338755362824,0.7365669580207613,0.3387857282138854,
 0.9946674575952464,0.03309409517834627,0.039991845034397344,
 0.3970048662844796,0.7677867914774241,0.08180492117015725,
 0.28220739604195344,0.9475722194082464,0.7041273025054046,0.5224834533977998,
 0.7451270940644428,0.9835670105537833,0.7681541554526617,0.02169182820182236,
 0.5324957713642897,0.3875158018628361,0.5942052002100993,0.6546242842335663,
 0.6618011306883872,0.9091453746137342,0.0744385223658578,0.7242410716695822,
 0.2912267458539214,0.8705369982930591,0.09882559056308504,
 0.09893622995674844,0.248176919593613,0.3126863659261865,0.9141157164908801,
 0.40380917652788095,0.6816175725029365,0.4213840643863196,
 0.04658837935555438,0.6257521771772406,0.6921983065920855,0.9912358144808922,
 0.7641886884849196,0.6902000439424754,0.888740006681094,0.5908133654311043,
 0.87654615551428,0.9909283811105746,0.7154619194328827,0.9662663945457497,
 0.30076444033971994,0.06793792291883305,0.473123646989949,0.1390841556905047,
 0.6116283924213154,0.779060534970078,0.7907114709683536,0.20200229996355945,
 0.9002452039124761,0.28005310860886934,0.7498489064546939,
 0.46528330097060244,0.008241989536084349,0.8628190077971435,
 0.32142651501008246,0.9889572822173399,0.04404208157393075,
 0.3491262576707339,0.3062955511011445,0.8579471552322485,0.19548040528910748,
 0.9380375683126182,0.3945297401463622,0.6993411330388407,0.01874568830677037,
 0.27192791648963777,0.9875433194358756,0.7338706151793478,0.6246924090712715,
 0.6441583243496001,0.49477375504130305,0.25496413010244234,
 0.44095724809035886,0.777583249239423,0.1768137897727169,0.30487682301796926,
 0.08024829037675207,0.2886757377083249,0.6688154643416347,0.6772819547589745,
 0.016829136576157033,0.456199402593388,0.649460913228857,0.135430126056316,
 0.5987313916124458,0.9466923044123908,0.731150965351404,0.12628865975232295,
 0.7030760397653588,0.35013692721474543,0.7487814963318467,0.9288856556908215,
 0.6763341797756659,0.9658426844499524,0.34511999007740535,0.4985624193919519,
 0.9200747352307879,0.33719050315291155,0.04804515133722087,
 0.8678904381690244,0.9746105542924697,0.3509316728110815,0.593406158371573,
 0.27740588930599075,0.8005885358259341,0.9889597312611423,
 0.028713591007923922,0.016768194551761928,0.5184904680849226,
 0.8731218872944034,0.6684052378137998,0.03742781454612454,0.4484115982885877,
 0.19254938544008215,0.1451105275140132,0.1632888435491695,
 0.39443668899840323,0.07296098272706641,0.5943780820764354,
 0.7679913648058032,0.6136767977383071,0.41505404943062274,
 0.035163903355252546,0.04504644295689664,0.9277545157545983,
 0.772159698718754,0.3373508587830485,0.24003089043443404,0.894908122279031,
 0.9816268725901729,0.5349610788268327,0.9606430920133783,0.2101397796362221,
 0.6702171733276479,0.468161008454706,0.10620039922614943,0.41105456253755446,
 0.46045385078357826,0.8833540832077452,4.045323820128788e-4,
 0.5632229696763315,0.560877117780414,0.7330432361280047,0.3760399517378157,
 0.9418910662178123,0.6903493460012973,0.4796062636200962,0.2850058270200625,
 0.3660235695348968,0.9250186914001444,0.5378225605072803,0.6519696909092026,
 0.12655131944764075,0.03397478787185704,0.6480784487173292,
 0.9921685178119914,0.9912466794231726,0.05561240960286762,0.9975682237464397,
 0.22900514335810807,0.673708068271894,0.23821342478716034,
 0.16978620463305905,0.45767564393578075,0.8604263827136381,
 0.31479737926944906,0.45486127240526364,0.24918843233528187,
 0.7593079339540711,0.05507224477758055,0.43162088546770094,
 0.01115910048239932,0.4240651671756711,0.7041521733843732,
 0.04414206950553734,0.8203526859252775,0.15427310122886262,
 0.17474443896633052,0.4170877471876454,0.5071555320452125,0.6045096874095457,
 0.8225528053813917,0.9398464440584862,0.97447547370722,0.11117394435164274,
 0.3374262293152096,0.050196094245787526,0.900216110296468,0.7136415078437576,
 0.41767069286602254,0.34333347169861295,0.5379345358925659,
 0.13454877417489353,0.5076824629270972,0.1711814650346638,0.8481639257885215,
 0.2262513140800242,0.09357454354610129,0.17320794136000872,
 0.4442629930691022,0.025167074425244906,0.7405046813290543,
 0.9501340469850265,0.6143893151780228,0.26509645960023753,0.284651329340709,
 0.6224810641851596,0.22510034364433062,0.0245387611712109,
 0.43227074745032934,0.894267877613032,0.11775848694097946,0.3515937041517446,
 0.007708804469949482,0.4098633461056267,0.007140617045684361,
 0.4619689796882709,0.7522278443982465,0.34842713616177456,0.7901343608332515,
 0.8305041172966237,0.7293802481183852,0.8958804983719076,0.354331133088873,
 0.8881860061410555,0.06957762283763314,0.8033791091743537,
 0.37314670806322425,0.9072384006130302,0.8707378945681844,0.6318684547738537,
 0.7568142198325857,0.03417456584695189,0.966718686376195,0.38648109007434306,
 0.7310532493782556,0.838342308716749,0.33900157929802033,0.4395837607989366,
 0.9519714179362324,0.8177866442810787,0.3200882180240391,0.12543422034170637,
 0.3601009883547476,0.3283235976892358,0.6129194710675091,0.8257126924743278,
 0.21395931959051118,0.3776774686473946,0.14972739078664954,
 0.6066234665091519,0.393348533218165,0.16878731905507616,0.42201996240395134,
 0.16294348284019533,0.5559829935523972,0.5777886012536451,0.8609025430659216,
 0.49800879822396527,0.599289034301018,0.01691329145671494,0.3087895158224211,
 0.07419867595960483,0.010276742596791677,0.34624998339211877,
 0.31509303433170077,0.09037219634774996,0.7761234612992587,
 0.20947666600182746,0.4017242242405137,0.8012127453159258,0.9472312190401438,
 0.09868486302951263,0.02419624387111685,0.2067331571346286,
 0.3401256911726911,0.7793745164888501,0.08867927507142592,
 0.45977690140419436,0.4657516245094532,0.47690226002903935,
 0.5987407643066088,0.22066076829361525,0.3054013240270219,
 0.21900973828416304,0.48625128408991125,0.6152481593599635,
 0.2232277099567047,0.3448472650732398,0.09277805953340235,
 0.08913174799478096,0.3750459517097874,0.8763522340695314,
 0.10017197043119208,0.7941013052519348,0.8264993648654065,0.7336054870583713,
 0.05326746536698024,0.30597495189646584,0.6014882768297585,
 0.7671188777774868,0.3036479434259888,0.21228949027570887,0.7606172864331652,
 0.6103125518496274,0.9740412526461066,0.9550680896794821,0.5985565265566244,
 0.1718548607703977,0.2541692922904166,0.574048860809868,0.9419656512420325,
 0.553574180274749,0.3303059797940251,0.6740540479911239,0.4250562456520759,
 0.11105373617376879,0.46968545947998486,0.0927488145107519,0.527991901562107,
 0.9336494745971652,0.025779085956315173,0.18945773916126996,
 0.39830250673340195,0.2517803757842917,0.07936993232688672,
 0.8435388288177806,0.08810858978524816,0.5936003326715424,0.7516252266735671,
 0.015188589051994605,0.2704444926736278,0.6837205604257948,0.884850373802988,
 0.014187406600550179,0.16532540187702505,0.40842448675514853,
 0.5064671602883459,0.22073519838819688,0.30540147462160094,
 0.034133882175621744,0.26652788188479537,0.7649643844432037,
 0.3355683932338759,0.33522927068153685,0.17136115834374233,
 0.033320181203199906,0.11965579599984477,0.8435642768972369,
 0.4734834845008954,0.8020017330652233,0.35856718362468354,0.7643530455568621,
 0.3956598557810026,0.4126386976068416,0.6923326825009241,0.9738041678233361,
 0.9093650207016484,0.6417056133698964,0.6238302170920899,0.6917260110306266,
 0.5332057116784024,0.553567915716221,0.6520152740627464,0.38742400780881847,
 0.5560955463245812,0.49990134853310364,0.4264397202538963,0.5520492660195806,
 0.5579527962687384,0.0563567241642684,0.7097322623974747,0.8536317094722556,
 0.8258207972481149,0.08165452549312469,0.03400092530614218,
 0.9634022418417438,0.1591939719700085,0.5308433501436282,0.7315367660652625,
 0.08063513475023265,0.7245903270537184,0.5181705097054601,
 0.014782404408903216,0.43388026932884904,0.23408636260948146,
 0.7060897321917491,0.18382568344657868,0.434747417884524,0.9509047715076844,
 0.18467507040326536,0.15320781304906494,0.13561462168417393,
 0.794048261173855,0.15571315493309712,0.8849040106347872,0.7914436282872029,
 0.6138233435837134,0.36336081871659565,0.6586733373525753,0.7507719117849454,
 0.9082222216997716,0.7326644043702019,0.7822088916367012,0.4607202977461531,
 0.8798829470927833,0.7314992606365376,0.4406479856466522,0.9854573983029368,
 0.7807039853625539,0.3247232823186278,0.13851746505137474,
 0.48151068914113215,0.08060600267285722,0.8501690646699506,
 0.2143280688106347,0.3526224231724985,0.7941685066162254,0.8240905053121134,
 0.2373444932916926,0.4814332040421896,0.2301869310709478,0.9861983620957783,
 0.18648516104401913,0.7846244118314515,0.010192703784057344,
 0.37922346630019854,0.17704989582937258,0.6654818416988048,
 0.9686162424322408,0.12873971145938168,0.14844487753496693,
 0.15856657831727905,0.6857801409854829,0.3071254594027166,0.3160955933435754,
 0.13058513709810116,0.8305336061945016,0.059852428511811495,
 0.24545205924904756,0.48416446179131434,0.9362440654702385,
 0.5799987327815326,0.18950458617056531,0.011032805392472422,
 0.7703453817324533,0.7268659306974846,0.033144347484251435,
 0.39595089236367387,0.7159774091311453,0.7400373985377126,0.5885690642007557,
 0.8023083879749762,0.9652339291377015,0.5545883519740368,0.6452462308854449,
 0.7422353270268143,0.6561059202388322,0.6113758680478524,0.9231193130127562,
 0.39752969616179135,0.24992678214204178,0.4927611960026308,
 0.03426470917090951,0.7837707528000535,0.8212164153970363,0.2555991839111764,
 0.7794934869674369,0.31824447004898193,0.19014582371478284,
 0.5399092152821897,0.7284798160398549,0.4192099053998346,0.5482807599641518,
 0.8698798314689502,0.6882055794597239,0.12475538948655596,0.5927286562126884,
 0.12170003487874659,0.8223877625137376,0.9750712440550018,
 0.10301120887394466,0.0672234979444164,0.8513925979545697,0.2900219766912563,
 0.7157108794800052,0.2517147252254692,0.7778140641243843,0.09783747897146666,
 0.40904514158913574,0.4099714884367216,0.8749149071376944,
 0.024377412789215347,0.22096976591005035,0.6445711585220468,
 0.8168189186114118,0.23647933179474023,0.3970726611291925,0.4793580199721924,
 0.7429942393049755,0.9582954385666731,0.8875473874477003,0.13157619815346877,
 0.7475655440984794,0.44208935853116893,0.645904100394406,0.18986679454674982,
 0.19280226671817857,0.191783412711645,0.07077749437882175,0.6686748069513314,
 0.1677722254531374,0.9554704500179756,0.5891841306870216,0.13042711665916662,
 0.05854621744059074,0.9938679218966381,0.7220472520553584,
 0.10489150257426894,0.5447395652122743,0.13037736627695518,
 0.6040731597881206,0.6150370619363329,0.10016502834175733,
 0.07007628030556423,0.18319715906920497,0.8275443220981819,
 0.32219953208332897,0.7747531121867072,0.33881836458866454,0.171326279954933,
 0.49218062685038966,0.16556242957368894,0.012069131427680513,0.7910981636746,
 0.9649552375134811,0.5242326612059215,0.25266985991758684,0.7046692306927738,
 0.03077298717028043,0.3025513680016195,0.4573516677846432,
 0.008387539553311063,0.440135069611607,0.5344823953694684,
 0.48700459101605253,0.42409689610624235,0.39233353500729573,
 0.9571654076580985,0.8341782491605679,0.13788372118914383,0.4859282009316165,
 0.2437344409086839,0.465343638454388,0.9414927154894573,0.041929628519056905,
 0.8805664774194678,0.9380369419497945,0.5542092124081349,0.6427290293746475,
 0.5675793786984735,0.6020767470221953,0.4423137225450825,0.7158377551674218,
 0.4270351311816092,0.001772229504002576,0.41569315471392665,
 0.5277616568507344,0.41067458901517184,0.17050122255624744,
 0.02216497104928372,0.7147836228517043,0.6206444557306735,0.6014299533964113,
 0.5597443388966306,0.09832964378516595,0.9818989659040844,
 0.11967807706930578,0.08436439943682184,0.4493321771291324,
 0.23102515381762856,0.7840250660833926,0.8753240997101589,
 0.32968621699912726,0.09152014099072847,0.8355108761408129,
 0.11060416989668165,0.3588219008184983,0.8859817973894938,0.5880374347831686,
 0.6389607796285033,0.7994610710796671,0.3961004636004437,0.5688657921672584,
 0.1108935880112597,0.1950781343682686,0.22477272799369996,0.9854673090099456,
 0.6005890912192853,0.5202357669617665,0.8317613892187193,0.2057407346696608,
 0.2858091911566901,0.7331505972238599,0.05580445520229649,0.5640068525522217,
 0.13812954801728916,0.6830550595417711,0.6754991335039762,0.9140893733992081,
 0.5857375891002128,0.2119883675526575,0.7487728078290314,0.875731729050925,
 0.3962249170642934,0.7850688815049761,0.47880788913965777,0.5706820522214493,
 0.8269407637075744,0.7416413895091346,0.11688221704204993,0.8224827183438419,
 0.7597679846339164,0.944072434409707,0.027004250626823545,0.5607825132914831,
 0.9941319408857423,0.9691030937532541,0.3961650888417956,0.8542302960171375,
 0.28905687352674614,0.7790694060943801,0.8538884851749996,0.9884453651701496,
 0.10192686285676644,0.16511432995933928,0.6464007035138698,
 0.9367371215973133,0.2712602414048333,0.1447564210154142,0.9540836752279637,
 0.07326051441514092,0.02213530601132241,0.5068388881407373,
 0.35984467742169557,0.05492384160334762,0.47963089374113355,
 0.7711161054426462,0.8790416919311761,0.5462495428186883,0.6076054089303677,
 0.6200159557888499,0.686897580865725,0.9656332732645244,0.49388734800106426,
 0.325345109895823,0.6965882179534144,0.12974722108142378,0.41124955054137946,
 0.49917412523093974,0.9970716811265505,0.7574798464319541,0.5825552269063504,
 0.2028973153513257,0.10316968099588641,0.3789295735815058,0.6645895690267345,
 0.8096350714580622,0.9995374584733491,0.7262427626650881,0.18185145522330515,
 0.9094087884636275,0.01201838634341823,0.5944593519472723,
 0.09584840589690191,0.7500196660042191,0.48621131456490674,
 0.1416553433481853,0.39846773896092824,0.6808654267254246,0.785535586979917,
 0.11797547718718782,0.6805909651422628,0.09674379298219249,
 0.010623838967256938,0.006167114957876141,0.20016518518315496,
 0.45802001844577256,0.76940726925827,0.2034836878488534,0.8680632545928281,
 0.882369733797493,0.8129483480910098,0.14941763912499817,0.2484994508437819,
 0.29517385170349686,0.7148950504411813,0.6403581766402038,0.6595268215190765,
 0.09445942731453816,0.706273780506403,0.7163010996677431,0.6560724461453894,
 0.9561166617752904,0.18589214257695297,0.6542778435833764,0.5092994192473249,
 0.42956461933132895,0.5962743239346439,0.12391835864429301,
 0.6973535711264708,0.8891880854484736,0.16327399109106489,
 0.09026252618474662,0.21957789579364095,0.47173428226608705,
 0.08175725500696629,0.792465325970863,0.2519581966287344,0.518889097165491,
 0.3446513733438702,0.6915258422081936,0.8756733256676205,0.15927727098667877,
 0.7394739861641016,0.15221582871759332,0.5771522328720979,0.9638657796149621,
 0.6743644225831126,0.677007591589536,0.3266472694480559,0.24060151970134924,
 0.9391744110901024,0.3685794710081468,0.1789658200879869,0.675636428916804,
 0.3641826351118781,0.7293038991147558,0.6959184290077707,0.23887539055036688,
 0.5048615058717794,0.16579345566825499,0.7681775113287793,0.5273747593285105,
 0.38950543745777444,0.08258550273080001,0.4495847960885564,0.267969557735504,
 0.8919155990267538,0.7021126882194466,0.3598903650965741,0.3173887351541448,
 0.9846115513686091,0.9776342843013773,0.1281526863500484,0.18332463422641587,
 0.8406276998259696,0.9845941284044877,0.5081234453875584,0.8311087794848638,
 0.8713191145659792,0.5907028432545824,0.8241163270740248,0.4914750290288388,
 0.26215187633938264,0.16360794905804743,0.5498966784371302,
 0.1728888399507985,0.3178775013277738,0.6205134452337012,0.06719202508999056,
 0.9234783594845638,0.33960870443024416,0.34177007247196556,
 0.5794487724842864,0.45684378081128885,0.9801681480205884,0.7476563771367197,
 0.783530450488624,0.6462681939084584,0.6583970065911011,0.09898861945383364,
 0.14282453508050774,0.9895795225427936,0.05387093739342996,
 0.45260447019369987,0.10421605107924681,0.5791117074933714,
 0.6539401986795871,0.8425346533493695,0.4083848828805118,0.06980498338609209,
 0.4371290654509965,0.6452511702869473,0.33798368775068177,0.506859135028589,
 0.7162023661713841,0.10214803741867373,0.935064959077081,0.8115444229074379,
 0.3996201888349733,0.3394367910672281,0.9118852407723306,0.6500944094716888,
 0.13666140682902495,0.6086937010724636,0.6811221923795622,
 0.14788466394359379,0.38224418961676987,0.928422740021901,0.7698080128328455,
 0.9688234065673278,0.4949603491217228,0.16564700873928262,0.4324286063812244,
 0.1554415284423165,0.6294082943319783,0.35270569093004245,0.6658480434198137,
 0.07036348256057223,0.33601395003635703,0.2413096854097314,
 0.4413202768020794,0.6558110127614226,0.3987677474647837,0.7648290040062422,
 0.9723291530940698,0.051224884226497625,0.19424256499197723,
 0.5809531834834991,0.5906292500482567,0.6862857515493233,0.23573722542095799,
 0.5850037157657197,0.6108711151389495,0.2607534103392688,0.6227475067492902,
 0.5989596959292546,0.8529980314164836,0.5742325657978448,0.485105767103132,
 0.007153308838036887,0.8503705421486306,0.444786812153273,0.8723272497540074,
 0.027844242455171142,0.8751434925480919,0.7476402369750965,
 0.6140379231938721,0.015023649530594096,0.29276744748592454,
 0.9959321769117512,0.5071869435442818,0.808120282966829,0.3984349492454655,
 0.7186885020483348,0.537170329996647,0.5765340464048575,0.5330745264286151,
 0.3201270104485223,0.5738047308407667,0.5163999570416343,0.6455302153983257,
 0.5179723283196719,0.5656081240895281,0.5918457283289066,0.36488194172196364,
 0.6730841528204251,0.8053091850614823,0.5175820431728427,0.17276759588822954,
 0.5201001396050006,0.4870647012156124,0.13279144014073652,0.8260880887885772,
 0.14776731687889422,0.2707636060778592,0.15049309501494612,
 0.9040601235697868,0.4376440997924096,0.8194160007206132,0.12499566654770888,
 0.25320406157721576,0.7917164501624813,0.6829354430665533,0.3279381302090292,
 0.39445127399429425,0.35667866627245537,0.8043397812741881,
 0.7653921099334867,0.8229308193657254,0.6927671043859044,
 0.007428313618477289,0.4403456827733241,0.25081189646443636,
 0.21777239441286955,0.19111681478074238,0.0034722062781851193,
 0.40170821559275804,0.7837170355970018,0.12049700535032637,
 0.16763153774120032,0.10919594168559599,0.3379307292333682,
 0.41404659869257865,0.8206638214329823,0.7746095634447037,0.831957818346808,
 0.9798063869538206,0.2418471516308217,0.3481176983996477,0.3302992417933823,
 0.09351039249330362,0.1773902182019217,0.25151979622309695,
 0.8308141254095802,0.6992517962616197,0.3528329027912166,0.8997676564561559,
 0.5091900011265732,0.9514786880819112,0.49935126179759626,0.7102638407002051,
 0.8605416080984591,0.317339312652994,0.7449684870185357,0.3515203481715623,
 0.991882164724185,0.6369992003972198,0.3603350044049739,0.4417372862469616,
 0.9244911248488108,0.8540830767151143,0.8832923168498282,0.4128008790550608,
 0.4607451245644285,0.4984084242550023,0.534914664593803,0.5320587813726371,
 0.6058311218160838,0.7058153313073401,0.011532176711131936,0.49848963826007,
 0.8422340362326064,0.3396878335205109,0.3472484719127644,0.5385727627169897,
 0.10315956426658424,0.9114763068824183,0.7418204229988827,
 0.33082074980386944,0.5321230695939883,0.43410830528927746,
 0.1842021702588843,0.08395361395920753,0.570790267942706,0.22187803574881504,
 0.42615646564172704,0.8185314447700199,0.7517653819558525,
 0.012655576357924914,0.5755426895663621,0.016064880592068898,
 0.9057642562384514,0.7322963040615406,0.9718347855942552,0.9440975161268524,
 0.6634950638847398,0.5436393298606083,0.1035211899028986,0.3905588518183838,
 0.16228299462857187,0.2432477668531824,0.03181203975810154,
 0.4878202925415015,0.3427829880196688,0.9085190466777492,0.7997601521286953,
 0.8030247176780277,0.940759026681143,0.7838385056565146,0.18676329314430062,
 0.2369072263159201,0.1516258053814884,0.1743634312369834,0.6787191665567179,
 0.9711448993185323,0.304924525648805,0.26735780040167934,0.36503083371221445,
 0.4027926862470078,0.4706496334616648,0.1155005146668786,0.5568807369995998,
 0.428501205939945,0.271353778700108,0.32233821094568105,0.6623765396000546,
 0.05971892435787529,0.799359688298654,0.0712797073205581,0.6863416478477091,
 0.30324621841351684,0.9985991011908509,0.2710941764812447,0.6610019840374165,
 0.382361942933082,0.7587593989474577,0.5961737803350119,0.24395216831746824,
 0.28217310818589114,0.5793093531503313,0.32567088147139633,
 0.8963729200378169,0.8394525088100294,0.35529705643954823,0.2523650488083087,
 0.7986842478740459,0.6343182831121863,0.49551994256610665,0.4248145533702623,
 0.41012236435004734,0.7309097451798439,0.26727960676674234,
 0.48988520373090405,0.5200962446102554,0.6484285772280879,0.6675052241531059,
 0.31170340249868633,0.7060888322656241,0.9660612990669614,0.7614816878920013,
 0.10371664120667101,0.0368180841971788,0.7761156069426974,
 0.33713248530664985,0.8878352663346216,0.43298111763264446,
 0.4332487236636451,0.2154468563451437,0.16923725549964508,
 0.17417838918714612,0.996871291474736,0.14684708036726946,0.5832202945663625,
 0.25958342397329515,0.14547009842049763,0.34061680722970744,
 0.9454023509263368,0.6790356909504826,0.3846745973037735,0.5653233928105581,
 0.9843299624526404,0.4207304616382984,0.31586823969559963,
 0.021999275248958172,0.24222158632625945,0.41098263852749384,
 0.8455231450579465,0.6608163794232117,0.490906439991435,0.933349309981801,
 0.7643296055398939,0.8104567320439632,0.8703254945096172,0.9451600378287623,
 0.849130670943802,0.03327792152840969,0.26992164781637507,0.5278742851268484,
 0.961065408587339,0.5739312675575922,0.3025174378407901,0.37147682595189,
 0.4442787945095992,0.8802791531340536,0.3129413995199528,0.7933408919099668,
 0.7474492226295557,0.03968913302439647,0.7653257044602411,0.3699072342207387,
 0.5406346831467234,0.38405282168999477,0.39489160331540063,
 0.3004784281630606,0.40155013518739824,0.4941062551603147,0.5721250874299246,
 0.7524371754063139,0.3868604700020104,0.1721510604701606,0.8394325148943437,
 0.26559677438864837,0.27063653728993653,0.041675778575041234,
 0.7035930192492322,0.5448086881922185,0.0908484280947689,0.6478941639517092,
 0.7363967483981767,0.050752602005178504,0.9329164361102942,
 0.18386392691450482,0.7540016113972778,0.2680322274695738,0.2605221514031004,
 0.14432170507738962,0.016480965092179556,0.6284296156623075,
 0.6380643932326899,0.22576533753700456,0.5574824793543981,
 0.32742239384323424,0.5324486970170377,0.04593997977871889,
 0.5840609869107887,0.19407873232771577,0.631292176828618,0.3820199324222425,
 0.8484819451370461,0.6831009091608938,0.07886074866947179,0.5885132457118207,
 0.355923054069075,0.5982923543518563,0.7491833769974623,0.5172160940806746,
 0.7403061067391065,0.8667646610534364,0.7147633321722038,0.4328730784928828,
 0.12926268869088098,0.5353181591700276,0.35797463637430493,
 0.7646182156448585,0.7691185946057258,0.9798748292030505,0.5357553094687617,
 0.32437393646666024,0.2684204195045904,0.45874025394349327,
 0.8980103587222781,0.27720128035481517,0.7575726946838803,
 0.014056756487448308,0.7916726800251173,0.259199602987241,0.5956744391522949,
 0.6032034549164818,0.04620433749960262,0.7604226754675525,0.2760164109623551,
 0.5596898194058451,0.49893584863016305,0.6569755650457554,0.5039278619136712,
 0.476111009515342,0.9655856679175735,0.6474933163830225,0.6428887132317656,
 0.3747890151898663,0.5813941010554421,0.3612936849495211,0.32277348373731196,
 0.4368426739891813,0.8522730836981168,0.3169922337149548,0.08526966837377692,
 0.6093650250083112,0.5077007819087963,0.7365179570969039,0.4160363967109324,
 0.33056469802671773,0.43273471866802593,0.7007192017569219,
 0.2565582104634284,0.014033568154372067,0.3310516661072599,
 0.20399485261364025,0.9811068963863906,0.3297645357648379,0.9703013734546644,
 0.33052328713745127,0.24084383878123072,0.9095903720766183,
 0.27958553857496027,0.7726282824090084,0.33328079050623005,
 0.3588151698589179,0.08887539902324648,0.9218492341568346,
 0.24032892309889897,0.14051438638201308,0.8056056367924909,0.596142751391713,
 0.08116619928856161,0.1532528460694731,0.7792748951823387,0.9803288618426864,
 0.5052177050992938,0.382665784052837,0.7628949844511212,0.9371636841774276,
 0.4935892285493192,0.8412762125615924,0.20014431795622134,0.8553890726149835,
 0.22615222164194837,0.5248648843595161,0.4925151529271722,
 0.11998716807097842,0.18819746277541072,0.9253063276065409,
 0.7644664282289374,0.8470125707657872,0.27778520219817904,0.25857648679914,
 0.2259806146777117,0.8820249276830515,0.28949292042954977,0.4697269237957553,
 0.2139916830580193,0.5169691190422111,0.752152706799091,0.48476875394303287,
 0.9044033548520781,0.06141041517101531,0.9709197382102941,0.1503106881107017,
 0.5673882835793277,0.5920648295096874,0.9066962093894168,0.9207759639349943,
 0.9628593931293509,0.5797841191043964,0.059696725534269635,
 0.8157025605507938,0.6679018960793076,0.4137622638434191,0.7933529873614248,
 0.39002535184903453,0.4010994688305699,0.20106825336954615,
 0.4173389510582409,0.9582146175425589,0.8513125467699115,0.5945771760841533,
 0.7212931733367642,0.3737887713972333,0.292972055803008,0.322012732534221,
 0.863876187421336,0.33574035119082857,0.29116563547682406,0.5124557997031386,
 0.12056779487557368,0.3289345288532569,0.9940118138856889,0.8626541943436556,
 0.35885737413811536,0.9237317817612298,0.8280478587298826,0.6140975007412408,
 0.10440119197686404,0.5104471396214563,0.2744617691269453,0.6484291645650064,
 0.5680540211785264,0.9354994570482287,0.06452664781153583,0.8310579179111879,
 0.7103402161545987,0.9351465065097604,0.702322573830243,0.7985899266010692,
 0.3750923133246198,0.13852327534135156,0.3167170630579217,
 0.31221512738156243,0.7282942523958389,0.894750540865897,0.7468107603993364,
 0.9682690674423987,0.5860540794557484,0.6694060412334341,0.695940197726954,
 0.40406428130475114,0.32184487094073655,0.6338667917262445,
 0.7298497759002944,0.972370806020555,0.41711201283387134,0.9185579133511004,
 0.8168248773315413,0.7343195162634333,0.8749480176388461,0.6180075219005412,
 0.335765301338109,0.41844681322353283,0.5520897095208609,0.7352994141068714,
 0.0766847930522645,0.5959998384721451,0.8987862568641948,0.9182879380021075,
 0.47895825588568264,0.7395530799806449,0.017561130348481058,
 0.7787145598519022,0.6865975177756328,0.49716144906249526,0.811640245108915,
 0.07813877686338455,0.7000710070881949,0.7847857621704066,0.8931604379006236,
 0.6514101733799245,0.29030146482876185,0.9694486610987568,0.5747647788769896,
 0.5824153379257131,0.8180488184340331,0.4101629343701223,0.42549705065545407,
 0.023216777105059205,0.3921545481553059,0.14241880620037795,
 0.41855054350746224,0.13761849557376227,0.6027388313499867,
 0.8164441983957317,0.7378804238346394,0.7185590331454876,0.2765340194217768,
 0.2377249246729617,0.3299658531688816,0.0805179430025117,0.5787688611610272,
 0.015725900408248727,0.7785316147080301,0.917216853323294,0.8207621261797271,
 0.5744464823453892,0.6105495672921735,0.5893285991433804,0.5442112323960293,
 0.6334399354711877,0.7858905416132789,0.14368787885561074,0.4729900130768613,
 0.9179078904928674,0.42776398111908565,0.6217288588185161,0.6901220397962152,
 0.6749295668589285,0.5641962921020452,0.1333568160275469,0.9790827117297304,
 0.02573572652533651,0.9857246087322582,0.1405254865572111,
 0.09234354375771092,0.22261325149160005,0.4242185314914454,
 0.2160294686975287,0.4016867691409115,0.15087512083377708,0.8790112278744613,
 0.7253079211598963,0.6510668066453231,0.28611339290567106,0.6653582399744843,
 0.017509868220863978,0.9117470846916835,0.9530008140579931,
 0.8756560342746363,0.9788865478408588,0.2989065873880681,0.4746226035912473,
 0.8068647383117399,0.7140276044141096,0.9919858121587057,0.232579295227899,
 0.3885435183866266,0.8168053429334338,0.18978962806256328,0.7881143875383286,
 0.5128978006459083,0.34587327921562994,0.8424894117531377,
 0.06575423510790823,0.03609673618936682,0.2216525156420992,
 0.5047983214441341,0.2653860519531466,0.7189555729567553,0.188388084375253,
 0.740339467255887,0.3754674665197103,0.8455300288914614,0.8881882433230514,
 0.7989009249536186,0.8228579587760722,0.04760098372308885,0.6258271527183451,
 0.09685548103778185,0.19629663927311647,0.0742606757249542,
 0.08761512122673087,0.6793037846231605,0.8561863735769139,0.9582399595423468,
 0.5519248606717266,0.6586544509712995,0.26278465783135263,0.4516841132562621,
 0.35885913356628396,0.836837948832086,0.7136277678050991,0.04543014162164083,
 0.35194132087080654,0.11172732853708622,0.5537251651372933,0.648786170318702,
 0.2935205478524888,0.08311830058290959,0.6303664747910789,
 0.08673182449275951,0.6287070434763425,0.507411599447328,0.7520621851811021,
 0.4272080459258871,0.31408598488297024,0.27179908383155427,
 0.9341936107720445,0.6997524341199779,0.6583620403508748,0.28048230408867636,
 0.08668006144274876,0.8611701677124745,0.8859378515082075,
 0.10008411765553504,0.2608680560433001,0.49416818886205494,
 0.04265639118524567,0.40903335210295166,0.136257919523046,0.7782393543838438,
 0.22948126048810935,0.5359136446190255,0.8788805205254258,0.3728718670611654,
 0.6641090367333803,0.8515483615683598,0.3895127872859847,0.7554641893195053,
 0.9004513274122067,0.0864822330944699,0.11157731543753002,0.8001041577025538,
 0.6369589901442815,0.10864637796400156,0.253298712121214,0.1755693519803294,
 0.6275505133932726,0.2240784558007054,0.5212535613627735,0.5475217782863662,
 0.9845466291675927,0.755671161523651,0.3108542662069371,0.23979940442835224,
 0.7049931946855691,0.8862455607667187,0.8365564825184436,0.5954082072225961,
 0.5546866005426463,0.4753306750443801,0.44835901638011655,0.2663257086564056,
 0.7145299085336696,0.8223200852160851,0.3913898382057118,0.7815818208312142,
 0.32896242825483923,0.5353071650549837,0.9419652570314289,0.7776597586175015,
 0.2736493499440642,0.7009712371056229,0.4943014375746595,
 0.047865115178120754,0.49623711356368727,0.04923238595335733,
 0.4528645734468343,0.6545036911025296,0.22415601003227736,
 0.28870491940589305,0.4014288793524968,0.2939169816376137,
 0.08047122080741476,0.5230784727141541,0.9712677024742737,0.8430677934402429,
 0.5243226207865446,0.08441094108904434,0.4108407826357797,
 0.10289481650204502,0.3318650978442549,0.33345890146628954,
 0.5389939905071871,0.7671387127484687,0.6681628580970267,0.4727216386991042,
 0.6807492453432777,0.5195039769236678,0.09495188630516238,0.5069015324454216,
 0.6375090862970192,0.7529209059448883,0.7823928784729557,0.14188542852696184,
 0.8349550730289855,0.09226157421156178,0.8612499873370397,
 0.14107473303903517,0.4834175354187713,0.009963369761703866,
 0.9665835198888482,0.4422284595038283,0.02211191188666206,0.9014131458783852,
 0.7774042776599095,0.5843551399213812,0.4972366625769157,0.3182439257240324,
 0.42333536597910204,0.4856693555747199,0.7739166822905645,
 0.37111188329021283,0.667315087251777,0.20568816635817988,0.8651621045253525,
 0.2754358773689619,0.9107041979477298,0.7741768313050801,0.18114236605327605,
 0.4051606926719594,0.1949387017098072,0.5954089324395895,0.0769904987859873,
 0.863983964769594,0.625292712080215,0.08571730700743219,0.23755576717337057,
 0.6440525704434187,0.0016684584517163703,0.2517962116613167,
 0.011514497495967446,0.5783645434398635,0.6449505538996275,
 0.2953962475390999,0.0858355068055654,0.9402701736616507,0.3132772530579383,
 0.35252454345113693,0.16265242026897953,0.5359849061674725,
 0.9801150817282084,0.5149575718766664,0.8984357607257455,
 0.0015976765400637083,0.9794312347700125,0.9102379389379718,
 0.6154673147814015,0.27876945592367386,0.9944660298909763,0.8693791672965663,
 0.4816399929126869,0.5786119664393717,0.8000760986747075,0.4667631457332351,
 0.5642774950423773,0.811229995108783,0.674732782267848,0.41556555520420546,
 0.9442379124976211,0.3572455546578943,0.7641165449996454,0.9171425827284619,
 0.20407665652071638,0.5023351941819351,0.6262717552263211,0.2746482717967488,
 0.15818943177958,0.7058908325980617,0.020760300271130294,
 0.019051158195439122,0.6141417837807279,0.8347993726646765,
 0.7275644243990355,0.5045647865586642,0.5724276539081921,0.9924268855891789,
 0.8870801603079385,0.565419071174303,0.3656089726625331,0.581680446763137,
 0.1356371535913672,0.7468926510257556,0.4443622627090964,0.1906059842744885,
 0.8928905392119041,0.7799341301777764,0.21170476191842802,0.8668336714886157,
 0.7981827876860982,0.4274771630043752,0.40988220325559754,0.2676653637307441,
 0.6180263438247974,0.1563072237099894,0.14799521628968704,
 0.24267315396721767,0.7290877340026102,0.6559020774877805,
 0.33538929261241046,0.26391944291725367,0.6673059417645446,
 0.4986035411377384,0.5952855195349597,0.5594262685351068,0.6946698346921243,
 0.5931163776925283,0.0052028562167835135,0.9551822853871185,
 0.3264741361546146,0.09218481751639207,0.7387932083038208,
 0.41091808676792996,0.2413325440926697,0.8155620433267272,
 0.42933921753981896,0.7891820846408337,0.5756459174004268,0.4765870652893648,
 0.9036339158637148,0.22604435951081003,0.023388335758893652,
 0.8168914744590045,0.6060156118042376,0.5616807682732461,0.5931564431704937,
 0.8453382580971077,0.9059139731348153,0.18466456651099217,0.3376986870477243,
 0.8170694271240021,0.617250108100977,0.6439054738766498,0.5810768796182362,
 0.10540801907739239,0.38743100687622023,0.6951666611545404,
 0.5920590330030281,0.32899858100309665,0.8565551865493796,
 0.23111531972858212,0.3674889808398094,0.7989758449944435,
 0.19369548935832293,0.5807072460780452,0.593610361077217,0.922059647592126,
 0.8466727332540498,0.5156023821138768,0.2990791124817963,0.366733547152611,
 0.9394361795959927,0.5783785259261237,0.7224953722702784,0.726125881060758,
 0.08731608144667136,0.7432277177121686,0.048918136601786455,
 0.047785285952436474,0.3555915169869427,0.3056592680423764,
 0.31833870668385833,0.6458347193369073,0.499890547366557,0.1418122542042115,
 0.3212277699516438,0.6147123165171197,0.07210512109301181,0.1323998758145134,
 0.36737322426630203,0.24594456465793324,0.7596135139857371,
 0.9871030995168046,0.5407933108904184,0.9507423937376946,0.9923977042991474,
 0.9245960148240716,0.7917500090581105,0.5627710607210997,0.2290554918785208,
 0.12032017789006755,0.9142232725465735,0.6986062862474686,0.9897362257055877,
 0.785210942967082,0.6971643305478266,0.2979400987066513,0.430055744655805,
 0.6050563600186412,0.9775255764360585,0.7114370649503816,0.7612311909668847,
 0.26302618535852607,0.5744888220718168,0.20077139689750334,
 0.5022840825168344,0.615897683073742,0.6732330314740138,0.17521542566274717,
 0.40259955248036494,0.5041596033247516,0.4789320340730703,0.6618399507807217,
 0.1650364569681375,0.32057926450840524,0.9764451356452742,0.4002582828503285,
 0.003792991204787155,0.21841910145773946,0.9308569228993125,
 0.7607948411910824,0.471700618272776,0.46866441425621874,0.646070818370507,
 0.4561833199838682,0.03437807517672986,0.6247828185515844,0.3961589645975505,
 0.8609875897844536,0.09074346875160832,0.9980021811422614,0.9695121039480722,
 0.019785084809791464,0.047769203636872426,0.8176808138655448,
 0.6526765971886209,0.015536047188161595,0.3177199109538271,0.804899572968378,
 0.7498439700607908,0.28559351269333244,0.049918247537068194,
 0.7707080691790225,0.7142740825773207,0.5121856349454075,0.20446429607312877,
 0.33959982242606557,0.1848063119542379,0.9380268810834305,0.3622710673400761,
 0.8868412706464343,0.221239592063028,0.9282196320468034,0.5312100248914575,
 0.759974757401566,0.047156692312884774,0.898644852424852,0.41961135597312205,
 0.538069527665197,0.10172133926078619,0.20680867520047164,0.5869177384080002,
 0.5916519336197408,0.5064362062787375,0.08549927370254362,0.9513460198582327,
 0.8688941038502055,0.8571526553759016,0.09984048460533457,
 0.07612141292811592,0.6458115059689143,0.6583320075619641,0.8333863298984064,
 0.03872248474138795,0.6365383763875335,0.8731962914232944,0.7220976156494532,
 0.1361504872931426,0.49624084079355335,0.24120912665383776,
 0.3891029546617176,0.6630332960738164,0.5378814275341413,0.10653291774235774,
 0.46774521703946537,0.6568149006201797,0.3245897047994333,0.7762632961981233,
 0.4124299712895758,0.16024297908658258,0.26793312466938735,
 0.5339019689653179,0.39443533334258996,0.19394865634292302,
 0.2485713539312624,0.8469810097791677,0.2715610792863046,0.3071373858855623,
 0.38204140823010657,0.9342206749611113,0.3928802667097201,0.9707507730567178,
 0.9088120904341291,0.1197225668061761,0.08556018445594848,
 0.002068569930277264,0.05325892269544363,0.9596053565482423,
 0.6979042387389912,0.29676231987098756,0.1734069002555071,0.7876030437888788,
 0.5235081767739962,0.6798017270496061,0.17117174289964798,0.8283608670362066,
 0.9732903959395219,0.40262329057474755,0.18630277997623346,0.437061074845432,
 0.1912577093253447,0.9576322885293999,0.6959652459338208,0.6144714856278212,
 0.9557833349908729,0.30518094788045613,0.08573006449646103,
 0.3413809962563874,0.4502165639412742,0.2774176505707069,0.8732898170078165,
 0.19293206067689006,0.20390700178512877,0.10470849454834774,0.96917902028409,
 0.4549564185304793,0.7781444149037963,0.3880446493491767,0.5566883283369558,
 0.2944758658196387,0.13789452742649955,0.2995141905457587,0.6436584096460793,
 0.7662528873057536,0.8900884639994768,0.45936270911483545,
 0.16494909716841843,0.9322481425225343,0.5697779959922371,0.8543503692210261,
 0.5131396259227063,0.16057512295900966,0.5605413363025046,0.3263970127929955,
 0.7206303350512406,0.43289115492174624,0.25027508709719026,
 0.4653972752912773,0.37442115360122863,0.7314702061262409,0.8107406259912856,
 0.7881770522222905,0.043408255832279696,0.06464101499183039,
 0.28813024213336413,0.646457540254227,0.9383648740815156,0.44880142758456965,
 0.3074619988765228,0.47857348420228696,0.40011210209696924,
 0.20755521082360417,0.7334519326423975,0.5216373381996033,0.1483194360241058,
 0.3084868383365945,0.7638842479375401,0.916360577865133,0.40744668569370857,
 0.33946985870188395,0.3006224924682017,0.4814001922067149,
 0.11595646621734534,0.9627175038503514,0.06442199976157548,
 0.14828279410189238,0.797084016954713,0.23974652450727274,0.3799674437568116,
 0.797177183387601,0.9898717815927707,0.851566816709878,0.20887320108201124,
 0.7148008916312046,0.22245278797471357,0.5554536736248413,0.8226912636715529,
 0.7302302370169748,0.15515969108750194,0.8095257945920517,
 0.09643690852799469,0.9824936926668129,0.5942070503448076,0.7399074179015683,
 0.08355402528645461,0.9931207839469973,0.2941414864603171,
 0.40334097281495995,0.8716948187868874,0.025091886980781908,
 0.660133927345954,0.7083270825842273,0.5810181530719606,0.3701466343732247,
 0.31241753590782595,0.681054994725365,0.6958666362319831,0.6368079915980089,
 0.817539514220034,0.030938258704666266,0.6515259838667373,0.4292673565521258,
 0.2975973272892489,0.35789591583865143,0.6126809818219652,0.899135456768983,
 0.5309898029599818,0.5755536275168838,0.4491156336236565,0.6251075007432478,
 0.17729731986432318,0.939194735300087,0.9316970941000574,0.8455180917415802,
 0.5367611046210541,0.03840756720354199,0.6633944500059851,
 0.17340615501595857,0.5589557483493992,0.48088159646017425,
 0.4797217840518766,0.708788218303653,0.0786539886398252,0.2851084742710239,
 0.25313843431727356,0.17207244381837994,0.7718452170660349,
 0.9180581563731646,0.06899043279314798,0.4059046863781389,0.6336594774341248,
 0.47714103936938723,0.10399422990220808,0.4745327130809962,
 0.8710912569384568,0.5283247051818621,0.37591411096043104,0.8389154071474174,
 0.7062069623920821,0.40099314748413295,0.9080849712344823,0.7537124949440401,
 0.4491490099109344,0.7869427364569437,0.1759127350225569,0.9622467820512529,
 0.10446395627147609,0.24223166116243977,0.11946385534529147,
 0.8211421350306737,0.6199922298854158,0.38926920263383524,
 0.26558393624553533,0.7163983789061269,0.26042045015393755,
 0.6963115553300222,0.8588113664625605,0.6248217733533399,0.06415519210725174,
 0.3751898342681368,0.12680449030992758,0.16817201673511684,
 0.41087365202330717,0.7768889843012889,0.18973701510963525,
 0.9179548786613753,0.6876897042267714,0.9144136377343179,0.2234363532275061,
 0.9920776152125964,0.05859580717249102,0.5883807611883793,0.6817360241613937,
 0.695675229167789,0.06807424969821008,0.9639722592000046,
 0.029176144156219497,0.9120215153972712,0.671883389139657,0.6505069740376864,
 0.4272480632336455,0.18680632614871218,0.16098093173682393,
 0.8753169581255109,0.6013473456000284,0.9240121069720585,0.7206796281600183,
 0.5570351260240511,0.35968527049924504,0.11361264802166582,
 0.4721544966373046,0.23611430954566204,0.9318224146468485,0.8622323651303816,
 0.4575735166956596,0.2635799956070094,0.6000908545570992,0.2779720752913747,
 0.30941005997145354,0.3466951211239665,0.8393065422966027,0.7461710421257576,
 0.6127477541673599,0.6935700504505772,0.6271126846006023,0.6299240310696962,
 0.3685396255710044,0.3101528737098551,0.9592263035918033,0.9796738897811712,
 0.824008455816368,0.05916977617810848,0.37313811907646777,
 0.04019971121838717,0.8214588204097605,0.11259120942385359,
 0.18843626403711433,0.02997837496010225,0.6225681622394308,
 0.07508521118945666,0.45227851241651673,0.010626930239460375,
 0.6343073709335837,0.0278367305466638,0.6919122847302771,0.19314388742804445,
 0.4489209874074428,0.7217474865859231,0.28718321271736746,0.5787837204317043,
 0.3662886296167339,0.4373245659808541,0.3872603171419573,0.7331488879514899,
 0.6799161421807515,0.4361074263613822,0.3614362405082774,0.11012211228303148,
 0.31594866225715834,0.9693889712889893,0.37003706705588935,
 0.38774746619537503,0.38570353033475946,0.20689689031685488,
 0.3409163272026544,0.04456709143686444,0.39014621041158615,
 0.16489788114525195,0.009774157944721917,0.4613513270810452,
 0.04468039541794322,0.11092149161632714,0.7586419500311965,
 0.18999031003498512,0.18132328973995537,0.01385656721913453,
 0.4581714297169096,0.958923416570947,0.6378493930588849,0.7076754686437463,
 0.14501545326358167,0.5300514282385365,0.9949447538074265,
 0.34894702224921215,0.5227778920326469,0.9030998207899987,0.6975997756105298,
 0.4408092510914413,0.5191589604207194,0.905553042296939,0.3398223840532655,
 0.9995171764540755,0.6528559479805695,0.22658710753277722,0.7216968056349039,
 0.7552914611446834,0.858721868499345,0.8335343528197586,0.23391968090564386,
 0.7710026254157207,0.23914848119638998,0.8945026181113496,0.2720981757173353,
 0.27998673157816756,0.09633019804614695,0.9650124303676064,0.687098431441891,
 0.5017536516875009,0.7245404350631482,0.3512684301233975,0.8904904090619516,
 0.9744120978001592,0.40554699968908425,0.7596548260657516,
 0.30751439060898667,0.5111028719175623,0.49143948095171586,
 0.7668370485769395,0.3386015897632403,0.6770447587086924,0.9853638785312939,
 0.250554220885819,0.6085845705417979,0.4892302381247502,0.8835215276922819,
 0.7104899781565898,0.5614234705974259,0.5191745171201041,0.5549444766860607,
 0.5177065612666344,0.8237407310516119,0.41979890730962066,0.6849119216157643,
 0.5208295991231594,0.12958090697765157,0.7397544081414584,
 0.06974199963087346,0.8731900402012496,0.9589786632592662,
 0.34790999290074753,0.6793735620366621,0.31266269542126,0.9561456630521555,
 0.6280057267750244,0.2325120182849416,0.024090804766215612,0.502458996333889,
 0.8105240182566086,0.299851787816404,0.07430867465841307,0.8936890844228484,
 0.7787719967739851,0.33731675985902676,0.0918300927460095,0.9595898123884921,
 0.7078539620364674,0.15919499338801057,0.4168578789716969,0.5893414554258505,
 0.7414884409032003,0.36999729834103456,0.516276357116548,0.8164912558289773,
 0.7161027719959827,0.8278730568345476,0.3093607063770363,0.7787851766513572,
 0.9489207597296073,0.6260365606774694,0.12172953315596846,
 0.04237917415138226,0.5899567044229533,0.16254032003331953,
 0.4543193828470339,0.7193775595665248,0.3624677319178009,0.6022691587657885,
 0.7210353021016683,0.9729571942087144,0.959879825932281,0.3974660465717226,
 0.7523006775031852,0.7554230218808791,0.6512370868092782,0.5298165531181009,
 0.27579495669932474,0.2281112206597814,0.40954095864027285,
 0.49015035740526336,0.07629496983406558,0.24513598840777806,
 0.016181202252619875,0.9276117527175365,0.055750352074225185,
 0.15843218207673582,0.830617407566325,0.5215416375508101,0.30954170618952814,
 0.6737605995144367,0.9410788628303268,0.46841299077596743,0.6553384036941934,
 0.9213658111103361,0.641843469903837,0.8332525729988866,0.3716741857954131,
 0.8210350580485574,0.7082986611651153,0.5002132951194206,
 0.042400482096118575,0.4664253920075685,0.9094945182069364,
 0.09332625791472648,0.3756457960069546,0.6649026680579647,
 0.16687180315745376,0.26248591835302715,0.7885970987973818,
 0.12192904256203863,0.9375372640976535,0.3857762115970145,0.3656824452235783,
 0.5506258115320777,0.4838025629282603,0.1556916949416708,0.839720439051393,
 0.9569925415555866,0.4133877573259722,0.266911981624551,0.26882866209013656,
 0.8146426895400072,0.5748242542741768,0.6716985300766558,0.6417305378148418,
 0.34090995300025406,0.5282244051365255,0.12661889598555975,
 0.9163702797460666,0.3960237764156427,0.8627075742830903,0.4698309110109362,
 0.39170158755922035,0.8338281288439122,0.8398623279443013,0.6393689020955412,
 0.8569684953668757,0.4360833308592248,0.3314307633066478,0.858708416878024,
 0.8535216383545673,0.9968294129135973,0.40666371222777864,0.3671826034853368,
 0.9814798206511113,0.12200564184975515,0.31439895489918457,
 0.9190197587371242,0.18069256422379731,0.4812934035402008,0.6557868119773164,
 0.5167783164097024,0.7992077203936045,0.6979811316776278,0.3437227369588075,
 0.40590828854211336,0.46972564259654637,0.6720785205362108,
 0.3548081610840431,0.5875024190782622,0.8998350631997785,0.9043610731066805,
 0.004718332704258921,0.46818228590365296,0.32334510958799256,
 0.6757410507367004,0.9447681788083528,0.05780239494477957,
 0.15644532225521468,0.6188543141737577,0.05677460921340627,
 0.9274681433198673,0.1637510583915972,0.49877330023325983,
 0.05621027076224616,0.39571607696365396,0.11581258228047764,
 0.1816977474647844,0.00604286219577177,0.5699252249393882,0.862961907187316,
 0.34350008764855844,0.5388605726933546,0.03860574601986966,
 0.9185289941386291,0.393255307527532,0.6513299199096946,0.6520263017071365,
 0.8439124325036222,0.9283141915578357,0.5021811142574304,0.8403218134720377,
 0.7606871361151748,0.8529211104204228,0.8791237993278418,0.05933018678201896,
 0.20384054503940452,0.9788048680334034,0.17614014481563223,
 0.9185771846782689,0.3101690480601962,0.0604303631631784,0.16593322178545344,
 0.33801189224463324,0.39980472897967,0.2584018218631754,0.9820259311757868,
 0.41549039384204767,0.8983826702758692,0.8801751210893283,0.8177661345801828,
 0.8109756420639437,0.9076776670421317,0.6260394067306448,0.4958729420833127,
 0.2206455526219403,0.9899662051763386,0.49616808705447946,0.3067738817246992,
 0.904787142024936,0.66038682250292,0.4847778600305648,0.31216679395003466,
 0.6814286799545767,0.3705104860962155,0.7585987080890167,0.1414445430250154,
 0.13194648914465945,0.632848481034578,0.22218855822877148,0.8465170508064817,
 0.27896244831428285,0.7312588903721671,0.19315528817852523,
 0.03242861593451263,0.1608386971860961,0.6647713841490287,0.9200260050481706,
 0.6685996606123172,0.9068882310845066,0.821779988969085,0.8831270226197094,
 0.5276641000263796,0.7983344046245009,0.21666770917231704,0.3509307071711072,
 0.028908627665619302,0.22598959319613843,0.17453189288934712,
 0.40659917172480553,0.196430342237507,0.5013921962543092,0.3540203148316965,
 0.1549811178946915,0.39227746353229587,0.47045895665802107,
 0.8586335153650956,0.9671247904757191,0.7770887342366725,0.9383914471694106,
 0.09519632907425923,0.8693831375895238,0.49696738335394197,
 0.9474517188021717,0.07655293951838471,0.5289030054046974,0.2159355810325505,
 0.986859362757913,0.09281882947550424,0.7234267469509345,0.907320208303491,
 0.9769870077370001,0.787998603485421,0.9129788781338284,0.9405199554525888,
 0.7989913618204446,0.6585585972683428,0.184011698104099,0.5653077014337647,
 0.5610816921457454,0.6593959267997663,0.8807198893313988,0.8027287334137909,
 0.5940909234728432,0.8420770943648064,0.06623346372254668,0.3417085329973366,
 0.37035045921696874,0.8720653031297879,0.8197165321548621,0.9347105782688055,
 0.6493399102219218,0.899285585364942,0.9636723921580042,0.40579202821299665,
 0.03399664014995751,0.9557981867009435,0.4247176022850826,0.3606729314398427,
 0.15548256720137799,0.2622532422329795,0.5387095355582437,
 0.08202673249400205,0.17528696248535747,0.8373869203019466,
 0.20415505008970447,0.07367691236785079,0.8252221409913233,
 0.6543008375119904,0.7800271918795264,0.8917946544084672,0.7975407614427066,
 0.9280135380862139,0.7988013348915903,0.058872200446812406,0.239215815436837,
 0.6330928152673315,0.7617885754402102,0.22065364832040413,
 0.38286533315699556,0.010956924054479233,0.24764534910898361,
 0.514792555480783,0.0542987073418022,0.2480961570885366,0.4781377185365503,
 0.732451031515198,0.7525612439790844,0.46051686035803774,0.4980579343566993,
 0.49798177732183957,0.2652606138596687,0.48843804527393886,
 0.8830962324269052,0.22089208802191562,0.9222677452461744,
 0.30305422632089796,0.8040659585027626,0.7605539810824103,0.7243987016121678,
 0.5068774777923228,0.5784619862709288,0.16493952345703544,0.3413956874913473,
 0.6399965753195184,0.44614339192642083,0.1779413073175523,0.8479196391452095,
 0.250349391163494,0.48004307802968715,0.8649458449005837,
 0.010685951868545707,0.1931118274247371,0.5134681325547069,
 0.5623157298053303,0.5446244935979798,0.21107897731262448,
 0.17012634319031328,0.3690164398594038,0.17873096537536104,0.367714015565082,
 0.2995504493962665,0.32839260581432916,0.4341075240341904,0.325927291303755,
 0.30319501720365394,0.4316310086296513,0.2337683624672654,0.5023031498677591,
 0.5842392179084662,0.9764519240115895,0.055724921800452965,
 0.6645645207835013,0.18612462559287002,0.07966820294547172,
 0.8177594341661459,0.32260967247797045,0.7344935897222802,0.5062422716954129,
 0.7304398095171393,0.19545597037045792,0.8966775659048362,0.5816490281093668,
 0.3999644072814118,0.8781116059117904,0.46684105350410754,0.9247521698773314,
 0.4330400426620249,0.16169350842118246,0.3947907690095981,
 0.27138694695734744,0.7492993010296879,0.8602265030058065,0.8436183224336685,
 0.8396615088437713,0.46707013043020706,0.8996715038034097,0.532956296218377,
 0.10245276804083459,0.7603855211624074,0.7241963992168302,0.7454640007026869,
 0.7078819561278883,0.7247775891726,0.6196882956518367,0.6198460717856481,
 0.6787063317926799,0.8418356797404712,0.31120790305623747,0.1689764660998383,
 0.6032177921607428,0.391895178675183,0.3367011783267557,0.7217214611611729,
 0.7211735558264945,0.8320839526847026,0.9811746286404183,0.0930234001857666,
 0.5998064489856629,0.9682422224214202,0.24551958466427748,0.595655621677563,
 0.28179149998853514,0.05336205813222539,0.9866212876960553,
 0.4592413013234756,0.4100694307222622,0.7344306109382358,0.5298220514227211,
 0.6915652024889055,0.0626530741921516,0.8862417028663379,0.3821518009179402,
 0.10220097361148073,0.05272341252500801,0.5036455432564025,
 0.9898956138387698,0.8371332001257923,0.5730821007532467,0.3865977577840135,
 0.811300350628011,0.38728046143004513,0.062233267631118916,0.763944903947551,
 0.31585181791639017,0.3537331516350004,0.06242741862263401,
 0.6215436088221483,0.7495545778096491,0.22395408046824894,0.7182158593537415,
 0.31352839889101414,0.3455007137723598,0.04682475439339351,
 0.13890682260731857,0.05810952115541479,0.39573981170023576,
 0.5802172295160161,0.331627709312295,0.18364471294907891,0.4690647147628908,
 0.6628367709992922,0.09988070132449345,0.731259033686725,0.9234234682197761,
 0.09169403240570362,0.385190695665125,0.4438910022274585,0.752700950339168,
 0.9340447986228533,0.7754506482537181,0.46246852018349904,0.417872814455188,
 0.3530488722361316,0.24549862834769443,0.8205148494724834,0.2594484136946077,
 0.040831057500697376,0.9658194898747756,0.3694677913059028,
 0.16889520161497018,0.28791040139884905,0.3099052528551859,
 0.8135245164866528,0.10335665692734497,0.12834960794986627,
 0.44413367712979646,0.8470090443502176,0.9553956825157887,0.1772900020806728,
 0.7499926420807902,0.43864744971166825,0.22139829016564838,
 0.7106812138662468,0.39742116959083307,0.8511674771622627,
 0.006338268511632328,0.8887033500472262,0.32135000221619225,
 0.2827696930874254,0.3377348956186549,0.9940591443525767,0.876132227351677,
 0.3701506224614034,0.6668171330754389,0.88785727100961,0.873328005894908,
 0.8254757169724493,0.9899778466540465,0.27079979359655915,
 0.002691175066927176,0.9738901175201993,0.32410410745749363,
 0.7315383723892797,0.15332569493643944,0.35263161625073247,
 0.7044213383436922,0.31575466179839096,0.10233026223453834,
 0.6713618403720789,0.12343539745461496,0.9187825992939764,0.6305436980115836,
 0.10648737035302835,0.4779941540167838,0.8205810512397065,0.9615581997448643,
 0.18843029224220165,0.8013885242267066,0.530891338234743,0.8793312025430848,
 0.7901152932792476,0.40544177518573565,0.31960793077362615,
 0.3460469073732506,0.2590867438949108,0.25030959127226526,0.3974471210695498,
 0.08425445896257178,0.39397851651068283,0.5227879794931586,
 0.44724989597957365,0.2865261801661745,0.8253216432059897,0.5506980821979005,
 0.9722288389495851,0.2010728272526563,0.0996300636578944,0.5479407824260826,
 0.3751882775475861,0.8467418675461842,0.04510808270090361,
 0.28610656058073414,0.24851768334699798,0.3148299874466205,
 0.5621166134375982,0.9974367135424275,0.8779538663097515,0.8902040974987423,
 0.4140558320518637,0.33041985287703957,0.17953878743790253,
 0.7727194130806716,0.7303191872211031,0.28967300753809433,0.1668821755581399,
 0.09732578177441165,0.5680225948099861,0.9601470697544693,0.6674681821689894,
 0.09815797574723895,0.8404705321525784,0.4744420211531951,0.3032839091521937,
 0.020204215920083035,0.5686350781519893,0.01961823338274249,
 0.06401104768954713,0.3209194185414955,0.7737396866753077,0.7963558523210186,
 0.15453583999900633,0.1501001043933622,0.1590123553620434,0.5924332174173946,
 0.4794889737874547,0.5392705371565756,0.06661037277595261,
 0.11439384484447968,0.20067237782349834,0.6113909482161443,
 0.10277195076420431,0.2841288706418694,0.6174488165959111,
 0.004978009204404055,0.9972885108010195,0.7025357587177138,0.69304792214621,
 0.4055606181840694,0.3637931793178709,0.6538290736735315,0.8605104293070402,
 0.2872783383467553,0.03636767521528639,0.053708833238361064,
 0.12613659297324054,0.6510136865813387,0.5993763844054795,0.9890533627503825,
 0.572617646897722,0.32949776362052385,0.3341608521022842,0.34761970425777644,
 0.10861326183404962,0.8865626689319661,0.13188422680472345,
 0.5085793433063506,0.35709848639980524,0.5265636664253379,0.4365331389487075,
 0.18721541295588606,0.39355657146823175,0.47114800283952896,
 0.8539425045734542,0.5270238595680411,0.9886251987523946,0.10147470487927657,
 0.36740391855133725,0.48034934073835533,0.735489146836837,
 0.13060702989387596,0.26625139944312004,0.071423104274555,0.9760112982798401,
 0.3487626312536537,0.27213768970576835,0.33064706317537407,
 0.6036594508673421,0.7197965391034626,0.4000964001681644,0.264578763981996,
 0.730004425832482,0.29240928916944386,0.4962825830026729,0.4581199296616302,
 0.6987507554238694,0.09725280976377015,0.7341891457087946,0.9652577871649388,
 0.9254518616876214,0.6119192397001213,0.803383193280289,0.5482799016176658,
 0.7309406335464782,0.16505595027088182,0.33262262069207704,
 0.7451889144779258,0.5060691756108703,0.12624168795141177,
 0.43021116596280007,0.9261553003341169,0.4187325789715005,0.3314009772166727,
 0.6005206927367491,0.25374844969065635,0.6968024757404522,
 0.09161361029820825,0.797932387220055,0.6191298806357619,0.41165546183990065,
 0.20250822117299205,0.9059259313305459,0.652832222365205,0.11425327317684086,
 0.540778091128658,0.2673187421719454,0.9661073867011936,0.46425621454159594,
 0.5251386717335876,0.7013995212101569,0.7792506263258635,0.7730941602811069,
 0.35685782908113206,0.3356122943241445,0.19248401804573478,
 0.019033672909505417,0.18763747589501079,0.21725582314202418,
 0.7147082809847998,0.8777910595066534,0.057705504609797065,
 0.17089237875232288,0.011484044658489623,0.55850918113248,0.8361905046026195,
 0.6170669186453184,0.1433211634110414,0.36967024336947607,0.6634679596918864,
 0.8513576882835212,0.52968238829644,0.5054925912321864,0.042176309418604396,
 0.9017220603699808,0.9542154347843557,0.5550054545755371,0.3215130476897985,
 0.12427517167031787,0.9936882945912434,0.23590470052454937,
 0.3393752165271048,0.052237848275932475,0.696933828511373,0.7706339049358647,
 0.1520092245908219,0.3035594283714578,0.012789138705993519,
 0.7186009875557462,0.9486658112211593,0.882475380365046,0.8853424493483275,
 0.6143186026130475,0.20488208533647323,0.591654849338089,0.4152096559593026,
 0.7825724370865254,0.8172873244501574,0.06768725804196885,0.6587796968294053,
 0.7023045794337414,0.5362774393867751,0.9894526547373348,0.9582356229596856,
 0.6172970561951556,0.011000451967844271,0.9822232025379687,
 0.12011207062245055,0.22173967301446773,0.5660154943835088,
 0.006560478993328012,0.19762406221570683,0.5593314073995614,
 0.6897240801465028,0.5972225907487883,0.7253319558885415,0.4294457186314028,
 0.9500510023653466,0.8669371537550618,0.26904022857773147,
 0.12231400514418944,0.6623155846781273,0.8511773107226928,0.3756350771766741,
 0.12157668487609663,0.6766814745887187,0.33116465611170187,
 0.8992690224835719,0.40336413955234596,0.332938718561665,0.9260712820547914,
 0.8474689679588321,0.5134291849412294,0.47107832863226184,0.8955301886208272,
 0.40074844414761,0.31679767341165976,0.5433993549390226,0.6369652122365691,
 0.7705098622590246,0.8807919791914532,0.44557671849271874,0.8383553409575295,
 0.6163098060628461,0.24569574904293168,0.7166093962486345,0.4864191545277041,
 0.6022558156085236,0.2255910211784692,0.16167780030444634,0.6705569380067187,
 0.5780430432250241,0.06924237992804971,0.9258341486746978,
 0.09273567138020122,0.5722155887589473,0.08760670527291958,
 0.2837040532746109,0.4087174150465436,0.8734174804683187,0.296720170008979,
 0.8288702611871439,0.6234179224443256,0.5903052243714684,0.09496758342452649,
 0.9602265794262786,0.43200234624395484,0.6711204285147283,0.6279155993114376,
 0.7125172946398375,0.2292444417416052,0.778681537625516,0.08330473741176336,
 0.5055087662780472,0.38737892228187243,0.6425385683471247,
 0.023115069240626562,0.567943454278596,0.5391499171660684,
 0.47326913463497333,0.7399374955072495,0.5497494359705842,0.9895350823437259,
 0.3810549631167355,0.5199516816391787,0.4171318050311976,0.6761538291474892,
 0.9284471070395554,0.22986376800817432,0.11623040371090265,
 0.8968450040321937,0.16745687431549217,0.2186569496182994,
 0.44577608775430777,0.6476803863078874,0.5252317059008672,
 0.40336592262756255,0.9477833198177716,0.8654249673187516,
 0.35765783242634064,0.8397089189038291,0.982127317640118,0.06670242170988339,
 0.010363087759511891,0.9765100174905863,0.9844488925266556,
 0.5724859696295672,0.053352328826980266,0.6123807470405075,
 0.0785962583043126,0.09171579763971704,0.08013614708738193,
 0.37840521781504444,0.33922745659019715,0.07839641372346762,
 0.3337769379419415,0.09857018390975758,0.13920288987266005,
 0.8442592680080377,0.10301054347770022,0.23124952904883533,
 0.8424876423943616,0.4633931589270921,0.4855170852694686,0.8730119368824136,
 0.30149348758316463,0.4733998865646364,0.07337664278880074,
 0.6046229496371804,0.988535623110307,0.9224110707487962,0.9718043394460569,
 0.8873012539978834,0.928893783261133,0.6434146277719492,0.7530789905395636,
 0.4893042935020726,0.04081149904158521,0.37183572366603046,
 0.7902986891589026,0.9317337630113797,0.08373104025110145,
 0.11633262364606956,0.7894725680454876,0.466297258617959,
 0.015498578747275782,0.16291230900976328,0.8099100679814726,
 0.7242966624586724,0.5396682305571436,0.8572487501928127,0.8692388105237222,
 0.5098688675217187,0.2539641137206933,0.3194586610223167,0.7022808017293863,
 0.8069823473900614,0.03988417071131922,0.8414373561284684,0.0664551205268501,
 0.2678419851232259,0.2937921514950481,0.6647718129113829,0.9208376073677893,
 0.24324262081789672,0.45042704764068675,0.856774153563227,0.8424897019616395,
 0.7057515713472047,0.7004981023967434,0.375171470676091,0.8965161749111589,
 0.26985156848247605,0.6330469435187528,0.10179650554852859,
 0.45947195341286795,0.35131251598058777,0.3125651560756806,
 0.13389038547420362,0.430153468874637,0.26519480973421294,0.8824566147976974,
 0.2062537549672958,0.23868364358195981,0.8188289737316113,
 0.26961429150759064,0.3575225034800673,0.7384872419498704,
 0.44092506584056623,0.07332127506183694,0.49947710092398856,
 0.2601903428401211,0.11378592442365143,0.5188995748602253,0.7325480845501474,
 0.27857384891259285,0.008140509470210167,0.2137854950556265,
 0.5375171993474439,0.49326556212688044,0.6910454334204158,
 0.048941026130530085,0.385208775924747,0.8235434883612082,0.8761305646014272,
 0.16494033098374083,0.6399713086319335,0.10581664736142615,0.836012236419834,
 0.42009199362714456,0.950599380705213,0.7332081486151282,0.43357300377824126,
 0.08373205191020683,0.9060201903069602,0.9058654616687453,
 0.44845684126640706,0.46299423044697974,0.03682876042323202,
 0.5257659319073723,0.8459079032492093,0.8766068860685523,0.8410103874657107,
 0.5343133301262283,0.1459926643663192,0.732766489441139,0.8928709018183394,
 0.6553191112062905,0.4091098553003055,0.6491677390915267,0.6912837111273312,
 0.7723392521149304,0.6150128114907539,0.5472753672856226,0.5642444558389953,
 0.7348474657045253,0.6129373413345526,0.7495072478741183,0.8371017029424354,
 0.7080868282668464,0.14671982598606192,0.5313549042680652,0.8795753145433115,
 0.507170718928437,0.7436755880384027,0.48079662087133657,0.6944100319406017,
 0.5840649069612475,0.5388845481209159,0.7484047564859524,0.4531347574740089,
 0.7805419915084627,0.394600874658823,0.3975353958915443,0.549699445889162,
 0.4550020006157439,0.8635611794244374,0.40424914470316065,0.8861000529201644,
 0.3885958078865388,0.756038468600309,0.9881699872105513,0.43521265299108347,
 0.11133316940828897,0.5256011411960344,0.8002058830362707,0.4055869559020977,
 0.20324307316423673,0.775508416724552,0.4659797221578641,0.10545708371804441,
 0.6365821500697937,0.3357931007145353,0.9478418250453478,0.0392615921349726,
 0.8180355335925011,0.5606875259571674,0.18864913091676272,0.7667836756405497,
 0.10267605305823091,0.6851769042351962,0.17805926816691353,
 0.04161574477254759,0.5438984917473393,0.056193310525206486,
 0.5801796738765603,0.3180773070967027,0.5370368530238699,0.4074414741750476,
 0.010075332539897675,0.11791801050475748,0.09627392473802932,
 0.1984823084677465,0.5052006376688596,0.6892915119727547,0.23207172946193977,
 0.8794964538536485,0.5111255453961092,0.5042811252345414,0.9127118131800956,
 0.7919625267930431,0.014217556855206426,0.12802879244269383,
 0.5843190926103085,0.9605321997906973,0.3152933837282603,0.37479917933942664,
 0.5765678684693235,0.3887624111211665,0.502690489093,0.7731375841355216,
 0.12374886686678754,0.2742492145273603,0.35609928798446877,
 0.34230003165692346,0.5990503666099919,0.9054165548681333,0.5188953352674391,
 0.28146735512791554,0.45047379814081245,0.1956904343055168,
 0.9617339674231283,0.7308871986038912,0.22048616606774862,0.8580956612641142,
 0.3740982554164137,0.3130426154291741,0.9988225315895296,0.640094120922779,
 0.3520463461156451,0.8540798810581991,0.7278282838992075,0.5864205502234703,
 0.7302034424167115,0.10590039714717614,0.42260302370260017,
 0.28836595175758417,0.9264493640801831,0.11898490136111739,
 0.10153066638101715,0.9831286398898614,0.02110859255181463,
 0.4851847863768657,0.10652576194964403,0.3687256889673072,0.6010871209936534,
 0.48558641268320213,0.11599427900557613,0.5614728640925671,
 0.24205588870271588,0.14438008661249646,0.33248931786492564,
 0.8219701157624059,0.3409620121946264,0.2510581974335706,0.7259238565736973,
 0.03411671204047928,0.031700851247653095,0.3913694445938246,
 0.7341740227058975,0.17145860961127168,0.8001394397290111,0.7015555827423783,
 0.6859987031840536,0.643132873294446,0.03476016630998602,0.7113250520201104,
 0.7088259686705977,0.4747331789588075,0.7557878486202305,0.37379564033851775,
 0.8295494187668716,0.8738317871344901,0.813103603055621,0.9355855529513364,
 0.8760711674795278,0.9223422656446787,0.9746163950108963,0.4369292397424859,
 0.5464398020849965,0.6326057771414755,0.16779104859044747,0.6360897888064625,
 0.07303718664985293,0.3316623373785774,0.7598229823892315,
 0.08647233768323004,0.44851556205846077,0.32151282743805787,
 0.5251216701079255,0.6291232178147176,0.8605788087973963,0.38391746876244204,
 0.5010745877875109,0.19191926283110905,0.846061915906016,0.5798422099365751,
 0.026115448417022846,0.8888088977742394,0.9814024605631219,
 0.22927571523396206,0.5329535982942104,0.4985200946911148,0.7005983406434174,
 0.5968184589908634,0.724972443508243,0.3200049733064789,0.33628437702255887,
 0.5593954122107629,0.12833424522280368,0.0754700964074767,
 0.33270107004168314,0.30010148011325843,0.9425147464024095,
 0.8545714659254132,0.90117287297032,0.8562850292606405,0.8309038455660662,
 0.9294149601816643,0.04318091059750084,0.4121931118313972,0.4048015807670643,
 0.2206578959879929,0.5392246198670877,0.9149150010620725,0.8011245253683055,
 0.18832197140672224,0.531466780688133,0.9597081504277515,0.610331650703137,
 0.8660353081835778,0.7377937078420325,0.655140760206405,0.91904987498077,
 0.12003276308554811,0.43024365257964337,0.04478018443216736,
 0.8674008484620136,0.6338063925986989,0.8649896255919377,0.8508679974764994,
 0.49820162842201865,0.40850135636129636,0.4810340020115751,
 0.4331736123081059,0.8999998095801444,0.8380600076371708,0.7292738968720002,
 0.4281494908888387,0.4570612059559307,0.7092098988747781,0.8173073290973607,
 0.3913187260130808,0.9108983273524611,0.23635657016678002,
 0.08580951454277042,0.036463236922948616,0.7900005922264098,
 0.16331769030786025,0.1435164820667142,0.030939205527094393,
 0.9022321582960529,0.11567074199167249,0.20894478534787264,
 0.18403887893327342,0.5794537298472129,0.9597479393912228,
 0.03235618138601504,0.47082350478843416,0.5424611740370093,
 0.22476492506749857,0.3581214682008459,0.27191514180634746,
 0.20913735253116617,0.8437824999276233,0.03135862291437308,
 0.7589768982901587,0.12385285117617495,0.18177418922762645,
 0.5279787107805957,0.15752001433766782,0.6707406598470727,0.833874529845827,
 0.9745854408527842,0.546465796858901,0.9412071418474264,0.9287966206349743,
 0.2653991141770077,0.14258598278806967,0.7400920729142204,0.855337586951842,
 0.9153900197983084,0.43698165865991556,0.9802205268315674,0.70412850085706,
 0.7758087349225142,0.9553176207603504,0.15299347519355622,0.6702261438945117,
 0.027918416743785035,0.5702333456448129,0.012148979239013924,
 0.25306013451237097,0.9875237065185637,0.48632465003729763,
 0.5876883025551557,0.9242910491088229,0.9047226523806815,0.8808806719097563,
 0.041531895805345664,0.6607604041204497,0.4968481543822961,
 0.6075779779158236,0.07821775905734552,0.5232540158745673,
 0.24220622255578328,0.9911652287536824,0.3305168446974891,
 0.08788577366700634,0.578559055278637,0.8992994237355921,0.5140914256011749,
 0.18316113521786237,0.2177594959017094,0.34529999102582,0.6780304422309323,
 0.06298568997310461,0.6421976813400124,0.5494446408287771,0.541054895410473,
 0.0832695296384457,0.688478083279907,0.480330024637333,0.09339265718125778,
 0.8762220496672669,0.7312048116121154,0.4587613098702853,0.7529618404058508,
 0.1809240223820333,0.2710419888278288,0.36349984472958186,0.763700022492004,
 0.10480988721172613,0.1055364788674612,0.36764604362555164,
 0.6162122162323387,0.6698070566355884,0.10716259021783492,0.6994867101371249,
 0.5322651059886779,0.9836218242679656,0.31139871227266414,
 0.23011081953581447,0.09105028473726029,0.8885021515672628,0.935078403857736,
 0.9101332910504303,0.4065997129911455,0.4742151405796746,0.8144064476084251,
 0.11415835435310062,0.47172983734758256,0.5143684543269078,
 0.1881264478084641,0.7404250812230199,0.5846634416860993,0.47411421345485805,
 0.05204822367663864,0.34373799137713223,0.46433846353838404,
 0.2887297643747708,0.5661826251738847,0.7823973965453365,0.5877004423581117,
 0.7325291611307136,0.013381643788550002,0.017954933968661657,
 0.752620986604401,0.32570067366879474,0.2752340002814049,0.09746774024685179,
 0.9936855524929304,0.6436772367910703,0.42088633359429517,0.9324913077923982,
 0.4484932155267828,0.2187150669109772,0.3543278455423057,0.8368047217827201,
 0.6595174381803357,0.7798931961313004,0.2091969117810235,0.333420624789048,
 0.20274088227535092,0.09930811462153588,0.4734446140502562,
 0.21289030653613972,0.3918281353496854,0.8266958849666755,0.8887351758531,
 0.8342087474083233,0.9912848239534823,0.13409982250971086,0.040596103541924,
 0.4130373505276076,0.527842527761915,0.7683122440912363,0.7789412835646955,
 0.45260855290394697,0.09529859871729063,0.26209513146301,0.5575326174557228,
 0.4923620048844132,0.703820346244488,0.0973236396562912,0.06671266495416206,
 0.0865000755344002,0.6613068510014383,0.704661474269073,0.8688918367657671,
 0.3407848888685734,0.733363147038868,0.18399979285183654,0.4568849542882262,
 0.4172530719164729,0.33334279694457825,0.5447052069763988,
 0.14027623737160377,0.40566645394346057,0.5711738940457417,
 0.9016366378651238,0.3639815465485522,0.12920829105246534,0.5491142988582902,
 0.5550702891071403,0.9251583727504484,0.8175850889234,0.40072896742528175,
 0.9222930824894024,0.738082037665435,0.16672255176055328,
 0.041626387648310725,0.6337243184597046,0.673017482860738,0.2052257721898657,
 0.2146715719689697,0.458228425339434,0.40323466267268193,0.8119945667970538,
 0.15384893630388974,0.9572056010305671,0.816661929445654,0.39625114868248545,
 0.5978192762466096,0.3946095961629753,0.691842541346149,0.21576815402494132,
 0.8169871884336541,0.61664022300807,0.8965789765491898,0.17538065018832627,
 0.7470816207423557,0.4656522330966739,0.6785334757047043,0.34367515662582915,
 0.81953937176551,0.37408978152369454,0.758211624964288,0.09631048834760336,
 0.3428176012999896,0.3665381594646372,0.16828113344044393,0.8070493712103214,
 0.4722076913786504,0.05113019086364612,0.046415048706427764,
 0.6867248759202637,0.05386966714333963,0.8912245092370561,0.6631697114689901,
 0.863094610360493,0.6062058548532561,0.3071709310379007,0.030986370669264907,
 0.17194813476851722,0.7294465221492135,0.7833118207023159,0.5366848935884033,
 0.10554743598538252,0.16968662395810807,0.44608189663992426,
 0.4077795712507656,0.09343799829297761,0.178299285227705,0.2963415933186886,
 0.7826515859463737,0.572051025808967,0.3442886674167429,0.8594558037685491,
 0.9596705726984173,0.8365579295594028,0.28179568532958277,0.7718991642846976,
 0.7855435921231864,0.21382145097372285,0.46105140808928813,
 0.46650328398518837,0.09686399267174395,0.8307551967649671,
 0.4171424102623198,0.668229355186764,0.5228857250826002,0.6857894271407132,
 0.8580239191501804,0.8663525353183186,0.2679906955412038,0.516986500147158,
 0.6525490036255419,0.8259329521126525,0.9991093848415867,0.6580979110946519,
 0.40216634558842346,0.29325218191754854,0.9624975027165885,
 0.6430684310377516,0.462145742595249,0.46993161448396137,0.44267379737534773,
 0.5160670122901234,0.5870118218624206,0.4830922928432122,
 0.039800276679639035,0.9652469392268093,0.06259521799392709,
 0.3276394353475478,0.877107601980414,0.5279311802276424,0.9191957488690354,
 0.8714828304333079,0.04706151578031159,0.7821486010392836,0.877503890000582,
 0.41490901264075086,0.4155075760363627,0.8247186840591344,
 0.32647912355091024,0.054772586333656115,0.09803802198463651,
 0.9449691765300816,0.6582003629451526,0.821405712862273,0.5547807093691712,
 0.6545440013970806,0.5879579189188178,0.25533633727075544,0.9349889874149595,
 0.4336331875671213,0.6606147639697474,0.9343166304217521,0.4894489064469578,
 0.013120987774821402,0.3935075834437849,0.49777106043550023,
 0.966847314614573,0.4381169630734716,0.7262236748366434,0.34948578180663215,
 0.6325284864215683,0.07944939381661165,0.19339187865638807,
 0.8445223713807777,0.7202786274595053,0.08881781666097177,
 0.20392125813672524,0.5255893817429311,0.42377874516644665,
 0.6029210760151412,0.1202670694286384,0.013312127609501512,
 0.8517721896097802,0.34319140598578857,3.1824679123459987e-4,
 0.7342716883212324,0.7090910933728876,0.1415331737394654,0.6128020069776741,
 0.26705107103019954,0.15048947904075594,0.38459807852847727,
 0.3045580981179129,0.9986446172450661,0.31217422955475116,0.3270311463361285,
 0.8679261568461489,0.9606962661131959,0.20880971010934202,
 0.20582548930042144,0.5918389131611373,0.32933485214516534,
 0.10975558740577507,0.9186374907551706,0.10490094102244729,
 0.40834891058434375,0.863853260120428,0.19737166728023037,0.5202356668723401,
 0.6544773218127736,0.8411109188507955,0.806172204674245,0.9768616540888002,
 0.9280541931681066,0.05061115729241983,0.343140724975166,
 0.0022484246653390327,0.3542685191593501,0.9054932341713324,
 0.06461567265854196,0.9490066076586867,0.5265769297982378,
 0.010239759599951714,0.9822040657238913,0.3426520560097477,
 0.9629712616658374,0.6097630122581674,0.9615061767879399,0.8277284474264786,
 0.23101548445014375,0.7921001524068699,0.7370630215674998,0.5146015171414573,
 0.5008351814122721,0.6575628531769213,0.3403876575344883,0.8624185315905217,
 0.45956081972670926,0.7871088207375477,0.644508546795658,0.6569584309773052,
 0.8213552826258474,0.31336874486914335,0.06002590385678275,
 0.5338884441520145,0.3051664264869406,0.7578834159287837,0.6785887055078191,
 0.6540484042420177,0.19677173656923286,0.6021340487552378,0.5576815736123469,
 0.0073502112016856636,0.3362056045654067,0.2762153197520939,
 0.03845125155432094,0.18827575134481778,0.34712499252075935,
 0.3946227295676451,0.90063829517818,0.4959737210166818,0.22703213005116152,
 0.3653058103360787,0.6402366560634699,0.9105358015983883,0.9326371189813691,
 0.04573284061715366,0.7447537340641691,0.37122168878976325,
 0.9121385955374194,0.13362382572162135,0.6885014688964236,
 0.11597590854091244,0.9114037303791266,0.0174133606750273,0.3624225257992775,
 0.09161967448234132,0.1940006892683157,0.6176411814655842,0.131547779636206,
 0.800553262575551,0.4752945099377225,0.27549226611684263,0.9314491625091087,
 0.8274438973378995,0.7145872427115171,0.8487283559564325,0.4400479543440199,
 0.29262564381077605,0.36955353405175106,0.09055704703194878,
 0.6552947232070981,0.7653900642286873,0.6727211910782019,0.49996854827301895,
 0.8700713200017595,0.006047197396132908,0.5977623883573344,
 0.9902251406365742,0.30550858666770797,0.6351145533532341,0.7937945468215379,
 0.49109954311068205,0.5235223356647551,0.5978013569201696,0.9018023972878515,
 0.2001312441555123,0.09201328986523594,0.38621715345847185,
 0.5783493750758433,0.8563724616287675,0.383765444677985,0.73737669566771,
 0.25971253435778374,0.847992759215364,0.031688758204493905,
 0.2674067656838307,0.08261551235502385,0.7014633424154051,0.8775911486141714,
 0.08271110166201956,0.5122253390204747,0.9286667596136953,
 0.03174799061933231,0.5830528880440464,0.22710438706578362,
 0.12419312026978402,0.017868167069632124,0.6563516729774815,
 0.407949253352361,0.35310687363410276,0.15035972711460976,
 0.12080287625476671,0.9969722279234543,0.4769594337673706,0.8001956726644849,
 0.7920040611686265,0.23819215300346386,0.5083463687094922,0.2909312176394532,
 0.7111921607568193,0.8273958270287418,0.31732389735881295,0.4943516827167047,
 0.6690532933111014,0.5347064882508005,0.24288003602047825,
 0.36603173351861973,0.6821021493692642,0.15860392646458898,
 0.6974820765563102,0.5623332022973972,0.21108107546171606,
 0.03298374297795892,0.7253008264862512,0.40761986733586897,
 0.017152334670373803,0.3549359992572212,0.2692044958624543,
 0.5802474369102875,0.1311243803155704,0.20988364801768933,0.366662106328534,
 0.7520387913006753,0.8946344089038615,0.5380173794776121,
 0.035488468887507674,0.17987672563310309,0.6447744873240701,
 0.2201743873873886,0.007178545011660931,0.640997406398687,0.9512215005943664,
 0.29007168636110525,0.6791569161605373,0.03147417741349967,
 0.44553959883718086,0.1081565860279291,0.01873929625703541,
 0.02399088555449569,0.46508304925423927,0.739895001010584,
 0.043524006852139197,0.7548028924145412,0.9032271253694049,
 0.24245247778018553,0.9343181287860733,0.26531226335885616,
 0.7098812239077663,0.615225922496934,0.100094724697262,0.5667327911444495,
 0.8126159275488153,0.06432900025989363,0.8011625126593677,
 0.30619954416105244,0.5193199328728152,0.3242751945625565,0.9719680165906646,
 0.19546915212523963,0.2822496795072823,0.15746111378092698,
 0.47821721772658765,0.9706705134549078,0.9189651111217932,0.3939349017241731,
 0.46292762217535854,0.19042187054312476,0.32318868264402323,
 0.4975950451236766,0.5865205247057179,0.3791773016945523,0.9954206825498471,
 0.07035207939053445,0.0020577641020633752,0.08585955665481504,
 0.7495002474956469,0.6926927764061542,0.9887182339322043,0.20146775442383036,
 0.6549271362256001,0.6675810148145134,0.08012511240551723,
 0.26597397230844333,0.6073555896648093,0.8279882845991818,
 0.17723421222634494,0.07654525881828333,0.40034096955623144,
 0.8133540908478949,0.11665732389452999,0.4555622951277919,0.6422109260698279,
 0.3532668405831615,0.26168767985337726,0.6974010981513653,0.7336682381323684,
 0.3435510078554578,0.8677327780581198,0.07483881789312608,
 0.09444359225713417,0.5592038551078066,0.15395721986173783,
 0.5638760234958512,0.6508696880382494,0.19188514368299958,0.4515423704976713,
 0.9764151355211943,0.10600774554879366,0.22386863342724694,
 0.7835992117008557,0.8619710598248508,0.2336671459673073,0.16512676514354063,
 0.44921434766315915,0.2103763311439596,0.7738769637650506,0.7627326248363525,
 0.025359319390347745,0.686112506965677,0.8855453101585234,0.7913450831974527,
 0.7770800149332716,0.705833447828426,0.8658440239606626,0.582881280657427,
 0.4422444330624502,0.5099916672484911,0.8991190094578337,
 0.052545754729613536,0.3274758679380714,0.969742522213271,0.6378293635186945,
 0.4976369279012901,0.49169907019327075,0.181589654205786,0.6513486348819907,
 0.09943524149758354,0.5218852073885163,0.9037790041615232,0.5662327139732212,
 0.7742235886634842,0.9112894428358315,0.4694735677406121,0.6220277066170343,
 0.4932690070217036,0.19850995734273857,0.0773800189978553,
 0.12480614950239388,0.20891362773119282,0.09225077198911324,
 0.1296935388590612,0.7292041363665813,0.3077853745926389,0.42587783551127023,
 0.4047315262731126,0.5326455740131264,0.7962033355219773,0.5366629150691412,
 0.8328795280138027,0.6508623179688081,0.7886379880748969,0.7839669141347699,
 0.8409734234695161,0.4812927412955146,0.4300915346029176,0.02446283115800818,
 0.8254315985656769,0.43690408787417057,0.9664489718069065,0.2222661262065514,
 0.10448512756335504,0.4550194086395566,0.5230084227110969,0.5151085280646714,
 0.8237880869305628,0.7841012571497512,0.4241761841142243,0.7625490817225591,
 0.3194020611103209,0.6069655779622691,0.6291500054748747,0.1945898104158511,
 0.7451030118121127,0.8215361538627345,0.5795112587878452,0.23939353981877953,
 0.10135960273961031,0.12216995601247016,0.1558810765581249,
 0.6693347397793049,0.00115678735659841,0.8551499338823371,0.6059581713743026,
 0.1729320762881379,0.776543424438384,0.6356598254996776,0.8460754611462843,
 0.30523119344313754,0.6672772151188076,0.9213283968286934,
 0.22639716021665834,0.7635271831352317,0.34499818554249206,
 0.9366126263013649,0.8792657166419334,0.7986013818409065,0.2704316718673683,
 0.0986432613843159,0.0345572112303848,0.636768697332605,0.5283672806865699,
 0.48985876137745965,0.8691348323764431,0.37872100320742563,
 0.6616491397176021,0.7850971458602315,0.46979214395604674,
 0.11776865749938847,0.5453972638556991,0.4079166661187026,0.2337269253641865,
 0.2954653949475168,0.962832310302099,0.5144172755466847,0.9256186833866056,
 0.7889653234265582,0.8232732244403163,0.3239438599915143,0.7818708933290437,
 0.013406005513374453,0.4765103006445909,0.9344242043360144,
 0.28395788540890354,0.3231152342495244,0.452293524963852,0.15255567066310782,
 0.36039985312630973,0.8009981767979082,0.5581438367281145,0.6862414880332732,
 0.9462549010626002,0.3467422356083367,0.4392726818544753,0.5195229267459444,
 0.7318012330452855,0.1811295642605768,0.21737763227388074,
 0.03054688997015742,0.5431902803421005,0.8049342051593021,
 0.25656524211624987,0.43712565599010356,0.5008617198659286,
 0.8275853122443139,0.009443006361613215,0.8894025542833377,
 0.5573159786053961,0.9934450917547986,0.012010900674600089,
 0.7049436618812013,0.7858834949115693,0.9874241064060111,0.5067580099764919,
 0.6675262704124797,0.37980871552572015,0.9188979271732902,0.8726467622447904,
 0.44069846979986815,0.7296227675504421,0.7376994249121832,0.667021155907938,
 0.5497574509034778,0.6112918787550621,0.19764775519862388,
 0.07953251072212453,0.18637519930957125,0.37146598998622216,
 0.7370246489440244,0.327899016873646,0.8968828249530283,0.9142492230875626,
 0.06655120076507293,0.7521501735102886,0.40573609281166134,
 0.5372555193521512,0.4177018497958733,0.6420587342850796,
 0.0011720059597889465,0.7612714458856605,0.03476217719608943,
 0.42424846143197636,0.571836410539152,0.40984291193234024,0.793045040758555,
 0.4455382551486735,0.4199382031087221,0.5770578637057187,0.20474584102803828,
 0.7083339775624033,0.9670823770364192,0.6594859895353331,0.30448183992751954,
 0.129368668593984,0.6554086751835211,0.5167932085988625,0.9392323422743162,
 0.34820703232206196,0.2520371080381989,0.4508893307502069,
 0.17158445926719812,0.7682291950994089,0.8918672676583618,0.5327266550964306,
 0.8874199614156478,0.23315160165674886,0.7125422323071839,0.2685595603452955,
 0.28578628985903376,0.2412249717479823,0.8420039230024662,0.5480523879856959,
 0.4770071303972665,0.7960067245808864,0.18159038868251276,0.5036370320788771,
 0.7342065393975062,0.8245913703505847,0.577797574064949,0.20865359817105855,
 0.3790744916251584,0.10041292388956446,0.10411563077684305,
 0.08335253042168833,0.5682968276813234,0.10889361567870337,
 0.10944466318538093,0.9480625778096553,0.09318102308909337,
 0.6694470491036485,0.37267619196985724,0.3911616227354484,0.4040103913271491,
 0.8742440533838374,0.4120269117412977,0.27445231548318794,0.4864843802280622,
 0.15820726201158364,0.04442026718328418,0.18539369748942114,
 0.5453694577459857,0.2648999525178144,0.3372605333927625,0.1896175489026959,
 0.8934983065215384,0.9703908621812292,0.8083287979631332,0.97657684246217,
 0.02678016552107787,0.6464969326339238,0.12529496035759236,
 0.17020699310464482,0.6759754915489639,0.9143635558365029,0.5043269576739985,
 0.8513481615224365,0.3801923474714154,0.5881884018393874,0.343898251462182,
 0.02056720973712589,0.33621006708431367,0.3624300231473372,
 0.6758657337547472,0.8813706172451509,0.1264974817960025,0.6006756620203679,
 0.49107851615200393,0.10630325089689197,0.6488729423774552,
 0.9572172063605714,0.2325283886871199,0.5114908735707339,0.26153308884522486,
 0.7348694831819222,0.09360714014941818,0.4655368895397778,0.5464935125827273,
 0.1477676787760528,0.057955952809539024,0.4479736202729595,
 0.8981032413598887,0.31802367958354194,0.4746567155886122,0.3607760316843063,
 0.1128154045586327,0.48457004805913606,0.4769572448375732,0.9611265013506195,
 0.8679258246497148,0.8168783006000938,0.3129250061692902,0.9884007944842956,
 0.10407352283255156,0.33410230839050425,0.12561603203700633,
 0.605204038322674,0.7331304836135014,0.03425497315876813,0.1120693374237357,
 0.4767087847064375,0.8139807661538696,0.755095199074696,0.5381668960368201,
 0.6435517332138544,0.22692093178577322,0.7223371438667628,0.7345789724087362,
 0.3416151830410039,0.1527325134961427,0.07114519424735766,0.5028378703738533,
 0.1459891769003302,0.44607514105745816,0.15257639025256775,0.880098496981389,
 0.4361541712879582,0.8367074478549851,0.3110286771786035,0.9001588923424453,
 0.2527593984027242,0.692821034195032,0.5248278839628115,0.32686219512194004,
 0.46622306980931416,0.848673913347657,0.582181577182585,0.501560621377898,
 0.7342463349788101,0.7115676482129235,0.8602964407717367,0.6167702406779496,
 0.2045126779017694,0.8242483522131622,0.7873348946702363,0.7584533449383724,
 0.09012386714543763,0.5019164975982677,0.854175776837471,0.6159401482685838,
 0.4850080989971888,0.6660030873991593,0.9387880182031161,0.5607523094280376,
 0.772774014920624,0.2997169119412608,0.3751752448395874,0.44047902970002084,
 0.2695572383079591,0.18850056666323833,0.2743650740159401,0.5679018591767869,
 0.6025108552377106,0.36486980704088134,0.3000465812292502,0.4596158136993027,
 0.3959190322939976,0.9851524430602961,0.9541079595687083,0.8732538568308409,
 0.5498369279311557,0.12819412179537037,0.2403199487255009,0.3839456825282731,
 0.7308200251248357,0.06988140264024756,0.7595135335271433,
 0.19692310964928506,0.06878870853965513,0.3289184911851021,
 0.09169985270515268,0.5643360995938316,0.08648234352701312,
 0.7950077349069271,0.40167907754561116,0.7643210721218899,0.73444333756554,
 0.04251155308845467,0.20125069856297917,0.8952774466844838,
 0.9842788328602262,0.5594625135383564,0.7827413912220513,0.903102319114232,
 0.7856717315616651,0.6609874443586348,0.08979655892410043,0.4187720914312467,
 0.14199400581532584,0.052683830424627986,0.6184725649757566,
 0.058215969877673523,0.7700393911007763,0.8304384783868984,
 0.5685616902299468,0.13236827066900303,0.1837139440867862,0.788315466807603,
 0.6712222594215695,0.8769461096936266,0.8401986960109034,0.761931560337653,
 0.722710889411634,0.7890527122740181,0.1696747921797721,0.5513916057041445,
 0.14931281887880776,0.6419069981255192,0.0863597384552457,0.6951477306837525,
 0.7299150303351686,0.7707969483975813,0.00481972407417941,0.6098890667873107,
 0.3268460881151083,0.5220828446655669,0.7889654511056056,
 0.042685351724479514,0.8012000304008039,0.9887979310953402,
 0.3766391617161675,0.2679268676863483,0.8860268679466572,0.04163028747018027,
 0.297055726065776,0.7697031779274036,0.732244939193037,0.2072904493997958,
 0.3731315805897475,0.04017646354227178,0.8793848003843782,0.2302670920152904,
 0.38238380021549034,0.7505163468693542,0.18711275087671198,
 0.7422131873892267,0.5361864518627278,0.671250445584968,0.08236922119879209,
 0.281621392014871,0.6431990944814343,0.7548747132317009,0.28822482218778445,
 0.136380989645717,0.4336407174414151,0.5316765882684255,0.03751602768847062,
 0.7735849410959551,0.12709249218484986,0.7265557175748791,0.6796162760892006,
 0.6088007268261393,0.7202526188389993,0.2373105156304629,0.07679016334275046,
 0.3494595188267928,0.8725287528573947,0.43465830871646327,0.549690509075224,
 0.9216038676395095,0.5308218320375063,0.6248500106500783,0.07630338842349016,
 0.05036788722230445,0.6820069203048851,0.7723743148799868,0.9169486243055196,
 0.28354397365574413,0.3275013478454285,0.415898437719714,
 0.019726584152397586,0.16875211061988782,0.4654917551699852,
 0.4776070781692002,0.9777746745624686,0.24110052074933774,0.9461097498042799,
 0.6704892839100818,0.17409779695650407,0.9158414125829074,0.8445573699227024,
 0.4125922995496244,0.193498837793886,0.05627460137007145,0.77330107624666,
 0.38208064556135723,0.7716306715637987,0.7618673237477491,0.6606222583629201,
 0.1623127929602154,0.14923186080193018,0.4265773474696436,
 0.10733204649859762,0.7536750732137599,0.6381449704179696,0.7921274248682801,
 0.2145193015215795,0.24037695450697527,0.4861704266368305,0.9321449585235316,
 0.2577163636266282,0.5677213943275967,0.40766070496503237,0.3311863091437829,
 0.9024192266155057,0.20026279269510228,0.7624841634053955,0.5343835849304218,
 0.130665641430474,0.29501081807535745,0.36143321811894713,
 0.29455644949402915,0.6706514171136789,0.8350879740186076,0.0455815740635972,
 0.24132613455476792,0.7209625592859848,0.606051477064138,0.8151228087119614,
 0.6696344395745877,0.5754723486733614,0.13193455397537868,0.3191992974105514,
 0.12827565679988895,0.7273885773264392,0.20582941185290837,
 0.06963680949420903,0.5321560445029998,0.45486711710152217,
 0.3886683813904198,0.9220593230795865,0.6194296912424302,0.3285972700059596,
 0.2991876507576474,0.32934013048364785,0.990099459872661,0.08432148094003211,
 0.5996544996656544,0.041455985817911434,0.633463327097273,
 0.37733470296600324,0.39267723953714595,0.7466016374174556,
 0.4195298939258484,0.4512858778064074,0.1447108718672503,0.30005061089148666,
 0.0109203316176395,0.9580455785527635,0.7447324118834444,0.7458853486801335,
 0.20472738970010762,0.06256599300676335,0.4327037925592363,
 0.8539035244922333,0.8685473551189594,0.526811979572677,0.40204874566972837,
 0.8273431162211067,0.2818291726029529,0.5370441691583692,0.5793111692536465,
 0.4635850981455929,0.9641600773144189,0.4930312107617695,0.6233182138266142,
 0.9472235580909276,0.2795730976572054,0.8774756976760174,0.5969475121936196,
 0.4650924555162288,0.45050682920212903,0.17312144183977973,
 0.28933536581314256,0.24984710638170382,0.4179655163438192,
 0.9611475267507165,0.9796767453384758,0.9096665545789111,0.19882892955043852,
 0.3018877108085546,0.9870443989239852,0.61426559653458,0.41621655721878015,
 0.5984530335506745,0.17199750785697143,0.00284497040695042,
 0.8563514480102404,0.5661108374829736,0.8992391762667675,0.8095567354059015,
 0.738050496750097,0.05197944302957391,0.6080822550599635,0.9594572370000217,
 0.2403411818180754,0.40461247920003884,0.23617781120555437,
 0.9350255746833985,0.7010481322985114,0.6978143074106576,0.7431768720753669,
 0.8691202239627576,0.9038725923667491,0.005901403711178088,
 0.7671697817079401,0.4313941646924968,0.8526670304867184,0.4693789531818673,
 0.96300102662723,0.1460236364893549,0.07767275184453548,0.8602227310875609,
 0.3761229789292745,0.9438238811875617,0.6210855013025023,0.20972924046159847,
 0.4874931720441438,0.9087724620676462,0.28489597998239935,0.9083136306367845,
 0.04600616332436247,0.49008636470922906,0.685678130866066,0.4624687311290451,
 0.9216713021662388,0.6644885890707388,0.23011546343827516,0.5985404262303843,
 0.45475349187568126,0.02922497363632992,0.41325382938893485,
 0.9871066844572579,0.21166322241152213,0.23537579474149029,
 0.01648010146989254,0.316870824936491,0.4471829437207919,0.12452564075511274,
 0.6846539419669749,0.6004843853569756,0.978720192273671,0.8855659384471969,
 0.9646248536010379,0.47277200614081005,0.7246356326518284,0.9247388886767182,
 0.04585425926684705,0.18187251525525172,0.14016153533097286,
 0.42974646501941594,0.031709726247740044,0.3088417930201037,
 0.7823281374651964,0.903137814269353,0.21332966374103013,
 0.015598669746879623,0.9044796083363306,0.9166565238915136,
 0.3607562196543612,0.7958071858878911,0.04678100268035723,
 0.40547974292715916,0.9712191723239922,0.6509436027796145,0.978528360237175,
 0.7044932066575551,0.7325929561516161,0.09357430883050688,0.7515611801028417,
 0.026488923644617635,0.32753903406677654,0.4772067359237627,
 0.3146857550613308,0.13490109630969005,0.31778990361091086,
 0.7320614153644819,0.03935757342745827,0.5383959359254106,0.506080425433978,
 0.9416159508084001,0.8254080686126568,0.46332936379108414,0.8218955221500963,
 0.2682661630223545,0.47780397408555786,0.08706446801147427,
 0.5727690772786056,0.2640923800176186,0.3218108119441936,0.02641577516634419,
 0.1326621145886121,0.22924639850255568,0.6706406996188713,0.2117797650234864,
 0.25822801465766143,0.14500741350376856,0.9347508721014985,
 0.20289912789542097,0.9890137323326915,0.8145404643984472,
 0.47717195443561655,0.8937313730809352,0.3859510084974378,0.6192797848155487,
 0.8792381783782276,0.5173180289335798,0.6771104241843233,0.6793200924319206,
 0.46881352312127733,0.815137606914887,0.8036323501830662,0.4625618074564224,
 0.5470251081775083,0.42120062082911924,0.2938260156874647,
 0.41163992356325996,0.32422466754506396,0.0889633281017167,
 0.5521936268173158,0.700330533600335,0.6499889741689152,0.30680087038912807,
 0.973473414157229,0.6486048813637915,0.9679703540879793,0.7060212379315494,
 0.018982298722641455,0.9295788679333333,0.5287493116732638,
 0.1934956335602429,0.2928596843963288,0.8320276471792875,0.27377994567467645,
 0.7833297875707654,0.7432859423533423,0.5974312579980807,0.7422538552300245,
 0.5732462558507151,0.30820819609247085,0.20976407120922547,
 0.6288287724658387,0.3760491114301382,0.7631982193868568,0.7064688630306631,
 0.9069585789982514,0.7645417934527902,0.7324127790468502,0.170872169395062,
 0.5644954834491335,0.6531075038199949,0.04994545276839368,
 0.23149180513117895,0.9953991300351477,0.6759125060409495,0.4003815266908489,
 0.8572555614666034,0.019583310504258766,0.3386863411343226,
 0.16126034127365207,0.49877169811304956,0.8401151082786981,
 0.42099444068301417,0.27958871244369776,0.2013149315205941,
 0.5501171280766955,0.6043494436309529,0.5414873582162165,0.7559428781194999,
 0.9846049450054644,0.6835707578075567,0.30858371714255417,0.4289683613055937,
 0.5818826415872792,0.2684335288488282,0.08008423753912486,
 0.18345373689370126,0.520554610472328,0.6326119573255234,0.01865950772185221,
 0.0016065380489029835,0.599165830790308,0.06187172857322143,
 0.7353260337499727,0.25887001316044667,0.4612120049925218,
 0.13228577885910175,0.048963534941384945,0.5789517308685904,
 0.8465187210729377,0.17195196815777003,0.815062560740044,0.0373699585140187,
 0.9580644380124528,0.8845810624232877,0.8541288743410004,0.7670370116657874,
 0.34312769238908236,0.839454986641816,0.03608977321959439,0.2993084905129699,
 0.6403260844842614,0.6515022243991586,0.021379493621358048,
 0.12341975740881195,0.9178054930241991,0.06047953653570537,
 0.28894300681777974,0.3582819772219281,0.030131661803727616,
 0.6110593339253397,0.9910915277372668,0.8660931335003528,0.8159638481480814,
 0.7439312521564662,0.31740027474975696,0.8686251134563323,0.8902251979569693,
 0.08667303904657286,0.3838238725506874,0.012215126340599514,
 0.7182142870739329,0.857280478109493,0.8748608651952993,0.34359274877594836,
 0.620608449522333,0.5437529828251163,0.7148782914867162,0.5693563888717496,
 0.016122419942494304,0.8500788454280348,0.4835843570213274,
 0.29403097621841134,0.29619578801395907,0.4903754904961045,
 0.8120362105604139,0.10826766526361653,0.5287625214563485,
 0.20313686575672973,0.8029173073501228,0.2011714771953278,0.7149407414479583,
 0.10748387259570924,0.22864094461893603,0.16842329714577375,
 0.6127584072224725,0.9212729769354668,0.8601988859072494,0.19674379315477641,
 0.2745098275993,0.7216745573788751,0.12040111896733108,0.29430930484355566,
 0.8215067868695509,0.5406176933694604,0.05486005006012318,0.9243003908046669,
 0.42489148714481384,0.7779100173187848,0.4119168571896855,0.6635047411216451,
 0.17845848388753038,0.453771519990943,0.014007781229651517,
 0.17365409680454913,0.5099927183288409,0.9601322910975718,
 0.10664523610489995,0.42207816366604056,0.4347432733679185,
 0.9100675316344322,0.5353861200725937,0.3225830835622494,0.26899294200338697,
 0.11453649251017695,0.212144250764164,0.8359122643668269,0.10728474336664495,
 0.4952573768802906,0.6412469244223714,0.7730230164012828,0.7627334341673411,
 0.8657161379527718,0.2789651801063352,0.25954065200164156,0.6686985119909273,
 0.934356261791354,0.7093675294709485,0.9775240529003665,0.967000292889595,
 0.5603598002160989,0.5285139255641718,0.9639806709546137,0.8269302112141929,
 0.8557153124055547,0.05700233278463274,0.239950705415787,0.4435641141067401,
 0.2475438671901412,0.9916521074706933,0.6486000894730491,0.6752362592414337,
 0.10690583926444808,0.6901312905302512,0.5992729080174202,0.5333967595760685,
 0.36525091583039226,0.4759159787795523,0.8243241984552636,0.5968608868269901,
 0.7979203532605663,0.47796979067029954,0.062099302167642034,
 0.2681471083948185,0.31847386054346505,0.4151497818974421,0.5715271553119338,
 0.6403967942158153,0.447633354272291,0.08452803385999363,0.55675217288696,
 0.7812119033825528,0.7571705231714435,0.6187704567655956,0.9020080437425104,
 0.5316490761009158,0.41431074935070344,0.508503579557914,0.3140100836443066,
 0.6375556221358196,0.3402307024463759,0.1980300708976882,0.9794680258725689,
 0.9286795246417814,0.019902723491193708,0.9603041607691783,
 0.5720806360269227,0.29183647834126625,0.10984884272633844,
 0.5583989267055545,0.783698506448712,0.039338883403378655,0.1988752107444176,
 0.2592705106604378,0.21904931522996507,0.3422669971098925,0.6044444309597088,
 0.5726398378505488,0.3674230817427635,0.29051489720809376,0.518182792989409,
 0.3433854077534483,0.49101902565025246,0.9599272767292422,0.92190563527918,
 0.7681618928409077,0.5828444874374433,0.22407052156059704,0.2595523243399974,
 1.3786733236675897e-4,0.22131796184137786,0.4067795058532703,
 0.2089549014696862,0.09314276572684255,0.2560773775147053,0.9319313233132077,
 0.03823135214480233,0.562843666797059,0.4836019876048129,0.19118554544191702,
 0.8967765361211322,0.6926093310959509,0.9245503225705839,0.6417996220120722,
 0.35620633797417844,0.11830624862663286,0.5089332224573742,0.693989454258651,
 0.8586728307507029,0.6766779995303185,0.2937210453939547,0.9774568215101405,
 0.7813037700966387,0.07382230178608962,0.3977899466061693,0.430283882591841,
 0.7282794269862296,0.4988199073148972,0.602717646366662,0.6814526419126217,
 0.1976448880379016,0.8537282483250233,0.9873701245093203,0.12967517396656014,
 0.0016734156146436252,0.4960782830097761,0.9246972138434952,
 0.4888017305429484,0.12926903508269383,0.788906319578297,0.780346258311341,
 0.39603446516888674,0.6304121407471639,0.2415498002938049,0.1602471600722779,
 0.4554190038628162,0.7220763524466103,0.7705392290033062,0.14308107311912666,
 0.7523468036992904,0.24832315960930829,0.46229810046916475,
 0.3487197447657381,0.743941752496133,0.8098402035872367,0.5402540748945673,
 0.07538437041508794,0.6648882441047241,0.7351916520165265,0.9830756617346904,
 0.5329774181151631,0.62979437878485,0.21059417442150663,0.7016840122871264,
 0.16652402027754865,0.4097780250926233,0.371274906628413,0.5057520468428562,
 0.5488098089675207,0.9676697960932735,0.7423968762145795,0.2961802931450509,
 0.9871887530550002,0.35446522773009503,0.7686204856864832,0.9453531951078306,
 0.707476085099064,0.5380998630047265,0.27559523243124673,0.22525313295367355,
 0.10856504340909667,0.07096485422956567,0.45717618094003076,
 0.029249021242115147,0.5672474939624399,0.10628032184609604,
 0.5498328676449792,0.5022138991786743,0.11764708444661254,0.7709609900395845,
 0.08434130067408052,0.04411609807089434,0.9771927483515817,
 0.8431636534839976,0.5309688020089103,0.38975833198295873,
 0.23381237974810887,0.19341171762154408,0.40564935085736054,
 0.3277816204889714,0.34924776005370484,0.398809087834191,0.22980997462153174,
 0.5583455468828614,0.37431170759538446,0.3852127898423606,
 0.40311565962629814,0.43267072950131236,0.7935492684364053,
 0.1899251482217189,0.4656496518384855,0.8334239711475238,0.5153225975442604,
 0.2916054939672419,0.5163606989094217,0.023272795508681,0.4412425780638969,
 0.24617103035406096,0.94609071211724,0.7862473840128361,0.37214833852451235,
 0.6937711055049912,0.3458013139878673,0.4203559130512675,0.7543798625774448,
 0.1958115819062748,0.4887621239144532,0.018681834159910427,
 0.11282924532953587,0.5075459197196295,0.9557522736488256,0.6734989112356619,
 0.8524701602241086,0.8716693592226137,0.46443594338049143,0.3849948570852947,
 0.3292732473424047,0.8625540492931779,0.17116604541601066,
 0.12132445248926094,0.43133414444460694,0.36391771251381344,
 0.17987269815814855,0.7770156225193106,0.13738193003011745,
 0.4414514393022828,0.7354787749192515,0.7009576173296085,0.08110740403958283,
 0.28675983433823193,0.43562605405084853,0.908546114628596,0.2844531891420288,
 0.7154512467695509,0.48782910571691085,0.10855571858673863,
 0.9837749553649686,0.39316636531885374,0.8922715287129617,0.6740924971009384,
 0.5548470398035297,0.796243254061969,0.7665631878757844,0.9445336160204985,
 0.1371321471836734,0.76834337957693,0.9630615152261299,0.44003814454809476,
 0.9515723609161291,0.8492200649220963,0.6555478310127121,0.5828309380559364,
 0.7456549379946049,0.8167136373569254,0.1139549282676533,0.3674905504576915,
 0.46629504873108396,0.20893728766780006,0.8649065978125307,
 0.6168115382817719,0.1784897267347314,0.3515898707520503,0.7131878341509921,
 0.5304913950808583,0.06633337407391982,0.25014079850048376,
 0.7534280671318637,0.7669726696307612,0.747770773710069,0.42960540933855274,
 0.8408977176770956,0.14154870540929287,0.6424303738373784,0.2520125224756401,
 0.18358177459196323,0.6740494871695856,0.15284988530651267,
 0.7660387244436454,0.6951449173010872,0.6423814336856258,0.5982297115903803,
 0.5360081901106162,0.9444178901193719,0.15266159301572213,0.6245238395411933,
 0.20313924411546436,0.8458212244969646,0.4722867443960529,0.8794312942863358,
 0.05648091440798231,0.5428660210008963,0.9797772009904985,0.2609961886782305,
 0.21269543657310586,0.33939213031894033,0.6183538661104526,
 0.09388063777962863,0.47387674368932065,0.07651721897693631,
 0.9104173060931385,0.8872024819970825,0.8417110674187207,0.6004417886776512,
 0.02979038085432717,0.9972967080439546,0.45044394631454554,
 0.09420567703307725,0.6131643878318253,0.9395419949307414,0.2746080060859727,
 0.7705587104381965,0.3371950744680583,0.9824599792498299,0.2688729551927347,
 0.42097394780574915,0.09663060362320552,0.1415358359271668,0.775256319209134,
 0.22288722683734374,0.5527050938124988,0.2626890390051574,0.5001086708104487,
 0.4141519332865937,0.2633437160905939,0.8503246609830737,0.07044883985711325,
 0.046099741717567255,0.8047732508846238,0.4986161150039399,
 0.7762037733167257,0.9807407895857475,0.7348743435151677,0.6201209143556979,
 0.9660150863471777,0.8284220307243195,0.6269178997680552,0.8473954648126495,
 0.8545054436802177,0.035486122759689875,0.8216912971530069,
 0.24458413988671002,0.5763034624529222,0.364500361604743,0.27734975904017556,
 0.12856059088931804,0.5028604117452945,0.35803175754433547,
 0.06893829467778567,0.5875493268094731,0.6449403232549505,0.6752115775667963,
 0.09736860220917087,0.1536770031320196,0.49402671126874,0.3142615222547511,
 0.42040987586229217,0.8692057041495843,0.7021712334749809,
 0.47505853906448336,0.4093350380379276,0.24489976479078956,
 0.4533245504207306,0.040984203530775365,0.23055005070942558,
 0.9692146633543937,0.4968449345720627,0.023277602491684535,
 0.3169668579893118,0.12727299271186565,0.5982693424392248,0.7777097200251247,
 0.5459372906224627,0.009791783402183118,0.4712706973192834,
 0.7694705817007549,0.7590931189016368,0.6428346847922968,0.7473723252171466,
 0.0978132811605843,0.3371063802055173,0.11854910843956323,0.7240747789937146,
 0.7215121999702124,0.25766133441748684,0.7315694343934773,0.7341330978461755,
 0.5932047322029056,0.00364903404811745,0.6840928290943168,0.4969970552635551,
 0.7720063457574002,0.01334862355061206,0.7532571857718433,0.7612783178082438,
 0.11869185053268472,0.4290579648799948,0.8594493626605636,
 0.23046645448393466,0.6657669092873062,0.4517277671285407,0.2822577658716764,
 0.27364767689242775,0.02510273898781834,0.08086092857594296,
 0.5632918986404851,0.3526950772220985,0.4697775705151068,0.08936276984324865,
 0.9582690846954689,0.4111594665943936,0.4964435402869969,0.5540982236997853,
 0.8173459580517198,0.7868667856514708,0.7914157522391119,0.3841461915359794,
 0.725163718260958,0.9671983419363813,0.4993888742734671,0.14424459736026174,
 0.6447160947525912,0.1905141666975937,0.7060106104788881,0.32875774893194154,
 0.9465331624357246,0.7959722857038414,0.5579525404773493,0.08307879695569242,
 0.3815602457626064,0.600697055277442,0.817916458272614,0.22462908371606183,
 0.9952689037166215,0.6519875260212714,0.9750011921074303,0.14731871388289042,
 0.9126916168649402,0.1964567024830406,0.048243413702183324,
 0.18694888471609072,0.19430086729291918,0.08886078125103669,
 0.1363109467612903,0.19032797087748143,0.13267953909783026,
 0.5687570063503743,0.7588839086124558,0.1503218133441242,0.6441458084731966,
 0.8561831011085881,0.8487994460755024,0.3275875843026982,0.07644518955265467,
 0.2538239159516278,0.9131662957595756,0.1588076837250635,0.8133604139391246,
 0.6794626238899217,0.9714789741843077,0.6370782760030547,0.2145366495805625,
 0.8726350114999224,0.7501226400684611,0.08197443499753498,0.7071200478772074,
 0.281940450692025,0.0896557111648355,0.05383193449929091,0.2801142837374947,
 0.7460563495436124,0.6542164190691635,0.9506204899218201,0.6250312039437333,
 0.2331518152285489,0.7441872833392994,0.2707775396518477,0.9845151127923775,
 0.3138414901116574,0.16757607381035233,0.9260313790203614,
 0.11219618792751984,0.5942107107451994,0.3702456622406569,0.3148966656110398,
 0.449137623486095,0.6036120275799435,0.8939053580123216,0.5916790654230836,
 0.9583934765656712,0.23442439385020053,0.8105132505521023,
 0.024983046667707343,0.7021634235630139,0.7060964286980979,
 0.35538224825959097,0.8794674211569512,0.9897740936938437,
 0.42821881337494194,0.5816244400733009,0.6290141383181669,
 0.20973780119060337,0.10981230359209104,0.5759024394057212,
 0.04281114748045023,0.02984762825428855,0.7522862013176114,
 0.09922947109288982,0.7370300909981112,0.9737621264837638,0.7252251179744725,
 0.9143333575212741,0.7638066124681104,0.9819640798384951,0.4304191625871452,
 0.665306654111194,0.20187642525195604,0.1523903889412599,0.7636517660924557,
 0.93519811367559,0.27719617464270874,0.26212434035141285,0.8762837835459484,
 0.5416747396280763,0.35665877041459115,0.43027470007329693,
 0.40341458905908734,0.30692157188270264,0.1377054523942589,
 0.08596757132669852,0.6731429002964183,0.5252929625461014,
 0.09742599841740307,0.7830938293726614,0.9248534978356827,0.6704046045253128,
 0.01617161852452964,0.9949840230148761,0.4056206081413639,0.421587643708803,
 0.4323399690619124,0.7200003568780418,0.5534712129926961,0.36757669344448696,
 0.43465799313842135,0.05476865912276263,0.1544253183130294,
 0.7876507073804441,0.39879470669585193,0.27289084599577385,0.761684940180157,
 0.29803385963635487,0.3158863349142169,0.13854414961958006,
 0.3094714325938843,0.3469881598568325,0.2152076897414419,0.5199184174693868,
 0.8897190270642804,0.21124135362948282,0.4703921551542586,0.4406749092874408,
 0.3497017564801561,0.6571919579369286,0.2270617733616388,0.3859593010932869,
 0.5484291038765506,0.8260435042879379,0.9834293777890004,0.1760658619207217,
 0.2963311842130555,0.9318414165873024,0.18867926813628733,
 0.20249148741018397,0.42451057032876083,0.029965748554201532,
 0.7034258848963674,0.4598097740816258,0.0367652260408351,0.5011893656932217,
 0.8844347503112792,0.5351753164046409,0.785682290210199,0.3218930638763825,
 0.32124762126428497,0.33759600405861745,0.42511300701799193,
 0.09472700354116581,0.5745453320781034,0.6217847594981386,0.8300000896501275,
 0.7736959256829787,0.7225271501776083,0.43509902509273846,
 0.18703019019918976,0.5013582296267111,0.42329524684080577,
 0.41096991053444676,0.6939297106683416,0.4356366102520447,0.5226258680064755,
 0.7641750114351914,0.10185794086780964,0.018967171023856366,
 0.12837164028289394,0.8408395980255888,0.09606763856646638,0.801816891804281,
 0.5958311582216176,0.16733366544574213,0.9153086275916384,
 0.20031938695906726,0.22157557011914064,0.03288149545475039,
 0.9728607233303124,0.018295781146577705,0.10341337426627284,
 0.38130607727974275,0.1854854267964493,0.3549609489549166,0.9814487555361464,
 0.748334434962967,0.36691585649330616,0.9193997890855202,0.878961982715364,
 0.5957364167756084,0.44507609607325627,0.31410145284708324,
 0.8294803298804866,0.09185655315342822,0.4548796054557115,0.7659029713124941,
 0.2830962631687026,0.719393681548023,0.6948926386561554,0.9054362613925891,
 0.8909111710205682,0.6858723418277461,0.42472366482783375,0.661335858764547,
 0.8117845446434733,0.5688871020311499,0.27916985593999766,
 0.09968475997907522,0.5754489742965009,0.6178453141544704,0.7449455660235401,
 0.8019050185528749,0.9326500436234031,0.6646827498703701,0.6579053908389716,
 0.5428374281575609,0.9282593946884466,0.9733289087684884,0.7528994349479972,
 0.09098678515384839,0.2302337255796174,0.7065388474204355,0.1667879999916162,
 0.21334926051730496,0.5258620442360733,0.8763378500871376,0.5684575516178968,
 0.042049668233020476,0.30306521591981006,0.10990760205121597,
 0.8552705109647026,0.41628571675945514,0.2967935463153005,0.3672055376826482,
 0.9143726109234587,0.3386955176300588,0.2715420006521113,0.2534112994609655,
 0.6161263856479557,0.4610590287426287,0.19844147987461402,0.5640725957853838,
 0.2784925708634396,0.009616210618126342,0.9918036824195589,
 0.6988214509798617,0.04042278030545787,0.9778674443136826,0.4369002505506778,
 0.3512507433140468,0.572687615462151,0.1895242970264709,0.5806103383638076,
 0.4931695665042335,0.7402208198233181,0.9723246994898109,0.9419433910460769,
 0.759010137644125,0.2870679655663795,0.6944105948413919,0.15369125791178317,
 0.06269305166547079,0.2952895708333001,0.22810040329370596,
 0.6088778149694845,0.023531254581437677,0.21606819343614259,
 0.2858007315116705,0.622461571373735,0.6255231898810064,0.8827224104075783,
 0.15017467734857215,0.9338326707440151,0.03724161362393663,
 0.3635717367130429,0.7739906555177012,0.384139228118353,0.05314692454069325,
 0.466357553453003,0.06122263695032415,0.39618001698843974,0.725197091077659,
 0.3634677704812179,0.06535678301210102,0.6037700864487493,0.4964653151129851,
 0.9329368055429104,0.3807840267085225,0.18884428936216047,
 0.04459397087728689,0.9065360354602098,0.028464634642192066,
 0.24702644372957983,0.4481700822300855,0.43075367754336147,
 0.8961558408437363,0.9735952396942914,0.590936015752687,0.3895621333970698,
 0.4421414851056288,0.6057833892005231,0.6760337052749635,0.19257708011901542,
 0.47092576199500635,0.17977427915222188,0.7100024545866225,
 0.7023245024274989,0.9020110497500157,0.05798426431031778,
 0.08864846750162259,0.3238009346211572,0.8820979666530638,0.9743931383158666,
 0.15377330624761654,0.3243297053166265,0.8336741188320818,
 0.06027153029262755,0.40753951360173357,0.7567992395741823,
 0.9482318574802476,0.3447508095175733,0.627414546118531,0.5626842389358045,
 0.5334679241791851,0.09878983280763909,0.9426073894939311,
 0.06668579014917086,0.7271890916352732,0.06648348068861609,
 0.25784093049773604,0.19118685468900254,0.745885853255658,0.2972369047464569,
 0.13726921914974566,0.20257486551721138,0.0807300073989562,
 0.3588899337019179,0.49141797468876325,0.9042332268257621,
 0.05456309704787832,0.7869484887725545,0.9015711282401374,
 0.38839877640831366,0.05791306421558906,0.752050123600893,0.8060947351395036,
 0.31697357494255674,0.7941177331366012,0.3179238420731305,0.4103820162798397,
 0.04676626713357679,0.5321949926047255,0.7243307434818546,0.8341216131718474,
 0.10621046216085672,0.20081205449798445,0.10258989431674614,
 0.02051953792349781,0.49305188475282136,0.6879934316188123,
 0.10182507820385767,0.5386328991126543,0.9690107073411128,0.5446248603983875,
 0.8874060019919121,0.3470373244410171,0.10241200223389169,0.5459243054212894,
 0.24220981053555934,0.4197087751620986,0.6492285439932667,0.6844763124722841,
 0.5301563560859444,0.5065075159120082,0.8969497087149613,0.3559236801176281,
 0.14106515937256625,0.08013782574502049,0.10347358795123518,
 0.8839460483798152,0.016517238528965472,0.5814943312166495,
 0.8530288243409813,0.49543510621000586,0.9369899269695856,0.986134069603684,
 0.885251174758686,0.803945234051937,0.2691097380839863,0.7231888909928732,
 0.8501047857708965,0.2710063146894348,0.25418923928370163,
 0.10959089062415128,0.4740824215107946,0.8556718787004821,
 0.03249493890674471,0.9964883513972748,0.4873057091893994,0.6741119384540727,
 0.8346208181676298,0.8788783314533787,0.2796691508578597,0.366241208482295,
 0.0923740726194241,0.5047876813682471,0.7174972456550724,0.2424735654847867,
 0.5459286897716502,0.01564272876180428,0.9672471637879485,0.9978943027357158,
 0.7157947547494664,0.5235076817017272,0.9002985197096168,0.8980893182253058,
 0.2828820226848433,0.6413822479543869,0.9291397025332363,0.856491154479907,
 0.6810666528109324,0.11019848559197343,0.725969852193848,0.21205504606790437,
 0.049863812789776474,0.58600952159781,0.7266688671252213,0.6286584289467455,
 0.7322110876007687,0.050201548708791344,0.2701118765204411,
 0.16800284562711298,0.5991418623141515,0.3352626516980596,0.92965603536684,
 0.7540022057658876,0.9719301621341989,0.9046428933364596,0.8623440476438937,
 0.43770076437272953,0.09023869080806746,0.5724594479944256,
 0.3824216022766189,0.5055926947888663,0.81179896566594,0.4338086849581927,
 0.46057191358224225,0.04444503918213982,0.7694941094561917,
 0.6472036685597947,0.2619640089886319,0.42049678608269936,
 0.05514155181526725,0.02556848433685921,0.1424674471866738,
 0.09259404065935362,0.9706314001167786,0.8129068369874771,
 0.10544400433917256,0.5114151649982159,0.5263845432281373,0.2167803388840468,
 0.3602959149098305,0.9493678425131148,0.3932026775288108,0.9240065301771607,
 0.43550587626641324,0.7133157877768339,0.6588146918023786,0.6665033765293387,
 0.07192716099363972,0.8147759950743515,0.6293789548381662,0.4462578537461499,
 0.5581454533351234,0.5670077889640733,0.21442593261306375,0.9874006092469049,
 0.3397192188861873,0.8479794304721744,0.5021991258466052,0.6411978845269314,
 0.3345034388252084,0.2202969266815865,0.019068485900717635,
 0.43896907456113965,0.5406373327368589,0.11955017002919144,
 0.8069706965405951,0.6264882409385463,0.9219479916930939,0.9758606912606661,
 0.4805296186813264,0.4332540456516818,0.4137892757518251,0.1811323965766044,
 0.39585996963148007,0.31528835874725947,0.7248766971781863,
 0.6422133811347204,0.7992125442170945,0.6191671524013533,0.1500078350858074,
 0.49475498971662213,0.15341089460565271,0.06772958508112947,
 0.14964200873333666,0.3952761805566216,0.3261091526432238,0.2492145800033606,
 0.16624844965047625,0.5806202990685567,0.747789753318244,0.2638691024818469,
 0.06724819591471642,0.7534629301497386,0.3910540194381954,0.7012319870939009,
 0.6960945120694099,0.4038043498400836,0.7287438999943641,0.4029481579180926,
 0.274463746813448,0.2305378534011211,0.8828003231570677,0.6778339247819734,
 0.4020377902305947,0.8416207643239938,0.027768154132086553,
 0.8048165517153629,0.6213952537592161,0.11475445902120807,0.4702375058642442,
 0.8150730305159702,0.3170170434069304,0.6727654404190809,0.3962273071058823,
 0.5862331552245598,0.5316755365735932,0.3632156399952755,0.8376390360841921,
 0.2717389438874521,0.648602534004947,0.37140795606147314,0.8019637793700273,
 0.7919082886756649,0.12721408817971314,0.24505497545797428,
 0.11790948836605286,0.15005878823716368,0.6110840874874727,
 0.18485492960882532,0.36604851029875396,0.1356887654827137,
 0.6277312127324357,0.0915166727367529,0.7151328664637567,0.39143472849644234,
 0.5955097544254596,0.4088416263075716,0.1390554487599468,0.5718951712079227,
 0.7872203304684682,0.399152558622072,0.9494290561987514,0.40229627647835675,
 0.1972568966753263,0.616455678695198,0.21421078100132274,0.8903546848295504,
 0.8645608385542134,0.6111397135785426,0.45778547524927415,
 0.42333584993916784,0.9540742462520487,0.950444709210448,0.6650400825286136,
 0.32698212207012356,0.08115367761810977,0.3039834222772092,
 0.44042838798944584,0.8639687262751368,0.45847475796560366,
 0.8065072189975495,0.08388554606074194,0.5361880746403623,0.8731585340077627,
 0.05039963769545619,0.1935167473681244,0.7580034381696937,0.8351812124017604,
 0.8263049666251421,0.6303537556674076,0.7197841718527846,0.533970906614352,
 0.09666313660214054,0.47707139584288527,0.08535071158007135,
 0.01748952807937576,0.5561034415374297,0.1091591385249181,0.7441550558487899,
 0.40708569519523796,0.9226921646183794,0.9979485191686992,0.292401673060796,
 0.839449340825922,0.3695422169543091,0.7825380244357034,0.81123456839216,
 0.3542606207232124,0.41174082352303876,0.7917974533174528,0.5341509513070586,
 0.3168044180519519,0.3955212206412855,0.5295694411030558,0.76617041160612,
 0.6156296842436694,0.5818528650529445,0.24832659016106673,
 0.44522865047967086,0.4570467314639921,0.2518294874510062,
 0.11955676339412957,0.45845757362587924,0.6147000781674771,
 0.35017640551304174,0.026432208321167328,0.6424701278960606,
 0.20491351081364206,0.6158860501089801,0.9287162979864844,0.8733237383009826,
 0.5985798673529025,0.5709133372951575,0.750359803974477,0.4681503388140621,
 0.3842407515446775,0.5291909612759067,0.2766347661888364,0.18144305259912719,
 0.6525630630825118,0.08336664796083182,0.22189099599809026,
 0.8696696066340356,0.7451036363357051,0.30020446454989735,0.2263998790097156,
 0.9446543967731628,0.9248753088893862,0.8485740373174442,0.6125623409368521,
 0.6771475512900957,0.21887787922975865,0.10741533424755279,
 0.8254663106493815,0.2937469860548265,0.9427553267874272,0.2151780205212852,
 0.8652478437078135,0.5346427758336598,0.08012591672564007,
 0.016822924190924904,0.24436481409282274,0.8861955988714338,
 0.20932472923924061,0.021342391063536525,0.20193441193903428,
 0.8313889917867715,0.9283960175339917,0.6114540209227747,0.4657802399872849,
 0.8564328868070602,0.7714829499065801,0.23197946902418887,0.3898766777471959,
 0.6778865129759699,0.788029369971721,0.6357545232478125,0.9486989463512224,
 0.47384050710420444,0.15237703149469106,0.2528546855671713,
 0.3068514114365869,0.9932292032702013,0.3264857833136058,0.39973070757305573,
 0.4041522994714971,0.9505672447212024,0.584044230368961,0.3118408611340099,
 0.8678277592569614,0.7149704230089314,0.06707575297431645,0.4905622937829034,
 0.16052608048788208,0.7330996209150197,0.23792595600119926,
 0.27819994920066116,0.4023048819332308,0.2003419410038515,
 0.009690043052332675,0.48003113227999195,0.28490550738689246,
 0.8886624600364803,0.975451691618397,0.1510961706405316,0.49854820035510916,
 0.2748818546831655,0.5486169381147488,0.8913546532932115,0.5166206905997474,
 0.41166811209264975,0.9492747413289562,0.5016589674093302,0.8455694971884886,
 0.9746386147100244,0.5012761977889155,0.7973689042527937,0.425578174287341,
 0.15510507463943235,0.9564508048579947,0.9675576497787575,0.6177265088743774,
 0.48477103750578543,0.7774492291557823,0.37678464696856506,
 0.06298812765287276,0.12269624024159465,0.03874810052971611,
 0.4255341365153489,0.8430796421699361,0.8899126039764336,0.8904551236314987,
 0.9572219954704118,0.6291909111169991,0.6169468990274891,0.71773955166497,
 0.6903028114245554,0.30439408957959646,0.011023267857694696,
 0.3675692565213212,0.062433608563371545,0.09706483257871656,
 0.8959820661285959,0.3472994360399203,0.12821562311849133,
 0.04594776978777315,0.47692113689051885,0.4948473841271591,
 0.8338047059514607,0.6322731818188908,0.45778401177242656,0.4754873976505958,
 0.06783734597128943,0.9112755449865287,0.5582215090801474,0.2825410624090102,
 0.7463277610479648,0.3706816415696823,0.6325941374253304,0.06903123770192854,
 0.741141314323783,0.052327986814989624,0.7358946792919057,
 0.40844639404728433,0.9215965518409154,0.10998242595727248,
 0.9531639574360098,0.5295282773327862,0.4235290278715568,0.8646206241776095,
 0.8354713216119329,0.7060069512065696,0.34317377690880524,0.6830074064963237,
 0.5057934900453677,0.8658396582774112,0.2564014699923852,0.7558953594592924,
 0.9807851013974735,0.6744662753697205,0.8879649582294804,0.8159702257283007,
 0.6324763153147559,0.3683386467280385,0.4253360648900706,0.5744036771224109,
 0.6900447515240236,0.6024229820381517,0.5738856568906869,0.8079444279615879,
 0.009750580965846356,0.26359451514616794,0.19186046287636982,
 0.5284794761118383,0.4414565268464409,0.02652820688891333,
 0.43561733423465476,0.7547984372154225,0.5621630365497818,
 0.03555859172716991,0.16509699972588443,0.9762997885919623,
 0.8886146916270076,0.9192664190084733,0.9317911045685323,0.3200985125009851,
 0.6563408845266577,0.5853925082609664,0.24456280338016267,0.7140826734922976,
 0.34173610850583935,0.9805026308540068,0.3665015829551459,0.8951338351048199,
 0.7241681919860258,0.17220256500189834,0.7293892874850207,0.545925846123974,
 0.6589428742278732,0.6920354566829169,0.9217805706982088,0.5039429305459306,
 0.35525170146942464,0.9301953241038212,0.16507580797346644,
 0.5705711368657685,0.9561359932761909,0.06273722035363682,0.3776437641633412,
 0.09878555644854448,0.998629041374317,0.49754297219969257,0.2296178820054876,
 0.5941885853668865,0.9601261784153615,0.5054584699954365,0.5835620540247649,
 0.39613842033546387,0.1468912912424173,0.5956733206459712,0.906001142046761,
 0.19266100177162704,0.9803265287579409,0.9313819563831618,0.4919469884490627,
 0.46423009183439756,0.8756049220515172,0.06096248839870655,
 0.8114047024564893,0.03705494321327796,0.6210687608819779,
 0.30932569043992864,0.9008051155732141,0.2928597973688156,0.5708993602375714,
 0.05485608968234712,0.8536984566168304,0.8607985003507137,0.5507540043310106,
 0.8903645115335984,0.052468357520827236,0.28604311057857856,
 0.9848125318185244,0.24784813398160233,0.618219576525902,0.32609266340684984,
 0.8087351454595942,0.5233429132636584,0.462258319474003,0.5451453815244861,
 0.9739059155975027,0.12017163814949627,0.47113301834807275,
 0.6079749242551297,0.7461227959488869,0.926874650728626,0.8036220555395939,
 0.11519299012807449,0.9625597058994476,0.8563157107629366,0.7488471274246868,
 0.37929013635664144,0.05162850390561824,0.44470161016442544,
 0.8502840256565827,0.25985649602178806,0.2878492115205967,0.4884506097524801,
 0.08044999200830283,0.4482306393944444,0.6044523935562239,0.4451951236252263,
 0.771825422953218,0.12746142536806615,0.21806698623386367,
 0.42407716750256563,0.22636782028408453,0.26956263706586725,
 0.890988302244117,0.8715190495101992,0.818992362881157,0.8601905450990287,
 0.7119469086945431,0.9481552268332076,0.2857970581358307,0.5050350512646214,
 0.6891227027530632,0.3434943190487507,0.09465637595718668,
 0.43509136189116626,0.5329485997486233,0.5997048324609238,0.6974658722324906,
 0.9424437857389676,0.21962113035565667,0.0052611301392953,
 0.28491772742166555,0.5045357582057239,0.4887656564700119,
 0.48954414803028057,0.5294744702658805,0.9185957777746224,
 0.42039016596601564,0.5387144164607859,0.8954080995783775,0.5601257199280754,
 0.618998577467343,0.5672225064626923,0.14984628138315714,0.8319828817122485,
 0.7693210252903524,0.8483243333832262,0.6178577950848374,0.2832317303776587,
 0.1552928642518424,0.29228901463187473,0.74480208450472,0.45415790853516513,
 0.25419764723404004,0.29430323875116626,0.75223420828841,0.9089580522324766,
 0.9291639064742014,0.07550001723399014,0.060801521341406484,
 0.8096190121293496,0.5842513352252781,0.29856346654682053,0.8254170317169888,
 0.8240626138612439,0.6440824387167208,0.1809615654428216,0.5964731974025239,
 0.06008513615039368,0.6718657099544987,0.87448663122951,0.010885876717131038,
 0.9592892873610758,0.5143202424779114,0.17994564005871339,0.7178676991006758,
 0.39044073887741215,0.34479698909855283,0.06518662421129684,
 0.9323667947033465,0.7307918782895254,0.8236173545414596,0.20894352874821787,
 0.6434164443529096,0.07386099598198492,0.39437697992722454,
 0.02159139610140126,0.47918625988828534,0.8281569005881768,
 0.09208408180322047,0.2660017044285665,0.7105494003272873,
 0.026328287896118407,0.019538438487653687,0.5767487379415284,
 0.18990222475226803,0.8030601245709466,0.8367440970943146,
 0.010980090611120041,0.81599644297754,0.6158132249039503,0.2182326590542043,
 0.6135158158876276,0.6410081419611591,0.45287843101166547,0.5723646520446575,
 0.2988374141351192,0.8198880424458358,0.6581336038922229,0.8463455196968606,
 0.3855983046797514,0.09392881856650304,0.7969100402695133,
 0.37555841713155325,0.5523423747710539,0.46061912583728726,
 0.1176629414928515,0.5343825595399092,0.3173993587689319,0.705932888140793,
 0.466713599528809,0.799204866546441,0.939045583758011,0.11295458874767828,
 0.23261247206456648,0.26561218006809606,0.8273208095016837,
 0.5006929294900223,0.5837684293539004,0.838603300979345,0.6673623474224178,
 0.6101476579353219,0.8336086434283237,0.10999064243243395,0.4088016061617348,
 0.5462031942419348,0.7165438479607817,0.6802713006604951,0.8314293991050157,
 0.7016489113177186,0.8942847795870338,0.14596294946874222,
 0.16056658192562556,0.15477702438886898,0.3269108569289705,
 0.7770465306445193,0.07892214415555299,0.5988432979751522,0.6606533075134357,
 0.4102497828086248,0.37324907233697646,0.5953187534036981,0.8030246103927132,
 0.17913737857558965,0.7339918337500476,0.3490907828507719,0.9724371283577098,
 0.8627489559359114,0.23998682132333315,0.8574126400300274,0.9659559817282524,
 0.5530485197652866,0.5087489483092674,0.9020051863311734,0.5547646955253109,
 0.13594023628360252,0.39512656702629567,0.7480477677134528,
 0.31720847616375747,0.4982510572857842,0.01137999327362793,
 0.4374397578118334,0.24799285827259476,0.9403737034713108,0.9271919386555139,
 0.23545629374874044,0.9105786582905059,0.8038547199830437,0.926417182545063,
 0.7141669217730081,0.6748226534465639,0.48485410179507715,
 0.49180924147335137,0.6111049619894053,0.5968270588766619,0.7806513379969474,
 0.9972483994160719,0.9682283739214073,0.3225099125118235,0.8802702972910028,
 0.394889540136399,0.12780670748793777,0.7430891998772768,0.7687364063176563,
 0.4804359524960269,0.7486295578620096,0.4982490595196357,0.8362041597982529,
 0.9650207139456055,0.31313836697215314,0.18570518182917417,
 0.6818258828537955,0.24076977320731008,0.07223595107188474,0.743970488128638,
 0.6246087763447264,0.04771127800491293,0.7833534900735897,0.9068672420048155,
 0.17434021558218216,0.2511534645087554,0.2585898117329375,0.8910641552720238,
 0.20950963475019013,0.9222195839853228,0.901536754421409,0.9234359686564995,
 0.5778458928263481,0.8088885464480283,0.9168173758709222,0.680331193136724,
 0.5700772368480731,0.4938391295353385,0.2914827252251704,0.42927321356827897,
 0.6049975196363827,0.8886370835954165,0.4425626697683466,0.8010535521509703,
 0.24487787993422372,0.05052225583139869,0.18861449225442561,
 0.06696687948302371,0.7627769772727947,0.2287809461336756,0.329750982906867,
 0.6568699347116534,0.5033474871002043,0.10051485091856804,0.9337242143537818,
 0.7280891206631207,0.08806673661644182,0.9491309102961946,0.6151254625100243,
 0.4544717740176605,0.7093672796557055,0.49255806056643836,
 0.47684675380479513,0.6153922929094505,0.8363823762082865,0.8923591442545198,
 0.28997324318764683,0.3419570568675431,0.859574761720278,0.3899297786687297,
 0.9889032663189306,0.07728030065974101,0.6236150632998332,0.4969105908138618,
 0.6679047269273197,0.7413440197306504,0.4875929897972573,0.5068125404983136,
 0.7743535725324873,0.5501732197967614,0.08770133565816485,0.6538091291266206,
 0.5964095822753359,0.7714309972203355,0.5823665145085378,0.24731167882413851,
 0.45452588141575845,0.7971607084476294,0.19260731627984207,
 0.32542048888131603,0.3461413363376792,0.36255688106242456,
 0.9565395354698434,0.8332540366615023,0.5656243926961033,0.5384131952954528,
 0.9802274965828305,0.8507111474080007,0.7072059271328572,0.5936206419795299,
 0.8578860914124992,0.9505158277574465,0.9285482913964398,0.7946159529869756,
 0.5215944758426403,0.7415991374079726,0.34878432076777577,
 0.07566690406164767,0.9892139933979633,0.4897423300108028,0.6415823099939231,
 0.8610075492090972,0.2822951434086066,0.28889450083744983,0.9708268824996988,
 0.4353313728820014,0.2233219806543031,0.9197944241465932,0.9875740990872813,
 0.3789296062986949,0.7595689494025993,0.19634077902731684,
 0.11376781754009158,0.26080279673479057,0.4723822782940983,
 0.5788069546061497,0.667060114709048,0.8791205274339999,0.8239724434299688,
 0.48416157529457804,0.0018996329670302181,0.6027441621812181,
 0.41691505908106574,0.3288064840828061,0.43111676371260277,
 0.1610368072652707,0.42664498285406827,0.60758068209153,0.14964279182203244,
 0.48970501502170183,0.8207602462079573,0.6018331341185553,
 0.006247075791154799,0.47809201432364945,0.4536878280624339,
 0.12537284121791237,0.9867753151843248,0.546851405785455,0.7737901241341952,
 0.07092471690360769,0.9734977990396814,0.796246105946544,0.29150695904662927,
 0.04186380270394352,0.4414693923865942,0.8230696775361219,
 0.09848705539606917,0.9174069703374717,0.03104747466331481,
 0.39273532475171236,0.4155301583955583,0.1293285589333546,0.3118870434880915,
 0.43729619521656904,0.049712372962245954,0.1418112250765976,
 0.7453588771667383,0.6828176709542083,0.9334043789236752,0.9298614298615712,
 0.6039967763157549,0.858095356210253,0.5104814419491411,0.36272067328867064,
 0.6995277735635171,0.7987165818848265,0.5349314058201019,0.09518518118147823,
 0.2907615115529294,0.251760658911099,0.26070969760910234,0.06309156380092817,
 0.9458797971150099,0.8290278027873093,0.06578207663269142,0.7757540262859736,
 0.9188384328334203,0.738698763842756,0.6751307365437451,0.7070895966257289,
 0.7011748300675156,0.14376232922762344,0.974258772716717,0.42719081000192893,
 0.06999390967909158,0.6415572941162694,0.344791872498293,0.25640874859085494,
 0.926813916011445,0.02957097074198345,0.6261515150600726,0.30144409460084853,
 0.9469055164436105,0.7184985003482276,0.9683387927443541,0.3812448656487891,
 0.14645803622624742,0.5541193179670507,0.8546343228189851,
 0.028903245707377234,0.4319238600624493,0.7792554192135828,
 0.9722389780318181,0.13905598415781528,0.1752620577052455,0.6148261604732874,
 0.5803915957165923,0.9512179041145465,0.8682469763243106,0.08016624764777436,
 0.5291888692159714,0.7273471441400888,0.4862993791388779,0.3852500256791589,
 0.4037117321115087,0.01120678738480585,0.19862571462985978,
 0.19619825975690175,0.06729646201574935,0.34632429006075927,
 0.9693606224783619,0.8023247173263582,0.6601527608224306,0.08017594542895534,
 0.20425715502942765,0.9972712484381571,0.29716938739109355,
 0.09207677527691904,0.2008566326928407,0.6874470085028372,0.6281676227126192,
 0.8838383392163034,0.43815007872872014,0.9574211915465057,0.8512680264397202,
 0.8174729484485527,0.9265473582424701,0.3041464724191445,
 0.059202858307308315,0.9351991518169677,0.3555220965887701,0.68922901595505,
 0.045834267982785226,0.1354944196213712,0.9183516876686032,
 0.14265582163177737,0.9116106158096979,0.44218732138586814,
 0.43798407592687116,0.9993907722267268,0.20029994057860523,
 0.9769176467156016,0.9716086804575158,0.4816409409820557,0.2464784679225237,
 0.7627178760784376,0.2718128596712105,0.37860450274385615,0.4846747071048171,
 0.9612033753748962,0.23474071816272624,0.44807753051628263,
 0.35650580279738897,0.11165976386704357,0.8588588194682658,
 0.5322223008382742,0.2830058092776353,0.697022781404239,0.9460649073503956,
 0.24037088187472988,0.15182827873707905,0.13123515394050245,
 0.6355051068343166,0.9897008438198459,0.40810867612374113,0.5426882129500804,
 0.3153813904091639,0.5484173453669402,0.42793945329570426,
 0.32603373048196016,0.021446711689151687,0.07886359398960185,
 0.30232202377306683,0.5163775313405602,0.5553450788329493,
 0.17389000956119705,0.7682304868246113,0.23947162074170447,
 0.5211982352081513,0.48913866354473745,0.35643929314314493,
 0.5106557811946193,0.18690967843268258,0.46783462790790264,0.808439823213721,
 0.7037242415297453,0.9212044386836361,0.6092614154983093,0.2501560818914018,
 0.6656864952690189,0.6666574810833207,0.20179893155520445,0.8214475681116217,
 0.8071042494783177,0.10176071545710896,0.7264474375068024,
 0.35478188101452324,0.9666034153969505,0.20190579299587563,
 0.4674052884604961,0.6610184564716041,0.42767013393007147,
 0.045608486417776706,0.3324583810190024,0.40105929738849555,
 0.013452743805426604,0.34999499065889494,0.5383639274595131,
 0.7843502351639374,0.7045838807731017,0.6308067109678203,0.2437620257978259,
 0.714856985697931,0.6597486076354572,0.5036436792837349,0.24235069897718564,
 0.941508652841628,0.8546611458073258,0.024387685808366477,0.4412891180742585,
 0.12933279023177302,0.026471502954034998,0.09524899214172244,
 0.16442191154176944,0.15005393993094018,0.5875249351293583,
 0.4827383087024103,0.8046094904455279,0.30870824370579286,
 0.13659383377458567,0.0813087492658785,0.008024527132923676,
 0.5981454451681426,0.6199997316367464,0.40658812329688643,0.7405203059506847,
 0.33126916440584453,0.6413481779326614,0.20115681019109777,0.248885002487077,
 0.08771126853350364,0.992274345072164,0.31028461250581896,0.3383521349754641,
 0.16246672156038156,0.03501123041023835,0.5762629822979557,0.505601840706865,
 0.13549817794538555,0.3325284168676228,0.09786019753034247,
 0.9496623232946149,0.5987078863175503,0.2669256259559304,0.9127529269014394,
 0.13668721536821904,0.556590749847667,0.5739615506322874,0.21420169705603964,
 0.6939303002413494,0.5928045142563498,0.7848990043625736,0.8620066860052967,
 0.9476069919201418,0.7433448984844961,0.7561544304825987,0.496723970867471,
 0.018805953351643145,0.06128185534094788,0.1636373884718204,
 0.20814779700449115,0.20790816411120294,0.8341709843172592,
 0.10612825229637335,0.24622904757334207,0.34578071750765393,
 0.7128490632885293,0.5734143257461259,0.7306382281725176,0.2565098348935009,
 0.45966261564231725,0.6278826169827367,0.1840670036630151,
 0.00872555883328252,0.12255159883086308,0.5994362357313744,
 0.6248632278021221,0.3856761103552262,0.6882294130384424,0.24079897388004223,
 0.7130707353006802,0.9018800594776417,0.13151282256329144,0.5874684564102426,
 0.3325902637670779,0.15148214411715066,0.9920333430508904,0.588813904280087,
 0.6065414335057707,0.2716600488278137,0.023768659290232086,
 0.10564017606453302,0.4170532584847202,0.14062222794690504,
 0.9904462900699409,0.15592306586692795,0.7961546273373857,
 0.13908836484650888,0.42021877696562526,0.0992758000460563,
 0.6175485936718581,0.8240050556309451,0.8416231983966285,0.48116565914787335,
 0.3482286607152194,0.09555287294003856,0.34417468428414455,
 0.9214728282294977,0.24993813083967675,0.5452589223799658,0.8032953657510162,
 0.44775106208432436,0.4178845134070156,0.31180950630544857,
 0.6060508561256632,0.4010492974551718,0.659869796890836,0.3763104641809414,
 0.586956471091727,0.1622835459911769,0.7826821647318152,0.41781098646485926,
 0.1318593910953777,0.005912754081903837,0.13579044874288404,
 0.8767695541950219,0.3344704247516218,0.5335755585399228,0.10144956855162135,
 0.3093905423909986,0.6895294904159163,0.7052133124463946,0.34893827996340443,
 0.8669635386596336,0.3138218726395363,0.9672111503641746,0.8740347874194017,
 0.41672686469071396,0.2449039742191622,0.3355488800504266,0.3526606018723659,
 0.2933228731873384,0.545660124931887,0.34400207534730987,0.25291030883524535,
 0.9488323161232785,0.7219798661236887,0.5256419982655194,0.8162059397121341,
 0.4287419736547977,0.5952328731318234,0.11933609288059088,0.5494777481016213,
 0.8149709965329717,0.9870615155977894,0.06857113696519856,0.7087632638064666,
 0.9461684309667505,0.7782903997776208,0.31808087433008214,0.3467240577830588,
 0.030734365116518725,0.41208292680832537,0.3633070856875049,
 0.10082558099698169,0.6219185457254917,0.2659625282011491,0.6973820959629266,
 0.48969331105970604,0.9973960178908033,0.30061748796977705,
 0.6222475502484422,0.10115434013032787,0.8814205043527076,
 0.09326192085580332,0.17239416932198337,0.8450861054202141,
 0.5749555410602404,0.7549912575884756,0.9869067123860645,0.6870123460317646,
 0.5345939536863532,0.7493231855029329,0.5693561913012638,0.42038527673820114,
 0.22483015043679955,0.14040616465773925,0.49375961845755234,
 0.1985015703462959,0.3569256861990435,0.7055116638774654,0.8957314783244903,
 0.5373933068984506,0.191855212397958,0.252921395558959,0.8506885341257004,
 0.6051015555059427,0.1716745720799162,0.9205413376048692,0.6237168537255535,
 0.8325196188699413,0.5268691594292415,0.4469894371591391,0.6122478873568173,
 0.1742871719686373,0.39903597979157013,0.2628601318770626,0.3379593161244182,
 0.14533105392135304,0.9051518347268983,0.600075779096773,0.21212719533232183,
 0.16896138482438594,0.5412063568149742,0.7679826260700535,0.6600161942108814,
 0.9656004715206024,0.9175610565474953,0.08837296171067299,0.7165998340181072,
 0.8080881071014563,0.34473662329520827,0.3288718092499221,
 0.40650112931173954,0.6746119536830831,0.3216317209970848,
 0.42339518320507863,0.9936669719668522,0.2766995585528854,
 0.35101703124345707,0.15111993481490404,0.8185444341165419,
 0.5476052524602705,0.5544642647137606,0.0047771996828223084,
 0.7577567768976808,0.8954252731126898,0.3540099948731713,0.9237006271483243,
 0.0031787869438921756,0.05045410264999162,0.03201275756521427,
 0.8138925773558969,0.7470588066858735,0.8187125832712825,0.9172990967902237,
 0.4018367556272262,0.591274888938577,0.821961950221551,0.6890151891166281,
 0.31119975418500034,0.12281465293441174,0.053540849475259056,
 0.0030777850499051906,0.011954973880258679,0.6993621903662685,
 0.7803995918479615,0.10132218925533487,0.41458900123422915,
 0.12886138232124345,0.22160308627061986,0.6925035546881259,
 0.018248312322271087,0.37888360490492723,0.26625709900760564,
 0.05141535923042151,0.23693119069728974,0.36045786490322484,
 0.5457630586628286,0.31458577235421614,0.7489784536075159,0.1952999733633337,
 0.3861242495316196,0.8292231568286063,0.510414576567835,0.17583120515932782,
 0.1005561626584518,0.58574028322147,0.7465841092901149,0.4225427468044236,
 0.09061828435641794,0.24704974369786092,0.40657216569631205,
 0.9535642091609102,0.9077590236691495,0.31711262231802595,
 0.35351995409308457,0.58158340414688,0.689402754009993,0.03805665154251914,
 0.532938037797601,0.4483625391790308,0.6710889800494448,0.8627525308165945,
 0.040424913991699674,0.3774707668888304,0.5976105488704768,
 0.9962182321694069,0.13830658026181042,0.4551764752351315,
 0.08511401926314166,0.9918688857723905,0.06995174559563933,
 0.015927725055159603,0.34058745916708977,0.18470906036710066,
 0.36618163491015676,0.9478083427848837,0.9534897892439393,
 0.49991434590554473,0.6336546522255538,0.5887834137933935,0.9071863356217871,
 0.446941059453571,0.5275514688542283,0.8242742574431148,0.10655585086695873,
 0.40923693638179814,0.797010641000716,0.9263927764399987,0.20275362956168685,
 0.30433800039931314,0.9584094700765026,0.9888128299342515,0.5156093211410976,
 0.11216395536503154,0.8229409654879946,0.49221363227098003,
 0.8268452576787783,0.9173428559805515,0.6658657235682247,0.6036206797556487,
 0.8689367757177978,0.8640020661640634,0.6137458134372291,0.7251278229936101,
 0.6786562041445432,0.07868901653719118,0.24824915922885604,
 0.8348988708825344,0.3097256173854207,0.8906268734880343,0.2755831537964888,
 0.5010512344300091,0.007621442257000788,0.7382475925477514,
 0.43765196070095513,0.805370324686919,0.9410628254161646,0.9052225529514161,
 0.5722035439885899,0.4022891393543917,0.19397347534666176,
 0.03125241544745183,0.44874400576736817,0.8332366783207332,
 0.4938349528515621,0.14796430643735237,0.27841999696580555,
 0.6804588976300241,0.5497378405040214,0.5580731220518648,0.5167303306256457,
 0.5139146487171995,0.31987778844374803,0.20683007402714526,
 0.7165399404637569,0.36048918518775064,0.6082718203780819,0.8812974174073541,
 0.7543410337646543,0.018457788699736244,0.6748298803454418,0.985174428432587,
 0.9810982978991138,0.21008301595251222,0.580936937755365,0.6927578487071592,
 0.9607608604720839,0.02153336423463048,0.48748472844485835,
 0.6922928699345301,0.8233928938921751,0.344491144753993,0.7017933163247099,
 0.1469711282538052,0.9348916672594793,0.18272672105421517,0.5236163237012506,
 0.6257507614041349,0.36164298863681066,0.12472378637879211,
 0.2620458652523817,0.6069539057713101,0.9527590324318311,0.3851551898955089,
 0.2503979580272524,0.26782687124204185,0.13136933159035802,
 0.02287988633882887,0.07265101036434429,0.1146377343017384,
 0.11860579235793478,0.39365387436356813,0.11880590970500482,
 0.557491485005418,0.21688659142322564,0.10241245609929717,0.4829427795149215,
 0.6897954830740264,0.8536904167959819,0.9396258179917965,0.24205028926386152,
 0.14703982118569092,0.11069917375986349,0.6434925577636967,
 0.3343373171601731,0.4274595073874834,0.7078105028777885,0.9660954534951547,
 0.6103901176569787,0.06265471378694998,0.1882359271518803,
 0.44111426521215624,0.7344500143857857,0.6282026607884668,0.8689046029576468,
 0.061145208820485575,0.8886473141484578,0.37195963609043314,
 0.23651334449699357,0.8296128697250832,0.5045525895316111,0.4758174691877822,
 0.135917842401735,0.017230790052831124,0.8002371238352559,
 0.37388193219258403,0.9415059679896727,0.017542168224973897,
 0.3406701870081632,0.8735454073520702,0.6579310177259673,0.8594526825812165,
 0.8756060115560725,0.5073769965671456,0.29450476617900945,0.7005176043734949,
 0.3543796895547193,0.9925622360345463,0.6249253390434116,0.5449334440946522,
 0.35494476244479467,0.9248430585922164,0.6869645205017448,0.9961067874309443,
 0.3829307423916084,0.7094181354983689,0.6619831340464869,0.44402791473465797,
 0.5906699346421681,0.3835210678977302,0.40321133384522545,
 0.12023650469261715,0.8726709418457328,0.6917177748029463,0.7853160896932891,
 0.5903515613434968,0.9271292450219379,0.49634171849916553,0.8931616145122901,
 0.45408742036478145,0.892990870235419,0.90341938717762,0.0641093878622937,
 0.3280962593804666,0.5741024195009227,0.9145919121538745,0.9000757875397482,
 0.519077502873423,0.9103611269984316,0.8287521581141531,0.9011580666760706,
 0.24469895212110404,0.2532855163230765,0.8802761299760384,0.9954651759456705,
 0.04184692942995383,0.7080492440324495,0.01536612040666474,
 0.004236870558290473,0.4838586401862379,0.8231773651232315,
 0.4684300070915868,0.5255895018507314,0.0563781911510316,0.3835324313342864,
 0.955480341570763,0.3528173347983502,0.644910330120262,0.28947390487057134,
 0.9652323099983844,0.9274343279190543,0.900619040556709,0.6393413673701229,
 0.4125528223242998,0.2967142613218289,0.5673507762742682,0.7057304596415648,
 0.6419148810406856,0.9709373703988523,0.42094834968769,0.40389699925022793,
 0.565048602770813,0.595630144901144,0.916696853890945,0.16257545645411542,
 0.4941379720129049,0.1530466001539883,0.3950263385826197,0.8202766894489593,
 0.5197761109362586,0.444611593061933,0.5482802172036065,0.11534319400813353,
 0.9863019559756645,0.9976343272849442,0.648270114409095,0.6614464992684213,
 0.21378840694507717,0.22632162958714208,0.4599390845560345,
 0.17381156087598848,0.764222466055225,0.7615081331587823,0.38273311117121045,
 0.490027206250374,0.25555032405196243,0.9420050867107717,0.9698086179320595,
 0.4117439163934222,0.011002182017985973,0.24444859754379122,
 0.4758819032369762,0.48281097794778516,0.8963453119817072,0.9178487258203991,
 0.7023128357464697,0.5964187599194709,0.7526909139731148,0.4187807333289635,
 0.6472162164756831,0.3719230607098538,0.2976906528161437,0.40845970608233295,
 0.5185585476270782,0.6179330427481335,0.7738359638517549,0.929552306983504,
 0.5238534859612705,0.5765848913602257,0.42753669947691864,0.3395225448621131,
 0.5968129080421598,0.3493628990697454,0.8114194191028453,0.5202431929657372,
 0.5956841812056436,0.10796848256766856,0.14893565136196474,
 0.8855764236692739,0.6803417847660181,0.604302195764293,0.00724785731897426,
 0.2539302284701628,0.3148579448719544,0.10402030552213548,
 0.31743395476323777,0.9779606196370059,0.3579241601677996,
 0.27710682677378207,0.35129179289806345,0.7733072990919316,
 0.5669809146851038,0.5364605769739325,0.21458902635957355,0.530109614700814,
 0.23667654166035246,0.6966095345007787,0.4786396035694649,
 0.10278316117097275,0.3997445437203312,0.8650847877386886,
 0.44745941369347475,0.49579040237335803,0.8283471848506814,
 0.6255409552003707,0.9335919734966462,0.6720848552217638,0.9089542149589394,
 0.3375454855563349,0.8102474651644191,0.32157447155609464,0.5857152314296087,
 0.4225917864872567,0.523272248084459,0.6211766826262588,0.11435576412086035,
 0.44192431212198824,0.28006159142921927,0.02347193354668342,
 0.9543167792407259,0.7364665821114152,0.3965789154287017,0.6555498130729405,
 0.33775630901443965,0.35194462893989964,0.18469420563458794,
 0.6099647617333085,0.3499061748856056,0.22381929756504582,0.6171876996610869,
 0.26686183258043994,0.5261765399291379,0.27316844997316503,
 0.9200323681880357,0.8403887694969268,0.5607325790742174,0.3962939586964036,
 0.0998473429393456,0.46989593625527404,0.8532851710040399,
 0.01446265038587946,0.5467388527910257,0.8543527953623621,0.9000367302765986,
 0.3042687865113076,0.8136182311274097,0.7214940415503421,0.3801927020608358,
 0.4430994140430663,0.9922689713422588,0.8622393486826305,0.6091061509267071,
 0.5969369336892323,0.1928458334093741,0.33970878518686143,0.6241364380795531,
 0.9272242198447896,0.4409035945942048,0.399193490340214,0.28104209734059304,
 0.26697296502143775,0.14954080900865674,0.6493182793658634,
 0.5938740662826666,0.5450843470579643,0.4624923945409538,0.12553199800018,
 0.7278591305486186,0.6962801649421115,0.7421503868624562,0.7293393588985,
 0.8837251755437096,0.1410658838744545,0.9559157029212668,0.3002392744651602,
 0.9909751127333392,0.6547562006799066,0.8124566474707355,0.9341397286777483,
 0.7714573969049802,0.674562443095107,0.823152338194237,0.8616592891756278,
 0.45249360449687415,0.8224164310802748,0.9512113249496621,0.2637867959437402,
 0.8679531175320602,0.6058726604367823,0.36541806181078174,0.5628065747621523,
 0.17150579406807653,0.13247287878461067,0.331557011823561,0.3335406662699143,
 0.282137227749922,0.28373643384632297,0.14098503947555574,
 0.21747949561040325,0.5001876759118786,0.10232209708481554,0.799305183995443,
 0.17472011377383978,0.4128883310603533,0.8150807827813619,0.7961526770570281,
 0.16419438972705847,0.7840692790144654,0.4630237098003386,0.9947249325402937,
 0.052738824625755854,0.47778682756916746,0.6828791548233308,
 0.7909862456941138,0.5568123206173003,0.8081110608115607,0.9145929602958377,
 0.2565822601406885,0.9287928033946826,0.36848266910567584,0.7408222901814039,
 0.4907447525932751,0.12644663162257164,0.7358714190619786,
 0.28831867665367483,0.799665896475544,0.1250274500406643,0.3342332377370184,
 0.577878131144274,0.4156544830047527,0.838997487863825,0.9739228445663431,
 0.10218488733549924,0.43746647215116474,0.3601893620921718,
 0.6872030250501467,0.8695437781685349,0.5629329533436082,0.5163172515714675,
 0.44978894415513726,0.8714936200982631,0.7182985361608188,0.6158290411680527,
 0.4078508874824853,0.04128820875989592,0.003240366219245394,
 0.7513959812094037,0.8446528410127067,0.17174328096852387,0.9743028867114346,
 0.06269960951046671,0.33883221105135397,0.05371740901644395,
 0.5211259106084067,0.7655513639717324,0.529093617386009,0.8221411321499813,
 0.1939010768913848,0.7444642896355029,0.9207373669848157,0.5733743861204691,
 0.7255669269559146,0.5907015916859188,0.21622534393329595,0.7915539986822984,
 0.15214289410586446,0.4532829270360067,0.654770720967784,0.7967311879636091,
 0.6145520256651104,0.9388425680632402,0.674866362568441,0.15222153621890266,
 0.5993645988670732,0.24747560665285107,0.07783833573134635,0.993537076901158,
 0.01712159065970098,0.9675245857401463,0.7582218862084272,0.6945412357985765,
 0.3728020496237884,0.7199853566620061,0.7163847408965553,0.5319087136204235,
 0.7880319008993654,0.34991050335368024,0.011366408607936251,
 0.4694263106467229,0.23183196707317077,0.6653517646296934,
 0.36271289134745166,0.6811306031598052,0.9680730394250197,0.60558383320487,
 0.11373223813298583,0.5014576329481363,0.050090047733255205,
 0.320685842099771,0.36942731461170963,0.7434014447094287,0.6434732845744399,
 0.28295361271154695,0.7252078723364193,0.3546923615643096,0.5843538188775118,
 0.6206639438341569,0.0042557263173256255,0.048678434947617144,
 0.0635123645460548,0.5217741228824314,0.07439241575487954,0.5549449300506407,
 0.3492871241987784,0.43920266976243705,0.18338893539731382,
 0.7971845109865838,0.29216586296461045,0.6036003179999159,0.9569434098899525,
 0.9596513471196274,0.5581982447078412,0.1926268489673798,0.1444028518882644,
 0.5222699464843759,0.31478182491414874,0.32149213906190444,
 0.6745412372330272,0.12943884339825296,0.3642872625365844,
 0.38634368752238857,0.3478910143010374,0.8153904608289383,0.7256353200660051,
 0.5463504089085197,0.41310538581499723,0.3435573808909671,0.8073082207679626,
 0.9143568063087248,0.7486725923113187,0.9485436655386214,0.07141519709958621,
 0.7866225471534471,0.9296457383497819,0.2585917897387897,0.03837849135890503,
 0.03529139612795407,0.727222754738561,0.3178892879245221,
 0.025875311893760244,0.34869102378207617,0.08368465708064998,
 0.9916451771523829,0.90389638877555,0.8635395275934276,0.5235522874372152,
 0.8380594093014578,0.6604882606994384,0.9565853795612769,
 0.019933241608231844,0.8983487571069915,0.38818754583714454,
 0.6344957160980764,0.08114935435387882,0.2003466011359436,
 0.006773574342534672,0.9692873866591067,0.10389216804350787,
 0.0755195072010385,0.448226916708248,0.07120091060264078,0.47291167870706174,
 0.695547139495781,0.5136029895710856,0.9941601639937544,0.5286349792654382,
 0.10516685959546324,0.9907362627764243,0.05235246898257029,
 0.02065807128448216,0.5811691613077321,0.4751875334156227,0.9690964797320324,
 0.7650849837245944,0.39465283789455907,0.7228106879531466,
 0.46670902937419234,0.5914916821272707,0.5082432199527026,0.764174831140819,
 0.3365113136612089,0.11311500187582713,0.6386855212709652,0.9043367200778492,
 0.8834700541503294,0.6794972993908386,0.542750898276808,0.5535084805239069,
 0.06217165477572373,0.17737623940421576,0.5121136213778849,
 0.0873217891359388,0.08088551199429661,0.3247045734867158,
 0.47601956329563766,0.46739708622263443,0.5204393980227483,
 0.7093922179734339,0.7362491087151853,0.7200133701773475,0.588370416947634,
 0.3478174121671581,0.727843151880423,0.7687082953847241,0.016721106401988872,
 0.37334717614842405,0.7123132834427311,0.15614503418455883,
 0.7884612266628324,0.34639069522544796,0.5031535709792838,0.6336397018004748,
 0.2605721354131113,0.7536831809760589,0.8017088593589611,0.6165429555153149,
 0.7939923284198916,0.44843590035232905,0.1821622075208117,0.7852778802839984,
 0.6189405661628187,0.20541067398229007,0.7617577330509866,0.4030681492700354,
 0.818007137459728,0.7321071959932802,0.19525388252151166,0.4612319083891371,
 0.589836323182505,0.533039749795173,0.8368210744733258,0.44028846909955954,
 0.8669600495205447,0.9576695630172589,0.1613505468618871,0.4077947097129395,
 0.667214268375031,0.046785033960640154,0.1276330785289811,0.6714424426950464,
 0.9056749039644744,0.40649600570098765,0.37987875689654904,
 0.5766822855714797,0.5423246084126329,0.9687659697733261,0.4265674660116924,
 0.4252679274808636,0.21188940024307268,0.5532039967016843,0.7379160477521953,
 0.7097241485131977,0.569396465288728,0.10794114580653136,0.09179960545377097,
 0.9017603801641223,0.960961411534807,0.4771678451458585,0.49319448517451836,
 0.6918106660288608,0.7168565348695042,0.17354068622710184,
 0.08241589680835493,0.6932697491669179,0.8957794680396761,0.639821065649351,
 0.20373761969988247,0.9040104308739868,0.16148944458720638,
 0.05984189956810826,0.12012201365119302,0.8730195987646258,
 0.2503603522336205,0.5920568340894263,0.5081289719531082,0.1129322408167498,
 0.9746513613615392,0.889453406557003,0.4368496885124876,0.5683519825447,
 0.774390064631022,0.5850575041282912,0.9714543835065558,0.0258172893645654,
 0.9695624346856994,0.16319757818707514,0.12418025674522071,
 0.2372037300301446,0.269139398299184,0.48304385976527353,0.3838533738641534,
 0.9787295632716326,0.07285979686101907,0.28099378470001457,
 0.39486400228140106,0.5999793013811057,0.6896991449340264,0.7404684338139532,
 0.19043176068993617,0.19489360588376703,0.0350242076147681,
 0.38979789265746745,0.7975061168582491,0.0499657222047879,
 0.040893058437169394,0.8953699387616115,0.25722504878122654,
 0.8794883873416749,0.7156216854985968,0.40671570360138376,0.3047493369247126,
 0.15453131535742537,0.5764901741858275,0.7546683089587447,0.4468945349516408,
 0.39542787397599266,0.45183011606399803,0.7422088881674086,
 0.5548189881363546,0.5024958460907587,0.5654130747473363,0.41763540840785274,
 0.11977815507512013,0.9461241948601573,0.9551264185581079,0.7496672837859161,
 0.624912903667211,0.4044384911203984,0.26581614373143114,0.6080216459050876,
 0.4021914608535141,0.7070478784329645,0.6036868686907346,0.5329409854194286,
 0.9676175312295989,0.6962412225292661,0.22693857340358126,0.3076592120582363,
 0.5177417848518878,0.4738559627304807,0.01157096413390224,
 0.08389401324961177,0.321401601126334,0.004846004795180825,
 0.3910749051805402,0.6602213006439179,0.5693206455856393,0.06025746560036127,
 0.18069141547625733,0.2510452521021679,0.41407416012086096,
 0.8550565854692163,0.5408541898362603,0.4378417754882562,0.6599285529998196,
 0.8802949103236819,0.3180974082149327,0.10933802640395263,0.3015826912200391,
 0.12112610881293473,0.8687745320020253,0.2899352790357884,
 0.24480013768643383,0.8424914528534888,0.9286365103538187,0.8623973032475977,
 0.21127729070011314,0.9641657325265541,0.8164148200448571,0.8610622878553532,
 0.1853978322396732,0.8151991075554064,0.3813849756090324,0.9848904027299161,
 0.5468235104951504,0.032112950092456294,0.9550650186034922,
 0.2070893743691513,0.7138861319559715,0.5745086863069212,0.11064175821325528,
 0.09241486421966316,0.9928724506543177,0.5890691267455588,0.8181505133892986,
 0.9586579627103523,0.9406756214284823,0.26489952578546794,0.6950794740309358,
 0.4227406666397262,0.27620829570359307,0.378756601327501,0.6959126736669654,
 0.12775762174052685,0.7707193159156042,0.9185617008896891,0.356256429319626,
 0.9024468035597913,0.753599339308542,0.634949637983964,0.49230600322242357,
 0.9974420084759686,0.5272087490261466,0.20336295908515778,
 0.38006254604364065,0.6107569814694632,0.40889163948694973,
 0.12029419090749016,0.9561052148321787,0.8083693638471972,0.825657505389763,
 0.025323331668282267,0.08470085477211331,0.06555600957191365,
 0.9142761485041945,0.20633716042993666,0.143422176192728,0.33312922602587525,
 0.020605191346536933,0.9682071757634632,0.9180461260622259,
 0.13971680149729404,0.7991367946223072,0.3180423951621819,0.5673044138456006,
 0.6331942822112266,0.5815879234065727,0.019045292004918002,
 0.8218087912439208,0.1749600447097538,0.8657527351482636,0.4139984826128458,
 0.8851619555368893,0.4534846590824677,0.5437906288541875,0.8723443328248086,
 0.6701355951633196,0.032371118541196786,0.4597710958501494,
 0.26977626009408406,0.6577482127198544,0.6582639864053623,
 0.35392312369158374,0.28054543422057343,0.18463305498219718,
 0.34901331159326965,0.7798508741131072,0.07088618403945757,0.631164200797868,
 0.9341150042533715,0.4789967105510117,0.542232366618951,0.1436289152493494,
 0.17160828996376365,0.25168136441554867,0.8038766703017843,
 0.4543287016356842,0.8468984956072246,0.714003673054329,0.17688974215122388,
 0.6786283701801035,0.09111150318298378,0.12702674750612097,
 0.4014580330720813,0.6075795614562435,0.581366889187612,0.1428442923061467,
 0.4746326153499899,0.6778293603030188,0.6898476701867278,0.06130784293170555,
 0.12269756726172754,0.40365054407500855,0.05290189524126021,
 0.7989390537725048,0.14029041520004015,0.2882162693393604,0.4943027704138355,
 0.6716212761218512,0.8341992657787389,0.8006844253135953,0.47692310564023743,
 6.808638113545706e-4,0.41473488729889463,0.5519516868198509,
 0.908265046596953,0.10158207526798457,0.3670216980760713,0.559561268473421,
 0.6755438649459542,0.8738836888528249,0.8532106033285142,0.10992457742806472,
 0.3569011674339524,0.7141124973071834,0.5863725882620403,0.19719436096845167,
 0.16790019166569548,0.08736467904703948,0.0769477914748834,
 0.023306546186940058,0.019092288039750827,0.12680879664228706,
 0.28075304565382786,0.36764164052067194,0.10109169083104053,
 0.017674368494105086,0.26501450333933174,0.03714146334841428,
 0.7444121191619434,0.5919271774634425,0.8959014871321003,0.4201013345109895,
 0.871442799017163,0.520808263989409,0.9096955971369314,0.8145710075021115,
 0.26408522336747375,0.3947347022433312,0.7769279118436851,0.937720140751622,
 0.2524956325023008,0.15583446172098836,0.5488490658146904,0.4338754404757671,
 0.7891238157504772,0.4892493919864134,0.9856914580322279,0.10629509579661334,
 0.38088210254585564,0.19763211582343632,0.24695212822453405,
 0.29390614772774015,0.8180660908063919,0.32879204118685745,
 0.5980970559113679,0.5896123897846016,0.9959511087424799,0.6229384892484741,
 0.05839821345752971,0.020062493171629336,0.5005727051337067,
 0.8415679324440637,0.07651427338358596,0.3619700412584844,0.5697689269162005,
 0.8058508637684837,0.40651051711304886,0.6074158312038037,
 0.15352873268192102,0.018873515796510643,0.627933051222044,
 0.9572824608059277,0.05889693895953341,0.913107114125855,0.8499109149902873,
 0.679532316742414,0.6716557959602918,0.7176934239729821,0.45283014799439014,
 0.7448612619769142,0.5058315106865187,0.6243229369576287,0.16982238553123774,
 0.850094318689366,0.2581443188902127,0.9299354739604779,0.5828268583871016,
 0.8496722435684326,0.7194903647472766,0.8535129895999124,0.4240635622797221,
 0.6609832048295499,0.8438406527060538,0.7345110915778303,0.4946867638034388,
 0.23639080294670423,0.31292908736350333,0.4472969475622658,
 0.3499087085719912,0.38510472020302355,0.6525833613694805,
 0.47083128619054926,0.5486580844627621,0.8626663121877451,0.9736117720224019,
 0.005339070209320451,0.08662228679500483,0.6320963068072358,
 0.8273066497649131,0.9437223385620914,0.6520311786910664,0.13741758980015073,
 0.6511623063037841,0.8158408444402354,0.3121411946083158,0.8057091365947202,
 0.44848288549840154,0.5045441919736967,0.5237262686008867,0.6315612952772556,
 0.9272803627709266,0.33502131778279176,8.512806617710922e-5,
 0.7570603083762949,0.42143725515808195,0.9219543622988735,0.4127538737621338,
 0.4080291370570065,0.46547479634774236,0.7964976044305359,
 0.21114981256442356,0.7593107194670041,0.06208372401815865,
 0.1522058963698435,0.8037633914570474,0.8868018672916209,0.6963251735435478,
 0.9832356984776662,0.8058340005586516,0.11687187694899015,0.7311576497190281,
 0.18024618273300685,0.031179978760406235,0.6595122195656699,
 0.025183299995402142,0.7693677289308368,0.08261785711025349,
 0.22681011671743145,0.09134872669188487,0.4558588525379428,
 0.21892368808217966,0.7333789164266995,0.5250149161497487,0.9520124019644516,
 0.583158246672745,0.32621935058182194,0.012945071557337484,
 0.5297633503695439,0.9086183738651167,0.25218011113215866,0.3822532542359629,
 0.6406775956304862,0.796237873044952,0.5399135167348537,0.7694439188278422,
 0.9933614356439826,0.7746477983264263,0.47437089332283233,0.8379080894294301,
 0.5513640578148719,0.2969491919777265,0.8836353271577063,0.42351169143820644,
 0.9075909584883368,0.243737216808947,0.5845726097531196,0.9895422265174796,
 0.746956192386366,0.8950519720232476,0.4951156760205725,0.8813937186439925,
 0.7144785899147694,0.4453588930241461,0.009170742742381055,
 0.9199457734811438,0.17077380445017953,0.2255267482971105,0.5129292340849829,
 0.5248979792864108,0.5351192277328262,0.3117469245611768,0.3870183237896401,
 0.3855118040823615,0.5547103748057982,0.19944481214594945,0.9022915659810313,
 0.1711040558828525,0.9228118199623244,0.9494716532852145,0.5009563440697192,
 0.8043366238480615,0.824582854946677,0.2328634816156585,0.44573627838130836,
 0.5493180504104392,0.624376509445414,0.2679231011381069,0.31685782452308353,
 0.5070273924253778,0.9140606173456925,0.9468632917192151,0.4951249794309721,
 0.9453296338764176,0.6081494941024637,0.4942299471244076,0.43865200969697293,
 0.25254613093805367,0.2074272412732533,0.5135492293463573,
 0.21065965064874792,0.6573579087948165,0.3065261769664366,
 0.27973874685272393,0.5914794920772536,0.06282633132919291,
 0.4738020400131401,0.2559703339466548,0.023150512123022104,
 0.5678328365922714,0.7543261825747171,0.6005646520077915,0.3903759610862966,
 0.4665836109636905,0.941037623945749,0.758934848218022,0.7712538958698243,
 0.5820249936496898,0.5497715020938884,0.21829538737610732,
 0.05595808944320768,0.14653394347081383,0.6635930673147181,
 0.6864600559602961,0.025332214665794717,0.5087552773788998,
 0.19162971132547424,0.49390561335771077,0.8812835257474898,0.436698345406895,
 0.22869298536164273,0.7415473251357924,0.4187966384455917,0.6934583097588781,
 0.9556434747228338,0.6610524136446013,0.016558995040950597,
 0.8895471379298073,0.6078337213922466,0.5415277475303375,0.2712128760741208,
 0.5309768021245965,0.21201612039908346,0.10942016128251753,
 0.6202847977004624,0.7739432566714404,0.625483412013456,0.7030713134416735,
 0.4217909791920812,0.7460340077898934,0.36876149382942647,0.2780266272883427,
 0.5988675098492947,0.1837161197183872,0.10623909035390822,0.7016814109604833,
 0.9870012935028096,0.9040682622791505,0.38788214107503816,0.5205515822027807,
 0.7228053161615674,0.4408040699287249,0.2880075204193713,0.683676892386341,
 0.6354431183279797,0.5028165483723581,0.7235330644729371,0.25962129996166394,
 0.7158933993954789,0.6139643297493785,0.6114380411279161,0.8442886582359121,
 0.4742502405127651,0.057656163693938445,0.3318460460465614,
 0.16178011217454924,0.6219520824140317,0.9857950400133239,0.7451905558904982,
 0.2610894333996262,0.9968162932257615,0.6035973936260541,0.6677751716453588,
 0.3978996102542326,0.6825768514703812,0.715651835732074,0.6485575216680821,
 0.05880447259223187,0.5955842597554295,0.745817728955974,0.4792898699150403,
 0.13750870794462422,0.34156811366341977,0.6024981711817063,
 0.3936474588800798,0.7407652831370368,0.1610731737486205,0.22779397055234407,
 0.3687254453654427,0.7772418677347532,0.4873944056871762,0.6607791025545947,
 0.9354027688927722,0.4335760846998651,0.29914320627315116,
 0.0021574928754196643,0.3890617275902395,0.9492737380130425,
 0.42802626717055914,0.10242987730901654,0.21622617799849297,
 0.8856620603769476,0.9439798666450131,0.6634479945742174,0.9092348114244926,
 0.3667271291827632,0.6346436785278051,0.40047731792855357,0.518970738275237,
 0.2995127738128288,0.18235476517463678,0.37090077694864865,
 0.6098229662089947,0.21831022345268658,0.33136920652179946,
 0.3095500886346936,0.3100495066377855,0.9128669923934987,0.6806794103689274,
 0.21832458985263636,0.3996150629540659,0.10662048759140708,
 0.3014174544050505,0.6214540549730215,0.6256451567269953,0.2501546557138442,
 0.2539410820362257,0.3950004124364168,0.6381848717592878,0.04867341291676841,
 0.4051046684089714,0.7664504420468878,0.20830435833793814,0.6115443587428366,
 0.7047071928488644,0.7537485702268344,0.030762249562161448,
 0.45958332278812175,0.6771243926892824,0.573104551213958,0.05838385225971532,
 0.3713762845928079,0.2613950266979971,0.13605289576393287,0.814707403261921,
 0.6311817912310463,0.20522504325128144,0.962232789543051,0.7406818236497072,
 0.6655297910045593,0.15507008469879424,0.536910569822695,0.4470645765204381,
 0.47337005894216033,0.904851848356574,0.5779429718215198,0.974886284774156,
 0.5695495947015663,0.874434990467802,0.3706816460637474,0.0896784484775206,
 0.32883611336986485,0.24344772346245913,0.22301460735536083,
 0.773636754497979,0.42661253554888845,0.8245536403901361,0.3307692997583589,
 0.6219130397033218,0.09616943567932779,0.9056976753722406,
 0.34260630834321004,0.8753132879701891,0.3698389419602073,0.7520202298521503,
 0.5121940246429779,0.36492122032104024,0.3646855764122735,0.739507258795986,
 0.7539956438561196,0.6159521302036038,0.060406401034126644,
 0.030653556303928076,0.5217372396740831,0.28370384970669593,
 0.3681276112834354,0.1548370884673118,0.7415961138529422,0.24608670216834416,
 0.5123590649368597,0.10492008089945515,0.9905225464500614,0.6348287792742567,
 0.06904438824370185,0.8598921236467779,0.571537766599385,0.48362785892353044,
 0.3366276922308826,0.03533803204018404,0.5100030881834214,0.7707945192639981,
 0.5910055764898657,0.28358117931489946,0.38729893284564565,
 0.5249392005530892,0.25229534337230386,0.13096655314338412,
 0.2897378349745031,0.9706537819561826,0.238565632865966,0.38931886472446464,
 0.5161364837803482,0.35046570176132397,0.2370608806204758,0.7691452963230379,
 0.8203915419796822,0.3985093080320985,0.4075708414743642,
 0.010410146899326178,0.4562437550685249,0.1562695792056038,
 0.0464051774063462,0.9424274317139314,0.9740907176339706,0.7293274742049496,
 0.5867631856647283,0.3437194336368655,0.26118949068001485,0.8250431931782213,
 0.8708278943195806,0.7518925897642545,0.28028003530311885,0.7808349395337194,
 0.9679215712015009,0.6812058002709832,0.279649028721657,0.6708264577668798,
 0.10753303698317218,0.7095695170856032,0.8238140409518937,0.9333815127362562,
 0.3594663489269019,0.6182621984980938,0.12833413299783836,0.7974332995046038,
 0.4047648026769821,0.43177955106536625,0.44764550959108707,
 0.4402445680217143,0.9342118087418622,0.8287697334544223,0.4446381844666537,
 0.710657395825378,0.7827693621752521,0.3004258233732464,0.6797827359848884,
 0.30518023190724697,0.23737773673755402,0.6502353148724853,
 0.4242115322013763,0.0975694502437251,0.8080056471947274,0.05556131017365418,
 0.593498834486296,0.23700584509714084,0.23458270418153215,0.508472947819205,
 0.6518289480159493,0.20443302459992108,0.6766396656321446,
 0.34439603953663867,0.8725713287636144,0.5006523537448565,0.7080941817833011,
 0.6991349014604078,0.27755438514164044,0.4911765508218441,
 0.38363112330969074,0.9433995936321558,0.2642828504490149,0.9105840414734547,
 0.6696677747507265,0.8786491755214446,0.4507528494205256,0.7349611736316312,
 0.4149977953390682,0.8510796531126844,0.4149872918207387,0.7419116290844959,
 0.26295218216227734,0.05689674263578115,0.17744218398852551,
 0.6803852171291774,0.08974236158118665,0.38774928573163114,
 0.7154914010169181,0.9871838194313605,0.32520505263461874,0.8100573050694183,
 0.5847048007140204,0.8922938956717317,0.9549335701616756,0.7365981054920175,
 0.6589565947489004,0.384056168618391,0.8517008159097839,0.9434507578936063,
 0.02578384068653461,0.42083624321180435,0.23316938823911637,
 0.5224651157810842,0.3659037509618114,0.07607374670631717,
 0.31366692112852146,0.9456405083548856,0.2436122600803875,
 0.39269485828964557,0.73553729459426,0.6905185766128186,0.8274481223175695,
 0.3033487986367729,0.44313174182329707,0.41393596390147747,0.531355402845984,
 0.6985855678266752,0.02707000766722112,0.360473286223414,0.7782498524789103,
 0.9015329417604233,0.5030585655827484,0.8654325216531426,0.4462872842385788,
 0.9475840685366889,0.8739198280471017,0.7663753359066322,0.5956813961086279,
 0.9022728337054716,0.36849316956099143,0.9969998828622539,0.3430000913087111,
 0.6318750996378072,0.3551558859004056,0.3006815548812488,0.46127642277293646,
 0.8755589201663343,0.23104629044512515,0.8757751117696058,0.2930544826131096,
 0.9644957103489992,0.4358776335472023,0.14554320097988116,
 0.008134667488662872,0.9737231536257432,0.5459098623330065,
 0.8187504341157205,0.634480613632624,0.6526405455851543,0.1697269190960342,
 0.7925033445325778,0.6544945670092247,0.06763149566808557,0.9258894961881136,
 0.07042649375252763,0.0741052385270159,0.3528615601493925,0.0780141223770846,
 0.7845279121847684,0.969546656266665,0.4413122863194361,0.9163237564175812,
 0.46066743241398067,0.20315422482957313,0.043591038230681,0.8207936728657471,
 0.9031387011482174,0.4899428120746059,0.8579360383286257,0.24269596968877738,
 0.8388461023100221,0.8068670264895692,0.8222703084574858,0.34639274611731075,
 0.1309996055724424,0.5118059279577452,0.948082609741743,0.5637371555852528,
 0.4508904622713632,0.995564506165967,0.11115486020735899,0.7644539253368566,
 0.72814498368815,0.6064345151793094,0.7068262514057186,0.7017107469899546,
 0.7648563653566017,0.6818688686117477,0.22067281520673032,
 0.06442828322145244,0.4432011343007336,0.5039385941343348,0.4963575521888297,
 0.8563969927493869,0.1228069136854204,0.7368800036579842,0.6276594041619086,
 0.6107767901437211,0.740281317546664,0.8012784385443201,0.53012771890936,
 0.7992977858485022,0.09386872580618943,0.6425305917976751,
 0.10591482157911347,0.2563473813226482,0.46040532873614104,
 0.4713005250458736,0.40208296015907663,0.4325911278909653,0.5511918263014641,
 0.6545146851563525,0.3527457827227829,0.2708547927313325,0.5270350910042698,
 0.579936075913442,0.8989810709568662,0.48143072494308337,0.8943532974733988,
 0.9866656562292788,0.3462972545173222,0.6433503236741714,0.5075912794507371,
 0.9898483015964314,0.8322005943584606,0.26450473677445774,0.5443871077476408,
 0.3289692042071378,0.8064270983582363,0.5218112160866581,0.7368916960395782,
 0.5075132744237094,0.13415692143083346,0.6900851092380434,0.6629912921759548,
 0.4190751456945454,0.38989787572020607,6.950907360899761e-4,
 0.13346131989443966,0.827161870284452,0.6781773035330443,0.846442573624901,
 0.7525793987261612,0.42432481122111776,0.9441154251482304,0.2345996261844775,
 0.43551494012971836,0.5492112883261665,0.5139587836345569,0.9480793591347167,
 0.861336909900359,0.9596458886034017,0.9000372721630108,0.9976939753815173,
 0.43484427034213313,0.7646913663136552,0.3441106518160473,0.5471574987559267,
 0.5135640751027755,0.9374631806256486,0.9908729249100813,0.9416291290220986,
 0.1323963284260402,0.024258664348867143,0.7723336331884426,
 0.3511835280746478,0.888157025101103,0.5202490445162697,0.43210784851529704,
 0.4814036855577589,0.6087220170085322,0.19813439568295454,
 0.48266377760138135,0.0684176796394288,0.6381152726521265,0.9022637029004439,
 0.2849311465180089,0.9702181183411722,0.2603287205506246,0.6657924923014389,
 0.8329716825012631,0.33760780411580193,0.420876876432857,0.2967260650520701,
 0.34037781368752096,0.7251354286886977,0.2296965572018249,0.9241029328492328,
 0.9728396047508022,0.4274131189880823,0.001008313404751271,
 0.7431320450882468,0.32790981710856204,0.04400993125554542,
 0.48681446301585307,0.5509917115431884,0.4621904124091687,0.4682091647112391,
 0.81257659777321,0.1310847519948931,0.05817988782783412,0.6976055288350802,
 0.8305924333599533,0.01881777624343639,0.7661710741018466,0.8089116274474561,
 0.9674436474936149,0.8238959988704015,0.7075498752207631,0.3630933071149045,
 0.4692842161440368,0.3663093483346288,0.918647476929275,0.38579567997552655,
 0.7734290412846829,0.34033781545043373,0.6284505490065377,0.4008466507607077,
 0.30506415631779815,0.08698249293704408,0.19783388910821542,
 0.8379281124525773,0.20305892236893197,0.7007363082888509,0.8063270880887979,
 0.585710943983067,0.9716449171729422,0.11118608569992106,0.4562635985423482,
 0.7248710775842427,0.9078031069706901,0.6908132305528256,0.6337476564833392,
 0.6513336223889606,0.8903727613043513,0.010380410431790499,
 0.10234064982687996,0.05004433698840327,0.08729972866371216,
 0.6619601328511153,0.3883085729724929,0.04997773114029336,0.1466637988199766,
 0.7727160169891084,0.8865961642735953,0.6065330648626488,0.48160197380841785,
 0.9280831225780551,0.377984583447875,0.9641573518222418,0.6170937590262655,
 0.9109764774303565,0.2267705434649009,0.7621908621869775,0.9754927528506813,
 0.43220794266563034,0.5612520260562597,0.990895619427246,0.8648543253619636,
 0.5020304487330355,0.3971353847489262,0.5344559682358354,0.9887192424183985,
 0.3659984850828535,0.5317462778284963,0.08478075814169739,0.8311734288657263,
 0.2842227911640811,0.15317373291516012,0.43304908985844637,
 0.26473384543655487,0.17383242657115283,0.5048567570366753,
 0.10886840867411929,0.2354316574646591,0.40308865543396477,
 0.7291686795799869,0.017154041091287486,0.06621620492003166,
 0.8403421755416001,0.3311103090306037,0.14763448773450838,
 0.030593607049687943,0.7683127485697554,0.06654750355939532,
 0.1198418636747306,0.00609198265281663,0.25030654098100336,
 0.10444927202176268,0.7662693523481992,0.22844343609673912,
 0.7386859423130098,0.4868268314966555,0.609572895744724,0.7837551400918218,
 0.8963711818157118,0.5929810413603628,0.07992675384895764,0.5213965826522188,
 0.5439854515065562,0.90116805664683,0.6729942325118049,0.557424723164771,
 0.3614866234927252,0.9013618128629829,0.8373687099354601,0.8989692073603422,
 0.10349356897687134,0.33736487100442336,0.25057527930830215,
 0.4280041215011934,0.09032771680450391,0.7760529681259779,0.7321569399939173,
 0.48731632296525285,0.48914556500001005,0.7598248599097289,
 0.0283038166325289,0.812529498360499,0.07252213990165957,0.4325944562836115,
 0.8499279699764174,0.4051409183485327,0.8819924911959299,0.659736586704915,
 0.9118903339349332,0.7775784713660691,0.29380776312303847,0.5888338922879945,
 0.42979509849589725,0.7548376645445857,0.6007447556228833,0.8025007294612683,
 0.5545273116545121,0.22615616450584708,0.04775182182866855,
 0.22255677459778944,0.7482447916795797,0.9408143190905617,
 0.09207796391760703,0.2490807256200802,0.7157526132711592,0.9182505585099657,
 0.18904668960312254,0.04552287343135519,0.49153205115834686,
 0.8490840078252997,0.1185810468271995,0.11118522209031256,0.8655863897153301,
 0.5923417711512496,0.46855784791708355,0.8123807801105711,0.7109134505658864,
 0.11223270708253952,0.07238717654351545,0.5690021449665414,
 0.08692126802027511,0.538803738381913,0.3530745422885162,0.18871321119028717,
 0.5990740331791656,0.27580077129539415,0.43095935231722804,
 0.004826191041827466,0.9100194926344801,0.14576894000740725,
 0.7045727975731751,0.6441209239935309,0.02347636330163283,0.2291000698220672,
 0.5115721059063246,0.062656622057078,0.2014088219680027,0.9960642249885311,
 0.7606655922417264,0.560667731503502,0.09204027555899408,0.7495977594153725,
 0.8110660824524352,0.25937519646365537,0.6295884601966334,0.9632740538908359,
 0.38618399872060527,0.48555189764147344,0.6837965627386892,
 0.9981245266856228,0.27999533902857177,0.9651262284523359,0.7620964157048207,
 0.20182096628643242,0.4520388814027386,0.375990051244641,0.3720059071414469,
 0.7169579094312548,0.6960736826587289,0.080244295315545,0.5076833039880684,
 0.15119507991072767,0.8281048291921749,0.4817192197691338,0.6395811963998775,
 0.3164690106095255,0.4899731105048628,0.5537851802258955,0.7316234684940449,
 0.8135979433954059,0.5112667035291683,0.07574767408481398,0.8812181389510947,
 0.4791058638190814,0.11369722482408817,0.9834736040500031,0.9562511905045451,
 0.164940433519136,0.22814492478570325,0.0011073416394239421,
 0.753591105810977,0.5499432660432384,0.5598593185308038,0.7084722220055213,
 0.4878642317860371,0.7886575037283254,0.24374618499667466,
 0.26415867402446125,0.5367502527052778,0.5414176061331387,0.3596596812573104,
 0.6197644627067692,0.07060042270924993,0.8558805101884045,0.9339874044549704,
 0.12371017344243784,0.02273262288669453,0.09285268681224856,
 0.15458405951016496,0.29760396561209446,0.9780180780554936,
 0.7900942695268349,0.0252366594186324,0.8954598870517745,0.3516104757408065,
 0.04792047782666753,0.36409051604974874,0.3439549707758417,
 0.7906858637410742,0.6324738165313682,0.32563398523624254,0.4067890416597977,
 0.7643040612036296,0.1354509438922019,0.07186177584963649,
 0.01305412093013869,0.44956737280900294,0.8306680379206376,
 0.9667733637041402,0.643517104520976,0.3830170120524393,0.510719075702872,
 0.9475272892391586,0.4459148772991476,0.39548905421167246,
 0.15773337547106525,0.26808517473546534,0.6240701024169155,
 0.8737618348232042,0.8608660783230345,0.8391388985277524,0.7327726343396359,
 0.6084768374946803,0.9990466368284786,0.19312520873855155,0.5091291283446977,
 0.5008395519805563,0.2896465472535662,0.9625840323804521,0.7946477092836145,
 0.8345875510083737,0.6584856896796728,0.45162864684995163,
 0.23399731252848288,0.9907402286260821,0.2855006504116493,0.9704399995083626,
 0.9209451769447835,0.45758944939935675,0.9958890139837772,0.7749378489324453,
 0.8360946972155517,0.04183341675920693,0.46573611384600455,
 0.00884853873467939,0.45151372624184694,0.9096487686508998,
 0.8492332231789287,0.9168555242245804,0.46859370553538826,
 0.37039170130545807,0.5481428874060984,0.6380705949083736,0.5417541869927809,
 0.0977365114985208,0.30493024132270985,0.019174780425282867,
 0.9868848119846897,0.587606208212141,0.0910900695628617,0.582666258799547,
 0.02526821399528767,0.04958613823492697,0.42491379898801473,
 0.3147212449565848,0.4239070611351935,0.02690529115824858,0.5440224873226649,
 0.3858186008192156,0.15330697153877804,0.4145712596157288,0.4742641893226327,
 0.2704229036398219,0.2596048242984236,0.2935101637058021,0.07677535296125626,
 0.133347901342441,0.86851272546956,0.7933291221235308,0.26241258623826336,
 0.9106378014753613,0.5960624923501048,0.9996894829151233,0.764253446960344,
 0.5650795619752659,0.4869694747461871,0.028420747131784108,
 0.7690287062466424,0.25192267378268607,0.20642044399316228,
 0.3574576640096213,0.4140161336295686,0.8138716418071991,0.34504598211604587,
 0.388329134244742,0.9295080999731965,0.6540027918159749,0.5708703401786526,
 0.6519716912108736,0.7634819154701541,0.630953291678968,0.75267155292597,
 0.16814859189854914,0.8107686518974797,0.033398795080195165,
 0.26964196674245655,0.5370886886516177,0.2870521675796527,0.8243137358935975,
 0.7357757347374965,0.11332713782554682,0.660331403417904,0.15136245537216497,
 0.7406951580978021,0.47687654216096986,0.6469093620693803,0.1524338299592204,
 0.47498812718013883,0.18262279409740612,0.6432528089214659,
 0.23508563155194007,0.9991770350797493,0.2244879859304415,0.6358524826598363,
 0.35203713654229785,0.12557129568467762,0.2979100734251563,
 0.7864113646304708,0.5637200923349444,0.08075612478919336,
 0.14399586338009251,0.873210345996822,0.37762290187150405,0.8254995043483739,
 0.20121460468295138,0.6831310419056008,0.30338914161184793,
 0.6276094986835278,0.08843258136076404,0.383202658701696,0.27116626360202045,
 0.17099371942028363,0.677931528874811,0.40549502139333926,0.1736642262346605,
 0.6260681097123707,0.33569334360178504,0.27652091109892973,
 0.8114309382299707,0.21689747082606137,0.2732440696078109,0.9243181249474202,
 0.14086506455572767,0.9485755728341688,0.14331405568907085,
 0.4935083706352107,0.5198511548295565,0.5705654453385107,0.37463601958529447,
 0.5218774285472823,0.2586824615956804,0.2871894159593483,0.6316517161410427,
 0.803199824816118,0.7058608057198404,0.8334824895762079,0.8471754201058339,
 0.05445800836720782,0.7603510259342614,0.6796157442350254,0.8182143546652368,
 0.6745839622229851,0.29355837786411176,0.04939293007405665,
 0.7288929386066944,0.2555935024131689,0.5203313001421903,0.6465284916526104,
 0.09650616668825751,0.6716757185058406,0.48481190610382374,
 0.5801053571226414,0.10883951590468244,0.9991322105464282,0.2357817325442143,
 0.03985378885447155,0.878675707088738,0.5763554926023107,0.3720287696459271,
 0.09959714454138024,0.7993879737255473,0.6749804234077533,0.3922492378240223,
 0.08358643832920631,0.3932028241717982,0.2733286186568177,0.9104164885778359,
 0.7973240253739654,0.5686826869360275,0.8257120541268004,0.7963468974409897,
 0.490530195894909,0.4842777658817182,0.6597970646504443,0.43730447656961224,
 0.8138840000981867,0.12487831701313035,0.24819327201785557,0.83077601955222,
 0.039546961953456194,0.5552219414565108,0.05283626922520157,
 0.9839561204378078,0.5663455742944427,0.15846348891288997,0.8318614548478022,
 0.4159574374497301,0.08445867574320387,0.9030449834226245,
 0.21403342294148642,0.48311523711999926,0.5157251736577381,
 0.8654858298442212,0.9634899366322539,0.019662005496401957,
 0.4036186807776251,0.4492274026182046,0.8949032434147132,0.49421147577784263,
 0.2995244322652919,0.8084610123285313,0.5884328505391614,0.7625136296374296,
 0.2512568408247541,0.2948246968908811,0.11305579625463735,
 0.38782810964562664,0.061787499651936195,0.8067526157520624,
 0.4762299936181883,0.48730381355354346,0.632780194120607,0.8815315281096997,
 0.8988129826176275,0.1533666620895323,0.3519095707574995,0.41683439307138714,
 0.2644690655571925,0.6943973693797043,0.03818862277547508,0.6798173607655489,
 0.005332043032446054,0.2877245628159928,0.2049489023378981,
 0.2550984157781334,0.22671258507236736,0.9828591733400454,0.7157540455751198,
 0.762237479386509,0.2514255510533473,0.7480972057489665,0.2704320629448338,
 0.10493661340868699,0.8731832280573588,0.987981822518238,0.7757545434665581,
 0.0760418394745036,0.11547179560280119,0.8888732535227458,
 0.32409726477414413,0.74739718138405,0.4753225000690876,0.8174826662219788,
 0.15449828757353257,0.9517342244211265,0.9717193008931849,0.9966855826085926,
 0.9179405644153146,0.35561823828827854,0.48694661462317146,
 0.5149062532258847,0.06716132681948628,0.4122051804026986,0.663411268782966,
 0.20408843203214644,0.5072214332418721,0.840386985757237,0.10250340349952136,
 0.8392934733849315,0.9829119063655165,0.46953371991214,0.8916476291595793,
 0.8971770517204446,0.8293231377274172,0.7813509346800737,0.7744902583756103,
 0.9446119124795274,0.5044814430268922,0.6266646608981915,0.5481018775763277,
 0.16569186712990092,0.15283631486875027,0.20565960921132653,
 0.7486505883870833,0.103475215599512,0.9805982913515834,0.3090771880189801,
 0.6588863218188964,0.2806266456451043,0.9508742123606477,0.2714001912037266,
 0.8459160218794525,0.7122646518026184,0.3190475954976826,0.8133229352483009,
 0.29952918893254754,0.678305484836784,0.03758091080955106,
 0.06419790661463731,0.7334226661444683,0.7294794572176855,0.5353447774905389,
 0.7437125030242119,0.5389521058584708,0.09632100209646388,0.8810455943190225,
 0.31193373652378176,0.9925488546655956,0.8490125736144062,0.5898568567923198,
 0.6220385143946326,0.16034493835778163,0.054406966495317954,
 0.0057447108279771175,0.40316905339251585,0.43724009439967526,
 0.9762359544568077,0.8097767845444122,0.08776798226947446,0.9843999941852865,
 0.23150462210239453,0.1373113647100347,0.27753870766490385,
 0.18858305977387424,0.7692536134891554,0.9408114331908011,
 0.27212374403168527,0.015610520588477561,0.5842192456265989,
 0.862167757376469,0.7285928758086385,0.8178439449254407,0.9441146755297936,
 0.26996324228166546,0.7548116527802636,0.11683230103918574,0.819868722069367,
 0.8761954511474871,0.8658275242231804,0.9772928983910665,0.15759110680639166,
 0.8058205674842821,0.5883494278384179,0.672167113934909,0.00586545251018078,
 0.35581711689383144,0.05612532658214353,0.0403200224196576,
 0.5323265782007576,0.584266701962599,0.47019094196424005,
 0.028664894842377775,0.13864231714532038,0.5783289700084784,
 0.30259390942673936,0.16403311778169238,0.08214263037617309,
 0.13099006654457312,0.9028357602990599,0.8562796827206611,0.4426380287527616,
 0.12373610830095927,0.5939231641158842,0.06970818878561702,
 0.9940304851472024,0.8308052862393122,0.8246134240930911,0.6725407502567126,
 0.4731027871215523,0.28429652662115323,0.05154634759031129,
 0.6142475776933372,0.9164564564942256,0.6729256261356951,0.21463903044800037,
 0.34040839738763695,0.7513021338350461,0.26899777565076044,
 0.34749462488907656,0.087854174004801,0.03366586337113786,0.9925263220854984,
 0.4135890303033881,0.9293849637491526,0.6439294985031886,0.6347280041895483,
 0.8554228210127437,0.9571243755259129,0.7101195503883664,0.16386675354706137,
 0.6879204756746707,0.138482052131232,0.758729295550715,0.9019625875420524,
 0.5205353361763951,0.5011924249277857,0.8262764078320828,0.38063995040442966,
 0.3566638245836913,0.9689287842806934,0.7888829521401175,0.3966204348390737,
 0.33428210455799245,0.5454535327557203,0.8132408339002601,0.579026548890685,
 0.44878032055584705,0.5237984104818237,0.6134263097047759,
 0.21514921559666467,0.10228319040184508,0.5862922832661721,
 0.45813330540230013,0.13419369964074723,0.8876697632986646,
 0.2403519368502285,0.7848903125851424,0.7664802294762215,0.5517744377162667,
 0.5562522776633351,0.8449496689164421,0.7433063040807277,0.9022849192102169,
 0.9088073460932783,0.9388546754955889,0.13069247807738393,
 0.11224776217569299,0.8870289249725057,0.5860668911285004,
 0.49169018761028394,0.9883032919894678,0.5221158143962238,0.0760382018688962,
 0.0731484651205837,0.79147492354065,0.3489617755411303,0.9087785892217204,
 0.5803458373467006,0.5758448577068286,0.684992544833997,0.03913372243627877,
 0.1527804494718712,0.3727451861264778,0.35853470873838844,0.3174418376608017,
 0.3596299169869811,0.14848570017635787,0.5915854285850389,0.8920956998329047,
 0.8743448897656891,0.8944942412543866,0.028163969177015247,
 0.3621374619329809,0.8737348732980492,0.015441282958323588,
 0.4514021827742357,0.3783771158991953,0.3013096046888074,0.3234419951515187,
 0.8221038682356356,0.20646718756512517,0.825804920604594,0.28865205550976103,
 0.8319932526050385,0.5441740208137006,0.7111034210997683,0.17020061209205528,
 0.19759121624645515,0.7904898525648859,0.5216354184040476,0.8777873295954842,
 0.46476558691657266,0.8233860556752224,0.2818081196417651,0.8273069716003117,
 0.2947210765995665,0.3478970954058338,0.006697715797935544,
 0.02792524611042624,0.36305461615368995,0.40681379053899047,
 0.33725721467807257,0.02221718595486777,0.39641494841216973,
 0.08739597370301444,0.8562324990517213,0.7803235294417415,0.6114713053956244,
 0.9506583466504344,0.12390220429085652,0.008652251594615357,
 0.17433648518433253,0.286088295703153,0.0696388558573402,0.6940478355114077,
 0.6115952393949105,0.1174596661154621,0.957166687878853,0.41191994738134197,
 0.582718737702837,0.9161552180064847,0.006923123785863683,0.7318708474840894,
 0.2862492623361649,0.7984911228049973,0.870982768712223,0.5402829805974508,
 0.8832842058543925,0.12713490889332335,0.37462634763640656,
 0.6988857895001637,0.5333679429687158,0.5513754105844638,0.48880477883455287,
 0.00474686023257731,0.5167575376922224,0.8703709489453656,0.2994716885254336,
 0.954935367211794,0.042750077939698095,0.30016519939113295,
 0.6624187788367837,0.30882664393619996,0.8209728221040374,0.7697457789684149,
 0.1487679091763069,0.29912433594379273,0.6708101345416645,0.3065227368767913,
 0.3783574284322775,0.4731088987495806,0.1182021369293571,0.8779505169624395,
 0.3007965296290538,0.17157504338339602,0.3068771821793571,0.3843587131492897,
 0.5731547833967424,0.4876828795620347,0.0472005027179605,0.548169696606226,
 0.8303845877971776,0.23188555203582817,0.5206574234310862,0.803027583609044,
 0.1556567422090963,0.9163078739777902,0.2209287924634188,0.8376807247255146,
 0.8294594850141253,0.05760672062942329,0.06480171530419276,
 0.4358916264785534,0.9728888398963691,0.5452142795739522,0.6999023612150875,
 0.6838865479712282,0.8168333991342042,0.7033918530418084,0.1213975378625669,
 0.513952118813731,0.681020701805633,0.3475927928417404,0.21727311507935798,
 0.12824006572854119,0.07680234742885605,0.6312367065153923,
 0.056081273945378984,0.49391927928460366,0.12825522375872622,
 0.5393280958143039,0.7785713455791183,0.45022082782917827,0.5475504643798558,
 0.9064273255487163,0.7369739837772435,0.29053423980766147,0.4866663654214596,
 0.4943858133502267,0.34264805930735553,0.15770772262371768,
 0.25167036762703343,0.39119460062188094,0.36706005684028997,
 0.7887536660761029,0.3531718328645659,0.02269049242125698,0.1531007292955111,
 0.26801016062189675,0.17101544995927387,0.5446290824361477,
 0.6628162148478125,0.9742278545739476,0.7401097703274329,0.8104840949558363,
 0.8150967659981296,0.7785951188461474,0.930002469190671,0.23371442652508667,
 0.31378452330849305,0.0820074929310739,0.8306963127732329,
 0.02277813084530711,0.45733893598402475,0.6124052718362453,
 0.6306524797199634,0.5036184998711706,0.17791615703862396,
 0.29127933980851617,0.06118519098468034,0.5408220627133273,
 0.1903101086415997,0.2512768497556188,0.2847738967742305,0.18706989297161047,
 0.009940071282627194,0.1294971758987966,0.3354335777666124,
 0.35189372008341346,0.2730982125190242,0.5620415666230275,
 0.08822501470501842,0.7915406150937908,0.7988745813277838,0.6134461178928584,
 0.03382528693830289,0.3583300977499997,0.6212117120258753,0.73248789795675,
 0.389544561019013,0.7257753343971975,0.9811929988677754,0.12146977456958041,
 0.6592195906143029,0.39672944794421305,0.3085258386796753,
 0.20014178361359536,0.3914371578499345,0.5659409032147711,
 0.28172703325236326,0.26582168289294683,0.9196181602915559,
 0.35444315224781864,0.5546120047694914,0.39846966032427644,
 0.09949904218169847,0.983445673960975,0.5722296701361383,0.6547702328362769,
 0.631762074051455,0.6705982771216156,0.4172738003546863,0.7790175493875977,
 0.02440049259588395,0.6039442344969425,0.220085963622221,0.5859195662395003,
 0.1011128461931734,0.3004582561119924,0.5855329572469529,0.01587749945382555,
 0.2058316044649412,0.21389186446878317,0.5969826995327172,0.5972884942475297,
 0.9690203107702781,0.8540443442834832,0.03524120654319918,
 0.34103278827699535,0.04738120543851032,0.2755675359388804,
 0.5351902407781747,0.9590632092761008,0.8319121217861446,0.09789690997491951,
 0.25125453083154736,0.33376185204832687,0.45613257090075066,
 0.9149775337945922,0.5855138765576542,0.5482801207286419,0.03791382228795814,
 0.02236245756321631,0.751748043300332,0.5424812824376067,0.19115433325447018,
 0.19754850190588247,0.4819635782591637,0.4865486230014231,
 0.06493092898801311,0.05324929881954843,0.9774302053793656,
 0.8787725910388069,0.9523372885328818,0.43981705214380284,
 0.025138434406470678,0.27980786066312846,0.6308951906056514,
 0.6344084330445138,0.894660727371197,0.41641117460451416,0.04308118115540449,
 0.9278322067452774,0.2142705083929909,0.793222423467845,0.08815615196109006,
 0.5132233199191905,0.3771536817241594,0.17102411679384655,0.5038459967219321,
 0.35280782405129574,0.49706312984193324,0.8711483275165381,
 0.3309852318896258,0.9605554331344437,0.05343949838199402,0.6727334952735728,
 0.5591571256858826,0.1556888598994992,0.9403884168308342,0.2897135774113635,
 0.5136912642795256,0.9479314675808467,0.5036624598410379,0.8507999183806485,
 0.9048049340274781,0.8647538971518808,0.5346187919742833,0.7638046422193108,
 0.5696487092476722,0.2847883225113107,0.713831706436927,0.8589876050747466,
 0.4789861663213083,0.7515291345473721,0.1499217186604509,0.897333843762342,
 0.8796734049724997,0.9581423649417355,0.606645627681655,0.934877181414088,
 0.021721857064135097,0.04016028617306333,0.18053950182930434,
 0.6086330010048366,0.1116252226882758,0.9218778236327313,0.9083674577148737,
 0.6579858459726972,0.1754506815620075,0.045447150482944565,
 0.4153876073591767,0.07488588323473033,0.9928352077130107,0.5287399788774186,
 0.07061453307251597,0.29329903100900534,0.05305211425284084,
 0.7259336504776573,0.716521714104946,0.01978659122341797,0.3931347718069145,
 0.4784227546421582,0.47596736696928743,0.939546530326494,0.634070978614064,
 0.5002780208520128,0.15479438970594828,0.5982718225199357,
 0.37262939610035584,0.8800937779207781,0.22791144934424,0.1160850595294568,
 0.9620211488438779,0.10215730039556381,0.7725827009765912,
 0.09473477915867079,0.8234086955034711,0.3636699276544698,0.6074446822467405,
 0.9353990998117686,0.9976239484654911,0.3383535751189931,
 0.035464535091652594,0.20464350957551236,0.4362654493132876,
 0.7372803494722727,0.24062527230883823,0.1747061936701667,0.8297102653161197,
 0.8504405580544632,0.9338374828383843,0.049478631405621365,
 0.7899946790921673,0.2079422535555202,0.49761328873377797,0.2630478988533832,
 0.799992434121507,0.9064696121390996,0.20758275246227642,0.8309923922129864,
 0.783761812924045,0.2631262726124989,0.6483117943115593,0.7599239775952933,
 0.5915075128051489,0.15987904821484455,0.22423169732481008,
 0.28866516983072676,0.7840013730711014,0.15542029098387233,
 0.4839998209790991,0.297851097489341,0.8516815461303424,0.46209175712626616,
 0.48442856135334944,0.5127986916041926,0.480502149376252,0.5603838243352448,
 0.6822154791353175,0.39881498778317126,0.0781019734027395,0.8639485841485494,
 0.8786405360922968,0.3990180708727028,0.2430751083731707,
 0.025462065465514883,0.0714444590190243,0.17087529003587365,
 0.11561275665186532,0.6292554758571699,0.4697144443255905,
 0.16889796889492525,0.07296070204850236,0.001557092579058894,
 0.7696106904466657,0.6069341929729313,0.4717177890844564,0.286697505282449,
 0.8359948295544173,0.43116756266298695,0.13521629152950143,
 0.44527428099814703,0.6048171453535314,0.47366289534130557,
 0.08349067240547292,0.7803309631008944,0.989795261912473,0.07294990229028886,
 0.7834744076587459,0.16030670958909599,0.5039017302435983,0.9099522154814446,
 0.8013640481719627,0.18109812843099793,0.03236725527396589,
 0.2921891028876774,0.26549730328356247,0.765691679384941,0.613487950346788,
 0.0024356326214626245,0.633380834393152,0.8194028801357703,0.350472962771477,
 0.8461238847615272,0.30621200267544413,0.2261317075099626,0.6173927186431881,
 0.10464688923577314,0.7465218121584059,0.20692002123440356,
 0.5483855285974574,0.8771738025820204,0.25504572597914654,0.5825655320883876,
 0.2040481054087122,0.10106528139222615,0.8472889236150098,0.7647851799336043,
 0.10638663018904104,0.1392076679858002,0.4894381701841968,
 0.07481581205139343,0.845077312800436,0.8114295228739928,0.7245648323217523,
 0.04478140966420263,0.9652399307174686,0.1224261704918489,0.5593889754136285,
 0.010678766640470072,0.49082618454308824,0.4024433887390848,
 0.9777575940821502,0.34936274761202624,0.13474456421811376,
 0.2394026888538615,0.3694194474739423,0.8212704851536355,0.3478956613393518,
 0.5572228835458205,0.37341952135284573,0.6288022448769991,0.9368383147032692,
 0.7223562478496368,0.8913438428478669,0.38519665271247194,
 0.48603368178387185,0.7138819608413116,0.7724402780851635,0.5506747249097248,
 0.9466134127000988,0.6926060545297198,0.44146440083217586,0.7938568186286311,
 0.6765172668063227,0.4109149971568621,0.2525341131609923,0.6835642344606548,
 0.053325465064354094,0.34350021099140293,0.7940319839877246,
 0.5118333206005092,0.08468796250733024,0.041812524720647004,
 0.252576194282894,0.8925599533622881,0.5261001198318594,0.10650794419566656,
 0.494628732145973,0.6781878980758231,0.1205188328336605,0.445304048845076,
 0.6355615789263326,0.7377582229386609,0.5461999206497578,
 8.386418649263261e-4,0.47033190703309724,0.6876613645072222,
 0.7132386283358848,0.18299768633323787,0.5899459364144903,0.6412366698910117,
 0.6499850955277062,0.36166913602340733,0.19585626990607252,
 0.060664671350338595,0.41964122931650627,0.9938512519072823,
 0.19636479596487233,0.5424845628855498,0.10163693458589307,
 0.3123794208424038,0.32376553226463745,0.49612574410611765,
 0.09911136704332024,0.19901058609435662,0.2648909777289332,
 0.0795123183253924,0.3289381047643831,0.6124761461869661,0.2612798530248711,
 0.056834367720885015,0.11069342406641314,0.4743593431248022,
 0.3533318108871447,0.7873912378532665,0.9209075828816083,0.41838746580366903,
 0.1757850461385021,0.6004045735834063,0.4448438063436022,0.04528268783584133,
 0.9058250840507874,0.3544389138642051,0.21754422651111416,
 0.007392902329548812,0.12604337185183212,0.585106896972091,
 0.5335623156869602,0.8009163996071265,0.34033781696300947,0.7958349708197396,
 0.6397991283874108,0.42847693086998795,0.7015969302534945,0.6251717054449712,
 0.3314176576274882,0.8906736059397831,0.5209363340197402,0.6616951553488603,
 0.5233147352117432,0.5038634492512586,0.874259121318915,0.7847799581571038,
 0.4767292829536195,0.6955634046329521,0.9993905323301814,0.8910151787086126,
 0.14223996639457515,0.6911304325129833,0.7001565530564886,0.9023717103126002,
 0.547538992105127,0.4349824590081407,0.6245351876686884,0.2077928209615264,
 0.4123125013775546,0.18753353027230613,0.20601262283638078,
 0.32269750202887293,0.3749007429597785,0.9445293589111,0.5487078691211492,
 0.9357613063060694,0.2635990005145269,0.48843100432542375,0.1086155228268626,
 0.6473808304683368,0.9294722642103215,0.4914885483660325,0.9919847380838182,
 0.28282242239897537,0.22737364764325585,0.2352486771704685,
 0.5327097318222037,0.9258472207043839,0.5379635154897451,0.1957621656923757,
 0.2519817901098198,0.21754884700069832,0.28134743097050907,
 0.6063448827642935,0.7061362754363774,0.7889659668209342,0.9516172312384797,
 0.9856129501851437,0.8906321673422147,0.7304707906569816,0.4118191805214143,
 0.5312856579826,0.2724476451211453,0.8994718539966701,0.9045784601217443,
 0.2459435241512945,0.09309279850368823,0.3464226001316373,
 0.06822376751353842,0.6121237024771968,0.020223726439213996,
 0.3632033857280984,0.7120735755974225,0.8468427997244262,0.19106369821513647,
 0.4245351551499812,0.34845369445680907,0.13565606244588946,
 0.052252705668355715,0.6643316525854552,0.8318671288177995,
 0.4212466751918794,0.3819039687184369,0.5831665319588771,0.46858672313744354,
 0.07763260797160554,0.9054281507566918,0.2956527870161224,0.7918151683672955,
 0.8323930120410119,0.621516458265113,0.2928536839690915,0.09838711291209301,
 0.18360282022436247,0.19760156384153682,0.6117841467282339,
 0.7026010018416607,0.26808189861615084,0.31529820979243117,
 0.7247828051639376,0.1313713913611233,0.6147431937139168,0.5497135020925443,
 0.4678569793586882,0.25956356094764876,0.14058500076951885,
 0.7531855644077862,0.14665934097044608,0.2919920720540573,0.1181103539297117,
 0.6456595076677387,0.2059818366767634,0.45348790415085893,
 0.17925621576482875,0.32753309733518843,0.9859691222726936,
 0.4761035480320345,0.9559434278917641,0.04556276954783367,0.6270277644406818,
 0.0607029028611249,0.7455362841963994,0.727825335938753,0.3745840018022222,
 0.9238435272216465,0.22639166724165916,0.7257845340699227,0.7981248535544211,
 0.38911637496379026,0.8704903381197676,0.24976914323401556,
 0.34492731583796654,0.5759488818119614,0.8065381375557884,0.69985696556159,
 0.9290246577842334,0.7969706971798104,0.784780863512498,0.574564967818962,
 0.787350122100865,0.05978349114673687,0.28215061414619635,0.0810189542512465,
 0.09742503567534855,0.2857041135038223,0.07477681586255269,
 0.9794725466801502,0.7604121035120444,0.33719558155910534,0.9429240890475008,
 0.38032443433581786,0.1672080650911395,0.20638736025497495,
 0.30680603317708377,0.3388842228175025,0.1114820353244006,0.2978358334730691,
 0.5854343983055622,0.27754641594741225,0.5566345758372799,0.8921483855353164,
 0.8635123210314904,0.14469627015114894,0.9841841714404468,0.462183668775891,
 0.4426366243741472,0.4722402825402915,0.9341715046164688,
 0.022273037239714633,0.3352315043510915,0.659299800504326,
 0.10168187739080814,0.2995213224261384,0.5053412638108294,0.7806665006901202,
 0.3761374650516399,0.6937643277619916,0.9285830038655701,0.5059641498903475,
 0.39995799575219626,0.04749319007054831,0.5289249010919237,
 0.29083461137057165,0.9075556696813741,0.6418535262725632,0.4747721959177217,
 0.46397718935752974,0.43497756373580854,0.13527267558069123,
 0.5699988605907889,0.5478013535267144,0.8741229335952075,0.6199708588400109,
 0.6040557441444128,0.24096036037424928,0.0781514796678241,0.5012381006390574,
 0.8141954969710324,0.37135919877754164,0.4702568188285674,0.6566054485802874,
 0.5796940513538896,0.7675287660888641,0.05497521918489434,
 0.09115431762960624,0.3785045958296477,0.6436424383967668,0.931911595800016,
 0.21915549289267588,0.3862712709043715,0.5902371353853801,0.9921912904506077,
 0.6922186564735137,0.9945217352185971,0.2618247958229476,0.9919554741983792,
 0.796879610718123,0.5343907076408267,0.6716092973328606,0.987785591900425,
 0.7116894661605065,0.3252614680050646,0.25611556622202003,0.6677793334506559,
 0.16230963116100305,0.7990653400018813,0.4924781452576118,0.4526233194062207,
 0.5891666830777597,0.7159920788446793,0.14591191391895275,0.47583300040181,
 0.517741802168924,0.6652497829657101,0.5185688543583102,0.6884983485807307,
 0.2945465518648216,0.9923824908609692,0.196315847604414,0.8181086111440965,
 0.7116061006034513,0.9978815742126156,0.06645143934419306,
 0.46579968391138715,0.02613230855300419,0.050855960525622246,
 0.3758837018273826,0.8121806681076295,0.5388568456931818,0.5562254386109946,
 0.9861001496832545,0.6020478841966608,0.49778416255178115,
 0.07203977126266237,0.8981258836620617,0.59255415236966,0.6644325302837524,
 0.7366330465592008,0.9876230025648748,0.923234544626425,0.26052869613498497,
 0.511498938041501,0.3562614839095668,0.6609739213053086,0.6760704039572369,
 0.6610262732121132,0.9352507965498095,0.4268089544436531,0.8243123961095078,
 0.7156756205816486,0.11370173954621143,0.7585295928434655,0.7505299337987582,
 0.8374637334270869,0.1950668926595308,0.5936493237200183,0.10333183717609518,
 0.4815992391453111,0.9384675316309123,0.6229616274016075,0.19205609270121282,
 0.4051847003165443,0.9877378406935979,0.7121452848506153,0.563337708074731,
 0.7774381539310786,0.5510668547691896,0.5567734238088698,0.3685068454673305,
 0.4834215845855848,0.9656705785892395,0.11156016844932237,
 0.39439216500151486,0.48392811807398073,0.04055920066629892,
 0.08893000819284325,0.2868048165138105,0.7507420703413582,
 0.15960121950700867,0.058088298109897574,0.9486178644397053,
 0.855182382743759,0.2272693296363797,0.5782110772818105,0.5281182345763534,
 0.42066556691260903,0.7547380313055426,0.8626512519267738,0.8833434234105488,
 0.9980510086632939,0.6731346585471155,0.23677640200757033,0.5792876766669361,
 0.9785003068022043,0.6156293439599653,0.024901252928047235,
 0.28342526268032386,0.9753323022203567,0.7552373301471086,0.5075642898741443,
 0.7934167808743051,0.7944438134258079,0.4145599243125695,0.7830889729329857,
 0.8202048105273854,0.7164783036060255,0.008286348742204908,
 0.05226884717086966,0.5899780023329623,0.9397684928142384,0.6867974845002116,
 0.4530316759433015,0.26320288325071794,0.9090482080177935,0.6919405171382956,
 0.6717308585000697,0.9085348824087996,0.16338500934971467,0.3827822324914866,
 0.1944577124777539,0.3307575380595146,0.2838940871802279,0.4679554776449305,
 0.41657761118754555,0.4094681999384506,0.9428048691725698,0.7041868469811218,
 0.3962105742914609,0.44461095349079494,0.8952977198437277,0.3335121002518019,
 0.5275280314267973,0.14190893695398676,0.7085404253414739,0.6503082366962505,
 0.1798476359066703,0.1315952804643965,0.6038586027085925,0.6800925729513548,
 0.6212617633077722,0.9631886215690443,0.0284727688364389,0.9197089732217111,
 0.789019722880975,0.842528447503369,0.5097134824251742,0.5357391497622234,
 0.7279705901562408,0.25896576625209045,0.8319098558856325,0.7040682083707331,
 0.1319055025634358,0.9071101022842127,0.4516469374879941,0.7630106853147998,
 0.4952304607672331,0.09999177723260932,0.48338603584301354,
 0.6829850946819395,0.7910142424565944,0.30362826208322236,0.032004032938176,
 0.8037942244308879,0.7952069432063283,0.6072509207921499,
 0.017847447795347526,0.7548336707752898,0.2876172468485154,
 0.7376516333102832,0.45552899416628,0.6981701857874597,0.8849915906725841,
 0.10281241590491852,0.49153183969438985,0.1783571413660039,
 0.21348140374256142,0.18968695692099302,0.227083794475903,0.5190894858703964,
 0.1594907893562637,0.980073991758699,0.3829202515342768,0.026281106813298094,
 0.10257018262959938,0.3865109598903178,0.9934724683883115,
 0.29228177128382593,0.3900774756651255,0.40538432489685894,
 0.9956883104937819,0.8250714479820804,0.8174594290393052,0.2804062638192124,
 0.9127550367906248,0.886243196443539,0.18874060138404025,0.5768265222335194,
 0.5496329646797887,0.531563769610917,0.521503126064241,0.995263621920938,
 0.4649599768453342,0.5153589222983244,0.5458794615095892,0.5794844190829935,
 0.6781356304952456,0.8722058217004791,0.08116866583969862,0.7365494919295423,
 0.26880197888372037,0.4232241546388499,0.9686392508215289,0.7301375586895936,
 0.15084597334251248,0.6674922615319122,0.3394726917554388,0.7033963482578184,
 0.7843321919302202,0.14301434922330636,0.29791563914232166,
 0.2289455153660177,0.12066960122540293,0.8977429103045893,0.4513945538243307,
 0.3940307466657317,0.9596922351489716,0.6679907962910521,0.06535198469068726,
 0.8739393565482968,0.4260377969208313,0.6674173997718899,0.7922409741566674,
 0.3830816044982218,0.6701008066811066,0.941486057362061,0.18463596288317974,
 0.32350042231478093,0.9615895088588616,0.9140022194123274,0.5561399429231254,
 0.9231861318021796,0.6158496289135396,0.5095743416314923,0.6607458369863226,
 0.6859963476257712,0.6359218159049503,0.6972222157738499,0.12975648940493878,
 0.8161035475398705,0.5969813975163855,0.8172511466861583,
 0.004799926037551638,0.4228512610500549,0.26645097815731056,
 0.7158966768709861,0.3101331498655244,0.957111947791601,0.5663395350947112,
 0.3702280900073419,0.3726476611955747,0.45991893012667073,0.7271308545801372,
 0.8842923816584334,0.8713562966809248,0.41435401631966606,
 0.04255197364766927,0.3860606402265604,0.7321613556997923,
 0.015612528846257634,0.4527582127329487,0.35869074217738195,
 0.9774115254376101,0.8303354740717651,0.03800828902372222,0.7524740663128195,
 0.5098704743230671,0.7497820741062338,0.3180171341182485,0.4494162382235597,
 0.7264297731847639,0.16746572753638433,0.3103249863403631,0.5471195228616121,
 0.14884397325499854,0.903498421720868,0.35394454585789337,0.8152684008194365,
 0.7642191860837344,0.8438131826408929,0.38551267676210255,0.3996990674344997,
 0.033406760444868566,0.36825918273463665,0.2006965013685551,
 0.8974475305825288,0.45939301178303504,0.31555473457282834,
 0.03667081839178343,0.25675780285102234,0.6276763550402452,
 0.18693290440993038,0.9611809073034001,0.3989682913967262,0.7266626871595695,
 0.9675399864933139,0.2382714336727939,0.11245735067518425,
 0.041626758852761125,0.35480661457663454,0.014702699943182185,
 0.6005960150325956,0.912373310517744,0.376992183331308,0.4732968391088095,
 0.6494228943019235,0.0612466517490966,0.5896377643184542,0.8918165127215848,
 0.7300994584472776,0.11539293576678811,0.17179258209082537,
 0.7232815905718473,0.01193733719732648,0.6928957012861715,0.5698734691906481,
 0.6710487799314405,0.9786647358922158,0.6605966923477397,0.03877635382946121,
 0.6442767603511271,0.47620009609235847,0.6716298324712213,
 0.44810818764540805,0.7357363945496771,0.18010182933155827,
 0.8726408914512637,0.20745261743456345,0.08588899978185449,
 0.03901118759142941,0.13764710282370796,0.08220662401024392,
 0.38636039574235326,0.7030575082281367,0.443282332390624,0.35131509504096814,
 0.9566448281332246,0.9842741870738941,0.05425890713656345,0.381762068658152,
 0.26668281614394684,0.192142195012442,0.25095992128624867,0.8842519298485372,
 0.0147721100439977,0.37242840274477684,0.610211000086981,0.653672682608579,
 0.6797076529651711,0.8896842186630374,0.6749249174664684,0.5558285086123974,
 0.9039578328633577,0.47618636557091354,0.6322031782083378,0.606935690184773,
 0.7281036440593819,0.7080341133939814,0.41001555122503863,
 0.20168002776170868,0.46770328424068963,0.29250947603558397,
 0.16992477819553342,0.9299022720858429,0.7464124550536778,
 0.30426351677763375,0.35784147584214454,0.1326214742739683,
 0.12352085214900121,0.0349994451833372,0.6815227813884706,0.5207723943669849,
 0.7301555410168932,0.5905782580945477,0.8659525810288399,0.12559819431490016,
 0.4346284732684651,0.42915705485026623,0.06018751227420438,
 0.1549254439733183,0.24182223127571978,0.5216420831730989,0.6065907808709098,
 0.720623928730175,0.0755871863283335,0.2773436817434699,0.5520341062676088,
 0.18862832825059472,0.6633686925254091,0.5418108198949886,0.5746942324406036,
 0.820166602897941,0.7707824494530184,0.016607581301693597,
 0.35382021205219594,0.2447797762262378,0.3582512176421633,
 0.36370343949001305,0.6175724389511242,0.39899853950784536,
 0.3344638053831187,0.9572918945948344,0.7441747090616604,0.6514864502372597,
 0.29064326572722365,0.11975806163296898,0.6620637497431295,
 0.7372824434282936,0.9807683354687163,0.19740891727411214,
 0.09786678988180741,0.9508310214514042,0.7708157752131374,
 0.0051734293415467425,0.27748448170050843,0.26811482490940974,
 0.3738955411539565,0.5970448060613575,0.8691518770091455,0.49640859285230654,
 0.04009001791009725,0.6581690217567522,0.9000892538570463,0.5262867135479412,
 0.19921927070735035,0.3545958925063266,0.34434725088775764,
 0.5349109376906351,0.4383825488263835,0.5024807190962043,0.7299185802991047,
 0.4939261604350791,0.1189737192798288,0.43605676567757345,0.8646358794315515,
 0.7023525163329579,0.658719547704084,0.43643594120115714,0.6336159043020391,
 0.44790869295333846,0.6988080243849251,0.9803163786163502,
 0.31022389578158593,0.6328828535977551,0.8555104588157673,0.214503142104419,
 0.4268483527356921,0.6032648326419059,0.587721620501787,0.2844854649527214,
 0.3744101101293461,0.7693689749315671,0.25103056877855867,
 0.40735955996942064,0.8987831425110494,0.534354676418369,0.4671718378635341,
 0.632558359395685,0.1710908999433136,0.5785143211317683,0.7721285636710326,
 0.07890989956355232,0.8080841030180554,0.8307416146965261,0.7747561879474323,
 0.7026675248831298,0.6781277379255083,0.6145100103512176,0.6793881024193517,
 0.8032864238718376,0.03661947176105873,0.36942866588721324,
 0.4332116421335963,0.26172739989484084,0.7189530717152154,0.6105349598808966,
 0.611741915500415,0.6661449745022961,0.3720339224270046,0.39705839656954733,
 0.7477484628437854,0.20987903088802207,0.09763875609583761,
 0.7809968476822616,0.49077554698535986,0.20257846880018193,
 0.11902298529294153,0.3153003538993814,0.7942063954835956,0.7353845603837923,
 0.9580646594774838,0.2828076652096494,0.29337401161830357,
 0.04573334052281108,0.4925783325303629,0.662019838865677,0.8718618838022026,
 0.3111375405262057,0.3214275521686102,0.2016960972927223,0.3161384703531491,
 0.953172633549461,0.2268003697951878,0.45791798353724544,0.9350794418951258,
 0.32607244794106205,0.6734875565075316,0.15305880088702362,
 0.04866369210788646,0.3156784097042914,0.5118762285385441,0.5457224198155666,
 0.26781011536484867,0.3218783217424458,0.599396027663375,0.6314823433914993,
 0.5064185343774453,0.967028864145204,0.8185481604091536,0.6521279181743596,
 0.4921904662579907,0.26657954633196435,0.8935359997213783,0.5404320289285068,
 0.30606523559716725,0.4455903303887083,0.6768488661199515,0.7476631760262288,
 0.5442669049421518,0.3905997129457903,0.7775390233309554,0.7345046703672464,
 0.5197919546731778,0.3100868080618091,0.36762389874869894,
 0.36766159343680316,0.09918607562401593,0.5635336520647963,
 0.9394600746763287,0.656452853151757,0.4272011699642111,0.17960789392324716,
 0.18462363163971118,0.586695100572078,0.734319197100732,0.6518868397236147,
 0.7818479674809433,0.2600282296591031,0.7972183276107526,0.14247723263146783,
 0.2115235902997521,0.9994818113012347,0.734934841852041,0.8705341308664116,
 0.34301981660232406,0.723905946061697,0.7231883334466063,0.667924443182084,
 0.026557534759521495,0.09596128587945041,0.8136491701894109,
 0.1359599974179878,0.07914590576295988,0.23495081777944993,
 0.2840060905226318,0.6870248062284573,0.12315346192166976,0.3311885465552219,
 0.16763150047769826,0.7058852395375004,0.3865789817116272,
 6.028589049235089e-4,0.3786092001896937,0.31622678186652176,
 0.6994537705717092,0.9029418976411219,0.6732341521852379,0.7370359904639338,
 0.9193562455859484,0.4822114240079428,0.10455968518131542,0.8771341461324673,
 0.5730951967252101,0.38722843814682495,0.08204465033731245,
 0.1064739225138639,0.8454012583432511,0.9382217816478263,0.20393335118299794,
 0.7622027685789515,0.18172190640447783,0.7857919752841843,
 0.20785257201240948,0.9459496019992581,0.24395024368561147,
 0.47691870176725837,0.23375710435393537,0.9089500068729071,
 0.22317128982058698,0.3993051529130275,0.4651806461444067,0.5064753815254818,
 0.5514335261787296,0.46738573256842864,0.50754589639029,0.30310703767825753,
 0.8136168131169369,0.5568974685928598,0.0727327086156977,0.19970520082226084,
 0.9179958945566599,0.5287994240341637,0.340922662730062,0.08461096791455014,
 0.2031716054399373,0.7764714504459951,0.584911153494813,0.8119501630095654,
 0.8744825553848747,0.2525421265606962,0.802699889070884,0.7870158880764929,
 0.7149559380423413,0.5658214428039597,0.583253714931737,0.9171676939363683,
 0.8809622985167306,0.9717992677862877,0.10426063740528024,0.5856816108975988,
 0.4537840061204903,0.6165674703308214,0.2991002076685998,0.08623332126131455,
 0.009799320653997579,0.8229322608204657,0.9645443414887108,
 0.06089348732854871,0.25704379426982193,0.3305741556656101,
 0.054841574057791775,0.2213976826863885,0.0573219977028272,
 0.19696662866387082,0.24099538344939964,0.20303231625691387,
 0.7081601157348208,0.7169894782150416,0.7927710164290758,0.6672355397484884,
 0.6750467570245424,0.9159339358196534,0.8966504813343819,0.6036951281019076,
 0.12098713780200998,0.16435396581956763,0.43607006013746163,
 0.18338112011991559,0.3685046578320865,0.8759125869183447,0.8798687927593852,
 0.6690211560758478,0.28187500472742044,0.1752906983241861,0.7250414603975562,
 0.2315272093793853,0.030514045200300433,0.9087224863894559,
 0.48253844946725666,0.8789487940897842,0.6322888436213935,0.6111577392798118,
 0.7672880741496173,0.19482270399367196,0.8535562811247808,
 0.32979127214573767,0.34578614929431306,0.8269764410939379,
 0.4295031233722413,0.5740120456714195,0.7780484639841596,0.5044388378697845,
 0.2799267722365185,0.290124093554604,0.4794646851167228,0.3192551094419094,
 0.3889727418940807,0.5479083657056437,0.8655898840308368,0.5561988360468205,
 0.30481724987078573,0.6295368706947052,0.6982097265229137,
 0.058874829752954994,0.6770688378310862,0.9858858370381808,
 0.12136122225034474,0.5936529988920766,0.8196569912229718,
 0.14165535876516078,0.7463929834774092,0.8641817288741096,
 0.045758604043083406,0.7076041744395152,0.15251198634326057,
 0.9119115809309242,0.2018085686354678,0.8828282872410688,0.9528121510237452,
 0.2545202223371408,0.6156535857990263,0.6887772993536441,0.6404126727892124,
 0.38329793786460475,0.09170985373488216,0.015496422474671645,
 0.2742433259083903,0.16133348390315327,0.58245999524331,0.3637783362877567,
 0.7418218483845386,0.6562387475643668,0.895260524697177,0.47772626164879295,
 0.4679417007281549,0.38638933203711745,0.1999001193828912,0.92870593936591,
 0.7013936892788611,0.20548620775783089,0.27931043584464,0.6561396352943681,
 0.41655567213738975,0.40165896113160515,0.23693010317918117,
 0.7113294023940272,0.7682035370106016,0.8636332977296441,0.5123291978159508,
 0.6350982186910493,0.8045799059656569,0.7912712763822481,0.8631243041921313,
 0.5471666160030061,0.18140232633392694,0.6557177452516719,
 0.15218260238973214,0.30188670143739493,0.6172503480459096,
 0.46696908376748214,0.9284966540736438,0.9153592318999152,0.8216254798664018,
 0.7764056219370391,0.9780382814584636,0.05565225137221519,
 0.16732581529709556,0.519838572376479,0.40291464222492457,0.8105138313862534,
 0.061472633778712926,0.20195556473160547,0.9935958413782554,
 0.4058157741605619,0.2689652474855254,0.6282666562045757,0.04165056685116686,
 0.7581192909140706,0.28286357716106914,0.3651608065280748,0.5867560253895313,
 0.6789432929265939,0.775940788871767,0.980085069317777,0.1723921971245045,
 0.03843313571859397,0.7309302935633523,0.3637645067122447,
 0.10092374639700841,0.6693821640848107,0.45764054415036226,
 0.2699382870892205,0.6053446563220208,0.3148506958851788,0.3711640550419175,
 0.6872166338285461,0.5441229009946362,0.03325968293413051,0.5194294011705097,
 0.4500709191371174,0.7588725575879913,0.10519734310136675,0.1037367767804499,
 0.3626645033683985,0.7086480909870784,0.8029029674715552,0.1756332065061419,
 0.5463893912262168,0.6550324083400356,0.8032081202812668,0.3622840132622046,
 0.25551948504825606,0.8431734818354654,0.8645593924486784,0.6708666266191222,
 0.12923816191875337,0.01921575355188654,0.432785716020867,0.1219571065228846,
 0.6562209372535429,0.720849665216826,0.7892141390915838,0.2720201636611703,
 0.05443074617772037,0.24990672093459165,0.8084829194323269,
 0.4639887568353641,0.9915030760013003,0.43163018894676286,0.5903346350898332,
 0.8515488081936288,0.5952037320232331,0.1151732845169463,0.1580126669627735,
 0.5525180916321327,0.4805434877577093,0.2668521153299337,0.49332390974437423,
 0.5002693347354927,0.10059666938675005,0.8320237056377138,0.7355664732838043,
 0.01637160106630564,0.9571098330308095,0.570049515962579,0.9687424153472356,
 0.7316627595566457,0.11897719424664088,0.16529002508863455,
 0.9841473613014438,0.8367071791055394,0.4428063360917499,0.30853515797702735,
 0.18541885292813487,0.017257821729090406,0.6413902441725666,
 0.41340881051587464,0.767235731823559,0.442363350451226,0.32421982754223555,
 0.6743156023411927,0.9996770297084926,0.14692094229732966,0.8381399189479196,
 0.9430973332561348,0.03342970034855053,0.7626281694216318,0.3494873363770641,
 0.1564095286509224,0.7346707905834556,0.04384417975075516,
 0.03815695989508716,0.34961545203934885,0.9654684880035598,
 0.9173709866080498,0.604582734568792,0.5558150493332726,0.3752571911663318,
 0.7057101491894355,0.9070718437995631,0.9456600022956688,0.4671042666350992,
 0.9433274178000403,0.6109233406167379,0.9814933655454173,0.9828946166931709,
 0.5074919704307286,0.8426860528007077,0.41972920516538736,
 0.15181427272036663,0.9732797402945292,0.14913578213984469,
 0.9279964439340295,0.15601270577823845,0.9601813923351596,0.9638656165905818,
 0.562593422531793,0.02902953223624094,0.11483259333768125,0.3138701587339737,
 0.0923580961555246,0.6612336729949726,0.4115409397020239,0.14172568563385668,
 0.10387268926167825,0.740064063138594,0.7824535450852294,0.6107039372149111,
 0.7259387193187955,0.9313450736574486,0.7560140172751454,0.7887025042833812,
 0.8357550359756556,0.2423630547039486,0.13344016824151217,0.2780786034029352,
 0.028477895088463434,0.487380928126358,0.15738815822140173,
 0.041420884592553775,0.580305667461027,0.20151141766320668,
 0.19614215081368025,0.06859801721098646,0.5705766426184707,
 0.3475022788014439,0.7992831690551895,0.5515659436114545,0.5190173646996689,
 0.16922472806016053,0.9425525579220319,0.4100530571499783,0.6572299291550979,
 0.8999622122535523,0.8883659269311956,0.029188706999558778,
 0.5736195374851547,0.9688375557494068,0.5341777692032492,0.47423547813651035,
 0.767514931130399,0.556121820175077,0.1515928101248215,0.1417449095864105,
 0.2738819528532024,0.14761506341350494,0.1204166259497379,0.2810700556085758,
 0.7960159268856877,0.8964844724293296,0.6611772168218943,0.3481008511386372,
 0.7687741765890702,0.9227731765975073,0.5024523256120094,0.7155924554680205,
 0.9001800101531628,0.19055148343571437,0.07009047908898858,
 0.4600915380816504,0.0044856308721719085,0.1332907702491588,
 0.18279619926778656,0.9353655443818857,0.8933771752907744,0.6055618182614012,
 0.10387544934900217,0.7807483814765285,0.47564384281836125,
 0.5428994474669505,0.6068178282926423,0.060656009758311646,
 0.42584028398800233,0.105618281366882,0.5254876195589576,0.19901824803955348,
 0.9179918374200111,0.8198000353421033,0.9383299751245993,0.8846982251539242,
 0.2847470286870454,0.8742789243697883,0.7147660244369701,0.34836789380654243,
 0.85802488360489,0.24332205645867022,0.2641295855376866,0.03402996569173222,
 0.08507285429875522,0.2850305961718411,0.030084372404867166,
 0.44687320547681453,0.07912103469075393,0.840099822719952,0.5917302668393918,
 0.6446220006668557,0.21479519894993893,0.5372648128411801,0.8653311953247372,
 0.6756789090629293,0.5993186968995345,0.9243093149557069,0.6146831100695292,
 0.8717518351558989,0.5933710187542659,0.7578431530007301,0.7103472848414035,
 0.5887967722523074,0.02901473366779761,0.45514684453750365,
 0.9574539914768185,0.21780904437876203,0.6213656831820498,0.7193987628506463,
 0.8124713024318867,0.09896713145608149,0.486366776188333,0.04503234697917724,
 0.4044867666359141,0.5099435441241719,0.42227875347549526,0.558061729880304,
 0.764517077125302,0.5643858223682281,0.8053421748031544,0.2156875477610386,
 0.23947289185397336,0.309253696475542,0.39836495783267356,0.6874211490850801,
 0.08324525003788985,0.3063704048799071,0.6422645171036333,0.9129974757058505,
 0.21175387819701352,0.42287109112101795,0.7067200623905144,
 0.03383422765933042,0.9990023020595119,0.574609458534491,0.35234605925390683,
 0.6098136429889746,0.7284318995890645,0.9417950580290108,0.3426276021548682,
 0.5374803506227576,0.474670593170867,0.554029565054783,0.4538866029202373,
 0.8085463104076651,0.5581239445173863,0.043131194272713724,
 0.8418384323380792,0.5049347476838363,0.2162445028992538,0.9463192705464629,
 0.45393718953486284,0.8329505189648344,0.9174550619161072,0.2079048152413282,
 0.8518866024746394,0.6500270264005898,0.447465996875705,0.619128604121142,
 0.5705847755852221,0.4420959531636396,0.2946271968656008,0.38215921811829334,
 0.46681157259018746,0.7966939252798296,0.6445352035869941,0.6113397357393427,
 0.5889756161720369,0.20282519607940563,0.7690601636565255,0.7676391143976726,
 0.48513551260358423,0.7505080332661159,0.3473589479943573,
 0.10249891590226823,0.4707509502183518,0.23692734016823214,
 0.4166949655660537,0.850282452670167,0.16655953459177164,0.19593433277953975,
 0.4099446688319107,0.6912806435062153,0.93612325718735,0.8745906977076655,
 0.2478929943846675,0.44380776692642254,0.06084254512044629,
 0.8504378642255344,0.06785489761151042,0.3586947306484405,
 0.38058300456148686,0.43515352064429014,0.307144073858191,
 0.37113348576241945,0.066916459955769,0.17737417154057555,0.3904796867210154,
 0.6349027739114698,0.03684056277026426,0.937435062365297,0.16740283467930184,
 0.016435819753862213,0.4243655918756265,0.1389622442855939,
 0.2456412831065935,0.2753541169804625,0.5471867459896892,0.43153503870779863,
 0.04064660317076585,0.7456498956715392,0.39638789538785313,
 0.0714986427155272,0.7143093604400539,0.8318106564588863,0.9375731464674153,
 0.9244439778089721,0.32256469405738586,0.16617401133515486,
 0.3335476064461488,0.7893369583322908,0.9104227176260407,0.936566409805608,
 0.714142356829685,0.4174107060451081,0.1839923773699872,0.9308255922254118,
 0.0711785248464557,0.26907401937348285,0.36299741996084534,
 0.5850900609781767,0.8755564278484629,0.6403432379779267,0.41622776819570584,
 0.1401243813693296,0.776440490885359,0.23506243231238932,0.7999499066354283,
 0.5970689994848458,0.43176987354959584,0.030361453833807284,
 0.20009805811353554,0.20484744321893944,0.5287098883222823,
 0.03274104540543776,0.7361130761726441,0.2559212794605842,
 0.28422664774916007,0.8770043053510783,0.573630304974513,0.17119604349067397,
 0.48881895614425963,0.3446738612180221,0.7527085465716056,0.5577384790313201,
 0.056303419990590076,0.3979606906027551,0.2745860802315533,
 0.1604377706195821,0.705783367644601,0.3995936951206275,0.13752001876116815,
 0.26790528949960835,0.7719799193953623,0.2754902487964025,
 0.006801007350592747,0.570678961366925,0.3897227228548139,0.7258541898921073,
 0.6153671759048879,0.7317701736609475,0.3615554304418699,0.7288624208653505,
 0.285913458601061,0.10265116477457814,0.14781900774320567,
 0.30754458432119214,0.2282291637598458,0.6938175281940491,0.8235959497023109,
 0.3487076238844018,0.9411066390727265,0.3491352553417517,0.17029236802711878,
 0.37810130408898146,0.4454444857186325,0.6075221288064109,0.8756833806129037,
 0.20688378594260537,0.030401950830272195,0.7401876915825975,
 0.32667953972918495,0.27807888375479295,0.26109180791542386,
 0.5214098767783153,0.650560830532751,0.05022741447128187,0.1499888634113098,
 0.725886423896748,0.09904913153848016,0.5062701383502792,
 0.016416136906073242,0.8002217364377917,0.5799919948694888,0.576497195434564,
 0.38702659248958793,0.636833829971634,0.9393880471910728,0.4249934483940384,
 0.6912030336010117,0.19670362640433425,0.2905026369127337,0.6479897795885001,
 0.5752985730290795,0.2779153064137393,0.8534745021712439,0.40923354310920534,
 0.31924847159697867,0.28110846315675597,0.06373833308379473,
 0.10067239214939827,0.06597430369608981,0.15463672163808273,
 0.8739821430159334,0.9945292888861431,0.10057845876568061,0.2840900313940201,
 0.5983844912188827,0.09472112655181308,0.34747889245699015,
 0.20850886654202816,0.9943943457888197,0.17893203647989075,
 0.6100479409532372,0.9158760547895912,0.741931040485964,0.1678595954793214,
 0.8597667261260842,0.5720332246962001,0.24165809895156098,
 0.24019759891147563,0.6608942106632446,0.4299138229201217,0.5229456922285337,
 0.5668269072689044,0.47506895640571656,0.23944067619829235,
 0.2002807238421257,0.7770879686205374,0.3261803846030995,
 0.023905937612655137,0.9160617172415851,0.745895910327037,
 0.09395592406250919,0.9451647132586827,0.5463004211569161,
 0.17264815051003313,0.9082422791212373,0.03628166472810468,
 0.25736300908445386,0.014409084257125393,0.5427889633994758,
 0.15032456709429887,0.32131626178552897,0.7892057774245057,
 0.8348320308397201,0.9846848083892246,0.8394885017232252,0.14697909067073756,
 0.05369326619465575,0.2065652256206173,0.24199792176809165,
 0.7228019270567524,0.8864491799626263,0.5914464008899126,0.7559081758193759,
 0.5477136933990793,0.789957486746027,0.7936271266542783,0.1168915031208152,
 0.6447360087926725,0.27383303764370603,0.1345059708872347,0.4297047230491736,
 0.8281067173466381,0.8443697919727339,0.7854519400760575,0.43124011822999275,
 0.9907423533886394,0.9353485753232348,0.3130375158995986,0.9359229646952053,
 0.035969246763249973,0.3527175608234845,0.6032198937941008,
 0.4990352012697038,0.4245461585778806,0.7773503060150575,0.5746528820479492,
 0.7174347553996784,0.6667823198509009,0.6293069266633858,0.40903530379304165,
 0.6288686195399844,0.8622215825610993,0.9040095223680653,0.8989838338995856,
 0.28551233598249126,0.9612643341722598,0.4035975571250294,0.1275305415814686,
 0.6738768438442361,0.5106453387738089,0.642946228076756,0.11322912514271277,
 0.02402019266672195,0.30140838162238115,0.03891426438711598,
 0.9183212519628694,0.8572602028901775,0.3180702558730194,0.326342918585692,
 0.6063600083361956,0.6258435828471424,0.4850168001947166,0.31750929094519487,
 0.1177257713950276,0.6500668299799455,0.48746201380167864,0.5006555395630923,
 0.02762323502405517,0.17274077934058152,0.61796963089686,0.5214123745226282,
 0.3083789722108027,0.386162653091477,0.39119851143006124,0.04044281723699572,
 0.7827487251394536,0.32791642544897215,0.32201209181143564,
 0.31029113383789686,0.8385965861846286,0.1169630435315443,0.5997640829013924,
 0.26647890316749423,0.04797114640914257,0.8964167997832995,
 0.06974105883050918,0.12224691913433139,0.2640851899966332,
 0.2963932780808477,0.46070824127400345,0.8458448900039985,0.4270699036980439,
 0.9670280422200035,0.6427282910720155,0.04930006599267689,
 0.38142568609340566,0.28566689776428733,0.9846344069846029,
 0.2575057923797841,0.5782623113120512,0.550900660950722,0.017646765594804492,
 0.4818954134628911,0.16229958679556045,0.5096858311570129,
 0.42227472027936686,0.1534378613136833,0.7110470290525437,
 0.22166125083647648,0.015148290967065603,0.5975977549311268,
 0.32853147857802323,0.5689582656360032,0.9072360131265774,
 0.16175479176144347,0.40153935888991366,0.2114167719049822,
 0.11987436454039413,0.2837091078228876,0.9690744356540264,0.903225967234867,
 0.6923031076945012,0.4466135762566912,0.42093380916771106,
 0.44599047508285505,0.907187475608335,0.40933106888554016,
 0.08815090397619996,0.4339322137399665,0.7254080958859066,
 0.08168752411634861,0.31381454970006617,0.47743829352520073,
 0.3773062680315675,0.8766894796017005,0.7772608687899738,0.6735645636189104,
 0.11902525738525649,0.7143972713433798,0.17744251306557446,
 0.3881987517391934,0.16803784572078773,0.7315017751755799,
 0.17263338591586197,0.9534069062881675,0.582156867010845,0.17441093884801995,
 0.8042479719525214,0.15775942302990342,0.703916591544267,0.8141846997066902,
 0.6795072547688008,0.20144437508026736,0.9350658355421748,
 0.22799625913542532,0.7735147619872558,0.9124473911587269,0.7647275339609443,
 0.40701284397883153,0.7910404411554551,0.9619394692816527,
 0.20963609327397337,0.24565369480684718,0.3977823193283826,
 0.7521458412911143,0.38554827303838857,0.40896715396986005,
 0.18682983689572707,0.9248993809872605,0.927998985218304,0.4462152988162661,
 0.9757981741407538,0.23869193468802574,0.17890867728369508,
 0.9195140494196007,0.4313292583831634,0.35962909178908253,0.9108242375170352,
 0.7917431653558484,0.1890776544406716,0.2902549763677108,
 0.008480010731564569,0.26102825868316015,0.6055845505388555,
 0.5116538611513001,0.9534790546530135,0.7032739441678797,0.2432310214942962,
 0.2920555061479916,0.307943419627121,0.38830646790698553,0.4195539056833839,
 0.6411006911338744,0.11216710766595828,0.13008402134844466,
 0.7491742081934191,0.8927379724540889,0.9985754300611555,0.1211776508295167,
 0.7067765176405765,0.8194021881018211,0.5961426565902337,0.2809968365820873,
 0.26280217963199204,0.10622012166068373,0.5269575297678979,
 0.3431329785574442,0.806473524574702,0.6481347885647338,0.635490538223775,
 0.24750677693689271,0.95460543857866,0.8098167541513129,0.7651814392379424,
 0.7259742327566512,0.8073101906706421,0.9965918369841936,0.05849744403813606,
 0.8368605269259214,0.3874386419603649,0.633793382093115,0.16991720938099564,
 0.8035424007424227,0.7433422716002919,0.5645688788689939,0.5977656152597988,
 0.4198788136823638,0.7589326589486611,0.01400064929143885,
 0.26581569505227587,0.08848552675103849,0.40168710225343573,
 0.7230695522214209,0.4713824893640838,0.3960028594521211,0.3764484749699134,
 0.4869975666632841,0.2740135615216589,0.6680902576905354,0.4877255933874234,
 0.0010565219001350545,0.8153074430336565,0.33684706198980163,
 0.01312251367770001,0.6595731496788897,0.9603193640378394,0.6274858008191289,
 0.6003342735590512,0.7483786388898518,0.16287717679985136,
 0.23511568526735704,0.44975476039172646,0.33172883562806,
 0.011962060423735599,0.8439291752965562,0.8058339467177912,
 0.8174424996034368,0.4942026676356778,0.5614404443938773,0.18235028844631707,
 2.619730164890033e-4,0.4094015512253324,0.2564020219217853,
 0.44909791770370733,0.44417613663935474,0.21473848448787325,
 0.01127166734681595,0.1959931458682287,0.0014507993684800269,
 0.43935291329995274,0.28808846505132,0.5968815472046952,0.2638005432839189,
 0.1931702457386154,0.8725263170687645,0.45502465193191843,
 0.29070687744284596,0.5195392269882997,0.2741525451641226,
 0.45599121535277054,0.17412178638466902,0.21510061477394404,
 0.20611828082604688,0.8265165008251939,0.4783464853763769,0.2330452360006129,
 0.7186713391388426,0.23651673609720159,0.9612595971163704,
 0.08063750437127659,0.41453883057090224,0.7516440319087467,0.523555009606908,
 0.7268293401244261,0.6832418118657131,0.8603228633400053,
 0.028797912277214843,0.2428054585260938,0.7523336989790904,
 0.14012843088759175,0.8109948681619261,0.9119795660263916,0.2643244803251745,
 0.5035126257024976,0.29026487511909904,0.22770824891858898,
 0.24448651913728603,0.13822850657378727,0.19531122627432018,
 0.7412752564500693,0.06514067979893268,0.5697148659382864,
 0.47486538870011785,0.1119094000993538,0.9966783017968636,0.8456707667064971,
 0.022320583074983746,0.8883112857313226,0.692759804498155,0.5150763176189231,
 0.13555477621151846,0.5051127013952492,0.9064711778880906,
 0.16453275785207988,0.18924958714489393,0.01591797723612376,
 0.17297948869548452,0.5276532973311456,0.5688033724149058,0.2435321584461767,
 0.4895286420328848,0.6886427471022756,0.7334116624794678,0.8734062861793558,
 0.5977855295525188,0.08977620384847673,0.4254276937406284,
 0.09228232411184822,0.9451460339087426,0.024649589081894296,
 0.6898338875586043,0.110385039650714,0.04235809800742829,0.1952427041188698,
 0.04454874807524867,0.5695664961056633,0.30875407009150846,
 0.19439753658797798,0.859137570399709,0.9447705056149429,0.6897064722391392,
 0.42401741526464565,0.6872947038924897,0.8100343925121102,0.597207576543104,
 0.7128186087096282,0.5407002040028813,0.29930437543507094,0.5794728198597161,
 0.13713863015743188,0.7786405440684961,0.7640362472625781,0.5034410721294045,
 0.7979380347221777,0.8203437091989605,0.0026225270801154554,
 0.03362371611992809,0.705000643286799,0.8414250946526477,0.9948703577985447,
 0.5037223388315435,0.6946266294975691,0.719758355608125,0.7837992229335402,
 0.4245895920184066,0.8598854294132926,0.6900050499112961,0.32348420167379255,
 0.07217197934556618,0.9071803657663833,0.6337481874611233,
 0.32780055013777065,0.4964125934391791,0.8377061272964218,
 0.37007124898234245,0.06844641391696138,0.769165098526576,0.7031330347253767,
 0.6886977578137022,0.9899267926850688,0.428304737687101,0.45815646745411,
 0.03452003227920275,0.49348101100040775,0.004402236440166685,
 0.7145827240775372,0.8667886921765408,0.42575025407644473,0.8474931351251942,
 0.21897386488571147,0.41449209507992424,0.9304886854020691,
 0.4207599189793262,0.09749957029794354,0.30789669788996443,
 0.4165408336904236,0.8727431792521325,0.8619065682928899,0.8134060101311927,
 0.3848309876269904,0.803406041282617,0.9354100488551201,0.27419601107355185,
 0.7316106054247183,0.517098613136431,0.31872617526746705,0.5672070047203484,
 0.6360907358558694,0.41565366716862717,0.262931608783848,0.13707656593848117,
 0.7164490471444691,0.8116131112940864,0.059331792817001294,
 0.7454967620932977,0.07827592108275105,0.2799387142005527,0.7430951755047922,
 0.0860921891422809,0.9417295632044219,0.7037410886345946,0.7782895928120729,
 0.7421370309329032,0.09710829290205081,0.3209625009363667,0.8480530167925263,
 0.9385903502625285,0.2103051750313446,0.31170938496774725,0.7031405728370232,
 0.05749339510555873,0.5753114612334058,0.07143236856062196,
 0.026332011276734568,0.9181852180371081,0.6140395310682237,
 0.33750104658668334,0.003818913946886493,0.03655046352043745,
 0.1658432690716416,0.5540997858573398,0.9962176830133834,0.9798827923766512,
 0.1626807583853146,0.8992016979609568,0.08842605232208012,
 0.05262399942103957,0.21532187381437284,0.23847881808357285,
 0.12273536032703714,0.24036605393616262,0.7027545297110637,0.982682482775684,
 0.9307200415297362,0.44594602679477946,0.6625828364153923,
 0.02722348109887429,0.9104645944008696,0.9311579545315049,
 0.20069287143862358,0.5313181746925836,0.41973487523192266,
 0.1623352313262817,0.12939587269739716,0.6447266710503543,0.6589934344185177,
 0.559059102948931,0.27658065842953905,0.3955385250754302,0.17548410549806948,
 0.9481003429919412,0.32528021472533475,0.2317954653465334,0.4219469565944426,
 0.11064834163664228,0.8204723060044881,0.6598518248914707,0.567413427909438,
 0.9909493006754987,0.6190706736559823,0.6544385866660639,0.4077164081157052,
 0.09108721725813573,0.48578926970982905,0.0023585392366708424,
 0.2738565258545156,0.1665684089401437,0.1460331354254597,0.8252870184856039,
 0.10884556223093478,0.5090352295398325,0.7748632792944583,0.8950544862558981,
 0.22099941824995617,0.3130001672332593,0.7099309041908292,
 0.014247844950981192,0.839399056168211,0.6414663770237833,0.5936898222552521,
 0.3991627750615889,0.3803070078609557,0.6595484646681217,0.18740879147754252,
 0.09955463876737936,0.9512294134712284,0.8871716020383886,0.5243747384973464,
 0.8630863528108209,0.9867164482390415,0.4237456976429542,0.15820712032486628,
 0.13780786567932257,0.44782617755486787,0.506509285122682,0.9372728918402954,
 0.6889738409293166,0.2248774811568911,0.7130988468100335,0.7250555412089622,
 0.5060900274395953,0.10842962102206155,0.31759999512063186,
 0.0826790744968503,0.34052521446676254,0.0658189251254877,0.5353787986986929,
 0.5232125341968964,0.3424432517834559,0.5246716325741595,0.5254678444862375,
 0.9737379811393128,0.5627761019063494,0.8421145705267655,0.5469953019451098,
 0.6090196891668765,0.6506539858524916,0.344213961610918,0.4614175987200895,
 0.9302093199775419,0.9344854310978387,0.1388596858656025,0.4977952258213848,
 0.18376030174638736,0.7409973206907484,0.06784594570281055,
 0.010679590629681157,0.6103557004303676,0.8981317070363464,
 0.48334583381577256,0.36874749682921215,0.34952114112306054,
 0.8383579528887667,0.9688727845467903,0.45162939528947654,0.677619687286218,
 0.8393802417017306,0.003778056115964712,0.6619216150350103,
 0.059381795552919336,0.7726432009567656,0.699439782246468,0.3530703288056488,
 0.6024407160937494,0.25778207061298486,0.4601866434263421,
 0.011728401919992137,0.07846953372332055,0.4576227450772815,
 0.2300759622255849,0.6350687297821176,0.8104803306684494,0.4319253268586142,
 0.3143601179900263,0.3954888354286362,0.04486922529436732,
 0.24686657187449912,0.7147855812639559,0.37756561525057586,
 0.13581993991464758,0.6691461023529739,0.9254024390606229,0.1570412599669857,
 0.1762356901400648,0.45859766017766745,0.4653701825603932,0.6355100478886264,
 0.014991543561353904,0.003997297557816859,0.9841652626754007,
 0.432044040452338,0.47415702587748876,0.09176751361569657,
 0.08905443837486826,0.3401790748600576,0.707336651300164,0.6611646676039763,
 0.38257679320618865,0.09617450330564592,0.705678286041494,
 0.016127653370264894,0.8792737157296768,0.8927658978384423,
 0.36924302562177835,0.4425254052545756,0.9261115717884186,
 0.07239976450867769,0.47419210899456776,0.11826974784320476,
 0.2635478894214658,0.06566340374070745,0.44608169875942816,
 0.22814438043460705,0.9602711559449455,0.07703682083736707,
 0.28788315214929927,0.9245391902486556,0.36554106941510156,
 0.6494900204439393,0.516270497253512,0.23413134362504118,0.7967064068408113,
 0.6465554855935378,0.5582151468830859,0.9968308196998967,0.8238264509457935,
 0.6616524156247836,0.6385202694050638,0.43414526274715526,0.7596500903515685,
 0.9898323486655056,0.8296365686838358,0.7430533633756943,0.9226343169626715,
 0.8104263671062346,0.18579821660234974,0.0789790073253549,0.3045680520018218,
 0.6113122487668692,0.44279079771099983,0.8496899364842679,0.8664871435164334,
 0.7653182547085752,0.42065799776757873,0.34197399712703214,
 0.8304298083812572,0.5888202340977469,0.5878052399950682,0.7196059798746868,
 0.08919741180323393,0.25956443457280143,0.5218154128762388,
 0.9135772407520778,0.8184783401918829,0.6108125181086911,0.9241489396190592,
 0.4055539967531244,0.5922917379479856,0.4335717756691757,0.5351973538722608,
 0.48952754100816787,0.3304438104663868,0.9164667046043792,
 0.21350148557149817,0.10016313711499558,0.11918175229366146,
 0.7893476589338991,0.6692960476302958,0.49208387512486673,
 0.16145176160852703,0.31342092230183904,0.6566190247244231,
 0.16822964628903336,0.2167277340374838,0.8057888892759558,0.6822930814998475,
 0.8502940120858276,0.38993562097420076,0.23168238236099958,
 0.6430921332563926,0.9139781797877853,0.36819448234560326,0.4853543972236538,
 0.8158589385991365,0.5857088659577178,0.9381512176274641,0.21115977235851513,
 0.3896813099613623,0.6097584977669839,0.6391596217504455,0.18604363375929456,
 0.6042256966655577,0.15200722744846007,0.6594227168536984,
 0.44125381654920093,0.5175124189887854,0.0738454170085524,
 0.43858978634748935,0.3751411901891969,0.7454468727398236,0.6110238873796854,
 0.9402859832983671,0.5247613201313706,0.20886151093609107,
 0.22371689502621617,0.6776502931318593,0.9296466905206983,0.9469492512835926,
 0.9710668293234168,0.3676932950817633,0.939364809776563,0.9244290720615734,
 0.48453952387692734,0.03162986611204677,0.7741919636256835,
 0.6009103753830357,0.8156225747702605,0.303463912542311,0.6692174090368082,
 0.9813097217959498,0.6630153079704939,0.2564828271408963,0.29692654814547403,
 0.8508364056605431,0.14889221445629786,0.9796527521964467,
 0.31739283110611394,0.1365623244897436,0.1818912400369992,0.9531678367956735,
 0.1654107786073975,0.9654536019339645,0.002182545478394804,0.345733172045936,
 0.5253854725043899,0.3573211458172264,0.3201946333589496,0.21907328498131895,
 0.5607448288793255,0.49949598918856153,0.55346935410319,0.3149078782598622,
 0.9223118185490535,0.53749260640729,0.5003084794773675,0.29573918916143893,
 0.5923357459330915,0.5140801643626428,0.37702581318724004,0.5099697267121445,
 0.3419424247856454,0.9284745145659316,0.4516833132001884,0.29449109498771553,
 0.615016179552726,0.48128363612309444,0.5992878945400689,0.3114889197969908,
 0.2714902713123307,0.12194361115735464,0.5724341653738562,
 0.12907967609005433,0.8369809360819828,0.26385075174389494,
 0.27272120106086317,0.5310890312363825,0.5419631700486678,0.9878823646595792,
 0.8997547269084122,0.6258356499571727,0.7841885053102462,0.21175925612319224,
 0.6385297333244196,0.9902878763659807,0.6066147413900214,0.09584659096850123,
 0.8962208594130198,0.21393571999074024,0.82381062532816,0.6852739463021685,
 0.2585403822166694,0.6211559134812241,0.7115945667133382,0.7740210204047892,
 0.42866510589389295,0.8361645083667123,0.11590355337213198,
 0.3852795895783595,0.3108648728252209,0.964298115629099,0.738629244288047,
 0.01632059918514317,0.9201143187797368,0.6961121628981949,0.6494362793563724,
 0.16209108254055238,0.35333559630906974,0.07927266711328296,
 0.527835298298194,0.28866269753812956,0.4869835639761687,0.5570799158319827,
 0.15971059921160913,0.32347322461794503,0.6375599158404868,
 0.27337280841062184,0.975446182105111,0.5570325365254761,0.7879577290904866,
 0.9276142895325402,0.8387117037231115,0.3930962074522878,0.7084745425766554,
 0.18893324846927506,0.34068222303098117,0.9976109029057395,
 0.7598086826911553,0.35969777672197634,0.5400566872580697,
 0.05817058024980237,0.3067311489888056,0.29633005475140184,
 0.026754562114748803,0.3936599285968677,0.6830100954050303,
 0.1526697161065914,0.23706061837443204,0.09632732197496707,
 0.9055103802397944,0.11305117678830967,0.16097163664888314,
 0.9089101755818683,0.8751389997550502,0.539153901819452,0.9823647530622529,
 0.6378513213042791,0.6869083004820995,0.5906845562686007,0.314866473015474,
 0.8777768851999503,0.9359252702194939,0.9772280692318782,0.5309208464041589,
 0.5897431315033131,0.9137260017631806,0.28189997768708197,0.4070215355349863,
 0.7238059654995511,0.5067079861076612,0.103232009750719,0.21426467114197045,
 0.22697860123438263,0.6353010103720513,0.21293310686302624,
 0.029795806475208853,0.6342073920579848,0.16833788402093997,
 0.9036956149451302,0.6944010206196446,0.11563439159859668,0.3690941487254523,
 0.861151359775222,0.42929256689883555,0.35213979218789815,0.641045491568651,
 0.503255912152286,0.2062722159676087,0.24401353786958957,0.6084509567108439,
 0.49982061800925515,0.5883631544219411,0.36469128734582645,
 0.9174235361106252,0.7121357695593882,0.9270302550860656,0.2834516921760786,
 0.06061784136544546,0.6409079847225841,0.7075646632310781,0.8655958373797025,
 0.19306718583333438,0.2530551533540383,0.31193893851309573,
 0.6221384701555824,0.8852134261295449,0.39611316529542007,0.8987388979149511,
 0.7821789936855383,0.0461774773916539,0.9938184100686588,0.4107014391000119,
 0.741638516690538,0.46210726924081025,0.501272464833369,0.9744385351488734,
 0.7891005519464859,0.3335777991972192,0.2024317262042432,0.2452140098907526,
 0.6008552414373627,0.5651683328950124,0.1780316761150964,0.8137787715581779,
 0.9801692718629339,0.07789526725755724,0.1487732887888502,0.7776019712447523,
 0.7084234741650378,0.8802045962247435,0.46893589202450503,
 0.32681122997731993,0.4879769457841805,0.8743280963662391,
 0.16181902959876437,0.626024221237247,0.7144740108517558,0.40940056567333816,
 0.9872198570509871,0.45519211735213105,0.8737698165561569,0.6481705146562002,
 0.713336473000613,0.9945453689265829,0.19479047202507394,0.4797755176525279,
 0.1348783711812327,0.8222469227585452,0.5381008808985248,0.9749006476460873,
 0.6912814589366413,0.16019682316174622,0.8719048140726284,0.7376715750865273,
 0.21631121816992704,0.3909646269455246,0.16279820623200525,
 0.21478662529066728,0.7227614928863192,0.04118639087210485,
 0.29880787741852166,0.9072495540181315,0.02429706106398788,
 0.6182450907924559,0.9384245058184938,0.3260907246120859,0.25367426645425295,
 0.6887178535131446,0.9222872478947067,0.18729149633913122,
 0.32762573516190874,0.34691477632231926,0.5436888377569337,
 0.3730505655588273,0.06624047436050606,0.903639885656262,0.5178568399255782,
 0.5012209747641587,0.9199531887723982,0.8574092417239406,0.9851136359309476,
 0.26484007219410255,0.1371179124911075,0.24766138516432668,
 0.08178907138076097,0.007569853788204406,0.6809084079946792,
 0.5928194489863707,0.7281395444268093,0.05206811112248666,0.906693604084254,
 0.18462409455944007,0.37362860037929313,0.6690323426148069,
 0.9949237002622051,0.3901970237951041,0.7459926671171337,0.4638545750906151,
 0.8627441254049201,0.3700226019241064,0.5734537491613768,0.9928226860333221,
 0.220816425440702,0.46161359909800426,0.3601171814720219,0.20186183014424564,
 0.6617277703173087,0.3282297864696163,0.3968079111315286,0.9392720941004103,
 0.2042730931450346,0.6036797712079989,0.8222901771071296,0.8515597415537436,
 0.934055713834536,0.4643131663888682,0.6199143152309738,0.01846104359029488,
 0.6100115402260714,0.20573405148420787,0.6331851237970234,0.7203131615131322,
 0.5778685543968259,0.9541597363548235,0.9778128534890259,0.19647865506723616,
 0.4960457381750609,0.7813255267859867,0.8925292066058954,0.2393051682712637,
 0.6037663434898863,0.39745626573172044,0.15106599172206991,
 0.7081607214993666,0.05170868571174858,0.9037097897041524,0.6256161105658804,
 0.7653622825313033,0.9330948074196951,0.8645196910540975,0.9227296736868599,
 0.29639144855753563,0.40142159631555485,0.7344413289904357,
 0.7343472378135792,0.4557246828654382,0.36925069005890376,0.535422466577906,
 0.08461775048332154,0.7011954072413036,0.20274530682005365,
 0.16316934784941073,0.1429645314843735,0.6595350973449389,0.7091594733274565,
 0.2005979731132146,0.4933404052274366,0.8829899399259037,0.5105751724384183,
 0.3939851899218909,0.7851795486446452,0.021416220563866717,
 0.6074343094312192,0.02606945962730456,0.376477141655982,0.4203684441753295,
 0.07696357139692012,0.28907604348247573,0.5228148949672102,
 0.11006123733042716,0.8305092371708207,0.10779044486934819,
 0.25889228935505393,0.3603436186576632,0.21914260758894477,
 0.27426960507981246,0.3389598426107052,0.7738843635309908,
 0.03697593571545099,0.27024982733036274,0.07396564079006829,
 0.26913574387426786,0.06579926673092296,0.10163593324875042,
 0.7760627786905305,0.16490406726517348,0.29003606852780106,
 7.206646201163291e-5,0.9843211035911282,0.10647518132242662,
 0.5291238646254175,0.7346457833909927,0.918069583659632,0.0858422991299902,
 0.5746248188334666,0.6018741575459182,0.4032449152896157,0.5794532506065239,
 0.6650870747258185,0.6954111147260695,0.6134229937620677,0.6282894689378559,
 0.5037438272712752,0.539396845091573,0.37078264070499856,0.6462621695775083,
 0.7913151352159598,0.27550986955140566,0.6536380650894368,
 0.48523281296479714,0.13561508416999168,0.15101570688002275,
 0.30784902613975296,0.8218749434655551,0.3569145497276679,0.1391343348843873,
 0.3772183264658211,0.7611714253583894,0.669572053220278,0.902519452817262,
 0.5043801578566599,0.061194645425616034,0.991185586977825,0.7458305262033598,
 0.5255339292552247,0.9531367519185643,0.08311617793609527,0.8652738562058715,
 0.9760516681412342,0.6514706175469874,0.3578576427129656,
 0.005762392841298314,0.2866940117959256,0.7392942364038884,
 0.5705365627801651,0.09723012122141728,0.24956644894904534,
 0.018330810044801638,0.6506888374985018,0.42349060738958655,
 0.9163722958310073,0.4773394335367417,0.9317402996799081,0.9002376330490982,
 0.10576688549500024,0.0890181868384311,0.11049265682612441,
 0.6400872124164771,0.55607644106718,0.24124960460808087,0.28196557771814224,
 0.3228865789841113,0.10407262908660675,0.19192201657466557,
 0.6016720996453216,0.3216629357723305,0.298338292594044,0.8521246749399684,
 0.5989883963388968,0.7723293100388018,0.2999165776530547,0.31817392495718333,
 0.16247606789714963,0.20850684348986614,0.22933259925771932,
 0.628032693757437,0.07206670807282467,0.47138934237205055,
 0.30717095832412655,0.3050276357712115,0.6321520336593647,
 0.09024839135289675,0.4172468660897789,0.7994745225005061,0.433165747514327,
 0.7451596529629019,0.11087886754183829,0.4229854616645097,
 0.17866827605889146,0.31912723952075517,0.8391559502707199,
 0.3592797885351454,0.3245122319941096,0.7485383096372874,0.9531343970216448,
 0.8731905412322688,0.6551168577018204,0.637695180797467,0.8605479372190926,
 0.19555707434119296,0.5798991691758945,0.7242807396981376,0.1324726771579443,
 0.7871887049186382,0.33255892682700694,0.29020519564505876,
 0.9262215411204063,0.5616459142029887,0.19947851912157266,0.2991050414567984,
 0.39337600442984433,0.8777326224802238,0.2949113116382658,0.6954268103195249,
 0.10531627964454582,0.6161850596729289,0.30554354929510785,
 0.5810587709888464,0.6460827016410895,2.5712746957462507e-4,
 0.08842176852701122,0.8189602167345202,0.9151629072191596,0.7181819087149216,
 0.006946639613541694,0.9341727793780145,0.9131901400214645,
 0.5298322993552604,0.7171976856008857,0.5192924612795752,0.7377345599829864,
 0.7366509538991706,0.5414131811398009,0.40833408617152867,0.0258188663128327,
 0.604203551871512,0.08927192888808055,0.8504139324412066,0.2497421135328035,
 0.7622672449289,0.6656629695867994,0.4420106722799487,0.4964539941104067,
 0.8797805370994842,0.14714248512624273,0.2778946792987026,
 0.06684146059767426,0.21255808255331166,0.4027002919459799,
 0.02282305908880866,0.696678084780029,0.6535734150311568,0.5696832954586036,
 0.8478375445622345,0.5519070951622189,0.5039633118489458,0.12567006511276568,
 0.7500368176001914,0.6046297334421665,0.38456698853130145,
 0.03208598190267109,0.492380702684058,0.4583322155479812,0.03950636019445286,
 0.2367178574424994,0.6914778333098005,0.29826123484715095,0.6877257557530236,
 0.5047465635076768,0.694078781991193,0.6287678683222983,0.929053783626351,
 0.5518182926888838,0.11364947198219255,0.6426557409748176,0.2938152962416636,
 0.25523640871057385,0.8274971532272943,0.5934618612885847,
 0.01264771706662482,0.7423764949489697,0.24508967294805795,
 0.9597331693516643,0.6457193777348837,0.9670938382718114,0.7880345773793801,
 0.32663340213222036,0.31325013721181927,0.20732916396366963,
 0.42683861538064205,0.05819103515417767,0.44360306821314865,
 0.903148817962527,0.43577955648558364,0.1596806797608279,0.7540823973934918,
 0.40771504937797753,0.9642453265604282,0.37695570298783554,
 0.9361021155788668,0.15028995800420908,0.2519288313103718,
 0.28375262405420865,0.26286742940096797,0.8619577085509504,0.5866683607895,
 0.08567402548914194,0.45367087551208063,0.04160940168696081,
 0.07379426676256495,0.7436219896635824,0.08805773908406866,
 0.8559852713084088,0.5764353545674448,0.7897419740669456,0.6836200228926769,
 0.010776789828883526,0.40520999491882914,0.29387712236507957,
 0.6144137569057792,0.54296517904612,0.44645016562474427,0.3345710292342239,
 0.5812754440560807,0.24771120297130012,0.11807571153450414,
 0.9230849057090288,0.3046469452433904,0.708523685757412,0.9483719811028507,
 0.9760487803611255,0.08710864089132964,0.19717214283021378,
 0.5493110945328462,0.39559669304544054,0.033330926780638315,
 0.4760859065971448,0.9070543882275038,0.602590281076211,0.13814932174019212,
 0.5522708280866202,0.19602606638155695,0.6548250148313585,0.6532130440422882,
 0.02770207534767488,0.3409273748296954,0.17480779088172937,
 0.5217141071370712,0.5317844927434741,0.09140999927956306,
 0.17498020392852653,0.37413710614758144,0.4759579341488084,
 0.3138719329158203,0.14674022152001776,0.6661728089284277,0.6048337291585596,
 0.043196195949638416,0.5287856630090209,0.520788200107729,0.2910457202627831,
 0.1312752904778206,0.45955086148310376,0.30083987674445223,
 0.09054452635168209,0.0720559707082713,0.2090178587037831,0.6800458489846436,
 0.13691289652244942,0.7523989771362904,0.3435698058526385,0.2570165911025144,
 0.785052868948944,0.5430266452554928,0.004550265024980349,
 0.44656538957147307,0.18284594898338313,0.9243476463536474,
 0.22796937255278737,0.5559291075245977,0.4487935058566349,
 0.022101148680928162,0.2636631493746415,0.6333679589180066,
 0.2587489746911006,0.8951388583442753,0.7803372441474039,0.67850626118426,
 0.7562672096860638,0.15135697594797548,0.7617196657523524,
 0.09526613353958313,0.7989145997703041,0.9455746393495178,0.6493137424206351,
 0.9113071563882601,0.33738297961049535,0.9548212588121435,
 0.36174602883536144,0.06753974270346363,0.3182254628829,0.694059096634757,
 0.5492978611108852,0.9151616454877001,0.469986865367112,0.9181812767484459,
 0.09073999604460459,0.45727798093536853,0.7707401157072652,0.905471762912216,
 0.017859816954325147,0.8140659473991873,0.7543547922227642,
 0.9184879410498669,0.1956304098304974,0.9727279814946238,0.1908255620607231,
 0.41201652188136384,0.3887680409891454,0.4981760491820141,
 0.41353095875187784,0.7385475906065165,0.009994531912298377,
 0.6556195466391845,0.1306561744430037,0.5290417473036151,0.7805238719809324,
 0.636077219103595,0.888454976772314,0.14231888594119113,0.29962211397866734,
 0.5316532808664245,0.7592023581141817,0.3190036660748313,0.7130897960804105,
 0.11385707379618637,0.2520817918941648,0.13549403487132128,
 0.14458671083165098,0.34092473717805216,0.2217814810168952,
 0.5470957691325048,0.1032806168552155,0.8523429713164787,0.2789900424924896,
 0.7577342840632982,0.8191512490734484,0.5324697642370735,0.7900759355919903,
 0.06736250020530954,0.9709198467320329,0.0947881122619102,
 0.013594580577764193,0.3765026758601583,0.7488271691158694,
 0.1353146580899116,0.44944780660701955,0.7125761926139839,
 0.02385506095123635,0.34973158983856745,0.6042312373704307,
 0.8216394119940175,0.8254904130423972,0.45864130637276834,0.7091981448923631,
 0.7812579876803513,0.7380006427022493,0.6656655232365455,0.5883712551212237,
 0.8477650831762293,0.845212187921097,0.7938699855270839,0.9026878981526166,
 0.19538236995058345,0.7117923192191092,0.4723647611009164,0.37121566166136,
 0.994717909208697,0.18851510683854344,0.3198213112562712,0.5875392334944608,
 0.9776464365359319,0.770910950436833,0.22556041991812692,0.20848204849386698,
 0.03194097256330819,0.7634448446416263,0.1925816421624732,0.9204237936016006,
 0.3378139880787113,0.5502196682558779,0.9861985845408536,0.42185588969991683,
 0.5587467663681418,0.34557445483812865,0.06052042140473668,
 0.8128609498497902,0.8767215001923524,0.18971951936275133,
 0.009158597879215802,0.4502674593552787,0.06411176951859043,
 0.25789003740063077,0.9400584769437318,0.37048036614700486,
 0.25722256638368957,0.9958452269407644,0.31005614569484563,
 0.1458431376923267,0.5268437144991003,0.44142701055415123,0.9878414072622412,
 0.1385622185600638,0.7575804004886935,0.28724820327132006,0.580942259470902,
 0.08740619766757407,0.7892492023906015,0.6609637778098313,0.4682282436498454,
 0.5399446517206945,0.35406836181759194,0.3470470576491377,0.4729943422892565,
 0.49860963487166887,0.9215565363187798,0.1253852882760783,
 0.10367457711702399,0.7883700974327411,0.03153445834417645,
 0.4370885832919065,0.8811263765392204,0.12983867163112028,0.9734656759556526,
 0.46298808933615754,0.6303821420432938,0.7089854078184665,0.394729567875827,
 0.6259390237962336,0.6107233190019097,0.009645720064542895,
 0.18348727570419632,0.8817345547026046,0.2741251477136616,
 0.07146032971413574,0.19868685361789307,0.4319687956830718,
 0.49170281018130313,0.7618235734781793,0.3308417625219481,0.3607691682849298,
 0.11776235390779832,0.3973484070260519,0.5927570786567065,0.6006817593177296,
 0.9927981222406261,0.4684140338442736,0.9720202451568267,0.5434479680399111,
 0.2575911232905902,0.1617002836629866,0.8595096031638491,0.2499359870734411,
 0.1877215011060881,0.6505455924007393,0.19788041332826833,0.7876236748855077,
 0.8218800706258338,0.48883711646297734,0.3255331642930753,0.5643445273461137,
 0.7557248508172146,0.40622232338955144,0.14469004308383315,
 0.44163245503333004,0.9725618994854138,0.6778533744245911,0.6455320065872803,
 0.44282012304528773,0.21597891377292222,0.9776061510150832,
 0.6978225246326712,0.5221360266641571,0.2648267502316193,0.5197368718497325,
 0.10191975553100685,0.3832265690105975,0.6749081926445283,0.6928511428052789,
 0.8414439344074853,0.08198875988329024,0.5303077097542133,0.6850082711239607,
 0.6097862320431958,0.2605529793419755,0.0844316207195871,0.48209427545559025,
 0.4870229213761278,0.12737856371339706,0.914164514433547,0.9743872413023091,
 0.5417013497702199,0.392616807317536,0.2897872574766921,0.4789019800876986,
 0.6147558624630948,0.15387467197448113,0.21483757592542063,
 0.2815168835555443,0.7607474599080628,0.7989519125163027,0.5988617956287842,
 0.012060211865989157,0.8279338979103849,0.977853047136379,0.9711929721613436,
 0.2967886776034354,0.63482213590913,0.5684835253424855,0.928540593536362,
 0.4328902240054877,0.7920465229184208,0.1566735372241873,0.8599999222303664,
 0.617098275917875,0.5033525681395303,0.37927019612769264,0.4312834296719167,
 0.20283086783624882,0.6655596793346839,0.9194955506647436,0.7596983932279416,
 0.28520404316574843,0.29168814342869454,0.703811437097439,
 0.29621333086390667,0.44521585233587824,0.36361301983652206,
 0.08575947920740135,0.47984700750353326,0.2339435459174216,
 0.020557123759486373,0.7046118898787664,0.7411578580240191,
 0.18456490035877737,0.5955849652921353,0.1785672492303575,
 0.23770888609535334,0.924498577982974,0.7505561645161268,0.3847552441010088,
 0.6169319715922473,0.5321438114797742,0.1436331196850058,0.8465084771027824,
 0.1469921265325842,0.4110167895321156,0.43514619519907205,0.7230295586680974,
 0.6909501498131961,0.00657459492224477,0.8529960032832703,0.9740488487756751,
 0.9173770341775955,0.5216302416719376,0.3467878543225712,0.8803242473008592,
 0.6405270759298862,0.1491919642094076,0.36899629435191417,0.4738509639704146,
 0.4597106836540197,0.48785180293839925,0.33995328183981566,
 0.032924237843434456,0.32771321756070215,0.554524799719167,
 0.5576446973771391,0.8779267222534769,0.09232888113972093,0.0834789351244748,
 0.8873510439958749,0.21424096149446303,0.235192570369806,0.7264324029545731,
 0.9661123349503603,0.11995225666021614,0.08626296529282762,
 0.7879213715321295,0.5753249332876211,0.5117113216328548,0.9823035232399167,
 0.4125533349374205,0.9372383404548602,0.20122781623979114,0.8736804041910977,
 0.7518954546911162,0.4359822702643952,0.3410919118762258,0.11509294617276145,
 0.39674912830330045,0.38561899719167547,0.9138787238151714,
 0.14276783661386216,0.2537078329845057,0.3579546861574644,0.7632194631542019,
 0.13054322796047035,0.938029976510817,0.5187970052886051,0.7153991676210619,
 0.6099545190051018,0.509187439138547,0.6223465567124824,0.09054962358920304,
 0.33850678309193993,0.12802583691846192,0.6899040701264542,
 0.3805827652435455,4.224806378366752e-4,0.9245200445785563,
 0.42504311117157356,0.3844547604386497,0.17720072295877753,
 0.3881591921554589,0.48187561468443374,0.20227471480207238,
 0.1220430274443789,0.5882953317822662,0.45090746308557716,0.6782770356552436,
 0.6223274490149039,0.9822659386187298,0.9053230364806574,0.13477262242282273,
 0.8250471143005352,0.81276458722791,0.6014588983774372,0.002178783882340496,
 0.21119951590518193,0.2477925860703869,0.5978578741534548,0.6729270916132722,
 0.5705627456653174,0.07368973535303469,0.07310765458755653,
 0.8020940289972457,0.4482931429128981,0.8762833031490542,0.5026094742144291,
 0.5551919258369659,0.41519467374706864,0.89494657687663,0.6145617891994746,
 0.4034492511827985,0.9764218343149416,0.9488665760387417,0.9664391541240837,
 0.20214826170736577,0.4620150814153887,0.13883120681086325,
 0.5128250291733522,0.2623229598479082,0.6347199667368929,0.12193407888255381,
 0.26009036029565147,0.12857255323729733,0.6253394062469719,
 0.7920546932506914,0.9345772182549553,0.08805692301997026,0.3119890365037602,
 0.4175803394589368,0.4814026371070059,0.276127606610295,0.38477681367487593,
 0.6564488238848463,0.8940381094713441,0.716473658719649,0.7058773841544966,
 0.5247615314207882,0.1512779812654902,0.0362756829068821,0.45297555576414283,
 0.7383109834906159,0.6543563057485845,0.3498098824651642,0.8258547745324039,
 0.5223935007220218,0.2540487377651427,0.8992388764395988,0.4649424682053277,
 0.8344453966563109,0.36119233232092784,0.5626125316980354,0.6901568423584048,
 0.607866437870791,0.7718483611119491,0.5172519170060654,0.3708414223789396,
 0.4619212730731552,0.3520403705893359,0.8143243192703647,0.9797644734031768,
 0.11797849290574236,0.004600874219222728,0.6278767795189799,
 0.5476236700514007,0.5345775848082582,0.8357235698385416,0.44750170265734657,
 0.41224420770219505,0.8948135627936674,0.19915694983664034,
 0.15210839160454992,0.05565205173698062,0.44707615561993896,
 0.6947291101398309,0.49141345207251774,0.4660208158139246,
 0.17262343544195224,0.6152472695120134,0.016351691970883486,
 0.46716336013368476,0.049147534759608646,0.6205100094096607,
 0.6164770272679433,0.9685162127452607,0.005493885130945508,
 0.9280209612944174,0.9847581047639656,0.9917794595644162,0.8244948258159579,
 0.5854928760525662,0.45156115831964205,0.771708180862489,0.10189071076261602,
 0.4269074504492586,0.6125284411672741,0.05304066506878802,0.72093733533658,
 0.697068059086282,0.9323862849341988,0.4860187542848817,0.18173468845389007,
 0.6487145932402527,0.8323141519662146,0.7026175290859282,0.6070176997368737,
 0.9634868020062137,0.15574743150478454,0.9581296804170237,
 0.35926601887655885,0.9691021264791972,0.895147650538553,0.7910211887205099,
 0.42757556415899656,0.11502618254134134,0.38160820438726106,
 0.016584762765551653,0.6286372719073203,0.7949980343495879,0.784233624616361,
 0.22806900427523558,0.13359399833318497,0.2160132792981445,
 0.23664731863884747,0.7983163595109959,0.029824803741819617,
 0.013726291596444007,0.7407856926491223,0.5786255823819463,
 0.7210761305445024,0.05396404387363607,0.6698323814494975,
 0.006001002962616853,0.35603685965851906,0.7018463226372673,
 0.518651190039447,0.8537396904162384,0.5478513816425702,0.01939174573096636,
 0.7500751363146227,0.5099839515285032,0.32644828620592925,0.5660906599519928,
 0.8785659415484963,0.265246156764671,0.11027528004478082,0.26711371722712185,
 0.3705877484466298,0.8123344488977106,0.7944785238864658,0.9309940922688571,
 0.6340495825617358,0.23264574305952168,0.09211184268198336,
 0.7884468037979152,0.5537147186698619,0.32579072302792755,0.9959588413433264,
 0.4093914477904015,0.15612642625636675,0.4130990329751413,
 0.04712278837013384,0.758496260332208,0.6612955323989448,0.45237798661282913,
 0.9030788024651206,0.942019297242506,0.2674092102783435,0.014346288722275835,
 0.8047034168455796,0.8496445135603383,0.32156441400533486,0.7439508165023908,
 0.09299535743749288,0.2543716048037794,0.44587209615329126,
 0.19176741953129417,0.4968416002026136,0.9762209956437011,
 0.08570686208187828,0.4137889877924953,0.8852384616838922,0.2768235374194631,
 0.7349278112698103,0.027708024382951058,0.5794832985333397,
 0.8497728302863078,0.9790189037738027,0.4462225363314618,0.38191275414517956,
 0.7134421105047944,0.014866469573959429,0.6862538302191703,
 0.6694857323401956,0.8330545242623493,0.7548704021706691,0.39848552299803375,
 0.9079709004674863,0.2458411526359282,0.07023646409228013,
 0.14354553499764977,0.24974056351255147,0.07418416825163066,
 0.5977729040757042,0.5633716028366615,0.5071226879140107,0.14177020231323234,
 0.13850863116356116,0.9789754222690016,0.392688571328751,0.7979130417178596,
 0.5484939755398468,0.38095041311233246,0.6119098022897458,0.0405285727895468,
 0.8750273876930705,0.6130522634685208,0.9370010024529315,0.43585476239751164,
 0.3900990292554538,0.7652829401019502,0.4150034642863849,
 0.057318826960694014,0.898752211544614,0.399499736169429,0.6807494162682982,
 0.8423631651682665,0.09109506607741025,0.9087643336557946,0.8631590778419628,
 0.4910074185809401,0.20793051645173732,0.3966973788775532,0.9486027503719273,
 0.30987753748515523,0.6307602543262286,0.4859168491060317,0.8571045188825545,
 0.7087166898206629,0.053893137358695276,0.6477418901956408,
 0.3232223144707581,0.4453482195018087,0.3595169723741871,0.8333149887524203,
 0.12711090294965632,0.6868974188658744,0.6425996663492374,0.6089761603837706,
 0.7979328283004631,0.8581028055866373,0.42801297836852836,0.7559568142070521,
 0.09361569301295902,0.5888379307084602,0.06757394302632402,
 0.7189519627342925,0.005599817467448642,0.34670614630282526,
 0.009450565914583264,0.3877434843139146,0.9705720206861436,
 0.2116053369232227,0.42955409889113993,0.08676394083275574,
 0.5291608251811454,0.6388937013029069,0.05027126623454081,0.3372559173541698,
 0.23641638917972296,0.2370199963835118,0.8293942820062001,0.2026107771930844,
 0.412212147423471,0.6560648876491092,0.8083321356883145,0.6608785199906949,
 0.5019617809717491,0.04943846243673988,0.23583746586954057,
 0.3245558164705692,0.44556268689636647,0.6438088910187066,0.629927751176111,
 0.09855097133576707,0.1897489656275756,0.1991547370893325,0.6437657276912256,
 0.4021738222248624,0.4117798753588653,0.556013672310408,0.3747216499715822,
 0.23902307358245656,0.531277473915666,0.2706609865187558,0.23052409253671358,
 0.637506972084282,0.20707956184726084,0.6437410002640829,0.9714176337392456,
 0.6043976796790977,0.07650176426964272,0.3209550511561321,
 0.18537287924227863,0.5595764148674621,0.4858534294201945,0.5031189596740455,
 0.2822205536834701,0.36794542313407763,0.8784332764843855,0.4366287698164719,
 0.6107114981601134,0.31725959507312396,0.6142246468927455,0.7524815252680562,
 0.07899948288664527,0.7142903461503679,0.16565208935164477,
 0.7815507889949622,0.8322236613718004,0.3746865116453898,0.2888030002254788,
 0.7039365931862963,0.0471355505904838,0.07438948311469351,0.9300476944542139,
 0.2824976412738148,0.7496897975791982,0.024509618229993757,
 0.005482200229923295,0.5420601712062134,0.021113157813349215,
 0.19822327371228965,0.9654054350530554,0.0438430077286468,
 0.26487809462527534,0.761279401002344,0.714523631903997,0.9596252483961832,
 0.2541359104844878,0.46092361714459784,0.2720458824805557,0.9162946788668307,
 0.29992109292707925,0.11769205331220278,0.7788840772351957,
 0.6935036597286237,0.588270735128712,0.9014283317994699,0.28745144190811245,
 0.06627343119445428,0.3684730149219263,0.6107723544217998,0.6619960317627366,
 0.8404030502030294,0.3923901782098489,0.6499367393554097,0.9383416984731205,
 0.5585730700529631,0.7966125288107869,0.8411859728847404,0.3151938649565581,
 0.11796294858106626,0.3721293817524749,0.11633865965692491,
 0.9847830473868888,0.803339182835886,0.279796168929921,0.2428204236598812,
 0.19364092073375982,0.23892241829268934,0.8278770378687531,
 0.3713078756146231,0.5884675977450513,0.15626731162754703,0.230640353715711,
 0.7516069682366588,0.07230899794737922,0.2948925610748474,
 0.42161536260084953,0.04293120431630015,0.27975943025115313,
 0.12682196357339914,0.18803875256037017,0.42923148542853895,
 0.4231957170941544,0.6753297362062507,0.00309494064926763,0.6284004104009668,
 0.8299338791184165,0.9848741750433269,0.9131677086350322,0.662109860862236,
 0.532695091256595,0.7659254969359157,0.47100569300095096,0.6541804220157846,
 0.3616517298562737,0.25916312696464483,0.4627414120065988,0.3145410212743329,
 0.48639933429797566,0.2586086842483313,0.7158651118238727,0.3148821473033061,
 0.7986890679794666,0.7115345463592582,0.1788011964739179,0.09570820370042687,
 0.4586928173087177,0.6601124263092897,0.07886353559450698,0.7273078285515114,
 0.5577086011634709,0.5599873486535892,0.8000813973521105,0.4891664125980495,
 0.07691897613494514,0.35254826281865337,0.5228617675515113,
 0.07025027591916444,0.5073371361489276,0.40292544379071915,
 0.4689517505814442,0.48922661830963654,0.17647037254845388,
 0.2079988349917895,0.18765397095926617,0.34168567121743054,
 0.9535398393414063,0.5556773801960924,0.528212516491726,0.4768556799773538,
 0.8501008625254624,0.8191398280132572,0.7615686139983802,0.4714549847317673,
 0.90445479986552,0.6598771873760627,0.5858161423123261,0.32636028541083895,
 0.4624105414165479,0.8729469331430253,0.008208467783712092,
 0.1295404782548848,0.03947592022920765,0.1520430397269701,0.4508754897189673,
 0.8002966593732607,0.5402941462294343,0.8055986589611727,0.7626571560606237,
 0.02157730897426502,0.5365801552379085,0.906462939745257,0.06399247739271363,
 0.9713195039869891,0.8471599856622485,0.03987406958015893,
 0.19284400571459093,0.5704727199414898,0.3743797424573222,0.6113151001187352,
 0.7142088500512167,0.607815658760838,0.03186900007875082,0.8809208313949246,
 0.41729549011855815,0.800846840382995,0.9013229233727493,
 0.018121217755501795,0.15237075501362352,0.0012855308846264624,
 0.35614227062042736,0.8110989023661094,0.9661207118919986,0.129023224149316,
 0.8497466180145814,0.9404159471182714,0.7520803241261081,
 0.0016460158195460428,0.09827347713884249,0.7291189985972885,
 0.6891703451940008,0.4804314242673783,0.7296350640656272,0.3171142178831151,
 0.12286842473061554,0.17772831943707101,0.608764515349322,0.0214616340639584,
 0.9407098421544388,0.5313092102982933,0.318413646272396,0.31507048950517513,
 0.601303075327476,0.7694206545474719,0.3859718177608271,0.6399575748163123,
 0.5922101695846348,0.6962811857317059,0.7303144798758301,0.4145711938813923,
 0.45161045566870484,0.6020382589741167,0.6487494248779133,
 0.34667674286258054,0.6672867844971707,0.35557718095640456,
 0.44862605859027094,0.12173248812773152,0.7561747141788192,
 0.8324305009444872,0.04432881635341723,0.8236852390558348,
 0.10495527463279053,0.562352653855164,0.6065707594438581,0.08273221704586597,
 0.4373656014583296,0.9717731434661496,0.6632150061400724,0.24242308167777793,
 0.3978662577872839,0.8833869443524843,0.543288891832965,0.5477953928573863,
 0.9029271662923806,0.41029205354821685,0.45935626338756963,
 0.8040998939111755,0.8632847355391099,0.6189080409747305,0.39199353725017905,
 0.0581474098676672,0.9665621414276558,0.5314762764084132,0.09034225630549697,
 0.8932455633854133,0.37168280875050663,0.27478337699998623,
 0.9802242686076923,0.659645961107812,0.7233673325314997,0.4354369230374253,
 0.4940738828250818,0.23757076312900705,0.25666347158274094,
 0.5406785633576859,0.10712752895701916,0.845430478759209,0.04889065623903843,
 0.2920326858292388,0.7885327849092859,0.9861305828733908,0.5410555293140148,
 0.16537970423591164,0.16208044319868875,0.7060550149879637,
 0.6812445945728002,0.34582353940725774,0.5038621842079622,0.5206340852217813,
 0.9145349784508476,0.09296961616100785,0.6584565124342306,
 0.14555463539315416,0.4165662089205018,0.4916223079085694,
 0.07182433950613709,0.8710686069050304,0.7974913032774736,0.3744492030717076,
 0.28423914783105153,0.8134683138540492,0.6148198091724698,0.4157603233045527,
 0.9393113124057592,0.18754135188197452,0.5254080250749052,
 0.17073419715999705,0.3091478875399186,0.3558979265221276,0.6088196539168115,
 0.28853845057462435,0.7859148129072275,0.6572506783454641,0.1426886581284602,
 0.6259857431223097,0.3619473206608521,0.6778313140050548,0.10228161342793907,
 0.06366663266783046,0.10083793307696498,0.9391174830318544,
 0.8656107918631221,0.029360858644648657,0.32175638144127716,
 0.7121836763289265,0.6648609557031157,0.43930188845799156,0.9948872382933676,
 0.7466749877769971,0.198412963175979,0.858921918112892,0.5131899237451085,
 0.3523622331461157,0.7763311504240783,0.7270285235623412,0.43426085745464604,
 0.5295392551235604,0.37503511728568784,0.9477566274770111,
 0.036695001181568654,0.5155451557251007,0.9317874806154871,
 0.48262675761658824,0.0781621769001003,0.9077198304212408,0.6760905606735979,
 0.2745120466712979,0.4552039009398545,0.053422733327197824,
 0.5756776142583303,0.15709446128152504,0.36282988879056044,
 0.8115096643691235,0.060366462383453445,0.5536702516766229,
 0.03149117691928505,0.3847512383406727,0.7405902513919854,
 0.021386629415324654,0.06149223118222502,0.34255835653827293,
 0.7497871502975403,0.748219292038785,0.14975173295487798,0.8422595267495842,
 0.9292742953412144,0.6078549288086424,0.7598920023682117,0.2141367275284881,
 0.6176155311479771,0.9754346412630723,0.6198684045012498,0.1667802226786036,
 0.640223871658995,0.9457869064444231,0.8863652717839003,0.506580995909215,
 0.3300927034995572,0.16754286823364484,0.4983557032394337,
 0.017423123937715945,0.05899079843354915,0.5558950623862832,
 0.34942957491492277,0.21529835065712688,0.025795780529924595,
 0.6730277727825892,0.22649835443617494,0.6340211008738806,0.5937594379833256,
 0.1352643256069196,0.6590169752510026,0.3612198977681901,0.9337705107158935,
 0.2887172227328846,0.4700708053737147,0.4196871063250227,0.11477978290509172,
 0.7459111887627657,0.5617560642080667,0.3744004405173099,0.18635802513147548,
 0.5466002238902519,0.3660993572983279,0.22422896389109837,0.9466087865151604,
 0.21305852168379102,0.5417200377991838,0.3827393962504986,
 0.21082621138842617,0.24116952198218988,0.39410500758678146,
 0.07361994410252048,0.9521335938817987,0.5080201521880003,
 0.27048247858930063,0.33376447173866497,0.4591768370937154,
 0.1923162400289592,0.7049628521364759,0.8616665711974312,0.8510189142236197,
 0.14393575565756,0.8319031279090713,0.6959557760064785,0.3268869428779758,
 0.9211156262028412,0.6277572249105638,0.7172310822571958,0.2807728893804161,
 0.4536533891129315,0.04227263615656596,0.630918910036766,0.923881234108486,
 0.8835237976401904,0.5422029622733731,0.3668732081294419,0.8634476349656908,
 0.44019813831905963,0.008329767707528246,0.4376982399400343,
 0.5584876865603998,0.22929270830564108,0.9138530905296149,
 0.13025400516394325,0.6382813611819367,0.3349410486748794,
 0.22478643826239564,0.46734535442277203,0.5143883995216605,
 0.7993795622208029,0.5316967437997144,0.1973918709580953,0.7889872379797133,
 0.8674418948013843,0.35488247134843054,0.26252315008142724,
 0.3757930810843595,0.22204443927618822,0.31665724238542237,
 0.6383268719450761,0.4682431401656464,0.2174028131314123,0.37369749445332046,
 0.11376066872206758,0.8315466584731559,0.9547762467794919,0.1642980214014903,
 0.5741428257341356,0.4762857343991356,0.21393087400167587,0.3320601643848259,
 0.9781034372903954,0.9419818809271215,0.6033077947903673,0.44498812745129956,
 0.015597711545701776,0.9996696614946187,0.8763454881285835,
 0.8276074647215828,0.18625287144281422,0.49934778833576954,
 0.6124455824924633,0.8911405611942602,0.18969725498814305,0.2658096894138636,
 0.7133018917055877,0.9309311485853861,0.009809849498818135,
 0.013742543514682626,0.8378478472673222,0.9017545669866166,
 0.4556937041097754,0.8799894115956337,0.8370973636827961,0.5321697729849284,
 0.5016174721129629,0.4969188238979928,0.7992315129260127,0.5750230678133279,
 0.5210395744486266,0.8152674625493087,0.27527462254926427,0.1186373490309457,
 0.3350445709840253,0.8917603110139938,0.22825745588970547,
 0.46121906696194115,0.6109300333154744,0.9365644341020597,0.2173917191026331,
 0.6022019768185105,0.3026271999910166,0.6218704017018502,0.2950995569871341,
 0.8924310881191624,0.7742735009956483,0.2834883081883157,0.8468245536720401,
 0.4440232361766333,0.0075980791085291655,0.3547282305417323,
 0.5316066820157689,0.9559584152518136,0.4301404335266108,0.10650525649934772,
 0.14627336865401985,0.06254272304889152,0.022009271698445687,
 0.2614533974828781,0.3742173359496008,0.054760875973907774,0.813673798548394,
 0.9929713202466169,0.9175194581526053,0.42977339268789083,0.6232273357611419,
 0.8186588088830797,0.32005396660476293,0.7874790964267173,
 0.08200946488139271,0.11724061922648199,0.4635171846961249,
 0.09386757290128633,0.7740491361191475,0.7929948080976922,0.9026456603246874,
 0.34085521451337275,0.1800839164172977,0.7908091286891108,0.7809537980113593,
 0.19680383343116192,0.5077160405676818,0.8148539767713291,0.9513403237797549,
 0.4415568588809581,0.37846075389726264,0.9175260416583209,
 0.07571821642682619,0.9298609736525456,0.8679495749431709,0.6586641751285034,
 0.010593365716042191,0.5609541340286686,0.7074597684848728,
 0.6879508874825052,0.6898059211161588,0.7997802085824695,0.5042391249374997,
 0.8904309528776791,0.9946069991355575,0.03709264188683481,
 0.48621083780302676,0.29228970022711775,0.28720289489861106,
 0.4951974666386072,0.4689053764871295,0.738622967701364,0.9890287732847717,
 0.22629565574890043,0.7649325893257364,0.8817620200690699,
 0.10421840286648165,0.6351260164230719,0.10828079205053487,0.892436089173315,
 0.07494564079240218,0.646303565719819,0.8132283132187003,0.06498089059435819,
 0.13341809734028798,0.5516859270731361,0.8533297598857972,
 0.49352752524993826,0.3880318558070587,0.22355037684702128,
 0.4673125194129444,0.9201203958731045,0.15775388832594472,
 0.07534426785561343,0.7575162144233348,0.01765019465029727,0.157220085368728,
 0.8798506674229167,0.46433661765311596,0.13586575054633387,
 0.5116579353335984,0.3158341824253076,0.23551044789332976,0.5783504053550594,
 0.15625505658562178,0.15674932306890366,0.31828743380921387,
 0.46078343630673224,0.15191599749064333,0.35496306086072704,
 0.6871968709422209,0.7824566150498966,0.09604907952480879,0.3256833686499587,
 0.7594117568340416,0.3913157313212773,0.5212763820504963,0.3058679512485877,
 0.6183855248914244,0.2748011361613043,0.19095706251652333,
 0.48607360315538595,0.8086430939664241,0.2559096333188924,0.2439693894372038,
 0.938441443967686,0.6371885767903045,0.5785987364983503,0.46394313309509116,
 0.955861406879643,0.37974377327103803,0.7529026677002024,0.1411891440145976,
 0.5706403778455695,0.6756740036793173,0.6530826288090356,0.4618916519232694,
 0.27787501837940043,0.21584234597022012,0.13120458672138446,
 0.15032677811157658,0.7525409476593515,0.41873552310332907,
 0.06693258332968699,0.8013471081193615,0.006951675191734785,
 0.9821989347657527,0.5887554765733413,0.8033183919618034,0.21464218851633388,
 0.4403416428346629,0.3584639536422888,0.5951834540597426,0.23158065760046587,
 0.6780455663226002,0.612757136663152,0.8622232911001718,0.31104542897820253,
 0.916472062081038,0.17642529825600284,0.353742882935405,0.31206590976128445,
 0.23855485892224482,0.9205243765683794,0.9662395199176137,0.8089365810904736,
 0.2037473828195513,0.9712903430348998,0.766802783407876,0.6181914038829706,
 0.6218509543068668,0.8138188647821383,0.12426813416252802,0.7570859358470595,
 0.5328860297320513,0.19953283993782256,0.7721452971907954,0.4800066537778519,
 0.08843388648620298,0.6256229069953312,0.9071270504811317,
 0.07831513007625501,0.4107121699712055,0.17260135615064498,
 0.30475610321294666,0.68994642856463,0.3107262063180245,0.219037924872291,
 0.33932711445830743,0.9710901463776157,0.8386123289098267,0.5692455854095154,
 0.7999283381269378,0.02262561642304195,0.09178381165081917,
 0.1373307615332453,0.3943807584697989,0.4753901953524132,0.918966125102626,
 0.3097124072234422,0.13408268918265387,0.5996715903535611,0.3329760861517356,
 0.5579307592748889,0.9206478315550503,0.603016192552047,0.720912420744183,
 0.8339378688220456,0.06343904106199538,0.3611458548082209,
 0.33007807614797535,0.28242688991512654,0.16725225406697164,
 0.06500335557043135,0.6221577047211987,0.7382668813399134,0.5040445235746714,
 0.3005445979863389,0.13509158545579536,0.47844841665462257,
 0.41542965725877934,0.07288098574598878,0.01660910256304593,
 0.08611627892623552,0.915000296871449,0.4481300753283065,0.07876469247073192,
 0.6185727208947118,0.9758146940230443,0.23236900399267446,0.3056910065278635,
 0.6338554907788548,0.31879021261041585,0.9445343799341938,0.7377031522040413,
 0.8302700140729683,0.7523337564415283,0.04069080677752979,
 0.47323853737610855,0.25344930372218466,0.717521143967533,0.4656749857291823,
 0.5269600809294346,0.5828677987135269,0.7604113105909739,0.4252000160186421,
 0.4669127961818169,0.33596713102166065,0.10368199415469803,
 0.3006882679777394,0.5221703718395021,0.547579820642276,0.18936821547734561,
 0.5699424272249344,0.715258853848869,0.07116338107137599,0.9122647394964549,
 0.3447744422375635,0.9136175204256783,0.19431333063258194,0.6176007214108888,
 0.4733643344463838,0.5708811260284752,0.41536089366613094,0.7733988698100691,
 0.2230367941685143,0.1788329996044894,0.5092448413639764,0.0448417935409986,
 0.1857703841474213,0.9650403216608566,0.17503030772667394,0.9757695033582314,
 0.07212767800823883,0.9487081618823104,0.8358704039420757,0.8839726841384343,
 0.7575742254236716,0.17454665528418056,0.22003881008275095,
 0.4845426160200045,0.6929138789181408,0.8670061795689351,0.8313930115715225,
 0.28034878486906345,0.13953836505201522,0.10660792536005737,
 0.6139314794829375,0.9059241811288378,0.42548259607258054,0.706180464213821,
 0.2799793561717102,0.27757794061444585,0.014330005333238027,
 0.9731674158901383,0.4364413303716347,0.10283549542529946,
 0.12111713127050927,0.6169165102654434,0.07581401207802196,
 0.8829019284757721,0.5700212027963535,0.8875564088214841,0.41861639864063926,
 0.7670465065804343,0.7681166842846833,0.9425523912676784,0.8277613900434797,
 0.4274376742968321,0.6164807680946085,0.20534098888485286,0.4823741680041195,
 0.1729550355380679,0.44063005289097623,0.3204603927702413,0.7780357394221269,
 0.7379644699456472,0.14354523287684906,0.10361703628270436,
 0.9070854718147245,0.6691497534783293,0.586146664928598,0.2097139266449567,
 0.27859714898914245,0.6528824232654858,0.6813092724546995,0.6219301694693039,
 0.6137400729121851,0.10983811977783775,0.34205821215564514,
 0.3938465312544953,0.8255311723716584,0.004765477266571616,
 0.7470647558450522,0.45547025907775196,0.2349775026471813,0.6853801581767083,
 0.20492008047776833,0.17172855784925234,0.1814649919161041,
 0.22541012483693057,0.08776563313987651,0.14649886694412684,
 0.6326959072015792,0.6633824980057303,0.4142797775913054,0.3181584563604015,
 0.9584401903003581,0.5688465019150971,0.31074616518359766,0.3776454973893637,
 0.26463090424515967,0.2334922528934803,0.02934446443177674,
 0.8094723171656492,0.9372006062267004,0.9026412306211503,0.1288148843696596,
 0.38357216508929415,0.8860852973353924,0.035130980208588625,
 0.22512505371633862,0.08478644795687817,0.9850746844304604,
 0.8391706264432561,0.9377395768223491,0.030815569828426814,
 0.8343163277420795,0.9999403005693392,0.5763903204343791,0.6309729217671488,
 0.567189050554482,0.368702515695416,0.4645354697962263,0.137596391261118,
 0.8920898559076607,0.37826984993324886,0.02191590702835511,
 0.44979468004820056,0.0190844958252383,0.2609886951611655,0.7957034738292339,
 0.06324037941674088,0.10365338953846714,0.8939920777881064,
 0.07173652744089332,0.31762047641198254,0.14425764602594926,
 0.5802804775364716,0.47978578753142753,0.4488081540798601,0.9825562853118722,
 0.4165276064808563,0.5705590083806192,0.10688306883830112,
 0.46127305237134986,0.17306925113654614,0.25632782199769033,
 0.8882179445530018,0.6423186165272892,0.7113306829921024,0.7625992866565351,
 0.03782636496669489,0.6163555177349991,0.16758956710525053,
 0.4825807440610057,0.6399313696324754,0.8845868404432705,0.8262760879861316,
 0.4971256178535851,0.41043335572075557,0.6963274961292276,
 0.27590758441976637,0.35115029630319905,0.7114223590571066,
 0.09844820834667756,0.6272987406942435,0.963306992206686,0.9965668693887201,
 0.7387671238283551,0.8900030601817228,0.6650075945859065,0.5434234763537199,
 0.05850327287977852,0.8518713507286729,0.4610686987875141,
 0.05648804570256971,0.784000131605912,0.9128415918838239,0.9217689704156364,
 0.1393001536286531,0.42883405816585296,0.3090409065385795,0.6644033296691129,
 0.26182495810347806,0.5406515279004824,0.34283784563219366,
 0.8707372774231636,0.31639123975443295,0.8794929685523739,
 0.19874559041685658,0.96765891338051,0.16777623979664402,0.2242271809667915,
 0.44399137539086586,0.06681159932010816,0.6512664018546589,
 0.01684322500059998,0.933923731833998,0.5179337674278172,0.5706535685576561,
 0.7101759065825783,0.23887028355748285,0.1991503429410244,0.2900809933244085,
 0.7719223592611844,0.9735607410091206,0.4648795926413154,0.359663333600431,
 0.693547938981167,0.682797038754494,0.05408565762937945,0.4855269535951189,
 0.34487293254166407,0.7608020430373497,0.03192283566879517,
 0.12813918233746324,0.1928595700196828,0.9909700630921842,0.6324048231355809,
 0.7401116547700864,0.2114871706692376,0.7394067666456103,0.5031578390262001,
 0.7224061907908339,0.5165321185072967,0.04755271272655859,0.9889179224117388,
 0.7985980280681213,0.35945565949473834,0.6605844003446709,
 0.05798705581120378,0.23272312447749066,0.7717421930314178,0.549731445245148,
 0.16013442070211337,0.6807218677493838,0.3032960768801327,0.5369672971501358,
 0.5711324938833552,0.6086156142253641,0.7092209195638045,0.9191004787741047,
 0.8150255216817818,0.3717363952005588,0.9954968592775991,0.9058607448137069,
 0.3540507126104877,0.939753943462839,0.9510500585029587,0.1148875397499215,
 0.28436980586573746,0.1457531927734559,0.8096023023341851,0.8009385393190034,
 0.7202429593769892,0.7946577778602294,0.2879080266853561,0.7829611273370274,
 0.21516398596655784,0.9126310776159535,0.5018090683895687,0.630284742957442,
 0.7566982589448845,0.8006485958728924,0.19782323388266132,0.3288175698780724,
 0.7372657589066333,0.6054370655768201,0.21027269722377806,0.3540468510330098,
 0.4309207564252274,0.9036048288016397,0.998594927070177,0.7663789046942077,
 0.6222484053351854,0.9814974931644549,0.24513914097463485,0.4404229984542859,
 0.47093372674171097,0.7158730048917922,0.9193738515446154,0.7189506611573075,
 0.5905897135259768,0.9790067580698298,0.6107099039470836,0.2784681966122049,
 0.39359980666650085,0.8080407216911677,0.9380079012698788,0.6653814707827669,
 0.15193087187541165,0.9185119451429516,0.1118561696466639,0.5016896110178334,
 0.7964505762422934,0.6363449258668663,0.30259188528776026,0.8685662301946679,
 0.7895984280279892,0.9578774297442821,0.2531665344552084,0.35090483117941484,
 0.40781930614933937,0.5134251171679342,0.8790607846073183,0.8362914445737747,
 0.0072026402062728145,0.5680062515097066,0.9341255849640538,
 0.9328291464594888,0.26260794936187115,0.8067658915620795,0.9868746871448647,
 0.2568975166729264,0.13620740309486257,0.9799935252572048,0.8435969478798186,
 0.16053792010117407,0.0306325153312208,0.01576635779515767,
 0.34685905096363234,0.3524272516960665,0.21514951153666506,
 0.2035356374238252,0.9416990151569875,0.3802431073186803,0.7302204829316885,
 0.8269363353952089,0.8558482848238664,0.20460341835566098,0.6113911347234579,
 0.3104945295364082,0.6813770994552835,0.7873833509655798,0.02072420466312852,
 0.41431256266027605,0.6611647726692981,0.5546523146428144,0.8238679970118159,
 0.37299510069865516,0.8269613104888541,0.4700846495933235,
 0.26849855756632623,0.312689813669626,0.8560810461233246,0.09506282662974952,
 0.7568625449779348,0.23514504021181215,0.31285625631409597,
 0.18699159928082776,0.07870929091282164,0.2681650516018913,
 0.9726306961064048,0.7212699233915021,0.7368676327936298,0.4808697848789376,
 0.6268182521747543,0.5039153530948823,0.7876783252225856,0.5302470989234257,
 0.7260649103251351,0.2850430637237842,0.9398473226394066,0.8485950430328474,
 0.14693756550148918,0.007855402222766816,0.6899012819822619,
 0.06198288937823404,0.7066819991572928,0.3589875818151187,
 0.24986117312200007,0.8720207890105351,0.8760322801239989,0.4536371629716901,
 0.26291913188936855,0.728798521873522,0.7069776511740452,0.8024291506928692,
 0.33817079797598293,0.0990461321653886,0.814509026010175,0.9051873419147248,
 0.4218128154975197,0.24929656457498217,0.8846949503762311,
 0.14813912760030556,0.0932176208669041,0.6625189284717002,0.5410424598014572,
 0.03293779339374103,0.5867447558866599,0.5017562076688573,0.7844994368264578,
 0.09281692332427194,0.5287584237276549,0.5957623763710579,0.6257168155077252,
 0.7681518282463673,0.1515379321705872,0.25875492833312363,0.3545492767562646,
 0.43969390221444005,0.18305203472963827,0.3617524401036857,
 0.16719576203697506,0.001281831884879292,0.859124725099471,
 0.7479677032012094,0.21301494772277396,0.49098838413578594,
 0.4602943673731813,0.6864366695374688,0.05586842547086324,0.2390436690386668,
 0.969735650870198,0.5790334963945579,0.2446837684678228,0.1998656085905337,
 0.7093043773607296,0.8144710194977295,0.4510766158102575,0.6655672891088398,
 0.7686472437338827,0.48898867630426623,0.2415205740304901,
 0.37814406035723025,0.6489691504343968,0.264287685205182,0.48305412819031635,
 0.47219644269328964,0.6648285382530041,0.0689200331090335,0.9130929716002848,
 0.9445753755658933,0.9468628150598546,0.5314881549503546,0.9464245957665574,
 0.4034237184104825,0.7669865895952881,0.4931095248621079,0.9049240391078253,
 0.7802511409410341,0.15671662171261858,0.2399848452448783,0.3947949154257172,
 0.22673618589677358,0.05434416955626953,0.925743880554249,0.9631282758735857,
 0.19617940566510583,0.427275709135958,0.5789413851156846,0.7570405358248695,
 0.9614860730658357,0.5228424309173073,0.8226117009425364,0.48837594546063623,
 0.43883216128930336,0.33880685225464546,0.9764271858780299,
 0.4594550266583056,0.543194563219901,0.3611973636011383,0.6647501828271009,
 0.5175020930491107,0.22378503019299611,0.24423617992469449,
 0.2752085762132189,0.15448449476092296,0.9516831549557438,0.7441569686642028,
 0.9676060894168919,0.40400066394172984,0.9692981420273765,
 0.49112068783674734,0.3444957311609992,0.6332329752890107,0.5325020240469924,
 0.26860352094707896,0.1363983916829673,0.9256468558753271,0.9731368201208634,
 0.4813104873426335,0.600485501625723,0.2799671853588447,0.13728892252081726,
 0.18060710564795235,0.8025013222094927,0.9273026564732574,0.2912129632555228,
 0.15036081644576293,0.10557231910186327,0.4558728139565551,
 0.6882484785386064,0.9449674593021674,0.41763680983873785,0.8226933726201098,
 0.6644002405522075,0.8896462644290721,0.07962507351535064,
 0.18762562476447397,0.3937194268562042,0.560320625188447,0.8346555127454137,
 0.6453535165552391,0.8870248256616027,0.39285927279034827,0.8165265453950536,
 0.5318064698786602,0.3976053635630832,0.19782961667211785,
 0.29586993140537254,0.743280376236803,0.15377189589098827,
 0.050072197222502335,0.8065706906146862,0.909985524744454,
 0.20295971938865065,0.7482084724552435,0.9800215287356916,0.2570837544354646,
 0.018683113133146967,0.24931641247029956,0.9463551687109937,
 0.7215751221375687,0.430403378112008,0.19105939330059607,0.17158471762711158,
 0.9385750589627853,0.7918166959721575,0.3689073853265842,0.24818775173888197,
 0.0995765242475628,0.7003456971256976,0.6117038702226981,0.10495707891995076,
 0.841147922021093,0.6980269257077015,0.4060214379683428,0.2666927364755023,
 0.15299811895318438,0.0883700242866734,0.20516905194483726,
 0.4106792999848409,0.7766617998337725,0.5733312589260564,0.9394929099239819,
 0.3874295424426285,0.9200573471507785,0.30073691774923417,
 0.18049683550606976,0.3674949651392354,0.9485981930262319,0.59643272038132,
 0.38155432398017375,0.4841569310710504,0.4735462858633115,0.8753208487674696,
 0.5245211548180606,0.14084767357086625,0.15697155451204559,
 0.35770857335407824,0.7708089856406217,0.7824035301810589,
 0.025784106025148024,0.881372427784372,0.7996052958728428,
 0.11036743061001386,0.6153056061510653,0.8598053861143979,0.5130678959505552,
 0.515346624743628,0.8433297689703536,0.0035590981821482792,
 0.39483891038889296,0.5501168279255889,0.5753991870863855,0.218276168876277,
 0.3850534403740712,0.47895594989968826,0.4365457823195865,0.6674760830937406,
 0.9228737092961901,0.8358484063635448,0.8731803996118616,0.6056771228381465,
 0.010789144663837291,0.8848968095551974,0.3462284610756854,
 0.5880431858008276,0.06460634000576504,0.6608621826367133,
 0.36223046569022865,0.08197562146730897,0.20617126162921717,
 0.4968511026554546,0.537853492377401,0.3509108797015621,0.8759847915024512,
 0.2502250816246818,0.9762439672667725,0.19678680148822536,0.7253618356082017,
 0.47136526688031877,0.9806805777091016,0.03792703669019992,
 0.15720056828140072,0.09986236723633934,0.8816705971847411,
 0.13279225905526626,0.7044154741941474,0.6208146965239282,
 0.18333015645661344,0.09556027180733118,0.6106520932887803,
 0.7720530596723122,0.2215890429379338,0.7602983414252023,0.611532586756875,
 0.5675234617026401,0.6206922556145632,0.6198569790742923,0.766616512008514,
 0.7886381518631986,0.19352973895899028,0.4191591210497808,0.9945857852259502,
 0.9333664549788313,9.298459624563771e-4,0.45576466883140787,
 0.10296906486733681,0.3713680259799601,0.4293485620639157,0.9093626680187279,
 0.2571810820047645,0.793051532478102,0.5986312333040625,0.4683810942275545,
 0.374792499871755,0.3835324155843921,0.17526666542500446,0.8512815177144898,
 0.7290793856604894,0.7647609439965642,0.8363759853063288,0.7435393295472594,
 0.5554861048911602,0.967433362168552,0.8130848549930252,0.3497161399215747,
 0.40360653571350935,0.8959898174594032,0.6456671689880326,
 0.0013130554215694534,0.5861439657116898,0.8043773069239629,
 0.802214240189669,0.08006141896361851,0.5426049201856155,0.5459284724121948,
 0.5475155266885596,0.5406250330194903,0.38853300722015227,
 0.41241531057397307,0.08577622568619026,0.5063764939582933,
 0.4939930083200832,0.8610957116619158,0.38885331821487434,0.8577342776060226,
 0.16104470163770912,0.033421579579666294,0.5933599325181264,
 0.07062166858815866,0.28283133747935885,0.9643891915173805,
 0.5887380260610484,0.22560597239176827,0.43126834408349257,
 0.1671234253231012,0.957672941587722,0.20830880943404895,0.3658778677573984,
 0.7944891547273911,0.6874894555700748,0.4566194580087598,
 6.732762879435583e-4,0.8183247587138105,0.43864209430450984,
 0.6599643444649868,0.6173370163275993,0.37368480869983434,0.4039806905930222,
 0.29153823209774476,0.6010153427277967,0.6469518918302714,0.8982736571255583,
 0.7365928873283504,0.39154308337361277,0.7628479694985985,
 0.23165516786248463,0.468235842574258,0.8248731790513915,0.9556882446229304,
 0.8324499482151358,0.24530339926057454,0.40147171637329904,
 0.39909405815876353,0.789396990379148,0.4857591586364717,0.44410611000641487,
 0.9866002012867597,0.10345324264778866,0.2973905878304818,0.832571171359992,
 0.6729357058547772,0.2200637762343014,0.6736934022794314,0.49792894759154793,
 0.08736005823961823,0.5462908473538965,0.003261898510675154,
 0.8837832847405166,0.11536738428713833,0.10801825293549105,
 0.7481265181349199,0.2592298152205842,0.2956332745886875,0.44630728771265804,
 0.5733511606294707,0.9729369002633418,0.649421181299758,0.3588985970116432,
 0.005588796222820269,0.9388473268072595,0.2292653785635822,
 0.28476290437936513,0.23443623903417066,0.8830606489560099,
 0.8022611124633545,0.5405988038105347,0.4845348538928854,0.3213004622103468,
 0.11676084232274797,0.23656387137754775,0.11991555335631598,
 0.7802963743104379,0.436657225353532,0.9080099261421682,0.01538617620207361,
 0.43954099359358745,0.17779440340299169,0.6717956488762039,
 0.35447096535891376,0.572888310383802,0.06804395099950722,0.727191363848978,
 0.10100939188864111,0.4653765674996475,0.32726970533716904,
 0.6081052459128328,0.8224174384979882,0.623885498287668,0.06115867738029701,
 0.8972087460161838,0.05783564605751623,0.8686599786405486,0.9228732394193078,
 0.1543841694948861,0.0692583694276353,0.19241893027246648,0.812627450061095,
 0.02744897605697849,0.8629521694497869,0.4049532010783904,0.0876382446678351,
 0.8259524007529341,0.8056419537554349,0.3113353794116255,0.44791592063325936,
 0.3990500920653648,0.05594959877607908,0.9683295436804111,0.5812360584341691,
 0.3630061921664386,0.3462560863203843,0.8286431402854575,0.7327674048690528,
 0.059736749508339404,0.8517167462174675,0.05311869004016123,
 0.6792233472422691,0.6202722748552347,0.4743437163396914,0.1255023623052779,
 0.2469710809914848,0.6625591479356417,0.4734355475157671,0.5915493573220794,
 0.5521245016274076,0.4912800127481517,0.12835981007563158,0.3137644689198721,
 0.8724065326659003,0.5551152908244932,0.14981732802807513,0.3483920605035925,
 0.8716651459193926,0.9935790651856696,0.04506233658564329,
 0.31632592797015935,0.12414800023610806,0.08746009838066504,
 0.036777815250180756,0.7295116033945913,0.8003837895259088,
 0.05313158380421501,0.8761110444041647,0.2016350990559641,0.418900818724834,
 0.17137936052686142,0.3225632656547962,0.24009582806370522,
 0.1448405902407699,0.006924152677775797,0.05744280782721389,
 0.10411095489128863,0.947905591213838,0.936849385721582,0.35649404944389074,
 0.42258028359242017,0.5981760729823655,0.05562383210074151,
 0.5107970266406904,0.6063159632704883,0.48714651611857895,
 0.48904861491948515,0.6363915188869237,0.3477585974909819,0.8182833872020225,
 0.422210932836121,0.4616347231254564,0.19012233045205196,0.6528029549042291,
 0.3471243018247938,0.2795626504669647,0.2868192149651543,0.5292584494200221,
 0.7350676147712814,0.9865637906339648,0.01699085768034836,0.9893526400719662,
 0.6736330909066222,0.23315405812749146,0.25328682420048454,
 0.14280569351674555,0.9885737021540849,0.7581917410772058,0.1915455736413444,
 0.32178030316474393,0.5675170337496935,0.16567063348207012,
 0.16500733067367124,0.01539510230569463,0.0914448382232318,
 0.0833427406890358,0.8348365945583245,0.5539576808707688,0.7700649279940094,
 0.5333704736122566,0.02033273409770575,0.767056550956665,0.28462038801562306,
 0.928696154457363,0.5382979302313742,0.8410917658583025,0.5539612000289044,
 0.013565532199380392,0.17376669721854687,0.3718641252037106,
 0.7080356443434469,0.5935383257123619,0.10452407226587956,0.5001658059408775,
 0.2531042372170891,0.5023915498810863,0.12263642929080987,0.4662041526276587,
 0.9246843883544931,0.34827784090063485,0.42870816717588434,
 0.25815682885541447,0.7148839993626911,0.6437813501950816,0.9181939427073457,
 0.45866530896315905,0.9677167047931048,0.308372919704752,0.7939370220939271,
 0.9172709803378243,0.8115027652334702,0.35154700642749526,0.3783252696742869,
 0.9463699271190513,0.43623693671169217,0.5019335113984148,0.831021442878409,
 0.6241898017653366,0.3968738294349139,0.6236653201004795,0.8368019869945088,
 0.9162648282892244,0.33650842129755176,0.8058449312846497,0.5174739359895381,
 0.941672301684509,0.9502768059092961,0.10274850999677754,0.6508990360309048,
 0.9210329653750771,0.14826333059118968,0.6158415838596676,
 0.23795214539945364,0.5337137114691746,0.1644138652099082,0.2899303861621839,
 0.6027067153412045,0.27286060784025046,0.5784170039238334,
 0.12898923664311934,0.7418374912702678,0.7025586693794074,
 0.22648362276044287,0.4779030379775332,0.48469812448396177,
 0.8762734234938273,0.8517857313133427,0.9975363224475623,0.9561907997260626,
 0.6793637971619636,0.590632490030472,0.5478686079021413,0.36163212453978466,
 0.9960226515274908,0.3085083503716889,0.41162644244949687,0.2478477571866895,
 0.972604786923914,0.12857457059747524,0.9962205893566377,0.2919722119422459,
 0.1839511310954728,0.3288895491651721,0.27084434809444136,
 0.10748403950577334,0.7070833815958215,0.08210319016021117,
 0.6001516909563362,0.5246311274981131,0.47785356664667966,
 0.40261804235230525,0.9191650581267317,0.7090042199529117,
 0.47556386830019237,0.5129241725884289,0.9530014312124718,0.3781584645946303,
 0.71389613488365,0.16447962470452826,0.801110077670365,0.22902049316752715,
 0.8395623421216114,0.5865505483569741,0.49713739064500795,0.5220903957044997,
 0.9231587881304048,0.013475900920040473,0.49008595023043755,
 0.33898662054452866,0.35095953263025037,0.9045257894592285,
 0.7148821370504314,0.053475961527042704,0.19564109973635568,
 0.5833319297873776,0.2725688080448042,0.2679235216935357,0.5198203732240472,
 0.2957664188346141,0.4039796807840539,0.14072251975077343,0.2379454085399083,
 0.9919452564225953,0.473766811924916,0.3535415684191312,0.23649398765661167,
 0.8392521983648041,0.647184284410109,0.75480860348897,0.050560251779579124,
 0.4437531337352968,0.7563241851270522,0.957837164960173,0.6960878065662423,
 0.5579123665659882,0.38591185209722345,0.3015068988801425,0.9028861061538309,
 0.2935711155485412,0.6172926738404808,0.26922602672805207,
 0.12072759549865952,0.07923716468291064,0.6073505050187045,
 0.4243802534089718,0.7761944667296298,0.6837099360795753,0.09773297919968615,
 0.6969134673558168,0.03490702354794739,0.9734277840573704,0.8784755068593467,
 0.044343689293791955,0.5380727502055683,0.7888958473664518,
 0.6534475720385672,0.9030637053517828,0.007422028576933082,
 0.9598276389607205,0.9604088146103851,0.523896130432497,0.5036355923934557,
 0.9657572788955797,0.9246017001469111,0.9522540948863193,0.11170624752952862,
 0.3127125070279494,0.8270258603306211,0.7213327797661122,0.829912767887261,
 0.25653745044654186,0.8400370302484771,0.07339283527585272,
 0.08253859981965828,0.15070968714419009,0.8263982280021309,
 0.6244268796442916,0.8826631891829665,0.9244498435712005,0.20907745672521016,
 0.12861941887015904,0.3807056976706613,0.808172042943101,0.705696226462077,
 0.9613475961976451,0.909275276279836,0.19358532964158348,0.510165801371816,
 0.5055357591875449,0.09813124221498026,0.08771799780381628,
 0.39123783803881973,0.8939277803669329,0.017906826513223395,
 0.5027369911076421,0.6879504506687936,0.14090675497528637,
 0.15965787483724392,0.11477269437779358,0.8856124362026102,
 0.4305931766386121,0.8125697126242902,0.9659729433508749,0.9712757491321573,
 0.3845430101590247,0.015327979612589138,0.9781314167070495,
 0.9128749218722286,0.9578514115476996,0.7168050854391921,0.10789464095513812,
 0.30809391209487347,0.012461635315782038,0.09844630950854838,
 0.7171940475656661,0.11014261077703769,0.5485104165181256,0.9899086901645737,
 0.840162908484371,0.31696372375100057,0.6564921007669939,0.4759649138367702,
 0.0919911367288121,0.7612254214236401,0.35629361414501204,
 0.45388130900780155,0.239161942570992,0.31777177207681784,0.4611720425010265,
 0.7092049842463806,0.8351148292624577,0.7745338338248822,0.3896803019506928,
 0.08236380488396344,0.4862560167451371,0.3583305373946787,0.7145377049127838,
 0.8458244561143081,0.7192611441510082,0.6167170151155065,0.5637963660853843,
 0.6448899087594024,0.3732542517076023,0.881796957532508,0.9523524456751085,
 0.9150764888901426,0.9286686854249577,0.3465702259099128,
 0.0012685893237480705,0.7888253134569887,0.7614279169356344,
 0.8084229031546056,0.18074541153208756,0.1633162578372892,
 0.08535878682383691,0.34270311971560175,0.5531172886675773,
 0.013480454954492682,0.956996380926976,0.45885944015864333,
 0.05717502425890886,0.5743302067870942,0.18976648687991815,0.601343348064042,
 0.0036250084352803524,0.11085107308592757,0.699955986252425,
 0.5361180168468872,0.29687612388228135,0.23135596176759,
 0.0011845047849484702,0.12858639116704662,0.5126096867002919,
 0.8350000955293194,0.3327211601543074,0.8926268961556652,0.6902013884603373,
 0.9364674235691683,0.11782354953266794,0.7526132530015565,0.9922032673918968,
 0.04535981048889981,0.7624477114085695,0.9974112871051198,0.7216415834148128,
 0.7715455268877182,0.9706517237359517,0.8046216661733384,0.5644556788475708,
 0.07049979920917837,0.10661810136999705,0.49985186341080956,
 0.3053179327030596,0.180698104494581,0.27431363643748896,0.10986000352947067,
 0.0704631913200523,0.9195920835214957,0.8771692553465306,0.34752308879472804,
 0.8573061734912197,0.7788279636786797,0.9671001300809449,0.4425743050116221,
 0.47865479603247696,0.13975837232919486,0.3771978689681995,0.706013717120596,
 0.4813336014724935,0.15235897841976298,0.9971419124948615,0.7905819140067667,
 0.37170134140750155,0.3910336708283487,0.17068671242657407,
 0.5409343830992918,0.23071878581683014,0.6334866295492317,0.3581702398399017,
 0.27965589412414793,0.9653588866591686,0.4211960652315013,0.3706353973775054,
 0.5836047293710096,0.4037000981908525,0.8907277822435531,0.8167124063836827,
 0.7219606319506225,0.7410913481661775,0.73565762700817,0.19699273494350178,
 0.08259051773800241,0.2894691990612869,0.9638822797644736,0.5731148645384316,
 0.1492665943501701,0.47311908311858397,0.526392029171526,0.45273226509044506,
 0.9919777252704138,0.13175416953271502,0.015850238691747287,
 0.39269175950470947,0.23831639497454749,0.9758860871489642,0.372480237506129,
 0.1006125596204408,0.17813164427066708,0.6781368571799224,0.365513048410781,
 0.23686933630256934,0.8447985576383372,0.7595741683011955,0.8428257826085437,
 0.07646751023389764,0.3388473060689836,0.5633927092808091,0.6183328972459933,
 0.16999453692481425,0.24984493307417233,0.3724327524084128,
 0.9670997247358968,0.010846854848970855,0.9903372822741727,
 0.9762906640490198,0.9057992677054301,0.42505746922844856,0.5046836769110262,
 0.6059556216779629,0.6064154949560082,0.514059120063602,0.2352648924119407,
 0.9767836586946195,0.4665018974900461,0.2320467438333833,0.9524210615617753,
 0.6591578145346274,0.39625688141212145,0.19566041628337172,
 0.06202644735496532,0.55575179611483,0.9658099001333603,0.9547583970749731,
 0.22855277915256889,0.4036002658376099,0.027286612552066325,
 0.919461414326002,0.8107075393287684,0.2935972820166699,0.07850417374520435,
 0.7468209190849076,0.6590297547615017,0.31924719135502966,0.6987775439050632,
 0.13385452183832647,0.6444036394506247,0.32266660016638404,
 0.8169592014146722,0.740275257525169,0.6197108692132798,0.4599062127384549,
 0.9685864363235537,0.4838721918413983,0.1371432448892832,0.5895740943639283,
 0.027804140865502647,0.4467621784417638,0.8685252340607029,
 0.22608206626399352,0.09279756659766414,0.10441917376796062,
 0.6840972471210127,0.26725946353658325,0.6777803449568842,0.9351771920760366,
 0.6904862441381486,0.7487197924312933,0.15783767251472647,0.6460641477443373,
 0.009489730948069819,0.19876057268781533,0.8737140327563824,
 0.9447525120141552,0.7328237108525916,0.8460150238181646,0.48934990314674554,
 0.5442873449858303,0.6056543606361527,0.924394545391133,0.19174734145577488,
 0.8854447080432304,0.4012448437309315,0.7724569826360619,0.6234594772662628,
 0.6671153776812416,0.3581061089346146,0.28983940963427157,0.6320830141898268,
 0.9596259342464498,0.07034012406485035,0.23455894061443439,
 0.9244096349630282,0.053676556052945745,0.9341280676326268,
 0.8294474895441412,0.4911505595773803,0.335313625979509,0.7943226686523456,
 0.9115261857977862,0.0993612317144551,0.31373392136861455,0.8221506657875535,
 0.029254963138492518,0.011032148203892644,0.2177748263771715,
 0.30870294059669046,0.8450396829504814,0.46060034591734444,
 0.10803418948506616,0.7785675965547628,0.9134450699204398,
 0.03725514386680684,0.2340305341837451,0.5543828306954409,
 0.016543404242798654,0.7088744842697629,0.04367997933931955,
 0.09646767010309443,0.9472853911258863,0.811748597301486,0.9829167594026115,
 0.7799904220776803,0.9267292744006019,0.9677937384131043,0.062100046735447,
 0.6652322109968816,0.7718754233355871,0.46789752818821984,0.8320208148271492,
 0.21498238770643752,0.3800995185092457,0.2437119391703767,0.6058959344515792,
 0.8812451507275405,0.3151428142575954,0.7191340598109777,0.48692183827731994,
 0.060891653504454024,0.9056758747167872,0.21595367838473079,
 0.053209622194785866,0.7037118428215954,0.4902124330883456,
 0.8168262294360967,0.6433913827032121,0.23636658057553273,0.7577623374821075,
 0.8513983058890683,0.6179696574388925,0.09984522614974833,
 0.16463306707845038,0.20783675629706522,0.237691095238464,0.2941220870210809,
 0.8541250623956458,0.8032723155698038,0.4918633115811728,0.34529689972285327,
 0.8685504443883997,0.23763359266026496,0.33416972741769735,
 0.043172677615396116,0.34687910095891694,0.3792130317242829,
 0.5369173338498046,0.025949675286855878,0.46782843101272764,
 0.6263387250580428,0.5904268344645397,0.9638868803753603,0.736008132545272,
 0.7481374840807994,0.9114164887216628,0.8099878564461579,0.17317294500255476,
 0.6368273112002573,0.7034545902325107,0.7263256572615335,0.12942245456961932,
 0.6476005177780203,0.2611763951285063,0.6587369682219999,0.8605878306593622,
 0.641849506124713,0.3144298160169686,0.7877189157558343,0.30616896617295364,
 0.13278042052938244,0.44746126100891703,0.3463771276060973,
 0.6764957348987821,0.6815818696874223,0.686089079259744,0.7372937465898979,
 0.0790729036346205,0.38325513722959403,0.18092905719053298,0.984631121820918,
 0.568797822119546,0.44217009713513333,0.635266607847255,0.8618936997780734,
 0.5379606486249956,0.04973338646937531,0.8422231940849456,0.4674159327364956,
 0.7254724084491421,0.7667100444182878,0.38311690152933076,
 0.10570684510858208,0.746154415106874,0.16240438821837944,0.6123233674014383,
 0.7734420493287425,0.6197124006854604,0.14168895976657858,0.0951715141451217,
 0.9635604483203198,0.9059305753908267,0.9275514128430684,0.7041879191268097,
 0.9541798546252115,0.9954402477534129,0.2298434198318413,0.621376362230995,
 0.9466560744012439,0.2320637364630148,0.20673455148039355,0.6209750662210898,
 0.9535281913204301,0.7241690989602508,0.43050913400883617,0.546556462627336,
 0.22985952129267606,0.15071452522177986,0.7230778654266694,
 0.9382829595478672,0.7435262084378454,0.7710564233709314,0.5127943489124462,
 0.9933737633115433,0.05820365523775728,0.037598147746986554,
 0.6858015079937835,0.3740582982673173,0.967032404467496,0.7995858289616954,
 0.42094279986146177,0.6898051445003219,0.794359698261426,0.3072845217629985,
 0.40514098150459665,0.32568949647452905,0.5818013366407386,
 0.5046447422963979,0.3910201556865872,0.9947735708652996,0.3175835277128237,
 0.46029252874898674,0.9591538060040181,0.3510787583957016,0.6856859445310792,
 0.9597202544770375,0.9230777407118247,0.35812057085185045,
 0.013424826327521955,0.0454058612210978,0.33556337186536667,
 0.984136589643481,0.7204934721219157,0.6273374134701296,0.23344470269447337,
 0.13559758764057805,0.4798634161600073,0.5690727667026908,0.6420734858999315,
 0.2356716079946347,0.7170158351962959,0.05485085119493194,0.2922930610845409,
 0.7683473842550153,0.3121133499263413,0.4468842801018059,0.7305314645524005,
 0.9408573156300786,0.23344452584571462,0.8423720662003676,0.2074086685444526,
 0.1191289142239863,0.5304472849321883,0.5935471341748889,0.7497164349607157,
 0.46865461791993923,0.9955210156638181,0.41999656744931413,
 0.8660042612522615,0.02585026205561003,0.34908291802990543,0.835895366597539,
 0.3351832659111097,0.940624288491616,0.018328477938150955,0.968739631389381,
 0.9244247695836622,0.7899920490212211,0.9946274930112938,0.5888784943693892,
 0.6747540619247606,0.5598762191279318,0.202596646041035,0.9346555668088596,
 0.17879317201638556,0.5143279482226645,0.44849529381200504,
 0.05559335262734455,0.7046834335031924,0.8881225228738561,0.8867096454120337,
 0.03984841290493435,0.572109022496536,0.093163260177621,0.5034513458047969,
 0.13642644297069328,0.8381346525525305,0.5551582491549532,
 0.05539366210554442,0.792573558746813,0.7009575101527586,0.9695656239839392,
 0.4782871502738315,0.3769306298945291,0.06272681969327543,0.5881791647034158,
 0.6552617084322798,0.623779254209896,0.8748102962866963,0.7816039049062165,
 0.3180517837150926,0.8010906386775731,0.1020192271381527,0.06239226542761245,
 0.0012147999107996466,0.49387319427417675,0.766377693044598,
 0.04054779251910701,0.6416635059177999,0.3501786720551372,0.5443440584139528,
 0.37755102310356037,0.3699119468662434,0.7916105196811669,0.5651965173652832,
 0.3457503825781302,0.6671917053646219,0.9376816506511452,0.630733474476021,
 0.0916658428480811,0.9457042240216231,0.4500154627807573,0.05646339292410951,
 0.5672409338252105,0.7992199118113412,0.0067679634630716246,
 0.8991016142510826,0.6456713823478465,0.853099365809054,0.4311553537219137,
 0.1765219979430489,0.7308535947634244,0.6875658386407152,0.8897857125075247,
 0.49334145829172404,0.7689231478672658,0.849968240399584,0.16240249112500194,
 0.8829493834724368,0.8565323353187373,0.4730881927426922,0.6657738450530855,
 0.33737140462378745,0.6766418411401095,0.9138509631086679,
 0.45462512781332653,0.2977942372846135,0.7903581175712041,
 0.05236393018959107,0.7879800911656778,0.021655188878434872,
 0.35430979595192336,0.18641437467998145,0.23826559308629314,
 0.7199987464021769,0.8239561229728375,0.26588295870398815,0.3163251126847828,
 0.2487953817759646,0.5148882922324572,0.5048511569833088,0.2235733016894984,
 0.6625030611446624,0.4700753809638879,0.11541227308374824,
 0.34589000777701606,0.739229913741516,0.7470387865032877,0.945346022150938,
 0.002195712065131916,0.0683577266879134,0.46554276567099373,
 0.5600233740512572,0.9981816698878452,0.35054991904638344,0.5395767237685871,
 0.526009145916176,0.24571008239893588,0.42236870368409873,0.8240245214561658,
 0.5277582983165585,0.4239374990781173,0.43159254654784596,0.911099837039958,
 0.7407618411055794,0.9112072458122593,0.8109093815595545,0.07367083305786859,
 0.435219011954013,0.209578067335373,0.7256347270403355,0.40603811293167114,
 0.309256387351747,0.7994305160696985,0.26027109354514655,0.45049274724384847,
 0.7044159420891069,0.2921890588013042,0.7309352553287706,0.20387087627689016,
 0.44719912164514775,0.7996980813177794,0.6649367565715634,
 0.28665729645436233,0.4713060643665239,0.3072392361330304,0.822632677910615,
 0.8025768641700585,0.054212148934874915,0.6375964501303528,
 0.4510522467901258,0.5805136836645167,0.01917954738827532,
 0.015543156304518746,0.9606374577011906,0.7583951740225521,
 0.37369948704059064,0.848388945477465,0.5148181719878828,0.7544514602304169,
 0.18000075787740266,0.3673068842177354,0.2728810703054215,0.3022064784054096,
 0.2504740473853929,0.8716347005331417,0.24926439397748867,0.461064326500103,
 0.5189608049459229,0.28530004730080094,0.8520651708391292,
 0.48379649741751174,0.5723660713253492,0.2821773149013169,0.9032265790636852,
 0.12513148905148408,0.019265298884973303,0.09889457571940363,
 0.5702977316370137,0.5814237493887313,0.8830613432246147,0.10612532134232233,
 0.16579668360500455,0.9931554714447496,0.7042957571459253,0.2656759983569018,
 0.06885661953480826,0.08974279469400959,0.6747851597783079,
 0.1848882563393851,0.020878241691777655,0.7315625454483173,
 0.6834423395849312,0.8496163108491673,0.32882287192782755,0.7578012610853717,
 0.20827871837175105,0.24294394000428232,0.2269006843630822,
 0.38884976166184215,0.13016955348799641,0.47017211698866135,
 0.6868349203644842,0.3392138825209844,0.07257262009397691,
 0.11223725543569707,0.5359172418040163,0.3513393991870494,
 0.20902527513616287,0.941861210761446,0.6884194811880757,0.12043408136593514,
 0.7311575942051292,0.8134883590642735,0.6928924583315809,0.9402399186667313,
 0.5201530103142984,0.5897569964408313,0.06085480132804222,0.9913526067197284,
 0.29400008670827815,0.7172210133013002,0.5767258172624127,
 0.49613527904024135,0.35787637570338904,0.20488412733889716,
 0.7233598262294749,0.41479354882969033,0.11343440912248326,
 0.25101098539303013,0.7188705268958798,0.5023733651063914,
 0.44384490086898587,0.26890924340917965,0.17923511036646989,
 0.018220622953566745,0.28947142750640276,0.5054686640646472,
 0.24040030759136277,0.8679469282421816,0.010312534200750956,
 0.943861364353805,0.7331173052663986,0.852860377273372,0.8669740045898899,
 0.6391748467938142,0.3717625691359401,0.9552187074214807,0.3096602284087211,
 0.7031984800826581,0.1766151910527195,0.18139358132843908,
 0.07138023036802255,0.5705447343181145,0.7645149179980966,
 0.19744421403135304,0.9159485471589873,0.7506569861407769,0.5545385390303956,
 0.2687581078644464,0.954870285825832,0.03154350906480574,0.49040534549805326,
 0.8683530698343438,0.24905115486761464,0.13365878290746014,
 0.5231671446458203,0.8374215714533872,0.41724863065352347,0.7585591435212462,
 0.9730789682359259,0.6955120705134126,0.31909640049608845,0.6731567485407149,
 0.602589063839536,0.538945587506436,0.8179127196409536,0.8881358848599588,
 0.7587362482392553,0.2570505628674342,0.43418217240832946,0.8453222876900235,
 0.8768143810530941,0.822911498248435,0.4208725520977793,0.6047539550740171,
 0.5379790189346287,0.9494949343937328,0.25098742728993595,0.2755054806656161,
 0.8103333815592921,0.8689709756206552,0.1923311370839631,0.8234370719737917,
 0.5641926976870398,0.7194994391863742,0.8640007450337043,0.2053408543092481,
 0.7943647797706597,0.5844455115331068,0.6216709984851935,0.08804609397449015,
 0.8681998619588442,0.2933402894197651,0.6932083559542913,0.8942093406920091,
 0.4089770160436068,0.22776269474099176,0.4010670805301909,
 0.14003196346799407,0.15426461716016748,0.5258942493109883,0.156652314229055,
 0.5607870742305989,0.7994076683593199,0.24158047335228772,0.6856343496883572,
 0.8552320453965779,0.15117219029435724,0.1505357231742096,0.9012339808822089,
 0.07129025726153382,0.8454048405824044,0.6850113478014714,
 0.37962989988166596,0.9969008641595571,0.697528398647875,0.394746968444022,
 0.8562018706444603,0.569969906948872,0.541813100645115,0.3668235264810814,
 0.8302739458773795,0.9750119352599803,0.6963656184106859,0.04354997414340522,
 0.4830173737819844,0.7307068310555616,0.5253867619502645,0.6363469790598171,
 0.762436239721122,0.7428182938090009,0.4210661374002782,0.5967781208778625,
 0.4666345749702677,0.2998398023812129,0.029540728038153774,
 0.4079982384969236,0.6364807688116199,0.25034901546070665,
 0.37995199258071777,0.2486518479922334,0.1853219900627957,
 0.30138589451145015,0.7422882522429208,0.7258770743120896,
 0.21788943223022683,0.2485297738028759,0.6930446547312722,
 0.48831591924842266,0.7344149010455737,0.8494117739120765,
 0.43151446741650545,0.7038761608242042,0.17988349364538114,
 0.4039763994277791,0.5382088981722786,0.8082677450645459,0.577289067964387,
 0.6741283543457603,0.26976149796126503,0.24173311772767236,
 0.8612606839666921,0.8064014493081023,0.07291741956242737,0.8420768248930289,
 0.7969497957756619,0.27201026993205124,0.6259993241028715,0.6488422755022212,
 0.4688752865386948,0.8946765087094639,0.9769301515774409,0.7946597498316573,
 0.4576557430498954,0.7251744887420244,0.41219228853312406,0.8561326634729809,
 0.17391144380813162,0.2835361980199518,0.7887053962659721,
 0.12190098159358509,0.7229385271754112,0.6680705690459642,0.5482702347760626,
 0.7493307675421204,0.5134910282573797,0.5276175182823558,0.26693640049037715,
 0.5659072471033894,0.78839831599242,0.6116277400498229,0.6078790253270432,
 0.9242889296850901,0.3896956879226977,0.9339519937252333,0.40017542104958004,
 0.7206560379053727,0.5403483822336869,0.38196599155344757,
 0.09113419545210977,0.4930978086525133,0.6486886353012904,0.3861307941415526,
 0.830319325525573,0.7510949511400024,0.7848356131242478,0.1914811999909486,
 0.7341066560777416,0.9881716679880477,0.34585088744817294,0.549614365730007,
 0.544583328595525,0.2531087974816514,0.999114869750286,0.7801976153741365,
 0.9586443036896589,0.13400722061110737,0.9581245644306653,
 0.34383932732660716,0.18862027675852744,0.9000838029160813,
 0.8556479368148547,0.8545154896552551,0.2559061897875444,0.2801846746708664,
 0.19546373003822937,0.5097539596437954,0.26163612236707046,
 0.35626834701603893,0.4039107372767575,0.01228194445844566,
 0.5766666139007678,0.21335852248277665,0.09323877061253538,
 0.8220805754336354,0.6654486114461877,0.7944661569390613,0.3080595406189264,
 0.4816643790583566,0.48890474478705503,0.6772512263177122,
 0.11444170282903465,0.6957042217173932,0.8965649396409259,0.9816230633492156,
 0.9926610834003405,0.6802219716240163,0.25016597565523724,
 0.26679872760707746,0.8249280872343296,0.32571581743970257,
 0.01446276895044063,0.4086969289574922,0.03467083845762997,
 0.2841359932364994,0.009495683378137998,0.45322333178617946,
 0.5027860180279674,0.4553296630860034,0.7597412563072178,0.5424312399385035,
 0.7551128641777309,0.6976588321888727,0.9659011711413811,0.9938186030342794,
 0.09275941397373544,0.518718921716976,0.5781326092380147,0.4604220545562332,
 0.49751958007529584,0.2522344562344697,0.9514072893541228,0.4507341557793525,
 0.07682287318237324,0.22364525134898194,0.2154322235181818,
 0.33741739444638486,0.4178979034765826,0.13073671939091647,
 0.48742656882416935,0.5725291096894405,0.4062582253843372,0.5694295526025298,
 0.14347916267772476,0.9265830866335236,0.6476289454424432,
 0.011047630409053655,0.39095783968437237,0.9848701321328028,
 0.6301549899123308,0.10780037627283545,0.8988863778982779,0.7066940427039095,
 0.6592377334201348,0.7616468605139461,0.02179669473960444,0.5435286904302802,
 0.013575313721328586,0.996489713316039,0.9644244372112978,0.7165844728196301,
 0.728678852433158,0.9028977444478921,0.17660078790895983,0.4543836983343864,
 0.4795993167149798,0.5292758495625167,0.19942104665315463,0.808089004395149,
 0.08372522878641309,0.5856278970226326,0.18573740996080113,
 0.9604448004993901,0.09084967296428859,0.5194467469398314,0.9141913547702081,
 0.9080693010665711,0.8427979009150495,0.2722182325227579,0.4655177905579775,
 0.5522077993923474,0.19538759822950424,0.9991591640428326,
 0.38296563887607094,0.8065933097472914,0.517721679307095,0.2855719659628292,
 0.8254870724043545,0.9866071447122127,0.38376269841875066,
 0.17863576096213274,0.7130421399861216,0.8726218347387198,0.979727299638165,
 0.4955918875505795,0.013623302589525266,0.7930171566891877,
 0.6335234290372482,0.32065148576087565,0.28845787751214547,
 0.7865405331696949,0.5725183785773872,0.3586077468745824,0.9283757163558115,
 0.6242527236952027,0.47518745536225493,0.12005318246313257,
 0.14074702090807856,0.565841948172,0.9843231523755922,0.05329996573415474,
 0.4348544299084883,0.46935786564308757,0.027167622881338338,
 0.19675562034035277,0.6854269030045307,0.10984566415634522,0.513297095150847,
 0.2417840704546086,0.9594464877743262,0.038709519740043885,
 0.05314947260381808,0.33665759955929364,0.6548183274035737,
 0.8461797363310769,0.98459411297197,0.23054926173452794,0.3159162508615607,
 0.5568353941787622,0.6238459169377794,0.6892914445221194,0.3066649754964579,
 0.09772574298634007,0.2774823804250647,0.3018780103469637,0.3404210593325302,
 0.8666924899706635,0.8242871344168172,0.03140869124033474,
 0.06409531505329025,0.9445704563031591,0.955364980935529,0.705182500632264,
 0.49286788119472824,0.5619203254355367,0.28326398713372847,
 0.6513060013414581,0.9154529508344353,0.9223571875404694,0.8964420540759435,
 0.32880985798200946,0.22014019552750774,0.41502805563861145,
 0.06694482475132224,0.7190242084557754,0.9504999143113307,0.6084064525535293,
 0.904985755265752,0.8127344154910756,0.037789838051552005,
 0.05784765525747926,0.6383516590740469,0.2533292000200561,0.7833408831850566,
 0.7145821979230771,0.8686557477854676,0.5108886097016655,0.7042092495364135,
 0.7406996332645938,0.10946353735281367,0.5002828407711439,0.8528684239569925,
 0.05157332874278142,0.5033171822685361,0.6618223666632447,0.9758340167076763,
 0.7593740285235379,0.8669812580519138,0.7084167995987632,0.35610647325708034,
 0.44595949733118934,0.6654393752533182,0.5974167510754745,0.8208123941333353,
 0.7967275458360874,0.49706438289018495,0.5289850348071223,0.883100918303933,
 0.5671986993505382,0.9079742467817888,0.13031712450529043,
 0.32479858416578145,0.47992183366569385,0.13718130969859876,
 0.11614946995907971,0.7999901198244568,0.11063623604259809,
 0.25857628861252857,0.04211290136973456,0.2920275552150091,
 0.05922051520947558,0.2868775879085076,0.7830335895303727,0.5841855720838013,
 0.30763333000828497,0.8195332188805219,0.24405954474411096,
 0.1433527407770171,0.5815872213811768,0.9354534215205942,0.7840536981937324,
 0.6260884292402982,0.2637170229583794,0.2545368567513452,0.47909781982551525,
 0.7815508733567874,0.06226614011760334,0.9169835567689635,0.7453600840274969,
 0.6560378182612918,0.6114963243710807,0.23861832929935944,0.828341806824035,
 0.07754828699138305,0.1160483875788696,0.417095584141589,0.9150877075623841,
 0.3088624303880383,0.9942948501639653,0.24226774624197245,0.5000979059783416,
 0.5031451685133502,0.667714340470343,0.09677663544889625,0.8676661008514152,
 0.4610424308399651,0.7823071075680379,0.6823082925661598,0.3974615445489087,
 0.6216397947748252,0.34567082907710145,0.06842986256444306,
 0.32035682657813547,0.2306322146156793,0.43145809694639103,
 0.43531697461014196,0.7198892786413074,0.9940167223298828,0.2525904799128119,
 0.46098359382731724,0.7355938963966168,0.9853780107034679,
 0.04098310788529025,0.43980089858307425,0.9717876339096932,
 0.8049299257871193,0.8657250901579557,0.4987816199854759,0.983780883965926,
 0.1441067873733991,0.9850886102320813,0.919252897255051,0.5349380025134394,
 0.1492111654861733,0.7423307630184834,0.6682187624546978,0.5167611829819907,
 0.33869117836640106,0.869626842790394,0.5910793945308138,0.5083829600239003,
 0.6687869388997897,0.4711980141054898,0.49799339107573815,
 0.05858726004380699,0.602146271279466,0.4739193614955449,0.8086458681235595,
 0.8756029376345559,0.8926442766601115,0.2963761149660209,0.9293329881605257,
 0.8035071173969863,0.4082241167647135,0.9551286272377274,0.2138239394669943,
 0.28381054850052745,0.6699227122436291,0.9463042834062215,
 0.10337391016227593,0.32488776539110553,0.7683212390179943,
 0.7914883031999987,0.6353625682633064,0.8467778579496106,0.5676960935765704,
 0.12141113336617015,0.23201218042859695,0.4442383675091168,
 0.1563515101609172,0.8398273022205341,0.8741467800663975,0.0775373857455165,
 0.45820225147761473,0.4982188812917766,0.2028811427457048,0.9565720232676457,
 0.32180515194617687,0.998751583063681,0.1260503303264885,0.8711168670016126,
 0.37484013283475615,0.3559214735648546,0.637841921049644,0.3315494298174645,
 0.8145511018615139,0.3242991100148038,0.8108511670988231,0.4354008344923137,
 0.16668643222041568,0.81318380193039,0.31182314205615436,0.8936784221900979,
 0.6649742648079578,0.18086856091227,0.6496128519652846,0.4860026857898142,
 0.5517222700075978,0.7672995628834856,0.4434385413452556,0.6168673906573005,
 0.835298176367027,0.9724599839418682,0.09301611472659399,0.13602073928609282,
 0.43777775604886005,0.9044724315327365,0.6677995668587015,0.9254140953992934,
 0.7183974396780153,0.12147209544021842,0.39064711184914136,0.762534072175286,
 0.7032630788177984,0.9377303978349079,0.6350170465370439,0.07564500620604597,
 0.49622683007050306,0.7815226100839491,0.8858019566233686,0.5724863405960513,
 0.5579103827961821,0.33493564649491836,0.7952750975367042,0.5289238549413596,
 0.22628915405798744,0.6677175746545028,0.014198107421851236,
 0.6564610864099532,0.14340711789230298,0.8750029551147084,
 0.42734674610000667,0.5216773929823042,0.5211874570168585,0.693824859614144,
 0.7110572076132813,0.01818737231681755,0.07286730931865415,
 0.5815700804184616,0.6573626819915728,0.6384391323966585,0.5812682911070943,
 0.8041192503542286,0.36729829357666155,0.8805349644000027,0.3987168931998645,
 0.2424010492466071,0.15522114299916423,0.549138459432389,0.19386442096680867,
 0.6089788953244648,0.25981879922837825,0.14543657317247083,
 0.47423340189655205,0.7345469606376464,0.3660411010065977,0.9807371554131441,
 0.37658571518244655,0.07429428635727608,0.8874809270798376,
 0.2645823016484381,0.6115038201254722,0.9874909146533973,0.3841782421947998,
 0.5594689463989051,0.8971992035917429,0.500482542652379,0.10138657810439722,
 0.4784504528917546,0.8500975196207677,0.016623912670068464,
 0.9878483646939866,0.8996783824623398,0.22577394285299257,0.3158606265090861,
 0.8918202155639826,0.5321256790813009,0.966318145019532,0.8609328750477486,
 0.7020824581351424,0.384454533511172,0.7742576287220649,0.1531596548876295,
 0.9817951814128019,0.3002489106138404,0.5070493299709052,0.5947434638363881,
 0.07568195861987403,0.20273035756781965,0.6428892406325861,
 0.8278395248331364,0.5001556634230805,0.6882702056974108,0.31483497075837996,
 0.34435716634925007,0.6131130978726073,0.519379231936519,0.28898513931941794,
 0.4371572805583883,0.44644591028496705,0.06283837687437854,
 0.6813470372530314,0.4184995864862364,0.8926369496280185,0.22790609792670902,
 0.1536802345697167,0.718506318661589,0.10796480411307907,0.2705560644022895,
 0.9210113633756737,0.6688800507534658,0.7339053902259443,0.26328922018730616,
 0.006464844397815583,0.8756627247033506,0.10807390153576957,
 0.340560734019449,0.14159563562524546,0.8579237701632485,0.8342096568555044,
 0.9953680116109472,0.9565127193285943,0.5597904112076503,0.37223197900956384,
 0.20669125586097925,0.5573396565820073,0.5324904142495777,0.3246053058196394,
 0.21670426914080554,0.8167892087483097,0.1031026062353152,
 0.41798145521556274,0.5397792243250386,0.6168841407982731,0.4417827957757823,
 0.131546638652697,0.0808617162388211,0.3746735977392698,0.18579604977415143,
 0.9036376109708444,0.8470881446402151,0.7068275557673904,0.5849446880896947,
 0.07473584029827351,0.08269425843503453,0.4643446105901057,
 0.9773044587324323,0.10497914923249807,0.1276391328369607,
 0.38556873607883213,0.3110409831447697,0.36426269671563816,
 0.07719586988327354,0.6943345540917569,0.3951540125919524,0.8616407409185731,
 0.5312086249561597,0.7113586872228728,0.616217367241469,0.9386760642894123,
 0.7322392274946957,0.2093009991598601,0.7570721456974898,0.8086263812428331,
 0.6731896637296,0.888936426027057,0.9082702656216975,0.4124501653380104,
 0.49153888711506677,0.37923776566521195,0.03191242526502458,
 0.15815469727668963,0.9797020442528609,0.1574088062604453,
 0.051288596301075295,0.7713023522653938,0.9753851252821177,
 0.7988818646917847,0.5357518507037188,0.2861127102913995,0.7642577805176105,
 0.389642741577671,0.8267101916583817,0.609853054184502,0.23458544995389563,
 0.21698966064424674,0.6119463382449369,0.022766777301076457,
 0.9707691630344379,0.3307064347032119,0.4100169158371212,0.29425000659514755,
 0.9696639407014106,0.7876602201261672,0.14975145307671012,0.5643670829235637,
 0.7095682517855213,0.9289520028269587,0.3921195742293859,0.5170309931336693,
 0.6907765806130857,0.10734608311753469,0.09816349370796695,
 0.37407146676838865,0.0216208846256829,0.4484418266171294,0.3661503783239618,
 0.036682310195759626,0.9839550189098993,0.45399143501007144,
 0.9707652141558823,0.3436348083664368,0.837670248661581,0.898670270105564,
 0.8695663181778958,0.9593803730888781,0.370728733277912,0.29070992222078673,
 0.4543427969610052,0.14969270035130589,0.6667891986043142,0.2941152111658787,
 0.4837013526729361,0.620558748214473,0.9962212938761044,0.28048546110935924,
 0.630283348640525,0.9564118176146755,0.6707810956862268,0.38899322125556113,
 0.4551195047541784,0.8188244136404491,0.5737003633906328,0.41944345017363815,
 0.679653815927977,0.4659767626957956,0.590043216470165,0.36192063028136334,
 0.23859408089285628,0.8650076658474252,0.10069474001618062,
 0.4145633829331983,0.5221861386471893,0.39399356591412726,0.5135861206451011,
 0.29710085217375637,0.8289856066842987,0.4358817640164162,
 0.07691867393973861,0.838578096331804,0.5307213358061563,0.2659330335432615,
 0.46114674340302575,0.9344989908616866,0.6202757398412466,
 0.04068642679921364,0.07863121222875247,0.21768014389392354,
 0.2937934027588638,0.19921144028506654,0.731362908796071,0.9271710866598258,
 8.569682753303276e-4,0.5236837555922176,0.8703399028996947,
 0.6210390922769405,0.5696960539185086,0.3763932196408375,0.3622425449164397,
 0.47315926318356905,0.6288034380769004,0.6923783122994976,
 0.36839757869596723,0.6603762780760266,0.2058726590725921,
 0.35336260035089073,0.43788232961592155,0.9609710468990713,
 0.08586495933912919,0.3304580360417565,0.3525393690259252,0.2729859880106106,
 0.8287241792432389,0.9532861479356908,0.4158804148950813,0.3739930773716008,
 0.13666752714258812,0.8607435591803161,0.09510362835188202,
 0.47749614745164504,0.6500118648986297,0.9006879697703085,0.3464244349617478,
 0.708075127758599,0.13544921237643393,0.19441567207514787,0.7803073726171872,
 0.7568940683846138,0.5207119865794027,0.31243703700133263,
 0.13730488200774738,0.2824255284909286,0.8869368488598487,0.1191578038677642,
 0.5949410229435784,0.7378219235285046,0.7273344378587245,0.8453263296986722,
 0.7595214709360912,0.7563105363764943,0.9834799844954347,0.5457927345998359,
 0.5576152487115227,0.3209960741441614,0.037255573319413426,
 0.8469120508897886,0.8516259897441262,0.864728110824587,0.5945768336569022,
 0.6706187450491264,0.3963118387678679,0.23432889065553297,0.8231984895099799,
 0.14550378442813716,0.5711350546866617,0.0547298857016586,0.3299110981227863,
 0.13209248087757441,0.2592935925818005,0.23527704379975323,
 0.4485217414989845,0.6573298740656512,0.32589292561665684,0.157857582370597,
 0.9159210350101518,0.6227843789346604,0.5178521292434357,0.39754781088520774,
 0.01899797293230343,0.14107036589686373,0.3777222894193466,
 0.9426234584162143,0.9310167750846681,0.8466382502356984,0.9015455687791931,
 0.11757051051667056,0.882121288608336,0.4119935498495857,0.428271935813314,
 0.1906450550384301,0.23014125600251922,0.3660398696487077,0.8385778378902469,
 0.29711657791237633,0.5010004742832901,0.7168087200934288,0.9328342086337134,
 0.49483307111099084,0.42893006351668495,0.48264617553432854,
 0.004368291314340689,0.9979465734523107,0.3568264427065013,
 0.3600710597350276,0.5602541573980852,0.09516768877779369,0.386874759443778,
 0.8078469978913151,0.29683815955170023,0.5466815578316441,0.7899691798709169,
 0.2243590572773475,0.25802109774951487,0.26630637532471246,
 0.8503600860041864,0.7138158106387523,0.955470006409283,0.06100515916100324,
 0.5633839738667207,0.4263468010481589,0.793057477489649,0.6271252086762463,
 0.336661963775684,0.7337433579739545,0.4984594809665418,0.8671776800791517,
 0.03371964982825004,0.7827058710770246,0.25221022213350297,
 0.9195294181408626,0.09386426181416552,0.8418712202314306,0.5067891486937425,
 0.06492848273005691,0.2877690469744695,0.19480396118853005,
 0.6681678818815847,0.24903551360911524,0.9421693072519363,
 0.26598720854541014,0.4099286499751448,0.15126133948011244,
 0.28782800503410544,0.6110920278164317,0.7939700970655554,0.8940100202829181,
 0.43319171111256705,0.9699305501971448,0.04524076043365943,
 0.23529271676118424,0.9430286215821031,0.2648640199860305,
 0.28970524305266376,0.6382020142769309,0.4639691294269821,
 0.32779289434675296,0.8228842314650077,0.7080201700239603,
 0.36162123868377605,0.3756137429129266,0.1477775740864532,0.4915522884521125,
 0.878141457495865,0.046572890094551456,0.11788449491853703,
 0.9793199580296313,0.31892690502054577,0.6891836138273976,0.9631017179663589,
 0.44684640981437107,0.47493187786400737,0.9487535038007557,
 0.41208110140567444,0.29795902498449034,0.8479931875441148,
 0.2624828691531317,0.3858362988651136,0.8219486549346273,0.6087854608607867,
 0.22306051150746198,0.2024823874489977,0.04714544598991155,0.511753699046241,
 0.5439461861886524,0.25244129617930067,0.5815434483333913,0.8793996003474344,
 0.5380951895886867,0.3438651955378156,0.4982506908252684,0.6688754036734844,
 0.26559653293705665,0.5419411970826902,0.758841783785958,0.1964763857816747,
 0.49409346643783214,0.49182806525179257,0.2959169154795286,
 0.9495217137024454,0.3702757868946239,0.04677426598894652,0.7595966398297621,
 0.07681653994401616,0.9736425993020816,0.6173178285550273,0.5931255081931787,
 0.810891407927369,0.21850501935593247,0.7438194765614379,0.40955644031205285,
 0.3777738242465021,0.4119445988941387,0.8517618519958983,0.7076987338688622,
 0.825976284989679,0.4943777272586317,0.008647721403054254,
 0.13234540611905898,0.17462304649520166,0.03967751114343632,
 0.554827655389454,0.8970349254941462,0.24515097336105018,0.10343868368984355,
 0.9473111407866549,0.03670420203660829,0.29979097097042884,
 0.12890753909165442,0.7948483959671498,0.6936953079773956,0.6400350227750398,
 0.8404899222316563,0.5173668544053298,0.44665784534316577,0.6306886647984564,
 0.6493683343826062,0.08679824702134997,0.3097545947431196,0.4535609927091486,
 0.7479711180977358,0.9809625545405546,0.5173272056462692,0.7825378142639487,
 0.3529167375862575,0.09099451122188196,0.7863429464877704,0.8698857082825725,
 0.036102971829547914,0.5285984114176796,0.03202498369858453,
 0.5234327093832611,0.06213684031640576,0.5665758211910891,0.8101489277346091,
 0.1651067555893876,0.17189089653363465,0.615423120286089,0.9139266928886449,
 0.045688793860195975,0.7413749061375624,0.2970405500854273,
 0.08993110326128995,0.7469806002811039,0.22322546026123924,
 0.9348118446757898,0.11710366303683295,0.47899405890126734,
 0.2891194621640235,0.0739501881229786,0.40638439810480775,0.3048250349522307,
 0.3214199211965729,0.40687852924411516,0.12119491273401928,
 0.3278024129334707,0.36211509544346654,0.7364366971875669,
 0.23449414221049503,0.9406586491123506,0.16567638480004976,
 0.40255428573271335,0.6707509921308578,0.80619021238487,0.32633820762971766,
 0.11721174636346443,0.5055312446633422,0.38043784382780477,
 0.46621776030066975,0.0469686551164743,0.0499672965781337,0.6675562511974494,
 0.29470529013798885,0.4699251566333662,0.3696058942262017,0.8255335310352659,
 0.1200970859420668,0.45291119704873606,0.9717281147111847,0.6586347907851117,
 0.30877121836287524,0.40175690170819667,0.4287657112532707,
 0.3697500835855766,0.2639066124540673,0.5207058682093453,0.30500331431634886,
 0.8420126302677174,0.82602149909036,0.8759322285217896,0.38296279201285766,
 0.399660137789956,0.8241375500865277,0.9500224037595472,0.8663190929090907,
 0.37799622770119123,0.241234726722377,0.623666226165548,0.4599549686444362,
 0.6143080124981394,0.11657704860559281,0.8662430802769246,0.7535374652712885,
 0.9012815519695325,0.2536718529324171,0.5766630543892162,0.6119376499008815,
 0.46795380618076954,0.49207872462521074,0.7349300813850603,
 0.5829087060310288,0.9888893207150246,0.31115370709885204,0.8357508283337881,
 0.8890177918065745,0.7807602510437681,0.49451150388792475,
 0.03285172512964285,0.07732580658295873,0.3561645577363093,
 0.5645294174207026,0.5567733555783679,0.655524015222738,0.713672608433047,
 0.37867273605912466,0.06974721615362667,0.7259985718200052,
 0.13582184381186502,0.47228498322521173,0.6425376668506388,
 0.8895939762291596,0.2820179780865524,0.6141939429493363,0.8776586531414544,
 0.13676038485707043,0.7404383095490285,0.9890480563082932,
 0.005407137842949772,0.40233995246382936,0.800542303316624,
 0.5983842511116564,0.310959663355343,0.3045467531542103,0.39093376019334214,
 0.16341693959606085,0.2186386731590223,0.9171443123179225,
 0.20737657522810538,0.3638780582070631,0.9007211722139209,0.1669469976741631,
 0.24718730582129989,0.11179843134769984,0.6598528525772847,
 0.8742863986726345,0.6301679794132312,0.5523682223624292,0.10732622980204087,
 0.19344676252775106,0.485373825147702,0.13556017927174313,0.4862421469646505,
 0.5202150510936667,0.6807672644565679,0.24326253324640612,
 0.26339781706085175,0.52043009858211,0.2745288530815476,0.7315916614778281,
 0.5169240394042055,0.5715323947601948,0.05072203350029714,0.8030955914615155,
 0.7741753804054848,0.5154354698332129,0.3620950113419722,0.5368115933794485,
 0.9338933925768803,0.13253365948671192,0.9924442296997422,
 0.39793992028367986,0.7527428012957339,0.5559211852439814,
 0.20226578842311715,0.15370390364805142,0.33112251169567375,
 0.7590233541787056,0.8283057966440331,0.6985669095130247,0.6368425876243529,
 0.14216879218366074,0.2480187893699688,0.5886694928577454,0.9889814363161895,
 0.007668347995823277,0.7442425300819435,0.3369170959989146,
 0.8143757986992456,0.3141978621681598,0.14223897512896844,0.9617686308170158,
 0.20372545078309878,0.33929811080760774,0.18793633025126177,
 0.9282932668032119,0.2466730518357616,0.30399950808816734,0.8955494201469486,
 0.05812682994019591,0.01429654605192765,0.35435920826244394,0.30164752495614,
 0.2721990720661047,0.2150362808512961,0.7226383703840988,0.660442288151352,
 0.03835311144781217,0.403260404101598,0.5621226501074124,0.7788896750938689,
 0.16519016370613027,0.4129630935423081,0.9890723720723222,0.6150657331118263,
 0.6967739215838186,0.8235234010028384,0.5980197525403841,0.21778522640284037,
 0.43379151852877695,0.20268961616682202,0.71068349412698,0.643740377473279,
 0.6544776161943661,0.19428495063430362,0.5350416636699575,0.9975571285086854,
 0.7833652230392837,0.10825392155160651,0.09284700916379285,
 0.8635338684774237,0.4830778094314496,0.5584592880100793,0.6974816148662455,
 0.507648755455323,0.9175721064839619,0.9925656321178489,0.9121543351060734,
 0.5705208389507405,0.9333802493861963,0.5410837556201198,0.8794667876275724,
 0.54985441257882,0.6285422254326166,0.03218528364749884,0.02355518776517751,
 0.07095523696452688,0.5479477783716872,0.12267629360678167,0.994444912343877,
 0.2406312635641562,0.9236336919021308,0.9986552238818531,0.6081342452954129,
 0.24654285747633353,0.3592562683853804,0.10104997142503447,
 0.28617509932692886,0.8756080138462892,0.13301822761580995,
 0.17444293332753913,0.8736504316707413,0.7207983527044828,0.9209921357660503,
 0.5733575573024775,0.3075803963286313,0.2781912914019572,0.8109097160611619,
 0.012533319543850396,0.7912777557614442,0.5372892086682302,
 0.5549932167434412,0.7002952468892978,0.24252488867496813,0.7026821484833272,
 0.04206071907349862,0.7468245843043633,0.9108758329254767,
 0.21570581195522764,0.5430791260336569,0.2758717842936581,0.9785227314498505,
 0.6599560954803584,0.23099934930435229,0.7632808573077402,0.9434150524914523,
 0.554693737502654,0.681395215087653,0.5056751673270262,0.29751009754013324,
 0.38170614743849274,0.9094124424767691,0.4594133379942431,0.8464047717309067,
 0.02466034415094609,0.5799280445756435,0.06608487863937951,
 0.8667827469106064,0.1940151172522261,0.5252596756212773,0.10784984315154278,
 0.7331145950384949,0.3582831762847062,0.8651360813768086,0.22185276511308893,
 0.3480900039325674,0.15125996967166766,0.4566502028689422,0.730138632510937,
 0.8711514649732931,0.9673924121618034,0.6775014555110599,0.7226645759383848,
 0.6479002663268333,0.026493134956709685,0.04827544186263488,
 0.8522007135453572,0.2283509598261363,0.8107205507160532,0.9339810843306008,
 0.78273259439851,0.48188209622234257,0.9569842846131774,0.7141920504343743,
 0.9199570475163186,0.2062864653219738,0.9342279282830899,0.21988846101410486,
 0.15989118457315243,0.9306880588450915,0.6354096805480882,0.8180495365063513,
 0.5081461631919149,0.19695255834787506,0.07026993035389095,
 0.07501469525441196,0.0990011660798451,0.5186799504192067,0.5449792742173305,
 0.7888659307468697,0.006373789904996374,0.2643015409410626,
 0.4292466865646575,0.35452068568851924,0.3595991182378515,0.6688618921752574,
 0.6626878472789262,0.4590907172585672,0.05839944724259338,0.3962348076595028,
 0.7042132976639547,0.19324179326349922,0.27411003629208175,
 0.18007390768623377,0.8000420173349587,0.053683251063040616,
 0.24378405868408548,0.8543027300225517,0.6512377399411244,
 0.08096500892483094,0.5144634753075702,0.6683493562048368,0.8818021377792296,
 0.2257896463562432,0.41111385800973,0.7373538874370715,0.47448589361169624,
 0.6996279170562257,0.2048552596704657,0.3070244208933105,0.6438288197220655,
 0.11500690196960162,0.23990637452168306,0.17500692026101072,
 0.9893202180371063,0.66713970243578,0.7147747560495971,0.6736321106177483,
 0.42390550236117,0.9909061466848766,0.23715853557984856,0.3427329528628862,
 0.8339898794160284,0.12943779985782689,0.5593655952790597,
 0.06268313288964977,0.09620459361409439,0.6517756748715363,
 0.5918340775524076,0.41090025419665244,0.1514259811756199,0.5733233195880508,
 0.46842187147003345,0.6173588356125664,0.5437148261615931,
 0.26646736567851803,0.1265940409052454,0.16852136787701055,
 0.38885169727290236,0.3063347709304989,0.6003722569320139,0.6413807160440964,
 0.22125742041099294,0.6808299949848561,0.49993296723350034,
 0.39535739480675947,0.6418901383224395,0.4646158619897405,0.501950436921623,
 0.7018742327592324,0.6501769685678814,0.8792113099615683,0.5176755764893756,
 0.2919506757171836,0.20504144940330327,0.23845151776075468,
 0.9215659026037768,0.4913674798571003,0.06823187580606382,0.472642834450987,
 0.7829169990580712,0.2958533610751721,0.2809408018882804,0.8179696057936248,
 0.8908868235462057,0.955629482377952,0.48742438952364586,0.23190441459231337,
 0.7189189342092502,0.4106762385477636,0.36117762455587477,0.7787163603493705,
 0.5852001756586566,0.5706805023692966,0.44341855279530884,0.3408620003534303,
 0.017626292699501,0.9460226370609365,0.9997777321775467,0.6509179037012177,
 0.609082512523992,0.7756922076031472,0.802452739573452,0.8485574669541835,
 0.5254170644459598,0.9457644182743355,0.5991911315586106,0.3425169797317916,
 0.4914434621825028,0.572193216187228,0.05772902042915451,0.08578328236864852,
 0.3747070033309723,0.003299070361159173,0.3672945912417118,
 0.8934398297783311,0.34845151407158004,0.48662098623682426,
 0.8186939787812492,0.16060978544639248,0.3383330787555897,
 0.22679480712458466,0.996572241316304,0.5612718710379908,0.09657797798354584,
 0.06597650278235756,0.5342923567297451,0.6962319927126341,0.4971401108964991,
 0.9561564834844861,0.9721197269319493,0.9834953270479421,0.11581192119608352,
 0.5770273297583439,0.4712397431548747,0.052380560616336336,
 0.42425102091716693,0.8880127255864485,0.42051923693954973,
 0.7595946111174656,0.823401544632449,0.14640724749580114,0.8973948774611132,
 0.9639095011170272,0.8413468199808642,0.8134065165617044,0.5428995304374145,
 0.8984239783945611,0.1728550988993467,0.11480712340288957,0.8535508650808256,
 0.17971300094287512,0.09335626499785576,0.3310080935947032,
 0.6168217600149571,0.10992353299376478,0.7307805524700105,0.4488741401599534,
 0.7314889572026377,0.46026726257743,0.23793303451918713,0.8460567442185337,
 0.13419417567331293,0.6864195446436345,0.6474828628065539,
 0.22958610286285264,0.01612005844762532,0.5682556076388143,
 0.15123574748514568,0.26090705070164855,0.3345404761663656,
 0.6637718818381222,0.985214694936942,0.44197041525434155,0.1289053384607104,
 0.7468342917567661,0.42062114730677047,0.29066541240001253,
 0.8493972946243058,0.2849650230703109,0.9042886009746638,0.48677221615965227,
 0.334714184202461,0.44011476434828456,0.6224745503520974,0.3592287151340068,
 0.3333735467667013,0.7504676582894336,0.07109665067887716,0.835189736251466,
 0.45093812741763484,0.8280080384683636,0.3661911244257604,0.8306030268973902,
 0.0969460374289206,0.8366245075853972,0.4105818910407868,0.959967680328866,
 0.05760011285537947,0.42623278578103385,0.9236018505283775,
 0.7451867627672644,0.46116555835454665,0.9570067291536567,0.7531152806154425,
 0.37665878148849063,0.3829953821068385,0.5647499198674216,0.2775913298740281,
 0.7931609413268808,0.935697495069466,0.4399741314274265,0.9288029426855365,
 0.39477293974101446,0.45858211445028707,0.7271298525823863,
 0.9199693920519978,0.13025150313011058,0.6401590649688811,0.2634885936865019,
 0.39168426029120695,0.7402748732066301,0.18763228255076958,
 0.2515787723836739,0.261454791933889,0.620621778141805,0.5979533983153393,
 0.7475449047682958,0.5755256275712846,0.6602409942748131,0.8142988514225361,
 0.4891234312697955,0.09229053702279766,0.9557640563815941,0.597849692589358,
 0.33158508547188614,0.37875297254014173,0.8509557728410387,
 0.9685571428203191,0.41932351384661926,0.6816572254442173,0.3546978993795654,
 0.2625894819661686,0.6813521124735001,0.2921073977531835,0.7659536192084703,
 0.28928033034624534,0.03910528630925503,0.8837548770077025,
 0.20587590542746126,0.8725145602058252,0.45521525734452095,
 0.5772424797094577,0.9478502389284698,0.3684097049634215,0.09879051529086902,
 0.6194471408341605,0.22851539072414662,0.08661444949398922,
 0.28139008451994696,0.8989196168278113,0.8201770800886362,0.9129471151696102,
 0.47776693512860213,0.6420624630355343,0.372004118400759,0.15670102004355724,
 0.9579660110232093,0.11494536139837863,0.035768882876592634,
 0.14030111497949338,0.21500027550807355,0.002016638616535893,
 0.8342300822639648,0.6575316535431609,0.33681880000983433,0.6532286545039429,
 0.22502964155199556,0.1300220072073366,0.4359352240074925,0.6093552688030793,
 0.2583165526644907,0.8030343811257823,0.5504157494091981,0.4374352601271141,
 0.9408242945572907,0.8796413342445044,0.4921680046191357,0.8739781505800205,
 0.12516249995530138,0.8528193748040891,0.3801404520143429,
 0.47024539017658584,0.7192312324760026,0.7375732853268137,0.5693253937938734,
 0.4803187013377326,0.8641077409197285,0.41217758091263357,
 0.23861283044406834,0.49248105387653907,0.8173274110121249,
 0.7069083754237181,0.03097838440085965,0.5216524441443343,0.7380194059557857,
 0.8863292759509142,0.4454145388574934,0.6472537978386383,0.8409227379025963,
 0.41527506191646846,0.21391204430413402,0.26830422917554597,
 0.9065139310526775,0.2469509345438251,0.4630629619630152,0.14109342021657423,
 0.44471167434136316,0.2582922960825915,0.00954229541657381,0.579692162350834,
 0.2962383385916014,0.7937800762681922,0.21989722323984706,
 0.08152067483721492,0.8771411087603465,0.3107258551362079,0.7031363261521275,
 0.6929208755817817,0.9940735864499265,0.4156942510538587,0.1856188368019005,
 0.6839802595214696,0.6342237297168082,0.6027977458817608,0.3762206910746019,
 0.7145700672426049,0.5655552278347268,0.42730641937287284,0.7643862121704441,
 0.11784861277234748,0.834343741156414,0.21761598731073883,0.2984389285917983,
 0.4185081121515992,0.9759735062217885,0.18667584101606738,
 0.006337416727916767,0.19702147111167112,0.014591540318280471,
 0.29494234953271714,0.7201892401859212,0.38260270427413856,
 0.6307382250940979,0.34517330115177436,0.6058085548670885,
 0.17227154466626382,0.001524680622711383,0.30762442011517654,
 0.45213334284162154,0.9923093017395955,0.9019436576746247,0.6165996630179985,
 0.7944486566807731,0.024481481168514785,0.42845158335567635,
 0.5838414532860811,0.3166830913145354,0.7782865733527028,0.5676099731922053,
 0.8417130279845441,0.47686203588482967,0.8865644416560529,
 0.35479473245941895,0.5759323186991689,0.27583351999063,0.11698850241845782,
 0.500124590983049,0.7092171055389208,0.9462415326975077,0.2322014491403901,
 0.8044168142973448,0.20003018049025156,0.11246775450532764,
 0.8453145251455776,0.8986160408433601,0.7416292936645703,0.646707504252914,
 0.1258186037875223,0.5153811581104654,0.2641485095389816,0.610924289176517,
 0.03105433620548992,0.6526267818791496,0.5854526033113951,0.7863422819692226,
 0.4831314914506708,0.12536671777101882,0.9552161627803444,0.9666306891793839,
 0.13408747842901014,0.44494928680317325,0.645586647463619,
 0.47913772745680006,0.48876393415216646,0.6664601234510913,
 0.12918995745410589,0.800014052194177,0.5335105348253236,0.2168396605638394,
 0.1619743546209601,0.7664430704584756,0.1265632288056856,0.711083442340348,
 0.05984399035242938,0.7981916289230515,0.27193735754821446,
 0.5448033028910008,0.592176534475568,0.5827914656521437,0.23453252011632353,
 0.6966891711989843,0.8004613791959384,0.9108901144204916,0.8538629404545219,
 0.05504422549281218,0.7872390127849531,0.6187752123639244,0.6537151820196694,
 0.6717000255046156,0.4844155911472625,0.7104847585904293,0.09326838285663497,
 0.7140200447095827,0.8548263472902357,0.11565781806998521,0.9856132993673418,
 0.1588244143457288,0.8720776098333772,0.10021109879980505,
 0.15266647988926985,0.18556828137017722,0.6108698120129094,
 0.9177614114594114,0.7772383133608726,0.40284649148571805,0.1768597064504731,
 0.9384196133916104,0.5741475154473118,0.8460880556220117,0.6160734880612165,
 0.7198390241526316,0.47677573781102967,0.7591286696990989,0.311587490459701,
 0.1595829162462652,0.19178402625499835,0.8918783603329846,0.7284511518438719,
 0.39167032133615964,0.8784458598659152,0.1137566002686532,0.9418579534080866,
 0.8327679034237282,0.6978282417510394,0.2705285977052033,0.9466615051728483,
 0.09788876207705344,0.6044253519871609,0.6118413489850026,
 0.017745697064588506,0.16397783150402134,0.9247163596462047,
 0.8862905033579777,0.9919797164889496,0.22814435320572657,0.9576556180393251,
 0.8099191151279489,0.1899076157648374,0.5041423236281063,0.6195515881260135,
 0.3896850878018001,0.00658308148205311,0.9103605840248501,0.8213117322245512,
 0.33446502673450307,0.869001166885719,0.8109021367509772,0.27368437163328596,
 0.8482875363569993,0.46900093715671376,0.4810388174264646,
 0.13051633354456593,0.6127288194902639,0.1886642333949311,0.5870440391780202,
 0.7450494372774195,0.5927042673748094,0.36064436538599853,0.6151871878142482,
 0.6207240551481629,0.8910570657758418,0.5937431410544203,0.16376672973097683,
 0.6587557992519235,0.37879668963258095,0.41109495814134633,
 0.9760691399029986,0.8112362999783382,0.09415429742837422,0.797740687290034,
 0.9807681756484853,0.3873483217129893,0.020417956654196323,
 0.6409481211892841,0.049646813017511726,0.35239945411915075,
 0.14902329921189494,0.810386439376173,0.3796477984585368,0.6306794553395366,
 0.04593516588148805,0.31523259978965834,0.579901508185568,
 0.15729971061127868,0.7752821003878094,0.5469004478386251,
 0.03995769306876451,0.2032367986097714,0.6136664321210961,0.5481462901701599,
 0.7474112329536036,0.6643855765230658,0.3594612017824772,0.48952417027917394,
 0.14590473978368756,0.6835105017711846,0.18208810903509598,
 0.5154653357345622,0.35248786717210234,0.32420486763979994,
 0.6645529984988016,0.41949000987800555,0.7079091115144627,
 0.057468126093876526,0.9972232937790455,0.8963485936685102,0.416160341898692,
 0.6499321061998588,0.8841687659217077,0.04871846630777665,0.4421229383047325,
 0.7072975837237314,0.8832560422496448,0.12098920177192762,0.6963100646435287,
 0.907173038867998,0.2905451387757675,0.45743886929233357,0.6293026162108823,
 0.35663137849843163,0.7080751438375672,0.847651727112013,0.5151302063159915,
 0.9243876074042218,0.688870115600842,0.24865029612272138,0.5389626563248697,
 0.41503327806476253,0.07789728131328799,0.14264536788865023,
 0.8342987933457263,0.6067398075358904,0.9915991043595399,0.18980689597401423,
 0.7297444512124248,0.68370377006859,0.1352623191454949,0.18148082751217454,
 0.06311418000214852,0.5233973599052268,0.9493321234062064,
 0.08708151125474739,0.8528934249237511,0.7167228949916973,
 0.44993856611433936,0.9132527870267821,0.7999469406169065,
 0.18294372353656496,0.6332638139197947,0.07323255014083419,
 0.8931374545382137,0.4652779074056207,0.9772510192360383,0.48363938578996324,
 0.03729348493660656,0.13706145220591537,0.755490039201441,0.8839374921671315,
 0.3363344518779978,0.9190191504649503,0.4883208846694179,0.5762378693741066,
 0.003557152966328303,0.07293587508255239,0.2597521193431832,
 0.2867700382643332,0.30289691558628895,0.9535222544960028,0.5058332931247003,
 0.9200418293718183,0.00936188341685773,0.349979628408412,0.8189353560005226,
 0.5683218241744564,0.275508658065878,0.4209177727443395,0.6504131979935134,
 0.9166601074203234,0.2818920325260156,0.7317616021685015,0.3185604843352413,
 0.9756490180833536,0.47449397652123393,0.6408657323540238,0.7165624107407056,
 0.19372837598093218,0.15067141511053672,0.9851956690864668,
 0.8621223690090872,0.9914232170050813,0.4876483376692682,0.995042910386133,
 0.05614411945524722,0.7072412254172467,0.8603854786383431,0.9638693958100844,
 0.23456232798103938,0.6255539895849165,0.9059480692418203,
 0.00941674579398466,0.37735549289853476,0.7342910963510476,
 0.24571555738496065,0.27242145732492495,0.8270233983845746,
 0.020637560255625576,0.17989388339937817,0.5815428598599346,
 0.9892427431578581,0.5568230045864512,0.0697144062322429,0.7214833910924164,
 0.5845135056330459,0.7296954204774498,0.569604241239484,0.7921387545804514,
 0.538160517475548,0.9264024381062518,0.9351311911234571,0.9247298255715497,
 0.9665670628544194,0.2466135295933447,0.026164057915481154,0.489681317436675,
 0.8444053613062237,0.02633064440413424,0.08196911207331226,
 0.4425325796668995,0.594964814827896,0.17714754316712822,0.6679123125715972,
 0.9035918450163937,0.9114177880971231,0.02785386072941609,
 0.05142973041304644,0.05022346325495153,0.20311925820902732,
 0.19223238501931073,0.540191697301701,0.6043135969090405,0.1344081961167506,
 0.0885835282171985,0.10791495630855719,0.974731166043809,0.3535305466606259,
 0.8161747527880021,0.051289241391138884,0.2513759583172299,
 0.6913398836035847,0.7261980035184921,0.26907356369174973,0.1998047597215329,
 0.5381149905853556,0.4836547774546106,0.289860622506366,0.24637287961069443,
 0.7257462460515024,0.05850862444215443,0.7295792106939594,0.4829886525221537,
 0.21437114480946817,0.8177685880718151,0.5540861545616854,0.2990037218363438,
 0.27817191618528303,0.07080977754487439,0.7159230743765717,
 0.8179638150001848,0.5822838358697952,0.9656907781266466,0.7662283166821994,
 0.46775150201291116,0.7690324649268345,0.08076715890374586,
 0.6134025394237516,0.17175013247819415,0.9749913759027443,
 0.14259980197485977,0.7464111095845227,0.3877929100200873,0.880939393277632,
 0.11103483372720857,0.9740202246858517,0.558260829713348,0.22879348100910535,
 0.9095223099017231,0.19664487644303708,0.4034324162033981,0.828618928563627,
 0.2689814450527699,0.9660947626912737,0.8259927466972707,0.19392876331632913,
 0.6601017520499584,0.2852993607477574,0.787003978015015,0.5566812998618841,
 0.6340428580699917,0.8709557501183683,0.5661404135194588,0.9782826760438561,
 0.9644477889701336,0.8084511983886709,0.3352981285182485,0.5884795998894552,
 0.9859796362297973,0.8036570379897228,0.5806126925291245,0.8980629648542123,
 0.08193972262830373,0.6435244756586042,0.338501009116953,0.950209487901794,
 0.02654763907816937,0.43461568360788116,0.34535942126333596,
 0.15813920650784352,0.21170734673694291,0.38282445198212556,
 0.4303438605998531,0.07602391859356539,0.5074318571663465,
 0.15957281470230322,0.578304857568989,0.13693107474853505,0.8181948961111264,
 0.09276396209890625,0.14425246976185047,0.7006089320742479,
 0.9730723365031577,0.24053293170090817,0.45932975224158035,
 0.19129971065160822,0.3097823449562277,0.0575469561813865,0.7858264977628736,
 0.8840050428994578,0.6976365039687283,0.11652406998131726,0.7554519602217392,
 0.20694785706355845,0.5466948444132455,0.6195304965631536,0.3784555680829793,
 0.21666239919038788,0.44028328657122195,0.8001904581429727,
 0.13225131657420686,0.28542805269672045,0.5128118309226462,
 0.0332021261579154,0.940066121460247,0.6808730095637783,0.2363787461325465,
 0.8819106926007105,0.20458398401994893,0.9345220686397089,0.9075306863126095,
 0.12152264149840541,0.869177621174526,0.9358707781963524,0.7267420253256476,
 0.264408639264862,0.8993453060248737,0.09781194846895913,0.7579649575771361,
 0.9470375422554491,0.6699160917085024,0.036112314085442576,
 0.15602858916534224,0.05556112337544805,0.27300106810924274,
 0.3872517216910256,0.8213752513856041,0.08057108387375234,0.7012981243169932,
 0.298396654086583,0.7528042216513406,0.609020343616454,0.23190198311394516,
 0.5112308078816531,0.4406565540333681,0.41996870959071053,0.455559374348557,
 0.42766743736368484,0.5156042610952594,0.04620459885144401,
 0.9036798254112958,0.24582489465810364,0.5255145012291323,
 0.36707691029390577,0.5990365555403652,0.26925414768213585,
 0.6491641264799761,0.6885305651744148,0.7388961549089915,0.5616542858502319,
 0.5195198979213136,0.32214454416336247,0.22233538488733073,
 0.2884366742222121,0.8308989101355658,0.3409679006146349,0.5313991633700973,
 0.2507442438560883,0.612662520363236,0.5576024626074105,0.023497310131975757,
 0.368498208762815,0.6269566265267185,0.6799006265129547,0.5660373021758496,
 0.9415252505973006,0.5802751469481349,0.9870482555197525,0.7070403205088946,
 0.05438228078731644,0.19361439526677127,0.30582891954224833,
 0.38366009113097965,0.7105983520009815,0.5764672161362833,0.3976970143202071,
 0.8077008107039813,0.08791399207428519,0.5228298030535805,
 0.04136425425259205,0.35945032807587396,0.34542441215279573,
 0.6690780968997946,0.39752067037056404,0.5700352548432063,0.4717024575431735,
 0.8186484975674716,0.6229256238322427,0.722081648758465,0.4143725922886191,
 0.5377355904045346,0.6948371913997551,0.47264669854889674,0.2977324978714432,
 0.6459637736592145,0.9337497815060374,0.3973195677195207,0.5966098321778786,
 0.2944361382346439,0.9435081999276905,0.02782520566868829,0.6134601245543765,
 0.6983436396572127,0.7030398314310532,0.6374128867512129,0.5622790282604181,
 0.6089161262752448,0.1645503401166305,0.49941924428701023,0.7744231028226585,
 0.02761004980123627,0.7970441580607605,0.055240833526733146,
 0.10056319445318951,0.8490433207618348,0.9024266858873938,0.8941095409413019,
 0.581087802005232,0.8268047857592237,0.36876162081927844,0.7240824469831164,
 0.054528210669266275,0.35767818663556306,0.5253212974960622,
 0.6135646732452198,0.885891455600061,0.15358902253065043,0.6423228788993687,
 0.5660324084268252,0.5616442925715321,0.45205067013214384,0.8804699541730319,
 0.3814502965755392,0.6206731136517061,0.19094495109686782,0.4219887135472149,
 0.027707625922444246,0.09735913703310306,0.43030417141167354,
 0.5847623780793805,0.3774298016819424,0.8236198050739946,0.6566142563305508,
 0.5401890164529749,0.6268561323375618,0.4357769613043203,0.9577299944631129,
 0.6155593975083034,0.983243771928711,0.7020656153155594,0.7036878028330292,
 0.2456147207906818,0.9333632298221491,0.5412877641560886,0.21431708201181132,
 0.9756203953726553,0.2600014002493417,0.7075596655888694,0.6764095701886286,
 0.029075595197262327,0.09575781595788169,0.6395042048282527,
 0.2985811958087162,0.005590838919898533,0.7676266682326529,
 0.6312269814455622,0.7877930519382141,0.42725982702687915,0.5318851093462089,
 0.6869653951495449,0.6913168376299093,0.02517818951649864,0.905790275225919,
 0.4156387149343159,0.022190801164122753,0.7992261659775965,
 0.5937747016619097,0.37183874484410395,0.9116219843500116,0.9688686153302495,
 0.43211232592043713,0.09513607108033162,0.7784865587576995,
 0.27793278814793654,0.6743368327695032,0.34041746141373996,
 0.9538451954136749,0.4723972493348746,0.4642004889193717,0.6704225978471727,
 0.681555982009344,0.8213957092445551,0.04872148633208362,0.6427446304339249,
 0.7901307920740243,0.22625921811450478,0.7124106452448512,0.9761285761657712,
 0.10950500545116448,0.42934869758156013,0.849397767587477,
 0.07427770383822785,0.8725985885420825,0.10321716390906344,
 0.09230774202126335,0.990605868808728,0.5998375595007895,0.42896971814243423,
 0.8494704826901625,0.953417292508122,0.6087913410426135,0.8833849962056698,
 0.6958263203296063,0.37039212766001683,0.8069968526532791,
 0.031186111246359727,0.6781973454622077,0.22450204511302752,
 0.8061192406141747,0.33328472478278914,0.028368002252542435,
 0.5623092076165717,0.06279577288245675,0.8735595887177566,0.7683922424514115,
 0.44499555917169575,0.48862912596055197,0.37600428512470596,
 0.8504335416740054,0.7094321657077184,0.37884303103788075,0.7585312526404766,
 0.7991860754731994,0.48017029346285267,0.7133683899402204,
 0.009585064480306826,0.539857304589196,0.0779455791132121,0.3847580493474372,
 0.7624705958091652,0.22186604265134502,0.855432664237007,0.6853507714434897,
 0.45755418057236896,0.4165664964327471,0.725096273724713,0.4378652987761933,
 0.6679744149234587,0.3895135824595032,0.9123598352044849,0.3992927134058557,
 0.9024250123712444,0.41602485055572636,0.35052187923236733,
 0.0547410883798225,0.0338271756636388,0.2996127757618584,0.7946559909367797,
 0.7720514673671235,0.9105626773555134,0.590213832916792,0.5984278731201446,
 0.6982087972809927,0.9097539288005352,0.32043104104838604,
 0.07896504075303956,0.2982396777897163,0.24803299280650182,
 0.7369369506847188,0.3726800536246001,0.5838220454005686,0.6555316952956293,
 0.19327106883107692,0.388491751181679,0.5512769411931894,0.620432566125844,
 0.45939796793638055,0.637106578415427,0.41595409662028227,0.7885278066683347,
 0.8435147158444697,0.7554668266729634,0.08643032948771245,0.3489614415674719,
 0.17041861888209298,0.40366197254758573,0.4287703524541707,
 0.6745375558078758,0.05164394404247141,0.6296214454538912,
 0.33222458879628236,0.37701916946571845,0.7285910721956845,
 0.6719690838407524,0.3494379648996052,0.34576758669163565,0.9518802275959471,
 0.16903564778418512,0.29851687071718147,0.3055338903428335,
 0.18435471718669905,0.19566429718823908,0.11385887573903464,
 0.40264572447197067,0.021771720708900655,0.12531418614903844,
 0.8323226469203625,0.7224646240388402,0.3850041171766645,0.3216338474449997,
 0.7817710773076788,0.05960941077991744,0.9833288016571701,
 0.011455009625333079,0.5919534395977372,0.2902183099450908,
 0.7033656962459586,0.27591212171539947,0.5951929027348994,0.5874674876106039,
 0.3966471827740564,0.07301192540756862,0.3065066774544457,0.6357495706902765,
 0.5387607987548957,0.5273712001896214,0.8741606928773132,0.7331679830269527,
 0.9387664078955407,0.8691929075956453,0.3757615313749194,0.9251931351630895,
 0.10878621809838895,0.8479845453008725,0.6165146455422006,0.834001737327929,
 0.5436447485604604,0.6076748742942487,0.8499478553613441,0.06599524362360676,
 0.044154872355715136,0.6620622868265725,0.8388761176661724,
 0.9760326240343772,0.048657021938548706,0.29555883794995963,
 0.8292603042164981,0.7533821060218713,0.5816489158975726,0.9613760883540369,
 0.2336631087737603,0.699117374990887,0.8957675429868879,0.5228319680444868,
 0.7671329953009676,0.28442147311419863,0.6377241449291289,0.9326334055585505,
 0.20801440058320986,0.295943220694629,0.3222597399517436,0.8013353237082351,
 0.5683448530116831,0.3310127894178868,0.4797042763115277,0.6263069162926507,
 0.5771771162333978,0.18354696242594115,0.7718659954791747,0.5094212875221602,
 0.5848320542803453,0.7210242036335501,0.7347194805673647,0.08199659008758609,
 0.24322280851592903,0.1493957218402252,0.9557114196587334,0.6686136034376814,
 0.0657008515596198,0.7677877572511769,0.5281537604775415,0.5463650119324771,
 0.04638071761744622,0.7914061717468878,0.09030563759267363,
 0.9819064621654795,0.5365440161781622,0.4887621388025337,0.9390514097872542,
 0.9476061858093053,0.22157821767284736,9.467061921496489e-4,
 0.06083404509319312,0.5825524299737173,0.3787493353755036,0.7430160446393361,
 0.6668697165874253,0.028643714932298336,0.6422942562681353,
 0.5695907492331105,0.9590818431102899,0.6253376176611397,0.39244548089390074,
 0.0016465032137773505,0.2674628704394477,0.13189174914288426,
 0.028457945759649883,0.9627786394953713,0.36101909858180325,
 0.058584326575691015,0.2523608941478971,0.6159994275044598,
 0.14872198711842355,0.41759312041186347,0.32534213987203686,
 0.43464335233994755,0.7646854821509332,0.017695087393283848,
 0.39838006210449606,0.8742357011690545,0.12033448645098599,
 0.7574977791859032,0.7706715294933737,0.45885687931596053,0.9376986989044439,
 0.9693819508554364,0.9454632383344562,0.8994432243871511,0.48547860119262365,
 0.9720305329400877,0.20706106568148452,0.5744464306152351,
 0.09991663167967613,0.14522239450369356,0.30686367413855486,
 0.927784899348135,0.2332277570861676,0.7764365677613276,0.3354282986828331,
 0.5994115455665049,0.4612321237830814,0.3407730279070048,0.8647881385100484,
 0.17169419659045926,0.32421518584795683,0.12260632126868176,
 0.5681315833085456,0.11717792069581001,0.30064738815781367,
 0.9347393096611472,0.6581683879559134,0.7170343892836177,0.8330739687406472,
 0.007627742380989511,0.8508705915679303,0.029846039206043784,
 0.6649798406901621,0.29902186633567696,0.7801336092165164,0.9594859797616131,
 0.08099168295868475,0.6928230491885166,0.7594577687432021,
 0.04883762974988526,0.23851189114377525,0.14653043644568753,
 0.5683931764933136,0.06864384457256167,0.9720882349343335,0.7092695403428956,
 0.015470742277713723,0.996123241392981,0.07110739489411855,
 0.4663313165068538,0.9599931387174323,0.5813914725884635,0.9568624061410574,
 0.7436220023688698,0.21937205339071886,0.73321745016304,0.17981804539577595,
 0.8641732763964185,0.8342942438957835,0.8616813825730438,0.19005410061893624,
 0.09165564149907146,0.29293258126756516,0.31884363877613797,
 0.4873748204383669,0.38814086375356305,0.013571685128343058,
 0.37040655970996084,0.23819420065115893,0.7711743609276447,0.223424916429673,
 0.12473198825505949,0.3916583291911899,0.4226378508159496,0.8925909478220511,
 0.649023696235897,0.9413134871767788,0.7804754179272866,0.09469799366595333,
 0.7023554750838786,0.4581006326513999,0.6576217819855787,0.1925443975968767,
 0.9838788761544003,0.44090936593390084,0.5184001971045911,0.6937763848708569,
 0.8813267107156088,0.8273572475290963,0.870243049119042,0.24322710840415807,
 0.10588848095706194,0.4491693403994458,0.7485699781248037,0.280233899974622,
 0.17013841413705044,0.874065553482713,0.19786715273731337,0.7340223158282353,
 0.7223489836679065,0.526465287295801,0.5473720348436923,0.8734033056009238,
 0.8898147439652477,0.9370892615120895,0.13027149609936173,0.770127739680485,
 0.11777136373750896,0.8642939395737337,0.1097008284370784,0.558758597090031,
 0.47322986813990553,0.11943417344249163,0.9732627066975316,
 0.036313311040713114,0.3870274545892257,0.9818117338751003,
 0.7589157850301501,0.7523283254444448,0.6306785217653863,0.6793061270081568,
 0.5630743265803213,0.8656483065782026,0.10029614480234927,
 0.13072276546105988,0.9954965539452606,0.5630519283815603,0.2237445138890244,
 0.3350608748219598,0.23875424077445662,0.6426526586140124,
 0.058120067991750735,0.6572079277767919,0.35338088686056224,
 0.9712541066395719,0.02717027951269546,0.4032811525066575,
 0.42898289083651897,0.3046714507193725,0.4772577103592711,
 0.12007102716430362,0.4539354509065271,0.7890397963890964,
 0.20201957141944415,0.07705703325259172,0.1002921375283119,0.546932408981127,
 0.7331016760983975,0.8064658782793528,0.7625111255364416,0.8547940036535377,
 0.25396565424392703,0.8212612453785522,0.7826939966751661,
 0.051096000155114285,0.34132791576927557,0.9542899152548656,
 0.6940008192607563,0.48478944229718535,0.19184370227684,0.8373567962054892,
 0.7698372937893682,0.4814874248918706,0.37009313184191994,0.519293634707633,
 0.07229958288765337,0.6024653954090784,0.6373185933484468,0.8274531289981685,
 0.9313379783863183,0.24949818518234224,0.24209445129071017,
 0.9791223914543181,0.8198308279435015,0.9267907797208491,0.694641545191167,
 0.7626332527369153,0.4978193619612282,0.921714157576882,0.5620377721649711,
 0.7776696367171512,0.0012096721489467388,0.6697859019995733,
 0.9552622698454063,0.41001026721238176,0.48143626421656605,
 0.9094062422366077,0.026405199620804343,0.7321814120757192,
 0.14566878479136502,0.060710501048574894,0.8743603636727845,
 0.22991748715898608,0.7944090909962156,0.5069423037229426,
 0.05175063824333681,0.5998917524966747,0.5938547436054441,0.2863377979507733,
 0.20374395924779132,0.8650631078180009,0.5151122641478638,0.6664227348291505,
 0.6025639470721783,0.3064472218684552,0.9953311030378783,0.8323664813341377,
 0.5454123558409024,0.40629166891433055,0.779531137695852,0.02439266000831375,
 0.02934616475176066,0.6522030538064474,0.6756089346836794,0.329015715952792,
 0.6915488478946221,0.42272803294527905,0.15564439932972707,
 0.7391941742654652,0.8921445344841167,0.5397878904816047,0.21864905644431643,
 0.022635589336498563,0.8368691468603494,0.3019479714412262,
 0.27632940561612485,0.7887555026262928,0.3635445979065358,0.5873283413301765,
 0.6255594596179782,0.7219156914492126,0.17083466084981713,0.5802888846456191,
 0.25620623837101775,0.32554129956086375,0.037143026650369504,
 0.8343012608428608,0.5859166649448423,0.13518413943302698,
 0.019360327258702736,0.007308228219508606,0.7762934248454738,
 0.12393159432724077,0.722670294252613,0.5715787034727254,0.03938471115023512,
 0.8233091874044502,0.4192540864846981,0.46907492096431536,
 0.43738545082705627,0.06326593709792783,0.6897996047231317,
 0.09417430059111742,0.9245618977400695,0.10000667832136845,
 0.5483733966420783,0.6587541961526796,0.20421247285451471,0.6214228890938605,
 0.8224312222156787,0.4388426611868519,0.2783284145400132,0.5080273983889301,
 0.9634335362193508,0.6477505409810607,0.49632942907238553,0.808221703982852,
 0.6896346256725822,0.9670356039656368,0.3096444189351083,0.47199511317310083,
 0.4613044878205058,0.019250489867957077,0.22263675047497145,
 0.6928070136706472,0.5723740170858915,0.5165887198155313,0.29765532963460783,
 0.7429027510762035,0.0953297564708615,0.8181310966041725,0.43784941474814065,
 0.39495538653081597,0.39031615124648755,0.5699443794995404,
 0.32765280215366976,0.7628033520060185,0.6717027803217079,0.5547859097709497,
 0.5762822989485945,0.9565258781694342,0.8356155976271612,0.723660132470789,
 0.41981787413478033,0.2621127283258057,0.4977238297513167,0.7827877227888599,
 0.6317442249513402,0.6512508987558702,0.4043462689588252,0.4415702220621889,
 0.02574720091300986,0.09295519830523757,0.15410236534868083,
 0.26725473192260507,0.13764073169978341,0.06789838384206104,
 0.5636543544878269,0.9412427185907577,0.49008769921564177,0.4746074400864452,
 0.574289439070403,0.3570727905513166,0.7385524519434995,0.2910753901279115,
 0.5821590398253585,0.3713682728331996,0.19487723484376585,0.3672565333847917,
 0.38257838437635217,0.7551848504654254,0.3854439873018488,0.4532092772582158,
 0.4229467961911473,0.7819449376211763,0.2795076122518536,0.7181347226989936,
 0.6186032487552855,0.7820694071876851,0.4908267936186015,0.28686265649642095,
 0.15604311112001534,0.9286620282376636,0.5950701035558089,0.6846332692921874,
 0.17611882288269498,0.6028540326353856,0.7508272389773125,0.572591926669238,
 0.11134895251857568,0.6174100833324194,0.8556771941759203,0.6985166422075948,
 0.1740171753566493,0.35580634526487154,0.04113125965334985,0.723620322354723,
 0.5182595161917841,0.10276000911613259,0.03434366281053303,
 0.1281610042673259,0.5551007830166116,0.4656706381238963,0.03868090003105362,
 0.8302068181721238,0.00955641688959108,0.3444352550145341,
 0.28040470648569205,0.8191606178860839,0.613498743682132,0.5702553774692637,
 0.9087597813651122,0.2698639984901413,0.48196259838602173,0.7361587316613921,
 0.15191987147163033,0.15184582719351114,0.4693706866174653,
 0.8008260848818642,0.4189088007390972,0.6160450543702817,0.39583810301384437,
 0.3546210164582365,0.7087857071530124,0.6778400395444508,0.9853600946608212,
 0.6932657218380675,0.794967622748933,0.3804247805109824,0.24122617009659386,
 0.08686317244204045,0.8088538707963399,0.5852730402619611,
 0.25413822691203836,0.9174171729077236,0.9506203747036649,0.9538000851430403,
 0.25483335677113195,0.9265043332494991,0.8981942408822203,0.2539181634408404,
 0.539818821447038,0.6455109452189931,0.9357091550234633,0.8711257595525157,
 0.05530858170139419,0.055322853172559805,0.5435181588602367,
 0.19652091284676299,0.31654013512239243,0.5720665381798773,
 0.8395671993821973,0.6052564041639213,0.6147550975856059,0.17554922735821865,
 0.9111094991708505,0.7969656626551742,0.992655181654476,0.6559106173898697,
 0.5597752334010666,0.11835792107547083,0.629702613862711,0.35296245177049523,
 0.2638619130828448,0.5185427359807028,0.07689008367227856,0.2176439714065077,
 0.6404210048748917,0.6185625857668374,0.45302304707811003,0.8164178264390888,
 0.7278644079992321,0.8453342282644563,0.18465492385015236,0.8947708912111122,
 0.40076984039794405,0.2147092386040087,0.924289427165027,0.44025522150511387,
 0.8130364873175479,0.11003873397575392,0.8314156727470824,
 0.12995025179176656,0.6324135201293418,0.20558498992014618,
 0.38647485439265317,0.6023904849301297,0.36065358435599804,
 0.8544367443680582,0.20065471327946582,0.8632453991190754,0.4668317224253543,
 0.36968262127946827,0.7855152950454324,0.9875022471812082,
 0.06481672951570916,0.5119770694400452,0.8923227809285951,
 0.16927726242243912,0.8066213290206646,0.5899964342712243,0.6238233631090453,
 0.6216629634463559,0.12618236953776651,0.2745433053977834,
 0.41887180121078227,0.19530650924390291,0.79746458285354,0.8633240697694762,
 0.2705648545712863,0.9558643682842328,0.9852843170051505,0.708877509434605,
 0.14991805889331944,0.39996434966488437,0.8416217237197329,
 0.8619910667750962,0.950240641926911,0.16304970900360016,0.13248146842050057,
 0.06377369081291495,0.0404792304606717,0.6108176600381254,0.2620341304884892,
 0.7224042379767102,0.7580770874718948,0.08177059749225868,0.533484575463631,
 0.8848608076561079,0.22605020202890502,0.7160861051285745,0.5005185239378265,
 0.9020635269842797,0.4628834930152699,0.5674437280915612,0.05985545468809228,
 0.23809433789770929,0.8201806186604619,0.35804409795126646,
 0.43886856047090594,0.012548144118159632,0.05675458392437993,
 0.5454181098431077,0.7513538543485817,0.29301548935998956,0.5697991284701848,
 0.6338660315041672,0.428321311078926,0.9273576131386174,0.9840981857022343,
 0.8864323636538299,0.19489102289068538,0.9646979345634878,0.7709332155145469,
 0.5587318653680882,0.12722550233783514,0.6691850260592027,0.4982227570088099,
 0.782871532220024,0.18274810096389926,0.05506659559273597,0.4976758284533257,
 0.8934884582586954,0.7253051401802638,0.8182827693200825,0.7439009959219942,
 0.7790204591327088,0.05932577706262776,0.22838540433285504,
 0.8323435479258673,0.758913313155136,0.07100914438938699,0.4500824836733862,
 0.16803481957703692,0.005810128287484284,0.9410618658720746,
 0.9713875804640129,0.44896203244527866,0.2686762818181576,
 0.40974120335849407,0.9819451091166761,0.1789680056215892,0.463249063817892,
 0.6837135665159443,0.1668630060904811,0.3658409524017381,0.08490828829050623,
 0.9705421389396349,0.6085768227007997,0.31375728426541216,0.5067526698117534,
 0.5631727922065818,0.7356088411539903,0.19999700637389017,0.7360932219689482,
 0.4717091091036184,0.42671185647525467,0.24648527319901936,
 0.5401009837608242,0.17860625652314777,0.8130090508167023,
 0.31120106493293453,0.5382377924092612,0.7912904603347435,0.6504037521892834,
 0.5259298749406832,0.46441178262059035,0.36450078814360143,
 0.7544012957806947,0.07012160081724206,0.1311288330413376,0.3532981870208012,
 0.5080617762801926,0.724845707491018,0.8735138498971947,0.1549443391104619,
 0.9754450840527196,0.07060733374312245,0.7659622096857798,0.9537647715010528,
 0.03761253631157102,0.20705842808736963,0.35794574903505216,
 0.24052678896692248,0.8464653236072466,0.40101032000583225,
 0.8991892081190924,0.1482185638870377,0.9125436859877791,0.23119131328094522,
 0.6709504523673734,0.5598350323746678,0.9560562202048767,0.9132451876071158,
 0.16559034502377656,0.7668458018711141,0.18763503576175267,
 0.4932365982436211,0.15028636208925283,0.011003206605446514,
 0.7074081039700465,0.72618453512603,0.08514953035272732,0.3553498815372429,
 0.9567964559774758,0.08483686321774231,0.0822224597025012,0.439979731144829,
 0.0808375733469745,0.033913678993015894,0.6912128154442876,
 0.2538809122994411,0.7738817076652068,0.10902607478719861,0.1525042552848369,
 0.001978780001047009,0.7894211281822412,0.7925983151847218,
 0.23118952331693388,0.8481860087908939,0.8354331589681165,0.1476508207909104,
 0.750814555510396,0.909282971490993,0.8274442958333782,0.9510155159142655,
 0.005862022736213612,0.9261080193739155,0.005778063257870669,
 0.8542627683541614,0.12688527215435097,0.982890454397317,0.6100802487773856,
 0.2569470480944007,0.948927330832117,0.4376838055920661,0.663405811718378,
 0.043202255064293316,0.29748101625540657,0.9475144079642819,
 0.7332268659666799,0.2834646471212561,0.6404185632102548,0.15967427796921474,
 0.1441899169186337,0.6512447846016417,0.04037748346843273,0.9950438216039776,
 0.6146132221428348,0.9134458210863403,0.21686077850075847,0.4987994527903621,
 0.03784727758164552,0.8955708911510777,0.6352455111665639,0.7771605336855917,
 0.6697952462418608,0.6232041379987731,0.6785075316887025,0.4825858180542103,
 0.23389494891434426,0.1610388101200989,0.6160462869116462,0.9865590990440917,
 0.9993542509187221,0.8938612256727579,0.04563317636130937,0.51757039844891,
 0.6723862591709016,0.0034274520314031554,0.2376346379532055,
 0.6607265265772341,0.5945057410110692,0.37635563492691504,0.8207967391563715,
 0.02826878891904072,0.7433650624619634,0.7844415279998075,0.9420694721878481,
 0.09451593373827372,0.5921422552856792,0.6471129380117392,
 0.23643930853393802,0.8491371700589228,0.44916399992573397,
 0.8477194153763588,0.9785816793411933,0.3063883287014809,0.4003616897424417,
 0.2746996952816916,0.8882455297240041,0.8482932914418391,0.826286481069995,
 0.24279757287211878,0.5867178106724825,0.6142052772617794,0.163205507962758,
 0.9504503464051404,0.7733779917958817,0.7281780041456746,0.3359226699296127,
 0.8951054437691314,0.5704520399469997,0.7185067347518154,0.35702544510860723,
 0.5999219982862575,0.7438060639890827,0.5579529858849138,0.6231729684773125,
 0.6745479678753988,0.2210454954048361,0.4207430341516456,0.9026605020567515,
 0.4241943236820158,0.7846563675714019,0.07804565380103612,
 0.032428059780273744,0.8931607565242252,0.6244104055962373,
 0.1408251714122909,0.043929566766073846,0.45918733794834776,
 0.06171517943920346,0.5980124207722268,0.8985493346883339,0.6260801799391716,
 0.5786483278182689,0.08804407014303617,0.09252367880344525,
 0.7445075458872388,0.9788530388915594,0.756089953461448,0.2417073543129874,
 0.25218597692407274,0.12501939605890322,0.4279933854431074,
 0.7269937184237077,0.7747550694745549,0.10325748073457963,
 0.35529276396377785,0.049743643496638956,0.42312093423582886,
 0.33676751658208537,0.7488539346422975,0.4099830504411177,0.4561775335062954,
 0.9694172700205425,0.49527386556606756,0.6526634485496475,0.88348371245208,
 0.1599675012091346,0.31540446466791805,0.06437542510763183,0.598819562525198,
 0.4234007755024678,0.5171096586652473,0.8307501085606245,0.0959418394374364,
 0.3635697000853352,0.2693095278764759,0.2712600024468731,0.8074433598837873,
 0.2820135088187573,0.789009115639988,0.449911617473429,0.016684046591398327,
 0.08433276710068038,0.9045748583589307,0.8690533262909287,0.7612800566148739,
 0.9650407827943462,0.6354991021815475,0.8770905171245885,0.5116544531940879,
 0.7279125830457989,0.5945761234011989,0.7299149703398007,0.2027366995952138,
 0.3867832231229952,0.05767099174226553,0.5008540546232139,
 0.40940322833105003,0.30653386001124516,0.5591041391072522,
 0.6130979839144591,0.9341372957349908,0.22829066881280924,
 0.42886094653720197,0.3146556045768313,0.42227580593362213,
 0.17339257283995135,0.53563936262725,0.5187338249243183,0.044680095397192976,
 0.8185904757540999,0.9125821829238744,0.9867853723734236,0.4355768174896304,
 0.7296261668475363,0.4553532785572576,0.9706058282634431,0.805190845095657,
 0.503859333519018,0.2600996783160091,0.352823758149326,0.5472630409622254,
 0.400158692848238,0.05541648191668558,0.2603116314121592,0.6208609832736984,
 0.14572398143731735,0.19303250648373937,0.0973463786839257,
 0.7014611714749067,0.5677116946817845,0.887111264872288,0.4366511337013309,
 0.22525549613031903,0.8514247227711036,0.7097365826870133,0.9198729447357671,
 0.03944763239428095,0.1646181397559111,0.9213298477246171,
 0.23968570913847753,0.9076889926603264,0.7149247395657584,
 0.08843729616394325,0.44938012685004747,0.07872073578000238,
 0.027589189390270663,0.18118616011706978,0.4700963787946023,
 0.5954759038127204,0.8490945511895986,0.48150179301219587,0.6077135543343087,
 0.6320553017328787,0.9865384936636925,0.02640709032105848,0.8992090849129252,
 0.3139576132757542,0.8527630405888107,0.8806092999385092,0.9301797617394258,
 0.22575316793173028,0.4496645415121019,0.3563557232974055,0.7695304624424841,
 0.6627318324927545,0.023599582819190612,0.033857151995022705,
 0.358516248839688,0.06061635166720347,0.08276160805025257,0.9486952560772988,
 0.3647915405567891,0.37439574601211456,0.3596970112520278,0.3701514971343689,
 0.16771687021519288,0.1363712547656285,0.4082098854751608,0.668084375964656,
 0.3800169658241179,0.9117296872143749,0.26824988844138586,0.302284071349167,
 0.4693789039754448,0.467677565831353,0.7783154566028186,0.8327288561788398,
 0.2917477474508491,0.5760575726264827,0.1861177393481166,0.8936672154971891,
 0.4350355371055261,0.7505994782131657,0.08555100348488098,
 0.49618993420715996,0.5118139693506657,0.9308520162023817,
 0.22116981479057762,0.9974269043702257,0.46451025186979944,
 0.3702049111685938,0.24534712407705028,0.3027967196769693,0.9896579497751614,
 0.9917293985144305,0.04786176783193419,0.8435998238517659,0.5818761281880471,
 0.6198905410867634,0.8336932418224763,0.5020310706497831,0.8671407064479613,
 0.6319016541498735,0.40823364391844297,0.6288071939529561,0.6226165833823878,
 0.991013079875512,0.017159170301800697,0.00904506944854111,
 0.055345843614022394,0.4508996991647196,0.4737317496005482,
 0.13881289545178177,0.19771647223114408,0.67319029122699,0.1639429898665531,
 0.02934384835129089,0.4668670110208264,0.5489017483165772,0.59372512549592,
 0.6282877513244683,0.5682803482893076,0.805472061681313,0.063520610236933,
 0.5579449031160392,0.9069679867766591,0.6890388199719306,0.5539857227383278,
 0.3187893305462266,0.6384883372012119,0.32362466097777254,
 0.18600368250007038,0.285035335786341,0.35991837165591234,0.7841262997668847,
 0.07179897533545454,0.27769807538753777,0.2960434088905557,
 0.18777542794653124,0.9183150297989352,0.5272458337315319,
 0.42969554569453805,0.47883537156192857,0.6729629316051158,
 0.8328543531935578,0.33058417337134516,0.9308995570812677,0.9406752207310458,
 0.04274575949624615,0.4956921813657956,0.37336949155928023,
 0.46555235210661405,0.8491153620996886,0.0481787332120277,0.5679046935885373,
 0.6032561417596563,0.5728451768948175,0.10739055771122841,
 0.13571354152032022,0.44836529163271954,0.8376675321245071,
 0.4638697963482953,0.3125257736263478,0.17612131318659352,
 0.32616040229475673,0.14697767719496035,0.8633535156043134,
 0.6329463000122226,0.03529880725911205,0.14854767312890937,
 0.2102823510248415,0.6070455172980601,0.005534779836323445,
 0.34809418400275716,0.7194328414032464,0.1223006015874164,
 0.36412260360268156,0.8905161682599253,0.43355607496815207,
 0.03385515011152895,0.501612041387558,0.056743914464815304,
 0.2663713872254285,0.5223346310425454,0.6108954694328327,0.5610803828547274,
 0.7985094452721259,0.727280198177536,0.5674762428553746,0.2488713863237283,
 0.06358962056636253,0.9029092976785759,0.8961208935519703,0.4661450539755897,
 0.8712966918433027,0.9907894834184657,0.20444152454111197,
 0.39652896679175953,0.42214386875453047,0.3472649453497889,
 0.3944339869592012,0.011947799265771386,0.5586686637449116,
 0.5876765151088242,0.9777996901477161,0.4767892498223868,0.24240207502557765,
 0.639197764805528,0.7051269723598333,0.38682345681744246,0.40587645144727136,
 0.5165401036434498,0.9891380942007947,0.7291626123342103,0.8983830529300835,
 0.8690194155504519,0.23882714495433788,0.6299840004628988,0.9129293724772279,
 0.9265993170605457,0.9272868426223053,0.342999394773855,0.22162430912164566,
 0.16673540448246738,0.05210589165448343,0.5137810527242859,
 0.5865393311940963,0.6309402809953029,0.3763973481584456,
 0.0020312699838024306,0.7417863912532946,0.5075592395037966,
 0.0835750732690197,0.7044168715448591,0.47583468965856845,0.9355117195356257,
 0.27026732430806943,0.5552762807756346,0.411427812649252,0.09462729925476887,
 0.7649400515828174,0.6918143975517181,0.6827254938489155,0.5645094956040136,
 0.7844535360207128,0.47292201232894804,0.9623573588537272,
 0.13328332999890502,0.14464495453562176,0.9771298030612289,
 0.8798539628482662,0.36922833878520245,0.23472260105134657,
 0.22649479618759005,0.35288864795241026,0.2926741762403804,
 0.6802541654774094,0.39269599206283257,0.8815349058261259,
 0.30595262148520563,0.5333536373865625,0.47069817674010506,
 0.7274875233455834,0.46516852896782424,0.44645590824400516,
 0.3153909930329588,0.7569350343521054,0.6934414102714531,0.47438360010118874,
 0.6584894293582538,0.021999953208967526,0.6892957498397064,
 0.8243870185883851,0.2223406697334943,0.07476994072377452,0.8689248422406317,
 0.7931058202950008,0.37310496672387505,0.7494842266846328,0.5394402853756279,
 0.5725566826327204,0.5253133452035131,0.5777294897490974,0.8070730384085657,
 0.44643138955642825,0.46688353124211013,0.9470924091642505,
 0.7273779415710547,0.20812065203022634,0.9965728887394955,
 0.26915858884666655,0.5879421843558585,0.8113847526761986,0.838737225511786,
 0.9589924493853447,0.09728767401304561,0.31350779364308623,
 0.7302673189075323,0.46451352085451125,0.45014659136558743,
 0.6144104822212229,0.018537372141189112,0.7537424383206488,
 0.21633685960390403,0.2880281010105106,0.6066471729326834,0.3643528308153883,
 0.478896470229907,0.22421041515053952,0.8143812139692047,0.3573355625529281,
 0.9302106935108143,0.6791806259815908,0.48851804424227974,
 0.33989923674723244,0.2964901271450817,0.5563131540497559,0.9237875118635205,
 0.5538139017474573,0.23196437353159594,0.24018281670904051,
 0.4960542854711001,0.5026285595764144,0.7910050079184312,0.8829926904977413,
 0.9000258507799402,0.08354756750641834,0.09824622236074809,
 0.6453436237455669,0.39248181292856765,0.17091244847679465,
 0.4201887235750206,0.5210508632137414,0.4186962262009851,0.47674979506488324,
 0.3762290076393512,0.20043919727198145,0.025415205027861543,
 0.6683780289860228,0.2222518305966431,0.5877007609502295,0.24106187093387887,
 0.8603959305206529,0.7820889774485708,0.29330482888705156,0.6365245522507484,
 0.3919131122417705,0.7001377024176512,0.5250100079268262,
 0.026815733862811664,0.5858060430627527,0.3036897096344662,
 0.03671103292886224,0.39442348664655025,0.5982755510161276,
 0.7404018849374147,0.8610293327498776,0.7425690122799283,0.6598673166553791,
 0.22761874688207295,0.8687949430161347,0.40768372628744953,
 0.6420220765456289,0.7241455552896082,0.17590022483805245,0.2782835369444625,
 0.29074883491238246,0.3656437263818906,0.4733415590427297,0.8399297716738963,
 0.7890264707693099,0.060801277508817364,0.25703876042489304,
 0.026246903303003388,0.11039483264941652,0.588380423887232,
 0.8327741549214098,0.07252440608051036,0.12634315819257358,
 0.7866189832975728,0.4564452525276298,0.91345950414107,0.6588037083738109,
 0.34481129730151017,0.0911625397671556,0.21063249445405985,
 0.6009721064853064,0.6152931765593541,0.5351931251569981,0.8340895869706739,
 0.3143312722214662,0.9281414976799177,0.6391210717421181,0.6749966390851346,
 0.8844504528771056,0.9671780521734996,0.02602827744600955,0.6836542997583414,
 0.1796481670737734,0.20662514028211632,0.9301438589442311,
 0.16719120449309766,0.13033654647502824,0.18403856663413865,
 0.2881669145774798,0.777691368245389,0.08559159839674324,0.5350849773874444,
 0.9552283832628179,0.6589011944516875,0.7060426051372957,0.3564124118415536,
 0.9548641455725045,0.23466263542143628,0.46110240530822055,
 0.5858658968659174,0.30855822279009315,0.3552794415873407,
 0.01903197075585894,0.3221239148157511,0.7052105764824779,
 0.33270517472452454,0.9683367518983778,0.8125965301386117,0.3363548599258632,
 0.7165473812827545,0.5363290297252676,0.8557920024181166,0.8389777690254732,
 0.49517396410806047,0.9004601459381509,0.11085576279786769,
 0.4893518601314101,0.46407951194943753,1.8517842720577882e-4,
 0.5089388848456378,0.8451617072173133,0.7803009285601435,0.6628440463947709,
 0.948026963790581,0.22823743019518095,0.12357760771124515,0.9723064341282531,
 0.36977305864094856,0.9565016855351419,0.3526076667342155,
 0.02347118787409319,0.17701253727184485,0.8678347420625127,
 0.46813876406489796,0.4023835092464073,0.2527266146264674,
 0.12821854513448283,0.9380206914498554,0.384420126552687,0.7538763826915784,
 0.7241651727745038,0.5672080663572989,0.6334938896820271,0.45695760445945155,
 0.005911225459613645,0.45821444549412627,0.5799515297888118,
 0.5428464391086882,0.5434433709595595,0.5907222509884552,0.7137084176816495,
 0.6439853807713564,0.7261053955666633,0.5113928461789823,0.520771590477696,
 0.8979816735229953,0.6084736469377702,0.4545441770869252,0.5700647374212456,
 0.017830066100426034,0.23135394108648177,0.15092921583358043,
 0.34152162735329306,0.02986879804563025,0.056359229347797246,
 0.17703334737970666,0.20488762113556647,0.438552565032052,
 0.07710091736751967,0.6326446740900641,0.5517842759885679,
 0.16162180878107602,0.33627744084456135,0.494294359261892,0.6675788816197686,
 0.16447057790781594,0.9309245883780217,0.9101528974860875,
 0.48836711508690756,0.8269290773012855,0.12158717127313756,
 0.39608699349957144,0.1971922409240593,0.12587477133442349,
 0.38811539665665207,0.7601633027079308,0.5967282890341715,0.2686603876560112,
 0.28779872982399224,0.17928542939368608,0.08095497265252871,
 0.27650220564915373,0.07671814697378343,0.8596638812205332,
 0.24817539691044377,0.18780115594394853,0.29954446791968714,
 0.7344648401864052,0.6890719112680496,0.35724516911649384,0.4801301869009226,
 0.4803381037980432,0.2760260075929213,0.10133088592749406,0.9820637347598229,
 0.4743956007697975,0.39373112987937586,0.5477151300591963,0.8679470514497538,
 0.9867351520194356,0.19458290963829983,0.2762317797390137,0.5486872595762425,
 0.9004508971206548,0.02408145077883772,0.5094443517400232,0.3971500692442198,
 0.20224415034637655,0.31258595737242656,0.653850721025008,0.8905075439219783,
 0.07171791180584042,0.8011288286945293,0.2636483511998228,0.8136204438596192,
 0.04981635088888403,0.8259386723582868,0.6880889834323544,0.6604459864828403,
 0.608835957014839,0.6058219334890307,0.7922634749724209,0.503930715171162,
 0.07551027378415509,0.2846717385133639,0.4995635364312567,
 0.37757455398227224,0.4995959129222134,0.40918569522834336,
 0.03693411130959745,0.5962509968568017,0.7991541236721696,0.564705054586549,
 0.5177994616846672,0.6607809838490621,0.03693278047152515,0.3394759313585475,
 0.7678939253986191,0.7832122929396708,0.42789610406543366,0.8899948105274235,
 0.3267668276041471,0.5913162980961761,0.49762729135865735,0.7974988320118414,
 0.3060627334324044,0.7155234281511824,0.42606305610095707,0.2121886183586964,
 0.9325409498644356,0.6619481753362401,0.870550117650382,0.0439239893191663,
 0.38403860034015236,0.3552619242290136,0.530616437503242,0.09240313591089455,
 0.005222133744322122,0.5029819468938312,0.47581477583605025,
 0.025298922800203183,0.9574387726383127,0.6367567614272754,
 0.2439846460817432,0.6881446187211796,0.5202671526735078,0.5316539973629293,
 0.3160052833905025,0.18382652469640715,0.14606748978037587,
 0.07170348337705112,0.9856038258256373,0.8301738577464977,0.2900898396308086,
 0.2516087426349577,0.8162410926570285,0.17547698974947035,
 0.0011997027316318354,0.8504474646851959,0.5337186427071028,
 0.4634942503313697,0.9440542099207674,0.6596077876892855,0.75823312947197,
 0.745126167592067,0.41107814471376924,0.5033747676282806,0.6466629123569572,
 0.38736361803450126,0.3418977254618065,0.4921637927228111,0.4844899829096866,
 0.8880999365164745,0.5105690180964465,0.9518901074543042,0.04466794567389598,
 0.6058134677953916,0.7404636608349854,0.5803921902169601,0.2905811331368465,
 0.3211412778026168,0.8577985197288814,0.988920119010249,0.7624035254693654,
 0.007876146192560163,0.7552982018486537,0.7609180622038083,
 0.23304464602300068,0.13211500351844918,0.9174326819247917,
 0.9172986027709324,0.8371868230807026,0.7740185055050995,0.8041832946498143,
 0.4721962223367313,0.5902423449988801,0.16688599027788054,
 0.014664799602113088,0.14365368526000033,0.3500983635531357,
 0.37458382287920067,0.8977206082749688,0.31121714873618844,
 0.0781429991607594,0.3373006850994449,0.9294987584043393,0.3056764810669488,
 0.6296897064081638,0.8434799681750353,0.9758967701442465,0.3236699954341249,
 0.9668282231946752,0.19043804099124118,0.10278654960625011,
 0.8510319872962369,0.2720458011523582,0.12147129049047156,0.1075056992268768,
 0.7893172679396075,0.15135201131580733,0.9444452337981561,0.4276480544874819,
 0.21262657934311413,0.9946055213908644,0.6141242944024063,
 0.47664723350376914,0.5276264079885935,0.5524327260086644,0.8785487169029526,
 0.6318427575171289,0.985385123890984,0.14922192243427523,0.4016024009461816,
 0.3908050150884056,0.6258660586950163,0.3718020147043566,0.21660412492387626,
 0.049123912238871914,0.18796069876277954,0.2453624533048625,
 0.9432615898565441,0.7734772410565124,0.827771078004115,0.1752413700373493,
 0.8385235278283311,0.9596581084399584,0.036418187051328044,
 0.13871713513041417,0.6370138997867913,0.47638946184606723,
 0.8546064865787381,0.7938146437020307,0.46151905303293894,
 0.13493681757399978,0.8932745765959555,0.43627563408808956,
 0.4803432257429111,0.11305800908250874,0.8787213986213195,0.5004537033104297,
 0.06263387708372459,0.22843293371251416,0.036830464177982414,
 0.2222440889066809,0.6760246041738459,0.3502930684683586,0.08053878610655052,
 0.793362423132551,0.48654505572736095,0.1943375819319616,0.938213266338863,
 0.25925001631402744,0.6407353505325211,0.6247540517214529,0.652835089170138,
 0.6054217549316153,0.5700410345937535,0.47273680944265806,
 0.09270631595474088,0.8417863199837854,0.6309809124581649,
 0.027581159661789423,0.7042894743910535,0.029296029547344893,
 0.7928622841636678,0.2829498831253424,0.5342224426872539,0.2465102484809676,
 0.5573112980975886,0.8965505966393245,0.7724815421071853,0.4441033124764091,
 0.6263547642823732,0.3692634520871442,0.7378876208161134,0.48437194100304626,
 0.5833566797721835,0.9150463474496501,0.03758387952484232,0.4445849208057753,
 0.22440314816440798,0.4329491549311453,0.39490994696545134,
 0.48239851863235517,0.33623120482556224,0.38491242621797295,
 0.0867144794404836,0.9000953044621288,0.6904075703861032,0.7347930341460506,
 0.8214303996863057,0.0762186293532563,0.05117849290060528,0.590511960276058,
 0.9688630201806557,0.4988483736186635,0.7294075158499901,0.9374071486855304,
 0.698860766484777,0.40411241525669006,0.9326414350717989,0.8497195871850028,
 0.9685393856252688,0.29722202574679535,0.15408480267427782,
 0.7627735350128274,0.19838623214942408,0.6656980218192772,
 0.48559206592064486,0.11330165981615303,0.3732477447517163,
 0.03019887744749039,0.24616382425955907,0.5281864009578197,
 0.5565575188774968,0.6063578793060966,0.6578882730436443,0.43862457615726624,
 0.5263338840628584,0.1270508623463259,0.36344895564513313,0.5305046083171856,
 0.9955063884258276,0.3988439347899604,0.6266554668477898,0.4979020451070061,
 0.8218987308960225,0.0694877365104355,0.9420606228532561,0.6652205991845099,
 0.9480838750743655,0.19523653519232453,0.14689287029382259,
 0.40055744132243676,0.6075311940001333,0.6652190884354541,0.5652263629397732,
 0.7074604500117638,0.7820603511645333,0.7177484083274878,0.21111290377407643,
 0.30749981203458643,0.08569265407701099,0.5271596986919829,
 0.8601889519706627,0.21082064496880626,0.8749111484145801,0.957000278698893,
 0.6921740954432488,0.6712061591262776,0.17022631662180265,0.303598374516476,
 0.4379529887485295,0.8527189751294881,0.9841388241747308,0.6404530659509996,
 0.45061504532475216,0.664093894045812,0.6841672554231736,0.499673819823638,
 0.8287486192812585,0.7568089233625991,0.8967556500186373,0.46047835506376056,
 0.5147791912299937,0.9087743653210745,0.5843851458556714,0.00869744856708643,
 0.11731626073367063,0.6274268794992207,0.8754572936386789,0.6100520785247945,
 0.20182673897522763,0.1722166627686913,0.2627011779776135,0.7441071115044368,
 0.34782314687576954,0.04444284381879904,0.634310128252738,
 0.15885644306877467,0.8571038235279927,0.9643261794310758,0.9985180497343769,
 0.3386504599826943,0.7306991737075856,0.1976540912100394,0.41914849250993025,
 0.9766812908132673,0.25453673790505327,0.8125597176561915,0.8006505090824212,
 0.23413283180990596,0.5751109020353226,0.5705137789217277,0.7616833571535179,
 0.21528717353422466,0.7753279886869492,0.3137529975703366,0.5839131572473233,
 0.8239013090193681,0.9716597006419301,0.07146923377580228,0.2505042398314018,
 0.2387501360415648,0.2345085922810206,0.18423616225965334,0.920919327112534,
 0.5980286019191888,0.7090776113222715,0.7100019955669753,0.8405251399552737,
 0.6573741948884825,0.31433497995006787,0.8299550944609088,0.2157476983987483,
 0.46404659266669046,0.9545773823803307,0.7006285557332734,0.9835780858282914,
 0.4266638844466233,0.3964977194129984,0.01926020636946968,
 0.15755843758473909,0.692941211070458,0.5009148226162596,0.1643691508980727,
 0.7608911276343991,0.9366702181120299,0.015521016091863737,0.580792117035195,
 0.9417319270867587,0.27342453830152136,0.05425405493910018,
 0.39118916537204074,0.001088913087401,0.21832718219019986,0.8014008037284761,
 0.4486476503210479,0.038587860145652964,0.28672303110470065,
 0.46647331947839177,0.7688984891266473,0.8256868179715763,0.3005414230604678,
 0.42506715577255605,0.0175179656843113,0.4518551616634561,
 0.30961748201116535,0.23229471134449609,0.9480775879462277,
 0.17927126645387115,0.5703118227255611,0.8051596463724136,0.7176245293127937,
 0.5776051559110297,0.025535040236626894,0.6959436961330702,
 0.49211171891909244,0.6682277025629452,0.3917919071641873,0.6503633693097425,
 0.9573441213786705,0.9978636699305115,0.15277117246150607,0.6269666388388823,
 0.7404805404903844,0.6801258600751353,0.799699624215835,0.15092497798282267,
 0.0011597236121537735,0.2610265064495064,0.5485133357226777,
 0.5627696278894423,0.41559431727643903,0.8912522892435266,0.3539829289101267,
 0.2530724342674705,0.34875862647337286,0.5201897297164448,0.7718055956216417,
 0.4571609705791939,0.19245131654305436,0.039332760746432216,
 0.6553550898684286,0.0011280654536219847,0.6836968419316144,
 0.4843713132361316,0.35118500213428516,0.4655249823964529,0.6572948788724765,
 0.8900120901814192,0.18536073920370955,0.2983353083795828,
 0.21079138608340764,0.3236560589868046,0.6386342828889927,0.6471813425843435,
 0.3332887957313331,0.8475217716895679,0.7399544270834052,0.1664783214148713,
 0.7663116328269254,0.08701809736494781,0.10950131946023778,
 0.7199592159202749,0.41384031554580936,0.20208289619806508,
 0.9662598937092922,0.6396665488399806,0.9115804448425987,0.706411234216179,
 0.6555091636407311,0.06342745162910535,0.1768154477935895,0.316383728581874,
 0.9397015593169307,0.31742097642809397,0.9184061266284131,0.9680066894828349,
 0.3058756041005808,0.13437351751000026,0.31361023897343265,
 0.9781869576545605,0.6634291749021779,0.9126225083225035,0.6339663929474032,
 0.7262495348339155,0.6785165533514697,0.5905007962952619,0.466767555413782,
 0.49212718658375887,0.8624806998480062,0.5697209869699444,0.7026167802955282,
 0.49753740589138074,0.5531373395505395,0.881834518682026,0.8898103422321192,
 0.5519954520657893,0.41990297596117665,0.7922911290020014,
 0.06853764549966956,0.5982052506835616,0.6410460909442794,0.6066018922376942,
 0.23010959644710682,0.7290817777381067,0.8817143937875772,0.9413630561751409,
 0.2138369535147716,0.33210082460353885,0.06081885762498074,
 0.17248954914147352,0.7216705278323319,0.6867899955168169,0.4659030740928791,
 0.11675477449960936,0.7881983530513406,0.28402829708313215,
 0.03449513811426841,0.052807789835033536,0.47608239756179815,
 0.2688994394065949,0.5243044874236861,0.5349947313568121,0.9089854395818175,
 0.1693861081939203,0.8541366670454631,0.2523001486342671,0.9786507276069444,
 0.6162158284599835,0.37172909470368065,0.34894921503033627,
 0.7346727555091431,0.49654028856655086,0.01651298721523331,
 0.7530729738641565,0.34607684155734253,0.6854297767946675,0.4849421704720639,
 0.8191007353948231,0.3150343547791171,0.8938051248827203,0.18588915363372305,
 0.8241912108334961,0.09713423510770147,0.869329114074147,0.4086772725959378,
 0.47918030950919666,0.6467760014508249,0.8035392436211336,0.543969433512383,
 0.8235696180290262,0.2353818487476742,0.8461396452174265,0.32137429418332986,
 0.9035296577567349,0.49115038253409216,0.6916416531466568,0.2837032716930783,
 0.5410161165338025,0.41950345720259097,0.2874426895170237,
 0.20745909649298322,0.7860087318097193,0.1851761368716196,
 0.005986906066093423,0.9433787442993345,0.6038961683404158,
 0.4257481709674099,0.7394417437781,0.4832161196034812,0.4646801829850349,
 0.6801619696171297,0.4366252239422914,0.010412738768241225,
 0.9941707458640281,0.7691341536760145,0.4755075441867058,0.7917791292133975,
 0.2866765466154503,0.5399740656109017,0.6522832560958619,
 0.026308737231882318,0.6097617002062965,0.5198872781524803,
 0.5285881440396992,0.4923740327408207,0.09402119942169218,
 0.03823977791468369,0.10803530615927603,0.006792709892818216,
 0.10121075413525993,0.39961308929537465,0.8341879752443253,
 0.6443057404589212,0.5784693534854005,0.09029119662363738,0.3622449188703688,
 0.8585006796086238,0.4710634644468463,0.22818272566309783,
 0.22048387213522536,0.8898533103770155,0.9142449096686547,0.7527474108929948,
 0.6070007704387869,0.8527197484240466,0.2531799652931115,0.6516307187289854,
 0.7328799568655776,0.9568144949250836,0.16490467184184332,
 0.42834107224188844,0.04230891761091182,0.07002509360728748,
 0.39995128715823214,0.02855934165520857,0.5873545360634314,
 0.5320924595918297,0.5060741359201022,0.5287129293956957,0.5557957518988551,
 0.9075457990826208,0.7147624188296089,0.07264894387636833,
 0.07364007958656638,0.01679381712406236,0.6460609241935845,
 0.8061756023823874,0.708984466080726,0.17868741765557183,0.491032143409571,
 0.20500891947124233,0.07205358242935822,0.9361997105489558,
 0.46117292899146767,0.3397235480280575,0.3197205224628564,0.1286883989746601,
 0.29452631830509834,0.8095099658061587,0.5854489152256054,0.3599630964470164,
 0.14338334491013294,0.8832780777818522,0.8354526215725387,0.3842776678956463,
 0.7806203996078783,0.8899056803163462,0.2942245820448576,0.20211497016775937,
 0.015642844995178318,0.2157477880799049,0.4475871732656663,0.979637250944973,
 0.9253213396585611,0.8260776166063245,0.05597409409943299,0.5410262087814599,
 0.4144533687167208,0.2835317953251162,0.47479824436292595,0.3149082862306496,
 0.19160182713932572,0.4538048783818819,0.33278861049890496,
 0.17896355177118162,0.9227078750816836,0.9473083536399599,0.6191842661957645,
 0.7006972781710548,0.6046615470833212,0.3576882544958674,0.20667613318197275,
 0.6053921902068825,0.2622641248777545,0.6977873623276177,0.5294146643723772,
 0.750015986441467,0.30284793732400916,0.3193436243461686,0.5403895745619944,
 0.7375840703615963,0.6911607841119506,0.19953941794057373,
 0.03927288134170581,0.6495947862645709,0.739725115966864,0.08546317491671651,
 0.30157792433738906,0.26132034975102714,0.9601663792373113,
 0.13315328494114143,0.9361233800350808,0.9581560773703983,
 0.23772790175563308,0.44040519788867494,0.817394316320603,
 0.18419317321881712,0.25844369056237226,0.7614099742529485,
 0.16920087927960115,0.942306003683068,0.9302600267734455,0.37247143734707844,
 0.6741020804168846,0.4626309503029131,0.7043695930295462,0.8650725586848609,
 0.7305425955056839,0.6350100473812844,0.9610855331303253,0.9100488681582346,
 0.45476279215232884,0.0978235802455123,0.06864125032038315,
 0.35337205438682295,0.17514016718173586,0.5988637613108384,
 0.7538187549884949,0.07193377013409163,0.6354583840795438,
 0.013039765444065488,0.15099042685168318,0.9979795958203516,
 0.017192371100285042,0.09470886184410854,0.6583452010976946,
 0.5334104646290445,0.4510998933895791,0.12918908505867788,0.6733687362019397,
 0.839113797362984,0.2847664335504342,0.6936053218971696,0.8260056636773991,
 0.9938331435425684,0.06440566423287475,0.06916243569984903,
 0.5675625072819961,0.7242314885462398,0.5398089656529205,0.9559432861735625,
 0.9261006117442862,0.6796386065828525,0.8664275889666501,0.9967773805731077,
 0.8505960969914096,0.8875977185022312,0.47295040836073765,0.6233522473928665,
 0.4473753169774618,0.9377278966393914,0.5582494680972405,0.07802327266464093,
 0.8850356918295149,0.7859274867886288,0.8114144667907218,0.8860055914102987,
 0.5637535588393119,0.5657664157142099,0.6218069586388644,0.7620395445427584,
 0.34363715310873877,0.5245590017872065,0.138254441009667,0.162046470202578,
 0.35592500021970286,0.764171945768397,0.5800583209345651,0.5987882211748038,
 0.4130194728574281,0.0916662728557538,0.7204891529267846,0.5712632003645588,
 0.5507373509602972,0.9110738889686189,0.0421836537807726,0.9305670261690943,
 0.6142405832604464,0.2732889110098331,0.8039792276759976,0.3658159743541506,
 0.1187041247224917,0.10894710110556771,0.908275139164286,0.1676898994117919,
 0.7775380644500083,0.12536778229917112,0.4757397456238792,
 0.054187019410626425,0.005792600406997583,0.06357698311110416,
 1.8581426772878995e-4,0.9111037543658315,0.4743204833233208,
 0.08570970529530109,0.3262652894743062,0.5435256087809504,
 0.06527150541756388,0.3037284448019292,0.9448994241061932,0.683032841473064,
 0.44709727800718047,0.6766564980565325,0.9778989419518943,0.9835724316012121,
 0.17682957028988222,0.3840829290301847,0.7063751326682054,0.8156282350967583,
 0.905267355621966,0.06612589499389065,0.9209702764189442,0.09757603526224246,
 0.9256465249515939,0.055018281787833745,0.7457315107517027,
 0.9761798088601651,0.4232874106159622,0.6139813202598237,0.6497186322118195,
 0.8428185458516634,0.08365754205111753,0.15437746035578448,
 0.5695216398036138,0.16518744771899696,0.01769240041102582,
 0.8662236558346846,0.746270413836332,0.22592232724035644,0.9504171967922551,
 0.36298753730903366,0.18173504310649868,0.4697754288118492,
 0.5347047402500718,0.06779518783014293,0.006429957661082675,
 0.9846322246508104,0.655733992028586,0.3459899856348432,0.07935622616161542,
 0.6739466869285832,0.5669405434059609,0.39172427963299083,
 0.07785901386339922,0.4103530764547686,0.9853184182937602,0.9134707086644668,
 0.30120714181451214,0.8352902493257931,0.13961282321057078,
 0.5748907607401056,0.6192739963959781,0.878766711534434,0.09202655694451622,
 0.5819748162799183,0.386176673520642,0.7420852448424087,0.5037789805579508,
 0.9981652823962079,0.39364578681207696,0.16072481081001633,
 0.5016785226490154,0.06391691442117875,0.7882607228250883,0.5733048669752742,
 0.11846707731219008,0.6561687208314303,0.26212742266658284,0.722201006718398,
 0.4621437695132158,0.25551630392416047,0.4253356193031812,0.8993483116930346,
 0.8286211274217195,0.024973448660717046,0.4843909925834363,
 0.7995250512833676,0.24502632532554147,0.6879639348267782,
 0.11025125482531217,0.5323845556348185,0.9957508611385946,0.3780546387629835,
 0.5733400094658628,0.49513874364354693,0.8863466237803156,
 0.33182676470121986,0.782239423897251,0.6456945324313279,0.8511335490622913,
 0.7530825571839398,0.2834419101422767,0.6749190977512514,0.2591418716295141,
 0.20152339596417623,0.007232218854333006,0.429599265121529,
 0.15803752635699875,0.19516159371130803,0.4706968801783735,
 0.7536218570452127,0.8384362041058286,0.2525826644275422,0.8253429893271057,
 0.4182107532221624,0.1612960501267372,0.6086844318083539,0.27849405802316374,
 0.4835386934656891,0.7827270188464222,0.6748149303103246,0.0777846616004767,
 0.4446496823217131,0.5138894720439239,0.4438016555639146,0.13009736991572263,
 0.33701109239492455,0.3429861946031211,0.6206530142506742,0.9181912278672635,
 0.2963481440932918,0.21433768222106075,0.26138694856806743,
 0.16692018112815943,0.9043675343743802,0.014085314770592272,
 0.9981502484680774,0.4044647561162137,0.06437796684059705,0.5008254708764521,
 0.9787403477712269,0.14958198946025125,0.9781762865231924,0.9694465953936126,
 0.3979096642692853,0.5167083293654458,0.40506647495558634,
 0.35360620539276455,0.483064330671921,0.29532217723855936,0.8362274534418312,
 0.21789175207215106,0.2200423111919816,0.2401672855798992,
 0.13156088163911117,0.7881252354608637,0.7219164801299413,0.5814855360817162,
 0.589787413519632,0.5635033820894049,0.6399683770617943,0.6880147636418258,
 0.6673953848170889,0.27490239728863475,0.45432921084412825,
 0.3873707078552472,0.4462104884969387,0.9782109224029416,0.6484517690193126,
 0.05215040483312683,0.7695763726562554,0.4835051917482551,0.7663654491239931,
 0.49298143258252847,0.8564259505514411,0.16047153798502378,
 0.8926696124204078,0.44548722782115346,0.5418571268910474,
 0.45335507060920044,0.469858489246353,0.7005125161179342,0.4995761832322707,
 0.18755459636227834,0.035895896277133854,0.6884853458464854,
 0.5483215507907655,0.33438155309544304,0.9980549242925427,
 0.40663145621544283,0.8377256395899466,0.4008427358420641,0.6282422168442306,
 0.12114435562459147,0.25146570086301645,0.026737302136310648,
 0.6666858770403207,0.37178813008286005,0.36278822866144755,
 0.9239540376090463,0.5077392076313235,0.6772406265937203,0.6154837704506443,
 0.579758736340134,0.588547889147954,0.2829500948148518,0.8284139026829787,
 0.8989592888251139,0.6998510980211878,0.6296984570088666,0.5291764724842403,
 0.07198479582155348,0.7230081549722102,0.09293204373767994,
 0.4020116042123526,0.3272428208964716,0.0016508901712379842,
 0.2029192580463316,0.46191912284113146,0.14859819099036609,
 0.8230210835751051,0.15992777685225723,0.41854159354350395,
 0.7190993017051723,0.2770788479007742,0.17786812633530247,0.5529747622433827,
 0.4578997708872702,0.5138490858020824,0.6433420603087419,0.8894219832577247,
 0.2607863462855171,0.48215967566055806,0.206386641466352,0.07942768982084203,
 0.8748261998743769,0.03884412758438438,0.997638855566553,0.7592532961782452,
 0.9113425006689253,0.4418421644102849,0.5691035166648367,0.9717802276341928,
 0.5436680291023817,0.2956426950655594,0.39410122504813627,0.4562537493445131,
 0.5225456601892562,0.39141419460672155,0.6266760481545455,0.8012113816007959,
 0.5258409124022934,0.13937372444187346,0.022946999477614027,
 0.6115044848191418,0.4412134856138076,0.433792477125603,0.23524700559617817,
 0.12552200591691143,0.35680606654667857,0.8576118229525385,
 0.3560612751232478,0.8415774743539696,0.8596624042979403,0.9751568902164259,
 0.9947978915543382,0.9300263755662801,0.10165945821058553,0.3199587386589994,
 0.05170190544937203,0.7597938457627761,0.8262163569507674,
 0.17459748057654167,0.0700704121131992,0.17939149155460665,
 0.18024568329863366,0.3722150391764171,0.8249445715236565,0.5038545896888211,
 0.0818902121565821,0.2693598068393588,0.4835829760279057,
 0.0011743115924206515,0.8401788685969239,0.5003132312738366,
 0.6369962897518513,0.8942692966495726,0.5275116558394739,0.5924868545607792,
 0.6817377587651355,0.15680414062232073,0.06968950070860247,
 0.4934221845038872,0.7742292942258827,0.8551227354315042,
 0.012976586961549352,0.27661224815272156,0.8192448354248252,0.82383710906472,
 0.06617637030564372,0.5338440321139236,0.7008694498303223,0.3588863385407144,
 0.39467722660256643,0.43966024276849724,0.4037330028573245,0.991516528961316,
 0.4024244269340763,0.06350083506237791,0.10760336280880256,
 0.4543486984296544,0.5977632424496441,0.073030935952767,0.9634143954644645,
 0.4519331178590338,0.07721595684046056,0.1099204602805618,
 0.25457354698686485,0.922344783537717,0.4253198857788665,0.15800050293282664,
 0.13236609715706105,0.8026796877737131,0.3157129180537851,0.1290965793575516,
 0.13575452058943271,0.5503831272022892,0.3645445943336837,
 0.44847954907030885,0.14430607625204983,0.9723581767880259,0.726367055477611,
 0.718035599147186,0.4938050948597674,0.8762616158143579,0.7689480733167331,
 0.7989999781314573,0.28902376942619057,0.3598725251623853,
 0.28650749493733674,0.414878361487902,0.6755176629962332,0.37967134504450506,
 0.7888881630587298,0.4121539091212374,0.17521309351117575,0.7608056450486004,
 0.9413124430032435,0.8633059263319681,0.24661008358230496,
 0.058163151446835926,0.6124731560003449,0.34715919939231465,
 0.33974681375059834,0.028546429594389355,0.042856389370950376,
 0.5220122545066147,0.9406127247631584,0.0489217650206174,0.4639214609738609,
 0.2018236090801574,0.8064470960746594,0.03118121066094437,0.8285365589787769,
 0.22309931221896662,0.5288449554568333,0.9386518012317113,0.517359583168966,
 0.7310208298499405,0.3527451347074827,0.6869309316368319,0.6111708765696152,
 0.9157359824397981,0.7877732159237283,0.8332578737892197,0.8219174670579227,
 0.1677379541510371,0.6436999366861137,0.022324200862875898,
 0.8314914718637839,0.7170254639533358,0.5209375708206918,0.8886453746344329,
 0.7881804576319847,0.15814125216575947,0.6357328540005496,0.4750522019476221,
 0.5350563113628954,0.7806134734333737,0.5999511875508934,0.3948713390461882,
 0.9089797103329113,0.777575506788643,0.026579918481199893,0.9081724901411847,
 0.8066766220761278,0.9749772544871953,0.07168356899632922,0.9103013064512262,
 0.2657098637264884,0.47228043172540146,0.4629705932900441,
 0.15190793650516243,0.30159305400352143,0.9916413067205995,0.798651099819263,
 0.39098587126969875,0.7477391044953381,0.5600924480469716,0.461615802906562,
 0.6915187376927144,0.2349887214813272,0.318309880921847,0.5152510631886426,
 0.7327734390648242,0.9462394058431798,0.06464620710788638,0.6861439426282759,
 0.17207028173160477,0.3568369598076907,0.2361160209092843,0.5176249712434706,
 0.14401833228821248,0.8711932269572444,0.29250494654975967,
 7.437881755931741e-4,0.5342612092085363,0.9215140619213064,
 0.4051217256780225,0.8683627562015329,0.15800747025901307,0.5850369989312202,
 0.7060782102825987,0.6065641977447838,0.8379762880730511,0.43024490176610597,
 0.34216193737254463,0.9948715572467985,0.33755728751454706,
 0.3739478894894913,0.028518854416699546,0.6025770737597815,
 0.7553683553459583,0.15336813410916683,0.7135192931280699,0.7499633178425614,
 0.3324525421031992,0.33934331521187844,0.7899961316089275,0.6284069742431161,
 0.2289486943313631,0.7280646025788996,0.867578403159511,0.3980087898014345,
 0.17226588227030937,0.5198960393059419,0.06125496468918756,
 0.7044743957344776,0.18167818577095174,0.8306067487010694,0.985808182415805,
 0.36380655439021437,0.706425275749711,0.49023886854804666,0.9746601454329997,
 0.4771161140330392,0.9553741196334674,0.9480698325504613,0.06210354468396062,
 0.9590718985054035,0.5776641911043747,0.6066310552209103,0.9775913535004634,
 0.544575683369815,0.7711445124348659,0.23222057995082923,0.0652333995791482,
 0.06742837376639221,0.11114610021171822,0.17842356181053143,
 0.9087265276453222,0.45744142296755874,0.7754877101849529,
 0.10934115111394827,0.44274899111195665,0.8857111247525066,
 0.49090109008590455,0.5235177738175966,0.8609371002277377,0.9009240796242909,
 0.46367223029287097,0.10962030311501691,0.17808872574905033,
 0.3853100059921716,0.5586531632602905,0.09509457960775136,
 0.24289598666073497,0.7108314663785784,0.8042500398020611,0.6806785542574477,
 0.2931146652365413,0.5865534089986019,0.8348399631349046,0.08970541398779963,
 0.6370652232485109,0.49470833310480145,0.7474240551166573,0.8500547738052073,
 0.9275757248410111,0.13806095657897077,0.09787387571600781,
 0.5046582893627339,0.5896615335383889,0.2119389423343473,0.6203790155219661,
 0.3167174575070484,0.3623410607626457,0.27695724143134426,0.3809116792671987,
 0.11061417158485831,0.9929047243876554,0.3751919379003772,
 0.040281368350884805,0.9671171046525062,0.3915369420389424,
 0.2879000283885237,0.4526391360483736,0.08291319620762505,
 0.02563497771789902,0.2876288947145177,0.913217878839931,0.7730502294482157,
 0.43537191310346457,0.9791209749889546,0.5451709335062477,0.6881188688527266,
 0.007938455131312305,0.23379068072931064,0.2415312448728516,
 0.6972602593459732,0.3251851224098363,0.5777284622854832,0.7579720922287324,
 0.4921429463654059,0.8062295241989936,0.9141228471829953,0.9906725748177434,
 0.5281804599928961,0.23130378024956688,0.7487269521480463,0.8131742439923828,
 0.8753082933458214,0.10129941921036756,0.7773330324214618,0.6570401648810046,
 0.7318896358728995,0.5085183907816702,0.5576522455949047,0.8929032679257937,
 0.4729701829818178,0.6493556984132787,0.3645148075555297,0.8294712506417029,
 0.4421194287248963,0.5595063954156034,0.44768754887924567,0.6202046340847895,
 0.866669163701675,0.4688566966439258,0.5592600423264501,0.16602589552736535,
 0.8982020184519854,0.5779780256683027,0.7204418426979126,
 0.023621593997117274,0.054580171553494705,0.5149095859522906,
 0.48161806064976465,0.6686049767296959,0.2987832902491393,0.5097323734242407,
 0.733484958168705,0.118409291829636,0.15908272721239225,0.9618399298391144,
 0.7437320126057669,0.8533098640764529,0.6682775381733017,0.10712967059370582,
 0.5609642388740353,0.27026530747796695,0.8576886467625355,0.5568425261875767,
 0.49643892654583055,0.1855093411641141,0.15285012818128352,
 0.22863290327179076,0.7624535996357354,0.7216833534434429,0.8705574030071421,
 0.6222802233852462,0.8343717530718402,0.2740318353069743,0.5731593077892583,
 0.1498062957985512,0.4146995660457059,0.5348224698959183,0.7020286017223748,
 0.5980354590446446,0.624009556149588,0.8876449260845216,0.907936310022535,
 0.5039115663123059,0.09157149970546774,0.9315283644272974,0.9957065842403859,
 0.767723910664076,0.018929684148369375,0.4146330361677595,0.6809121844336741,
 0.4343382141583012,0.043205787599436385,0.574394327596755,
 0.07960827114945335,0.2907134789907694,0.8384132356704903,0.5131038278393686,
 0.17628229034409437,0.6706161150549734,0.9516961634647768,0.9550429830640952,
 0.04159269054111747,0.7541015975453942,0.7170284494119762,0.1912658706182817,
 0.14142090475871738,0.42530416721562014,0.4704594958334659,
 0.17043362818730612,0.29317154021924463,0.5052711172606329,
 0.07134767764361971,0.6677749407663373,0.4179294976641429,0.2868013442793155,
 0.6136887316102241,0.48144560135833225,0.23257774152327648,
 0.13781576409727125,0.9808888947175555,0.679721139387295,0.2961371898830105,
 0.43906563503652973,0.3571769261199994,0.4742906179917341,0.5444702117204235,
 0.09117271277045647,0.8421864844181403,0.10774839408043202,
 0.4339587020362943,0.7113050059570003,0.41565250955912614,0.3148772285616467,
 0.7198115805101081,0.413967843814662,0.5228841451568286,0.7721500341999982,
 0.7964478919573172,0.219201527056626,0.8720323445906648,0.681037118734606,
 0.9360293216128673,0.28079984526563406,0.8876613007124883,
 0.19911633618816982,0.5768145174250577,0.6142591065582363,0.3020445019159197,
 0.4604991172688423,0.45784966387794235,0.9693781159365564,0.6721340824490845,
 0.5502443638599183,0.3736901917528864,0.8461471836585299,
 0.004812632804057526,0.5888646465498661,0.12703211325943475,
 0.8738808091762864,0.5254056532756376,0.9019396021762256,0.44061681395295216,
 0.45698723357365933,0.40622413112201594,0.06880494553048166,
 0.13790710887312285,0.2103195778435032,0.7294081166311122,0.6611496902309171,
 0.9735245737854615,0.4623092619036294,0.30799308908573464,
 0.26207860559879925,0.6493833674552018,0.14542229332502887,
 0.7261404581024643,0.019094484377937593,0.5303518664935516,0.872798544333152,
 0.1936975543386158,0.8075241834477334,0.27809955538110953,0.6346778110693652,
 0.21523727713359653,0.48275976348357375,0.6587247289073559,
 0.005776535280120343,0.3480650319096229,0.2294019931000525,
 0.24157510023526108,0.9994748678188564,0.9437978607039642,0.834611277037456,
 0.6631025700191782,0.09730442964590902,0.09539487673546698,
 0.0028503205098288076,0.1021573527962727,0.588037605724997,
 0.005246420523823625,0.27704758149229314,0.9470771367869918,
 0.287057046938481,0.6195274986966035,0.24553360455633133,0.9598984698995708,
 0.7048086403187812,0.6873658371627532,0.0477589187224885,0.31680468857475486,
 0.5234482868386042,0.7961086950614388,0.9360656286566712,0.9989614340650035,
 0.05884103899128312,0.40023664633846234,0.9592435470764538,
 0.3663585066157019,0.11853582934096463,0.9647083600828644,0.3753234124364855,
 0.5353526350674814,0.4044327626892863,0.2212252364830961,0.15171867935724836,
 0.6762831229489643,0.3862706193123848,0.8497799149143542,0.9505764309049111,
 0.5290235334298531,0.640900097782514,0.7947369241037139,0.3828779379773688,
 0.7233557907811234,0.6135294042412213,0.9762864601051223,0.11329421000834561,
 0.01864818934591783,0.02975306444363157,0.3220971176543851,
 0.3111118019357987,0.6454944499998951,0.49368962265090477,
 0.03613875809935585,0.856454366998431,0.46388911026740676,0.7230234184182724,
 0.4223137003911236,0.07257853350875998,0.737809547241791,0.47557866142604355,
 0.2819477716811165,0.6921318078535542,0.6039852533662674,0.42592811559254296,
 0.6506990012177776,0.9674491550034139,0.6009412264288639,0.6533866191693353,
 0.1338107241571952,0.8165166548273559,0.747033328169596,0.5399805700231474,
 0.7177714634966171,0.7493276610633322,0.5920522176808368,0.3524230538660298,
 0.43624601624386683,0.9517526856408711,0.01628432647711542,
 0.0809327869583426,0.23967864477154677,0.12669863902081502,
 0.38122281705615535,0.0049236954851624715,0.01667456271935286,
 0.3950580622245654,0.8887445079146626,0.3827334186243003,0.7803044521764115,
 0.852013344554643,0.522032672611999,0.6801933253336107,0.09397341087537281,
 0.7456541224394192,0.15785704300660264,0.36031104936935066,
 0.06422374051440927,0.1912838025369865,0.12812273691205056,
 0.7709869610894124,0.8762460300496967,0.6550089027492916,0.19773717379575306,
 0.9465361451245943,0.7347468088867922,0.9723212246126147,0.9056156503358131,
 0.27884347458154624,0.7819250612595031,0.6324148917085542,0.5348080922644269,
 0.29589782416579025,0.3061559701259391,0.23305570344167525,
 0.2901350111596991,0.9249439067106717,0.9096153090181668,0.7900900759875829,
 0.5652221445831841,0.9089244379543777,0.05801997391157376,0.7179966983699608,
 0.8022401559733985,0.8268239965888919,0.8503387368527708,0.687144780035583,
 0.8467608096833839,0.250658959203369,0.8019309187078316,0.5439820718330349,
 0.12086553291784397,0.04093680783139342,0.29790532991457475,
 0.8767960349095552,0.6850279184936657,0.09719494773852688,
 0.07532546054655909,0.6873710281616974,0.6913501063544909,0.2746290497908229,
 0.6290365052208085,0.013706608800971631,0.3764227398588026,
 0.5139881203373478,0.9799943037630203,0.9175095885071248,0.8394928250314498,
 0.8174530719327746,0.620241442231125,0.26812067630998915,0.10248398466465192,
 0.9439191299101225,0.8022155161940295,0.4141326519087276,0.0745822557295559,
 0.8807027607270521,0.8610058254156842,0.23370232450863038,0.6504907379695821,
 0.0580889713417303,0.9654064164019185,0.3184392107497406,0.9884606493147367,
 0.10429970894731033,0.46133375785503017,0.3313190316200746,
 0.8953639087422169,0.029909021477097286,0.045609580339693534,
 0.3843455786130372,0.5100372036595544,0.7233945218771244,
 0.023398204037640813,0.15501141604158164,0.08431885505675707,
 0.020275297510836565,0.36124121314928825,0.3135318905577642,
 0.7167329759711579,0.05779896926785821,0.06505399485969365,
 0.8766303580117417,0.3605878025653717,0.9678778654755867,0.3750213873845336,
 0.8040268592794635,0.25187951883612447,0.321988741851873,0.7495653328212256,
 0.4958865779471032,0.09253575779538636,0.8529038576916208,0.7528157149737115,
 0.6432305587289269,0.6802281617593181,0.10584301664002715,
 0.33120436829968025,0.7657585036424891,0.17643817458257793,
 0.4339188415589811,0.5914740602772588,0.9250166454106971,0.8285886022446287,
 0.18522447847478019,0.2636848918694893,0.7527478817939441,0.3355973006092735,
 0.4399470789982364,0.8719327206437599,0.4660212390502304,0.16295010692364198,
 0.6384077346110585,0.8307262965127954,0.33432289124223113,0.7419788669769594,
 0.7266455792853015,0.5863412704944589,0.03501258094782167,0.2708434823850291,
 0.2805604094603157,0.5007784481696442,0.12932860604622687,0.5969170810184156,
 0.6783429736619927,0.2071904705908275,0.8643029403136413,0.3783898736371294,
 0.6869487640755099,0.5187682691669722,0.11581871274852096,0.7171520399144872,
 0.4145593662755721,0.03857397147351427,0.20117561041817555,
 0.2419966606807339,0.009646840737757456,0.7406512245210979,
 0.10154150022118469,0.6224283528141884,0.5210524614082382,0.6998110847220245,
 0.6550707067994539,0.6334356307822535,0.010041173613703602,
 0.8996371722339971,0.168520403447587,0.13223150688802254,0.2621436280289544,
 0.9061047492995432,0.578869120081326,0.9491212067061223,0.8924783662993804,
 0.33869376558360975,0.9890584258068754,0.9023908862717064,
 0.08742379903446118,0.7110218814071324,0.09458117473801209,
 0.3559421195160003,0.28060529907657505,0.15870057927275505,0.616700793735351,
 0.42507188681526664,0.7690298940224957,0.9630511107926907,0.969045294874882,
 0.9416766820905287,0.5167906601455309,0.9158095227435314,0.4595104980575416,
 0.7949864032277723,0.8309802779957655,0.9597054296234447,0.653788842927637,
 0.4589229090253155,0.8077137742496369,0.05358257808535272,
 0.09354384686640116,0.37477673620781216,0.13771229650354977,
 0.3811319584174935,0.7382203453034631,0.14499474572441728,0.9855272025649661,
 0.22655880679646956,0.18196616341246907,0.8490308171379125,
 0.06454986718075029,0.5112774806283613,0.9789225447377855,0.6035290828515869,
 0.9113063811809309,0.9792785708204079,0.9574856880738254,0.4487943436121443,
 0.8155895639038034,0.8692280727686066,0.3265819043939512,0.28584330002192515,
 0.6045251089051374,0.6415909946148878,0.09668283036098302,0.9641421351772068,
 0.7563434121607229,0.7617454305572043,0.3130664034944738,0.2594151436342481,
 0.39673858839583376,0.04876523757882012,0.4608974296707513,0.962657042250242,
 0.7454018860249141,0.6723459209264441,0.8010024044290812,0.24967509779342834,
 0.561872125889505,0.4835508076766597,0.861059953716922,0.31365658091191617,
 0.3365949894616851,0.8385179257222668,0.36512369705803227,0.9382798930703119,
 0.6028743808678564,0.6723607600633382,0.13504666137418753,0.9721991856868619,
 0.6345690870165308,0.12574327207949576,0.1463768214272242,0.4260019426324394,
 0.25558110380405796,0.7869300295690642,0.702155013081923,0.39078786583575564,
 0.05915663592544762,0.04739497637561396,0.7978030163250736,
 0.24177971052581892,0.9063360378291562,0.2576429021639355,
 0.02966657927284133,0.4508089865967584,0.007354666621046557,
 0.5874229951719885,0.7473643357696519,0.7989584672695105,0.9907004594137809,
 0.13540477707546306,0.7447618197705091,0.5666216441666879,0.7886250164905787,
 0.5814825930535383,0.7229399334341257,0.9359061637834118,0.3020741272538915,
 0.35073234386452856,0.23529091260285584,0.5287877376391299,
 0.28620447128933135,0.004086465819607477,0.5792569388508828,
 0.35907674368265874,0.8891089454787116,0.9361452928814787,
 0.35031765276831395,0.5477812482968457,0.17123897383773323,
 0.46529398442175895,0.8376216047547422,0.7427154935497127,0.577294602788649,
 0.2787809300432652,0.8214767505679488,0.9020971093088154,0.26468686227288807,
 0.3447088459773786,0.13792938542726385,0.954210120509206,0.9182018302667244,
 0.5701628400734345,0.9227846323370982,0.8820419436485295,0.8497425795139641,
 0.3058400090400205,0.23152092394116808,0.13779210334676395,
 0.5685546714087173,0.2761772974089487,0.9742743882940456,0.2815367408562496,
 0.8405035078033335,0.6371500912375707,0.20692906267851918,
 0.11883724711960897,0.05481040445231533,0.634250708024278,0.7417327002111249,
 0.5771134869533989,0.15468874932498455,0.9261840725151187,
 0.08533809077216825,0.5574973988863219,0.4132016877303374,
 0.32068698430506404,0.9288549789630645,0.7377227045689602,0.8451716169672988,
 0.13234906799894597,0.2085969258581739,0.453798235971124,0.97767875288874,
 0.964878231481458,0.15916497594482437,0.3353806798107404,0.8708379355230772,
 0.10632081740999333,0.13751070504599605,0.036272025695680776,
 0.5690118895262233,0.7877149149012971,0.3502084938486324,
 0.035016336588104546,0.7464096212649387,0.2000753883562736,
 0.7074225678595268,0.17345301325033402,0.8006172567490506,
 0.19263875584100465,0.5439942987605955,0.37955781574834824,
 0.3611001993127029,0.04402617430205846,7.734269460225596e-4,
 0.146837653160948,0.8060090356316627,0.15389273307750195,0.32280301777329223,
 0.6089502540698343,0.3546378496408249,0.2711301250703462,
 0.001405164035088724,0.38348992741702337,0.22600370813598258,
 0.016273229881785256,0.4036814241354838,0.31226597746263973,
 0.9724985139761387,0.5699537372739709,0.6864987924067514,0.5098605410202437,
 0.04939244457024716,0.5390643779133293,0.9972030961595238,0.9789625886314264,
 0.6267369192443999,0.42033817804043316,0.522986411263787,0.05427493710181808,
 0.8508574042432218,0.15166317360491455,0.1241515802551173,0.8507223942894757,
 0.226848199304866,0.3660392126088037,0.21318372934631324,0.4058107541942514,
 0.192919565574723,0.7889634992785042,0.6115677038400463,0.29020769644447175,
 0.1607482354696258,0.031241270671360843,0.17998138195188684,
 0.4313743017872538,0.5507577595097726,0.43037327283089843,
 0.055509399954311256,0.18399845760040745,0.17924067471773597,
 0.2523976942872945,0.5654602047907616,0.1673987639317559,0.8413894974089188,
 0.6972683043953908,0.11365446965317028,0.513533081279736,0.7856655843554108,
 0.16769070445088186,0.20131007900675002,0.749924255949248,0.7827687235219235,
 0.12809500425484488,0.4836050033265241,0.20298485438642588,
 0.24398329190373635,0.9140079820720384,0.7857503203980591,0.6782475386757134,
 0.16714392016175472,0.6871878352056263,0.3220204346419271,0.3731229157704855,
 0.3831888813399771,0.506359098231719,0.45715386627024324,
 0.041733752121921985,0.058642176178998806,0.6532049946738693,
 0.9185345128469111,0.4535598555441287,0.8218810685481291,0.42736101662353176,
 0.15405618988513115,0.6322194704535442,0.9873429762388737,0.7095829693940604,
 0.6854574682167222,0.9263543299701092,0.2773318166462654,0.07887866755009576,
 0.3696713766676535,0.5089966454343777,0.3196720220606172,0.8434987233688624,
 0.8133457305027348,0.12549736814023416,0.2559019632197843,
 0.06455181020299927,0.11230159806054285,0.9984687435569213,
 0.8085980676292782,0.23094500215230074,0.33006156353575467,
 0.6594330179256541,0.10299462021957928,0.9121794574214255,
 0.45497672159089064,0.6527445928724275,0.15172235924017663,
 0.8757303326732339,0.8666892121814488,0.6086717864735331,0.39076858436869033,
 0.24116515476591105,0.7870465036768366,0.4956602347012192,
 0.12436875313561202,0.6635028175264706,0.10698255414320501,0.989052343315071,
 0.6571864850928433,0.1487178656370517,0.1697913411050398,0.6881226095829983,
 0.3629853932734657,0.6641764277635294,0.5951420581129729,0.6719258371938239,
 0.9352548655154775,0.8732399477716415,0.4108973543959962,0.42732184857106503,
 0.4735019657826417,0.5236506147885494,0.3061535794849053,0.37499553965149257,
 0.15319619026642484,0.9200342907062032,0.1044677138354062,0.283042593453551,
 0.8068588025766172,0.7187291058711265,0.2915134358653446,0.16177247994101196,
 0.6385256798047234,0.1732010981986929,0.6004083503349984,0.1666945525943574,
 0.8641955773939172,0.7061699550224363,0.09972889457998346,0.5354269946222228,
 0.63059968560452,0.14866986256595327,0.4739007555585484,0.2052316258426209,
 0.06277569062239863,0.9368037510675296,0.43253790737608877,
 0.4277592118672222,0.1477216957189007,0.705286099797557,0.9540147020115805,
 0.1839662388553237,0.6076072836322131,0.9333863799339608,0.2367186400955129,
 0.008406746250430808,0.045520885570495606,0.5208576103214657,
 0.12535573145396928,0.6073748351890658,0.19014319988631645,
 0.6390656568605287,0.9320624194140927,0.5229166009106051,0.1004084776538603,
 0.4214258482576987,0.05761057947122439,0.03921849560107171,0.898085860137712,
 0.33751693865091326,0.7513859662766119,0.9475779772354879,0.7149346256530424,
 0.5316402570630421,0.3262314115785464,0.6835348855162822,0.381235780779128,
 0.5785501097258673,0.32173875065115987,0.835152101140087,0.5191639820250491,
 0.6322387981882738,0.35124534105254746,0.999506367519573,0.7818793836797895,
 0.0762901066117766,0.3967838822612697,0.8120373721260251,0.6388744538574187,
 0.873963603075698,0.21848885535747486,0.35985658399088927,0.562576637717846,
 0.9317099159766548,0.7830036057278856,0.3819811018672038,0.5022332926750237,
 0.6705378667210125,0.6293302225976187,0.7912628187662412,0.6533689729847578,
 0.6481948990571821,0.07353818611833085,0.6847292839192767,0.539680406400367,
 0.07923789051432806,0.5231359308467501,0.46588267646815307,
 0.4014931039556695,0.37635366142547744,0.9779881901473246,0.1737257707321127,
 0.466994022556797,0.7820231955238276,0.30222203189259034,0.7520627882272672,
 0.9557030320453339,0.183182502746297,0.8514707201904578,0.21584931898617887,
 0.5093913801180947,0.8110502536218296,0.9581111748441917,0.9571001749704514,
 0.5978200171747113,0.8556874693092568,0.3045242598364267,
 0.026200611674247365,0.9162696101879633,0.5480977465263167,
 0.8488915640290416,0.3312435202415056,0.5788002474797656,0.2535539479942124,
 0.21674395883788167,0.6980958914556429,0.29851391793448456,
 0.4301566391637647,0.6985134555833195,0.721807565985154,0.8335956303130665,
 0.4840785771299809,0.6441547839482282,0.26224067894075354,
 0.34836509153964723,0.48409535272618065,0.0673226950639284,0.284440979687432,
 0.1428851483636816,0.2693454198524,0.3484412140442634,0.3545439357012816,
 0.45871399457224005,0.9206654969571936,0.832837241539093,0.651724756167531,
 0.39490351899876563,0.8823184773144271,0.06293346578173642,
 0.8643909041444244,0.9652537402990923,0.22167927319059522,
 0.39745644883854647,0.15603769149605173,0.465655962251048,
 0.08343615355596401,0.3261240167099858,0.5815841781359714,0.8555817491814546,
 0.40034994403274904,0.996843244868903,0.14577410977038674,
 0.010241773284583089,0.26078615480440703,0.694309575529927,
 0.5194610560130231,0.2480377715208193,0.6970067491803268,0.6010283358246601,
 0.2613208635981784,0.7780129724992714,0.8072419991157275,0.2803899457881326,
 0.34300207434967067,0.8814219409969938,0.37958799064110016,
 0.31765799304428577,0.5429217463116616,0.6680165217952638,0.2298415447509421,
 0.025003812898984457,0.7123591715755305,0.9300316978440683,
 0.5259492466167366,0.9173848107188183,0.17705594624410348,0.680246137684774,
 0.39943710224182705,0.5342987536106393,0.4294386116002422,0.2888358892234698,
 0.43290162093573115,0.3337985704547214,0.024095908869504856,
 0.4697631238508012,0.2316722359419493,0.981719141924913,0.26414854632957363,
 0.4612363842623429,0.34652035263103154,0.7890865621361008,0.9863645978868596,
 0.8267805659217713,0.10465980859277546,0.03509120746951033,
 0.36355794595716306,0.906057819699656,0.12442996994670774,0.4502960035919784,
 0.5118720323712078,0.7504851836910312,0.2648169433798466,0.3779297223792588,
 0.08610344926704472,0.17172398689661073,0.8702298394367393,
 0.25082246952048426,0.01681431613141482,0.8417060066461096,
 0.8640447046242142,0.6632548306247859,0.5476225858282995,0.4177790894511555,
 0.384054551406649,0.0024055169396260956,0.028170861504080436,
 0.560912912841379,0.7472224127720947,0.8409999034145113,0.14501557644152585,
 0.6130200047761017,0.7621168530516668,0.07153182918983825,
 0.08495105194780128,0.12950555203105885,0.9552050318693062,0.643988950234963,
 0.038727836129908044,0.6785650584433367,0.2926206129430369,
 0.9772993227132059,0.012338983061387698,0.1938337309954452,
 0.1389575124666862,0.8068927003341184,0.4728519232119891,0.9813444729607532,
 0.5394676751746811,0.8709050371199041,0.27442193878012333,0.5157707562297917,
 0.6827585527080491,0.8788742117155142,0.11348929010498177,
 0.20261215301807445,0.5745756901840192,0.09262615404262853,
 0.8630633570576217,0.6560593452194665,0.88224759554804,0.9513950912517279,
 0.6760662519976316,0.8683427973422996,0.3985088841795674,0.9669698598435296,
 0.7033665253633855,0.2788676285566771,0.9288892042437554,0.2692821717176125,
 0.9794938914607227,0.8327664476954282,0.7081070796458441,0.11957791299951015,
 0.8478811291872753,0.5018336823980091,0.3185939959101808,0.6527527001868761,
 0.0945522352593795,0.44683906327558065,0.8788713307130944,0.4032909543758951,
 0.46493612958059316,0.027255773439284203,0.2692913481971255,
 0.6366496302604141,0.6386199702284268,0.4394447257473927,0.4045890934372528,
 0.22239434484566478,0.8535967735324628,0.4258006216672946,0.6912331474407073,
 0.43456561794627113,0.908521667977315,0.7497990973417312,0.18218918535027107,
 0.5898111561761004,0.8986905799226048,0.5479342647381825,0.14739930895609454,
 0.367244829157619,0.3951224253975366,0.07677141009784638,0.14642546006299506,
 0.27260189801583934,0.8521795247684936,0.27067358030630906,
 0.7310458160167957,0.7845302314704908,0.6615401115438743,0.2559008298773794,
 0.1329927937163603,0.276533425781345,0.7739755630310584,0.5672000112999179,
 0.7885411821167874,0.5128496588278156,0.3443129849545036,0.8311484495950973,
 0.010651174789221507,0.27881249688904985,0.6076570670186714,
 0.785357593246207,0.5533177764696537,0.3050129585668211,0.7463953434153243,
 0.5196324877056916,0.11352934022267636,0.30329893675187125,
 0.44558227792332183,0.8844723836463259,0.3941881336428227,
 0.20292942913808265,0.2409461506047208,0.03725877074956352,
 0.19646362640296267,0.7614049403642734,0.5787070021899076,0.821629783541377,
 0.4376145218616623,0.6142145023534593,0.21403804703412477,0.9796604822746761,
 0.5840560184811486,0.6367083303138045,0.04779306574123876,
 0.08904191549168394,0.6365402497845887,0.20250808882638716,
 0.1066958078127821,0.005726872070901834,0.23288649302457842,
 0.7113009833634498,0.4443811393281625,0.4943238406473872,0.5597747655925858,
 0.6018876013859754,0.9398279132254441,0.5344702363973222,0.8823408146522764,
 0.6398267101833269,0.8473270540113922,0.06368026188938603,0.6611801454018656,
 0.3539632964164784,0.5556742755730589,0.46730448802301683,0.6851492260068283,
 0.9238706059737323,0.2776716486904607,0.22528345400941985,0.8945243772872105,
 0.8128183075380883,0.6275008708686141,0.8197195132091919,0.595849820629184,
 0.5934647250710946,0.6645470092467165,0.9193976886654434,0.291283230689947,
 0.8961226500681331,0.8301734719969595,0.296295625341657,0.9701336737051294,
 0.7046297492228256,0.9564321571068813,0.9572764279066146,0.03017710702829357,
 0.10863353921583119,0.9324343758094529,0.2942414559016572,0.787430223278841,
 0.5507281772905955,0.1789758327520714,0.23631922614754208,0.371823228712003,
 0.2094127084969204,0.3546725587010876,0.4479637908656099,0.9868692398990678,
 0.4993069097428109,0.7087223641197461,0.014597065274140064,
 0.008611311022349355,0.888470382172903,0.9015233182035773,0.6023000052253735,
 0.7854118427139828,0.2828500603555262,0.07757353668757705,0.5609102182114493,
 0.7711373477215725,0.4374309028429776,0.6032908733078507,0.5714097561077992,
 0.8753229113810747,0.4226934624541583,0.2717425132821265,0.7987559703823858,
 0.7318667660011224,0.5031148933441281,0.09902027217836462,0.9634731852316631,
 0.29947163336935345,0.38339755934719977,0.6727302641010824,
 0.1334377053757495,0.5547378330418358,0.06276365357853098,0.9456463046404435,
 0.8265646320234884,0.204383746588363,0.18139137774708058,0.793655053933842,
 0.843991457150483,0.12412612506068232,0.37383035727186553,
 0.11595465404734084,0.1022192496929123,0.2642475987301912,0.9727034433171737,
 0.9410901294921347,0.4235091185743499,0.5907366149548805,0.3939318936527063,
 0.532189547737596,0.8324323507192196,0.6189856573485393,0.7511026710485823,
 0.26657707944295705,0.5216814731375983,0.8480521637367703,0.6994811825222086,
 0.983681135461502,0.6241263694959523,0.6387058880281238,0.07454939326697319,
 0.2959726951373629,0.8629263668830143,0.5562534134782733,0.9187835150997332,
 0.49341305567575644,0.5370618718532232,0.32627687033285246,
 0.2703685856374557,0.06689645778441489,0.3381315687792942,
 0.11840716516550698,0.028653959322072264,0.18550066130647036,
 0.80879293145645,0.735617283372545,0.8117248100422985,0.6116927002422171,
 0.1389030030278443,0.4607005853754026,0.22126838250073066,
 0.09042939319590237,0.8708387804795629,0.6965535439869558,
 0.04899430562785534,0.5688372476047998,0.46683070520403824,
 0.7720078336941062,0.3883904775750876,0.3298093624145636,0.4734061596005089,
 0.7983479505869433,0.0032299581283661283,0.5317168509105227,
 0.3305393640103148,0.08169543168925131,0.26870335331633954,
 0.4676792591745116,0.6448041407614168,0.8046449620487501,0.6458859651356966,
 0.27888097093449793,0.8370743771078594,0.6692636693068413,
 0.35690710730294606,0.8823160929369005,0.6443447913907127,0.9463028252139366,
 0.12069839093135082,0.6976004678368296,0.5931802166603269,0.5763358205889114,
 0.9420582401966019,0.008675541769931905,0.1546591107216695,
 0.5791790494455897,0.21400202695728587,0.5542877386672219,0.726350919023287,
 0.04938120669965407,0.42543304434949536,0.2429896429556625,
 0.7539883703175643,0.4802872390489964,0.4466198739480378,0.5987237119791876,
 0.7312104877316234,0.7121224032057987,0.36579660097591904,
 0.14646369158546402,0.9105574453883215,0.1875075466128584,0.8319308191197823,
 0.09349862062023029,0.3374191948449642,0.4556358704998157,0.6870088915343476,
 0.2625042723951575,0.4112138408345358,0.8944844145547057,0.9952736261217725,
 0.21326795309371482,0.587462476667191,0.34985545420376396,
 0.08256364311848174,0.42842591137984476,0.3843580560850154,
 0.8821546543516826,0.5149677876304579,0.21916583634659814,
 0.45134836289669167,0.46499593440183484,0.17800267151816795,
 0.9508125507758711,0.39454297007104266,0.9747197599747719,
 0.17312318678145244,0.17262322536639585,0.47313544444398636,
 0.8931944310766224,0.2917661874515054,0.9887910296615349,0.1112099658503336,
 0.485675825890866,0.3028865822600746,0.39278114366304084,0.20320440035733833,
 0.1960492868353978,0.317018193716346,0.008685281773467834,
 0.44223325694353777,0.43515277425770726,0.9138609483790621,
 0.38902741034022015,0.3076682562809666,0.9016625918456457,
 0.16615580384352424,0.6165056550696711,0.7421982693829023,
 0.15323953110898658,0.34588564942058675,0.23390066820733102,
 0.21614429688120773,0.16501372802965875,0.07893157497770109,
 0.7866517461505547,0.2274110464821012,0.6527075965843995,0.6345836572494882,
 0.38121100222352133,0.3670556702061146,0.3780298114748387,0.60246093536943,
 0.16479362408461506,0.032284685985135675,0.819595900553987,
 0.1245561923281493,0.8183832360765971,0.6618365023715664,0.9076984560425532,
 0.3992308771834452,0.5132809772337714,0.5088341125257096,0.1820933950921585,
 0.5863277475814925,0.7068740603087529,0.09797041587796074,0.4055383433557431,
 0.7533090753338425,0.9479109226280731,0.24569241009251352,0.9812452506356572,
 0.8915689091479042,0.7726956809342255,0.7616639035685963,0.5874585646159484,
 0.1774586116995961,0.8627677570237727,0.5118445627106857,0.1057607668594922,
 0.7107004676027756,0.08376609471804741,0.9060234765756139,0.6229382022637946,
 0.47702353603990455,0.37377260791269074,0.6948887635604675,
 0.1879508146369664,0.39605694923106083,0.22996296080760315,
 0.9361180339002854,0.8648592267141375,0.2350938329278357,0.5893388570407505,
 0.96106900083562,0.8801403541206642,0.4737504105592927,0.022881810746266273,
 0.7008917736936853,0.6041371943599948,0.6009181440297496,0.4700746625419262,
 0.56686681584613,0.7984905940706063,0.6214514111844214,0.9482046217367532,
 0.16691038846147266,0.42844294093568935,0.8641474150983343,
 0.2568466656421009,0.14087022447925812,0.9099009928461391,0.5435548640789954,
 0.0421969025311042,0.5140667441792349,0.8705411706511493,0.8868182930791817,
 0.3780499135456178,0.6872689250750708,0.673917653183445,0.4214914256181208,
 0.07142240207022987,0.4991416749890286,0.8819757133039804,0.4037652767412744,
 0.69133235239312,0.0730586932048225,0.3556984022806339,0.25949052309042575,
 0.08956197618416595,0.19556924593789438,0.09279063241799657,
 0.4189927057572228,0.6012880014088491,0.40252196022337217,0.4573773295954635,
 0.3603557642370079,0.9603911168879938,0.5583917547872679,0.5360147890809873,
 0.5435618303910753,0.36205705846686864,0.8175756831483862,0.487088433406362,
 0.43324547583166995,0.43747026604644146,0.06970544059748363,
 0.06730089754904556,0.65071501647645,0.3602047535633882,0.3395336387562421,
 0.5246084505374163,0.25338091014521313,0.03575532653534064,
 0.30106979118338184,0.6356432987409646,0.1739016880298776,0.7958632817640752,
 0.9890646611826918,0.18210951261233355,0.16862733001801922,
 0.8642527143254128,0.9978722367449457,0.4012730550116237,0.30809236090513226,
 0.3032471314900097,0.14644268621407952,0.3036986796436917,0.9141979065706999,
 0.05270286105874222,0.9796477997708469,0.5053295280570003,0.7050675267226741,
 0.653341072989414,0.38585920678030616,0.30573227469621234,0.8329201527831032,
 0.14039165965200495,0.47068956169942167,0.059162170457332186,
 0.10346758180945526,0.12115977713419057,0.24967460231862104,
 0.01051996323092741,0.960101028971563,0.4950955063972353,0.9452186071251989,
 0.8360731668171617,0.8918784354938075,0.6863216189310293,0.02488965534115728,
 0.11369929699953696,0.6943591601228712,0.9445372268038947,0.9042439951362644,
 0.5235616690465738,0.857340390124433,0.796883580642938,0.44299528983829806,
 0.4949833152849369,0.8010698363202069,0.7253460452742534,0.4220559806957752,
 0.4878100437097581,0.9020738749429651,0.153857428688008,0.10481563395567528,
 0.15296546962355662,0.782079137492037,0.6552177795410115,0.2247763195844562,
 0.06353386750589451,0.587554446671346,0.6968850228964762,0.6538904786582443,
 0.9316460155060085,0.6970818309219582,0.4477305119579461,0.6882944267180733,
 0.8095017927583857,0.5685307243244861,0.49568660356050187,0.8643464768752518,
 0.8472924548716434,0.8432189792418097,0.26535992357825033,0.6345052984645879,
 0.9796949241904923,0.7431258730268363,0.40685172635408373,0.6469514281114378,
 0.46037844815710216,0.06865730342866527,0.5915433718511153,
 0.3117879042030882,0.4337950983892759,0.4545136547334587,0.4375056836389958,
 0.7763581235564517,0.5761005278996786,0.17366576587290627,0.9223368947774674,
 0.37538921342930043,0.27150854665172974,0.7693764038156368,
 0.8526078800237875,0.08249714231655647,0.1788194046422824,0.3338645222390988,
 0.2882771544669549,0.7903768031060894,0.8988106892838398,0.5618858835256287,
 0.8171569658611693,0.5288999559152839,0.17320955729584053,
 0.45926790714065546,0.46321762516713316,0.7252381413372856,
 0.5413535208189485,0.6598972236126966,0.4688982891190916,0.4507386658341822,
 0.47648938456689305,0.6462151416711964,0.2809959962607208,
 0.43833191058839405,0.1495082397770211,0.25844301821006277,
 0.3217515511525615,0.47388669766548286,0.07265102955587499,
 0.36409612823685344,0.46922392533818036,0.2408970365199753,
 0.8574551370952002,0.38254794841287953,0.6716767603039019,
 0.46305068388849835,0.36524410712158706,0.7663256973324224,
 0.7892633413345118,0.6559393283122489,0.7478334468768436,0.4765136052581136,
 0.6411106448440883,0.6228893902977429,0.2534332482430802,0.34340377005054634,
 0.6322130315215516,0.6415836984186731,0.4745420850861104,0.41097488836325213,
 0.5070891001486788,0.9353878185875192,0.0970180435275323,0.9726843476519479,
 0.6948652398319152,0.7871915505288073,0.16301239342315865,
 0.33834176832520446,0.8189715397708756,0.4976009894193766,0.6183523722745387,
 0.6532520196546964,0.07165200369250468,0.460109534262067,0.21995507420309834,
 0.2539860619714981,0.19783612551252494,0.23879439596515561,
 0.6804160297596328,0.8029245626781478,0.293950241584641,0.5710315930689764,
 0.052189972265375695,0.6824196095403607,0.0285153629603071,
 0.9885864774832642,0.5177931141846814,0.42320182165946374,
 0.26050365422546573,0.017042154077246408,0.028571893531560866,
 0.4222363303065948,0.6970432062020239,0.9339403064085373,0.11503753584587606,
 0.8543641757251215,0.03749158857675383,0.48164940380785315,
 0.13641661946625983,0.5262880248187826,0.5262439274951216,0.5550497031181307,
 0.5299418585309144,0.6028832654044689,0.673982340741879,0.8279621359489667,
 0.7718669757284077,0.7101635253093954,0.11142631338716924,0.7361858316179636,
 0.8587649365771006,0.7743078546422927,0.6446414583774938,0.31933513564282057,
 0.37440931127938115,0.15186586316117145,0.10415992994522849,
 0.6491566252602269,0.05596951170427589,0.6361746098461678,0.2274359836366343,
 0.3053056006760254,0.8162219481880107,0.8818933548204845,0.6987261456133044,
 0.6905361976556234,0.527219344345536,0.450234249143027,0.8362155072082468,
 0.298149285382875,0.05080491284130639,0.7829092486437152,0.3411787891605046,
 0.6178591801034647,0.26404220099207,0.7598550189593439,0.0474104031169853,
 0.38374503142498007,0.9582754684394703,0.21014070489765202,
 0.5667199432179313,0.9080531862655722,0.4572635618877766,0.0757267647965002,
 0.33783704767279654,0.7593000224361988,0.11576053086524418,0.865830454574794,
 0.15396250469677175,0.9553178968704288,0.21287231959762365,
 0.5925115791845295,0.1997637372421298,0.48449764068026646,0.3983407005530776,
 0.6416926121567177,0.3503208549590845,0.029255399172410712,0.907336977258013,
 0.020418228176119158,0.9086623911008058,0.6346138154320438,
 0.8804933858210724,0.04017549640055318,0.7474319515430433,0.3883613520434348,
 0.7721686310344497,0.7490630319216486,0.8852521845987802,0.7822227175992477,
 0.4054533994560584,0.8551511018183896,0.3445936382407856,0.7728378703394424,
 0.9748187593579407,0.6937844198907768,0.3231910327958747,0.8463213706636153,
 0.6880300701573174,0.019566914105132005,0.23318482863031686,
 0.0866206781012856,0.4471091688467056,0.2683319528509307,0.09205731264996242,
 0.3097475539514021,0.7277467879525161,0.7500851547506608,0.40879712948726843,
 0.38448152926311907,0.9864781934223267,0.3501529609821179,0.8205892568274991,
 0.5136233775072268,0.22611798964564844,0.3092037487060034,
 0.41165417926775527,0.8206470158677835,0.7917833747570264,0.8726126415857413,
 0.34117827974530923,0.15854834754496272,0.7176304113114461,
 0.03905805678840368,0.05969935338645915,0.8305541907660321,
 0.30856676012197704,0.2771829498314088,0.8964452479626897,0.8101178334845098,
 0.35052452217391483,0.10118311532365254,0.6773515333850403,
 0.7694222692028189,0.5947195864123781,0.5779964126536793,0.3376797062357757,
 0.19365111362527154,0.5718123894167852,0.03519037962274385,0.521384733818797,
 0.5552294007109635,0.9795225590161897,0.9763564969101971,0.8303239043221435,
 0.3393833882419609,0.4026774905836574,0.9306675510418572,0.32065848412833786,
 0.5653066263328799,0.4627530383415829,0.2827588531111155,0.23550728089449857,
 0.004394293389184956,0.2550136864294261,0.20298118288471223,
 0.3988054967091431,0.5357850785471492,0.06932704290380343,0.8247684135246601,
 0.8507628475895215,0.5840346569423875,0.26923331731609945,
 0.30645669058630853,0.8401561848897058,0.10625703518931773,
 0.46098203981370145,0.5423689605417992,0.40433037462220955,
 0.11405116218421973,0.5049489909769262,0.49174810708410877,
 0.7477854851436846,0.7265992684266243,0.963045318851048,0.3207041358527535,
 0.15969090834782307,0.5466010596627968,0.818419922445294,0.6112291153433757,
 0.7637813099581018,0.2490629476628231,0.45566962737064154,0.013614610249018,
 0.7418202135727787,0.2436932042722073,0.9208270650250282,0.574619304825398,
 0.8730744357185123,0.47867590638436164,0.9381795610795477,
 0.03601975432702964,0.47554342358136437,0.5769638588078977,
 0.2795849573324093,0.7782355004018363,0.909191985293919,0.6629654192249327,
 0.8947263993792837,0.9809034701003778,0.27108477874168124,0.1964805348214178,
 0.6909754723272714,0.8507179305056587,0.6281948281837282,0.7186355686251908,
 0.3008642290075645,0.530594786641349,0.3759276819786789,0.011981790295111705,
 0.057778492272276605,0.8376797489099269,0.8076823699677355,
 0.30920202074923375,0.8028144156385071,0.31246739695454673,
 0.8315814042754703,0.5686171929595097,0.9979394392363872,0.4814447639640105,
 0.837776255156827,0.8003239493274145,0.7092792846596898,0.648716551153758,
 0.8060743651252259,0.5240555892002703,0.39460599361836857,
 0.23323790604952377,0.9099011456745215,0.9431828202286878,0.8603482465802244,
 0.0985988358580816,0.028133616402914785,0.1986892933166191,
 0.013131912791978118,0.07389050840946965,0.31728617696108996,
 0.3849627825704175,0.986584940836388,0.18954865588697278,0.06558680871166012,
 0.4496365661846582,0.7416077049309039,0.6504041888086236,0.8263140659577513,
 0.5371421991341191,0.3371000968391701,0.25155800249939053,0.4296357866132191,
 0.7660816078790805,0.8523753889355938,0.6901989116086299,0.4850505052473594,
 0.6716012927544036,0.9806732190291532,0.09750753799783851,0.5723061848205967,
 0.09478471119906628,0.6046998224508732,0.8422524293123326,0.4070524591703568,
 0.15525491995454332,0.8512360976914961,0.13015304028221153,
 0.7402012130967462,0.22087582997281396,0.4180779507734157,
 0.36380466346914764,0.9739308001994388,0.24225672566111822,
 0.6896463062837296,0.8782622549252892,0.8624486110728437,0.5655241942493623,
 0.1223833593445065,0.5625184793320183,0.020513347602807763,
 0.19615605646426382,0.2484264978845504,0.7854046504473657,0.5272440213547435,
 0.4369552020361003,0.4965111909714932,0.4267080024340536,0.357014446043054,
 0.6753150887918911,0.8097731356203298,0.5942847807283177,0.09938712919817805,
 0.6474272976955223,0.8971462638475713,0.7054208064577437,0.04827627034804194,
 0.11806336741421986,0.1872219196623206,0.14420793695833678,
 0.6222087147345713,0.636364008130778,0.4848204607743626,0.33597315196223854,
 0.43389483349480673,0.6273441794061227,0.41151573042593836,
 0.9454463045938558,0.7574152180276321,0.5661493578215497,0.6691992482286175,
 0.18132387080329593,0.6073986928834001,0.9899223618180373,0.8740543644959088,
 0.11105603568677858,0.11076529903940147,0.44149951347892835,
 0.33131882620852493,0.12892942853139688,0.27265871248520035,
 0.6140098769664858,0.7801172238224576,0.4918351147429316,0.05765778336088612,
 0.513247404150286,0.6356202132338897,0.13105853114821064,0.6391131132720684,
 0.24157800745916513,0.5705224985185998,0.631610980043965,0.5744423288155155,
 0.08447717994961013,0.5006013238521876,0.08192464300415492,
 0.21890396020449077,0.12754650341989016,0.9315245976059047,
 0.8235395902717837,0.489941543833707,0.027634214554216663,0.8311135986941041,
 0.3420241273616904,0.06733273366936521,0.6371928017896455,0.5040520895220352,
 0.04459893618462707,0.9821573082827346,0.22514478245542868,
 0.28509713755226174,0.8961078130110803,0.6355887547258574,
 0.03657998336629309,0.8645323550147679,0.9714215131661896,
 0.43050344168563903,0.2501432343373262,0.6364258957853786,0.8516571078084645,
 0.25101079501914925,0.046615157784364425,0.9993278005466983,
 0.2770752501628644,0.010746138421974138,0.6642814370722658,
 0.8019391047284272,0.15741129630746942,0.024480382885408325,
 0.4824250404809828,0.2893129247693593,0.9073974375613572,0.21400858731182648,
 0.8684228303501194,0.2762507772408417,0.15832560185756006,
 0.04565739923570811,0.40490900633677274,0.3900522514264708,
 0.22374245090868194,0.7050976461035419,0.34512830422356566,0.563542503780131,
 0.4881004052443395,0.006293334951953278,0.37476819277508167,
 0.012241537118518386,0.9554785101448394,0.3971308990187076,
 0.1347021562417694,0.8054398875483764,0.8005816968029756,0.6379259608713537,
 0.165760861968265,0.7373620067578196,0.012025982525472156,
 0.17873935286055753,0.4531867062410666,0.3982808568729298,0.6593913796947178,
 0.4060898791822257,0.3014939755175102,0.03213829370545662,
 0.25235594598405353,0.652502140237169,0.41896561621370276,0.6750281725891825,
 0.745963875892707,0.611515927956793,0.04629601680342521,0.08818051792596049,
 0.7569736255913699,0.001164626500536059,0.7992671640599441,
 0.1082037946275397,0.4113241330988865,0.4388372438029907,0.7376745325054932,
 0.14426811366386116,0.6043798371218451,0.9436697318557531,0.4534535876237139,
 0.6726827402716917,0.8301467576753776,0.10494427483388014,0.7258643580261582,
 0.7924152417140978,0.18218692939158565,0.205288868102542,0.5081358438697862,
 0.20720956271198265,0.5360307797747015,0.8018070483312467,0.5725709100557141,
 0.2558303999532172,0.9639581010464923,0.5251935833046332,0.7096461841831005,
 0.197957357494572,0.6208857364518721,0.3239195100950507,0.18586799281120403,
 0.7237462620563694,0.43699583807413966,0.2497288671796919,0.3076510283830971,
 0.8565312035574081,0.4283428759210741,0.5631838415804167,
 0.007400999419921672,0.460262884932894,0.11247955433779067,
 0.46226302427339605,0.5660672490973613,0.8487488585235505,
 0.09971749925260898,0.7334898304050331,0.903262675921323,0.9965152711427842,
 0.011973834215686319,0.2633552941304198,0.23719945722012947,
 0.11112764608167014,0.8839344785605742,0.23712051439030557,
 0.3247518567269091,0.02804474857203787,0.3756433776623772,0.4682347349020929,
 0.27344735738467907,0.23743701677678186,0.13457446269490114,
 0.017064908873358542,0.6823639536338986,0.3084308093348306,
 0.7291992099300045,0.8202314546568025,0.5503606755225315,0.20885158589310993,
 0.6583184245770229,0.5337222943788652,0.022352943441773188,
 0.4702105520352431,0.28697130063000187,0.9907689648760322,
 0.12726070477073748,0.9211403768932003,0.09854235578670856,
 0.9344735319462932,0.968024131287443,0.7185978677073269,0.6708338665297048,
 0.054727318312812634,0.33579119883766584,0.9196227671118067,
 0.137427050325878,0.5224967567568242,0.9192163603183383,0.5283841696138565,
 0.19286588845356178,0.6853410166166779,0.9503832091275655,0.9194500633706036,
 0.7641205644027098,0.10111018437935337,0.134814905245911,0.8959677667816754,
 0.9158636399176056,0.5537691393652114,0.5162807684276537,0.12781167085023226,
 0.7280027257460021,0.7678062939331158,0.39461989146084797,0.6088948837378539,
 0.9705417871866315,0.3033724246535282,0.3206301453203282,0.18069728137130658,
 0.7772079235540581,0.024015873491800455,0.5436636882339795,
 0.07211376770698552,0.09922088137170237,0.9551087471361875,
 0.5168212413807134,0.49596260883806276,0.264285004256011,0.8024425627215827,
 0.7364666074088895,0.1032728209052094,0.352207956247488,0.36356368979673437,
 0.49551858346922484,0.0381286766768788,0.6336390528915394,0.6994711248286362,
 0.8562654794242702,0.4475799525040304,0.611604561196764,0.8953213051475826,
 0.5361249496947251,0.8985471330223093,0.8502411331342322,0.2616862417740078,
 0.09058045881712462,0.1082575421889076,0.1106796953552508,0.1557721058416559,
 0.21403128083828327,0.7506385112475129,0.7410189793950374,0.6936756213628009,
 0.5845708298793372,0.6383166829890069,0.8197204514604509,0.8269903913760026,
 0.24736360556830417,0.5740414063620523,0.2891434470050026,0.9003225556086012,
 0.7601390721452567,0.16694138930262392,0.4403698071953922,
 0.15155391286628217,0.029229932618811794,0.8347650499657759,
 0.5958475431989255,0.451000541667012,0.6192981892054965,0.6721480498848451,
 0.48964746356883293,0.9400204631996115,0.30004233162913974,
 0.35525557209607206,0.8297721576229405,0.24347623487053327,
 0.39526349183395415,0.5608014466182585,0.5174699457406526,0.678539053686781,
 0.9918106693647296,0.22470090243195295,0.39081942040787543,
 0.1831980089656856,0.38167154305498696,0.7408543416096826,0.7616730223634435,
 0.34279772787877094,0.744935545413918,0.9375406034278153,0.8770290983820945,
 0.8777552290065465,0.2955252104164884,0.8733139261648649,0.40578846074992136,
 0.8623291849279758,0.6010109319694692,0.1880824386249368,0.22209443514674976,
 0.9433946338113938,0.12545122741480919,0.15354787028336436,
 0.5769251782363217,0.930159187845331,0.7292728127254425,0.47787805351585844,
 0.37761339333353294,0.26115969962841634,0.6058994977881619,
 0.24600120453769347,0.6162674812562762,0.15895104736593524,
 0.6809019759921618,0.8415313982422117,0.8976531440225713,0.7602667184034024,
 0.1939953281710699,0.9362290339347776,0.19810022645542524,
 0.27838784411720485,0.27057835860922197,0.6020814913840175,
 0.5136911191961998,0.3197654245992595,0.5697744178357849,0.3760725901481915,
 0.11408337824977821,0.8544263173755969,0.35052758155876906,0.590797067610122,
 0.5158950902246607,0.5563693933575304,0.806895327103839,0.17179339595763843,
 0.48884830468756946,0.34475711374765716,0.9823915353381021,
 0.20717546541759524,0.4978459889463446,0.7823941302023957,
 0.07580653656923653,0.631414549624571,0.39625878516534874,0.5202026228521868,
 0.149801603045011,0.6781804339102967,0.5301655521275277,0.3767597049590191,
 0.46811060988176834,0.8580698729894136,0.8266650420516545,0.8408810820269794,
 0.06930812286046051,0.3379547419837523,0.09680186987956763,
 0.03573853849544861,0.4521153178243694,0.34032954136997107,
 0.7491769131189637,0.9125493670226188,0.26306579709374756,0.4144653932779604,
 0.6732644877426983,0.9294953312400921,0.879205457227081,0.9376957189089403,
 0.07596563141108792,0.13327331740266224,0.3804176826890687,
 0.41616343468128636,0.5856386796055061,0.08335539012479465,
 0.9962704733847265,0.7928998597522794,0.536561758940905,0.8015438609468379,
 0.18514021788297508,0.9708810243001964,0.49089205840799477,
 0.9862696543286021,0.051747199440714375,0.8008942041703953,
 0.5491828947213337,0.458379978129836,0.24141693701449973,0.29703990471586184,
 0.049543559979471796,0.7808155610415183,0.4965018118486545,
 0.4362011676231963,0.46844001015585834,0.9582674868756297,0.8640533498529334,
 0.3109543999944505,0.7762002220921884,0.8610245824064767,0.5168996618538817,
 0.361544660607659,0.3185912140937939,0.2964975884199673,0.42963918992336636,
 0.4456147867180885,0.4039412379967451,0.027502303723107024,0.406225314608052,
 0.8951645644829908,0.2956955635257137,0.43153558482968446,0.9414901632340005,
 0.3081400024913068,0.20414463849050668,0.0793457590059633,0.2945410709774934,
 0.5089096148666152,0.2894605458316195,0.5016805682325055,0.09520849002843168,
 0.08726480676376747,0.47253862552078385,0.5958386231635476,
 0.17902426262735416,0.5491864979604066,0.6188758051993183,0.3354267531734877,
 0.017638976101378767,0.976549172521892,0.04474882737316621,
 0.36207978965515786,0.2785847008810294,0.7759459923954781,0.5871295213296328,
 0.47313052014442714,0.6635843454222535,0.16867240306025388,
 0.0022150049058413668,0.7605822799812958,0.22574825319206215,
 0.9859598862124238,0.10815671473964217,0.6479735463604092,0.660900830362974,
 0.008617160614623809,0.8192064217519726,0.7120208424746769,
 0.5370515522027357,0.9130950372404163,0.5726635176360603,0.3305987527916706,
 0.5523926196979458,0.3974908511865993,0.9314986580632953,0.6039665180797654,
 0.4937958728978482,0.38088959139206924,0.5071820679558264,
 0.22452666273361133,0.5097984114546393,0.23466487388705204,
 0.5875460699289901,0.39739092573799384,0.1452921142724829,0.7106420633624729,
 0.5590289212832026,0.3332562341702978,0.013187583736749442,
 0.4672894829576322,0.0015617487939099117,0.16892594559734186,
 0.10866377270646832,0.7776502731970248,0.6409534058529971,0.3669628941951919,
 0.41904828242373726,0.3106682557191899,0.7575288097642661,
 0.41672844277659815,0.6003357186785289,0.338057983675326,0.10302518674613936,
 0.8132737956414019,0.6624925775866075,0.04164352352425449,0.735203227849029,
 0.42695903875832264,0.8277472215076915,0.13896104619157978,
 0.8653531225655562,0.7526033584650003,0.6182144228359855,0.9960347878241497,
 0.2735325255018164,0.32064996234069976,0.18617235511525032,
 0.7370103148340288,0.04746554729708152,0.5354936225510514,0.182530688069223,
 0.2802216556105932,0.05943906195679771,0.49507481193538827,
 0.8290653143875846,0.32762039760377204,0.4680212955943994,
 0.46017397330670584,0.04487450483379285,0.8683325608398872,
 0.1501787203587751,0.34068022390981734,0.7479189539803539,0.7318587043867472,
 0.7541207515455274,0.40580193643362716,0.7207901889118273,0.7414880708962517,
 0.3080079348526268,0.9697110739268506,0.5856654816960537,0.28907589126453115,
 0.570250908085817,0.601188437672532,0.1062308028731005,0.8822199396689788,
 0.4318742912969262,0.313081300444799,0.39738319661031196,0.9217201343533363,
 0.4886748632651744,0.26704574781793566,0.8339581165982114,0.828194327910658,
 0.3694961852533768,0.844786100923456,0.8414686313463414,0.17451532710900375,
 0.15659587097574412,0.7989924964741549,0.05014707453238376,
 0.2537849266904433,0.8743680839315233,0.574681948406943,0.8923410238232221,
 0.017706233479699574,0.7694010147602393,0.5608738115853565,
 0.2595158643878781,0.8612051523654833,0.8066360716912415,0.4784425337042493,
 0.6953674269266229,0.6173237778234549,0.5162526151861292,0.5954118323354198,
 0.6730978023761836,0.6229790339109088,0.8329527303119407,0.6511126395930139,
 0.39759115010787294,0.7698782001257437,0.5879200602634651,
 0.11376926051456016,0.8775858112532304,0.48154824570585675,
 0.48148870069688576,0.7050733350237127,0.26613417297084907,
 0.7031069483973943,0.049507607748543125,0.5908421561595394,
 0.9130776956072809,0.12217974022284639,0.17509987800149351,
 0.8806459830795796,0.259307348361847,0.19719431272717958,0.043893577118572,
 0.5155085099610207,0.8080688961737619,0.09946854626399548,0.5370729957316502,
 0.4677171353029659,0.5803925571201722,0.30651484709520005,0.4440417258136885,
 0.5536922824237486,0.059871225666630834,0.4927587963075387,
 0.3253762191550885,0.21977372130881234,0.5992139013542438,0.561454420163936,
 0.887967082967824,0.00930863799688475,0.8682458052833397,0.8685834005558797,
 0.6519251014266376,0.4745285057129798,0.6477839744331718,0.860530503050574,
 0.2887527510433264,0.3886845525402247,0.3023815896560773,0.7722574924519794,
 0.9359135490604689,0.33829381331585173,0.19376352304941835,
 0.043538042651206976,0.6369986603081008,0.6213006090635754,
 0.28439063542847365,0.8828262733991168,0.00613974272118023,
 0.31760724048046274,0.3765839904589543,0.3748052611113545,
 0.46943656810365497,0.24833041257735855,0.6662017033873003,
 0.5857617031485577,0.19194633880359024,0.03301821801848115,0.61872901840359,
 0.8561243684323364,0.06614970404069676,0.23004027290134366,
 0.35638741286635134,0.693414284974027,0.7022993020293877,0.1902463981710305,
 0.8461727358970066,0.04033563569525245,0.40926266420423985,
 0.4854374568076174,0.09962108040331119,0.6588803429764213,0.5109785722773365,
 0.11781302617722134,0.6597156350031592,0.3374485819317654,0.7700700794774743,
 0.5569327040831434,0.00608649610217693,0.3608031532916992,
 0.06182415543102049,0.2460560555432153,0.17131768332985242,0.978723455685886,
 0.39639854785251516,0.09325292146751862,0.8685830537089868,
 0.4591379406673253,0.1783478791772889,0.15992559201248635,
 0.23175440584327767,0.34081789109041305,0.2902090245982608,
 0.2244400729794389,0.8578663935429008,0.9541873474875032,0.6264527696443057,
 0.5475451336270295,0.9944132704161737,0.8279986089831735,0.9576901193638231,
 0.6215136510904491,0.9573223898121216,0.032575436590391416,
 0.26402254685997006,0.618232091659168,0.7252504549296683,0.4647453611076501,
 0.583540208091405,0.735296585866329,0.6120013977431766,0.9067708881585648,
 0.14443676105233327,0.7796444930897865,0.2346415962152566,
 0.07407193218830022,0.7038964431181146,0.808175081360222,0.5376005983930641,
 0.44651579616156467,0.9893517607864396,0.23467803167256596,
 0.44061086086927304,0.14330758483462291,0.3968628720272982,
 0.5900813354474758,0.5356159981981958,0.323874329960278,0.5613171734477,
 0.9614923233818924,0.37287652669656257,0.2701945698805288,0.8478738484900786,
 0.9741638787255549,0.5986417137009091,0.7334550343633613,0.2743023611636414,
 0.19590386947354022,0.44108569979293666,0.2327334162047302,
 0.46027767359186056,0.2627139272254594,0.292551876500602,0.7929789327057345,
 0.10021169283750786,0.046088721835053414,0.3010148705106187,
 0.3857386665151865,0.01973574475887563,0.05099648764738429,
 0.3296981197386524,0.15970487152580792,0.9383631868210778,0.5177744750961555,
 0.8766653187325812,0.42152662331626234,0.7255773622403024,0.8743606457503842,
 0.6641815822771558,0.44248652569384983,0.7305572894656082,0.5050821755674022,
 0.356380120572683,0.7746172355888348,0.4775545001731495,0.46210449318525204,
 0.9779406367961787,0.33872056644160375,0.45940743295410125,
 0.7938513543718068,0.5235930602164122,0.506292950405655,0.3903677746745884,
 0.9085597231782803,0.5464411134679945,0.10561546523437615,0.9859508626494715,
 0.4515768984659452,0.6143694998121603,0.13542126246622477,
 0.016707161871935328,0.39611368352273524,0.5916331859695256,
 0.358525418455534,0.404693464326505,0.7197479256317837,0.48276685489651716,
 0.9165338274802116,0.6232887879034466,0.26872452143679704,
 0.45985286065071795,0.6395945400263663,0.16670397649111368,
 0.2900220557702514,0.42059286302390597,0.8942022508342506,0.6091009121844286,
 0.8083309709626026,0.08854704522536705,0.9471880235812403,0.312185823966044,
 0.8133644336695547,0.5658785966230403,0.6759544372602168,0.7983380884949585,
 0.9799249642451278,0.005496583619612694,0.053676722457176695,
 0.295036747951563,0.5962303035101765,0.6117246700476364,0.35190735420693087,
 0.9883076899361798,0.5340214236387362,0.6893873995209362,0.9469171369534629,
 0.44408722269360557,0.5896845991052275,0.4514725067911518,0.5294620617480339,
 0.7089724536944353,0.7494987258089779,0.4536719133143191,0.8233040600713294,
 0.5308275035681546,0.16764034734212996,0.6969191797744867,0.905954368344575,
 0.2966072704199634,0.8878563884028838,0.6224275104417645,0.1227713751293108,
 0.3727259952343819,0.5129705011199492,0.45939727348687875,
 0.11150468663272184,0.6614370321673503,0.802233111981936,0.8775939803458269,
 0.013325734997858496,0.45235056150755215,0.6464603955648581,
 0.21271467770602392,0.19749691600106722,0.6873897742062043,
 0.1671411367175193,0.7655963738612772,0.9588028873866524,
 0.019326991301034657,0.11801969399921963,0.520180222944982,
 0.4264880388122768,0.6133178531021378,0.6406617701765606,0.7758299700697302,
 0.9361665885532671,0.5960448985484303,0.0011946733972173496,
 0.38644883777541905,0.053850913082407414,0.1774862899890537,
 0.5283622087463615,0.9561833641187746,0.004182043389615031,
 0.06052412961179321,0.49308612894714465,0.728774786302654,0.9087937835734324,
 0.1750759016648007,0.27352179077242467,0.8132587794951394,
 0.13474798571683055,0.8997009753801501,0.6386740291600902,0.6277096918590428,
 0.08073360327195511,0.28962413895487826,0.7389414692089264,
 0.5035496410052326,0.7027899508252062,0.5015870472595654,0.9663310858819876,
 0.16687283343146275,0.13959274920547504,0.9620515708310472,
 0.15666054337823665,0.8503218675030849,0.9285545596420774,0.9671411178871768,
 0.6759592079161663,0.957780742798789,0.26719137862280795,
 0.007419676087883964,0.4871124036818657,0.4466342159468558,
 0.47475435033426505,0.029163373118747615,0.7961868296041843,
 0.3470566662932787,0.06563268673534783,0.6380391228987966,
 0.35951431288286617,0.04504919128409479,0.006818425433795849,
 0.13479352198591732,0.4969254310047285,0.6455515198819124,0.6022857240947282,
 0.7138000876068963,0.8100619972553188,0.915187565182137,0.7245979875873926,
 0.7153619236992188,0.8941956151400415,0.06763629246155689,0.1754868961817161,
 0.3790111254615923,0.5578738297879131,0.024687860694269315,
 0.26293398396834833,0.6506952840793908,0.01778579551166748,
 0.11225466913823223,0.8926791417039606,0.7494934757081337,0.5281119513435935,
 0.03931552123114401,0.38127844673715583,0.22325403097049046,
 0.05106482736899065,0.08510194232710422,0.9139486785139701,
 0.5793193887679785,0.6448921222203001,0.68893780171591,0.7887032923517006,
 0.8139956217066786,0.12743295778952823,0.7958200908611761,0.2981738245920661,
 0.2214246904859627,0.33068402953402976,0.6161271775469767,
 0.01734195907753111,0.386072130917823,0.09937879116248904,0.5756433787586565,
 0.4503832051238138,0.23031406581199632,0.8450403160838528,0.5587457877978114,
 0.5671334559584387,0.3174716188742086,0.12494089026735455,0.6461438567088248,
 0.18377260564147513,0.008684373600902684,0.3113929309348994,
 0.24422277574743068,0.3114578833530036,0.7666155781289741,0.7566371591849829,
 0.02553224451879732,0.12913087910328358,0.18334562296393664,
 0.46870727496678055,0.805473607900574,0.15420109925254522,0.5087293301556455,
 0.6086642166306004,0.5252611591032939,0.9360365711143178,0.07187415435950961,
 0.6471422657017644,0.4768297132942735,0.03208997762196841,
 0.07472049620944254,0.33034421400880787,0.9361987431566277,
 0.35600697384493785,0.46024758667276655,0.12551126000131363,
 0.7046640883672585,0.08017050809617343,0.11301742098587919,
 0.1527470580385546,0.08068360202985503,0.7757476290036407,0.6950332602445882,
 0.27839770126696706,0.9460029821359018,0.9094166460943927,0.5982607149634159,
 0.9103579285836984,0.40756520010105124,0.18737508270603476,
 0.5818917510553735,0.5521651818589798,0.29937078618724067,0.4245819354643321,
 0.310800394407851,0.028045632062930537,0.6543757781913089,0.3353046975723477,
 0.5028391771139604,0.22728669937206702,0.08036163790555328,0.940567060143593,
 0.6950917910034515,0.7818051369571347,0.9401172580249573,0.20094088050669312,
 0.03825346266391416,0.7535935645866036,0.8128658909114558,0.614969041385428,
 0.7428165198321867,0.9219324508351602,0.755477525349676,0.8522325626438446,
 0.34736130566637735,0.636130072589591,0.21688179677728614,0.2636005528055787,
 0.5575484242806444,0.1491028882182922,0.4880279453301102,0.02066948269247959,
 0.9157949323282119,0.5174567731811183,0.9903112079656508,0.7095266695467156,
 0.5881906616508289,0.2754997155981844,0.661661665189492,0.3964234369766956,
 0.18515243827061245,0.2671763191860046,0.45714078172148853,
 0.7183589694230418,0.7791109552863646,0.2219065621351921,0.4663124012014631,
 0.7563367818751839,0.021301834623291604,0.5134279769403962,
 0.2729998157799298,0.5117003475208507,0.603164554620698,0.310590406403668,
 0.48995672665895706,0.09912398736475359,0.8634773393102132,
 0.7042162033690521,0.34748743321723474,0.44579174173545044,
 0.37344008122602923,0.5400321742623911,0.7453499656985433,0.9416525730986425,
 0.6974123193065391,0.23472262221474283,0.03911968541011579,
 0.35969036662482257,0.4116980234266101,0.08364372328918968,
 0.4957190471827826,0.5908678443821005,0.3504727009676756,0.6884637922558177,
 0.7978002655635678,0.20126083617785406,0.9966992283359497,0.5232167769058335,
 0.12851164458929198,0.4983483706152464,0.037293710925249246,
 0.05097136706859684,0.019356786832338102,0.16756508449427776,
 0.5298195142580829,0.11314681616933642,0.8571501093746398,0.5281669358139235,
 0.7239738962263124,0.14212887257125328,0.38284351412641804,
 0.6531709162461744,0.301153371297743,0.7952071841197309,0.35378547982262754,
 0.712776476349338,0.327139006041253,0.40184247533425,0.8060906558183333,
 0.06298608237505265,0.2929556872683454,0.49050553037935396,
 0.5568792420080736,0.08307030075866562,0.4547117781811867,0.9012927776265529,
 0.9230442819862957,0.7990057639876907,0.21801139419171478,0.665750124586157,
 0.10799171228482751,0.9545899431456855,0.10665087281272845,0.981972787416171,
 0.9526265534358797,0.022205940346121777,0.2131145220455557,
 0.6676733079232806,0.7624439874763105,0.32553190265239884,0.2075040004671027,
 0.9187381268696198,0.4894493640998284,0.13390973578891707,0.7163580787934177,
 0.17588241093826906,0.865248051829817,0.4519643972081202,0.970706051640858,
 0.12297371285150249,0.8967267058122427,0.7124227415264545,0.4498467509711307,
 0.9427022046172799,0.7292379722765939,0.37934299702297225,0.3562015798220802,
 0.7040569920018843,0.09312180486124721,0.2458271836974909,0.6752743661181946,
 0.8771642050329536,0.9597553231455327,0.9166109072299126,
 0.040918073625542006,0.10729578652155769,0.9516802902635132,
 0.15208134373910076,0.10265500309197273,0.022157859422693327,
 0.971036710296213,0.2740494737954703,0.7618130287842318,0.07598715053162453,
 0.5716676978990695,0.028536410529639955,0.42899252155081014,
 0.08262248985911236,0.6340300387781715,0.14611737079475973,0.9030384001452,
 0.018491189472831793,0.08431762850234992,0.5956123607491761,
 0.8653629407119977,0.44736961985147583,0.2998924239940779,0.9390187635969386,
 0.3847809166386409,0.5251340339029181,0.7032035044938918,0.5689932486715884,
 0.4639660613397172,0.3987191623829277,0.3516071710653895,0.2980924997429768,
 0.05565389058425119,0.9290581707507115,0.638043462225025,0.49440915442183697,
 0.3200990549120514,0.6073308336112165,0.2458857605800454,0.394041972576444,
 0.7028287283331381,0.8552311211686866,0.8357274902844826,0.7807616646584425,
 0.023680437718133884,0.9210471280909283,0.27276130906017915,
 0.6233156472469218,0.9933952002981605,0.791458812074111,0.14932434489323498,
 0.68303706113559,0.5954504393841611,0.362704313676774,0.6516843385733073,
 0.9336142497108629,0.736386746116321,0.6159252058575488,0.24322345788588817,
 0.5647817361896086,0.699535851784491,0.912444796265789,0.4585100297432154,
 0.15795187951293,0.3562364249590664,0.4659524319362969,0.5380931425501481,
 0.22654549850835315,0.9661935732401834,0.6892631063941995,
 0.09684387588987853,0.3742778252164731,0.7073530262387554,
 0.13116071831164322,0.4281290928367174,0.5547357928848614,0.835655296776062,
 0.7095544631262041,0.37797447239981685,0.2658278454477192,0.9600020294822602,
 0.8945836073669089,0.5714567467767812,0.7207126668614385,0.3093650912427309,
 0.7901555534619304,0.48520689935796546,0.5863678725850497,0.7815714071034174,
 0.34583077780174754,0.7041567486184589,0.8603808966492337,0.808611020354803,
 0.9921527318702918,0.45701651560958545,0.5657764635724908,0.883843130967243,
 0.46139951680437796,0.2458964102201141,0.5773950195185777,0.7645881407203898,
 0.6637604770818758,0.28067947253923475,0.8908444972530521,0.2375490397057154,
 0.2877086287126063,0.1311493634420603,0.45970013356457895,0.5560752295101062,
 0.02734510471097218,0.3186476285179062,0.4305781953208233,0.3019150650362167,
 0.71969312046134,0.21494543548218192,0.5241307386626797,0.8044476744306001,
 0.29859960339355995,0.5026623079611086,0.3145088372923184,0.2510673486190409,
 0.3665516997984092,0.5427576865425527,0.8975928730824146,0.8330574006421564,
 0.7951547219585927,0.7371187963334107,0.8603038997411907,
 0.008322131391848897,0.22871932852969423,0.9825672635080482,
 0.3687907483421825,0.12852313124510517,0.03116629298655149,
 0.37732012286364613,0.10401317851254988,0.051532883673049756,
 0.47463396972134486,0.9889118195845692,0.5902926559386941,0.3923151347011269,
 0.15777954774009895,0.9348141680359571,0.7074522425886048,
 0.39558055807544656,0.1414881853506571,0.4556315542983904,
 0.49755662954773816,0.462367448319957,0.04914239772711637,0.8854421877923455,
 0.5358921106659795,0.4145027726479884,0.7025704667979258,0.46927487067924967,
 0.03607885332841809,0.7816019728504224,0.5795554496915545,0.8149787223655227,
 0.3608916434518381,0.039523041368019984,0.011938264833411871,
 0.9566017286649894,0.8982712134835698,0.46363758988372605,
 0.13914066279138637,0.6381749620481058,0.7970194591395026,0.2602711325952165,
 0.4056537568351619,0.9083798005679266,0.09974320759272715,0.3135334499403579,
 0.46236458147049464,0.36901760578719134,0.8398528576358306,0.545196240773826,
 0.40517183332730067,0.9954102658043523,0.4474130790637165,0.563355375556853,
 0.7500534574707014,0.630611087510808,0.36017972051764013,0.09633069528864979,
 0.03308092299660337,0.008135267848051964,0.07333609806571784,
 0.3017054687705213,0.7886268620226962,0.5955532004531051,0.785116335868313,
 0.11464538235941324,0.5029527244172151,0.8506264243640315,0.9939993615484215,
 0.9347403560518218,0.7032030457926072,0.4865028848241737,0.5980159375104088,
 0.39749430389835916,0.6478382086292626,0.9349086315729306,
 0.15584545281061413,0.6271953024491898,0.09772738832249162,
 0.5923132870415272,0.8729460043803783,0.30306723472960606,
 0.27123175341651207,0.5548095854626519,0.32511181180431825,
 0.7697990955826227,0.3200839157728591,0.04795800077064938,0.6853711683354019,
 0.11224507055027722,0.4503402602965525,0.7141514893253244,0.9951791244456418,
 0.3024690611690716,0.043582054602680276,0.6534032695933925,
 0.19655993298962815,0.44621011637770447,0.6037798644855951,
 0.9467650165207411,0.7361496690556679,0.862241479687584,0.9497135078058109,
 0.07071213056538272,0.07455024005716426,0.25108091151989365,
 0.5487866423036711,0.8011086548004772,0.22765654290751558,
 0.19411953405056492,0.45326917203328,0.6080265747634949,0.9377474745514773,
 0.3889543474409187,0.9357452513987757,0.45811859784455944,0.7296708355557295,
 0.8448847476587483,0.04235370323923515,0.29353237575368596,
 0.3791703429050238,0.10561040611515315,0.9869158992476236,0.5516523530730011,
 0.40624981979108155,0.03433402207493219,0.7815787197028352,
 0.2536075896723795,0.25335053735793456,0.834008052540556,
 0.037114128447903405,0.7071829069769462,0.4061735939016362,
 0.8446501287799332,0.7293512692156269,0.30041704050684326,
 0.45035835474070063,0.7164421234405809,0.09112886693084665,
 0.4750529064313541,0.06885326719495621,0.04910489851616595,
 0.5039211146271274,0.4209668215709129,0.5924877602396443,0.3786099284672275,
 0.3757825537030748,0.6021507701831187,0.7565488186053739,0.5085957936421269,
 0.1925135226844895,0.8479637599268139,0.7418414332459553,0.07708024208977837,
 0.8906573891526025,0.8026706377007752,0.7485549276440627,0.3494735058903239,
 0.2753687583110135,0.4785479749986463,0.6414786415694376,0.4726476717033561,
 0.6989605817528054,0.34891129842805746,0.21436314119538447,
 0.18794951324657216,0.04659464959324877,0.22697945782843287,
 0.7630328966025764,0.054261156989339976,0.1629334458461622,
 0.8186722014092871,0.8153152906124296,0.41681129701093167,
 0.38953810127951116,0.03679139379145202,0.5935440630157807,
 0.6943586343560283,0.2644744253605664,0.09707005686886105,0.8210761839471223,
 0.2976834274113206,0.0903895852921841,0.11305682916876869,
 0.31051578188647844,0.3056265322769419,0.7199825314634165,
 0.06826614856857582,0.1252503218515536,0.7268400937937931,
 0.009784688946820275,0.24130653774685928,0.4687998007516162,
 0.3246557106336624,0.16049744345020123,0.14905081252528907,
 0.6559366770853544,0.12735923936146176,0.45413401886558535,
 0.3661793001441791,0.5330544370923715,0.04998593801783957,0.636094094961333,
 0.2144282511602108,0.06663764403131534,0.9728806868290365,0.9627266879152098,
 0.9799951385449371,0.2610157826868481,0.44525649970641445,
 0.18366558233054575,0.7593108432303171,0.7820836258268612,0.6727361440975934,
 0.14614143505390054,0.5169907125381513,0.5775578892819893,0.3838283776796203,
 0.23088340229850757,0.16938601635218498,0.9430648857997319,
 0.32629398283662725,0.858980374206584,0.10477758228713141,
 0.21106621580300544,0.8448193618780593,0.2574775336606512,0.6666919616156515,
 0.6156918871267003,0.22107894353195734,0.17296176014850229,
 0.9290560775612522,0.6954600150423347,0.01955926361657223,0.6688851633917648,
 0.03074842475307482,0.7956078349836381,0.2015331299054336,0.2306025137414046,
 0.923528090331552,0.3612280951861313,0.054281150729228345,0.507212855030047,
 0.4716352075193865,0.5605863033568668,0.5863209112342657,0.631077195275859,
 0.9999835261049435,0.9002021603189108,0.34698917762494697,
 0.26558723620173297,0.4282145463319056,0.03343414751890006,
 0.06906655507671577,0.8086007289807962,0.5413721277500141,0.6223993120530341,
 0.6170997755591022,0.3719537925204208,0.09435554071264862,0.649312336472458,
 0.99797233728459,0.15152849353068787,0.06366190999712262,0.1921315989983413,
 0.21517896749508825,0.9423251688013897,0.2529014310101759,0.8003054968142136,
 0.5246453480211554,0.21712951839202932,0.5193502211262342,
 0.06762841917873474,0.2212828983040359,0.8055008995590466,
 0.46623617635584613,0.41931732163145197,0.9968531577096823,
 0.6881120394838619,0.26001109122653543,0.6359155999596595,0.5101693419758541,
 0.4687478128759903,0.4887950864859161,0.9147151705600537,0.965258637346401,
 0.9779399434030351,0.6533013480814455,0.5272376669512607,0.3495754471908219,
 0.7195684018170491,0.5641797926213963,0.502956175210181,0.17347734803238768,
 0.1568099166177428,0.8666281686628953,0.8562196328200439,0.8856201542246506,
 0.11655669274582325,0.06593165246135524,0.3266112527126399,
 0.2394898476958075,0.54890626731319,0.34087228268630243,0.26828737236111666,
 0.213935264091748,0.7712712040669307,0.5976692183994672,0.7055106897640772,
 0.24160234741223163,0.33202186790596344,0.6471640175542258,
 0.7433087667767049,0.521760828525481,0.9721730504061581,0.24109408684840927,
 0.7868629392583518,0.33116881326825354,0.41505458743855084,
 0.5092270978108557,0.6059724889459767,0.0418466706152284,0.5690902588745218,
 0.3340401873955019,0.6207330511492888,0.4808507022486108,0.19550366069560854,
 0.9682659805076986,0.740165942836143,0.19260200018332596,0.6059390872272623,
 0.5378595077597362,0.9662769082747245,0.7126835582287805,0.25744818426282834,
 0.6677555020364697,0.5345547439616659,0.6600186877192475,0.9996058413654038,
 0.20373861256929793,0.26782599591823075,0.6321991773153754,
 0.2739673657475201,0.5004445034466921,0.29049856052040324,0.2970285957493687,
 0.2080839434531745,0.24979739434117632,0.07548127138338734,
 0.17635515316574502,0.34653854489858504,0.17836318697017206,
 0.704468758263198,0.7603197530676535,0.011855557280988194,
 0.37844110268179554,0.2727072055966485,0.034471790233499555,
 0.016336170492412205,0.4704237622342381,0.11601095172568698,
 0.24471072611475442,0.32020103281015055,0.3452212956955928,
 0.5069791794084602,0.10530188437112464,0.5233179705263119,0.9228455158391271,
 0.9425996229361071,0.9539632867192686,0.6437114795954333,0.5229209295658305,
 0.45231069154535364,0.41464569967105325,0.5402901944466155,
 0.1519603848666783,0.6662670086583953,0.529729260379017,0.7568727953388688,
 0.8605017010813837,0.0647084256443694,0.058126821982333654,
 0.1315115883243796,0.5342863849241027,0.9994597444655023,0.4735994596264379,
 0.3909419948127275,0.7869161256400831,0.9426360830716901,0.3909707068242878,
 0.34555798741247834,0.6655166562436496,0.6945022495568399,0.5791146212169155,
 0.7880459496444913,0.16113710528048075,0.1596904114132922,
 0.19169583643311872,0.9299255829359427,0.0749829167927134,
 0.39324832242345475,0.04774302306254785,0.8310941695200438,
 0.02433443711356187,0.819825038718692,0.6556442096277063,0.2095187999580752,
 0.5676068396366986,0.4151998631885817,0.34260851216321875,0.3324486723903243,
 0.7432576055929035,0.5354139841675264,0.3687232986406571,0.8936206045898127,
 0.6945273063062714,0.7290832098063075,0.8264344092902047,0.48917511537443215,
 0.6401718422293254,0.9179688767452194,0.22792578130833474,
 0.08939416318992166,0.2644887362104762,0.18032985249033295,
 0.20916996501686147,0.560180298522201,0.8926219554615948,0.4883030124498511,
 0.13382646215737715,0.20862603314244177,0.8591312777592368,
 0.16084477917558282,0.464664858176126,0.24155262592597326,0.6067693295772186,
 0.6399789180213868,0.8514637045647198,0.5317391123386604,0.8473270961860284,
 0.047535879390549596,0.07604463347029344,0.07647910928632429,
 0.22849774668268008,0.6333718338809405,0.12102531534339744,
 0.7861264786475926,0.010533866108650591,0.9498821011697915,
 0.08971268128394538,0.9970031290836627,0.44778456012009893,
 0.9563767906775381,0.2981126771320153,0.2715315091402125,0.23796746884453046,
 0.30238095703604095,0.21588709747843615,0.05794625779853557,
 0.996956381391145,0.3210942155596177,0.762564901890528,0.02190383429160092,
 0.9953986530315255,0.22519754764430222,0.2624999998269306,0.7099054947369224,
 0.17883906494891588,0.45064449651503835,0.6665847910739113,
 0.3047796022697281,0.7823311418320493,0.6353193644277387,0.7750394445618419,
 0.7901201216562529,0.3938379640520634,0.154975473745963,0.8303902619766526,
 0.516621857169348,0.8215610961388917,0.6883575511469515,0.07024296670816865,
 0.2783041462867072,0.7936885309340601,0.5559749637363428,0.3758485428950882,
 0.0797536884884773,0.3627916725190272,0.4212280307385661,0.23377028633607455,
 0.6645793704124902,0.334199565975817,0.6616737434147838,0.05383877109966151,
 0.8582839805772786,0.7613657160704144,0.4391489932220701,0.2830396444429746,
 0.9749378337183309,0.6054467198012112,0.5528565012107133,0.3225018158621098,
 0.21264784361288802,0.37899573783949636,0.8809985254252947,
 0.6329192264200977,0.9088849774068597,0.7137875412088155,0.5217796720741009,
 0.256735171383213,0.99314536418415,0.10599541233229282,0.31481564477379964,
 0.8611386848775083,0.9819718773318198,0.7095740786270887,0.416509551161406,
 0.8963311681717856,0.7734096291394217,0.07828627434812396,
 0.28912508931054615,0.180723406028243,0.11100158638900981,0.8438986240097086,
 0.4326779456973301,0.5813155395261433,0.32471128933137583,
 0.21215280169964956,0.7203497351624405,0.28947560602146927,
 0.7431224794681082,0.6445055482651046,0.3207191692410547,0.15819912818418647,
 0.5857737323830983,0.007301679643862191,0.5281245274367531,
 0.08017147908933253,0.9907431034546881,0.4362178756620503,
 0.27118872927092585,0.15114757862858874,0.49784217567938405,
 0.11359812841509465,0.048787273927166086,0.6489192973072289,
 0.8067559630099717,0.8804322374736545,0.05835785222614942,0.5771521943361319,
 0.0804216384493659,0.08380563584536305,0.25369360076395986,
 0.09901381733527848,0.1033128940241771,0.8752940591873551,0.8862506581793874,
 0.44472084324113037,0.6847582091605721,0.6453198701006708,
 0.053501942715879824,0.46251324646322944,0.22770640891041927,
 0.4440447336407778,0.3355133848506763,0.1694074902237247,0.2667874086225075,
 0.83881790233876,0.14174287059521018,0.9033536481701827,0.5693317550042396,
 0.37159643912339513,0.45168512861268195,0.5143721281644602,
 0.3775833758288719,0.47651690134048885,0.10883172969975896,
 0.6239191315852011,0.8334575435451024,0.8539122081278447,0.13933195984651525,
 0.9340953794751872,0.8207589163486536,0.8443584851118606,0.46985672606835627,
 0.6162471894624708,0.05415049391690252,0.9151549081273693,0.5998463265989205,
 0.25318392135411283,0.48640240496383347,0.3972403318666665,
 0.8570961949621704,0.024872615591299718,0.0845562388051403,
 0.5287677880469661,0.5239286023555921,0.5808028599815137,0.110632827926611,
 0.2892936888597437,0.6575433945850894,0.33950328498044946,0.6003312300059471,
 0.6869227459191335,0.9010865078714061,0.07816885918570615,0.863531686334025,
 0.5719415047824635,0.9994308190033604,0.42982190072361637,0.2120972135413468,
 0.253903314060686,0.2146019704600678,0.2743573813300286,0.9804280289015025,
 0.11631328955329379,0.9069503475597965,0.8232434234944449,
 0.15432781705255683,0.2520106843806177,0.8457974491692877,
 0.020831821019566932,0.013656486980563987,0.6873236850074961,
 0.6156124078224342,0.22408470138159298,0.8941374123043836,
 0.021169225364884303,0.5416460690980593,0.5600102816757423,
 0.8862868506277003,0.8373256989849311,0.34937212849660204,0.9812574830906733,
 0.7453816643480439,0.7534738529597294,0.21102586335988338,
 0.10706657206085068,0.4920922473649346,0.1533622913097067,0.6787815362677634,
 0.9575805663741056,0.6437512544297341,0.8007087431179947,0.8711725098744835,
 0.20302644369814826,0.13351110410450617,0.009811976852597049,
 0.3395317621727637,0.6441918220228713,0.028493338904498566,
 0.9959647276153174,0.6119105155250939,0.9978963240652805,0.37696361342600404,
 0.6484762080262109,0.2125925267652068,0.3366475296685656,0.5756307141199051,
 0.6894018824595963,0.17239291318745187,0.9754942709592532,0.8707535690890998,
 0.043246412331200856,0.7646048398951565,0.023745511435710243,
 0.22132891730745408,0.4736128446854736,0.8400390285188067,0.8463807175978626,
 0.43190690540845567,0.211699976574803,0.06887168276651723,0.94451608017073,
 0.022077944958874718,0.24403947670580328,0.7785513511304426,
 0.8302903418396747,0.16987886452327272,0.07448553892050103,
 0.2509543372330657,0.568038474292432,0.3479898642001552,0.18710227233767407,
 0.42561844598023146,0.5838094992524012,0.2287091454660437,0.7575015543952587,
 0.5777677729686331,0.6809333922690461,0.22754516632109734,
 0.38294890612806065,0.42016632158228084,0.13289182427753454,
 0.07528869684178752,0.4406421551216705,0.2211390781682826,
 0.10419746904891947,0.9433275649257151,0.4180880307333066,0.5391316272258455,
 0.4447865780328826,0.6904382292667958,0.9214758706146645,0.6915992684148912,
 0.49663367612599885,0.5459869127684822,0.23979933100400497,
 0.39285765765934333,0.8925888285171008,0.8235807263444616,0.8635833375878977,
 0.8549743335722793,0.3730435080528586,0.22422414538363633,0.3196872253342171,
 0.30947164361914237,0.7125455238992647,0.6730736637451138,0.4755993393115372,
 0.35366289372424187,0.5133161914351319,0.9848966919707427,0.3249531295608654,
 0.4649562534264762,0.1641425352380831,0.511134306659255,0.8513270548024777,
 0.04476071957300376,0.49510333285212094,0.9552435049166783,
 0.05749825615606596,0.08032442034858589,0.280379475889725,0.5162012147029666,
 0.3999245438061988,0.586524692281191,0.8143985883801749,0.7888372233324612,
 0.4300714698626589,0.32573196446921215,0.1080657236579824,0.667512791524241,
 0.9669405339863109,0.2755803777485549,0.8220865673455592,0.9418518344331207,
 0.033342867007592814,0.12574548344512182,0.1914168005585971,
 0.7193053388805081,0.5182483728548611,0.0453178287660424,0.9742385297569202,
 0.21011444695512238,0.09008330551928712,0.5495846911782263,
 0.6943294858996735,0.28911973695446014,0.8746392194797177,
 0.36466300058363943,0.4419356953591098,0.16118596339054447,
 0.4462845594340623,0.269602190968912,0.9373654817143182,0.3299913545404476,
 0.7375114173938265,0.20486533474230018,0.5413253119454625,0.756582999600268,
 0.6678454583397369,0.09198543426677352,0.08586886697554075,
 0.3734935009066971,0.19149404772741074,0.0843485821333817,
 0.08819497683674582,0.5274712782689142,0.24581521683416807,
 0.6356861939385519,0.9903426352488149,0.7885186157270697,0.13703649317235578,
 0.3424366575723752,0.534990298632434,0.39809761440896285,0.9955581673252812,
 0.2461604853103595,0.8945007643401324,0.410770005572329,0.8765867649015474,
 0.793150453691176,0.8843261248361852,0.17249162572552812,0.8161962720939309,
 0.8201994898428517,0.46630123065585005,0.22126512705783197,
 0.6759672383324443,0.5166144741827207,0.5520723687928358,0.39343429213242986,
 0.02992844220735935,0.48494925481841733,0.4536400850033072,
 0.06493057837507044,0.6557144821377783,0.8891699244838397,0.585615067323355,
 0.6824084217922252,0.23050879711277406,0.423618644641905,0.1166981979377183,
 0.48235051529436335,0.4040235503315204,0.5542232919178927,0.4048640225400809,
 0.13628626510888953,0.9498474359504535,0.46711460821254447,0.259750863806959,
 0.5514247433528865,0.4989706420061343,0.7029806610387034,0.6994985878389393,
 0.17643257749735275,0.8738831041117393,0.29976707329082974,
 0.12295624218171375,0.854633072857985,0.8381968410423476,0.5023571710373704,
 0.6936139639259046,0.574403612557661,0.45431443893974655,0.522684036272426,
 0.2295093165361467,0.3791155518418091,0.6611106862615542,0.57938755974637,
 0.3939745152670322,0.4834819842182083,0.2270763409239802,0.8855658405961745,
 0.3559525860527337,0.7944734408334752,0.5688263651693674,0.17943056763073273,
 0.2011578490751217,0.8882834889777258,0.039921053384107674,
 0.7325909492034365,0.028614711106949242,0.741459545872288,
 0.03753017336287323,0.04295380184037367,0.5032410602344118,
 0.5832106332498779,0.37322090009216924,0.2544748080080612,0.7771725957994114,
 0.8380821913758557,0.22885544335359997,0.5880727369161395,0.5774634898583874,
 0.9559553279043473,0.8175675055099664,0.45526950212965667,0.6260632329388405,
 0.015336137837803632,0.2953149770906168,0.649750784794959,0.2712235908508024,
 0.7321263312823655,0.05816004542659914,0.6319305371162569,0.0653259216345553,
 0.9809349102717537,0.3713322268495485,0.6198055742605229,0.6936625268493869,
 0.52330305736706,0.9157676792586871,0.35811327321620834,0.5181526977905421,
 0.6606653638529374,0.7296038890106393,0.3951609060574002,0.28877525209234056,
 0.9323827091135753,0.5149749562509389,0.3427362145772017,0.9607513340864111,
 0.3402592828407812,0.1893391308698089,0.44723396732987153,0.120060216980306,
 0.8696095320405395,0.6933187834983725,0.417496889048098,0.07225904496255431,
 0.6573623415727272,0.7922092889377162,0.7827060021455314,0.6122425827074677,
 0.5288191138984629,0.4186151817127792,0.06954855839346596,
 0.42766860003961293,0.6770496776575619,0.4308894875464945,0.7813227119069798,
 0.7067225908074319,0.13533160052383275,0.7399374344789643,0.8130821589070705,
 0.24724568608638275,0.01525499984251566,0.9808290859486704,
 0.2758758096414038,0.34191323167292886,0.09745104570552166,
 0.5674020898391094,0.2416383516150028,0.5873292038123079,0.49996809920877494,
 0.31210112236980836,0.6758950031001859,0.6966886763267883,0.2485208943271926,
 0.876146896477733,0.1342309679167878,0.7086747519660711,0.39508743479291386,
 0.016138786329075505,0.8846446878902339,0.1551755477468042,
 0.9862268020036483,0.7711503549066487,0.6051804321274723,0.8664527102585722,
 0.252071630239465,0.1721679695777677,0.9295628202361796,0.14492137268814986,
 0.3186528229605855,0.7453044870729055,0.2032345664548887,0.0443401888707613,
 0.2577695342184477,0.29378789236973724,0.5339621420822807,0.3477278327772364,
 0.9668656106719492,0.9589561858537892,0.1425622609439906,
 0.049122475190501724,0.6991678777926685,0.014375288981870682,
 0.21721775819632538,0.671715973637541,0.8922365179544904,0.2949092126519819,
 0.8313630276329106,0.3373833969817479,0.24877914011692515,
 0.49563717134628715,0.45200270517694974,0.017252933934549075,
 0.6005909788566154,0.6431967073104707,0.5381871694111479,0.21506288632121073,
 0.022121862743377685,0.6321447522949667,0.20559954490636423,
 0.813193327049983,0.39165064396393423,0.5237044790303788,0.31188092532389056,
 0.7183131508250137,0.6336805804444541,0.2943693237926214,0.0655895201120884,
 0.03596207406003582,0.6319226615205235,0.6179106707733265,0.9901549128351063,
 0.14916027768849272,0.08348950825992851,0.7340141310474557,
 0.05887218637201619,0.4466374964387143,0.6432107391317259,
 0.27717704327975456,0.1916086526711489,0.0965664058177873,0.5672472673159825,
 0.08373033487479575,0.9081458287073355,0.550937207334078,0.5995101704177148,
 0.5565271679690286,0.9031994363683049,0.4037632151485646,0.7071757108060657,
 0.6699775401661494,0.09538991735033142,0.8505712828679561,0.6909638347917342,
 0.9999428263819261,0.9555628173305037,0.5602987114441219,0.13507681748980027,
 0.4709786915405957,0.4647012607943357,0.7541776170239242,0.33400284021028426,
 0.8272157375912919,0.5478080260829371,0.7152816967344968,0.5031041654473982,
 0.7630226508493502,0.7160037089943845,0.4888878595164192,0.01570954225204979,
 0.3937975593996631,0.7425566061297018,0.7708906150758729,0.968092039537109,
 0.6843642767346394,0.5206582674152485,0.47514454050552213,
 0.17917681749851277,0.8656270106844016,0.803976673722133,0.3313469051323259,
 0.5967426401730381,0.8666658320142439,0.16998725571041695,0.521446521042695,
 0.5264852324982068,0.6437029175024077,0.32916823176160404,
 0.39760805666783494,0.16620046702368585,0.42149112393062704,
 0.6670363454867092,0.495998263986295,0.8074045674335146,0.17224417034269124,
 0.01669819671187467,0.018413971709747567,0.6780821806058314,
 0.687088618469977,0.0661445688943384,0.2102207503479081,0.9239742946893237,
 0.7362147607102383,0.48104163486634977,0.9934436100598476,
 0.18087583554458808,0.47461292275747113,0.5839194084688689,
 0.8906636985994523,0.39434535529944725,0.7840300846642139,0.7027659429774515,
 0.4728480318419317,0.735976731037806,0.952766467996725,0.27015612601602246,
 0.14695589463241954,0.20859335812426272,0.047999661112825054,
 0.4375913097633941,0.15909973805537392,0.050392891667680906,
 0.3349571041180158,0.7918670502123621,0.34566032854156625,0.5839969338632972,
 0.616923042638549,0.605158292926739,0.8227584837806315,0.73505727783652,
 0.9286174632642687,0.15812311187355854,0.7656451300856884,0.4383420333800696,
 0.13605149330268174,0.7144592470197587,0.9294538762016329,0.6936887024785053,
 0.8651333208217824,0.8349263911068685,0.22414216275510246,
 0.21152312185267907,0.43233363958814497,0.7502339499160312,
 0.04360605035943732,0.6909959822254034,0.774500954137512,0.2222682535359949,
 0.749211295934967,0.8592006100925811,0.1477595359284558,0.23798876088832754,
 0.17062112963193132,0.1275101627459564,0.37882123553880664,
 0.8859584768826986,0.21698692766425587,0.5963196186261569,
 0.11356342685251775,0.970973234996458,0.8666102611240666,0.09217014261134204,
 0.5030259526414719,0.21818013014901305,0.8130847910814314,
 0.30429900978980584,0.057020372119721,0.8720240531702035,0.5219611314088771,
 0.3996324495223147,0.7562374701745534,0.22662374420830333,0.3975168848925199,
 0.8548141306546251,0.18065064244614304,0.4543915650857887,0.8050976245534637,
 0.7338221445666373,0.8304115105514165,0.012937477541765774,
 0.14460893717046175,0.3698337404767382,0.9537515524779033,0.7254038265316953,
 0.03978570800898862,0.17952876523953576,0.4571444125310067,
 0.22628706656264153,0.6750734421788103,0.7537750963002222,0.370459513066151,
 0.3361308953419412,0.3325120828235617,0.9271335728307707,0.3667113471530049,
 0.03739976006496237,0.613717781034674,0.9448809533792001,0.6720176243608815,
 0.23454915661193376,0.5908131080949746,0.2613673571836008,
 0.05035193804624849,0.4731203786760424,0.4302054524945551,0.7646489686989474,
 0.7436412398889259,0.35447560393607624,0.16568787314057154,
 0.034031416023606376,0.36982675163863976,0.5932047097768841,
 0.15341400266502347,0.08584223164220262,0.1060766668629688,
 0.3589598112451998,0.7365336827609834,0.5500702805718313,0.49163621693122506,
 0.45735273541265786,0.1293524598263852,0.6752033065557551,0.619707740780412,
 0.8202219775291965,0.10069186743415837,0.7643654949750205,
 0.27378882076704003,0.22937781683219582,0.7656890318759151,
 0.13456271103196493,0.7193129585755612,0.3298813071487451,
 0.06714930754811177,0.24114144203766363,0.4358767183141393,
 0.6493438204648231,0.773652171249304,0.4544911268980715,0.981460674652629,
 0.7525970305768439,0.4060559472161017,0.7366367788028115,0.398077905891033,
 0.34555690938142736,0.773914117412461,0.34991205174595297,0.8803369348316288,
 0.12736755614832274,0.5527242815530246,0.820450830141286,0.6198474425538035,
 0.8016462424040525,0.2839947841162994,0.10722325242097008,0.6097126454167223,
 0.9000115229190423,0.33141057220514913,0.6460037296507083,0.3126568724570553,
 0.9053351801654191,0.40013330178327966,0.9046741088304671,0.3255860308640006,
 0.4229488526924646,0.6011958709626163,0.002255089007628497,
 0.5677159357610914,0.03245114621174805,0.3246857069957543,0.8330361878491291,
 0.12907571924992922,0.831144823265743,0.48536288653161186,0.6987092511214248,
 0.017071495677112525,0.7751485954507729,0.2529383716009762,
 0.9002009626032729,0.9933393688340239,0.38671710811531923,0.6109523847134776,
 0.7556761091420447,0.2367418168969443,0.22015070723511643,0.6094410887356592,
 0.5096146730209189,0.218497448537602,0.07737852164570314,0.7144838367439391,
 0.16724658711412177,0.2955529973434655,0.3469112013286678,0.7368534388085802,
 0.30877103390761096,0.5379327814457942,0.1992932392188066,
 0.45398992250802406,0.17079650303133187,0.5411209871544128,
 0.5344607594116484,0.17196299730817566,0.1425799690356614,0.9157749391135613,
 0.3230936532114981,0.33817350406228586,0.050738580840617886,
 0.12418953340299899,0.1455596821985743,0.8123334512135225,0.5027452945821819,
 0.6606474156910712,0.23864209758264232,0.8841166106820066,0.6772122642964574,
 0.7736014339556951,0.7626415900539119,0.5997190442178802,0.8969553128720094,
 0.20454338001246097,0.4783494807083438,0.5501437969028246,0.9457205521941918,
 0.3133813966296781,0.29141500515356944,0.8075022385910898,
 0.018818349293883263,0.011891371238406667,0.677155958911088,
 0.029610264702804256,0.4744723999080874,0.7036490179827971,
 0.6725055649352403,0.42212921478834775,0.5067375148333801,0.7688692028968735,
 0.39418439840435016,0.11570163859636595,0.6794343096417593,
 0.8127054165186363,0.19209513376429443,0.5282011912561695,0.3577428681491328,
 0.7707763494051972,0.472490578578455,0.4596587601883395,0.03404944073809035,
 0.3051093055330928,0.23134826123663121,0.41256915666703886,
 0.4749627691124483,0.8636382431714223,0.3036845941595685,0.546752016269182,
 0.5878072925989722,0.8757842866041127,0.5786040868753977,0.9899846734533447,
 0.4998587270251347,0.7873464158756915,0.5011114386972751,0.235159379228857,
 0.06076602525436747,0.009323892605976325,0.5133644593774341,
 0.25294235015489175,0.7882866543395524,0.36748435312434624,
 0.7191354499929276,0.06384581733478856,0.6181697080762137,0.6337338174541524,
 0.45115424545777194,0.6492373653105544,0.571278522871329,0.0723458559269522,
 0.44435638246960407,0.9842715205606869,0.41363326125168653,0.667931049113272,
 0.5355227143540662,0.4165017710209653,0.3909782414024098,0.8716039482257879,
 0.42650498599061915,0.31274539362717657,0.28600792721962254,
 0.8875615378801343,0.4364820139800064,0.08512416233520237,
 0.10488517991172852,0.5916202522236068,0.3402978960746821,0.5543650512719535,
 0.4296668506568557,0.156095781397396,0.3905344481302725,0.6627171626611937,
 0.9713761192866073,0.4480258447177268,0.7673655373237909,0.8018189401833214,
 0.5105456107843236,0.37436509410553354,0.7469840456575615,0.5047803430018301,
 0.4595019318307926,0.004309323532251219,0.8792438070494892,
 0.6751786721196511,0.5856228052754402,0.5715978231364262,0.5848063825837312,
 0.6558808023696303,0.23979097834237273,0.29887522015448587,
 0.6163903608002017,0.4578656138092818,0.6452956683052281,0.21845566698968966,
 0.19340564793427806,0.6968386778898947,0.12143832664217058,
 0.38226908945568017,0.6195293094736125,0.6171860563743929,
 0.039253946316465727,0.459658302664981,0.24527499069572767,
 0.14231195390832074,0.6906988142436047,0.4692708555679549,0.5269512827409155,
 0.027630455389524466,0.755733769480968,0.27756752514302696,
 0.06650635952088724,0.20215959634409997,0.7072346057733279,
 0.7846086684772039,0.6850105630591122,0.5600824741021596,0.0806140005812267,
 0.3233551874628793,0.6780362783941912,0.4590116932339342,
 0.005767656977284763,0.35399528902600796,0.6725935274607473,
 0.7827042481321206,0.013551315129372687,0.3975436833434499,
 0.4989791495912339,0.44596414637563886,0.10174255513034003,
 0.7064252267019515,0.018180210027970323,0.1493605764501249,
 0.8677350699093351,0.8570591399898583,0.28754214257811095,0.6464633326570453,
 0.704755926726091,0.936979149770813,0.7400074525370178,0.37271339361579725,
 0.985712199913936,0.023105050246758352,0.46527082624201843,
 0.46509235730928156,0.5141199686692466,0.6299058995917712,
 0.20234656690036856,0.8194328490872708,0.5425201315311597,0.589427762619078,
 0.8024136295726881,0.5587838758106538,0.15394736471658677,
 0.11592879677131684,0.48832721251316347,0.1112968499468896,
 0.8631388328537126,0.7758215314992085,0.5977706653351512,0.7062780642730757,
 0.3853398770346317,0.2106363104638107,0.8256549083151103,0.5435028429951831,
 0.6392215751806374,0.8425291893341076,0.06502992529275231,0.5023962535786041,
 0.605788078616015,0.7004100153647521,0.6641997904483422,0.6920543470792064,
 0.5885160362633536,0.7132470140568489,0.8207948136389944,0.06154293552633128,
 0.6584885524514574,0.08176832986896398,0.8931105305777135,
 0.01503224636630307,0.6805136461678883,0.11457326698780246,
 0.4329616842148354,0.8250207732188177,0.4284588942940939,0.977423053086335,
 0.13263572813190316,0.21815194898700951,0.24329598550240272,
 0.062232552615887005,0.7882548429150417,0.09467080991789689,
 0.32482912860348057,0.9239819961694543,0.5412460810198527,0.3116269164000188,
 0.23048588096236222,0.10787885799438612,0.39888406110880337,
 0.14725894870356737,0.9096868939327292,0.882584980648168,0.25742103508596803,
 0.1101544533872647,0.4235529025942393,0.13297418021696195,
 0.06371148128672621,0.013769407218732876,0.26323564513053843,
 0.20227013355276657,0.15736240108570876,0.789307610527075,0.7011215896985907,
 0.12316995414520582,0.031944331077508714,0.4186592509782895,
 0.8388157827874655,0.2059081596529415,0.9166012290468446,0.9174646230164508,
 0.5933155736734318,0.8348352917201213,0.56797454793157,0.13170767772166292,
 0.2670938350183465,0.9406008539719336,0.4561524139541552,0.213819293382694,
 0.5449446066701304,0.7296597966580227,0.9409771366011578,0.4382083642097555,
 0.6206831227283005,0.6624209659106332,0.555830105943875,0.4892783726289963,
 0.46667829436744845,0.6092774980679354,0.7666013332973984,
 0.13022450852933398,0.38263411730556796,0.19067567325204393,
 0.5829466394807482,0.31466734408589836,0.23033989714436842,
 0.30166707875852306,0.7296310588766113,0.7212981909226969,
 0.27745050480576605,0.047150903994698457,0.12053087578819988,
 0.20633795465257676,0.8620880375673581,0.28085154280131486,
 0.26907770906936035,0.5143831545444192,0.3628586175531401,0.8422658703133128,
 0.21612831234070118,0.7152606019441354,0.5801260120199683,
 0.31219360215593284,0.16490418371486315,0.2907236802927117,
 0.6886436326625629,0.43709886965859757,0.6280629658193742,0.694643654896903,
 0.9006378661763325,0.5529164896528566,0.168510151035823,0.6611727843829266,
 0.7294004413934189,0.3354795882763577,0.5129118334297347,0.6709077659953947,
 0.12308530507484117,0.21642039122794865,0.7952199603409928,
 0.3954700113556374,0.42798296035750716,0.16708676102784814,
 0.9288654561263534,0.1739607884737865,0.49631764552597124,
 0.47110065499380804,0.11002036218935785,0.44417568276788555,
 0.9795868827012327,0.5976772587098526,0.9539810284128483,0.40519698597050646,
 0.7634347599085672,0.5437971038008143,0.6568372517476524,0.6017544674108032,
 0.8145611675177823,0.8036144024120591,0.7161104868000847,0.19944346401673763,
 0.06765538028254481,0.7103246486144545,0.2895147391110031,0.529713605123311,
 0.5808843029298381,0.16376389828026594,0.3880687035125858,
 0.32580558067363796,0.4795684107795398,0.6736154051455842,
 0.11519575067078802,0.5913635497078835,0.4472962793055827,0.5618809764249801,
 0.273618487402689,0.5774587435208561,0.2944541946164231,0.1261045528464373,
 0.5206412819398727,0.29643397404982474,0.9600655229693956,0.7529519920186684,
 0.8853196857122316,0.5581954893258383,0.09431852344706526,0.407009672089633,
 0.15250750521698397,0.21960661878317467,0.5325822776657448,
 0.3824131237471764,0.7533729411393562,0.8366971816552541,0.8289532521040311,
 0.01791208139287659,0.2839255889412965,0.13094400151273022,
 0.5782488538634284,0.031384684528365736,0.42797702037246566,
 0.43022495467587396,0.6959587325165744,0.7357602019057776,0.8961519106975625,
 0.9078406704967239,0.6550903128421806,0.007373316394794327,
 0.028276292277396142,0.7888979076802193,0.08321280840389056,
 0.7847928670552793,0.8687269496292811,0.5937414494705593,0.39981654644154785,
 0.9021202199639629,0.6349459605380339,0.20431986175781014,0.2498676387594475,
 0.2603511509416099,0.027678480265771466,0.9318847587658318,
 0.5054417732643727,0.5445856396239506,0.0700646162237939,0.7210222180583996,
 0.837250962875529,0.6188979741380507,0.6164092163624882,0.40922388015292277,
 0.11723283642815929,0.7821958680743387,0.3068420921102647,0.3361352683263924,
 0.507104357030053,0.2744187357820306,0.6924024496233883,0.20805675996342732,
 0.8093385612125797,0.6449693640770711,0.07470638093677617,
 0.11830555729047876,0.9365349243327329,0.8805549689974623,0.5725007114381327,
 0.8571721270957947,0.6106034692529098,0.05945040327743967,
 0.18968221075321223,0.4773274658349542,0.8118819548209721,0.4602941339257415,
 0.5841644726480628,0.8511143911285434,0.47070844019268265,
 0.038850947028808536,0.41290138881184524,0.6947671068052789,
 0.8072979465817731,0.22613805906282414,0.550283837062775,0.9985906215618658,
 0.9637459540471023,0.14059765976001515,0.7324096240506958,0.3784210717168437,
 0.4526798042140253,0.08386863054860871,0.5061317834497842,0.8645972215626264,
 0.8593942587154337,0.3306190663299612,0.6954752736654091,0.8288380480624084,
 0.05929304900622423,0.6836631570982683,0.7418805742859969,0.4428956725089169,
 0.7334905779708534,0.7593975319290728,0.5052945777633214,0.8003413199913616,
 0.6969743815907203,0.8412561695694172,0.36822855154175627,
 0.28111092247699676,0.45576462746310664,0.01350531640618624,
 0.7023922609259374,0.981222099056546,0.6494478929204702,0.5304531408009653,
 0.6730412111163643,0.6536372877120291,0.20743211698804853,0.5721621730083617,
 0.7009270483690164,0.09352911712182455,0.4060379373286347,
 0.39206169454033857,0.041392164931542186,0.09400499917756155,
 0.15051367485874678,0.5277159874549638,0.7995767664465363,0.6184111950391133,
 0.4331061597460285,0.35082199218957366,0.1388283009841499,
 0.17499273119972247,0.7258516577516572,0.5329413934224677,0.9751695409472315,
 0.55141068992899,0.2912035161207367,0.7066811730533116,0.7555394186633525,
 0.13142479222802228,0.5079311397491172,0.8324928170059501,0.8317183248845753,
 0.9450699017476926,0.4793768873242179,0.5717434242179721,0.9842539774590557,
 0.10456185750604752,0.6379568543215105,0.7742231373231127,0.3328258754597533,
 0.9895414161076929,0.560899547796867,0.9127369285160198,0.5612535725028283,
 0.22707930904191576,0.05730549371578564,0.5502648916939966,
 0.04845882179431382,0.9520536570278554,0.2645452552426589,0.9417261395528859,
 0.5852898780472345,0.7881957377657703,0.34917352764811704,
 0.04400429077492918,0.8390557998228909,0.5361163006356677,0.6924046015757077,
 0.02395731234650922,0.5550745062848483,0.6275286053431636,
 0.026170961118714864,0.6996655399181083,0.23994989874544193,
 0.34335905437351233,0.5335052708992035,0.45263060764534113,
 0.43007839509286294,0.41513162036096496,0.7474205528137311,
 0.7025304839299888,0.3044328540848844,0.8539492016865797,0.6099511865807393,
 0.8645529670142275,0.07659917186547749,0.04555877848559753,
 0.16164192907454633,0.015223974307207566,0.9189723503589011,
 0.9562903494804336,0.3634787954270856,0.6464454711401137,0.09355552151495905,
 0.9649107661745482,0.9732600450210686,0.5295439311858311,0.20321765013586923,
 0.22101363496121273,0.9516785409640195,0.49925436728269446,
 0.026090304553432198,0.16337402317507743,0.736361340708631,
 0.9606545732638087,0.23319627674088717,0.49165714504231917,
 0.012764174515589422,0.5579782236374261,0.8683362570819171,
 0.32291866829156546,0.6000988463789241,0.7256303395283579,0.6231873564923704,
 0.5760366116539628,0.7404686804947714,0.5312700050467469,0.19128980314214217,
 0.8596105956309326,0.014788300860478887,0.898017517278819,0.8338489729072753,
 0.844548318997767,0.7258053185239706,0.4363272839686574,0.8499236983379569,
 0.12787358572615504,0.765967298095468,0.7376194006602165,0.5371914913726201,
 0.3047454150988079,0.11341199732548701,0.9944911528590479,0.9608950192894657,
 0.20719340133947572,0.4467747956001663,0.17731747999582081,
 0.8250569397356989,0.652474253809985,0.025508282342067368,0.5802812488907823,
 0.7028632089386653,0.9161449895002571,0.03137564704004736,0.7867392851683235,
 0.2666416603375765,0.4141565956089772,0.4583281572137393,0.5755895705989743,
 0.033948578789511186,0.1018506614709912,0.5001066693512594,
 0.21287736823793957,0.5794605520563598,0.6892020287580064,
 0.02695359076316073,0.38301648274711,0.8832938066558534,0.13466912507384787,
 0.3582712268583078,0.801829953928385,0.9089570785018382,0.1490887533892835,
 0.5652280588326299,0.19687489741559072,0.49700369384886645,
 0.31148102574806813,0.6680271851651677,0.32971057950791716,
 0.43539218391688106,0.7946110204060898,0.3145611342447231,0.5340465783589742,
 0.4553589349402548,0.13200354833860783,0.12775859215010377,
 0.2923609220457509,0.5275353139043786,0.9878104380293485,0.12630120867390193,
 0.014306919893707248,0.7894427695114752,0.9654835552788545,
 0.8162368136585586,0.4710227213377737,0.5971134937494148,0.5487264972344702,
 0.04222055535417668,0.1510719102258944,0.7131039620632019,0.713202444056369,
 0.9878090077123758,0.5152237315566158,0.15368480836179965,0.5828927823468097,
 0.5731573100042546,0.34568061478725953,0.66218464786209,0.9872895412142133,
 0.638419985790577,0.2173738647285386,0.010835035332651288,0.955731955885188,
 0.6790721154964451,0.0911534916193677,0.5370374860221054,0.887029312539234,
 0.7246362145141614,0.485389360033467,0.22702662560207532,0.91435088866525,
 0.06556735715796058,0.4026476124758612,0.38159251230474767,
 0.9910014312618796,0.7113059438902203,0.3121624746211584,0.6384162877979548,
 0.07133037780584983,0.3636940111479505,0.8923865978503913,0.2787188540969917,
 0.11601706475886064,0.591629446516484,0.20070826245575635,
 0.36879054301635894,0.9852422537140125,0.37918656535500594,
 0.4849369195510802,0.6005028277334024,0.0511429845135849,0.9172859811535659,
 0.1797577800400516,0.573399907425861,0.3011859277130893,0.9977599083351667,
 0.41940168096383046,0.9462658250147062,0.757112714103669,0.8287352679078044,
 0.2053064308354381,0.7210901440976386,0.06561945364073074,
 0.47238230211277665,0.4398144508747649,0.6727966141615077,0.8939551125217264,
 0.2181245512554444,0.2365476930638747,0.3082679218622896,0.8159142638604964,
 0.1550301698733937,0.8071127063492,0.5481103962074841,0.722836485722288,
 0.07931274108867803,0.5475482807604927,0.7392828475969698,0.7592605108382653,
 0.007123804107052822,0.8698393879615768,0.01872963451561427,
 0.9548335134999919,0.42085655831104085,0.552798097044242,0.7230020960259423,
 0.237187654432772,0.901514522178909,0.5875009975263411,0.20727265193965816,
 0.6893011052482659,0.2243652901989298,0.3256159202187431,0.3362528476222456,
 0.06327947171236337,0.5542846989740268,0.3122489501315344,
 0.34777430829427847,0.966236651763499,0.6885094250798551,0.1311818982540659,
 0.45189032849507194,0.02579746978089892,0.2115870332897275,
 0.8063875288011111,0.9211520334951294,0.0985361350957703,0.829970925310483,
 0.32123462993579466,0.34058852553213004,0.2722009515300219,
 0.2985479079733927,0.38215401584761044,0.011189543446861183,
 0.13104441345660667,0.5608624263059833,0.9916381433731508,0.5214392568622706,
 0.4205138350099705,0.7710676620122996,0.7489778422417025,0.9756711747595824,
 0.5157412332996432,0.9464496747401024,0.45936915176552107,
 0.08025961497683376,0.9985252667738218,0.3519876752397475,0.2245988403506658,
 0.030763338514065652,0.8358659024527042,0.38548085553365563,
 0.2808919386381016,0.2502736168668457,0.2991316219759417,0.18817520797342024,
 0.5420059681869847,0.9229571186356781,0.08302341238408917,
 0.010751964600117181,0.42956548889516954,0.029256847405154107,
 0.839836270509692,0.7084749111762847,0.32108837767198706,0.9164951663340465,
 0.7851301565726662,0.6582661144748069,0.6983298820723565,0.9231362293615522,
 0.20916780943417368,0.3539345567134893,0.8802002731794611,0.8387827255147644,
 0.8134522998558094,0.11778109239446388,0.40019588980982046,
 0.2113581277452271,0.15422521722495652,0.38110984740999165,
 0.47754142156045687,0.4627924217024947,0.22597366103071337,
 0.062468099621421636,0.7859085387336359,0.42926846983297484,
 0.8327352418154942,0.6044226605236199,0.8941421569390857,0.01226704165553294,
 0.2085656334192043,0.555051258396238,0.9661637225499068,0.14157549419675808,
 0.06278062585756106,0.14723781958980087,0.31459318398657987,
 0.9134639457599109,0.3436759865410687,0.8785868919171591,0.8239403685552197,
 0.47495877586322566,0.01207541204439988,0.04307771323979703,
 0.45132573564138245,0.7701215160897615,0.0224127379265755,0.4293805480451256,
 0.7068763304646802,0.3898171384752851,0.6057213266488645,0.5925744053066822,
 0.5483292508414059,0.6582450109095668,0.058584181624618425,
 0.2934354937924357,0.7227855428791443,0.12438620450934912,0.5712742646186052,
 0.803775606213392,0.8319981540480845,0.9625836615270394,0.697921057068428,
 0.10707768335762513,0.40003524132037227,0.4871860842320245,0.80805477927635,
 0.4301402997924709,0.7369976368249342,0.13337105427797047,
 0.05069737792779894,0.19064216762140562,0.8746382067366266,
 0.38793531213086874,0.054458693340984854,0.4602103960023861,
 0.8434684169289541,0.3504763607688226,0.04708891997581499,0.7300751376936093,
 0.042765607021111096,0.6240546128343377,0.5748145538155934,
 0.5394322775273651,0.28152160820989525,0.6308922790164364,0.9419370862144766,
 0.020145929398904916,0.1675958495419012,0.3342134666267186,
 0.33005208114451956,0.8838559184901921,0.7429084359294633,
 0.38088121093315963,0.33012800759987304,0.17919464873875068,
 0.9471480752652295,0.29659833100253774,0.796541544097532,0.993273909987751,
 0.6719970259026485,0.2670258072348536,0.7691244877827723,0.02598590994973249,
 0.19973649947403982,0.7687900544860023,0.668115198565616,0.988297033207979,
 0.537764551048286,0.8946864883401585,0.24285253557357622,0.2356692590884445,
 0.02716479564465235,0.6652349161044534,0.2386313926697642,
 0.05994369970870417,0.2364359487415444,0.15791921023578648,
 0.27477866469219836,0.26792665061129173,0.6742382321449438,
 0.5145973342645536,0.931463006068306,0.34780660942169306,
 0.0016110936807525178,0.7118165078951318,0.42581613626018655,
 0.967381264617367,0.8864507709392292,0.7543691700167008,0.06276890526611845,
 0.47400023789904655,0.84198184986679,0.7267096870476377,0.7072285345482315,
 0.9118724825943652,0.26609430600138706,0.4486218323780845,0.7270872964066752,
 0.29926545511096614,0.22043589259982588,0.46346541352511417,
 0.8126363865361005,0.03701926786727783,0.5056975115303726,
 0.09595307361947114,0.1559541168567551,0.8150723626994062,0.9929010323656594,
 0.6146559090299544,0.25318978232630507,0.33115261494316206,
 0.5850066721015401,0.5550088125300261,0.9339826255722223,0.8897438595393995,
 0.08877925162068145,0.2244765600994063,0.24316502631942816,
 0.9164735408693293,0.8766347979889871,0.38328413107463966,0.8409345104778335,
 0.25616301505931793,0.14749185533844433,0.7238230635473474,
 0.3559598777650863,0.39318184170034154,0.5225096555136428,0.5730349913903716,
 0.803198851936794,0.7751793912499061,0.7539171754396503,0.08329434121758306,
 0.09434536158223561,0.9755185417966579,0.5245641680274694,0.3946921811908838,
 0.780285877633627,0.7577739795262463,0.18051125080759434,0.4832181999073276,
 0.5068043478454197,0.4497614104968457,0.13562983562218722,0.9653496974114025,
 0.9920352722268087,0.9905901670861781,0.16568229282316482,
 0.08198707212008971,0.10203265696077757,0.8234435046832289,
 0.36392151891519753,0.8517500267383429,0.9671031574101593,0.6938631862761552,
 0.6635317153462464,0.5126261834091729,0.6583708004114273,0.18815995810098307,
 0.5294622094528529,0.09784542335212776,0.7655054177479319,0.3340906599553536,
 0.09144782802266338,0.3034236075138883,0.6061710696401742,
 0.16150272673413701,0.45485362154826503,0.8824685305163026,
 0.4065262952680181,0.33147073719202347,0.321304614974953,0.8307371453630832,
 0.32320545107996657,0.06900587561491223,0.11503969113406498,
 0.5673600074558054,0.012470183481246289,0.5521378678779626,
 0.2489156966847188,0.9710226953348412,0.2441598157594771,
 0.014098851219844222,0.28130339935192583,0.13992727119797532,
 0.7360892452905082,0.4913551635055502,0.7678378227951952,0.2889508177049872,
 0.9575878042051522,0.10813085090049057,0.9159907859163318,
 0.22354362639472314,0.2523982232423676,0.7653113909920152,0.7639430561948418,
 0.257668708861744,0.200719542982972,0.23585881323359092,0.704778156372815,
 0.15424861603278883,0.5338883762495871,0.8218648488878558,
 0.30626237881632556,0.5518009501925548,0.5084030560001696,0.8624547841585622,
 0.5009819601681358,0.07080624632791821,0.8875665499182072,0.9767081142515525,
 0.47215573052977045,0.6846765508788183,0.39587736666783435,
 0.6576423034939348,0.5425705010594849,0.38629257762025426,0.3762541679424267,
 0.2459864259727652,0.47086314754271896,0.1846570839806193,0.7135846562426202,
 0.21585567930574168,0.6175331858473004,0.44263722567128183,
 0.7576115754247648,0.8554081953716016,0.4693916174946252,0.8422501112060057,
 0.7427153676153065,0.11029987324970503,0.9955694880220284,0.6102716137505,
 0.25116782597711784,0.25050796248142515,0.11090645773374397,
 0.7900761624138294,0.2314075097610876,0.39024100968581277,0.2711540220871962,
 0.4226534859044222,0.09982212246062416,0.5144675301964425,
 0.42300007947826157,0.8048342661962236,0.8255373854596121,
 0.10142681102260642,0.22723746744217854,0.38216326198310724,
 0.9354252856766797,0.5330541994982013,0.20904103458412826,0.3027538635744026,
 0.28552332116650947,0.33480403830733274,0.9552925151848352,0.587427547945255,
 0.940233251047561,0.7139306191970518,0.19321923068046776,0.12007219612321274,
 0.5085326747405459,0.3634169622571757,0.083945145910293,0.8615244973217556,
 0.19861089341410643,0.034546056568968586,0.4208009790877937,
 0.9879220876097646,0.9409164917422186,0.15424343521677875,
 0.06408700871947326,0.12243478472068459,0.17123959021447666,
 0.7448582013110074,0.6199427228024489,0.33774295375891705,0.8602985294898913,
 0.3214806609939328,0.6116679411868383,0.46807655192093567,0.9306992473030871,
 0.3203304302527338,0.052936155210135616,0.27484658486288904,
 0.19440636644812959,0.5540795884044798,0.11471924167431324,
 0.3301095473064558,0.44383912204772324,0.1290696416756003,
 0.18351418979426792,0.20863302224217994,0.5855994988164642,
 0.41512689910053036,0.5560582602386276,0.10557926569277609,
 0.30793170224771493,0.800658413246489,0.7984136351190458,0.16533325513417818,
 0.5974955165898435,0.8811148073058852,0.7461756019239438,0.515698168517209,
 0.7881540635816381,0.4099722196935405,0.9877050011299354,0.5044180740523025,
 0.7070835607107498,0.023054978221080158,0.4348333257047593,
 0.2480815237915599,0.5920839206320843,0.7004889725559376,0.45651846570561794,
 0.2403854786367311,0.4323415012626206,0.8259379845483122,0.7091125028983406,
 0.5214110616235376,0.8961800998574202,0.4375178102749163,0.7424495019551209,
 0.3959813877265508,0.9944682934071942,0.12907521383019072,0.0567809560870105,
 0.6573802092160629,0.5741025858971419,0.25692007395326466,0.5728740046958488,
 0.03155731128814998,0.2725303087270349,0.7732199335347791,0.3167218818376323,
 0.7095053401128393,0.4037192155497923,0.832601666748055,0.4940718172005405,
 0.3672266218048297,0.6577297840687298,0.47651785560864657,
 0.21168651212023637,0.8858173066009427,0.8964452871672943,
 0.33809962764634405,0.26430155374855113,0.21616639943849492,
 0.9936160483902677,0.5962211474632456,0.28050982997516993,
 0.32382650222847464,0.15266387029372797,0.2642095862391076,
 0.6675799209205989,0.7060131632503451,0.11098745057715298,0.6003121223156084,
 0.19614437502577753,0.4479443094647818,0.6061383360215056,
 0.09767810848114356,0.637208436680048,0.6590123468688766,0.8451331835726327,
 0.31664639624726587,0.028533821804424764,0.06104377133582184,
 0.5866778436848894,0.23168926579199145,0.6580065010231899,
 0.47361956654216597,0.624542267698444,0.5694191671027942,0.8434810682699352,
 0.3403763148578106,0.025939703987308826,0.8992333671475219,
 0.05777143896399928,0.3274811226034282,0.15719969109614074,
 0.8744989503838747,0.05899328446321861,0.24358082788354196,
 0.18625463514820906,0.041948617861696014,0.9999101697872655,
 0.465384099818284,0.4574708902559995,0.39570694098582504,0.6365956666399806,
 0.5903414358633964,0.5342479277471694,0.6794512915711445,0.24908059534782923,
 0.8059006621497153,0.6760135687455731,0.09894680450890914,
 0.19898255165049772,0.326545953410713,0.3465997277177414,0.08618260458004168,
 0.20139635120680122,0.3475649777469816,0.45362343276355244,
 0.6033365095693307,0.852306511762602,0.00376727939750221,0.5133346349376521,
 0.7651599777318705,0.45185036678802193,0.8981109793998022,
 0.10133011182295115,0.02144253862544998,0.6125100811069618,
 0.8555996622397486,0.8094221372678887,0.7985256160760401,0.8631483613877143,
 0.6009288534884226,0.6116224155042315,0.03947472885422121,
 0.03634675308673918,0.6053700808644896,0.31903525926382925,
 0.3572802664570984,0.1258203086848142,0.3952024619823077,0.6660232472275552,
 0.6862636885677923,0.9648849425277567,0.5565612311693777,0.7804540598733165,
 0.007134339336719808,0.3426131794564352,0.08316124403820746,
 0.2317773553804553,0.3102360860192208,0.8221133892682483,0.3591984173045387,
 0.5155801184805363,0.19349441859117333,0.15548456401391342,
 0.2910549880363853,0.09426510858386827,0.5842933191293171,0.7324441890934323,
 0.2624759637901448,0.542430036847047,0.6027263328573743,0.4896386672949634,
 0.9214208672880675,0.6699542168281083,0.1429449400227213,0.7375933800232736,
 0.39323365067075755,0.46722997193126214,0.6521948611507787,
 0.06327475391300275,0.9512197144960657,0.32053415729059553,
 0.6807355164618494,0.2945214888746813,0.995171620108863,0.4320008948603079,
 0.30473980716009275,0.48055028942575007,0.9201094074262808,
 0.0765846682552227,0.11233600932461951,0.601917173037419,0.24756843123346317,
 0.3711814351331145,0.6644916225681498,0.37760541814918214,0.763601545147905,
 0.41304243610716995,0.5579680870576228,0.28296618761306414,
 0.8044270763152955,0.5900900168800306,0.45028058920895264,0.5322465851077332,
 0.9269022669344349,0.6461024848156901,0.30801619943732217,0.9213169373495863,
 0.8149242537878445,0.07048074671194227,0.05223086917558928,0.471776628588611,
 0.898524488183412,0.9803523569700414,0.5800309221628305,0.6441127983941549,
 0.9549773160770462,0.021160567619455728,0.954501346081285,
 0.38895787369253565,0.07418971293708998,0.2511928948216422,
 0.8193561584491749,0.447912762719873,0.9376170167066418,0.454361710420948,
 0.4554177325403026,0.9852983638770934,0.6218848378378686,0.0337515345179944,
 0.9610303775318019,0.13107731525908717,0.5472470834847014,0.8623003710927158,
 0.2818441094760662,0.6796717268752197,0.6669227353641662,0.35883870948551055,
 0.44572881674878806,0.5094524574346586,0.9033801931397545,0.3723457801811464,
 0.3500671916057986,0.44274258503285957,0.6403845724108009,0.884086101746571,
 0.26563785473783597,0.9477303393009615,0.4316677478573554,0.6284317424013104,
 0.11124134778380301,0.7719501387352813,0.6928335510611255,
 0.44083741572436186,0.19081678955359194,0.1664315235005689,
 0.7409330872077421,0.32638296697341707,0.6484465702642266,
 0.17473985488847166,0.4029111233015882,0.11074521058986497,
 0.5370226006187607,0.4897545887063631,0.5283624294926638,0.954129449738915,
 0.9855447217353405,0.34192786730018065,0.4011910887224922,
 0.48825187496516964,0.5174988132112718,0.6389817054156596,
 0.47188572440531273,0.9464449892627761,0.7356247556411412,0.4756311068634811,
 0.4793384171697038,0.6364073527745236,0.5245343987142181,0.375492687259944,
 0.7418254203620448,0.5219766293791239,0.3321875911361104,0.9145464268736011,
 0.5347643818171823,0.8847507236453852,0.9342257239162726,0.7294065085445793,
 0.5967175047301726,0.3744203183313283,0.37263629124377595,0.1750654839522352,
 0.15139435336807394,0.021117020708233514,0.4026681676356335,
 0.3811774717952624,0.5469585427729929,0.4517559449700945,0.43252667057055116,
 0.7798654969728092,0.8063401815068935,0.8083491579458398,0.2366665718140295,
 0.6825190840212105,0.745642562379098,0.26695473003175063,0.9120839669005636,
 0.9200672456060528,0.17919231764721755,0.18023481944064568,
 0.5177722512365245,0.5534048540633201,0.7643711840785122,0.42414878491510466,
 0.9183516251497688,0.6378480210253813,0.6689432377152863,0.2993158876920261,
 0.9050638165111601,0.8998622700304002,0.6988583514054841,0.36725773927980054,
 0.993322420942781,0.1509349932066988,0.9464599343504188,0.9205004129744948,
 0.12292582593675849,0.4765606638611086,0.11055113216520751,
 0.6414354461126348,0.213362811211108,0.6784538381345262,0.12438872453339367,
 0.016435047163302685,0.10146240925727856,0.8570891093776125,
 0.3140959554223054,0.6229511600696878,0.5627746297430799,0.9790382827485867,
 0.7222516046605771,0.3871580844310851,0.33387308976683217,0.7096602515906154,
 0.8765174472752919,0.6641963310239763,0.01746111732355604,0.7111092399789088,
 0.078641344464633,0.4279520279302238,0.8486400491342716,0.039733492010894045,
 0.5972243967600274,0.42484435653764807,0.45116930388821264,
 0.8086698371491048,0.47961259243682697,0.8139523918332985,0.7643802867928896,
 0.3500816776531599,0.15469207407482677,0.2731618298057741,
 0.38203720198310986,0.21331153430325622,0.17667047071471076,
 0.6099648359142719,0.09129171333229635,0.5145005274795328,0.1028975382394246,
 0.14779533398985845,0.4794738746673274,0.5915558472544594,0.3499850747061457,
 0.01682329994402365,0.12925429281439627,0.7932265615991648,
 0.8332727965295611,0.4843828738568826,0.7047996687686263,0.9073075687725757,
 0.47697214063007576,0.03566989066678339,0.009512677794069946,
 0.9413477899100385,0.25338031462685096,0.7390291372442207,0.9814722418884061,
 0.4216658777128396,0.792481427626725,0.40598952755408035,0.8161943060984267,
 0.1337293287619047,0.2752485205827405,0.5983626658933936,0.8455885671795327,
 0.03606291443829979,0.83081860039177,0.7254346940013578,0.3175108171595471,
 0.4522714169803561,0.8147164006688461,0.7324103402575536,0.877472266160177,
 0.8411724868759924,0.22900489627982712,0.08772691844077574,
 0.2889204518881062,0.25131021693992717,0.5543620551983547,0.9968081516588809,
 0.9053394628105639,0.9480659085275248,0.6250927191553568,0.37785743789521764,
 0.9630704618654425,0.4449105254320346,0.7892597563432839,0.09673749928673148,
 0.9942033750496005,0.5027612679160747,0.09804651687019483,
 0.18047731943871115,0.7828393697649697,0.9839420579660374,
 0.31198064059816244,0.6924745186648236,0.6509830254595133,
 0.03192464361285152,0.387448123030521,0.47266973687390856,0.9686937603476169,
 0.5432062349056573,0.6132596036670255,0.847576173976622,0.9686871057488031,
 0.8046000906353243,0.5227466555158052,0.28808651278874864,
 0.45856320750830126,0.12536706665260244,0.3669931012356421,
 0.5431181339832112,0.1558996487337938,0.3561385698390951,0.7804531585607667,
 0.8427606076586792,0.6753483716432291,0.8799521167611806,0.3974000516210079,
 0.34933554333368755,0.3844580007681726,0.25823700071616273,
 0.9355082785633371,0.27976853198221285,0.5983602516576068,0.7144537339314626,
 0.2719495884398675,0.26133394147768796,0.34574552361155964,
 0.5519827600452755,0.3513522465233081,0.001551360139391611,
 0.39330098153799675,0.4931662503703039,0.279178405043895,0.9813206299768301,
 0.22273966664646938,0.2856460753605654,0.14964549225442159,
 0.6941135443945972,0.6786750809137125,0.5612801283639894,
 0.038430364537238804,0.9880579380332565,0.07980440577760894,
 0.9035342768789969,0.48780054059995626,0.6759693914219213,0.7674812461434961,
 0.3493559019606449,0.04786873452661866,0.5995348512212756,
 0.27036897430450957,0.9375408558091788,0.3156756732551912,0.6404410084735693,
 0.29964508790986566,0.46907705443087266,0.017361579019934403,
 0.3719162576337296,0.17193918770129246,0.249243651867194,0.31509811075421557,
 0.7311418580700058,0.6116622273883507,0.49061907728193255,0.7801329812118837,
 0.06337349501724387,0.4865822347924055,0.6809913841410826,0.4854855147663908,
 0.6788625661834713,0.6774928168618921,0.40970957660130547,
 0.30040018258677526,0.3997517718425221,0.27021060386715723,
 0.3564594183353973,0.18896115722029083,0.35576073586522217,
 0.41667139260669217,0.8944900408465984,0.8883906394332839,0.8015497126791,
 0.5095280174250667,0.4100750123598451,0.6038092615350309,0.7551279511393831,
 0.009835426846220074,0.3441352241642993,0.9663893638536222,
 0.7296886589861352,0.657507250604824,0.21172865429153576,0.25015682989098365,
 0.7470031393971078,0.4684599002789228,0.07913840949532969,0.4088996992193168,
 0.5561725848914322,0.3778291991974543,0.55970894099446,0.9655686092900756,
 0.895070313022454,0.4856415230278325,0.04948944666574662,0.42853458947121814,
 0.4500911946193754,0.25377931970766865,0.0037053398057772524,
 0.5114576072933417,0.3634909216183264,0.0783388250202266,0.5152768067765732,
 0.5323320642282995,0.2708293531982009,0.612732124909724,0.028775212182138743,
 0.7603306089134452,0.723033970241677,0.6008420953987249,0.6609081086386842,
 0.33176562387553865,0.7384214740890188,0.748368403553634,0.7378564528651457,
 0.12063868856833748,0.6585654905327937,0.8563745050811109,0.6273359575754042,
 0.8995252184061001,0.06296211998004031,0.7846509956904965,
 0.060461436965024595,0.06210847418508236,0.6086135337368073,
 0.24603677184603645,0.15654352097550372,0.43147651914711016,
 0.018037870692233615,0.7456201290901335,0.9516947900518355,
 0.07346734837822977,0.8094208694698046,0.32694714639596234,
 0.9458225823084097,0.6567111780295166,0.9309347434281424,0.590261105592808,
 0.8676564079699942,0.5366116541447425,0.3269008261512856,0.3338565468250455,
 0.724020510058834,0.08820954456555041,0.6942674027478658,0.12827934157527454,
 0.15761959881026266,0.010806388214867346,0.19848711809673647,
 0.554110094901731,0.23592305943086112,0.23271646173422822,0.8311047145548116,
 0.3703539517020341,0.9884240327161564,0.11126215316377619,0.3922489512385954,
 0.04823877671908372,0.8563484724920017,0.8618769889876778,0.725302409154011,
 0.7206307323529515,0.9955277568174737,0.38797281824710783,0.8148323588521258,
 0.22302254428464185,0.5140410973377243,0.6975726687536761,0.5353660905903044,
 0.974575144196463,0.7222851173126077,0.4439073359976441,0.19615881457426887,
 0.7949635041607064,0.39198494467141887,0.49231197466151866,
 0.6781144626720054,0.8085887718088461,0.04313867222606249,0.230541375513459,
 0.16138095638172822,0.9408242201049845,0.9800706330296343,0.3198619307742999,
 0.14402882662971328,0.5483465803536455,0.6977786572360926,0.654828319727953,
 0.4255379898164664,0.9122574121197581,0.44501613773906745,0.7741673728369395,
 0.8394331927188213,0.9597050237481668,0.03790170282127592,0.3300098359233452,
 0.7707797309196998,0.6865264810663229,0.19658967490068446,0.9580317362475456,
 0.09160288747114143,0.3183739292500518,0.8109572553062386,0.9518468596684608,
 0.17633802410896837,0.06295214564737861,0.7567120870437808,
 0.8291214774394531,0.23146806651511498,0.45616371644154463,
 0.4099999913761434,0.06744238532017532,0.34506568016806094,
 0.9188177589581603,0.010269428710040218,0.5617657099426487,
 0.7753449504582016,0.01331232052228093,0.302017572781221,0.8227374532000244,
 0.026751760802371738,0.5670120806001985,0.5442751377451001,
 0.9804026413833294,0.8874114548779735,0.676042849985541,0.48185744837791866,
 0.5828926578395915,0.485747715757436,0.7757247842002792,0.4866410083727515,
 0.2667465771264549,0.10330181266719896,0.9272326925854851,
 0.12616809662051354,0.5855256657971875,0.4444627312580831,
 0.0046216555345222066,0.9481124141226964,0.8961705858869501,
 0.07787933378581835,0.6636522850406156,0.1750007504325034,0.8171857168731488,
 0.3815680749159308,0.9137994792063054,0.10406207284226865,
 0.021263553780337063,0.9755154400530448,0.9096543462655886,0.368909515207443,
 0.24143082568093732,0.09678586996634075,0.590368029611876,0.6221335500154334,
 0.4626717300215537,0.2250513425345677,0.05581196826338308,0.6195828451142685,
 0.5942917179393489,0.9291407574785914,0.1551604288653765,0.9484740053486724,
 0.8096481154784216,0.27197054617113525,0.9886478180909634,0.9026866818581323,
 0.8520738347887652,0.9215396695670474,0.6169982498761422,0.5495953002014418,
 0.7748373374809827,0.27935860813441404,0.33494046172538394,
 0.7914442434138673,0.6981378486858411,0.5006129067354268,0.07627676113581205,
 0.9334292186118288,0.3414323857711754,0.02117024520148425,0.4231568586305672,
 0.5911657196339158,0.6744243847725058,0.35235889049908564,0.8690593394478383,
 0.23311134125711497,0.8577371002678572,0.18518529061461209,
 0.7660366894790838,0.2510539476217277,0.5251794352863011,0.16666125561693768,
 0.33227346768863675,0.7559049404286277,0.09521695079488945,
 0.030947297278149266,0.4501720592671198,0.5873408194874046,
 0.9981034618568385,0.6507816024941082,0.8428757637832208,0.9697090049537784,
 0.821982034126994,0.6793715599478295,0.8672908233003866,0.5953999241650736,
 0.6844856645179115,0.2546936100288324,0.39442037745786374,0.89065810414697,
 0.32576910544003845,0.4970402996673191,0.3358958631339789,0.2928639188622282,
 0.6780666436626854,0.5218071091984395,0.8483828331125527,0.3697000325278782,
 0.0013905529412223672,0.2150354882029667,0.9200520088337647,
 0.32548774662784996,0.10929931403629933,0.11319060469856268,
 0.979995255134333,0.021058330673298542,0.9424406663164021,0.4468036028924105,
 0.8189830243666973,0.20655036147470585,0.9800146842897139,0.5144397355040867,
 0.9066391050537311,0.28000370009481246,0.3364387098442525,
 0.05634169894687835,0.4820687634512426,0.5150261949079337,0.5944677408998437,
 0.7097381416498726,0.5558953797620415,0.5571871564826747,0.9660806662558857,
 0.9205018024680451,0.7859419571910538,0.8877485026893867,0.42294405193568263,
 0.8149461133226332,0.4062909941027675,0.8971508281295454,0.22275954404500065,
 0.4399880438644961,0.6247770764461179,0.22331975629587342,0.2590439500134819,
 0.8082428533041652,0.7193166205879229,0.18213663429450444,
 0.32389042923845424,0.09608254692223361,0.221452695705187,0.2918687329890136,
 0.5155409933614342,0.5947793992419413,0.01884502215003636,
 0.06824136286966087,0.9700922301651954,0.5687782639294496,
 0.43557683140011105,0.009006327429223582,0.36991198712374296,
 0.6133397690602038,0.902487654629346,0.38104317693248635,0.07430188047455288,
 0.03461711075085716,0.5546757717141733,0.8141006602303027,0.8356333264114177,
 0.4058693239297885,0.34274275151891076,0.5345443019377066,
 0.28115575353073896,0.8184801914210053,0.7246872554867693,0.9152369796695672,
 0.7765653222971046,0.3687095151831685,0.9283485899707081,0.3391694628084089,
 0.9949151015858749,0.14222599246387435,0.07594170057328109,
 0.5502868567864514,0.8559599656803776,0.13573661744734844,0.2818848118325429,
 0.7635739670155626,0.6246680973386879,0.9369701966153207,0.8178576098730843,
 0.9156363115068645,0.9831779345239996,0.9924013219929035,0.3964318211638197,
 0.39509564827680155,0.9482681315664262,0.17272893981317883,
 0.3138883391507553,0.2225227660028164,0.9178657968500861,0.8552646686505245,
 0.2586609350278781,0.1806030113219388,0.527415390359346,0.8421770427881161,
 0.005325190158304016,0.8802701732366711,0.5942743919161046,
 0.3535279709709894,0.2704613421368255,0.6566053931643492,0.44879669038913095,
 0.7051860882665794,0.6763134457750795,0.927551014721794,0.18506780442957171,
 0.17367150580170798,0.5330180725128145,0.38426786996994067,0.496613841292181,
 0.47965529633912474,0.7545097305001159,0.9095385251960751,0.7453882130591195,
 0.8391437959561545,0.7372724696682154,0.8531558695871871,0.5701654916041216,
 0.7723296789789836,0.7493530499922105,0.31831758450681313,0.2998729238866368,
 0.8559926101409224,0.24638579710825503,0.03888915469016452,
 0.8193621203001709,0.24324897512239452,0.28723897580608304,
 0.37978225693719525,0.9207680728701755,0.35369049379581596,
 0.3265576217317683,0.5791048449402137,0.6125820632491978,0.46801694618899314,
 0.8457870821580865,0.35215191773269716,0.7610700960397762,
 0.40879158849793906,0.007853507283755334,0.02385748669555854,
 0.4070289254628393,0.42692960000342717,0.6066800450733193,
 0.17240597879435227,0.7310427595030138,0.22694564884125212,
 0.1933085729251054,0.37966982231935664,0.48870560765848114,
 0.6838604905451984,0.9976321941867674,0.23698102515219988,
 0.013195126906281107,0.0641001947221026,0.9711488252011085,
 0.08125694110622916,0.9998257499954459,0.3071633985560207,
 0.26599617282698673,0.6502210689813532,0.06331687441614708,
 0.6408175195181716,0.6920882972106694,0.8273793750975124,0.6460867126621244,
 0.5552953927506314,0.35870839428064916,0.22145091099780045,
 0.3788969946150138,0.6596370725993047,0.5117199299174437,0.8335651586008618,
 0.4959688299883993,0.145805800485313,0.49492564255868254,0.9143596770097964,
 0.4535567159881624,0.06774939985972361,0.787511480253876,0.9590092615901862,
 0.8242419502726852,0.9987433240942124,0.42591104442979255,0.9686423130718161,
 0.6557697518396071,0.9286511368889545,0.5382547470290588,0.5845889879975663,
 0.9133584500478139,0.9245831061083352,0.2328644152314081,0.8180963972293324,
 0.8380977127868959,0.5629326225230995,0.4746691263642713,0.8975138629533349,
 0.7491411397998416,0.543819622223436,0.80971515634509,0.20516117717457483,
 0.4121738415620323,0.07318601086992127,0.4392961966136094,0.4389859518152872,
 0.9063780613739859,0.3787274790075337,0.7434831320990252,0.2990976905061037,
 0.8832308395548414,0.7772707951309171,0.600665333296375,0.7750912841287648,
 0.5897115416953542,0.8034538654890101,0.5747336449401984,0.8254562834937365,
 0.35226068971654123,0.035932224090350684,0.1853908305854356,
 0.9972374266783491,0.6512219500928137,0.19805211924819266,0.8218121343124047,
 0.4716814793116628,0.6944838332882012,0.16777666359082233,0.6616299001345763,
 0.8594198129031625,0.5853681421486724,0.9171710218368608,0.6607656551884135,
 0.6815321375211497,0.25892087267447317,0.09915620473827591,
 0.35494735560085333,0.16097146366836323,0.3553505720035679,
 0.6207691444870795,0.8107287833391023,0.8192445638857233,0.06965281409777635,
 0.3083658297319942,0.47539551299363614,0.20835813270519377,
 0.7271968153065704,0.02182375063797859,0.05419799531635461,
 0.9102654268697907,0.5846837969715138,0.8587628494836249,0.6562710870651673,
 0.07587957939239565,0.5959049053344678,0.08940862440414155,
 0.1471264149138487,0.7463115793893057,0.6470998415220718,0.696865271036116,
 0.4410728813436938,0.03401416085765163,0.8321475047740704,0.4512618081603137,
 0.6165340141044644,0.3926392275977093,0.3488889430907054,0.9983064082747748,
 0.08608149077737759,0.5796138194224573,0.5154589290224095,0.6901620562977069,
 0.7757380268994863,0.377114928328353,0.5467817287832661,0.24129113652982337,
 0.7929702983544499,0.17458096088768005,0.47223360042291884,
 0.05867390234641667,0.8255392230027582,0.2505881182240688,
 0.16772709302598787,0.9591232485600314,0.07897365676890034,
 0.5817190372184147,0.9870526986200611,0.37251057217686434,
 0.060066709855845724,0.9882949808977695,0.1915610883871608,
 0.9833370763588256,0.6751760223306242,0.9949041560597378,0.1797374256727955,
 0.9797459178530616,0.3946335206497408,0.4129318584044319,0.8254737576501575,
 0.3588223815399525,0.6015844837019556,0.6110352536093154,0.893964349945375,
 0.7561883718082262,0.5033153717408141,0.9569994152905644,0.7804075653824393,
 0.045639509234859044,0.4448568239845502,0.1461696013680862,
 0.7672443090575001,0.4381219881595433,0.5203237987358007,0.7005962765913366,
 0.45752914509082776,0.4573255833186324,0.5327811883323276,0.5741446755826392,
 0.015245256090558157,0.8616472147533045,0.7499436850476916,
 0.4033287397156533,0.3793765488671623,0.6660646575242646,0.3127815321642482,
 0.6072608905719561,0.07690958050805063,0.2218028388830524,0.5577439964114977,
 0.6611113922256351,0.4593743440743272,0.2417934806635803,0.9720273169683961,
 0.7943720425945241,0.28963059514039013,0.640079864596177,0.9581027742346185,
 0.3596994980093785,0.09808570281600071,0.6091640864936934,0.8473262416119811,
 0.7744357130959132,0.2813202090863287,0.9258520330043927,
 0.009751954624815262,0.4607429549739448,0.7961582401876128,
 0.20786001717118763,0.7649500806362384,0.3018259457271657,0.7263331355414384,
 0.8155814789456687,0.316710282961034,0.4647297027901559,0.6869123213270754,
 0.9048648063801548,0.3006440608437474,0.11584267411216698,0.656951781427133,
 0.275602980186919,0.8407400525513467,0.719255729678232,0.0652770682358681,
 0.3034467289443553,0.9264863448573871,0.6274094178327926,0.7929245527106671,
 0.6297004656599196,0.3745022032867573,0.18272761119274805,
 0.13974830919892867,0.7575432469108401,0.4393169843918139,0.6761637863126404,
 0.7048694972346217,0.08949069055615155,0.9256747873379652,
 0.17353164532270804,0.7274807669223959,0.6879100333193187,0.6386779055104257,
 0.17041537823638364,0.4083697608976393,0.3079443303870095,
 0.18825188336048337,0.41420849887134414,0.20248992311001046,
 0.22170700982943525,0.695284740443636,0.5376266377643802,0.25480613634397775,
 0.2667497133599157,0.43223289016191,0.9807376522030773,0.9685161222720499,
 0.9591163095244599,0.7141748074205277,0.12403572108853878,
 0.009931084450182204,0.03930615081106048,0.6535437865578714,
 0.7028557856787127,0.21356613542271607,0.03496360633979567,
 0.7654563990097746,0.8326021998187991,0.3034272765660019,0.9499651378971463,
 0.29205313854075043,0.015125306011151252,0.8699985894441665,
 0.7634249195259549,0.7198084993307697,0.8795693948926769,0.4187213236535585,
 0.3715196417844896,0.9749530081337953,0.4586889938006501,0.1423423694846181,
 0.6595334097693447,0.8348430811250687,0.22320760952993635,
 0.24257979432007637,0.24796785156182888,0.8985108167188158,
 0.2774849500180434,0.6388691140804794,0.7945815338506496,0.7508874922575117,
 0.44697179697561595,0.8694209334728699,0.004981216071410777,
 0.946606222614869,0.10616781397069386,0.49967775663925573,
 0.04948856633578558,0.6660101986897176,0.7052190636041683,0.5508931986065007,
 0.05045504544247814,0.8826822611025537,0.7754381424551741,0.7111304286482045,
 0.9667532936778208,0.49997959174412354,0.4226036333077938,
 0.20024144428202062,0.9047589466363647,0.6552318588655196,
 0.40561709056314177,0.7662502131535105,0.9508795059748343,0.5367508387073892,
 0.6982658470561591,0.8079621116846298,0.9674666864949649,0.5921600079223832,
 0.8248923988040398,0.6399428912356474,0.04719098509807418,
 0.31644148499878244,0.5392153331912118,0.774597358047174,0.4968483864806752,
 0.6380727393636524,0.168426714909925,0.5814505826950583,0.19734305632286575,
 0.7651590632567222,0.8120904925164061,0.8586202605778358,0.00795050870630376,
 0.5007955570124805,0.8396986666800971,0.8981861562543818,0.46445810287893907,
 0.02109469334262859,0.7465273005352535,0.4277847830419348,
 0.007577477510910535,0.4241062985332039,0.7821109709217375,
 0.44988569304717796,0.6778360498387812,0.7601838480480219,0.4867330391677269,
 0.6352498701199758,0.7906916627350804,0.9080007453982011,0.4611801623142361,
 0.5971851760610815,0.615724058704797,0.7877865314017259,0.6917256680812851,
 0.4398776962130482,0.2944628902960583,0.22187559870206852,
 0.031889509508376546,0.4680771264679917,0.4303394616073233,
 0.5332538259514972,0.3947869701757548,0.08085903990013699,
 0.13899742189116937,0.444986850242449,0.8669072405054447,0.2580908605314248,
 0.3529131470740028,0.26859759595162824,0.2217258517794165,0.7930175162736774,
 0.8725599644587975,0.6529565928562235,0.8608947685431071,0.06894073116249688,
 0.5740165799511171,0.8551686943720982,0.7917844784939325,0.5332430427137779,
 0.5202873411689188,0.047798924505689344,0.07750179747554653,
 0.18478665528444627,0.600040792988618,0.06128630925693512,
 0.03998837307480263,0.9312268793760092,0.2017077042631955,0.8038644240374423,
 0.21116509838326858,0.08015328882803763,0.7403791495083766,
 0.031115357433900925,0.667618415201634,0.4027593335514502,0.8544205936438594,
 0.39421645560754004,0.5956345224311218,0.7919792550134499,0.8966909563783517,
 0.4567154340383214,0.2454005372437742,0.839477243828528,0.03610909017574082,
 0.685725935969557,0.03468149171911117,0.39880221279016936,
 0.41245875465776627,0.5318849859827944,0.835174817390765,0.3455335538867279,
 0.7295668875453146,0.574492542887517,0.5724992186206349,0.6860247777538727,
 0.8473259652481997,0.8189106167250653,0.7229768026640214,0.30571368478735983,
 0.152809425609859,0.3221040878290723,0.6064764488158423,0.783662678622763,
 0.24408316380447015,0.4538790531424567,0.8397872180671698,0.8440425983741192,
 0.6892668875155834,0.4179687169931685,0.8455503433805647,0.9761362375158725,
 0.12223035418982031,0.7627968820924117,0.37862041858200934,
 0.4871954709135916,0.9498244505195945,0.4334147403172205,
 0.040116171383861476,0.00207892786131747,0.12122281651852684,
 0.21093863609723418,0.6548836766548455,0.5399929102185616,0.6959869718472756,
 0.4103280535435905,0.2960901953379014,0.5702506956434484,0.3059740614648577,
 0.37571935255360767,0.5907837579837527,0.5918033467725774,0.4103440214249373,
 0.984180703990119,0.06136787765533158,0.3714403315449145,0.7031518286170733,
 0.55491828545508,0.9673141724578135,0.9300814859011466,0.6728272989559282,
 0.6581993252422182,0.023421532477778784,0.16091934119503737,
 0.7144858883806324,0.24586225748729706,0.09740353761283205,0.296122320184417,
 0.536653438180182,0.8380720540753755,0.8354582116733877,0.6486810155652711,
 0.12360073303564578,0.23596612350767487,0.6989205594877379,
 0.9462333757314277,0.6797019693016875,0.6026567652330663,0.9086755562196076,
 0.3670547652630418,0.18545656125963406,0.2409296153446786,
 0.18670733331337097,0.14292197457476574,0.9233583118537035,
 0.011720386663579219,0.17978116651577591,0.3190056800870732,
 0.10440454706630975,0.7670696216245412,0.46409620697563525,
 0.5053091933619502,0.8785592988417792,0.3795863580359432,0.03532652254251878,
 0.9347521073130213,0.3021943675177574,0.7796166547499598,0.16563334396087992,
 0.5508042315378683,0.24758162414089835,0.7410292037969395,0.7933482932431418,
 0.914194306793747,0.8264792349725059,0.40006219045745117,0.61514934259184,
 0.6766621327225762,0.6122435580158656,0.8846728660683165,0.4692716232577169,
 0.6488871889190166,0.8600362232757185,0.14007325459360265,
 0.33063736695057555,0.7092475512605435,0.19498153861142942,
 0.9550386028939565,0.5183799240878185,0.598135624063935,0.22563928560723512,
 0.48337381106830013,0.9826052166423984,0.7456832682668676,0.3064910146750871,
 0.19328926154845616,0.8144692772656024,0.4223120453474142,
 0.053240515264504906,0.09073050555015028,0.7469173589742508,
 0.8439740219376111,0.30314382498701997,0.7111372226150674,
 0.47749509891336184,0.6898764885903301,0.5977724007311112,0.60753516890568,
 0.9663649496980812,0.29492730769804143,0.3184985169032353,0.2570844050662562,
 0.5615455905258032,0.7608579042687367,0.48045218938517953,0.9011045521629213,
 0.5763311824416595,0.9092108706651058,0.6520264087938754,0.6460381380855938,
 0.1832877544315754,0.42773701379742923,0.760704209971408,0.7648664635140002,
 0.9691378805145454,0.5750621343085415,0.03030432690616003,0.9914013471409769,
 0.11683911541240155,0.9316070674198794,0.7105320380252395,0.7333817307966096,
 0.4068253284705424,0.606165240328038,0.9414418280696527,0.08564790389790544,
 0.047762793004777926,0.7756377682246988,0.039039455085359664,
 0.618815174817961,0.7961063063359288,0.8693574531234715,0.062403785781783405,
 0.7715087919759797,0.6984936070453397,0.24083621711470438,0.9597233224433313,
 0.4713001059306222,0.5522509064134254,0.7566500893860921,0.646990214478531,
 0.9414127534689437,0.3311916854384911,0.6056529859207148,0.7936843600753021,
 0.5181725980113561,0.09070752971502205,0.6050870793615747,
 0.36358850354442696,0.7171409585815134,0.19677137285967605,
 0.5315601720454846,0.9453125838865594,0.33003580724483095,0.6242926775472458,
 0.6157828579302527,0.8593598349225147,0.22844887942341607,
 0.048037126342721836,0.13737895002835732,0.5279250958921877,
 0.09338997123163817,0.992890961044345,0.48242712055248604,
 0.15598188757882225,0.05262813717539161,0.037634053161707204,
 0.9239713568092481,0.9647300947992997,0.2133392146169122,0.43515854637485407,
 0.11204013509867494,0.246615142211285,0.3235126996335945,0.10309449526794878,
 0.3137781561207562,0.25221109579358314,0.8506920704318243,
 0.11435472119115575,0.7671408775759126,0.8766192369587383,0.9206536693077265,
 0.272057439261956,0.7227201850160717,0.2872711367688733,0.7120269003580614,
 0.8704818973296813,0.09753532807662535,0.8262585749006875,
 0.08744184906325558,0.3676054767052648,0.8228380180424552,0.9415123961143153,
 0.6023538693539896,0.7299803495539602,0.09209384583005031,
 0.08976622502956544,0.4455886958621599,0.04679702151813003,
 0.12751691150399824,0.3001226808492242,0.7620923639593802,
 0.18945101634538064,0.4185438185333884,0.9952350096872802,0.4591493782511883,
 0.07725640063497219,7.773162193728256e-4,0.8152335477181167,
 0.9800305443079746,0.729277095477811,0.7172965458623323,0.5048142203388011,
 0.7784052280511559,0.39248291561722204,0.5679428383800212,0.4868500996418329,
 0.4847932140146173,0.4979283656818727,0.07845522536075245,
 0.29244609074439865,0.1739503335493623,0.8925242399855573,0.5359636540199613,
 0.4113644292313582,0.01876271655677131,0.3019507179336165,0.856073572149194,
 0.9220429379404694,0.2883101658384193,0.870294868860124,0.9607760959096092,
 0.9040943263657508,0.7109965904899155,0.26394223574446984,0.5013448717726412,
 0.9423975191454994,0.7185751157883867,0.7719040184395477,0.13994886681409546,
 0.5026312880714566,0.41614582973068825,0.39300709824696156,
 0.6814465219219672,0.011531536966743805,0.6264378033522455,
 0.1065582987982584,0.08122564588803138,0.19545001065884127,
 0.11945889585169933,0.907923147793287,0.5024461163144309,0.3219538586128806,
 0.731330024972473,0.6940909042015798,0.5003723440228812,0.7360657545290737,
 0.31542321387633143,0.3295977053714577,0.325405580525016,0.44823368101990935,
 0.4986125188567696,0.26577874898773535,0.6964115849949022,0.8448775159505901,
 0.2507519407552139,0.2505406033268468,0.4794468720415326,0.27815652321536344,
 0.5563462607363907,0.696999252053586,0.8800702891187024,0.011800148926597087,
 0.5839022198180848,0.007160257831444472,0.21896220535184352,
 0.9023184090073957,0.31354453543424,0.26975849234060467,0.4265200334939331,
 0.19622123277419956,0.16885924223780058,0.7434890345163183,
 0.4751356311565692,0.9685719902951064,0.20344229450974471,0.6022905777199647,
 0.8727259974320124,0.8081999219382902,0.5601041447385121,0.5249830117228027,
 0.011141402857408056,0.4625669027121687,0.6561300359411972,
 0.9790138496186205,0.9177849810828137,0.21837553149021427,
 0.05669153705185437,0.2987736601949481,0.3678424674853904,
 0.44277834293209695,0.5980421946119757,0.29535085388625604,
 0.3392917528335434,0.048343839756475715,0.054173556458925076,
 0.08388935698550815,0.9503376600878806,0.6659471098669519,
 0.034580511920949464,0.5226222275195446,0.2779795345002093,
 0.7648097581544407,2.8155598465251508e-5,0.126833563006718,
 0.0384820342052955,0.5523036698215232,0.6933866848190313,0.8277278015997646,
 0.2709771496544777,0.6139430370830604,0.5930238711668488,0.5296146357553421,
 0.10910162104794618,0.04938475758122274,0.9053909852558324,
 0.2814185337348707,0.6564548147801902,0.3552490772708169,0.913755498483353,
 0.3377728380390421,0.42477296505551043,0.09186300314323537,
 0.04491714078261533,0.6947985752722894,0.8041467894587409,
 0.10303373484208433,0.5047807560722036,0.7321458572424009,0.8441880185683208,
 0.399102452378969,0.11507381374507475,0.927934345195233,0.7894982529334319,
 0.3187604876428485,0.4300061417473585,0.947704422512478,0.9946946771843582,
 0.26579190386525775,0.5619902913469035,0.5745918408965136,
 0.06628296122651944,0.23582843317039626,0.936044010619751,
 0.31370104880448013,0.4326399290432206,0.27278365099628277,
 0.41554350008944185,0.33176056587870495,0.9217975651341866,
 0.05379173645947821,0.9211709502826628,0.8994763476810318,0.4244912692134447,
 0.4983189349881585,0.575829371449428,0.41223867730108654,0.6742735381663268,
 0.5796474576364757,0.173068809131667,0.2036174057648703,0.2087612188452267,
 0.11491713726475983,0.7895393314703897,0.6589952864732477,0.5086666015644501,
 0.2055553808995484,0.19001813055784533,0.8143342861974936,
 0.33645178508457885,0.9364305791987159,0.12237984406313918,
 0.1414966983119753,0.05944638694168247,0.5574847129384272,0.9013129833335943,
 0.8398162849297921,0.09618912955514813,0.5772593019625712,0.4496601407270008,
 0.28464510298768464,0.29695687157708894,0.2553274071242283,
 0.7108258148060526,0.22030708377606523,0.1280414547003209,0.8482425965896235,
 0.2604123860426836,0.004335119075125071,0.9827084751455567,
 0.2518600455990224,0.8839467846876001,0.5703525642391408,0.3369564076671654,
 0.3650913771714458,0.6638304624571074,0.9686139435146218,0.5124232441721507,
 0.2636687398013169,0.6394499524678195,0.44804746005642415,0.8741496656027118,
 0.2494985938116312,0.8215174652157148,0.7286768128936444,0.544359868428534,
 0.5251276537657861,0.5622274276668385,0.05526317819652671,0.8011815837187225,
 0.040595191939627684,0.6644394825747623,0.1467202824145475,
 0.8326796204835599,0.48548058534086513,0.7759488916154123,0.6011301747161056,
 0.4858194995078524,0.6951596198478469,0.36740854734398476,0.2191586286349052,
 0.3155327694684744,0.8936738917945355,0.8118224773541276,0.9774289319441969,
 0.8377118115308562,0.18495468002182858,0.15245162715238725,
 0.7090024284590715,0.7355095730770753,0.6576871361728123,0.8355850034022251,
 0.9700604663165829,0.049092668716465226,0.8510912540169709,
 0.8805190701604875,0.7085668887423993,0.9553961629927659,0.39872542342345296,
 0.24548882897276503,0.8606030526401283,0.1492302584047952,0.3757754824138795,
 0.9835894076020242,0.23883531954546366,0.23988850610643542,0.914211517412657,
 0.7257510692434586,0.6223135335724888,0.36090619602860485,
 0.09494783279538421,0.6518136884360323,0.36180939841321846,
 0.8802634702492615,0.07832209866271489,0.6933902728968808,0.6171667200801437,
 0.23557517346121215,0.6902810283755408,0.5645849999861088,0.6323710790535928,
 0.8318923223500895,0.1863822812239011,0.15076797011975152,0.9662037483548049,
 0.9374368880910194,0.47649350616723485,0.17352782672010025,0.762689325970873,
 0.49359969395898406,0.7244615504148686,0.5748163902397334,0.9241549532058247,
 0.3752381932478135,0.6772028201636601,0.23576542306313109,0.7248545877037389,
 0.13978209848833378,0.728804601373349,0.20177165131237695,0.5502090687837922,
 0.24015581065613356,0.08279044549285164,0.9345492085970193,
 0.023029235899939104,0.8963476898336362,0.5007057802053967,
 0.19041374245360387,0.23487096218840853,0.38376924427652204,
 0.28181926652470324,0.7624315614311268,0.3127042073457128,0.5141429047969017,
 0.21613325800547634,0.7813214114364633,0.8006295716850841,
 0.13435982218929954,0.8852349168558509,0.5241008595218446,0.6397316774574083,
 0.09137476997882477,0.7573053878043097,0.6178384869407949,0.695121031572929,
 0.25498973771817,0.1716210778046674,0.7370290295202518,0.6626803304131901,
 0.09543245187087768,0.8838724877120361,0.49622636503745876,
 0.19090565953115002,0.930578778057596,0.15139654780852696,0.3353652228315779,
 0.10920068234959446,0.2775053562868428,0.251177376811331,0.11900038259561807,
 0.6572258188097122,0.17345009360477115,0.8883006096191771,0.909549806324894,
 0.702787409999685,0.04713863314990441,0.40117723171827047,0.5772850831079629,
 0.40723051580241965,0.3697666881355274,0.611250988752771,0.7695093519295122,
 0.7936527401612199,0.6173470905367933,0.1731028057809354,0.42582754244337073,
 0.8647998097497993,0.5407824131919233,0.5337191547499228,0.39503500690269233,
 0.20695106157137366,0.8881299535617191,0.5747124929065012,
 0.11104579328042985,0.7270704488463183,0.47578212238095174,
 0.9072897245905873,0.46415024128051574,0.666651223501612,0.08023882078941036,
 0.9201825517663438,0.6025397672364903,0.963958542369724,0.5929914276477449,
 0.5553896642866726,0.4047932882941907,0.5190151832029717,0.32960615554752515,
 0.4645667926746999,0.7756792647539575,0.21102678948102727,0.5414546014047981,
 0.45309755019915077,0.8965687643812621,0.922618867824583,0.04700246178952194,
 0.5383140678616168,0.2949568885966032,0.2356262275773987,0.20921151588449738,
 0.04339518646898155,0.819437218194178,0.29088348038169354,
 0.033536282993049005,0.1566138628148258,0.15524231804670852,
 0.020958784941126796,0.9634204188633195,0.6730098009074655,
 0.5550838954533138,0.9052468750228979,0.016779143911938688,
 0.7232581667047863,0.7210237899064704,0.6927537403686198,0.4667760801691836,
 0.5662371227516325,0.33610421294755854,0.3842950096265386,0.5056706148843315,
 0.28186970288887514,0.005225882805305004,0.20600890673894867,
 0.057229001511315825,0.47826551271598694,0.07195442418525255,
 0.9039022108781882,0.4255644562129388,0.28999038073682726,
 0.49455650744316215,0.5371922744794482,0.6895294290618638,0.1333669818699923,
 0.029436265313769083,0.337574827020926,0.2989529851631719,0.945660959766688,
 0.6578679119418391,0.27620500168227036,0.10327777092346002,
 0.18753637399667344,0.2512783137960348,0.414023936305137,0.10075504687819259,
 0.7819307037849641,0.5467840885248638,0.950841185091639,0.8444435088927338,
 0.32390574210847545,0.8485425620789874,0.9163529347320188,0.7043770498244634,
 0.906002124529371,0.34685673919959903,0.4426731588506483,0.37783354863639784,
 0.27972625196889955,0.06269767413715543,0.05553716371888873,
 0.09138171369465065,0.8252731257821005,0.7221671435938763,0.8107861546815368,
 0.7845066771015398,0.4973163050297408,0.01791631653301906,0.8104905463774142,
 0.34346787266746537,0.4892147919607869,0.5186848818372365,0.8771798809539095,
 0.13592646844596123,0.43964612899974764,0.948884851920864,
 0.022827411632048378,0.497959085718388,0.46344158724638884,
 0.7997027701103276,0.6858241771035268,0.9387211987500923,0.6740858859761121,
 0.84001237663352,0.8428536928130277,0.34979655687357436,0.4793195443607944,
 0.7769069531120378,0.3946611123878978,0.7295047341462129,0.9803870865587945,
 0.7334603190941978,0.4563232434089366,0.8023631080393313,0.9674956551454567,
 0.07883715207616859,0.05605880851140066,0.7612708742097379,
 0.9917481182555679,0.7367940787260246,0.5384073269506154,0.8354363834980002,
 0.7461023345748659,0.008208515632277313,0.9902386232605607,
 0.5470505539476679,0.29779581252686527,0.5735085223719282,0.0724665144135952,
 0.8829363510620118,0.0272254370972419,0.48524175942707237,
 0.33762570144851145,0.08977204786931736,0.8755438120927381,
 0.7002777500444242,0.34227856040187443,0.7908488855774795,0.1928899856446833,
 0.7243529213161444,0.46715754333026227,0.29887645915874306,
 0.33001018811992966,0.8002473099335286,0.8300987029146677,0.2701842139381463,
 0.4627195051213049,0.3060654101170688,0.41101170832069167,0.3619611097524038,
 0.886319573926995,0.14527121493403028,0.916749294415333,0.8677111226210759,
 0.88322638939713,0.6374401512758474,0.700965179296436,0.7582026464522796,
 0.33812623879055914,0.321427820946262,0.8004583353439841,0.6947358921472413,
 0.21358762660954111,0.05261829132623963,0.2500414304568561,
 0.9670788017136975,0.41368645361143475,0.06528167815798636,
 0.5212208186854668,0.666701859097347,0.7720605810274752,0.6005739852616242,
 0.5480173238047491,0.5163834196162507,0.7719689070540229,0.6009362012221776,
 0.5626327117255016,0.931715127076127,0.44365918677357086,0.6212364965598838,
 0.43330184421600604,0.14581747762201136,0.32771767699818977,
 0.39719794397433106,0.8830811252175647,0.45023087523315275,
 0.17087863562455186,0.4647275209069859,0.06512299933158727,
 0.8145695882358511,0.7768373793854306,0.9924313166669114,0.27967473187331326,
 0.9721103277318457,0.11014488469657369,0.45588388260309043,
 0.07506696359834308,0.19393293791901933,0.007848029303744841,
 0.6737479379048505,0.3319863311739688,0.8542112898569558,0.18754109229111215,
 0.6363719106439314,0.17488622475148585,0.4881699278916748,
 0.11370908756690534,0.3121214154208496,0.945994060231794,0.08567473022816197,
 0.9400317524316856,0.9141604520653228,0.7679217472910728,0.7621305476285188,
 0.9123696152191263,0.26190356233232437,0.3686131606638975,0.7494976082819378,
 0.6277923016033619,0.14839469775773884,0.24003825549810076,
 0.7466246452211749,0.01809266570878565,0.9987379664250833,0.8189861188559063,
 0.5519340051921718,0.1812396736777433,0.13563766772978236,
 0.14872382368073178,0.23762921314466856,0.32197342495961045,
 0.8564837354270077,0.6193340561778125,0.6251517150546242,0.7082288968154824,
 0.09412560944202575,0.591663816940686,0.701278072896178,0.08266515933406893,
 0.05269383283404039,0.08308466546315052,0.39842362574837475,
 0.6951349367007752,0.0802313111514601,0.15283561826390368,0.3873651675433094,
 0.09172416042558251,0.9994556402047725,0.6811983662293446,0.7056121379585258,
 0.4104110596543662,0.7304491210735184,0.027942534591244517,
 0.5361118639598388,0.3517966986789637,0.4023445139725512,0.12939222743375556,
 0.6172942549122101,0.2656499008113975,0.8276200712060188,0.7842347629285754,
 0.08858907011654671,0.06185966182457514,0.6918593814601623,
 0.2604471613126721,0.7378912819573594,0.22083271397731552,
 0.05911760456436277,0.218054295564727,0.06412419530680247,0.6762164054693278,
 0.2854430438594625,0.2803679475153704,0.23529736558070513,
 0.08528249523713882,0.1301609616779078,0.3211506874807538,
 0.34672962892281256,0.9881017400462164,0.09599658750883655,
 0.8282577073623356,0.6891663234317688,0.0029665130976493925,
 0.3061050861156286,0.5398361116106201,0.5641018572534086,0.6993207476755595,
 0.3501477227752605,0.05126349665833407,0.7528060470382064,0.2651880586355285,
 0.4437865456822916,0.7155286192782521,0.5625676915351157,0.6666442580417277,
 0.09468192521046237,0.7235556007310113,0.8780623886005714,0.5681898481346676,
 0.6581127616554708,0.10175775864453551,0.8282018996055605,0.1916302813302978,
 0.38734845009158947,0.962189728355021,0.012589902732910385,
 0.41867230311468806,0.7301564490235605,0.13551242802308405,
 0.7145542688896583,0.9336421970540096,0.6560667056053018,0.569195381809493,
 0.7472153889303063,0.7759616295428988,0.8898975708184006,
 0.051476239277140845,0.4242491425121315,0.9672277845075221,0.742940011022253,
 0.7068026359282599,0.3143424153940996,0.4434743669061304,0.19836986742624285,
 0.657002883679973,0.2323659423641079,0.9638343746242958,0.22518045575178514,
 0.6319165339790043,0.052593756075280385,0.48684262194179584,
 0.9107240602932823,0.12651146965138316,0.02641623434862639,
 0.8910365305296166,0.11700698365550466,0.9590419541714982,0.8377216849333142,
 0.5082935647782885,0.12238851147973062,0.2896096818732704,0.6568831976433536,
 0.014018045886661112,0.5881641464466227,0.47811406601081363,
 0.48242875599249846,0.8839709040687426,0.4460034974728402,0.8194116620577327,
 0.9377970340438573,0.4372874783375811,0.5752146470098111,0.08772820974920237,
 0.19166143798607904,0.8969519158842214,0.6242862813296002,0.4714479619653077,
 0.4150573962516878,0.9761585733812306,0.46475993026894113,
 0.18721411589579662,0.8933866361335994,0.21507846414297804,
 0.7464141785724204,0.8252795258796986,0.21963682103261772,0.3963390547263742,
 0.2100556501713663,0.35861775835233345,0.1407801333296521,0.293304740449562,
 0.9549925310024562,0.8517878883960788,0.008588868629548774,
 0.3944287679717482,0.23368451370497345,0.29092103877673525,
 0.47978883551131624,0.14444997330610854,0.8849772980910995,
 0.12663351315617088,0.3489356281084047,0.6116887680650616,0.7976610724489355,
 0.40708054675785976,0.6561914287323314,0.7712694910302809,0.5558458927080728,
 0.49588879246833095,0.9112707956928107,0.3493680748515873,
 0.07071959278604611,0.45704307549926504,0.1763929632982948,
 0.09411973631459558,0.44506531259649895,0.44389978143402875,
 0.6512029558558458,0.2618883704959437,0.36468903011877574,
 0.11691730229048969,0.7201472330868323,0.47518751233218426,
 0.9776324355205009,0.3327480644231209,0.593660181033717,0.7051522641453716,
 0.031702002337416935,0.6530137271717854,0.42444942649850514,
 0.09677542365861647,0.5011937540219905,0.6086650786527792,0.7064520786240287,
 0.4858802185367246,0.3000007253301723,0.3202504550175179,0.4918958156264426,
 0.415658445676796,0.19065033500211326,0.3659273614674512,0.3478256760360966,
 0.07522401393768552,0.07544184829477407,0.3909569443711678,
 0.12982651324762717,0.43751168845584765,0.47426346801070074,
 0.022567163092721443,0.7821684541901652,0.21887473803928037,
 0.22434640676392004,0.9146867129186332,0.6844694594669144,
 0.13047812768585432,0.35686540523979826,0.07602511547089918,
 0.6475379202621775,0.04473893631918968,0.9426112452548442,0.9575077177528453,
 0.2721379209510776,0.9578033116169089,0.7756220430578149,0.6900026993726411,
 0.05236448596710841,0.459425652365013,0.6487876208742683,0.4563970271762119,
 0.2565483711955894,0.5050811302043906,0.14200678333360217,0.3861362856659505,
 0.40899296636426874,0.21596963961072402,0.8772604302507103,
 0.36260653338716653,0.5202970992944965,0.88408950349158,0.14520539388106646,
 0.13927113122695467,0.5753307685259548,0.8565065969538963,
 0.11034392346146571,0.977075937384396,0.8992168210526127,0.3237194660706487,
 0.20725607685371494,0.48119053895021147,0.6233068706207912,
 0.8100816772426291,0.18081135775731738,0.16509419991724927,
 0.34101662129658705,0.23321340390995915,0.32199271796892237,0.53517686431127,
 0.2101069167022962,0.3449910070939042,0.3563245218517699,0.31361984674825394,
 0.23376822984632084,0.47454870591653153,0.25749001456367315,
 0.08652402131050452,0.8157073692683093,0.9416452637661764,
 0.38616825408729494,0.2986865704600413,0.9549407265643332,
 0.23893170006062525,0.664305392740004,0.004087493076135207,
 0.8611462324609579,0.7160706784516828,0.8607701298642524,0.08498231030384407,
 0.2857899246771217,0.7491285320461724,0.10658730368266067,
 0.09309236077562943,0.7758023672918123,0.09685723229903231,
 0.38586649868790057,0.7143330330765303,0.887366227887857,0.16863158306932013,
 0.642976752128253,0.3300027694212919,0.1079855648423661,0.25855368005655266,
 0.6477404367101146,0.832409285358442,0.4334655479556956,0.39837840973383964,
 0.3409403833451574,0.5447937843748857,0.5437567405775449,0.9345718415393347,
 0.8447892949537215,0.5526349023588221,0.022821443257441842,
 0.20419037077799018,0.7468510954603578,0.7011830684636668,0.7267890280598008,
 0.058444249729956876,0.7610900254056725,0.48567017424932235,
 0.3391822779617152,0.18612860755943128,0.5587286777356842,0.5934437972726422,
 0.302897882607402,0.13354025273849557,0.444904981391698,0.6814526694380235,
 0.17535737641160765,0.27635719726750263,0.04125716434712301,
 0.21070377800567786,0.0815259133894215,0.25284113703154265,
 0.25749169420657336,0.4614030599623816,0.4873647324167204,
 0.04057745759951581,0.8239171224668284,0.792375099090705,0.12687636199289554,
 0.6787674962422162,0.7373893962382186,0.8865961304195324,
 0.045396462838554585,0.955532837072664,0.6400037420813384,0.5080460640451689,
 0.26649340843585656,0.9640053728362251,0.7476406988274727,
 0.11558413570458426,0.9858446934642308,0.615053085844899,0.3677033414586832,
 0.07940042253109914,0.5341855116024096,0.530566740438014,0.7832928687816352,
 0.3354237336354178,0.3889594995796878,0.18806640362210425,0.2935047398519698,
 0.49713971594785644,0.546896033480008,0.4375530184530254,0.472812326133802,
 0.989186644549271,0.9200379258862967,0.3196054249339173,0.6061856866519999,
 0.6154734723894401,0.5015649823956739,0.20546391182651402,0.9267232954524612,
 0.5642516109311669,0.7762683162973917,0.3805320556865672,0.8702681022761786,
 0.8058562376425609,0.04598727686089632,0.14704196674529912,
 0.2691643110189329,0.19394589103776305,0.7112845069218316,
 0.02903649364576988,0.8388923185913718,0.13107063400082763,
 0.03193720692843993,0.8196645114079799,0.3702344089244203,0.6712880175478972,
 0.3667292627516172,0.16529481035455862,0.28801806099260907,
 0.1446428803721742,0.5004582996718042,0.010067764883799457,
 0.2047722191264426,0.4761554758999812,0.7018734566475981,0.3427275558031029,
 0.4156397371656819,0.3274693720238846,0.5835348102117448,0.1830116324226172,
 0.1477088829865345,0.20818276757961607,0.6737847990004256,0.6868891416286643,
 0.21895055654343987,0.7282807975270446,0.05534362159001682,
 0.2685281949365296,0.2537677430484129,0.5263376976874665,0.27478208631902323,
 0.04866293481687345,0.5791235585912592,0.04549151251017436,
 0.3823446725708981,0.7872309576431754,0.22435534668056534,0.3323450361139849,
 0.5991781778419133,0.5613188841572829,0.41877204551912595,0.1020646448615139,
 0.49021095632159617,0.9653256788003564,0.9576477814455494,0.5215708794074143,
 0.03921666565409787,0.6085728438821315,0.5847903590809053,0.6169449305606589,
 0.5687954359904462,0.15953463291803216,0.12824992261224033,
 0.1527575303733425,0.46853247505557094,0.37489703272467645,
 0.18541664689431503,0.23563049319460339,0.456171005851174,0.7966173579646694,
 0.6176537825870376,0.5757946098451839,0.006710401698236446,
 0.14846732970689858,0.8001202590282993,0.7998329422433403,0.6145107988018217,
 0.4875657466294273,0.4051191929310075,0.8237453597646401,0.6421405304617359,
 0.6195976836458331,0.3612265021567155,0.040112251924327214,
 0.5371225659703923,0.2618048906977939,0.9670623729944687,0.16144046543299118,
 0.9112798678371914,0.9673954034435547,0.8639264677585458,0.8686327251026644,
 0.49177193222940974,0.04888527332170112,0.7751709544786677,
 0.10350385813236895,0.5926774682186643,0.8914002546592413,0.8053297603151565,
 0.0753026693615324,0.5836809374742047,0.9324301905599421,0.36082333838672187,
 0.6018414812709583,0.4369269343380989,0.26675434311735124,0.5169328726018235,
 0.7549388508963727,0.7585418786849899,0.07279696079526965,0.9970572630560949,
 0.24840445121949073,0.8511484126060878,0.42034024421407745,
 0.31523887683923557,0.845025954747407,0.03056063992751823,0.9474965915816342,
 0.6444297810707662,0.42732236818781266,0.8430145368807478,0.8496519177720797,
 0.89299475820021,0.2946278402530378,0.1393557357432781,0.14101474476371567,
 0.8460274901436351,0.5985831261481238,0.9728523531535845,0.8994964861208083,
 0.9479721461597156,0.019458384431227937,0.4978161863641771,
 0.6568800677469007,0.13009498210811077,0.11965625352988174,
 0.6688407666192864,0.9886773445629682,0.6547197312554216,0.3307705857046145,
 0.12276870252278471,0.09947665908676628,0.2677358331394952,
 0.9807964606614735,0.5368832624502958,0.45698535296313325,0.8662790648054626,
 0.7012706802517995,0.9583050536811968,0.3134573924230165,0.1793248983371174,
 0.5837900103301389,0.5554701908953441,0.43004848670878404,0.4783172642444399,
 0.8622278587056109,0.23096952653493585,0.09318922459240064,
 0.7344505538976925,0.5722259306977366,0.5862507032128589,0.95450910756179,
 0.6588945625980654,0.509964095947351,0.46124452488141254,0.6525134541555881,
 0.8267392208413437,0.19628027648008484,0.02437032443225451,
 0.5989181359075054,0.2903338850895052,0.24710836741921993,0.3005490647801049,
 0.4685741078420861,0.5261602175187461,0.36225454878502816,
 0.22921652008423388,0.3010995032808288,0.5228654792233447,0.7635519442970166,
 0.8573392149966109,0.2580368124242052,0.11522500876509378,
 0.49062558568245407,0.9014814137427446,0.015282258935026105,
 0.2461714863859953,0.5442546215351225,0.8606975457386273,0.24173862481814723,
 0.3030032574508279,0.7978596436827925,0.7777223351127183,0.8477100839019195,
 0.06951391673449636,0.4842103467416219,0.18042521748411922,
 0.3157710507431577,0.49321259151395114,0.3799689908574063,
 0.24986048099003066,0.37102888073546925,0.5969032551344966,
 0.3878657847978697,0.22383946358331375,0.8035972197482041,0.5045170120648476,
 0.29440863796560723,0.3380050194087818,0.8675754390900507,0.262464578057213,
 0.7120139910124278,0.13416552727384878,0.09875505277173391,
 0.27293009116664824,0.5019657463945053,0.48528238412299274,
 0.12376392744532794,0.5794502275365809,0.8760489659393128,0.6463051341223852,
 0.21340104872891452,0.15727508735883444,0.5537723092695026,
 0.2891127706332335,0.5881359679369428,0.19679841586009128,0.9391143663787116,
 0.3772889194276331,0.2505344055843348,0.4632393707375657,0.41337686017721875,
 0.40438012305709437,0.7954692526436778,0.35909144659271575,
 0.20838771385472554,0.02059622465544786,0.3060491637430478,
 0.39741234370106315,0.20118024067448237,0.31364661849918873,
 0.04969056289247087,0.17256121877680408,0.6383803236315199,
 0.7265417272398739,0.984177405168196,0.9691494952072455,0.7987160458000551,
 0.093721205076024,0.8301806015926205,0.9936250633265512,0.17453953328884053,
 0.1886920230741027,0.6794975970318581,0.9022290424995885,0.38003322729389766,
 0.4069827665220884,0.5182246477556013,0.5055300045019042,0.16696300997402835,
 0.7014495631816303,0.2887903815779429,0.8653383476447685,0.8214446911813184,
 0.521512168143079,0.09120481806545488,0.2428834715363044,0.2626596905031273,
 0.3378413699865056,0.03772550376488337,0.3956121589849748,0.3946407180556105,
 0.8211162993884937,0.9529460245824112,0.16655737494875766,
 0.05114475519996642,0.8157992710888002,0.393254992992049,0.3800644692300944,
 0.01444463223102388,0.7767193608087923,0.8205731701076808,0.564669205503125,
 0.6589428762067484,0.8403293619988764,0.9381144051386497,0.4275806595867926,
 0.191478300905653,0.15754221285590764,0.6540183444208711,0.9038450986401739,
 0.02882738930665607,0.979678680980171,0.7413282627557034,0.9521262030594514,
 0.9314434184409596,0.7103977879845154,0.4146364598354658,0.5387443412380633,
 0.766108341222511,0.3589004963492226,0.07557209632182627,0.587106203532562,
 0.832329312747719,0.07541038991988747,0.6841877065437083,0.3091208352997956,
 0.3429460567684329,0.7280085724732177,0.08629856327606389,0.6561667128488524,
 0.5385418780977757,0.2899984778695873,0.899781692101812,0.4728832089488635,
 0.6490203591382251,0.4904149313989301,0.0784038518342568,0.6598951047810376,
 0.05083123091573327,0.8080768015149419,0.25155332046959755,
 0.9856819480507776,0.8608197494685651,0.39590456451133604,
 0.31702487567154347,0.9064130237814296,0.640407176642663,0.9371541319330785,
 0.9805031671656698,0.9620359443750439,0.4452053538364642,0.07697589028469631,
 0.36888088281227005,0.31750286052319354,0.8948965041493915,
 0.8705133142764726,0.4111031910861328,0.014673054084323844,
 0.34285433488017913,0.30194182631939737,0.9186778440402962,
 0.7372204919126831,0.9217697289380474,0.044177935623933684,
 0.8228285266901485,0.3810819034234332,0.5103145550383623,0.2082330242250885,
 0.15129889819076292,0.02685465435945089,0.4897790278189855,
 0.32949795825440875,0.8643966780894187,0.19946928978236944,
 0.27020351470250514,0.14387829055310142,0.7779458519600166,
 0.7086363615444473,0.42821576178399545,0.058008048507006915,
 0.917426626560574,0.040162838706088944,0.7969471760484946,0.5029042300553089,
 0.6405666120173115,0.8626463438725703,0.2593406054641023,0.5221752163780987,
 0.8051535551210315,0.3807943928733124,0.6839061947379512,0.3100569040706356,
 0.8997058440317689,0.9945455278061917,0.43793591488347305,
 0.027642837876492576,0.6578070734262407,0.9858008878146918,
 0.2526613774856218,0.8811647334051125,0.003882942375936338,
 0.41333445484153064,0.20419173489277553,0.7857260147525679,
 0.4704782977955567,0.610933292377127,0.21689964887792823,0.8889930822001708,
 0.854757519069946,0.45267427492166834,0.004928493193429584,
 0.6121017788370365,0.4102812023668896,0.8608368414793446,0.536175352673215,
 0.1827031635944637,0.916033342489617,0.11158839048100888,0.2088280617746069,
 0.6471035220568905,0.9560462142653765,0.36024637185163355,
 0.12362704532000313,0.07069640176365422,0.7750855364549832,
 0.8917319110310581,0.9559784134182069,0.43871028462772554,0.6931102401497042,
 0.8680830308421427,0.8861576930367929,0.1887880950265991,0.8776386481708899,
 0.7960719839190069,0.5506021602847689,0.8429117304384074,0.4201115297262704,
 0.6582507377400095,0.21947754677055342,0.5970227601039861,0.2406461794628526,
 0.7827875524886742,0.7052814327405466,0.6418298016686813,0.5658058867869521,
 0.8651942761493322,0.755298184501225,0.9374601329807788,0.24457512480306587,
 0.44687337707918384,0.3048186233380581,0.5668379511092749,0.5598806526448211,
 0.5490569046655962,0.11426537478305476,0.75278293267805,0.40498728348696156,
 0.6311142471290483,0.4825698734979287,0.687320616048827,0.7072730095358247,
 0.8735471901237147,0.5682627646122982,0.8807587222194999,0.2963831648901255,
 0.6828972691157449,0.19761457457355802,0.9216412547230034,0.8902812051573887,
 0.9056575260425302,0.24470808343949435,0.9638065088647249,0.4575050753112304,
 0.4770210846446279,0.8462332378671572,0.27883780473593056,
 0.04567085703106577,0.8492462329227041,0.11986460750588435,
 0.03313230000989731,0.30905492448653227,0.2198747226481419,
 0.6851626032480769,0.9293184522106024,0.07938082791364331,
 0.33573341403501067,0.1518689612834716,0.6524683947947396,0.0262581094806974,
 0.10375170936032574,0.4609386899179746,0.4442161585143666,0.9978196364082784,
 0.7560360909577375,0.8122743376055532,0.77777838866983,0.37194471948257,
 0.527951951789547,0.5095989664950935,0.23401196428364868,
 2.167380986156653e-4,0.695945561321286,0.295486005904403,0.45457484681175575,
 0.04129440150545305,0.427697077288257,0.39873308337894786,0.9981994706444516,
 0.7971124145067316,0.8701456544302152,0.3747831548526529,0.7270271429481819,
 0.8549863220810145,0.6171579994869516,0.08058580880446176,0.835458219035202,
 0.1345978918672478,0.6973654067077382,0.823064529498036,0.9429772373792599,
 0.44995613731939177,0.7488683538910343,0.42023636882453785,
 0.14282428328856456,0.6034321326466995,0.12686852914437752,
 0.17931852477760213,0.3957644179034182,0.12080819110592655,
 0.1632345963869375,0.2063084787461379,0.052370808239592836,
 0.3933912477486313,0.9658315785233348,0.6449710891303342,0.8968154937107474,
 0.18692834935596525,0.16574734254077583,0.6189594347854673,
 0.5825072249783835,0.36382458648642235,0.44321222807641747,
 0.8835437420334956,0.49310617463196393,0.3447135126973556,0.3163136468729,
 0.6924585688083359,0.785613571115243,0.9314490069297788,0.08536267198464986,
 0.07553464078395111,0.10019248587964202,0.9901164677002138,
 0.05821803992532726,0.5014918099871503,0.13609019429188496,
 0.44383602118969256,0.9112247756983626,0.0381193103544456,0.9814688961117395,
 0.06479912393472576,0.14666790859453505,0.586441035084759,0.8280842708034002,
 0.40095622683293364,0.5869733038711844,0.3162861163454921,0.9145273452488507,
 0.8975233405141071,0.4156230918354674,0.8981175434982555,0.41976333897031826,
 0.14254037762745864,0.4759608411587416,0.09547163973716488,
 0.5855346931679631,0.2603539934567882,0.7068652726331816,0.9183505559119699,
 0.4509085693285403,0.024987610409725414,0.8238340019024699,
 0.9322512339261706,0.46875491138569725,0.8390989727002061,
 0.02517452020073041,0.1935789954501661,0.7175548785189082,0.6243040527520564,
 0.44008991693357336,0.33442392501380347,0.3943347657443147,
 0.28405431632642975,0.5654976967593508,0.9431547212232535,0.9878079381978596,
 0.8558426796020466,0.785419595657511,0.7942147265534337,0.1791646558387343,
 0.16877080888297014,0.8454890961632698,0.1923892568038852,0.3116639700508326,
 0.8886449314590557,0.42621756852109527,0.9647178740744242,0.5698253417044598,
 0.32335448402335976,0.6291435143476741,0.32096941539696844,
 0.12097648628392343,0.7514494669023589,0.9554878617189342,0.7146972486503801,
 0.4575756020797741,0.0808982190586428,0.019387664593164877,
 0.046806392465547186,0.8561037605404291,0.06364592553282578,
 0.49765270149691665,0.7224918205975784,0.8892455704931918,0.3810767128183803,
 0.3434102526932944,0.70530139833594,0.2219138510383375,0.1644468445153614,
 0.6817478447284314,0.7305217391529175,0.4015662491729223,0.9473618552003104,
 0.5932024412060033,0.6433144928678081,0.4265201786503874,0.23680107111487547,
 0.678670646566786,0.6034211710913495,0.611209593644244,0.40835535898436015,
 0.5415674900182311,0.18911513622411313,0.41064453553201063,
 0.5886657855471628,0.16636806828895467,0.9777224017504954,
 0.029198114127553465,0.3141431682795329,0.17990698984612652,
 0.4217202387805544,0.22747838450526325,0.19233633759775481,
 0.4796223990112555,0.07773858240464404,0.8971929263814297,0.3854054676847738,
 0.3814309817451189,0.5765255787886787,0.5811840026104859,0.4884703287421821,
 0.008505711344028777,0.6501632510697161,0.9594122243735869,
 0.6415063488777989,0.643825510525528,0.542516442730885,0.382245257124755,
 0.5335661592778985,0.6814076903654038,0.803959850119192,0.9085744815872172,
 0.019731487416214777,0.7529018984734343,0.27386168468318617,
 0.6287963462804494,0.0634965897373202,0.629345199992301,0.7152876299824282,
 0.8398642707210354,0.4851031697248531,0.2377291269575974,0.21202076594446406,
 0.6942372155528311,0.24799614162555883,0.5036627185726181,
 0.09269899162062045,0.34997342623359773,0.7404776917152385,
 0.9338709237121026,0.07353338438054957,0.7690491303284952,
 0.04627515259654402,0.9549417900674866,0.18726411242457264,
 0.7829147544981018,0.6059208661282987,0.592698878353019,0.18643883640747638,
 0.8637133514531914,0.8580821591210194,0.6087731224780324,0.8432809864011475,
 0.9919088825557751,0.7194036560196434,0.8118385175945426,0.8090323480731848,
 0.15961989390702092,0.963494445205081,0.1778959105051373,0.7349632130522992,
 0.9923034049425835,0.815460886092922,0.07715328575959057,0.2394931542619697,
 0.2125839291397531,0.8503098379875287,0.45418201768210276,0.3163523644825792,
 0.5143999873621888,0.6134856942806418,0.08938031175837136,
 0.28035950879559557,0.8576196626208262,0.08015285715906073,
 0.08429215448561456,0.004962966354523646,0.8951430639835807,
 0.049620861436865726,0.6141271989487833,0.04212811329148758,
 0.24277674054404186,0.025230748165048777,0.26389625987412835,
 0.7211630642734019,0.4737355173939637,0.696447639752383,0.9641357291461116,
 0.1345807768402657,0.03415468801977739,0.34823227166886384,
 0.9173073501117935,0.9027034042006216,0.7742696740276384,0.41968960656147736,
 0.6301252236338722,0.11864344748715272,0.2351915316579095,
 0.32791789019209283,0.10417467421389648,0.7948016724678487,
 0.029607494390773703,0.6405286977085419,0.6710754605562127,
 0.35683042952154886,0.5823812337461106,0.6155876857873236,0.5311671639303585,
 0.8467643256395772,0.16188010804736142,0.4720835914706829,0.8453897647975819,
 0.6772698041827901,0.4980937350953588,0.5343000476114984,0.259954072108741,
 0.6830111214635071,0.5725377894977374,0.41596740202973304,0.7324589400824635,
 0.18436468740990308,0.9496476357807323,0.12862336233957106,
 0.8799512857408768,0.4945679648182836,0.024437788708603025,
 0.3316397032667312,0.04549539291667226,0.10125999995982182,
 0.7473812733299503,0.5273896086327942,0.015306628590151306,
 0.7594877266750881,0.5700308419081953,0.8660033480919781,0.7310649827844072,
 0.9059275258468126,0.6338797317623107,0.8279143113871795,0.366673889591694,
 0.10658668409368521,0.13885709151263137,0.34887367106347833,
 0.17685147312926408,0.6341479172806122,0.18789212223627116,
 0.5785463324365039,0.4940815456476456,0.7148154645557441,0.6346895892337708,
 0.07015116783419484,0.304617312850192,0.09142095032929809,
 0.43049034916692264,0.8118823843709079,0.7334773548316568,0.5820637800856002,
 0.47558724722541834,0.027457430360175783,0.1683295098875197,
 0.5506283973589206,0.2377225573648597,0.13202410624168803,0.6786000204593375,
 0.9765753269224805,0.8898673434717331,0.8910153052957674,0.41582404305199583,
 0.15003196998517299,0.31255826114627216,0.8028234486915327,
 0.31176935031848885,0.9995591321521573,0.4425254037555789,
 0.38120962459814856,0.4186450710213966,0.21036878757799238,
 0.013728090565131135,0.7088174448006632,0.5814654181073153,
 0.7829050856196138,0.5033948603037712,0.6523904133118335,0.8407467907349735,
 0.8144744696219294,0.9335900801191152,0.23606909301003742,
 0.23415519605116364,0.042402755130488856,0.3581676436878707,
 0.03405518712227007,0.10267117995562447,0.7107676453871434,
 0.5937901555414707,0.7870124739064053,0.5820685978100425,0.21076544892066595,
 0.6660926011531054,0.13493740206654548,0.5134321379991537,
 0.43421272259899935,0.9548243845751101,0.483914572117323,0.7647034559701037,
 0.06933791434475933,0.7462826493990397,0.8040082590117341,0.3637709272850663,
 0.2563224560688502,0.18925272136637417,0.35821823082753235,
 0.7609744301319866,0.2716441007589637,0.8406454144828928,
 0.019108521781556864,0.5076356071834256,0.9576531816008285,
 0.5215904350251681,0.9727986396516853,0.5627193592871758,0.6728371697850071,
 0.7992800746741825,0.17950378109128984,0.9895667621937123,0.4465650573582254,
 0.812470812141252,0.7031148626458054,0.38134005647073377,0.8533021764155717,
 0.7352963719200728,0.27873823212899973,0.33260944141363924,
 0.13034550628985386,0.44750884045254535,0.6095857326585454,
 0.33020325948596874,0.2719880501303623,0.91615288972278,0.39731122029501376,
 0.06939040652048112,0.9733646649171926,0.8007599977850398,0.3226541642276035,
 0.9573097981877144,0.5187547824427375,0.5116791016364785,0.5261170262911363,
 0.5037752005815598,0.49395435776037977,0.532132552705971,0.5995604444678356,
 0.7162564063841548,0.5873121389056633,0.7634680514870267,
 0.014825857946995646,0.06584666256800753,0.7653693116608136,
 0.3966115566763573,0.7987653831754192,0.2697070277513739,0.6549269443930371,
 0.792903236101443,0.999731302567074,0.2876412495860835,0.15937462487200504,
 0.15205609394757735,0.37194885978664693,0.22482030175227563,
 0.7702482545795617,0.4187887779082281,0.3584111959157369,0.5451698316090251,
 0.7984891699218178,0.37410495094417073,0.7335272205072288,0.9555930807063285,
 0.5035831257147066,0.30984180503362857,0.988573295125343,0.7136181529894953,
 0.245973334961462,0.6421983093458933,0.7784676097666141,0.053957371864058135,
 0.2228541692004815,0.43027431958139684,0.7073901661426815,0.7269056630712327,
 0.6405320295455796,0.8880702543270286,0.8862322184708192,0.3190608270595535,
 0.30493204266347074,0.18769956556545864,0.7633291917443342,
 0.9723402997816576,0.9143085782865175,0.13667689237311342,0.7538293698538902,
 0.46552870780432787,0.7196555398431836,0.1803921872514891,
 0.24177513924922456,0.04600461966329017,0.21833713117907239,
 0.3292638998088936,0.047870477122164856,0.5229923148439877,
 0.8336911342938367,0.727842231187747,0.9349041004877781,0.2894735045562483,
 0.5445922620292913,0.48776132180317466,0.6744887712719886,0.2807800791770052,
 0.9884541899580787,0.3240811863176618,0.9442099765684406,0.9525863962682432,
 0.5145480662961041,0.8459050904168085,0.9748741257015819,0.5864042547810737,
 0.5994468577085077,0.9495386994564463,0.9200704775856056,0.47822154259432803,
 0.046218500658272155,0.3378298744570166,0.7591624217002182,
 0.9550444894678675,0.7585544686134391,0.7957606266038404,0.8831096879334477,
 0.7612525159270633,0.8299278397108951,0.4768577407438507,0.1416022873116769,
 0.9469741413500178,0.2038865813424633,0.24316386536557721,
 0.46835961048501673,0.19742543619960295,0.5046789834597768,
 0.8267708504943642,0.7342820904752555,0.22217350185886053,0.1402814312346986,
 0.7767732860786526,0.23230743297379097,0.6741699270061635,0.9855832087298414,
 0.9626556248819089,0.2755269816161272,0.44475696344136095,
 0.11508702358151779,0.503488249807193,0.7254204885399215,0.01817069867248533,
 0.6654440669043793,0.597001691817922,0.3875681868554267,0.06378970961932284,
 0.005972241433486847,0.9171805698881678,0.11758365984545427,
 0.7493605183977936,0.46149013281812,0.6128037764546203,0.19451422205462732,
 0.6632441154492277,0.1464339088695623,0.557554788686681,0.9501214215611383,
 0.7890214183972931,0.654045319656139,0.31053419287112105,0.8270769898769094,
 0.5473921342799424,0.9801785199454576,0.6887506506753316,0.3013440614888516,
 0.9237540279503017,0.6417425707652537,0.50845855891918,0.4220446693562868,
 0.4300296804246151,0.23988744473724488,0.028023509596766916,
 0.610945274918417,0.4474262951306548,0.31011917423863644,0.5390206276809656,
 0.08061982154817005,0.7623440032318192,0.3742344335769814,0.5233080879783999,
 0.1178508030255313,0.8302265895730584,0.10310749973520561,
 0.28816474012968574,0.30415959325129516,0.08780524812109558,
 0.7762269768644471,0.8324631762638692,0.21699715991532398,
 0.04396099705342804,0.7650443425746954,0.8401846386514177,0.6651083963498082,
 0.5812257719235142,0.19479872187657976,0.5417051187287438,0.9985379306199318,
 0.7935393546508458,0.549561022059129,0.8948737941004519,0.2604177030272462,
 0.16227753364441155,0.7181256969074821,0.45656113674478105,
 0.3694413370652392,0.19353719895490396,0.8097471833129203,0.9078627488665959,
 0.854601620604968,0.6797195105165975,0.14418501231201986,0.2868361340322928,
 0.8166824997643445,0.6926178107826121,0.9406137151128768,0.5555263006533503,
 0.26808099603262303,0.9790259029203919,0.137348946562589,0.4708985583546407,
 0.20549543668367554,0.8133829430283014,0.495872901280184,0.24571174878955127,
 0.1650491115932523,0.858412484189078,0.43538043035897545,0.10292194163059143,
 0.12309906579248653,0.9636426964586197,0.6130469607487005,0.2592193644043672,
 0.4966110571624698,0.5278049923939292,0.5139498362160035,0.5939019321859247,
 0.6891728431228703,0.5488402498548034,0.7368742020072797,0.6489550995544893,
 0.7074130258027619,0.24987749240389545,0.25827480963355043,0.655303987846379,
 0.5173960030700295,0.5487422277445072,0.3887666161263008,0.4150395471262953,
 0.6474007699499168,0.6367701738650574,0.9602452242347492,0.5121419137300275,
 0.8818652881864767,0.8303132273924517,0.10810734396482269,0.4883070408538708,
 0.14033095026437403,0.9268910593758773,0.03060429432151668,
 0.21558769407543965,0.6775132777159293,0.15624124050159482,
 0.6687685401863392,0.9922240884290953,0.8464257195802639,0.13220392277693638,
 0.9639616254006411,0.1982066798959331,0.7043114003175832,0.8108731645189643,
 0.28543740482412316,0.8297737256195491,0.9272233883942136,0.1381034009179587,
 0.17321155846512393,0.5258521672165907,0.6691907912698862,0.8141570405331964,
 0.7992511362530075,0.7330453914082642,0.3537487874685622,0.8391095085390364,
 0.31751973474037665,0.3069204434122008,0.770499807263454,0.17360982989708929,
 0.12804555968915143,0.48547630277484477,0.8145543417931594,
 0.45853243848899605,0.43656001445798176,0.4662982042542656,
 0.6626024443009131,0.7251657382063422,0.9783660452807941,0.6636592340240132,
 0.9731317602388126,0.1327258655944622,0.2553190191212661,0.52255280611187,
 0.7162239147410676,0.07594744333356829,0.405892590656318,0.26870542085358473,
 0.10975901735734395,0.25615963008546005,0.7608562793394688,
 0.3784497891425703,0.9738442655364253,0.46410578913475853,
 0.16561346430834845,0.09092727982405768,0.10705600092195877,
 0.27188071333804886,0.8953844372027359,0.866896584800733,0.5957416553334511,
 0.17477505400676985,0.09271705476942227,0.45673009928754493,
 0.2273782821555418,0.43339653570513936,0.8596576521385901,0.5709058099286118,
 0.08453026438142774,0.7630713639520545,0.9722902681124403,0.5671549378707639,
 0.16114109846584213,0.8009098168947644,0.21156668826385483,
 0.004039656157789895,0.4758596149120662,0.39725666156954764,
 0.1181907780507091,0.01968934153910229,0.5324655682505988,0.6348658701769456,
 0.07423987777976293,0.21447398644394577,0.857414852063949,
 0.07167480829584572,0.1933014108568103,0.14468008587515824,
 0.7072407959258784,0.10170344586900415,0.5009060569823839,0.9093216566710514,
 0.32269549522488616,0.5202927530816406,0.0993588955169521,
 0.002275638597961316,0.7543792152957249,0.24114978703888024,
 0.8426129248587436,0.41884313638628545,0.9818090115719748,0.5139985402290698,
 0.2965811297556701,0.2989951900961296,0.2396434440623687,0.394203475604067,
 0.5293856126897369,0.17298896535872665,0.27004080215610504,
 0.28315418615524424,0.6396759382220425,0.6489451571044111,
 0.24935623654250294,0.7921306226561421,0.9220769047385285,0.5890134992987656,
 0.0683995299593897,0.8093130476739476,0.10859342570803798,0.9009737214564911,
 0.7331573584484904,0.0897334035710664,0.962202302347971,0.07605184871076631,
 0.16967717754626221,0.25733393703855,0.21414096264746763,0.16328360722923263,
 0.5699092879248389,0.27069414911883716,0.5327832023451535,
 0.21142033877606892,0.2104124981194352,0.5580322599341027,0.8166382904731685,
 0.3905358428943849,0.8697752076729737,0.7736317035195366,0.7100754525059582,
 0.9021792405041448,0.7642326472764034,0.561245211320414,0.25565716771661917,
 0.0851265327591979,0.7241749570586458,0.8980164379030583,0.35555197596941135,
 0.14354321468156395,0.7937935155731866,0.7247207439139598,0.1671887689467655,
 0.14708126773191244,0.029576018718066366,0.6820142640373735,
 0.5512081007486819,0.9316861798705549,0.028840231881547274,0.465596140766922,
 0.5383839780704384,0.16682661201499716,0.26044724431394284,
 0.9847745792387336,0.4532604379669528,0.857071635577353,0.413615471775554,
 0.7701347327813686,0.2593489078206759,0.7888156099960547,0.4341445573660856,
 0.16434988086742097,0.4793824517805573,0.05013817835172796,
 0.044532083681558365,0.4469310396237267,0.3390246447387816,
 0.3446143018367914,0.9193023665995246,0.5893782663126274,0.16691893785213985,
 0.3423762181977069,0.06686408478457528,0.14845526439599455,
 0.5288178598314317,0.6358964138966162,0.6890247823308245,0.8090713958528302,
 0.17146287274728966,0.8863867329504667,0.7318034419907736,0.5767967834450664,
 0.25420655292406447,0.2372990138619565,0.8830493470245274,
 0.021886085809184164,0.6844966923962498,0.1564232988831008,
 0.45609201590395976,0.5985608279937845,0.2017467678654652,
 0.45063770393814906,0.8722222133170017,0.16303355856790347,
 0.3984427066286804,0.7897448174039894,0.0910172892160186,0.9783269070642959,
 0.23066044227437477,0.04665851291855461,0.5823382374394057,
 0.6521670005617254,0.32714661204333684,0.7714563015161959,0.2486977608367298,
 0.3430157534607754,0.9051453313527771,0.763399502102072,0.4030594015982989,
 0.24813215806544675,0.13072831466601753,0.131445332970883,
 0.08211024289685687,0.899388059088604,0.05588574001017388,
 0.48405227294936337,0.33869058875886304,0.2627021832264098,
 0.5089493208562217,0.8075522465430154,0.9540977008478936,0.8977867355138228,
 0.568977410857423,0.3129090722723401,0.17103285049589867,0.33176126230507164,
 0.9424582750670947,0.8272890391754437,0.6999933000704711,0.5409041890196986,
 0.1984074473470706,0.6660253164921095,0.5292394268299251,0.6779135052559597,
 0.39948611974056025,0.7946040620346504,0.9050794588714159,0.859040614033472,
 0.014626459465053143,0.7391613208740516,0.13068439998038994,
 0.3349179434345688,0.1300107478879381,0.04423470018329412,0.1826792736439563,
 0.8000371561956887,0.005627582508876294,0.6459029268379607,
 0.1528036529620852,0.009051404972821309,0.11638520215351034,
 0.8080416965309958,0.09569961963838602,0.7673641380883642,0.7094103407141045,
 0.9474861133485329,0.38037507587396213,0.5801794411063258,0.7238437646482073,
 0.5055953246165173,0.9905153193711912,0.5093717174321688,0.7628232009872862,
 0.35260147788825513,0.7952665578826168,0.3856549975691529,
 0.10209820838518235,0.43531394951604485,0.9175090281029953,
 0.4712135078340385,0.7121215517739712,0.44003295879788196,0.6286251965000869,
 0.8616321442429341,0.9308178158810053,0.17787298487764103,
 0.09415664409256186,0.6764030909365704,0.7935823548348072,
 0.34699908541378865,0.8156445113928632,0.38016444934454274,
 0.09442624878362527,0.17359059301582136,0.3673173671887938,
 0.22492956530154484,0.05002587832971683,0.7941257730987817,
 0.2582243943755649,0.0059219213253234315,0.08066315626933229,
 0.599185916558299,0.11227398493541896,0.28565803446060956,
 0.15908260528056262,0.983545742386358,0.7911562295365416,0.4783889293156236,
 0.9268828703810656,0.6586882433083558,0.4296909437214309,0.8388672020054048,
 0.6893197323083822,0.028582043760488007,0.23894665066939058,
 0.11266227977937615,0.6057433959975237,0.10251607915717598,0.889072030875011,
 0.07640986148498918,0.7265285960745276,0.6580356644085255,0.9112670084457017,
 0.8554464312204834,0.9938326186373937,0.10347614150321682,
 0.23693772496533982,0.9477119433313503,0.2897835187871338,0.2642802162409864,
 0.20386324698966563,0.1740797862849831,0.34932130413357343,
 0.2630775312523963,0.5500545242917102,0.865452949211943,0.9560497645254039,
 0.4581362957603089,0.13395097601158878,0.2140320839672495,
 0.30992576026781327,0.8247189520167177,0.5845046573822118,
 0.012871475936201193,0.6902707050139718,0.058685320189712065,
 0.9826015987906302,0.7926868910809294,0.8669470343821214,0.7490279099733298,
 0.7407101608222917,0.3653729509110537,0.4226942773782906,0.42192892716992936,
 0.9772654763058535,0.39438809268718833,0.5754411277163914,0.9428192628778503,
 0.49523204833004275,0.07665703750903452,0.8389225101117332,
 0.8379680446085949,0.5477532220352945,0.49002488599915206,
 0.025460439066861573,0.8999549358553941,0.17145337507921488,
 0.6041044806026087,0.12270447948563268,0.3802788468307192,
 0.25466132452603696,0.9508867814202342,0.7791804345194874,0.2557581506074458,
 0.09659432633248333,0.1664793269225303,0.014524839938339862,
 0.5341491148671695,0.22177846330931517,0.31590220347557896,
 0.5525863736665738,0.7826187737573662,0.2834275703217938,0.9406668453568225,
 0.35436009595372975,0.4871704889333306,0.6588230021148157,
 0.48899734906672243,0.06028864450794602,0.11139222454574915,
 0.14917651300141033,0.2510015361505914,0.6097836982553213,
 0.10194896888964045,0.6955413193013718,0.846349889929122,0.768104741054478,
 0.7779447481845551,0.5307501944007693,0.13521699999991443,
 0.11251487261917492,0.25778025412860095,0.8095215610964159,
 0.18713739381459726,0.7747669481843485,0.008283458744198557,
 0.19802895915595264,0.16605248903845193,0.5009240304020731,
 0.7379945394832155,0.8143492850210916,0.26850424948877405,
 0.10636211880373814,0.6338760003027352,0.534585875751058,0.5441963226000448,
 0.9483643650285289,0.7310586456945254,0.3115255960451535,0.379116642314117,
 0.8507062055151198,0.3860601037271294,7.879857853240191e-4,
 0.7552738646207872,0.45451275261153423,0.38402681596297517,
 0.40480005799552876,0.779299349340045,0.5847185862006391,0.19066473065244238,
 0.7981057239371117,0.587755998134279,0.9556630199677546,0.8603325449207424,
 0.7621229560907656,0.4714781080888415,0.8281721736576041,0.45033557031226945,
 0.2302261059306785,0.4288854163532194,0.4488582651617351,0.7649322525433063,
 0.711187676757127,0.6996638858184367,0.8711680225471267,0.19812288361103692,
 0.14448853612830315,0.2731544722571019,0.5353769160475857,0.78670551422394,
 0.235116752724137,0.10984003660232655,0.9402460739496284,0.6813878561834722,
 0.9330642715279588,0.5275868540121151,0.6048950245827092,0.48327711686714836,
 0.6189284530331738,0.14363218184369064,0.19409418978804616,
 0.9190458620658156,0.48934187443601695,0.26437136607080136,0.706252204266759,
 0.3545435962353808,0.3685936433367065,0.32899424188411674,0.5057268469086758,
 0.8268069058694785,0.6717039168144224,0.4204525474388999,0.30771149987148105,
 0.8520139390807352,0.2819591196891372,0.6514154535425699,0.8373672116789798,
 0.5655559680169614,0.9627696988729743,0.8172529347392513,0.5442070554351126,
 0.08043070638015204,0.63243100071577,0.15362111237612697,0.3727748966767708,
 0.46269245738399345,0.7924279573082713,0.7038154645958346,0.9468779795586764,
 0.2360629823871523,0.13711530190094467,0.5332723838483043,0.9556845463567814,
 0.3037231868565987,0.7134810288775164,0.5632599552004934,0.05305966873672141,
 0.8040487288074719,0.19199821766442582,0.22890751451705815,
 0.45134458260144705,0.32072850680996934,0.07572642265821983,
 0.5570545837001795,0.2763583340469632,0.8582510408444439,0.758659297675666,
 0.045923288228253534,0.2167411417088434,0.7867866793928533,
 0.9348861649712458,0.43498750309628365,0.014825729868775284,
 0.13122893172280803,0.39589580112219314,0.8160268857593979,
 0.10709895649763523,0.04365237039746217,0.6471784599004864,
 0.040003108031450796,0.5310745282098666,0.06282521302040278,
 0.6712722953360025,0.7080056460315939,0.709114135460424,0.7471174340083735,
 0.412341104817904,0.17901336547763402,0.8196485129173097,0.472821998784383,
 0.6924612239137173,0.7796772176602427,0.8883197616893627,0.9117316109915949,
 0.6666913646335922,0.649222909960391,0.9325896603039554,0.45600097687099195,
 0.34490935444163984,0.7768871809010202,0.13766517739062345,
 0.6595178812205328,0.4398460722813978,0.6483086946059797,0.01486016478404939,
 0.08364209082192231,0.8015544244887527,0.8328826000156344,0.7025270574180584,
 0.6137107311877529,0.9771225378165032,0.7706753378318714,0.37752775947046335,
 0.6325664540172944,0.9847979943410874,0.697526837099926,0.14208714498317387,
 0.3961573061322705,0.8643539025638505,0.13881085368657753,0.2659055412813792,
 0.5138989291695093,0.5459750962463248,0.5687006724905216,0.7720236716595467,
 0.2311413949359794,0.6037175412980393,0.8349702090269692,0.22191987925462464,
 0.4951013745661801,0.9977667544237795,0.08831477897521656,0.8907415176774393,
 0.5154581644110735,0.6203666820191187,0.507240609389211,0.43104490317749955,
 0.45456074487360376,0.812528412513081,0.1871552427566665,0.4488825872185266,
 0.7421553516595906,0.6428924341026766,0.4389557141284324,
 0.005304927090070244,0.4104121818440323,0.12713584264262978,
 0.4141058410234737,0.13705249011744147,0.148298381697077,0.926904995841113,
 0.2654862828094351,0.5435178485683728,0.4775637484931776,0.9295066274678145,
 0.9129878534905745,0.5897331763858109,0.3878799678243704,0.5125546377110862,
 0.7421475241768908,0.0961972542118339,0.03771248071717448,0.5061772226799879,
 0.8313181949369447,0.7504437613062936,0.5836952677969024,0.8006864655454117,
 0.40397486977216235,0.5106668518929625,0.8282929997645985,
 0.14577249870581144,0.5803240335215878,0.5831402373917906,0.3921605493867891,
 0.5125708415765811,0.23936539860688666,0.8032860768783641,0.3841388213379625,
 0.6763203547600839,0.5816369176044432,0.19058690278200363,0.9649303971460701,
 0.8702123467669681,0.6884080296755475,0.2029347965431746,0.6298022336600477,
 0.5824078112588015,0.9066752352325138,0.2600939108318252,0.9076906268170244,
 0.6399950857215089,0.5578540407806493,0.04127340987943451,0.5474340829429408,
 0.5775976898073347,0.3275107119719809,0.8929680817397617,0.21420495914379134,
 0.5554556007800983,0.006016632335324235,0.8505173824914307,
 0.6137969164631843,0.9099844546264239,0.032109073674879585,
 0.7226416564888242,0.19040570020471015,0.3329846608317808,0.5173184275065169,
 0.11559188737890647,0.2758406675184416,0.5043921713471482,
 0.05127786556799563,0.7917058758224984,0.6932051777746076,
 0.06944585946417459,0.8946476185257373,0.22656234871609893,
 0.9866578281128734,0.690884968770217,0.9080233055277409,0.35268336782408005,
 0.3131757306391767,0.12418540443569359,0.8603158769336344,
 0.34500572854962486,0.7281151283643607,0.29141949310551063,
 0.4117045570553993,0.5291464993087788,0.972818189808641,0.8074580711595393,
 0.4559259534584015,0.7238372577901449,0.4171499733600228,0.7522423455758552,
 0.959872662883,0.9037769644453466,0.07739758213038139,0.4080784637011634,
 0.5565144538887781,0.8450227855233742,0.5874473753426311,0.33946042824178235,
 0.5053882950286183,0.554777440996439,0.3456264388403091,0.18583714268069962,
 0.9961330792671512,0.10039157935551846,0.4745897785840347,
 0.23951430227112314,0.8386870783115664,0.7684248078043058,
 0.33221566412820613,0.035242704942340275,0.3257795291338119,
 0.1289002612295833,0.5337206441530865,0.7396801170079137,0.9303059812945726,
 0.3240544297812957,0.07736244173224383,0.3973589344366779,
 0.022684305366577617,0.10335972045263564,0.31766326109101084,
 0.47646447564350203,0.10417407814999191,0.1367809262958325,
 0.7570188436361369,0.3038278486108146,0.47679556000036627,0.3017206820504277,
 0.32378520909138986,0.9313018999917426,0.2252657274971035,
 0.12484442118832006,0.24852294762735283,0.059105236549811146,
 0.17826183540023544,0.22410073874581604,0.6219547159899812,
 0.5444696209325048,0.37436419978181923,0.3564345366682382,0.5569347819122241,
 0.14812730560554876,0.1861522867615133,0.6510357081242907,0.6394868122459166,
 0.16580545456114992,0.5038585991761629,0.5001841151946058,
 0.014660038004665421,0.3625208978772425,0.20779235889332615,
 0.256564974653149,0.05211949544091188,0.5240243767430262,0.44268436561678826,
 0.36537766092155666,0.7020446721942325,0.07153239961156477,
 0.9891417345186558,0.03712003085939697,0.5871068469186047,0.1378783228997873,
 0.8985996554242835,0.7909407044389456,0.294540746943039,0.9257070809248378,
 0.8002513448773394,0.9932543469331215,0.6627949471288518,0.4122348049428385,
 0.203157802380739,0.6661153083771911,0.7218352080742723,0.559011384473076,
 0.3777198001391392,0.2173353955613644,0.9181506954571206,0.6834626285490202,
 0.11196821304306614,0.5781043477026145,0.35990538676638695,
 0.06826868595987867,0.5576445837628473,0.5423638150549317,0.7014462354160499,
 0.518164940913521,0.6007152949049078,0.3633552831517147,0.669353485729375,
 0.7714182856171673,0.8283526219507972,0.03105423964371068,0.7047812997956513,
 0.22972602520516183,0.839216783437079,0.16930811519519473,
 0.28988252452848706,0.7866509330818049,0.33607420480496364,
 0.7067946586865493,0.9896723173817547,0.011078731252633811,
 0.1255018142540537,0.9938227417932813,0.6862232342390239,0.45662214682854096,
 0.7734547607919727,0.301424539938697,0.10671604245043254,0.5653051098086898,
 0.6244109277625108,0.922438130046037,0.575708035994576,0.14045376473142446,
 0.4453217828303665,0.44603171050966983,0.9458142795250032,0.5310884102713623,
 0.2562234344634493,0.7845411416860302,0.7374483167113465,0.14715972489655849,
 0.005801622404826401,0.35844211677924176,0.6860321021925926,
 0.24140469824861133,0.23200620982766507,0.119095428279968,
 0.39115420552334224,0.5822232610990428,0.5810331786923414,0.313076723366551,
 0.8865050615024737,0.7133180562460586,0.7349353465091282,0.9517212034657064,
 0.4323358322550739,0.464254619350852,0.159754359684684,0.8912823810044352,
 0.971736254580318,0.07056172909017544,0.08922161868882772,
 0.13300220546471708,0.7499779898750781,0.7039759123666978,
 0.49399654187083186,0.19579953178863596,0.5425343767012492,
 0.036601383122165765,0.13507283788021418,0.33841593574349604,
 0.9718937178249869,0.006155043924824466,0.5986988196666729,
 0.2954645230194729,0.1978602762670051,0.9656524856181559,0.7127394681057122,
 0.3651744836160262,0.1278856939065024,0.18554694794795568,
 0.03302481834147042,0.5202535491392547,0.8568146050223967,0.9024235270857314,
 0.2833170220919572,0.38889293394860924,0.28342412314081644,
 0.26104987133943147,0.13084700900503954,0.851136805768992,
 0.27928647484435276,0.787098060663642,0.8905089672715254,0.3238865576296438,
 0.8705795061823404,0.3627428272779596,0.9995664470740485,0.9522073048849127,
 0.2670890753548586,0.05063338127130945,0.6230605999557668,0.8465664874561425,
 0.2503322650476658,0.6285856010517429,0.905165100602382,0.20106494389863705,
 0.038290047836029406,0.15121461340598763,0.9968303907528404,
 0.3391761085503242,0.6000665981298143,0.8176967291665517,0.08292698920785213,
 0.8291828948197546,0.39252080641323817,0.4329896779149778,
 0.032262810154554344,0.25442990192408366,0.5011327225079135,
 0.8854305027845413,0.36572064205843713,0.6409871699158672,
 0.17182580115495805,0.9776684096736764,0.417304473113727,0.8293413746135307,
 0.1393546126947255,0.7630576505591826,0.8294357344901149,0.8988410928322146,
 0.2104212996743695,0.5548163314428424,0.6081379098760565,0.8617483452998207,
 0.16565492431461692,0.7176821544896986,0.953056076179691,0.3637598667524914,
 0.7721973528026792,0.14967939691155252,0.0010808436818572353,
 0.9438939343043246,0.38233292959990917,0.47750352560271414,
 0.9650410712291021,0.5028801781626416,0.522763281415438,0.7268144576061603,
 0.7761927065147052,0.285508951348449,0.24049840952822366,0.8800780303082181,
 0.6253143220787334,0.7147777957345973,0.09499214877366091,0.8102766733681046,
 0.6830981434732794,0.2806823718940539,0.4889246836937261,0.9134166484045707,
 0.4609543921771869,0.5280382821971675,0.8075872300698369,0.6300741703763906,
 0.8862559034805776,0.53471529896846,0.9466757826775406,0.8485706046069232,
 0.3896663311733435,0.028651976271161117,0.8459132216219207,
 0.6218969648839154,0.7155715521688594,0.09176985181618891,0.9932488311031572,
 0.14925232429011326,0.9468545876674818,0.8736892390761374,
 0.34539688592827744,0.9780264896178297,0.043339120472975834,
 0.8419756806097978,0.9353057630773589,0.11090344676600783,0.5748291314437132,
 0.44485433705419075,0.9923897746355514,0.09655744144918699,
 0.7740612412863982,0.6858966452291465,0.9737729990891213,0.4669676901639903,
 0.01365727144000528,0.8077141791079651,0.3630775509197185,0.4286106805277409,
 0.6379196416453075,0.513893784065096,0.7661628434440724,0.3780708703161847,
 0.8589153263675687,0.8857714850217718,0.6764713870702008,0.6083411524904992,
 0.24921546456280097,0.5686369763983465,0.525493231009363,0.48930390473646346,
 0.9888504043506154,0.3274472629962394,0.9206143949577978,0.03762765454097983,
 0.24139292716410543,0.2290339926288013,0.4131317894330624,
 0.34336167105411164,0.6310115003935406,0.2931783096179129,0.664941187667644,
 0.017682383037657345,0.8999665264685865,0.574518579872588,0.401930564667481,
 0.3198262707405939,0.5410863448400902,0.5643898746034524,0.717096107101513,
 0.054648320380503046,0.8855291218348094,0.03594205651456117,
 0.8627447680768308,0.008763848538134944,0.5380638929521242,
 0.7251372612839445,0.07698265108036662,0.6845219373587313,0.3401235475225105,
 0.039733515840546296,0.2119264317434686,0.9286617449068433,
 0.012810421990435916,0.924647262896356,0.32312722034743746,
 0.6187145952457073,0.7621916453862814,0.5005003327992416,0.3672650937272177,
 0.36600185021606724,0.6745120146171442,0.06467463009411278,
 0.5803235405740658,0.4820512430290752,0.1028787866075539,
 0.028296530744202242,0.31745181130353384,0.8586722539190975,
 0.34876396521998243,0.7774818322484888,0.21515560266763367,
 0.4767011225014752,0.21313484419168793,0.5464642840720718,0.6168995701258949,
 0.5607829395017445,0.13611483059366658,0.7372424716673218,0.6526273324464827,
 0.005813891198002907,0.19086868971606497,0.8787636982098218,
 0.6841640983776491,0.8838443809005878,0.7838667398033669,0.16863547867461104,
 0.5433545161998679,0.3724323711001164,0.8504125231870323,0.34287350917502324,
 0.05026171142526625,0.5222565255332702,0.19001371980130544,0.381503287467694,
 0.8544898645207605,0.43280244646316646,0.0522912936860116,0.4410520283815209,
 0.46802120179647355,0.06792523857922605,0.25411478742861493,
 0.6665170431015154,0.5836300275784785,0.7076765394823277,0.3714285305617253,
 0.21549574706127705,0.8042161826664883,0.46763304629887414,
 0.34180532046297885,0.9940823090554126,0.45192846219052396,
 0.03489305792947961,0.015990752904884165,0.46886783209431915,
 0.8624151032035454,0.7365102384980236,0.21180051601620042,
 0.23488812145827564,0.19384544058990072,0.7129293958668388,0.699686099497182,
 0.38272934283613325,0.785686811461257,0.4273224890308903,0.8330650068897588,
 0.1209555158036073,0.8895008005908013,0.13630854674575632,0.4904833671774319,
 0.6426925616035569,0.6992014890309143,0.2636142038741812,0.5283356697824737,
 0.8743578950529252,0.5180127884244398,0.1586292600310688,0.8928040263096401,
 0.6587824473283386,0.009617727366558748,0.12066983849132873,
 0.8105528309525087,0.8026109520087272,0.67606161421371,0.6046962276316975,
 0.9409372762785416,0.41950808507497994,0.13517852896691462,
 0.2649493886507186,0.4294849189623955,0.7371119639476198,0.05510458706791699,
 0.0063199837635873335,0.6120992745836481,0.2771634638239794,
 0.38850259151830024,0.20936536325766986,6.962602501869863e-4,
 0.2567447145004813,0.6298823192265208,0.9197138498414037,0.8936786627532604,
 0.3743812464854298,0.4061789116945993,0.19304820556802738,0.7581137599155219,
 0.4726821999469617,0.03547550687187184,0.3506014404492592,
 0.37690763176215214,0.06452379925099849,0.08384995849391907,
 0.8736952642717207,0.07931175815935054,0.7330628463463401,
 0.31873741691201574,0.7659052867948085,0.23231500526280524,
 0.05219856587170635,0.5995644099994277,0.09346927966557372,
 0.013413193360829467,0.7864414133538336,0.27453115055448174,
 0.6633344937347957,0.3803650459000838,0.6603747157768286,0.3664843690685281,
 0.7546516402725008,0.38944948643615884,0.7399350212337149,
 0.36451435037742663,0.9213142487163231,0.20976629842267158,
 0.6790670513473718,0.9318824817966203,0.5061629187839407,0.5341135168872119,
 0.18127176657890187,0.44200059536576863,0.7447265198718565,
 0.9913969116488306,0.44490450070624254,0.5158183951952747,
 0.026485782922301748,0.766535520403102,0.5142385216061192,0.9062373860817923,
 0.6432021306849021,0.8901025635914155,0.5821209731374271,0.17797333858572395,
 0.9594865896379899,0.3827122121423211,0.5605674217338149,
 0.028235671859340172,0.9282502325920634,0.4225065387501351,
 0.7518770468261688,0.6676144449354557,0.6241432784527391,0.8333495837677217,
 0.08233892415911015,0.9841947693169836,0.7536174453465324,0.9238648197511512,
 0.521927267951929,0.5704655945536903,0.9438565683222166,0.852082700614948,
 0.29373449006291286,0.6961273055179759,0.773904788418602,0.7239511053230345,
 0.8636596405930891,0.16526484988807655,0.17258377350483134,
 0.7822213521569349,0.4381394721013757,0.35421042413090587,0.8079215717186393,
 0.5856421467957581,0.7999031530851579,0.14301379969961958,0.8096752017810764,
 0.01567714966782563,0.5294967179559551,0.9654808969949621,0.5939873228482445,
 0.5126637189762894,0.4701924372258873,0.7116107039830384,0.3738610070229694,
 0.5789374995163193,0.40151555228589964,0.32929408799726334,
 0.9354029216496352,0.6390383122434917,0.2805884684135571,0.8695526992067597,
 0.8704875092568017,0.4081465008951619,0.8353912323093166,0.4075331838215246,
 0.6791411416608234,0.15661555735961952,0.699726164211148,0.5237604530779797,
 0.4005808358678917,0.9530134523487755,0.20818729641868328,0.9023290328289846,
 0.758971152924597,0.12039410829871844,0.7583660531746415,0.5610086411901641,
 0.7266470196366532,0.8944275644967558,0.4094957162865769,0.4935620699452671,
 0.8470461022435979,0.39654040845686156,0.5747921549151809,0.9576836609267491,
 0.5557872869598836,0.2546237576248801,0.2864222724447494,0.7590592254748286,
 0.2333755597659416,0.7455931434326494,0.5164172544778621,0.6310552356739727,
 0.7374991271548096,0.3086051199914154,0.8264806176451293,
 0.044079949471744984,0.7441992213237947,0.09403282568637117,
 0.7098924053811886,0.6371886991133284,0.41758265961335805,0.6950098696457097,
 0.41034162076381775,0.03672171246880673,0.8224850388079499,
 0.8713977317517515,0.081645287867527,0.6055866029058518,0.8957340494792916,
 0.12125671397265547,0.772089781723687,0.8588757068714731,0.6976465649172319,
 0.33934667452581835,0.3203079092445478,0.23264882217720328,
 0.5792712134666312,0.11926936171330549,0.8904028745470961,
 0.41000605608156804,0.1200904387046714,0.04570756683438304,
 0.36845270735046187,0.6186778130476059,0.613731470782725,0.26247158927006775,
 0.050736204686487985,0.7521751654099277,0.057461107675936596,
 0.03978469010719499,0.8723825349147633,0.5660808844892775,0.5045908254967132,
 0.2611092288875486,0.43324887904829357,0.1861594378992233,
 0.38360086430564655,0.9729631160568597,0.13666123234482364,
 0.7187118111452961,0.8224746431094088,0.5527528015610509,0.9016774665922613,
 0.5057065312219428,0.3998045269750421,0.14272477906567316,0.6019637386785466,
 0.9085623900397615,0.5856188171657492,0.4613122026250548,0.5288466173331576,
 0.16544382774348054,0.652547625071887,0.9183905600606557,0.11917139629957951,
 0.5381818514505783,0.42166173623095615,0.03652916911679005,
 0.20685546908424401,0.2352021384844355,0.6969162314707279,
 0.43945381818354656,0.6345587115838297,0.9698139664476357,0.529035699398877,
 0.03335686926974979,0.40834073011582983,0.49532910369768657,
 0.4923882547894749,0.13428172865504784,0.42715577784347847,
 0.2725350205952757,0.4399326347718575,0.9572883679245979,0.6669691556300769,
 0.7414498616806372,0.8465383840404319,0.6303802841095991,0.8944772982996998,
 0.29115938056354107,0.5114699615282149,0.987326420121278,0.7416029832863393,
 0.7736773896915459,0.9257247754027996,0.205880448289269,0.3861371188288877,
 0.3325378262012175,0.41497251169085514,0.3986561342802263,0.7013082350912472,
 0.10153009276969194,0.15923415332148738,0.9189510650193884,
 0.25834990300685345,0.8021123539130652,0.23925025012244583,
 0.1456343291277138,0.7397867472066926,0.6950088284320713,0.3406846129728536,
 0.7572539417285987,0.2550894684888262,0.6862099337174965,0.7867007071976493,
 0.3848942087886622,0.3702183020530634,0.3421203524577078,0.44337130624100674,
 0.02394298609993406,0.09556746485973588,0.995728427706005,0.3304474029430087,
 0.2451037147957601,0.6880006265592062,0.17917156773428577,
 0.025426957855102295,0.2955792207878746,0.18164626475545542,
 0.3663126283967951,0.4477003429769487,0.09746295817584871,
 0.12424111124854587,0.1716003034678355,0.2353959623689592,0.5279052348878165,
 0.24853451533513315,0.15192444173058905,0.7704569420472092,
 0.1422886347712975,0.11937589887040512,0.3991590449828468,
 0.26608384623210246,0.12371730225976452,0.38291364426642516,
 0.11338205549311009,0.23537460410814648,0.38525661090974783,
 0.26904598717014017,0.7544372770067043,0.014801017406895323,
 0.515288735204309,0.8599726688052781,0.7051316376294201,0.026667866536823626,
 0.24368684944006336,0.8570847233712471,0.8853336016418194,
 0.053069903339668545,0.40605338980278116,0.9039491231278962,
 0.07322856230684605,0.7128909539066172,0.9861791610663866,
 0.07778570281911079,0.07098633217420769,0.3265205659344752,
 0.6500148068261716,0.12259745706114855,0.8592777835320909,0.7795435439340523,
 0.4898752580258754,0.3717666851864959,0.7035822654447849,0.04407197845902677,
 0.3519135347113592,0.6607201070239052,0.3624242716983317,0.5286241258726658,
 0.25880683327524323,0.4957786068468921,0.6880153966842656,0.5327649821872447,
 0.3293108066657294,0.2373450612700042,0.38956157761890753,0.0408848156398226,
 0.9829353060703755,0.8249266559645981,0.33103958840419234,0.2573239925215832,
 0.8474425558312296,0.41848009979256506,0.0825303227178934,
 0.30426636961572673,0.7565770819963946,0.7633564362158338,
 0.10002482164836213,0.8056180398348324,0.252838906940954,0.7359401205057574,
 0.8083662631633521,0.7662843474457985,0.30313882523431324,0.3983552780947668,
 0.757556196120527,0.48960355670756095,0.7944813245597856,0.09694862695047246,
 0.11013293608603048,0.10965037405003097,0.07565273951934315,
 0.6289465891608945,0.8775696523798082,0.28417463245763275,
 0.11499360969048877,0.4495894223101916,0.34502353408206954,0.922625457307367,
 0.021249013356548274,0.2593044109022924,0.11474448310309249,
 0.9763900459723536,0.7096063569417866,0.6705159182588583,0.8122571698304558,
 0.10288208489763517,0.49739260685233144,0.8975353908265424,
 0.028732417273528044,0.7222254666234805,0.5005884514997058,
 0.9119691649315921,0.41362243999619797,0.5366674963344279,0.9947077762805934,
 0.03273823576645141,0.21917759416329174,0.11200741751682863,
 0.6502153031203293,0.9671101542934134,0.6984956289937978,0.6480503540182474,
 0.8614356196598176,0.7711591606560224,0.3569822139488217,0.02148469108153328,
 0.5425199179730424,0.11056037438281463,0.18519217311738168,
 0.8610987423901256,0.39312091040752606,0.10173673373189729,
 0.48556678800397957,0.45802527709137664,0.8736294916228681,
 0.2652139803233402,0.9606423605332586,0.7408207921922263,0.5143695957722417,
 0.47152658225584554,0.13832256448482338,0.35164468557878825,
 0.727024221309938,0.2270688813476539,0.5780658047664621,0.9406551679547097,
 0.2339923454237448,0.15223248553998725,0.29629893160467446,
 0.8135187495868528,0.988938761935478,0.9178532449961955,0.4638514735854484,
 0.6969204814776087,0.7944580915441606,0.4999676244943265,0.9391747004059994,
 0.9663285407341187,0.056476990305210806,0.9172395333343152,
 0.6932149450212399,0.7600319308821066,0.8251694080778184,0.5650425568501897,
 0.939389337925512,0.7289708512527735,0.05871319418034893,0.561242851963339,
 0.19086861777728759,0.08602898438692974,0.46420569575818893,
 0.4416301671359172,0.6328506226275493,0.24312682523905527,0.5427321576717693,
 0.2019252465067647,0.5178404822732953,0.11975591459869406,
 0.38249754193738916,0.2696246720488227,0.3745972860045007,0.9733476883983841,
 0.28192468496002987,0.6313296227530524,0.19105315220399843,
 0.9011151419087333,0.11238644616319271,0.6266948011107872,0.6664114600091459,
 0.2614403624590921,0.028737840161216166,0.609307817170011,0.8925065290776701,
 0.3033101254609152,0.8222737777699415,0.9803340063828462,0.38583459080507077,
 0.7203854840373802,0.42603687370494703,0.3056706479248623,0.898203119578374,
 0.8868198223178841,0.9072218283155671,0.40416235122700916,0.1488712923836879,
 0.15286839942605512,0.6495624496388699,0.7863763049118838,0.8140205826227468,
 0.6456142823343846,0.13783625269559674,0.15345079772650327,
 0.004328315431977092,0.2015704248671103,0.348627754735088,
 0.47991142867924785,0.07226412994098397,0.16225638736031378,
 0.7820866195492989,0.47576178528119295,0.008680019222000235,
 0.5682527132270849,0.7552147474078268,0.31443253416838646,
 0.31476937875296085,0.9019050374908567,0.46554275327176387,
 0.8532194282585033,0.39065261026586595,0.8430911392211382,0.8209358055055648,
 0.2935497659964813,0.8376082466272883,0.111283123885348,0.7349309027356543,
 0.26583703930257196,0.426069071901618,0.5018791490650917,0.8113886132379935,
 0.6866617414572878,0.8432649679600944,0.6654903979429964,0.7831009043742513,
 0.34563340738084747,0.5712715149387977,0.29048363921568277,
 0.02333295705572802,0.5886040630382002,0.7472285988141018,
 0.13745500292829838,0.28368968946589346,0.3548186016030671,
 0.35285946596467876,0.36631880407943307,0.0176896494453066,
 0.7205104979101322,0.90213223522522,0.25284511668211895,0.16027617525229854,
 0.949415116137241,0.08770229463566748,0.9961272940397246,0.1163641572884564,
 0.2900045526817352,0.4834938184290807,0.7816811058347791,0.5627376495479967,
 0.23737460446962233,0.24911091837450372,0.12112714879671271,
 0.9790975831564263,0.7763415611604644,0.048627949563858075,
 0.6428213166318478,0.4701885457995131,0.48890205995804237,0.123388124539077,
 0.8242757387856003,0.0708694203520931,0.06070766722966903,0.1381344931373114,
 0.3274788813902453,0.6776154837150378,0.5262849863824259,0.5284793019467917,
 0.700512251290715,0.9490334971461666,0.6960349406078978,0.5717077223176887,
 0.04493223795243262,0.73139414880854,0.0636833854679576,0.08120319537190068,
 0.9042965787510404,0.44926076543669524,0.44911494235202165,
 0.12048934454522697,0.776933587568827,0.6227998830379189,0.3792248006936658,
 0.8978073527241864,0.14723593282933634,0.4174447075968025,
 0.08967634520952994,0.34927046881786195,0.21095793165566346,
 0.7188555127091165,0.8862298349589678,0.07975885361515467,0.7641598670255525,
 0.7050007890833708,0.8299377931936743,0.7026246725738359,0.17996523650873708,
 0.1587792545780543,0.39072131658193676,0.2968026982952211,0.6474428788166795,
 0.9697873200716778,0.18029272790831186,0.6092179546702347,
 0.10017306806082238,0.5207132755917433,0.8225498311122799,0.34134732787585,
 0.9737119582403404,0.03513643583580239,0.6379196418325093,
 0.29380933882122084,0.5291938644230941,0.46263270131788625,
 0.31560791946837785,0.17416662406944128,0.7460727602784619,
 0.4308658173719956,0.05803683929983261,0.1353147897949104,0.4459173575639205,
 0.42114891392783305,0.11464359376620104,0.14835332959944403,
 0.10559177971133638,0.9500879750756867,0.8421758859942091,0.6502324129811609,
 0.317396364602213,0.061701369399864925,0.2939529197875318,0.7145462016435791,
 0.42263507056029226,0.46577597606335686,0.7847642155129769,
 0.21696613495249129,0.5387958625763134,0.7538975937198957,
 0.14372099870725796,0.19959218080663876,0.12563400908472655,
 0.7283949627788814,0.3621481506514048,0.2365066795196783,0.2303282311758288,
 0.7502383797116035,0.0435699433920087,0.24548027447954912,0.5161270939935171,
 0.635058371459607,0.8010998841954236,0.85013589863963,0.7876105989778494,
 0.5270073549063619,0.9106675201217838,0.5991438159781426,0.7275115476353922,
 0.09921458551189499,0.5818973817832749,0.2903546738862115,0.8462288437458128,
 0.2866507824936263,0.8655674043046023,0.19636678487063652,
 0.28793610724733243,0.22552711859180208,0.7774162359172542,
 0.39837521071409554,0.2928500426090662,0.7805567304054457,
 0.48152941617337675,0.32991383965893906,0.9931397067852367,
 0.8154669179399862,0.24219612934275203,0.7911552169893672,0.7319408091175713,
 0.22560306351825288,0.6464397539962778,0.08545604463101597,
 0.4991057430188923,0.07898847083724903,0.5786689291197026,
 0.39469332479327013,0.5248338885298054,0.07793286613810047,
 0.0390246654960017,0.609451385345332,0.8298481093120134,0.2642044506461937,
 0.12580614890758057,0.18543376225156438,0.8463587802316029,
 0.29591809256770607,0.4601696264537203,0.26061310796038606,
 0.23228694494369218,0.5306370593816111,0.05344805231517025,
 0.08971909147727519,0.9275848357793253,0.0940429944374237,0.510740119896625,
 0.1860604661525027,0.7973669970223759,0.7876614437007756,0.6539471612884692,
 0.22595416037467697,0.10188827868936556,0.5553791366161104,
 0.6387758025457331,0.017502095900045478,0.4513491717000232,
 0.2129911208648837,0.9000782487999882,0.09909371471065942,
 0.43232429343409007,0.09657105583788972,0.9400466654022763,
 0.6753198675536871,0.4548958320681612,0.3132978384589945,0.0899919537756888,
 0.20581645370431279,0.38707495530064784,0.11671984339994612,
 0.04741580520355149,0.013631766708647852,0.45190866745062286,
 0.6050800926848552,0.1442789163362449,0.5329509190015194,0.7907113086200652,
 0.6215317959786304,0.3911568311988469,0.6909645866223161,0.921225985485725,
 0.7191189155121704,0.4986201138226861,0.3197446454451415,0.5100677965893746,
 0.05899538652280223,0.3047767007663923,0.26781066139228893,
 0.4858650023228219,0.47476966189374925,0.18262248554568283,
 0.8402330333560484,0.6744170272837147,0.15233496104617994,
 0.30402761128750866,0.11852806977713448,0.10582153269027683,
 0.8947844531146348,0.11479414817749611,0.13982770321933213,
 0.5103576027688833,0.8844055758433567,0.5328899309740845,0.26014747997375054,
 0.2057625909289672,0.7171086364419588,0.10939170431564416,0.313788457252436,
 0.11292937972158235,0.12268892945512133,0.7611067310559884,
 0.01914936950766344,0.9168201074001612,0.8147017500613231,
 0.17467582155356653,0.7185105873021687,0.6384838711909047,0.1860917957382935,
 0.8853563653729251,0.4613646685673958,0.6568843881464638,0.8653051916032958,
 0.18672009803281775,0.07159765827723874,0.36214796454041664,
 0.6605268264706745,0.25058334457969567,0.6593784968962724,0.8861718605879239,
 0.33738121106596997,0.9334582263223755,0.1820232127305258,
 0.07038639892975807,0.07625326016555345,0.3553358657965934,
 0.36622633381803793,0.3455617999915602,0.47967819741829965,
 0.5340171929882391,0.827627500181796,0.5171501440310953,0.5479617775904634,
 0.5944664269964935,0.4572404146792306,0.18101413386526133,0.0576260756700514,
 0.684292335490802,0.43079307302875214,0.27411869745366113,0.8987930675751971,
 0.5691638757120144,0.9149884254466327,0.5475795975844179,0.03499724075768029,
 0.5819317683597468,0.16081270218412735,0.5142354709556851,
 0.044199690263891145,0.7825711017004726,0.8302077865479808,
 0.6391168836740498,0.16001622034838014,0.6673223394824201,0.365644380427249,
 0.03212172791223411,0.6019465696811411,0.9353555434537713,0.826535083663864,
 0.3918459833485879,0.18741137169027278,0.7667730682456462,0.8396380539574815,
 0.24400212768939722,0.6785598598005517,0.021811304734220327,
 0.27918886120519204,0.4106633231689353,0.8929292590398187,0.8381187790394614,
 0.9377849898981556,0.33675743772292,0.14737179253354438,0.6156966558303639,
 0.1417742017637914,0.6147287897872868,0.6969742241439718,0.36597686747846825,
 0.13172853108794402,0.13531507914838536,0.12350382207910227,
 0.751691336978266,0.43336899186992917,0.9702770173242736,0.16922458044054556,
 0.2394736570182605,0.8920836063691033,0.07083767590985723,0.4751664302709043,
 0.025253442978061535,0.4813569130810159,0.8110256104647597,
 0.5537097328443706,0.9433364281233064,0.36089666358321026,0.8771547219381615,
 0.4919024575363218,0.7985207428089678,0.016696813179981063,
 0.33593990473978497,0.4193137042834597,0.5633501445884121,
 0.10727827980424776,0.7297579967718808,0.3575117925235519,0.6285962469651469,
 0.3473698561706804,0.7307470621011581,0.8956263100643417,0.8172105753785948,
 0.24519460003509386,0.11228838303937261,0.7727490421694796,
 0.2106923725154729,0.4762555236907732,0.2074439975888971,0.3953874652015137,
 0.7354316015198546,0.11764142855190751,0.33332492778157097,
 0.1271131960579862,0.47671101246112213,0.34028083910473206,
 0.8499741919727425,0.2854229629096885,0.9212235254726844,0.3558607221899175,
 0.1017955020948389,0.952650232430564,0.8108376443585368,0.19292047741179064,
 0.18636797177898612,0.18526542074138885,0.2675704673974346,
 0.8116131569783832,0.2707776766024941,0.395040788064161,0.5729703578722196,
 0.28220486124187283,0.6690498637008078,0.7052336901598508,
 0.16920014169932918,0.9652647953781024,0.07839658945783323,0.55986548730645,
 0.29910816524571193,0.3584461886097827,0.3821149431653592,0.3949608655885398,
 0.4410613841471518,0.1449425677034376,0.8530651243327674,0.20629385841375492,
 0.11502593262841332,0.719942193148366,0.48542361251185717,0.582844307827527,
 0.7943223617668951,0.2140369739730631,0.8429452381586465,0.7107775676148682,
 0.5985672440320249,0.5492614332735499,0.12812055647783938,
 0.19948922752713028,0.8920568397814512,0.07459967543299884,
 0.35098461309525086,0.6941581245301811,0.426768609530754,0.10930477498168123,
 0.044596978467305814,0.20862088312477534,0.5857723968158609,
 0.9449122482033717,0.6486875542159274,0.94370136123704,0.2291654585742514,
 0.5559318529799285,0.5191107019901272,0.4212310112015957,0.5937708068513166,
 0.5365729083267301,0.7438345291349217,0.015395299726412404,
 0.40450756232936946,0.6621252189109448,0.3506112134551037,
 0.22056731578947852,0.4644684530259673,0.1981690742898361,0.2842985125813883,
 0.28353976622892507,0.9329565742045018,0.5813981944381366,
 0.11566192546527032,0.004698599255334113,0.4204207082560609,
 0.9058918551572993,0.9500144349587031,0.9765863012533954,0.30456534010128256,
 0.20497121065270085,0.15252105267162608,0.954065749978233,
 0.25827603542875743,0.6449910395046831,0.027296129213343012,
 0.6620355855889696,0.9478392373303542,0.9859072620413311,0.5859274008566102,
 0.8525330551016101,0.2585384875185768,0.5130951949390998,0.32411785586462616,
 0.9448919474755165,0.9465856546301318,0.28435446759241323,0.9592375134626937,
 0.7534920537169802,0.7511628707357627,0.8475707683958233,0.8467603792819751,
 0.4189499086804753,0.2556841527492908,0.7313955567078076,0.8758606928151537,
 0.04982113812168576,0.24346857425620794,0.9135961338689411,
 0.7502167266230187,0.8500701744393022,0.5452041115824021,0.46625035835086104,
 0.5361596041497287,0.2456556715420548,0.13698759496997792,0.6110102186394203,
 0.4762894114563716,0.0301451849054456,0.242741314102208,0.7505039648228916,
 0.2791276663365637,0.8363649322254214,0.5757000971162487,0.3456575050910895,
 0.6979875998719564,0.5253804648944965,0.46927562595392347,
 0.10135415718930507,0.4034652737577866,0.20203570904658774,0.878959470501849,
 0.04985152114135405,0.6662159059800019,0.4185879393982569,0.8376158859652889,
 0.7567481324839963,0.670458681142165,0.08969344490288755,0.1523758973698096,
 0.41912970650195547,0.45556766082207534,0.05260597851373394,
 0.6184373980180373,0.771562807056233,0.3060196126020025,0.40257702504564463,
 0.502004175484187,0.09647390610715678,0.30358852467498687,0.3696894656912283,
 0.8903690937210393,0.7930099513015001,0.8099854477613849,0.09559607650965807,
 0.8471248210364113,0.10482128237892074,0.3198910192408426,0.8792109673171458,
 0.35135531233402784,0.6292521457606368,0.35302183758759853,
 0.6408258062097565,0.6198512627656994,0.6842081374981985,0.5236459700326818,
 0.5119335775490091,0.6967162405129143,0.2636897231644195,0.44187964798116897,
 0.7049007278256527,0.7290631495980079,0.9111216829366461,0.5091383149648473,
 0.009024184722611461,0.08403246230345518,0.35768166859080686,
 0.8236516800028418,0.5538368539122204,0.9068589499676484,0.46667065081873005,
 0.9980707716134259,0.4926517146154408,0.6533001818764848,0.46938977208444144,
 0.7867223073804571,0.5159321501782632,0.4614611780703276,0.47281723649039065,
 0.11164917422505605,0.40063370025056066,0.5097430764227482,
 0.25168461000921005,0.7748216317267084,0.8840239514946181,0.1649976921869608,
 0.47988544844059033,0.6448277228280502,0.5366886064744951,0.7845625103137769,
 0.07269246335658075,0.8411863680992758,0.3639252391690192,0.880837322207707,
 0.6925477784978605,0.6005220299725967,0.2072993277565338,0.3554699677146176,
 0.32704171124777115,0.14601958893794031,0.20515424856869902,
 0.7092600895082459,0.35286441363107335,0.5178542318611133,
 0.00873288997009869,0.7867339946973283,0.9537929071654622,0.957120152218479,
 0.7459643798164213,0.1438280319245424,0.5423029691169091,0.9145213464598102,
 0.9779862937450114,0.9758975113820577,0.9851779494995289,0.5646413233614714,
 0.8841909471763647,0.10498540852235927,0.3739148687578675,0.9516862888587183,
 0.4313606611236681,0.9002737653328623,0.2651339255485603,0.13145748444079264,
 0.34626070240464113,0.26569484617101835,0.7769084413778935,
 0.3841648218705165,0.8452703710345598,0.7473874253451718,0.8403161553854984,
 0.19537422619786626,0.1073345921820349,0.9632558174258943,
 0.23502264237677806,0.43019113220189564,0.6857191542722549,
 0.007570836946981574,0.11243840248447012,0.6292369495235904,
 0.5812250392855699,0.8322986420802277,0.1426068414964361,0.19337574971353522,
 0.5199908886260307,0.19804401398366,0.8635701363591892,0.06722046129543591,
 0.04784077780134044,0.36271875505456563,0.31872630903830834,
 0.6879767245015036,0.7620459642686371,0.14801300502503578,0.0579357224404915,
 0.2925046662813926,0.5828647382562753,0.9053330606106446,0.7823016675781294,
 0.05573487985942149,0.806482213286908,0.7569021288629031,0.33956191869333097,
 0.9490769954911882,0.106402681135912,0.5604236280458903,0.1322054816313818,
 0.9971461043773098,0.10276676714936439,0.4962649902769726,0.3121733008672358,
 0.9863615963298389,0.021882789581390116,0.8526035874900373,
 0.6622149833944302,0.15829368968717605,0.82803113787779,0.8263345256335431,
 0.7786663077101855,0.4605792984645223,0.21072834406455998,0.9192665117494878,
 0.7227646809227243,0.05986007347416866,0.40547980394371486,
 0.4620268044725212,0.5237154269579002,0.029020613190096634,
 0.33434213527314305,0.9982056260028871,0.2835610254669704,
 0.39151294075185106,0.8863376367419921,0.46898483389840384,
 0.9655366449429399,0.9056091737269578,0.8615388939688863,0.03456510147280001,
 0.3019110736274775,0.7718197275845448,0.7217935924729844,0.40376360406406775,
 0.5638723819400445,0.8716836145177329,0.5621256552474446,0.4768619523429283,
 0.917428668456356,0.14605243843439808,0.8612948970963541,0.9214781922007319,
 0.7673747902694764,0.012219657913107085,0.23363257234509938,
 0.9805226197635492,0.3917462457186992,0.7416070135726055,0.8845489236307742,
 0.3834431258163382,0.29358051298967747,0.2358233654334131,0.7663565024326869,
 0.09458851649732758,0.04224955211420511,0.6878942215877147,
 0.6509414555171595,0.4992461778767978,0.7608688477467375,0.0708743393003225,
 0.2959362326254076,0.4683603758157662,0.23772352041129918,0.3099857022491752,
 0.8969139845174408,0.11536907414119657,0.05474607350964511,
 0.4502063605188227,0.3674301871288046,0.23672041828500268,0.5254369637397827,
 0.24550849441344036,0.5534490396725468,0.11088714783656628,
 0.9861928500731891,0.08438041688728204,0.14464011110024944,
 0.45632431751556213,0.4635348538463986,0.8686791100520977,0.5342687777195052,
 0.32561483199809693,0.8953806092454978,0.6466198988814479,
 0.27071509028563523,0.16083747798357995,0.06683060150330977,
 0.7153195337430167,0.8213351584470417,0.2931797550206584,0.7053527420081788,
 0.8727704814096396,0.6598647858864304,0.06463733144624806,
 0.12478861283937115,0.05175615276144008,0.5344310839023615,
 0.5936531228941336,0.2366828004661562,0.9008618705185045,0.5662572430765359,
 0.8177322989256841,0.885123448824974,0.13898684267410055,0.489911282903179,
 0.003083506332133723,0.28177464665097074,0.4995534659654153,
 0.1416464362242853,0.6225220822062175,0.066191229761557,0.8585550966168688,
 0.38521245070078547,0.6133098751035809,0.3163844536467667,0.7109126518129005,
 0.6687077421160471,0.6100266426629132,0.8355245128045983,0.16652719716627962,
 0.5933996451470263,0.1283663511439499,0.2731578448873153,0.43798013187481644,
 0.7652604032968036,0.2707986250901917,0.5901982579866066,0.9565320172879608,
 0.8535583128400727,0.23097923901316464,0.8479228628360481,
 0.37897895619306876,0.5833130017821452,0.7109625874307948,0.1608632213920238,
 0.5379670596053734,0.9835712819157773,0.24272480913756733,0.6807946418801697,
 0.5005946602153253,0.1158528844459541,0.8586558623967976,0.35583335496184987,
 0.668285479564781,0.4447755701759416,0.7003260163632932,0.6591541315505504,
 0.21874700073709774,0.31073044023950935,0.1426702653637536,
 0.3729587277936164,0.44205364454318263,0.7270045324863776,
 0.09952100203523018,0.3307090366815615,0.9011144506520126,0.9151629011208239,
 0.053809264179472516,0.7201795594797564,0.15911466689863607,
 0.6226926865576332,0.9590703059875216,0.99906889191563,0.6514541230624322,
 0.16041965838766126,0.7081926948447668,0.15936794778491825,0.456052626894933,
 0.08167650844929086,0.8500896506432221,0.04304953578752091,
 0.9168466436681765,0.33063505978123153,0.6082787119562433,
 0.22337184930460918,0.48036191483748014,0.535284374092546,0.3650183995464239,
 0.7738158323576338,0.9828376661204985,0.02918466783698337,
 0.005233207992771716,0.5217553233927166,0.4632578553817802,
 0.6530504352654452,0.17459017058085158,0.6603621458291538,0.2727998960927692,
 0.3870799152978853,0.5011162401789215,0.2866394690319126,0.9175782584712985,
 0.1624420332132548,0.485374548482043,0.23564434920796884,0.6412653032599169,
 0.6725392660984842,0.6042601143082245,0.9774883399643685,0.7313165966982229,
 0.039956310264054884,0.5015183241901089,0.05153816444340742,
 0.42848955091347274,0.3629244045175394,0.44490916315948514,
 0.9058412723533846,0.9062501023203912,0.8233657722171087,0.9661085020323629,
 0.7723347168294361,0.9465676289234601,0.047724631996637235,
 0.25372018244609085,0.5602400646480337,0.3857842278867578,0.3271832461286275,
 0.260945537107775,0.12212543934831832,0.03123066919783335,0.3072794827878462,
 0.9009834648444466,0.8608106986176922,0.9592404326138174,0.3080955980034433,
 0.8338351804911217,0.9877788079170391,0.8210251882460646,0.9089522745050733,
 0.9838787240573892,0.25652021459403007,0.015868941798571234,
 0.6855040366327477,0.7363022177672829,0.7057851996362468,0.5808973472819692,
 0.38598485621143896,0.7955874543681736,0.4781052538948798,0.5709386811898004,
 0.9444923050959492,0.6799752872358185,0.7306151261009676,0.17459641879861332,
 0.007694639292647798,0.47766735798226156,0.31470992126076136,
 0.704404529428402,0.44071518511687635,0.7891284850588368,0.30424277316617276,
 0.1503069090586252,0.11074724958403337,0.28771346291196986,
 0.08001720392416345,0.37743507908981055,0.001823057727218691,
 0.16419380242168613,0.41737780006868896,0.16854080942841443,
 0.6125258943565339,0.4169422739616999,0.5662298068121898,0.2128364876973261,
 0.7154777549483613,0.18062100548281323,0.3554550434772391,
 0.00357852927820157,0.23330027095285055,0.5697769023473989,
 0.7256109762044091,0.7696413484057861,0.7311094853227239,0.8016979256990147,
 0.764449679488736,0.6390894701628688,0.8048161632182151,0.3736600524725604,
 0.705190727987196,0.7423770802454318,0.10479247752065812,0.47960727452080787,
 0.700344863142816,0.8827168569468515,0.3415915621613814,0.48265839325820914,
 0.4678242017172821,0.9058071667563503,0.527308315969274,0.28967671178142584,
 0.8753384578985174,0.45463077605679136,0.6350541319052984,
 0.37894367307747967,0.4977332562745621,0.8217592507406409,0.7371651392387942,
 0.4094111856301641,0.5169704280663023,0.16881425277648754,0.8548499196577178,
 0.7501593554749998,0.10408633063792816,0.737691517468973,0.7751848560968222,
 0.8617232351928038,0.8484704314785331,0.22124744986945166,0.9298450859224648,
 0.10134620408137607,0.6987360554960501,0.7239557273871401,0.8497318726196983,
 0.7233473353750242,0.7911041109738679,0.2678010675882817,0.9038611193250818,
 0.639657098692513,0.9909400724350835,0.22182157138125136,
 0.010385698254591458,0.8193408845817542,0.01758418291629582,
 0.7939466915115847,0.7750560944222106,0.12558694005859772,
 0.33644140017937685,0.8181300587273382,0.2671127909209846,0.249824539387124,
 0.7520048476399449,0.5320206794316248,0.73136222472832,0.24539983649530384,
 0.944634951601834,0.7157573882650811,0.766281508551864,0.26116508372434477,
 0.5014990034625473,0.9731934286252785,0.6307715494099987,0.6919991572381254,
 0.7067934503339132,0.2836743827761393,0.063583680171869,0.45267042714631633,
 0.5905903308914926,0.59628629089826,0.8482204822154277,0.7647037318798908,
 0.3867473498152899,0.06274140887966229,0.5270195224545492,0.7911041175247081,
 0.050143395339017016,0.9362875426822717,0.8545980685110482,
 0.44979414931365913,0.44522223356418666,0.8467275355159979,
 0.14107006909285114,0.7004152724857902,0.726531787936412,0.16980867730493676,
 0.9393710354451809,0.5891798774182188,0.22877170175692918,0.6653925678893582,
 0.17850665046286496,0.20022802444591903,0.4831549248842627,
 0.6278870204122299,0.791568842554747,0.12720684864389598,0.881507757683163,
 0.01022378392811385,0.11287076677643326,0.002831944916713236,
 0.36612101827864385,0.8868765684402862,0.31586811997404196,
 0.09939812708009343,0.26665457664181225,0.8257106332569653,
 0.9706588628128862,0.30173426726769903,0.10517183203707337,
 0.5982134974162443,0.7146745927206668,0.7896899677065573,0.7527229833683227,
 0.3011375448572461,0.39747667377046314,0.2930659205613566,0.6771243916349993,
 0.9986354252967228,0.27586907694443574,0.5525607197656262,0.7740231793307573,
 0.18686541004515678,0.18894280051017254,0.5002845351490643,
 0.07436428394518468,0.39056773265801503,0.17997523449698138,
 0.40397184442226597,0.6907341254499179,0.3622884791244094,
 0.21044560266336454,0.6004746636298589,0.6124978951585246,0.8087245078348232,
 0.20677693720666512,0.8259283797262942,0.8081147474979707,
 0.035496513932842144,0.2307140045451308,0.10045375201434359,
 0.25182257553418586,0.6595050116373478,0.8759712500271692,0.8919639890235399,
 0.51100966970926,0.002911521581975385,0.59014879385706,0.7745201591612108,
 0.3550087062984806,0.6359655031961217,0.9102021427099386,0.04970009477274584,
 0.944250816703789,0.4002076441312741,0.36010071001867827,0.838910302486763,
 0.9917174429810395,0.9231485244167495,0.4935900612495886,0.9407671235112699,
 0.5051074863772594,0.03344068803811284,0.5696904073791511,0.9690938411166724,
 0.5048499441631942,0.17067904475901463,0.2547268070903741,0.2654127671289821,
 0.5512028644969036,0.2503589180156687,0.2505953863009521,0.5696300999856418,
 0.959703314811174,0.37208798628869477,0.08424105991837127,0.5171399856466667,
 0.8483057901926659,0.11151404340916615,0.7491736224589784,0.7830460818702345,
 0.3203222240741672,0.7109907709960228,0.2832027247821264,0.5941811217807333,
 0.262688365748718,0.6715167604539046,0.22935786742467879,0.20036482537789657,
 0.8127889470766055,0.10345740625912347,0.5316131261096708,0.8289083675243274,
 0.006391238573796421,0.8116656633249666,0.9271407677363821,
 0.9959726899975919,0.0738897211797529,0.9806443306065307,0.5016797307809048,
 0.292310932818838,0.3493797948338727,0.6162754500319273,0.12784272658674,
 0.4685883273405902,0.9946875652449099,0.19141693104965785,0.5834219351012103,
 0.4268662061764519,0.07516374997058373,0.8962316942552458,0.619392921743964,
 0.949886587758879,0.5460408903254474,0.9865145668366601,0.04414533296826795,
 0.36168170283236845,0.5696931722309027,0.2338917869651237,0.8958364218877829,
 0.5217174638163421,0.14166509047294062,0.32895385808169053,0.833670089256803,
 0.1656217208199925,0.8004146295671487,0.8742145273165517,0.40980867272724597,
 0.11410027433707237,0.47056152071080737,0.4274601092849425,0.585664633954706,
 0.9074123285312539,0.6193532328564793,0.1168572856190715,0.3411885420481269,
 0.6084677047000387,0.3697612252981676,0.35696368283112867,0.6716012602573125,
 0.22215342276408656,0.14736331985428353,0.25483153275531445,
 0.43902881985518827,0.9904797016804129,0.7776127113728141,
 0.18872377160114184,0.1450934546599284,0.9182232876682976,
 0.28635684779372805,0.46554041221532083,0.4983354022406751,
 0.34353244678851913,0.598387003029953,0.5741107431356348,0.9089345990927935,
 0.5553452358744074,0.2968544283536926,0.9087553635735037,0.9998055712349869,
 0.40464222248568815,0.5691008008202953,0.04503822890453757,
 0.6646456098858934,0.8779888520519853,0.6676573133643678,0.07459489082490123,
 0.19738749042376935,0.5810056881274173,0.7647686238725082,
 0.41309382262690375,0.22538253147253018,0.4768673776988207,
 0.5153400028479335,0.15323716372958485,0.45876909535378263,
 0.7601013295987324,0.9022089383215406,0.17365639845487857,
 0.48012262546960227,0.8841331862136607,0.5992940086350956,0.3857307823163092,
 0.7591141177149403,0.03675749646228421,0.70089027962324,0.10754057030882169,
 0.6599842892293656,0.6254648110071124,0.632105169086089,0.9330842516801238,
 0.600443230319211,0.2435726434713188,0.44618663520645196,0.24646264035089183,
 0.1897770689195608,0.10754104632925288,0.3925532714515322,
 0.06809234629860828,0.1428658690588902,0.28746817447160744,
 0.4542126596122561,0.42946289085673883,0.2779226667115491,0.4868316269190479,
 0.9948517373384924,0.6473703913820658,0.99704039623039,0.5609603262379539,
 0.24885351432659342,0.4810724346550146,0.36701386441178074,
 0.3309440423897041,0.28241382908083146,0.7625810728760918,
 0.33128794777583037,0.5800343566279441,0.45239415793308746,
 0.007684827101095082,0.06850444462341587,0.7120700504952994,
 0.8303752615442834,0.4431329402883256,0.004153887354406249,
 0.6729002943837584,0.04710906452923269,0.04821998128714011,
 0.9551925851103382,0.33445879249740784,0.07193836385303742,
 0.36040906444352405,0.6291044163493327,0.8337901994850573,0.2251460326113156,
 0.8468244012684304,0.6115167536417755,0.6405966227122769,0.9536343745827093,
 0.38959864419070744,0.40087641793695994,0.4093632596160476,
 0.2356445288619755,0.3338304528967462,0.44723171707285947,
 0.03294680270723083,0.40640908337525417,0.29525632929216516,
 0.6807738516105311,0.29936494600456987,0.04286465055215927,
 0.06234451866109181,0.469004142288187,0.17222120311211153,
 0.42231367936031206,0.36051642654479454,0.4690937651713104,
 0.9459438396490908,0.380926153172224,0.2043096033692715,0.14989975513568143,
 0.5899444260985373,0.9429006146082776,0.6006756768238497,0.284973200054798,
 0.3139979259246558,0.8296759159171007,0.9291854288540746,0.596210457499344,
 0.4970460303505799,0.8117287679327438,0.6000738423461955,0.4772625437203437,
 0.04429252780680237,0.08411193293781594,0.5539516616564018,
 0.6465448270404799,0.7743277946533471,0.8121382075671356,0.608954899395171,
 0.8313500479953471,0.23231752047335508,0.07157289438698089,
 0.9122046702802918,0.24678476629638946,0.6648471738988081,0.8523501690977107,
 0.8885548052260243,0.5330289269590731,0.5265405392537036,0.9170498339639355,
 0.5859413515850217,0.5512247412229893,0.9215185977174283,0.16470521349130007,
 0.5265854644788254,0.3796630480334665,0.3774389523446069,0.9794247983353431,
 0.9562658190515272,0.30036092817312765,0.02090702624525942,
 0.41832798639116453,0.8961895579751713,0.7093941045578314,0.5223390065753565,
 0.4188203402574707,0.579680678331853,0.48624766747385617,0.1558765649085394,
 0.09747459823304938,0.876329970240191,0.27474393910231987,
 0.005560043983252938,0.3233393758786861,0.5460913170370063,
 0.2381423836944806,0.8783183541836337,0.3157660236960823,0.853878909226126,
 0.6220845162310658,0.2785330067992725,0.2973633131409462,0.6429276708612386,
 0.25730590523454244,0.7026473183657489,0.5210140253470726,0.597571742951619,
 0.019021401512888114,0.43887311462081624,0.6120972534846773,
 0.7977441868728518,0.2567174776917943,0.7719379770951065,0.7828130616340434,
 0.30523283578890104,0.45726814962453466,0.4761146366484412,
 0.4740665602239428,0.9618300495864309,0.12371991532086168,0.8336782870497111,
 0.7733245852675911,0.2109082432513414,0.8302239368003503,0.736295237390531,
 0.9790607508681548,0.054409044961798325,0.7829045298442603,
 0.22989453100882296,0.6165279165037012,0.8361744936219191,
 0.27216187359078825,0.5501125303670144,0.1609791846099048,0.9017850837307892,
 0.8891291290300893,0.4363355837989199,0.990379291470109,0.6895351621320567,
 0.3266502130658453,0.46038931201605815,0.23361251066070043,
 0.6876006783599262,0.21215164095116001,0.18359816134308948,
 0.6842619873277648,0.4178272083355705,0.3978590382811289,0.42056151889589843,
 0.7100291012209345,0.4588166283864912,0.4900970124950922,0.3446413107655407,
 0.10526589323074598,0.6215236157977673,0.09759392922643462,
 0.19769422219492727,0.2956519042727026,0.8165954439890373,0.9709408737386707,
 0.023114776419765337,0.23739358236155095,0.6031831043940572,
 0.11341420376276379,0.6312182288728266,0.5523971079088149,0.5718841243540963,
 0.6568699808729563,0.9056177208312406,0.1421937296729805,0.6392374464276529,
 0.042170546529862296,0.36911385838795385,0.8740912111024041,
 0.4656532884250718,0.16975322229140066,0.8154716242157721,0.4607783625775054,
 0.6701981222728168,0.8927215709824113,0.4052668011067224,0.7959389929823855,
 0.7201455492848013,0.1365880085722847,0.7068121699711738,0.6031662841033064,
 0.1327799870647386,0.6854143385502081,0.6228321052817674,0.1883903465911645,
 0.8626934859192917,0.9198830178158989,0.6014434341140309,0.05516252545899669,
 0.12095251700525456,0.01581461659782557,0.8315480707407769,
 0.7776098614218191,0.6520747724800122,0.7472201839035808,0.9907667519886273,
 0.9770071087890888,0.2660348921472706,0.8464159875885655,0.668378515764798,
 0.07118375640669217,0.2803851083855372,0.28826547193069163,
 0.12798313378920156,0.8554387273946072,0.9629712807873826,
 0.05653538065733077,0.37134316384700417,0.8962711173157011,
 0.36132406692347063,0.8816214057279935,0.0010108760138386463,
 0.743749771933073,0.03385635733135772,0.7360018629407512,0.07666043773182984,
 0.06269995410186668,0.282150095526459,0.02015153876607778,0.650738323099973,
 0.5076249702030623,0.7930053365559686,0.33870900771611157,0.6873391226537559,
 0.2707752989370463,0.7915506479976736,0.1924212900839377,0.602765441457966,
 0.15533097895047288,0.5666087425359304,0.11170072710487311,
 0.34184157014272587,0.4044018098089045,0.9390131783942419,0.8008376845457411,
 0.372496464639007,0.06030599843983053,0.6955346161224493,0.5719166102741718,
 0.4627432563549394,0.8433900736105423,0.37035150301549,0.3925258206325014,
 0.37958053706996486,0.7372271366720967,0.5021970429576653,0.8381662492158685,
 0.43843028981383814,0.8060467577756817,0.917852368370548,0.33411779562012145,
 0.47894588568261964,0.6413631408342316,0.8493267261935115,0.5440156930707568,
 0.9954896288391901,0.2027418397131403,0.23905019480160417,0.8887480567365901,
 0.971664048624828,0.63913724844101,0.4847475229187209,0.2707978428696304,
 0.20629156492301637,0.31482859554744935,0.9868446120364043,
 0.6340760029827855,0.03838820660748632,0.3102873215163817,
 0.11736341190042188,0.8271479804737067,0.35507085432017194,0.435180352872629,
 0.8144271268430898,0.05531714695624179,0.536217717177588,0.5462706362816933,
 0.7156006151540757,0.05570434093839072,0.7286088760060225,
 0.46824131565841987,0.11780145527289632,0.023387138784811357,
 0.8796748671305615,0.5103924609554608,0.7083419159250406,0.8319107718237203,
 0.010310588732783232,0.3577111898590096,0.5836148656549265,
 0.3462546988372589,0.9553223722544405,0.18104626561232862,0.5236832717427152,
 0.12622970490338725,0.2161185370056028,0.2309189670173658,0.6271565965346555,
 0.24531374805150896,0.7415619993356986,0.5638679417400246,
 0.49118040477749025,0.03482605239556824,0.3729751766877418,
 0.22216806731278738,0.9575553217268461,0.9999267790532897,
 0.009900575622425078,0.3637207615683442,0.6356976868295326,
 0.7289206044129942,0.46248995246164304,0.9560267064812448,0.7752436386797383,
 0.0014855814946578283,0.19379357617123116,0.9147796714648748,
 0.6263513340131018,0.8519255611029241,0.7728537508361634,0.5063104642695789,
 0.6424093632724803,0.961115189245664,0.6468239776119278,0.9408637050957088,
 0.5593152350931258,0.9195471822960558,0.7888776559870261,0.27542292989921513,
 0.6875211892734179,0.6510807407679878,0.1667665927344404,0.3775134942241203,
 0.12011873385067673,0.13743311644816503,0.5875083144224034,
 0.16206565555084596,0.38985505947338117,0.36174108428457585,
 0.1634493715021802,0.7021190112468888,0.01864292929744127,
 0.35091502337805647,0.5720661048420741,0.4459895113482397,
 0.33974484898169954,0.11001553024674238,0.1456790922608423,
 0.3086988117300661,0.11898595316026539,0.8169532585321148,0.7088214812308347,
 0.5794142897335552,0.35976649646012293,0.6345769502096529,0.8078455865430677,
 0.6006642473676691,0.7151672556867792,0.07432260114349298,0.349834621211234,
 0.574473447015002,0.5824605315603543,0.9570762223921504,0.426916138750539,
 0.37769679427188396,0.9430030577093752,0.5578556667344987,0.5600783661053227,
 0.6642395220753121,0.5153024864895809,0.8208230482014175,0.8539287882356088,
 0.0969116597299029,0.0030190523834434657,0.5279368107482956,
 0.6392777788187368,0.9588168350156281,0.050649116079838685,
 0.10620503251802849,0.7257407791327497,0.19009740841610878,
 0.07441416962615366,0.6288594557410203,0.5878008164570617,0.7809613189425039,
 0.4062862174696572,0.8699205287411805,0.06949347643967418,0.397541630372625,
 0.33026033684434686,0.8608107055869443,0.030107960184762472,
 0.9963173069779463,0.029294320919588945,0.09336525333469381,
 0.20109998152915554,0.6639107341427548,0.6265550874518645,0.6807591179557144,
 0.2245683569508682,0.5418385844298006,0.6242605828557855,0.3162726641897313,
 0.019755868860824477,0.9616140077794536,0.19414350834900473,
 0.8032820771325392,0.2144424800042818,0.6453014584425819,0.6059416305274463,
 0.569021254099104,0.6785851239389116,0.921538942603624,0.01905380217788133,
 0.06619244042204853,0.8674854108384732,0.5530233846805115,0.9827977284868523,
 0.227248437217085,0.37742485330814,0.27278519554874336,0.6718913845632412,
 0.8215142608680518,0.3019229286044143,0.7132518171634122,0.29756570424130946,
 0.25280840494221013,0.004043009564835054,0.642785850341546,
 0.8903158142714448,0.505767431857762,0.43464329575421,0.10035214858894825,
 0.3193952127768045,0.8168947133368045,0.9762829498512745,0.10550499777976352,
 0.11374221740462198,0.6975204208616372,0.9998434777177697,
 0.24709700673496826,0.7300802185603253,0.6812130673121874,0.4358585659461891,
 0.6991078208546938,0.32322428922262825,0.11754269349595337,
 0.9023020076909816,0.633010316252098,0.8701230277481786,0.31266304456311045,
 0.725071622324954,0.22369940875931826,0.8886108029023465,0.09592822104953747,
 0.16803175820026406,0.04608852082516368,0.9885423172456017,
 0.8660932397389474,0.2357363803300262,0.06145027191806407,0.1863886653017781,
 0.8735991255281866,0.9111495322969511,0.870975538441023,0.16985430345496622,
 0.10527452990604691,0.14848207672119818,0.5496190129749482,
 0.7120099940270826,0.3615583528883938,0.36773020764190956,
 0.12403274407327436,0.8306497625473587,0.6209847709544619,
 0.05881561562660642,0.42647471883972654,0.41325228427835303,
 0.36008258479880906,0.8765069071711152,0.6152685558625801,0.1234661871397239,
 0.435635642860732,0.660432718685485,0.9802474123210393,0.4071375345936157,
 0.4600180909115035,0.4769351550429878,0.41448471132234765,
 0.12600357417692043,0.6724177542539843,0.07655150527055454,0.830760123263313,
 0.4065899629605373,0.3815736377893991,0.8259828072741452,0.6846630429333073,
 0.20560102097888588,0.9844830852881465,0.9621091615670172,0.5653462561689733,
 0.2543305951709842,0.7559285132337359,0.3310881371213974,0.379193563936999,
 0.25348551865961,0.3912198428992062,0.8928520185564248,0.67378072806235,
 0.9058271428069808,0.1781980317972549,0.41693718596460305,0.573841411443472,
 0.3013896345814745,0.7394559068358939,0.19443499549085175,0.7915724655503623,
 0.17385257919385222,0.0678608569995931,0.5845151967154079,0.0355378656076234,
 0.4658371552372522,0.1903630427595544,0.45552477507192035,
 0.12471340396477537,0.8509389905025728,0.8338113580516404,0.6213590438729562,
 0.1473181541855686,0.5991581349213967,0.829109198730477,0.06096106092338682,
 0.4973861558416487,0.7806165013418452,0.3455465924899974,0.11234277849009888,
 0.36813053000098583,0.9103534057983029,0.8163997361607525,0.599638224322972,
 0.14853643699866037,0.6845607961491657,0.9295822938850005,0.7397253553808069,
 0.9185621554880326,0.359423685351382,0.8272066931635814,0.44956873130337405,
 0.7055872799356384,0.5200048150808463,0.39246289667302103,0.8715316706002093,
 0.9999914709163615,0.616602213915708,0.5485400452864801,0.23363646275666194,
 0.4363938614804722,0.1282361602457055,0.7100152112223521,0.21488700649049575,
 0.6473527334000055,0.6874155924950646,0.9429023025440059,0.6859097172663506,
 0.48868552055388526,0.38569748135895854,0.5276094439049079,
 0.4776158159775028,0.4034280571119827,0.2865192666323949,0.5390897374485931,
 0.19756673634302957,0.4407745417982779,0.7617104276517379,
 0.0017345270964171933,0.8829511847463298,0.7364799772666988,
 0.14533478176366843,0.7552556654287657,0.04362168109140385,
 0.4544383668310723,0.1522109225293914,0.08567606300707431,0.4389197574639603,
 0.5918584844633403,0.896342624498831,0.8271574033693522,0.6976200509001351,
 0.7063099330435698,0.7573567873854019,0.011116184801940876,
 0.9299067415296967,0.04657124110043365,0.6390918263794957,0.8379064081142729,
 0.06470634855703039,0.07513707878765818,0.1373116679705647,
 0.4192759224410748,0.9342861821186708,0.6938441847241354,0.7672961685441998,
 0.539535917956317,0.8730117170917573,0.17570176161890694,0.3973490390407951,
 0.22143943438317593,0.40611830262251514,0.22321912237853903,
 0.8322571944057945,0.562604395477672,0.8087017116336138,0.3937044733686432,
 0.8099505260529319,0.6450920200863792,0.09507404039716594,0.6952423585917518,
 0.9497521430200555,0.35088843022964544,0.1555836837418484,0.9545345214418044,
 0.58444630364699,0.8128365290347751,0.10695055727968206,0.16461439977504289,
 0.7502221837345358,0.14031240278420465,0.9702611245856297,
 0.08473926894375217,0.2412061945706887,0.1422976910553505,0.5232160142722071,
 0.6569423815300448,0.6051283740508199,0.06379099846951253,
 0.49639927759267277,0.16274434488156342,0.7150510561704597,
 0.6639679129968747,0.8988823223188059,0.532748401084491,0.07507264076234421,
 0.35963108405293287,0.9595789281085048,0.9531440886149296,0.9229958612159869,
 0.9722868582096632,0.8225149587732202,0.3218584324181605,0.8915273109106876,
 0.0843347175305571,0.4314880868128056,0.10802375158030408,0.8799400549585751,
 0.46490773121600204,0.1597678230425429,0.7813943092925887,0.5889889264381816,
 0.47742117026550634,0.946686358332844,0.19719033299159727,0.5270397220147618,
 0.11105074732759979,0.2884545482717408,0.3158346842460703,0.092072196623876,
 0.22696290182412093,0.9641178961652991,0.2345187110200706,0.8517922247316124,
 0.5913665393305787,0.1261988022949062,0.9307310565408136,0.7083459517206138,
 0.8767519783123983,0.6533690841113127,0.7691782880688726,0.5363561772030836,
 0.8333855048797457,0.395304154882681,0.08666941463804156,0.8749381673915722,
 0.5143358564551589,0.13142591075417476,0.8137136404526756,0.3817059774298516,
 0.48426326040686885,0.03247966350308995,0.6030614748380301,
 0.6809270883285535,0.46759554565325623,0.43389502675634906,
 0.2567259659739236,0.08980933345988373,0.0829000743645123,0.5654623099120388,
 0.025345738467427825,0.0034172048879378414,0.22158396563567606,
 0.2014012179675363,0.05987526673451682,0.6600045769570639,0.5637126617758462,
 0.24992883092378848,0.9061162896890208,0.48910146839822954,
 0.06408225844999837,0.5699916248830412,0.5232971709049157,0.9221335040558883,
 0.9446602732230956,0.4780056734985906,0.5478014380074533,0.9133553846061806,
 0.6517069492750073,0.15193074119604588,0.6579701141566906,
 0.059704403845153076,0.652119228641299,0.43444044596106757,
 0.4398323254450063,0.4426896888551507,0.6547760670087951,0.99353228826042,
 0.7204138257484805,0.750966544612817,0.8227354987936809,0.2204272230198825,
 0.8400786374662887,0.6414226093032568,0.5115513162515141,0.2799634584832402,
 0.11290147566206454,0.9424129368936524,0.013034152430447699,
 0.3237176213440459,0.5399728508439259,0.6398611490217803,0.9420001765117616,
 0.5529864430147586,0.21809178563146003,0.09090825731816776,
 0.4196772291904024,0.8449303280284767,0.813696575675068,0.23226360184809858,
 0.45345773266677936,0.20508133038056497,0.32168662439087126,
 0.8955897269185984,0.8249015467138551,0.5979456968304026,0.8215036690717262,
 0.5301148306310212,0.6311853578547866,0.37774444420072284,
 0.28637465196021394,0.5569554223646561,0.010372150697970373,
 0.012927311285291565,0.6384194073876193,0.6065479385824464,
 0.3441881713775603,0.4636812942422507,0.43005298229962885,
 0.10453430504326788,0.14522062391966717,0.5515260147004395,
 0.6747682076590404,0.3610162430212789,0.8728755366680946,0.5328194811544443,
 0.8233161853155648,0.42030113291770566,0.49204513357283564,
 0.9361992002162564,0.14571698794513632,0.5760199738304344,0.6824253581155398,
 0.6260194369245908,0.39534560998843793,0.10151514868596379,
 0.2857908521336301,0.2676194937001134,0.1500876385268758,0.1204915484582244,
 0.02571623758118624,0.875684162196925,0.7372872463311795,0.6738177737154805,
 0.1692057737645033,0.9988371353562178,0.8491585020358707,0.7319092691291392,
 0.31036445867585527,0.17848407048918524,0.8360212194790614,
 0.18276173656452666,0.7488837171655609,0.9735757031366379,0.3708193948478006,
 0.9403910164573708,0.3105085080208366,0.5242887854465194,0.8569472278590659,
 0.8385959807450304,0.5350807329956762,0.5543843495522195,0.02431276972336127,
 0.5387106356286894,0.9787989900358728,0.176593867826781,0.7146184554791987,
 0.7492119448606244,0.7140996817701519,0.8668752853936923,0.9334802332708136,
 0.8009474743454968,0.055391366750877635,0.7699498477977282,
 0.2299584488959041,0.7578001846967131,0.3093868939475084,0.7639498494686338,
 0.7982011857012791,0.40411363876893946,0.9126943729414424,0.5817369702515865,
 0.41116029797990705,0.6589467993426011,0.629258465733979,0.2145915818673252,
 0.7070074525511375,0.31658899036599053,0.7466515651725262,0.4386853426714905,
 0.3650844915587085,0.32974939988524465,6.489272121479317e-5,
 0.908921626153672,0.8140111506354856,0.27711812234992383,0.6628142335580769,
 0.9708290818943535,0.35706027872244794,0.8008028365295738,
 0.32791978571666924,0.13055734803104213,0.38124179457739693,
 0.0993319089496818,0.31026932244065764,0.4744077882031511,
 0.46017121023503793,0.03305129775791493,0.46376980697439596,
 0.1524300809795468,0.5779368595610601,0.6618491979025272,0.06848265533548359,
 0.16949834493994642,0.8807461177532057,0.42088328940631614,0.024948260238707,
 0.2497592578131017,0.7498289157568004,0.1711256313301438,0.932844062591166,
 0.4828970226944722,0.16852057936543008,0.7939038732519879,0.3340610877471547,
 0.3925569434083861,0.9483535266420569,0.7336780539514787,0.1143367588414983,
 0.7720114963941076,0.4678099209532859,0.34196408550214097,0.2727730676814192,
 0.9679492409323581,0.4439506629659501,0.16632981216407539,0.9017940195271076,
 0.7758976983205594,0.923682495365896,0.35399364595064287,0.5794447107758697,
 0.6349488959241092,0.5356316289886273,0.6687641554104053,0.8067574728425364,
 0.04172610066002702,0.4257330864671677,0.05741941470888379,0.353075611877685,
 0.9513515532585755,0.9596233703167376,0.8002231179781577,0.28697551017042966,
 0.26634146581418816,0.5456862441055398,0.02189721706100267,0.934075302256612,
 0.44630914391785015,0.928172003212855,0.7238662568414077,0.1377231091204597,
 0.6453410865740803,0.6243485384095409,0.9332171908727265,0.3058139379348014,
 0.23698496967719607,0.4849797074710894,0.7388400864349514,
 0.32358670147081847,0.30699029044675574,0.9039129217425782,
 0.26716635035696046,0.07138726334610979,0.03710722310687398,
 0.4680875364207412,0.8969271048206893,0.9930810895604236,0.930122329705925,
 0.05231524670451693,0.6399453746714505,0.5213736753838218,0.7351491562656558,
 0.6854690540198894,0.1325151115042541,0.8948402162500555,0.10946899444381519,
 0.6936682925820141,0.8538029401402281,0.12400505150109109,0.6385947211337887,
 0.695755873540459,0.6063036222428086,0.6763555063801633,0.49967765316009016,
 0.9646448739298847,0.32367052272878505,0.406752118217539,0.8339821011664802,
 0.5772035656970589,0.6026012535140045,0.6044346795060754,0.5285482191145385,
 0.9155708705332174,0.1516199384649698,0.677444962003704,0.815576653264244,
 0.4589196817028517,0.6679927111077223,0.639369258705031,0.044657951400384976,
 0.8391529767769575,0.40319364954045867,0.3609533975045949,0.9475586791472546,
 0.8303423123419148,0.9826280348814345,0.7113601918167335,0.9254897076417219,
 0.05123201246477649,0.16504084272082953,0.7613794874222176,
 0.6741451333315706,0.4413556124855178,0.19742466792645985,0.4586481378399996,
 0.21567112288039414,0.3629736515229939,0.8126540363985012,0.3260333336691632,
 0.6907524658711923,0.5530333097927042,0.584115332181052,0.13620153881678354,
 0.4924746886465555,0.7177839457463314,0.25185849551550876,
 0.06136063055091501,0.5075671662835544,0.9473447987758391,0.4332402162601976,
 0.05854102637611358,0.4667965771280578,0.6249851362294042,0.2658683019853899,
 0.31263394401332434,0.9413539208092591,0.7055779052150748,0.7147519007220282,
 0.6451198573232324,0.1802089076473451,0.25651663132811975,
 0.007451918303694516,0.4117533599823572,0.09473903189479549,
 0.7750859572978901,0.8682721561999164,0.012361316220349194,
 0.7667733700063627,0.6428438130104343,0.12976928958460154,0.9792210526209237,
 0.7418444401757256,0.06255040770776954,0.08271821555465819,
 0.42685182535060995,0.6153586730201126,0.5212857598433367,0.7319764088927011,
 0.13841793263787971,0.31662718800191614,0.49778604971988016,
 0.6244337690925129,0.9384961316948868,0.6903802257034161,0.4060839891950738,
 0.09813054488869555,0.743944458630315,0.03772803570905747,
 0.32683064788854344,0.3414766446078399,0.5056481908774586,0.9817536404936569,
 0.21753944076426213,0.20172955311816643,0.8405904762923317,
 0.21332080964424693,0.896445768220271,0.9704855558578909,0.11225615348383333,
 0.40001805529379386,0.23141725975398253,0.04272330542739988,
 0.608341866290814,0.6242896688013483,0.21461381875362617,0.4132606551109579,
 0.4514449219370298,0.3587925269695378,0.2214680814282044,0.45069392835307664,
 0.06153726264052062,0.9237998549453463,0.04891516857000899,
 0.9308749801345547,0.4638225593051756,0.4955076440567383,0.3207367040759246,
 0.9191757291477847,0.003512400004085592,0.9873336178076921,
 0.6652418569634028,0.08060785796363468,0.9367601483797827,0.6648352080793467,
 0.5885443266946224,0.42065113829762446,0.9866571932371964,0.7939233050366571,
 0.9711086402901975,0.008153364497015647,0.47886113689320675,
 0.2969726349632351,0.39860024664246396,0.03913089044675511,
 0.07212000383882033,0.530799206351361,0.5037233322349485,0.3712179057636912,
 0.6708946538929035,0.4728018710037789,0.7440912503646617,0.3903994464327023,
 0.17064878428613406,0.9461211127537078,0.4250696006876912,0.9341096061642568,
 0.8882202952548431,0.6310242252273726,0.31846294731503555,
 0.04948812188138518,0.6285112526874592,0.5131981650990868,0.5613779226050652,
 0.33575521919264684,0.3611455723625314,0.9409707939982332,
 0.35340301904741955,0.691293747413201,0.7429944843310464,0.9363917124541046,
 0.4584066720421204,0.3012332214896022,0.3417669086514188,0.08182213543998151,
 0.43870312842265946,0.09793924768369877,0.9389954947800119,
 0.7802359992367979,0.6669854846940017,0.8067281463313299,0.37571091574956006,
 0.4501985932571456,0.17860204460847595,0.2065306041744489,0.7982019590246076,
 0.016337649499984397,0.09236890990225644,0.6860474879700601,
 0.5687598180309392,0.030054858068686796,0.6426396184478752,
 0.8428783005882738,0.20447449913529103,0.4483577638135485,0.8001211243416154,
 0.6821829676570669,0.7755576865492697,0.637861031617981,0.7452091779521315,
 0.6710709584044975,0.4587741161021852,0.09644307537179952,0.2331391481137879,
 0.29346147677084156,0.053949679864687905,0.2735461064346547,
 0.02405263546727865,0.32806307851687067,0.5580659738254956,
 0.10595322136605623,0.10704728732157975,0.840505555168702,0.4483419243752661,
 0.9597743527389198,0.7553485345200043,0.1605402126894404,0.4561091960496355,
 0.7980529466968285,0.9276135873926104,0.5390774077938328,0.27874624071862064,
 0.31590482726011127,0.8148838701647212,0.7875568028486086,0.4635758283050081,
 0.6020194740696879,0.4913307503147957,0.3233837631850828,0.5104504716293436,
 0.8930632357055565,0.8324270576087174,0.1405832932657829,0.8642522280949263,
 0.8441549194933708,0.9543001717013977,0.049379881428323724,
 0.011833979420902807,0.2620843721845746,0.8817814038657323,
 0.8025292988802567,0.6559366913956448,0.1945350966381474,0.5575375142835926,
 0.7392471424463016,0.9594401067411985,0.0448413192782392,0.9369908642012895,
 0.07853840640894849,0.8388335906699231,0.6297410121505278,
 0.21235336042503006,0.888005153728862,0.2124495626333135,0.1541762844954775,
 0.297671212303928,0.2551120549705179,0.8500251099147782,0.36485320107812347,
 0.352900459209344,0.7236842368452413,0.05474142449873998,0.592842185580479,
 0.3890514103254975,0.8098431023454707,0.686758189058527,0.9744538027284437,
 0.8769999334612821,0.37961816677959737,0.7948505306735606,0.7531319203278855,
 0.1634281754091469,0.17220265485449615,0.808024607171858,0.5558593966770896,
 0.20767444077371366,0.7505968182789522,0.8750268835764848,
 0.22175450619009673,0.3586475934921516,0.49443927022776646,0.834903836413102,
 0.6717314847293354,0.8035303190252904,0.154636131177306,0.9772651469755319,
 0.841263867935405,0.32148420366531383,0.4911152888852506,0.9591204483219457,
 0.1525213407229894,0.8133735917250657,0.376242076189518,0.5396920092353666,
 0.686746606959976,0.3236800977283866,0.9841654351405527,0.11025233820590485,
 0.4442950068343924,0.7632415410507875,0.7843539769131787,0.4118510027285399,
 0.8559447005367194,0.21414475734272675,0.2489781395463152,0.4005174249132638,
 0.7043143947710677,0.39211632452730405,0.6368553819694163,0.2752694251273431,
 0.7522281927402875,0.6566137344871559,0.058591168837974106,
 0.3902782369467781,0.6290702727555355,0.6437363531907165,0.3991328488716148,
 0.07034124330415681,0.034577964826063234,0.4696827084648898,
 0.6022426752628903,0.35804882880489974,0.8701578187937095,0.8667765143047708,
 0.22660513334712806,0.3124667560130332,0.6509751328451912,0.226366123011981,
 0.719967266727722,0.45564348551842815,0.32589890060303595,
 0.07783234854492505,0.5019546668060733,0.3087473258569009,
 0.14004206328814928,0.7558129968817437,0.1186266621558193,0.9383877112118594,
 0.7852130685845252,0.6985372633957714,0.7833645160852779,0.3845513042496982,
 0.8856569062900859,0.5986896141779466,0.7396248672597974,0.12717590052664585,
 0.8190421641745225,0.5967271442515286,0.15782656437888753,0.4736102790376647,
 0.4934951342029774,0.29804454561291005,0.9984722580835066,0.5551366068704662,
 0.5629617819482108,0.0018343663491303674,0.4473748228318978,
 0.9891304402330457,0.3638740318515,0.3839298099782137,0.975208433432898,
 0.059339529307505745,0.24865614389038115,0.08499297129538685,
 0.852176867780625,0.6716306818641878,0.6442228055793386,0.6376008540045858,
 0.23755683154526797,0.6737979331448267,0.731825635511319,0.5510132312245234,
 0.37978767380356304,0.7288740913963998,0.4449543969558958,0.5313429987067568,
 0.8646337014279787,0.769881643733172,0.10835985326553876,0.9121373316352392,
 0.600985467623523,0.28867260313275545,0.1861069057588578,0.6271724810637427,
 0.5595921546667177,0.07691537149693983,0.4734577225866774,
 0.46129640457367305,0.46729426954300723,0.05498316988061691,
 0.17286046903869778,0.9896740696739268,0.31478048227916666,
 0.3225600108503368,0.08984567982349223,0.6079248372861793,0.7086691737136711,
 0.6947747823362703,0.9344951060215222,0.5407638391718042,0.02215165927968199,
 0.46120147792675814,0.5563505256248872,0.008817389635836559,
 0.8839491853897907,0.9316615402158811,0.7556205144809014,0.15693558585460277,
 0.31250570237714037,0.7745234658484892,0.661889142483655,0.7450070494446972,
 0.43464120689783703,0.7719659496223862,0.06451646342918449,
 0.7371810530028484,0.0019516800150616298,0.7902677674482657,
 0.20079682846847036,0.2973463592583534,0.15240101556396246,
 0.5734315497324713,0.25914206207579815,0.7410131933051756,
 0.12025380388050955,0.021868147902187496,0.44091679011527907,
 0.8900930241076291,0.32352708705727395,0.2711141648254064,
 0.31419183262239914,0.44042509763520377,0.30943907096905465,
 0.12188517000487831,0.7307239492184503,0.792571986991684,0.07053317410518445,
 0.010079807908207727,0.0627345104483713,0.7981407124977302,
 0.3616839830199706,0.7587126254784279,0.5271172681206268,0.2141127548659587,
 0.5925904406245915,0.6454909485643646,0.5789209506892661,0.35854032394241087,
 0.933311096151718,0.5374368793477267,0.49215374457584715,0.52013995903131,
 0.7715151152529984,0.07938859126342157,0.6239824296967081,0.5734249085909078,
 0.2888657636175572,0.501426329948573,0.10621851412541639,0.07526667043464463,
 0.17505337631016027,0.4107139888179471,0.3057586280730986,0.9308888021219162,
 0.08266169597854645,0.28822272881094146,0.30868528669769013,
 0.8888739543217862,0.019842434236447715,0.3951349827162435,
 0.8160142886656866,0.21261173505338804,0.20886687363067613,
 0.5480314889437382,0.8474622478818005,0.9823384696459688,0.40668573665524743,
 0.4036444969765388,0.5152048452932729,0.08189805304123565,0.9024922050410208,
 0.2396490466610801,0.5689915828066646,0.17158424175929898,0.5351004249768501,
 0.20976745599234692,0.3073146186137993,0.9962546089075326,0.6348732193140076,
 0.34258432098428093,0.7236428314734981,0.29196272385814026,
 0.5105162908712135,0.7973198944478179,0.432685845938935,0.5467402756438692,
 0.9741660598490036,0.6876828628251654,0.3509821396195112,0.2689327081607905,
 0.9826131184018594,0.49714665016202986,0.24193548056855885,
 0.4026860532477009,0.45820884921855864,0.44140248129842663,
 0.2656075605176125,0.5643246749617591,0.660412735737925,0.2358553504439247,
 0.4932906510044527,0.15252870028692223,0.805767396590005,0.6987876524531264,
 0.3274963943056972,0.7222267787130785,0.918929486574192,0.35944493677356537,
 0.49894694279991814,0.3186741079556046,0.15353974216110944,
 0.4066025542147971,0.4948471623557452,0.5097845822539483,0.2628601334737033,
 0.8606562856024186,0.9138534555115966,0.027073179611778575,
 0.4142746975890872,0.7875546165900262,0.7737552646633861,0.07112251216998366,
 0.16700566013618023,0.7044280786136157,0.24975877063967988,
 0.5957509507710584,0.3755337702867606,0.29180799604615415,
 0.47363859171859835,0.9833249812433248,0.291924352102081,0.49994744227614785,
 0.8331761729016065,0.13531468789391587,0.3809331494049232,0.4421975894961575,
 0.5240487837619472,0.65672623025481,0.287209156844382,0.3083635861278864,
 0.7532091208362007,0.4124587154973367,0.8229307552194408,0.43705377705078285,
 0.28751062849440623,0.9019261455054869,0.8598882712538382,0.4700905244552458,
 0.7792226006348413,0.17609114552767513,0.8987039526113412,
 0.40599229526863523,0.32656467420144275,0.1989462835610576,
 0.8872712641213161,0.3169498168334206,0.41095910050721074,0.5125610488703033,
 0.6620645064685143,0.991391937791259,0.649258124727228,0.6332856658199468,
 0.532218686560281,0.32959563168431805,0.49747256007425533,0.3303915958858089,
 0.013615524679904012,0.15032933223017253,0.7194799262931675,
 0.7520951357262072,0.7306121830544842,0.5317556398760243,0.2683615577162064,
 0.29799684550889893,0.631925741481291,0.800392644860142,0.6370448794514899,
 0.7929157606379331,0.6991571285819123,0.694100577252261,0.17015276162185933,
 0.9585948732550704,0.17369794680268202,0.49629719744296374,
 0.8468718147951491,0.9540255225561145,0.09615860413602917,0.5762503928217279,
 0.3393737301715821,0.19271617179210832,0.9840419127187128,0.8500046719887088,
 0.5847468871307077,0.9252456785967453,0.41286263012178603,0.5478889962294611,
 0.7207677379968052,0.08139086276021823,0.3634634678790152,
 0.27597309487561283,0.9642595798015702,0.7318363459390765,
 0.042290581128826554,0.28780369432010255,0.2946497722864163,
 0.3407556627838926,0.31763369346964354,0.11829395551259703,
 0.7858968349818336,0.9728062072029738,0.3898802663688481,0.2871764687615238,
 0.459058046998045,0.6746285953861904,0.8196683215933722,0.3480690913124689,
 0.7703223005418294,0.9570657945054997,0.05568117804507322,0.8966121063078268,
 0.35591830891173826,0.6881375855413304,0.29736313954990956,
 0.3893884803360018,0.07767287028697434,0.3492371750777228,
 0.26205188498219134,0.5459524171248079,0.6832023845401151,
 0.05010988146502626,0.5670464570081284,0.24385760033817094,
 0.11147874700333493,0.09267312914872083,0.44337992340400545,
 0.25961662208864755,0.5696391335186292,0.18932325574254794,
 0.9032494668982619,0.2872170484167882,0.10749018800283194,0.5231519158744509,
 0.18184325912079558,0.5979276585457758,0.5904747244489832,0.6615519723182016,
 0.4043360092454583,0.7926618443083866,0.12499966577026211,0.2809817169512392,
 0.9898881667325978,0.9665313307763443,0.9333791057765498,0.6141685946318759,
 0.4225601343774549,0.38111842808845997,0.9148872307166525,0.6987841774563944,
 0.6464768825413935,0.5164980867391095,0.4197260185661149,0.0915265877994399,
 0.14259967968748466,0.26284950479978797,0.2946499493547702,
 0.35415108618924107,0.6728391766148648,0.03457907516755354,
 0.5762886088488431,0.5897450736425484,0.6759339241193171,0.276895803104218,
 0.16769455420800372,0.16628630951415813,0.6633392755411819,
 0.7846183316288929,0.8312128085577603,0.03725108638550967,0.8945759194632045,
 0.6142807994048165,0.7082888231121668,0.5705366484233334,0.6059380360130674,
 0.21952140341795656,0.11401924934077212,0.13092721261838547,
 0.8353626259180066,0.8822368853007586,0.7385327197728417,0.33342705884477164,
 0.2655176847244026,0.7396715093637558,0.5651174917977415,0.07580120472617902,
 0.8973871503773716,0.9647553756319617,0.9329687088407898,0.20946957983041192,
 0.9123188328370857,0.5361928324346473,0.41715046027001557,0.9224536867709351,
 0.7709600497290735,0.43045387895751575,0.5720645097066339,0.9673144696176094,
 0.04921490533665984,0.3357908246717758,0.20732386078194331,
 0.06996798536504478,0.653421311949337,0.28824962737876425,0.7137459484713686,
 0.23750175802594387,0.677922079593745,0.945206977585786,0.35306743082406766,
 0.8565243212675274,0.3604781082454438,0.7668826376012674,0.16375587614389642,
 0.49762168264818074,0.03249410650434903,0.4223264617176261,
 0.8059309212715554,0.46975814146396655,0.6455561398197407,0.6826619275128787,
 0.7261796720146974,0.078273466504161,0.10521260454154369,0.3931334796537147,
 0.07456537245278383,0.3748184610541344,0.21000184196961658,
 0.23889007190352562,0.48312913632433385,0.5083826014595216,0.547603857391796,
 0.48187612915326994,0.7558152772589753,0.5082651141265621,0.6242920443188258,
 0.5535531928553931,0.8907937822227989,0.1472734055328484,0.05318242895968173,
 0.19811533174653417,0.39672182288527236,0.8518621316798543,
 0.13348386609739094,0.959802110735201,0.07328111326793885,0.3763325288527599,
 0.3025337908535466,0.7891253977041023,0.3007032286905278,0.22767591580982582,
 0.5778969665674653,0.48875629325653236,0.12323200747021223,
 0.7258145996765489,0.34800488341761526,0.5270422440590455,0.5289800599427563,
 0.1574100468069366,0.38738354704509237,0.17532978600841398,
 0.6387494449378942,0.32604934771004435,0.9450704566089301,0.719783704021652,
 0.6892974541894894,0.4691854420938665,0.11794026109047706,
 0.061451500957496075,0.9148299152920345,0.20921666824941576,
 0.4242257163326897,0.813645486606661,0.7645492226316909,0.9581153657027057,
 0.28370267844731334,0.848495918828225,0.24304648479395485,0.3917990070286218,
 0.20160830255327844,0.7332343234311737,0.6720532852153517,0.6248300960907258,
 0.2986227522637058,0.8201705792375216,0.9009384278409708,0.4013738764442788,
 0.8318260996792832,0.22219379185987898,0.17483661779749948,
 0.7448591774527124,0.322477372182697,0.7106932915769733,0.6254144719409069,
 0.08824951329331299,0.8917190617304805,0.06494829145286629,
 0.9385427046828467,0.13469919716019674,0.3827399057828512,
 0.31062013331167593,0.9412535388819836,0.39733995914251397,
 0.8300130723033424,0.04370826315539833,0.8559995863296015,0.2076979521444937,
 0.7311626251181645,0.3660942152930635,0.32273833889648806,0.6796529285417917,
 0.8844679274056434,0.06878308557828294,0.5706197447882748,
 0.15366652276940163,0.5636985738659565,0.057588132171357054,
 0.7578301232238509,0.27446398461708243,0.23279153350345516,
 0.8648429252607044,0.23141584875384313,0.29909321484509577,
 0.33299229911468475,0.33657642039196844,0.984715352817802,0.4700303736964958,
 0.0715931794064425,0.9656604842177401,0.28523628494512854,0.3172443155220508,
 0.9505103297412387,0.17936288507687292,0.1264644599310124,0.8149194907041136,
 0.4593075698381949,0.40676162522673587,0.950868007099604,0.1031593736488009,
 0.2681044483768372,0.656388665406371,0.9707759538451418,0.4426835394636327,
 0.9203710332754809,0.7051646796378307,0.2602494749941999,0.2872115110210337,
 0.817612735861359,0.6334597845065585,0.7598002175290485,0.24655512735382423,
 0.043608608009094385,0.996623712034438,0.5809633335812393,0.592146072548325,
 0.11410433195901759,0.3241595289654382,0.7032098458672663,0.5025153094708987,
 0.40424353248686573,0.7896142087070342,0.12404953032046717,
 0.8195760039137068,0.4140579894558657,0.27528396324447013,0.8682236887118192,
 0.4817213964432675,0.8930132930253825,0.1849580895021944,0.4301702425485736,
 0.5503190903596661,0.773603153325442,0.8814232156367415,0.7152488280426088,
 0.06572643020542279,0.7431840977890324,0.3592447882236979,0.9538570533474564,
 0.5157667857504582,0.19392410555089057,0.04738797906410164,
 0.7012814823250573,0.7308396509344046,0.1646479284972594,0.6122493226778852,
 0.23384856790173902,0.6306787369989388,0.4203899908414701,0.2873713465390211,
 0.613260514414913,0.9798309383909221,0.9342923837231845,0.68363727845988,
 0.22982781454327517,0.9919692064472985,0.7743095063795545,0.6171230505594482,
 0.279643348541571,0.26957304963761375,0.4699679651627757,0.589639633275849,
 0.812104528035202,0.009190119060381141,0.7393128995517775,0.932292910744351,
 0.5799952648603697,0.15542301152451876,0.40781519211323813,
 0.02256501841209424,0.44227526682884444,0.797679433568707,
 0.25778243855562366,0.2702895422566012,0.1895956419915209,
 0.16819355884250786,0.10959278144449147,0.3665835420598408,
 0.6563332925064851,0.3540932010604042,0.47817763564738425,0.4747988906925913,
 0.1502023843194702,0.31299851527740297,0.7414499698118635,0.6581059270362806,
 0.8427130811044733,0.2272953094848199,0.5810156630514973,0.16234533480676117,
 0.6264325230425221,0.7850170153119854,0.6435011357741864,0.8147791476454963,
 0.13293708936456883,0.34187970911304677,0.3552616302311016,
 0.2811861306668466,0.31973420399075436,0.48898453548962084,
 0.04301033312518343,0.8107988360745316,0.7408338141707169,0.8417313903511815,
 0.8732390327894172,0.17550793767472017,0.12591297162013637,
 0.7850594643033111,0.7754086742027627,0.01888712497404399,0.3659718077999433,
 0.859988671550004,0.7483464742108987,0.5762859508857624,0.2611577562192613,
 0.13455110443841556,0.5782247372248641,0.5984714259534294,0.7355607870255644,
 0.6860388795694478,0.7110710027342346,0.5363133795896765,0.10413016308307332,
 0.17800627663427038,0.6035873175223643,0.9276067990889398,0.6496920946740307,
 0.8576004778077199,0.4754121518887239,0.7223618949935547,0.9440820922887434,
 0.8465451210177865,0.14618252468802206,0.2831571439419235,
 0.36648087012904074,0.7458776331975319,0.27958588556575337,
 0.7942800398138555,0.5099393503812255,0.8372122976610282,0.6336468453329013,
 0.9090166588921197,0.8799126811084793,0.23107296019475088,0.9076691285366094,
 0.11117564737805163,0.496749450162013,0.6948420094159052,0.3674460083736464,
 0.09742873958660536,0.9706836209429689,0.9696382397576879,
 0.31071958604644456,0.857564836830949,0.8995213385635419,0.8558998408903703,
 0.4953596317822324,0.48375330902030234,0.14952181971826883,
 0.8806622323191948,0.17952724652110108,0.15839610553359262,
 0.3433550687187067,0.260287021219622,0.30073110424053173,0.2176885583955911,
 0.5922839001585427,0.7375997039413617,0.9167261719805792,0.2566269133734945,
 0.36258290485845834,0.4969401166794317,0.913958277209304,0.9109588398841972,
 0.803777254758609,0.8154640725139195,0.4020018598023106,0.4144652631240675,
 0.9772423612685269,0.9428265186825263,0.5042852381411762,0.8365434311884089,
 0.8051031252592701,0.2933633388941743,0.2625860376155805,0.9713658038454454,
 0.20366899962794005,0.03991480417233184,0.10904789645732516,
 0.4041404829418439,0.6746816535638385,0.6432460127982665,0.524103383976074,
 0.5542582101948541,0.04251385537424024,0.6666415352200932,0.6422187373342289,
 0.793613467689078,0.15229870103795906,0.7667733587857798,0.9901192899823801,
 0.5017743250698948,0.73204322512766,0.18522678694054806,0.09345779549265076,
 0.9055408889289799,0.23437362914255822,0.9696134137531588,0.3289125030772262,
 0.5112494592000499,0.7470957244407971,0.6084278634435023,0.8327932144640336,
 0.6841857730253516,0.5262138174839062,0.184618930189836,0.4103921903185245,
 0.7890964996684463,0.3110705837456911,0.934125907634382,0.6761731089360524,
 0.0605695892234166,0.947462867235607,0.2571779015943244,0.7241489282520898,
 0.2175000447243557,0.1994100909043801,0.4539454300481254,0.01801678486148217,
 0.06229664237152577,0.8137640482628794,0.32619313712137854,
 0.8679250199961098,0.9016371175871759,0.11547818352508332,0.8264346592409069,
 0.5988052683672337,0.7951189159364245,0.964301317350843,0.9533212527271668,
 0.37687973356768434,0.9227042901531176,0.4480921624380657,0.822597549502941,
 0.7403026772094209,0.47102408539601726,0.18285798045697854,
 0.8608112988800106,0.8204005901502588,0.01996595530401002,0.7049252194299506,
 0.11223564344585435,0.9791272382575293,0.40207486853785257,
 0.49095007033821114,0.9607065074233874,0.6116106540718432,0.9961777108565165,
 0.8464282399739088,0.47616126440661444,0.7907171745210578,0.8979631787028212,
 0.5630556769821669,0.4242719606671713,0.11155750303075607,
 0.18652761467992418,0.5239206386501095,0.06603653089667572,
 0.2458367652147464,0.20236959076009442,0.8856730116033918,0.4265924161406546,
 0.5716106786925992,0.34921615146177165,0.4550863360107056,0.4346168422419162,
 0.4577717428753024,0.5583815626254389,0.18678313091857668,0.9027662920766911,
 0.20874964901400705,0.06438864832751456,0.9718483805129123,
 0.4933461788778303,0.9423335114565837,0.9223894542249023,0.08867337366814665,
 0.8981785030850803,0.20997951123883263,0.3813162546813972,0.693109393700828,
 0.5166013738050272,0.3663845860477293,0.28705976550832324,
 0.15896335240526605,0.6000342937351113,0.06791352894527458,
 0.1661343606397222,0.15723242068253218,0.6340659583637488,0.0788781309890213,
 0.1461824841900976,0.14441557596582388,0.412509514474499,0.20033186336546982,
 0.6183475982477972,0.03706371019979726,0.5311326189152791,
 0.11118495788211435,0.6367035952395614,0.9007691535952588,0.9300908037217296,
 0.31842604735272717,0.5821220024368887,0.297590894017688,0.46995183260129714,
 0.7203099822475397,0.039812087469846325,0.57304540796414,0.7635358416385027,
 0.4697665441046718,0.1566964478421925,0.3508149307002547,0.9952719979744052,
 0.30611182935040326,0.14146765573939823,0.3861434622787472,
 0.06724122819181289,0.4784482085165759,0.37564545967749097,
 0.016881006900476336,0.8324320841089968,0.5648125613866323,
 0.5665608196036963,0.2610415536456595,0.36944447889007315,0.7022866932912416,
 0.2649743022642298,0.0920847746568656,0.8394571093128956,0.4852419112526325,
 0.05561863600835959,0.5271521215204645,0.1024774199484062,0.5161223890912936,
 0.15688577453192676,0.012451650176571571,0.12902394710590648,
 0.41937701910478375,0.6629555423327684,0.5889652840586824,0.7934853367785794,
 0.6264712034557555,0.33601958801309384,0.44082015112408,0.33884944603032463,
 0.25716704952808944,0.8904286000975223,0.17399098976154326,
 0.8953145866953593,0.8638640240600634,0.0545740181704466,0.5224984628463055,
 0.7314547902633589,0.48434193095690037,0.03602913034276356,
 0.10899018652373116,0.5300318779746325,0.3103603470378271,0.9964051683675776,
 0.5021776644476976,0.8141660832628069,0.5709847938834003,0.12997717388684815,
 0.30556430045816996,0.14403488500436468,0.13515179710900138,
 0.6548296663419497,0.9423312262012181,0.02775872741859886,0.7167411508149519,
 0.5194802121343458,0.7791616079562877,0.9548900219871003,0.9316197803645813,
 0.5822536729516942,0.01586710008337633,0.2908175096296851,
 0.055221578954748884,0.3898484677420302,0.47942393583750664,
 0.04004803468288787,0.31171376748080126,0.044040104990129336,
 0.6301986438751348,0.8543821645345411,0.7328362298680199,0.11613162501805085,
 0.47888270640250286,0.3258811552863097,0.829329473693397,0.4610298585793964,
 0.7738001206144332,0.6402605486244489,0.33092432900696644,
 0.39537171190584575,0.5699157181517395,0.2925830526250801,
 0.34135190763286194,0.03315610086737497,0.22588393371031623,
 0.6878376009025395,0.6870552151348984,0.9369146713835153,0.5303031478671958,
 0.21121185756236704,0.32302949573732587,0.38055402411036177,
 0.6399932664369049,0.2997332934097495,0.47342011287605246,0.6114968622910045,
 0.5871200313534021,0.5185773296608985,0.5013515176820726,0.6937348622756196,
 0.6647976379640135,0.44578521403899785,0.23693384573631593,
 0.9221277624115474,0.6436526920335834,0.8301195673025658,0.7303834148269958,
 0.45092053288475353,0.5924883486108157,0.45818284383399255,
 0.9418937826584534,0.39510233051103283,0.5718163632136863,
 0.46422674421391286,0.7392299988890628,0.8913495651982412,0.6426174164797829,
 0.5205231894258732,0.3388021792528809,0.624920922805795,0.14987991477684892,
 0.19854292824899433,0.6237595271349989,0.17507632314177757,0.917943487860797,
 0.9439202951456486,0.2409304384538431,0.5898232989524673,
 0.027938059531762427,0.21774140613520654,0.6564044842110627,
 0.3449289937280303,0.5183871083503654,0.9026655737510153,0.4524374480493384,
 0.7920206962598183,0.05321340965802368,0.3687731080271832,0.6218506196435312,
 0.1755496123501975,0.4425264573020059,0.57081516238944,0.6884276038256206,
 0.5546374986633404,0.21032548699340048,0.393620017303587,0.4803129006842546,
 0.705683930407432,0.22314181371615882,0.6406548487447608,0.7871276897435879,
 0.7779864214050481,0.2837830182127544,0.8467550160437013,0.8014733350460835,
 0.9647756992473362,0.9135128717059309,0.17668918598813932,0.5068765258087983,
 0.3083206542467791,0.7475508749257637,0.9197140086617065,0.24518434107122022,
 0.6719566362388253,0.585516356748066,0.7141082240594325,0.9232095159292739,
 0.1844474501096347,0.17955118292039596,0.219168148658756,
 0.0012321625066905595,0.7200258245443253,0.7999402270335072,
 0.14442064373767616,0.10806281577771715,0.9389938792760016,
 0.2921132956307406,0.09884975377534311,0.3573543565920789,0.3148851609217258,
 0.9639541070552843,0.35837228052958725,0.345343394699827,0.6095642232897995,
 0.21540809269827155,0.9087597397676532,0.9946645396031017,0.7887603342359099,
 0.07070831176669601,0.4350176495909145,0.9897758471839473,0.4300521492259721,
 0.29866142266955253,0.4033305862938532,0.26986275726779296,
 0.7559032830964105,0.2563354243360192,0.25802523651630926,0.2940335385752647,
 0.21015335323754059,0.8660997388463946,0.19903013171548456,
 0.09445066233830057,0.8654479750576434,0.5911086307150306,
 0.09904245426172753,0.33981367049961,0.4252546269827775,0.7551543161955586,
 0.1662578229630156,0.3121067776820552,0.7282955900711957,0.26394335669635205,
 0.30707425109181874,0.3514730979955263,0.27244855265497947,
 0.8229630225536317,0.8927982806904907,0.38543706388455234,
 0.07643326488557467,0.8588093379049596,0.2700148289435803,
 0.26285079205449213,0.2965013292360281,0.47791241115955607,
 0.9868351032815095,0.7642856882520618,0.9065885435647498,0.7016169893184298,
 0.7260666736412745,0.4882335437614983,0.8360037967917983,0.4874282556806754,
 0.2684902064098823,0.9624999146610834,0.05387438476671935,
 0.009806947053188741,0.8430437207328546,0.5191363461792566,
 0.30812475471270795,0.7256941081257001,0.22061721163503079,
 0.27614273342619133,0.28559230116475864,0.13382181669539162,
 0.4476213020332993,0.9321620228461546,0.4975325678999063,0.25266974325644653,
 0.5599225341601021,0.6266633139069004,0.7061901489400677,0.2539993151530864,
 0.6540384239683659,0.6297974959560136,0.4696075311946646,0.8292757724019025,
 0.7935024261033644,0.0026040790033641237,0.2013947823471398,
 0.6349638980248198,0.6269680094649671,0.549829392328863,0.1664767204592219,
 0.9353795343683274,0.476741627502288,0.06713652949871118,0.6434768781758302,
 0.8468721666508033,0.09133996991480131,0.46804764367737983,
 0.3482810892725895,0.9968454370200439,0.2934448787906019,0.49443689902591326,
 0.7163830538188497,0.4792763663146138,0.9056212168833464,0.6325973342633108,
 0.39358300946424407,0.49288545406180084,0.4465830534183792,
 0.01967274143311981,0.4250953607378668,0.938005123445378,0.07009266145427306,
 0.08829892216562019,0.14891916848944164,0.19114966953371404,
 0.9671067466265642,0.8043812292676388,0.5475938070270295,0.6520376780604021,
 0.8076959624560689,0.09667085312485568,0.09971525736825826,
 0.4677942718360154,0.5409575381429108,0.9495280713147719,0.7013912763256249,
 0.2490688749039156,0.5322549248888232,0.4682889000392778,0.7020404551220513,
 0.9655966376769148,0.588331155158522,0.9249794220239024,0.8142570346575807,
 0.35934844933787946,0.11416916786823339,0.6841217766000198,
 0.8725957108710709,0.5913723192399389,0.22288202524546097,0.9262671874830853,
 0.6513787583241185,0.107469437961906,0.43581393186207285,0.3623450766907309,
 0.01918794077823094,0.9510409149938036,0.6352700682313848,0.5038170160768348,
 0.1287136186300144,0.28647624123364024,0.21861826507547688,
 0.4261707998610419,0.7903858160635647,0.886022365873046,0.770123855312509,
 0.02092921596795072,0.39870316706297826,0.9829400439085149,
 0.6183191702072818,0.5990221774653357,0.6233961795653822,0.24328223077193134,
 0.5436412690761404,0.4482109051398653,0.19555317539577521,
 0.12213818435112422,0.11847793015935215,0.1635589424956707,
 0.41329354794665185,0.49042698298124143,0.4679037247102261,
 0.17685845617161877,0.7992058729817867,0.7790295183654131,0.8260013330418635,
 0.7095013934211228,0.6085066475912586,0.44451914004860427,0.4370384727980684,
 0.9100887588623242,0.08069020082895428,0.938738017476041,0.30628443454269094,
 0.027591768666221594,0.45272918969612685,0.23285276932569654,
 0.6988432372587885,0.9936708543196909,0.6052840269929947,0.26049819271675134,
 0.8208303357886174,0.176163812276943,0.600275209743334,0.8567380405884341,
 0.1382100832430937,0.634459326734897,0.6265008917534249,0.21159809984328026,
 0.5770279069253577,0.543505218491473,0.942630764952363,0.7508388211452652,
 0.657972060472934,0.38057119588970767,0.23796992382079285,0.9968265790880873,
 0.6746781786420823,0.9561114926039301,0.04577752750394226,
 0.16001240979028286,0.9206269277423912,0.10091651561824411,
 0.7482433654028758,0.680396507712594,0.32763881890564683,0.4864015037699613,
 0.7892068418863697,0.8280876783298897,0.9976712702023979,0.7825345561088358,
 0.9424170063781668,0.9850560390479279,0.054086884289113346,
 0.9280563185375259,0.6797189580289406,0.4201366576542106,0.3910207188740684,
 0.33724546370325015,0.5053980478578322,0.27740232795144526,
 0.5905352040230439,0.9869029913316084,0.053519269891004306,
 0.8838802052388195,0.3196446135894966,0.8614492144316666,0.7693092583676968,
 0.38167433680889584,0.011552360799135797,0.4081501319533709,
 0.5202392411232334,0.8935941868357999,0.5077565034178895,0.9032739325731276,
 0.8839363470343442,0.14940204526881373,0.1615567077894734,0.8662482604300504,
 0.8875795183263268,0.34576029361494587,0.3780964747039417,0.287540420997637,
 0.8293314756092098,0.14046466754029763,0.24748729073776898,
 0.02963000445387809,0.2679672931297955,0.7813489075574375,0.6023999046051003,
 0.825660260427296,0.2576517452217374,0.4147068346453743,0.371770001268531,
 0.8954351476362435,0.19851581678696562,0.8263689389178717,0.4685002890969452,
 0.40493270726001707,0.32816999113961376,0.35681753923910464,
 0.9811688342487691,0.34018283549031414,0.17709667608881507,
 0.6090940553742975,0.23124043657054427,0.8707247656946312,0.8467743615215587,
 0.8581263103430838,0.14609057109055512,0.7624273734605522,
 0.42533688820870563,0.28940099494304794,0.07693946196099943,
 0.8407023050656118,0.7752679635135824,0.7289080359713469,0.5860623535481972,
 0.956635957650834,0.14871584422733786,0.43685116841968585,
 0.32786802142448124,0.17357334773703414,0.7608045963155357,
 0.49305371915524177,0.18500932259322678,0.5605500976354785,
 0.6628082196574765,0.5369653211415774,0.8078737264559178,0.9879077437815522,
 0.1377917582200897,0.18794064696786994,0.3813180385211926,0.9070495628556311,
 0.6453719578809509,0.9598691016484869,0.2767871178104099,0.21608447960593524,
 0.29288825073607605,0.5205918773315885,0.4830660466665666,0.877849577699984,
 0.1108345491622762,0.12672330781332575,0.7016340389242112,
 0.42378772126681497,0.7163681941753465,0.5901908289135884,0.2605478956987478,
 0.9401677418452911,0.7935133526141187,0.454755193302394,0.6670275480655755,
 0.8081456889066795,0.892533211966116,0.5400440673034735,0.3143620558156589,
 0.7365927456887931,0.9949264046820329,0.9836875324042959,
 0.007189676381848531,0.6750351732083353,0.37921789077463186,
 0.8836484311678248,0.6493467811506837,0.09124199952649881,0.7667974428188354,
 0.7446763065929612,0.34963563108411544,0.6387371779436444,
 0.38625996245535255,0.11458557116786483,0.5436644516510585,
 0.5174569471494389,0.07423982845985533,0.40853049342486836,
 0.8292767637438545,0.8901412953057429,0.28662870501632776,0.7451908536897797,
 0.4816414233244234,0.6987450752594674,0.5057790763068081,0.11839876828209983,
 0.3835447206321946,0.2381602680341678,0.11934292665047086,
 0.038093653938633065,0.5654308336484378,0.8522279826935525,
 0.8084042865042391,0.8479099511140294,0.04388454643709383,0.9816323262007836,
 0.519247519165212,0.2800995948230228,0.5261034159584324,0.06401862807150849,
 0.6182828576925823,0.9188922441177633,0.3901629760498502,0.6809589870764151,
 0.7869084939516717,0.7937591421076725,0.627140315557632,0.48884055223087874,
 0.43958003170491367,0.13781163879022107,0.5088605601542661,
 0.4395291181763039,0.5478693940221733,0.21301762704784566,0.8559695868703853,
 0.3535432152534931,0.34026762413814793,0.7878484006825698,0.701717035550967,
 0.3112511361278485,0.6294620220816285,0.8077145313076458,0.9229577457792942,
 0.21012121422573948,0.4921575637942991,0.4244836220200269,0.3834981971260024,
 0.34659048825385674,0.8455637597788194,0.5519790295003824,0.8967651613068132,
 0.812346965676064,0.4784498924451458,0.255185757261992,0.4502106660252152,
 0.9907240416785559,0.290924776302269,0.5155552596435399,0.041490561992443165,
 0.8368750436537302,0.7554617636823627,0.2633096574586741,0.2583846405582567,
 0.45398519428470074,0.052026670575520906,0.6920769029790123,
 0.7194803441233867,0.27600003929581596,0.43198845986356604,
 0.35863814586445875,0.9060569132207362,0.34240941530349556,
 0.2810317314836337,0.9705989452150522,0.3309065747924722,0.7698296628009367,
 0.35362039188435657,0.11713836940643318,0.33017703995054404,
 0.4004589427033767,0.8336916074457583,0.5409609740579828,0.651993623521086,
 0.7717622918496793,0.4105030800332934,0.33612410152147776,0.7725716580546768,
 0.5165498793924042,0.13480761014642673,0.4381855871802053,0.9147806436957561,
 0.43171237194939194,0.7890313295668206,0.2573398553725982,0.7083069717278728,
 0.21566330872325046,0.15297685330571786,0.18432217689010666,
 0.6024762389476328,0.35113334910820737,0.2696333093760195,
 0.21485304801088823,0.13954260756173908,0.3999625700839998,
 0.42622950873684984,0.5710818387228737,0.6893254162592132,0.5066498707334469,
 0.02919762864872355,0.8752067489189671,0.3834307943760974,
 0.35493567108485746,0.9179862270165452,0.3789314091152064,0.8963597792633947,
 0.5376207397972375,0.21323890405455104,0.18948356635652341,0.769576573633813,
 0.8756100313429638,0.38746176547092126,0.37810306018146833,
 0.03181045900325474,0.8286574171618308,0.3649331915470517,0.3842772102138472,
 0.960965361607153,0.06690626651587472,0.8176158543571764,0.18932854976845992,
 0.9582645688397856,0.5527964765134673,0.7043662709797754,0.21558284346520584,
 0.35926033931462964,0.4752603998657494,0.5663023398567678,0.9329463816607853,
 0.394308388822745,0.525599749985779,0.9407781338585854,0.9631112031927578,
 0.9115178873056556,0.44359833838802165,0.26417004485088835,
 0.2421220009034597,0.6331333177426117,0.45531774511738576,0.8630134709766561,
 0.859268353551391,0.8049343808982174,0.8907650198452177,0.3226482912795111,
 0.8273027990380567,0.9021196071655814,0.7461239971868141,0.3259845341948654,
 0.2872073144033618,0.07878703278429444,0.2502721795713676,0.6173853092734569,
 0.7101973499528236,0.9335612777305852,0.6572874375514036,0.9575615875615844,
 0.5561876249424068,0.03179791382489017,0.6673950396959691,0.810507420745227,
 0.26442685500132457,0.01811873275741127,0.0749708557450931,
 0.5692266038664132,0.4878797745240998,0.9153122642172455,0.513674470284847,
 0.34672895939687387,0.5168186080035319,0.8087910267809825,
 0.26799328435837255,0.26645337624797105,0.7765052242412384,
 0.0510039257597148,0.9330111629682298,0.5658508111688981,0.5640061924795641,
 0.5926343395590908,0.9546723888065953,0.5615543754001225,0.7474597374110855,
 0.8543507173397338,0.9915993505353244,0.10279760838562235,
 0.12268937462037843,0.4961797177656062,0.7015271232963957,0.5773157278109451,
 0.4028186520036333,0.9342241352212171,0.807780087561283,0.24630093270091125,
 0.04633012614555965,0.5728375849916204,0.41444141599890316,
 0.2801867016996038,0.28212910460281193,0.7642497841573167,0.7534908160969376,
 0.14483058719224473,0.7933981429977006,0.868680271718223,0.01400570741115914,
 0.4839018142702386,0.3284366446975955,0.7088651454346591,0.686577574634022,
 0.42835881088628724,0.750910282256827,0.5192276950406807,0.7898336419044398,
 0.9055655407691017,0.9999638381930804,0.26539474135036,0.7198797086491455,
 0.44985963193782896,0.8840761723528324,0.5559552184046292,0.1937457882628727,
 0.6194313387719634,0.4709654301447196,0.5266895596676755,0.9040445333544357,
 0.5721278503817175,0.7057844906544282,0.8171058024970463,0.40776573658649085,
 0.6725702129801818,0.1929546391076865,0.5623183560630612,0.6763652402057476,
 0.46268881360792136,0.8738402562870544,0.06395074325225314,
 0.4616188402928383,0.3651650902822984,0.35430022938470096,
 0.16455683391115972,0.00518395528862059,0.45784186057299214,
 0.31534240364216093,0.4814718867302732,0.5758577267878837,0.7458743727988597,
 0.3787187074088084,0.24468122737690445,0.9343111410381826,
 0.11733413601187115,0.8147952799421124,0.5193299866097828,
 0.13284005515698805,0.1618945755971819,0.734777767484592,0.8593395846631366,
 0.40240491438314097,0.48673039335629675,0.9142029798404425,
 0.2334845653488773,0.7167166453019548,0.34514452030780884,0.5223020528928026,
 0.7066971889125451,0.24175671266705667,0.24945845172036107,
 0.9805871684545203,0.5769980807969574,0.7189371427124459,0.9508995135690042,
 0.7555100715520293,0.19726641721280402,0.09615482757711091,
 0.4221096962064411,0.07871100685401311,0.596200616565461,0.30281651891786787,
 0.3762453840398643,0.2037718823239143,0.7867982183179457,0.9008498382081181,
 0.25589031237907667,0.41939787258016414,0.5131374507164695,
 0.31305622017595836,0.9500803842104332,0.8796646617679688,0.7785623669964346,
 0.7988217258665384,0.5649236479679305,0.06619489986751326,
 0.12835526146987053,0.4227620399039188,0.9415420801053446,
 0.03608133279173986,0.15996382954047328,0.8738254867923031,
 0.5852780651523634,0.7529685183050632,0.8954431627630041,0.22852084559275188,
 0.6865366438055275,0.4172138040490161,0.9480088763027245,0.4518627071567446,
 0.5198961233631453,0.1147973643575183,0.2933457768253269,0.9888190873179062,
 0.8903582802731608,0.40897431840596943,0.8456049010559826,0.9471324025873925,
 0.2514028737492937,0.197287111128153,0.5322203904300693,0.5071720667770373,
 0.9132798360951455,0.39372314321595625,0.19597464376509088,
 0.06678230300702037,0.14395081320142344,0.8588463432981048,
 0.9620842266880956,0.5375797816401654,0.9221548516456078,0.3800567814103232,
 0.7206379344116417,0.23180535653478618,0.8093590054022366,0.3995254918987243,
 0.9483780003803965,0.5827753912241845,0.002227948643906519,
 0.0593828434835062,0.2764396468795537,0.875087684045611,0.764915790819902,
 0.274793966808611,0.006021310266131841,0.5638863751272701,0.6636357584336625,
 0.07974736993065679,0.05773299784662578,0.13096210532785924,
 0.19055959353477558,0.3455980346010883,0.297489354190804,0.48771347027673684,
 0.33932045238295594,0.03486110368278639,0.4610198666624993,0.528185247015112,
 0.9861411735596918,0.2170576887534852,0.07766482530604957,0.3632388065879708,
 0.2902475524881907,0.7424553137760307,0.1657419059036287,0.16344551999893975,
 0.6880321540622909,0.5394120906542149,0.6081973669311562,0.7662742122866149,
 0.04200077735873997,0.06315294372272817,0.19697278465864732,
 0.5732497757120157,0.2886469383111151,0.42770555367041463,
 0.20960192714438342,0.1725148827769547,0.06496195687361828,0.261934177318909,
 0.9238391681555226,0.13430101383538506,0.8655280362997695,0.6142422732070275,
 0.8850961569624537,0.1856723394672931,0.3645566818536722,0.4892284404267556,
 0.28207524183530175,0.5655118420437536,0.811909809987933,0.1650628075884053,
 0.21136969480142986,0.9046150821702241,0.3648675846228331,
 0.11962671620829046,0.22747831049681289,0.11154803635090826,
 0.7921198982829731,0.6672077973085135,5.2724954172966854e-5,
 0.6494673896155063,0.5684484991595907,0.3666355037925296,0.6723858010764896,
 0.21242363766747785,0.3619447147508501,0.6414969123688186,
 0.030775935320559777,0.8094896118046421,0.5378182234501235,
 0.4365846984355789,0.5082965751611516,0.2983802639546491,0.6222991387526561,
 0.5872494051111786,0.3769116747333019,0.13999941964997464,0.7935617215559114,
 0.9081382192004869,0.05822545567328741,0.6462236319331545,0.3458080190754699,
 0.8517184874049335,0.5216955846446365,0.41136395211469123,0.445498669535326,
 0.47004670100830626,0.7608194154382769,0.8267522754803127,0.6045394063196804,
 0.3290195549980133,0.13871799306524757,0.4767938491942525,0.6494129978174309,
 0.666651518112107,0.21251674405466803,0.1619016250914631,0.36740164621991334,
 0.32539115598762813,0.651393584101724,0.9258064998450313,0.44122897162160657,
 0.23322830370696068,0.33506194346427953,0.6926048120838008,
 0.3234265935525641,0.5140183891379642,0.6216819632649304,0.3339317800336228,
 0.19343462478884266,0.8086641006603612,0.007638560706202835,
 0.9239133921931465,0.9419484411535956,0.06660535732466777,
 0.40332628236958434,0.5081098834146774,0.45657415733107476,0.734694072608962,
 0.4112939586632426,0.20466401953892333,0.17567085489673318,
 0.20291272126778537,0.400439852236928,0.3148104176154044,0.13078795649889385,
 0.09743052675622299,0.3593519410800069,0.6388921082063561,0.9507105928107913,
 0.9908382075189095,0.9497149980961564,0.2870189811133529,0.8605839199659635,
 0.6214627858156306,0.43941350369696586,0.8725056776060736,0.7096848197179018,
 0.2212926041429134,0.022851441350279056,0.8426473811007581,
 0.08287127898155444,0.9000069288045776,0.4068067944354171,
 0.21141182684284499,0.7571903622923138,0.3004044467893543,0.8001458333541235,
 0.6334414121173387,0.9840168601745971,0.8763691523126861,0.20206221930958912,
 0.7511097104742082,0.8013934630066991,0.34446476566870987,
 0.11068441497025783,0.14961010143417708,0.32819408758429425,
 0.4643245610138993,0.268273938482268,0.17944128038227203,0.4099412506579997,
 0.01880285017472816,0.5436674868964255,0.19009420444208214,
 0.7216054492183815,0.9694742713451658,0.6495686125338055,0.4089896416101382,
 0.5688559852616819,0.5029969549269667,0.7526773550522872,0.23553713692093403,
 0.9609658248533928,0.3214732299744908,0.06276453504641333,0.7982301958578248,
 0.396908498029518,0.9824368465822311,0.8857240262085908,0.9872672671913638,
 0.008067970924421686,0.26858738223237577,0.3597228418078512,
 0.18868458293749124,0.7296685409078116,0.5082614827973307,0.9309661315668311,
 0.8494296996013727,0.08856554294591334,0.46613906055020304,
 0.7845519875849465,0.06456306311193494,0.4353249717241203,
 0.34247524432691145,0.646284103549866,0.3304986273227115,0.4660991173000427,
 0.12688032610161137,0.9548837624771644,0.8292407182119079,0.7631173550620696,
 0.14723284746404164,0.6659361202269648,0.7053557924616335,
 0.04160146784482788,0.8883508287499526,0.9571265912879758,
 0.14894768568953531,0.08724665194835346,0.20177435213926942,
 0.02826392496637442,0.42428388825557484,0.3841987015685089,0.93469249095152,
 0.12136903872137328,0.93171451178894,0.40117870128373745,0.3730833319995339,
 0.4794396644326153,0.6845808527341574,0.8841396564868506,0.7856516815800025,
 0.43975163491575997,0.8529396698677779,0.16133256669772578,
 0.08844237823372669,0.22218802128440118,0.2940853577856136,
 0.05976921286359094,0.35281014260409194,0.9714183285860999,
 0.42422826560753046,0.5214158108736837,0.3352451203209277,
 0.06825736244798719,0.5368054970356472,0.506275888230365,0.7355904314218449,
 0.5168376937981031,0.8571274164696722,0.2366185879490983,0.5046547775319759,
 0.9745026960358742,0.5495135921395806,0.2941356783399539,
 0.010940855217560053,0.3318131058746656,0.6047673920226244,
 0.5464576325563281,0.08031522134542057,0.22999600113382868,
 0.3439940719831668,0.3449227061389323,0.43219404494566493,0.8589918161879405,
 0.46114020991673654,0.3425030243901389,0.5000414107292919,0.7900903025357319,
 0.19433224920007164,0.9037246171020276,0.2593925802089313,
 0.47204540017608065,0.6883086443324268,0.3878734241903805,0.1479125663805756,
 0.24887367307371866,0.6289968066044885,0.7680316992523225,
 0.11249483455431042,0.11708126313095157,0.019608629700374802,
 0.06201248634838841,0.03853500762562556,0.5051235377563557,
 0.4466253988380976,0.2249129086215661,0.20502838096316878,0.736114153725388,
 0.30133947302076836,0.2058854961353318,0.08297520232350586,
 0.8097929964525316,0.3627406136013116,0.6091592173477038,0.4174703043025694,
 0.8904663300924529,0.6004564999637099,0.9962013614656627,0.3426719063628888,
 0.22107877118465546,0.47476722237947266,0.3805744003571938,
 0.5419042478769134,0.17431439154135375,0.42943485045574914,
 0.7562062974445124,0.130645171844313,0.4100375123006217,0.6846029609347499,
 0.030404115957730515,0.4957803960162882,0.18100247138203962,
 0.15847100437894535,0.8130894287123431,0.7983942476698469,0.5406596073914763,
 0.3121493113563526,0.9950689164881908,0.4618873642737793,0.8175780595096069,
 0.2281007762585302,0.9644138834660019,0.17555846888082183,0.6296497464640125,
 0.6126788632338531,0.05720422427120153,0.017983865924349085,
 0.6692445153305159,0.4954992973723521,0.6654469790953389,0.48248035216978735,
 0.6008200675479466,0.47236954776218754,0.9708153681192475,
 0.49406627494429056,0.23059643229017746,0.2180351290507262,
 0.6711383971586381,0.9936565709936541,0.8342465924473035,0.46803169315453663,
 0.315934982131132,0.33426206557401517,0.2289838516743996,0.20596699956381403,
 0.2713322987016462,0.9235585887252173,0.39645283224311223,0.1260625979371076,
 0.8988977773668634,0.28442805140453353,0.49001239468574376,
 0.04900887261551368,0.2497545188447482,0.7595188268090282,0.9459575778607895,
 0.26018510796768635,0.050902901084647234,0.6184897926476598,
 0.18830391640963662,0.7785467311681362,0.37492017336200656,0.714299363131575,
 0.14022715545866216,0.04134730323688185,0.34156747569156065,
 0.8705265715890763,0.5315394402330346,0.5053546111800962,0.8021344894405683,
 0.05032309108211119,0.9290102113974681,0.3141750616815995,
 0.44009550813515375,0.8517950847864078,0.44852452493226136,
 0.06662636086127283,0.860163525780353,0.04118903921554962,
 0.42967722400212693,0.9372854430941795,0.53219877116121,0.6446033505069407,
 0.32675678525367435,0.20250046808229938,0.7124813667174551,
 0.24306244378105668,0.9560289700878215,0.6549871646471209,
 0.12878303688727422,0.5700526117145996,0.41850341681726044,
 0.6206392600964363,0.5431951701265492,0.13693663882712004,
 0.023681484453626167,0.9063483644186541,0.06086553492799285,
 0.8400678501947654,0.794448185527924,0.20755078356974255,0.9598069292972333,
 0.9353225091148804,0.3113943798586718,0.06317518674727196,0.6193364581422667,
 0.5536032443556022,0.65327114666176,0.9877310122448406,0.4665343839928029,
 0.793257596714794,0.1733473369942834,0.5673551823953494,0.7742868905645468,
 0.18351100741515614,0.9111878028942914,0.06267144126511415,
 0.7704240294899953,0.9469342755895435,0.15655940372921573,
 0.22878078422341142,0.15095577130241122,0.36300124692395697,
 0.6943019453234207,0.44092779111096725,0.40862118268611847,
 0.3070264619832228,0.8159970314923388,0.5729995276241521,0.15692503191491292,
 0.4039763761208356,0.5300696909628945,0.3659706354430742,0.4246451569387091,
 0.793769338554895,0.3087023812059536,0.9223163963799094,0.26907506434470574,
 0.35407387708379634,0.6768765523098992,0.7236390872183777,0.5219603140745444,
 0.2496323086069313,0.5754398617521741,0.7920090856513204,0.6358961893604735,
 0.1291400560060998,0.36841897661280953,0.1820491297407003,0.4505427711130341,
 0.6943510769316945,0.9414898781805531,0.730848527672132,0.8751533581339656,
 4.768624852785619e-4,0.3255364133631804,0.2885346711671066,
 0.5080068538672976,0.35773223754486305,0.3759462806386695,0.7439394745314045,
 0.13083429262946922,0.7608063717128661,0.5525082074202375,
 0.18923628831270678,0.761165061068831,0.08835553616283487,0.2733748620550702,
 0.6193896457069348,0.2061539152614863,0.2172730660162997,0.3210594357223089,
 0.49915854050358455,0.8981980833525369,0.5360493907022779,0.3740592833793529,
 0.33787668928038816,0.7944849889210432,0.31740058043061026,
 0.006830096128192853,0.5697973369124453,0.5619016533038494,
 0.12204625417054016,0.8541628302302139,0.49474677421702273,
 0.36786700279278284,0.8729322721076157,0.659677710358094,0.7198250412324566,
 0.8090546290903254,0.9679575771178648,0.33380082103742204,
 0.38148707245328695,0.7874648639709897,0.32551305325886615,
 0.5537987217495184,0.09747211361986519,0.2931442768501742,
 0.43469119830534253,0.5572718299539999,0.7221854831912089,0.525396122989669,
 0.734422358488773,0.8287813773399529,0.939012382384646,0.3969051730536951,
 0.5268810845505305,0.22200110608371104,0.15293637129209203,
 0.5952935491862312,0.8749759252891485,0.19771320697668926,0.9386951514257486,
 0.374475126098174,0.489067181676481,0.6981506376441442,0.3871165798548101,
 0.4229794861362627,0.876787182714124,0.8366759335210587,0.8087391986792696,
 0.06179451843710526,0.34214543645397955,0.6318363080810342,
 0.0654762986306574,0.7565652171107617,0.8632596503057737,0.256792415736244,
 0.7133922768535415,0.7724857464342371,0.5819696234218642,0.9310846602838878,
 0.5228607059515455,0.5745436130530837,0.3037763534662259,
 0.002504989471256769,0.41117258161321557,0.9038114701502246,
 0.6963787090374518,0.45785853949164257,0.887116323982494,0.27926111996856195,
 0.9915128544909876,0.725318882119732,0.9742327927830198,0.09857320540580261,
 0.43344084121168713,0.32988184082116967,0.3107657153568941,
 0.9567142840011044,0.8846726817425811,0.8661101951263368,0.8991224226590122,
 0.9024848634020549,0.3534244124300887,0.28284778044627057,
 0.42458674699432664,0.5943121658894814,0.528227602888469,0.05215492493606777,
 0.9008842901217609,0.8649773876980904,0.558160107617177,0.2948155977049266,
 0.7847643322357224,0.1410725236354069,0.4877216544586037,0.48223568137123207,
 0.09135325506875991,0.7611508967311275,0.16444505329112913,
 0.11988650192746797,0.6931434940848449,0.8095485336235408,0.8902932838943804,
 0.5874924102824663,0.389055863857256,0.4373693732285311,0.041693214286324554,
 0.6304716793525998,0.4981382372899672,0.5701862577245225,0.1594214439402708,
 0.4454973644114064,0.8558374677529154,0.1579882812794695,0.46346283125345666,
 0.8966883833473883,0.4015400212695125,0.7635010112878058,0.20748105182155496,
 0.7304778005022357,0.8382671676532188,0.15740116501985013,0.5156964818248648,
 0.61394809170451,0.8696093823999056,0.914004679193872,0.36689803548284206,
 0.8810378852694516,0.9371179987974909,0.12950928980977772,0.326396373884033,
 0.2701083932243812,0.8176931434445711,0.22953118845588827,
 0.10894103085202889,0.029607044567639518,0.7052378825916267,
 0.08055834908308324,0.45545978190275815,0.9149358650746342,
 0.13328041767272714,0.047200589848113816,0.6656946854538536,
 0.5232910384837813,0.31327975206684244,0.6732022504709639,0.9291852569992088,
 0.00895282492168148,0.429854577058816,0.5541344168481687,0.11579041790897626,
 0.6894724843878667,0.4335133548540767,0.22875221966780734,0.1731532940649756,
 0.8395179810080649,0.3744756019104223,0.6411542903561108,0.45455695230853976,
 0.8595092375758068,0.4807113899379031,0.8763315903952592,0.20447978620247775,
 0.005593604958765375,0.14798408913066174,0.08511949589536316,
 0.2506907065804337,0.6011632417918314,0.004538709662354146,
 0.5305809442235464,0.9666104105780396,0.7773293381370787,0.09535359868078475,
 0.33199385017490557,0.16620746371477524,0.12819454039041644,
 0.3107307913745603,0.24653406528398225,0.036411174590064155,
 0.6773496100049942,0.9530678380805732,0.33508301763775006,
 0.43404683813318745,0.2877419189173709,0.38475026249736577,
 0.8942454502454686,0.13402683364769685,0.4640476493061463,0.7756666698376948,
 0.49098768010246363,0.29830884666266666,0.6979887031350941,
 0.3792913958505283,0.159568108617818,0.6943831009102438,0.9497487537189954,
 0.9356387976083413,0.9305025568445481,0.743212185916732,0.6158048924002146,
 0.476298528406755,0.8714897301053243,0.46676523383787727,0.7977051783550395,
 0.4006295681317473,0.5835692629968512,0.23690381457603538,0.5792925628461794,
 0.1992238526821195,0.9430068548179201,0.7373354479439047,0.9686604438170295,
 0.18644310621908866,0.8206976998089546,0.5908652642682015,0.6661070564051718,
 0.7048165325137327,0.05347284422684695,0.8870678424439832,0.2551615372143555,
 0.1083360505922587,0.8353970143591513,0.4625379128836198,0.24805557064265182,
 0.6912506682414852,0.4416795886233412,0.35314370634805525,0.8010211741868951,
 0.2285294477432246,0.36132351838359644,0.8808544064927476,0.8017932794137888,
 0.23559469477645567,0.6714923269462526,0.8677808687087177,
 0.02453808053342621,0.2532145060495846,0.7826950465041088,
 0.37552637406079514,0.24481714898533602,0.03160352537641453,
 0.4726594157903312,0.55291308512961,0.9975082604854791,0.9719067265805922,
 0.6699304288342696,0.09334310380738564,0.5820852182964992,
 0.06576255711701617,0.3086917395312607,0.19727672108586947,
 0.9444724761214512,0.6346785792408812,0.3186798269535669,0.8718442281922171,
 0.05529591245444443,0.4605603746606497,0.0527925290447776,0.4341255338236062,
 0.18460129963712835,0.30733875683824774,0.36183533682753266,
 0.4536427659590405,0.1002714301532257,0.501502675076498,0.9750251301549991,
 0.06666752090736561,0.46995696800666953,0.8202364329917303,
 0.9429279604275231,0.16456805391007845,0.5314880158901287,0.301432669395287,
 0.8843578833986995,0.40309608094351534,0.7835579446250422,0.3744285370581949,
 0.3699692136551831,0.3605633077528464,0.7874314979007058,0.5777621548090177,
 0.3397818667789946,0.9931300154390013,0.15584637429675774,0.7743218917899131,
 0.17113137942718437,0.293892429167923,0.5477257337306322,0.3750701817491634,
 0.3426996527527193,0.16330778499435816,0.2173263919156233,0.7749798149477414,
 0.24916290170698752,0.7304332171757799,0.5846400065381978,
 0.24723710218388828,0.06680723805312294,0.8392778730545518,
 0.9323944625343474,0.217612931821214,0.6576680553531822,0.44949428213719955,
 0.949079032190211,0.2330828699353601,0.5011676108808847,0.35994773617216425,
 0.9748886644312066,0.8743543853474423,0.8804442495548763,0.5544967903701106,
 0.8517015545246595,0.949488122784123,0.953163333554085,0.4724102934221076,
 0.8763005869001178,0.6808729993182719,0.3767351466288539,0.3083577630652563,
 0.35722544345841,0.8267664508410423,0.7605205690599007,0.37628252407271234,
 0.639090318960355,0.24963133041331897,0.6111815475253068,0.13517817599643378,
 0.998847594243655,0.8416792086389198,0.5794160579010915,0.49268370448271204,
 0.7713850347638683,0.8380242429357083,0.5154700508315189,
 0.052278435126604905,0.09768132168714638,0.8262449741881571,
 0.48557158399673306,0.2585795863450402,0.4703638596279974,
 0.054189467236631095,0.7474152774625114,0.561407382086847,0.6581635314485472,
 0.3532760848245431,0.5898557372124269,0.8264418049099421,0.4263869337909306,
 0.7765014296967743,0.5271306540535379,0.3777511433623375,0.8287505592388995,
 0.66749035464354,0.8530338665059934,0.6969894120869419,0.2661329094086895,
 0.3212395113551719,0.9084282317478618,0.18625182129148343,
 0.35517264482849553,0.9297649082120867,0.5863867372556919,0.6610306768881042,
 0.323859577850916,0.6876790784712268,0.6020019965948049,0.0633365523968823,
 0.5176732507094025,0.4795460752376044,0.030823465660936744,0.664878259698463,
 0.00795544887396693,0.5920956098802389,0.13461281717067952,
 0.4879542631632642,0.5569665687674881,0.04721077667339524,0.9769076810948486,
 0.2817980095751472,0.883682820689927,0.1251193288977137,0.216408778340902,
 0.22280402522643705,0.23010923346804069,0.626544834733206,
 0.15937102643616974,0.34439032115179313,0.043821389057113413,
 0.14702335654317944,0.6361053637579487,0.9307924932280943,
 0.38501120762883123,0.9302393841304537,0.0189672332090004,0.4818766853088504,
 0.1198590796625909,0.3909580103871714,0.9364317801626761,0.15310587130867814,
 0.40622138725252943,0.5310149762054194,0.2154633748119713,
 0.36132578265387116,0.21299223618289742,0.29498437497305807,
 0.5421429636876721,0.5498364601068239,0.12536707088021093,0.8323147031407341,
 0.42478638294896476,0.2000790197901997,0.5800840113574237,0.9309673032110383,
 0.35523383352144133,0.9808864152890565,0.5510558571242927,0.5516021433848994,
 0.5632955396667325,0.6859387699579528,0.8154659400760442,0.11279852834795978,
 0.528355472508269,0.27093909116351056,0.49861397938754226,0.5382375997604885,
 0.12199793219207833,0.8495077126657876,0.37912937132552127,0.578758326576239,
 0.6739798262402943,0.9474334085449413,0.5807364046462997,0.8766190993312499,
 0.40516012980490657,0.4723545077803052,0.6342592136077643,0.600990885840374,
 0.9864103556251322,0.2878843886754727,0.3695795155808963,0.07302847400889922,
 0.23233359269652853,0.9152613984821968,0.9128258962169875,
 0.40332695147219877,0.6297567414896994,0.26228556269470527,
 0.6065031107150375,0.9480799774355518,0.7826079254076337,
 0.015201572175858935,0.3733672293420456,0.392275474343068,0.7307936453143502,
 0.21338848431528867,0.7465024251964727,0.01749457451439118,
 0.07296419209265481,0.10283021532189429,0.5646202380020438,
 0.20499446542017175,0.8250882042200179,0.4784805526732473,0.3944547064784454,
 0.006533407809996228,0.7289094624037238,0.07879928605145414,
 0.9147237829993677,0.9697624402899178,0.8865901513953177,0.4976719792100681,
 0.20805042210538305,0.8919598554975212,0.6048861907444189,
 0.06849739435352609,0.493629743140226,0.08827265294317432,0.247356155886145,
 0.38687610771933223,0.3186746200451924,0.4829155678861252,
 0.29551141306330775,0.6957166184882895,0.7218537923691105,
 0.32096618219122763,0.7480579378090418,0.35952414537142374,
 0.15257545069154121,0.6499096363776105,0.6640237182051905,
 0.36761193463934305,0.08407482362247254,0.2755338841461271,0.684112789062688,
 0.4721600622393187,0.5353970358825781,0.7787171345734264,0.6580312372542945,
 0.7650243324011634,0.03933063067842446,0.5930120983972557,
 0.33893640854074814,0.6414821274351301,0.2986614023911486,0.4219143803722225,
 0.3463516027709924,0.5884741468538194,0.2470748242181543,0.6302111641563583,
 0.3901974450736607,0.7972131955007521,0.2822258424517812,0.5029781870811126,
 0.6266840538017381,0.1875912229843789,0.7006426592970596,0.7275954497834538,
 0.5469069787383724,0.6275253756229838,0.9446659151859226,0.05749721119992568,
 0.03193387988915475,0.9988255744910186,0.3393236394522853,0.7776422099284122,
 0.6094584107246176,0.5164892425215127,0.2602964535814095,0.8544820925643702,
 0.19700177591924695,0.03033683222098138,0.606281775324757,0.798360163638862,
 0.9952918204064015,0.5490047468556186,0.7478091831776723,0.42384329079954586,
 0.5169377402153369,0.27001851397325416,0.5978360831704426,0.9653865297872806,
 0.7034674983385156,0.9101075421006902,0.4554146641505854,0.3367714534125891,
 0.5634410569237586,0.4639069819250542,0.6216032799587883,0.29532677356487125,
 0.9002385062356006,0.36351115562090264,0.0991611485042947,0.6223212796077465,
 0.8001259123813653,0.8050309814118569,0.321517072565782,0.7698905022880713,
 0.9294200088685365,0.12497632484805421,0.8877561615911663,0.701566875821293,
 0.23462043820414968,0.5394803389785789,0.1513216201642047,0.678586637774121,
 0.20697593016650273,0.8771228942649865,0.9507867514971259,0.5117462363193044,
 0.9611131473766179,0.8460260556669746,0.17907345382243944,0.8663014193932291,
 0.42217669393653656,0.2447606834993542,0.674575396575287,0.03684779756546874,
 0.23623184729044588,0.5867974715409837,0.026732101619687088,
 0.14958633641963204,0.946757619298038,0.516153677364162,0.9856492138217665,
 0.31095946849261036,0.7202392490560745,0.10872652591876619,
 0.6312577730194623,0.5344930672341557,0.17589699188199537,
 0.06799448335317415,0.7467387686984491,0.9255112779672708,0.1908628984632753,
 0.9561131590112386,0.7732235303362747,0.7639862277849214,0.7338714489852474,
 0.6715228716010843,0.6495287345076637,0.772246652797965,0.5039825136558418,
 0.9292490125707251,0.9987457930958054,0.3005165734238491,0.14615040446564076,
 0.6450646637684138,0.02363552266144342,0.5581884557980126,0.5706198045602047,
 0.9622131546961215,0.7645643932903566,0.32995400754746207,0.1711998845766931,
 0.9785036437805539,0.9917145039874653,0.40583048613226336,0.8293467417341737,
 0.765192060275155,0.8304241057120232,0.10912151248621482,0.17301095858990678,
 0.34362921937387536,0.6462763684903569,0.8143713545572873,0.6032317807044754,
 0.7583873259572662,0.0236291134605372,0.2642254181815601,0.2854346327820353,
 0.34890611112554515,0.016300320093383558,0.4875786027126412,
 0.8151085639653115,0.06342503180174452,0.24771552878939518,
 0.020989303952429594,0.38366991625360913,0.46846437830674925,
 0.6390964904958059,0.9889687177915223,0.6775206704568393,0.27673158605778836,
 0.652465204425773,0.4225100842357391,0.015788770408877384,0.1299118369784431,
 0.7330709515768272,0.2958129275111584,0.26955831572524014,0.7031857998201446,
 0.6596058578191558,0.8928494663873822,0.5610246423385078,
 0.061423564466370095,0.6019396606694046,0.5106038191051332,
 0.014470153374925743,0.6020302253358407,0.4130847871372809,
 0.6343592460823286,0.02350371308550281,0.16921221302002087,
 0.02732701071546606,0.14372207191037045,0.5720724021588326,
 0.38391953327882955,0.009005389060961488,0.5058452625904718,
 0.5898759114606655,0.17341041094286502,0.3496169659230879,
 0.45209857777578055,0.326922584590863,0.09644374996370508,0.8922430609915984,
 0.7028886375987753,0.7762837941089703,0.3441750122562093,0.06016980455022003,
 0.6902157779536225,0.23351304925145755,0.9201195684685793,0.4909662360668955,
 0.15567351231464366,0.8830703987177302,0.7210358154794333,
 0.23506259761200232,0.3428449080692346,0.9169456861703995,0.8428389545552282,
 0.811152116031553,0.1753355737076011,0.9126803422892963,0.1074257307268095,
 0.32465773958696986,0.6935104533311609,0.07694640924285678,
 0.46895542165389603,0.37742020699754675,0.6582475761771533,
 0.7881814870241661,0.853637697095312,0.9735377534003301,0.3589592976526078,
 0.5180572585883461,0.3916654922758581,0.7934771979086606,0.32931167321035915,
 0.05237652840055873,0.9852874231715274,0.21930594280819504,
 0.8977863275465268,0.47321640774370255,0.36521947594107873,
 0.34950424969960436,0.06204085940443138,0.5289862837547781,
 0.25662519738170275,0.7659078418628242,0.5192966650729689,
 0.008801516039212087,0.6195361791604016,0.4090195224532672,
 0.8312477032454977,0.33608685570477115,0.1287295286125001,
 0.26124472412936983,0.6393109555966419,0.769819594913365,0.2700484508894633,
 0.8813854084761404,0.6238825520134,0.25812219166116457,0.5145503795432438,
 0.6808218198774888,0.15579184587624295,0.9083488151097039,0.9130668618931973,
 0.19679057686207646,0.6665095800892473,0.5415698366070303,0.6648683220584922,
 0.3166837914031845,0.9999733478670879,0.8390915861749074,0.6964577431345982,
 0.9349434389859153,0.4043126476653931,0.7725316732382931,0.36081087682482704,
 0.7860624413772515,0.3953394833547239,0.7037299239147273,0.3479173352594711,
 0.5809672318141061,0.48450146682387485,0.082825448063643,0.1776064830933468,
 0.18829493006808318,0.6869167391660299,0.752025185220935,0.3829387612744537,
 0.5151287061692602,0.2710632679066355,0.6219692319229301,0.4822820535022796,
 0.6240072382177176,0.5132111029634284,0.8246994533443208,0.9866599703055883,
 0.5651549140248004,0.14750552073846135,0.15322060496357293,
 0.5048838927564681,0.5539662976981937,0.5791569841487815,
 0.029793904687434036,0.5594928710636711,0.5050295312893267,
 0.9638631796243445,0.9733557101707729,0.019314679058493267,
 0.038061187334233026,0.8150189095660292,0.7630862187470722,
 0.6857379702089326,0.23934528323694793,0.010659498018585678,
 0.4328845363007283,0.8508591532889933,0.8951862785770713,0.8298218284659896,
 0.5825196403243414,0.19939398891201265,0.2788674104466009,
 0.43022884481671575,0.45024948265621334,0.05378945506517652,
 0.42259706926942986,0.22964699111818931,0.5724539992968655,
 0.9418938926295072,0.24102789159049953,0.7405064643730755,
 0.004332255000595309,0.13837466814564903,0.9691801261042917,
 0.768018952632169,0.772916296601489,0.9800539376216303,0.7014101829856277,
 0.27634663853354,0.5183890934856842,0.3576377450237692,0.7155346544048905,
 0.3016547283296599,0.6183280072198348,0.4709365827454848,0.3202937482598006,
 0.8162660145570986,0.452063807164757,0.3592533482204603,0.5508417730592252,
 0.6650737086438607,0.7094088366156213,0.029851931823587252,
 0.23098716798704833,0.7085702764714862,0.2776137313877177,
 0.31914133593724003,0.326134260243773,0.21476384414446792,0.5945054125386012,
 0.6804467138504986,0.4487665805383748,0.3587885092698794,0.3919801838848249,
 0.08302734815118229,0.7764273958782473,0.9099453661787064,
 0.059280064244159436,0.19764650003248585,0.07591186775046266,
 0.14264215687300952,0.49176087541459557,0.7718833414376549,0.42423585225508,
 0.9585580663076236,0.6125394022418378,0.19426269373761962,0.9236281238629707,
 0.37172805765605743,0.43749273938595423,0.44552090146004386,
 0.17008455189747618,0.12135443245784217,0.5318115073297194,
 0.49120539354506176,0.74753338300286,0.7968758432007763,0.25845134558346006,
 0.5670581940706171,0.5647394373917431,0.9490999319990615,0.13670817201927699,
 0.17084186970242565,0.7970337387602336,0.8822477013810408,0.1874831595314823,
 0.3359224693748377,0.26276317466115806,0.7700685416672702,0.8265217834406582,
 0.7396397684450488,0.07031100709486915,0.2201082972818128,0.6774890573967428,
 0.4631426430204174,0.6007735110195844,0.8428859446607078,0.26833796581245684,
 0.1296039676442754,0.4986441115895399,0.11058172956313062,
 0.17341129330825766,0.16521677804086599,0.9308883490919933,
 0.47801457859335594,0.09089022120405843,0.824489044670584,
 0.43324160328956063,0.7354254806521721,0.2114584309697075,0.312141142799763,
 0.6649365512780406,0.8925077490926685,0.834419020040901,0.13252871663246601,
 0.17023414230430112,0.2591544176586489,0.40548860980283674,
 0.3339155044698612,0.6903654062618275,0.49381183364779146,0.9906465585832743,
 0.4924198937654233,0.08628926553138738,0.2302804292931101,0.8363819898891219,
 0.2568515140590728,0.9983482996301313,0.5137204185484351,0.8901486012929011,
 0.6962900600032806,0.7246307515594835,0.9086958219657546,0.7849418490368733,
 0.5108877521099567,0.7859760626992748,0.11600255103565227,0.8130664950726758,
 0.2723856509972149,0.2489669078912199,0.8091312922871444,0.9446610902014351,
 0.6660770001988973,0.18316944534952107,0.33275313979489163,
 0.9936298879291097,0.26743489138711685,0.7422440088015635,0.7422087701626194,
 0.6877492714386103,0.048145006960419456,0.05373230831886211,
 0.19215280161724724,0.548689175824343,0.29171251149687205,
 0.49936914220954864,0.7289974712486321,0.6763960418940904,0.6522719652099569,
 2.3502439525036056e-4,0.6200705391903844,0.7413363882215517,
 0.9643411220296924,0.5264358475261456,0.755791213405872,0.45704017571976663,
 0.7351844734136128,0.17914998987755038,0.11720669794409211,
 0.2491725167555211,0.7859689176230366,0.3830768333052602,0.6516588916388834,
 0.11429233009015916,0.05999674684417999,0.17530170348217378,
 0.6829796524670353,0.1535170161160807,0.4122765992445151,0.5011226871280317,
 0.7213563669044232,0.9376453187855991,0.6255676308973627,0.35150947196253224,
 0.45039473157652343,0.8635694137571233,0.9914017957789726,0.2419457244358727,
 0.5831883500346581,0.08725338625886736,0.34989785032163423,
 0.8926172541978382,0.6398215219841356,0.8637619670991805,0.6167407243323134,
 0.49404920485866655,0.8620123448465964,0.4495694658562629,0.9465230875589895,
 0.7951003574590582,0.10653872459457836,0.3646452163758358,
 0.08625862087654179,0.35637781989800243,0.6846212329987545,
 0.1718881000244482,0.19767490791162368,0.13819351904907018,
 0.9719837947242966,0.1716897226503067,0.6592051325955398,0.45810059377125134,
 0.254876023823032,0.5689852098025345,0.2486168226466411,0.33340188902098955,
 0.026306474414381942,0.9673323822706055,0.3681060754563088,
 0.7256765665541259,0.33057408730006976,0.3840884499340418,
 0.45928909161008735,0.1879015147698223,0.46521177262526625,
 0.4589897531044188,0.34075351289123423,0.061178624380160196,
 0.3359042556481744,0.018032364343919194,0.7729527538886465,
 0.6639623646072548,0.1278730991041792,0.44322387639259775,
 0.012183215127385849,0.3278461989982019,0.5998362880298858,
 0.28744940282263676,0.3141305341929307,0.6161509619560266,0.9948343335204892,
 0.29219711873764087,0.9346632090061522,0.8244481132134851,0.9801150799817793,
 0.9776106807633392,0.5641139565073072,0.7984261570384878,0.8118657173099902,
 0.5558957151889429,0.6557266313168598,0.5607218966538978,0.8608690900026658,
 0.45029109541434453,0.5685400831999883,0.46546581423162314,
 0.7752172330776685,0.6712427679301334,0.42763971098366016,0.2987490865170076,
 0.30078336330287236,0.21899029008767879,0.04770907160809379,
 0.29431606867275095,0.6918617122996504,0.905910513731445,0.6915519597675487,
 0.7284824802174525,0.6604722377852039,0.5603580567772385,0.1564872426248236,
 0.7176868865360543,0.10564130531436422,0.0909148296858926,
 0.46053323701131454,0.488101828976795,0.08127301211488391,
 0.28868264827368606,0.2585253214043459,0.44478855687701013,
 0.2199968558113271,0.0324919520285778,0.07805079675055937,0.667458620957193,
 0.5073916794941862,0.9525813551790926,0.6206852764340527,0.9836945900970896,
 0.08728739925500317,0.4141549334132738,0.8443399135357625,
 0.010659911195908589,0.9467325764973642,0.7260014112658036,
 0.26443616680410154,0.6530812327084334,0.28166379790527596,
 0.024448821683122568,0.5119790180556149,0.6715222159445122,
 0.7445361993573799,0.8421795211369512,0.19680703912258946,0.6776913003007309,
 0.7284946211255539,0.6942912846873621,0.8913353252493025,0.9678244851063387,
 0.3177443546545813,0.28799494016039584,0.870019052273729,0.9329448658259992,
 0.2578316969643307,0.2668938846375163,0.685270966698364,0.31142787060939003,
 0.7295110990908251,0.29647119388974874,0.8418404705647894,
 0.14279642686566785,0.6459523557675961,0.8350628619172505,0.5913211612133108,
 0.5012666538817037,0.30186249685713373,0.43468736035296207,
 0.6524436162490356,0.37333395927553203,0.9140353062744893,0.7905143071570545,
 0.6055350945135431,0.9399439459330768,0.4699928798211359,0.8300212604903228,
 0.17612496478120476,0.8610558461192025,0.8190268283170401,0.6553261514895447,
 0.48712390387985516,0.5161096457059234,0.9371964628108344,0.572431101138274,
 0.46511939764107,0.9549468335938825,0.5174585717889412,0.48925998734152853,
 0.06283332742232872,0.5347932571250268,0.4076286945087133,0.5296384402779736,
 0.7572995349348182,0.28729335454011773,0.3915106357332875,
 0.23348275423036302,0.07359901204818775,0.6568678168782528,
 0.15145878149554037,0.10366132929876348,0.6761234799595478,
 0.9716772233686028,0.0432902127298066,0.35915850827636964,0.7123444361582965,
 0.6652464594681132,0.2754138813774567,0.755287399917938,0.7796798087406593,
 0.12297297550304664,0.06343246177694639,0.9313175724083309,
 0.5789575071196532,0.080490141956721,0.5606689918122475,0.9563623006067701,
 0.06011478748952703,0.31574217630178236,0.0722231531658285,
 0.17680077924620663,0.9119249860177712,0.7712323321314479,
 0.40553532670358383,0.7248941294565644,0.7553513227834271,0.8092078290981977,
 0.05280886659322093,0.4064708468285597,0.5802202423971973,0.8598927378426175,
 0.2348616272931957,0.7052606290674697,0.1653940526855281,0.5557113811987715,
 0.7531198235450718,0.8919586601097443,0.8623842202280976,0.05279620513047256,
 0.27109872777067134,0.31912996398433047,0.15837635315188758,
 0.9953366789094873,0.6474275362341975,0.7176179554802541,0.8482224300463714,
 0.5221205215346887,0.8744720432965629,0.6652757100114529,0.6809368670184107,
 0.8997982598941235,0.49969986926516463,0.16852004020728306,
 0.32688685598244793,0.5262232730587867,0.639458978872584,0.2941686069910884,
 0.8853224692934913,0.11903045653408983,0.5884218424408799,
 0.29319189657001754,0.10127665768884941,0.8726339645344583,
 0.4614974751498728,0.5848484503238995,0.3664811746146668,0.5185707954876952,
 0.9095160399765863,0.45487112696879006,0.7126833758381554,0.9260206696756985,
 0.5740573196261702,0.9250803314961069,0.9586289982018503,0.2626357616735342,
 0.9531773618696819,0.23573480493911403,0.24239498416912286,
 0.22694113110143022,0.10120280635908552,0.8009177660127222,
 0.031196062886887366,0.6399477601655511,0.09858796765786204,
 0.7605134367075268,0.734876280665417,0.31218942419376405,0.749423268046453,
 0.6651990122391406,0.07126567359598579,0.9107583339906755,0.7374277000603193,
 0.08936928205150596,0.7565400652819729,0.09738718491050224,
 0.7652357928338626,0.8395750427981475,0.06010964460229451,
 0.09580808706418353,0.5690402896400499,0.03615542761123358,
 0.3276964087597527,0.819855047686251,0.1970256843421032,0.7388629847230838,
 0.5953895711589033,0.03766741321444711,0.8304616535239514,
 0.28067836107415656,0.8254854511006869,0.14795984870937773,
 0.028178203209820896,0.8335790256701876,0.7841037950237695,
 0.9065502449787576,0.9981048646466062,0.6967145597901064,0.5017416531740051,
 0.10574883309090433,0.25200620658536843,0.8624626945606916,
 0.4759873592330469,0.5534330330919377,0.3018477390292068,0.42239645244333157,
 0.09498932027459162,0.21378086771280758,0.7018329404259208,
 0.12226590139878435,0.07487004986426604,0.15302965239164668,
 0.3395621826234638,0.8309417946523316,0.6443244560862695,0.8714467658280478,
 0.16301414703989858,0.5301165441875216,0.3597026662011462,
 0.31495158698235204,0.40713454664572624,0.2839593183165482,
 0.35311416860157174,0.02048998343097752,0.3377608321133194,
 0.48735666716150694,0.49828720275151317,0.41228188370167285,
 0.9955645493673105,0.76186243063293,0.6963640203634281,0.7533325320119052,
 0.18813387959976247,0.43776805496721083,0.7441072956805578,
 0.3349545177488862,0.1897930368345806,0.23317820996135286,0.6829402730891646,
 0.5678915017480568,0.24134781864950838,0.43845661675210346,
 0.43754584204275404,0.002610139861539462,0.7833368989095303,
 0.2212536705734306,0.39409437669080494,0.4629950836975185,0.9577193823894864,
 0.8176663557820025,0.42867954975669254,0.34899073134434555,
 0.1813695777759371,0.3333711474972936,0.9403936306303443,0.5780765639785674,
 0.38865062430342145,0.021233121665719,0.23283133280414747,
 0.22900222529446612,0.8990257866060427,0.44418485626926973,0.263608962345634,
 0.5176632037444235,0.671405054893479,0.7785797494774803,0.8596681502652886,
 0.0074030494861274665,0.8588213030586126,0.7293119649882556,
 0.5123560238305873,0.3232057639100909,0.5742571623855188,0.2920140327908136,
 0.6591149882257898,0.020420796138604924,0.41402903846038563,
 0.11187992363480292,0.07555369965321168,0.16143747702985756,
 0.2623861089859856,0.8317475730775329,0.03439642166677315,0.5721527278568571,
 0.7720387896953519,0.21376214465208088,0.9917078644493137,0.6729450672412085,
 0.5407614913281824,0.23285574975094336,0.9544136761448724,0.3909998150732771,
 0.8271185685144864,0.5707044556612183,0.7359432147958517,0.581532445436016,
 0.30172317291792805,0.6231386603607096,0.8857546967846335,
 0.044902977053458915,0.48878718275170974,0.19415619300604348,
 0.6428846118794691,0.6455425945286897,0.2745999368033518,0.1314854991208818,
 0.36041708595641664,0.02685109220265458,0.9632287723342393,
 0.4874556841003722,0.3793668212880573,0.4433556390558716,0.5920681782870453,
 0.09999814106598826,0.9491127163814472,0.07229647896020197,
 0.018530432564635047,0.5676599403160436,0.11868309400511046,
 0.15112144105255954,0.9399107889696526,0.9054528418764382,0.6280456251399259,
 0.19592765046257388,0.5029372486678855,0.6740110795634007,0.4873361285107298,
 0.40939991744009363,0.37459081725422627,0.45147958698070934,
 0.08231123641394711,0.43641734360509354,0.9969934261255474,
 0.4101987441563828,0.5572848714813414,0.6693348001307,0.40820906483164854,
 0.963076152273469,0.508151501980455,0.9018030804283398,0.016448946920190943,
 0.23113327203894718,0.5670079581646744,0.5116874478667884,
 0.002471978841856348,0.8819033363500518,0.0946672684841594,
 0.7347748607772155,0.1745056559284197,0.9997227100737445,0.5882808728175511,
 0.8101407700194739,0.6101071072870744,0.291682303619879,0.19213499229318032,
 0.12903843894393718,0.012634015889485095,0.6313360300399209,
 0.16761239674913064,0.16853039010648152,0.8497611741141957,
 0.8107710200693459,0.6772732314406309,0.36942105361250754,0.8802790867169344,
 0.6224516848427433,0.9526850571669522,0.052190233673411135,
 0.9122029813771135,0.3267279441179042,0.568656074977169,0.7337134538101932,
 0.5930503168636229,0.015164055821105249,0.7214791719863246,
 0.9145852915765643,0.8111937599696493,0.4435932205818578,0.1272086460299049,
 0.12999025863976543,0.7912391716298592,0.8573013216478619,
 0.43504393744211833,0.5183946446100609,0.5814005632983329,0.1873556310255695,
 0.9845383605171113,0.18876770527381181,0.6190304037396537,0.7138140135213608,
 0.963503216430993,0.8380314539909502,0.27144842777764344,0.06662781388337813,
 0.5422988653488999,0.18925810919879282,0.6615138405108152,
 0.20421623762330343,0.41003861918458806,0.2519313601389317,
 0.2069283218276584,0.4961891108100025,0.7197901244182778,0.6535375568985213,
 0.01783862419290716,0.7080352926211321,0.3493836082469619,0.3468562085999536,
 0.11083282837369823,0.16870971441744487,0.592912739057252,
 0.09017843783949941,0.8868699219703732,0.34008985047057305,
 0.8500983907689528,0.32972678064401384,0.8430980554984242,0.2538850604401861,
 0.011195500170691042,0.29184566548365776,0.1411714190464255,
 0.6354737803296366,0.17939208628786507,0.46055221792655854,0.714269221697477,
 0.9150615858546467,0.37107214595344795,0.4825248716955368,0.5416599885623776,
 0.6126496808881519,0.5575830741511424,0.8734369918739103,0.44426802096526086,
 0.8898611642713011,0.11883284217090939,0.722513245217568,0.36999133968723286,
 0.3298721983157832,0.08958386975124522,0.6182347733379202,
 0.45690680094146163,0.8440063735200193,0.20148989947375662,
 0.2205760235310571,0.40974683766567765,0.6947020030055008,0.6573920210442465,
 0.7131152579546192,0.8840060242950025,0.3436805320330769,0.82966924640934,
 0.5622025509586075,0.6972017632757771,0.07783665976484755,0.2972568790304263,
 0.5576398827942455,0.14336996290157544,0.0021193587287198223,
 0.09638308340563195,0.9095130085374095,0.6588873267486086,
 0.21520605752304434,0.6267289827332171,0.33214348785776415,
 0.018106065162355245,0.9310988479136113,0.8272811435712805,
 0.1729546812680144,0.07632587711211836,0.4041327801689726,0.3727081131848935,
 0.010054877263332096,0.07065163976308808,0.04752408027316768,
 0.7819567015397487,0.7573777987047934,0.44860989794098605,0.5013594286669708,
 0.12699571097080709,0.852605987292556,0.6320197327327264,0.35000936237992675,
 0.7106486137537231,0.8703667633003386,0.9940642505934169,0.710847452057658,
 0.6556164920965885,0.9017711480795759,0.7545533931921091,0.8724250697166336,
 0.14993507439214537,0.9387486180540109,0.9791038794227329,0.7833247409531855,
 0.5152780645865146,0.5204100465029533,0.8274025145375667,0.20781952926966485,
 0.3905661596004756,0.5925413562940282,0.20163972417843568,
 0.49528342781635204,0.48082892944367717,0.6234544940051031,
 0.4842436474400911,0.7323801172151482,0.8006046795081692,0.20627298790708315,
 0.7394049813795595,0.5241512758893544,0.4646817656862804,0.8062024354441607,
 0.980564457535547,0.04666865547074006,0.9904757997757657,0.24786326022764826,
 0.6827897430841965,0.9234511285247519,0.12022886328976734,0.5403667495127162,
 0.8267723752653218,0.872283889397795,0.5656830123006911,0.5100526693521377,
 0.5770368161734403,5.250141056402091e-4,0.580122263773563,
 0.23702549546765905,0.5528965880069834,0.8596800099537969,0.9878876744598187,
 0.2743618662206422,0.3835745650845893,0.061564656586978295,
 0.08324134380258175,0.2882251559816098,0.6360273423459154,
 0.024561730154698536,0.422257551885992,0.4488232144046683,0.8793356470841321,
 0.08628498696814124,0.7174831812533159,0.871774343655196,0.41612952506410994,
 0.6536142782198209,0.7612789619833128,0.8275040886568859,0.7314455670345062,
 0.845911309515215,0.8734613937373359,0.05272800933246391,0.1184607816320955,
 0.8439866904118494,0.685308599664229,0.26819494311745806,0.07335654785177692,
 0.9777073465109971,0.15640462183494586,0.5423816353255171,
 0.21767909203958746,0.21602461979434862,0.8287730340147509,
 0.6783462922882568,0.37911596631657285,0.8717016730119319,0.1716638676193757,
 0.21894333144332334,0.37492710562839615,0.005391905479413928,
 0.1957056503774557,0.5314488150536205,0.527250811455827,0.1939635543945475,
 0.9559062884663012,0.6881959428819086,0.8542284993306895,0.7283172425189853,
 0.21600721880399665,0.45741483608478295,0.645244258985145,0.9336926770842561,
 0.1810396438902533,0.03529349299188833,0.03362784949847941,
 0.5056865892778835,0.826002938390851,0.646637352371172,0.0337953461884537,
 0.995370064895248,0.18185308482670925,0.46764649017181403,0.7499439087804947,
 0.9268581777182816,0.5365343139917964,0.31301018567427363,0.7495442632189218,
 0.10809793982028826,0.865228067450368,0.9353304348372344,0.9922822296206637,
 0.8848788233372274,0.5861473047182626,0.044912016331736915,
 0.5440141743866296,0.0820933656896572,0.3533062844844712,0.24728226370719109,
 0.2618765404316671,0.025997274666796577,0.9426619834102141,0.631565742194143,
 0.22637368264930058,0.3866978798698839,0.6329909331180837,
 0.08763546642893016,0.13049698585160108,0.3325404016180056,
 0.06050087683111549,0.8894362846811275,0.7726247642657275,0.6551252330213729,
 0.04440167017803065,0.06799213410511051,0.4737752733399806,
 0.6713604724518195,0.8219048839303701,0.7134796441674676,0.9779366987625842,
 0.6465220347841925,0.7161838879544405,0.5832991895467399,0.19569550126063529,
 0.19964338403991122,0.20560272994714823,0.4356709010847124,
 0.1291307188862244,0.0032372435309040437,0.45104244489965784,
 0.027497202957392886,0.8456823317971143,0.5688999385614236,
 0.7664168435223823,0.7767470708757832,0.971587106635855,0.7691309011797097,
 0.17145743471072372,0.24855796747486136,0.6911870397447074,
 0.5842141509673846,0.6476527814669786,0.753696550816309,0.5721464595813415,
 0.3783220933884519,0.7119911101147691,0.9099153833887635,0.6083262233213665,
 0.6526991993254969,0.9339822506106169,0.8228603207467197,0.911327626603501,
 0.370238697428383,0.1842588499138711,0.8148786409114376,0.05287095244633777,
 0.7207745728975817,0.7542935969604144,0.30537728447151014,0.6234961402738171,
 0.7104560510667076,0.8525108671395567,0.37744896313552245,0.7867792998797926,
 0.6912895418114611,0.297824478417103,0.3786929348819772,0.3082054706621844,
 0.8204161449362207,0.01636348551832678,0.13111207363665878,
 0.35837305878168424,0.042173250353798486,0.2505657663314116,
 0.0671580429924804,0.8616480341291177,0.22054163978548136,0.6039477046417712,
 0.9837432399191537,0.258888444002931,0.29193893696382534,0.7759624701772059,
 0.46616399405948394,0.26416706125961964,0.08993678512036585,
 0.6704744037510094,0.8340480930800664,0.49712907048784116,0.28898881019599,
 0.797230500659818,0.9315396538078841,0.9928795892203751,0.9940284131275855,
 0.5597236557926011,0.9635995012755223,0.27817359238509887,0.6708296729362578,
 0.7488990410477072,0.22798968141408482,0.7423409780441875,
 0.48937721244089205,0.8678246328806422,0.5155753354796344,0.8928608597455151,
 0.6874896543732395,0.7662685680143123,0.683278506697369,0.7012964187455297,
 0.4424493191811395,0.38399578566570003,0.09345995799579054,
 0.3974301125461396,0.24589538799780308,0.2668587539824766,0.4067915379917757,
 0.8807004698869162,0.35776958064503783,0.13404165946195967,
 0.5487695900932537,0.7627517937009578,0.16262079324070577,0.939680716759618,
 0.8716364785831581,0.09949617588174797,0.2631567372786281,
 0.005316863558594287,0.2730567777715991,0.3343119584907909,
 0.4326918302089181,0.1902288955635909,0.22166068921346338,
 0.29023594252820717,0.8074088490253674,0.18140431122231526,
 0.3639974922543987,0.07320498287978927,0.2933354157935295,
 0.20656404703846953,0.18967778940033975,0.53017442957882,0.7799515554386034,
 0.4843840041775943,0.40592298437915786,0.13826043606372218,
 0.9513829651814615,0.5328017261718048,0.8476540106300263,0.21627067447388892,
 0.2820762963227128,0.7565440233854169,0.7879567468230971,0.9361745303812603,
 0.4883373656111373,0.01366710492814971,0.3950998090546448,0.362211407095403,
 0.6247513602122707,0.613289725577825,0.8397645498048192,0.646502820954332,
 0.5839454478602509,0.6975828299120492,0.6073297247749596,0.7539000517727665,
 0.10076839163367768,0.6502590838763304,0.7420619487141923,
 0.08930513335098045,0.8946014382126058,0.5451716524115222,0.4611527557663341,
 0.23427205985904664,0.11868110583771341,0.08889719316597566,
 0.9945965362819322,0.5856947623009845,0.22759268099069674,0.8253117520294172,
 0.3537219016048544,0.4333212782875019,0.36075669011396183,
 0.04534283463058608,0.7460109055414884,0.686303080768535,0.06549158614068284,
 0.9564755382372133,0.3288371797884587,0.1554587347265428,
 0.034374647315404516,0.8602343261080766,0.5259692607768365,
 0.8179904836335179,0.2597541608632412,0.18244963918085144,0.9682460324744812,
 0.11897286458237999,0.9320316479460824,0.0809145740736249,0.9464192167610374,
 0.19361475225736258,0.3544417206915983,0.09099936770015082,
 0.02798877391879917,0.09978362704708932,0.32212341680893075,
 0.8065149548499795,0.25052731903552405,0.5626870093673679,0.2199964485863133,
 0.7777453192615355,0.9558939744385593,0.9295925275046221,0.18951914244128343,
 0.5874556802671833,0.2237189407021114,0.7188717124553736,0.2421738041624104,
 0.9748146937020522,0.3227820694368878,0.4729934789542478,
 0.004761541779798595,0.9312116570038552,0.8508356562363628,
 0.12017534779160399,0.692351212322908,0.7613188469677679,0.3324902397609044,
 0.8427558232651947,0.27263878290547905,0.39949861899335026,
 0.06454216088623044,0.3008010370813927,0.6740715020554626,0.8935322595979942,
 0.1282470855508243,0.5013189571753813,0.9194541479166043,0.3363807060000825,
 0.48089631735130145,0.15833781354382892,0.16107044043585217,
 0.38750008266847913,0.6995308533565422,0.9731479531285032,0.7479159526477167,
 0.06046833637788418,0.3122498587695016,0.7082133487628376,0.9884412842613478,
 0.38030700627467984,0.2953296940839757,0.7940002262575342,0.3297184036718094,
 0.8038346829562986,0.7816249237920285,0.7778244427825125,0.6871226518491187,
 0.7608294221511751,0.009307071799041888,0.2992425430953849,
 0.8162104361091858,0.3648431794693312,0.05338421846402008,
 0.11657334733184788,0.48184685073281674,0.8611212773650925,
 0.35886803165380265,0.8147381030348324,0.44287519594972635,
 0.20005432394270517,0.35953645289955793,0.5089006752085904,
 0.9394457134874515,0.9540445744142325,0.2786709297105414,0.825250765958503,
 0.28563527755156026,0.6668221557195442,0.852200443139983,0.14836385879861969,
 0.5656420065892042,0.7913461671513923,0.8359907427475924,0.3871589079428889,
 0.9725837176893252,0.41230886347874796,0.1962201395620231,0.8705304303902388,
 0.53199893764533,0.24284461131078583,0.8788371970489193,0.0356270046738173,
 0.6357142951054762,0.09527475076550507,0.8488673185754797,0.659525486607487,
 0.99822450906626,0.6893932697595754,0.6629585952595016,0.47607608906656207,
 0.6260759193241396,0.5997638913250676,0.41387865233716425,0.5812295819793476,
 0.6848774649750615,0.00482320179778728,0.24403275060379281,
 0.8871577721699093,0.8063560098650698,0.17488764184360217,0.7775522516906385,
 0.6372221978020276,0.7808589360450847,0.5061770904607692,0.2107741293597737,
 0.23485755142409304,0.5422703362001595,0.4482603751336284,
 0.010479311650831935,0.9107586698256381,0.7039837642197898,
 0.4971280559365234,0.9768569220652775,0.16911789159713542,0.6025051980700327,
 0.8752967927426561,0.6148557006694767,0.7639379856669665,0.4928340360086506,
 0.7461179672477303,0.318623324460235,0.9973345846873234,0.5498130674212554,
 0.7544448384362002,0.262655341955085,0.8680478906597086,0.3346223835217065,
 0.6072022255821257,0.21086251244401666,0.2530795635170813,0.5513031524435553,
 0.24475236373321396,2.1643013991423654e-4,0.6257493611462164,
 0.2060413817794331,0.42321569179076424,0.7475911034705877,0.8747335495977344,
 0.590251748809106,0.1672524402937905,0.49346762531223887,0.611479675710631,
 0.73880575080261,0.18237156150393863,0.6269883272841167,0.45067149952430663,
 0.3819779069424113,0.28319447451518354,0.24757098685788537,
 0.04876586449389353,0.1339045684890513,0.24031508066873153,
 0.9753643225540296,0.7969250340924636,0.30650525718418775,0.718746370434388,
 0.10189617432666609,0.38699566175461053,0.15258447319621604,
 0.7085557181187618,0.1972399593657005,0.10529214013287391,
 0.12078175936046576,0.4820242189083874,0.8335873637091942,0.6762187350196328,
 0.4572944913702195,0.1314177056336614,0.781513576142153,0.7580116781466143,
 0.5917772372030996,0.7168620551835527,0.6486234960876327,0.17710913748973411,
 0.5111291524898054,0.7834868944618851,0.24247358060339397,0.5203313372538951,
 0.4004582678816324,0.8631009323215827,0.0826818312608566,0.595623975109354,
 0.14446977741567735,0.010846585747929072,0.5783435681496145,
 0.967774642247061,0.7620861711545583,0.19342761109067463,0.22053108176849756,
 0.3502143262004632,0.7195500956342716,0.5897570350530332,0.45884283362526734,
 0.9345968835103802,0.016553043005868573,0.6998014320741934,
 0.1858071754764361,0.19036326185508845,0.0328440520988329,
 0.042833624196841225,0.23464592063712153,0.7665483804180594,
 0.1109523889176739,0.6971128340799126,0.5674388374409517,0.5044555191993728,
 0.18917929147997659,0.27058356729826305,0.04867870937860164,
 0.46617470474263467,0.7144961030888273,0.6645896310940921,
 0.17566237669053097,0.8406071942357375,0.990195456911918,0.40414750798088406,
 0.49181167669305514,0.18076788752434947,0.6183144654764007,
 0.26963589285098233,0.23044186130438649,0.6565212977022689,
 0.14502690179441902,0.11839974200478809,0.7188152109143353,
 0.7657107594535428,0.5744071839267428,0.7766294131105019,0.11744546359809271,
 0.06795790516621658,0.6384736009688952,0.9419965476589689,
 0.10573253802324867,0.6491668304113799,0.3049422444591342,
 0.19800946864024593,0.4747240584902417,0.13893572613751048,
 0.02503150811709509,0.05183821273953426,0.09873842446247794,
 0.595004784283811,0.7931156577953244,0.0278872816692336,0.8004954359933418,
 0.16270703513265827,0.0225549316584287,0.4211771436446914,0.8605066863442161,
 0.1348553315062323,0.0748200566314855,0.8500307943696879,0.7161960372313703,
 0.8904392394892663,0.7937142480424837,0.6811071356066722,0.27609699172685176,
 0.22264911197266035,0.2992497471822335,0.8844178476837058,0.552466338278531,
 0.6508158691971194,0.038632300153796484,0.3810479454913349,
 0.17758360247751537,0.41964064214665964,0.4866609765830713,
 0.46294072430102495,0.4198517835315696,0.9035227686423066,0.81492342590115,
 0.39563962167060085,0.9035980173316354,0.4403315821792235,
 0.38916086321463805,0.34052026140424974,0.2937016814426222,
 0.8175801114640601,0.5753773025592864,0.9649522858797372,0.8177864250741541,
 0.9656589085328908,0.9228081356710496,0.022316724623394513,
 0.1738489790445048,0.2782041387466958,0.4661494639196113,0.6011050223358027,
 0.09502339667294668,0.3239982417288936,0.2153249512261065,
 0.0010640026293167182,0.05942019775633112,0.19455011660226496,
 0.8669675772301119,0.39058347901607515,0.5542894229619814,
 0.16871338886698972,0.1152384084474104,0.8779370482239096,
 0.47340821779535947,0.7139647445308046,0.7910886142175801,0.513032381497468,
 0.5167866378908501,0.8572192642265986,0.5656977640269208,0.8968295371980669,
 0.9336772019495478,0.5514892976858259,0.4556337799798865,0.06300954015114746,
 0.7645331293010947,0.4926533980116419,0.8095944267467957,0.6678639144371391,
 0.9366066734856551,0.5526701087310812,0.705190260970427,0.454198130011489,
 0.7489516448034454,0.5724672563560264,0.26320764355653303,
 4.787305412332632e-4,0.1299455686773081,0.863223584743466,
 0.19436326198035048,0.9214899415764531,0.7176905628498005,
 0.42634632727631483,0.9588412695837082,0.29314593793601973,
 0.6920152686989995,0.5611775084965542,0.12704081393987066,0.6685389208334467,
 0.11862758656198402,0.12501864893984038,0.8450960737851079,
 0.7142341612051999,0.14157357803348392,0.9637432792988321,0.6520200459197908,
 0.38268750575837174,0.8751178273765008,0.49478205044339374,
 0.7010964544074367,0.5210654384175026,0.5321993885159377,0.8780948314833952,
 0.2679842522860464,0.3130466274112559,0.21717352307356674,
 0.09297323460438611,0.5997950832751601,0.6133809488590518,0.5954297154854076,
 0.3692150988166495,0.2728379735751555,0.8979549621937923,0.10231783735440869,
 0.8158006880361658,0.40398118231021185,0.2636077320975482,0.5990036556778568,
 0.021539150843703836,0.5217196701721131,0.06738944361360011,
 0.6255005606970446,0.971456275632576,0.25188527631554947,0.10149845667340429,
 0.693339469405667,0.9747971006355249,0.7546635002191758,0.5366776175620078,
 0.9229326944594344,0.4677880509543839,0.7115186184610888,0.47259652846662525,
 0.47318704920230203,0.6738064821831371,0.659723281392086,0.7276496313653918,
 0.20059303349662982,0.19556462850265566,0.6680429011297544,
 0.23094692564019736,0.22573789051486182,0.965060996567614,
 0.17208026812676624,0.28620623729084194,0.39010573517116975,
 0.29226674055419843,0.5940613618059725,0.6083884515718153,0.8297528503545272,
 0.10694006765494835,0.772470642568934,0.49137254600783453,0.4025357410156546,
 0.8321764498214685,0.7849530394892316,0.05891154744673932,
 0.14447924210944163,0.6147492888458755,0.1949492600211935,0.8720915887844765,
 0.2902000003999472,0.054367382177648804,0.507965970842843,0.996554590446369,
 0.5362546575060179,0.7820134922426384,0.1845296780350404,0.07168814551106584,
 0.5483891465354197,0.987548311662496,0.5707101307463129,0.3955714371521978,
 0.6705170168944646,0.012794524615834968,0.2863456397735352,
 0.5421130788667129,0.7064821044149355,0.7313652175777513,0.3387342893900872,
 0.05959536148965172,0.7966228648503902,0.6277117443563631,
 0.13084173527836562,0.05735027160829975,0.033625259115088624,
 0.6646405108364248,0.18414495006792975,0.5137212650756209,0.8594389605882274,
 0.3648843647914871,0.6861970363243997,0.48471732957649716,0.5739158177128689,
 0.5288824689472242,0.029452822147773006,0.4312960805471181,
 0.5409432068114727,0.04836000857315825,0.19451421718998874,
 0.8373001776504726,0.28665756953949906,0.2506523179943042,0.8920083296078203,
 0.572255753376449,0.46219840188587347,0.616369210146,0.7701950845258481,
 0.8426169792776976,0.6606942142241186,0.20098276227462997,
 0.30853193841460014,0.7980039182398309,0.18125315287834165,
 0.019027693722490024,0.8622263657119621,0.8730799063596378,
 0.11893851694703228,0.2679692387321031,0.6553481828737521,0.3642040281441208,
 0.845442808872241,0.4914833722457135,0.08134041945969028,0.5560906746325092,
 0.2324490219364077,0.8519711727878444,0.7086854821194333,0.971175320541108,
 0.8785998966956631,0.7020032122175305,0.054992911362392705,
 0.014538326546612024,0.18847394797660233,0.39590706207971216,
 0.9755402432864636,0.31339111340459314,0.3452864367548846,0.9900175297799171,
 0.21197095625335605,0.7074878333818074,0.2237789573957717,0.6196108486399025,
 0.8047072410734837,0.5670714985887507,0.7517357518828727,0.6153833098177284,
 0.14314497677858418,0.14304008377137079,0.6402236917961512,
 0.13773658700910385,0.4818676150543021,0.11184057567620964,
 0.9635081051139014,0.8482921593753063,0.9703179840381866,0.3046555571811417,
 0.415392324716365,0.686038324059142,0.03403655825154983,0.7927117062314492,
 0.40881663182432937,0.16994830548323014,0.6657034969028187,
 0.9705772317548662,0.8722999788679916,0.3606075944929834,0.8904080144746899,
 0.9671641165845446,0.10416547091999007,0.6848342264420437,
 0.36733239128660544,0.769170555668359,0.14435703491359797,0.4904010035319304,
 0.5113070961369977,0.930875944354128,0.4377091052982448,0.6402007563432837,
 0.8614550440164045,0.98877506130708,0.709466060098445,0.29187093709616685,
 0.060892005540667204,0.5409607121698563,0.41886393172175074,
 0.5505077835292308,0.7269213084093605,0.5707143751247598,0.6457923499231457,
 0.8950711629340153,0.797149040147266,0.35097991711479615,0.471623108381788,
 0.1361429319759503,0.28055851704050216,0.305451021380137,
 0.007507040538607956,0.6385571435892918,0.4861188629325207,
 0.6831645100785834,0.48013954116739666,0.41382096100381816,
 0.8019767416334106,0.178768806353378,0.4674770240617643,0.4391160400913703,
 0.02438260841368345,0.5964495615820737,0.23713552240299274,
 0.16004328915920518,0.21971108121305427,0.014340131152216218,
 0.9729627352698641,0.019028496951355667,0.5640062113859337,0.822364841230826,
 0.9708576824014246,0.3974766837497371,0.7816524453503078,0.9440533221253501,
 0.2543903752441199,0.30062040758679176,0.4987957071612944,0.4107554839222145,
 0.7846463300988388,0.276801338345658,0.13277159123952476,0.2394366780543813,
 0.5452304878744388,0.354783864803735,0.9355106385213106,0.5653790421277425,
 0.5381437259807358,0.9622854725882789,0.04858596521708092,
 0.08289394801782646,0.9447561067492786,0.859816012179261,0.2648126684520352,
 0.9349274559720595,0.7031917144344023,0.7302235460505536,0.04810783458405443,
 0.28614240189058315,0.864561884408026,0.9894352203426461,0.3424483856299074,
 0.7807598332260364,0.6678169222349699,0.0851987371482259,0.5841957765968835,
 0.05242359935483898,0.28700546957663486,0.3387688733816444,
 0.2554814546532913,0.9894149963687479,0.8899167261768165,0.8484734685535855,
 0.7581429721710129,0.2249770664690195,0.3125566303423978,0.5032687099928701,
 0.10993897671386388,0.8068473973642594,0.13431817290956294,0.803660512720904,
 0.5798499368903053,0.7317842460063145,0.553894314295281,0.3866140766439681,
 0.7040184426695282,0.5873484171621272,0.646503801240721,0.5908842738651245,
 0.38766381948162354,0.5953213753205596,0.1993598074362779,0.6090851990473771,
 0.7923084494251853,0.5558889742639419,0.6286680145010255,0.7935165653555976,
 0.8736968339467962,0.6528016355179397,0.3590348052137695,0.3056960870195732,
 0.08231835639923002,0.6696297618891852,0.3746962288279303,0.8716088880457619,
 0.513352165361825,0.4684104535926481,0.6218231443758933,0.5265552763017782,
 0.9326402636940649,0.21124559233214157,0.14321244930473198,
 0.8733539425399506,0.6408583168699319,0.5275992633237075,0.9558174972475805,
 0.9816931400633668,0.4198217388099179,0.631838629386403,0.2973833321397138,
 0.055087448062922095,0.46275257588688123,0.15148992770329317,
 0.984646694172492,0.15345232334545833,0.8815351831951797,0.6190063098207808,
 0.21548909371227631,0.921459882805251,0.868786107443474,0.7024840600895386,
 0.8901423600687233,0.18784239207437148,0.8215197476829817,0.6047002991120648,
 0.7335038478238998,0.7002930952981253,0.48415399757658506,0.6539670721707153,
 0.1109951919943768,0.1195870085330234,0.5117973632212801,0.6679247191274541,
 0.194362110750944,0.4430999169220091,0.7051910871547422,0.698487984286507,
 0.6802144113729629,0.38318139702888976,0.5685705248986009,
 0.008460652798752522,0.016779924994064888,0.5704358742875126,
 0.6484366453884659,0.30437435184860395,0.2090097100722459,0.5254774618980665,
 0.10542979532516017,0.5991159421839797,0.5308255359853853,0.6175373458842085,
 0.3402036689372897,0.30608990232910005,0.6437515806118141,0.9186918833493898,
 0.5981773905008805,0.06935159930442036,0.9914871264124122,
 0.22360931532134076,0.36058154153805133,0.6604641445019747,
 0.5051233629578055,0.626134362311543,0.8411676985966778,0.6415935402002673,
 0.12895129625384938,0.4267680523814452,0.7369422888378965,
 0.20289590497967025,0.25050700077698407,0.9482099842495153,0.571093761313507,
 0.11061612742079241,0.57817208788287,0.7297835267858253,0.14915718890145402,
 0.09045162938182738,0.4058668288622572,0.44160149649170144,
 0.38234340589322036,0.6984576748932689,0.5319151485384331,
 0.15135873496487673,0.2840785915541397,0.13504239987661149,
 0.19835683882567012,0.6852421745757705,0.05752017059724135,
 0.1327573783659265,0.758625235030734,0.7802711123112989,0.5977668978467462,
 0.7606821869559856,0.0706246021768181,0.2584329138260538,0.27446364263206957,
 0.7863421206616796,0.24133916231661834,0.19060674880135486,
 0.42915081753912965,0.5337473176637263,0.6191855670575489,
 0.15724977335851398,0.0035385705220754016,0.269638691677209,
 0.6291867570527772,0.5691943581300604,0.1406042977446179,0.7438098007866032,
 0.3991763206740876,0.9214852458414514,0.43450287049590963,0.8861450339151361,
 0.9794686507712633,0.42197214442663744,0.10861965188945566,
 0.14966510696232327,0.9054179722662714,0.8325754601630456,0.4798325125101719,
 0.8583604994711426,0.964009120666913,0.9341165827859933,0.5060370187781089,
 0.7738596522484662,0.39971625886717743,0.9385603498061376,0.1837274467861515,
 0.8256876067712927,0.3162440309343302,0.19036187800789517,
 0.10337903024108919,0.816114638285457,0.4987145239406202,0.34400382035869714,
 0.27399142167736157,0.2848488091806497,0.904549469609644,0.6120260097429359,
 0.011436542105736414,0.20561057384813253,0.645370885810385,
 0.5968806027437328,0.12276532542496132,0.2119415745635742,
 0.15366634428946313,0.22104509091135394,0.03254025469966548,
 0.5428903192836053,0.6618079435721361,0.9444972090828625,0.2601039281651262,
 0.08094424020189761,0.3316552759527589,0.6732774908064866,0.7963993408958512,
 0.4630789662755329,0.4058989960596864,0.49373423386107107,
 0.11024396666840586,0.28979927337748834,0.15034820185594855,
 0.9132642518926853,0.841364184075098,0.5588211542601659,0.371550925415164,
 0.507871147844981,0.6796248655433408,0.4592016117074015,0.21490797333554557,
 0.07058993913036471,0.6001514053397163,0.357021733517523,0.08036246583494,
 0.5152361554281841,0.19230730040895705,0.5051576737665631,
 0.37014725301286977,0.9781894233845521,0.17900046293414495,
 0.6928663931459467,0.4641179339304844,0.9728916340785834,0.4311174278316018,
 0.17508042693821224,0.6987588605054673,0.2907662937552685,0.502168682637568,
 0.012281620704157836,0.8313924657951709,0.7281576470710038,
 0.1895287329131613,0.5606391565163404,0.7225873727791463,0.07682666378970837,
 0.362309340903774,0.9630230715312741,0.9106986247984719,0.636857550667382,
 0.27921529100191833,0.05291921559716506,0.4433705813885458,
 0.06099743020357207,0.20796432526149045,0.18629568233029004,
 0.9717648077289269,0.7206196813210278,0.15018020234486273,0.7283750115919517,
 0.43595293949630165,0.7224254079864797,0.14876562152624206,
 0.19896739124522878,0.04156482443755527,0.15058638907761956,
 0.6025792936767259,0.9818402960038706,0.8775223510219017,0.6401727211581327,
 0.7980867046523825,0.139882468032288,0.6119746941523265,0.6818608134336165,
 0.2944800546514743,0.5843533539037576,0.5833664330541527,0.37823448578877006,
 0.13778331188335935,0.9872752635346156,0.6267423323993038,
 0.44158507356769794,0.7114304644295628,0.42398763178054877,
 0.1981193574708876,0.3562077581469498,0.5378210451251502,0.6144174188094456,
 0.35588117348368087,0.6283469973358172,0.053429090318045436,
 0.3603733379302293,0.8269196231983769,0.38237640976769904,0.151680711691066,
 0.24740598301884464,0.9270945108100418,0.6592552254948925,0.5467130843621505,
 0.24588933894375353,0.7448703685699407,0.8131248376786369,
 0.21714805251059888,0.1965200189998434,0.7567275474191235,0.4118598476760068,
 0.05106224728472353,0.8149994207054503,0.8659384525454719,
 0.29965102650997666,0.30065274449781654,0.34252229152633307,
 0.9259828489203749,0.9489925939310823,0.2231376182581396,0.8163468908278592,
 0.34726648285243944,0.2584638282546663,0.2697274025574723,0.4021772817651805,
 0.3130293269426744,0.8950097014242345,0.8399225114293042,0.32829384318860344,
 0.7262658043479217,0.7605566290633476,0.7989648021840857,0.62522719357315,
 0.42094581236932727,0.28982666211260244,0.8686112012715813,
 0.3701746826197503,0.458073324464613,0.14506721156342894,0.14647154556599318,
 0.6921473191091262,0.19346390473709996,0.42931689935789974,
 0.5111087348424574,0.4152235851406232,0.45827087751901696,0.5519785517243674,
 0.9490585635279624,0.784978550508423,0.8383565971792,0.1260197834215312,
 0.5345167050181084,0.7053162823143674,0.4574869103353649,0.11214823577486432,
 0.1603905743824432,0.5849879499232131,0.18391572875677764,0.4437834897265216,
 0.3633446469677475,0.8459331981470494,0.7448113333448243,0.3263335400435987,
 0.5635816260591376,0.029815660267702104,0.38955072708311256,
 0.7580104300944545,0.770316166907883,0.012609080446999266,0.9161814307430857,
 0.794247781270408,0.11154091259168342,0.7063377524672054,0.04396178968049236,
 0.6967126195113542,0.289163412325134,0.8373069666971494,0.6391412216467609,
 0.10793147396690161,0.06964134613759787,0.4356105498504881,
 0.7684260882002619,0.6883080909760276,0.9204793371325168,0.1599090704078081,
 0.2679721410330881,0.7136282076063291,0.30783911010242626,0.5577556581417065,
 0.7253437687808646,0.8361121687409067,0.9166306502825755,0.4978210675562251,
 0.5611202310835471,0.07209393918363305,0.1924744100757163,0.8860739080501298,
 0.0014704448050294472,0.0016738574764659786,0.9804545399856129,
 0.26651598991567216,0.43901680756012107,0.6868141064905685,
 0.6634620790378292,0.6604837940207038,0.787315612678087,0.2663979271146746,
 0.26939853097595456,0.6755211462758988,0.06457440311806173,
 0.13155876964655144,0.19067945414750864,0.12402003520333282,
 0.16013638163273025,0.7135375894930058,0.572674664943506,0.9954357265472625,
 0.7151661507323875,0.3916266677836374,0.29541447562256073,0.2817311427994419,
 0.6756980411793708,0.7714991140250995,0.12156980978838017,0.7950625844816812,
 0.0010167787035538778,0.3328305800380046,0.88077486775672,
 0.36863501060003423,0.7620158476321399,0.5910754047167276,0.2057093703069992,
 0.43322496006059596,0.2502603626259695,0.5240773360737152,0.9234589651279335,
 0.08936586626120746,0.5849392761905796,0.1710964580359171,0.7618423410541614,
 0.2303528964252226,0.6278079191470697,0.618786480447786,0.8108162959662811,
 0.12411328231762542,0.14585354913827586,0.5351555356366872,
 0.4484437198382394,0.31506080000032033,0.23286360149994034,
 0.3274177449672333,0.5660102710573817,0.38068225807671263,0.945808702651865,
 0.08258395024651533,0.025760123456626927,0.7123676554365009,
 0.18351101169926154,0.5363233729709835,0.666535878995408,0.3098264487853791,
 0.9320126773016713,0.6409604492120459,0.33647499935752867,
 0.14647962993996033,0.8355003682165217,0.04433457270358565,
 0.9249712073778801,0.8916267190482029,0.7030078552341837,0.9417353740978183,
 0.16087290653280362,0.4097984193538662,0.4219684359058701,0.3240101549357576,
 0.31933920182423214,0.6987425060090809,0.1983745291665724,0.5130782680030879,
 0.977804744603218,0.0664576526879268,0.9004620320676261,0.4409845885431838,
 0.4331494147899724,0.4762804386744781,0.19426893519127486,
 0.12709612421490557,0.9589684204718768,0.7740720561917096,0.8970636518193476,
 0.5851357873727246,0.19848290073938668,0.8310214365095631,0.6058985761961809,
 0.5724717977139596,0.5849946487446599,0.8553393224590327,0.05896451044462958,
 0.7407451889663758,0.48951543261500463,0.5356458423778978,
 0.23215482281110955,0.5160122743688379,0.20672484339925423,
 0.6599687628497375,0.7168557384188969,0.055746111942160836,
 0.44709115193777205,0.46002515441622205,0.22179653428925594,
 0.6847686344062394,0.8307790093676936,0.14701533670648803,0.8587947786049461,
 0.7311695929830501,0.8903624053570762,0.42330915814453185,0.6755149559618935,
 0.7720744413468057,0.2704947136416128,0.8832018132039956,0.12911386980457928,
 0.352208113874365,0.6927838951426862,0.8673902144287668,0.32355118296084306,
 0.44967109728436816,0.46850181433082394,0.8939272566613583,
 0.14281549471977606,0.09709378765165688,0.2224933926867495,
 0.2873480324673351,0.08682345134967528,0.9107079899148894,
 0.05333083772973013,0.44777358214450025,0.7051300047271783,
 0.9249868479146506,0.09133165973826851,0.5346728431547492,0.793088446283304,
 0.7183841500937826,0.8043365858907904,0.7886310665359031,0.42234851213004776,
 0.23307417063542624,0.7582019752531844,0.41926809731474135,
 0.9416341394198772,0.11958169268781338,0.2973029300927845,0.6018214425181558,
 0.9097914492769295,0.7827474114342972,0.6569073747277305,0.32953217966828485,
 0.08896542130288942,0.1075807399965063,0.10553649124630038,
 0.0914026551708852,0.7230094674979572,0.039581877130345,0.9966440859673847,
 0.13090738061224116,0.0205823475846445,0.5669848890263676,0.1801524080744722,
 0.45942573476848814,0.4676217829252972,0.2826092219793034,0.5199307246784007,
 0.8979971934064923,0.8372646166809143,0.73347241519,0.7064429815869608,
 0.09827794428316394,0.767731024463552,0.4582957863303496,0.5285714020710673,
 0.8681570086467862,0.40375558409185464,0.04443853676928944,
 0.7133228081168906,0.5314049098640746,0.7707847180453884,0.05910521548939651,
 0.7324852950589996,0.696080383893325,0.07965685329034367,0.37032533816232427,
 0.5764548782730632,0.2020501446465065,0.5214973522062507,0.5039224067532501,
 0.4304565616019662,0.8639646224524989,0.739315892586003,0.24347075398942164,
 0.40620366201193614,0.5169911007024249,0.20673557362882555,
 0.24739080327933383,0.7889221126823166,0.7622117925599882,0.3188746752229944,
 0.8499456164653899,0.4564680828196569,0.7522850122728658,0.5736963028864038,
 0.8752459553047044,0.681225893129233,0.526196038079226,0.7195128821296053,
 0.11830400215093195,0.6416545549059395,0.38909625734123066,
 0.4338648360745888,0.4481184072375066,0.6765272423606866,0.5758315218413346,
 0.40215315203947344,0.9581793074231575,0.5752488099039327,0.8320154998671019,
 0.7835530107096342,0.6647482967837974,0.6344596969300162,0.09916655811299602,
 0.24498710509373145,0.5285370484662545,0.5493112081786675,0.604526168638261,
 0.13995995794455673,0.9640111624200991,0.5126874124755824,
 0.40460122835673096,0.6315083805675857,0.7541564418144531,
 0.047189704655535514,0.6883136857234288,0.029888718826750482,
 0.5064201174355472,0.47752347337126366,0.08612196324092591,
 0.22566988498716445,0.6275202960692287,0.42823903217960047,
 0.9961823864834424,0.10998291660679282,0.06631890142359331,
 0.5232541992439728,0.8610112158512153,0.3254840435148363,0.4452452572527592,
 0.12431412217560356,0.8856724928079248,0.5283719330795353,
 0.0023819224095918967,0.3649746384618491,0.6678200088933907,
 0.7089190214132839,0.44091210883301724,0.9920864782683931,0.4327014973958282,
 0.7424552670359064,0.035273047312525785,0.4714710267993206,
 0.38586734003136847,0.5276541996418264,0.28165067898108975,
 0.7336660410982195,0.17101512141552822,0.12365219658323334,
 0.10954294042877455,0.5538283941425954,0.955366011145602,0.25950070899534505,
 0.22321637855658993,0.4157320462143844,0.22131630561094862,
 0.03050221066793679,0.7735930658945331,0.21407975256982326,
 0.8847522448969823,0.699396921673775,0.24319433908448984,0.10757763175593937,
 0.7750412202323943,0.6648377374973912,0.49924236089894847,
 0.002489635103273602,0.08183938986811234,0.5609308052049077,
 0.7353214637671698,0.34865125415786336,0.576132001802496,0.6774512582045775,
 0.29171106846794514,0.6835451262700084,0.6419845216158334,
 0.21703078136175014,0.09752293811685075,0.949282161981107,0.1782805673318676,
 0.5932619223862526,0.18983012220625584,0.134645793635213,0.7227258090386812,
 0.5502394138709722,0.7252155080270427,0.31355484050952487,
 0.05793296541238391,0.6795899727196377,0.8727978412975719,0.7226942720846787,
 0.030831138241730227,0.6968509178982389,0.30808934826615275,
 0.5452432724570284,0.38536808633929864,0.10996618585854412,0.844512620783848,
 0.07647480612688451,0.4727505125870415,0.7697471133109829,
 0.19677696724395055,0.9501797961614354,0.8753952567255721,0.3168462952513639,
 0.3306021185336113,0.18144394686098764,0.6792106149200483,0.9283197313078546,
 0.5460641459740079,0.4329984404586238,0.6822259286336058,0.4063921837722273,
 0.37399477269506853,0.7202531103361662,0.286206562644032,0.2860544023656699,
 0.48120286511756893,0.17363554052311891,0.724028711785107,
 0.37044658712214895,0.4692746668031793,0.2934859047554198,0.3504767796249222,
 0.0573639434619424,0.9476032845776758,0.3342025118720502,0.8604454366815598,
 0.7639061085758598,0.20399422808645284,0.6357092305766305,
 0.45142797779668764,0.1611527891448148,0.9385765180884295,0.1438797600705949,
 0.8862912458092911,0.39672577667674014,0.2021197601672653,
 0.17143143647203618,0.12183438759548781,0.569403317196961,0.8180202470141503,
 0.05262228078026654,0.9290035693450447,0.37479974673712835,
 0.2747561160221066,0.6171259965247907,0.45041702166814623,0.3823082547983391,
 0.6209213106605553,0.35897061601715463,0.48150663366001767,
 0.34913820000158935,0.45081412693273815,0.2147714678667776,0.273998427521452,
 0.10986971971473092,0.8445766420108055,0.4059437191744102,0.2258992869467641,
 0.22138573928421346,0.07031059281833463,0.7992580198733963,
 0.5819619381602142,0.3175378934646158,0.30950224512416724,
 0.47823363220173587,0.5464826869536304,0.6853332366793455,
 0.20795763715745763,0.19001778933955882,0.08900251266594761,
 0.9742137328122878,0.5829830473804892,0.649542475070716,0.1999433278621503,
 0.6740698471418622,0.590317420583246,0.6972509076845451,0.2951369867249102,
 0.9642809021841969,0.6716294967464325,0.5445437085054228,0.08720945759117604,
 0.5389826718057095,0.10196388027124437,0.7586919620444309,0.5546603884179717,
 0.10743085624807136,0.6061355966748468,0.17500398436129327,
 0.4785152052414434,0.44954225949870774,0.8865952628996253,
 0.16944054514831752,0.15468118815281395,0.9562784953799466,
 0.8124989088468407,0.6112993183687943,0.38303597511629717,0.8836078019517907,
 0.7305066407669214,0.905897892436301,0.162282634179437,0.5386930332183621,
 0.9365312615588347,0.23057211436920888,0.20829817755897595,0.227527252446861,
 0.8248544307785499,0.7621983679439378,0.6857326857797439,0.2371313698288472,
 0.491268094196354,0.3402064311509482,0.01870958990264826,0.15146814710643108,
 0.9040308512948847,0.6533952223816327,0.508470040331156,
 2.9580990335109636e-4,0.26077984334572757,0.2144520790433464,
 0.7030727757343335,0.9276602983731251,0.36500585238088257,0.8896094654625214,
 0.2879822373411972,0.7146765566209103,0.508334180539669,0.39222509637280867,
 0.1662619191925343,0.40979932516615225,0.09297683905277465,
 0.44390130811184747,0.047152044066514076,0.035452961314489384,
 0.3441151451158043,0.04159555403476067,0.7445288057349835,0.7161282215817628,
 0.3343836863097499,0.8713604530106936,0.7966707538375837,0.23060659347806278,
 0.024429873162864713,0.895669194584702,0.3584141784690933,0.143823902446939,
 0.38333150714961817,0.2088382332222828,0.3559262597270224,
 0.04582699141862123,0.04223114608555867,0.1133758009294682,
 0.6240487787431479,0.5641306460596196,0.8885996448845408,0.45151889349802876,
 0.8840336565751336,0.8134914268705015,0.827384869408215,0.2796824882482385,
 0.5556872545804724,0.6872090408492789,0.6530987013731006,0.6058433804251192,
 0.38583045747209777,0.21852478013232923,0.5140691626854877,
 0.6690647254753488,0.04124666375723063,0.8632805955688451,0.1824743648538426,
 0.8179039906495005,0.735617393952342,0.7137950290927951,0.7083915619493263,
 0.38318708316570294,0.14412379362400984,0.16315134918751928,
 0.08134392266152113,0.09907744340779001,0.24625721118098576,
 0.22869602398818478,0.5114284999654339,0.6961590449812405,
 0.36088870529707706,0.39028310712334036,0.7424113315791488,
 0.2577960465172109,0.6029256716357466,0.22414799200708357,
 0.31197588453136627,0.2522909477005665,0.5937682752208955,0.6642838279903494,
 0.7321201944774907,0.9660785552980617,0.2288763014443499,0.04885252224860215,
 0.8688460545588242,0.5898791268436301,0.696535166904345,0.873656373853862,
 0.8821091872145539,0.5274925344702899,0.36240352179235547,0.6338239657635422,
 0.6219816278750662,0.04325369182215455,0.6512866074233029,0.548799005473789,
 0.926954567871958,0.04284530012635901,0.7888288771296623,0.5987082760789441,
 0.11009273089795335,0.8189135176331417,0.08808496089657858,
 0.6783814065219673,0.9544401643104434,0.2783166147009497,0.12427644031777785,
 0.44347665965942873,0.6033543346053247,0.6360928749232357,0.1438655187175688,
 0.6899063356527724,0.9167295035528423,0.4071212101404159,0.9769825885597027,
 0.8850145732055327,0.01581193060326902,0.5491887052388551,0.388920499684124,
 0.31942424796138513,0.015511538155824178,0.861402233206182,
 0.22375440438999805,0.6020298905201672,0.011863861996947,0.7882209942283385,
 0.4974999096901789,0.5486346105132668,0.1819370161613777,0.7303952115746642,
 0.40118637422465586,0.40275469476241466,0.2787702493294357,
 0.4323085216579944,0.1594704985783506,0.3690567253023911,0.4819928264521006,
 0.8968062078709202,0.018892431546741806,0.215974715190133,
 0.059528269593371874,0.5345144704500704,0.8621946789131952,0.561345059867029,
 0.7143150249801582,0.5777316993116562,0.10321875473924513,
 0.28943749349096315,0.8373420717318175,0.2844459973582574,
 0.34410420641506656,0.2437263922191224,0.39768199221504597,
 0.2564853192160852,0.9277689020062131,0.7183156095922645,0.8603969278379657,
 0.1285155109114302,0.16938288726726514,0.29391764723272507,
 0.9281990297140822,0.9401319545280067,0.34162052308526125,
 0.41628483268805627,0.7049400410500897,0.8820688482513815,
 0.44641269814789386,0.34695721479633745,0.9639465079454811,
 0.7209748597515402,0.131561834815054,0.8304191481547221,0.3479128583549872,
 0.7795664930051418,0.3081633207242942,0.04214659280797961,
 0.19119466349755398,0.24058253072859803,0.9381844738820277,
 0.21792711221326244,0.7856333826056934,0.9502320917113934,
 0.12222889471690146,0.24238317604483928,0.012037501342499701,
 0.5370232731518313,0.46544091026673096,0.40183331695882196,
 0.18988271307506321,0.934805203652758,0.850399435261184,0.882914991684194,
 0.019188305235432246,0.054259134284717625,0.676511329247155,
 0.8356575537704737,0.07391771619869802,0.9355081792711724,0.2314744128868912,
 0.6553256847640018,0.12105554305353616,0.7282487259925906,0.5656888546521284,
 0.09305173455081572,0.6998605466822463,0.05461702435885174,
 0.26249974147397837,0.5442518023511727,0.27861435123191336,
 0.21981950060759425,0.7498493617436501,0.8464241137439317,0.4142784935835946,
 0.05787849445161142,0.028058120929297825,0.44972661679719383,
 0.39426601131583094,0.9174549958495322,0.8339123145414817,
 0.18274285966978687,0.10725953511177355,0.5257156504368004,
 0.20028791991349926,0.972875358088242,0.9315083670735529,0.23986813538498597,
 0.16005306449877932,0.6935093322970027,0.04508974850777803,
 0.4400711142042819,0.10935078690399813,0.5241658158651685,
 0.19051580255265366,0.8862005009481793,0.1311415141148392,
 0.49608744855862874,0.8052993649505589,0.8372649501250413,0.404444995696037,
 0.7686519902387237,0.6126589588689114,0.9617389408946556,0.2457187835574216,
 0.8056300320520109,0.8232613937555303,0.07498058221611703,
 0.48688364398688144,0.7577502895076205,6.368839606010201e-4,
 0.12030945840268303,0.7723567728449097,0.11878878871310161,
 0.9603637518095302,0.7640885968108632,0.1909247743750071,0.2138768175083846,
 0.5066317342552796,0.5023374824961179,0.7653158457328351,0.6041426831994746,
 0.27597242528047483,0.9447075356702866,0.734414120782635,0.6555265261280139,
 0.4377353749507531,0.36404714200285715,0.41352022152348766,
 0.5907151808012819,0.013150075504824943,0.6910112966486804,
 0.3716054780614413,0.4029683726667741,0.26445254743259994,0.9571608117776268,
 0.00525054441329087,0.5981262413201637,0.7624554347040334,
 0.35752875037946197,0.19286367690529937,0.40738470208996747,
 0.42963833000006846,0.8421688441780605,0.8507678037031845,
 0.13627526778454158,0.651944456178127,0.31163010446991724,0.4435899666198617,
 0.858052829903959,0.28489558585956154,0.6157083765800072,
 0.017912623233009684,0.8375024529568202,0.444872873817021,0.4348513035315015,
 0.2226534309280257,0.18440570990416627,0.40200148746240383,
 0.6271049692973572,0.2480637504751847,0.0945104247417049,0.8353854024486295,
 0.8447797876551807,0.060634040811420986,0.5695431607574647,
 0.22446238615025083,0.6479702392818253,0.7479077118946135,
 0.05189860030168092,0.23584089065828373,0.09423954583722616,
 0.6671628180534855,0.8879576882174928,0.5016682593305978,0.4396860208627231,
 0.7473234195284592,0.5127809856569245,0.11723830452167694,0.7607668032920011,
 0.07199512561402388,0.82681498813893,0.0836202475759944,0.9262132172570247,
 0.7934828273229364,0.8790578422462529,0.013159352308731331,
 0.017805245387130575,0.9084165043996992,0.29338439510184877,
 0.9869251179923798,0.6814598905665809,0.3564666622896655,0.8590738189381608,
 0.6656554515485847,0.4768436657579558,0.7772890630781172,0.3390795756417885,
 0.24959745871139505,0.5807477923838399,0.2082839248557795,0.6789060855584642,
 0.020141370119983293,0.20854305794591355,0.9824155167607795,
 0.13381321657716205,0.6761777565185768,0.5721101731080288,0.7626323191235522,
 0.46909302582858525,0.6252650878652797,0.7028122688195645,0.5478835515482439,
 0.7839462493882955,0.062471166574747494,0.21906832538409504,
 0.4202260557076781,0.8123722369733561,0.79876175984828,0.7131994095310601,
 0.4930609918001887,0.4264606314321041,0.9759561432708135,0.8822692789032859,
 0.27546016320103417,0.8860259610593925,0.4509128357133364,0.3051158623511747,
 0.8932622368139504,0.861858423106607,0.7795034956439149,0.48058857224599905,
 0.2043047778705559,0.6434316602377833,0.8572394901750048,0.16166613228189075,
 0.5044506074989209,0.27934206636390846,0.27559779344467794,
 0.19925811337825827,0.40155007769577744,0.07607645119170892,
 0.21651645694504784,0.6222639795081734,0.5627900588790307,
 0.09905276787647088,0.415230377686137,0.3897255387390196,0.11210510496474924,
 0.029440110283660337,0.14547134779528315,0.2657588841117706,
 0.22048573350740075,0.43999221280309286,0.7030239976069318,
 0.9330843921129622,0.13688028455991852,0.40356698114558054,
 0.2053399252759187,0.9842543390545075,0.8424899510705477,0.18694694203791362,
 0.7814014279495343,0.8546384354040371,0.7614068296750838,0.9151340727843786,
 0.755690312856818,0.6369958480192145,0.6843313752670483,0.026359153551045122,
 0.027421109297011692,0.4292827806774069,0.6937596556930528,
 0.8910586786129719,0.5693304414157468,0.8787383153792956,0.9999336263814667,
 0.377000216224896,0.6976299049935373,0.5957835658758984,0.4088650902538714,
 0.03973284926466014,0.8593157649627908,0.19795912532682225,
 0.48129218652190453,0.747376093724851,0.10397934897715222,0.3967084790707146,
 0.8229901928930405,0.3571048877334725,0.7446213387574597,0.19169865830976618,
 0.8580409486962173,0.3732000978202653,0.8881536192474586,0.8198484197177983,
 0.7680241946852777,0.8794042779048568,0.6378361545852527,0.44124212462160706,
 0.7339038081011847,0.6607518631244854,0.23874917278206043,0.9240385650192341,
 0.41838228115192133,0.8953139808239265,0.7573081402882872,0.7333645183242735,
 0.2390535763439693,0.20588275754080265,0.15355042580318146,
 0.5142243988442826,0.11776428322254262,0.9745033063106544,0.7598040263134344,
 0.3640482680570378,0.2972184583087465,0.3361356135886301,0.859768188918479,
 0.7310060433814379,0.5907198343996858,0.38159316606158833,0.3724626953635596,
 0.9623574945578279,0.43516159670903454,0.7340190382180204,0.7626032262409614,
 0.3352944116523394,0.5976644183107804,0.8745509350776033,0.338286119107686,
 0.34296235115909224,0.20848575508855416,0.5480346901081579,
 0.08376621416655872,0.3495504413081836,0.6482851113350685,
 0.04565134503873003,0.20895248566884816,0.10444984169157012,
 0.4193333757981019,0.9689539871719093,0.7237091423637194,0.8903998244713246,
 0.17775042600291702,0.5260979307052267,0.3909735237355365,0.5708834118271501,
 0.9984504588848562,0.05941874703394101,0.30654371056104135,
 0.8253273247653388,0.2649435566860123,0.38772743598796455,0.2872572942789198,
 0.3140954693311553,0.8010775835862521,0.40890572991707963,
 0.22587033629277464,0.4591113608441074,0.4741835296541317,0.9364025072355381,
 0.7409182247602545,0.4751609944604437,0.5919111174560515,0.541721105985944,
 0.5367796062516592,0.14331487854166614,0.9403348598874776,
 0.18863154835466567,0.17364588046623763,0.6349666610665017,
 0.7153047053008063,0.06825153467789735,0.20718180676645614,
 0.2998238607156626,0.9743210412802511,0.4074291500234827,0.7805154508009656,
 0.5514490388986028,0.44545656791421195,0.37343891293105413,
 0.5369859920226722,0.6773565863404155,0.8069064316553719,0.8149237446712068,
 0.4713564883778508,0.6645115639329088,0.27681978582136524,0.476287890562366,
 0.4338851794393037,0.08856995870956119,0.6679639256107383,
 0.14048240205593732,0.7950774785469453,0.08814218269989604,
 0.5653694972396588,0.648805920904277,0.12174690038218428,0.7455532857030361,
 0.8940272490565958,0.15828734732693572,0.14872028342468968,
 0.9441280286539397,0.21866006321115572,0.22846201417075052,
 0.6425051386177225,0.2470236858925713,0.2830138310113587,0.522299791955387,
 0.8588105367431721,0.5488608067226708,0.6161628751612169,0.12032872389717841,
 0.003509191516234633,0.458806906473856,0.2321387659180496,0.8586942316056454,
 0.9703400352197042,0.4617086582925888,0.5402393008448946,0.8817785900214326,
 0.4731711948125664,0.37457569758776144,0.7418551405065292,0.7150413157078306,
 0.998957275025925,0.28182426107997816,0.3444311706779998,0.6283065603051192,
 0.7972821124430418,0.7790550315310145,0.2855914059596758,0.2099015091010762,
 0.903751738268286,0.6276726963547882,0.9616083648265504,0.30185062550311004,
 0.799101592140865,0.32565514166380516,0.9019716825477379,0.4504887282230269,
 0.43426501006951035,0.9418269525736279,0.25112439399512454,
 0.48633518036688717,0.4910098537158136,0.2022214867336306,
 0.20011455243300402,0.4674833018134439,0.16008113633391763,
 0.7772841802879843,0.6336314073238817,0.13292733488825492,0.6522010182271399,
 0.09853340213050221,0.24015646970479432,0.3545360137921847,
 0.7671415421551147,0.9404430996777505,0.5343622438112311,0.46834120308414073,
 0.04486892260829678,0.13775979288210005,0.35056656506683204,
 0.19259266192795144,0.9288730950147256,0.2799712811330648,0.2408505824211401,
 0.81939197813315,0.9333721275408516,0.2913554743419594,0.3025796312777921,
 0.18936116061004782,0.6328705495689225,0.368038632599371,0.4937548345534415,
 0.2015288545372752,0.09030976008331448,0.7946787850777097,0.5662769363095532,
 0.3684223282942156,0.23283350198912134,0.9880336448772921,0.7910545183506051,
 0.5941163930386839,0.36961426968993794,0.1819964991326659,0.9847272109598965,
 0.5950928187160328,0.32696552324345796,0.6226684117890271,0.8135783984324461,
 0.4380859241104107,0.05374632115688571,0.7226478271445417,0.7252473865788821,
 0.9074162074229588,0.14926047456827107,0.2480774654992961,0.3589693753468888,
 0.07234436789170684,0.5936518068577181,0.20407467020065473,0.297484830872387,
 0.8665902242717171,0.18977253819629136,0.8167140509612214,
 0.15067559003693143,0.12078416438352013,0.6288824822601244,
 0.7014481226214588,0.5782487224376015,0.22849011621562054,0.5932813646485051,
 0.22038716664042435,0.3969049125780283,0.38056033921122134,
 0.01210944279440529,0.8495027090003814,0.549464950556221,0.8070719120331448,
 0.30021828591317834,0.3249947273964825,0.04360927655295144,
 0.7102723836498197,0.5081952200386185,0.7322390543792495,0.9557328457547059,
 0.8516994124466525,0.866312672454503,0.3423629488047838,0.11262276549022454,
 0.5138608716016656,0.8969986889405644,0.8352910001661455,0.8997089680906968,
 0.9883746682777155,0.07782826362893436,0.53655214322231,0.37971573396209113,
 0.9762194863063348,0.27548475878994205,0.23366760197699776,0.593966004295684,
 0.15084346556909567,0.9834811450508234,0.2777708085474714,0.7696697436329111,
 0.5048048511106841,0.38816320942360993,0.4934482765661461,0.370930083877277,
 0.5559471683514754,0.7366974458446816,0.12680205657822374,0.4975560632613255,
 0.7371243206727984,0.6198766634601451,0.9497426386425002,0.8906167632287172,
 0.6589952941033985,0.4698465590891553,0.9412040039350074,0.23153986449063213,
 0.5308129966937375,0.41753237082801,0.03478345347792744,0.01607257902917203,
 0.38033938295242886,0.19687593849120166,0.6185585421061929,
 0.9017464093964821,0.043918038652466906,0.6744657473753324,
 0.7415326505357982,0.2818842684420366,0.43852292294015904,
 0.08732544641620343,0.21454160700569858,0.23464223580211852,
 0.02886393346912708,0.41345833274019583,0.3356315392380377,
 0.4892851771200638,0.13780079464663098,0.9333293063075857,
 0.036849256802047425,0.2687211349973586,0.9194929352787758,
 0.7808725155242678,0.07194553533032366,0.3382110892013739,
 0.20161848948361016,0.7151592837941071,0.2517175995814501,0.7074418158338995,
 0.3658103702072959,0.0356548552049516,0.21353110038571366,
 0.006634586262401765,0.2402237246816452,0.7390087285103222,
 0.1703271476800614,0.7742219670818508,0.2168356518003608,0.28675939524036675,
 0.20619462956259005,0.17708350638493076,0.943601948917346,0.7057148849748307,
 0.06284264181224297,0.6251194004944534,0.29826261327673964,
 0.6014727052109428,0.8104444559650651,0.15263734166163123,0.6584162854852901,
 0.013777933702466605,0.9342795158906139,0.2618603453046735,
 0.9984432653213928,0.31148188524960063,0.5437086305814395,
 0.30936963192439537,0.5058231217514322,0.027091914226171188,
 0.32092712318236827,0.3534317938749384,0.34095914185268583,
 0.3125505211534193,0.7975809525281133,0.11328872636404397,0.4487371267958572,
 0.008869641460610422,0.18160226769199206,0.13807402411312208,
 0.6208123641918233,0.31883273177738214,0.062269536184574426,
 0.23967915090060876,0.5994780963559403,0.7868946989166293,0.1738031096943563,
 0.4954232940518214,0.19068409360055705,0.5155239678321214,
 0.46649251534325675,0.16182262094591748,0.7779206938818937,
 0.1235237070518227,0.2758016203550162,0.03424615248971952,0.8058126456666246,
 0.00604415658773072,0.7610674560786883,0.8881099478045917,0.5650793943534277,
 0.2626146644275402,0.9914944236017825,0.5291489793938687,0.789424866028523,
 0.5568731518596428,0.34392376431027993,0.8332088247582525,0.1440382982260442,
 0.9044703755019201,0.684438853253912,0.6764737171991517,0.7395234121036905,
 0.7643482991157133,0.8463958060922147,0.7428557326748415,
 0.029220472250275797,0.43297879539269335,0.8481188903598191,
 0.030156947819175883,0.03607517194860377,0.027326395097462375,
 0.8690540940727585,0.6392119738586517,0.36442881751932055,
 0.19344803546733114,0.8637998300727454,0.7431650367963414,
 0.14242343964517012,0.7862376750901916,0.9746047006026906,0.3738879985304385,
 0.1619166191372069,0.25092431141829474,0.9304811013014639,0.2731432951413373,
 0.7193222824691093,0.6092485896187433,0.7675874881992966,0.8071803995216651,
 0.5318624845280103,0.9799206215589122,0.23437002482645208,
 0.035697584884386235,0.5549724713207107,0.8774308098458696,
 0.4944067624113395,0.8418225442150001,0.1466960285611777,0.18194422215944805,
 0.4597756183413899,0.9977327365440511,0.3554745161041284,0.8538617957694778,
 0.8710095269179885,0.11903611584799223,0.9610176557036151,0.7955242590367796,
 0.41041786216033216,0.9528999235197871,0.5769614693195769,
 0.09059335448841277,0.1685971917576533,0.6776651782732088,0.8636723119361587,
 0.7562354622045597,0.13133942836621615,0.7662614358611406,0.7248025886020779,
 0.6597484463592498,0.814721134919114,0.3707056865744983,0.46407144633649633,
 0.9319342326168408,0.5649161720668102,0.883114078039955,0.3817387391298812,
 0.06961760730624866,0.560381148932544,0.6864831427394453,0.8295679734799446,
 0.27517092421171163,0.31001451887995,0.19568236968328934,0.8544848849637605,
 0.5600369544129785,0.6459978760310445,0.9492778715098298,0.03725316938505019,
 0.41536728568047965,0.19130435229951925,0.12462309720987319,
 0.20691111027106968,0.7559952232131528,0.5299820789149348,0.815995647700479,
 0.30531036080139684,0.9671589752228821,0.07002672321824,0.3641655936286998,
 0.5406337900131182,0.7963856064934457,0.21403330553818578,0.8133317789209453,
 0.27969217587481254,0.03431129301443547,0.0310893496104111,
 0.10508468592798403,0.6479118726233468,0.9749370023909427,
 0.13266399061251533,0.8351688075229341,0.8395890466331186,
 0.40866564399246486,0.2954859873470286,0.6265842811891534,0.870852696738192,
 0.6361044523870018,0.03274876330043264,0.45156720877810863,0.904731606438451,
 0.1067043550742669,0.8816032379638327,0.9809520517198103,0.8142392381063578,
 0.5219470021940351,0.7750481324292122,0.30180797578723595,0.6985088157546223,
 0.7437652472821403,0.3951092174522398,0.9531876438998104,0.7604650899349666,
 0.7258105910603005,0.06102554536518301,0.1429600058194087,0.3164173154995844,
 0.44057962060175493,0.38325352904400145,0.3085569175310481,
 0.2121851181162021,0.5105000827098427,0.8241915862973089,0.11417102508149346,
 0.25540005805821486,0.3012744027502108,0.821317956438867,0.7445723639168316,
 0.4924129026195585,0.7108966640196299,0.2535860876904601,0.19915875664661797,
 0.09410101887180233,0.3925743798690967,0.9724569703040814,
 0.22042992772079473,0.9152552353270641,0.73857116173953,0.5597518262923018,
 0.1530184814374963,0.9665774629740692,0.8223236229478873,0.22066959893171845,
 0.7209784647088772,0.8863264333839126,0.1330759619495715,0.21331217035173577,
 0.10614001867153255,0.686762424519403,0.9189965706892205,0.3123611409248256,
 0.21059377904546428,0.5463290813748305,0.9132748922214081,0.8797619017744996,
 0.3380060655785101,0.5584831610125912,0.8214243058426337,0.19150416729648756,
 0.2888091632763108,0.9577401883048203,0.404467530409684,0.31514293970454066,
 0.5011341333770067,0.3854539495587692,0.10984414271719511,0.6022662401358917,
 0.6316089882706157,0.15855220481442744,0.9038197276130063,0.8496707339326643,
 0.2577893295560566,0.3950764674510612,0.911892171993165,0.7685126292147133,
 0.5144973676269878,0.9094540618441823,0.9358080839049381,0.39714274337241817,
 0.38812939147758163,0.477593586490057,0.43560279650178535,0.1463312161824749,
 0.021495838785385524,0.6312443010813131,0.9284065357125121,
 0.7523441850046231,0.6995573622205299,0.1892836543041807,0.8365274956271781,
 0.9055655917870604,0.371164560936734,0.500280440635345,0.7619637406761612,
 0.4330805594135748,0.5934278854645546,0.4364165946144868,0.10543982877129232,
 0.021700012386359102,0.045773977637498464,0.6379114302722482,
 0.1455456610364384,0.10088326398972913,0.9045427247858281,0.7509079305893945,
 0.24739237382465262,0.2597885017691157,0.7808281598977008,0.7642231473978408,
 0.7255234142678335,0.2603142291976934,0.11654595937716006,0.9413257788556684,
 0.9942471387194445,0.035967387748557345,0.4846555009460497,
 0.5657461950811971,0.7712299613110681,0.5149104080053254,0.6125200838732671,
 0.6340281777374143,0.6157702669444237,0.9805057455547613,0.6539386619451963,
 0.3639716436609295,0.5136670667234596,0.4373650571156287,0.7333019100245521,
 0.7477428682107404,0.5431265211351568,0.5335225003021431,0.2762085871799065,
 0.6869333653934497,0.15232813619838048,0.6647300571956749,0.9282013910881364,
 0.26372635136765477,0.6554062177899304,0.24120054428711124,
 0.4236404029525489,0.9055646522089436,0.06778736448148426,
 0.13418048301248053,0.903252790781822,0.7446178682278255,0.21255453833073845,
 0.25119631386459496,0.2740072041856604,0.17069708701832575,
 0.6813682808231708,0.7717741835421748,0.18318528103899234,0.6023010264966123,
 0.17649557303969232,0.0057003359634907325,0.5360716651215163,
 0.367223111432769,0.5506664354036662,0.6070939518977163,0.45134294576153833,
 0.7408927996721733,0.4601843782197832,0.34868619079901464,0.823471287421444,
 0.02128679341049544,0.08153827619081166,0.08120281990529443,
 0.7848589047100297,0.22775045609263822,0.8654872230071234,0.5779452934167243,
 0.5651891654337414,0.7697374059210952,0.7381163805131763,0.23641220839361843,
 0.2824042094431044,0.11856823909394953,0.3500118296294089,
 0.20250570358091013,0.3818582435855785,0.11165270410357275,0.554124975426729,
 0.7603455475366191,0.15162379837143858,0.5860635808748792,0.8454618347108142,
 0.664062430705758,0.1713834139464786,0.4804812766406147,0.41211370104871803,
 0.3098248809586136,0.7915302172185494,0.1853211193332719,0.320576546622365,
 0.43775706716865515,0.12832489010855797,0.5391527292669721,
 0.8848630201632501,0.2675362949816217,0.19548767374316645,
 0.07424011896672655,0.5012406200191045,0.055257315685932396,
 0.2640211569375157,0.9831392850205832,0.6763596449132729,0.5555668066470798,
 0.8455229515799872,0.1169000146985173,0.9913432084718858,0.18908706886900895,
 0.016192373286595174,0.4040379697659854,0.045733952333423034,
 0.5229872191884114,0.7605103923852579,0.08374156111974262,0.6465255345083384,
 0.2942761315048248,0.08819407456405715,0.13825333397491857,
 0.14604534225472954,0.7719277772380284,0.6755927141088568,0.4664418298183358,
 0.27219050932375377,0.6893763651225572,0.5641287470741381,
 0.34828268098971726,0.33613750600884607,0.9110554983172218,
 0.0721022699957593,0.7297187427848433,0.7988159425460442,0.41504025443401116,
 0.3138999861032803,0.8564004690156022,0.5347072150981336,0.02262897395175356,
 0.3912842298107605,0.8776936554619319,0.9464626310968526,0.3651573451459573,
 0.9012992327303707,0.6367740964530461,0.7366509375896342,0.21687048146684673,
 0.7083447638881158,0.14441293902762897,0.7325816020799356,0.4904391758722072,
 0.9574544110768218,0.448003977492464,0.6296102939658873,0.4567665845173787,
 0.06495139335017808,0.049793874187556764,0.869937733856618,
 0.9173011538050324,0.042107945953045345,0.2917000549404777,
 0.5990343751027238,0.6453050737333729,0.4395637786952201,0.1963560391232384,
 0.42146388273219393,0.6064374259468108,0.0958723093014463,
 0.26486835268154674,0.661053385959321,0.10981463353527554,0.5256085920366462,
 0.26366731965896095,0.4129157695865242,0.12440148809418472,
 0.4592562747573259,0.4658136572776541,0.9273223031482036,0.5969116578386192,
 0.3178056164261637,0.4478034873458918,0.8336270377179056,0.5096641807345621,
 0.7964290255329765,0.8893548866838098,0.16615824509085597,
 0.23092559868206153,0.06355666869692711,0.3131116135466514,
 0.2768432969961687,0.9573805161990475,0.33304991859628597,
 0.05991788035421752,0.18735282366095163,0.13346307990792872,
 0.6272145411109389,0.7492341197621964,0.2053881587691918,0.32550331470197613,
 0.009713542965713251,0.3184883217836806,0.10653951248350269,
 0.4177989784710434,0.33396548415213856,0.7227479437752801,0.4804412430257504,
 0.05303593738210821,0.4029947138286082,0.12364178836525663,
 0.3734134306809058,0.38616963582097574,0.16249718130967739,
 0.6822125154483518,0.6948373287501356,0.9338315540634041,0.29099801922880775,
 0.30618342249541963,0.9747736953941635,0.03728164620063845,
 0.42135739748225587,0.8074654670732776,0.9403663939470944,0.8561732974731601,
 0.23335727566746522,0.873638393079083,0.004340515880208895,
 0.9444278432831484,0.44806235638337444,0.12947567877397725,
 0.5014979234248667,0.3521237618665789,0.5557098320563536,0.8954731672704038,
 0.7486148377262006,0.45297155246353993,0.5463215922970772,0.6025027773041413,
 0.20203735228159447,0.4507238070150533,0.3177739870072691,0.9941941835625148,
 0.9189357001881843,0.9552313914394541,0.5165270990338818,0.2835266210942332,
 0.6995471599008446,0.72609277215082,0.11299129591941703,0.2517488277047417,
 0.6768107210273532,0.7506242025309557,0.22564628342266635,
 0.09771102181781488,0.159215220937707,0.9917911857908557,0.5368941618500638,
 0.2034734644209606,0.20206462896530042,0.26968010610004445,
 0.9828322167542649,0.7382506295308991,0.17829334097544802,0.7844456551567188,
 0.29580503742699926,0.34329716472227545,0.7546362229893443,
 0.45471238719575824,0.8244889876121397,0.911733883522884,0.9882794694900437,
 0.32776534156210735,0.5832211177205143,0.6813957470236915,0.4685648716931716,
 0.28832227320725556,0.9941680989217895,0.2929072450025533,0.6186256148846505,
 0.3173106868879678,0.7596713322920722,0.1803836423777473,0.08917450571454,
 0.7322182684612542,0.8452593721789328,0.24112746982559763,0.2461294580609792,
 0.994567710229997,0.818630847090558,0.5349056512114229,0.4184445318069109,
 0.28549604723208555,0.7727319223047874,0.14753036326331215,
 0.8479569392008559,0.808300320145359,0.3991642429332586,0.44087845686352345,
 0.39307161901014076,0.5083258817977981,0.7242000602407453,
 0.33871248962105904,0.7907683538576635,0.2024337741504867,0.2526476605665541,
 0.6326841326982001,0.10392988751062338,0.5780422993663755,
 0.20914844534700072,0.4396546495813506,0.5294587755256345,
 0.08109715874369627,0.920716348034645,0.5943933143312184,0.36839379897656743,
 0.551054131362678,0.2746980564353031,0.2052818432549829,0.22519664309503695,
 0.9652210128906861,0.26785512508351794,0.7825977843156379,0.6437752596867677,
 0.06666601493496394,0.8981359096937723,0.5223718529873972,0.9098487563389555,
 0.7107424266001654,0.3800131597662665,0.6711790532469826,0.2395962566050135,
 0.8623238518329019,0.43197428808109695,0.18599363592327145,
 0.5324445701231351,0.09915219273285436,0.6462552956545667,0.6052054589848258,
 0.38609457373130485,0.531096160944189,0.5417714622335207,0.9402281795477861,
 0.44964682873991735,0.17618607688645838,0.18864430174367697,
 0.5075513818980785,0.1563693835882114,0.1733276519559882,0.6276399604936345,
 0.8360633393695303,0.4417723708852107,0.19413077293543246,
 0.47162811995164233,0.6278479950192153,0.8346882353623712,0.7167883802774702,
 0.91351625521193,0.2615610311552241,0.3315131718108911,0.5460290203054641,
 0.56750063134654,0.8540057637962888,0.9915563291550694,0.18146294511226557,
 0.4922509546141378,0.24729286087474445,0.8103393368824177,0.7664327427682174,
 0.0728624182533989,0.20965525580877453,0.5283880253069171,0.5016625742086838,
 0.760285117600488,0.09374562266056374,0.6535997573811019,0.5879933547671236,
 0.6192535915775013,0.3660174761073144,0.6303395742124058,0.765503042587848,
 0.9803391146303742,0.8339095577412136,0.8226065544405231,0.9549330872001052,
 0.9803781898275298,0.33640335378958497,0.4042994388394281,0.693230912119382,
 0.2192559551769213,0.7074498720156747,0.6634638488525819,0.04421485780959755,
 0.9762623623917038,0.09500083650936997,0.37747356065111537,
 0.03447114723477274,0.01244098178191022,0.6636872853891688,
 0.32290798089492245,0.7332566473829772,0.14306277257689348,
 0.27414631374752296,0.3160176728672325,0.7984566279668994,0.6157163137484823,
 0.9818648293827831,0.18256628570687616,0.7671597117147408,0.1641926211873589,
 0.20485623070507242,0.5742518237238415,0.2779259779186709,
 0.42855312934616896,0.4461771537311182,0.6186358860637509,0.6535779520021401,
 0.09308868364075207,0.7107568072336586,0.02688271811277032,
 0.09480850766083582,0.8763735736440764,0.20153738004929034,
 0.6922409011529048,0.2932052642726566,0.6745305785893337,0.7807853378121198,
 0.8681479566338441,0.6924235400699585,0.7734990963682774,0.7216370348506826,
 0.5609174457928408,0.5867583089377075,0.7465129508000645,0.7590916615802787,
 0.5325333520308255,0.683259182723595,0.2285811334628287,0.9995046385089323,
 0.7491969559497077,0.28727776782861814,0.40815892799433506,
 0.8158758252605317,0.2346734224017435,0.670492951929335,0.3131956743366061,
 0.749847485870037,0.11460428273432088,0.7923186062182205,0.2360104506732137,
 0.04505305449414598,0.3930218234765517,0.6679484084506241,0.7138234400459634,
 0.3484789488625005,0.2748379521478408,0.9169337581161486,0.38989774400064825,
 0.9547518445159738,0.7038897111136054,0.2672564986932032,0.42449248347498225,
 0.4497522778315248,0.612510143241625,0.8941730639578687,0.6624880510737725,
 0.39709399692132397,0.10605343301125414,0.6323456954731493,
 0.9699642437198198,0.7499190025551203,0.16007187098365427,
 0.17707509432962032,0.24996265111133142,0.8184775232676114,
 0.8456143451788155,0.9396146530411108,0.0719152444407638,0.898222043179402,
 0.9716210776042844,0.17816207662905714,0.0526863598917634,
 0.39122306364138837,0.9844774832692387,0.3248814842248149,
 0.31480150473651036,0.7972130944678023,0.6517150713944901,0.9137835892056804,
 0.5191341778746636,0.9781618155383727,0.9448705100859277,0.20400946116493124,
 0.40955176036421326,0.6791456449718348,0.41157843246515136,
 0.04558447746049554,0.19066381257528298,0.7069511128520829,
 0.9631825495332759,0.5206769144553661,0.4036647766442627,0.647713855333096,
 0.3762397056953395,0.6271542378961327,0.4147113047584734,0.5338430507378135,
 0.9637087648905179,0.9261362678420023,0.8056822382473849,0.4947204167851439,
 0.448535705981594,0.21446823214334743,0.9566763653513914,0.8370978880807156,
 0.9923291541041872,0.5920024391339166,0.21355650942881854,
 0.033790090501917946,0.4044589984287743,0.10501739318822834,
 0.2389958386348391,0.5420193625553585,0.572896404966126,0.6411168678761193,
 0.3929162480709746,0.8486443257017947,0.8040465801850734,0.8007774721196917,
 0.8530649698590118,0.05854774954577413,0.9098311122931604,0.2019418097219272,
 0.34190469229630394,0.7581276967085838,0.496372463477989,0.8024907231284156,
 0.620255013387667,0.9654336052138753,0.20006373197005134,0.27564807469792696,
 0.25989927187618156,0.7472149093360031,0.22233190963245741,
 0.38105044288491685,0.5738791131945524,0.0770651761601785,
 0.39743837591219255,0.35052696240814885,0.042586331400657484,
 0.8842240571581315,0.9405689901595927,0.00242825834064464,
 0.28371850490940564,0.8836274378420864,0.39219733278728863,
 0.3984371989426141,0.2356990290029719,0.5821782467752518,0.5282773044869505,
 0.5733293288212022,0.07675997165076243,0.04353685395902407,0.932387039891718,
 0.1336884764771361,0.5234128599366661,0.7584179617356994,0.8799090853028886,
 0.04117213460818492,0.1254998814708087,0.8849222966516346,0.9871121848424284,
 0.34290441276986616,0.5890679731181495,0.26136823195539016,
 0.9445069566232634,0.20576387734185575,0.22454776349485317,
 0.7928036672704248,0.743110446047124,0.3093711647121776,0.7224932010451598,
 0.4276896190717179,0.49530697135414437,0.5590428498987264,0.9656541054937644,
 0.851538158469344,0.20496833286653415,0.6089922477873765,0.10758302982428192,
 0.7622989072072661,0.8188002860244215,0.6679706372733709,0.9234427928569245,
 0.9450770017183348,0.007142199522067072,0.19919051095082463,
 0.39795349665493407,0.995667551621245,0.6872472606002494,0.40923871530713496,
 0.43557886382697075,0.009976537182438738,0.8149295422282741,
 0.4264183157146376,0.2646020281865997,0.3473367591473161,0.7534808868746742,
 0.47745825901187544,0.5237859949897605,0.8157783000154439,0.6412647113579656,
 0.7280298671536773,0.14412526978248408,0.8424340733920715,0.9109781265908862,
 0.9557951851835157,0.6085094064656221,0.6570973318060964,0.8607094793854022,
 0.5636521362688119,0.7256307856169102,0.490752069736598,0.28096840722039385,
 0.076942359220601,0.01227133878613429,0.570336768951545,0.7526347109955115,
 0.2660319540565076,0.17634995108382956,0.974811255823488,0.9707240580194898,
 0.04969778953982784,0.9547110011971573,0.19682453011694356,
 0.045692006418988185,0.48282211405670006,0.33224923259331374,
 0.501003018701254,0.09647960807642242,0.3021466703903295,0.3113679312316314,
 0.1347706919191446,0.007824856353407128,0.1530301067947265,0.446695399427939,
 0.3504194479556386,0.2024755545927136,0.7169605837229959,0.6383441439148397,
 0.7734728260238003,0.5765893413167956,0.23893413765877014,
 0.48900119808326126,0.2561569665580976,0.08143957627863085,
 0.7438333284865587,0.3785629862768999,0.28908948397886525,
 0.03115486463191297,0.657380393629763,0.6340607598070299,0.8763453025776006,
 0.7425214216266813,0.5341037496314897,0.23804956125497057,0.2813768429218044,
 0.8920701587747585,0.15816088704420617,0.8170380391726859,
 0.20658219575776215,0.6701784034281723,0.1977510229806395,
 0.10171904978943691,0.6762792602324098,0.018263333164170614,
 0.03641339739107241,0.5941588842386639,0.7370960948083518,0.5767211417965492,
 0.07092914750709567,0.29359650589504915,0.5952272687713864,0.578739162676501,
 0.33481846854694575,0.8349525432337424,0.7666021014577544,0.8715355242716545,
 0.12406385598817338,0.13105854421777574,0.09339300240299386,
 0.0888743854022973,0.6681822820846425,0.6405660589521216,0.7740338900590757,
 0.4870588383839556,0.9183598869223889,0.28796720971910694,
 0.35573419535822226,0.036637371970348886,0.07157307914895128,
 0.6515286372760805,0.6786787183370184,0.7913036103419417,0.9662123823524638,
 0.7404881129555642,0.6621707905456545,0.9599312615488284,0.9531704394713081,
 0.6535934860919391,0.822985189516071,0.0337192402906642,0.12037761507307321,
 0.9169674512874346,0.6242921002651258,0.3179971181831095,0.04847689193104954,
 0.32146918467409114,0.7248493209221555,0.6912684584036348,
 0.34333666694880005,0.06673987609765664,0.705239384151125,0.2718271136205287,
 0.769672990317553,0.2945821797461151,0.5089605292712106,0.6454421085639149,
 0.7136801433920248,0.5712228460827575,0.042263981879045565,
 0.5161532176243845,0.7061965103466027,0.7043283303539427,0.9381762512947452,
 0.6137593133948249,0.6795210007418795,0.2791497957244058,0.726143901597531,
 0.9724977303796519,0.5719948559480325,0.785588938786738,0.8155449486213939,
 0.13074485349160014,0.6683301672900993,0.10498179030538446,
 0.2665321557818723,0.1907852385033615,0.22803059682338744,
 0.18409470649985438,0.3651536795403445,0.23241989319632572,
 0.04390836495995709,0.11042929347812558,0.040758279132035447,
 0.04789512802745444,0.3121093009852667,0.27013766182111065,
 0.4051493271166664,0.005962188972465032,0.9596400736695211,
 0.7186466602873384,0.7286297258103275,0.7408975543015747,
 0.010942872372271872,0.11276274282968513,0.9239272133831545,
 0.40384084735236647,0.7497308672119866,0.41682068451861976,
 0.3499702671746303,0.8549673392878647,0.2569238529745016,0.3458665173313802,
 0.3401381048017653,0.15589307451598333,0.6452006002885704,0.7484762167592608,
 0.4218009900494326,0.7159726325305593,0.7418716676646443,0.09244152312485587,
 0.337616717537943,0.899034592127506,0.6472044636888595,0.02493941850324699,
 0.30292763878889917,0.3210210631319905,0.5815206424030518,0.8010556381347123,
 0.1813546823858524,0.817327578419169,0.953333639563821,0.1276127265725232,
 0.9651304838588188,0.643776126770667,0.4526670220470007,0.3782075381240133,
 0.05465954828169994,0.9629400352532322,0.06591055745162266,
 0.8352626283586143,0.04804287373069194,0.2719542149879808,
 0.32557915565309326,0.012991022713443767,0.9593107430367617,
 0.20018365563776508,0.4866617376251434,0.005810867737557568,
 0.714258184683062,0.5949952487490163,0.30626181350398707,0.6688067055585556,
 0.5521059346227004,0.46335348308797286,0.9704887171335808,0.3419875649221047,
 0.9294461493590165,0.6207063911501227,0.43477150136615605,
 0.23365731031957504,0.6635870546613234,0.35302866871021554,
 0.9740385261396595,0.09731106517067467,0.24053073321649188,
 0.1968304623296977,0.7334835875621705,0.9872575335523979,0.40446098892880167,
 0.7559130936543773,0.9048558897930894,0.33178278469226974,0.6729257941422963,
 0.1516328803391077,0.09843988046611418,0.8691707745433098,0.3157878904171735,
 0.6645853484981168,0.8302351262535314,0.1260690074096139,0.1598245551962021,
 0.6388936781682396,0.6877444840273581,0.31786883763740514,0.8491817360426791,
 0.8629217854185987,0.5246303814361913,0.8957297313827265,0.41610395924142907,
 0.04126138629825837,0.39936792081789085,0.45535674215521404,
 0.9093956196546085,0.615740289389315,0.8081752558734145,0.47503622036390714,
 0.043047728779318395,0.7803685455602339,0.8006120707254782,
 0.5398552506413238,0.3280409991156856,0.9681615040582674,0.6256960283354304,
 0.8791124902278197,0.07174645942498847,0.2378996509599215,0.6948351903740541,
 0.9861403333950239,0.3058720779432058,0.6406431438615017,0.4428923010247632,
 0.8858208619255213,0.8507079978844401,0.15984814889029353,0.5040275811224714,
 0.7113859822445378,0.9229103727402547,0.9300115558436083,0.9920715238992541,
 0.33447426970496485,0.3590209949829408,0.9223809364464073,0.1429318677702811,
 0.36903523403564986,0.7038132440760383,0.596030198639258,0.6687289174204204,
 0.9807467016726134,0.6386317590613104,0.03829181970677642,0.061673460871978,
 0.04236692989919277,0.9653245589403168,0.47642353177740404,0.747699115124561,
 0.07963305512305963,0.40796839116808437,0.22606479856301442,
 0.8007355847812165,0.9292610081830539,0.6097661270403935,
 0.057872491550849836,0.3697572127689335,0.5587776732894273,
 0.26143359637857266,0.8444301786203084,0.13548239042153898,
 0.6961188782400547,0.8927988438073267,0.6195747838858349,0.9680573793122348,
 0.687022465423039,0.7933830934414521,0.22176346648366563,0.7169129242973692,
 0.12806130268794202,0.5479540877155208,0.9976766428574791,0.8453129866612019,
 0.4116410072262835,0.23235472364935372,0.6422860994981117,
 0.31645180577104837,0.5606489993523435,0.717994828676708,0.7527462932510116,
 0.9015562712698583,0.6823573164552617,0.4169283438939849,0.5922660825349961,
 0.8905218961261967,0.2518014660822938,0.345543978347577,0.6312789200985515,
 0.46927945454333797,0.9492374187817584,0.3035322965936298,0.5360717843573678,
 0.6552155114986025,0.42948031304316847,0.6870960367321359,0.6218015537165432,
 0.4550144872948205,0.3237799814146344,0.7036842658720026,0.5987504496301181,
 0.03491332336894537,0.6541833688735055,0.7528097099647126,0.9820266531242241,
 0.7606107904434516,0.9063752001575821,0.3974070556969026,0.6603324059435086,
 0.4018917906030116,0.4122392502423563,0.8296371438094813,0.6919381276362282,
 0.9027754593342875,0.18007912134427906,0.264522805256148,0.05788587857751171,
 0.5118293424044148,0.6851027294894882,0.7421868571612817,0.7138566361799916,
 0.09330319096020945,0.6272123281822739,0.055018310626629985,
 0.09092081348207637,0.5679448195428523,0.727241929246644,0.5231520574768254,
 0.021418238404272483,0.6877572424842169,0.1291437083974436,
 0.0362380271224948,0.10389680029600878,0.23511859936495805,
 0.6720035643813441,0.9659635388380966,0.24429778708047334,0.6388082999487239,
 0.8043100826310351,0.6285211897536398,0.3569052169642477,0.49928813049675913,
 0.12401215375874441,0.7543243508315598,0.27857130678511854,
 0.5160061424243632,0.7799228163942972,0.28966837630932085,
 0.16869586137064008,0.4261268381952129,0.7117804528993273,0.556419806154644,
 0.7655724293290388,0.8896628260451015,0.3642952138409852,0.7955487661287606,
 0.7088654662943153,0.004366300665234135,0.9128873281418324,
 0.9910515240900261,0.5088382542796561,0.8359960770790094,0.8258292569174714,
 0.39274888826795984,0.036907620305115074,0.518136829445528,
 0.20288398730237167,0.3729357151975048,0.5752601612429803,0.5776796544019622,
 0.47237600191912865,0.2521814538615479,0.8552873012375927,
 0.36870925990376224,0.5596900250022652,0.2051703066093038,0.2968583769243238,
 0.7547966173232186,0.6362986656829861,0.4105790124995057,0.82571807814953,
 0.7472082762373188,0.8270327309349191,0.6509105369318179,0.6164306470272827,
 0.33721005722179476,0.36246936985072187,0.6311455146234879,
 0.6921253828452356,0.019206825689275675,0.9112348133256155,
 0.48691772682099477,0.33038289439160895,0.2280028554398864,
 0.9364143552023066,0.7168665225390957,0.5461041514227607,0.11727957556750179,
 0.10215816425072766,0.40531533050308455,0.7926196038963168,
 0.9698057601941988,0.08679396711442233,0.7164183866708069,
 0.11595814224948908,0.8576222808376154,0.28254408222787286,0.78757840389432,
 0.26563814770305394,0.5952561593790702,0.7066619951635589,0.5707653584851289,
 0.7311292659938771,0.34773906967732215,0.6079201748680509,
 0.11949825353280741,0.31298789530477444,0.630496423611983,0.3447614710037682,
 0.13631138232138518,0.5657846585374243,0.5966886041115661,0.6353986197581225,
 0.23812851169670068,0.05583574956460147,0.025787340059154396,
 0.08425820500424108,0.5762720477333855,0.03278257516437111,
 0.8557883358688103,0.23600286464908127,0.8232803774074717,0.4848057686250159,
 0.6714051781646931,0.11185508538537126,0.009212329838170552,
 0.17732098031962717,0.4073153972155953,0.7908867014906367,0.3316536118404392,
 0.45913559299526263,0.3339191895290443,0.16920410887845805,
 0.9308124307723595,0.18213093593292928,0.03200294136566495,
 0.7926317022533368,0.7908953042967757,0.7416362768657518,0.9785101725768416,
 0.8902926152489572,0.6143763236837932,0.6468422553878548,0.841897865969505,
 0.23287778752818677,0.5606464602247269,0.6065869197269554,
 0.39876171519551074,0.9848336482783521,0.24297566822619032,
 0.31865147101216507,0.9777925567248988,0.969121009139557,0.39490534232807406,
 0.36265576289207146,0.06492708417213178,0.27977401905147015,
 0.39594664746642516,0.4363731481081433,0.7799220754663336,0.3344702834849632,
 0.6469811085630044,0.8867853147158055,0.01300283376180078,0.5945640975010537,
 0.11151332475276854,0.2603113934757044,0.8336699049370001,0.7565786317164779,
 0.19139777775877875,0.272897727465807,0.4523091211421034,0.4446013311207653,
 0.5038486105793535,0.7664569141743025,0.21384925905850627,0.8955917230009832,
 0.5547114931706744,0.012819411139388781,0.9017627539415892,
 0.8272483068488845,0.8441483698917935,0.7812757938776873,0.5685512755060602,
 0.8051952555366612,0.06058838319114335,0.15861271839498414,
 0.5587546443672627,0.06971621521640226,0.4957626645567196,0.8219565608156029,
 0.24782384847223538,0.5463529980821513,0.21183494863361374,
 0.09242231077023866,0.6316991481837739,0.12079511246242047,
 0.7994654215318426,0.9496540607355434,0.7268633765805588,0.35795086337414705,
 0.14584681752292972,0.2692260220381488,0.3848221765391483,0.6648999830266197,
 0.30896735612614223,0.6855815965501156,0.3292369635070358,0.6933241818488035,
 0.8029447360202301,0.23374023964130208,0.7695680452783477,0.718668653501233,
 0.2723235840393706,0.6958986799434659,0.09685015623364035,
 0.20735131178232802,0.45656658722704135,0.8623409994581765,
 0.11070502120177295,0.2038081785730108,0.8227628044911102,0.178839909371851,
 0.7573714198677856,0.7035686170611016,0.14000480072539612,0.9686748971901166,
 0.823046088050954,0.05791175815266081,0.1185556267565342,0.3662402723982382,
 0.9718364999047625,0.6133084958253717,0.061949049039634385,
 0.003149810113913021,0.4766162273173039,0.04317347459161991,
 0.9063785274660466,0.22038235143403384,0.07649178594305894,
 0.37628601012642793,0.8386105420419907,0.948916680030915,0.8983423119063103,
 0.93478871207904,0.5883379769893711,0.16825330401273597,0.44620121057258555,
 0.24269411332374968,0.33871852070294606,0.8897592747300512,
 0.6856847550035289,0.7959600362983192,0.3328453384135559,0.6139774424328497,
 0.21251941279441222,0.7678916614358157,0.04913900687340278,
 0.12320694857870307,0.724259154741084,0.4497323833167588,0.28965967267305326,
 0.8457737150622938,0.2746470971004873,0.3185693063940098,0.2543271255366839,
 0.08903361961140166,0.9562970119071207,0.3918638670111052,0.5529079729605199,
 0.6064800495245949,0.5368700449281467,0.0031917438684460064,
 0.4396008053190851,0.9162290910654389,0.3399525604479927,0.8653675989477535,
 0.09931139673432055,0.18217930331754362,0.14048127775301866,
 0.26331475728031284,0.6415755156837815,0.23034031123267185,
 0.9328539213768289,0.6796745364068664,0.6506107335507436,0.10260430905648266,
 0.9752695667344191,0.6393975130876668,0.6505542650913959,0.5347094051658434,
 0.44234934622288713,0.7029296301698926,0.6147987798637786,0.3632067596029228,
 0.318866340993479,0.3193165452768607,0.4050202310623263,0.15175259256273543,
 0.14933781598880136,0.6722827507904273,0.5639000086178797,0.9006983002980695,
 0.32495429657172126,0.6796148385756555,0.7635978266360673,0.748645702614969,
 0.10158927612258495,0.7235364337037558,0.46234071577797886,
 0.9502818635033606,0.6704969874992649,0.5015928288404591,0.7181070615550824,
 0.5275195922358857,0.8987262940983065,0.7242743873146924,
 0.015800999898797463,0.21000313494587974,0.4194060821432256,
 0.33281781778431896,0.7762744291323085,0.9739382619988944,0.5245902264599434,
 0.2327086549530335,0.8268720605098826,0.17030445243641523,
 0.02659878283025585,0.47151299913713585,0.34630973139025434,
 0.4786037642008192,0.8880280924056912,0.28347643458138394,
 0.043269538353552184,0.2728708081053629,0.42936603980497035,
 0.21531919441636438,0.6707831143574933,0.7276906917631835,0.508175725988273,
 0.6976242575468966,0.6949428800688543,0.14416801895190046,
 0.06296733688494227,0.8503641167894818,0.7016752949963534,0.5522501041368806,
 0.37879198998348707,0.20356372264524292,0.462634685888205,
 0.16071569036797195,0.4780419015666897,0.9766424782404013,0.2503720703598884,
 0.8749470397381348,0.979560832859381,0.8395798243305381,0.8268379977799954,
 0.40512863276705413,0.5143976640207054,0.8462020197136434,0.5152009845001284,
 0.008232059579718398,0.052499437302144014,0.9005968733802793,
 0.7096766764218979,0.716232884150498,0.4322399864001546,0.43308752201259604,
 0.1591672696029801,0.8525598892322749,0.6324922796860194,0.9064124406895502,
 0.033347693746552,0.1350421581494089,0.038706663782566375,
 0.48031277236523484,0.3300919952491287,0.19651954828836693,
 0.5103006512120934,0.8101098449548838,0.5202300609099271,0.7046988763825482,
 0.09641493648224753,0.6528912188571283,0.2809442713690664,0.8961154985199992,
 0.13731331185825346,0.886227425689054,0.3124377263193384,0.5463529249379588,
 0.4131815854026246,0.2529068486247884,0.036461875204641836,
 0.5561695805275165,0.449981851300308,0.47630989334087315,0.6439099234401988,
 0.560136085719057,0.15276539677676515,0.3218578519154991,0.330142750871961,
 0.833463183237576,0.5829753884165358,0.16454057375522835,0.6227374244771138,
 0.9589305385472664,0.06342881001921552,0.20087907003664662,
 0.3192075727728616,0.39546630632266555,0.5155609702512236,
 0.35125799157707427,0.6510235565610485,0.9285583728615356,
 0.12296932970981267,0.2105678757156727,0.8904424971753307,
 0.014402392157685018,0.7508326569849298,0.9451598311679529,
 0.0632778172103341,0.5423441053240242,0.37320871738010386,0.6773027629724254,
 0.04988769673813864,0.8373296317432372,0.7351844883020384,0.5245699982484938,
 0.6432949962744325,0.14618287031789,0.44839167045668016,0.1541335937186572,
 0.3936567753012431,0.27639755577716363,0.21076199672565188,
 0.9705024870544996,0.1737800986561937,0.8224405383252434,0.335858632886176,
 0.9780212248280127,0.004704997095285268,0.41932274943215253,
 0.6596189510399628,0.27074382557094145,0.9867651643706802,0.9103718200822842,
 0.6859920764746194,0.5880935919811152,0.02502110910890798,
 0.32018195669702254,0.13201274461776868,0.16897889001417177,
 0.7785439507226167,0.81707380276456,0.14635228033850312,0.05461827393033758,
 0.2860636243632094,0.33470952764878403,0.0184042165119207,0.6926573056230443,
 0.06822057773108325,0.2459785844171336,0.19695145065691388,
 0.2859219640639167,0.9419524575207324,0.3951451854518052,0.4676681166325326,
 0.3303822794798075,0.42073333223426146,0.08900733461243515,
 0.03834684206026384,0.2578604203616226,0.5538322042345573,
 0.16022065396701013,0.37340944843457524,0.7967562999999355,
 0.6784470065088485,0.5970364152396549,0.5010220845688889,0.6164253623896119,
 0.36318105243189214,0.5551274851604635,0.5681854236235884,
 0.46418294808483906,0.44348751292471233,0.16875831472531178,
 0.3045116729062447,0.3879075969432968,0.4204032591994078,0.3279509968423656,
 0.1386674974966028,0.7897540082308634,0.4814027987578299,0.8479999270580424,
 0.43693858915896067,0.5952900319469504,0.39851554914573284,
 0.4960687121709412,0.9182966145314151,0.7825458422349291,0.6026558085935296,
 0.6450596043669478,0.4236789120645764,0.3206808322872954,0.48815188457510506,
 0.26358688856465773,0.5507862536227414,0.9520723707676928,0.3786796833358047,
 0.8704821167374895,0.011026856755244396,0.9088678180398632,
 0.19288381733507087,0.3111938252680907,0.57024429002245,0.5320454131171476,
 0.7518144137963237,0.2541903444498612,0.30955620253135496,
 0.11650804991995145,0.5827545858613103,0.5526597365461194,0.6956120360018667,
 0.6325232552336162,0.751712749649816,0.8780286529490925,0.7674608069987292,
 0.6713314383865622,0.9334788161046531,0.11451961067461658,0.306013160952209,
 0.704004254634505,0.6226141199989337,0.9617270468588589,0.43361928521836757,
 0.897057220043848,0.05137517310131978,0.68254850002323,0.7582960188237404,
 0.4242520161070784,0.15385055871475972,0.4817086437382523,
 0.46226321817390803,0.48126234728169304,0.32594401388449457,
 0.29613570656143673,0.6452426655970216,0.9081627784094961,0.8000965380372126,
 0.9136057676529445,0.9960422685384182,0.7539301081648534,0.5200726466750295,
 0.9149872233727264,0.052534623355749765,0.48643226732717126,
 0.5600765586003651,0.5296499938025983,0.6527672831449729,0.3429173515262245,
 0.11390540343001343,0.7985108598868135,0.42858996841538854,
 0.4356720233591278,0.7491642807201362,0.3414629719930876,0.21400812462533292,
 0.44817620564052296,0.03105840846650132,0.2982306448159121,
 0.7030058253238063,0.8766606860393521,0.292089810067205,0.44213454912223027,
 0.3182803724671749,0.18305890359034638,0.9650672744300165,0.5162570693654216,
 0.01201422036738764,0.6191226566712379,0.8051360530241657,0.9433759599454584,
 0.8687168171990562,0.8153454243818595,0.9622234530263468,0.833891095575689,
 0.07736159507804308,0.4362227164934942,0.5611005413484778,
 0.38701652768644246,0.013422179556922265,0.9245461205146632,
 0.21879777214829965,0.8114647227663012,0.8754131838247514,0.9275517839217008,
 0.6673391278279658,0.16646368460612992,0.027221867413231404,
 0.054239931766828064,0.7904583876857585,0.7716174989752167,
 0.7773960206126838,0.2209636752649966,0.32698632545977324,
 0.03449565944180055,0.17810048452087746,0.6433500315963924,
 0.5861497938004258,0.3468812700498751,0.9632441456106178,0.6966653733137294,
 0.9679931936561071,0.9597783737308903,0.9450428679346039,0.5392780167979654,
 0.10735792722194881,0.8813212949167568,0.9248088514725926,0.8306678829449807,
 0.16008155663372303,0.9719493426490284,0.3004196817539587,
 0.29827477622373555,0.996508363419942,0.6418587625258034,0.44128465521949434,
 0.35174331273754844,0.47135537457021104,0.25293241645815534,
 0.7320490340079538,0.7922627598338485,0.8446080468911854,0.7412287948235735,
 0.3902457676266381,0.6226999493316889,0.35017283005666844,0.500995667556884,
 0.21968022049162872,0.2791130793752623,0.051715163085512716,
 0.29598681103475033,0.6294086673953948,0.7888976996173089,0.4044466066757817,
 0.747032077214347,0.5647398728710847,0.23616252431822993,0.19547957881916123,
 0.6852354091607148,0.8458051819131808,0.23876694449481572,
 0.41003818387257107,0.18682287944294762,0.6535255260032214,
 0.8239182949279598,0.7181559828413037,0.0012061433653210996,
 0.3687820418835388,0.26755731557853446,0.6356868176973371,0.7877333846002315,
 0.46204861732500657,0.7914985269452622,0.06918853180747286,
 0.07480124121790943,0.4868580789763304,0.37148174804202816,
 0.039607861568116465,0.4597454977491088,0.08263279965322778,
 0.7724318591708368,0.11579361746570227,0.45152839130307953,
 0.3066734440907912,0.4753088451835421,0.649880786653353,0.7395295608102145,
 0.7392887633477749,0.2832913863257701,0.4875033675076098,0.6375346629350875,
 0.5886717681922444,0.19591296730551777,0.10590370753846301,
 0.4190278847874339,0.1530013633966878,0.19717479517466607,0.5779972071575737,
 0.14612397168007585,0.8825116342079741,0.515444786317371,0.852242102732652,
 0.8647111574008935,0.7002355069895463,0.14858680388782386,0.869087547189233,
 0.7298175451971615,0.8311182880853316,0.02181751739543325,
 0.39973414890612635,0.5522611891395776,0.6398032106222662,
 0.14810797407588605,0.10224596840440958,0.6313757136623975,0.403715362927928,
 0.056220189841004964,0.9441250635816698,0.073532647280109,0.3199548228116047,
 0.5008225141020857,0.31955697564630225,0.30554015115011446,
 0.16565375118352943,0.09462838605327006,0.7648056413734707,
 0.6747391895838724,0.6217197639201738,0.09467729325565821,
 0.33969822309592457,0.7951376767048975,0.23321858687687402,
 0.3627658688528942,0.16494654903615524,0.24171427621840547,
 0.8139927320402605,0.07540912926903431,0.6947996969264121,0.8377032032450947,
 0.8004706683626656,0.33673933119405586,0.2158591436867815,0.5787552340792919,
 0.25417626045533503,0.2923702277402225,0.8802180340863418,0.3670415350911044,
 0.9049003083632109,0.13488839235617933,0.5631387455945357,0.7417420646145222,
 0.6377894574373777,0.6630439089204511,0.4911461676018589,0.579206455207568,
 0.8149546150999255,0.801025116626442,0.635619158271739,0.3848811108089101,
 0.024274229662173186,0.8468487408603277,0.9788607043447718,
 0.9832610099133474,0.0669717344531432,0.5595123597308014,0.7811720043683922,
 0.10607217034759986,0.5258806567798455,0.10141004083756489,
 0.08970728231872815,0.3855191911037351,0.8440592685481743,0.9831065649444244,
 0.5243911342945103,0.999206672249225,0.7206421005947057,0.6403090438080556,
 0.7150093448872551,0.5116415552174368,0.40069333047201294,0.5910626478735538,
 0.3038080363104506,0.46734348544268944,0.39958843534125965,0.632976376010772,
 0.8647177384700483,0.09662571425483879,0.9766806144465398,
 0.15164476214402156,0.7241494756973011,0.9829553820558048,
 0.28148390375057486,0.8581038739546247,0.48892551706131926,
 0.12027415923010443,0.2202682118547692,0.7989640426986919,0.828428985559625,
 0.689840478853436,0.7165876130010422,0.5136866437589994,0.3801744969644828,
 0.18369120292622287,0.9401443057043616,0.33582488132958,0.7266124801683518,
 0.5170668475010519,0.4738788606943948,0.5400028562753315,
 0.043635831357034416,0.4205095116434444,0.0801049343079101,
 0.5805357680840539,0.18600910308658047,0.285618275263135,0.4511957387439912,
 0.2325411669469698,0.2638428489496544,0.15483085052184464,
 0.16730498897271673,0.1635042553680317,0.21617805828222525,
 0.39227086359384566,0.9988516053387394,0.6269228282615374,0.8353320006327232,
 0.7932712197989747,0.6479327063174936,0.7523617686033859,0.1383987914546263,
 0.19512568941392305,0.25127090325359913,0.8608875520291089,
 0.17784774651294444,0.6120633229012717,0.5752152485485352,
 0.43845503866874647,0.38271786525301976,0.7839612728727379,
 0.25474716187082025,0.5336306145714649,0.48200471570923675,
 0.6200924193765587,0.7938998661776795,0.8105499789261124,0.20009808965248932,
 0.6967518178265566,0.9206760285051411,0.579765995111397,0.4302241272067162,
 0.4719518206710497,0.05364793863831818,0.2558786027653607,
 0.10054459377103364,0.018725002875957797,0.12926201176466234,
 0.6318293318788957,0.4662820532804345,0.11247023608234441,
 0.19722424102158553,0.20337940668580412,0.7804638698092968,
 0.3701184279028117,0.8153214713626488,0.5335463812682777,0.4243398491494397,
 0.4883950203800348,0.38156949244769967,0.9503192826492748,0.8575620445635441,
 0.38794320422366335,0.23210642465302855,0.6444497224558281,
 0.0457602256820987,0.9365282939140198,0.7379477535419744,0.15398656011398737,
 0.931373077823352,0.08801811548580374,0.3244022527509617,0.47196939493547607,
 0.7580336899347525,0.5610187417149918,0.007066931635885676,
 0.6000025776555601,0.7780175829221466,0.04502079406135939,
 0.44537180565723244,0.8834945729507849,0.008629143806057504,
 0.7613009105116315,0.09652269716246975,0.1906842436155709,
 0.42288764663460066,0.03441431947493434,0.8442884378465234,
 0.6389572885390851,0.9211260939059954,0.8230274085813302,0.5140976407137329,
 0.1588262322509857,0.33936919592108583,0.9504653834049418,
 0.38695742111971865,0.08119819132975625,0.5188202820553006,
 0.5711663071892195,0.3016156173112854,0.7559984726351717,0.4476305403466758,
 0.9688399536535686,0.7091120008179013,0.47466300308906684,0.7725104493277176,
 0.9235142063646917,0.7698946123105596,0.49921495809859506,0.4448274453111608,
 0.6200538860284394,0.7969534915026963,0.8880180210356599,0.38683403417395357,
 0.8045659546835677,0.7924974844585454,0.8588436117631212,0.568570217785274,
 0.0205898516926315,0.3968128149489488,0.07858127260180414,0.7353972804184638,
 0.4311601122111435,0.814027336707203,0.2226243298610241,0.45136077974613065,
 0.8240450482340531,0.6952725694821993,0.4887151466475004,0.32246805192565053,
 0.8938472245889164,0.6641307969794774,0.7960318138495605,0.9620283485755613,
 0.7212361215271472,0.331210349429509,0.451383719828127,0.022859540541478318,
 0.24573870773773143,0.7564857027137591,0.7007575055855711,0.1022460702419713,
 0.2994107738009727,0.026428447499377894,0.39699660294205685,
 0.2924144146859804,0.14014007280331853,0.36634253975045905,
 0.5968528816656368,0.9551576918173355,0.18080910274905937,0.2726793024262274,
 0.5839855971532673,0.19314131021602765,0.5162149002904067,0.8461715436878278,
 0.6036004987941097,0.6063163089128498,0.42098474320349333,
 0.19807848664803396,0.3541328879356538,0.26506076292244773,
 0.053848726371326094,0.3739424661158941,0.9711564882617315,0.854658217094605,
 0.1937916692652475,0.691284184427279,0.9128354355706928,0.22956870547539776,
 0.6310554986872043,0.2076022447499356,0.034378774487002106,0.969797537232139,
 0.2507629750931195,0.024390100845707363,0.5720630427624935,
 0.7030716342782645,0.766697355283137,0.8848501272389696,0.6443420800257147,
 0.7002438865172287,0.023078847664658286,0.5888116293965229,
 0.7604475244761325,0.5631248048646333,0.8283914709655231,0.0876812890766423,
 0.9556529477510561,0.8841683932939084,0.2880552206519943,0.3060584071290544,
 0.6868850514707983,0.3788705104873571,0.6707060614010211,0.9788505091344477,
 0.025517510230388507,0.37824534762007533,0.353167879888061,
 0.31214986470045825,0.01658534988374072,0.8362262447412131,
 0.8542721999673918,0.23259836635046394,0.020292768650932547,
 0.23968595656242142,0.5776403061572182,0.24437609569423702,
 0.40159498833497687,0.9882307953128892,0.4342578613909399,
 0.21034959360230665,0.4425721273115717,0.888474669655137,0.6123516973572123,
 0.9896516242652574,0.791274967587953,0.6970013137640462,0.9032935038723909,
 0.002973117093421296,0.8671451440071156,0.9232702155226369,
 0.5470834932843878,0.3552137877172432,0.4876261823585344,0.7800929762928518,
 0.27222108174740345,0.7997481355896292,0.2237398141555681,0.7241409947759192,
 0.21564483464560769,0.1569407510986649,0.9158797465400019,
 0.36388540337638886,0.46970017073115833,0.42306355135547324,
 0.06198932788589118,0.412658144993874,0.31255564578692396,
 0.49068504992973827,0.3923090815775617,0.17171319876267077,
 0.07722359762617581,0.4772304642137907,0.29978146167839614,0.591746951588519,
 0.9518825989043559,0.6618216869437628,0.9888970862865286,0.7770796858516883,
 0.5348825080349112,0.5942058355267279,0.5193009239672733,0.7880632001109675,
 0.8507120339207529,0.4102796248360683,0.6129489944067331,0.7208680655894258,
 0.8605980348747239,0.11390518821378293,0.5609057503864318,0.8711688666900019,
 0.4750143063357912,0.13828229560673477,0.6997847475306618,0.6605644970584712,
 0.4951119304345808,0.9012987913243049,0.2128568608793301,0.4133310416569017,
 0.4145560678238648,0.127290543914361,0.23887355059033544,0.13536690012274588,
 0.18510714700176167,0.13165972432770523,0.7426141602852989,
 0.6867195689490854,0.5005215007719311,0.6400243896304446,0.28744365916517634,
 0.36206161230269696,0.261961399586095,0.0721444321050988,0.5174037451667961,
 0.23895334901202714,0.7791145408413893,0.7831270784052099,0.9140156928552259,
 0.8091928650685062,0.08741818753250563,0.8308154852730338,
 0.06908494498823092,0.32870094839435193,0.7032271686313781,
 0.2540748180716782,0.348039837399883,0.6345102075481012,0.8172713304945753,
 0.9295950425095894,0.3108743248130029,0.7041906686780891,0.5251260989786926,
 0.6469854124969266,0.6717727744595767,0.06169948415745219,0.775412019076188,
 0.8160440806082758,0.1752068321364183,0.110212851494921,0.39466760159136793,
 0.2726208650416725,0.5257227934113917,0.37434050829958854,0.5061374951593013,
 0.3341087915835764,0.7386260086411375,0.3115506464562361,0.37342106795255847,
 0.574184047529805,0.4732923281123146,0.2721034060463413,0.12952008411547833,
 0.01567690580758918,0.46117300568141384,0.01632960945409212,
 0.5403807214176112,0.8362929158857124,0.29173437553591564,0.602288048669661,
 0.9250040430290206,0.6180711824547217,0.5008046982538801,0.3933945345079435,
 0.6007344235067759,0.3787439261562068,0.8436056063311068,0.28220974189697534,
 0.4755714886723582,0.5365507635394324,6.330444763181786e-4,
 0.7931078727642173,0.9049814403407789,0.25393756142348356,0.5611793279001929,
 0.033831275648960824,0.23328975513249361,0.38115684916665593,
 0.45742985565906014,0.25744812081178226,0.9363810707884183,
 0.8911550598737593,0.5609611452129442,0.570047682279127,0.4294021081070577,
 0.43726891573770543,0.12392174550990742,0.2788860218186455,
 0.21936469774678868,0.13776009311201376,0.10889131131746976,
 0.5913504596459761,0.2241683747894736,0.6325723649132056,0.7980823233115116,
 0.18942655175561296,0.9326236903103046,0.05166757070273542,
 0.9017972604345775,0.5530281977257089,0.26657155826899837,0.3572217964133574,
 0.7362372636229673,0.2605948996688102,0.5277677622016708,0.3707088026165396,
 0.48981248909765673,0.16733509082097342,0.8581218622044103,
 0.9327077663396123,0.153070043149925,0.8307168404157755,0.978631796550039,
 0.2805946712315398,0.2014151791698059,0.5604015918825963,0.08814669859560653,
 0.18753870444624843,0.9655993216234524,0.7613588746178644,0.4834515513288461,
 0.22558107285646856,0.33109906230761776,0.525768657740156,
 0.42147291775449014,0.013139933843495322,0.7554677824476378,
 0.9270234589538869,0.9503928237875211,0.6729385304096496,0.7408750268815399,
 0.8350531396003259,0.674699632114687,0.16553196668916415,0.62564169431144,
 0.10750047096540138,0.46858796511815404,0.5432965095608492,
 0.5038931695696305,0.6074796312728232,0.4284105650631995,0.37419804024780323,
 0.5263564743422767,0.7544977560533904,0.23241193342788158,
 0.35360994208030094,0.06495925908552436,0.6225065895724612,
 0.8857562170635666,0.20950901054343818,0.37287726072741345,
 0.21396412092757988,0.6968689878892284,0.5883350723387326,0.7387689689211422,
 0.7103217085573269,0.4972132437798755,0.923232366093161,0.24373849477897003,
 0.7229015146072596,0.989915335858892,0.5651421154526427,0.9786412533340169,
 0.5742757447660984,0.4472529212023312,0.10383210081468575,0.7394408295501593,
 0.9903200860000092,0.7851559835946196,0.48379148529262817,0.9428907015242595,
 0.21692963844007918,0.8141087538513503,0.020824307614537352,
 0.8908037401259302,0.7145893821747263,0.37512384025037315,0.9550878287521081,
 0.5417305749354726,0.33737587245052125,0.6527161258168321,0.3356717800459803,
 0.8138594640864905,0.8962569507707857,0.7565404645945933,0.16092255956995993,
 0.13058509614823266,0.3540371839949311,0.6330688460204807,0.1591236849714618,
 0.4505224794094034,0.2372797951993999,0.6881197050672639,0.1827331750840253,
 0.8897705789933789,0.48704119249478517,0.8479761431623651,
 0.29025066762576024,0.8046718923531153,0.574469175966934,0.44659163244124056,
 0.9416030192473813,0.8511231806182454,0.9660622777148813,0.959985806094046,
 0.8482672196576607,0.08588834218015795,0.41194336613799754,
 0.2550481389548439,0.07119555164389424,0.2545464502146889,
 0.26185268339807755,0.3777977052940671,0.18711202245248115,
 0.8874995789260246,0.11432937664157627,0.3691292134169146,0.7969284113244729,
 0.819938093151755,0.22959526394271623,0.5171631988733505,0.7264043545707565,
 0.15663149647642727,0.6449315666628922,0.4697091548074088,0.7712960943091869,
 0.6809320934418222,0.390233020326328,0.8662990255626184,0.11451320843135948,
 0.4394426403792496,0.05308724789384489,0.49911577244080285,
 0.6271599078556644,0.2912068688087597,0.038158514606534966,
 0.3721038120874498,0.04885137454462596,0.09835645315968111,
 0.08563823218147637,0.5093065858282094,0.3965430784212244,0.9409967591715354,
 0.5430122421330279,0.2950956637900135,0.38778990270075653,0.2253425290668314,
 0.9331473856852597,0.0665075057249902,0.8004506356604321,0.3539648224105266,
 0.8076407990623053,0.5082959175246843,0.35536887877267853,0.2685329066113752,
 0.8221624454000656,0.9070915692726843,0.31938038289998527,0.8984466346472492,
 0.9594826209492611,0.9171786847464716,0.9270224150229252,0.13624203172984517,
 0.7239447380136858,0.2218964863537508,0.8889281781843615,0.1178351310281388,
 0.2681084726072239,0.37506563715738295,0.03793411633665822,0.823369650387354,
 0.9454390425336304,0.8383744363150912,0.4531633109392835,0.3365633348762068,
 0.8376940371325493,0.236309479978709,0.8993536604318737,0.5201592673461686,
 0.533440366715347,0.7060298271009637,0.7690104894927893,0.814466362759425,
 0.32892502996975725,0.5405611133309564,0.025687921407819503,
 0.9805673670318619,0.47513926230705517,0.7050323733977367,0.4405275992392442,
 0.3939439723139413,0.4527036992137654,0.9294350937707986,0.3739427840419175,
 0.12288597424339331,0.4279844905756244,0.5578493630339371,0.4114986439781936,
 0.6753005443920037,0.7207618883987106,0.9895506978479517,
 0.014746488277451064,0.11822773517053364,0.9773880448486828,
 0.9726564044985374,0.5280101362988499,0.2883487252975625,0.2853182811447522,
 0.5633577789105595,0.8740818141185541,0.3233772499584847,0.3465470212695493,
 0.47244988658243703,0.7050776631982577,0.8512571691242338,
 0.17367431905572883,0.12483756169905802,0.10946646816406935,
 0.9335160400988897,0.5656542517625527,0.5689874179372312,0.5848378839358633,
 0.45363377059181753,0.8798733120354256,0.8280299814292215,0.7030787310515468,
 0.40082891393040443,0.7029111549443179,0.743456570951376,0.766509584831379,
 0.28841092953847136,0.6504737666458196,0.026101436763420027,
 0.1022890211818096,0.31895292415375376,0.7075548915346308,
 0.33062046194862693,0.3228999915511136,0.016477103696360462,
 0.6589955476640446,0.4169382904787473,0.33335505929096176,0.3509952910805142,
 0.9624000199060013,0.2709128689152349,0.6003872660473046,0.9516242027241041,
 0.26240298856755184,0.43975258184217125,0.7705221965627164,
 0.03431403070025474,0.4270862699136273,0.12767002987631348,
 0.4699714336689027,0.5141727755499944,0.9137838030666258,0.47906108514047296,
 0.8265563764397965,0.03638910207675683,0.37633961357026835,0.71009093071376,
 0.29564397434048206,0.17803250048111208,0.5724085137248848,
 0.3092776749305346,0.9118288072715989,0.1627448638302298,0.6556674760358187,
 0.2407227499675466,0.08686172198274122,0.08839123257122261,
 0.0401867939064775,0.6908400947322731,0.2583271058475337,0.22214702829552935,
 0.8108777680127323,0.21213605280486847,0.9987414738028474,0.960111340300096,
 0.3036351313954312,0.3373576352381136,0.9241919022485426,0.7507717388237618,
 0.38854851731502926,0.0782947410853505,0.7822559297285727,0.7519411426509788,
 0.43545082057957674,0.8034489482357403,0.29244158264365494,
 0.46240059059847516,0.14715999218067877,0.27000186676393634,
 0.21496595350189673,0.8069063115734498,0.9189984004619484,
 0.35032650140568955,0.5172435621072276,0.16996963954531796,
 0.17242886852996875,0.09360997962958184,0.8727304603952994,
 0.6441981850130809,0.2546802271512286,0.15551104762951728,0.6770797318653756,
 0.17217013779455226,0.22488531780474508,0.5789391650505179,
 0.9059069890181008,0.03909528861812317,0.4417194543666456,0.9302579080859709,
 0.3255946437206463,0.5964264402273791,0.4677584503594203,0.1639225491069073,
 0.36221719031954713,0.18191313230220718,0.1570495638924384,
 0.46931041678417507,0.7991391121770647,0.4926083702320523,
 0.15902465190698556,0.2595273928402557,0.7778002410784214,0.9880085805021206,
 0.4581160069968089,0.6441539095752463,0.5619855545213985,0.2038778347283161,
 0.8062782096880398,0.8558307062759192,0.5156040825903445,0.31856861443686146,
 0.40455627160493224,0.9938182406400496,0.27226088814085503,
 0.8322337765147974,0.47009295166535203,0.12375356785835812,
 0.22632868405824602,0.2073179087747532,0.5543320315791098,0.4383692520904142,
 0.13022358310337312,0.15239461591937165,0.2602666748984195,
 0.6938073428617639,0.41588977312552333,0.7359304083871449,0.3627835809373212,
 0.7142849362501968,0.35918898868110904,0.4442445536690445,0.43480449108828,
 0.7540582508708432,0.5118456002786824,0.5857501855231672,0.923175435206447,
 0.9041770948215597,0.21379665837381978,0.9080974132421985,0.5354833153337839,
 0.8736182105016967,0.7383890712942999,0.175265053347887,0.10903246090554652,
 0.866256672646648,0.49623153479007964,0.4365117202354847,0.28444866860610096,
 0.9033178170011167,0.005802747797065022,0.032329769406166387,
 0.7286130419500776,0.13312579354454135,0.7269659607131198,
 0.26007265983088534,0.2434471783828056,0.10025425978407132,
 0.13852810174073948,0.1680819214750947,0.355043901291642,0.8753638158478426,
 0.8635486976983608,0.7200576773061466,0.7150403098918405,0.7346344808440125,
 0.12681225239288973,0.5179399059168874,0.05855924665646861,
 0.06566221537828976,0.11131929003993846,0.9258640148703848,
 0.5252768311178039,0.3440743023022136,0.6614655926252531,0.40639409956327566,
 0.29284415740657943,0.9208968901522635,0.3019158304191023,0.5022787009644128,
 0.7321935526541101,0.46627213653807886,0.8476373281461723,
 0.09489474043778012,0.2360353342160173,0.13837406274069897,
 0.6251231949458838,0.9269306949513337,0.13983427351680122,0.7366245929604182,
 0.35775298268599864,0.5096966868439387,0.08561613940363322,
 0.8112796110451457,0.3446150432009438,0.5915869929444332,0.6101719075567209,
 0.8381521607881987,0.8619772055623256,0.9509757963851578,0.6403706463948955,
 0.6208430061330382,0.9221683807659609,0.15691639640974175,
 0.44395952221457236,0.6029404907455239,0.46087899298728097,
 0.7468414934646164,0.14274205500565174,0.8824697917454662,
 0.07194404307846125,0.00901471015620714,0.026091340203003077,
 0.15636830390505962,0.8198032457745502,0.32084461439924017,
 0.09451507201864973,0.3655664164717807,0.1071343189770656,0.7640473891123521,
 0.04174044089128082,0.9024575161544355,0.42579165701142196,0.760619954792336,
 0.753379515110441,0.7126035560112003,0.6446923010429314,0.6889292995262597,
 0.5981594860641463,0.5643512400678612,0.3493263527604954,0.7322287793057859,
 0.9118711279591616,0.18402644509155508,0.9239739565305752,
 0.26530665921890617,0.5595504367219005,0.5829377763032124,0.0172634080168933,
 0.1676538848579684,0.4541676966870668,0.2264085312628905,0.31791725876945587,
 0.32580599037700253,0.0654191630147799,0.7319985646092284,0.1490469453503006,
 0.22878627239734584,0.4505267156477534,0.19612574999273003,
 0.38476275169947705,0.3025589231655934,0.7601714697266259,
 0.15039045989923217,0.3391115349053494,0.0417563911202492,0.9474728782221813,
 0.14077383272382105,0.9498143404504613,0.2085052756210446,0.8730819977113057,
 0.6886340919123538,0.9355467497133457,0.4279995204610757,0.23854519361657145,
 0.3591176128937529,0.673376529671782,0.2670941119221878,0.17346070430930227,
 0.5078946566651397,0.21064844493138246,0.7761666325136816,0.7924081439308253,
 0.4100660624411868,0.5310916763912698,0.15002557930682847,0.6219733629105997,
 0.4626593996466497,0.2079973199210241,0.07769811789887926,0.7687015858355423,
 0.9680234762919561,0.2550209655380541,0.9254836381722168,0.7998187504903196,
 0.6954440176692294,0.322790224813431,0.4485603709907948,0.5557372825607478,
 0.4880257938164727,0.7774662888339605,0.08161065216901464,0.6370022406296358,
 0.5149291949955839,0.04855152066006285,0.07149797313974804,
 0.8524643617358015,0.6412948655602213,0.6402008783658839,0.18026822890641736,
 0.3018733015610686,0.3746584752050621,0.9325805220503012,0.15914606932563313,
 0.07473502551140188,0.9040786847112281,0.12607416308882125,0.66302448743722,
 0.47252730495988426,0.5554281393756482,0.5196462458163712,0.7420273142632883,
 0.8980559485683249,0.10955864873084564,0.2266979613668837,
 0.36242952477524326,0.9522425600920097,0.07442646370614181,
 0.23511157428807494,0.809027718109757,0.45348967039716837,0.668625259329626,
 0.16779004709312728,0.988605595974443,0.3999047967913737,0.9513513220659147,
 0.8577223551977703,0.9830058671421907,0.08942151751220306,
 0.31195133132758157,0.6518182088132611,0.8626388470594724,
 0.24229829803386305,0.27142803025033957,0.11030394282776212,
 0.24181058626663765,0.622001335781361,0.10390804423099931,0.8852255092432791,
 0.7421926485978363,0.8069559789698381,0.25001192087585333,0.4748258189258415,
 0.5464938336593728,0.2608235155765073,0.09012271955237172,0.5256201108477349,
 0.4477962093425668,0.22388957366359485,0.2271263471482481,0.724137023018645,
 0.0038265967104894605,0.8191696547010903,0.9478563821565819,
 0.13930576370946302,0.3148293641519878,0.48045372169172784,
 0.3981055956420614,0.0987394201947267,0.5810997687721857,0.18138467576467743,
 0.5310495097337289,0.5126294240427662,0.1691872874195376,0.56680872679129,
 0.1252112832654414,0.6656912888222563,0.1928843431414191,
 0.003155285208015668,0.783831920354303,0.20496150401925095,
 0.9163355818318277,0.9424552014558832,0.7950970700499891,0.32538453587132254,
 0.18835481344976412,0.2120570211454098,0.9832185649394214,
 0.05921716828290524,0.423770912268258,0.9250823937976654,0.9665886043502391,
 0.12217543484410776,0.4305096094473978,0.8842635168598215,0.7573423524276404,
 0.5160962038773842,0.0651161425287975,0.4266301736911742,0.6501495560385566,
 0.3203058823821252,0.126877698980697,0.5928678049976855,0.14743674360379933,
 0.6697425593669032,0.6036704770209392,0.5033522558205473,0.17654279743257262,
 0.7623776858939753,0.8097481044993899,0.018102437443247354,
 0.4295092177048806,0.31600020258105643,0.31039398617985914,
 0.49563341134603867,0.6600385056926592,0.9660757339606489,0.8973343967089044,
 0.08079021938421838,0.3529514516273441,0.6516699991085453,0.5913435543213968,
 0.8862009834578409,0.20957138388032703,0.12683999838087845,
 0.023807432739213708,0.4597760816717662,0.9907201990179783,
 0.22939366070497702,0.09177273868568037,0.16624378085040548,
 0.04372949319798014,0.23023713970070359,0.33127705610945246,
 0.8965231138599795,0.2938778529247462,0.21200535658073616,
 0.026240312743926686,0.05100561715968421,0.8647934598387496,
 0.751039372990829,0.2510374152633257,0.917458458873183,0.608244282120721,
 0.7305367106641218,0.37427848492945126,0.9413760443841267,0.5463490731571707,
 0.9627098130773479,0.6763285352874181,0.05552585519624542,
 0.25541286218697934,0.19907255726995554,0.14561676543875068,
 0.6219221787520482,0.37028043867541194,0.4394613245125736,0.8142476202796387,
 0.7679279548930685,0.025792960678445498,0.9263831522269717,
 0.4233289595612626,0.7637909428230043,0.9772223812904464,0.9222098650378223,
 0.9820568600044269,0.607159795887115,0.15934061509804143,0.948116233662362,
 0.5587917636426315,0.8950704656524994,0.4822044889747932,0.48253518662311423,
 0.2413178675688974,0.4410780835391763,0.2299722382456445,0.31721440638600384,
 0.46230868148678983,0.04790016480238024,0.7818987502616297,0.732746781241143,
 0.2911023799186653,0.07904880154304894,0.319024632839494,0.9673633749836433,
 0.7639116979254855,0.44918589688057264,0.6401523946731346,0.5293700761476453,
 0.8903251095074399,0.6258289247621215,0.8036293783452935,0.6506938691411135,
 0.6209727666539588,0.08079260311674352,0.6462363922078663,0.1890977924055911,
 0.8318840258366824,0.5721216436170222,0.6031467955145415,0.610173964641572,
 0.2315164878944227,0.6351453392447769,0.4031604918189208,0.7086408259007375,
 0.3567006824243385,0.37523250747472303,0.839431034249397,0.9804931927634092,
 0.0372398501895187,0.7956132123881853,0.01867927837496186,0.8932816756734299,
 0.1344288416490429,0.19035362025495817,0.35385301299683836,
 0.8433734478337778,0.5857419622732092,0.6838348797276769,0.15617773428533233,
 0.48968834314317944,0.6669231611155046,0.0019017686948430423,
 0.7344147545169452,0.8894932225099519,0.8903953879960063,0.7479905619835191,
 0.2283901796872161,0.9430150732190794,0.15134920708145128,
 0.49220759265721326,0.991499832347173,0.3344948060367676,0.2473014614703839,
 0.7227673190877197,0.3022682310154682,0.2765783674832707,0.44456105924562084,
 0.5548125089556252,0.13281277746893105,0.6591426437065637,0.902512824133008,
 0.8827970743873388,0.8723934541528428,0.5423869719205253,0.3039409586424222,
 0.7497948405371287,0.49988159739427396,0.7021519738027817,0.6138581372443808,
 0.7750251054999284,0.5363519277632449,0.6500490148713023,0.9598504962594014,
 0.4082527738203995,0.3629546526113343,0.667299198249928,0.8636914327091181,
 0.578066219856814,0.1400953139910608,0.07413310547333973,0.3306394298321488,
 0.19449060154354125,0.638743405418188,0.5784325282045503,0.9400172898035309,
 0.46981525995452234,0.41837358387445156,0.8861827938590459,
 0.9396598762484181,0.08181883682231228,0.5099192537214856,0.5517698259853506,
 0.17764181464865336,0.794841464959123,0.471870369696458,0.8656320175430833,
 0.40593959873359053,0.6117079382766386,0.627593991162909,0.9800925997947432,
 0.47801658820087867,0.4581541504641998,0.12058724996857118,
 0.6281474185512247,0.5462931915913287,0.26975465789265396,0.1463135022226637,
 0.9243517492265094,0.01504456660292286,0.03797459537426506,
 0.9336873837963645,0.13042399969560903,0.6622717558809211,
 0.22923591508604757,0.761339011352754,0.20603573923789875,0.6317918936679392,
 0.881515988219024,0.5907093224210158,0.345242175464901,0.6370674203738944,
 0.6210354562651429,0.5339074989222541,0.5746913347035923,0.913184981482717,
 0.4221332186995873,0.9229286835323487,0.252835039201052,0.04598434950462582,
 0.48905057329404833,0.22394928108519163,0.901923973791211,
 0.01801223288380749,0.5757885662529045,0.22547298413038097,
 0.46720225959803263,0.4359055652887792,0.230354341780589,0.6868910573332061,
 0.39690015966173053,0.7176069122344741,0.8293852433132142,
 0.43441423592108874,0.07113349495356891,0.11524602292892994,
 0.8124585617618982,0.21836558686883623,0.8350911801000142,
 0.29153881656440184,0.0852538145099484,0.30488728197089165,
 0.4420023004671896,0.4236808546782366,0.792866248316287,0.20227149347483964,
 0.9371913870375586,0.6063623758917219,0.4443938516987005,0.2592182083491933,
 0.16830776788840418,0.7282916607340717,0.04648862824502942,
 0.5299095026241089,0.47940839528105816,0.3517701770636471,0.4520761517478612,
 0.4040822649283812,0.25505688571427976,0.026268494393683395,
 0.7613300629453393,0.711410929467855,0.1443345022120619,0.4444199710347446,
 0.6763960596163732,0.07279506119966384,0.03691670011117143,
 0.9507962450088568,0.45804298958630607,0.32073997935608123,
 0.8786514976565049,0.8342047477810894,0.1032711801647016,0.25325452591795516,
 0.27124128539614,0.6313186027215336,0.3199397894562468,0.6144914816082416,
 0.3033355360740251,0.3055641872191819,0.6591289025931972,0.6668120395485315,
 0.2399025329593022,0.5403828832604215,0.9691286038620684,0.6657754582483154,
 0.4016181546588343,0.5566160389872971,0.15079775100433057,0.9653619415724414,
 0.4832208128384754,0.8175166412405712,0.3692580190053121,0.3228393154359336,
 0.18448744400540096,0.7231761580122779,0.35645927721137805,
 0.4567478474756935,0.7098904452668436,0.7012664314116792,0.2251219259711874,
 0.15857942191884555,0.6375529116530363,0.060280658023762146,
 0.23371179555808363,0.5155418311232991,0.12966481114164485,
 0.6745012834782356,0.31277384804230923,0.12347580763941857,
 0.3310595514092951,0.41275613603872396,0.3574969944186587,
 0.27677231107173295,0.1674924706315356,0.12572369045018805,
 0.9652323540997484,0.9313620713756775,0.8094385533051871,0.2868177909590133,
 0.22187765476412313,0.682252385259899,0.14073319004801066,0.9249479058758288,
 0.3884471082877826,0.5391743172730626,0.4384926292662823,0.7445370703538376,
 0.43135158770981075,0.7516563732737229,0.573840404456236,0.5061284896630912,
 0.1722491019340885,0.7380703401070989,0.14175264193235992,0.9041373754533314,
 0.43378718908324243,0.926843328620415,0.6349021727502953,0.4764780582504786,
 0.5588756749150948,0.7803216596914181,0.27617306287735344,0.2742094920873902,
 0.5106893272916764,0.794894357497954,0.4580645306843798,0.09608479518720581,
 0.2501865994418335,0.4822682653390573,0.8377864959353822,0.42974210511808375,
 0.5318513654406255,0.6982055419160151,0.9479822938941873,0.19123798540581072,
 0.9335859499625125,0.6476609663482543,0.07574986020054986,
 0.48966474140755356,0.5610281388281082,0.2332821596589879,0.4902533201192051,
 0.4524068324855416,0.18025068520687904,0.11438876777190199,
 0.4319218884734848,0.8708758703214636,0.453781127217054,0.33772710691113994,
 0.11574097924734039,0.9895682125391301,0.8565821806950685,0.3931547791977532,
 0.14892795065123665,0.9932955886001491,0.9337888358812958,0.3296691798536783,
 0.1564285197459926,0.7875943032289581,0.8455185576981118,0.16651677889393168,
 0.20473718819665954,0.47878029266506605,0.6365042764310025,
 0.8046481379847156,0.28434877935742653,0.6707084237327197,0.8382930863176574,
 0.254220490116277,0.5827897412939894,0.11910579536157972,0.9281587762636369,
 0.6421479070317238,0.0011561832757922552,0.5911525820604558,
 0.2450383580206803,0.5563105925161338,0.47392396339856746,
 0.30365891386703203,0.40329066559038806,0.03293543080193384,
 0.547968805081795,0.846059890316283,0.8311005617349356,0.8542599619846554,
 0.47052308767408313,0.8282798365040112,0.5714756308254312,0.5999890037692786,
 0.5743064837115494,0.8920019193354125,0.6121887471361059,0.6719524522251897,
 0.7188331815833652,0.04632284300552636,0.48752456336530314,
 0.9416740525108798,0.8372635673852152,0.1966168978158499,0.15006084681306753,
 0.9879718788245462,0.9911623771150477,0.8789148541501915,0.6900793362996397,
 0.44451717462684126,0.1211676224213194,0.45914192195890646,
 0.45350689513796394,0.26185896383956764,0.9775317266556797,
 0.48274294333108847,0.7643196861727872,0.7244876694485023,
 0.24822025304782813,0.35616313534312083,0.15408353087470472,
 0.2125185544312721,0.6096566163184446,0.9517995301180091,0.3498605556167157,
 0.6908647505542773,0.8056077330176956,0.6734385761039828,0.7024681603293078,
 0.39033041912022604,0.5908214173817036,0.14097674981915775,
 0.8368108972209929,0.7511196608996634,0.7523974631567425,0.7963853349986039,
 0.596686465866749,0.9632794089999711,0.9363671569289044,0.8664553617421962,
 0.5674620561244555,0.44964007486768365,0.9809172681457325,
 0.25009841142094613,0.9091728021569123,0.8406058081280187,
 0.013501329646843474,0.7611801997639772,0.8846191475232789,
 0.22622778194236048,0.5481967374738151,0.41915847076353197,
 0.8291481006589969,0.8516250528364862,0.2759357685911449,0.6461214973890282,
 0.7806274499613066,0.25459994462490476,0.7675967304922074,0.9491348781463484,
 0.43535867505005854,0.47471963443763043,0.1569041677114853,
 0.10334092476656864,0.879403652910355,0.5619176164214974,0.9497731274022805,
 0.39541601341357613,0.7953897025628708,0.41822870398140377,
 0.1944112317674137,0.9821943093392982,0.9633031256596315,0.783771141680835,
 0.7813383092047316,0.6496158156082896,0.06495481364668587,0.4640343593313023,
 0.7761231265523791,0.20382821445189336,0.4135045344932482,
 0.02016935081548421,0.7805389865210954,0.4655105672721531,0.781782169454079,
 0.9581684820797574,0.15005134356035443,0.38919367430321095,
 0.8101463847659125,0.1477392070748691,0.38331092597254346,
 0.019516338942672906,0.5008133145246118,0.4151293355827461,
 0.5884354177232424,0.22806774120593254,0.45066282632706134,
 0.6792694986380964,0.8196112396767115,0.672919679578268,0.0948903295879,
 0.1154852695769839,0.9229521828686427,0.6984900282604333,0.5260901730707986,
 0.472189265126603,0.20144617521556307,0.7799322040015757,0.18495430402813784,
 0.7679882893903891,0.49130996537515215,0.6199625726392903,0.1650080916311939,
 0.8658840122596998,0.35686697222817654,0.9874889540778745,0.6159731245552618,
 0.5309625651222554,0.4857838928888838,0.5676700529684883,0.6209597474055557,
 0.25910731176286705,0.8147644912741794,0.8633663624179025,0.7228676292617765,
 0.26848245707216534,0.48434914932800543,0.708222233104825,
 0.09080285763530171,0.7251898191604388,0.38871743059237907,
 0.4679539415355748,0.5844881830612048,0.21087474481151158,0.8346517223687027,
 0.511626969794083,0.6755305339960201,0.7782862819006803,0.35297445083582535,
 0.3904199599515594,0.1958340449674698,0.20044800910212346,0.5586254139571145,
 0.8339663485884616,0.12177342256097345,0.3260636589132001,
 0.19072007044295125,0.32215934602074325,0.9640588311383214,
 0.8642045082739542,0.29688663759176837,0.03272132404234285,
 0.8534255468060917,0.2077497744391441,0.16214574558229422,
 0.22672782570721406,0.8673023670716211,0.7397622343375156,
 0.26685441808061705,0.2273125720935002,0.9904913350696876,0.8157191180213745,
 0.2221054659301246,0.1634370591049057,0.17458882662435504,0.7694864811696903,
 0.32537920415850174,0.3673930648830068,0.4430446736358079,0.5199096055702666,
 0.9682740943216254,0.19106841899141058,0.9869958424548536,0.5626234576717772,
 0.684278962164833,0.6719035436636455,0.18717634280887996,0.8486717114367158,
 0.14710858182653863,0.334431814633392,0.9264593633079635,0.20486395194811025,
 0.9179823234211132,0.5194440670646159,0.23087932645875042,0.7001446058917234,
 0.8120968623738787,0.1880018205018803,0.8626757909788536,0.35277955426164165,
 0.7026000898859509,0.7576340837641736,0.59226073285305,0.4560934623171198,
 0.19127918797862464,0.450272633870641,0.4157100646601406,0.2386792769379088,
 0.26545824490196746,0.7819688360099912,0.34953613089660707,
 0.17996249473880988,0.8285714662619433,0.8660753944944631,0.5280960297300606,
 0.9620839767196446,0.8689860890093366,0.33468878406714797,0.9334854822757769,
 0.0024816909718234914,0.5610458811446054,0.4431935021389404,
 0.17708315767768168,0.4298352799551478,0.3524814062992596,0.7904126106041095,
 0.8075401098209931,0.7137338312549105,0.38924816232171755,0.2632579319028594,
 0.26310969039512844,0.3819236319244004,0.7985530365559539,0.8566725302817193,
 0.713305622617737,0.21060809241931855,0.6619072849055616,0.9679648513016703,
 0.09113248307125832,0.7420863147571038,0.8922641697162115,0.9168425878497802,
 0.578359511162397,0.4612308375153188,0.1758011695218583,0.23865352828452135,
 0.3806368265862651,0.004510591648672957,0.658363607861605,0.8088420797766175,
 0.4811112982903577,0.11931996926507893,0.35417923649231586,
 0.40008157164533964,0.9792201712165933,0.8103227365708867,
 0.26788077770514473,0.5958763742254807,0.2936090163158055,0.53514063421953,
 0.8856060226744578,0.8915473051782951,0.8952699338619194,0.5580980008518089,
 0.9326341734546305,0.5623316354911134,0.6747378854470498,0.6837516556020676,
 0.875142148743551,0.17022351388116896,0.2916802291147579,0.8800393647167312,
 0.8766664056882324,0.5537842088826429,0.8805234336723682,0.9396160519650077,
 0.16716325290930734,0.9285726036557901,0.5136080935892849,
 0.17074610806178148,0.157700667085377,0.9209793146856223,0.3851577345130983,
 0.052715170201668715,0.25159411576547464,0.09711099055955985,
 0.25316753182033724,0.32320900094859484,0.875115480217736,0.9196420118676699,
 0.4828451083322164,0.8728514671334443,0.09335443214045613,
 0.009290286641120638,0.8142210723130534,0.24688677955026606,
 0.06704700674797048,0.29447056144211026,0.7168086560816005,
 0.7323840777214796,0.844900708782657,0.6204823233123099,0.6360385833463686,
 0.7209315664755392,0.7034492392602005,0.44977632744529883,0.3227146336689717,
 0.33724185409918306,0.0063209103376576925,0.7898618917155468,
 0.8419208310958074,0.44962977776897917,0.4198227854586545,0.8696611939369725,
 0.4952764789565065,0.5427295358994828,0.6097675218188059,0.7069869343959428,
 0.82309512956427,0.5758430283494964,0.7204708149451327,0.8506796323158108,
 0.2324227693041684,0.7338637649488094,0.05885715859913143,0.6231399207506777,
 0.6600525649434362,0.3474721264694104,0.37959652078481365,0.7044556170775281,
 0.06851333102587787,0.4245765800681295,0.4686200291598045,0.9360662564236465,
 0.1549474913490253,0.856731367494798,0.15161162276582207,0.47978917392902887,
 0.07397608174539894,0.9176127219363895,0.9037917309291323,0.9757326442607605,
 0.21487930790847942,0.0695016574321955,0.5652317953622167,
 0.19820396361970316,0.46396410408313493,0.15210018203866615,
 0.6926666999355513,0.6901042372220945,0.534526294504545,0.2396724503078812,
 0.9702659713849291,0.8325695310288557,0.7142802336279528,0.9610344251674194,
 0.4220318522549761,0.7187252951239913,0.19566866665117089,0.6533204888950623,
 0.5072907742603611,0.0018696506718728356,0.040128424151253395,
 0.34504935856503016,0.6813320046430106,0.7117857784613324,
 9.95404795939625e-4,0.6769798049034316,0.6475950143675302,0.7947853999067509,
 0.41110590334773656,0.8155750275200517,0.3276291263514558,0.6625119446290465,
 0.4825278736750107,0.08655559019976944,0.9415611826298604,0.7295332543098146,
 0.06607161496850067,0.19342538342714333,0.8933480065853463,
 0.8392638087710812,0.32795302793534775,0.5125968726022035,0.5782008382079263,
 0.9156259933230113,0.793542663684857,0.09640549050238634,0.753592492879437,
 0.42764705133587766,0.3194988232882825,0.6247843627055895,0.1013606907076634,
 0.5375496687766683,0.32707874326480535,0.27944661182481334,
 0.37742240522891823,0.4885398878370363,0.6997395126545496,
 0.34159765216622384,0.5215035419780122,0.41682989212693455,
 0.9015717281848565,0.4325506834992453,0.41062712890878394,
 0.05304919299542299,0.24661882327297457,0.5629217280258316,
 0.9021254173124015,0.7067919229154145,0.45890677151121617,0.5122689839412542,
 0.38734739249859784,0.8245176785300525,0.7693933067062377,
 0.25365723667757323,0.5883004805870086,0.15616272445781043,
 0.7937952202191656,0.07889554939894772,0.22203554984113372,
 0.5157830492588859,0.35862006263299273,0.22669776300460243,
 0.3719117495792429,0.09289017238020925,0.9927345462912174,
 0.08061590024673426,0.352210397709788,0.7251134758953584,0.11953130695473357,
 0.9009792533980963,0.6220015556808206,0.6613451868935402,0.6166910583314595,
 0.8927525648553649,0.9232740323756163,0.36543386128601496,
 4.717839094868781e-4,0.699809082603375,0.649464706030359,0.24962933254758246,
 0.7926515755532794,0.023828751171973694,0.4464773794321022,
 0.5504913891399943,0.1659079403232916,0.14600862076036192,
 0.13249764925779262,0.1938927446935398,0.3432989853813381,0.6308169990118345,
 0.501510278001105,0.8098534108901944,0.40094258884271883,0.5146828282662668,
 0.4000739290504406,0.09420936147700565,0.14478972917836797,
 0.5039622846084817,0.7230264948729356,0.8911160165283226,0.1409820888512111,
 0.2730580842315582,0.8441798756695563,0.4359158903071043,0.07330121171442905,
 0.4336084157186515,0.09808630871623403,0.4609795351952587,0.1712795013782189,
 0.6173594338241593,0.12496795882277012,0.3739958156199201,0.9870103444781226,
 0.6251969534890928,0.9384755614433007,0.12426325679676509,
 0.11123909397269616,0.8766885425054357,0.4627594760846552,
 0.20367027572684393,0.004427588353471901,0.9091458855346171,
 0.7380716986563517,0.7299974861746771,0.9543726492038703,0.6016010389194346,
 0.47505063922421564,0.0026270800032528463,0.2544363321133576,
 0.1958319757234498,0.44861922036897284,0.7936730763546425,0.4498491230776738,
 0.8106006405941927,0.462879238678749,0.7260660085677857,0.36808453086160786,
 0.6687868008619074,0.5203938955299177,0.4985846479027065,0.7989140452420356,
 0.68857019361937,0.4630243627913637,0.5362589556360067,0.7718398044407512,
 0.9947868999531077,0.16083673103293028,0.2891058539934078,
 0.48134776303683935,0.2949406332758886,0.2556882653540784,0.5038039418989158,
 0.4697818878206691,0.3246340545557116,0.19539973114418208,0.6023889133314613,
 0.6262332974644936,0.8330577118420692,0.20165171330099574,0.4272431928800109,
 0.7402646729114446,0.3880317532561408,0.4861630341189347,0.09254517674840634,
 0.09696460858422762,0.5538879537324564,0.5242210624918272,
 0.12837296209188676,0.36317024649725416,0.3473439254017767,
 0.9035729348206394,0.42563682892499644,0.8327330275752011,0.9307606626209115,
 0.1728569311885888,0.10831184717981901,0.6988817095451592,0.7242127731359717,
 0.253345256367005,0.8643705253547516,0.6258705085192396,0.9889945679372152,
 0.13234212726471484,0.7247700017129931,0.706892027554701,0.5021162952197218,
 0.21190024192945034,0.7867328872206539,0.7183962898099893,0.624183082913535,
 0.5501918765919818,0.526801468374832,0.25809119034938777,0.8360733203608646,
 0.03691899711799074,0.31430861569421586,0.28560036518252296,
 0.633420363672959,0.30961017639000465,0.732846517561669,0.775808052798516,
 0.6840591354860346,0.7282934656146834,0.019254590316035713,
 0.9465814416583342,0.1620429569311601,0.16445245596979663,
 0.21232763505245256,0.8199616786524133,0.22241499896351125,
 0.5935828155025544,0.8329362883070207,0.3442488829832231,0.4260902693362426,
 0.36418651685245984,0.7032912518818186,0.1744243679405041,
 0.47243310079426526,0.789446099471931,0.9160194933625588,0.4103078330207348,
 0.9102541274544003,0.663491934458572,0.0613933606528303,0.29135046148052957,
 0.24216429063315204,0.3952340592171606,0.2559481298188031,0.7722309645090725,
 0.5443398028400814,0.4590352523209932,0.8225089162737929,0.16805734231707148,
 0.4430959600731289,0.8649871186678467,0.8537918542017262,0.0785540128766592,
 0.7122821014293484,0.10848169115222972,0.8853510079413942,
 0.20216081142355447,0.8155834181298115,0.11788221798604603,
 0.19489972493610286,0.572825498524972,0.6476153463738855,0.9870329553037658,
 0.8460864793493011,0.8204164870318335,0.9840304368200719,0.2336904606365312,
 0.28131832586623207,0.21782631265297225,0.8451149543988443,
 0.7868392388361946,0.5054798611398135,0.5807490586106109,0.43442176490238327,
 0.006143880807785829,0.8360961916968646,0.2947728049634437,
 0.4693877023782863,0.7671935191220403,0.9532356428851327,0.6450189109963927,
 0.7116196059817845,0.7259196194781039,0.726614767747517,0.8063732118825365,
 0.48480428358165206,0.12009562639552007,0.526712048260873,0.9107420500990182,
 0.492941872254961,0.41122687866445307,0.5216965042259112,0.6146919177695029,
 0.09376691076947297,0.2925705595762843,0.4662588975587313,0.9668640685693983,
 0.15061636604037665,0.120631458997559,0.6280900070529657,0.7037140166956882,
 0.4642790692877652,0.6361861171493066,0.39306449141902255,
 0.46957396025253084,0.8946197722141244,0.3737742310648585,
 0.16429515197711764,0.25672615575689206,0.9867792851586398,
 0.12302489959331864,0.24205918997431275,0.6292510260547165,
 0.5126959003561244,0.7380285189543726,0.0656221886954419,0.9420058604188403,
 0.8232650305744293,0.22254510349726864,0.17315961222330833,
 0.4054386809791669,0.8929909432758629,0.5797110438940867,0.23037529102919152,
 0.7007789853428917,0.3081410126376052,0.7947152023901948,0.2723776693188589,
 0.5789175675889733,0.1574849544113347,0.6348849821156832,0.7589628312176779,
 0.8073219983863329,0.3157516102901481,0.4119199427845224,0.6898102346606749,
 0.009320564646244978,0.20144705304172617,0.9206156028251392,
 0.5980922340383219,0.6943537008362478,0.4877104476765721,0.45823047825446583,
 0.8032128948737106,0.3059686183196536,0.112316480903067,0.4191499176246608,
 0.03386347910839915,0.5784012197391889,0.26556557473582765,
 0.3675879592561966,0.12259173739867302,4.8239055295029587e-4,
 0.6554499205488332,0.4210216057053291,0.7885865817964964,0.2762622801744793,
 0.2312836671549311,0.3638994656176313,0.29851528287226103,0.7214472770871474,
 0.09040684128789389,0.033739070422923234,0.03100169401760189,
 0.4604845979143162,0.24551607378380147,0.6135489924710372,
 0.17947808246469943,0.9207898666880173,0.2409692804147311,
 0.12204675064476811,0.9839197345222855,0.8441467752758282,0.3943638530424075,
 0.6417042792671638,0.9516471285015675,0.7430633088246803,0.5875810999419239,
 0.4766940118022476,0.3380712213544289,0.22384081526613797,0.7530729772768544,
 0.8568244093840796,0.6646959731294872,0.9760822833812405,0.9407711186670341,
 0.5426241362471873,0.9900975334800751,0.31390053891438674,
 0.004523782135379739,0.7298297001087285,0.46583814890751774,
 0.6117215146639067,0.01544413786060117,0.3106767531741771,0.7572915811908895,
 0.2936442687610017,0.6174329333590073,0.5399672591878484,0.44448754917897726,
 0.9158923703210377,0.20587039114494232,0.816078052050384,0.8406513915765566,
 0.8267180932545628,0.9418155265558926,0.08027946040991907,0.5551038372180653,
 0.044881226088378745,0.44974156899326956,0.6526068174512434,
 0.05698969494163397,0.96596301730151,0.8642732635499901,0.9981080085107615,
 0.8492273111673465,0.1275371444500054,0.8972166493002867,0.5531211321899614,
 0.8822007521092556,0.1805092116560787,0.2870244223520362,0.6353482003209152,
 0.9843216038384445,0.05556868835024176,0.5369520434918039,0.4770325543687802,
 0.27939456452253797,0.8870719473568885,0.677629326447479,0.32476062397548333,
 0.20038143025577482,0.5352336453044942,0.01946286499831129,
 0.7603522657097849,0.07546222340273501,0.4653039198724087,0.1930882857612659,
 0.037480321870957845,0.3186190370486036,0.9268464931887712,
 0.28179501420395203,0.4519420280921822,0.6971595961836081,0.5489031102084572,
 0.5776326497105714,0.8611816230489823,0.7047519169745812,0.10233082089913992,
 0.5323973173278976,0.9124915967644651,0.37203365797358356,0.5408861745513123,
 0.772906906790711,0.60103475267054,0.21874822308599073,0.7519642016467796,
 0.33538174690101097,0.710366242492676,0.8407451008516299,0.3979665241578785,
 0.5831306116145745,0.6632084482186253,0.7894158586222422,0.33518444915571577,
 0.5047872321225658,0.44511779331850054,0.5250234311439952,0.717683413130759,
 0.8825409860928602,0.07181474277661093,0.5760897647059866,
 0.059530108140402715,0.2262886474989192,0.982331798151863,
 0.021306745225216694,0.5819494845997397,0.5430037887579643,
 0.051660625323055154,0.14760463400831303,0.5236393883141514,
 0.8383959716298717,0.41379232446918024,0.040106058881755224,
 0.4307095150197511,0.45767807778303493,0.8478004436678785,0.6117108207669721,
 0.3791801888675468,0.3519287185647989,0.38071908989973957,0.3549294817292271,
 0.8183525641442219,0.5643390713081449,0.7795578811460268,0.8467980535520134,
 0.3556525430201639,0.9522815655528084,0.6309068918985458,0.6770890871490882,
 0.01764968469164052,0.6566000120538786,0.9117635654956836,0.5374072618019263,
 0.2694076261094721,0.5476034063131027,0.8571879757339587,0.8401067915034844,
 0.6727607119467636,0.8005603121420383,0.6772710978994722,0.6372096137001396,
 0.08946590840240143,0.8672835556402798,0.692755959140164,0.5290118832909411,
 0.3636955139444311,0.44506642506974825,0.023920885192198815,
 0.17785478191923254,0.7818602095378783,0.4138791808203769,
 0.11165980089372951,0.41760553765815756,0.28862951371949286,
 0.10063220828197117,0.4236034297596736,0.2065864528510638,
 0.44894516350814295,0.7136018853315146,0.6252631001816186,
 0.36119938535987206,0.7061690901346617,0.11984997906543104,
 0.2867713153977931,0.9530311465758774,0.9281458105485034,0.35318858505524287,
 0.6246393061257788,0.16936819232603173,0.09340086336679776,
 0.44833903694145105,0.9288964456142051,0.8156451559321346,0.3816147996365026,
 0.5556470173010458,0.24869077301806686,0.8055502245688028,
 0.37406139800111937,0.5557772609713622,0.7763494004406961,0.6315670877453722,
 0.5961365054734503,0.08144104169392219,0.1319227663533521,
 0.26304629286922776,0.6582748255576282,0.4952441739490111,0.7381633073359032,
 0.5006244874315485,0.3705002696152834,0.25677588081962965,
 0.06974108096730658,0.4341720634773282,0.050754955383250366,
 0.20296714961784623,0.057135843293265154,0.30830140611499657,
 0.9846822022972799,0.41104370146881264,0.0107204345993549,
 0.35788161544954455,0.6754395300675392,0.08976552876286126,
 0.0076233011194886124,0.11669638180505304,0.6856460526906982,
 0.43023369683981805,0.8356450720999339,0.46757086160814987,
 0.22463781122422022,0.7139128027296167,0.12988294770350095,
 0.44904637529598723,0.17419098261577104,0.19062841046636458,
 0.24090539565234004,0.11594120510836703,0.07561383181867165,
 0.20057908862256513,0.6491181496657807,0.25029972271913725,
 0.9409523323331166,0.7594253295385358,0.30915853469181753,0.5407389719532716,
 0.6688103911882859,0.7655532349347135,0.28137145654345586,0.267749250204005,
 0.08282661260788216,0.8153352239551949,0.3050044644086125,0.975874439973593,
 0.9275056019852943,0.7091353846057815,0.9333834983218465,0.8482781994022932,
 0.9666053878805048,0.9731061854989025,0.6263042131202229,0.9071861444772639,
 0.502805383496394,0.7914935396367649,0.9365764276339141,0.5233816133442533,
 0.0647109575268141,0.6292847430512533,0.6400503125242487,0.8094324046497411,
 0.016307644029978308,0.8039587656281388,0.6730543691474792,
 0.6603916853455134,0.5457234486849883,0.9247171227894094,0.1371426791054341,
 0.9911660505526182,0.5390372376565915,0.10355546753975975,0.9528695918765078,
 0.12150096525540666,0.003368280874717522,0.9784124555503914,
 0.11909140334561474,0.11074438476282655,0.6528142288720619,
 0.29755643048699165,0.8649942716474657,0.6066079904495766,0.5846075647409091,
 0.4977220730805382,0.5721167574235667,0.9878443785987723,0.7196820809020198,
 0.8576440096672806,0.5456092119317091,0.6453724838158532,0.4589362849035534,
 0.8866926345402115,0.523922785676485,0.13111921279413008,0.6658376498937629,
 0.6736833249529448,0.8982255434017024,0.19757504615506716,
 0.26170184544729136,0.9469253488532383,0.5307172854752954,0.4265363215460225,
 0.8104469598066045,0.4286439615090135,0.4943593270533505,0.4419126251285377,
 0.2244999247217624,0.8333764171763082,0.17590989398514456,0.5950393086708086,
 0.12187715928203169,0.6224833884630375,0.3654160114870326,0.6150351990775429,
 0.881848634301813,0.1511206248303658,0.5121877906180297,0.810644675794834,
 0.4790736488242384,0.5578238951474013,0.15878112748705142,0.7566627351009639,
 0.34045913733982747,0.8214859655979937,0.2946125866460694,0.6083656965390737,
 0.4214672940405968,0.8528450549936741,0.5053467037204574,0.20883955760734219,
 0.9378498199099142,0.1157612419223345,0.8032599622244027,0.30264711770689945,
 0.6661435862429251,0.8788513066780483,0.010563790698251307,
 0.09250913504983838,0.056930733741373696,0.3522350480080212,0.19632487195158,
 0.09782756626359601,0.4177044682667532,0.9876365765526068,
 0.22641397031164634,0.955552090034714,0.37903698841024036,0.5671530271469247,
 0.6361668864293761,0.23105251857987452,0.3162253236118052,
 0.04516969537505522,0.4990976654661212,0.7287850749861449,0.5558706966766005,
 0.8833099841929087,0.9245849150264094,0.640602154862283,0.9366363747833251,
 0.8730162085054078,0.5850146141790467,0.12273440921888557,0.8577293734859376,
 0.9172325364653835,0.17934233808086042,0.5872597348975663,
 0.18683936092152154,0.73917694501829,0.6287229685166978,0.4870452142558108,
 0.5106642917633065,0.13740570536367824,0.1392344052000627,0.9309529432835894,
 0.31549249939584156,0.5682268331076783,0.0023813877819873763,
 0.9605584130315756,0.5038414892591265,0.5224055161407966,0.3821702390990136,
 0.03573855733527376,0.2998338732752027,0.20249737883384922,
 0.4389686862074418,0.9495482562017217,0.6300882606801744,0.8157862660635491,
 0.30378245004619775,0.550347992308989,0.6571703101825518,0.26082630359391323,
 0.33539742716587573,0.5001221558911623,0.043800111424784904,
 0.7358483232525285,0.5534803109741878,0.9341936050576088,0.6740243839843949,
 0.6238049982750717,0.26930788485957136,0.1858860365299272,0.6373880905159197,
 0.049769780753430526,0.14118609435461726,0.43516147300906327,
 0.6380222492875594,0.965856965308779,0.42951806497105105,0.04065591803373059,
 0.2502555785366226,0.14965408738186303,0.5358746317440317,0.6225393260362431,
 0.0969768846738347,0.25587660338114293,0.8881781213968957,0.3656753960827128,
 0.6349222877963472,0.6625760965507125,0.1324409916250655,0.6964076579863451,
 0.16821417622852497,0.23050682395505018,0.07762659431325025,
 0.8575959162914969,0.2550609378626919,0.07209517132691967,
 0.10687166714806162,0.6640488928232876,0.4188771371836677,0.5025226183393298,
 0.9171179235983148,0.2975574545002718,0.9440022324043006,0.9502812816685027,
 0.1927043932605227,0.650189378283559,0.3913986213562054,0.47256371248474216,
 0.017765120006150936,0.27436587016024505,0.9581344805366396,
 0.056332980796060084,0.42424638586420405,0.13877678220748435,
 0.6641243994920101,0.9557163695546494,0.981782654025719,0.8651785492949382,
 0.2839813323025563,0.002719950811082561,0.9126330110353577,
 0.7961214826878724,0.6606156710124013,0.5232503676189907,0.22610258453915866,
 0.6144144692107674,0.30920816262267803,0.6980223611761467,
 0.0010624897312871973,0.69984477506892,0.032023775057256065,
 0.09342151170456525,0.24081946436494583,0.6131054358825824,
 0.3386347881224111,0.4387890850626984,0.23308697489719532,0.362658664347691,
 0.9229281950349469,0.8196455466360049,0.37292368356930217,0.826037631062027,
 0.6376845752837313,0.20139290416126765,0.4640713215331209,0.7430721225835901,
 0.6873774202725892,0.6977121841945644,0.26855596312352104,0.8052482368621869,
 0.634796537064603,0.7903853634787776,0.5819434576828957,0.1913326736975166,
 0.7732904607960209,0.7797196178776827,0.6223809953191551,0.5978265881038392,
 0.08710149161503733,0.31982748013244044,0.8189979323462738,
 0.38279442512254835,0.9935029079404879,0.05105978890491658,
 0.22835527836468794,0.44838192047767045,0.6077970469872097,
 0.6377940967384137,0.054189242146003624,0.5602152180963567,0.580165447817661,
 0.16705207164067604,0.5028469176873426,0.3745856556559545,
 0.41594676723893786,0.14725370622604028,0.9671953363572345,
 0.5336269272721221,0.5131724674886633,0.3831961690471225,0.6442872790374135,
 0.5059933809035823,0.700426120078945,0.9398581881550679,0.9692035439160396,
 0.9448113223971695,0.5026376186709642,0.040812178536292265,
 0.12226011188704466,0.2842661631970568,0.2222937712728007,0.5178204041735706,
 0.3853599974787095,0.20387923887153536,0.024295589965906084,
 0.46005886703871013,0.3224776267378243,0.10450699905426075,0.886220072300581,
 0.8649796634934402,0.35991730741476946,0.8603856999225009,0.6809185572192918,
 0.2159623131595656,0.5942857814136645,0.8283393571755822,0.18005283690765617,
 0.33702046795817275,0.3600879830513195,0.2288913035442025,0.2075451050604189,
 0.16141976653508605,0.75604676156595,0.8106809800025027,0.4365173478009323,
 0.33301585388040744,0.7313116442285678,0.0901498820820471,0.6988230079758163,
 0.18599159586880343,0.739313544409417,0.2573726717479984,0.1455759625362426,
 0.6744977382587544,0.2563204631645547,0.6479563823201709,0.07055613031419972,
 0.7601045341474773,0.6473159584232596,0.6240707372225599,0.6189650022313823,
 0.22814288309397646,0.5712748557031341,0.22461156271221683,
 0.8732608545120485,0.6713102801882127,0.07999239011880954,
 0.06570013654395868,0.5545255301713324,0.23216631428408008,
 0.0029378748162347534,0.3199415401396468,0.3525370671312107,
 0.8996096719342779,0.026736072856164528,0.22241683571881998,
 0.9540824881904978,0.8828658525226584,0.30954899400331826,
 0.08842177910471882,0.4857749127671157,0.796572027079395,0.9803049234694577,
 0.9869359455887843,0.0789877044201206,0.7113936278345202,0.9059495556720382,
 0.7108742604046349,0.2873274337073659,0.9343482956720179,0.09216468308368919,
 0.8880223868557742,0.06574058214577283,0.3095432002631183,0.9746284928713236,
 0.09243841490696691,0.4428791779828808,0.9069752213604498,0.3986674098288664,
 0.84476801730247,0.48920321881541484,0.9329292137917446,0.8860616337631989,
 0.732148177029636,0.6769175080006027,0.8485293997473551,0.8146897555465525,
 0.8351319877318955,0.089313621117736,0.9676625715985082,0.7253829451138718,
 0.5928075997564772,0.6313238335553334,0.31162735965236177,0.9474872063298244,
 0.35745530132919046,0.4799460418585274,0.6016145287535803,
 0.23617295362520951,0.7442284505015485,0.39998456416986716,
 0.6715532575136356,0.03227952989675731,0.20230092855691373,
 0.4266395870634204,0.19888361256005638,0.7276039798706435,0.5400859493239698,
 0.8592718394296812,0.702267125110386,0.3526659309401966,0.302644768545011,
 0.8522846719374464,0.16928460670746215,0.5374912858230576,0.4433677965007744,
 0.011499879205681407,0.6609554840095273,0.12117327696128222,
 0.7750599814525365,0.34495981148410626,0.09772590169049022,
 0.04838858428825865,0.06617368932968892,0.46986365606652913,
 0.6522519055940826,0.40536101704792726,0.2386522548761365,0.5847075390966145,
 0.36204339995729273,0.7857067447740925,0.9039932426747065,0.8127481761909661,
 0.06581339373079398,0.1417281902888432,0.9809281068067798,
 0.46472597937016435,0.7780075192494672,0.8967614704226982,0.2807377473985091,
 0.8500372705004862,0.1775155455969789,0.9005367622838015,0.6792044513703801,
 0.9273525487901582,0.28634364206519425,0.12835142016576528,0.853131918735118,
 0.038364588076004735,0.612735345399435,0.6214010282320496,
 0.20119877378645734,0.03313274203973948,0.8446200280512897,
 0.7672621934969799,0.30253441214771726,0.9018790027646467,0.7398278560940742,
 0.4680372340071566,0.72544624420032,0.7797316405303213,0.39717509897404435,
 0.46923218305822,0.8819706423700977,0.6874437570394982,0.5669873311542949,
 0.33286555740657886,0.8480599824967838,0.704271362351618,0.7473166783986226,
 0.22127115945851195,0.23598186528070608,0.7784932398413654,
 0.8807556574754615,0.3210749322516243,0.9302430531884524,0.876411786618151,
 0.017862409504574073,0.5980472669778392,0.46255924076957855,
 0.45039130066032873,0.31702302958195094,0.732166302203429,
 0.09060103151615273,0.24799229681718482,0.686100373966506,0.6163906006169897,
 0.8478332744889248,0.7403054249068255,0.3830784328990827,0.11064917256565665,
 0.37973496467955503,0.7058463607128913,0.558404671334538,0.30852220056976487,
 0.7444386524609166,0.08033536475024794,0.9101476055404236,0.9539250843182598,
 0.3825304851980561,0.5734342676639383,0.7621729371967674,0.6580123786791447,
 0.10133233981211265,0.06388126058439503,0.3302370438361558,
 0.023297950828107594,0.002354774607157273,0.5312732534464294,
 0.37311450719142525,0.061346517814692136,0.3759069116076318,
 0.3525247153711949,0.9248149473385288,0.21885611836542418,0.7787579229869162,
 0.27525285722606707,0.25396907909940314,0.8077543736287953,
 0.5348112092475695,0.11732688718030027,0.4083550031920735,
 0.16731857996112476,0.6616945118540873,0.9346358983610552,0.9136267540308307,
 0.8143492869752095,0.822958674954155,0.38963132250171995,0.6928214013547064,
 0.05091697437463383,0.3808443016896087,0.3682687284062335,
 0.20226547460090127,0.17967974129356057,0.25926387048182054,
 0.42445606086637266,0.1488922233417389,0.6776186206107657,0.8460649140841926,
 0.7119815925560393,0.5656031542921761,0.6708951384056252,0.7477304797316994,
 0.3330674341553118,0.3080979285121944,0.0044589136035301236,
 0.9673621909517036,0.8251993694896891,0.8967921564317777,0.4186111085735712,
 0.9585306460926937,0.0776653463091661,0.909541917810443,0.8619790223253766,
 0.41727254896225907,0.8659933007439725,0.5194708718966151,
 0.018836191571532823,0.40443673480472697,0.44192627297413334,
 0.8583608023685303,0.6310716952989123,0.12248648341102031,
 0.11929588289213788,0.3016573487478016,0.98025442680785,0.19487382266465492,
 0.9634967384557983,0.2515561340652902,0.7852939278791652,0.6344563606128418,
 0.30881325489232236,0.17290958938610446,0.2234044661205099,
 0.7043469789322353,0.5253250321820586,0.2822002092741628,0.36226583446223115,
 0.34745888960429705,0.9293488851995602,0.09704640079897608,
 0.33266073665116047,0.8062880155446623,0.7100062947444855,0.1744290032974568,
 0.4962877438455461,0.757825908801466,0.5465022118519004,0.7803885058254266,
 0.9484536281692192,0.8399796743698008,0.3178764161919875,0.1719764677269734,
 0.6510184351237093,0.19869033886706033,0.9178967376119814,0.701891154817791,
 0.6726902284906012,0.1059446990706352,0.8792027367918779,0.05204575895082175,
 0.22992917013620207,0.13701195342710726,0.13035012443653782,
 0.2138896247272234,0.11551204440005947,0.10033535427073437,
 0.5981871873509199,0.1335731869131319,0.1333732235126479,0.3185674899127017,
 0.24375919170164595,0.20447513368115966,0.17523296616555784,
 0.09808486540208672,0.7701146021318086,0.5182622924404268,0.938555371655628,
 0.7818117441111967,0.08988521179339959,0.9519751000792738,0.9183845786149827,
 0.7747669311012744,0.8458476034094145,0.9610398543169194,0.13473966618027355,
 0.3617213970622153,0.4312687155607713,0.8342668219578554,0.43355674103881514,
 0.19148604820021764,0.16248573340324934,0.7002800854503868,
 0.034748726501976135,0.9889531951630764,0.5886019198241662,
 0.052343739231383135,0.16796402767923235,0.7107491405667578,
 0.7801921811118118,0.43564835729098783,0.3193965966307275,
 0.06210624905703921,0.4864333577604122,0.2474649022006551,0.9416724257616641,
 0.43614662344150457,0.9401514292141998,0.2939362155220987,
 0.27902119467887476,0.6904784075932953,0.5006325720133905,0.644216660319908,
 0.7445220050667147,0.6446323914579738,0.6508555428485683,
 0.006163994199214984,0.8255578691811004,0.7666236359088139,0.13626713417004,
 0.7218769401233096,0.06326037153802078,0.129512520274915,0.13164641348543804,
 0.28972178185438935,0.24552304495548893,0.7538557955536322,
 0.28912760300513185,0.4085462656923362,0.7427688785685279,0.8465418577975334,
 0.5943059049435517,0.6519278485096303,0.04594719431283467,0.6931393735087922,
 0.9129609183758677,0.8133637338984523,0.52734350407219,0.9598258554275934,
 0.7116717588701705,0.8604145253965924,0.3075962938194752,0.40996779872925926,
 0.5783996605032642,0.35085656874323856,0.7573133697559223,0.7054665760028269,
 0.14644323742244614,0.4824267745640136,0.7451161635217687,0.0335568882948708,
 0.5569614906186698,0.8375488426910258,0.20258307237617146,0.6035910089834883,
 0.25484162201392546,0.21291005934820728,0.07573493832278333,
 0.41023696458344194,0.28326123930525793,0.7494830082766328,
 0.9838554815310735,0.5238314382165239,0.8607042134067175,0.662516272980486,
 0.3668489384235486,0.34244934052087084,0.5643884203285561,0.9295326822206601,
 0.37703399027850515,0.36898352631843445,0.016350527663928083,
 0.4527351489816476,0.6395451219506866,0.3644157323016892,0.36391447080305817,
 0.7362470731566043,0.21886466930560466,0.030265775424492047,
 0.5681723149193175,0.6154556016935603,0.2143463495455773,0.5867565591791171,
 0.8289512076731332,0.1224405542574748,0.9631744846323257,0.1284111864934655,
 0.7741612222060337,0.25743700603624864,0.2770802845168203,0.8285300297417584,
 0.2498172497497253,0.15350914206060895,0.41191253065738653,
 0.5107204385654478,0.03085139143605453,0.5093223578550021,0.9513417320462061,
 0.70346011160359,0.1845763353723976,0.5704300011112363,0.12503697776113198,
 0.3651646330417744,0.7432396671683668,0.17690410857646288,0.80964375144819,
 0.4743199406769498,0.9291248827780741,0.34028601738570974,0.7648289794853571,
 0.6842578986588783,0.9671991376794024,0.3883729272696915,0.6868278474978371,
 0.524153196588497,0.24338530001094194,0.5627671102029892,0.08634831720546665,
 0.6794771263019275,0.04186454029536679,0.9677409460287805,0.2812699389344768,
 0.431669190810278,0.07345551735500522,0.024500597978657668,
 0.4577694599997304,0.01661057807774724,0.7095007219344756,0.824580377185664,
 0.7725527454397345,0.1721862357634938,0.8472150507065545,0.9845369332474952,
 0.15136926977214815,0.42844328949313326,0.7267089533574739,
 0.010707508793822651,0.03910655725903012,0.9775833425030722,
 0.9603055181932402,0.9820268916746109,0.8229588506516301,0.8180067350351748,
 0.003030687885260952,0.3034168951056241,0.2717529178914818,0.526951564745301,
 0.05208219606841602,0.4043776441073229,0.30818029093098276,
 0.10118546140776297,0.8824985144913353,0.02060766185506102,0.320159636397807,
 0.2826458849809008,0.41344781300601996,0.36761513414209224,0.851772504231384,
 0.8053102255621204,0.8388962820421165,0.500785282915931,0.4909450252920845,
 0.984837273272928,0.9410168489976882,0.04385549842702685,0.8965865293313732,
 0.5834297929163565,0.47577176695510937,0.9625241263873249,
 0.42265226872838113,0.8594779105233493,0.37854393434080036,
 0.7682184819391555,0.970884580198542,0.18972222247070591,
 0.023977632485295275,0.525713943035078,0.5665338452625899,
 0.04224016882759385,0.9216631781451697,0.25331291003626066,
 0.7291729969638601,0.36610222708449613,0.36797273251017637,0.824335709021343,
 0.43846831138976083,0.7459508945289626,0.7298034496367587,
 0.39518213508877953,0.28667901084555336,0.5377586002126925,
 0.9991417859570284,0.4865806303591207,0.7033204201910075,0.10030099423709382,
 0.9739657552459463,0.47181174161406336,0.1833292627655848,
 0.31976799142466095,0.11940694652161987,0.0648715131027715,
 0.9316865684766981,0.2406632891419238,0.16802322182861307,0.5223318786374769,
 0.5880557602168089,0.1228145397010062,0.84837754385802,0.749869497204596,
 0.044219266326148476,0.08334962769378651,0.05253834303616833,
 0.8520398732480159,0.5068579467641362,0.118491855739698,0.20997040336567685,
 0.7825542845499467,0.8172223397257355,0.17324508429508145,0.7048102872889627,
 0.054641461216817455,0.05250200436199648,0.4512139218317911,
 0.7275460658537802,0.36818383969735247,0.2014838807799707,
 0.14188881654739638,0.029495222943869304,0.4758662311989331,
 0.21684000436103046,0.20553255602335485,0.7770344219336127,0.36072422752737,
 0.9197105561076085,0.9263125597531523,0.2647948708251545,0.7374338830773569,
 0.324868746291439,0.30306661037205285,0.9692974015348244,0.35643496464324964,
 0.18632721009699915,0.19798742228991129,0.7367548863963275,
 0.5677703533517445,0.5371085585353977,0.20785079088724534,0.9128394149210636,
 0.09674358952839401,0.7215038919823897,0.27586116766819424,
 0.11872459636420518,0.20885302724406196,0.8760193005716247,
 0.11679377001398639,0.13265597719603772,0.13369600880465426,
 0.6824645439714585,0.38807451200543475,0.8212079450645683,
 0.48408040464287083,0.7198230509659216,0.7510775603286916,0.7417721699707934,
 0.7945308130273051,0.27644792559911335,0.49024205423478784,
 0.6829605224708054,0.06058850223808532,0.6015978666958548,0.9884465608013261,
 0.1323691708608452,0.6750457144638304,0.1395719387387785,0.23825325509540707,
 0.6073804849676515,0.9126714277238307,0.5639800321895001,0.4345273430055392,
 0.48455562128041285,0.8618052454348641,0.17850072187386956,
 0.6943361041673315,0.3500473253360179,0.3430474739690097,0.6641008034040249,
 0.05581076900838755,0.8672605028919401,0.7855357097567758,
 0.42035612890762075,0.5770931518185834,0.35364571861191085,
 0.02785018570492781,0.2187868002136557,0.9055213884842039,
 0.03385106322475695,0.8455147873016573,0.44957940274727226,
 0.3567197043625998,0.8200793628638904,0.9461300521940238,0.7313369302878374,
 0.07389870626590844,0.13560378733333578,0.07729666012797798,
 0.6430581779667887,0.09732127600160023,0.1827863376384562,0.8349041825256949,
 0.9094963274087392,0.6663105185870001,0.5239122519177339,0.08533567082222926,
 0.8940869423002128,0.7327097829268817,0.04727255743168013,0.1623879762631809,
 0.9096452325491527,0.6873008696371214,0.6922961812979337,0.6696230396496538,
 0.8939258942612065,0.6478546928219046,0.9443286178807917,0.49470384223774566,
 0.9546375736014072,0.7260967828548935,0.06404520554725812,
 0.24639792651406778,0.06266904861915368,0.29640088455399893,
 0.800705174643335,0.003442786568655376,0.008618470223117347,
 0.9002605105887779,0.33819234689369737,0.21331922375709556,
 0.45196338729710317,0.444064676512673,0.1657226998072826,0.7586093504893296,
 0.9220342448672234,0.7548552532437177,0.28570509192620186,0.3569519161604654,
 0.7517735570242531,0.7233248828945011,0.22736290890231836,0.6605898871598179,
 0.8522896763083617,0.28275531472336035,0.7192979033780154,0.9697929001684559,
 0.20870562571072038,0.5334126897623505,0.8725703637102266,0.8102551209817627,
 0.915687932551748,0.8921455180842988,0.6969038846728757,0.9585617986938542,
 0.8302306531098375,0.6305170850427814,0.37780611832347966,0.9951210791682548,
 0.3175780989597178,0.5429722700163401,0.742069265995099,0.9390770449190344,
 0.4570871803234411,0.12762273328350004,0.8930002582107811,0.9476394545872667,
 0.25531952049560225,0.004143319209485852,0.053964173552267125,
 0.720064837864953,0.7122736602427807,0.6739984387733218,0.35981040451211077,
 0.013678025979791841,0.07152328596183866,0.20884681225146584,
 0.11664244374917054,0.277143818732585,0.5802409812675444,0.4001235066923151,
 0.6260484527781454,0.5939815222047826,0.6223331841028171,0.06204952702047806,
 0.590176092225616,0.16180917423588803,0.9339368311243241,0.3055281101193418,
 0.6806954423833973,0.20423920805592977,0.7918787298509201,
 0.48300434048894025,0.7272371898436961,0.02461038215257816,
 0.9958309577487592,0.47523703060110833,0.9340730281980891,
 0.11460882638040804,0.9250728556444101,0.26968816299220644,
 0.021074289196790108,0.20710605261882242,0.39316565955277805,
 0.908229599681595,0.677748609729326,0.0574347129424182,0.2977938468147312,
 0.6177123548544928,0.4092219283566497,0.21857233067532622,0.6891407913603,
 0.630175240360706,0.8171669134395976,0.26930809587702254,0.11446623209939744,
 0.6771823690919956,0.680483376111852,0.7346111210465189,0.38387234880059684,
 0.5893679012817511,0.3341655425003197,0.5492299848428001,0.9251653457625602,
 0.8467658121772131,0.1132780427223189,0.34316316884290055,0.5645554141287081,
 0.8667293118781964,0.8591256820190685,0.920269093770602,0.7478862179355529,
 0.6866076266281884,0.27267586365919505,0.5149293078894221,0.2764796738039761,
 0.10356336560127066,0.3896638627734782,0.4181822892921835,0.4480564827463961,
 0.5444378361110306,0.6314335451966194,0.7244599964374201,0.10241231348939184,
 0.6973370941869536,0.8468888678275411,0.17652760782991794,0.5208690275769016,
 0.5732497307318279,0.9643750099634977,0.9486176159144553,0.3409881035036053,
 0.6916007884435845,0.31308112187694115,0.5936849319928269,
 0.33112317802631686,0.21355545152018207,0.4291242296554021,
 0.38456268175082753,0.34412016366409337,0.39599927564732806,
 0.4968317023780764,0.10565089678884998,0.08326790156309871,
 0.8451801520898852,0.8783763415764312,0.21247100107411976,0.7103105184122791,
 0.3740731782462883,0.9851305718648807,0.9461681653000757,0.17239910360169308,
 0.07537125023248159,0.3352009198591357,0.19360806829326704,
 0.8186194611508478,0.1475507693554382,0.7811734956866928,0.05427322113665922,
 0.942643304331412,0.04586988383081847,0.5342396309304294,0.20946844125619468,
 0.4434480770040087,0.7702162765012378,0.02014658170942696,
 0.29760771322757207,0.9231003393431302,0.8930214728204497,0.7697792921242137,
 0.3407285322191138,0.22219637559455052,0.5575122709142858,0.6049555398990748,
 0.3120531082651059,0.8894148612076576,0.3465424376875198,0.3067260377774037,
 0.8830533954491733,0.9812015321823205,0.11162497396755222,0.1428193182882963,
 0.48374208125071505,0.5527201298640312,0.6384038112144139,
 0.23270110461891202,0.2608336136171716,0.9701328344873105,0.6107989936232718,
 0.6828068406520744,0.3546188887463869,0.2761289583489597,0.31163070707915674,
 0.6579788189006317,0.12465740342139675,0.2716893113830374,0.6501073520617386,
 0.6197440921225428,0.6593009564453843,0.7882002331009244,0.6789428267221719,
 0.6395043779840637,0.37322284607359035,0.22048694762864518,
 0.9227284155307715,0.046997189727118416,0.4825462767760632,
 0.6108829827978863,0.3142436714075464,0.10857854651482157,
 0.38381307729960423,0.03016016355060025,0.23120807051168646,
 0.07794969632989157,0.12764759890972377,0.432153506616091,0.5410870198910345,
 0.6634046309819163,0.1967416893913989,0.6520345889745165,0.4767511272193067,
 0.11440632418934749,0.3870300241664041,0.13646698314181682,
 0.06804620663580152,0.138656520462616,0.7287063669679593,0.8886649749086434,
 0.837251343180886,0.898626947637807,0.31225898940071606,0.29561093021047324,
 0.10307272316946214,0.7603647015519133,0.6284151261108818,0.1557199971839579,
 0.5121106311294147,0.23461400447712155,0.6152954249554047,
 0.10099650162373741,0.12846299332985467,0.8921343717100033,
 0.3132455364962301,0.8770734948017921,0.47307362843353307,
 0.09927005144489576,0.1831628828139773,0.24691490984724596,
 0.5103010726908649,0.5163675729085245,0.8243146220523425,0.7919681246165523,
 0.4534175859503021,0.6443865664719123,0.9130515239666072,0.8261411251115726,
 0.2534935750330113,0.2730449103637065,0.5590015667021382,0.2944996323489746,
 0.6110466672599215,0.21857040661586868,0.7079863320235507,0.5268010136011061,
 0.8606860381938752,0.11583368300121155,0.8274943620837271,0.2339230501149926,
 0.006702637044366577,0.8451828844979914,0.8804533549475899,
 0.7383459035996949,0.08226961222603701,0.131189141687731,0.04903751986298277,
 0.9678903886940281,0.07992729777621144,0.7983678155152757,0.5356527796729306,
 0.32476441459646643,0.8072916895490978,0.40040475732315994,
 0.36541153240848434,0.050662368857538165,0.19779324857837044,
 0.5762648464660642,0.12014895824604166,0.01594596591820996,
 0.7126757574039543,0.7865419950330785,0.2511192831110668,0.5039361634566044,
 0.8954636929247877,0.5935361823068745,0.3658838533577534,0.5969484201901165,
 0.7200620902264602,0.0359515324592351,0.16840290922525947,0.996409919818137,
 0.3955772423015944,0.12723142126851816,0.09751894888195511,
 0.1915235949418782,0.7369968795717174,0.9500873605799383,0.9172120651652212,
 0.1262619944705643,0.7940222598165173,0.27165583374550617,0.6866129659937732,
 0.4121069966679043,0.7488301485654203,0.17004321024339297,0.5700509946617467,
 0.11458167960590437,0.3410111500507709,0.5548859329630266,0.7612634885340965,
 0.920731381278765,0.12854173451995898,0.8216519993944456,0.2520026601872044,
 0.8873398774511979,0.5326278566457708,0.10097203691139367,
 0.34067058666809147,0.37895943290847456,0.3641064387981888,
 0.5136742492265487,0.915088408323451,0.7825037159764303,0.5330955468984869,
 0.12288270302096349,0.985146101209108,0.8981915159253376,0.4426173881136892,
 0.2573051933865589,0.9223144372931632,0.5530904658171787,0.9777719156803196,
 0.04596019101609927,0.7112971143144061,0.7944975032217744,
 0.07501705558072735,0.37816678598335207,0.7577526641381659,
 0.43633225466547854,0.02226107271782185,0.6600180528211768,
 0.3041623449553438,0.9093127517774736,0.49329281657435664,0.4594993555269511,
 0.7361988608884583,0.9839063071805365,0.6485671675263454,0.8717508461092391,
 0.7146140893556482,0.6734213827064509,0.47569553521350527,0.3062830694386287,
 0.7428536343859762,0.9989449765748194,0.1808641765457435,0.39608092966595276,
 0.5585401710397698,0.33271857076956435,0.3801959803759859,0.6522149307431035,
 0.10139307582001733,0.771406935834671,0.5014302879977228,0.4031859416443796,
 0.8534909614045603,0.33633044754169383,0.8402058084231813,0.1955784157382775,
 0.9447014162649637,0.5066457566077218,0.10409688923017846,0.7826309829935381,
 0.6903972322837197,0.5484136175460467,0.8068315873798987,0.9772374942989855,
 0.0024552488532697536,0.6355209538589275,0.28860592436848065,
 0.6755946902265754,0.6376629469101611,0.14894487562146785,0.8260282753389593,
 0.8513939449330061,0.5431119423543007,0.15702122044178213,
 0.45713311627767717,0.8628625650399281,0.7518047541420725,0.6395818001150294,
 3.567886714332946e-4,0.13107383380065618,0.2884672434895462,
 0.7903016747771819,0.01732456063550647,0.11341037219190375,0.419681135595775,
 0.9259146303360501,0.6617475803938088,0.6480029638642935,0.23395447739294906,
 0.46931423554563956,0.9970681313642947,0.32824989652401426,
 0.12320926749390615,0.36412523408334474,0.44956643210879876,
 0.3979166768634964,0.4059129403841736,0.930503137116092,0.9326316378532534,
 0.14251339391763762,0.5137842457151732,0.5622998421164321,
 0.07372800678711655,0.7713278176996754,0.7368539703705828,0.2249328073258814,
 0.14437693982279287,0.8497186230012912,0.5001967378737593,
 0.35665557290582917,0.5802008200658125,0.6915911226324647,0.2902788844108559,
 0.38361231486720737,0.31281024080970743,0.3125551789775274,0.558091389292102,
 0.19569139480098996,0.6319666564077511,0.4271649940328145,0.4566189106190771,
 0.5712916399655348,0.6957862228557988,0.9795671017212134,0.8921311704081036,
 0.010001956438212223,0.8814834888084659,0.028446050660587363,
 0.6741195452552468,0.9132381835233186,0.4803428598788575,0.3916587648369363,
 0.3535920717673082,0.27945437852986543,0.09843302183069214,
 0.37298491038912285,0.39570141170616935,0.49464274235643213,
 0.38521500573706524,0.12512913529850112,0.9767202055885459,0.742315989046311,
 0.15470221842414744,0.5433069821318572,0.4219959950325748,0.9565208575913365,
 0.901595206240807,0.17810658768616916,0.303536142771778,0.6166008887998856,
 0.09894806688789226,0.9946586458396407,0.4668309206771447,
 0.43171580007154686,0.6039539352036362,0.5172869641248182,0.7065541742624227,
 0.752652091128882,0.10896532488422173,0.3881771343155639,0.6652157093008639,
 0.061303806204454525,0.7211154280459225,0.37939231832424636,
 0.4231538470880175,0.31262237148025596,0.913582840690185,0.900877884138077,
 0.2964845625831679,0.6248516304224853,0.7748353353611829,0.937340621785702,
 0.8261549809023885,0.26423860790232934,0.0750286244396966,
 0.06980130868086909,0.7590885200831512,0.4942600623797908,
 0.23990537505553655,0.0955580939349654,0.8515233810179826,0.669522097603868,
 0.8019387376499946,0.10194603905052779,0.4027513588363566,0.8772294268244131,
 0.9325152200491605,0.6381666160688252,0.5224692086116086,0.4932986434153237,
 0.5538474812869908,0.5860146775386229,0.45664047763313864,0.2743027144538499,
 0.7798695768142513,0.7683387277424333,0.8217082707863219,0.3478793931353662,
 0.3123734676130957,0.7190982220874571,0.10795727984771686,0.4809982612730257,
 0.9590145273173253,0.3994339749816128,0.07547729755058996,0.751279006124475,
 0.15114750752966566,0.040670282166265115,0.38776469669436864,
 0.30884806716974,0.13676833345082473,0.5831028181485756,0.9321702385274528,
 0.7504689411460584,0.24820933840198134,0.7335771316841709,0.6392232134764168,
 0.6468890422259574,0.24822709819570332,0.560161956717631,0.0750563181481183,
 0.02563050254332544,0.07736937635519314,0.47171886928754103,
 0.17364326609554415,0.10787699767001147,0.6707743523161158,
 0.1674214465548335,0.6738591338944068,0.04238592287760612,0.9383300726829771,
 0.14940788386477766,0.3475031156000238,0.6106270907063969,0.8770210683882241,
 0.2577062081926802,0.18669707487820564,0.1836609326280051,0.7155227058688725,
 0.39058404656030765,0.6413881794036936,0.5155027236212579,0.1482413654238679,
 0.5858361044657572,0.8301619073229849,0.47733903618763573,0.718655308633416,
 0.1832895796183348,0.798069641105418,0.40271429622053345,0.0609572936091517,
 0.8072655064002032,0.4282225327477165,0.6689568765693767,0.545746845724768,
 0.7477195201269449,0.7534381845427068,0.6097607424488339,0.14188435596171523,
 0.3307343432355059,0.47276884958743015,0.5935192883049627,0.6934490675800293,
 0.41667453221921,0.9496518944015286,0.058355661514209456,0.481815300781845,
 0.3396849500568666,0.5702526879374353,0.15046090718431626,0.4100632365871941,
 0.11007057582191127,0.944695309740401,0.7314543078319306,0.5525088559440503,
 0.41770489376408815,0.7722342965275825,0.29455413882356585,
 0.9513390084938443,0.4940221827606287,0.21167286049757272,0.6060732494614448,
 0.4434059315799188,0.5711603325893502,0.6391355606113538,0.6832848856342247,
 0.9952960165714363,0.25433796442986834,0.4272262140840126,0.8495913305524067,
 0.5372934212982524,0.945419931666748,0.37924892059918525,0.07033645876534522,
 0.22854595315957837,0.3974777687641389,0.29401306276907435,
 0.3052690192961517,0.4904178189623086,0.3197336285997756,0.1298823073860852,
 0.09571592144683916,0.5202792768462945,0.5328395494896794,0.5645073868801149,
 0.8006498344160696,0.3914426017068213,0.30403790621255333,
 0.06753953238121024,0.8157354410157364,0.15089281909908303,
 0.7427783067520449,0.4209305359236706,0.28724547747036144,0.5467629012611535,
 0.8573780692882923,0.6688884330224533,0.29380371628682367,
 0.05286283781902934,0.37254612125478437,0.9537164317763505,
 0.14790952872140894,0.6711175176801898,0.9466756128000622,
 0.21882472975217448,0.5409412945640831,0.325257957137141,0.9405811062535065,
 0.673502390854074,0.08236620913116544,0.8246996331905508,0.2557950385533473,
 0.4702753577122599,0.6951920795857469,0.35804476905389526,0.9385088250014639,
 0.7804189588577111,0.01645149177677996,0.1406915512402388,0.7579062431895026,
 0.8269238572831092,0.3325611949609053,0.040798264655453775,
 0.9646222379197429,0.285646527027573,0.2602358473179599,0.9258980654404712,
 0.42155765499484765,0.6642551322887943,0.2179052524846513,0.5860270354040574,
 0.906377245185193,0.619205427865884,0.6746368073737847,0.6531914904800029,
 0.029479093054723626,0.9738973381007957,0.12131150254338707,
 0.9024313604897436,0.6183129517598581,0.42327294237485624,0.6026738023682188,
 0.5317246701533291,0.08693178500587195,0.5258530951453033,0.4645631712379812,
 0.5050065201769905,0.4493067716901553,0.7606567575567348,0.64218355815969,
 0.13189202572291447,0.7419760055110918,0.6382168672847764,0.9325710446749608,
 0.16812446730596609,0.47005679363200104,0.9394442663692033,
 0.7409853106883892,0.23061016396838319,0.32595505132571323,
 0.6639964389419072,0.21727240824788663,0.5871613773673254,0.5299741834206211,
 0.07896101556983237,0.6553769097088434,0.4032490103938188,0.1936140002987694,
 0.15822582175422328,0.7089515529048176,0.6249123363270578,0.1614360333604381,
 0.6063481239643648,0.9635729978075402,0.125522545454254,0.38435124706997226,
 0.866038819002793,0.40352150913438956,0.8674090558932562,0.6255728476083686,
 0.26520602549481676,0.08995581294511179,0.1664471322653509,
 0.8880759393539166,0.3869534558213842,0.28029005860172806,0.440270021949393,
 0.9122194346544666,0.11081394068763906,0.6868126269226897,0.8353729578746885,
 0.7786207453516212,0.08549666690077085,0.9992885776591928,0.6793754227172181,
 0.2688961751267972,0.06582308119033697,0.38708952155808807,
 0.7675496101578883,0.6785700273577087,0.7898845614445276,0.7607968757516521,
 0.6753787999464265,0.7283544860535506,0.2672266844201684,0.4481052356767111,
 0.2920268498778502,0.03588923098539176,0.9071426106737743,0.656717799729537,
 0.9909405517357478,0.1531909517657719,0.5662286475244902,0.7214584285491671,
 0.1564209165107623,0.21408577042236832,0.014840301950425294,
 0.9931016906291319,0.5476761840233785,0.8685204376709867,0.7158149349566523,
 0.04939488011451554,0.5107620551928027,0.8838279891023163,0.6020372635539377,
 0.14452128345717397,0.9404414062855034,0.10712781603180477,
 0.5160799615621516,0.6530110149089564,0.17526011689446774,
 0.42893257086890413,0.42187309091136005,0.9429131111566265,
 0.5582474589053423,0.6060020473300269,0.23353044056144434,0.7332502507852234,
 0.40439422643799416,0.5732249167236303,0.41570406484872313,0.578304810704201,
 0.29523711447954737,0.21312457998660939,0.11136291519052155,
 0.13999990855421762,0.9502519998347251,0.7177702321832653,0.1863844096508427,
 0.9662798985282712,0.9762136530748208,0.2115164577241062,0.24525140241273458,
 0.7360783152508701,0.4421498874778578,0.27919474434151914,0.577742281075521,
 0.9045447977148056,0.3991717666747714,0.13929470075212924,0.7191443157182772,
 0.7851453223830054,0.025669117420402232,0.6182435880068771,
 0.8913549426031243,0.4210398960724486,0.5981249822035678,0.967158861818701,
 0.4470489661629111,0.5595607574857695,0.729475023861377,0.8863012295710696,
 0.07301013013021374,0.9134343294314724,0.26206208300565237,
 0.27693135534510804,0.04968780494251446,0.9714198353659268,
 0.6677245430180604,0.979401689995664,0.7779510650629993,0.7899015899567274,
 0.40383951571126053,0.3210761731887829,0.6344129788545452,
 0.40786001943310457,0.2681192748638038,0.8380780482110155,0.2305078173497649,
 0.25525565141660966,0.6786268020106779,0.42063327138565765,
 0.03834790777296693,0.8472346099074985,0.8196891825488332,0.6488542776941661,
 0.9390651979427976,0.5681408220928956,0.19185798906080698,0.9920084515370711,
 0.6603292274811935,0.964740053310611,0.9770154982695547,0.308965256875642,
 0.7374476721655161,0.12631039302181635,0.6412590591178126,
 0.03675140743604133,0.3124786603597237,0.6769707175038101,0.7433144905996188,
 0.14721315785729672,0.025659415646721917,0.6485861909489857,
 0.9718572062021036,0.29059976109212604,0.6026990364252411,
 0.08462255271642549,0.0790658724771971,0.8013472589539862,0.3061233242465097,
 0.9155521266512204,0.038333558915403354,0.9874099034600685,
 0.44683518904088604,0.5818103360110082,0.03517324307686509,
 0.5488385775454976,0.5951554636703479,0.7698026175121185,0.14676610732551576,
 0.5368174378567869,0.4437167558757379,0.8372742872021508,0.8332685657597848,
 0.32634451246444796,0.9142718544065951,0.7281349070230195,0.9943105309383898,
 0.09105243587131107,0.7684192323798191,0.04941996419953609,
 0.25359185167867615,0.08897354976896721,0.53702347171216,0.36978198830335407,
 0.9908674782899933,0.6042817564826899,0.43378902179951095,
 0.46863769491337737,0.2817906378527215,0.5635509610843572,0.9177278095802339,
 0.16003622196587453,0.8853492875795917,0.15013571235547296,
 0.10531157490235543,0.43621001749955646,0.34667302163482894,
 0.5636247294943268,0.7329072353250381,0.598228200328573,0.592251711217336,
 0.2920256415306103,0.2661033319846985,0.5446247928343112,0.21664928652564686,
 0.814889597214974,0.08755115581596662,0.9585200878277773,0.25181312700816605,
 0.8037065233499641,0.7810026595904448,0.7523092784605947,0.8908766481836015,
 0.9135961955073661,0.9595356823237422,0.5369442111962722,0.9005728341690081,
 0.6002272015403808,0.06356316457187405,0.5632503870038722,
 0.19304795600359537,0.22846900281375682,0.5165483370064922,
 0.5026823561417834,0.36685648607662447,0.3700788005417358,0.5617703614133475,
 0.5297826199165534,0.8723826948656879,0.8412577794214843,0.8107314230579123,
 0.873856715049478,0.2862017539051187,0.23858079107000796,0.14688046813761535,
 0.5171502493605056,0.07741170208561488,0.2418420789370418,0.8442798141127833,
 0.667901446819579,0.47396076340465787,0.8145651151229674,0.1841084303727365,
 0.20342414972787348,0.48391007532998453,0.07182237990679219,
 0.17244849125713568,0.06427439921871636,0.7670716854623406,
 0.2965439630051573,0.44870679548371406,0.5855110604647039,0.6290653347971022,
 0.9628283262138991,0.7873885820246069,0.11297949286609388,0.80986028838331,
 0.05267961666230825,0.33019197511874965,0.7012658197867391,
 0.5648099711199632,0.3715944531560508,0.41475919171391706,0.5865333606603947,
 0.3744290851428794,0.7709735878929086,0.13707139216248576,
 0.048189936367010744,0.9307142404910315,0.7979557349760854,0.685910509864756,
 0.7959777891222375,0.36974060693511646,0.5104749711713988,
 0.05241467114816356,0.1729783021761424,0.9052989614366961,
 0.010812164646613276,0.025856100701472706,0.06460587462862677,
 0.9803766128492235,0.3267406565976567,0.0023040495334324396,
 0.8727317359742921,0.774760802298476,0.48180722984803304,0.2133374319408443,
 0.7065150838008616,0.6871020899464807,0.560438580441289,0.5357447934948718,
 0.7689958378027186,0.03662823602567167,0.7859491301266618,
 0.39064361479308374,0.27278590696415006,0.5617689740999314,
 0.8824083644813477,0.5856358445672967,0.3550918834582182,0.7106938917889886,
 0.48736066652992427,0.830236794661336,0.7533179290040726,0.36159750501318655,
 0.22757213743577742,0.6171234696267044,0.7371833698219779,0.5353477535769041,
 0.5513561751575378,0.38027588203816076,0.8019834292620263,
 0.33447626140470343,0.10725866336981071,0.2736453169929257,
 0.4168531407224093,0.9744411230864215,0.17610716924916037,
 0.36866906286036927,0.3814779638117495,0.9029195498146382,0.7913367055138236,
 0.459377111426751,0.5074884595174504,0.48035783733353754,0.5407785745923439,
 0.5290925868592051,0.35743910382194044,0.8841299440700234,0.4043479369090972,
 0.20126470347582193,0.24639627035268763,0.945544453412336,0.3761816255873027,
 0.869397054093291,0.14019424065084685,0.7203964672569879,0.15114998393477608,
 0.24569741569005188,0.8519890920725687,0.8866240883507301,0.5582321145478202,
 0.09648100576651575,0.02942444168305769,0.6095725339666386,
 0.30170739545691805,0.028648786966041007,0.8488977652990308,
 5.874488198926997e-4,0.31712841075438736,0.5344793084212756,
 0.42815108155481774,0.2399628757872142,0.5470811408643358,0.7197991850183848,
 0.9741753806422696,0.1701241498654269,0.1152554456688849,0.5084102317836519,
 0.028655321514077126,0.5379532571627925,0.9353972952776677,
 0.6928372628650485,0.8695531530021955,0.8920104332519232,0.9591630297454073,
 0.06804021274624139,0.007920972221384992,0.44321883578474897,
 0.6932122827211074,0.7303334872424136,0.3098107539124373,0.78649605126791,
 0.9655457455699575,0.48994740854324925,0.6076381664582596,0.6361608768596947,
 0.1857442759801744,0.04638072291376916,0.2874808181521793,0.9087766025905707,
 0.3200879330241537,0.9270700759505076,0.5463632833949551,
 0.014697008937642586,0.9984226028220444,0.5141170482124722,
 0.6653892670804813,0.419930607107622,0.5141427458771569,0.4692239383715002,
 0.17759484991777785,0.4817279578743855,0.5960432246666726,
 0.21636527948666873,0.5211732522520285,0.3118462847079526,0.744173226173728,
 0.9285758041948421,0.6324198915781007,0.8071165236236132,0.49785869951056094,
 0.0074564471817470945,0.7302831267069023,0.015393447899728296,
 0.35624829888151344,0.341631291277381,0.48558819738713843,
 0.020570806115167817,0.03591664797828331,0.13745589428542526,
 0.45756591504081,0.7947048316625388,0.5755570449069104,0.44096888031282444,
 0.7646063066975711,0.8870920432188392,0.35945423637938057,0.9845081040403333,
 0.9646563241554416,0.4864407710164108,0.5437766982019986,0.5432333382451964,
 0.430310943406625,0.09248396260756075,0.7058084086862261,0.7477729194746602,
 0.3186486647743605,0.665550655484112,0.48043083592329616,0.8635681750296457,
 0.7304296275519568,0.9355701335856187,0.1578912896103204,0.6572414716670457,
 0.09607380389407916,0.7889242471657287,0.8597546747634557,0.6667637336292865,
 0.12431703240909786,0.061005598793503646,0.3114018779406377,
 0.2503756330990837,0.6981133051108754,0.9141253358323868,0.6447873894715151,
 0.08964263699706498,0.30053820760786554,0.8585238879954423,
 0.2399157153943443,0.7552317199207355,0.06358723450916604,
 0.06260645323663927,0.8760119269964012,0.25477202940051724,
 0.4127330693426895,0.9040474063192063,0.35769915658806173,
 0.10733196352119345,0.5959631423777088,0.23465877452811656,
 0.6949263103361821,0.4540613062467358,0.72724272160695,0.6147494458170173,
 0.05433277237959966,0.5521771717614319,0.36048058154617657,
 0.40169300334721936,0.4898130810027178,0.753615789794554,0.8799060791512961,
 0.6031345650419946,0.6136181648332679,0.20236747329673943,
 0.13760135093462156,0.2597279539698347,0.5635853224349225,0.6916229844514992,
 0.30217081862154627,0.6641730157390584,0.22684258108811717,
 0.6137740360527236,0.3824215187735083,0.7039379020182982,0.6420721904462381,
 0.6997630272832267,0.4554428055565054,0.3739139357102863,
 0.025553042688135075,0.8707008829959564,0.06893965475364638,
 0.025147779895868028,0.20287296277145206,0.7706530015641118,
 0.2266094922075671,0.2222861691501865,0.9365444789910966,0.22227644731721297,
 0.3959928588465398,0.2621612616180088,0.6686695433957639,0.7226160716588022,
 0.18514428735002675,0.4974972069352216,0.05907585697137674,
 0.5138141513025145,0.30652762318487947,0.33386943584289586,
 0.7876949967084836,0.9288051690233298,0.9774065256597667,0.9601067055374704,
 0.49630372131418443,0.6813089794661403,0.08517187377450242,
 0.0162992154674394,0.34460857546307755,0.5237265058084727,0.7338831373764266,
 0.7527263123616851,0.24376317138380532,0.4033485180567695,0.7095260793738725,
 0.9630771747804294,0.8909380501752803,0.4473814231384958,0.7583939414252171,
 0.5500404342010029,0.4729590249585599,0.05106685516641807,
 0.41924517675337447,0.8728582776040106,0.13891258761268444,
 0.8233157793762483,0.042879207117788476,0.76891712132045,0.09504667656394106,
 0.5802325313692381,0.48803491612928784,0.2333677468235299,0.552868133665681,
 0.9836744883571603,0.2569987847573908,0.41093838507945235,0.7665006075386062,
 4.818357997974365e-4,0.791793968293763,0.05394306524184145,
 0.2097899939180322,0.7965930000262551,0.5891305935886624,0.05300673860386251,
 0.9050043526139252,0.12837254195414605,0.02152471657138666,
 0.24707255748019774,0.4460551059725212,0.47685791490199514,
 0.7171525968339676,0.15656804495865795,0.03613737367947123,0.986376629573071,
 0.7920637549646327,0.8446983437609575,0.6270520105682774,0.6623061969228027,
 0.24946979205099984,0.009764815859800935,0.5304861045891531,
 0.8326496788642296,0.9703647654512825,0.8638596298721264,0.9715090005942317,
 0.038123126095598314,0.017942887782081418,0.18816887435668117,
 0.008340096902933025,0.8534645122729999,0.8404596618770596,
 0.008236669864833623,0.7180752674534994,0.9547080248302858,
 0.7824168372711081,0.08101906420642158,0.6176702638480395,0.795924730111887,
 0.8414044890340425,0.12380950377274946,0.0873893623612022,
 0.17976035211244612,0.40730142814737513,0.5302693150271404,
 0.3770694197956582,0.9397768770915053,0.08119047505649246,
 0.15008858799237668,0.2829877882340738,0.26019873516398695,
 0.31955609305387656,0.6863542987280922,0.4903921247850105,
 0.13046290778261804,0.8768155731439347,0.2503722771779695,0.7680047852565625,
 0.9919554030369062,0.3441390893619982,0.997560825692124,0.8541624810930004,
 0.2587500314108888,0.380938283764807,0.9331452046217129,0.19753983114038154,
 0.16456468902561094,0.11566770640513613,0.42674165579452805,
 0.35444641549449185,0.7292007256472145,0.3788132305169447,
 0.35923329228373757,0.5893972004946838,0.863518521457055,0.1317652305758097,
 0.3181261453875822,0.34666516325640584,0.4949111999639404,0.8175122776846947,
 0.012842508639522743,0.3091691407191616,0.11711624253295028,
 0.6968718719490181,0.21415980197628826,0.5053452121081518,0.2901765309691683,
 0.8569758240419305,0.20835139725412077,0.44077454936012583,
 0.09620383787516307,0.9584821856421489,0.3136102065538632,0.5440130005830418,
 0.387928006826319,0.9037780461360261,0.40231426007495297,0.5026262052442325,
 0.5258316458683496,0.7290947300624192,0.6998253347743412,0.961445491229095,
 0.2369836575227906,0.7147115825479083,0.5605558553238863,0.379144893574732,
 0.1467331876785456,0.18884538118173055,0.9475549798878828,0.5880235468117917,
 0.6146620404238108,0.8148511257363819,0.05682712772857057,
 0.24542820892035033,0.7531279647695267,0.413597817257517,0.02318740142224618,
 0.373696060501004,0.4700315547663052,0.43114616282545604,0.30716071709398585,
 0.9781441341921508,0.6147934635516202,0.8601311907419188,0.08491457530210256,
 0.13181307347310423,0.20693501120259605,0.7116251680715913,
 0.5220737740590721,0.18420264142322784,0.7490126618671219,0.9708795989968393,
 0.21313510685714032,0.25155093088242164,0.2876260194194571,
 0.2029779113828567,0.8951063183848341,0.44967698146988855,
 0.48150608443282394,0.3955679293068992,0.37036131498692193,
 0.6633846413243106,0.22965074449926026,0.6296734625657464,
 0.03274787852118788,0.0700515755554798,0.26523853655007634,
 0.41272103876578337,0.680049161051856,0.14138716901431558,0.2916932401931258,
 0.012423431266817664,0.03528184683067803,0.8652454281800505,
 0.8891961596521369,0.18803556423437717,0.34387092215425263,
 0.12187713859834781,0.29100141930583945,0.07288462631714232,
 0.9985321345841318,0.22792831951691062,0.370546245570761,0.962145229973563,
 0.5731365463496139,0.3003588297685862,0.43231853480289695,0.5182811888507147,
 0.020772799937372022,0.5502927104539951,0.2103694106445131,
 0.35655101143941004,0.9198145019443419,0.2665955572003089,
 0.37584372215008377,0.2546927220084374,0.17103805505078862,
 0.7493284824776765,0.4013416013690889,0.3036091497436946,0.7772599698971004,
 0.11615249498890512,0.31665041943687056,0.5671826954631123,
 0.46934863662727144,0.1727287890723581,0.575419880522751,0.6480215466014289,
 0.8964814461957188,0.3910414236726081,0.8230172777507192,0.3994163233907786,
 0.26286326173214525,0.609722614344965,0.6532195778723459,0.4668218135154596,
 0.12801216258750536,0.3278240971230786,0.9343036167491049,0.3449774067885253,
 0.4427647250503879,0.2574890102220372,0.5228868710957781,0.49425572341212154,
 0.2799532222821943,0.7921906562693181,0.9547773524585627,0.6627948603532696,
 0.010682041781520768,0.8508800056727307,0.4184098398591659,
 0.6514004849428018,0.9330853354058333,0.24307089098468249,0.9014706796032035,
 0.8422406514517373,0.9663088454947024,0.40614111648324724,0.6769496547331474,
 0.37258790580717505,0.2930092687032726,0.4899041232766258,0.9800843017359773,
 0.14907690603002843,0.2721354513793748,0.5629863753015745,0.1776993063881952,
 0.5231032593624447,0.07589207722272717,0.30839894236326004,
 0.6014673706484146,0.5257250854922911,0.4985527776168208,0.3838784468067029,
 0.8013090844300456,0.9183894656369664,0.8199838282150542,0.26344605430506485,
 0.03483349107176159,0.604891215438862,0.3035052623290039,0.20328976177009517,
 0.7259585567784506,0.4823770481520928,0.961270236850604,0.807001661240883,
 0.3207672263130721,0.911155852894908,0.4042111500548188,0.8347865306270044,
 0.06450631567196652,0.2747243541632902,0.5478851345724444,0.3616541418005046,
 0.3282020875877173,0.8952811325230864,0.4363339193357155,0.772009936750115,
 0.19933098512355407,0.48090466192570114,0.31225456920013006,
 0.5577251235446058,0.33575212926024045,0.818419805008449,
 0.023474355029063254,0.051989230920129303,0.45588442404912977,
 0.16853709499037983,0.5310053985495762,0.9114853802357403,0.4878296915944226,
 0.9268579837135681,0.7846935342116695,0.039783261456106556,
 0.15511925309123076,0.6772572553669246,0.16606817404105068,
 0.3290965372796656,0.16513448643155781,0.6006446735602079,0.7932260321597703,
 0.19667794592583032,0.5370859164474108,0.6932082512367744,
 0.49824621846363104,0.5902498510593912,0.1050173150915008,0.3198386288770174,
 0.13911005398763165,0.5808294324268348,0.7353819698049966,0.3369428002327918,
 0.44101158548996455,0.859709658009677,0.44736599952902945,
 0.41407385751113457,0.3203120913817239,0.07961673693402138,
 0.0577920008951911,0.4260749024231595,0.7617297257206562,0.6509480921564204,
 0.31976119150210847,0.3179209153733069,0.9874045265787929,0.7204631276819877,
 0.014776394862966113,0.7144584161346499,0.32062819961335676,
 0.6341897043616297,0.5279557181336898,0.9169354199572086,0.7667156098603929,
 0.07317484299200439,0.9359621256308964,0.9663704361275989,0.1999520461169617,
 0.12995554229618855,0.7225557523499915,0.5799700602968667,0.3921136002662514,
 0.88099837093248,0.7172533761169702,0.47055736610543875,0.19108663781624635,
 0.5331652409716895,0.39309294408617734,0.5719869511956353,
 0.49464064797340745,0.8764616666248479,0.7061679433442462,
 0.05002188107499761,0.5931957779518728,0.19852415931780232,
 0.6986682442994622,0.1467100262876596,0.488537537848742,0.8073563217483382,
 0.8887942892913248,0.4445519820670629,0.035079287606514464,
 0.9491302189873504,0.007645514191676583,0.1263174617031556,
 0.12301673828962367,0.3403168064397616,0.6760804327396165,
 0.44059476272683606,0.5133411652044595,0.23574205472228682,
 0.14397916422165657,0.6628765263622363,0.16257007082564512,0.482425499188182,
 0.9488575950982119,0.1545984481562367,0.949090580759433,0.986099962188679,
 0.039857278137941266,0.6766281827902056,0.6684563574831485,
 0.9959760773884703,0.4801304749571055,0.1195818882540095,0.21092733674932285,
 0.9465609382187392,0.8278532006629246,0.8044760386666137,0.23461303235828568,
 0.9610688774031305,0.19553703056052574,0.1891060502627936,
 0.003416575108498243,0.49852984382362886,0.5455713089885296,
 0.8592835342825329,0.49697427407019,0.15214460090189508,0.09457331142337033,
 0.2562965187952232,0.06088366319576832,0.8182946154848434,0.2618163643802086,
 0.841437111377695,0.6790047729227582,0.5147497869297951,0.31200319017608713,
 0.020349595929732067,0.6318116038377792,0.18402290323941894,
 0.09784065419705512,0.31656496694710967,0.1275095199190982,
 0.8709948895304862,0.7836552343698092,0.21433898636406468,0.4718142841653252,
 0.8754053115530058,0.2315050788689641,0.4614532702731564,0.8799501557058829,
 0.9305355264203944,0.38137412234569923,0.05385702093708533,
 0.4299568745802175,0.11244788681333229,0.5962094349472078,
 0.22403983979838304,0.34911383931810425,0.2906385297754447,
 0.2887168550894361,0.780518730220798,0.966535892633145,0.037028929429527735,
 0.6075827915068559,0.48787621200727493,0.8412739808838857,0.9347496622712395,
 0.7939010512292268,0.009454690764967966,0.8378392587160881,
 0.6068029447712772,0.17489257280024428,0.7089832414218394,0.9798344547285277,
 0.47443933906455193,0.34615215386613807,0.37286740130745377,
 0.035793559819444765,0.338444970449317,0.17776389919951674,
 0.2249737901691199,0.4922262921710391,0.7992910580453945,0.6771915809180088,
 0.6260768961322488,0.8295940563821582,0.3127381103660547,0.5847415935752355,
 0.550529763632186,0.6996574062970439,0.5393364588134882,0.921471673851569,
 0.7451750694115906,0.6548691909861548,0.8551119957030805,0.6741302950639267,
 0.16643998365030832,0.3280184671124795,0.7385035844481332,0.9260979266933234,
 0.6609128928966651,0.08007028674240713,0.36034969991722554,
 0.7039899212452317,0.3197768387639863,0.12126130760491749,0.8507613872914672,
 0.11223163466522235,0.4526824382244471,0.0014770942410111676,
 0.7692273710044198,0.2574821712576475,0.4221863019706704,0.8413503668362007,
 0.9613007336410597,0.8168426058394476,0.3679134111078691,0.244594287157343,
 0.7005676086409676,0.1189741742550453,0.38377020833075315,0.9109173616928257,
 0.739407560424572,0.8719760956049584,0.7253321891034399,0.11355973673718989,
 0.5193392849834662,0.5845080954082534,0.23421675987282542,0.4584965841983427,
 0.29554232355566,0.15187791754400182,0.66618391103049,0.2816591656785401,
 0.9137547197920185,0.02997261901733972,0.9812060056810414,
 0.11897681156922069,0.6618567802311234,0.4801746137720343,0.5799939446646616,
 0.680744934091417,0.060139844540647025,0.6197924470472259,0.8104403199476713,
 0.5932461077823807,0.49706594082261346,0.777998756975911,0.5313676886385823,
 0.7408086057547314,0.9868300920449651,0.8880060710147978,0.20338128881406858,
 0.23505310225665754,0.37298497653978746,0.3526981871912067,
 0.4091574425310839,0.12798108545296438,0.2587612354105827,0.2746977542884771,
 0.7930638320840552,0.35311993364472916,0.9940600346403153,0.4842256854497896,
 0.9632671483304506,0.3033438944695809,0.8089188561107635,0.0331010456198183,
 0.19764417211706464,0.4424360518065342,0.7473115672076942,0.5991727927970184,
 0.1281823203423268,0.1948403047882532,0.4162380391197734,0.29852660505793416,
 0.009780686167325081,0.7017097591676,0.610743306766361,0.1029175245662135,
 0.3633179147189095,0.9646339912385696,0.26963383789633033,
 0.06821545219525471,0.07089718093280573,0.16747155451315998,
 0.5436954866195247,0.8442959801212376,0.5392519404146958,
 0.045169999831027874,0.3196804294258452,0.48206166923252514,
 0.7054806115118485,0.05736310603603734,0.24546815788542253,
 0.4640401859187554,0.4373795075843564,0.5791473590797868,0.17148346267522518,
 0.9053103910598734,0.8715682683301834,0.2627992962472464,0.501365578862691,
 0.6531379193375039,0.6650983174822553,0.9905560964718324,0.6608753701788318,
 0.4486603689709783,0.9446091302509271,0.06584317721025634,0.6061333006467999,
 0.07957458986987476,0.6860382971582217,0.6975496631506195,
 0.40090393592153883,0.0740697082578331,0.2653287668300285,
 0.48623284176013337,0.1613445247934936,0.501705387074785,0.8299414871410064,
 0.43959540054133106,0.012968508748828822,0.36327235835454463,
 0.5346866373274091,0.3621279884355876,0.3336067531531096,
 0.018613140934926276,0.37500595410317716,0.997239019059846,
 0.9415251406303236,0.8412130599848862,0.521793570813417,0.36410121784570215,
 0.3130576011904478,0.22331943146663957,0.8431822165138836,
 0.33476881803695313,0.8318651403904377,0.8370280073145758,0.8296673649145847,
 0.6906583144237459,0.6491403841150443,0.7384152537718305,0.8695552622416418,
 0.4773567679586128,0.10875383518378157,0.2303343560281732,0.7168890019928578,
 0.9319915208712983,0.14411260932807046,0.49949729955528377,
 0.11874375463827758,0.8577536275078324,0.15123200764871825,0.783457401849004,
 0.32126849952603864,0.35656196099987514,0.6315151310046712,
 0.6231697559319437,0.9382694695998136,0.7081656083163865,0.1745838393252006,
 0.432716015883166,0.5060477093906495,0.49082450185971305,0.893559656656508,
 0.09910658679605266,0.04898083435365108,0.9701515023607635,
 0.5180154952756629,0.18671360606843157,0.8470110520593244,0.753171369813078,
 0.7354776239813428,0.15652465013228734,0.5827581937679721,
 0.058348889465302034,0.09615755568612705,0.3386465008979933,
 0.028628564634938125,0.8100370662036497,0.13166910532997367,
 0.9158853147238235,0.6866141491783067,0.36442522847250514,0.3080027492382592,
 0.2851302992409047,0.03724183602687492,0.9085880386158292,0.8420501988252946,
 0.8748114589343642,0.630661048115803,0.45125040302457065,0.3189303982814664,
 0.24077819748254414,0.33152756132463446,0.9700308066400668,
 0.4655546206877812,0.015745376986650128,0.6909398186575144,
 0.9217875115112653,0.17472367685467438,0.8913481176373133,0.6668259928887471,
 0.23614155789671076,0.25870283496629876,0.060314173174149555,
 0.7869132651330178,0.42480939882183744,0.42694955661082046,
 0.9706424302871057,0.2619730356668164,0.5852046405877428,0.513076992737553,
 0.5120669523111098,0.1828987256633391,0.29801171650721714,
 0.10607921671646114,0.9642224622205101,0.139879143599846,0.6223180681255241,
 0.06971688496609196,0.7418803838186372,0.9034610790125206,0.3171729783120618,
 0.9986277833294079,0.6921713867050412,0.34052311745732766,0.6340873618259566,
 0.3716427710915182,0.6065508274921091,0.8243223566916809,0.5251408334981178,
 0.4313716671554346,0.539975231725091,0.15767979288314715,0.8269188126778764,
 0.7517420017313889,0.3130445767969069,0.5914463605552074,0.7595367581356123,
 0.0880541425414837,0.5256484122092037,0.31579350853138355,0.7473570876642746,
 0.5412539032186047,0.8970975821945013,0.56240524615717,0.8662652083556226,
 0.9384429990998763,0.45115563055911223,0.9310975032228159,0.4137211676251181,
 0.42774525041604583,0.5586425865341103,0.4367938262213752,
 0.10019890808894583,0.45880362188282364,0.5921265357693308,
 0.6748042582575704,0.4488456317860495,0.6356538781297081,0.6987765425469968,
 0.8866345478105113,0.7419145530103204,0.9833069247822944,0.45676263779075843,
 0.4619424594507916,0.8665054561689695,0.6108155933551055,0.29327754937576966,
 0.5504915587599116,0.09402632979633674,0.35936804099450903,
 0.2453374117315621,0.16691870293579048,0.1733750762470968,
 0.32610780498061764,0.1396925082587014,0.9733884510959032,0.4295455753119629,
 0.7875239653112005,0.3333172343184807,0.4981838554657344,0.9599568448356377,
 0.6097494511062822,0.6940793675280424,0.2919106613804011,0.8017118577070487,
 0.9698056454270455,0.46457844306604423,0.15480358050367526,
 0.32828702170471413,0.12154375237491444,0.34106217278939754,
 0.4274151890465252,0.567138709273896,0.4113147858062779,0.5217012477190317,
 0.0927561173366166,0.7165721680306989,0.6454337455389794,0.8858612714168986,
 0.3277048027114581,0.573803898422336,0.7295429345714932,0.8223840346960035,
 0.526400344855534,0.19659284807735977,0.223775879851971,0.2870419155191879,
 0.1519725715137541,0.9563852575628939,0.44566072204263896,
 0.20520639088790393,0.6484575645054405,0.8474368603832809,0.6334262593424953,
 0.34690151406172276,0.15113112758552538,0.16274471327678153,
 0.9236457865660159,0.4755065203700618,0.7146350482072166,0.4386626383419743,
 0.06698377932077382,0.7872993811470854,0.6380303800872873,0.824564246464268,
 0.5716483992658766,0.9372716032547631,0.9187471575954672,0.20845074650869455,
 0.4882837342301223,0.8217021815659735,0.13790374086485457,0.5037688828012553,
 0.07211175956795846,0.7626177227665162,0.029353647072615563,
 0.8827795924576758,0.7624975083287096,0.7388566198786758,0.18717650366826755,
 0.5137823541110805,0.5950545286599767,0.23716549784558094,0.58613414826617,
 0.9218329209036016,0.4096568461281149,0.5144975286463961,0.3904988133092495,
 0.34691318427500295,0.8564345830406171,0.7523383947122101,0.750335131128686,
 0.7990590310415255,0.13407933238546244,0.008281028885705366,
 0.6902985832418215,0.5576038430003506,0.16176209723453983,0.753379673747912,
 0.9077316940515183,0.8899095365742622,0.07086965615713225,0.6172600403953321,
 0.25263648925891,0.24537647686427533,0.29070130598627375,0.8416680852751621,
 0.619187813509321,0.7712048662225263,0.25390778308169704,0.07937578759043584,
 0.37798710724271467,0.09153725207108954,0.4549347474741763,
 0.5373153788977975,0.07164309884238829,0.7453187894439922,
 0.12911805650739416,0.8116022475907361,0.7359528227431437,
 0.13094903986143025,0.8359394767584476,0.6707291300255107,0.8289979478686352,
 0.05170339014840564,0.14545370322309936,0.8194993857937909,
 0.9105138198446325,0.5788401949823353,0.5198412197140015,0.8348579045087781,
 0.45981434527039355,0.8825927596260784,0.5581268070867071,0.935229646589401,
 0.7160812138385195,0.06084497332537453,0.4642757493901324,0.6011309531640784,
 0.6791253872868205,0.230094600744435,0.3584796958027926,0.79360150227946,
 0.3335680109053164,0.8366681331032618,0.785614161662444,0.8964840341540354,
 0.6353292970367634,0.32956840509833396,0.20374601022680194,
 0.42946732112737396,0.6355217744833923,0.989793600651959,0.37407526528097657,
 0.7039281060556714,0.015409901832343142,0.06576310359981696,
 0.9248842695674397,0.5529384318756623,0.025286339031177052,
 0.3487551810655234,0.8337785082207357,0.48707795807013043,0.5744812768113836,
 0.6014217793883817,0.7897081681808692,0.8515299675964558,0.774616124100413,
 0.21775474809360063,0.3662076041199275,0.1076487838503074,
 0.19285121618746037,0.625461970053558,0.1288004506191781,0.5553695094923834,
 0.6926818632364328,0.12149493901485431,0.17887224546082736,
 0.5326898753447219,0.2976671774641725,0.984199007253522,0.2599590872500529,
 0.3081193936451132,0.9656755079887953,0.5058621115235225,0.1935852281654964,
 0.8906431664750345,0.9857075521166074,0.7474858006206414,0.7943312806711473,
 0.7783356818805571,0.5101948276535427,0.04423243967284829,0.3954053739536565,
 0.5517731810266703,0.8129917461565588,0.7297756596896814,0.6168461165433736,
 0.1700770380750194,0.0339614866808835,0.18619793926136985,0.7015984030153196,
 0.6730110651301371,0.5290040726437917,0.24533879433054107,0.3766294848624665,
 0.08260303516275458,0.48117229601361566,0.8067510210715495,
 0.4275395452364029,0.0779722138643758,0.2762684912320591,0.9900933385587987,
 0.022906676770285128,0.057337531251382745,0.1685408128112872,
 0.14568850809799483,0.6691503725751067,0.2268947820186341,
 0.24589848107777568,0.03996057221961613,0.15277046062621236,
 0.1411160359043191,0.2940057865854776,0.9465163637892741,0.8570328839402067,
 0.601625107851508,0.13761030733360158,0.9182360128620825,0.2831295246075781,
 0.042409116463919205,0.7277721373942083,0.7303954397768725,
 0.7980728555841998,0.5719218238231107,0.21844945973691976,0.9395750657296702,
 0.8724643553990039,0.11699451881172662,0.5058607758119552,0.6271970089774812,
 0.8184306001261497,0.16183564762955993,0.9347768559507367,0.9525351719225065,
 0.3944860947188241,0.00630573996999606,0.780845907488677,0.7441961853000694,
 0.8236419361999938,0.31980636017651687,0.32503862466994915,
 0.6003716290826762,0.3396946866518945,0.659714397471978,0.734350001211624,
 0.7637078225198873,0.012640995654699783,0.016485143420121195,
 0.13764067906480382,0.6373374856669304,0.8930980265458865,0.7976659890038468,
 0.38790248115827564,0.09355893539053328,0.8449842134716603,0.934186948213425,
 0.35198595203287,0.33474903546891055,0.9319861003045953,0.8364497438128732,
 0.7880276324659251,0.06402075099827187,0.6285335695402754,0.4288346813706656,
 0.9083054106057364,0.2457547269281256,0.7182941986449528,0.942978633533043,
 0.9953033011583747,0.5078510221889337,0.9628334779928579,0.20365411626620578,
 0.6326829011559693,0.059342326325037775,0.8830612015750798,
 0.4616167738437872,0.46092677535276105,0.7222795282903187,0.2072904511287631,
 0.6428143581100211,0.5948592035831725,0.11791132727793019,0.8901764389493527,
 0.3901757834388191,0.06365052708975227,0.5085879706775154,0.2613764447260035,
 0.7880722231436946,0.5963837587184112,0.4955302150469447,
 8.810384968235274e-4,0.8489732883040598,0.2458387152641348,
 0.6950785745581247,0.5095453564093889,0.9070402571053725,0.9242001819448838,
 0.6520821162246283,0.7901760701192768,0.18986298193957218,0.6561508432336824,
 0.6218116710408158,0.8928322502279509,0.761157003020276,0.1366880543764406,
 0.5512542590817195,0.8321787606977978,0.5750545467072559,0.7067748223304503,
 0.7213909799105174,0.24938713302631843,0.9151682517861439,
 0.22892781170829513,0.6020028635751232,0.48687372489666403,
 0.7024865038555751,0.20486249312851013,0.17357617900481592,
 0.025767795978916208,0.3796569219180339,0.5692120350119241,
 0.6962779182793934,0.18610628630842585,0.2159695480062852,0.9915443458761443,
 0.5115598108761978,0.8395253391768271,0.8118606382149002,0.8194534008680175,
 0.724278918172759,0.9166510369583131,0.9826090477303767,0.2645585099430393,
 0.6886846721777057,0.5497690752129583,0.7341939199685792,0.6722518713287406,
 0.6447643219229438,0.18009006177166642,0.604807283782801,0.30151211686072144,
 0.617195343908462,0.11422317666616344,0.8738543286279922,0.44063458853994075,
 0.3938459081963659,0.7825168948497341,0.46724503440077636,0.4391501488707247,
 0.6791784458392659,0.9602138342934855,0.06946928543425612,0.9648438611211442,
 0.9858503652649216,0.7461376941610391,0.32273271205751186,0.672479968733355,
 0.3826816211201397,0.5186002872638582,0.6765581716617556,
 0.016361538425882505,0.17609178199596398,0.9372314199621374,
 0.2192373288096695,0.7643464889983342,0.828081340651021,0.8558834125557206,
 0.31311077234346774,0.6455181361013523,0.9246932801783082,
 0.01797325943596595,0.5145003009854101,0.9596295384985789,0.1978503280474997,
 0.9096263038846891,0.7094760758467163,0.2154209229916081,0.6127224189065212,
 0.929993058736617,0.053292878295274404,0.8342930882735763,0.1451132228069032,
 0.006434508161551777,0.2965444347462869,0.22012342254212314,
 0.5900016381684109,0.41180910152637396,0.20179665319617168,
 0.7439092492121182,0.7080307496891546,0.25102176979961577,0.4235486479264894,
 0.6079054539184725,0.44051627038536567,0.5763544682978482,
 0.46672455007252556,0.7012175029238726,0.18785195056345882,
 0.2628858833568567,0.4863032083672081,0.6375310049164581,0.4250589816553286,
 0.7264917470940827,0.7314398507255464,0.42774919538247214,0.8466543243419483,
 0.359881125303247,0.06594396785924694,0.017960309670050245,
 0.8859046445817329,0.027106707083824788,0.939124117892787,
 0.12090350244763237,0.6191515958688083,0.5350923020674305,
 0.12594283005425932,0.03059970853742222,0.6171490893508748,
 0.5951311077959317,0.008418539094104394,0.5522161132370442,
 0.43489419607852986,0.9721991613083906,0.07281739753124616,
 0.3739579244954335,0.21793185298115936,0.9714796731601008,0.994100513598538,
 0.6161824836140847,0.09855733178045534,0.1956070748384604,
 0.06283731813909332,0.41487707451435474,0.8991548628492656,
 0.7708143836098144,0.8242743277043107,0.0386352031501499,0.11326392216077208,
 0.23831857225359743,0.622690181973988,0.8239147841421652,0.2885335414907505,
 0.35541546147615133,0.029024225199918208,0.5426508366839057,
 0.3273706676842171,0.9248888709937327,0.9269459412674853,0.921963340218584,
 0.0878781277967896,0.5786635461871298,0.12166891506676192,
 0.47651140686680693,0.8136000232662688,0.622112336143651,0.8538851707959039,
 0.6989816277322785,0.4620967022389172,0.8478389958048712,0.31969549867204194,
 0.4247075065902888,0.24065240174369862,0.12439968877368945,
 0.20591972398476144,0.6436023711506491,0.4745968171959888,0.6574305469511841,
 0.21859738428091013,0.12844903913917427,0.7304794635868324,
 0.5795549874150109,0.804270930981015,0.5676385208129616,0.19845475396492385,
 0.989810274955972,0.20643558984710897,0.5662224937617724,0.4724752189362674,
 0.18211934185663992,0.4105084541555598,0.11222446548944287,
 0.09696072186196181,0.9344815174160956,0.5199955735966734,0.5652975694392487,
 0.9856122322064322,0.6697571322768362,0.09232866036663467,0.5880977214023128,
 0.8632297261072877,0.9032582061724882,0.7176444298421603,0.3608390468756011,
 0.3606751120484615,0.8147082496717212,0.06075327372260986,0.5265597234373313,
 0.46276208751519576,0.5489867466246786,0.5485761579314816,0.3167020176630144,
 0.5631493938994151,0.3559964492712754,0.9781243878695544,
 0.025887874910663355,0.6507799672735998,0.298585004569349,0.7891787411774952,
 0.8710681044564401,0.8014030304667092,0.11526822165790118,0.7218589006912407,
 0.6644917043290817,0.7534030684590426,0.09609129727666721,
 0.32745124362504785,0.658206916029667,0.3275545004523003,0.7656801581471747,
 0.2753819479315994,0.25565745684232444,0.20934085563974536,
 0.44337705195959953,0.24938855661753534,0.24089966288775588,
 0.8921414387886861,0.5766751882148322,0.725090213177639,0.5588643870885525,
 0.5297418107693324,0.03704427390679965,0.19282439528056783,
 0.5940274274458278,0.2476292587933534,0.35566036232359144,0.5409056878797492,
 0.5108428193918396,0.703924425698485,0.8371273441456283,0.8166171402822033,
 0.6578448318447299,0.2375323791786189,0.32589978479418225,0.8409733096786077,
 0.005046451287015197,0.7870922909939081,0.5819510863459836,
 0.7044077945799625,0.5376529284332361,0.48313274239731796,
 0.36962561559999507,0.5916379812515251,0.9825591551698426,
 0.43737840050151966,0.8423496615361177,0.2641854863072769,0.8662488803505638,
 0.6478477895547274,0.42459674685382953,0.9759154480324604,0.6805228336353123,
 0.559024325605859,0.5726800922890105,0.8669377055070768,0.9312472433855135,
 0.1172638477857374,0.401483313457604,0.7619074787221771,0.9455514077097645,
 0.08768118092373378,0.03774507625193746,0.042756516333715244,
 0.598261352942582,0.19245713809731357,0.8679809866384156,0.6343106313088586,
 0.8227811527119974,0.13449351948474975,0.21319217667625812,
 0.15417518744090752,0.9187025655848757,0.5200634036145872,0.8417504920903647,
 0.9611828369342985,0.671818457828318,0.5247319292710981,0.8942870210763945,
 0.21334632049205948,0.6181311152659321,0.059748227046564906,
 0.7366262263603774,0.53026865040952,0.0375781535205445,0.8173568371403201,
 0.912838066777641,0.30330947426074095,0.37402359451923795,
 0.13563517260811297,0.6227310963420027,0.7149271068010549,0.7067993516493406,
 0.9826212200081861,0.2079162133365825,0.6030633433290045,0.2048018995307247,
 0.9861164316858734,0.4634296783662438,0.964207665471038,0.6604221049741998,
 0.9903863115672643,0.19448386691343378,0.8133132527089635,
 0.45733824317592287,0.8909310232650599,0.8812316245776872,
 0.18856703918213158,0.6328925315477354,0.4851693833996784,0.5861211412351153,
 0.284241336105888,0.5301210671461251,0.17159265832384252,0.4462525099098543,
 0.6283018728668796,0.3831649554715628,0.2211805588726653,0.06685550315849065,
 0.269236699872621,0.7660005520570479,0.5910476405969068,0.9212237310981327,
 0.7122801350492132,0.3282269466189617,0.6533984380390934,0.6264857660275956,
 0.012667537885012181,0.6039019326963833,0.12481422497292238,
 0.8209699029805236,0.558959290552738,0.4503634034313615,0.6946320926060302,
 0.4636197296336886,0.04161206590207989,0.8213222550095245,
 0.10944584210912434,0.20026435355797834,0.5995894165571908,
 0.34655192797961154,0.4235817513352038,0.8056395915627093,0.8455935082088999,
 0.13736285428413908,0.4532645975396151,0.7518387675310455,
 0.04940589658329697,0.3921112819908885,0.8622935567740593,
 0.19804254942637356,0.9020885370929563,0.11776091578181846,
 0.31614379235012624,0.9532681783530799,0.9515778104055415,0.7652271235689809,
 0.02853363520887442,0.3190983568374517,0.9396279186871397,0.9546792418171879,
 0.3048511471469424,0.5767301278270319,0.5519031866306953,0.6557662377103259,
 0.20830356414830575,0.788891011145994,0.20965540327806664,0.6464636734750376,
 0.47512755075415924,0.819547818514681,0.41635099514976803,0.8867281101256224,
 0.44583095035144615,0.46079844521421254,0.7446292300512295,
 0.7123752704036758,0.4415556406194876,0.47319989724933653,
 0.08621225283164236,0.5788623017096073,0.9655864047038679,0.5677235031904582,
 0.5409804795735794,0.5297788029321729,0.6488970594559732,0.9320312435818415,
 0.7082612352568546,0.8727122009553826,0.39254492294758725,
 0.45954223732948174,0.4335999811368537,0.21645340203526753,0.901868049703855,
 0.5125423453876381,0.583823977887578,0.8441573763432522,0.07417293499798139,
 0.6552240244482638,0.4626662931892098,0.6948835790436769,0.9849475122351906,
 0.5139746074248892,0.18045946723803008,0.5535393174147053,0.8278116467134361,
 0.07008484581728247,0.8978552654663768,0.4227644158818994,0.8066196283512119,
 0.9615539108800265,0.8354172767412533,0.6591424757579765,0.276017298240284,
 0.5244787852497776,0.6892044254323658,0.13329225039373926,0.9950372029280716,
 0.3165741190486989,0.9725642756999822,0.32780253776614254,0.837515737503313,
 0.7435373851632607,0.7696081926519414,0.09346547509021463,0.5787324898251104,
 0.05549325502006186,0.31515867750284776,0.6556370495842352,
 0.9238985207219529,0.43415474893562145,0.6700181035975045,
 0.06118771839272008,0.0059159276774002345,0.23542170372932614,
 0.7553571002240336,0.3543882290839655,0.7655169396230621,
 0.006140867943263517,0.9432508379175669,0.0862254632604621,
 0.11701645621427759,0.5949970097126054,0.8644945092926678,0.7539487654246463,
 0.520006027293147,0.0874738564800549,0.8905099494091608,0.5490141485304104,
 0.7179551409216539,0.7038394057935701,0.12326430729550497,0.8094589422665847,
 0.7877047143122694,0.3273701528389017,0.49375750350896896,0.3783001983615236,
 0.04291347340676721,0.9758154318952232,0.8095254475693298,0.8087861376339437,
 0.9804822430530132,0.8001206719992178,0.4566037441981845,0.7843281450243911,
 0.9221347415024181,0.1269269248603413,0.160839494838119,0.0670737770763592,
 0.0760596512744246,0.3760837862507187,0.4349471494284872,0.21794255791275885,
 0.25706683610676495,0.4386767485690005,6.910124201179313e-4,
 0.1127085460663042,0.5755515773007032,0.5409655786522095,0.3618062923847635,
 0.8665315065561314,0.9615848632057193,0.7665650552904114,0.7919787370980468,
 0.8145567521989763,0.6240036766027885,0.7858948650751623,0.00425863413012495,
 0.351669622741987,0.808203071700318,0.14486147108907355,0.38084316599333223,
 0.16940666374776447,0.9259817789521796,0.8043640968823723,
 0.30324827986293335,0.8611348005939503,0.9975040081612467,
 0.08646449548986468,0.4084399311176925,0.7868400080041984,0.5141040380373345,
 0.6468788657647873,0.22730296999202293,0.14785242453814396,
 0.36070695248379225,0.4908945133570666,0.43243779523582293,
 0.7818452147836491,0.5823389885674021,0.9392828750048098,0.4670958279630687,
 0.20114722701966148,0.09483927954642246,0.9269456769773238,0.951564105784506,
 0.6154462189065216,0.6058061474854878,0.29917067689989496,
 0.40013828984962874,0.8911410810253156,0.7061179387189771,0.7280595291281122,
 0.9714308082962542,0.4094607832676247,0.37368812286325415,0.0879181161056447,
 0.8867201692978907,0.2498189765689559,0.22208335039235982,0.6110190977356328,
 0.05347268217684531,0.7970954453122866,0.30140843575641796,
 0.24610835063303227,0.1747246892521931,0.21753565238401973,
 0.27505883105144635,0.4404441025108926,0.8693958199653782,
 0.17605126355923284,0.1097378676922488,0.5671613131612657,
 0.33268508407017827,0.08264935419854036,0.8408494128018882,
 0.8924043675269667,0.2876279675549772,0.12147829698736357,0.6223363423569165,
 0.5237326592344088,0.48058440753757015,0.8099215893781803,0.1523281428859381,
 0.9419044079010092,0.8355789265087175,0.932462593999222,0.5597384989665612,
 0.0443630186207874,0.794141141211813,0.6149216562037071,0.7285202059940579,
 0.8447179890013249,0.6692306282389089,0.06945555769635148,
 0.024805698808665566,0.8023621102627237,0.7043848518037267,
 0.6159100205125664,0.35884646129934206,0.10652264227598274,
 0.2857234936268669,0.021564513771838212,0.3733595033673319,
 0.5396409712863142,0.5857069002121699,0.10623889105017958,
 0.44743653965041097,0.10635530883873989,0.9838561012679987,
 0.43872070892732373,0.5271741389038688,0.8104683436325231,
 0.17123320492209004,0.04921561525794915,0.3388444487018526,0.510166643945325,
 0.7707149883271012,0.34680049350186337,0.3559737460467103,0.6330276625967786,
 0.1737219160438307,0.6097855252957249,0.08364681109738659,0.3361756250897201,
 0.09894965867340388,0.48947944705851043,0.9835153207645323,
 0.7558341044079409,0.49159378843472123,0.9569204144243485,0.7858887426115077,
 0.14786010346233946,0.3097925054076358,0.11851499142490296,
 0.49838392134952236,0.1536373069651562,0.29912782759869816,
 0.37200793060986026,0.7008599548114017,0.45585051485057015,
 0.26706422942078867,0.7299096393558435,0.9857033781422428,
 0.07503085284790278,0.7511896026730848,0.06953758168115065,
 0.9449609863943036,0.9843071041454164,0.7630912932539285,0.2893087354654307,
 0.16261895968596762,0.5792081097246667,0.8716276444675894,0.8241295221504471,
 0.17515928987589402,0.6336763458470744,0.05255733951728481,
 0.05811458156200444,0.9817286168209702,0.5851558183670836,0.7296702076659268,
 0.7347409550573508,0.07723139175024385,0.18871023115003172,
 0.8187077105985991,0.9163773258738427,0.7238386857397654,0.2529250569664523,
 0.501574863919243,0.88769989795602,0.4083828965678523,0.6365142711807564,
 0.7362188923432685,0.6546916298991351,0.2945229519406014,0.20093388988143357,
 0.23219519518023635,0.9504011771620114,0.6591335053586468,0.962078613298063,
 0.6956581124933884,0.6254831876455812,0.34770275724106925,0.6422521056105217,
 0.8959647319807551,0.6931866425449033,0.49738139448472485,
 0.01677286141335188,0.9483033259649721,0.3059563367497672,0.9699795688064142,
 0.9895185939149561,0.13829370428051702,0.6453488739559892,0.5170434670710479,
 0.26637117563567236,0.09518673935487199,0.1144945902179586,
 0.7055909653834248,0.08120358302039843,0.7117675118673154,0.851138260410069,
 0.9351963439808854,0.383985291133106,0.3027556899223405,0.7377967750379159,
 0.9693630504845571,0.5340584139837063,0.014818844114167007,
 0.1404605897565676,0.09284941895949783,0.49453072944054777,
 0.35256312094261344,0.6226076614391485,0.3596793168764384,0.6041363870954295,
 0.2023503936761878,0.4385390862662737,0.7657703835561518,0.7572850837077565,
 0.43419997622660933,0.731219761416763,0.3908233977290603,0.01857020515023322,
 0.5893635446923092,0.3659940299212219,0.7692690043228879,0.562275576305071,
 0.9525379573557694,0.758646046002557,0.5155650065929783,0.8056898405156431,
 0.7728135099368312,0.2190128560129857,0.1422012117183562,0.20491961581824492,
 0.6412945495149588,0.08419943974372124,0.6549233968871875,0.9205729261225385,
 0.9290929816969208,0.4063126480419913,0.21645430353411987,0.6391294861416,
 0.1586578470029668,0.038457659278968936,0.7519482190563811,
 0.8975925926865163,0.6491969624254191,0.9539353785143181,0.6397859237802817,
 0.43844511247737383,0.5144589853352851,0.770180155571191,0.7993048576057713,
 0.6076256077007557,0.38812343426332063,0.6904982706143278,0.5068235101802558,
 0.4788154346638279,0.1756464996046938,0.43946056939012923,
 0.03609218578507167,0.16210603557030956,0.7721983122957212,
 0.2393870580407621,0.40736923043049156,0.06988539759359769,
 0.7683864613462769,0.9158106866391851,0.4482447202737465,0.2147858799207124,
 0.21012751204128122,0.8600164841737312,0.14165356788256112,0.965801534095653,
 0.20579440434386564,0.7033903773399107,0.009330959587135257,
 0.018058948888834414,0.6396432973580326,0.9913423221505007,
 0.7133254427365392,0.3783578053325507,0.7388169215751288,0.7398326528347896,
 0.31848997564602355,0.009641196495718485,0.7150800387797,0.911288961530508,
 0.8453090004109794,0.5080011603320124,0.18743627132118512,0.3173506618284647,
 0.5931045150500855,0.7244425498778501,0.8089814702460961,0.04934927575439507,
 0.7298753331967852,0.2094043374171568,0.30609987261915483,0.7011623613906428,
 0.2472039691545158,0.20493531384300812,0.4466874193417989,0.4341656462531577,
 0.5965948742164283,0.5426186162731342,0.9763030833394002,0.5072300334618471,
 0.26420712208512354,0.899896275046132,0.5557702999573325,0.05914628807580591,
 0.03533197813154154,0.681234198294901,0.20621691334795322,0.5251272482377208,
 0.04079694318606941,0.8696973669789729,0.1636057003425524,0.4180434766526756,
 0.4916713903741389,0.2880859674387537,0.2738799727308098,0.8276107966674398,
 0.7859855418603547,0.18847924159014673,0.3361290327430363,0.8875037049607484,
 0.3373225171261107,0.28191555801617996,0.27977921897927693,
 0.4688265238049486,0.5418304952579559,0.06635778492875921,
 0.01516046494028056,0.8441038711353133,0.6100546259006977,
 0.08143852522049444,0.12392185327965466,0.4765019607856502,
 0.21757797026070735,0.9817042955542705,0.05707164951076338,
 0.8191651154758699,0.8954608845584736,0.9681056265003519,0.1717017580571023,
 0.313533366333296,0.5349871274441897,0.872787708617405,0.7748201613249923,
 0.2551937422320534,0.19890201213842362,0.248204543267585,0.7412885265715482,
 0.06168306056489248,0.04327825875663993,0.8165439225864733,
 0.9430458930487454,0.7024380145517669,0.9668623676542832,0.09929569575704511,
 0.654161643489264,0.7985404387053754,0.536895216196538,0.1140163255026776,
 0.30666104029119534,0.8880177852410865,0.1916151728715988,0.7943531594337647,
 0.12938348535967026,0.926637866287466,0.009192480007827908,
 0.2647239482722734,0.48448633308590516,0.873102726622368,0.1236164969676311,
 0.14880412603283055,0.46197952465873204,0.3681104507016544,
 0.8461469102033167,0.9242318122554837,0.34685676865733306,0.502867677908327,
 0.8975420244230423,0.149713169666422,0.764916948940379,0.7677144238898328,
 0.5780876514158374,0.9307837099684296,0.035644618518317794,
 0.7267217273606159,0.6565631498271328,0.03767857017501064,
 0.18218853109597555,0.80202101523797,0.4502460109257239,0.18070202706417804,
 0.2222814691028484,0.8389194051036466,0.32111168691152703,0.5262838916677406,
 0.05222889202829606,0.25872623016970864,0.9107269821787378,
 0.9411841902462184,0.9399943228954709,0.6338369366071636,0.2054267593021185,
 0.6946712407518886,0.10806988083017122,0.15474269776451333,
 0.6298723436569849,0.49034607545576725,0.5017696875922998,
 0.048808431592259116,0.5527985031502208,0.28759265335836326,
 0.7431114043996813,0.8257031291042168,0.8568545642346767,0.39465415498416767,
 0.8322371030262954,0.9695921853614221,0.7531158677570009,0.7194832705637709,
 0.7234660349087568,0.901572536911668,0.21797967247458439,0.9702735160072059,
 0.6541503019165673,0.04626455253083084,0.9279955756092797,0.6064059072299878,
 0.6364871581839984,0.44315652298918085,0.6083145133127564,0.9421811198913901,
 0.8829344171619029,0.7898791816525287,0.5398729576852397,0.8823105574889899,
 0.8356262977520826,0.725598289214565,0.4980122482255199,0.4943606820549178,
 0.34073936387140447,0.7822450500059676,0.4433406093131248,0.7123036068050832,
 0.19840004941692213,0.2858926901912986,0.8549253267243238,0.3413163836175779,
 0.43593873556584417,0.06829417207476929,0.6355910593162104,
 0.17674538169735154,0.7599055991500235,0.12474677686108882,
 0.41127246565779063,0.8021597567922827,0.990195320417286,0.4260452422300556,
 0.8502928314874695,0.9219151023829786,0.09598278993912357,
 0.11063944671558257,0.35447923232068457,0.7142900079048561,
 0.3780489930559434,0.2939397087917037,0.2548217581135297,0.606095877905885,
 0.5773866842598067,0.7556695292233521,0.6771767109785228,0.16360110880814482,
 0.9053969662709094,0.8905687513162743,0.2710185250615893,0.3815376844560203,
 0.5443650500802383,0.45170757773265335,0.2817726835511086,0.9661978986752916,
 0.6819837946764167,0.3836276760455385,0.025663680902705543,
 0.09144270200562521,0.742710915474942,0.8321055998647365,0.613517298824881,
 0.9279745816642202,0.394289035775806,0.6574481844947093,0.8944897538934858,
 0.4954607011047504,0.5812048983122452,0.699511901945039,0.7758224715785987,
 0.12313650428008338,0.19264905830214119,0.22728651840351355,
 0.6080106248029306,0.6612737365709702,0.860322191344823,0.959166888877677,
 0.8919029441712328,0.23085568812269575,0.09926127348357272,0.622929018336583,
 0.8792071691137494,0.9561873696491224,0.9937467925409615,0.9994227313745307,
 0.7274676890267965,0.7500767170574772,0.16049306530273366,0.3300011853357482,
 0.20531025944389802,0.6897015782196182,0.03442207563806998,
 0.14757275645120727,0.42501648266983094,0.1742887800140689,
 0.11477435392349744,0.5723847288089492,0.823293200330493,0.9733075778314279,
 0.08804254567832914,0.7254145920879987,0.32890314052016234,0.683476058083312,
 0.5623584815967477,0.3648414913947663,0.5014649793673138,0.10623316394696614,
 0.16613742364185136,0.7772196105838538,0.02812761138168407,0.56759414207289,
 0.07459013447812124,0.1845939162703888,0.7387953116398404,0.6762377801070231,
 0.4625812664272424,0.9068091742942593,0.9259240639559301,0.5055282020317273,
 0.04189851402238254,0.6970101719275064,0.29460447975039195,0.555848717582032,
 0.40441348512434905,0.6475705371065079,0.526146384615276,0.35789504677027184,
 0.574539840743271,0.9333048058889105,0.8161480827156535,0.33066080836394685,
 0.16493789773166712,0.5002325154602622,0.8770850239463894,0.7951948278246822,
 0.4239671030329064,0.4800402022916771,0.5949553328756495,0.16661644430657896,
 0.3194710217175306,0.04887463769788325,0.32940917417241145,
 0.9515770904293109,0.07301884012779314,0.6383403261908033,
 0.13328684907288724,0.677046074065033,0.9694164681435637,0.4011680152854573,
 0.5231982406910085,0.14364577323343963,0.6417891020076443,0.907087444766374,
 0.9575271663199504,0.7812604707178357,0.46387652033992954,0.7759216425957167,
 0.1763718536804143,0.09919288196701848,0.8398697319308183,0.128168575125574,
 0.7182506834181431,0.46103090791554624,0.036488837801146645,
 0.10424605364282569,0.844648048703345,0.8824425876054631,0.8271378307029068,
 0.29045595209148223,0.12604300733470208,0.6158758986620003,
 0.8734264820548391,0.564285190208773,0.5986995457427434,0.9782683384426069,
 0.07695104621009918,0.4287787017415252,0.70518150211351,0.9405527336702364,
 0.19445477230214425,0.8934774680707285,0.11613292076424453,
 0.44579839594754134,0.020425957015240748,0.20187963904116057,
 0.3825420579433445,0.28241920101325946,0.059575042414108825,
 0.6924210102527265,0.6650787915571502,0.14264279183056106,
 0.30137454639782835,0.009108749877373534,0.13556730327661315,
 0.8922680329152479,0.8676068711226765,0.8980854195841913,0.6387614966768685,
 0.5111766625993825,0.9154261851990311,0.1306893966868029,0.13017880786695224,
 0.42574258116895813,0.6423002617761889,0.7779800878514734,
 0.11352542618762307,0.9647858681098356,0.7865982644429154,
 0.25952911759639996,0.7382328619487152,0.8908711187595646,0.5461851904809829,
 0.5766003908437668,0.897746538503382,0.07958964135246405,0.808487903192535,
 0.44209933505130783,0.11755083297549096,0.05505228255923045,
 0.10780200378385874,0.5277862024227897,0.755876430059751,0.09602592713897629,
 0.7600654637697987,0.7367129923640847,0.06107575378808083,0.1469637499126848,
 0.8979944479333195,0.08166923672928239,0.841673194352719,0.3525270282347247,
 0.3072583765564604,0.7092359998078555,0.17175237504378194,0.7380165535165552,
 0.8260791590044236,0.11244852607192812,0.6996293608238315,0.1836852055113639,
 0.509752955104805,0.6732955845518752,0.7010369661285445,0.5639261973426604,
 0.4584461324088507,0.7536853234567674,0.20676986556370403,0.8499852063077254,
 0.6802948321895425,0.5059634807318185,0.14185693904742946,0.6812684616989665,
 0.4321616274350485,0.3802039436804321,0.9004894211028099,0.45891860066161627,
 0.8811145826803836,0.2780678253448664,0.7670967396350425,0.9556967213290924,
 0.11065197206968436,0.6109961091956267,0.23187090535695754,
 0.10522004949828545,0.5814823024913305,0.8820192497567678,
 0.19293302791766354,0.8844265568427717,0.43875860324567495,
 0.8988981842808202,0.8862339344720417,0.20530338811555404,0.8278892642175698,
 0.18624353590765153,0.32641342116193717,0.005694438700227122,
 0.04002580667945155,0.47852276558828255,0.5236915467381518,
 0.6478842915459827,0.9459360629116825,0.7423795694537252,0.5508150690504459,
 0.06362738151772884,0.42557765010658655,0.13041214943003288,
 0.9231755825853349,0.8127665290967301,0.9210284995195596,0.9835168648508231,
 0.5626137504544478,0.7184824791690739,0.3019426469454707,0.7703417805737788,
 0.5062215982186784,0.17193689166696358,0.33389633352827863,
 0.020414219926244925,0.9285273003600797,0.11047290892493733,
 0.734996052514485,0.2032929680148443,0.12785744245164088,0.4176373746061026,
 0.05255546906278019,0.5717549979440811,0.3801023833804483,
 0.011325862649292362,0.24740833976573195,0.6055520695600718,
 0.8831593725119113,0.9451520244333078,0.6007956079210734,0.99984398550296,
 0.599279590311473,0.445938506933799,0.15290409229212565,0.1850642741763462,
 0.35183132585330124,0.4850286073064235,0.562909193145286,0.2915723544162603,
 0.4107533267615665,0.5780824386799658,0.5938462364698279,0.9715702177261405,
 0.26915372646464075,0.8781366996463225,0.3398365614955403,
 0.22734114597344046,0.6403483116723608,0.6817009670245843,0.9009444627899081,
 0.6571107126064251,0.24287177816020689,0.22875318934639288,
 0.9984543232417304,0.9764776375871906,0.9052155507959346,0.4986263549343267,
 0.4877307950859786,0.8198952233039277,0.2768001065345458,0.6101908064236365,
 0.35445484819898987,0.048566939138846354,0.48737679182441096,
 0.13412660100437745,0.8122857032608973,0.5534802716470206,0.6196140546274922,
 0.3183951937260352,0.8462755632425232,0.4541041710732526,0.5654286438360666,
 0.6461742497386557,0.8901415384369559,0.9821011908079307,
 0.024574499759414836,0.3848668625344378,0.01737409793936273,
 0.7711025197557564,0.07810853466255707,4.8462105129543076e-5,
 0.808525916252018,0.13826632881605438,0.5732847505767427,0.3310588918348364,
 0.8914686191869612,0.8040725884637776,0.6020805840317442,0.3492213921142483,
 0.7615903608978329,0.7125001370701002,0.7022414171416715,0.9368398242022911,
 0.2497095469307954,0.670599682344608,0.1642126145805315,0.17411393376693185,
 0.36765105385542696,0.5067956803834505,0.32714331853747447,
 0.9422432860153421,0.09253415679376586,0.23196604543033905,0.900069195537997,
 0.7358831756746783,0.5632666725384358,0.023261144222802566,
 0.6745056878172742,0.7999022566245563,0.8294506768608775,0.06746259853944858,
 0.713248719637585,0.1397477737262386,0.6033479960820607,0.2791518647728325,
 0.1926785916562731,0.09293421418735726,0.19049634936359305,
 0.3865769075806015,0.09869060828802789,0.24176846591052126,0.085929062795577,
 0.24062248445700485,0.39858242237836927,0.37213384434745633,
 0.7879578925005931,0.20996026324261063,0.41885343138923525,
 0.4265204359735626,0.47919669310947155,0.31241342773201886,
 0.6364715102424109,0.11485796044775853,0.014454531365221457,
 0.843222778146124,0.10441421496421766,0.830970477556752,0.05976862457898169,
 0.6291879909726028,0.7144385370276892,0.029412880779960626,
 0.7207598225634878,0.6289184684228066,0.3014596116946736,0.45607371775841055,
 0.8314941988164211,0.13183997714834428,0.7601387593912924,0.387193537438813,
 0.21492291894342008,0.026965211308252135,0.8411048178010395,
 0.9146393614260893,0.08427957042462375,0.6951358040865517,
 0.43214602716106953,0.8852076215195861,0.04611355423473795,
 0.06955529315266751,0.13838966811208842,0.3933326924060845,
 0.4444848542620127,0.8581772550553907,0.2099266530983419,0.28915763863165933,
 0.36115490256052507,0.9308046529704567,0.6612110730824847,
 0.14866203842021908,0.5071377236476561,0.8789072329388767,0.8262791730503565,
 0.42981810684368627,0.8991720898879503,0.22875638966970868,
 0.8404005726642378,0.29563533412223764,0.754781181189806,0.5500925954589775,
 0.11084249901806376,0.11953532878766293,0.032957272197764786,
 0.8983451607585468,0.6978866633628691,0.7182576495742324,0.9661651687572337,
 0.8717947760135878,0.0034229551089252075,0.721732639592513,
 0.6623405411041041,0.6674033437446134,0.765464138684896,0.9986218772216047,
 0.21696830870946204,0.9452329656397,0.8276416166132661,0.5658641866795264,
 0.7196386733173065,0.4071289475096781,0.7259369148035378,0.5348119511244704,
 0.5383435559094113,0.6056464325978138,0.020302137964907097,
 0.9761289446633684,0.6447000723617211,0.8721264357991958,0.2591042350313021,
 0.23747451510071405,0.002589422152118992,0.6603662806730386,
 0.7163362897381662,0.8140024741315771,0.3745167207961191,0.840287524805347,
 0.4894293056312764,0.5425978071550487,0.5592033190103939,0.9349484368206377,
 0.18665624563935357,0.3928663622886358,0.35134116342172883,
 0.9651956024822347,0.4020945972330767,0.5645271912251889,0.0719870816379995,
 0.1285849432492635,0.12235333778315594,0.5441331629214795,0.0836623557632401,
 0.7761423455434704,0.8985911674821616,0.367127317522992,0.022373139225597916,
 0.7378070129220666,0.0015724863291606272,0.6626746414403307,
 0.5543135965564234,0.7113296669723022,0.0462406883363814,0.6588357165631671,
 0.2530475307210396,0.5128102508453509,0.5583945663369596,0.6040958651737919,
 0.7830828875209616,0.6656781780727928,0.13739886300062998,0.6693395007676678,
 0.19684369155887294,0.6407954306945507,0.19830104891469102,
 0.6851560194482633,0.6762087210145851,0.38116983985369657,0.6114877861764382,
 0.7663111801484951,0.8996435837067586,0.8466834161225064,0.8525428784500334,
 0.9700146756280718,0.5470618612709488,0.6511303025936127,0.6022210792757806,
 0.711481947016823,0.5316405042362132,0.9023963547853281,0.24885369939157564,
 0.18815704367157637,0.5521318444631776,0.4011764623991796,
 0.28475787532557373,0.3347242656139898,0.1258911823908696,0.6120234105688065,
 0.030519796609819405,0.9382950620053634,0.5986007746943144,
 0.21799877760697461,0.5302439889556644,0.8255433310261665,0.9113141285436293,
 0.2136508583279143,0.5487554728556209,0.549837042984204,0.6719963350960917,
 0.858910645309933,0.9550408822236623,0.6003857064768824,0.17256656725438302,
 0.9503624560425865,0.797700848245292,0.1637169041258043,0.36629394681230987,
 0.5980867281993304,0.022560793414730718,0.09024323569793435,
 0.4206779456411882,0.5709438522470672,0.2256948920210512,0.7173612548413145,
 0.1920296910257896,0.2944799844806657,0.3939256689109194,0.11031229377126639,
 0.9549364444687718,0.02611193349092167,0.23310358886563953,
 0.22742582352540874,0.7395832304266784,0.8611105194999766,0.486312135846536,
 0.6814077800133428,0.4118018142249469,0.004431967336203102,0.18533289028737,
 0.3439276802983119,0.806987245331203,0.31510945168421256,0.02508806262667629,
 0.19291992423323068,0.28346858962070265,0.3288855066836759,
 0.8103539252441022,0.049416057618209885,0.89703912785449,0.9466588287106701,
 0.40863814992128245,0.1525319142925915,0.49935308041280935,
 0.7899030303058832,0.5398341573037904,0.38921436875942184,0.7932270251031323,
 0.6053370164155073,0.2934251869082971,0.9613493356313529,0.12891237025099012,
 0.3001342247836337,0.3266291855531053,0.033372165401132636,
 0.9610720341533685,0.6808795178828045,0.6090307747270988,0.843400267041345,
 0.31969845594256485,0.020301975721135235,0.2979529378905704,
 0.0895627485654873,0.20559903346627784,0.4729554332528585,0.0665483679616623,
 0.7620448250899053,0.04195926359108293,0.1909870018820332,0.7178127042950648,
 0.5488038075068743,0.9766595523117324,0.4711355073192609,0.12364837588216482,
 0.5406023464881052,0.8564897357872856,0.5618789014505553,0.0168468209380086,
 0.27156654177659045,0.42887334342085925,0.21962988087736846,
 0.9772417275546469,0.1359570286600219,0.0603901857238035,0.5159461291524007,
 0.7965640844402585,0.38617913003879134,0.33535539732038233,
 0.49635759352661424,0.14613393840926436,0.10888958963632656,
 0.8781426633404872,0.8989775176347277,0.2388327780251025,0.9846501317480997,
 0.813892398468612,0.9714191043595685,0.257943158727755,0.20607347396302556,
 0.5668519867183778,0.9420628212636224,0.09833468741410978,0.9006444291539922,
 0.01301500258668642,0.8231659811472477,0.48574833790022987,
 0.20257052383348306,0.7475480533402801,0.02351487283988818,
 0.6480824063075081,0.7713710547366126,0.22547644478740625,0.5420023750148898,
 0.8136182321856928,0.1412947565824515,0.9460953567723007,0.5554535803208833,
 0.38109741052536306,0.48767555567778564,0.8858031594851062,0.436968592754879,
 0.2891812535798699,0.4995233347555691,0.2298387003136818,0.5170107793443608,
 0.9931208122706218,0.021552988366588255,0.31560557104714304,
 0.989247338275656,0.18081596566927083,0.9012637729529444,0.8760734059629248,
 0.6521207336248893,0.34424961919640695,0.16082698231192072,0.990820844237598,
 0.5297555106719234,0.5397575616888534,0.7832288128961438,0.14919175972656218,
 0.34564506006407025,0.5440259533154924,0.0032145535933508773,
 0.0031938786875239655,0.8986765430906761,0.4372786368303574,
 0.5537684932195183,0.1758902759943335,0.43140744305038625,0.3988389075603911,
 0.2710907686134644,0.38960974263041426,0.6895618014807001,
 0.07808548417133121,0.32128585526017783,0.9665067707659327,
 0.4341428160222258,0.4280683934763114,0.026313420917914776,
 0.5309456949076505,0.9559800448476612,0.2176361393788968,0.5407460995488784,
 0.14770954246242385,0.6353639117007596,0.9203824369158087,0.7815521114752497,
 0.7785664121836564,0.4888915732871236,0.9196826067899503,0.2302228967121428,
 0.6635467783228652,0.31279755433435463,0.3477691556287621,0.908354618575415,
 0.709547825331926,0.8987263797608739,0.6553571296623228,0.3341345458885555,
 0.1870624531022308,0.3938454836848504,0.10574773345407129,0.5777602615632317,
 0.5802791315013297,0.30867177858162564,0.8599461989833531,
 0.32518774371895975,0.8897203480887219,0.9779938474332155,0.2802532419713525,
 0.4810747389905755,0.2513000348561121,0.15681672678638325,
 0.29578440928142025,0.8796612302635641,0.49826864375750846,
 0.7478022797908723,0.0175713290616184,0.5145874275444988,0.3672831040789979,
 0.6059387660166515,0.015612226012941322,0.6478969641598307,0.766845230328889,
 0.7636335655503607,0.16352972311478264,0.9423036156713502,0.9097481883507097,
 0.9515708388565585,0.9142365811649683,0.8671994742238562,0.6652177505124293,
 0.178615582598197,0.4991431727186032,0.1464108331417201,0.48568375782470274,
 0.4550338403158428,0.12363540395031081,0.12064213490882825,
 0.5663496064681293,0.44783708886475493,0.8453994565558071,
 0.29364339190580013,0.6127023739117627,0.21896926738342348,
 0.1313726485154596,0.5534096990804406,0.2716420610557645,0.3664307170334842,
 0.6647349933982563,0.9452286446111544,0.4577204780208255,0.3909703468283179,
 0.17345602021431794,0.512875706933159,0.6663713820498283,0.7417080588243817,
 0.7859368476233881,0.3679774738767744,0.5931600689202844,0.31062684004268126,
 0.034085100950577485,0.1328502236903193,0.5618789777041481,
 0.6100462169277823,0.16357703206495544,0.7200389915991332,
 0.060330458702077094,0.615953576339138,0.11219262566278443,
 0.9695791088061853,0.9533856387229453,0.7326812770970992,0.5655897570904301,
 0.19341043051346418,0.9846864376124065,0.9234329859334708,0.9481916839978095,
 0.4244068266710327,0.10543241420883853,0.9209902825279722,
 0.02734043747706827,0.12273983979969483,0.9546022414361182,
 0.4116247126881406,0.206060955121404,0.2230542573186569,0.658863378988718,
 0.9538933449932435,0.2746842823811396,0.21903462356065828,
 0.11265923495903185,0.9865218297579209,0.29984237205524245,
 0.4128843874905693,0.5364798474392709,0.10322750939508196,0.3071932422767376,
 0.6924440262489512,0.23729065538610117,0.39950738813197917,
 0.2584964060250462,0.6785194139349845,0.8239071712442603,0.8488068891636222,
 0.8005976570819733,0.567098229828457,0.5848077834687022,0.11729978712582245,
 0.8568811452216922,0.6710740557835153,0.6278463184881775,0.8583003421311857,
 0.3236794864924927,0.15701194210762792,0.9001342889613673,0.882272834440772,
 0.9640711105226617,0.7702587545006592,0.08374190869595699,0.280865818479649,
 0.5562978832054085,0.8355664811877667,0.4936430648325447,0.7852287685541278,
 0.2327796983619872,0.9069287394399003,0.3450220126675685,0.9132521519383804,
 0.4315497573179553,0.5618907666337293,0.741095779370018,0.5409807839122575,
 0.07514836358259691,0.9943544974888333,0.5922586604939158,
 0.24038826166080596,0.7158128087492269,0.3042157713622383,0.7669531011077761,
 0.09421677330623157,0.5712832201281188,0.8567361399207247,0.8795617896357674,
 0.5815091412740309,0.07526728712106556,0.2754388182951514,0.8970970312789425,
 0.31528923884323357,0.2526142205169123,0.24094962764093064,
 0.40159975793420444,0.7276506864411019,0.28566012441700717,0.161411440051473,
 0.8663016987301244,0.7515035132479301,0.12905340376608676,
 0.19255123376137084,0.3671357967010984,0.13131915349827494,
 0.9910206856713583,0.5188533685102306,0.6847392452834147,0.39573631773144813,
 0.7898752803018484,0.39479476844548544,0.18250990361683161,
 0.46064604597817127,0.45274226208212487,0.1687653029285906,
 0.8782005256259098,0.17568153480138393,0.10161653233511536,
 0.6061552455998502,0.051915755145425124,0.5602359068777707,
 0.027183328268993168,0.7351158171767811,0.41373786006505975,
 0.5488653662106202,0.8242689403332907,0.7704623253038265,0.46175791241977726,
 0.9970324168748553,0.6388207033577009,0.2935633225962616,0.06213978025764033,
 0.7952675189617945,0.4581440011589083,0.2614814068980379,0.12349513678603774,
 0.8467253277536946,0.18275686137982972,0.9716606569714891,
 0.49959399229544965,0.9789423401280735,0.4197248627625736,
 0.10099035605272189,0.6751162079414671,0.8769259858861745,0.8756825852021843,
 0.5921640597042018,0.9438465398494951,0.08395343044576586,
 0.42996171849167286,0.19815311903681387,0.9318448739875849,
 0.009613717685200918,0.036653855568218985,0.5012714953699603,
 0.08870264277060824,0.9245331892126996,0.2947086239621297,
 0.47221436901882385,0.185433997752802,0.3830512166435207,0.1700110805679913,
 0.584363537381597,0.2957793355349554,0.8227593098127384,0.5451607002415968,
 0.9813116042496931,0.19483861997668384,0.5583028724777965,0.7667950364583996,
 0.5026585012605103,0.004003160481304113,0.8258322119749517,
 0.07958833679585287,0.9121501033125103,0.0366418369467173,0.2522866595357425,
 0.8552547252644934,0.6841867448659379,0.7514531688758611,
 0.014717672020776673,0.513960447928037,0.5013215890371665,
 0.034533347018566696,0.04646344312317441,0.7126241375139525,
 0.21202642587998877,0.9305966771098881,0.7272974075309732,
 0.05963064324395184,0.11521618275812329,0.8871037013712211,
 0.4735770656200312,0.048528979578262854,0.894153653772675,0.8908752383077331,
 0.9177254614864954,0.35498775776287383,0.5761177801168463,0.9484136414253236,
 0.7847217854798961,0.547206538594742,0.1744945218048387,0.23220447088809615,
 0.03796459052832535,0.5141991768615131,0.7322429159891443,
 0.16414619743274805,0.13111229743144293,0.7974825630294046,
 0.5586218150563781,0.9716007735511709,0.3741100621492085,0.9833142460594116,
 0.9832585482399183,0.035988082378970776,0.4549959680436574,
 0.5585900432302037,0.20095162164112013,0.1541191358720655,
 0.30560218079568824,0.9198026899273883,0.18241853426317067,
 0.1526009922466237,0.9462141555201078,0.8447678718417082,0.7440108095606017,
 0.21687396543884097,0.9880970484169055,0.650836809933371,0.14772529744303875,
 0.0877886338534799,0.585755450468821,0.7184703577255798,0.6074068397455771,
 0.6906121598685961,0.5954736354146013,0.6907995274195213,0.677071492931866,
 0.7732730073811815,0.5642954209501164,0.32616182428518925,0.8014943247920074,
 0.2074821747171347,0.8537436450659416,0.7971790997580803,0.8349971716162299,
 0.6524267905630087,0.7985375613516702,0.17088022495215505,
 0.46121941554769486,0.7673005291578381,0.21538191027925935,
 0.9869291281412063,0.6678641081228236,0.8351744056048188,0.2342467253406928,
 0.14538714743479464,0.7706445505894639,0.13739639619482658,
 0.5244998569810702,0.5046764949711664,0.9595182401531434,0.907782093129292,
 0.7416859478815524,0.7004989860291957,0.7823104383422904,0.15617912614930707,
 0.31032273181906544,0.3382666453706288,0.4427268422723284,
 0.038993475255341914,0.46733028685684763,0.38199417178991735,
 0.32775646692453897,0.6681733324510626,0.3505425173726321,
 0.10700496006110749,0.056970704052238585,0.19003335538174015,
 0.8904371044749255,0.8653309963795823,0.4726672314245297,0.6733032243456021,
 0.5200264215802419,0.36298864697087097,0.8352946009251673,
 0.24849552463056934,0.07331538099700441,0.38147501123402083,
 0.7721520669237641,0.2823413835227322,0.8633300232490685,0.9008057799846992,
 0.28143061311924866,0.47666898087090304,0.27089013347772706,
 0.6723286608677388,0.30279779966375964,0.7069893958632981,
 0.18996982140503804,0.7770191085787898,0.07603758997382992,
 0.6018761001777779,0.7515121316517193,0.4751517612461632,0.8992149012513063,
 0.3691035200181805,0.576001139941567,0.27127874510338956,0.3052439829948067,
 0.9985084130326507,0.6392384803003741,0.8837069342159226,0.6943640918472748,
 0.3014399284534188,0.5477528255328905,0.4908320916450526,0.7399539825528828,
 0.2895694301475946,0.05268832761162434,0.09769798478432679,
 0.017665961717206913,0.2588602675887368,0.9002937584522779,
 0.12457599039862333,0.4204480580049448,0.7256140310997617,0.547468238712115,
 0.08313300867369522,0.42570526850943524,0.03535501208365743,
 0.21945979280355776,0.6921687376842168,0.04360460388978482,0.942524914018349,
 0.5484222801202611,0.9957038876739105,0.8606143014776705,0.2876185921750942,
 0.7186033322504352,0.20745192797125667,0.8833231422245336,0.65287867740617,
 0.7884711649591535,0.6996191040368394,0.11802456958890761,0.6663841323591139,
 0.008470931689256278,0.8036888051772213,0.6999417838123845,
 0.7168483654393246,0.9172862467331142,0.04028052619618383,
 0.12386836898172171,0.6196458130066078,0.12714502095057556,
 0.6342233249911284,0.8215909488760623,0.1205663240461109,0.5626929897252827,
 0.7674378124328795,0.35357765695596954,0.3198817044933071,0.9165784907124809,
 0.18706139585507375,0.260201569346587,0.9552864890333636,0.10020760132653339,
 0.9984632524089534,0.38160980925746724,0.7219684059671078,0.8652695783325458,
 0.23123011040200214,0.2998338267251074,0.09609163084747852,
 0.5361294377881619,0.6915717673718549,0.25262540467886074,0.5691741444707463,
 0.8273395967937066,0.6267303277218842,0.6382483926970537,0.8388708595470954,
 0.6248356057955367,0.12153907646112805,0.48603107318755545,
 0.09810142605447347,0.8354133595412055,0.8384786654372564,0.9871981056813001,
 0.3360743770454985,0.8971317891866918,0.8328562788739983,0.8824971065117759,
 0.05930670840937846,0.2843690926462399,0.14697136509186293,
 0.9924378194493423,0.6558345557164296,0.17214993923231237,0.9601355493969934,
 0.3741791519071376,0.6721047180770496,0.22302213015003913,
 0.16427692016384776,0.23104619911803148,0.41788326535664677,
 0.07789242887160208,0.13544760203736872,0.375494936122583,0.4109829735638957,
 0.4888809543631758,0.06269532637971209,0.15842340939400956,
 0.03169334060600859,0.5778250039258124,0.9393989530053762,0.6399162444829741,
 0.47757935091446413,0.6318368454668704,0.8145028198368667,0.1737923340273143,
 0.9993264726826961,0.5198390427143328,0.22698617280436806,0.4392200029682949,
 0.16319406118062196,0.5641270518673168,0.29963650411548215,
 0.049329697029590114,0.07233603304345926,0.4594357792296868,
 0.6919461556032293,0.08391555783975602,0.20173312807400512,
 0.08869793872371357,0.6988531346445773,0.5471649141489315,0.6869772945790058,
 0.014665918155576607,0.6317351074938581,0.8757419872005646,
 0.4646746293124795,0.789372913999759,0.30214313190923436,0.3858033195915047,
 0.07790757798500092,0.02462370730246144,0.9590067151974686,
 0.1585785097431044,0.10674323827821341,0.9503358107875,0.7577743030704728,
 0.3381883636466523,0.6350613918363055,0.5635620683969831,0.8955371370129018,
 0.46935366577416693,0.9206808410418907,0.6922905814213401,
 0.32887884147058666,0.6744167330594981,0.9792666316407326,0.7243276275293847,
 0.8794537383507263,0.9059770124874534,0.806221406732349,0.6613153880308764,
 0.643521040572805,0.9128718162141188,0.6170449821097451,0.6141651806498415,
 0.5213053978149386,0.9251094885681463,0.41824252449576915,0.4870158495633179,
 0.11069448148020222,0.07713972876483932,0.762152733058651,
 0.12646267332531214,0.8375604544366296,0.6896894015517139,0.5032048162907583,
 0.8471363736297494,0.4151286441032169,0.3998066635010914,0.25262162778420993,
 0.9176787681497025,0.46710394656030685,0.7231178239652806,0.5826690801479631,
 0.5125747420231812,0.3309679019471613,0.16999507797517444,0.4630516269632118,
 0.19729292914748153,0.2654188057463728,0.4282824616372789,0.5129981826927508,
 0.6026001200254439,0.7922131379661762,0.12703439619782642,0.3989761905923461,
 0.20922163634684943,0.36928725219784164,0.14963836742871117,
 0.8130500192079979,0.5082434084348886,0.4655073450445054,0.8766467991350588,
 0.2245854294169678,0.8633373872901517,0.7738568152003172,0.00965392489447403,
 0.2861067359837741,0.6891838854446838,0.2176590066684716,0.9454155491121838,
 0.42045202217715605,0.015680931527038422,0.4876255121580303,
 0.584616584255208,0.3561811428444811,0.3450064097483201,0.7728673896458838,
 0.5640025509557166,0.30263367290677445,0.025023956044372242,
 0.19522137377022397,0.8181273873024728,0.9858261202171149,0.9072845330124357,
 0.379012075614124,0.5239173279916052,0.5667561967530483,0.7054067233390601,
 0.13017536675524777,0.15138748129540547,0.5281659697297335,
 0.8198259309385622,0.7355295874410247,0.5797745109867545,0.96232279838018,
 0.8545789685879737,0.3442406328633171,0.20181745602407553,
 0.48372891749745683,0.14855471826766176,0.08764167797338351,
 0.009804277857402899,0.6781005035620422,0.031917638330575815,
 0.635983064913484,0.008627692793612751,0.8683531490616059,0.5907414386019589,
 0.5586094003644764,0.49503051911314344,0.922212428221361,0.1357497407637377,
 0.9134602314967338,0.3228103768574919,0.40677626391727884,0.1749750072759899,
 0.8565492987944239,0.017507062803444273,0.6205163157596152,
 0.5344713003707375,0.3435713477977691,0.3389167880841112,0.55674342195881,
 0.3444937774896677,0.8107463960779316,0.5410771959507947,0.9431932669213414,
 0.15172433014521625,0.6345472338839058,0.5518200922400004,0.7248921450250606,
 0.26506304754453835,0.32403207610858886,0.38437236180770173,
 0.4024256954088944,0.03956370546346588,0.6678097113336053,
 0.18097316154191045,0.23614291594153003,0.08483612135019514,
 0.16383043770363026,0.21124692042718385,0.9298211512507784,
 0.3427459580507306,0.6876640978190215,0.004269522936144687,
 0.4993296308589814,0.7376961603228753,0.23858580794905682,0.8697362377585981,
 0.14030168853318092,0.11712799817574127,0.4579697243232176,
 0.0033001882413201677,0.40535783250706925,0.5574450662719731,
 0.6387846203055336,0.4621607653809261,0.04034006691582881,0.2920209900184385,
 0.18200816813413234,0.5302445658166879,0.5583637274568115,0.5891590762425719,
 0.3996918614762557,0.5274729713987224,0.8075403954495419,0.8086437025705157,
 0.5731091666626388,0.6048753237200272,0.296324414257888,0.11973635853661257,
 0.8714127979670592,0.14707699664446727,0.9606973725223178,0.9147025142815959,
 0.4310525764593043,0.10884348685013501,0.6100230161038769,0.9677584714232681,
 0.32641140362110543,0.3138883754044185,0.16967729763279416,
 0.3568626387378312,0.6173650218937418,0.43034362956830297,0.408355489912796,
 0.14527485073182633,0.7566484289551156,0.7290904594849252,0.8650464057623775,
 0.17581973605973755,0.938658041872756,0.3700369791679181,0.23504195589218224,
 0.48667710429315025,0.3613006357402866,0.9912136624211779,
 0.007402509633042231,0.44779425925134064,0.4896169263078355,
 0.8724825188507896,0.045420288920417495,0.6785959610395816,
 0.12186340953240227,0.8160742967451161,0.8570999686526392,
 0.09033381425944575,0.584926427968705,0.1707471409437593,0.1282771502675838,
 0.7767492642641249,0.9017654371422242,0.8298988308664412,0.8568811299663028,
 0.9872266463166478,0.43735410508356054,0.0020233096263990724,
 0.062052756418745414,0.8565801246986251,0.4311474565946397,
 0.8682395119504777,0.9065138474801056,0.4954210605887781,
 0.016963175868224267,0.6506314689906865,0.021112839583143694,
 0.6140876349538356,0.5715949211128583,0.8759016647584048,0.81222774472759,
 0.40224067766523675,0.16607044477577226,0.5591886514070554,
 0.7135086469230824,0.732768450549785,0.572497214951745,0.07710413128635418,
 0.828970370977241,0.2507788515122882,0.657476182329293,0.15546391553465932,
 0.5700887420229739,0.34125145935588574,0.18879683359365285,
 0.4045737830281967,0.6730931539457868,0.2858208520565292,0.14023469552655687,
 0.6672396040488018,0.7308506517393717,0.9300720694672764,0.3476718621559196,
 0.5804446254722624,0.9854581180029354,0.4105878017852447,0.6010385272522145,
 0.5515903665763963,0.39906135709185164,0.3827685596508438,
 0.23651918786369985,0.23223400681274353,0.4030550588992533,
 0.5878837833611232,0.49755525221987174,0.4182232992297662,0.8421884793961427,
 0.5669522766135054,0.09821107443285304,0.4905509344238874,0.8824136426285768,
 0.4447984979960501,0.9617229408420255,0.01237439181051836,
 0.005166382718571616,0.37212847651825753,0.8555798578973227,
 0.2293165621654225,0.09034690061819117,0.2540924339515569,
 0.15579148570396528,0.4024678284779275,0.26149988469576746,
 0.7560922255131042,0.3570559633329312,0.03467073609925009,0.3600708722620818,
 0.8290028871025505,0.07369548275450821,0.844376143717954,0.651650519323024,
 0.2720479141899056,0.9206705883189475,0.9183120489021397,0.38678562930554883,
 0.9384682317511738,0.050181384340476365,0.9587654235517606,
 0.8433028944787951,0.4201401136312911,0.1664984790757078,0.1890076278072259,
 0.5622011993543071,0.018166712234188753,0.3644195679050525,
 0.7842744589586838,0.2837131376294195,0.9690152890465796,0.9749467151255408,
 0.5077928432955787,0.03252591109298475,0.7249856209238554,0.5915275850628442,
 0.07859193267410602,0.6235985404323996,0.4406702686302548,0.2170858905741685,
 0.994418539874145,0.8323141748393553,0.3208545361617313,0.40269999457079997,
 0.4058998391875245,0.2969727575957527,0.11762063504706022,0.7050389176425114,
 0.2800076966671552,0.7755529767005209,0.33455876997660217,0.21794716259281,
 0.8699032210094559,0.6746310493755752,0.6213718971636361,0.48117364022016185,
 0.2935712821074038,0.10478595741708596,0.9293553970545572,
 0.32110329390147996,0.37454430852236187,0.4370582046331273,
 0.7702081448208219,0.054319535037799516,0.5042056894177381,
 0.8482284521698001,0.41219124859264733,0.5590600928765418,0.94141280922951,
 0.3263095207808768,0.8547954419488645,0.6769190583064167,0.9712058208221598,
 0.18796672585442975,0.4615346200465643,0.8345484153990115,0.5060696085210148,
 0.5438631525593554,0.41767039938596295,0.4153116237531389,
 0.24848201068529585,0.6017191755570991,0.9255301983142251,0.7562426246685837,
 0.3286903651343088,0.8800129036021702,0.6035557035658282,0.7439019988871196,
 0.6099712741729768,0.7548822321233007,0.2875701868188798,0.6469268410367157,
 0.35256225091198945,0.5442445081950521,0.23243861425718604,
 0.5253946990461857,0.7019108885952835,0.22557921788425994,0.7901471137715758,
 0.8432866525505739,0.10852576240629541,0.13731030912353004,0.722841308567841,
 0.787304705433822,0.5258952324154217,0.8364148095019183,0.13988167942924679,
 0.6135854355496924,0.6095582373871795,0.20739255665615197,0.6356913937631407,
 0.38673449584386554,0.7225043240654171,0.5985209810160008,
 0.012901601413581765,0.3863887344552669,0.1731248129882096,
 0.7797621933542546,0.3692392242410506,0.19793252228922142,0.76647579503456,
 0.07636739955468963,0.8022612000632715,0.9211982692182004,0.1519207192591352,
 0.58932040298115,0.01940393262096085,0.9691458828757283,0.8635268287780453,
 0.06016941778642848,0.23011092080023376,0.2967113530178377,
 0.9172063731577325,0.4428858898383883,0.06925432621022776,0.6503725910681393,
 0.3558109498200759,0.40457936787010695,0.11579379310182047,
 0.7771534990450208,0.3059756004639398,0.8106170101447733,0.7486122600201474,
 0.6887484463098456,0.07822623712985255,0.1840984840513612,
 0.060973034831871775,0.36194109554269105,0.24872106903432467,
 0.8323566629545829,0.558324624663488,0.5133872919789004,0.8720288819602575,
 0.1754335344347565,0.4426213297070455,0.9926736462972382,0.5173526793322988,
 0.02313832900998325,0.5268740155432656,0.35497825541357186,
 0.13311550035149144,0.6763850564929819,0.9611446965639914,0.3983025877434485,
 0.8748245062177595,0.903281747396206,0.3008679597890065,0.8568691650131133,
 0.3660870726724891,0.4511016530832944,0.35716868303097493,
 0.10087436715467135,0.5666301019044804,0.6683966558209292,0.7027926040754046,
 0.02168480092431979,0.7320145305744624,0.007705106841884302,
 0.7905225226105571,0.91788451933297,0.9602937611774232,0.4254963704460678,
 0.9618040491842467,0.8717638760266186,0.8938936154451805,0.9427296257721982,
 0.8554814297972805,0.7740330727346735,0.0193646610918069,0.4593975352690256,
 0.06600868089593892,0.024781944090591868,0.7987200095768751,
 0.15604752581744585,0.4463796795005849,0.17873458018266686,
 0.7301831770956845,0.3703317722732765,0.7720576497145247,0.977763658905699,
 0.031684013784007314,0.5009919456671276,0.40628722650814075,
 0.6886833060900495,0.7161203200105796,0.7947386376873994,0.8561709849259338,
 0.31600891374039275,0.7928810539954964,0.03793043981518274,
 0.32190993231219556,0.5559739726758154,0.8724115978778657,0.1317330170048675,
 0.851195248315338,0.4484265048164119,0.19766233133767486,0.7630789563313893,
 0.4376148087496643,0.6098928572149267,0.35258670796833225,0.8946829654213155,
 0.17145332029087768,0.8129487818653091,0.29780928759440983,
 0.15222566512991242,0.579795028593148,0.6578185148314973,0.6575867793015966,
 0.8655381206222961,0.08762152127791856,0.33874079830489967,
 0.8144648643745679,0.9745942228510617,0.8609543878932799,0.6560123081052543,
 0.2320374964670171,0.49047177031634825,0.7484807886714866,0.704351190697795,
 0.8563981423434619,0.5543290926577732,0.5610729004386137,0.7788658031391696,
 0.6123060244321168,0.4776664775863114,0.3458987153988261,0.1863085793702265,
 0.3061352999515433,0.8672417779805429,0.28723798084778607,0.9385791202452449,
 0.8903166478371027,0.49191490838415897,0.720828134981646,0.3054195465477487,
 0.854621105789794,0.3028584965749235,0.33239697234592724,0.265643804135523,
 0.8777164031736326,0.21045486365326183,0.3255240443364925,0.4755558861953173,
 0.8165608539784162,0.6584912880413004,0.047475329439234604,
 0.7351119780354427,0.23043731373704157,0.40776489632408996,
 0.3487952765136768,0.9615368780560484,0.24653339616356973,0.7890019862696681,
 0.4431908340944024,0.9046408999326971,0.0030187297354067333,
 0.7402594161916941,0.8802143259414559,0.4797002675870251,0.8737027065597229,
 0.3055518845161943,0.40760913439023405,0.11008158839669013,
 0.4377797982344107,0.8627617129870948,0.6719191086767223,0.563302807160974,
 0.9681816918618324,0.1643379810303589,0.1697365404420883,0.6482443748593334,
 0.2749543077321456,0.09461182285261871,0.8756810478736295,0.7053832536950941,
 0.11950651730323614,0.9139263600820157,0.3322530808780144,0.9321330024271097,
 0.9832949253753732,0.06996085822955467,0.3327784215525832,0.8121980895563806,
 0.629631985216521,0.4890403893298696,0.23387573928750882,0.21759554778511137,
 0.5389721369623591,0.6334238038215583,0.5057842413123671,0.1831144425669834,
 0.3924239474490301,0.4906894806581479,0.15776776449813823,0.9285380602752683,
 0.16733120971167914,0.8741539904111408,0.39959877432514,0.587761976402225,
 0.8108793819493856,0.9477163285284798,0.007644423946796319,
 0.46746499362155525,0.5799782142861717,0.9268304896144758,0.6765815758441618,
 0.8353113661789104,0.7472536260527056,0.026428437566245022,
 0.6575274759559215,0.3870965770289563,0.2631613226223264,0.9824014357385251,
 0.29862243666828814,0.4988975702213573,0.6853246547219195,0.0676955795856905,
 0.25407345529503145,0.9295197862485033,0.3893715057212772,0.6128599190832393,
 0.836411265176678,0.16998561491477526,0.6684059866693638,0.6868456778170708,
 0.6700214343837767,0.7064573866780162,0.17424179575882104,
 0.20327618967327266,0.8363405418247738,0.8553779495473058,0.9324310811999102,
 0.6554802560243025,0.7141563233974382,0.8629148935459495,0.146260929940765,
 0.6547380689594915,0.8390081557966282,0.04793024564473847,
 0.19199353950578169,0.6825149310242928,0.08101283836703621,
 0.39812696684674775,0.6712007890801484,0.2996166276931832,0.6329677625809317,
 0.5272746482230181,0.25420680880047264,0.021241574453552348,
 0.8382981063414128,0.6532846365812431,0.3316335535741064,0.3792248042029325,
 0.1351019963848673,0.6929682270367878,0.0020279425177727184,
 0.6599444865462964,0.8942377290478913,0.8666723871898016,0.9925769826953943,
 0.42690427620215066,0.06684409524866075,0.054991737116276695,
 0.3832216269400681,0.9120531998699584,0.03879736061961592,0.9285328604317117,
 0.2779684464015342,0.5136173840785979,0.05952734041331653,0.694300540749152,
 0.9113020078169576,0.47621679417478613,0.27981311068975434,
 0.7656625273734352,0.16528079718778443,0.9452277438524295,0.1130467185954106,
 0.8987436086692142,0.07363821771053869,0.3439329339946786,0.7693753118038329,
 0.02596330358919352,0.7010599275502956,0.037754237665379775,
 0.25971477665297993,0.16422308101892757,0.17697579448402467,
 0.2081899585731014,0.4120730618200581,0.6275795244612136,0.2269525393284907,
 0.33142061432158065,0.9838011719642417,0.0824791557624911,0.8179437047252692,
 0.7667700835450293,0.5397168939013846,0.1194088782647168,0.5624368846105923,
 0.4963857121181282,0.6840560385379869,0.78100902197396,0.9271384811670783,
 0.009644884004162857,0.11471048054969868,0.47672710682918773,
 0.950245833366799,0.8185602667994155,0.03733175370309416,0.12252231083497278,
 0.554092252938117,0.6066001543440331,0.3319001539881852,0.5896339200274312,
 0.4020340450155607,0.8353177729840094,0.06808332917051019,0.3742555344916366,
 0.6282544631785681,0.178016801243864,0.8981383439387213,0.20915786575313844,
 0.3055664696644049,0.8335413292707426,0.2335484122668846,0.6800251726916572,
 0.5514845766924602,0.756620042979156,0.6138996406634872,0.3408950773880487,
 0.7611086376848686,0.014123582541213597,0.6651842825557971,
 0.5038454000884101,0.009223236261037604,0.5395876487447571,
 0.7179547352056114,0.6988547009361497,0.9615674190984042,0.22031279733310583,
 0.8767454037608434,0.5626650997864651,0.3796657038502099,0.5712806579019868,
 0.569716469743374,0.984110603262971,0.04185565880026054,0.63368460381515,
 0.7435147524611225,0.4597436546324879,0.9870795229376271,0.8757763501080007,
 0.9062244570987271,0.5603091370484843,0.02163831116744941,0.6614295392558217,
 0.5345495781235521,0.9370993267342028,0.17738329500889272,0.3602480599719794,
 0.5765680879995192,0.8058828751063559,0.9973106995751505,0.35398914284173666,
 0.8640765800080131,0.4221272283730838,0.4671180912600642,0.37080698279640345,
 0.8867886529229673,0.8194284166914663,0.1820668187996043,0.18861010521097088,
 0.3184839482453665,0.3217336384867201,0.7800566241149871,0.26197161520059487,
 0.5362768291669493,0.3995674202164874,0.3447252075101753,0.6907657187584532,
 0.561038544955922,0.9641542423708486,0.8222251942857061,0.8955913932958428,
 0.586699419127142,0.06143583020337784,0.8438748150078322,0.6264560355155655,
 0.9525268156747915,0.14482526121792116,0.7040884028825187,0.5863381624564307,
 0.6693706229390096,0.5083343265565929,0.5923696093618417,0.11397286964061537,
 0.7994811667907198,0.9842851911634941,0.827336459828911,0.47350829112471093,
 0.4280687780984226,0.07095611386342549,0.5087111559750661,0.8010197348330068,
 0.2518264898273288,0.32005020462851685,0.05396991218570946,
 0.4325260049180797,0.538679398512799,0.8503591043152567,0.6806141151787326,
 0.6640362529532459,0.49094449903393966,0.09170258115060959,
 0.14796123888397583,0.20530850870991935,0.9212182055296982,
 0.7723073712045285,0.835627392866231,0.6472378173647189,0.9889335793641739,
 0.9860303557929117,0.5502503568393116,0.6801308986157527,0.48944727896494244,
 0.2526089307369984,0.43598565880275275,0.47821666517970096,
 0.08585073569120483,0.019451073903305915,0.9479311055283555,
 0.11806253699661118,0.85034798573938,0.017385474890816988,
 0.04649743780156901,0.3548316338703944,0.04591249801243036,
 0.9485262165267385,0.9337027364518421,0.8532653492395149,0.741839754609351,
 0.927596625001483,0.9129253758856396,0.8632026201133264,0.7503171212422778,
 0.37317861217034964,0.6429823391849561,0.5781535207168356,0.5154352149313005,
 0.991430634824834,0.7615177604872786,0.8366013874005674,0.4642060436123012,
 0.5171907923579718,0.3742690056539886,0.44520176558130664,0.6113278132307275,
 0.7990412660466287,0.9040750532886919,0.7367356183012214,0.9699432570957618,
 0.0631535502441396,0.41879155916238353,0.8026268051212145,0.8892475350817581,
 0.31276299634215254,0.2393716585216441,0.20658878011167092,
 0.07377456655319115,0.17659535728607845,0.954314244990974,0.067992797427056,
 0.43647456201781476,0.3766786893585937,0.5272220252831561,
 0.08019568140467492,0.13128309671914773,0.19179607182982017,
 0.48347542204172367,0.7584979956954558,0.010878258674335672,
 0.16607664075956796,0.4803068132605165,0.5078885953285949,0.9324144514754724,
 0.08360791387909627,0.07925750199447668,0.600786300021192,0.2709220557399711,
 0.6830665201055096,0.1214060975557496,0.730960453094195,0.9542825890727482,
 0.6040616008144156,0.06757249884254923,0.9571277016779691,0.1536788270653705,
 0.11355926276747241,0.6044838872644642,0.2203311234058699,0.9649099443332313,
 0.7160916848581469,0.08070343774155675,0.8454378309200502,0.5875375283757975,
 0.26407978867664894,0.3619046782176246,0.12300446276501087,
 0.32216117483031936,0.05290791423154728,0.8077784106346715,
 0.6012397088810817,0.8546433464414402,0.4770413234965132,0.47911571652473317,
 0.5304851320601703,0.06944858292906875,0.5665087542362904,0.9088305522195798,
 0.3641530382923831,0.9170445050577872,0.07664868883364973,0.513448417904567,
 0.4177762136875014,0.3857195307590069,0.5973163745392935,0.09050302702057567,
 0.6830285575363326,0.6319764279041135,0.23004622936444596,
 0.32713868073142927,0.013017084190043215,0.8520613192305047,
 0.9814161338204319,0.10992308164477482,0.3134005371958767,
 0.42521877252419393,0.2855856169214901,0.8277783822992287,
 0.48578918563062845,0.5998511057217912,0.8856792357211486,0.7878542279863991,
 0.9649917531339723,0.7043027128142425,0.6453409630850047,0.9360712083978451,
 0.6567520319064686,0.8666936494661557,0.2921935151307101,0.9366534218677992,
 0.9349260449284885,0.19800872731258246,0.17376436791512484,
 0.30240559844223336,0.06704742421759724,0.9360775462756233,
 0.21374578736577976,0.9278372639525588,0.6240794575920079,
 0.26250769612197766,0.8655023497264309,0.27257299967101234,
 0.9316842569053194,0.9270801496564114,0.5044175999727961,0.07877877384923793,
 0.9761001985569829,0.5609457990266746,0.4290072760724736,0.5923530132436562,
 0.7331198592772374,0.5238532813859367,0.38918091635555685,0.8834274391400212,
 0.3949811327013577,0.057547743998528245,0.1463404981125319,
 0.17424251435835125,0.6703328665067637,0.6421518184938477,0.8204819280593076,
 0.8936182329115268,0.8009907890332034,0.2830436539494958,0.4981281463281243,
 0.6172675964396203,0.5751744247483807,0.5331942371671433,0.5290594343481123,
 0.5087805581161194,0.010757720077938071,0.5099423021279562,
 0.1861055653957031,0.4655433356592016,0.1152598840162522,0.6472580724983058,
 0.04815670102989644,0.4300832494063733,0.41673125514827136,
 0.6179683541560541,0.06932679510260231,0.7131912857651126,
 0.016078854084208394,0.189831803439653,0.08602763241427813,
 0.43451165517445456,0.5353930845934197,0.4314805870944892,0.2741521801316259,
 0.1974748702540151,0.08782188109857791,0.16286450414976528,
 0.25118884478749937,0.29008147548229,0.9716522799646451,0.5736211214651847,
 0.4446538866727303,0.6871617967690021,0.7887891244316724,0.2993440607236387,
 0.275623972037779,0.15330435311644297,0.4897480352013811,0.5896071211599837,
 0.9696930141631334,0.14224295039298462,0.722885751865887,0.03278794961640885,
 0.9860089099660807,0.5602886889438734,0.30910245394827796,0.4440980130757177,
 0.18965526176815528,0.7130520980533799,0.9033540976994783,0.4139407206495256,
 0.5808390773414538,0.934747365215808,0.6414056312616099,0.8706513031015417,
 0.4607044240432976,0.041176759538101404,0.8824072090965811,
 0.8334547311249157,0.8709031662825147,0.8662459073049497,0.3185656906792004,
 0.9517521325500518,0.10670496704928145,0.7056330440200365,0.632349582093344,
 0.17093181194720242,0.3234052632926223,0.8353436679936163,
 0.38569021355626676,0.07954142828289856,0.9744013300317829,
 0.5734903888134191,0.78281944329291,0.6702535787126084,0.40860865627550524,
 0.09400168109456097,0.5769619046974513,0.7654049924876299,0.8199947535702423,
 0.9371495882369603,0.10977798819137907,0.5923369032171826,
 0.26195209519724627,0.09829438940574153,0.41309772333278694,
 0.5042615005806634,0.24879710468625815,0.9122819504950294,
 0.01850930318522681,0.6133827018672036,0.7489487010236153,0.8889524257306742,
 0.5460509987058786,0.9055733753512675,0.7719370980270133,0.5341078914351026,
 0.8801486163157696,0.8707993758676303,0.3817313421714321,0.45387039457463196,
 0.8573325696521359,0.9142593853076226,0.37430950046370615,0.1446732487956527,
 0.7642967871859304,0.13407382212898145,0.02055799562808569,
 0.6301581909578824,0.4820136263518413,0.11455333679528223,0.4773031277668984,
 0.037571309711998846,0.8117634928250227,0.9565051863126799,
 0.24482559793038028,0.75411616091689,0.8869070140112126,0.07873287995247114,
 0.6697779316690474,0.1413327350518988,0.9154332930778374,0.8702148542202722,
 0.8331815246658334,0.15284662445006525,0.9818963876200916,0.6300410923911077,
 0.8491275244732865,0.03120173546413041,0.3158893456756281,0.8817361377219666,
 0.5961177822722143,0.667214944418177,0.8117446195295359,0.28127953790146953,
 0.06018022451287286,0.26328286163928427,0.7172002298066638,
 0.20126181274560662,0.8689726487953762,0.8305392564832068,0.6062895426230491,
 0.19942321828732013,0.22719515887309694,0.18321763507593536,
 0.8736311810492998,0.736386386765369,0.9963830331174383,0.25627132778652584,
 0.6914159656366918,0.9990261254382915,0.23046021193649746,0.6282438055292711,
 0.005037774602938305,0.6135012877655786,0.6234471898562062,
 0.1457180727350581,0.3896497235750289,0.21362714404312233,
 0.13917602133552975,0.9081583846793765,0.21238053395529255,
 0.031332628758176884,0.5007733781966425,0.3938336416909386,
 0.4466330043129443,0.09645943507408516,0.08630035719384366,
 0.4432974291294355,0.005597000679436803,0.6312640434509974,
 0.6287606960275894,0.3416758603023261,0.3587152774330419,0.30362461784796013,
 0.31669483088984784,0.05054084102055145,0.8197878873959562,
 0.45557772664373863,0.12374083801770897,0.620005061705232,
 0.021105350253611488,0.20324161583912825,0.9726078399956312,
 0.7401282544158673,0.7957570189198002,0.5555136528997665,0.9324660108602384,
 0.4298377872834378,0.09081732610782423,0.8249667518982636,0.4152878831320055,
 0.4317436404125883,0.43769884550664107,0.37877225619114796,
 0.9718229866019947,0.8386887966374411,0.32192550079926674,0.9816305069941192,
 0.7872155936303361,0.8670630374437044,0.31683462098519166,0.720591508141961,
 0.5955930797152417,0.9691743875479318,0.29747015844773617,
 0.16238831646470353,0.5340890681523167,0.3308771699735814,0.7774218736281825,
 0.5890309628176046,0.9759548271567999,0.596040879433068,0.5151419010043703,
 0.020016448744397675,0.751343544899008,0.508530619187388,0.4560983961322781,
 0.6062589230965885,0.43692076804234714,0.8818730294180871,0.9065955481283973,
 0.7360213403072265,0.36104091001608885,0.13003756658187848,
 0.3662395566198877,0.912856013167906,0.5532221289023503,0.940712729476624,
 0.2612893655611772,0.4803755214003036,0.399042407598451,0.27248645237679026,
 0.5547112561748202,0.9897491977993422,0.5839206307983955,0.7645049609145479,
 0.9892609962329679,0.543348184628386,0.6075460611842828,0.16873087375530516,
 0.807580923533118,0.7970845561384812,0.7678375394675971,0.9743011722999263,
 0.729742701628159,0.6028876025875298,0.06504848601240998,0.6566980133472609,
 0.02359862506925492,0.36140465715783376,0.5226109632468429,
 0.5925115023366869,0.2803267546938836,0.2754088731753286,0.4986636993039125,
 0.7912877350796236,0.7950692155096372,0.7982245397190714,0.44137915235162317,
 0.053135155966602676,0.25390212286065617,0.6498685199956475,
 0.07998855892584289,0.6323201930091125,0.3135097394674162,
 0.016958416687808286,0.3677429094874767,0.7109082175719857,
 0.0764711129077823,0.6929710851365267,0.18114821046268226,0.571059428260571,
 0.8194445826161172,0.2776713024191275,0.32314587044084764,0.5628767776534392,
 0.29780762755188384,0.3252802927633539,0.25411477598774923,
 0.16502347039744159,0.02511271479591981,0.10314998578576418,
 0.9562400026474743,0.5824579061331457,0.08228660690746314,0.7675675904296125,
 0.876624169263814,0.16575851276466536,0.7491105489951645,0.6254671250507623,
 0.6783871834041959,0.23872725738917855,0.7141046547846196,
 0.23850775860888185,0.4968946140156266,0.5633072381373174,
 0.17042508078485452,0.6441742237850265,0.9024913689024346,0.4277081676342056,
 0.8985123383063028,0.28392784867051335,0.08256129630905984,
 0.30545175174717976,0.6314922373314513,0.10147370104357878,
 0.9282049940969058,0.49649757395470306,0.0397487092364287,0.7593348998366894,
 0.8021767803531968,0.06447708425508314,0.476325918732743,0.25132385116297107,
 0.28461133672559635,0.28070652837621457,0.0025911103725209106,
 0.6803691671550247,0.3032892146180549,0.8526131024817372,0.8598706903578709,
 0.7782068930334182,0.5380867835903179,0.20620866732865295,
 0.36764220547622295,0.5330166935784568,0.2929260526523162,
 0.37277538633604723,0.19368214938408812,0.03483927144352872,
 0.26258295444937085,0.5803905138413938,0.4084299937324556,0.7381919203964936,
 0.752590549083104,0.7098918894198641,0.7222601420620377,0.9424572086752147,
 0.6426634332394903,0.38972145641690736,0.6109743234680902,0.9174953005340012,
 0.4265870865356265,0.5600749696016101,0.8810646253632821,0.9263415451083478,
 0.7829726348934205,0.2033681546067897,0.6627710429160165,0.8465594143442801,
 0.7078939401726835,0.7128098547495877,0.039838095687534825,
 0.3898923485146337,0.3226248239629279,0.9597292337500032,0.10384338471131738,
 0.45372868952130363,0.9281327476604919,0.0685229882763908,
 0.010421284318523849,0.14607712184901045,0.6706530205515066,
 0.4792305555005901,0.2856159086301695,0.00844508645239972,0.2578822012326899,
 0.8485584199749155,0.15503471881787256,0.39245450721823916,
 0.6535298870637241,0.45024037932025407,0.2903660786583604,
 0.43905666981402136,0.5851397459346127,0.26224656570558935,
 0.5932346679848686,0.8496291510607009,0.8213168427598644,0.6622336551712044,
 0.5215537843351972,0.7087502979435244,0.7436060286095711,
 0.006742829571998232,0.28159019408051156,0.6908174727997983,
 0.5488395518995747,0.6339787475852683,0.5287094859161235,0.24249879977051303,
 0.7595293918005737,0.9429737683327553,0.3952938137914641,0.24861119720695002,
 0.05671746459153071,0.6366549573834391,0.6156868052045495,0.6969062249177586,
 0.05056988050195643,0.30816588553360513,0.31155884566209846,
 0.2627198154001156,0.5842739025884999,0.973010674699179,0.6568153909072911,
 0.2905088242780678,0.03260811790679896,0.7383330246846537,0.7156639257332045,
 0.08305217624396932,0.6466466972534904,0.8220918957715875,0.5772077306390547,
 0.3482724951369457,0.6124982950185707,0.7126572518970891,0.8816673443585746,
 0.12135590962486364,0.10204684812525677,0.32229379946707315,
 0.03533498283193348,0.8168154459467618,0.20635079453272465,
 0.06657755627565332,0.08942494209801022,0.018467822546034518,
 0.6501860816066981,0.7360110866407517,0.9086600976148537,0.9288182129102452,
 0.3634591515575587,0.055924636625220736,0.8345344421614606,
 0.09052360405411786,0.8951279473111498,0.5562386084768859,0.4642338288071325,
 0.9868586745272301,0.7673445043731458,0.9621531489672344,0.5993825189044798,
 0.272054307396119,0.2669903479386385,0.9282584884643447,0.18972792106596092,
 0.3457551195916265,0.7508310933055301,0.3844225976473642,
 0.019498006628110742,0.1398043691142733,0.11545818745693281,
 0.2719337506925483,0.14647743588545747,0.31662344783786645,
 0.0017388986138626716,0.5720814959975186,0.3392195748932736,
 0.45117152537285254,0.8455028624031445,0.041688811610432364,
 0.30049414756103987,0.6953389655944904,0.4308613648868601,0.9788392701697314,
 0.9614468657741597,0.9067914403987254,0.9006132745330366,0.06809941433759203,
 0.13561295872824985,0.1863816752765739,0.17878167311828408,
 0.5596213153927111,0.7691814884974806,0.5118912156296668,0.7810404023231898,
 0.9630012368453299,0.23956818316180206,0.25050978119963185,
 0.24757047153727096,0.7048989743974042,0.2364866179217504,
 0.40166242949693476,0.022242715941846902,0.6898035292667855,
 0.11763384773429064,0.7887316826900747,0.4495009914481308,0.2509271683101535,
 0.8805353428581035,0.5699532388424993,0.6604174281502184,0.13807468201759154,
 0.9243220745163482,0.9400688690624234,0.49071916124906434,0.3867585124188433,
 0.7137121425639192,0.420261559859337,0.4863920759722268,0.5303705885604197,
 0.6708046393764677,0.8568650477468618,0.8890969903591819,0.29875076835195113,
 0.6505998726649146,0.7114596181745103,0.6521907658176058,0.698681997399318,
 0.4487303536163447,0.3377021420574402,0.9989173742912816,0.5214907089313103,
 0.08288485254020561,0.7780819285462551,0.426509907602867,0.35970342361924024,
 0.3170804370906407,0.8544559774871897,0.7357642417867621,0.221008677416624,
 0.4268918259694282,0.18661979836878928,0.7748964988938999,
 0.39894311101869484,0.08387254566637747,0.31093866051966756,
 0.5806794541296076,0.21841305137835176,0.10460981009630665,
 0.4676248757691752,0.11714856813945462,0.827686969008552,0.8991216425460874,
 0.29306700161151644,0.3832919950723258,0.1667813301936507,0.899842294803868,
 0.8324459032933909,0.479997636359786,0.09186800051938171,0.3521497485037729,
 0.5107568648912245,0.4413522493813604,0.23974840130393527,0.7790155239843797,
 0.7721416155996336,0.31239844669583194,0.4918104065929291,0.1082577650346902,
 0.4074312522142156,0.383664248902713,0.01577614665493121,0.08174480434647291,
 0.2589606170037412,0.46647578248720434,0.3344467294338672,0.9576493322151752,
 0.9216266814388507,0.98844583593564,0.9023488487407809,0.25062866766486014,
 0.008562525630978146,0.7298904101134327,0.402193234310413,0.7048263782557044,
 0.5051028768382666,0.05673386903546751,0.8497679039709242,0.1117664492020658,
 0.2347665371891271,0.9870063872173116,0.7660690982776813,0.5754150702096323,
 0.6595398854538974,0.9920529506861049,0.3961726227911565,0.09381039899113841,
 0.9460989243299768,0.19164540018980336,0.4772938756527143,0.4529037077149529,
 0.002533102850479585,0.6877400888324082,0.1334751410532068,
 0.1910415503619558,0.31243930114830465,0.8101041612486642,0.9174557331484516,
 0.9740680967550098,0.13366890140137455,0.8802030802207511,0.5124769204689473,
 0.4978365031081502,0.7070521412691665,0.17961768292353952,0.9640452638529886,
 0.8089545161479583,0.8426315065511306,0.5643668203430082,0.8134087223959237,
 0.6097216385533473,0.010546973467733145,0.09750500882451785,
 0.3296983349767817,0.3863073141063901,0.9401536259585356,0.8537869101398826,
 0.15550825137299373,0.14834103391695874,0.9798848415518636,
 0.5129494449733453,0.0511814062162419,0.17547641129183777,
 0.35716087271568586,0.3181613130019161,0.8883945515471272,0.7025084925973468,
 0.9202504248781267,0.3297798542833005,0.22558597043113648,
 0.32201696721651174,0.3081156746880801,0.8134272962662189,0.9493427997375644,
 0.9053890327115965,0.30579286247998316,0.185126715917732,0.456989498134152,
 0.7911309845479988,0.8008781292069242,0.4633920882362572,0.5033254711001088,
 0.3984634930133494,0.671367306014047,0.5989573749926953,0.9267290005026797,
 0.3481390941345548,0.18799930439126655,0.178190383731055,0.2223194531699898,
 0.46416489448106335,0.8205503993277227,0.36048064839012306,
 0.9979463185101619,0.9972048518818806,0.9960110344600547,0.5669344123319665,
 0.2228428915483628,0.9619212020961421,0.5168507293334554,0.18484162858618036,
 0.8390170861751326,0.24800684394860661,0.8557541575417716,0.4800011917089788,
 0.47883546199392135,0.6354700595765591,0.6482047357238417,
 0.18972306845440245,0.18317210997109923,0.04669436058678156,
 0.42223981922851594,0.4662651308651242,0.6812417575653056,0.690327959918619,
 0.2090099761590517,0.7578026521878992,0.02264782072190119,0.7581154338343883,
 0.9130035790309164,0.44524183257451544,0.49677318141946014,
 0.12599845230373774,0.7982636489404653,0.19382864916865994,
 0.2758515808172115,0.2067239183895596,0.569818414939444,0.8738986663262098,
 0.5718351680305924,0.8878228936605884,0.15481373331392356,0.5455910160900256,
 0.933055076854704,0.6841975834969483,0.3701893855152616,0.21181851636061677,
 0.9658708708623274,0.25176770757165967,0.349893723868255,0.5995143831652626,
 0.6606116212959003,0.4530564303201512,0.085655955453118,0.9916951968720646,
 0.48167496327004566,0.5275991414920149,0.14729807018765917,
 0.25148145022420143,0.6899182099066052,0.7516370628728067,0.6750794186691162,
 0.32196795799137434,0.8112881188057716,0.26217063027963017,
 0.3780914178531143,0.9288286580290824,0.10501637612161441,
 0.45938322861705494,0.30345901376280926,0.06450246095117074,
 0.5934579095955572,0.6175365009228093,0.7379766503600772,0.30049335203941197,
 0.9220561699132452,0.7521034704837739,0.7437709954594489,0.6086537723499782,
 0.7455922997934223,0.90574833319864,0.6134192800467125,0.4180384456664822,
 0.3826015644175059,0.1652643736932068,0.07550069495939737,
 0.09822203055226364,0.27638768739395425,0.6999755569130055,
 0.7755592363025079,0.06998152034420135,0.39398498837065576,
 0.6961696118305329,0.19943826441389434,0.6005409948233381,0.2532058837590726,
 0.6118076657758741,0.7067517453975759,0.6756249447488325,0.8607124051611442,
 0.10285303093526743,0.41406332461056095,0.13581374991488393,
 0.8346139810044522,0.4883088132826203,0.6919118236510678,0.21179845904809558,
 0.4680401161444083,0.004756815360556832,0.4482589472839048,
 0.3418876889408333,0.9440281556490434,0.2199683592204883,0.45899516426493325,
 0.822811793193334,0.2372627676881085,0.8735339112614137,0.5958026279856785,
 0.7834521276176978,0.6356302534267645,0.9464360973534899,0.6858574551816574,
 0.021353105681821694,0.2664948611549872,0.15085948387143633,
 0.30007063383125676,0.010617744488534697,0.05361107783565766,
 0.4802892946526053,0.042138349031304645,0.9313798430948274,
 0.0016370053086526282,0.8321298790939127,0.05597945972938223,
 0.911936082311874,0.23994210178837605,0.6852041750322708,0.5565101213277764,
 0.6172455771652519,0.6095791696501704,0.0073856139740258175,
 0.6878127416479918,0.6214264717860298,0.47188654015775033,0.5082037806040359,
 0.8314480701047764,0.5041416126877134,0.6944560674945418,0.08660250966126082,
 0.1268756605055068,0.7699793930132787,0.013026839307459515,
 0.6787251055266276,0.10851162476485587,0.8927049992931309,
 0.14855885961837711,0.9799911728525187,0.8712710082946553,0.7636331757554703,
 0.4695771535356168,0.4225550402288858,0.7698401022840713,0.11551170086652696,
 0.6765021018032623,0.6328827647866312,0.4945727368636308,0.19386935879774225,
 0.2963287739507039,0.3916516045307359,0.18807303948212317,
 0.40823786686111896,0.23983898553543637,0.217520708287663,
 0.48503129265944933,0.5847365290720105,0.9208755829212574,0.6730242539853369,
 0.17188967758602458,0.058541518914294155,0.3386254491273887,
 0.5347853280350978,0.8032716603800731,0.7984975267459047,0.7449548360536971,
 0.8114659303710945,0.8187354218938535,0.5423325221953469,0.5223505083152294,
 0.31185298516853477,0.16166788615057748,0.983341314276338,0.9915382354436341,
 0.09907920522776481,0.2172168748287464,0.8243697117592987,0.9413956735779458,
 0.2986557531310764,0.42848558141972726,0.09267829127647431,
 0.7485806775989476,0.3485481555134302,0.7819732284467729,0.4208228000034314,
 0.23131391006266555,0.5541358129606376,0.8596287557762251,0.603786300462054,
 0.46603203213039146,0.7646539579360567,0.9575512093138796,0.6754856791387273,
 0.9547648852579915,0.9470258213776028,0.07565839693917153,0.1723809414747529,
 0.0023075006704287304,0.6462514734943147,0.9415460009906783,
 0.6083314080346965,0.8117750619917623,0.6380293274682656,0.6351361649004658,
 0.8092189603033455,0.22494466214204845,0.2103788991196871,0.8998380034728475,
 0.8134089030282978,0.9061105506186831,0.1554161109383999,0.8015636942221515,
 0.922378935001209,0.27362429072537775,0.2259157815086007,0.886630745670122,
 0.18751837781684944,0.9677819887880575,0.8071629934364638,
 0.32217554244285174,0.31913336081953125,0.3827791405416666,
 0.9928893856670653,0.6288206386836201,0.29792125246448176,0.4146092165917693,
 0.5972388113487211,0.1439722691885038,0.07644810386782841,0.6065967289850247,
 0.798084280149926,0.0533018422461381,0.08297782208900673,0.5790794463843459,
 0.6202161521461508,0.9998848994880974,0.16124898148765487,
 0.34197215365405165,0.4170228731851472,0.2736253754665059,0.7978739120762481,
 0.14842243589757775,0.3156761659160676,0.49947632220591565,
 0.6439286767976627,0.03077031710994671,0.5968955413297565,
 0.25407583800374955,0.8187719979130431,0.1355013306981726,0.7919112297304478,
 0.46174985990216455,0.32787369699928404,0.30627026136528235,
 0.3732114893841242,0.5058213006001435,0.9677382814813201,0.2774576766740846,
 0.4029019896953776,0.7915486544057416,0.3531312146355411,0.7064628863544372,
 0.4992090135740025,0.649878618296058,0.5091914941471389,0.2930333234996161,
 0.12672433308370823,0.022232579988874623,0.024395684867204892,
 0.15845439533237937,0.026335748642871804,0.3048711587388514,
 0.5954034949904818,0.497531557065106,0.4641335185624851,0.6392684175270678,
 0.6263104832302807,0.09557079043513383,0.0870983714910264,0.581310056670175,
 0.3544371808093567,0.9923300711625145,0.6293459983293127,0.32350329350632434,
 0.6105188277594199,0.9267742073845101,0.1771998473702009,0.5840107397657354,
 0.6450783198870201,0.5095789844629602,0.4536817311768544,0.12118848960090012,
 0.6730167349576109,0.07709871278495517,0.19685178563854677,
 0.5266837270172045,0.7174657312642764,0.8951684437787715,0.9377962807803799,
 0.7552693883120392,0.28438290673413935,0.45512508691174913,
 0.6117356768536251,0.30202549908332055,0.12890283290183513,
 0.6080599462005443,0.2293293434068585,0.28813620269712337,0.5160401277213524,
 0.965509904338212,0.267916919258153,0.13770029750314128,0.741774545391648,
 0.748244102843753,0.3735021171841709,0.8260860249688056,0.5974280444207022,
 0.5921589963525489,0.7597571919700978,0.49001283312778987,0.8269556998595421,
 0.4761230782247935,0.0555885736426861,0.5293275166743843,0.9739284204159986,
 0.9664792660626125,0.9315051941490381,0.25265005878922453,0.3363305927932684,
 0.9016025423616474,0.6668221006034089,0.9174457250799916,0.03531502801891906,
 0.6254030580541933,0.3028536392641875,0.7517638727210947,0.8635226832684091,
 0.9338687349592957,0.829589039451845,0.37417512269628594,0.8725036422515399,
 0.19397931239563382,0.19720124949269868,0.22735885797897426,
 0.9291291711236696,0.7593934656776276,0.5727907805629308,0.09532560292253434,
 0.4930757829752397,0.7251377112246528,0.6236757681516695,0.08721268134520221,
 0.38239638614447435,0.4712536840087671,0.9318129324199808,0.7516962354697717,
 0.3065439904548034,0.4353220511976297,0.9877294525188071,0.5887761669746131,
 0.0452631835666365,0.5349430124011054,0.9555073427393248,0.9877941420340781,
 0.2323656232650827,0.3185766254845306,0.5153599780030118,0.26440918992993523,
 0.13101563330624288,0.8554782166207422,0.8479549819446425,
 0.31651827447516934,0.7156516345729063,0.9693633562450645,0.5948556450294207,
 0.28238992105479854,0.7681201476575561,0.834017230856105,0.0693030619433338,
 0.9903482151488057,0.024888565484521452,0.31616761428400453,
 0.29234549043709623,0.6676127657660643,0.21298205618475596,
 0.9143342041720914,0.6498935077783322,0.49154158584181706,0.7857639247770439,
 0.426662227226164,0.07853972344454274,0.5546312124154408,0.4393208510658877,
 0.3504586129843964,0.845647104368422,0.8898668818663443,0.5933362879831907,
 0.6537948551388738,0.6574760269141906,0.6587357770503078,0.9891632670071512,
 0.4498593785615864,0.6768465116128923,0.4040848298063068,0.8566656839405434,
 0.4820766378221457,0.6855539814187275,0.024614547378317364,
 0.8604172634565087,0.1467902820680521,0.5270762674390347,0.9968484440549283,
 0.6302187430585073,0.3539716612027969,0.5977886034762739,0.7577374005939097,
 0.8474565413713483,0.4291398240439781,0.1642697401769938,0.303298118196644,
 0.12840788868043052,0.5719479910329855,0.8568120484066558,0.5334944354777412,
 0.23746978350172288,0.7075395010680875,0.40139603187200734,
 0.06207504591554369,0.17581260609383687,0.2089054395572092,
 0.23807370262845495,0.7727716985001536,0.8838498163901941,0.8438864496534251,
 0.3561896522179767,0.4476330116295114,0.7998296579012558,0.4078459410436236,
 0.9154914702940938,0.4104278076066824,0.4077356421844964,0.7670597605154865,
 0.45748560037347186,0.2998617454804734,0.03194764137841366,
 0.4680857204692819,0.3157991275297469,0.6308038809900088,0.5332917739454194,
 0.98065970489043,0.5693706440693566,0.36281921432318687,0.7076497688409312,
 0.8324162667079744,0.9421732773520218,0.5197763262176216,0.7626375564846061,
 0.011428434471061899,0.599728881041276,0.10088487112251865,
 0.3550090462419847,0.5846472306973838,0.6921035618402764,0.11481755099825763,
 0.932969687434871,0.861670551982901,0.5975221718137302,0.04898850068833305,
 0.16919139690622342,0.837205623816887,0.6525948059193895,0.04442916990270296,
 0.4643307564194047,0.7833567740675453,0.3826296124357881,0.513816429805297,
 0.7268175566390116,0.37548478320575174,0.15328090382540605,0.297465219026387,
 0.596495069259762,0.38530266679430647,0.6660022506738792,0.6736265741375879,
 0.33299819365849936,0.796571179170546,0.4143833729399319,0.523533222907792,
 0.77849365585784,0.16454374755509305,0.7709509271116273,0.01476980569339803,
 0.5374521515334395,0.4189488042399264,0.17280055756455748,0.5785504517925739,
 0.3183424763946179,0.43951923933088555,0.7420630549765733,0.7105757070514821,
 0.924474290210684,0.09701731660653323,0.9603778325580529,0.20139129124545174,
 0.9584505775073124,0.25151919032237413,0.6004961910283427,0.9973021068903664,
 0.7110566036990671,0.1221945034223263,0.41064710216859435,0.1100489656998338,
 0.1682299779324774,0.5720251356448552,0.41591774916749746,0.8780867158769963,
 0.4085708916083517,0.8086531951945977,0.9605135925273063,0.22669471358901438,
 0.7590784689769929,0.9513444254172133,0.5006721169858563,0.6057918878425611,
 0.36370644889783965,0.6801180379736719,0.24506145905591858,
 0.9525169914944174,0.2565369758667003,0.37948390294473355,0.6411735339607455,
 0.917155487409792,0.8178266617538245,0.17641254465187528,0.29958390316088446,
 0.12421597632076375,0.8452988708820796,0.2395174839337192,0.5305489101107779,
 0.06858055790125073,0.15275076091577267,0.12108214710001464,
 0.47999670981756015,0.35342751846243925,0.7980310848366333,
 0.09576346351847087,0.5527329022624056,0.32620744478491626,
 0.5865203552032636,0.2515792524632381,0.6496089941064748,0.9639536655114366,
 0.11973384565711576,0.8755158631118269,0.005961837550987892,
 0.5163807598358497,0.01720558578639242,0.5061484696702897,0.7927765364743542,
 0.4334208190050224,0.05819236930876326,0.28345442773236806,
 0.4870545968218337,0.9345088496321854,0.6877622834849622,0.3116584918685458,
 0.6687860602818666,0.4263827818906648,0.8136233010682611,0.8588843912479638,
 0.056659698669146774,0.046786800439889586,0.6258750978837957,
 0.7449448744110809,0.0834442787520544,0.9718967225403444,0.2040755276829992,
 0.39497648399623186,0.8299037011060977,0.4994811319980571,
 0.13797236521709133,0.26112580135628793,0.21748516657557537,
 0.5876786795258874,0.8882946884209906,0.851216571048817,0.10004975752750911,
 0.7314831664146595,0.2262127013263756,0.7821785967409455,0.7720148595469658,
 0.05526840848478054,0.6469864636717764,0.9387461887289672,0.7633654435928793,
 0.6233491246020896,0.6807033793647608,0.44329096479159624,0.52561760595879,
 0.6359433009298133,0.3341396248118679,0.9403370717408194,0.14357904096899488,
 0.7339059961973079,0.7018298203793736,0.6721464279262683,0.4147619785812088,
 0.5237723396412641,0.2132775488468005,0.4230866364317236,0.9011977124216889,
 0.69066753726496,0.9233478222263286,0.8782097658676924,0.4869406207301017,
 0.06418626437725394,0.4299459561120098,0.25023122679172827,
 0.21931241488275732,0.43614948675016973,0.11847025486056928,
 0.5939189113427806,0.3490433683387282,0.5889187223926153,0.8875907404433601,
 0.6915241679367709,0.45117339099385945,0.22805841173126717,
 0.8381137088573276,0.9795390896327056,0.19088999498570236,0.8929476551875593,
 0.902925742750438,0.16767258848119726,0.801793952898083,0.8375577500024132,
 0.05546598740192943,0.6183602763374382,0.2726210641351955,0.728162055079592,
 0.058385714382434784,0.5299618782112285,0.3665880834745525,
 0.2958374133607563,0.7709323557419624,0.90426042844298,0.5863044700683745,
 0.48830065552504787,0.7193906677373625,0.5817137527568135,0.8214679972565595,
 0.39374060812971134,0.6111247749150505,0.6552243239140201,0.7749277559063706,
 0.4098264106872246,0.8035870891673325,0.07656784467126665,0.3092851306523289,
 0.9124132834285842,0.21921860273276095,0.4668200087889268,0.9396343588024538,
 0.7610419317720514,0.03913634555432988,0.2634586008468953,0.687774632036102,
 0.5196289244022299,0.19296892192097503,0.05246651319525897,
 0.17393378115663316,0.6702501814375447,0.3416385641824715,
 0.35664759825770054,0.9627263166623861,0.5721841434973837,0.4066378470683071,
 0.1882846075044452,0.6303717878619761,0.8760013193179745,0.0963907549494748,
 0.5942291749156074,0.8271517176196094,0.9788479043711154,0.4250512299767658,
 0.5913410083121701,0.6730193350533921,0.9133999167484248,
 0.040340128723060364,0.513617066641094,0.6143754695391959,0.8551331356115418,
 0.36356848724201674,0.9985929377284191,0.862445882274794,0.43507744094766754,
 0.24470865731270208,0.9479447835301165,0.3369620865446258,0.0727710560627517,
 0.2812859251710559,0.7192930868376195,0.7984033989806189,0.3626905785311143,
 0.0709553058255954,0.216200535378401,0.8714909994933355,0.15309242831881031,
 0.1751410280055562,0.6805591838873497,0.7032766599959955,0.30086837642931286,
 0.4189393080612679,0.012183515694954683,0.12324524032046347,
 0.30620493859618336,0.11487489173662668,0.037764216161859876,
 0.9723703478077097,0.2511049475311974,0.4885282039828276,0.14630553154483994,
 0.5942447766848006,0.8688281848371928,0.1750824952217711,0.04953614313839427,
 0.6176470698755477,0.8461214055151709,0.2829669268006514,0.756946258343865,
 0.38750800065724056,0.9513066941785463,0.9299114766115932,0.789359076995706,
 0.19919392756172916,0.020870293713900256,0.8348136549593547,
 0.8198467080186699,0.01405764918268071,0.7428897203186977,
 0.47379584964175525,0.023381794342431483,0.3627291060871194,
 0.3967242238493801,0.5698249651833841,0.9774867955736192,0.49954008063041455,
 0.9355941920478827,0.44094424477738425,0.2757125075880694,0.9755176943942321,
 0.19564020029739648,0.5688652703002094,0.21579787643988624,
 0.5522999665690501,0.05749039437802328,0.8973725608197483,
 0.45014595606751995,0.2950076603439943,0.244392629659393,0.16853894217551812,
 0.8366343218786328,0.8191728203878864,0.11152741599340965,
 0.27957663003769095,0.9473769132683807,0.4999647836914882,0.5414070734980353,
 0.7974970821394195,0.9063915987271833,0.2928768539587092,0.30058045987800697,
 0.4907623148578878,0.2462878717976016,0.35540921496041766,0.7354083553618465,
 0.7726120791995366,0.4469150875615191,0.9395800785487116,0.12744746940882845,
 0.6120857819882577,0.28622809956388606,0.7133822913039956,0.7398063525645054,
 0.22018832863515414,0.5346253733677995,0.4666814308139682,
 0.07928232649412492,0.9645519254187346,0.2517354681301027,0.7870075893240693,
 0.6781516439647836,0.2742290805088411,0.3730426408590828,0.4321940640472334,
 0.6577949680535137,0.9193914775945565,0.6501785892189103,0.09137764138655102,
 0.9041398925807592,0.7892998590749692,0.1619886544504423,0.8284643842461128,
 0.5760240306917096,0.14395848860228821,0.6565698304285985,0.7542078941017986,
 0.6325985218378152,0.40907778067160017,0.28722714972164465,
 0.16748507365626042,0.44829117179426214,0.2770965102592472,
 0.10049085889650766,0.06318796533395254,0.33305618565074924,
 0.31244866029418816,0.5544150939204647,0.9768699934952679,0.996908268464507,
 0.9202648415645474,0.536220078238865,0.9666620736466484,0.7463767698799069,
 0.9211357574973851,0.1875378409197379,0.40935783415035165,0.7177755422385769,
 0.3311690763108658,0.2718699241120035,0.9989877343721493,0.4584676778630723,
 0.7632803810812667,0.49052610949524045,0.4794445313982366,
 0.22282705581270068,0.5464937049505307,0.4211462994902968,0.950866671377075,
 0.6841292953363558,0.1466656242867418,0.8486149165374538,0.3846265668361398,
 0.2799928047972763,0.6162209295687585,0.7368281575771705,0.04820386280934563,
 0.8481597142634573,0.4860732613563985,0.212111487854443,0.13425535129489607,
 0.6157913936847507,0.6554603597593616,0.06868429450095892,
 0.33825834858188997,0.79300830771887,0.24001137455543076,0.05039026835023708,
 0.7406631109525953,0.9955414478393007,0.6948578162080173,0.7134643144339764,
 0.030360266796674362,0.5834593872030661,0.9642914593846574,
 0.9807844187658519,0.5695470586761776,0.9182828669058034,0.3773392602451884,
 0.5892781435975821,0.6025739980473651,0.5853452789953704,0.8426359380680191,
 0.7270769778415486,0.036325387015156085,0.13479835049439004,
 0.28965903046232533,0.8285637991470683,0.30342838461346977,
 0.4834874015443803,0.4519857573325451,0.47259342268138405,0.7566642358385184,
 0.5419907860580065,0.6655562317877104,0.013908360632724825,
 0.9298197648018912,0.7363290015830836,0.4350258203873234,0.3300815442215027,
 0.823518002075317,0.7448204703634106,0.05055549538305493,0.3632749252794466,
 0.011138146256660097,0.21774732400264374,0.459681952386527,
 0.6561747126344405,0.9490819111516062,0.33198665351901757,
 0.39685065880371495,0.2802738202644597,0.8848209695022228,0.713230513423671,
 0.05383166415349849,0.6786497533199302,0.49112354448589035,
 0.7147511949422165,0.9233129682557178,0.1692640401457015,0.7593973368818484,
 0.4576643900780172,0.41354957414310123,0.5569234882739337,0.6853542770174672,
 0.7565218450839994,0.16254990125888336,0.9532060803040032,
 0.16764836635409575,0.9403671718414865,0.023108598623327814,
 0.40510905488527693,0.39200252861756724,0.22654664828462778,
 0.15851926287209428,0.07094700112708296,0.7061413742876731,0.725525795897665,
 0.7612133627541997,0.6709160927285559,0.7693504845548601,0.13802546552489214,
 0.9054081224164727,0.3061900344647094,0.6507242915564906,0.5885244163678273,
 0.9376278708461324,0.4992893330443414,0.5753934324449534,0.17712068767979572,
 0.08383166006143805,0.05828403400184379,0.585189325133074,
 0.37729160077632373,0.04447548253422151,0.7938187515921316,
 0.8567936240437722,0.27006197229961326,0.011289262650028387,
 0.1438596705938877,0.044174928231346944,0.32620454156818357,
 0.6776964513503275,0.22119971074377698,0.8012414487314635,0.7589508362364927,
 0.05473699302368984,0.6131219098878564,0.429453462083359,0.8527873628202282,
 0.5662123941758581,0.7274079014961995,0.5119481247492972,0.73138958600275,
 0.5999653810721712,0.7574065671807302,0.6746263146371501,0.2577043731167762,
 0.2187404054427856,0.3425846130988023,0.2592452081063522,0.1182954850012139,
 0.7376088424801149,0.3526610012852236,0.7054476256936532,0.22985755167270006,
 0.7328195385445693,0.5507567094409574,0.11501065286084489,0.8310287232366491,
 0.6849586647148602,0.15061240572824708,0.9755815915998964,0.5369162362251241,
 0.3062971667079828,0.5544411480470626,0.05916480384638867,
 0.047500896959601246,0.6195946071614514,0.5725894334028339,
 0.5278480677486914,0.04733823181718411,0.012182946452028331,
 0.7815313842728503,0.060110245709248944,0.9267005611553993,
 0.7627881757970268,0.7357963540625156,0.9199769887487007,0.12373255502141589,
 0.7064302388346106,0.7864004710795359,0.32378630085269156,0.4025478738824716,
 0.008823372759707349,0.6110832577749276,0.3153071454160912,
 0.17722788126834021,0.49028369525419735,0.8401894773975049,
 0.6484122781663516,0.595926914489724,0.08168921617410008,0.13309129084857618,
 0.8347393111816285,0.7670845064674632,0.6859129753438511,0.6394797321435619,
 0.10795058161399096,0.6006725735013677,0.6763362374338522,0.9187534110817821,
 0.8447555199912962,0.7696156143442453,0.9511218759910509,0.4371769632704382,
 0.7779647106125587,0.698062788191003,0.7369173412784895,0.5829046384891657,
 0.5443880782130678,0.45657984838088417,0.0519214233240467,0.5288551917278163,
 0.19648347001901412,0.06961313320123563,0.6686346339587697,0.545647861570471,
 0.8683465856646518,0.42706211927740156,0.6075206761237241,0.7551971009923831,
 0.9923535204259384,0.8890164620537439,0.3716408743964713,0.22946878084358346,
 0.5806191235937219,0.3746215166346647,0.5577855413809439,0.34915544694804845,
 0.2025322193431841,0.27783519745027474,0.5137681087716859,0.8062165400832446,
 0.2597944004229419,0.04275145268621294,0.6282129524571023,0.524232124864518,
 0.15702358335385058,0.3622274571381018,0.43744262903458553,
 0.6652970537517768,0.4732596921249915,0.6765916888371225,0.2038385982350821,
 0.26778926347290055,0.8791847171363214,0.09394101514568931,
 0.7938211778537113,0.7216387703702943,0.26856254188464745,0.9182646372335149,
 0.20990202738112607,0.31486464916262813,0.2680934888780948,
 0.024212935029509053,0.597347198524432,0.4219625478079545,0.3503625380512395,
 0.7229539348359658,0.0027807493325965327,0.4273633558550989,
 0.8141429436877987,0.638990452930845,0.6782323428210282,0.4167200017433502,
 0.12879843520960588,0.2037293715077848,0.24862442275164498,
 0.6005181573778219,0.13722125339260638,0.5868276970939769,
 0.27951078600877977,0.5874400366875646,0.07882257488124833,
 0.1754675004377193,0.4036981024824221,0.4117237737523918,0.9149938071619236,
 0.7202472067770019,0.42648993995536233,0.8382982651614759,
 0.22701393747820764,0.516158962982156,0.9491326514333669,0.18338820043146342,
 0.30009082623072436,0.3193565184106093,0.9910374166866548,0.4131286440606421,
 0.545802944535685,0.045452154799546896,0.5012370906381832,0.9153055178967552,
 0.9158755351309205,0.1630678030435041,0.9736834349743827,0.5753625260570319,
 0.2012800892829545,0.5739945746250773,0.16230942815824523,0.6911147447705146,
 0.13645389276965536,0.36320506157464105,0.7294201418172307,
 0.8870891433712209,0.08138324134622688,0.2619203324988638,0.3447047301769217,
 0.6039287306339992,0.8474613586915861,0.02959360433182534,
 0.46946395124663054,0.8927122808366939,0.31033305018130486,
 0.2581280909553393,0.6230888413754719,0.2035541554850976,0.07797455504726702,
 0.558728346964652,0.949378141973023,0.04677393590844048,0.795970289016458,
 0.3329937320895625,0.9839137376047578,0.8172703014585218,0.08296596754313845,
 0.5752661898802599,0.8202180066191589,0.8117680492752505,0.33488940609747553,
 0.10730749665916484,0.5828326965785642,0.4662121783167602,0.5179089424941339,
 0.33784560784816686,0.5590626430510325,0.373464087298798,0.5081934705133089,
 0.5762716641179835,0.006642267854855555,0.07921430401989715,
 0.2925894594152719,0.2047704516768888,0.5017933850573252,0.5632722230227833,
 0.2006179607535582,0.011477362096254717,0.010365351650852839,
 0.7141731045374106,0.9679630310058932,0.5469412351478179,0.39420163586205004,
 0.5688193494640571,0.877876702800185,0.17080669370228418,0.9041727590455209,
 0.3765790519448754,0.33473348789440915,0.27903199644065346,
 0.08276733477442966,0.45958266637215284,0.2545021955733424,
 0.33493894718622796,0.3890164961921708,0.16173591869911408,
 0.5869626253144697,0.6867153288201497,0.006804876047146968,
 0.8236422696400557,0.406869333661651,0.7705236521394123,0.481068193440214,
 0.5861885424255017,0.3281934080343716,0.007168450189583231,
 0.7569912847130976,0.48937836773057697,0.590933055573597,0.5477004126801573,
 0.03243977847020718,0.557042663853525,0.5641875209481017,0.7098976753162126,
 0.761671843706997,0.39205779530607954,0.16646283555905228,0.2527800928067151,
 0.8301449177592888,0.14947248001037883,0.6617655109055897,0.6938554374654583,
 0.8877672221336053,0.15587353591444897,0.17629938051824445,
 0.1754813004636715,0.9135988533848791,0.869672259270839,0.3228064358429744,
 0.17386464530678292,0.9274906563182277,0.009666870421166651,
 0.561403218448572,0.18337747406928162,0.9894748233783779,0.7845758991402779,
 0.3438658597816414,0.16414965264037806,0.23043740275394764,0.208996537947007,
 0.4001797174102265,0.6527326715336332,0.7945603992589063,0.8784126100729431,
 0.004705308238675543,0.535135434980803,0.13122973479097622,
 0.5764082193986093,0.011987382852337913,0.3266755871432928,
 0.09172475170558081,0.849194037919904,0.8393681212600907,0.3716297235798567,
 0.8622659485186414,0.21149988669840714,0.8982591782640035,0.3643433909965551,
 0.32127872556504766,0.7280940827451715,0.5832522990163764,0.7030091251029659,
 0.5984939749432251,0.9904369512944192,0.6028531616272345,0.9822989258792377,
 0.661389444761563,0.15327095577027827,0.4813149988960569,0.03891282803004681,
 0.671369972753973,0.7152438647126823,0.5155403915624341,0.6302220648544371,
 0.045766056411921,0.9708259365207809,0.6668786617760812,0.9795584903614768,
 0.954739626034467,0.12440697291465863,0.7014787863955421,0.8838511541514249,
 0.9217709262482376,0.20355338085316976,0.8329650079283452,0.6850627590983631,
 0.701313508299963,0.9228532295082831,0.2633565408943632,0.28760561884350655,
 0.46949246116282184,0.35465168788738355,0.656460107948802,0.643676504855054,
 0.2941572533444823,0.43372910592030245,0.7198764350672048,
 0.26783781028902487,0.6115327618987855,0.7127093922719747,0.5296633195191852,
 0.15561327301673777,0.9504263593891621,0.5789334649361251,0.1847251508349449,
 0.3525901233391745,0.5904015633919214,0.9334043616017823,0.9851011112984976,
 0.1512097430910695,0.9870720698749188,0.2672420389488305,0.7757876570807869,
 0.49932040994193216,0.4357445220166176,0.10831444409666047,
 0.7188646786084038,0.9955774219900398,0.934866304012417,0.44507741812758184,
 0.4216980796827563,0.22527067894480768,0.9673641337719162,0.1489627269381001,
 0.27237669693722044,0.8354816711437194,0.5503143807552768,0.8304385988380416,
 0.9303210055111577,0.03685426165713246,0.5998298768822626,0.7140598729445119,
 0.8225658560949031,0.49255364270235635,0.40840151597322216,
 0.6083391628815882,0.7192389668240337,0.704584176036849,0.4873000896669585,
 0.679749689541381,0.05904837422416276,0.029036826958368178,
 0.7072497095338953,0.6455532025906967,0.050228635871407246,0.943154139441911,
 0.13233445688789436,0.526778031184608,0.4942207446211996,0.7943972590343278,
 0.6892079310476634,0.7991539734381234,0.3757193829531671,
 0.025096982719207683,0.8633479674233295,0.9961164676392819,
 0.32854912225122934,0.3794385284158899,0.3574157169439047,0.5851494154546376,
 0.8601727713669487,0.31410922126512086,0.3457699618280292,0.5724324341837096,
 0.8491318225084127,0.21872028581023895,0.47051944998166395,0.646269875494375,
 0.9797829706057241,0.4059110007090504,0.4260980412487827,0.4346435032198428,
 0.401893053845095,0.7477805691573867,0.7837390881769072,0.6658646482743855,
 0.1222155735606183,0.8423316523485421,0.09804582924896732,0.6003135353000565,
 0.3481246268384469,0.14804751617185763,0.365160082399262,0.9460268167689628,
 0.5891379512968524,0.3666168223601284,0.09716849559094665,0.3946070437793796,
 0.05927617999289869,0.18227385391613582,0.4923612022946252,0.965831000443549,
 0.9751173300169609,0.5453299068348957,0.454021200480884,0.4423327393771306,
 0.9196096563836514,0.20659818296591403,0.09434676248673435,
 0.41663981362078384,0.8686335082836447,0.6195425929091886,0.8838905748731096,
 0.9666753637872145,0.16046287015669158,0.43884644269313255,
 0.8748180683689032,0.7979807912719702,0.6559845330855731,0.8133997730129877,
 0.44835640087376905,0.5056293936861008,0.9193113548752336,0.4044211700741728,
 0.21994549128448881,0.5574356987593659,0.18545775713619628,
 0.8107060373713434,0.6617740920586469,0.4086117403165316,0.9716840440442726,
 0.24515089591590677,0.28161353368442066,0.5310305010577347,
 0.7015280449866855,0.27265131482441174,0.49319845466730083,
 0.9290497193342082,0.5431778007376119,0.11100620640564296,0.9008537405283046,
 0.8379049848403971,0.08766124290846344,0.2445754924500857,
 0.24248947918339125,0.19192084056957281,0.32738102259286384,
 0.7640701796343378,0.2763323897703104,0.3532962890319251,0.1322925046430976,
 0.7491750950931862,0.7076179268667357,0.23651504906564025,0.5178382064565925,
 0.22062413674270842,0.528639010841422,0.8919969460712952,0.2440126048928949,
 0.6373473589844971,0.9881125677807476,0.8270404546727247,0.07240064912325761,
 0.7121373679178686,0.7073548949478378,0.11297609302810574,
 0.09238076930809691,0.6893995092809732,0.14772364671435723,
 0.20056730508108822,0.17349333409567813,0.08053111547409852,
 0.43124844111160465,0.46354347191912715,0.41759426488882867,
 0.9991937927971662,0.29682699240861454,0.5238838651127071,
 0.42623096731172494,0.01333824955030527,0.01553255013295951,
 0.24986501173804176,0.5003874945923883,0.10340958941863054,
 0.12478853141950941,0.7539944054597317,0.8650151367153168,
 0.24183482389146094,0.8110560037704416,0.5845951871645003,
 0.08673898336014019,0.9183005907742844,0.7134861339329551,0.3993395364634502,
 0.43057877470023886,0.18945521640136076,0.507683311855177,0.4885117504146338,
 0.09790925312593635,0.8386024961122948,0.5690916259760529,0.8617724593378479,
 0.21344023057370876,0.008226554201941405,0.3227544895662755,
 0.5915054913170962,0.580907580987873,0.28142841090825865,0.5647353720456719,
 0.1553601232025511,0.002512128969030303,0.051677841631856736,
 0.46950662257508524,0.08210226244074337,0.47295043026072503,
 0.7739372028272673,0.8748501998494638,0.34930097712156205,0.2834336358447199,
 0.6944851368568731,0.15741901614014453,0.8496969322603041,0.7769527483142474,
 0.9134317982331701,0.23360490027622205,0.9720537534407229,
 0.37443920504676553,0.5876730999902219,0.68291500669027,0.6680042673206388,
 0.4937288269064649,0.8252401421226456,0.29402552374523316,0.937393030710276,
 0.5041836820708726,0.04481550528013445,0.8401552121598407,0.3140843098738191,
 0.7977874863176118,0.6124246151714524,0.5977015106047674,0.11207726458186473,
 0.18196155036210615,0.8275103598094645,0.985321769413023,0.13036322355866836,
 0.313802265077158,0.7169509327538378,0.9922002923298098,0.0886879567238763,
 0.4595991067784254,0.37263173868578703,0.647440106736478,0.5049100455636015,
 0.7527621415653982,0.8204204322470645,0.9088870066368289,0.8354190759889648,
 0.3545362793451119,0.3460128909469896,0.3317021388379433,0.993011603813868,
 0.040971497551270275,0.7683404221670423,0.7927444263027434,
 0.9585612000008119,0.3197202544395522,0.5383383858150452,0.4339095437082814,
 0.19088702316316597,0.651912163864522,0.27382833952372354,0.5530431249647384,
 0.38093585677122255,0.5164960029846771,0.014979536724584741,
 0.7447555627905725,0.10116957760527878,0.8069962708934639,0.9613805085567227,
 0.11628614026869788,0.5033598839539896,0.9528173541476693,
 0.17952546102605227,0.7479019578057922,0.6701780283804524,0.9246850468065063,
 0.1653058801445798,0.1033891993572984,0.800987310137172,0.9509647899393057,
 0.4913361679430964,0.9800735175774875,0.46316998635853024,
 0.34984031099613344,0.921717389248297,0.9871668777494824,0.9426393820881012,
 0.7627563165873593,0.8579844798515176,0.17324707561037045,
 0.006220351977485161,0.14707649939181977,0.5811452509926777,
 0.16155118047026407,0.1880957960511941,0.06900592082578672,0.855798537050305,
 0.8050748384781413,0.2562741368013186,0.7885180460762073,0.31394833057267213,
 0.9164607044054451,0.35602052146461227,0.39317620899055605,
 0.7475371249771214,0.633689509344046,0.9199921058741188,0.8542092844432475,
 0.542583988560046,0.9860543106025179,0.50864121573333,0.5177291872315725,
 0.4876256541212498,0.8637693644507077,0.21258896304504005,0.9436684286213621,
 0.6915722582021179,0.4990662128488146,0.9738276428569168,0.15688174016740541,
 0.2554354885491572,0.23766112774749126,0.06500445009672373,
 0.9791841372221836,0.8930745647290904,0.12143853981461894,0.2937748785411821,
 0.5598042571356284,0.7814878031243958,0.794610085002249,0.8245716816985854,
 0.46384592590234996,0.021873967458575838,0.5145331931251593,
 0.3522385467735145,0.9391220276553316,0.6807112332494997,0.5534293617060064,
 0.7419665081046802,0.17953871682807,0.2231397920488597,0.4262072927201166,
 0.0690968573476417,0.4730999062701474,0.36544402692625355,0.7324149187362045,
 0.04757332461290942,0.03141929297778856,0.715845324936837,
 0.10984715267111045,0.3192288455474537,0.6544175531000566,0.4303500202257028,
 0.8489519836727516,0.8531597714987997,0.37012110481335214,0.6372915609868472,
 0.48409480062549315,0.7481682474970239,0.7231401116912304,0.7632883663876255,
 0.788173355576385,0.995573735054572,0.5359851511580799,0.6079010958699184,
 0.5934550474299206,0.7699502000392494,0.04564514367924688,
 0.06876483425672286,0.4810576108962036,0.002897108814282401,
 0.806083259936858,0.28057231848947806,0.23206045388103758,0.7788986615414661,
 0.43018939039424875,0.08243475802671296,0.7593333441201874,
 0.5899973936565853,0.7101797199260437,0.9285945345568389,0.6309575491938102,
 0.7600731862999079,0.7451126521698686,0.7485099857485904,0.14651440415554184,
 0.27159663229233544,0.31855740620492323,0.43949525366905806,
 0.23806155279498498,0.38470601302288077,0.999753362738999,
 0.004372921048656719,0.4797724321095488,0.8085347439421642,
 0.5219063856946285,0.6728963872646957,0.06591784765931497,0.7858953446444222,
 0.4633524600739869,0.07493724199209095,0.4700065730481401,0.8471720703787458,
 0.47292751611924133,0.2381227689194223,0.5225117597993583,0.4924407043812934,
 0.759729500260406,0.7335548408860745,0.10627406619655977,0.14363729420520865,
 0.25575551187661183,0.6839620016457252,0.9595547929847857,0.4962539283119024,
 0.3432999809027002,0.5059532845972747,0.42521173341597596,0.9427410872184967,
 0.8395484435223489,0.24349003748953058,0.41326811388243,0.37226611516630503,
 0.14045329921433003,0.6036681236314658,0.35936230901746524,
 0.9655546468893506,0.8873324733758966,0.8104582789006626,0.20781896604622263,
 0.5949920359048755,0.9354064516681595,0.2226484415788711,0.2798160729846474,
 0.12628430830695014,0.6036461625096641,0.8797576506309513,0.904338966156273,
 0.5601622455789134,0.8115237306788256,0.6474351506596865,0.7125180857345766,
 0.5150892054883442,0.5342867472050333,0.8224946619921103,0.4214426962178953,
 0.8304055505445451,0.351220276671102,0.1067328289394418,0.5309329392054815,
 0.6995023624330374,0.869585849512618,0.11190001827816498,0.25848840751328783,
 0.28841806351016475,0.711476183760568,0.10842726855035789,
 0.019409875932569376,0.31724886993686974,0.4133892736193674,
 0.8395229397690501,0.3766318982154083,0.7368486432224471,0.48237731808881823,
 0.7097114935934612,0.6216256842324435,0.10969530643779646,0.6924989164490059,
 0.07547619869052591,0.6975500586702262,0.08155902902322609,
 0.6232805902562875,0.2753998661544933,0.048562005271759134,
 0.6195739860602348,0.8589536468497844,0.49836616884149965,0.3516599149987577,
 0.7557465126831947,0.07153937156411816,0.12155216782310263,
 0.9781876216305381,0.5593494077891336,0.20085688195735862,0.7245476466302316,
 0.7429218637601785,0.5111387022065651,0.7390255236609313,0.36614039223822736,
 0.8405814484092335,0.8379198298399508,0.21233873140132942,
 0.47635173658995267,0.8601366002484498,0.4366906604250532,0.7197618331117317,
 0.5637534042799269,0.5103089614454406,0.47208785202232884,
 0.15306443266582626,0.5420251096555634,0.946952682585333,0.13780508410652226,
 0.6610401271088641,0.15888500505043834,0.4787493379874692,0.2661106968700573,
 0.26249760227284735,0.15313368492742507,0.2279889778196723,
 0.1113914738620182,0.5549556445591948,0.8162245829429101,0.6064652996385637,
 0.4550223146468023,0.6078345778991368,0.15808963235889073,0.9349593560855303,
 0.9253599411505729,0.18557069846033858,0.474353641618633,0.6872014025132547,
 0.5841254232862184,0.9011356321595188,0.8680078985944188,0.29995723732625024,
 0.7208640291365541,0.37834454391084216,0.21370853744980856,
 0.7706893042646001,0.18468110272006577,0.7162281797785661,0.4493865945708817,
 0.16161114672137866,0.2334101888319673,0.5650770587142101,0.1309521413248469,
 0.186142868562705,0.24166649317883282,0.9580808127774694,0.46005313676077375,
 0.2185323527705615,0.29310497695465676,0.8196661540938995,0.0411434618290909,
 0.19626999729266725,0.7575911831353935,0.4589423832158803,0.933458007383139,
 0.6893608755740708,0.6027086767163523,0.5348838213745541,0.05815167435935931,
 0.5473755973898516,0.2817472396046069,0.6041729969782923,0.1897050013538084,
 0.09486110921349056,0.08102958756844048,0.08539941714778454,
 0.259524422884721,0.6841198516804631,0.48419520575774566,0.48572710226249294,
 0.2598547508337816,0.9370088129324226,0.26891448649560423,0.940560382718203,
 0.18224302216175137,0.5489404265505659,0.6510442946550112,0.3763669412345454,
 0.44253176400314476,0.20392954921406004,0.0051735889377398925,
 0.30083676041793217,0.04735460958755522,0.26177397630657107,0.2884773847565,
 0.7189918256749234,0.761712354049104,0.18821359827832218,
 0.059455393256384936,0.5210551560535153,0.5256263337925945,0.993824973461055,
 0.872525952808148,0.10848421279693721,0.3114242852446374,
 0.0019473639663182274,0.5848040186615238,0.12843110591746054,
 0.8162723822410898,0.9109480757261862,0.918912740389787,0.3847214003079913,
 0.9532459880165258,0.7632186463864067,0.8063754502351679,0.42015323000637994,
 0.46221834494018066,0.4830818324205305,0.0817941547999288,0.2549644495511141,
 0.11147572396923433,0.0664128284637786,0.1073412999246014,0.7903597298751455,
 0.34131281813827685,0.9711742791427974,0.24677970919275405,
 0.5111696245260834,0.4712997592510365,0.8514556815906812,0.6413689794991015,
 0.3217385768731832,0.2210160370775185,0.6896525553447016,0.6804424374909274,
 0.08524053701708945,0.2794766692117733,0.7567039158443385,0.9006212093069702,
 0.6175657016826199,0.5902890549690822,0.011437370316760398,
 0.06631300307365617,0.24384018671859942,0.7106936665303247,
 0.5553621981311125,0.958828323560014,0.8021687189619269,0.6230798114661424,
 0.872625214652757,0.9376027392870462,0.4788202546068042,0.05295903782190392,
 0.6231523184679045,0.12986115652126118,0.2322586604326411,0.5363763638822958,
 0.5027000014719368,0.05679914580196943,0.8717265679443592,
 0.49492689443085425,0.0653011738746514,0.08505851568886236,
 0.9672369361590696,0.8595126521432752,0.7319797362119989,0.366257014540014,
 0.6611552477000872,0.8584524932270166,0.3810006445213946,0.0720035922541793,
 0.3511604672109556,0.8271404753940618,0.37283368675241746,0.783647372342355,
 0.5075509128325442,0.6720602121321511,0.19476133405688822,0.5943338888023395,
 0.15557518552974592,0.6009518382013085,0.2981549408509485,0.1876778029473738,
 0.08708457663158553,0.6560797726173735,0.6150012543582617,
 0.20138731559554424,0.6099220419803297,0.2692512039414513,0.7944867387947538,
 0.29658787185046676,0.41307757928019095,0.14217267326287641,
 0.766361572213416,0.8596736009086026,0.3016904787764577,0.009090958973614227,
 0.35254820890106986,0.6802278265338737,0.06796114014203891,
 0.9192431724726974,0.35972195347224,0.33560435193815363,0.9514079080173672,
 0.4278600855514515,0.49903916535641724,0.047523132989721106,
 0.5544356535055891,0.28341968266564854,0.28676873948894066,
 0.2162573325162921,0.8996623472957367,0.7855166601259358,0.4057720290637441,
 0.2496616617196028,0.0017080644089461083,0.0726058136986788,
 0.13709235347724158,0.995841147063157,0.9031959445630914,0.7636343133772471,
 0.4745532662811658,0.0807689950559849,0.6156895963686111,0.6719418481143566,
 0.6625845952141878,0.15774370230690382,0.6006293559589574,0.7206154680595669,
 0.9383455741090089,0.8074008605633559,0.7898870238016222,0.6786583427154826,
 0.2594982372752951,0.9522424056253239,0.6346658633862622,0.34587584027101836,
 0.07361789183760215,0.373398386848508,0.25498526885783956,0.5965911564848169,
 0.7188196624064549,0.8666662051031016,0.9612807195323199,0.5849066891828785,
 0.5976843534027044,0.23943217907115266,0.42615303899998724,
 0.6863530129274864,0.6497670652360065,0.6981128849346663,0.35651544470302154,
 0.6522374928528775,0.40366287951390367,0.49670312398582406,
 0.9259466478558475,0.5347602518077237,0.026250185157486006,
 0.18508144199450882,0.02688720448851778,0.48049799940934756,
 0.7409391569877517,0.985716478773035,0.7833398159978329,0.15353386541596847,
 0.3767462790118249,0.2737748321118394,0.6480585970033855,0.8245306433045474,
 0.9840073654689372,0.0961416308728168,0.3471249976424774,0.3544010295470328,
 0.6185474316502744,0.17109767920470864,0.7073614800813028,0.9459039309493338,
 0.06738029888247898,0.9816268408384077,0.5233787232261797,
 0.22615160844919324,0.47585911564102734,0.4968640424780637,
 0.6064448541232641,0.9395897350738813,0.856171841957915,0.9510534757019146,
 0.6707749297006138,0.09182450707200196,0.666605340316242,0.7491300789998169,
 0.8775716446313025,0.4039174050321932,0.06402851491181194,0.8115675259435375,
 0.9214395419329815,0.26308470281584095,0.3157606622368675,0.7768746772402275,
 0.7159179370848966,0.3018654567593484,0.7540578370784565,0.9279051200961804,
 0.7741964647248591,0.8485826338726911,0.28036485942501244,
 0.36809906730536524,0.2417876980080862,0.054393147818322385,
 0.23914742343648632,0.008914243713907366,0.8669170068653014,
 0.03374293423629604,0.10184818738140233,0.16553738213768077,
 0.3544059614071067,0.4011890461134425,0.1606753976559684,0.15333065354673658,
 0.29431413910200255,0.5991556863810084,0.6979440282906757,0.9793366977551518,
 0.9677643928341177,0.6151749109918734,0.864619437845925,0.6920881279798538,
 0.2325005894129476,0.5580715001572193,0.34360884284055426,
 0.17157014875580712,0.48337256236762816,0.7866148739709877,
 0.5731053077610427,0.7336482459034708,0.5270744961915137,0.47254307629372816,
 0.9408390995970677,0.6156202457401024,0.9497757864016811,0.143340040937727,
 0.6410786247660202,0.8785376614801182,0.34003645150747497,
 0.04948185620207968,0.28841839637461597,0.08787215651606717,
 0.5556396805566082,0.8024102443952829,0.3700814208904507,0.39807089404826046,
 0.6381758889873204,0.12804928801932824,0.12318488760763824,
 0.5035437563983823,0.6600252360679926,0.675273891524911,0.604186074954377,
 0.6849387621928905,0.5121219117092737,0.7316798046852245,0.26708181820459254,
 0.1778759081718675,0.3674551244041965,0.6227941214335723,0.8952167159290575,
 0.5752424730601048,0.7735491585755656,0.433104838620222,0.1532027255559466,
 0.06616118427288997,0.8961321864035372,0.41820128205495566,
 0.4156750232057618,0.5241066974825197,0.8820900204409785,0.672202219564732,
 0.1293570113854583,0.5289911211943827,0.7683991119197874,0.6488363334455498,
 0.2888672531985933,0.668538535809449,0.27380144025867714,0.9176520209586205,
 0.2744263142975608,0.5941664370497556,0.9368564169890484,0.6544331386740618,
 0.5211643714568526,0.3225510378104912,0.20124735087453605,0.4059210562137725,
 0.18275629496527723,0.6839405606878662,0.7510075745393493,0.3853703541193857,
 0.6399377812723406,0.7086327916342355,0.6689147219030853,0.2865814210520473,
 0.42862609657261874,0.728505717650248,0.36407516056083034,0.6167868832911354,
 0.24862191723118954,0.38305632651171373,0.7470614324842194,
 0.17187452791257585,0.6828376171215099,0.7658239906691947,0.8089217122159664,
 0.5300572408160097,0.5656949441212017,0.2037600176486739,0.719469045254245,
 0.8309975900078774,0.13144049377762346,0.4392858759860432,
 0.044067997699153705,0.33164481721817907,0.049649315190361754,
 0.2689861997005828,0.7560373692229446,0.19636521137643903,
 0.45797807210603214,0.7690482138914795,0.5252489415223349,0.9035900290330585,
 0.5633670598010939,0.2713953521320511,0.5173181583362874,0.9296477562442326,
 0.8049302349497132,0.855424034638838,0.5631527215687618,0.5333884672839416,
 0.4167632622808507,0.0376950288133544,0.09815240323117191,0.885217137475132,
 0.4761422235194531,0.8834357847413359,0.10520758368542449,
 0.38736013931910596,0.09683991228135125,0.24060374361765202,
 0.37952934771052216,0.014889084480339543,0.061146406460384384,
 0.4887071257437986,0.32233235445267905,0.0101012227755366,
 0.24087685452308782,0.03229729552526783,0.7210488509854002,0.847961852959631,
 0.31717298079138234,0.8099089780041924,0.05705588840622,0.5713074375749859,
 0.23200650491441688,0.11845732585304802,0.04976579993816066,
 0.8927087522122792,0.6237939288777778,0.22966261307763414,0.476994012998292,
 0.24985365010388572,0.35344635273853686,0.721041491135271,0.5230157689158601,
 0.5779262642575245,0.9438872317444111,0.5046508648840855,0.6637699281920603,
 0.8021703689197937,0.8543051704172497,0.8918505701165315,0.9003305237011623,
 0.10051223298978884,0.7376658043351537,0.738067169798828,0.07158000524790087,
 0.43280622368967814,0.66712085591574,0.2158079102835232,0.5850668612255233,
 0.2726499787861397,0.8288633536669684,0.32456763377726205,0.0850835604302157,
 0.5183392280241048,0.6818455342627001,0.011425127757661357,
 0.3720515777961547,0.1994819192679692,0.359674994107409,0.44738207572365274,
 0.852047676144644,0.897917140289668,0.78172252967546,0.8339907464776956,
 0.9109070949280871,0.9352443362449117,0.255083271299271,0.9539985829248636,
 0.4269407480362025,0.1415154670557427,0.4166342819203741,0.5876093797724975,
 0.7985630849906217,0.8522048243914403,0.6094638458310734,0.5662057192336787,
 0.7225194334270673,0.8904904058141974,0.9027635195847511,0.48428964192802904,
 0.37166188553003887,0.8874158909314291,0.2985825661951137,
 0.36045014523927044,0.6830221679797676,0.2688014987279814,0.2576831714491876,
 0.5470583722082196,0.8895809855003803,0.5790324405637022,0.5855952133038592,
 0.6567512840387966,0.128506890482749,0.6882706120170712,0.43490075532285255,
 0.360298031942921,0.8070503603136356,0.9947574528049018,0.2894616714518291,
 0.7873008615815944,0.6127109186693112,0.5224525750936418,0.7641062155345885,
 0.21134840681172395,0.18385365587380242,0.6653014615461086,
 0.8990881127196175,0.6692506293534584,0.6359762783362157,0.9875111060821145,
 0.6624783181424092,0.024211385699149823,0.6582404999739091,
 0.8896986670223064,0.579315326814635,0.2788199977034299,0.8481812702937148,
 0.10587069955099271,0.45699623872698103,0.14767447739391815,
 0.3080425772664659,0.9135572860674985,0.011140934117348622,
 0.8120473519714538,0.39263258127056155,0.15170862446569788,
 0.6474817346650055,0.014452896282601824,0.20858434850178653,
 0.8677862848831137,0.055118777000502606,0.8517812918871047,
 0.01361254837477055,0.9980935647973831,0.27981151217683875,
 0.0424155311041603,0.1746309809276757,0.09663234266343956,
 0.20261797245553226,0.6350518800061554,0.4513619484292394,0.1987665153649668,
 0.04665305405266984,0.768117083660595,0.6260201487213869,0.10726460663186532,
 0.013850893654379415,0.1747581510045929,0.47963493582659134,
 0.9155486574661783,0.9428047085617474,0.800982108414932,0.2685388299529827,
 0.7997936587498266,0.12311438754872295,0.7256398643389015,0.5438733429181141,
 0.5799897547625261,0.03688686453995127,0.1311485518445279,0.3223721337029394,
 0.3064887443755201,0.5427950829047483,0.3163630085861542,0.7128504889140878,
 0.05234106429757113,0.9656909576655395,0.33046209006545335,
 0.4967327839806135,0.6283496933190277,0.21411445853372546,
 0.23693092552343764,0.976177110245227,0.2959359775132451,0.45305783474142935,
 0.8929364749389703,0.3966462852283922,0.8242146951355537,0.5287625425213401,
 0.0949554587663537,0.3596564600686576,0.0867499672173081,0.08019554705637522,
 0.11261845528002912,0.19616020338411955,0.25957843788167034,
 0.5391794346116341,0.8858599629883029,0.9742102839990427,0.7287720221916034,
 0.44052646775045556,0.3188627707150937,0.1894681696444548,0.2521690076828119,
 0.8131558813489514,0.32741190491111116,0.5243547229140118,0.4486038447794143,
 0.08379658659124445,0.9258384435416219,0.3380555126852134,
 0.45030580405913945,0.9368276671667521,0.09231844012381196,
 0.4363444852071189,0.3840428750790853,0.899310953968458,0.7692916123558627,
 0.3290885863673487,0.6466475033736543,0.7793050266180865,0.06807815291510999,
 0.4626627254093745,0.5343503534390184,0.5683819168029212,0.2877287109904314,
 0.8286052493146892,0.19852417505295739,0.34276058318034885,
 0.1228692729356009,0.9182925451142395,0.11172268606326352,0.9463717446023233,
 0.11178385537896585,0.629586955557765,0.7753848188226183,0.9593117156211096,
 0.4579085521967652,0.940393690459434,0.38515352009920767,0.2868475308656142,
 0.07717770366001031,0.6634217928226483,0.3768289010363761,0.7696182119390746,
 0.6310590707544094,0.5368285521750908,0.23908388166236116,0.6712175602054458,
 0.800363138121974,0.5784291725998376,0.5083312504579891,0.5307296557871968,
 0.17728258971633848,0.5801967031790859,0.8433560943277664,0.5450746688523848,
 0.7500490142463055,0.2023052103430314,0.946028610129612,0.3540883261379665,
 0.22131370530938665,0.5176495553144083,0.08534376053674464,
 0.5815889066478017,0.10633439640105335,0.5656517961865796,0.6554916835097977,
 0.8974006542145698,0.293731125702561,0.45230902387001837,0.9824730456424513,
 0.5516119519680845,0.19687533387195355,0.13504756243294247,
 0.7914425042850062,0.475381555010708,0.7502941905842819,0.1074167701838099,
 0.80243322109237,0.7942157586869704,0.9695564665749538,0.1947913445003847,
 0.8489298839651689,0.21354705410325836,0.6867903724129044,0.925340182331238,
 0.17769757436614642,0.64022087277411,0.4302513705721246,0.052482591782724186,
 0.4163062366083159,0.7048231327435288,0.3266792197566425,0.3014622560322031,
 0.32262903713328817,0.002937119233251795,0.7232859785687467,
 0.19478914834515526,0.031933696422251745,0.11005694951954625,
 0.3237674826428365,0.3881848133368311,0.1426472134281348,0.7693879643701037,
 0.354657803453012,0.823478889787659,0.9183256507019804,0.8657209062851109,
 0.5238813917088776,0.09840349708017981,0.5553228044145212,0.8530775147926065,
 0.6453444036333661,0.28350795786668836,0.6620486857643118,
 0.03521375047964592,0.5074299938316293,0.6221793697811876,0.5728525537430389,
 0.09880049987092543,0.8826843994503276,0.7669441562621416,
 0.29294123287887075,0.1245856915135688,0.7528297794261244,
 0.36878516263807826,0.985542966745039,0.9195623291873358,0.17659222292861754,
 0.47505053017129417,0.7590887472263939,0.18336222504222188,
 0.6518632562512078,0.7459726264715656,0.31805572709007945,
 0.16585498606779603,0.027790496520668784,0.7743815126559447,
 0.6171961448690618,0.39667779695946526,0.9629283451183249,0.1414079025071877,
 0.24872874689311775,0.1467434992610137,0.2378074174228456,0.594522572620125,
 0.05988106858245679,0.7912015121259222,0.7242101828535759,0.9473679656292554,
 0.8373630641634372,0.8307782064696182,0.9216389534107357,0.4062162099624316,
 0.7703693358122857,0.4286061240785194,0.30998377593781035,0.9550911690769326,
 0.589187995756699,0.684892274017619,0.7249585100789916,0.057872706660487716,
 0.02089343381224551,0.8873992867946414,0.9947121335822521,
 0.10314396705406825,0.42705695167604496,0.9179050923147923,
 0.7255802272415585,0.02673832001729981,0.3504864514070616,
 0.13057015051637655,0.15897512264060254,0.3842737103773406,
 0.16132438169036079,0.6449221997430798,0.9793000412030987,0.5480430381799719,
 0.17149168663068282,0.8465994878233828,0.7906186914517122,
 0.42336661448545476,0.8453587937113966,0.05440105874777279,
 0.49968138924973704,0.7350639446755126,0.3037568706508256,
 0.28500951617745907,0.41370815629363766,0.9723899421950771,
 0.17927105845729496,0.9433723827890448,0.8882490144356707,0.5156927025564779,
 0.7966045917252762,0.2925974555305235,0.7009025933057612,0.8026444597153275,
 0.22699998766049756,0.443316268078237,0.6274173969031434,0.07270430176967446,
 0.18022206088748327,0.10158519798192145,0.4690382732585088,
 0.3776914128260509,0.4188626215924174,0.1636343777706475,0.6740622938167548,
 0.36168692976521566,0.8949517428087553,0.12367704208589336,
 0.9700612828397324,0.1044007962482828,0.8568195547276276,0.30357494219918546,
 0.5195206102319795,0.25016766330717866,0.8541870007466541,
 0.42317693718869653,0.6965450985651205,0.6839641373577094,0.9351509609419978,
 0.24796194064862753,0.6398772256616532,0.7108944160116477,0.621403497649613,
 0.5022263767447044,0.9915892668575332,0.6850466669166482,
 0.045507428507883696,0.8100754040313843,0.10430705493137371,
 0.6577008985173586,0.40688886818370185,0.34839145838100316,
 0.1484962888583492,0.32915499554978556,0.11313551928534127,
 0.8347174245270639,0.9727338419674442,0.2694350723670277,0.8434117633545256,
 0.7837718808016574,0.2056589214834037,0.9559572582772067,0.7132535946839154,
 0.1124484187976742,0.38387049368859094,0.33087157497162634,0.394510392650278,
 0.055387885157453076,0.21342525910626975,0.37201176884698317,
 0.5696411445029974,0.4707511704198062,0.8200661928370686,0.5753077544812089,
 0.654476868360409,0.9620420605843243,0.6822504408565104,0.09002177081540419,
 0.19337425848116374,0.022691606302136513,0.6033328905396317,
 0.8248876506743937,0.2297776258484856,0.6815823201030318,0.664671099401581,
 0.9800293844975209,0.7003959553538737,0.4032883694263101,0.7157220930675209,
 0.1533964305864126,0.6046527029865117,0.8295533614438773,0.9275993829643464,
 0.2755798156384993,0.6566427295089028,0.23858382430170078,
 0.06244737849340298,0.4632015918875122,0.1622940869452998,
 0.14367650648322383,0.23723036139236797,0.0970387158965772,
 0.8626556014146675,0.07691695489832906,0.24949856599764508,
 0.6456650329610779,0.11665835881884001,0.9065744556594247,0.3248959765505238,
 0.1971081718563511,0.8870642944190593,0.3494806517490824,0.9375669101616436,
 0.2405520050664679,0.41933370465454656,0.13270659833309972,
 0.3462185725050728,0.13107143286806433,0.9204389691044819,0.6391621993471871,
 0.5072280422723431,0.8299639643416384,0.7787767311489728,0.06395196229421052,
 0.11946527998159928,0.28756428368051634,0.8518851788634321,0.118695866783329,
 0.7823860002850296,0.31715032162342705,0.26444376250213575,
 0.24232812276314097,0.4354581397141739,0.9908028131193999,0.2899770650287453,
 0.8627391105733051,0.3966632942484408,0.9335665197351797,0.7514584195064302,
 0.46969134376203825,0.8078690218569714,0.9485950841827347,
 0.12379088174507036,0.5451000518485455,0.835396741242416,0.6582671189059135,
 0.6204282663839885,0.1175890281518599,0.7027514826379074,
 0.006651689960450513,0.18859233044502643,0.9179668607839288,
 0.761982402815517,0.871407669750072,0.5075826608313341,0.5150418919602525,
 0.12121942064879498,0.749411229646275,0.0862216820249434,0.19323063762651327,
 0.16181347306908878,0.14606791809902964,0.4817268004907608,
 0.8914403394186334,0.9047015548835998,0.2819646061118246,0.8427028515475419,
 0.7108579746259346,0.5697532690442738,0.16110589644628615,0.7218211262214513,
 0.5021988296188679,0.2663600579179789,0.8294987930412867,0.14128959056768353,
 0.44245682600113756,0.04738812909378387,0.8074412606864239,
 0.15985970101681535,0.7381394638125391,0.6597999473937605,
 0.08327081091008648,0.6313076245690916,0.08557359246526797,
 0.02756578887181893,0.20937270707356292,0.2300772839007642,
 0.15559135890152653,0.7764070072396576,0.5262284647948018,0.597745913166361,
 0.9134384756186388,0.9303031807763286,0.5440613727350355,0.4889634326615351,
 0.8539064677482293,0.8407561582759534,0.4252193007966967,0.6274026858850207,
 0.1136447249454805,0.9642246793339021,0.9571623514076633,0.8720270708832447,
 0.54237586058873,0.09714075865804817,0.8085583729133251,0.41532937750812227,
 0.0017551967910828248,0.7263763272145797,0.7934535546249218,
 0.3068702937410307,0.9198596909160719,0.2746899876673101,0.13825605253441187,
 0.8409011259683745,0.7338754587183941,0.028732550530054968,
 0.2710464107538425,0.042141959525885064,0.8581897975550378,
 0.7383183372433051,0.2527614728757017,0.1828052721208876,0.1865414164972825,
 0.15772962591610495,0.5279133629390803,0.5438440625477122,0.4655766844785365,
 0.11665299050115818,0.9153144665981643,0.980315279492726,0.15695667549549097,
 0.9259756617463895,0.32090381773479293,0.02525076158062889,
 0.5767296407149353,0.7433516682092193,0.6247456922221545,0.3616325687969004,
 0.6360837658256215,0.5515949682728,0.9578934774071876,0.30070829602769733,
 0.23531161195506167,0.3945634188083992,0.5092525280744179,0.5349377315143469,
 0.01988610997357454,0.938907437349251,0.5172372321957892,0.35825678698024344,
 0.1932915509348737,0.3446291282753998,0.43210376975256726,0.7399611716563723,
 0.6471835002474776,0.5755809536860637,0.006634409165329647,
 0.3617640965364468,0.14184923839286356,0.7999970039355523,
 0.17690143179637763,0.7822278519755158,0.6305918475470995,0.739814187384112,
 0.7193570515719849,0.7106133061282605,0.4837206818495703,0.09698645973857434,
 0.5527789169172543,0.970243485280342,0.9230321406027715,0.47216728799925656,
 0.13670318318514596,0.6138802190933402,0.0927615682126397,0.5383893352423766,
 0.2919035652255326,0.897504271699468,0.4271422292828877,0.1593291812799682,
 0.26697536257992405,0.776011270601178,0.9375912665822989,0.9779927694469261,
 0.6542151513098697,0.5810463435005297,0.999162314089298,0.6026501837946608,
 0.5396629532942251,0.8282040903465043,0.07807392442629535,0.6993085963343696,
 0.7014135371851917,0.017239572967474137,0.48110040257036457,
 0.43119622007654734,0.5425262801832514,0.6734201372553771,
 0.16306479217459635,0.45923426853695093,0.054333523949922524,
 0.9379233264723361,0.0336608675061546,0.85460848359571,0.13857148042274947,
 0.999640020526687,0.7351632516166103,0.2386596692535966,0.7043783358598041,
 0.9863213231881397,0.5767681801818263,0.05442194592880356,0.1613542077666842,
 0.9622272163979546,0.1601700543021235,0.2563904549034203,0.7928709107129097,
 0.26757440725102277,0.3259503625946337,0.985517351136622,0.8938227624543184,
 0.9688690995095318,0.9369664390897033,0.5836621691491201,0.8269069995322528,
 0.3212253422934187,0.4982754969113695,0.4138365141166448,0.04669358394555656,
 0.5848947627321076,0.4114042566251527,0.44789776447994667,0.7306120653857088,
 0.21204629047307255,0.726428649603807,0.06761004939260351,0.7248342489307688,
 0.26555801917916955,0.6459123401526435,0.33069935625609204,
 0.4557281114487479,0.6387120424107536,0.2398582014118099,0.4543336575888869,
 0.9536725002041684,0.5640110993502864,0.6965692253419152,0.7774704533623756,
 0.014048939160637541,0.2104908526805751,0.6590292497460049,
 0.1393025927132745,0.19303812072325233,0.9162075802025188,
 0.30245384498966144,0.989282987135627,0.13570229272102075,0.8706392449686245,
 0.4722595967662624,0.11843812014937072,0.22809711054659476,
 0.7006054947905261,0.13459364627717907,0.3772728550427962,0.9235772475969027,
 0.3729228645806867,0.8335584415060128,0.619411827052405,0.01357361569777582,
 0.20465728182868015,0.5265553984174117,0.9038765485774405,0.4371401176330427,
 0.8197555305476277,0.9461413909424561,0.54539711478762,0.12077867047437962,
 0.12956641600387464,0.5320104218681279,0.35917931734611075,
 0.18246004560904994,0.5759414975033746,0.031453139112894046,
 0.5134843604218221,0.17838791779888508,0.45137651277132695,
 0.33352998097515374,0.13579650604717386,0.3339141437097113,
 0.9558335506342109,0.4543167710595547,0.21672196620409212,0.3208540481321205,
 0.5790600502281896,0.5034306743155673,0.2845805216602968,0.8030034326172426,
 0.23239872836675357,0.9530357510066021,0.742709092613511,
 0.027897747179680743,0.16336106892649183,0.30283628909205873,
 0.23481251586026608,0.4003005619747072,0.5369590026228187,
 0.29206249790155336,0.4341429552397642,0.6002317456569065,
 0.21031484303261594,0.3273289515402822,0.9496204571044026,0.932974469627037,
 0.08084767307999685,0.5481099455776887,0.3168495364238586,0.9768661528994466,
 0.9755655632139867,0.3674833775230394,0.128154551278718,0.7670594248071972,
 0.1411794896786568,0.9162067731562303,0.8608597482816086,0.7824812675920044,
 0.3004435029414626,0.4603507650328271,0.5655127703381232,0.39773602303029887,
 0.5674150624899801,0.27604888602899325,0.748774567503898,0.9230884418699546,
 0.5928524370942646,0.3119152095320944,0.5955910931599485,0.5281895777673071,
 0.5110247467654145,0.6881033394161478,0.3574878897867442,0.7720479101987927,
 0.5071375237500433,0.06628920000161276,0.3795622031585073,
 0.16573843480137573,0.05246981431764619,0.49690471121896373,
 0.6642667170217579,0.2211204527626529,0.03664398447452211,
 0.018207636434886987,0.5862674587830499,0.7273726373428511,
 0.9444094850223861,0.4897812483626891,0.22168147151965145,
 0.20670354331865604,0.8063146747094133,0.9450050209949273,0.5257684780815953,
 0.3642604414980173,0.12719395762744257,0.21792969552767905,
 0.8040590719632472,0.870685719423693,0.5139993447916372,0.8055708224400825,
 0.5593875542100497,0.9169835560166383,0.3747498498307582,0.5297549969941624,
 0.22612629092101277,0.4568708803745565,0.7857361542600352,0.4978432651602184,
 0.732596729454579,0.49988347152191026,0.8842985209859655,0.9156121345742572,
 0.8690144613971461,0.10361793002141761,0.994875567707102,0.5578520611077148,
 0.5765360235630443,0.5029443001284652,0.35704807692387397,0.8060755998548707,
 0.6075919651498798,0.9781444575835531,0.9183903139685465,0.26828627097998925,
 0.9022206197411052,0.2609841677107446,0.8591376400518584,0.4225693030668032,
 0.03424267434300976,0.5272071820444747,0.7908917469622668,
 0.22318284527488175,0.9757545543012023,0.6668319017709519,0.5591875361802974,
 0.10219910498950457,0.3349163601332651,0.6601727597823972,0.8116148514966536,
 0.6098625526499534,0.4789539165807266,0.2017326997046337,0.2802137391951556,
 0.711002759111048,0.7014599716730672,0.5718061171280947,0.4559069490402087,
 0.5624259888778436,0.29455077437576827,0.29044388164601964,
 0.9114098587872091,0.4270075721924955,0.2436447311484038,0.3546621936838703,
 0.8408838919070223,0.38010112320855416,0.6864472000987314,
 0.17191967408826647,0.7293451931839869,0.15573940443892254,
 0.8223382296262068,0.5592531591373208,0.931415077713446,0.3241612475768304,
 0.6064648687074695,0.2951553980239352,0.031492601270625586,
 0.6517290781112858,0.19512092967166628,0.7474540828938782,0.6577401154057102,
 0.8640659422144068,0.23532351080414293,0.5839112894067834,0.5836482566598882,
 0.9830538030537646,0.8747699455995696,0.4603153821489784,0.846857881941845,
 0.9603773718231546,0.5526319888494738,0.9297917903539524,0.8133919938332155,
 0.8208628860747111,0.11730710611044859,0.7884543716123505,0.8902888368437802,
 0.870646741221705,0.5785990993695924,0.5496940885684712,0.9686389371508471,
 0.6256709101385612,0.22677147212531323,0.9941608387407727,0.3860354863230231,
 0.2361413414966158,0.3912958652017794,0.9112470811760405,0.8129890815416773,
 0.5017944886924713,0.9501476077046737,0.18050390336390432,0.8030690578924575,
 0.1981947167180419,0.8726379072948653,0.5140842055668933,0.21625276921499714,
 0.0991480787631549,0.33152507044383195,0.6102537185848378,0.7935622354009602,
 0.3784949261265116,0.8527455102381081,0.46957975232020743,0.6904919878713621,
 0.3599230080667978,0.013848512560899362,0.3867771009835673,
 0.2520955684192372,0.7329760064673878,0.023873713223884985,
 0.5426958332870699,0.44386707080271126,0.7701065877259083,0.9988029296407613,
 0.37953438129307315,0.37256501094037986,0.5135901886214884,
 0.8480884631962791,0.45946532465359224,0.5907234213323473,
 0.42397688715384896,0.4694232689247668,0.2920939369936483,0.5365595988142045,
 0.9710503098725055,0.24648165181103587,0.785648700643393,0.35282695424489074,
 0.07553752135514746,0.27358685887914114,0.6365224147915893,
 0.05540234685179757,0.09975311395239006,0.7407521227428358,
 0.02199648516599495,0.5297446204370742,0.12741744410969724,
 0.9900560804985107,0.4960364674824742,0.6137955658984958,0.9290796332940748,
 0.3788194889230809,0.3366341075480106,0.026833208175145007,
 0.8840999354731925,0.490282459470179,0.32292258208933133,0.8821287271396886,
 0.783097094909492,0.1773668691299839,0.8871942829778642,0.8182460571362229,
 0.1651014842544113,0.5581517990660597,0.6702306021185,0.4235672686748805,
 0.37504180211838567,0.5337973171861088,0.2260021278781782,0.7786487541949263,
 0.4341104069207654,0.6061000609495786,0.9318889824857927,0.7356742659992029,
 0.9108021477563202,0.33472991360848936,0.5671890772271201,0.8193053216518179,
 0.8410026493805346,0.8981954994239325,0.45725619601922307,0.7389540817786516,
 0.6000080258022201,0.9358103165552987,0.6735108329387881,0.7425986629493624,
 0.4717705751146063,0.30399337570308826,0.8571145499863949,
 0.21994181532448753,0.1735306676107592,0.4485630081023796,0.3527985819852507,
 0.7840824969137458,0.2807478408039362,0.3858464965088142,0.6832680839303853,
 0.8251364538147062,0.3725672745334029,0.8757978930693525,0.2918529111244327,
 0.6509363526033847,0.22806726887186368,0.6935358782063964,
 0.14469245919698792,0.2296924115125495,0.6662876204314967,
 0.20203195505846094,0.5428550836695895,0.31182883462384514,
 0.019419810188960662,0.503641042459505,0.25785702422800416,
 0.6835429987531221,0.6452370997024887,0.006037581877353482,
 0.7663143461436827,0.2795811638500514,0.9788691195631442,0.48952356963801924,
 0.9737492964832203,0.9327641303848306,0.06125298053810546,0.7898483202556245,
 0.8956212523217959,0.043623374215188915,0.6949549985544021,
 0.43599562942971637,0.43177146821553536,0.37065835257926205,
 0.9217663281529861,0.3644637380163003,0.8791039558133557,0.7726915141590577,
 0.10855236929451662,0.6378307579242279,0.0418797462216115,0.7982826779068624,
 0.6920160148410319,0.6972270409298582,0.5952817691376974,0.31253937735558635,
 0.9407618945778473,0.10898362091535807,0.05777807819540226,
 0.8586203542805079,0.448608695833667,0.09072694969424555,0.8207759875506712,
 0.5402416375226444,0.8428753599031364,0.09532839623992517,0.4403617452537596,
 0.7618521083856082,0.6970556994088666,0.2926320487746126,0.38490991641898664,
 0.8684358092684085,0.9065704556964016,0.8880221990932029,0.9237021717942845,
 0.9614356726118233,0.2813623525869735,0.1518773816995853,0.8586318601327559,
 0.7110542315784143,0.09529587701431674,0.26316155322564116,
 0.4078205849645613,0.04114679128910437,0.9677173919632643,0.1907718267507924,
 0.657125762147374,0.530222737335688,0.23703416204338945,0.31291939875795693,
 0.2313710376173993,0.89680192655381,0.48786597961568,0.16732882926487724,
 0.5723359597545173,0.45910359505330944,0.7368685761338981,
 0.04742223067803031,0.5008947322214277,0.2512460131496852,0.7163396552878756,
 0.9577689589239159,0.21426361410491168,0.24482189687625755,
 0.5302852130942372,0.25421973846222234,0.4233590416119316,0.5892612758488573,
 0.39299012357030083,0.5581317556356928,0.8317881879121352,0.6324108659796245,
 0.07992460004538748,0.4858212006836218,0.43713155404747595,
 0.4520002536108121,0.674767731039179,0.6980616779560092,0.3399738443318757,
 0.9581520686394531,0.5180599637815989,0.7586852176293181,0.7167512281695266,
 0.04387987905452073,0.49995590724316963,0.3438265774486813,
 0.37183383665329584,0.6702149421538561,0.9144374157738855,0.929890507759431,
 0.774801941707241,0.35569494554684267,0.8969007392015771,0.9041214596985232,
 0.9591787205074958,0.5048070820382989,0.18651804197570776,0.1195074968240376,
 0.3037506589132203,0.8422924468347381,0.380800704741862,0.8262043494287253,
 0.4404677224110003,0.2587432350515372,0.426043157061322,0.07813462195267251,
 0.1753670480603906,0.5089826187788115,0.5314072034376703,0.7252717400078401,
 0.5500528806717676,0.9113926535659901,0.07351967023206929,
 0.23746545665992924,0.9989779463753398,0.7502766982994755,
 0.19312155502549955,0.11255685956962302,0.8375921923751689,
 0.4793848865366852,0.3823020549688181,0.6135027052842577,0.7480568141233392,
 0.2192379788324076,0.9873453181256326,0.5624605360668132,0.7040078050889119,
 0.49551075836768654,0.4407302350781377,0.2679350991993491,0.6658922565454136,
 5.475259413180167e-4,0.3206849833314417,0.9467081713157848,
 0.8891554647741196,0.4766120956989336,0.09651939634919793,0.2563885753430283,
 0.7838476787715758,0.02531695041690385,0.7089927051591951,
 0.27800353024938607,0.4837743324075454,0.6167454383704016,0.5140790520769388,
 0.2519524594000727,0.5508875360290307,0.7892456898047392,0.12481853120763442,
 0.28177670475203354,0.8128344061899304,0.14776095111572018,
 0.9412932908980326,0.8943213013445016,0.9961311021860606,0.9670136501644969,
 0.7475379192970236,0.12950154247410728,0.14797540250017815,
 0.5757254804359849,0.9083388967672065,0.14769164120016554,0.662603847188905,
 0.3341546323987026,0.42598003304316545,0.5535643807337418,0.9089280983493672,
 0.4077843637718346,0.5670978459492044,0.8254213755088589,0.7073023830881238,
 0.8513019089808725,0.7493557970110873,0.8298448772501693,0.4495480725333738,
 0.8957092656799437,0.49059568874472403,0.8751222446276035,
 0.12291491536063492,0.02824517570021623,0.11109327980548334,
 0.4363127486173347,0.4106653631168995,0.8141408782510544,0.04460369910351103,
 0.007332141604142417,0.9175809822775647,0.33615321123372743,
 0.710820105081579,0.6834042870604915,0.8018611126789452,0.7345648964891199,
 0.6591723289077592,0.23249593434137705,0.8344073628105829,0.1270623836197955,
 0.770500672141296,0.1006628341965975,0.625165478916722,0.06275622153960148,
 0.6907785036752776,0.8015052272360457,0.47792939779952104,0.4484149762525689,
 0.21471611068941254,0.4235480349998416,0.27665828058065145,
 0.9679797912542613,0.5958485167065215,0.17783724093823683,
 0.10857335120467315,0.6053658757916361,0.029625600623022552,
 0.5569343340988335,0.4904962478072624,0.08087577926607936,0.1580224872986108,
 0.17652560345080182,0.5268690067381299,0.3284739099945623,0.5917064480778275,
 0.9695996488627857,0.3651798859855224,0.395491500625319,0.15070649716834972,
 0.9507535605455041,0.5776418799403719,0.83919616021589,0.21392361494539325,
 0.933484207976835,0.8950510959690321,0.12911085935321998,0.38948754321955525,
 0.15299482309531953,0.1555763417151631,0.810203182662416,0.7459949535353574,
 0.18304488526974405,0.17217537342097988,0.19531814174649298,
 0.9298952276974505,0.1583172453653563,0.973136190786488,0.2434804587917072,
 0.21741209567931952,0.17340979117045374,0.6032157184214113,
 0.8837761167499969,0.6221955121750978,0.8365024062116251,0.3162354393913164,
 0.9844435291550977,0.009754334996834979,0.6508623445597973,
 0.4719050452837261,0.8239847225222403,0.9274944062863406,0.775151409158401,
 0.42307054522418264,0.010533457524561252,0.8128123055996829,
 0.7103177874578164,0.35476676417021424,0.7921014467752602,
 0.23299980955499466,0.3931530198453477,0.1616249039824439,0.7206575247419266,
 0.7029795357515759,0.6825043844784338,0.07087058207566932,0.7883727343631937,
 0.8485907510434078,0.2271042426169344,0.30864194145158963,0.9060513519455013,
 0.856973711732607,0.3889923820257746,0.6294005667694186,0.24755621517513293,
 0.7348512508475183,0.5815661271436586,0.21649428658330583,0.9861527497078326,
 0.5596917841433285,0.3889471462770234,0.014288273846976507,0.721414892133376,
 0.22078515397745024,0.6419707663001133,0.308924511733883,0.30393733364132924,
 0.47208465275107336,0.4311526568900307,0.14956655091408855,
 0.24913574373987324,0.7149656218077851,0.597763916704665,0.3599640019440503,
 0.10445429870501854,0.05328676902390994,0.9515172983638531,
 0.5899791624893278,0.7601299582914836,0.08987334671702552,0.9104102614812327,
 0.5176210258290924,0.7184049873110718,0.4165663731939654,0.40522426737230255,
 0.6819337497106306,0.08226972189784099,0.7156646290410094,0.5995403255210353,
 0.8796041442385132,0.39524913621199553,0.7886648544063343,
 0.34413618809009616,0.632314403378622,0.2549042592952039,0.9725109926245215,
 0.8353672595941264,0.7327333550406364,0.583344537023334,0.5526637868504715,
 0.43005608498202763,0.38545714550908683,0.039357714552127554,
 0.3445850257512999,0.20262844148307324,0.8379146739809952,0.4972191258488421,
 0.20823258072912976,0.9247123917529907,0.024519539108855293,
 0.4123288087700899,0.14999701349726846,0.8692475713579785,0.9923037219572827,
 0.5394867650525215,0.45877079418776023,0.5557387553392311,
 0.04809190707767719,0.041117376419726015,0.5314325690582642,
 0.19216153669571967,0.18627602067630533,0.485209548015442,
 0.45954310923985875,0.25869262678490557,0.3795555686042732,
 0.07660174823149757,0.9779228508414435,0.4426162441933825,
 0.0025001868132559224,0.3199251209346158,0.5817020487594595,
 0.04657966506827882,0.34873828326392875,0.365411067990443,0.6601201031808763,
 0.9219560716425982,0.4554938144682573,0.6693976793329327,0.15044291475446647,
 0.8962658183856718,0.36125399538089176,0.45984104056380337,
 0.48570637637494907,0.38516692643058736,0.2734712858592242,
 0.49717263772192777,0.02369262980041531,0.6328883704123477,
 0.0754404415536476,0.14478854660235196,0.18935475471183583,
 0.11511883532878409,0.3681539451840645,0.22948751962723968,
 0.10367929039608348,0.8820350564734081,0.5526146180308557,0.6791335481778731,
 0.32052113791702297,0.5757413576703722,0.7964712076975273,0.2635664648927115,
 0.13593112277816322,0.41300555357610735,0.22020934102420497,
 0.3062541757621451,0.09916555229573665,0.22721762944428114,
 0.2442845485925783,0.7145112721361762,0.6761241306183078,0.48006011891790656,
 0.389008199589444,0.4960408179287532,0.0305945839694175,0.311129763775028,
 0.6613288240664497,0.2868737250118084,0.8452109951564287,0.9930055853628174,
 0.6258028657239437,0.07181579397893523,0.8432621585581023,0.9513347518805862,
 0.2310206294668926,0.24560410245845143,0.8543935990001614,0.2837578261822906,
 0.8008382442313622,0.32847498015482735,0.8222256669897816,
 0.39548194828840144,0.7567508223869739,0.8348343725182312,0.7594893490195415,
 0.6597182190115275,0.10336552606710936,0.12415057966072585,
 0.8379850680636644,0.6928907270577704,0.24190504044177896,0.1248348274599348,
 0.7831687652020828,0.47821335867548165,0.32053402205340176,
 0.26221140043254576,0.902008758816239,0.41945365605708484,0.9780983466678616,
 0.9560386418209686,0.22135956687268454,0.4768606455582539,0.6923985495463031,
 0.08162885278203735,0.08720478730147596,0.016737672198510745,
 0.9401485030591076,0.277647484326089,0.18366719594835812,0.04744404532312363,
 0.10961683834390312,0.05658628100855523,0.9604233876040909,
 0.08772055749943619,0.2789588333383166,0.3096344776954938,0.5711542121219817,
 0.1814821745487879,0.10231040523878399,0.17387191680552028,
 0.3860579014179958,0.8319626956040101,0.40684322370612047,0.0708672783592047,
 0.5943985039419908,0.33474117638729683,0.3902239746127536,0.7811559791005038,
 0.7724594725878664,0.19842982445116952,0.49277182609983017,
 0.10982659406154693,0.6503237998052274,0.3629155555865101,
 0.45089132161734824,0.48748012283932307,0.10774332364267358,
 0.5667275853530304,0.5903753435522635,0.3761999592454004,0.8688916436060005,
 0.8754397825364203,0.2922902177508182,0.2544982141113119,0.2349482622878678,
 0.369145547356834,0.19941136709580498,0.5976968222047221,0.21361160159714654,
 0.32781270542541585,0.06620623811207849,0.4473006282834797,
 0.22242899078408762,0.44916874664126805,0.39938501297557816,
 0.26134487743629586,0.14522875355171339,0.7504724174369254,
 0.3301762700089975,0.5817314976312601,0.32102089300952696,
 0.15690664103620877,0.7375271735588667,0.9077079033662668,
 0.09774464296426022,0.05313142884675848,0.20687751596825,0.6485179868371415,
 0.48157051205518087,0.4249160784771643,0.42105233719422264,
 0.6982234061807301,0.7470445517632783,0.944390118591089,0.1211039234531377,
 0.8977732452118683,0.4209236436205588,0.4489322591381626,0.26637439338276325,
 0.08523588578415875,0.10149361210945851,0.6456947082010261,
 0.5517379891075739,0.642358918327963,0.26927376790969904,0.6434998807372413,
 0.3240614445832932,0.9340867080261235,0.3654139380794825,0.400347739156648,
 0.7014385939383158,0.49074829769537776,0.5748036456280948,0.5666675948203944,
 0.9368735444174128,0.2596637182143937,0.28330905197724454,0.7471290919198358,
 0.4694907052262308,0.4058588812010502,0.4836658812431188,0.3025413502035955,
 0.4955368653860047,0.4748699829006323,0.7456330653211451,0.35664603757602753,
 0.13298981651204345,0.9431045323038211,0.9674355422227487,0.7728749687941436,
 0.5098844281922019,0.2556147561861054,0.7793720288066138,0.6112789880651402,
 0.6821694223110429,0.37134475998518324,0.5528375045285107,0.3697890017981271,
 0.16509933957261813,0.25437290157707215,0.9068240915348419,
 0.6310227576433387,0.8657949480421816,0.8063424422709993,0.7148299453489609,
 0.5300779295052035,0.08079627546857548,0.06256020054072264,
 0.5055638118455567,0.8413992680132194,0.14125644254185366,
 0.03192971388669652,0.40448758456392786,0.07038698898619322,
 0.3892139389180759,0.8985760961035896,0.973576159205916,0.9361736608350089,
 0.694686667532055,0.6337754983006434,0.5187837997424393,0.6259327050539525,
 0.01763266336849545,0.4798521780345789,0.7785903331748714,
 0.21564134361608656,0.11357421995019368,0.3280966310591976,
 0.9432828253382116,0.6561821152603413,0.22821933954076468,0.669394532699915,
 0.40349379296514865,0.1419626810826663,0.37854979451477166,
 0.6664381987417748,0.6701749797913166,0.3391931113676817,0.6423655173224677,
 0.06960458184085105,0.6913837105010661,0.4722769620028229,0.6984630347033551,
 0.02619334982320154,0.6259094603402292,0.7739373593991165,
 0.14431976616855924,0.8329842687680624,0.5787088434450467,
 0.33479512553116264,0.7061399436806158,0.8936133275562498,0.3716277540520333,
 0.6713665031896445,0.6886278535794457,0.8637981213232104,0.41596548548147705,
 0.585397198710829,0.4792543962509157,0.460851035095612,0.4003617936292345,
 0.5015834300891172,0.2856214818825539,0.1020075675068185,0.7910678569647327,
 0.4076421239824729,0.516185507761622,0.9626693031595627,0.9523305557722082,
 0.871310447363147,0.12005630467852468,0.309025941326207,0.7826933919715084,
 0.2459101472166796,0.09847289182019603,0.8571496463540921,0.8174801596915712,
 0.21339528578617495,0.7041799600064598,0.7708064445184906,0.4438666730469699,
 0.6344170381744211,0.9583713554870349,0.07644600276729918,0.940758067603064,
 0.3294417379895226,0.25176220456245846,0.8032806331096447,0.9386999555726782,
 0.7216786257895786,0.28496565731343504,0.6651739853109864,
 0.17361841881103746,0.24171795365544124,0.2182084474273368,
 0.9756709383445694,0.8130793191105442,0.8979459028086844,0.5934606744194859,
 0.6417266644980295,0.17234029510157545,0.07169929453942037,0.531526295904612,
 0.29553731226648794,0.5205938138246218,0.35826986066948807,
 0.16704037756276302,0.4252395924905541,0.08413867195129476,
 0.1434622014455964,0.6466778816928249,0.1400324918215149,0.47417852562937657,
 0.9897884840516006,0.29366799093509144,0.6415294606719468,0.6048140856693042,
 0.4063752769016502,0.3261233166300318,0.7720186694682393,0.8758737865923225,
 0.5351744819801965,0.29039695454609193,0.36800766048341926,
 0.9322859230560101,0.06756925315552709,0.2674937295541804,
 0.37263793031252124,0.437833304600053,0.9869160744680847,0.9767483684048831,
 0.48369951842839476,0.8780563100068177,0.9831065892085712,
 0.042655908747051496,0.5450330823885939,0.06679332430355456,
 0.47963824236472297,0.08740254807542791,0.6839622156829738,
 0.20185166398747667,0.31831843464778464,0.7835055430372263,
 0.8318754568079743,0.8972277641495562,0.6455172557684323,0.6213160619403008,
 0.46737153176353796,0.39826659194249925,0.26266802669286116,
 0.6982237275868665,0.2619383199439246,0.9678111147436705,0.7901943886967783,
 0.578298166817811,0.5071437694958663,0.4471090739499425,0.51549900347162,
 0.9083760059875214,0.4716568642069674,0.8888722650078489,0.4417584094373418,
 0.3140238819391208,0.7419990286469249,0.11964529867408336,
 0.017909172301772358,0.8716119981490503,0.5118459992676914,
 0.5419016183918689,0.5128377307408768,0.9191782504053646,0.3200130309209994,
 0.6384322327385706,0.013358074656820151,0.37438237199543256,
 0.5534884848778587,0.16287310454694404,0.2878788590776368,0.5438727736651356,
 0.7727103992831996,0.3780339108529843,0.24866632854976511,0.639051940858482,
 0.6351486062434483,0.6837993689385478,0.101126380946154,0.013216031220972702,
 0.9818013050327348,0.45845722938902445,0.9566858849375927,
 0.016587536303945716,0.9341058742647819,0.38852510164085075,
 0.4205170385051785,0.9049392620751731,0.46432019213475495,
 0.08147811918415104,0.24195927432935238,0.5523215319154362,
 0.05874815424461299,0.6428482350707075,0.9397161480033468,0.2758635608260659,
 0.39466149730335465,0.7160940587843742,0.789258414725713,0.5832796316287465,
 0.164869138193124,0.8774972618680832,0.9241831358801352,0.7689426824483833,
 0.029785994371493874,0.1112938946579487,0.02489096224177645,
 0.13304714271551144,0.7533937755535303,0.9182824840138026,0.1691888892299055,
 0.9094668769632763,0.4739558395783273,0.6860052452229257,0.9810038385906478,
 0.6978989207328772,0.5852602282138863,0.7953797532551596,0.2574849239113417,
 0.3399994596110636,0.28616704538149407,0.42818633149019447,
 0.8139828624867086,0.23029340124676334,0.34314401565698016,0.806671846338158,
 0.28478653561244727,0.31914900271308355,0.3198818789521458,
 0.7300928674527518,0.6743601989446409,0.09975986794547864,
 0.28346865807034505,0.5239982020494905,0.16443480048291215,
 0.5691719218803099,0.13483707500241404,0.20203027667270101,
 0.4659937627020412,0.5763751824933184,0.08070061812289342,0.5634687621039027,
 0.19633529963897034,0.6745648892328447,0.8124056431165846,0.619740497152844,
 0.7363258981133176,0.7594675312215249,0.9302818876390458,
 0.034960555821609285,0.25754092325123346,0.7893411479631738,
 0.9888081417320826,0.5589199966495698,0.2702825132793534,0.7812161095438221,
 0.8757940506224339,0.0026507350727315027,0.2667389747455302,
 0.7347070361626752,0.648999865646086,0.40754651736832725,0.7023983813711281,
 0.7457219639765484,0.9708108992556657,0.04087296181590283,
 0.08419853145683298,0.7339436484179857,0.1710164591036658,0.2266166988222197,
 0.272975067911427,0.021597627903039318,0.1043175221162167,
 0.26711242098195864,0.8562830509777117,0.08281892449454364,
 0.12913155423627964,0.8585261695086339,0.24195754414600557,
 0.7669522865072542,0.002567495776119033,0.6460196983126476,
 0.7760583855147161,0.7883728152216378,0.952592985340497,0.41378708019988186,
 0.36099972773680444,0.3103163302068844,0.6811237537525372,
 0.47257832755514606,0.15033103745733556,0.27449759731405987,
 0.2896659155035517,0.4713623056569758,0.6232305115041569,0.14844505871919167,
 0.9713691321905638,0.142966963941907,0.4615829854091036,0.3019123653176573,
 0.13150320933352178,0.30583926920837434,0.4389959862803742,
 0.17297818250102637,0.26571852317840694,0.5749278566167881,
 0.8643869425257134,0.85023183972809,0.7447426416901547,0.559455268259014,
 0.6102017577896738,0.6530132870092605,0.6908365685033927,0.17065903598808596,
 0.7462373530558188,0.2731299793787769,0.840746772268214,0.44767998615006954,
 0.8794780881507572,0.16603750866292666,0.4216360608922115,0.8201927525291244,
 0.8810441063045871,0.4569065214777912,0.12070597115314152,
 0.061433594697837046,0.7957733506431456,0.7791259174178646,
 0.8913521696286406,0.7316603894277387,0.12494228949114994,0.7116958973249763,
 0.5819074440978227,0.035957066697594575,0.47686373779594116,
 0.8839195457794048,0.41581043415194174,0.7916471879413747,0.2181229117505613,
 0.8879632762262395,0.028779287963991762,0.18929878784928256,
 0.23935269119462554,0.4173936162525149,0.32960432048361055,
 0.5335560664260959,0.9155059199822408,0.5896104125908949,0.29069942944928084,
 0.5231253536255001,0.9292608975991616,0.8184973020893977,0.4883571787542563,
 0.5648834660558698,0.044085879106067316,0.7322588408388502,
 0.8675679848912751,0.8800273067894293,0.24450368785179855,0.6922360084059394,
 0.3512794241656142,0.6917270213817307,0.008435440700521823,
 0.13262550581372878,0.34041439313459,0.8915438242481125,0.5041401805376587,
 0.5098704146732943,0.586049345440637,0.2881995147791774,0.9460110110279074,
 0.7140369713231158,0.1598412765724645,0.8753176148578311,0.44770876894806366,
 0.6846355301700514,0.05987917229412465,0.4902276940454552,0.4924047037551623,
 0.5471452841010938,0.5459338612100458,0.13896706782653379,0.0723028626755533,
 0.9993611526108297,0.49133186498575465,0.9119297819199382,0.4090365911877385,
 0.46239981063992,0.6157136890453095,4.002666162142843e-4,0.5833619566073605,
 0.3763309210510669,0.1566674272040287,0.944301659954979,0.732060254369929,
 0.2392237026303864,0.3704383315535389,0.2400517992159844,0.3619508880610126,
 0.9518396225709885,0.1887268364828052,0.9383290025495189,0.2751862808119787,
 0.5821362970799077,0.36769121897293133,0.5474149411484166,0.517330887811848,
 0.5914582563919323,0.9797545669774549,0.34198429083534143,0.7131217673426608,
 0.2581306779404372,0.824894695198791,0.7296604943246853,0.13127492810089458,
 0.7234914573926259,0.6248763875029548,0.502387536127833,0.8244301534835855,
 0.24817185806940623,0.24803609623004985,0.4853230599231858,
 0.9745421400644119,0.10222347402086929,0.28090721526929663,
 0.4540289711779797,0.8900666635723462,0.2929139540429865,0.8429774481757022,
 0.2584646803509484,0.34768631953517226,0.5745254447063258,
 0.09955218726681636,0.18994337879795353,0.6177320511315921,0.4406161555186,
 0.01232045469357712,0.8103611848291735,0.5451981656350394,
 0.39721821034757965,0.08557778441528407,0.7315232365416211,
 0.08942077522423553,0.9290606105850803,0.5655074385473517,0.5944015983738625,
 0.5502929118532818,0.3621608361829952,0.4073478519330319,0.03391342055425872,
 0.5132360095812577,0.25297114850279456,0.08844265933331652,
 0.6520617890710196,0.1932082211345112,0.8463474192293053,0.18848608437495962,
 0.8706663603138474,0.3021114665253962,0.9907852161019984,0.5965183463295203,
 0.16598002984339422,0.3332673688756148,0.09897188272841828,
 0.5001968330498038,0.1085812726743135,0.49687929363083916,
 0.17898475642709577,0.15689887823762483,0.5069579446463728,
 0.4663712064606952,0.9163659428922213,0.32469640829006663,0.7872078929124594,
 0.4562154621956874,0.44321041725855215,0.13773931026421882,
 0.06551665639284399,0.8023656790279279,0.8993151708126614,
 0.30397401872729046,0.06717023611719852,0.3090874366453006,
 0.9344024577735296,0.8390714870675001,0.4061649100815684,0.1209502864971499,
 0.39280653128961374,0.8368742849404138,0.7044806506465041,0.6533425314832231,
 0.9155520565449642,0.3738648840195261,0.1833216749763349,0.719264244539666,
 0.5295565723229877,0.5844440584358461,0.9412135880564608,0.9715660709995146,
 0.1831367676672344,0.7572314540827795,0.7839665318418393,0.25379372075291895,
 0.3373205632249444,0.7797853360128917,0.05202624022721514,0.6601359065452443,
 0.711555650668244,0.8358879858788595,0.4450535631874196,0.24348828816774204,
 0.03513338795748655,0.4676699604477459,0.09204936712084599,
 0.8951261496605573,0.9918059654862008,0.3239867877756246,0.18149977997418787,
 0.25983333852715607,0.48396494871133944,0.5166365993801609,
 0.18438250796324163,0.7711388883043386,0.6437160351662117,
 0.41381309840871294,0.1477306815261964,0.24008656216088076,
 0.08761108960464381,0.18559951022123544,0.8051315547950099,
 0.21632607025236483,0.30920823618150517,0.36436306778153205,
 0.6809014663539391,0.9063214545915076,0.9213353737318313,0.6154543988340965,
 0.5434855577775364,0.4701925282403189,0.35026895870869823,
 0.33401587558371815,0.22496647030355743,0.031155654895801743,
 0.2957160656246174,0.048173909828682304,0.04539975667470153,
 0.7212096542293388,0.34095971230822675,0.6338809439185026,
 0.26181509584393536,0.48819881915938534,0.7173763545496391,
 0.6697160409712442,0.29224056508136964,0.0048306807020530584,
 0.6386480628031161,0.8092733666919922,0.6268756727302653,0.13711645852775833,
 0.9059643185699032,0.02629750996943847,0.03035885654198467,
 0.6495265472047553,0.33060352394819614,0.5455839903123515,0.33618339641606,
 0.5618491697461367,0.016476698896917386,0.20859218950697822,
 0.11305157916231301,0.5288126181957905,0.7475126505449021,0.8555064963702463,
 0.40960996329040134,0.814778097920966,0.1327132932882833,0.33145276139352164,
 0.8550680005139405,0.1407211834476778,0.3314683637775012,0.30938313586647714,
 0.09463786793158302,0.8585829855967214,0.142630759847734,0.5004339749955433,
 0.09228465980344114,0.9491753335624286,0.44774132023461644,
 0.43535618907568086,0.319853506587812,0.3331255057665903,0.12491716931530006,
 0.3861277061046279,0.1837051357257673,0.02752441376970516,0.7488469881490455,
 0.12579554852742336,0.7443555225111972,0.772988281877603,0.2600056675433001,
 0.3476326511163389,0.5158619005874019,0.6882508153614859,0.21876986916922692,
 0.6104665467940832,0.8170512669405582,0.9634385190323922,0.9413738135684062,
 0.09289566443272235,0.975031968773077,0.7300547930181912,0.7870083083452495,
 0.6620928301571807,0.9581560506661132,0.11130473277290831,0.6062903186515548,
 0.34232028315962704,0.8079159151725502,0.47508316902065684,
 0.13997398525101423,0.6129583092780349,0.0980521695849137,0.6364547969085991,
 0.881142917798156,0.7571108508250164,0.7691264327862302,0.5695912160332516,
 0.035820889316523274,0.3313598468456592,0.059133528351217435,
 0.6297883993718282,0.4653439083614066,0.33078945220826594,
 0.28253731496537104,0.9786296980290324,0.48929850124008856,
 0.8446600431788788,0.3834317347470443,0.5650193992662136,0.3902963935478646,
 0.5150645745588038,0.3729013739615319,0.9580465356036242,0.8975321372366502,
 0.8599542681091761,0.27580578293406777,0.37325267548077556,
 0.9601755105744019,0.5637843789000672,0.5165142442686305,0.8595070574619323,
 0.9281205851686363,0.8371737603961847,0.010172355475817296,
 0.42839254183045294,0.9799751055429694,0.37243797989869254,
 0.9964501199827835,0.09883264216848453,0.9181396152078996,0.5718528593227722,
 0.2580086883533179,0.29175430387381673,0.7389346207589199,0.3611429355774035,
 0.6420404544230199,0.06862497616778529,0.9965911744634189,0.6258273303839321,
 0.9430576723264108,0.24774963225871183,0.5378031700297533,0.8411096209689343,
 0.4787526654301705,0.30164872243188035,0.8687143397518633,0.6493461934266661,
 0.7400533337169772,0.18411761628991408,0.0018190612139518114,
 0.032793201865027544,0.8940509040761881,0.9728272839805371,
 0.25468748553622933,0.8564274313366664,0.6958896591147281,0.4678912317453613,
 0.38265491434742827,0.5818259722587494,0.05225305188733986,
 0.39279978294413886,0.1004002338221488,0.20563888360475213,
 0.06494576207634173,0.5546795104784854,0.080067806903662,0.9804696455996255,
 0.5450287293046902,0.46710096137794066,0.7667942588368247,
 0.29032606237714664,0.16935605609278015,0.8855457308368081,
 0.7005691688712975,0.4577144307861375,0.23258909971210906,0.8391892746964792,
 0.9498874487084327,0.7985540190625089,0.8871299204524388,0.8027474912128922,
 0.7529283097721504,0.5339592211054656,0.9673805411675238,0.5631450204081155,
 0.07181650729126465,0.833621474935824,0.2976273919339437,0.7771643423432455,
 0.4664094322025465,0.11996854898474496,0.5446971185712871,0.6223167499773152,
 0.5400865034107248,0.963675444402945,0.7823594554042659,0.147863196300263,
 0.6809999135568285,0.08824329329951164,0.23301190297767194,
 0.16369789434555226,0.2528855081764919,0.5581220271582011,0.0721131351449601,
 0.3045765885880132,0.8091586515145204,0.9153920885114055,0.31653368148206196,
 0.19914912504962767,0.5406555612974474,0.10690737031151853,
 0.2992580746645511,0.27643113960800136,0.2726767013364263,0.5655293750365434,
 0.5203341809656724,0.8737355367135142,0.08305051961550602,0.5290762576684659,
 0.45920234466785903,0.2588612325939842,0.7062832608584628,0.4349942787669767,
 0.19192911110160393,0.3274754153482705,0.7435675589529305,0.8877021629407628,
 0.7183073893655608,0.6720115851643639,0.558790469417247,0.14309244423980794,
 0.08621996712182545,0.9300871265712555,0.15101220828688933,
 0.09436979961876346,0.4805119736193466,0.9988629818734803,
 0.18437780214458888,0.8168312500530501,0.5079724974160559,0.582490646034588,
 0.8887945692918517,0.17130467864440513,0.7379655629807756,0.3196498154273214,
 0.6222120808638627,0.6732540916227926,0.2918503315927714,0.7921531321857553,
 0.43294777800669737,0.4036583646503945,0.18981193633815208,
 0.8360959344293778,0.539743710361125,0.008701519248634937,0.202156313087816,
 0.13602068821830982,0.9872006661298529,0.746141486584437,0.5268068676825284,
 0.9275795824107584,0.27534363544196205,0.47363768807555306,
 0.23776576356321077,0.36402707477695195,0.037691906170158385,
 0.27378905282827126,0.35376316260425456,0.48009320141270595,
 0.701923248989529,0.68588929240476,0.71234509068344,0.10740998931643997,
 0.34719708039496733,0.7248297828686014,0.9395677905906482,0.4317508201738922,
 0.9181365816910734,0.9771330543224837,0.33175600728479726,0.6879714719875796,
 0.9536073675534733,0.3327663944835395,0.1797508281968283,0.7160769121165075,
 0.34871571972055515,0.28889385288607583,0.23588255696038574,
 0.808469824246221,0.6931207972554454,0.09051674047848335,0.7492768742686847,
 0.913187882644811,0.03126870773388424,0.04139274717352481,
 0.11299841314433545,0.7401912640521713,0.3390551746712196,0.8296371466056243,
 0.39910373955127854,0.11121833275225623,0.9435332288841592,
 0.5088468678127834,0.08733853377415468,0.23893217007972956,
 0.01887817167050454,0.9397543701368956,0.551612559545916,0.10766150286315757,
 0.3016994429757328,0.2722843415982058,0.6055354514453853,0.4681129340953226,
 0.8149334353246136,0.030164549112076444,0.7291007152544854,
 0.5953117369911206,0.4099524897710788,0.21194860378553837,0.5761177758453554,
 0.6762770256888212,0.0177791808754173,0.34989575298261244,0.4441683315790077,
 0.3161279923372725,0.8780653281861418,0.9547806757482766,0.27907802313527186,
 0.947967813099236,0.3115082202623082,0.21201133324800137,0.9606967720846498,
 0.5540887279093368,0.32129534898930223,0.2516724379941536,0.3128380128463627,
 0.369287618304888,0.4975644974449913,0.3893039930781861,0.7500942835332539,
 0.30444265600904097,0.7650925586248865,0.7754506665780804,0.8756132940139765,
 0.3380884244937732,0.6921877173760285,0.9925973880495876,0.598289156413862,
 0.05123479044124657,0.1492424214176873,0.027863019983891513,
 0.21210211803050805,0.8502939118154128,0.525675089514203,0.706352693708383,
 0.18740131866650722,0.28971667386248856,0.7941929669256514,
 0.9284182905019412,0.105111024291354,0.871542628485291,0.5052041015356549,
 0.09110583317564536,0.9147899478446965,0.5571605329817819,0.5219908310576953,
 0.9221896337383719,0.618021734037061,0.12091017301128859,0.6165446695892319,
 0.9501170456852672,0.8653097534268804,0.29420105617109904,0.9119513804635966,
 0.3668549370182824,0.028828879925604012,0.0013418522555194157,
 0.7318690621198394,0.5363127988659574,0.9413418497961183,0.2710921951109816,
 0.5242435364370005,0.2838602355694544,0.6466864327771473,0.28068549844154733,
 0.49843813390980607,0.6241864076960391,0.017640393671703336,
 0.8446397252184913,0.3992617358445427,0.9688540314546238,0.6532542349036893,
 0.976371897946896,0.8419945358889342,0.1609919682655636,0.7057228849950874,
 0.6502940468011433,0.8144798871172583,0.32196279072773937,0.7243401779208569,
 0.7228238312794971,0.31984513366862255,0.2539769039920352,0.5311008655438879,
 0.4940317335972755,0.9505929554422812,0.5511501500251674,0.01740135069163906,
 0.42946512667210124,0.173839380209168,0.9958184540571657,0.6966155854829672,
 0.44530135878960264,0.6651892454592453,0.04037601446890693,
 0.4899513742604571,0.13952571490166965,0.614691581366685,
 0.007730104397192572,0.8687899670458219,0.04082344294476581,
 0.11479939534945027,0.13786333680761953,0.6278127612478597,0.75333480874438,
 0.23211774698224286,0.5405416680686421,0.7865216206531902,
 0.13692305803280302,0.48395062917445153,0.30998134325037285,
 0.5139109797033444,0.5249656397244034,0.6117853973196353,0.4344094936777061,
 0.26282404681459604,0.06417578429042248,0.3723440939058049,
 0.6312025350009979,0.021581427260776165,0.6092748805292006,
 0.5155078735240829,0.8003774086131171,0.09800988896102714,0.589963285558437,
 0.09042276759562284,0.8172362183037207,0.052420719048273834,
 0.23928231676757972,0.7396114071175983,0.8485447384476292,
 0.008360895358832221,0.707009515859425,0.5177706385767712,0.6307672721406989,
 0.7815909218846975,0.12343466744101039,0.3535996255289474,0.7310563098762126,
 0.4003510344112573,0.08264283556434637,0.38793488399156395,0.777093863281674,
 0.6512551947006754,0.46339408949433925,0.3841015140834394,
 0.17937516901256578,0.7364072579615757,0.8978200375877873,
 0.24796450637560286,0.4061156545633875,0.11924897116998234,
 0.41720715770206995,0.77713729570199,0.9708172254920683,0.5684989583712434,
 0.8251593873783807,0.6247419448545685,0.08265607310224254,0.6161709707185404,
 0.5259491418152298,0.025982666832807255,0.7039417165243609,
 0.1804200819127444,0.9224031963107362,0.7346783668031271,0.973441785231496,
 0.7502331098627413,0.5474809635269984,0.6383706915089575,0.8035939431992299,
 0.5294047855568967,0.16770832855048778,0.9401341390255818,0.5436163471925026,
 0.6408905670279296,0.7961373147738,0.42721919948503373,0.6498315062906644,
 0.15257676736649028,0.5359794560149594,0.916719821113751,0.6468321962561074,
 0.16251106126183013,0.20861801670757718,0.7473673579753903,
 0.8850702163554796,0.14390839465171623,0.2547383788604245,
 0.10210898939068036,0.29529748920292753,0.7346180771430848,
 0.8320433338433921,0.6125261783728323,0.4813806301676544,0.629547684475563,
 0.5153134892947776,0.4549162171648895,0.05885930984013721,0.5336715593118339,
 0.8179888460238247,0.7042300014018861,0.002840455494667493,
 0.7166876388686692,0.12779263972422572,0.5610995585106713,0.7897001356676036,
 0.5498295218952841,0.8208383514624954,0.5304531707568706,0.26336609097163377,
 0.29617310618488235,0.9562372777790114,0.9017519269326925,0.3159889680799367,
 0.49490433442628,0.6976336389928487,0.0326289874439919,0.4176948066113393,
 0.8726911384288912,0.25422778376644684,0.07922596638002999,
 0.3188913776473923,0.42632025760799175,0.28681564381747293,
 0.15790416276155783,0.8239051278638533,0.0026541879997067364,
 0.6571425567032451,0.814451027432979,0.2884838201596305,0.250347695241119,
 0.3288219722191197,0.7944268330137766,0.03021645325350708,
 0.29802155024561794,0.6045957675852085,0.29362504909350096,
 0.5955014538735346,0.19517297299956227,0.9386575687514443,
 0.39736576143135927,0.8605070359066225,0.9055571404554706,0.012128389020579,
 0.8722741606340972,0.6905676378370426,0.7443225094515475,0.21447123254192502,
 0.11370082025982542,0.3441012487077405,0.46523297987964796,
 0.6353035734642705,0.9168412624661963,0.5078255485063358,0.3744874374102247,
 0.3724736474845197,0.8808420681499506,0.6960311690799356,0.733295403343996,
 0.7114993381151989,0.46777895076638154,0.26993469294409,0.03165547819956849,
 0.7187269373617285,0.17877031318033143,0.0689004493690154,0.600642962979675,
 0.6179342901608651,0.8240583498037943,0.4139997581835615,0.46740029587717535,
 0.5824830391419548,0.20063332474604723,0.6712308632238582,0.8768943247735842,
 0.14784241298077871,0.851360553873333,0.6809929019347618,0.9771801886983283,
 0.62001940275917,0.8710128716421552,0.7193397802492404,0.25330995448478877,
 0.5049051265484055,0.9076040705441163,0.1943433138694619,0.3595256420383086,
 0.4229908796381333,0.049487949969949874,0.5317312087947411,
 0.6333951771345534,0.5046023394623457,0.29260518113254974,
 0.33962206498575487,0.6613862515640822,0.36097268636465973,
 0.26727487804421957,0.8080068190998595,0.33187062430959635,
 0.1855473107354535,0.3606347026353689,0.38004541109529455,0.7122816549947448,
 0.12794540292324297,0.660520922174119,0.6795476463002801,0.36562374847343126,
 0.8822520490363757,0.10570289320725479,0.7527690972533114,0.6404808075258678,
 0.9388384786297667,0.9525082258363444,0.5513155386308852,0.18766230127482408,
 0.2396350948969549,0.722421306977744,0.27370149967610335,0.599241224641894,
 0.9867181929845656,0.2792219541865608,0.41505459596562155,0.5652753188782611,
 0.6337439055991665,0.1536598836205766,0.8950879875345671,0.5196655817725033,
 0.25924728419087517,0.6380543779918221,0.36129729146342293,
 0.2533197857508708,0.8807263885129807,0.8437838827445032,0.4069504137239547,
 0.17411145547492002,0.4698839403970332,0.07180475622806592,
 0.46946529473809684,0.15324797741877116,0.7515466393400652,
 0.5415569208950883,0.20182764736483283,0.28334277075181746,
 0.4874696107777171,0.685957749185684,0.6878434456345929,0.02185915442195885,
 0.4030257316167326,0.994985068238489,0.5310373088033725,0.23190909790932424,
 0.9682535506390976,0.6674377970982064,0.3852465777976095,0.500789167532055,
 0.016237931910460524,0.29858233167948167,0.8912782866231251,
 0.22991892323290608,0.2894855773544017,0.5317334488755672,0.6986229662469039,
 0.09233573912419557,0.24685627585535352,0.5763208080191684,
 0.18102094020762016,0.14794011594309564,0.3092358532460405,
 0.5373413583873412,0.9503184973663368,0.6859946035090321,0.43636298288201714,
 0.06730453501038347,0.4147516888488435,0.059773915099520014,
 0.4799667694960925,0.40301601591238323,0.5166242489153524,
 0.25003665414587417,0.072773487765596,0.5069276794028125,0.1953435562993786,
 0.8670694642019164,0.48881799567150275,0.25292777250904086,
 0.5921916800322211,0.41341378889838376,0.6067440364760247,0.9195137630442304,
 0.7192791330766356,0.8714407529324981,0.8098715970838135,0.18388766585677985,
 0.7870653327461123,0.24080843927763218,0.31457366921844887,
 0.8675257098081827,0.576466324755043,0.25746956503282437,0.5480780055566397,
 0.6382371739851499,0.049010667095953586,0.21642169958323132,
 0.49310657161748805,0.2028694382039109,0.9489795055451972,
 0.37946012672405727,0.7352353616209067,0.05074302141201448,
 0.08127924561251365,0.43692573454410255,0.9122134842618604,
 0.8342825790808484,0.2833665400902984,0.7389298134195793,0.34500031664124997,
 0.6433889746742135,0.09137295225374109,0.38714000842112783,
 0.6702868707204661,0.3074288012248271,0.4397996974960191,0.6439184997658174,
 0.3676727544472538,0.5745187684526446,0.9694463526233248,0.7573630009651023,
 0.984560076395746,0.8958807459035759,0.686152272070241,0.7117429837953525,
 0.04315629108831254,0.7508202794523743,0.2476078587121372,0.29540801113701,
 0.4635724666897637,0.42447707584993855,0.02223364041755116,
 0.47996640588023365,0.19415475020031736,0.8879149650895543,
 0.3375419950906393,0.9675879946838949,0.6189825617377518,0.4946441038638536,
 0.011438040953945183,0.5231653402413027,0.11868878621985568,
 0.15677622992120654,0.6469591528197004,0.3563232719837912,0.7952064763590244,
 0.872715314313106,0.15555181167110765,0.90350638186115,0.14646676643973655,
 0.9340819977078677,0.6039959627726872,0.5519851288267982,0.9647491262812011,
 0.7312060369582258,0.04799073452168845,0.5538842067283145,
 0.09082869141421701,0.9591466896127292,0.7750920355488832,0.3487228812339642,
 0.4914865403091837,0.7218971451491019,0.8953645396572909,0.39922736046606033,
 0.07873719411687552,0.006138869403646496,0.7694663961563016,
 0.6611299088943163,0.6690585933591293,0.05392878243934196,0.8494428222491973,
 0.03408497318402093,0.9714800537168321,0.3529876610825903,0.6586335734146281,
 0.14829571164704114,0.7925134329637367,0.9880804811421321,0.5533580765580821,
 0.3603129429113915,0.5079368209604913,0.538079281856716,0.985087092556425,
 0.9290334072190812,0.6205964544329647,0.06110814083618055,0.363255153612166,
 0.08659333138664183,0.13023410731041307,0.9353908913570511,
 0.3913433927234359,0.2984345166745672,0.932230814286556,0.055381775579529724,
 0.6037079221658415,0.19377893277768288,0.21858555093545395,
 0.8043759349664394,0.5120744134342249,0.14105577779550071,0.8281328213522519,
 0.232434227888763,0.29100196328352634,0.27683751152999014,0.8022963901773763,
 0.6027459768292004,0.2978322173819675,0.16051348270506982,0.7643050102112022,
 0.512204043782633,0.8188945287023744,0.8305853037264478,0.49272131610149106,
 0.8553757665179019,0.7641852811734033,0.31982998447133537,
 0.47088153678432243,0.006232757092568986,0.37395279476798826,
 0.7569370047052025,0.40513772872713427,0.19247211083112015,
 0.3383663044231773,0.2540212405741391,0.7560205413304623,0.33074277774605687,
 0.22390632760890894,0.6830493527719731,0.16454127638875637,
 0.5719366191638449,0.7791820703920441,0.6453320210862288,0.9958331411602235,
 0.8161065576329518,0.22676722653236547,0.9969866409552516,0.9979954061265802,
 0.7060511514072816,0.7183052304762226,0.20801850144663214,0.7787228025696464,
 0.8040436477205714,0.0417871056079846,0.5953275969154413,0.7682172528474812,
 0.3297381495230842,0.9204696936924631,0.9723063606476907,
 0.011113266142425693,0.544128224126777,0.13610224858530662,0.921559288422794,
 0.5220172489391636,0.8192539816626946,0.5544910674708962,0.4247470079336324,
 0.5529458486785713,0.08836829254731982,0.24736572004827573,
 0.19230128235200494,0.20454856689877443,0.15189833516041018,
 0.7791971847238648,0.08673455298568322,0.06230179388673651,
 0.21388332400204768,0.7944826274802168,0.9809691970649218,
 0.42836730110463817,0.7761291579436093,0.6566063691946459,0.8323147640547156,
 0.5207947357129045,0.3635439117527468,0.6441318064652559,0.33323372220180236,
 0.29577052364068546,0.5104904751764762,0.7374864987503061,
 0.24833107453001713,0.20720425974296675,0.9812124149285247,
 0.9086271736960194,0.037559280349511504,0.447657297551999,0.0772541529676608,
 0.22460175387992976,0.0523547206596201,0.39074668042155836,
 0.9788515652191802,0.6565824931523669,0.7193189879266602,0.38252368809549675,
 0.2802179346821354,0.6560166058943978,0.7748006616298934,0.7049649796308607,
 0.287767358186107,0.22004114117207396,0.6895531759332817,0.5782183911999395,
 0.3584364977850658,0.09299307243185428,0.5843919040207534,0.375169366899175,
 0.08805646675513124,0.34019577365013454,0.44667918943564155,
 0.7325962803110596,0.4197999371323421,0.036837388819445316,
 0.30224706045700644,0.13236198512408726,0.8877588873883056,
 0.3882488264869275,0.7375940977765275,0.6709997518880366,0.2832038044962817,
 0.4755692617448224,0.04177155488407325,0.663152227380524,0.3489983642225788,
 0.847175853397433,0.7662589288323968,0.5050712770104077,0.26999583424225615,
 0.801719137736381,0.6534208682787027,0.47554133168928314,0.7933570746699586,
 0.6867093735179431,0.7830892470298179,0.33615512997465424,0.6310872169272115,
 0.7767045375712047,0.5341814266959536,0.9997627202432691,0.8568756889035802,
 0.2365995901003939,0.6345349316512517,0.7572994260169457,0.49133295384038733,
 0.05243342704878684,0.9455642903685264,0.33942112685630665,
 0.6169405735671397,0.28976053419084236,0.5293555471119903,0.9556167169634012,
 0.6348173515718857,0.006793836576165169,0.7969247487884947,
 0.16453076214496054,0.2227118466367476,0.22723308997927638,
 0.5929634086522351,0.011823857583941999,0.9069462990379895,
 0.14741690217557957,0.1489161832215048,0.14102378329992882,
 0.13971280335248737,0.543945369415542,0.26170747029155084,0.8797772681649116,
 0.36764595148669066,0.14040099126671968,0.23986974863328936,
 0.5554176567236719,0.12175478944826967,0.19159020944592975,
 0.6301140461814588,0.5371775262735492,0.3971097845168021,0.7732809842029211,
 0.7114678027582886,0.7214227594365014,0.7284796683032003,0.429907000103408,
 0.3849046431741484,0.004189884244835684,0.38499650350761516,
 0.34230582612645055,0.8067886563968678,0.520173783030309,0.6633111085508467,
 0.7421636260236678,0.9249957146423871,0.8741877079153496,0.336818147902921,
 0.5292541068540937,0.6099204918180011,0.8298486962538114,0.09458609119214656,
 0.3486277121409199,0.8238870119970372,0.9873515253403534,0.09747663285190045,
 0.15015555269411573,0.7891286008914028,0.3874071330183383,0.8609366432768223,
 0.12730537277416376,0.6226856661033261,0.6684704560643513,
 0.13865173597148883,0.724054756776814,0.5559009949977622,0.5696688185148618,
 0.1683569256558164,0.44684850205576643,0.13901330103454546,
 0.39152564346875696,0.12008364115176218,0.8245371073322368,
 0.8449890011827658,0.3991772535607786,0.9574848451546287,0.9228254502194093,
 0.5112591837810325,0.21705113940155984,0.9412355998415937,0.6448166481879177,
 0.6016286775623028,0.10835882055133061,0.8708133062623777,0.5353323672918228,
 0.7093892763240885,0.13180796278135132,0.9375832363106883,0.5663863125247307,
 0.7874030457060242,0.37575721255637484,0.6752907679654008,0.4382087381974894,
 0.5573428945593835,0.8621439845897512,0.48132156150721295,
 0.13354561570137635,0.3687858283809564,0.736064776132308,0.5446332374153486,
 0.31181454604594583,0.6410675984448639,0.876615971343964,0.7299074810604105,
 0.9150673504582161,0.5663628471346358,0.7288626864414357,0.4389502535573162,
 0.5077919754836504,0.9877572004569738,0.5486068827996546,0.36379664113208154,
 0.5200304983116494,0.21530012113813302,0.594403839309531,0.1697434397930072,
 0.23021462712873564,0.9755125879319693,0.10521949607063574,
 0.35405666333712915,0.9601252132020002,0.6702854587573471,0.6078284292761127,
 0.033937816289847156,0.59176438638139,0.7885077087041121,0.2709998243239625,
 0.665595405531785,0.9250859447226176,0.8339391367389916,0.058705673160420746,
 0.6490516899210361,0.7227730279992755,0.8893330412132414,0.9612212239476514,
 0.18040880399656778,0.11254388164680351,0.6454357848284921,
 0.5855830695444578,0.9397269343221621,0.8053834946079128,0.37681335008734596,
 0.30908360271572977,0.18893015489266962,0.7604150611190253,
 0.03926108594778388,0.06444499613547738,0.7970787989771696,
 0.3534083889119265,0.8742710450773861,0.408963597823348,0.6376264942022879,
 0.5359592532358554,0.5175142317187698,0.24244238685006603,0.8701838266400862,
 0.19253922240499355,0.09429771279171872,0.0591064744564157,
 0.8132947501448634,0.0912307290902653,0.5192540940491671,0.7515058992019223,
 0.8412757861839685,0.046255436959262264,0.34501007948951945,
 0.636328985784555,0.797552034194146,0.3258786776107757,0.33604743771706225,
 0.9636205376271126,0.9967832816042663,0.8109048829206891,0.9773182285663368,
 0.4686385965567229,0.24175454237170957,0.02847587755812142,
 0.17027011663064517,0.8104720868104144,0.5535713935622331,
 0.33248276063210047,0.521341138721772,0.9836679987721845,0.47548529536328843,
 0.9572192628216571,0.5515648444267495,0.16267471393241728,
 0.18554768838517766,0.24502506030644022,0.8449547964189174,
 0.7438175004026035,0.9001351812190055,0.9987873427568945,0.8925086739336532,
 0.7063053514820715,0.8751946220425914,0.9257345090049307,0.03660845552918213,
 0.5766599973221201,0.8972210906963957,0.37173325428375037,0.748910790326696,
 0.7944546371032826,0.15145746403171945,0.9477882163027879,0.7251179134344738,
 0.6228690309919525,0.6045330044885759,0.8724550477086701,0.920549106737426,
 0.28851868980321693,0.7686520451065978,0.9785088974141024,0.5893155951116328,
 0.8855675274113973,0.016866261230423785,0.26496559833783717,
 0.42789124660151534,0.9297907854233185,0.22542842706529942,
 0.7162448959882014,0.9723066316242052,0.7093451188512268,0.6448013668004183,
 0.8863591205715539,0.9235794323053372,0.16546410118858257,0.6723749048580712,
 0.20814919415747635,0.20492453671273803,0.5699434872700125,0.941710519969363,
 0.6819912164793914,0.648433888123249,0.22022581508238925,0.13454135537538237,
 0.9861112232065847,0.7216827850645969,0.14590596360934793,
 0.23529505563163067,0.28391414763985057,0.14206104544952058,
 0.09780117962455293,0.21779577512066262,0.498047524473895,0.8390301747634519,
 0.3541040519411438,0.7978981560353031,0.7599153046256701,0.636225982939616,
 0.5137263380825029,0.8774752745661871,0.7946913100379571,0.3110615254950375,
 0.00717481666837394,0.44255439853493483,0.6519072610099841,
 0.45881025612838067,0.3686592678936751,0.41306708609510145,
 0.11342674326470614,0.7936855362288827,0.9648711354207729,0.1877511736983366,
 0.8071395548539504,0.3302320629220834,0.16602660296891802,
 0.45750159350681363,0.6717823846001064,0.348922467769223,0.518230329125966,
 0.3506319581266025,0.4354824247232649,0.7037243547064594,0.24320230157435718,
 0.012206903477800513,0.33319345577377957,0.5343028887001581,
 0.7017991096865848,0.46816557056285246,0.5841504842688012,0.8647313294993193,
 0.16384178032039165,0.7709254167458119,0.7468534739038544,0.7415922004380476,
 0.1689270804266756,0.35272009216225664,0.35399007690238804,
 0.5348728167514233,0.06606857419150014,0.17385244532819188,
 0.6614331012294549,0.4273241848531477,0.5047716681879179,0.8949362185500512,
 0.8055239424624546,0.5349349022182078,0.7335406764310973,0.48590859975474743,
 0.32153850505745396,0.32885403329467444,0.99299279392446,0.7101918948467578,
 0.30432043509345086,0.6739435552258025,0.042920228658963944,
 0.11066387511272513,0.13857703634953433,0.4460554924367778,
 0.3532888390881367,0.005121659595531569,0.49395345908487276,
 0.8957336170678137,0.9478346267716358,0.7469642687334767,0.2478126841734246,
 0.7147849764677984,0.08088554752976895,0.5698189289493657,
 0.37064854189738916,0.34145663129611004,0.8486436967100941,
 0.8286414649005831,0.12358753503649278,0.21807344328399414,
 0.8382413759261955,0.15555158423278526,0.7603266207427626,
 0.13913404210648128,0.7047367852966113,0.11715454673368819,
 0.041038390745029574,0.9459612931661654,0.6633022054810258,
 0.3510842784137115,0.04480439207087139,0.2744511267111611,
 0.05770722996289546,0.0409746681645774,0.3837129862752918,0.451616857527408,
 0.7356243563094579,0.799689161437918,0.21196234634931876,0.4077411461426763,
 0.8991270624226054,0.2868196044204191,0.11469583486654955,0.7173060672503223,
 0.5445485577780009,0.08371277279133871,0.13293048866346435,
 0.5160376683047678,0.08150469833788043,0.045872630763365785,
 0.47741127195963684,0.26766753813456645,0.9812744326459071,
 0.9876805232867883,0.15901963140443076,0.43383532079446885,
 0.25057110875000865,0.12099103468274852,0.949547967307558,
 0.26575054433140866,0.01719882140094442,0.2766269994992454,0.493045392402616,
 0.4631059608561797,0.9611641358880807,0.2797738517545918,0.576858623097814,
 0.6864411030365956,0.764634344750585,0.07491038744923328,0.20478794501083475,
 0.7059338738152955,0.24394365579303012,0.2487246920342202,0.7664702695897211,
 0.31039985946421333,0.4636035918595256,0.19629799583670238,
 0.9010979926601802,0.3851720079710519,0.5127748434226685,0.7149548843191013,
 0.2705522939124766,0.6141690874197879,0.5430271755965319,0.2529714923567904,
 0.23110480266435796,0.25912373159010904,0.34926308375657344,
 0.39647072635044756,0.46038146482831077,0.6949614958365062,0.584144744476139,
 0.725600728452304,0.7093636026000874,0.2720227474588822,0.12318733238752633,
 0.9288262618128074,0.4420934735354274,0.38377677303834645,
 0.0037473965576488855,0.14698939510820874,0.17977258586797618,
 0.5801993129540045,0.4437306815450913,0.4262458752921938,0.49057188064426016,
 0.3040723830237885,0.8624874208236213,0.7846692890096477,0.5459691509522329,
 0.7075355711293654,0.8791261090565728,0.5186401523037112,0.4907114650300498,
 0.21074151356850457,0.7531477551068391,0.16940130266974607,
 0.5105835133279781,0.3586244214105079,0.5693036374026269,0.2984843990845022,
 0.35251898605726206,0.010558286564920305,0.8225792722449805,
 0.7919168882722941,0.547393482444224,0.2566325603061963,0.22090153624776754,
 0.21070161894117345,0.3567316080997184,0.44212894077267084,0.447542503056894,
 0.2788152152597324,0.3581560316223823,0.2964228506157828,0.853743552466346,
 0.15124362456110896,0.4651028203977561,0.5203328353515477,0.7020280277505502,
 0.29555369996720926,0.7946663825153908,0.2639815995342387,0.6740958595974874,
 0.050765048973529514,0.7658978781547139,0.25775605027138715,
 0.3464817707504376,0.4934390612111298,0.4267436878581016,0.4159595904739012,
 0.41409646373365094,0.8754672071564386,0.4259788918070351,
 0.11543188931402182,0.6854123350983927,0.8280467990471357,
 0.10242427202996995,0.44904309350581295,0.4929260209912537,
 0.7634210034110855,0.4068201821815034,0.9934913000085059,
 0.053002363999854696,0.22140294120978044,0.09959372809513967,
 0.6671420992066645,0.7220210108907914,0.181330049424277,0.9064271637417414,
 0.5811590185122661,0.44435719720372424,0.969066708661514,0.19273037035813567,
 0.952767124452152,0.13607377386788144,0.4044802154175384,0.8469226673136405,
 0.7520640712694721,0.4238672310248284,0.32481092305398285,
 0.05914029851673408,0.9152699914804486,0.7813902599781515,0.6403455923749023,
 0.7603167432634366,0.8141010773937292,0.8790227088014244,0.2896562192541623,
 0.05141741171486791,0.29454727920947,0.9504722806303688,0.5177873892348901,
 0.5459753506110312,0.542787779336548,0.12811793495228863,0.5542082402244475,
 0.3293095181065696,0.6837579655023731,0.1554940139675497,0.7033420670845594,
 0.7525313306562543,0.4467276773534593,0.5443117817862015,0.19121966229074294,
 0.6502873361720629,0.16405607467502425,0.6291041809821115,0.5564887679082823,
 0.3431503102294938,0.5246884909131765,0.006918119685820123,
 0.13072557993978562,0.45694486909374776,0.009701929370249718,
 0.041155714249566766,0.8755596939282263,0.6983535955192087,
 0.16660915844120028,0.2710590637698983,0.7391570034049508,0.1699433725617555,
 0.19345453298140736,0.24836067948663176,0.2126329907316299,
 0.5878767373572082,0.3219409158245644,0.8438955025131657,0.7307026522958104,
 0.3894862372359249,0.009429186176978743,0.8714756584477368,
 0.9426868023879308,0.8795223157343341,0.7931704424503786,0.30721542211109987,
 0.3243208300861109,0.14977315067011232,0.24453785898474156,
 0.5547591127295076,0.10857424296883111,0.9772345884898874,0.3662962218669348,
 0.8128273568057056,0.8557236822669714,0.4558972148981354,0.9443075958016411,
 0.3185908055918102,0.3549952297426525,0.07444634500171099,0.5032206548079542,
 0.3430901574933475,0.13004491822424324,0.4221345099411399,0.7144623478231793,
 0.5094884413822954,0.6710104464895454,0.49030748859521134,0.6023464382233145,
 0.6757259569720272,0.6594342869028338,0.702843938108168,0.11367578359226616,
 0.8473388229300122,0.014841967389141897,0.26881697811438465,
 0.2553545344971415,0.23941812245155902,0.026520962481816346,
 0.5348733863839027,0.2227724768742192,0.882703714416397,0.980339365411127,
 0.07393127570973101,0.029828625065746017,0.11804124873597754,
 0.7415144063123528,0.9499805728165757,0.36651676182960413,0.5670600806780753,
 0.9463098891713848,0.25578068796832165,0.17067598307024656,
 0.5647720678028286,0.7305589875283706,0.3527227435184812,0.22469069415271448,
 0.49338459002947643,0.29983125742901806,0.046288045564412905,
 0.46543969783712535,0.40938166420972877,0.1434060366908817,
 0.4650707263907914,0.9462475276305842,0.5038146451719984,0.8293935132831342,
 0.7596514886392524,0.08362514319310332,0.043998675470572124,
 0.9118610668415966,0.7178784402945241,0.29730318958460233,0.7013269626580585,
 0.9700882118107018,0.994467676176569,0.6672416499881001,0.455830880405674,
 0.5643265697862426,0.3502398597535308,0.06340289577563496,0.6504604449657767,
 0.026913458117619138,0.04326049985184377,0.2710828984045096,
 0.33814756463537865,0.6438738751680532,0.02789776310701919,
 0.5167934169647564,0.3328115491990078,0.13880295226105432,0.3221299027314215,
 0.27525345536638157,0.41232845727027734,0.19141855193909185,
 0.06415980548943245,0.2293639226011207,0.3303067699214086,
 0.08080087461377516,0.860676255218325,0.5412398981321552,0.3584983994984111,
 0.9453571104773475,0.014478223585966843,0.2860623330361953,
 0.8969967999666456,0.7600094038595705,0.20775119738901426,
 0.39288375227040084,0.26166226335227016,0.4838531739602583,
 0.7233256645451082,0.29134919019505834,0.7730064979537585,
 0.04228666289848282,0.3357094170236391,0.29319149699944036,
 0.07785602327229124,0.579704027074879,0.04567060977159054,
 0.16351855709340457,0.5969900045407187,0.3656348654640754,0.8875785367322812,
 0.9769507928549999,0.9240700953406196,0.8681039473558735,0.308947658453825,
 0.2753623392480613,0.08646101046565102,0.152553989870718,0.5146244381070545,
 0.003878698144386128,0.5348151193264441,0.19763754644113918,
 0.8250405267658868,0.4129060665590899,0.9396606458192773,0.5537244625295861,
 8.574598441377645e-4,0.4940574035013344,0.8890707842267054,
 0.6099771128512442,0.08398904957722253,0.8901745542409367,0.7873700979305565,
 0.6703987464391977,0.08039233866344864,0.35137309719551413,
 0.3469513573988896,0.15600531276425011,0.023927804776533,0.9648505159407159,
 0.8694874643241318,0.49666271289218067,0.5500835373006324,0.8622244882015834,
 0.5884693013620075,0.34897038226971644,0.24773917308144922,
 0.7459441764142641,0.8963132267236863,0.6205302020172079,0.9542911979648746,
 0.19858825495313526,0.6477115346994026,0.27243261648205686,
 0.13603551752266174,0.28265940500138365,0.637715142092034,
 0.052929335482694406,0.9568935037702493,0.20491573464831803,
 0.6821547309011176,0.2198318450586939,0.8121474790134215,0.19258715683589422,
 0.05231395868516342,0.588439646079432,0.8500251438894068,0.3009192771725715,
 0.42550859098648497,0.4696628824386154,0.010607359331175625,
 0.6023239984812194,0.27454226741721155,0.4550951777690876,
 0.05288500972083865,0.31775397129452565,0.23430942269987465,
 0.4717071306926274,0.9713734105467193,0.758398786825986,0.7001681447556717,
 0.5490707105771944,0.45675596192160006,0.9295149906038641,
 0.14217490991449555,0.9508522145378933,0.46067455829698434,
 0.3793511520703896,0.8299013991433365,0.9251649208926312,0.04138076975073879,
 0.8185657164577914,0.766043822619739,0.5182553211814895,0.32162325131703695,
 0.41034611602676674,0.6492683720689413,0.4366087297607353,0.8529047262877267,
 0.8570906518409195,0.6246070442917269,0.5025775327486135,0.6096383941779359,
 0.8852929521451309,0.36305333786473903,0.8123441041311287,0.8391376724406431,
 0.5392277522191831,0.21681748123313016,0.1653335347196263,0.9875946912953931,
 0.3824117700829368,0.8798932209219539,0.6335473373073186,0.09185779996337995,
 0.07037794092373517,0.2319760363323069,0.1877503714228369,
 0.16796233243693093,0.7811604253158186,0.5519537919080235,0.4442891714493221,
 0.8684444274618675,0.14654131679358529,0.931846797439094,0.3675988240379656,
 0.8427421968134693,0.665955715305883,0.5927186730352889,0.2716963842461112,
 0.3686581197344342,0.16810114158515677,0.8291239128932366,0.9718391931542295,
 0.819622088155778,0.7919769378550923,0.8699927997838046,0.619462565436265,
 0.03769344193926394,0.03255238666046367,0.03753397451397633,
 0.45259163391549306,0.8047890840705918,0.4236969560593824,0.5285602935107492,
 0.50515816500466,0.5978561337209911,0.8677422588162375,0.4248411533901322,
 0.18225529626104575,0.6939644619336155,0.36608290347168926,
 0.0617975348319203,0.8511089196168776,0.6856756350908001,
 0.050292346285100153,0.36113524259951657,0.8472751116127064,
 0.1857215461329369,0.1748776762468962,0.7067050752117362,0.7203131792192189,
 0.839106792783796,0.38780839837577785,0.5505684405829236,0.3595946449435039,
 0.9352888563448695,0.6197097178910012,0.02122896366700655,
 0.11154635687295333,0.7143806905014958,0.29961901750772746,
 0.5554772000699469,0.13174303307256996,0.5888922593723618,0.8669896483556023,
 0.2632913881030222,0.7407253570958641,0.4955329614883427,0.931787218103215,
 0.8851892563209637,0.7001110340828067,0.2788331561903633,0.2804665794511584,
 0.43343120242927935,0.41655346397589477,0.647904433725502,
 0.32668501011678497,0.7992803371310652,0.7935458864610392,0.5379202698563066,
 0.9972161457414541,0.04091999375814248,0.49394860650833494,
 0.006920869195938595,0.03278644977468814,0.5554637365035262,
 0.5053143370609793,0.17989716528808497,0.9260931344809534,
 0.11745795065317322,0.07456618888951161,0.9736214089676873,
 0.9796473773632911,0.2073203021790867,0.9619061483236129,0.4943388448819832,
 0.9087169885206212,0.0798606896181484,0.5137864263956045,0.5627529614681683,
 0.9680303710388405,0.7930787197624467,0.979498815694326,0.09671058741085237,
 0.2024901718839609,0.8504477361677313,0.07805761254837462,0.6643115474855413,
 0.14580380431841622,0.9872488643894318,0.6307196342990569,
 0.39790226003149065,0.5482061803359854,0.30977262014546636,
 0.4797507116897385,0.0072249457265201045,0.9209214102331643,
 0.0053702005975569955,0.5387431563590235,0.19431363707274885,
 0.019536782709258516,0.7453253576541478,0.20750502034082063,
 0.36171429811852496,0.9157368397321998,0.3285745687249957,
 0.26385128610088737,0.06962231243471595,0.692236061749072,0.8126302911271844,
 0.21261306565087879,0.5351803169031493,0.8416999387203828,
 0.04670730417693769,0.2566647476047108,0.04118172045749324,
 0.42930323048353725,0.17930157541144964,0.5095958618639,0.7649389941053651,
 0.878930485936512,0.3772738409189864,0.8592453810561018,0.02829863149121281,
 0.9881579314687733,0.9529772297748873,0.0321366527181044,0.47483490638141546,
 0.23350960977698998,0.010504672968426165,0.4926254059409836,
 0.6613312628846859,0.4259689943010676,0.36081598114938884,
 0.15163535252624927,0.8679177752258339,0.9519075271359505,
 0.15801897795079767,0.2507933097043783,0.9637439301224949,0.6366232694968514,
 0.5877085878114564,0.48679039205366714,0.09029317338534171,
 0.7911544393156945,0.4610722579216102,0.685895143533345,0.681326079347877,
 0.07312266591718597,0.8395010851022179,0.5089147353322357,0.8573390116883561,
 0.7453714983332859,0.020994006072067806,0.388521097385781,0.7196636442600691,
 0.20874414045273926,0.6177012301140545,0.0465680186233669,
 0.36610009331466387,0.9289900423084146,0.7127211865971983,0.6688498556805074,
 0.541919400483772,0.4798738125922384,0.11402066863679017,0.35773265706954804,
 0.4683401104632746,0.27934390288948685,0.9467737964411891,
 0.009621285351366082,0.909858596905438,0.42479181657637866,
 0.5809542858392733,0.17964883612337568,0.050565927482305595,
 0.2049699722234588,0.8126291076457813,0.5051301100566857,0.8960126959885261,
 0.20736352713640593,0.8301674571272567,0.41236469759014316,
 0.2174232456633588,0.6735501499098592,0.5837919135870255,0.9377496481361535,
 0.19904562901052802,0.7967414527903369,0.10286580200584872,
 0.27293183634850593,0.5152037728480238,0.2510575948137621,0.1655537164968135,
 0.2535932549172466,0.6492901751748943,0.053948015057073584,
 0.010778796078514619,0.10781698140299179,0.3087950810419985,
 0.4695382807754225,0.6602885341132582,0.12495149108533865,0.7542899642305426,
 0.6221219994698606,0.5386823668255302,0.9089406133150698,0.3284341830036147,
 0.7884261469987417,0.7182645642427085,0.2966861610997579,0.45291374760228986,
 0.8275987219534578,0.7912615367781544,0.03171460533341963,
 0.00464491249119825,0.6312482934142486,0.020307239318586134,
 0.665330315741636,0.35936480521451086,0.8708393041717825,0.7035215469262313,
 0.9174480807478018,0.820004419301448,0.9462835717131561,0.7626768293761811,
 0.6810009264701616,0.1454014450639416,0.21157459014550994,
 0.47175728885643353,0.8073144074134467,0.6453811105667835,0.5878636202255242,
 0.6342268912144406,0.6646601916997691,0.31502905959020033,0.3697756714906546,
 0.6202219064860568,0.5914442693419143,0.6152119139456752,0.26704606029470157,
 0.9102705630154853,0.15737481331095104,0.5563498429359396,0.1257074832400813,
 0.8443129460365753,0.6120112477639281,0.5359790479809191,0.9759419359755295,
 0.857378451078488,0.38517904080532706,0.2555524205335975,0.8977036537927642,
 0.36776895411637534,0.6176825947791035,0.43049620635910735,
 0.8100881337521343,0.7035992121516137,0.2116113659743296,0.365868019082517,
 0.05103975646584202,0.014502494930288323,0.9004562689407686,
 0.18058994650124738,0.8445800243768995,0.8988898350046971,
 0.31298623966623385,0.7332908932078076,0.18945493369128963,0.534025424545259,
 0.9278340605002033,0.9584221198650418,0.33379227313744286,0.925228633093634,
 0.055840607715804125,0.13251715257967056,0.2890625276966833,
 0.5745783522840796,0.12309368521031161,0.8666351220529476,
 0.27676777935782704,0.6194705973049196,0.3192843909576415,
 0.04864500531228644,0.8188192069561901,0.6047286369783444,
 0.10904434087296444,0.5980360824938027,0.21645983473699026,
 0.9719596431093507,0.5802670934704391,0.17023233297342355,
 0.46283592593657275,0.9008917926987314,0.02135605490805703,
 2.656419658977338e-4,0.03146802547225136,0.8790178611477297,
 0.0652129294515007,0.10547401513345678,0.9375620087316845,
 0.24101412915883547,0.9652744979467307,0.026140442491757065,
 0.7380927198044507,0.054167753181569744,0.734991048051092,
 0.31450593644190494,0.4350166986508043,0.039824587957246416,
 0.5981830778551179,0.9879054933834576,0.3078199598344805,0.15623270213828333,
 0.6495880196577128,0.7177556898015356,0.03395637315258082,
 0.48073201798109155,0.03602250945976715,0.6932952855573006,
 0.2383863084842972,0.5178680955508145,0.35821290252393156,0.3392127537122861,
 0.7209216725424744,0.010234877981936785,0.8576251361676379,0.99736695266473,
 0.8019145806704044,0.06000174220339216,0.7256389085817684,0.7973197750627837,
 0.3354692713403542,0.7906416847341649,0.6241216098171287,0.6960637750750481,
 0.5665594766584633,0.4235897303343932,0.5510289021507565,0.9780910450723032,
 0.5945148732207053,0.5470844107577731,0.5957020243559887,0.04533262250161019,
 0.3613304912912465,0.5187743810562973,0.8001359575252953,0.2181884750488321,
 0.6593256443819978,0.6941822477618025,0.7118376407727872,0.23334513902300769,
 0.8930769986209701,0.7529815701318436,0.6438453004225752,0.8890452953277777,
 0.32278690318333947,0.2368771878489212,0.146832352392111,0.9409686591084337,
 0.4667843424003367,0.15087498963744328,0.7783293509882453,0.9031646013432274,
 0.6786314163409313,0.03602152759635491,0.2245027760206032,
 0.31165128315668855,0.9253192200667542,0.345105188208628,0.35307931882711163,
 0.11486129034462378,0.8481008834330632,0.28796253043021613,
 0.7970261755906513,0.004038874831211903,0.5889411344729085,0.375238190002945,
 0.16206061793162851,0.47077264560906007,0.5095001582159235,
 0.4536583798895091,0.889196548308349,0.0855440676676208,0.4772908035699558,
 0.21876456014313894,0.14897203883938215,0.16039664126758968,
 0.047801759180643,0.144664083824264,0.1411112627438973,0.42024623191737864,
 0.004897638293637097,0.281164154199268,0.1013613353515469,0.9289095030911452,
 0.7355723645760361,0.7415464487487177,0.04509393210880108,
 0.21480209205930456,0.6255580054703044,0.5748200152579599,
 0.09895932070963487,0.13680032188355917,0.925385509212125,0.8036182226794432,
 0.61509241995384,0.6529117388607985,0.5751199120709642,0.9931355123192379,
 0.5074134040534863,0.052921199636043914,0.47645489554782716,
 0.6856658719657563,0.40005663907582656,0.6087104912750904,0.7474320039450681,
 0.3335345578161838,0.6517001841081066,0.2515536646386435,0.5748053676864603,
 0.6162409513002264,0.8960223121472888,0.33690046978385013,
 0.38274556683036753,0.7409809211061431,0.7473120664795655,0.2751605823254064,
 0.9135462186217854,0.0768821647855239,0.8832104528778645,0.1797263797259666,
 0.7854280624199071,0.5738150939700738,0.9059274651072536,0.7540996064962131,
 0.7015887031156738,0.33426513150417514,0.16547580345099622,
 0.1045004980765738,0.027915307237789122,0.050920180431116346,
 0.5224533330715468,0.45922781691149817,0.20687843385113014,
 0.9669163672963345,0.2551463012799893,0.2151928079376747,0.6137910204150335,
 0.7275263402022762,0.1526878447286827,0.27529680741619345,
 0.28748683221401783,0.20102180374597423,0.4641683476111208,
 0.9111028113623125,0.5215350547513345,0.03447701969517036,0.6796864172694442,
 0.7930079625616566,0.3890501387434987,0.729722692223253,0.579086179530999,
 0.037518615638516994,0.9723711150380817,0.2668800644918147,
 0.41754621522454816,0.08626791771312348,0.06988327116750748,
 0.9718717512341436,0.5716278404866955,0.36747525393054936,0.830630866713546,
 0.7559547938085361,0.6484672238362464,0.7711461981768788,0.7427620743087194,
 0.3609826399881998,0.7803371241140356,0.37084766360745935,0.9463024569476612,
 0.4186207301262145,0.24040404244587532,0.9582364561224846,
 0.28488682172113433,0.8980504100777694,0.9043148363164457,0.9385564628285893,
 0.9170729269067777,0.5376363582929004,0.5836031507742893,0.8058517839209073,
 0.4705239239289811,0.8831079935584017,0.6859048315788894,0.8884198484735873,
 0.5168587769760978,0.33066133354095784,0.7833604020077783,
 0.15603991465837463,0.6368835863594895,0.1523345258064145,0.5935003116631533,
 0.40061560309343325,0.8775384475431843,0.4563441031460723,
 0.12039686015960938,0.26624271662785004,0.35284251126200084,
 0.5945952920816292,0.6496933259807847,0.016816467247548594,
 0.9258700336457822,0.7621501516799011,0.2137933995385174,0.32329880240724074,
 0.47469886831812313,0.914810672882267,0.0653468930265032,0.8668167360474643,
 0.9314571952848253,0.6013088132357163,0.19991487485013215,0.7781029370540578,
 0.14756276720150818,0.26989593847051885,0.9735770769549971,
 0.020129006463162846,0.024032518858356,0.5690673503508507,0.2700314444174091,
 0.6528147135272858,0.2514891254017254,0.1068560775609918,0.9823880761519783,
 0.39042141235387073,0.7705405658715184,0.8439174575066579,0.577872653459853,
 0.7849107648728576,0.11965271042935688,0.014458054548143306,
 0.9936957644022366,0.5391513012691235,0.15910040276405946,0.7811542533250643,
 0.5051459393761291,0.5863931846019114,0.08678688382482425,0.2397155553040281,
 0.5210946757866174,0.6789242485564646,0.15401127254438207,0.282731444444129,
 0.38891682689816975,0.003872866130025021,0.10257614967114059,
 0.22837186221286554,0.7085323216592478,0.7585776524890739,
 0.04698200904342653,0.04610278094683595,0.852844467968864,
 0.09616111319129317,0.4337486511006535,0.7512443710088305,0.4569817094445944,
 0.42361029850787346,0.8261651331844814,0.15414836255483633,
 0.06851656434425091,0.31946653134186564,0.7759058822004671,
 0.2804562974000133,0.039638603599263855,0.28894088669702034,
 0.5898534410676028,0.40332533362684175,0.81352899707727,0.26909408259619905,
 0.7740133806284305,0.8850384206382489,0.24267413864617093,0.6598509714639007,
 0.3434283551348849,0.4074511056095942,0.004181978331461833,
 0.39551100881363266,0.08036257146864123,0.46130802363975576,
 0.5756666219348692,0.45512392094774623,0.40849619801834214,
 0.9167750180841547,0.6305903766504503,0.1942183010222056,0.752104623788237,
 0.8052830073524376,0.8034898892486404,0.5503135985774054,0.45144145230382937,
 0.8769292146148384,0.5011977563825172,0.7275297450404833,0.2617628172621045,
 0.8732353576675872,0.9622507786739263,0.2292679260254391,0.6063188201808047,
 0.5000692451544292,0.4279820875387804,0.012444765642499744,
 0.3317536827947172,0.4673746389807535,0.314097209275052,0.3366378162788034,
 0.6723536176120692,0.6908758118117057,0.3627075919870105,0.04592868449408927,
 0.19214315095607448,0.2774669924878188,0.5950189382938836,0.1455729842573058,
 0.6405515568189978,0.9839611197293279,0.4363168837838366,
 0.030097189236995225,0.7188236991656736,0.4186201683678161,
 0.45883392899907005,0.15775152273635296,0.12649936881139556,
 0.9087507730759059,0.8966205828938805,0.20990065768518984,0.4586879149143631,
 0.27444816442400166,0.6596584799284378,0.45490021129218194,
 0.8555234346771523,0.9230612543347012,0.437141105122128,0.017639897993236953,
 0.1107184404210988,0.5489143460823975,0.5057006727842662,0.4455395237653723,
 0.753145792015365,0.37678808048838663,0.7184398657879015,0.2296844105273591,
 0.8752851258892286,0.8752479193405565,0.8177135757744938,0.1149484998092668,
 0.537585031006516,0.5018097455057774,0.6770154933777395,0.7515846595116692,
 0.24126727386457147,0.4659389024734999,0.3389592806558104,
 0.010911008655095977,0.3221368861127758,0.8823357251196235,
 0.5153515285074702,0.6420375696209221,0.8103448386433929,0.23523366426849224,
 0.5593844295358412,0.7074037167960884,0.06323852523449869,0.6790247444953519,
 0.12534482705362415,0.7605244824056824,0.8715228125513383,0.6195087330268124,
 0.839446642860027,0.588936974382987,0.26816054592056393,0.591693317368062,
 0.15950024409203378,0.11108788103805678,0.7657886425591814,0.871380917985661,
 0.9808085378589697,0.8790496645286008,0.48553180481266656,
 0.46131875247539056,0.41673380598823184,0.4646433120435882,
 0.4085385956934008,0.7382288417009383,0.027163366261554245,
 0.6470494272411399,0.7707488279461738,0.22890251796073124,0.545493847125123,
 0.16875031552741282,0.6970499019559787,0.431962524841367,0.2487910881702552,
 0.40491951755376177,0.645554904662464,0.0031404440067017703,
 0.055007525650202616,0.928754711573603,0.14304717465509387,
 0.16358092002135316,0.8478283777174365,0.19965302354664194,
 0.14731101688146508,0.5455315115408431,0.4036759117330252,
 0.19072950768252694,0.5200528365935532,0.9950198182909207,
 0.40251469268138373,0.28328105386300473,0.43854372075722825,
 0.21103073038841724,0.8922637626672605,0.44803952737183605,
 0.8987312220003597,0.06222404676580473,0.2688798668097311,0.8142151490598856,
 0.11290261477289243,0.6742886287806668,0.9194081684803793,0.3788410500129016,
 0.7117192464575143,0.7532381858379456,0.6127808209576625,0.45139047028451695,
 0.5945498961929584,0.6256185263535053,0.1110794017060398,0.3544926025978985,
 0.7268544033333179,0.08327064075533741,0.9189399330924642,
 0.27385596491796305,0.7601504773387802,0.4735638654917943,0.7871149938017716,
 0.5751747184135373,0.6566812419345065,0.8353467754505672,0.4008762441155297,
 0.8233981105874302,0.3504214185277317,0.14991496915599256,
 0.17873065044007808,0.18352353240741914,0.9639836999082964,
 0.31283594846171403,0.5306309829207266,0.016516634388373475,
 0.7495513283147965,0.46871465235368226,0.43087583016613573,
 0.5514299566236506,0.6149908363893041,0.8418190136282138,0.9081080490714399,
 0.644207397471427,0.9247893637528124,0.5997946886273597,0.578121598940807,
 0.12620353937499662,0.5246379636077364,0.2952305622216643,0.87335265058727,
 0.06588254071769939,0.9602995886483865,0.2303016632395517,0.4496137736210961,
 0.38312868370048736,0.23790862462567075,0.1181369499428162,
 0.15470898737130012,0.3726992835208869,0.024325258428858798,
 0.4078521704132839,0.48649691165119147,0.9612981732445395,0.9286594322304397,
 0.1588869382380258,0.6382469548958623,0.34837343160688894,
 0.043204119046839184,0.7318732833588149,0.39191276266815445,
 0.36943468327748064,0.2622343048033736,0.015176496001073736,
 0.020478445915805388,0.6855053069982044,0.618000398719768,
 0.19818318091820275,0.031117440252159123,0.14052784436254262,
 0.9348021389967576,0.9313014301942836,0.473038198535502,0.38554851585109706,
 0.7121460506368036,0.2392698509370944,0.14755504251310547,0.9296391783671413,
 0.2843535784529623,0.5976463787037192,0.37653027424746566,
 0.47041546020587544,0.5381015477447447,0.9685083196194082,
 0.11061824067108916,0.4665959068853759,0.16362315823275508,
 0.3809955173235988,0.33002934770444936,0.3668610904534302,0.6487923619940767,
 0.6585463060751608,0.7942978212911801,0.28574666542711746,0.63658034304557,
 0.2558173251896473,0.8801493485671333,0.6414710567735271,0.05008670397704085,
 0.5089278047137072,0.050478502433881745,0.38180739556780774,
 0.28908382877862426,0.1270835963089697,0.429002680678816,0.9641684652991045,
 0.07678229358932642,0.5548255949916316,0.3920688485484807,
 0.04124099064631359,0.001894244628935171,0.6969873370663034,
 0.49428735371199684,0.336077956687466,0.6696565116578773,0.6373412914025176,
 0.8172900952349422,0.5035743188756177,0.5409059056248547,0.04766198049682102,
 0.6090298765650166,0.5571764023762714,0.6142717348561453,0.1476997975988178,
 0.13354910612449167,0.7278748380902204,0.40515311754325034,
 0.7318839153431829,0.5231544919279342,0.9181105027350025,0.34673455568442124,
 0.07297027396779177,0.25408123495102686,0.923911718891793,0.9458095848204296,
 0.4025296487928015,0.15452035119807617,0.4702773916230988,0.2773513663866082,
 0.08679060409841322,0.059705966668604615,0.3744490694722702,
 0.24334939739238903,0.03119938004422207,0.8100684380874145,
 0.24058885804657082,0.1270163660035496,0.9458994007480573,0.3043729475783231,
 0.32891776568573394,0.9745361561756402,0.5133749555116366,
 0.23358871894369015,0.5176569036449706,0.08053562216098209,
 0.8440392204898985,0.449816006408307,0.7950266442202517,0.043070384540591755,
 0.7235855517002668,0.4799121614708307,0.5516535775070183,0.49022445097937806,
 0.5167647908310525,0.4670787175349629,0.9847611728445725,0.4280589107210695,
 0.8112210768540545,0.7839475741630443,0.05419809929428665,0.9215244088685429,
 0.519485517631735,0.11601113573067467,0.5986002436333407,0.18349253841546675,
 0.7588553229552942,0.897378954854858,0.7521193323808479,0.6373577816527124,
 0.11367465081938133,0.6856549633289317,0.3145024759513131,
 0.09104697627885572,0.2408081590316694,0.7675158726552248,0.9866993136566822,
 0.7550632902194484,0.5508449430032357,0.9320239805734818,0.9216277146720214,
 0.5427414093585088,0.42988779183297376,0.3737648987278622,0.9571913393398616,
 0.865271570277856,0.4168666871174096,0.07570812206931243,0.9577786779183639,
 0.4940748861314854,0.326810918055106,0.8255815361566242,0.20664036033171218,
 0.11542061516621782,0.8321100920742307,0.29565626016599433,
 0.7605310882885491,0.018711340388474218,0.021980648456947804,
 0.6742206474899346,0.5119457428939157,0.26013841799071113,
 0.03766734068411448,0.07972080504596057,0.6764935895960258,
 0.7037510105215253,0.35074793143692007,0.4057549360718574,0.9152398631781409,
 0.7736726153544087,0.8114844034792409,0.4767149084678053,0.47483533930128463,
 0.07483278725508524,0.8544995226698362,0.9332578637461321,0.9653659063488303,
 0.04578844418021577,0.3846218299640558,0.36184424571009743,0.600347296612852,
 0.8190896529707615,0.3028106919834159,0.34610199802504493,0.8284886693994461,
 0.8203416084876041,0.7316025737833575,0.5165478942748858,0.6078612197399867,
 0.10948874490922023,0.43671267340094366,0.2232927973918225,
 0.07902066278060094,0.6799524666020869,0.6886744401449278,
 0.07824824482419168,0.34536290198706376,0.24046549283951102,
 0.4865665706778335,0.58168894102598,0.7108837729771607,0.10357070483575848,
 0.23596470555375837,0.5604400952118965,0.639641505431938,0.9039220443541033,
 0.32988960448015336,0.6774453198197283,0.3760171788675828,0.9053783405042345,
 0.6046883149268142,0.02549058297966944,0.09397682870452484,
 0.6426925480322521,0.8654253059997803,0.3637169051026222,0.8128005953890587,
 0.1696237545737872,0.5799664797809957,0.7578236389269154,0.40048170448726195,
 0.07702599931349785,0.26558112280978874,0.02196431144496902,
 0.5585540482965932,0.547727246948342,0.7121644386923831,0.806084600129846,
 0.9890457758534013,0.9467053766445982,0.03829753200745045,0.5279450107787401,
 0.9653531807564625,0.07997466580738466,0.1121320639644795,0.9508537120116675,
 0.2377954694498613,0.909703825811564,0.17392335807628034,0.5583113113368413,
 0.3933895191194964,0.33200187266755976,0.33317110204197875,
 0.2779914629999184,0.7134597729825266,0.860190189567558,0.7087968435114738,
 0.11293444084157578,0.04796400575694537,0.328406088104234,
 0.08374721578173105,0.84782595293009,0.01571575449411866,0.48770648588193877,
 0.41895672247473514,0.08414011199548321,0.6426337782666177,
 0.8869881436458411,0.3029366699829238,0.5206639918157091,0.07401780901170851,
 0.14222214756803453,0.23658801159749487,0.19949499413627847,
 0.19167901503941198,0.20559064353482093,0.13966858600842724,
 0.7121021325438006,0.24142721213793017,0.9215768712442984,0.8194559468630689,
 0.31702638554933693,0.769438945726593,0.5043795609922801,0.49064570941852703,
 0.5429272684726094,0.5882645605432876,0.4890551161725649,0.4170945942029165,
 0.8240977251092009,0.35633737545882016,0.584981898267362,0.1302112295710518,
 0.4581926935066456,0.5843460553699529,0.1838548844732989,
 0.041984204546806514,0.47663333475640224,0.4059471356817721,
 0.4145388959101096,0.41377862556833334,0.0774228565301267,0.9578834856744801,
 0.40152962716905816,0.4352519797125247,0.4679910099961082,0.7786858188199706,
 0.5581647018114676,0.5591272955463576,0.4860238227489282,0.6738947153481091,
 0.980995706354276,0.7703670490639898,0.2551835896144098,0.8795538854122233,
 0.9377422522500034,0.26734439790283315,0.17877652164446223,
 0.6766466530338379,0.43636021719987983,0.8388919777761773,0.9294560607398079,
 0.20827396275780474,0.5752824508567619,0.8747086621555011,
 0.09398003002115818,0.9489315341281147,0.8397577894757294,
 0.32568341864642003,0.1475456737565457,0.6315418402651208,0.9685430897601085,
 0.9202511552988466,0.7597234100157565,0.45915197341709735,0.3728687316302107,
 0.41653145288138427,0.3122866647351913,0.048843116259482566,
 0.7314368543828839,0.28934482174866827,0.8336620332656871,0.9341649190240993,
 0.7115537907625495,0.5345634289109724,0.410729053144394,0.2293489867842291,
 0.5506525185579654,0.8222706701353389,0.4166690222444245,0.1823774510157692,
 0.48194334714601084,0.044422487078185635,0.8634162543788849,
 0.20771610303457855,0.28074559008860245,0.5289541242653439,
 0.6588037931239391,0.5701128084964071,0.3461449958970164,0.9969867379346447,
 0.5961365057841755,0.7878968533653251,0.16702659849427226,0.3586266296773254,
 0.3023123553041731,0.2774194600594707,0.665704117661577,0.9499707136887559,
 0.21975117191246185,0.6615194073576824,0.8125247699695197,0.6130006273665651,
 0.2182259958312689,0.6096052308300265,0.3891171005024575,0.36733018011644436,
 0.4749682156111503,0.020457287121666168,0.6845827326720819,
 0.9977540175707104,0.6322619348183403,0.9854141778765634,0.05824862756393856,
 0.6211121326699028,0.8509545537260583,0.008103572116550994,0.519156188154435,
 0.5392288880275984,0.967328372706628,0.6352072436066365,0.37072885020958113,
 0.005268368935859802,0.21213124427762897,0.3012295508933297,
 0.6904630793116873,0.9049971759541664,0.4832468324747584,0.5511587587617042,
 0.9022313563005571,0.458848564102736,0.5712497950571003,0.5297610373284242,
 0.14590874226357142,0.27393689222125794,0.4552264885232131,
 0.5569465470610999,0.8406721364423962,0.9182549998948938,0.183585334027076,
 0.07894966383206048,0.3032991310781822,0.8802680312749201,
 0.28175701090718164,0.9933233588012843,0.10464693272917569,
 0.7948987187429115,0.39817226155014585,0.9134203599570707,0.7298859381517415,
 0.2737961570841537,0.8394204242574307,0.7152312202301448,0.2519494027007352,
 0.03205194475637807,0.8511963547100055,0.08944579664681351,
 0.4909126804406819,0.8706523407107503,0.008553067890980426,
 0.05335785361323575,0.6864556350607732,0.612393788922665,0.5034047060204523,
 0.37622264769700264,0.761445385493177,0.4429400897960625,0.7033209714277957,
 0.2655410836847706,0.8635232679413078,0.6156195525768033,0.9488186396026945,
 0.9668109910548894,0.0017487483344758026,0.9324699911288495,
 0.5612645770094161,0.8874138142278338,0.413795520944201,0.7240545123882198,
 0.3379371161538094,0.028723706355099465,0.395479775180293,0.4979816666067648,
 0.7431240195630825,0.4435384656288408,0.5695596939995757,0.4773687245938858,
 0.5126435824288322,0.8138017703816298,0.5056817009519189,0.9621258991876238,
 0.5720032667144143,0.18952772280806396,0.7829763993940452,0.6897074358709188,
 0.9546088616556757,0.7658072099713356,0.9383815089145048,0.5039202012158263,
 0.24816652394028627,0.6514841010243874,0.4627992888678636,0.6833413995132392,
 0.024181021097934186,0.2707306724414892,0.029353057550854822,
 0.6706723373313526,0.9331627090543907,0.38908025944751534,
 0.25974256633992976,0.16992522010245903,0.7332079784218799,
 0.9079460596471065,0.48940016466453073,0.9613625793209071,
 0.015507810082378448,0.4423645894226068,0.6596420913112649,
 0.6175673041090918,0.6029458668027639,0.5877521781661095,0.9148757146291613,
 0.9207275887544356,0.7249899861650174,0.029962429503709198,
 0.8231378064588769,0.6756969525208418,0.4329183742067034,0.3788606807865633,
 0.19762961323629202,0.12852100990939896,0.5262028588289012,
 0.7513305071670168,0.0022187851144590542,0.2384898292501958,
 0.7285118477703627,0.126209429368212,0.38193422863768844,0.19827179255804805,
 0.01645724893897471,0.5744625128931257,0.09080619357695907,
 0.06985248402533506,0.9185052131974409,0.558213219027267,0.15969362766015605,
 0.35695524300190895,0.7070159534577937,0.3160388969693092,
 0.03790174626282761,0.9849823940017058,0.4023551858269291,
 0.19602127428593596,0.054589331579324996,0.7843234285649668,
 0.37348304852339054,0.14855934767504397,0.6834416014465591,
 0.08904485738761025,0.3104069601895705,0.008306262038693446,
 0.5644294105242423,0.02472725746246196,0.8604792365468635,
 0.04094973001949165,0.3855778193003708,0.7294912961675123,0.9072565454511818,
 0.38534440889186083,0.9105864146222401,0.0031470070859923593,
 0.018052114876671133,0.10547179149751096,0.9397709501326893,
 0.5104507687648505,0.6771744595778154,0.8615276981157833,0.5041778573018935,
 0.9847386638017457,0.12230605599169064,0.7992264816400574,0.9048290621917333,
 0.2648638453375707,0.8736204628201487,0.41718949954828977,0.2166164144516075,
 0.4502797141369205,0.7045398679369117,0.8107936544382511,0.31270637194911177,
 0.5489138651462582,0.763879317880121,0.2210701269954346,0.9000388204564838,
 0.6233794749992576,0.8208476805518398,0.8779945415770376,0.572774853961995,
 0.728601117194954,0.014484331128414185,0.6465658485971523,
 0.18866369774611025,0.08315798524104479,0.15392159332497712,
 0.13860956020149273,0.11827127532347848,0.8661736039383321,
 0.09417117730217894,0.9576040365321294,0.4127127923902737,0.9040966437276706,
 0.24478541475247328,0.5388065724085903,0.38240880695887647,
 0.7933240168761924,0.267725605394892,0.8151690249756692,0.7493471587055273,
 0.5337140031176375,0.888573226959339,0.11210939761867045,0.08289817650401532,
 0.7800791233767226,0.8469959777138728,0.29875284922729106,
 0.07539686020645131,0.43231088579866206,0.732247777272479,0.3353457608048691,
 0.739452998878862,0.061545152098194,0.9656367007084913,0.8155351012221207,
 0.4242111780988963,0.3128220264408109,0.4329322262620935,0.3790578878240012,
 0.13704860854936418,0.79080919896015,0.9523434328912642,0.8900216500691368,
 0.766442718462838,0.29236923396455883,0.06943603474461713,0.5492099785698638,
 0.14011494647613498,0.18018633255457006,0.20151348518322887,
 0.706405998612177,0.6107709273008398,0.7731883112351358,0.6135095546612859,
 0.5014736429035213,0.9166354866294727,0.21142453531974337,0.2378224694288733,
 0.005855190596506743,0.3304614963656568,0.8515055597292861,
 0.6813482366379541,0.9056294996487606,0.3579447065856527,0.08465183457345027,
 0.2647424369111878,0.9149805228338335,0.8094343263511767,0.9619223652360253,
 0.8478162667172868,0.9404968054433277,0.2089293172440001,0.36888459885262126,
 0.10836070814541776,0.838744173646786,0.9532138405318574,0.43638423074350685,
 0.8038182212232124,0.10244516322945918,0.42609510981493603,
 0.2723450272877911,0.5747808139348821,0.9964687080110924,0.584665682601841,
 0.6863470333123639,0.14531549373736286,0.9362673541546266,0.5905413271522092,
 0.7364883117840832,0.6126440822600291,0.8508203510476426,0.23587988855504483,
 0.5332380898322427,0.1414733754809494,0.03720639847692797,0.8168229213046347,
 0.733242451500975,0.29584932561404054,0.005476842925854819,
 0.9214674567771212,0.9977916460955689,0.8447558854520315,0.738986154240221,
 0.8454533890174017,0.932792927857485,0.32255919067022554,0.676442978741499,
 0.09637864473559399,0.3149646232150385,0.348590060312835,0.9577846702833513,
 0.8514936646904189,0.9931616182280327,0.36943235516491035,
 0.22388410825923588,0.41838354754304663,0.2605237306009546,
 0.16436369495858083,0.01646265041955408,0.9641990258941355,
 0.9232754281097078,0.6704314021389765,0.10897100761011957,0.0558221993764888,
 0.9355633489562699,0.9065824202289361,0.2125314899033005,0.08962842382801084,
 0.5806073128114763,0.005555670798776946,0.9045920527643608,
 0.7652210339260336,0.9222568200581234,0.6683865887552419,0.89545019467707,
 0.7363907870304661,0.4235032310699446,0.6891702863395313,
 0.016684093261467137,0.27525268943054926,0.24570134890440787,
 0.534581907252124,0.11295236802475472,0.38365040069440737,0.2677686957508847,
 0.8911155680684184,0.3676945331174224,0.6596131545008116,0.15606568479507077,
 0.5976431898337573,0.033095763776124665,0.5131678883906791,
 0.019526761589124364,0.4245266400886203,0.9940452835444388,
 0.21207203533310925,0.7746050120598091,0.8009297922050084,0.7971858911991616,
 0.6479548696666502,0.10996677447969816,0.36263920021774765,
 0.8907423659983877,0.9201818071278609,0.20151984332641193,0.8266772033938541,
 0.19351691193162246,0.656499222730904,0.7146944010724998,0.475770151199884,
 0.483527252033355,0.9126043265693566,0.345402621321936,0.08064763056712831,
 0.26104618529294354,0.4939267247568424,0.09704145708012046,0.679979547165314,
 0.07373470837152862,0.9949284084892407,0.7776439594424525,
 0.06844786585910634,0.060033541442299776,0.40932350812074236,
 0.4455572210916897,0.3607003621717505,0.09381839063549768,0.2468791859494337,
 0.27197113012845753,0.21796864663380777,0.2382375740408732,
 0.8242161135727281,0.597483797039955,0.7076729413445273,0.3462268881274112,
 0.7966378506618302,0.6520079417294897,0.046706488787810585,
 0.11353765084906087,0.7582699198817552,0.09254545827970417,
 0.5704008508716822,0.5290894298590717,0.14064606587560458,0.3216615152958543,
 0.00533970301104969,0.22907370626861967,0.15241122108476968,
 0.7847360097878071,0.9765456671045565,0.2740053935298,0.5041397857616426,
 0.7435864536459321,0.11961421153464635,0.3539467234855975,0.813829616216955,
 0.018509500103903398,0.09846375751738445,0.9563262627562584,
 0.06264108575838334,0.22114057523957775,0.5818451855501181,
 0.5650938791926501,0.9223950120526774,0.7580994336614155,0.6459430198099073,
 0.9439336712857266,0.07531514850586807,0.49200568385654586,
 0.35332792697658955,0.8954148301524333,0.9729911588453795,0.9623708582616051,
 0.20396832542922405,0.6084674955635525,0.29969860132257575,
 0.3067431548193855,0.8744421121572626,0.34611670266698547,
 0.39203539320377256,0.38494810542407276,0.8345663247338962,
 0.16222043177864176,0.1197989843534154,0.026965725306992372,
 0.8779455309473239,0.8436850652404231,0.5256523110971341,0.14582144442705425,
 0.6399232392212082,0.38763189054429825,0.7359313042794744,0.8038364903150377,
 0.1544339291162874,0.9209733070202352,0.9264583316973458,0.7072932688851874,
 0.10936678999093263,0.48368078911997836,0.4795669000549988,
 0.2551431914282565,0.028248274111875804,0.6247094710789005,
 0.6082033435368065,0.011558408786850216,0.6356030031226607,
 0.16084011994269454,0.2887019336153772,0.3197764735921822,0.7727237012653309,
 0.636149241083436,0.6484083892094585,0.693647572400271,0.2742896849247324,
 0.5524523018646563,0.21870926685622805,0.910841659826036,0.16028558027749096,
 0.22711251770601315,0.8957316539797274,0.5085342582540843,0.5124305786937561,
 0.15013274020855938,0.11763156447970824,0.3777110147696694,
 0.23971076944588154,0.7035688771108675,0.5628918334309867,0.836466480032543,
 0.07613145338058003,0.4335323621112026,0.7967229083869523,0.5517391436016071,
 0.05933222458842724,0.08170653020604624,0.7369244614050818,
 0.9898649247702098,0.806453017394156,0.8782731003085505,0.3749859035136497,
 0.7715219651119234,0.2816090638026173,0.07131334381339038,0.5685612892097393,
 0.952016136948232,0.016493253057685675,0.13919730419180432,
 0.8461063470744646,0.29204234524658157,0.32284960251800066,
 0.08493516939180301,0.3352221579401521,0.41022148735636166,
 0.9489660851664854,0.6491483542896197,0.5406547008558851,0.6423838473703267,
 0.05823615166972029,0.02852852697326913,0.9666411285211257,
 0.5427568678237065,0.8817368211519169,0.4109820785853059,0.984659000635407,
 0.9569215024144954,0.6405199706542481,0.6140958181809228,0.18069855286101022,
 0.5260631101746077,0.7585392113575213,0.5269545284822063,0.425786120795,
 0.26552219035941016,0.7960196471262777,0.17946264888630425,
 0.05034436789182928,0.4714010958071424,0.5191823446498881,
 0.02410882322306218,0.1303662228092518,0.938593995176737,0.23946247591166603,
 0.34011999193529385,0.8480125394675265,0.528316070039455,0.9947967002914351,
 0.6036007007254413,0.7495562027997641,0.20418436083813218,
 0.37965015362684007,0.23380633597872091,0.6102070864009448,
 0.7467847843451199,0.687108304485556,0.8386111904328044,0.551468647437859,
 0.09444945733962062,0.7067954484455956,0.4419801320980832,
 0.28850628720051763,0.11399282986492265,0.5782191450703924,
 0.18760655731182363,0.5491491132844046,0.38093300368403393,
 0.07892261820133661,0.4467004381382391,0.21968733066924095,
 0.7158501567290058,0.16221325210936477,0.014709037147317527,
 0.8992318936549569,0.7435436048430458,0.35029425851602713,0.2110211043638457,
 0.047459863005613334,0.6400773375795941,0.4776341402134975,
 0.27294328943650537,0.5775047958012568,0.02270887562422963,
 0.3883702373842921,0.4662358362037634,0.17074066821343103,0.9069465091393571,
 0.07245853146126258,0.8160487780267353,0.8524817724057497,
 0.15982812864122353,0.7552885597887153,0.35494745656971227,
 0.42785035823708717,0.24773391674750245,0.5709686289931986,
 0.6836989181731741,0.3769289565167404,0.8519753753045964,0.6762015755898588,
 0.36561052380747916,0.23862882931280327,0.20902663117996623,
 0.9272073820319735,0.7588396012183717,0.07341837608327995,0.9726974957099543,
 0.12238518701441958,0.06614636881432712,0.6530214832954783,
 0.8295566154700894,0.9170214792077653,0.07969160361966154,0.304308981897373,
 0.12448020583564456,0.06709666786645918,0.6298028635182747,
 0.6045147917620497,0.5880121093132361,0.7571239720516788,0.7318454077821719,
 0.34247764631938604,0.8405427836648793,0.5359347087394074,0.8511838369217921,
 0.5245636533638502,0.961666517573022,0.19088142604483627,0.8527480175975343,
 0.9391075682403839,0.9590365459685685,0.26610681445261763,0.8605074791804791,
 0.5029428936809351,0.4894438056741274,0.5549000815905804,0.9182216606810196,
 0.9570084048564927,0.027180188113588022,0.8412840134156815,
 0.9420879520753798,0.5166664246332268,0.29919270571894896,0.8452869985562403,
 0.6935991215272497,0.7807340391177179,0.15275768568567827,0.8442815879599018,
 0.8317309194851029,0.7908447352206789,0.5338505897262512,0.9360375844428157,
 0.29610737243295193,0.40534887733456015,0.44748847460328844,
 0.5097695347029416,0.48134605851472034,0.530892430945347,0.16864157278773206,
 0.868442427820799,0.5867194017476899,0.08313377115408838,0.9211007068314521,
 0.988736455442171,0.6604072711339486,0.661389901627473,0.011012462417383073,
 0.08710273690665271,0.028131424349916112,0.29604544687160494,
 0.17486003517743343,0.6172922013220801,0.18843549198450715,
 0.11640888611056743,0.0689175896951193,0.7447105300887377,0.9039426717673523,
 0.5605441809571843,0.8207563782422508,0.5068626142893997,0.37937428281687957,
 0.12118216120785608,0.7417956823088827,0.9134131279466887,0.5366571305959649,
 0.27038974716373565,0.24969264392695523,0.7399440404081815,
 0.26863726769124385,0.3599286315273337,0.4033609041006906,
 0.45269820788322446,0.5733623500222306,0.6533567522214999,
 0.42614098805607903,0.7371618774686354,0.5690399634411756,0.4524960233527271,
 0.5551690912838684,0.9651348777155676,0.4669591542599929,0.4509863909170472,
 0.22794661588846132,0.5360602265094649,0.4829570110925603,
 0.26736749374840274,0.3493620121948039,0.15537710979253339,0.930496365318247,
 0.9924590185158022,0.39524871279890705,0.9182372337344514,0.7254773686044946,
 0.3207116964644481,0.6629824263230638,0.640089903350182,0.855534103873557,
 0.5915438420524577,0.7377859437901935,0.012276375982920418,
 0.7966734516249468,0.652576212642419,0.29407078457004,0.698082756046118,
 0.8942749215591385,0.8789735175357152,0.9659311535744134,0.22276173366994612,
 0.6387202772596239,0.8367810225918724,0.21538941128186928,0.9052987537121634,
 0.5607500153921982,0.5378420577192425,0.5908626651345281,0.9167285337320549,
 0.906505787728245,0.514776294153854,0.669544974903227,0.45108571618807747,
 0.928953322123982,0.0489468298816389,0.7179918871124398,0.10033950708040262,
 0.5593945967023458,0.6137509969123808,0.4815108610014146,0.7130082176646181,
 0.3085979953950464,0.6774441835878982,0.2410409833298205,0.8129887656432617,
 0.9960628583972883,0.6456651235752854,0.22812280100196636,0.8956743897299148,
 0.10954261345791838,0.32184904146652105,0.4681033484481931,
 0.13189887467840922,0.05025215054126786,0.43449468589064266,
 0.1434988957282074,0.32685648447708604,0.7419557047737887,
 0.08049043004470502,0.13906874150516868,0.02415340913665176,
 0.21141890985409528,0.8207493074543875,0.478203654366941,0.3158426130062486,
 0.5414675705352928,0.877745967092366,0.5286423995118265,0.07205822157665653,
 0.8764878547326802,0.662080638656197,0.45233173183846376,0.6067378951989758,
 0.4551509732827985,0.8508264571837449,0.45548735400565177,
 0.020469233506993034,0.27027610289462833,0.9378878954818509,
 0.7010231937970202,0.6966766883785118,0.19783174500680523,
 0.45598546607225965,0.6945114096095163,0.4004217200631114,0.8876587703380397,
 0.8758057514000352,0.7264399488824648,0.560077375381442,0.41082999168711454,
 0.3457004429975814,0.979857630753464,0.7942396100015169,0.4109016104104134,
 0.4540048008662123,0.7221455543062063,0.13828291279355165,
 0.08899551539615036,0.2948549872746673,0.31869874031470324,
 0.9571500060302889,0.055566832711902436,0.7576882691172897,0.736563072187896,
 0.050525317217203214,0.923937573835434,0.47856146290215695,
 0.8700139098999708,0.6488300788895495,0.517990405049974,0.47491231944844303,
 0.8661351440880736,0.8584307908508658,0.3968566839480211,0.9520797695920099,
 0.28015095939122325,0.017341732853337954,0.8735309992927436,
 0.2629787497041526,0.26044380348693874,0.6422163799555742,0.3500921824989248,
 0.07300337202578278,0.385626274372723,0.6511905968716153,0.3740442237264715,
 0.3807283975573499,0.16337669356831552,0.9284955382886105,0.7532885088453765,
 0.6101153118497596,0.07953346167432762,0.6432646747273882,
 0.05789939454659032,0.5838622273068403,0.15053198497146703,
 0.24534450710224553,0.38001949032939,0.9072624230339431,0.8588606691881684,
 0.5741606131753079,0.9231630986405681,0.030061978369862685,
 0.9925565468518599,0.8693157881478155,0.6795939003156152,0.816488099942611,
 0.7558249578644397,0.2793573630298062,0.3129923438998876,0.44117318041915354,
 0.8398791272125399,0.49919853784056256,0.5315044745186599,0.9124945295447106,
 0.6577998352147119,0.9591335055314032,0.9099367310457617,0.6688381920570199,
 0.03607783720097257,0.5260244403029343,0.1382485446669871,0.7376561135128866,
 0.5147172427611773,0.2023045024610174,0.2235954858004927,0.7082891732467858,
 0.13367545755299837,0.5136277820876435,0.2306298653446589,0.5260779432955643,
 0.5491981829545644,0.8234113457826887,0.45284763694817565,0.5890091951104218,
 0.7276012926765176,0.35921400594009345,0.9028012885281393,0.807010393300263,
 0.6011562966390922,0.9636146911910713,0.08794460446050045,
 0.016275180162174552,0.8261735671604876,0.657441029175977,
 0.26752950959756483,0.6353307494064357,0.6248258985485431,
 0.21103322044518424,0.8624277429428766,0.5156798076440507,0.7126508798217602,
 0.8313234418885799,0.2153050853479912,0.9969616433777619,0.3573991280087465,
 0.3565296329430355,0.08832588383032836,0.06287761655789159,
 0.1356418516953718,0.8536409299532808,0.09246792031724649,
 0.11011316745071387,0.9377653479167802,0.4714312208531972,0.528080285534566,
 0.7434211122185964,0.12185240677259379,0.31068307640119397,
 0.37098252891663863,0.5109594944189461,0.9668423631018807,0.8318735910220081,
 0.4687258013574068,0.596879114618901,0.8603159494290638,0.7269907169667967,
 0.9931048450676239,0.13082449581479816,0.02834780350750088,
 0.8983684026517483,0.27156785471290357,0.7637049220265326,
 0.14487188385339655,0.800116182241212,0.237159270545923,0.1467263118501021,
 0.3717006829169015,0.7147823090906703,0.9653780919884308,0.8775515184349261,
 0.7256712007094748,0.8960923251078783,0.6454788006387385,
 0.010227387513369934,0.574802712832935,0.8427244156586056,0.3664595243868358,
 0.6144824015338204,0.10516372421467779,0.9203997492342835,0.8310778568568414,
 0.30045803798442416,0.048053563676981104,0.3757497573845374,
 0.9355316372229795,0.6409777326356042,0.5239844145266722,0.44581525804798405,
 0.21000438057111914,0.24710901287964593,0.9776386474029886,
 0.7927875688345699,0.24713739919005862,0.2930133374971178,0.840259803870529,
 0.6097988928925874,0.6470394701073912,0.14129672716339914,0.7707695478862769,
 0.840225242622935,0.22898163598101307,0.07381825123782049,0.4404712318603208,
 0.5785591057427081,0.6756029834851511,0.24194362357985344,0.618752544245055,
 0.04080070623787713,0.17845329066426252,0.9768172684303383,
 0.6466065757545294,0.0423231858204377,0.7031660593880491,0.5063270992737561,
 0.13572009780709937,0.15155200560646087,0.361830644693103,0.8018594119795844,
 0.14463332249455763,0.3165324322051375,0.5677198707426805,0.4757213173223087,
 0.5507960957698036,0.9607623578836862,0.34504701459927734,0.6118364771020355,
 0.7452935418070994,0.444132827417567,0.24085989286485665,0.5451504259762894,
 0.8690200965682782,0.09579846447262075,0.9740680423673965,0.2836517957412408,
 0.3221306782452755,0.8953751183580325,0.9909971670003287,0.8282314682612835,
 0.9941854756563272,0.04566633113691365,0.8830993634797624,0.3084789193148728,
 0.5308532924576622,0.35325837540085914,0.4638063197015173,0.7111258905245381,
 0.47172330930363526,0.12973040940139247,0.13071052405527284,
 0.6160482889134534,0.22081961708335096,0.29279281449811756,
 0.28888086654120626,0.4761957555390386,0.6869674701577593,0.4739265073086879,
 0.27250335645971613,0.7296285886059235,0.23824893405920367,
 0.3047772137108653,0.9334543557011371,0.9859479184250817,0.5698087670427996,
 0.2883067487149089,0.6989775699533302,0.08464937387340654,0.8548480803804615,
 0.9231662326195306,0.06023686693241226,0.23264435553560514,
 0.7186490268322533,0.9288976604681163,0.241867952333497,0.2079606100047312,
 0.7008302774603858,0.807489200321943,0.1267390066739178,0.9061030366139942,
 0.5248021371440749,0.14088529566213148,0.6154342212554181,0.8057291166634445,
 0.8831529274707678,0.27992814987776315,0.3646516902589827,0.9316575997511859,
 0.21211962763370085,0.7930246671116975,0.8999953585208448,0.8774136240416446,
 0.6186239502427253,0.9269073439935651,0.5331844394892031,0.5758564799312998,
 0.14116174880374788,0.34857230684226304,0.6461913227893972,
 0.9456891503127974,0.14131574807804959,0.6682167965412958,0.9569817731120266,
 0.5756064094644079,0.6620185205326818,0.14144829876177167,0.6374464863727787,
 0.9167577515374634,0.9451353945833931,0.1313846344488322,0.5247480783078445,
 0.7188971424046162,0.0737196463741111,0.21718667905150013,0.6889778820885585,
 0.3353004385415067,0.855489050883934,0.6558213910593702,0.001087444551215122,
 0.11326057210805429,0.745261915593913,0.28161327364980904,
 0.42945010254561833,0.5397822040855781,0.3494728402693389,
 0.14403633929536896,0.09209320815217181,0.2832433945345787,
 0.7790888015485279,0.03504227137800697,0.4753507433818005,
 0.09563265616152183,0.6284438233596202,0.22625777826188065,
 0.7262035846528996,0.7237127542546841,0.24036307000685997,0.4934455818012,
 0.46768818621403174,0.18171067203307978,0.4538822855970298,
 0.27406853202433834,0.49953632113648927,0.6997998175113224,
 0.7503414003267844,0.4638229342461727,0.57906093211311,0.19184231503056015,
 0.00785928324922991,0.24355940248750474,0.18232485914859853,
 0.8596340250140976,0.8819467057710763,0.9486490344370666,0.09552537462316524,
 0.8181615946568094,0.41251728132402066,0.2830875320828583,0.3008736328984397,
 0.1417266693387692,0.08826835545319489,0.5525460834834718,0.8071141540678195,
 0.785779739829841,0.4104288659971458,0.5443685013631631,0.8132116975305655,
 0.955571583482562,0.5367225942110571,0.7669109671679261,0.3646938877165753,
 0.839000096980435,0.6916848179131959,0.046220234907954905,
 0.47185103341181844,0.4087117166014852,0.26259474149677153,0.725783408576659,
 0.39692259258460316,0.9731184014347288,0.7821249659032348,
 0.30131396532002885,0.6033334578518303,0.9355584238826085,0.5679524318082994,
 0.48438258811588675,0.681615608893881,0.456236757794132,0.8120474564826587,
 0.8479240384605132,0.756383946162052,0.5997989634824578,0.030032782422661608,
 0.19539925178319162,0.5413733267983247,0.9308493436461714,0.7123379706215069,
 0.9817980239547477,0.9675524194845933,0.8679440800293136,0.20658179121849496,
 0.9837543492319302,0.9184644151897706,0.29099071064635784,0.4234208493320799,
 0.579397413851735,0.34855877503915034,0.26484161375280935,0.2536964880023206,
 0.9954193261730956,0.2648786480365254,0.08318773992371398,0.5829761597661427,
 0.1263792065289684,0.4084056393872224,0.18039661044646482,0.6778542606041295,
 0.38611116015726143,0.12703541188019551,0.10146447829668448,
 0.4362908565532081,0.4048544346391808,0.7515386846334139,0.682525074978206,
 0.6033747784205943,0.018392682671720784,0.21996481637912046,
 0.4058267283184007,0.4984978441850971,0.9761505965460582,0.644381722722971,
 0.5433245596965336,0.08554858693491385,0.32494517650961974,
 0.6346610218208848,0.1250550029017814,0.4480264809427207,0.6223843068716112,
 0.5038985536232068,0.44066841879242813,0.351343267025394,0.7564380736305598,
 0.17365134113878722,0.12970500104004024,0.1420566308211595,
 0.6395803794896335,0.7315810812708761,0.3940371451362704,0.38925285797335396,
 0.9902695425400895,0.5270892256414632,0.6758536573785592,0.9893716522708964,
 0.49651270352197585,0.2707067626724625,0.8776880723289979,0.4077229082176964,
 0.20317635619239016,0.776910578767008,0.6116814906373582,0.06706027737138132,
 0.4888260401922677,0.044214744640674986,0.23577304568278334,
 0.3706958576994799,0.9490368967320584,0.1555267396842226,0.7027463142360799,
 0.9145647130884178,0.8254359334976286,0.2785524261888439,0.289670504455522,
 0.7969217333792036,0.8253307760968365,0.09437660929705449,0.7820294192133846,
 0.061162721384729224,0.5363657967431508,0.9481432576259083,
 0.5043374010818913,0.7031607144179235,0.2915417642956849,0.10339237093824716,
 0.13815128649959552,0.9893522469676546,0.07687201521897102,
 0.7482239133082147,0.7447623826171071,0.5472864412290518,0.1616741313635347,
 0.3888952582445869,0.084597931012934,0.6931651928225651,0.4770226198053584,
 0.39336193738680814,0.9899288950888133,0.017366855361929168,
 0.3587584638456215,0.0965562718835331,0.6037672975712756,0.6460999519266056,
 0.8456828592898507,0.8540594989282978,0.1355234635326349,0.4023887551875722,
 0.23726693146064715,0.28074599517081367,0.3953566930102326,
 0.7916437094600186,0.7637559878547263,0.672704595906766,0.5814325746463903,
 0.8198313596465394,0.4852396606115068,0.5032998024496859,0.2753719177970496,
 0.550453021258021,0.15181798397596802,0.6242745757955096,0.11055896801110576,
 0.005483070302392257,0.056260745989115124,0.6434155640069383,
 0.3774003329513915,0.15484613307268102,0.9590725960298456,0.5014512040662991,
 0.10261288845504835,0.4010623553170064,0.07485118070347918,
 0.9935027215848844,0.03890391525713588,0.15647148670719424,
 0.3731851978301195,0.36629353689994804,0.7191166634830272,0.8757029923199006,
 0.749005501762969,0.9278837451760839,0.888626401921277,0.06667124101376631,
 0.9427135076082598,0.09698086734920186,0.795494622065023,0.478160362311572,
 0.46708419980446125,0.5572331364772933,0.9880587477248293,
 0.11424489148622263,0.36971334228084296,0.3815545544396063,
 0.10122876035511219,0.8594999292501668,0.12218798463482716,
 0.7078826677549321,0.7496552463412806,0.6074744882679447,
 0.014856395118546084,0.7133874490893642,0.49111237752354897,
 0.5697205003222848,0.28211648567338876,0.032959915205535206,
 0.9825828719444717,0.43789708630233737,0.5791884900922025,
 0.36744874257705507,0.7894445986776883,0.11643856580947554,
 0.6179908099274288,0.9554983186057474,0.32223369540171976,0.3192288807904713,
 0.8122583049173293,0.3991508095405595,0.047866370940134684,
 0.8093723371905293,0.5208945674948018,0.31364804348705844,
 0.17876317036669753,0.10530368342427043,0.4935186677590906,
 0.31075059214477085,0.9235910178640423,0.6810427755987074,
 0.06697873540685106,0.04883664591365722,0.5575337162789212,
 0.9567333380941826,0.8741413292388688,0.30766822286718265,
 0.01141104136347082,0.49056045522816183,0.26938807595485503,
 0.45954153290648336,0.880664037938474,0.568395944993658,0.6073634902215351,
 0.1969003803086966,0.17576178145935872,0.5844368838229916,0.760337918835791,
 0.3610601038669974,0.4651957983572276,0.9243167697478861,0.28205140265102335,
 0.4361847423750259,0.747085274230732,0.7595831740067358,0.7850539190651917,
 0.19097630214403583,0.8555466729359793,0.2248650644896516,
 0.18629425770652264,0.6449743859362213,0.03513987748617953,
 0.14821855207122814,0.20976421336031292,0.7891500833026723,
 0.4276889303485726,0.5515746350239564,0.38453537988993414,
 0.15629956070437545,0.8561421991447062,0.5416983490235661,0.1763303442468902,
 0.28116797163753016,0.7522847189537036,0.2740398072135348,0.5384121969285053,
 0.028884834054361708,0.8112484192709235,0.6826919738306017,
 0.3344112970000007,0.9264624791489056,0.21031664300846897,
 0.46707334988266824,0.11780305761278642,0.4874247939127665,
 0.12226460992501809,0.6572096981881523,0.9449654446000983,0.8709473046815228,
 0.09111120591903321,0.35552480099053996,0.6536701391638463,
 0.2550918567396341,0.2645554072869837,0.2192973258641273,0.1928023054711443,
 0.49146624740935674,0.3746531568917213,0.22081538992508865,
 0.11740853037435439,0.4081765823585567,0.1521818498926799,0.9050966679783792,
 0.29268146743158807,0.9825343512654535,0.5965284712935555,
 0.38527668786146074,0.21697002067418203,0.5318372772916957,
 0.3560076335557476,0.5276006811862821,0.5019956982274041,0.8517752144863229,
 0.39983817436580527,0.8640201609529329,0.7020083691130482,0.4789618557042252,
 0.4596371367862293,0.4028406985820756,0.5868356994094921,0.7031568845262784,
 0.011136138600123968,0.36119588848238804,0.8068954913240214,
 0.6573092754429389,0.5091846126846231,0.6388740970421103,0.6156253571833368,
 0.3831856683178513,0.5424753614878836,0.7190120168513929,0.3896934939660849,
 0.4276431691140229,0.224100150025202,0.014088500931698111,0.0891382771155258,
 0.32662701271471606,0.693005794801785,0.8018824978548297,0.15674973401726944,
 0.5133058007610987,0.0319278238697166,0.7589272762858064,0.8135460966505955,
 0.9939988012442108,0.8613571399930131,0.8245175334435985,0.5123350192053472,
 0.45518647701981796,0.16420743174840446,0.37402888367510223,
 0.9997793083670979,0.7741327303949187,0.9306383678818408,0.60829237550223,
 0.49650470067134955,0.19342374236416948,0.1244910919504546,
 0.30593520103006944,0.5982966898612332,0.7807433665064185,0.7772980878491572,
 0.9981454320904559,0.925903777601168,0.4523877269621457,0.9902610973272639,
 0.17611840356072472,0.33109455322785963,0.23439828060661838,
 0.32656034077962204,0.5582450886593407,0.5694568460096441,0.8026068405064594,
 0.34162825853533196,0.1611518246722886,0.34106394320391686,
 0.5668897533818386,0.18237801998724967,0.3088965631711601,
 0.04721825835540505,0.5182270054024951,0.5848032822959047,0.2987593850415945,
 0.35060484828526073,0.1314102121798627,0.21778618065867417,
 0.17964495709211759,0.48656717689222917,0.6013127288367305,
 0.5592726134455053,0.7958461951607491,0.37925344689035223,
 0.18947050934686527,0.28333078284909285,0.7452814953518515,
 0.04915583119279343,0.8216280801468759,0.15358841111523025,
 0.5542731105134289,0.4552264857399887,0.6176959927813184,0.7022822933441977,
 0.29096084623013085,0.7400446278628425,0.7013790120444274,
 0.14726518395877652,0.6042540638165774,0.8485366612212807,0.7648666857505692,
 0.19552521514566223,0.09499744938199106,0.8227051242071742,0.655252090250577,
 0.6500869257214197,0.22531377270300756,0.013123790527229562,
 0.2766422920030599,0.6915423673887716,0.15754470008979793,0.4413999636081891,
 0.07238230443016835,0.765275413178506,0.27857553433902427,
 0.33998962418029943,0.1348495531014431,0.010063276665506438,
 0.249419025367149,0.43593441873241257,0.47774976193414553,
 0.19000276634734514,0.20207780122200514,0.6911635625173617,
 0.5118838598601289,0.5092760092415263,0.21381034998306803,0.2396221206301612,
 0.4556286590199212,0.5971995815584721,0.8518235897528211,0.2976102510151166,
 0.662303408031926,0.45383722932741555,0.008463044430217948,
 0.8037056195764448,0.2110234810177063,0.6063961988584581,0.7338453358643257,
 0.6541587041146627,0.2485267667328962,0.22779108673566806,0.77366172943332,
 0.5740751100881372,0.22809537379520461,0.27942984056013775,
 0.6114217771357525,0.4766960593387646,0.656685209527702,0.6709330498114227,
 0.37863075721340156,0.9834238067923662,0.2322484840883806,0.5449117987433647,
 0.654388645901455,0.5761473118226558,0.7969365044391745,0.7503885408225129,
 0.7065981694648164,0.7723207732981046,0.29646691521606483,
 0.17367319138327697,0.27059371117029885,0.6110378585394258,
 0.9062809152953748,0.4616583807125664,0.09033000388796963,0.611079394173969,
 0.8477242823599315,0.8335579448609198,0.0669765223298684,0.7965348514354638,
 0.40950969165908846,0.8371205053666393,0.2647299589585871,0.9655326993220266,
 0.9433460634689552,0.5852963521738468,0.9306896419014887,0.05930618379148711,
 0.18325595879401124,0.7695373882434587,0.9439018816179366,
 0.10805010266234139,0.13159784255453333,0.8237990661792619,
 0.9842314652310624,0.22549293941914572,0.014880006906274446,
 0.27874176625125746,0.6928966807751097,0.291140759372303,0.0856951842637762,
 0.7238809240925427,0.6150323552811545,0.299815143799282,0.4826834447273346,
 0.847574779796156,0.7939573709525009,0.4087778242271264,0.009837648978537006,
 0.8292544343124201,0.5997273736911233,0.7892739503501012,0.24684188391307949,
 0.8144739924696665,0.6455894058316329,0.7025399871426912,0.9231187314852345,
 0.5798168184592458,0.9298476072841144,0.45676805993215686,0.2584632980188647,
 0.8650759982502859,0.3239692003598703,0.7955391123541548,0.41281418545587223,
 0.9339123167033816,0.9263524972871688,0.8861775830429748,0.5753220747340789,
 0.5223911727786146,0.3826350648683805,0.6743839962245813,0.23169587501940503,
 0.775558609589532,0.08736771784228547,0.9636412232337002,0.35307899186576,
 0.9115605494276761,0.4378687667520831,0.017264040495270527,
 0.9576878626913523,0.5110969371498596,0.3069388759542383,0.7961220517237594,
 0.8936388998059075,0.001047901402288498,0.18505309507397572,0.48974469985864,
 0.484643694615129,0.15443472050771212,0.19427468644403578,0.4749172344226461,
 0.9096220905436919,0.03506109024180237,0.5255446734195681,
 0.07400221046571198,0.9921276524387493,1.4198257811826043e-4,
 0.06127230402949546,0.5814908448475021,0.6751887155787756,
 0.49307558754358216,0.5021831827284922,0.06409473305513269,
 0.14623967876944932,0.7030011162216635,0.3252858451428151,0.2172480610850268,
 0.6970152949136031,0.03653341875097704,0.2247392558485789,
 0.43508881193886706,0.5013307550827161,0.8300899002414724,0.6832268372990178,
 0.6949149127723873,0.1714544153339933,0.270654465729689,0.011966114591736798,
 0.3672400997424189,0.7329047904313556,0.8999319897115045,0.5620263710936866,
 0.956446516061809,0.5150639478554686,0.07072233414308338,0.4408270337389143,
 0.9919376867241712,0.14557558374465684,0.6508120969664178,0.8199058839078164,
 0.017084172793917163,0.08669060158079933,0.027300632910967404,
 0.6044396698753566,0.37017977478976116,0.4087394607324524,0.5115239468983179,
 0.7473715454551337,0.3200673756430322,0.619356217163692,0.1660328595756615,
 0.8974390429090761,0.9200600800153037,0.6491564933047524,0.49282399528625087,
 0.42420485096848626,0.8563883877130243,0.026297709677847836,
 0.2696925974383134,0.7005177112271869,0.09138860125559445,
 0.34147789277912044,0.756115893361406,0.05055949405440152,
 0.21106251534222875,0.3744739626652672,0.31682205121444496,
 0.18371610207783118,0.5696964675929976,0.0036428731766767752,
 0.8147320091989946,0.14834238874839145,0.779451297809893,0.250011483307104,
 0.646450088885339,0.8528996436875171,0.6559742430678377,0.74584760640472,
 0.5730418743692021,0.7310723279092289,0.8625299461900517,0.8404149557319779,
 0.4763796117261434,0.2375458005263974,0.6827509087615062,0.43671024004727466,
 0.33044736154756726,0.6131104013910429,0.48262755430381044,
 0.9530313087456191,0.40923671334746503,0.9188192070814752,0.6738292227296994,
 0.14541563397612323,0.12167151005472898,0.45792664465046695,
 0.31134798100259076,0.5676364599006644,0.009453190122756805,
 0.6170143777939174,0.811212278346785,0.7788077776172657,0.7630499913607334,
 0.761774813659918,0.8612486847160365,0.8009443714105817,0.8691539943478711,
 0.3110780642335008,0.7680563377013031,0.9471244202930149,0.3786646058013784,
 0.43640547414597697,0.9788533795990435,0.6789331651886452,0.6201287057595753,
 0.4492316197897527,0.04304003356978425,0.619650317588425,0.42723201433731917,
 0.5283765471219898,0.4652617363908639,0.424890710654362,0.23195911841542605,
 0.56689816425417,0.5864790631176107,0.4257969021114141,0.7098726667400226,
 0.5634429374262758,0.09798913110728946,0.6272834954791919,
 0.44620434060296643,0.44585903451223174,0.21357537574534335,
 0.3180157348469984,0.14282971502345088,0.10122202565593663,
 0.28639013304473726,0.7987798616465344,0.3449779093712313,0.6260324311277595,
 0.9868118516918459,0.2600593828058799,0.5315519168110983,0.04149776970342545,
 0.4882946142197473,0.30744847236422024,0.025572661514647987,
 0.7366636345030226,0.16557756051248684,0.27385633674871845,
 0.1894502999487746,0.5155063994713988,0.5777068183769892,0.9432093801299205,
 0.7398115818550219,0.779927399664823,0.18302889747427087,0.44095936846114203,
 0.6744084052191249,0.1623708464966218,0.36546931647307235,0.5064102094807652,
 0.6351555688404128,0.9762708748205512,0.3136393031041357,0.4203781312955259,
 0.4512478210800167,0.5673833550032749,0.16426231259875956,0.8719667995315147,
 0.9238573208766641,0.1629027966662473,0.13958777973851655,0.7208003705353283,
 0.13872264171450133,0.7891626896301415,0.9445230227835688,0.6351561631141858,
 0.5622077539702932,0.010383613030393146,0.25698384006748587,
 0.45797012280660276,0.19885601403790076,0.3041318209918369,
 0.6352064166440746,0.31257915330290853,0.10883431791843856,
 0.08338170813780688,0.33545711253969646,0.13007524507574242,
 0.8997228008472105,0.033974004920560796,0.3383184321465026,
 0.5347877854443186,0.029222426548040192,0.9005334684924051,
 0.5377190949454939,0.18156303159915255,0.0499649733375408,0.8265427208835623,
 0.5213749455517772,0.4523311661448227,0.5216391536491147,0.6355684262228691,
 0.16145965517543992,0.06416789455681216,0.15260900970213664,
 0.691285391865111,0.266627387281103,0.13983912788596153,0.5951845794742175,
 0.6102406369523939,0.9059658241822742,0.9839564085050206,0.20235360796859725,
 0.8433492010976161,0.9285577380612567,0.30046070335536235,
 0.12636948510130097,0.07920193420313637,0.010461089848374572,
 0.4643262004714709,0.5992845553425106,0.03653195850051505,
 0.42154259040704833,0.23767274368134295,0.8070743933832509,
 0.4088495023404144,0.3472428379468745,0.9505951580228449,0.4425383381419702,
 0.09916923819075263,0.8520638400291096,0.9255430453783954,
 0.02750020170969636,0.13769144712113657,0.3345789143727377,
 0.7326355329981753,0.6380874067880755,0.8871233780605368,0.8628092900986771,
 0.5858032345373199,0.6127913073394011,0.2687915657728537,0.5178203617195221,
 0.6717187882273092,0.4598521991026905,0.8065993889596783,0.6594765742009279,
 0.3094922551071394,0.34408874691428026,0.5450143140683654,0.5810050672894335,
 0.03992464586708466,0.7884407455358828,0.15880853965763064,
 0.5501233713113148,0.7302234882886606,0.2613431976457887,0.7629648765417222,
 0.25030057315748966,0.709304338842521,0.7560705718214915,0.06581609415789924,
 0.18883750652622877,0.00891229865176324,0.27316862701583855,
 0.7597245146344116,0.34583552957321173,0.1456182360262076,0.6474548513450642,
 0.5900098465338605,0.6856393654884839,0.15754031647930677,0.5180828586207634,
 0.9449834476423591,0.4044384036772165,0.6524516499207803,0.7396439523964329,
 0.0025978261685368764,0.18694604201646858,0.6409089017681728,
 0.20477514468094526,0.2170708004161288,0.5462897033548189,0.1662410600211528,
 0.38552648437234593,0.7602991707659614,0.5441090872394513,0.9591083307216789,
 0.8340463000894607,0.8371783052971064,0.6560829032192913,0.5400370512130342,
 0.5166581992747017,0.8274222222276892,0.7442029648017483,0.5799197561004084,
 0.6125082086417354,0.17572441477307343,0.4170783269205106,0.5969389495714584,
 0.532489952725658,0.716069243919326,0.0784999136839789,0.13945141441798548,
 0.41964326810895713,0.43648961349611604,0.7067439488125904,
 0.03901468213368353,0.2289821256024397,0.37620217330721595,
 0.9248459453158502,0.8428835295551953,0.2515718681875335,0.529842806767269,
 0.26082781755972595,0.6012636279050798,0.1893186591380862,
 0.43901047917927605,0.6985294304412126,0.9164565754961013,
 0.27652233290951134,0.41089866392497765,0.36979351241928093,
 0.36251621780671517,0.9421675051292694,0.6331460209068771,0.9865114038201764,
 0.8732024085937781,0.10341056612160537,0.29008556137033137,
 0.29867314950569124,0.3300892603257536,0.7401797107570768,0.3562434111936824,
 0.6252460599037949,0.37677345825311215,0.7659037834991274,0.4550464811658057,
 0.5155834691660748,0.6303517582019601,0.12677006299311677,0.3827452523298903,
 0.7748605215791075,0.6710161060224035,0.7103845920724418,0.953445138709113,
 0.44006376370284495,0.6143879986716166,0.9687647085561905,0.8254562322752775,
 0.39751427055597144,0.5948571428721621,0.46889061673278265,
 0.9575409691764583,0.7225823111483981,0.4730914345309848,0.40938056202608974,
 0.21492204153011973,0.09701742408749525,0.22191891203159642,
 0.021353923238709438,0.5547272817007908,0.5559003045487446,
 0.3591429349324837,0.48568714886211106,0.8918178723997108,0.5982621121127777,
 0.518317886724281,0.4453862870101295,0.4210427621648878,0.7252870613635448,
 0.32011208686274206,0.6277013138894457,0.5954602967796507,
 0.046569648969238875,0.9244594142849542,0.5454091468314627,
 0.23904946362457946,0.26040013719679866,0.33964840666945506,
 0.09840573941829833,0.004132341308818899,0.49522095921872444,
 0.26776720835049006,0.37548680538999224,0.7814115214789753,
 0.6368307358844898,0.8714801817100867,0.43384404179119374,0.9252748909971303,
 0.38428945255492897,0.5743658814173133,0.1636002872493334,0.3707783349842033,
 0.19333681063190922,0.9616113918099385,0.4144014623488237,
 0.42832191928882635,0.5338701918543606,0.10865436846711718,
 0.7044957110566565,0.6139748244448267,0.10602578359674952,0.3578525273828004,
 0.057813562457894596,0.04233683172955738,0.48001374020606147,
 0.6095181515895193,0.4337802222042262,0.591181682446947,0.15728696308348555,
 0.9156455641480558,0.684604912000293,0.6994810831918711,0.20531603751399763,
 0.057101396050850406,0.47337143483363786,0.7985500265552725,
 0.40858269256778235,0.5440772989547257,0.832554520740254,
 0.004191557463071649,0.6767061584635476,0.10115781628165499,
 0.5980818566918342,0.813463239960748,0.14563472088949514,0.9951344287281453,
 0.8348392551181725,0.5638915572869545,0.4411811315795998,0.6643996220208764,
 0.1288700997331632,0.9440014897862943,0.16594221725219838,
 0.08213669879563767,0.27800343438067254,0.1919811063392164,0.820204889079434,
 0.9292895973713413,0.7114705564194116,0.09279680563642145,0.9545809937863796,
 0.3632613678533363,0.6280219788813687,0.692802361479391,0.7825600859171746,
 0.5096259009614786,0.3095427104450138,0.6697113894087857,0.14474998875086376,
 0.09686507463677585,0.7734030051659161,0.511683048371888,0.9929096690620529,
 0.6879217052487149,0.1428613305467611,0.47006115044369845,0.4381126385099269,
 0.09066254988477995,0.974744241338343,0.9501735865341081,0.22456709829583177,
 0.12719390740282543,0.9634678315427605,0.504846317080873,0.9210512482810608,
 0.20366198541067793,0.20514484662424382,0.0730982111924614,
 0.02487291136896097,0.7624472308517657,0.24992132288181024,
 0.31253286649335255,0.6720390857760479,0.6695495286067803,0.7700085954640419,
 0.07033811052142669,0.4625581748409824,0.20665006256999063,
 0.7429164122032283,0.49307746044595846,0.1569814317245336,0.544143650841161,
 0.3590284899406432,0.18845850403413844,0.5913833699207519,0.9381835994901094,
 0.3091717605126876,0.7766259323405315,0.3510196128975198,0.6897879747058697,
 0.1423753283694832,0.16251743882634817,0.6399575225393508,0.9719900814599931,
 0.1945424431597348,0.9773845405991598,0.6545049081015265,0.44344234528741544,
 0.44500220011603386,0.1436256188535714,0.4986908456491781,
 0.30418682842999223,0.9562464624307961,0.3376048311239669,0.0847472582515032,
 0.8956156621187048,0.9168279418054454,0.8268009190080589,0.8000779751548674,
 0.07316835400194,0.36516830919325693,0.13866301832820926,0.3715954634612426,
 0.5447943041921468,0.7926929733773525,0.11778275171140828,
 0.10306346784665188,0.7350652838128152,0.7179615633026455,
 0.45966782495654124,0.15330745443233063,0.5798157428281699,
 0.43308660992563786,0.8649023793946993,0.7032218323438048,0.931866451066799,
 0.3475207209693306,0.11556392980678687,0.44315859558827175,
 0.8901905202870796,0.0466095202690201,0.8411748635582105,0.7424388720598234,
 0.2916946733345458,0.12014557022777472,0.07182774830125838,
 0.5400179176436519,0.24753102233426205,0.8050570788965008,
 0.22697212352450635,0.40153639143246145,0.570350096205554,
 0.11768670579142737,0.687640200390807,0.41283460335797006,0.295596200467091,
 0.5245997098310985,0.14719633583567993,0.3231593384433624,0.6324291667509759,
 0.34985468173880563,0.7392682916747315,0.11137678513276117,
 0.8379983382855128,0.6087136618356375,0.9101212663535909,0.20619233196869557,
 0.4504283934486051,0.4641360588789677,0.7117152279648146,0.13392579199913546,
 0.464363580111432,0.8652484124561574,0.8800421531707473,0.03202411959860274,
 0.5998391826796091,0.1259880684055712,0.8194427098606096,0.45355645402603684,
 0.5570869864727593,0.45622044275475615,0.9891675099055421,0.9630836214909325,
 0.5750636619768992,0.5985192142910892,0.3712062219034553,0.8063547869478607,
 0.34517573632539156,0.9477921991767019,0.4489577398270512,
 0.23905167613088096,0.16517443582890923,0.8229330896527783,
 0.7135950499298375,0.38918887396160473,0.9659952011090318,0.805429502880863,
 0.5282156435627265,0.12160428438691884,0.7535273655467642,0.8489154521234031,
 0.6703963306606251,0.7676770139728556,0.47161444658154483,
 0.48214208218020727,0.5017883212804041,0.9785393146783019,0.959776595181699,
 0.6588012936624275,0.8904990264671971,0.7916424947509824,0.3463493056700515,
 0.3292657987396692,0.20707976959921748,0.9236418986697195,
 0.12839627870606274,0.3108627196654291,0.49758438051399345,
 0.01979032502989242,0.5998802463644701,0.6563809182794687,0.8970086521118816,
 0.5167451424245302,0.017846881973625717,0.40176219559340565,
 0.850404342985559,0.954836442102687,0.8834908948909599,0.04881712992675458,
 0.0783365312993618,0.6839126742999747,0.8185529662894749,0.8429300547758033,
 0.6988527310881906,0.993346683177157,0.8126750745182129,0.33631796197729724,
 0.20117566837368583,0.9608463121060445,0.6185815594394193,0.9044353774179967,
 0.8686246410105353,0.2749733084510745,0.3127340037082772,0.5717817967472896,
 0.15096680254076178,0.9193963126966196,0.8415306187053054,0.9209456192354328,
 0.484900619619826,0.026900244925500694,0.5882613812857189,
 0.12499864528882443,0.23324094801497086,0.3316076521104653,
 0.6326463754227775,0.2842020352078898,0.8761319548344463,0.5571851677204276,
 0.8572291918258585,0.3805479016766632,0.3602559992585955,0.2507825593870481,
 0.3236658951622141,0.6300309915292817,0.4692631273235066,0.36302275723521493,
 0.045568370748239606,0.591071555002396,0.5053886841541809,
 0.07788736693386011,0.2021429251393314,0.4252159160622704,
 0.18292047693349522,0.5894235979207602,0.517562696047535,0.5325920396162749,
 0.01812921532476197,0.7856639642204918,0.8577567469074716,0.7819079424993108,
 0.43409367138086175,0.14109283650609716,0.08510826299797947,
 0.6419879112783757,0.7017892165473792,0.9543785505730263,0.47052068852463647,
 0.013572149098702302,0.34557249165660064,0.021621780533398804,
 0.06670196435371245,0.8415202463668336,8.322571369550058e-4,
 0.7388972781883674,0.7302404418038194,0.8590922442670762,0.19313943806571254,
 0.8208386802634998,0.5990438463958571,0.6011074851337375,0.4029329808176131,
 0.6537195070566609,0.49793694893402807,0.9157943404869491,0.2519118082587487,
 0.9088106246708696,0.5249248727947715,0.5384377672802177,0.14747806998873325,
 0.22412487124411884,0.5915681844312202,0.6167950363478373,0.773528571335053,
 0.5478592651088792,0.7818619564241027,0.10653158478502966,0.610254452515194,
 0.3756924401229482,0.04266499878401364,0.7665189453549719,
 0.05445157479356433,0.5267432284999679,0.20991474917457875,
 0.13853763456551682,0.037698729302897416,0.5502847176832906,
 0.4820883774397222,0.4347118244626359,0.4984718332737258,0.5329020156950146,
 0.7967966031996581,0.9471120909419264,0.7332040582604489,0.28204241014245,
 0.20687116911896042,0.4797422551317341,0.8848101343792403,0.969753302502243,
 0.31771834269904375,0.9750735675867758,0.4498956190276515,0.9619342872890294,
 0.5921797243973037,0.4070402115629477,0.9611005527711594,0.21465981724934524,
 0.8139177422324106,0.4342842310074363,0.5307916938459264,0.6557565430599674,
 0.3189766183134086,0.7524299065641097,0.41789860517358945,0.5058219401482298,
 0.6792495144217167,0.147201790607555,0.9328455632343863,0.3367926379820043,
 0.3738100602886454,0.7454623622812542,0.05179762579646896,0.379131034214121,
 0.45513130516763034,0.8606835565843968,0.5657195013057755,0.5175486219775791,
 0.7800688022857771,0.9566523506513511,0.6489325634426113,0.8201175977370931,
 0.49243762516618095,0.691738957853349,0.21589274131918423,0.900956470945682,
 0.32196338215624487,0.0636136063574636,0.8811309044189868,0.570592290079735,
 0.5199097505478378,0.5279200896145483,0.7577989279875725,0.4984910105538205,
 0.08515022270587003,0.4589031748788077,0.5607859337998196,0.5242887157243853,
 0.22910225475465262,0.031073534563895167,0.8201855562163473,
 0.6032921490245059,0.8150261343236624,0.7020399909238337,0.5047339229439268,
 0.5039458857486213,0.8069700765107936,0.9685159800078381,
 0.061571168679316224,0.43714941644040073,0.7179512653802286,
 0.11305246623907539,0.4136627709178964,0.7445087498848801,
 0.33788019343758113,0.272312140343275,0.5325609865935024,0.6441235011239379,
 0.8219676691232171,0.766019042051869,0.49449771441569434,0.4172956293893191,
 0.14934226869353662,0.3065063341398625,0.9185066455533611,
 0.26585944776233095,0.9786655563554818,0.44170664359511347,0.485163309651619,
 0.8276694344523989,0.21983935518336917,0.1363954686453901,0.7840334779172037,
 0.6756096362997738,0.17370545189417919,0.5306702970759117,0.816250389329835,
 0.0419782191967174,0.9629793229499783,0.15020155810103697,0.5944199136536064,
 0.5430569829519613,0.4895969363713174,0.2939234844576295,0.27851218598756555,
 0.3197528868912911,0.1189739546022941,0.9673326913854237,0.9508765935480865,
 0.5862110884350015,0.990381863092888,0.24255516193328497,0.9971136386527073,
 0.0991619101684802,0.7511025123475316,0.10742700217794887,
 0.16446198891311392,0.4023090159096271,0.011625149028106185,
 0.6482760649631856,0.6469273130934258,0.9318352157538081,0.3157163209681383,
 0.9397757438095793,0.56105527192543,0.7473023660157223,0.6040068911894207,
 0.6982251938454448,0.198987132043835,0.5827439869334784,0.06416000583263592,
 0.5667171477441638,0.3675025454684441,0.6264843673398829,0.4785273710834128,
 0.6059867973658483,0.4170016219011664,0.48676472110216007,0.5361199389716361,
 0.19998466162996364,0.59806195696758,0.6872800704276564,0.9612750907740999,
 0.7675800731205111,0.13622859623423578,0.5125043082952236,0.8996181400825627,
 0.2749292118498179,0.4382429455567446,0.23602946045368978,
 0.041477992498238136,0.6856373328540624,0.6185918663866368,
 0.9113092508178389,0.8979579080280853,0.02963515792086624,0.3841706636311951,
 0.7411545887423533,0.0487634174347793,0.6820701404213425,0.4851248441357876,
 0.9953357163793715,0.5233177371943789,0.34967396499944636,
 0.01532533950808801,0.7396992703267979,0.021993100375342634,
 0.28172308889369024,0.17716756517469856,0.571586483732353,0.5469819766831,
 0.0665166553591604,0.37795735286472,0.2816440199760828,0.4285125409135986,
 0.9352631464581275,0.974553556799963,0.0933227607699032,0.9047746075539269,
 0.9599564237457806,0.6287196560973005,0.6824676004344906,0.6294173581036514,
 0.13807107830231302,0.5764926764385954,0.45287810493733094,
 0.6160890551151949,0.3713086489193813,0.5116692516112398,0.6522940960675226,
 0.1673991246804255,0.839318366374017,0.9242557546062337,0.8575290800549842,
 0.2605423250634592,0.31930313368245544,0.2075668983369272,0.8337582408341505,
 0.6717051834500671,0.8929301498071932,0.10040636880649478,0.6311578444946162,
 0.6200162373754453,0.3546420011004624,0.6833424515112635,0.2201136768474612,
 0.49369863755690724,0.5459422974320078,0.02730092827901709,0.9191603369018,
 0.03903577942711067,0.32568776248452014,0.013911156503479294,
 0.30951122551828214,0.8142682562414776,0.7533405435951808,0.6122081941552037,
 0.0032000519434370878,0.21795324964051876,0.25536285883321186,
 0.6995680913250741,0.0030949914900723074,0.4171000232076725,
 0.5597393796418523,0.8709011502997932,0.3087377427400967,0.3487939325375483,
 0.20459982742326488,0.302165643442603,0.6589680380164011,0.9679069607718306,
 0.6031850812650655,0.208394214829415,0.5852310441487726,0.5712680696487307,
 0.5568003988551609,0.05568856675754097,0.3418909060351165,0.5368084159503426,
 0.9246469768039633,0.2237375400891417,0.6614846538001563,0.11025510210865108,
 0.1431257969019991,0.5780514302151077,0.3762495809545313,0.27297758242120895,
 0.42407366147086956,0.8002146747467583,0.22462817010584846,
 0.9045913475797401,0.9155674966234715,0.7690085489910433,0.22630797818595816,
 0.4150760255196734,0.6553704813999046,0.05987456321859397,0.5030540585473435,
 0.320545208550869,0.7041219642275534,0.9693056316262595,0.46751071885436546,
 0.6735644780018066,0.9095382352942362,0.840441983303918,0.817771485064154,
 0.2770597981654863,0.8336898801303082,0.047493797384067116,
 0.3861722004753537,0.4333704408924127,0.4022720045200978,0.9297216784064558,
 0.5291078621344794,0.4843293077039469,0.3512671206786314,0.25460567438764314,
 0.3917047461846349,0.9585337605083457,0.5830495409547015,0.752984582930148,
 0.8035698330625136,0.4917969497802679,0.5597568764296038,0.5384240213880116,
 0.2840237164625511,0.782378462377896,0.6706917114590237,0.8141492438430409,
 0.1596331070497088,0.16591728518370563,0.7675768620694517,0.9113254032443266,
 0.7402491171944733,0.33312478629499853,0.26284710544620227,
 0.1333119139675104,0.7091868945290856,0.2155946639626417,0.05560517702615875,
 0.14068834769831995,0.5349605397305555,0.8525859979192209,
 0.07536282523707694,0.4393305326985335,0.8966615189047125,0.557365248110291,
 0.44297286278400916,0.9227420550592726,0.2736385914775116,
 0.40695833663779113,0.28599959284774357,0.3097742181729807,
 0.7494337433192129,0.0314059933117985,0.9106250580826618,0.5935833890377047,
 0.6250374894669701,0.9137202146123647,0.04267917278194222,0.6719290570447103,
 0.4891106397522649,0.06507440172909051,0.249621704920858,0.7137656836540898,
 0.7637472056441376,0.8367515732072881,0.603291594513706,0.6671794923484128,
 0.8189786396198824,0.11447884147457943,0.870260812574555,0.6350996960271436,
 0.7441678527235239,0.7459219382196066,0.5986961075040348,0.13563694405049986,
 0.90381188001733,0.37024637463176124,0.06765730511866463,0.40932849297943175,
 0.04991588219114029,0.8757744170509932,0.555547706737731,0.9773071328905186,
 0.6590251763976889,0.1957624801409954,0.03927599296742457,
 0.35292033223434205,0.034527653729281904,0.8071276131484211,
 0.8757633721226846,0.46941181332076054,0.8122964319473535,0.7911907756910442,
 0.7181784214966221,0.09131899644102626,0.555834930979863,0.15606450608861255,
 0.6593055239220085,0.8199078197902502,0.721039137076414,0.6301545349345812,
 0.2710906106901305,0.6643084978444392,0.7056404527026268,0.0055235527819959,
 0.2880390275689493,0.7862525862143749,0.0989623745628152,0.5300898862448513,
 0.3700356594778722,0.06003586044216791,0.4655028590807818,0.3613647981163025,
 0.10514255646659942,0.7816480682098863,0.9968686812497048,
 0.08110123343081677,0.29883287147931703,0.25667853523185713,
 0.18966616891675225,0.5967448725490824,0.20718169607217118,
 0.9407709959438347,0.29665739888887765,0.5365587336904412,0.9154914449401989,
 0.34916626216560487,0.5743535676793413,0.44729370558785086,
 0.3753058458866865,0.14660232981092292,0.9983730101158982,
 0.42321454211055853,0.9667364872999354,0.3132643687945197,0.5690969405553998,
 0.05383894718938931,0.5133106914566775,0.717258490765092,0.4915376264475664,
 0.5025932353546428,0.3925276864598417,0.6070432755617331,0.1436788706127443,
 0.012127039521716898,0.7564796636496093,0.5800471348337163,
 0.3516815519176266,0.3569220411758931,0.5921319539842285,0.24729982509610682,
 0.07214239781692333,0.3243086045053911,0.39780216079632136,
 0.49372748016169554,0.8503607402335411,0.006096432812584096,
 0.3144242903827106,0.7390091451442878,0.6886648128179209,0.8329977354371961,
 0.25886680141252594,0.12814296554162752,0.1434617220703598,
 0.5863987768468568,0.8903334115535347,0.4013082090327168,0.4760267648377614,
 0.957053621802624,0.7906095596175591,0.43016898185330366,0.500715705982567,
 0.7281603855968823,0.649912190758127,0.5920048836893902,0.11258343524382108,
 0.6793601171353079,0.09132352177408254,0.11760002461847985,
 0.7820521540368661,0.6620485803001672,0.14831529944700272,0.4169861066834334,
 0.8746379097705164,0.41141446980162255,0.03265301712790758,
 0.24086655040622107,0.2410102528199256,0.3080754319341664,0.6216070840841562,
 0.2413238982565552,0.91972883047511,0.9513538420024703,0.02903093978687743,
 0.6398150404994161,0.948063759089584,0.2769512842734252,0.8784873595196031,
 0.5443738771222038,0.6649263464650313,0.23745294870877964,
 0.11682160207542203,0.457567725464853,0.2218364973291811,0.6581337237264792,
 0.39236966549006036,0.5670791891611418,0.6875925956422805,0.5462604490846148,
 0.7123814757021855,0.4730299621616455,0.25411246813043253,0.658110677907117,
 0.6104301475579195,0.9738334354371779,0.07798046755081145,0.8724319406417134,
 0.20439032688380177,0.3082808000019136,0.8321382898872536,0.6359435295167126,
 0.4721478414144251,0.5278688718848082,0.03453834960276392,0.6517797069028674,
 0.46866983043977517,0.9165508992030389,0.2425064411240705,0.9086904692704909,
 0.35340190135875016,0.13397575816531793,0.48914977604421006,0.55402198766859,
 0.9291010124692947,0.7224858508334073,0.8087668784329162,0.3392294228383623,
 0.9805571981379655,0.5998927931442968,0.5359804370204709,0.06228159109171172,
 0.6291183729171662,0.2868235834644718,0.9376344608916446,0.7145797151393917,
 0.41172925358934087,0.8371100411729379,0.5538876776961776,
 0.44672947849483613,0.04296304535975304,0.2416665772751979,
 0.4250872728589734,0.30632400555112516,0.4236389549547075,0.7422419961481536,
 0.36012473967554226,0.4253819098541469,0.013341407365252245,
 0.764927588011776,0.1969948198606033,0.7031088134472128,0.03198146837455529,
 0.8430611362491587,0.31026729695534905,0.6341904613834684,0.5377401485170786,
 0.1383112525427238,0.6262527735231957,0.7026387114890545,0.9610219863979553,
 0.9018989791394821,0.5275352423451947,0.584372884373544,0.9507802601837794,
 0.6906765446378852,0.1076256103379879,0.8388634272330125,0.44268971089903786,
 0.17735938174047816,0.4324759228306878,0.8709148913129405,0.4494564939705288,
 0.6833724758875261,0.35300013254018814,0.08739091197669435,
 0.03495112408848655,0.34617007187360127,0.06913508413123937,0.88713483900759,
 0.9515902379428018,0.37799710439113365,0.46127729463743994,
 0.4827096742338194,0.32756357626055965,0.3603389323372207,0.9693630408487192,
 0.16898819578600255,0.8738593811815752,0.1150279616988592,0.4714766339805039,
 0.43075638134676675,0.22843576802234722,0.12686323766253094,
 0.5875908475574139,0.024952737531976132,0.11275835074518825,
 0.10255399648959507,0.3074039162388349,0.7260574534377593,0.5668320226104551,
 0.9279767839978087,0.19295868284391926,0.21686110980618023,
 0.8342617197968328,0.30040379246548343,0.9793338089707311,
 0.38113881275022155,0.2270770175682213,0.6489904400523612,0.6157191663813193,
 0.059693165342972865,0.3889294212435964,0.01679350320241646,
 0.39940192914382044,0.25744602011126827,0.8877770123133168,
 0.04619538106901577,0.05543923775786663,0.9011797750452928,
 0.19678192317880083,0.6857762411525476,0.1900242376979352,
 0.017723234930607346,0.03776516203979596,0.46116548745076047,
 0.43772550411217714,0.04345064151023248,0.11886481021000883,
 0.7621430944898125,0.03953340930177207,0.6098616053916911,0.8512981209315665,
 0.7681934568031478,0.4802988067893793,0.9781430672853997,0.8294889219577086,
 0.8790052826907541,0.1924995347469899,0.155497771624574,0.21595049174586112,
 0.19294338074231288,0.9295245842666313,0.3057380752740033,0.8338512190855876,
 0.07747077098588773,0.45475558722345666,0.5373581909257645,0.264721277047657,
 0.42184539384341935,0.3813009339177931,0.41781484018313664,
 0.015948097313235765,0.5246038283091653,0.5040292698760364,
 0.8086438734822197,0.9036236660576519,0.7688597731136593,0.38734993557451425,
 0.01504175056611734,0.8939050779637121,0.9900866504794487,
 0.08758592884742844,0.4342787488155119,0.7927330450137727,0.4932233040385997,
 0.9236042561362499,0.285822551904825,0.8883471796295874,0.42971483655100484,
 0.045292949390350334,0.1532648586388462,0.9146803288699519,
 0.6725556825282011,0.727244313264516,0.4644961906456827,0.2625176089040444,
 0.5498175045427971,0.4364984922614902,0.11000313139347795,0.3079993649463847,
 0.4080800575074488,0.7639717305755354,0.022120389476223856,
 0.7977219260188768,0.4590859356050131,0.08824646877874798,0.4639048318725878,
 0.8928282499989302,0.7632923769797236,0.9348756144398713,0.7829869516737734,
 0.38843621762078406,0.714129018804039,0.29536768886900333,
 0.03038488282868579,0.8356815599477102,0.9181976620132714,0.5350879150517666,
 0.9302179889832238,0.058417314835933265,0.48263601051376204,
 0.10320181889465485,0.5411619253280081,0.22706909445544454,
 0.4375517229780901,0.689180032609382,0.23715907167493433,0.553769995349655,
 0.3819064594045726,0.6837069007166608,0.3292534394714357,0.8093018481901036,
 0.7843102211078938,0.5734752601081373,0.6603906247958948,0.6884912003264748,
 0.5237384738819921,0.44182709833520695,0.9472833830164809,0.4064307319814794,
 0.29095179259692117,0.37714350985507616,0.3848962726389238,
 0.9397203109269228,0.12597583937304413,0.8567173400683911,
 0.18649478939022257,0.040215411596836725,0.27515921888840145,
 0.6500665459312777,0.9362762800806749,0.1102867097700243,0.31510185179263017,
 0.3991181424315958,0.8049828940956623,0.13398673427217478,0.7921564315123579,
 0.1662124899577393,0.18656661104319106,0.030892038367976826,
 0.9945142490191631,0.20238023603617106,0.6197269384444783,0.8371058790192241,
 0.4200439229636824,0.7058920127063902,0.5258260010851092,0.5249992293726692,
 0.4314356528599739,0.4852672808583606,0.815948282257764,0.042957577037714634,
 0.5251419933175023,0.9382217094156189,0.7678262605515824,0.08892518943806604,
 0.6012613335202662,0.5624091461273176,0.9718906586910281,0.26251780413188164,
 0.5957865962550571,0.8956897971826745,0.33256660477515787,0.5279508211849147,
 0.2545177740617772,0.42307102055503154,0.3165035668850915,0.4446205062311628,
 0.5065166145390848,0.3990898885425067,0.6881333607180579,0.3826002670090176,
 0.76662162840267,0.948219440522391,0.12493708291299421,0.9033991859138749,
 0.5217117904586107,0.24630644633070187,0.8786144620192471,0.6151120809924854,
 0.6789440501956944,0.7837487010229539,0.6226909400492225,0.10781420150811694,
 0.875440468464738,0.08241858762094345,0.36201973197059556,0.9404195439897379,
 0.6997215450418536,0.8827796224964084,0.965554918879429,0.9925916873968624,
 0.3821763228433278,0.5628577215527386,0.964787388507761,0.897428259790614,
 0.13308147861029573,0.7425097387393916,0.7042896652414651,0.3768347077169423,
 0.12466987436479948,0.1578599625757099,0.7705394774241523,
 0.09324906363866903,0.6133985774503629,0.7523236496883609,
 0.34631814408483275,0.8936086994427989,0.8354239882859275,0.2500664599769393,
 0.5286273134917345,0.8725475034825791,0.7832084061215804,0.39305590758349773,
 0.6060114702309688,0.6978754241086159,0.8582451683057234,0.4887374176051612,
 0.7984027651922349,0.43273118188109694,0.5085049462839593,
 0.22874992617251977,0.3959980444627821,0.9953866111669237,
 0.12155105463951865,0.3395143838670599,0.2774331027058826,
 0.27506368749190047,0.1945240443207018,0.9560219098187647,
 0.24104172966661508,0.18605257299009115,0.8535805006668643,
 0.027638208652117302,0.923731555774759,0.749778071981249,0.7435185029775369,
 0.4843118423653896,0.7432037020469077,0.7689983198445781,0.8439710370123156,
 0.9914623560465384,0.8736926204313064,0.3747260280064999,0.47100373571553167,
 0.3583366902141577,0.12279360387749982,0.8183700699063086,0.8349312088041992,
 0.07594095765433106,0.815872642886843,0.17689851649029953,0.8804691414084935,
 0.2534641055688217,0.37013491965010525,0.8052390276540843,
 0.28893613267094653,0.27780039093739184,0.118499415782364,0.6045024677897208,
 0.6070773390036754,0.6058178403156402,0.013335162238880982,
 0.7787274233009971,0.7007444079313078,0.3156135881754162,0.7803890479813306,
 0.20712861594800136,0.7706761307892629,0.2570391896191754,
 0.012709914398159383,0.24713004480109513,0.9891479564624283,
 0.8647710198587465,0.05264299449159382,0.6515863436843606,
 0.29867219345918516,0.07973292245116015,0.2556683225448002,
 0.8764877026140406,0.4947019150773114,0.8322308382745808,0.7774280525024396,
 0.3086288379564508,0.8711758266317488,0.03619317317152537,0.4595938394344685,
 0.07837771643500835,0.007890965648184367,0.018341036790102017,
 0.9482889283681969,0.838510875064528,0.5008744697649513,0.0540854960214594,
 0.8000074768127401,0.054580015107906554,0.21842522570071132,
 0.5145621598958835,0.09228282277084021,0.7281969585168139,
 0.23907121545183774,0.5984511036288666,0.04990114741124618,
 0.3916473861014865,0.5848848215906643,0.04606127319602138,0.8149565899632918,
 0.8365271780308301,0.6550893668844425,0.41937084484844245,
 0.29378855596546594,0.9064962324831667,0.5252958529233943,0.4630507308267343,
 0.9415271828469093,0.46632042984106203,0.4465938826920909,
 0.11826010544406096,0.6362716667836107,0.8882650910438387,0.5935104216658739,
 0.7945698807500515,0.8713262992470344,0.12511340521980008,0.3737950129320252,
 0.39392872388669153,0.6739333188422523,0.002430444956127098,
 0.7997587796338242,0.5815976375266896,0.8117989473357248,0.6048288079109123,
 0.014366006306069323,0.18483650406994034,0.5424632350728543,
 0.34851505064514443,0.9067820765861925,0.010726011326883667,
 0.5780259422802281,0.4654742909180487,0.1671962675961277,0.4003950684126979,
 0.8133542701913223,0.15272360067186652,0.4766847371071504,
 0.015318066696006771,0.9406120792623218,0.8428113643980537,
 0.33510599281518705,0.1814817115566264,0.7219543818380428,0.794873567523859,
 0.843321943124742,0.6280016913116285,0.9601643003130409,0.43025808081032346,
 0.6835027183465738,0.44276203361277,0.01641465296688749,0.21329416261925882,
 0.18444342405021374,0.14151818413197237,0.297420105841006,0.5396831666993641,
 0.5745715837792336,0.9298123141922955,0.13927821820935204,0.2572477143135684,
 0.2077524217082164,0.014833896799737123,0.2611133276122344,
 0.1333950980181705,0.7737439098313029,0.6247308806574682,0.7025749719952694,
 0.5781172474330931,0.9202596624690124,0.4766395580016094,0.9198132287460972,
 0.0826785623488433,0.698837711364442,0.21648453371797782,0.943527977988842,
 0.11729433375482778,0.3588487436766833,0.14984890265647433,0.517273405788179,
 0.010127324119911796,0.5324961144483921,0.13317307138469325,
 0.08630525687076807,0.9391589337607071,0.5043269700447108,
 0.08169656995327501,0.9878355645319049,0.17731488587741062,
 0.09156166651218411,0.27307063275773313,0.44072895108820276,
 0.4504292870455864,0.2097085555008239,0.6821102549697291,0.5249053625876696,
 0.21815935670002362,0.6974361951969135,0.8170030263089074,
 0.13536497546711845,0.9058411423882633,0.621360241361307,0.9113539165164142,
 0.05564962050316258,0.8759948765517199,0.379034692268305,0.24062652586155153,
 0.6981995631907236,0.7160111937503995,0.20122591674291668,0.0966851159103641,
 0.8325269613299311,0.3692554538339292,0.1652870181156878,0.6815595260767012,
 0.6744744920605741,0.7451875756195079,0.6840403214339931,0.7152625039577875,
 0.5505565303084435,0.0023986011059352608,0.6640879806045011,
 0.7177648382771157,0.0373934689317037,0.25174791792546214,0.5920683714042514,
 0.9685299136995029,0.5984052430981087,0.050075732672718254,
 0.31395471297829014,0.025821905490239883,0.24919325523990232,
 0.06754063689677225,0.04922684926824439,0.3084936339957892,
 0.6644044101395562,0.42854009046567587,0.6706727317477734,
 0.34214280121559415,0.9559986599768654,0.8584929692095812,
 0.39089594406227657,0.9274073117094055,0.3653584780900616,0.6447376130626956,
 0.8985031624089496,0.3018232409021364,0.601582627881765,0.9258068634911297,
 0.2064820371116417,0.4149704440689459,0.08306215013657758,0.9613971573964978,
 0.5682700122431916,0.2196883672036054,0.2376725939094192,0.26749780136541734,
 0.8694309122093878,0.01283597189964969,0.3641334695114562,
 0.36576773195244705,0.6367860911541426,0.5977005286675923,0.3970126572179228,
 0.9835761957717348,0.19867866252841282,0.1361607077061654,
 0.39264845262175985,0.393365523125367,0.06678733179382002,0.3084920167697677,
 0.5940203461966017,0.3531082581917989,0.58316860107004,0.9889859023493784,
 0.9392044751466533,0.8406606549748539,0.5176499175556608,0.49106637682054277,
 0.15551932709917282,0.26235947551442707,0.7690170735236046,
 0.5034318364231014,0.5444740251519573,0.8626341210629872,0.43882120184883733,
 0.1743520340261947,0.4193366764549131,0.9653403842803497,0.4159162206778927,
 0.6549558713112033,0.15483993580877364,0.21557756438326814,
 0.7771697204351544,0.7516313081210232,0.441858021930562,0.2469355676605074,
 0.1002876633784685,0.09009826050099679,0.15645024084474934,
 0.009100512546169393,0.5646341718645631,0.2702250260599508,
 0.5926483108962781,0.7251133174960053,0.810555901045214,0.6854785204878778,
 0.6574710101526473,0.0037445044934240057,0.12725760002789066,
 0.6008852532725775,0.947374608802591,0.5903846035006769,0.5918837820391939,
 0.7769599972262077,0.039923089956691915,0.7338940803008103,
 0.7891091766996173,0.4362714952631084,0.040572476989700235,
 0.13759307123441955,0.7021607232336456,0.7499588843225079,0.5181877060166172,
 0.753624178087949,0.5823429153577396,0.9422598508765071,0.45810351709141184,
 0.8465052136524389,0.9982565625643728,0.754767125013592,0.999838047047595,
 0.5575511486820162,0.1323872777367482,0.05651538117093369,0.4273487672869064,
 0.5140064803745558,0.7553262028170336,0.07931441536262063,0.4960913540984274,
 0.7461218690559737,0.37302022111394073,0.67689358798721,0.5059247382961273,
 0.7588977378493664,0.20617292116245944,0.10758377884095938,
 0.5953971480548849,0.9525950221715582,0.5922937993190328,0.771461245802424,
 0.8802601597837003,0.1639728585704947,0.809342956752979,0.7154795591152522,
 0.4481843247527515,0.3281793067140064,0.27849681561818795,
 0.035915132280719186,0.6589055601274363,0.23397026473428384,
 0.3612294059269101,0.37531460039012465,0.8789622160942465,0.3294316821742612,
 0.5684972819030859,0.9139101830929277,0.027452156328078292,
 0.01570528593265219,0.8478204761752073,0.18584996944512677,
 0.3959204508270968,0.24429144503386435,0.30128759889911205,
 0.10394364355807129,0.4113792239180888,0.43447638598176064,0.135664041419565,
 0.9627696138894857,0.8557502330805742,0.04930891585690578,
 0.04215876410317221,0.6940553420583302,0.45814744622582526,
 0.20822127315020145,0.2642315727264619,0.6247146928927425,0.7417119877729627,
 0.6166291463269638,0.5715421313120063,0.6256495436261078,0.37260037348222497,
 0.49963774747538936,0.26670755780898514,0.7028730268777047,
 0.13450933170548063,0.4938680637230134,0.5151741913333919,0.3061104835164776,
 0.18870100487352393,0.6032471339086783,0.9247589766264804,0.8367908438045488,
 0.5490005586704928,0.9567138962668017,0.3267177829896377,0.22386787269083375,
 0.3445055638716671,0.29938289524861805,0.714345966808119,0.2114414227635817,
 0.20397339437593698,0.7368208712120043,0.9738147086825484,0.4671718321966891,
 0.11229380760339369,0.5725742022335668,0.3199770481893669,0.6061417078561615,
 0.4349662721597872,0.9413754514802767,0.6353024080317689,0.11191968434700283,
 0.565775555825712,0.8568813201393171,0.6362507543828977,0.7168975756081489,
 0.7231145827286533,0.33497122225959963,0.1541909788605884,0.0765830024576103,
 0.7484138795893964,0.4575173761397936,0.08762527558578226,0.2559073479664189,
 0.9447305428210624,0.7115536778729489,0.14159208150876312,
 0.23749589147429484,0.8384395961327658,0.270192629743511,0.2958000717912226,
 0.7723119249580116,0.6683191461962928,0.48484842937724826,0.5106619451888609,
 0.27792238144668924,0.5251567011436209,0.3346872831742824,
 0.34973661453183913,0.8848077833867971,0.09447183457783837,
 0.7687240809642412,0.3605538520770206,0.6886326512334852,0.8029520726582243,
 0.8311456922830188,0.5748113764396667,0.6587117280829285,0.3232145225132993,
 0.4969852555808757,0.012150438639255912,0.3889948532759143,
 0.7380667467350135,0.038795215541946315,0.6926657326002551,
 0.8878216434226217,0.5787971519073005,0.39358527615999306,0.8130917675462885,
 0.4957854026704578,0.07744885214266528,0.9576705902846718,0.753578141718164,
 0.8810073601189462,0.1619842644868491,0.26435676323878643,
 0.40157730983376483,0.006173299031084323,0.5975284130275049,
 0.21481843916679444,0.43821332000629987,0.004652442462054984,
 0.5694285246436029,0.720465129594944,0.7395327962471343,0.5382286845164744,
 0.9805334139752893,0.03948331884629497,0.36096592667125627,
 0.5725680587261968,0.7871600357915381,0.404149669499442,0.9511623109261315,
 0.32934692228598916,0.7627656651434571,0.30748866382965745,
 0.1963548879742969,0.6082381274171238,0.36711086840067153,
 0.18287335214015532,0.5312217374777201,0.8268332076495687,
 0.30016884304931024,0.5858026580733322,0.8692242378685586,0.8275057119811193,
 0.8023274799526497,0.5298439672663467,0.08629640191750632,0.6874319730925139,
 0.864481570032648,0.10112028090423486,0.6258742959742976,0.5493471958705419,
 0.9704040011843666,0.25442175168293135,0.4573721002567557,
 0.47453079317160995,0.01929467188409395,0.09414800291886638,
 0.8427744497371866,0.004757132773744788,0.5798334753256263,
 0.8333773362396016,0.940602453154493,0.3477523569296356,0.3947016234123297,
 0.1919145829678266,0.10319707702491099,0.42513489742595867,
 0.18788251487405871,0.028393867322404454,0.584689467403646,
 0.0349565892147401,0.877982471782913,0.4346980391513412,0.29956772305759005,
 0.4595233651588846,0.999395732614395,0.5353446260653475,0.5897781632874969,
 0.36825923535224614,0.2806208681012369,0.776704564818643,
 0.0057248432492251755,0.980812478036032,0.9667074598055441,0.589994830720538,
 0.45007438118840715,0.14792079329881225,0.32085095234118144,
 0.2915000788225517,0.7402374232321854,0.292119536918219,0.6084671734115418,
 0.15308646212523058,0.9548448157341972,0.1473966138716699,0.2600196835690127,
 0.1187891390607414,0.8563550576168294,0.22206486827465166,0.7008456043278145,
 0.33522954578273145,0.3688402820961282,0.3524837263345859,0.5383110822337243,
 0.2742138000131831,0.5025501131335094,0.7922957133765199,
 0.014506134738805465,0.14455535308087286,0.13054337164052487,
 0.5725827212733322,0.5893793522108063,0.19942581005357707,0.8829972941505267,
 0.18120965420564228,0.6374599076893447,0.9437852638816406,
 0.46278840953203537,0.338231692929153,0.06224191899462306,0.1327450334543443,
 0.08314110221865645,0.13106587583535556,0.6282883249257475,
 0.8839661291280293,0.032089484638373134,0.5712564910013859,
 0.2711443759829203,0.23510534288583607,0.4856142081455618,
 0.33826296202770245,0.2470801660902806,0.5686266827174967,0.4429872080194417,
 0.7479121252604576,0.4585463636643672,0.9130182876296076,0.5456976513727968,
 0.9197596337502197,0.24792586881190692,0.4043776458618209,0.1293265442499909,
 0.27965427367770646,0.9954613430667901,0.5974592191252507,0.8173811388539195,
 0.8581927630805489,0.2515642190563332,0.44743971600699206,0.4617442798920911,
 0.07082340324541447,0.5600254599236107,0.760458074952144,0.2916326908541682,
 0.31781206161227504,0.6488263411311924,0.8677742683326493,0.6939368369537832,
 0.3876319951086513,0.2910324666092019,0.3276316075624085,0.21680351663428876,
 0.9128807610954767,0.3168839511666577,0.5976855738681647,0.7516543219720002,
 0.07038724761804294,0.7213619628313668,0.9984906240365271,0.2817181944212247,
 0.2623306225026486,0.4134337192994211,0.4664753362202607,0.4683351223592027,
 0.4305562603990505,0.11892062985879492,0.6676043769074004,0.9662232293210934,
 0.5389532896894198,0.006649178521807131,0.7384207502936033,
 0.8535888388607087,0.30644081866025086,0.04909116875996067,
 0.7768223554980488,0.3689163273785515,0.3434767265615425,0.0826202288752319,
 0.66623100763266,0.12264087881962937,0.14495135489546573,0.6627306947846081,
 0.3006172973801221,0.3077368467625643,0.09918520601801495,0.5178620559453936,
 0.539426989166391,0.49345410609441165,0.3346057955876506,0.40678051037855023,
 0.226891112017525,0.6590968854216285,0.8161657356600756,0.6176680773504102,
 0.2669920997266261,0.7144361979803023,0.020106743524167414,
 0.020507496507707335,0.8360779683437273,0.3740656750512593,0.732135305242704,
 0.6933056427419996,0.7256081371945794,0.9878566555589852,0.4026975075586562,
 0.20509454997857812,0.8467365466674555,0.25037015344164015,
 0.9131947477732341,0.8679055554037328,0.7996024443535662,
 0.027063755837277603,0.38825880224943643,0.825229844208427,
 0.36248741573568166,0.5498486088304138,0.6335416505387008,0.0824431858546355,
 0.5775732183388255,0.3255285837818124,0.7859031127446964,0.8765315399065,
 0.15821261791756336,0.629192887784284,0.12298088028278131,
 0.29962500644846124,0.292440989699179,0.5569797437310265,0.9940281833677392,
 0.13642833319592673,0.586040433497191,0.22194119235656562,
 0.25403988618917117,0.37259711626596426,0.676463689582828,0.9586615678956781,
 0.6773686921516849,0.3483606005109685,0.7547808999126042,0.10287989518385088,
 0.057732717771352804,0.7900242178736364,0.49162183257042535,
 0.9089053278182161,0.9988091486006977,0.9572872506702574,0.45184442632912913,
 0.4121318209126593,0.42798098238973703,0.052351012844215616,
 0.14210875867423112,0.6284363594317788,0.6038187288765696,
 0.06356836261006671,0.03250226820101376,0.08107306158580452,
 0.40534053135159087,0.7077073773075712,0.39743525010261693,
 0.3200795235928877,0.006064879796820133,0.42160708666859226,
 0.2773474503769757,0.7694746290749627,0.03465960953130853,0.5434665283725278,
 0.4825665739186059,0.3553139294754516,0.3761761880341389,0.26162694606439507,
 0.9473825443372506,0.7162639674948371,0.48731281129168846,0.625113074259823,
 0.027483399855248725,0.5691706152698591,0.8013774054460593,
 0.05046791140784901,0.6045057574346365,0.6929094483721998,
 0.48085782755363937,0.5404279113655396,0.37247822516989226,
 0.9088277653526527,0.08431395637872241,0.07845064437005511,
 0.8620325502564088,0.29559746560228395,0.7264262916774686,
 0.13298329380137197,0.21346987636732806,0.20711490425715295,
 0.47031245453061543,0.7808937743703284,0.5305793629048787,0.5730769991247823,
 0.31643477258822983,0.031951082146204346,0.6796157684334789,
 0.1849569674213024,0.4339912207710701,0.5778356201126761,0.3414582724573535,
 0.4149638575828426,0.30758162241830533,0.006626492336807566,
 0.390552676702837,0.1940403345332713,0.7060692881128222,0.3054718893392342,
 0.6264022685733944,0.45728582114613736,0.1165603398768027,
 0.03416377527358605,0.6408973729737264,0.47316064523284784,
 0.2312613221648101,0.7760210464868856,0.9292576640757535,0.6646379204574714,
 0.08561830249271418,0.9777752207914685,0.024230377212940857,
 0.34409396947761384,0.5112030768554751,0.7872853341282894,0.5492426544146826,
 0.16219646549514877,0.9357477702685844,0.12184727581890564,0.566811422570948,
 0.8382755496229064,0.21794496494759907,0.0676254279423153,0.5948632023708743,
 0.06624030320398289,0.6563424875162831,0.18684224252782955,
 0.7370421589989731,0.30820138489665705,0.5373019996034056,
 0.47955651045784775,0.6460194440510337,0.01644536700059207,
 0.7993457474923925,0.13859548439791092,0.3848389132444857,0.2912255765064349,
 0.29187183502640146,0.09263157246513576,0.724954036230529,0.6378305037538871,
 0.06921167891495017,0.06526225839345068,0.7150732941174414,
 0.40986208973594496,0.6126122816768815,0.9213695981874287,0.6438843567394049,
 0.30218273268281226,0.9342380347186385,0.4402490186512704,
 0.37823213728961647,0.1264498816523375,0.4427790888621922,0.7843869109495657,
 0.4231767094838611,0.01157018741681426,0.35619459590152136,
 0.1490348536836359,0.43826776748460305,0.46777406017278744,
 0.4691149694725891,0.7600870168694691,0.8922204406613871,0.2184347184850104,
 0.16237710830387342,0.9294804224031742,0.9968215164055967,0.2929441567226153,
 0.9976500302979922,0.7597565555937222,0.0901897139103297,0.1460552288948178,
 0.7931747051723135,0.16785702230192712,0.42373729593222853,
 0.5664640280755919,0.2958412989312673,0.961943047263382,0.08186507734539028,
 0.2186397613865776,0.040651986792823336,0.732922988591944,0.4472793411074767,
 0.026653680622151565,0.8656739506141243,0.5903631290525349,
 0.03686637354669897,0.2817643724015515,0.6082268554266107,0.8425577606887434,
 0.4383529621118637,0.4429950907248986,0.23377698280186798,
 0.10494531071246527,0.6353873010064368,0.21428369941035408,
 0.6924707349692539,0.47379709212768684,0.7351134969650117,
 0.39574622675444515,0.25920828183354794,0.6192311509193197,
 0.3889235061066161,0.0614540535620951,0.8035704068648412,0.8687915751136386,
 0.306431098761214,0.8247719750345789,0.5198143624602538,0.42243876084779386,
 0.9116709644550858,0.9396518271503257,0.35982084607120945,0.5780670220880539,
 0.41661478686225584,0.9824246247953456,0.9050384849984914,0.271547568802346,
 0.4985505529482318,0.9112006462490143,0.6083903453664871,0.31865650438475246,
 0.09496610759884705,0.891480528248918,0.9512905424767711,0.8539385527485044,
 0.8180094665724746,0.3538753608128037,0.41328189411802896,0.5101134223459785,
 0.24115474389355074,0.27851034295992705,0.6543244026842036,0.359786926334032,
 0.6520001359681065,0.15819380466568145,0.8084639927913139,0.9041054424238693,
 0.6193211016055472,0.5528878996335153,0.7547480443390734,0.16647374452087782,
 0.8897547040636626,0.7590918548670899,0.22247208065845925,0.3542980233852555,
 0.04201462099432185,0.8613977448777208,0.06707516886879306,
 0.5950973744002291,0.3051378362254953,0.03376262713581102,0.1695935291395586,
 0.9588276981130107,0.1535377134841659,0.3329391004253951,0.7950007928729308,
 0.19298532198892981,0.6600108997405768,0.8464772870191126,
 0.15318848678050756,0.6922640371890126,0.7127142852888089,0.571160704519188,
 0.7284341566344195,0.7280089540360719,0.8173841456896974,0.10082828123149501,
 0.10417340593799806,0.4463027805116887,0.8307771566664416,0.2863889449198983,
 0.27590528463000363,0.692942426820266,0.13919971827065347,
 0.45078394730816485,0.3428379871484103,0.8365766119391982,0.3656021118786791,
 0.09086865564407354,0.1934170740546517,0.45132206510982475,
 0.9201820225038847,0.39237080881206454,0.5749008700974537,0.7568921989534606,
 0.9324971805902846,0.5365476126764261,0.6199791624384103,0.7421567592381783,
 0.8195112973109014,0.2829072554765264,0.03563516565250491,0.2770709797863089,
 0.4920329986380093,0.7939768426673175,0.5981962758946806,0.064997447566007,
 0.32246781783088097,0.3437348235793829,0.6952990092326162,0.9504182955973021,
 0.581294203163333,0.0024647658784990867,0.16034624230690897,
 0.011605359153144312,0.6058677947887607,0.02329111164395803,
 0.19981447969323107,0.47196439070120955,0.008662950117145796,
 0.02743987851168339,0.780335939067175,0.37985063089522053,0.3143594914128933,
 0.4318601486384521,0.955093866650625,0.7613364933950983,0.787344480433658,
 0.3326151888320209,0.8636648221597061,0.16301566671079448,0.909684351890712,
 0.14923955861049853,0.6045685509983287,0.7506472123517264,0.8492155673770523,
 0.125745940723769,0.5035108379649291,0.7123549496149498,0.11059251075447785,
 0.1919929005024157,0.18524789556004972,0.25697848508916155,
 0.9752145143189391,0.45441482669238265,0.2598923358204721,0.9947665731563181,
 0.5604100043939164,0.13699591825991675,0.8164231460163067,0.5193940673776543,
 0.39774155980183024,0.9184936934714776,0.6090050887561478,0.8048373839796386,
 0.6575066616730236,0.028113597120391853,0.5937352200039306,
 0.2445617925464113,0.3793381364264665,0.538879223479426,0.17244721861832113,
 0.7827070760268826,0.041957424887468564,0.7280511203896095,
 0.9309225156454879,0.7277949368363408,0.19542094209271166,0.646200352306528,
 0.3178797145392981,0.4438792357659881,0.6130540795143524,0.37000937648729326,
 0.4548575334175715,0.618172621701754,0.6183331695095754,0.9146468359943237,
 0.9388660015397723,0.226625350921575,0.20737644568338398,0.18725953594953915,
 0.4292919794420962,0.8461755505665512,0.1356430437965801,0.5355046153409848,
 0.676475822630589,0.9010982576776997,0.1336959450008075,0.8617618798935337,
 0.07277269229254213,0.9713299070610615,0.7246997025196842,0.893759956757847,
 0.37522753361326433,0.7837429365105669,0.2212423769646622,
 0.34360394871234023,0.1210641764100383,0.4902491564755862,0.91391947649849,
 0.3045090132972479,0.6446511955596914,0.025394762171659035,
 0.3288501496702124,0.8060464607508235,0.3978518241737494,0.05156820424145914,
 0.7908661337445697,0.42710646778892125,0.8843165260501702,0.2102055383123459,
 0.641188493281003,0.46202565659410333,0.5990481329917473,0.2127578489051618,
 0.31181682215475937,0.390034685383705,0.6438501399918557,0.5972217446557104,
 0.7343664499384591,0.9286038326031294,0.8994301839538332,0.7573057071006879,
 0.41517784769539146,0.38501645788306504,0.7949829914874779,
 0.29939509764383343,0.3280970439084432,0.5187744441270168,
 0.12991465894921062,0.34978819239375436,0.05416965428578324,
 0.9424733152756934,0.028385236434646965,0.057957420828886,0.8676357182323824,
 0.15322581537861013,0.6068468008219811,0.8121136842216934,0.2355328897205906,
 0.1776862678336384,0.4777230173954813,0.13047024781312155,0.8530625970754983,
 0.5734365761412965,0.8780051964072719,0.10332891544010636,0.8543475553768429,
 0.7738107368518089,0.3825929062035278,0.25294420313477173,0.5312918277647097,
 0.9064596630070586,0.9457230349297024,0.8737759788321925,0.5294486757350215,
 0.11953098113590765,0.9498560457092633,0.11202835354445362,
 0.19091247663740374,0.7628581410290547,0.12260951243880747,
 0.8820993725243083,0.5545883358458954,0.25067801744136864,0.1767034759633057,
 0.7333572542648259,0.6517247663351592,0.7887187159556638,0.7693105970070115,
 0.9701133735546289,0.4124847181020004,0.8840252671048828,0.42027412729712,
 0.3636572203370274,0.23068923110176687,0.32292946939573397,
 0.5611198008577154,0.8471675661494469,0.9817187469169791,0.6603575835425439,
 0.36463117224413255,0.04035922734451958,0.4196804167748014,
 0.10614079690079514,0.4450863627216466,0.680926524319914,0.8561516561743134,
 0.5570420534664956,0.6152223570499491,0.7398700175861406,0.35107557124172295,
 0.0922663407618608,0.7157852837335105,0.6921225269365551,0.16241521068738984,
 0.23074310168055778,0.8059202032827326,0.4270361013589421,0.3288093845275166,
 0.38300805816777406,0.7718377709319733,0.8110111963805884,
 0.26037108712778423,0.041607246940002396,0.45949701209102334,
 0.1797263773667852,0.8061891876966739,0.6634402117671298,0.455391321542975,
 0.86376266330763,0.22856509708423633,0.0380516278424905,0.5266280275371791,
 0.8835897608020701,0.3366810913310263,0.05961064529122573,
 0.07035217363319113,0.5041578122546279,0.7331767115326291,
 0.14961622559374455,0.8991522910832475,0.38632939294052493,
 0.9449475204505459,0.142881628175353,0.5371005237598456,0.8743673859994968,
 0.9516355591438856,0.37709072947131606,0.19263273585045326,
 0.1012058157345399,0.3735546632382084,0.9652261592671824,0.3473127924650612,
 0.9880965017252504,0.69214253158699,0.09409686508816917,0.5515566758769018,
 0.5339677314475826,0.9272336544212598,0.5706692086428613,0.8094078654248791,
 0.9472806801034596,0.9720836552875923,0.3939952514701395,
 0.009339558940999027,0.08956222017129023,0.22238001437418076,
 0.9242709275627603,0.04173032856530878,0.20918239614918543,
 0.6919127502130377,0.7340660575234291,0.9468959736979023,0.6432034891164066,
 0.029376508535411072,0.7316421355060897,0.8855306700785384,
 0.7578736130879604,0.0972949758837236,0.026649545976381916,
 0.6313239079866042,0.17950012882498567,0.31992082809647515,0.344502600836173,
 0.519652434412092,0.8128781101570969,0.5712155680016299,0.7300745137498527,
 0.28079224433496797,0.23691631308124395,0.09494616498449515,
 0.2913831557523202,0.13705439219358595,0.5230632125986788,
 0.11571446270268926,0.7381529529831159,0.6853612028242717,0.7628294671006515,
 0.8130046842425238,0.22354783340979645,0.059182068675338706,
 0.6317909531116876,0.3908289665093204,0.16123973802943425,
 0.43288007057941946,0.5629406540892984,0.4987632321740697,0.6227285323874686,
 0.3176224986227385,0.7746971589702453,0.9796743108836524,0.7479022668104099,
 0.5753089977615397,0.13784736040320233,0.903006447027283,0.17592556667920933,
 0.6639252578479867,0.8356072415059725,0.25935601767537986,0.6711418927719054,
 0.37415159567267164,0.011836571597396817,0.18131919145897568,
 0.8597430749210484,0.2946237382071649,0.435724709427446,0.4195930946866072,
 0.020048470695870346,0.41758185616290966,0.7924619681511808,
 0.07387988696790515,0.23304566487053946,0.844236103399054,0.492482724716798,
 0.24390325947951197,0.1713332254305716,0.6636786701066678,0.2765796414518429,
 0.9957024578640993,0.891529821557963,0.38180293385728215,0.9745440978341033,
 0.36889622329906535,0.20476969313619053,0.18395201595118837,
 0.18244411477744926,0.0060135473431131015,0.9498916704589676,
 0.5250223224094535,0.031192575037073178,0.2634666291960005,
 0.3830359965548221,0.7229002649496212,0.5068720474926062,0.7396105059912627,
 0.28291853771236253,0.20136305330240167,0.12590404280585954,
 0.5063596554167752,0.9741923733913653,0.23512025870579423,0.5815862950187092,
 0.3396624077649655,0.582123123124162,0.06724546370593898,
 0.034046485157844586,0.6138593556540135,0.18544547763448394,
 0.979153789295834,0.9881746477363317,0.6691292972715683,0.2535523740567289,
 0.39012410960112953,0.3751733937495416,0.45749402362442215,
 0.5776282730703518,0.9610519476248623,0.9262043601300103,0.6477751451203628,
 0.8878355608576783,0.7425133635576207,0.14924041113579434,
 0.25382790856295323,0.2164439488659008,0.13340256511055926,
 0.8586890035065841,0.6494033225406793,0.7501147033256619,0.034200568576363,
 0.8328579839291335,0.6937187972143558,0.33049454930863986,0.6232885366347135,
 0.9123660864881036,0.5271326736445534,0.25772943503955836,0.4124015252032355,
 0.011744952705441847,0.7421792674709179,0.18745964654662084,
 0.4094978477252499,0.7808019471751063,0.15463495243479153,0.4689724026345845,
 0.7761284550446799,0.8564341266820059,0.5554113002411242,0.05081428499675966,
 0.13819137010640115,0.17504462571097923,0.4942225712775101,
 0.33066373169791863,0.3365736251778255,0.10875448845744373,
 0.48222429005542744,0.2543019430714476,0.3984260468911508,
 0.09325508322738041,0.8355662729194079,0.5683794490518032,
 0.007204075100956553,0.820886401214903,0.1468190351451687,0.7701211955102111,
 0.010973800495319819,0.06139688147866673,0.9828770003084625,
 0.962645910828075,0.7955993927190136,0.9371215759560966,0.4841460539077648,
 0.8021450364930026,0.4175612325143857,0.9771127542205488,0.49757654313132493,
 0.4714618645906058,0.7550526107789456,0.2403805044197701,0.930088097835614,
 0.1961649527041891,0.5127737334508208,0.272774795135193,0.3350899520014837,
 0.35513204039098567,0.1703960545595743,0.7107093400271769,0.9149185369044608,
 0.3615088406503515,0.7311192535252079,0.6517942085024694,0.23479179947756168,
 0.9221980489327956,0.8375097655204082,0.9873948794909567,0.575393467237977,
 0.8647271643156897,0.6177304295438483,0.46147981707946506,0.9687277846377857,
 0.8998095360045725,0.8092394988702662,0.1109279138750211,0.1954110070175189,
 0.5510247125813399,0.7792099032556248,0.9507355644070141,
 0.0012983543321919422,0.8891311459393053,0.9537796868274763,
 0.4302252565906398,0.09450241651276148,0.2815971264663214,0.4388069083995594,
 0.9511588884732933,0.6342180121029165,0.18546533664665144,0.6901019015417474,
 0.6852325690659078,0.6832075968788207,0.89354436776226,0.264958737474306,
 0.010421090602160454,0.1320549574557075,0.6357581069319359,
 0.19031801405122117,0.4127951453252219,0.4443024498137288,0.7403813938374337,
 0.3622930764959227,0.32924100272921775,0.4538500552578395,0.8178955522445192,
 0.08160842217147257,0.8325324288931086,0.1704926323491558,0.6836518349689698,
 0.11511472726365923,0.33855982625821224,0.8132237495890922,
 0.9805022002738939,0.9091909031223697,0.9164791195611592,0.8370611330329497,
 0.7649160337900264,0.5273556375591814,0.6578538911331436,0.6348068602314818,
 0.5881153493655596,0.39103992675705146,0.2508347344225067,0.928248451643694,
 0.7254326616719663,0.3469805483656294,0.0537371279502018,0.24527914950728558,
 0.8389425086218177,0.6679973416347207,0.4188669128062942,
 0.009934424915905127,0.09804648817680395,0.9991396800879163,
 0.1707580876295275,0.8704072220149254,0.47692236644404706,0.5917879036865307,
 0.9849899596709037,0.04292321052590997,0.48112771193952253,
 0.6202744089802481,0.7521032885920234,0.17497130761432533,0.6950751693108682,
 0.9431626964807086,0.23734293565648545,0.8353276074990407,0.2678247970623987,
 0.6912924907150078,0.5628976184552433,0.8573647484868745,0.595313003580008,
 0.9710354362457795,0.16931665339565471,0.9040283126969162,
 0.025811852072171404,0.052468836228827254,0.8002702619215115,
 0.4557687845120947,0.6860865942939767,0.4985062364027991,0.19365714554340785,
 0.4245466404364474,0.48682512360046126,0.3221199978566911,0.5831780559819202,
 0.6642973545544709,0.04535093576367688,0.15066727740661034,0.724795780870944,
 0.43870924476682915,0.41446095286755724,0.959633966649599,0.8665468646950409,
 0.892247058945162,0.8862747171948021,0.39873881222534013,0.6882650588405137,
 0.6524164640015664,0.9550889025332512,0.9203839078423849,0.5229311406104917,
 0.2845218403450296,0.3062218265266172,0.013242947678476558,
 0.2783525564252337,0.9054863543301092,0.18885510667790895,0.2628638733566422,
 0.8350490184962542,0.5550599179246791,0.15943515533840213,
 0.44339770101874043,0.14451706345596482,0.2372902629801814,
 0.8581101668495639,0.03664163443268753,0.9538821828785325,
 0.26595793607609797,0.921228994731619,0.3286778092741469,0.12656625273679423,
 0.3481814441629809,0.4352886182267548,0.7244287801688207,0.7551476741951416,
 0.5930056153219085,0.6564910939029853,0.19267810402072982,
 0.06778927885147001,0.45913512197707285,0.41335323952252845,
 0.9242065829328732,0.8943233154742296,0.26433191283377155,
 0.05224124054268142,0.46767631009694155,0.5999855091150198,
 0.22163967332024737,0.8330821277805781,0.0194980433753269,0.9319670806189781,
 0.05394142332134588,0.015952016135173164,0.9123014561276492,
 0.3453796205522275,0.5938172650729058,0.9621306856154848,0.7881682626705451,
 0.972023258318355,0.7422799397310953,0.9984399798345165,0.4810133801243376,
 0.57546180214663,0.6984424448740112,0.7709977708558217,0.38453290788637795,
 0.12178371458847492,0.10130707872857281,0.786748925791393,0.4053107845379399,
 0.1821804772512965,0.7428196935643043,0.6502617579835052,0.823573163776208,
 0.7979669050722491,0.3353540112943658,0.4963244645483581,0.5654913962799115,
 0.44402328288565696,0.8792052133224172,0.8974505069209278,0.7199622469956799,
 0.27813929537196036,0.4683476065980272,0.052638886537592744,
 0.29543854940730374,0.16374633298494,0.5774537358513137,0.5113731679658942,
 0.5781128958053696,0.2588384221367501,0.053281286358596525,
 0.9403163649540072,0.9721800163046163,0.07078939246897431,0.4696506470448393,
 0.9296605536822301,0.12235575351787031,0.6747378842084536,0.596645299412354,
 0.4354792958152556,0.7001274875347756,0.9471523576054511,0.46236313501293713,
 0.6462387852474712,0.32930425073196956,0.4759994632661688,0.5232770010203794,
 0.7356663493918751,0.4109875619473654,0.18053668660530886,
 0.22540986264226892,0.8602819332946277,0.7181377295619262,0.2995725773709117,
 0.7730606427441926,0.5471263900617045,0.24318803165002933,0.3191852864221757,
 0.902282284289865,0.3435690288095671,0.7783351811099125,0.4415087218373871,
 0.8442515737844388,0.123196880787111,0.014059729992200953,
 0.025599707617989353,0.3181134460491649,0.9722654744392081,
 0.7030528955280907,0.9383355892377112,0.40264133560072046,
 0.08735229585366477,0.3820898547794729,0.2888892367686333,0.4324762929116204,
 0.20947308615191873,0.8579617131225635,0.8521759782729887,0.9077653360518069,
 0.15688635269541507,0.10068506501132846,0.46994120136569884,
 0.3153290638674411,0.5443710074158367,0.9547943905552324,0.5775162316425994,
 0.5647482695316299,0.28474144728985007,0.7027192594648277,
 0.039721085450785454,0.1285325442788524,0.8641546184886404,
 0.8207436556649809,0.42963708316849336,0.32288300770760736,
 0.5207613478942101,0.38573626820172324,0.030817989568006854,
 0.14718539763799476,0.35033327049316443,0.5498105161834491,
 0.010559299659981258,0.8429126954026583,0.36358631239040706,
 0.11655138246054308,0.8370308632648358,0.3890951161974123,0.8303431929279044,
 0.11462843221296859,0.3754804656863061,0.06861971891809238,
 0.25074634756915815,0.8870926138637278,0.5691788840175389,0.9715868567907321,
 0.6987690476494484,0.3686016391634538,0.7674518553716582,0.14538576552453075,
 0.4679133623130124,0.9730858721459887,0.05507741929785792,
 0.22284443322572767,0.8552708151975303,0.8654987062902779,
 0.18425285380184142,0.254432762291804,0.7247622161251318,0.37480220469062,
 0.1107650208520844,0.8887187411095248,0.41430458990612484,0.9539239382084266,
 0.09977129704761911,0.6563550116645447,0.9003523521444573,0.3748268510986231,
 0.9698852305005683,0.6081123342627266,0.3290790959076514,0.8492192229874503,
 0.2231011599821784,0.276348017911891,0.3324494922998473,0.2573180598020337,
 0.4220096278243731,0.29229582114396124,0.5259946748850975,0.559431005785065,
 0.5492638095981488,0.6283175026857691,0.20147176892659047,0.6763619075042521,
 0.7065972838658434,0.8128002042099445,0.13963210820383665,
 0.31129427892857386,0.1703378887659086,0.8731569691619268,
 0.39270540224604367,0.9695010504983882,0.268613450715266,
 0.058700470193464915,0.5580863863569278,0.7799362592947718,
 0.08842743945645914,0.9553771629720375,0.49420329907561733,
 0.2993915957338944,0.7608418439832677,0.5759266829773168,0.03884022058186676,
 0.716709043226493,0.41372851432593705,0.1509435956866596,0.8164234235329145,
 0.6141735761817506,0.4415379021437479,0.2861703712534087,0.48091730927421794,
 0.9796568797536498,0.7705164905440962,0.6185662752228431,0.07135140306863996,
 0.6317127055485702,0.8261695886304775,0.8377341452910696,0.5485646736493481,
 0.7992849399418945,0.0752178925275645,0.255533736858895,0.2051072075043212,
 0.6292398229713139,0.30681834208725833,0.9970112628365678,0.8052615760510344,
 0.47979708565410484,0.09989248183720167,0.5478166885681249,
 0.9734744626184403,0.7874703918588675,0.6879327031793276,0.29436953303825497,
 0.4435049559707531,0.733360068108262,0.6026268794389662,0.48097099753568007,
 0.9004632856242061,0.9046905041154026,0.48856865328368815,0.7840711587439947,
 0.3533548089133106,0.16956885867847749,0.25684767943245734,
 0.2405403034957475,0.9483244280596471,0.051250092997171315,
 0.1293489119272162,0.7283919572672091,0.6666854012273999,0.5641959460378527,
 0.22649245993428702,0.3039161792723415,0.22817624987503538,
 0.4305122543641843,0.3980611743352518,0.6728199290211464,0.5426657583076798,
 0.5280399326032736,0.8867606928429843,0.024104748596320058,
 0.031208444698999527,0.719019030564112,0.18360464529407972,
 0.26621835688656825,0.48767401033782865,0.535071742750089,0.7305708123309156,
 0.2361736390755802,0.5791440704622456,0.868573541002772,0.37800152606451787,
 0.09146146613192285,0.9636417179466863,0.7728391981278334,0.4532897298111255,
 0.6051174758660746,0.5541495007788528,0.2471923330518705,0.8974862083664971,
 0.7833085202172949,0.9921463832429395,0.696278789295082,0.909566874286336,
 0.06292787036993486,0.8236152299596996,0.05777031793261023,
 0.8625774183282839,0.8477325721200558,0.9668278194361721,0.49271635503071076,
 0.7730479720838817,0.6614059846655365,0.6810836317668282,0.3237760622709257,
 0.664116269818024,0.944828925309041,0.7172028844271613,0.9845058349585418,
 0.6857501988500279,0.8805986949679551,0.8776102426093592,0.3741490233876871,
 0.047147248392889196,0.8405484298221769,0.9977758245695698,
 0.1980524406789521,0.29022664080884275,0.8180256304950191,0.1694268606743109,
 0.6157343643814692,0.34625317863513716,0.2855551600943731,
 0.0012502377950923682,0.5993143678412428,0.5471798399714768,
 0.24527438195808515,0.3767542509377907,0.15694436239799858,
 0.5315898110086519,0.1783367058904437,0.728296251723866,0.10334285014058697,
 0.6287120602793063,0.7691555339464511,0.3640623447614355,0.05986028686757727,
 0.01753525136679146,0.974586736531571,0.08263762545287845,0.2588204116802053,
 0.7850191818200982,0.8482608099941603,0.9042096328381625,0.6006363469979986,
 0.8101242144382842,0.3662791246638637,0.5741680251870799,0.1465592970754166,
 0.37187476957294585,0.3348933275085203,0.7600051181672309,0.8384759684237768,
 0.9801413635962556,0.24786545945666072,0.6033739750155652,0.5360267356998921,
 0.6626400719047968,0.31602570285125064,0.15994064057846957,
 0.8102769833745452,0.9080357187537605,0.2625314232722076,0.8240946839274365,
 0.8025331787974138,0.22557982231331064,0.8917935132605119,0.6441231218763398,
 0.3355296307471023,0.0557721030055337,0.26405174019162314,0.872223634431945,
 0.28776123575740853,0.7099576695951929,0.48507331803328635,
 0.5805000024413811,0.24184784264029213,0.55226290140047,0.19040369441925287,
 0.7904760435099819,0.08249556199224073,0.6866856713020656,
 0.19754539501689194,0.28661379938401954,0.24850840951450137,
 0.9357420779656231,0.141189160452042,0.23020433464013312,0.665870534295901,
 0.8399390754202178,0.4356830884384032,0.2523650083573028,0.9799498289283514,
 0.09219443965052188,0.686914504490908,0.4089938156494055,0.09415709798476235,
 0.937392667981557,0.2666029924634835,0.029156844004694538,0.738775531445969,
 0.9109506030294895,0.8972512255875382,0.104175798853827,0.5628953181474037,
 0.5634764214919076,0.2651566975373424,0.9875531357363843,0.04904703649858866,
 0.3785214165021018,0.40623322186927435,0.9749169303340806,
 0.11228877878710741,0.4176445633203575,0.7592157812169094,0.6103534191995552,
 0.9617999565212055,0.08681712854402512,0.953262859376867,0.5389823037190039,
 0.8633306332223724,0.1894977828671689,0.9449326130589073,0.8336353605573916,
 0.33401173157468933,0.22761806094348225,0.44087449462048656,
 0.7973327538487388,0.2296577810138053,0.7868946826697525,0.7461423056579052,
 0.6503251232086666,0.31903572730276264,0.5361508890132383,0.4926546118852959,
 0.378417599267585,0.8750692477239183,0.3911006091685115,0.259157232258437,
 0.8034081192293105,0.3221039022728752,0.4404292744849012,0.3660416081357919,
 0.8144671916926667,0.8072729328469723,0.5908681135410052,0.9524479949787605,
 0.5376969693793396,0.5557428918807832,0.2685845759455845,0.9572782776959833,
 0.6392563995055407,0.3656932726032578,0.1936170889311506,0.1598937855930168,
 0.8309723625401367,0.4750926725208292,0.6335952535732461,0.36486626402518996,
 0.9862863314887148,0.41800571306419776,0.962609480733718,0.4787824876853014,
 0.40053696194269317,0.8444292382260846,0.4245087700414296,0.9506460813499942,
 0.5482189632808385,0.5844830911700385,0.4594923716843765,
 0.013290332290944451,0.7282978485513019,0.7870420185333218,
 0.6791161730375849,0.20917607009943862,0.037047746667975945,
 0.09986849467347847,0.19196911449176324,0.2980052538095954,
 0.4974597075264302,0.3938096179461279,0.9273572120107362,0.41032022756387687,
 0.9102662408862702,0.07212405505274988,0.44647696831767947,
 0.8892523865269694,0.20972115705380523,0.5694817770144044,0.7391009712451247,
 0.9960840736258201,0.5618067124267203,0.8652539638543797,0.02044053897919862,
 0.04187204800286981,0.30179000905003817,0.8338565080130771,0.40673520708293,
 0.1304860467088087,0.5693118632095714,0.9029772663682513,0.8145221016118044,
 0.5643917398831982,0.5084177179123102,0.6861250493201972,0.4111420645905509,
 0.37428319837805646,0.6548055483957634,0.8562910182410101,
 0.06851335173702011,0.03659511261675008,0.37253439175524194,
 0.10645826110043866,0.5050673946761204,0.3738407797143607,0.4035879724312156,
 0.01938597390946506,0.04354032358750903,0.8526830840907642,
 0.39098663290233515,0.8628225176927562,0.7279071474832681,0.9197413272568542,
 0.35219311514441987,0.02393144085887744,0.3117829072121392,0.539241729855341,
 0.33139412441798033,0.1395534739055162,0.5085681757706039,0.3800271354035776,
 0.29457765930824575,0.7660574901363206,0.2575430861097463,0.2817293135722808,
 0.5139384977938618,0.2365430864146607,0.9471988908765295,0.05928883484857861,
 0.1589284850964725,0.2622420095310646,0.04098060458611985,0.9582697667773309,
 0.8869021411980079,0.6432507922370039,0.3222801219920822,0.6298572244517129,
 0.6371493651372646,0.19074690886708978,0.01724260668911448,
 0.9006773017759993,0.21708085787560683,0.6762767105334136,0.4168938691819237,
 0.850046255995042,0.42470224239148957,0.6232498741401313,0.18729630756128468,
 0.19656945232184686,0.4644563679152388,0.29049297634851656,
 0.5273800055363159,0.944275248340606,0.5238122091989434,0.9097650194137317,
 0.8091767100590643,0.453056458947639,0.7830291421517619,0.12434440843109329,
 0.4470901539175336,0.021679999489894186,0.7050861327988376,
 0.15673195914530036,0.8617170596429444,0.8505673969594348,0.1369075245937852,
 0.27570601107113757,0.9937044609052982,0.728355190506526,
 0.010376064529859064,0.630707850925833,0.44011102881004605,
 0.8503096337195551,0.14438079390761738,0.8614792098884577,
 0.12352264130476298,0.9086191085573437,0.589123153461103,0.853620784731613,
 0.09884596970286552,0.8955012762929988,0.6532556739566933,0.4464721143605721,
 0.9736308692959132,0.19442787620620283,0.14080183326637008,
 0.6977468475014872,0.9082964006064305,0.9118244458589158,0.7620307064709141,
 0.6622545515528355,0.26767492585948005,0.6606057773406055,0.4616416286102816,
 0.04827769854473163,0.07764192561655925,0.9782027701954229,
 0.2106577516890089,0.6341020686571348,0.7243898242897592,0.8920961836534541,
 0.927629100442748,0.7253009749302589,0.39549770013852314,0.5402580383631502,
 0.034771032889367004,0.6381852599724388,0.8321058723703335,
 0.6426977380598332,0.1684581112025324,0.4998782359656224,0.3538775257188129,
 0.8210076879955139,0.6276133819309159,0.9979067377947584,0.3124936206929583,
 0.02283912380028652,0.10799530106415844,0.5945647036499119,
 0.8020192525852856,0.16776533566394491,0.210950456645854,0.07442362823642656,
 0.8097687227349823,0.9882325647866291,0.9144387444636578,0.9633859969400757,
 0.38770743823974496,0.7940845383421185,0.5333772791862684,0.9076427354144915,
 0.33544900063238703,0.4416204551179309,0.19108098925589323,
 0.5717285588869966,0.28431635468147454,0.2294545812185188,0.6283029258557967,
 0.5085876038502928,0.8014617321222746,0.7516524331285269,0.62199192008479,
 0.2745669684821139,0.733700360044626,0.38798018889946895,0.44670686136016935,
 0.3827847193665411,0.51469817342972,0.5793998043797717,0.6794435406979092,
 0.4506636454565681,0.7582332503034269,0.4343527294249454,0.09480223738465754,
 0.6506589537560156,0.5049517120823215,0.7138682821613038,0.10680190357512287,
 0.1450409665063057,0.3833698243248027,0.3654652259344574,0.6050732515340573,
 0.6027130377706069,0.228924742091548,0.22674568892282743,0.7397215588700253,
 0.19605131965014877,0.4201117317909566,0.7747459524626992,0.930733347560456,
 0.1782533415262243,0.5130213424666242,0.7866594196054412,0.7003330841606091,
 0.6587114186434471,0.07982501179934309,0.6242721068932218,
 0.46158424117690156,0.6298657761988462,0.9882374046862973,0.9500210096586363,
 0.7406922779853007,0.7521139167885729,0.7480686783210928,0.03886551416807005,
 0.2744397364117561,0.768365812153587,0.775228446988634,0.9061145222047369,
 0.5540117435873934,0.32970925291193187,0.4021692016705054,
 0.43476474000603105,0.9375486561348855,0.3163804143060751,0.7316232160429037,
 0.892234008608787,0.6485995864666467,0.21774774186347035,0.6518199879317619,
 0.2959350531440478,0.0788814433095264,0.5441482854861857,0.24504615311970634,
 0.6147993190170038,0.39344112563498146,0.04898400240312428,0.59481623478668,
 0.7952862054785699,0.3937586669428452,0.8295969637564603,0.9620906051939144,
 0.6963627671025473,0.750832819112179,0.5723035822984655,0.7066009621851668,
 0.12195433696646774,0.21048760895100782,0.1249169170811475,
 0.6471120808506405,0.9271039029319063,0.9126985625706203,0.20876534976428507,
 0.22652225050466734,0.24296927292088488,0.9619472427301798,
 0.7066264811123816,0.9593623670673923,0.022136419366516114,
 0.24210129362762944,0.7013115824947408,0.3464048544828495,0.2507620660537663,
 0.37003952167513654,0.9899790655892086,0.2515125938784021,0.6072261755805227,
 0.8352439895367203,0.4093730660692966,0.5264466765235137,0.7322573451425468,
 0.623207698452951,0.4914058578100695,0.23141390856544275,0.41291119703399937,
 0.417689689300061,0.9580573041191817,0.2313633606845149,0.7618089569592881,
 0.10174465381368203,0.7125758156863814,0.05103442230426092,
 0.009741135627656106,0.7469507142894305,0.28609051907588656,
 0.8211522484073562,0.6782400415024837,0.2481194029586522,0.7511466384580429,
 0.9247554032601748,0.6540128717377036,0.879933991297815,0.6257535283250744,
 0.9269197257850216,0.9425398321068487,0.57606854705783,0.9899048742682798,
 0.9554835005656642,0.6802425284659867,0.6093791971556164,
 0.004930803396035244,0.03632169985869427,0.5793815073380103,
 0.9684472968432382,0.5451904316974484,0.7730581212664913,0.3591282713725782,
 0.05466783850851875,0.6731922414762476,0.4010038142342488,
 0.49743356576218156,0.9956579497571141,0.35972655245873175,
 0.16206645521564056,0.860915460733793,0.05832207388425581,
 0.11463414985448195,0.9312594418219947,0.6127629446529564,0.8291717408740513,
 0.1508085321574153,0.9486927171405493,0.34452783826336575,0.3775531158232428,
 0.3141983455145321,0.44342584801605134,0.9767676438874994,0.5727827605725407,
 0.37667967816828385,0.7242765021935386,0.04183040736834087,
 0.1007578048025104,0.48075936464636115,0.5399225156434706,
 0.35203835161799124,0.49530772507093146,0.7276052797300998,
 0.7016904820960369,0.15045093635634288,0.3262261033698819,0.9657298917642215,
 0.8004818246936749,0.9700204646755137,0.8731542669013853,0.5183633391659905,
 0.9725886507037782,0.898056171870442,0.3536214610971887,0.5011430201811274,
 0.7293747108934494,0.8935182209436641,0.7390792922699992,0.7513899766228596,
 0.8958566818177721,0.6671485437589079,0.451758257310825,0.636076197240377,
 0.1870898558145161,0.21440108085408927,0.15373723849485055,
 0.5584887618946438,0.03649702564481194,0.9551383335973501,0.644191600342087,
 0.3647875116796272,0.36282108575765404,0.0016657287964034762,
 0.6675736487860726,0.481623268623556,0.6328248414414211,0.2942620430070516,
 0.9796377941930952,0.8966501239901912,0.8335663747390571,0.06008527832483668,
 0.972647248638798,0.820826188988508,0.06563243073694425,0.26462799979911666,
 0.4943742099677856,0.8532697979375925,0.696579160336417,0.49270503033722857,
 0.8301384715650623,0.22698365792145614,0.9312043025287162,
 0.08892986999323316,0.6592539061365775,0.2830126291820153,0.3806177977597667,
 0.034781592299572894,0.8708365391621139,0.28888923976463377,
 0.05528555089028331,0.7753349914833522,0.39930955786887445,
 0.5337692173334275,0.9133988861782772,0.1282078992960971,0.33404146012533686,
 0.509633298879134,0.8408631646526333,0.9850269085238937,0.4580608127345982,
 0.3910654654378285,0.0482616290773108,0.5843090153592813,0.34104272860461937,
 0.8863218362408043,0.2816731143210185,0.8340950942381359,0.545280370003001,
 0.15715846368536068,0.7506710417038807,0.9438602814395468,0.9116250048610542,
 0.69958983294918,0.9834740736584654,0.8493585397235003,0.6118084957237618,
 0.744087081854905,0.5180462807258367,0.8723140196704527,0.33372047814368744,
 0.7047445655326096,0.34740309708807504,0.46492027190533247,0.245360974594532,
 0.8831680626544782,0.8296363523316691,0.08959161845512154,0.9826809858358685,
 0.7597003668973474,0.7699059381260298,0.11166976867486489,0.6134242474057472,
 0.535429604797633,0.4034169591650215,0.48509301134183813,0.30578094667021893,
 0.9352216294826387,0.6011213225891922,0.9412661390537163,0.6279856417300359,
 0.11275430726695246,0.8628091859910703,0.26059984569799266,
 0.4523463544488062,0.1787331126060454,0.7115639488039757,0.50702882252347,
 0.4050642190510869,0.8524502914637013,0.5957249206319557,0.43675395144108464,
 0.28468154420468117,0.4140364849360535,0.8136512490915024,
 0.46652629980988725,0.3926286502319376,0.035518188851982324,
 0.7191899428061468,0.3027548879980959,0.27324498892916993,0.3543500759973407,
 0.47883685581320445,0.09087201050707006,0.7165222987879357,
 0.3676773080294974,0.3123956133556941,0.15926540808890355,0.8030563010965825,
 0.2251582217854019,0.9112786956430994,0.024486636256500494,
 0.4883603974493824,0.7271622142932881,0.6289104220479843,0.8980922244089407,
 0.05486014710043041,0.8445238511517981,0.92952721401481,0.5177087474200536,
 0.7941560364350573,0.5532772254553686,0.20146208557821643,0.780082968344248,
 0.8934167137545922,0.8771952698696242,0.6521797453849665,0.06332615581538492,
 0.43001060707041316,0.24764243525833796,0.9470945089066124,
 0.6121835936254142,0.05264759561427956,0.3251611772533869,0.8981621463253837,
 0.7767915141169077,0.09532643816719821,0.9940391614279627,0.757968241964717,
 0.16651198690964497,0.0646909750972966,0.4122149985193405,0.4961813544251664,
 0.7760077316192423,0.40583009213524024,0.7728078236736854,0.1317698862077873,
 0.905483135049914,0.9564157646438284,0.37459744808656525,0.8458332021103127,
 0.8606516241789022,0.8279982051740359,0.22229710446165674,
 0.45361781602295215,0.8967393929424993,0.5600425373427458,0.8921204408971005,
 0.2590268667188067,0.9018166743833719,0.8670882657786287,0.4381692344091592,
 0.7341708937303728,0.2359081699759884,0.1940882038273688,0.3987785341966418,
 0.6337827394609247,0.8268263008348243,0.3760719369799065,0.5134065792848151,
 0.48320560468201124,0.07286824561962824,0.5327874048573789,
 0.07095259142401433,0.8697031386260803,0.38383015083494754,
 0.2834310280873952,0.5981562727217875,0.3566587235081651,0.6599678634484956,
 0.16057986177990002,0.4692404558552896,0.2735156576289758,
 0.11540262522181877,0.7373908863067568,0.4756194056877957,0.8890600629312247,
 0.23207151779477364,0.3541740246235252,0.5025044319805755,0.9051688597705153,
 0.8490039751180818,0.9503627927679024,0.6045143391148768,0.33125931101428674,
 0.7680191897588613,0.8285774447017404,0.3972947050812693,0.7376533265871114,
 0.2462550306418957,0.9545997548593401,0.7698406026352589,0.6704218201832245,
 0.2446816653552144,0.062499837503631084,0.7932387815859875,
 0.3807707944240919,0.8407712553344213,0.002750672214490102,
 0.7574772833421217,0.6362230512962334,0.35887180712622335,
 0.37015680469352175,0.5172260289295656,0.25094236583309026,
 0.6128021965077218,0.22728630663299398,0.5119194714232119,0.7842748284920755,
 0.8682870086747372,0.40023367718723923,0.9018607811079274,
 0.015726882030854372,0.8407610316608514,0.46191116866775184,
 0.6093407135529328,0.8731933183939722,0.963918948692417,0.4375624229788577,
 0.4723506475486353,0.29163006698129834,0.3822040699812803,0.9055202408414575,
 0.4047191691831541,0.2969259505720341,0.8594696177747488,0.9348245800045936,
 0.27679289659680906,0.7184681654410144,0.11387757911453111,
 0.01734710028444797,0.717882472800365,0.8488279243054948,0.6684531412921217,
 0.6393664846833378,0.6598484810677894,0.3286306652300641,0.8770652846421688,
 0.4675544429593824,0.9273086689046937,0.12354386771500436,0.9752596391521192,
 0.7335512560804526,0.050455279215290916,0.7683222356646744,
 0.24953852243788588,0.8518925948391741,0.7773351939438621,0.8414898719885209,
 0.2649366263804258,0.6043469736156497,0.30195710960373856,0.2757191617772665,
 0.40212619578622266,0.09315555644066531,0.10471331323854827,
 0.28810200654118445,0.9737317600217584,0.7317098247132988,0.9236052580712182,
 0.7391230093954296,0.9356240216785349,0.12437293018967244,
 0.20720993233025797,0.4471759028816745,0.714843680290424,0.7142876278101877,
 0.059558084608204664,0.050020455177641865,0.4254040162265791,
 0.6340371349134216,0.28096796816171377,0.5305228112603031,
 0.05380038906248108,0.25360244981367763,0.8519772780449384,
 0.9094705894577586,0.5641111255922788,0.1500223149014147,0.14948549368366582,
 0.7328072483399523,0.4914716532092813,0.2612603161964183,0.1292374303228856,
 0.4542822156524382,0.23326945728813842,0.5127841441254499,0.6855545941635304,
 0.7543952793400613,0.8440178322129349,0.05715768969456336,
 0.16198619211673693,0.14821601953651098,0.5004614419179013,
 0.021782797943344256,0.33196423475885506,0.6847364617948534,
 0.9907476124829595,0.04309955192534942,0.11651881976756728,
 0.3703088882650507,0.6375256729284687,0.38785326797335895,0.5315402238276449,
 0.8749745614937879,0.4889172579324168,0.8171723518425472,0.7903625168912667,
 0.8916909714886253,0.9037554035656976,0.6557263856230011,0.9208139686101185,
 0.6301363748496579,0.25229322474931515,0.30217340404220927,
 0.15537235498548185,0.26759677629938683,0.1482353657349068,
 0.3643194015469705,0.8025946697238557,0.6994509954676009,0.7903624302458454,
 0.4013076398072428,0.010142691986157915,0.8779559739362186,
 0.2724375721782666,0.3817270674947256,0.5063736595133765,0.9725207571188771,
 0.28672037979516074,0.11576125574782326,0.07709878722199759,
 0.3905980289784453,0.8345308348550767,0.5046493258443701,0.8942104518425165,
 0.5854413877288637,0.7623176099993891,0.7758435394333539,0.22510518397863588,
 0.30028524554413516,0.44357152778004094,0.20607575092740837,
 0.00601676959760955,0.6035862299497579,0.6935536024706089,0.4272665932375691,
 0.13454786133948338,0.3707068656037922,0.9485482002534062,0.7840718134305406,
 0.6452782139930074,0.21192821223285208,0.934809999461742,0.787406891052709,
 0.8734141858025066,0.07404716378145915,0.04194221036020307,
 0.9392386164015876,0.8891533629210066,0.13277378270235407,0.7324996618363413,
 0.5719229776771769,0.6411604179213197,0.996149850295785,0.53065817070767,
 0.738486273722327,0.08023956813191735,0.04395857360924238,
 0.34440336102484226,0.9371492219965807,0.6697826812079625,
 0.29648050929470715,0.7816377807621457,0.5114637160283108,0.377024765481604,
 0.48026693328979375,0.053609966434005285,0.7255721479372463,
 0.6506577819654944,0.8196439910391313,0.29623661433012516,0.5381140440643921,
 0.994138125977265,0.11288120193024753,0.12317942687146277,0.5942226680962265,
 0.5958301014818257,0.391047876941539,0.45701795301358006,0.2752267718196504,
 0.4525820771216499,0.8677564288359283,0.9285372675675421,0.1928664472478988,
 0.6591641585852241,0.8732330332049975,0.9326226458373768,0.8679885014263814,
 0.44574003418627395,0.9703420476917033,0.45754775839032014,
 0.04166832937448939,0.6984236994897202,0.41576551839164755,
 0.4927034295397881,0.0962896949496842,0.9265790275207062,0.3467417756740243,
 0.07264221137604288,0.6757518383773914,0.7327561154120084,0.2599324902532162,
 0.5335400555024178,0.8542558896229219,0.09456292541805544,0.5010821332696831,
 0.3809419650414345,0.24412170326925808,0.5102089775171604,0.6930325095106319,
 0.17901955239663636,0.3950880854631972,0.22326343885187472,
 0.25196800803907604,0.11387741079812674,0.27891691469454705,
 0.5285641147608973,0.45756659161718216,0.5892894152859958,0.3027526560946211,
 0.07472492534688835,0.670964259270759,0.7097679022485743,0.09458648880311404,
 0.12987291924846356,0.7397104367558055,0.5113105215131531,0.7533512179870455,
 0.8829915187820654,0.14242413002589127,0.3989731287714885,0.5607051611788798,
 0.2270216054299542,0.09195815384024075,0.6012268064357204,
 0.26934710531189754,0.336067849073156,0.7665869129239145,0.4505169900757158,
 0.1802108198993474,0.24365809446798492,0.8591921120415418,0.9505220178675617,
 0.13426055768914957,0.7685961420270828,0.2191628066852489,0.8039447411161381,
 0.8057248616007073,0.4386642138809287,0.02749937332894503,0.2112162639733951,
 0.4620868240585909,0.1000958652066406,0.2395626139460013,0.22676706466543095,
 0.2816349634205697,0.7473050748380328,0.9301493940005602,0.3237626141701587,
 0.516933321776331,0.5618682068666558,0.9187504699273568,0.6859607706457888,
 0.455189602892654,0.05717513712066036,0.44118684609283554,0.6956278053707377,
 0.5415168042833098,0.14239663056287277,0.8809239422705895,0.5184564204901964,
 0.04920545927941444,0.6898217531695943,0.362951957032864,0.09833573490192682,
 0.4459964485554173,0.3423967671883642,0.6401623009106203,
 9.498225545215844e-4,0.5849096438699376,0.0526286069888271,
 0.47339600816955196,0.8623680500404565,0.26046884077028665,
 0.7132258226448427,0.2568947970316532,0.6596154610874981,0.1917251408652162,
 0.12787028691396685,0.7269696682831329,0.8328624694810014,
 0.20244226977889102,0.300195165640201,0.29615438769499103,
 0.30690128935487604,0.3622574082626122,0.09364689081230071,0.719756612914718,
 0.6881147610496334,0.47249747442031276,0.2511851755583083,0.6986912288519304,
 0.3513383432688193,0.8415922533209357,0.16838687096237848,
 0.20355064486604013,0.7494670845096318,0.5170971725206307,0.8313695292345099,
 0.44735132826911916,0.42206816365773814,0.30009555656784637,
 0.22626130394400812,0.7157299179031726,0.26948343801224295,
 0.07460412952674655,0.691454411896089,0.9248029273331563,0.5514995002149666,
 0.14789619699402523,0.9178022224434995,0.5814885741808267,
 0.35501188978282183,0.019456038803378317,0.21266929898024267,0.9203318059902,
 0.7739286055195106,0.9917256621658925,0.543971366035657,0.6450044263902978,
 0.3289767010610398,0.4789473902984579,0.15720196381883234,
 0.47481619460506264,0.39205590697329984,0.4342295166379635,
 0.41611315438375907,0.870080693311737,0.8408653605183896,0.9421753914680774,
 0.036273736052836236,0.20952671152974722,0.4247445032292194,
 0.7088915561200312,0.6015266912992863,0.19583495011553897,
 0.08506215057302402,0.11557923570069228,0.9171794855649863,
 0.07613451333479171,0.9984347360677772,0.5287238510977413,0.4971217776947484,
 0.46657500577181144,0.35944279763353126,0.6302565320447415,
 0.46196689922523504,0.0866633828918737,0.5819276310658266,0.5704948545940405,
 0.7828390440709936,0.8517580007309914,0.36816300312962513,0.4470882859634381,
 0.22362991300307278,0.04159231488969206,0.5142189016149444,
 0.3370548479197455,0.2894943705589159,0.44643325158059355,0.7800138196309933,
 0.1784645952653624,0.18430696061804186,0.7388723198967186,0.467162180434312,
 0.9611358767720255,0.5360784190413207,0.30154676700598637,0.8520059992730453,
 0.5678728910597508,0.9039413637700978,0.6603183012035758,0.38866403201942146,
 0.8182556588609935,0.605026953110597,0.8350664658503281,0.9235368092580224,
 0.12772553463344627,0.8175666673176945,0.4780406349485804,0.6016133452048991,
 0.6251567541047578,0.3693835028267436,0.4350985618602954,0.6821101252105077,
 0.5516152277252893,0.4893286963531577,9.431607401880759e-4,
 0.8670626980072548,0.14135683562759693,0.2089195342085438,0.4194037983995713,
 0.16657432649711945,0.822257517322898,0.5483045403497236,0.6536319301923832,
 0.004248236518528015,0.5861708526974791,0.4881962961084738,
 0.35037133394805886,0.06579184591121912,0.4802676368306841,
 0.3857710938451794,0.7359012530238593,0.6256462202925124,
 0.032226301461724005,0.44890335706344375,0.16491999876095542,
 0.8887192133201467,0.8568929290742356,0.4423306462612988,0.3428415259385479,
 0.03988501727450211,0.28830906730249295,0.9561355261273726,
 0.2725560245152516,0.7409065719943629,0.42190246836514633,
 0.02190624883691683,0.4539980947703094,0.03007085350299532,
 0.6069848240951163,0.8104144425360251,0.31802428623901524,0.9224112210570456,
 0.6777017148920556,0.06983299609214955,0.6961077730029486,0.5808790089166151,
 0.47028974930590073,0.6447456123744276,0.44910022512025827,
 0.4312018315100411,0.8247376999440784,0.5613581541806376,0.5110245089458739,
 0.7740667480837197,0.5891778118808152,0.133622125299934,0.6007954133770057,
 0.8613361230892471,0.9427766499570466,0.33541942019702187,0.5499731979974068,
 0.4358473862283868,0.569563043395657,0.822910709166503,0.7022180587750462,
 0.47514758074538244,0.3744355825608522,0.6172539392816787,0.5655890734418569,
 0.8653206712423874,0.6636026228966659,0.31079423962956143,
 0.12922147713433418,0.2503739464121584,0.7040664942920777,0.8698673420009022,
 0.6237300975781009,0.03357377685522178,0.7896258121047212,0.4214666347628867,
 0.9232054619138769,0.04873660778288258,0.757614739730152,0.08813519859444119,
 0.04945863500449865,0.9992707560242968,0.5119177111419997,
 0.27278073519593227,0.4162850347964251,0.9326273349629288,
 0.34380978575476107,0.9245289947662887,0.7881968433779744,0.6524685200846946,
 0.2816065276161195,0.031012813223751046,0.16228842178908243,
 0.17690553763052563,0.9822360053327963,0.5304148343398886,
 0.20714845016215822,0.002656561148385417,0.563306693074744,
 0.9318696929232443,0.6269226964999226,0.8472289346297255,0.25263758951666415,
 0.1921913106338925,0.43787766051821175,0.8839866784200207,
 0.04552699286916295,0.758536160477429,0.47242364284514726,0.5434803140061636,
 0.7061806415845318,0.1608179318255114,0.020359482436472343,
 0.7962169047267457,0.6970103527217527,0.15817696838526973,0.3376230055231366,
 0.34502953250305746,0.2601545696932691,0.4381597746586724,0.2503837959921206,
 0.902354692470945,0.5046188491238967,0.26040926771547435,0.9069289171789667,
 0.7530558030734041,0.37459644182120067,0.2651615277986892,
 0.11529026319862767,0.815812235645788,0.09111083946009002,
 0.20896983831076998,0.55710420571297,0.7983061902799745,0.3747580524679217,
 0.03517171636750471,0.7771582373935845,0.8190893920100667,
 0.45556757016651117,0.11295031164162916,0.1902182117671518,
 0.5764409986095098,0.22580434367198998,0.5670561730238308,
 0.13540238699349672,0.03902902973571665,0.629062092344907,0.7705944542132203,
 0.35879342639808687,0.6821014158221885,0.18399634761850592,
 0.7462452980826462,0.5322086926889946,0.6364833405448705,0.2382910331165251,
 0.5593065473831177,0.30765796755089037,0.6420351645225582,0.3824668833333824,
 0.8732037157383844,0.28822836977377997,0.8639801114833777,0.9222811209880875,
 0.7868761946806945,0.9126890561439253,0.8033158643413103,0.7483415405526815,
 0.15112796484858937,0.46455762887201324,0.3766688583897142,
 0.6072843206545886,0.2914760761763616,0.2448583619129836,0.31248907620966415,
 0.30074244853232934,0.9799697601150271,0.4773267044327899,0.4418214000164129,
 0.629412786258693,0.2076257178115194,0.17847045756227842,0.7059448849046251,
 0.3250789327906636,0.15865071670984576,0.9881399836816275,0.8405320279799704,
 0.8721221568469263,0.960009779120881,0.4029920055431613,0.7272653082379643,
 0.491618354958143,0.3724306329328001,0.7348547088746752,0.8521306286720511,
 0.17675653967447857,0.7383317410855986,0.10540469158431043,
 0.4358410879887612,0.5065642045134151,0.35376440079663707,0.2357083593904864,
 0.028046731316514473,0.7748263668847315,0.3077802466459417,
 0.8218759778417641,0.802106722172612,0.2792950917734124,0.7170629562962576,
 0.509484526150132,0.6329254853758122,0.7330632302532495,0.5463543058996404,
 0.9341394429124336,0.41781963810828004,0.8956246174232044,0.8331130374134719,
 0.38290228860731435,0.47780099924387265,0.3384750740051078,
 8.430159578926011e-4,0.6523923760221483,0.0446918452087226,
 0.6138911285888553,0.2390108440818245,0.006584640089136173,
 0.6867603666360229,0.8425079582651482,0.9260577459108491,0.8551003153564871,
 0.7301463206869061,0.008549605235285807,0.3024205832904697,
 0.05087361079855246,0.9281722539665173,0.810826668862549,0.896548348818913,
 0.061978225903884576,0.07061389806325724,0.34189572511266075,
 0.2942701997799697,0.024067337202686945,0.5819924785068677,
 0.055096470001145814,0.2481292465081223,0.8451849853742458,
 0.7357254483189595,0.0022759676558050668,0.6999822646959627,
 0.8096880972259791,0.441165893076641,0.3590850603802579,0.5437207990069034,
 0.24606154619991705,0.27181946681577995,0.12013409327280644,
 0.025750657772780183,0.8194207163860424,0.745900746055924,0.1911644434989097,
 0.7250082294898147,0.8479801405028534,0.6493541535312249,0.20434431474317982,
 0.763418337886184,0.6463990133140839,0.2224555750879711,0.5925159306058531,
 0.6750382870798594,0.21594545875625337,0.25772205884601374,
 0.5385591998341284,0.9394414751337586,0.6881633138843655,0.4626119226662406,
 0.937679936072357,0.07852611595970094,0.6164209401168774,0.22545433991398678,
 0.12707269348500294,0.9296407979011562,0.1647559576721105,0.7768493641382777,
 0.34073496126188785,0.3021480308676112,0.025148843991580683,
 0.3102614499039096,0.9040424579360329,0.08432876461716488,0.0737934760402521,
 0.4623918329541653,0.9653331900376011,0.5155783115836355,0.07476816519221907,
 0.08333283571916572,0.9778427648962786,0.1407767144997737,
 0.15932646270850137,0.5097483875321435,0.31745434070576395,
 0.7627854345491607,0.8990840819083923,0.6329680148381617,
 0.016909364279523054,0.12019147221915394,0.8205173208527197,
 0.8814407636763635,0.19294946584911932,0.13330547681380145,
 0.6358938786968626,0.4095223022651858,0.6901469195829726,0.41676740102536314,
 0.08516826634143937,0.9334817101118511,0.3650555113317807,
 0.12129659912348933,0.5076277059933761,0.7880105913289321,0.2887019527997703,
 0.5189586367311797,0.14724026817992109,0.535883833218467,0.5302727946685108,
 0.4292067103349275,0.5720474010042147,0.5696616245909611,0.11621433582476515,
 0.20972468944803935,0.499364879934067,0.5009640675896443,0.9677547865911031,
 0.016492520276969236,0.8165110590171056,0.8113775866562045,
 0.6911515335613365,0.12533512422846815,0.7615795758655115,0.7729185591624742,
 0.7522975197884534,0.9488814364486247,0.937780142562381,0.8704013137964981,
 0.48444673981856134,0.11269027390002284,0.8948423800231321,
 0.30961298371007095,0.42428396210733665,0.3832376095811184,
 0.5672922037358084,0.3069936890173177,0.4135240811594675,0.6232267594770238,
 0.49716644687620604,0.712314753716814,0.7801404789726858,0.2411748447513351,
 0.9000494167452945,0.384879064540118,0.8597733904813283,0.771068126941099,
 0.5371024630482674,0.010308167124501444,0.0041964229005728715,
 0.9533916704812113,0.6654124890787096,0.7027704523704842,0.3228107087997836,
 0.21117650830677204,0.40869830675643015,0.8985807792411371,
 0.23540063191357208,0.9869191196893343,0.047951769866541594,
 0.8455065606713796,0.7360482072151219,0.04338577847239278,
 0.26428158709357197,0.5995737154512744,0.8113844867167503,
 0.11567290657398022,0.7980625991970556,0.5778069551875556,0.5817772612420762,
 0.9287780381706168,0.9010730102993801,0.18984966868687803,0.990465692247636,
 0.09145548678356763,0.5483870098025299,0.611688357893751,0.9850053596913242,
 0.7322963787261089,0.14392943679333703,0.966637157933878,0.7610927131372218,
 0.09774025011958509,0.6201563623188048,0.7084142171022906,
 0.32713626819332575,0.1561152021960297,0.8607193845696064,0.9001852891536796,
 0.22730926737912838,0.5046767340455847,0.13874956247346737,
 0.7190662068759451,0.5948355709730694,0.5049642611869406,0.0778418278007853,
 0.12031158237036776,0.07517214276516793,0.10383943730453349,
 0.3638369861006363,0.08937648661777509,0.5968690753453836,0.6086227838210271,
 0.6560233973619161,0.47395175881139184,0.4570156123158423,0.6066528392179195,
 0.7243421575990641,0.9228648661892553,0.84768451981587,0.013992079011392877,
 0.4741706510737571,0.5488462981260037,0.10968034175914998,
 0.029122375822624558,0.6180108431380418,0.8505917822336321,
 0.9450477130005666,0.2286931892612083,0.8468794721701938,0.8672448530398852,
 0.6566965718140879,0.4111930722277207,0.42058724442865636,0.5759038101428605,
 0.8678730975718603,0.9076983571394801,0.7883882782892747,0.39417069028346374,
 0.9563291252160904,0.5684577993900962,0.9244744146606714,0.1489636895975157,
 0.43785856909990817,0.9791492447378564,0.28316455113013594,
 0.8372509817814501,0.9360229070681418,0.4458225095376047,0.33671294458035905,
 0.1785380276094286,0.02314622440319647,0.21140757257329357,
 0.3294436542320134,0.5732664593018858,0.16707476938833754,
 0.43643023208008014,0.8797842460258225,0.6653223908073749,0.3577822610237559,
 0.3642292977157361,0.4200719631401292,0.8408330101590203,0.10740069422594922,
 0.060250837176147654,0.1592080241854672,0.6068194688970848,
 0.11438746364099917,0.37039006529913177,0.24144126258075316,
 0.35062498398754116,0.8928000333127407,0.5584844663834027,0.6750041947598318,
 0.05745396586236573,0.289351290838738,0.8436424750730208,0.5805640306825081,
 0.8654334635559964,0.560400200442888,0.977952618364739,0.04475314798131813,
 0.3140610983866925,0.5479376790566366,0.3797146802719531,0.48975777286199473,
 0.5294439723588609,0.9444370501844287,0.12387005809269791,
 0.44024487424475245,0.7188440002811943,0.20504087657026804,
 0.9407119702670318,0.5906644166606527,0.13132656506542773,0.2962089736321212,
 0.3951116920890946,0.3431969304386896,0.8223239982467316,0.9060384517806853,
 0.6545856792216511,0.7045153246232095,0.8122766851199535,0.7033576035710286,
 0.2039460590253852,0.38910438280927273,0.0868690562675678,0.9957240677701631,
 0.22183195872689476,0.8765159164874711,0.10872899230524136,
 0.7133648290872099,0.6467346834029724,0.6821656937519945,0.8701307293108904,
 0.8587697877024896,0.47098819927596813,0.16879946968509119,
 0.8766000918753298,0.9008517238012016,0.8749004404979258,0.898691345159011,
 0.15795514657654974,0.33541810512442927,0.9956630760516026,
 0.2741434030899482,0.44219458922508414,0.12135729008741425,0.63809233288907,
 0.5604801628269597,0.7424424856534166,0.4561707822064627,0.4984629247481318,
 0.2715767316365414,0.8995399017772727,0.5711753254790093,0.12264964964284986,
 0.9207976747299921,0.0019660551764585366,0.9730445482481296,
 0.5050882439753709,0.927402962356616,0.9743609431758435,0.8049931134981937,
 0.4031869717148452,0.13743764648018764,0.6577962652409836,
 0.48196507623757323,0.818603873798864,0.014830130712403244,
 0.33895398661875287,0.5912754261260195,0.6082672454176998,0.8482063808986292,
 0.6286154447388571,0.9520124261467688,0.8426026522497382,0.12705051023325453,
 0.7358029282047189,0.8428633681641349,0.5527025967768087,0.3970313032458057,
 0.8359665565638476,0.5839294948453082,0.7333231511371622,
 0.047734472131219174,0.6586651120575908,0.674869589790529,0.5599836850296785,
 0.2897126528196128,0.18925454863018143,0.1562789423090919,
 0.13123098774901543,0.5428628528522323,0.4402206455377592,0.3452085509608771,
 0.05567336446174498,0.5995288592233845,0.7918601556491893,
 0.06071017815325652,0.35744339014812254,0.24631616685903135,
 0.5356638069854558,0.06085156260731972,0.7315913429789952,0.6050243163670961,
 0.2421888111371414,0.09288186013131217,0.6209030745723607,0.9607439279193314,
 0.049714719642549476,0.8758790663769427,0.3521118898763489,
 0.9801742644827245,0.6978093157204451,0.8298671909005045,0.17889882176169092,
 0.2712525777843502,0.49389086078110533,0.13937373930293862,
 0.09209469988687435,0.16646338791618276,0.12145317293016167,
 0.3827663174095697,0.1020178288220005,0.711489954094818,0.8096072308483412,
 0.907720805564725,0.5926801948385118,0.8831323345168718,0.9748203623163303,
 0.4305603059389057,0.10858918940480655,0.9042431557265183,0.9265447227536159,
 0.36844018421008773,0.2749371907228454,0.4772848841867763,0.9159759698060209,
 0.21048821674498186,0.22097637781922652,0.38598949243610603,
 0.06735544851422648,0.4218379605160688,0.6890408739413739,
 0.29867606464358853,0.2706613809477907,0.5548561690214291,0.1838853493238497,
 0.5021285441326055,0.5434876865628643,0.4459301468128437,0.34167299447261246,
 0.3103311802733182,0.5618550415058782,7.475697885764854e-4,
 0.8727323496473912,0.1712543320414901,0.6885292605807581,0.4320041373350094,
 0.30366980298126034,0.40310760467990414,0.29249359288496957,
 0.6300569718726599,0.6363935965180159,0.9480235670679031,0.7420846684395725,
 0.5278192709565133,0.7869488050729659,0.18753216053823174,0.6858416997283506,
 0.15166816229249214,0.2011161557026251,0.5329034402743552,0.9177617851449563,
 0.7537635902141708,0.08872088461516281,0.35662758327176847,
 0.4010319162569629,0.9660554361076337,0.392822412398082,0.8439877878269793,
 0.017727070064772255,0.34430769701887254,0.8225697409581153,
 0.12790478446722864,0.6408814718692938,0.8336925987429241,0.2164187940311436,
 0.9252052044353188,0.7849676453711231,0.6003929623545876,0.9141440712908957,
 0.890250860599044,0.890348096681124,0.15487818382053664,0.9183693916622248,
 0.5214388741112387,0.9434531849369328,0.2944971465771382,0.20994952052279614,
 0.9773248804025748,0.4204837270168187,0.27849413831310177,
 0.15011403512377597,0.8440783959507134,0.42379922916275514,
 0.8338814043908852,0.4718336402561555,0.5916394789014118,0.7864239447522422,
 0.42402114348143083,0.38991701628790143,0.443042969158274,
 0.06845054160529873,0.4516176138927339,0.18111712121557633,
 0.9357738577979493,0.8679900892121719,0.8622974596763504,0.22235640196845108,
 0.41982853068982895,0.5585908330115172,0.058646190965088496,
 0.8570583016275553,0.8792457830532825,0.9565763492072672,0.20001227709829394,
 0.28722694235697244,0.7165647889064839,0.29251574871643393,
 0.17312816079635607,0.9111585903191549,0.7696650747575648,0.7177587122413908,
 0.1100940381093769,0.11555340861466157,0.5475265245737583,
 0.12097219707241824,0.44466835777800817,0.7244718875487988,
 0.22096317735232396,0.2143904033331836,0.06699949721005949,
 0.5402258995997011,0.21873891581179838,0.4510419382176374,0.7384338299333141,
 0.08905365031571477,0.23325265611687307,0.9431669554188822,
 0.6708972623118804,0.3463434214680329,0.8881924399258434,0.8732316021251922,
 0.9208111715279103,0.9884920619105229,0.9916768512356691,0.19375981793861397,
 0.4756766677980888,0.9268294759967978,0.35526572245835075,0.9120716554785468,
 0.04274480872368269,0.20303425715094237,0.5749647957251146,
 0.9562730334063135,0.033206901941776334,0.6129311984629447,
 0.23710289388562822,0.20908377795575994,0.28637750385392935,
 0.8337494007049147,0.08316157539076219,0.015611798729464832,
 0.965140390148957,0.4335676961151136,0.7239385621045038,0.48034572224220884,
 0.7569702661138269,0.9635067462676818,0.33492334696596837,0.8676111292315898,
 0.6482468883573046,0.19570495891975503,0.6430438492516973,0.6828813109026459,
 0.5962340072091429,0.8785008897632265,0.8288774282261935,0.17320244385348393,
 0.8378779390575747,0.5685999526022533,0.7220534379274158,0.8866950962772843,
 0.2393891188479017,0.7366897469208933,0.5228863401480891,0.4137111126997126,
 0.4126330153658072,0.05700865407502331,0.6584425431446688,
 0.31136437080191937,0.7716166953258113,0.9157689004832714,0.8873212973997661,
 0.36436032701943044,0.22641023289913276,0.572336277097141,0.5518428616183244,
 0.9522289813496231,0.39951764048775285,0.8152687745404303,0.3461417675638342,
 0.5999191575863572,0.23618771838120012,0.6453164388269141,0.7657539429981991,
 0.670460906373858,0.5568123752422467,0.04431527987573469,0.44304704860615485,
 0.9171265846709099,0.8356028029060542,0.9628251269193685,0.5549862535003015,
 0.41510830173864566,0.761539041039244,0.40517099620594277,
 0.16257975518053192,0.7448765513175462,0.38269954458901345,
 0.8324857411032047,0.3812806930682041,0.2489062690248388,0.5246714306742476,
 0.4390348113493665,0.6713329457519521,0.9943703316851206,0.4358916653941197,
 0.1941824713628455,0.4151545922003793,0.6101556738640496,0.5536347021279525,
 0.1496447535116604,0.1079692110428474,0.1669077232396713,0.7052304714749892,
 0.5019482380659214,0.27366764788701425,0.17698311114770138,
 0.9470205493418443,0.47536396356332145,0.9098748237188443,0.6702786717338904,
 0.3113250216394664,0.7040259023858306,0.008553485634577584,
 0.15848484419816444,0.3623225394840871,0.7600858012143525,
 0.42477692321318283,0.9624516332391795,0.3291398299994793,0.6736366946802171,
 0.84567248090713,0.7045588800424408,0.9485209845978034,0.21848739826552976,
 0.7336411097547583,0.4251794316004742,0.3415853328811166,0.201742403780664,
 0.5013297383996631,0.42715730910146743,0.7067954641371607,0.4136208316855088,
 0.838749299751772,0.3707491630628804,0.2531003995760134,0.6008426401444322,
 0.8639956939968346,0.154752064911571,0.7631172068894345,0.7886323025190227,
 0.2753332577376605,0.33833223406369206,0.30531875371578987,
 0.2979185558292533,0.6310666729311528,0.13478514760415705,
 0.23107031504741216,0.6732683251443509,0.3377815203525676,
 0.44937909030817347,0.026379208916123442,0.1765805122728601,
 0.41401780569933244,0.8987696818056874,0.5116985670869563,0.5784417868419738,
 0.13492006732475226,0.3231166056284641,0.4657395410874632,0.2635515420874467,
 0.04251329547585547,0.7224220834922768,0.5365887870294264,0.3907506433079806,
 0.38229270441612195,0.3298836529865272,0.6617675706056229,0.7948270181353108,
 0.009776497943567719,0.17899338253362007,0.9883596132586643,
 0.6952821715985525,0.929418340708763,0.16871709580637673,0.9064129498391511,
 0.012160688643598538,0.171692072469267,0.022227586016637657,
 0.41003128117117493,0.1993869774948871,0.75982071717733,0.5930287439848482,
 0.46727089514472897,0.5955059138475689,0.6911556239486344,0.966689043483929,
 0.8828343403121688,0.929278958831796,0.2512760403774754,0.024917302781899187,
 0.22458727443578164,0.06923816350338441,0.014778740701742898,
 0.9554856232267782,0.7080478215141564,0.8130204369245524,0.7851496153131057,
 0.6953554348183536,0.0202183127446951,0.26781153568950344,0.3482262481050491,
 0.7576354154849873,0.7242903428612605,0.4685166778594302,0.07252988558681028,
 0.21452816472660896,0.5914821585090054,0.689363546577301,0.05973840287609267,
 0.793368399239799,0.19136592054257717,0.37961768873354307,
 0.14604170533082617,0.301212516554602,0.36562763855028135,0.397210801436377,
 0.06604732002500069,0.6309724004171964,0.34535091588901823,
 0.3804231381558213,0.9322453552605188,0.5823266740214826,0.1213795551825354,
 0.5903851998756482,0.758909972133173,0.12144337585148968,0.6022170525741675,
 0.764898932434603,0.2323035148166127,0.14412904615401623,
 0.011048601009317482,0.4555940741227007,0.027741270456891454,
 0.1441892276972957,0.9715868405252834,0.30020824017534764,0.2775168426446998,
 0.7533155317724402,0.3298118028525209,0.8928173491932541,0.5797670530999249,
 0.7897977288748617,0.16734178925828547,0.057555215287257044,
 0.7454091962509815,0.24109609619809214,0.7541067374254908,0.6639874328169388,
 0.07342554830119008,0.37689138719803894,0.15943678117452276,
 0.7122602474840525,0.34862115660515647,0.9732987447482132,0.3274035714513541,
 0.4724318838290944,0.14176230740772988,0.8191715236415049,0.2622644804956733,
 0.43018184971637474,0.00607684322236568,0.3189929862566139,
 0.7910608581818853,0.9353154910942211,0.4048323336780022,0.6194903304991897,
 0.9294959805129801,0.9004539093219613,0.12804183211258757,0.9156739543771879,
 0.12453477708207916,0.1342200185207102,0.17996668687704287,
 0.11202241795801748,0.25417625101219454,0.006652438070374633,
 0.814173494671713,0.949043401590006,0.2563924121954432,0.9782296657864233,
 0.20869224350037552,0.2986991839605768,0.12341879079960749,
 0.8014505879180057,0.592513270228748,0.06917272039089473,0.11827298093846583,
 0.894032828419339,0.3520180411418149,0.39722269553275036,0.9814066111316079,
 0.9112713982818558,0.3162869887660935,0.6403166925482356,0.2913005725821156,
 0.08951644523064517,0.2662609297406333,0.45064103174595627,
 0.015601187991780563,0.34304518325647204,0.4215991286461479,
 0.6007350823927592,0.772109394589798,0.002065393601834886,0.3274313492229639,
 0.97212989351112,0.9459484863683272,0.6552214480369598,0.17795195292408728,
 0.29697710444870384,0.03121891244246533,0.9069565331376135,
 0.7037649291491153,0.737990631411777,0.2291357443322315,0.5019579079617239,
 0.6479916537172085,0.2724061534086215,0.8045609255784474,0.8845313272654146,
 0.09130139700044315,0.019473551046952253,0.9986872585551566,
 0.20691953918754902,0.6916920388189844,0.38920254264872123,
 0.9289866578567377,0.8434537030706737,0.4954535358924447,0.3027774476154349,
 3.2669010917807384e-5,0.21487701449593566,0.0035552767321336276,
 0.8723884501636331,0.2215857408903844,0.29972270300831505,0.9255784456014573,
 0.13844374352243305,0.6854419167914094,0.16880053143394547,
 0.5272876322464872,0.5380555354294637,0.6130659943710887,0.9898087182601558,
 0.47825802654628813,0.43497293888865707,0.7822258483752856,
 0.6595347384172349,0.48431121023702683,0.9527663894090429,0.8818342439449457,
 0.16212756811992646,0.5105139425733014,0.7434685088080808,0.8972765138442376,
 0.12574290381209408,0.28427098779508075,0.6195339782041415,
 0.10163096965541563,0.017203903849362434,0.6929384523939769,
 0.36261402245891416,0.6036127124657472,0.15250745715868808,
 0.6225557499337341,0.2216367123754528,0.7301644757345467,0.47037215631112306,
 0.1597870795555345,0.8120201585048094,0.5902165705804895,0.09748011588077665,
 0.0642787772343838,0.7428948950459685,0.7145215309639571,0.9922176393220308,
 0.7234423755788059,0.867786461767277,0.14092560227583983,0.28168256694930927,
 0.23276717461401786,0.9768044984635017,0.9378448459923302,0.5227286216830699,
 0.20762060841139762,0.6863682878463743,0.7728975265130239,
 0.08930017804431678,0.7593111068922082,0.03580744551643533,
 0.23914625625354136,0.7923937710323169,0.04285042615611945,
 0.9056408064264029,0.9481296227780508,0.6517024677799307,0.08621114270942432,
 0.7767390926190646,0.05848604184509998,0.09841087142378668,
 0.8301365924401573,0.3505599979347145,0.27882618677897153,0.9780150037508032,
 0.9516829593892361,0.05767259296746352,0.1036306493143746,0.7221968875798875,
 0.21321054165544506,0.4816884586872353,0.42751530480951627,
 0.7760549060076574,0.49739730156961615,0.45204653365556624,
 0.8991695110411436,0.20487285522665677,0.09244820840164314,
 0.9990080321338204,0.26760315709250304,0.009761519495898319,
 0.32628181035965553,0.841458018564424,0.6121405465799173,0.4650998369387802,
 0.7136042551167554,0.4922608305238835,0.9223923826691555,0.0545773022875119,
 0.359759675922134,0.7240469931338964,0.8133086117662242,0.7585693053265535,
 0.3162159322328968,0.12049741442453443,0.009797376335340724,
 0.2393467612043717,0.645737681099564,0.5854150405701416,0.7387702576654035,
 0.2815221955923356,0.7553681674289833,0.45131714339990736,
 0.15847886882873374,0.6209743929017578,0.6884211388364543,0.3656779248554931,
 0.16083667761442055,0.5368539683496008,0.23973971222370005,0.831289698489687,
 0.8344183719541444,0.8139201548387451,0.3313721630172636,0.5044515159160153,
 0.7056425067889122,0.5009604163416677,0.1485267614504302,0.317674197064575,
 0.5733906708072407,0.020897059487437364,0.5206120793301737,
 0.8283665612554638,0.5375018476514567,0.10191114768249099,0.6753454945529552,
 0.39514263026406593,0.12505677139467497,0.8466873634363525,
 0.4845153381785603,0.6036229778118511,0.9569681835897513,0.47810871040596825,
 0.7210736712377606,0.5055905140063252,0.3551688073559961,0.8791686172583213,
 0.13832410488329172,0.16197248327069436,0.35329034610086674,
 0.05293121087803232,0.8192994546454583,0.5194126757484953,
 0.30982001111978397,0.4288611408832864,0.6845237639388905,
 0.03571202171412258,0.3312453641565186,0.6233830596178745,0.9352962705327332,
 0.7191377076334937,0.6850568830914968,0.8433042636809674,0.643582811452413,
 0.8367397049279097,0.9508064705583842,0.975406682307982,0.22022129455686412,
 0.46206497149086756,0.08034483292166861,0.918356619243897,0.868040339585819,
 0.8196774872737013,0.20677851643595568,0.8188245482096855,
 0.30372997287371684,0.715102932343042,0.9634011044430426,0.37843105013155953,
 0.2850556927655773,0.2838331419966693,0.7383239091158116,0.8734112954378895,
 0.42045273439222974,0.844817497973406,0.6296375961921721,0.12205880549141435,
 0.07093960811072075,0.1193440724762912,0.21967835538082314,0.435193394736614,
 0.3608372838753906,0.09098513553312304,0.5494901872240358,0.202420853463247,
 0.09713794288994293,0.4684453929024983,0.39192563732762964,
 0.8680071312258678,0.7589199918079699,0.6250309324790828,0.20583546863829205,
 0.5246819853726545,0.06978748388674405,0.7930496454760465,
 0.009789367422076256,0.4986695339767655,0.008356120788399757,
 0.9279011287455355,0.10139985836744658,0.6150096262336584,0.5305391031439288,
 0.16237264390529182,0.10725043224733255,0.8426710424479629,
 0.23297281500929012,0.4131729296570751,0.6580740284529399,0.7374164651312477,
 0.2988966460242207,0.5797901867314992,0.1125102901233791,0.45604390821558405,
 0.726125996259382,0.7387403704818076,0.7250949444963815,0.22934872694886643,
 0.759461150369992,0.4191365130236788,0.23431793690244074,0.22875898119335758,
 0.4682518196519695,0.24286013720031263,0.12431349391787383,
 0.7426816450045027,0.8597237546289066,0.36671739093926825,0.8323817429088198,
 0.8848778003600006,0.2430995838301866,0.5262343127212012,0.8301956016626835,
 0.5072091076053982,0.9187442213254441,0.9654620518643277,0.8247421560742495,
 0.2593948486316574,0.19452041223612992,0.4588817179518182,
 0.07542152569991256,0.6227120325595877,0.9830703775032293,0.9813698377700781,
 0.7368274405056049,0.751292789310289,0.26813321669418544,
 0.004471509160538845,0.7067159777673031,0.6446914368021219,
 0.2317371299756118,0.6149362765777895,0.1719291948242907,0.4781997384163985,
 0.735723096016884,0.25594866130396776,0.30692704025962003,0.7612370988837512,
 0.6728529047350585,0.9718220274290251,0.9813972853826387,0.8300053770368415,
 0.8360836892583721,0.34514814416970174,0.3626544127412069,
 0.23956603013305433,0.4483903292054283,0.01474968789084008,
 0.011423503851425765,0.750596861669412,0.22739184724264028,
 0.6094022838973933,0.422610547967733,0.9283398587225901,0.15343631805486657,
 0.7804840638547782,0.6452281984965563,0.7704271124243831,0.21024816898406795,
 0.6786297613848874,0.5816966153147686,0.33094239688194393,0.7162094701863149,
 0.878685494887319,0.35219495151977853,0.49845854083741636,0.8690875266851706,
 0.5270351022903618,0.22509197173737827,0.47961928567069345,
 0.8194122683051835,0.8412150868059273,0.06298879175692429,0.7274826370974075,
 0.765140343590995,0.17966592271228365,0.634670238613619,0.11021775531641365,
 0.800526279894037,0.8943631789591342,0.390169931648382,0.02767178066143372,
 0.8405679589504532,0.7132708817590888,0.9378667686427882,0.2739415382701871,
 0.884198254667746,0.16634536269005173,0.16215302860601288,0.4501846292998145,
 0.22622585927999306,0.9882048236204273,0.966470104136543,0.7486389344644244,
 0.9811354095198228,0.9623717488227502,0.6668657084627581,0.3957515009449677,
 0.46894953121332517,0.3292488144684049,0.6406451209333284,
 0.07231043892007283,0.8001597850747688,0.7372849566252664,0.136136135411558,
 0.582419898484627,0.09002707043616898,0.0428999821827466,0.5756490452492747,
 0.8734013990532287,0.039554676874920314,0.17473921953160632,
 0.15578964043552423,0.12211271122912348,0.88522658063735,0.7481628978539892,
 0.5749223301067038,0.47946131215876364,0.3897356905106717,0.6452187508626315,
 0.7270113910292418,0.16878408142779444,0.6891121182434687,0.7993891004252969,
 0.2796490817192059,0.9305389259059422,0.6561622237820993,0.2760308417602909,
 0.3963890297530387,0.041446850161687476,0.30706605502701145,
 0.5644051998480974,0.6294871252414512,0.24723461981517203,
 0.24350019652707244,0.8205600241696399,0.6380926021468326,0.9446516265177646,
 0.7676432407087632,0.009692561397663835,0.18984773087016826,
 0.6686820480147113,0.9548563950489918,0.7029330997054679,0.5010112984254891,
 0.19394657839945995,0.5567474206186097,0.5369225339817705,0.7036416917871727,
 0.4093411826395952,0.7463746489719199,0.4371955973689252,0.6095673018088874,
 0.4934036873891551,0.9398071716163187,0.030628894060683587,
 0.43665871070803464,0.20157107442946365,0.9218610568653904,
 0.41084695808702953,0.5376725044500772,0.4898303230416826,0.2646368369353721,
 0.5163230512035767,0.04445690649523093,0.32891694086926004,
 0.5453837218967372,0.4064352442086405,0.8209712199194535,0.7149863515188897,
 0.5415060219921917,0.6293938447366126,0.27884465555564364,0.3286345927719144,
 0.21570884016199776,0.020123449226288392,0.3709156180750671,
 0.022045248453991362,0.7281755530657156,0.514966729032002,
 0.08837558725135974,0.4044316652009444,0.8653257490870704,0.5552877575456816,
 0.4974040588289108,0.7044442348180108,0.6988125302208926,0.4441680444065105,
 0.3919827119515953,0.515538497812436,0.28771287767210807,0.6483730912807809,
 0.781646784237297,0.9753475582865226,0.0572659761903368,0.6805159084945638,
 0.532445408127178,0.322254820985024,0.6181132492069784,0.39368174450065874,
 0.7809326596853235,0.20654028749334064,0.11535440697965382,
 0.08468994316434109,0.410557340123368,0.5453746685416553,0.26491123860225607,
 0.047635949090766405,0.21018766542154743,0.13282916696357494,
 0.2419632387666072,0.4425374754905351,0.15899589085464783,0.8195518535592519,
 0.4559863743929867,0.5509908889246488,0.2922145175857509,0.03477076838043969,
 0.6477986942193842,0.4583958273668487,0.6357234576886033,0.2017876509882658,
 0.5648711525629656,0.6606769552307054,0.41538356423932776,
 0.42850260972436505,0.2579424172108591,0.9525163697414818,0.9846636508138008,
 0.41613018639710597,0.5232087986215125,0.21087258065545167,
 0.014962445851066541,0.7516271666708633,0.7231001539129729,
 0.06503341051421663,0.3831777061764673,0.21326627583678337,0.912032835157464,
 0.9739001869477627,0.5379707877623015,0.5233221780574446,0.2504693966446916,
 0.7080714441196363,0.020969493979659304,0.4795558126929176,
 0.7744436447799434,0.872562047563731,0.27976035643745756,0.5622445603064444,
 0.9962210788002087,0.5407062703738842,0.6290391770763918,0.9952911003813631,
 0.6373466652828201,0.5745535454867811,0.0351024589633695,0.5667110837873257,
 0.49768223318339466,0.9095872384093757,0.48613513986328927,
 0.31270792929395463,0.8964505424821069,0.6752645558734869,
 0.13975524188464172,0.5013283029845655,0.22698588360378225,
 0.06273585152249617,0.7167108785473265,0.6756471022630424,0.3498338944062134,
 0.7785842031702204,0.4823370123190911,0.6955875246814566,0.3441955334621315,
 0.7514448047897763,0.4729509989448113,0.711223707029392,0.6790615000143961,
 0.7521606492770535,0.17793176267957067,0.4300875597082824,0.958758033605467,
 0.4835673253095848,0.6962836261232668,0.08072464973481253,
 0.03649699982341981,0.022293936486855515,0.5723043775899476,
 0.1664446101160597,0.48931770251309403,0.9922150560490047,0.3481677722059766,
 0.38733243333348233,0.8503912191683624,0.002171810414564046,
 0.32764376719048915,0.9297518258797184,0.20349840424520416,
 0.04267699877153064,0.9644351916686354,0.40087053332325384,
 0.2948727486886277,0.42725441202231906,0.12407162235332958,
 0.8351807614888699,0.2721235422798497,0.7517113301962062,0.31608062842448015,
 0.07656824284319541,0.582622334123932,0.9153696007460879,0.6949964508365245,
 0.1073577249574934,0.4890927730742739,0.2899843825481868,0.7117014243709799,
 0.31947833343306353,0.8299702316887828,0.6409926222153131,0.0926737420036089,
 0.13846742348539698,0.9100478847148935,0.9718734653638476,0.5945169527557692,
 0.9205865957011564,0.6715332135627595,0.854155933796324,0.2627849784739499,
 0.9530194724687017,0.8317908946724094,0.06833452184853028,0.8400133257459492,
 0.2771414341556603,0.5362961117458874,0.7338081683517901,0.6581091186455348,
 0.8516110784238164,0.8388454231324507,0.3742653557702168,0.5174448556150413,
 0.9748710912044692,0.4505122517063147,0.041884041512855075,
 0.13385854526192198,0.63367462047091,0.598292289536221,0.18451306726059913,
 0.23420810008243953,0.3106965021749244,0.486471491302104,0.8353630945226695,
 0.7767999098713254,0.7442173348000323,0.969049121747402,0.6714075560719797,
 0.9313866217729065,0.47382618975846924,0.5578558569539935,0.8276511632547635,
 0.3962215388089284,0.12447182537155188,0.1041784084883387,0.8462760165315378,
 0.3094763912561075,0.3737563754004405,0.6064080642380999,0.9728891264136517,
 0.09823081077366269,0.9717141924717962,0.6442830160664853,0.9685545452201648,
 0.22937650301846946,0.4279100179205704,0.7884010734262239,0.6805316250095719,
 0.5482321663207957,0.27628362219737734,0.3829491799251966,
 0.07192639918771859,0.5650415408496787,0.639368032471746,0.1101870181214405,
 0.6463059987590385,0.33823034245209116,0.6491880330318094,0.9382815494981483,
 0.94696280249274,0.27208853461771165,0.7764504622779735,0.16096424716156488,
 0.15410127007378283,0.9586321177939937,0.10178288494023613,
 0.2000086680762727,0.24999894108373355,0.033364561559292016,
 0.41257084667273747,0.9728518875661673,0.5087096483586064,0.2069384650768441,
 0.9765229272169813,0.7178426203002122,0.06631495489797135,0.3161260585348775,
 0.04099295916693568,0.35280823901922387,0.5338529332139313,
 0.7173058010951393,0.7621427452950995,0.40393943204485916,
 0.33247185647057276,0.14559482498001208,0.38669225839194277,
 0.01263722406056067,0.999011503060482,0.049994289287090155,
 0.7181074157828915,0.01916568116715811,0.17488733973639348,
 0.8287976160582738,0.4014665752119868,0.0896436651328163,0.5186066491415714,
 0.2577149006089151,0.33599383409645034,0.03576888677558798,
 0.3284944823115209,0.6266727171171516,0.4465073065676036,0.39657732213912966,
 0.45097860623065855,0.9855667518428747,0.434277729834435,0.6914727702311482,
 0.7716157752639009,0.8632143347171669,0.2493490835117288,0.4665519280680015,
 0.7170172514224235,0.494084168185946,0.7951948104761195,0.6104071956238976,
 0.7984548041514719,0.03563310681141563,0.8097921693635984,0.7019848402923188,
 0.3017532256160982,0.6973130603838457,0.6190628433092409,0.5819603210808061,
 0.7165792153723232,0.695921054511449,0.8796884351920268,0.3774882503500828,
 0.4065618857624749,0.6361970319440472,0.07129936456611696,0.7623115903772552,
 0.6249045577838817,0.21644483402442427,0.19170598742216194,
 0.2735197715064017,0.749770807633247,0.3282810406656783,0.3446895894734726,
 0.609438989192581,0.7185271704716533,0.0599361544932534,0.6443868442261089,
 0.09783053102558892,0.5937803035379374,0.09386941323524134,
 0.7812666152075789,0.4394392580245443,0.3450778414815848,0.49728850242966693,
 0.13334420003390013,0.9205683110111897,0.0965947482761016,0.324693707822185,
 0.20171488380418467,0.7852905115641676,0.8981152125540179,0.3976581549099337,
 0.01692567564242753,0.9469315234031394,0.04656485817042333,
 0.8350860397603557,0.4845058442734508,0.1641827556996095,0.7120635351538723,
 0.37780861888681616,0.16294675017824722,0.38073813781058197,
 0.10529232966398172,0.492611238917388,0.25116560531093346,
 0.23032620008695337,0.4793370540196461,0.7664489446345606,0.4285070437115486,
 0.4672266261520547,0.6896244121122304,0.21566244398399004,
 0.42915114881196403,0.7686977530896101,0.5699867073639449,
 0.30866376958798003,0.31709715114791226,0.17710457395547108,
 0.1114545770283446,0.8499217171350344,0.38526167203865114,0.5950346766513616,
 0.6841351536693504,0.7700320772267957,0.09219198480195645,0.6601575087240059,
 0.7428342362658374,0.9542500060754583,0.43800150657309334,0.7387611310031916,
 0.9739986616852867,0.5649287478674823,0.2800693510640452,0.10595621365335728,
 0.07083625393935777,0.1387215478114352,0.014366455185670146,
 0.3144342395977696,0.9917933104335709,0.26166418921854084,
 0.03533317453287699,0.12997995566634746,0.7120861397946603,
 0.40784442574140756,0.6238020061149634,0.19140913068426624,
 0.1940623006151042,0.4427738153088192,0.011782024038500039,
 0.4760143041059275,0.12027793838755207,0.5820311145184077,0.5129585680229676,
 0.8755004760817926,0.4392191722752947,0.5478324292911001,0.1861350561474635,
 0.39795868161313896,0.17895887679142108,0.07317462354324888,
 0.2775409760622928,0.9753137258823602,0.5579440269670327,0.10950865106480812,
 0.6370434473136738,0.6430530022108587,0.7054438651153118,0.39650196560727524,
 0.7840622460443666,0.6679332381223232,0.40409947289182657,0.84003908660778,
 0.5286895120432404,0.7448222041075037,0.338903020457896,0.47960907036361766,
 0.44951031464740876,0.7600714623211299,5.182618570515052e-4,
 0.5958504508814575,0.9677583247883106,0.3097376168747844,0.5398177853110537,
 0.9391897715272138,0.7778136209446215,0.5907853133856644,0.4855451235935764,
 0.7694099830827473,0.10712314124311262,0.06122735560875747,
 0.027503432142328754,0.04875884715244472,0.8776716925545289,
 0.14742435571256685,0.3298532160803678,0.21637038076905823,
 0.15457220140919492,0.9618408144424571,0.40681391052132276,
 0.7893366248325575,0.2899082057163309,0.6699485536760877,0.26158242950919375,
 0.6002403134240426,0.8881348039228694,0.5653624038510349,0.5141743439482493,
 0.5141369852492244,0.6408420183812035,0.756246903132485,0.0932945583594249,
 0.9583406810926773,0.5248115741103797,0.0633698879885739,
 0.012542946185462922,0.2204850709298296,0.8437651244178197,
 0.9332534676926102,0.04092131497796003,0.2697178165487174,0.8891998139187064,
 0.5939702261955204,0.28007426772395616,0.5876173209721803,0.7621691778240999,
 0.7454634292060981,0.2832012026393469,0.63193408296319,0.11462636263693193,
 0.02495860640914649,0.8446585799538923,0.8939694799938194,
 0.07130999371861757,0.8602456528492789,0.4457074496692851,0.2820165366176177,
 0.5879826377905167,0.6481812378936921,0.9944260821776638,0.04342045855539778,
 0.29145058758088715,0.8880486213542599,0.45252573455574585,
 0.6237959662377597,0.2854407197199429,0.7155088160962109,0.37936573466976,
 0.37732033664906284,0.15431406640719691,0.025597816107967297,
 0.44817581962413133,0.7731855656555142,0.730772650012811,0.2715060066992764,
 0.8993166729595855,0.3435553574918746,0.820603104521979,0.376452228139769,
 0.5128371685421658,0.6875597942452165,0.764254596975718,0.14512410819768196,
 0.4752468223012738,0.4389175868260674,0.5702011558049378,
 0.009043135879025499,0.4256004850782289,0.5047084131610499,
 0.04234076090180172,0.5083352735286221,0.8789687939990036,0.2869525724740326,
 0.9137254973466132,0.41451751121583535,0.8038601398746018,0.9379392093908521,
 0.641762376460874,0.16101529524456482,0.31595904381997864,0.9138136492456563,
 0.5906417118450922,0.8369778741690467,0.09788536005807491,0.4624914328253491,
 0.799234657350391,0.42406137582044856,0.011885134495714156,
 0.5674409714772418,0.7280049267097813,0.03462220710916286,0.6599335721314007,
 0.12788343429878646,0.203519559057727,0.3574782140962416,0.1016143365830765,
 0.7118468853752734,0.09765171809451834,0.5801806295316722,
 0.22465355603050385,0.5656080633705284,0.8682287040794414,
 0.14597400005744776,0.1366191408487737,0.22890363625288312,
 0.17633804601234193,0.04132566382807934,0.10330033940535488,
 0.025003694409537203,0.31503576013686985,0.7278198264606476,
 0.32447051206606514,0.34204016403948134,0.5333009284421143,
 0.9352312511974846,0.06430721296446962,0.3649257883275119,
 0.09915426510360947,0.09402358230457353,0.3704298422813802,
 0.2593838200755685,0.45224844537773845,0.01698385075454889,
 0.5775343023378858,0.5059142305847141,0.5471406531444184,0.6882937023325704,
 0.1423548693629345,0.3697635049871866,0.9271704683826897,0.8885146299557521,
 0.5943731109791894,0.22287390521900574,0.7082732360206433,0.8879484642877118,
 0.7816872856766671,0.13311262918180566,0.1280425787581404,0.603219495873025,
 0.6279742669528146,0.7294215186544444,0.36757375958612937,
 0.19236421199656495,0.1966833507977983,0.3291286393324039,
 0.08624629497955572,0.7227105587463201,0.9621905569531457,
 0.26042378491066653,0.04305928225687394,0.5506700829714586,
 0.8195645108002447,0.4738350731112263,0.13454121051549117,
 0.47123741481507786,0.6927964147692484,0.9613819524810447,
 0.43191754553388695,0.29645846889836036,0.8562919343849277,
 0.9010515577515594,0.5122376323133825,0.9628291880855008,0.4693257203811979,
 0.3603169755644343,0.5084226123263625,0.6024500013342331,0.7422840376197724,
 0.28080854548964973,0.46594933670894756,0.7162358857562308,
 0.3499825461959579,0.7776988396029044,0.3525441758112847,0.49938595861226,
 0.19461525696561677,0.5794412310367477,0.2797929949461192,0.0691336949264284,
 0.9824629117557596,0.392011346863568,0.6950471809180823,0.41459145281494747,
 0.04640893503244059,0.5562579014213702,0.015753997306313994,
 0.12821079968063898,0.8152348194040648,0.7807332360257785,
 0.12197195979653475,0.7708667892814269,0.3153530071330085,0.5071943054994773,
 0.6248704075617474,0.3917957711914699,0.37713959987053913,
 0.27931197243485617,0.5604457756291454,0.30564608968143425,
 0.5015292332825054,0.22107870873887658,0.849544864973038,0.8274706030854164,
 0.7224908735191243,0.9695659233449649,0.6599768660037091,0.3557381334523857,
 0.22836226605270693,0.3431814529758367,0.6400853748947494,0.9074690629060156,
 0.149790690009743,0.48848891361144464,0.32348249811399143,0.554361810178359,
 0.37685641929650104,0.588892380711583,0.8046252148356746,0.8869530558709503,
 0.6333398211118441,0.6940404607623769,0.3109218166481962,0.419100345206166,
 0.632875642474711,0.4014586879164692,0.5755105837774319,0.349533759596774,
 0.612522588358366,0.13119454212376847,0.6688420983056143,0.9295082320740783,
 0.7314955093462212,0.3905350336889394,0.46907732571621175,0.4512951362065919,
 0.9526824145928162,0.6842362634142509,0.1860260832399513,0.3183788971036068,
 0.5365297286809015,0.24426331654326372,0.9168917203969752,
 0.19507849680738465,0.7684903492690496,0.9851515569080969,
 0.38303123461797006,0.0035473557592813343,0.7844451155972932,
 0.38872643222704517,0.4382291443640317,0.27232720116697784,
 0.9288428069088025,0.46409355113759343,0.23565871991865117,0.928277562604122,
 0.2555986682764323,0.4277008030778826,0.4205485693658296,0.31628633062054234,
 0.2533849237350887,0.9859815667300694,0.6314749028782939,0.4440520890897428,
 0.12061519333795001,0.12378251891943515,0.19270036171160532,
 0.11788174954774855,0.9426162931385726,0.5321289909432824,0.4124068819911303,
 0.012565701117408401,0.6822811053047875,0.43849917049153253,
 0.1796007555638116,0.08069932499012578,0.4863450232147374,
 0.05069471981712448,0.7984864527009309,0.8434068211203352,0.4736093142886941,
 0.7983345372679662,0.2831432607847991,0.9203509395731023,0.5128155058979293,
 0.5389599615942646,0.22410528634895188,0.9390922986807956,
 0.22591119466471832,0.01625849587843875,0.8787790727750381,
 0.015945986846957916,0.5548601174705267,0.3656996792612536,
 0.17915424704636362,0.9565648231538584,0.06450556266827889,
 0.1450042454887399,0.13816603011094097,0.07414977566815428,
 0.23611284186570036,0.14752841858094667,0.16540098571284878,
 0.37936835317050577,0.7833297460839732,0.7765543043327543,0.5639387875883635,
 0.9321239554407402,0.7476459156880502,0.6151767736642998,0.4641152576496087,
 0.9359611279938324,0.6174865324536642,0.1356871353823177,0.30232611278946087,
 0.605519729441504,0.5758702663233711,0.9494707939240282,0.471198174005404,
 0.3109353050717889,0.7190712948561452,0.4236096024856204,0.12406615776821928,
 0.3787248170498415,0.015507656424392557,0.7676301502669433,
 0.28852356131115287,0.21464219027265397,0.27250885441446193,
 0.3955241501444574,0.04657123596666801,0.0843374690618296,
 0.45355125535526286,0.7082292834345689,0.7332581112500132,0.6732171888832411,
 0.9284253101726865,0.213813530334939,0.6127330146208307,0.9025363686902163,
 0.8649565647009612,0.4275775433633735,0.21207615099315635,0.7647684162191221,
 0.9720125233988658,0.9100423717891893,0.9816294989374612,0.20472273946281838,
 0.6157716481346096,0.7699067339985998,0.6127968303990204,0.1884806411097274,
 0.8591506807029464,0.7886671203331673,0.5424639017253783,0.6882376652102801,
 0.37882113704678677,0.8973723546271921,0.6828601400593192,
 0.44391184292832964,0.16320772015177187,0.7620191989761835,
 0.4479443705430636,0.8872764157453704,0.018255384355206794,
 0.7611762617910633,0.33159308654600383,0.5010153762859103,0.2615028301125064,
 0.9196784505056921,0.8465476128391035,0.22400355687574935,0.5770444203919121,
 0.4201678249038763,0.7076462027324065,0.9012981782173276,0.361007641306909,
 0.4870167636322079,0.46868340980200285,0.272391199982014,0.10309576203566473,
 0.4212957243883957,0.2635574873708455,0.35072174424445546,
 0.16818945984043476,0.860368225998116,0.43371624050306334,
 0.13856705284337556,0.8888965518326903,0.5526465514123552,0.9641252180537269,
 0.19532920185007907,0.697639966598058,0.1988541445860531,0.5544626838524278,
 0.11781656223039239,0.5608268238732872,0.0012130970407134845,
 0.8263705551559702,0.3386700277526479,0.5395353862189077,0.06554374136211294,
 0.37896705402204345,0.7954588075145701,0.6335801606637383,0.5607347909425278,
 0.9653774494339625,0.6277638056523205,0.12206811117461869,0.9365040535399772,
 0.9023207860543341,0.28332519197642225,0.2963287484360072,0.7868615273134438,
 0.8523347769351022,0.19240006473711757,0.2148398241031293,
 0.038334949529368556,0.43133189177453113,0.14043123253016132,
 0.9915649899590714,0.5818554816003527,0.5857272749713288,0.0428847078490866,
 0.26334784884278173,0.30850203411453037,0.3169382194678877,
 0.3076794832081369,0.32253270707832227,0.364090842119557,0.04500302786470589,
 0.8949662932912777,0.6840320561952105,0.5252178894515654,0.9856600256991844,
 0.28593096232174386,0.23951617112328338,0.1825871749524055,
 0.4076517214613493,0.7796015568459119,0.6262231651519377,0.9976863059232894,
 0.6121936882510204,0.9847423804685324,0.12034045142757555,0.9870361797616495,
 0.7628779211320869,0.8068313947305097,0.884263319656046,0.5708687373381421,
 0.3219544548270965,0.8067502955617702,0.33793363863692527,
 0.11486129190534133,0.5578741889375866,0.327640010423237,0.26412613699252796,
 0.37668257327583576,0.504828668297084,0.6293900271477558,0.45980166870704386,
 0.538804963141107,0.9159665121992256,0.5734838003050048,0.1881127412925343,
 0.9203730967502323,0.9265378587742947,0.4694243318467868,0.0968530676288688,
 0.8702678460183007,0.9345583445647243,0.31023921112017394,
 0.06960458535847214,0.8998533829333717,0.7556747671153158,
 0.09092957728540263,0.4815649088998406,0.05865414488597609,
 0.7030849613727187,0.19680554635065928,0.40675598416517667,0.909300366148757,
 0.01427031416295721,0.844072728725885,0.37154992645749185,0.4198345269097121,
 0.7201701113494516,0.43356917407174544,0.4581412480607586,
 0.41092366361847477,0.670072332961768,0.006300597081295711,
 0.8437802073545767,0.8190012380289259,0.9805843025436587,0.9261096484335607,
 0.20147458962147413,0.42136160020549873,0.33917110546158535,
 0.9510144837046094,0.4183317874585112,0.43853025960722125,
 0.35622864424213674,0.02506675498583555,0.9863229430451859,
 0.2623828114797062,0.02215002538211186,0.5221545338884502,0.6702451095425113,
 0.5552747940917924,0.029854859645904397,0.22229912261762463,
 0.6991599467239313,0.24445732322952485,0.3146838785938879,
 0.09202376860864814,0.9640035788224068,0.41398839263292064,0.568797284326296,
 0.4944931250633203,0.5887505584568987,0.9351384347593072,0.9488689489632877,
 0.934557832310602,0.3760988319734444,0.0172120975332557,0.685144562938187,
 0.9146167827211885,0.34449786090449086,0.7025669127353631,
 0.004022696354006872,0.6921823427882033,0.3311081363131274,
 0.8080085597065588,0.9985829939384975,0.2896929428670283,0.5892539737245869,
 0.3451548404701872,0.6760210815654306,0.8208860541159501,0.24666619787198152,
 0.6300200560698,0.19144550959887263,0.7114293373474866,0.7497751150404895,
 0.0666692992781176,0.250822289423729,0.24769584792859933,0.3791749127226911,
 0.056216868830314626,0.4994802606030756,0.47462519983265694,
 0.8360428500593513,0.9715427381434285,0.9274895965233918,0.3946985777365093,
 0.6212385053539456,0.09237761812216483,0.19203468621981756,
 0.38811315042027905,0.18633465101828628,0.45942436948994425,
 0.47999274917136725,0.9093789586426059,0.7814191527219023,0.2543370691477119,
 0.9037353924102189,0.23107383817820915,0.17606957621753205,
 0.9306160417342538,0.7558797636732499,0.2604795263211074,0.6449125060922849,
 0.40731519571909913,0.2538220770962446,0.2834745689945768,
 0.25624243131631996,0.22459565382142155,0.7812224060736329,
 0.4140453263092325,0.560791446250493,0.5762459224194041,0.20097050547521345,
 0.11452196517828805,0.12153575186359833,0.25148295666744724,
 0.4497383432563442,0.7121822649234583,0.896829115845188,0.6579735479860824,
 0.938398444461826,0.5646724023838078,0.012944272120965827,0.4750934721950659,
 0.16417953530925544,0.8581971354833483,0.6085761989443046,0.574411224711087,
 0.5713720293946003,0.6314291722788724,0.015140171088307897,
 0.8415096461009995,0.5233146014095844,0.054866399303111546,
 0.5884064999301687,0.4255162478919341,0.5960676132442884,0.6047318814639348,
 0.8055853980812869,0.679268851450254,0.40198491966145466,0.5959898138733598,
 0.41552303454366857,0.16449720203045326,0.6906875461706815,
 0.1965776595871368,0.9589307845618862,0.7596219280210987,0.9840440409875016,
 0.9241320874292586,0.34059606135178,0.9927962750169924,0.37145465908871,
 0.7351865362604788,0.9183200299991099,0.12382821552048129,0.3346533176253901,
 0.9394727208758569,0.21030110217393916,0.5821432035542315,0.3459104212474705,
 0.818340309995745,0.57402782471654,0.5748902662422091,0.4365422404917557,
 0.025485186348372824,0.14872003417043367,0.6158640137368079,
 0.35489789745620515,0.58401321877124,0.4966945468623939,0.4325203713571816,
 0.668364879089575,0.9625131153305961,0.748891269906048,0.1364623762760313,
 0.28382519299461884,0.39512166882889266,0.40047611717166454,
 0.07955539760269725,0.8393028428653144,0.9382157294421989,0.892739715211824,
 0.08505869594298066,0.1613736653215826,0.29178795266443525,
 0.9860787427605844,0.8660855582504685,0.3975284914642814,0.7810087841958061,
 0.8123866320651066,0.8561661702484051,0.11560435905521182,0.5684490325675303,
 0.7537299992297094,0.0926390825889769,0.5766926407644936,0.14982606650939034,
 0.26325392309954776,0.3700838215803248,0.0539476068142859,0.6081656526625796,
 0.34538307729618145,0.008030370968537537,0.8265241107865853,
 0.9714396743157248,0.12319634461950335,0.7548994988961338,0.6443894166883274,
 0.7906514454026521,0.9548102690318822,0.8847754947469091,0.5020923192419089,
 0.4323752347205627,0.7661772466888842,0.16863625085750544,
 0.014989903519032177,0.6218569503695722,0.35894220404678034,
 0.3713416732247774,0.6590696084166003,0.1429309261628957,0.9225170400110166,
 0.9845678939772567,0.36317477067070825,0.42956110598898134,
 0.5386782178071647,0.5177628969948839,0.4693460320305214,0.9434179383676147,
 0.9414139961650325,0.5470457923567585,0.389799886211609,0.2430454056658098,
 0.44702422451008483,0.9703058443278932,0.3439991297195757,0.9766823243862578,
 0.2614074504462579,0.5811411185395825,0.5117209117794943,0.12071770139556381,
 0.1285770516508632,0.7832462603216231,0.14128152022456664,0.8322861347139026,
 0.9191645962306335,0.8293369540263328,0.9503395342878848,0.8958852003155018,
 0.28496199305111614,0.9096824932475667,0.08225221017347428,
 0.8897057945661807,0.24389285355568968,0.450892600627943,0.7404554587350511,
 0.38341634192793383,0.944700094884517,0.6518121585441364,0.8790063547416788,
 0.3666514318384807,0.6772346791230877,0.6189855322277712,0.5574373778296958,
 0.6958801636451969,0.5541588336339782,0.6663927057975473,0.25122115661556665,
 0.7861132802365299,0.9970452272045044,0.023349120294818393,0.671960615080044,
 0.8478818155272891,0.26503525146218154,0.957808828997477,0.1251636083397467,
 0.7977731366889227,0.689589769422022,0.1867759060464559,0.15690457430675886,
 0.39945908135001384,2.834407943337336e-4,0.7628929232075499,
 0.9491814019420679,0.9967578574774073,0.0690100451514658,0.08075243269288856,
 0.7416901371710306,0.4122051156314177,0.5474391014058214,0.5378082102225414,
 0.37564438239642195,0.4737281982674262,0.8042672863362199,
 0.21240802127996627,0.9624762147878577,0.7217580079451311,0.8913677848908464,
 0.5543357375712061,0.2642040551745247,0.934077638271563,0.8114587602214336,
 0.3599064832587584,0.29853122534123333,0.365708576100488,0.6796513962425157,
 0.8124274385197099,0.7450469446388961,0.29196463177900334,0.668303438073875,
 0.5524758978705716,0.14287886114835058,0.49609787400217187,
 0.34842626404843224,0.6340629761907322,0.9783583803438498,0.6543012381822939,
 0.027801949826336214,0.81478805386753,0.31188028871665163,
 0.43186687029243453,0.6963346502626235,0.17454310634127101,
 0.21909933347872002,0.3673229081572582,0.9389965654714966,0.791429421073992,
 0.4908577007030771,0.4552174262204557,0.6646774198311259,
 0.015750667784356853,0.8998306224474052,0.9380913575955047,
 0.08487387149906833,0.40676654227316467,0.05709813557747079,
 0.020246323280663225,0.7796662758718245,0.17151955393490825,
 0.07386651088847562,0.2861136186499409,0.7612530069694969,0.7183197882182207,
 0.44514524048056514,0.41244331365165066,0.523000016320665,0.7076416359961741,
 0.39714606165882793,0.14399886602547884,0.23067149373536833,
 0.26709420370443815,0.11341689354004147,0.9843948290118301,
 0.47481869765993423,0.8187248776512264,0.3422079662850428,0.7781365587327295,
 0.8321905143917384,0.8648477470782047,0.6796493095727943,0.7963767229008851,
 0.2183861683841174,0.2764713539358217,0.7542878993899488,0.5928270363438255,
 0.1825098181225292,0.6379446392443383,0.18338535127998612,0.9991577152326871,
 0.188507641267033,0.42013593672536975,0.27128477097200066,0.2800303605724883,
 0.9698700687268906,0.6339128270203153,0.07588401182654936,0.686264080638981,
 0.30574575672940774,0.26676205145302256,0.9868527977017402,
 0.17912410256564926,0.9300879293045412,0.3840275610960783,0.7986120787059356,
 0.3949719157285304,0.8453491610005229,0.4881074255458735,0.6922709117804782,
 0.0315143420025481,0.8895116610477799,0.9824027749594526,0.6094450815571385,
 0.05039662060687733,0.8922521212353601,0.3956662642179034,0.5894428417416216,
 0.9358888844167675,0.14327966806098524,0.11030917030268061,
 0.2966752772181339,0.8679410906367186,0.8170495953909463,0.3983838093876758,
 0.7087898177348361,0.7529913864161982,0.97846035111576,0.4832897113678013,
 0.07311455110316067,0.1150224105218991,0.43319066817327145,
 0.6090878582924785,0.2183990177952576,0.07484679101150205,
 0.45675963809940223,0.7346496591537124,0.6027721293734922,0.5211832645405183,
 0.779640061174029,0.8531691057620954,0.14161031123545664,0.7564600775288913,
 0.893733166342333,0.0470514935661317,0.19822492714036755,0.792021396678463,
 0.6554669100501391,0.5400262406910808,0.03347036966782446,0.9519534225577652,
 0.26852425347356845,0.301345403868008,0.44789210780879407,0.6781841654228604,
 0.73471877574573,0.19638559174570502,0.8283209882714982,0.7864858742010576,
 0.4772793095035056,0.5012331082185084,0.07642853599893595,
 0.04472438952659974,0.8558402155692343,0.7415516697367516,0.1947527120598075,
 0.991623263363641,0.917815973121277,0.6880020921620849,0.4206523439768395,
 0.7624103430055617,0.19423202910824067,0.8385127566751034,0.7062606377281627,
 0.44288294843133,0.5014576250399216,0.8244899134997312,0.3663193676799332,
 0.9951376204674091,0.5301985331867762,0.2739127142137906,0.4194231279016404,
 0.0819090033428389,0.651122084903624,0.36244164447697247,0.9171516376637242,
 0.30640600736375045,0.8895428625368175,0.8396713892490351,0.4551371859922627,
 0.3008548244875966,0.9998985209695921,0.800377342785373,0.48752915490500637,
 0.10702227709047596,0.6131562009384761,0.6871691856092254,0.8313845902735669,
 0.1735945067841006,0.7971854920780755,0.8880697021951055,0.35767314796995153,
 0.33620649624346566,0.8345094437594719,0.9750357632240544,0.5391247733391736,
 0.8382671325850211,0.8324717172575057,0.32956318690514264,
 0.40342375380469253,0.7260716932615427,0.5221046133691234,0.4590292883316013,
 0.2646920374443015,0.4325573533070759,0.09187708127354166,0.291120659125912,
 0.610111506737123,0.9008700522736829,0.8675351693653552,0.20186315925075837,
 0.20064959745990796,0.7865340865278967,0.9256327291658133,
 0.20316824644446174,0.092182410733336,0.19989238400499476,
 0.05885612599316015,0.4298193826571547,0.981609616360972,0.8378889319456013,
 0.03074559241209396,0.7526087459832115,0.06088709287537031,
 0.48637739056159257,0.805217553266572,0.10534198756761282,0.7546157500817146,
 0.40084748349354826,0.28330085353480194,0.3678148728310452,
 0.9274404843852828,0.5486028387556109,0.05153253103040123,0.6713169087299348,
 0.8068490553569292,0.9781075173273789,0.13191177315122782,0.8680685215434499,
 0.7100835786110036,0.8270745212155173,0.4861885447511717,0.4473712267139519,
 0.7556669489254705,0.36159613230415955,0.36235163435666684,
 0.49045034401087806,0.37114320405033385,0.3164653788839109,
 0.49445423983218173,0.4048707190313412,0.7749361537846121,
 0.47936984078360023,0.42293229350010697,0.6403471512333708,
 0.40447498894336054,0.2719567967161266,0.2363575598739337,0.8591878981246942,
 0.4335220848148478,0.7811064465537124,0.5846344908189733,0.12796614668204231,
 0.13594409966279108,0.9860711925507376,0.9500784785638,0.061627681565749715,
 0.2985614762233565,0.37180966075631017,0.40312800363431756,
 0.3290777511820645,0.25992259843368515,0.6430868965785282,
 0.14304490178024964,0.5826363943446696,0.9684871683051437,0.9315653728896385,
 0.33615576421041304,0.4694251915518056,0.21275597370692967,
 0.6495090485973416,0.10642284079388575,0.031155902433011362,
 0.044507957463900905,0.5227611725657033,0.41124604279459864,
 0.5308655379290609,0.6374657367347698,0.3926625087105522,0.0429722976504604,
 0.5530112838802227,0.899942022539125,0.1472183547813508,0.4302873752559695,
 0.19029354370235352,0.6518305527314864,0.28429784630702537,
 0.2591828402275874,0.08045612804371893,0.5615376995818803,
 0.17495368338603046,0.5688848363503913,0.22327745881764527,
 0.5611014523590437,0.9468311327643267,0.421746088765242,0.9136265262448388,
 0.3336938969051606,0.5292036215964007,0.8024080326200175,0.5864571006624366,
 0.13966102927766066,0.9935145181165044,0.5715537214037012,0.1197036832544679,
 0.03296509795699254,0.2885491566839983,0.6846897357919405,0.3774999761985517,
 0.10704965974119052,0.9711626624608855,0.5227793037002005,
 0.14074034102250632,0.1629239014358398,0.1254743783847162,0.5578586824509395,
 0.191177624506492,0.9669822125186907,0.13829990031476436,0.09792874905951965,
 0.3218992156595487,0.10479662699739256,0.32605525825588033,
 0.13817546494985067,0.032036659421109115,0.8061292589635662,
 0.32859728440150726,0.905699360421328,0.5692466317781554,0.11288826444303501,
 0.9700245944539267,0.8210987126637734,0.9535932762266345,0.1798742421469277,
 0.0339648849270614,0.7604075539606414,0.28848105695632986,
 0.48003036072245997,0.22011268141486406,0.1798232815306381,
 0.16701842928451016,0.17265764752113522,0.17641351381784576,
 0.3960333454944306,0.41939444421492234,0.9790672245529064,0.8123035936672386,
 0.32403964953263176,0.25670786734459994,0.6357452687006944,
 0.3899390729451311,0.4928837868893531,0.41140031978123437,
 0.12896906993237012,0.5498643853140288,0.18938765203029162,
 0.8520194675307795,0.8440522952452965,0.5271583006968642,0.23904332651609272,
 0.13579967703967732,0.15935524453312078,0.8259766088411974,0.737343708117901,
 0.06142330756452774,0.29342298347232243,0.45381729994004505,
 0.7831261634048674,0.5944274434608389,0.32643576204652514,0.83459343338035,
 0.5837331093581027,0.9773862021518105,0.9971069866820792,0.5342901263052804,
 0.12166095236688745,0.04722785650295247,0.5859531406979803,0.45949671951617,
 0.7431273866870594,0.33830782382744395,0.944802919855071,0.2880842509013557,
 0.3986362356004123,0.6724762117403936,0.01746039756955753,
 0.04603661839738471,0.24701636307998176,0.429477814488781,0.8745004716007941,
 0.7003582779372172,0.8947334916786472,0.9584736728215443,0.28268565449055494,
 0.7159485285157119,0.8598958286112826,0.1846355948245335,0.11153128633530207,
 0.2752699425282894,0.07496678088750797,0.795476654049415,0.17524606675041488,
 0.8077648312211001,0.0620622755162088,0.8902089157516426,0.26550794203036565,
 0.844971780415357,0.1258970476745278,0.42846367104851535,0.9667432409926194,
 0.9168073500488777,0.33044431457910484,0.15682582828515867,
 0.9575453680220242,0.47698686184430994,0.22399188418634874,
 0.09517088361733661,0.07145074882365832,0.09203698985865505,
 0.002286150521793817,0.10871066169848353,0.7116129223950205,
 0.9811693930388083,0.8738634121102188,0.11883389279379775,0.846092496932135,
 0.8449082681439761,0.4157996733628295,0.07706949548526221,0.3271534373913092,
 0.04011301616431273,0.7618850211333891,0.922887930670814,0.4426577268062274,
 0.48937757841892304,0.7008554648960634,0.33539568831545663,
 0.18919013893682202,0.7263672686448103,0.11388326485331057,
 0.35941008965786525,0.6420120668353586,0.9581025285834056,
 0.42287013250748884,0.5815055130334221,0.14563714141059447,
 0.19925718641943746,0.862218722843185,0.047140425979121936,
 0.5467300384369187,0.49121842140053185,0.22910254024737597,
 0.7015529845425772,0.655015498636449,0.8506294919646649,0.12004497651075596,
 0.16207150021327033,0.405740413280564,0.48739082474841455,0.689425882458596,
 0.9273205809499945,0.46103771334275356,0.5932648009461835,
 0.41908004348584116,0.737108608345699,0.16543260425812834,0.4087923585032066,
 0.2787769666650639,0.4852452296904286,0.6313242921592966,0.6800974107847872,
 0.7255615518885201,0.965762531027448,0.12551762145774947,0.9616504141080601,
 0.2858277149600663,0.7648433014075104,0.05360133251465382,0.840730680023864,
 0.6948907722228538,0.9572415238719255,0.6070308952990677,
 0.055164368679136655,0.49435039977290196,0.577428112590948,
 0.34620411679796803,0.5212425588041674,0.9916934491533266,
 0.36486420004142106,0.6653649147021986,0.3116069374672187,0.7267030605105812,
 0.09188670866881321,0.9804269169800861,0.6022236164443506,
 0.12190965534945963,0.24095483743417345,0.30696737274053854,
 0.4510078009449682,0.5999735755888826,0.33703429886306047,
 0.23880621474209962,0.670011451993568,0.8500134777205619,0.251156132305731,
 0.670649087908408,0.0756361401802439,0.03573313295935121,0.2618868428701595,
 0.45920264275312617,0.59688253623849,0.6164578109402933,0.987344872974535,
 0.818401981897147,0.13359585671636354,0.19612128985820976,0.4806488810879376,
 0.08686856109127472,0.41734952033740147,0.02579056739181773,
 0.5853458187296968,0.6048109975580859,0.235773306577735,0.2952565686877725,
 0.5394661873624056,0.8809806629355245,0.16267609538243344,
 0.49399986234387006,0.4448717270730037,0.8445781834716418,
 0.14735002702511746,0.389688239053086,0.6936942232943807,0.6131455285160148,
 0.45726651755389525,0.40243702521351676,0.6971959705680715,
 0.07907124631902163,0.15610789784047385,0.8730753013903922,
 0.29506514206388357,0.7462625623887101,0.05659976311717485,
 0.04111177049108572,0.14528977068618742,0.7223145616425753,
 0.20313596028512249,0.16786157784006417,0.7733182799099578,
 0.1905194741156423,0.035639087576657236,0.12027316994869186,
 0.2661630580401271,0.2932251032035197,0.9418285229648992,0.5957508436769814,
 0.31474200988965684,0.57790653231619,0.29827587546725465,
 8.822122932480303e-4,0.3355458019724431,0.6153625748049174,
 0.35302790856969035,0.46775597722015805,0.7519151477782916,
 0.6879541526620725,0.42287453689660337,0.201366796452203,0.7865904477797506,
 0.2988775428332273,0.16363867460812542,0.9451142143033376,0.9823976845575055,
 0.8377337228006727,0.4237387289749027,0.29695438979016764,
 0.24252029652909268,0.002683043828119569,0.46162537603698406,
 0.3519874273167325,0.8766819312870288,0.6473690710059715,0.910158555701624,
 0.12457971154986491,0.37042309349394686,0.44122913385350304,
 0.311645377732796,0.8445483518234911,0.9307620617157146,0.3177437611664873,
 0.5261872333360886,0.4389251769530925,0.40022139029658976,0.2650456265488619,
 0.35970997280033457,0.022875051126241108,0.06966083085034902,
 0.1968990137624037,0.4705825494432969,0.3839126479700191,0.6426837337572555,
 0.37289998464860985,0.8310364212807974,0.9818944933990058,0.5815877738794902,
 0.9271601170552473,0.10484033282672789,0.3198804562630012,
 0.34960155711925733,0.2448923536081541,0.5838993928199662,0.9843674964146432,
 0.1371342594299192,0.8359491077478094,0.47707082305750004,0.7080512786878145,
 0.7282594950348301,0.053804085184898875,0.1416254904758849,
 0.17904177513212804,0.7291868572427473,0.6237861165718609,0.9407124987235476,
 0.6182485558997918,0.023287684260696606,0.5930099052639483,
 0.7396805088189878,0.17701260673747432,0.515941349487731,0.11273775783824691,
 0.9224591572722342,0.8049027342169864,0.020764333278511682,0.744076382299435,
 0.24206626890923777,0.8562773953625356,0.7772920920642864,0.5196749511632259,
 0.5383630481390463,0.23521800400702764,0.499956982377633,0.5144535159133146,
 0.9095642993419308,0.5402850996716015,0.8366920731895146,0.7097746920789119,
 0.31884899588951465,0.8979005990103204,0.5865047363707198,
 0.00282860213061098,0.4099826030108097,0.7981288298135729,
 0.46821996428910684,0.9741967357686879,0.3091267956482122,0.8215505801491766,
 0.5400265525778144,0.19412121895615075,0.023726872388905917,
 0.2495007572110437,0.5440110405665034,0.8443736250925173,0.7966582472086668,
 0.12495018328351903,0.825087050334571,0.32202863523637815,0.7856473002165645,
 0.38942042091421436,0.5266063097482219,0.5902469384712637,0.9205880648712373,
 0.549295600479137,0.16965111614510997,0.6839145128712318,0.9801158215226639,
 0.6766934021780263,0.0732510577330252,0.8303278005293857,0.42908388675869835,
 0.9721301745574397,0.98002946681026,0.8469427367329168,0.35530955796259145,
 0.05589918713221653,0.10776087520799538,0.8890190562815039,
 0.20689586235224655,0.5782882315372914,0.5678316521310914,
 0.33887236360428385,0.5270887476059868,0.9232382731132817,
 0.028127431428764615,0.39005565573301415,0.7089468537493875,
 0.06256525182825912,0.5837468024188929,0.4462313562468795,
 0.15655236052440946,0.569308779186481,0.8144504851661171,0.11641900248076631,
 0.9693018055407007,0.7983871265825172,0.9518102379967975,0.6885661593048427,
 0.1851364383849715,0.8928793653249938,0.2767972854012186,0.1260116843876462,
 0.3354760606637286,0.48589518407941146,0.21283811728722501,
 0.2425616397649758,0.6222410383996502,0.9956026891912408,0.7495777612205594,
 0.135050890302024,0.8632507234511215,0.23286579906346916,0.16907206166410926,
 0.9769778352897655,0.5205623753874978,0.6385354396896538,0.4943257939771524,
 0.4551959682921134,0.2528018617269935,0.2560326187777544,0.9270428775060079,
 0.10982152208255214,0.7045047520458133,0.7051035905037759,0.9270222590442728,
 0.48108642370680843,0.6627947669202773,0.5917932102775852,
 0.48822862172449233,0.5737187905596531,0.9493481643848861,0.8346815434029536,
 0.20393853390189343,0.06821884077599771,0.8048948437210213,
 0.7532272784379933,0.6334680846727837,0.25519030079970817,
 0.31525091012014594,0.07015042765493451,0.4483695819370749,
 0.11080571030320796,0.19573277794334298,0.9948166502339386,
 0.5081349271591404,0.2738130898757519,0.936691516842761,0.9645273616582619,
 0.4721786255110253,0.1606327492547167,0.36537384523529526,0.5875256939942948,
 0.6171498608070957,0.8106476956435592,0.6046053189182118,0.6063759444913461,
 0.5178405222898621,0.7268105521651539,0.7189546014187745,0.49242548879520676,
 0.7653753226058673,0.35162516607485284,0.37014419076987237,
 0.19447625584330752,0.5699650845359715,0.5792388784576905,0.5489594568637746,
 0.28489664713413543,0.661767945817775,0.8151977417054369,0.5716119378696326,
 0.18656105047693017,0.22011544061773813,0.0368041234696912,
 0.8305372930369633,0.9607819717796808,0.26523901319668164,0.6421625520144035,
 0.5574065655617492,0.0013783795691321155,0.19307026046125575,
 0.4485106772325399,0.5907105608657027,0.6617619499906457,0.5617874051953007,
 0.41780410483541885,0.2378912693823817,0.025586226309847415,
 0.6809105830242079,0.34296395905973887,0.3587039456682861,0.9152638178488901,
 0.2695297997856708,0.23716372173497813,0.6690003638800992,0.8730413326340064,
 0.6598689714747329,0.35126015735477456,0.9156761139353876,0.8194320204476339,
 0.2168316576855399,0.38655202281674483,0.7607566057721444,
 0.17784669638248185,0.7972869790966576,0.48632059645251324,
 0.10396435746069754,0.7212066270429636,0.41754440063804266,
 0.9660989523969062,0.5110556375605535,0.9111258641042403,0.541524110395705,
 0.12622635756820566,0.08208594620027188,0.909761544518502,0.6733398177798827,
 0.8198798473661568,0.8458575834707508,0.5416808174751845,0.36149292349376005,
 0.242788454918348,0.5348263634794809,0.43572198580285504,0.9643884459911841,
 0.46150662136143183,0.7988079567419109,0.6617331084463002,0.8900734429517152,
 0.6088156758717707,0.5614130554591459,0.44066020023866403,0.6043430894077552,
 0.25641606239290826,0.47272902555040497,0.43587630297733526,
 0.21942302025329807,0.2954772660584465,0.014089075798498385,
 0.5632106569924497,0.7106700211492418,0.32051762014496643,0.656826083571099,
 0.35216661758577505,0.6737786400657254,0.20353792544621951,0.53195284027026,
 0.5392847112000368,0.5963448867262165,0.27929984390749574,0.4415810693956973,
 0.44866540082205286,0.28495081442507575,0.4522854363803538,
 0.3840187682358098,0.1722706198999493,0.6471526585387052,0.7096392116128865,
 0.25324915546221927,0.4693241221809692,0.5815233980714891,
 0.25193495319905646,0.7904499112885576,0.0057380895125339615,
 0.526741179145573,0.34325298968136164,0.03949715862390191,0.9066769845392165,
 0.4045568194111536,0.5308466431621981,0.1560555003905454,0.9655452235558852,
 0.438143371578951,0.45547778311513454,0.019544372843752433,
 0.6781919525729042,0.24574925449773943,0.38652474672821047,
 0.4482580059710968,0.4955820465060532,0.7588115278260399,0.531853196650785,
 0.6300387883024898,0.5317922411959483,0.40660065551627667,0.6679357715747917,
 0.5335227369483927,0.30381835754781605,0.6923389114894606,0.3967337985930557,
 0.20166686320213956,0.5842450276962029,0.9095807892235036,0.6428022845198877,
 0.45762231927707386,0.86059179169673,0.3488612257377036,0.21891365535667462,
 0.5522999615689632,0.9157885427989045,0.3674653080099465,0.1848540687311928,
 0.44631918471362375,0.7568694392550218,0.5422869872019416,0.3624661710963445,
 0.4054687119121475,0.3568707602413117,0.2536870962014035,0.22352263817317486,
 0.6919108883451218,0.48808948899491755,0.06682118353890854,0.514207158820727,
 0.13228136320186423,0.9603975593094061,0.3657861079597692,0.7860557488329929,
 0.36366655379013035,0.2084500485498031,0.7794817224360576,
 0.0026922957907110634,0.16028629031956765,0.6749255775872599,
 0.9280635755580301,0.7237306998396579,0.8548650370342007,0.14924268469988988,
 0.45049852199655194,0.2597277693964033,0.4153917523457962,
 0.060762825883949656,0.8241458989523138,0.10220480729472059,
 0.02814408272018465,0.34894932737267415,0.25354814495456746,
 0.7630373607826214,0.742329777497184,0.462079432880661,0.10919752248286951,
 0.41542902919292224,0.17018943519609986,0.47885281493290277,
 0.09870822087608744,0.7967388418747866,0.22555294118294744,
 0.6180071523868403,0.21898435308736608,0.6763828328005782,0.580256414048728,
 0.7395842501491908,0.23243563515036325,0.6828906503764173,
 0.22621792646731376,0.17229335865338358,0.6484189639559252,
 0.5002296014627243,0.5426165030406989,0.7042152476160464,0.8639754409268093,
 0.5033252864821125,0.06581326576106195,0.39875158086361917,
 0.2732883635414084,0.32172411530279743,0.7530296160540154,0.7873985676826457,
 0.2785932583439098,0.18204593296582272,0.1670240912455696,0.7357080517794384,
 0.23832394250705458,0.8421981631448207,0.5554628274651103,0.4344733670003671,
 0.4717899506690165,0.7360258231142449,0.8498545268950548,0.3194317737965029,
 0.006064133270320848,0.24110449977203186,0.014083246665932636,
 0.09146895143348155,0.5647336240052616,0.8644853484005081,
 0.08778408169144292,0.7317821052437817,0.8785691799139769,
 0.39608237177471295,0.11165349485300391,0.7272275306083595,
 0.5636904099856026,0.41538313009791517,0.9381770292966154,
 0.10654817020421337,0.4239740035587508,0.4333847026121551,0.667947288364213,
 0.4413839963648045,0.9292689145474317,0.551315470074263,0.33524076699782657,
 0.5118553678464665,0.3621127435013819,0.2950451199243562,0.3364531266235552,
 0.24390169225961833,0.7671342325135265,0.5360449237545623,
 0.13198450752225144,0.047480550634032004,0.055314670558146095,
 0.6363798590215468,0.5315889330262973,0.0405002067521687,0.817581871860555,
 0.403533744467622,0.8315711286086992,0.21836284411634477,0.2643733200420241,
 0.7474589191118404,0.3348449690852575,0.8011123219424983,0.747612580326795,
 0.28603526718600014,0.40832793379204013,0.08343705149881453,
 0.1626096315314004,0.7382018476657755,0.9626079796153643,0.9847357569533041,
 0.6983178018858487,0.43765034581132845,0.10264661652368923,
 0.4001091586720307,0.35346796237340083,0.0012526347774864943,
 0.19497881104198767,0.8955039087095985,0.13901109471355444,
 0.7471526841647206,0.6338397168394678,0.6429062757274713,0.5595455077324225,
 0.6901639685547464,0.6016496946285744,0.06118804427502755,0.8109992192611905,
 0.22157861513342747,0.12927149864639054,0.17710602435629208,
 0.09983824981901568,0.36441832766736704,0.19328312324446806,
 0.9772461190579438,0.916853429753894,0.8740408711039753,0.9568398895895982,
 0.8807490741856374,0.9807017888611952,0.45244468359531465,0.8162484578228012,
 0.06637204275739994,0.3645227435353735,0.993399532400715,0.01535863190896003,
 0.3297182379177328,0.7126013905640038,0.1133310046510928,0.7277985401558632,
 0.21015918575320192,0.8466321797894895,0.6638275971590223,
 0.08169103382921172,0.9423700235368817,0.0806338473500906,0.8945611113604303,
 0.5599202155345896,0.5277588486057637,0.08308011434055618,0.8360214937326839,
 0.19381464882753152,0.23264319548854195,0.48621489708404264,
 0.5270561401732548,0.9805254898640027,0.1763541334412826,0.5726511967474315,
 0.7015406926749875,0.9113483319025942,0.3399966326495324,0.11423375394593782,
 0.7140879249919566,0.384470123185809,0.16862505385665694,0.40889147311786156,
 0.6248415252494619,0.30610838708932775,0.6189632290591558,0.7626840208095717,
 0.84398449251791,0.5828253492684046,0.7999999775726948,0.4411142524475512,
 0.8870304726211049,0.11283919064386849,0.2817110635634097,0.5537267182152741,
 0.1772358600938565,0.8072483302680953,0.3665686566280043,0.10218472057679007,
 0.9256078471443447,0.8318241157364372,0.8888167630162962,0.9705265876547758,
 0.5078276493734218,0.9866528492853602,0.13611807976084478,0.5386332088295942,
 0.24533200591822757,0.5088150318434497,0.35960348385728824,
 0.6517294475994977,0.14204915659290163,0.22475144956688786,
 0.5727235040272955,0.5411320498250356,0.4324985797165215,0.8564217757916924,
 0.09648352322019105,0.20712313576070052,0.4866667089083998,
 0.3580725206062898,0.7077721736569267,0.7688551880483202,0.4322380395262302,
 0.4230977476015786,0.2735610985143927,0.184703725074145,0.6381248508474305,
 0.08279365394978988,0.6237365058796996,0.1811323853716621,0.8182214709379092,
 0.28835617308169104,0.22109034858133547,0.10771659893731167,
 0.4903806680153465,0.8788544955142765,0.4083441901649607,0.21363079323489853,
 0.972182757952789,0.034115807117781616,0.4065189646255061,
 0.18577068906180083,0.2746839590129331,0.22827290967282954,
 0.9823577020519653,0.2950999383480226,0.015834442520935244,
 0.8650226209332535,0.6436504242685485,0.7959845988531942,0.04349973113270822,
 0.7686361759768812,0.31721003478750764,0.8367017636980094,
 0.15112451751871514,0.9430508074572735,0.7632432653262042,0.7361153497022934,
 0.7625278407556478,0.673976179393165,0.1796721254647745,0.09399807011699923,
 0.825488522044415,0.7714421486242421,0.018158787266016052,0.8695731558928427,
 0.6019648107779141,0.8961846728561684,0.7248624667530823,0.5779263549794627,
 0.5365912420627721,0.5425476282483919,0.09677123407449562,0.5348960285379093,
 0.7362245730871524,0.6318865932716806,0.8887301534024962,0.03897019197623941,
 0.6042382728393637,0.11681881858834786,0.9444614543717391,
 0.25269808891397527,0.20114143570433685,0.0339746118732116,
 0.32753318827401734,0.8979014898524668,0.1965438790622518,0.8725367860897855,
 0.23595819211156932,0.14072067589467752,0.2810557869232929,
 0.3126405831152269,0.8385614573483824,0.9448880995774209,0.01020717806324245,
 0.21301304868015347,0.8191419540107017,0.8359051012257136,0.8847863119799878,
 0.9973777902822104,0.013477994356751344,0.2457599762250382,
 0.6780728940216674,0.9873950426950276,0.22369505761655006,0.4661997367763996,
 0.648895016229212,0.35065887255522865,0.9044651897619574,0.14882226032290735,
 0.6581476187902975,0.6021122058442473,0.6454223520062606,0.900619603949898,
 0.6024292688662466,0.2658565619695209,0.9735806899195473,0.713768219026375,
 0.8679877490785466,0.32653151853059526,0.15857801876310917,
 0.8703091177902672,0.5770207894217146,0.449633143028195,0.13771534405202102,
 0.6856647611167705,0.12680633466152014,0.5731979469862657,0.6776178299302218,
 0.9851522540996259,0.24399946818320595,0.48519510247261555,
 0.6081868731406137,0.02326039455072959,0.29741275722998717,
 0.27379609742805333,0.4072322740013544,0.4913131640291186,0.8450536862210783,
 0.8266286265119432,0.39692576887670306,0.0713290194917835,0.8320072116080391,
 0.08532045045088887,0.628328859609831,0.4942863180647439,0.4010328495857963,
 0.1507631957489215,0.29218358268329014,0.6911920318486516,0.2452243841837478,
 0.8413178443977826,0.23622691824102904,0.5656285983011834,0.825237958054015,
 0.9586045045812751,0.8927222752949615,0.3760380435552316,0.4249176992733015,
 0.13043056705535716,0.6645189549657213,0.06730952036311288,
 0.5919787085867243,0.40288774762050594,0.9320833427010369,
 0.15386528339198513,0.2498579170141434,0.5202929611325949,0.5264966173592462,
 0.43680065814355007,0.2750071228913742,0.9280148747894768,0.8783799084226518,
 0.9842806889824722,0.06618017982011815,0.6718283637228758,
 0.49358946824678807,0.28435465775361934,0.27891699645268975,
 0.14616119376911996,0.3483912067420881,0.25511001185202686,0.712051041473456,
 0.2069265988855833,0.8608799523390779,0.23819081760477123,
 0.21007613875904096,0.7469396823270863,0.44902994581258615,
 0.3368672597112484,0.9557409264678469,0.6968326268881224,0.732085399820676,
 0.6225399327658876,0.7399392706717166,0.9961591893544645,0.7262429898752166,
 0.29357954128238284,0.24930134308254004,0.6391223232679448,
 0.9979386116109998,0.127752436273178,0.24322842877222495,0.37915872979227916,
 0.8384009731755929,0.7189929152127232,0.12473294578049643,0.3946798897386449,
 0.3022141791981905,0.880849439722001,0.6435012427532708,0.08238409545138659,
 0.5994294274941265,0.42852969685052666,0.19472504443292382,
 0.0695191411764211,0.8144275514120126,0.23113786273970027,
 0.09001605774630839,0.08755354191894282,0.6396713852321381,
 0.7513482612732837,0.40301791972439804,0.9641739035048387,0.7763477589897547,
 0.6328598566484078,0.4659428476267562,0.022926850282724565,
 0.8685996900587452,0.7776278101028297,0.7571393324556668,0.7641340036387488,
 0.2474687677188777,0.5293922701247549,0.8315540685408239,0.8478133940103073,
 0.6969615080354707,0.020763690775394084,0.8567320769653319,
 0.7880418333872071,0.6287109318957036,0.4312935055739384,0.2218594432195795,
 0.5601754615595841,0.9952706182332134,0.7556494595590454,0.9701303104373281,
 0.33583732058448235,0.6936962801643733,0.17223148995876458,
 0.07368215225194885,0.1557848721700562,0.8257522879493463,
 0.43446071008973874,0.7560359767771595,0.7691404290875924,
 0.42451447739531356,0.5250921693047068,0.8397269994632569,0.5554281334923963,
 0.441046984806696,0.7012488881350917,0.9963677466743756,0.34044456861550876,
 0.8748347820472278,0.47419447896508904,0.6039078561508113,0.7138054908487109,
 0.31061317898847396,0.1644527174913365,0.7446519189145405,
 0.28756007978011744,0.8700422570079817,0.44164940020612786,
 0.5582971968500815,0.5917209996142457,0.5611592364732229,0.8041362790569421,
 0.7365181096979879,0.28263964330027747,0.98005896368848,0.17654284246335572,
 0.10571197644769281,0.8305095871438328,0.5616132158173528,0.2993572715057531,
 0.18605286970204093,0.651582525053082,0.41619317230256725,0.8336615476331071,
 0.5276251773080852,0.38126317633815043,0.568480542934157,0.197295298484519,
 0.38055962552734435,0.2593597539086434,0.22001865877807514,
 0.025294745974704713,0.16569919281538092,0.6009568749644245,
 0.8879394036554707,0.326681078451069,0.5999767467784718,0.45755307085926467,
 0.9286292651203302,0.8771963639727781,0.41548382073333123,0.5518532726871936,
 0.25862234205034806,0.12847360717423306,0.3412512263859152,
 0.9655283989276403,0.055779189537429,0.9418056507216699,0.5787659946180824,
 0.20260322328658142,0.722772696049046,0.2385757045777308,0.49294420854946264,
 0.15478212196745766,0.2027124805821242,0.040294119946281515,
 0.29037249796908593,0.38656242816094344,0.8668176064508009,
 0.2834302930182415,0.7262884303474844,0.9785012424730118,0.24483846537976073,
 0.4748129255899729,0.6767402843782352,0.7023557228083984,0.8388021669651845,
 0.886939106661175,0.15850371929739182,0.41746030996600725,0.5512762619983672,
 0.052787211940322654,0.7632758369285091,0.9416745809516273,
 0.9741686200762969,0.15218035174351818,0.4847598535606382,0.3307904587550128,
 0.6347318185063402,0.7596344893044258,0.795245509723295,0.24857091971444223,
 0.0384685697472843,0.25698085898296275,0.32295153219264994,
 0.5615603011683241,0.816628290830438,0.4517354493040536,0.5601968307402712,
 0.9303865971797349,0.6198779648300529,0.8438638409511976,0.43219573683307255,
 0.14703160907439483,0.6871732510987154,0.9562849087366694,
 0.34995877159823063,0.497716831401109,0.10885820958343495,0.2928799481772335,
 0.9620256688106404,0.48970785223135205,0.9240814801123777,0.971689015719313,
 0.9855423130935401,0.12200525687838992,0.691377115429809,0.5470620627439166,
 0.9095518764129773,0.49310795145885766,0.6884871949140867,
 0.35725211996166295,0.021453062261455713,0.0835263568473621,
 0.3197672418334503,0.07240031621639043,0.5687687434238182,
 0.20712838057478244,0.14783344407886878,0.5187635513062803,0.627300015128767,
 0.5237298547103371,0.1701772455712799,0.19692558781880753,0.8648075102427477,
 0.9164419455419435,0.1727266342781676,0.04403689596351701,
 0.41641437618390975,0.6197308847612653,0.24953395813429313,
 0.05968736015420506,0.7041093351052111,0.2493685066717789,0.595975962357221,
 0.5249103962693128,0.5013764156232574,0.8111204144398927,0.8483661885473571,
 0.600876644927276,0.005765306866509223,0.7021439043722388,0.6043103632002159,
 0.6674678132576689,0.27487243627617886,0.8783998337353343,0.7444914937629185,
 0.5391120066381038,0.787670012912652,0.14729113452627252,0.9065923649186368,
 0.8297329500077646,0.24527242794280457,0.5297757930859586,0.8465246455508723,
 0.3063477126246116,0.19048321201208496,0.01047993124058011,
 0.8165861935654913,0.2729542775973125,0.8372790937700285,0.8221171704977722,
 0.5543534579347758,0.9777582054487213,0.5137254345271116,0.38437355695782083,
 0.5975506500800921,0.09976244983377403,0.7792745856827386,0.379300147899848,
 0.34108832276264756,0.20213690088826097,0.5367763376627626,
 0.3432338131131164,0.8107801239213961,0.2822403968798004,0.1290307574261581,
 0.9608228203367105,0.3065056339382056,0.022997668356453005,
 0.7076566845073341,0.9681190797395436,0.12927404596485725,0.5332631215744421,
 0.623543387339548,0.2057994394966549,0.5911030902487002,0.08009175310558048,
 0.9964635972981177,0.26412773731086403,0.867383682122335,0.5961854504147582,
 0.937650196257553,0.7365679551313038,0.3110979356447402,0.7896413523482324,
 0.262910598703317,0.45696122612587275,0.7154270380760949,0.22234875124024678,
 0.7381943152391692,0.1415065223790578,0.23103440721402746,0.741304605450683,
 0.8870205277746681,0.279857080363793,0.23451554198110813,0.9553984020936395,
 0.6544194728376879,0.3322255577972165,0.9915655698733212,0.15588485536205754,
 0.7959165273282736,0.40154378100673604,0.3734783734445706,
 0.40914831673082486,0.14938454946411228,0.5858078163419186,
 0.5899429580039248,0.29864229978288226,0.07238232245472587,
 0.5479425508557457,0.05286944244255054,0.9259995619599646,0.2971921771908764,
 0.3186568318472689,0.3112154745591874,0.4879204160877517,0.7600134028909914,
 0.24981556547257866,0.7020732070246115,0.6828677386283034,0.3825251325240955,
 0.8202597599240037,0.5860312184756142,0.08514067449049034,0.4004465052613133,
 0.7577363409697625,0.2038496187537946,0.04558586834523071,0.9485789032542499,
 0.2459806652451193,0.008730158162195911,0.9975267626841471,0.507488115126589,
 0.23441669210313942,0.8537998391467735,0.2240170857544398,
 0.08601339237721817,0.6401354617934883,0.8460234523426691,0.7159908159616026,
 0.8971941332080741,0.8102268690161063,0.660280577836223,0.5878646389062298,
 0.27105834804368334,0.6612714128391185,0.11642113042861568,
 0.7884429580904833,0.6243528875859711,0.45261788133749514,
 0.07591053697695416,0.572625967478398,0.7842397476293499,0.17615456891169123,
 0.04229679487704496,0.1897950214767079,0.18595067382296637,
 0.22529436521623347,0.22636395592042446,0.7915430825914146,
 0.7351798874545893,0.4480620171521794,0.12944835951920763,
 0.09503211648094267,0.41734168625829515,0.5206200237848797,
 0.8200392543159285,0.8252523002687291,0.5231678032209811,0.15375738795380633,
 0.2673570591600276,0.8678039874995724,0.8018593273493339,0.7480869074471357,
 0.8828237706399702,0.8177461631943758,0.5786121337051897,0.07609615273347392,
 0.05826324070597111,0.8272778396191773,0.9461209292025959,
 0.48557528521203885,0.3503555241835138,0.8545779066690917,0.8701514566474595,
 0.41180864611905177,0.2645063163228786,0.45208762823592796,0.800767902897896,
 0.9437766568706205,0.1139990974231222,0.3940905780273142,0.8247355053682724,
 0.683128039164386,0.4230635335366798,0.8060529539868607,0.23923401718785509,
 0.46720188106892613,0.6134103732303969,0.6909421433283682,0.4794999524597954,
 0.8812127825220274,0.677361380514721,0.043814393728862644,
 0.38149658506043116,0.9826458088652967,0.9497978111150568,0.6170799867020804,
 0.11009887040862443,0.5135817430448119,0.3250403676488297,0.1413448749167282,
 0.2750159394404599,0.038201292831053335,0.8117244123846243,
 0.42062556018176345,0.23516098445996192,0.6135376764416547,
 0.3028235149500721,0.0858945817638963,0.8368461478900179,0.18671739516659802,
 0.10059059595778752,0.9074871576914556,0.9269809387490936,
 0.20722870724600861,0.8818403649333149,0.5772172743472326,0.6612685939868992,
 0.33102829980782267,0.6388588188260907,0.5121037613169418,0.7628767996324178,
 0.8258908147581275,0.2657979792183638,0.2727702842639542,0.9681684550269604,
 0.2583574491016263,0.053680532131337166,0.46467413858074913,
 0.8408505613389412,0.6840952008865366,0.42180327056694544,0.5819487289530367,
 0.1538768683796503,0.653349452654972,0.35314581986337845,0.0867518503603052,
 0.07958703206529671,0.8259424357239868,0.07562783125442674,
 0.16759827658940973,0.006881350413179277,0.0913912852936829,
 0.7751980106864498,0.6895341430403656,0.29028770278838534,
 0.24819930522734923,0.24682001359341255,0.2066029763445828,
 0.37356033385879717,0.18449746344300666,0.1824453261224236,
 0.14082652729180856,0.5192142167641866,0.44183682954752,0.6046367895006334,
 0.9545841596490485,0.07086166975940777,0.3724221043067336,
 0.10355675481087578,0.801331442014793,0.16130227226073124,
 0.09778682892864632,0.9901717968318735,0.6217221350245704,0.789266636258103,
 0.5416626594080115,0.49266226577654315,0.0015882387327419423,
 0.566576333276179,0.8999320037974579,0.9493985820983997,0.1905285091328437,
 0.607772018718891,0.20824006772990744,0.685747883011265,0.6576072219065549,
 0.42040477539134236,0.36437382023650733,0.7820080878929166,
 0.5023562331122818,0.6226259492500793,0.8992093192472199,0.04393950323619977,
 0.3511283827103695,0.8235107761248153,0.34746972230934403,0.967464564595493,
 0.8282773026515664,0.6952672721442097,0.507100576362756,0.7060369874800257,
 0.6150492934496495,0.8315064070485951,0.00240580070147578,
 0.31601682947634324,0.9541927711965353,0.5121620854450769,
 0.13157894682903148,0.9420851669226106,0.74812569590818,0.6258218332988152,
 0.5638323400138171,0.5390128505831528,0.7601484259419778,0.2521069608461216,
 0.22831490693916112,0.4106581907195582,0.574807176801249,0.5739538221204903,
 0.7711875998351678,0.5615296719938693,0.8695199759713443,0.05440489245438873,
 0.31805326711467574,0.45303836788736174,0.34551103095103397,
 0.18824813812905838,0.3954980347508963,0.4975510873810065,
 0.21630720364581113,0.4753358814061207,0.6567371842262955,0.6579908492147911,
 0.17357292762763998,0.8267771345990976,0.12198417141806595,0.009252410620051,
 0.08279054246731554,0.07227010171360715,0.9753576566100428,
 0.5975929445617918,0.6489993839812299,0.2604960198372335,
 0.023007664274067974,0.4830035567517892,0.042548191193651386,
 0.5109537482590114,0.6559696762952308,0.6636374168915951,0.19652705014525862,
 0.06702696750461012,0.5734036362000456,0.9575134602826174,
 0.09840939787685832,0.9897159131844091,0.27559241078935803,
 0.6182192501989833,0.6713853328647389,0.31797146852140457,0.3056379423851515,
 0.6216078020977822,0.6320581137525532,0.03297024046614905,0.3282824546811203,
 0.6049669238013229,0.8987718450788653,0.8366254817776463,0.5195567700270577,
 0.16364438485459443,0.4655774902832236,0.7300998662274115,0.9019502094816373,
 0.2446458593575136,0.916173909392621,0.36198571949921743,0.9930998941081767,
 0.4391185319324137,0.999018609242426,0.6904425454817739,0.12083735767850856,
 0.6216405338528556,0.3911539777958051,0.4976554666961268,0.6838517802970348,
 0.39669215178452133,0.06555171883170785,0.25376453384254183,
 0.6934760387115377,0.632554125121752,0.07826606601697605,0.21486016308629485,
 0.7042660412641752,0.08928485901411276,0.800693741633171,0.08236025061560626,
 0.6132845856127895,0.4810024645811224,0.4839104930604221,0.27237411825547164,
 0.24534451699659465,0.5356641899367329,0.04879025201801113,
 0.46498319891272655,0.8831343681366643,0.3413051275868003,
 0.43211284351826074,0.91600521478336,0.6445311464924134,0.4949346490305313,
 0.634352183439434,0.21991108177627072,0.9992416925814295,0.9769840359227636,
 0.7766281394705746,0.06404286334881304,0.9882077449267205,0.9983168877933474,
 0.6855423858535262,0.7522161089790553,0.9793185267632368,0.5383676342000807,
 0.7159642483695374,0.8079394480239404,0.5156797474427866,0.9814901945361332,
 0.8313922415534644,0.6576010210368606,0.3798367999419635,0.16266455811538305,
 0.1038884817036666,0.6260474811775096,0.14152898975852446,0.8079917700728894,
 0.19291747849143348,0.802164956883264,0.8914500152090901,0.15159699425121653,
 0.5726828082489753,0.9855110811447214,0.6474523442962228,0.28351727266216464,
 0.16046924777146765,0.773897962935911,0.5830677941376493,0.7773088142692599,
 0.22865604743057988,0.2903699820050951,0.9703913032454016,
 0.17097210499707105,0.9958433904519614,0.47750198346260864,
 0.11408474648399802,0.7858080876500645,0.13444031223160524,
 0.7449641623146649,0.35672118422959087,0.5438400944433563,
 0.27320622103776426,0.5010973373953158,0.4512280959721978,0.2503777730044431,
 0.27966873482351406,0.6227665911275572,0.2482569792632816,0.883327704479838,
 0.6145850782053581,0.8363376702672405,0.06159101520610921,0.9394036974231569,
 0.43602938426165716,0.4282188021630309,0.12700713884696413,
 0.17290094520600596,0.15103878855528918,0.9697277223396324,
 0.7692920785494236,0.4875142758816502,0.602067876473004,0.5569087328828681,
 0.03180374087260729,0.6149528718263148,0.34973142400425716,
 0.5447922692210689,0.8541028168784791,0.7925994025932974,0.448659043313531,
 0.38485424816860136,0.42016007998799065,0.16944596623981012,
 0.8316967376313564,0.08373132150718621,0.08413341760255144,
 0.8782706402049518,0.9524190267335512,0.08744725296970735,
 0.36797320033664416,0.11194810054744553,0.9796073988719409,
 0.22009810838564414,0.3117399113005268,0.23323196490281906,
 0.8353223449644173,0.2505296824329847,0.28988451999347986,0.3637770575105478,
 0.31175257394987266,0.38675180037665957,0.4234664449764679,
 0.3106141597740305,0.9482966795639879,0.8327200702035052,0.10397063887578528,
 0.4141406981257668,0.6788592642888014,0.9068853958917101,0.8124315156942065,
 0.6466477259804084,0.8354417811358329,0.5139725748336738,0.04142722026610468,
 0.7605899973678475,0.5570739074723785,0.22122542566999503,0.4267227775998328,
 0.15450003283554148,0.33577046538540967,0.622676855655508,0.7670399429285547,
 0.5979291151192653,0.18927664223440455,0.11959921719038957,
 0.2597910623247388,0.11869177338820291,0.4149036546232202,0.5104240332485656,
 0.5441438644550619,0.4301669245777271,0.6999623730770184,0.3186397621896089,
 0.5524731950882966,0.15637398639321576,0.7260726368390628,0.6000414340725558,
 0.8573227431038051,0.196695388583211,0.989783119157947,0.10459985475684253,
 0.42621806214199753,0.3545561553935982,0.7748290974196084,
 0.02336560861572523,0.997171244519105,0.04600834613307647,0.3110911505537196,
 0.479245063385107,0.02332615083380818,0.0186425963798158,0.4932589014274025,
 0.6295487206737664,0.508646836248969,0.5734178825747503,0.7118986653425678,
 0.8834686425302275,0.6543792669913171,0.1035592293253711,0.3473491405702913,
 0.8903519995811844,0.6847362853679253,0.19012972331467848,0.5187371486549629,
 0.7947528593933357,0.4917370000738174,0.1892886957595148,0.9304850382426815,
 0.21132260384552692,0.6410192731950071,0.9326083694853443,0.4188182738612929,
 0.19473278197937616,0.5142386666870199,0.5004776561888686,0.9544203174429883,
 0.37214972400347734,0.505798984997296,0.7590500902400886,0.4859349475635575,
 0.8097375022453224,0.08157441735695814,0.3083568189692736,0.5952121211761051,
 0.8065507179927222,0.3749395298636471,0.2709295283810489,0.3548121666859865,
 0.9027618391440435,0.8914770256160888,0.8140904474928288,0.296031790174471,
 0.09020783317714698,0.6822520989128827,0.5700429434908858,
 0.24284930243243308,0.4976899586130844,0.05776990694322559,
 0.44621614689748224,0.4174317658133051,0.077595077911642,0.8934336433001852,
 0.26513345069565597,0.40240144381740883,0.39318869892974384,
 0.4666655911015125,0.3586457305822869,0.9547002339821548,0.6974154345864555,
 0.8259989313957211,0.02315059354025939,0.19755248225667332,
 0.8411809030152555,0.8441418374137982,0.9183206858791403,0.49569892562856066,
 0.6004619179225392,0.09548260196233127,0.9551217939801964,0.4412271316853952,
 0.2023728787258765,0.5839016134032733,0.2619105405665848,0.2115736570715595,
 0.1541962492376182,0.4899953711376225,0.3342523746261693,0.6885490124951914,
 0.5571722382816833,0.8990087111906634,0.9450342082508421,0.13193438428132365,
 0.6301314843965538,0.99279429045097,0.6347880208585345,0.7497604461764898,
 0.8283191041074554,0.07779590425240246,0.7361295955328869,0.949303753745905,
 0.2924073562826588,0.8494133828368378,0.14044175802841874,0.885158155154987,
 0.9818865508937977,0.12425722075539303,0.26404666122968656,
 0.34343285904918086,0.7744800926031811,0.5650371785706438,0.9292377094811132,
 0.6349451452710799,0.9255549925689122,0.32339507710835935,0.8371988743018851,
 0.5410902265999716,0.579305341782455,0.9260988602452905,0.538703562904825,
 0.7161321357536939,0.9764160043725165,0.8189824145927409,0.6041143985991764,
 0.06696726610625203,0.2880825432887505,0.2228804081431972,0.265583967867981,
 0.5635371709822555,0.15302639086786352,0.04732911265222617,
 0.5865520400774008,0.4836627489965293,0.339380166415726,0.4431000181644582,
 0.9412399105975009,0.554400567807068,0.24128923074013753,0.2911832942393847,
 0.4702542403641745,0.48627173788580436,0.4855867264059348,
 0.37453055918267386,0.04310202470313251,0.9337416680374099,
 0.5461265106116914,0.1733990183835652,0.7131586439303316,0.7129315894766737,
 0.5132734704265984,0.026417902143011207,0.7386459907667808,
 0.4309383450588248,0.21346550206920145,0.36701157271701657,
 0.4016646992190802,0.7929800888441007,0.9842660029161464,0.8138163998083867,
 0.16289601897592465,0.3821773053754174,0.069997780869963,0.49854112708793596,
 0.6184643115076538,0.5646105228056143,0.7844202768461735,0.7667693330371871,
 0.39587064511985104,0.16566346748526584,0.7250234409297219,
 0.9901022346129045,0.47513819593349527,0.7703874465552712,0.726082493672498,
 0.12589064168358965,0.9933185702375754,0.9802208375002386,
 0.43038254961513367,0.19589200582614175,0.16949014189644207,
 0.012551541365349639,0.6764816694775181,0.4669613758350124,0.212484763606959,
 0.4982407515509778,0.2987303497630325,0.8625057228188429,0.6613650339301816,
 0.908104273260429,0.9658941461729936,0.027325270796119905,0.8933696713778377,
 0.7008082578380223,0.8383297748522113,0.8488409002464219,0.5956976420698396,
 0.43988677414124444,0.825644686245292,0.18150742602267478,
 0.22174810648548327,0.4678635653891756,0.11147559518894012,
 0.9226095815037154,0.48400017043281474,0.6421783322682052,0.3348814054436533,
 0.3320687681122617,0.3791060858839004,0.5551558961715907,0.7868542960293173,
 0.6102569842564851,0.12436481902321184,0.622022465803659,0.03875077777478109,
 0.507823263049354,0.18248717503161038,0.840309767214271,0.18639813262032556,
 0.5063018530352997,0.5710695998014969,0.5009892798754678,0.7222008788791269,
 0.6973057610688431,0.036028300072262875,0.15969747609467833,
 0.10144398942651578,0.5611556088530832,0.5730275952297983,0.6126498561962236,
 0.37247077080973656,0.35487396967985174,0.7029486604795153,
 0.8390931914562341,0.9113997106096791,0.3362785746452044,0.4024134221769742,
 0.22148222159224784,0.5526989691103158,0.33155609780038753,
 0.6401266283483095,0.6242263697415624,0.7239529251295751,0.8143448130100376,
 0.46985542521194146,0.24133281519131744,0.8978676105156942,
 0.39114950723999997,0.4741257182634564,0.3266824828289435,
 0.38462184308518854,0.6442452698837686,0.8200881268109497,0.9589243010368391,
 0.15601117854541746,0.026337767977565152,0.6731418863654854,
 0.4375168663635952,0.5908977031775039,0.11771603610427961,
 0.15952079643370354,0.026141328171886635,0.18006413852620706,
 0.16919795787483882,0.30430890194018423,0.33446199365711304,
 0.16429816409942666,0.8755463625938401,0.40142269266539765,
 0.9646320550772582,0.6133114802249907,0.5683430270985487,
 0.047228068020483516,0.5111420073484451,0.5534324691677309,
 0.5292082366873545,0.4588562514083472,0.8560902787456115,0.13576290539517755,
 0.6701085751569508,0.3421420342359782,0.8174697744750188,0.1177086107484715,
 0.11362416357949257,0.7407897503498925,0.6633435527105171,0.7377947580595235,
 0.9504584190209059,0.319022848633712,0.12780968425906902,0.11663309388536602,
 0.23344405155374415,0.9194124529898267,0.9759580253531802,0.3710620088451958,
 0.9693184237853195,0.872731591134144,0.5687301655344383,0.2121391110152555,
 0.32089598315497825,0.2040937884170717,0.2464955202025202,0.5487526262275535,
 0.3640072619784228,0.418205460844712,0.6837892639026291,0.6363924836491843,
 0.18475679740256346,0.634683274507158,0.7191627403293404,0.10637500112103915,
 0.8522674037636385,0.4766660178229001,0.9568838921648715,0.4021158042428823,
 0.5650318553870511,0.8096541672094508,0.15657021766850843,0.6506369349402317,
 0.9101899633701145,0.24181303087732653,0.10696488108106139,
 0.2132093675748985,0.17778916046432136,0.4982537332885735,0.5778732508791765,
 0.2856196065150679,0.3402827184898509,0.011144268019779857,0.841804046858745,
 0.8205254306002198,0.9616605118728792,0.13610561717354075,0.5587294581473837,
 0.7050731567040751,0.8303963288522958,0.678113108762434,0.19174517869093022,
 0.2715317494721997,0.03354495184356954,0.8030221476059847,0.5187412925741943,
 0.6962583015542753,0.993189168272272,0.4714687164276975,0.5786010139695388,
 0.24049526130818255,0.612944281494199,0.9064442076702562,0.7083460659762332,
 0.6718691758027376,0.3035124344575837,0.8637746704924396,0.11750833111006675,
 0.3557102570621151,0.5709646142268723,0.7968080685059015,0.6267015032134852,
 0.5203458714948058,0.9157147674104892,0.30140193950457406,0.4667292619450405,
 0.017622996708730243,0.9348266382224701,0.1231708488012786,
 0.2237993559751965,0.6405646593493952,0.44760281967612314,
 0.21746778165055236,0.469891441352973,0.9704565775605744,0.2264095355674075,
 0.7968353069066243,0.41797095114398064,0.09680031050653526,
 0.20780239572119685,0.24267842741785606,0.5689693257589427,
 0.8341740574508785,0.37323844304589093,0.9831012976527462,0.593144864697489,
 0.9617205632409979,0.9062181109620849,0.19324402803044216,0.2444472111211119,
 0.829041237069602,0.014875421975034442,0.6847165113482603,0.6972208114566235,
 0.6771825197732644,0.26846127749916926,0.1888984090953375,0.6751670439176723,
 0.7086509106546677,0.808562564502831,0.14335807934139344,0.16579018403087153,
 0.496612506571245,0.046963037661528984,0.7092842396682651,0.6941932220554754,
 0.8522645778642686,0.9537032917802731,0.8734403717548712,0.8280447632609738,
 0.6187673336282211,0.7092429554980791,0.3114391833705519,0.2960072381860449,
 0.578223837838229,0.3643971388480799,0.41523105460956333,0.6995263179071529,
 0.5783766877930601,0.3367209033954829,0.07454954832727212,0.7152389359476414,
 0.44874241824728933,0.8428304387792311,0.576614664705614,0.16349151742603607,
 0.7916166662429056,0.7596148501002634,0.8639805201166405,0.6234852758501203,
 0.3527008587364988,0.8706777589039351,0.2781353784775772,0.5669355457823476,
 0.33045309904770725,0.5425729344117048,0.3458463802031454,0.9078529527827146,
 0.22379508581556173,0.02458567879064466,0.6553430578014272,
 0.11751854152303653,0.6637146015585237,0.8116527201487722,
 0.37845653154183057,0.9408814790536889,0.5139220494484612,
 0.35326358202595254,0.5980288336333229,0.5639640937788446,0.5261234775536038,
 0.9484781557737708,0.5143581881318166,0.8234345731727978,0.7870268429398984,
 0.04810333103634512,0.22280144516084788,0.046007132666044726,
 0.4871059403026101,0.30570255866668394,0.29096872609149005,
 0.16125570568454795,0.7002705135512816,0.7154865657135154,
 0.41412061240467346,0.966887817979934,0.9664679845375461,0.40465637294076784,
 0.691959343955772,0.6665713132514743,0.24562216150611382,0.08672365556189132,
 0.9680818596871161,0.7361439691244952,0.679795837425595,0.2849057524231613,
 0.9448809543992019,0.23838819648196874,0.9603196396872612,0.7248893510957707,
 0.7420146986912078,0.5072407245158664,0.7985772210726898,0.7629306436080721,
 0.9336663939274065,0.7834470112783376,0.6287462131874282,0.6375956864473359,
 0.944759261516406,0.1781983475069473,0.6497480419130125,5.236575102266494e-4,
 0.7266906701885039,0.652605183840103,0.28516337712675455,0.52860641992563,
 0.5499978845252607,0.813351881051348,0.8357642782433365,0.7286052388409255,
 0.2217392206244425,0.7913090211302266,0.16989378388718612,0.5845290118732794,
 0.8893017189294173,0.3890931578664826,0.009322153619580886,
 0.1973400988073618,0.21746546596244087,0.1296752222560199,0.8352050684645318,
 0.14267592120019323,0.12158758802345992,0.024852630650988194,
 0.5444027393405604,0.9717239616714318,0.5314961889689981,0.4163774256931825,
 0.11764274596626112,0.006689647152625766,0.5267521326486762,
 0.592751808589094,0.81352846339217,0.9768756282518957,0.13209919830984485,
 0.03622294091001976,0.5158953965627651,0.36038449463505795,
 0.6013166260247315,0.4754043431922631,0.7217577355372123,0.5169356765508221,
 0.2590204021474224,0.03386907828696972,0.07449112354092602,
 0.8068257998473839,0.32812024517948,0.8989568318671051,0.27250850288849104,
 0.0991554757094616,0.9616410321979589,0.08870205687080168,
 0.18602739524476553,0.09630623858466403,0.16141704022556647,
 0.6454300873871304,0.4453305179805547,0.9392769404419288,0.5783993144964384,
 0.4366766497731759,0.664937503655464,0.09245196193696659,0.2136785555228179,
 0.9966876070450805,0.5935910319816228,0.9702647705956957,0.3031493262680629,
 0.036801073600551404,0.7579949000605715,0.6281477901609417,
 0.5861209184379641,0.9270025869409321,0.6809245591476545,0.9591847049711754,
 0.2883472345754857,0.6953680520424232,0.7767788297733461,0.7002932621926489,
 0.926219838150528,0.08069186938313566,0.4294109540378248,0.4688227734197763,
 0.4217324059866534,0.9734092552271454,0.08521673222934556,
 0.42182490644116055,0.07756581277143282,0.43971438927130535,
 0.9126906195708466,0.915814957987325,0.05524758075938796,0.15223113750986927,
 0.13601664286228232,0.26979799196013887,0.7064685183509157,
 0.6912734416326507,0.559241188714875,0.9856750915905454,0.6468048312692918,
 0.41925526691143855,0.7037031357859334,0.38895206159746365,
 0.7931193122208469,0.9258976011245429,0.08191454810071586,0.7763458055611697,
 0.8267810801472967,0.47271624042359706,0.4548428070574264,
 0.02432680155014899,0.68654374100756,0.9458616744277678,0.3345776231054205,
 0.24628113468401813,0.38507503462001924,0.0716909259248466,
 0.8482251609072176,0.5868994658470539,0.20042890280608205,
 0.09908727383822502,0.8593221712314167,0.0982213895066083,
 0.37164995399442846,0.21151659964725633,0.07169861609452521,
 0.28722846639416233,0.6200911458725746,0.6990303934407799,0.0899372986324144,
 0.7339745606387023,0.7113460829301795,0.9850802570999178,0.7676564701800008,
 0.7712798075528118,0.46350690446158394,0.5479627821866654,0.6588886004126069,
 0.8009150328220719,0.7922545034475762,0.12991548006157816,
 0.10763737199739798,0.42738931482525544,0.4549250668348902,0.862830373096592,
 0.8044579742823358,0.44492541993491197,0.4728791189324423,0.7950270627362738,
 0.2680236573349619,0.5348138383213225,0.16087696824924003,0.8419246392204345,
 0.408998209895985,0.5413354039842888,0.1300113003054223,0.08426949877569745,
 0.8682830645261961,0.28918556215236924,0.9832556409166865,
 0.047021429666928016,0.8845299865981577,0.5594186541151775,
 0.6722649756428047,0.4959395177905015,0.9603242289680088,0.47401658435333693,
 0.6163692714206248,0.05581576230892382,0.9890438397776827,0.9275851105196551,
 0.9194033709906029,0.8576604875297447,0.20906768232790574,0.7947883588596862,
 0.06956763594652349,0.8738786215126727,0.7746608815388453,0.5936605709898182,
 0.879228847522623,0.3575751112441312,0.07587511654749746,0.6564174014526651,
 0.08566772728917249,0.6765006076241755,0.7495138413440336,0.78706469845373,
 0.474515951259884,0.12395329360494745,0.42489709330344494,0.4832273788036491,
 0.9551448662405694,0.24214464584933315,0.9434843998388472,0.8201613380637341,
 0.13691620782624014,0.702127715510814,0.4195068112105953,0.7504900040851952,
 0.2743809440472498,0.7318761936864149,0.39730304508159675,0.2796937419456924,
 0.6665593342030686,0.5525835546807385,0.2689193089978722,0.3874374392206843,
 0.8453499483300664,0.559962863705663,0.8683196975867215,0.9706152882286254,
 0.6780628991833999,0.9377660729077973,0.5216450312294894,0.7538173250245176,
 0.47313372353254357,0.6770370300161295,0.3283434035518399,0.6223577347834524,
 0.42626618326996935,0.2390838548951506,0.18584872768029603,
 0.16297107849014814,0.16538830894345313,0.8862628647100729,
 0.9000424615441852,0.05489403385092806,0.6966075211865512,0.4921892589173379,
 0.6898316355670123,0.8226876008213796,0.7873447817614864,0.2830314464241437,
 0.8238752127161761,0.7556994548620596,0.9883937667257375,0.42810707268473425,
 0.04157138912044778,0.6446440522239526,0.4289842000080446,
 0.31341992692555376,0.1447655513132884,0.4740052632683719,0.5910632997248705,
 0.9728759021706798,0.9731170855928165,0.23931262677938747,0.2651461252856354,
 0.46151852689077466,0.7422932586091578,0.2560493373679078,0.5251582188507423,
 0.20884518458961399,0.3250614037473847,0.9973184661429694,
 0.03721102477280658,0.8281227327456163,0.4500486759238881,0.6612836816324147,
 0.9831494379471191,0.36289617738347935,0.23776185676476258,
 0.36265886339839914,0.6158164182006893,0.6716919440810621,0.8169658024706474,
 0.08450203486925922,0.6317384666738668,0.301146890344808,0.4727059372583545,
 0.6846975343973761,0.5119439784302418,0.3353780827304327,0.632997501623303,
 0.4621600431263665,0.3031840000131375,0.6520866833495815,0.9010896814804525,
 0.08654167817278235,0.5491058811441835,0.41329905635622277,
 0.5243024567933373,0.28394384022470587,0.13172086790699256,
 0.0772437565323254,0.6727708178175262,0.8957578308457516,0.7359721401491361,
 0.4604457602496902,0.6452712961229805,0.9313406112006462,0.9013013278499502,
 0.46426366691463117,0.6040165825726848,0.9505556060496515,0.9338280314796538,
 0.7223600276800234,0.2575015853424443,0.5583599801934845,0.8094930910191401,
 0.3067023736866511,0.13797791341895715,0.5966601102465806,
 0.27730154824549835,0.1778955431916569,0.630320331231082,0.09445175993769805,
 0.4059205264007545,0.5188076343877067,0.45246789746208527,0.9596703607692051,
 0.6613585044688942,0.7070709668053213,0.6455917455163188,0.5856329497741227,
 0.925910359917797,0.3717865039288595,0.031292211238488576,
 0.09998729047192845,0.6700510826882442,0.08227064028467646,
 0.020472891389867853,0.981656840685161,0.280570064678087,0.398195542349288,
 0.4780302887032998,0.257990170120602,0.5456429035909935,0.7107306306870577,
 0.16861548885589062,0.34145999400499927,0.5220228167715294,0.838086511655596,
 0.3559881104800813,0.6288523211499302,0.9292816350712793,0.6771305500109543,
 0.1297556379199205,0.7958929942585465,0.030609532817209684,
 0.9382102933732729,0.21483521508606973,0.9070201729708808,
 0.26708120240276256,0.5624892113349277,0.6572460400469514,0.682594534058552,
 0.6062517608270694,0.11565608682961537,0.5777653305286629,0.6937455728006773,
 0.09610282041773044,0.052913851916045584,0.9076949918808238,
 0.15312008540227084,0.5042035227259415,0.32535799252223097,
 0.47543990120667046,0.25180947337588755,0.46062921511870936,
 0.9418303661145657,0.16912910326468555,0.4372468911072682,0.6552923725486706,
 0.08500555148097277,0.6529417853300928,0.5753665380979239,
 0.30453133486272377,0.1426734892921172,0.8563115277637431,0.5890675795947486,
 0.9232502395578319,0.9148586358463484,0.32821087923126424,0.3067981513727783,
 0.3687139986807366,0.016964213196969502,0.1387764755539449,0.024390942589537,
 0.32516170448143666,0.5233642743129819,0.8395156428906256,0.8520066448529653,
 0.8401221403195661,0.5864459160171056,0.36012874658487637,0.6930353801711108,
 0.7069550945635069,0.6110529513067222,0.806703384963,0.9311411989419754,
 0.8251435676877208,0.05632108901293842,0.6263687885407133,0.6079994369261568,
 0.526737325818642,0.7308056954880642,0.7351611868989895,0.747849508125066,
 0.4653461391596423,0.18967482009876369,0.3947749844631432,0.7508927041635071,
 0.058271931687165046,0.6508492041378748,0.32466369441694554,
 0.12455217955925313,0.6997904093227871,0.6123739591094134,0.1983155578736473,
 0.6794114888466961,0.2165316314166591,0.34292552040445756,
 0.03567717478577814,0.011131060267120851,0.8198300350208785,
 0.2939370699752233,0.4831228681328896,0.2270076847339646,0.3465278234053998,
 0.4016055559630404,0.18508408811011012,0.8257121479993806,
 0.13521316504497471,0.4331898590216491,0.07680501444353882,
 0.2572682829635371,0.07852779529787057,0.8602807403400434,0.9115150713892329,
 0.7783940247011584,0.7254146132184641,0.34355973358715586,0.9275688361729355,
 0.2678761857749802,0.4095441248609123,0.3469901194580187,0.28405150758200004,
 0.3795279561233126,0.5000341513149098,0.31298833444490715,
 0.36869293965281613,0.8036493385206135,0.9754914381218227,0.6998338529838409,
 0.22391768635710085,0.47513353186528307,0.4636920106413268,
 0.4541857367319494,0.8579067621948423,0.5172435448134054,0.08692512256113782,
 0.7513947540963419,0.009626877561038083,0.386972099761017,0.9135160212700904,
 0.3971034106710549,0.7638416414564607,0.9966356587367845,
 0.031242544612265166,0.15027389602914631,0.05421320596376622,
 0.30556803355950235,0.1162645151349816,0.6351882627563905,
 0.21985989098219794,0.07737318317386888,0.10698750434863946,
 0.5104989009629901,0.3256289238794404,0.3424324135427057,0.16283748784108398,
 0.7256074357608666,0.9655814183845933,0.6844778020073279,0.19598321348369063,
 0.31938001353052337,0.48091173736110593,0.13804322887411058,
 0.18665614895519533,0.9152076370567007,0.9548492288014897,
 0.28694791196593317,0.3648081595917977,0.04957204378978397,
 0.1723623772711509,0.4952879422653872,0.128731110333566,0.42102850691824434,
 0.2710664190615383,0.8558800072653225,0.08608665505161217,0.5166347105386982,
 0.8029183064185087,0.32314245217605964,0.3094886161023995,0.85970950299409,
 0.9032196426345314,0.4842384608126207,0.8512371439823483,0.7012876184431678,
 0.025073963075834538,0.3082930610366549,0.24744835045759883,
 0.31201788786342133,0.9152533520050348,0.9393695449798505,0.9379828783436976,
 0.5855503382592807,0.2430434471683901,0.49803665020042087,0.6641858287662404,
 0.32740607493968554,0.7977738704358802,0.7654006366233695,0.4238365748563815,
 0.5937103950944834,0.07639918223746323,0.932717922705698,0.7214851087184103,
 0.37906688543679556,0.8774576653301592,0.569415628776402,0.28750997957422575,
 0.6044186983961678,0.31595366270444947,0.06882985306098954,
 0.34060004408508915,0.5193461597852842,0.7647603910820256,0.7877682396147271,
 0.5603478368272459,0.1555484464799377,0.6565309217406999,0.07144886212878698,
 0.9201984808048165,0.9385671536590366,0.7937745972919891,0.3751647253362893,
 0.8413418358384281,0.7599124343779847,0.11080701290888928,0.9939667219448317,
 0.781633159023417,0.35106328855187874,0.9448832321368548,
 0.010383364686498808,0.17886652988239415,0.8929714877352236,
 0.46284394347782665,0.9787192204139724,0.08091888192060603,
 0.5972875445956909,0.5148016281221597,0.9133843196106807,0.4795271673251169,
 0.8281784720641431,0.7112423193703781,0.6525513181608323,0.6880901946303615,
 0.09356765728320338,0.35694652871425925,0.060420896309398175,
 0.11112981136733646,0.6837744959442904,0.29813174062077896,
 0.1899040793050839,0.7714698231049142,0.4001256118020541,0.118561238675843,
 0.7601394885662087,0.3292982545198432,0.689312277169948,0.2065166569290543,
 0.0748260775914924,0.9579798134811977,0.7850281352079015,0.7441476431551061,
 0.9220629298144488,0.639889189195717,0.8886757611401491,0.9845656852197558,
 0.9942094089320669,0.5256080003808933,0.9563221428777764,0.6907953642562391,
 0.04343734689148493,0.7642696811579769,0.2929717203655233,0.7723859837004762,
 0.6080437455517822,0.5592974300220815,0.9430002304788836,0.8051568358462847,
 0.955743176729363,0.23528974243419598,0.4927627400732556,0.792758863262901,
 0.5007784209216583,0.02752199676864653,0.4231874040494473,0.82920727283293,
 0.2734461004054136,0.41263608449320976,0.6899399887651055,0.9893208402751179,
 0.76376414608949,0.46080197929526323,0.11163963461779702,0.1770066128752863,
 0.044762468250148624,0.2833874790307246,0.22921845091038373,
 0.31894811241504695,0.4517417681952669,0.8209520015237349,
 0.14088562070694277,0.11674578377353928,0.1466189437288592,
 0.8910769229847803,0.5315076382460976,0.5615883395091541,0.9589610822957952,
 0.9203016498088208,0.13558591655970342,0.37947038298586877,
 0.3361111292037735,0.14403069704031868,0.5586375186193445,0.629034650650606,
 0.5797183599452729,0.30360520728446894,0.16373613348789107,
 0.8824099701878065,0.5062734745231864,0.3350451050770098,0.30707176384198853,
 0.6631485869107624,0.8840979478031499,0.5549091136226852,0.272081463643576,
 0.14348155068454238,0.7059629987568304,0.921743206333669,0.9297049722476423,
 0.015629670442587118,0.37857390027873994,0.6129254186321504,
 0.9062882249721128,0.6326634047966249,0.20699807679854576,
 0.44833692443338335,0.6991975383619862,0.3331522344259775,0.179649058158602,
 0.2106299281083871,0.5794941994379782,0.07428774209369848,0.7948906146043079,
 0.16884843412517747,0.02312758307438767,0.1401272149788405,
 0.9535122087018333,0.0890669602750298,0.21555399857817692,0.6894347220518626,
 0.5296954642105786,0.5224866737512338,0.697318487034845,0.3342444555755303,
 0.3658892914583811,0.5200330494808826,0.1349737766376654,0.5616316006049246,
 0.19355837848957813,0.42957729796527977,0.8105630248178753,
 0.9281896985074756,0.2048592055653473,0.6818083900196028,0.7114416265484269,
 0.20071422417131224,0.12060565365757936,0.16214966445395906,
 0.44813273165828693,0.4390126652990902,0.45329058625910157,
 0.5034976581239857,0.9772438229219175,0.4814571503181134,0.4833359101633691,
 0.657977318063489,0.039059338741048166,0.8418437721445533,0.8379735703210482,
 0.11524047395601356,0.7366614862564533,0.9418982171166435,0.5137492865425557,
 0.4086934312663102,0.26704931908542917,0.1451057093576874,
 0.24841113399901393,0.004653531509441811,0.9180896909369923,
 0.9589571899129478,0.8857263008866882,0.8016641113392633,0.18587266641255462,
 0.43442618963161395,0.6286867488523499,0.8466400223228687,
 0.018519184578023373,0.4795184350976619,0.864434386077188,0.4295499720346133,
 0.44107969519571144,0.4359691356482305,0.6795937938417125,
 0.36488601764841777,0.3554879729217506,0.8428173852456676,0.7460635027944742,
 0.035680159115741406,0.4257092623245182,0.5678962744012401,
 0.9781424791229086,0.12333608785341588,0.6660592244308349,
 0.14519030780107012,0.14370794305565782,0.8767442373396089,
 0.8212130628511775,0.8140115220057542,0.7305066628554069,0.7752370397886148,
 0.5072857252814582,0.7260480656761678,0.034103731874013055,
 0.3303908754563841,0.8422941607209263,0.46752685250548043,0.4740457633638616,
 0.7791591843104141,0.5925622875777522,0.1447626461061049,0.7653490371241836,
 0.8036590733491724,0.7033412389749994,0.5383705620645189,0.8152113789458886,
 0.2832443289600892,0.975045454017814,0.5411826968831722,0.24862923102255252,
 0.6129673749720006,0.8671715146111993,0.8510252409201469,0.9872337066108421,
 0.8528888934806479,0.5066125725792987,0.9306549948344403,0.757954483696189,
 0.8101508620108064,0.40958507779766207,0.18764054161009946,
 0.4626206399837375,0.9211700971385075,0.8271241003581828,0.7377717072744236,
 0.894026488537822,0.9997207124114439,0.38257867257501343,0.310923000750969,
 0.4113182816971779,0.5167351728773637,0.6431803672787705,0.18533947541047402,
 0.0114564044839649,0.922392392225742,0.5340507664305647,0.6104479381856,
 0.5330607026343491,0.14361213740256995,0.7069690810829806,
 0.07929963366770232,0.29520092582639235,0.7904084044008719,
 0.4398915306185107,0.09318079725984463,0.15139310195145061,
 0.8678846323352403,0.3561021787729711,0.2117974158676328,0.05796703389922042,
 0.4306545355297394,0.13212778937037972,0.9887390944567676,0.7038901728657683,
 0.7453327773179516,0.006748769465233062,0.6505904197021163,
 0.4303985873166264,0.8268936455284628,0.7406920615572997,0.9980327464593625,
 0.856385416281891,0.3156472264879939,0.07579057635534725,0.4394672251182311,
 0.07876663509652226,0.33266861085875454,0.4748959055292039,
 0.48471433401853836,0.9281074740391888,0.6851987501773013,0.5906694396120252,
 0.48553114042802903,0.7889478165011741,0.7943701700087087,
 0.39356305086354226,0.08401296383394097,0.11544433892328165,
 0.9591235180185038,0.5956850725998856,0.71751373708468,0.30385660641907,
 0.5925447760179517,0.7691432228823996,0.2853776318565764,0.925120324042111,
 0.5478913087504911,0.20896717696000167,0.8277424802098385,0.3955854539866397,
 0.10627199815464705,0.03441527488810692,0.05498682653856146,
 0.740984059496157,0.7260926951997356,0.03119725412930885,0.06987268233726396,
 0.44517063017865877,0.4399193787354603,0.5916060499347627,0.9006763477443944,
 0.8669059646514866,0.2973291188465036,0.5567318102840386,0.04946376511530548,
 0.31303340151260906,0.303443181116791,0.310079301373704,0.9682365268533084,
 0.7049311524420763,0.6942904468484327,0.07876804945867077,0.4186921896944459,
 0.2844036818227098,0.638092256574314,0.060517167675536054,0.8889458299309473,
 0.43370680509325876,0.14077539949177786,0.40445218430703456,
 0.6900948828584275,0.30492757931110626,0.4243513835864884,
 0.23490058578094464,0.5049602575768727,0.2537962901210268,
 0.18268557156694332,0.08837617643235518,0.655599995843416,0.0687868848692681,
 0.17352320348813333,0.3529109209339921,0.7745448082761464,0.6550990527617943,
 0.32875605481167436,0.7736259600343991,0.0992226289317455,0.8060230686419845,
 0.8260765579240263,0.8636502805444862,0.12880809982462393,0.931662744432845,
 0.37724819029049583,0.43823606190155473,0.25393071370084863,
 0.09174333999159956,0.288557603504804,0.017910357467622928,
 0.04340890496575189,0.4739216756025989,0.9004725885981427,
 0.33332824472870637,0.5851626843877294,0.9999720367633418,0.5046693930763503,
 0.44895938579250316,0.43247065103265603,0.05763339217686114,
 0.6718472109564312,0.8012453788181537,0.394401586596935,0.5300872904198243,
 0.7614059427977146,0.08114303848752391,0.12127858956930848,
 0.8552424174162623,0.1441534319937453,0.27389594409085816,0.5852732380919002,
 0.8325307105260581,0.04168505953318036,0.472355682181698,0.29835392266019845,
 0.9651661557015248,0.716625571936889,0.1696894863689219,0.6414624810179342,
 0.6954477296316698,0.1973072517492256,0.9835154398813283,0.5945645783050421,
 0.24056664934064298,0.669824787446978,0.764036408759623,0.004292239614789972,
 0.6756865722821758,0.3610584989532847,0.9676620550961964,0.7956545880792294,
 0.2983098678353855,0.7577656034499864,0.6350407552791002,0.16942359967963505,
 0.4565079450689611,0.4651974315500309,0.8526823924920074,0.10336838579968421,
 0.8694094344150116,0.5732705003455777,0.4488321546270373,0.633023858349378,
 0.9177877888688143,0.18408360042598582,0.4770279462594731,
 0.33076234022106454,0.027681805135710502,0.8027991921697809,
 0.8221943742431232,0.9138573370597648,0.7878308934195539,0.4305167443442741,
 0.616360442017486,0.010775608836155248,0.04060084823004306,
 0.2962997431343125,0.41318143837240007,0.5406204339461282,0.7514564992789371,
 0.6539735871820967,0.679086174574866,0.9001488686993606,0.3904415770133849,
 0.8485818239098195,0.5206444763977192,0.4407621047065192,0.436799095106034,
 0.11818388702689808,0.6598323214464544,0.3983248198050967,0.7133311078206548,
 0.352796758101082,0.5558219564301754,0.8841061086899539,0.47200568074449545,
 0.544248453949968,0.42879112032456634,0.9278528511169358,0.8194517863692298,
 0.7717121585890295,0.18438973192643338,0.5930628259007577,
 0.031487478932190394,0.3012881079154821,0.5546940458318653,
 0.9809121555387866,0.446532367968421,0.0853830519638038,0.8354643949056442,
 0.21578791033630174,0.8058947489186443,0.3063080165476044,
 0.19965167435175202,0.4297447950751152,0.28590035995811025,
 0.48336751499617114,0.5545163817064255,0.07117250871072767,
 0.05713832344055825,0.9949395113100801,0.0542621332434009,0.4538837608320455,
 0.7854651012408058,0.8162115581935006,0.8898450548623875,
 0.004273067916597761,0.080749823112929,0.06954348289037271,
 0.6973675287352242,0.9598378640288137,0.3272426958673432,0.8035624410446572,
 0.5157811605794468,0.5917099816411814,0.4752437698657721,0.7968997886186142,
 0.4525438196591035,0.5076417211071427,0.36665974575602456,0.8416792355600536,
 0.09069812637768804,0.24310481036999898,0.7034782302281288,
 0.7573574543226416,0.8613517875305936,0.35142902554703337,0.7167043707867758,
 0.7531784228603021,0.22224392097614087,0.7886034179242836,0.618682275971542,
 0.1813919831781109,0.30904596006979124,0.35000921375087335,
 0.46600827869918515,0.6929778490918155,0.9021725822217004,
 0.13957862513458086,0.9196530821446277,0.9907722881043303,
 0.38575667082486687,0.8745080259256782,0.32673001587255834,
 0.4284542340204569,0.4523837623485164,0.9329350747002668,0.7165967183904339,
 0.430724607887749,0.9698516286597132,0.6568544044484376,0.44008189108516926,
 0.411438867397502,0.653761854038932,0.4847946790397647,0.6522875724997478,
 0.18490733933886283,0.774175993501479,0.3887551097598311,0.7604785428312097,
 0.8456749191734938,0.33183128126607975,0.46445155898052737,0.90188771420152,
 0.3339263215926651,0.4546755179376494,0.7376736225239099,0.9885335624274915,
 0.8072468863981619,0.07152247358500685,0.9107586900027381,0.8602020833353271,
 0.6485886403646682,0.9939792780935937,0.6761357833115744,0.4465355052475979,
 0.17370342123533578,0.6706073952760347,0.5588491933445243,
 0.006074839411199462,0.4732642069092762,0.6944650628748695,
 0.5599352630861506,0.08342404466110642,0.19353989699529695,
 0.6222398700670114,0.5281631217570518,0.052308586286959224,0.464023902338217,
 0.5041655915553781,0.10562593208495663,0.4932845527375255,0.6253483323856907,
 0.12843567692427538,0.26942111029380167,0.5432336627238594,
 0.6972900790248079,0.6233096164032642,0.7371924268056443,0.9599943915409153,
 0.9986689586595915,0.5303643163476484,0.8153848820351105,0.6076947901175855,
 0.6566720820556002,0.03399124708220369,0.7107663478003218,0.6981214516238066,
 0.3403598274998191,0.5014383115196771,0.056473837102207325,
 0.5295987098864184,0.5752176206579832,0.6944878936325133,
 0.034209771749592166,0.2453564430116919,0.21515653114899036,
 0.25580856556334375,0.20582128106386843,0.8513791573895803,
 0.8059024376325621,0.9227443226923449,0.06672989989085976,
 0.16254664068754932,0.39124494129646625,0.9488827698630984,
 0.049320064988586654,0.4683597914162352,0.8687097940861234,
 0.5653097324382009,0.8632402661200436,0.8826917656185059,0.7543006165956879,
 0.18629249130775904,0.6715033003207785,0.6205411985107894,
 0.25443012620164684,0.7342944446219123,0.08087195354556942,
 0.7450195726153732,0.40568267878472597,0.5608908587656808,0.3363297788353562,
 0.3836559585249095,0.1663711562954724,0.2930007701082812,0.6717780774905003,
 0.09177313266619513,0.7406217793855148,0.4232206340406801,
 0.09824406452358803,0.7852662164498617,0.7921033454378651,0.4278263220480406,
 0.5743849192917795,0.5776115412720727,0.5922202475810256,0.6255542990322691,
 0.9708923570963484,0.2626771851442574,0.5976309573451639,0.5144201371259077,
 0.16445872564531594,0.3912798162332788,0.96688246356532,0.9849514236986349,
 0.9897949158793913,0.9159820123759853,0.7177679138992703,0.5103306710215694,
 0.9086071846543899,0.40544078576304043,0.4137210450173414,0.5060106210003926,
 0.9754758323528125,0.2632698345468272,0.3767907607006913,0.49901818759397976,
 0.41780295859865035,0.11570341120715599,0.7796687299322879,0.130855029842615,
 0.2207812071558818,0.03095340458885598,0.2777633969918495,0.7146634142153174,
 0.49753842387834224,0.9074776528178939,0.2661304859778285,0.5615837321125383,
 0.1703986006801459,0.5877116298050585,0.6997152222408013,0.15952598347942937,
 0.8721842102300661,0.7545772963227398,0.04971126896701783,0.908867405500731,
 0.12352487484707386,0.19973703323651026,0.14352917375690033,
 0.9004537142416352,0.4663306718603427,0.2593833070912708,0.6982340077760234,
 0.7150783965922616,0.681037267889062,0.05343394438797733,0.5476946936448487,
 0.39088265777679787,0.43861427665093866,0.9784231471625023,
 0.025371331329064106,0.5640131367427623,0.8542306818731905,
 0.5143996758497479,0.7872138509931732,0.8044424483316452,0.42701193922960357,
 0.41289928660646813,0.1617608173443138,0.9411392186797621,0.8310598262286526,
 0.33778641754792926,0.8668074380570459,0.4679038499957162,0.8079193567112153,
 0.6778141410603683,0.29114561820437534,0.9375637767819047,0.5511081026802374,
 0.22056764080952318,0.9328462589951926,0.49820485390217817,
 0.5749074054083483,0.6563346052359221,0.15559485107140725,0.7707261700564086,
 0.339739596030578,0.029818280841453215,0.41326024578821663,
 0.9257496604728306,0.6772269796247403,0.1242133714395276,0.32426234286929,
 0.285815138769824,0.2546909408349234,0.8833777270718055,0.8017614022223619,
 0.2861511849968764,0.23456288388894841,0.6805141016711097,0.3401707852318995,
 0.7501167548909476,0.7798095861508054,0.39694648475821015,0.3315543478453292,
 0.8543666474625521,0.45798437353128074,0.5668052174460085,0.8476395343389194,
 0.10361270014758928,0.14780607782180077,0.951437690059236,
 0.29672439796409433,0.26612485953696563,0.6392283025980717,
 0.6165325255737744,0.2623380142968442,0.22527240399186754,0.8067394931156129,
 0.3374662091583154,0.8717873318188383,0.8226085117046922,
 0.024731541317398187,0.02016262855546702,0.9654057023484098,
 0.07633427238229162,0.745420240958224,0.35063122715779094,
 0.38466255570837626,0.4018129069813674,0.6086500958525387,0.6049529837125196,
 0.1640377684799884,0.19355543794835328,0.45943927295881815,
 0.41515448167392366,0.184265753589251,0.6885359402420338,0.16440162408143444,
 0.8069915669700312,0.9975994122677183,0.7178867313143973,0.13029265436332105,
 0.7087280289264378,0.9906162345081301,0.5436492693119761,0.9691339403738262,
 0.5729716090216134,0.05811395648216866,0.551217948806679,0.13586756871120687,
 0.27668071719086007,0.7212172843777629,0.91303127752905,0.7231643860913475,
 0.26658006889080166,0.2741658974656328,0.4954557778910369,
 0.11378270529716727,0.2592267867492517,0.9489205432489103,0.7855598039785421,
 0.1705079496100932,0.13521652995651,0.99940685510999,0.3282343307048925,
 0.10326746518093721,0.6900721495207642,0.0013171203897934802,
 0.8690350089834594,0.4393070358550051,0.7301050887157194,
 0.002429088850340544,0.4762650453751621,0.44823676857278716,
 0.4099714795710977,0.6236758526757885,0.13848221280741557,
 0.26097816798293305,0.10486269902161849,0.9269416276505282,
 0.6019374641714839,0.4987301814957763,0.19483155441012456,0.8714381643904876,
 0.10415982188725392,0.5291206215254961,0.8921459910264378,0.5063700708973539,
 0.23704549445125567,0.8214434030478657,0.6013179876043694,
 0.14955517417123898,0.6291901246851703,0.5411740701286004,0.3863555589703548,
 0.0046885312331643725,0.5581498964132534,0.48187050634127204,
 0.33011164177151886,0.5746193416002143,0.8685419411112821,
 0.27521630339740255,0.5157919280464928,0.2825747868347742,0.3763727138736349,
 0.4652029677190713,0.03742245628111751,0.4026578575539149,
 0.23141942993957854,0.17016863277076444,0.0246515000575398,
 0.7386187614168613,0.12651933557871364,0.4550263122233975,0.5819025003724693,
 0.22645913946407759,0.5105899596831013,0.10751682877676705,
 0.0256209860041835,0.9226276627642642,0.9976834347039181,0.6672553701683226,
 0.07131896017079309,0.3132817456685658,0.06642182806572272,0.223304873557434,
 0.8547281961704063,0.5298777725010603,0.34094197948874017,0.6381250950323781,
 0.6099556877752689,0.16669456880926103,0.8243657575743146,
 0.41407150078741906,0.7369010869712572,0.5042034912211221,0.813054627407708,
 0.32951547668813497,0.6392610592748338,0.49920104235521556,
 0.42282441569061735,0.2635474478333324,0.9642506098364967,0.9449348921480605,
 0.7707345074470648,0.6229895168490603,0.5944799565757075,0.5538474587170366,
 0.7246188297522042,0.9931874792011813,0.6463391318098426,0.6856758939776455,
 0.2725486391898011,0.4788759958000668,0.6802223311318404,0.39688860324795905,
 0.33991188652856363,0.7704372700794011,0.8031001883870761,0.8987972144351679,
 0.8254037032383583,0.8486526739400891,0.5890426942806832,0.8153397161688642,
 0.11528585119294676,0.27780853939297745,0.8164023289023193,
 0.2121582683905625,0.19540517472822694,0.31407731849688436,
 0.7494812013145917,0.9099608813553407,0.3299373759184163,0.1054174349243151,
 0.358684961682462,0.8805061523321134,0.8760735532913257,0.5616251757727648,
 0.5028871638061874,0.6416705722787928,0.9396284433279215,0.2935391418141855,
 0.21287789380060884,0.14455982440965165,0.8745457009823454,
 0.033736532773500416,0.6274848954000334,0.5336581221529453,
 0.8148631006386231,0.6339142517024918,0.5063576502065799,0.480772242624788,
 0.5051857273654676,0.19982517869489502,0.3977210374553056,0.7810403777293996,
 0.1987916768261525,0.05592736718018343,0.40132662486002924,0.310176365369617,
 0.07104157453741666,0.4702368970100682,0.03986694690989023,
 0.2076139110774219,0.08491469961325337,0.6892076460915755,
 0.40715833705175897,0.1401916191555228,0.26216240782824307,
 0.7601081946359004,0.9166616948444243,0.18720800847663066,0.8330599068089484,
 0.10713770278902901,0.005962597586492535,0.1464988626328061,
 0.969249939742401,0.9703153901630308,0.044557877636341336,
 0.26569883750343026,0.5944792916052941,0.3886683395731355,0.3038486116604783,
 0.5710793378221181,0.43856211106482645,0.99975915020889,0.03093256280427692,
 0.3453434343792027,0.9328035999472974,0.8720506286469153,0.16619969895998188,
 0.25014098650702987,0.6616976731068377,0.6327988119403813,
 0.11266776033210701,0.06884368128831897,0.9036547493247467,
 0.9988627791822527,0.9561805349285786,0.04368743150091792,0.5990342140405941,
 0.24277974876351116,0.158058514470648,0.2995860366687553,0.08591975698943721,
 0.2167477017110111,0.28536138911922027,0.12132468428982601,
 0.4222791268970426,0.36580642047007295,0.2381495567897024,0.2668223933946027,
 0.7873551756168553,0.924413082724262,0.7902264946688005,0.7800351750273781,
 0.07721272541442037,0.10477227210955775,0.19780683367761065,
 0.6744127926988551,0.7679406626390234,0.7636289044350707,0.8679486316511498,
 0.07747367335415989,0.4503562485682082,0.051296401020396676,
 0.03255015470526812,0.7049079700331096,0.41629673399555533,
 0.9529813559437731,0.032036801825475414,0.23837205607432876,
 0.8263583278596145,0.2107226682732002,0.2624223682479234,0.07542459663069434,
 0.4231316578999289,0.517629436936421,0.8149119384373998,0.9957130082584553,
 0.5361307268011704,0.33526874822338215,0.9799451362697678,0.887562173137557,
 0.7559864387546564,0.8026497764817407,0.8328879547142066,0.9716876029419059,
 0.9986331632902231,0.9801573734296902,0.8791637742255975,0.8048807901246577,
 0.730081335842693,0.4618591585546701,0.35730427053217717,0.7612612586237479,
 0.8578962728468864,0.6337494670973361,0.5988297407027477,0.9911364139663085,
 0.9983216992659095,0.454782949453412,0.6271789161713984,0.44194968774873633,
 0.773324976112228,0.1788176170793382,0.8147008690938917,0.215021710554403,
 0.9688825024283483,0.9603659508499306,0.25401481674564697,
 0.33426933203590603,0.092587769809936,0.9919360025570648,0.7094951758417597,
 0.8321702634818662,0.3310702336201451,0.7946782770087801,0.11100793908932005,
 0.37467534014123205,0.6059795854873125,0.8886166221697274,0.756401143747058,
 0.041932844800070534,0.2719470521323062,0.246581934831967,0.8548781157331161,
 0.7329939743585085,0.6629947441336129,0.3564530305159366,0.16113357285499674,
 0.3437390383447171,0.6860405453036231,0.15856536195181747,0.8467475482748221,
 0.9725191916263163,0.16409101123085867,0.6643331064029251,0.9980619969237683,
 0.15124693366689712,0.9244503090319961,0.7104844793602254,0.6022113074531894,
 0.6244946224068854,0.7761930075591532,0.01720883433413134,0.8440238096827648,
 0.8924588751309811,0.9666310714407624,0.6620329805696625,0.7662814472787707,
 0.11197343378162006,0.18665675518910707,0.874993745916803,
 0.17995567434978732,0.4486675010365302,0.550088587683165,0.4044501248333131,
 0.5856227910394513,0.8131335570524808,0.758731616790701,0.8926387293801858,
 0.025959472425129193,0.6547925396036164,0.9897783842756065,
 0.5196828555390468,0.9379763607298839,0.6101093012129887,0.25232716124842225,
 0.4431781079437068,0.9322008413670244,0.10849620948114913,0.3386387759751669,
 0.11284914040887584,0.7845389633140464,0.23257725885467384,
 0.7189414692063376,0.42557877783709586,0.9898158450085496,
 0.27676686672260176,0.8311507118857802,0.7018334610046348,
 0.14443186036936806,0.9925127833396099,0.6412159007749076,0.705347888681951,
 0.992082196369296,0.7543499817065517,0.7471806253502986,0.40977006350041634,
 0.11266950634539019,0.03250795909779758,0.5968088946668823,
 0.8282347565580358,0.623596177761968,0.9804457740659389,0.8264347850114149,
 0.373384957678765,0.4583698891801943,0.9388056464498498,0.6926865148809923,
 0.9589633087575798,0.18884103114913742,0.11404110759212649,
 0.5488835129868944,0.14077401379018228,0.16164376077148723,
 0.14856292296605622,0.9669753304896882,0.5451633795212325,0.3622227996031453,
 0.47569187390196166,0.6533700994084826,0.6021741851924358,
 0.14277602395416977,0.23464091558196987,0.08372247143312816,
 0.8163421264627045,0.49419873747423937,0.7337295659723877,0.4840880030064951,
 0.6639611008779086,0.7279118887331781,0.4868933643236881,0.26535641133633303,
 0.6226780273612796,0.7993157711519318,0.6451181017508584,0.8311400089853284,
 0.05757995601006094,0.6569863811744885,0.0178335761700118,0.8267953054979683,
 0.7468559366015336,0.44861278389218695,0.7477522701232773,0.4766514725028619,
 0.07761595689570067,0.6714298129879904,0.8919720842422639,0.643676050362965,
 0.7681074403901577,0.3752228621934408,0.1907997615700392,0.4795462801219109,
 0.6095915638938736,0.09439739981636552,0.22631896606169288,
 0.30663543551362415,0.584763828653998,0.980362313521539,0.5113834717713495,
 0.5602590005878653,0.5352773557469536,0.6614846776447129,0.9946973728494778,
 0.6541860667504917,0.9177421823074752,0.8690058589007496,0.4641927334984335,
 0.5071615326774886,0.19229254730507472,0.5362483209911081,
 0.07154743189830326,0.14383657800127092,0.7153550843182048,
 0.5927780378666304,0.12169275405416125,0.11553969085583804,
 0.5304901760237001,0.8290177995465318,0.5394567551271592,0.5768955044749894,
 0.15341937807977746,0.44874647759650943,0.4633606959373686,
 0.5691399360390791,0.43513270439406615,0.2716134791014919,0.81866174963293,
 0.4227856850198576,0.986346883484159,0.5389502246551209,0.9604659048414705,
 0.6822845631826326,0.622698198482268,0.13479244140295055,0.3935342279529688,
 0.6180412215291842,0.42109184993186677,0.6038550916785188,0.6952980562687406,
 0.5316775748496511,0.10370805494757362,0.9336180809003299,
 0.12376763529293999,0.4741029504334423,0.5356164746946926,0.742174057827147,
 0.18663282681601834,0.6898837748514838,0.35193842212131166,0.927081804516278,
 0.3083469551431065,0.22599184296787167,0.6759504060503891,
 0.16767374846911376,0.6889605217489716,0.9058368194323811,
 0.22200623139579267,0.7963634217190326,0.7272347874693228,0.5332789282416679,
 0.2893442099250998,0.5608107737586586,0.8911307740777833,0.8019214786857426,
 0.265141649863609,0.8917983556371266,0.3237254485276073,0.425833355415699,
 0.9721946013787768,0.714925793264506,0.9977755007983403,0.7566466285830898,
 0.9911245785326268,0.8104961555478899,0.8809093557954523,0.2036001133340618,
 0.4844065407192861,0.7064697178180521,0.27677325457842683,0.5786120958959191,
 0.3025291321923059,0.15157215166923388,0.08689964101295733,
 0.6103881141595825,0.33681465449375003,0.6497410507878989,0.5469867558374774,
 0.35969123311651385,0.4380670180211411,0.17034664165697444,0.606976136876828,
 0.1322658346531893,0.9727399689869265,0.25785397668719257,0.8449424768813544,
 0.7007152604350174,0.673782645124066,0.32505635053648063,0.9254879376405852,
 0.7756963752351328,0.928971063379877,0.7939574054748025,0.9211665924961628,
 0.3346356949766063,0.5009112956618897,0.7709519326010046,0.9175354726392253,
 0.047655671788129905,0.552960927961748,0.1535625237992162,
 0.17884756891575726,0.9072098437496596,0.3071475515047376,0.3631926328762348,
 0.11575930363856501,0.5665300832950609,0.2951876972841142,0.4145544985093338,
 0.44985943833258957,0.7959618174828782,0.6117487224853797,
 0.11861052668085315,0.5874685550220308,0.7019333178925842,0.6724190885247108,
 0.53416205776369,0.6358266264558757,0.5092566999637977,0.31377391053520365,
 0.7077971937953886,0.6630339308392846,0.6681832784246254,0.35699914345467476,
 0.6342369196237443,0.3264309714280975,0.5498133122553106,0.5110891917442546,
 0.8595782988560795,0.5661841874186976,0.07364519349797449,0.3168993102854528,
 0.4615682908592771,0.25603208469874494,0.09318692711437104,
 0.12127891603360141,0.5539589681815946,0.5751218925826221,
 0.03229557955022189,0.7010052362234048,0.746747752918913,0.13324651195363169,
 0.4570141005424303,0.3497467510108153,0.47562813415423555,
 0.28643341601076355,0.3717684590785576,0.28296185157660003,
 0.7277433535073865,0.1957896222483172,0.5940398921506072,0.8743599543604349,
 0.9852126035203327,0.4366989579370205,0.0012081526037220902,
 0.609164144378007,0.4698676907819549,0.7561145149710701,0.32086851491930446,
 0.22553239910697964,0.8687799767860335,0.8541696206336038,0.578589146612854,
 0.39697102461424627,0.7380956317122952,0.7230977035565124,0.6597552875634901,
 0.7745922001294135,0.6929732791193268,0.29656017435837534,0.5685810631358386,
 0.27175905311302695,0.25184046392839454,0.13612413345774688,
 0.8817330142369968,0.17943602081074117,0.2602620818870959,0.0869039751992694,
 0.6759531214469856,0.781740160979983,0.8753002501186417,0.28915192448830385,
 0.0725457263354562,0.3007151439162772,0.1748756235827209,0.5163013061541765,
 0.8046531697706709,0.38298566550429036,0.5546472098843961,
 0.32772491946292426,0.2444845844842941,0.639371959885434,0.2942261619277686,
 0.07318934933124688,0.823467672944515,0.9768473876089362,0.5217211841162377,
 0.7883290949767776,0.9539163213222609,0.5668473368808036,0.8071208198828841,
 0.9250548606251806,0.6175797073094668,0.15515048678805365,0.866524330599196,
 0.5910387559563843,0.0408839202054736,0.5584175364846742,0.6407664375732465,
 0.6463674356309657,0.680987410721487,0.6237876729381324,0.9084420518981806,
 0.4509864048197384,0.3802954071436323,0.8941822969995001,0.36708682766161094,
 0.7825884618239971,0.22233111291022667,0.5836041884494548,0.1763394980937104,
 0.9184380854831402,0.7606278898942044,0.97769792961839,0.179281558632729,
 0.14477342318578124,0.5828170414712953,0.5267787760881695,
 0.09426416250182634,0.9191646278608477,0.9482420226683114,
 0.16365057412584172,0.7287917125794864,0.5144353727857481,
 0.36171048365768566,0.6302774127840203,0.055020000631076926,
 0.8845274347395395,0.5474536533835767,0.7378776039268373,0.832031074888439,
 0.5808743739269537,0.6611527441779463,0.2928326650174611,0.22082857827456337,
 0.9186435248514756,0.614516655073033,0.14380688407949482,0.9422500191872409,
 0.261851590716766,0.8292539663617929,0.34496584113549755,0.4176914415541241,
 0.7653000302127065,0.18514970395040964,0.7582577923331105,
 0.33648690864492137,0.04566778458685361,0.9558773622238691,0.179501833766899,
 0.7115283511626863,0.4414568491389532,0.42824968829968124,0.958097299860454,
 0.14646773986735706,0.3165182015084482,0.6905328701316582,0.8525751445402133,
 0.9116523719891848,0.23661422176690416,0.8746162520362286,0.5942126348855464,
 0.13244725514324984,0.12573666557923058,0.9110228185092574,
 0.15042196550433717,0.29015600934442387,0.554929529031088,0.3661618700033662,
 0.008679827593266687,0.47853652017357373,0.9203648899654772,
 0.9038098964766309,0.9939640440965432,0.7833341762615501,0.5672697401708701,
 0.9048911708011308,0.2892385709940778,0.28632894232540107,
 0.05514895708170087,0.27510829907278345,0.9664669176883104,
 0.7868367439019489,0.873121306356135,0.08760041605234281,0.45408950285244676,
 0.36754955251197385,0.07054356699469577,0.9207603579248251,
 0.3110387431586953,0.9739785836386478,0.41301245276636356,
 0.26406801125878065,0.423116006933549,0.983544784118008,0.9299984286819267,
 0.058236550164154854,0.3616484779151756,0.7580744098235014,
 0.5855615176261626,0.4365012322081534,0.809458821967389,0.2792737897301517,
 0.32738467348975053,0.16510358511024747,0.5918884624544921,
 0.8039488421795519,0.38320079181210676,0.9985010214816032,0.1309237314711802,
 0.5706762189791095,0.7294436954325051,0.19857274362078337,
 0.18090598974906502,0.22139826386791006,0.15839514510851893,
 0.3621710808433324,0.23281298870888256,0.13396600005254922,
 0.8392399156562163,0.9264918222241436,0.05495363009603893,0.5600439399499209,
 0.646540156003495,0.1958078128240095,0.23716094576334712,0.5466467520300704,
 0.36191693078404774,0.4871482776077468,0.3594376643103925,
 0.45008076003632724,0.9726359917107186,0.3385316104970648,
 0.11378922568864824,0.7324869354137952,0.180098892685158,0.6986376526601964,
 0.6260030416567677,0.4690081186451429,0.3822125563221691,0.12886464149548493,
 0.7519226851046129,0.4774851937685599,0.3502842160595184,0.7215898664144194,
 0.4043153748247097,0.3227461430847107,0.5575806338717519,0.17780218965982464,
 0.29143351743483525,0.9870463120072557,0.9085949629587671,0.8838285532782529,
 0.330669607704841,0.6306125194029771,0.6699737009570248,0.125925911961434,
 0.3439536734206152,0.1317870444532786,0.06610340051316921,0.6882859565612698,
 0.43574232537859714,0.5924640777875895,0.7814563253462933,
 0.08430697750367644,0.9292575428538885,0.7620612029042857,
 0.40279657258193546,0.33743633147426566,0.9843418973064939,0.634890492027523,
 0.8131939203189715,0.7315304161934086,0.5846297968175972,0.946062342370598,
 0.5730323784656463,0.518021650305212,0.6490241815520995,0.3230894091484605,
 0.30740568862851325,0.9625920867894162,0.175883304398466,0.8926824678715503,
 0.22264550042190878,0.31644246159618583,0.5330873849938493,
 0.5660673537621543,0.6937048300593941,0.8070580304437112,0.4097539445746653,
 0.6773313034361246,0.8659000776362797,0.5633084729637036,0.9339543210934274,
 0.7119541839133915,0.28607017380412014,0.1807471567982072,0.5052302802292089,
 0.22715112750490418,0.5925741643507981,0.6139850610316618,
 0.33370024405243026,0.22791017562844784,0.9095328164104525,
 0.6280158184961858,0.5297452208761945,0.34164105602652306,
 0.32561124111675743,0.6096588208821204,0.6398975288554215,0.4692992841674227,
 0.5515702004878665,0.34891424629667833,0.6610564886779021,0.3184506914667903,
 0.5088908965642176,0.03562883612406431,0.27654137470779216,
 0.5747050559525403,0.3072235101243963,0.4575723190315476,0.1492292180313769,
 0.39454970739413164,0.3778017008841914,0.8704569239907255,0.1689771926652167,
 0.024005306832144746,0.554486029256033,0.324235954878077,0.4845643059857784,
 0.8842506587358072,0.04508010123912409,0.5458271481802545,0.3706564575689246,
 0.8557932181326573,0.11590693861225332,0.8018525799545827,0.8963421628127406,
 0.0012418667304626707,0.18234266860201942,0.8061857041295263,
 0.565867206334076,0.08495709431308218,0.9654083231620108,0.43811401578466325,
 0.21479097239283773,0.11688763112505329,0.5908634417920179,
 0.8780438682143296,0.7260527865127298,0.09017994905606319,
 0.01896091534153288,0.9993844837318273,0.2379678794401766,0.5997207761049739,
 0.3043612465163421,0.01585354669257555,0.4746345983794539,0.0416209373134897,
 0.42383330248589846,0.6417424810766337,0.6348923955166909,0.6824202877186075,
 0.7698957919065821,0.02138120552221645,0.9836706096907142,0.95760541894299,
 0.5739311231899865,0.7721973111091175,0.15174975425258563,0.2158379014773023,
 0.9006335551774838,0.13994591159769765,0.35249781160806704,
 0.6326333898309574,0.2523915638465821,0.5663565757149074,0.14280586996495492,
 0.647415223830036,0.5842380191331925,0.3867099076629149,0.8561540170439398,
 0.042850308042469365,0.6236363104583923,0.8911302602673501,
 0.39990515163009177,0.5946276135906664,0.5813344653637752,0.4096170938619975,
 0.042986929320069756,0.4992263501261349,0.5275352719085659,
 0.24185984911838054,0.1286592043412288,0.5873820006403974,
 0.10384967810985091,0.8345192246863843,0.17356790982613834,
 0.9905502953741903,0.5833209535999877,0.43392912074812706,0.7659382637259512,
 0.2788836400275736,0.6374785948481172,0.4496337295507723,0.2378586128301191,
 0.4811679720575395,0.1428199364606424,0.6870152959780483,0.5918099689914502,
 0.6656119147948149,0.7823825674307265,0.5439589261026668,0.3708115566842497,
 0.17437017934835553,0.35955767186802756,0.9481449607165279,
 0.2283893018881723,0.027374375276457807,0.6470728826383267,
 0.42174658686116195,0.05079621156811587,0.6858242080418258,
 0.7597810087468929,0.7073328058474592,0.35064908849933885,
 0.24132332854602245,0.767825443752539,0.6837029699053093,0.30139041527048827,
 0.9941556076407977,0.6130960817782485,0.9111413867887984,0.989071585389031,
 0.5604655004778419,0.3559596958053546,0.3411917944106376,0.5082105872635193,
 0.8283786432427611,0.10655116454374314,0.6595083055510917,
 0.14393253872801304,0.5780676658471866,0.041954876614393455,
 0.12013794951378154,0.9642495033102648,0.0589816684772998,0.527882467775982,
 0.11192820324912822,0.506551778698155,0.848849853556332,0.9844491828558699,
 0.36785831179162665,0.2590202236499244,0.02613725225960506,
 0.3814257321075223,0.6395700745219074,0.4629181908176838,0.36379406271797965,
 0.06255563119733587,0.3254490894814378,0.5943285054337695,0.6737310162285753,
 0.40901271318288435,0.6340134337164636,0.04228589529286919,
 0.8257255317932573,0.7171649450213665,0.004505829167624098,
 0.8543875470509577,0.5129438842507008,0.4135911688437671,0.294510557165241,
 0.7880746492668382,0.7751170120895298,0.41441630268061835,0.3992407503172717,
 0.49196731788463055,0.8511755726995606,0.36778352218467136,
 0.8721139245957304,0.3012690089824728,0.0726207047797317,0.18194954452661705,
 0.1118131381402222,0.6550150610556076,0.32711635805950634,0.8803034551573872,
 0.5915493342996538,0.46887895176995686,0.4591348779576585,0.8163769440019812,
 0.9087334818476756,0.7169973640439617,0.6143907766037069,0.735520955573473,
 0.42728102515266864,0.9122194910813226,0.8716442299515829,
 0.34979981735076815,0.35282356768510525,0.1286661098319033,
 0.9075823823474899,0.7948236592590635,0.1424926615044938,0.5196459930759083,
 0.20236818697974623,0.6721006879342922,0.45353111853762074,
 0.8351706206176439,0.404154613661332,0.9291764649024692,0.5243099132811475,
 0.30294316388264686,0.5033411913919977,0.5427967672793206,0.7824942689261346,
 0.18569733579157977,0.2943992075339974,0.6527650421145238,
 0.18565095489419092,0.36854774107865784,0.480221885920431,0.8048107204494449,
 0.3239098327854766,0.4154628567847445,0.5093457080106031,0.32948728789039994,
 0.008970612984159465,0.2773453754866204,0.912981105084561,0.7000895395365491,
 0.8577702779974135,0.9294611990758606,0.5638049935683708,0.35976495167313016,
 0.48155678447368633,0.12963050609807625,0.740324946750889,
 0.44651575277822264,0.3812090769733172,0.6096859354076908,
 0.27065393539495153,0.6722289341918484,0.03177451685433774,
 0.7167867065697978,0.769585110015138,0.06078719479925887,0.1396677894518883,
 0.9183506614997611,0.7299643172994627,0.3837306960472193,0.9360678552766238,
 0.22987459472257987,0.47221080834440077,0.9613524393821886,
 0.7531965772800318,0.8346504833480622,0.05692059352830936,
 0.36195604025062666,0.28875671804967196,0.24074332089596318,
 0.6927351640725065,0.5526774167555326,0.7119849664478339,0.3984834164359765,
 0.559284015992586,0.07487827101764433,0.05723945896877458,
 0.47100013935968454,0.3619160672462386,0.7884947949329897,
 0.42974607748750715,0.32138676434696645,4.8672110945724434e-4,
 0.7194186889875754,0.48040123785732025,0.739283055162208,0.5078467421486172,
 0.542699885988837,0.7042002178102186,0.4886470288128455,0.9733880157827822,
 0.2163592320817802,0.025470039578732617,0.8662599284922959,
 0.8332450959064376,0.13359593406853232,0.7556541374123091,
 0.028301239240743126,0.38746858606834433,0.05084728197899058,
 0.20899455609429185,0.08435466926608515,0.5849096929118531,
 0.8575481036171365,0.6276231368978582,0.9772535067254917,0.17008309140333944,
 0.27055005810733534,0.3661789235381021,0.5621354592215835,0.6955289635285163,
 0.989863435260276,0.938957284199056,0.8413477340121962,0.9189193792772404,
 0.8690857970380981,0.5248628708086214,0.7234541325446318,0.28623910555359156,
 0.48542886068823377,0.4008748544613707,0.7890388345295717,0.7554633343844497,
 0.2525724414082393,0.03826201567151477,0.16954798455335374,
 0.22105708595755436,0.27657692622565944,0.8345677137894194,
 0.1359795875796349,0.6085462463232554,0.17923854352359303,0.3041434022969257,
 0.9912587355168558,0.07560612742027073,0.6745061946338177,0.4427747858477443,
 0.6425986461447084,0.5446956213585207,0.24603714641489938,0.149125128300026,
 0.17457405302692108,0.8337145312666931,0.5318746707649386,0.4882813079485053,
 0.6489318051079314,0.460025410610587,0.0272333256648688,0.447360694882979,
 0.2818051067586308,0.32391585245548193,0.07248418385421584,
 0.7640254482501971,0.06553628062948658,0.7762548075697836,0.8442465553188856,
 0.9160899612393104,0.07455955974674944,0.6177747381091091,
 0.49565021673006426,0.0033413090174267968,0.5635119989090541,
 0.6698837419866209,0.646461611175398,0.2772159819868888,0.17109212293068454,
 0.1337241936792145,0.852087765757909,0.38265361664801956,0.1390032381065811,
 0.6963960870716814,0.32382483656122807,0.6626773012965191,0.7398998680805208,
 0.3351873842067995,0.3460358089255886,0.08584707789469348,
 0.10987070571294699,0.6534561510068585,0.6109996825006107,0.1054588976019949,
 0.705299025237138,0.12570228299767472,0.8022037558837133,0.340049114277572,
 0.44498083061068827,0.45955997376957436,0.626812787730754,0.8163818147040334,
 0.4666201892154934,0.2404805556504982,0.9493696529918649,0.5418242260632994,
 0.5603667477075301,0.7011192477525044,0.9267755732045618,0.8043075451254236,
 0.9842922914956616,0.04212832198190952,0.4760390383022324,
 0.03210110466321259,0.14370990364227765,0.9750344908285594,
 0.3140923273914309,0.19313790466383807,0.5531532794715658,0.7122576003890309,
 0.10113704225577325,0.9846950912719652,0.051716546357313486,
 0.06372537780668874,0.6931695497147927,0.13123131333411953,
 0.7983490840827872,0.0047977915172412455,0.2627630960874333,
 0.23310915719425318,0.4268459070869106,0.4581311333143756,0.7884284697270924,
 0.5858406966259359,0.6745789771010103,0.031591188752334354,
 0.8879636160098242,0.881931544337767,0.8199110133588465,0.7862451686708227,
 0.5102866454988425,0.7471898581978368,0.3704647863842784,0.10360886859190466,
 0.5285740148252536,0.7536152481817172,0.31215834385623487,0.7775681582819975,
 0.25462698377091886,0.6864029383873802,0.12876606798278434,
 0.01283155833080274,0.6097972378790131,0.7598292895765856,
 0.11247562791655596,0.5731235964963458,0.7121509423449123,0.5547107675807861,
 0.9422218031427511,0.05178882789594397,0.7749876608825188,0.3026334684289642,
 0.9866813390256571,0.675339463442338,0.8648985735329838,0.9242522638177552,
 0.10862635077853644,0.21935793934606718,0.4700672178812588,
 0.9347620239509696,0.17540330990182973,0.19952439396171184,
 0.043738555546208424,0.8051965758555804,0.7489848501053331,
 0.37914671187554083,0.7518458463889368,0.023153566099151424,
 0.7845425220911257,0.07601440248514668,0.6803289724569587,
 0.050826352995037793,0.426563140352339,0.6225122043807865,
 0.13804530153400119,0.32453638359173653,0.94260751818655,0.8176541677462488,
 0.288465331203984,0.1855055766928596,0.15146817265960744,0.3958810231819091,
 0.9079351266051312,0.25405796689311977,0.04476055396514067,
 0.22374301126626983,0.17661014075153114,0.4152780976143774,
 0.6021037777432727,0.9565796983149788,0.5812470386052145,0.41226631164769134,
 0.39436011347322464,0.718926815629919,0.7121935672757721,0.9648383176641026,
 0.06275460870758776,0.33540672728059306,0.4184073513193697,
 0.9652555555830391,0.7742492504769148,0.23634004840525458,0.4607427710683544,
 0.6160151379158307,0.1146465666613431,0.9584379933744446,0.6769504710412104,
 0.6088863874064293,0.10531240385181351,0.1178231327887942,0.4743319632829617,
 0.1556855426016276,0.39995609098418605,0.7530410888294078,0.8112197772645129,
 0.9441919585209289,0.5888227058355872,0.9971329458591389,0.4706626118252829,
 0.2884175083129863,0.028597807479593063,0.17918890693173983,
 0.5046905333159454,0.6184686796247536,0.6724079015778133,0.5222105553862968,
 0.33627272478130654,0.6285952016586306,0.2784915041140713,0.7090402245943023,
 0.03386069622507759,0.7920403960837672,0.22218972984030894,
 0.18791099436409708,0.6932095481342048,0.40412905757686435,
 0.3711976100504376,0.9398126312545063,0.808546401329818,0.2651211251783373,
 0.38261059845196344,0.7367879660252941,0.14931277938414012,
 0.8295790680764207,0.6375933213354504,0.6753880816687284,0.18460580841841367,
 0.8940187528446512,0.40303585014267107,0.6279493337269363,0.3002380893690818,
 0.02272770806898028,0.9588165955260382,0.7765499008097093,
 0.011525306422325632,0.9263068714319401,0.6172342319434383,
 0.9167604841335363,0.730123671788204,0.5896184380301934,0.6752926332954216,
 0.3952446435994751,0.16839645087742294,0.5127923169968003,0.9949126716508085,
 0.8211012453718288,0.2563887066267925,0.002945984973828808,
 0.41436104337449986,0.3838300151114251,0.47799397099756324,
 0.4419012092514044,0.36475681513654945,0.5042641351952439,
 0.10474803792560516,0.786138248583582,0.6989671884065743,0.7048884326827038,
 0.44871957028750964,0.5172149104369681,0.6198043237331912,0.741313665964566,
 0.6919229198149758,0.19431540550246906,0.8760955510177011,0.9160729368104984,
 0.9528706020912836,0.34437750001742995,0.22327415137520457,
 0.7708808655212347,0.8263655472595226,0.4436241715337713,0.5568785781452795,
 0.840227178780272,0.2013083783346321,0.3336931037757228,0.579841027831109,
 0.06749353514378142,0.8542988706091823,0.5171586855524983,
 0.053363586154291265,0.9454374554916853,0.8635481022358503,
 0.2871235081162419,0.8362275840850116,0.05992059918123793,
 0.01747508439759926,0.9786046820718682,0.26291455785877516,0.651585161469708,
 0.28685819247807565,0.9446774965097926,0.12713755462220644,
 0.5483340762496518,0.051036598430240554,0.1197970857226105,
 0.24812959732815698,0.5591821680002675,0.7580410628986035,
 0.33484876723721724,0.14246159604775577,0.15099431527136098,
 0.5997540429222692,0.3787752496441099,0.3790902475512692,0.5820474786485847,
 0.480285277827729,0.12353907799118091,0.9850197849970506,0.5977080214091939,
 0.2280275105662759,0.6877702636097627,0.9916753111030826,0.23293872102737245,
 0.23357797304680605,0.5551180812777269,0.272829796880103,0.08326861289637388,
 0.6043763339976214,0.7076846666054285,0.6122061072815306,0.33259213790602093,
 0.9807806722029141,0.25424304610821813,0.9334752063496302,0.8622947617997702,
 0.4524476764212564,0.2630188517725832,0.7821857927450928,0.5539191752734456,
 0.31247179267054326,0.957363786579158,0.8856754207445336,0.5010583330396009,
 0.237418324509594,0.9258536995589729,0.9251140873659173,0.14278839006746147,
 0.449572402374814,0.12076979961951873,0.9811547583604923,0.724884256069479,
 0.22361434129992042,0.20188867269461708,0.3792311419131661,
 0.053036648331657665,0.2093766729928208,0.27113812954788186,
 0.5380720075198302,0.8552399443455182,0.3504466165153258,0.8897983955850044,
 0.3490004527285413,0.23918994347098643,0.8582786768494103,0.6416780005785452,
 0.7020124886679955,0.751785525260551,0.48621073470685683,0.3561483853796411,
 0.8581479717008421,0.12734003692586893,0.13606439507325296,
 0.49434794660231207,0.6568866518687525,0.5011834312311925,0.8365247062012742,
 0.16771014395381367,0.5485606893822642,0.4444698646531535,0.7264150306504599,
 0.58962694493476,0.2045130916105139,0.7044497918326197,0.8631985013220369,
 0.7699235254793206,0.5061709528696042,0.291270942711788,0.5574755938573573,
 0.7592316553486239,0.22778185832272513,0.721304173556308,0.936214138488099,
 0.8361888722057942,0.5863749095291622,0.6937837598119271,0.5601948588401023,
 0.6540126367112471,0.5545658153699304,0.9354635130550368,0.05080811310755673,
 0.056465230468235705,0.5278176969491657,0.1367655408765821,
 0.7349516669859388,0.28606745833762637,0.5661106556835579,
 0.18779004412076095,0.058093775090133914,0.9509084311084106,
 0.36632257161716597,0.2659944298504098,0.5327000570430164,0.8819257887032957,
 0.09970917303673654,0.3805936050077793,0.9985784737843403,0.2237620690408343,
 0.1041915357522718,0.0556629004428143,0.5772692183577834,0.8382635668745931,
 0.5642698774623471,0.33414382370849116,0.9767376980757317,0.8422450656126096,
 0.9015355241642774,0.5887377595247713,0.423144197936939,0.6170325362050497,
 0.26839752779465864,0.8220961451293143,0.22668587128435413,
 0.49693795754363057,0.7587666046283916,0.4569487550169118,0.4427001041656168,
 0.2561534526977164,0.8460030196573085,0.2616993920776415,0.7414510353085374,
 0.571265654788992,0.8359142891699839,0.4367081955699127,0.5804193029082747,
 0.38175047761526104,0.9478255325659478,0.23644023377439405,
 0.5781091105588199,0.5306102834618984,0.06713863986075497,
 0.16803215623222556,0.9450272745369505,0.4252273584537203,
 0.30680958480373577,0.5378948815461877,0.771124185992421,0.5987897428048918,
 0.9995923493762168,0.7018212898242646,0.2581871812423506,0.3755690142015259,
 0.9792116076285644,0.433557627168487,0.27722452793765917,0.47815343079455874,
 0.35044416109275744,0.24488791526376408,0.3250197293024395,
 0.5898954694627421,0.9196766086895155,0.3237631724450888,0.6622323886964382,
 0.944541536849379,0.873331821061859,0.40930071957207226,0.38300903402784425,
 0.940517657842766,0.27722184063296396,0.628419170921437,0.38317402832959435,
 0.5619477964210182,0.8082156621925227,0.960102096665433,0.6593547983039584,
 0.16563277563490042,0.9487961836579831,0.3957143414672263,
 0.42413941442926106,0.9918415044613436,0.23960406030190162,0.204179223336593,
 0.8657016700788167,0.8549873020745996,0.7395506657062799,0.5231944480877925,
 0.5688836325562365,0.8127144492184244,0.209259480760276,0.42489191475514065,
 0.5352630760765861,0.36827376559044667,0.8461285652780735,0.9638699686509742,
 0.8319286132840725,0.5334274635564116,0.964099433592724,0.3356181957523713,
 0.45192292201826434,0.15575888860982,0.9239467405947924,0.43724485416756853,
 0.3446875987013138,0.7844672046893538,0.3251419166751287,0.6742533242013923,
 0.3864013329880177,0.9073273739995383,0.5521248997045239,0.13967592861240385,
 0.4718677212166349,0.11408656951250329,0.17102707830311492,
 0.02618426559134268,0.21567509680746244,0.8124729142030067,
 0.14344494099011518,0.8812169865967144,0.5852048225579931,0.7369791293811309,
 0.910390526068264,0.5796494535439763,0.43985049833426826,0.9556836610965582,
 0.5692710702832587,0.9108683058667965,0.9354538462936134,0.43283135829685027,
 0.3211521675169555,0.23564769891168225,0.3976346441863787,0.4638570009512085,
 0.6659636299776504,0.23038418153668871,0.9493655962038307,0.4344645061338609,
 0.886580215049816,0.027887349946999374,0.8402761315701289,0.8609458350911086,
 0.02004259465087288,0.1195557427052154,0.9153328748077753,0.9920998933818703,
 0.5973528084417252,0.7559749572398353,0.1894834445455883,0.808368008619359,
 0.4758116246722437,0.39417181872571105,0.12872551081215577,
 0.6076090833919059,0.03844784725339678,0.8362601843678782,
 0.39937552493915807,0.8872735394789166,0.30446828227126377,
 0.8742959676016622,0.08884918205710468,0.6257772118458055,0.6054496870228987,
 0.5415270549231195,0.8222550430661861,0.3951846527801003,0.40409317750039964,
 0.514702432099557,0.8609698292438637,0.7387605234508026,0.5444131746723919,
 0.4494918086455969,0.8568840657805186,0.5292320387881568,0.12318023614621865,
 0.19788818921156037,0.36258447809936767,0.9582656451088959,
 0.9115590426401157,0.04804921773331383,0.9325558056222487,0.1818802877761666,
 0.29228916154130546,0.7459820848448047,0.904209049574801,0.5411242293681786,
 0.5809367207318694,0.8841524975992546,0.7452908575370308,0.3404295996295099,
 0.9884308152127812,0.4812552918868712,0.3552175959248294,0.6575357485902364,
 0.812588623423585,0.22479640093277087,0.9725456867602539,0.21657041597650029,
 0.7313195939093412,0.3309807082206284,0.5522538347345982,0.7727566520070602,
 0.7060924232126671,0.49700363169539186,0.5121442900124967,0.8966741067770587,
 0.8753392932182554,0.9115080776051017,0.5918494068598493,0.0873836246421319,
 0.3576647290152629,0.15582327928125428,0.35732441344707044,
 0.8716787624778098,0.6001237277887224,0.47343386181927405,
 0.12299762254944047,0.5890663950671273,0.8584466152116447,0.5275166053496246,
 0.4020945217586064,0.9895765851991842,0.7170146830567328,0.04959482669735493,
 0.8321676305697006,0.8216297501616873,0.553660977791052,0.03451538630210893,
 0.4406096431945197,0.469359481181691,0.43986851555720463,0.43174702557736333,
 0.45030475647044543,0.07583211993279038,0.41504253484946185,
 0.9280067140054106,0.7967142487389491,0.8925664400491192,0.01751482857101183,
 0.30217024993575337,0.7532864162998729,0.5135723474171126,0.3817069051037906,
 0.8724963184265628,0.894484899199807,0.8520447244111534,0.11413677494805285,
 0.3445007974827793,0.6273367181536219,0.6897985126592739,0.8588829109671681,
 0.08141258973167387,0.018517693765102705,0.48976592293808807,
 0.2788294903457553,0.1609015059244714,0.6355764445895347,0.590525825363754,
 0.01337091819492664,0.6051053565811361,0.13676269197916813,
 0.4735602217489884,0.10221418626193457,0.5882225675135806,0.9904967846250379,
 0.5726519481628397,0.08426459705319256,0.9629044654811343,
 0.010757112664782742,0.37947648128397216,0.9307187788957993,
 0.7323235380284884,0.583548185846984,0.5726828972942527,0.9231016358987474,
 0.15689181451379208,0.9112152733487607,0.15541546992351496,
 0.17413719638295888,0.7404922300009543,0.2520784930349812,0.8456196746900937,
 0.637480056820899,0.6680448458937193,0.311962915420687,0.9146335480079526,
 0.41874837774768814,0.8894856567209009,0.12029081605040315,
 0.7905397127935235,0.1142177621294238,0.13956023497478776,0.9786363078617947,
 0.15470674010584018,0.9366036524881944,0.007359812501472773,
 0.9669994890668218,0.5097067515825012,0.22361503796293913,0.654619158813119,
 0.3753035746254906,0.1687070461088096,0.26115243688143597,
 0.09705797800587268,0.04778794231020278,0.9140186067342547,
 0.28238522081937545,0.6988188123204768,0.33638979494062016,
 0.7299944931993883,0.3683389211791165,0.7665091402822416,0.627421282351023,
 0.0018735974940584654,0.42647494125265994,0.2669454568881414,
 0.26424255209430203,0.5957053573664929,0.7206879839319459,
 0.40635442528306553,0.9224116908460953,0.3215188674327103,
 0.12594508309904018,0.27558650175507604,0.3267669111801541,
 0.12055679534463404,0.583369025249446,0.34753846798188304,
 0.23121931817186236,0.15736825110405472,0.7397597823446689,
 0.28569871293693005,0.028938082377734098,0.5283588787821765,
 0.9767107534628988,0.904899659402103,0.17374148467635375,0.20202366655635773,
 0.785262776275816,0.48393132688851226,0.9680126607970944,0.12843676967162898,
 0.7818515447218715,0.6953803588043617,0.32296041448519597,0.5619650455585953,
 0.36349719823929316,0.20531215155242521,0.46697928736200034,
 0.709075238170152,0.44085302501050516,0.794260945938703,0.6988687660876124,
 0.23346020769971387,0.33403538895374085,0.06878499977321895,
 0.1479258920769052,0.17187588141486687,0.002065689481787536,
 0.22834011734569337,0.4676621831126673,0.5671136427644248,0.7144541166027754,
 0.8676018088126507,0.17154916848089197,0.1948189497385232,
 0.001975331745940778,0.47875157736964213,0.5557426368191042,
 0.12326345358739332,0.8701270651147989,0.7985729769976904,
 0.05487856017428172,0.005238694260496279,0.21685514282214247,
 0.13540110331050936,0.4680110285023642,0.8557664061754309,0.2371783001843808,
 0.304928034649257,0.8806668503208941,0.13611722645292001,0.3855884391287324,
 0.36534946824337977,0.4424737904087811,0.6606301121136223,0.68763065740993,
 0.8331766065650398,0.6398070946870252,0.8438140251390612,0.7268724686507838,
 0.10652375514412704,0.7396977783953771,0.10420810152132232,
 0.06589819144881959,0.6921800791446046,0.5666593915717119,0.389392692827199,
 0.7016056095857365,0.5244268436459901,0.9972370669520376,0.3584028662155003,
 0.35478505137850336,0.6557355955498935,0.9940187867704555,
 0.01884853536376907,0.6608143317643168,0.19851491630338147,
 0.5332700019649067,0.1700279841604302,0.9336013445318906,0.1755976293946938,
 0.17374535108724287,0.42182344468268296,0.5876451736609287,
 0.6529145997795147,0.33727871797277786,0.7707006398309796,0.791668629023745,
 0.6724375179827055,0.07364292169624143,0.05832580115044561,
 0.9713295962731344,0.3416563016137756,0.29184908439758783,0.2664130660924743,
 0.0613334991615907,0.5309991419505159,0.9226096157009379,0.31582326730549126,
 0.9205287373593335,0.31064954726293337,0.8498684145042248,0.0719570176760429,
 0.5693948936082238,0.32603993449062585,0.03277666991970851,
 0.5333538001168797,0.2104921694678531,0.5652959532586678,0.11091232364799686,
 0.10748599754628607,0.808060176626011,0.9533081062051255,0.03508044606549843,
 0.5998652892592014,0.429381994801312,0.9627030978699067,0.9403875217435628,
 0.43265515998537096,0.7183408657515987,0.8398091912816257,
 0.16296715792729122,0.3622480176468278,0.8592036950056271,0.5046134370878603,
 0.23801769693092267,0.7756450429670533,0.13034365524825497,
 0.7134916239628641,0.5633508659077072,0.0689745905157183,0.6159618092597419,
 0.08623616737703255,0.40314900891764616,0.21638898034954024,
 0.2728863144284661,0.9445113649492856,0.1578016468181942,0.6083471166815183,
 0.4470169793135438,0.34762928662902215,0.5446245104700993,
 0.16056965247183408,0.5595094619421515,0.6677447523978947,0.8090845539412276,
 0.44426724436566056,0.028573756795438054,0.9468382161902976,
 0.29266986213066815,0.42719683887588156,0.654236101945369,0.6283383360203428,
 0.15716304221119792,0.35013251429724956,0.3220834910053435,
 0.36482587741892136,0.9987901556450197,0.9906793915789788,
 0.33094808453576774,0.9960980428512856,0.030845435332018223,
 0.036305780332423154,0.3569930160751733,0.8970572338540548,
 0.12606095018823826,0.4170445432642075,0.13025943190010225,
 0.007006434779332271,0.8531496873872261,0.6087953742027479,
 0.3099281753853892,0.21248800585667094,0.06870831037975844,
 0.19868677409675006,0.5517844030797886,0.9409953769661841,0.7999083365165361,
 0.3945050826383366,0.8547153050122457,0.836361469133911,0.6430761891186383,
 0.9897820671889331,0.06820284068750937,0.48493251188672337,
 0.4556993837877431,0.35872197777497605,0.4156538032180188,0.407939315144874,
 0.8119419347064569,0.19202810580183283,0.8434007137990189,
 0.17784796792081825,0.6388289536303052,0.8662378419795534,0.8847233675072874,
 0.4215071795048264,0.414766896465092,0.32012644811133795,0.06769382816831582,
 0.8235246358988025,0.5896078604517556,0.36681571215712383,0.4270137369242535,
 0.17133724489258273,0.24854115895095985,0.08644441014410142,
 0.692554869281899,0.9092361043866879,0.8411278592258506,0.33715923031433725,
 0.6052757403036672,0.5107510357240701,0.28736187636411537,
 0.17177584602812712,0.14578671040007085,0.5309276412553652,
 0.5340040489563689,0.8110263933012422,0.5227270576631341,0.6219519771045954,
 0.7160067628358234,0.7144018518697368,0.656348649972408,0.3657531022241257,
 0.9631856227953164,0.9778464011399803,0.33625354933603335,0.3986693102208285,
 0.9800377427242095,0.7493966454592726,0.1800759279133346,0.06256729208129153,
 0.44958349677312626,0.9816432492702846,0.39766839913566643,
 0.5722852054721456,0.6997915176966759,0.7412300954872039,0.10189867315188939,
 0.3120284320301743,0.26167628853645164,0.285156013440384,0.04237993154782105,
 0.8419308961776136,0.44225261721420195,0.15170372664451737,
 0.4076860991084349,0.031611632627378206,0.8131732720903752,
 0.5593430626649452,0.7624184614571387,0.4255537976340359,0.9358192041138755,
 0.881691018577612,0.7330885455615418,0.5386546026051032,0.06961928093058589,
 0.21081721273766973,0.537960883290044,0.8290250578683321,
 0.045703139419746575,0.5413184917485662,0.4941094747081305,
 0.2729636481279356,0.11453062175335582,0.3698332634709304,0.7042152587714476,
 0.24161550332049786,0.9856480357816462,0.653164688214169,0.94645012226668,
 0.6806495588829444,0.9251483670879633,0.9986958906542807,
 0.044930580918976615,0.9126989844003849,0.20338193317739228,
 0.6250989396723644,0.14485717907171802,0.17120726208139547,
 0.5709741984526985,0.23087601820013426,0.24859406720715205,
 0.9420650691839575,0.9977053904058525,0.7566277145713438,0.31300117538795935,
 0.5384013840511737,0.41378550177673556,0.7644444959839015,0.2610825932812857,
 0.8888050922431555,0.5350809383395706,0.5773378811488811,0.9815892488203697,
 0.632460603011767,0.8876256117798551,0.702963209281828,0.948102927546799,
 0.8620137050632172,0.5897677752961862,0.40638858407543477,0.5686188352251358,
 0.1415134241626692,0.6915463552892825,0.39216716067775825,0.5098248228106426,
 0.3318771770408908,0.16484810645644477,0.6649930128587704,0.8671642259285792,
 0.44156536238631516,0.505721628230863,0.34352576537564683,
 0.39275219903852954,0.25592666091042826,0.26759532410604203,
 0.46970599865039275,0.5244676622963761,0.7754396276661149,0.3608118950561676,
 0.33323204993626376,0.6796165109844521,0.28691829959069337,
 0.23326956986944758,0.7031332972339479,0.7426739943956018,0.6829675814821008,
 0.953428256353154,0.477923668424434,0.3589296633612793,0.7830048079280071,
 0.2565958863112253,0.6723773488716053,0.9280478253933598,0.5936692144770384,
 0.057311000718998195,0.409020876449663,0.30580831924712937,
 0.8949827833256422,0.6044293673744848,0.03697571191139393,0.906518872880053,
 0.7392305988589964,0.4175088007743435,0.09599599334841047,0.5781091970944434,
 0.4248943752785965,0.3420925079432373,0.4236848985505318,0.8542420064737224,
 0.5501291574598368,0.1933882176618099,0.27602572631764877,0.9646972916946833,
 0.6670275384302778,0.5571284340237143,0.9865994024910364,0.563723023039775,
 0.4743978347906247,0.7333758098772689,0.8877281521999134,0.2444231216963314,
 0.7339338414576373,0.6460007947273178,0.05811459025344423,0.3661495943537326,
 0.004499094560242778,0.3419821488921996,0.8027729541480376,
 0.6187620601840877,0.4498570834115514,0.7869319659750273,0.5355370958932008,
 0.05625583186792882,0.9908912824318139,0.6413226057316833,
 0.45163139232057403,0.9818594953051195,0.6693729012346047,
 0.36459955282528556,0.4700018881858651,0.6164820624640609,
 0.17736896722813567,0.6754302339672177,0.7092376259293655,0.492598075526812,
 0.5646342166765468,0.20308369515916613,0.3373457266224835,0.8779505714409948,
 0.42929728218171637,0.492904586556677,0.30269852783857853,0.9866506472042216,
 0.7747099872468632,0.6758794376581198,0.25975096634131645,0.4295725719350487,
 0.23303207359653377,0.5517227673652974,0.7446299701258337,0.4225480848700526,
 0.5889108189468278,0.7201095066760481,0.842022465589972,0.9892194070062741,
 0.12256837508220542,0.49632405137970115,0.23123478960074761,
 0.13700189629666393,0.8935553535378634,0.6002560120612987,0.3157216554129928,
 0.13720836888703913,0.1995180954431479,0.4090412169197797,
 0.40878835669392377,0.2705354112918569,0.04963190328145828,
 0.8326430843716621,0.49455749455073894,0.935714820064151,0.235668025955457,
 0.13851670253974968,0.9461657562860387,0.5104233606378745,0.175929959413328,
 0.7924075776381657,0.5325529919080809,0.9385813548986586,0.17323715325154598,
 0.16272735191044896,0.1615911870883876,0.8415153493044627,0.1607578066997879,
 0.5138539501417035,0.6477917643341432,0.07656325121621299,
 0.10375913713501594,0.48821167148836475,0.044927412878567496,
 0.3820771935363293,0.9403794807390448,0.45098518493599693,0.2814357148128699,
 0.20609223066820104,0.7720664342126256,0.21817900140432955,
 0.40018138154669647,0.4358867750826534,0.8934419550570379,
 0.08406278155492042,0.11335055519210013,0.288278941655589,
 0.31960065975626106,0.6482164652327157,0.01688913037603701,
 0.9345287699602559,0.9776577913060863,0.07511636970389501,
 0.46912113927767307,0.7863474393883257,0.6032769507430557,0.9775454336770004,
 0.10492925497777994,0.06159040745203226,0.38467879227112867,
 0.9765438515080933,0.818821394541329,0.8524483554125827,0.7429683138278715,
 0.29062732207883624,0.10732209875447718,0.7608251335656087,
 0.26795183672269685,0.11846171417799278,0.8789384401187933,
 0.4433528242356929,0.21763360067486726,0.8406133587120169,0.7593665013253017,
 0.4932819144111652,0.9160334501808831,0.9372281205868862,0.9445126610201543,
 0.5564547791147991,0.8733056283618671,0.44378272667962204,0.85490024946825,
 0.24670567942774457,0.04040461625641778,0.5021379185435993,
 0.6674842813429851,0.8141856412202243,0.3173797180060409,0.02318467696958959,
 0.10864482045453294,0.7148046929575537,0.8337767422864304,0.6479980199866683,
 0.4967366297626892,0.2264236555046144,0.7620887391708847,0.8952001109870115,
 0.19087294912923458,0.13396670593494864,0.9385680552466169,
 0.5030631665181403,0.48748038763510226,0.9138741940353216,0.326533539192446,
 0.2773263588084196,0.14173432335656322,0.7660188278702796,
 0.04695957553269636,0.9017104975340945,0.17935455401710132,
 0.3435966193640574,0.3262555200790668,0.30319594420247264,0.5395724579053836,
 0.1648761386893023,0.848063973977977,0.5960695022319403,0.32305117693352403,
 0.8905008477741814,0.8146290178152485,0.6180652514556784,0.8461182873609485,
 0.5966412355770088,0.5128005458774502,0.3544903840197403,0.3533485017475009,
 0.31740599424645444,0.06429072863896024,0.7272670663566058,
 0.10963937744472718,0.9053152620593453,0.5433627948655071,0.9275627490291912,
 0.5618935548715683,0.2193420733638909,0.10963388192057133,0.8821615054995211,
 0.1194295105157046,0.8982202661623141,0.9662532478018041,0.17085563705234752,
 0.8441295724938004,0.7899557742244616,0.416846384491111,0.9969976194471681,
 0.47311766416468826,0.5635596073251459,0.24534803633949642,
 0.8723935508954216,0.6271812519954687,0.22697891228627154,0.3449848849135603,
 0.051488669684943034,0.7023482255828027,0.4614867894868203,
 0.17495152119783375,0.40146466192300523,0.19612360869085677,
 0.3458957343919059,0.0928065702800066,0.9590059983789654,0.8524912320662764,
 0.8419702762783535,0.9042587062488232,0.6420423054231664,
 0.020875203846644874,0.9852208269690702,0.004097663790769701,
 0.1771634275142051,0.10033144409228534,0.6789965451701774,
 0.10849449831072511,0.3479620129846581,0.09384648196514045,
 0.45216953910756374,0.7513010202984444,0.3666198237287599,
 0.41741846352618617,0.18979397257165131,0.6690487685668877,
 0.44277815392922903,0.759754830855566,0.7129726752115948,0.7017792763888143,
 0.8697429820697278,0.5868580056549364,0.6647221096882261,0.21035849200911927,
 0.541307370710813,0.6542291401587916,0.4488021850845558,0.26149900483982824,
 0.9672518599461774,0.7783276450308773,0.7948574201130726,0.19953944859931005,
 0.9795750023576354,0.284787066318285,0.9450726410851014,0.7804931541318422,
 0.5182848607935145,0.2113447577891061,0.34037444105912884,0.6735077248412608,
 0.9986295567614,0.49973937637822285,0.041298612998595496,0.40672860469996486,
 0.9761955973460539,0.20675679718328488,0.8050243809896235,0.6711007483614445,
 0.35851185928184137,0.9383859159606631,0.8061998254507602,0.7918189495961379,
 0.28961568192063103,0.6401530306411586,0.866010188331228,0.4116248665835738,
 0.5496075933486125,0.9897145093710733,0.9823148996522513,0.5713481656786915,
 4.921324477327182e-4,0.009013638795248413,0.6990519267429857,
 0.8005749352418151,0.8431116526592985,0.24192907875174574,
 0.08080958356264034,0.2696553449937482,0.5594240786581137,0.9277032102295789,
 0.09347462301484588,0.19841569011605908,0.39196811877054627,
 0.8453291527793927,0.3747159176785737,0.8714968355697507,0.752673064491116,
 0.04928586662322354,0.2309489998629688,0.18265657367710142,
 0.9991027558552322,0.6135564904694372,0.49294329995192976,0.4037421565958169,
 0.8430289289167173,0.2791799848808719,0.5380172682160158,0.8431549684944041,
 0.2608436954874981,0.7143464559354678,0.6833109620891978,0.30261536692366897,
 0.8321299276014729,0.7519716995251213,0.8500383122015338,0.16670507691883063,
 0.7200367191851548,0.9721393585243018,0.7788923164439351,0.23259230789481544,
 0.762141388615597,0.9360428478431841,0.3958374243846934,0.1761404083269903,
 0.06671680433284422,0.9799872925800648,0.015799259181599856,
 0.3389492764691181,0.8146678138584066,0.04578400035268393,0.1204975044426867,
 0.7981045205990306,0.07718450646371366,0.9960907258772589,0.7182620190215717,
 0.1433187463835477,0.49545755858136076,0.19285101616732447,
 0.013882736446429123,0.8303630495958364,0.5586989856060115,0.95922329846192,
 0.7075075988655264,0.701241140506722,0.28564112661821806,0.4469667188517812,
 0.7826993684389079,0.6662065123939064,0.6085840036812575,0.20356908683896868,
 0.7721791253271273,0.1364918511141362,0.9277096610204202,0.17531672802286458,
 0.9264856944962027,0.30761419506830645,0.5730559447725645,0.8533186564051569,
 0.9404597665268848,0.4448814886627641,0.3331748019768218,0.21851072210300881,
 0.04382145104769242,0.16093829545416183,0.5343359518985764,
 0.3484236206797814,0.035019699130483195,0.44367974891218365,
 0.30602379751184805,0.46181823548312473,0.5342222100785065,
 0.5466492870272719,0.18618665072701168,0.02161039249262009,
 0.5710627717150765,0.42982313220701396,0.2295649252380334,
 0.45008054453250745,0.7240998135635155,0.17305031071444077,
 0.6231356207198725,0.11050458751952508,0.096900060181913,0.5777209149285227,
 0.8488547626604197,0.4261161093168466,0.2360903413701514,0.5704380075556221,
 0.147714975042877,0.9113546747497779,0.0614446733225511,0.9772330057755417,
 0.3531342273565782,0.437791540641596,0.9911499861182719,0.6096713258310736,
 0.9137270403197513,0.9398638902219901,0.14171952246856057,0.8789943626714966,
 0.7217011498039028,0.0678528775113878,0.574173530090338,0.03142494203041313,
 0.5464199322279466,0.3521220307848366,0.7979825501756014,0.3148403507589075,
 0.4920421446630201,0.16470782470991818,0.7635763493903852,0.9622923646508045,
 0.5275412203042487,0.5519320455598868,0.15305808824042277,
 0.22114669923628483,0.40551837393365975,0.06631158356532241,
 0.9572756814977132,0.9686142282653085,0.3781082031606837,0.9018172827403508,
 0.37485501281555766,4.0915181525358335e-4,0.1697772350134119,
 0.6198169591208135,0.1757220080038202,0.5236046974779295,0.8969356193007043,
 0.37323229723759066,0.0762471848209032,0.8646786050555114,0.7136659195680518,
 0.9166384699106074,0.19465376165614323,0.8378279188814773,
 0.29802317835136105,0.28528387279473544,0.7598163335379343,
 0.07195845404395984,0.6997024688244816,0.6696876728508363,
 0.21232853951947028,0.5328743142779834,0.28567151687848213,
 0.4650301110241475,0.40105819145356636,0.024128433940260097,
 0.6579913876010003,0.8681803844069944,0.836993001999225,0.5984490744950581,
 0.03954023875999735,0.28354192906861,0.32160982317288667,0.5120990187978064,
 0.36980195137271854,0.5563588258651975,0.2644085842630255,0.950800645815031,
 0.9496278490945163,0.7906306932896742,0.678413007741871,0.7440922383646581,
 0.7311319976979161,0.6226060993221936,0.17025107711190846,0.5704325117987488,
 0.23276812505587352,0.5720643358506303,0.7166529764111979,
 0.01214673336618799,0.8679393122919071,0.33582431812098223,
 0.9529837357552604,0.7427247735752924,0.9701377645553153,0.1854169366944326,
 0.1742755512460299,0.5127080399178685,0.6312692097795976,0.03544284795985064,
 0.3217591215672272,0.42254250114321434,0.6603039036746223,0.4261105773495202,
 0.7129499642178593,0.5269282805337147,0.20732065430098967,0.1611336589848228,
 0.5702582812047546,0.911294691839562,0.904172474681187,0.7491527282814153,
 0.8551569289188228,0.45191353342334195,0.8433139137343638,
 0.15743855073883717,0.7442203943190682,0.7153935984051755,0.641295459710442,
 0.5493704821086465,0.5989482288446953,0.7595950029408929,0.22044547665985914,
 0.7459882378336284,0.6407146912907123,0.3811122496555066,0.1823308576088788,
 0.10014222248984095,0.36933927654691256,0.0273893734224655,
 0.012726969720964765,0.41970390660205936,0.28989731132279817,
 0.11054352554154567,0.7514429658785017,0.983204150327568,0.05283232435773133,
 0.8308149119491802,0.15003149809037675,0.7992587124788452,0.9406860787120342,
 0.6367169450622914,0.3252815434367179,0.7583377224952675,0.9966227925816181,
 0.236132027952081,0.565711361762886,0.7493279042607359,0.2426743184018627,
 0.27709784691191885,0.31720748994130477,0.9981394877218288,
 0.044751487376036625,0.4927337173729155,0.6658379053884634,
 0.19598302610212215,0.01148475231603907,0.3528627940458141,
 0.8896869605270674,0.5977770239835233,0.5786110578718453,0.23387276121987988,
 0.46057000354741795,0.19584239505717016,0.975245882706837,0.8268973768942839,
 0.19816178784599436,0.21027032138895818,0.212296645065446,0.9993785098441063,
 0.06341310878938966,0.508315200826496,0.941093803004962,0.8600781846971705,
 0.36437200481001,0.548931545285076,0.9786376662579724,0.5623843089598046,
 0.36614301300139884,0.1207308859052374,0.8078292918611207,0.8491116684966505,
 0.6339154572252491,0.7709174879600003,0.2747594491458708,0.2336059841976159,
 0.9499535401925584,0.38915468835491884,0.12245440422901355,
 0.1795805936754783,0.6332181181056807,0.0579094518831722,0.6179310418284322,
 0.9465587329620638,0.37377498379596896,0.36465317049969315,
 0.007654153335971303,0.09833390597676739,0.927508787219611,
 0.8751325578183452,0.4886857309198456,0.12803953145073732,0.9372300774791449,
 0.09369981013169715,0.2674484123246922,0.6903395718947407,0.3291817980981122,
 0.5978746248448208,0.7859446585895067,0.36297889296074626,0.8378498642749138,
 0.9706293236362967,0.7408984523213702,0.9738570265983868,0.4476066645047777,
 0.07122910754217149,0.6087500027799094,0.8189181661398094,0.6057669298189526,
 0.49883100807780867,0.6242366525480278,0.3699076962931761,0.669588090623589,
 0.1425440976822594,0.21286665482135936,0.9985776543242374,
 0.07688982141604139,0.8876658973752296,0.06706665007326984,0.833063046346297,
 0.5448787368994917,0.771328007807998,0.21697534883835856,0.0986614742986246,
 0.27778059579485703,0.43553435277320796,0.03110633993666001,
 0.22252666853192238,0.1250793005123323,0.7138532528748048,0.981676051263639,
 0.7680361673023743,0.20291870632553022,0.9767878570487047,0.5683110562027099,
 0.31192663847039526,0.8781078856899307,0.014791205793220241,
 0.7874316911621614,0.6599851876879296,0.030238648834688076,
 0.6535013315420798,0.054521275656514434,0.6360432866050834,
 0.8784352187674547,0.8398096196660721,0.8906412996412256,0.5259943883398657,
 0.3839281569343066,0.5433812582089235,0.17301329176195013,
 0.10080190513283283,0.8989523994934001,0.6803795963337315,0.6741926111139211,
 0.2853152841734681,0.5995468977638819,0.9185651667230794,0.4357114299668017,
 0.8386305747727683,0.7758524621127695,0.11497687251908595,
 0.26741353642555077,0.17294455295467492,0.890238340738873,
 0.45548549704173313,0.4650743646067542,0.8771328206011187,0.9345964837083914,
 0.6040870215572042,0.22427471425340184,0.19016020689841873,0.741289598189065,
 0.3970501007144812,0.16361202219102733,0.19127191223457696,
 0.6356057721502235,0.5174515434244781,0.7374441418914405,0.16986749338682794,
 0.5004045135083589,0.025272638549057147,0.37481744830945996,
 0.5315681462052511,0.6978949596900108,0.2332860577112622,0.8234734854069355,
 0.8496536966461798,0.9350120834707274,0.6025957261933048,0.01993599737051044,
 0.925924358682612,0.3126913332118647,0.31798536609275296,0.4051897157063802,
 0.9897994832958794,0.024129889739664234,0.5050674589352773,0.834417220877494,
 0.36826414138318897,0.3667153333668708,0.8032659494533266,0.2630024329022602,
 0.7415767277977776,0.19381621808034133,0.7531425458330006,0.2842099925963949,
 0.658291548849069,0.3885193486820986,0.6888126037657195,0.7087334606275836,
 0.3255294625633375,0.6684231273092353,0.8949676872719741,0.06559555217697233,
 0.05626491903874109,0.1502162560921203,0.27851110300667403,
 0.2524355315924317,0.48141959601848394,0.6112152055802262,0.9550989934957891,
 0.9507123313461798,0.9555531166736118,0.5517358166076106,0.36559501169106456,
 0.7338229110145675,0.10235274560327412,0.204785821849092,0.7014672234880224,
 0.3957715540966875,0.7216097886209433,0.21416767890377386,0.6938511281126497,
 0.23547440238873552,0.16052053454860526,0.016740685794411347,
 0.7527643818668357,0.1848760616340257,0.6823916127928666,0.6188862707425874,
 0.2949488750503777,0.5812769361208273,0.19927908322012056,0.9527635261107308,
 0.3639145668276138,0.43528236776810325,0.10974821499327758,
 0.31071484393453763,0.2532426446412622,0.8022685609234537,0.3587546953659595,
 0.04764875789654599,0.21734523994807375,0.9127845690193114,
 0.4908259486002077,0.35809302860721637,0.6487176555418328,0.9517730561020553,
 0.08155961929831523,0.9933953678738239,0.5107989257466076,0.4568257037163851,
 0.36780469943632044,0.833854756812284,0.880538754618903,0.056730653348231797,
 0.09850096836652245,0.12286508355257164,0.6760153940395529,
 0.9727725478397492,0.9185893529800403,0.4813445319139621,0.15180104316556653,
 0.9616124077934616,0.15931911044556157,0.2926198678915908,
 0.13911053674020146,0.7011074492299463,0.7723273299034982,0.2743621962417301,
 0.31651765881591076,0.5877481015852951,0.2537352872944495,0.990488119937895,
 0.8700044792779489,0.5277334196621595,0.35589916688981815,
 0.36721776518804916,0.6097219531973154,0.3249608979739417,0.639112500687135,
 0.4007606131803576,0.44185986617714124,0.09050016193063204,
 0.8829352354007789,0.1710856637000301,0.6609565044867914,0.9163831292641732,
 0.9718813079677949,0.18167887551091733,0.6202460371385305,
 0.27393684016647823,0.8836983882564126,0.6872460386819952,0.2722065117418786,
 0.763913867804031,0.7344755688269271,0.4455563768924624,0.5247834346196956,
 0.899798517347914,0.8592246636583302,0.9268046615573133,0.08772639358126932,
 0.7147821528659702,0.010688935032268088,0.6228584746616375,
 0.4852094559243896,0.8631731090041745,0.8854234467393254,0.5584060820203417,
 0.0883710911302531,0.7328508409590275,0.4949508018559635,0.6892431400812852,
 0.7893815249628611,0.5065277145073843,0.6665206716234335,0.4287078469166794,
 0.5004345491361961,0.5516185666731448,0.316188470954617,0.101571577394591,
 0.7849229170957268,0.481758270907771,0.7756401426787314,0.7667977597951295,
 0.5125622681076787,0.969201793132974,0.24205865315928765,0.44996688801858875,
 0.487530115544352,0.3382220517536272,0.37836579117473923,0.23559474690929583,
 0.3532748197054081,0.8443494387002524,0.6442243857808334,0.8902805170178156,
 0.3361646731398862,0.2394907432279153,0.5500531812696998,0.15271356555106463,
 0.4551321742476446,0.6835599453609997,0.9107394254825317,0.8684692641760731,
 0.579946483205286,0.7242282410060737,0.596152494191995,0.08595004131984318,
 0.5665693796988593,0.00750253141130619,0.951003595490778,0.5373717119973231,
 0.4428362413955955,0.08962491033395281,0.7842303868004405,0.630660445466357,
 0.6492213452019987,0.6758355353079956,0.07938033513277754,0.6787032189208748,
 0.09273740168254674,0.2215396491319278,0.7899974352361944,0.3050400118331035,
 0.603141366563501,0.06602594657288163,0.6714931010144405,0.9112992920535284,
 0.4984884199183154,0.24846222869501866,0.8306790169015567,
 0.09233258728352478,0.6082769387779581,0.3561081865735386,0.7954907738478151,
 0.013603824971744727,0.9468514869508933,0.7825642464774076,0.93393336750632,
 0.5629292900275124,0.8411180844098989,0.7709868819816162,0.5044695174079917,
 0.7669811827262714,0.743170015002842,0.8380149189824538,0.9200495922740962,
 0.014849689916356823,0.267394373493799,0.2626704025041615,
 0.45877719900790326,0.035451676432956325,0.48225271004358516,
 0.7609829721444042,0.2561952262559376,0.3135202629149244,0.40745736407803534,
 0.9130735104028286,0.12607847489348378,0.4533008347958004,0.3131534193857878,
 0.5528565278611663,0.5704245153638257,0.20686665588748954,0.5975163802165943,
 0.9675830433105352,0.3669051196888562,0.4001104704403099,0.16789860155489278,
 0.03246488291343763,0.9864296755227049,0.04536845873923967,
 0.4707962650006172,0.8394320519976045,0.35175948369283006,0.3363279790622402,
 0.7422835910287353,0.19690107747612862,0.6397727203344679,0.797034133236634,
 0.5862535100387787,0.2552880966411346,0.08396096605863379,0.8711630988288664,
 0.40434070493656016,0.8612141258795493,0.09388775563741902,
 0.9859508445464275,0.32184342134592403,0.6242089170873827,0.6495510437482273,
 0.5037982794035453,0.06095520203566418,0.7418884095333002,0.8000381558917943,
 0.11798349043209433,0.28391180725577514,0.6576705176162552,
 0.6195165191030386,0.43760985615011916,0.20204712700469185,
 0.7629671712863294,0.6462537996874387,0.019522730652033582,
 0.33416594186387727,0.5464142268033045,0.7684356909374439,0.7594504998964445,
 0.17659303742163535,0.27158665806307647,0.8349065492477624,
 0.6444522716513995,0.2627487193916207,0.4949576885808302,0.7310236202587267,
 0.11877360826702121,0.5337693816214345,0.6224912057414912,
 0.46516538718952793,0.30473185456595975,0.9742292402084527,
 0.22082652673839942,0.13287853811583872,0.26430395533048334,
 0.49264206383298514,0.23626760197566465,0.3902842602850911,0.657635697095164,
 0.6138896628938938,0.5538968102891442,0.959161245189106,0.03371367285480065,
 0.6714889603579451,0.21427324162366124,0.4870574443197845,0.5760324888557373,
 0.0055641791614107206,0.30951460267781494,0.5336255012263905,
 0.8289113588033098,0.3314507762683263,0.04870485806570146,0.2519846096805398,
 0.41722161222044196,0.5530062320809055,0.9862296916701432,0.3482319180844591,
 0.49686145590946396,0.2974179320300748,0.7811207748355767,
 0.07672272651040091,0.327809301684157,0.8702659966152416,
 0.051840051645137386,0.4804393924663809,0.2611291518565819,
 0.25757465573803207,0.9221533925957334,0.7552305041783239,
 0.13626998184725925,0.5525002934506525,0.14453491637728966,
 0.6555229403065187,0.0603736467833349,0.4417782997373886,0.8051603465498307,
 0.2368916942074426,0.35837522575896996,0.35509144339371845,
 0.09818237166449284,0.6567311298009844,0.0627002057053152,
 0.07739039491054067,0.5742347729421595,0.5875956848648833,
 0.44254614847651463,0.6649312657698943,0.3553925867321448,0.9639731359944597,
 0.3111115683796495,0.42865496908740575,0.10987963128463063,0.635471134810002,
 0.6607596168329996,0.08415806243334589,0.4472048033964736,0.5384611577305357,
 0.9541286275822612,0.4362779917080547,0.2929972241819264,0.8416409237529302,
 0.5167149339142821,0.9093107162997001,0.2647750590811856,0.6984186931651484,
 0.026519140469064006,0.8462433692580644,0.6891579320240634,
 0.9999146974281511,0.009504300376283692,0.7306992092289307,
 0.4115217118486967,0.07691984013230924,0.33331815234066364,
 0.7875135230579686,0.44991353772636977,0.2853048381794012,0.2174474023311438,
 0.566691842273961,0.7426080403922234,0.09633065102599347,0.7850379044481545,
 0.9289717746817802,0.5355557265371998,0.17758875461508794,0.625960592939675,
 0.4900293570363595,0.25700053502679887,0.3003256445573169,0.8353749144341034,
 0.3357833057082835,0.008919820781899457,0.24188395070450985,
 0.6543636812673951,0.8948059721903061,0.8463043049448399,0.7930397100819468,
 0.8880304363049888,0.3505870496596406,0.7750695988034028,0.9002385835356,
 0.7388749375930995,0.22194583664109935,0.2690534149374689,0.9580066207510032,
 0.9330402808218214,0.8109995003881313,0.7202759347005347,0.15281445079298395,
 0.21884998160270652,0.9432783088742045,0.0707867446926631,0.7379934127046027,
 0.6406750573798877,0.28183826979640614,0.2495855132688064,
 0.32599879174262336,0.4513251106355426,0.7679304713477927,0.5035432395413849,
 0.7911871328454654,0.9447296952015533,0.1265852617234221,0.12292308222194881,
 0.6182562892312452,0.8742106320228316,0.8540554333913392,0.7460512941729311,
 0.3576014966127573,0.8637540990391592,0.9265070821731047,0.8820454396288371,
 0.2838126600480654,0.0972843326001015,0.019606346793456297,
 0.10654049974463287,0.6365331191979573,0.6520748090424429,0.9677583254514297,
 0.2798800237776933,0.18259866247864598,0.6474671662634075,
 0.48565068720406923,0.3516180160420569,0.4615594572261996,0.7682763594033901,
 0.5911833038236162,0.4188398408443007,0.4267774128552507,0.5646948413228468,
 0.5861610934040087,0.7744103772592072,0.4188278009373213,0.16651253897999907,
 0.7851788446966158,0.8354824031451089,0.5793712123729405,0.9783662821216592,
 0.3359861340159488,0.7584769043244491,0.26576427911500944,0.6793748839639495,
 0.3251428417526987,0.721555874315941,0.6388434183625508,0.38232646792570313,
 0.37890787477813026,0.0852503760319876,0.6949302071902681,0.5311407395609586,
 0.200347786587593,0.6664808836895668,0.0011711657018574906,
 0.25289391302381503,0.10307159816513223,0.06725746041290104,
 0.20897808517707694,0.20612012181283867,0.7152068499322671,
 0.8263736687679559,0.5032024730363569,0.479023144044761,0.49061143965668086,
 0.019592427099563547,0.14522767605130804,0.2284550762112545,
 0.9810473718590542,0.039308441930641425,0.056679147217113646,
 0.6766640575289792,0.7333316562172533,0.2526447113966688,0.9833989709096311,
 0.7771077687924233,0.7104655604594015,0.5324412614132497,
 0.002334347925812108,0.6541264793395589,0.04746253327034755,
 0.4805794279628962,0.8768064130850178,0.5443426413956107,0.33901268112454386,
 0.504503165400601,0.8252126953252521,0.21573314719184689,0.3384188531670056,
 0.29076929376350935,0.5567449940101353,0.7762576211240505,0.6350308673298669,
 0.9081992302342927,0.8325465476385213,0.28866781658897933,0.738571402294077,
 0.5670326418036367,0.17304512307902742,0.4801785435016045,0.6051212611305015,
 0.30300035569350026,0.8155038722885206,0.33435800488744805,
 0.6637427681183576,0.9607704497295996,0.30594157894519036,
 0.33192609221749836,0.49477648403043184,0.608032600158139,0.7197341653281348,
 0.37861861097949556,0.00126449669286266,0.8496976324624632,
 0.6596394522828563,0.26941967607271766,0.20164747985350973,
 0.4299150609692477,0.8237049015839101,0.4856850108986075,0.5215378954700898,
 0.22380492806364938,0.9626729939879682,0.4999375615748427,0.0762449128718482,
 0.2005635684702649,0.14304746852767758,0.22585469568715633,
 0.16874517161371994,0.4839568960378464,0.77578268878299,0.4014185624051123,
 0.2751571613894832,0.6287019830256629,0.7074450074770993,0.358432074068311,
 0.5917379975391602,0.04435115668932876,0.37229774192643117,
 0.5021220443188736,8.624890150783493e-4,0.7081769392466837,
 0.02412762550248826,0.05097093236789785,0.7118802053565503,
 0.7927523665388788,0.67805330114566,0.6951447662214295,0.619068149256484,
 0.5178460580598017,0.11252619112725104,0.4822271096058235,0.9910037402970379,
 0.7171126792080668,0.9913325452620377,0.3719629536189816,0.24200957447543858,
 0.002949634474280205,0.5019009006180024,0.14898793217311157,
 0.4369120298455236,0.08921484826391979,0.6160542401247875,
 0.47976019765715594,0.7366505624568155,0.68615844062657,0.4945566219524884,
 0.8335623451031928,0.7408790245705524,0.7059892288837254,0.606523770441048,
 0.9763601640035521,0.47126562897323154,0.882378848179368,0.3099033554746221,
 0.43283731690571314,0.22460860891097034,0.14428905293155292,
 0.22330660747412368,0.4826840020051846,0.20894991985115963,
 0.28327769517652923,0.9041249502225827,0.8719704117687893,0.8552501269847013,
 0.2605843059064086,0.5567002253278935,0.7115642261428358,0.25670484812941974,
 0.2929515710318723,0.0059180202727120435,0.6538136346888925,
 0.5471566748895512,0.13000772967083996,0.4716136728984751,0.5004835800441759,
 0.4922174850912715,0.32322716079316016,0.7475443635878154,0.2700679932444062,
 0.9736651717550722,0.7323302696794097,0.7404710905273464,0.2749793872300156,
 0.11795936537243246,0.37605231567851094,0.7594760240393456,
 0.8952822559335131,0.9965678542930015,0.28938795109084403,
 0.08739544851330938,0.40704415107065883,0.011199284255076436,
 0.7848671631089619,0.11686382424110586,0.8840974688858327,
 0.054460971752101184,0.3810839801116376,0.4912184694325733,
 0.04352330562049023,0.10774863157549375,0.7448986151745852,
 0.5238515406844533,0.5293753438641078,0.29302946088242443,
 0.20777291137427334,0.46277679089081014,0.10252231292189729,
 0.26316440004863295,0.31502575845883407,0.9996172656736642,
 0.7900829178419946,0.24257668490376838,0.6812132636471927,
 0.31486587467212357,0.11289509397445241,0.6387147691744081,
 0.5565947576624414,0.8475963295924777,0.07609546943389744,0.1758440543640808,
 0.10436562175376474,0.9347469590678271,0.5322592570589735,0.9199140088818865,
 0.9569369483854692,0.9117801491665811,0.763916469889871,0.5294553169983384,
 0.5854695697579635,0.34041533718876815,0.36388409587330384,
 0.38303550214158477,0.07455875209593088,0.3722151780016766,
 0.5126915658191823,0.035337827433272295,0.3006742131744057,
 0.5327288730491713,0.48501285319561993,0.5338454024816102,
 0.19468856593430173,0.863831834926449,0.9632381720139284,0.4406484591381342,
 0.6194200859092125,0.760909489160748,0.9050575278472436,0.08721603460655081,
 0.42785986916921537,0.7528469844150263,0.27083584183895815,
 0.8120585104892758,0.9432320504236239,0.3427771911572697,0.5366475649883001,
 0.9036920412082642,0.4923364462943537,0.13785156960155331,
 0.26242308281911686,0.368519647868529,0.12046730579792209,0.8247000538229583,
 0.6941901647127686,0.6671667095016278,0.8982855079734074,0.4950169675017886,
 0.8473886351882368,0.1915644874381749,0.019270409224835472,
 0.9669736415089041,0.8635685619918285,0.1160729939735472,0.17741795204254418,
 0.7830309324025706,0.025081262197793386,0.10663317145212492,
 0.15593599691411297,0.5725857674181726,0.9091357748799355,0.9030466790190276,
 0.5877081379018269,0.7301781166423056,0.23298675544429825,0.5378491839262803,
 0.5738112545588323,0.06245293728553525,0.17772908259298548,
 0.050866139114140396,0.5435285084012175,0.10460039341781668,
 0.4836368915397369,0.3409003257503943,0.18672083889554403,0.6919119303296997,
 0.29595946961652575,0.0930031245675863,0.15849971390223816,
 0.4291490395648718,0.9485600316371496,0.7714032543541542,0.5618603379747501,
 0.06040317839081122,0.06164362517831756,0.07238686720335563,
 0.061838988177515875,0.93614357590188,0.9319950428972688,0.8400831706151966,
 0.02638836525791821,0.12602339871900303,0.40097124428603215,
 0.883008837571065,0.054620816226785496,0.016051416435794996,
 0.01719168253277803,0.9145027913068943,0.38664081443986575,
 0.4175972911867133,0.7932866134344783,0.9139609570514171,0.16831044223954694,
 0.5808774070077501,0.6568070516771852,0.8519916516585397,0.5488057880372345,
 0.09656930878189773,0.4495375117181284,0.6804433503893504,0.3702899137803545,
 0.17824444994407673,0.9332717948282268,0.9068576283472393,0.8732095883862526,
 0.2891755663790623,0.37247934283452033,0.56814225188448,0.4883800227662789,
 0.3264159694270201,0.8280553542343693,0.6279984819843936,0.25964612812707966,
 0.2035497552425044,0.25476539163568557,0.8048795851051053,0.697298874775419,
 0.8779904824887962,0.481200338675285,0.1425348370807542,0.7854321829733143,
 0.7724383544521362,0.22209829271424952,0.2529995306287868,0.5812290050482075,
 0.22971130681224983,0.4346117581686054,0.2126665304245644,0.5720487941550162,
 0.14569755183644162,0.9560647001172039,0.4296305616320073,0.6171382205729699,
 0.8830318360773739,0.0030121390557812378,0.9353171193301828,
 0.4873121567455896,0.14576603791040343,0.6384852137977832,0.1860175606385004,
 0.37412075416627744,0.3605563565580836,0.3930162959316167,
 0.35232511204621986,0.7347129632723264,0.175756033825322,0.29626952115214455,
 0.5738699174832288,0.15176493024566629,0.1333756014865638,0.4680408849535982,
 0.4523572424328569,0.24010219199617033,0.03930104236272769,
 0.24151959452833272,0.4849412469426774,0.8980857832768061,
 0.07874285991970642,0.6693062041113205,0.08916522806553695,
 0.09939727533840426,0.7879795654360086,0.699131794197154,0.06106532433966949,
 0.8429137490517188,0.5847808462247249,0.21418554293167114,0.5281262088120651,
 0.5772008413628593,0.9204422788191231,0.33306202230209514,0.5465256588192878,
 0.20899253083579583,0.07870140404339199,0.9428066447914779,
 0.7451116631098724,0.045798073359410774,0.48761528656659847,
 0.8666242166868462,0.6746523454055001,0.938632759484947,0.900197127248763,
 0.637101618379994,0.012161527365268432,0.3820812365138434,0.7986360205471725,
 0.2715588603978075,0.6806320927963112,0.8354156531670933,0.8914006507252845,
 0.7870102925167954,0.2614795310735789,0.2269062399695213,0.5206860215539395,
 0.16101001915370983,0.2358817691764047,0.8030975800457142,
 0.33392861321578726,0.17235227941135522,0.37460923730568796,
 0.8290527885596831,0.17165166248057984,0.940582397078566,0.5916163907747434,
 0.0231951899056293,0.7525642028410379,0.6609441016381745,0.8914992431468495,
 0.9396783643550058,0.8785547540720693,0.406968955503752,0.49602021462637,
 0.5661280549661656,0.6496366825790529,0.7907188570247533,0.8033857174235302,
 0.9640044226539404,0.359430905624599,0.4597892821253916,0.039721204633008655,
 0.7231804747735745,0.5515719660350378,0.9850455983598991,0.2508705948179957,
 0.41161152237790877,0.3707521379004389,0.3956767303161409,0.854386124522887,
 0.9487351218086579,0.7839702779305185,0.7796468899088767,0.20947687171706753,
 0.6980095315103173,0.32413384521150934,0.5482856172930297,0.5520511726764809,
 0.27191015232134763,0.31188060741083823,0.5271752183625488,
 0.7116708635416162,0.45745542379594584,0.27846506302972396,
 0.3122154444491687,0.23391707849765475,0.24207914742514913,
 0.4371476256702759,0.5741003051853659,0.43516774905094235,0.3355251428747976,
 0.7695554346244416,0.3268206799183919,0.6668003719011384,0.18594552265528186,
 0.19561168363328685,0.6699907154791536,0.6508189728887073,
 0.18644946005018126,0.5111238231089511,0.6364840864407103,0.8384192799142627,
 0.3421081198443089,0.7949053881939083,0.8085352360431258,0.26993334855535367,
 0.029870647603784883,0.5627224682628309,0.761365657982937,0.8059072271255939,
 0.9880043719042372,0.680962701056032,0.43913787455562436,0.8498009332590838,
 0.9988076380812436,0.19740103357360417,0.9728092283276097,
 0.34581295557753955,0.3359898605401015,0.25564056462615536,
 0.23583618136956253,0.07702651916591907,0.2475321237676661,
 0.24085084237491627,0.5773108152760372,0.3222248668450858,0.4765242999089848,
 0.6061281828327907,0.07453619843110426,0.28696437265102914,
 0.48989978796978284,0.7012855925713081,0.12715685893822382,
 0.46141450525395067,0.3058703857021443,0.4480487141573245,0.8836957938676487,
 0.38435019400657044,0.01035908932553875,0.3084520715665495,
 0.4332419267436163,0.26012486570716953,0.5138073904450426,0.8170334142269444,
 0.7722903876673728,0.6595142694074875,0.1749429976259793,0.18749237638716676,
 0.14281635762751077,0.13788182941529026,0.20044902184567026,
 0.15431023572933467,0.27639631771729545,0.5725789083789042,
 0.9489230380205382,0.369003040072754,0.6148432076510377,0.1808153662235098,
 0.9604562234932862,0.6215214534186911,0.6738139533493068,0.3066292904073479,
 0.3075269617172808,0.890219929921805,0.386372300567116,0.5647336592202244,
 0.11574884071670766,0.9077236613504237,0.2728192277253413,
 0.13275873649811143,0.263386176118236,0.7029834676543012,0.8810989303853446,
 0.24525841545839688,0.7884840873175927,0.16913085049399434,
 0.4126697287826032,0.14820505889149693,0.9445257704972686,0.6979561386209894,
 0.9299531272420513,0.7452276506135573,0.7952530931326687,0.3137694042212247,
 0.8042161805513501,0.6744327480224077,0.905480240671683,0.7322318677589457,
 0.5517633068387957,0.5189077599751383,0.11188099086865977,0.0859729101121105,
 0.9587592982130644,0.8176010857273742,0.7902181574867232,0.7078348050912382,
 0.6235225795574674,0.978009629043349,0.4976122954112836,0.11840953099977403,
 0.05871906026274032,0.5590635969000751,0.37855389947308216,
 0.016865661252286723,0.6551456181381208,0.9777123909070365,
 0.03111800045558044,0.48381767777113815,0.31341655707281024,
 0.10163163961012667,0.5085333599007816,0.32217239674288567,
 0.8962948201373695,0.5700508006142634,0.13751617188931786,
 0.48515039707949037,0.17467494637817338,0.6141925671844638,0.642547833219813,
 0.6318474143335847,0.34082039422309507,0.9460781483353221,0.9561172367772557,
 0.18260537476363714,0.025688474524274474,0.045278669986761555,
 0.5458974934709167,0.5262440220182815,0.4750772533852021,0.1969431401638152,
 0.024498912097131065,0.12863289943119982,0.9652645589964274,
 0.47125076292384904,0.8285369720213923,0.18372972329649484,
 0.6259389095014385,0.5216849533605962,0.10408721696305123,0.576045416257564,
 0.5531595348474725,0.8094782257701743,0.07511825165577646,0.8307591696192695,
 0.3480782975109372,0.6130407735990194,0.08267770722116685,0.4723208613682892,
 0.8453538693005784,0.38426403523828256,0.8717295212532862,
 0.37773827954058015,0.33064614274713977,0.4833013684277546,
 0.34311732113135374,0.35623810731314287,0.8508647162358887,0.195902310039071,
 0.5956397833183766,0.9188488341590644,0.25556286399901085,0.7430694500838826,
 0.8189655581146992,0.5571691952554942,0.4222702393967378,0.1042461597617621,
 0.7630114928134646,0.17087240095645695,0.48515220009116367,
 0.39515452833272413,0.0127253412583751,0.8178895953586004,0.7101248138775125,
 0.3604457267905512,0.1545351258584461,0.4975620457687008,
 0.026332632264893352,0.38342159420968325,0.5018513048242259,
 0.7411229862722721,0.8672455103190915,0.4860175867395692,0.8184616676997081,
 0.04329339802000387,0.20851183592239475,0.7864646561746214,
 0.36766256826165833,0.2125086620473905,0.37453278631725306,
 0.3512131265602152,0.7373921489175923,0.05715965381751398,
 0.13477747789932248,0.9394024344178953,0.4253291160250178,
 0.24237548910429652,0.26638969012282443,0.1942674490062234,
 0.9280880741679856,0.8835093847846283,0.07783805331847282,0.4855066803308665,
 0.14620213997649656,0.3302121117796101,0.185289423425388,0.3990303451507453,
 0.4556158981892958,0.7800019466223425,0.056030738365356214,
 0.3557647192047637,0.6438668882628257,0.9748466352515324,0.419246473951702,
 0.70084448986353,0.27829156001731326,0.6017268786445078,0.4345511630495944,
 0.8537612011871418,0.026764726916622128,0.47561310686867253,
 0.8938767642730411,0.7314868415191805,0.6382470103493255,0.9752475857203741,
 0.5516024053456602,0.5377395078089313,0.6518456379456795,0.725282505525769,
 0.1347279705568042,0.10523026828550774,0.12802105868244162,
 0.013894294253580242,0.9846397487341146,0.7929405871174191,
 0.6982684213855349,0.6046862422030969,0.6186058604767536,0.6999798676054229,
 0.3242270927568217,0.045037305964631735,0.47109679191662523,
 0.013423320546671724,0.7693399505254936,0.3518630164593226,
 0.36184840952696906,0.09847110934081627,0.9326852533343593,0.580788508523384,
 0.5225998094340026,0.06961803571293101,0.004037205754020001,
 0.3725342102372674,0.7087756230018701,9.215923122209979e-4,
 0.8850532196270282,0.9388822120812008,0.8624587774687041,
 0.004988491782076165,0.16612875458734733,0.5174775028491794,
 0.5947596279023574,0.22658328857869425,0.999421621094412,0.284050082774123,
 0.7984790964313756,0.3734722074307607,0.3362168848232985,0.10910688077244324,
 0.4578471585817021,0.30013692456355756,0.20921878819133322,
 0.21960227477572514,0.5606173935217373,0.644172781963869,0.43414226525180544,
 0.3365392812724839,0.6144529018255919,0.02244249061632997,
 0.028866806687428137,0.47295843305524,0.11467587316692773,0.1493177129601717,
 0.6974643503557911,0.618022687609892,0.2551266548691957,0.8516099981797722,
 0.3236722070803357,0.697875714120661,0.7610748965628759,0.8880533359845084,
 0.32578356812601117,0.9072856846272898,0.5189627797549787,
 0.38162485090526244,0.48525730885211626,0.21685208941186596,
 0.9429179202803022,0.38570437512565714,0.4355606072708098,0.466966079583228,
 0.845644689429467,0.22111042650712986,0.9987211824316884,0.3825340357404817,
 0.04186474463261591,0.005887353712282772,0.2779118719788295,
 0.564217236009186,0.6884339718400887,0.7614317358402477,0.9915418907899706,
 0.7833844035120952,0.5782980436959198,0.07462105994303747,0.492791897288518,
 0.3412053359370246,0.33836924552295633,0.017001480196561225,
 0.6611848757675632,0.40250841079866995,0.16247254939499456,
 0.9149141470757359,0.06496229148541977,0.33364629846414073,
 0.6788286246609105,0.012416936980156668,0.7963865797767532,
 0.43561171465491166,0.6212317281499368,0.8562822074823058,0.8956936275255576,
 0.5618436018631042,0.6249326918851135,0.010381236778610248,
 0.39233914283126003,0.996414191238637,0.4519060978254158,0.22629021744304145,
 0.18288661851935595,0.48252497243345593,0.040164378564816916,
 0.6348374789074009,0.11502149573607712,0.17332987433052505,
 0.19501526798093383,0.47818631544371115,0.9939381401264255,
 0.5805963453890473,0.9007242168783091,0.7797492487472824,
 0.026891030508749858,0.6821012915084095,0.6611473756983633,
 0.6780542902634651,0.6396207884318108,0.4251628921706523,0.7040898555090958,
 0.36517393538689136,0.17764270816054706,0.5303955622381793,0.599381506888722,
 0.08376970374707182,0.026411987723402586,0.6967712125674648,
 0.17780755036650953,0.776386008779619,0.33374227144491353,0.9913668069351175,
 0.7338568492605537,0.005746370000048406,0.30887309087544834,
 0.7376627424318951,0.6107041482885841,0.802157460903089,0.27686332214698095,
 0.4084502050436736,0.6655857273209895,0.014379837168748466,
 0.9055947372134633,0.6738687667034697,0.5805737124429802,
 0.014232112530774721,0.08067362207766982,0.624856763278826,
 0.08647402939034621,0.6618885312856644,0.12426060412556661,
 0.14309437586069085,0.7250851489614681,0.4632223425054569,
 0.30457910836807656,0.841329815053555,0.27825331258404584,
 0.42792635821209657,0.07873057063262279,0.797854992350165,0.9351461908682578,
 0.802917880155146,0.5863644491598631,0.8132629731523433,0.6431555151881276,
 0.9146917085511489,0.7308015851335815,0.7304709694576226,
 0.005642143877298111,0.3063250461082815,0.7805771495372779,
 0.8593331714775682,0.8675081727976618,0.8419574099357401,0.6147489475291374,
 0.13728169365455078,0.19250106649744192,0.0018269662489555083,
 0.0037591446003377715,0.12034131321193176,0.696912138598278,
 0.7329072766516614,0.7357822412032722,0.5256760161463414,0.678996905152961,
 0.596871352096364,0.7017041785190432,0.12405387622882358,0.49288538215028654,
 0.7790880979240358,0.4254539467825582,0.5352830596436964,0.12712572175968984,
 0.017480609326997576,0.6224155408795293,0.2512909295364488,
 0.7286233339321502,0.4667739331209707,0.3804374521683004,0.500603910622022,
 0.36840047716905266,0.9808928994139209,0.6433381897838657,0.7987134393937851,
 0.3314468599989131,0.7141577767119474,0.8571611007999447,0.12262102301533306,
 0.29168691212219433,0.7334206872646247,0.5920968101960534,0.8356960484303394,
 0.38950858813692923,0.6230078043508808,0.6674913648395315,
 0.45257367085727507,0.4826052415679153,0.3022184429903446,
 0.003336270531727248,0.019150359114905258,0.389229170161437,
 0.5387976263236113,0.6812438825133408,0.5759196076117644,0.7388419293631983,
 0.9963535343572935,0.7433462621897317,0.4736983006601807,0.32570923130686924,
 0.9020512061455727,0.8219052559833859,0.29723983404149423,
 0.055023777018413766,0.8125058480688175,0.4129416661870514,
 0.6622135183212009,0.9492432450006708,0.911027860204417,0.08845550382166079,
 0.32203424863445596,0.2977348139498991,0.7379500803430832,0.8901704272128316,
 0.9082289879332746,0.4100360237879864,0.7003664956122011,0.21683317290408421,
 0.5350055879193862,0.05766145760994934,0.907903323374771,0.147657142530916,
 0.16625970752252428,0.5762623617647957,0.14080592691354044,0.545818152283564,
 0.3655139050847166,0.38837673965577635,0.9638821448365057,0.2486787672844284,
 0.9579562985206773,0.9355379873726064,0.7166876861911582,0.9487042911975498,
 0.19104514550733342,0.9167000297574578,0.27022014189383947,
 0.5198944301607354,0.3487478174349783,0.37906760366601966,0.4810015271605156,
 0.16927031902144152,0.18060137246892638,0.16552731224982065,
 0.01295186276299165,0.9231342628946344,0.5188005989796268,0.8619090465913868,
 0.5500931710285045,0.29169665062197137,0.09634931020428739,
 0.9940768476610211,0.35217680274570584,0.029298595426566543,
 0.5697986205815007,0.1619896239619103,0.531765208689656,0.3872901448075148,
 0.7232501305397234,0.4844576746230266,0.2012474962541585,0.870928472490903,
 0.5078374163688196,0.6193271018574561,0.7728384410778574,0.3816136479628368,
 0.22799158125518115,0.14286528320864422,0.6300647862322377,
 0.6437593658090032,0.4500757170367331,0.9054111277781212,0.2665704607176659,
 0.4837149669435632,0.4153437467510215,0.2061972038696871,0.6118811809174207,
 0.08210987436334849,0.33332849902429496,0.044229631468169606,
 0.39718729849031886,0.21600089346080165,0.9610645079064071,
 0.4057657012285256,0.11677916319280146,0.2512279082957447,
 0.08225000205658917,0.8598638606299588,0.8540736198286103,0.7520565718022056,
 0.03238023739252138,0.6549519754831676,0.5235562934580174,0.5318636954384497,
 0.26375683384570536,0.7724901083300554,0.09769441677222734,
 0.6630930483464812,0.407326169265236,0.013908797855696209,0.5299177620640397,
 0.8464358165143963,0.3838398116871712,0.13865307446993136,
 0.37590196896429706,0.7686669832483646,0.8665467933868899,
 0.005944195233164251,0.619182936355446,0.2567520426998863,0.6717012124996213,
 0.1851999918627586,0.9056473927457049,0.8805249242699049,0.15645358547567978,
 0.031086591814235343,0.3759264576853002,0.5212420152679091,
 0.24610403530459068,0.24163774145389816,0.9346795739634487,
 0.4847173750737791,0.2823255480893778,0.2058478455932734,0.9174739024959674,
 0.7152356395618804,0.9051397714735786,0.2223635753903056,0.8958813570188322,
 0.14540733181193755,0.14728503203017684,0.5944970413490529,0.983057042197897,
 0.6976159117813976,0.1257571493916121,0.7513110903729278,0.6721121804274429,
 0.06270976551078822,0.7403774637943366,0.7371625912714334,0.5510805045555518,
 0.17854563455639982,0.6528013303312934,0.6289009836827043,0.2437863409567993,
 0.06334944914996521,0.22015984892383533,0.8767371233490355,
 0.2647521920201852,0.24066244167390094,0.01108319748248976,
 0.9524897853375767,0.13353486890850508,0.853013555855478,0.5520744838917295,
 0.9909238868113395,0.9168851730951522,0.15485220506482145,0.8172065299978862,
 0.25881947066305355,0.8375758785255318,0.3916752493072557,0.8521918635094512,
 0.30307565845480633,0.3602748017356654,0.12651379416563968,
 0.9468240167375543,0.7209787716054096,0.21169767042877552,0.6368713837568112,
 0.17557942119485614,0.5897294834508578,0.3568429071003757,0.4831423318802185,
 0.01839215726041643,0.2290788354598885,0.5361348063034869,
 0.27273986538930783,0.021600688383359934,0.3679060344527616,
 0.4351862398891666,0.04373481636756438,0.7191500042887865,0.5697586619648874,
 0.9376165235323679,0.1442765382455541,0.3496522847320428,0.7226276129355279,
 0.7136010295626681,0.5633027247072039,0.8394383217673589,
 0.045115352773138606,0.38163195731348143,0.6000530914574397,
 0.16255601942526265,0.7509634309793102,0.584959893781975,0.3812291493215496,
 0.4458745752410681,0.7393933468605713,0.3845881694869824,
 0.024595984417173256,0.9942735275319633,0.7357941937149839,
 0.9314674010614616,0.5997992779389266,0.32966182872839744,
 0.47406146889077205,0.9112276462147436,0.782724700941879,0.14397870004364655,
 0.40699302286271744,0.11039466315061275,0.8331804963895977,
 0.8615270930179133,0.9031164309503703,0.21999342309004088,
 0.14417851526324588,0.13481127560862805,0.014538347788185635,
 0.3035490465842615,0.5386567184374552,0.9852142441394345,0.15287968820526332,
 0.6461313060152302,0.9170709764918403,0.3736777079824505,0.9185238852458739,
 0.47034615953196157,0.5747187465194875,0.8242103636123718,0.7447372586500646,
 0.6814050850290918,0.6310520001374865,0.7438571852252447,0.7186416974022293,
 0.5589426928125364,0.5547655943418736,0.7094650060469853,
 0.043951441667856916,0.4536373576871263,0.723477630868699,0.0660337560559412,
 0.37440358568826815,0.33067310440768094,0.8626173349817415,
 0.5425003294973334,0.06256822573884924,0.11478522538682379,
 0.2502354128628985,0.5569476865612318,0.4213694164356522,0.7682379008166441,
 0.20581283928983451,0.4655626525122698,0.45962651470178884,0.737444318403016,
 0.3627628181538082,0.09263600242780257,0.5882328591752103,0.6540748693319054,
 0.7457473860427914,0.4312586919656606,0.29642440007613924,0.6007889764311043,
 0.9976513619016486,0.016516111943103318,0.06933410479097046,
 0.1804620076000183,0.1476897091157433,0.4512696432082036,0.8155047151705821,
 0.9124102794794785,0.8031977799282612,0.7386186790648942,0.4742681339722523,
 0.6438949762146657,0.10726187461263303,0.09179666545733212,
 0.9898586383946638,0.8447454668870146,0.4168939846538283,0.6181325896137017,
 0.8476997506732927,0.8400702634059716,0.5914558849455298,0.679472980344567,
 0.4548713480295353,0.2434664697683081,0.44584355005438026,0.4915870518911114,
 0.8525001101128473,0.4364332767590633,0.7339311007494708,0.9583727476968597,
 0.44042518991622104,0.04775382913777493,0.35875248013878314,
 0.02653827354063365,0.1720352669987495,0.7355122066479259,0.6130170942527903,
 0.30466676996632636,0.18140462341364938,0.3976203726476545,
 0.9244868827927073,0.7220199803054317,0.4917069310485639,0.727322844575175,
 0.8076652520924722,0.4291763718815894,0.5228219832020037,0.476147404133065,
 0.7167896184432643,0.5587621784182372,0.8560002376802436,0.5400797574265035,
 0.8888124503601659,0.014038532766743606,0.24871384637137928,
 0.22615507197128037,0.782321301400754,0.03464518566661423,
 0.34410831768822814,0.4303814216008033,0.9830796126106337,
 0.08381944385314544,0.8732357588754787,0.9931565366955003,0.903040990296507,
 0.7848554912308734,0.26250215940597765,0.6959062110693314,0.7542603257093666,
 0.9378414229279554,0.3087560428700118,0.6093022009446843,0.6654619962993968,
 0.17023853778380493,0.2372957012260416,0.18189863137842865,
 0.8591040509992776,0.40948264056589434,0.7652740348822518,0.5904799762578911,
 0.5584466821451198,0.2159494140343291,0.5181498884624245,0.9864027860598671,
 0.7018440243914814,0.8787201813501179,0.9750922791143585,0.5882988893363986,
 0.8782815593981093,0.7972175921374243,0.3116661049351318,0.8156942198671587,
 0.3148578093692226,0.8628737480924046,0.206953252515106,0.11596553019809297,
 0.2963895682052461,0.960202078384818,0.9301550792866793,0.5991437283314157,
 0.7317676678342142,0.0774955009552023,0.4346654630002944,0.693832731911074,
 0.01057693782778002,0.23061851646374898,0.19947597028567876,
 0.8294587108915076,0.8388083398036853,0.14306189183073437,0.9722503168993163,
 0.2764484046065353,0.2622073123252736,0.9085449965383627,0.2585261240505984,
 0.4415810741186781,0.03506392734139907,0.711908145271071,0.2675273046121277,
 0.7672487957265804,0.3516910542134878,0.2737114819108184,0.3537110249140718,
 0.2884317607765007,0.6183356811532569,0.42939039718888916,
 0.39001869644389486,0.2865743004941582,0.06878645887953294,
 0.6354659123093621,0.37392547229853235,0.7510015199225253,
 0.28077515848890233,0.677543116502107,0.038070413289229776,
 0.29192946617944404,0.9934402082942808,0.07594258749751404,
 0.8351519963121674,0.06857876934472296,0.8043072366544667,0.5837138579272559,
 0.7981496976716942,0.2518403361502908,0.4402176886761997,0.43754323512618887,
 0.6917857331150827,0.9652373917191334,0.6270917884667637,0.17011912738378931,
 0.24448644065595637,0.030249939596304287,0.32322553333139736,
 0.14915812686663166,0.31210627017286563,0.037331279099114156,
 0.6376816271461514,0.9847842114831689,0.061735469560990075,
 0.5947996824891988,0.10665540199953072,0.35030435353963696,
 0.6749104789817424,0.11309125253497487,0.992092535210572,0.673212191314783,
 0.8100930316973812,0.32523497601728657,0.18553588703986745,
 0.9734039421499727,0.7516578484063942,0.0637051727603013,0.29856232519438497,
 0.07748121305704214,0.19453384317764755,0.8537853295323931,
 0.8459881312195241,0.4795273172198158,0.10272329635162625,
 0.12451931821175855,0.09777400119179014,0.6238785705032038,0.436076513379151,
 0.5450033992680121,0.39081781350815037,0.1884924760133101,0.9965650913215206,
 0.023846881780176132,0.06948796622020725,0.36760099477930397,
 0.7788045035362807,0.4846100313904309,0.03321798771548845,
 0.38425792837580475,0.06635940132293527,0.5179376716280166,
 0.7584270512307099,0.27516212492873526,0.5643228230082066,0.9053679153847889,
 0.780020528473064,0.7266918922727523,0.8977325659767224,0.8342499772010803,
 0.6623379732776008,0.04820562638522474,0.7985309112193819,0.8024681221832287,
 0.05443016144820656,0.5182731795528241,0.36197711682246825,
 0.8314065678575703,0.7450788947177207,0.8808589858994686,0.0696430412888639,
 0.46779744904232523,0.07456822683037656,0.6555566678542247,
 0.12341023462527978,0.7506953782801494,0.3453067140109183,0.2817054092442025,
 0.06287252589544401,0.6245999798487274,0.8751856445430591,
 0.30949737072792294,0.1771470326222071,0.9252751528292471,
 0.40304277288742785,0.600763655936737,0.5067333257640348,0.39711049410494204,
 0.8117391835784181,0.10111953293959897,0.5362731488407689,0.95524671338771,
 0.9536299976396251,0.11131930839069204,0.8106880344431014,0.7026333154984588,
 0.3218800131505013,0.8845681971264774,0.7996243028463731,0.5320378092483061,
 0.6367843352139791,0.7131354547409683,0.7289740491634397,0.2906850364491549,
 0.3578201620470759,0.9328822735554274,0.4673844912588099,0.08425455807633764,
 0.2937773099703678,0.4669473515053929,0.27950657208767116,0.6195987205380767,
 0.7244614325765055,0.9050372614811306,0.9520450435052387,0.30666265151323735,
 0.3998492747640816,0.44210009470350964,0.5095977471352278,0.8817649619401744,
 0.9203383528399246,0.6965131897464626,0.4983418054730764,0.9222991638993706,
 0.7706188419298985,0.8456259877339339,0.7046022324082826,0.5199468351088666,
 0.18978108751972422,0.8466723876432494,0.8324931102853901,0.9037854992460836,
 0.8816605291242696,0.4399458777698906,0.692400400046946,0.3449626343806701,
 0.017160185103690107,0.018075895675995324,0.9457442794658921,
 0.7148484901484748,0.5971308044080152,0.8396593209192919,0.9980846839544063,
 0.32154632643428216,0.8570264374466059,0.37268312382137025,
 0.6141342700352416,0.48563619148625414,0.7847860924422974,0.909888674672897,
 0.7271594142939639,0.6869532526762485,0.1212632261389811,0.1553973982117104,
 0.14651231842796264,0.683595555114086,0.15415796933872394,
 0.05533537686456824,0.27327376028735395,0.5243925627267374,
 0.6285284668427716,0.40902560091876783,0.8529828404770282,0.1739603490050704,
 0.12584151757815876,0.5784307452136571,0.7941200538784964,0.3081845290184144,
 0.10361810281162243,0.20403126225137724,0.5479417402036458,0.902938001360361,
 0.653149228097519,0.20587939139933287,0.7550591386921354,0.9438516708114,
 0.42519317132572265,0.49396003176846437,0.9000391425455843,
 0.05076274155471938,0.3844366830955116,0.700156196342004,0.9334656800293856,
 0.26224991303258327,0.061996379862420126,0.6086582132923117,
 0.3924707927668941,0.974781834686148,0.5466475366468583,0.9507986569046079,
 0.009804419243545248,0.591472636844702,0.2903193543830689,0.8604797790497556,
 0.8562107893569454,0.4393612561237068,0.08252031497685575,0.3570296428743046,
 0.22980024893065643,0.8518104176913435,0.035520531458989035,
 0.7994382001933177,0.5616132631198167,0.40182935644199813,0.2504607403666309,
 0.252271754947313,0.12245968437623611,0.2535357815398134,0.4195381355794403,
 0.5321991591909302,0.4527144893909494,0.09006831532893811,
 0.16431433805832785,0.7801954820356667,0.4057259425383971,0.964926495372337,
 0.2131122315166073,0.5923659050727371,0.29370941195741695,
 0.01661320800737187,0.02770262070299012,0.25610250493919495,
 0.14683096196357182,0.19857002825463255,0.8357520132047811,
 0.35085954260873276,0.7148887686378795,0.30918853506786226,
 0.9105630162813231,0.28952050820549713,0.735354779943672,0.24715087018750703,
 0.8353882061649587,0.33966243696306986,0.6864003694650606,
 0.37738415724637997,0.7426730860963023,0.40830013633617157,
 0.4049401130893918,0.6390338921269394,0.29228694446341486,
 0.11458331937310262,0.41899963003270335,0.0799088581824895,
 0.5928067429328013,0.16240449614517338,0.04733331146100983,0.283014732470523,
 0.670494321399694,0.647470296056261,0.2558400682452884,0.8292695609383732,
 0.017536942410306455,0.334649600115442,0.1503748397944079,0.399764425063729,
 0.42613289920857855,0.8842670964239212,0.678527732518675,0.7223217638668973,
 0.6791043690783435,0.4538941319835993,0.8548345701953898,0.8435528110365136,
 0.5013550530521851,0.8603923958614272,0.46607886875399107,0.9247660230082503,
 0.9228634586619527,0.41928293540175887,0.9672051567482441,0.4273803583528437,
 0.43785035527708915,0.4181184977375745,0.7878122218983447,
 0.025877984684600785,0.7147550547039999,0.6506263211240775,
 0.4220053337757732,0.033583653788485046,0.8708056316802849,
 0.4222689620566463,0.2511261385495006,0.6085580201659458,0.09076256298863217,
 0.9055910186205506,0.7252641480653741,0.17998145329260695,
 0.059732252934651786,0.4893881812085439,0.05549904304987763,
 0.7306009389239088,0.34354277984484527,0.32432847711009793,
 0.04688690024312203,0.36753307417235814,0.5043244807161427,
 0.46497197190599504,0.6822544962243656,0.24167518178194902,0.519348916674912,
 0.9149264580487367,0.3591622000380499,0.9347404465380104,0.238959164659027,
 0.8540243960306875,0.4806778474393446,0.781614627073092,0.1288070867847798,
 0.58063279546898,0.926286449758796,0.9676074503767017,0.3240284346592579,
 0.9270678659172489,0.0620428343314412,0.10772308010689546,0.1610276288732646,
 0.4592601425311297,0.28782992328938306,0.8329526583604562,
 0.061846445469283196,0.7943134981484563,0.36224488917602393,
 0.6671592941451645,0.6254861077269996,0.32375104907738295,
 0.21483469068137584,0.5762305636687478,0.7135828177963881,
 0.07501183902738051,0.7556139661964535,0.47967478763971416,
 0.01866861290014643,0.08190342738548873,0.7288595418881834,
 0.4544223421507261,0.7137481839141621,0.1058980933482716,0.9473858636228343,
 0.7800430468997638,0.37827007907506593,0.8372705280423685,
 0.15401285236855022,0.5315643245962103,0.3171070161834709,0.5385149540361303,
 0.8079139597567211,0.5050188440900528,0.7216324497757058,0.775421177492825,
 0.973674346974466,0.07626101370707383,0.44851524001750875,0.6961642046700744,
 0.20209372015419624,0.7040325698784198,0.2605950451039186,0.7033044711307076,
 0.7769154442128909,0.9606280730164143,0.2611174198933852,0.6887106845639623,
 0.8634233801448479,0.9173682358361461,0.6122781213987589,0.07075003142379976,
 0.18815720736506591,0.40966868864018025,0.813831477421897,
 0.15123732875328133,0.6153422261861626,0.7307175036738378,
 0.07975781813372596,0.5151217342984349,0.4004397446799727,0.6629362516264609,
 0.9970724017757562,0.21554917067730428,0.2357945272473011,0.5653633569771735,
 0.6675235031063355,0.3040505046189691,0.9934321049779204,0.47676900904330777,
 0.9603261527515987,0.9158619183005775,0.022407183385077745,
 0.6698716614701266,0.9768367343700436,0.35701750322728054,0.3446072871625774,
 0.40057524111717535,0.9682947391703708,0.005564875644412992,
 0.782712739186379,0.3095713898218262,0.2718050505613139,0.9951233106058195,
 0.009581866531483851,0.40821796436597657,0.06997206075377849,
 0.20578535889617433,0.9041365403815323,0.6086115883952634,
 0.08629986363787445,0.8474157383063936,0.2811779794638658,0.781054192299461,
 0.342992155833852,0.28199825007169277,0.728305055737303,0.7075881663739014,
 0.9375386020628929,0.8304426962741545,0.7989434461430958,0.8736017158752613,
 0.6617287217574903,0.39603888793757736,0.4170975080565189,0.3921577738409653,
 0.6858976027244769,0.8689748336133784,0.07764317686706312,0.631149393169909,
 0.4317530377518436,0.6456573665574337,0.6354703062856645,0.5132035407998007,
 0.07650463613909186,0.41436996792317304,0.9299328647584789,
 0.4262713214800449,0.8715232653633451,0.10862645599124976,
 0.038486257650063305,0.36346561055284055,0.8946142939648005,
 0.8261734249752049,0.7217804576443996,0.6952474413559471,0.10082005073963685,
 0.8670437649087478,0.5757419087261723,0.5497129981377955,0.8348598715047643,
 0.8815490381437164,0.6673867594415167,0.7838000550244024,0.9483190077267372,
 0.13944426259460208,0.47891838219629135,0.919794068974164,0.8218167054938353,
 0.3995121512982208,0.7925764121631773,0.7816376530526943,0.5391995482161996,
 0.2188912163601744,0.7419320777722985,0.016947947191456647,
 0.23128780311641373,0.49387124027667983,0.6247837474646258,
 0.007919316306564617,0.8491505602237795,0.2585616281177818,
 0.10605876781310752,0.9433943805842293,0.011312171964446116,
 0.6448331338473967,0.9112926988229496,0.6325326887150692,0.3713742118511658,
 0.824883730590751,0.010028585571577708,0.3354029315133371,0.9006365529453928,
 0.9489574094849159,0.3743222933733179,0.603376418329306,0.014511767671670395,
 0.1127288665827193,0.1371238779510333,0.3572847517385064,0.5803824913077277,
 0.7304905985275314,0.1859674191085241,0.3179013626339592,0.4149415225986991,
 0.09013890109094957,0.9580495487682227,0.029501313519800898,
 0.21340236652318667,0.5524421396637483,0.7786709152339646,0.9394550327004283,
 0.9684315528992171,0.7590065994411345,0.4623063659222496,0.6051422671293314,
 0.7766056655966994,0.5013974883365193,0.610832133261771,0.22422040734253657,
 0.015049854535753804,0.6850439981824901,0.19362555478161347,
 0.28008516079377876,0.41766878227714466,0.20630125521754894,
 0.7517054078875361,0.05066838436546406,0.2478073493702535,0.1145589781812466,
 0.28531496723569316,0.7864973821174497,0.4155590818874537,0.5490195181433288,
 0.9784498506456687,0.2188007227008597,0.4250075382766111,0.5943118683163964,
 0.22678013868189806,0.4881937002256239,0.283332090018965,0.9300640152564518,
 0.928957451834593,0.9069108087597216,0.8433103446136745,0.29295131034788746,
 0.7071639581867428,0.20886963321269392,0.50108892588687,0.18665595866230122,
 0.4930998704390033,0.6519159616707867,0.8905858535024724,0.17859341050173483,
 0.712203997299069,0.6100995216588267,0.19862565911374708,0.4102847730919783,
 0.5808775933771234,0.6787811564896752,0.020572729914288157,
 0.5562757232446286,0.5514997788587055,0.6857446343944552,0.23031691958071,
 0.038757122384181275,0.1507395078388124,0.8492377350777681,
 0.12875711178215732,0.8457976582034761,0.29574044815503553,
 0.3506906462161159,0.719614467812311,0.9848967572539128,0.5429238465900834,
 0.4360766508790481,0.15686611386248273,0.5977211447970379,
 0.41219308482967687,0.5348734084122688,0.3352974680208647,0.7825241773622928,
 0.23069525724188567,0.4940743647728223,0.3242916206532912,
 0.38418047732592353,0.753802409587497,0.783304602863106,0.12302137163177151,
 0.30708977346173116,0.14500018203899656,0.8875903418888904,0.817381500982622,
 0.10051181169766943,0.9735025151411842,0.3241611176538419,0.6487821473914113,
 0.7019076596828643,0.4999283688710362,0.16990866938909488,0.9158510675459077,
 0.03631609753763576,0.8839683873939069,0.9690141608196057,0.915458153931265,
 0.1898198565642083,0.42709549036163186,0.3584431448091373,
 0.06599886311785386,0.6944144496322263,0.11696433839249876,
 0.1733376997552717,0.21545525996093584,0.12277834069900573,
 0.7486653507608001,0.37736654199875386,0.8764502779296452,0.6832937769719265,
 0.6800841191242283,0.26620805426467253,0.04449091896818658,
 0.4470716719166229,0.49231235930923567,0.8522699836944146,0.7774132337928175,
 0.5191120356835007,0.5935009582684632,0.8565537907386285,0.1819620234306467,
 0.2495711939580506,0.01221893929084572,0.7871058038961437,0.9190872235203574,
 0.7389386835088004,0.8203903553107018,0.34518455169022044,
 0.29890323624525494,0.7628405317419784,0.18901721001497396,
 0.3793901900840373,0.8483670537113127,0.9366924977520964,0.5487983303453488,
 0.5261962468279744,0.3059611529285756,0.23042053713066935,
 0.13545461834466443,0.06886814868816393,0.3114330518471933,
 0.7533743917258033,0.3774569014040614,0.9548621572911877,0.47822046493415515,
 0.47907919169099944,0.4755238824308564,0.8549150999970153,0.9930919271635344,
 0.7412239749090946,0.256653382136556,0.9728951573352893,0.3606604514084927,
 0.5163506484054813,0.931974148583547,0.22571913281333134,0.36258903220125205,
 0.14656932109946919,0.40070385532039055,0.003241814368825424,
 0.6217571540679765,0.912664441990808,0.5449244887356494,0.8240273874134505,
 0.639056530996627,0.39295204491222835,0.4186988975699888,0.08560843284103048,
 0.564712937485891,0.9380193268323364,0.20622640086999555,0.9315598073238235,
 0.02215288197998333,0.0610057130053252,0.1654732982352468,
 0.12164350221884224,0.596368403769271,0.19072359971249453,0.5390607644004731,
 0.14296548680249965,0.49959217615358587,0.5751894929268218,
 0.7048936212458804,0.08363853504894292,0.033647703898474934,
 0.28742269446093727,0.14434719415612984,0.3372788518998453,
 0.6766602074653908,0.8647018412745632,0.5526360132990004,0.7677583650214472,
 0.13942558585661335,0.7583553149658364,0.27845916176434415,
 0.16419570388122684,0.9757243489599,0.28369669774566275,0.746945781386245,
 0.6072230221850181,0.17040479531832053,0.012728737609155272,
 0.8728980288195172,0.6885354891176879,0.020388235975839164,0.457413410285912,
 0.7755639998316217,0.8145344346756871,0.1592055036956248,0.46008443720030134,
 0.34478364949374607,0.9220116076020847,0.9053691679853891,0.5749118866202096,
 0.6689662673036393,0.034037714698212485,0.3431755106249341,
 0.4694604149216557,0.7832885073932702,0.633572720169859,0.776520866616031,
 0.49848178387216313,0.7417110850287632,0.3680784220500757,0.8999790305589489,
 0.8624638319942689,0.4581441205536859,0.14867996650174997,
 0.25486987393477156,0.528236131136787,0.21377578712820133,0.3959991516012522,
 0.18296913963912076,0.9342047348425648,0.10679926350859492,
 0.19771709996179043,0.9061867304155008,0.1374083390093498,0.6140468144177569,
 0.7493144963703585,0.9473217585296821,0.018868853146265963,
 0.8378344459695738,1.8619215035808878e-4,0.6832133889665616,
 0.10890292826814663,0.6221664128280747,0.35731714219047206,
 0.4345592700873574,0.635607175875153,0.350974365638015,0.7483622127575716,
 0.6251446325247965,0.0787759817692304,0.4261975386591357,0.7552770903459897,
 0.44805365228726146,0.11473064354530083,0.9584614586397285,
 0.6887370132644344,0.8466545865389766,0.17295802184082743,0.3744748009530052,
 0.3356730568428459,0.41756321454882517,0.3232903642177538,0.0839838977494406,
 0.317132344541318,0.5178846911234546,0.5470197807632103,0.39221084087524494,
 0.13873893694128248,0.5544593611934274,0.6268999470287985,
 0.39327527259695794,0.8494316599022228,0.06838195706209205,
 0.8016479689196753,0.8598257372079094,0.498957647864697,0.8729596161120262,
 0.27409760175056186,0.8769959288467082,0.4116637553121768,0.3971741945635614,
 0.6313177964267906,0.29307840977112865,0.622197371400063,0.2941290328620654,
 0.6623559234474893,0.7299002141951929,0.9171297734898451,0.23175131696191076,
 0.9830814222370545,0.8208917432605338,0.5967044932124517,0.8427584741268341,
 0.8741885215584301,0.7523767728230047,0.6128199881304546,0.8545932592572711,
 0.08187010884752677,0.7804684486449748,0.029080169609903117,
 0.14457874944485571,0.16785900971704382,0.632070684371311,0.6134199280274665,
 0.2259499307892634,0.4020864996813126,0.7916123654455818,0.44745232830195714,
 0.7017600515535284,0.27102012647078055,0.4168509543800558,0.2188652968973197,
 0.10852364566589334,0.35602818484380305,0.9856111087608533,
 0.9959492976580041,0.4779537052427828,0.05812708887448215,0.7192872155364464,
 0.7450135779526526,0.7338673678031501,0.8913393720739359,0.5487042356245307,
 0.43818239952183446,0.7291650828589679,0.8375913989283387,0.5844122332161972,
 0.20857992500231803,0.7989343503479202,0.2653997841224145,
 0.31366473097293457,0.895947571115653,0.3401940799956201,0.9969309020106688,
 0.012309398356029067,0.013820760853405112,0.15913544474266494,
 0.9685898954242832,0.5896474672435841,0.7952076937561479,0.5932483392867046,
 0.2724819272645058,0.4046037753777858,0.9749861480885725,0.24118421029802517,
 0.8712262255346106,0.33978076068859775,0.687306516283152,0.0534180309737341,
 0.1874781256421737,0.5701049012409668,0.2805700310473034,
 0.029865876973667826,0.6027223520344198,0.2136576454799325,
 0.3612987222852255,0.18501458772710722,0.5223834873546219,
 0.20485402236237638,0.9052505971019599,0.11089579201580879,
 0.6482469676989187,0.5210210357214033,0.6090197269862377,0.4033537350731049,
 0.7655565315908269,0.25217713792786167,0.09997814098583901,
 0.5402004420065588,0.9573017810818775,0.7163748104851281,0.14883075370776144,
 0.5573703646185808,0.1399223628429327,0.4753159122625802,0.9066431575409917,
 0.8402985513655276,0.6917337772109615,0.5114361060380834,0.7112200068936944,
 0.6617089316194159,0.15489432852984697,0.473018675740854,0.24709918722279767,
 0.37573987554835897,0.5165798695490248,0.004030728996931354,
 0.7999718787463682,0.07216766150206178,0.5409142075672921,0.8706498463257889,
 0.5583637895107544,0.3053134332631551,0.47955794178303257,0.6678649572947017,
 0.7475306816359744,0.3675725699005049,0.057852023880722525,
 0.5397173643893356,0.6871009062350686,0.6604061625430764,0.23866922722755213,
 0.2106406044135759,0.3581135720074857,0.7995862269169728,0.9558383429386141,
 0.9179504437601634,0.5175396055609338,0.24061468358092652,0.8566823865660574,
 0.8486643474541443,0.8356807785744333,0.9464686315129524,0.8591667975472354,
 0.8853031111232591,0.166466575394149,0.4029130782191631,0.8505502893701645,
 0.6707158691070321,0.5381465329084266,0.18296739648763605,0.3059611839779195,
 0.595447154292528,0.6647822781952739,0.9336060405340452,0.7366467463143105,
 0.7049139031378664,0.8926884642025034,0.10918474699783753,
 0.01117313983066659,0.9676609868601771,0.038658205751051944,
 0.08903592042725783,0.5520064518482224,0.9452482089573216,
 0.03221948596490798,0.10111930868814811,0.7016322960387968,
 0.9771228966730251,0.7082270855643213,0.6337017770253907,0.23600127526140124,
 0.8811019958429545,0.2508545596578705,0.3843087892535717,0.5269028633572169,
 0.6455318456865492,0.7785094531908755,0.03480782052186049,0.6150461000316891,
 0.3423298182760294,0.46623699199889,0.4597377634499169,0.18333148514853848,
 0.3825185316175619,0.8467799031053332,0.28890415982257767,0.7015793988633026,
 0.636014466457237,0.3005082771450649,0.017271730011376163,
 0.11173095072152084,0.7290526515573885,0.7043419908476314,0.9723896190230881,
 0.34135516835180046,0.36453754454286136,0.5307281770321942,
 0.5870901162547656,0.5405662730391243,0.7393625643464528,0.2388553479278216,
 0.030911338808924937,0.45039081022099214,0.3090905957271982,
 0.44233050879450775,0.6341879315069979,0.0027994035150253893,
 0.42239319384810825,0.06321907390054549,0.2224596081957131,
 0.8262282296664653,0.6117691428680059,0.525502579059482,0.03943649822324413,
 0.7601240506774036,0.49279315945951363,0.8250086645163693,
 0.26267557632875327,0.7360761183678591,0.7477694785513096,0.213161375243487,
 0.9782742545826699,0.6185270993490619,0.10084900832038113,
 0.11615983497580662,0.8993380780915219,0.7243526484737502,0.5992174076878412,
 0.231522130114997,0.3349222645191565,0.7029798516718659,0.8469798491302286,
 0.11511352102210426,0.1079370346799513,0.49698054692758176,
 0.32298513479420377,0.7634255364528383,0.5104310058062901,0.3153327835138914,
 0.4915198412703877,0.04538029279394573,0.36943821926719655,
 0.7888746798004059,0.899522286565904,0.4580493509736706,0.7626896131277555,
 0.9900603690720869,0.8937820051460739,0.4322417637839132,0.27082257589567793,
 0.40756948145900185,0.29568284187463967,0.033557186316509946,
 0.11777596643856691,0.5306685972356475,0.6328271188292152,
 0.25945791762913895,0.5893444594178705,0.6932663304826565,
 0.08872561856866468,0.7354577646212684,0.7639923921637369,
 0.27861084363328814,0.36745366021893455,0.33245878899643055,
 0.788189867124375,0.26754695260047967,0.5349972708522086,0.5038173208443699,
 0.2871128207171286,0.16901515938141554,0.4942661722195075,
 0.008198805916833973,0.27675674427421704,0.6217300519065236,
 0.5477973011495587,0.7651925704484336,0.6481339489535767,0.13296330322859562,
 0.04916193001413893,0.39509655280527856,0.9007296438553004,
 0.2595531871788126,0.29472591162830486,0.28486785133449977,
 0.46196684686094946,0.32940747211496224,0.39804393803242455,
 0.5440930754568534,0.42358738195965706,0.5419437265206288,0.3549141446427523,
 0.5012937807385653,0.08494685260211277,0.8858634761075597,0.8167700808824633,
 0.026325733417607577,0.44718261491200206,0.26440199744302706,
 0.7073080871379972,0.6998183540772082,0.5735060965236799,0.2706003791672299,
 0.93721786340144,0.33946906467257887,0.8902490038579979,0.7740913031798734,
 0.3751419287000515,0.35713911385420016,0.5523265718411098,
 0.12035196302337703,0.3586657309615454,0.2960403636721576,0.6059450655330257,
 0.6370172936127211,0.22774207413280312,0.3737986379856987,0.9310866827324397,
 0.3947516004603703,0.4987725766668658,0.24628133045800682,0.9917008968566833,
 0.43777947853429344,0.010602290192491037,0.5818660052301013,
 0.785847827124576,0.8896764418386178,0.19956249889951083,0.9966345845036026,
 0.7849222396923312,0.14825620851539667,0.22982495231639843,
 0.24436022019888282,0.6990467117448422,0.8745295045327649,0.7362491933373162,
 0.06687622720169983,0.49310750453478636,0.2545911829011557,
 0.6275109302260463,0.644904268057119,0.9999853565582131,0.2764375226654078,
 0.6513954477997608,0.4178568329987814,0.43853300120518446,0.9610292211511112,
 0.7447847954186526,0.4545572281903889,0.285008633895416,0.30718871555556615,
 0.17649624587366286,0.7079061808252093,0.6259567014905633,0.5825127184623643,
 0.8056161843087877,0.6053822315270795,0.2272955031794367,0.6852533599756873,
 0.2648065630334737,0.005495825726313908,0.87116048813342,0.6375271065797142,
 0.2781549687131365,0.41755361687997605,0.05801372912873071,
 0.47184480847228705,0.2884989055444458,0.9116855027435624,0.8682358159675148,
 0.8152863882863949,0.23724137324533778,0.8395619466223533,0.798949786282054,
 0.4968882845406809,0.6503096794109824,0.7523719288916502,0.1341610543094307,
 0.8633735919805228,0.8504479160803056,0.745725358728503,0.4371884211773752,
 0.1478520339305247,0.794804961444876,0.8739093410658954,0.2604690746460394,
 0.9545396703609189,0.5701516616167597,0.8796966879166064,0.21006751654384792,
 0.8306421693485087,0.16182055086206915,0.33101699867733925,
 0.8158279919283247,0.12936722026071612,0.4844627387224779,0.782790509640118,
 0.619070645382667,0.27748630036295174,0.5564160220609738,0.8931607395047206,
 0.03547210753841035,0.6434424489090425,0.8148775697870372,0.6235240455802818,
 0.7586069896850279,0.05950767982829319,0.4151134426707772,
 0.014694434802429712,0.6238267034806994,0.5313279912867,0.044941078915243016,
 0.8157337082573148,0.9174556049939707,0.0735098373353793,0.18692290722503668,
 0.8130128126136984,0.7952032930437588,0.8898611324284844,0.5769257082108747,
 0.6645997351589977,0.38472026666632064,0.6192360890355401,0.6143281978773638,
 0.983613584895161,0.9292153202584991,0.01792120644378581,0.25953361661994234,
 0.8300503749804862,0.07187892566458398,0.3864193487313077,
 0.025177003585847957,0.09898506212663072,0.41835961791805276,
 0.39452815078761627,0.6496386311016555,0.5150851340247207,0.2720079512914608,
 0.25183357080048796,0.24826910743799058,0.1375813220072487,
 0.6656328665637797,0.067807318395402,0.6889797060755314,0.3260183938372181,
 0.270066227124107,0.8424607450628626,0.5536234855747145,0.049775503184365855,
 0.2962925739943896,0.11025857603196232,0.6874071408626281,
 0.0013496992418426945,0.4948811630909181,0.35641541251998266,
 0.09884456845703604,0.9120454547679921,0.2960436497954593,
 0.16606524140036305,0.5625822615269959,0.3779096920483911,0.9140271712155087,
 0.080768612766978,0.8214647297074187,0.4162845162059008,0.47294953783392724,
 0.5446566864353516,0.6057038228391903,0.6457739085596623,0.5940775416295223,
 0.23486328681347335,0.9949346181838883,0.34487011564457404,
 0.29471699563262244,0.4565744360075584,0.8891650858605964,
 0.30164992100077204,0.3930676576215384,0.1669538674040828,0.1590996777171486,
 0.3424326404775857,0.3769566597726114,0.16568486849523067,
 0.25371140901070277,0.4410273725847882,0.11204187652778896,0.815112581233996,
 0.07024969025771421,0.2904990678211309,0.0559283156933138,
 0.15825269338389347,0.18949000663315574,0.22702002744904426,
 0.3136704986021317,0.28806954818801245,0.46485684707922514,
 0.38596899522284056,0.7254436587429426,0.15806765943228507,
 0.9157531251213846,0.7360356677167605,0.5706290547798012,0.14949768312551082,
 0.5101195517684592,0.8313737094124803,0.11563551739220257,0.7412266087508139,
 0.41648452321804486,0.4572421466462532,0.9243861789784655,
 0.36731553301091235,0.13760264755446583,0.6726649677851164,
 0.15181520457407682,0.030355680377617267,0.02062934358519708,
 0.2744743551289037,0.5244177614526423,0.7280756051819509,0.6190892363950748,
 0.5542590289506724,0.5395247126460381,0.43957614567367653,
 0.11550419092977093,0.9749099184965864,0.2848213004184931,0.6278450423094144,
 0.13919225430561877,0.15441405125899788,0.5980154985143179,
 0.2965271998013055,0.42162494430755504,0.2558226928619518,0.8547919958016105,
 0.40779339604631226,0.9027017055296526,0.2964730240938661,0.9468786313876001,
 0.6738525583220134,0.09837827836534274,0.06482250867585393,
 0.10812083574408149,0.9581629444669573,0.9591817182165384,
 0.02012179371184253,0.24423372311764624,0.7389045173623181,
 0.14235060489285656,0.8203002464085933,0.4506429556808378,
 0.05079893911681255,0.21603575695560495,0.20749220027194637,
 0.10848083586437762,0.10630608365343241,0.12736625886106778,
 0.9068866434869528,0.6985926169797815,0.07869267209373865,0.3630936546150342,
 0.6992824343362329,0.21162056711494448,0.5912162628509403,
 0.15819199714918597,0.2124378831843463,0.3906552985841423,0.3782340987863687,
 0.7815516016262098,0.4908656043998373,0.13440909717718597,
 0.08402070387568805,0.015530621251803711,0.2505165796816218,
 0.6979835968965824,0.05600054545299038,0.19698830967857228,
 0.011550478282215781,0.5865413669460677,0.3663601957247338,
 0.34811153774026227,0.3336824922719688,0.4006137562305203,0.7830271801307219,
 0.5425841510349503,0.7606768734629804,0.6670025387719803,0.49595065716552633,
 0.712289711968899,0.8048582668334512,0.07099246676373494,0.03230967747683933,
 0.9970719712392299,0.42572415821127196,0.5992787935900593,0.7601121814961629,
 0.12125065751427899,0.8580037995955425,0.9638324510908161,0.5643836182951175,
 0.07739650563741785,0.10817833856341164,0.04257570267957955,
 0.9491299898840317,0.986631974249615,0.9079500613934273,0.4495281614970761,
 0.4327166566068258,0.004581520688886731,0.8614649113725378,
 0.8986351265571705,0.27208779871669997,0.46744434480680963,
 0.3964660795408582,0.11243119396775048,0.36493761809356795,
 0.8998333210667775,0.519541728449981,0.03059515885838615,0.9206040919583394,
 0.07654843537430911,0.5556796675772927,0.68240813569145,0.2671697863180864,
 0.1285305895462746,0.8418637059473772,0.2506537705674765,0.14664142580301098,
 0.854416264959273,0.12980059238339003,0.7233675578954317,0.42495692825193165,
 0.4324306998451395,0.8641546414442428,0.8033130021973313,0.7601836597255657,
 0.5642294948316361,0.8669306315420805,0.10703040723281765,0.2869200203231116,
 0.23629372927452585,0.576181755985172,0.17323005596922592,0.8149339341846638,
 0.3909623949446148,0.8579240620238824,0.9137688144990281,0.8337644870992128,
 0.7466893561176865,0.7303452884645939,0.534883542647914,0.8873237307736686,
 0.21236922697313387,0.9857920907168232,0.6137438861560873,0.3427887513539214,
 0.07865805542671822,0.3723253283101444,0.08199823522814642,
 0.3988899691007227,0.5132004066730144,0.6887666717979334,0.21445147652914265,
 0.8139115503504716,0.7379455105048099,0.4143238389506616,0.27039624915692606,
 0.736139020359374,0.9318971263601342,0.25125974626381675,0.5234604058518162,
 0.7963681598586219,0.33180885367110047,0.5785319889662258,0.675343569399985,
 0.8947864241554963,0.97019559409899,0.9077441568568927,0.5175432439972774,
 0.21826044097303354,0.6253029172138849,0.1833320985486464,0.3797542789434749,
 0.4331549954158871,0.44741595365976883,0.47128174311229787,
 0.19639465830912017,0.872515049805632,0.905218265185649,0.40809149290585855,
 0.8499487695229241,0.250115039003836,0.7035323250621182,0.4173704419499491,
 0.7134470369381543,0.8578947521713167,0.7020995910598393,0.8512191339028203,
 0.9719753327725296,0.8233692922419792,0.034403841986729855,
 0.04522018517394655,0.5738994871441591,0.6000981341193876,0.9759182262872634,
 0.628459683056126,0.6962649813872372,0.908514779297568,0.867835496685147,
 0.11396907505725029,0.9368939981160621,0.2877150117918166,0.9136626614732947,
 0.7080415905721059,0.31732877599537235,0.9606656547310399,
 0.21236474915092896,0.7385325872223035,0.4115095732021532,
 0.062431040084334755,0.22868528468106597,0.05269072325175428,
 0.2037250759390875,0.39932878470025346,0.0532852386244852,0.0431265855880385,
 0.0870596601787974,0.693066521870747,0.8645129334157519,0.7940362179298316,
 0.20031973642845013,0.6534470358858748,0.3860260796401388,0.6294252753975829,
 0.06273947286508452,0.1995106247328341,0.537303494477422,0.5357508978471296,
 0.8361134208185665,0.3352607856212134,0.17542397784515684,0.326580209478472,
 0.47937608023336853,0.018319006851706843,0.16395067389093898,
 0.961866208128817,0.9137976069734085,0.27518553120511324,0.17017402860758046,
 0.426212488399923,0.8015675716072087,0.370089296667939,0.204664214395198,
 0.8556233642990545,0.6293175881770388,0.31460299491560717,
 0.36873993376218217,0.7722505976293014,0.8124434668095255,0.9643754785888671,
 0.29841094778024135,0.627218616128041,0.2059549981610148,0.6260274326881146,
 0.25587442693750984,0.6370250106017807,0.8306570612522064,0.9104813614450717,
 0.20030192058095397,0.09303436756460903,0.09170174622656524,
 0.6986229485442722,0.9252483363502806,0.6516374891727053,0.9606809664622549,
 0.4382862622737814,0.5331504027547724,0.854294906274225,0.20981325445680754,
 0.8911485502089145,0.05115625026562842,0.02468872506550912,
 0.7622381002201383,0.7936982712805952,0.027444657039947473,
 0.5995874921357375,0.8215143538688005,0.7265763772456387,0.2716644222703857,
 0.9626108012753434,0.2272178437431417,0.868910464623696,0.7217098953656909,
 0.30521030351411305,0.18645474804696127,0.3040256517207499,
 0.6082964204852055,0.5493803660006804,0.5194074997355654,0.5760957156601859,
 0.592684521198892,0.7759754805106296,0.9847373664434542,0.4054776330105211,
 0.6518898297657348,0.6646344424989135,0.884168199180857,0.3311901220933229,
 0.1267297182634629,0.3176860817874231,0.3949865041934203,0.09090883273730732,
 0.7959805009548173,0.21893920156414326,0.24529429931629732,0.518268701571907,
 0.9417392374697189,0.8713633621460777,0.5256661992719993,0.6620015139365233,
 0.2203927065443676,0.7350602827644731,0.2539808076058915,0.33595422602894676,
 0.5285291057547198,0.28535411929913435,0.4521820450820233,0.8100040984756908,
 0.538676182514965,0.1591172735662214,0.6432933983018086,0.30469861138578536,
 0.15780266581839852,0.35430508363782987,0.15788540666518514,
 0.9445129428151611,0.40876783785101045,0.6016716586758887,0.5156939911805166,
 0.15002158876162863,0.9580783412881424,0.2918627278261724,0.5336223877261878,
 0.8277585831725646,0.7474418462528921,0.2544970749893385,0.36219836571095054,
 0.7510600917303997,0.9832580150173095,0.9150172258354557,0.3259263290979413,
 0.7838004969715976,0.22275427140784,0.700297884431801,0.5344126441168819,
 0.8207970424859463,0.5612546633551962,0.08063914511339709,
 0.49081598823545214,0.3442119623630233,0.17426500834388103,
 0.8759959442501163,0.8848252447196086,0.6091519047440173,0.7476778225154209,
 0.11664687289722664,0.8375893917749427,0.8648439335044471,0.3293067048262438,
 0.7138313346021946,0.20458586652194277,0.5838073280500709,0.564377860617282,
 0.9187061272658474,0.09610950433000531,0.9166126283469472,0.2675842969303035,
 0.8058356233753229,0.26820657794352265,0.17103017706262014,
 0.7409726052436985,0.7092249682085159,0.20236078527459356,
 0.07976603081710598,0.01988416775697399,0.05223440980438199,
 0.7322068832437412,0.662998582930075,0.9169647920478301,0.11755046575143802,
 0.6689216956445319,0.4572142348912128,0.03305675023733823,0.953433373712519,
 0.3245272004211013,0.49001434386214926,0.36604590925874314,
 0.3218812729880691,0.39747534709070176,0.7161051768364856,0.9511384491568836,
 0.5654142922074756,0.9274816305856377,0.8645369579145956,0.23092308095771796,
 0.1578031532702855,0.11277062963066176,0.12201515766172955,
 0.5856105924816366,0.17643743153249825,0.6419528046633998,0.6452788350743197,
 0.017753676231272997,0.26082605910195455,0.003085148547381622,
 0.21224259738356166,0.5450883008422314,0.7263109489346111,0.4774490068645648,
 0.6926690947911305,0.22047579814683815,0.41941918463706096,
 0.9545191525022565,0.35527478085744746,0.42531903463351295,
 0.10451955518588485,0.9241330980602281,0.2444805121000334,0.6965667250069871,
 0.90233943420203,0.27607249273253853,0.28953573386849685,0.21572269055730686,
 0.4555997448073441,0.2734429835145,0.43998210703996166,0.4902305464131206,
 0.6943965345854055,0.48654458092223896,0.10412840448999283,
 0.7236578766262637,0.04001491456391848,0.9233909774472884,0.7837531282861899,
 0.45857808480231954,0.5714488737036457,0.5959411405036723,0.4754976226243858,
 0.15105752989874832,0.8295997103839788,0.5634905232151729,0.8737989140155921,
 0.40261345053378894,0.40961770273282616,0.24967495568339537,
 0.20434157477971748,0.342908998827957,0.4095465433737344,
 0.006081044474481434,0.6770278346800317,0.05759412128884933,
 0.6148019806082555,0.4341611692302647,0.28404770066803986,
 0.32195523637707557,0.05266139609017939,0.8403571673118412,
 0.37267115544600116,0.04519405273105903,0.8318777070894174,
 0.17751634988730824,0.7363017029929031,0.16442869963584272,
 0.7838465726582993,0.2721992418895428,0.15633341721895744,
 0.41415890304709446,0.8398830790748042,0.3289630153999913,0.5584290862801625,
 0.5252595578693372,0.04388440431148355,0.8700251432451169,0.5283585614558169,
 0.9947274338501766,0.8338268478213158,0.9268764544698561,0.6897644407549117,
 0.9315323202957648,0.17860878888999077,0.5312967156237333,0.7228548245582539,
 0.15722934642216746,0.40070731298920437,0.260965572265066,
 0.09867154804938483,0.9360300735874427,0.31215184854486944,
 0.49800737694018526,0.16823040400414946,0.5793467902396263,
 0.011518017238516864,0.9689046307181336,0.5276460443025286,
 0.5295065641525412,0.8125771013605878,0.26492899626985955,0.5984826191643458,
 0.4487888167877896,0.244651013522326,0.4019495910912798,0.47570864095959786,
 0.8424551329431582,0.062365106450908314,0.6980010452103577,
 0.33066558381196953,0.3425826339474223,0.5669698356651007,0.8094719218600432,
 0.9971167028115311,0.3507228777624156,0.3455320414232056,0.9567261749379262,
 0.725887688259189,0.24204831364709567,0.7523748953511599,0.3867803640573003,
 0.50807793767538,0.8998666028015878,0.637005844437343,0.8061609079196675,
 0.12428735067885521,0.16637609975006207,0.36549496669714643,
 0.3710616488516719,0.3324162451668171,0.9484546979102161,0.8576752639720933,
 0.48482282439719504,0.010288494708710738,0.03860653784642565,
 0.9557728518210629,0.08025458515563055,0.04330025491122047,
 0.9123651184331766,0.8538899006136151,0.6907520055355745,0.6755333682087409,
 0.8263460060542671,0.43545606803073333,0.7721009303298245,0.5347993697849309,
 0.5890156353521478,0.666347095108293,0.6732161082819288,0.3419513206121432,
 0.9742593016899397,0.7103512892724093,0.029415480758245405,
 0.8410278642478568,0.7359966898823513,0.9479299007033758,0.7009969531926824,
 0.5373161755406435,0.12645641052885426,0.7251103189560806,0.9680362413290244,
 0.5408631262772459,0.9993410379038354,0.326666715756994,0.5598628290829295,
 0.1338041341955185,0.826047360732458,0.8187590484673781,6.662062334612529e-4,
 0.3683027396864931,0.054669070751445004,0.41909930090793734,
 0.6412838187935413,0.7912275867697152,0.5056468397806059,0.0395246843092254,
 0.404401558598126,0.01638592744067524,0.6390782916135572,0.46238343264360926,
 0.5455969890761593,0.3215470681101795,0.5100916561450382,0.3377970733139708,
 0.4667931032602136,0.7644044122130995,0.14910504151284076,0.5197640669553584,
 0.13250543133194137,0.30391597782574464,0.6180099001547346,
 0.3181193542628378,0.27277223711586185,0.25561637540904414,
 0.9150595907441923,0.809065589686455,0.30257127166834863,0.682584457707199,
 0.30324363980142266,0.6546687662558608,0.16547049504657096,
 0.6850369160504288,0.5370433844073602,0.46366469395985477,0.2517008784215162,
 0.40130489225584853,0.24615289257805228,0.5446610836359226,
 0.044213993887970204,0.9923719012542385,0.5407469517575688,
 0.39290052782694795,0.5038807472279386,0.22724779475488244,
 0.6902182504270395,0.007766497642601622,0.8113901573531719,
 0.0382998807799374,0.7968496742619897,0.7920695541811017,0.9284749333706502,
 0.2001273819459587,0.5525782107842817,0.9827363258818729,0.5360337764364496,
 0.851690965813041,0.6796218772919735,0.5630459647750368,0.19502859244175197,
 0.40513176191537403,0.8319732331935757,0.39187609398085355,
 0.5750063714687872,0.9240538526576644,0.22439485220898836,0.8268686363056037,
 0.5322440024444997,0.5747510503821878,0.09246038983787708,
 0.14625865003882654,0.07627975448694169,0.48754111918766374,
 0.8777584520707519,0.712445892075409,0.09751162744904185,0.7143124111720968,
 0.49993210448108805,0.1471064524841046,0.4944136165244074,0.8316803020640436,
 0.22110570348186953,0.951777365915492,0.30374965811736276,0.3660353852082723,
 0.17427213101605588,0.4343113580518436,0.4017259398979951,0.3471000828475471,
 0.7011466958622792,0.45784255964658505,0.07662264415945774,
 0.16872811959651113,0.798065699667817,0.8389476803878252,0.2831279899821597,
 0.03563334951867614,0.7354641440761029,0.8570993462940145,0.2015622949651774,
 0.8908093274626815,0.7822753653955834,0.21378709657143025,
 0.04622132397396572,0.8397165520623848,0.5578334388137625,0.86997677557518,
 0.09393333842870166,0.9968733685733525,0.02020726975868914,
 0.9064443709899033,0.647253129010743,0.05741066298859976,0.23455749734943443,
 0.731425459725106,0.7671327656056255,0.4091874638751569,0.9702716401178957,
 0.5251009919763102,0.10591681680629539,0.9588904653148084,
 0.22564237562568845,0.45141082623848805,0.5623491940994483,
 0.9441099059139857,0.17177399704009777,0.1248671970344305,
 0.12236083895972805,0.10343738586907769,0.33062642062165515,
 0.22516994717102223,0.33838656495415975,0.9903609893455007,
 0.12408439770109325,0.5513269861228179,0.5122334331175248,0.3301395529374873,
 0.31979391783742106,0.03315093575048178,0.013828712522892972,
 0.3448852459608691,0.15431762355147793,0.8813530733322636,0.8134854138887435,
 0.3610846056341378,0.9932578204645314,0.0767042066627247,0.33488226339469707,
 0.521158523200615,0.9123192900720497,0.22452187340983,0.40438777801786097,
 0.9102987793679531,0.6199082929363586,0.6170870505963075,0.7550696210617305,
 0.1778106696953884,0.6652175842167115,0.31783984637162477,
 0.39672623586168465,0.6033914458539095,0.8370994605552884,0.6742062409450368,
 0.6313060368753345,0.8124773560084798,0.5946246155583058,0.8858710742869009,
 0.08698758258520545,0.0013863660966237923,0.22338870917101894,
 0.9348354129244215,0.5885673543614117,0.5305258119535232,0.9493924120165118,
 0.7834904747851589,0.9009183659225706,0.6254841516842129,0.5117142569713583,
 0.4057150186551721,0.44276130436122973,0.24114633922803286,0.524728348896741,
 0.30447450628016237,0.9676970076338718,0.2680323173900617,0.666351321292201,
 0.2606784703902496,0.10218190989605946,0.394319962602353,0.4902244802402268,
 0.9042832031726654,0.25590897343675867,0.3932516699243298,
 0.10145537639955238,0.6433348343262909,0.4087153585392803,
 0.24325100202900451,0.31380681172278324,0.3827740656500348,
 0.9946824229890623,0.046915101972888285,0.5694095030025891,
 0.18668423778900667,0.8410381082647111,0.4404451328844301,0.8547697956640921,
 0.9897204688516373,0.7565082040375872,0.17574923587025915,0.8613376747260962,
 0.25278104147757086,0.3310788503222126,0.5626481309324677,0.7878554447631209,
 0.44835268750398816,0.8382516228065393,0.08260647612781746,
 0.29935931461223964,0.5212356294286886,0.1683866812862861,0.6872140901065443,
 0.4809490785872096,0.7463422018287197,0.447606486185695,0.5079070493961994,
 0.05634244600227489,0.3052519477096396,0.12031042856142982,
 0.1108636500131851,0.12962138198787598,0.9745582925754467,0.2717315409939376,
 0.5469962667157191,0.3187967342745943,0.3773479944061646,0.04116398273883615,
 0.4668843290542616,0.778234566852748,0.25768675618396286,0.5962824118613251,
 0.5363934997979494,0.3328336335444062,0.9201937673639686,0.9716200299205542,
 0.1372419278351903,0.22605079000995754,0.025517559490987107,
 0.1969250116224459,0.6658355720021242,0.4225756567007972,0.24675953600617495,
 0.020296454693651267,0.018348014202773433,0.9883755053509341,
 0.8041302375544571,0.5514625073257072,0.9619868101483766,0.9348391461528642,
 0.17852192180161564,0.4426389036009277,0.6349118256110391,0.8691571093846188,
 0.7695876409607783,0.2611757932603682,0.5706400903111672,0.34849521742787026,
 0.26716455617428714,0.48383925001475747,0.4303303841871441,
 0.21301311217542795,0.08541461157619057,0.7041176105736463,
 0.05017686000889943,0.8624082080831335,0.6126854058170458,
 0.037951529894612146,0.14618107108541967,0.95111983009637,0.5551058140562043,
 0.5626576126829336,0.5995201560266483,0.12554454824454342,
 0.23793437246788385,0.27541110206577246,0.6806802526249642,
 0.08446756914175368,0.17568617271591824,0.8670664611737577,
 0.12617762491886098,0.6543395120355029,0.06473942582818237,
 0.23746221527440914,0.2830303275825089,0.15169653021464247,
 0.5208423440713396,0.5079390157135865,0.1303311954924814,0.661851030958244,
 0.6669995155606419,0.24175389822896987,0.2685325344423809,0.7720573003750387,
 0.4496929476208398,0.5924284994310126,0.9493195059717382,0.3056873021201999,
 0.7725018023231867,0.9790209669858381,0.5509706032782001,0.7004888526564954,
 0.2978562697097765,0.00500737449078692,0.35826293573300816,
 0.0794693212120986,0.3828687163026874,0.6081828811574267,0.8588343247380795,
 0.8009482939361229,0.6626342014312291,0.7481523001625368,0.15092724428169235,
 0.8064423505177254,0.515351507275037,0.8153151800206618,0.463185666527279,
 0.6373455315288806,0.6582529719580199,0.23039514633410063,0.4596516985567123,
 0.12551695708796995,0.6741544653404189,0.6808825277087791,0.1680462090031326,
 0.7021998822591176,0.08632134268985692,0.27052213891037746,0.746174565114779,
 0.2741324584480922,0.25050006575514605,0.5336823172031628,0.3344258988946123,
 0.5959367845401116,0.43632813749567223,0.4293336201792962,0.9642437805762957,
 0.970724503808915,0.8554366822020648,0.22419073714448212,0.9313734012772604,
 0.9966843180157364,0.21100512181450182,0.9103823030189687,0.2698676877055558,
 0.15517528441473083,0.618239156504116,0.7022054346232144,0.5347730008111382,
 0.09107508862832303,0.20072121820674793,0.6305719452455049,
 0.4081208977155084,0.27669261719029614,0.7951155491044598,0.5800517440319737,
 0.6119539833751407,0.6645235255518398,0.3245257768884231,0.8987802427779573,
 0.7549224283673605,0.8297908189713676,0.752881051687815,0.9693815907842744,
 0.7568718487929499,0.8154381099727028,0.42630456176757936,0.7300884346708161,
 0.15392461536380908,0.5754886419864371,0.1471511700721232,0.9701392369216697,
 0.3893154642087413,0.9807938352406965,0.5946047056488256,0.5302778475666969,
 0.4032672640964474,0.3029441588310898,0.1987802232015049,0.3356581850593716,
 0.6948764622868937,0.25892554308812765,0.018363045495012997,
 0.5546585093275849,0.6833567057030611,0.7919263614899951,0.9054765242631999,
 0.28914070964987726,0.470539195806972,0.7800486490810521,0.38758769040837215,
 0.8287127272790303,0.516697157801863,0.02053278961324656,0.8314870079285195,
 0.9313760572479568,0.9579511830621629,0.7887212023312423,0.14780277499028027,
 0.7021012411892797,0.5050099258333551,0.1426431977155097,0.26369706828386164,
 0.15677546132727027,0.8307175368351222,0.3980354327932636,0.4551441945222733,
 0.3645010996415041,0.9511579493328897,6.235746183036461e-4,
 0.23360120774497684,0.36655953300282684,0.7067907070149015,
 0.9211542242947837,0.011390777818905962,0.7824792216863473,
 0.4366217812583686,0.5006135339958534,0.7757241320532429,0.625730091852367,
 0.09033906768159461,0.17822129397860276,0.18382463375088143,
 0.8955833541506814,0.15042405025437044,0.8490353006774549,0.7323980774337854,
 0.5984627972493721,0.4405599689207449,0.6474474797124221,0.31844369792139315,
 0.3165870326428254,0.4850756692199941,0.2694151370799801,0.3472406868835053,
 0.9503795897880345,0.6084345929300117,0.5615591353781264,0.34172495166033523,
 0.3943106207070002,0.12756520136713456,0.8497135153627858,
 0.034881618391774505,0.2591471619824526,0.95777404346611,0.3566546363858746,
 0.25611822082939617,0.09020685074169021,0.9896225059173253,
 0.9636874509846839,0.40732418188266295,0.2900671993048921,0.6890282920283558,
 0.33522750921953526,0.12132313397040961,0.33665205470367,0.6228207196695595,
 0.37690335944425224,0.23493375677080242,0.9136967703022083,
 0.9427026323730726,0.0681743745102279,0.6902836553958207,0.10528110962301107,
 0.9766526592769926,0.7308704358610985,0.6239367523681282,0.5321899212760418,
 0.9865838315690308,0.6858612903032933,0.3745499041650633,0.1544600197478132,
 0.5311299649036314,0.6891340852201289,0.9177493231878817,0.22799344558724788,
 0.28526333181359664,0.9350565725301336,0.2134928942832235,
 0.020994250853716534,0.17147145222370141,0.8731890052936888,
 0.999542075511246,0.5746141144109197,0.16453513366848005,0.5898018699073649,
 0.41346392636295615,0.5171447958443278,0.1573318188249797,0.8561775310067757,
 0.2398811090468227,0.11748272957283412,0.2553645356800628,0.6275895357356226,
 0.8654788180696908,0.030529155946630193,0.4199124900695491,
 0.9561573405601316,0.2844351927534471,0.913533663856811,0.9485469764658442,
 0.10223410064077476,0.5549776942624477,0.817598369968353,0.9576599750148242,
 0.4534325726224302,0.7800350138212923,0.4103513221831566,
 7.305780638935389e-4,0.8985069661288024,0.1368030717158495,
 0.07574899270787872,0.6176954454473831,0.4787108560707387,0.3660375629677004,
 0.33354062043643196,0.6026089175607051,0.8800061025900234,
 0.15704904128358743,0.3530098819055473,0.18312792421772273,
 0.8984565806577927,0.8524755535157231,0.33644897087446435,0.7865884106653733,
 0.55030773060852,0.8600763260626436,0.05494159104002194,0.9313274014528771,
 0.19240124454316954,0.7178475464021394,0.570945384594969,0.8615437476796931,
 0.5904803944899775,0.6131902734632658,0.1677201257067069,0.3929151953644082,
 0.6563526094352197,0.17816956040808063,0.8622841288521934,0.8446021249016982,
 0.6026805933118182,0.32417993090112707,0.2880191851203532,
 0.05169965938862242,0.3387627847069028,0.7454471016196186,0.0243432476958777,
 0.6867472211705075,0.4115891115561874,0.07817781191035045,
 0.09666084120606844,0.6635550422496108,0.43292430488058187,
 0.9579913605577127,0.20628409251241098,0.8132869618065979,
 0.10209122755510236,0.11330124170708555,0.5270480345281374,
 0.5806868260806933,0.022873719931376768,0.9491952814456696,
 0.7728307274077998,0.30704180529296266,0.6260335725176776,0.8132331375951359,
 0.47819530815505906,0.7423231803755466,0.6001731509711054,
 0.12577259912518268,0.39393682455481627,0.2942936890573332,
 0.9241326353280954,0.2984207067215574,0.27812788529416055,0.583161705458382,
 0.6934729236722086,0.5776956856698116,0.19757032615446113,0.2887455103590538,
 0.5553341189958709,0.22757296327108345,0.0827934142257668,
 0.10457549192928584,0.43254126708491425,0.5665569175276636,0.850988257050718,
 0.9390827560353604,0.014524223925869606,0.16116946489549666,
 0.624162201212116,0.7104064104169161,0.7695178784549206,0.12402256062944894,
 0.3322869408514505,0.683552672357981,0.5825635176028635,0.6439858363401305,
 0.8317662774425175,0.40330566198704443,0.034430338523522064,
 0.5762352813645591,0.5192600655231586,0.4477256974192838,0.952482675138634,
 0.39331371457451914,0.18784065271247452,0.19074929724115297,
 0.8687037178084548,0.552236695137152,0.8559680018560694,0.9702876075286067,
 0.3697338053244419,0.8965268937336093,0.1288372764505039,0.3066125875864294,
 0.7102143225628164,0.9870048090418134,0.31107826011999284,0.3650329304839959,
 0.454188751678718,0.9758108367793068,0.1331195786006808,0.7842618160271124,
 0.9735475567500085,0.9892759263427184,0.9744054515646164,0.25738574071013853,
 0.2571645268999109,0.5772292767184839,0.31090108444048636,0.9618145682156775,
 0.6209314297036845,0.8058527694982086,0.12701974542967376,0.3670733008145719,
 0.6680633178052591,0.5232802460956396,0.3240674675860923,0.9350693911706539,
 0.30525702463906845,0.7574911830379588,0.1782656736509558,0.1874296381518431,
 0.4657368561309777,0.9865781221381615,0.012518258531498572,
 0.11183733141367747,0.93820853778217,0.76186808287075,0.18659964480425362,
 0.43428827939751113,0.28734177725999777,0.17117197440800924,
 0.32845271589724634,0.395619347523706,0.06882452554103335,
 0.31340741747710765,0.8713912210474485,0.3808985799918001,
 0.22542424457134635,0.2276709094309186,0.47027562121283417,
 0.9400292267545627,0.7320260233691124,0.6123357584737148,0.4390776892548446,
 0.6317452653282047,0.6096015486599944,0.15040057951001806,0.617923308592295,
 0.7743847109755962,0.7040609741863055,0.48477784554999004,
 0.44198535641319847,0.4535608618975898,0.02493486671666978,
 0.8630444769082185,0.7332852985290281,0.4930531896754945,0.22120831722416523,
 0.47358556380723105,0.35877932215975483,0.12320251026778517,
 0.5434382036102072,0.8996572772057438,0.26874076225396615,0.2267369908168796,
 0.3038980698342496,0.29670056463373273,0.9425322797299284,0.566380633723217,
 0.5586919737928113,0.16696675535456384,0.9654821584692497,0.9218746851478453,
 0.09067672138124139,0.9986551705926059,0.5878149104678996,
 0.44407031273456576,0.42311488220068627,0.3252946715543037,
 0.5507371215812519,0.35078386461703204,0.3746728334444984,
 0.10731722148544298,0.8899701361470622,0.7829337963549012,0.2644771940299451,
 0.879328991932635,0.2534427791949527,0.227242201678731,0.9399982375537699,
 0.4693225027154442,0.8951008978856064,0.8179305987539026,0.32043713993958745,
 0.3908869581947966,0.2677269866954056,0.15468170170305817,0.2220541865586808,
 0.8886399270197517,0.5040119691438193,0.6576446473943292,0.26527374867033804,
 0.45344641380936357,0.907603209308128,0.641804720074135,0.6729679300586402,
 0.7695437436376141,0.504671022769577,0.4712525294191967,0.1243205853139161,
 0.9765734278777601,0.16774267929677766,0.7109156478373457,0.4206212512575226,
 0.5530059333119728,0.05184399693394748,0.21973483698280716,
 0.7245035164361744,0.3450665851239183,0.08816341650528825,0.3747918587097253,
 0.08381495757335511,0.7494195994597643,0.48787512329834015,
 0.41935244377698055,0.9488749036889659,0.6858162634889039,
 0.04556763137359321,0.8336465829930767,0.17942788560126843,
 0.6154926564506868,0.1337421645321496,0.4307792298371603,0.37489748240078025,
 0.050226944306039245,0.46270912832601363,0.7702613337993444,
 0.903303705569938,0.7334954889119766,0.23337316318470924,0.6170277315471445,
 0.28769194203599635,0.2164441289214707,0.23827268807071367,0.571840851913676,
 0.325656724668025,0.8146515112656195,0.004528678372301087,0.8877432825383426,
 0.3539556174999995,0.6296691144937395,0.4741445652358226,0.9824378189954072,
 0.5669128368139705,0.6548614988897081,0.26227100313350815,
 0.40710284370645955,0.7263667203486441,0.6655974903822407,0.5395392754577645,
 0.5306324461618634,0.3764243035207563,0.32346845640809074,
 0.31843275770443424,0.42849392085798443,0.5162631057617221,
 0.1280374797022823,0.6210985684787529,0.9205109192884634,0.9297350680935793,
 0.7096362806003037,0.6668837556001472,0.3825524286914421,0.7953803929122225,
 0.059269625956726246,0.35809924576394647,0.5132366340364526,
 0.7259882678381883,0.3666376656892405,0.580498978658418,0.48437379660780555,
 0.4402817443206213,0.8310285936277613,0.4049458606131826,0.2037876505113606,
 0.10876934717802045,0.11199266551829912,0.3410397215578673,
 0.8941455247325645,0.8917423321510166,0.19626122473612528,
 0.08120049438033461,0.29908215949348504,0.9541800987193918,
 0.4659039075227158,0.17199265348045678,0.572179218748289,0.7400295641205836,
 0.10693630678154287,0.9231643186739896,0.4114057654676514,0.7923910260155882,
 0.5364527888745362,0.9633845131024824,0.7714747375910354,0.6870085887672363,
 0.9620670250508115,0.083456522888811,0.18221173006151115,0.0203364200629933,
 0.03328274322456992,0.2770582053455357,0.59622112259259,0.8179022406274853,
 0.6727322714171938,0.367631638271002,0.526657260338105,0.3164832434178919,
 0.20765191983772102,0.5734011511469642,0.5647975920158342,0.0988686190823801,
 0.762743389805877,0.37205163615698067,0.7189730356771343,0.46991112427435233,
 0.8224484866092616,0.9629999632141529,0.21272520435949038,0.5676320858882705,
 0.2192768348517362,0.15906081797614113,0.5921398315125671,0.4704524824791493,
 0.17792456608891427,0.08428947369230455,0.4381558046018571,0.582595834003588,
 0.997598732104008,0.5829925228193769,0.9894263704183744,0.5127387946130415,
 0.2853341188495233,0.9346942864054045,0.3110549631542542,0.5449200421866979,
 0.987068383749489,0.14947321172320027,0.1704648350298963,0.7344014604808018,
 0.30117150799334147,0.8163920505733379,0.14754727625369746,
 0.10750896886282002,0.6346399434653872,0.8589674440549057,0.6566280160145386,
 0.07773566213256411,0.3699215078605338,0.4063931028268962,0.7877777923587678,
 0.21378469183065496,0.36464740724250644,0.4642205473921147,
 0.30253929037743044,0.27277130936604954,0.08713313896449926,
 0.2810423474982988,0.3921401549893039,0.8710238916155991,0.2358724429389224,
 0.6596306331157901,0.23536615005246375,0.4391442470385202,
 0.24488048831178622,0.27591559883258565,0.31540203008463574,
 0.41527659924965954,0.05075136954051762,0.42283226475719193,
 0.5056074346339939,0.2857983659562685,0.15795007431404817,
 0.17917452274764967,0.9154654879797393,0.8843601038061618,
 0.46749012137636003,0.9771075347093592,0.9914192078689423,
 0.01694689215442402,0.7045425620932101,0.610934714705544,
 0.004640049883250175,0.06176528349840227,0.2211074534618076,
 0.7394368641610451,0.7575465684349062,0.42342629696270095,0.7022289593176031,
 0.7428513320359211,0.5652175247343626,0.06509677335633635,
 0.0018612218751091758,0.570093691562907,0.0453435309171506,
 0.9902972840014608,0.49592483148250976,0.8797784670021325,0.5440714855819183,
 0.1274085981106421,0.003432359289917364,0.14076021864466026,
 0.9311102661491512,0.07358702488618729,0.8627744515282624,
 0.18696774536477379,0.969024665015405,0.4953497194616374,0.8347845181104309,
 0.10032954129074179,0.2079644114634276,0.16060934523926507,0.354021032682577,
 0.46388076055349103,0.6990828936601967,0.8365582087081146,
 0.29346008377690513,0.9078716597858048,0.9976042206001405,0.5881816263190706,
 0.0635262115209273,0.47271332152976564,0.6045508859737063,0.479093098829391,
 0.1795409926990017,0.027738393177973775,0.660630293966905,0.8565486906638866,
 0.9116568891690838,0.22303675078834662,0.29848999184232283,
 0.6654819697823953,0.3978971177208013,0.9616924868828947,0.45740373269495993,
 0.12062964815964594,0.0131386528314551,0.3310811822244275,0.2682759635285631,
 0.9925557638121651,0.6911054065035535,0.10400472649045522,
 0.47664755887676535,0.8268987414002031,0.45008021955036237,0.449451885239031,
 0.20522609013980675,0.07668731337680335,0.16872854873842025,
 0.6747886071750178,0.553585910421786,0.18087983209021585,0.1928453310590844,
 0.2669579612319952,0.7597222094708441,0.6301892872617154,0.45502537780805075,
 0.6068451006565219,0.045406327924115475,0.9087713186459456,
 0.4112423902180181,0.5687242703375502,0.6571986387105155,0.9754707816737226,
 0.19011236794125064,0.6766976619793931,0.10824368178246557,
 0.6082486789033045,0.04272893521317622,0.8173617353702616,0.9967506593881983,
 0.7884370529551418,0.20986216132584046,0.023096454649542908,
 0.37938376634276394,0.21450134029421475,0.26474892133104566,
 0.8954497046556456,0.30513662420188803,0.6779187252550405,0.4237507730968614,
 0.262060757034309,0.7697186916177056,0.6059029196436432,0.8806539162639885,
 0.6254478183933707,0.44035852688244237,0.8805935531885342,0.9547880127048688,
 0.6358861494746793,0.3072788726826122,0.4602050424135704,0.4465982050933798,
 0.0014065429634402271,0.11767167133075518,0.11841781939766027,
 0.984251232805756,0.25649157429348424,0.8643988835395916,0.8892404005058167,
 0.9602315005541756,0.013852313980260433,0.6057021672637559,
 0.45429663750488214,0.4957827303679091,0.09388569137804914,
 0.9512649586519119,0.5426165314902218,0.62085071846847,0.2976438406394508,
 0.36999109305732114,0.5950291217542908,0.26680436456102474,
 0.06994947894379655,0.7812521133067956,0.3029096106822047,0.5663475112573233,
 0.18801660381962737,0.32476071756573843,0.7557097907606581,
 0.36897780094361854,0.9966809662051634,0.2614411951437242,0.5520867228947933,
 0.6909354930654604,0.7254411967949268,0.6233689418974058,0.17612314438330012,
 0.9226270471440905,0.3785045639013386,0.7941064065643448,0.2623179536100114,
 0.7455860489208785,0.3703770906269259,0.5217666099052723,0.6326479563188971,
 0.11005089599521811,0.6841925509914134,0.0991798156425785,
 0.11737205028930509,0.46627965129837934,0.5458592544923355,0.254232255710385,
 0.9602043482705227,0.803353728541891,0.30828503135036367,0.9368477599117424,
 0.13928220538764535,0.3978950495447371,0.4301736491870797,0.599046767166771,
 0.2797435371472127,0.28476288034882435,0.06741464940566166,
 0.6965165616906874,0.6363819028413554,0.6950516760555265,0.5122096467500041,
 0.36775142702758434,0.18350078132997116,0.15409110241316892,
 0.8744939678919177,0.1997763229905849,0.6559001457928693,0.03327111393949933,
 0.09385252850403714,0.1753161439995764,0.7596685442129318,0.6478039921410452,
 0.8521223416431041,0.017606781540476457,0.44151060225912964,
 0.7458752577200012,0.30913612219104325,0.8867931039829542,0.2218299069343086,
 0.4114133045852296,0.3794451961337918,0.4702810082235592,0.39135008762139145,
 0.3615249301682245,0.876424654633291,0.7772129106392298,0.7429133168383988,
 0.4613405609792699,0.39804848589529873,0.23328915317422116,
 0.15674510353725424,0.0679092896466601,0.7581399043495543,
 0.34489045725841394,0.7633866981412382,0.58159665671399,0.20343093162392312,
 0.02473139974592653,0.807478194967783,0.9589692955673237,0.8882481326281314,
 0.9878654639518726,0.9245406780002424,0.3185385412240149,0.4018737142463352,
 0.7271454721791439,0.70618455464987,0.5215086056774709,0.4304182737173672,
 0.886704197934998,0.13394948396239026,0.19895922259213716,0.2625997754469859,
 0.7224825611323868,0.28330390201878164,0.4243753269137297,0.7963776098811041,
 0.37681510107980654,0.23539742444411005,0.40630063483213597,
 0.8815620215924747,0.05173208758983561,0.4135409366813376,
 0.15415248593455355,0.044195446611401734,0.9031498721906379,
 0.4597359036981443,0.36855920539038634,0.6722855633200967,0.514307207318044,
 0.01736207226766795,0.2538809454831463,0.6450949983773625,0.8195576561813981,
 0.6756953702671042,0.7015887422888717,0.05813237079368072,0.6691098799052625,
 0.3640622006521992,0.10344188617881112,0.4361870915905981,0.6450299312040781,
 0.0995158246053115,0.2964403776827961,0.27640234977179023,0.690781704454811,
 0.3067170475513251,0.36896812068338913,0.7099052885649562,0.8345832010679513,
 0.16834340755240518,0.23906151381880159,0.6238904401954054,
 0.9985239405184374,0.6895136270106048,0.7257383539284114,0.6605912639596319,
 0.4878849172011548,0.5176389655008098,0.2351092591765792,0.6718519784050265,
 0.9388017767589386,0.8294265283959793,0.8277569395252379,0.09253723490051502,
 0.3958908215540925,0.1606034224105658,0.08932740602738176,0.8996250608184619,
 0.48331788065326153,0.7104667992184639,0.554383927037092,0.16954953360709757,
 0.7984971513490434,0.6142338134935292,0.08588705960752652,
 8.752993850544843e-4,0.6697663548777265,0.2737080119301869,0.480608383164145,
 0.24912522372463541,0.08640769602107212,0.10142597814742527,
 0.7948839259982351,0.7621466809766102,0.3394077474001248,0.9244963213981049,
 0.5482284878034381,0.06465196595207157,0.8268305203776816,0.9110896918703019,
 0.31074206221119627,0.13634929843423105,0.837180226456469,0.8176891941998443,
 0.6812876030022841,0.818627240088035,0.641646685326443,0.19368621128109131,
 0.5777066377494553,0.0230296909388924,0.7192903617456772,0.32179793775917365,
 0.9678181834169596,0.8210888226669146,0.7860570221172697,0.684023992495313,
 0.8275603050265702,0.3082285005116753,0.4861478516409492,0.8375564531659987,
 0.9885643547422277,0.4081429676383932,0.6041109679973564,0.5506676478415166,
 0.10615002459476763,0.7107826796467656,0.7649330985216403,
 0.25663265393725254,0.6867532782597756,0.0067362511108335354,
 0.30364876813668895,0.37150542905911077,0.661514027997492,0.6950873848969443,
 0.9866847570280586,0.7678245919260732,0.9764561414659152,0.5225937485497772,
 0.33784666502448324,0.7011737722575648,0.48596673517125444,
 0.24009182907962912,0.07041808173307307,0.09121845373286475,
 0.7522015358336761,0.8199310248233148,0.5498288535762539,0.3812122417407502,
 0.8695334637106957,0.5738543951785239,0.3776524116587441,0.987170622746007,
 0.9489046898603154,0.042473216015668336,0.108696951066243,0.5964157566312278,
 0.4752490952245688,0.6958760489510237,0.3740860890811267,0.7155876452761538,
 0.22002843669133387,0.9274292934358911,0.12237388367235791,0.22447870715287,
 0.285880870196374,0.7769331671075501,0.0727424319767247,0.710167774086294,
 0.3296386450572847,0.2817961407476025,0.011072320610478137,
 0.7624692558277347,0.7565425011350011,0.8824718030010039,0.2427311362568665,
 0.9548978279866751,0.8523236239272218,0.3847867921005952,
 0.0018905646775257967,0.1775502517132711,0.5801144654029334,
 0.32003687917270107,0.17934384470139964,0.12624439779789243,
 0.2998172393666998,0.9700323751297889,0.6098499279322042,0.4963220582594313,
 0.015998826727920745,0.35555000375298995,0.9791633672133271,
 0.5083464976010683,0.7745828439363689,0.6740219655367989,0.766782123985463,
 0.6935996365942239,0.7282747872155991,0.8979346282411433,0.4950162611035286,
 0.6064790746216948,0.7304799769776962,0.8037975203287481,0.40990079121795275,
 0.6254517587864669,0.7298545398754305,0.9451037888369472,0.6003043215978253,
 0.7604333635717907,0.8559456606053093,0.7030420400454669,0.7430650361578192,
 0.4846585261729207,0.16316535231642848,0.7475261755756062,0.5006928527092654,
 0.838053453154356,0.15285310406599684,0.42226529621748743,0.4029135347863021,
 0.9326278618904662,0.9119429611383995,0.16901545585602318,0.661497845337077,
 0.6105134204697896,0.40754189696049203,0.8402235087202563,0.6751095256333619,
 0.42754276052675655,1.2484753924113523e-4,0.3912303681822067,
 0.14630773363095229,0.30275339345107466,0.47096053212182665,
 0.5642816402189312,0.08939519695106457,0.37502324263710407,
 0.1660551579534537,0.22672081903372754,0.07440844463731433,
 0.10026272006153036,0.08641696829314094,0.9882611733745023,
 0.9267721619963513,0.3063328054139163,0.873900331878018,0.26564457872336944,
 0.4832968105819829,0.1413608508048193,0.47320217987954116,
 0.47675295894789005,0.1751077698537632,0.0321160789893874,0.6116322163616379,
 0.9355080772496235,0.4911461597932063,0.540893805590827,0.1496234532571291,
 0.558536048244029,0.47502944940196346,0.1738990725499886,0.9891711037190241,
 0.8196651663904768,0.7750909158853213,0.8852825272823345,0.5999918145071744,
 0.7954715271256487,0.13226695397111532,0.2514818374696337,0.4992836681564098,
 0.04362471831420023,0.3493791336452917,0.48953442790622015,
 0.33828339023287723,0.1211287213788853,0.7547584858885112,0.7208051955445908,
 0.30827032540701316,0.8409917702840847,0.9089676333187404,0.8263745793969749,
 0.3530151883044135,0.09791382419080497,0.9863181038314331,0.0363361244807775,
 0.9508544149624095,0.5172279036512643,0.18463059845227392,0.9226664285630682,
 0.35717191772261303,0.3398828397033318,0.6504493128623576,
 0.25238320808352344,0.446114265107579,0.8876816397130569,0.6838847455152953,
 0.43031119140702145,0.5600932496008475,0.9892158820538648,0.9282860123045051,
 0.5163947022868113,0.6950568968996985,0.7489742705034692,0.816072570135592,
 0.6534345675141678,0.11456980002969352,0.38965597384729056,
 0.7177251000792075,0.6604140418051828,0.9132642264522534,0.6847612752720345,
 0.7768281230266387,0.7080489547930197,0.323999464049071,0.3943410600569561,
 0.1429106571824641,0.6907146332736549,0.37565332108621874,0.8113793622961941,
 0.8467905989824995,0.056426514517069815,0.42145052886887346,
 0.28682221551202436,0.28780191136780986,0.6918725573235298,
 0.8682575545649117,0.3288064751633446,0.5703028571845868,0.98559142256204,
 0.9783775035262984,0.07998348579902237,0.9697780125365214,0.8884353609394443,
 0.18599846618059024,0.7803934273214285,0.2860749699979066,
 0.09650988971527952,0.5719604317277621,0.5480319179693053,0.5790131614764303,
 0.7655712476091332,0.15005992680193891,0.7541265612475798,0.6623287642420101,
 0.2866986760834579,0.9731295165297091,0.11898567579399832,0.2765650701032064,
 0.8903163989858855,0.6282455825874583,0.7207905070578454,0.1452980109354225,
 0.4138310985008451,0.03973748716654635,0.26558259557533526,
 0.9055089693775904,0.23017497617539306,0.03821459510287062,
 0.06524067892386609,0.4140425777465718,0.022960291864528526,
 0.822687263525844,0.16911722363338622,0.04048345102770745,0.88501889440464,
 0.7418884026013411,0.7153391908103588,0.3413614781770933,0.32546760092655946,
 0.8599376484383684,0.577962463236241,0.6296531639584471,0.20106874436567101,
 0.35450015745727514,0.5480923426792241,0.3747894838215434,0.8953051398911868,
 0.34113607903666987,0.8393946900387991,0.7484911399156053,
 0.09264486451892628,0.9402937442258901,0.42263056931585996,
 0.9988439705195327,0.9676969989357811,0.9183752029526535,0.6870697887225542,
 0.18482450509652049,0.6231482990392544,0.16946219013854624,0.867408796508972,
 0.7193563704707451,0.24945478451775271,0.5955631480289227,0.6478089272289977,
 0.4233387062519133,0.97681518106311,0.933359171686021,0.8298025225182843,
 0.6987510575475783,0.5074252928546545,0.20676880273733766,0.9182313332581453,
 0.895580982475368,0.3065854598255888,0.785479581568798,0.7370370444517139,
 0.5639692141622334,0.41487490515974623,0.7893304936911428,
 0.25283923971913214,0.3042678904686843,0.40757026482050496,
 0.3915659231231623,0.5592186657488915,0.44507827627628593,
 0.38198091445584115,0.21320689448158392,0.2428803093812527,
 0.23392951785575744,0.823378445741576,0.5037808222324875,0.8756280159618283,
 0.5785642728412043,0.30484801898992264,0.835862650421665,0.6216274309748344,
 0.6884903925107502,0.8568657490414096,0.0076174506248913865,
 0.6253388766553662,0.19458032715002227,0.9594021089713463,0.6004925376991841,
 0.14261244627381875,0.1593958801232347,0.21005847241054,0.5979707861570697,
 0.17198638180579373,0.7624467332820117,0.791302352001724,0.6496012426505664,
 0.7067768576090665,0.09277552990394167,0.8948286957994552,0.107386653263288,
 0.41485482335005996,0.743162007217409,0.8868489130366068,0.7880291731397203,
 0.5647210075427653,0.19327748165438496,0.6470817916995975,
 0.28648520481276296,0.5635005037454779,0.35570278419428547,
 0.8674248623507641,0.7782764860857111,0.7256137422268382,0.8137744671156661,
 0.21483047922575904,0.9106219101284009,0.7192014644080386,0.8946102033553699,
 0.8294870511663914,0.9950879183768316,0.04284233698551376,0.9318786246805362,
 0.06689491518637447,0.017746040624481263,0.569108940043441,
 0.19337386952845714,0.13254681023111292,0.8171506301189138,
 0.09027523515392433,0.8561469780241567,0.6736517802334931,0.5591851125745921,
 0.1886415681661262,0.028346992926331582,0.07937846425187012,
 0.44666247428723616,0.5258062329268632,0.2466546679784014,0.5304763917091273,
 0.4691673648949244,0.424216563601276,0.18794236909079676,0.725089103931743,
 0.8439868111302515,0.9284826381356862,0.6793383814762404,0.6208775368457516,
 0.27371840278181814,0.06814318609464387,0.15384865628098876,
 0.047255147445999124,0.020202093745789296,0.5838510089971969,
 0.33127749853397803,0.7788071190174747,0.3987006735999349,0.9905213125634377,
 0.3282984541142645,0.05679730473044786,0.4090671693546134,
 0.06751545745879117,0.025196921107209058,0.7960734313875992,
 0.935556059604723,0.5558859270900725,0.38163814768404847,0.05101329652365716,
 0.028568906241826397,0.5379362632114002,0.4216230556359821,
 0.30598738072319087,0.7363499955301233,0.24292902791739812,
 0.8253503479808979,0.43410579858137777,0.582319691250689,0.651317672288526,
 0.42924445885785645,0.2723377761202568,0.41217784309329186,
 0.4082283055831323,0.22822925571862807,0.40470029935584684,
 0.6056017016983527,0.5969572179187326,0.07315038210361413,0.5718421820104296,
 0.5037371706723336,0.45098291572241545,0.4259302082804959,
 0.27682258738530263,0.9734984922990872,0.4169381385517471,
 0.19279111187807796,0.9000236948556994,0.013416439876753672,
 0.6492986408844734,0.7628598195384514,0.6274894816422677,0.6077157541393262,
 0.1681873880377983,0.33036403811220183,0.9647550344461543,
 0.03977903330286825,0.9003768256588451,0.08914281246018574,
 0.2517515806100489,0.4928848117113891,0.8018961830353704,0.5791463420675534,
 0.505274843542391,0.03646702497581289,0.8724891871171082,0.03609032935562495,
 0.15336320636391187,0.9467418174107941,0.4084029436021943,
 0.16545017252054872,0.6060443117320988,0.4147204997529743,0.8141378639517386,
 0.9923030982538522,0.4194824886517017,0.6911472710975838,0.25553127132704456,
 0.12463948137619252,0.553535387030845,0.3208876323538197,0.9188830168185153,
 0.6822193948790817,0.14839843475206294,0.25655956300290184,0.756956404526377,
 0.7912561739613373,0.12981925449741993,0.28020876432381636,
 0.7596639314735957,0.3899447687914652,0.8525009245763248,0.6877049529225269,
 0.3025927258246359,0.6328133388704019,0.017431410768409705,
 0.26849384715549307,0.699629958444456,0.4063289219068752,
 0.025007489960617413,0.40369606398374014,0.36321998443605175,
 0.20691872840275483,0.5902702713095123,0.05211331296962518,
 0.2082484026232515,0.34466766639112334,0.66543820020901,0.7986330718585668,
 0.23045734616913938,0.9143937154805022,0.17950056230527722,
 0.28888942638845616,0.013890718640938715,0.5499242626293477,
 0.580442013290771,0.04625286418780272,0.2441659434261949,0.9817662478533049,
 0.6825258118521773,0.5619821546202322,0.25937298731163605,0.6190473652961196,
 0.42769880841168195,0.6198530487170898,0.7736918611407075,0.9196181502194267,
 0.7065525893932785,0.7855365182861126,0.5024782798945173,0.21539447150540858,
 0.9707975783499758,0.5631180897887094,0.9958866743398171,0.3765037124543196,
 0.7706915571846712,0.43948047307179294,0.7509743955379602,
 0.43411188949800894,0.7834044915299593,0.17538131541679447,
 0.5632490441590082,0.3301750901507774,0.7158265475751291,0.661075408371118,
 0.36092060753610333,0.404986542456621,0.1033206399991059,0.0926263045699206,
 0.7317494124692212,0.541096762639528,0.08083974764824153,0.13520951964196448,
 0.06874975147384421,0.39801513117297416,0.5841138929522934,
 0.5857286300598817,0.4851998552043866,0.25302210618650933,0.0371400041527985,
 0.9918637511908679,0.09146739245309221,0.22972621360226264,
 0.01481468038035838,0.42704063516413915,0.1120150007759737,
 0.8795782362024595,0.9327033776290021,0.24868384247877673,0.4214660199359239,
 0.39458437196595775,0.981776150106557,0.17975272786229324,0.918680735990704,
 0.7983492991610569,0.9836965453107707,0.6031000229416925,0.8561046971006757,
 0.30630321928025006,0.331683509524368,0.8484037397492388,0.1403877766203504,
 0.4084811204096932,0.5832741245896877,0.8177310517025894,0.6813365090101701,
 0.8140723232987641,0.8257565177986831,0.9680239685279198,0.8835110793893038,
 0.7019577319915113,0.6595350967630796,0.5330794845863143,
 0.059255588736817466,0.44795413713137466,0.9950545495510947,
 0.2592129359900893,0.3106084316197226,0.20827782792285865,
 0.10100698919175133,0.3375424140404447,0.5239660513404722,0.7759018027300545,
 0.15354856907809844,0.9856331192700182,0.585717560950018,0.6872756789082005,
 0.7328697595254815,0.5735775778484041,0.6176305436301881,0.20360190955783797,
 0.7607731953813396,0.4321702523888329,0.6247957160274566,0.9437014233116433,
 0.616021810560295,0.2531733388110208,0.8213970644625378,0.849450323586976,
 0.5404865299683993,0.4033887625923942,0.9865736075979568,0.7854200566956808,
 0.5291426010451172,0.5430632322463698,0.4926557115033736,0.41286240694274656,
 0.8276713812550776,0.3152574659319112,0.5264975058139784,
 0.056994482390168355,0.5283378387029506,0.752737892860971,
 0.02134740248077238,0.5235353063944124,0.8731632583752166,0.8235942591768652,
 0.757280826385933,0.14472441588553098,0.5113774879849124,0.6967358907806479,
 0.6301564104095116,0.9220726267228224,0.3347068898092044,
 0.035097914428936616,0.6105729420744032,0.5645678477578568,
 0.4924060422410199,0.9689670909183843,0.6342950857746352,0.6417798030752786,
 0.7629964462491298,0.8854569632327708,0.4139072201947521,0.5274465119804089,
 0.17532765506494186,0.8005410880394462,0.06650120866802389,
 0.04272740380805096,0.33376096946053446,0.7215120704570976,
 0.1705242206896691,0.898650672264371,0.6952595390523684,0.4142769871153845,
 0.14045428909407742,0.5856764789806606,0.37122127939075644,
 0.40902221136346606,0.30348349278921727,0.03355291276471928,
 0.4084796453491395,0.7151926036626687,0.1788281767483113,0.11318285047330967,
 0.060471952096684434,0.6963575893296303,0.07293222743475591,
 0.00542851528737287,0.8289045738761898,0.8471457816161613,0.9279675119744909,
 0.10436258081935412,0.7759715269948848,0.9155449765570092,0.8301362231514717,
 0.44177029688743275,0.921063734634031,0.6741301849717822,0.41189290558566194,
 0.09411534658915244,0.6345262133618765,0.04470151785068288,
 0.7435948060373233,0.29241833945995643,0.8630645065350397,0.5600383962461223,
 0.6796555921234704,0.7908042714212549,0.059665187559291066,
 0.6618452755618707,0.4805290822214443,0.5479830614585915,0.8558072395546148,
 0.5574507312901532,0.8824350545716241,0.016911378033509727,
 0.2990593030424882,0.11135690895346306,0.7263060727099127,0.9712760392890912,
 0.13488122695863314,0.4736437424778033,0.8604017071801588,0.8729078953738667,
 0.874603263643883,0.48954076351257403,0.1639813694311819,0.894590006024935,
 0.4262119428957951,0.012930223814454966,0.19589410432214693,
 0.3590794534629175,0.603307549296543,0.7202619523092407,0.9748915007570734,
 0.9845623190301678,0.21310100729185577,0.8415343082462388,0.8242621522885557,
 0.5906753342632299,0.5765897730934008,0.5538860978944299,0.8477783889212692,
 0.33009239117722755,0.24753450049160997,0.16420757797798413,
 0.8871389135458253,0.4022848729440285,0.9553779377441471,0.8710085378931155,
 0.967974125162242,0.2259995390632329,0.0027868985906754906,
 0.6552825193474511,0.5672357503775056,0.49865251692549895,
 0.02679049001174927,0.5077723772790611,0.2271999955635613,
 0.33727587822843974,0.4531777840926865,0.07354089569015188,
 0.7900943830235136,0.45437143163526206,0.6685385225834403,0.8343257650731435,
 0.2100357116491509,0.6215152290384653,0.6988952816930217,0.12771233332815246,
 0.1299033374502856,0.98873677296587,0.5250452127514501,0.17464674685020776,
 0.8166161857231787,0.45856826116581073,0.7141713071984364,0.6036688399456922,
 0.32091429673640426,0.5830819526560317,0.6329218432845681,
 0.16855819019855256,0.8670045024481267,0.23768432564974395,
 0.6410560278019102,0.8222492019959458,0.23643218760036278,
 0.16963074552228996,0.7721267767637388,0.34886083028227755,
 0.9231351262455573,0.040881770020786634,0.564357045406417,0.1675905178156104,
 0.283050269029821,0.8585137371991431,0.25684630426756316,0.47462403217644944,
 0.7003442855336683,0.4753398315779518,0.049400540569503826,
 0.7833895667154266,0.1803648607950754,0.4191190056388172,0.5739082324852613,
 0.5271352847195062,0.4885222661291586,0.7622553716937853,0.8463831719328491,
 0.572674872987031,0.22297393396587784,0.4688965466327957,0.8260468580447591,
 0.035701844397776794,0.8334540419303437,0.11096693446230099,
 0.8705983073032254,0.11820288309489158,0.7210467697913903,0.6111044016303702,
 0.39087026749244214,0.47695893963330926,0.44077851737837614,
 0.7749420447883248,0.04784077048238844,0.8462230398177977,
 0.27665905328702034,0.34786878548764066,0.7391872514122781,
 0.5057273973404327,0.1156401615392646,0.9599444853067755,0.6709661339815873,
 0.08468731068215063,0.19568374125122978,0.9460784993799564,
 0.9682812412784115,0.9801323840053523,0.6647467057265509,0.907861242557153,
 0.1769207739450691,0.7382065948704407,0.8789051862234085,0.7023203616686172,
 0.5992310343396088,0.5746810155675178,0.1754449326014309,0.02030202005561821,
 0.6614524983123001,0.6773352266952087,0.759172537088299,0.7609570792066728,
 0.9634529622543824,0.9255955412333299,0.9964412470438792,0.02296162295024473,
 0.48055477315756234,0.9955096806500273,0.670388653009064,0.34634165814497453,
 0.5385675105059486,0.9742466067233323,0.35360761578613853,
 0.19842122757364167,0.9799177712052572,0.9666058194696118,
 0.10503008843236139,0.3097355593573826,0.9575828705320504,
 0.45589872924994923,0.6266810077002207,0.11297258628425844,
 0.008455416726531051,0.7290223030723954,0.9342977781748049,
 0.6220419857463152,0.017184601801176425,0.9898718345368303,
 0.5905096769611733,0.8062074081884901,0.5996208955796756,
 0.047543005850568365,0.2748222674421362,0.1847182944345619,
 0.5862883982149044,0.1633283355822488,0.6889340001878033,0.8885759622585186,
 0.7631899765073592,0.5068436637035982,0.17835071459943996,
 0.26401301166404756,0.6450294648473683,0.26101891474634153,
 0.8990734888921694,0.42686439333474524,0.20829577444060676,
 0.0041481656241632026,0.19208424685780523,0.2913838553976864,
 0.7968880418708415,0.029517112969550285,0.47795019890211976,
 0.8514656016649435,0.33705045233035114,0.17447336438813343,0.934719647129168,
 0.7035138268214245,0.9988232367416158,0.7054826807742983,0.9558776466702393,
 0.7148289888201802,0.7435864933773638,0.8635265730867028,0.18657632529782608,
 0.04072064579396928,0.41249380727173324,0.16885832272198475,
 0.012215415120404582,0.33398069818478104,0.6073095130580425,
 0.4890427429575991,0.28799558707249795,0.41965375387423576,
 0.07711300327955317,0.8751534515593388,0.9211157822323685,0.3067315884257371,
 0.3047001759272374,0.30210455123862856,0.9617901759225654,0.2539047622765531,
 0.0863014135588509,0.8581737618793914,0.10973135676722179,0.7622499374680644,
 0.3808147317808086,0.8313579477508689,0.48729140563133744,
 0.027842382639514107,0.7270007072654469,0.77497466335821,0.5898036369437754,
 0.4415185786591658,0.8662734234130652,0.6971108313604552,0.48129405706266026,
 0.20765652844227223,0.775401772888822,0.9905673584018813,0.9537647211342235,
 0.2883002557794595,0.7510447423842673,0.3993754920719276,0.7877778037947136,
 0.16666201845631035,0.4412152590388535,0.9694074199791712,0.1819054751781637,
 0.4140798459020202,0.9671508900957625,0.29859122418841944,0.5387422896842633,
 0.3672415520220824,0.18993637507657013,0.09588177655416119,
 0.30725392953730424,0.9631946224056522,0.6195597357668339,0.7104310686554007,
 0.6350080296746105,0.4412924875938996,0.6978370911308531,0.28058138052216064,
 0.7160920551754545,0.4081491482761962,0.7493463910430965,0.9148094939613877,
 0.6223112295884503,0.9466546394140136,0.8122143789171167,0.5867871827822563,
 0.6029680162692114,0.958979480491778,0.9226397858299069,0.7887099011008472,
 0.9520483546509365,0.9168819300367719,0.09961394200937013,
 0.09931079064896575,0.8028381020613203,0.15725871727981655,0.993854757440828,
 0.7264099346052924,0.6559245177535113,0.2764990852428819,0.6592204026937102,
 0.6510226003699137,0.5512598770659262,0.22437232982585398,0.7007578759027188,
 0.3751838581408584,0.20099285732823347,0.7225393743404634,0.6571276246796051,
 0.20929601259319164,0.06469385057965371,0.121424704714305,
 0.018383899245262625,0.1689524040793463,0.2865462356459797,
 0.4308952801438155,0.8561890960827832,0.1438235892499684,0.14157078118310862,
 0.4730981038284423,0.6935653795030001,0.6882904798112865,0.8602706909591209,
 0.9916432797594958,0.33560083482356806,0.3166561185650062,
 0.17794540692704697,0.2888112725738271,0.4987139818115903,
 0.21239962597409356,0.3490902524251237,0.861263914384141,0.36309878410761043,
 0.27232878966478,0.7916161163908058,0.7193728152065687,0.23240389284768492,
 0.2118225488138643,0.6317208093101873,0.08302983571617872,0.5085234422862259,
 0.6435874199310676,0.5883457893700041,0.8285695156154428,0.26310986731425234,
 0.985976473235147,0.08752028317158367,0.17978911965039557,0.9199048705007719,
 0.8422683222652156,0.6045908084877386,0.7333541833784046,0.4501418887619353,
 0.061298426683349616,0.200728562356966,0.8197938078611665,0.8267173790313428,
 0.8411307524114001,0.5514413547057019,0.8751785742759087,0.0885602314399695,
 0.9714276671503351,0.457752648996389,0.5357717469748531,0.6368782461983163,
 0.6843577972140006,0.3033729147260604,0.05168370182052506,0.3645713684516768,
 0.13788343968588845,0.5748994693860968,0.003059937787702305,
 0.8881596263597589,0.7478861311421651,0.373395381220582,0.7252338328749987,
 0.6656563135509225,0.3114368396928764,0.9535798434779427,0.4821913388587683,
 0.6701085992891246,0.6481467306333093,0.5173247688366585,0.8901309557590544,
 0.02623348569009487,0.25941726115235697,0.29597840382394147,
 0.8691848383596599,0.5334109627307333,0.5640184357693139,0.3555345183082077,
 0.5847198676181915,0.8657746109570672,0.040444095383939116,
 0.9234411535070886,0.12137427540807255,0.023285647548461852,
 0.01906860074465988,0.7742007456994427,0.1712033347259987,0.8526115350777289,
 0.5254464416081917,0.3557564147314244,0.8559741738851642,0.7710429089716627,
 0.41975875241380534,0.4893375314514514,0.09071946042958112,
 0.7540031578367793,0.8984986448347708,0.01912968579618679,0.8084959038210748,
 0.1450235020017715,0.8277673340391374,0.12163569385346262,0.6775077559357161,
 0.8480771639331312,0.6687075856425729,0.6301178320573192,0.7813377744483044,
 0.3874986366583779,0.9657605952597651,0.8566647398130959,0.22491673291485026,
 0.637194322104095,0.16890713977026994,0.871619428929035,0.9804450619621982,
 0.3736672682968958,0.21619897347048833,0.18796997592791198,
 0.27839675304582834,0.3640010897056549,0.24442158546944204,
 0.7221334014118491,0.4648665762790931,0.9825300143351043,0.13383714425516446,
 0.39395232017558934,0.4738265616569306,0.5323158549307805,0.9977461261511289,
 0.8386403654814393,0.7480097251070189,0.8925696776136984,0.7499382821014833,
 0.6301383933235459,0.060898997539099664,0.1641956211122828,
 0.32445515835120153,0.8516692937453655,0.2609842335277841,0.8008335238283792,
 0.5679864995834693,0.9876432179711638,0.350894894351792,0.5069380273770079,
 0.2551045373867953,0.11910768784324088,0.07281880093185311,
 0.7856404349290355,0.9132101532959828,0.1058741399122104,0.8255375147368488,
 0.10725866801054607,0.11847872813084392,0.5175310394944658,
 0.34569270681444597,0.4919401669754404,0.7351237452745543,0.5172709557797499,
 0.7976152574189245,0.14793340198792504,0.753405004767707,0.5547646251551955,
 0.6596543676715053,0.486978830600695,0.00903933594064854,0.6605525607285688,
 0.5081401504255924,0.19840428370080854,0.5749194135858253,0.9720222716470052,
 0.16062621681353062,0.08146541293794118,0.9237259280274948,
 0.6344574106064943,0.9242068667197767,0.9869698400688399,0.7103335532194679,
 0.5664048363670824,0.8363656000583426,0.20761630030026912,0.8970018717268371,
 0.7921924053751188,0.016113672432494175,0.5834880152334002,
 0.03466735471052529,0.6853736183528593,0.18420845650295747,
 0.10650025884814784,0.6119065974450053,0.4307311949667151,0.8484047407529977,
 0.9632257913314166,0.15553199147036267,0.660448022589612,0.6411661467278826,
 0.6741096047722424,0.6906410071586083,0.9070488680612951,0.885948457971614,
 0.5396452349105024,0.7585991593878828,0.20995752601397788,0.3982970334745559,
 0.6233234057967193,0.9287302886981117,0.30113184424051265,0.4193324445030967,
 0.7951858453759576,0.5173000952190976,0.4384237317264298,0.43943150890060967,
 0.04921617584499183,0.10545970230957646,0.6352908622970395,
 0.35096610384364557,0.09121971537976648,0.692808975710949,0.7356808570782406,
 0.7477916803105221,0.7064776031270811,0.38144477085528716,0.8885855185998801,
 0.27247676062817516,0.8152124015579697,0.7333369882287303,
 0.06788831665229245,0.7607333017837762,0.2722017979889967,0.8880338000676512,
 0.9850327123079585,0.4444753442307552,0.09825683358870774,0.673876929735787,
 0.3179539281384639,0.22011819714237435,0.6008715987264066,0.4375982772223106,
 0.8933038381004922,0.28060834259733625,0.9727357677853217,0.2856556294795769,
 0.33210397294673066,0.37925153201017603,0.6323842385787157,
 0.5830016323996606,0.5542129944153352,0.4923222880250011,
 0.037508088078501034,0.4324223096444848,0.5632767270083103,
 0.4639428308000759,0.9475318437152518,0.02693168071741825,0.961646058591314,
 0.8855042410459715,0.03392083194262974,0.6451748235002531,0.8340317940383578,
 0.8156909898387267,0.4339020129203838,0.8759783328892453,0.2911616912753632,
 0.24807345681839743,0.49519251752857874,0.3546484536576404,
 0.2632331065106005,0.8971769095938404,0.018871504728550037,
 0.39991328147934724,0.9698116365193418,0.2565371978323304,0.7568432193640755,
 0.9323393646499676,0.9972873228933187,0.25125937376847607,0.5993186384216658,
 0.33014886793577713,0.42997096913455457,0.04055925759654089,
 0.37693802656869335,0.09545409068390365,0.42374827981954855,
 0.6915345699085751,0.764115691805414,0.17626710660862843,0.5914190247683994,
 0.053002280535058066,0.6378970781217986,0.9467220353087495,
 0.4589809022754344,0.44391881940156463,0.49652259029528634,
 0.8771176818302217,0.08993103307680272,0.7263067489627354,0.8992310705173872,
 0.6211530655172933,0.9148359423860519,0.9540426876439585,0.5733899425899649,
 0.7887780194221816,0.21538695271172525,0.6642002700101672,0.5307942590285976,
 0.3732957031302766,0.04674265505611608,0.16392965973437867,0.421941362794931,
 0.48422474921928393,0.3003650058819739,0.35505686263507763,
 0.2934773354053264,0.628923297653038,0.23421417998909688,0.9331254224337433,
 0.5041098517660234,0.48748791158500504,0.38763705756140976,0.952766613789605,
 0.13569259745263407,0.13794273624428843,0.0813801224586197,
 0.5419643503079543,0.5560894498272754,0.8765638898377733,0.22936328723759014,
 0.3849397147669288,0.6544377199585603,0.8117874542523302,0.8178103219826599,
 0.832027646554689,0.7771650311296456,0.7951726840691999,0.25320423164445194,
 0.4516750508928833,0.7394089029952078,0.403236693505608,0.8986512407592916,
 0.5867532348885519,0.9063062437041226,0.5656194728608099,0.9747793780766482,
 0.40500929854082757,0.043650077093866035,0.5798032707399041,
 0.07585554714733056,0.4349646243534161,0.07250953887947664,
 0.7686189015229655,0.08174551798977658,0.277686534171543,0.753296808985568,
 0.4306908547415006,0.5847244048902925,0.6555710922562897,0.20422528281501418,
 0.4408355316562209,0.6982769018933864,0.8156599373028992,0.24226524230641566,
 0.6268111266191978,0.5454848631601507,0.9952704413775421,0.9418144342618212,
 0.4177081092923225,0.6397439414124202,0.17512621982853005,0.6903581026412241,
 0.9379062665081617,0.21350933768379865,0.18242865683198617,
 0.005827394735226532,0.4481264057160549,0.11206567790942323,
 0.7640267218487405,0.680904026845197,0.6588785470397444,0.46847441849877247,
 0.78873625223144,0.2830714424116151,0.9065266942124303,0.6162153889351156,
 0.1857293128665759,0.9000806376538426,0.7627880058481415,0.02051756688750761,
 0.7781782735973685,0.2963570341076718,0.02051580645329265,
 0.18037526632502932,0.9753730967124296,0.9549244099902112,0.5303686997353907,
 0.055629899464036625,0.42394104451660986,0.16093179888295617,
 0.3277507304071643,0.17708430003772313,0.40473803955376997,
 0.9464531455655073,0.11505049025811731,0.9108808762925236,
 0.44149771348536404,0.02957479385746309,0.22566596685341755,
 0.6260107636991941,0.41605750768102023,0.9429651235259487,0.3898619241353305,
 0.912388157051464,0.39829074428253275,0.5035769974932084,0.9747865580852539,
 0.547565905535622,0.2446844345311605,0.564672228115476,0.8275975052660933,
 0.02745448733811917,0.7580906688355924,0.787088832091354,0.22398677507790676,
 0.5614901399681682,0.07917165204479082,0.26319015629397025,
 0.026472763657521825,0.41789868943583197,0.29410586651825554,
 0.25125927601246134,0.18002192819423946,0.201335729307213,0.4272487973901794,
 0.8533499882863543,0.05177583587572043,0.5280124430736409,
 0.36703579437707146,0.44321729063954374,0.8561102833364075,
 0.34859217255037644,0.17274654752493612,0.5968529239429321,0.781723433208174,
 0.6496686020505308,0.2282229741559696,0.4546227529619369,0.14634849228024027,
 0.5387173847824139,0.4020165547867155,0.47929634235467,0.9735147996348971,
 0.6185263893852171,0.850086261893826,0.3935558883770336,0.9572886238898664,
 0.9324078845202923,0.7526623475346429,0.8743735599216543,0.9769208117544907,
 0.5134839905710057,0.2106775650172159,0.30947924200199983,0.3012110119716839,
 0.2978797221497762,0.9127424321878372,0.4304323602782373,0.5782526341944606,
 0.7031003021225698,0.10225296178023713,0.040418852072862665,
 0.9603231511489538,0.7849840864966638,0.173963322610306,0.8432293887729683,
 0.3684243568731408,0.9199789195709427,0.6548237524870993,0.7415132195694187,
 0.33999454796944184,0.707858092357222,0.9555879833684044,0.553858185195343,
 0.17660462254912757,0.24463249718763214,0.40538604857789273,
 0.3433732402601892,0.1375834451547292,0.5652281212307716,0.25213446676972673,
 0.18813486022017378,0.32372579487854314,0.31976053228583545,
 0.3568231807411243,0.4158645109585706,0.13585203061424422,0.7702265065309745,
 0.8062543750797567,0.8389313036402033,0.5635193346249369,0.34129169106073987,
 0.8764845642397642,0.4750283550652723,0.39701607219249113,
 0.13355590232504966,0.8075699924673853,0.1685834232933925,
 0.06950165606431147,0.321832774336539,0.2510660051139121,0.5292891645369315,
 0.43043613029990546,0.2544454793231472,0.6888250442182476,0.5581922557259831,
 0.6117333624387244,0.14168909656695095,0.42695684846445403,
 0.17241237076775362,0.8648419391288866,0.7307195055962925,0.3899965227815694,
 0.9435308363002031,0.039088402401734834,0.8842218298525474,
 0.1672977254415582,0.22000515334223691,0.4800647244922932,0.4270397943524804,
 0.7796769092666681,0.3587025577865325,0.896684712946186,0.6116118248573668,
 0.4903190094389649,0.8224681567029167,0.13822225098707164,0.8840010914271862,
 0.31665079079175884,0.1639573163345045,0.5628420614019252,0.896570932748961,
 0.9877706119484643,0.30467523407716957,0.9446625726260788,0.894469343290747,
 0.34643614204223405,0.145456254508593,0.10261819073979506,0.9845115041995444,
 0.3321201726062787,0.10386229757675347,0.9292989962203381,0.8232727412554988,
 0.7244995029052133,0.8030144247774703,0.6701939839259432,0.1627021917469662,
 0.02489067022282443,0.16378278099075227,0.3209811984547384,
 0.5304597026934406,0.7174809307812878,0.796928600664889,0.3995668930288665,
 0.5374306037224307,0.6070248480309054,0.25793774292196314,0.9529411890053363,
 0.8404239024790843,0.7951497894904899,0.06376715676316491,0.9307742458684297,
 0.4556812851754064,0.876929231163021,0.06567920839324448,0.8832097876243343,
 0.5631668228003706,0.802138529709026,0.6740907650031603,0.5160231637983717,
 0.9837672426186564,0.7262387233785483,0.36152632340232993,0.8624124355821478,
 0.399496153635607,0.6899830066328442,0.020797219978049508,0.9274045102692443,
 0.046706733997034844,0.361787285295342,0.38506001933610434,
 0.21765452828362042,0.4107201963982907,0.568605131782742,0.5144779811877441,
 0.24806197921621698,0.44142679203589485,0.3572388640393763,
 0.7828212203788986,0.024580375906724683,0.2918678064060698,
 0.5914323307987723,0.7650957165475715,0.5782015321123412,0.02758326946661893,
 0.32847706423673984,0.8894121166068718,0.26343320536736003,
 0.11747261770944706,0.37305160303291696,0.34587667996139726,
 0.5369560893975334,0.8283630625554801,0.10502720793906706,
 0.12280338807944624,0.42310553309325993,0.053104878204512374,
 0.5568868785195631,0.7865078041464189,0.2541370901452339,0.32810130775863844,
 0.4764802189885551,0.8822454440341007,0.26537941092840334,0.8815988811064583,
 0.01532129644969138,0.6745474837856484,0.42944709261657,0.40741400207821976,
 0.5918941101754214,0.5861099369355465,0.201108647122191,0.4349357562398678,
 0.3582171338160274,0.949466358557268,0.2868399642621783,0.6269999207344991,
 0.7339004048037389,0.16952636138760535,0.13529202776569726,
 0.33411095903739696,0.6413171794976487,0.1323892806101492,0.9396412464505025,
 0.34291814601302417,0.3086270116214771,0.3737715694467706,0.8756123042341539,
 0.7114557114439014,0.3590163430912161,0.27355094457734985,0.8723745673342825,
 0.11264080919877861,0.24804441907555042,0.6378567099845923,
 0.32302048170239916,0.6466239824192773,0.8887852948995992,
 0.33327307680253604,0.44693037431516236,0.5225923706739106,
 0.4103014636272282,0.6520311427778964,0.46111683817938864,
 0.16337855603104012,0.7410315166449992,0.31304832393063786,
 0.23492727572793692,0.9511200942868742,0.6534143853401998,
 0.28198107389368865,0.5455326081012075,0.5596685369189472,0.0813682138547579,
 0.9210401792808376,0.940578553296988,0.46423926087889544,0.02396463673657334,
 0.062338347624740664,0.9673473741051811,0.5718232653062663,
 0.28725810204276037,0.6302004744939985,0.6204806545409531,0.4872556101682949,
 0.7042770188225228,0.04894886954713684,0.0931084435653653,0.4399190213261339,
 0.3905071591089533,0.5876652223263391,0.7479525829325978,0.5295978691835851,
 0.969787139050775,0.7513546304586062,0.11217156731249911,0.04424470282640469,
 0.8979833325038262,0.5129961147916607,0.4274616762773511,0.7470272370549758,
 0.9434596104282288,0.5350654886642267,0.23623254298982044,0.8421449001926489,
 0.7503116097808544,0.500902417798577,0.8085669551601737,0.4371366359310238,
 0.29919420228843474,0.40296757137385675,9.467273647141283e-4,
 0.727285683256412,0.579459453285761,0.7288005864159012,0.6554651544208456,
 0.9956900288763629,0.8135157188307917,0.1703298358933817,0.47712532618382886,
 0.984749517684559,0.8587978062828144,0.7949577306678886,0.49629870070275217,
 0.5107499438916467,0.3988135183584973,0.9197842329164584,0.5121046539492836,
 0.32453256827162025,0.5905772279961378,0.20709313607610602,
 0.5242811612325431,0.8944840655450994,0.4110935624821701,0.27670707733618716,
 0.7572745934527711,0.15271417879186444,0.09531786375660423,
 0.6803195919637036,0.05520714049142139,0.7037496207758202,0.6186101944121403,
 0.29537094312566725,0.48812723339076014,0.7315194433593178,
 0.8339651179775198,0.5852348839163011,0.15117874424881983,0.8774838884581646,
 0.31360563114643103,0.34129166727636284,0.19151321780266128,
 0.6063428483937281,0.187336167051113,0.7824548486573721,0.9462978813330226,
 0.12411214680975546,0.6845771348327588,0.7152924713016412,0.655174216740309,
 0.5346301805160333,0.3402879111852527,0.2120326950153293,0.08605665728950329,
 0.03934891048506506,0.4026061423864431,0.7953413654589636,0.9134444944254173,
 0.5928680357495272,0.6257179765834343,0.5601810601764705,0.6040657519114296,
 0.3577964632245212,0.041963059177318573,0.9796433353369753,
 0.9645698316028304,0.6421984153179853,0.23733888798403502,0.1593993641454614,
 0.7280620697832425,0.3557227869274864,0.10719925040621825,0.5666872021620972,
 0.5775948303234819,0.8531118112167506,0.9555833335932863,0.12078083492802016,
 0.08212012769800192,0.47105965423442053,0.5479916749684193,
 0.9524038046590421,0.818058807677281,0.31425516862795533,0.14289236838203,
 0.10651682884252567,0.6771226465364342,0.4331841828949605,0.972901588814037,
 0.3811774592688323,0.48260892780171094,0.11101869776735795,
 0.8707614982841423,0.627839093336949,0.12796013565683428,
 0.024297400180857576,0.3085106772267219,0.6723014552508719,
 0.1887598978758659,0.26826422143875495,0.2726215019172733,0.5764967072845345,
 0.5953573963001186,0.7749063813009626,0.4990389561184132,
 0.021390155072132155,0.6904253164242121,0.7311555071761867,
 0.9963808498589105,0.10242801144373215,0.6958011581688949,0.8401658237439656,
 0.9957015728120242,0.5608857143173486,0.9789105753361081,0.6240770726027182,
 0.32339489806173105,0.5359620198060684,0.5255217500425811,
 0.45802891329954565,0.6639604109035067,0.5032596372470922,0.8702603549205308,
 0.6608306490727893,0.6866959347012452,0.5456898731470576,0.5129427575240956,
 0.755819011075987,0.7959568787177452,0.008800744726458642,0.8135921466373437,
 0.23953604878787949,0.950763279146259,0.9180145983500717,0.7406929377517714,
 0.6073914944657642,0.3715065551670067,0.05837900189452416,0.6619766295411396,
 0.419109399710232,0.6585691746951899,0.627426652516569,0.055090128146686124,
 0.7785566106749997,0.5235552870837552,0.2007266061597458,0.4681719947721731,
 0.6336125598511736,0.7658425812926459,0.2678538737469428,0.7317109834491016,
 0.11277512848260662,0.42134215066701,0.6990191950659714,0.17545905155188923,
 0.09953519014414292,0.523866038294095,0.7626946157797999,0.09931035333502658,
 0.8477119530161842,0.6543267035715061,0.1966094266190529,0.0851613231074676,
 0.22557401902392216,0.8779065994206687,0.4802346038157297,0.9685328070460076,
 0.46464668369044004,0.6602042048565951,0.026567515297432287,
 0.5976624709181924,0.15465913224856565,0.81789370417828,0.10174669706458705,
 0.23413236473900623,0.20799599751734876,0.9792852037046174,
 0.036840710550702926,0.7650786630234908,0.9434467707564006,
 0.3140669750318107,0.30130431915144096,0.37427660673199026,
 0.6112778024768888,0.8042947441025686,0.059885587287466846,0.378002175052067,
 0.3653705326598178,0.6254405303982515,0.8461667364007932,0.43720431149106587,
 0.9318150365206653,0.9114224706453374,0.6814097186587433,0.9691627286452966,
 0.6216207721642134,0.14574731978103206,0.41520745393280034,
 0.11178229498325287,0.6516158896596193,0.6935869789904935,
 0.40122134328788506,0.5623033096909724,0.5598013489407377,
 0.44920611853030457,0.30415899503416544,0.7140499074343735,
 0.7863926395040809,0.7067087645100756,0.4587902398903023,0.7604121457559915,
 0.8424608244638601,0.9024599326304024,0.7599984259365088,0.9876941572119983,
 0.6316587476775243,0.17262482024644465,0.09879928453109299,
 0.5259943110035757,0.9780081530909716,0.8048455062659555,0.9683535020829177,
 0.17309526835222577,0.9828411659575651,0.6619457490815557,0.1401501845769444,
 0.23696265979790565,0.7434729791937145,0.8665107108488106,0.7708174653668508,
 0.12174446167504416,0.09234381205593278,0.40752578834704767,
 0.023167473855172482,0.9180087722909301,0.17468601692196906,
 0.9274781276191941,0.8070779250600408,0.7752540814579658,0.42425406383231323,
 0.6806390466130348,0.5250113347061057,0.6329688175121326,0.12648051914058756,
 0.5882448717216122,0.0028777808153743756,0.4791216310701658,
 0.7883106880798614,0.30957474050862577,0.7000389116794842,0.9007285996511102,
 0.42811429871577655,0.45031717310272557,0.7825171009021705,
 0.7009070910705841,0.789816136046945,0.5920911220463029,0.2750262762352499,
 0.6076954382621064,0.6455007220523781,0.7405567394625022,0.40664564554256777,
 0.3210564696229383,0.5262298378917679,0.964174143397234,0.2928179648397564,
 0.3533137357827497,0.44682609950516006,0.675606656452838,0.6398363129015475,
 0.5730817724009566,0.885305798543241,0.6369125867661927,0.9431891224236018,
 0.5812325323463846,0.7383919246538343,0.5655518233918125,0.06924053792158369,
 0.9356752545681097,0.48044973622607956,0.5822545579894347,0.9842517064539315,
 0.21422814066139395,0.3508975537134813,0.21069441098605346,
 0.40943061432210137,0.7527206386755496,0.39933255096107934,
 0.009670739104514436,0.2360227822116988,0.2000232926675225,
 0.5233451485982136,0.21706284856961233,0.2877954668068591,0.8075359474862785,
 0.9138963353643674,0.5730477996341325,0.22364529369071884,
 0.05802336333714575,0.1862781762937521,0.10635703569406274,
 0.38844507515349824,0.46370837824990474,0.6040556084891162,
 0.7029435773634974,0.5907342453695719,0.42789912863671087,0.6650451768900786,
 0.9252259709703674,0.6146060347768959,0.20219111412018986,0.522903288981949,
 0.20283341060121918,0.674689620437477,0.8381362867127774,
 0.059333486602231145,0.6852460841670196,0.4059351355644836,
 0.39779224966737037,0.5725730399290199,0.4804059589004184,
 0.31401274873345975,0.19650446939204858,0.26270648883797676,
 0.9101044326307971,0.17146730845809033,0.09998810968639849,
 0.9099605794249224,0.05848556099127744,0.35191663970997467,
 0.9308858623968943,0.06406926100028054,0.08390722912529225,
 0.7669531594600811,0.8520260602505751,0.6362658179355072,0.24029395112061114,
 0.05533879341041026,0.4580157486478036,0.3684714718189057,
 0.16624884637426818,0.08345990488479949,0.1541830315143773,
 0.6013349471836086,0.27409125626315345,0.925537336100406,0.15840256967655297,
 0.6378935841682996,0.679762821798329,0.07683732677412591,
 0.019015240894312835,0.4984776094539395,0.5169965737184362,
 0.054869926608750896,0.39872099228246705,0.4211297363763742,
 0.391763882768025,0.45315561070398347,0.5897577744247131,0.3800044726257905,
 0.46129002211300885,0.6669426528363871,0.4083729168078033,
 0.20072521098192353,0.30459045834325793,0.33925438489531656,
 0.05355417188047307,0.9476303775927244,0.09987863321203683,
 0.7238326228626474,0.20847752951520915,0.07530192651093515,0.486041425952238,
 0.00394367546416885,0.010500621489263118,0.991717993350099,
 0.31270052075026444,0.6121186581284989,0.9651775702546367,0.9071849232015574,
 0.4616834363327966,0.8576005001607843,0.6235906473331068,0.9193691978142451,
 0.9292164995420928,0.12302908610499586,0.19741499920191308,
 0.2507786279313312,0.9205849466249019,0.07683549675299073,0.7592675685819644,
 0.2629873209274284,0.6867440000279881,0.8412398496374286,0.724270584355269,
 0.6513522839392492,0.5705527858797332,0.2636407961068248,0.32744997894370453,
 0.523880541737049,0.03757218312984412,0.01891969650994685,
 0.11643170360301669,0.5945451460835673,0.30522301629713167,
 0.14272288075903516,0.10405928348583138,0.9529934356802102,
 0.7018468114302042,0.3027931034790675,0.5135106089632385,0.6943393822126964,
 0.5004134511388107,0.5740203792396146,0.71670939660199,0.43495627816695814,
 0.7039930564619132,0.4358715266442602,0.5706177883391226,0.5046855273812468,
 0.39255477838813224,0.49267026117786095,0.345433564641099,0.9052531717327642,
 0.13911964531867704,0.10147205760465561,0.6654943843961336,
 0.9563481207202101,0.4136795800955281,0.6851025730159349,0.6079516856974291,
 0.4569731987721789,0.5892718197657498,0.8599591561166676,0.8704420026708417,
 0.643158417347475,0.0858513060442535,0.24650659872657532,0.3605785953120323,
 0.6610896289598182,0.08414194800797414,0.4435369132863939,0.8067626525810127,
 0.8698262899356904,0.9534273309182077,0.009093943080834888,
 0.18811473224397313,0.7279208000018816,0.30912007782794826,
 0.46238128375562493,0.02982902528850917,0.27549463588241985,
 0.7745328087333165,0.7170687747322794,0.09088978307202966,0.657115665545833,
 0.16015722692242595,0.6499302592364923,0.6925286593932778,
 0.29599235509680966,0.8563159752897574,0.4209287149448585,
 0.18018864009317348,0.6364133155489314,0.9618004551952495,0.3371231396788006,
 0.8278555930063982,0.10547728736449469,0.5292692283082112,0.309255449385751,
 0.5104874720740277,0.360687470590469,0.06927054248511522,0.7274967681987603,
 0.9144499540744329,0.9907480908357161,0.8182870191402012,0.2972927297319372,
 0.6513350937451919,0.04017658396086787,0.32906591029074084,
 0.8569406843025461,0.5493664125077116,0.9405021077044653,0.9419442486828865,
 0.3821364675735537,0.07480347548390753,0.14424302016267698,
 0.37436312684078565,0.44235333492879725,0.9682174321837407,
 0.40360435626279534,0.6356439814345793,0.7833214232902864,0.8217691977409477,
 0.3030832016754672,0.37453387252847914,0.7105540582692246,0.978842178006481,
 0.001586582571902917,0.09976887864252781,0.7313969407274138,
 0.17806083002735995,0.1354401560218904,0.7245027893721465,0.599527653582291,
 0.49167918902424146,0.3835927628565282,0.4952947272423628,0.3244514864713649,
 0.5143893368383732,0.8743053895570351,0.3223546270731975,0.48408421521777656,
 0.3834802667652516,0.8577246054412283,0.5619397951136883,0.8488600755935689,
 0.7474950299740065,0.271267479502331,0.3526018490595938,0.15056197966310025,
 0.5686375162056244,0.2790645646101212,0.28039912344829054,0.2208270155220391,
 0.13686409632086205,0.9819535682790546,0.5918962815262157,
 0.39234504310726415,0.652087090023748,0.6383971131992002,0.2594034725919253,
 0.1446203629797721,0.048465283847579754,0.6169567852225595,0.828368383859468,
 0.6376365757328594,0.34677884034173057,0.06942274427387995,
 0.008404085711474885,0.9637628834178624,0.9910986520476472,
 0.8172267190121322,0.5611209508361555,0.4602157437950132,0.25700290836091166,
 0.3911551420635766,0.40709274100105575,0.4038428732844779,0.8718044928378337,
 0.7369855087384116,0.923163716764283,0.8464900047848235,0.43278103183966354,
 0.41843307512664896,0.11156812611731526,0.9743076049607161,0.283840692840039,
 0.7988057211833737,0.1544488863708925,0.41512099525552393,
 0.44332640582318106,0.2618569598102206,0.7429197171954841,0.8081693852616716,
 0.42504212188606283,0.24282824814955017,0.9903699534996165,
 0.6332308561598222,0.087952805925953,0.5323275509836256,0.1951510749694798,
 0.19270938637094115,0.0657901069965302,0.28699391636059235,
 0.1553752464058138,0.4145492861934402,0.0011859611435613981,
 0.1867748809592239,0.20090809739230164,0.3392662936620934,0.3087733594911779,
 0.2563078205758048,0.14790612705066197,0.27672165396950466,
 0.6803797510435844,0.44188358432020414,0.8640655022221776,
 0.10832352232414955,0.3022556585155428,0.7229768179295354,0.8484136683432355,
 0.8207774888155555,0.5455066310413387,0.9900337183192243,0.6250940197389672,
 0.4399326964250946,0.20398808126011558,0.5823861201291611,0.4102645264549091,
 0.2762014311222504,0.7440628617013738,0.5008977376017955,0.5192538263568012,
 0.2428573422421918,0.9685293061635086,0.38903120734096863,0.8699233048658238,
 0.4418861275255187,0.1897410446721296,0.5377447066227979,0.6497917866159618,
 0.5830427954031672,0.9339104142522605,0.3981970892245976,0.22616489714824461,
 0.010774366708306005,0.13349248094285115,0.9521093922641257,
 0.049720017240567294,0.014624529639599482,0.02414828816661241,
 0.20043962335958976,0.21771822016571196,0.2961357370853114,
 0.9293476934451815,0.06820394309959021,0.8880023429292365,0.3331155575193908,
 0.6265554956552024,0.2948091965839037,0.16816025710244464,0.5859238391538947,
 0.8697962332294875,0.4979403235141743,0.4120042931621918,0.5783208511543555,
 0.8204171725022896,0.3809911008269439,0.1504409121501007,0.06055718391070419,
 0.6783928363922158,0.9434483409071581,0.6853928944243853,0.16210291788139142,
 0.5282834030267035,0.28659008962584187,0.5828452338501138,
 0.37667896427579883,0.6249514735029285,0.6230960292011828,0.5675425685176121,
 0.8171181899019855,0.697251655402953,0.42456384949844417,0.9768513679041477,
 0.4744655661125725,0.811866035609851,0.6116424579237475,0.9776672396846465,
 0.25675614714566586,0.3752818793508016,0.8904403583738848,
 0.23269806110493685,0.9990967137615135,0.24659038732000604,0.791435216929949,
 0.6233413861151692,0.1542637846453908,0.432252150866159,0.3441358411257259,
 0.15863298651211788,0.46847849783776796,0.6005562507762287,
 0.8161880690861536,0.5707595627717506,0.04830939599648587,
 0.09241614263105391,0.76735414494907,0.22813656779797997,0.08287177721819272,
 0.7772335286147694,0.5504149745099768,0.27272471133331033,0.6322550475896286,
 0.06193575756745706,0.5339235196034355,0.34362830356809526,
 0.027653237314905543,0.005254297083074566,0.8043787694710676,
 0.2935837679315988,0.3863426744928383,0.2703266957381878,0.8658341637616723,
 0.25791506946918485,0.370589397343291,0.3444714687728888,0.29874966318517016,
 0.9509110153635483,0.07769767053261678,0.5428962233218587,0.5948366797959472,
 0.3527735170298627,0.9813613695799688,0.8242807814715665,0.5874558355442909,
 0.06715686720726766,0.3886377133970449,0.13590721172937337,
 0.6995491769541967,0.5817583876721909,0.4935412301715919,0.7531609781983946,
 0.5260384696119806,0.8129516054398944,0.6800258499961246,0.8614795309291248,
 0.05375418218939243,0.21127830599521424,0.06714108718137879,
 0.6318144409516587,0.4763316669926957,0.7985124341503994,0.08334133626471507,
 0.6087186629001982,0.43847696985223605,0.6906387533942465,0.3744866174564382,
 0.3904315425588304,0.4646096869115284,0.2910332140401771,0.04326881809096661,
 0.5260853979808986,0.8206993085528856,0.8635613651185579,0.6310275119627904,
 0.23596764966069717,0.8821646493503721,0.09969033415141348,
 0.2352996685618437,0.7238053729613082,0.5167674912472551,0.5018155949346143,
 0.004482250147999189,0.214113903659789,0.96792192867206,0.8295586172023758,
 0.8578241665804291,0.6786904250397172,0.024109206351415113,
 0.5757466736196223,0.9556495166315352,0.3993302897688388,0.8235789458840272,
 0.20696138607264458,0.9965879658591725,0.4607526574482437,
 0.29938510502610916,0.6725211175229503,0.390761532522235,0.240337057590866,
 0.34183551120612565,0.31606944534636805,0.9612219273331061,
 0.9468274301399183,0.32859495750146306,0.7948599584628425,
 0.34852326805052625,0.326252177348887,0.8398866897118731,0.4517904497707397,
 0.6295287940385103,0.9199012150572573,0.20712779380249136,0.4486966147828023,
 0.30175934882472777,0.4281087109951287,0.6102325701257169,
 0.32646572643578375,0.23794742779818245,0.8100005105204351,
 0.6116850870726307,0.5415852828715758,0.02525949096328106,
 0.06285608647755325,0.2201454949171655,0.8157052690423515,0.8960004904541452,
 0.4263673097250934,0.728731386812838,0.10169483646512145,0.7433388572877149,
 0.9081082373991574,0.020017790356306908,0.22122561722720524,
 0.9830768694805206,0.7865675023362589,0.25733592463453514,0.6499157574905583,
 0.5476839867958937,0.5014089641434168,0.24933747114450167,0.1617776994090211,
 0.8616063886719423,0.036514115034038896,0.8829490257968358,
 0.03336804493889456,0.8464931376637683,0.24629870203997772,0.630233373657854,
 0.5213114093822899,0.4017215928007004,0.07257216024341462,0.5616407145013783,
 0.1931066920266562,0.15018555308805914,0.5083009982170559,0.2892536355487738,
 0.6656485415563513,0.5188245889621314,0.807242392732435,0.5228495829019636,
 0.6165140773837461,0.9324352268720775,0.7352224360700417,0.6953846828744199,
 0.48720556689441374,0.36183458857497364,0.23658586776445134,
 0.7979501109958796,0.04928066743656312,0.4722111699399023,0.7175443238365563,
 0.16970512474825938,0.1794896056631513,0.28352756266764423,
 0.07227816286702404,0.34627525724149166,0.753562383978097,0.7265519981758302,
 0.730949849227617,0.8330812601420217,0.18024763891733508,0.05488883510477771,
 0.4112327145136453,0.8627355072392702,0.5095982970625507,0.3076526113649125,
 0.7233769367640911,0.8236017848752516,0.9538750818637495,0.6432163853409983,
 0.16041903746221386,0.40755241694894473,0.48028484014641204,
 0.5026724848664426,0.12538133366821747,0.6540398661846253,
 0.07798164006778219,0.03736976642326584,0.9532098914202497,0.755684882886481,
 0.13815630410180813,0.46010497530174477,0.5181649317654815,
 0.16421815473358037,0.3798619437449512,0.07582768434838605,
 0.3622458277971957,0.6373937995034947,0.8663966923900759,0.09100304457329489,
 0.22586872862751128,2.2291441479471708e-5,0.09352477328445685,
 0.4602247670115275,0.7110015240163605,0.22944184951986069,0.3579970386079022,
 0.6377916964519154,0.6291151660560051,0.32026939385110065,0.6744505324658587,
 0.3434898062473999,0.48199228165180796,0.1707089054055566,0.4240556062625067,
 0.634011541597649,0.7947135758278953,0.8445870185868619,0.7582156513220524,
 0.5089974495416082,0.24655623750975675,0.9493766080597955,
 0.053650758427181144,0.20524236793580042,0.24250526729231725,
 0.24841707879288033,0.12025752007909452,0.7193008205680959,
 0.0870453734409119,0.6222208527305783,0.8047763641466802,0.10133060073158773,
 0.4098966093753432,0.8471645260036841,0.8354909605332601,0.9011293942390425,
 0.14076938903655245,0.9305835755519941,0.5642853364821183,0.8408737098761109,
 0.35451056148978366,0.5783391576989818,0.8167506893345741,
 0.36612393026211243,0.736130633151764,0.3908895178986628,0.3423491778808505,
 0.8516954953834162,0.9150109423215291,0.3399498082818836,0.9520905003637578,
 0.6225196330405974,0.11659958897174205,0.8111682624322449,
 0.11553613443372401,0.7734814725489156,0.7106555704694469,0.9492355132155701,
 0.18588396051456724,0.8386452918565496,0.5209621467748745,
 0.10088697823505721,0.37117608158956794,0.7942465567120242,
 0.7027845205532065,0.8757713590146554,0.3792149325666323,0.3132034484493349,
 0.03746204132705222,0.3605111745511531,0.8463044050424424,0.2128497563755174,
 0.8860982655650138,0.36421878466381485,0.3430596887377666,
 0.014000405194196985,5.470053888310122e-4,0.07227888038162722,
 0.8125832569908902,0.7113635119789905,0.4170510346119063,0.08697000295283397,
 0.543629187270584,0.5606787935666218,0.41867468548553044,0.4693175489113428,
 0.5084455361072182,0.6247249230393012,0.42293107152555465,
 0.06974487552773845,0.430024668245116,0.4891789151885564,0.19484608197762565,
 0.849451281557108,0.8016968228723627,0.18995879267161342,0.10164327631826531,
 0.3324320734477142,0.46896676653669656,0.9840012524462556,0.682689736802609,
 0.4194936464723723,0.15683427361468782,0.2633145868028369,
 0.09256555536401534,0.2816498119217572,0.743075176542132,0.8494325037394864,
 0.5031606776775166,0.397284324168222,0.005464594159270586,
 0.16171985131794464,0.6114927500084895,0.5844689886251023,0.6410723474572405,
 0.10613927721520688,0.8610907257466022,0.8472710488879784,0.7847260005189937,
 0.6443261846294784,0.9943437476059828,0.3273338902918468,0.3533175940905925,
 0.9241183859035175,0.7157219268715675,0.513067805543691,0.6087979701343886,
 0.431728979597894,0.953665125397976,0.19142492541684153,0.8058976798083479,
 0.4103158890552796,0.9207787044275252,0.47514452198848633,0.6028289265598965,
 0.5714477261120754,0.4017246388601081,0.7607747365460092,0.5920315019581939,
 0.981656171186162,0.1382226321184299,0.7077568810702257,0.4046923376446363,
 0.4338649055045414,0.3348623409536593,0.9306407933487573,0.09312751606025615,
 0.35752329948751915,0.30958734484832917,0.3865540090228907,
 0.4895291805321007,0.8918123735119369,0.31519350932274515,0.4187462585047065,
 0.24704329634375377,0.8209171511649084,0.6790087678673412,0.5554323162592462,
 0.8887155555851826,0.25066205277723963,0.7031293305812328,0.4211143709726224,
 0.2135521718230069,0.6123534987566125,0.11279573307322366,0.206835788067271,
 0.1368165834286783,0.6502262218728262,0.7904456003010065,0.9594646175011619,
 0.21563021719900766,0.9876515212551157,0.9637971921195195,0.9828054962741706,
 0.576070032851501,0.437846444375934,0.27570121503296197,0.04410906730101649,
 0.047905904077616746,0.016333967989758946,0.8540998704204011,
 0.42892488547815744,0.04221478934113887,0.7631161660458983,0.976537115503376,
 0.7000026417836572,0.8896361717584934,0.059679343911238636,
 0.39330175653534194,0.2179992633307335,0.9509315556515293,0.6648292127091451,
 0.24881586653918086,0.7049520969326482,0.3743398254595447,
 0.10586076263998878,0.9024308169818552,0.6617471848700203,0.6014066175876822,
 0.004691999248795153,0.059258799867349166,0.4986888077387771,
 0.009570089883056299,0.599947122271744,0.5873980247974917,
 0.24289992108161673,0.3072536396637183,0.41490858243376194,
 0.33217048537083016,0.1420626222683825,0.47854331485397295,
 0.6445766849240038,0.33837932412151983,0.020654154009215553,
 0.6231391469665718,0.21841280591090828,0.2426094560228973,
 0.24444182480201548,0.31201864144431934,0.9282969519044136,
 0.5647104109153803,0.351649776194656,0.28948429274688514,0.361228380211287,
 0.6886871546530516,0.14495714885219235,0.520413832587912,0.35759386496391365,
 0.6349529827605925,0.4893859289275547,0.5558339619782682,0.32251130925060423,
 0.7867410176131228,0.528796217883552,0.045431907436841557,0.5321609000951923,
 0.9878150579637192,0.0588726762951326,0.19606515344418318,0.789302554943089,
 0.03509582164094982,0.06752595359657587,0.7360292228718349,0.761779439155636,
 0.45116911024429074,0.7076881456532725,0.4560652240512352,0.8638188932674927,
 0.2391214214647981,0.9998760251682672,0.060340222553908585,
 0.8520017454247504,0.9454110882405627,0.708247262520236,0.8383979534810126,
 0.9545413984117049,0.6081557233863001,0.6540763568661098,0.2695466476855264,
 0.7290047072586778,0.5201962773633764,0.4396976913088424,0.41949817058413497,
 0.049673440320913365,0.3979314983811816,0.5643427689113305,
 0.41284348834839324,0.7873346490030322,0.9866121817538807,0.8854098686813725,
 0.4347442188639983,0.5783999190329638,0.31321456573264206,0.3493480585627641,
 0.9719934406859193,0.7147579688611101,0.7603706081822329,0.9077262109077072,
 0.6171339314674992,0.21431873180523564,0.09320997626480831,
 0.5671728394236668,0.822140849388703,0.3046478226840408,0.6112941220712319,
 0.041581083411802444,0.40794325739799,0.6439844431171025,0.7499312819762695,
 0.46658573594071073,0.11955146645324966,0.316622363678533,
 0.28823355020289976,0.8581813899203401,0.5247177092942655,0.4217387495425059,
 0.12304723298067843,0.8401571683039986,0.2552096597202418,0.6079003267710538,
 0.03248450176594875,0.5455993557070894,0.5996335474391165,0.9212507710130703,
 0.0038418167518057267,0.07655280854860458,0.33615358108917126,
 0.5781090083766688,0.6819109080855631,0.0747662873278595,0.1702148294351542,
 0.3462348590952543,0.7462794257563422,0.4591114752623827,0.3022908147466695,
 0.6720005092523769,0.9353675594273745,0.15423798190041893,0.702882684614793,
 0.44268682657686276,0.0755209156417942,0.33649764382033154,
 0.15479258350147673,0.16741259306020662,0.9900881306551321,
 0.20207912494959912,0.3223900469602833,0.28191064316803105,
 0.9338351377880455,0.10140280190733275,0.4730406906481375,
 0.16014080265130615,0.7055114182836496,0.05623506098487918,
 0.31651341591052606,0.5996663645674303,0.08699887692825714,
 0.3022215430514724,0.19487329005150378,0.07824362808965901,
 0.20657713177772508,0.8965115582098968,0.6876490087486323,0.8618244829704133,
 0.8738862668360846,0.8484489939649948,0.5263510558324263,
 0.011933475092113843,0.7713960028421478,0.6879499521341873,
 0.5039146197067285,0.7774739345958697,0.6891933493558638,0.7796585506165143,
 0.6236463513680519,0.9460763396627035,0.32233046577041735,
 0.40592664127104305,0.6688939814204992,0.7267718973735421,0.6829113620353123,
 0.0858938027668148,0.6265924311238619,0.7522348115237049,0.8708498839345556,
 0.9012452349116016,0.9449467017421592,0.6877609488223954,0.3066289908797679,
 0.45507093025847856,0.22603438056480607,0.7828329473364704,
 0.07793516407265144,0.5899478677421937,0.313567183483597,
 0.016333123513572745,0.6368791207164151,0.39973429836708396,
 0.06839681532158903,0.3891849913335925,0.938408876787381,0.45015897289643503,
 0.36892937194979114,0.3884541371138145,0.38562157491010873,
 0.10481652926893381,0.43790449827809497,0.5939240385600033,
 0.35326607440926183,0.12297420080404786,0.09289688198840074,
 0.9413035429162837,0.1332280707165716,0.9959394215230206,0.4610914334026698,
 0.123483467093217,0.698088330270072,0.8074687148223891,0.7008275709141013,
 0.7044990013547843,0.8237596553304108,0.34120728234924624,0.5722346479440309,
 0.5242895810549356,0.5054540254113495,0.48357370790617393,0.9279015945822914,
 0.9788104493826822,0.32755898809610895,0.585822998146749,0.9095160391856134,
 0.8757249023382944,0.9175293664695215,0.52349859290194,0.7798406571584825,
 0.3359486743203651,0.022056541291224807,0.8979064887951853,
 0.5101693307993663,0.7864333094040972,0.8204849638401059,
 0.030684944867957142,0.752912985898363,0.7227545712040466,0.500733112496285,
 0.8847015076594993,0.029028833775169405,0.2910348105608954,
 0.3904341106521896,0.21973471729119953,0.976657480627356,0.7922425642166523,
 0.430061803865329,0.3812739489119912,0.5736917397833339,0.1840713598022693,
 0.297528301582511,0.5327914167510376,0.32178449951185517,0.05232781415922405,
 0.07286034614946013,0.6911740944942415,0.32619779989751785,
 0.08706176920834308,0.441939229386633,0.36219932812660016,
 0.48886383605361594,0.9345585597869385,0.1760699493028871,0.8637810593289492,
 0.6968316023566054,0.6705392020483992,0.9484185149398274,0.29035666433509655,
 0.6797009320626335,0.9300452474517354,0.7269852794428113,0.38192319765454386,
 0.282820699704491,0.19858164922259258,0.4233599885941104,0.7881760659836413,
 0.2572931998076252,0.6444989596824832,0.5416744572550686,0.7448110682382586,
 0.032282580007399475,0.7648639412835179,0.8811509967860754,
 0.002168008392743448,0.7549009053108592,0.6855847287329828,
 0.6815611852840366,0.02081948772922637,0.2336853015504513,0.5437455033909574,
 0.6890830804429156,0.9763384865147776,0.16049871819729444,0.9178493740911176,
 0.8437337678751131,0.13784740327610223,0.7461510440195653,0.7283768881081562,
 0.03542968179151618,0.2165535089206826,0.06670960242601387,
 0.29151468213410325,0.9260186678731429,0.02651064134378678,
 0.7971752602672628,0.9575359116318766,0.46992851881877384,
 0.44267578131313157,0.6353874798996865,0.9544492714983246,0.9733944120260866,
 0.1250128600782433,0.3064050771871304,0.8224367283295342,0.08577273701373289,
 0.622988091315627,0.9610827934471835,0.6115032432969779,0.08361465425834647,
 0.014565651529234502,0.30849776996604394,0.5939056632665239,
 0.5359766299301706,0.9447133271374137,0.4208666953220188,0.36967002477683997,
 0.21214595663342706,0.9546596516500154,0.8459268335856677,
 0.35945487958778977,0.2919545707009463,0.5974312246321571,
 0.15031277577748992,0.5701927344911891,0.8063364787761419,
 0.10207168672564193,0.30722139841735807,0.6767851125514251,
 0.9001667998954533,0.1958998063887677,0.3028017623281256,0.9382710063887028,
 0.33704464415850943,0.8030992816165313,0.8670442948214607,0.5755503895165928,
 0.6077604848754032,0.8835845268640222,0.7517083625553738,0.6769191510646986,
 0.8655401495475871,0.7314290972127591,0.796031681471892,0.8639385241267733,
 0.051618511981770254,0.5976231987306662,0.9249926230662349,
 0.6915787540459631,0.40937887942015383,0.5808213965906476,
 0.01680034514698925,0.19486411989711383,0.41422192357683607,
 0.47639007577930925,0.08212916054902475,0.9421699229067287,
 0.07484378958576476,0.3111711579080422,0.6021622408869276,0.960825729616495,
 0.8188253471875875,0.4471238760914532,0.8935379046481262,0.8240459521550464,
 0.6284439864137324,0.5013180962512662,0.1547445686086958,0.11573420635803433,
 0.8088108422823321,0.5079850368776205,0.5361382586872494,0.870021087926614,
 0.18401388308296662,0.35415609944712245,0.6492156927682948,
 0.5122781780496778,0.7881580731907553,0.8944962164457343,0.6415837326888619,
 0.192599334565575,0.13604682491489906,0.5102750986113537,0.7157540382112801,
 0.4208078196210919,0.4607367322161625,0.6331040725828253,0.8944978097801254,
 0.7322995117961952,0.09488377129211933,0.1253157942409655,0.1918908427652214,
 0.8601429580256921,0.0964940955629977,0.8896838261293775,0.8510403799943419,
 0.2810081496888813,0.9877477545239962,0.2761038829860565,0.9449396894049226,
 0.5756003958122244,0.6329265344933458,0.29853213636672815,
 0.012285683486027654,0.49426693407299416,0.7211068544323574,
 0.7780750743389632,0.4571205824930443,0.23404347378415602,0.7441952592811402,
 0.48293851848775815,0.7985976068414241,0.14558375352325292,
 0.6869697235626401,0.4888014552299931,0.08370758516134114,0.6181278700836897,
 0.45359327605761257,0.11005166213946416,0.9471661130299294,
 0.3712533491342396,0.005250965742117586,0.29893642095749884,
 0.09844577933604215,0.15416504147067278,0.9091718902266558,
 0.7115310262335428,0.21966024593371258,0.7511359782610183,
 0.22048332454269515,0.5695772201962992,0.9197268118141688,0.4227811835809925,
 0.2307370385500077,0.6706331588984794,0.42385333365397604,
 0.40413960733287435,0.5722095763567704,0.007883023087643881,
 0.270260416363289,0.6311741607077734,0.5222085012851244,0.7899466221498939,
 0.7147895874219128,0.035896658259561565,0.6634922509744741,
 0.5539644013816467,0.5749879548489666,0.4580871298685102,0.6666528669955423,
 0.9077942271886374,0.9538630570173723,0.4822207222786593,0.9100201938886904,
 0.7130200375849672,0.818934980109875,0.39466432053341227,0.523680897485964,
 0.10108425583478498,0.14243390712428683,0.5832721472364438,
 0.7837370060562436,0.14408460009605262,0.3022819412113964,
 0.19113570152845105,0.9469378026262764,0.44144031510267046,
 0.8458767441951472,0.8559325657243549,0.49419302537352705,0.5846097219504461,
 0.9838882567622468,0.6879543646383091,0.5425367508654175,0.06159555628350455,
 0.6513994819123988,0.1982245339434383,0.8231570323381728,0.33476636213412925,
 0.42581080784751235,0.9520645252425239,0.4030137945706681,0.2799644088280385,
 0.7198582052689058,0.07956981219147474,0.7072145385235106,
 0.33214380310841063,0.45979839051923543,0.8205148525427274,
 0.10793099698382347,0.13904577756504377,0.9012144427960292,
 0.6681143711001954,0.3387632344606111,0.10486169259351064,0.6263464205486227,
 0.9320612131681438,0.9998197688319352,0.3509397998788173,0.2878368472881849,
 0.039095108375486354,0.5927758291806082,0.29632203093308196,
 0.012926661803454143,0.7866968628367225,0.07817899991231148,
 0.820118696226971,0.38140770790023304,0.048973828176511414,
 0.34786177009275776,0.912153720949501,0.960525673754328,0.3160791928260538,
 0.38002435214181485,0.7060989613257139,0.775311512303239,0.7087620471547966,
 0.9020629467752862,0.022344767586045533,0.28128487675540526,
 0.13180564390199845,0.5190167192226833,0.8120440356181393,
 0.45603856416851773,0.19094218847186517,0.38654944190105045,
 0.6281381799309421,0.16897130106033487,0.4203812845067194,0.9703976881226383,
 0.6123195747417592,0.5849049605017977,0.3062102706627413,0.6202228660816679,
 0.8453566723987853,0.9330745038451731,0.7357029204962584,0.5981480437654545,
 0.974209473193389,0.8970936816899856,0.3126277191288318,0.5097867011299821,
 0.16417090420149583,0.18372062067691608,0.1145300250057053,0.59908860234417,
 0.28777316340378634,0.9787137014658102,0.4426221153071649,
 0.34979746163021397,0.49166963160822863,0.3539538423091043,
 0.24983503488437553,0.001605156098209548,0.3332891124490236,
 0.3898241775279876,0.32801881059483895,0.6390280817577666,0.945847204436465,
 0.5976839767161763,0.7817068485985504,0.0765817921027222,0.5468693326707317,
 0.8667380474276422,0.1291990180565058,0.3840322071871938,0.33248866514652686,
 0.1143546897532115,0.466915454735914,0.011327767852786663,0.2146206885899019,
 0.2801567595728014,0.33459934116168333,0.5236374984453196,0.9897427948129379,
 0.023110080011189527,0.6104174419739208,0.4863676590070827,
 0.3252141191918063,0.8096159109042601,0.8120169150395037,0.11540085374739362,
 0.8363112095662424,0.6282720735389236,0.2995497854024648,0.951298362530081,
 0.679104256759145,0.5749173565737484,0.6320649897575574,0.5788435713615541,
 0.9970637059376971,0.024212260122664087,0.5537260785442037,
 0.9133211429638068,0.17805120515110273,0.9484729456241793,0.9985491277947878,
 0.39997531085632143,0.29441463940360757,0.4030054512027835,
 0.7064756191317174,0.744354947399745,0.1421318177838734,0.8462430052192427,
 0.8008090194138933,0.6933962585600683,0.8595350107825798,0.6277982931705142,
 0.6930057954388772,0.9858625735205125,0.15628602620879128,0.7498388260816573,
 0.47889640200728556,0.20839590447855383,0.662175321752616,0.2137473191115814,
 0.4687800893480939,0.7162284713059165,0.9965349494442264,0.10862753870615716,
 0.47795600909784874,0.159375135341579,0.787944572086828,0.06779378112082701,
 0.10786909227145258,0.49627682890652214,0.6792862464736162,
 0.47111290812106277,0.036528380693800155,0.45704048748423454,
 0.40574566220896346,0.1771102607807481,0.8594546638953843,
 0.21263006374310456,0.986348775514348,0.5292897077343894,
 0.029489355114720617,0.31762832666065755,0.4970753757098407,
 0.3345006261763901,0.2885545134297589,0.8290488092364529,0.8760817790629953,
 0.9819667087859899,0.5240158869158422,0.5799316010555678,0.6671694868064582,
 0.9613320434157011,0.7118442508326608,0.6703736515722626,0.21822630985451508,
 0.0931107943865298,0.6676678689879316,0.29303238155151734,
 0.29119499621122813,0.6494694538631711,0.31670778118357734,
 0.26743864479100055,0.6497502712960544,0.8634419478866057,
 0.31602807693271506,0.025594305659106475,0.9071804139097991,
 0.531560855388047,0.026083055042468772,0.1381037974909073,0.3251437774184157,
 0.4105543290834692,0.8816901846983599,0.818312426829085,0.4166840844712755,
 0.40920792768845216,0.2850901154853519,0.8205664724887848,0.9524707139916128,
 0.4522583393460454,0.23909586158954932,0.9012475111551054,0.5852666286550103,
 0.935167060704083,0.12728271131395685,0.03716931347695951,0.6084190748117412,
 0.19694784794762488,0.2709291596476687,0.006735948323066232,
 0.3290387739999169,0.41166367444276386,0.9778751771379485,
 0.36725332267767685,0.3066784386881277,0.4958196094643542,
 0.23715877671604024,0.8773038959398535,0.07646407441682612,
 0.4489014414287442,0.9971208296684804,0.22892979157686033,
 0.48827910636284055,0.623463246527239,0.2634475916910798,0.3922265665736183,
 0.6878131950168539,0.36614619037407725,0.5983342917732132,
 0.08296637101046667,0.08106443432978949,0.9355883189117433,
 0.7588223714760813,0.7099229973185164,0.7620746753578327,0.6853530861028818,
 0.1401711075771781,0.19690139313346222,0.24831123454073112,
 0.1281020893917929,0.1970668644788136,0.5017735164633441,0.9933275960950958,
 0.5531229673410532,0.34719437439372236,0.5888249524707502,
 0.01850166695972355,0.9606779748714424,0.5233022652438382,0.7721810365710693,
 0.46443021568331,0.407250095631158,0.14742977961787684,0.7193299646590919,
 0.04192790572531202,0.9349376252671718,0.6210093937475328,0.3633630797192897,
 0.007978548091682525,0.3188401049711993,0.8031058226197386,
 0.4195197562798191,0.37754271301331976,0.6913535957430813,
 0.011710887854766705,0.6626777117853471,0.15806608041663506,
 0.8685845193409166,0.5405855382442506,0.5406581269598802,0.06432531684395115,
 0.6820864121272171,0.1334223023696306,0.5755857707058014,0.4049832215323911,
 0.12149941823945043,0.48413852945681857,0.07351439596312126,
 0.8968646884551105,0.2443072088166044,0.0354417418375742,0.3756428531828959,
 0.6197156003766048,0.5468108449591791,0.5435406777935458,0.31762417393671305,
 0.7195135818433268,0.6627347891457704,0.9241428059539132,0.9508720665522146,
 0.5459690247593701,0.5881069524790097,0.8035306248929538,0.6128371529714736,
 0.6744513420418133,0.7523325421905509,0.6803750011867089,0.3250386632782333,
 0.5121038677935806,0.860984014421023,0.9035329002977365,0.10344147673292414,
 0.18471566743079615,0.1683956639794877,0.005338694820105294,
 0.15960053127112817,0.2584215549017198,0.02051385065948419,
 0.8616692527522772,0.8382949461378058,0.380649621496433,0.1211475790216694,
 0.49901592632486547,0.12792281223915902,0.07355544813813175,
 0.08028967170363799,0.13987509074249027,0.8448731191524738,
 0.6231709141799296,0.47294308059520873,0.1475981331096976,0.9432831921943935,
 0.5958740550882515,0.0011054492821716,0.9834814434426022,0.9456049563683518,
 0.579403659555983,0.7589732036100874,0.3217709787235362,0.6053286714244765,
 0.27115392558942986,0.1586596086936254,0.8527100063760261,
 0.27419721454003765,0.9095951071954108,0.9906095682091971,0.9338577535483475,
 0.7451099758706222,0.34441247865703106,0.023898273161051975,
 0.027312689650655808,0.26632743835218553,0.4368408485097093,
 0.4448984066328626,0.17338594657552797,0.03409629073862208,
 0.17227115187651043,0.7489322406842975,0.5180222543459513,
 0.005718667930818655,0.4723973814562267,0.5455516271126513,
 0.3022152944090072,0.7529527848793536,0.10172124922615833,0.7266665022418751,
 0.8826085702103332,0.7108063680285347,0.6824600437492736,0.3140209931366401,
 0.43993200539370503,0.08168397617131584,0.7717766150423891,
 0.8562868689693989,0.9239774816435844,0.3537568647988302,0.8627179694807003,
 0.6193410785528971,0.05022652358875945,0.8862302334533189,0.5346079224728258,
 0.023324747670447454,0.11804161279375547,0.4248722506937812,
 0.8435484962685205,0.8859898813126882,0.5202472880531107,0.8287716137191707,
 0.6793103343211148,0.4564072317379213,0.2150185808909818,0.6721431597573573,
 0.3640834954357739,0.26265960966111634,0.6922950959238027,0.621075487793102,
 0.15677557470073045,0.7900665925824839,0.45350707851525185,
 0.6377295223396824,0.7141461910098942,0.4305822615844448,0.8063900637616813,
 0.7323746177705855,0.1913524365279452,0.3044273131981876,0.275503316662256,
 0.12933673364013876,0.06787483512249703,0.7326453384267753,
 0.9258933435600882,0.29113877436424984,0.31045212439557857,0.845964100808674,
 0.5842099073477016,0.14641818433722809,0.24874645314775157,
 0.7768862157363713,0.03784953930420565,0.7455413799971744,
 0.01703534975008769,0.6270686935223612,0.13150366505307887,
 0.4964738301239161,0.3214465985855999,0.9303184033335957,0.28460281355647576,
 0.8887786074841498,0.7172529595578431,0.9697688482131018,0.3473220431233406,
 0.12841411788043455,0.5289189548465976,0.6369009496176166,
 0.48429131375432233,0.6947608629218344,0.44135190502273236,
 0.1942575020907643,0.12462440617484583,0.7293233204981732,0.1696785702247493,
 0.051475271559680236,0.2038800178689576,0.8768415006816989,
 0.6057431817374721,0.5849937237277655,0.5547839008799238,0.586275190561532,
 0.28407208303333775,0.2884114243530004,0.8302006703343427,
 0.39613392342617015,0.8302469617395372,0.4882271926375529,0.5394294114265502,
 0.2990963744465178,0.809169000407316,0.8612639422434082,0.3992021885084752,
 0.10679488831485995,0.9917520478129763,0.42570534876875465,
 0.09066278798580052,0.9708329074295761,0.6788198475888244,0.9632928598111152,
 0.41485097786557656,0.019163001578312078,0.8718635897785748,
 0.5654608650959411,0.12985538581733214,0.6678897429502363,0.6396534999081626,
 0.8652988800805099,0.4925522176036131,0.815201961051745,0.5136370539296866,
 0.6885715167917804,0.6051842698787179,0.9135854287747087,0.7965444169876496,
 0.8441477305487334,0.9751262353896067,0.30863769084131254,
 0.09483863159871331,0.2584211233121396,0.29405175832493635,
 0.8543658050225782,0.6339913293853788,0.520816770028847,0.5665024285431043,
 0.9008532993283274,0.8159375723445715,0.9683923008168043,0.787553739130365,
 0.232562455477198,0.44874999686756856,0.5775443538323229,0.28275710513291274,
 0.7906549747182857,0.24561758797511068,0.38751710297625075,
 0.22488833453424562,0.859995041511703,0.3035105821464227,0.8509429038240335,
 0.5378228849574893,0.03206140955003911,0.12397577461109699,
 0.1109782387313204,0.4459645480281562,0.9700901600055634,0.36766677440790263,
 0.32522410921662304,0.42444089806217533,0.057830758967355234,
 0.9851347284953472,0.6777099561374466,0.7189487468294918,0.2967822601784593,
 0.8775753125474194,0.6525541664591588,0.5604863499593199,0.16793955366364743,
 0.9435099387636228,0.10855456872021751,0.4884883428153891,0.2516793844492491,
 0.6259377276624484,0.11068409376089483,0.7149091677410215,0.4917485865260226,
 0.06412362033036267,0.9876773342304509,0.8053190713227744,
 0.49285715058321344,0.09994130144060198,0.3110080027965211,
 0.3271623751298094,0.02320499309939761,0.19035283325084573,
 0.40824186081873726,0.0463038768485412,0.05960466368075801,
 0.41376103813555865,0.964154157024502,0.288027523881232,0.5925738433318337,
 0.32926158389790605,0.5256772802364786,0.5495026273790797,0.5018181687363298,
 0.6652813638633243,0.1219117027014005,0.41530357546261576,0.3320156845533405,
 0.7272187054685333,0.7576040244856161,0.9620101911262345,0.740291525146213,
 0.7895117700904276,0.9068237306322489,0.9763055088719633,0.9888573381056509,
 0.7982621977802242,0.6219576357312986,0.647760950193079,0.4080286759354751,
 0.5349135356599314,0.7127605540608084,0.056184423157884855,
 0.3000485131241526,0.5673297547500114,0.9221267166601516,0.35194017828544044,
 0.0797355875958442,0.9441959635990658,0.1148519121503433,0.30069277509477876,
 0.0648022223859267,0.13282508808636362,0.13316743300130818,
 0.6468005777475948,0.8648427516881148,0.21471386622246413,0.6735728396586784,
 0.9593559247375305,0.36989879672492076,0.8311108099401782,0.5620527600807225,
 0.9675367852655307,0.6266417792482483,0.7080354017207667,0.9251093300342618,
 0.6403101838864362,0.5246674344541685,0.836000623464337,0.49394659385501993,
 0.9083871116531226,0.27587485077257123,0.3834841719205174,0.3136537391258226,
 0.9282753203040106,0.020779504752686417,0.34801401075910987,
 0.7170246129824467,0.014069276902584416,0.4845978262646765,
 0.3915037546725667,0.20149290186786073,0.5708230041284551,0.5806738475646201,
 0.3754925307786534,0.5164016402251435,0.9983675489567385,0.9077563819217421,
 0.15765093066569935,0.8601358793900031,0.37026505138737154,
 0.9532829158294871,0.09200227187941001,0.9056633018437595,0.9330632536694046,
 0.26714068032128035,0.07553930259487185,0.0684895275902524,
 0.3733137363536073,0.4002197983110477,0.5503049490821624,0.24529387985120688,
 0.8092479424878402,0.033299426102402,0.34459335944957503,0.5005424424519053,
 0.9538609466049077,0.11852908841041798,0.04519759739061935,
 0.14414584393148955,0.6659849969217225,0.35376732184173576,
 0.9160901935724615,0.0427312964223171,0.729813032153607,0.9056410548393564,
 0.7492804245859971,0.6019996606916905,0.2695172605617663,0.07886303565203034,
 0.7104767742368389,0.6615471092005779,0.6964512372576206,0.8029112818122304,
 0.816160146053936,0.9191591099608707,0.8902608976105738,0.6847460844057522,
 0.9528927181690011,0.7811758476782835,0.025747155611279382,
 0.3075177588502721,0.6905461434743397,0.022701368845707615,
 0.7996096258439953,0.9473097842326563,0.08083269007602356,0.6432091120307564,
 0.20838582624726998,0.8963066031506892,0.7292203831071011,0.5386205407294034,
 0.5282711127996256,0.8423794322900839,0.749416219034879,0.8456881145059535,
 0.936239456858025,0.5800904049829189,0.8203450438756199,0.540869699858881,
 0.13129697785243621,0.4317588219279569,0.7922272650466099,0.9238331405048124,
 0.8101918118794046,0.2810187687002703,0.32223880278170647,0.3568135788051162,
 0.2999135390727028,0.4673688245906461,0.4563547763174399,0.6422527903093622,
 0.17719458147754008,0.8475641898697378,0.3282295274258985,
 0.13414815697427362,0.9852543429792501,0.8189600687754004,0.9652198254345713,
 0.09652764855711193,0.9587483717086076,0.2677431551538294,
 0.41267081174984255,0.8211260922917414,0.5205764866132716,0.8881533129294075,
 0.4890192286063374,0.05903923423103108,0.46985971454368425,
 0.052653229803042545,0.5488379395240797,0.34743857795779387,
 0.5078884407498088,0.9290375835082151,0.4596834903696081,0.13535510180950228,
 0.27346453636911483,0.5596221042584884,0.5020863524626116,
 0.01870448372572464,0.2405887526479552,0.8448870337191251,
 0.12882940285253897,0.6530292578050915,0.7582118932892924,0.4790886372205275,
 0.18942850381338472,0.6548733694607681,0.39217439662785114,0.890979867444833,
 0.49630911863673877,0.5617366421771254,0.7679973977843864,0.7149293994121056,
 0.20107076277707892,0.4715584257145048,0.9317156917981756,0.2776550119663106,
 0.5026913849698229,0.06624277534403011,0.33254257235853235,
 0.7544312253064989,0.8764965715031023,0.9250394430076078,0.3530665073671265,
 0.3442713550492834,0.6767499528523271,0.8927431261927565,0.6912563484580874,
 0.5202656155986042,0.349177287317886,0.566606773121288,0.6114194419476303,
 0.5180885158111498,0.5605276741023114,0.9957617874265449,0.893212584609479,
 0.34279073931635196,0.9728485368956232,0.9814672409640036,0.4497248040062838,
 0.3236990112218293,0.31787542572832217,0.05591502682911842,
 0.47697093885060304,0.3756815588613752,0.6341894966806527,0.6772944499846082,
 0.3732648391497103,0.8477184447230054,0.06523273055395196,0.2378392236960215,
 0.8123707457774392,0.5944242896820288,0.8642056036336633,0.6489929271428219,
 0.8582991068165727,0.8816248584790188,0.8051286551979064,0.9222462770506873,
 0.2437859607828171,0.5581300339188999,0.66267806155115,0.36866866030913015,
 0.2559155641685107,0.3315774832380751,0.7534147285691066,0.2682138404918821,
 0.8305500132821925,0.08983287535740281,0.9761988583282609,0.8961879910359069,
 0.6130364873623778,0.4596486169198424,0.7693348082830185,0.2866437113415895,
 0.9675379997748899,0.23919079985917235,0.06215229531798594,
 0.19259685772127544,0.05294285922467634,0.08544858137745492,
 0.07181845194013492,0.017944833911229807,0.5284926409922697,
 0.9762215603803168,0.8650676449598016,0.5245228480384013,
 0.020708029430788866,0.7082595696767658,0.34221208151865457,
 0.479895722387211,0.5115323316205106,0.18870569266780857,0.34868003032611183,
 0.5786580032078363,0.2515265290935642,0.9567776300192653,0.09560777699507295,
 0.8565990616102086,0.6556337759504184,0.5232886072906091,0.10018530702496276,
 0.5177051229283061,0.3031893976095057,0.6884687675004908,0.2102939646606965,
 0.23582856812526465,0.6665914223506817,0.4506334020130297,0.7945273151852138,
 0.22434867549164006,0.297130368224238,0.5581549473164934,0.02642014564732642,
 0.17302464390059025,0.40393120949053507,0.6364182498490976,
 0.8635082156531375,0.8383808831373972,0.7090611737741922,0.10114773476685779,
 0.3853842580198352,0.8676114845585403,0.33745958762274286,
 0.07905585428208661,0.984216924073344,0.6230042807309123,0.520461245288855,
 0.9996959566187236,0.23748767918696734,0.6018688278853915,0.9731769934075111,
 0.38733686170709314,0.8687053899402604,0.3636065788032725,0.9982419667699358,
 0.05643525353477019,0.28501827517462375,0.832378288897227,
 0.24313002716951482,0.2620484380417315,0.16889592908139128,
 0.0921484869832151,0.9712703550396149,0.38110802897197893,0.9668623511048297,
 0.6476794427077058,0.9744144882749219,0.853066244372509,0.6228495531857946,
 0.6965609917382215,0.08548005118177016,0.22988950209689818,
 0.27303272016334323,0.0804335005804162,0.22940217594810142,
 0.7194275598900954,0.4245046540164734,0.1991947749753734,0.2523336529643363,
 0.5998079616899165,0.15312731488390685,0.07286519185801088,
 0.024077482743404932,0.5642633255757671,0.47047409018220876,0.56771983738876,
 0.3115842598476589,0.10748752402390649,0.5703588931241214,0.1471736350520193,
 0.9790134500407772,0.7556350316931854,0.21210319047238124,0.7250066689543245,
 0.28866843806849485,0.050808776881666184,0.6992227659043753,
 0.3542425370501374,0.8601982086177192,0.20810952398445148,0.7434998778930483,
 0.500579069883638,0.3045461329135216,0.8231674606341788,0.363572802258592,
 0.9172139593023698,0.2298743264086598,0.49835875026857823,0.5495236498831085,
 0.4497623950889772,0.07730603765101662,0.44616150500269647,
 0.6628125608049182,0.47186204279542776,0.23060589955649413,
 0.3225372822102658,0.8628168324985532,0.24626366860920268,
 0.0032983178226538845,0.5111428987383473,0.27626103987912864,
 0.924114183161293,0.2726808251248001,0.14665980115995103,0.7081395779190574,
 0.36120940274574576,0.5580162250269145,0.36646563747643945,
 0.6980497957820844,0.6304383399388459,0.658346511458485,0.5840163196604086,
 0.3014301995953834,0.6658898869086376,0.6410225446444454,0.34060585784395203,
 0.5822957124983207,0.42586172294091795,0.04073514202446715,
 0.3827735390146696,0.5334787468483605,0.9336167734492429,0.8631082627467276,
 0.3077689609381673,0.5006165615086737,0.29682478316956884,0.5345566212111145,
 0.36726514188719583,0.27449169869946843,0.6478418377449302,
 0.7750133023906501,0.05549326301423607,0.9997098630371377,0.4830618591322564,
 0.42765480982709114,0.6987064514053289,0.441046308507559,0.7943598473331883,
 0.5261340923682626,0.26305928267687084,0.3464941560963616,0.1686136453648609,
 0.0760676730213059,0.13370137391443526,0.16042206535658876,
 0.6541528754347061,0.9189742603437502,0.5209067522465006,0.11134059913453465,
 0.27563505647904163,0.7506735969441378,0.5435184539777184,0.6862783338676516,
 0.5167228816094472,0.85182289632268,0.29099995320238514,0.3382916576810332,
 0.17477833311585,0.17691830193989433,5.9250474755701e-4,0.06336817872512157,
 0.7966701523518104,0.3023300012422204,0.5244083276862066,0.6446104406866536,
 0.8434251777918078,0.12017498559549189,0.9438049734770174,0.8869504708399919,
 0.13761587448030077,0.49254227735846323,0.4741601609295234,
 0.35646421636558967,0.8447202969651337,0.3182449826727143,
 0.22829490914370132,0.6866883665161752,0.8781393185868058,
 0.03234754439176757,0.5860771270046169,0.20018769066271724,
 0.4510261341058568,0.7631814250412047,0.7328977535753929,0.6856217804234799,
 0.7927476156381954,0.3188949151843893,0.982192491500292,0.6164815211726463,
 0.5364255276320473,0.856114573926731,0.3874558845122723,0.16013134096731896,
 0.053227509747109436,0.574219703001711,0.7230537836920229,0.7318283444812973,
 0.6973179347458142,0.7874010868652536,0.3734312580603897,0.4928935844868314,
 0.9722861631206989,0.6642716632053697,0.2701974551289531,0.17008906330963525,
 0.54445495435057,0.04545186650893174,0.414874122627771,0.07650244918126037,
 0.7025174406652213,0.38190165585171987,0.7789727066113107,0.5275226342328334,
 0.6160029709072957,0.6647448946328056,0.01357842170079926,0.984142005601861,
 0.6196140278306961,0.3769744258887924,0.5710740757207533,0.4456397597396865,
 0.4989038753933057,0.3402549872744949,0.994552680344708,0.7818152547797186,
 0.6905427166361202,0.7886636263929084,0.7979310719052417,
 0.016852725093447485,0.22039805431883974,0.6463228656378165,
 0.7593633048954564,0.5055857938062187,0.41824827527499475,0.2012709114755814,
 0.604485505519757,0.5586580075392678,0.8425371515753315,0.08877293609323988,
 0.6928042426240136,0.02184921588623212,0.44007109983665493,
 0.9998280241090303,0.5144340793084932,0.47626796458620435,0.262128474069387,
 0.35784947256349975,0.15802698275730487,0.2490277806796929,
 0.7485348971365836,0.763795168182799,0.908522717731619,0.7431748552974361,
 0.9321944225689226,0.49013358907516935,0.8584987519566695,0.5649149952018838,
 0.8740107433226019,0.2650390830897541,0.9215403927639202,0.43965840187360683,
 0.7476435975859324,0.4586486946038423,0.4059292544654388,0.5758394094591712,
 0.9801801139435685,0.06622541321147246,0.4067311645513991,0.9659333185671743,
 0.773525406067603,0.09252657512230444,0.5600246837439633,0.6962316407567836,
 0.5288593678057073,0.5424874106559662,0.6331745099153925,0.17755131235116528,
 0.09334191347527476,0.2668338766641073,0.5943613365504381,0.9471876884165153,
 0.2698945660569194,0.8311415832386546,0.6804115438782676,
 0.044246739095718746,0.7406925432605356,0.5955791161133435,
 0.08615661276073106,0.9774487760222076,0.1728382750487928,0.5735441756852799,
 0.567940195355004,0.829163106644913,0.24021458905047133,0.36729482056953044,
 0.518232698504646,0.9075240525982005,0.11938622239937347,0.6942969036980313,
 0.6788312335261434,0.18940248280809924,0.5405513177359565,0.613762275351195,
 0.5748220086920982,0.33205392968115643,0.4109930544392385,
 0.10187348784195227,0.8240229535277959,0.9339844402532959,0.9960310964544415,
 0.23644530834364796,0.4987134956497976,0.7613148896050325,
 0.16326123055185715,0.49439674086298724,0.23545585436719862,
 0.9950096825223036,0.9315742091014251,0.4493104346189821,0.24092851850432062,
 0.7433754361032097,0.04028420892474782,0.4392922955005317,
 0.009118790832365442,0.813608377627133,0.0704250903313609,0.2379971123973491,
 0.8661202665697335,0.687904139513978,0.1913551402109427,0.9667782641512475,
 0.772340151366221,0.9299240675899519,0.8671874624226645,0.7942908562871193,
 0.4138559484890587,0.21596934785371857,0.38291021193484254,
 0.9458482540241654,0.24006924689703324,0.16740031648188158,
 0.5607684147988543,0.1742125495916852,0.3081786145669868,0.7362654834876698,
 0.32727022103895353,0.06909355118643012,0.11894520535360653,
 0.47796163978310613,0.6197198588370424,0.783761461751263,0.7800034941812174,
 0.11001080883954084,0.49944884161874237,0.9304067285059143,
 0.7451149761787706,0.8545013137798073,0.5211669882629183,0.370322669794663,
 0.0017951070213115816,0.70173757056662,0.3536298530208486,
 0.46533478145947615,0.35824231220969793,0.11359381732575691,
 0.456989110074855,0.7710910507177035,0.3829912302883939,0.26833029417615717,
 0.8410059148919676,0.1880615564775996,0.1934675180013059,0.9329935559530698,
 0.27050799517147583,0.10585986193136687,0.6833797360259,0.3349151955219444,
 0.9155557344858143,0.24699700407441205,0.7003370190686152,0.3468774616800615,
 0.4002204409822948,0.8290244074175512,0.932330952580178,0.23939129074754717,
 0.03778746528231991,0.7385067740726216,0.63456829557833,0.16150379680275861,
 0.7465047841842983,0.838238106694166,0.26320346707234354,0.7468099074964905,
 0.18851720640749148,0.7289913564387656,0.4047440670474035,0.6045471361600775,
 0.07011544146873128,0.2932812567388189,0.30518987507045825,
 0.2518930698835632,0.23009753963043078,0.47803185304928264,0.858780998491679,
 0.4292058053805121,0.8939909509237747,0.8051700503732795,0.07305727635063508,
 0.8593941599597226,0.1495833094472543,0.40796518730843156,0.2497210869289408,
 0.7812966108607227,0.024875225541524326,0.3144379213055781,
 0.5851187503432428,0.22323745422640762,0.8453617378509242,
 0.03925070398424424,0.04261650376081361,0.15530497716551273,
 0.8369306679135363,0.07277941076140493,0.16374033866098814,0.802953431367535,
 0.20337200340550265,0.18027600135748845,0.7967154555011782,
 0.5047310246581869,0.19388911442406587,0.0808229142929553,0.9347224961604044,
 0.2886006293317557,0.24539344461259494,0.22650454245307416,
 0.20018297011609898,0.6496595642746497,0.6772438500197367,
 0.25300674496465536,0.20874942842363264,0.5581243768994705,
 0.7525171452757596,0.29562624603272003,0.1998399723948897,0.5910812798203069,
 0.9082828075283621,0.20684773808786294,0.2132422884839028,0.9665845064796457,
 0.8783539194335038,0.8601625584231015,0.8177781688994541,0.10819124162814397,
 0.39585412478607784,0.20525038578519683,0.4813095089728312,
 0.19032149874483406,0.9354982639617111,0.07528676600084616,
 0.6247570988793762,0.18503307234914146,0.8299399364356005,0.9010627700377083,
 0.4433995358900831,0.07825974637069477,0.7599583843767105,
 0.046945753990051964,0.008470938848597176,0.48344371257600083,
 0.8965375695023386,0.27745989101600077,0.06673548899942083,
 0.051364683836831015,0.6558614619529803,0.3167351892912754,
 0.0012399987134251722,0.11493066060397172,0.7762388710916082,
 0.7616457202223181,0.9249295815155372,0.802626765113452,0.5114235329128064,
 0.5372425896756621,0.9046995655082879,0.3268203314172279,0.42435765777135925,
 0.024682148675922222,0.7045934397967668,0.5729508045235359,
 0.01130374876536977,0.20843532761166383,0.26537582040126795,
 0.8885373001784769,0.04609291994548492,0.44126918262481485,
 0.4098148186586761,0.7296760106020396,0.5092654711490111,0.30385516006584834,
 0.4252812339583065,0.6064058149234708,0.2730135342794804,0.9925829548275878,
 0.3404279468016279,0.8288029409015072,0.6218229394620911,0.8534729132190336,
 0.37843385973312027,0.9006899044236012,0.8606438301049955,0.68612326432185,
 0.7664070228332077,0.7456764478845763,0.36720172694676245,
 0.06214988567927415,0.856931994309852,0.18297324676992055,0.5262668865143301,
 0.8795898003537258,0.6221051215844566,0.6842207235026264,0.24778664011511387,
 0.8782677213945759,0.23632190614513593,0.24737775172700927,
 0.14845534382686398,0.29986368075247805,0.057036890499780335,
 0.09954491002252375,0.8355098966001256,0.7523119291836831,0.537259032143943,
 0.7316785842532331,0.2783747534799127,0.21410478075904216,0.8607747142052334,
 0.8551502777797677,0.5221636026420824,0.36040519213181066,0.6652086560796944,
 0.42118032546811845,0.8868792903552976,0.886763351814868,0.9253920108131757,
 0.6619003111481814,0.4479005472168539,0.3828658601199495,0.5328416523477528,
 0.02655540131833145,0.8955522046519382,0.5145858255666564,0.8146731085158166,
 0.01754861631526261,0.2660957701708171,0.09780177855031424,
 0.4133719501481483,0.56871904535392,0.40600122249673576,0.1585702324572047,
 0.024119327017120273,0.49552749445054367,0.08381973280579291,
 0.7536393558260134,0.054892174325719356,0.7659841276558643,
 0.8309037450724248,0.9643155347026018,0.6281939495021718,0.05542193591995137,
 0.7060007574542428,0.07830478255775453,0.6724827912932383,0.5696093490653823,
 0.36931915834750517,0.415283008651378,0.7788438157528024,0.6827838049840276,
 0.49572262934578615,0.16138148546400666,0.18192976699497732,
 0.7120938025156298,0.7188804120111996,0.8930994512352476,0.12408627856587662,
 0.06507082129710651,0.4324916651484355,0.4497472449457599,0.9094249438623032,
 0.5320680352951919,0.608651166614635,0.17677084391812292,0.5050464552798248,
 0.3252232208288721,0.33272869772295344,0.5201878231922429,0.4365972072037162,
 0.5608539350092061,0.19862545361586448,0.7908492185308177,
 0.28908570633112973,0.7318264373002326,0.049141754207200905,
 0.49274957779994055,0.21749222397725698,0.6438835435995124,
 0.5672493516415965,0.5232729249058033,0.970756330997423,0.430314444080644,
 0.06226763210343478,0.11236112395322895,0.6028255321418602,
 0.9862934573635357,0.5283855082165888,0.8256746228198695,0.5414358877284584,
 0.2514815445949241,0.7249183467422199,0.5028869639843863,0.9586854664579582,
 0.41535527257757654,0.011348467116418681,0.5131153575908666,
 0.45300490051828135,0.6079706865985831,0.9887865587656411,0.701469389103156,
 0.005253228014421518,0.5882466845918732,0.04136703194232705,
 0.1544598490514788,0.3336273964436457,0.26195297071612256,0.3486992159453097,
 0.30304216367372105,0.9815181600567741,0.6935073994557288,0.5531446327949265,
 0.641791776076127,0.51236045065755,0.8928972634090421,0.7410341144689353,
 0.8672279780392621,0.7270516965479025,0.4945770443312125,0.3575311320955822,
 0.5549248138546855,0.2406958537595828,0.611784718824153,0.9024563995811973,
 0.6457122518336991,0.05726901046746302,0.11374625217617307,
 0.04262648552287729,0.549577337219055,0.831642671341164,0.1520699575258544,
 0.37667345798471663,0.026334781613807423,0.5456081913314612,
 0.6202820370292569,0.06014334520770437,0.2199263785625336,0.7079890620466462,
 0.3383220154499954,0.720957649873495,0.46463651807668593,0.9589961417983982,
 0.1922595251292526,0.8464258523233992,0.1352017213546921,0.5692002216886153,
 0.28618498126623815,0.8561555554285837,0.9447666021226956,0.5576438691747393,
 0.2124635974522785,0.2483821290420608,0.7508402187697746,0.04029258768674937,
 0.5617374623532284,0.06066154062025764,0.008600893144394273,
 0.8685752820750532,0.331630542918779,0.32127622897200625,0.9413384497626599,
 0.706680429582462,0.17220862637229484,0.05452275157097264,0.2576274367089326,
 0.07760384027619094,0.6954911331432577,0.9944887746733668,
 0.34521518410042684,0.24107069303925022,0.35517687966291733,
 0.7597073557680362,0.7064880659178007,0.009043901243096286,
 0.9754226163929236,0.21596536753917173,0.3824739418637092,0.9833523434270043,
 0.09633562698321263,0.49772217951028175,0.6069662798068766,
 0.47466060480583117,0.9799055719083622,0.8847788902797881,0.7425819447712164,
 0.6231613020884703,0.31820543155005354,0.8560344591993412,0.4478313641071916,
 0.8817629562648751,0.6590925728954351,0.7365483394059279,0.08848605944640697,
 0.0020831027505519017,0.2726948117549287,0.8699346364720948,
 0.8027394773039933,0.3298767088599298,0.3359614276519297,
 0.053039373067301554,0.7207415756400061,0.5105352052440097,
 0.7125356883828571,0.23587171891248948,0.6418293799659958,
 0.007215441013302626,0.8836504399462134,0.42072673449958375,
 0.5067686168677359,0.050811706175501126,0.0451836006302313,
 0.9428756871841028,0.7133885961744589,0.13131618394202438,0.4536910467854003,
 0.25800927634685733,0.4571922047165041,0.3936567579396333,0.4082852066868332,
 0.379637830976853,0.33847443540797617,0.8417566148558896,0.8715299856218958,
 0.8969537398293963,0.5397662117860622,0.33891973700826383,0.7304315392290753,
 0.9836102973213804,0.979663174886936,0.06715037756703424,0.6763838933380907,
 0.7219424656752519,0.770712332734682,0.9064402143760234,0.5303593254467504,
 0.21924581070971172,0.2493796029866343,0.185524147111912,0.39830238446255795,
 0.5230701595131437,0.1353534289291285,0.3151308835277157,0.9453376546662503,
 0.797590289398181,0.6174853299166609,0.8783814925001494,0.08523315069027326,
 0.3447173998986507,0.04128512029650944,0.015088079350698114,
 0.05977355980918575,0.19687633671138383,0.2423484795992228,
 0.28879421545014194,0.915940168931051,0.19062044156030178,0.9494822909502852,
 0.811970271045342,0.4299573443819068,0.9251856629446412,0.6043325484484203,
 0.9197436208662294,0.1018714984686434,0.9440183423320402,0.1092728401440447,
 0.8967783626656816,0.41303177495719234,0.52239556059167,0.5344046602498889,
 0.5526178438470022,0.7477034318431988,0.04519725580539857,0.6391564176507272,
 0.6324162551245534,0.723397757930504,0.052901281798272026,0.287544318363532,
 0.4296670182283243,0.6830853638049683,0.7853195945197734,0.22527492392155024,
 0.3797474939809188,0.7727322680466558,0.22199154282076272,0.3235715452422563,
 0.6071418939768431,0.5508897896593792,0.5654200752211382,0.635513917981795,
 0.9071828615227988,0.8763337308663677,0.5005897603170278,0.40026284759407693,
 0.9351872101276701,0.8269332657415833,0.5028462009599552,0.37491097313743205,
 0.4801031144349872,0.5092689998994194,0.11401452668011403,0.9976846707132241,
 0.7967387696487732,0.5957725475468353,0.10444716418342326,0.3732538591096328,
 0.9087715821308732,0.9995893005031082,0.5039669278821017,0.3080671928218738,
 0.40423118709884465,0.8745261463434107,0.9034605443818,0.34963118068441057,
 0.19786528796837322,0.9805449465771416,0.36297271904034767,
 0.8352931714993503,0.2624411319303227,0.060997726425071175,
 0.16642027358711442,0.06637336895603763,0.6808949562677065,
 0.9686520695339063,0.6482074134828846,0.9100324632395208,0.8635019183405253,
 0.8575672205666403,0.21879794460610147,0.9389582627653565,0.9408014264368754,
 0.7141856428213659,0.6856141843990207,0.38798831772156894,0.5206311903204532,
 0.6426187057115692,0.5156862084078394,0.7040103964195338,0.6252327284262635,
 0.689793459831567,0.7401541974288142,0.845625007008619,0.07745069147355099,
 0.5391197653623601,0.4116380547813385,0.2040978463631834,0.08860946889955257,
 0.8645811938798983,0.6543008568918419,0.7201689632949173,0.2159345021198945,
 0.727518642831871,0.5311810004324699,0.4720964050555094,0.6092236321811997,
 0.03001397799691663,0.1505683859744248,0.5196657458038761,0.4550849675511798,
 0.30589431553216684,0.20716251983454315,0.8113398617420908,
 0.5734113522504526,0.4086399718663706,0.9075825037835785,0.48017156575356146,
 0.23932196261651262,0.3096435740596578,0.4201091788240171,0.6387476514109668,
 0.15239174057522353,0.7073861821892441,0.5993080858383768,0.793432757382137,
 0.10671041079911636,0.9727754226175777,0.7505260234417324,0.3272384613134439,
 0.1943020036002392,0.6719292958043495,0.3763570686888541,0.47477195557535057,
 0.41627051107778434,0.9189875823796536,0.9507763402124653,0.2427224202131375,
 0.7335661274764232,0.4000875245175495,0.47622073453304525,
 0.23391727907780513,0.1569821613058867,0.9371788826776749,0.8400651765160114,
 0.15437193460216825,0.9345714894883355,0.5823971547334935,0.9967864648231752,
 0.35628649941113355,0.022737421958900828,0.23266307303434408,
 0.5142551443212678,0.7027091368415656,0.2305861316290806,
 0.015572419399177972,0.6720632383975232,0.002441700441704664,
 0.40865196841218043,0.09890899414229759,0.8593590364868972,
 0.35913889921301423,0.09155177013639981,0.017680104508771666,
 0.6612804359109951,0.4847063469446544,0.35271924279580025,0.7518966472630714,
 0.5300374985297812,0.34671334314965996,0.76396664330014,0.006721247872658931,
 0.8027766341755802,0.17171959591754937,0.3558401001691339,
 0.16627035711131044,0.42530596956908506,0.6413305090948178,
 0.8071543548683545,0.8407017789628038,0.15699701946255495,0.5655985897025443,
 0.47180211391688653,0.6968246524990676,0.315855906366153,0.5733873989747883,
 0.8068103563413762,0.49808225582168153,0.05547840388610714,
 0.4452664313221196,0.26698754636946875,0.5479838427582436,0.3911879977452192,
 0.6143264212649029,0.3946994186454307,0.7863544926402621,0.1926323055690643,
 0.9151691208750578,0.2941749724355861,0.642663084865022,0.07214789225268636,
 0.3598365284186019,0.6355773980860795,0.5393115051872839,0.824346699006093,
 0.9115495497217554,0.16452293789495986,0.436780991540902,0.9342824762256786,
 0.5922856819255548,0.47309513852226437,0.9204023225958204,0.9738383123720271,
 0.4618731374214049,0.11558467255415494,0.9897083925800277,0.8389713459048423,
 0.9285228508652962,0.6796005367761871,0.046040638839917714,
 0.1303043340184401,0.5540332781776879,0.28122396286148543,
 0.43514284261143843,0.4881271310955051,0.6443850078932335,0.4497680133861388,
 0.7093393966225198,0.15376095901686637,0.9991002430496075,0.9052721458674748,
 0.42120748255938123,0.30852735362722106,0.801656339359008,0.7525979617747467,
 0.8344757090300337,0.2016400963454228,0.21862253725913694,0.9478241125279805,
 0.5239553621079501,0.6308984697519954,0.20472506069444196,0.7736684950576458,
 0.07132384999103691,0.4573161247059524,0.16413946653981815,
 0.5535164231427591,0.9880178103082858,0.2561443165622912,
 0.010606517237337187,0.9229683450021191,0.24405277757797306,
 0.03573336808567107,0.5575084259030975,0.9780641931634313,
 0.47999382099910226,0.4840670220052047,0.8105964643805827,0.9070956048455852,
 0.6835551486452961,0.37374560097730625,0.8569392912667214,0.9054565793101611,
 0.5746489507904734,0.1824942124278346,0.19839676424829633,
 0.09524336885023144,0.5050054651034505,0.6979160899469671,
 0.12473984223828816,0.4516072425409159,0.6974412273490597,0.5855552506959775,
 0.4536890130105151,0.6818486278612277,0.8604053778295647,0.518470235305951,
 0.8663178361270754,0.3512035573897777,0.5729692072416244,0.9750417771390301,
 0.34158056990498875,0.7888439802115026,0.847379638442918,0.1480832850788154,
 0.4568527527483286,0.1999536206209922,0.39868174506770426,0.5984618235913562,
 0.7479921097748187,0.13859469282486714,0.9555471727084773,0.7380638698182935,
 0.6070959988050787,0.5368215110544953,0.16789351511536743,
 0.11365558739802051,0.6482931960651694,0.2780232657862989,0.5318034705290181,
 0.6156043990818575,0.47121532160306623,0.09116887895650228,
 0.8930572657593467,0.8266739897732266,0.8896406723911261,0.8167916135026315,
 0.2832940203567583,0.9583491696969418,0.17279303609143892,0.6684453527106512,
 0.3056414954133845,0.14538995842797597,0.9013646478679054,
 0.46587379243036897,0.7460878295208325,0.7777557801548931,0.2974224381660425,
 0.8514556932726983,0.9646745664835858,0.12223908795026583,0.5294030207741132,
 0.8448833997747093,0.040210678411454315,0.15548581910196602,
 0.6270898287001831,0.09863645498443918,0.5466014404831705,0.2829586189390825,
 0.367046488507787,0.5800253304017845,0.5892769077078904,0.9741041169031751,
 0.46727761655868794,0.020675093613216955,0.06226597452237459,
 0.5321149318315044,0.5839152752456526,0.041125468828903444,
 0.8097975368456736,0.578485508334403,0.6664133605136603,0.8561319620306584,
 0.01874520274989977,0.1660061236962369,0.9142915484584491,0.5618664445088598,
 0.1377978612258387,0.15502998250017253,0.01168830673599408,
 0.03629698170018336,0.04330648732450526,0.09592180900530356,
 0.915669592405287,0.32463466354622006,0.7652451064074713,
 0.025364623398219988,0.32184182440396125,0.09917035123064866,
 0.5013508283744907,0.041053873609043445,0.2177600126518392,
 0.13802129777274297,0.4486720769092587,0.19717288373768582,
 0.031983339761118224,0.1752480165777759,0.9122863252775536,
 0.026924969244299546,0.7600585595188283,0.9195226138970877,
 0.9937832953168246,0.7918321217190204,0.9359543268539883,0.1749861194849872,
 0.3488830987282595,0.5477721810574867,0.011136448256207188,
 0.5143318130973917,0.29986271632309947,0.11258482832273498,
 0.6353254819069248,0.6502110525286207,0.26394716503089366,0.9699129621565253,
 0.1382734721322687,0.7180849061921742,0.5646680303035351,0.09263152690950838,
 0.8938465164970082,0.2901547536703609,0.5206998974736724,0.3229089595545849,
 0.09100053146251141,0.0034695173352101705,0.6652182077179167,
 0.22428711999105633,0.12087299871907264,0.10739726026429253,
 0.711353134057434,0.1056538400659921,0.22067652437338736,0.4468623498016706,
 0.12451927015290054,0.7278664561563426,0.3431576651164938,0.7674042646504006,
 0.4320988204806008,0.46222242876681663,0.8143968191409273,
 0.36898407372217823,0.5570061182391284,0.08637292291044285,
 0.9477530549197705,0.9702900157754504,0.02797584327392333,0.946746137865833,
 0.5991794768222292,0.3155979973925712,0.4538735070443668,0.5371747627523085,
 0.6584897309517348,0.04035514094534387,0.9333740646639557,0.3449742299500168,
 0.4736432204194384,0.057987567467744694,0.9304660592928748,
 0.5383906559775011,0.4423044728942348,0.7686009643250331,0.131974106679284,
 0.44270409458592064,0.2724867995636715,0.3309874193718215,
 0.15121849939586363,0.861590204723781,0.377506424214088,0.8543349179363946,
 0.082961403986469,0.20052217211390078,0.42804737411976057,
 0.04316668290384096,0.0378770726133727,0.5830781576530442,0.5063185000573169,
 0.0012534460192485941,0.25134505948870944,0.30619196503519164,
 0.5850948125952408,0.237403808950704,0.013607736115918367,0.2727276620647795,
 0.6707939699686982,0.44354563955178894,0.9704360568112237,0.704284315242078,
 0.8899961176949397,0.975323189849647,0.496777222054843,0.0231415535505467,
 0.6815897069977174,0.45662185982326564,0.766686774538963,0.4433455740249128,
 0.4347692069417165,0.07129409810200449,0.6557386982938902,0.9647640940022504,
 0.007168099069438383,0.9169532487087421,0.7927648058322843,
 0.9223349896993819,0.27340529962340177,0.4791920432800463,0.6959350327828161,
 0.09025041059879058,0.9460807589911937,0.06497902513317566,
 0.05262086745594596,0.8544713683382367,0.17141489849814673,
 0.16950222921014224,0.3471113241058603,0.6271982974401259,0.9421697323251755,
 0.5761830417327751,0.9201724069748097,0.4732058614186342,
 0.006822507882841222,0.27498936404879926,0.8075532756152645,
 0.44922265451165666,0.8428056878280504,0.3483962663299659,0.616923083409837,
 0.5744693591592369,0.3706430127808964,0.8203169944147021,0.7459292294879879,
 0.42576711530195843,0.22653368054082357,0.32516337062407596,
 0.9180751840594716,0.40797212142635675,0.7075180122530539,0.337242240702639,
 0.6940307556139209,0.8425901218232879,0.8990289261219333,0.7375612188701355,
 0.26319760620224975,0.6841524675416348,0.5737116907818477,
 0.17439975627905902,0.9920478728886049,0.9900336447327155,
 0.12527246128902547,0.38999235684878164,0.7189054827265613,
 0.9287540971521973,0.8814267571561749,0.8216312462065587,0.639681785380118,
 0.7649724695653034,0.24932913063051942,0.0324190203293403,0.1666782053499949,
 0.3808386193443991,0.8375802929111047,0.1747805157848763,
 0.014711014349908735,0.7951919556523077,0.43599389957144175,
 0.3448512562711812,0.2136036454607706,0.4689355380952456,0.50744533462183,
 0.3315310445512596,0.1862163909600142,0.39648563588175967,
 0.24560660211609464,0.9420881214258477,0.06071017032841142,
 0.4537459389988797,0.7007767348629791,0.12334956548832188,
 0.20446311763769287,0.7958148542485473,0.44129116593412454,
 0.7298831621709451,0.8447034045627467,0.30676891833354814,0.5780392532161216,
 0.9725978750173598,0.3622263671517174,0.9179426913773259,0.3476567999663247,
 0.6311168955148552,0.29189000572442336,0.7692996363732211,0.869588951799375,
 0.19417927013793157,0.47859940038901205,0.4157669592377816,
 0.6935542511127617,0.4168927619527032,0.2607595395286656,0.5413727165689216,
 0.1576522806092221,0.8143856485119225,0.16218554287693665,0.6616207773476825,
 0.4586492593400169,0.13252944240330855,0.24196144497072525,
 0.9460116610394866,0.2863177698660043,0.6689540376184469,0.0656845825369593,
 0.16752949550079388,0.7895778861646139,0.5746281124242781,0.5274087266248584,
 0.9226925830382776,0.2765591121962162,0.23615338133935426,0.9948149014725988,
 0.6278708170582813,0.46292281495385024,0.2043185115349886,0.3272058499888719,
 0.6434682190271741,0.6262344610247034,0.17613116316410637,0.5791752822385307,
 0.20219413107669704,0.4187313582023382,0.45106568508627964,
 0.5850622948674631,0.6484608425240729,0.6331172851377359,0.0639352209654459,
 0.5222332665069066,0.9706943922373803,0.3637510960902347,0.8075607530563222,
 0.6354004516316826,0.0976883796580501,0.2106423909031111,0.37783964221647914,
 0.8986848240133103,0.8735045364952516,0.038467020169529076,
 0.7926593110516172,0.18098221145851134,0.9754663374435477,0.2118351759250332,
 0.08276478904504925,0.9441128922850708,0.6514257412954773,0.4412057156663345,
 0.0724912942004301,0.05901227997814584,0.8063014461994304,
 0.011852877103327941,0.8557815440335752,0.22135173638008254,
 0.08460008066957947,0.7709485965700107,0.11531421121377827,
 0.04453685555647313,0.3871279600078744,0.15349251865743285,
 0.10097500746191124,0.47990600069163025,0.8826227752218306,
 0.4077278082665119,0.12275870036260705,0.15153220561180536,
 0.4473994279695782,0.37454717253769043,0.2070519862312321,
 0.31192975813482693,0.5401981668168929,0.6316555001412228,0.4813785483749027,
 0.49943215217820436,0.9625534462595472,0.5226269938937427,0.3493105621829642,
 0.785645417802709,0.9878695127512029,0.43254313308507686,0.12143191231896633,
 0.8655480881824097,0.35136106518508237,0.7979315720072104,0.6505839591907884,
 0.8524774252847355,0.9581914430012178,0.335950709184047,0.6297579749520404,
 0.7173298248813327,0.023175300507983776,0.2202945605278892,
 0.6556276834218904,0.06052592069647389,0.39563911483140846,
 0.5436404567246806,0.9284198577013241,0.8984962460127027,0.5899608176633977,
 0.35395445284769245,0.5166856390623902,0.4185955895212927,
 0.48705447173990557,0.9794934331436043,0.965360453623023,0.6348337897916425,
 0.4980049552520467,0.829383110682192,0.6751999682882596,0.4782454962692235,
 0.20370344122587536,0.7786995471002467,0.47039451338539406,0.415893983941111,
 0.6976843948903303,0.8128242505134281,0.15677000688626386,0.5677986096522827,
 0.07406135892562149,0.35402505833026865,0.11324402654212651,
 0.11035879561682793,0.7002788429803412,0.85340130654627,0.6252069778821369,
 0.28882151413246304,0.06532786137477709,0.9019381129242975,
 0.27427608158847694,0.32361797079167265,0.08215236944661108,
 0.26153952140265846,0.365882451266635,0.3876726054864045,0.722361875821356,
 0.37483015286715105,0.13013163202055456,0.11211955545128693,
 0.3549506581792602,0.3398005822892002,0.35827714474717953,
 0.021807078010431913,0.41551881116775424,0.5612606255319312,
 0.47823946711442955,0.08707917069001625,0.6223094861919545,
 0.31591794081178637,0.16568985571734718,0.9859451427901467,
 0.6405265229430778,0.32212068620100087,0.7034137613858664,
 0.08267577915189028,0.20427945000874526,0.3819373945002008,
 0.29320068597377136,0.28731491992199376,0.2909624654626046,
 0.6348259412779419,0.7750204617335112,0.9967817834815621,0.0649283135343629,
 0.7829856802582499,0.20150186089265176,0.40716703528233833,
 0.3143090127758116,0.4740368226024294,0.09593203844266596,
 0.39757023831912064,0.4535231013941282,0.5556907126529057,0.6891914208407371,
 0.8282733184045032,0.965491375423542,0.02312296385887258,0.3430171467579912,
 0.9232665660433748,0.5370115681201968,0.5492759304790737,
 0.043023720439748336,0.12683532248539187,0.5021713415722772,
 0.49919267320160676,0.5760427756017203,0.674990770027256,0.3208279595941014,
 0.5002416719248726,0.35472159077964727,0.24565491236916293,0.766561370767356,
 0.6822299692502183,0.4366230378714504,0.8090828964988548,0.7727863014989323,
 0.757474718933717,0.4470783173547489,0.16167239561315438,
 0.0011301916726788042,0.08467381602374724,0.9005348833273502,
 0.25529795098536623,0.5214003498858181,0.6222100977711631,0.9994917521443396,
 0.07885889952881264,0.27238438196536385,0.5773775999120054,
 0.5886018645897473,0.12233835479102773,0.653746268392991,0.899670123269606,
 0.9038114367964493,0.2883708035286394,0.25345685751763136,0.4511853699037558,
 0.05828757683390118,0.7565931293887125,0.936302397964203,
 0.0066108628287665505,0.23517458680925196,0.2221222014996066,
 0.9931634139200348,0.9247290694831664,0.4893385695164034,0.5526565741879217,
 0.18927647171821538,0.2762494446930782,0.5058832255027204,
 0.42145669548856257,0.24388940906663492,0.3086297572717628,
 0.37475803021310605,0.6802235304523182,0.14507891457905042,
 0.40465718257544336,0.9644218703785328,0.5599658928431875,0.7930184941161834,
 0.6083720739482039,0.4887489718266196,0.9173438041162923,0.26882787784864814,
 0.9752148438421714,0.05466089412216857,0.19926913679980274,
 0.6965738232109774,0.46624059762347514,0.5949910976524924,0.892190950078589,
 0.12567241548969565,0.773871510343509,0.07749629273246872,0.8528689023079093,
 0.4596526728796636,0.028433383352428176,0.022120512147065563,
 0.5085553363056392,0.4919538756976335,0.08884866775081157,0.3502546843612653,
 0.32268265489021264,0.8816038750492388,0.955889628081018,0.6857059976000136,
 0.09257693256662392,0.6511829314603789,0.0859498300082363,
 0.20292120490059684,0.9871296889182978,0.18286964499606795,
 0.23684860727552282,0.16654702492943074,0.8735236073225271,
 0.5946774852742842,0.3106727289982475,0.6922201969738775,0.6033688190859315,
 0.5140152018700938,0.941305611875821,0.6361507135410976,0.36532809332925753,
 0.4614088189516542,0.802675285554556,0.4561331958844671,0.5985818403031012,
 0.353565092661125,0.01766666325482058,0.014697280403153323,0.321831382889162,
 0.5298537464498276,0.22182253283589848,0.7585828319948755,0.4279092119729082,
 0.48547399005722736,0.8277950967819288,0.515774337831828,0.3308754506150481,
 0.860866358540156,0.022832846539685203,0.27547967943662544,
 0.8643744857590585,0.1518566736412944,0.2629823202545961,0.844605721481153,
 0.5428172077074639,0.9999478406541474,0.23733277926915541,
 0.01333722365666834,0.04151284934395272,0.5486999942067248,
 0.0764977090099529,0.17352191801691785,0.9193323368577602,0.9299388470790965,
 0.5803102129005479,0.8876170353860001,0.12845436304913516,
 0.014848952985404007,0.7126480861919172,0.7881381831084667,
 0.06967806301051815,0.17954446593474915,0.43961062611582225,
 0.5849925124675752,0.7566155706288138,0.04250341408674663,0.997641176142271,
 0.3295523420564419,0.27147674283839995,0.9167181276188298,0.3438239520609252,
 0.49237846700311727,0.0928184001906025,0.18527260218632735,0.379354514745319,
 0.8912008929426459,0.17549048312373094,0.5746472523122524,0.5384118385299376,
 0.4046626590543312,0.7623773489429191,0.7750978569089516,0.6161440848807156,
 0.5857879980186848,0.7181384783566542,0.7673671101448309,0.6461801480599678,
 0.5644494116196038,0.15120742498142847,0.0943875060157433,0.7294179851558337,
 0.5830827394849273,0.7077424563664377,0.5197073708293802,0.10584557594483024,
 0.5298535783874331,0.07415515862798294,0.276964012424448,0.43253228698084045,
 0.5531380637832839,0.02963114017671631,0.7472528092701978,0.7453681733621005,
 0.38476263892756624,0.8653406693777711,0.8336283276338121,0.7600660759495319,
 0.7222985912850727,0.03371156974504996,0.4791191971040254,0.9898993142343412,
 0.673416463377411,0.08959621913635696,0.7260047307111084,0.1393436336161874,
 0.9534573620469906,0.31920209636273555,0.7806525938941552,
 0.41238928413129305,0.022504773864517325,0.10108590752740587,
 0.9110653889899537,0.6515379440592034,0.9755243330810947,0.40450418887103234,
 0.32759896262559285,0.4117236545366384,0.27928458975436565,
 0.7199060450751243,0.48548726149993726,0.2787147969645767,0.5451828489424667,
 0.4668550910474656,0.9925713889470352,0.8153253638965489,0.737964937057375,
 0.3132307028281759,0.28611249572427655,0.8688345869032988,0.3540189662073695,
 0.6198507852159603,0.5267691864030162,0.3508921764155829,0.3995152780803458,
 0.6885107715644195,0.6029958450901007,0.012977826231453049,
 0.17586135261405333,0.3253081771415677,0.28150192539145436,
 0.27605415728875127,0.5132132308952911,0.46033374532755844,
 0.5637695573431122,0.7934581516158914,0.16675844636706771,
 0.30041476014633306,0.563806815939274,0.7178663224116275,0.6511608516120899,
 0.9497429684394908,0.07692281971188586,0.052078163899558105,
 0.20997799284217178,0.33951751640819217,0.2900664529526036,
 0.2260282069779327,0.31619898243947686,0.36979186754760474,
 0.3233342701219202,0.8738740869158249,0.03185470588290562,
 0.16555944712397386,0.9083037623089913,0.5052583303630721,0.6661203195092109,
 0.8639899245764819,0.24897048520243992,0.7773212134529858,0.4971500002695527,
 0.1733301025499897,0.22075461609905245,0.6707109464812685,0.5326980251796929,
 0.6451794035780896,0.08211873334327824,0.6143565180983245,0.2531080621330668,
 0.4492243496107913,0.6041363729210363,0.7561741732703458,0.519105109612579,
 0.9169352644340024,0.8420604707463096,0.039362609126163295,
 0.6517119363426306,0.15589058321179672,0.17154296741792208,
 0.7790701113149757,0.8489031235968058,0.22551496192486176,0.9974212172269245,
 0.0665720432912158,0.5577157893870843,0.7787199976978232,0.8170918783759213,
 0.010379289748305043,0.08183719244986809,0.17646794367430052,
 0.2545373657718284,0.5812713732359359,0.7110861400116103,0.1964159782065893,
 0.4980131135228748,0.353955129750672,0.3151810949958863,0.7852568535134831,
 0.785023449322417,0.9218727041712714,0.10990170916227948,0.647040988976563,
 0.4868941424106028,0.06131646683444858,0.04002221932741279,
 0.5002285200611172,0.014477675056199013,0.27786232859917437,
 0.7151731864586498,0.08707933737208817,0.695003670810462,0.2746865580647191,
 0.25603472063160115,0.2125382974439579,0.35725863049263107,0.282370025523919,
 0.5233080908676944,0.9263300594465654,0.4664100868142118,0.13580190366136358,
 0.2379268084782309,0.43395249686566195,0.03351445810288345,
 0.6390997801006851,0.7838076091479667,0.4429330083085411,0.6268398989303939,
 0.45260682224436644,0.45864250807234974,0.9009502455782719,
 0.42469222857466116,0.48566723555053803,0.47969740664762905,
 0.21664017917328748,0.6738743508971822,0.018562416106560398,
 0.19237614752084697,0.3692270843243157,0.22763963131409004,
 0.25891867306984606,0.6951542208592346,0.23045236724411478,0.967391813133619,
 0.3267035589207101,0.07652796318927413,0.47791538490048713,
 0.8775471606402483,0.49182244681111786,0.3661707367946061,0.6806450634115668,
 0.8813452372664855,0.32602945628086377,0.48661093914539644,
 0.09877303648882396,0.6110210938334364,0.7761628051418517,0.6486226563239257,
 0.2396962284574521,0.49410792507900325,0.5797638359573772,0.4273611198837012,
 0.6341293178046661,0.11266527250276248,0.14490049400005145,
 0.08977038951422744,0.1157667544178469,0.6817905735223541,0.8928508353770758,
 0.06535855529024792,0.9721767045866344,0.08652831586854015,
 0.9253115219494921,0.3499882816335843,0.8199086317021557,0.15936648919819485,
 0.8969141563570291,0.6362178699474317,0.08561202239686028,0.8980590535933661,
 0.5600843956086021,0.23655257127663243,0.7259099539178813,
 0.020209258358420912,0.28314619246093675,0.5742909019385172,
 0.41394802576180156,0.5387362559022526,0.29763416245881136,
 0.5283946003261834,0.49023461034203086,0.46717779793818115,
 0.5888618975311967,0.38642789916683584,0.7802148802401798,0.5655998568341746,
 0.22729117994948633,0.08152518279749488,0.16291294843919046,
 0.7784705487826813,0.09978562438056271,0.7736541093127416,0.5611663558386007,
 0.02263756539046491,0.018573967906380107,0.16848514527276004,
 0.35811427469789026,0.27645287978533384,0.845683344033664,0.4384971415012323,
 0.32946369589653635,0.8960916946557483,0.690108384415301,0.3318901493024674,
 0.6376459482100411,0.18342679389871686,0.2578812847598453,0.4788971323239326,
 0.43640452695329546,0.4479978365575844,0.419127120653772,0.47209421760593073,
 0.13231953116748774,0.6293059021214196,0.08627686088327458,
 0.4355256102270253,0.8806342555198167,0.39521543650892466,
 0.23628820719272892,0.10164754628385053,0.38389480587470726,
 0.696253313525945,0.37996487225128783,0.036823468258886494,
 0.14160490302824047,0.021260608201940756,0.5405419323590424,
 0.19100071747035474,0.10831347952121917,0.11340929952847478,
 0.4786981192399322,0.5808778960199549,0.21521933671941618,0.2970377870723635,
 0.2286555226350394,0.7124478694117831,0.5058513503928797,0.6080644789121504,
 0.6324881737100407,0.4854908452464083,0.9991092747244159,0.38799784500977474,
 0.17026263062182356,0.5261425479863749,0.5839558099618577,
 0.04466440355597834,0.7212655175429075,0.33011836402684724,
 0.5301483284882053,0.9878226349722723,0.829838486299461,0.7239338706246772,
 0.0957785088025741,0.2555701449158524,0.4731971387335171,0.22597757574463528,
 0.2853851599356968,0.12153634969571969,0.3464599981257488,0.1311658286768309,
 0.20204241485102836,0.9752690382624842,0.2673019428934744,0.5652795748153923,
 0.9237787358257133,0.3832164440049516,0.15940565343653068,
 0.05907516546523994,0.8231730854669196,0.0322226655033192,
 0.48081384025442453,0.5200934388776197,0.6464581371510387,0.8383549103371882,
 0.5548537604920071,0.9952100617929325,0.05208278754965945,0.6115002420743743,
 0.3747330373524985,0.6773382504862772,0.42167204622423227,0.4628897422509398,
 0.6299540838704375,0.2691509468275006,0.9476953521887894,0.8315834802592359,
 0.9985383591967706,0.9569862931404094,0.0766041345781674,0.3573754140292248,
 0.4760220632680663,0.980344782902342,0.5066570928968945,0.6221228688246471,
 0.9028730716325809,0.4492112183852328,0.8671452033918996,0.9307521380240851,
 0.5700375468754947,0.14570994965700157,0.35061692304369774,0.808799019059927,
 0.5124209508227177,0.9712684757235287,0.41918768327438904,0.0885627809811953,
 0.48738036362763193,0.33897181343840443,0.36513368999374995,
 0.19732200638627373,0.2517704121420118,0.4282148367320727,0.6829266408817518,
 0.723585085621597,0.4597905349865692,0.6614809418248181,0.04850681446269456,
 0.9094751847504968,0.3077655554829749,0.8643642216741135,0.38837747934355715,
 0.03158532022141114,0.21620392500695385,0.25270455214923915,
 0.2527821911364835,0.7540631888563674,0.7463816830123465,0.18033851499628795,
 0.9017276797368715,0.8741965121947344,0.7370834514343125,0.35645039260239253,
 0.25924916594999425,0.7749575111894323,0.6178249017735683,0.3639104326442437,
 0.40279461233668024,0.1265905730073198,0.8937731018917657,0.5478772584037236,
 0.8505345322074372,0.6210542582827026,0.2289291298381907,0.2935368164311424,
 0.8363406124314471,0.3319130780881465,0.6616040022913433,0.4961160364313264,
 0.15587483805859836,0.623696818536824,0.5207788436372498,0.4135435835896193,
 0.7583847992941694,0.7649032171997094,0.7120791469549764,0.9590024096835538,
 0.9104564245647383,0.6712085303913441,0.7302685076402022,0.5942810203464383,
 0.840817362829032,0.9839705754596064,0.4750028631544442,0.45287792943572425,
 0.5407663949937213,0.3495462327798483,0.26814711157589355,0.4652135458162501,
 0.8092577821053875,0.009198201918324544,0.7564654652987792,
 0.40591081835164733,0.1322078482289052,0.9422298870024705,0.8186995034031425,
 0.4279291151884016,0.43236321750334017,0.3613655923814041,0.4559453701298519,
 0.9312840215935979,0.4380894999083582,0.8763449152081593,0.5770720323473899,
 0.43861305655920624,0.6708234120359726,0.024741672172904794,
 0.7773769380641674,0.1229827291284844,0.09435220021871094,
 0.14221015850652285,0.1472010854201119,0.45323872167851187,
 0.2703255397591119,0.32110790619540497,0.16021614230408254,
 0.5822819203161376,0.6233899863490038,0.8625184119027174,0.12577906768243996,
 0.0062838459551211034,0.7219824421963253,0.2511495848499865,
 0.8153737870791954,0.416818015360754,0.3641091621107647,0.6227840517879156,
 0.6974100070969728,0.8665245844934594,0.4317950653309329,0.875153970555065,
 0.7373936785928934,0.7147358284638146,0.16530446996774661,
 0.33289003868020295,0.4115632562043867,0.9966860569425896,
 0.03601295226024581,0.8340524898885189,0.1528092480404688,0.5135364025318192,
 0.8641504118263352,0.8401355939848373,0.39577364606213106,0.9995959082749988,
 0.10532119757033942,0.8412041207209084,0.1590972609230734,
 0.08683635071853024,0.15213185502875948,0.44610226731652614,
 0.48567255489390837,0.6067273566757391,0.3208116700669753,0.4259208211753419,
 0.5665250246548967,0.5781510904238737,0.6193492977070107,0.8756907929761543,
 0.5007461431792034,0.5799599972339143,0.1769527817839861,0.5862199027439005,
 0.9596221734251726,0.1343667826966486,0.6132820323869702,0.7496301735827756,
 0.9326770800611507,0.6740527733067644,0.2773877658735192,0.6368164712639484,
 0.8706978438441022,0.5589029880442183,0.15720956795486474,0.8306745145275419,
 0.960184655844084,0.5264685238034793,0.938947076422624,0.5786395652408287,
 0.6533926570243516,0.7925541448098486,0.7572757181283191,0.4097234745542958,
 0.14503630495098052,0.6531078576584439,0.8271915700327209,0.8097951613470422,
 0.8458114559668032,0.09546752172918971,0.17421767402696497,
 0.6414669748611451,0.3790771023417462,0.7128949231641016,0.6457261201974379,
 0.11643096067846226,0.4261669563898135,0.2984560774043473,
 0.24584980615315555,0.5220094454001407,0.3086100721177053,
 0.35405676389157636,0.5718425877559575,0.7373821682775837,0.8960103943653113,
 0.2562234217892957,0.7075247766578275,0.7858754499072392,0.8788289787872814,
 0.03550243804874453,0.512785938736012,0.3728433089096014,0.13737111238914956,
 0.5914613576444715,0.9032413436998538,0.41080659786907026,0.2983392324401216,
 0.11120271390912861,0.6958163265392143,0.4836504160713485,
 0.13206480304630797,0.27242634179441616,0.5713809641055341,
 0.7020852543629856,0.5522203256341552,0.7894325688705174,0.8085055273918318,
 0.16800726761818574,0.4560777610392006,0.22416961057430473,
 0.9457830479523971,0.9220254427816671,0.41994248291958847,
 0.43675449845622416,0.8591346524870765,0.07448417865641788,
 0.6422815569357525,0.5102473504882251,0.8116282880063747,0.7982291311343697,
 0.8414486276939634,0.04618500879473686,0.5278201641857994,
 0.18884132018428734,0.5768939712438405,0.4445481635098065,0.752283379026839,
 0.2903099886091033,0.7589074958863697,0.5911680345050432,0.8714703980343366,
 0.6938637899992438,0.08875767048409466,0.3165127455118887,0.7398925298782925,
 0.1758280714034035,0.779642297600329,0.2850006572853586,0.709782749049843,
 0.8692721961273353,0.28125588337694996,0.36490394347640853,
 0.9863496625233401,0.711312351593554,0.10905634622110982,0.9957415851326286,
 0.1838029226786343,0.4979186735902552,0.31418698571022374,0.3859972637336808,
 0.5718712409473458,0.6357856516570702,0.11788847093643573,0.8824272095705964,
 0.6271993351933081,0.25783740032931635,0.15287196798665048,
 0.9180797071577714,0.6276337172210377,0.40302693220525376,0.8154998685206963,
 0.06404240575999887,0.8630666137296713,0.8127893787933581,
 0.11901782247039383,0.8501870030750254,0.819963758884574,0.9911316679877651,
 0.47387327786826405,0.5669156791369098,0.05372220056044985,
 0.26368408696789924,0.7920548117048806,0.13713202670964186,
 0.11945134659157575,0.5462786023094879,0.877393178472295,0.3989830767363114,
 0.6152010376175555,0.41619226312401225,0.9982115581442851,0.5312965449561727,
 0.41354605079652884,0.004312701614691661,0.505079306790704,
 0.14185112912177278,0.6556432171710398,0.7802698980210848,0.588910592602982,
 0.7856771590121022,0.15615453363184806,0.32946064602446734,
 0.8397375898542818,0.2459656890008346,0.10692014301156905,
 0.11606084041090792,0.9530899173266041,0.2192425582976144,
 0.06210165473364215,0.9957111857868934,0.05597618132542237,
 0.6469159514980635,0.3138545676194501,0.37527123343858193,0.8945846854180982,
 0.5482582757645543,0.9626079004306406,0.5074616382356214,0.8662167986933599,
 0.6394750632836282,0.9158863071881622,0.4855117956467305,0.7895825017414326,
 0.9925934613807017,0.3016386166746736,0.3820991146844642,0.7175792753002204,
 0.9494468911866877,0.06945852977760392,0.013966376168900618,
 0.19479596414340383,0.009321580661401319,0.9101832248249905,
 0.08233895314917694,0.5846061628072127,0.9674053051452757,
 0.44317923282912197,0.038933092412808956,0.13963566482378398,
 0.01108780879786575,0.9425526991699944,0.35935241053887457,
 0.3107294789075449,0.10804327417246695,0.5888182902771327,0.7566680058065788,
 0.21064716312255427,0.5312917359950974,0.8220058181820189,
 0.24103759071177666,0.35819754261297243,0.010660528140454484,
 0.47671538186452467,0.24970661607360922,0.6049004066059954,
 0.1484622063864608,0.11830662572003181,0.8892234362101614,0.7018140016242406,
 0.24892797507277065,0.09549867856823369,0.9559380934894232,0.635436569338908,
 0.8761574051363606,0.5857265279173566,0.37588469908334343,
 0.37511551761362394,0.9890029166881745,0.14556287183468697,
 0.6660010654105969,0.283186163115805,0.5146529873991355,0.3433587791712217,
 0.08412846880347213,0.7199692516087728,0.7592823938502694,
 0.12711476917113895,0.7168328348571639,0.3573763129747878,
 0.10007903170543442,0.44953505923618853,0.9680335571720124,
 0.38258425482889113,0.5550605702980052,0.9751296828898103,
 0.30569413700647396,0.6770003158339496,0.7205446372460865,0.5781351425195443,
 0.9266296786978387,0.3597528638478351,0.08619303556696833,0.831089197821488,
 0.8794322642952391,0.17001728550810602,0.20310319890635553,
 0.12378339291722229,0.07046257107699727,0.7022563637744255,
 0.3123023326324771,0.5198903586964526,0.4860469572858067,0.6030108041410517,
 0.49430554412705163,0.745935970364695,0.16715382345523733,
 0.008840032950465204,0.7844729528671158,0.26808922566547455,
 0.14134436359230307,0.6857760218061564,0.29326901051396914,
 0.42363432704208537,0.7946025976210558,0.1373510362783662,0.9485526503570079,
 0.3607869512155041,0.6587280606880058,0.6655462204465886,0.9303044609377836,
 0.234369662225953,0.9075103041029572,0.582227486873542,0.5213693814510957,
 0.480170279305009,0.6768035597668599,0.023079400599270716,0.0598046681347264,
 0.410205295313149,0.5968134809060945,0.03590199394189486,0.9099348301931054,
 0.1691791538934898,0.20123412234347615,0.5059436124246169,0.5230202143991322,
 0.7223962789383481,0.7857622075636199,0.6017274436383192,0.3854512803279839,
 0.7313152753852702,0.8910300079541317,0.662765155533224,0.12162034851553605,
 0.8355834360171405,0.5106505256036495,0.08319370685064842,0.3885656107543457,
 0.5402172278926426,0.9513795936795791,0.9312162734038498,0.02102989754411322,
 0.11508900978070447,0.2907814881358701,0.9851834693464959,0.5025704925599451,
 0.8593431246524192,0.06388449981672761,0.6406999635655403,
 0.47192222273415596,0.36756589463676614,0.5242517157860294,
 0.9742092669600284,0.9316588476512876,0.054666337687014344,
 0.5277939432950447,0.6033720999150753,0.5222487674837631,0.34204679057675913,
 0.09809474867831292,0.21107177882423,0.6689967218284553,0.04469962591145926,
 0.2758951574824612,0.6068192865336637,0.06525724093694041,0.8028160279319887,
 0.4308562583108926,0.4639199611237923,0.7209267074671932,0.22729014811931958,
 0.8434002482471956,0.2683964596461016,0.7792347243041143,0.7584991631866103,
 0.702890643337462,0.2790758149252033,0.24663426975745373,0.2541238749145006,
 0.09225757499919407,0.41547325910636146,0.5518390844983305,
 0.11622688291213312,0.43918966664961834,0.6825025723302116,
 0.5464498586581155,0.29451306145079637,0.6804940901328509,0.9225571326596448,
 0.2026597215691529,0.011477455823555749,0.630449115777975,0.6109283067170206,
 0.9193777404283398,0.08943508460001903,0.37443259430686426,
 0.1756818222563412,0.3426238528502381,0.0891369602449128,0.17146419500662113,
 0.6339208514541921,0.20061774915618624,0.6968470173303589,0.668125198888909,
 0.9268877944527183,0.820033340032961,0.8636877110291922,0.7706178314645885,
 0.40123408475522604,0.7306723917533055,0.0791895635269555,0.7639852875725375,
 0.4025608851656304,0.1728242328488384,0.7591121405760066,0.3795683051867722,
 0.4732013417414489,0.8691329110149739,0.9633408701213745,0.8169013033843276,
 0.9637381682473355,0.27435894522095294,0.21764721248545316,
 0.9240544480217329,0.6186205244975432,0.761905841260396,0.5761014226286523,
 0.6066090970909886,0.5427963207320384,0.11677043664512343,0.7995834882850938,
 0.9055557226165803,0.8804059684513149,0.7589953813624473,0.959915472829241,
 0.5458039635439462,0.548672926137149,0.09695273746614941,0.27719224938202414,
 0.28329527344116223,0.2550688355007138,0.6686370296312403,0.3836830593653979,
 0.2348921815110302,0.01734175949867356,0.9534830393113232,
 0.15350130116902205,0.9691971458702062,0.49480753301258107,
 0.03327057636210384,0.931923859830316,0.7616690219730761,0.9765476200398677,
 0.835986154616712,0.7174051637292139,0.8357348868988097,0.9383964060575437,
 0.6310010550028908,0.8496232673380549,0.08773588857247039,0.7265755356991214,
 0.13981607291575993,0.9211394745177374,0.6540940689531215,
 0.17266833953891925,0.33735974250149703,0.5472515599034298,
 0.2237683389059537,0.997032076479134,0.182776754368971,0.5182768904949935,
 0.5277016391760276,0.8627225623537155,0.14915155746009712,
 0.16478954325403927,0.3947158074928129,0.7646783055676759,
 0.15621064267335083,0.39508908827195965,0.36505481933173645,0.62638143639447,
 0.35102257143835747,0.34886357245663624,0.3800021470763514,
 0.3071130617840422,0.8012567225818624,0.30457445728026156,0.5140345945367167,
 0.9638949644393155,0.5662377634700368,0.4990427577128429,0.07813473856011977,
 0.5716593939122885,0.6115857103390945,0.3729203925022405,0.1812819394642139,
 0.43525512699336444,0.4754029698841137,0.7724733124364335,0.8474420628470292,
 0.267845266056412,0.6451439002892396,0.6750978196090506,0.734612335784324,
 0.0260066649148335,0.7468416679858227,0.7901633162989435,0.43573693402870173,
 0.9921359314379896,0.27686561506098395,0.8877741781588893,0.5706699791864853,
 0.4985466591036144,0.342328572017769,0.95684258108974,0.46531240034092214,
 0.704463479757519,0.6751107869140671,0.8373034181187021,0.853200576451284,
 0.43312417621874655,0.7635509958771839,0.06070797971149111,
 0.15507164545952967,0.4417651213687618,0.11789504906855364,
 0.6085785860313901,0.019671041786859078,0.48131959668498736,
 0.46776351589805865,0.628393239643922,0.3434608592023185,0.34195880216405417,
 0.05455118339865783,0.15849614721430794,0.028521959048103396,
 0.43591239695398276,0.40086812821231954,0.990609224395137,0.3720463790369771,
 0.21563968941702494,0.7862622678940754,0.4136784593335405,0.7487801531920567,
 0.10399289204986573,0.21819111919913292,0.18944455170020646,
 0.9977871992610541,0.679556042826022,0.6684193239832903,0.7902880911016859,
 0.925434223166384,0.7282768110866914,0.5296245636244156,0.4716957123675698,
 0.8750269013986645,0.168700332544243,0.1636024909278101,0.6380869734522543,
 0.464461821854275,0.9937067955787698,0.7098137973031594,0.6917587005488617,
 0.5657224447210433,0.9667624365918244,0.015935402197326365,
 0.4580060804128052,0.8334398675190774,0.043127816721868384,
 0.24423114475223473,0.5257892787836046,0.3442877079770603,
 0.07020646248761553,0.20441893349820617,0.7808298522686543,
 0.3672866773225122,0.08382644690545915,0.6854076760849069,0.6339907394527694,
 0.9994587092885024,0.005349255265311603,0.5456258517700564,
 0.09280124028265357,0.7172926719118402,0.5901512926352345,0.9434561031278773,
 0.6080392692173401,0.10917871757330433,0.9386782517093286,0.6146990129033388,
 0.322490179351444,0.40259276416969536,0.2227120011239958,0.2402375776443092,
 0.30290778306464317,0.6613308375213578,0.9317304512531706,0.7760984143707026,
 0.5527292649191368,0.9040957410831545,0.2742602384621269,0.4175284778785032,
 0.033285829092740804,0.4124185298732739,0.9808870766595467,
 0.9852745365694726,0.8191199034907034,0.6346687215310522,0.5890935495193461,
 0.8359514775329971,0.7468144234252825,0.7370615590957132,0.9475043331863033,
 0.21801180909264894,0.08256536604800568,0.7819486137644409,
 0.6810853613538407,0.19614933478467722,0.5898162909651619,0.2702123598718126,
 0.4892783767774612,0.734219882534182,0.44902026850428556,
 0.005431184198834149,0.3684193524981265,0.7488764420650639,
 0.9465374058190434,0.6396987775815891,0.8903551081551059,0.3378292378402208,
 0.5431021202216884,0.49739770142035,0.9382362975667526,0.7481589652992857,
 0.6911169296913746,0.33742196418709014,0.1428921027493909,0.8482551055410614,
 0.30499318576545686,0.5153349534777908,0.07732052305055881,0.76853415796509,
 0.6150442302770586,0.022662690292260312,0.41383485565294864,
 0.7566007158191662,0.20520630572104304,0.020347358183035014,
 0.034432625476774725,0.8310668376717008,0.3379038811079442,
 0.027976977439983286,0.14186985610467517,0.9703944908988784,
 0.22850276731839625,0.6183230314978779,0.3109101982161203,0.6951481475948743,
 0.4759090038846837,0.309097543257119,0.9891099837023428,0.5439412139994251,
 0.68626046371515,0.10434677486080723,0.7205853243960534,0.2794469087077398,
 0.5037046650134277,0.5607770050097435,0.4562463826759571,0.01341334580984177,
 0.8327299482454593,0.6579534339186791,0.2729871018301755,0.3171610545689588,
 0.014152705888052619,0.8143186549515039,0.5885675616620719,
 0.5157279052651733,0.23623635498518203,0.8100086463415241,0.5351075619610727,
 0.2317689476566399,0.31608326185615354,0.7764034328153328,0.9717026018135381,
 0.2675809210959239,0.8547009988133761,0.5381312200873941,0.5263777567801644,
 0.25735436427161884,0.33525785573472766,0.12300786640244898,
 0.9050801267604061,0.4454679684537669,0.6809750755289744,0.6734901962404767,
 0.8916077745997839,0.8511866737810044,0.892658930890006,0.8813863654043774,
 0.12402434331553702,0.3936990496853524,0.823075781923964,0.1304617992386904,
 0.24717230077878194,0.35789329224649147,0.8070600847851711,
 0.019018075479097396,0.029724720975042973,0.7201563748845475,
 0.05993593853214041,0.623565895804936,0.7039362825257761,0.6578253655337983,
 0.6056616073713623,0.6815010762775235,0.3038263467259139,0.7981892347870821,
 0.3213541486638475,0.7817981391588642,0.1283163998748411,0.27131779311159476,
 0.7707502271394922,0.8837132047049945,0.2716795850910899,0.31337576138091805,
 0.5625182773212623,0.08570093388402855,0.03666707633703403,0.082974288058975,
 0.6180466602333842,0.7500047591187836,0.15677882896568307,0.9395820531542088,
 0.9982994676715723,0.6986796880191524,0.854213471832431,0.5101555825343039,
 0.5654619271496283,0.5815974114158703,0.6280480179710368,0.07943690791732783,
 0.6976807816577983,0.40905898993590895,0.8635667493427642,0.5082413955414515,
 0.3175875820706777,0.5513425397213789,0.18472118424450779,0.4595005362352659,
 0.9469521850882447,0.9105505548899897,0.9294979090856264,0.24638711676949654,
 0.9008982779230239,0.3803696243727831,0.39976646461551535,0.7863963886016907,
 0.19785786277970652,0.8688943197062797,0.8330103736132801,0.9550388493025066,
 0.3604748839954197,0.11340478294984524,0.6591297698369021,0.8206543671681727,
 0.11107650559997795,0.2410903877170465,0.8263142364183839,0.2272265195299249,
 0.7310155785483388,0.3010198128666689,0.2940544952807078,0.7991948435445348,
 0.8511210178664115,0.02247368041497888,0.7171070735129444,0.6729045185967156,
 0.015548901691404216,0.8611576719782491,0.6004437315226714,
 0.22190533104442645,0.48029643969727465,0.19970294279562406,
 0.8360923677579023,0.6781441417648165,0.4190717031427469,0.3355303518637347,
 0.42274826989928493,0.30019145129031155,0.4650311249090675,
 0.9377429557243622,0.3915288627886715,0.13390193677496354,0.1710551801618837,
 0.4975023229827751,0.06241810887001564,0.5103489687844327,0.5661155894289265,
 0.34242733258985414,0.8440062646485303,0.7696297992453571,
 0.28880306815875645,0.549063742517059,0.31168442818563025,0.2663988731540853,
 0.3363515661838393,0.8357903405928958,0.38332954069723,0.07910579956023878,
 0.4092014826692558,0.3463769124653703,0.7616657917135586,0.1318886469147439,
 0.8937540837659875,0.15571990634218957,0.8981589110341432,
 0.06522190119202564,0.6374506139161905,0.7941548725264637,0.5661300395634318,
 0.21308781372645025,0.00455656351854028,0.03840642735869282,
 0.102793034187176,0.23106233779874252,0.38810345871761487,0.7112825922230759,
 0.4259584623422279,0.9018595854171173,0.8827663375224063,0.27732627220415196,
 0.6820451798641165,0.37628182485367667,0.049404296938958536,
 0.5668519592344217,0.5885210699609754,0.2003382112128782,0.44316293991681066,
 0.6051785541167131,0.11679562123666098,0.08575899468942538,
 0.9108904081655498,0.16323232881942196,0.6574033576312259,0.7257157381366287,
 0.790196818753861,0.029987406941054284,0.6771885812123564,0.5209779989001263,
 0.6581839613009189,0.9351858763721462,0.8926613081563279,0.3737622298311717,
 0.38057248298799584,0.8976225465973251,0.5910658731365709,0.8114705679678106,
 0.9621634231052583,0.8100010073780801,0.2733820187028738,0.792495476628887,
 0.7097275901734582,0.9293550116133361,0.38655489105943897,0.5387986584135909,
 0.2899241933197292,0.6988011998016411,0.1392296578206934,0.49302668321239795,
 0.5511791317154275,0.0036713662959622084,0.4798550782871013,
 0.2759325460457013,0.70634844921041,0.5939265469824125,0.514588492805464,
 0.42793903927093224,0.8838739565504468,0.2852599989466672,0.545973112020713,
 0.5873812597493869,0.9352394553231114,0.9985294760831622,0.9944268764297519,
 0.522458705308596,0.7803846703156114,0.017966377108183718,0.7689209524563553,
 0.0836039591718748,0.5714638601948454,0.556005127824882,0.8863742100261305,
 0.1814310120736946,0.7143983766225934,0.8435158199345658,0.827265157970451,
 0.8149219058467221,0.03414335471393948,0.8080366406308709,0.7262495616518119,
 0.7730254972783258,0.6214979726214437,0.08896877684758209,0.814743442379113,
 0.36360906720832553,0.8184682231290266,0.2093526460777655,0.3237681961263724,
 0.8388814937154875,0.517551164240658,0.23046586837843763,0.98650602413305,
 0.3563189710249124,0.042178986375408734,0.7175201846615096,
 0.7045619630626184,0.268315136113383,0.7204160017436583,0.586724752796737,
 0.7884809182579106,0.42566397516626575,0.7634018062058476,
 0.0044544781071946815,0.9033859125058347,0.735539730125584,
 0.5014787703758279,0.13390163716746695,0.9779048912240961,0.8720993158662613,
 0.5140540335951584,0.2681019596324238,0.05327683647913928,
 0.41065894259547864,0.7753225845530273,0.021064171971397427,
 0.49059705667553966,0.6404250098714399,0.4208224294435654,0.8759831399204344,
 0.6433180167956591,0.7840314558277948,0.6593223074074801,0.3149898018444004,
 0.7751212655522589,0.6950263701752518,0.42299529709278816,0.7631713928384403,
 0.11716126986709885,0.2850247221368526,0.9856220603745013,0.7764827988787827,
 0.026399240210922326,0.9377344293514541,0.08616258876627403,
 0.35171323872437843,0.6907916633884359,0.5892390553467661,0.4326683129936354,
 0.14880936407157053,0.2004165567419739,0.09345213500449678,
 0.44425686293284605,0.25918619122198405,0.9815800121894555,
 0.6593018269413774,0.3422691547791562,0.918861731287544,0.08279387836994312,
 0.9241484256575222,0.025028161078320954,0.10565280179356062,
 0.2172080995878125,0.2583010594374743,0.8959340096264232,0.05121150234262706,
 0.9188706289475859,0.14072090731093567,0.09282576654100472,
 0.7162556746509259,0.567428936737738,0.15643627352633882,0.5301651417474892,
 0.6441977388372095,0.38744651763869165,0.5706391847358847,
 0.36419995426696483,0.2666556017386581,0.2624127479282149,0.9191447155286321,
 0.3750348621462639,0.1810292569917723,0.16445316237111074,0.3346039530621331,
 0.7518172338972372,0.7577534572998087,0.11030344989649532,0.373887351091464,
 0.7146261052573729,0.13750290591922165,0.7605647277901497,
 0.09524949930296012,0.12300344051063439,0.029048506720216083,
 0.11669630774394668,0.9723304642677274,0.397671135498355,0.10280031139154144,
 0.05341987017122818,0.550341874609868,0.011586612943878305,
 0.48228723261139694,0.7474946306635663,0.9763083081686033,0.9050634739862973,
 0.21597714496220066,0.13250593079586848,0.31104161803138963,
 0.44569719040441225,0.4425703322819212,0.4672604901225569,
 0.05681422693223326,0.5850795859570169,0.6307744705458393,0.345180794982758,
 0.09609191291046837,0.1743250349743498,0.6516483372984647,0.6197447163228185,
 0.9207888806617982,0.9476576905467565,0.277764099981644,0.48112518286698824,
 0.7811810480479725,0.895189049354481,0.3282492653947777,0.15138192676935847,
 0.32025935934626604,0.5236397284223018,0.5784947149254397,0.607327018612835,
 0.08500776394561793,0.07524017221347512,0.6353628240889642,
 0.7456433663627621,0.8357757171166599,0.3959111100772398,0.06254807750503022,
 0.24704572211488274,0.30697717671270175,0.9575493459202323,
 0.7921223733183161,0.0031550139479823436,0.10339950529276631,
 0.9485825694266616,0.5678728462149307,0.41941370228938746,
 0.49187609968665624,0.9403969248614492,0.4181567268017886,
 0.09299484998384455,0.25683404258284614,0.6693910163851977,
 0.16494703596135252,0.6473906864275578,0.22318595533271124,
 0.18739038999807345,0.08007537257177078,0.5185818740266841,
 0.9990845343730793,0.24024293653302453,0.5982264669523127,
 0.19315295239700403,0.7321937153542926,0.6689597316927498,0.1594346866782781,
 0.5406783790361214,0.21343456510160685,0.32920976165181404,
 0.13801677108045596,0.7873435798423787,0.017668056138556154,
 0.5657891673161304,0.08038007273347503,0.43478214103372403,0.762531142100357,
 0.5771913266896668,0.5612780816686681,0.6673055373065099,0.24860406920558276,
 0.9630201676470698,0.14835928622494354,0.8490133641822529,
 0.12935330421155056,0.8087011328103193,0.9398575795495089,0.0870836839018776,
 0.05526027601254879,0.40420925101099303,0.6383224312033503,
 0.1787557744012246,0.49289038750997727,0.9109120208706581,0.4542824029373368,
 0.16235944433509408,0.7783502967997488,0.9385638194835245,0.7714686674123095,
 0.7868130707400377,0.23320660585819808,0.6251076526715121,0.941078959119723,
 0.8790744586725583,0.13292930139087833,0.9066237224871051,0.9341196231994509,
 0.15263137826894224,0.8445371841776291,0.365378855145588,0.10199658382691623,
 0.6340163429003663,0.01465059873707475,0.7251203356258529,0.9489416003714859,
 0.564792569140267,0.33651775104569503,0.7059194283048166,0.8184581551708613,
 0.5906660779183116,0.401685819691622,0.756661598535454,0.050662797177211116,
 0.21621984468462196,0.4587540435456655,0.29814960141242053,
 0.14251367084319166,0.11622367079411,0.7231497790523358,0.9649938989854079,
 0.8209499364571644,0.32536829564412706,0.9185691590862041,0.4479708882185004,
 0.8664091605360337,0.8226171904872799,0.7996577121318277,0.04461249242969212,
 0.09552623896626511,0.08831739554229168,0.11300145380025062,
 0.17992622936969493,0.38921011415297013,0.7709018603034107,
 0.8876043290436574,0.4803457285916327,0.47030690344803006,0.3597761772035907,
 0.35475722814514254,0.046278390528439406,0.6018407965284904,
 0.4023553807033906,0.08574680454103856,0.4614847619358525,
 0.10974417287399252,0.3456911726417158,0.4995578807948078,0.805921526377624,
 0.32213610119105796,0.5052470962761098,0.9221291703547123,0.7336654941577014,
 0.9293466434195657,0.09828726235299423,0.4816790492238999,0.9217762237517761,
 0.013090289092760887,0.9426018945956441,0.8467177762692172,
 0.5991849603055177,0.2637731352628645,0.3124486752472396,0.9898194456105703,
 0.9580976024767061,0.757010062604461,0.1457981661181087,0.9119785632716834,
 0.9707876276343193,0.5116892642863211,0.07957833191697705,0.4635410741812075,
 0.6745848474590557,0.38093325176079096,0.8337087285160694,0.3130935402677202,
 0.5261727791179992,0.736677803283592,0.4040245420166758,0.4096869366810276,
 0.8796097265334871,0.439791924768147,0.026585621160596595,0.391548677666124,
 0.03011750278074876,0.9698682194085162,0.6041608002116056,0.9748765104997105,
 0.6929843777690478,0.7929797955131848,0.4407246852200256,0.8437906697626077,
 0.5538287776636504,0.7940946468303066,0.16636395139173954,0.5619696067504606,
 0.2766098851965487,0.1773937805600706,0.93327696629325,0.16715113693875172,
 0.7486420246625117,0.9694657694874983,0.2519933423079348,0.22456652335883676,
 0.8446343449295775,0.9009671467211214,0.7131203771401449,0.3554596384249472,
 0.8875969574205529,0.7248989836162676,0.049787005009195306,
 0.3182652188934958,0.7888012847376792,0.770475961657299,0.9449922722708404,
 0.6954716709739746,0.8494612242142432,0.5742591249755487,0.4235672441334283,
 0.4867680421161644,0.32200672986434176,0.3782772473079101,0.671548543270448,
 0.3364597650985449,0.7558909675015992,0.16998475282765313,0.8433657846713455,
 0.8102184508048944,0.21977652335379017,0.6691183938522339,0.5218829983440784,
 0.1657198141171976,0.5734175190197357,0.037473605284922384,
 0.3721547343135533,0.6896655503609691,0.8402206076176162,0.6623083522609982,
 0.06085868134943773,0.19753278241091043,0.7737916484202396,
 0.6012680945721185,0.75736481364505,0.8059740953347674,0.8358919513810493,
 0.7215699650594141,0.08505052485348297,0.23349352187599548,
 0.6480651164031229,0.30373974036267415,0.44729536201672415,
 0.2854144879606457,0.5434990090570491,0.2759382530948217,0.26175657468302393,
 0.6971499242800006,0.45937258518493174,0.7452991556281845,
 0.25600377781276573,0.4037691093245105,0.8598436544221428,0.4372611847442892,
 0.6134020535605593,0.047438053205814734,0.17935561079937312,
 0.5656609903109343,0.33987589155321773,0.6436681985608361,
 0.17450832202043687,0.09084094587745273,0.5484886523324156,
 0.6821740098461107,0.19622928131701456,0.18221747664647192,
 0.5927632790601378,0.8551242564263581,0.008760981974572646,
 0.8974858043105852,0.43250473155395874,0.31724453784686635,
 0.6749349711738264,0.1566640867575927,0.5437277884680671,0.9278468600711014,
 0.3645456401651934,0.4504360032826664,0.6742349784789223,0.9709212150886686,
 0.6618085040264146,0.6687718946703,0.9962444384836607,0.7826343958472878,
 0.7598996900129994,0.22762858453091128,0.25330753004956374,
 0.7068725528939303,0.13147418805764632,0.27979068759402215,
 0.18159229834291446,0.4894276326021689,0.4288508945265571,0.1273089302408107,
 0.47884682946718893,0.8723402787022281,0.9004683367559674,0.3841392038150877,
 0.9511219983394256,0.8978848336395714,0.8956423096491071,0.730178389828972,
 0.8622055858701535,0.9709072364913598,0.824740696184519,0.4470177153606597,
 0.060731406789775955,0.1501098505792171,0.8940394009295556,
 0.3049796406641553,0.9879708779514073,0.36018332399778785,0.8739493656027343,
 0.5811362835908779,0.7274379906456935,0.2147751162857654,0.7052815634561541,
 0.7348544311523758,0.42400679735064306,0.6429895778723937,0.5740708374495918,
 0.5408069954976449,0.22327120606485806,0.6002344049651764,0.4383928417999491,
 0.4949407591065863,0.6263123538537321,0.4836497100283722,0.20895794316128946,
 0.7206332392772233,0.910945517680331,0.6276893975021018,0.11247317412754199,
 0.1637060701028924,0.7234563832034785,0.6734784324689206,0.20103246848954825,
 0.8610412246849183,0.9423288107048713,0.4820981936023223,0.3736355629298477,
 0.2046852724218413,0.7830954994630792,0.9686165314189222,0.9307348285588416,
 0.660351995687338,0.43444627656589296,0.5172860011215442,0.6389264389208319,
 0.5801881058508399,0.06481511489123448,0.8679551229189324,0.1209716603887454,
 0.4772879193658034,0.38171549349950895,0.6449746643217401,0.2415893789297605,
 0.7059313507291848,0.6373315552703787,0.35163600628175107,0.5606684217242701,
 0.4592811916838696,0.4941113445438846,0.744036839222755,0.9153319883372756,
 0.38080720987063554,0.6605852852159776,0.08839585836185426,
 0.9906302806651719,0.17410441996758297,0.5095046774020127,0.0375675908074915,
 0.3178741254888634,0.7957462917109832,0.5950966359638095,0.5599950288640045,
 0.21771215548290657,0.6996895865174769,0.9730569380252144,
 0.07750780779281041,0.009103121226896826,0.05844613427956458,
 0.9634986537105239,0.9666873173167289,0.2570753179136773,0.2908247429745803,
 0.836939734300396,0.5651888854975373,0.38706610682870846,0.4047023778800142,
 0.5235206341018289,0.5955557888307484,0.3020154521425913,0.9934527914650284,
 0.7165899023033094,0.02180868169339467,0.5518239020320479,0.8956017935286514,
 0.8654800895646066,0.06534634447884113,0.1327829203938954,0.562403294746352,
 0.12646896429032095,0.6590247477390265,0.09430727222414814,
 0.8095581002477359,0.5978109041101323,0.2650215645341868,0.2179125660733543,
 0.018032383143432473,0.2872001645751958,0.340555896215041,0.3439447368450387,
 0.34592899792290055,0.011367723753895831,0.4423071821133937,
 0.5934463075395414,0.5911100974096967,0.28632664013098674,0.5223612284901951,
 0.6053076141067666,0.5646856096365215,0.01069585382747007,0.3542816086609333,
 0.08059697980920943,0.3436772643378162,0.7567514875875219,0.6380382301864885,
 0.0268393444799927,0.7413903205617306,0.14219427882500357,0.3076183743604781,
 0.06324191633704834,0.5480880163328778,0.769014511767534,0.22715815742909395,
 0.17965004868325007,0.5938952548908735,0.4405714576571391,0.2637330600123571,
 0.18806647630094564,0.02599394387233389,0.525568204469941,0.7398670168363548,
 0.5039600768446769,0.6153314885958944,0.34803166661369467,0.7752759947145312,
 0.24558815907392906,0.9762907650477894,0.06570962125317337,0.339902936531858,
 0.4324005804463932,0.6798097082950993,0.8563962791853835,0.5395211573418253,
 0.3797153250729065,0.12113821567323957,0.9215058826623233,0.1551256609621804,
 0.8717575297955825,0.48505692512636955,0.726875196122117,
 0.001623780905077754,0.49100422249633235,0.30177176244442816,
 0.154837829285013,0.6335733558941973,0.5068573904869524,0.860102105986796,
 0.5467394766022049,0.7818083343490922,0.5800453443755341,0.086483883650054,
 0.8527542881981544,0.29842777871381354,0.3958450893400386,0.9357220895977258,
 0.6593650049901397,0.023281277191112504,0.32600870064842913,
 0.2830039645571131,0.8725109789329344,0.295097068855684,0.741985369807999,
 0.2844651147174312,0.6064374355159489,0.5210158991188786,0.858256429962976,
 0.5723454798645577,0.3858291637436908,0.9890606903560425,0.8104182866581379,
 0.18809733080972624,0.4480291968490613,0.692498409537139,0.8956490292327884,
 0.3955455809300583,0.5729848875806708,0.8552986481497804,0.3179517896226224,
 0.9922594343075364,0.6293755819127148,0.7725581182770331,0.35814635836253206,
 0.3457378435264593,0.8977460654366439,0.8157328772512693,0.33253263817541157,
 0.1566635246015725,0.006363126888323967,0.4183014397626196,
 0.6183532038165263,0.5437929634598258,0.12045507132531053,
 0.017074896789189276,0.08029659451263826,0.25841771135492075,
 0.45471138604269734,0.9781454753583293,0.58578317647574,0.1222840676643916,
 0.1760642117163832,0.6048181478344709,0.5774617192890406,0.9224629991323505,
 0.36540678512310676,0.8488904095772547,0.8404409508529421,0.7584919445353383,
 0.2521662391375329,0.07032593640106832,0.9142520171699625,0.7496615964634986,
 0.5477689700235341,0.8464279657228283,0.7144071022221447,0.6620695630728667,
 0.2181686741677894,0.3428449937886795,0.2850722334229836,0.25161601026501845,
 0.2138627762139721,0.41902832917719024,0.7782137229903643,0.8664863814428002,
 0.9102758182961733,0.13307831539227244,0.7757133838874138,0.4202451006235838,
 0.6639984120178656,0.5163104934454842,0.1233962556498015,0.1094111462612094,
 0.526095581670147,0.24384416326009994,0.6808066812423201,0.9830210403058954,
 0.11661992809236532,0.06185990163719235,0.5634697478099759,
 0.7531081842977743,0.3825725424104144,0.7853312326912044,0.12041892335283555,
 0.21960103648389317,0.032602860002818757,0.6570611633439772,
 0.6702296170766153,0.31310636109933954,0.6483425513405954,0.8098606733429613,
 0.07806451158128003,0.3000175901125627,0.1374922736854094,0.6240320909264412,
 0.40806161893446136,0.8325461590759904,0.9271547489658608,0.9167353595171568,
 0.703516154313752,0.5416008367664289,0.07046811871338166,0.9676098159154943,
 0.5210600607353688,0.8601656274003067,0.7947941998534385,0.39908343159050397,
 0.9054437509418787,0.13211001309194603,0.6167957626469606,
 0.07473796048123205,0.006868368264793312,0.5155033514790104,
 0.21980426738889236,0.9157020892714831,0.4575064268514334,0.7745354131013681,
 0.2833977423295001,0.20718892662160615,0.9153821760933203,0.1711704538815494,
 0.7664056897310997,0.9663109102997406,0.4004381394285711,0.8984376141064103,
 0.0480162800587981,0.8569403247416909,0.4395552661149036,0.9512894298477134,
 0.6624320592317314,0.10577688266492058,0.0745384349987267,0.1789818187190555,
 0.877379807231726,0.6946380940283313,0.027800504411974347,0.8587965017535029,
 0.732495213628722,0.19028945430095856,0.058722083505571065,
 0.8390659473152604,0.9965693005736816,0.5725251818475943,0.3343020029735083,
 0.5894045980258303,0.2545935849251386,0.6641348761220041,0.8941059677569854,
 0.01006472028045402,0.3444014569033659,0.4155935478099112,0.5916408362193241,
 0.09118960024700495,0.37247988713708047,0.7239615821640961,
 0.8190439814464366,0.07965813653239373,0.24671818272822532,
 0.2699743404790709,0.029096961780994413,0.4690881846956513,
 0.3931325849885161,0.9875183117030987,0.6320237786422029,0.2639537263787155,
 0.7283710032951921,0.8854603142808157,0.3226471447361753,0.16301490505588123,
 0.1546094694309914,0.7081643230584219,0.8718076349387182,0.0632982528596513,
 0.4439780929659658,0.3653426711977378,0.5112273307537618,0.7481950638711937,
 0.2658823871026559,0.5718320584859196,0.8604858608670863,0.6518158251751089,
 0.958383260121177,0.5710648466009492,0.7379817819300952,0.7299543977702858,
 0.5865253646166333,0.8500578241809537,0.3637902805017168,0.30355712042098293,
 0.4183803465607564,0.75877032009663,0.0768613508542878,0.17699052741947885,
 0.11070012168454535,0.14274356387183185,0.4146365134834922,
 0.7418572010983131,0.8365419824558769,0.3223273092656157,
 0.020318981336574393,0.20193963289527317,0.9988043337148972,
 0.4208160993098794,0.14529698879352304,0.5926378671804371,0.5462357880879412,
 0.05782279180070504,0.8931840579121288,0.43840508115649757,
 0.8786107069696845,0.6264905467160234,0.2691468518371283,0.8985583885105253,
 0.408147376980703,0.19949392686945977,0.4453213699514834,0.44642664211155825,
 0.8485812452576705,0.9725003258650154,0.07568848814224582,0.9450468803421631,
 0.02562959665023301,0.45477141692361533,0.5750574543042773,
 0.4306267005426032,0.8540715132291566,0.7648494882890768,0.3461886138982322,
 0.5513096240486639,0.45079678393373945,0.9676845332309946,0.3668116503519101,
 0.7474793760508031,0.966220887985473,0.473173929314859,0.7844030824394544,
 0.13816480334837755,0.21466995398353372,0.35974338110862414,
 0.781638400413076,0.9302427599298868,0.6216419147093734,0.1158507916124688,
 0.6806019329026152,0.9576334548230026,0.21026331752767424,0.3928409448004163,
 0.7029160731135567,0.03358574273115611,0.6040001642387123,
 0.07523067237956427,0.9889033029298406,0.4237117824982406,0.5385180844791556,
 0.8728330877713162,0.6730779740947054,0.24678480524515112,
 0.27427334685947613,0.9126508611554007,0.5636108447432647,0.8257413032363636,
 0.101906940603667,0.06242558876453541,0.1974708190726049,0.29295726467612404,
 0.564095275161179,0.8962050199964223,0.9701199525955945,0.9938291076089918,
 0.830439276040399,0.09841896835134833,0.5366921648981067,0.3546712775967412,
 0.22974434227777907,0.9763406352727456,0.4015077113267673,
 0.42107825529516385,0.8986714686035084,0.7879339725048798,0.5384794076980156,
 0.8341527681906526,0.8127018181100002,0.606825510313946,0.2806697679013822,
 0.5933133465689046,0.6178734616385692,0.9690502794078609,0.14197206585659483,
 0.6458718373605074,0.36374537374009963,0.5233587214898858,
 0.22920446689210205,0.05886715924413355,0.7409970756807547,
 0.8609872755656419,0.671488574731883,0.1890804154902097,0.13795611966412613,
 0.9454269773422049,0.9975583076493425,0.005221870204177748,
 0.5432748567863324,0.314340006669781,0.9397485940231133,0.003174732292728355,
 0.10030633523924815,0.9046047081516865,0.3377422572868505,0.7338388958096898,
 0.16236346616205355,0.2471357521972013,0.9522268534078018,0.5082641149627597,
 0.9338496759406546,0.7723813113343879,0.9090518537038244,0.3279460486164584,
 0.6303484605324718,0.8724977396085333,0.15862460775396026,
 0.05390328176683046,0.6938282563301837,0.18086999028166262,
 0.8999763247654408,0.2299509507634404,0.33097501290002274,
 0.30551255346445294,0.05309184274476342,0.40871165350832506,
 0.8255910813085268,0.07090463056125906,0.7065567153648336,0.8521621502420909,
 0.6498780725337712,0.1956486083909319,0.1085729150353969,0.7053903347311924,
 0.19893464885697765,0.39808784148379106,0.7636781706457226,
 0.6188202642195257,0.45858280627427217,0.4568159831889852,0.7153102401207891,
 0.7877813603902157,0.8310565674413216,0.7534258385948098,0.0656404537026598,
 0.4157625045204716,0.9387867550081442,0.17817984415822663,0.3144967435226189,
 0.6954294425410857,0.49940466926511784,0.3860871805721048,0.9172633516186369,
 0.8954411082491045,0.47832381827989356,0.8848734177143092,0.4941063833026136,
 0.3884547344516768,0.8102729497235412,0.3753612860224017,0.3489957353999825,
 0.6362244410604728,0.4158922995493587,0.2444194827883467,0.4691713460734682,
 0.5936549381522608,0.6616176843876104,0.5088923011890045,0.45706046611323403,
 0.167275853349657,0.5401269382690534,0.41826348889080256,0.2925221934889055,
 0.18104481109582993,0.3826064483901921,0.29462026258151797,0.216177803196977,
 0.4974965103700807,0.3527343860384514,0.2276194431336961,0.6371732511986989,
 0.11294935962544861,0.08523668318938027,0.2992083344903197,
 0.5901134729491095,0.9203144049578811,0.181210164742461,0.5939696079554796,
 0.3304559055662708,0.9718791704173333,0.4049108476882788,0.8004287378826627,
 0.7874949191434282,0.6435091645575519,0.5378964438753913,0.46490725188544424,
 0.6156060872107562,0.5957341156689951,0.5465539078684448,0.48337760071761515,
 0.3891733665884044,0.5683222929717698,0.14864730388073055,
 0.10176550151262054,0.8121434337195703,0.40536292719388634,
 0.4424389818413157,0.25704214914918366,0.6521422503862243,0.750228237746084,
 0.21181908797879434,0.9983652735512939,0.17408451075165554,
 0.1581956942013385,0.794302448969773,0.5535798784875927,0.8547694058612343,
 0.5046201707639344,0.5153422345106973,0.9152393980878153,0.9225651874529868,
 0.3236009820794906,0.8592429996703363,0.3190898478570088,0.7551843509810832,
 0.26870905570597237,0.5646088726168891,0.9883643149027601,
 0.03240277021874283,0.4366801846759407,0.8018434609502858,
 0.44777748231879877,0.45949934437386486,0.16907318123746817,
 0.43504499504194827,0.7984390106512173,0.6897767623812211,
 0.46420923569301786,0.8411028438431506,0.8258480434032482,0.4541213332480589,
 0.5849746047324257,0.07637681775245309,0.7464337655147325,0.7992166967924487,
 0.8310348492803726,0.5370082268562242,0.5666167463335875,0.5741961398152213,
 0.7761515206619249,0.9396679470931518,0.7709453740986751,0.36560056915767514,
 0.14822845874051005,0.10315573672616285,0.7766376810515738,
 0.05423787560804849,0.06965687482820504,0.9345878043427069,
 0.4332955347181209,0.15902519022658956,0.8773628079906037,0.8628155745165006,
 0.024507737507645455,0.6190715742020331,0.6836383551390428,
 0.7534563271722184,0.8740419426859148,0.863420582086315,0.25938610702482456,
 0.4447434023662409,0.4886979081392945,0.8712627930668675,0.3937084494275498,
 0.5181300235561916,0.3107109763732072,0.8931545970171476,0.1143603246156003,
 0.6329025178370945,0.5302753694885614,0.9449908109964664,0.6892720798723728,
 0.9235658078658603,0.6953030051141681,0.394913864768488,0.29342519325156924,
 0.8618733175369313,0.2294029684533665,0.07013986558359442,0.632031283335331,
 0.5653887234761003,0.38399712582129253,0.6806065035390945,0.5107230389432191,
 0.13066267691321531,0.8335207737050772,0.23688551044179285,
 0.7394031457289604,0.29176230990765506,0.07852377198698024,
 0.5166474916848416,0.7728674375033379,0.3956773043313273,0.04243679002065415,
 0.46724343642155697,0.7594346001523524,0.8883491607038381,0.1718856015657345,
 0.9776897923987962,0.5099082385407461,0.4625969965684814,0.6838054168535691,
 0.27293097075736805,0.5525391641901896,0.9091431986484961,0.1085896045033462,
 0.7808077283549272,0.6509041894555974,0.4428947519483897,0.5328728740202567,
 0.6552553378437155,0.4591383647851517,0.2194818268564247,0.9523487600474225,
 0.021400908235395755,0.36071147205661636,0.7481928705215837,
 0.18921282454290977,0.5870539442018508,0.5438739271148445,
 0.30403805931960104,0.12216305585780074,0.3984629831950316,
 0.08465545830064514,0.2097227900855192,0.2391851537098837,
 0.45308541764177046,0.8560904441600224,0.7286466457211895,
 0.11033376306871123,0.4878809100343242,0.8013740121485393,
 0.25726168955086326,0.871384333511207,0.953761108432167,0.019737174275797464,
 0.9392370938409542,0.4908342574804577,0.11060832255783226,
 0.09066402530589135,0.05579631684017394,0.3319941411112982,
 0.6940880372179129,0.8425792416391381,0.4876503804097173,0.3640226104838682,
 0.980696722103307,0.5358088981342988,0.27634392680192055,0.804212289905257,
 0.7505022723858226,0.17191615394122783,0.1437833160688431,
 0.18308976357441653,0.5703368581172146,0.4032397182359748,
 0.49997930984352745,0.731822025631984,0.8748868267771186,0.3263563009215943,
 0.8922217242212163,0.3771585155311634,0.015215894795229334,
 0.46257410037897373,0.5800725037091393,0.7179609421336997,0.1641951326680524,
 0.9808059855973215,0.6136245742044857,0.3944249771892099,0.14978497158677195,
 0.13867460518606511,0.8742240370009082,0.6260003746478575,
 0.052314012456596726,0.474305778136214,0.49350936764408615,
 0.6921921085200511,0.3296440736043539,0.554255131635863,0.7883206352119702,
 0.28602738550850815,0.44156573205217187,0.898003419309207,
 0.09072560316263567,0.4720740482322122,0.1380267731864958,0.5511513122269137,
 0.3161886623328839,0.5284176224092091,0.3468706231693188,0.9813303661062315,
 0.5863136068382986,0.8438828597289659,0.2709447990223919,0.823310493095867,
 0.7604943441067591,0.5165258056251223,0.763423658068289,0.4560240162594966,
 0.06749625617572219,0.31700940558992696,0.4402505115555275,
 0.5434890934236013,0.6134603883247501,0.6506877415058769,0.8076555184447569,
 0.19190767730177305,0.4865128040101412,0.08671967898763122,0.9117890821442,
 0.9012323592742012,0.33979421030355506,0.6951099628928156,
 0.10054495168134492,0.46674301442442834,0.809773311494409,0.7177031829213635,
 0.3992786626745617,0.0456370664354413,0.5571544828068639,0.7851520863084558,
 0.8885962901320043,0.9326617745874364,0.10680654307290904,0.2802026098925162,
 0.5580359487153406,0.6862488624135723,0.15366001426444098,0.8956604744024436,
 0.5005094144695043,0.348417599349351,0.30603457097160436,0.9461307898852336,
 0.5591963849670031,0.7724754409577936,0.5261422172088159,0.3178664095873317,
 0.5897206944686859,0.49111313715618177,0.6763181093332271,0.5994184934999107,
 0.6925057043716931,0.15045038533807908,0.11616091832481001,
 0.6150560718400067,0.9779879434072927,0.8714867080654595,0.19289312256466518,
 0.45778871292201206,0.9936703727541679,0.042331219012584964,
 0.6574780337516377,0.20200920689481017,0.2825244742115002,
 0.20167684565311528,0.7504979302156775,0.2693913356553407,0.8579953540898891,
 0.31043878212748766,0.5722956298739611,0.04150707954374688,
 0.38827155072570774,0.8317649660293467,0.4766331927812111,0.8091249441333427,
 0.8034246047832267,0.6290033266870141,0.602388730386676,0.43649649306148397,
 0.9918910504656993,0.5480745275807635,0.5772758230835936,0.6851046551815471,
 0.5874255600405506,0.6749677057063452,0.4131957102146827,0.3331968562929273,
 0.7072416077639097,0.2916505269963733,0.3274803559471302,0.3066900145745354,
 0.37405566576302696,0.21863237901015597,0.8456754616941774,
 0.27964361532864135,0.2996930852798846,0.47928848746343167,
 0.05639839484010256,0.7349145329334439,0.12829456139037376,
 0.0696927800749374,0.38914755034022364,0.9836788509109453,0.8584621989245249,
 0.9283068912782637,0.6782792634216716,0.9588080515292775,0.0700257185328298,
 0.3429287440289336,0.7561197905186703,0.9066809359878397,0.3451826488285836,
 0.6307435286118843,0.3472707233204062,0.491886877495182,0.910133901948547,
 0.2817498905842879,0.8233652676233613,0.26984310630796937,
 0.24566977324419625,0.977154484388923,0.8978422184291802,0.8338270193717459,
 0.3800944475043522,0.45344421216216535,0.41733186394206845,
 0.034700113157172474,0.7551214727095811,0.43962332859598474,
 0.40470616071623267,0.30057024896490814,0.4730745373770736,
 0.14066463847729216,0.4193539426020819,0.8484567053425998,
 0.47023540913922046,0.759422565043459,0.09267310712996002,
 0.41777163562762776,0.7171041025617463,0.5381391959207551,0.3240202482485598,
 0.9133956169809344,0.7003851338949945,0.4471075286167727,0.15551815932223778,
 0.7713261023675729,0.744069395570486,0.9006949058748839,0.4384142631315805,
 0.028805193951574548,0.7809753031745019,0.03192275103441056,
 0.0896936748509859,0.5493310137103519,0.3017438910439798,0.7815036414286661,
 0.5588807034984538,0.13451965139729627,0.4463135578488594,0.8258305809921942,
 0.7717285088701334,0.5918567496266525,0.9753763489037724,0.8337731070791946,
 0.21083618770478596,0.2278559652679618,0.13156393364182184,
 0.05650133864206097,0.21717749722466506,0.8036885966759688,
 0.29055570743179815,0.8170793958700503,0.12500283421302594,
 0.19771590531168026,0.9494676595464112,0.2624995585785239,0.4112002682881063,
 0.446837873366541,0.20314075639040796,0.403896300122657,0.5517261003498586,
 6.108490362533026e-5,0.7802389818400004,0.4226195061624559,
 0.5636503869774546,0.512373327672084,0.09104475925950284,0.5537338134715731,
 0.5254392996171349,0.276430153366366,0.34307222385558,0.8844537647018333,
 0.0849025774814437,0.6580500499975521,0.5467611344553815,0.813671394651129,
 0.023540058172346945,0.25134624522766436,0.4152746576190609,
 0.29464866139381973,0.5494274087781709,0.8401970220211937,
 0.34168080529075195,0.14443512493580657,0.10419719478386202,
 0.4003798954708141,0.5614843412712553,0.4127802931935063,0.8822854942518087,
 0.705680343026523,0.5732063879878447,0.8597252705170491,0.2739997894802434,
 0.10203746666503899,0.38166475668851396,0.8483068368735809,
 0.021941856294651174,0.9892637783130003,0.23592630693633787,
 0.9588971252276022,0.36664123835194173,0.41913606660774017,
 0.14825021752938627,0.5140748243695437,0.881752504511686,0.7084318831502026,
 0.815852691559764,0.8673370842551472,0.13307970263037416,0.27111981385962625,
 0.5404523761917184,0.8542173976808203,0.17821222510478973,
 0.03971762588778882,0.08487674261713829,0.845757736406136,0.5607560438649787,
 0.29440859856792023,0.6090526975610139,0.795540920286754,0.961135546252533,
 0.041611347526521736,0.48405725642980535,0.6565553275499298,
 0.1457898698466391,0.8907104544024463,0.41024315030226444,0.5067777439066732,
 0.029088969326067526,0.6304510029901176,0.5756434129911447,
 0.9663895507374233,0.2819495451082368,0.566849258434649,0.7744001924267843,
 0.7514900314239942,0.46610262874605124,0.8629490043032619,0.9751704224294615,
 0.20174141519314714,0.04134009274557837,0.9863967774456912,
 0.5968766309751061,0.6422758095858067,0.898561716058593,0.759193714107716,
 0.9626161027957982,0.8747174247478504,0.5266000140158588,0.5721766709656168,
 0.45783645764948167,0.8623396546018182,0.3198701844369439,0.6688380033593315,
 0.9713818474982101,0.5426333968738667,0.18075254786616068,0.9564373765008504,
 0.5543816165730718,0.20443120053793817,0.9359987864891551,0.2873397756174624,
 0.45044429367029837,0.9348434520809631,0.11348627705395009,
 0.009610274395348006,0.585276374562664,0.7040610574291744,0.3862971668506058,
 0.4262466639043211,0.1962566223379919,0.03880178446676563,
 0.49842379735383835,0.777377096059919,0.6056127393745647,0.6843031283202773,
 0.7895136438976263,0.8914058389798886,0.13985774441708487,0.8935139718185838,
 0.38030090631159263,0.2766910198585717,0.5531670011661778,
 0.37953544759249624,0.5207821056024156,0.274887431130448,0.14603261129237444,
 0.5199792483113412,0.8044046809859934,0.3775015700719183,0.5922697798319807,
 0.6546467451864462,0.31878732781636954,1.0497689512978248e-4,
 0.15833351447386035,0.5076580658683073,0.2857594403980275,0.1628959353467473,
 0.5307991781718692,0.15640765730359474,0.8696526584078487,0.4876238719429965,
 0.5619150824508308,0.9961233435596505,0.2474670733986477,0.05034430357537556,
 0.9190010297842939,0.7276973901495609,0.6626710156910331,0.9355745778711445,
 0.7108765483792239,0.8439913165518056,0.30329465816310897,
 0.42333740279122356,0.3173876145663578,0.7831133397735148,0.9733214310732342,
 0.49700019219079283,0.1000642099160527,0.5124225770851916,0.2672200524406637,
 0.7524563854022233,0.41977841987932185,0.36952647159965246,
 0.005982478090435395,0.27377121180819586,0.5965264735008988,
 0.07141413599258406,0.0724003799414279,0.07236108869583158,
 0.1419395775310448,0.49401040214828496,0.6003637497523787,0.6793282800366206,
 0.046803254179107356,0.4700334523205586,0.9239129912384085,
 0.5235696323414668,0.06394722024083699,0.11928443134214162,
 0.5338458578484606,0.9516725823815056,0.9119362653577219,0.21704951706773423,
 0.062182961342867715,0.9716390143408574,0.564755834124631,0.1538829921910232,
 0.7502825582651184,0.4171481298392443,0.044092681095159625,
 0.6374533758972464,0.6886302975099343,0.290187485715393,0.09205117752885184,
 0.8879568643545874,0.9973387604625481,0.6843866939117542,0.5317566745730415,
 0.2877686407320391,0.11318106989072374,0.8892268223058023,0.5449025403799808,
 0.5644958817936033,0.4485469621274545,0.9298792013131418,0.7636833900998958,
 0.7679436591660203,0.7706355025585315,0.7815872892207032,0.6826852714155074,
 0.891290857275878,0.9607053020770411,0.3203976483340871,0.08570867143602823,
 0.8566395056270344,0.1396778689924998,0.2113285988273107,0.9900943514351719,
 0.7072810377609035,0.5032502670503026,0.8819550878035278,0.5731400709632966,
 0.41526831186126445,0.933408768520212,0.892175843310197,0.4024492647757667,
 0.984786979171395,0.08890776134183465,0.4215058272401808,0.27860944073847826,
 0.1259232294655227,0.15805353620690776,0.39603588685080227,
 0.40555197464085535,0.21912114146836714,0.1991405123754706,
 0.5803880932543282,0.37408948674911213,0.8553093083192093,0.2023749089469391,
 0.5236803415457593,0.6074460086422515,0.060485278856034386,
 0.49731350899335025,0.665672711651938,0.4629118560555715,0.17511563876087244,
 0.3721150220178785,0.8046967023520297,0.641356730841693,0.5011712127151674,
 0.7341242321759208,0.7091829706708985,0.8248339831078186,0.5768648267676609,
 0.3187738181379959,0.3628310365095191,0.6051624321343332,0.4895953448556688,
 0.6493218475455999,0.5938705552221655,0.35730314248363726,
 0.08139641618365101,0.7314261940800073,0.5415386281301251,0.373034039672988,
 0.5582909589648734,0.9834346411436063,0.9266393010597644,0.9090242940867385,
 0.38735978202463106,0.5162665617598805,0.6767271658526229,0.6974537379879938,
 0.8273379709380096,0.28974160823457573,0.39294014015321144,
 0.4298031967394845,0.7095866990795432,0.9003198662955364,0.02126868394059922,
 0.4635190847313563,0.6374677929410374,0.9281074035054546,0.6041558802522468,
 0.7535876839382124,0.977572505371879,0.9960960149315518,0.7813516715075919,
 0.9430770063804328,0.21006999296850104,0.873333044969738,0.2129003604886952,
 0.5361089660543757,0.8695745288461545,0.44021332446766526,0.6857393456907663,
 0.5683548639194262,0.591348627636873,0.22574468624548027,0.29823450289226616,
 0.6269140760513673,0.6430911356692238,0.9279999352970069,0.14089755627436484,
 0.5672234851246828,0.9526193853387249,0.4555379828413103,0.7506012006700995,
 0.6979839253996479,0.04516581193018809,0.26236062357856804,
 0.2775016346277407,0.43297029031965273,0.309308430565329,0.5954442259150186,
 0.7549508485090906,0.35135468153587934,0.3759096381122795,
 0.33890953970090665,0.7512928698608101,0.27707697623166494,
 0.9436713368448069,0.7340118503075584,0.18046999181623313,0.67169378497354,
 0.7306292829744262,0.4920890218302779,0.9467917690477554,0.3020607358826263,
 0.18808052436825295,0.09056980246860302,0.9845167649979788,
 0.5979975354707081,0.0012460248953838349,0.482970119117366,
 0.3516749435340403,0.4287858448835218,0.7360737058788595,0.6350315401945282,
 0.5721243544065232,0.8591191534600104,0.9353923380623359,0.16630272887396902,
 0.8769203801606806,0.4190461410459585,0.022808393041022335,
 0.6128982490799099,0.3099296144716275,0.8734374939602648,0.23360388555917777,
 0.9133369269498277,0.49902482943075444,0.1850735786985468,0.8510798834359469,
 0.1696755968062963,0.5705854995181804,0.9462291873366436,0.8940479667761456,
 0.5970468723626288,0.7430812111612434,0.26718391705904165,
 0.02476801059169098,0.20828388968004186,0.530573144491306,
 0.29424335003886026,0.053947775277924714,0.570963259141787,
 0.7686203868290429,0.3625506268006682,0.3642781157986116,0.42039502834409526,
 0.5204560507131113,0.2545958670183437,0.42303759338923563,0.1562954383970072,
 0.9316822962178286,0.237130041322427,0.7375133230965605,0.16284150689185628,
 0.5281681986503152,0.9550811717013898,0.11126472304737434,
 0.25311831915146243,0.6094836430668128,0.5589296566677115,
 0.08035926496572054,0.27026504099878723,0.7777078256639748,
 0.8829564054327739,0.5229920105350475,0.5925901544385929,0.40841537685051676,
 0.9593551208486015,0.721076117820525,0.6492355940606216,0.7847137071452673,
 0.8183278972294229,0.07327695341065421,0.031415148420101846,
 0.4622952727963423,0.8902697468188672,0.01144894602344626,0.5387509610225318,
 0.7543282710126502,0.5759962360369609,0.3022291742720312,0.8272021645284725,
 0.5885445801629916,0.8059702461583438,0.04663482602171887,
 0.31473488910836667,0.9567347946766714,0.5465416042829585,0.5854448195117372,
 0.17977013856453083,0.8822880168771574,0.525550319716432,0.9797305304276845,
 0.9943854042895768,0.13011723518337726,0.18285235198050942,
 0.9151485839972326,0.07902183425069786,0.7853947296952664,
 0.10814829731872355,0.33550397120622444,0.7293649764434956,
 0.9447157956936623,0.06822828452169938,0.2707864042952295,0.9656871459963408,
 0.9725487009498892,0.853456261125513,0.23903786638381763,0.8549952032865762,
 0.1552825929583671,0.2290628878065032,0.1815185954868267,0.9068269835650986,
 0.8945319802630463,0.8871888980889523,0.0711732093402846,0.521439134008843,
 0.8335863687439675,0.703674375493945,0.7926099050730525,0.6089460946943291,
 0.13556067898233626,0.4285409985235441,0.612806763192317,0.4927099988034117,
 0.6743738750635155,0.5520807187070936,0.7512417010618597,0.46409647257247777,
 0.9842866464435991,0.44748383998860375,0.38131477427036176,0.815376825294867,
 0.2569165218993401,0.6029682111818998,0.6064295022061741,0.1196399233671479,
 0.9600803738737025,0.4907727000138917,0.8816376828869893,0.16631336208923264,
 0.9653664673374089,0.5854013127226151,0.898550056994647,0.9825131360129646,
 0.9768950137388772,0.48044540680459447,0.5263607750400324,0.5027844284297791,
 0.047572536283469136,0.5114278082014796,0.820215489631564,0.9539283543359454,
 0.385600019003946,0.4503507479046853,0.9073298279920418,0.8001039630086417,
 0.6735636345574108,0.6381865382967414,0.7434663548223521,0.2915506854396378,
 0.27278498265272755,0.9401567946412935,0.05212690211111248,
 0.0037658163511992138,0.6893416769683984,0.07311420435710236,
 0.7990440373970308,0.5193309882719707,0.15506283472922044,
 0.023267537630882407,0.3077902372453709,0.43973206030635026,
 0.104675064885553,0.15722038503026992,0.4464880164654883,0.9133079151311297,
 0.7483328836091006,0.48674155996389035,0.674556058501871,0.4457083916252338,
 0.6953784719729902,0.029779725968625437,0.4787930791174715,
 0.6791487382676732,0.31431205439449006,0.7667238658478693,0.8988646163249803,
 0.132296911371658,0.08268460935893007,0.4067545960189618,0.8881232117422234,
 0.9681711896338953,0.49984757000220303,0.6923594856409663,0.6201671983085395,
 0.8924352403456401,0.04423188837725589,0.5975007968164827,0.2726222204539236,
 0.08160120186493225,0.48660489183414257,0.7909870272278766,
 0.9927016004151714,0.071367312769262,0.7135913249274468,0.09799594947117174,
 0.8658769102268816,0.6746490894718387,0.5436400355775721,0.26903881283760633,
 0.4832564228043803,0.8581725029021248,0.569908309708527,0.17982547707563445,
 0.7804751798749795,0.972115629946293,0.4512464226984838,0.6993711127941037,
 0.6466972120682005,0.6046499823204784,0.4616339555343565,0.47445172161239124,
 0.399463157620051,0.3314002424492928,0.11755789362283464,0.47471762781340876,
 0.8846577470872375,0.6555190126914802,0.9129010922553656,0.05311024445412227,
 0.5197390003787629,0.4971818960187009,0.6399791831383237,0.43371965449455707,
 0.09358551691269235,0.9719505251664131,0.24554792121151103,
 0.4391627142985669,0.9311380839790898,0.9314369619515802,0.30049829793177085,
 0.5151161531856926,0.9145321857213194,0.9522513605683843,
 0.026120429147905133,0.17321913364847463,0.7378200418040968,
 0.2048058445455342,0.9208568557455554,0.43564118127833407,0.3860487406320462,
 0.23570062708948492,0.4156052802882998,0.5632444905329274,
 0.01817168625035559,0.01433322242898305,0.9044578473188181,
 0.4736480014474501,0.498120367263148,0.47193194657250703,0.10919766040129064,
 0.9663728690006804,0.2538300494910719,0.7102281353565745,0.37501282705646444,
 0.4484120075274598,0.7151911753852906,0.2663697560764463,0.6721364389983416,
 0.5672465359101754,0.831445301987733,0.3783405635553393,0.6894356833069034,
 0.8351065841869044,0.0960246964083159,0.9800050786204811,0.29754174794947597,
 0.7837806581845776,0.5265368269420062,0.19118709630323905,0.6043143359645673,
 0.9894816402412424,0.7753248451858089,0.5491085333748218,0.2814600314189891,
 0.941591654926332,0.27767527720214014,0.9321817518575192,0.7029895657141219,
 0.8672600674696002,0.33577472793723906,0.9305468281413501,
 0.011204756948510974,0.7342851227266924,0.7196847170492934,0.94666691394521,
 0.28007522508342886,0.9704821493396846,0.10743710242625082,
 0.25486752381249456,0.5371625013140197,0.8932607057583798,
 0.01711825736743089,0.36849557171031777,0.8121169237943141,
 0.5376852861498028,0.34158667967475953,0.6161066175271537,0.6140155142432548,
 0.49685087934738714,0.7379105462521771,0.2901185861419281,0.7294055694324062,
 0.5136128855337211,0.2698602347435667,0.9440578126755595,0.49058769415990333,
 0.48273117194949255,0.5456988411415454,0.7338960498554097,0.60138563166262,
 0.4514591722600031,0.5514986832512521,0.6265368293839739,
 0.002998612656193833,0.7253029054118427,0.4587622004507558,
 0.8429197741534763,0.7901692631025018,0.6437852878047587,
 0.012038465925551356,0.6888030067142481,0.8767140792222904,
 0.09253984717439745,0.4154784847313212,0.7539608086293532,
 0.09009221546407487,0.180285165892201,0.2694207615847467,0.5282522541364255,
 0.16590687908002133,0.42493269646244913,0.29456827779768224,
 0.6354786947709835,0.477782370589794,0.3805662636177143,0.7158071729437844,
 0.30833664760130575,0.2391893983923733,0.26010788433742715,
 0.9951237111560503,0.4070777136720668,0.6423207323766705,0.9596252076481687,
 0.32202357743259125,0.4462259176247738,0.885472238400844,
 0.029786243285147562,0.033036789886525364,0.7057801630099156,
 0.8744649506364978,0.759069387990363,0.9818302916273641,0.06127548605636224,
 0.1611465216497019,0.29265389820557397,0.4852959325835089,0.6764405042684538,
 0.31409703682524504,0.9760974263993294,0.44327302452438555,
 0.24237484816942956,0.4196208730943849,0.8198113623559223,
 0.01298054145261629,0.5456026001821795,0.6205702485156325,0.8044030583900144,
 0.039431747949529194,0.6232421098217228,0.46838248494759716,
 0.04708075740966666,0.33315613284159407,0.7278609357528544,
 0.4207153153177705,0.6144555276380121,0.035243357910018314,
 0.11260560129620645,0.6583001474308192,0.8876474669833543,0.7894069372823547,
 0.5398317770408148,0.3764332237225677,0.9506116794005479,0.42320715946771703,
 0.13115095012773803,0.0703977941894669,0.20623539641064137,
 0.7543473293909364,0.4295355087480235,0.06070326676927895,0.6742885205436961,
 0.9400700224442567,0.4093409208553579,0.07925563920247014,0.1282527596377011,
 0.21583363997753358,0.6806698704836424,0.449056292458034,0.9989116444980645,
 0.9667360342038028,0.7985585303540963,0.6301857069314958,0.231274642876095,
 0.04130866907025965,0.8379293874726665,0.4562549233141734,
 0.09718669441065686,0.5775545001853414,0.8329726588138741,0.5627177183813344,
 0.38152454683735637,0.7846968727845144,0.03776142380478764,
 0.2374267193148829,0.34082394496479773,0.8897350249622729,
 0.25217096981994236,0.11174922982788482,0.09258162377470902,
 0.3319486578911335,0.820331021972585,0.46305308997045014,0.3874610708083521,
 0.8828960832589484,0.18230475487281117,0.4225893413148082,
 0.24980657468991785,0.15857892288135222,0.39712718722693785,
 0.5275135941116463,0.5890643475813306,0.09856230115687437,
 0.38300796188053077,0.8140256118157762,0.3019288024415845,0.9998151764137746,
 0.7839755092813385,0.1874016882533529,0.7885078207849698,0.25259039586247345,
 0.39591561386048113,0.20878371703244536,0.41516664882106014,
 0.05225716580384421,0.9290264696985984,0.7451669713693256,0.7466508189772517,
 0.5502865301167138,0.30333157817925493,0.9122558251805564,0.6005303175672557,
 0.24089557529084948,0.5975662690796529,0.5185267019100817,
 0.42083643116426184,0.26552335760649337,0.7342772450740658,
 0.29732963796097023,0.587670646475827,0.01867196502260282,
 0.23568582113228753,0.6488368117388886,0.02624116167260282,
 0.8072424304897197,0.6383268554004964,0.1619143730725675,0.1878161111426785,
 0.5375033708510487,0.25688890176695156,0.5780165027733963,0.7923750257073389,
 0.8448406227114198,0.4856066552375423,0.7115953445627919,0.6819964539500728,
 0.8074810869697663,0.9606944677482653,0.878812780312527,0.9708229810297184,
 0.9714053831997881,0.30874716049630835,0.8220848964798065,0.7617188994010041,
 0.9544103332180551,0.3067659212552808,0.1689428774699282,0.5145067021781426,
 0.9483409381351897,0.6195493664186612,0.5846985073763734,0.37702208354654554,
 0.9500990850613716,0.31630641617262345,0.40887838146162925,
 0.9993964950540121,0.5528335752783925,0.37670256680963166,0.83099944601201,
 0.02689181956090725,0.5280151306138716,0.9052159045016379,0.8887228878168246,
 0.42899612964137923,0.5563185056145792,0.7170948560052857,0.8639499915297365,
 0.1335326896977328,0.27845639766869346,0.48572056811296505,
 0.8594672082411788,0.1050787469349892,0.5573621988360745,0.4352475779963493,
 0.5567614733277423,0.026937353793747842,0.47479838894186,0.26297122769857184,
 0.24485849334006882,0.5103437723453278,0.031121274636537466,
 0.16781042759322462,0.1476801898668142,0.4677606250966797,0.5663899681866312,
 0.9444983773703913,0.7163860241504643,0.7755329928787434,0.18157174871275772,
 0.846468481428258,0.8557337164010689,0.08025824016091876,0.7763011379836358,
 0.14193142916047774,0.3473983821766786,0.30839800969613396,
 0.7453314131466859,0.6039470584189813,0.030920954675654144,
 0.3605101230639549,0.670956695477721,0.9252215573686893,0.265208897355118,
 0.4010366143772266,0.7877594806675867,0.6288624231333341,0.8511360702694009,
 0.23133280781688503,0.14542619711101956,0.8594044084016345,
 0.6900539248862122,0.5658309171252435,0.18393454465416503,
 0.16369717278618445,0.2658694407068156,0.25560400606453404,
 0.7221206020194451,0.09942207077187881,0.3055175241775837,
 0.20049789813016272,0.5650471961485798,0.4201111836104464,0.7458390604121716,
 0.3350306904404091,0.319917017279848,0.5451306576444735,0.9901453073714702,
 0.5469900998210849,0.714536207722828,0.8702591443368822,0.41027106995082285,
 0.9235690894120093,0.5449893141594475,0.3417835432901275,0.094306207137413,
 0.5814596194916478,0.8822065371464717,0.5683233304249565,0.3601311484695283,
 0.18094236622182003,0.04668517233822811,0.9941435177018687,
 0.19802341685444824,0.5024307052281387,0.8076351471070798,0.3695863517109661,
 0.6812995584480779,0.6350865000863382,0.7813800202011524,0.9976850727378297,
 0.7760798242324801,0.018825914060751736,0.07883326597852702,
 0.8017040860589107,0.46300706512393175,0.976121888084689,0.17376100841348463,
 0.06159491390432403,0.055849039721119986,0.7724225284652767,
 0.8896829828973029,0.01575378292402263,0.6056589814033974,
 0.011758038159159878,0.761598805560275,0.922214454957197,0.3868429330919725,
 0.39666003294042396,0.8347564109173763,0.8915802958056266,0.9357366513396989,
 0.24838677859633562,0.03325508339989636,0.9584451767325309,
 0.33977719159902875,0.6922727701151593,0.6039661414553299,0.9739655163077514,
 0.7262568680778232,0.24901452263434065,0.8350571209997012,
 0.49281945555714013,0.9623464000510069,0.6949099895638793,0.6107890120107939,
 0.5677882100677545,0.4914115962182902,0.6523896501431555,0.7137832495041828,
 0.085938615242653,0.48942332442816294,0.702539156643353,0.2570668246364738,
 0.8589524384669782,0.08085371298234945,0.6841046903945871,
 0.049202224540386585,0.41909835966813425,0.3421035994254388,
 0.011638209900188046,0.8307615360274955,0.42275850189723696,
 0.46615157322709455,0.21690951580889417,0.6618385532093835,
 0.9470811253282961,0.8345696843708291,0.37458335208486127,
 0.33631997614025777,0.7445681090543925,0.979851646328466,0.6587229145451878,
 0.9818883144178462,0.8080901261630073,0.9311173351693968,0.30003665870374097,
 0.39288657938412197,0.11284748817808454,0.19794401164683706,
 0.02665296546444129,0.983626040288831,0.16953751620392443,
 0.25085345268657055,0.13340131659201637,0.7451441451351252,
 0.27333852544923254,0.14693734921927248,0.9138147332156917,
 0.6249870285972453,0.6071100986474076,0.76783578822638,0.11373283130519385,
 0.005862990254271283,0.98852868390687,0.3770308132985085,0.6674872649697295,
 0.1129059705477875,0.5218009880572241,0.6628530381803139,0.13243718468719878,
 0.46196210328895027,0.5921764574418141,0.4062784818100782,
 0.10613051843921983,0.7397761517388554,0.9477476348842164,0.4020089356921749,
 0.07780885562576778,0.06495732797989773,0.2865047083123352,
 0.5600377531773795,0.9709865498541677,0.9731256425801447,0.9508207039750067,
 0.21975671625957494,0.7228685508872429,0.36006235636325357,
 0.7115970457321263,0.6567042402019506,0.3382616463391652,0.9427149871263377,
 0.7720346180526867,0.7039256921694087,0.8706113964104988,0.11679466508710179,
 0.5100078828939275,0.28019218501803533,0.9056863462623238,
 0.48047325056369306,0.006470398676305611,0.08556637264220368,
 0.38574571807664004,0.09643552352589668,0.0944663969251548,
 0.04389495644159869,0.17177869985465843,0.5954671880488774,
 0.5229196760593332,0.5486396130601945,0.811608971899961,0.6340448359389145,
 0.03634400736751997,0.13353227054286498,0.9996957191863008,0.453700517175627,
 0.6080366269689587,0.20564560995006986,0.770294257961628,0.7026866741982638,
 0.42468285241853976,0.5515260586257382,0.1920907671653166,0.573484222776413,
 0.20505651624789623,0.33946732217520126,0.9394887455835977,
 0.38160135341014634,0.9879881905311281,0.8004447846111424,0.9495713556189022,
 0.0850933220918626,0.15667566501731478,0.9780271504809168,
 0.26024005677124196,0.18328314882654972,0.2818758815247687,
 0.9292215168410614,0.5611211691103398,0.6808871933650947,0.52058621001287,
 0.9112519108235249,0.3862268534140516,0.16033216126962058,0.2449754355264261,
 0.876857253071561,0.9126908131952194,0.457364430141858,0.53601441508124,
 0.6786527453735242,0.11887660745295847,0.6618061033515292,0.3326634178523906,
 0.1408288872834227,0.6568160260937441,0.9860799477230038,0.31764330800549123,
 0.367334381122168,0.73863530671734,0.27648043931563904,0.8497187122483227,
 0.8680795494599995,0.10864148833544729,0.9082901708863529,0.7379002188217615,
 0.5368180770082803,0.7358336627173567,0.5182134910627293,0.27552175142369084,
 0.07789105517703343,0.32669848029869086,0.498133213372,0.7304222531604007,
 0.6220094987476771,0.8699003458322933,0.6672667935968636,0.758969418470998,
 0.2675726972681479,0.04435618527212626,0.808151091060151,0.742345868826991,
 0.4713762495024415,0.18759794815948438,0.6726019175116669,0.6704953449983957,
 0.016012080268614337,0.8925046164514261,0.21478758152098865,
 0.25670437333451246,0.6545674741081033,0.47569131547053944,
 0.15535491470642848,0.09790360893133288,0.29357737751401913,
 0.3802506141155757,0.3288554529442105,0.26626059066180896,0.7184731658439429,
 0.8318617418490107,0.996411177154291,0.5301718428088876,0.42234703592846834,
 0.013247063916735358,0.8897016225975575,0.26416714778741,0.9044703123390546,
 0.12430867271012425,0.7420329067852492,0.4880521220103565,0.6975220131561756,
 0.457444999506485,0.18588177948110263,0.5452155867934786,0.5364081243116322,
 0.40814407655186036,0.5394265701532721,0.3381117708214808,0.618289973578227,
 0.9243873920503106,0.22471472690280825,0.18275272227144423,
 0.9215356272983972,0.2654346025241491,0.9071162604292007,0.4097093803276848,
 0.5255617893271721,0.48015432882869147,0.49804379830853296,
 0.17034809358357295,0.006785868022039732,0.12095534121028118,
 0.907766945383147,0.057662123371209906,0.39452165896707814,
 0.6650942960059253,0.23300425477433617,0.44719016490510755,
 0.6034897812214287,0.6671150732059297,0.16570602170864168,0.7726835582221225,
 0.9472026503949161,0.9815763785501507,0.8125862043774614,0.38508341169170845,
 0.13970095135154315,0.980761426046537,0.39904742086935285,0.527729130911067,
 0.8598203506673573,0.30343286840065,0.3786915638187893,0.1460313363640603,
 0.42435246501989954,0.3269157343703738,0.4582057289344553,0.6679849844770598,
 0.6523703666333892,0.07795257730529959,0.20082177029179227,
 0.9533038946523977,0.04658288793106091,0.3619312295416637,0.8290924215278668,
 0.5764659855460798,0.5160462453217942,0.3215501850575766,0.7757853427581892,
 0.13453271066887895,0.4817121328342474,0.8958948881421747,0.2148235836638711,
 0.10481077974769415,0.845593431893351,0.5166671609115212,0.6855231888247505,
 0.540808030936231,0.30719736475260717,0.690209242301371,0.3186683821672085,
 0.7640012549464192,0.8404278187985841,0.7050373091103235,0.6350001004931985,
 0.7666533318556459,0.8654378545589089,0.770172337958338,0.48057867324683234,
 0.4960670519772369,0.8816232553124436,0.3861263597826675,0.8054071731809502,
 0.6030420281381393,0.7289151846885801,0.3694483649578332,0.8038486447781708,
 0.9091432136375434,0.34807891394139856,0.8288060667936983,0.9705575246539573,
 0.7064756526331848,0.0893397021248743,0.025050894620319952,
 0.3443171471765931,0.2020751670748746,0.23647514605433606,0.9135616279448538,
 0.6547034552421869,0.6661920700609817,0.8656292992686258,0.12996604669987866,
 0.19709086047906632,0.8605730096610501,0.035268709792275366,
 0.5498361688240624,0.1762957832590285,0.2899364918193139,0.9831710075242418,
 0.9885604347069252,0.7126084151064218,0.3626034350436197,0.9572633926434777,
 0.9103761919109336,0.09047733854139384,0.2102884369975434,0.4519135060573748,
 0.048165228806975824,0.7442699104861317,0.5413408874311759,
 0.10721273944729448,0.43897208399535825,0.7160669311972699,
 0.2831317449609715,0.1144114991173033,0.1963580874556069,0.9397216373227676,
 0.7997372295193266,0.8523731506559303,0.1146718572005061,0.9520076134210466,
 0.01506970451413514,0.27935166264818945,0.019088462096286785,
 0.6398152646252077,0.04330181820845025,0.8323260430567168,0.876160129559961,
 0.42341531854730075,0.752230788984389,0.0637513989258024,0.276100413291426,
 0.15033636592641453,0.7838098864654768,0.7128990176591077,0.6535299656325605,
 0.11189725118301219,0.33829367108217145,0.9391434368661404,
 0.3386068351776177,0.1509850371929028,0.1964496353144798,0.9322320657460297,
 0.8730704214933036,0.5885734827940832,0.9050948187148677,0.5605418158651778,
 0.6312791604728175,0.9821908402820102,0.8333364241821518,0.8465051396684518,
 0.8153646710044417,0.2746603729347883,0.9540744422307098,0.5913050062174249,
 0.3577024167701085,0.5616840669247967,0.038661772929982186,
 0.7810865497080534,0.4176371187574772,0.04162667066949943,0.8783679898693671,
 0.9828699489788638,0.8348055050475081,0.3503623407403331,0.2738202093798713,
 0.6027710538464588,0.5213875447091246,0.15809395442423202,
 0.41499753199807976,0.028118979839630742,0.20463555994739868,
 0.27692931740231097,0.40434204220639514,0.84053317095447,0.1171811815618029,
 0.12008088220885504,0.21327882451402458,0.8979738239101687,
 0.4741359322445162,0.6968811794573878,0.8828569186447601,0.8437913802359036,
 0.9154607698786776,0.4139793419069069,0.3246683065688316,0.4627903306738176,
 0.00996816697039471,0.27199807959438327,0.06261058084832105,
 0.9651570301752953,0.14755937922021278,0.668663011322262,0.6079944518423461,
 0.871758408141078,0.9768384552446467,0.9823679794476826,0.6831853259533973,
 0.42984494205048474,0.9105218401527702,0.9879422910065386,0.7404415806201288,
 0.8589161029583275,0.9268634032070393,0.36079681387084694,0.7370571210822344,
 0.2898125412167055,0.8654992349128734,0.08517040112844598,
 0.25636662744533534,0.6249315775954222,0.6397002116867058,
 0.26717003156752794,0.49388943456475076,0.7722975210545998,
 0.9929609014520493,0.4269901515155804,0.9333220981700241,0.10722146069503902,
 0.2009848661869088,0.49581759535125614,0.9811693015760931,
 0.09475104825877745,0.942388802808285,0.39030967210524525,
 0.08315477973907459,0.09126091146884796,0.10396616808814507,
 0.655696056629168,0.07105371300466856,0.9932435410854733,0.9758607434817353,
 0.39051575120874127,0.31401980787264594,0.1718505041105458,
 0.7097193959885225,0.8896168096740944,0.47422784759161885,0.9025900494586395,
 0.4769435671819563,0.682584663043683,0.4627208351133656,0.24877519315792485,
 0.789489021062823,0.07706116277920638,0.5841187828993764,0.9484923330360341,
 0.6683106639520527,0.11573770896553004,0.9693038466365272,0.2966534960373258,
 0.9870984412725643,0.5283460171956925,0.27090877886500087,0.5502455196656206,
 0.1327859459606493,0.9597817132727966,0.18466411903744961,0.8434407979157187,
 0.4855110690311727,0.2182738898677976,0.5608824468183848,0.8370195935339545,
 0.43921666821354166,0.5762681465452204,0.839292929073451,0.4113474557281177,
 0.34302129245720825,0.7365505071457128,0.7877744578235318,
 0.49913702017369366,0.7007235197984014,0.9602385158747714,0.853496970287434,
 0.39698083362012304,0.7731107818787541,0.23798944778802977,
 0.6250615312043853,0.9723248098857482,0.9653208239243501,0.8142670227465105,
 0.6192163045304013,0.41069699644696334,0.728947357028064,0.833034513909832,
 0.7293725825389122,0.5410115093626118,0.5371760711167531,0.5717666610534865,
 0.45358248964100245,0.4287375636667653,0.5343043595170691,0.7655046233451112,
 0.11574923663031222,0.9722527838235041,0.7566439343387563,0.8075155527475977,
 0.21657325144921535,0.7523639047632255,0.49761077447015434,
 0.4520189097684286,0.17433876151687167,0.4590059943895277,0.6413867812783889,
 0.7873896914726322,0.6902326282765434,0.12356445427464657,0.5726460323145856,
 0.41309384690127127,0.037296735831319316,0.41247949567884523,
 0.903134489998881,0.43825670699053054,0.8353129456395345,0.7845277662748551,
 0.346472718771333,0.5279099361568921,0.7465635396595336,0.8637189153146574,
 0.3148221358510477,0.581822530020608,0.4297219581857601,0.07080255348701192,
 0.2660148449189562,0.9924290442408175,0.9548469710876547,0.6319629053925705,
 0.5178711679898024,0.493486345072059,0.37675352353787894,0.3733675111347713,
 0.3098418240686155,0.4723528860715158,0.5497227935578848,0.4677697969875242,
 0.5022861646234322,0.13062283575081546,0.5074923636712204,0.3600937471211181,
 0.37659242860909503,0.42396576024366583,0.7395772518268829,
 0.4733992255140804,0.038811269691979455,0.8958201528463513,0.947321429175539,
 0.9445923989460513,0.7851381680939151,0.2342594716652988,0.08921241893680565,
 0.09758960467091371,0.1422682366671788,0.19422815917453695,
 0.4578063017725865,0.6140103086805151,0.514192492917563,0.04238462602232662,
 0.7403040366452609,0.7529065236548281,0.3590656213452105,0.5218509495903593,
 0.057358354620222984,0.4851323546392683,0.4531299441652471,
 0.9812616365872141,0.830685615125032,0.1448430555571496,0.49550920767674667,
 0.18626837661984208,0.5464106034947858,0.7549576829466992,0.1110801212525443,
 0.11580363861125176,0.8379134871831052,0.04675611967069371,
 0.04131512102855017,0.8412971964385954,0.18075073544222364,
 0.7391761014910742,0.770263493693156,0.012733906432739195,0.3886692972855934,
 0.6605639093148261,0.5008181412479572,0.24778075056653948,
 0.25084086586961607,0.974484130545185,0.541476581821031,0.1550002519054533,
 0.13104764083121156,0.36770538526586827,0.017621859814583085,
 0.4823591009984014,0.4414815465917993,0.5118310760935604,0.14557220034529272,
 0.21439924035910762,0.9692772214889255,0.45218001759007787,
 0.22031387135467184,0.8661920253401726,0.7071514334402224,0.231224460470821,
 0.3098450718399782,0.042708730767158753,0.81766866454463,0.980495291311239,
 0.3126342120901342,0.9578444639239545,0.05682199303601687,0.6042156446984202,
 0.34656256594695467,0.4873369499775193,0.1472927838918936,0.9793625494689359,
 0.5446206199110827,0.14898730617231004,0.0660901933930268,
 0.18337308006675224,0.8974808835339924,0.2313611544011257,0.830286595815595,
 0.052053440430526954,0.786399844802364,0.20033336000439328,
 0.7580833726142215,0.14491866737772385,0.7503348854774423,0.2045114205861316,
 0.9777908987308648,0.8801379636392715,0.9801515605663786,
 0.027686900117536828,0.7743142051184468,0.9632740224591116,
 0.6542122704179032,0.09210488601726696,0.584560317054084,0.8309640450338459,
 0.06225449790251858,0.29136775122989966,0.48677191025319033,
 0.024539648866295116,0.07341888881556513,0.6834529232360681,
 0.8901857172387017,0.961474297995811,0.010744850336816492,
 0.18193971560885702,0.014488573537592808,0.16431346693238502,
 0.5105083969080226,0.06981552891276399,0.8816987869689566,
 0.09958949895388258,0.8583619207430744,0.31940341250580795,
 0.09960844226764609,0.06178324625196452,0.4897300991957466,
 0.12140933805604581,0.9343250937462902,0.1385526353725166,
 0.08185756252056131,0.79490140100152,0.24552061763847577,0.6175848731182202,
 0.7869385847178625,0.3681384774278441,0.8547419156264067,0.09084415174682903,
 0.21535427966979515,0.4546865997170223,0.15707806812055947,
 0.17637033767664678,0.6344875577076511,0.9018277587706268,
 0.05676005095709935,0.46954275418934366,0.25665085550948785,
 0.8362380926559349,0.5281707234778732,0.5543300428039197,0.2869808412848349,
 0.2177524181208068,0.8648245378920136,0.8456476689718899,0.6333370362622792,
 0.47485991342420275,0.1553918987626025,0.5810274420644731,0.7123933763746438,
 0.22880214734020554,0.7137858694241227,0.2392577161442342,
 0.30769078353786306,0.779549897663962,0.5991877126485257,0.433930777616775,
 0.8775031836635929,0.5189942521695019,0.10490486621479944,0.4863321533906748,
 0.9729201588602504,0.3730229492473638,0.037009390863969926,
 0.5553021809374646,0.9779442615469321,0.40021665299720044,
 0.20277238062637792,0.6215783093892476,0.7830339183115584,0.9752923875759469,
 0.6588032349414936,0.25963736519969083,0.8117710635337605,
 0.015494304777731838,0.5877787617221499,0.8907686633398695,
 0.5836580007207924,0.35831474732769053,0.2036176183452988,0.5749795572845722,
 0.2543061632647359,0.1938979239764641,0.35911672742065426,0.673237404568323,
 0.4708827892669093,0.45293823322230764,0.2811609994566562,0.5936104202370962,
 0.8249271646341659,0.3938623927428341,0.9776103855882025,0.15320319285368855,
 0.7982672078461888,0.39212292983404484,0.1719179976739632,0.9782959098768369,
 0.48515453066112435,0.8470705631647171,0.28641294743632706,
 0.29067257130171065,0.6052655814453486,0.8264396572484464,0.4472742064878579,
 0.5817351102091488,0.7384858600408709,0.13075845885547877,0.7249160670950272,
 0.3902928514155818,0.06127336355390889,0.142374520756403,0.1294097308938833,
 0.8284156805841962,0.8075597309828997,0.17947999061569408,
 0.08585168039183455,0.5098207878015256,0.7139577383271433,0.5771251524028883,
 0.2576849563392931,0.5520064437344482,0.711480121403779,0.043391090783774566,
 0.8042704261327007,0.007801972219064979,0.19920146669917105,
 0.6470708377607158,0.3613810702542646,0.2738692350426809,0.7211047423757226,
 0.5765372794857707,0.6343686265751791,0.37986532999716227,
 0.22827918711648387,0.3312657206656219,0.990181917044249,0.16579269345191794,
 0.43030578818470744,0.32677489251265346,0.5811990891671815,
 0.9113343417132848,0.19759736907980197,0.01461792546130336,
 0.9422506278640905,0.20355308445106335,0.8442653997140487,0.7570459824528873,
 0.3968787309408982,0.6729431632720392,0.17599895009565458,0.9747896075131709,
 0.9870806672922661,0.19700493464415514,0.008547696732342902,
 0.26683771016356606,0.3805221398746301,0.5094792173657123,0.8971089154551891,
 0.661704281248426,0.9871805630343989,0.7626843247431597,0.3758412579263267,
 0.8066158159232161,0.8089476016163296,0.784111069152698,0.2558667489650548,
 0.8552452499240586,0.8532002583098955,0.2703672057972317,0.833137492811474,
 0.2063118339711697,0.8222112427680975,0.17856620258973743,
 0.03043272062951008,0.4425685206471386,0.1291982277785897,0.8133911013037383,
 0.09145615767648474,0.5173252342249564,0.3153357234288535,
 0.18579972617638252,0.046292575461115426,0.1514098983788451,
 0.44641202771665833,0.7720608810625278,0.5225028879741623,
 0.11075908198350282,0.15680949883555562,0.9632353506547019,
 0.0587135183448666,0.6566971000061907,0.3281117987112778,0.679264525425984,
 0.6928424167739464,0.9340939256625869,0.3356313036720352,0.14941155953273544,
 0.5290763556124983,0.18756553825537403,0.8029725920701752,
 0.039070937461548216,0.7322798525654316,0.5496690898990594,
 0.9932485631067158,0.8289491504263216,0.43124380641239013,0.4521639420728545,
 0.9539811812261636,0.06843489359520805,0.37115058754922314,
 0.5576828648523172,0.36142647114203674,0.7141271302962511,0.1405864168332429,
 0.5341207245884002,0.08882543799709675,0.6422375425872547,
 0.02027501503832574,0.4073503877634176,0.6980139272392782,0.4435983821886862,
 0.21776023732802063,0.6886879639339657,0.0384264714313165,
 0.29421433436640965,0.014510419772805427,0.25344274666018196,
 0.13468600852082013,0.5998103826544295,0.7715092500539076,0.3686894744548177,
 0.5054411532942749,0.172501945301621,0.5551756050916311,0.3614805492850911,
 0.7940276112394878,0.5461919326572753,0.5546834854140537,0.19040865900611526,
 0.6110970851420799,0.3313543642136495,0.09965301534237292,
 0.03332640264706732,0.9073111968150992,0.600913007392032,0.994166657285207,
 0.9153503523229086,0.3978831540430221,0.9343104895911715,0.28720450002389963,
 0.5683350448670941,0.8045932859993555,0.5570964510982266,0.7759078955732024,
 0.20151782871087687,0.48764176498940837,0.36159048262684856,
 0.9556004389742078,0.04857199232637477,0.6795387840135613,0.9487791170348462,
 0.1943267433871625,0.9818015535617256,0.786145233896555,0.5599559851470597,
 0.7226724350678251,0.9068402822192042,0.1095287264425775,0.22444677055448747,
 0.3909074197354032,0.8652175293374964,0.717925086652017,0.9428067682711255,
 0.44754182255103725,0.3994397156810069,0.8215526438053533,0.2361043109683777,
 0.8785507018019473,0.15314312372136785,0.0656731858004369,0.2474782516281323,
 0.17010686489008386,0.7458109524295847,0.22284486531134506,
 0.9340196961297881,0.3816622079327956,0.0923671230058365,0.22506348959900102,
 0.45398124046639254,0.20661290051377423,0.7542835919917218,
 0.6443261182188481,0.36942341680522983,0.32163144318121706,0.590314509246873,
 0.29470242690373083,0.9739078411884892,0.9547337048003133,0.9583226738507711,
 0.2669222220044398,0.7629843989459649,0.6733078971675371,0.221978927537702,
 0.5653552277633563,0.02896501264860285,0.514962692979501,0.2562791217096726,
 0.7146575144695778,0.7213112902716226,0.8591773019129669,0.01587756368402171,
 0.3750721449007348,0.5076750440913989,0.838131825447119,0.6224811206382536,
 0.7042895950378862,0.10772048707340842,0.32010413250977054,
 0.12887765633027726,0.10858948257287238,0.7473120722695517,
 0.21480821551734786,0.6304475916825508,0.686326046457788,0.9093119432341745,
 0.2549746873560328,0.09179808842280468,0.4772370050055731,0.6627969693044587,
 0.4610239791528862,0.5400142905427943,0.2747305826408092,0.5621879347190093,
 0.025983243062352224,0.6315269266341406,0.8631097549050796,
 0.10893445874418894,0.9462436099726581,0.4524648514822541,0.9526791859057742,
 0.1807119694679895,0.6046338873453199,0.417461499191203,0.4059081409165326,
 0.9608132952494183,0.48319719825477414,0.2597018779247372,0.8631828750786277,
 0.772195024906093,0.5132534239695501,0.659161304137947,0.7225923615773232,
 0.674267140449192,0.08553483005103957,0.9639398988808993,0.942457629122989,
 0.2646680827527379,0.2388132126251833,0.39114327459829623,0.7882049640692871,
 0.7667678369106845,0.7373103657982385,0.09793426519599202,0.7337180474052083,
 0.49386871133426524,0.8913128586432766,0.16847862463197438,
 0.5689324249861751,0.24299174284753877,0.06190517079730473,
 0.32435924536816463,0.7132800359177657,0.5810996311857513,
 0.23807819277823772,0.756462525232324,0.9459281342502678,0.10560196652106157,
 0.3530032974502707,0.037121477635449596,0.6186587313231202,
 0.13583586220106791,0.37925062206270277,0.6188707158927484,
 0.7439401791738196,0.344064781391765,0.43290049843115597,0.04939999961985486,
 0.8453224723432643,0.37148798111039916,0.7318538158973479,
 0.06445127498185121,0.6056419445952782,0.9628710090507778,
 0.36097149724346944,0.0043015139849724715,0.7868597167498627,
 0.4531781466558009,0.6018722668327798,0.6883389123798404,0.0334866544982112,
 0.2703572283894887,0.693887668664576,0.052295877770861976,0.8803542166647049,
 0.3005917300766787,0.6254329574331629,0.8613993687272464,0.5965156649921072,
 0.9782410110317614,0.7393729351772717,0.6872234418198482,0.08461993871137219,
 0.4745233004905073,0.8169394616965949,0.09169267008792226,0.8982618495495255,
 0.13602400193508535,0.6927698665871106,0.6206642421433817,0.8601688138954039,
 0.2780650947324609,0.8804252469802087,0.43069475995041284,0.4232442416310592,
 0.45371384712050333,0.28121537622985826,0.9182172926595511,
 0.8988307407087406,0.2282692788836691,0.9483516414042922,0.731167446685577,
 0.06068989464222452,0.18408551452386845,0.07822944852729774,
 0.23072215713724253,0.08490249080173151,0.5713367258717403,
 0.26494103768087673,0.9226928667693155,0.3559048328095509,0.9443084549226505,
 0.18988185958790293,0.09990020345082362,0.3822547845738067,
 0.6747394252872573,0.46047286549759847,0.41095519567558203,
 0.9252967094088502,0.9207212942298253,0.04233440002916533,0.9592486128351676,
 0.8184688136368694,0.05831977362494922,0.3236274270969932,
 0.16988810785358832,0.8119059602129787,0.6214483795385545,
 0.057378175840472556,0.9921814671797955,0.7971619221156648,
 0.22216923413041745,0.06064124977920171,0.8642139989407261,
 0.05169557916109013,0.007228219806605929,0.9583553617425973,
 0.5537484495568653,0.5542173176305207,0.9012602000016785,0.4750998144327365,
 0.14694553425723478,0.1772854306462195,0.8856483772146182,0.2709513382083969,
 0.30005890022218895,0.4231841075177658,0.8287381902764465,0.8537773441306945,
 0.01644046629444118,0.9625318598458991,0.36000644894000045,
 0.40962503711196274,0.23669947743485098,0.6010861463921867,
 0.9304876653555447,0.34066598889041166,0.7144389013702788,
 0.19158708048816675,0.04248566391966424,0.9820991270183068,
 0.41047102550469694,0.8794702992687032,0.590434859534052,0.41668618024374915,
 0.06669003814913732,0.7801342506942692,0.5615480018434933,0.9156862826753545,
 0.433895593764288,0.7776085388178751,0.3420505307915904,0.3872482903963519,
 0.8484560656570569,0.29103594003276223,0.40373392222319426,
 0.7395668086304186,0.8526439851672479,0.6211876109622634,0.8368402028210066,
 0.6321725603555953,0.010595840444359128,0.7198030218729508,
 0.6978089863864162,0.7815612049145596,0.7762667323931661,0.7479474551107036,
 0.5003504517466368,0.5537602315856096,0.7772664922899089,0.32116098873177434,
 0.5238770855489377,0.5409339147394864,0.4267364001278109,0.1228703739217305,
 0.12503536558692618,0.7682986028810187,0.3925202193562908,0.3782049638571443,
 0.9779465173157572,0.08775888484114835,0.9790093977072113,
 0.04537809687648864,0.6012129722504982,0.8679703020960966,0.2769416192763492,
 0.9968655598672637,0.9943827428523981,0.6750223380226837,0.4594127558861476,
 0.7791746335874574,0.16682675966716343,0.8304648582011772,
 0.26256420651882706,0.9147552753268767,0.12910860692468595,
 0.7286690195094045,0.8548884187102406,0.16767192826737365,0.9775979086530046,
 0.06428681450377338,0.38344234261538257,0.9685805760471006,
 0.8667549686127846,0.5728668588999437,0.4038824869263845,0.45392931989564955,
 0.9704470423145611,0.32487644378918,0.9388137166612543,0.7257377451888817,
 0.7015207016971716,0.020951929543326253,0.8553060969203663,
 0.8471112281137723,0.600374443785805,0.14123668954752733,0.19683817159679096,
 0.6919840398200795,0.4223994526363718,0.44887520424274996,
 0.21430034218485172,0.5924072998470915,0.9468354548844693,
 0.14408978684892115,0.5600567185986525,0.24400156755733315,0.017896823681399,
 0.08068278898631087,0.053020621567131165,0.7820526333494942,
 0.06668633479467412,0.44117322513401525,0.6825717682120395,
 0.1339269419393977,0.7353133514060461,0.03651109924236362,
 0.043475739407626945,0.8292455122166065,0.42087787218048567,
 0.5302382060518472,0.9683696836046302,0.6388239667076335,0.43555503715366684,
 0.7388869102001933,0.6591043241538798,0.2594456434275163,0.2778734466366306,
 0.5610650087086426,0.2130785112321425,0.2799959917467024,0.9436739689347774,
 0.42214945416839966,0.7092547565034534,0.8500269990644057,0.6921183515406658,
 0.6328298186094677,0.23309058476077582,0.5161027686675396,0.4214372203948825,
 0.5923418578693598,0.7940943868908695,0.7030785185927683,0.7523154584138393,
 0.40384836288524817,0.11614683012451488,0.12965502472799806,
 0.13779237930840482,0.37361331449087876,0.6098236292499841,
 0.04912274658043181,0.8485057102759928,0.7203880214275534,0.6703758912739356,
 0.7641232214058584,0.36474691429621753,0.4485824434555369,
 0.08072368691839193,0.11990283949733449,0.13177148531997018,
 0.2817218071713282,0.8368671733002626,0.5671876272487338,0.6852053400547412,
 0.8841667417211146,0.5391623898973412,0.7268185782645291,0.813598008553339,
 0.1467845771446461,0.20762023391978512,0.623620060255553,0.04524564882502413,
 0.7891136241425498,0.21806684010821953,0.9311348745065108,0.8922246515559703,
 0.191369014422474,0.12810189215520307,0.2403619184175685,0.8257703658251158,
 0.6384421653488648,0.09909213979892773,0.04794736112161291,
 0.8820295082181872,0.0978343041241374,0.8843850804648397,0.8374948019510488,
 0.052679663257415354,0.703672942651415,0.6502542493987901,0.4003431098044954,
 0.2055755654659308,0.05751145406481606,0.29006425794844115,
 0.5493730816811324,0.4141438993685487,0.0404673698192638,0.9675253534252489,
 0.3197872630440559,0.1799431209286123,0.5541250341975019,0.45418038895615087,
 0.18942610194468723,0.056363167459165586,0.6458382405442392,
 0.31298302926981614,0.9708768295604236,0.6606481111812879,0.7032604072130554,
 0.9740763033844636,0.2711592249527035,0.7838246089225263,0.4060214731130025,
 0.9014844492958067,0.4409861896013334,0.2047289033310442,0.8797899658738341,
 0.21395135427785672,0.08735524689776031,0.12861435288497902,
 0.07746628944118561,0.29672424536806263,0.6238872561462925,
 0.038999853642254445,0.390882096745081,0.2433334309130567,0.3012197386447908,
 0.9317280490281822,0.5173719255363171,0.716492628566429,0.22942739846939375,
 0.6200270357183745,0.675727493557379,0.9619507733278944,0.8157881517494938,
 0.7644387452069896,0.37489706555599334,0.9787435856990577,0.4817225597264495,
 0.950850352666454,0.8681430318401495,0.7162786190027003,0.24058378059020225,
 0.7652710220128237,0.7019310582710054,0.6417698830611578,
 0.057652183034298976,0.44414356287282486,0.13816440311622147,
 0.27676694658022205,0.31984912661805276,0.7306792101394175,
 0.9823823692941333,0.7418557297374998,0.14791605239701888,
 0.27652016296583926,0.11068499474949678,0.6901719910867645,
 0.6778532367889427,0.40122157147634807,0.4606360883581053,
 0.28755157034587164,0.118332781794243,0.4079004456638671,
 0.034528178706798474,0.43580263694727195,0.9504130640621679,
 0.5220326549329499,0.8027160080248265,0.03712059416912983,
 0.027929452048733983,0.9061524556262063,0.48276955080318384,
 0.006928428538643727,0.9471600220644995,0.6460713453305638,
 0.42640492512719885,0.246159199548175,0.8920272052862531,0.49610401090173906,
 0.47769004684559224,0.5920809677382327,0.8063378624418451,0.9699614270403671,
 0.31255549410846195,0.8460920763258988,0.9911305902628875,0.4084490148701756,
 0.7072103152126139,0.28802983203589017,0.4534954031647227,0.8970738400289454,
 0.03444288466767664,0.8021893413186209,0.09646434127102477,
 0.5238222494485218,0.7027854302002691,0.8119685926419051,0.33155690778729296,
 0.1735076182780012,0.5658226697630994,0.7508250091495678,0.4110066906509564,
 0.7901043723088282,0.15815227057671044,0.1366519252209074,0.9030039344557841,
 0.10750178480652306,0.7218517997490973,0.38675732516303896,
 0.26872828664865644,0.7773965221228845,0.19300848449550223,
 0.37044383015185467,0.9508141871906032,0.668162497618061,0.3088316781035637,
 0.9628690526766217,0.4734946560859399,0.8753368680955312,
 0.003144436637027792,0.9188984891749714,0.0707199286107898,
 0.18917316078176893,0.032771304754002806,0.28927855633483135,
 0.05427546801757388,0.3966665819351247,0.18818671167174117,
 0.8237947201063553,0.32316602299047403,0.5298871688924568,0.9982190874698756,
 0.9502133872307027,0.5590329004851718,0.40809576135877723,
 0.05318525057596024,0.10097567338681979,0.15223477190866008,
 0.13532962191979925,0.9247879367929185,0.8783736728356109,
 0.46349052664328105,0.5436367307455995,0.46434025200598394,
 0.3780704934265755,0.8035382183046267,0.7370487125588913,0.824602494636793,
 0.03904017177145935,0.13198051285573054,0.3129768751021573,
 0.7512488593704836,0.7786703377004397,0.5861072588776549,0.14631632998939437,
 0.5796093471287681,0.6639602788618666,0.5655884253075306,0.5428470370909283,
 0.3808755292156416,0.47496436219457416,0.9795845959384015,
 0.17050413861629643,0.2460419803955204,0.9105622336930859,0.6080581917679073,
 0.5826525970952429,0.7839131660050125,0.6555265037662337,0.9876435917935752,
 0.7827441030286282,0.9946630298143881,0.30417491473291747,0.3877098376253967,
 0.030441615593108096,0.8258322476913108,0.5908014192812422,
 0.03128327950627108,0.017483550456053765,0.9072468419926865,
 0.11426640421313072,0.46089202931562134,0.8501071505058648,
 0.8361130143853307,0.7233908028015955,0.3867175042972959,0.5357693001183109,
 0.24881837670818852,0.040675451429555864,0.3342202025510168,
 0.8937787530527634,0.11256644507656444,0.12548279103719995,
 0.7771514022226694,0.6092699832501973,0.7291491278531352,0.5494150025302229,
 0.5115721202933367,0.24811017383661105,0.604913322825349,0.8492084956675016,
 0.4752653754514855,0.6815922794051263,0.7985595551793048,0.9877948021584357,
 0.7178809420951447,0.2172433985687291,0.7517528501211421,0.6058207458158599,
 0.5006448869805454,0.1817638014318168,0.2909958219409654,0.9024633816025328,
 0.6866685378976587,0.645641073041763,0.07130446084790987,0.7208769574614583,
 0.23189449250436567,0.058151733778338155,0.3648020676312027,
 0.6274809594832114,0.31707439812695815,0.5995276511599963,0.924141511146051,
 0.5862130756747349,0.6897615076637353,0.8870806017822974,0.5599372751601541,
 0.03136565461934937,0.11363554320308544,0.8520627875630905,
 0.13419633970777245,0.5675400576680149,0.5884024827891974,0.6824424083061166,
 0.4970463086115847,0.8808930141166849,0.6526752194392977,0.8750417463604937,
 0.54989636913263,0.23781647277021434,0.386021262383474,0.04727344382039533,
 0.01647610404102684,0.4009368633477405,0.09611951251134532,
 0.45877203225597796,0.39574239270143496,0.7767134790330094,
 0.04059798204784626,0.612271724919286,0.859148422312595,0.32008786724438143,
 0.12001776851271884,0.20043819643246952,0.9060688498991866,
 0.8616738636903705,0.638826215387561,0.9379085373307297,0.4343008728684774,
 0.11681491634147045,0.15250504483140137,0.03402750648552755,
 0.37740883686105675,0.6905129892425198,0.7417684558903059,
 0.35933849016134145,0.6680252837010181,0.7432335783832373,0.2297592431138551,
 0.6871895905241654,0.8786459008162955,0.21431995531574688,0.937256580486806,
 0.5280303212048059,0.6522490087011283,0.1851739927536803,0.02312340138589919,
 0.8554463265629374,0.25816590599692946,0.30519409211492965,
 0.7877448180953105,0.9555157128555042,0.6817712469406766,0.800301790431448,
 0.415352472050447,0.41958852938370117,0.28879790280122886,
 0.15316416108724296,0.7386338460642581,0.6770935555877075,0.583516310724089,
 0.4961869075480686,0.18188320929600998,0.3284458014681979,
 0.48970922313743215,0.24733784933758918,0.29680419252912804,
 0.6274651668332147,0.8654679807989617,0.9382490380600121,0.18204535380098985,
 0.5285610494649423,0.2977584733121589,0.0594490502710896,0.80407999265841,
 0.23399674814074578,0.21836277601276888,0.7031384960362703,
 0.8402810494432854,0.8460404436733686,0.15318837845860822,
 0.17390308866208237,0.5323361842918691,0.5617045813275131,0.8447705890643364,
 0.22919226330272724,0.6869515499462081,0.5684315802877825,0.3128131136146719,
 0.5772862697625929,0.3651205562212415,0.5662481712323527,0.015970898769328,
 0.4408768561868649,0.2477870468434491,0.6364112242716669,0.6787384104773614,
 0.19167101120458896,0.0982593330554492,0.16631038923243435,
 0.8338621533501829,0.2147303403688674,0.3158385675200325,0.33718355923359855,
 0.4008437126580522,0.502633114396817,3.776897933231993e-4,0.9165449269844171,
 0.2864904326769596,0.6634458733025922,0.3332288652887835,0.12852542701790615,
 0.9464980181204282,0.4551200168602788,0.20508446787692747,
 0.22275354723606455,0.2689482643437826,0.21966962986543315,
 0.9443862002774113,0.8963381128015367,0.46765019205711567,0.8764975940276358,
 0.5456386043166961,0.268454648039098,0.04535486201020422,0.6952680311509739,
 0.03701814049131491,0.502586016027019,0.7478469435994427,0.9212699857528726,
 0.0566819324324942,0.9880365876150047,0.5163126483867458,0.07184705672992397,
 0.523460434288521,0.7791410479631302,0.5307103860527903,0.381452734532327,
 0.8668106063039565,0.054012432302475544,0.5596650674691371,
 0.27720625862685067,0.7190556650431669,0.6230202180406685,0.3547593576650372,
 0.9494556652131273,0.8335139894830567,0.31856863158414483,
 0.38615067127843195,0.5411365487710627,0.5092997058893709,
 0.36127352126519796,0.1423477873714385,0.3940013001356657,0.5826265616254546,
 0.2696985887495591,0.6147208190933525,0.33990241963395496,
 0.07656123401791903,0.5055826098607225,0.1534459905286295,0.5961280768699322,
 0.05251481081086107,0.6703839062997811,0.46581583484641986,
 0.5637357977876226,0.40635127507395674,0.49501095726739053,
 0.34450826989025907,0.9678046439620771,0.6971602325701033,0.5573185283232963,
 0.43108452986317647,0.6739348832856944,0.8317177971346141,0.1823833163134746,
 0.7327470714810718,0.7351769019566335,0.853907459056822,0.8032561569174813,
 0.6661937309939059,0.26150285503906023,0.7657641594641479,0.2518154277300588,
 0.3677637726751364,0.847144454471198,0.17655556584280985,0.3674896796723436,
 0.14017103160727395,0.6727730892636284,0.761964847764773,0.6494648871933539,
 0.396840497947732,0.47834726503079594,0.5980590371372624,0.8742532860593287,
 0.1756619416014139,0.9793650165487566,0.7906712405716823,0.41761434192840863,
 0.6630462318194638,0.3008600987082456,0.9185698854039398,0.1724061397610439,
 0.4914858159687746,0.9958787247933268,0.5721282309023668,0.5820341452289596,
 0.08984555819556683,0.047072699270821805,0.6691769015913769,
 0.4093498924241783,0.8225328292792576,0.25159988980316506,0.3933065051850956,
 0.3791491721882848,0.6037106630560394,0.9273748474543326,0.7973559166487735,
 0.9033519917715835,0.20212809014699873,0.8319232727095113,0.6931748594426622,
 0.5236892001381723,0.35492610763330723,0.8625721483624119,0.9818833487951183,
 0.5342026124229828,0.5270859624081128,0.8010796026463137,0.12639696305570913,
 0.901075175860967,0.882917137823017,0.4741914190487454,0.41009716441391186,
 0.2534878269485662,0.9278290879716726,0.1828408834489117,0.18394016157983117,
 0.9946451597555978,0.5045829075816698,0.5629007276699992,0.931112935593889,
 0.3522481770845348,0.9877595630479555,0.9615809772297783,0.7533241147854217,
 0.7207249745961873,0.6877477708546844,0.16625968194633,0.2942299523194851,
 0.5224455089243736,0.11034200820800133,0.4492958161853047,
 0.28777247395772376,0.3842881205566664,0.7163531485521711,0.4925718236452741,
 0.3771967742713458,0.44508578441125257,0.6640659193015577,0.5645186942293355,
 0.002096749073192239,0.03005750253491435,0.6912170529265486,
 0.689163084997037,0.8894428774450339,0.8147949263713699,0.025256068300345658,
 0.3232060571925186,0.7072438070314768,0.9612246715504296,0.64104060592005,
 0.10617994875260062,0.9651897878633562,0.31033590824579715,
 0.4486792521167864,0.4124117507045708,0.7853654607659926,0.5211134416296505,
 0.7096124022176856,0.892083080856981,0.0434358465616117,0.7297376688678257,
 0.42840978727878853,0.438526612942841,0.3853830167409732,0.9399862428047009,
 0.5564951619713145,0.5001268367053457,0.4171553289173727,0.8756947853607552,
 0.7780668054845833,0.05320948286790517,0.185635854471202,0.6560619150869235,
 0.32765898689312445,0.6032608725094724,0.23495587971419607,0.948453324114115,
 0.6088050731278902,0.5398898152314855,0.43716480168412863,0.1129186512356819,
 0.3336681207815988,0.2538758287494369,0.5634223112924966,0.20612739964733162,
 0.3645402111415763,0.13214050035026692,0.06934314351640802,
 0.04754403009291197,0.7746346460891798,0.027909687545323303,
 0.20141980990506914,0.5032723208097121,0.38945124760967387,
 0.5989387956707846,0.11113430956911496,0.5500603649851348,0.9819086569763923,
 0.18300063460105132,0.5802385616496774,0.4776990324512371,
 0.38519460049643994,0.5993580119201197,0.7933492586674326,
 0.16017654072141396,0.18411880828028016,0.42740558874190915,
 0.9742263065952372,0.29179182533936465,0.3262504334420624,0.4868679367231935,
 0.033355552022962254,0.5953912877016736,0.3278093841663485,
 0.9746828082678359,0.01609065179777147,0.32362465195513845,
 0.9124399805079428,0.4651469754493003,0.8656578925880014,0.0824369414723648,
 0.47974743777525586,0.00821346344617957,0.062245622892242114,
 0.3391792084709111,0.6352830461593307,0.5465177853877875,0.6920392268280888,
 0.8107813207348051,0.3521171005767664,0.6405831901519833,0.267663278367533,
 0.9801570588435518,0.035613059914102574,0.8580493254694468,
 0.08032423882198125,0.044226742501714766,0.1570971279116533,
 0.9737474743258685,0.9436704858046896,0.18850880276137527,
 0.07365389669394318,0.8886364531464761,0.2100745601222177,0.1698580668805606,
 0.32387960721279097,0.8177804956342916,0.39513023889475996,
 0.40213879706186106,0.617748997172016,0.5798358697796446,0.7327868936940785,
 0.5280247353866058,0.006486665733932528,0.6137911810380804,0.139633606892949,
 0.32272754893336986,0.912112812263593,0.38310610894690844,0.4557120837499776,
 0.44334552066971256,0.3576739219953199,0.8354993127081921,
 0.023868937112116684,0.5479666815635666,0.7361308805195748,
 0.8342510791620279,0.3397924353578754,0.9149736239022579,0.9653565280922415,
 0.8721402518133209,0.9122487889410089,0.677950937390112,0.5269036268070202,
 0.32851073638172357,0.6033681562840362,0.33278730373704013,
 0.10261806309112398,0.9743009932723233,0.9588201984963783,0.2061713354585759,
 0.5658688680581966,0.8650914277641595,0.39834386410832234,
 0.28228852886316613,0.28338558035208916,0.8672066904395347,
 0.8377801011580595,0.5088149432071704,0.7461028018652951,0.3390757299130508,
 0.6767247612021672,0.4884314264272811,0.6884665313825861,0.4420079889082933,
 0.9350861114139235,0.8058830931489339,0.598548199755183,0.7434196072171716,
 0.9840760318433434,0.8681083267360176,0.299715130933097,0.9546055058008001,
 0.505277621594882,0.40553873557960884,0.6931021256749499,0.9724215921382016,
 0.13147731723866873,0.35520492314037944,0.19595592583809396,
 0.40030899133892495,0.282553187626261,0.2787924990726396,0.5639028852790302,
 0.0035251567165377473,0.20341321687522518,0.5493343882075508,
 0.9136135777887668,0.8427869622440266,0.6895151781216907,0.4163460084100109,
 0.4050338580871302,0.4688882421088264,0.3974444977167493,0.04073004200546393,
 0.5903670523703981,0.6996011881240329,0.013091628226647889,
 0.08860596010441502,0.1504938356425778,0.19925060029092778,
 0.8011564014401683,0.3286485132999134,0.5093517524382964,0.6132115238358354,
 0.8964116294147166,0.5556833294414949,0.989943463403566,0.2992029166857757,
 0.8475324334256789,0.5393432790292685,0.5437782053525383,0.27023340238953397,
 0.8228847793140662,0.6631278097860082,0.8305763059300285,0.8831255183347848,
 0.6160526334433456,0.9399763834515934,0.6946368132898575,0.4091973743101076,
 0.8789949707910274,0.7120715520488389,0.9621202183249535,0.2743078361828506,
 0.46029341930247925,0.25558930372377275,0.09894211965618183,
 0.019225014387546313,0.20095812040848327,0.4680862651261537,
 0.18380928867007196,0.3369733632675076,0.16928943061152057,
 0.4638129345375408,0.9643344639568802,0.5273050129805232,0.41963474566684356,
 0.6962872513052145,0.8629183105994203,0.9475423898028471,0.957166744786307,
 0.7248661174343383,0.3354476797357281,0.9984131860552986,0.8355107608499208,
 0.42762858926813774,0.5332753501323405,0.9214245610417486,0.6380462053080874,
 0.8732102068228108,0.41685507776399233,0.6234989278111246,
 0.38457341950467516,0.9197273387336686,0.7261650214388504,0.3645349239391532,
 0.21428984534883988,0.6779875191962975,0.9531895765706928,0.844294318747265,
 0.2075871325067833,0.10248683726766572,0.695080429879385,
 0.005008807714082186,0.915119408060748,0.44988509168252333,
 0.29257017205588354,0.05225536219700633,0.7001113841403241,
 0.7654324146529545,0.15150938863554153,0.016184465102900614,
 0.5765674292782514,0.8488327446659025,0.47505426198491474,0.4234190347591674,
 0.5369527129824428,0.07411780458838357,0.9793378697698537,0.9526798380691825,
 0.5935172151614689,0.38914857036905515,0.8449379894731555,0.26935978724694,
 0.2982874720539459,0.3414760040166651,0.3971969761275894,0.2350740372415232,
 0.08745293979922608,0.3275166957486826,0.2827651200192236,
 0.36032648655995225,0.5421000726894137,0.8828604819383762,
 0.22481721148461165,0.5025045326485232,0.26843290042947743,
 0.7562862037950394,0.7586598498085756,0.5519628292926029,0.9810634099023661,
 0.24204437504383536,0.7643254295186067,0.9409271261908332,0.6349711663772485,
 0.6354935648420942,0.5902787490384598,0.5836662366950698,0.10910879248058925,
 0.8927030318881858,0.3214141273149719,0.6577861624543647,0.4693016732410269,
 0.5043798420338462,0.574890595626989,0.5771811720111535,0.2241848691561954,
 0.7285828938391272,0.2558764610562507,0.06503933336499135,0.7452668875728236,
 0.7514402193809604,0.9154890439386728,0.5313071483516354,0.9939284319622202,
 0.6989087476052281,0.6239783551833001,0.4917649064822974,0.26028639417908805,
 0.553470370560844,0.7717872363188886,0.07685167104071766,0.39240256088060554,
 0.29740373770365247,0.6534014525104793,0.8274723226780798,
 0.44232671864074624,0.20325836452920454,0.40751240940138156,
 0.5108867324307531,0.7164130097093881,0.2749022198588348,0.2809615676523355,
 0.42365065558705595,0.24291320918858716,0.5620251512785458,
 0.29480914839141803,0.616630198956598,0.8503352253794061,0.17303961276879787,
 0.9157559109943897,0.48363633135709194,0.5073747542123438,0.8482800256723194,
 0.33004600138285134,0.9727860322954061,0.48687566916168923,
 0.6779944286413033,0.4956590089916497,0.6049467434548292,
 0.001254560871191801,0.1955183530481951,0.011207591815640128,
 0.5347312990872657,0.724957411950807,0.0843047482028364,0.9666774222408704,
 0.46056395222127666,0.8866194299615917,0.6631883848535141,0.6232496435045309,
 0.4516662597883013,0.33833122868269117,0.5083034960850364,0.94070579268497,
 0.942274202591901,0.4522408503268226,0.20944947733672004,0.07836837861658885,
 0.7499240836922868,0.9520881523352626,0.5648036401750807,0.5166849392340891,
 0.030699538410031146,0.8022893221213505,0.744604272413577,0.1797136839004888,
 0.5477403149986827,0.33504326725599154,0.01482176420817738,
 0.24811681876554748,0.03679159472779736,0.2476860158669716,
 0.0862870647596421,0.9314839429894313,0.22992218487708416,0.2547659797712575,
 0.4871798182561301,0.9039502048738389,0.6088031347575473,
 0.054889044111693375,0.4118193025046809,0.10589730388618546,
 0.6327715514085293,0.37811003452983294,0.1413714604010754,
 0.15942870166667789,0.27420567673365337,0.54949527127769,0.9991174960848114,
 0.5266256908474526,0.19974033409711234,0.49420971498306165,
 0.07559374132999674,0.06696872166290269,0.3479936630664341,
 0.12920782836593747,0.34488649624535195,0.6044470827982448,
 0.9979433790611278,0.6335668348377477,0.8971401710334339,0.511199054005803,
 0.9436049612442271,0.7067563836589932,0.9807144799196968,0.9229072681695056,
 0.40734892366346065,0.7009611763089573,0.7572534689535483,0.8754012944992907,
 0.4002735927608979,0.47166329312226274,0.4847911778352658,0.3093447420491393,
 0.3123530116439863,0.10019146690941527,0.8934352365149936,0.9150026264605767,
 0.7980126827465582,0.6639412924792989,0.007404491229698107,
 0.1536988985296026,0.4716728651768455,0.08343343284006144,0.4626501262480406,
 0.8045310712413389,0.8703657415154614,0.6486707124164101,0.3383318246072403,
 0.5719611492664798,0.5696384322806843,0.14691445621338262,0.528342407175347,
 0.35388476710895544,0.6734733534584504,0.48177962945379316,
 0.19792837018376308,0.3870776965155742,0.6841318528032146,0.7713992138201266,
 0.12976796941988444,0.8263315373389057,0.7406353451608967,0.6197226201863199,
 0.05429013435514751,0.4254486958354895,0.4315107015328373,0.3637698371171051,
 0.5585225442203204,0.13789780880518898,0.7274226962735898,0.9593548371645034,
 0.1663667802852571,0.7629390004983931,0.4118195792102809,0.4859817836364805,
 0.2657144013588505,0.21899272675009707,0.9933627671858635,
 6.181612306793705e-4,0.017978046424815153,0.49509624144919573,
 0.5558948076899506,0.7822178659852308,0.08645508582878103,
 0.25997037921315813,0.5572024138165622,0.23178205737941715,
 0.9859322995280547,0.5153222626932858,0.14050433076772118,0.7037399656783798,
 0.42360228115311727,0.36793878259592316,0.06371840736468304,
 0.9810848399151111,0.5381800538279334,0.9088558777522927,0.1671227883391515,
 0.4499901591556037,0.37455086164772733,0.14418471223344942,
 0.7878316030419795,0.9626438798962672,0.40182867918137766,0.8241283412374728,
 0.8432693993737161,0.127263418559025,0.2829400595765176,0.9069939251516257,
 0.3419866467693534,0.29821952490018944,0.9133428752362598,0.9251966954594092,
 0.19794821543949426,0.33236117084702943,0.5403039044493038,
 0.7445975188221396,0.5559452398397693,0.9202409359461362,0.0886422613545933,
 0.3697639234974448,0.039743779626661424,0.5491136741078005,
 0.22695954844479982,0.8145561320242174,0.2712588622312504,0.6888390950587928,
 0.5228467642150548,0.10194437580220672,0.744006833779114,0.9438879428889421,
 0.7283794484122883,0.8927915831688369,0.4312877345667949,0.7084207465128635,
 0.4185518721045076,0.9637571004411828,0.3317489976331691,0.7639962523591821,
 0.9601850708957644,0.5300685504952016,0.866731191430875,0.7299536707465603,
 0.5505484489672487,0.2025788050543308,0.029677339871049147,
 0.17988307500520473,0.787133469748156,0.8614166590166952,0.25086216313033116,
 0.49512531688809314,0.2809368736344231,0.9966562935501696,
 0.23960210171739826,0.5200994465342303,0.9715821189983134,0.9185309320586577,
 0.07731239247275201,0.24741248476193534,0.3133519061092729,
 0.5186387123002875,0.37782171160250355,0.8753185332453833,0.8667393859088806,
 0.9426040506370104,0.6786712321360148,0.9974865362197213,0.6610060922461641,
 0.6546703801098315,0.67135085128373,0.9079321208323452,0.5226297278419341,
 0.8249059242207467,0.012387441749056927,0.9923629000388304,
 0.5652604174824967,0.06298201086407051,0.6786882163949217,0.6275956631165104,
 0.7686338431317891,0.8329377306145173,0.2918794496144982,0.4093329920804576,
 0.6638393755882979,0.895475486634703,0.3928710624506645,0.9671973020980567,
 0.9851267584018355,0.9924315532129333,0.6464514863783483,0.5590918670683526,
 0.8580948208370468,0.28753393085086654,0.7151880072391298,0.9150571321255138,
 0.5457324052245183,0.18103420560597971,0.31904875268033905,
 0.4432356098172451,0.1644186132889035,0.706721036749071,0.498435944444375,
 0.7232043062448156,0.42725411230723387,0.7289269460083263,0.6952410668888903,
 0.39116260215706133,0.608076638110459,0.34686080362761784,0.3896365700827158,
 0.9927377414673175,0.030353531087248142,0.38004923021878756,
 0.5570356955263203,0.7174186851948309,0.35862239086933156,0.5148042340132288,
 0.4740643917511056,0.25743037121537504,0.40088496289728237,
 0.9787197119961204,0.6343165827124498,0.884436643287623,0.15288889884627166,
 0.8275362014502947,0.31439040253787987,0.42446338168255326,
 0.9899745679740287,0.11368270821294968,0.7927192693301448,0.8698216585850698,
 0.36573053907816333,0.6957099433900723,0.2780157870106583,
 0.011000538546478533,0.033038388830950116,0.7219658319383426,
 0.5196131744741536,0.5921527906419389,0.04529279995379709,0.4450037324651146,
 0.42805770032158486,0.4577296168439161,0.6897633387624197,0.3945733544255846,
 0.11319637588026654,0.21373156433905482,0.20071781012300927,
 0.5318947654136756,0.9132012978345591,0.5217561267876778,0.8441354087736547,
 0.9912223485218408,0.5613580403244384,0.9238472367357646,0.20391897016537652,
 0.8305099492196899,0.02454518652528359,0.625417815345325,0.2118817389938179,
 0.45830265424839467,0.24813126515584316,0.6386240777726051,
 0.5659516663509248,0.6493977184876647,0.9316376000195464,0.5873352600569477,
 0.8622813236317117,0.5505524246237691,0.478519535275597,0.4155455446741928,
 0.43343803938482117,0.38758393992438733,0.8167472080636818,
 0.1329236136594607,0.9339907715606666,0.8284978681848723,0.7864680268328413,
 0.618069468044571,0.6832630448655044,0.07929683810888011,0.7310766640612362,
 0.9820776576362236,0.422950493703942,0.9264592968826473,0.9299108812600985,
 0.4088179140491116,0.6613683772254433,0.7590904491230833,0.9050268745587271,
 0.11048294661508729,0.6654815370439702,0.1594264760310411,0.9891301817856331,
 0.11514461164938661,0.18255306811661298,0.7214683389622717,
 0.6955772734499223,0.09739109421523306,0.4711809980476299,0.6437867623010846,
 0.8843141595382612,0.8408461142586451,0.9989863969099022,0.5888324345426055,
 0.012596440782021268,0.23555272434038255,0.6484069828456611,
 0.6868693392142159,0.7040225585346821,0.4830952160565961,0.11309982109202277,
 0.9650385687364085,0.13645066713089837,0.5287185330343127,0.4379134379245817,
 0.8337588311896145,0.9055180230305061,0.9805700582430762,0.6487839159393447,
 0.509812923203326,0.3230371597715449,0.4012650135751489,0.5623425558340841,
 0.13415222100781765,0.9397088938868721,0.8531434914945069,0.7872669860197732,
 0.44061042540198436,0.9198102415984613,0.41977031465849624,0.099962847127952,
 0.1485085548113546,0.6034131018541876,0.334285858889666,0.4329014733068034,
 0.5270336838579551,0.32243973682196825,0.3976765115461153,0.2929102947506359,
 0.11048137131429514,0.0988568402620924,0.8599212799314072,0.8428886538258047,
 0.8156024673279259,0.6574951068674357,0.7884136287830971,0.06278332188532632,
 0.5219631465568875,0.11341509475054179,0.8351643659985724,
 0.38346454930645324,0.32268328116544354,0.47118754743119573,
 0.5405438648916701,0.927384091893899,0.45985939015121813,0.42295867560877065,
 0.13534673741655256,0.8239800640380827,0.09722099155161312,
 0.49865385682876207,0.9704241779947205,0.8801865059822902,0.8067166681485449,
 0.6880058957610038,0.11518516392091716,0.6207169511503515,0.5759113337100706,
 0.30338911313144856,0.641479478352673,0.2582586570178136,0.45735105846191937,
 0.6979722772821696,0.2977575587295358,0.31908920291642795,0.860865352805736,
 0.044957386065963445,0.6082721069019845,0.6636630823874022,
 0.07682042121548893,0.16915433033807759,0.023657261359318782,
 0.6541864227425018,0.567266489046927,0.1567245578196398,0.7063971162651361,
 0.9502874473809159,0.8430072436449594,0.8401033232278312,0.07074555842314145,
 0.9110774233684024,0.6467683921948132,0.30049108017086534,0.920912703667466,
 0.9291729636947921,0.9931664277098118,0.33467434559055187,
 0.033811072186080704,0.32536257116533474,0.6079098489456312,
 0.2547625182055814,0.629375688296655,0.7404780079127824,0.5532465715226903,
 0.12189467953614053,0.08365856972372854,0.22733294906455137,
 0.8852342300885423,0.9211268973215839,0.514148581549752,0.3900125726846937,
 0.5690730320401559,0.19650637140538718,0.068526044928507,
 0.008465920172645691,0.12232973869918806,0.5007862309080612,
 0.7863766120218354,0.4807757289323663,0.6838855318441246,0.4061904867505437,
 0.5388316189375678,0.6065712015580418,0.974807110984246,0.7802676417578936,
 0.201628851214804,0.018429181295537944,0.9299250383113256,
 0.09422967932608639,0.741855625472209,0.4644203495254273,0.09008350265120646,
 0.3386072045395583,0.9089704485349381,0.7052335512831657,0.45323087100749215,
 0.4425228415832937,0.9961760673494915,0.6957615370699718,0.34445359808452436,
 0.17836298221329183,0.3665041504472819,0.279431787188834,0.9434513284001081,
 0.8956617335584317,0.9356531244549348,0.2003660558558742,0.9238207774794917,
 0.19976160087519657,0.19114078479171748,0.9299691309547443,
 0.40650473380861174,0.3853051322755251,0.14566181902473507,
 0.2516949336559723,0.5856139432547683,0.010871713012192474,
 0.48035379547172774,0.5119746572348096,0.5800328623780467,0.702529163835509,
 0.23643048013156565,0.14476565033810962,0.28077033293779907,
 0.023248524321319186,0.42094720515600736,0.029907305678319895,
 0.12404113089824409,0.08152413550656767,0.9721182336903292,
 0.8029851897654057,0.06872445377447389,0.19613193086696623,
 0.4065074259717112,0.4589696904752718,0.5883319163974498,0.8700399390802382,
 0.09994832074321747,0.48809154547408906,0.7481811112510457,
 0.3750294532177443,0.949688644323344,0.1419048746830759,0.6729593285328692,
 0.018045323038358707,0.42633178032448105,0.09272150397809908,
 0.5568027988181463,0.8068537981484397,0.8174532432174868,0.5530026427271255,
 0.4494961838639451,0.4515125703585452,0.5346401072992724,0.795019505769664,
 0.6800793512178684,0.5567223683442176,0.517421478564731,0.10559765088399653,
 0.999221048673948,0.03352801102529801,0.2558178404813374,0.9321459037183938,
 0.5986747228704584,0.7180374790624234,0.20182705508221999,
 0.03410842512635637,0.2611573962183862,0.4339593014012225,0.8718152908712309,
 0.6399535289333401,0.14598042563121072,0.7382587518031684,0.7249410021535414,
 0.12952975277957712,0.9661470286663049,0.5155118089895024,0.6852124278939721,
 0.02810595344924749,0.49880003762983494,0.12511718061218025,
 0.43253655878695585,0.39838253543410607,0.9416704500179955,
 0.49446971135181805,0.5320663195368182,0.27192519927603354,
 0.7177921235040021,0.21177431318813023,0.6241423632404363,0.7080420175591569,
 0.42412659331878244,0.13470221880148014,0.24464238254963344,
 0.25442239184403037,0.9336044947957108,0.5059031381230618,0.8590322256574912,
 0.8450989437284068,0.7485765474135782,0.6131191426339284,0.6598340744125941,
 0.5160642606112771,0.4548598784090452,0.11383273844877018,0.1384861027046479,
 0.7533043618667109,0.7908057973953684,0.6156349824714924,0.8422953864778195,
 0.10853012817244034,0.42597959972977595,0.5246945372246807,
 0.30008920442143106,0.9467036238167665,0.8622883214677045,
 0.09687478888459433,0.15194664508023237,0.9515962613748729,
 0.040663405597831215,0.014104360695709195,0.24667318486876377,
 0.3339117385338386,0.26763815940240765,0.4151917126667949,0.7473004445025049,
 0.7459746959035373,0.4301313338814129,0.6404214729655133,0.7687005123311629,
 0.44437296443652796,0.15601335376877545,0.47270429328727637,
 0.2515584227326447,0.042985760538236994,0.16573588670058936,
 0.393085529076339,0.9244491520150062,0.7543198165220794,0.47606088264563995,
 0.20885470053091204,0.012121918039775026,0.3151611096367942,
 0.38027890986723234,0.39663788170173375,0.8966482793615965,
 0.30429754457303204,0.41542125304132616,0.5662872917621229,
 0.3513031304018903,0.6111438197350908,0.661477426927456,0.5117461879992267,
 0.5634552588371992,0.8694605475079026,0.8058000124942856,0.16870382492088398,
 0.7870695835508702,0.8945878793593045,0.356393506249539,0.7235717469426094,
 0.5586937160465008,0.15237572929781007,0.21976102034669298,
 0.5143324879728881,0.9794275195560604,0.6968561478671218,0.23222541020670306,
 0.0075044489692080996,0.20052437022094527,0.4759754786824618,
 0.2598012036959638,0.3265631497451268,0.537920115895837,0.7833024028324582,
 0.44084162168333973,0.1864024730378796,0.8566812214221106,
 0.10366007402497002,0.32439338692597364,0.6075589937395742,
 0.9329534969758524,0.9240340029965095,0.29914415940243877,
 0.14278558177516398,0.9448425439904472,0.48891992817567975,
 0.6348673100427026,0.15015737543887375,0.3658329957538684,
 0.025838846501301194,0.006646729985265387,0.48809579858749275,
 0.4348452402883277,0.2057122473123293,0.17688011385244473,
 0.04360873454829217,0.6404256470819156,0.537207143106723,0.12428552863124254,
 0.3743819991751225,0.6235074939331824,0.2885505654382472,0.3506163291351534,
 0.9176061230169049,0.1553129533158063,0.7734391455880548,0.07770878110639745,
 0.15260936463367547,0.1366816042300194,0.7098935200556142,0.757678207540257,
 0.1655625920884728,0.42489180150209827,0.25057819707134366,
 0.4622332041341517,0.5060366266063331,0.30398812680881426,0.9967249805304573,
 0.46576964359266637,0.9768949270585037,0.18371030170585267,
 0.8057323978294191,0.5022229177947525,0.8972900326597847,0.9375079130249111,
 0.9522698163711429,0.6598692540730663,0.22392359814936114,0.4202210706253219,
 0.40849039436997425,0.8958251132945123,0.3531003762535043,0.714774644950592,
 0.5416853467174714,0.8034910998487919,0.22544029191729198,
 0.08336824292055545,0.3423516588297514,0.15133033991047662,
 0.8128002159780765,0.737099625831323,0.6770617968902782,0.693374343361966,
 0.7365025580804894,0.050639564012448735,0.7655928442862526,
 0.8985553328541296,0.4105628949570166,0.24754160984130136,0.8518883209809005,
 0.011776462522910092,0.24548620378767105,0.7743448523866645,
 0.3646805818504353,0.6980888655912063,0.6298206408526785,0.6174308437995549,
 0.3400369863942515,0.7702125784797862,0.24425792590204898,
 0.22814649812568233,0.9958724034039417,0.5461017426692132,
 0.38126703890835967,0.47439253152645877,0.7052606789647917,
 0.04650736820812862,0.46572749171087613,0.9405964611950018,
 0.15394844320239687,0.12923624053284055,0.8944894931682333,
 0.9976503133024492,0.4392113595246424,0.1919235377910664,0.9538277326013208,
 0.8901456821613484,0.8516288476486978,0.16650323834622138,0.1328349070489756,
 0.6475555194478217,0.5684536533448464,0.014679994426409992,0.516955288087813,
 0.5015196468640828,0.17711425105887668,0.11399699918557449,
 0.0994451761228673,0.309353260219404,0.5601500570414352,0.979805597434747,
 0.8849769365688572,0.2371887756966662,0.1293864099983929,0.3150784528290468,
 0.8987677095917358,0.9171364863513953,0.2961765965537607,0.8922979597293201,
 0.8061039825506406,0.026545045394010036,0.7570337678012916,0.78042695509746,
 0.1268432628460867,0.6683820587253755,0.446407334412789,0.9130820617828149,
 0.5045725416480984,0.2207247325397328,0.2513238380701186,0.22602791921516052,
 0.42213206653740354,0.6496813420319649,0.38775288677044883,
 0.23066896550321114,0.4699019633180561,0.5851542595287158,0.7917219413957964,
 0.8787871175189688,0.9679425082566872,0.35147827972048595,0.530118304040647,
 0.8763990692902256,0.5182499611261443,0.9382332439295682,0.15027999044313334,
 0.8722915231115415,0.577409270180661,0.9660001126100117,0.16284993282055238,
 0.059565188357956766,0.9324206989359286,0.6727095376994863,
 0.8990979098072415,0.5945119992850292,0.7554211356719709,0.13201846177997045,
 0.4522017531404884,0.8157903902742224,0.7831259469849055,0.7646914449826149,
 0.39017684215083703,0.8478429432287806,0.609874149907878,0.9756895695434231,
 4.4352041023576483e-4,0.7088590268126082,0.7129080364924619,
 0.6195653067918963,0.16523044634338824,0.47764798992515745,
 0.5485438282431823,0.4845654704932605,0.29345702096096826,0.2810581625267856,
 0.5617301722601905,0.6871124467420521,0.8693737632828087,0.2311015901729282,
 0.7096331269175855,0.4934567129031523,0.43429868585908393,0.9507486697399182,
 0.7607936395520531,0.10774007586326939,0.576404071453796,0.4304590225316266,
 0.6057923204487389,0.10004266558393249,0.6327253777094128,0.764293158525127,
 0.35398544281104516,0.1908004771698517,0.16574016718283413,
 0.2939792347152097,0.738411827527181,0.862422641551166,0.9473749454285838,
 0.117327840421521,0.9325112643780769,0.1751546227381322,0.7984401971029604,
 0.2382556751502296,0.18909941282364617,0.7612822802868302,0.9739565680383374,
 0.8164344986412587,0.08444863587237206,0.13188786566207034,
 0.9887601148200109,0.03669858969786732,0.4977506839604636,0.7971596327961407,
 0.1448016255704473,0.6267043804360954,0.10544925796663607,0.7691386412658784,
 0.5433162921867615,0.7227721644749064,0.35103418644223805,0.9571624637405123,
 0.6044544883256211,0.4448846518157781,0.0946290380734327,0.297380441639176,
 0.9472553284111277,0.5193585964485624,0.621803662392581,0.6049589035114546,
 0.1471967980933564,0.1894403478764255,0.6133544553142215,0.2756350505277352,
 0.7840751140118931,0.738523996082755,0.3448633495555202,0.5194242996280615,
 0.48359361593689143,0.16457539874070004,0.5431433699399247,
 0.3351369743470941,0.756342724308972,0.4798257000083137,0.5504431119147775,
 0.9062696223722547,0.8335645399972801,0.33911857212141805,0.3340858546386035,
 0.9166440196647208,0.13050251346045172,0.4220083687195769,0.8882222416632425,
 0.6854721805033172,0.981587892056503,0.9067642883031267,0.8358929873820516,
 0.32552480831681785,0.8870317591115348,0.2931137530483898,0.642822112459351,
 0.6397766503769495,0.9161509279073867,0.5235828857835975,0.05009733302664732,
 0.020735356951774753,0.43009206211839435,0.3301575782825926,
 0.4976602720990734,0.07620405080372017,0.12337059135286799,
 0.8760894998106098,0.0051678867565140285,0.20544067336270255,
 0.4638853688496847,0.27408384721754886,0.5144919762396747,0.360522879620226,
 0.277388631296145,0.4656278090392182,0.4410015590010927,0.0773110183334762,
 0.867353892113881,0.9390982972742775,0.22616280084403673,0.06829837523299576,
 0.7361333013606354,0.2862218166794337,0.8236391192267754,0.6019608277667909,
 0.04044526242640567,0.21532589517054357,0.9401928087050252,0.683581567414689,
 0.5355578958619504,0.9515440347909995,0.5426798447191408,0.3804134767225831,
 0.9390453728447475,0.9654454367476973,0.2906673084022765,0.8833277442036307,
 0.5718749622355581,0.0976057410468909,0.6131285793756004,0.6127537819900825,
 0.29976768849179314,0.5793782352869737,0.32214320037644417,
 0.5043644773768674,0.4755713645108052,0.749216596236591,0.07600889764830987,
 0.45767563520710486,0.5314271074286351,0.28410410430644895,
 0.8772134816052455,0.6734432224012658,0.34969462122189887,
 0.061770071245980884,0.630629990466276,0.10875688260201377,
 0.23716966750788293,0.5819447593364292,0.0988344516547276,0.2672553610125472,
 0.45221156392122763,0.5927572476681041,0.37426725931500326,
 0.7739595018959097,0.34213588360025327,0.9606268726062052,
 0.16477039089932777,0.9229363841506807,0.2569952040940576,0.5809035770308953,
 0.9928309127863785,0.020207064504417005,0.07542004024319371,
 0.7668546274669594,0.5453146446775274,0.6095574891163158,0.13439459403669207,
 0.04046980609231687,0.8976181094724132,0.45314953601804353,
 0.1978408435834792,0.5377453283233331,0.8414718963929807,0.3399705249949395,
 0.27430363725036255,0.0348733321446012,0.34216350691933717,
 0.6333125927876365,0.6319693772613704,0.2878823813498227,0.47330739295040913,
 0.9892636468423077,0.7787910447960721,0.31763284158264204,0.9123073963525632,
 0.42171141905192533,0.48816957309543596,0.6375225567701646,
 0.24466882440027649,0.814382216925966,0.042334526404852135,
 0.07214786067293677,0.9496346147564381,0.6552146681562976,0.9231619102303396,
 0.8253637748309485,0.7898813407349629,0.020953775212771086,
 0.22144237659301969,0.14041643899470535,0.6573655888396713,
 0.33781406285617954,0.0011057978314366057,0.3457928141632026,
 0.03325111491776633,0.8303126555377836,0.7608730884063951,0.7654372080613022,
 0.7716314204057654,0.5225289280976358,0.7612128513254931,0.9138494690549211,
 0.3176784884380258,0.7082502549027829,0.054443052741546594,
 0.7067108067800347,0.1740312177021397,0.021249491241384533,
 0.44989690569152396,0.96496414137185,0.12037877577836475,0.2391104844665204,
 0.4580168420009758,0.03862960632453427,0.27995258307470106,
 0.15232356853065232,0.8188764369002919,0.4787052210661492,0.9627500955051566,
 0.7583519548074512,0.10486481660604452,0.006428569298124409,
 0.1107228582135531,0.5186084821260274,0.7695779956759997,0.3732195139762665,
 0.47147397720952466,0.09241712506892064,0.8469490115247121,
 0.7046365278080088,0.21571455739878742,0.7947625124051725,0.5964495657454005,
 0.3930033832395994,0.09590455857574853,0.8489755465334903,0.6063853958122997,
 0.5133600468660628,0.8873568535042744,0.851421914458482,0.519913204612877,
 0.0022266739737328844,0.8308613255748925,0.10762130692940841,
 0.25763427275975237,0.8584879580830943,0.5495270695892984,0.739733957657728,
 0.7646632331911773,0.8330789504050743,0.43444293047253757,0.8905987048310698,
 0.9432177899362945,0.5211834932632493,0.3200344123496197,0.08468771607539749,
 0.05737593393211049,0.9330099195563027,0.058114805447667894,
 0.5928443126452684,0.8870200890001166,0.4077365225755857,0.24806119225016243,
 0.5590091283542767,0.8378567746431113,0.39164016634521326,0.3996500275200483,
 0.5886007134552184,0.5634391325477878,0.2789935954956323,0.9024680194392984,
 0.09939139722019774,0.2408130750272648,0.5731456079725692,0.6368396741591252,
 0.37893379414013306,0.16612714921762428,0.4320769512056247,
 0.02533976878085198,0.6016444653897978,0.27601098240918265,
 0.40948768886404774,0.42924744173051754,0.3268970274408127,
 0.8981674337429233,0.9599068439424845,0.4788371673072609,0.4894954779960019,
 0.16871711999424766,0.9164039286497387,0.7834479200513433,0.6324854140625582,
 0.6994636753005782,0.49381759122562985,0.9430384302687213,0.3876842008643012,
 0.023747008077238396,0.7490545002901746,0.049530116536408375,
 0.10286625128254434,0.46066734446789603,0.06635647148237189,
 0.575510421434066,0.5817689125075065,0.923876216055451,0.8292382249000622,
 0.7913425591258791,0.253073221788193,0.8769064779391992,0.5807647766532918,
 0.6335627303753666,0.3506469064001685,0.3311073970844217,0.02703275084503698,
 0.9135179954860234,0.09398300080860034,0.2993676504516598,
 0.054627960028240774,0.9146602169377428,0.8292795928749208,
 0.45539594464393196,0.6705370057252339,0.8209400556559429,
 0.35927925057468224,0.7604376026000068,0.6602685555994617,0.8565954427350461,
 0.23225250688313515,0.5726483870561547,0.7747260529006571,
 0.48074660103882894,0.4538106386668629,0.4196116790474771,0.581008462012352,
 0.2722789163542696,0.32993024025485884,0.8126367787811296,0.7394269186673005,
 0.263562744137672,0.16924470365900046,0.5902304732776794,0.5422884177343041,
 0.6523845514280101,0.2719794926287935,0.0404466494485779,0.29010257305061593,
 0.9571542147029223,0.25563793080046004,0.4747520538219563,0.7889874253089413,
 0.41453202670458644,0.5736740901130626,0.31068202448651805,
 0.32627330902222285,0.49641672306575035,0.1480985086668024,
 0.2639585124072894,0.018982259965748316,0.2267906369589695,
 0.5975441721210101,0.8134395749022315,0.895073841334761,0.8899776660704984,
 0.7387006643025438,0.25573710155768326,0.14988477456888005,
 0.5651830096229326,0.5100990749306641,0.3002272532137822,0.1381931543861904,
 0.3659467880189724,0.8985965075981384,0.10768264454773768,
 0.19858690516134936,0.4629921858627911,0.12172414557893452,
 0.5880648550711184,0.2926404029320363,0.4149846880612357,0.09832807108395047,
 0.37826346624777485,0.7695754414969297,0.3020717657960912,0.2588610856136039,
 0.8127768570747027,0.8123311336886668,0.15776528886483543,0.6989811102068888,
 0.6641465632441416,0.16360908285519027,0.6116956039604473,0.5857198796674358,
 0.7345802746292319,0.7272750093446546,0.30906601182160065,0.8891014249379021,
 0.25262114569502936,0.5032798216624339,0.41725237229352763,
 0.9829428167541847,0.19841536866062626,0.5320388343418415,
 0.11488683207015349,0.2693279005721566,0.7061668243459166,0.8493045725241585,
 0.050650050050692785,0.9828391056994836,0.48780116562483256,
 0.4946320265160602,0.6250959962362739,0.7650563849712514,0.9235430559114269,
 0.6015673087279825,0.6449584193884146,0.6216519465458417,0.5654620515132368,
 0.9816328142965395,0.852432881049658,0.9588009801214742,0.6270233233428126,
 0.19553654634835005,0.0965426022449164,0.9133142943581394,
 0.26041562769694404,0.4177729857356318,0.737767052230585,0.6857768850412137,
 0.9391805218947823,0.8159747943278978,0.49223650629199356,0.8086759514212709,
 0.24108304100666766,0.9237041170779505,0.3008137757646613,0.7119576689052873,
 0.740613232412508,0.7903350798375812,0.4232514940009685,0.520761426746118,
 0.2331946390863181,0.13480986179911048,0.5628293408562991,0.4307896371394938,
 0.14702276461631714,0.7337051498354878,0.2545939941774086,0.8999014414234996,
 0.740362735639138,0.08301105177756596,0.11706547054841976,0.8864760395754884,
 0.7860697576747573,0.5417715757203866,0.6418385512973159,0.6698864098000025,
 0.9293529997848788,0.9858324243078143,0.5972418498204376,0.9121628548437586,
 0.7297330010750952,0.8165176230698704,0.7793362462455736,0.2881941433740983,
 0.31987921532626296,0.2531414371835524,0.10135577386707184,
 0.9538954774665274,0.3853238948556176,0.516139144272664,0.041204877630835135,
 0.9988721640218209,0.4736702766709996,0.08175101893412551,0.1212109701083306,
 0.045725502237217164,0.9556286138510969,0.6814053694856397,
 0.6839394841874762,0.6056050276562892,0.11466825260415436,0.5540665400812872,
 0.19993823303857738,0.8186323904083861,0.5097668334285328,
 0.10519623422383328,0.3638814566661882,0.030498055278772584,
 0.08344312998714909,0.8504700031008804,0.5007806049061155,0.6933433027497934,
 0.8722697766050054,0.43203924518850534,0.6156581622526367,
 0.06108193113065341,0.4221399455257797,0.052230018014334245,
 0.28112169511356555,0.9104485531791258,0.48235838866260405,
 0.12738726251036292,0.12466634994394887,0.9695946210293969,
 0.47075695441401555,0.28128557533702336,0.5396982300896768,
 0.9204582027241345,0.3814380860594403,0.023228117582384522,
 0.11933409206858447,0.08788788851319151,0.792927364731358,0.9217072229343661,
 0.376699572272153,0.9219854151941362,0.8344877030562982,0.0351215858377012,
 0.859747110119692,0.18807777081119526,0.8637606343399332,0.6505777211538144,
 0.3015405941781487,0.6188477907250723,0.39534416839483444,0.2225446226593807,
 0.5982782536895246,0.1629436424274734,0.770562396830374,0.01216513436209088,
 0.6808262218980945,0.5567455165290095,0.41925958524277884,
 0.40435061341042144,0.6078331582741332,0.3564948277209984,
 0.21631035639759166,0.33460441475901004,0.5945747217676043,
 0.07891853844193075,0.5115697237380245,0.4654269153475312,0.1743507299526308,
 0.01815802953490453,0.02032363497589451,0.07045998772215678,
 0.5942996577957845,0.4307765814083343,0.6469211727416557,0.12497568664510927,
 0.6314418605017148,0.22453136978955746,0.8677399331954297,0.5121011893955391,
 0.864964034540373,0.6208676657535257,0.6546181476262741,0.8654575836621088,
 0.6238057475918833,0.7115815096341956,0.10607081894536718,0.5504589456638657,
 0.36223421512379617,0.7622152848187844,0.00675447380565819,
 0.9233934723297289,0.9315158682955755,0.3883137602811142,0.2596338467135363,
 0.02802474950143552,0.18477895700406055,0.673447200737728,0.814012192720885,
 0.2789995555666466,0.8239765877390445,0.2266586224846815,0.27130376424436364,
 0.5354297369750296,0.045026840436146354,0.1479917935786762,
 0.5354832893762974,0.5010092537160943,0.9644697385462042,0.7244694420104123,
 0.472504059275877,0.8238077277729264,0.44727843082729923,0.5943085165512009,
 0.026950024256741312,0.6090399609000186,0.48509417706280233,
 0.7865245804874621,0.6784427412957221,0.63488900702669,0.7820383451938644,
 0.5978087531069777,0.9634669466921219,0.9937250898853376,0.5546528646293326,
 0.79840137591925,0.35667137262277915,0.15146201813312443,0.17156523978331495,
 0.14451123214007255,0.8017725374687259,0.6653014059448892,0.7636073322024156,
 0.07667253621261627,0.9001138733394913,0.40066668426764174,
 0.5712386576523977,0.8419905964925604,0.6570484433307113,0.8681356467618578,
 0.06305538807080713,0.3199742552381718,0.3858025149185885,
 0.30189681720036843,0.6323932211732288,0.43181292560789897,
 0.6891256513486514,0.6947162475240721,0.1983753948229482,0.25880892230963837,
 0.10183455348919968,0.8489458972930174,0.9913686223157787,0.1754367726842475,
 0.2505343863758872,0.9060885670002886,0.8320297166118122,0.43815952601964436,
 0.9955671644297363,0.7367010411386969,0.9245032918781363,0.6334836601203806,
 0.8106317160877684,0.7733728471428725,0.4336391439587308,0.5805537982056346,
 0.11159409940349474,0.0053189952488988945,0.4605341305479127,
 0.8291763906494498,0.13107125674214593,0.2934297959259592,0.6814042722444256,
 0.22659687153694785,0.6954490498994985,0.36431820282136784,
 0.8405930838027846,0.834721038426996,0.5452810642904932,0.7712960940229056,
 0.5936044670490762,0.9905196596452022,0.25046586272734683,0.5025940242586667,
 0.28131272687042375,0.6870409776850921,0.20130600741552795,
 0.4601371522061849,0.2983218378082565,0.7134056984650127,0.8662578716794033,
 0.1466748319864779,0.6246660672230441,0.44385113049711167,0.6144239416956713,
 0.9733880575008527,0.746262781201556,0.8445206787407004,0.07919165746176726,
 0.9887360582055281,0.9464713049919885,0.5902617548019404,0.4040380453766703,
 0.7421534916345839,0.04579108503123108,0.7640913324516411,0.8834604875980409,
 0.1718065907497056,0.6287921715489309,0.6690298754070553,0.8469019262747275,
 0.7791345126989705,0.48518255668812404,0.2447689510096731,0.3464499445484692,
 0.659385089227564,0.2047600135273291,0.9464473286243402,0.02591960223894041,
 0.7279744621899351,0.4148786442796166,0.11702960297827703,0.5089163836050782,
 0.9166243320505046,0.67596109579011,0.9330541242782778,0.9864246838110476,
 0.921573844124369,0.583599356826751,0.14438038504542794,0.5004370553293312,
 0.3680351458661665,0.5987231231871651,0.6401603637101979,0.4917195626734647,
 0.763597279121471,0.021209722552795962,0.8236494399037919,
 0.012132808514171312,0.10194752253526607,0.7214020808741699,
 0.31951503749295507,0.28856466805638803,0.29676849641401337,
 0.6748216554504234,0.23420478494934793,0.4942225059934694,
 0.42742068717552484,0.7912034292798064,0.316296863386082,0.7993216119548366,
 0.8259652745143644,0.5764800230858005,0.7554192348312626,0.6718509728896713,
 0.8288182259711724,0.20559377360754916,0.3649443207794173,0.2558786363678788,
 0.07797925305090359,0.6397529454060691,0.866340116302905,0.2396175212507603,
 0.4416430932061348,0.5206775724476732,0.2660916495372314,0.8742212506423138,
 0.2528090980542149,0.8593100544851844,0.6782047448289893,0.5625407626301079,
 0.703392712467505,0.33573461314212927,0.7996285844030689,0.44712824010200025,
 0.5692048136221721,0.7629981069094359,0.2102318420142597,0.7158436372231486,
 0.09713958250458199,0.5763615778709097,0.2838239198032424,
 0.21631293432813348,0.29530525127529994,0.5230873161185534,
 0.8758347565523331,0.4220729990734602,0.8882535496921498,0.6407043268971935,
 0.39367765822734746,0.31291342501959507,0.7020875161731511,
 0.7768358392589273,0.22112530647862394,0.9842026578569142,
 0.21540717677185883,0.4641551181493444,0.5809626865897368,
 0.37314420716218466,0.2784840148384483,0.07139812598766959,0.935881604119093,
 0.9303129641459499,0.698496128193813,0.6430061900138612,0.694532334100931,
 0.8452353808213363,0.6971391054380986,0.2618455920402021,0.681684904888845,
 0.2620223347475702,0.6985707273449788,0.9833769085333149,0.6574072009433645,
 0.04438647424911246,0.6488602608199088,0.11644656197031489,
 0.3788536902447339,0.7724000591016225,0.7911577029511642,0.13072178799932943,
 0.47122357935356707,0.18237270140754358,0.9625313059350761,
 0.9510153347336601,0.9155616370719896,0.04169036320220265,0.583337333077472,
 0.977241348679435,0.9576795299328229,0.31737937154967666,0.2061770991814802,
 0.7789479717831045,0.21027566992112023,0.2579009277452414,0.6827547923461869,
 0.08216787676782822,0.30107257267696996,0.6278099289542736,
 0.2519810434591978,0.479114979103631,0.9518529107118849,0.14233468570802577,
 0.1760575893683214,0.2187485104992577,0.6725193715756139,0.6743702276560557,
 0.08551426843379417,0.48685791418428526,0.9364854135893884,
 0.3566810167926395,0.6426648220494128,0.1566919263136599,0.35156108572103484,
 0.9666765155614672,0.28766696774926404,0.23833747133447936,
 0.6784167647661565,0.009904618467970194,0.40173923679745294,
 0.7856999419101452,0.6115475806895584,0.6583094462780567,0.16897554819379934,
 0.09095805756139586,0.11541092320887403,0.41373064623562483,
 0.7152975645490945,0.4699023241262418,0.37337306285805205,0.5181835372663947,
 0.9223813413605338,0.7741217560758791,0.5042821564161035,0.4831920155128271,
 0.9236467235011655,0.979584082966394,0.5838532452568693,0.4632577456661596,
 0.8108555682974008,0.24578084520572663,0.9228690329146969,0.9333580917828511,
 0.683593681319777,0.7578125267585502,0.7531502642122153,0.9455136885461654,
 0.3432648791995835,0.024096932792404768,0.2167143862461549,
 0.29205049945699146,0.6426429920507601,0.2695716909337855,0.8812961247004516,
 0.8406683869503171,0.232122056328524,0.5177170298103937,0.9007779689634899,
 0.7823889229501381,0.031213724843589974,0.19659514320932603,
 0.9092181052348693,0.4947354863849922,0.2424932889396164,0.08563731891369364,
 0.13349888684260824,0.26926109501843365,0.5646908634082265,
 0.16302226790860608,0.5288773132138306,0.8353959314617263,0.6201465833240544,
 0.6426534701822568,0.7182770238777082,0.4464492979450334,0.9203748528181745,
 0.9864793642844367,0.5874315875389048,0.28048017878190645,0.8128630463821982,
 0.3403597590646532,0.38347120000198576,0.2932933597473165,0.5685443791692395,
 0.014343756871424529,0.16251881004496593,0.49997393743700647,
 0.3670063974041078,0.4009792290019819,0.17888393224026133,0.7581949387660711,
 0.8494974273626311,0.3307941509398029,0.1235141713751563,0.6864500484652196,
 0.2700969131957026,0.6687355606431585,0.28653901612292054,
 0.15464066303632773,0.685850967548949,0.9021929293444272,0.44594561315793424,
 0.8204608601890873,0.505206489827413,0.4549436818351503,0.6844086513575148,
 0.7379555504483435,0.27087800430869635,0.10847547583591544,
 0.4042698223341503,0.8003700340051073,0.13136586336048406,
 0.09432178627142196,0.7592991982912555,0.2549963895929451,0.429507938871902,
 0.4546716594229877,0.8421762703789497,0.003980572180891762,
 0.7300923716002362,0.3060417567067949,0.7321410591482697,0.883968908882978,
 0.4727148726215119,0.355616230980051,0.14972047370649788,0.235758463583021,
 0.15235688346993137,0.33653419142387453,0.5172506566248918,
 0.7248598509149955,0.8564673336719854,0.8018820567491378,0.05122328869443893,
 0.113512501886605,0.9221736450200799,0.15893960600697188,0.3839544170897473,
 0.1439664471608888,0.370481643828812,0.8409185073364404,0.5119550198318347,
 0.10847461226147304,0.45465931077276545,0.3575528442605602,
 0.5493810094053673,0.43094910921930285,0.4822402875538626,0.5784384589556125,
 0.6153409541485027,0.17956928291849494,0.5711816516611093,
 0.34460004186909454,0.24584204664201137,0.2600289442242928,
 0.8740595847124232,0.5770209131039014,0.13832445709772978,0.4127098740471493,
 0.3244755127604061,0.17125118349290802,0.4963921554913103,0.9579955695521329,
 0.6495546251327163,0.7991181169616777,0.8272727051432339,
 0.049491714355107796,0.7824908867916562,0.9871834757529268,
 0.5841226375388968,0.4130870019448477,0.04956956773019827,0.775465671446512,
 0.693046887878964,0.056886047550606,0.5791880325810831,0.9809897982013388,
 0.7790093039525243,0.6483540425008856,0.2694836996787635,0.16638210308933155,
 0.5943573587822741,0.7742609675677172,0.2830861882928156,0.18932680711909566,
 0.6247557140410167,0.48856907830344565,0.6960641465163752,0.9297656157755357,
 0.2873474628099679,0.7770250462905566,0.3626574553772871,0.17171494055736913,
 0.6394261438032494,0.4029915167214283,0.7502572192889168,0.06687117861485437,
 0.9170188492878203,0.027093403476171063,0.03372533807942202,
 0.7916967306313704,0.4072784822716582,0.6736398567986567,0.36897037668487387,
 0.9535911641424425,0.6622198337882841,0.01952436734422036,0.9329044069272536,
 0.47054574904166846,0.852104755658885,0.34507552112495055,
 0.12374223268181606,0.37737634524137564,0.8814755739744417,
 0.7234073974809136,0.2131031735679091,0.01424006696332969,0.8718756735962756,
 0.5238838940932844,0.198621175903562,0.1277328470281096,0.48211339194947345,
 0.007793492614080244,0.07840877514856459,0.8958063776375342,
 0.4365251113570414,0.8662005601018351,0.5983981368234002,0.5101521271255183,
 0.35596185281592985,0.7178704372262662,0.006214592096248861,
 0.7173660439426942,0.4852827793845049,0.5609073575998093,
 0.055788391558256434,0.37858460390392334,0.2888759059832612,
 0.023960780487075284,0.4788514491772733,0.4268871974375693,
 0.9721464488404892,0.39611227854157616,0.26748034776578744,
 0.47575346211897285,0.41728143844950927,0.3387406946434869,
 0.8480406691412415,0.7229627218524868,0.8714867873192416,
 0.033600138057069406,0.2116084997886195,0.05767269485398818,
 0.8124543905141579,0.7510607778806937,0.11461525499929914,0.3346369269592986,
 0.7197310718357809,0.6938258640600516,0.005984785076500132,
 0.4202192383885469,0.05761049810177754,0.17009089990002768,
 0.9748039391055235,0.43046532595494946,0.02836995141074594,
 0.6651190438147715,0.27418934552630614,0.771633443721077,0.7559227847252961,
 0.5467558393206144,0.3115084272511397,0.07248187883438761,
 0.22862697743558025,0.4623375852114252,0.0707006477891251,0.4057076992020565,
 0.1259439638797617,0.7825004509879592,0.17980147432221938,0.5018740311623622,
 0.2987714466417625,0.2692374470289569,0.8790618997687912,0.9552088462465732,
 0.8374669340206964,0.629581062939127,0.0445842535264227,0.9519877059007307,
 0.9425438239076759,0.99471194761737,0.599428515837073,0.03977010144352666,
 0.21704298369831665,0.9941991942869568,0.006883082521201511,
 0.8589843301186488,0.14450868169507392,0.9683018540949222,0.5949202130085318,
 0.04733798049495663,0.8537897862552464,0.5065119385110276,0.733345985492245,
 0.443451808537191,0.9925634015139735,0.8681988398814914,0.6343154204125847,
 0.23489279122626205,0.9637087907369926,0.022991349604014055,
 0.2630015232184647,0.8621047822611017,0.23782395365333786,0.9887856356367346,
 0.5776642809336939,0.33879884175438346,0.8948114324850384,
 0.12071170218982952,0.24130439638190004,0.2861658343523087,
 0.3323601903966439,0.9447876715797169,0.01886714369827891,0.5235896903407008,
 0.9001444241366668,0.16283563012168345,0.8175916886867891,
 0.34332387103306117,0.663480371639558,0.16191576716297784,0.2998597819376021,
 0.8872033503148424,0.1681957903519199,0.2997252333137975,0.15778367549596295,
 0.5624580160178803,0.5965550141874576,0.14457773559800446,
 0.10507058291580729,0.7766058046036668,0.24516217037038124,
 0.2091652117034999,0.8295820880503212,0.4160109144345522,0.9158857241405463,
 0.771461354023013,0.5052285445120217,0.8498536071295553,0.876729280009141,
 0.18806024353356954,0.9580106372854351,0.7522403679268378,
 0.35191791502613967,0.24696706407519764,0.3553338183205468,
 0.008655671063555137,0.11639183445741386,0.5944146701001604,
 0.6902772484626059,0.7518959943301231,0.048887618369964136,
 0.15969224454227804,0.8792197054910438,0.12742176820513718,
 0.3181621694189545,0.7792274458215066,0.35104885369651373,0.9567560095036225,
 0.6340267506804541,0.8074235626259527,0.6257293898302609,
 0.035960028739462024,0.04407817266717384,0.1537281064287408,
 0.2572060362990838,0.5114131936921417,0.3453915609405871,0.12135873394478147,
 0.8802909393838563,0.45915549530286626,0.24298657686152259,
 0.6539701722296178,0.35237947437036965,0.9688409864718204,0.6945882758501041,
 0.4886068101621128,0.4357812456327488,0.10298292573250012,
 0.21018241783419034,0.9544162262628239,0.6564819687230289,0.5295822632361815,
 0.28446751636022594,0.6527320801606694,0.5232597096413941,
 0.31984149019414987,0.19815708222418316,0.9949424955185743,
 0.3803842949644214,0.7721276292681057,0.1909075254848398,
 0.044688681680383424,0.5823818627497246,0.7203877894986261,
 0.5325356749663842,0.9852155847852112,0.4912632761546504,0.8007626724246575,
 0.38623721227348384,0.16052625647595553,0.42890693562454085,
 0.27619281765673853,0.06721049607660123,0.375806563738207,0.5564126905799109,
 0.66121062122598,0.8279725508351218,0.568682185817086,0.3081320332412223,
 0.8425449395050554,0.27744626871679445,0.9865732297438897,
 0.35906120539835296,0.37086750331289353,0.44012499135154737,
 0.18076742477044117,0.9861568018192538,0.4840630363261196,
 0.24777884959515628,0.7165067467388547,0.8359030975112128,
 0.04260932288245445,0.9582135348938333,0.8333473243226941,
 0.17116301030128023,0.4469603820065635,0.9921725449922225,
 0.18197757630038736,0.453484939707019,0.8431302896326506,0.74602699920981,
 0.8257556355866396,0.04488477241917255,0.6420165977933949,0.5048296324622055,
 0.5068759131065,0.8891546227678603,0.6006269226901122,0.17116737044903596,
 0.31060282429296326,0.7224594506175664,0.813158294650709,0.29351686755847894,
 0.39645923013037443,0.762702060275713,0.39579952398988283,
 0.13068516483408743,0.5130579938235043,0.10657387242545024,
 0.9039196694224227,0.28185878800038533,0.37616546876578616,
 0.39803947573268283,0.14066076881157352,0.17423746636388004,
 0.7638634654988272,0.6131886805527066,0.1929803359929353,0.7879311317462502,
 0.8160831314564253,0.1995377495482652,0.5351096654734228,0.6039447400904361,
 0.13019603039668937,0.823071469754699,0.29799182937820934,0.6955084935952924,
 0.9415187561401871,0.17551213820403444,0.5594931584375644,0.463103240653265,
 0.6674144756576237,0.1488109994894985,0.5284266053708392,
 0.011339858054830287,0.2667255957005542,0.595944671760762,
 0.12693359666665227,0.1846209360109945,0.4311537257108794,0.7441779693887997,
 0.5026584532278862,0.9895344024378092,0.26516018097890737,
 0.33786178365480213,0.5331152813567502,0.30215703173388275,
 0.3738380625316756,0.9988612368166101,0.13379234778921534,0.5391154049356812,
 0.2794283693092128,0.04326278699458419,0.5737825711353879,0.6093527701816593,
 0.09661866728568846,0.3822462019243651,0.14424386199606865,
 0.6886536352677997,0.2795526461176573,0.49084105724317917,
 0.009277772662487638,0.8517289352206451,0.2970818200832579,
 0.18082820832725421,0.9775818256961702,0.1898154161147626,
 0.20033240240950811,0.4441424514240606,0.4195622608176389,0.7015050352239657,
 0.4226704330038382,0.9418488593813523,0.8658046586607342,0.9196001823349065,
 0.7466133932571504,0.03429965246318367,0.25796316386537377,
 0.7898036276576802,0.11533640251768063,0.11929107190834332,
 0.9225533746186855,0.7068594213634567,0.6002651772262386,0.3546633233808989,
 0.35132168507980965,0.3110931856978367,0.5824550166755831,0.6944677600627656,
 0.6102565631524471,0.9782629599697464,0.7584897038053612,0.6520871079349226,
 0.06982606427141702,0.1008547128800843,0.28910955742640654,
 0.6972172575164075,0.5782015830603746,0.35030163949215176,
 0.28031024088162415,0.5066858943669434,0.5510414183833248,0.7041644642765963,
 0.1403388131850063,0.18468138173204673,0.4678733680882736,0.7794922207366267,
 0.30066558094541573,0.9852711731863724,0.5136461065982,0.2888726749672198,
 0.6244744472629856,0.6574156256817202,0.4730064857878855,0.41473267068825836,
 0.28163869558239996,0.3890073975249452,0.5061656808178665,0.8582139609936917,
 0.20794641578902184,0.19136517846706602,0.5776145673119653,0.649205262268338,
 0.23241162315144515,0.7951341475723022,0.1147566616868142,0.7069208378149404,
 0.6280483329233018,0.3231212358790323,0.31777028097425053,0.9366054646100922,
 0.779279404892326,0.822408032928377,0.48588332345502516,0.7070365694714998,
 0.13368509499404424,0.9314265729285088,0.6168706350860855,
 0.45295042983028944,0.23296276768706037,0.05089982518508551,
 0.8685826846035396,0.7809232684567656,0.8006002121367296,0.4124853208453205,
 0.9698940762855524,0.7675985276267959,0.4676806484445738,0.9971342678563448,
 0.32294551883848943,0.2235740508169176,8.782254404506862e-4,
 0.5888945791566381,0.19267439149638166,0.38754469878814535,
 0.7784633375103814,0.47234422235206197,0.15592544187898216,
 0.043122127696150425,0.8683191658504893,0.37872129177032954,
 0.09926098426665209,0.603913695812617,0.056517325786060946,
 0.04582020729605796,0.9733922474679078,0.8541578215912028,
 0.31010555771382364,0.391691265110101,0.09933530865183837,0.3357458035007972,
 0.955440216705838,0.9803510683313594,0.8957673324443134,0.5921962346233661,
 0.23103763495517882,0.4161831018691867,0.08840518036921063,
 0.9346116903506831,0.34408871426297294,0.37125439090670254,
 0.3166944789595928,0.4898894184691347,0.8141017064777133,0.02999970366120064,
 0.9755802202257615,0.5608109844746577,0.4254472225549958,0.11208065274644319,
 0.11376793729723933,0.16772769672814491,0.5779623348484819,
 0.24072243963768925,0.5148809820747773,0.48743981054784147,
 0.06683335362465637,0.49044559068943994,0.1502501308266465,
 0.030873591292672753,0.8936517343840329,0.5717831981441489,
 0.9456392188155669,0.3513064735186,0.22951095041217995,0.2321781441072862,
 0.14192730570336554,0.9291242990815448,0.3260699070728028,0.6799996003589533,
 0.9207713440313372,0.4126586932461862,0.3645906920595855,0.18861059457006146,
 0.39034411951186143,0.700094809986143,0.9416564822790844,0.08625600400586353,
 0.06886642260519271,0.17909420407294785,0.006208591559122345,
 0.3404321263078408,0.6577803738582507,0.1499741226711242,0.5426363287385662,
 0.4566973765603257,0.7509285977041409,0.13310656838328738,
 0.051210611826409735,0.7527653237796872,0.41075217168165923,
 0.27811034184417516,0.01634000974673666,0.4164789365297277,0.982806595977769,
 0.9368255134743247,0.22977048370573427,0.6518489823167471,
 0.22258067862094177,0.34546404750307613,0.28835639999354623,
 0.4033540244810788,0.5256438739187682,0.5590402475383696,0.8528069415057958,
 0.5669173732502018,0.8248055092373674,0.6572686165607777,0.9320134901640466,
 0.1995952548008587,0.5675761125232883,0.198731995008545,0.6437544804306021,
 0.686506254321637,0.30790028073840825,0.5289759018157107,0.5127550465860526,
 0.3947589357346688,0.7724325751492231,0.2123209798691441,0.9772191730365749,
 0.23534497311063574,0.6743071977377869,0.27116542321257375,
 0.3329882148623404,0.013776144989143924,0.1778314410735906,
 0.1321394292424123,0.36445211085962226,0.775746748268913,0.3659023148765391,
 0.35150165152383694,0.9022488268629608,0.24809522978255139,
 0.8047081424512641,0.3043956671867305,0.459896541705353,0.14562134224168166,
 0.5173913852118532,0.17432295039772072,0.4024015038552935,0.2622970672187779,
 0.348096208825982,0.5432833845456501,0.5061009435787565,0.44287075629297146,
 0.5640169688172947,0.052274929105134316,0.20978957567636713,
 0.01900698940180856,0.6054358499078586,0.48922069379392485,
 0.10216118926437079,0.8538307399006589,0.16199912631865065,
 0.08579558314799274,0.2918402349372107,0.06999118040668806,
 0.9523579710879903,0.4676564525203164,0.7757039278323333,0.16325100153462768,
 0.1444722115319511,0.7700180778845651,0.7644586414778047,0.3716000113593648,
 0.4969010507180792,0.6547745701330148,0.16498020914406897,0.9546133180812278,
 0.8387552090144483,0.7665366003892473,0.5073525803992402,0.2943323269379712,
 0.2871985377076136,0.6654634018047796,0.9698266322005239,0.23231499458458693,
 0.4374884516396593,0.030727715305860825,0.45701808188694737,
 0.07443526834168346,0.4602350362016081,0.4784916768008708,
 0.18664205098768405,0.574661502079943,0.1577479966898594,0.5521845338820323,
 0.15778679687646258,0.7663472229076946,0.40615346288059895,
 0.6967032243119324,0.8301000272604665,0.10888035982771793,
 0.45408871871186196,0.2142176532789749,0.8667255588206225,0.9958519516290514,
 0.866566975210822,0.02149512859030467,0.4401873331027314,0.7251218135652316,
 0.09646532728762836,0.9615550172756607,0.8280831071904526,0.8282414070949407,
 0.08232424724175325,0.140300777247247,0.8816759146841221,0.2776223046073172,
 0.5777343398322466,0.03057136964103402,0.09376778594333901,
 0.9340679066167016,0.5633321549206779,0.9895089986133839,0.5337785979147182,
 0.33473817861418687,0.3219958965654357,0.8444153381185059,0.2651317499292011,
 0.0016080863755559838,0.8534878626623644,0.2310611856765531,
 0.5997892633745896,0.01418776665491106,0.7563239150037786,0.4910538030355941,
 0.6952527424674045,0.737988791688743,0.32414380705990686,0.3488554025159515,
 0.7519931238113244,0.48877516320065495,0.9099620389954511,0.6164385445762879,
 0.8189986153327686,0.03761680796555977,0.6466922525219634,0.0311351497357647,
 0.01091519671387986,0.8563240327667003,0.45234832476367137,
 0.7515638467278709,0.9989579032626593,0.2197864074906707,0.5721712073614758,
 0.20053450313140642,0.46587585515718344,0.10752204088884676,
 0.7428124374625504,0.5618635950765817,0.9426708774386728,0.6343363468685714,
 0.7184048888584963,0.23319435440710712,0.35274753937303815,
 0.6589013120650501,0.0720250034181763,0.6322452386924118,0.5144271503259902,
 0.5588129722777926,0.8645237737666551,0.8621811267390787,0.8212418151741735,
 0.9858990565081959,0.06196162280201345,0.30800196003436986,
 0.2405564392352939,0.218575760235997,0.7065556624709652,0.9445129531032657,
 0.3432597538779375,0.6417465026638421,0.2114153525370277,0.38903797644535365,
 0.5115093639580689,0.6783462864997755,0.3518198371146809,0.4051467577109822,
 0.8379146921153682,0.5332160158403906,0.10234496252040326,
 0.0012577879822199822,0.1124786996397018,0.0830793871428287,
 0.5760286320551765,0.01062542103169406,0.6740954484482119,0.2537222381593045,
 0.5420933543927803,0.7648142787891463,0.38577432164333136,0.8832658723195451,
 0.7819439857599786,0.4564286749812212,0.08224136657454295,0.6052806249669483,
 0.8150032131679437,0.32998377135607915,0.0689699954470514,
 0.41582509487075137,0.05382886558119593,0.9244372381348875,
 0.2411135175086352,0.634957347325604,0.7459450851632936,0.004820092196082704,
 0.33335970615303934,0.03719541588735786,0.6332433567757373,
 0.4499503562677375,0.9945832448411188,0.4009824084893311,0.42555327196341886,
 0.4060658506116278,0.5218333248239755,0.7028090560497839,0.448200968550451,
 0.9200767429909424,0.18276183904859145,0.7511800805198646,0.994123560898848,
 0.013040609525764912,0.0027164963443948587,0.20143904310508265,
 0.8917703131363885,0.022326325653696633,0.312946547381456,
 0.21449999013456833,0.5472506495823342,0.8447300764850977,
 0.09268037887756586,0.6088088395009119,0.6247455639021434,0.6827684501256377,
 0.8969631692242312,0.06137141496351739,0.831987218390642,0.5915470745773571,
 0.44729190100656513,0.3857802762462158,0.24268219725331075,
 0.7739687711210296,0.48010486072431924,0.9445237745815881,0.7688538201350918,
 0.41293431520890356,0.83606627314512,0.3742777909638282,0.4126820587003023,
 0.06033933063075425,0.8944962130609058,0.7005867733616674,0.2528342086970634,
 0.6950824554964614,0.39638045334394456,0.883691617997571,0.00335369650852102,
 0.467924362370046,0.21015546935283957,0.21580266948065763,
 0.02965844372389559,0.8707617030754381,0.7223126907804838,
 0.12345187573040584,0.9327097796504472,0.13707677200618218,
 0.7238504956119169,0.36761748230274816,0.03538629073067889,
 0.3287324449012443,0.4238792390713161,0.22202679479369858,0.4146512087109311,
 0.13197594431438475,0.816276017563489,0.2462918571455458,0.6636700618913323,
 0.5396329582054047,0.8189196077916172,0.09210060879748183,0.1007700383064285,
 0.7928053092540084,0.4247034806170191,0.13291037714427179,
 0.31617120707463287,0.8940177422326221,0.14451795624535024,
 0.4594182194582085,0.5381793165000363,0.02090683845269803,
 0.34982242745874403,0.6388906552808855,0.10883171044335205,
 0.8907815214205086,0.6776505133489661,0.8933482195631861,0.178650411018609,
 0.051788360058251603,0.4644233414333754,0.44798931851869483,
 0.9933624507976185,0.7056929960106787,0.01749599818276215,0.6599654833981623,
 0.7502744740851323,0.4667705979607484,0.3391445030676584,0.30102516911366195,
 0.49869911577347514,0.13119668735911572,0.6311476504824581,
 0.17542736890234178,0.47516824390930756,0.4927729199776373,0.408758395562995,
 0.2559146778766299,0.04090883547256097,0.7476874591200316,0.9110343227535787,
 0.707517037409156,0.9162457657798816,0.11125894641400114,0.46345113670483706,
 0.5458792534692751,0.07657336142253279,0.24368376567518102,
 0.6805601479938982,0.5845818458969414,0.06145523945769549,
 0.21921736098207012,0.14184792692707637,0.8278238790020337,0.99467680760675,
 0.3870772845509506,0.12064925550013295,0.6635136732096222,
 0.06746283971528055,0.46216752641450565,0.9803315244214859,
 0.3746109224849581,0.6561445962005393,0.4124578841579567,0.6725037036826379,
 0.4250036332404482,0.07625454046266888,0.578483461831868,0.49982644124869935,
 0.49188705225607776,0.37612756231906386,0.3124918521201594,0.947041225217856,
 0.9400882054967923,0.5379915519600325,0.8797097960713282,0.18178681625423343,
 0.8262166511948877,0.3043255096272791,0.7789812104363024,0.6886330746655548,
 0.9426286623316371,0.6489246959450619,0.5224080114547456,0.8044715764872744,
 0.6234036320772942,0.9913905669138451,0.9010652329860477,0.8948590976969927,
 0.399179240071718,0.9070786669704727,0.4935881673812741,0.16546696595955135,
 0.7702963716423482,0.5306436587430717,0.36172665481548993,0.9388342620327608,
 0.7103093692090324,0.27389779161508376,0.34649922602371264,
 0.8362189623708965,0.7195324177500798,0.40167448446505183,0.8029372275694856,
 0.9031401777201014,0.7620176039791338,0.005785848499639368,
 0.9270313700227731,0.32121385927004853,0.33585629606534684,
 0.5439731904115872,0.7805954400783354,1.0834120786507029e-5,
 0.8117648094238812,0.17186737017029563,0.9785287989198602,
 0.04825007543949311,0.5380035587912361,0.2821310219187734,
 0.43992735270810335,0.7841378148062986,0.759984888318442,0.2317982083107195,
 0.1823258346550074,0.1193915297425312,0.4518941350347089,0.7964978237503952,
 0.6193003931428922,0.7914463468554385,0.8961638807352303,0.5597105218231339,
 0.7873883116682194,0.2414169383589524,0.6397147528409336,0.9926818880090529,
 0.6472964401095047,0.7652269678379309,0.24499323603091672,
 0.17513807768763734,0.08383152390846582,0.9020066937346476,
 0.2996595136752269,0.29938508194169433,0.34017777257909954,
 0.5745221722158859,0.3788644330733123,0.2615882243384706,0.0087623341675368,
 0.10086933117802754,0.6948834430684909,0.2543493716244193,
 0.013542969492780443,0.9506525071393189,0.1515158830533503,
 0.9356861880996074,0.7018090548589642,0.702211163320202,0.37974926369822015,
 0.8182195542272332,0.8235596299582703,0.6561277027882407,0.5450918221465589,
 0.5164274430777964,0.6982525933088968,0.9272859195472007,0.0474194896946033,
 0.9783325892737025,0.018805598435500115,0.9947483769254727,
 0.2822418044891706,0.13163559527625324,0.14241320726227025,
 0.9045527960385408,0.9896681435914189,0.012478013810906186,
 0.8071538667037954,0.9919274753279499,0.5086245285659092,0.23666201823684374,
 0.39448171279517685,0.5413236657888518,0.4022188212119324,
 0.43299380923199904,0.4796475248913006,0.8630839510196746,0.8387371014726682,
 0.24544816980463147,0.6136160073938256,0.45798938628717567,
 0.6460319246453394,0.2034738489829092,0.25574944471727656,0.5840385867604139,
 0.23574885474476137,0.3877448917699984,0.9230848898108356,
 0.29494416582591076,0.014083600188733114,0.4840506114915176,
 0.05720237990965482,0.8167917386795497,0.1345110456361185,0.6977465512174613,
 0.8104388551425258,0.7164741469095204,0.48184019478927187,
 0.08956889876114582,0.1516078111892596,0.8263013048628962,0.370605138079918,
 0.8284087222877146,0.3323947713911306,0.6388628348014903,0.17467393403555287,
 0.5151932395415141,0.7211266396794047,0.8718970538764069,0.4513780352673431,
 0.34045339545994135,0.6867046730638969,0.04227704423767187,
 0.3547463332125401,0.6180630732713039,0.8826474748300762,0.03713477491770101,
 0.14235594104324356,0.3107706079891126,0.8613180783037188,0.8744660569889853,
 0.3074589561115,0.4426000840302904,0.9606234230708958,0.23305320396733742,
 0.7360701411287929,0.0961516554242613,0.5722296429502887,0.39476306428915797,
 0.18536171389611844,0.4646255617841468,0.1546007962510384,0.7388365937251624,
 0.30330735963071564,0.9428057823692552,0.11968857339384475,
 0.23585484914895305,0.6810237568487055,0.75862520467738,0.02666419020186428,
 0.7046112882665574,0.08259441841424198,0.7489022291320672,0.5869092130785312,
 0.20226061497495795,0.10845939492787537,0.9749395309839674,
 0.002949401117940842,0.7598512919343959,0.09974879115054225,
 0.9750011851918517,0.6287685758312667,0.4137641618990009,0.3704915533962353,
 0.43263260228070344,0.6678070638929073,0.7215661739886694,0.4761675627232661,
 0.8725913659088174,0.5954085819784533,0.7957675062916023,0.11423524270062291,
 0.5191569056473223,0.6023166486928433,0.13115321293615012,
 0.25237398662687527,0.674068135461136,0.39433577754442195,0.5143921099282107,
 0.0784498642324607,0.4147497362056094,0.9405636403668137,0.6063599534140203,
 0.17799423642659407,0.13633824508962145,0.7747199303957345,
 0.18080626531854382,0.48180639502816025,0.1630911103113155,
 0.11900640109906246,0.8069829283602128,0.11490841706800536,
 0.5619981662134667,0.2071526027726791,0.4565462366187656,0.9146830300245903,
 0.873613168622139,0.7989139329992265,0.8819643390556606,0.37194085336354665,
 0.09244361331337758,0.6065563145869161,0.444932324077176,0.08860310988967757,
 0.425096958515898,0.5155813824559181,0.06158013951906205,0.3312250027950978,
 0.7495114877503238,0.5703551629391175,0.8065384499678014,0.9265863804532095,
 0.18193154824066948,0.7017963640812204,0.832618414224883,0.01030126448068569,
 0.4125423821252694,0.26174049200711547,0.21894296371142996,
 0.2799960473136024,0.3953285268247806,0.6762936718023569,0.342260641409873,
 0.692082304347253,0.8117783878501641,0.5055797298723378,0.2610931476665346,
 0.45791606686772557,0.8035218070069463,0.4795398262842314,0.6790841004778381,
 0.9985702076384785,0.2562571562990673,0.8770725275101119,0.35857353241975165,
 0.24540361404610422,0.9865218237033229,0.865489734135946,0.10865614112478839,
 0.4887591342218911,0.013413281844119163,0.36746591977653165,
 0.47196074947853794,0.12456494305519028,0.7807303068979191,
 0.03195028640150133,0.9542494909112025,0.9485237561503653,0.4475297413781606,
 0.11501850938211422,0.19169655759296567,0.6449592622578064,
 0.37690975853790154,0.023293940460151097,0.004767139485042038,
 0.9930596091309163,0.40522157830698313,0.5826794957853301,0.9532212718376236,
 0.5229713658717723,0.9859246481437764,0.3744733321796381,0.311132219072752,
 0.6456610294192273,0.830497848265474,0.6503472524848541,0.07035838517650506,
 0.43728797456967616,0.4910771503119815,0.16664244699258568,
 0.36329584040099416,0.5719224331537677,0.9827297534592092,0.2955075449515704,
 0.23633691560864578,0.5328603655748918,0.7882218711706567,0.5563342136370872,
 0.7022807692870605,0.7300810773098669,0.2311326704010892,0.8482480875463673,
 0.9599993960171144,0.3741142804302148,0.3104037530770134,0.618783679927072,
 0.08714511390182045,0.21655260369344453,0.6375558298908617,
 0.4570326742576455,0.1644580720645582,0.6352500996779213,0.7646846645531571,
 0.02186002385135033,0.9222344849173578,0.666065009952998,0.17392289711951592,
 0.6519036085848318,0.07702658227033965,0.6937867793467192,
 0.24889574263418157,0.9286181380609364,0.11240080426116661,0.538882568300842,
 0.621345048337442,0.8613253377766668,0.17680307931438,0.5527840968629318,
 0.7846454731837099,0.5957490131034888,0.9395182214239989,0.20008816831079224,
 0.24761558188867672,0.9677496159667697,0.7249796236840524,
 0.32235851975921914,0.8349102148956759,0.00394418642737604,
 0.33289441810642484,0.20035961147170556,0.47943038244422154,
 0.6403826576208524,0.09199349944210855,0.03469721062810904,
 0.6511807498339325,0.7063385610951836,0.3952883392732639,0.5151183233145574,
 0.3555577782438236,0.09421904708948958,0.9894910471544682,0.8883614975935855,
 0.046641124052329896,0.1277429996253805,0.6112607903965026,
 0.8237400648098541,0.17289665521455855,0.8375216461799632,0.4112813780799157,
 0.005567993682980776,0.6775425718524982,0.8170230559739574,
 0.8916497098449749,0.41855066878937064,0.3698121507187859,
 0.18881780319096553,0.1352582894153278,0.9034384619000213,0.7421445194396957,
 0.8547239869177122,0.19482059990387957,0.2159324809207862,0.5526674256379642,
 0.08520642007213086,0.379384071816508,0.8374437514725965,0.7955997201542642,
 0.736700233120135,0.8438780491310031,0.3262626983088768,0.12284981667335515,
 0.5355184528029513,0.2204933886995203,0.642932868194605,0.10449149387176049,
 0.8440450374585052,0.740711118141319,0.6365835121543311,0.36486478114838083,
 0.27580948564668695,0.7591851953558262,0.1502268249919132,
 0.10340470875893071,0.5176423840921989,0.09310668605680128,0.463779684823471,
 0.30530130910946207,0.23626298315375738,0.06427880480930614,
 0.050451499856568494,0.04551689693213923,0.07271335586887373,
 0.030340797978191003,0.03075663415154528,0.34476180084049546,
 0.633224765708318,0.605571097711756,0.12059106824247223,0.3658020978443278,
 0.21312458780544397,0.41879342041399525,0.7141597472797006,
 0.5155529339443827,0.23409704897635164,0.8296145812303157,0.8190380973063293,
 0.14621984366270357,0.7230378382153756,0.8812494878690279,
 0.08709851857580664,0.8776557735489584,0.28519068728699304,
 0.3508379270641291,0.8244929791461869,0.35169777009141845,
 0.12989048929271652,0.6650347244962463,0.4353438833629304,
 0.06629375927331127,0.7856058360486262,0.6507114314037415,0.6610627619407069,
 0.9797956624128885,4.96369509177863e-4,0.39144070460768676,
 0.3210744446425784,0.8414929194279055,0.36352215872770255,0.271550687540385,
 0.2565458835551071,0.031110858736003055,0.05882696763637307,
 0.9486194078056335,0.5182013126227514,0.8465218275910855,0.16115542230815905,
 0.044400201820490315,0.734708350346612,0.21139083362492594,
 0.8705554316830691,0.5320390079794988,0.681740604492554,0.8028088639790302,
 0.9896232793346147,0.5597504555004721,0.1969226860756933,0.23409963518521637,
 0.7224802121944431,0.9876131750799206,0.18995065376928,0.6783535408133606,
 0.9331004460892253,0.4618386642277291,0.019344140680398803,
 0.033092603208449645,0.8549352071705585,0.32590715028238326,
 0.09470246471919164,0.04725589447533329,0.7706780185564422,0.715290551279596,
 0.8625599747163802,0.860409377046948,0.8771234068035202,0.48171000509418227,
 0.9229008848881501,0.8872453930129396,0.08065787173561556,
 0.03410219911855372,0.9503610997427682,0.7557018443807566,
 0.16375608746969517,0.4631675428531494,0.7650216280351764,0.7569794153333695,
 0.5579924969573127,0.2327378429401299,0.3583068799994007,0.21257605667750812,
 0.472268121723468,0.5323937728404708,0.38148585961162407,0.6513465733313409,
 0.40945241848446534,0.8539462620769804,0.30399896577921126,
 0.11371653040222318,0.7866928819345154,0.14739860825459672,
 0.3200120335520863,0.10713155892241066,0.054812211227034946,
 0.3858053509020688,0.6418085412776134,0.9153919394020317,0.8207382197395549,
 0.035198964905122976,0.4881619187381607,0.7635939697167504,
 0.35323853562243945,0.6316923699234397,0.9475615962223924,0.8899163077682899,
 0.504251428198841,0.9673478371941842,0.9605021835008848,0.18952278094009345,
 0.6540450003079166,0.19021620984866072,0.6262169901051757,0.5320867517937676,
 0.43568856822021496,0.9341274387500175,0.6124780309641669,0.5525691243026654,
 0.8171782104415858,0.6548382052407576,0.248619299416318,0.7581333885981438,
 0.9280545977928418,0.6253613842352684,0.7111464441421121,0.10665357797179775,
 0.890684976715897,0.7247764429263907,0.6845133242836398,0.11690401154729546,
 0.915149861026716,0.7049817903924719,0.2796978480200759,0.8813358023512115,
 0.932218200371305,0.5831674216927774,0.12030412126195444,0.14906624583330452,
 0.36087145192304226,0.5683331282443901,0.5325079546902924,0.2869335745661038,
 0.783347436341652,0.3969545988311456,0.5976659951378578,0.6634224918710184,
 0.8921448490533233,0.32459644689027756,0.16124035577126872,
 0.5954924953866612,0.4434643279817183,0.7455704448335468,0.46397363773857936,
 0.6064311992488238,0.2075297422881177,0.6233258745789543,0.9322946415385847,
 0.012688944599436702,0.36624298564915114,0.4321980097062269,
 0.6628862534223489,0.6310231889618826,0.12367614224158652,0.9577849462277839,
 0.5212196925042528,0.5919255657442803,0.8124888707071173,0.6210031860924459,
 0.19095912607105436,0.22043412417784947,0.12627247863500635,
 0.7012514542246366,0.7632425128347674,0.8569918152667712,0.9828841248113017,
 0.6904243950481793,0.8360190785610804,0.634920338038047,0.6294062881642815,
 0.6551537183072051,0.13063679786892246,0.9199895670841248,0.7901807010805079,
 0.9248021328607917,0.5762472743278235,0.9114101735427017,0.34690800101711905,
 0.8904155946225313,0.012232328954085547,0.8018596492840863,
 0.19663686584539442,0.4981794032822141,0.33729210298078016,
 0.4182609834202271,0.04720472885119609,0.04763165125853419,
 0.9495768259699198,0.20675639179714922,0.11737454460523289,
 0.6089071134326307,0.026512292832692408,0.35401720777652956,
 0.6910494561858804,0.25758819585788495,0.8489900708587194,0.6136771309879414,
 0.8087929230719801,0.7028930647357858,0.6328812410382568,0.18474185339717397,
 0.20343978476116198,0.36594762109959833,0.3217135509956396,
 0.2286708787286128,0.639518503445053,0.45376219719200783,0.38062226315118786,
 0.08753072932807071,0.6663486312009897,0.3697965491042704,0.1869845386001988,
 0.691786823587886,0.9887087325908346,0.3203681745490229,0.2145262076545652,
 0.9286732997708522,0.637790160646923,0.03469789142563895,0.48579563224373123,
 0.4219018005621946,0.8995972743915598,0.23663573991655595,
 0.07643146810427592,0.6173051300864093,0.60256289965067,0.33440648590926925,
 0.861678220971597,0.5770944094570961,0.896117923517524,0.8058502948896383,
 0.9730255715028094,0.09561594794864214,0.4903678729632919,0.4824189144495007,
 0.11403461658707004,0.33033163295608015,0.5022185277912632,
 0.5720489089910511,0.19913041484946004,0.40458592382513914,
 0.4175489152079892,0.19367357395931473,0.07388383748532268,
 0.24882511820479525,0.8917112627595492,0.512881258220498,0.38791804007852193,
 0.22164751844224329,0.2280088492021669,0.08222945721928099,
 0.5645479170293243,0.256947749636355,0.35726731675916124,0.5350985711751391,
 0.28847360831349433,0.6411245652984379,0.6459167396761444,0.296983359300906,
 0.3171967142196106,0.6226286843421656,0.9204030987367068,0.6497256169989738,
 0.7257389684351891,0.22653661507270906,0.7839263146601286,0.7394347857800905,
 0.42450727644641184,0.4840040193578493,0.32093728256020204,
 0.3449351396890229,0.8010506845423822,0.47631125540818386,
 0.0038694597741928405,0.3220057778726665,0.27868639141266727,
 0.6103265997648117,0.4797322693062891,0.5390921276715807,0.441481803183998,
 0.8834242147325241,0.9399399207923188,0.8926056288564832,0.41779539873988303,
 0.6866023005493032,0.2328816905240495,0.12436206470304312,0.8142144066848119,
 0.019550765493330347,0.486699461003378,0.14935385063408047,
 0.5055612115462148,0.7312412609930387,0.30778609697808734,0.4911703058560456,
 0.24508233984163097,0.40488631119999463,0.20335517982053397,
 0.11466299582229078,0.04481858305800479,0.41610200344287473,
 0.4269338341061718,0.7635106970743377,0.7788290569533983,0.6609481330588092,
 0.8608251180243434,0.6986251397224449,0.338937219581413,0.23928177533519945,
 0.5537099955778467,0.5619391767225912,0.936115565911175,0.8032343203653454,
 0.8057240963044272,0.8566562973061956,0.45218216032285896,
 0.14488325427361004,0.20605941293046415,0.28283236745338813,
 0.47619450123444906,0.5838980111824155,0.12941180743529057,
 0.058154235984982705,0.27461240448853474,0.34767591249890917,
 0.004509690330912641,0.6535519256142356,0.8312332483952193,
 0.009194945391690301,0.18341762038965204,0.8097993593281996,
 0.6015112828033724,0.6455123043803034,0.5404180462764455,0.16261412426032185,
 0.12719278902018238,0.5132885942889359,0.40638534392107006,
 0.1719441992355949,0.9841542784610446,0.8833764997297363,0.9396891824289386,
 0.022452575727465285,0.8121336807477527,0.6671906947851165,
 0.8032279182856299,0.7227599484643946,0.46897264150252016,0.5881823239731012,
 0.2758515840746071,0.41931427554341594,0.5389546421599954,0.4834989269248876,
 0.36130575552623334,0.3327789723332709,0.8638093914291174,0.4699012996716494,
 0.49436069308272046,0.8488554545031227,0.7835202733505071,0.7926333265700718,
 0.5193622103267358,0.761021118880558,0.3291648619532377,0.284645509045669,
 0.8421947919460311,0.7600723881954258,0.30115126637029477,0.8041776541639029,
 0.8120813694834292,0.6632926533128197,0.808096377439681,0.9646911428471246,
 0.95516118886453,0.9887526219279754,0.3206682246262289,0.26043673484310337,
 0.443175425971141,0.521798637564032,0.29935452737232215,0.653596377388154,
 0.28967084923978903,0.34165393098851804,0.6405651870416093,
 0.9121112390722874,0.6534317399618592,0.5171975751445878,0.9653947678971383,
 0.6132530247230158,0.5417565867070521,0.9555181064604571,0.6631352437785407,
 0.7387760144334643,0.5761228494816591,0.4785178255040107,0.17452974291658274,
 0.40327379210005776,0.010547207490614108,0.7033951281953048,
 0.20493349655541127,0.5182304244341007,0.7599623778538838,0.4098417051607075,
 0.0816949869068424,0.18863514299705386,0.5337417229695567,0.8411469086370768,
 0.5073909453973205,0.05564345636227053,0.6966380940952753,0.9328490454405765,
 0.9768407816077196,0.12818886160545162,0.953558384312414,0.9411696100901963,
 0.9003986597159946,0.4547906174993893,0.5934815428685596,0.3399949874267567,
 0.7582501117469075,0.45612747797600617,0.053878226749311864,
 0.14138173153454903,0.8791854440890718,0.5493312521307465,0.7096666885173424,
 0.4556688697615371,0.5146071960751752,0.785784757109607,0.5916380225876477,
 0.771118183798752,0.3192056861773942,0.8818568772468635,0.7624831621178688,
 0.20740153813290296,0.26736048225046016,0.8641017817757143,
 0.19174802559412074,0.7627058191763713,0.011096479182017727,
 0.18742938092117456,0.362841375771927,0.8911583202863639,0.5321689761257545,
 0.4521572042481693,0.7063135482550235,0.9706143894303371,0.8214424768029329,
 0.40568221680148986,0.803136409683107,0.3976629119533609,0.9682251710474814,
 0.357576532062561,0.2858878410513723,0.6022762913028478,0.465209495572554,
 0.698908650260743,0.10689999318335408,0.8568383126417379,0.6402329471451447,
 0.3363573731951176,0.7191224642156645,0.6728635590500882,0.5385330708396163,
 0.5174700510637477,0.9171002496125978,0.6652900410360701,0.03276933435923912,
 0.9628709520492048,0.33847651216733543,0.3783476714904105,0.5097793833708905,
 0.129232593331973,0.42493567934223475,0.10006264886781258,0.341000793161084,
 0.5044752970964756,0.8419047797712667,0.750062269893669,0.5776816686855573,
 0.8616888768456442,0.8261533709720145,0.008840567952544842,
 0.7083089490225182,0.7957444250537478,0.769114508059209,0.09638629793833231,
 0.11323169237935049,0.8578287793916533,0.8400200964590805,0.8390662049477038,
 0.6069643160975335,0.6477473941449172,0.8168923945158216,0.7859160371301936,
 0.44101501280978406,0.8193726163598036,0.9920523570518245,
 0.35230898929274834,0.17269419600507552,0.05419007499092765,
 0.027933366442171148,0.2959794532356733,0.7093326339582826,0.562338394653089,
 0.9695484091842548,0.7860844205447826,0.30867425593509235,0.9634055043465459,
 0.39946487455195645,0.7563671637796087,0.8805458815629608,0.1932322546219909,
 0.2800669621206826,0.20119531400935697,0.9262816006820833,0.9343413843929087,
 0.8281127078295503,0.6718812029081376,0.7654773044044687,0.9959909192638626,
 0.3485321361397138,0.21014094235702785,0.9664018678359836,
 0.13470520684060283,0.8728195729836985,0.22014207172597966,
 0.23714218323721625,0.17802444488113955,0.37631518826835075,
 0.3164223802263697,0.9878434653532578,0.8559868224356149,0.9873439417391355,
 0.03460574498090441,0.8506065864216521,0.6028018368190089,0.781574677080226,
 0.12186727742583492,0.6782115045871759,0.5045368019039748,0.4227345694251313,
 0.7450289328322379,0.820036112915465,0.17197811396117968,0.4279623241238123,
 0.08954476256576971,0.36600818603078344,0.8399532071237523,
 0.48256414776461265,0.9493276799863594,0.061475484810674885,
 0.3853128084302333,0.33865295117707583,0.5491509420511335,
 0.22682494110389984,0.3394365532571567,0.505964967217409,0.7796597064068634,
 0.8567493688414803,0.9326096776773872,0.4720728542694347,0.416438594965279,
 0.5557349704274588,0.722921818395224,0.06562673230294003,0.9428787501829891,
 0.7208567938421587,0.08334430007207194,0.4278119965826558,
 0.45432958652331856,0.8696563572338354,0.3101723511134383,
 0.18585979316181223,0.44384199104208955,0.7711624456304615,
 0.34564176095650634,0.707901119803902,0.8052667108121608,0.2679045309640681,
 0.3985824462473627,0.5876346305477778,0.8049900549034603,0.6960114019971843,
 0.46154462273439245,0.6790904979160066,0.9875361270468959,0.9886992215550445,
 0.42098355863428105,0.1750156681645283,0.9463409313565575,
 0.022187485214689384,0.5736096405576311,0.9260116739779803,
 0.0028398498781411208,0.331052017852404,0.608150270510601,0.8082868376556905,
 0.2668583448279557,0.3158849721144764,0.4481922057557597,0.1681102083734045,
 0.9673460485822224,0.8226950763709178,0.935034568286063,0.09890133603039919,
 0.4615192416179177,0.3966817380909372,0.5888794515217097,0.08301687278590408,
 0.6220173638862969,0.7558952245305257,0.5418617143138642,0.9785248811174414,
 0.4073530462023174,0.5278193128230605,0.7879006328847792,0.8020101023657303,
 0.2930300905590931,0.6993100486858953,0.361577179636132,0.10240641554667396,
 0.9003417710375677,0.2933201679304053,0.7176502771863765,
 0.039752782902268025,0.35346553106971335,0.4552037000885385,
 0.12381687489246407,0.9285203402882657,0.47766324747466304,0.523852743681395,
 0.4933431467162943,0.32330240491185536,0.23521836154041942,
 0.9602905803720345,0.641573499062025,0.4636436962602086,0.3430951134825365,
 0.4087622046451952,0.7768063387143368,0.08469539925057734,0.6903249119914692,
 0.8953768063366737,0.08165209068623014,0.39108489159442983,
 0.32688531043219504,0.15373180527450248,0.8100409645694643,
 0.6997986768601402,0.27820827259726977,0.4845737249426416,
 0.09946881112234285,0.24382505354331707,1.7270030311555828e-4,
 0.40719936573110715,0.8239303783196389,0.9207448603215692,
 0.07342852734361704,0.9825917299785543,0.2593265556878602,0.5626190161863319,
 0.658831200095534,0.7721553778940611,0.7350599949014676,0.41351614802043946,
 0.7488272853704536,0.626411066765416,0.5247143220820346,0.5469975999293405,
 0.30598803838268174,0.4011287786182042,0.04544882631978431,
 0.9308312464309414,0.10441460048364082,0.5766195305144203,0.4612918760331248,
 0.4919622231470422,0.563788767347892,0.19844984430188273,0.5695618937790056,
 0.3699619454244666,0.8837837475284659,0.018979505513274986,
 0.3742232249199223,0.9985515318445839,0.5046264402618966,0.00913825928792733,
 0.6907939805405896,0.8559416650733175,0.4437502009689289,0.3101674440558232,
 0.3151749991440099,0.346286773489923,0.04713060918794776,0.9678410566655354,
 0.2900953080464579,0.8385059719646104,0.7833060380923902,0.5186630076215342,
 0.34383294738718917,0.19155337216433221,0.055378380767981406,
 0.619649060415972,0.17816454695551565,0.9780436377962246,0.6027440223952825,
 0.5457689857474503,0.6162592992280751,0.7037740785184,0.15723376525500043,
 0.5270793781342087,0.9688569202835661,0.7396230259935415,0.8679901867277404,
 0.621965699429666,0.28609666399649936,0.685682976145967,0.9010471420749453,
 0.9207350711798697,0.9653691478988126,0.3663322888135745,0.46059649967277017,
 0.8458761707478609,0.8325395021782833,0.374424076133249,0.33314808133500806,
 0.5805616512607062,0.5670336822390399,0.1435544491630858,0.923876601995879,
 0.2801937346936867,0.03197262453613947,0.8342451673011826,0.5014246433018559,
 0.1558417973074192,0.22061208915265196,0.3061671418295797,0.5883763564166323,
 0.18705486975835817,0.03468361162944089,0.0847259974610608,
 0.39338777300022576,0.4394996824136401,0.5069090081944765,0.2322584964031864,
 0.5563968722040725,0.3530664917611962,0.9568982307596517,0.9159281659617923,
 0.7854907447397226,0.14666521845880298,0.11886912958877816,
 0.6367510061629001,0.32761267690172924,0.3451313577070775,0.393838429091692,
 0.4375859122358433,0.4725070237154544,0.13332572582065783,
 0.19954837083457466,0.486842752800383,0.31872849003166315,0.6547954597840148,
 0.13255334540443553,0.2533250099935851,0.793700583775019,0.0653812601592858,
 0.819888484547841,0.3829877213726789,0.42928689138514353,0.06762193072946299,
 0.4758967524130966,0.5021264679308283,0.453591895423735,0.7101313337413282,
 0.29568167074949026,0.15947910116993314,0.5379080114336873,
 0.5450266027761149,0.021101293029495594,0.20200478059671112,
 0.4236392799930737,0.1659556726671918,0.6498611611563384,0.23056451347124884,
 0.11698189987064445,0.7774461326710589,0.6494017873483919,0.569411220148205,
 0.9028542325775897,0.688830022893055,0.8252505021920755,0.9274131924678208,
 0.4408986392515095,0.18674069940580995,0.9405740404553946,0.7159048428087457,
 0.9798599643160052,0.4718893718117668,0.3485067775828684,0.2086173637820834,
 0.9582488726823833,0.925366946396364,0.4075534249675199,0.7720915083546458,
 0.3518036432602659,0.2791877744332022,0.8430229263281016,
 0.018496138491649106,0.05028930647973229,0.40718094104933233,
 0.35746031609857254,0.48024861151365716,0.3727477901879733,0.674819151500111,
 0.472648113294039,0.6685091736984593,0.1295799050641273,0.3608144124574487,
 0.8444526271166661,0.20624686426665717,0.9716920248944968,0.6767371540015948,
 0.2993447602748801,0.4415720842952554,0.38507126256804036,0.4105612139550898,
 0.5856240462329128,0.39259943284284216,0.2499190279348834,
 0.48744553118351286,0.09306094079444405,0.8877119330125294,
 0.9873889261612001,0.02042328009270089,0.7335002291759307,0.5667351899677391,
 0.10537830883070787,0.41773729756691624,0.5800314218807929,
 0.5131356174771777,0.7644170479838326,0.3718634172997548,0.24045436786444774,
 0.7652021382443641,0.6030555999589539,0.0076330919609675085,
 0.08063453401055443,0.4609368735976688,0.5258255170708076,0.6927487137857551,
 0.2524963475144931,0.2849461816506943,0.19344238288644222,0.2828273892853106,
 0.4454521003282236,0.6543198256740044,0.9299737789711341,0.7881079179005486,
 0.6954761488240727,0.538302925129301,0.5511564160991375,0.9517251172147635,
 0.2744742853381992,0.021682965885621464,0.9119231817662697,
 0.7534681748437677,0.7258525918331287,0.5980201513627053,0.8913775618934245,
 0.40759392378002657,0.9669214469782419,0.12812600778247585,
 0.3425745369876656,0.20467400845309458,0.060055692422981255,
 0.8096202574539533,0.7474118656756592,0.6522337917244568,0.9287605989845359,
 0.15636902117891627,0.28815388456474944,0.7633726828561589,
 0.5539152635057472,0.6127295266251401,0.21840993506257333,
 0.21751602160035577,0.9475492994170172,0.5800012469871413,0.829949118150643,
 0.20107935312745695,0.3095737982823088,0.7457216009049237,0.5931907006734736,
 0.6200458174956481,0.2079302791377805,0.040680835319933495,
 0.2017916480130677,0.6096278347055929,0.3614328638983243,
 0.020693404937207394,0.10979440593886836,0.8179074012378029,
 0.3046557711279142,0.6470650354900096,0.6605260457255312,0.955092222404317,
 0.08673227736106282,0.06430105090012139,0.6750852190298511,
 0.9895254073973105,0.5135336822833305,0.028081446581011416,
 0.6438392917203007,0.6683410568870177,0.30269002112868126,
 0.12010713371261927,0.031036375592136345,0.15152062788140608,
 0.7614704020842104,0.02726635730522009,0.26733320724614007,
 0.8819815414803178,0.07347482364479263,0.8987813970616862,0.4312941423563914,
 0.6227177853580906,0.061200276569893575,0.7260186191337525,
 0.7072480659133993,0.4871278666201838,0.997639093552047,0.6908391015930005,
 0.7335403178990019,0.3747688809469394,0.6528722170750221,0.6046286819553386,
 0.43595793377197267,0.4681569479364561,0.6492599030497157,0.8271101391797502,
 0.6498583212710664,0.46147483762521824,0.26817033926230593,
 0.13079195440609115,0.645955115663748,0.40886314346180286,
 0.09925676399251981,0.3662734750867317,0.9313236329373238,
 0.0063037853848756376,0.7155484273194919,0.2600309957230986,
 0.22964418720309643,0.47619927740444234,0.5750361610579889,
 0.19484265884482366,0.6212740254954914,0.832238851950817,0.7545457315585224,
 0.35752406367140155,0.5014187548240708,0.8545517641999261,0.1798240718613433,
 0.45938658598081616,0.4130511144718938,0.06591099483875051,
 0.7495193652157446,0.3297643338274837,0.9007733000571185,0.6816632447356366,
 0.2339728083882754,0.6882361907936521,0.5625502262299185,0.3832885538915146,
 0.8023788350987116,0.8951438766045915,0.42632904307242925,
 0.11162340273541282,0.43810083488286744,0.6370120046465003,
 0.7842744586972485,0.03387875300349086,0.8109404635802693,
 0.12206343635762784,0.2434783045323753,0.5708626433280498,
 0.023260643843922457,0.6330580920075287,0.8628053669241204,
 0.9400301225903045,0.5009635030532864,0.9286186680543562,0.664884892350054,
 0.04540995620871724,0.06510428727716289,0.5690822947895884,
 0.21420911449443125,0.5142769037182588,0.1811627346794349,0.7816647870477758,
 0.405412822356793,0.8832339007815749,0.01801825365656118,0.7098865157698276,
 0.031320728458420155,0.05924656140313167,0.6189557299078056,
 0.5619460082574493,0.9821274820045173,0.2707140458653524,0.4369798579781735,
 0.3166938926880144,0.7876626839879352,0.22295877476402504,0.8305213350696129,
 0.3432424991387889,0.7095455932902274,0.3227994100787078,0.15117372796979112,
 0.26557866891497994,0.5307818009572144,0.48154887092500276,
 0.6265029749747579,0.2360097977780704,0.8210668729926338,0.49964117768965477,
 0.8657875481622709,0.008519219419010682,0.3229896174869995,
 0.8712345531521504,0.0340360204360356,0.2663191251593815,
 0.014851376815746886,0.24507510064752625,0.9395274770091824,
 0.06577672496643783,0.3984974018865408,0.8688573561663333,0.4050648575571433,
 0.3419409825274694,0.5962653477376466,0.23364621477349823,0.4599547991745597,
 0.9344549435039646,0.5311266874314294,0.2649429261357683,0.1245920732166863,
 0.09812902445757443,0.7186707773468118,0.5100234979364445,
 0.04784104261340549,0.595451337223582,0.8216475173600702,0.7953885435896679,
 0.020378711698569085,0.7397119442364235,0.3732848227308756,
 0.6281274065428629,0.026148733592612605,0.7060573074183065,
 0.33383634314879673,0.6940203811111134,0.6590544288092507,
 0.42807573354301853,0.7019026189647255,0.020660505334620938,
 0.05311045299755546,0.061509135429394624,0.4334830479696371,
 0.559240722084543,0.4153948538118626,0.4794485862523874,0.46823477615311615,
 0.275418283798205,0.7641606375692345,0.8045074356992438,0.98306539449335,
 0.5114412544830359,0.55007539747395,0.8287397842335689,0.6968718693074681,
 0.9035330060903864,0.9040742576248425,0.037002169029129894,0.449504088511273,
 0.28058373422977056,0.18057206052749375,0.6465334864723843,
 0.2701647618561328,0.1993995265627896,0.3442943799219882,0.521266308508159,
 0.525264510293301,0.6272288992578567,0.8924479515506039,0.34595469672398316,
 0.47131860933952985,0.8345574056356901,0.3211281681426028,0.1694559191742968,
 0.4153768559425153,0.12434375228137373,0.892673206026606,0.49472411276650274,
 0.7410209152078705,0.27216817342157495,0.9588384592021062,
 0.18294714057852068,0.7328425757148375,0.21681288879170635,
 0.15030506816587363,0.38845013405384043,0.798448226375216,0.5767245433165955,
 0.32061137866750467,0.9373397347809866,0.7287848413087358,0.2875399995218365,
 0.2815596847394236,0.4767337692245185,0.21746029344765805,0.3778306318227309,
 0.9959149190116922,0.07375762853948542,0.28804607233387824,
 0.0823219156886732,0.9810858066943445,0.9697532015982312,0.33868301507052656,
 0.5885418957740078,0.5866588248437679,0.4445456615431188,0.5994141371411965,
 0.5469204932527887,0.6824571255517788,0.6415402420601902,0.38227469036216843,
 0.028054482365986178,0.9220876914659738,0.3087645630095327,
 0.2770376852712185,0.4434248455837061,0.2297069288752166,0.4596707968854279,
 0.7108095922772228,0.19128832010559382,0.6906525266029783,0.9505972825861595,
 0.2842226523679069,0.5970681701427386,0.7885293998822756,0.834854023282959,
 0.44219233100075395,0.16952304539546015,0.05271925745906103,
 0.5959800543157107,0.11634309163729684,0.2840249006403258,
 0.38515208190053174,0.765200095837468,0.09759548341289781,
 0.06493702371639154,0.4331877870061054,0.9429483409832174,0.8042442257680084,
 0.9424558004409499,0.6897919717046127,0.6235260162570312,0.9308054330999899,
 0.4464194615538085,0.23225413377256687,0.5584850526525511,0.2594005259471205,
 0.1723228600957173,0.9117129165035688,0.7073744191920529,0.07784585916363818,
 0.004915153438941888,0.17230205725804837,0.5566224755192134,
 0.8297521551379772,0.938252569362949,0.8690595174043516,0.19593213139589116,
 0.2150951115865618,0.2067181626523198,0.8525560773518559,0.7519634561471755,
 0.8243559607907359,0.9097543362857522,0.4139351430883398,0.4283164585169176,
 0.5057581338923101,0.13741304817350242,0.5166134715500781,
 0.05358837646418735,0.7775749570037555,0.18239273780674325,
 0.44060197613884766,0.9081479805228615,0.7114292522156396,0.8925290765104269,
 0.5682986304592107,0.6634516977742271,0.8589508753275835,0.5481708215970538,
 0.3051429411241787,0.3914174560043545,0.29911502483883856,
 0.06419003759350861,0.0730821024742232,0.4672897240050208,0.6096727140022171,
 0.09988789040221724,0.3783432701866082,0.6863679764691087,
 0.14816577229758932,0.7733635643066042,0.7796168547569962,
 0.18690564210247496,0.5202879897120193,0.18573779845153637,
 0.04273305119117521,0.17378414465821024,0.4781991279333264,
 0.4901276901875192,0.4823161295154086,0.13531201490516387,0.5743053795577079,
 0.6938003614221148,0.5573328464827925,0.3139735812262362,0.5383232993073035,
 0.4149484517045374,0.5752307723877866,0.9441695059563511,0.199994422204954,
 0.834694408604159,0.8346040288983976,0.5519871365562095,0.6162318612121522,
 0.718336028210991,0.4828580080330286,0.4281446563444581,0.4507185594427767,
 0.7511389871664724,0.4704330451139146,0.12607972223388175,0.4138997660413235,
 0.456449431872445,0.3749415543996035,0.733013486874436,0.39330849750076746,
 0.005897660870700205,0.9395210849353571,0.5092355593597314,0.093309687176924,
 0.879823633896428,0.6075374225319687,0.627066948437537,0.18598519936184976,
 0.4480246634181406,0.7263555562832144,0.08267903297314017,0.4956175072894242,
 0.6325999209970204,0.22241474404820094,0.14181409128105726,
 0.3802685907785853,0.011114241562674043,0.909629482301152,0.6489046790349594,
 0.29384001605860366,0.15710044787238908,0.6394316755385334,
 0.09793356965225775,0.12799771262810256,0.055130172070909644,
 0.4813047067397661,0.004878430818532231,0.8382442248336207,
 0.7859359141755573,0.49102845295263065,0.21029203292583143,
 0.8883591421517038,0.15221419501757838,0.8240570178645014,0.0622818284488702,
 0.08990503971487696,0.4277773319166578,0.5524790420049426,0.9988917080164995,
 0.12765907993363412,0.3640187693812411,0.908871973883781,0.8625108560481426,
 0.597441185931337,0.33421039652427487,0.7901361023439142,0.4484802353757932,
 0.15673197762616764,0.017837296012912063,0.853642026958818,
 0.8948457311670196,0.23802999830851745,0.5591749656984728,0.7328252316679047,
 0.5062710487477852,0.949650786173886,0.43714483196553156,0.3475320892183831,
 0.830635526631937,0.5662986862875474,0.240859992866862,0.6319975359914507,
 0.009976286651697475,0.46825266361587203,0.9400751865731182,
 0.40209219777282734,0.6939218395254068,0.38696007941702726,
 0.4126342574123416,0.14850268235661268,0.01608179268478782,
 0.2041578358864864,0.9356275359791137,0.12428739350155515,0.7022607116428707,
 0.5996204738787188,0.8305640090758101,0.05625577665169956,
 0.33918293457226234,0.6663674090741247,0.9194285642248677,0.3623446094408208,
 0.25122241391319955,0.415723040428173,0.7119544836172126,0.8832656890008208,
 0.8583557193773867,0.6291861916150773,0.9337313080823357,0.8517434923728879,
 0.36463207310419044,0.46078600270567494,0.17478976191132023,
 0.9317593099632105,0.9354488333588573,0.7838608421713743,0.8509476104808855,
 0.37215363935672585,0.9782773503108968,0.34699837308485826,
 0.2954580343334171,0.3434968933553313,0.9944421046017102,0.4793757808430642,
 0.5801938921285665,0.07784256800085498,0.3695410063626139,0.8222604343596577,
 0.7789050057291323,0.6110900307493365,0.14404733475779574,0.5574659912118637,
 0.8711854431610467,0.8480534261166295,0.005121239261098975,
 0.03378253775270956,0.5715982021048275,0.2775811135117925,0.858813132540148,
 0.42371751081352405,0.7759508726441283,0.9719461564340948,0.7610020682246295,
 0.9474351965365924,0.9017434001408675,0.05133583290220889,
 0.11407415588285963,0.7253876306796077,0.6329710792155631,0.6105643170593483,
 0.5608193889426765,0.8376614842945631,0.13537767611673257,0.7769813162682312,
 0.5748946221608366,0.1678882129726491,0.5962077159049635,0.6447007562039753,
 0.7710828049205393,0.35547209892227694,0.9640861498661794,0.6776706737992658,
 0.3921724071553816,0.04763226123341846,0.6446431266570667,
 0.19425361803186703,0.5705052715588442,0.5393733254916409,0.3302631530618507,
 0.45549145493008236,0.18107726222873688,0.13845438353532313,
 0.22503375747166443,0.04601186105760313,0.7176696598974062,
 0.7558826854451475,0.7421958463627754,0.1454529742613685,0.921821713097243,
 0.34119452772944847,0.4931815413735193,0.2706800093321038,
 0.16258427892726546,0.17890507905234998,0.8991010483539719,
 0.2347746431099378,0.7985054344152206,0.42474451430293436,0.7523235185777151,
 0.46353690131177405,0.8798839355844664,0.33997568740302553,
 0.8200679934789054,0.7948013565395681,0.8734551438876375,0.17176707706878536,
 0.07817366432223805,0.8250756047108281,0.5449623507765805,0.5793746097303015,
 0.25813258150540064,0.5095883161962458,0.49504303288015983,
 0.2136776936963518,0.7444487674216861,0.9861508197237884,0.966519675058876,
 0.2501911324688947,0.7167497367308339,0.7013498583088267,0.964165937813415,
 0.7306684502766242,0.46910018975782186,0.9662746240509903,0.960864242800024,
 0.3755418989832561,0.494599558617986,0.9103994113768845,0.3779773496921385,
 0.6100931513551863,0.9608974159005217,0.4073074308307333,0.24458811472547015,
 0.7257460685743117,0.05054869003321616,0.9324432151030275,0.6254384189528068,
 0.36647242243156,0.5257110108880796,0.9534208103827437,0.34871909957651515,
 0.54009462768991,0.21260669009245314,0.20857294296320117,
 0.009159212126683691,0.23382067990010136,0.5214823612785924,
 0.12364157799901465,0.3595012587890679,0.24717898725697418,
 0.7003643876146042,0.3182499096972028,0.5549904914375917,0.9445607460373749,
 0.10544273276777405,0.8862885796321047,0.7509721343106479,0.9663520763995552,
 0.23342349398714635,0.1395814251955826,0.6014070700745056,0.6944439388220026,
 0.565110462082642,0.5525528150231565,0.9586937790671024,0.2656190833766159,
 0.16963182880856909,0.5396799485210342,0.45970765479928377,
 0.9091689102754685,0.7629831307653455,0.30327072324924664,
 0.47977390968688716,0.044274625883055196,0.8305721824406032,
 0.984549666165457,0.37266394659657354,0.540136376788406,0.6554831310085131,
 0.22183566648283903,0.3309641657869752,0.8349713431513605,0.5423008490419259,
 0.6689681754835459,0.7069314753226548,3.6395818020462656e-4,
 0.775431704050208,0.7564363073762661,0.34952512321587315,0.4915592665628634,
 0.8925080480032971,0.9649985488813212,0.5908876730750812,0.22863239239218824,
 0.9137762631781938,0.365736708512719,0.38781634337124127,0.7538496068157756,
 0.9228047290600626,0.7831113156464173,0.32832056885040983,0.8603051555796424,
 0.018133687632116202,0.06575838691517155,0.08931112276726949,
 0.4644800665042552,0.7970014878059452,0.39839283379617785,0.8915963592716819,
 0.9221392700026482,0.4244632509299625,0.5937371957098259,0.955200602484583,
 0.9020519236274457,0.8219961993713847,0.06781871643207193,
 0.14970250892758208,0.8530943625758737,0.6079596964190767,0.9311357293822348,
 0.6353668764746896,0.25516488920028046,0.596214905580367,0.9505715359169423,
 0.6188200376718571,0.21183776908294893,0.40876989529031493,0.594774011975983,
 0.48225360798746286,0.7720861025437002,0.5766147117281917,0.5371276527388467,
 0.31406484930479217,0.3767335497683616,0.6522229863490983,
 0.37505807536139746,0.9469334763546036,0.281021029722878,0.2005033912768761,
 0.8414264842887731,0.3263399460263653,0.6550772666078929,0.1153635887981439,
 0.8430657848807077,0.826953845781274,0.04803840671036286,0.6612989991882263,
 0.7866344856630267,0.038397221669046355,0.5508165798457668,
 0.37613943217535384,0.9656072539993709,0.07985305805947651,
 0.8408436555923338,0.5479878786244515,0.9510834722063948,0.41799279283736934,
 0.8691495920564118,0.6925583992472742,0.8767726950948435,0.429834251714761,
 0.5334563292544254,0.5084180222992061,0.2843528145261065,0.29577600414554284,
 0.9254993099672182,0.540216034529876,2.1371032421391156e-4,
 0.4660998664316395,0.20489515611969833,0.9206956435578004,0.848645702513215,
 0.5967534529904777,0.9220271142063079,0.7056294046534448,0.9178972400398177,
 0.7407220502651156,0.43059760796716795,0.1805725663992498,0.1101235011131132,
 0.5812614230717317,0.9485792358400067,0.3268831540315027,0.15553159409349226,
 0.9969259308604147,0.5648341696326211,0.6012852961773767,0.8536266156087162,
 0.20189030013771958,0.7298481889386143,0.6803802949174985,
 0.013442871245773369,0.02117282354638461,0.6631398609470768,
 0.934669203033992,0.4155656173946878,0.7143524777957846,0.06968887461072715,
 0.3547509994225442,0.2830450410632468,0.9107227353846363,0.6768719658816147,
 0.4049289354899399,0.12408240475583676,0.6001681173003344,
 0.16769340735109006,0.4035655830252194,0.9428639376398344,
 0.05929351057577281,0.5930380669642028,0.08693744292133432,
 0.7585870033329134,0.5545676259419431,0.9613414886636893,0.5049322135533558,
 0.8681766113430055,0.20287270890654352,0.6044203443425812,0.6882298122117715,
 0.29388314323311715,0.8208392319049427,0.9765478298102774,0.3178825610238116,
 0.053426776656865016,0.5761774271467003,0.28695038734200995,
 0.6136853834817092,0.17362894377242866,0.32448300728694013,
 0.23125422208482083,0.3025370538169825,0.7880726998860741,0.8016724680478409,
 0.2075283410403549,0.21493457735440957,0.35488663833081935,
 0.3987794373299435,0.7646292450758689,0.5352030595659519,0.859256492152778,
 0.3560101191564079,0.6793461988312851,0.8508388401558,0.6301573467046324,
 0.33874258455774986,0.9548894458795011,0.15168538030824041,
 0.5713222027920797,0.2633488159318008,0.7594423019156943,0.4316116317522083,
 0.5589551371752628,0.8216504028873378,0.41595292001258466,
 0.17878626867675573,0.09540949037754198,0.440395441554315,0.768021624429711,
 0.15832733119157094,0.43259027825779717,0.7315781528843227,0.719017916945575,
 0.18155517750582184,0.2696236295134813,0.17808586748237965,
 0.08191635834292776,0.872767412546072,0.612528261444344,0.9002390578832672,
 0.7142293823430349,0.3371533185432558,0.0613042075529191,0.9699302458812732,
 0.8355620437888324,0.8119243511213075,0.2862956603286493,0.9152122411109077,
 0.5139699154249908,0.9528331928496862,0.7525546717946745,0.40746803998301306,
 0.22227137348398351,0.2744157844810573,0.8396561059055783,0.3283847429629314,
 0.39769071018261126,0.8603335286227138,0.2577759688244048,0.6314925360621179,
 0.47018176638072173,0.9329138830106,0.2020384674638459,0.4046808280932638,
 0.2822268759121176,0.06253275983683437,0.7727266831878586,
 0.13364652951889067,0.8265686397176603,0.6571537166706676,
 0.10643954812155487,0.268900864862722,0.3448588760519653,0.9483041355368758,
 0.30600762464421305,0.04807145634942567,0.7362813156849566,
 0.5768182269941847,0.5562726144029191,0.5030760001752423,0.8711598914881743,
 0.7361787312185827,0.8756334397528334,0.3493659589105611,0.39889973993539707,
 0.8434296011383613,0.8485497889416536,0.2972838916086732,0.07503859420191594,
 0.8479051807758392,0.8923586192370979,0.49506086185854237,0.6994237746455617,
 0.8108897824235586,0.9892732074819428,0.5210649474173,0.6353316382634953,
 0.9266093788872601,0.18751296216017488,0.6731804226236149,0.6494153644512267,
 0.6287086323121822,0.5350321992092409,0.9618908620303668,0.9021168630499884,
 0.41589288428666427,0.31043393679871356,0.6492214658649381,
 0.30383680123391843,0.7154163224448473,0.34789448412665425,
 0.7975352800189798,0.9911974270843323,0.7033058385713256,0.9240200649576796,
 0.6134026193730209,0.943966827888169,0.7214046268850123,0.5715680959555773,
 0.343093622127976,0.8560037242303177,0.27225333782268524,0.2898583152563171,
 0.8309001872089843,0.012648239761871238,0.3333357780221774,
 0.20397778932840516,0.029879061703854792,0.41613807551587734,
 0.9294557305877861,0.39899648911173646,0.5583552484890524,
 0.05914130850810673,0.29789005605134367,0.5049994553572461,
 0.22005440217500793,0.2921148165284352,0.6727414409491813,0.3642032781304372,
 0.4793721532018189,0.6771949577785686,0.5042048787438856,0.34239765322732985,
 0.8294494878069579,0.9001585236366851,0.1259237224213885,0.23183284542169347,
 0.6390209266045174,0.9512980567174872,0.4982363233067678,0.24313126214935932,
 0.061352034929696764,0.852735382473782,0.5380850649120211,0.3829233956149538,
 0.19398912929763135,0.5417955425001572,0.12737327836742263,
 0.8857174386688148,0.38268900071302947,0.6549890187243682,0.6028252957136736,
 0.8914747900474631,0.27487712640636786,0.016595639923578265,
 0.6472197218591217,0.27541630928539806,0.2878986125165688,
 0.43467244315172526,0.06811091738332498,0.08134670467533256,
 0.8923647572796061,0.5003761796089979,0.9613515185361241,0.2722765726513311,
 0.9799897611323937,0.8479645066953766,0.8903757425177808,0.8325121763761041,
 0.6820606249928899,0.417857096485703,0.9738018932199197,0.7585442559792608,
 0.017220000893614307,0.10135199082935253,0.11792197984712605,
 0.02985601220787648,0.8111507914431726,0.2251489243229884,0.7870279757714027,
 0.6808227331356971,0.5725835066792306,0.958804715362776,0.5155533333120813,
 0.5192358551305436,0.5332616425132556,0.327644110750552,0.3284138953817948,
 0.16098353687779676,0.9184102018973245,0.05631529661614798,
 0.4055762882919737,0.10215230505786543,0.72098961545846,0.5001694943170341,
 0.4142406344489351,0.2824390316871289,0.9321131526019684,0.9431442027524546,
 0.9680239219091473,0.19389202251476212,0.24264215905857456,
 0.4222175754505052,0.1583083992118205,0.5731534859940997,0.6700235471492563,
 0.2812260451870475,0.2509430452561413,0.30674334436497996,0.9419456986343079,
 0.5669563729183714,0.030745677502500013,0.4085339560432967,
 0.6590817414695306,0.91442914093692,0.22685782479811178,0.8497483483528133,
 0.4988677133137377,0.9634665096475732,0.8829166880370085,0.3088821489679814,
 0.1304295633174678,0.9880539104984166,0.27214058228759885,0.6749772023228634,
 0.6163252010117373,0.841014083762544,0.4852764758546706,0.17862068798381392,
 0.7681934160061278,0.6593957875571863,0.12770670441777454,
 0.17284149087883094,0.6140816939827664,0.25731733314348637,
 0.9576735036625523,0.39129057676180734,0.48011122072891066,
 0.11002557490684761,0.4634392929785831,0.23115388105650725,
 0.032504612633524776,0.5068677148834216,0.5554032639709139,
 0.25316221292994623,0.4508167907468932,0.056564111367785386,
 0.6554054653024204,0.27024714648345416,0.3615601615240339,
 0.01835920989990003,0.593066126055582,0.9074337497569929,0.27232108875295546,
 0.9413650565433418,0.6200697045192899,0.22175842148929126,
 0.15629457781720468,0.8620488561603052,0.06861234323966325,
 0.7396271914105376,0.9743011982928914,0.4104825129874089,0.46592068169763323,
 0.23641489736294352,0.7944358533593974,0.8004803004976007,0.8053663129325649,
 0.4304537519428433,0.33777795320635007,0.4202747073095422,0.4316785711235164,
 0.3512171604099319,0.36401142214237503,0.7413681104139035,0.2746682300052484,
 0.8943785626939895,0.24257194732217935,0.34144698288980424,
 0.7790047570677319,0.573233360621754,0.12530608690291656,0.03217273558822653,
 0.1088446766142126,0.0358980798753511,0.2515204318841222,0.2714749641005414,
 0.37286247698805997,0.7900813660324106,0.7734145310533131,0.7507779146715714,
 0.6698455679395848,0.1321731017894413,0.5889420252185018,0.47311521100577647,
 0.552126289672958,0.07936529446936758,0.9511679483530787,0.7106054041408497,
 0.5944458158081154,0.012712205356869455,0.7883523082617947,
 0.3630944945196488,0.34470004850702884,0.41276381150115105,
 0.013459147332343546,0.09865769611429565,0.42941111805395915,
 0.547247969634376,0.6502034146814977,0.4164673526514183,0.9666183627789854,
 0.42151809039233434,0.1362036488086521,0.07510067574988133,
 0.5738705255624298,0.5976036756446571,0.6661180524332485,0.8862075842057353,
 0.2978327566862431,0.39878536589489677,0.6706698343839977,0.7655606894579778,
 0.9335805453438542,0.30647700049679805,0.4106459667396102,
 0.31467042761103325,0.2399756318433438,0.736282484907735,0.8509558108730385,
 0.7647352355539818,0.12527596785209227,0.3099555047845365,
 0.021417953624334296,0.09692786165084155,0.022728840542509232,
 0.9183765882983986,0.5903776515920771,0.2975593406845336,0.6292146477036459,
 0.36925605615570234,0.01056761204476675,0.6649083159020819,
 0.8013816418937384,0.9272432713878296,0.8019870038868157,0.15954168500240162,
 0.9668629042113441,0.6063655581398607,0.6756855455130772,
 0.029415835723158534,0.2678643874065877,0.899005341353073,
 0.45761223125171224,0.092006579916045,0.12312408193473445,0.7203138230113925,
 0.9979337658511874,0.41093452148057785,0.16896316057451966,
 0.24955003863296443,0.3955934602782515,0.7554654842444145,0.2701407574916964,
 0.7472573334416814,0.5735837479140181,0.3714725934622455,0.5697926180737585,
 0.9916948056912189,0.9824326593666725,0.7782406089922597,0.1609201390982269,
 0.20089918090902625,0.05732050308349901,0.6038537633839232,
 0.06283200304085501,0.35679992077985156,0.5003135733045417,
 0.7226379853715161,0.42417533082298453,0.4920820939853231,
 0.22774582700082568,0.23464522110381358,0.969660893150421,
 0.44593343164017907,0.42883011418402006,0.40878296749042264,
 0.5112646703464183,0.6258683089350849,0.4294145552907255,0.5100787428086535,
 0.07104449260534307,0.14163434826875876,0.8866051961963206,
 0.5147041265837641,0.7413777099666601,0.6657370048800908,0.5208700471536404,
 0.5532507564452348,0.35372632039108154,0.49345737413660473,
 0.7846626829556034,0.3201972720079138,0.3371708183058052,0.7080781596509996,
 0.0281668464614534,0.74919961242876,0.2728292762947073,0.7091934369116671,
 0.5857267230810429,0.5300089339831238,0.5976729419234141,0.685804914539851,
 0.47303064973028386,0.48616695711819247,0.7939874703317363,
 0.2569499656132114,0.27260276673890316,0.5839330328452238,
 0.09422847636682197,0.9351420721497653,0.04592842161644384,
 0.21397910709395396,0.08342483997337258,0.6560245575830818,
 0.2598598438581369,0.33954627711504437,0.2029553284027893,
 0.27317013629761056,0.8696776226529787,0.8808744096322632,0.962615036951076,
 0.43799880665855717,0.7727760591530495,0.8138788664736735,
 0.36967866071182987,0.834628863478219,0.9264644383025776,0.8056338884081679,
 0.3852377688482903,0.023700114643628867,0.03752640843813837,
 0.06361294769382808,0.14679288052175476,0.5160576195811964,
 0.10215462301788192,0.3136941536275689,0.19039360549511464,
 0.008664888287339356,0.3276538144052987,0.14334726075383097,
 0.5457977876025815,0.7396664208334313,0.6688986531638619,0.7549186673041417,
 0.8689775865891615,0.7816828108761309,0.47151464233494833,
 0.15597011280469086,0.8931310216284615,0.2285990347154987,0.3156832106457639,
 0.8800149585819078,0.10690891200647368,0.4187757651868741,
 0.05933413999440351,0.6878394996725756,0.0028685146969280595,
 0.4200526589800171,0.3682954255966858,0.30220933181365217,0.8527131034056487,
 0.7466058653517598,0.028060585993648246,0.6025785660802752,
 0.6020079162319938,0.7310265210465183,0.004670972476626178,
 0.5084454468081432,0.9378969853759523,0.9546544400003804,0.38397922672633744,
 0.6135170459137416,0.10076814446926863,0.7557100713843264,0.9189485885248525,
 0.8777702065660471,0.7807766560424442,0.6202687912378909,0.19063973318509364,
 0.28304682271273907,0.3490929388311046,0.23709228140896899,
 0.018612496829471747,0.2141551824996204,0.016764094892838366,
 0.010002237697905092,0.7243723645866966,0.359464299624056,
 0.27898990222442666,0.14610144761165755,0.9767161773221739,
 0.2448687749627242,0.9377129435722563,0.40770609682726966,0.7011866103872335,
 0.8255139537712702,0.5262471173869734,0.6214578692163231,0.7446252853819192,
 0.15403763803751358,0.13639325251107948,0.6527838614747035,
 0.8392312793450591,0.5858932521354321,0.9468374521725782,0.8738399731290912,
 0.4383602865773597,0.2569201017032108,0.4224957204096963,0.2105006838220801,
 0.5572478832995738,0.6298148933440653,0.2718086169250862,0.6116797175310665,
 0.1634380131924248,0.9045412284383579,0.43538441172850884,0.5655289665780157,
 0.38978174571648727,0.9624233494792457,0.5099966226777869,
 0.048501269811126235,0.06215100199813628,0.9343215214101577,
 0.24557642360691445,0.14504859389055025,0.26548985386505464,
 0.11476113906900887,0.02051116411506604,0.7747336421342709,
 0.9254371562312524,0.9079130967314731,0.7029245745310486,0.49831924953424567,
 0.2685939084990302,0.7130019744465881,0.6360010584111447,0.08493600842122562,
 0.5227683126162378,0.8405490748198772,0.26019690805220574,0.0909947630524146,
 0.3328559767407807,0.32543441337232426,0.8131453177788669,0.415392423701647,
 0.9329351281245748,0.29518720375598795,0.27442371243839936,
 0.6961969516089599,0.9580280637764078,0.4932917937786274,0.38422667722205983,
 0.2914062386385954,0.015645387204862637,0.3173771889664633,
 0.2917050681392521,0.19863328208073494,0.07560209676827256,
 0.9469487673309135,0.615849155209102,0.22928987928825761,0.6663789377220012,
 0.4312869784617923,0.044864539803144865,0.41420925363620664,
 0.1900407540466832,0.7088385048380379,0.9941836667674044,0.6681999881455213,
 0.7206583442940008,0.8318827693871264,0.7201976927097096,0.3442357704386019,
 0.17359078828184937,0.6773397373180081,0.6131409414515956,
 0.19314001509630074,0.977169769060281,0.898243851010919,0.5072823450336408,
 0.8522076691349096,0.5476646503558368,0.46511288000628415,
 0.16376206985014452,0.42087008797276493,0.5641105489324285,
 0.059021063227013504,0.6936565457926047,0.8018584265406314,
 0.8137566984889111,0.6297171943698245,0.43536550685289777,
 0.0038152784938321194,0.9208004591136489,0.6727516442181855,
 0.9154719636172599,0.20888192352292256,0.4780045302243511,
 0.022682883653965202,0.35699396868428246,0.1246547189619076,
 0.8395908891740825,0.18917897697807606,0.9378916124834408,0.8033953642477336,
 0.30525334160132733,0.8844789591544205,0.1520534998864599,
 0.04852224761737567,0.2872909034653548,0.49373005853201557,
 0.5973581782383398,0.8099372533285248,0.3659964359770962,0.07589511105172375,
 0.13197272426085438,0.4404784489635033,0.07646048293482921,
 0.6089494159408783,0.1972656619369756,0.04445574114362605,
 0.10543938778056627,0.42548171208799757,0.823801343638321,0.2971011888889882,
 0.8433805150520908,0.3502310698231734,0.8533925641769278,0.7452272037904533,
 0.25677809171803245,0.4498696206161388,0.281085308654898,0.7856559153757391,
 0.6926980651598789,0.2510091494068809,0.1918450112476886,0.2503195955107972,
 0.16190052173863934,0.5381774133328132,0.19225715130067067,
 0.06698226401063134,0.6779213354632082,0.4369205119093784,0.5951648628447204,
 0.8152160279297466,0.4724299356858598,0.5292705079333877,0.5398095041279142,
 0.08267910556934699,0.1578369351863702,0.3638462282402225,0.1580656535182272,
 0.5195609905308219,0.688694918614241,0.5564562549152827,0.5596624429254913,
 0.3177890559647326,0.30514447283859625,0.21187329833543522,
 0.8463236015502411,0.8376217344586939,0.5685227630932028,0.9963281304657797,
 0.5217390235837867,0.03041303451739441,0.5278096103877553,
 0.43088856364347217,0.1219370985400765,0.6653311668864167,0.625884484733436,
 0.24433638528169155,0.3360050653093687,0.2034001209229107,0.5777118068226532,
 0.7361648163687604,0.3870016919278383,0.9347356289227603,0.5417005563354399,
 0.8307661798895464,0.48790229914950245,0.5302002506035977,0.5919402038440316,
 0.040850210359073214,0.002043880603551873,0.24083121114961437,
 0.4246804127672821,0.21665424910571218,0.05992263418480148,
 0.1548323739328632,0.43242550411027025,0.5459408428463619,0.4184065672756244,
 0.101651066833362,0.6320576663121843,0.08048536084967362,0.5274451127546157,
 0.5290947293788739,0.9868920393977533,0.3296199191872011,0.3268429821635209,
 0.9775001680208124,0.5966295018023182,0.20625956009731838,0.4452510852961524,
 0.22913064529882377,0.727233784336309,0.5060414491520961,0.14759053688269996,
 0.5452723027586923,0.026366885193450873,0.03472866235865457,
 0.3903677985191104,0.8646129134899133,0.9927234775541792,0.3326920365238897,
 0.1507890269236053,0.7948375646626158,0.0014763454315057656,
 0.27611218097088697,0.1511205535494209,0.9680972648848372,
 0.020981154393772794,0.3146787928041427,0.832571162528885,0.8830096934518158,
 0.2151780359317318,0.6053812670232981,0.6215036553689403,0.6831280606925603,
 0.18229253882766594,0.9658678064537837,0.6739599794125015,0.9878883740034172,
 0.21549324904291445,0.4999146295256429,0.5501099429591032,
 0.37305548000616406,0.4497381224135458,0.9103736759508332,0.4381782942895598,
 0.783368004676907,0.8869602453830773,0.2715462641418219,0.13400922245215519,
 0.44179153404692983,0.8873083376611972,0.17302344000761194,
 0.2957653055489261,0.5455192138740965,0.7951888464837171,0.8996883937770744,
 0.396964947716405,0.421523182877638,0.47922392003540004,0.9413846794208103,
 0.3466191942251562,0.3753360910604462,0.12769122431474855,0.8463904109039972,
 0.23874791383743788,0.9291688839510237,0.7161462611383245,0.878660009963817,
 0.052559595749274046,0.6310109523737496,0.8680980887426887,
 0.7560906122157262,0.6428917833876646,0.2438270297696724,0.4907224387376309,
 0.10985890037134283,0.0706186893233034,0.8925360118953181,
 0.022831252860435747,0.049928751493560464,0.05822877891091105,
 0.9631285657392691,0.4333124361667824,0.48699808689717106,
 0.21787456912733671,0.45790912203067,0.5632248263024403,
 0.0012446839659302489,0.3849415115036454,0.2846926354518702,
 0.11908003850889526,0.14721010505179188,0.5096359195522518,
 0.9286736680353687,0.23057010471122852,0.3997111700271413,
 0.11488411714381308,0.9317938970582642,0.3080833032348985,0.8290569164096374,
 0.9398009301591477,0.5984961677373647,0.8633622592590479,0.403493902876755,
 0.49680247848918613,0.7929191552554034,0.6276876905670655,0.6084335597160797,
 0.7132972224427065,0.7207774959528845,0.724874172678564,0.6392369462240748,
 0.6778284235591056,0.7901612405347952,0.04782401752142795,
 0.36050082419486296,0.4954999667169776,0.22538959339409625,
 0.9532179505796569,0.31158824263268714,0.2993110302036275,
 0.37422634286019285,0.07046241814179544,0.46295213704396976,
 0.7091916647876048,0.09711459607286488,0.6953125968516201,
 0.07361933552245592,0.0014461593960160002,0.5105826497451732,
 0.5365743504456921,0.7303803552763417,0.7831503601150055,0.11775667672251677,
 0.8381821188804705,0.1399364771034861,0.5309625428466983,0.7649758264382513,
 0.6984487906446101,0.12218519254129234,0.7008894119745114,0.850658083944569,
 0.588227106919091,0.5369451698519911,0.5495076601443866,0.9301085780459807,
 0.027014220899212127,0.996194710963705,0.4258946944768587,
 0.39685774836904397,0.41266542645399884,0.6878997564840351,
 0.5174575826231682,0.9167705344700219,0.8720677301959212,0.04508829064598019,
 0.4613193155341647,0.3547050866427899,0.07489152940913557,
 0.07978217930813003,0.6695623748849338,0.4838079030048479,0.5608640642127994,
 0.8398807374430604,0.21577292979221963,0.7231905721425806,0.756207484143905,
 0.6343222421226191,0.3116225287820863,0.24980262966029998,
 0.14682002222400448,0.5915903899047749,0.013605978206691294,
 0.29527234971839167,0.31009801349582033,0.10095441789748971,
 0.9439125448793966,0.9300620368941368,0.6601460791151736,0.7265465234338422,
 0.0032374254821204707,0.0012614723126657124,0.5393053332593201,
 0.2488478732707674,0.059067803445164935,0.034009881602510506,
 0.14783698619718055,0.7078971734738218,0.0964722583135813,0.685400600779946,
 0.14362444351098202,0.09818507355728423,0.059183656792638484,
 0.014927740975741122,0.12587781926528518,0.8881448369677405,
 0.7393655876811102,0.097252676228343,0.33155381238822546,0.665146576525573,
 0.7281985273025813,0.8349698892821724,0.499406282985686,0.15110693721273682,
 0.20630875236730495,0.7218830893672978,0.15652525870996958,
 0.3009461873018837,0.11989635566675694,0.24528138625514273,
 0.9652569947546287,0.686565356208152,0.706916511795073,0.42329756733700075,
 0.957627628390425,0.6677313354809343,0.811574689692736,0.6265509805621559,
 0.6826812056058064,0.5757114531700349,0.03545809841108205,
 0.30943776676514034,0.8602365393842156,0.835918759315659,0.8417157791129067,
 0.7124171407482704,0.33311470981611746,0.8402965690379366,
 0.24592337676706866,0.32753858356907406,0.4297869024116545,
 0.09207244357991451,0.5104400028327767,0.14769665694104694,
 0.8029118984703758,0.9835465494026716,0.5421652982534252,0.8013784789411358,
 0.5565707949796697,0.6769801385983407,0.9359853948744161,0.7937425553410453,
 0.6705615838129032,0.7441232010064297,0.6118847949143714,0.9346252178547866,
 0.2411425599489151,0.1049535937503917,0.05043712758997554,0.5128370738447359,
 0.2655777335260935,0.35093745516206765,0.1345347161443381,
 0.03127890435451952,0.2657810650293686,0.8744648283310258,0.543057952799871,
 0.4316936392282804,0.3414445698150701,0.031062035051327208,
 0.5723046191913708,0.7398482820522758,0.3445523547938334,
 0.033663168326947046,0.8113391716006935,0.37844457710033397,
 0.33384142266017347,0.5512323835693742,0.19078308313110737,
 0.23720296116596606,0.19644101675425218,0.03329666686264232,
 0.5846786628106375,0.09175900944278081,0.36706279445646717,
 0.6600917973438579,0.5444881519232294,0.6077335275790646,0.5059343287433885,
 0.4419805509894186,0.054542880538675065,0.18628070678925723,
 0.9220848016975876,0.031402857045892185,0.09084944109469018,
 0.11851527625765701,0.13927947255367001,0.24009434266879737,
 0.5917391618132607,0.5398550273072081,0.46606760626378796,0.5039119686625597,
 0.5102754768998701,0.22480771136848854,0.3622609368500842,
 0.05271086749421572,0.8130270210708653,0.7937734747551926,0.589489117121886,
 0.8262548766862421,0.01948390116458476,0.2597150006326847,0.3047019769545076,
 0.08702527473864818,0.3720169790735631,0.6956355610828271,0.627313506204844,
 0.34589383473658974,0.10305194792262884,0.6287054610895733,
 0.6823857031429128,0.9743340488609032,0.8958788640092721,0.2942825885413509,
 0.527682530004615,0.1327028169752672,0.7342553901977286,0.9021854531626307,
 0.4797796069455953,0.5369636682689836,0.6368624510127902,0.4052709497834165,
 0.25151914144548315,0.1686232823841486,0.9918956560609848,0.5639686444253895,
 0.9687851810509208,0.8586113227026794,0.20992413089033068,
 0.24144632059924787,0.540460549050553,0.27283069220102085,0.8272997652275631,
 0.18253485429612903,0.01164283129348953,0.3930242251170817,
 0.7092521656473179,0.2371859837607856,0.3066051728693687,0.3778676201627561,
 0.8238344070260282,0.805419249631931,0.4705228001150136,0.8976022004324015,
 0.18804097168715284,0.2336754357500147,0.02979906448305769,
 0.3045706069041889,0.1809652930284813,0.5620655564995813,0.43940130038256575,
 0.30938044370882256,0.3729673584546034,0.7079213889722417,0.7205080522620444,
 0.20077166814088754,0.1924647851794149,0.353816338812434,0.542547001030414,
 0.8517929220653585,0.1310360476609288,0.07688291570392358,0.4805155547972968,
 0.6617727970220962,0.5285330355139869,0.7470460553423792,0.8835438873077311,
 0.2984632884573951,0.16944297890800353,0.842377484570443,0.2764301739717958,
 0.3858867364909818,0.3259342465495709,0.24444800518419885,
 0.19689129149231954,0.08555539815770796,0.598035561750565,
 0.19575921098859017,0.4619484316512936,0.07241820071305793,
 0.5442464467744463,0.9113425319802568,0.8058904073551403,0.5970062242655326,
 0.743827913570512,0.8965488716441936,0.33329194054499056,0.6645574586707113,
 0.7763970458206196,0.01010166118081135,0.7405938653329062,0.5333530733179879,
 0.8066617509080798,0.04083505542727184,0.6085540406250973,
 0.04486250660733293,0.9884568613769134,0.12339030160959641,
 0.39221205202493836,0.6860022126297249,0.8179354419903758,0.5954020494442024,
 0.7287200291486884,0.6110314993396597,0.7543136941023213,0.3879595040820911,
 0.7677638219145565,0.497733203719682,0.5778472057637629,0.5716039047318238,
 0.7962694016847841,0.2511311482479752,0.6326895607962904,0.09243436546463202,
 0.46772576573681335,0.0106578343871222,0.47159357997222096,
 0.19359881038696602,0.9495460618499346,0.5122095121069643,
 0.14820209268691054,0.23259163933254345,0.9442891074118144,
 0.4945017290469106,0.10773507812225203,0.2766518844372392,
 0.08922679620077034,0.09069529271453147,0.6405071052812509,
 0.28526109136557687,0.9511648147114602,0.8485095920622159,0.7032922099602137,
 0.9809413530540673,0.48008699651749775,0.5154284559692734,0.6016265491587551,
 0.6402162317868092,0.7363722710078646,0.07501019228907535,0.6859764212572598,
 0.9055536088370798,0.2889452875097842,0.1566446142032547,0.5431920800987262,
 0.903547775149998,0.72798236397659,0.07139050588399964,0.9029608257048178,
 0.710341997100905,0.5078421830086056,0.8514727809935793,0.2896268310709059,
 0.48014500260162185,0.3115142954406587,0.7547668904179707,0.5422432800242071,
 0.5636135830366531,0.9073138679709036,0.9742143113671252,0.22149040396393427,
 0.32046349742095503,0.5734783049252079,0.608448761787879,0.3039334328762535,
 0.4637904508318659,0.5499333287497665,0.5649275905523885,0.7724497514779248,
 0.6965676553201182,0.9740846386227096,0.6402016221851895,0.18403147440420453,
 0.2662126898602445,0.2483472777064074,0.5885901623866961,0.11152996484500377,
 0.09091115174785813,0.6005014254004084,0.7544359047129315,0.9119732125524321,
 0.5098231180018963,0.2532942523756587,0.280090645415076,0.3889675070763102,
 0.6453309676351285,0.876026342346922,0.9661078486455936,0.08671474259404588,
 0.8588843550524055,0.36132460055481896,0.22190776101066867,0.673960418934037,
 0.7442029244175068,0.5069203073268073,0.7954450686813878,0.9999174568602927,
 0.21888899258527372,0.22217934434355846,0.9588523840425435,
 0.1150148041940976,0.9715785424106107,0.5215888082844174,0.20785423132374026,
 0.4781331509642375,0.48410308016403936,0.9244260006379778,0.9449017508991999,
 0.21871965788832848,0.4041070255350594,0.49108076267984757,
 0.5038042469804959,0.7093128536619404,0.07567924527433656,0.1665581271166705,
 0.27868029904773095,0.6952202760836912,0.3043971552019257,0.2174777797018721,
 0.5300958141195871,0.11374934838040351,0.3150564218840298,0.6914759082066803,
 0.20639051685470577,0.77527412155443,0.7496970755534176,0.32694616563337375,
 0.07000271356802756,0.9477644596678368,0.5196951494402161,0.0446781148420734,
 0.5075551633889236,0.5052844456101221,0.23484098571255518,0.2612589599189745,
 0.5173902541261322,0.3977523616237937,0.9383279352428283,0.7758640584168778,
 0.05041449831798683,0.47714413411586265,0.8933823674381152,
 0.03732289942250344,0.26027041288826436,0.3886503056522238,
 0.011119515063569674,0.01689198154797711,0.659351640438931,
 0.6284482611789242,0.59798920163016,0.5918730998972943,0.0226882515382264,
 0.764913951520097,0.3051120291449443,0.19096459639722374,0.8474922047685736,
 0.4062401747916006,0.9469068695517804,0.8532310638377407,0.36150801241771735,
 0.44305673403320933,0.3231396349065009,0.3270358876149221,0.831203503108449,
 0.33583424320361166,0.9538715611209919,0.4747824024652417,0.9597978730328242,
 0.6704660745192732,0.09138117834849202,0.7940479997847902,0.2996688880665562,
 0.8066651685811901,0.2124923088965357,0.6404234616027995,0.9508121226071498,
 0.5864633713310374,0.5405059546439797,0.396078218594588,0.010040478273142805,
 0.4809945357943044,0.19901238622064665,0.8549351935869045,0.3315941341495655,
 0.7849946054350683,0.6830449705948933,0.9765418431984622,0.12855603375379032,
 0.4896360673889453,0.692028210646851,0.7701583086915691,0.5550907697530365,
 0.33931076307383357,0.2974002185172565,0.48738357944887045,
 0.9338388271683121,0.8225964711881879,0.24006788087627917,0.8248196758961452,
 0.45217233373697585,0.6716967457816725,0.06691696842993267,
 0.04372074065414289,0.9934538406382071,0.16073872174123183,
 0.9562912452543894,0.8953915738231606,0.8025447938682959,0.3232263123089395,
 0.35870130789644517,0.49389863989751925,0.2995427918761291,
 0.041829782364871115,0.5744385300443658,0.05975240274252824,
 0.8995111024294627,0.5359822174384659,0.7713422590319381,0.5272134629404632,
 0.7909417658488446,0.32150365482686794,0.03655275127777502,
 0.49254783587478956,0.0920536112032797,0.26201279041845793,
 0.8249032893976604,0.4587537055442008,0.22145037732839523,0.8630597628183256,
 0.4654286551036142,0.47587077350760254,0.8194419346756303,
 0.10745836717881396,0.4843950310966286,0.8373764269327362,0.4217941521070747,
 0.5350471428124003,0.5338839560274452,0.9277165124352347,0.302933825677311,
 0.9503444263604972,0.5946117839936763,0.45887392258482285,0.7611195891915887,
 0.8762558570625754,0.861459458415947,0.44244562598824877,
 0.006240625338268635,0.5617915368760701,0.1507313919763189,
 0.38154007929753064,0.06299720028549993,0.5812185476258419,
 0.6829937325370989,0.6944311389127155,0.07112814197252226,0.9808242266232942,
 0.7344362864119135,0.25699354593500623,0.9184280310808307,0.9812447411677403,
 0.1099692497492345,0.7762033712647521,0.9696812544434263,0.9712657084218821,
 0.8712729183849842,0.24975757555501565,0.15326570205125278,
 0.07930267666083413,0.3452157861072437,0.02481759843971365,
 0.40852023248527325,0.13979399584338803,0.06360410625900359,
 0.8266519767407102,0.47096050313003734,0.5263351965903524,
 0.03288071748089105,0.47330699865010195,0.7844195944058756,
 0.8013444703881955,0.2161572992724088,0.47217375860636435,0.4458276687967021,
 0.3311989883014309,0.9591163836206601,0.2599877413571724,0.5509931388179149,
 0.7093193912140892,0.8860035221818433,0.1316520683449417,0.7540591970174935,
 0.7022902806442896,0.8505132403624206,0.03520202483884438,0.4291286092220109,
 0.26454690808569925,0.6014255150578987,0.3795105611201782,
 0.45505397391197533,0.8680181995188477,0.04218453778891218,
 0.2806637770788548,0.2985624991853376,0.155835915431755,0.6798370640778633,
 0.5979884509341947,0.6377681184387044,0.72227220376586,0.22434449676715718,
 0.7969196771349877,0.32434375132292637,0.5066517193421607,0.7319347000454468,
 0.6083856010686839,0.8641490719412077,0.8973606460794544,0.5230757261477019,
 0.012062935117465923,0.06342200189452829,0.9984297916052806,
 0.7953914465751384,0.8403395456976065,0.3582844150164708,0.2861154160102879,
 0.6902338496651871,0.6760381239030036,0.6242216779379898,0.7480931896811464,
 0.40858895956444174,0.6887918701938498,0.8986378285777477,
 0.038876619306778326,0.6505187433939766,0.7593716601181559,
 0.5337980701008646,0.6267622355670364,0.024553134424046608,
 0.9792904714409801,0.9988054157619101,0.8443838687116715,0.6646556639369053,
 0.059180321825088145,0.5874745687635126,0.43051002237588354,
 0.6053740603887081,0.5212108917650304,0.820606655497894,0.2163504769709279,
 0.005911236244109497,0.7645577427948617,0.38384402804166873,
 0.9801214659980741,0.9032002885733482,0.6668351429869054,0.829684259419303,
 0.7970859689922999,0.8705603107536751,0.9509676646007342,0.2442038691107956,
 0.1007828619271326,0.7740792975067287,0.09672957706954677,
 0.32943701575483064,0.2620192176224152,0.08962756631218771,0.763169130559612,
 0.3750469294080617,0.6756573482035032,0.8301622625783917,0.7814026552247137,
 0.7340596292607436,0.006005214927703562,0.3920816104105129,
 0.5930181075957226,0.44919620071851907,0.27334183795259814,0.382083675681381,
 0.20955057767610397,0.13956834231321336,0.3200267009923602,
 0.7994181846457792,0.5327848296882834,0.5033404696701466,0.34560343607928334,
 0.8978087899745983,0.8891544324497714,0.29962753398870967,0.920423172243129,
 0.5301274850638014,0.8333364953996102,0.3429043990593408,0.3762499219492831,
 0.03455727802259434,0.2966429912344698,0.3297433727845299,0.9529225759543201,
 0.5128084705589645,0.5464292444746015,0.6088616115464683,0.6581625345222172,
 0.6848585864455297,0.9301059859222747,0.9550781465573991,0.6240116693558183,
 0.04627339723986901,0.5280588461849812,0.8489834159014557,
 0.08538695160128007,0.2988510813129015,0.323805048471446,0.28025188915285193,
 0.4498470415341601,0.7317502729437534,0.760116475013552,0.20203214817367954,
 0.6771595320298331,0.8808568033552544,0.142342120729126,0.010268008296603903,
 0.6461412796421456,0.6590620765920869,0.568501401415421,0.030872495390069532,
 0.294316194876034,0.9730152765728415,0.703917791919137,0.9219675917355231,
 0.8103410368712778,0.133278667349936,0.08408783189308389,0.5859564797107805,
 0.5552503775794215,0.38886301350883146,0.9446095071135598,0.9432421604660771,
 0.8839859091249722,0.8532035693577703,0.020042166664784422,
 0.3109613995852667,0.1737610893106757,0.29922720407427716,0.5212091849587162,
 0.15126746574133787,0.6735704369772553,0.8702167312806162,0.6029149145039812,
 0.17011666818370053,0.7690601680462148,0.5596821281364032,0.6161875383036112,
 0.4428322164395605,0.49720366342154976,0.2888114718058189,0.8638697302624646,
 0.9728314813038874,0.42717549849751635,0.5904558214251379,0.5067810529871181,
 0.8467347199294524,0.8609577036498436,0.9652215640687933,0.1142845137671824,
 0.9832804667709993,0.4975924875536838,0.6304006946358226,0.9869841189753896,
 0.22696315403002842,0.5147887362672777,0.5289489803782013,
 0.13859649560275578,0.8334713728393318,0.5356883108546261,0.7562324396368635,
 0.2686068843577535,0.9234283051061368,0.3444550705211058,0.9528351878350958,
 0.04863392774281261,0.3883312616960958,0.9618945682968415,0.5251686849026039,
 0.25410315811504847,0.6078655901881307,0.8458162366539782,
 0.48816645986704377,0.6376929618057282,0.9602386893816044,0.2777313407499693,
 0.5262818414198674,0.40840581740953275,0.9845112220108742,0.7808176600762129,
 0.7738697873229963,0.0931972274637769,0.9374035376295913,0.2820995872477515,
 0.2323289120693106,0.13129227037544633,0.2903824478844511,
 0.45492763508636025,0.2846134529036025,0.8273144220510641,
 0.31320188952318395,0.6559371618459537,0.19862429333470955,
 0.11955705514322901,0.7791934079134765,0.6627073802919112,0.9065095780703827,
 0.4982951058527292,0.14320697268123705,0.9289585515468364,0.2741425422553704,
 0.5939057509247807,0.1353202168734614,0.31885334131529497,
 0.10255107195945679,0.4944318720506051,0.35349841016350025,
 0.33155903478769533,0.496239603501474,0.8821823866281383,0.13889386608279897,
 0.16422014622970038,0.669900347820574,0.055942914572791214,
 0.9590051208875312,0.9327473405170301,0.40854571349827007,0.8333728667042345,
 0.22492793031742675,0.734422850783064,0.5868922645468174,0.9773396052551561,
 0.5756801389160929,0.3991318798142477,0.5619524621254445,0.2740000846902909,
 0.7666440847547811,0.6105055860786354,0.256393676578501,0.19525867096630134,
 0.06282621019012635,0.12819252186781116,0.7807223821801459,
 0.12081105892053146,0.4219655869722354,0.6985022766758074,0.9233128409769723,
 0.17963682063223108,0.09889690910409166,0.11781779705400763,
 0.7673257825945049,0.6379229494859637,0.6548531031618304,0.272600094665722,
 0.11315583091274073,0.6600143666444857,0.9471502815889357,0.5761289732908885,
 0.028859262195889523,0.7358323186816843,0.14863370790564434,
 0.40872197539640764,0.6855141904903894,0.6386884178664806,0.7878878649819854,
 0.5776790014633152,0.879478785793933,0.8152944098483227,0.1312938913743389,
 0.727217326224899,0.26802718321844776,0.8238130444183417,0.04090942436846556,
 0.0814699209195412,0.15443252352579295,0.09265182938805061,
 0.6539855169644663,0.3877687936485187,0.6920282039889577,0.5219556826461983,
 0.9736694646871663,0.4011514802895234,0.08461308554800728,0.6815293385361075,
 0.9758960590309511,0.7592257130135098,0.36783552039135947,0.2631036375167236,
 0.15620446511602704,0.6423515444241682,0.6058645019724173,
 0.19859959190489695,0.334700352952901,0.5884500468648376,0.9315261526630058,
 0.9389859811043119,0.08786838278493747,0.6978317547805604,0.5605923780636073,
 0.5133199425129731,0.8663783488361035,0.9053453974537647,0.03393605216678641,
 0.5845464282243157,0.2100735243251295,0.6620324333089127,
 0.037364559741691505,0.9931776972840942,0.330124885035609,0.5969825685130694,
 0.3995010882482525,0.6329398214441946,0.9340814462639395,0.2711415345248265,
 0.046527353558871054,0.031068140344237216,0.23087803894251346,
 0.9413315795394744,0.4728087714808644,0.833754422631319,0.8412323738926945,
 0.42959649527438193,0.7721690656681959,0.017144313917273224,
 0.7641308895433968,0.5040298437831192,0.2329670157647742,0.21877852544365395,
 0.3957094208378936,0.3441469964088064,0.3912848469876965,0.34162508703372485,
 0.5309092216494742,0.38023583029708063,0.5637438123276997,0.5027194611644367,
 0.40392528680673556,0.7819915458617896,0.9428841603481508,
 0.37934555507469014,0.4938106357592905,0.5545991373246735,
 0.21910651276762672,0.5339000645304672,0.23292046722442006,
 0.5400986791731163,0.8216835392474453,0.533398540941328,0.9474465803805003,
 0.05610992424433481,0.9291420541143994,0.7706007557523232,
 0.30593958575025415,0.9969635131075163,0.3276989970469658,0.8013665280117388,
 0.5652688865508241,0.09842388340247499,0.5158069971703286,0.6266238204470397,
 0.6565937693861753,0.7929518244868278,0.8138449838567064,0.49100048298733434,
 0.4619624401277743,0.3303494258861933,0.64185113948253,0.14674877445721024,
 0.3723820625565464,0.5201667470203135,0.8725726230823178,
 0.044546290085751505,0.8266110176993148,0.9253609370818882,
 0.5514992554095205,0.8035756855217229,0.49107364567479006,
 0.25979017450800845,0.18571280493991627,0.7855581881713011,
 0.15373085897732963,0.7509645579604406,0.19023396818183202,
 0.2509271371656694,0.11807420874337315,0.9591329512904292,0.8612954084526345,
 0.09167347238668355,0.9793744253594534,0.45288856849963155,
 0.3253822807711235,0.8691416369928114,0.18307333591666286,0.7158351202901166,
 0.7265119595431097,0.17655085429058248,0.7428910112900333,
 0.36206688267291376,0.20203708029816236,0.4744122901248927,
 0.3328989281840372,0.5956630207315081,0.4383806871362209,0.9162560189439085,
 0.4835494635452907,0.39604172051286035,0.13973801422328436,0.556552902190477,
 0.0546216128394752,0.2246208662432012,0.20275773154875543,
 0.24904081708406722,0.9581999093568733,0.35966287987442724,
 0.6910472437254068,0.16229197909803184,0.2947431217302464,0.2562104128207888,
 0.5775530330931125,0.9650930008668444,0.5262589475218897,0.5703196794374639,
 0.6706754433592754,0.2621282541570731,0.4895328649985732,0.4737877256878803,
 0.8579691544944543,0.6754773903799928,0.9988381605301908,0.11369553965483592,
 0.1224376294949282,0.16046286775166196,0.7538418742785212,0.54821375977637,
 0.9993894516111896,0.8018469313632541,0.08544102338894144,
 0.13290107145934327,0.9593802152734072,0.7500438506362468,
 0.05555199387669951,0.5302511863468173,0.9794091780323622,0.7948498608173745,
 0.4448205753388348,0.4406070777671538,0.8220822505923228,0.8055361732317966,
 0.6149594725497662,0.29941361357881346,0.8284578320782479,0.1870066556867953,
 0.48973506872319816,0.5240034288178738,0.5511790209636942,
 0.37243494899873664,0.9110689873862282,0.7016805863567659,0.3985972525899478,
 0.3145301494438145,0.5319153221320884,0.048010109466859796,
 0.5185012516255794,0.1952117394185019,0.37651206556784744,
 0.19759943215898357,0.8951760904746402,0.25692472762378216,
 0.9734504327067122,0.02702443606662419,0.42971167328578963,
 0.6116594161566713,0.1769027420344963,0.8534339430670878,0.3248420057534537,
 0.21684850712899273,0.07672317912084547,0.9097124390628714,
 0.4881625915700394,0.3266962664066184,0.8902792642812212,0.7961898072269176,
 0.0738450434174357,0.4506442222479735,0.39428096853768924,0.3429320800150515,
 0.22901156856244298,0.44585881072734257,0.8698811789980927,0.999071299593728,
 0.6212075038777835,0.6294502354616397,0.6583585785156484,0.242261700287714,
 0.5910659823006617,0.7821511497076189,0.7679545643561329,0.28308978366939574,
 0.039349934213220306,0.4773276691032139,0.18087587032958097,0.45255400725035,
 0.3228502429024034,0.5655030895760711,0.5692992963906479,0.3750133475041143,
 0.4442782506348163,0.3455572007124117,0.13697543916665855,0.7139521653122362,
 0.7459730428413784,0.025514975423285624,0.8029130415934242,
 0.25278662304339217,0.629921729544449,0.664265994941773,0.30694313836685794,
 0.7222448218153767,0.2100463716959392,0.8884997767270998,0.692495351762497,
 0.15870723762236738,0.14075348486978523,0.9644540984069253,
 0.6403712716647145,0.043142899712950156,0.9682981539880511,
 0.16957506091953622,0.8448212840797996,0.6960574913215919,0.9989178624969197,
 0.9667296550042941,0.781753454692686,0.8422304059852438,0.3446970201969657,
 0.7325044089846257,0.4872532572414584,0.25578219921167145,0.4425843966930013,
 0.1893389381681546,0.5289033670393081,0.9767459464395393,0.5807519339493188,
 0.294068998109149,0.17937393765115128,0.9109033458319452,0.646008329883133,
 0.9595048437215021,0.323064862099224,0.1522416723748662,0.36814678936107903,
 0.09734678610722947,0.6645877984003282,0.32659024226729816,
 0.6701429163345693,0.6837658801139239,0.929563969610413,0.7997817473801437,
 0.8388515220247657,0.681818752824972,0.984994667223605,0.9190077068095934,
 0.29127685904701317,0.2100123926866332,0.7935452091115008,0.7228506968405162,
 0.8726933922157094,0.5760880898008877,0.3612168797734363,0.0890660265940002,
 0.18538112535095708,0.6356757977745128,0.20208761307419532,
 0.6245508634425009,0.8873152851199311,0.6442681321572172,0.15452207754187663,
 0.681528880112997,0.5446459178537506,0.10315913122901132,0.5550561580142881,
 0.36110213829624993,0.7473711925944212,0.2843484794344934,0.8028406236713368,
 0.5248606798695918,0.28747635661556303,0.07173514228905487,
 0.9324562578472273,0.6324016646069592,0.17468248771716144,
 0.18545582369959845,0.5664795222964999,0.8370899289653018,0.441179514883108,
 0.15262512261878447,0.27151602599065083,0.4154597812066315,
 0.27443756571482747,0.3202238921865239,0.9092402519306385,0.1835851226522529,
 0.6230060972598126,0.8538273562641463,0.5728020825240817,0.22430829814078102,
 0.8961227164639576,0.20794723616069255,0.4813549823759832,0.5468390362288981,
 0.9789388313190751,0.8567388112020118,0.9398041609612217,0.6338415050440285,
 0.20471291053535756,0.009383711399709727,0.5509745481028959,
 0.04548945706020946,0.3187567303322736,0.71343157050084,0.9592106008033859,
 0.9069212736446205,0.4613092574361979,0.4279975942539095,0.10562642486632079,
 0.3404283873759272,0.287390388018686,0.33956308067910124,0.5304260843898287,
 0.27996578984358955,0.7455253680282589,0.57702908374189,0.5160770025986016,
 0.2869157381991737,0.029591850525051644,0.003918906724349691,
 0.5541367275747622,0.9985181263797179,0.38780147041514756,0.6499942109267943,
 0.8427734497809137,0.9440554745568023,0.8434199270674255,0.37080205236599384,
 0.11647427974575397,0.9323150113207701,0.3070938667312806,0.3660740724926267,
 0.10961172925181462,0.7780335075483427,0.8062630853682344,
 0.16492622971421944,0.21180205591443846,0.5386708821140285,
 0.01133748412280855,0.3783581646424502,0.5073087294779297,0.7218667068011322,
 0.10964933857280446,0.6087029723448093,0.3598483611799178,0.8471150452390414,
 0.44966173645124985,0.5281913809282108,0.7354914768849786,
 0.006813608067815946,0.7264060345136854,0.18284158925286054,
 0.14510895072688412,0.8949027204578014,0.4697678959260775,
 0.23485244065185618,0.6942132654415046,0.13648786692395165,
 0.4367814844799932,0.9675445822903329,0.8460232763716133,
 0.004501094381157755,0.3682437305677675,0.4927776389000392,
 0.10868590996779359,0.3321011598521859,0.15656575871401546,
 0.031697686456865215,0.8084526160911082,0.8046278731030441,
 0.023617869249627166,0.29636866317690436,0.43196405816744565,
 0.5133784101566968,0.31140585669716503,0.4665900503328554,
 0.47984772741845394,0.21647857176410956,0.08504862758665421,
 0.26192856332576375,0.8811178008099798,0.05468729253842031,
 0.9627699110385443,0.6639777086867735,0.6869983085540148,0.8738701163419661,
 0.38790292127038406,0.11006298125723502,0.16747472500760652,
 0.4906872942307229,0.9211746057907986,0.06319721385078259,0.8618682409233202,
 0.042816162601183705,0.3770568684296586,0.5657936505420436,
 0.1553912445732034,0.2749209808984122,0.08666389643224537,
 0.09153731988104441,0.7185513100193255,0.2645292409436256,0.8299304637550173,
 0.29699723422502966,0.12152290969842483,0.13352946794862341,
 0.29328940585724006,0.9612499628244042,0.2201864913039,0.6535732374926145,
 0.34636873113643707,0.980299519217527,0.26623361002903234,0.3720013853530578,
 0.7990961111346009,0.66618591266686,0.973233044276341,0.23153581078611318,
 0.849959445557563,0.21616123281311006,0.48099643018356686,0.9952569440764968,
 0.8418387635930025,0.2889644291233926,0.06758173243478849,0.628215735273384,
 0.6913879884572298,0.715969088265346,0.5142046754809415,0.6791455961268767,
 0.5523387050746571,0.6438502008905027,0.2266953737091324,0.4287462108935429,
 0.6992777263496901,0.8924146604380774,0.7460769067566532,0.4389680643279008,
 0.02766729414212843,0.14362790933182468,0.4750409527066555,
 0.2779629267272984,0.25181067954706327,0.4801369957284364,0.7740485323316878,
 0.7124513346309471,0.32814712192829243,0.8043173687679649,0.4753892260725088,
 0.17059590694080573,0.2578074102403024,0.46916334142839133,
 0.5131082086841912,0.5193876417265643,0.9351358342362602,0.6980520227682943,
 0.6313920270875237,0.7894984939610179,0.43724167837764827,0.3519992025911365,
 0.6123586150221592,0.8528728783975413,0.8117032565625281,0.8966538896063377,
 0.9450378638964941,0.8522457686022619,0.7461359316638296,0.05178948103560821,
 0.44226323253231115,0.9796902117674889,0.016464490457072656,
 0.26609138211854966,0.12639926701544446,0.13511152586307673,
 0.46084326381093377,0.6746549267304501,0.2032506904333502,0.9603572194211835,
 0.671289303169145,0.5486075915435423,0.3791343438250119,0.7676322048154441,
 0.2020269145000354,0.3388193832933558,0.017862330038592833,
 0.005016927282515349,0.21262595965350828,0.9066577978411678,
 0.18099348015388605,0.5045814080499388,0.6300082630279102,0.7974212938729319,
 0.23836052960200138,0.7148599870684407,0.9320169166058961,0.3055672110998575,
 0.4647556143401115,0.9424817095603744,0.3260798661862758,0.24913039224652456,
 0.6004681850298249,0.77725703240154,0.57201091268838,0.9661622600567104,
 0.39475783743523607,0.18583208651080918,0.13526883024498304,
 0.05413746939368125,0.7223453499108411,0.9275000618830884,0.664231515340424,
 0.18898569253931607,0.9019614792369812,0.22349841696573947,
 0.3649004188773961,0.4288733635889127,0.8235004748738967,0.3570202278050636,
 0.375037216244558,0.3116007726930101,0.14095817521231402,0.6174574933978866,
 0.20803604357478833,0.613317350352884,0.3721969620203367,0.42878441252103183,
 0.014959704912631988,0.8831983241254681,0.25290207738852666,
 0.02789064640236394,0.6762269738960611,0.6235642789976433,0.7772879948785832,
 0.4635454491729821,0.48140031067524025,0.028237109213192224,
 0.26420315760614177,0.24425430109993296,0.43266111241897864,
 0.18700658542676896,0.016575555143073184,0.44828830773326966,
 0.8010859482825339,0.8577286169805821,0.4980099584858019,0.10096472805060441,
 0.16039195105069282,0.9343110679359171,0.42441001582540194,
 0.6041757102281292,0.6502131616071636,0.9717063347304701,0.6732552408934683,
 0.4931461049287351,0.7306758250710481,0.3046413604824615,0.6470035478405873,
 0.6126562552409427,0.23236250481840182,0.02410291546017307,
 0.4806238555443355,0.5946128724255205,0.9745558022913694,0.9842850383286145,
 0.13874995361337417,0.21861957725979742,0.5682925588972007,
 0.1098049551325131,0.9169905604405973,0.7258778217245052,0.2206920703466626,
 0.6305316759016448,0.031400776313238454,0.667524532880664,
 0.04571851720242204,0.6794689978787205,0.2034545807379753,
 0.40059711477270654,0.1252580122252902,0.3160597135178681,0.6885352907103665,
 0.7380041881603152,0.8718291273892834,0.5570056159532137,0.7015102484425244,
 0.8424892530034839,0.7881753176816123,0.928576369675806,0.33114254386744746,
 0.209214213016381,0.8217370969865326,0.9006486560544275,0.3267993088505732,
 0.39153093760036106,0.8199377514564448,0.7153052940648899,
 0.37081455182986767,0.25439584760144396,0.9414509546948606,
 0.8604779338614893,0.4102735282248582,0.42846245668096317,0.9081859868198872,
 0.9833966067276263,0.6102602389879761,0.025935656540453045,
 0.5509512176628838,0.8047477789052586,0.8504833195965977,0.8584536094965629,
 0.9620793992585543,0.8780967171696636,0.35377927040361556,0.9016001635152068,
 0.3444255442944476,0.5232042135533777,0.6577936083070469,0.13335637481935847,
 0.6809570834621583,0.016550968459646875,0.47949732723059924,
 0.4170169179232974,0.25282056412788123,0.42804478157251125,
 0.4000214974047477,0.9393905859290783,0.3207020663646767,0.6955033106488268,
 0.10841386128018582,0.2272926139353142,0.47217323371399245,
 0.9180182158034058,0.784609853360704,0.12094772334123127,0.5864942614470461,
 0.6481956101414763,0.2331056221057901,0.7531126209134542,0.25849954423000987,
 0.5268813128396428,0.800789493044186,0.3307964766340724,0.35458064210198426,
 0.03241504177867893,0.1957968219311934,0.27256985878720486,
 0.7406865803733532,0.8804928059177664,0.20872172222291518,0.9283264784358797,
 0.405951562714431,0.5856196820177829,0.945818003957087,0.8575195628209407,
 0.5873843990097509,0.8783033742716149,0.6889559285005966,0.47887844515371425,
 0.19366847856678637,0.42510790722781977,0.7991209381623373,
 0.8842389062126834,0.2335013108225774,0.47907445646159075,
 0.45041066400622376,0.9005950690251529,0.700865621098839,0.2983602038728692,
 0.8358914920309692,0.02926355044766804,0.9777586270664836,0.4187022684006305,
 0.31297772264417656,0.7111945700229452,0.29551937529184147,
 0.4600396411126584,0.6711375043078118,0.20951111957091784,
 0.09898023062879868,0.43181627433605985,0.09631565077105686,
 0.1978681363669772,0.42014123235299927,0.613212575699628,0.6830574966757108,
 0.43456664982151705,0.2676005719104191,0.521050341726328,0.34404195629894796,
 0.12159620880485589,0.5399405804012025,0.19923046675527178,
 0.8748010138209261,0.18574639679198612,0.8643091871976766,0.5245653198137032,
 0.7311851600074795,0.13149093134508716,0.6959124785656241,
 0.31660710189933683,0.356387707539068,0.1618104542050245,0.975405649319415,
 0.46573018378828024,0.7142242914120669,0.17805968596980293,
 0.3140436615218465,0.7626649873069817,0.8208909704479007,
 0.038270186424347874,0.9975547919309066,0.8476870074208331,
 0.09736310476247279,0.9979680053919979,0.22440361964783506,0.98642545772875,
 0.892433041775934,0.30645813436802705,0.4352040520806034,0.3675528528398133,
 0.4039111334068888,0.33818147029422285,0.8475276887367552,0.9800883547936142,
 0.36508907826724,0.027536714682698138,0.7148657937320518,0.3790490131736941,
 0.27437072693095055,0.4439784001644087,0.015253431114209404,
 0.5507298705335488,0.03434904478522427,0.6934568416009175,
 0.002109224638066398,0.08566886255446438,0.6255900988928447,
 0.3847796414079441,0.47326080598796705,0.6569673179606703,0.0989514353651757,
 0.6704677087491393,0.3297659044379053,0.29549477842876015,0.5067648999715387,
 0.6780234661056126,0.41417474087075246,0.06921782092118756,
 0.9495095794930916,0.4251213001718842,0.5553545909735541,0.9480610832033073,
 0.4924153143464719,0.4950221962468552,0.8336400969667037,0.5491823403760758,
 0.949278211430222,0.5610170210916855,0.9830313884542098,0.5435694971307937,
 0.9792436635313373,0.16138575289352108,0.31145443707703235,
 0.48503824595087397,0.16723047738446783,0.38777302644670275,
 0.6421908580300408,0.9883995001357875,0.8080742200813229,0.7329534479667761,
 0.722872636548091,0.10898565149892847,0.6998917228450089,0.2562431210167627,
 0.44576053471112553,0.9886376782131744,0.8048234868807196,
 0.20732477831926766,0.817868157343975,0.966402285220458,0.9576354859360248,
 0.2582418797149518,0.458931926533096,0.4388688615198232,0.8781225929371157,
 0.18650811064595518,0.2887710469807361,0.8029606178956158,0.8430423548344309,
 0.6436473383297979,0.5194790763966102,0.4790983222391053,0.5507980627171887,
 0.9884164200816835,0.40351493846665276,0.03734319163320987,
 0.9750666141938563,0.9079188307920395,0.13817007421837368,
 0.12231926225239231,0.45645656849055893,0.438528512424018,0.9049280731397045,
 0.3119165139737583,0.970835879799295,0.5780991869741303,0.38160984803586506,
 0.5150389171350864,0.5735334907414749,0.9381781588465796,0.5304938599494758,
 0.7935730475279242,0.35909469764099644,0.1922563744288439,
 0.22607921907889827,0.4713286151095789,0.5278755369489323,
 0.46001978304451807,0.2473704702220142,0.466098238105483,0.7206360891533033,
 0.042026423540156066,0.3195945508197964,0.1972737396532166,
 0.2865881682452417,0.03779036375176581,0.07678119871917821,
 0.48111118561050414,0.9530167643484782,0.4473295511030876,0.6826486496200271,
 0.2765644833088946,0.9689515680741251,0.7691881676204838,0.43893622947317523,
 0.7660884005477101,0.213742717509269,0.7306702177625466,0.5757631888834087,
 0.5561753768603124,0.23103225817667838,0.9782579403201813,0.9454101559637897,
 0.7548436815184416,0.0996464972988037,0.3001979039888195,
 0.011449165446318355,0.11904656485871679,0.010319902566927275,
 0.12131942205610069,0.20889818152134576,0.3397678891878023,
 0.6184353106355276,0.5444021704032553,0.876595888164028,0.8174756235287207,
 0.9444051259276925,0.1533998508014044,0.3953828760749829,0.9972031773076039,
 0.06466668386984264,0.9261707616582964,0.9949517517871864,0.7212839424980221,
 0.7643154445364815,0.8338524536604258,0.5387007705241591,0.363115833262844,
 0.6265404004694214,0.9619985936657288,0.5399996183554926,0.12305920391500935,
 0.4834823649200737,0.05305329777237888,0.30854555758615876,0.097645663856616,
 0.9625695681281495,0.5114457549599023,0.14052489312655891,
 0.23379259618545867,0.09864625850702091,0.14400366233838624,
 0.4333817970208269,0.18714738176767687,0.18439140387471076,
 0.18872670547693082,0.4838022916963023,0.1174515355459066,0.6712551349943671,
 0.6109133093039122,0.25294860808240227,0.2665069341458278,0.1931901306624103,
 0.2605667364743951,0.8924040596514703,0.5517657280134527,0.11310835366136818,
 0.45159605260680513,0.8802605771972744,0.37113103229787525,
 0.09074077682105985,0.4279193874470987,0.3996153538176084,0.7658778520860832,
 0.19187367268929145,0.4354870003638218,0.18801698555897894,
 0.22864243893367353,0.08377962997685984,0.7353043837115875,
 0.6579406549853606,0.7312284293057334,0.31754938290167223,
 0.41013480588745177,0.4872244329621741,0.13398393626039973,
 0.32807735458035914,0.12580966656130899,0.9195714561875044,
 0.9588604852963255,0.18715508504990608,0.10309266015115803,
 0.5538341670529049,0.6326124990350726,0.260851133268051,0.9480513985181465,
 0.20632651887708642,0.36957745550065424,0.13567908181160604,
 0.5586911529305288,0.36497428000934506,0.9526348519221092,0.8141164558834736,
 0.48598180982254513,0.588453150201419,0.8435012946839687,0.8552858160760721,
 0.26032107782287905,0.08544331892186452,0.2494084307197325,
 0.6580443687259239,0.9288996202178468,0.4614933756037438,0.7525445723323616,
 0.5016162377456209,0.6464063515388359,0.5037766512833193,
 0.026002702685237744,0.70622139981579,0.35634240662801364,
 0.08557772986244472,0.9036016527096479,0.7157875571943169,0.3939731210602828,
 0.1547561025381733,0.3262944413649609,0.3017324368874591,0.5140760650075284,
 0.036123209711614424,0.45758176342094015,0.9689454613740541,
 0.11204986417345564,0.40121813599474654,0.0811922664218524,
 0.8033684876136257,0.6290105424551736,0.4789042824493184,0.4880527005553009,
 0.15210535335397168,0.02803559249991905,0.21767076741429514,
 0.8575513050946438,0.23046587995126555,0.30683361480616234,
 0.5359462211457015,0.31094120068536957,0.9311970554679191,
 0.39552103737700206,0.5947251406700355,0.07356421145661085,
 0.8601844608575768,0.9812885603354375,0.27685373249105427,0.4693195862401338,
 0.6067317198589999,0.7146774898694154,0.37700237547272275,0.9187984309417244,
 0.08753198426707165,0.5987944498542144,0.9351530164539216,
 0.43361512711004757,0.10517361279133064,0.4161257609304906,
 0.3579412755855351,0.23395692898767917,0.08811361791982775,0.620430837958595,
 0.1229002797015245,0.41313945323757817,0.9102593155620121,
 0.06080578008362758,0.8305274508906038,0.6661486208906183,0.774579288732875,
 0.4228688041201436,0.21216323263629422,0.4295854832356423,0.2113148486053612,
 0.7494185373851838,0.999853144178393,0.7192518382621493,0.12122883236186122,
 0.08415280941929604,0.8126916099771029,0.5555068270761051,0.2101287544731042,
 0.9031060745346097,0.10116942483545854,0.6217444486379949,
 0.19104619955981506,0.9436850381554187,0.8769564016071424,
 0.04317270630664749,0.7796800507199207,0.7173652309057732,0.552363729134587,
 0.6941541648968431,0.1285716753679763,0.7976962959131938,0.0657061306060116,
 0.3540167150674832,0.8039936687383057,0.9498879606743273,0.9166174660808527,
 0.6007243056254888,0.42509543382055215,0.5754586830333375,0.7900629621223376,
 0.004254815988397054,0.8352954725912137,0.5310454208882593,
 0.0876093124120183,0.07989957991494856,0.5520428580419258,0.720133448035365,
 0.8505898243611714,0.06079554367446971,0.6510554747258575,
 0.20475179729585224,0.5506710117264321,0.24119665935121726,
 0.8069194431859711,0.8844003829321174,0.308295734400455,0.34296970416024075,
 0.48548719260136497,0.6076482528271334,0.15170096961176938,
 0.7132524689280796,0.7421179914552841,0.008027400621260061,
 0.18847978976299662,0.15484612785102514,0.2614610119511225,
 0.21033553481311618,0.9721907264180177,0.3587875400278635,0.5072275935471497,
 0.3124543793412512,0.1691785695433483,0.09973738972289958,0.795552012300959,
 0.8693428465760462,0.87664249047924,0.015256067352553027,0.24294439225724773,
 0.28876160743554546,0.4686106333482062,0.8822686972018835,0.8689891004760775,
 0.20858084520631914,0.7119313226303002,0.22151821782461922,
 0.6805552570997444,0.6248300457267482,0.5019981035457595,0.5109024521309601,
 0.06708131460523037,0.40103833162152713,0.7890987581858744,
 0.7099628686377877,0.29246426547456505,0.0867146692009374,0.5734166674607056,
 0.9649042098207691,0.3090789617015567,0.18616094429712793,0.7223579275502022,
 0.4839911584382709,0.50967459775825,0.2538988453879659,0.28371858499884517,
 0.9188710580642855,0.7440981355271747,0.6537455442824648,0.21734718531769093,
 0.2855372557453506,0.9378354425996092,0.8612287572445104,0.43197036116246745,
 0.7094176745526568,0.9071464553429254,0.4023909596756674,0.1432314407771731,
 0.27253820309521315,0.974952462826708,0.9373663183479777,0.3342207035557502,
 0.6885348325193055,0.9631133377982528,0.7633819678397407,0.14022268262441473,
 0.2639886005703376,0.8592442829976394,0.8578839224609505,0.3286127484535979,
 0.20362175701762464,0.3817982532588148,0.25693052679749084,
 0.28043296425128894,0.8435129720866543,0.04079573184641816,
 0.4193709967088022,0.8690772018216784,0.9989787249586691,
 0.016075879658263292,0.053297537212227186,0.2072244851294608,
 0.011184781111665432,0.40393269879577387,0.6292211495350192,
 0.9247388297242201,0.9382540679234335,0.5176363374667384,0.6322866520145156,
 0.49148978451045966,0.5331278273129963,0.2789564826698192,
 0.10071818210213201,0.6699219333850422,0.6359223348969364,0.7871664956904623,
 0.0841273509329965,0.6041251773019283,0.1373112374501324,0.5046492340353478,
 0.5533515594772306,0.7746231441845739,0.9015533919844778,0.00912362338825623,
 0.654694038246572,0.1850647090198878,0.1129082551417544,0.5284865179977173,
 0.8390376309566512,0.632691951406439,0.9380668640623885,0.8284240608518827,
 0.7019120266752227,0.5289428789202036,0.6020294255351387,0.20681912203390573,
 0.11210860340637141,0.7556349324444698,0.7496649177838233,0.655709575176568,
 0.48214161681410084,0.6467519294121797,0.2605945409895185,0.3168144548402726,
 0.6742284458804202,0.25000257812920756,0.08214561061442549,
 0.13845340331741707,0.5693390662840533,0.22753791139445245,
 0.3281733933805412,0.724528553427973,0.9250047972473815,0.07733875688621561,
 0.6846013288031602,0.37775093830409123,0.48087662931043906,
 0.21444265172255916,0.9415146884281661,0.3028361928848883,0.6727122470961266,
 0.8025778734240246,0.24016050714263137,0.7108303057978091,0.5420287410345355,
 0.2657293555068261,0.31439988326331236,0.3676339827704146,0.6488628706168825,
 0.41487172688470064,0.8733620398027583,0.7731721850749971,
 0.34797361556564677,0.20538398171116778,0.2115294035268107,
 0.40253425300082835,0.9136893661846115,0.6499863538683633,
 0.19655298331558801,0.48828514578001037,0.49476322733489275,
 0.17570236867472855,0.6317971950208069,0.8060365513519191,0.7207761839777814,
 0.9847041390928065,0.29161485420745126,0.1691642879488433,0.7952693423421299,
 0.40284606009704504,0.8636490144811246,0.7679764136058205,0.831298182465718,
 0.8543846917619644,0.39516056354062257,0.917020284064298,0.05188550630336264,
 0.3558335470029106,0.6069301180456271,0.511667855588104,0.9522958377891588,
 0.19727431350950198,0.8227915095265459,0.770016540361988,
 0.012490044568298497,0.7096064590846237,0.9072304217667368,
 0.48648637191358834,0.9743295761693005,0.22339119851857903,
 0.3235991179522697,0.6807182982122512,0.6509620335031179,0.2990482356016305,
 0.8771727330010882,0.12675590093148592,0.456049961021771,0.22368680031847754,
 0.48148927707875844,0.1387159605133078,0.07233303215349696,
 0.10593820771677498,0.7750924265687691,0.05848032907756473,
 0.2636753188893557,0.039874285162943335,0.36339491932415136,
 0.05387876254780588,0.19297572333607393,0.27504315104270016,
 0.02781721356345257,0.6294201484108354,0.8356862412300727,
 0.0023345729329131704,0.6074724088394735,0.3578550358051854,
 0.47268493053347216,0.8240121008472978,0.4240963603549507,
 0.26722609794708596,0.2820762022472353,0.9403963226871331,
 0.35251506922504716,0.0800453632368946,0.3176523091967428,
 0.45618694159585604,0.6552904832795113,0.5584672955818734,
 0.32345106560726633,0.5318467880346284,0.19691365175998077,
 0.48163421315102306,0.8289666156085832,0.7296059579090882,0.4518410534897256,
 0.6322568640468805,0.4064564592585507,0.30250789965166236,0.4463657588200194,
 0.5199256831958776,0.2478887900120299,0.3593637606707206,0.419683989657324,
 0.2971602845092898,0.6307976920026614,0.9014392477936081,0.5403111208232806,
 0.21757323819671592,0.3463166848873137,0.5951713079353251,0.8415743834050871,
 0.5299125238602785,0.8460087027153627,0.33326599073499874,0.4984250848677193,
 0.01684570919425421,0.1658578932978102,0.3421636470246957,
 0.07254369542674122,0.9744251899407101,0.4845262139367589,0.7243278052030231,
 0.6126416078854943,0.3545437946462173,0.10437969253283674,0.6119637631712526,
 0.7285217435998328,0.4508084104735428,0.827288267840594,0.27858624275302646,
 0.06504199682320155,0.22360887879682245,0.43201784737673055,
 0.4509829261280407,0.28234657714166034,0.36431879445163395,
 0.42180956722473706,0.6232930843234606,0.17625059656869024,
 0.7100881877138842,0.20651771683436515,0.3797432612925833,
 0.41876132696967094,0.21909807126507175,0.24967909744181072,
 0.8561768026307751,0.3271251292130858,0.2961404273590197,0.6218784056302307,
 0.6047853944389961,0.5778255544139453,0.6947455780811105,0.7334701755205772,
 0.23667598684903224,0.3606662742832014,0.5678314517112656,0.175435255280037,
 0.5536858907792612,0.23727649028575293,0.1345190880414605,0.3504581517744493,
 0.936133122523257,0.3095421014624913,0.10617528830721035,0.9497097572056079,
 0.6394516749241103,0.8566620411040937,0.9621257982001106,0.3291346989749172,
 0.8339330135375985,0.0995021832939682,0.8172281705964984,0.49856403868851484,
 0.6745521534161548,0.10753006375809349,0.08266207932358904,
 0.06604820501358788,0.9768620318779813,0.3942075997530594,0.8248478629722681,
 0.8237560378959996,0.8036937753516917,0.7827132696601743,0.08910423606712259,
 0.3488726877885874,0.8169464795656313,0.5867471708208801,0.5834709468259565,
 0.43276437151510183,0.6375975789822673,0.8989287006787597,0.7566050954545505,
 0.07392700233344873,0.9492542543099679,0.7077419666716342,
 0.16283830803743582,0.5287659291941139,0.7875661099785999,0.9609504543466927,
 0.12281944674784362,0.2812248960311051,0.8268353001674948,0.7207841837071691,
 0.9050081146511434,0.7685411075682065,0.43331807809017453,0.4699484594812453,
 0.8122868227715652,0.14014924368796777,0.4077456273328314,0.5929673859206384,
 0.8951125267932859,0.8584133716228451,0.5914638587250114,0.3246364005792377,
 0.8364167866326672,0.19852604277655184,0.49628817629489463,0.118365939460495,
 0.09513652728861899,0.9112207635644131,0.9135115043135564,0.9009303057905228,
 0.9696604130779708,0.7723824205810362,0.9559120842110282,0.6023792285241295,
 0.697645074086693,0.7688431214640336,0.9558649238978182,0.18337356761408108,
 0.10306580075677874,0.6167068218278339,0.7263630970799636,0.6529405758149174,
 0.29575538107413846,0.5965312558787308,0.8747474937345197,
 0.23194635893151494,0.3223196473695362,0.7775652303279149,0.7516511418020498,
 0.04304507602218066,0.6717279127186688,0.6407060895416432,
 0.19260006492529058,0.3434117287732724,0.35630389502476656,
 0.7990562435516555,0.8531202582192323,0.08306313374167451,0.7632939143489645,
 0.42398511523015014,0.6203988425884831,0.3439582625281199,0.623501302031316,
 0.4544056290173061,0.8150306858003279,0.3282383569003111,0.8178131771415855,
 0.9053908521516246,0.3116241973233067,0.3595125548096677,
 0.010652484765865777,0.026328181096319847,0.9855584394874082,
 0.44677599803254064,0.32598547184981663,0.040566629493077766,
 0.5492871284830434,0.5327568714174945,0.46669309107545887,0.8953129809795883,
 0.014305475168458148,0.4918915903739749,0.732366648692266,
 0.004884227826831622,0.26985075286763405,0.990415810382309,
 0.021706748455635916,0.5925443235496273,0.31942070451294813,
 0.785035950574292,0.2721902307451449,0.8759397207378463,0.17629154594173435,
 0.2972333890741705,0.8431503138846065,0.8072273686592687,0.5080322049591206,
 0.29813177650319234,0.8810885903923563,0.22962045752184712,
 0.14278286500263837,0.23222256270971797,0.4414121526859798,0.313310399460616,
 0.13969234091136962,0.8644630362487198,0.8114324048732259,
 0.06388228265964047,0.9756286077607059,0.5408319513119655,0.748357807453275,
 0.5545301717382584,0.1591660143332827,0.9195950290506301,0.6319171807698094,
 0.19222254576400477,0.7207973042213187,0.9810578948188666,0.892684521195519,
 0.25491356494435813,0.42378459354246534,0.250240381864968,0.8339160705878149,
 0.4851642462669771,0.3288777750885288,0.7026548758100751,0.3016233813746839,
 0.3847726644622098,0.3995591562963695,0.11612121249669716,0.8948290361042236,
 0.9916275645206706,0.5763477699516968,0.3280487117675064,0.9468472870270823,
 0.2524091106825005,0.3018721635967271,0.9284923213086917,0.12136521529506217,
 0.23416729728643204,0.05882816739654573,0.9588320690402237,
 0.7855454112794183,0.5134825454083025,0.24778115451950866,0.3388183776242496,
 0.8830271209638093,0.11794751219033939,0.9307728019816108,0.7347704573605046,
 0.6480047781490702,0.5557951829410155,0.1658381244248065,0.6810613560722762,
 0.9742879533922122,0.6349473040420528,0.8418427111731843,0.27854225694236,
 0.3995862417822429,0.2736781654777516,0.4267624285566567,0.5097704897287673,
 0.6293251733910928,0.2020013073543192,0.2738018202245519,0.7989608669054427,
 0.9784317117846912,0.6258119016327016,0.8229475712813434,0.9525867661932506,
 0.8106801153318679,0.4928978034951954,0.7871309046754346,0.2968769175158513,
 0.3742593426312375,0.07755153311279062,0.8961759877435057,
 0.18623371960775637,0.7577193283930722,0.5377516516341484,0.4321904269297071,
 0.29182580780317047,0.7160434271791987,0.9233449329663919,0.8574644972906732,
 0.4890589403102966,0.7244298930828835,0.7651139186302858,0.21170684109514926,
 0.584110000934535,0.00954572198367043,0.1200476789958449,0.48167232990483333,
 0.9496444962413146,0.22812089251101708,0.7752008527558127,0.1914351113627153,
 0.5506513795961013,0.7622836087197985,0.582450904604124,0.5306212933473975,
 0.20794344064018644,0.8977218320934744,0.08242963462369868,
 0.1788575907961859,0.8294730933514055,0.16830839381019258,0.639338510021726,
 0.07823681292804874,0.7698656248249082,0.38447601969841116,
 0.39023731483760504,0.49454429825652735,0.23293896123392488,
 0.5022142210179478,0.8034710626477038,0.004785271799764401,
 0.8293006456531747,0.07196173417091312,0.5002676591860579,0.5428074428241749,
 0.18242484814189508,0.6383415091353741,0.7535792534383474,
 0.30924873597178526,0.27233428646685076,0.2904854811815667,
 0.33095677484873887,0.38884896921488077,0.3388166499059079,
 0.3625785688170432,0.4552311397711789,0.13170239210104828,
 0.14746041916763408,0.5691116610653284,0.3888266050647958,0.5900279818247475,
 0.021373638145832596,0.6973065711604655,0.05615463343013172,
 0.8318282689764862,0.5684183982899172,0.6125014276235162,0.8500209508987826,
 0.14648575712325718,0.22163825507187174,0.7418705193692884,0.854785800481909,
 0.27260180922805455,0.19487187999862277,0.0842516306707245,
 0.4146623834026615,0.8908025013642147,0.3677574593250674,0.5992497442765444,
 0.23549736467655913,0.7028805496639372,0.4170005841005937,0.4260043997111188,
 0.32823330485086044,0.4731808067200517,0.3238043505643966,0.8734933969222515,
 0.7634281511953899,0.2440006270182863,0.533686537656504,0.940393915159051,
 0.7433157317887497,0.5993161832956414,0.5323955395335769,0.5853712897339295,
 0.8133168491045496,0.24350722128068136,0.17243532328181232,
 0.3539589557897187,0.8330219505122171,0.8716519601696167,0.8421846440605698,
 0.8593010499153811,0.10689921176313888,0.5601173905421691,
 0.21609129203524924,0.12928733654795987,0.9126672613582947,
 0.4820283348677771,0.9732313925978063,0.7004532985295207,0.5171845882945935,
 0.1662785496119915,0.24761306421101426,0.8271283320614349,0.4285046772650847,
 0.08330721320056167,0.12451692956057214,0.27360734815225896,
 0.5976982250415441,0.24758979502997813,0.9689287527137344,0.90914297615063,
 0.881520965430514,0.6992147982293522,0.5467951292996241,0.8141450815565464,
 0.12309339159822008,0.5734219494151517,0.04590646373375018,
 0.023797607669927556,0.5595728633328927,0.7312540638985049,
 0.5908946453049503,0.9159144497594697,0.1934579554611261,0.4076349264467807,
 0.8309865700953365,0.15067674712969403,0.6696762985239,0.966925768279613,
 0.46732813820285435,0.6213042707391229,0.9346317923664583,0.0715950635140643,
 0.19791221826437888,0.44509771032134904,0.9067791494530912,
 0.7102280347575465,0.056020400547321714,0.48753816778765635,
 0.6934402668122451,0.10221958883520188,0.5760783107500174,0.1892495662277509,
 0.15452658554813714,0.24272506335677546,0.060012460266688605,
 0.487497608567924,0.2693341095189633,0.7259110748300238,0.12254653966015694,
 0.10139504429699187,0.36225759088920007,0.04011732592983863,
 0.5241415802262362,0.1490332154529621,0.42007318806030536,
 0.08210916847470384,0.8560483055878594,0.5926436914703026,0.2995801728092564,
 0.5174829036308821,0.10663674970092607,0.9874672149086167,0.9817193918926236,
 0.2961873694021344,0.7744422536838853,0.9242921497378599,0.7930361591197169,
 0.9656857771191395,0.06756614650117565,0.3690716326286759,
 0.14304051385440064,0.46260011685167723,0.644505141976501,0.2653675441295176,
 0.7528423940907228,0.6926850703312296,0.010259467881598483,
 0.9224990809739786,0.6943360984492006,0.6507402719649139,0.1162588012416621,
 0.8238741196036713,0.006968986432929758,0.3375361306110197,0.759505671243207,
 0.9894213880852377,0.8452382995431275,0.9215633930655512,0.5664566513726088,
 0.44578609112077716,0.7184034072520702,0.4160710865855002,0.7501826303224725,
 0.9952753632985284,0.4422476831767539,0.4778688061090287,0.04644799361680518,
 0.9941211471965303,0.19277588853797922,0.18206398186840533,
 0.4262994164430022,0.021584626987581124,0.5439870603900606,
 0.42884620240421345,0.6507954215274326,0.991328695980805,0.8158104383727294,
 0.8928376153229,0.04145101476587443,0.4797449686371017,0.7253802159142186,
 0.059555825581789734,0.26283300263153087,0.31584627839803026,
 0.9785324044728062,0.8008652920505358,0.2007243493101144,0.9650620863158719,
 0.1941411762380988,0.7988775658513267,0.5324719110787534,0.8403596653287785,
 0.8107325428017594,0.1470015745886879,0.38332507947216266,0.8628462847675249,
 0.5002507351297512,0.6108120884206658,0.9832478182165726,0.10474227179106,
 0.7834522932701758,0.5529087745160314,0.6807013396647983,0.05739258986630491,
 0.018089484789358012,0.4557511706371682,0.8434704023839326,0.429666532088931,
 0.3926656942511304,0.8775467338016057,0.03935056030490047,0.8701035556693417,
 0.49458502964234197,0.16318023496500333,0.8979269085823677,
 0.018916778864503603,0.6246316943934818,0.30915079982711324,
 0.11083185603888712,0.634044557724496,0.32433609318726797,
 0.28975762476081324,0.835118430743119,0.9786125379342042,0.04971033310920214,
 0.8352209870259333,0.02853415291877781,0.6764608427281736,0.9020650038903093,
 0.008203038322109668,0.49601679575148516,0.5153177409377333,
 0.4044271258351869,0.740408250428824,0.648566950477395,0.44622194738602794,
 0.4317303371866432,0.6485342677443917,0.013227332769629174,
 0.13790046056196492,0.09516984458203925,0.6152726221803588,
 0.6660239718309063,0.5365503104866839,0.6708097416895612,0.9483630956152296,
 0.7338499416205667,0.6947877549575555,0.21707816770219068,0.9664412930518046,
 0.5568075984669971,0.2587267011347423,0.2940731830097614,0.2119381647637827,
 0.7911067259276781,0.44353586340917894,0.21975383536388093,
 0.5201240525019302,0.23398451666459186,0.024548186921694026,
 0.10595417138922048,0.39794235782590515,0.8286747735466287,
 0.7750818258031079,0.2607261479315405,0.8354388001518375,0.6751307704759658,
 0.4375092125160456,0.17920000934991276,0.20452945458130045,
 0.32104945527525053,0.963602979446179,0.0011616880038021193,
 0.7570783126706486,0.09662283721438336,0.853600438444054,0.11407750445389442,
 0.8930826623920962,0.1099207077831027,0.49722427724843166,0.5420665841118415,
 0.3484883966254374,0.17074715111803684,0.038615410478702206,
 0.8691277553647403,0.3702506324742656,0.2689959514298428,0.9959867126584968,
 0.04079040517521726,0.8084367714323659,0.02728001715465156,
 0.3856195237625931,0.4365484369439767,0.614652555126212,0.7307178191847056,
 0.010906928052144327,0.8293304024078614,0.5582909801681031,
 0.7190487889800159,0.10964599309276635,0.750680629386465,0.93074217419993,
 0.23102265151887247,0.026180073076432886,0.1983109444851553,
 0.6429778082687798,0.40949702140977295,0.8389287244523198,0.2858867046317718,
 0.8001143670915263,0.8791669085739076,0.4600604276173429,0.9144536134425387,
 0.6649101978958779,0.018442477460523965,0.18401378517218736,
 0.8776403475447211,0.6773512948392264,0.8877012386671348,0.619142758034124,
 0.7217437351243885,0.5096428623704837,0.37756241591692596,0.5191756562985053,
 0.17882810259142168,0.7204539578452911,0.6972651831900628,
 0.39151314766517864,0.6432746273743841,0.44061487073692296,
 0.4262581956926408,0.7383619384388453,0.7033880744897508,0.3546089261714369,
 0.011523724229953403,0.3389570212228862,0.5220033920975258,
 0.1579792507639839,0.8166891076380365,0.28535937067503736,0.3578261944304737,
 0.30757072452288226,0.1649256977475133,0.8048587283562477,
 0.21341986181444794,0.928001155603761,0.06641108563119269,0.980198893415697,
 0.1465920935902304,0.3133313008861379,0.4823598132169672,0.9215762763674835,
 0.36518169405634315,0.29082961796977447,0.34916090253037046,
 0.6531672406701261,0.5717745559414927,0.3993933708205737,0.417917162915704,
 0.6728864656388011,0.3742085471396548,0.29477513752083784,
 0.28071780009811287,0.8390758343417419,0.6448455813994148,0.5591667329372076,
 0.24673979199425955,0.9764568662673787,0.19577569132693085,
 0.29314975970823265,0.5251770071058522,0.5801162682125776,0.8782855881872194,
 0.7500795325887755,0.43472068567241406,0.040222406215389195,
 0.9252338034933776,0.7450264371968984,0.5162926190875592,0.7840911882185727,
 0.09064708340311223,0.4603573987740901,0.16364513748389342,
 0.18591772856285882,0.2885281648608147,0.5841779302893613,0.6507336702625179,
 0.3846060095939805,0.6325755359798765,0.3436027971434391,0.10709733651649378,
 0.6860175879726188,0.6431210100580896,0.1299469168265207,0.09533824724113615,
 0.3191314912599841,0.3560582120192122,0.4301376334393906,0.5484723014906051,
 0.571730278107668,0.4738860178786779,0.2922450489633662,0.5816494192484605,
 0.7701274174987434,0.05539994026648032,0.08516321168126151,
 0.8771037725362996,0.5225676391489971,0.4204461006710042,0.2049769394311992,
 0.3485845835265535,0.22591075707524721,0.41074039614898017,
 0.8230682804308861,0.19358395217123459,0.6376046410606464,
 0.19675705344825667,0.3946482083866014,0.40799914064325415,
 0.5992009086733733,0.31905522659012087,0.23552721677261979,
 0.04209795605080713,0.41710025995706457,0.72103032200076,0.04485001330209937,
 0.14784618444076802,0.8105502760560144,0.5344811655093362,0.5629802738690681,
 0.05642244107925609,0.19021009912668052,0.7918016479480642,
 0.6839338409650391,0.7301906163983979,0.7840383206795423,0.7676379840413577,
 0.8524612058471202,0.5453123225700822,0.8140081490561017,0.4494410772182287,
 0.18487468527309203,0.5301327320054908,0.8335933589955203,0.9006920001819885,
 0.24447335722070618,0.38988527630430503,0.08271810531134738,
 0.5800054826037165,0.06419428234075686,0.9612653291049722,0.7043711867133293,
 0.4173028423702887,0.5415622791616767,0.29996894119281214,
 0.10771666666978275,0.8762807356646956,0.36768096884542,0.16532292401827842,
 0.7119616580474553,0.07840313815320898,0.31987091296238546,
 0.9170573016352394,0.5915810974740809,0.3968159909095357,0.299626757613614,
 0.19686405718926503,0.660966760660376,0.41986283396714064,0.6401072742011324,
 0.4713971136106835,0.09328508749366238,0.3435728616899666,0.7448990500405404,
 0.6560027264823861,0.6826790536824022,0.45816351633170393,0.4790277255510804,
 0.49661713363802107,0.8580489374753034,0.7413635880486269,0.5795523863735083,
 0.3693904188124123,0.012611415484478306,0.8168652145786407,
 0.8374061937826509,0.8100658261909965,0.04664752156115315,0.767532999622987,
 0.46401324297459334,0.5119990857613168,0.3631429661429182,0.221498656061892,
 0.1319248528435808,0.04561904567274977,0.5852046025860099,0.7927636451697803,
 0.05133780131508214,0.6987586650533347,0.49171128881671644,
 0.0944186032783807,0.26225512989832633,0.6276383032837958,0.8639333798997745,
 0.8384114590212847,0.5738331527243791,0.8860632721503442,0.8021333994476829,
 0.5070632589576624,0.0799199067338302,0.5193459321145605,0.4202837987046817,
 0.6816357375800518,0.5718570285746123,0.8103657994126023,0.6952071299954689,
 0.3195359755298681,0.5602359834844709,0.07947792316906466,0.6286396087107173,
 0.5328239397548756,0.19655723448003282,0.8350509394124489,
 0.01905477686705226,0.1852501137558561,0.04169016396237046,
 0.7251978380720904,0.5242247986751145,0.6804146089278776,0.8193182900087775,
 0.43960375838428156,0.17993204873791568,0.8194086025393497,
 0.5196143465148857,0.40081271013138686,0.34590641042219095,0.683181705953109,
 0.029301121240414396,0.9363150522045824,0.9081471112187766,0.886537548347419,
 0.5488805668576939,0.42771467308868993,0.6206463467267186,0.900542191344132,
 0.1530760512013989,0.18884723920435664,0.08790330050557937,
 0.07299560469875188,0.19454227851873507,0.6002246185619701,
 0.7325481021936606,0.9309266459392072,0.9515010079379631,0.5937069046596243,
 0.690358800057953,0.9966762881383244,0.4201589905720703,0.34919419317736333,
 0.03145659455555816,0.204441246358399,0.5138482193387338,0.24500916265980865,
 0.9380882262878416,0.2493544828070262,0.3269831945087448,0.29333624177026096,
 0.1603822966940417,0.9608271221493068,0.6184147349246846,0.1254487882499583,
 0.5992619735582321,0.6297972713079558,0.4942797713990802,0.12186439917290204,
 0.37610830710578025,0.25871586766393617,0.28879714912989585,
 0.7036282045212797,0.3181962737374917,0.31347917031030637,
 0.35277578719743086,0.19344210775818438,0.923634073873891,0.5121455345741985,
 0.18055141863239055,0.1760830680032901,0.7483952661907064,0.9716719239648929,
 0.8477921366942393,0.32169418787897874,0.7668315168476705,0.849887406274722,
 0.4248870465953082,0.43690153636520124,0.28144790696320976,
 0.30594288686753046,0.8556036049492227,0.16936267167861951,
 0.1254312347557457,0.24472594412215576,0.6738590396227654,0.4978621774799882,
 0.4529181552533229,0.23120012555871006,0.32126279681165726,
 0.39321846172304575,0.009878347615505323,0.08628408022327538,
 0.638072771349823,0.2989595026375176,0.5229044469694408,0.5705601660240751,
 0.12321492227654762,0.38512540293719555,0.7343910470123556,
 0.4871202281855982,0.653684559018575,0.7436050262400962,0.6715017468655283,
 0.24085325858069695,0.38861249256483044,0.5288841172697493,
 0.1816571325251406,0.8271785043696137,0.03677324407012039,0.8160098159045013,
 0.38204551664445296,0.7968781553580949,0.5089939332566469,0.7131226888094293,
 0.7103989952412508,0.02918496471559795,0.4574636204379615,0.9703963583563376,
 0.9469248669545757,0.8691323225383117,0.8621254379786747,0.8393227363278704,
 0.91877195176836,0.997026650241602,0.7651610576831555,0.26550119837950625,
 0.23015146926615104,0.839175430954928,0.2635287112301079,0.6541000105326767,
 0.7877606084945501,0.48783052674528216,0.07235273576486356,
 0.7291935020454635,0.3005060564542258,0.10496053036921327,
 0.08925492319554629,0.5506776031124141,0.247247563221219,0.10882080885307421,
 0.09559699223332352,0.9935430772431861,0.8660322800155256,0.8997581348106818,
 0.7031876663699886,0.6782558684182717,0.9878870549414226,0.9433279620851495,
 0.20026467956052652,0.01051363095278024,0.27783095032821237,
 0.037635939589675815,0.5268367263729622,0.6796646001959256,
 0.6260037199292432,0.14921183228514234,0.29080086601307176,
 0.1324784670219239,0.5087393067292155,0.5182801012121336,0.02428293221427924,
 0.4184955955409,0.37146894915975626,0.019917394598976568,0.7744634665725727,
 0.1435246284849253,0.7103308130378704,0.7906525229700738,0.12829151741048794,
 0.12490118427057528,0.9429517088524992,0.20692532956454623,
 0.13564638093425352,0.824448598380238,0.42034182722872293,0.2925676193063418,
 0.7420115048251397,0.6706180450517529,0.19735405029896858,0.54645562857463,
 0.1995863409701525,0.4310995445506304,0.7711335217971362,0.4560794712521924,
 0.6314565833880457,0.6448137995366644,0.16561930377263967,
 0.049810617003373525,0.17022208576353393,0.07208325973125607,
 0.5179261363086747,0.15215740059281713,0.770068832649281,0.855284736609016,
 0.6182338278352747,0.3425181509105164,0.024227327397510767,
 0.5353786192266723,0.14286340100074146,0.8877119847004573,0.8420376440699949,
 0.8919293103696164,0.3085106974226163,0.8220448389746099,0.11629814434681551,
 0.6037952209218221,0.06989242007148999,0.4076753712882457,0.739483994856548,
 0.619614469428325,0.025382245000203474,0.0023583123270747297,
 0.5888088578214083,0.19060921171286882,0.9532965882909537,0.6556809019437793,
 0.14790287334686547,0.049682739629015704,0.3857381210391718,
 0.2659501545797339,0.1734803224278978,0.5753541332626847,0.31148128949311327,
 0.7819951689941407,0.36946437792652154,0.3624128974992715,0.9486418066163064,
 0.7344452156117124,0.6551183823207437,0.8581811247329351,0.8012376069050591,
 0.6855804934121589,0.8205889901198131,0.531783516532557,0.5956882853959488,
 0.7882043103419621,0.9991729235626916,0.728638054692641,0.38203749228541073,
 0.5976345431321223,0.39695933175013665,0.3133445597613116,
 0.16115497734981477,0.3229429501594574,0.6839864159869141,0.3593234479621267,
 0.7733720992484852,0.7598167048894077,0.18572343583935103,
 0.05492854912266443,0.5968291257412514,0.8303727700225997,0.1183176850263381,
 0.7682304661242046,0.6502420534652918,0.4797390202677614,0.9648791183012333,
 0.23536002536997136,0.4922157957550122,0.4793150969728691,
 0.21428421259828656,0.2523817909429774,0.6386068344535795,0.8729282421807604,
 0.14770405044507418,0.026652951406005987,0.5376954186803294,
 0.28402454457034676,0.5559473592326303,0.03631760714441479,0.691394634794493,
 0.4877358083177449,0.6720211625360393,0.26668419638290863,0.3552350539614929,
 0.0365027480656287,0.6222092502654707,0.30877975537124114,
 0.21646292193249628,0.8605516189803789,0.15854483280664933,0.853775869751351,
 0.4559155999942174,0.5759649316046207,0.33653571319246467,0.6270686107706637,
 0.4905080256348343,0.734733285566086,0.4459314208558538,0.09603534679865178,
 0.8241756669968304,0.4217143746274443,0.345756793160399,0.44449456612863314,
 0.05217145178790106,0.4862363235048185,0.2381961808948031,0.7975223846620842,
 0.6643006834949059,0.6205801476705822,0.7667596092929142,0.03989172044438527,
 0.7936975485089719,0.9281580457330051,0.7615931513879118,0.6406457424344392,
 0.11996032514673693,0.3026614921810631,0.8963097125029275,0.7007067252347925,
 0.49876227975793874,0.9163367236093014,0.46578490501059155,0.830785110242848,
 0.4620277703669735,0.08084475754358045,0.3380403421348106,0.8473213235829319,
 0.8095308414978495,0.8798727025501338,0.25435455337822166,0.403724934480728,
 0.09311052714259149,0.9580647555102079,0.792925271919564,0.48589363890511306,
 0.9890864025216923,0.5279200153617647,0.013337097992241254,
 0.9142844765932188,0.48196566705462707,0.8981091212882503,
 0.49326101739432315,0.8023668980553635,0.17863013280185047,
 0.3726722174761605,0.5240306727735067,0.04416771211115822,
 0.35412239571352133,0.4414706222086401,0.40039999003086146,
 0.09205481197518972,0.11171662020666318,0.6965562376514344,
 0.4170103630546381,0.4955337719859251,0.348322083562964,0.09542666735998484,
 0.0021329251421153828,0.3633332647328694,0.4268393816368966,
 0.06668368633683108,0.7130355752889873,0.5747207095238759,0.3155894689906553,
 0.26163085492036364,0.5507471197362762,0.5983297116460378,0.6783892509929231,
 0.3376964284007662,0.24257493468498037,0.663641561983243,0.22205595386451782,
 0.012821009946220707,0.25385950616057307,0.12738752388349805,
 0.1949218853200967,0.9485982452180194,0.6339234830002949,0.7995747270028407,
 0.9920633039641252,0.2540148013546597,0.4263237244946564,0.7118319199675947,
 0.510984032465956,0.09149832586553452,0.3975773079006121,0.9427817170197405,
 0.2040372005211999,0.25740531538481526,0.8493776126089067,0.8186580569628158,
 0.554523663029194,0.16185060100751758,0.2953680274177387,0.9383934671822207,
 0.4340461007118134,0.7733189815265098,0.6217972472902322,0.3001844338374393,
 0.8081762564422297,0.6763913892048375,0.5064780341894687,0.662302179592971,
 0.4665740049217797,0.8465073509125719,0.03753270827751287,
 0.11901649484072774,0.6328752580556116,0.7915355226550791,0.506879265349508,
 0.07807055601712865,0.5154983984560656,0.40738799966310657,
 0.011849192096269778,0.11982869537238805,0.44662460425577266,
 0.31787703292083724,0.16114501472132614,0.4016446697231757,
 0.9857857963887481,0.17092761794185696,0.693950537654516,0.8870522711809766,
 0.7122168158425677,0.2147116739125915,0.8475802064620339,0.7166603807614735,
 0.8100729269716003,0.5567254256763227,0.9376848536604115,0.5164710375135774,
 0.3304958581568108,0.11928736264371598,0.9544365444874016,0.7887179798683374,
 0.2322053104031242,0.8435146950612564,0.39106438586149983,0.7711109359332817,
 0.13636334065201983,0.02121454286212976,0.8358591706846838,
 0.43204154298484665,0.08963535218464047,0.07434950743845259,
 0.28874273562389996,0.6769125039512568,0.23314178652382755,
 0.18849044547169602,0.08064522827834475,0.8470480078446522,0.533020526067484,
 0.5628569831618689,0.5467774592500618,0.7980382501533438,0.4561767291256844,
 0.6503158492522692,0.5140090460084418,0.16479228519211608,0.8938785412199618,
 0.8253005607856149,0.8722278567595695,0.3112352723499604,0.8507850489108822,
 0.12025165958717277,0.34013257151212606,0.4336088403629328,
 0.23436047416376604,0.05274399441834565,0.8888627018678199,
 0.6180394275279136,0.2944047225278599,0.555841591222581,0.14342926365527953,
 0.2765902583455082,0.6732423171599022,0.8128133150361709,0.7709601701021539,
 0.585284130832868,0.08642165800812796,0.5743692615763882,0.06391567300230949,
 0.07902059932661765,0.8268952179202745,0.16018269361728021,
 0.6285935480107465,0.3444744965490424,0.10218312200166169,
 0.39104799064404094,0.23489762764745092,0.739341252402704,
 0.36419813125176015,0.691776921675736,0.26447858187971685,0.9808894887569729,
 0.908703273779251,0.8772825937841434,0.39608346391061666,0.45171922459053915,
 0.7750565007710192,0.41231507797484734,0.32402885150875504,
 0.1059620611633808,0.9554684714960346,0.0881044861573067,0.5016190026353985,
 0.6428183909263118,0.5893243426577496,0.7641246755328512,0.2350008042674533,
 0.06589354506630296,0.013513318049577339,0.9650290593858771,
 0.1906391674174177,0.36719862315943674,0.5398134719272544,0.2039983571692504,
 0.47626807963479134,0.07685280248807658,0.4984014656183472,
 0.43802096879634356,0.5296253771891618,0.5772964528282432,0.9672600148857251,
 0.9134935530554841,0.6946791582290159,0.06614350733912577,
 0.22549883021410855,0.543359445098138,0.6099921044256128,0.1338082305768724,
 0.30242591167442556,0.9099558851396359,0.4053933028975829,0.1003978656771286,
 0.23765124308602126,0.12224101895728368,0.8396873924817942,
 0.7349666308944169,0.4715824418980805,0.829141046782204,0.8861982066999228,
 0.7286279550723761,0.9590208546831946,0.19337105592779502,
 0.07377595402730841,0.13305563363133976,0.9087445390726601,
 0.44586737232697515,0.5441890349274208,0.06943119247417551,
 0.5273709667348495,0.41659786905449736,0.9660619975186344,0.4209400031580355,
 0.8373379168346934,0.781675666486557,0.2837813677581419,0.22717151832020144,
 0.05947102393729464,0.8872185163358155,0.5052863695762964,0.8440654934373716,
 0.6265797324930564,1.6225660151492072e-4,0.7688551702143603,
 0.23489511472339875,0.8450554573959358,0.3517214703004149,0.4904911451111299,
 0.9040205393568633,0.11245780610124512,0.11464351199783684,
 0.5148463321088551,0.8674043821006064,0.4949622796047437,0.5850052347027013,
 0.7492396182709544,0.7915377310133322,0.4498709059964665,0.1810401733320336,
 0.5435771158913503,0.915427460120924,0.09630569984618853,0.6498020426134082,
 0.21178160045160055,0.5753215790311864,0.27228474002129455,
 0.19961546395875396,0.48247777963607263,0.1655212009432564,
 0.9698774706250674,0.3251734757795577,0.84463461211992,0.9870782719391729,
 0.7579549371596189,0.8490449630001276,0.8012784819652478,0.6948082726770238,
 0.3105941800457609,0.963755227856546,0.13084176258504054,0.963637270249976,
 0.012849869134151426,0.9448115768411287,0.1798385274480523,
 0.7704694354415107,0.017053295032720794,0.924891800959636,0.8598374043828415,
 0.12691827502711484,0.7707475933886885,0.45933187328570024,
 0.7632145860003324,0.09811854718400226,0.9807732548245724,0.8410306635526855,
 0.4569531635679487,0.8515876194949491,0.23938258930665535,0.6840212439234372,
 0.35774307359897983,0.4836620443036468,0.8492854958300434,0.6821386257035655,
 0.12410516504694935,0.9516016576444265,0.10533205906829823,
 0.0043180995854849025,0.656203728762669,0.6263237689817721,
 0.32147559046508134,0.1593246229655364,0.1990785068737575,
 0.47764367035786026,0.7155951865240534,0.3503994362576257,0.1541181139540686,
 0.5716832004680521,0.5631230786133119,0.799436929549674,0.7887857400540341,
 0.15426883372658817,0.5046193835183302,0.477058508909178,0.13987713211535435,
 0.22890237727032714,0.528019674446443,0.3145693746629943,0.6243384262427627,
 0.05689821413856755,0.2719112242611783,0.16984169426775608,
 0.4146683842847231,0.2454845913547038,0.48238035906943366,0.1309433914874506,
 0.8419476027688436,0.34051583809862995,0.9770817660363023,0.5688962228448572,
 0.3880251444857654,0.9294082148057311,0.6859419103107606,0.5174006017194285,
 0.36967879688788097,0.4713990862224222,0.19306583690959467,
 0.49055766028379455,0.33096141685029934,0.5815209028879198,
 0.7045672960919948,0.43754026913800736,0.3378494362450589,0.6026492641860526,
 0.7402720313535981,0.244261969068998,0.903523488226726,0.9119990909829429,
 0.46284841729003334,0.63706845469099,0.47528437119219136,0.7123328654385227,
 0.16657250538075363,0.3525231778267074,0.1311744609737987,
 0.23006234506242618,0.09160100225299073,0.31757026311067704,
 0.9742818456915604,0.27384577924702747,0.9558355832411367,0.7516922741778045,
 0.11356239483720398,0.6612518954075512,0.01837556063962187,
 0.8819113341543385,0.5551094409853057,0.7166269988237035,0.4315681600373962,
 0.29147921524905096,0.7851748063036843,0.9910391003463622,0.425433289701999,
 0.13001811794609774,0.23861286089269218,0.8459196031375694,
 0.05166418301122688,0.8061132401817612,0.17130218400963337,
 0.02765473779036598,0.9797909539524613,0.6294461108322887,
 0.27712730342650593,0.9117133675886773,0.8689528281431577,
 0.36038136089782613,0.6562902573941329,0.2992773568022735,0.2886315473916399,
 0.5382378268065839,0.7396383051115918,0.022561867028672467,
 0.6937034369717342,0.15262949023456285,0.9315615795211971,0.3475241983514826,
 0.5136678187999808,0.5368939768817731,0.34171403334636996,0.5842694404707044,
 0.5496746252950245,0.16627374507817838,0.858086474341855,0.5395604292524709,
 0.5475931511355917,0.7422325705205463,0.8961625549842659,0.12854384646411254,
 0.8367793174166517,0.8093461858829114,0.06957913676525929,0.8973107787536562,
 0.884371851140292,0.3231711893896795,0.022153184807155557,0.2181764642924373,
 0.7180254926963572,0.26551432758908566,0.5384296903010295,
 0.30585511919745245,0.13503630052893756,0.23918401103702813,
 0.8054366276106565,0.6752823515494505,0.6883599354796228,0.228433038771222,
 0.7558487278356184,0.7407897759182093,0.4786708135556683,
 0.021524233880513166,0.7861102700066104,0.5949020731736776,0.481727911514065,
 0.44411299362148937,0.016629785760631344,0.26935312395175914,
 0.3920393891420243,0.17277162505570465,0.7987515960532195,0.9264973457507104,
 0.7079335340360267,0.13558633344895987,0.15840952684061416,
 0.2302020261330865,0.2694386801859342,0.7917498435563886,0.23742644004909275,
 0.1669108099389185,0.696604927434627,0.026866392117225524,0.6435567601118186,
 0.8789906221334923,0.0036167989606173956,0.2184905565768912,
 0.13505553661220648,0.139085315035688,0.5974575374302871,
 0.027799396747315686,0.3328895104711389,0.8154306490189958,
 0.44796897680490055,0.5890116974731113,0.14052007066145877,
 0.09277459311796543,0.7022807027067521,0.24289902837869748,
 0.6744850210170361,0.6487592078588426,0.030616734378590582,
 0.0833370121241136,0.6744815228429988,0.8248638177140467,0.48415387951270583,
 0.04281700343352535,0.2565847008360237,0.8309487730215023,
 0.47658292432377924,0.6597702815243315,0.6088781625365656,
 0.24690832715559763,0.0640860607547491,0.4210366424278027,
 0.13864108977089962,0.4029595826067619,0.5155998193412054,
 0.29907865444405424,0.0655200965740057,0.7642094429318764,0.4477154627747473,
 0.9635048444497738,0.9127787082429314,0.7877591419113289,0.31551416826109224,
 0.30963410964010996,0.7782754435142608,0.9595631040241583,0.7602714353528893,
 0.646706056617788,0.766550070370404,0.8283611375859674,0.4837444536254165,
 0.08890406356025671,0.7549353512323991,0.03235955869590601,
 0.07601241917941404,0.1620610575499023,0.7994674652158467,0.702703738662621,
 0.5256839344695678,0.3905552317785641,0.9905306271214152,0.6388748379823901,
 0.3417188175372833,0.5312579894562266,0.9670252972153943,0.340733354368001,
 0.30872162108715206,0.4420075747881518,0.38360179744537026,
 0.3829300678011174,0.8969405015931875,0.5606619506966363,0.8931445780190597,
 0.6875706004735879,0.9308540964064546,0.005359300094241237,
 0.5351409352820812,0.03433376368582286,0.2138256798362198,0.6656403370829909,
 0.9648070879335093,0.8929401447367623,0.35588291389082416,
 0.48242975848987124,0.32643244351694134,0.2768023627044336,
 0.39992055296765017,0.19840478507859483,0.919714711267022,0.5188518867642934,
 0.010618732074183956,0.5764938656555127,0.3651470916339361,
 0.1607445856521692,0.9380852856609869,0.2253672604961796,0.43557870216419925,
 0.3929702852839584,0.09744186487480022,0.5427723763042491,0.739102140566378,
 0.48975735010082844,0.21267728415277076,0.8844850865277009,
 0.9444032206503036,0.5965944781175302,0.8085887578604798,0.9040079960063014,
 0.15463661693926412,0.5352224047708919,0.25142796060747796,
 0.6650606951681276,0.42502894110080447,0.615091713046466,0.07175299284625036,
 0.18210669294224668,0.7857650806796246,0.012641721840909392,
 0.09043978694000276,0.12290172441295877,0.15844815350579367,
 0.8823873567098499,0.40329069476269064,0.9265951405725295,0.6425558468925924,
 0.7600474999474941,0.08853612040963199,0.517796414812488,0.8874073174785732,
 0.2777445166795087,0.894474408299552,0.45496972116674117,0.14333692284553168,
 0.3266147930812737,0.8955765977567108,0.688070971115142,0.6715108642412404,
 0.7921585739782032,0.6717094618915384,0.4816784895847597,
 0.047183891121357235,0.6597876149755392,0.312343601873043,0.5596746535658602,
 0.9833639615576761,0.2711523076719342,0.03575289637828161,0.5895761201507532,
 0.8638488011326464,0.9620962233647953,0.7771220958442953,0.9573348930423983,
 0.989477879400296,0.9823063320667565,0.08642027785843653,0.08924639283415503,
 0.4003525818812672,0.5711854161605658,0.2898949781112057,0.3366089970296312,
 0.839766501094566,0.5851586693043691,0.6028042086623973,0.14788349008105495,
 0.5067057097006716,0.8980217608867209,0.7016955767763599,0.7676996910568561,
 0.06028236567918899,0.20197779605537614,0.2816225588091883,
 0.2010542264538393,0.20455139392457178,0.24975867089443882,
 0.3716361387534264,0.03359069162924577,0.3329265935674902,0.2881793922604704,
 0.03316325552999988,0.05574814908228132,0.7961781774617701,
 0.7744467867783799,0.29706813844147906,0.9033839363111128,
 0.27446894809128697,0.9418575897004754,0.22144807393331356,
 0.7253864391047965,0.4599011295580059,0.7393380831757453,0.07063970237162898,
 0.6770474817990508,0.08147610403893368,0.525860164440937,0.7158200068995746,
 0.1999764969678074,0.1380609602243449,0.10345246596314217,
 0.21917915550693134,0.6034534466728215,0.9783446149483107,0.6905911960234501,
 0.2568795652774021,0.9210749443362587,0.07168671144177297,0.6959488363777763,
 0.14886856543504723,0.7144492738181744,0.02193345280828518,
 0.11151083120026462,0.2304725209069377,0.8354759783433333,
 0.28728072592445997,0.7479409523238792,0.6728552111332446,0.6408541495444501,
 0.17128731085807458,0.9512550634174557,0.6532344024953666,0.6553560125314812,
 0.870745478844095,0.923729853231785,0.5124095461584763,0.025341591195517377,
 0.8311299945386129,0.8460049411272199,0.9883733956119873,0.2088500184574148,
 0.34324059532201134,0.38367835678294304,0.5333360983716751,
 0.2216297816900108,0.5126554931225444,0.7692775700517175,0.31040318071361217,
 0.18717744390349744,0.06948912569462562,0.8977493112583875,
 0.5577003195805476,0.38770549331277515,0.06629299731961469,
 0.9535034033007214,0.6954334562616007,0.49471660933240147,0.4214802848795368,
 0.3848885126608902,0.2306629589088639,0.8379980235657734,0.10930467030310409,
 0.6857875408953648,0.29443388757156896,0.5925878089988349,
 0.18863444819051944,0.03435364951669073,0.07761802769589066,
 0.2936917463613279,0.01299594941616955,0.3364646600628365,0.7813199660183716,
 0.7877182981179599,0.443471207956485,0.5719609969122569,0.1541109394111405,
 0.10837439173444019,0.924488793272865,0.23803218510168278,
 0.04795385068560021,0.5635742730378198,0.26595246527913596,
 0.4516702568784593,0.922197437471429,0.48229439957964726,0.8342946213066245,
 0.9842376818220442,0.7699097375272176,0.5147310787439554,0.5007059126677925,
 0.7329207814057332,0.03386798481981268,0.5284696240977227,0.8332140048211949,
 0.14437689518027352,0.6603673594913523,0.8249828303959832,
 0.13084238510437718,0.7657435625757264,0.08980445029625017,
 0.24349730136191106,0.8746160049894667,0.40731387507979455,
 0.5150365788889579,0.1568490980716425,0.5299133723009145,
 0.029839704000529932,0.42605683007766715,0.9585629576801213,
 0.8980454770302888,0.3234157970475363,0.7562362307781006,0.1676575699917613,
 0.6007882405339057,0.8757387257878501,0.5792952793762148,0.3610686177866902,
 0.26395664891159887,0.18056181244050384,0.5433642592983969,
 0.2142478755950754,0.38255622906893694,0.2096527049770145,
 0.006090006391680447,0.26820632992555804,0.6513595979005253,
 0.3559315756151329,0.47841217093444344,0.6349644111957065,
 0.08039050467560371,0.9228644808987025,0.41724343767929484,
 0.17817520534330655,0.9744806262083044,0.48120779358899013,
 0.15504739043482785,0.8194182777461494,0.8428922455800736,
 0.10342101228995138,0.06746202034484239,0.4630569693102514,0.553784913595786,
 0.951610615263381,0.2062145445598914,0.8106223246400222,0.21134337186986918,
 0.5063936389712272,0.5620255443823652,0.8537428475537739,0.327170622194073,
 0.48321557650301905,0.2843839033347485,0.9304163086966162,0.6986078716578257,
 0.23087355661785258,0.40238363958190704,0.37875385284596996,
 0.6339052715286991,0.9914968024524462,0.21032686364215936,
 0.051255753527218006,0.23683854024484574,0.4844009365827592,
 0.4173985346787097,0.6023398398342937,0.003858001261088484,
 0.5502116368324101,0.6226465333953783,0.7157468890034788,0.8178971898120041,
 0.3923344583606351,0.06451513222319893,0.5256836509029322,0.1421543972763777,
 0.8048485004713771,0.8550906514110712,0.3751185296123276,0.02396403028844707,
 0.870044205328913,0.4452747925756152,0.813241645811596,0.8068588513533473,
 0.45147540030077904,0.09047185327786056,0.8438308490168618,
 0.10301315118674004,0.8697897258289815,0.42398245822383207,
 0.7993996054200164,0.5544244405415738,0.5253874057695213,0.8446883567801761,
 0.09190232209555715,0.6856002976921134,0.5674224235999829,0.7251960799854308,
 0.21645663310262142,0.19293208805325845,0.2521684813689119,0.27944720573587,
 0.4545519327680012,0.9384377711996268,0.718219522022851,0.8659444098313295,
 0.510817426626701,0.5780356128345274,0.6265559918550635,0.1729956565659928,
 0.8618566826357733,0.7192038725133535,0.26508508843322365,
 0.16843262082115595,0.03308558013331542,0.9214527059208042,
 0.8827095540753344,0.33333995174899334,0.13375786694963365,
 0.9946757866958397,0.6755218266393236,0.12423294329966494,0.6674901551621082,
 0.6768567405384782,0.7527709650132349,0.8448452300089495,0.9227313303361032,
 0.795272708783056,0.3890396648511447,0.37436280066868655,0.5165690333892721,
 0.930725782113581,0.22557486156358242,0.8642717855903045,0.08109124568127979,
 0.7214870845178635,0.012012655916592285,0.4583572781963292,
 0.9593082478622319,0.8011222444863867,0.9596092245456387,0.07636053713246638,
 0.051186843985273844,0.6177122843558924,0.9077654519058759,
 0.6661729249724854,0.41067000116895447,0.7193985692040001,0.4727069466919228,
 0.21210400599452306,0.7485582778970952,0.9917790466172144,
 0.12970116900181117,0.43921289148772324,0.9665564821069821,
 0.7272748521960627,0.4511002250658145,0.11740819343465647,
 0.49649341426577154,0.5310466845305226,0.8511103885996864,
 0.10532435649375771,0.5769285872758564,0.5834721619957282,0.6887807731325823,
 0.619324092518616,0.8075416076803502,0.7096709181936499,0.89050146268936,
 0.04668882511334227,0.8517188891517049,0.598294445119819,0.3372853286150841,
 0.002495599179998087,0.9571394131498656,0.2791355700404685,
 0.054198851478866716,0.6276438835545198,0.8703491333941129,
 0.5252490041274115,0.6556873840741781,0.5518535814822033,0.5075241239595448,
 0.214657806343016,0.31030256892941344,0.8526041030305478,0.5252144590450379,
 0.11266651542596917,5.844201941829219e-4,0.7361599920552289,
 0.6102410486102468,0.12728895440527566,0.4537609316667982,0.2382419116402349,
 0.0959630905691028,0.4640155490818958,0.2251940824931804,0.27162599828989564,
 0.7857902480880226,0.37165058952261487,0.843888532403341,0.00577324738863072,
 0.99304702823098,0.5501765543641339,0.7084465783252664,0.1017650182259352,
 0.7746724948078149,0.7049803746195129,0.3438080350982766,0.44404868988755075,
 0.5523178609376875,0.7381524239713495,0.7536671853037348,0.08900854253141433,
 0.6811595409838507,0.6904599990255671,0.11583307267954379,0.9641819622784414,
 0.6410429932207196,0.6227983275338066,0.4725825882229636,0.12970183466328788,
 0.2062724196410316,0.8080319609707123,0.018640792180347532,
 0.9611772514849048,0.8446771210824625,0.45376574842178785,0.6006928641161093,
 0.5231012156046666,0.10578897832551581,0.8171913827667513,0.5562594829111622,
 0.9422565732951251,0.3018646979781233,0.18280223365344495,0.7810289326915895,
 0.49888478281130877,0.4795899557461626,0.6703830787607359,0.9787868202445943,
 0.5383151080682783,0.4949761564212072,0.14916251653338242,
 0.34172741261131756,0.08750452716440915,0.8173414724690653,
 0.4192932033097265,0.3693721069386927,0.15777539198231494,0.6701170426516132,
 0.9277417354828876,0.42983911791008456,0.1269492461519489,0.8645071316735929,
 0.27869477297248313,0.5926329685708763,0.5311506207595982,0.427891133115595,
 0.6834352003505848,0.19046354467305238,0.9605829160687078,0.6864835385971586,
 0.35077129659592043,0.37206255932456933,0.47735186403632457,
 0.4082411572331598,0.45090997692268187,0.6711246017518321,
 0.14755977818783583,0.45063539106940664,0.6021610028983491,
 0.05237882559553697,0.516778850251808,0.9096663913553226,0.7951626662321198,
 0.21620743727507308,0.33852702227608067,0.5753794750967608,
 0.3739368042271861,0.3087145656534509,0.9783174792906109,0.07518247524205024,
 0.4286257030621362,0.024357992772866432,0.26801701504618325,
 0.5903977496010822,0.952048654846062,0.8062717423213833,0.9921599848389526,
 0.4023515105067377,0.3607016868813757,0.9951234507109791,0.4021740413040792,
 0.0801297093091291,0.37927275259865123,0.2595845512387497,0.4391477336014534,
 0.2068473228933646,0.551766251801867,0.7670407633688279,0.16654748270274056,
 0.04742820598126485,0.1003647042189586,0.5613926242769185,
 0.12955673776804344,0.3567725505342957,0.760689467610453,0.03754608184906005,
 0.6065008486415416,0.11228484528968052,0.07412649505368885,
 0.3626597446217865,0.4975193573900317,0.001860652605449875,
 0.1714793147371415,0.7853003791454439,0.5530407097913287,0.8420692921758087,
 0.9077956486365578,0.3277959465054133,0.5296318904551939,0.30030583998105154,
 0.8368465263961969,0.08551063737251197,0.7918999103238272,
 0.20238641613668684,0.004707219266617102,0.6463799920900617,
 0.7230487038972434,0.4717756942377158,0.20456665170590282,0.7946429122488305,
 0.48218593872882387,0.18648898258405877,0.6097708531583806,0.72159918694771,
 0.7755944485059029,0.32640372598810197,0.54338758992663,0.014214846047755025,
 0.8981876110321501,0.16608511366328005,0.4615709589442333,0.3314956378230798,
 0.6638235507157496,0.2247494870387231,0.35396683905091264,0.5062986330410966,
 0.9470449925191646,0.19961017098157863,0.49302054583017063,
 0.6357940366477492,0.8666960609584546,0.056744974810996385,
 0.5196667820565526,0.03626026409765282,0.729784916554008,0.13652620815620298,
 0.01727794112698855,0.8651401853062386,0.7595613523687086,
 0.24042907158674398,0.8022261734935471,0.4865068675578309,
 0.16085860378543748,0.99740611898446,0.840643036149689,0.715849886730901,
 0.23877826729884133,0.24500388282964514,0.9315927945232955,
 0.4811002460702184,0.9566061976864307,0.8095030973054618,0.2011147748784088,
 0.29280613977791003,0.7129160931239821,0.3080726399706719,0.994043720865771,
 0.033172152725823634,0.47471104588607893,0.8294931220710137,
 0.13992966228100367,0.8174053237037698,0.3117261800438824,
 0.21199591840896326,0.17075130245203185,0.5836088644362244,
 0.9851210711587793,0.9617438731288045,0.13212436922209014,0.9333219261176711,
 0.8926293006356645,0.42422394816087705,0.24347797050700715,
 0.23616949979589397,0.13647293708832475,0.3110096633898449,
 0.27944332439724184,0.48774178268816937,0.08257485362927297,
 0.12133350027515444,0.24139501658804152,0.07919260241058168,
 0.13272153592250868,0.17519785347560224,0.7006927856574806,
 0.17679416901191203,0.5683278505207924,0.7166405614883203,0.8377413885632237,
 0.9006112806470322,0.8728243683379439,0.11611402744358124,
 0.0028518608244378907,0.45428738265154067,0.10802474415132868,
 0.878234000332348,0.24916715511143572,0.9295700953528471,0.7622910192867759,
 0.018654204591678325,0.03526910195349109,0.7829736218979748,
 0.6088084690521343,0.6056279322233983,0.7533291956301704,0.7299850130862136,
 0.41867101184452626,0.3455153506542378,0.7990216189357201,
 0.17376769621426313,0.6085528680511166,0.3529628866604886,0.5094301833256614,
 0.5375360218998702,0.7864010395288838,0.49327466560895417,0.9880147191407903,
 0.9741473367326425,0.5208730307629311,0.7197706454048698,0.5946301566675612,
 0.10333358368874301,0.8448325910460589,0.4412427924595578,
 0.33016811921975686,0.04524860715335144,0.23202410068884394,
 0.19386078568656162,0.046980237147808857,0.1988094140649319,
 0.4339996331752687,0.6796220562046028,0.5879017485995236,0.8669982934683403,
 0.4713750143277867,0.8089011772969096,0.43310519111043133,0.9199415221798831,
 0.6844209423558675,0.2263362878986661,0.15973999409246198,0.383278615193303,
 0.06869584683113328,0.8453521886131078,0.2079104797005733,
 0.013609923739302365,0.2337336413643759,0.023666737359353207,
 0.5404869368082886,0.08364741657689612,0.35206982739942727,
 0.10367210059080689,0.557028035371239,0.034152827068351765,
 0.003316079155082474,0.6933123923261677,0.9392303689796486,
 0.07753697694601991,0.014906184346702545,0.9075816387970117,
 0.5863722561968091,0.3826187316185373,0.44814873125551113,0.2857194138509158,
 0.7831635461312081,0.40675564779640294,0.2924654085199456,0.3451700053565755,
 0.22898805251618148,0.01583238628992345,0.027310525653091333,
 0.3386681510369811,0.9800855923843789,0.5360057658359657,0.7775811207492866,
 0.5918129495589992,0.2754385764999556,0.13649474385896532,0.6999448191812465,
 0.8682864679828032,0.6287068310055659,0.35780303365712407,0.5430051547856641,
 0.2261771439798519,0.5285545783355597,0.8077217526110417,0.25300196086134275,
 0.964087204813812,0.03608120853653507,0.47832651850455954,
 0.42099920916408906,0.2872094409279754,0.7604453215048628,0.6916858700808988,
 0.4150982745059183,0.8545693886373483,0.18378880182710733,0.8740428031804542,
 0.8910895936464214,0.8992464689573434,0.3216831913491267,0.9059916555698704,
 0.0025574697891099785,0.7172276999523715,0.019701819608677784,
 0.7347713457645341,0.9258147763774417,0.039857676945579956,
 0.49456278252752306,0.40246021021305367,0.2493094931436457,
 0.6114067294734749,0.4113715319523631,0.4516396822714771,0.8882429997570929,
 0.3746280777196419,0.696372427058001,0.2413641584085483,0.967446198608868,
 0.7134486098422974,0.4396523972611356,0.5644067864444936,0.3037107583395263,
 0.751047264340342,0.2299733461232315,0.059226764020394285,
 0.13352235962190506,0.022535187458424755,0.4829961354263598,
 0.8851564870500621,0.8884087148163062,0.1903017961721738,0.8743294944568913,
 0.9835280902562894,0.21487458364533407,0.14168973549443997,
 0.6227138007942867,0.6394426564208591,0.6410959745526796,0.18001414497585166,
 0.014201397952283146,0.6214645524583361,0.26509986834754806,
 0.46180788443225673,0.9307208728700048,0.4402812578157411,0.2980472552413038,
 0.14120730437190232,0.8747544773933937,0.5686680275673923,0.6782347759617938,
 0.5413524517575736,0.8894557171286578,0.9905212469621069,0.04141837703520357,
 0.7526368860728795,0.28242379894187597,0.6935465046390168,0.3504950265991993,
 0.5960374078471241,0.7410586886005709,0.9369474757705633,0.6628207754326899,
 0.5310619121011637,0.9535197592518729,0.35946450954554066,
 0.020113721988036737,0.7564221979079211,0.07675317587622166,
 0.8816299766757144,0.298551623500683,0.04232513902500401,0.6760579877361513,
 0.03951435489084265,0.006720614620435161,0.3784585499515394,
 0.5026101427812414,0.029144131155590447,0.2855662316567704,
 0.27927307195590867,0.13123755340060195,0.40664406428170485,
 0.8122649455887954,0.32124055514672967,0.07881182315985602,
 0.43484236423440603,0.3897621090637472,0.13879588072020776,
 0.8029757140816133,0.1622347449251743,0.87323276007508,0.44678918873361395,
 0.5926692395965588,0.1709639987787902,0.5015657701849756,0.8241069821785438,
 0.6960995480414092,0.7300711841744844,0.023190013415331,0.8988004808044052,
 0.06416060314230077,0.13984154591901599,0.0689901561409717,
 0.7129607268680248,0.6358063447716267,0.6200486473658667,0.5441797951452373,
 0.3554274766302966,0.04094068692195396,0.5847587844938427,0.1613843505616661,
 0.3897121609753359,0.2049416560377164,0.5476998372896107,0.10297901626509942,
 0.9567388943864013,0.870231308039645,0.21544529267103907,0.10616307186770768,
 0.948346845759976,0.7619861470102165,0.2870936421402819,0.46446287825084787,
 0.9188587510062831,0.5555198601416591,0.26103403595694274,
 0.16729537569093977,0.6202802785300738,0.5318429387785442,0.7135799693777651,
 0.4349092259181031,0.8790218557265747,0.5437882773849005,0.7837906871942,
 0.7145422573115111,0.2608352078980627,0.5520627955561848,0.37189188091212844,
 0.519093258750601,0.7855462011692853,0.3525258822894235,0.2169686449486744,
 0.0019095739252066313,0.8011324303882351,0.42387776546959555,
 0.419150536884608,0.6944284889473302,0.9474418555145062,6.883416351676219e-4,
 0.8688885326314336,0.8007236245187219,0.8638916234004784,0.978321327393585,
 0.2259620524498881,0.18235422482756702,0.1696625772799315,0.2711421115377164,
 0.9979416309826012,0.863551994797189,0.18755925466932766,
 0.005440314103443589,0.6070551848982276,0.6566458012702632,
 0.6819057711416483,0.9697967938347278,0.35553153128125403,0.9207958741846806,
 0.03837834291025888,0.25404808897845643,0.4901279109589216,
 0.5831861406930832,0.7292321312704614,0.4575904697571219,0.19902627356404712,
 0.3034579879293192,0.09088521610640277,0.49231160386006667,
 0.9418449640945197,0.2564153202627102,0.30937112954882506,0.8727617264046761,
 0.21417807761094265,0.09214835324112547,0.11871174032319043,
 0.2452754994277775,0.8335939601338379,0.015007878522476092,
 0.9503209384682012,0.6611758689630741,0.2743111437896206,0.9346955509458367,
 0.19816804834482893,0.45305582197447025,0.8928783442978436,
 0.48110920768190746,0.8955132579658477,0.5085383886324434,0.7180407822177034,
 0.06469773826957004,0.15417725254319503,0.9769619870815879,
 0.23110081153877104,0.17938565707389476,0.3275151826917574,
 0.9002789614445799,0.6419963094553898,0.23651029752027286,
 0.07533822880536034,0.536311731369243,0.9785107987666419,0.3373965023907557,
 0.9681085913307274,0.3731645188512649,0.14109657907091044,0.7695282240103457,
 0.3174554964924371,0.686297347557603,0.019130189881882753,
 0.34586575748644843,0.7821093025291942,0.7091524243059284,
 0.26608184457026207,0.057582580496724045,0.524757313920454,
 0.0329907779145211,0.513996264456209,0.1083943394363851,0.4426789793122706,
 0.47468147265895544,0.03614260399055502,0.8127409337974713,
 0.8966548489851425,0.3426096501596495,0.7576872936760951,0.28757326975547737,
 0.21114087541616877,0.23383105956350503,0.8106312119615905,
 0.5915131051902319,0.5996345532262071,0.45046614003529717,
 0.22434352889506182,0.12237501976349552,0.725285569279144,0.9587736326796327,
 0.1427850475088247,0.5638732633882388,0.9294537499696462,0.7684136419615892,
 0.80084054891963,0.321492642192973,0.7875719245608154,0.5075642072055018,
 0.11827710304075512,0.5652736883322946,0.020173157822893084,
 0.9726361882912162,0.4300552187849387,0.5168866985510434,0.9827743186756589,
 0.43195192113357594,0.01676703792548495,0.1660337533092211,
 0.9979721054692687,0.27787238891519195,0.2363527303534403,0.5936633034840082,
 0.4861291408486281,0.017919941853096977,0.039158923556698255,
 0.3928211061643829,0.7706415939303947,0.6809660295698532,0.6667895334915193,
 0.7554214541744474,0.690738442964516,0.8748220950368298,0.5186240116446483,
 0.06670841994841104,0.8173601018624361,0.22916853021236094,
 0.13623270624382378,0.8652685732834179,0.4344187190885511,0.1361541655982691,
 0.8541005219055113,0.31735119594771344,0.12674749022393905,
 0.8757103333676882,0.5955383176204193,0.4674153454823986,0.8091464046251146,
 0.06817376868505565,0.2593023663395585,0.6780754340642802,
 0.27801567649687653,0.2926419521566658,0.8557574750546173,0.8397627973022439,
 0.2161706269158845,0.6215300711149236,0.779848970153015,0.004019788885003095,
 0.8843692418366637,0.8472089565324356,0.32560188858174277,0.6983648993889252,
 0.0802897638304666,0.4393185475121707,0.7748171925766624,0.20472287017063096,
 0.23709349274331815,0.30522894329301953,0.5164954705998909,
 0.03428640432529695,0.040261505622686156,0.29542598269211306,
 0.32128984050688025,0.8293268249967514,0.7052051970781971,0.8098909723789431,
 0.7678882318918039,0.7720492057797517,0.3460420395539735,0.7503815658449052,
 0.8661061291539791,0.8545474026572323,0.5173219409637071,0.3202837332061478,
 0.43485792161075465,0.6053963540284475,0.052116798558765054,
 0.3706896073914395,0.01893300377639684,0.6273090936791222,0.5990666214041248,
 0.7005832330082759,0.8487822889062928,0.339230418071233,0.8799857968929663,
 0.7262065626601522,0.08449918001417978,0.19063320818002116,0.93104585239808,
 0.7443466032943026,0.6866360353517322,0.2646111830609488,0.5460214317452561,
 0.7542244951409304,0.3542685960051595,0.4464958020937756,0.12917454794379102,
 0.46239015642076964,0.09340398617840251,0.3807526840997453,0.390735740688403,
 0.2704280875698112,0.7388155788008741,0.8254547238509724,0.5323480031676876,
 0.3513191033739673,0.25062428058260555,0.4454615357469417,0.9784545839424056,
 0.8474362774916552,0.41978501308054816,0.5977497922889022,0.4311302959950707,
 0.4393050201181057,0.5498311516306142,0.052335238275674656,
 0.6124221262504588,0.9060471886460315,0.16419762617103284,
 0.21388516713706773,0.12379864857315459,0.11690961803698496,
 0.6900147219688857,0.5512611248804,0.6140137401350962,0.1317873738761176,
 0.44263554027997987,0.8485951449860424,0.9071848422698301,0.1790867902740253,
 0.44064095208459086,0.4050769169487898,0.8177220671345714,0.3333362158608557,
 0.8694948278448313,0.5353523102437391,0.40745908802172404,0.3484438307144345,
 0.5591341603974722,0.22580994350003425,0.45613638346685026,
 0.9191018747142333,0.204055054958588,0.601920972489172,0.42278009130090155,
 0.0120528614486769,0.25744546120391054,0.4985945245670904,0.6375444484713697,
 0.03344412178475775,0.21533596600013172,0.5447188578537071,
 0.39059150503635376,0.17870727633342953,0.3760477158733244,
 0.3688282641901688,0.5284929570760213,0.762446359669484,0.19072007875313313,
 0.9924660137327059,0.3423907560908309,0.8559898268942344,0.9337513133524009,
 0.2662491249168675,0.2957119737907433,0.7754976204297803,0.374759922142448,
 0.39212420541038473,0.8581546482070943,0.7538127584725346,0.9425026150595148,
 0.7880539933050084,0.6085419115512094,0.23671970023413502,0.9684430101834728,
 0.5757517647858815,0.0885920799977099,0.3055259613344259,0.24269238413281014,
 0.19684648008121153,0.5354845472183756,0.8948946807056452,
 0.12172578496173458,0.8897993202041125,0.1641651101836058,
 0.30891279828586315,0.5720688354692124,0.38413936903000945,0.627083457202864,
 0.9356545832488267,0.9704520363114789,0.2699803886983887,0.46156628147173207,
 0.4500130446534092,0.41341482981778943,0.6132583926411007,0.6715090880679997,
 0.5032454285861352,0.4062962271642786,0.9535608434094468,0.11645432008105117,
 0.8603378538127654,0.5291801922244165,0.055410007697283326,
 0.26631373180922313,0.681195925799166,0.20444830203356312,
 0.42121951285332027,0.6208753627148247,0.44342672166126995,
 0.2436569847911646,0.41002880892738913,0.9590916174180155,0.6377773558812422,
 0.23000718108271634,0.037310847911090406,0.8812862118843392,
 0.5480010758335762,0.1713868485462564,0.07835309287026582,0.7441171334018283,
 0.6354880201864018,0.30994061605940637,0.9233975002650301,0.5591987248843086,
 0.5816856217556048,0.28379370300899665,0.17891652386301882,
 0.14379406619723745,0.08104311953806653,0.68047708948033,0.17024210470665768,
 0.2649709541963229,0.03024110863680185,0.10213050617562791,
 0.9597541458737124,0.6096209438129302,0.12617141289554767,0.7262071743638601,
 0.9610247110331785,0.34387362462934246,0.3313625570678995,0.5838601622357098,
 0.8171979918240678,0.7272020630905761,0.03733683448861558,
 0.07653072887352863,0.243197187785258,0.10895947086876034,
 0.14246325597082532,0.871390376412834,0.8789960276593736,0.6295432494382714,
 0.1599837921823989,0.07553363464342633,0.7516731997226348,
 0.26929751282185044,0.40915535883761944,0.8707361015658719,
 0.12197559212773823,0.365381436526794,0.18783733903325894,0.8438826526532979,
 0.9721692741211393,0.7182143257684954,0.7798034697098714,0.11902976768182139,
 0.33829605457629297,0.4683572142532342,0.4697007899315617,0.8868064573341202,
 0.259694613075624,0.2948126067928213,0.8607052416162909,0.6466380093290554,
 0.918384555104975,0.6887400600798685,0.7786491706756413,0.28539652913697855,
 0.808828562658878,0.55000896038058,0.7981517637505945,0.19656004279478423,
 0.598055139564114,0.7522548756360417,0.5216149109314633,0.07649046932565573,
 0.704234523293298,0.04661433578422047,0.17391388714995448,0.7688777466823524,
 0.891268961269515,0.4407413169983465,0.0225311790835645,0.08646590560288558,
 0.8198985337305202,0.5149853584851383,0.4069490741944537,
 0.025753780441017504,0.5647405471856449,0.31913912975782566,
 0.03171253910180838,0.7690680688846762,0.1957640726448555,0.8556492832576277,
 0.38243356309977405,0.4550899904654284,0.13404404392099667,
 0.9560477808472243,0.9764434218226891,0.6431737471677718,0.16489268308756366,
 0.14480569791485187,0.18276413256669566,0.7906499888726573,
 0.0866358979543922,0.3826715780772121,0.9807990850812045,0.463946057501028,
 0.4886848231940426,0.9658613465771491,0.15427008811157017,0.378678035975612,
 0.6471376842761352,0.5710117520710821,0.8235661457341955,0.4642927526741675,
 0.7222425321867023,0.12913258594634103,0.7073245366418284,0.9686464852431672,
 0.028368228343186308,0.5894819820689298,0.7331880842420745,
 0.33828751805076174,0.7754742385225679,0.5533279410173318,
 0.37717262506202964,0.9170790524629244,0.9833980074987103,
 0.45847896786219333,0.8483715388607665,0.94565452758128,0.5242975916014666,
 0.8512094754364393,0.4895148308663392,0.23341868552921108,0.4825994792939756,
 0.5285450903798758,0.6101522280572775,0.4658325324074909,0.7081270872242902,
 0.7156754470492491,0.06430935810538341,0.7248765696343634,0.9833408531871053,
 0.45011497221069063,0.868625390756973,0.9383518032085261,0.21663565428016063,
 0.9095907286634086,0.7733496786440305,0.8958394789675888,0.12988570168998337,
 0.5544819362975261,0.6355461941210004,0.7853451355908737,0.5277672813433294,
 0.1093326386411293,0.15525593046840325,0.3515916861065308,0.2008807742767844,
 0.6103063714870535,0.6047596792884133,0.8682712880660285,
 0.015389598505871072,0.2933959638670497,0.6760047940994577,
 0.9346863715889501,0.44996233272032427,0.2696757305258597,
 0.45863321036347593,0.993083971985433,0.040466710291077734,
 0.6403965204725176,0.7922605391160733,0.48006448905235877,0.2802093035328953,
 0.43625183209423446,0.19038122244490485,0.8389736130529644,
 0.8710365060707118,0.3536814536094305,0.5266361565208948,0.9983516284977341,
 0.9358271584052454,0.6568764982442455,0.9157065845480759,0.07788959161865872,
 0.6220205218099975,0.9231495909391749,0.3834247941496888,0.9812711881958036,
 0.7775454757652169,0.8256461239199194,0.5442413036992599,0.4232400335462254,
 0.4192252950888833,0.4628388233838502,0.7708404164325365,0.4707247803862856,
 0.6225808339362074,0.011385235246055059,0.34754653766669097,
 0.2994699101679469,0.6588737862956661,0.508429819743954,0.8739868222896968,
 0.48208805649523145,0.12333568748741264,0.28858178082469166,
 0.10936428602393744,0.14137358388458998,0.4635873657689564,
 0.8015690530339928,0.4841906312516765,0.12945470881323695,0.4859921953313977,
 0.6018893688939734,0.5050172599277863,0.8301562542333525,0.9123328796673809,
 0.7370990603176382,0.5401329843225348,0.5639171459623943,0.9941673231020075,
 0.8642748716240258,0.09642993105993602,0.6967217663171065,0.5224882478096081,
 0.6555977764127777,0.557148219983793,0.9339914144181093,0.5587529321317365,
 0.744325584666598,0.8900102009470312,0.6621418976583868,0.8799654544522875,
 0.40376252991103745,0.06616329764900086,0.7567240310033618,
 0.0034202477742387716,0.025633297256264065,0.009856451265093158,
 0.8743782482762872,0.19746268484979457,0.5895737644587061,0.6608748440500576,
 0.9041076920751294,0.5162253045369162,0.5902448030726293,0.36977105291049106,
 0.31485710678548817,0.71979279773788,0.8822151471589703,0.4356743527777299,
 0.494710797645285,0.9693994415658341,0.6645870491536142,0.3321529484372705,
 0.016875498966874458,0.5592244902192436,0.9883431953852255,
 0.017793306864770764,0.5029028165749541,0.9704396733657531,
 0.2571441352270387,0.7738880397386515,0.8791106602947057,0.8705951358694115,
 0.8209411339638172,0.7949660334264708,0.41401929462830833,0.6597764059552537,
 0.8734013375192002,0.6502767064370916,0.841272058650471,0.3758353925581587,
 0.1218517381262263,0.16930015096162776,0.10190818394018641,
 0.1966742949343674,0.2076359212808997,0.16553737701810944,0.3209049548042835,
 0.0421078151713109,0.5089803430886586,0.014641942542120256,
 0.050235750338154195,0.9137154262676866,0.7818422734879773,
 0.05384451477727226,0.7633323767423021,0.08334316041500889,
 0.6189038361955139,0.588377614910735,0.7685326248819369,0.12493830930839922,
 0.0524422090392781,0.948146428793528,0.024967870976605466,0.5544174632436596,
 0.37367717076144513,0.27474625413028264,0.47762846241778933,
 0.334954210121712,0.4908594544469471,0.32208028519697174,0.6991682424711089,
 0.9596490996674542,0.5864375244417731,0.26044050589374645,
 0.25158369557726323,0.32400520913121433,0.5230357438723029,0.456517098492784,
 0.7420077429418863,0.49796649520191605,0.9609685984713304,
 0.08282726618165748,0.054953528794791096,0.7178624252316878,
 0.09642195038171131,0.33504808979867073,0.6044646557998736,
 0.4402565968359424,0.042003044034699566,0.238225192615933,0.2972105947029071,
 0.1840621680672948,0.7555241272617916,0.5872711332067196,0.5408791668516637,
 0.46322412687511805,0.6631063436025884,0.044420677820317445,
 0.36196430718134454,0.8211286604428067,0.9978945487626892,0.6932439782444731,
 0.6376513728156064,0.10271969578349915,0.16963893055298596,
 0.8628931954952811,0.5495373324746561,0.0722606150396955,0.13017204890498446,
 0.965186101950899,0.317388602768762,0.4677874610605315,0.6482559441161833,
 0.5608963185921864,0.9651467327525818,0.9158195752953378,0.9652501025712493,
 0.862077646298354,0.14484053513621176,0.7239877746698975,0.6523588106501034,
 0.592495910601649,0.005494491686301006,0.9540970813261169,0.3058390500753446,
 0.9544355291391223,0.2203291234786483,0.19702035672108043,0.5793298895413088,
 0.2290456450522922,0.8797025454246653,0.435603070444729,0.8240304680546201,
 0.5373954699219838,0.7874508060274316,0.9084615913388219,0.8526351830139851,
 0.2591731274505439,0.5312128790872404,0.4957339522919548,0.9384791557997939,
 0.5727932587645508,0.43915758200862354,0.14112036918147142,
 0.21837773979681063,0.15986893613644826,0.3896860688187125,
 0.4278488329922908,0.6098319588889675,0.27522384041661396,
 0.17519320239872027,0.7274751900320479,0.4301857599442871,
 0.15335900832642957,0.38117060984606754,0.17208684555398435,
 0.0073980165247453655,0.724681583330612,0.5642790214299459,
 0.10747695287444836,0.3269752964635788,0.6885885947373679,0.7344447986194419,
 0.6436247432754756,0.6866459677541417,0.3771702628208212,0.04147127635739112,
 0.8351558871105164,0.5035545876208293,0.023332212721438372,
 0.6961128043167983,0.1235806172540066,0.725991819259507,0.8890974030651307,
 0.00961005222135558,0.7442997371650969,0.6813490712900739,
 0.19472629377059925,0.012827868272945575,0.2447447338797918,
 0.7287033964029307,0.3356809325808219,0.3350951355988443,0.5165601550717883,
 0.6313188800142396,0.22206735980337855,0.47820666045583426,
 0.8310914679126479,0.6153932257514455,0.891975655316154,0.3334792655099599,
 0.8692354809251089,0.16504686260754764,0.8978549533728397,
 0.04970061770183942,0.272337999105952,0.389578020676391,0.2854621006403636,
 0.8953650086489858,0.4680523559170192,0.24839379545562257,
 0.39091175493006347,0.8280568614759654,0.2973566938505948,0.4451296809475839,
 0.5651773733639024,0.01799175066794645,0.5923806917374442,0.8595215317167209,
 0.5136765788720672,0.31383869394598074,0.468448935092454,0.0738162783242946,
 0.8407298940179159,0.2765340219613328,0.6833929741202692,0.4114138918814121,
 0.03823462052671023,0.6217216383300204,0.5254961183390071,0.7771159517854339,
 0.5267657054943288,0.14429942215382818,0.46032027138221976,
 0.9571484085969884,0.6572471539309972,0.42738909590555263,
 0.042910025189402234,0.08900234977313992,0.01577012100629671,
 0.05984952644145258,0.5275712751629376,0.9112460574691533,
 0.016407630186820232,0.1185453624421331,0.9590230584789419,
 0.18599468214547188,0.9469155777934337,0.27645820677829525,
 0.8249104783069857,0.2925517231768888,0.7987710421076517,0.5336371659744372,
 0.0735523617514372,0.05276290846329079,0.6517538206909614,
 0.49081793626276915,0.5321926559194106,0.1678889731259866,0.9517206890986585,
 0.7102372161882121,0.6340780135591386,0.93543339958865,0.8222124615800477,
 0.9214061431632353,0.8910175859719534,0.760342230644311,0.29272001846398354,
 0.044151599046559076,0.3598123507843972,0.23032595772971787,
 0.5914780342877786,0.36586502591131964,0.7693847871793684,0.648675579387114,
 0.19988070468740682,0.1862263886466834,0.9535363997074808,
 0.24697073173424944,0.9224334819671048,0.5181112172997366,0.5675814262548964,
 0.5546814269802022,0.24014424883688756,0.5671671368310375,0.9430095390360729,
 0.011401889363930051,0.8760795527321106,0.30298914555834167,
 0.855783588260333,0.8495996942169821,0.17038332969252745,0.4461546890668737,
 0.41302741108233876,0.9762008916829434,0.5362947967817197,0.8295722792754207,
 0.042731787443402,0.30185407059486546,0.840658522205564,0.7515041902519981,
 0.6275419918562921,0.7552810143196784,0.1534968541467352,0.198252506463364,
 0.4766402712197052,0.4459031275879708,0.8502162131474224,0.43654511893554815,
 0.37792183262953916,0.28605344496341256,0.09698484387051953,
 0.13941774367981574,0.8214230453577402,0.36841175898257894,
 0.49001893356567305,0.3262405259852815,0.44128971100434466,
 0.15165538037828086,0.6687292814919472,0.7614360337150344,0.8150787485464249,
 0.2260273472951677,0.7917491923253503,0.036255157580324715,
 0.12263724744911353,0.9436684893157712,0.4898831245369504,0.9942959336662738,
 0.8058029575247757,0.44495970413319486,0.6388085544232736,
 0.29859784058228045,0.6298550810961535,0.9657427002720691,0.540485469152808,
 0.9592316338221301,0.14142453525073462,0.07700374915801966,
 0.1783327051619379,0.3648221936685827,0.22227682735786347,0.4769284278436695,
 0.038123928258543605,0.902724822685851,0.3452009279847662,0.803940969229148,
 0.20138501587757207,0.5910522202451441,0.5562377010742616,
 0.07399766109502615,0.09284189452040226,0.6437269906354023,
 0.30741520798219135,0.2607202144265608,0.5318846237189087,0.7208630043484038,
 0.49721422746101995,0.6283825056121175,0.402327356873057,0.44833803153687724,
 0.03667088284097697,0.32245696345460995,0.7768730626657889,
 0.30259220897091044,0.6178476665376387,0.9908973460125304,
 0.42526505734295017,0.5098652596345186,0.7586268107416198,0.6508402214889527,
 0.2175437130513591,0.9173637436297618,0.18229914022559546,0.6203542630476951,
 0.21987872208467207,0.5232330587796181,0.802060718506486,0.25718749470100033,
 0.09760943253241683,0.7957050981223935,0.31311114039729926,
 0.35617798385220734,0.10262148043273622,0.0593597167026223,
 0.26317495157072557,0.8409904498455711,0.42017817504316657,
 0.0231841189218831,0.19765419530309347,0.007886989172186665,
 0.4281068064314545,0.6801090957425996,0.5315660365815928,0.7840464843294079,
 0.11804305329917408,0.1429265595376985,0.054363889444677405,
 0.7021879183453793,0.2008763238423943,0.0913195245076136,0.707714783356356,
 0.10502238187305246,0.8381461521742684,0.11229660910992623,
 0.00594622992233651,0.11742272306103285,0.3126123884190327,
 0.9494326990055831,0.17901386222469406,0.18056703356027404,
 0.16774886192346516,0.48768086633630336,0.1765595360514336,
 0.19771800817924434,0.014804495819726649,0.7137724953495862,
 0.7058658460455397,0.4150272230164699,0.10855040235654168,
 0.04511301544615309,0.523388850458057,0.5946317311269466,0.8197494064290476,
 0.18386330128487627,0.8146609015992922,0.8533537698932642,0.4825990218080235,
 0.07338617387664836,0.7628944954075955,0.03698045371340242,
 0.4940724031157856,0.8306943942316943,0.8764292367257639,0.9587199313812168,
 0.02034157342622689,0.3962244665887701,0.9414495391269719,
 0.06297160499342058,0.5112460608999083,0.7395791392019989,
 0.33833504633078826,0.3341943764880264,0.6349235944239922,0.5978395031072635,
 0.0039034080076982836,0.5734750549881298,0.0042726449019135915,
 0.043311007265352974,0.9031818757168188,0.3865530468182694,
 0.6283224099446055,0.8841630209610526,0.7202831147517995,0.7889531164409647,
 0.6186655671111935,0.9666845977679922,0.8904991405374759,0.34018990687736306,
 0.5640597317709519,0.03848209403989111,0.08352360345865295,
 0.42041619853449774,0.6283660079906042,0.5671160877746655,
 0.29431068232985313,0.5997107338221203,0.3413042858704999,0.3698604098873727,
 0.629019139512812,0.9585028223805593,0.9170966908723359,0.45067037561238776,
 0.30953807795875254,0.975458010104824,0.6506703621041571,0.06337171885439008,
 0.5072598907070642,0.8359746012960134,0.3409737538057699,0.8665040715181573,
 0.6482115330350262,0.8367673051224052,0.19259499863309282,
 0.24126020570704532,0.819736283945935,0.24174858166095925,0.6251651948916543,
 0.813775535371814,0.13503202389235902,0.9124226900128587,0.8571479849820031,
 0.8147093146387877,0.03723372354706189,0.06712485888765951,
 0.8540540204597962,0.9098401362789688,0.7536329557965554,0.6748193073049032,
 0.6508414764811973,0.19860535947331515,0.759083595275937,0.7930223836545774,
 0.6633697645394906,0.8099617451476996,0.8067958839671104,0.04721249675968098,
 0.9113545353569749,0.5558012153321478,0.1865949931404023,0.707595251392003,
 0.06264153598902789,0.7068180686173465,0.07472597308482631,
 0.00931517643157842,0.660942660613914,0.8586030641858651,0.3650233852918162,
 0.789896199921102,0.339610905740967,0.6588185338978076,0.9946001234926691,
 0.9186575513888396,0.3036357574729849,0.8597617346297327,0.4588970546465009,
 0.6684856642769972,0.13021218586050698,0.45668757158592643,
 0.37797481702211855,0.517107332043904,0.9104525072165525,
 0.021061580855496143,0.12665676964029227,0.3923432173951015,
 0.3096680444823906,0.5678847734391511,0.2879489026890224,0.6289593051825847,
 0.0693587993670075,0.28846761453697356,0.10979585274208226,
 0.3564330464669524,0.26019754395790917,0.5827489836669641,0.9951399794845746,
 0.013513158361606936,0.5822209044750398,0.9884155113897485,
 0.9989923164822144,0.8003242541430967,0.10295893509202181,
 0.36731896430819344,0.43804748383764636,0.3250091089334083,
 0.08970016524098934,0.31911433446651116,0.8402797974928158,
 0.8827893472532474,0.842886629427742,0.11298746100853063,0.7647371209363779,
 0.23071943270275397,0.8595080107565977,0.1523579367100628,0.94581128056684,
 0.7950033718209658,0.34593930064521927,0.36566328260416947,
 0.18644260495015852,0.017004809259380305,0.8583964710968408,
 0.6822302455688528,0.47308787398323293,0.5561513938075137,0.6343640234718191,
 0.42395483487945473,0.7112473370081543,0.8388659775304419,0.6622368605439938,
 0.5787487680616953,0.9752641897088432,0.8389728238373401,0.22614762551597445,
 0.8651385467562573,0.7347645772143685,0.07700749761389991,0.9711816851931822,
 0.7943765554730189,0.6763543133342974,0.7461278449483916,
 0.007340092511052276,0.800930687054403,0.6365878913604037,
 0.30048750861309226,0.4204100610250905,0.28809708173872584,
 0.029178476015958688,0.6028993013246713,0.5352564780979319,
 0.5801824584956639,0.9979783318515077,0.7569897287556361,0.8461716211294268,
 0.9807137276538017,0.6416883667585983,0.43278816929216246,0.4016563430355341,
 0.29469546910935007,0.40640196488563884,0.5845023052402706,
 0.09842775574441454,0.4493021791520586,0.12136034537437212,
 0.9616301100434235,0.9045231567095556,0.8276265846898204,0.7807620911302758,
 0.8924839243300043,0.19242675272468024,0.8516896817690489,0.1914311440871047,
 0.9160700989126532,0.8172322782459585,0.5215419115473621,0.35494650808628825,
 0.13575543555739855,0.6618754760615734,0.3310381683223632,
 0.48269407721382684,0.8600597583925481,0.9882821833833652,
 0.10818138688481227,0.10010806252441384,0.8737282700847225,
 0.19822930691248597,0.9320456686506395,0.3224643277442494,
 0.12526295470707927,0.12596147577384798,0.698815948863586,0.8130712967306549,
 0.2719928367811615,0.4550950296969676,0.7119360039819016,0.66882739771959,
 0.06112350954487755,0.7374710130481911,0.9488153393231132,
 0.24069988746100857,0.780351788636668,0.4186991072288764,0.6765184588739631,
 0.6431725026538111,0.46607559481815397,0.8866186736106938,0.6882824185374234,
 0.5953389218701992,0.5332632113380124,0.27711133219084005,0.9022875839861115,
 0.924361604858461,0.8066127500242113,0.09011065217498315,0.6221153040012873,
 0.7849273633377307,0.6078888604204935,0.21133564311717623,0.711116275152425,
 0.9139561515816401,0.7663213824320391,0.33023180317717815,0.3818720809690468,
 0.8380274398609827,0.8600470641454873,0.9979061900053379,0.8729943453377391,
 0.46510998950868754,0.657633878402362,0.514808659889205,0.8290385908025658,
 0.7010096126041993,0.5041033928948127,0.11155737575588165,0.707044618308732,
 0.9252892974708643,0.1193579027608026,0.12428070068941988,0.8123164534373722,
 0.6857505893216954,0.16391441505196858,0.93093877475949,0.7546815422888176,
 0.783410111509971,0.2229961111219294,0.69611413730548,0.9873977876680157,
 0.17101949603918343,0.006284154535559638,0.6800346969725592,
 0.34268628532877154,0.66118735857426,0.5698264114511327,0.13950492498608436,
 0.4249780750273413,0.8440089472334673,0.8758645969734022,0.1295054985136328,
 0.0011835713890574473,0.4156427118034083,0.8232414927508646,
 0.9502577648952771,0.8505046139247565,0.8784452200218166,0.06235249387648967,
 0.05278217842001498,0.7720808941134369,0.431470276093019,0.7725211681065391,
 0.8090487429542885,0.1748446834624705,0.40014590250951876,
 0.026287966156824383,0.6794885158263585,0.33341567747614465,
 0.9536590567880914,0.1808632999542551,0.32198021575655045,0.5723584491906926,
 0.23098474661337787,0.14052000316570223,0.017061074555358324,
 0.7914777471992696,0.16011941707263044,0.3112655661520076,0.9427570952136997,
 0.5398190511178668,0.15820723414450588,0.22331209875417224,
 0.3403205544651653,0.6583132810716823,0.44380662011219774,0.9189963644893488,
 0.7342014275248322,0.31205767706075116,0.814180477635589,0.19623890678874367,
 0.6499788752214329,0.86972089636329,0.6637063749141503,0.15384413828602694,
 0.9106428756332924,0.15415096920444527,0.4930631934438311,
 0.09474547125290034,0.2218564420794904,0.2819142394042512,0.4388627907849034,
 0.045717253623975385,0.7491066633729316,0.9231767378375901,
 0.17805735378474774,0.5877164852603227,0.10262847233365091,
 0.022997255569158548,0.02560497384996141,0.8104009869823809,
 0.9972843767646224,0.6081135289938385,0.7588053130448656,0.29318466930172593,
 0.9001429787946177,0.28362798773573805,0.8115687270648992,
 0.48811977219953884,0.7064986804137965,0.8841937659307321,0.9741291806412925,
 0.1605993114338854,0.003554721849432063,0.05479952311757008,
 0.26186120804165447,0.07123020930584845,0.42804354037189685,
 0.06445783218799328,0.25563575680316164,0.7416314670732621,
 0.4140791394044261,0.5232275822527405,0.02934464560105532,0.4922507628433015,
 0.5598737045261397,0.5555007328107924,0.5597528526737183,0.9722278017822764,
 0.9775965868996817,0.4984588851774885,0.8143116518348985,0.0882045567651294,
 0.44772141040891245,0.5137887710876115,0.11705506567090262,
 0.9033892849221319,0.9588036615457332,0.4372345740657897,0.6904467494391703,
 0.2175824263822217,0.4241421238217179,0.6495149174549639,0.7840426980600965,
 0.17816309761852578,0.9972696309574798,0.0639874982152393,0.6299130361359014,
 0.11570925000904897,0.3226015252875629,0.5339435136834151,0.5145250089803342,
 0.5148020243370606,0.5154561722457854,0.4852029996088548,0.34889465524628893,
 0.5683121237095505,0.8987379189371483,0.7254903917608102,0.1331920705841274,
 0.20959389932875594,0.3672165677638657,0.07787566415069447,
 0.8362263072655424,0.3135144517009827,0.4849454411083264,0.4169543527624673,
 0.4283359673404612,0.14357028828150709,0.9911947824747596,
 0.47742160090404295,0.27538602647822674,0.737004643779532,0.8269507947851905,
 0.33327680901261447,0.957347415098953,0.6977275724347759,0.5194207275366508,
 0.6533113995559601,0.8666695212831941,0.29144942304267696,
 0.41749833876773645,0.5031705215958061,0.713017755883047,0.43065327577576806,
 0.2642811720102969,0.6225745478141284,0.4428883517382618,0.7377274085820557,
 0.15607160432642297,0.5446801041410045,0.7857605540792759,0.930495700528456,
 0.4609857400343209,0.2519995239813281,0.9261299909648871,0.8630945043913079,
 0.6984602999075739,0.8981751364558166,0.8990776508925662,0.0609073071471391,
 0.07347334416533835,0.44423604976795794,0.6889989214093177,
 0.3497126224843614,0.393550651894888,0.48917548124588683,0.9211827247383408,
 0.05834691163933303,0.4852886373654248,0.7266455685798046,0.7152365885812446,
 0.34692041997272505,0.7748429800868917,0.5070024554879188,0.2546664059776519,
 0.5988611892600422,0.2013682727776407,0.5669197650557412,0.5743240063523318,
 0.9624965081939643,0.31411609452814704,0.29495070789548616,
 0.4084891286066207,0.7112084710088371,0.8279378139968151,0.3265524805576452,
 0.6409290840387162,0.5096805572631774,0.4517210822206077,0.6912004111240188,
 0.2908094122708822,0.29553803954590674,0.8750969229157335,
 0.35739663357457896,0.9528583227242359,0.04906618776988081,
 0.22265300704466062,0.8918520357897456,0.2887232940799307,0.2566615370368276,
 0.23287422546793712,0.6496939065129321,0.296206077563691,0.34449592866263934,
 0.7187740111735048,0.12286465043789019,0.7753241483155373,0.5671912045550019,
 0.25324674628647503,0.571157696812417,0.5378002240446067,0.5088086529502863,
 0.5364559609685186,0.04019781073326445,0.6857259458407657,0.6469248013564279,
 0.8664752290844322,0.21538915218955823,0.1766921167923302,
 0.01877431764626436,0.8619545376875863,0.722234039033194,0.77256852131472,
 0.4655951312093758,0.3371480042393893,0.5716088472951807,0.8726144617319787,
 0.30435683403464875,0.7687323923644847,0.9704637616008633,0.2912247471586421,
 0.09128450594767856,0.06926616375796812,0.9919994645153191,
 0.29874213014607875,0.9157689273001519,0.777901476626026,0.1660279472116103,
 0.5844613181851364,0.30455855292469947,0.09602869560180305,
 0.33798650143184905,0.9989907723776766,0.7699632709355045,
 0.21839427543267664,0.9125416167280322,0.7781582389029267,0.7515632872013667,
 0.5721974040964642,0.8145508164096581,0.3533161542759654,0.7349105935457411,
 0.562394135849284,0.23658946848221551,0.9647638143449718,0.6971793509288227,
 0.7360073297489689,0.7331653508242915,0.6278360197518764,0.19035801167144994,
 0.36353794586147714,0.09576384654854408,0.9320643163757849,
 0.1532912242938228,0.27242338192320426,0.09035927283978873,
 0.7477319010263185,0.7616274347570687,0.8705159923640494,
 0.016122399816386745,0.4290880450913903,0.18614491717705806,
 0.8087726266703481,0.6291612061916995,0.06744402761653756,0.2945549249642738,
 0.7733300913987701,0.4786463323655823,0.2831748282966636,0.4857311056903626,
 0.059486419994963624,0.36590811936450207,0.97882298506232,0.5014460295578127,
 0.14650973431624292,0.7196533460012134,0.6503046594515614,0.4458357808569332,
 0.4545637453803095,0.6795264492826518,0.7276677042404534,0.9857540198028505,
 0.07191532224989272,0.011317877547427857,0.9075874067766181,
 0.35127694322067915,0.3538997497501326,0.7638891118432145,
 0.03232359141183916,0.6670952629948265,0.11526406307668013,
 0.14515593742857313,0.04599297944955971,0.17523440661904088,
 0.2474506159591865,0.27683212280301184,0.4668827174805765,0.3756170712912905,
 0.2098659997968204,0.767328502164702,0.4477210645821945,0.5696474669348904,
 0.15057558559414674,0.8418601848818583,0.37780788516859853,0.265985550772178,
 0.18020096060842783,0.5373079777746572,0.5175318740060273,0.2607247718321388,
 0.3969597497493692,0.6808507876084406,0.25477291100101385,0.7008132383683306,
 0.5934671918493294,0.33601361609229397,0.9533717741734387,0.6935132601188383,
 0.27618042273826204,0.26832379974210163,0.21446867651991552,
 0.5700529099769461,0.15059916044679778,0.3381563843426383,0.9973412843068828,
 0.7248259141769215,0.09592964480000021,0.2291145533972222,0.6614107094577515,
 0.6066192355521121,0.8598627925006703,0.020033234064054217,
 0.8931470920107166,0.9389421538436892,0.023966910214327508,
 0.0785114304562331,0.8342187002327638,0.4847129288984169,0.989904580929108,
 0.7842867237219594,0.5536768386431723,0.05553003502109177,0.2517229067398442,
 0.820827107480406,0.557138743183067,0.44732281640108007,0.5935786409156131,
 0.802934983150575,0.6066885965307898,0.8594915480063722,0.009861704995299925,
 0.07077319094869605,0.29582857930652195,0.5315907788839409,
 0.9684501863761066,0.6812377588408022,0.8339582426215214,0.6217141924724308,
 0.84785334172882,0.4900967568892791,0.8249998438112139,0.7904927563718981,
 0.8945676898213752,0.4525437808388736,0.6522436889863142,0.3142978805834439,
 0.8535316725147813,0.35834210182597837,0.40588323831711315,
 0.1343590800754908,0.7303105220847065,0.18914858237502252,0.8853276749863382,
 0.22178532168620024,0.12271586753583841,0.9714446434716257,
 0.9569683531898876,0.5711459114954982,0.9212921513507271,0.9412372491545994,
 0.09858206832463678,0.10641728588991495,0.5773994217848626,
 0.07508594818317416,0.03420804418010803,0.5944328536982049,
 0.46344509871001105,0.4534687243050566,0.25944250381044953,
 0.9901708716425044,0.7192828653147965,0.42618897796941946,0.5649613436598419,
 0.46896617130912466,0.2278184459761815,0.39317936911242835,
 0.5524238195905286,0.43238909871348397,0.5383444069480245,0.6799902008418673,
 0.36941889266501926,0.09100989647484858,0.7165214561564398,
 0.4835555163338746,0.17456684359751773,0.7091510078024295,
 0.40953480346787297,0.27594344504335944,0.7249536899857485,
 0.45218453694487093,0.32167149743011225,0.07966455613187617,
 0.013744218766544303,0.1322158545396488,0.5568421568200724,
 0.5123977897682357,0.9375743246115966,0.6576111743503722,0.7892367238918735,
 0.7595660234276052,0.9026020094454925,0.6044852313063456,0.10299302173355518,
 0.4190763633357002,0.6282023985757668,0.2941166778064953,0.48349647041293853,
 0.23541222023186226,0.7238010511501596,0.12987858595564894,
 0.23424009225360698,0.6135939846603607,0.2233728158401148,0.5865986787563686,
 0.0016524290134564135,0.41201028528434525,0.7748861359125541,
 0.5785134995459902,0.9243732453838964,0.7902508658651715,0.10368702331698687,
 0.008562392239440775,0.0967668298515928,0.27062047666378397,
 0.8832180516253074,0.3803349840441108,0.5886299554025789,0.11392253766230509,
 0.01235614884752223,0.5490564759640293,0.45125667048232787,
 0.9508509573036045,0.6520497988250429,0.8083952048255711,0.9456165265932581,
 0.03031198319208106,0.06539018010577735,0.9997467707314586,
 0.34466261891975636,0.7561645791297822,0.22755848739535212,
 0.6433644291502453,0.05855600015188589,0.2839142456156182,0.6072342782700916,
 0.5360148201659869,0.7218340355640394,0.15775878811947808,0.9879101851740458,
 0.12045008298254811,0.8562096490494315,0.6437887711371943,0.9753642616306412,
 0.052696315428695906,0.6538211195877496,0.36029076946904715,
 0.6092322908143268,0.9078434653879931,0.3294386106379882,
 0.009503297667518451,0.5078367789618214,0.005248346643923818,
 0.5635139766160572,0.9785693168899078,0.3378460725011643,0.2783569210926037,
 0.34025713902498667,0.3128916185616164,0.05588538721197567,
 0.8431245839016296,0.4404124449943848,0.825427572433532,0.3849614834986428,
 0.9995778696938562,0.8316340018464815,0.8480711684207322,0.37541132786744313,
 0.1422671599367572,0.9583894733966236,0.6996422298848366,0.46055080010671523,
 0.13931529335329884,0.43954588266227945,0.7378698045432057,0.838210957054036,
 0.4872550619697392,0.08808251708877912,0.3126293139364572,0.7833601192412604,
 0.7754653707285689,0.5359229413710966,0.16892140932307853,0.6954740452672836,
 0.8870034638963945,0.4443102444311845,0.3391850423978937,0.9964095399504915,
 0.305856257248748,0.32303260085912666,0.8187088222219723,0.13963227631426833,
 0.3702041029793145,0.07851012389604928,0.6756727029542401,0.8272585752634269,
 0.29996922263762704,0.1849086169123918,0.1625513427160591,
 0.07970041973171937,0.7373072032724034,0.6113510522570497,0.4774605957898558,
 0.19931283204965222,0.006121817478295322,0.05262354440840711,
 0.9258805330510584,0.7920914189634439,0.9058194452421275,
 0.020538049861778762,0.9807630352438863,0.6942225065598021,
 0.3379646950769657,0.8739671415889518,0.8778383150758866,0.2767929236297205,
 0.06281131960077058,0.09151537024402134,0.30221694955337486,
 0.600012996641066,0.05502813253941685,0.9315637986256073,0.7430744467792183,
 0.3929604685567236,0.24650554918668965,0.9639702007461874,0.6301201747889807,
 0.06034613293083191,0.8279697613465942,0.8177552283101169,
 0.11973031817535151,4.779381094629498e-4,0.9633371895748771,
 0.9171351798517247,0.6014099275197232,0.009128566771052338,
 0.6077780457852101,0.06857703169976181,0.05967449357031318,
 0.34674573348570137,0.08952482845560672,0.0011256883421264519,
 0.4662381698189595,0.6222553502617224,0.11288841416410866,0.7514923234992583,
 0.6583973484308818,0.25569508648955785,0.9306983196780428,0.2539598228630261,
 0.4856905656811953,0.5780757123647678,0.7590817804691482,0.9297924488039048,
 0.17469118089045577,0.8605861885591642,0.32783479539582294,
 0.7346880781533025,0.21089166676589455,0.20121367986937333,0.897133147397085,
 0.9556341129056084,0.008005304752447095,0.3067040039608032,
 0.1456064006415314,0.836342846606461,0.8681181236026034,0.5017846457690417,
 0.4747507362178599,0.9020577251382857,0.8437743399224111,0.7553690357902353,
 0.592197747390333,0.3851028658351626,0.2185608630541872,0.3685363857953461,
 0.44341043249889456,0.2134034021100627,0.35764928546776575,
 0.6461372568222284,0.9684654688919321,0.5754029093014392,0.23198342032371289,
 0.39891965992909617,0.8182822098667798,0.24586862021499722,
 0.7501100630536047,0.63923724248802,0.9533355193775189,0.5614192172466895,
 0.33488182299511016,0.47037062630990767,0.21644947288788496,
 0.8779727948502704,0.47037406361741674,0.6436399014870252,0.2982402891865129,
 0.5239854171235085,0.04215197952732386,0.05603049209646349,
 0.8115746949774869,0.8114243702430844,0.8163451608567084,0.4115609258797599,
 0.39752346896515023,0.8961002687399869,0.25319960849929124,
 0.9547028757851709,0.3940501939841189,0.48038641432316076,0.6589915430344147,
 0.11414715518148966,0.2732980702079665,0.18600688965680545,
 0.2673888570672899,0.06574258362925578,0.5172748685624299,0.8718332394041723,
 0.7711080067164183,0.09113101075106333,0.5178064499906839,
 0.050659759714976715,0.3879625705099178,0.23797226835018892,
 0.20797309758764426,0.28792271267606573,0.3885075565143614,
 0.5090471057693122,0.0863474193184911,0.07363064460689739,0.1218080036320155,
 0.6944237692510953,0.26357271055906095,0.9739089605386091,0.92174417703065,
 0.40497100946915976,0.5072500858025614,0.10953271889720417,
 0.09708446317370212,0.36816209720422555,0.5362443831868362,
 0.026940697158844507,0.2623745447077689,0.010331394059780785,
 0.826135417264466,0.26191570939122566,0.7501712749981689,0.49528838397353003,
 0.8626355084472007,0.20453565901969673,0.895015307970371,0.6395597753857237,
 0.13020096675472725,0.32903974816498105,0.34382717958726505,
 0.6160544982833814,0.12699694407065598,0.4251824171487333,
 0.21875846236825658,0.5315296049132776,0.6726636560184138,0.9731038824956106,
 0.5404861813195037,0.2255553153190204,0.3023234695429594,0.4289892348454257,
 0.7125244324889816,0.808563069928101,0.6889393794095129,0.8070436889968036,
 0.41690246608539905,0.5253689635669291,0.1195731512218291,0.5238842047831788,
 0.12520274416052368,0.236481189871244,0.4206448728278227,0.40008184979139905,
 0.27808360008244837,0.4985026632845987,0.6349969653541087,0.5925860473687684,
 0.25039006062906133,0.6461814290282635,0.4396264285190824,
 0.23221462549716498,0.35491695851670835,0.5490682167979292,
 0.5553131423426394,0.4033774056773156,0.7795958256070833,0.9138464247095935,
 0.27361304229911787,0.6474205135657017,0.9460938875608885,0.2036689162770069,
 0.7160865024176177,0.20630976345929652,0.7187607866425505,0.6002389218186291,
 0.08794593699238773,0.43319192918817195,0.5609692459711798,
 0.1065532835557067,0.6199810380554002,0.48794613558407773,
 0.48464792288949365,0.29682390927928937,0.06538526500383068,0.15202868135836,
 0.8580426278521508,0.34446178325164456,0.4101812916211771,
 0.10633316630442091,0.27780353242209066,0.8437001926931765,
 0.8782953671916549,0.7763241496187673,0.0738861498198311,0.10980826089563811,
 0.29946477130112126,0.9490781319045145,0.7863345636601092,0.5150221128996108,
 0.031218349756894748,0.9911354278950206,0.5795866265556414,
 0.7472921610182972,0.09678673247715075,0.49491416189418946,
 0.2968493949089961,0.8839849925040281,0.4451010554804842,0.986321870919791,
 0.3966555246020652,0.8617435806172132,0.5106489562567523,0.5866450500770156,
 0.6972435898745974,0.17386659897724854,0.47807735145833197,
 0.11852788913811074,0.2836402876788616,0.38916320391827997,
 0.09548090013189059,0.239682414902744,0.1113933944374661,0.23341120358099599,
 0.6710887210175178,0.3741309717149476,0.5010104915148399,0.5507129575245163,
 0.051469164654386,0.5571689782670599,0.7454710839749884,0.43164953213618684,
 0.3241763698132837,0.5804188942579497,0.03127211651517825,0.6203812066508423,
 0.16104252569018418,0.5518136526382307,0.018946670714436564,
 0.7225540508479205,0.8220424244149245,0.02520829880605946,
 0.04720828603781024,0.16254259242430014,0.8714187449570834,
 0.4340073017582097,0.8319872376104162,0.3418901660672117,0.5734325553967564,
 0.6369509712932057,0.9067473356524273,0.5063889621540432,0.8769331149363854,
 0.0574521140850317,0.6928137499151455,0.2626529814019489,0.1178935725432434,
 0.02431782163521823,0.41180829494952254,0.956390377523204,0.6641352423444171,
 0.2420671042614566,0.8218870765407729,0.6388858633897675,0.31442094854358016,
 0.9350391282032701,0.6081162668108612,0.8129136492068976,0.23363449544389114,
 0.6976612304094595,0.8610521499120241,0.8640913951591056,0.6907289006357756,
 0.2387710811234467,0.5877137564285196,0.14041765389846605,0.6014292636852784,
 0.0779865336344151,0.6634469018296607,0.01895740039379956,
 0.07102214056812106,0.141555737350469,0.5386669137388556,0.08046255354803056,
 0.08824454528843617,0.34493515866280233,0.8036515257166006,
 0.7292482814840556,0.1401795952987418,0.462726471962312,0.8351202307945327,
 0.031207257358623286,0.5473722456366716,0.39166869500139845,
 0.49624211172955524,0.8928394617030841,0.8880952565111329,
 0.15957764875584346,0.18375002184256195,0.29519147026227,0.05400786047290307,
 0.8873471595854814,0.8538367895177393,0.22318577630404768,
 0.17905586511086335,0.12641686132144025,0.14487800493875747,
 0.1976442480640621,0.7300759399441988,0.1297801143451881,0.06837143688418013,
 0.2505291161480847,0.5951959949464357,0.74855609709734,0.7096267339096982,
 0.2564986147616023,0.5206358161758219,0.22776751386258587,
 0.16360844760663973,0.026909253947995593,0.43880535712733926,
 0.7046457655057712,0.818514314385036,0.8046445487816289,0.5744341899489003,
 0.3970638210401405,0.4022911425065435,0.08832698773827352,0.3203596670200789,
 0.49424581838818427,0.040353191157728285,0.450234091102091,
 0.2617968214932159,0.9831668318776621,0.8170987119011089,0.9322981360317856,
 0.9640759368354023,0.1021493224558142,0.3343307576796658,0.6497701787997515,
 0.247061382641951,0.3941093619277879,0.7852267095313716,0.6866943071205226,
 0.6512710318609597,0.16198944334154108,0.3965975478588367,0.8106173182319,
 0.9184393299241702,0.7688689518739666,0.394917268916971,0.8946867595526015,
 0.4594539001870589,0.02623234704513011,0.07354757449889415,
 0.6571494150000997,0.35506185377315436,0.33918736350843615,
 0.8785912012895756,0.678240586042655,0.4069406581561663,0.5841263426294594,
 0.0343843802723347,0.9027071513826628,0.153975740982913,0.7587166985124006,
 0.9317986310415868,0.5562469337643023,0.5231486012510428,0.9902159023037296,
 0.34087930293793123,0.6690672184215667,0.13479750627437748,
 0.7491199365799839,0.4390740670192732,0.78287758670018,0.802858173240155,
 0.749240225351672,0.11004419633557005,0.03442883115877282,0.1318045061544254,
 0.4388304963165456,0.7479423786291699,0.5883045415158061,0.6269461746850874,
 0.7310033275592507,0.36077435984660544,0.6052110621220481,0.6534862649391575,
 0.5284859224910643,0.6904469727687869,0.6162245117099803,0.5665940057359249,
 0.09838663778766321,0.9044747048806744,0.6117558676709932,0.9728367826318036,
 0.4690927658748072,0.012000979624548935,0.32631085383521685,
 0.9035467870100162,0.701009375688802,0.9682979811634757,0.5945373683688424,
 0.5632597077254428,0.5498954717647223,0.10206554470473872,0.2220460611646008,
 0.07175115876455629,0.6920628842851565,0.3308620136337178,0.7639112945792105,
 0.610857679987776,0.6891822320900838,0.023600338764150686,0.7150400125000935,
 0.040297233767515483,0.8745514105187536,0.15650392701104276,
 0.08812538889351718,0.4905525658700499,0.356495891863836,0.11333763140294217,
 0.6379161650922976,0.22157476552054478,0.9938621843715405,
 0.07080604953224445,0.08303326633264407,0.5689934462626346,
 0.6793372504932992,0.9799148032005778,0.5292456673375262,0.25108404387524486,
 0.6661650795613822,0.4872782103888835,0.8710155768939952,0.32884157504473466,
 0.3059673635137067,0.6793032794673748,0.5461362531630196,0.4702822068252208,
 0.8352457412726676,0.6926598797171086,0.2621448084535506,0.8094926657186136,
 0.10738101690847013,0.22492164431184036,0.9046044591942645,
 0.18789231332841594,0.9467301686683631,0.8108952162720621,
 0.11082088810562851,0.3885136076199559,0.5999352071395602,
 0.15055854432714455,0.13047439409127182,0.1420378226093092,
 0.9737730206027974,0.8675839271342751,0.5875897916841438,0.8764703922411022,
 0.42387926470541426,0.7104748015543302,0.11173986101267186,
 0.5501331116462795,0.4045671355790478,0.6395184327756175,0.9506670215161795,
 0.8162478524822862,0.3181359477928716,0.9917589851486696,0.7542968984110567,
 0.8693274216026485,0.9351195087555666,0.810896287317931,0.38760283392573747,
 0.3306970021329396,0.19642874080434747,0.7189243328267876,0.8330766992984835,
 0.5523086011509202,0.9322077752679093,0.8365106435641938,0.4356083823988258,
 0.47906820288567753,0.7596658708669971,0.3624984511873277,0.772219022151703,
 0.5772911342462868,0.43943352627960286,0.9925734775845588,0.5249332374495537,
 0.8073675171557053,0.4009586407733041,0.23156328933374792,0.6463542399426098,
 0.1334866875762064,0.6517146963837175,0.7069361485794912,0.23728051626903868,
 0.4452352353866006,0.13432900651140445,0.09604213724850641,
 0.2608674404308218,0.5904845255262743,0.042398217542402694,
 0.40615695201574453,0.8773461478263745,0.8054793940702807,0.6096147726556441,
 0.19255046064422243,0.6442275497228295,0.018496371751204888,
 0.336112407103748,0.5239898132261702,0.19035609832795974,0.39125161887434723,
 0.26442677387185753,0.795574958017611,0.45512778617883654,0.4373623913569059,
 0.2596890294740376,0.5940462915328028,0.8803563908147032,0.08558292301546644,
 0.39597664084913076,0.7167416205207885,0.3367115152058249,0.9266575464945989,
 0.08137538567567404,0.8894956378704814,0.8849399589826372,
 0.06614129569428151,0.10942812387763312,0.700832237221789,0.7843856464758878,
 0.7903766714946437,0.8267911062262489,0.1891333619378993,0.9365978466532391,
 0.028685296951459605,0.8142775417880626,0.47329377103606396,
 0.6048267992847876,0.8143463893050797,0.7805536891946829,0.6299482700419199,
 0.6456200989210777,0.8981626182907894,0.2507580025038494,0.4523195862786865,
 0.03011113177242286,0.19722523716019047,0.9717990293159552,
 0.9910226595252822,0.7988525858617235,0.11898092266833593,0.2878847058745241,
 0.22768236258518293,0.7377668180393876,0.13001737853829565,
 0.026252960632601963,0.8116123700633355,0.9379115354768814,
 0.2500202949559751,0.48782818790502536,0.8742231866355343,
 0.057840123815431244,0.7033652110209677,0.025240745755295113,
 0.4122990515817482,0.37148620516665987,0.14126294111245397,
 0.19314657324882267,0.09285722659243778,0.4591338925130781,
 0.3704847781623781,5.06402040140963e-4,0.7369699723296345,0.7570110592969849,
 0.3496828943099247,0.8397212335494525,0.054578125133719624,
 0.8651753754269398,0.9314416979022815,0.9451527258145901,0.25997829036748166,
 0.8500677876877766,0.6312725806427756,0.32168380043446954,
 0.20579501508823927,0.10341220416668573,0.8272179623667792,
 0.8827615302244061,0.9673294105227822,0.26110336580475324,0.4623925973452605,
 0.7854314878773402,0.8115279540073121,0.8191536011014087,0.7388080415523753,
 0.717773411764964,0.9812781264212913,0.9973836875981138,0.4074895235138506,
 0.7655728258692657,0.04418905417425145,0.6584876638272733,
 0.05228189913902026,0.8906442157042279,0.9341452902153462,0.7428815417533576,
 0.7382259318783416,0.6192335520757386,0.5282814302667124,0.8448498560334333,
 0.39833199065826985,0.001671905433391352,0.2786920637263324,
 0.9119105309644906,0.0977497699208224,0.9247245362607825,0.8832800610575434,
 0.6027574059307156,0.39810861102466644,0.02486865001032379,
 0.7628973243310024,0.5683459064716083,0.6314480744865382,0.313004156817705,
 0.9329270920781281,0.9241654452514323,0.8193121325694517,0.8197563756287983,
 0.2331722243210027,0.14527169183130295,0.006584688253324478,
 0.418748423455618,0.4361731781459315,0.6671027379212464,0.5239419378583511,
 0.583936008586904,0.9830674757610065,0.4443653514791088,0.8448650303534163,
 0.7049047668133214,0.6008797157446527,0.974269550039745,0.6772301104293837,
 0.8323062290629263,0.0575920056560556,0.8262421869513349,0.6189551084301197,
 0.07718990564935557,0.9253202034548947,0.24881457506131632,
 0.7490095026160749,0.5580629177028875,0.40424660914890964,0.9996865059154209,
 0.14657291993646926,0.71775452939626,0.3465064895231421,0.5869748565559062,
 0.6663000012603177,0.7026769872340757,0.8289909597461084,
 0.044079766224660144,0.8564270889830422,0.16300998996656924,
 0.9942674250615311,0.0352881000513533,0.07857230528275816,0.7805301690236298,
 0.914679895292125,0.4324214830681369,0.6357251335848606,0.19663319508967014,
 0.1355974644530108,0.011037812388031476,0.3501348417753426,
 0.5782525096610442,0.46773775286547215,0.5545909650165882,
 0.38436261252601345,0.5026200517236784,0.14926031410026008,
 0.34895926363890517,0.8655585267247755,0.30375209273270565,0.844096009651443,
 0.9169919308089303,0.5900181336943242,0.9108315029287087,0.476474399361433,
 0.7405275046486473,0.8692293018609586,0.6049124077577168,0.2521731244147811,
 0.7438788198893382,0.5832689432733558,0.3391808283692943,0.9904928352074605,
 0.7481082860047081,0.14482334240883987,0.8830496420809172,0.576164495205225,
 0.9270101516547418,0.4566164975110224,0.43193729361628563,
 0.023226331776481368,0.6125207855597141,0.7557866773674293,
 0.5156902872805612,0.58961327937826,0.2348457776463435,0.2616104707484016,
 0.9133253075803244,0.864419888039024,0.342462823919634,0.3918474684462403,
 0.8039504859597826,0.6765790124717074,0.175733089762665,0.4009934397048438,
 0.9793859603841928,0.3117059974400217,0.1113085078360243,
 0.006463710285455004,0.5208198173710281,0.2866948363871219,
 0.03163310585131651,0.577696520623625,0.23844177241318165,
 0.31323319120368776,0.6604525857674226,0.2746376854843262,
 0.37199167101910824,0.5777590356224375,0.02746715123065646,
 0.0025490585948996225,0.26678752889767376,0.614530678880898,
 0.2601552437523066,0.3789940694531406,0.5983655407545588,0.7330366813021176,
 0.952282437728515,0.49809746056964455,0.1982009833419165,0.779466809015293,
 0.9580201160574464,0.3135597619642896,0.7995726486918416,0.5874645836688543,
 0.34302891093879795,0.40411267912559723,0.27062107308074834,
 0.521762978354627,0.4194615282155951,0.19302253032038408,0.33559112106314637,
 0.03052720717818269,0.6325174172138824,0.6815502009254636,0.2169521494210891,
 0.21979753263594815,0.5540310071868717,0.5965637182670576,0.5239600352540311,
 0.24940621789932405,0.8706230120559665,0.7251893864103698,0.4855324131731109,
 0.44695469520821396,0.3858204520972025,0.5906182863694401,0.5363049875757666,
 0.1096358915675788,0.003297592343942446,0.6964724567972579,
 0.8715159430664884,0.9208891750732683,0.8104466081370938,0.18581971825419297,
 0.978996158558111,0.33229497703387734,0.9816315354927564,0.13031107575717038,
 0.16877508275458886,0.08277721935820403,0.19309709811992637,
 0.30952369056212026,0.2859396130878997,0.28600378813090643,
 0.0906593940209156,0.5467494930998588,0.10557633388449872,0.6251280567735772,
 0.38388293468456713,0.5121741685362007,0.42963854022035974,
 0.9450112627429054,0.5962877040921502,0.889861958950219,0.42315361427329556,
 0.34426250169225703,0.3311053400976973,0.7687549969043689,
 0.06893373935284519,0.2354521958958239,0.0822553134498275,
 0.21336815282572563,0.4602456979852355,0.3338910954640023,
 0.02734152385212063,0.20863503553508944,0.9596622574275584,
 0.47236822419142144,0.9490078421796092,0.20672296867727002,
 0.4266093184319605,0.6723510999342689,0.33266033181067134,
 0.16053225121512904,0.6870834502493806,0.7270128043058178,0.6283813440216667,
 0.08089627681842593,0.20610066776302138,0.20239420441477618,
 0.6100342686172744,0.23525569969429427,0.6382371858604066,0.357724104170855,
 0.4295322702237371,0.976230544304049,0.8006165128317835,0.9918521376641565,
 0.7941680900987969,0.020683449651642705,0.7242629867532316,
 0.7898196601242927,0.3971010896423789,0.2697085393001254,0.25190909382192594,
 0.07045060994871832,0.946695628274231,0.3408912201416897,0.10908740089867652,
 0.10587807867487664,0.3608774158277923,0.0930227175916073,0.8616453277199987,
 0.9551328565987711,0.11235322060418174,0.554649886480748,0.20758888541415532,
 0.912558721811057,0.5069538262332676,0.649681121003312,0.6339360419098305,
 0.801810425329625,0.873389812171627,0.9746584269936491,0.2196736337694004,
 0.6739615713850343,0.9050103747031539,0.2658883913728377,0.8064876473440011,
 0.2841842727533981,0.8031254602580038,0.9051758740652944,0.9699608398507655,
 0.2708935720056438,0.3568516720892799,0.6517193545299598,0.3021788733460915,
 0.03763177306769616,0.06716325950017588,0.3184682420241145,
 0.8876403545918605,0.753882144515864,0.6885052203941259,0.9957302644077518,
 0.6753725556682458,0.9009610653181795,0.6552868377343289,0.5058376201730072,
 0.6182138640936731,0.4958200324250458,0.8242167345734859,0.0909376915208463,
 0.8356466655961948,0.2677303969811742,0.32493113532350115,
 0.24902364888729267,0.21336388753460356,0.031514860173941006,
 0.1640019597174307,0.862870003384338,0.48449015581981414,0.3954995287816847,
 0.46133418158376305,0.2644844544805599,0.9814154744605342,0.5721419351772449,
 0.2849093428919185,0.5113854394800452,0.13791660222295266,0.7439898061646607,
 0.945514455767267,0.33122320301498676,0.21496460996807176,0.7650141451533782,
 0.6549406294355545,0.2125892992861137,0.3673129990998416,0.21755140922879468,
 0.5141330559276929,0.6096443150312664,0.8991019790922272,0.7317362504882072,
 0.9931047451182238,0.2770027168779592,0.3528085149563782,0.7690216566485666,
 0.21787280681257082,0.5607833949072555,0.8425819717721705,
 0.32152749259314073,0.9007657927856976,0.790613781369872,0.11290490086608684,
 0.22363520863891662,0.15362900755414122,0.23823116535921296,
 0.13111419920990497,0.26963972319477414,0.08064053785032577,
 0.9810053976289828,0.5284822646625106,0.9958922771117369,0.9268516880431958,
 0.7593962208731443,0.33874479047828887,0.45004977430891735,0.966363250776444,
 0.5375754631482614,0.7812773780726567,0.4799041140103796,0.34425194438652973,
 0.9665473559639132,0.8126574159286434,0.7176807936561042,0.09189435687258074,
 0.04529177090250747,0.9239548335347584,0.6281603126355593,0.4132515342816395,
 0.6360501702731963,0.8340412379349004,0.6716099875836511,0.25625674137737897,
 0.2386922724900753,0.5743992267815351,0.33653480592167184,0.8777113477752988,
 0.972872694629061,0.8759239415615346,0.11820589909212964,0.9821873506201259,
 0.9864060230954574,0.44591893662498105,0.8974301526236285,0.9366308192649856,
 0.7088383509590138,0.3928386752658455,0.7407003526900744,0.7013608240141423,
 0.8307718228577936,0.8954957710695448,0.6455857329151347,0.522616844750173,
 0.22267986950176988,0.4000273548609353,0.5769007787212242,
 0.10619783789790349,0.7280291568296586,0.31730837330133876,
 0.6391401521115159,0.5955544552619667,0.4316960027134923,0.4840569233060905,
 0.7929103331836085,0.4705003560589871,0.9392282580638416,0.21206431641025325,
 0.20279325698181205,0.32327740247645687,0.7122285371816993,
 0.5760108703814613,0.9647147228394019,0.4538269951937105,0.12403195377303311,
 0.11763378894983623,0.42060354979380987,0.9310837939860881,
 0.7455783021905438,0.09163902498942855,0.38575743868762047,
 0.7285739103232278,0.14657669007624907,0.2611788058367397,
 0.050052355318940145,0.332641368923281,0.7626603778940032,0.3213399627271297,
 0.9863474944852275,0.03775037145218885,0.5983316148615437,
 0.19051085265864942,0.23861045454386365,0.04312233893803474,
 0.319975703949015,0.09276096599843875,0.9086381287933147,0.6883327079890206,
 0.7666891080337616,0.21504168753030595,0.4880128134276691,
 0.24240701141705379,0.7948297936332124,0.8112458993253029,
 0.06992797724897903,0.8780263247951756,0.47513952310608487,
 0.40939398684476214,0.7266377420523279,0.7159913610246533,0.8936537946357872,
 0.8806445780098499,0.5136014649797124,0.3128239846780311,0.5312726658079198,
 0.3638244969103325,0.10666896055419683,0.7548417529012844,0.7151976430151272,
 0.7472185560525105,0.609788486058366,0.981965560278491,0.8729505656390154,
 0.24159624404598223,0.802813293399902,0.5234207069151592,0.7695003986987486,
 0.9721394392792296,0.7965121612731344,0.5212697169771988,0.9257702399633234,
 0.41325752305582664,0.10443180391749451,0.6874138995873025,
 0.5329397024700574,0.9343251112728763,0.5413172067190454,0.8347663232750662,
 0.0644660285427916,0.9701145259065392,0.15309928486775248,0.9696233153024587,
 0.18855082171593962,0.09088532982317621,0.9668779707640234,
 0.07025655613640658,0.0367004919934345,0.8155415874707528,0.3476913689317529,
 0.5718658796815775,0.7242889241552704,0.7028646083694579,0.5417401149695518,
 0.9514742342831727,0.9164952020447074,0.3394214216728222,0.427586632855149,
 0.18599800816833167,0.41815625372459064,0.6378744232736204,
 0.30948986654432686,0.9103332758590215,0.41760984004293056,
 0.12137798862305837,0.2229158272353694,0.8005532895936873,
 0.22295001235847134,0.8558231601931223,0.4857909569253175,
 0.054057057395374364,0.9371484353460432,0.6042134692421175,
 0.07973658230924119,0.8736367576369715,0.2969797024591512,
 0.17695586649540573,0.1749319781413271,0.894406341988139,0.8944868453935305,
 0.4766674236967724,0.7598339509006867,0.26668058967750896,0.7617626488125762,
 0.966547756213374,0.10971169803580316,0.845575958464682,0.3476594992184394,
 0.3468201477913241,0.12495691950303756,0.6620727065031722,
 0.48915953601744366,0.23239318302027268,0.6308018362464353,
 0.6927079233939714,0.2202274788294737,0.9594970094124058,0.19140016668124804,
 0.4479624050627008,0.1611938674296396,0.642032857289193,0.7320794075504686,
 0.6723718031233341,0.4251682847453542,0.9801864637149326,0.7147816615964008,
 0.29713107538821726,0.4809189554870652,0.6620821340623572,
 0.04743853978921797,0.11777511065700186,0.23624167544021912,
 0.6914263325686408,0.026783311285684452,0.671719784543066,
 0.39471638457914704,0.5668290019785936,0.4174138593358815,0.9707586446122864,
 0.7542745540338925,0.36062746863470496,0.4181382412591593,0.8621966945077982,
 0.006993439284235459,0.5638354669034318,0.06027858186377377,
 0.9689473010217081,0.9584870032287649,0.9413857117109116,0.5583966573130382,
 0.6351686374231866,0.4699693017631268,0.5612628174396056,0.6612667525226561,
 0.9463390334051462,0.3696397229111681,0.2923220351534366,0.6286480766630235,
 0.22765518314588506,0.453122338688265,0.6104088030444431,0.6710037040932088,
 0.7617574960906809,0.6939373187739365,0.1674426099295807,0.8237130151561366,
 0.623223783085375,0.5852410771498315,0.9044048242215078,0.13867944508245256,
 0.7716087456061219,0.3030373294801866,0.20584358707347095,0.5163981424909817,
 0.1780704681429598,0.8962850129494939,0.09370771941860756,0.7510513202676465,
 0.8458403379173612,0.7461394593803184,0.7566336955840438,0.17866670433740262,
 0.05598741324028089,0.46460068959121836,0.17863029134335995,
 0.7375835741596469,0.22971404252699335,0.8231588080660829,0.5698703271324781,
 0.9855993190707988,0.3266971585854649,0.6343571961140367,0.5225388439346493,
 0.2960045214153342,0.8123329830543679,0.5176030716517172,0.12074593619505636,
 0.9750493501306289,0.35723971772921803,0.7479556676485365,0.9619107192167478,
 0.15995161706310712,0.3467693053617862,0.40547537580872883,
 0.4957804095383559,0.8182194205784243,0.2962157664172018,0.4876102056639686,
 0.9025873702880871,0.9943848911829903,0.4215294487173784,0.21613468872917352,
 0.7197189818396988,0.9121192525032579,0.31082349630585737,
 0.023366964734780504,0.4548674906798964,0.4309425597781694,
 0.6255473221263853,0.07319427759537855,0.964081275065089,0.5840222695838284,
 0.46365749071449436,0.6534037062903613,0.037507527743084634,
 0.9443805849077654,0.07675237149432135,0.7591372654427796,0.7261496894751983,
 0.6431431461240862,0.11111790648657682,0.5737897930588413,
 0.46562702925786126,0.6850867701381823,0.36760336706185215,
 0.07747841591154081,0.5648073718642358,0.8670991422325107,0.4908896619142403,
 0.9506190612246044,0.9115250863953248,0.1728905921450853,0.1415746369596541,
 0.8862887149959181,0.22008777517922795,0.4408128963158746,0.7120435785833719,
 0.7705064296785248,0.48768064302618963,0.3419228324679524,
 0.15996446692184418,0.0023015967060419085,0.733653977724693,
 0.9801611941946127,0.9883697158983575,0.8803313839749668,0.9167193728688003,
 0.3370875346072594,0.41838396217948925,0.15601921121749374,
 0.8062253003142409,0.9900417916888674,0.24143520509910443,0.2993026207700161,
 0.9624588786889978,0.8175709445231866,0.10030673444669136,
 0.32400636786419446,0.12155415409491943,0.03853597902551087,
 0.12868765491860334,0.23261870916695926,0.857799750442215,0.078453201499733,
 0.12694635759128414,0.531777435048307,0.6878863653770819,0.6348480043652575,
 0.31013849539405514,0.45519939201846227,0.8129629802302261,0.284062689412256,
 0.9811657505889825,0.4159683069649891,0.3259188870235318,0.5135888365531464,
 0.7966179947840524,0.580525867533191,0.8055443132873668,0.07688480170538015,
 0.33587330928198633,0.7964650689752277,0.5200339767482395,
 0.41722919549476256,0.8625113391722395,0.2922228643879813,
 0.24119047477451572,0.29530294318044426,0.24866598191811817,
 0.5757693629983665,0.8459511502332664,0.7419781907516437,0.4199635629017756,
 0.09918084173707886,0.5677718846006958,0.16386356353710174,
 0.08784766697944435,0.872455612832628,0.34878009595425896,
 0.04527069805527928,0.6192829256766369,0.5911676463444481,0.9664137091380521,
 0.2721409792227083,0.22067787967114494,0.5263474743176981,
 0.23283885612283206,0.9082322010231304,0.6119352002302219,0.9798127560375296,
 0.9397939260743929,0.25949052130328654,0.21090289286143682,
 0.9731798071152089,0.14210982995805188,0.4570647240180342,0.3163470544880068,
 0.430767966036372,0.8364189582364453,0.07960395812983778,0.8272420171310101,
 0.5600055290273586,0.8802520938900212,0.058999643399508916,
 0.9189134456511745,0.2380432741882973,0.06452396405294336,0.3542775925014825,
 0.2746203043985167,0.6270720155324091,0.6466892175740064,0.8662361659997598,
 0.5122473515807546,0.6120840643973618,0.5028997895294931,0.96056763963527,
 0.45081962505484774,0.554700425219893,0.9183140210432428,0.7665602554622152,
 0.8652340742368317,0.9364392017824522,0.5407075763244802,0.05394176829766362,
 0.5899916730214381,0.5791718055373812,0.5242125954864608,0.43428199998865646,
 0.27819362057319685,0.5353366760772256,0.22044985310487064,
 0.5574322459621648,0.2383872303058675,0.7905693058918505,0.7600044758014965,
 0.9602469117942932,0.04644947905001584,0.9108666652865112,0.786214065837045,
 0.3915985668591435,0.8177171264933403,0.2858815386228558,0.3992024425578756,
 0.6343446416393298,0.6003845751820558,0.013226572424412497,
 0.45429996727520283,0.5804945305935845,0.36826145207945105,
 0.005795548463160882,0.4705042464702309,0.39370848233007305,
 0.47954955812198796,0.5302166609861697,0.759176830783122,0.8504039262092269,
 0.8275679494597429,0.7947731632115808,0.05741685777013317,
 0.43334061301818516,0.6182900318235047,0.8142965044155769,0.0704435291373603,
 0.6614004607187715,0.7998848364523449,0.4314799773711162,0.12763762488657682,
 0.7956570972174822,0.6901183234719673,0.10796697372970765,0.5047435433515549,
 0.0729522797782427,0.06829689263240202,0.7421221814220936,0.9514472723491211,
 0.432825679785958,0.9480831616610761,0.974943483840788,0.5483363026153476,
 0.29258791446402865,0.2547892642264977,0.04861000855664788,0.855454632861288,
 0.06058157738078085,0.19722428854905494,0.3071662718161726,
 0.12278526583218585,0.5022958782961533,0.02933104489467875,
 0.8291938178895796,0.5505896048070138,0.5942409342538252,0.2843000670716861,
 0.5299856795329709,0.4805734193433129,0.5800000897183544,0.10352528972670394,
 0.7372342562121937,0.3470593811431151,0.9658746389753688,0.21380086469647475,
 0.23697018384129798,0.8229466749467784,0.32557149049423495,
 0.9809160445096309,0.11291526074785563,0.3781685696368762,0.803740809364148,
 0.4869657470956117,0.2902521520853678,0.24680728680947395,0.5431098425425888,
 0.37839251326990153,0.7196961259297227,0.19857382342660213,0.483141286816379,
 0.8238591900473603,0.35290857442902324,0.9097673596055248,
 0.06543404323281587,0.517495033403055,0.6058932102260876,0.0785591281329161,
 0.03938454074126585,0.02568313692851465,0.8577374958915741,
 0.5431440752212887,0.8261571025153989,0.42211485001360305,0.8702761736568881,
 0.4771799156196884,0.7621058664390447,0.8446118716049911,0.9748306220976317,
 0.6448450009705226,0.6580637888064065,0.8605763803017796,0.8532430915554228,
 0.08787244933459148,0.6166692634871527,0.8893394138505608,0.4413225616612132,
 0.39986950618159267,0.07009112277336549,0.8783722002825037,
 0.18686893041416253,0.26407051117849933,0.47482625118045696,
 0.29016172006308427,0.401031997153135,0.11390386536559594,0.9679828021159314,
 0.392470045406242,0.2540417522268227,0.10530678159582252,0.3810361622078967,
 0.8592862800268041,0.46807462719834625,0.9177117167087602,0.617074350348972,
 0.60177211131563,0.5813531440804518,0.23683708365424294,0.5369336490596252,
 0.46500325437110157,0.5796574482409981,0.7857649572661021,
 0.45597435389938046,0.4091748732095102,0.3749051641001555,0.951048445731743,
 0.6395860183131887,0.15937496413516228,0.17233294526970644,
 0.14448393989099084,0.1852298348361311,0.4932892580179741,
 0.19316117281436285,0.321082847745363,0.6490840438838485,0.1848185078114698,
 0.13917183449682435,0.582530276625089,0.1870588224051829,0.5521334570805547,
 0.03707417622482856,0.9933759035628908,0.06951627593708709,
 0.30305785019320286,0.09149131596437732,0.2669500298726546,0.208912980789143,
 0.08277688266490013,0.583579495857399,0.37487041011211497,0.8009096615292398,
 0.7028470331194965,0.24019367651342394,0.6682483587798731,0.5357985399884341,
 0.42855794460493835,0.20497459472727209,0.2794766927337664,
 0.054774846607806515,0.17246507346677964,0.2018251380296069,
 0.2478615090989864,0.8842229089899158,0.28953431975403,0.19519751185924006,
 0.4180512711249531,0.2418847156686107,0.11960615947776798,0.7929520824411029,
 0.6383680422049516,0.8030244671648524,0.33358354188720896,0.6280162708993946,
 0.5426109121359317,0.05564031979015083,0.27827552910671416,
 0.34069119611830734,0.12103122611221595,0.3548881936158931,
 0.7937540571128074,0.6148485887242074,0.9168309246159532,0.7349290278718263,
 0.48061235808063896,0.8243970030644646,0.5157025710963759,0.8537828704518541,
 0.584394730159028,0.5558986447372706,0.24823686389159594,0.3762220250401145,
 0.517088126611272,0.43918479176426917,0.6762785099303874,
 0.026075952453982087,0.734926387431663,0.792887819036776,0.24597478112129467,
 0.8065559956260395,0.7072608259648899,0.18102090799171455,0.3498740058487533,
 0.656618589878494,0.23251530332173298,0.4861128134474305,0.5204151472376816,
 0.6649350958646715,0.724603414412099,0.49499595160354015,0.6534113735494824,
 0.6548033086189549,0.09510249264882531,0.14017065548013674,
 0.9573824565416503,0.4803583598762845,0.8129236022793109,0.7205416060296258,
 0.9142645329297884,0.04320213819279728,0.39422988498662537,
 0.7693011811725262,0.8792411388497411,0.4127681963737304,0.6548350763623576,
 0.8634215369774222,0.2269532313425755,0.16765507853899575,0.8681604844311737,
 0.3163663293397928,0.4229492552471448,0.9932950047353075,0.1621424525508922,
 0.9782101026990458,0.44494305862606975,0.8739288199255513,0.2954328657347933,
 0.3424735831983836,0.4183812613077271,0.3919759459389436,0.17956021988955317,
 0.7075898596265858,0.3492368822891405,0.7064820896777649,0.7498583796793508,
 0.8608493187545594,0.2801710017588366,0.19385395216846646,0.3052027940674803,
 0.1305445092208899,0.21700613860101503,0.4121995482905054,0.6352647104264779,
 0.06619739375719003,0.7802167532452554,0.40800922621567803,
 0.11014845181618826,0.9188117408081162,0.0993869083557336,0.9371951391478759,
 0.5446992290743554,0.8601159657131097,0.8307524166832998,0.9057475373171922,
 0.8935830230794485,0.4631256885449284,0.5517846807445265,0.5596932038190654,
 0.9588859131550682,0.9495745369214184,0.023044133567355574,
 0.14060945775416334,0.04664810020097421,0.9447605536823835,
 0.8657390607567755,0.33071577883461045,0.6427366541968795,0.9437085361049583,
 0.7592315634567947,0.2455637660690032,0.4465747568102536,0.1518853686860936,
 0.1256216255307756,0.027219398097986347,0.9986430145585601,
 0.2820743643302387,0.07705206348024818,0.21605722139072114,
 0.011224864310201532,0.9401049066874163,0.18671931145570098,
 0.6120661475887539,0.42691745010030013,0.9092070006963723,
 0.08621687405943623,0.42889207391168405,0.6903621553889538,
 0.8930302582181902,0.8497045735194918,0.4520921402864867,0.399820001552155,
 0.6524026479023707,0.9692522223586179,0.826308939846614,0.25646878125976114,
 0.630794071251907,0.7533586015367522,0.055166099256750534,0.6313190048286309,
 0.42063351445753216,0.18330611002508568,0.2834284075468224,
 0.3392590011605412,0.15737957733249708,0.05847155016580918,
 0.28326609207377473,0.9972293983558613,0.28076672039328243,0.711076913125553,
 0.6002290741014198,0.5919628689973189,0.9390066043307266,0.6498318191108792,
 0.7963398996434766,0.4090592973207321,0.08710837280189765,
 0.10404125605887549,0.1538588330815971,0.9625513684037931,0.5304958599932135,
 0.5353102213744081,0.23425187141941295,0.13923715449329788,
 0.20625408909202436,0.8602924490090119,0.7272032662603485,
 0.05554715612461247,0.9236564800218758,0.8112840315408334,0.49103331892333,
 0.2917622806508168,0.4660687227135043,0.14399671943423875,0.3632696578088114,
 0.794668877758299,0.00405478411323712,0.9866868210899002,0.8952801756674921,
 0.21818313940114187,0.5952793808983373,0.010552227804258463,
 0.6608046813284524,0.033336289052033674,0.8280455853099685,
 0.6667495393458237,0.5680613253325538,0.3828495261884465,0.8808480967089305,
 0.10981827671495614,0.18750660650927287,0.31004687030058764,
 0.5416343845026627,0.5553053006331157,0.3730672138226391,0.01300377036576017,
 0.8941157620583394,0.057931077735857706,0.9788350607212498,
 0.08243502244841139,0.07715670193044033,0.45550824875790386,
 0.5180259295820012,0.610905889976578,0.16645783575611195,0.9790869397254337,
 0.9342877912944819,0.47643848978763614,0.9478028375630477,0.7333530751963262,
 0.9157058215490762,0.045771223470449485,0.8514562009613895,
 0.8366261579385983,0.6596352914964229,0.7596756164717084,0.13660904505409255,
 0.3695454129700936,0.11737107409538605,0.776095063565527,0.5203742615324758,
 0.2982811807429736,0.5767100265749849,0.10383297367198663,0.3278582150214713,
 0.06410507390365927,0.5197093639672024,0.3438670028744987,
 0.08837829344165005,0.13182256939134562,0.5276338039518605,
 0.4449316345478065,0.4787007927845458,0.6977370069023421,0.6232011777506924,
 0.5867687869298308,0.11409429276947847,0.8575057700260442,0.5080315456552856,
 0.5323840645436553,0.6391333312769547,0.001815480272759018,
 0.39846021250830554,0.3946509377989579,0.20820869052183433,
 0.49462334172167144,0.11250130862522933,0.6268884993967366,
 0.0991267019076304,0.9448101023884784,0.6172026198451979,0.43445681480416964,
 0.09748662354434978,0.708749674724567,0.46948823738891976,0.6589491795502009,
 0.3257943489184508,0.2881235996212197,0.9094692104163001,0.11271901645061955,
 0.9901855542613569,0.46735157279877804,0.4473737523230512,0.8988934595147695,
 0.7172971650099571,0.2970201694050493,0.8956770742321379,0.45239281314428215,
 0.3723238024235591,0.41890796334891844,0.6891432538696185,0.9185550548752265,
 0.535913097119937,0.21705282818210092,0.15365274886216707,0.6886670102420397,
 0.9196753884082965,0.44413090574839087,0.9877089430773239,
 0.39059532596588964,0.3926844182015028,0.18651306337816942,
 0.08130599766126179,0.12418872778636825,0.25954353679915143,
 0.5936040807378822,0.9672420340835498,0.6490623894776101,0.09925208981019207,
 0.6313217940697627,0.571724316243389,0.5165522851204669,0.6614404469277029,
 0.5555852910676308,0.8139576653652616,0.47874787975897237,
 0.10594497281974524,0.276353077453252,0.6045642921504892,0.6307696484071833,
 0.24311463192871963,0.71939633660058,0.6692072635256363,0.40661293628420647,
 0.4101326098613086,0.5355083578315091,0.8298430727543128,0.8162051349470405,
 0.4693898016579173,0.2776225326223145,0.6156916815932568,0.43556689737040943,
 0.2832565891264027,0.2617498065023359,0.11692191526119544,0.6917124403700696,
 0.4526756981244404,0.9990027587862916,0.18148063514545054,0.50158235950757,
 0.7149557478191224,0.6867015312727658,0.011681708432541571,
 0.04786255721643995,0.11733904414568563,0.17544622347254235,
 0.4280147266654276,0.9483970567551588,0.4460902406236239,0.07408144662619465,
 0.28204299680736133,0.03061864499675382,0.7041183500847481,
 0.3043247375597542,0.9220222514064255,0.12724689053574711,0.7766710793952915,
 0.2699159834255638,0.7352630788946278,0.6681041770793217,0.06898208293065311,
 0.6412712333854744,0.05336504077565318,0.30134780489506996,
 0.5393265272858161,0.9909350389368689,0.1841640497172371,0.4753144899650339,
 0.6627893773350074,0.06068206703908219,0.2180360050242004,0.5010575300245301,
 0.6558420585916974,0.4296177332066582,0.6305287520355884,0.5531730750601661,
 0.3748943198007395,0.898902986763885,0.14983134220723704,0.13749138293652985,
 0.8987856925154029,0.22943706108698803,0.8945474573566742,
 0.36365852370099094,0.1619398771828522,0.5163281545744488,
 0.47998365417301814,0.5138495592634661,0.8010550166674881,0.4524159213678475,
 0.5243631185847162,0.34157999960521535,0.5488516058632738,
 0.25771010841345976,0.6283267812663101,0.5473028350527219,
 0.22082704516573204,0.716495244839474,0.8890804117843729,0.7392708970501765,
 0.6636627623510483,0.7909901326325562,0.71010862955686,0.8428013681459131,
 0.20042970333800458,0.7773978737406574,0.7458113892092885,
 0.24389335809027035,0.21384012331597813,0.9772479525687897,
 0.44547653447705426,0.7080009541620362,0.12863373375886733,
 0.30977116193177323,0.4900436595652987,0.8915890852669822,0.7043897553046363,
 0.14006966534747123,0.22879632016613005,0.7535348136290823,
 0.8227488203905791,0.6544563250220978,0.9790231929416194,0.5807728373339587,
 0.17528420862914607,0.12941577855153963,0.6662113730682385,
 0.24460583742534292,0.5018496099250643,0.40509533395769703,
 0.8749301597195244,0.8067231697122268,0.07401072286734034,
 0.11759199033129575,0.6744897667635823,0.6165126334102468,0.2523727078036231,
 0.36220982566713067,0.04940752445014551,0.01741576831503544,
 0.40127445255915695,0.9953469952974343,0.5000166470708095,0.6624757771903212,
 0.11756578811393426,0.044426422922331676,0.2877814691518533,
 0.44016344026030985,0.9237143563025985,0.7428140955625868,
 0.03513962292294004,0.641960702594528,0.6149902889223832,0.42548983484696956,
 0.877236653311827,0.6585279813413853,0.026680376370903813,
 0.41225562259441306,0.33958407618001396,0.3560978539735806,
 0.27911711479953827,0.19370174650766359,0.9704307114051718,
 0.8016298523679886,0.8100813717350294,0.6655142232916857,0.16092039991202645,
 0.8885031471342024,0.3725688669079862,0.15730923560264343,0.7973105159207602,
 0.9291854131324814,0.3815005770201222,0.6919942033566313,
 0.008468078718426697,0.8799554560209311,0.9532632664045442,
 0.5284335884297535,0.7301147673136994,0.07355094184095179,0.5139435929297916,
 0.7017385804943469,0.04937763840310061,0.9331006436795061,0.9434188439217333,
 0.8635285792999903,0.4905070024070113,0.387008445908378,0.03977300800211514,
 0.4233699371401838,0.06238140505345369,0.4012832100687407,0.899739147489085,
 0.32637281192529344,0.9012520689789205,0.786497676539896,0.44501660283907307,
 0.6234539247493108,0.43041076870211403,0.7847894158581472,
 0.20598633613015416,0.6020744848947951,0.3876170560908694,0.9623190581270779,
 0.6904458344654066,0.9339860634146728,0.6561068893637132,0.7379853212416143,
 0.6282455941780917,0.9236763574174826,0.5410796857010052,0.5070339454721386,
 0.7111426270524166,0.5038722222373329,0.45850737852479684,0.7152065573801644,
 0.2020138800117368,0.9629855182125985,0.9107052187238979,0.2990558650288647,
 0.9611733098932143,0.48703891096098795,0.5867705853772208,
 0.21558949673423833,0.6430287413742987,0.9109003167553736,0.5616768332532388,
 0.2046751818980176,0.8888693006935232,0.17076273869028236,
 0.42796562572345853,0.9338727218428482,0.8841153129123515,0.3634137972001076,
 0.5724657461669409,0.19001970953849812,0.7335614403197434,
 0.28616099788423455,0.2755557645327734,0.3202007786827479,0.4713482366577346,
 0.36084733647922684,0.16538019771819767,0.3484930042821026,
 0.10826960182915712,0.29366867712845657,0.7385038547636668,
 0.7060252697523948,0.5625441165528224,0.45734782393580675,
 0.27190086796781954,0.05617523753021636,0.06051253878522833,
 0.5117481739791039,0.35875681986587693,0.9162272525915164,0.2668289323296792,
 0.39241570511248125,0.47704872690519684,0.2847894924802682,
 0.33996889827925747,0.8473607684088601,0.6898303793570358,0.2559646708584594,
 0.7116815875963131,0.3912535446787301,0.9883594775293092,0.8582231314005376,
 0.8217339899239454,0.9270774835678084,0.8485069152775377,0.7044417977306875,
 0.911125202716402,0.08041734223789199,0.6577939121117461,
 7.034935722881608e-4,0.038786913332973794,0.9590272993599797,
 0.3217205393523207,0.9323796319348168,0.7594507508885184,0.06687822545294508,
 0.941005828963674,0.3259180463080862,0.6365714674830631,0.9556468166574238,
 0.7757987675964226,0.7462690494826537,0.6725447261500411,0.6698876597263909,
 0.5617796375835162,0.824756251558112,0.274766473388233,0.8507815583958519,
 0.05044700073299446,0.9355775595136082,0.45757683990471765,
 0.7096624258811319,0.19722961028990582,0.08092731854545177,
 0.8667101943477745,0.6680365417762699,0.028278783595552248,
 0.6384919818109198,0.9489831148508259,0.11298166920660657,0.551156336110049,
 0.5990011331605017,0.8331874425749389,0.44738235765201706,
 0.07857956690771162,0.4412924108306675,0.865098309936636,0.49363861717494706,
 0.5019487491751097,0.9838726145065366,0.8980529954249948,0.43668822362762316,
 0.16168848417110138,0.7865165057437307,0.49436013747286234,
 0.44618733124254517,0.600260236658612,0.7712980895860648,0.7189095607166273,
 0.5910565200797281,0.35908984363987373,0.3817133780480242,0.7597080233811901,
 0.028558569898405217,0.8727074050276564,0.9604458494951936,
 0.5858297178140037,0.9971456298945591,0.971468197704096,0.9216873608837686,
 0.4894993459442887,0.9844968302947407,0.8630261425465799,0.5813950669229442,
 0.6177452274168256,0.8470444089877072,0.7239018981956458,0.5116552264987797,
 0.3867894926718275,0.40666398155076855,0.4993804520370968,
 0.005670436500015308,0.42201120011254145,0.08676086350581558,
 0.15011005814122724,0.9758507752950413,0.26116749740222966,
 0.5621975960670967,0.7474356724789802,0.480162996137677,0.7955667925487996,
 0.5859433965630381,0.33707621308281777,0.7228952141807313,
 0.45086958645885145,0.17590545520299072,0.3746555044822747,
 0.37625887058220486,0.4650924381653688,0.32278423410676027,
 0.6732110725694465,0.9358458490063772,0.6033770708998838,0.10305528418163201,
 0.03339708373614658,0.7304495285391656,0.47276204427787427,
 0.6840095941403183,0.7114148590812572,0.28009706446733507,0.5198918704865654,
 0.16532918904178517,0.6308313428398175,0.3940936772142072,
 0.22589670629703795,0.15699365864221104,0.3378948502492707,
 0.9236297100293895,0.8865401110127019,0.6736721128709199,0.8348311026273092,
 0.5187144214824775,0.8077382910225562,0.6435651688824416,0.10455132558933222,
 0.09657152000186975,0.5698402673639862,0.6106371800845061,0.3959675981669154,
 0.39806279230153707,0.7038384254109653,0.823601789884305,0.7066423924418093,
 0.6542643739822136,0.7368788412971587,0.04529503247638478,0.3629984082551432,
 0.12009052449447966,0.729244046797965,0.23655030334009186,0.0658333211488007,
 0.17273499486692057,0.3458857598789462,0.09130246012504362,
 0.4159473787643295,0.5416382692615156,0.6909584968858499,0.65193332854835,
 0.5702349729292112,0.3337541261190814,0.3665695114039895,0.5242383377405134,
 0.8529771471029075,0.6329319146547275,0.030756432091914032,
 0.1741132533742774,0.8725711384125732,0.7762866889066269,0.7741530211576575,
 0.8887322420298435,0.42091547731818657,0.9543211580819706,0.5447435309900336,
 0.8958635407052522,0.923718736081751,0.07785582520866297,0.3398338629566915,
 0.28542395970887235,0.8540949717849805,0.916334850455481,0.9700402293973966,
 0.3692231104063681,0.4510599796442927,0.4982478640025165,0.7604507492952451,
 0.6652639443793902,0.9696152744715248,0.4220940666892642,0.7039607464364221,
 0.539512304204716,0.9709478994705851,0.7599837261617063,0.48559365671733656,
 0.6409866538556077,0.3119038639123136,0.312953761626824,0.06978784527308723,
 0.9453525014774927,0.7782353269888469,0.6700123138971721,0.33232487592638704,
 0.7553403534741217,0.40866124024210704,0.06173190357929803,
 0.7350803461408751,0.36589200624331053,0.04119718639032455,
 0.5394865286654641,0.8120822908284147,0.44205577046154987,0.6005612013714903,
 0.2579515335395707,0.4458976671776558,0.14556023915097072,
 0.22621183091065933,0.16348588776923378,0.4663337628572237,
 0.27487461709808503,0.8295545512691324,0.813382730316466,0.4483709973308396,
 0.022526342198629923,0.7275924956938836,0.6850452506793884,
 0.7684470048600917,0.9617260260928341,0.9248240961262475,0.582733220634465,
 0.6881974375749713,0.9507834452014163,0.01832889305100538,
 0.07794542278822236,0.5086677221229081,0.29927192499110444,
 0.08587537610311102,0.8894140812252179,0.4946688126381682,0.4363864234187408,
 0.09639459396387828,0.6342866744956455,0.9393192623178452,0.3588842133656961,
 0.9086570738870816,0.6375534552022204,0.13057490040661257,
 0.25039008062188384,0.028348021087437725,0.81568279503813,
 0.15073584463062994,0.6476788434246371,0.12910690008915537,
 0.5030563418164853,0.7936131794969544,0.5996912415478536,0.4320729915701086,
 0.8257097611470775,0.2766268118150106,0.7007256764746902,0.2434905609105148,
 0.7506945884135208,0.14067308758299713,0.17276371654622225,
 0.5170903500639561,0.30768253420505276,0.5344502002013327,
 0.44789179205148066,0.12776503799277472,0.44105905857337724,
 0.0325844753490756,0.3394162038342784,0.8797746535162475,0.9480716915653358,
 0.19199876809375105,0.08085277543827663,0.9224109896740422,
 0.04269519551130807,0.5181340041403433,0.45764700291658256,
 0.9623949443972843,0.08957894501097219,0.706626159832121,0.3477105322082844,
 0.7757601452322267,0.3594123866866029,0.8715410769673972,0.4618634537545545,
 0.536353689429199,0.2707388812601004,0.12893385915898625,0.04598726485578397,
 0.43039430915112886,0.5003765043790989,0.3278301513647912,0.597353960498363,
 0.3312687468472212,0.887217656424905,0.27493955309285656,0.5503783456599189,
 0.05821551161685412,0.8175993085190552,0.4631810310643082,0.9280161659864828,
 0.5331603587920614,0.6050377802217584,0.7327643480505842,0.22895572050946533,
 0.9434419245680541,0.8150813253648878,0.5041100577813282,0.24980895380780155,
 0.27882188916550343,0.3410813842250032,0.9261213851705938,0.3790435758863723,
 0.7770768597354943,0.32192849240434673,0.3741685152478813,0.714195856376769,
 0.8955375978909447,0.9236160807358699,0.07214096993512087,0.3471758144851448,
 0.3615773812944527,0.7712460552959108,0.04432429866097565,
 0.44813413279953973,0.46432856303222825,0.42197346892566245,
 0.3608604461476551,0.057132564870267855,0.9615515369696742,
 0.7645013500010261,0.011384441522428568,0.5343957846875894,
 0.8861060316861414,0.674946525873115,0.8928411311779295,0.5614697157958102,
 0.5832406541665371,0.06813883395720921,0.025667233993875493,
 0.5184037881978282,0.6460960603317998,0.46730570142716565,
 0.49641667697108405,0.004541215632699713,0.9115013546972998,
 0.38206815010177886,0.653312239662297,0.9790812952913662,0.01835290085572261,
 0.6651844996605911,0.4384682102920563,0.09365859158466272,0.6153697267158699,
 0.9022503608963479,0.9793834000563637,0.4766447139052101,0.3358955497302716,
 0.2141842459934845,0.2606218230485393,0.9651971927556614,0.9342946688483981,
 0.8241400442731037,0.8476221549477863,0.8883540480244396,0.27539596797543653,
 0.8334530447266754,0.5209344176665076,0.03515410449558676,0.5294764036372785,
 0.9790121838990223,0.9727602030835735,0.529769875611413,0.13256911958715312,
 0.14412324709479107,0.35012633392449444,0.8108206750057789,
 0.1450485659521441,0.4969306170666534,0.15154605553623002,0.7395268725032992,
 0.3842079107267259,0.3279857135729175,0.838978877511443,0.825332673805044,
 0.438954792775629,0.15273693415907408,0.9079253799441362,0.5966787676792841,
 0.12844169770266434,0.4045053317615188,0.7538815957796343,
 0.013063086190849615,0.2421867644750203,0.17737598492905937,
 0.29259281673975057,0.25998474966991747,0.0121998883965867,
 0.5897009322206094,0.8068065047907035,0.7544360801185445,0.6454473464028968,
 0.031051484837952947,0.8195909882042149,0.1826751674216116,
 0.9696621591642492,0.9842475905207094,0.02121877757551871,
 0.34644285296441946,0.26120548321619863,0.23592392346648006,
 0.6083114837674761,0.014284417966371032,0.38249506670957156,
 0.217045709353061,0.16379931630636846,0.2682638156113646,0.39721896592730843,
 0.36085741180952624,0.38628182542551015,0.6138630514503621,
 0.4587384838564036,0.3926375297029304,0.11214321437763364,0.8045088633760542,
 0.3198133526744932,0.5147302421892679,0.30881660492306506,0.8862891530340633,
 0.4014319164070662,0.2581208887696038,0.2487381757353332,0.733038501922361,
 0.05682207573363851,0.17028242020324835,0.9639496387258872,
 0.16637259502090507,0.6845821872348833,0.05040135289794323,
 0.2285285980502766,0.5736071105654765,0.26340290645209274,0.3290373912739012,
 0.013494312431121647,0.12752668458067395,0.4994085939793972,
 0.6634120365881299,0.9273985766560843,0.7333371980781921,0.8412539743290796,
 0.9912834126365403,0.12032024575743328,0.6052863391168186,
 0.04940757157957976,0.7967837853747968,0.37852626356424557,
 0.6203097794196143,0.8128732956937181,0.06741723986556758,0.8000664783055335,
 0.6872131714597582,0.7019033786628947,0.19552173167496223,0.8038727794218354,
 0.46607630553142276,0.6152718189555396,0.9939735339301681,0.8352109639205945,
 0.24382589210762784,0.49133488705881334,0.6595370024404762,
 0.3146540381693739,0.19600376889505988,0.4416634061413731,0.7905990490614448,
 0.15670313203152242,0.2533944393703894,0.7377333107770229,0.7750869907492552,
 0.41818781674146344,0.946754330435992,0.9272301173143698,0.27617476455821355,
 0.06201276043371706,0.7766953169131046,0.653015056540841,0.5685683121342336,
 0.27101877051946,0.9408920532395527,0.7657058081250102,0.32214571524690716,
 0.9866520514328224,0.8713984209815883,0.15414510208833376,0.4978161212290364,
 0.4588925378011145,0.6263139238923812,0.7820348637494896,0.3410653468054127,
 0.6428290253327688,0.013720593523675273,0.29529662939383594,
 0.8730777370848579,0.43709794126136214,0.23732750407894154,
 0.17536205214265776,0.5152640339340949,0.27119900346038384,
 0.9172894222388853,0.6509241920214307,0.12168909271616801,0.9650659077893017,
 0.36416085119774844,0.809438254275243,0.7538935692028009,0.2874882967577236,
 0.14477550054966937,0.8528180778977019,0.4105356567905183,0.5640422606610859,
 0.6905516427077107,0.513169381350106,0.14218122590526805,0.7422103260487901,
 0.5432241497112422,0.060667560806440424,0.2443725575007759,
 0.9001101577391134,0.8020935612607931,0.7398019306098454,0.8833904043756688,
 0.8150662258858706,0.1265827764861367,0.8937221786081091,0.700114801243783,
 0.7771523370277933,0.41961050073322415,0.6536051391811885,0.866413629358352,
 0.9053210399604664,0.7376963892337275,0.059595153169914594,
 0.4773793299068211,0.5616081883879019,0.2810619846589143,0.3479976119530377,
 0.04164380898808162,0.04227631505527707,0.9716445731290817,
 0.6679572636950913,0.4568376410681755,0.6028382086648121,
 0.011842388064488452,0.5130827463623308,0.7348092204595167,
 0.5611319871495433,0.12160504303230235,0.40973726801314336,
 0.5428189249587909,0.4794142880089901,0.809271706146221,0.743531957164555,
 0.48127528036203526,0.2124011031651981,0.6270659447852333,0.9369759064907143,
 0.3969510819672809,0.8237681472722701,0.44899083605729295,0.9491645278113393,
 0.34145467855702805,0.6774823952140384,0.46998794229831586,
 0.0856105747338447,0.7749733408825248,0.24827650295764758,0.7705945973753164,
 0.7681448744045066,0.8778755148422875,0.8098039118697172,0.6431463344191601,
 0.5473412843738062,0.9983738379447885,0.6852591299032863,0.4024550922821305,
 0.7821564058546814,0.28354461162019184,0.9668897178280471,
 0.19108047663969785,0.8441828294120768,0.01073710215984125,
 0.2103091885707351,0.6185392948686754,0.7699044552948254,0.15704030697298776,
 0.03814313764744626,0.8300667387695257,0.8335779176899931,0.5821645540880929,
 0.9100467267004817,0.8720233985623496,0.2676542679313928,0.11599845991062563,
 0.8428271882918881,0.4811771422157689,0.82156586467224,0.5962728297907998,
 0.7035737024849152,0.18676387329930422,0.2010661445154205,0.4026531528486583,
 0.9793427184016331,0.6274341599002952,0.7232046172246147,0.8647779469076122,
 0.6420703290031713,0.9615597959258708,0.0427720355804635,0.57210806574233,
 0.8330834421661742,0.6331731733339486,0.7120159566872678,0.18250724861528622,
 0.5935794529899574,0.948512645695335,0.6827786742653067,0.06611245411746025,
 0.7324021260271499,0.6607982681668229,0.1111325130303118,0.8902863097925834,
 0.4273330290462568,0.253385527732777,0.71776923871833,0.5091406461330545,
 0.973077061320378,0.14455131395868637,0.35761393035752465,
 0.006079219752524123,0.2473388644233262,0.39498182644439184,
 0.1540994632297692,0.10773765745858654,0.952589278213056,0.1392424004360986,
 0.8090538540150357,0.2463127447717246,0.2462407797858065,0.22957814904399965,
 0.27747005041823103,0.41580604639776475,0.6829103827950423,
 0.05413748796563711,0.05190862933069951,0.6998960206437327,0.129177378320404,
 7.537344389516587e-4,0.6571084105473414,0.11389995423315424,
 0.41639752792092577,0.5630283607644267,0.8434751790966146,
 0.14826408004543756,0.009007511811271574,0.8677592939332832,
 0.2136120680145619,0.7274948015446523,0.5643253602654077,0.306877075238691,
 0.22741774040163532,0.23470470369141372,0.7333137663933122,
 0.25032330584060325,0.06863556986340946,0.9680259022143041,
 0.5058630753665241,0.3019015037902284,0.42123426656300933,
 0.39040772776137944,0.4217340490766931,0.873880360758849,0.5521482277833097,
 0.8392723848390022,0.29635948065593165,0.0859318383586195,0.5734595963498719,
 0.5182519321617558,0.6920553694214777,0.22996548912773218,0.6067901148669685,
 0.8658167493634286,0.7457208462996232,0.8872532291777697,0.9722600300259814,
 0.221121491615469,0.6659202894229664,0.5389684616073573,0.6630809242945439,
 0.28844724210352024,0.8071538498431546,0.8844570261659913,0.736330914078347,
 0.011997108405735757,0.43621870707345256,0.2064726075432246,
 0.965849201280455,0.976808825608048,0.6839944100945928,0.7703279058011345,
 0.10358540848245701,0.09442391008576734,0.6684769603876108,
 0.2338537521945866,0.9970741096099248,0.7020741689577726,0.6651118303042393,
 0.3965943435425632,0.9808346123286025,0.8994770008533002,0.9515853233268005,
 0.9432852266275713,0.4002308528764491,0.8854884152458344,0.8822983001331446,
 0.1825547143648607,0.8816946510131565,0.7927214619153049,0.21322891968877355,
 0.7142354425828981,0.2618127422437375,0.6404722341862275,0.07904894391660922,
 0.42023434671272664,0.05332453775605439,0.07153353502665982,
 0.560720814952861,0.7830412286827859,0.8366609608756226,0.6471858821633708,
 0.6744566844619464,0.19683566574070543,0.6426288092888931,
 0.39836936920136234,0.16078585023455172,0.8704390153003981,
 0.9700817233095307,0.30273256448592245,0.7962954411712543,0.5302161542297578,
 0.9986309192714028,0.2116181402408046,0.3036898224998096,0.4638986315996969,
 0.6068031270598431,0.2536049064590443,0.05291514875363745,
 0.23573442931774768,0.44343238247332173,0.024796283143222464,
 0.8484872079665741,0.9610059008531351,0.4347687293415907,0.7182055718944569,
 0.2462708737089,0.5027752446250832,0.8120268735363292,0.6050752511303051,
 0.0043389445053406694,0.8634407713184381,0.20352413301527217,
 0.7639982017116987,0.11742900633724396,0.31946680374862235,
 0.9974617837002506,0.018382752795960022,0.889463863368681,0.2011473096694737,
 0.37926511285132714,0.7268289325012605,0.6439451709142222,0.4403697561185218,
 0.7216126305170623,0.5498062170562458,0.0014540297565966975,
 0.7302300035178599,0.0026051482877791354,0.8347921077063203,
 0.3008545033113391,0.8280095110591158,0.6053153185210547,0.44210867893264627,
 0.214328845473718,0.8710297937321514,0.5668961001077268,0.2913046303008656,
 0.7498810041759337,0.3934290054815551,0.011846125971541333,
 0.11214151820535856,0.6001061662822323,0.20676011418376283,
 0.5446144538536651,0.6497933929795012,0.2539638955544823,0.29631831098061634,
 0.05842733213167561,0.23200418242246612,0.2469066742147834,
 0.29571871828470264,0.711082808567184,0.11743390527383002,0.4106440988933985,
 0.9614288978889968,0.31566750322318016,0.9752124879271139,
 0.07685314858687553,0.9803638720504603,0.7816022478527136,
 0.21858489387979008,0.9760162373775401,0.7731430112466794,0.9547645941095584,
 0.5182808402077016,0.20156715575005313,0.9690577593512162,0.6736181803361935,
 0.7360226711402446,0.4721617075078126,0.003425959508076004,
 0.6956631963720729,0.8985500883826952,0.6249884076752299,0.8159489088136913,
 0.5662476615349825,0.140828553114702,0.706915070683357,0.3615831952414219,
 0.7360422739504676,0.2761332830383171,0.3421587470746883,0.5347676080904008,
 0.1818330516704998,0.8140856539166956,0.3113786787799909,0.43305375540251867,
 0.3243492689658066,0.9787434200878342,0.9126616696841127,0.8170067081900482,
 5.831318792608453e-4,0.8224267861039279,0.9862960914178737,
 0.24601167299544824,0.753566220391563,0.06844648426815891,0.763735142984503,
 0.9837676325097673,0.09905849167816294,0.7573354522044189,
 0.032966555114934626,0.7337237809067465,0.26367244582070304,
 0.014219164689580732,0.4281537793865692,0.849222111807692,0.9313597740564454,
 0.9545405623920059,0.8864284137459851,0.6283781287229787,0.11365656140961722,
 0.6771375303527605,0.5857178575907919,0.733835474255713,0.9199632402450825,
 0.9247611635139764,0.7315730941803743,0.20204994226369966,
 0.31044450400059986,0.12241366051490599,0.32717382650296933,
 0.9596711332676925,0.28222177624304146,0.19215755993799621,
 0.03260952921793525,0.8747152051161781,0.16525638458934178,
 0.7549826514397738,0.3706340468889352,0.6930165820380306,0.9750587677496648,
 0.49282842985014785,0.43227549808612475,0.5191124705825004,
 0.8672374038107218,0.43595982098714536,0.9891302030160711,0.9346235673983503,
 0.5514734932963153,0.1423109080116567,0.8869263982709844,0.562348990826835,
 0.17085177741297042,0.6969637010193601,0.49707772546741913,
 0.08568391873869496,0.2317521304625869,0.7757862276984167,
 0.36452433121164496,0.2920272343656233,0.5556032012083975,0.6947644482823723,
 0.8975866921906375,0.6001268419346258,0.4614283542501262,0.14320451416751345,
 0.6136604122096911,0.07632491763024296,0.4441268739642652,0.9910972222160355,
 0.6569048097830257,0.7575742517424485,0.39650752276749057,
 0.27941672847018784,0.9493740085312855,0.6408918631199968,0.2595849479129433,
 0.8490109682321211,0.09454946726277902,0.10490650831854298,
 0.7884085592876274,0.3939404948130717,0.5397649478649477,0.1917829493732912,
 0.7964566499964739,0.3752673737147537,0.4573380883280159,0.7437553681729574,
 0.2241400848424877,0.2833679208415303,0.4965038124273735,0.0751450596836779,
 0.1322961125365718,0.9836104349444976,0.717759443582378,0.5989148666748154,
 0.04575680944213589,0.908272691701227,0.5198582896868711,0.5994352388076922,
 0.6370455027154894,0.16802363039113266,0.44799435992637815,
 0.8295946939782096,0.7098293052492365,0.002920242119507699,
 0.028336815698511764,0.6592700937728906,0.6793828769482774,
 0.3179384433576642,0.32317271370569167,0.0582296942763888,0.7725263364302528,
 0.21901976628920772,0.004727594241561883,0.7654406124153473,
 0.26025801252111636,0.5457760847394267,0.5979542308285068,0.7090794311125808,
 0.5617994985308999,0.40596823279044547,0.4612344434750173,0.6714433954760951,
 0.21083122141191413,0.7001408053371853,0.4053020609896911,
 0.051045498269483014,0.8290051983627318,0.9500740499661877,
 0.8976986831467817,0.08458429038311488,0.04822793118710533,
 0.15131253794725552,0.6705991502776789,0.17454860349265333,
 0.6073580206233683,0.6192708396806572,0.9175037682131465,0.3499576751317012,
 0.8662780356835738,0.9243316996143731,0.8654948220932035,0.9773403359815593,
 0.8039961415265194,0.7433299928163379,0.4152495651119049,0.5849223186681454,
 0.4606734752154755,0.4454738845037347,0.4418538333389742,0.12974534929573955,
 0.7574469792805126,0.8123330556145857,0.5420296270082423,0.3585517711686512,
 0.45722798773061046,0.817711634797517,0.5700552926539935,0.43731391860278546,
 0.18389323871220964,0.3009286456583148,0.053970987793624636,
 0.6174132710108042,0.30218040784463585,0.10868183434515044,
 0.24684725645355887,0.8247465266260077,0.0832626676608571,0.5672178044759099,
 0.8547166300210463,0.14798030442086696,0.16264099983294567,
 0.4777613819699511,0.3668822684216647,0.31164378617450816,0.8934873747359298,
 0.24538983960066973,0.7105393921575922,0.835968232274442,0.15713267036288214,
 0.41563229302406535,0.7152399577449646,0.20343004622056549,
 0.031801525062053915,0.42791971563890097,0.6913660359048294,
 0.3732101941438213,0.07439517500824011,0.3363844504901038,
 0.20604804776562158,0.9460517774677425,0.0597142292341315,
 0.44070182004466874,0.49106662572851323,0.09250654188074858,
 0.5333205773880632,0.2169360644460946,0.39993557233459975,0.1339246429501414,
 0.8208027084898335,0.2618020469747482,0.6937956809390127,0.4659961695662721,
 0.12591331427520602,0.8691679070511781,0.07154417262756241,
 0.6641118994089901,0.5111909123236064,0.9885843405647545,0.4970790834711687,
 0.09832262477912368,0.7770814995020475,0.21723032044795032,
 0.3793601681194608,0.8599717682168898,0.7977374704319357,0.5449601987375107,
 0.9556789375187318,0.0034990102610201745,0.18209830914684966,
 0.40459775579222435,0.27534243551761195,0.9466386939438624,
 0.0607915697158764,0.8792682641816402,0.3957913849974788,0.05017745402306184,
 0.22817091196326866,0.1692879910241073,0.41600897306732454,
 0.4731162157260447,0.22936995624772327,0.9261821879854206,0.4470560042293521,
 0.27898700943778076,0.9254320979590845,0.992367904692059,0.1300064581489524,
 0.18849144932225914,0.3379264693174461,0.008520697622722562,
 0.33376547216185004,0.19780969825927208,0.1781507715222862,
 0.7939769288156782,0.1890866088103541,0.14068984209705204,
 0.22018972766294276,0.9851509308946913,0.061941606836835805,
 0.539668489880528,0.7848917545552235,0.23978777147062802,0.39271407133828695,
 0.524324859250026,0.8992318596008504,0.7984397770637561,0.629613341967763,
 0.9069258928838372,0.21271025019786638,0.4518503337813694,
 0.03010647197331373,0.8203435041771386,0.4339074383237378,0.9313350492126684,
 0.6875248853331606,0.8780131789999798,0.9771746466358572,0.8901585475681341,
 0.3715144900555035,0.13750330677222933,0.4693999873918705,
 0.10668842249253507,0.1864586146947692,0.4883797864727315,0.6188381935598116,
 0.21624605641834904,0.06586777013619227,0.3888934058712491,
 0.12422523529402674,0.6807076167642394,0.4355310502782922,0.9174963759697751,
 0.0016839677424265176,0.8142826902104026,0.9968913911605926,
 0.08023605194356453,0.8002713067464541,0.36673935301541694,
 0.3238328673876617,0.1879368232030314,0.7361030288092992,0.2760269174266342,
 0.07479846757132069,0.24476677276984937,0.04443880260260191,
 0.4558863438195122,0.18365281085584562,0.4365970459275077,0.387736674548948,
 0.3168107714040256,0.7633023819528403,0.5525833531052278,0.18124579479631608,
 0.17749662966075952,0.3476592591703631,0.8420626228842647,
 0.23264979293638932,0.9531215534443769,0.9778463571636208,0.6795690507289076,
 0.7010366981376979,0.18262383788677594,0.4216448901760117,0.6822163997878279,
 0.4447696517541948,0.854328293458165,0.9440870229658097,0.2354684354220259,
 0.9699918304220895,0.6554834161925321,0.6473160428286419,0.7913909097732984,
 0.8984765547596271,0.038060545756365416,0.8470226880754308,
 0.04855601829758471,0.46942622048200444,0.38631942703091116,
 0.9921588000006831,0.3082126938611358,0.2973323141577321,0.05391234052911298,
 0.9863445784478868,0.7787110721828918,0.48074348640064124,0.5407208968180885,
 0.5186464248289304,0.9449356094515069,0.5395465438308956,0.11334300694034938,
 0.9686490858295768,0.589475976356978,0.7516631017107785,0.70293141517711,
 0.9865881736534432,0.8973647771829268,0.7728657988952142,0.17291260564601818,
 0.8625750055083974,0.4442545670870981,0.6074326689117195,0.6320278537086806,
 0.6200568225630934,0.20602548769857143,0.3557604428608153,0.8850309346323555,
 0.007699128672659206,0.1487826042074626,0.20270448375925398,
 0.32637985168415307,0.6676879223826827,0.8839028456141496,0.5105946072681807,
 0.8679167471362396,0.865027719256276,0.9436152400553859,0.46437418801670427,
 0.9049293355286064,0.9946147776580954,0.6378908261653888,0.4853246159856738,
 0.6280234462672991,0.020966744066063803,0.6465061877390123,
 0.5040931763900834,0.3581640669975852,0.17347335158471378,0.2925607749218073,
 0.4427455136769569,0.6699015861746344,0.5265594507970435,0.16819532370989976,
 0.7704597771822435,0.07279006321790138,0.4228594722643918,
 0.21202562330988461,0.4357611783767916,0.2733267230329187,
 0.24998408857014365,0.757772577837368,0.29609208502944506,0.6058478009415847,
 0.6458891553236493,0.3074202372681265,0.38177815415120875,
 0.13232806144664278,0.7200176972821957,0.2731542454718915,0.9776284226336643,
 0.8110895397906274,0.9150218779753089,0.48971741110162215,
 0.22734348695912254,0.8743323870077041,0.6310290214368361,
 0.35512627779744865,0.057839049796071795,0.28442974586414116,
 0.3297887689790794,0.7284665851048792,0.7442846040761322,0.9913542135328185,
 0.9419238466522528,0.42077714037719927,0.47919691202538917,
 0.7794347271866974,0.02116638763733303,0.750706326631665,0.3159998357807682,
 0.15943762500909475,0.9523908238174891,0.9141393885225599,0.5484418005650669,
 0.20190059490450707,0.7833803943004309,0.6392637111381528,
 0.21912226225960652,0.4774819954986612,0.9848816252512904,0.9043170592627476,
 0.6564501277239888,0.5780726432614173,0.14385955670499162,0.9314470404509592,
 0.765551283577644,0.7858657692684241,0.05195814460637782,0.8978489302858277,
 0.6454001406173181,0.8855985146872996,0.22508652507902505,0.617462648435313,
 0.3290832796803276,0.6234592094332132,0.043503803306309874,
 0.37749039715595434,0.08965298951533807,0.7911473796388567,
 0.28031484179628197,0.633747105346042,0.4341371126332405,0.2561961521720003,
 0.8333428523322002,0.6024828728389353,0.13381683006992806,0.2915069054827927,
 0.40815802842772786,0.07023741274228212,0.47491138490237095,
 0.3661797545956681,0.5970410116906508,0.12283086866462112,
 0.05805130694744254,0.5127261456359691,0.7124840265755584,
 0.15902797157855908,0.6037257298705911,0.15209667291906293,
 0.3025766872922001,0.6209109690746976,0.8385544090447535,0.2577188553584604,
 0.2988695596732397,0.5669435009255648,0.1583725790862851,0.7712656058825417,
 0.16932685249985346,0.2065698253383198,0.7326903589287851,0.7866059667814567,
 0.9000794065921114,0.9520317141335874,0.14412026456673654,0.6746346819772602,
 0.20022756887176707,0.2447274098483072,0.07127738941604245,
 0.31623743882973854,0.9928611265978923,0.16589550768873906,
 0.008668702177135335,0.7424454501498963,0.4468688962696873,
 0.4040465492336396,0.2577122229432377,0.45439880573550606,0.9196390174934034,
 0.47501691904997534,0.0974357658406042,0.4625635760833343,0.9549674772360381,
 0.09901631364906005,0.18198493060206378,0.40899611867824515,
 0.4716654901388051,0.12850259248264262,0.366455978800342,0.6595667689661203,
 0.9873264492083547,0.7450957348411767,0.44063690396198063,
 0.46588700842413044,0.5001448048148528,0.026832454284096308,
 0.8016742685026133,0.8052271350446283,0.05616920998487329,0.690949940185096,
 0.37798897716173985,0.5843845973627031,0.13947147418751538,
 0.5384900696133009,0.0443257682360132,0.593256502462406,0.5366778530175662,
 0.8112652575410466,0.8900877812176442,0.8279101193191686,0.7753717578803588,
 0.38741794374579797,0.24974466377722493,0.1549271161846717,
 0.9762294636969303,0.606125576939147,0.8423104159885784,0.6611694230865593,
 0.9585739115525966,0.37355231954282164,0.6128310730418941,0.8692479156619649,
 0.12342810445031993,0.11233569012188349,0.23932238580386045,
 0.6727742104437505,0.6866523847410726,0.5365749942827905,0.1896633583490981,
 0.35456298935052066,0.30326712257779787,0.7729910841570345,
 0.9655079291114805,0.22156187273945227,0.020569303016152896,
 0.8775590554290329,0.1401239133353328,0.7760788353200356,0.25858968987753617,
 0.7240320917462955,0.13947425713222472,0.21737188641661875,
 0.08993199909518657,0.7626190890158201,0.3411088560455815,0.234027956740418,
 0.8007052063453016,0.8162715184656155,0.3384527763081322,0.6946137510202418,
 0.5781288079297044,0.7023422020098896,0.7070096177509029,0.01253366153399127,
 0.23362250877006585,0.22128962140124298,0.9805746605309067,
 0.6926991581447358,0.10849982854881424,0.04158118371002073,
 0.7341442607366435,0.28514875856546484,0.9367553134778334,0.757652465816738,
 0.8691856316147538,0.8815953375685516,0.9082207688060592,0.5394877966908448,
 0.3775977899071694,0.6100233870591252,0.5735304706425932,0.34091906018923335,
 0.6867598779728822,0.43766052896657637,0.06774876552229347,
 0.9788007555500011,0.39299323453859314,0.6657083924641909,
 0.14471699621375345,0.225573022852395,0.5954165419707237,0.5240367365334994,
 0.4738789274258157,0.930217726953074,0.6923636765312722,0.02529583574752292,
 0.6990239325747104,0.5181151896334363,0.5360243914987693,0.6918729897565745,
 0.0030167669047996837,0.24231319781166216,0.5065010349159814,
 0.27387396140578946,0.5080392378832345,0.9318471538128061,0.5888228430951684,
 0.8824142850078228,0.35527724593984555,0.5211434686039871,
 0.12759486454806734,0.9084207210914041,0.5976608536856187,0.7155542925181736,
 0.9566431250251956,0.8046762451616826,0.7933063179635137,0.30280803497496245,
 0.042877316612005,0.1172855264931486,0.3856693862510764,0.5043770473117293,
 0.11045163659940749,0.043254565707942394,0.7559192150084302,
 0.6180700112059031,0.8856650635145727,0.38146814146057706,0.9755311121285,
 0.4865833597795953,0.5011353654923285,0.31164945302913105,
 0.49736226561651564,0.6828255211822404,0.9952884733719103,
 0.44539734201552617,0.8814645236413303,0.6981017151949005,0.3040906274560048,
 0.3004826741327741,0.3147896688139332,0.8103525645136268,0.6060335715763637,
 0.6428149446726408,0.10954701637387754,0.16810412460779223,
 0.21078812452051632,0.7892037306985301,0.2647526221699835,0.6512860955297394,
 0.22417482697887914,0.4252146704300339,0.8145016625797636,0.5954005391834826,
 0.8591342125209706,0.42939968739123224,0.12893687205227156,
 0.9699105862611123,0.28946578566091563,0.6467767755265672,0.4168937873514186,
 0.8504284829654126,0.6507143795002824,0.6716040827067962,0.9583390223791733,
 0.19094023788425996,0.9703198842861254,0.3141776977058873,
 0.04280841338062891,0.15591276219188588,0.7815554197854858,
 0.04453336182277701,0.09128266843674715,0.7527991260708045,
 0.2978349066894783,0.5959562282070765,0.7096413366007046,0.2258276844314242,
 0.5237275717845915,0.3285853972670433,0.9377534167805472,0.9453319662541306,
 0.08795783879335306,0.7026127335097467,0.6643779611191538,0.3534838246156484,
 0.34321033201049733,0.8721534943275564,0.537975349920425,0.3935337065565876,
 0.44858910046132094,0.3063051718565619,0.4511609453663338,
 0.05272460461581552,0.4548410918409207,0.36350976394466317,
 0.8272641289004846,0.022707478255817315,0.8804211868686921,
 0.39159939096450835,0.16069358319967075,0.9129878322770392,
 0.3847544518033439,0.8378427728258253,0.5578100682599068,0.5943936134035056,
 0.6691448957392919,0.710990480972763,0.6075870781902257,0.009888166962191858,
 0.47958390045148713,0.5467324388956009,0.5083328863932718,
 0.05928233635750779,0.3633603301411138,0.028132938581732825,
 0.3072699591621385,0.29461523860547567,0.3472165270885872,0.9482999495718041,
 0.18869025159479202,0.8587022670100679,0.3603747227262146,0.6966029920464382,
 0.8299953201052354,0.7898378677615495,0.34644398131679943,0.9254734629517102,
 0.167305604836812,0.6050495362088842,0.5247703446072859,0.4991716831751075,
 0.6601467557557541,0.1480652310670344,0.6406452018831149,0.6097480762267619,
 0.7893408037335098,0.09450715433188361,0.48454074439568007,
 0.24041656868172945,0.7440979436729075,0.5996426441962082,0.2386682425691784,
 0.8481352417964308,0.6403877024051544,0.6984845409931715,0.9692659641805884,
 0.019711995424418927,0.9824101054023964,0.7525542621754222,
 0.29059244603981216,0.3107065660677484,0.16605771818600068,0.524732112528398,
 0.2612126398804353,0.5539919771576027,0.2022419536919179,0.01135468054926081,
 0.2701149667090944,0.11853903249355235,0.7009332881922288,0.6733281645405063,
 0.1278295705579321,0.037059154836512676,0.29173213489232686,
 0.4829483546938781,0.10272688065027014,0.7742300611757538,
 0.023003703025157418,0.31228260031489796,0.7683423225053396,
 0.5353214547288039,0.8718066491169016,0.9025382835840136,0.23506952765963762,
 0.15246944041911714,0.4280405754560763,0.15819711188309604,0.727401779322997,
 0.08677035639948649,0.7992626362978721,0.28919948223418634,
 0.15288479603652105,0.7054536451097808,0.5667150470541235,0.8011627544922402,
 0.1990398762534713,0.9572163505079097,0.9579901110788207,0.6263882738409767,
 0.6118541566289023,0.7026404362339611,0.9103409063064601,0.39564738270696276,
 0.010712847443097306,0.45768868388634565,0.9533548642001998,
 0.5264387622583669,0.3581783268110812,0.21098872338561248,0.8281723889809587,
 0.0011956018560456139,0.1454982166029546,0.13887577019266084,
 0.9681681952341835,0.8374255927560684,0.5003720339706546,0.3666042276729595,
 0.6662482301780641,0.09083201342030867,0.8966844062600419,0.760930719408221,
 0.4057283432195715,0.5683978231266797,0.5202114752649709,0.7431130099459852,
 0.3287200218445596,0.42876643507187784,0.4796919910666385,0.626529731396853,
 0.6988364076928149,0.47344742713187093,0.9287069358916945,0.5965289407155367,
 0.9273236756029133,0.8923995769623411,0.7791307602277624,0.605957549593996,
 0.03703088709465763,0.5260793632366555,0.5609878049357673,0.9694990940280209,
 0.43211192548036137,0.8022417947792143,0.15503281658253731,
 0.7127991343852211,0.2870923549861708,0.4943095410072227,0.9243527412621019,
 0.9249832690704015,0.48534423098222645,0.4443783197386071,0.693800011935185,
 0.6850274557898983,0.8818129532373062,0.9602133994277193,0.9313913258661731,
 0.7195779334033154,0.5595297377423427,0.592741196946158,0.9827536917677218,
 0.41561423479647785,0.5693376146487708,0.14720155637333576,
 0.4933377333008717,0.7263590694045565,0.6422115213877084,
 0.010333316872928489,0.9292544902230794,0.009332260446113017,
 0.33683773163458297,0.31258580750024656,0.42094775128272843,
 0.17697733387316583,0.4544127073069447,0.6458775353241633,0.7235669990459542,
 0.23629581820194545,0.3051194563523866,0.9524679548344078,
 0.11436668310492026,0.5634148509360477,0.3824185716201126,0.8206586096965376,
 0.2571759939978264,0.5030323110355466,0.8543022589111042,0.6189375814670292,
 0.1380441200425424,0.858925181129565,0.4785098845507395,0.2390922586622719,
 0.3978087833271785,0.9073235398418966,0.5600847190397907,0.13989934263135828,
 0.9125031908209794,0.562857426815878,0.7253252675867308,0.16519134835360028,
 0.9063261829557193,0.7609096390804243,0.977126209481959,0.5937966226792575,
 0.17270889843367132,0.8604822999863099,0.4650806816484305,
 0.061340826756729916,0.7499524447893792,0.055949750434931556,
 0.5709065558132925,0.6361772663549429,0.29670046415208307,0.8436581667698602,
 0.38808400933445397,0.45031083364028746,0.06699119332488335,
 0.027414695212395035,0.28304779866607843,0.5333168968387393,
 0.8983751936643863,0.11263804684077583,0.3096929314728265,0.8977225882908895,
 0.03885696728493804,0.39978285017499804,0.055824238487085065,
 0.8354964389577583,0.7563976152019973,0.005621113040516601,
 0.24498934125363914,0.07867148650995504,0.6729882235026333,
 0.33939226499707287,0.15350077655291205,0.34499272355316,0.41418962962436856,
 0.9175773759815661,0.5922769833208189,0.8605505675332151,
 0.021643002306868442,0.9219267979358012,0.7278337134580878,
 0.6602130815884389,0.6510072294865108,0.587079934295846,0.5326307777188837,
 0.25464573686710856,0.1368908752893362,0.5395877312147523,
 0.22576371264124273,0.10675520459976084,0.8419952468707533,
 0.20756870818357764,0.5322180588859501,0.6060483131625759,0.7144718794665154,
 0.6164761056114324,0.1063963778114333,0.36829710078224476,0.946710894279164,
 0.43319914980585494,0.35209367775104317,0.05399633608855747,
 0.9367855327782495,0.6628342388511574,0.6403891032449422,0.5366935521373825,
 0.6994209005659857,0.7482080427845388,0.7127289017184332,0.22095454205193477,
 0.9487419477907041,0.38566847151151573,0.057775008589732835,
 0.5997998196889942,0.18447828703829328,0.5914121923069479,0.5999065420092764,
 0.8864992706336566,0.7493476890122446,0.19714701601690754,0.8948343544000465,
 0.4958311438901315,0.7197625168333787,0.7406520962293871,0.5470644942706184,
 0.21946089390718793,0.9167662134053985,0.9407396378079697,0.5250074827325388,
 0.5325102581930785,0.7460734443204141,0.46444376339393223,
 0.22419319405231797,0.9560547142126881,0.6860492324567156,0.7767574590215575,
 0.6774927655159135,0.43877660651755146,0.9404852451312613,0.8755315828004864,
 0.7580361558197282,0.1728985863886474,0.9567708187717283,0.971306077144836,
 0.06409927484020561,0.5001661323816528,0.7160507839461833,0.4762123678314666,
 0.7219163115292683,0.898652205773282,0.621439850846041,0.12432628482925445,
 0.8097732315198295,0.290012719231858,0.04770964753077034,0.7819773738191098,
 0.20155504388122092,0.24913531136157507,0.32909555746197316,
 0.5604837566261313,0.8754129990266418,0.12100169758127477,0.5738003791007956,
 0.2464297850749586,0.45240000463016516,0.940549198055439,0.7064428683547272,
 0.45592138213590927,0.1530158096990547,0.1911279230593963,0.7324960585209993,
 0.7091112076816442,0.7041845333873132,0.7644990048063726,0.2851263943101874,
 0.8887750043345928,0.35524869776998547,0.19433846766929463,
 0.7434780834845233,0.21765786771970996,0.24053123477697536,0.530147535757984,
 0.4005179828322196,0.5971332558489538,0.829284459543704,0.16644917004228177,
 0.25048059724757,0.006196560249946859,0.4240579747404415,0.6848526543331923,
 0.3805806341278729,0.5259155558441233,0.6680232794471953,0.12784095987096455,
 0.7631709802347103,0.10392473145919667,0.44741273005112936,
 0.17087245748210012,0.24595648990593644,0.8959289137697074,
 0.8921808082714302,0.6568458661235458,0.06418746841379641,0.7171115844453481,
 0.8251268800272227,0.47376295693205017,0.5537219055191959,0.5522356706131631,
 0.48591600251966405,0.44185856279022684,0.6840594211804213,
 0.6877705030418114,0.6616583572917404,0.003195365223806035,
 0.38889401632803566,0.2153188425678101,0.4020467575642557,0.4733221552246337,
 0.3299968928870065,0.21894176332249482,0.2717261217457726,
 0.09440277106875195,0.5539471638787862,0.16447025170385265,
 0.5859697250370391,0.044613895599032616,0.7227126727784653,
 0.6544946981699358,0.03857945980887334,0.5137534365937119,0.7421204232744487,
 0.3928491348872637,0.33356680553286644,0.3758014339986102,
 0.05996608993087105,0.2995662923780753,0.7864884427516476,0.4436185410073309,
 0.6944633488182642,0.03093470969186851,0.8805744771573552,0.7910318476202103,
 0.9203578575524839,0.9349172880856249,0.925708946501469,0.07960356470220253,
 0.7636007415560638,0.014853165440506855,0.3547976399070424,
 0.8512762522228272,0.14241744595578332,0.8234389353304339,0.8074490230997754,
 0.9225219760368886,0.5174181639024681,0.7488008801828004,0.859988979793583,
 0.8851547320511185,0.4698480664381227,0.9943485775428845,0.5838045034985941,
 0.7923103180011773,0.22374004359318544,0.18153379500572153,
 0.5532140709468206,0.8558666523621145,0.2074812941892199,0.4952334371534989,
 0.7536577505885427,0.9050537116711396,0.945375038463223,0.8470026506892939,
 0.5202330399446183,0.5988019230110742,0.6837630843425352,0.6549631364341603,
 0.5039041900659885,0.23622683219326424,0.5303856940753838,
 0.08831269986873658,0.6703943405274668,0.2510464797055647,0.7115769881094688,
 0.7200432358526905,0.37634789116424794,0.35888030087806955,
 0.22802488215428274,0.25236756787786696,0.7934673129662868,
 0.8556070263834338,0.5047008218808826,0.9827295684228142,0.3965054953914906,
 0.25407258688394485,0.48453903722384895,0.5226368903762242,
 0.9698977859040713,0.09360702692525391,0.9373218022308976,0.9031199531567169,
 0.2131856209606643,0.3622669486044696,0.09213457066712993,0.8851801400205009,
 0.10886570849248234,0.8169815269680022,0.3907076029093348,0.7826550833983341,
 0.6913603567084218,0.6476834483336419,0.34942808218465116,
 0.33923406278299073,0.5277165418155325,0.7051180901625583,0.5162923922548088,
 0.9321179918354346,0.10180670211052434,0.0023109800970073646,
 0.9438906299194301,0.1905646891863122,0.808457581038914,0.41530901789056407,
 0.923111403560339,0.5631285399480386,0.4705494034924399,0.3270067642454686,
 0.2033139424002809,0.5094399147546204,0.9264459484547227,0.8086368952515333,
 0.5722007479175307,0.5275882443255644,0.46304999351125486,0.3146554300807739,
 0.43301294314365524,0.4206977217362544,0.885500278013619,0.469931448945782,
 0.03800405251548622,0.6736355305904839,0.8046172521710481,
 0.24510994711052436,0.7279631246028637,0.11139464945969824,
 0.02805688904161452,0.14683968124418845,0.09136751791868014,
 0.7595135708533199,0.3016400022940927,0.6235263843741804,0.4553516369064867,
 0.7733698012022341,0.5004203338713161,0.5209808056843834,0.3576682281129345,
 0.6747728155360386,0.815808349704119,0.29437272024500005,0.5746085011154894,
 0.7189553209764493,0.1678404833810505,0.8088398778397857,0.782605247352319,
 0.9516887308231043,0.060145398918342696,0.5229714971035279,
 0.5562581504414947,0.0014611920309965942,0.6026942024398675,
 0.22984979772359704,0.7793397626010279,0.8676594904235564,0.8156000508661327,
 0.4743274474036784,0.29845009753125407,0.05784724330340707,
 0.6234209472182594,0.7574743893977879,0.9606005895973336,0.5696970406375483,
 0.20827048147191174,0.5322636494161279,0.4206874404672365,
 0.02845796712851889,0.8105364107179629,0.9960328697053397,0.4878370892330086,
 0.44164725019490203,0.09062351341630881,0.03612754430531395,
 0.5918534044285024,0.5801428535111197,0.3108657230356072,0.9190862818436013,
 0.8285625645903191,0.23741201025121572,0.6961279067276006,
 0.35893328563742233,0.6257803747075409,0.11157147695655112,
 0.8162265916323472,0.5457245854413079,0.8359697086289298,0.5065603919087667,
 0.4305071819814461,0.9113976643790995,0.9560678395597215,0.42307515111459715,
 0.2197607926981413,0.21477794165608233,0.03328045601846574,
 0.9563538257142453,0.6306107647629473,0.4528641887939403,0.928364518613812,
 0.9991692604742443,0.12906813076076074,0.3103764082844266,0.6115554952111213,
 0.3313378915918026,0.021227617246274866,0.17079716593630423,
 0.0743786742807665,0.4053071211251835,0.5817325981255971,0.8563932600091341,
 0.1648578153148863,0.40081970989829496,0.3538071615514501,0.2594318810755699,
 0.5591867422396812,0.5972992560326897,0.12060057737383245,
 0.011722158312246655,0.7227705515758518,0.3960277975420886,
 0.2457346445133366,0.015232134632103822,0.529573515884194,
 0.04517564334418955,0.7074632834000121,0.8038372257120838,
 0.25483835724779447,0.11333770766530982,0.7195513589456767,
 0.4687901145503406,0.5325122159858044,0.8584469390101008,0.5326672140775576,
 0.9268998404964977,0.3441942929427749,0.3941551994663727,0.8622044456596495,
 0.2323654811212742,0.2649736408997787,0.7153836845670805,0.25969631540043236,
 0.3108992623462956,0.8868845888589146,0.9615143201778383,0.4274152965727045,
 0.4353872577764525,0.13672765844775525,0.9130505185024843,
 0.16930177138223812,0.18433965515506145,0.9049464550895499,
 0.5858217540351134,0.854656457934384,0.037111221009242024,0.7638062757061342,
 0.7960341035927797,0.7374295947333201,0.33696128743103215,0.7417453789407163,
 0.486224595072714,0.5367977286712767,0.45880896412774985,0.268124465125663,
 0.6356151164176916,0.5241781054034279,0.8740424338911176,0.7428373084687614,
 0.6684293308728322,0.7369591730446834,0.7158978967818789,0.8518259935528729,
 0.20959324117768297,0.0882018633243008,0.48291106441982234,
 0.5966303805598678,0.6779071694277197,0.09861981150134325,0.4829068890675404,
 0.7269005209223567,0.2190946890855413,0.4003467231647687,0.6232260357552794,
 0.004016365716241199,0.4118518970351268,0.5682352261976721,
 0.27817713581991066,0.20788286848187398,0.8574106995351944,
 0.4250120662986543,0.47596617684334974,0.599344081993431,0.4199965258238232,
 0.4941038956246113,0.5050810850000801,0.01684183755342361,0.4304086500694936,
 0.10710383170731097,0.8490045298515366,0.346696148436634,0.8210638754408695,
 0.7788381394642715,0.08222760549239316,0.016515827592956223,
 0.8409136700609702,0.0573510822180342,0.03920306110661331,
 0.27961962875634727,0.06399801697667251,0.6679387361431526,
 0.9336522158125319,0.7977582879233366,0.031711505075398105,
 0.8973791271272612,0.6705419273760682,0.6944862915947403,0.5942299449048303,
 0.04716262730216314,0.6324393600958517,0.09179582054045188,
 0.5494503684022661,0.0017413639145702486,0.37653219528286386,
 0.7473974509651954,0.5561076841022513,0.7024878515334623,0.3004692245616166,
 0.980266393640931,0.2016134103043533,0.09186992105101655,0.6968520962846407,
 0.6499663342254751,0.8435144061160288,0.5372814878476269,0.8692555046326439,
 0.439553727887227,0.8402231360941743,0.3407422914777414,0.5019758077785702,
 0.43473652929039586,0.5185227381172112,0.28407111903966875,
 0.28278256394622514,0.8048504043151352,0.7403187090627706,0.1474520161516838,
 0.05756734979053579,0.8534253103096245,0.9714118770401942,0.536474959470268,
 0.8071734533452202,0.27168029982509867,0.7446471433140701,0.9561832683850529,
 0.48056134983236565,0.4982422005399131,0.19179186611117238,
 0.8025321377640757,0.4685847229029071,0.23230091473442527,
 0.18187737879192145,0.4485207316363877,0.17767428746490244,
 0.15966775834718427,0.33340449923020277,0.5592975053266576,
 0.5280262409082195,0.8601506622455709,0.37533892940627855,
 0.006663923559139584,0.20882354181777263,0.8875724643961826,
 0.734987354927912,0.34910996578605513,0.8223450953586057,0.7597729132612818,
 0.5068168149542037,0.9850936608702183,0.16321751896539471,0.5634261332784567,
 0.8879332037373854,0.8055726372254642,0.44854351263571257,0.8224469568613787,
 0.9819215949299038,0.19292258959710207,0.4385015722774477,
 0.35267192084179266,0.2174894623860918,0.41367302340537626,
 0.3338094821737122,0.908121157271204,0.060473166022395386,0.3795809542662596,
 0.3858385973141071,0.5365358178577283,0.678407003068341,0.15700559029826044,
 0.5783947427584684,0.3145859037738111,0.6829856428038064,0.7994238801304441,
 0.4546718052552796,0.7098923297798266,0.6042826131965849,0.1423210931107085,
 0.699610366746666,0.8174044554679125,0.013241472703879542,
 0.23846328860922872,0.07363841254096193,0.7291267509749695,
 0.9574793483404616,0.9383641586298332,0.8112981647270104,0.7330175088746452,
 0.61471452004555,0.33678286197582663,0.6313292245190385,0.8093138713934005,
 0.14984730877665264,0.4153603565712205,0.948378708927853,0.6058920409270067,
 0.5478065539592878,0.1812078965111128,0.26562452734854514,0.7666240666368306,
 0.6763847835102991,0.013350524031173205,0.06380724311142938,
 0.34270861817589715,0.33707446297081656,0.6548711130973569,
 0.2944711198603258,0.41242863905040295,0.03643388835321182,
 0.6908450206137414,0.13609853943674954,0.08332283635822835,
 0.7557934724777434,0.4401239603270699,0.7650826174242336,0.6410229911755486,
 0.7379273759361867,0.345458887904067,0.33664790660945954,0.5449500172543659,
 0.7912402662289955,0.614529893691039,0.686323275925864,0.41521871182351433,
 0.9281286777809092,0.037806639228815,0.2471140168830731,0.2092217855263323,
 0.5396613413487866,0.35988399633624324,0.2449927620244774,0.5360472356717938,
 0.9656166368998784,0.7704661258875023,0.7284923786167328,0.9956852404008134,
 0.5064529292479517,0.8052968834435086,0.21635039884151053,0.7089029533542672,
 0.3847298526457502,0.9755169171431168,0.3444993296261888,0.7132490490113644,
 0.10668453365775077,0.5970307891977202,0.5976515552981138,
 0.35574803326628335,0.6862033304327694,0.26678560364761694,
 0.9111086056716282,0.1909910453980065,0.013925941483614479,
 0.17009852614002552,0.9351948391362715,0.501106725738709,0.44671300908310596,
 0.8547431158338332,0.10054928406101504,0.03385667939347503,
 0.24248201450228313,0.28093763493540713,0.12489414825082767,
 0.40400675589960877,0.1436053644251033,0.266125137977799,0.5427958731894443,
 0.6747428876524744,0.8371349197167974,0.46815664191290196,
 0.40970066287404905,0.21904570487302166,0.004500826766775101,
 0.06963591899606847,0.49921913916649463,0.6621769200135224,
 0.7107270181815875,0.1542144218664172,0.2642281777427188,0.4019387759452647,
 0.7923746539410349,0.8362965734779032,0.2915539157108997,0.536243068953685,
 0.768462544133971,0.9827199190096687,0.7081617314786692,0.9546887049570363,
 0.30488464870381093,0.08310025035386359,0.10388081910014291,
 0.18965463725440113,0.8755141731993933,0.37749901600706726,
 0.7115218849999532,0.492624908764452,0.536968367154104,0.4351330040462712,
 0.6665854410410533,0.07665082555614533,0.9052996864880519,
 0.07012398805683406,0.6195553253296618,0.06017061494352949,0.718666393377474,
 0.09599768012350451,0.9044398648210208,0.6787470874427644,0.6383298994257545,
 0.19793664057748084,0.1954918885100172,0.9292741609349606,
 0.47115571992777494,0.7780434359458454,0.32668917674806697,
 0.5128195256466219,0.6664754046094812,0.3146286196340502,0.340510689602161,
 0.39844757493626337,0.20963795744868052,0.06186220952331156,
 0.9290211668607743,0.9662965430878773,0.9322042307588895,0.1948716955854446,
 0.7531298424253176,0.48588273711747665,0.35815211702909255,
 0.8917209027663937,0.11727002438917711,0.7903914109916923,
 0.48160181283559933,0.870651345443715,0.45691773483989606,0.8395607015961811,
 0.691117497128657,0.061555038478822044,0.38017313200552816,
 0.8194771136668536,0.8438780023695991,0.1977422248825793,0.06478777841748418,
 0.13908696992632175,0.7406562490105042,0.5166294687619583,0.2345177784788156,
 0.34064265358493406,0.8845193366847274,0.8967445738414702,0.0997427122914969,
 0.6506798751663779,0.05253251421119243,0.88686940716333,0.0613136183220121,
 0.7523517301320186,0.9669861623359766,0.9345539789242203,0.5496325216199314,
 0.5642481567631401,0.15831368787497324,0.44175126202724013,
 0.6980604334165281,0.8832680453461491,0.9563028856349265,0.1337323793307914,
 0.33996314666596217,0.8511969709411252,0.68655918973395,0.519733556148405,
 0.8391564770467757,0.24493020472184668,0.078328657943607,0.8062404660913695,
 0.9541338261051012,0.9265965246626259,0.7223713738005478,0.3291222322884615,
 0.2155730579359335,0.8010315603490762,0.8608888372205602,0.4705718535969575,
 0.0699813427312852,0.9525218442980028,0.03696217060246387,0.8872403019497173,
 0.6228780136420369,0.5527913233728454,0.17560371208436354,0.8071170253156863,
 0.5952779597591754,0.9985209914450429,0.3107051994539548,0.8399935346192804,
 0.45974744911822674,0.9970617150542824,0.5560157602934714,0.8041832550561593,
 0.77614910126427,0.04459704768992889,0.6708652455916712,0.32031956514200555,
 0.13369536136751914,0.36865324461489535,0.1645204455349365,0.452030883367302,
 0.7209126565146096,0.2878078938671911,0.2935711889739123,
 0.0035305811671876297,0.3060901079238133,0.5849603551001656,
 0.40809915511587513,0.20353521069517433,0.9520338034735778,
 0.1968558694595569,0.4364288526205572,0.9070315980506637,0.854986982897733,
 0.7860301760183371,0.8920254575808428,0.7040881723035016,0.48859672686462274,
 0.6197621088827394,0.6929995078386066,0.47184149412420984,
 0.33680774525505186,0.6081629942827322,0.5335498928684469,0.8606110475138193,
 0.6857723949037348,0.5373946370444627,0.24360757744155848,0.9709532074157708,
 0.3000374063510276,0.06372764803542175,0.2972819039124911,0.6588112162003745,
 0.16255664589842878,0.2888923179685533,0.14045761976933469,
 0.1658512414255432,0.7266604695588855,0.11004656185292427,
 0.15899373193332555,0.5365006042892828,0.9588096189789177,0.5702040415918598,
 0.6749320318491416,0.12993553675922398,0.4682998635533243,0.667381626460984,
 0.17359375485136885,0.8542924048085572,0.4741111659476809,0.3184181312622102,
 0.6790499289103564,0.07311267060538063,0.7633790107069757,0.5174047343962301,
 0.002028705788643781,0.7080620646735203,0.6831531826228517,
 0.05596663637212873,0.9857115737687173,0.9183325630048907,0.4333849542481337,
 0.9582632083176601,0.09167315646088126,0.6169404057938334,
 0.43969696477133535,0.2766907625308155,0.41736701535068355,
 0.5845233446758927,0.2973695627319355,0.6475823744211543,0.8540432311380641,
 0.7627582948719501,0.38258666388422746,0.9894111280585705,
 0.04144201310355666,0.9985833877321973,0.4926352898609371,0.6480559021741515,
 0.4312124985578393,0.6928970624215101,0.0786492258536512,0.5332934914135525,
 0.5066419390849386,0.5418328905473416,0.0664336800345583,0.8785366154524025,
 0.4451460297832992,0.34645713772752007,0.06960523399173535,
 0.7292998062584024,0.04201769678601741,0.8321749358473332,0.5196405278320426,
 0.10072859112152388,0.9977909593877083,0.32033081241451633,
 0.4620030239342635,0.48320283509175277,0.1616763691185913,0.953474214695544,
 0.7398181021165738,0.9279676173216763,0.3227434533144361,0.084955126717837,
 0.8568232155332988,0.8579735538747506,0.5123693841022063,0.9576627550446113,
 0.4980455227389591,0.8602114538638403,0.0204346164638346,0.8138952337031262,
 0.2764906452765564,0.476187202326783,0.8943429856180842,0.17608828025501722,
 0.3641944358388365,0.3518496510426826,0.7287720609546675,0.6540691652994512,
 0.5303363966743841,0.45351534252725845,0.21654661627595828,
 0.3574681723674832,0.3581280736869469,0.458709849025865,0.07482671749869418,
 0.015480034419785627,0.964291532762824,0.452208443217634,0.4952709308890296,
 0.5094692195954967,0.09711589475762628,0.1518308158201107,
 0.23040376422362996,0.29332108283159253,0.6437941845509253,
 0.6297690154010536,0.6212256815898327,0.295265461620171,0.8347340819982385,
 0.32013239515735104,0.5272723718766654,0.06290250072140502,
 0.22008719864560355,0.6299303948980429,0.023227293825118567,
 0.1625654785418179,0.7971404200178818,0.5922677409273157,0.6433494752857644,
 0.015031213140428767,0.5003739999946379,0.21486753285424032,
 0.8745770444897996,0.30516023174432827,0.917972976952346,0.36514562026083097,
 0.5240155678268293,0.44843785490871546,0.6087729211593363,
 0.016773041232493013,0.9989678721671531,0.9125334604046503,
 0.8878297416576585,0.7462642491144811,0.5320208109316749,0.8108475918580758,
 0.09113576937036583,0.823472000059625,0.6330156700961206,0.10423491621892977,
 0.0616075442068289,0.8883881190421506,0.10907916307773413,0.6498351133188031,
 0.22989932440706728,0.40841463445713844,0.641043006702718,
 0.23833288430540356,0.6229610951580041,0.7476427486623444,0.3906222998387079,
 0.6638243355894813,0.779319557224164,0.8756905875481898,0.7611195621533648,
 0.8997647484863609,0.1340557642368565,0.8896654849172937,0.21793981318746458,
 0.34677218991703507,0.7369569047061978,0.8254365403650276,0.9525251647261355,
 0.5351098541255627,0.25082199326203425,0.277626586964244,0.6975829569825742,
 0.3853836695704309,0.40970867622575435,0.5545917873781581,0.8720788229668995,
 0.1284748720487905,0.3033177360483621,0.44260866005762944,0.1366607560619898,
 0.8277289297734398,0.8447942316115569,0.1593682448319923,0.2351119811024126,
 0.3463458535301772,0.5341685236602108,0.13883532083215688,0.5687063532613911,
 0.7218119643891368,0.34309831257809464,0.001839328568483456,
 0.5881988410934174,0.5681843783877844,0.7142834396253082,0.08280433242081875,
 0.06393598781660847,0.5619346230252298,0.8354852500030244,
 0.028516869343749085,0.9591972971112306,0.05285945783065171,
 0.7860569815214616,0.08678237190511318,0.9120591452273239,0.828625490258268,
 0.21699836413126727,0.07711686555222697,0.8515820956414596,
 0.8211537523139846,0.19988917562237307,0.7472009209996363,0.9180931403490875,
 0.4104996084729474,0.3999243577827113,0.4398434269438557,0.25538866187938836,
 0.9894310862438774,0.746946194111324,0.8709443095106822,0.11794006307524618,
 0.34343733499899853,0.8293240251589808,0.23253334916007906,
 0.7379091276232341,0.13619853559137918,0.7392442448310333,
 0.25810299623810484,0.43940935599838304,0.44824401208698605,
 0.7168590385172507,0.7588053107513582,0.753734614402954,0.6595046231553326,
 0.43617988127476726,0.43177339985622143,0.4771987530986206,
 0.7087406593508595,0.9469047172653449,0.01134973004505202,
 0.29483080351188684,0.3785734950020212,0.9579415275298344,0.9414036968570217,
 0.4499380282668296,0.2227101739576347,0.1798147717492833,0.7216002796570894,
 0.7902270848413847,0.5139780951513724,0.8512650766505582,0.5302960771356122,
 0.1736493807482642,0.5351055456570754,0.1719437112642891,0.948500919625984,
 0.11308460378434826,0.18244090995216167,0.008383034297779646,
 0.32517028710128315,0.5804706988279055,0.14313366623688872,
 0.5815135943158519,0.04335648413371518,0.49610233531993486,
 0.9701413354141346,0.966133118857309,0.7634586051997098,0.6141398360420991,
 0.5199474344199795,0.4161831114997274,0.646657046326438,0.01715760722685311,
 0.4704201491126626,0.12825849690839453,0.2847034369588818,0.1281195399000581,
 0.9324328318558078,0.04551983768824175,0.19524017124484883,
 0.35469816464018666,0.7874616650340259,0.6001314088879144,0.3007043854554936,
 0.6511703002331368,0.8179064664079919,0.38822643492982056,0.8536066817285461,
 0.46974159980831276,0.2671676818583607,0.9439556786498088,0.5858464352859204,
 0.15531958737054297,0.23556989834675013,0.3315242049316143,
 0.0385613062942316,0.07909500689845483,0.2952165457342206,0.7866531280549024,
 0.8809070221092771,0.9055337964360213,0.44326820339436257,0.7026147827310377,
 0.31243396077882624,0.560588806028432,0.2870550773257975,0.4914455462810059,
 0.8778864973862969,0.1727017774405578,0.2086524692409829,0.6259303135180615,
 0.7801322996913445,0.3006636554501243,0.46501554474966245,0.7431524480415685,
 0.8117288104504756,0.6051785775796941,0.41159797688890243,0.9976302860290296,
 0.5869300256970201,0.7861799212115907,0.3472988462357882,0.6271914544095744,
 0.5461491315535285,0.9543951392697104,0.7010689057824904,0.478234914872794,
 0.6757201440252916,0.9371122639928665,0.5339956778823767,0.04395171571638001,
 0.37426612690448835,0.001067584821430545,0.36293659288006086,
 0.06875571748368348,0.5734752688333314,0.39241830132128086,
 0.3141481357885193,0.2076694286663645,0.4574639374803493,0.6740977660580657,
 0.7403732887548063,0.865851207421302,0.34533979446378893,0.29092972363313485,
 0.5757292501349301,0.6725028807910366,0.4062963618817902,0.5993546689832863,
 0.41328960840041495,0.452677342747943,0.5757670563218542,0.5683603448086441,
 0.47190416108205924,0.6765462594314828,0.11790455648214004,
 0.5655663431113201,0.9863157602757536,0.1963472120733657,0.1394398031118278,
 0.34561740310835354,0.4821615126064007,0.6928749890033638,0.6663938095818681,
 0.22843575635298752,0.7205735169822198,0.3879379642754808,0.6375265035248383,
 0.8057760972042332,0.7566495923990159,0.7155405807322696,0.07036958512488911,
 0.875715770533942,0.5265124936947574,0.7473969451414556,0.2174457412731361,
 0.8856773686090639,0.47936950140232637,0.7375936697606859,0.9769844968016642,
 0.9078057338116494,0.29316823012184257,0.17004987821469086,
 0.5313001208191944,0.21909954925085307,0.3260006149346928,0.4311371193907154,
 0.0625432496255307,0.24166301960401293,0.9084554621435839,
 0.16499359774728883,0.1257263526966692,0.1181320857445688,
 0.17360000022696398,0.10817046939464492,0.4220821237704785,
 0.43845779311669175,0.737760992171125,0.3254773245615147,0.5425600752731184,
 0.2412835919721802,0.42992045008071955,0.5161017599227847,0.6201769259360034,
 0.6849725413554638,0.04329445880466387,0.5408366538608569,0.4975684797202288,
 0.10037085105619037,0.054627228506731074,0.6513004711853643,
 0.4788934584969978,0.9208134045358394,0.643026125507215,0.33275873209464946,
 0.32187202773883283,0.9075241572033385,0.7488245501063383,0.7344621982221834,
 0.8226737997406544,0.6658487013573049,0.7357620522244606,0.0788995851457357,
 0.20179517664382052,0.7146781317415526,0.3415167928924563,0.8047786773240821,
 0.7714390345453664,0.39911858707178105,0.8379518867379094,
 0.05942980423706257,0.782079492220227,0.6021269709257198,0.5383861968316862,
 0.3239271229098266,0.3344453529003122,0.49274698438225817,0.3634404604048521,
 0.4812245956241633,0.770375717057881,0.04485071665247364,0.5635205689670502,
 0.003405110003553591,0.6068945657712924,0.592919609366128,0.8643785782882614,
 0.4875853970602626,0.34928370698405,0.026810805858506925,0.7428559649877567,
 0.3681694106093567,0.3092902114706729,0.6322037728620752,
 0.004806462090816144,0.5577750343705805,0.21158956321946165,
 0.383184125374315,0.2576593395957514,0.30613777771863715,0.6415414267875846,
 0.9836337579976113,0.8396881765266335,0.8734769816972228,0.5996177600675757,
 0.36690933559544914,0.8002533170015981,0.4042609528706814,0.8154448229516931,
 0.4674512714602195,0.2202837767874848,0.15401848901111403,0.6854460498648566,
 0.6988576380115848,0.9950968920476247,0.05668155497580607,0.8212978872059965,
 0.4119713393749457,0.10073644264058734,0.06344368133563238,
 0.7783728824837242,0.8405623524044122,0.6174332264854027,0.5181712129814898,
 0.9101757284283835,0.31605767329887635,0.12167828838301775,0.593755692918536,
 0.21015269781459667,0.9083068297157838,0.6210901834320746,
 0.34931909282449236,0.46098064583041043,0.7267932717862752,
 0.5130763935038984,0.6217770528163814,0.16719628905678374,0.5057359563867337,
 0.8017067554517779,0.06846007756635242,0.48768864440929405,
 0.10141327904922204,0.2923411411615676,0.09817010798170689,
 0.47919626958216943,0.07973257541129264,0.13460521922311475,
 0.935025177142496,0.6411193892215076,0.18208365218918365,0.5351716978514911,
 0.2617376904126407,0.660291719257343,0.6625349213055804,0.06931939334775883,
 0.1347994221045361,0.7094948410686793,0.05272951407216475,0.5197818937277203,
 0.3286477279104386,0.7387869783750279,0.4072280868533503,0.08364557301926301,
 0.8909042333940017,0.7358095088933839,0.13075723740331868,0.646713687093734,
 0.7344585397837009,0.3803773751174111,0.5104828996873494,0.4441402192842109,
 0.6919108269853572,0.9258579980933849,0.00976342774144523,
 0.31961233914449105,0.9687761329000063,0.8349995829894551,
 0.19088725583366384,0.2968545985415034,0.5295114565234305,0.6942400894933197,
 0.01786260379712734,0.5336455396219212,0.7404729747394392,0.5224147150311321,
 0.722350607157323,0.7700782853973281,0.9084106261011249,0.2612091056168414,
 0.46646752112066625,0.1502556598394158,0.8456902494620238,0.7179896369132304,
 0.8670181711908633,0.26191895955295585,0.22377201415024472,0.71146956920487,
 0.907330791988027,0.9248400767008973,0.6581392529839147,0.6588644113720462,
 0.07948936711748533,0.046195445561763804,0.6695621391979139,
 0.8502474238646673,0.691249613708152,0.6422585197597487,0.9896698584864788,
 0.7437474535783704,0.2694892597678681,0.08170777350064717,
 0.34928410692118317,0.38973223805122204,0.37836138126132035,
 0.39457210495177286,0.5724198827141894,0.5328618423504475,
 0.44074333078756966,0.3217209519874248,0.7203581369836252,0.3953189621605475,
 0.6184253418614929,0.21733154163073354,0.6022968210999681,0.8998825901902209,
 0.6491519385643789,0.5877756797594523,0.6039617637781355,0.30952224073284107,
 0.6454620084868051,0.8343716947941183,0.6053089383622269,0.4378770615385551,
 0.02275240727217731,0.2101517127212118,0.8647098054944281,
 0.20238443713470655,0.0946745998214249,0.5606977406080393,
 0.48600292532455713,0.5464576078929885,0.582056174724292,0.6618616183231926,
 0.06366711731232655,0.18855325492271424,0.5871309154758877,
 0.10215326120889812,0.6441085376018426,0.4288781169657614,
 0.09759609548934023,0.7043635155224756,0.8886004042364124,0.3454427826240888,
 0.6142643617019485,0.596983835949888,0.978104851068164,0.7129841890063767,
 0.9390994655927113,0.4529600555596097,0.8823579259015125,0.4182537712314044,
 0.7212932497650064,0.8527491391302877,0.0392781254554736,0.798482355830991,
 0.5294098343708914,0.035700874746498745,0.17467276598982961,
 0.6068154831409998,0.29055929133555547,0.8947877288433862,0.9898192210479972,
 0.15409980468554574,0.47773853258743826,0.6244972591536995,
 0.7331175442760077,0.026327624803572536,0.3257844993631006,
 0.9590376901230041,0.15259751425298362,0.07026616490116866,
 0.22743236221278784,0.12352883538059889,0.17955022845253266,
 0.431524498578974,0.05277417517420402,0.5372563942366306,0.5096976961142632,
 0.38806362101900815,0.543235406216167,0.3622787287890311,0.7025971319048113,
 0.11199263341167232,0.23180498077961154,0.18056092756417907,
 0.020764831878802292,0.2312100316045227,0.7859345828830551,
 0.22974585344431198,0.28100373657990485,0.42643198566407636,
 0.4172884470900954,0.926247877045218,0.6515617160051517,0.22206039399023025,
 0.5803775857640895,0.21015699870054605,0.7582736383943144,
 0.12151649826968436,0.3598323626273586,0.9563007666304439,0.8063298309305443,
 0.6219847945471507,0.7893898219072855,0.684005728808577,0.5969683939313315,
 0.5179526412226185,0.7860977982984526,0.06713053151648307,0.3954150954844454,
 0.6739785274390739,0.7122569247114443,0.5289379008858033,0.21168073175795754,
 0.332471807279606,0.9375917719076553,0.32225310545085595,0.10128358707153318,
 0.44047496505687234,0.5903519158726016,0.5710434478426636,0.8618579185023926,
 0.45864744926919077,0.24748314260284077,0.5010341214560917,
 0.2404589823589247,0.5472892971563518,0.3425494231295527,0.6569883741422232,
 0.11507497966131652,0.4875085307095327,0.30607072432932325,
 0.3863945975991089,0.7243668276870637,0.8871286510518677,0.1975054546689119,
 0.3743887906437563,0.7709238230572381,0.918142488434735,0.3235976928959755,
 0.6138085709702443,0.8351743879071568,0.655976688296384,0.3143015818081113,
 0.4562756922434922,0.6930639050148548,0.5924636567656019,0.48400956803664286,
 0.9770657748903484,0.6582006721030339,0.6447310474063447,0.06531260204383837,
 0.23537984845936122,0.4925599301493665,0.1895767402064561,0.6446648338370821,
 0.16609921915285852,0.82874968127206,0.04187728303114113,0.2348958013561504,
 0.7817984645054571,0.2163808686918588,0.5673967473333194,0.6978830028139159,
 0.55896731218294,0.24111636237045297,0.39384407196388904,0.2817508837840127,
 0.04191686070828615,0.36155515247205494,0.06283011193784482,
 0.6452458289218254,0.700811203893014,0.7588758036519394,0.08337210590973154,
 0.37501004301618335,0.37137878678062997,0.6132630388615172,
 0.11625900057383431,0.2570055642563679,0.7498729611015247,0.0920290191582741,
 0.18751376782605889,0.5828060976213543,0.4452639800210745,0.5574978297381632,
 0.9970731690886234,0.7575103587783443,0.5226296068140224,0.9162683835347524,
 0.16339854900554374,0.5117792043942939,0.7096204687093629,0.6595567466739976,
 0.7740111558094463,0.07746340011253128,0.5750478214524907,0.666826371006376,
 0.8914769268707039,0.8153598153800397,0.16058250969008236,
 0.08494447934858895,0.21223093613662414,0.034692494909164706,
 0.08555836397922356,0.6353982657799608,0.38837758452908344,
 0.3561603512369581,0.12140232254531302,0.7170413553442565,
 0.08345875872680919,0.7679086320794516,0.3012877016571899,0.7721305193191117,
 0.8660997986315576,0.23372376392939076,0.06167525226350612,
 0.07206217563400241,0.5982451563960527,0.17272260794065653,
 0.6244360610241891,0.8203887061069024,0.8489046535128615,0.6110993613736728,
 0.6838994566697394,0.6024922981570856,0.21168901353513325,
 0.26844667307522596,0.0831093378998976,0.7600136771940845,0.3512606155436312,
 0.03812389166118124,0.6897042974543219,0.5704772783770427,0.5411418551534491,
 0.10823673665761135,0.5028388332450008,0.3999172032025595,
 0.32219635827680904,0.26930905718500275,0.35167427257066386,
 0.28084936286758067,0.8467928039956281,0.7832460617286082,
 0.09023768890566064,0.5254519464575461,0.440099952803758,0.7762552050888439,
 0.007826421568868924,0.6641996088246919,0.8737379061344565,
 0.5892436123489525,0.2955611125067491,0.9086851708049906,0.7003317354011029,
 0.09524898044318975,0.932377837776077,0.10463694803641899,0.6848593301550097,
 0.8592833679291448,0.6680748135389328,0.3313080889489516,0.5807938616553814,
 0.5173991507063365,0.13540526528524954,0.09791991986351578,0.60148306412866,
 0.07542348245315067,0.044569105088664185,0.18026656234328375,
 0.07872253560256659,0.7860105415209057,0.5330234949751042,0.7805617033786306,
 0.604550549092449,0.3174873252400674,0.8427994113186835,0.32139832270368796,
 0.7108950340450783,0.035864820733383485,0.7374202126770286,
 0.9653353411300648,0.6963230341033017,0.16750613006034643,0.7094558792059998,
 0.27869907105097624,0.641820099047371,0.5038563754532118,0.16359007370577194,
 0.7882209410438683,0.5489237674443983,0.005296997880768317,
 0.7271487164669272,0.8249356734675791,0.7822675903526454,0.8721864611604933,
 0.0862284215217155,0.5257933454585957,0.5044045521168317,0.06201043989701949,
 0.9391216330733103,0.9664568390146798,0.8882092633982174,0.40693805593990684,
 0.3552734103945451,0.9455784756649761,0.22851939363544038,0.8242454872799134,
 0.5066661473369811,0.04919006427564365,0.9200888620371043,
 0.28531896299910353,0.7018858203254374,0.3641806144106655,
 8.619353128862528e-4,0.0693766010648732,0.06745346784356443,
 0.5310985603017179,0.4619407221633408,0.34801688394629715,
 0.26475517699270024,0.3672881589864493,0.5364059403539054,
 0.21640937600248034,0.6951072094333866,0.0447481981046155,0.9575679541676217,
 0.5638431488276427,0.13108559599956637,0.771234733311132,0.17101063958727136,
 0.2521538726298276,0.7278445753798058,0.5171072044582067,0.67299467260197,
 0.4784041564245467,0.8390821010050867,0.33776741877045424,0.656541742964818,
 0.8555372229463921,0.1563331800590061,0.8783696958726681,0.4626346962334378,
 0.20413933404257534,0.8353735502730132,0.24425142555143176,
 0.6119236438963623,0.8608744310232922,0.9105098677069875,0.3860900112698088,
 0.10448926126594682,0.5003546353746282,0.7059085484199892,0.876265473055394,
 0.9484003481698056,0.5966605264100473,0.6201634905618474,
 0.031490181199694156,0.24853909528441664,0.479453493464569,
 0.6563301316525302,0.8043950597880497,0.866234258354942,0.4947306575460376,
 0.8493823395313582,0.39154616920582164,0.12681300865224954,
 0.5051053969945255,0.7013783775963612,0.3169362232038644,0.6908889744709518,
 0.13898649591755008,0.7241910612919478,0.5267073769971763,0.9161249219831137,
 0.15140227142099905,0.3645265778956174,0.6998156642291561,0.9111640176145521,
 0.6955183381128738,0.08310216901512357,0.8646702273006994,0.2702175340569778,
 0.24458048204724725,0.09140924313879517,0.17590151884267946,
 0.4393396839062038,0.2021208418095627,0.15449461563591715,0.634970541674366,
 0.33280820128447697,0.001221903501067767,0.7841015636485525,
 0.4710702592086422,0.3573637474118817,0.360455350126633,0.26558279148831976,
 0.6163512157624251,0.4366249727434115,0.12786799382870462,0.5298440043054365,
 0.8253733670588955,0.9616393195081863,0.31474294802112535,0.7486206819355999,
 0.8290896712547652,0.9697930974044859,0.22045984812212605,0.7329333648153403,
 0.9403194594654638,0.506573348385933,0.42270012151861436,0.8006830644959915,
 0.3902137631761302,0.35257867539769805,0.6641622951742666,0.139945242849829,
 0.5336778727793201,0.16865778381255458,0.2395715848119282,0.221155052927815,
 0.8154919622085273,0.2072596341301418,0.4961441776773191,0.6757948334878247,
 0.8287321714709995,0.6895168345903949,0.08246432394945558,0.9303828396039572,
 0.7786291478515057,0.9202834605478841,0.9211282450643334,0.09701231282893574,
 0.5215962979434341,0.7292961013087211,0.5684820651660161,0.03561100976382903,
 0.7318296957747183,0.6078774064070906,0.9017270871527314,0.6055469643702968,
 0.3202951550617049,0.5644648489297276,0.2573963198685766,0.8817413060259086,
 0.26971995094054824,0.7771492922162365,0.7496696568725004,0.5637620671444118,
 0.9072268930408621,0.8456360222238172,0.6445649789599999,0.677119250747459,
 0.5986117662989533,0.3774065830227964,0.7487302578498947,0.3302425912921346,
 0.9528802158806252,0.39205267842837677,0.2417604734755493,
 0.06810734846397559,0.7860945797160492,0.6837251362758031,0.138235714556052,
 0.6959288294470072,0.8898476357572791,0.6539550116677854,
 0.004013634576608416,0.4431293372576891,0.05577136715414388,
 0.8408166495849274,0.6027243387077219,0.6413200780210635,0.463674837581742,
 0.03943186873170024,0.7851188642611646,0.9021911323756958,0.8522760348753871,
 0.10567173120359552,0.9635993192605687,0.9837109746500982,0.4132905211489061,
 0.589156514075849,0.2779293579650023,0.005022913843437782,0.8024417711856003,
 0.48315546340518045,0.7000378539589956,0.16777121519746152,0.219329115095938,
 0.17220445668710072,0.8508693734919592,0.7172751037113102,
 0.11593617912547571,0.8039979612802249,0.18779723031306128,
 0.8499624550951578,0.18904102334254436,0.6971903429470137,
 0.006518377995819558,0.30019593094134933,0.7822146908245855,
 0.7945918402498315,0.9522012058412765,0.10981506618287573,0.6344759485341486,
 0.4706667204703191,0.3023441172823652,0.8067543532509897,0.8289473776095061,
 0.06799917929970256,0.12937682810988327,0.5047729656795048,0.831267037020716,
 0.8473792052067846,0.4269906590582122,0.9097596928296418,0.6718995290170776,
 0.3903249905428062,0.45021939934496746,0.7208752744670577,0.287404768639579,
 0.4025804024182612,0.7141329831926176,0.12311977352660786,0.9605500550762861,
 0.7899354271419868,0.04368939876350264,4.200850866682426e-4,
 0.8987038422371854,0.42410255162284227,0.8402941675371474,0.4824769903814934,
 0.6704470566998075,0.3762175103837182,0.6884048206514615,0.7675174892803582,
 0.2385102340785933,0.8030884633858222,0.45339302895504474,0.645764812000735,
 0.8429526305465184,0.4378069792447029,0.8109995266665038,0.36389822484424716,
 0.5286337365480698,0.6785850457131177,0.775282770323239,0.8254063732512281,
 0.01484048490663703,0.5608144531715817,0.4360445410668975,
 0.41553713827117533,0.5315350887174153,0.9433117724028901,
 0.03424998177737937,0.40300056225494163,0.17592387141470733,
 0.1023666604977278,0.6899206535881339,0.8363469071837211,0.6250617416724767,
 0.8349582907614584,0.3677454029044215,0.0023260053994661467,
 0.2576502002027211,0.047596353743096675,0.44977644631879443,
 0.17619477581351772,0.5478477405280977,0.8027117374190225,0.9445640899736404,
 0.8330016594736113,0.051707510458869566,0.02448938988976479,
 0.6035679363817108,0.677748821167979,0.37379259846114654,0.13815611708672748,
 0.6571506077202622,0.054197763726010484,0.7042876004666134,0.277069444681865,
 0.9119371748914865,0.8855878158130255,0.7659086625244784,0.5567256658103336,
 0.6088837164606236,0.33787610123385214,0.778553673203725,0.5136815376395281,
 0.6321897647904503,0.26708846930497077,0.9439850564909485,0.5232843114845631,
 0.18295715826341452,0.1564410824042889,0.11190152144998389,
 0.40338647614823864,0.8346181951051843,0.5294635692812036,0.8188071804635875,
 0.36047833188237144,0.1751664972039212,0.761116820711917,0.2728703178890839,
 0.19817578543064107,0.8929398247174786,0.9111296330181078,0.4028295763411952,
 0.08223050715517044,0.332047709216833,0.837730096406221,0.3020619581868844,
 0.6769312898286018,0.7215638406765722,0.49294232577938957,
 0.11357205249918512,0.46434585257176864,0.25056277245125846,
 0.10012151820577342,0.8930298023518677,0.3630229780178138,
 0.057312795262425165,0.11975784715776716,0.9932914254752434,
 0.2803286706236101,0.3222487412014282,0.8498128013146362,
 0.016000623619229826,0.3770340644364092,0.9716293526575123,
 0.10005496989556895,0.6308908800048543,0.2952206627546924,0.1655312696724267,
 0.5675884559043449,0.9129652290712197,0.7410859076806229,0.7390680361592461,
 0.6301955182565393,0.733060033034207,0.666158011511084,0.23074959583249455,
 0.41240574078126513,0.6511212597524613,0.2269771123011639,0.9554019040496915,
 0.18777346490866798,0.8599885334593081,0.492328326307494,0.8270479986187663,
 0.4967862238841052,0.47825317080636776,0.2879502395476765,
 0.0016361935421705942,0.35833676082965105,0.13089983670825878,
 0.916276200034188,0.7106613133927911,0.17630260028812594,
 0.029625172417601875,0.4124207338120599,0.4048466100419439,
 0.5095262967348079,0.07470532334776858,0.048536122940383675,
 0.879247829603551,0.6371908922800502,0.8227450097979979,0.9511156846196392,
 0.45744610471828095,0.10756283934221966,0.4389771944572082,
 0.8251270747277895,0.8615765952963236,0.5844530119565492,0.7934813804738786,
 0.7407835993285998,0.12046876023827147,0.5160948652600567,0.0274612182178553,
 0.6869728755675395,0.6717855221208938,0.16828849230052068,
 0.20195940002101798,0.6493173202814913,0.5669534307781658,0.7791269402384438,
 0.165823726387747,0.05036405840190028,0.3462518311073133,0.4975201832244156,
 0.3603623541179418,0.2579125615691311,0.9930559999298132,0.15788791597698593,
 0.6388567652554062,0.5371316722163686,0.9686876836679154,0.39494587626199673,
 0.14130454081936916,0.3658290261836671,0.7403754223354013,0.505441609266029,
 0.5866810748271069,0.11038011652345958,0.07691688231751448,
 0.3219604933939596,0.7589566054022195,0.7509806126247252,0.346608940143735,
 0.38354280114040795,0.684424786393815,0.36867623493716173,0.703608248808321,
 0.7087430956573021,0.04868047024045685,0.30269473275211145,
 0.8488952663956365,0.41675326991100203,0.8347688164000507,0.5479273691833312,
 0.9865743757653154,0.1504320487807963,0.7803948578961043,0.43398659400975115,
 0.8445380508835449,0.6233765078432418,0.9031888749564106,0.9441223279400234,
 0.3215334374460105,0.9883784675116768,0.5435529708684715,0.06759869478700464,
 0.2639333898634777,0.403375704982211,0.8131650639721644,0.2293468145122891,
 0.7714802426908829,0.6063234230507356,0.9534854725679682,0.3982543634757286,
 0.7584780029970327,0.3330947456990936,0.265344697543777,0.6050825633026258,
 0.13591495733824632,0.9187786941673798,0.6796550181029452,0.2665213181094699,
 0.6137710208008335,0.23076735760557987,0.8763264842794759,0.9503734836666636,
 0.9150182648056633,0.1502921861401435,0.023320303906935846,
 0.2534740616752976,0.4817028537493713,0.5456198315019739,0.9740877477961457,
 0.7547940704881807,0.9538888778177307,0.13220807963984116,
 0.46514117778968833,0.5228641151379168,0.3318774511661572,0.7088546037203607,
 0.3101178488965467,0.4403343365990351,0.035057583444935814,
 0.45756048351487233,0.1419167712134134,0.864972107732619,0.6781033622774738,
 0.9920168201487433,0.1399049721490082,0.9319510146982185,0.8110084049124802,
 0.24408987845063257,0.97260773790888,0.5132678314038669,0.2689008636750977,
 0.03654328894674408,0.060008458923580865,0.18654869894765524,
 0.27443550219580537,0.09961836967121729,0.680200254540849,
 0.37127202839872053,0.48453899526593114,0.5864666630987685,
 0.1618095429057913,0.6355463922927311,0.2779588092104801,
 0.014117625363100239,0.7285874673601132,0.4150648832610798,
 0.9562469198533763,0.430326721134493,0.8778462465521442,0.2939534020311243,
 0.07633154604322123,0.7998264476827561,0.3765659672999029,0.5605281479626762,
 0.4325962842192822,0.8370023459447135,0.20020620745154016,0.1996007878340188,
 0.3965219764804885,0.8655862892816825,0.26366001260014427,0.8633500594085454,
 0.4358726645972335,0.20242732265810115,0.11392022412634717,
 0.0071917016323723715,0.7999932803625862,0.5476734507581608,
 0.11890083821296893,0.5770790078199536,0.3626115710179536,0.6111649009107821,
 0.22468387136682466,0.3200782170504318,0.24793351411591802,
 0.2692712115396023,0.4275490089504095,0.16873580092532892,0.339325807771106,
 0.6169998412461273,0.1279600798096001,0.3306742328545538,
 0.016151250152558694,0.5826907643027555,0.9332189870728411,
 0.41980375462504826,0.8814054009307436,0.9307918162570815,
 0.11382551968097299,0.7328235198286538,0.549711775127176,0.6183014730866254,
 0.9133302009066251,0.7358478111219791,0.7437331810982238,0.6710254465363233,
 0.3275557789564284,0.9145817095623648,0.5316084922489592,0.390428715366425,
 0.7814035442138499,0.0029098488714004445,0.690060179690977,
 0.2207685524137435,0.6446960953920885,0.11510818098362441,0.1514550316947434,
 0.3095980569415835,0.7573827759671707,0.010465537773465372,
 0.13149014487287214,0.7158852122938097,0.49097017948098043,
 0.41772877440921663,0.6150086322695087,0.5794136193737245,0.492496969152148,
 0.9380690240583818,0.9090117895636456,0.49857074060800666,0.5574240793066159,
 0.9821916408920133,0.8743026659549558,0.7507843873235278,0.7862485428946923,
 0.8704753952986046,0.6211895672232396,0.7787431477530624,0.4127189400108613,
 0.19917579688632936,0.5733060754758695,0.90256537255525,0.873576263702572,
 0.9158092916669194,0.4995917680022417,0.2615439698393126,0.8563500930777771,
 0.9285554436963629,0.8391255242535403,0.7698437984004742,0.7760801850161498,
 0.28274488218016447,0.009811272013695915,0.7601879305327658,
 0.5972710921929115,0.4205397091904626,0.7233639518821681,0.9406885600261129,
 0.13248708105440365,0.16496986226036303,0.10845137903743873,
 0.09057582609925274,0.2341441481707789,0.2576520188557402,0.8122711049572853,
 0.9709427669139188,0.39635216427676634,0.18613094919259676,
 0.8199067302335838,0.9091645448532735,0.10479991458376392,
 0.43655391177123026,0.012971748566978802,0.8643984214166893,
 0.16314546932195972,0.669711519763606,0.604629852128316,0.7377150905621559,
 0.08017901839522024,0.5840852514859836,0.17931199070725257,
 0.4651071883208091,0.7873408593642359,0.7828342801118567,0.8741257231110272,
 0.8434311942467722,0.7419381766752429,0.14860321119750308,
 0.29380036074441473,0.12205895548854817,0.32925634267827,0.9924992480013453,
 0.36036518930510686,0.8732514194040026,0.8166815253307174,0.7604986599802961,
 0.7085360256093592,0.850701726090938,0.43689916086406866,0.35681123970429585,
 0.2604533899892738,0.10126459705395974,0.6072759267144417,0.7232871190885964,
 0.38356547283931197,0.5008835573467253,0.9160913667583689,0.8765153300482572,
 0.10413295249989996,0.5716144417439157,0.9046819228997296,0.9371571741535791,
 0.9514171099028844,0.323209450694124,0.6368309353492938,0.7668806092394067,
 0.37180809883295063,0.8691984281796877,0.9621482819863285,
 0.15996736343100548,0.12378616228505157,0.7304473832562857,
 0.4619857852563275,0.4918461511542326,0.4216807699397668,0.5549657622258324,
 0.686874885360897,0.5232738827424096,0.35196808429118775,0.8658295571215702,
 0.7551787699917466,0.1845799685733408,0.381639274200283,0.02215384064606385,
 0.9627319978331914,0.6012358117587459,0.14744203903121245,
 0.18809305550989186,0.20759473737040401,0.9960045121826391,
 0.07167045493322455,0.24969864638530181,0.8732094308027601,
 0.5147184824440285,0.7070802390914437,0.3806075624535924,0.41183050199570037,
 0.035071931124942446,0.5937370413378923,0.2581637876311471,
 0.1990795015035992,0.024704874258862475,0.6570289750014775,
 0.7018516354160113,0.982044758136356,0.5728762762180039,0.8112827413042997,
 0.08477319087714241,0.3953766462540109,0.823142769111163,0.16969826350622952,
 0.5244494506204941,0.16826084161850496,0.3368710289325485,0.4845417248064148,
 0.7457565897062014,0.6581407457325671,0.6682917712060248,0.21835009074812395,
 0.014662786362359626,0.8700019160037056,0.13784895773695727,
 0.8446061674892467,0.035398479597609,0.7494235188927967,0.9721519437261106,
 0.003328219043945757,0.39334089617717727,0.5542976438006171,
 0.16900166295686125,0.31065456294000693,0.6239014914674924,
 0.7855798721028594,0.8674171279259819,0.5876376715526855,
 0.013392437997685525,0.9580760525782539,0.8491388449269607,
 0.9449931183743197,0.7773322896714495,0.847532743825118,0.5679859847691396,
 0.7690278091213671,0.039316343480843674,0.3838011406109475,
 0.2172484249594847,0.25170059275234613,0.18200264151600298,
 0.6370288782536021,0.9698219593335156,0.4054359294699911,0.7062168921206128,
 0.23317390743950883,0.014686109826043237,0.9307126696325561,0.30975395266107,
 0.005775220666831116,0.038919661486416235,0.1814517640303417,
 0.6556716159146398,0.5195857498163949,0.14707652977739438,0.8663922664615586,
 0.07722799454859741,0.42203901931860455,0.3457514137243457,0.773286796506426,
 0.9697110764016086,0.7947753342271616,0.9002691509452505,0.3648180639113259,
 0.8408043320989486,0.27948925898798715,0.8217360068024198,0.9716804499147398,
 0.9226945699198361,0.22092329108125608,0.03289134240379721,
 0.023930417119436442,0.365551175443372,0.651915756000107,0.22873580471876054,
 0.8405050832526705,0.8610679264424358,0.21571015410422123,
 0.04677513877528994,0.4807793709044499,0.5722774058982882,0.5269540047328939,
 0.14253929923179287,0.845604927865499,0.4549101107336567,0.9255535381811806,
 0.9396064119159658,0.032092258746933555,0.3704347336760788,
 0.10080958458767997,0.8596988142854806,0.7514330501312747,0.9419596981645801,
 0.8860332634443896,0.8660609318158247,0.766466061691167,0.3045216315352828,
 0.3290727927470427,0.3858016965556521,0.03316016895256546,0.9345773409728306,
 0.6705009894327438,0.1860000849581196,0.002522676611982244,
 0.20680724639458825,0.17497680770896806,0.6964962810272586,0.73046148554004,
 0.35778701544914226,0.7286457750770445,0.5583809940217089,0.3093100635408633,
 0.8463110194227887,0.04428243258720255,0.7022992767935341,0.3617092586557644,
 0.06660972908500506,0.5844930755583168,0.9707470684737054,
 0.16686430740044977,0.3634769742176135,0.9109425567131906,
 0.02466587175972934,0.16224385385194762,0.3236580421211388,
 0.6790921632626286,0.8020827874455478,0.6211375604115017,0.6351661954209554,
 0.9914489851159585,0.09032461357214494,0.7620856414865227,
 0.48879519554127593,0.5146860429355422,0.5785766399594872,
 0.11013012883406514,0.5930075872479015,0.3147562316777439,
 0.24027514324436383,0.8498836060220807,0.0419222432111932,0.3873484504642576,
 0.2708845442578327,0.7499999206068773,0.37677588959493213,0.6513490493070104,
 0.35607069903377964,0.8690887036714935,0.386825007197541,0.7166362267306704,
 0.5068254529109208,0.8346437853912291,0.951159682216565,0.40623173741806273,
 0.8081393480872119,0.0377295866944094,0.7408690022888808,0.7066425607758449,
 0.025024180509985228,0.32112401602781404,0.6405185570071902,
 0.7173889926876598,0.9065651440981352,0.30344482224603764,0.8850132136046422,
 0.23754928611498083,0.2687877742716982,0.5287587128511924,0.225916096967677,
 0.48293396405580935,0.2753395715121074,0.6006358934953366,0.6117932527107912,
 0.3096706700444596,0.2281990765968025,0.15673424543243786,0.4180711393226967,
 0.0902098823044345,0.8039129882729528,0.9955906411953327,0.27837507666898353,
 0.8316143262837715,0.5683380218860763,0.5651533667093908,0.17438468258093343,
 0.10227014253847622,0.6871992385612985,0.5143240599881891,0.9384735997300087,
 0.9166234937518855,0.610443484134624,0.6328597982304,0.8534515520768402,
 0.36678888020918854,0.1915526174919645,0.3052935123484798,
 0.028491862871145024,0.5252126418602187,0.7348315031264956,
 0.6787481491796353,0.8730153721300492,0.8847806879771988,0.28096158702144525,
 0.5693211501725193,0.03560678294819075,0.13526246351577897,
 0.031406662924682416,0.48596406806303216,0.9111123162923285,
 0.5704135511514971,0.8114888693463004,0.8934065284239541,0.3565135878109541,
 0.8123894447129746,0.3538487487166331,0.9186914433434313,0.8315366906911154,
 0.5006985659741487,0.2848985213641072,0.20036289663687157,0.3224339250264283,
 0.9862994869015068,0.4701675108407599,0.966908788906728,0.6536358688076844,
 0.01568719583406397,0.8837963681596871,0.24932672659555588,
 0.49453100716631393,0.30556474027963865,0.06410433039218677,
 0.257007347587363,0.690303573426229,0.3622835214723418,0.933570628492132,
 0.0881712048062947,0.9240892468269698,0.9532567427657251,0.701925798639698,
 0.33474740907017453,0.28532637069430256,0.7296823922095269,
 0.6587367139640162,0.7139293727734808,0.5377522450278556,0.02879546311605763,
 0.25538944652052553,0.7945832209102518,0.6648235041846777,0.5824144358030017,
 0.6575321363184106,0.38075417725716854,0.4493880425391579,
 0.48867404804857406,0.16394976869667688,0.14012692980820485,
 0.3570673119547706,0.4853760984898976,0.8755863952596473,0.5904796769548006,
 0.1707518658015189,0.28271853874898434,0.9174278209971121,0.8710299309207866,
 0.5361505856859843,0.22097997411985215,0.9946775582437957,0.7039239922597932,
 0.7638815146505787,0.8714761588897102,0.5184880433068488,0.5544730919343756,
 0.5822075425955576,0.8446358338535828,0.9471647935068163,0.5617458787680082,
 0.04886041069446989,0.039710667670204924,0.669830411664779,
 0.3737923928441137,0.14597763233920813,0.9265451636779531,0.7452242156353646,
 0.6987317292635251,0.11309801538753772,0.7648702136726269,
 0.35111904617479517,0.24000310066726438,0.01471457564061529,
 0.7208698845589097,0.2736840987470365,0.3983522115941016,0.5243343595127362,
 0.5278358115269562,0.29230224923844506,0.3069301644860336,
 0.07183468218218003,0.3986563671618474,0.30682211826408534,0.411931313525979,
 0.17611514312793208,0.6932129508519487,0.8651786839861394,0.4864851507612099,
 0.8341120851709563,0.38722873658155166,0.8074775381562692,
 0.48001802013345185,0.608943882039767,0.6721151453050449,0.7546287002817393,
 0.47453735747966197,0.6747956082077154,0.51835676324911,0.3468997441943873,
 0.12636532318138805,0.3434653650467113,0.8423334997525328,0.5699830405970306,
 0.1967790257586558,0.10356306536243676,0.7180106001816049,0.7714187649879705,
 0.7901468086707403,0.4723052282572412,0.2371510612022374,0.07112950495390102,
 0.3257962704711497,0.03577349684753961,0.4843201537546593,0.6206339053306101,
 0.5840302650471056,0.13787393682847926,0.6664468733642338,0.9494675121027033,
 0.6962992117777987,0.2632027318611936,0.6045976581830936,0.5977433776391899,
 0.7549857937892965,0.7368750617743849,0.9717874391965792,0.07921120724001418,
 0.8396178108790971,0.3580184467982619,0.5683442269083585,0.8591483205312136,
 0.5933238495619972,0.08023726386014385,0.36875059270287347,
 0.11743338539870973,0.13774152535066098,0.12336973673768015,
 0.4047486774385707,0.9952152557757037,0.4516158517267972,0.5604270181517188,
 0.7196702171088314,0.6133298600064945,0.7520237681719967,
 0.011720007752030415,0.748170133967921,0.9676217883937647,0.5441870255544613,
 0.44366791151862084,0.5257729264925171,0.8599924055543238,
 0.35115167663341973,0.1540374786345996,0.5791686926413369,0.5008912269396459,
 0.7296382566115096,0.41581552325968696,0.48188423702019634,
 0.3245017305054425,0.002445390885243981,0.42242489638363845,
 0.002969706630232327,0.47805911779131227,0.6257762569131525,
 0.1588426620896688,0.1191410960065592,0.2899643676055642,0.1486064049901774,
 0.7633176017536523,0.4080312277386804,0.0015638110642538505,
 0.005945720730709647,0.6445566134278634,0.6650474883414047,
 0.4007519808692469,0.13649725958205694,0.29795919059678666,0.386116977596227,
 0.6825251658007349,0.6145786401799798,0.7770135072305443,
 2.8873451993882693e-4,0.7951147880209222,0.2362805959609232,
 0.4512697193470515,0.36590751131862986,0.4989321910029486,0.6603116426105439,
 0.5168721841750481,0.38440177151877286,0.26655644090917574,
 0.5308173663858106,0.4688026054965755,0.4811003656577686,0.6307477452076602,
 0.8152450095751685,0.38681661247165766,0.22724588218720843,
 0.012960375906454269,0.04214495607142488,0.1734995998189951,
 0.2523750958208344,0.2856252280311221,0.19700437984645247,0.7295717016795842,
 0.005874016314147856,0.37973872430104616,0.26552755684618956,
 0.8280252246967349,0.9712330841954977,0.914643419451871,0.901380801634961,
 0.3213208258012613,0.5830524848882523,0.22376525742666997,0.9897641625659939,
 0.18166588641982462,0.7909211917643991,0.9805520309025015,
 0.29556017609051555,0.6188310323210755,0.977113982091304,0.5606536168427414,
 0.36473774796785885,0.5938384762467681,0.41890883413943214,
 0.7387313076632405,0.18429408092734345,0.7051853146063053,
 0.04558240569395533,0.739775072662068,0.057873275961902615,0.735072381006454,
 0.3563501122417172,0.673671046126532,0.07390231923038248,0.6143699597964268,
 0.5923257990777885,0.2169739352683755,0.9865863306605542,0.823301837466826,
 0.8007420433073371,0.34174030764281216,0.14375452915863052,
 0.03866204135003637,0.38341713533148647,0.1386246536644664,
 0.38235465490200615,0.19069138993398516,0.4259411597536761,
 0.6640482799707469,0.2653221992849144,0.5674519313071231,0.4330135788341659,
 0.45173155352082706,0.054194485614908805,0.16032719861226552,
 0.3800670472121994,0.7217436294368543,0.8339970362045086,0.9504612868119379,
 0.05717488072379484,0.17500238457320116,0.5201517995329874,
 0.3046123703289998,0.14358813042277996,0.009036887828624551,
 0.20390351066965084,0.8933489447036805,0.9422756468440341,
 0.14002709671774183,0.2509353447336429,0.9777281243569355,
 0.13393161189685765,0.4066062246922182,0.670389328879126,0.13359727252891407,
 0.7417127756992776,0.8230664197153231,0.941628562137688,0.4930905763036203,
 0.6352366514969127,0.789920773531508,0.8334962931271538,0.060297670215240684,
 0.04568517020886098,0.22065992035032034,0.6258297179466169,
 0.5140171099115056,0.31911598179539935,0.9513002589674263,0.944460724606638,
 0.6895362079614127,0.5047721934875663,0.4181944970295941,0.2446032003454589,
 0.37066399193604593,0.8807036559205811,0.11228168075493661,
 0.04091719648364012,0.17851520019495792,0.7100868281076432,
 0.9921391800181619,0.24647701161538216,0.3368184347513865,
 0.45314888510287743,0.6097374875291062,0.9780430757944106,0.9647225181087178,
 0.8492610134438657,0.695004293768083,0.9064607535296916,0.5005482711276639,
 0.9257377539003214,0.13208219802708088,0.6676511130469646,0.8887171328268356,
 0.2365487726650738,0.5475052850783166,0.8257032150979395,0.4644795360450992,
 0.2271153066516295,0.05297867461902295,0.4829369092379171,
 0.12955264345753947,0.013133153121577146,0.9762446158104933,
 0.9828863817131708,0.1525906508699597,0.18665543810313356,0.3820027351475749,
 0.05132748816220556,0.14929159247936508,0.4201563893520174,
 0.5980794023322032,0.3906771568104067,0.41963735793137547,
 0.25590355557387756,0.7728281987777837,0.23485162045322383,
 0.8669585685311997,0.31094023814280936,0.5544205235427174,
 0.33613002822112703,0.9789681750917532,0.1230066621273509,0.3186802415791702,
 0.8749196300193165,0.3295675016697992,0.6452740774187882,0.03415505726125734,
 0.09768103391183436,0.48640067689587896,0.5725557434492163,
 0.36382499841352345,0.1236018202204231,0.9589277346765228,
 0.16970982614422137,0.9326725013096586,0.7245293065690348,0.6760757181226937,
 0.021586361765476236,0.9314435607373375,0.7272792604521856,
 0.3319356699744954,0.6849586090797912,0.2968631775382755,0.9203182840564927,
 0.3855677435075512,0.8758908532321232,0.8883929997665186,0.12584382311269937,
 0.5673049692774942,0.12100165333091051,0.08160849460542241,
 0.058151538276884795,0.7947096009109522,0.7193318498496601,
 0.6015637656835184,0.07545533205507615,0.4882611458129189,0.0635406274378495,
 0.31633291031476574,0.3968640898527086,0.4223407863734352,0.2769766061651542,
 0.9632662194309672,0.4150677535557,0.26074931575816596,0.15656054199477476,
 0.18250842785140764,0.10625876090575026,0.5318269161728997,
 0.4135115709332603,0.31765664262101645,0.7483855864754936,0.5376096210463439,
 0.8474726937766986,0.0653498399692477,0.6964524105526935,0.47782725995351283,
 0.0830986344250908,0.2885092182788721,0.004956921324051455,
 0.9673024092948884,0.6616686065865504,0.406497764592172,0.5923699285117169,
 0.6507011873129481,0.12328202591927895,0.008843709509950282,
 0.24096286637438158,0.5463672483064562,0.6637767023589478,0.5143122558098917,
 0.7399857489698901,0.8843730563564299,0.6323190314516797,0.9422001103129599,
 0.40669164185921414,0.41795306494463724,0.17764648082684187,
 0.5563275212242133,0.18566698022905637,0.5478223569442177,0.2470026258219884,
 0.45294053285982605,0.8306003750472635,0.25985597297335794,
 0.22626094415217324,0.4735447784173158,0.9223842038145279,
 0.26389049980238144,0.9367929438233467,0.6576019020073157,0.646694444167237,
 0.4618289401375426,0.41149911974531417,0.5064280220310745,0.4071954292915426,
 0.6127758742481971,0.3736008342449756,0.8682482735415497,0.9171612348024467,
 0.20662146682326354,0.4342734776650162,0.7505634964022846,0.6594977756827866,
 0.6466608369293718,0.2079834198876901,0.8896797914619815,0.7754393553266181,
 0.6916615690887065,0.5136228524312637,0.11929491414879045,
 0.27541579154147733,0.902122108399878,0.831873288463106,0.8234665518298095,
 0.9787216116828276,0.2741206822461355,0.8826598996960479,0.43191936089049854,
 0.2244867958356943,0.24540313194098273,0.7933837407019089,0.6284693596818193,
 0.2189434287127332,0.8276101978300483,0.9676022136494778,0.21294820321465824,
 0.5449226448640313,0.9830991178022199,0.9679521801965422,
 0.010968776301879868,0.21967113535778282,0.36285254223026897,
 0.7764795816605602,0.6746462436071785,0.7447431915482464,0.9961468717197794,
 0.6802432194840229,0.030294282969645314,0.909320974017365,0.7791927961048601,
 0.8662136155469131,0.5657059301074013,0.785951065262576,0.699631122637125,
 0.2166504172408792,0.6050325749329684,0.27588741209816137,0.6565942894133121,
 0.19473558363739418,0.600813517337555,0.5682976932312753,0.13113996304212983,
 0.9391325676856189,0.28840370318582575,0.8759902404984441,0.9343026853225052,
 0.46734217986860505,0.34441526668213285,0.29941741655002974,
 0.5419803624561831,0.13293303273678436,0.7310331886455456,
 0.20699878214690148,0.671674434602795,0.693063060238751,0.14756355404200838,
 0.863129035714382,0.24671353919558026,0.18988523607226404,0.482217579772213,
 0.8342925283203247,0.29971326452025315,0.20807609609221134,
 0.1243283405433685,0.042070955417193634,0.47455471509024105,
 0.10834890555666288,0.2645012745743466,0.10008791567762299,
 0.44939057780590086,0.9138470112809162,0.2509791475830979,0.8439728300812614,
 0.6696475775146972,0.8074139818827472,0.8018323007231591,0.3379278838645766,
 0.5034171052028251,0.5049749486933455,0.08688014432821534,
 0.03604476886213859,0.403432780898063,0.8045218282544505,0.345311108960937,
 0.912359196785376,0.5357918701662309,0.20104553435907668,0.6834285232649581,
 0.81580134268468,0.15891405215375298,0.7356146523788503,0.6726059639903853,
 0.17195771978571917,0.3063020977216714,0.28580744005837655,
 0.19579262352285676,0.9101586736763998,0.14782719595090565,
 0.8562886267850188,0.5215195443388242,0.2105477614782887,0.5164280160752386,
 0.9405372127241256,0.6933449794263722,0.8955016011239161,0.06680457700561615,
 0.7152108604824414,0.9685449929406115,0.9158273199295641,0.6132280336911146,
 0.6420722169928155,0.9043309786011282,0.29910538305439593,0.8824447232178704,
 0.299147171548528,0.2483739689149671,0.9317242369980356,0.14584166937323706,
 0.29595886995044407,0.6839572922476693,0.47588740085510317,
 0.16213734331706986,0.8052078439045959,0.24836942050052702,
 0.5684342007375228,0.2799669196790331,0.2793852559929645,0.15395642868709114,
 0.3517578273856652,0.8060407139439896,0.649938193320928,0.3434622549187989,
 0.5294233427484862,0.8114086394277832,0.34230852080020047,0.1210467208427235,
 0.2776416806303901,0.9916467470632616,0.14858920767465522,
 0.22826384502279573,0.5759058274138844,0.8007748847816152,0.9257340834589354,
 0.32975617977180316,0.6629137875067206,0.7645581591859808,0.8660096813460282,
 0.5869254204986897,0.5138143218084654,0.7088616016496296,0.6972462738894681,
 0.9044875183415542,0.9403735233086215,0.43925927284432875,
 0.15726790370552302,0.3682040265591201,0.16817978826954605,
 0.23914296263819368,0.6050112726518035,0.21459448721286767,
 0.060362172887327814,0.12257213333746264,0.1695420286329471,
 0.3622953140187182,0.19031952828692145,0.530716243788392,0.11610312275702594,
 0.5659304154588701,0.9797443286820433,0.13052579111033302,0.4362296648975461,
 0.43259262781198904,0.008330638192383,0.9646549569852839,0.1820338171962289,
 0.34409663065169616,0.5982544063143044,0.9965361294085817,0.4802231215663725,
 0.048978416771364386,0.495153503486205,0.727617040418984,0.31692068984942345,
 0.3032674814532552,0.9329700504060509,0.0580343823640983,0.8716147884561599,
 0.865838925220341,0.6376276146345745,0.4971964433655528,0.32596112464251326,
 0.578366636702369,0.7471696045579018,0.3037331363729311,0.7061213372100957,
 0.3952229355972856,0.04677163410451002,0.14921324558181104,0.242642676937727,
 0.17346805981382207,0.06842413448601292,0.651212320664673,
 0.01767356021247446,0.4338499136974945,0.2160177120254928,
 0.08815770160561176,0.07087023048080976,0.5791386965066612,0.879098910782535,
 0.5212331297773262,0.7224039929174821,0.7404883263241018,0.6111624616397126,
 0.9360953765522987,0.8977938927258732,0.45001703292231443,0.6895427254828091,
 0.31567049918257184,0.6430971439130652,0.14860402044620358,
 0.8690678399842282,0.2217755824075519,0.9019600317925554,0.739743383971014,
 0.9742562492865935,0.8918248308007091,0.6491801217809196,0.8311149670715337,
 0.4291406725095329,0.6684012867318972,0.06904525808778839,0.8571937273045724,
 0.5597427974964749,0.3912088695024345,0.29551192288993444,0.3406132790661771,
 0.20286391891773148,0.4767349087185351,0.47142000588479593,
 0.3095416138296615,0.03209632966495457,0.06328630996921625,
 0.04203688039364861,0.7071091297135932,0.953872518997046,0.22876447478134232,
 0.05900274805926864,0.9057076575555985,0.8595665097571807,
 0.021411961015188163,0.3609656222818016,0.437242012728419,
 0.016134261526436933,0.3704453344388451,0.8696778179496383,
 0.5108051967577498,0.6799456902780269,0.41300682471792505,
 0.37679914675022375,0.8057285508488116,0.7259948426376366,
 0.05924175331852399,0.023608455646357274,0.43659817575208026,
 0.9715642554306866,0.7653138052388546,0.47975403145047757,0.7801635772508199,
 0.33872960840738364,0.2901321121943461,0.19562138736992196,
 0.5643143612092569,0.8412392816456415,0.8422028392035792,0.28889101759944646,
 0.6802545052465858,0.9342784131877302,0.707629909761651,0.26997639952846886,
 0.519187467591088,0.33898010702479486,0.8188025118580596,0.577532374524471,
 0.5885398673462517,0.3983549162554286,0.9428104755950077,0.5275249472392486,
 0.016569262072125723,0.19748338918490915,0.9899520081453408,
 0.22641622292206065,0.06715910747821896,0.5414981585898786,
 0.033398075714326514,0.2809553150699,0.7470772695667013,0.29494512260728367,
 0.9426842190778483,0.7744766598807319,0.36307372614887234,0.4592076272839982,
 0.7544577697687432,0.6670199432788333,0.24113442820635478,
 0.054283919328368624,0.4034295422788956,0.6323632433227933,
 0.8598114512863277,0.725148164962373,0.6630188606481813,0.0867720872296569,
 0.7884434172263787,0.13106207750581145,0.9403483953898294,0.9389001255080205,
 0.5843247854295383,0.5790820524111404,0.024894642511510745,
 0.7186197273352701,0.7133240634072654,0.05898962612459846,0.4175162118943061,
 0.8992506772603248,0.31500801871724193,0.614317633887965,0.04244576949413259,
 0.24279181755732504,0.19238264430580254,0.09146988831894276,
 0.46853867889684053,0.025673861690308164,0.2013709641671131,
 0.9357138343268734,0.21824348194764764,0.23393550217108594,
 0.46986405918178964,0.6905547284665152,0.2562025144168032,
 0.32390882835095813,0.4167370125260288,0.7441531521539372,
 0.04369314799908708,0.6984948995445688,0.10041812091009394,
 0.8482099670287149,0.5431846086732594,0.5446247851270151,
 0.019309258771082694,0.2910422057690968,0.9932070472685451,
 0.01566610463829199,0.3791565932543721,0.7459785849046626,0.6790321401523215,
 0.08644814055659134,0.14391782368553918,0.9997709863498835,
 0.5421303075108452,0.48045375220071407,0.45825067537943154,
 0.4684357121801107,0.7981630265155211,0.9801430392856469,0.9330734582046301,
 0.3431142435082153,0.005973409307507116,0.2901642175425677,
 0.29143499197764555,0.9722612328215803,0.4686623605608742,
 0.14550866589787548,0.5960425989181298,0.7798259942544352,0.767601382425011,
 0.6277626370880205,0.4538747731994148,0.8146809236466603,0.1040348415865141,
 0.6346523761159324,0.22546502804875934,0.9174136261078938,0.5442505399953713,
 0.33571265711618303,0.4188580739185159,0.3981602623444678,0.9997474385315835,
 0.6602677853386529,0.78837073276295,0.7167633281999183,0.7857474038461689,
 0.1621088936594507,0.36630153700701507,0.07662127981278266,
 0.6569473538042645,0.8730590502296255,0.19468544469993077,0.6368314019527704,
 0.4764625884081207,0.8797049788867262,0.5377554340893577,0.2432316275612343,
 0.0652296972238956,0.7167282250073806,0.9411685107018859,0.8256386058033683,
 0.9533259199392892,0.16022602686668408,0.14970426689608085,
 0.1480718652024965,0.022685569381736714,0.7692868666993606,
 0.9736751069960781,0.18813258138080113,0.7727143922739517,0.714189509326599,
 0.2662679458254258,0.5968538995882534,0.08866034905170972,
 0.46332540650784115,0.8541655265641537,0.3653702742714965,0.4212795546228554,
 0.21925051649849692,0.15675250207162528,0.529713560582258,
 0.008044276974853992,0.8222298595979303,0.6160113618424458,
 0.8698371321559921,0.8559137900335564,0.2894161405181479,0.8109848894356115,
 0.08812279309586624,0.33174640840119995,0.9680730355949732,
 0.5977595626486528,0.6884651016684596,0.5147709588067737,0.6939505891750611,
 0.15687180471160056,0.6712685427661635,0.23854830102007796,
 0.18400150335398224,0.16127480707521402,0.9432238766206971,
 0.9771214358788878,0.26109422779473646,0.9715802778034685,
 0.14036958905935393,0.9842947323601231,0.19380994961132902,
 0.7833757653803404,0.031748634880594206,0.06228164937799974,
 0.4351522259443322,0.9888831341188828,0.7266760447075966,0.6947345670221475,
 0.7663140842707875,0.8883171412482914,0.20177425453958697,0.5187933612250542,
 0.1740311087802564,0.5247037696666274,0.8778540820674394,0.5601736508133445,
 0.6011812046390086,0.8543122891646031,0.023258626307618102,
 0.13314383018341835,0.5986139304654586,0.30258072448015194,
 0.7669424419533122,0.9452222241059839,0.05020273194291447,
 0.09725614187071385,0.15962096541323023,0.8822690268247182,
 0.44076722347560393,0.4560096275259522,0.8440559464593037,0.9448634324123815,
 0.03987445317691529,0.6744962248991238,0.24144489809156677,
 0.4129533624987347,0.11641367140652381,0.10638655685546472,
 0.7285404638981842,0.7294812875511987,0.7081726923578232,0.8422952996182146,
 0.2005386631352647,0.29845131814426096,0.5895980070103409,0.7498407899874827,
 0.46984841724827975,0.8034799372300098,0.5454609800322278,0.9509222543083073,
 0.9061773104447357,0.060795848783194884,0.7837303430895152,
 0.7142720171772576,0.20913200097579898,0.3401918951500471,
 0.46330910057234587,0.21521245443448955,0.01777650628137073,
 0.641378594636051,0.5282697141476578,0.6520193685133261,0.37498626932561807,
 0.9321722074659483,0.0985583493156923,0.1771106503147244,0.4855324606892497,
 0.18764391977036876,0.5897837091711042,0.1889517236886573,0.9672498400223337,
 0.6083684035047021,0.2121871901982978,0.07283451588431644,0.474586728668394,
 0.46258229494057024,0.6732472365583191,0.9769034143842139,
 0.37832024526783525,0.6440699899543253,0.9952320967948299,0.2699002148568539,
 0.7953977895291912,0.36767936484511965,0.33724837309421984,0.31895802421997,
 0.04676834806621755,0.4359019481818378,0.02370700199179121,
 0.2692665800167784,0.851208278221363,0.5790317564657442,0.7061615550327281,
 0.16424246237839413,0.6058318228575024,0.1388495862836342,0.6868196671276028,
 0.3073769782389928,0.7478434987516639,0.08104667607845795,0.8732335521206772,
 0.8560856073497981,0.04320572178437576,0.2877436079856428,
 0.38313496918618906,0.48331313327076975,0.13191265705826272,
 0.4082326753755382,0.34757509514762785,0.025893148362452933,
 0.9985693352814398,0.8362146686075493,0.08108620596723559,0.588513541077328,
 0.26582624433569013,0.5372747169949832,0.3453677466140441,
 0.10989232381967662,0.06950654813636503,0.7305087732332302,
 0.9271340606043644,0.32270528795433207,0.4426704365636416,0.4013177825190338,
 0.6136570647024174,0.05079036121495406,0.9053375840198756,
 0.03932632318098017,0.5888954977245916,0.1765441652959966,0.9737500309794307,
 0.2002526889403613,0.3480255732695756,0.921881108932773,0.592483443075861,
 0.6953964706743119,0.03515502377057522,0.6859752116601738,
 0.22371958777656387,0.11064659489944717,0.7317621313172374,
 0.28491989066315737,0.238858513875717,0.8999832641774278,0.2695004524227044,
 0.03952843500194414,0.22999351875064988,0.1662667484567274,
 0.44912201226916026,0.31133378557898084,0.1400083139875391,
 0.24801150339523992,0.2875954174797285,0.42129850952831105,
 0.07547817806897117,0.40519027990759693,0.4055182012036267,
 0.4498709411831282,0.21942038306893064,0.11987393729022822,
 0.18624497881478663,3.363690345020709e-4,0.09850924449784437,
 0.5597937736510303,0.34660662297039324,0.6625755789408198,0.7497116636719042,
 0.03332225087919494,0.13821580813929446,0.017908390029554133,
 0.17379713790401197,0.2265483593920402,0.19332407239707694,
 0.4178264970233344,0.21599641746020326,0.06294233845000352,
 0.8368324232194992,0.11248090653730058,0.5359442897766866,0.8720686200179627,
 0.27090693600738425,0.5442166363807145,0.1365759491760008,0.2322933349068057,
 0.7714597500800571,0.7078139689681654,0.5449051957682701,0.12801958961814097,
 0.33714223502356633,0.13429386764749596,0.9343119751508308,
 0.2183971502078761,0.5493424017995332,0.4176285540412057,0.1135815794446986,
 0.3565596612328611,0.8304615684634625,0.5603005965768846,0.3792620858096861,
 0.68549153028157,0.35358064362604213,0.668597948080417,0.75906941727481,
 0.07344593300107216,0.8642790494338382,0.5283230533712515,
 0.16700356068206368,0.386120179869592,0.7367213598629889,0.8804182931496853,
 0.35024318994802417,0.023491569054752248,0.6551975381060479,
 0.1795765039970859,0.10183260996357124,0.7842254612724848,0.9327907377468676,
 0.6267080280112788,0.47637374908301344,0.8621000569229045,0.5295387531300181,
 0.5280248080265942,0.4987093211778105,0.8154437268432883,0.7650902232796696,
 0.5932144499846501,0.5422039375656535,0.17461413658232672,0.9996515669682176,
 0.9295605088978887,0.8237114147715239,0.5881103278831348,0.06811258929284192,
 0.9508334060437845,0.3772881942632179,0.4803496519632515,0.07589095714074501,
 0.027498999882335395,0.9146015477951536,0.23111375917281773,
 0.3341305950906909,0.7647221310433352,0.4803460240254649,0.9811940453373608,
 0.30849794006612563,0.840716990896833,0.2989251373188917,0.5898222150162591,
 0.06552120298151132,0.4412655368460122,0.8734249934547806,0.719112444580475,
 0.45444124478833137,0.4672866302304246,0.28153335242250266,
 0.6918601058685717,0.8192482119382335,0.5917589268402883,0.2957039166390153,
 0.7044220382330316,0.5995435776953006,0.540475680821835,0.7194298457164449,
 0.8573111485716491,0.24179772100739605,0.11025159747426083,
 0.18699775791034468,0.6121126694600555,0.5562729486234192,
 0.18988213239683127,0.3121449500891528,0.9228751601115671,0.5291307983186857,
 0.13638643024244956,0.6502516917085494,0.12805649333286373,
 0.5106463395894218,0.7579227327521647,0.8355186769643168,0.32648215160050975,
 0.7226601622357549,0.8807301075302971,0.322839627381586,0.4816134103069232,
 0.251048413787731,0.05066302471849349,0.47467744728057326,0.738972661285814,
 0.49499226679531816,0.7103356739132878,0.4422303878148348,0.5621640461368356,
 0.9567902761131378,0.1520865617878444,0.23000033564800748,
 0.052788211513876915,0.909363603986957,0.37996001201286167,
 0.3215152890447104,0.6203250027792764,0.8336848068913636,0.6892232220876817,
 0.2520414287026057,0.9296587654828657,0.5957397286595588,0.7559315259864858,
 0.26918674455860003,0.29561296595601616,0.7001637255107669,
 0.6560052348099172,0.07342871917956317,0.217490701930652,0.4765435243005136,
 0.789955441723065,0.42417788194329464,0.2671572155880989,0.7259974937239909,
 0.7979106985016331,0.08532716530288453,0.632262651246359,0.4471194997043666,
 0.436043909158327,0.07459285468188015,0.48984764682828197,0.6251905846574572,
 0.23416563192552797,0.9927021913107805,0.8442128310264836,0.534316867605721,
 0.21690240364614166,0.8017499682152534,0.5546670999455976,0.6462027463353979,
 0.5713144189531895,0.7279448643762927,0.2588954521418374,0.43442843202014914,
 0.4503805994142447,0.20983034740001927,0.12225798264407184,
 0.4361944147728183,0.7964154903129251,0.7065878170071563,0.1401578204263766,
 0.7302116791317202,0.5904872318177062,0.49975149213511005,
 0.17984141345977556,0.9925720430650822,0.8115229861653697,0.857861630263745,
 0.4816097472196993,0.8691091760721008,0.2934627408226933,0.5049996073408607,
 0.8967249122554513,0.9068677465575994,0.1798567730057692,0.9576836188262045,
 0.874374775066191,0.2577735668216251,0.5596971895107256,0.02989818662812016,
 0.8925950336096516,0.47129645373153317,0.13914093214639844,
 0.9391889995736195,0.5942045968462435,0.7567678909639012,0.37429845603104617,
 0.9302387917357207,0.7668326188045411,0.054738659082023466,
 0.22519034191002651,0.8750072818620503,0.39974591628648426,
 0.14854408852609846,0.022149572306029208,0.35066539680607545,
 0.997069060548321,0.33568797850492826,0.5220397240510036,0.03126166531850194,
 0.2902952922555908,0.2979691279045329,0.46106817572154657,0.8345669541137982,
 0.41644656458849016,0.2385830308986755,0.7080611174839017,0.9783405433313976,
 0.9477582467450772,0.7846317512381571,0.9696003886994516,0.4991414667374904,
 0.12866424333368764,0.7880975308181499,0.03269363218935284,
 0.6010850528720637,0.9021361322188454,0.6102308179786377,0.7342281411862175,
 0.8167923401288321,0.7612209272208202,0.42236450553560523,0.445098569568781,
 0.5240452036713799,0.663684101427356,0.39149855926574584,0.4345791194398331,
 0.8716321543454748,0.24468421877574198,0.3146757276414335,
 0.37264964074264084,0.6464924940486317,0.7757536838255609,0.774109226079479,
 0.0797488860382356,0.8529413192394096,0.7065972325154423,0.9922070138078376,
 0.26079041397349845,0.3947091930646608,0.1760542968637846,0.5201932604490579,
 0.34018186059242095,0.47891413999207677,0.8154312067421512,0.814649684958341,
 0.8333452852975136,0.42525583511211174,0.45873377837986207,
 0.8105273515454491,0.6088867169413091,0.38960730105760955,
 0.14386572476671056,0.16005805178056576,0.7212060007112071,
 0.13554126137024347,0.49774738936934004,0.14398413602606475,
 0.3615942652708761,0.955789111073035,0.09306866163613114,0.8976505595289084,
 0.23884884462696876,0.38413294320365166,0.8751745426016149,
 0.6287692050639571,0.7022138256609509,0.18391429673236925,0.3882293016208568,
 0.015882644466834628,0.19201764564595436,0.4792469192961929,
 0.9696737516425269,0.24453131255639637,0.4556710034061784,0.2119165942538087,
 0.23403595122014087,0.2620722249486025,0.894725619377266,0.764025663315229,
 0.40919304615977126,0.8924842531720977,0.640878459789393,0.3311688926032669,
 0.3818260553628774,0.009343264711780153,0.2144676843382981,
 0.7978649273470216,0.5935758558539597,0.19825446318663376,
 0.04961613299951839,0.9890184798290882,0.8747670518878823,0.313673223525023,
 0.1381095617448589,0.49647840364802487,0.465550382927297,0.03195795804225887,
 0.763690914145442,0.3272731437581944,0.17745741743300003,0.08936786337143132,
 0.19618900576550424,0.7995844002354465,0.9726029536566195,
 0.10192868349698458,0.9288595222807311,0.03693753342182016,
 0.7708533276865882,0.031687370147428506,0.1616361417939055,
 0.38591641452867387,0.4034678061186143,0.3878935694931077,
 0.04229198058443695,0.7323641085984067,0.38181286965032224,
 0.033054820439340826,0.49535388203205966,0.36409058396055416,
 0.5179972518408054,0.5530458024932113,0.7447483230742915,0.2861278990932842,
 0.2115567493959173,0.3182089122895979,0.605439225137956,0.7938645440930405,
 0.4625208399586096,0.6311850409056989,0.48311027516982763,0.2473044903557935,
 0.47286371991087417,0.9254211889504512,0.8141928513946526,0.8918422568488381,
 0.39689809283637334,0.4170007669994774,0.07026629010652463,
 0.45647715706933434,0.08493950538676287,0.5455997587327543,0.524470091445098,
 0.10146386581982036,0.949148900994723,0.4315282597565063,0.416665055321524,
 0.19893852984129645,0.8093638311314315,0.6664597423489339,
 0.17942848596313243,0.3494745561277145,0.2648394608854251,
 0.22630504984390976,0.4331118317235918,0.21933005767277647,
 0.48632924159462654,0.5668807937271131,0.3320042844994314,
 0.12008033360753023,0.5744231022906555,0.9787947745829679,
 0.37476230383134357,0.135012847213428,0.3616434937295313,0.5120814580759271,
 0.05208361316060084,0.8144008299021125,0.5312646543791474,0.864734682643228,
 0.44674990112087165,0.30718196302662437,0.9391177571078493,
 0.9376417863057445,0.5039249063929612,0.5626380895427916,0.8953055350790196,
 0.6141060092161478,0.012678452551224129,0.28363886656231563,
 0.6382351384422235,0.41117843386681674,0.12741007180909036,
 0.6239510480509527,0.8547715717551847,0.022395764127949525,
 0.36198401079789977,0.21102974807186092,0.5811789267596117,
 0.38948704539965984,0.08321700319360581,0.18456262309431082,
 0.05969335411108423,0.8751602593272425,0.2487081674900169,0.7705549876781426,
 0.8354641412226741,0.3477265358110977,0.6546317767198022,0.23025817458684072,
 0.5247762819313153,0.22968736039216675,0.4768932092088113,
 0.17378949119994802,0.11582333864134509,0.3905197844350151,0.598498593495526,
 0.8119527822736998,0.8188653045014521,0.3471228269889598,0.10483779523055703,
 0.7677741865199459,0.20396268172159548,0.6620687209946192,0.9832470019569842,
 0.9153629506639989,0.740008739117257,0.7778094254249217,0.17609638247931014,
 0.6619863410668293,0.641989778254966,0.2779506602476609,0.10800343582376304,
 0.9943994921749875,0.8051054210685592,0.210137880757283,0.7310198511704266,
 0.05658634609659563,0.09144796933066623,0.023114878116545556,
 0.4232173434424291,0.4104054803383458,0.6362658818309191,0.29946342418841376,
 0.04525250399835612,0.4863304718891276,0.5988181192758775,0.7423398294961636,
 0.7017353769746354,0.9325246617963027,0.21329315949464211,0.8983232383698457,
 0.5263087718310482,0.30662835217694395,0.9234789184613945,
 0.0011751172080883832,0.6546924966900356,0.12072551370546647,
 0.6424901748375033,0.024225109727288974,0.9521572771773845,
 0.03133895922096175,0.21135613242547302,0.7890344998060179,
 0.6250689327979878,0.2702871474539926,0.38647387460982063,0.7437059747420425,
 0.1797958543204755,0.8270381528639992,0.34259567836339333,0.8619221803628846,
 0.13212391298384152,0.566806946073869,0.4781865678719994,0.322347324864246,
 0.23352423046315873,0.7904023315400619,0.8269366939206214,
 0.39397888617556753,0.08357544133286932,0.9370429133975346,
 0.8877642152267043,0.42014348608309604,0.13118916421762117,
 0.7631844844710072,0.7385725612928155,0.31579811414821024,0.9894289329631175,
 0.8787158586049705,0.8169180099975972,0.5581738286623058,0.8811859703075352,
 0.466459566444825,0.17620547348246518,0.6668197956261006,0.6155031304413683,
 0.7183483858972631,0.9793568296482961,0.12195613579419806,
 0.28452219697087777,0.1457771845485616,0.5922321243154878,0.3664725845835246,
 0.018937078174677047,0.5413637567817584,0.8186227544115625,
 0.6321377815079875,0.8711792170233148,0.5731519806708664,0.15446381199855586,
 0.05193057857506622,0.8473542599198574,0.6849137417982263,0.9194687221461492,
 0.6587906966955239,0.3081846936688848,0.16624166408617747,
 0.08586358824509688,0.7413506798228437,0.5528999143399586,0.9710090614387704,
 0.18200240587097138,0.927666223851115,0.3306586707201422,0.8679198749399957,
 0.09255410969511169,0.922226452289817,0.13177282457878325,0.2566816911800378,
 0.2601768797332752,0.7340903379240497,0.19722775518177593,0.3555538864430866,
 0.8269735839264379,0.9395906503627729,0.843795853416628,0.5976771570098702,
 0.7760095652116429,0.41431413582000665,0.6444810914210106,
 0.08611456313878196,0.17834849968865019,0.5794115562890313,
 0.7069269129070457,0.7210096282167012,0.5917564821193205,0.28701446661551777,
 0.7138804529240605,0.8962269600551196,0.292909474889275,0.45837763282755906,
 0.8508003257621675,0.06481388152877776,0.9379115994593945,0.7265425553080763,
 0.7504551456856638,0.5841311816461879,0.06465908704802603,
 0.08756057971145981,0.6673307236592984,0.9369606705502133,0.3796670362808564,
 0.6670854011479249,0.8165734415199082,0.006207103007983106,
 0.26018685946896114,0.2603137530331303,0.22612056264446712,
 0.12607024887273965,0.29979147286409735,0.283831014971097,
 0.034222095539280906,0.9532029748695543,0.10497191941919015,
 0.488564270826504,0.1076327884350371,0.1508791031553094,0.021121605103064378,
 0.3741494016543626,0.6214842922468871,0.8598781033206016,0.7806880645853681,
 0.6756731505728092,0.21241416473910502,0.2120571757612768,0.9617456973267476,
 0.7658358889465785,0.2409445791410093,0.7909377633747225,0.28108988806524304,
 0.5102024587807295,0.8702731474771199,0.013417199978858263,
 0.20308796553895336,0.7736478170515678,0.9751910819572128,0.6173176291898401,
 0.553461372521175,0.8606238454995967,0.10054192907233817,0.5716182730198802,
 0.4923522605974676,0.3854214161428896,0.9699719091708987,0.12715727242672048,
 0.8460679774455379,0.1961313819489493,0.9456314886574794,0.16990304095188424,
 0.8796919376692129,0.7215839652744166,0.8362715051448227,0.4658212864777076,
 0.25344381112366043,0.6694588361756033,0.6273703340685669,
 0.21161523106791102,0.9382217071781027,0.3663298010072883,0.5115959437036949,
 0.15169420526887234,0.15163662848877701,0.1918653092849517,
 0.21728368406525744,0.5802886303205749,0.375820526223009,0.6137082158233296,
 0.1381639790699597,0.7210691949051641,0.9016076992450831,0.5155459829059423,
 0.8534521678310488,0.3913960983543766,0.7922876359038242,0.282588852789984,
 0.6341012360459167,0.615450494157403,0.6492555356823143,0.8557815491083085,
 0.020138405669176795,0.38876510735130176,0.15439343220811697,
 0.8342336843427028,0.6262704066464877,0.6551096793851796,0.5575412778996578,
 0.6359098493636518,0.43276798928080806,0.6242899025079058,0.6110879128727478,
 0.1382553017531104,0.9703298033744348,0.06255592878599547,0.9255268482947182,
 0.2598759600977716,0.84365647761509,0.057547150165876304,0.47522116625077204,
 0.1886615826096023,0.29622770486627414,0.6784534191740136,0.8595686643632604,
 0.2927703345503022,0.5558311415392698,0.9861925282227492,0.41257369129476185,
 0.5377814980053955,0.36048823725153856,0.7422028173964857,0.5954607505687893,
 0.9120488811535382,0.8075607076302065,0.420179065240989,0.4369048784688059,
 0.781114636005441,0.20199848207182824,0.8848650181141958,0.8615319813845771,
 0.9824945776755639,0.8731262683918805,0.8920360634494556,0.9874818155562927,
 0.6171467350910887,0.6693431498960296,0.23463403333370614,0.783451778570321,
 0.10995612403361976,0.9395822848296878,0.42600487105333595,
 0.010142448992479847,0.759283968014588,0.6794973096983301,0.9150067452916083,
 0.5101781975125712,0.6446899233009896,0.7492178624244267,0.7842454529478351,
 0.6286856336353556,0.41353675948103574,0.11396019862789608,
 0.8521696357540652,0.4740325960619739,0.7147365695017943,0.43332314662343974,
 0.4166399099669593,0.9028095830558924,0.3034471578917116,0.7796411826894708,
 0.3480184739602602,0.7770510045382084,0.2612876128598426,0.11500931085090849,
 0.20552259248180038,0.22676012540321366,0.7864027029873172,
 0.9015993616012301,0.3769963341518743,0.8046403091506673,0.037395568732842,
 0.1641258586840234,0.3413845345411455,0.5955198355863607,0.4278302327441389,
 0.16844290130570028,0.04157218330271917,0.24702760698191617,
 0.00980888220893572,0.6882130719743461,0.46296213126676966,
 0.008575485345960865,0.7907964223640758,0.81362889075351,0.10955636580595085,
 0.9012683973318345,0.27896796334825535,0.03281260637329264,0.641760799327545,
 0.15657771746394533,0.02198673984589894,0.7877177935210227,
 0.8502408935746771,0.8247293011399932,0.904549117794169,0.9018011496688482,
 0.7990315063640909,0.5292038546798143,0.77086914356191,0.3957578853826784,
 0.3623696748267201,0.08575432123480564,0.7048562912876487,
 0.0017235496019833985,0.7721160898341406,0.6570652776664127,
 0.8474309899383636,0.31341477979746324,0.9321266284979497,0.8964667965796194,
 0.19637166395096506,0.8327355038353431,0.9249038980701187,0.9377635163762863,
 0.7240640181536766,0.4201873061444269,0.23373147437549202,
 0.35154729441451615,0.8512426207292451,0.8143627649497188,0.316203467410372,
 0.7277791965805459,0.6877660220044242,0.8927126836754364,0.6296745501154049,
 0.11905277907045586,0.443856706491927,0.7883226876080032,
 0.015229060155417029,0.32880902694249514,0.7086022995989463,
 0.27390480996850786,0.4980725484045908,0.5490562097404584,0.62971109541363,
 0.4513098318202531,0.3248074437374824,0.051275476309850965,
 0.5248143493686497,0.5405373282509232,0.5466846466986423,0.12410035593314606,
 0.42022495899702283,0.18924543318948017,0.06522938904163544,
 0.7700663924722428,0.023198381194428297,0.5029014619010622,
 0.0623323582824864,0.11784204676638621,0.4085629396341256,
 0.23992422017238202,0.8718384988238009,0.09391013020830219,0.795064443287904,
 0.2724257891319083,0.3759804065129012,0.042144995086025205,
 0.42569152351236295,0.33646796508704113,0.24809609288615053,
 0.49180084607203467,0.17027174212549667,0.07003572212258202,
 0.5472483057795742,0.25278038538526426,0.7497007784270269,0.6181846065583524,
 0.16587940278760016,0.9681108318520296,0.7725511572125434,
 0.42568493809025154,0.1854373160102718,0.7360751953769716,0.8116110217890183,
 0.028593127014762798,0.5077828792497686,0.015831202151595503,
 0.38566939415610824,0.028770949919161626,0.7605755904044175,
 0.5183341805200166,0.8601240197722303,0.8762474246399964,0.24038553908753635,
 0.6884914519646361,0.2423609813977149,0.11685234606393158,0.8312639150801182,
 0.6844209164284909,0.8788071031973097,0.711983731666361,0.2663737972755402,
 0.02847565509038741,0.32901686840679234,0.19510339365076834,
 0.9376778378626547,0.9299797156938537,0.29997168976279853,
 0.48796446883121425,0.2960941620036114,0.6193231263705209,
 0.46912459586252775,0.9253982045116951,0.020331664353447798,
 0.4455338216516541,0.6231051420995045,0.3456923788219993,0.6092422684786046,
 0.4051926668757382,0.5254225839915657,0.8143091712035323,0.9392304169435595,
 0.05325618084374184,0.18344736735630374,0.1762913001815054,
 0.09071632859962775,0.53334854101098,0.3549822310655698,0.025370493477119616,
 0.11590907962508268,0.9162277024094532,0.35651457045402457,
 0.6126465987460479,0.39579610519496433,0.7903936116484437,0.6580376854325144,
 0.17872032242382285,0.805959254077264,0.8385640621785635,0.14533591797816436,
 0.043151990319274525,0.43687278353942915,0.7921436121115529,
 0.6658083204493617,0.895738413223979,0.16421518500618748,0.8709461250900129,
 0.9968345550059271,0.2772458676564349,0.9760161749747616,0.8364197667901746,
 0.5809817669096891,0.006347416397578742,0.5366968007451659,
 0.6463264374349279,0.20424058257653643,0.8221236437660261,
 0.20687173599243747,0.21239948415705734,0.8992405099302014,
 0.029121058142530187,0.4920182316385723,0.2633089978554062,
 0.44730046100679144,0.5272251470732697,0.14737416301888118,
 0.13452735381630287,0.5218626122362238,0.5033236388499559,0.9644933586358785,
 0.8411856254046626,0.36014920692565444,0.1295784221507431,0.5342011034889255,
 0.7135207206044419,0.29880864005749275,0.9282248409768113,
 0.34202210709048075,0.45025424095445543,0.12624116963978027,
 0.46638492473128246,0.4091367843385928,0.8867307229829049,
 0.18087937046986435,0.12430318204330992,0.03615440843686235,
 0.09640574777775868,0.3011568060474862,0.3148795161578183,
 0.20693078007921972,0.6618642530523734,0.9422082135506491,
 0.04844586605517953,0.42064816454505793,0.8887206311961664,
 0.3763433761150301,0.17400264560457712,0.7928173260704189,
 0.22481627101187274,0.9592214062188367,0.26726524824737197,
 0.9634241892760597,0.3777910503245139,0.9340770222518483,0.8984265237353402,
 0.9003280158083635,0.5022401565334484,0.5931664548676021,0.08322841085772081,
 0.9664313467797104,0.3896194254268296,0.37337988241954434,0.9825345820851807,
 0.7021238311112128,0.47326123668740216,0.49328487811844346,0.923300867150558,
 0.35031510796171006,0.2553275547896928,0.20156431268929376,
 0.06173069406502363,0.33139439046244923,0.6501908224591262,
 0.8801568257859607,0.2858504378064587,0.6288306003691666,0.808930075877644,
 0.29516631186581344,0.20429146688745214,0.4318549933811918,
 0.17633713958132868,0.49112258657137553,0.057795749463205315,
 0.8451651968144296,0.08238618003315024,0.370250529274596,0.400705945435885,
 0.2575188992148916,0.8081415372828757,0.20811134202758286,0.5604501112031712,
 0.5013241734974747,0.709061966152234,0.45622481866261744,0.15646245526056757,
 0.6055913367709213,0.8309658723832307,0.7472898402356125,
 0.029804845674298863,0.7514565482751657,0.20535550566245975,
 0.4734119633689696,0.15558330839652212,0.8913327440407915,0.4624227649979187,
 0.8298631345965563,0.7221288221686135,0.17295378276543638,0.5669884576222247,
 0.055908461286023226,0.6001927078840461,0.21773726219535428,
 0.5871183621943407,0.7719186954758918,0.2574955449690133,0.1122650255556944,
 0.11528796924807849,0.1019462047579458,0.937468724570229,0.06494554659777296,
 0.7242252256464382,0.9686298722909712,0.3713859329659609,0.5661074783237086,
 0.9236893248272136,0.048863405909598034,0.9095147081695004,
 0.8097472079858624,0.08071883070907726,0.6128372847228252,0.7233942285013779,
 0.39991830306804965,0.5707518062532075,0.3357670463333814,0.9034505336196895,
 0.24719376522106318,0.6834347289349383,0.2914239920697428,0.575492910789395,
 0.652424253141446,0.6598758345396457,0.1815797243166608,0.8588096737614495,
 0.17967579427225466,0.08828983826996584,0.11367925869165318,
 0.027133137804487117,0.3988168464618781,0.9674357422075488,
 0.7949708589870997,0.2364419123516217,0.034419439781733674,0.894320270589629,
 0.17280773958224938,0.1540869188807904,0.6469834740721049,0.3146153332778422,
 0.12296147294243331,0.5762201502074685,0.40615855592314987,0.325545261328378,
 0.7968877070805662,0.9007035878348599,0.9186406209463625,0.8274961307943481,
 0.4582041084984231,0.1623735468188544,0.1245077687399554,0.81589707326771,
 0.1546702226855481,0.14018127348767595,0.4520538492684181,
 0.26065560578671987,0.5277520490309109,0.966870631695979,0.615293329270504,
 0.6153963263063754,0.820510276717372,0.08895715960936046,0.2929010456241097,
 0.6977277811883289,0.20662631641821316,0.4049800912768333,0.330761594446806,
 0.3949154499792825,0.058824906578347935,0.9502321156393211,
 0.11486779038226191,0.02376922125880654,0.9163609274352595,
 0.24877287316409846,0.713159184762825,0.576608224983475,0.5666224328369023,
 0.10446782505400976,0.7962421963134898,0.7722728062484373,0.8849461373185445,
 0.09365998880026549,0.9159691840234826,0.31227775800000046,
 0.040535833371766805,0.02364613676331695,0.5271174790976267,
 0.4998054838675361,0.18936262614194688,0.5464243043566931,0.6465499684620882,
 0.7221982213783559,0.92768219002984,0.7842326810162319,0.8026370020226422,
 0.605792022528034,0.45304818725251705,0.011830225260142324,
 0.0029766694376117675,0.24886887282794012,0.6706604961541275,
 0.9509817373309899,0.4162442034847248,0.6503123193424353,0.6134583251600986,
 0.9800791193243426,0.7145679797324035,0.23868662768854199,0.6747443700430984,
 0.9458398259842394,0.7268111413499854,0.10333583664439638,0.9293517690269961,
 0.7205517603455525,0.9153976582302903,0.9170374633019414,0.967636607408073,
 0.7599096740222936,0.0035836192804592493,0.13412692116667702,
 0.8418388368371008,0.05653124472347104,0.9612320163593301,
 0.055862944199407405,0.057507746905695156,0.48002111667400604,
 0.7063774974364051,0.12454387874649153,0.769124466570983,0.845749568647186,
 0.13867789643824624,0.14008596683170582,0.6613746132617238,
 0.9988674596491992,0.5088237906133264,0.36231716345415654,0.7253403612806157,
 0.41373164268269214,0.4800697902885863,0.8534341652604942,
 0.02524338389295422,0.4383743579788041,0.36553610455896224,
 0.6557481565877027,0.5715335006239841,0.5712731576701506,0.5631859696493081,
 0.7925388941156495,0.6970317607754161,0.4482582557105429,0.22970340141765755,
 0.6016779364832594,0.3562469863007607,0.9260819917760421,0.38461516872378376,
 0.212594267786978,0.8612409254249614,0.347007231971299,0.16885159819150675,
 0.2939715649041241,0.4164903120699326,0.9274035848151415,0.34073757990990805,
 0.04359051589113472,0.06822072599087037,0.8681189375833788,
 0.9958496440851218,0.2274481865474537,0.3992615219442349,0.2270804377337714,
 0.8856771694127964,0.8920949393092739,0.4944627868223852,0.8020315237599543,
 0.9040122082853351,0.4233198922991358,0.7106632697052441,0.32382909521425995,
 0.980560994915071,0.2944951490599538,0.15030131058502505,0.9417277657766551,
 0.004659632173430794,0.8985709529544235,0.2309963429314632,
 0.2979737654694644,0.7786231045193529,0.4325442686934514,0.8979791783261044,
 0.31568545633188305,0.9043120509067872,0.3016784745065645,
 0.38019924288923845,0.6454802289970565,0.943019053190043,0.27079714293548673,
 0.15329247173486404,0.7399436542607037,0.46455831882136356,
 0.32160305054461324,0.18279879628342044,0.6929890734249275,
 0.8813793397375724,0.4438491835830841,0.06023660404121989,
 0.38496189161901384,0.48926391649468015,0.5025724963912646,
 0.32652053407670867,0.4075328689491229,0.3212474015329959,
 0.48509295408960296,0.7320095851320269,0.9121650578202325,
 0.08610373854473785,0.042660136111905556,0.2587710965939096,
 0.5481411917726847,0.5339140011081748,0.1207934721680809,0.4362316466190259,
 0.7632969139399866,0.791128613804492,0.13035244257248157,0.6300844433843579,
 0.49004641639494184,0.3025891535288814,0.5309189399454644,
 0.34140193723867296,0.9719913482521468,0.6654396757278199,0.9912971095543051,
 0.2884306868337023,0.21359752012395206,0.36233014676802444,
 0.9463609852449701,0.9998080941833369,0.9231162978034922,0.8688067924129058,
 0.8730154662401236,0.4537520530539455,0.06117951303097202,0.1296556391852044,
 0.8937065670760469,0.14767404489308977,0.6365410735111209,
 0.10119716145095337,0.5967733100803526,0.3568016430218236,0.5337572921378733,
 0.8729208683335513,0.9727487434743562,0.4876752593310859,0.9173840950992453,
 0.6337002469512124,0.40595143908884046,0.39938981255688577,
 0.02926781803005918,0.8012934110773166,0.13495098521807825,
 0.17735732110520153,0.23870000988139195,0.0923700392703124,
 0.5976482655489448,0.4465031338757046,0.05851175863593017,0.4825234012642421,
 0.9874849708961935,0.8212257307723096,0.9814152415531865,0.13021962522556751,
 0.36472544674525076,0.4927671921361364,0.1316244813777281,0.5227085982647489,
 0.11630174026878903,0.1469991026355827,0.09580630225488296,
 0.3431372161128112,0.09748460694554795,0.9960949845236459,
 0.16409575960901945,0.1472075584954654,0.04894983849397039,
 0.22111413647270506,0.7832561004393616,0.6418191835283752,0.6939311127550322,
 0.17345007781507504,0.3504983577516362,0.5798704516429163,0.6963417342943475,
 0.40597558741683415,0.5396424991392814,0.7773663882499169,0.6904725700637773,
 0.8291737968283536,0.3449861297973953,0.7595816166898681,0.6462051147754199,
 0.6834900258478622,0.26507454727651636,0.6871096935267011,0.96145129134803,
 0.6689760667058672,0.7261938438704922,0.15907130920013302,
 0.38710776178986117,0.2453740539128373,0.6118529390728417,0.3114896970704214,
 0.17064056216298584,0.507077819078924,0.5166603047754706,0.48416503014354106,
 0.8899799010873947,0.12111404499330525,0.23293240914374824,
 0.5693892924635935,0.8613082452927157,0.8814451036557777,0.5270690595012159,
 0.5058561689885152,0.9959350617504528,0.7242846108624788,0.3627762445329059,
 0.2575408920512592,0.9482022482159319,0.6269860220694518,0.15105202350869185,
 0.15442290235316092,0.9750225167058274,0.322248749539522,0.6443924236728802,
 0.9828003110389132,0.6184883587330935,0.24794353177111783,
 0.27077107506069975,0.09574174250667311,0.39960135894867743,
 0.848267656098616,0.9114814585850127,0.5658980000453292,0.9061362578245926,
 0.6030259029216305,0.9456848917714378,0.2616887241433118,0.26135509532223233,
 0.16711994296776167,0.3013194174727698,0.5688256244706134,0.9052021354251559,
 0.9215797505240056,0.951552892991542,0.902050827299678,0.6701713918441234,
 0.7012799027848571,0.5037052158940454,0.900920375266206,0.3494108970858797,
 0.3624013895212146,0.35903405248800924,0.10356855922560226,
 0.9356368282206358,0.1923283333220741,0.3863432390681467,0.7039256569709149,
 0.9945987740765777,0.722454672002027,0.5571523452146023,0.49486534851320174,
 0.6792901837607196,0.21180499375296447,0.6319154162915364,
 0.49190373205044824,0.04323216124076201,0.4267095796539394,
 0.44081957703339025,0.7279971400493231,0.8273426353918308,
 0.08103402300691742,0.9371813839508337,0.4292671433697257,0.1522963920839875,
 0.24302757666639552,0.20450956892191463,0.10593323043957081,
 0.5729920890857951,0.03921535212168903,0.5198455965251708,
 0.048349744416334994,0.7354827444123898,0.982161125458676,
 0.30423210597743944,0.8080547457430314,0.0939935316772188,0.3761937670334765,
 0.5661437366249357,0.07289204189214427,0.8783411909540695,
 0.19804525275776008,0.8268160130252766,0.2835041848100438,0.8227421148866048,
 0.20983235209204087,0.9855985873841965,0.6938564728726792,0.9808477161440183,
 0.7175961999504813,0.48179478876514237,0.013125824009340703,
 0.9932787450367166,0.965846100119667,0.13728028244872403,0.6952894846860848,
 0.6934020462758768,0.3118096342108885,0.09675018791819767,0.7158238864089822,
 0.2236208280049734,0.19307144049726643,0.5745351867370168,0.5520981289970945,
 0.8383877978821038,0.7360974494157108,0.7221113939183887,0.19139772706568026,
 0.354021414031367,0.38609156516989696,0.12275661335040478,
 0.13595575646658586,0.7753648730653822,0.40990540620584903,0.899243027120396,
 0.38057103853784,0.9357467009022663,0.6143489938227797,0.033271261055319745,
 0.3266626614595559,0.06248399162751739,0.006158173440786396,
 0.5034349900329855,0.48787403909444294,0.03180159401939531,
 0.11114535291234906,0.3536947344529988,0.9024664559457602,0.9111216645233144,
 0.3937765945276459,0.1928427153189609,0.9022500653566857,0.7774525111538282,
 0.4204535355766008,0.9004660694019604,0.6442663979624285,0.32550388997971047,
 0.6794164499120567,0.29109540459518257,0.574707814970434,0.6940227778888483,
 0.1539676399736063,0.8634452477764736,0.7619342071439711,0.6436921628583342,
 0.5587399986316116,0.22026129642633896,0.8688793961386821,
 0.30241369311481436,0.6106439578727483,0.4432478700925221,0.7796110125911021,
 0.2998743654077436,0.6464798563298053,0.10857168982332488,0.2079075902856724,
 0.22146431818071144,0.7104948547461635,0.16003990937882162,
 0.9824024816742079,0.6998636373284559,0.6292554455787186,0.41983312299084263,
 0.17380137401218831,0.9848925936901105,0.4256943360633394,0.4879076817564928,
 0.5230324205332434,0.04418446975040591,0.8483486921090937,
 0.37679416487602946,0.10714061761303895,0.5108904652537197,
 0.9795583861491878,0.6929358395686375,0.5657379259834622,0.44721593875299637,
 0.956200169395874,0.15155401552426917,0.4070025761056594,0.6980716301457677,
 0.8416730571704057,0.6265662064682856,0.25141859239099806,0.9900126149877466,
 0.7870415661768937,0.5239017679780302,0.5402940988099993,0.7400664591532481,
 0.23406183142263404,0.13990350454218936,0.543270630439645,
 0.08013263280233562,0.27952518679618543,0.29394762090427773,
 0.3402915180745313,0.44668119699035047,0.485703185519869,0.11545380831179763,
 0.7615362841862789,0.9768926184910629,0.7315095204231681,0.7517304747278506,
 0.4589685424233356,0.21010808915332824,0.20622939347077396,
 0.7202470917940414,0.6322308208559746,0.4459522419852907,0.45758300778465644,
 0.09779645886097643,0.76256734107886,0.8627137892483371,0.05794541178113488,
 0.4435271668583609,0.8673588452522591,0.10526088718494497,
 0.10626747056480135,0.21998311693530692,0.45164061039877756,
 0.5860654854095527,0.17303579342419928,0.9267349271468791,0.5486045349567614,
 0.6262253545265174,0.880520710482578,0.06560254883171812,0.21361454115732514,
 0.7830390200270837,0.21769671195789253,0.5551217937409126,0.5316592978534889,
 0.7894069612376216,0.8449475983020588,0.7190129302564627,0.63411646104817,
 0.35233162668733575,0.8133084929258185,0.7748273366409741,
 0.18577654033826085,0.6067854356042287,0.6065955524868848,
 0.18968252130675967,0.18752483655247487,0.6253065770085164,
 0.028966716558777472,0.27696524961155344,0.6469100463647838,
 0.177893517979812,0.6899960744017,0.7561850928655375,0.5837045038389349,
 0.942652909998301,0.8576640058439242,0.839616775169234,0.3150736229467781,
 0.46950396892709634,0.7002305609162554,0.436975013351256,0.5982122669339524,
 0.23823199850808452,0.49879851347886683,0.3441051870751539,0.596929729974276,
 0.37662837827753615,0.39514327935039273,0.39330317966322836,
 0.6408412382869012,0.3373844904033003,0.6140042566736997,0.8509002637708893,
 0.2760447131351611,0.4409172771944986,0.5871717056772685,0.1504610798218975,
 0.33902887149686356,0.44523825614364343,0.5490262296785694,
 0.9243703115903743,0.6459393837218542,0.6109054541666434,0.3653383443607152,
 0.8313948646726089,0.20594256161813018,0.8236704219374408,0.268727623331281,
 0.1131080837729651,0.4103456984903415,0.7044999711373388,0.3058449404266367,
 0.15862566987640225,0.3655134486871038,0.02323872191808607,
 0.5793359893082631,0.06301553750443145,0.9637223546444564,0.5123315904847746,
 0.46297953381619417,0.6453432367891493,0.7958441500061416,0.8758586281721071,
 0.045497899729388336,0.6061493513656878,0.6380718766232741,0.588909846251438,
 0.5525640066265354,0.43682000938335175,0.3819970683669348,0.7565262047677751,
 0.9823238998192142,0.07950775678016908,0.3414552248656455,
 0.03616166298004253,0.55464459862661,0.19785004617362112,0.4690081492642372,
 0.6351094360479724,0.22015877277542717,0.355921038867674,0.38010592144245126,
 0.8039643439312844,0.034664479664887504,0.9498883508967878,
 0.2673991112902757,0.3402878739187034,0.07679128832161852,0.6230391705321046,
 0.4885434704871984,0.5342287554526062,0.6518527129868553,0.7020423705742737,
 0.2720124989349425,0.7188247278158966,0.277266547104166,0.6530637533757584,
 0.6581503079488275,0.7066717636252431,0.5038139322515045,0.9102836608792716,
 0.2856346468100458,0.04924168075085311,0.040117191527837215,
 0.9430017460386733,0.5447118924782962,0.9265019839897365,0.24538655318437919,
 0.06145527203308987,0.10459865873743723,0.7570038991439687,
 0.43247085963328935,0.4158660493278522,0.22490991885009792,
 0.2930610625876664,0.3292090102949161,0.5174669917605381,0.8728553366645679,
 0.8934721747376587,0.5295558511174696,0.29132493309895824,0.3123707789878831,
 0.015958418428243082,0.549481781054559,0.048484566386572836,
 0.09750037226671204,0.3031387562393598,0.5658775939003355,
 0.014014321552248798,0.9912131870403994,0.2053860725618103,
 0.9570103164576622,0.811424150413362,0.5115017976500038,0.9860041567000413,
 0.7959483524922478,0.049490231799857565,0.29965157410905363,
 0.31484344284375987,0.08430348244288033,0.8535372319996524,
 0.9815432687785265,0.4568784935923017,0.9904905987612742,0.8230932972358376,
 0.28421129992238414,5.944800902446534e-4,0.5533023671440552,
 0.4296315587953229,0.9830133829889495,0.6305664184902913,0.7362992387322758,
 0.1830104139186569,0.24193632025165723,0.8991526719378629,0.6894165258379071,
 0.03790407733946011,0.8229035717105665,0.3165832246368484,
 0.10190593144854265,0.1858364000996352,0.5551843786410096,
 0.30697541632904723,0.2159308848465199,0.18757561702211023,
 0.8657047027018372,0.7047669704915352,0.9810496396562773,0.7660073091977938,
 0.7232303118476896,0.08146860319865834,0.8710717603341458,0.9895696204377005,
 0.09097193620609823,0.15945415048059042,0.0580954693197443,
 0.10541656138616762,0.5454003755242487,0.9355280431604333,0.4004825355786429,
 0.0087303630365414,0.2212991083989284,0.5571723812373851,0.6143797886182689,
 0.7971347326098509,0.9136024851038904,0.7231478435202325,0.15505105999232383,
 0.8404100126358576,0.6308687182863264,0.9759798390506654,0.2654950739468199,
 0.8092268615744791,0.3701721282021623,0.900174573858591,0.6975446737510955,
 0.5171424093992136,0.9290713301427481,0.5836761582653348,0.5316161202849673,
 0.24265914880077954,0.418536692878022,0.7107435343245442,0.16151941541473025,
 0.7574001131836077,0.1457418464803325,0.7299456174654855,0.22708095390526328,
 0.7701563686810237,0.7774715263952069,0.18115361804611363,
 0.04455118868867891,0.7312068589541671,0.4825572149331444,0.5840550649545783,
 0.14949599820627035,0.19218367882434273,0.7511124039676672,
 0.7435801414062675,0.10610031409497744,0.540977263934443,0.03961915684032047,
 0.44545273564663723,0.11874266021506508,0.9970331508435295,0.674325938024765,
 0.8939810188171443,0.6350333057105315,0.030466958021062918,
 0.32838680784029417,0.5648924999315414,0.8469611692345295,0.8685192829868055,
 0.2639167216416133,0.07832557922117267,0.3984256901030736,
 0.11960029961869179,0.7632828410494678,0.863373483610527,0.9887954898770922,
 0.1914791565933377,0.6532034297202954,0.5643825256438016,0.16215103125517172,
 0.716284801831669,0.7108703877954478,0.5544332381325884,0.322922138907356,
 0.1587541979167585,0.2692028021627285,0.1017337083062031,0.5962747023058154,
 0.9419034755652859,0.04410414844309374,0.2668291714835578,0.3852708442983377,
 0.19572213258264048,0.4834425433534101,0.4376108847082789,0.3473181013520522,
 0.31394806964693955,0.45384051940766745,0.20508665482694266,
 0.39876219144533886,0.03225137535089673,0.48872454034353763,
 0.5542084759920811,0.35140899678863813,0.7149720557499931,
 0.026750142854461423,0.383118233008802,0.7046782759585651,0.2944069233165475,
 0.19519764856165644,0.48390626878208765,0.8257596435314855,
 0.3250994027806189,0.2366294201920236,0.3584065455591935,0.07304197190664952,
 0.8295718187108351,0.28097834788295784,0.8333043950233439,
 0.12199967791020072,0.6181241068827846,0.005326257300505954,
 0.1335661341469363,0.6547139118967163,0.32144625140779903,
 0.47511322494748887,0.21611062726791852,0.8423586630830571,
 0.06155943178105672,0.7822087307489692,0.8262462068628056,
 0.039926409132192364,0.04082041656100477,0.1472902275433079,
 0.3110916027675893,0.6217111188312349,0.9544455662936591,0.17870947310349639,
 0.37226789892835166,0.5625229546619053,0.19273306226936204,
 0.7191914385995364,0.9752381371566542,0.7503763142363056,0.967588705060371,
 0.5800148977362075,0.08563229954548801,0.18523669764773487,
 0.6549555241512891,0.23191133493309612,0.5008139941358108,0.3866250532775841,
 0.7332512807846474,0.2777088869291665,0.8871807564900829,0.580856829342744,
 0.25429494409865205,0.5861136866133213,0.5817160642735668,0.810778301967662,
 0.884118377501055,0.3517770891937305,0.4275003527461517,0.12919149632340066,
 0.4476194457014159,0.5774024774570522,0.893617107812456,0.2776697921156126,
 0.0471882850046339,0.4490046140569307,0.058908981592856624,
 0.14386540945826576,0.8630092217262822,0.2780880626956366,0.4814169185365359,
 0.16740577007137658,0.6784447488815266,0.8477270466504023,0.7042840045479797,
 0.5466803441399453,0.6302251540315533,0.9029379985319634,0.5393244218372639,
 0.1204041160492777,0.5237948286631676,0.996335786556357,0.8501045309285001,
 0.1891959235877395,0.3679335854389527,0.666720406443315,0.9393820724766486,
 0.40295700328689854,0.15840442707724467,0.6913457423515585,
 0.022862284474304273,0.5481682000154058,0.002584958219933653,
 0.8970503220571636,0.2594313102423679,0.4519559662375502,0.6680456247875509,
 0.8726962890997172,0.5562609220413792,0.10350820544074879,
 0.17757236634036766,0.7352039180607289,0.9727673200838851,
 0.03991933859836072,0.9972469195751749,0.5069834882506821,0.9712859690130609,
 0.023747936934133795,0.3159890425977073,0.22170612037280168,
 0.504274414588213,0.6113305883271305,0.9160430457833629,0.14085163463568162,
 0.6641449780753593,0.8710205967415322,0.9363848092894292,0.6323595545440581,
 0.1738459977512533,0.12885695999631985,0.30196483359127635,
 0.30848288048116657,0.9278110334602964,0.9719780684284238,
 0.21889183795860845,0.8094878493337009,0.6963784496272659,0.4283382962927266,
 0.9587335639135314,0.6947600818923786,0.39513429963682467,
 0.24207398418234582,0.4315103294378341,0.7929951501682806,0.60414780831848,
 0.24879779261908586,0.9549507451334976,0.837340600003734,0.6755137216359384,
 0.11136826317056481,0.6546823104677958,0.6543390877125042,0.223154019927577,
 0.6883275943350469,0.7493008307740785,0.5567420210530192,0.8360610257240804,
 0.8077540039123464,0.03591971280371331,0.09330032707230529,
 0.6224690429053795,0.11639040586612659,0.6383285515729675,0.2496684339177535,
 0.39295665425871273,0.34585623851262426,0.6969646296247507,
 0.7563646104401441,0.41571412166211597,0.4819862512064854,0.5544585607179989,
 0.6522995251891686,0.05104040096082452,0.821602975774339,0.4367986817246413,
 0.2618418617404842,0.5071529890206453,0.1296437094419255,0.2649782210863608,
 0.5892702087914341,0.950434951202714,0.5559172170343303,0.6023737323555334,
 0.39838103842960637,0.53657371519666,0.8411204619740215,0.32620451548059304,
 0.6497631062831061,0.9545660599643417,0.6066856157305184,0.3690728576437673,
 0.026664500089225385,0.4234334250113059,0.6362198279831291,
 0.24124215337228616,0.07049460582965816,0.6920891611828666,
 0.4691565786956149,0.9776951812266499,0.19170582896716915,
 0.08076976934328128,0.3147380060996777,0.5479729708456811,0.6476424053940626,
 0.8616226796655366,0.771305444042371,0.16849128999507557,0.6759983213114831,
 0.1588390410925521,0.7837577553516507,0.5987666782119349,0.0504945515297871,
 0.7871266826923641,0.3175883877496962,0.5808960373953628,
 0.013089273799296253,0.5160922605904658,0.6726346626345461,
 0.3397541428174293,0.47738211293046606,0.3590589412060756,
 0.03330354342373143,0.12994144441446343,0.5881928455008237,
 0.7971828822121301,0.600722938602943,0.35726400723111607,0.2799978508669021,
 0.217206661396262,0.23603491605317573,0.8466369706263142,0.2938110296592915,
 0.8752336212423488,0.4220793100367436,0.40843923939697735,0.2110739958391208,
 0.5277895244384375,0.5564956771340296,3.216680316229237e-4,
 0.0397605230685264,0.6703771784997065,0.13751213393792105,
 0.16324713612223618,0.8671152741848697,0.5947815072505818,
 0.31559146548600614,0.4950401855356513,0.516965306758076,0.4461865710840558,
 0.7915035225204076,0.38419505124431697,0.2727261775123182,0.6476401724242801,
 0.4894472399675207,0.7055963332299173,0.36398002455127154,
 0.20411201737053308,0.7941526836067339,0.7566137734850444,0.6503889742315417,
 0.4052384601401944,0.24866299213083753,0.05911519219999206,
 0.6330777381767084,0.971229598664813,0.4984291785783517,0.1337913863281759,
 0.8444614265006292,0.9695384513332411,0.9834209280896377,0.04116193193678108,
 0.39135288006591185,0.9988259429942191,0.9137227293983835,0.3545647396948639,
 0.7051417557035934,0.10102706571390241,0.4775720675666192,0.5084642751880156,
 0.5098211754491152,0.9694796601905481,0.06658171502430843,0.7321584568747744,
 0.5556498557105023,0.9839330289155761,0.7720092411351966,0.7735087814528324,
 0.7632161532271196,0.8269197822261514,0.891310234307302,0.9203778926403122,
 0.4216180501542589,0.9435672864754951,0.830726983589687,0.2728867489921364,
 0.2798471452961555,0.3976795319407478,0.40754067834158114,0.8037660282663208,
 0.1923103937346482,0.701431337644342,0.11699955543913232,0.16100406498594444,
 0.6843124276253981,0.7260194870587287,0.06933963437851509,
 0.19304624413395344,0.4338598123391282,0.9786598220739542,
 0.44740236177672454,0.029074694035139848,0.38719299254050665,
 0.5708758627174981,0.8992275598609552,0.4992708493142588,0.6125031542868793,
 0.09493008642974898,0.5178759606880153,0.7678131056399519,0.629183175681906,
 0.01701609556277972,0.048752973659654586,0.8202432994821901,
 0.8243226775123607,0.26347442750141403,0.8793110870478229,
 0.47367883330460536,0.08629985224053327,0.1856480172324162,
 0.9232426103581063,0.2697702577107931,0.4236798432185265,
 0.031661932921618896,0.5852813742703347,0.7129159343640706,
 0.24455236956756943,0.9038375658858222,0.8186911994231973,0.1546755151743595,
 0.6416411100699276,0.5134580258750893,0.983864857817379,0.6156420882319328,
 0.5515103180014646,0.7048625479867867,0.6124613868291069,
 0.009269422152787365,0.29773418064169455,0.35769315808681046,
 0.928443635820918,0.191881558193281,0.9325624470028056,0.8043208062696441,
 0.14072201482163804,0.08112782705354371,0.43995020538652096,
 0.7124760576201212,0.14555439639544665,0.9087222677417005,0.8193940417067848,
 0.05261407458775236,0.10122672615320116,0.8365279294735218,
 0.7545821673334088,0.3826759862018948,0.6149596544553753,0.6403721498334588,
 0.19549494626985064,0.29369168450619,0.2421807546790573,0.6881238628208204,
 0.1617038108978628,0.5368277206213863,0.9010934956237663,0.03195984658804962,
 0.0705286196265813,0.5436953303270952,0.42648030528684233,
 0.28837292918098534,0.5519006071257042,0.9943185229932422,0.6113612311219955,
 0.5065825083295015,0.7553653130990744,0.08140930721712536,0.8722861353065198,
 0.8704177340794048,0.2993384080693876,0.7700408007008573,0.506155093714225,
 0.29858927730425744,0.5026863285339814,0.1158315457411393,
 0.19124590976448463,0.32806490347062633,0.7942922697301099,
 0.21228226352394397,0.7890793518758044,0.4837500240047107,
 0.15765993745492557,0.8190763755835548,0.6212224614231695,
 0.024473737805078333,0.5367679931668439,0.39842258952400733,
 0.6934637791193741,0.3819401247107108,0.927652436065494,0.29175992210187296,
 0.6005237751828679,0.516813122470645,0.7465372649698804,0.6399662521999061,
 0.7943140143197842,0.40786333539693675,0.45076130604186526,
 0.7076441622254614,0.5671268644595985,0.043723743148911365,
 0.5957356739077855,0.17359095316047668,0.1715273600117072,0.5802726946959319,
 0.03229649479399177,0.7329856521881805,0.9639996873866968,0.7027873134763487,
 0.6925288460742136,0.27550492740295396,0.5475728212654292,
 0.38134362659995724,0.5545653304671613,0.5382770365820465,0.7077139590762512,
 0.20020838944941244,0.6558323094500121,0.6118105683308424,
 0.04393900275576945,0.021230299440139988,0.4666458996556019,
 0.32587702271137875,0.8508802261744182,0.13712731737026618,
 0.9071470494425136,0.6197426991554561,0.640545540230316,0.18681311748164042,
 0.8384413436886062,0.0077944724996462655,0.5445189605753514,
 0.2123454236234701,0.19882632343840567,0.6357620760262782,0.9466240392153131,
 0.7273311132768262,0.890937792448603,0.960069665835572,0.4678250369682193,
 0.5874863393410789,0.08532332982064894,0.19995192015708108,
 0.5902124904819624,0.027295122085885026,0.03418597399677814,
 0.6743051798140194,0.5188650547273016,0.603788866216976,0.3551978138282965,
 0.3576225180093394,0.7046775121551024,0.4033654691252011,0.37844762876869675,
 0.11320477187392597,0.3360019276952134,0.8788283458162882,
 0.25083720004612664,0.8043774190283006,0.6522120599698821,
 0.17722927726626558,0.8289579415425039,0.6597313183195389,0.9021915435319211,
 0.3543165735526954,0.08321208124765433,0.7327637865262715,0.5748268796067381,
 0.5071326196092311,0.35043544706878893,0.9838379602743739,0.4826227917513567,
 0.793735984371593,0.5901943317360783,0.5659463124093621,0.01607030245150376,
 0.05391640060332148,0.935827737028023,0.20188758609409296,0.667451358541604,
 0.5213947367797251,0.9899909425789605,0.9642874941229683,0.8115737547321374,
 0.9137645551275101,0.8778565663290109,0.7340779107071925,0.975967938784179,
 0.34659637797527676,0.7035327913154683,0.1548879419800422,0.5171109450369501,
 0.6221400243035851,0.007192051568600366,0.8172675358699236,
 0.6643396019721224,0.9821389702349551,0.9281731961605317,0.23094275193201974,
 0.2026992573233778,0.6021604361487568,0.5956391250589391,0.34150578076925,
 0.38624282680609656,0.01603716012179013,0.992356109058851,0.8297065846639746,
 0.7017180064869701,0.36682165786465637,0.49902081388903474,
 0.9868353307490031,0.8786785961919131,0.060483728454373664,
 0.33364430165082914,0.8308057776475748,0.8634409190519415,0.721361972166252,
 0.424184739484584,0.1854733188262041,0.8966982603998456,0.4160211314584005,
 0.1458329663565514,0.7719692334179852,0.07425999055082433,0.6340589029855384,
 0.8641228524941943,0.7811003272082608,0.6632905241095642,0.45967280595992843,
 0.780244407883877,0.548481701679404,0.22553239456378293,0.25553477702605853,
 0.03822124131450744,0.3124359800752621,0.4181812340579031,0.7956656959850279,
 0.987544183664407,0.5647892620692699,0.3814784486588332,0.6165593185612279,
 0.7084945426188121,0.31518475068511753,0.828999782916287,0.7367566612476887,
 0.33923621742139887,0.30927649183289496,0.2604002933748447,
 0.7872886089102771,0.07452155428085594,0.7446004958320115,0.734680037664668,
 0.03354112118436725,0.6287362888454878,0.8363265439852523,0.0640774588570645,
 0.6573572573467302,0.7735969288155591,0.5095598273398824,0.1966987589664123,
 0.46554717311594485,0.6588280321160628,0.8403502746955797,0.4501245833324997,
 0.45496934073549355,0.2649543384658207,0.9605989723328678,0.8470477883855418,
 0.8912518632118662,0.8220187953551493,0.6223121334611974,
 0.016047393584856184,0.6386216733514023,0.5695476105660984,
 0.9629034479253649,0.11077702211971119,0.7432834622721606,0.743707857187115,
 0.5495577683713376,0.637241527531432,0.06581742685468128,0.17835595396900672,
 0.2774970012160577,0.09844402397958896,0.7413952349519528,0.4961606307738534,
 0.46987120670984306,0.7017345756979809,0.8710870170131751,0.8498424530706336,
 0.46279130488428,0.3556494304821881,0.12004491476603418,0.05142277535429274,
 0.21726764639987206,0.849107023713083,0.7073971116247862,0.6406572512306886,
 0.4128429443942093,0.2708301482126809,0.769500122745741,0.0861952138510248,
 0.03494154088893586,0.5302076273191597,0.3370565726870619,0.5929075731242217,
 0.9278260951777797,0.5459811459788483,0.11953546972512286,0.2910736735447045,
 0.6624750452275046,0.803535718213508,0.19655436791688652,0.860555219297477,
 0.6855032771802506,0.2268550065615098,0.8653793549927021,0.0203393565148835,
 0.03558035464196507,0.11765995022203068,0.6580481579998212,
 0.3644577156339589,0.9149521829781846,0.30011146173130077,
 0.43658550898509163,0.7942745633368566,0.5685288637415528,0.7884286794544879,
 0.03302315760648544,0.0728304788537627,0.7304074823611963,
 0.33555274753610453,0.09776548573471533,0.4182921007188728,
 0.8258938246132093,0.21246276391392027,0.4023030012379962,0.5553654831312155,
 0.11143556092577489,0.4120393719103509,0.9876246083340038,
 0.058821411353746855,0.7620365809525286,0.2983372844057397,
 0.05184735363224502,0.10559176733240072,0.23955262132627309,
 0.5414331507714211,0.4517968608468269,0.6460012932096383,0.08708381469403725,
 0.9948765628500876,0.8835815245591986,0.40763106229269863,
 0.34689469980451326,0.5671043038774796,0.5428086782483811,0.7141180266851294,
 0.8156063608315749,0.5587458843130155,0.13776790226854674,0.1357674980988991,
 0.2099132815943272,0.8759964805130073,0.4496605071285723,0.6093978778707901,
 0.9881212217087616,0.5955268682672344,0.2662069148813492,0.8053896713118697,
 0.5431541850758672,0.9603543425593012,0.2015099726536177,0.5885490579093192,
 0.14387571380364006,0.6045530639798192,0.6430305348043203,
 0.10586554245058055,0.9954515042437904,0.22297827565870487,
 0.15830782674422006,0.7795455980929806,0.27114712231672955,
 0.5304849362707929,0.118109330219197,0.22146244293188333,0.29629837398756065,
 0.6361799697437323,0.6298578873952582,0.5366575925766427,0.932279683671122,
 0.4346566075378918,0.04492472802086833,0.4913246812535259,
 0.49917959181041816,0.5684638529375142,0.5158200741123513,
 0.10391816140623145,0.028325947943652507,0.8282446552105885,
 0.5945129324204599,0.9670211164226914,0.010101787245346827,
 0.2507138307158697,0.13398443611538957,0.6322593485166103,
 0.15956454656637375,0.6925979151635495,0.13747341679030212,
 0.5997371971746238,0.724437704568178,0.6597545057029818,0.6445961132863013,
 0.6575666279804533,0.19358809335928073,0.4457684991222205,
 0.12618646967743952,0.9066809109101721,0.396835242836881,0.21679268100481652,
 0.8631497437383455,0.5484396012742715,0.9555726910738279,0.8406962292139308,
 0.23862917017073026,0.3292205724327316,0.3751736467683452,
 0.21069494241623554,0.9135487614232549,0.8618304511892011,0.6256723671284427,
 0.4087882304274727,0.1735251286852415,0.3407048369337172,0.24739051886436692,
 0.9815913231632805,0.3213097378143198,0.8377843853783378,0.23154731869392975,
 0.8550488025680054,0.9842941944170578,0.2696969082148115,0.4531233137952163,
 0.026979887901839472,0.7856066183457682,0.9223546406906153,
 0.9043328306718558,0.3738772936019329,0.1914939894553973,0.8507418256482638,
 0.9743313188387117,0.5823503562495107,0.41758855655084415,0.8926242120560566,
 0.03268630695805641,0.36972279931493235,0.789056530645772,0.5226865885003353,
 0.5030745721846889,0.8120219669949676,0.3537016162688926,0.6015901010557022,
 0.6332796188344608,0.26173462730132235,0.5136734487894239,
 5.641752827449942e-4,0.3655161439371758,0.9810249404531931,
 0.40982005679884514,0.08207729472205283,0.9773770963146956,
 0.7140649431690785,0.7269078365979488,0.6212451543384098,0.8945320178104426,
 0.3627957292875923,0.28495578085539364,0.6740220514004572,0.3511892412776728,
 0.9156622989763997,0.23975461766903683,0.607013667179277,0.3773848794531629,
 0.21597415295939348,0.9997506539445027,0.9826609408598431,0.9291047363978887,
 0.5921716175272835,0.4559323155324012,0.214745480465945,0.9155406438500531,
 0.4711968927080372,0.753089552363754,0.739250082150272,0.5048108132742682,
 0.15191238935311513,0.6508778624497739,0.05409795377247073,
 0.34090560406899784,0.6168734027526324,0.4561121152233475,0.5837288715457742,
 0.47346893541491475,0.37938176063435247,0.3864923523757896,
 0.03139953748386959,0.20553425170145334,0.4407743446273168,0.501031983582889,
 0.45360484925380384,0.7505747215218928,0.9518044486656386,
 0.45890648938434486,0.13862958548799054,0.9019275272278516,0.74245875083379,
 0.7650641837519055,0.9913853815227222,0.2599017776772613,0.4019806545288841,
 0.8373388466045993,0.8319791483155989,0.8225833694998868,0.5617612926224835,
 0.8089520901706111,0.9122416670903647,0.4435501210822136,0.858002571853358,
 0.7744763147161156,0.9058291397022862,0.7794599132880116,0.6117784771002351,
 0.7772514350674316,0.5746903671072151,0.10852319680282985,0.8045395472651693,
 0.3698801170183449,0.6382728463118079,0.4692577114062658,0.46182472498550275,
 0.9258934713973374,0.7150830943876763,0.9663164202029385,0.13959151216744015,
 0.6861279557994947,0.49487154253633414,0.6066456962597602,0.512688964144518,
 0.8329115019647029,0.37320413994177815,0.6647538550944119,
 0.33139874304849115,0.0648041664919713,0.6160862027828103,0.2293843796123215,
 0.2638042828885394,0.8654077905983466,0.3219496999077084,0.6569149718949989,
 0.437858819977177,0.5194963383000029,0.7179495454661182,0.9770375552300087,
 0.7417777391163235,0.7844539326638972,0.9853157895519352,
 0.014890226851662192,0.690311343623899,0.8028614621153112,0.7994483208907524,
 0.5142188015722919,0.415773677025013,0.5688758881289145,0.10168400778559727,
 0.8987089015911397,0.9784483099685471,0.9966286360635275,0.5092472786245715,
 0.7267945370198613,0.9237206888534926,0.8358832727327061,0.14973557814047633,
 0.568146015944306,0.2946954455073728,0.7730750210039206,0.5036337599228434,
 0.611171485629109,0.9257739253148918,0.7388867238946086,0.15010210372455757,
 0.2443764891716631,0.9660408155775213,0.39375635695441813,0.9555952814337227,
 0.021184586802505345,0.918796287793324,0.6461076098314137,0.532387981662953,
 0.26382160034771296,0.08951903702690645,0.38636695730099524,
 0.6464152785188126,0.16450663405977917,0.2752410383253764,0.6983082930327523,
 0.6769009303506541,0.8487390927871031,0.22059957380454998,0.2517987131897579,
 0.5961661372818646,0.052181105045025156,0.15544525402286236,
 0.8168539050209103,0.22209599765757038,0.7004518118999127,0.846014027141454,
 0.3257327379576598,0.4764816347758166,0.6487146650401684,0.2955380121710742,
 0.20175667524853247,0.7363729500046672,0.1718058345341652,0.5226185670869227,
 0.8341280765725758,0.3651437546706804,0.49708990572071077,0.5327835006958749,
 0.9038839608445358,0.33053966096622345,0.5733151596252564,0.3802423749081879,
 0.09372504567426199,0.1557458726712686,0.1792487626474848,
 0.019320067007078223,0.6548379003795021,0.19127910514855173,
 0.8726382676473566,0.909937773440164,0.7767507066626689,0.3500332181746114,
 0.29781299613779777,0.4439603296878901,0.5432899724477851,0.7726652253877194,
 0.15151278943331514,0.8477582130493538,0.393409606408868,0.23158760736216866,
 0.8848665428727555,0.2663630901855566,0.29319480318976265,0.794151906641072,
 0.44425334035392594,0.8611060701807612,0.027228910121792627,
 0.7333489746068534,0.8096702055191346,0.6772693344846655,0.8602448018017891,
 0.7430049305212812,0.20144521132202442,0.6386463969375061,0.5901327390514653,
 0.058463837994314094,0.7881828301700525,0.2244173057840353,
 0.4983452614230943,0.20305797525248737,0.3534280516161915,0.5809500526621006,
 0.045329606550618795,0.42702691069689847,0.05553572714456145,
 0.15384264031191675,0.3045070584402999,0.4228414252749989,
 0.44606966291750105,0.37261107429657536,0.24088739739829945,
 0.7944825241654916,0.4890367754029349,0.8710981816631319,0.03069832379157711,
 0.00801151420387125,0.024865096154401356,0.1512803030119001,
 0.7443095029543753,0.0634255505574507,0.7215442246547368,0.6634791202200675,
 0.45501299849244337,0.8166984202429604,0.27609574598662623,
 0.17890089582371138,0.18132760978675988,0.9606879079664536,
 0.43963476890659015,0.7351460577946399,0.8139298380740605,0.5353213776840332,
 0.12606729850479126,0.01499235851730829,0.6610260235255051,
 0.6377829966363634,0.513782306492386,0.023778026550595688,0.5099309523446179,
 0.9433110907925756,0.025163179363504096,0.5685390763140264,
 0.3843204105744179,0.08004472793052142,0.3555888439884496,0.5195215276005134,
 0.5177165023625414,0.6121974994419717,0.9509185878887148,0.8052200772780522,
 0.3641304850121627,0.4913202176087933,0.8442659944401313,0.9137634685815006,
 0.8981512407500806,0.6867008551363237,0.04349539557719306,0.4927192714167341,
 0.8164968653216953,0.5091936825987953,0.03414762705111374,0.8276462220671809,
 0.16798739169209997,0.12147657041240745,0.682288282096218,0.6414311247950385,
 0.13280027089090718,0.9578864709500757,0.25277224232371454,
 0.12680783801448936,0.9800982764015055,0.12390555200385078,
 0.6712216438355367,0.20064509952212395,0.6081518225243058,0.7457065780936045,
 0.49429639987286655,0.924866475329848,0.3963136867878704,0.9614966839498085,
 0.6393445586023493,0.0670572272876584,0.5138379557439929,0.4673367131646694,
 0.7207390193362423,0.626895268436395,0.4784567208838436,0.4928405696799196,
 0.07611922531863913,0.8817468549011163,0.859718338340055,0.7638732049628485,
 0.317886630754455,0.6972397759953363,0.13375111556072106,0.6591141052244772,
 0.3581447672202174,0.32145910846213055,0.7650986540486845,0.538497921063428,
 0.054324402254520954,0.24409340701822424,0.3603383148967928,
 0.5828118327445716,0.37596277729418326,0.0996833007505108,0.8545765622235817,
 0.21932818357267303,0.7663937685085878,0.1387628776463491,0.9592759414020301,
 0.07798218170296822,0.03952707952166612,0.9242957334376877,
 0.29512071742237656,0.28413207724493283,0.3486932283860532,
 0.29751172866902487,0.0397756349045536,0.6537963926775111,0.47637983042152,
 0.2965202092977963,0.8874644865160752,0.08365929273067929,0.7257484943546011,
 0.9646380564140211,0.5130710970835546,0.37012253838750553,0.9218337437379244,
 0.1294046485837711,0.764036151071451,0.3552915447918631,0.8416932077229164,
 0.009034762663190476,0.8791539388364321,0.19683171772422126,
 0.234745584609089,0.5662210680605426,0.027491807066745477,0.4226670156471263,
 0.25981237787001554,0.18155301776477217,0.7548374785049806,0.793638226546531,
 0.5839635441853727,0.6486658638949979,0.6642148651609147,0.34453434996223986,
 0.6410702889064062,0.7659770096167517,0.1521504229399776,0.8439127027699189,
 0.4347627797534742,0.7834026568716905,0.1422635397859171,0.12708853992872404,
 0.07284813665243672,0.5466787476432322,0.8163601389264665,
 0.43303830022321244,0.5746984303811286,0.04748008536517323,
 0.7748587691578123,0.4359928213062797,0.5913827285662379,0.4505695774837871,
 0.2607978047415511,0.23580309325673088,0.2482900853796769,
 0.15748300494064693,0.7167620627514953,0.49528819043538597,
 0.43529351851529374,0.9436319086427679,0.8802007117083095,0.7144433268642408,
 0.5572746497737916,0.7268017221893713,0.4025895018604233,0.7096947949483954,
 0.04906288208286391,0.8216490619384204,0.23253393590999094,
 0.06866508547849626,0.49439843367677383,0.20753805124718183,
 0.3511524257955708,0.4571593806834904,9.742129263502486e-4,
 0.9344456515337864,0.09712862169174186,0.4318429191413681,0.6811936902028063,
 0.6345989637560693,0.02605873789152391,0.5703703706421631,0.773920099482183,
 0.10932701676094769,0.7283246839259253,0.8289196892193067,0.913532596561459,
 0.45220174563662885,0.3528253583121288,0.794735730647444,0.44751915721864666,
 0.21175302103448213,0.8586036757051191,0.16405803305952416,
 0.016828653878902977,0.7128238436923742,0.8870024901338579,
 0.0468527636956213,0.2159646504406998,0.20124560247148748,
 0.19633450145586817,0.25201709682252926,0.3795965182049762,
 2.5825062732498516e-4,0.23444383948819447,0.7442363301397902,
 0.6415883263871706,0.05208060198102882,0.6388190950060124,
 0.12404857820486459,0.639725463826588,0.2851200420052452,0.22919984789783399,
 0.32345660614599825,0.9372785506775929,0.382624306654221,0.18872661096489718,
 0.733889730828349,0.06688716821992757,0.23866793361020144,0.6271858027764828,
 0.8986762969306508,0.7152250560768083,0.4991088380742925,0.7214920926409941,
 0.12468273871040747,0.6601614738838398,0.4753975640899598,
 0.008366355027959127,0.03224233883314309,0.7244255580602228,
 0.7954278136775998,0.5304189518034697,0.7044644023319783,0.720011009960968,
 0.33638073452443307,0.6764417060787841,0.6350385927520703,
 0.12686532309884757,0.5828462932303472,0.9473311605485298,0.7516321843893575,
 0.6625453745556652,0.07480233851238693,0.15498560063805733,
 0.13547290656931443,0.9989678074578862,0.6391914186238477,
 0.10718909991094538,0.1590816711545492,0.7784392947718162,0.8249937860456066,
 0.31291209557983257,0.6358217885290911,0.6749296318066836,
 0.13303169293345773,0.5879412374993216,0.9786982094645259,0.3428935262678898,
 0.8188509745398267,0.5996028412536276,0.8230382461159775,0.6502184297423704,
 0.6662422330858986,0.5142493986652813,0.6793163098172221,0.6392043090399282,
 0.38697605999444395,0.9969625448613314,0.4479400585291665,0.7228566329920777,
 0.3780589098404752,0.8692807914310917,0.7865107678454369,0.4971941292717813,
 0.4150436965866674,0.2624406089616592,0.41029462908621184,0.7708873076997421,
 0.7191492818783969,0.157301184486766,0.9739904359127036,0.026856397474549265,
 0.6767560489764566,0.8065698345379682,0.15105081495577466,
 0.15346993669789866,0.4122526546981681,0.41076292883975796,
 0.19160474480966672,0.24357985437977747,0.3268930313435443,
 0.7198454828042589,0.8914274338706826,0.4300723626999623,0.8176546925266249,
 0.8785752627562696,0.20458162890120968,0.35733378819895467,
 0.8686130642057249,0.17262470018448361,0.7936069823991954,
 0.17903933634212776,0.2037088820079347,0.6424406409077942,
 0.45834482889938877,0.24448088882511032,0.5302855149516535,
 0.04523967900484527,0.7479618569261413,0.486639611238913,0.9242426264058848,
 0.7485776968160466,0.381872684131364,0.7197553241724983,0.3781054740153641,
 0.8999775940866654,0.7229507499666958,0.8960890015859112,0.8071389179246599,
 0.9810257666769208,0.6241733276646031,0.31479116113508954,0.7613223466994279,
 0.07437526314134912,0.8380660282435957,0.18889676001829647,
 0.8653507999344792,0.707094005678146,0.9406403891878811,0.5111482751520542,
 0.9151113135742447,0.9994733618611077,0.7407339369581807,0.2863686330805928,
 0.09289652617226718,0.48054039948145766,0.9154854713912044,
 0.9761168308868327,0.4714098450053694,0.10688787105977338,0.8584681707586606,
 0.5183022025807321,0.4250297359540216,0.2848370274036034,0.799549686641466,
 0.30324804869940725,0.9815792403045611,0.3015579551372026,0.8702081823987764,
 0.8328609691648445,0.8807581097098258,0.7498372489673402,0.8394522825540406,
 0.7886815387003394,0.3418215960678824,0.23846693465861923,0.5529548611891845,
 0.4837692229499605,0.2835997562629312,0.32207497855972633,0.7537700983691222,
 0.6332255230980973,0.9244896154254316,0.649949010793822,0.32929978212643907,
 0.5446714512853666,0.8750127940956111,0.9727958876459264,0.5873505750245592,
 0.5109077443185426,0.12814397222154683,0.506616569755722,0.38854377868353895,
 0.23584691761426024,0.6003801848820373,0.5911891027548161,0.7981646900382907,
 0.6193790634921024,0.6955683850408054,0.7496736404048278,0.8464291662776721,
 0.8396464605748074,0.7912052697701721,0.9037877101996946,0.38386620665554805,
 0.37607202052352107,0.6750764628668424,0.6687010173478593,
 0.23523377107858656,0.8508059416137173,0.2192821389277796,0.5699754273695121,
 0.7447877521267215,0.9791950102962975,0.8983540114561442,0.8858392293578777,
 0.2253007913922982,0.7343235962517086,0.3176528712780081,0.3027543506365691,
 0.5397491187038463,0.07405482324972046,0.27965048390185165,
 0.42558101756969446,0.11479802858766897,0.9467823529303505,
 0.7326922642008056,0.20332653315231664,0.5656365055698376,0.2764586091222574,
 0.08471679885994343,0.3795044445633713,0.2616191955028905,0.9807530666064109,
 0.2450067114326171,0.8224761854378224,0.2033162115618513,0.625158927985429,
 0.9974714547861504,0.43065879149039776,0.6661606306048432,0.7729281413871621,
 0.8444141532432207,0.08692522121511548,0.27170150132432624,
 0.7784050277706372,0.3326415942771551,0.8661548931256269,0.2126533449049166,
 0.43507647445939035,0.27947282071254953,0.6984009909124865,
 0.9241155923289379,0.5932319724607581,0.5887267111434088,0.5140944151422113,
 0.08322411832472698,0.4316077396316179,0.12554831074384398,
 0.40635963439591305,0.5325544007649141,0.2889231171109974,0.6165875078847263,
 0.14688646869310407,0.17229703745214875,0.7692431218284614,
 0.9900224613682832,0.6565820020469725,0.7940027351840839,0.5902311282103078,
 0.19513428386640785,0.2911227943844106,0.7638491333734979,0.9790582121404862,
 0.05046097270786587,0.2614860236238479,0.8292872749851731,0.1991379471544995,
 0.9554623634976606,0.07494291541400788,0.44994866272203726,
 0.17996652121242052,0.21630913399181317,0.9279022471588558,
 0.10877668838053522,0.8072310397816973,0.5709798265795165,0.7027436787165857,
 0.6258422257355123,0.5534435511351293,0.611987933866417,0.2579102195483537,
 0.5217762235756462,0.44031897381342866,0.9414937759262567,0.1174325131219427,
 0.28155259770945285,0.3276294837078356,0.9874433316744572,0.6492286217599108,
 0.11349587161632557,0.11804255001371167,0.08715675113672494,
 0.8826359430064284,0.5449548515144237,0.25012197248428314,
 0.013775348998032677,0.7579316806520215,0.45078709862738553,
 0.055336959239755346,0.9340886574190816,0.43400713272463654,
 0.49902204248121373,0.7551146643930303,0.6358262462541969,0.6790262101701666,
 0.6579302476573341,0.379675779727196,0.005349323529100225,0.5219796003181477,
 0.4461409210154822,0.2765395451725139,0.42356292653357297,
 0.13686857448324774,0.5900652066812421,0.13089004512439273,
 0.21263122630999098,0.4509114394392264,0.3608428844860353,0.675153381022827,
 0.3647050082867028,0.06051874041243066,0.32341967670907934,
 0.38868057650872356,0.24783768252475213,0.9492127698835305,
 0.1848313976286038,0.6165277437848283,0.8180815376958512,0.807158205212152,
 0.45221944809469183,0.521831971477827,0.551596537183833,0.4544120578824071,
 0.9194529435108965,0.4920314945158686,0.5794322662163311,0.7473815207261043,
 0.11302529076093415,0.5964501405835454,0.028050531532483713,
 0.24441175038763818,0.3443863202642983,0.696031104783153,0.14998024087738515,
 0.4065951552628905,0.02094198627270938,0.22942871993957925,
 0.6456638997831339,0.14262236932930428,0.32507564992171034,0.908114567196281,
 0.2805729506285177,0.8284212480383485,0.8648904086512229,0.8332743016102696,
 0.5299875647289202,0.498217505387782,0.5011009633511089,0.13296473081590787,
 0.2024818009767264,0.3268514807988928,0.8649892860037323,0.9723699889215945,
 0.1314183650093853,0.9648974268003343,0.4674262272119891,0.9308395725665989,
 0.11891168564038201,0.27202544732266,0.896142277503431,0.34834678061464264,
 0.2750203976823996,0.6456065464782402,0.10247736472759927,0.7052785174358206,
 0.6210180781966634,0.9063175380425651,0.45429173290198055,0.7588524881437057,
 0.7579083938487373,0.5763969611824298,0.25862854421893255,0.2470916273313346,
 0.05043959686688115,0.7656928538005001,0.5562100171954542,0.3725012101003625,
 0.6686879553577958,0.8689287314743109,0.1450480840452677,0.4482256301419474,
 0.37791497851151234,0.8042092158877738,0.6355337178907421,0.6805376820754918,
 0.509017792400042,0.659686482603314,0.06799581633793028,0.41717092070024975,
 0.5438840868279414,0.3174886641335021,0.4498909087066971,0.6432671753702403,
 0.3160391109275965,0.5079243936003391,0.9574477106772198,0.5726443941570631,
 0.2293913081254877,0.23915560164252714,0.3441928711205582,0.5542377353525448,
 0.9023361095984732,0.027875427905762695,0.2597620626613476,
 0.20210341102718787,0.6911750156713143,0.06424643411690034,
 0.27923490142509166,0.8636080945569918,0.6916714547269845,0.4785772546719895,
 0.816977256938761,0.39350502634057194,0.31378739764934094,
 0.46369839622845543,0.41698142256339277,0.6227117976913744,
 0.40477804511010373,0.2695166118697705,0.68739748444901,0.5554190962581531,
 0.8917286979085791,0.41160170614431435,0.6924558486278265,
 0.14971043332206857,0.42825629479948946,0.18838713545561458,
 0.8861609166255329,0.4568266722409158,0.980147576648487,0.5341875931672785,
 0.04539895001910588,0.7781235127320565,0.3156044358756559,0.6507421115101766,
 0.09436708539269201,0.8961438489197542,0.5660729370390818,
 0.07757854461943392,0.6132487655982826,0.015850802333987057,
 0.5758476369585129,0.2845984553998415,0.9621514069771859,0.41608604371679236,
 0.9965080699613165,0.17508873318517326,0.8024417772156098,0.7976605235029918,
 0.08124759761489853,0.2318470652514475,0.6435618596282497,0.7654860811123153,
 0.7540690762966424,0.17880407264387155,0.7248653296852648,
 0.41304715468805164,0.8553277059058935,0.014004422978367015,
 0.7125772644869596,0.1270670476635939,0.5905414048220085,0.07865781433418106,
 0.7940330043677268,0.5332344993152549,0.8412762635400464,0.7345824113561039,
 0.3324147123669142,0.9242799519845326,0.16030763047948926,0.7094169523591113,
 0.10899886939532166,0.4006455302843023,0.2966848584599796,0.9221144767588114,
 0.1707357057931046,0.07304536867600586,0.1267353678789479,0.7900207495923428,
 0.08799542917162395,0.94935949294646,0.6209914659742559,0.007235899256534117,
 0.12489957565256271,0.12051262754052416,0.4086107980968955,
 0.8912480435792844,0.873108784056429,0.33717665079850445,0.6244928085039047,
 0.7141575054511585,0.547158307449161,0.12614358627809408,0.8755837710881411,
 0.842374871367553,0.5170950906729658,0.6877630676711505,0.2890120337742219,
 0.3285775483122917,0.18824925683678195,0.6373763652419087,
 0.05519612866140722,0.5097299840554659,0.11518538288243352,0.849887345583732,
 0.5836495857842783,0.9885201846964394,0.5254997435933902,0.9182564175820244,
 0.937722305041916,0.39473582020379927,0.7651736790180168,0.5503448102119737,
 0.4352554257805794,0.9262435312031505,0.24956494139224172,0.6719031485370089,
 0.08150188847840734,0.5482852311907223,0.4131205666583765,0.6900846486857195,
 0.22052603008501603,0.8656644080600802,0.5275192532287079,0.8412819941884525,
 0.6491200821867795,0.3316259476430228,0.4791085558365282,0.5920635593951621,
 0.22787866970374515,0.7790931478422134,0.45170771385959774,
 0.7712706907979616,0.3590460646124508,0.6084786744212879,0.13738160918548725,
 0.09162090967015246,0.38358194958648706,0.548557805309722,
 0.36134343584173967,0.7111091703439297,0.2199359518957673,0.780656418207422,
 0.3462082161409944,0.5425552959268289,0.779504626314913,0.9611990487529937,
 0.12371424153199773,0.07333394797470916,0.2250114941066753,
 0.9293925345232334,0.47580490130727393,0.05903400736089626,
 0.29705524095915214,0.31281407908584813,0.2724256137356613,
 0.8947280852464045,0.3174799065432551,0.009041220139701545,
 0.3620630156825426,0.957080968436063,0.47519502690769566,0.319725554899459,
 0.3327031843556205,0.7072247397512939,0.6116793326515636,0.42274151241383584,
 0.5926613606522021,0.19006887358076419,0.6364902079299752,0.6531382429777036,
 0.7521836911039614,0.486568563771445,0.4865454567145552,0.04217028182534488,
 0.10865733309875669,0.20916549994804534,0.16002504743568136,
 0.755219612671536,0.8648333557188596,0.8273322714281994,0.8563327450927811,
 0.12363707624508313,0.471512712975922,0.6898070832226748,
 0.019848620793116112,0.2337646407291819,0.6138234554601417,
 0.7112392498233157,0.22120295214145258,0.8389136858811278,0.9247158913707622,
 0.5356747536234504,0.6556467004442573,0.6952484259156653,0.3481147478878236,
 0.6814432672148351,0.2462012082178825,0.08779607710102533,0.4262733443843282,
 0.8396674637144561,0.8275498656844219,0.8976886098271888,0.9450290997066771,
 0.4921747192233644,0.24931098469142599,0.6999104296880154,0.6955033623368462,
 0.9596027156663057,0.20386708013470944,0.11336551750029167,
 0.6381025969352745,0.577783686567553,0.4924616091574625,0.36738621460524035,
 0.8031278991603968,0.5710163751018286,0.6273333157287823,0.40069898862831366,
 0.734074426037656,0.14595749332614505,0.5236424342259486,0.5436252042468752,
 0.6510450086424863,0.9527807264587074,0.7523001795342525,0.28920837856381687,
 0.15351763218380332,0.4898906448673441,0.06603201014522886,
 0.6616189760810299,0.042321032653108936,0.11685192797489963,
 0.48723700462472475,0.0546148275471241,0.5792107761020886,0.5349474603837086,
 0.4963870005804938,0.8480145435035243,0.18983575409785136,
 0.22555411244251944,0.6007462814175488,0.7920595437585902,0.9890296001748056,
 0.7954627352127369,0.014895859134938516,0.5778681142333446,
 0.1024541575847252,0.7711429241006462,0.7992980547114137,0.974583068709547,
 0.4512615145523263,0.3237737079388596,0.6316991735222204,0.2129961578446774,
 0.7503566381007559,0.17548977544590827,0.7885998996227684,0.3341700410842262,
 0.5497076909314366,0.3157519432694922,0.09208491129081431,0.578910432793157,
 0.38262228270993337,0.17687513931066234,0.1585497633101698,
 0.2631889433646527,0.32616828997887726,0.26039779557508147,0.964647217995513,
 0.809746795967039,0.7581116203862898,0.5207636090759701,0.2027294987705227,
 0.8675398325758804,0.1933214046868823,0.31562923548625876,0.2650393700908815,
 0.6423768558508763,0.09061791807653141,0.6532877494157393,0.7756687645716067,
 0.7283185934924821,0.43795939026680286,0.694684054293173,0.05286549967563137,
 0.633334757290787,0.6559513498806243,0.8813742431048737,0.8781167557020749,
 0.4732701370337039,0.11938966243997984,0.7516010410152821,0.7712444210238517,
 0.6386773189896282,0.45771594414281,0.28673448809845703,0.18016970864295,
 0.13699356044783972,0.23726512937137356,0.14814991296137003,
 0.6915972052668831,0.952986721928824,0.9709743763848817,0.6558078709734961,
 0.6721893853944656,0.393035617596315,0.2319104649175583,0.6848841578065576,
 0.8830531332815134,0.06508652155178751,0.07714156220096224,
 0.40180819758690023,0.4990524249796804,0.6742787171930755,0.8913117925547,
 0.10332938285430837,0.0036391922044453073,0.0032065129132886616,
 0.4401477733714353,0.02153835365139889,0.6878005978683452,0.4067033006254439,
 0.5002932935215716,0.06711333374914152,0.40429237225305337,0.6451777495538,
 0.9784569936128124,0.31207668495994345,0.6986634611537847,0.2378144517280918,
 0.9928025833671287,0.9319465006440582,0.006143211187681463,
 0.19371716918095394,0.2743131594999666,0.4974224612062331,0.3265045556080677,
 0.39608552164672284,0.4024763444580084,0.30199506605673665,
 0.25178797458690383,0.2802743287700127,0.5710637142375642,0.2481975142729348,
 0.1830917834952387,0.4130510271271617,0.8528374420208186,0.24306002531246695,
 0.8042347366599181,0.661224304901579,0.023989971495540274,0.7429175853355195,
 0.4946568909131315,0.048411183060160456,0.44131567546155737,
 0.20659255592524928,0.6381551456970536,0.7346062321416466,0.7708759401920773,
 0.009164607345300335,0.1413289683903648,0.11735559928284589,
 0.24761334110557764,0.9196994899102875,0.6815784160539748,0.6433835341217262,
 0.5854171568632907,0.17438421028828122,0.6893328509077155,0.8115867130329791,
 0.31105694532310235,0.6531252482371763,0.7251342652140413,0.6630879629879759,
 0.32968329284636055,0.5317737082591824,0.801560816488335,0.7372675139664525,
 0.12374740485494662,0.0068768861931874925,0.561232867662135,
 0.8648825118458002,0.9083478888602246,0.16934465331171278,0.9896791079314362,
 0.39699847303865343,0.7729427186687122,0.5201899922935728,
 0.47908780942126294,0.3247844692236287,0.4631886244639696,0.7936479531459953,
 0.3851468981390016,0.44797032009243465,0.3250886409050926,0.7414542301456805,
 0.3927868196983446,0.337384250343125,0.8181974510671359,0.7993594223205084,
 0.6625341218344195,0.9884990865802397,0.5108497576093328,0.1312349698332157,
 0.07358256828014764,0.24032301640478237,0.7234165069399255,
 0.5258651721511339,0.4356984349949906,0.2978032208923971,0.7054431677499218,
 0.7274518116554455,0.06330857782222765,0.5224492907447434,0.6563457169011274,
 0.5134919050668236,0.30575054122748213,0.6441068235721662,
 0.05032123846911385,0.17461270773414594,0.7480815819575922,
 0.8373232406773692,0.8480958648088035,0.7619298211451733,0.7274836826875185,
 0.5004191039317082,0.8469876234857149,0.6162144494686709,0.5318482408685582,
 0.30772831017334756,0.26074784123926853,0.3061197765167807,
 0.9061974490171215,0.4553843308239992,0.6309084476175558,
 5.327628012385954e-5,0.8612266875469026,0.8352067024850048,
 0.3472900159897405,0.5949906405345303,0.49946784312532866,0.6971396895566113,
 0.21034633614109743,0.9010636984753346,0.6210028600664415,0.3059296643133196,
 0.11304757216441397,0.36861571879651,0.7965946892096469,0.4696870603872295,
 0.30678819346878305,0.365413778259317,0.7627460246828706,0.6696200330877643,
 0.07331782253114683,0.5249006525176717,0.8380587006753251,
 0.32728445737516476,0.45348899062331427,0.04030777683091846,
 0.8388798374275676,0.4535764728547558,0.7052981715313553,
 0.061695799552100916,0.043670039645222425,0.18279909715825793,
 0.836179915921918,0.7470725672417593,0.04791353668172871,0.9342557428056854,
 0.2790616976831398,0.044424952428011366,0.22495757563776286,
 0.8826799540459791,0.1974448913592154,0.09312099663929352,0.8220727723187728,
 0.9550415199969678,0.09089659487488477,0.5969785808621119,0.5651936849438457,
 0.4880655480510201,0.0077849619379247725,0.5610314978229005,
 0.14929806879171093,0.5667744695962093,0.7940234860363242,0.6152508917888615,
 0.7826173373713353,0.9840341769245091,0.471153938437583,0.868942934219979,
 0.06477674260166777,0.7268215489545556,0.30702320555073637,
 0.3194857063865949,0.9575118362628181,0.44465472745288626,
 0.03547005618983656,0.19376701688386477,0.1342075221745811,
 0.44207828455116793,0.9807200659249796,0.3375965258821443,0.9605861054687078,
 0.7002269170940506,0.9476854877304027,0.703400331853087,0.6339252992229603,
 0.6052243288442718,0.7480877830994088,0.9392475283294995,0.13202820522562708,
 0.5543399421268321,0.1100860650614296,0.6281925000787767,0.09611962724129353,
 0.12589881926917212,0.14856704046886238,0.8340801855487598,
 0.3425289163434009,0.23914916282242515,0.585105746434781,0.7795026064489502,
 0.9874124811979277,0.88291880980771,0.20869456667342912,0.735530387109616,
 0.4892702293150548,0.4301227163295419,0.6194835067211768,0.5818512305431169,
 0.5121122248330794,0.8051266984063473,0.7672053623322487,0.5025221017809862,
 0.7254087518238728,0.44002187660378445,0.6137796089177457,0.7263831299493748,
 0.07492303545387369,0.7856054972308159,0.061185924898998434,
 0.7812588859692601,0.4137922305828384,0.48547463352930875,
 0.35124790299846476,0.41938347100757456,0.6637901936103802,
 0.2963495275184381,0.7889548417203646,0.9765646990953732,0.8137716441978557,
 0.7822665506791274,0.7726399698172249,0.8157165096132967,0.19601001740791524,
 0.026322973686136564,0.33477240316192414,0.6498330528467208,
 0.6804692857241252,0.47222199449620716,0.6147323773092183,0.7329103222919983,
 0.15517794810301577,0.14126497518052483,0.028128091841851943,
 0.15088111592043518,0.3282635317578094,0.7274206625752241,0.7457901321276467,
 0.16652386744804026,0.8552699814150238,0.9507634026935227,0.3365685939934804,
 0.11082990273580884,0.779468240728705,0.7423883360546767,0.1174249308009424,
 0.9417743667782084,0.40763643666726046,0.07506159369464349,
 0.4647848342845837,0.6864288874624159,0.23248581581518857,
 0.22641019249353977,0.7486001926905811,0.9773610802358302,0.6342999197593863,
 0.26626340055673836,0.5946408494915769,0.9320724404940021,0.7277155540386073,
 0.3314177842103383,0.3700072294018919,0.7687672891192217,0.33584664219499427,
 0.34588146189620383,0.6420173200986253,0.016938461929403292,
 0.8617001604954412,0.7029626197060753,0.7348740234084273,0.8003314576909896,
 0.08088435765782243,0.8286545295855278,0.13337665687644107,
 0.6786675712143792,0.9527286165723805,0.5113193543972266,0.6640997425033038,
 0.32589436232076385,0.0014075513843553322,0.7293726694548673,
 0.7381401261090235,0.5392698171086839,0.7644213867256752,0.9451445741981794,
 0.5411042807406626,0.694173707664622,0.9904225636235753,0.25526614308106343,
 0.18686095507219236,0.8828381800487494,0.7650218349410652,0.5433619957877245,
 0.7639202416120534,0.738497368593695,0.9763775564252527,0.08891194806185143,
 0.6153340515354598,0.8003080753586981,0.06381305528915338,0.3678925286001441,
 0.4739669323479493,0.9173872014071374,0.9665710571918088,0.05870956514385406,
 0.6100785922043996,0.6771691387854072,0.6227721277825707,0.8937405522799111,
 0.4064502192647753,0.9519678198835984,0.3807011083477826,0.07111782114049592,
 0.34796356840072296,0.8417770993974267,0.8639973320737266,
 0.01860990592790457,0.8043959327389589,0.2600372663451803,
 0.022994267108408017,0.881636712064559,0.8663796979923294,0.5626785744594199,
 0.44994254066253125,0.2156988684449045,0.632226832423573,0.2123023410451952,
 0.20636864464288152,0.9680618702357346,0.3364119824576961,0.6005766040194827,
 0.3168062742467258,0.988636296312038,0.6092533536856366,0.7610345484838684,
 0.7581390658177686,0.2797548948387024,0.10732833335809544,0.4678158356845934,
 0.03289946303981495,0.5798486010721655,0.9243476081028996,
 0.40351830108580666,0.15578543744407747,0.812433432286337,0.9982016639482777,
 0.6137419962251042,0.45892488344095606,0.6292519905818968,0.8812789960908277,
 0.656878450664841,0.8731568730139263,0.3162117361476241,0.48187501213983197,
 0.4166583895744169,0.16143444575055388,0.2418923072299889,
 0.29022606844702326,0.8338477600911954,0.13384926984908152,0.318326394673041,
 0.7966601747698895,0.9831183776803747,0.971985513124759,0.39710546285294834,
 0.7714295207073131,0.04283285217994692,0.6063221174133295,0.8237262679605892,
 0.9419990282490667,0.1425502422948941,0.9188315050984452,
 0.056408230497053125,0.9714289941790438,0.5858181653739121,
 0.06501103140998177,0.3468392282988555,0.8085086973546828,
 0.44800344921811797,0.192783838989651,0.7732831167381105,0.7453621068244244,
 0.25774145182957164,0.6067863602946558,0.815795967070205,0.7661008817969366,
 0.841546155145013,0.8965534140796692,0.41240709327592595,0.8433854245803575,
 0.9624795998219018,0.04660857188242917,0.6297504362434087,
 0.034535779654818355,0.8794993167360851,0.20002101030478847,
 0.707179227082154,0.8231756883170149,0.8100728051287749,0.26195263544348,
 0.1436163071106823,0.4698516147138174,0.2488616315085148,0.4018699019664038,
 0.29748574055485133,0.3840803230176464,0.6665471002011056,0.907190569488115,
 0.42698593876722035,0.11788752639784184,0.16101402596150083,
 0.36295939959238266,0.7711649427401992,0.23609756759731537,
 0.5258868175679386,0.795705919614787,0.26908037596594325,0.44266555571247745,
 0.26357374636705955,0.5444383855606334,0.7936256012559608,0.8261052515331658,
 0.7029563876448641,0.39132393951510225,0.026842824758744,0.39258084004950555,
 0.9273295704586124,0.9564110651693889,0.05653989231921719,
 0.43391621402378355,0.917358521421524,0.7738661006980653,0.23909958373355744,
 0.5906458221875311,0.7610743352511392,0.6471791107915054,0.8104304480389157,
 0.4202211829923642,0.5240544074274956,0.5538480961198489,0.6559716000977995,
 0.12316917684737116,0.09106605659967137,0.5490801176804742,
 0.3737988784506585,0.16753931008259648,0.24022979730127447,
 0.37721743847692035,0.5750238090453216,0.8097818087935108,
 0.13425795272143815,0.9557673757451185,0.2003342917434564,
 0.28322719712099753,0.14939002179614902,0.8698829503559394,0.790436339689916,
 0.44019917754235416,0.1275426323463702,0.8745101117141502,0.3331819293034547,
 0.6814782098892397,0.7563203534202639,0.11331288581547028,0.5147190816070347,
 0.47067173250674543,0.6112225797425587,0.1784609035191087,0.8988952553731122,
 0.009927377704550433,0.7774344187701325,0.9232477990047143,
 0.8500392993364414,0.5980001999428906,0.006384007211131526,
 0.9186730846902776,0.1236878525451014,0.7298682767895918,0.8584021233592185,
 0.2206768313391243,0.9157590331919931,0.360831142628945,0.8896766461074614,
 0.4399068776317314,0.004415678665451339,0.8500559139754551,
 0.1562818821462768,0.41382623228980575,0.28574168140524714,
 0.01025055051910495,0.24271147583130948,0.39457452762664125,
 0.5939884909000919,0.35946318258582033,0.9898358353534497,0.3651627856656535,
 0.5141382063625438,0.42335351343498606,0.1831151672551855,0.7087417904434399,
 0.26383366078865234,0.9654714649133269,0.6614821927161062,0.2818229610358238,
 0.1497231578979209,0.2677863558487015,0.6385108393143114,0.9770217403266398,
 0.2868447377177131,0.10398231117032075,0.4710914625432444,
 0.23967347667206773,0.42880052518809897,0.22746202148103434,
 0.9172435842347471,0.2776259607580279,0.2762840494434846,0.1927820569615476,
 0.8334639613622823,0.579390190727623,0.3257659665920226,0.44308573384879457,
 0.002036851571136644,0.7808477163975255,0.5219125901983093,
 0.3422520655010479,0.8555049192378481,0.7012033467969757,0.38201565784752356,
 0.7821965067717059,0.027611777965815154,0.9364896898348352,
 0.048994068988026274,0.48473751570160617,0.7446196867455207,
 0.8874432650035746,0.8798702644935931,0.9534420168947582,0.890761430223695,
 0.9215796458582651,0.6248576299739832,0.16519920084169593,0.8800244455461576,
 0.4360940174759166,0.9973423263222061,0.6697835232900516,0.770260499117832,
 0.914590638012774,0.3065900012809688,0.1319686839421571,0.5208295493361668,
 0.7036516822922336,0.3225740587346828,0.8741921401121455,0.27411524001948273,
 0.24597693266275078,0.6177555481208552,0.9709120393415104,
 0.27879772224816113,0.05715760215907384,0.8841908780230452,
 0.2988882129133271,0.36728929833938906,0.48375152763080986,
 0.6079691546652448,0.05818487095499214,0.7809853317338267,
 0.15087247319614183,0.5953530187025693,0.6639923567938436,0.9107691971511414,
 0.7066867614195954,0.9357497965284323,0.17661192988730656,
 0.25743516425517876,0.9358323622340441,0.574357535596014,0.1878583362508891,
 0.7141967669799314,0.7302650547514967,0.6930676609429042,0.3923574871894506,
 0.933125004374786,0.9045373686722057,0.23018136527938238,0.2247965457059491,
 0.5009493555725153,0.6934044935409452,0.7748349645181791,0.5500517420030233,
 0.08389343941186733,0.5863993949192295,0.5364828865248464,0.414565572100717,
 0.289657183071345,0.5224240807928286,0.9331492361150331,0.12355785995580915,
 0.49983645306487745,0.025598800303579083,0.028563615464058056,
 0.27417723419925455,0.5082346704482648,0.3636540521767091,0.6361618377881791,
 0.04818427631469224,0.4540503332608431,0.43311431483905105,
 0.47681945747007426,0.8251101930925822,0.3257406693009224,
 0.14285792037641754,0.14666730884671608,0.7260738176009542,
 0.7103249880139387,0.7697162172706103,0.16800573254264684,0.8035027891850143,
 0.1770119651736578,0.5724263540764077,0.5504062209469107,0.4791758427588728,
 0.001346788700470003,0.06427970062122457,0.6430838550127982,
 0.9442236141042454,0.7529674578147849,0.6048159614770807,0.8528020586687662,
 0.8087107236688964,0.43350784327554837,0.28732874695727806,
 0.03382575455492165,0.21306656825278725,0.4208663459080292,
 0.3042234482439927,0.5245305644974151,0.397826166861589,0.32252671417395484,
 0.7471485797171743,0.13067421720904182,0.645971897484612,0.159095866598285,
 0.21540871012597096,0.21010615799198729,0.8599263290597747,
 0.1619960982391382,0.02567224221717257,0.4105374759574937,
 0.11678528422711143,0.2515819384468404,0.3054767532067948,0.8773518823404519,
 0.2630636285168282,0.2616844049018937,0.019446954664449567,
 0.7340065183354578,0.9025960321801928,0.8666371650576293,0.5016377901581041,
 0.11163305405564006,0.5636693141006811,0.7761347966836264,0.3879504305345619,
 0.31462598654394625,0.5269903975995474,0.7285570903353769,0.8016577970017832,
 0.5778366980832169,0.9297757317811207,0.5575571752562388,0.5298323353702308,
 0.5236593595546284,0.7975922513409908,0.40275354853966183,0.8981170999017913,
 0.2347753109521562,0.4447319901083958,0.7869041006530249,0.9185690454391247,
 0.5235638222337035,0.3773990440560887,0.659889263218745,0.8260263279080539,
 0.7863805122455443,0.97317586371604,0.30343813179632717,0.5441490504533494,
 0.5666861779996131,0.9829737174658074,0.33157795573133986,0.9502366023174368,
 0.1497154033398146,0.8765480781298776,0.6158764288135513,0.10747753359237333,
 0.8489456735319474,0.8340005404051567,0.7418511366622087,0.7815932261711078,
 0.33209276877100036,0.47731741597082666,0.3057695527050043,
 0.7041537576629039,0.384530436853475,0.6126065487899446,0.1719329011462336,
 0.9625956483828493,0.8288870939541503,0.3311079536074225,
 0.0026486446453899415,0.2458178580300534,0.4085577984642662,
 0.29655433847132784,0.7349889420461754,0.21874398507521486,
 0.8240777572651795,0.7942023400957949,0.03388047763804558,0.2532990074970237,
 0.5525166296907177,0.970295009638709,0.34947680508004186,0.1767751827469588,
 0.5834425727939337,0.9954542914284322,0.09953269854579583,
 0.056208251323806246,0.5111430518596535,0.34561772285526116,
 0.9618973026954923,0.5029503139471175,0.08967166912826507,
 0.34287702343648263,0.9851977192114891,0.9194140509975439,
 0.34998026399633897,0.7195115005202019,0.7946220978343501,0.4102678734176559,
 0.23102374104795598,0.7804677531334648,0.7829435417757281,0.9700517491343893,
 0.016057777700803966,0.8948776534358838,0.791027412710663,0.3672460247733441,
 0.20951446614160907,0.33069016446891153,0.4720768494524481,
 0.4251176051032446,0.1013589504535124,0.9275190203912166,0.7018840024438597,
 0.01822898752428892,0.2634258329384328,0.25950594668505345,
 0.24338173044366063,0.18134186731692747,0.6096226892871603,
 0.3059693806574634,0.8908391590769952,0.42049965629176356,0.5898554287134612,
 0.8077980012223417,0.6937134930511136,0.18073391118925042,0.3865867816025177,
 0.8657168806154425,0.7105621989979888,0.6506798939497139,
 0.052908847259333425,0.01422793959863744,0.9764812566025869,
 0.043979449944707816,0.35795394783266876,0.2343724724743419,
 0.2941681742761121,0.6999869385723639,0.550482037275659,0.19262124716781492,
 0.29771458535166373,0.015142885066265155,0.19700056284816458,
 0.9678848992770592,0.7654215572065531,0.9546325525298174,0.50501372730544,
 0.789733459379411,0.4473832771158721,0.7111923024029515,0.2447014066915706,
 0.36599801894027995,0.32406092908808826,0.7083870633041321,
 0.9078651300906964,0.049357103198668484,0.6952637295884038,
 0.5408990679177637,0.6201641905933948,0.5351911358148621,0.34086137787611503,
 0.6991852044376048,0.49278833414837764,0.6290360354502789,
 0.34349917188080703,0.7595404904640378,0.9328595032138004,0.5845874718722711,
 0.25247820834162304,0.3267423592338763,0.5049484185854869,0.8734475999478943,
 0.13165176173078108,0.9608420914071047,0.5634305424993883,0.8022483593212347,
 0.6154964298233019,0.953133866800028,0.5425142420820992,0.5136649269118676,
 0.0965914156055695,0.6874411646196033,0.5723987695389539,0.27896059799625084,
 0.9764054397129878,0.7908627821495464,0.16556961646118695,0.5280428521938779,
 0.7847690020365834,0.2910755658172116,0.7387453699566323,0.6109358061470593,
 0.9310579238786568,0.34626471022110716,0.2896158139070635,0.3588741999959949,
 0.6964776466882279,0.5514202269428996,0.010882523571538405,
 0.15874257264019187,0.8538674981898114,0.3267428671735765,0.7872906073738953,
 0.7986863009615218,0.6973262077531773,0.06581741584328504,
 0.26431238821930336,0.5841446074422615,0.12146464364435461,
 0.22568587107053062,0.6922590586367661,0.13860685860596822,
 0.8544093812103077,0.33823682823141843,0.3352842657498958,0.8893791371728599,
 0.37430238897061285,0.9691142454089409,0.7094916379513775,
 0.24608546675235476,0.3323906222195816,0.936790719433478,0.20145846100784282,
 0.5391507603977094,0.999139792116245,0.671289763465511,0.9856881311834742,
 0.3649005580441633,0.9391371194108178,0.4258222256634504,0.7215073433349786,
 0.8123370853770479,0.6363847130251383,0.001257394390482447,
 0.7883271779068768,0.21525184365687589,0.930518559294711,0.6454518566170261,
 0.8490702578524046,0.24073105142294893,0.8701046120592825,0.9225660153426238,
 0.15272477706127596,0.20808340023212113,0.5312575307020271,
 0.6201358362909672,0.5116231560454472,0.052458963281226345,
 0.9374171744000661,0.2693108827838343,0.5725387832547596,0.07581621728399623,
 0.018658947067214937,0.5223089301906497,0.7600403661104567,
 0.4134633826781897,0.4636824724644829,0.5644072888334316,0.14200789168093897,
 0.3158967502759561,0.7137661531037949,0.9393220266169713,0.6252420488402761,
 0.44390743563557955,0.24766812552443174,0.3729847765213631,
 0.19134773302872654,0.3560883503453711,0.7277393617462917,0.2346849441761235,
 0.8357977158290906,0.1391015099916879,0.12332492571284859,
 0.011726948656480496,0.6525677821983251,0.43802070672247084,
 0.032421530928628206,0.07359132705245652,0.1933760015470496,
 0.9624536261875959,0.8889543209300553,0.6120371592292578,0.6579219116197211,
 0.17856792220846585,0.41488969048899393,0.6594272985747699,
 0.34003106471844846,0.6147519031228429,0.09337555041590129,
 0.9737126290265009,0.4144042224702815,0.7977128075992046,0.8047414278233981,
 0.317753144528794,0.47832392373474597,0.046984458595420886,
 0.01394934305765605,0.5684551655900226,0.2964380014393555,0.7769654782448753,
 0.01183715906720384,0.014840084236707685,0.7030892700042705,
 0.0985279373803647,0.7614001826247222,0.8046640685848776,0.05698558778643781,
 0.0954644605902838,0.07552692160908792,0.8982688503284225,
 0.11428524626622849,0.6059194143892835,0.23279391813359884,
 0.7726874191990901,0.3312443278578475,0.7317421640137871,0.887172804186302,
 0.3272690964327354,0.9967440656967095,0.08879482185766163,0.5755445288913217,
 0.17202064715274668,0.84018242952375,0.7997155175613837,0.32332530460359854,
 0.3608199006812002,0.6580678511253908,0.976904652045147,0.7858213858404799,
 0.6607392352590556,0.32038342583109125,0.5490594523682463,0.3953883994213734,
 0.8157698422997288,0.5318783744806814,0.8966881742616629,0.582882950932345,
 0.3720399474962859,0.9675762701078061,0.47119384624083116,0.5402629017188723,
 0.002159741612325461,0.4195546323946545,0.4798450494687072,
 0.33967662998635384,0.7138394704688892,0.1834000443322492,
 0.36660632024772655,0.6024120584048936,0.8977483503019931,0.8668378333078114,
 0.23068809009039593,0.39462886170999734,0.18048627272922102,
 0.08964083934856737,0.21416804045875926,0.9520631976829869,
 0.8402577454460805,0.1000953819767626,0.8292291860403613,0.3356185310418791,
 0.33741011407733446,0.4119705264318382,0.5307139305031281,
 0.31143350309300255,0.850152323297143,0.2778937087281631,0.5300100770999918,
 0.6384042519118014,0.4173726061608388,0.05207297833284952,0.9016203022568637,
 0.2861799208699828,0.8273530474572364,0.5196136620264143,0.9740039442666858,
 0.7241542972603369,0.9043872903944442,0.6366579821807413,0.24445107785914388,
 0.14972814106423527,0.8138031194160928,0.30314715775998546,
 0.9453053043058475,0.6913929378406344,0.9621365365503087,0.44358990508879015,
 0.03209094667751777,0.8353050575283916,0.6818415413017873,0.5204373490376838,
 0.9037626682435856,0.08623474548182863,0.5140458543940909,0.753556056400449,
 0.8903858633998731,0.7594746513921118,0.5447864735521265,0.2356732370906386,
 0.8347463016763887,0.18633357152485186,0.36036564603238797,0.461251237831507,
 0.5218219120373401,0.9198836981042883,0.8604512001770782,0.0407856053568334,
 0.19118400595065577,0.8696842306996501,0.6906885911208538,
 0.03586621036992821,0.3961078745063946,0.10881685814674003,
 0.9333720609257072,0.7267911566327229,0.9986786117578014,0.8933431355287813,
 0.8073397199204493,0.7104542839341055,0.5224090784913265,0.19226500971502292,
 0.7677285911975633,0.8826045145044746,0.7476493680891514,0.9000494724324002,
 0.24352180052289307,0.20525095510344138,0.9915587090554168,
 0.7739291400260382,0.8420743318788235,0.8415752980921437,0.27269066368168904,
 0.855068032751697,0.3234394427647028,0.631011407924507,0.4058591706542989,
 0.6283395228086288,0.8453841363798051,0.719749301957311,0.8579785217760957,
 0.9926613931885682,0.7299816935251783,0.29121384818762286,
 0.15665374434154256,0.8877693567986343,0.48960193093972193,
 0.8797455602648395,0.5737873758800428,0.4924153574464557,0.31188932007680736,
 0.9132930495753301,0.6406069457547221,0.36778960345640466,0.5439272140736409,
 0.2925001563358991,0.8197092403517635,0.24987961473076692,0.3759609322902957,
 0.31981923613999474,0.11673209594516543,0.0023213056940337706,
 0.7198617340730704,0.7088583271973441,0.33547383992995816,
 0.46207588737051164,0.6018206466141134,0.10821777616337702,
 0.7770551862081563,0.6195598373937848,0.03372806287363106,
 0.010800584089135645,0.662972295720997,0.09852166164194376,
 0.24708959546372355,0.03452727932856714,0.4494063337691996,
 0.07042486155476069,0.3674216391265651,0.2436831149729165,0.3055557912372454,
 0.6503556844892338,0.18314185893597212,0.6395368756944726,0.4512326156008345,
 0.9241373427659586,0.5411935922736998,0.2259538227597464,0.1827174147448155,
 0.3222023218109873,0.5700369471346677,0.4528704751909498,0.7547070316131883,
 0.0877719999050045,0.10780895309443639,0.051032188834386205,
 0.09322060704109758,0.6730477287292382,0.4698231340225836,0.5240235084375211,
 0.2507404762333595,0.11430666054085703,0.1009194894503832,0.8303266742239179,
 0.23652797277282245,0.5992214741472421,0.6119783406405478,0.2708530104263994,
 0.003176167187988521,0.6299195281797312,0.3010281582548999,
 0.6063986861904954,0.9892927830092654,0.9346573371573031,0.8870280727533211,
 0.8346275974258677,0.6981002320973441,0.4141807782676885,0.2681761614241146,
 0.1610508980026567,0.4162477445510212,0.36100248079625186,
 0.09214894185137956,0.11691317380987343,0.786353405302916,0.6169316225593796,
 0.6151134811997467,0.14079331505450754,0.15192277283939082,
 0.9389832736049216,0.9445495687546314,0.7802533580251253,0.1481562265405949,
 0.1826418798990337,0.6824542756516613,0.6878735140264463,0.4688189834541959,
 0.03559010967125453,0.2572467279647508,0.7584852142142017,
 0.019498896285171763,0.29381822418223935,0.0552772758659108,
 0.06080787169057955,0.322771104939903,0.11610868514838346,0.5094649152016824,
 0.3944469077768046,0.6311656260452349,0.17662350589224374,
 0.49100646594674824,0.821273646554308,0.7944986938665652,0.7537861040813132,
 0.7999588338752833,0.23540741757936867,0.06052980089950111,
 0.5275186514979037,0.544228553232388,0.9945501669766234,0.043412641585130696,
 0.7788468335939238,0.5571462032013894,0.29376812249215434,
 0.33524377460656096,0.45697354920855127,0.3663158035820725,
 0.7979526395912762,0.050621019771730946,0.6195447400155832,
 0.2661845801737145,0.32402084006648135,0.41785787066436564,0.293345528255754,
 0.5341651077547875,0.3687560037473724,0.8910893889261893,0.8167282046472989,
 0.8427032310919407,0.9725049291926923,0.7160504070059742,0.30850399327348765,
 0.4786385058257495,0.09641466286711331,0.15966559314412299,
 0.7602251019741147,0.8530511407638633,0.5628926931041801,0.3008531697225002,
 0.8599734035718236,0.732889216437476,0.9897949449015822,0.8607488269426211,
 0.9207540860105801,0.24682139254450108,0.30399742405162555,
 0.8839491779806699,0.6088231288802992,0.9161186080446466,0.07041959744523252,
 0.5749721429691068,0.7157027880836064,0.7347022025229294,0.581985752178832,
 0.7543188382589827,0.3851835125541433,0.6240778119863534,0.7647918630635308,
 0.8423955716832932,0.8957388798917875,0.521068720054777,0.23117378814329315,
 0.04468426919367374,0.44803446509373035,0.8968533274732934,
 0.22366277968696457,0.6625124704038476,0.9511256969587999,0.9590996238397711,
 0.8400537092158811,0.2890998177666162,0.32367530737029226,0.5372695562522645,
 0.1622995453125442,0.24865538658741326,0.6276471327410807,0.8866199478630218,
 0.8821701143663685,0.29441782060356814,0.5004125116761298,0.929545793412807,
 0.9268336650694436,0.5767513448416312,0.8675511629398897,0.6459112654673732,
 0.9666779020496632,0.7326595946221108,0.10582912718478166,0.7837345242902359,
 0.7549188035710439,0.7905288776695802,0.5849253613101959,0.5282005963377838,
 0.8587875601573632,0.673768017959648,0.12112148200297168,0.05856367490040504,
 0.5528837853634538,0.95759425458772,0.05050623776604468,0.534987351857144,
 0.12825406888543733,0.416913050638551,0.7828387033518164,0.5616668625686,
 0.5249626358001748,0.3941394439374287,0.3812690302914836,0.6333234359478096,
 0.5429193311311049,0.9172432252770982,0.48175038255757274,0.4342009949410287,
 0.6406610101024142,0.5743961800179124,0.3627732477550365,0.4329711013459451,
 0.493087538610371,0.8369630207889139,0.5906417406252678,0.33881960423907886,
 0.8843380525139921,0.5008808200051472,0.2751469514206649,0.5783879233466853,
 0.3479157935622037,0.35222274549672994,0.2223597487145863,0.6851343777611604,
 0.8173692183165518,0.20511211264824558,0.6442298651187874,
 0.03632425347212931,0.9912728869489865,0.46682097620959573,
 0.7670215719071412,0.7055202314507898,0.4385359368077585,0.3285929725928024,
 0.3935803616622059,0.8628452345686463,0.2679140039755,0.9600309940835734,
 0.7859461493379523,0.45685715462788834,0.45480996657404127,
 0.4161548531913599,0.626278026320326,0.7814432142001227,0.798657253601966,
 0.7522741847358999,0.6923928381465013,0.6319084562408773,0.07632329489613587,
 0.3463201259198513,0.09871056460389038,0.8991609835855828,0.5423317670431828,
 0.7906119660545996,0.5317613250199122,0.17009993791284728,0.9329172755237173,
 0.24319976282045463,0.9426399019531869,0.45623317623377146,
 0.6419993895560474,0.3623740884070933,0.08219732037103544,
 0.12755185065445218,0.7641290242294428,0.9406851746971545,0.5053806320796106,
 0.8366182250545318,0.26399591775315134,0.6207552375832386,0.5399528468635982,
 0.7386625717095004,0.9412356308859323,0.4654539545923677,
 0.0020183044302112263,0.5483931692321192,0.6862349897870094,
 0.5589838960021221,0.24960791599565868,0.26392098687646626,
 0.42685980469394136,0.866118674519854,0.4335006132482442,0.7636447498055807,
 0.5342875074168457,0.6066426396508641,0.8182741253107294,0.6127997983325459,
 0.848059485227894,0.7965908072255073,0.328479316664237,0.6730495703535626,
 0.2773369366516053,0.5191616484443724,0.16485395970358352,
 0.04070235344615736,0.3244166037901194,0.1316400803890604,0.0406345191492018,
 0.40748058275794485,0.43245346636426507,0.7667919074046783,
 0.9864718623453448,0.025976880279209746,0.05816746101289272,
 0.3338998736237708,0.8292570266814302,0.16246267960085536,0.9324165306067563,
 0.9630949950096845,0.8200399735353188,0.3642957731879498,0.9273181534548498,
 0.7931405277985842,0.8287770920714123,0.37280981100177335,
 0.12073678210218997,0.6178365162748135,0.11262534989796436,
 0.4767797461358796,0.07683182094595786,0.4975173022341952,0.779704136114333,
 0.44301103102843853,0.9499168619325816,0.7834654227873373,0.3340656933649717,
 0.4485913768485066,0.5745602482263732,0.3610089941840503,0.5022773639311293,
 0.09353625864985149,0.8782050312358104,0.18376175592981725,
 0.031122694198886358,0.8533907603535614,0.06646225467783184,
 0.42120699425889585,0.9094231091990683,0.352901284182502,0.8101792184329686,
 0.40071751504562647,0.9016111024952305,0.5722582728709891,0.5238765851716605,
 0.3696529443826698,0.34008127663983345,0.46790053715545854,
 0.9246541636223381,0.6171327355163292,0.3724470048173726,0.7681252858432321,
 0.48307201631907637,0.9828647024441272,0.16816953848688765,
 0.46354139773237324,0.495927872314587,0.5908888579145631,0.38277099971249295,
 0.49676598450625653,0.6204686163852833,0.8499024614474628,
 0.49577105601399896,0.48841690965928164,0.9594567161809534,
 0.13194885701294612,0.4203155249703372,0.31487525594129107,0.474713858287512,
 0.3787894707111057,0.6335366246193662,0.7554341601271329,0.24915062064496785,
 0.871901872750582,0.34044312674049204,0.04761347061578136,0.3929486522241056,
 0.697794618795358,0.6594143006440588,0.14001092480353938,
 0.013745781319158157,0.36268635438938435,0.02638633932360912,
 0.6899788095579535,0.9822550191108541,0.2742502611125168,0.2231399820851082,
 0.41300098603730917,0.2546207748453816,0.7239547290884469,0.2147158867828849,
 0.6819646234299691,0.49847192568020793,0.21819659506790234,
 0.2802170365261891,0.6537373645876647,0.24703811721746338,0.9421999420640197,
 0.5523230074884637,0.2688674474748548,0.49144381013413974,0.35750403036906,
 0.32761872094652134,0.9863163858082173,0.5971075442365387,
 0.17190117945753025,0.2588182540840078,0.9133414604843967,0.6190021877478384,
 0.4731465350774475,0.4949431732302967,0.743343336184453,0.9543719706991527,
 0.5771795936307298,0.5386872775549565,0.5000395106285315,0.9517613414118651,
 0.057816129179101505,0.7935928699313122,0.14980786777943267,
 0.9931869515406886,0.7137574208165061,0.3885939717270882,0.20367294413678183,
 0.978704248864715,0.016779079369892203,0.8692922465447623,
 0.013267457252872927,0.9185886826423847,0.699572734233692,
 0.04099680996108668,0.946552919935336,0.8669637742423049,0.2916319743484507,
 0.26079733929175286,0.7952894770693582,0.026388521257460607,
 0.5071178421327546,0.8229375603727223,0.6366603502743114,0.769841289265325,
 0.7446728806048597,0.4320937102937976,0.7365619785961663,0.9050527964212299,
 0.7242533884225117,0.09957792738267046,0.5959843233217771,0.6821604738663203,
 0.4484388096030435,0.36033893662177274,0.20296777337065897,
 0.9083252624304878,0.4620775178018577,0.7342592527441167,0.24376599706413238,
 0.8615524044381229,0.20946333791415683,0.011917731670037734,
 0.12233480437702737,0.10126754264140125,0.8688064590197581,
 0.9084535035231812,0.5608774917727646,0.23348903201332571,
 0.08843665355245267,0.6738599398881323,0.18881012951601817,
 0.27870067306926316,0.684462736413282,0.6884883842779935,0.311239744994145,
 0.04615728753116599,0.5718833954577177,0.02611029391790698,
 0.10985883494129633,0.3818182591894397,0.5783855755657262,
 0.12618537349549097,0.23104929461323687,0.39627195542982363,
 0.08674407655966498,0.2666590396294235,0.8329928966728108,
 0.41467995270732816,0.12487840426835306,0.07766228658183372,
 0.2689040010244865,0.07147386427690383,0.9401587410676956,
 0.09547173871972692,0.30524615905309915,0.9398253062009777,
 0.7631793450711932,0.5995684741851011,0.0636456842674471,0.5755827029689906,
 0.5067523755674893,0.7550702547991585,0.8541761571956628,0.6313288738836759,
 0.9854680174725086,0.09294731306440918,0.7726267523821031,
 0.08030658587096695,0.2509112902266496,0.28989201866875425,
 0.2610227282910086,0.8694320576029897,0.15358846888619193,0.6535980141995096,
 0.8124626248838953,0.6450306221158483,0.3327623346778179,0.5989449342222761,
 0.33796336337916866,0.7613316307223282,0.9652281364264627,0.5237895885766897,
 0.054549284010708354,0.8293154462099406,0.3948279904235352,
 0.8648122026208725,0.753264388491818,0.821010061887791,0.7628037226605074,
 0.9208591583871978,0.2369344254770297,0.4376284015503118,0.9210426713090716,
 0.969998156807066,0.21649044639454118,0.5489723620391022,0.40840682813260476,
 0.14112553409454798,0.6828655943379827,0.332550720178474,0.14841014497090477,
 0.30315543416501134,0.16749883377124886,0.8389701227494718,
 0.5282543660135186,0.7362183777122432,0.8937209985771113,0.16202102773366156,
 0.3022070359322019,0.5574984064040621,0.973728365582699,0.834860750912741,
 0.7281658149350289,0.9483189683093596,0.39924816557852594,
 0.33858632479485773,0.1967605286453903,0.29688516038169366,
 0.2217373700276185,0.6837331983436082,0.9697655900939937,0.08179634608653852,
 0.5522352813029194,0.08698771182648879,0.7788143179904615,0.1962167311953259,
 0.16277899697098053,0.03260633747524744,0.7837033210282484,
 0.33840624881046655,0.007272150070739225,0.3135105447227535,
 0.6535560594653489,0.008079464249340584,0.8600976737689808,
 0.08514626444507523,0.024599634463542963,0.9756872973621887,
 0.23012241057710525,0.29306839752743297,0.4051467953598914,
 0.6159970689109989,0.8091632875128068,0.1571595388159509,0.2935710317808161,
 0.6903796247537559,0.5796580465170069,0.9453025211635592,0.942032959353615,
 0.22185592922813668,0.7998356935217622,0.6158401407527249,0.8909012818115692,
 0.2954519461344459,0.448806357469441,0.546645123108082,0.4303007136254037,
 0.630535336751087,0.5916470049215569,0.7917241062952782,0.33510183080886,
 0.585005630341175,0.30702885094803933,0.8423803180352132,0.46068536520706627,
 0.8913100411755911,0.13707626400843465,0.6714171627878017,0.6311611204787798,
 0.1058892869149376,0.1616762852364968,0.5831519597838449,
 0.062311944474169434,0.7915263794286331,0.9601888507967529,
 0.09426184672174442,0.7003822772004887,0.9153428265722645,0.1998718445175539,
 0.7184908103621288,0.5733542938818124,0.736974816276728,0.7972792206889074,
 0.9704601293519942,0.906496585079576,0.7415162000204129,0.13282726817874946,
 0.8630891256161844,0.6772294745428337,0.17785626573297697,0.7801749950349348,
 0.1597208765143362,0.44849191801483934,0.44123798719177687,
 0.30797093817417454,0.9660943860476531,0.2522956534338695,0.8699525052280319,
 0.05648081487950707,0.1351646258262853,0.8075802091896069,
 0.048923699230531414,0.7738121531310165,0.005355461482465795,
 0.710420677558565,0.924336090400058,0.021964910822965367,0.9171610521911969,
 0.5827902514357731,0.458649321833395,0.1545375658486392,0.030932599033275388,
 0.35056695294011253,0.5940645445774959,0.7777369103312861,0.8247138666433006,
 0.748669601492042,0.9711224579972371,0.9756899819127393,0.34648627139976274,
 0.5500958809007448,0.7114095754129985,0.49305762564550226,0.5787608438204118,
 0.23761873195817917,0.47967354491322045,0.12979343504090546,
 0.6124838267381861,0.6998611852807758,0.9113522566905605,
 0.010018108654206714,0.761482253402595,0.3193902425251728,0.617078613073618,
 0.06947559885824284,0.6862954841418963,0.4943712969202626,0.4520722928954789,
 0.4090697985008287,0.432659318992194,0.2694221474745214,0.6964147120893237,
 0.02117745107787883,0.5726583473605984,0.1864600743744509,0.1358675791980143,
 0.7843677239561726,0.9374917969327675,0.3016186602037989,0.6068482821603159,
 0.4962575698564867,0.4887391755192684,0.7982593921851812,0.03496142732228913,
 0.23759318771131044,0.46302445058116637,0.6345887601605781,
 0.5953428000882481,0.788154669173588,0.5343826479892256,0.5169927173314235,
 0.40497881668575075,0.7496976604921564,0.955943216347909,0.6405845145466094,
 0.5028523453062018,0.07960253583633237,0.8240751206535054,0.3519582625113522,
 0.7711959029486578,0.9291465574145671,0.4239662680709122,0.46785641950615486,
 0.4724710177000684,0.17766102434956943,0.22129508193047598,
 0.3512408176172729,0.7611403300693125,0.5039026349159464,0.5907564149799538,
 0.15507655893784844,0.060724158712280096,0.5526828748491605,
 0.9665233052404619,0.20895066552185082,0.732117927726708,0.10977074883983007,
 0.190484016869496,0.19771245346111743,0.415978929759781,0.9921740389991754,
 0.5107627408095581,0.847170335498929,0.21067533832868934,0.6396909205410806,
 0.5212692638541698,0.4159342381792235,0.4690665538838711,0.35478914259240013,
 0.7836794075263784,0.3645192008913031,0.5484383014508453,0.5841807934607226,
 0.44161605012827765,0.3138342957604473,0.33765526579851146,
 0.6568190302235909,0.09984690782810113,0.1548990967086714,0.7314721451852776,
 0.04743713724564913,0.5750909080772448,0.5200846484401227,
 0.02621172738826172,0.1208741506345622,0.9276213062988573,0.4331600009023784,
 0.19966228839615285,0.8250001379191203,0.8175334520377273,0.2531132005277783,
 0.18825513835282337,0.5777913634974544,0.5857834079826122,0.6244886740711687,
 0.7472141081716872,0.5019590279139896,0.8399387882071003,0.7047916748137178,
 0.757950077314631,0.21213010702673152,0.9574652697406805,0.8687899771186142,
 0.4735302229024212,0.8091059354956691,0.4483056713410053,0.9272247826418683,
 0.5863412008431731,0.06374713736265702,0.8503619712314918,0.8067934610667786,
 0.5113860681314194,0.40478326320236746,0.3183359161348309,0.8578159611016369,
 0.6043557700513623,0.2793558638524225,0.3309694388040081,0.11563944549136307,
 0.7970055680628596,0.8150112271499699,0.41403945486411875,
 0.21238720866406802,0.8038636273500009,0.5469430789254233,0.1313753451016173,
 0.4508515121873896,0.1391125581096252,0.8093854913614378,0.9527588403684966,
 0.030023750650741365,0.32280336678965893,0.02438554745885857,
 0.5943498137442681,0.06066288363831718,0.9198097523183729,
 0.30811226504499123,0.9980847001895882,0.9454254598761023,
 0.48915845565833016,0.5874902934488567,0.9877272653357821,
 0.33956960709134254,0.6719941882592134,0.043824605816840556,
 0.4358868652529817,0.6743689392022236,0.3269680270587998,0.5093533417141163,
 0.923827062728415,0.21446086885360072,0.6901957732539549,0.9234311074157397,
 0.4664878103528415,0.24533985396032687,0.8351719453285191,0.3529283838586712,
 0.5683153749674403,0.39076207703971577,0.6661798469081806,
 0.028587723836070444,0.26801183560651454,0.40012671521170906,
 0.9445806577174242,0.1427088716728414,0.8795995093776586,0.11652167826061333,
 0.012016302413472735,0.26573745555733874,0.688928695616721,
 0.7491771585445371,0.3178831836433469,0.19015431770876734,0.9800368000726178,
 0.798318086231764,0.2844729794087695,0.9713278978340775,0.3009440302118369,
 0.9117377512383764,0.5486531792307288,0.7327359576209002,0.16786778676742475,
 0.44262923403540244,0.9399860073449733,0.9012089231656402,0.5068099008187309,
 0.04333026674323326,0.10718877156898055,0.17843962416940506,
 0.7112559856931472,0.9966336329385682,0.8139375495893375,0.7578284436927997,
 0.30729837358180934,0.5905950011738966,0.6562092556616878,0.7865522390338195,
 0.6934846608172384,0.7379776598296578,0.46309165266389163,0.7416584269065801,
 0.6138947594364716,0.3793355493963979,0.8993508147195951,0.5345271218235572,
 0.5573276978127816,0.25572389584714805,0.20870356537874546,
 0.6792300735558515,0.7393349282988145,0.8980872479257066,0.549004556927748,
 0.4669313265560865,0.5276799735379861,0.055159181062697904,
 0.4593081956820678,0.024951542111589897,0.11104967190965132,
 0.16446638585142737,0.40287389518004946,0.38914017491198116,
 0.3068246147565359,0.053647056628108314,0.6388029725228282,
 0.7648266147363346,0.2508410041302407,0.2561636017123745,0.6752000127284432,
 0.02593630777146938,0.8161892006515978,0.7357182522256381,
 0.13651568645084933,0.8942623147355501,0.9944806059717919,0.5616672624118253,
 0.9065697059880115,0.9459636036355488,0.5283519030375158,0.0691760341206491,
 0.03754190042989847,0.44526782948288757,0.35542326611384933,
 0.5604832996471586,0.8784658186445081,0.9897968929861702,0.8829745354661027,
 0.25200430339685775,0.1884366004024437,0.7177588493292304,0.9163396388672891,
 0.8612364993764691,0.10349859412434825,0.8319385820507426,0.9045386782533823,
 0.6594926996661084,0.46384181793352475,0.09376470416825256,
 0.4274185129267545,0.05064461503934381,0.13838210886549906,
 0.7727818462778283,0.3731342696839258,0.9422285425430834,0.2698959632686213,
 0.4825520914180502,0.21336638607594627,0.5786502861868348,0.5503744717502238,
 0.9006094646031966,0.9767228513500504,0.1294258975985566,0.7111772129383029,
 0.3268754248619239,0.7784855674925334,0.4673417050216714,0.9803935441160025,
 0.6323614006836247,0.9468472382150476,0.8977942410092115,0.914869747614279,
 0.8614076825898329,0.47774557794718653,0.4663681188703206,0.3457711817743162,
 0.3375156238959496,0.9782444879038514,0.2953977472619461,0.427475439976126,
 0.390345291709533,0.7204792656233073,0.24586260136962468,0.15391058547357472,
 0.7736769293638114,0.17677310706206062,0.09591844393112348,
 0.13694892027694827,0.12634468360116113,0.9012309448381345,
 0.8950370287889686,0.9843175179517538,0.4340912329474902,0.22472385766666225,
 0.795754894087383,0.9082923202910216,0.6221454536831558,0.4008301550569272,
 0.5533579501511883,0.8695426028935804,0.5961123397127455,0.7415283729483143,
 0.9870449433255857,0.9262860709980408,0.5317901129863031,0.8714157188281767,
 0.06446471478281701,0.4810884666653583,0.11968801160847298,
 0.2855021831611231,0.3064922765395415,0.6338988090162858,0.5056845578825173,
 0.8014889443242189,0.009331439547577425,0.8701286130486937,
 0.6716647454633318,0.8080577705022566,0.06482694436968806,
 0.44372214014823663,0.953426195298603,0.13689146099801364,
 0.04353151631622243,0.385181772998069,0.5168388742070882,0.32134906700144394,
 0.6576870440994834,0.978378337263838,0.7446173231846942,0.17203948663170787,
 0.8914999852347152,0.5717656790672303,0.9359478095105195,0.45254420739987744,
 0.48263513869380703,0.03960662759033451,0.3633506106936196,
 0.8972669738401835,0.32384356875837317,0.18881090687865654,
 0.8182396703315106,0.4074240512877646,0.9218199201991686,0.8383663659218032,
 0.47970079830998025,0.1285103470352814,0.506403995833268,0.6380239730995072,
 0.8917711886881062,0.31891254672018454,0.9281248626725126,0.7777263555223,
 0.1713858544625485,0.49764355890686085,0.30610055275804027,
 0.41461503896257557,0.4571169329886927,0.6462115065213482,0.8805591956355368,
 0.5176961126606205,0.8569616874783118,0.4113506703039508,0.2560687419653035,
 0.7801804175180708,0.835361123930661,0.3289090644830296,0.3223161012868946,
 0.28765045495379893,0.4350247684121945,0.5569547342793367,
 0.14675678714437845,0.06765050764603908,0.32130000016297255,
 0.3487700044890383,0.3419897213969577,0.9527000593725833,0.9716800425786125,
 0.4767411797733597,0.3173396733909123,0.2224650355169715,0.6538225755853251,
 0.19380827337261508,0.11900904317646144,0.1764879469732018,
 0.11532924961865199,0.5630578910542088,0.6370055011052906,0.6485959588513714,
 0.7458958568748395,0.3567184839349562,0.889464482226899,0.03636644552816348,
 0.4858820298153069,0.5354584786938619,0.9390557788403626,0.943560229540455,
 0.28449673298968126,0.6676408500014341,0.3072169877098907,0.9753679488100778,
 0.507784565156186,0.5054996926107365,0.8799790304346886,0.3623877568638574,
 0.13873263878150388,0.2666614840562894,0.16668723869426783,
 0.48925530337961054,0.06436985136774909,0.43331232569478273,
 0.37955362550994765,0.7927056929785803,0.6345495427647614,0.5979551050160095,
 0.5210216576321723,0.10600759144034,0.28384490157670017,0.08962181253080104,
 0.312414641595652,0.5157159076639629,0.8357202660758345,0.05052711004022259,
 0.5518835800947275,0.554326960238142,0.443485654883359,0.5587035558483818,
 0.966265660327366,0.3964039661261519,0.05618820193438501,0.48941526639262356,
 0.27683844696018656,0.972355203095796,0.2681611657276126,0.5603601419120045,
 0.3516936793134653,0.05353538330817953,0.7230595959462296,
 0.11432919534703212,0.1340177442046805,0.31518500655745385,
 0.6525402440987365,0.36173699434521955,0.1354443337155503,
 0.48645505001062106,0.024701331951672567,0.14544332734454435,
 0.3402869528325928,0.03842780575385907,0.13706531406339129,
 0.4074311394362564,0.23881716736445413,0.6611646415201134,0.8882245074620302,
 0.28058407718684075,0.9031859346661649,0.8159518789920182,0.6214568563074593,
 0.6435983894099534,0.7591331393417666,0.5381907971376462,0.3524036274042317,
 0.18541793332537626,0.5627965892629437,0.3786122651479581,0.7109218944935327,
 0.023804293434207535,0.13641052705656176,0.1701752768708673,
 0.7777910903520217,0.9592658815788391,0.2802630539527825,0.6121330693123449,
 0.39046388495863216,0.36697021072960057,0.3437055637455455,
 0.5180188995764221,0.8983221399821165,0.6522067324453381,0.4533861314102626,
 0.9452257359616557,0.3220818642239187,0.5293179242153155,0.6140586166223465,
 0.9018305604700262,0.6135937821618973,0.9437451612598147,0.298590732996862,
 0.2567917394666688,0.5415133105811799,0.9310207857834671,0.7364601477326648,
 0.7114034118781589,0.6428994102316706,0.8800778239313664,0.6317960283822553,
 0.5330701773026971,0.8803954497733467,0.9117308424649699,
 0.007147625271795577,0.704551427119678,0.8570048874383961,0.9616273703363959,
 0.8507752552111021,0.592957873264258,0.7394218776092504,0.7477725867806482,
 0.9080848521980591,0.49957793456787736,0.8443285368075478,0.3508725305908763,
 0.6316528815797092,0.8491573896594274,0.7066309840248658,0.25668313122083386,
 0.2694619636131238,0.009681592891155444,0.12111724615835062,
 0.9320159018349181,0.7655093549362002,0.9037206524640111,0.4718976392478337,
 0.16997696577137023,0.10919333403062803,0.2601927654130488,
 0.09080936026054509,0.9220284587612151,0.3863061446363375,
 0.30638533798093603,0.9550546579518759,0.5974104668878937,0.5678615995231201,
 0.8836931569272038,0.25479298306888576,0.9952394146394046,0.5720074973261658,
 0.4834006597648425,0.4092014199337557,0.834819610733647,0.8926610331900201,
 0.6859358142681132,0.0957129166392785,0.2681569898373989,0.6539880678616745,
 0.04983680982890415,0.8951639703498085,0.014574532602492751,
 0.22458759399009842,0.12529303598602537,0.9848679091891334,0.324418801820077,
 0.11746899180666193,0.6348812874459235,0.8415606683374548,0.8255407292007045,
 0.9712995540390404,0.5526432441003408,0.8396964095445869,0.15194197706303492,
 0.5789722017388896,0.33648332069582965,0.417019048670513,0.8040330403496931,
 0.21372397053347103,0.7795904326860482,0.9089166369302647,
 0.27101154633002333,0.10857295965842717,0.42712669923588686,
 0.6031352437507965,0.3877990628433581,0.9440518217377746,0.5573532094339607,
 0.481531787723948,0.6816933034028138,0.7757328477137837,0.5418300978768407,
 0.6211359543686934,0.8012098376430937,0.8546164809794451,0.6666583423058227,
 0.9076546573593881,0.16374818236137723,0.9673050513607447,0.9042228547797371,
 0.45927610659130824,0.6516087194219852,0.29202062484357916,
 0.2659140001048049,0.7867831700010467,0.8975675074083359,0.5438227236801594,
 0.5843969128047183,0.7192252605540483,0.24790087986612552,0.784083891453026,
 0.3691288544603526,0.5295563354967474,0.529452993199463,0.02903908747429551,
 0.05739888090119705,0.4862622878478925,0.7744845154567916,0.5815379883538344,
 0.6648333596811519,0.6763611450731474,0.756911318252714,0.05821883887434731,
 0.7841602064530772,0.9039492148649254,0.4342014226482753,0.8441258259349141,
 0.3650050419809867,0.21131705954610314,0.8438825465019238,0.3683629721208792,
 0.08435402283721494,0.04217746635475916,0.8850172758918681,
 0.19262213146649954,0.9469336281581832,0.4939084227531171,0.7438994536399005,
 0.15823325931797783,0.5767576455770049,0.22677043211088377,
 0.34775858572288654,0.9767214683698213,0.287116748854769,0.5741903219734803,
 0.8828783131300914,0.29965229629125867,0.9036621397259703,0.6455109062005149,
 0.04015941470366424,0.7510910722670255,0.9734668281753261,0.552845305076664,
 0.23586003546122614,0.14239434991117317,0.6157973919704746,
 0.40389195578304926,0.3052213236949449,0.77494228741242,0.07840477870560536,
 0.24402478006154182,0.024985869159660057,0.4160507187347312,
 0.1032168193115538,0.6165871416985225,0.45552281229931035,0.5982404182248003,
 0.9172497480164531,0.27365305808505724,0.35401167394359223,
 0.9214144821717826,0.25963546218247935,0.0450703958195825,
 0.12983828146517817,0.70603417456832,0.741500291685251,0.9253328118959215,
 0.7558508744810778,0.3794797158456209,0.010060166002769155,
 0.5891639216799266,0.9213702295483202,0.1121622972277263,0.8977291493151319,
 0.9846531667283817,0.27147469295218796,0.33277908117876354,
 0.9687623873510878,0.9504818707808307,0.8246259782386286,0.879298504480272,
 0.31671896549633094,0.7333909323732721,0.0777014812092558,0.9508459927846233,
 0.9096038656456914,0.7119510190251571,0.8294624987774162,0.33547849326555523,
 0.8619385342302951,0.1421050547851066,0.6655259922652876,0.8212028307425583,
 0.7611286500545151,0.23172985576077898,0.7831243774402316,
 0.29514182371417397,0.8101157430331664,0.03238265897932924,
 0.28642230830981874,0.34179097854508333,0.8704764914218595,
 0.9869555727566937,0.5536383287324542,0.09294314288587968,
 0.37599971178900027,0.7764093167088011,0.38307076615258895,
 0.4060438292947798,0.8768675761210039,0.014276295275620243,
 0.2275819299800671,0.8789503433846505,0.6471715802694764,0.7393002032890921,
 0.16801737389722649,0.8659447353893305,0.7141873309427407,0.3013309686673714,
 0.8567140671934454,0.16733765285017788,0.025266915679286517,
 0.4787089585096296,0.2831866940933353,0.7474739481620044,0.44043944756409437,
 0.2510097239553126,0.9732582942076782,0.919358938933713,0.20514513389696876,
 0.1259035471039689,0.6989479905036018,0.9017545652323057,0.701977375844315,
 0.9650519221178542,0.37325387172615676,0.5964180179416965,0.622369143046743,
 0.5180056623648365,0.7120970185018572,0.17161959115540937,0.8669968373825148,
 0.40838560572276805,0.7450608939095861,0.5666967099642249,
 0.42891857916388976,0.8306339636606077,0.3873337346162309,0.2030457161570972,
 0.22544285816692322,0.49040149847408354,0.36478515518896615,
 0.16658570538783812,0.08529313870648136,0.20711836432123898,
 0.2652688666987275,0.3774988709439845,0.2656712612756119,0.9523908988327241,
 0.29932451805233296,0.679236973645927,0.9632255442553188,0.9739341540799714,
 0.7457641135317662,0.4353564514244044,0.9636997345058483,0.5284984359220836,
 0.056899249305971344,0.6670716745876408,0.07211688934948768,
 0.5506356678368524,0.8118036823814296,0.3133965680852141,0.6721800882599446,
 0.11365771507561262,0.03520447517649372,0.08552272154866203,
 0.6171573130288313,0.9194826274964228,0.6698814214568335,0.9172788142195318,
 0.2563135915798668,0.1332704271694234,0.29389175710361737,
 0.19688464261356697,0.0789109118799497,0.6189114967600126,0.1596914369489445,
 0.4000923557507622,0.9532479463283223,0.14154033462595106,0.5298461282171139,
 0.6142867335104303,0.4856342225803304,0.05843553606300045,0.8282903727079005,
 0.5223900962886681,0.3213440504025046,0.2664626606371773,0.1689784729570224,
 0.1976891376053278,0.09240282378823639,0.15603696846623205,
 0.4689702297470676,0.1338267795424588,0.16249244001683671,0.8481611396484733,
 0.5811514304415757,0.18878026163631012,0.4900541328388939,
 0.043348373380076355,0.24847890401164952,0.6479159386985724,
 0.4827487779634748,0.5533161377841608,0.3780423282480233,0.5830003050117685,
 0.7176419754764377,0.37527488813885324,0.36485734440501094,0.276175918807711,
 0.3305809273728667,0.5361618802643742,0.7381562938202241,0.26682929664964117,
 0.9997906165891851,0.6144392094658829,0.6128227506222408,0.6545362719831133,
 0.6943412589238296,0.09314298184923153,0.0024237960356049415,
 0.394843782040027,0.6526766732772875,0.03288601189068596,0.8284263544709437,
 0.6170103145367077,0.7275304263138614,0.040598515790259415,
 0.7461043315388509,0.4883700792318504,0.8269901201926619,0.5682321621395685,
 0.5063842580666589,0.4481398611862467,0.2870765804604817,0.8730473134193663,
 0.75106596145178,0.7376185291864938,0.31393297338469817,0.8017557264716928,
 0.4747283189480881,0.22766302168898078,0.5020194576409178,0.658930114841998,
 0.46774330032967115,0.1880344296132639,0.7221965900601716,0.2791369210462421,
 0.9345186015879706,0.3427947706637279,0.9608925904991488,0.15096068220921455,
 0.3592068300384659,0.8584333107887117,0.02563312250901939,0.9321840779519541,
 0.6448625195393879,0.5918248029787068,0.8328270002097523,0.38401192681137697,
 0.17011762602527958,0.474864950352365,0.8291649315536246,0.8417415519232414,
 0.43958844571299527,0.21226709585632564,0.31422182814647204,
 0.0671158628183195,0.8056874578858587,0.7344815908676835,0.3455745695449851,
 0.445697665890327,0.366441218591557,0.5557078868479074,0.3673808443519144,
 0.17461375386011413,0.06776098199772262,0.026830413096858297,
 0.18691797003228572,0.800917229827665,0.08975870499439809,0.1633474916555948,
 0.536920910271456,0.5672125416501935,0.4166681074932528,0.9125206605825859,
 0.8899887974681557,0.3038871127784638,0.10497126699969306,0.9290152568155818,
 0.8261206584187517,0.5957530136119529,0.9971582173068758,0.6880522197816179,
 0.038913343858199334,0.8916850621802997,0.9430108420289564,
 0.15115793260910182,0.13678642011181852,0.6881843275170716,
 0.07690011198396562,0.5895196766087708,0.7881127957398714,0.9114443863654593,
 0.5657636667353132,0.03318155951504309,0.005273098572282198,
 0.7542997774171616,0.7696535448455011,0.5950423001714636,0.832164608045519,
 0.2421681388948198,0.8781314889680888,0.12996337315974893,0.7069691698064804,
 0.1295022701750197,0.25682875202725075,0.8168731860665525,0.7038239112951464,
 0.21569969252714916,0.26464951985697005,0.9571421263048606,
 0.33634400250584495,0.4985162142461358,0.23835411505388704,
 0.5035501230757475,0.3549623609875425,0.48768083660003436,0.2515882986123762,
 0.780392099120459,0.9998574610260428,0.2929891957515276,0.6655233989033201,
 0.9844367995081008,0.24873889779678682,0.3982251060597227,
 0.07030428777031217,0.541974809698456,0.856253389020077,0.25242623153005883,
 0.9270815459139033,0.3518619494001587,0.7118965582231903,0.03877661604529559,
 0.7871915513772498,0.7340474843325457,0.21588839209991018,
 0.14162700666950112,0.10900515490902196,0.03485185016998582,
 0.5397123738803077,0.122617328536202,0.16037841743454384,0.4579967843869145,
 0.6132271993504034,0.6060868737140515,0.9157693201210485,0.750385395798032,
 0.4472778852556405,0.4551254083917089,0.5323549107415041,0.5078363093276896,
 0.8167684052659305,0.9798467612950713,0.5034703790199183,0.05291587418222343,
 0.06298766397838562,0.6885515707417111,0.7332854288502431,0.890875265348892,
 0.07963642124638193,0.8937065362602746,0.5868049569739695,0.96278747063002,
 0.29275168854826483,0.4108820140719801,0.751322272011268,0.30539461989676064,
 0.9075354029428696,0.5423576734492325,0.10338323687981232,0.8263338946540761,
 0.5189442150010454,0.2459064648880488,0.901452699059312,0.5763739132113721,
 0.7548258916177188,0.7565378652154829,0.08773058794095345,
 0.005773521155219119,0.2923356403168166,0.38459250227664943,
 0.4302704687191091,0.1649711115726974,0.05278145831640568,0.6937489753184045,
 0.5531216526228647,0.6045197538901352,0.5558714893665614,0.849849189385576,
 0.9749262110481566,0.34094560846973954,0.4722473787560644,0.313107301774772,
 0.44431074233682477,0.5306411373440594,0.6924151532715584,0.8027387324578822,
 0.1815289491354728,0.15598565478707338,0.2215781582767472,
 0.09738195907955915,0.8682197592693458,0.6539357837263979,
 0.46618359136588683,0.9412640745436697,0.1410979006618157,
 0.13022922233997902,0.47666063547313753,0.4446840258977237,
 0.9384016737153463,0.8491862692815645,0.6569151681437997,0.40193232262653344,
 0.5405120763480866,0.6539841333737209,0.5072623514883874,0.07674280290033686,
 0.5499486742705332,0.16307783715286006,0.7678683288401702,
 0.07756332979041636,0.7168838188080658,0.7678287968020333,0.3287459292885374,
 0.6657169343570658,0.5750203672609122,0.7500876897558655,0.17564823360232573,
 0.5360437551504442,0.49884668775076146,0.43016864226110874,
 0.015008008004362527,0.19128736475349262,0.7448000002112376,
 0.879253825600237,0.4822305206919182,0.5156813144199495,0.06476361358861527,
 0.7209173256363025,0.3779146593132652,0.44938671157208687,0.5753269987643785,
 0.3108528610456476,0.918328957066921,0.079041648951932,0.11160493123445758,
 0.19133502472082475,0.3633478488423928,0.32906792293813014,
 0.17617475847360398,0.6772858350099156,0.1786393883639742,0.7184094580520818,
 0.7401983342919226,0.7254171484853826,0.7219253301039009,0.57827090492445,
 0.9609490111985329,0.6748374744704954,0.4311906681274499,0.8974219110957989,
 0.8604997819969715,0.990352963862049,0.6756367749074074,0.8486618693748803,
 0.9089038197843398,0.3026790872459034,0.21064618562411497,
 0.05919875384797846,0.824883139129504,0.9772301297410185,0.6507326913920157,
 0.8029933613188165,0.23287726179133394,0.3521526690649237,0.1228077041684611,
 0.4930490165370136,0.03112447059750134,0.3529256840038527,0.7369225697214673,
 0.32900371119737426,0.06112852759248055,0.0010669247273547011,
 0.9824320740508663,0.5120022378220096,0.9789883036420284,0.2571728881090014,
 0.16138632156551103,0.4149126712960749,0.23782749591443508,
 0.7052971471064395,0.1383910435202067,0.412629313982367,0.6621754337833501,
 0.5986541971521068,0.8573033219090825,0.40922190042136164,0.9107325894716318,
 0.5820677704909595,0.19143967451799648,0.5256777064119338,0.7230110865406468,
 0.10278407538222734,0.047975373583483694,0.31409030139262684,
 0.3191968287223216,0.03168372056333324,0.09623193148289544,
 0.25013515987696144,0.8152817231106049,0.052403730171943996,
 0.06455227459483293,0.9287444938457469,0.029028121309507715,
 0.07548479612997205,0.31174297205436474,0.6526443701058353,
 0.30232038200493216,0.6235337790841871,0.8693050461822662,0.7979871828551258,
 0.6525617700564911,0.4399514752750421,0.582321593765087,0.8465421923558676,
 0.6802866049346699,0.9651897773799805,0.8710246202179174,0.2512412770047211,
 0.44532865297642654,0.5289956683289512,0.08337191416690826,
 0.03904541253892346,0.9800275626619224,0.592091780150064,0.44990279777868314,
 0.8024617766468585,0.38572424346693757,0.7635327246005073,
 0.41399359529698554,0.8276544236595945,0.67953888511029,0.17230831800931634,
 0.4067890511837785,0.6735208198860722,0.8569819779696515,0.5834178141805351,
 0.985599725473796,0.9266352276473682,0.924764797312158,0.6995511867083092,
 0.8350079292115153,0.20373857795327288,0.33338560603356926,0.934526316222062,
 0.7112029728267439,0.5026915680061499,0.8360928198912811,0.01878709117686672,
 0.4649795862433842,0.7504517942236019,0.5397526098539946,0.9558126592278663,
 0.6256324027110803,0.19568771893452197,0.575163580903369,0.5538111768747733,
 0.09766998786289971,0.6183333590370591,0.70296555439132,0.41393753880568407,
 0.7907058284922399,0.6000938470118453,0.42450618007125707,
 0.004160140508473353,0.43287648045702976,0.20668673202297316,
 0.34460444851287675,0.25055041154605984,0.6071942070416172,
 0.1877113609961456,0.6809075193108269,0.5913086458752956,0.14482217841757683,
 0.9328036069911589,0.37159019138023375,0.4111501660705099,0.3273984665571653,
 0.8790933113808256,0.42023658160721566,0.9372105278386935,0.3247161356257504,
 0.22783393695314058,0.22024019839536302,0.7244459705501383,
 0.8542899815049727,0.6369423472301765,0.7102007209231677,0.485232104405405,
 0.2941144088697444,0.024578608182788453,0.8372832682973202,
 0.47977193087285586,0.335278537949803,0.12551470296875877,0.772825526708652,
 0.8876868498922967,0.17410451674335004,0.6657572290907862,0.7497525991262128,
 0.30019281689733124,0.49962900998445503,0.29471249677877687,
 0.47469125203327855,0.5634116317011062,0.8111710583542041,
 0.39630967581085264,0.36130445691793134,0.616844769055868,0.4055077337502525,
 0.199792795413946,0.10113990318530286,6.126526678521538e-4,
 0.6524546984368339,0.15251801695583833,0.573456579036878,0.2500754713992327,
 0.2959026549668441,0.18096801175277233,0.8919698324727345,0.5269412653333038,
 0.3074905020337031,0.0463357692888845,0.017975573336356354,
 0.18673954153082017,0.5642489527052329,0.20509265493985795,
 0.026768725364681778,0.21107888832216704,0.03037047271608806,
 0.11164696972874866,0.35466479498171954,0.8893317054600549,
 0.07530283663579829,0.8481888093678089,0.1060055612813462,0.9645288472099387,
 0.6287363244464934,0.7558052470414405,0.421266231030329,0.766197341638051,
 0.6807253871629247,0.8602879867210457,0.8840146438542673,0.9972144062621909,
 0.6033463673127036,0.2159939195413787,0.6738073552301109,0.7416982808689687,
 0.12729960525326156,0.5438454688671271,0.10252881095287425,
 0.03599093570169831,0.6102604986151091,0.9555030850725899,
 0.46057369144441096,0.7094565565500607,0.49966624404709736,
 0.48586250209871906,0.7911976280785488,0.31712847387216736,
 0.25878918203438284,0.8444038199736721,0.7659213342908687,
 0.29860437298906284,0.8015871592951815,0.2906335699372954,0.5034740702254812,
 0.1121960699450788,0.5837151556961307,0.3636333147238162,0.8061680230835688,
 0.7033466083527046,0.4477033183593271,0.8249370061783017,0.7873171305167089,
 0.30880411845786204,0.611031760989289,0.7937191290413889,0.739609760879936,
 0.327393338478181,0.03985305907873071,0.21041844908694796,
 0.46033803626367775,0.3841426488006374,0.12616774665306218,
 0.7099739825981137,0.07473787646972885,0.25631448370501075,
 0.9024303541046781,0.5128569595881138,0.44686350476378944,
 0.46010416443533175,0.7963154099199898,0.4202950374280595,0.8411691857018601,
 0.5824364815703944,0.5398935166752074,0.864220032278508,0.876486225334353,
 0.14540342889588354,0.9231211787334093,0.20403046094771915,
 0.12237903351085078,0.7130046318881154,0.5872610756517632,
 0.06659168388087044,0.33958668772603773,0.24841365973712537,
 0.1816473244744783,0.13632548011977397,0.43926645692776833,
 0.9987490806765161,0.046372368857064084,0.977616974096708,0.7619995504401342,
 0.766081515893366,0.9122739083476791,0.8104191973663009,0.9476858600904098,
 0.9370244646740097,0.7547401784699864,0.9524299113714987,0.19397568069756443,
 0.20591869178437672,0.9682915135040583,0.8892349702955644,
 0.26690061888529115,0.15897267905960555,0.974230376933482,0.6943607675161154,
 0.47063693241676174,0.9615094855413769,0.620732148880937,0.23665104975689988,
 0.3595314766743498,0.2536090276558308,0.0925230961282486,
 0.006648819572559028,0.40958909148304123,0.7000380921162547,
 0.9658692901490243,0.40175001418316936,0.3781274589814895,
 0.36278385823253156,0.7109816979141449,0.12738580597423677,
 0.2621024112952881,0.958269566412764,0.844988408428883,0.12489144391884377,
 0.9273069192751401,0.04508261471173802,0.918174027371618,0.24751913784477386,
 0.0052310765856196095,0.6431177287971608,0.49500299500848677,
 0.43705566017678465,0.7911012931406631,0.30511022336406823,
 0.0937726333011707,0.35288883939174753,0.41667933561165327,
 0.4196900200131044,0.8984439065359651,0.7906862518054579,0.07831303681299917,
 0.7578620050860517,0.4234596463977188,0.13136281612594192,0.3816254558303732,
 0.3165524643825771,0.2602609433668466,0.43637295441764623,
 0.20668683385710018,0.18412682909627998,0.2766073496768633,
 0.8322653327955355,0.31801788707472745,0.876702925481912,0.6986045219499043,
 0.9633119719227458,0.06021717459613052,0.6759279958134963,0.6663619218557438,
 0.9858445913872977,0.45091753897787434,0.18765216682641672,
 0.5408907288820647,0.2410634739103963,0.505962101312248,0.6918970811634751,
 0.20198811389701987,0.7837146998191543,0.7425420576922336,0.6800999034134957,
 0.21633289505232,0.691725672256664,0.5435910071880227,0.27653692376552574,
 0.448070533098608,0.09820749075089474,0.923006320599507,0.30961506959064344,
 0.4817949317015925,0.5201572048565898,0.9972948460964872,0.721014159230801,
 0.5950539315777829,0.8966120870848596,0.745747910535248,0.9103612957303654,
 0.7133533183234496,0.07054612617797462,0.0063299210771092795,
 0.3918096547433976,0.7245963585901227,0.11027457555083253,0.6959853572840998,
 0.7822617046360589,0.7306090131248032,0.5644650511521406,0.7810111425303055,
 0.48239611505390023,0.4254560113220458,0.8382295093615194,0.9393706251030949,
 0.5977522945619912,0.34173638156853337,0.029859726048969026,
 0.7564378817658596,0.8291905774601909,0.30410903284803625,0.9999625489436045,
 0.0857879376239229,0.2517752442362988,0.6902510396462508,0.784804672421139,
 0.0897538588468717,0.29500630926125393,0.6662481118902502,0.3895927926734156,
 0.8473576278951547,0.37931949795592557,0.21222850599380927,0.940242395913966,
 0.590199154163952,0.5484238128247202,0.4556004546682242,0.60343664742041,
 0.0853414130413862,0.12670575555945907,0.3887945232059128,0.580390600869078,
 0.5667146045874881,0.8337511237599781,0.7997214275678574,
 0.062415181028277034,0.44596229142800814,0.9555375695606052,
 0.4922402059211859,0.8684473715980491,0.1060979732523738,0.24466320234444672,
 0.5853897522098374,0.27213160565524097,0.5722864939111455,0.5716528536615221,
 0.22687879954838674,0.8983049412185868,0.7020583430119423,0.7445265738195612,
 0.826575952642717,0.3153225409879472,0.01756597287202677,0.9752462224630789,
 0.22495839180887467,0.721288441487307,0.8825904119231511,0.6752925619656348,
 0.9432911481916306,0.9615690258826275,0.332676310908366,0.7525582265558654,
 0.15418799676007489,0.5497866370748241,0.3882981666955374,0.7346396815430848,
 0.4593096893557813,0.914187727762604,0.5074769597909772,0.5855577479350104,
 0.7108814697290291,0.6963780087668876,0.12322847243410073,0.3579817373951917,
 0.6091161368020869,0.6217136821485186,0.344092062589787,0.17353528971333798,
 0.32746564418598645,0.48675601884401887,0.33560025514001757,
 0.524793858357408,0.7577190710085911,0.7563144813531066,0.6653790344629337,
 0.48105113476746497,0.6512136578610063,0.01301008176708851,
 0.7628560474220238,0.8211764471523618,0.3341869478555982,0.5895669024728141,
 0.4298343485109113,0.5020582270645606,0.02611242809716874,0.9929750863935787,
 0.4125393751675879,0.7187830150482296,0.3264490273204368,0.22927448061053446,
 0.21584125736198145,0.8918444476161582,0.8732927013088296,0.7894426350382846,
 0.637115756016639,0.2680666076257835,0.9179402716059315,0.4622043291092167,
 0.03275228981920808,0.7373605623851525,0.4008052397268035,0.75246886262611,
 0.5999073178289648,0.1927055983821151,0.8217818066191914,0.39235889164739357,
 0.6657554223688881,0.7714161487499561,0.0405602311693406,0.3698264693816752,
 0.7217241466941421,0.6564516796508967,0.22791212603492728,
 0.33203071461635025,0.7175166271860971,0.010982819799980392,
 0.9014618721183321,0.3709559953869902,0.545514522763148,0.9701512066057568,
 0.3118040772283782,0.8319716874260242,0.04799168197431336,0.5749391395789698,
 0.5273029252430778,0.4906315130796012,0.6579727815630809,0.01519528415544491,
 0.19117048255070956,0.7870584537069154,0.6093483263075753,0.5860204994914954,
 0.07945377575096046,0.7269352304477397,0.23995340727066716,
 0.6921419383053611,0.6633000510342971,0.3848903601382331,0.1742531209122129,
 0.2924370684328964,0.3156346250562809,0.5998790974928384,0.04492003198490535,
 0.5083870540769206,0.07222422572739617,0.7552332067420465,0.3554033356834829,
 0.9858314928252705,0.8258068874573128,0.13540877850317679,0.4657796206738437,
 0.37177909603526726,0.4707534954403938,0.6209785673387898,
 0.14100634343173002,0.9450209665440262,0.001055087621755546,
 0.5973382318094085,0.4667989563443248,0.6750790814792289,0.40667814769730204,
 0.35274533648555884,0.9106590307461476,0.47974268139076504,
 0.29918473172987803,0.8819893804971541,0.7072573054026894,
 0.26115283947384005,0.7578150345886663,0.3976099129711479,0.3676247213853222,
 0.8257025059379718,0.6591793164531864,0.7955475380456178,0.2053364264498595,
 0.15936568960421627,0.7083344589179472,0.8764493133699865,0.7175551163130756,
 0.9213139694415186,0.7637384100095119,0.04027996736667183,0.7772196478386885,
 0.0028846606312532552,0.9582710837319419,0.3899534411926493,
 0.6176855427639465,0.47962952114885793,0.7993039066672352,0.4180362032003764,
 0.18333481481667213,0.14540873206294957,0.6835392852420561,
 0.23319779342475577,0.8387280563920393,0.7069662092153545,
 0.18572918921572668,0.8001788239273178,0.509965404616033,0.9026243799926232,
 0.112902868559976,0.8629087381406643,0.008128681667683657,
 8.546801208813015e-4,0.5107316645842392,0.11725199334042347,
 0.6218220878617101,0.30483015153098336,0.8543166933810104,0.8849482260229345,
 0.5554587739586092,0.4377883589458337,0.8612594231885232,0.38989698392838057,
 0.13293206285925618,0.6772557557169955,0.39302026466051543,
 0.5257755729574679,0.2776973219449196,0.46161696565014787,0.2840190871818633,
 0.9878414961298725,0.5143399495025376,0.9755612821199807,0.9956448186054707,
 0.6083643868318728,0.4482884761657946,0.7311188172121723,0.5119797065397141,
 0.8862656340379469,0.076821765195356,0.358671808510339,0.6114680976221349,
 0.07425493689416718,0.3350783045184714,0.3861536746985189,0.1852600884501886,
 0.05507772211330508,0.17996326601903356,0.8219248723585562,
 0.40213316721144865,0.5225750122285105,0.991203675082427,0.30981763850739596,
 0.38089414566182145,0.7490533702034127,0.4344588283465749,
 0.09840950262738513,0.4481432481444829,0.8987952757296587,0.7970703505924412,
 0.7347717143168371,0.6157811953417572,0.7632357794793948,0.6811403129938328,
 0.6819612861007844,0.4667010557305168,0.24309951540790842,0.5062946775274367,
 0.35227206069659744,0.9157437140890318,0.8635698478390262,0.8713977440452555,
 0.17005978921470677,0.5770313283816576,0.271327933429656,0.8222136977617348,
 0.15133615432720715,0.238488956943528,0.8506870961501098,0.6766236875026609,
 0.626289707597924,0.18253381693300363,0.6332006432982202,0.7776731610046627,
 0.4468392274987635,0.542583613948458,0.5788937096620159,0.8895818232419629,
 0.9364272365886535,0.8457422983946654,0.4068443518666276,0.14132009734266393,
 0.6837710595110482,0.7912816565218068,0.4951885684119103,0.7592299731742311,
 0.9328256493513978,0.2983153248916267,0.45776311223965604,0.3401346481960057,
 0.38198612281909305,0.7864184035388433,0.14552771848240031,
 0.9994313594888593,0.4936753092841284,0.920408617499129,0.46941542290527416,
 0.7556114661185457,0.33839693015746597,0.7119132253567808,0.8537320972468289,
 0.23607476854636456,0.4170186298926817,0.37059070731058885,0.592734346749197,
 0.11587171170943067,0.9314062558728259,0.16587586160091217,
 0.3204423568938727,0.3445312830756255,0.37243697183577584,0.2103692381290545,
 0.16976225960735647,0.7533291870652622,0.7438585618819447,0.9214611042978788,
 0.12460009133675465,0.774560052214208,0.8991176395513484,0.5181675766639783,
 0.033513303667660366,0.11672164517419525,0.9733570880900538,
 0.9188336093977679,0.9543109644243414,0.8800555351424675,0.11597690850824849,
 0.9810971644556641,0.40098313269075025,0.030931966708858005,
 0.24675914479639594,0.49901822300108134,0.6888864762971911,
 0.5223905869255412,0.41741707133250183,0.1993852849704445,0.9306603825663724,
 0.19491621275150406,0.9714832227810863,0.3161112599021031,0.4500940385347535,
 0.0695511833086585,0.26843408776456934,0.46720828219531574,
 0.2888970531031827,0.3586481677926944,0.20048206948104375,0.9465743837649547,
 0.29667856956479843,0.18483905285243896,0.8332970701209981,
 0.04766906169330498,0.6494613345095948,0.03980923322003194,
 0.19462100249766356,0.5228644116650366,0.26196683741814175,
 0.8390577526898809,0.6002814217024104,0.13865308552376376,0.2964716585435119,
 0.5134235308680022,0.9798806791542252,0.7862649319617594,0.10814359187214717,
 0.6216311980858569,0.7324439896488973,0.9565954770168585,0.32195737696317417,
 0.854260610914326,0.24517045083436573,0.3649847243177653,0.6686370695398838,
 0.08633978271707854,0.62768120237049,0.798841758552101,0.7281776298228826,
 0.5610413573056998,0.1075829324197396,0.056414668087549136,
 0.18023285423847268,0.5083573382886456,0.31983402833334573,
 0.6871594299776953,0.00740398628943939,0.7845021046665779,
 0.06070174727269673,0.9223301323588683,0.8985837955122145,
 0.39215648918297596,0.4030334093485357,0.5924283332781337,0.9400560378964611,
 0.9011771560265147,0.5175463190132283,0.004118557344912732,
 0.23119895496591636,0.9034932313655968,0.3649196773755796,
 0.08201294285271199,0.6469943002648839,0.02527219464509134,
 0.2711639782061589,0.9964160463588333,0.07351064376308347,0.7327309935207462,
 0.583064754058584,0.9940809379153911,0.21138893727968022,0.7680142028764357,
 0.3299537024732848,0.5863975365966584,0.1937755942789493,0.5488937576845346,
 0.6429195305017653,0.6776563861792784,0.5984201530355221,0.688838111441159,
 0.0988396756905251,0.564947645535266,0.34036954149240894,0.02431113345466862,
 0.26655754083602445,0.3720463094611952,0.7522591664210543,0.5702230552181551,
 0.24615054445859286,0.30440666009349526,0.1143242467908373,
 0.0420315131147706,0.44714756677465206,0.6429954672694718,
 0.44965870976115396,0.22001523059031225,0.48978136791580185,
 0.6443322960021107,0.7335239823193405,0.9052459245359277,0.591244615832414,
 0.5480010448533882,0.6419617165443305,0.30628597120779133,0.7848338101065494,
 0.5941009821037886,0.5255543403582212,0.9677338040490557,0.8031368520593679,
 0.5243998265619595,0.9666790229997748,0.8284852982847363,0.18235622337941315,
 0.21246596462753875,0.7881612756690178,0.3179826671825724,0.5878424732982743,
 0.3702536745644327,0.09941647560511302,0.7807129793223547,0.6953859459566184,
 0.2812773969875413,0.9474265825292929,0.10798311689923767,
 0.40287524606836767,0.25644920606902666,0.37988457107974294,
 0.7794083344253764,0.4551296508907291,0.9571967949210058,0.8030009250207817,
 0.8076608727774196,0.5025665512295461,0.6243798855272245,0.6062928646126359,
 0.6135990178549019,0.8130502715339991,0.9778411070767306,0.9763363531889128,
 0.8526974208942744,0.8149723000067034,0.8116812381206169,0.8279433425058116,
 0.4295283828521955,0.8802895160610891,0.46874004038659955,0.288738701624633,
 0.4652468353228031,0.3738323120651065,0.25119743123320804,
 0.17666689991947537,0.6194909825356805,0.8068191045923832,
 0.43440083214252745,0.11296745059350344,0.5405960254438091,
 0.9514765144222708,0.15160262142058956,0.35111565450884896,
 0.26720561199510584,0.3251759512656356,0.5052401837777472,0.2148201127539593,
 0.39677186561819366,0.6785194220059699,0.6757318390142322,0.1096292757780245,
 0.8097401661027528,0.8980295298478236,0.8222531594925944,0.7701844063505244,
 0.27961579696854,0.9141970449596092,0.24801297699109393,0.13130420647024066,
 0.822385534408498,0.9374438463576287,0.39949334116950785,0.28410088213019247,
 0.4998517420526777,0.7359703145216414,0.07107458526491406,
 0.15829463079461625,0.36865685571585705,0.25837000517345665,
 0.287473461833053,0.7228559940016079,0.3540459567874914,0.15478213093191517,
 0.3688377048247391,0.2600677495012673,0.20212283204447457,0.1921282789144808,
 0.4910400600617746,0.42755832007717154,0.2488332283753364,0.7336922696453737,
 0.8877583379412781,0.12132127309669083,0.1114586351602842,
 0.044109701740779106,0.17595945432947246,0.2490405640244968,
 0.7260752365882728,0.9057556052376036,0.4856815335482053,0.29212800021363106,
 0.15255065510773613,0.6709916524350822,0.5971149990442728,0.8877100507585111,
 0.4723401341145662,0.8830659611113691,0.9530825420578116,0.23180697803656058,
 0.5958083301743086,0.383411422282941,0.9640843722383688,0.8948173277821494,
 0.09244689920986038,0.4413535358813625,0.653629471593373,0.2744196527014372,
 0.3597136605007095,0.7782585253299281,0.5553652761531027,0.7309998749207924,
 0.9108010543691576,0.9961321957011351,0.3008056838696008,0.0652886567063945,
 0.31927564789935836,0.8781423946751256,0.367144926417037,0.5711373839045526,
 0.8696692505929324,0.967214010573908,0.1501722548763358,0.9913158532159154,
 0.019863431652604402,0.8172517944132597,0.4390490138398542,
 0.16301975726649187,0.5324529011438194,0.32351256554155106,
 0.5318646843574553,0.8953104533140396,0.5117276216228731,0.1602827165699492,
 0.698321017121475,0.8287530397470115,0.8350801644253663,0.942666099545397,
 0.7886456425693056,0.2187280386936974,0.8571866027568519,0.7353036851142992,
 0.5012842479576557,0.8496174139986333,0.38662022939633856,0.2055291914207098,
 0.1345192489387509,0.9763839297455795,0.9811494616574513,0.6648923487923916,
 0.5400268349240523,0.5156453298806185,0.34559431014412045,
 0.01537121546244058,0.45225711088862375,0.9804447593926939,
 0.23915091125990096,0.6495001820128048,0.07192784402672825,
 0.012571515504394726,0.4161773365202812,1.5790439172880255e-4,
 0.7835386624639717,0.014575601551707007,5.104704095926227e-4,
 0.45178347634630345,0.8484803595472505,0.06507885887903997,
 0.007783018860463287,0.9605779896050097,0.39556200938202124,
 0.4569086068909478,0.1427183481938149,0.7974798369856915,0.2881220368877988,
 0.22535786687737125,0.7020708185009931,0.6345086155488201,0.8948322782323078,
 0.5737021035829264,0.16450709336751335,0.9442995105882243,0.546268425615522,
 0.8572418017575149,0.7467622623096467,0.4287874534981252,0.36081222844523053,
 0.9689104942030529,0.6620866830438461,0.07133983670481447,0.8130672564115434,
 0.4394140942261723,0.2148347913419021,0.1304338389444213,0.7904533897056113,
 0.2934544233867321,0.4438015021786147,0.571738625265296,0.9508717214755158,
 0.23023170727565567,0.36150852032844116,0.6530302354485581,
 0.18323815446051872,0.9836261881907219,0.1950317701838329,0.8096844083512752,
 0.5603267620113,0.6849344975899969,0.1481380554179219,0.0669223617837309,
 0.6716381905260684,0.024672030410766044,0.061468093073422714,
 0.40738759582614004,0.21344929568120163,0.9169656418689858,
 0.9301426981890092,0.5644984098339386,0.5730050422472197,0.4523542849004588,
 0.42323069177503125,0.5402855653972668,0.0633896356072653,0.3066721010489244,
 0.511640451620683,0.626282102454164,0.701796271554014,0.09301320059510454,
 0.7774729570036831,0.5057333436690724,0.4126824710005912,0.9475780949064069,
 0.5067171870562067,0.2852588842115221,0.536687507516675,0.21127517978044252,
 0.24623850042254047,0.8213132673014846,0.7557308731686035,0.7216095426496321,
 0.7988485378804797,0.28360789013711385,0.3414768087307207,0.7079356488820735,
 0.8583452581673157,0.21657419952492374,0.1556770819682214,
 0.18208178112883378,0.9144805909352655,0.12695087089671042,
 0.3691902358202275,0.03073916458336312,0.28681349274708023,
 0.7150391539083847,0.4313612766330517,0.5355237019584722,0.21694441018963517,
 0.34797101444919376,0.5736159015777761,0.9488751658737379,0.2921302597983496,
 0.7864044092241598,0.9317602103866094,0.8763496213653819,0.1599613846218444,
 0.3571571481587801,0.8227577043520945,0.25694345110803074,0.796382850452601,
 0.6038458479269813,0.6952839233402474,0.42684659256554847,
 0.25200926499269893,0.5263558663767767,0.12313431088224414,
 0.3900767418036629,0.584732536019262,0.5054457219044584,0.7757079344769591,
 0.4234308418953817,0.9988317606099154,0.5030511533325592,
 0.010479156352494856,0.8697778436460699,0.9461877415444057,
 0.011657194385970127,0.24499773675031633,0.5328890526943009,
 0.5417024384155936,0.09846806423654031,0.96492532248785,0.9350993955553484,
 0.3929991447639737,0.789578523045696,0.5812801199246436,0.48941648579605046,
 0.6846443410359703,0.3366915762989606,0.7804664795441405,0.8961541104621651,
 0.3151852032100697,0.3933580950756742,0.20468846271146057,
 0.10311336633455936,0.9963489843937114,0.14462246240866072,
 0.4150384412155552,0.9241049130971941,0.09141219203813467,0.8430656588754452,
 0.6259048677335395,0.828278363354649,0.11090663006583035,0.4307229822794898,
 0.10622981189207692,0.7450068252259398,0.6594187792643493,
 0.06299633565393736,0.8264290965686437,0.41582330595338046,
 0.9831124282500338,0.6686456265760706,0.035382494302935275,
 0.042840092873920654,0.6276534698754496,0.36662462232605897,
 0.48490870803882913,0.6391859948407284,0.8198159792042476,0.3239927921608382,
 0.21902568872347095,0.2645314124639827,0.2346322568641216,0.6542784624137321,
 0.6438891590680661,0.1597203572872009,0.2633828249642941,0.6071411578092091,
 0.4967566262380173,0.5210981171743929,0.26719906387347636,0.5474936870936912,
 0.6778205204543334,0.4414089858943483,0.39568439192852345,
 0.02160504907405414,0.0655739873624922,0.558502240349969,0.11303663494148586,
 0.622414265540237,0.8679417219078083,0.9796013427180144,0.14364080519107547,
 0.4823287786644076,0.22465575573646868,0.9498692126320786,0.895228489127554,
 0.5197347289826711,0.5768870527078097,0.38951940819641384,0.9878862716907238,
 0.8951038231762688,0.8497654443387104,0.3319319401056139,0.7027903016448056,
 0.8634907877932297,0.7207491264446602,0.10550317751289506,0.3649519825135531,
 0.23651541449043634,0.7821175911997473,0.6117799720113903,
 0.07404121079705428,0.2846563791694565,0.5528868148930209,0.6413121276039003,
 0.6456931300830114,0.10709394254534388,0.695300580063843,0.5324493550876991,
 0.13107036892297286,0.5368710309420173,0.6160298038833304,0.7603473446123896,
 0.8691965506904796,0.4655025783136799,0.24259946181334613,0.5489858584809748,
 0.21433610343474507,0.6176161391046636,0.3503258414690882,0.5372882740990371,
 0.049490789057313855,0.25047417739651656,0.2976310166580738,
 0.25927458663387126,0.6946821504596699,0.025599813535768567,
 0.8177492026115318,0.3895342660841171,0.6899106870204638,0.23776275011314452,
 0.8262942987921418,0.861556884318002,0.5698603443702271,0.7379132639307903,
 0.9386575262002445,0.7800630184218162,0.05226323377643183,0.5716388217008659,
 0.40945078013615954,0.3929237911287542,0.2490679308606516,0.5080308949622745,
 0.20997157684538714,0.6382028945658591,0.9765103358204125,0.7469926284786921,
 0.8835782555763805,0.40792024366189406,0.33228957182793917,
 0.4425860503105179,0.4463193074269344,0.6995024484769633,0.6954365514063845,
 0.33715196920790624,0.05188034827478638,0.012500924799432211,
 0.4517943181382642,0.2053512244496768,0.8360148853580174,0.5139620815299722,
 0.9360829419566553,0.19441714031152935,0.008345647089946229,
 0.7824123430962189,0.715969478993466,0.8518261069676163,0.81227368853945,
 0.17236105919459643,0.3497655560081401,0.5668009498287595,0.904075291211705,
 0.6715261197103808,0.15868949073966754,0.46564051678353413,
 0.8573809078178855,0.21535355590017746,0.03345589279018801,
 0.10819584410035998,0.2764257393574473,0.22861525833983076,
 0.5784166365339745,0.071151136604235,0.4869876826551036,0.23200375839906995,
 0.6136280106384608,0.7041923771862708,0.13287722853697037,0.6975823289201457,
 0.09887312375908797,0.3880984443732288,0.8899783323601774,0.8148792426591588,
 0.7587349593132138,0.9557401185735713,0.06526623993914582,0.8523512978201139,
 0.8049886546383354,0.5234307210536248,0.8139467742227308,0.7330646151184324,
 0.762713895843926,0.5830968036347229,0.6636157576828529,0.3942736023895823,
 0.28898107604241063,0.9882238991797208,0.38504408101534715,
 0.5115364566467083,0.40076978473720604,0.555499328574514,0.734048452649475,
 0.37549572236457573,0.545733385883138,0.7748825167971854,0.2238147771759189,
 0.33559947470681495,0.7251492191804572,0.47567824171152884,
 0.5863813659206875,0.03596786680313846,0.3783123143860684,
 0.18047353291852397,0.12547992565430344,0.7198985997439123,
 0.3189733994949594,0.9688702730714838,0.27885651789328925,0.9003833164850076,
 0.8962564036988516,0.2457450523824236,0.7355319479218347,0.14421655637144215,
 0.790181581521365,0.9917880035527287,0.4832515538320331,0.3472308104850149,
 0.547890521379844,0.5122216957006205,0.9483108198189892,0.2487273475326941,
 0.3123453645120281,0.9679246860601648,0.7508397283552672,0.12196766380184754,
 0.009072473274683857,0.3205035082775216,0.4511476766404806,
 0.2548610065025294,0.8985907613557537,0.560743645077551,0.1233583764155205,
 0.04126849480278805,0.712306544631669,0.8529310924143989,0.7265335083692258,
 0.13075996787761224,0.4774442896365395,0.722584196899329,0.5343423173036077,
 0.34226258495528306,0.732977587296117,0.9594206556696548,0.24412809569957328,
 0.9008689103481495,0.11488167830435192,0.7676584265972246,
 0.045353846690642796,0.48333670772774384,0.2915987280060772,
 0.2206955365809924,0.41260872347078303,0.6166635808765675,
 0.14725354587631423,0.031083555536337792,0.35471663659698227,
 0.3959486625747849,0.48337587870199317,0.2566815514074189,0.824564489758137,
 0.10971662757131961,0.19697259107925247,0.3161938665215883,
 0.5516426560025761,0.7106255917041845,0.02072601595311585,0.7277811414402138,
 0.5680017397153806,0.08351571757829956,0.4597229317872653,
 0.12437882923212318,0.030774564563120443,0.7213108450562655,0.43666240439826,
 0.9023155553851937,0.7982662198973732,0.6472128939271894,0.378154233174569,
 0.41283515757902844,0.6363200392730205,0.3207876725688942,0.5186513041012208,
 0.8976140790824038,0.754435918721813,0.9394135984485369,0.3205097770187936,
 0.3064101772095642,0.9541680828098726,0.9218843157730741,0.04555464020488187,
 0.05557725717216322,0.3086185650972152,0.6447053307459991,
 0.33957328554415045,0.22411046723665584,0.09061870160040564,
 0.33783921236606407,0.4985386646927239,0.5328483555367617,
 0.29961409175724873,0.15258273406254796,0.03026319587085291,0.90864360467186,
 0.42256288361820116,0.3941274887933769,0.7150173165300154,0.2930636404508704,
 0.035889726440238245,0.8283559992131762,0.25130124723700953,
 0.5255442625350649,0.856222906858467,0.9034308633326213,0.16977423343988274,
 0.8802428339446368,0.0858054983382065,0.23369269932499093,
 0.34836981249990573,0.5584387679503005,0.47405912132190153,
 0.8766462606165135,0.3859219485297274,0.6979142814595924,0.605873762850066,
 0.35544977636130204,0.8307863420209267,0.07997906393406173,
 0.6691876585967024,0.07314622547366079,0.6787655867421742,0.3535769688635615,
 0.007061623175173781,0.8942307389903466,0.30354316102546597,
 0.8052999010780529,0.15797849878685866,0.2329408682320765,
 0.13200357011608854,0.765658965513336,0.791073056729909,0.3706288423189795,
 0.43889587154502085,0.3618447147655336,0.6746558318655751,
 0.42843523037708053,0.9414981075935874,0.11941140770427883,
 0.6559442596684177,0.171241610353303,0.9173133908454412,0.04463166169491073,
 0.8422734396096673,0.6772629705728404,0.550934138241362,0.1726652452329258,
 0.10220686350179875,0.3501916896031967,0.5147071907748372,0.5960103737967748,
 0.30078397124204614,0.9042301982987817,0.8949479587790266,
 0.20382000856475668,0.682277812292035,0.8971569953308223,0.5982092958902314,
 0.22244709233447413,0.5432942243172895,0.04211919070665904,
 0.2226415925305698,0.5610731212606518,0.2859133132487125,0.7700613965736681,
 0.023976635986062922,0.2681241358916707,0.945500719993371,0.7388115780722794,
 0.4605987641573276,0.7461954792703889,0.30237924246580894,0.4221445332253221,
 0.8117963957448178,0.9544122122906921,0.5158828474860577,0.5788468228919824,
 0.9542972135969301,0.5518508460721254,0.8514289527273988,0.3034895845718961,
 0.16639887841035472,0.6160593914465476,0.8228694500828758,
 0.016732047033088393,0.9430436566087016,0.5282325650708074,
 0.7797312305815396,0.006588513801131235,0.853936409548437,0.7375851608783404,
 0.6860280445870443,0.37567999748907455,0.05560511816496738,0.307010280480698,
 0.8925938017162793,0.8208135457860405,0.033424933214558084,
 0.7622042518823933,0.6068548324653593,0.603776192156452,0.39976477133876176,
 0.39677565425083383,0.7683287372575076,0.8771318071223634,
 0.44234736314005874,0.146034307576149,0.7706186990082159,0.4551209049970113,
 0.4537901332710279,0.273413157402137,0.4401198452333813,0.15767204085303044,
 0.44400316463238476,0.3674823920233473,0.43519735904339585,
 0.46122253023644166,0.6927188216479437,0.7459668654626102,
 0.008004488450014002,0.3847375982282505,0.8517930346941626,
 0.7061810801282108,0.1783438648481228,0.9514963081825216,0.43778356908377647,
 0.2481769325357197,0.6654173190886424,0.6728921156680453,0.45661533035949353,
 0.12463421477557002,0.07513419594701787,0.6266000445535361,
 0.12054790419777128,0.8066152920538107,0.02155707400991891,
 0.47888982662880153,0.1551835049209883,0.039911681501516916,
 0.9103518359029742,0.015493823161673803,0.9116635561838061,
 0.9323860867614491,0.5984670681898334,0.006768425411911028,
 0.7101839639652403,0.9444621871103313,0.21691001682215538,0.8428534925139348,
 0.22424970707056446,0.09550384648823562,0.584706234812009,0.8699942083580209,
 0.659454026539701,0.42177897633796857,0.2548255909808631,0.5035702167584268,
 0.7780189471952852,0.6598926846492894,0.15807997832110177,
 0.13372570162610842,0.6969248119795155,0.8364845113817369,0.4558499993466657,
 0.3208834081776406,0.15048878611089211,0.7257805178107608,0.7454692257781325,
 0.8912851940178417,0.007109451545574963,0.7866530893836935,
 0.9145877853341867,0.7855690974703002,0.3529315919713616,0.7961041915142102,
 0.35843210813841053,0.002223085645746248,0.8579247614386513,
 0.9172991279105255,0.480387525375948,0.18980553796960586,0.1385212604546069,
 0.63615506946559,0.34542619318041,0.8984158947425503,0.8571503905395982,
 0.25809158952020284,0.47757221324685084,0.29308150519316234,
 0.7201105330087588,0.7437122859764008,0.6991849629452255,0.4124221630270617,
 0.4558418566583037,0.48497056295234175,0.35965274776765377,
 0.7678810921659008,0.3945471009184506,0.9778885188455321,0.7190288581577228,
 0.3297794148543176,0.6665673488890502,0.7831081699749852,0.8883974573137688,
 0.49520368502182,0.4517487785165436,0.3374082677585979,0.19481716874691368,
 0.5232991261175897,0.9505695142933296,0.34624355587976285,
 0.12129213207879719,0.054119637526531594,0.22708439746543618,
 0.27093694013261005,0.29712429024462106,0.6032724714806957,
 0.23517836033447848,0.3328576703624677,0.30282497942140874,
 0.10699376771557612,0.3723295123488489,0.17399819471280065,
 0.8795389995403962,0.27451417921983645,0.50552362067548,0.7922922149440006,
 0.6486530561227564,0.6755855652025442,0.5571440105540876,0.311446324301765,
 0.8931811623864228,0.5038815408505501,0.92757273337314,0.7874226098736792,
 0.34207232826637135,0.5521765730493721,0.8857643484818194,
 0.17714333270129923,0.06987580225543744,0.2609271571024928,
 0.08473695589721442,0.7038096903705154,0.9025902635678819,0.8966276831670263,
 0.257211849209524,0.7813912482208938,0.5353516054896554,0.6068145043352752,
 0.8503656212906725,0.36033304709563085,0.5096356412770169,0.6113881394524505,
 0.7887882390712403,0.8584501332092361,0.7537790523055893,0.6527652809026874,
 0.657594390037755,0.7601198161634394,0.05564249220885609,0.35083415937925033,
 0.5664842387642757,0.30785433628946246,0.2920012897670413,
 0.13591413293931676,0.49526294543075733,0.4182052017655964,
 0.5775554114656681,0.21571007418540233,0.5002821290842886,0.588721360703193,
 0.642490484494151,0.043489072034579745,0.0873865890519897,0.6579488722392782,
 0.5327768786172686,0.3258592592919829,0.9288015374663882,0.7777481040750358,
 0.11286598048948415,0.8324455002487744,0.2876284393490356,0.295199675986465,
 0.7355281751902248,0.7989013833771049,0.039497862020117847,
 0.48487254262102975,0.24226839291711832,0.4963716476019133,
 0.2341602082230756,0.4768907108925723,0.23400035338793423,0.6825200979628592,
 0.6453092967092443,0.9173000873882557,0.16440043731722342,0.9260259983329572,
 0.49030571375915677,0.21020134251220512,0.9363163339152226,
 0.2861210859749679,0.611139880663939,0.4254647348643812,0.9664845454935536,
 0.9533020169989905,0.5892974363272095,0.875803666030085,0.785374599595878,
 0.9440417978190206,0.20293622730561356,0.020342871546822572,
 0.25828416011520794,0.7813485575288579,0.12085652244609701,
 0.9483108422410259,0.5011339438409611,0.24182474364429207,0.7701065827115043,
 0.4675075566070561,0.9110119708935351,0.7513118219856914,0.19310031583595533,
 0.5508068082211521,0.8348084520406953,0.835723663779084,0.6222524738622044,
 0.21972111065377453,0.1635832340806075,0.07347567260922527,
 0.9625582750747276,0.9183553684045934,0.08019191674841519,
 0.004079718037407698,0.45150447221901446,0.9647062984576144,
 0.06514979367823803,0.5663077847720223,0.09698710255795218,
 0.47555483726586123,0.9729748372376553,0.2431799926452327,0.7148366756179383,
 0.847940277656352,0.3406232586466015,0.4884366154701736,0.44149745557025855,
 0.8233486622154451,0.046743563452184556,0.43304952166204624,
 0.3428646780296525,0.6074169960280148,0.731289682572836,0.429930674624616,
 0.09754396002827537,0.09681203504881608,0.1791098446308469,
 0.3444384043172637,0.4639232694044916,0.4939832691823992,0.2825517833939686,
 0.4261302919798843,0.4118683314344611,0.7464218193888593,0.6082746184039554,
 0.8944346212196395,0.9715145563910035,0.7795285628685354,0.2788532428151512,
 0.3581163027552453,0.3659696838311769,0.8362951281086027,0.4698012138857698,
 0.9794130289883655,0.09894669096597164,0.4161027867296837,
 0.27198433734226246,0.5571370544066953,0.24549148876164373,
 0.8487764034978366,0.40333931063789774,0.11646755829722277,
 0.9111547471617611,0.31950140750944045,0.6880282424678414,0.3265199545331978,
 0.29472308527418134,0.14419576169838355,0.5119429627967332,
 0.027971046152714396,0.516003041723454,0.07371675368643793,0.589660787073481,
 0.22127814100987475,0.32885276334389757,0.6910676797554886,
 0.33807255917155477,0.16133243004258746,0.3823755819231991,
 0.9382574429389233,0.08354304862790951,0.951623956356745,0.7888284869811906,
 0.16307112049009687,0.15608768465196876,0.8361150049387408,
 0.3073103459609181,0.03979742153401533,0.4645561024901338,0.5972638111995703,
 0.6398579247749094,0.2783366539976987,0.21177849661459736,0.6978412336095109,
 0.2753299775456036,0.9198635616921292,0.6520879115133194,0.7074693822377047,
 0.5180321907000768,0.03400443207881121,0.8691104910640951,0.9063420175621915,
 0.7300535743961624,0.32281926499667724,0.7919859906634987,0.5245450361949867,
 0.7226258312424853,0.28821364142952055,0.6714406855164173,0.2526897406203583,
 0.32579447565686803,0.9246256798195703,0.37529319912826187,
 0.9958336137006756,0.4118887793387124,0.39128652496701366,
 0.03666489416330354,0.5558357578774593,0.26895994592262984,
 0.18165637377065535,0.058883183526558525,0.9903898221935127,
 0.46765083966322696,0.6568435929180765,0.3303932790659653,0.8023840729877502,
 0.06586645853253437,0.17236427879398664,0.35306686265130294,
 0.5053875824576851,0.4504748345273416,0.39799035652016246,0.8103296994154892,
 0.20315544425332377,0.27508403193361275,0.5298341950232144,
 0.7152300259040264,0.2755373178575218,0.5263619303799991,0.9116986142476605,
 0.18434014410426958,0.33228914505536655,0.1787481904917262,
 0.7825221811183505,0.02105272772875788,0.576250294368174,0.1580294711543364,
 0.057838404926696896,0.58596734238856,0.2941890292638143,0.3912487874989867,
 0.9915392067976302,0.3362866526280991,0.06449951528172759,0.5209820894788258,
 0.18670540138974412,0.5191000402449063,0.8839405742165352,0.1191677469719562,
 0.8052768810638493,0.6857789041715886,0.6740222641308706,0.5100737213616139,
 0.6458617658152104,0.9876220877323374,0.5287939871840006,0.49841459785531705,
 0.9666199604380137,0.8801424983331649,0.6867724472328961,0.6597256570774443,
 0.12161802074766459,0.2398993889101082,0.7557181011475561,
 0.10942672199705039,0.8474654219098589,0.5716374716326039,
 0.07640137387568957,0.3277111145913142,0.6042132153867265,
 0.14556420071344378,0.14811630867110515,0.4261519466783248,
 0.6523742119105497,0.08786192480439126,0.6004197175197906,0.3481364007138953,
 0.38374071996553427,0.6516833534623928,0.5390329387171616,0.2551864619180253,
 0.740760646165513,0.5231846991621871,0.45108308779840667,0.1253903205735013,
 0.230032191638895,0.4579929354330915,0.31006378127330614,0.696770383166798,
 0.5621060796280578,0.2514979146855636,0.6282947425452159,0.13403917220801287,
 0.2886493813505039,0.0913884181291037,0.7609768449024525,0.16483024624946352,
 0.43479613122899796,0.3899248281115373,0.6758424947089993,0.5944018820649489,
 0.3823302416448837,0.17046314087174552,0.692785116129403,0.47989344195981276,
 0.9545424301377134,0.15597755346284858,0.8080889868979773,0.5025147145728466,
 0.9929769801646374,0.7815256354683259,0.23842240440124984,
 0.28835422743419636,0.5149765579777564,0.17212563222481259,
 0.5875711738597097,0.0316178480601067,0.9534282732364123,0.8652193420042934,
 0.7698195486712752,0.18308415342591355,0.09190438463624306,
 0.2897289432327076,0.6924894794560676,0.5143138940169394,0.9571962696680643,
 0.4982011876470783,0.3982393691687669,0.0868614606162369,0.28401000258610243,
 0.1315957299106406,0.4059561059181023,0.027568544868151856,
 0.0072151680672557506,0.37486286794583323,0.7357046391269793,
 0.8166936911258329,0.9111904437817779,0.4252462357737009,0.4940089556845083,
 0.009274945122161538,0.04873019731613393,0.4985838901556481,
 0.9644711167273209,0.15496910687398235,0.4730527434086331,0.2954155316684979,
 0.4015696309209006,0.33840484452353325,0.5434393496193108,0.8928081331566622,
 0.614171748556528,0.8086364632162558,0.2677078375591784,0.8470157313602042,
 0.21427330436388414,0.7112386802569081,0.4451775660881905,0.6053671097001454,
 0.459707216658782,0.9262607951983868,0.7031958960524458,0.5031487292124084,
 0.9570506783122461,0.4704833605845551,0.20317456671744238,
 0.24085991859072142,0.7139283642952929,0.793795227621043,0.36320219489370364,
 0.8107070661741704,0.75893582052024,0.7461423447456756,0.4381783674867463,
 0.7346432959630141,0.5955987011444126,0.03681138694541852,
 0.11156444062084026,0.798652162715312,0.27291876901150125,
 0.007691511330180867,0.35526506401154156,0.13423538729419926,
 0.7708578212891656,0.013379855143347497,0.1338571859413813,
 0.13730634431053035,0.39465294542383145,0.5575384423466377,
 0.06409678547269071,0.3802479087854438,0.05395012377976405,
 0.5308336229908439,0.27385576239015996,0.6738710860081545,0.5890941383741903,
 0.011805288795965985,0.5537826843294984,0.3226547861171588,
 0.0033570366834246856,0.11502253493834613,0.9717062302530644,
 0.5586583510134454,0.7515780397053129,0.8612646992978531,0.5131271173698647,
 0.8267060702918323,0.39613401209824306,0.3962305799465766,0.8321560441600822,
 0.16646594664120973,0.2079333575943243,0.38545610449728485,
 0.016084830610534007,0.7038528111876399,0.460083148801838,0.5886339715166726,
 0.9893862234448146,0.9791067606454231,0.07527466085378842,0.9839297750364238,
 0.2946600497430989,0.09879330329732716,0.39748994927210346,
 0.09785092266888151,0.32072599455566186,0.35370788476575765,
 0.4349913419531092,0.20450566813487148,0.142020356778651,0.9771558882171909,
 0.36024586179062923,0.8262377099606559,0.8971216956478364,0.9420191518627757,
 0.4899236469811552,0.015062081688553497,0.5959507463007375,
 0.6469664971187341,0.9726433468921639,0.5128183740220148,0.7114667546508847,
 0.3852079044939143,0.9710577099665063,0.777103374272595,0.162391199104451,
 0.7951166626705674,0.3708602974430034,0.6322313705523777,0.7366553850338271,
 0.869896702219943,0.9350524463110547,0.36090118391762027,0.7535640206769746,
 0.8813815476755943,0.7338728204443155,0.09817298178759704,
 0.09372049014535178,0.8488085468668249,0.7050174323216967,0.5536592406029441,
 0.38890418955269124,0.25186582322557505,0.9627331487117115,
 0.4573926826163234,0.27689047604064465,0.47950347887898326,
 0.3736784994900648,0.9363051477807304,0.5001514142905331,0.644147258457324,
 0.04813861076859105,0.933477969519452,0.7925460429153875,0.6658141766790986,
 0.5271018008734938,0.9678288971067838,0.3421080497606994,0.21231931591619269,
 0.8251684924597537,0.6214974680971375,0.04130224457039289,
 0.01753257884913928,0.023046116231250968,0.6186259866219426,
 0.5439990700293439,0.4990610690992562,0.999872666477698,0.9307922943776272,
 0.7302352367449086,0.3223038880750313,0.8026753343169609,0.13759514395113903,
 0.023570218449593616,0.05155716372847463,0.03621267450906496,
 0.29671327001374204,0.7731484719863958,0.980981177737956,0.649909207751584,
 0.09641617227325772,0.34803864545176677,0.555812392087089,0.7984061744965539,
 0.8840760267219073,0.006057735071899439,0.6152120087797827,
 0.9245662965927677,0.9517215394661802,0.5523524822007432,0.7529810376180209,
 0.7028658174867731,0.5641075181538684,0.3346415638499419,0.2826560917592925,
 0.4351457440760985,0.9252510711384099,0.7848596493681048,0.23409726115959284,
 0.9782737709974718,0.1766865031096012,0.30339255447999014,0.5783757104389896,
 0.9806568005796398,0.21808492621785225,0.5032108685941649,0.6890160845923617,
 0.8607834346213498,0.5734792225203464,0.32103037803863876,0.3018616945605651,
 0.5330797745531767,0.38215155911457,0.40663834566585577,0.3960665743727072,
 0.7138614971097248,0.7459916333981998,0.8109654740065496,0.0440511432541536,
 0.6012490115367469,0.8140026150966403,0.025604237134103758,
 0.29060266297805426,0.5832430706420515,0.8256660637280493,0.866864226882015,
 0.9887242189571864,0.42960146500395635,0.20174338969596028,
 0.20136504435343094,0.018119203754730417,0.20901490850589177,
 0.3655169099142104,0.11459965423772178,0.4929910523715122,0.9656894264017285,
 0.7616393982961771,0.17778758360559743,0.5083426873601529,0.8105619431626678,
 0.19288696740777456,0.9126349893329446,0.08159986416534548,0.495123761496827,
 0.2736861809660034,0.1742045106143495,0.12277138614691374,0.5587668373763578,
 0.17194475646071372,0.8146272836876685,0.04193830635514262,
 0.9611137106683437,0.37475706725905156,0.992899660726065,0.49810618563536746,
 0.46171501716864594,0.6514806232947388,0.4085286271545169,0.9163877251667998,
 0.6691591198616977,0.9034583166694327,0.21285807678506452,
 0.25231560599067393,0.1553711862509658,0.793900794605888,0.10549080363171082,
 0.24194785077970482,0.9453405132604074,0.44224290630707586,0.500072887592107,
 0.7192325424784756,0.07529029737575976,0.09130190000210092,
 0.2652747762425277,0.7719659889215968,0.3708383642903036,0.5738214554494195,
 0.8084679963263186,0.940357683299242,0.21960867423558894,0.77565499858572,
 0.6670818064840615,0.25836163415149627,0.7416454164053236,0.6464939960582853,
 0.7473601995516721,0.7895748579173595,0.09513187785708976,0.5806088793608782,
 0.45839916907698053,0.022515504726009383,0.769662334638022,
 0.1581573436311503,0.6906187233437013,0.8425226145853253,0.49145139550684236,
 0.654141412506828,0.09420287734394839,0.8068408495670646,0.594267391709701,
 0.26504361283920475,0.3658967635079722,0.40582848224092716,0.777523919599804,
 0.7000315859807247,0.2955733572597782,0.010776122003894328,0.849090883557064,
 0.9101457932599426,0.05695022812072459,0.665120805458768,0.0727331133810748,
 0.9402120497165356,0.3413844267099715,0.3834269251475745,0.834645135291576,
 0.8310520776754938,0.43609373889851333,0.6443159652092126,
 0.035135994326689346,0.5953375612214702,0.9028357312519851,
 0.14673603046131478,0.6841209271647997,0.5223878399992599,
 0.49546340232193176,0.07477640233336391,0.13444600194391276,
 0.22578785025558912,0.8486909795986671,0.3825900324767819,
 0.25345751990129917,0.7029981999062196,0.7823169203007161,0.9973616377042381,
 0.5129063961037212,0.7314077597258303,0.7586837966941546,0.3443700158798556,
 0.18443333454898858,0.5634171910451492,0.22329501978340294,
 0.40449385588684883,0.5776385156846687,0.38669207951975293,
 0.5136802387067879,0.35454128009262065,0.38040801666061863,
 0.7499247829892921,0.9442703661203266,0.1460152329826253,
 0.007271447149940546,0.9351066975000172,0.33960849118498415,
 0.8933544780558719,0.4849890600906356,0.8456824028513339,0.4834268732890711,
 0.8122385800433153,0.09419831562974135,0.8170740457712488,
 0.37335733640224733,0.1965624661944989,0.5036982428074871,
 0.09547088860297936,0.8945696341663975,0.33704888035515534,0.190216737914787,
 0.9489158117716089,0.8170970093766505,0.971091946158414,0.1245163280961803,
 0.2689369418257985,0.45003966564271924,0.2868086760576314,0.3677763455257804,
 0.7856720201851506,0.020829210575210144,0.47448237465074194,
 0.16159572891001373,0.1584235648917105,0.5038738001613078,
 0.09361402630070959,0.3109586408395334,0.3176796022910954,0.6036121026808599,
 0.43012460195762636,0.19130435443287852,0.15166591750795222,
 0.02657035249659545,0.438858621090111,0.9941414722037039,0.6491084229551447,
 0.6816615168582509,0.8072419595759627,0.943300433552433,0.3203975849596228,
 0.21427876223661224,0.07490863149514804,0.4474680126108357,
 0.2712739937729213,0.1117406601925992,0.6222445256290428,0.2147549098242294,
 0.602389754600155,0.593569331606215,0.027084831730244563,0.7896015499008334,
 0.3290769324079539,0.2898743057677696,0.7122440520721782,0.06616891932644786,
 0.8850252586014445,0.5767052362069305,0.11730702186627251,0.7885119627082438,
 0.791928480310833,0.5560506941868556,0.31527003543763554,0.38002475769427235,
 0.5459631959674262,0.22379517367259982,0.6834324736348619,0.760511898838642,
 0.056062384332359105,0.8903712572060032,0.3252197696580381,
 0.019364837761782994,0.2965942184422301,0.4199934586902686,
 0.6574505278362311,0.9510862285215634,0.8411040786832764,0.3518760610927738,
 0.18610086276952076,0.22584683828250984,0.10499869029825737,
 0.6492656112054978,0.10248266556078423,0.2263615613743607,0.569887899704221,
 0.40830697431397445,0.10047361199417071,0.6609608635173508,
 0.09559191073217155,0.636661232930897,0.2776561520795242,0.8808920719795789,
 0.014949763194650911,0.46719629982607314,0.8564232477085415,
 0.08644938970499239,0.5258436159005377,0.6525207295442734,0.8459042029933737,
 0.597185541288966,0.3371192378589556,0.20146107515996547,0.11616765805135987,
 0.16403835636960418,0.22759340241200365,0.7508724152327493,
 0.07145693890257943,0.14260878645856767,0.03021955419596334,
 0.27806867432442195,0.46138168161304594,0.8732004638461224,
 0.9529174521840797,0.621750506823738,0.47944624353997023,0.1745604112254998,
 0.3481435353065476,0.0613904453944778,0.6054813115666826,
 0.050692963897725374,0.6253217332603315,0.23062321014266773,
 0.2655746428605704,0.725146966888718,0.30736335251539004,0.21733570400602242,
 0.264451792573426,0.5030487135148237,0.7514588188981313,0.06178358977764409,
 0.028972998922992876,0.4208957693899861,0.5725701028033859,0.510922748761851,
 0.442256258663989,0.6073922887004849,0.15416394968652425,0.633566638271249,
 0.564472264799221,0.19393484275026585,0.861696102633569,0.6627004081358489,
 0.7751160115451746,0.3203807450350331,0.09975107866960553,0.9971590303870591,
 0.742075981526658,0.5900626301282385,0.054678806049315076,
 0.059247570195360644,0.7327813771477216,0.5146401494614249,0.136341339111411,
 0.9311954017103306,0.06997387231808938,0.19669902274999884,
 0.5321620066886094,0.8078011670280782,0.2632644451304341,0.3074325234350904,
 0.31250185018851595,0.8519374139013569,0.7520379246184101,0.8146795183359519,
 0.30445614143970545,0.4800041731896949,0.044473251485774745,
 0.4131082416855809,0.4972397351969674,0.46413868797688673,0.2270866534274918,
 0.05396161067273386,0.4546336273421042,0.40818482982720516,
 0.7210356495470007,0.018830100909411174,0.39679127158098304,
 0.4217538291635341,0.8421845326858531,0.365497064373977,0.37157234560123065,
 0.8206721734902828,0.1136243986642177,0.2909462149226938,0.7220925592480588,
 0.0845677473119103,0.4933037934971626,0.12905551614726007,
 0.18390114472034258,0.8801238340736102,0.7274672882082849,0.7248137781114244,
 0.002340208315820491,0.15500427219279334,0.060340042930895676,
 0.1398171749661703,0.2837226775777222,0.8664763038796632,0.13280320443733995,
 0.7814893782508929,0.7338323325844114,0.323007153483854,0.923687593026035,
 0.1895633858996828,0.682373339772173,0.5853139093003796,0.18945226971966111,
 0.3510654936884787,0.4799798343644166,0.20719177903956054,0.6220880933230822,
 0.17451762646327673,0.34566447702546366,0.5070313360587819,
 0.39367298355997793,0.5480347044645895,0.1263985962854126,0.6958016589335048,
 0.774222591162914,0.31292124882543126,0.32545193371380665,0.8008530760075866,
 0.8130380073836827,0.22663987395708152,0.39714221155578877,0.632231875617297,
 0.6844973327477298,0.16361545521474985,0.2552429264445407,
 0.0014200301685018957,0.4843583393377199,0.5362665017378994,
 0.4400086671783092,0.016121130476141388,0.7305031646624973,
 0.21034049115924724,0.8588227960503212,0.16719617362170003,
 0.5581524575877747,0.08872639274935268,0.8003208471266031,0.2044314297342893,
 0.5415345187583948,0.8880278885998167,0.06379889678705997,0.3846751646611971,
 0.523443370605145,0.09870141901460139,0.6070696096467415,0.20805023817192458,
 0.3091362511047946,0.6087362515957842,0.09275073391615363,0.564053609948055,
 0.02394554048712383,0.4936864650907147,0.47079461838777115,
 0.26057600367913003,0.6487289884692007,0.26451352134076567,
 0.5165518015950636,0.05680443168175631,0.49480956692536926,
 0.0071516384974759095,0.07444562085485473,0.13565852000954737,
 0.6633514113908041,0.12744793907508656,0.7027474325978834,0.9161233370857262,
 0.5809257075404535,0.36711087950987564,0.21376042893342773,
 0.35570066185147087,0.17831998605744592,0.17571872852951087,
 0.43333564937114677,0.3890589852435059,0.49605114964679187,0.768726749377403,
 0.486761186240519,0.949044278838036,0.6429384859961539,0.20512305328493663,
 0.8003978022067222,0.7775711919846787,0.7740390622835509,0.08735506406248827,
 0.2633063532114558,0.3997927934279417,0.17335033830651014,
 0.32168893927186326,0.23433972958082794,0.1199558519474444,
 0.3783295476121258,0.503915118290481,0.9626937623906816,0.3368795878880274,
 0.05484227217877213,0.5312728007006644,0.8710808663002391,
 0.11016979163714946,0.590928801111807,0.2646699867916158,0.22523771295248407,
 0.6197138447580521,0.924938482812252,0.7794084362168899,0.8047426236753515,
 0.9987683710388767,0.09895245318944379,0.7653995052736968,
 0.47021999217635224,0.07340273596986269,0.18309722816028828,
 0.6108430114994042,0.4108832417679493,0.010463422266037359,
 0.38610872825248677,0.5420054368649937,0.1964612830338408,0.5683618683114738,
 0.591629300748345,0.870533415485335,0.17767344615995295,0.30686202269559726,
 0.049957912198021814,0.7279255136343009,0.6714146612477995,
 0.06888638316119522,0.27649359546717955,0.841570765767222,0.5632434466498899,
 0.0235688195018815,0.9817263685446191,0.9423458397455142,0.9014274686892699,
 0.5237603663257979,0.42735064484852026,0.5371351227070408,0.8877014815498838,
 0.5100322484901217,0.9503682966911307,0.7926104221748076,0.1389308952763164,
 0.9240473996470447,0.9394831614172791,0.3476383370356786,0.08143475123126453,
 0.2813895639316344,0.4353768512272921,0.018904750743197307,
 0.7696892899738914,0.14460042898774295,0.7443851878942679,
 0.18391411115359912,0.7694293408276064,0.8407274761903464,
 0.047858872168123234,0.3796988860865297,0.2734146491142294,
 0.7248483893804997,0.28530183087711924,0.5216787620002087,0.5470486980171084,
 0.9448161971150872,0.5013430177798431,0.9221352951145928,0.8759586997509557,
 0.4480346271457031,0.7740471975261294,0.9881013295920505,0.072868233039875,
 0.642038474280641,0.8239913646678342,0.511504475154484,0.2974140578906027,
 0.888218163975244,0.012736030899141326,0.8478529739224862,
 0.22510219519809227,0.32542967968620673,0.49985376666794135,
 0.38107401094720195,0.7086011811151721,0.24107830865822488,
 0.20103927419847467,0.4646309205688629,0.7259426170852351,0.3237619680120909,
 0.8421248133916325,0.5639788215625283,0.5732330735044314,0.3835269744315024,
 0.8800283188151641,0.6388354343695996,0.28019996169473815,0.767047544694744,
 0.7086040313592318,0.8898675235162292,0.18107816065898152,0.5081431070116752,
 0.6063572133573227,0.5035075787632499,0.6396680272595817,0.3442852702353272,
 0.5071152834145211,0.7180992722266197,0.8850683973116215,0.9735823098993291,
 0.7319535134476537,0.44127515907454085,0.25882011472765065,
 0.6067220037963142,0.7323385928793191,0.014295110720494275,0.345897341834264,
 0.6298755848498068,0.8719908761861883,0.789528879278798,0.26013382331249946,
 0.8443431966173179,0.2974828161318882,0.5258407527194595,0.7742416597804535,
 0.9248371469346907,0.8349230594916413,0.5486620009536496,0.8025757324260534,
 0.25333460042046274,0.16915882711279762,0.3383951763136429,
 0.4030160025760756,0.30489051008104495,0.45483792778251686,
 0.20170026343542324,0.28481666763394564,0.7516305613969437,
 0.7469619391671226,0.004447812411408569,0.356722109248934,
 0.08216017471207371,0.24893500466007568,0.7211784575242944,
 0.1398381451218471,0.3270621393685632,0.8449304081065071,0.3637229598729008,
 0.39599830280726633,0.7936334184014964,0.043912078729697046,
 0.6390902561121188,0.4149202657121158,0.5948854337935265,0.9542407403804206,
 0.553631186954266,0.5335641849694274,0.3865621825062425,0.3340667934231324,
 0.4720081995385137,0.06455605666082498,0.6443629683837198,0.5570614415752412,
 0.6775074291967798,0.6693535908959316,0.9501374122088658,0.5147647889227258,
 0.8996680975077039,0.7113914013360285,0.5268517420330152,0.08365655690706642,
 0.6429892711637921,0.9799260599810992,0.040333988758251005,
 0.9245334320310813,0.8327097126305223,0.30567655206001787,
 0.09770522938963466,0.7689741547143809,0.6890462131886445,
 0.30316742633068083,0.609695459705528,0.14056411257750512,0.7510464339723236,
 0.46990771221263217,0.7797172827288827,0.5370612026333579,
 0.25654679113625656,0.2469315268234431,0.6899624546230672,0.2289546335867837,
 0.2350980354853226,0.071163221974888,0.6881497143461522,0.769547847215426,
 0.3194510461314388,0.1395531449665336,0.8280308578354233,0.04028340627147442,
 0.5574481908439292,0.7706670519983373,0.946137433590069,0.7385061650909358,
 0.8040504610930308,0.7696331725208438,0.9197116280601555,0.8453259660474102,
 0.13365423191558012,0.3838196516580862,0.5986926005168993,
 0.004787969998294228,0.6743911598741855,0.1361666065483823,
 0.12186977040235392,0.7943867242117839,0.8093047359482545,0.920127512417594,
 0.8257138842726265,0.38085019383463103,0.522636669635053,0.5647805059799204,
 0.15662130234691962,0.5089109720388458,0.07044552155698203,
 0.8035961310509735,0.014121245873385924,0.6819738816393255,
 0.01780664435882806,0.45293121485228705,0.648411490434443,0.7446631543527685,
 0.7326818843231512,0.9357025961127524,0.08953633920983817,0.9690983090908678,
 0.18522826585050112,0.5441787690486211,0.7779649794985057,
 0.34459507228204156,0.8978319893777597,0.25450074946875423,
 0.5095745021871597,0.6987317618840296,0.7357617057447675,0.04889048510185445,
 0.7945549856450971,0.8639255880464769,0.6029036596697532,0.8076780264687465,
 0.12846549247727979,0.012281376828607526,0.5595239588156655,
 0.5498204284166928,0.9482991986698888,0.1413900793117736,0.22848552660614263,
 0.8894543703019977,0.838737775813209,0.953214954757077,0.261138976220877,
 0.5868544053948322,0.4788974988851389,0.5577881332634127,0.47804655084063885,
 0.839986027854966,0.10172728584318391,0.1600690680111151,0.12312700786747854,
 0.6513533574459784,0.4356637297861854,0.9459978681519219,0.3271156494558932,
 0.8464386026981804,0.8422999696561569,0.6927677241574878,0.57406299749717,
 0.8008225495900039,0.2031807699169238,0.21791133032389742,0.7261682798981962,
 0.4729689231598937,0.34336216145148757,0.3203910733071752,
 0.25486937874362614,0.41264923383006535,0.6066855822452589,
 0.3179855421489116,0.571300272539668,0.604355373729085,0.12758926348972743,
 0.4872387005801373,0.0016239539253596336,0.461568446868701,
 0.24650646310965307,0.4550339883433068,0.9540517593583608,0.5644583447686025,
 0.7245224979375593,0.043052171881357904,0.7616388379322906,
 0.4785215101199046,0.0682966230630706,0.20707191814948955,0.682598947242131,
 0.5960378658682937,0.9906209555190351,0.43067113761346354,0.5020024059255014,
 0.8136383676987818,0.960963030307745,0.7288784565371665,0.5909493654950598,
 0.9958925000982048,0.7544451270638736,0.05061238665994494,
 0.06815235825480404,0.7531741445386713,0.1301269278498528,0.7299767674420221,
 0.18240229779819706,0.07390333192243526,0.36596248840137946,
 0.055620709901511134,0.029130592581728676,0.026297320211473907,
 0.7828629523910355,0.42566071274002293,0.10008638291360228,0.845390883337168,
 0.8536520602926251,0.41650055096737837,0.8523488685250378,0.6254508025893729,
 0.13791402902785188,0.43376803537066355,0.3667801292305566,
 0.14699586950218468,0.9151603768626019,0.8473835084830419,0.2761091380741615,
 0.5801950493569712,0.33819168702427094,0.78884154031876,0.9863101502116784,
 0.3475986679896167,0.521768745630574,0.4772503265199143,0.7996888835834344,
 0.8998121600158318,0.8363487982075629,0.7586606545062728,0.4688467053747525,
 0.18497831239993168,0.6330980750831,0.733262550329244,0.5278890086285213,
 0.3621132665639252,0.5312678336769387,0.5306320146666007,0.6605233506950113,
 0.518110574211767,0.7852981669149481,0.8201995767088502,0.17241548036167387,
 0.05683741146626953,0.029689171380314683,0.698281617975667,
 0.5761483211846682,0.8670633765306884,0.003680658116532065,
 0.7959354316237225,0.2812429125841268,0.587797044035949,0.24625459255301607,
 0.18962904555294635,0.6722483365404827,0.8779942773253521,0.6394289588796837,
 0.15655053913183714,0.24925544808948497,0.9736531360385382,
 0.6414623056226543,0.867119642300876,0.4930488173924925,0.4644521087995539,
 0.060604231748538795,0.021993876394757916,0.13834627542380906,
 0.1152664372668829,0.4720665653822499,0.91480374447238,0.9330737316126414,
 0.8360616274359897,0.9335040917514352,0.802137497745097,0.8703406324067621,
 0.6876730721389284,0.14451564187164745,0.9895064780005266,
 0.20352105182207392,0.19701198120218832,0.6765348736650918,
 0.3072660859602293,0.17772846067975456,0.616619256703714,0.8394862320734343,
 0.13415842519440668,0.8024934943981096,0.8917692095157247,0.3550789119324417,
 0.6172181540056776,0.7905253297211823,0.11612240004318897,0.72087733897603,
 0.030547974874364003,0.9344434065864065,0.8271345556558606,
 0.29164194683178946,0.6110285789702771,0.5865224060645455,
 0.05198276998236717,0.25246840474526233,0.43469475885650366,
 0.4980175608945021,0.8493145955941378,0.43994242535184425,0.8593102595754384,
 0.14809965397706748,0.27495419019931366,0.07686474466036441,
 0.515607444424642,0.8051672392798694,0.6735032629568299,0.7542381382037596,
 0.10185832090036484,0.907185746331719,0.5052288359264728,0.9223985383719608,
 0.20483949753145625,0.20731051718834304,0.5896104511768949,
 0.24902536155931787,0.8004238497474053,0.5700816907226335,0.2556197622080786,
 0.027123998130798355,0.9735165334278021,0.836733719431135,0.9976408274842326,
 0.4234501741830855,0.4076639519471591,0.8939863522166877,0.12321150988358975,
 0.785369041057763,0.49394467174976686,0.5847962402686321,0.6568060919291793,
 0.9673961293051031,0.6109392253810705,0.4685854630182318,0.6897153446809883,
 0.8425248749143354,0.644782225536484,0.25039565761418736,0.5993069140942769,
 0.025842537557740797,0.2682233103819791,0.13027502471764385,
 0.5309880655745387,0.7160754833800073,0.7564174084860839,0.1935929306837687,
 0.035851669764128125,0.893557855814531,0.19170458109467414,
 0.4309623605410773,0.42309960915474143,0.5177519084072211,
 0.15697998230584576,0.9187186092195679,0.2665165600141194,
 0.48499152042349547,0.7446159444787619,0.9349540930802599,
 0.36155472562341473,0.6346806005038242,0.07154944600745838,
 0.6278373887010402,0.4306418469895109,0.802828881014153,0.25033161887674227,
 0.15689606006833778,0.4178520857513468,0.6945024067882598,0.5923471639338402,
 0.6906213504299022,0.41567451037329817,0.6126095394675561,0.9429512832993187,
 0.7348178835932533,0.9126306698076846,0.5026167517535718,0.4217226365149097,
 0.6456770239773063,0.9533962828542604,0.31199989740431555,
 0.032820456133104026,0.2937598708105389,0.4055927809525901,0.654104815484373,
 0.30178963323462993,0.112974303861623,0.9620260578351262,
 0.045308478234769356,0.4561222802115077,0.6848476410587638,
 0.08577125326141155,0.2007410017674195,0.4462352108890614,
 0.33483165477842436,0.452172890465265,0.5955997330887643,0.28352534536133644,
 0.9762451040185885,0.4453720984955194,0.8457215247478114,0.5096582612068224,
 0.025921789697789466,0.8659196164533037,0.1729922498952412,
 0.4331383194296562,0.8072487958094389,0.8749203856430637,0.6452346045393513,
 0.3333144941109285,0.06847319557378828,0.9848806487912022,0.6681983332427441,
 0.5724819755608573,0.18529451060189917,0.7463560604438946,0.8065465266664719,
 0.7737819600383009,0.7624152517963819,0.4587452929251067,0.8541743854368067,
 0.5876434183898327,0.5473886344773946,0.17970780880875958,0.8816103340704515,
 0.8887176544140795,0.8368515990118278,0.7140058117232696,0.7822978623328634,
 0.5176230188199257,0.873458482153779,0.09736215452986408,0.9145513486338228,
 0.8612973882048346,0.9267538171296087,0.9028616613027524,0.7504688310232389,
 0.40068677570312883,0.15163470203972762,0.5080241802453049,
 0.2551222876624851,0.8510731156353417,0.7335669251134875,0.10273796279938119,
 0.7699266253366046,0.5162572606029014,0.7165554846855257,0.583363829896741,
 0.721983071002777,0.6584874868760844,0.05882388922119697,0.49345711400548486,
 0.8210587943948338,0.875093778677244,0.25887722202886065,0.28094253412924675,
 0.3652176548908984,0.41079974388252793,0.6205226767436021,0.8277328705796702,
 0.9675331931620004,0.7781324813047957,0.9601346730207626,0.3097223495669895,
 0.10748307526458623,0.9049231172252845,0.7384818959470435,
 0.43410228433500264,0.5192306989486098,0.43869497995394746,0.235662065442946,
 0.8813916217702863,0.0370840504572898,0.701916606633277,0.9755704279579072,
 0.14630617631429965,0.88893571641679,0.5247837729957295,0.02274063235931323,
 0.16562962745728982,0.7083170524732242,0.11010763414195135,
 0.7626273053476726,0.7247544269762243,0.4261885161613067,0.31020577394796356,
 0.6021111074418019,0.16383258344797202,0.6718335234140707,0.7780964421341712,
 0.9185442056179224,0.16753281690344535,0.7286185795719533,
 0.24850812789572585,0.04661619589954824,0.2901443211275898,
 0.2758324870500599,0.015110494841744093,0.04425525881373127,
 0.763515843050708,0.8347493137911781,0.32381333548025415,0.9054588086537578,
 0.424106632014144,0.9461428930724739,0.7037570171288915,0.5538628475737323,
 0.3560231980836248,0.47718271840569026,0.6687667995051213,
 0.25415983890871396,0.3433794233029468,0.8820072158633324,0.2942509903128142,
 0.7372611249026153,0.15484620364066004,0.33169739167270995,
 0.7522470232152773,0.32541200730684827,0.5666423077386182,
 0.04333037399321937,0.4358871914107112,0.9855489246648548,
 0.32227422899666514,0.00830050253809167,0.6030049338612128,0.432858736312008,
 0.422306907778384,0.5444640265141889,0.681670803422385,0.22799826368086762,
 0.31861418654678775,0.2468576303636849,0.5824542423522401,0.8230633864875464,
 0.7391811080013619,0.9356231100773582,0.02756595320829447,0.6254663176833776,
 0.6056044754972666,0.41174811766164154,0.11974826482822531,
 0.7561846625860162,0.6459264523327173,0.7301376676328679,0.7560674719095671,
 0.23950915270888207,0.6907393694985665,0.46857576814845503,
 0.4845432466706109,0.8550949889848897,0.7716512705810231,0.13085260263666,
 0.9844379308193708,0.23830161842852104,0.06402057735890643,
 0.8322166508514052,0.8358633975198627,0.22424846200725612,0.5566317486809295,
 0.6418925435991537,0.17490306876258876,0.1523043931319017,0.832053479040867,
 0.2626271502935925,0.2225947462336607,0.5515286708054217,0.01950555828053091,
 0.5120578164643393,0.7421931568751623,0.3810801055616594,0.15544855249792822,
 0.9124955551653832,0.24764398161179701,0.43069704083346527,
 0.7212945446501811,0.15816067096778907,0.34761037538230377,
 0.5165660643146721,0.3736991413889592,0.6354097417259814,0.8701054661898352,
 0.8255248947333466,0.07242135386680137,0.8334157366805368,0.7264517913936283,
 0.008245198863057523,0.4809621630939941,0.4747984104266768,
 0.12200403229849588,0.8427724584235166,0.30264494852243584,
 0.04132020040016682,0.14709968955468655,0.27836147681690593,
 0.375132909962432,0.7329801124980881,0.5033678199939841,0.003955002895301707,
 0.2676103265158355,0.15672944840795922,0.28343180967392767,
 0.15155073334471192,0.629754314636014,0.6582899009345236,0.760840999444153,
 0.11878841197747336,0.5174446500378749,0.3159990835386053,0.5910030602773491,
 0.43755048571462274,0.9278286501076739,0.8184509622822992,0.95524473262517,
 0.7407640560828053,0.04300759330547943,0.07303257888642212,
 0.6274985830909298,0.16120401504281456,0.4545429296537786,0.9729665457492903,
 0.15371349934112177,0.7152607199844719,0.26973547210316384,
 0.8513721414221969,0.469447292221744,0.8384727482750629,0.002647046971804068,
 0.06437469099361359,0.4070058182103601,0.9658677984822762,
 0.15649561942619505,0.33506135152770267,0.7636589693851185,
 0.5584769085332281,0.7738694746434783,0.572829513135611,0.9302086738643736,
 0.4746092440120937,0.21914481831508392,0.6545488839628044,0.702367716299526,
 0.9212918303649863,0.9707122185812448,0.28519114442450133,0.5269017353348008,
 0.8470030948418705,0.39612189090442407,0.6898570942745176,0.8800469846292958,
 0.19555436249283353,0.1813100673931365,0.8995877498483107,
 0.18271538955780653,0.9764148968017349,0.8702760437746951,
 0.09887343851753494,0.5546715610656792,0.702690408829036,0.2753922437029933,
 0.4280874609477061,0.3778984251128338,0.648879160603282,0.6175773134984024,
 0.551345542694236,0.04300284110398778,0.16925914445211443,0.1843292704331766,
 0.7610433050298997,0.21824894654797033,0.8539089946117997,0.6654095833799264,
 0.7071425563143804,0.5231295464978754,0.9369710627532569,0.36199231325438463,
 0.35819888014934076,0.3828111592429124,0.47237137629734816,0.642296604947313,
 0.2665663811034451,0.038373669125917886,0.26451461182046854,
 0.13117746679757258,0.060560767095518386,0.49625700777833437,
 0.1202443609405367,0.4727505953721023,0.8235199181200104,0.2987005314356539,
 0.5153928932065409,0.746872069738057,0.6894133699636041,0.6395363409871683,
 0.6020969937219706,0.2854447910352097,0.44202425634111,0.1741766023353094,
 0.6073304324180238,0.4024936475947838,0.825024715315657,0.09793611761939569,
 0.999104498623861,0.6146245758080982,0.7303872743983292,0.478653470293671,
 0.8821915990512684,0.815679010873211,0.8555850911239585,0.15634198782495468,
 0.9637342698686022,0.03157102166670034,0.6250465402203748,0.7169075875722698,
 0.5124032906746377,0.7870679455782482,0.878142975360441,0.8156906347080065,
 0.4971519230195338,0.41536222746062545,0.8357181524592104,0.6911498812026501,
 0.9968990926253548,0.35032865730647433,0.46502666029359796,
 0.5273114806729327,0.783697255463989,0.8221023520991909,0.47346846951590293,
 0.642115981009228,0.6066209557520763,0.5647015477342873,0.9465414860319135,
 0.8924303178303526,0.9117986228587996,0.5611859221250671,0.7357288661789572,
 0.3087865429294322,0.986782988111245,0.7116452319217971,0.10851277799336101,
 0.08013185240802301,0.8237808552161073,0.08173455998925683,
 0.8482693007825753,0.6047069287657432,0.9715769511683013,0.6330887028709082,
 0.8704689228774718,0.8657632773455926,0.893307401689649,0.36154346998826203,
 0.7605025847691503,0.06219902263348287,0.2443206402245569,0.7914431233172938,
 0.36198377133779047,0.1536097775087667,0.1596268173935378,
 0.24804368312337022,0.7458842963744536,0.986378699575125,0.5957363301166226,
 0.7664334666154491,0.5917350109904789,0.4136839266602478,0.32760519046294856,
 0.808602636776948,0.07059975397199991,0.13867686174372496,0.3310990867010941,
 0.4071231974068903,0.4833505645979508,0.29003320397532373,0.829519231670993,
 0.4499812840031825,0.6645131692384276,0.8378710793561391,0.13438585562641248,
 0.40544336676015225,0.26942439360101333,0.7819016214480802,0.601171165428952,
 0.9456366280324309,0.7113242794818074,0.6510628686748858,0.7017308269847914,
 0.787724755792896,0.5611963530111462,0.7143874826984037,0.5221590106924672,
 0.6896290460010347,0.284403419203071,0.19569647731883333,0.6135577100779339,
 0.3105326341326271,0.90202497166955,0.6311176098752638,0.9048345959131392,
 0.18441844872080104,0.7136641627927544,0.9298618953982002,0.6452948032598891,
 0.37238938232098406,0.12865037577224803,0.41953537296905885,
 0.8170604091046312,0.832398439423275,0.6298018451744093,0.785917335067463,
 0.6239986190501554,0.8944535189082359,0.30263786115872293,0.5791779277527632,
 0.4110822225372607,0.11137775198052702,0.30044218443156934,
 0.6752630421176459,0.14777518239826337,0.30129814327084914,
 0.6858149737155496,0.8035593800933063,0.42796292690112037,0.8163678198314539,
 0.6285981106057248,0.4885013500434441,0.012218326312372407,
 0.15808884940654966,0.1816490469321802,0.5339558896304033,0.9063552194149764,
 0.28109226406824117,0.5025469339298796,0.28664844804374745,
 0.49463416362644974,0.2198260510100366,0.47212287362809624,
 0.6950297255365449,0.6973242149438099,0.6200072798869566,0.01916496418035263,
 0.1971511884011392,0.3611027492147586,0.30185919914494097,0.861724336522019,
 0.8697894312257579,0.6346769591623882,0.2367208000202552,0.7371488697569066,
 0.5928455361590481,0.48036510221665196,0.6635074936454668,0.3394789528601916,
 0.7703466684072507,0.7295006285596536,0.16309130780605297,0.7414914991476618,
 0.9961323155811718,0.35129751791860353,0.7407234223307472,0.609629895163046,
 0.42430193518370485,0.8382795516990432,0.9545614244942743,0.8579894126039268,
 0.89891180274229,0.3510714231368717,0.1626113907399489,0.9681230700270269,
 0.2513584875778798,0.345131069768444,0.9222014951391402,0.957298553314095,
 0.139725719458052,0.9180901321458591,0.675005770616724,0.5749800357431528,
 0.3185224101537607,0.35048810952159026,0.4016152663652407,
 0.39555497300973674,0.6743658817485874,0.688263845958391,0.16094556745917443,
 0.4702187917871907,0.4392339871770957,0.24606167716812188,0.8639973908772695,
 0.6579407054613637,0.609613812957817,0.04938862829177304,0.5053867830628965,
 0.7028540880775596,0.30181128460473794,0.08832829354140892,
 0.04755565459311206,0.831678181268513,0.41884383974196193,
 0.12716725400821582,0.46864375060852503,0.4483212054090291,
 0.8301838311160118,0.9953867448795712,0.30393192007473,0.03478299504023474,
 0.8695575092861033,0.9727452052243564,0.27217450599673876,
 0.22327177780438856,0.42324866624749347,0.9884439311170334,
 0.008812842818628819,0.5213075006588485,0.8981538977031442,
 0.3892203799411955,0.4855496770369806,0.5428946654724986,0.11126507673825725,
 0.13035768092223932,0.004201938310797182,0.21927871196304516,
 0.8792521893439509,0.6361225677983566,0.7537210330937921,0.5887047203736303,
 0.16714833601955092,0.15040329405017827,0.5089621773401095,
 0.3217176763201809,0.2380551486880742,0.9421759490345374,0.7910036014407152,
 0.10976535263215237,0.3378489268871293,0.6734763714149244,0.9834012999985341,
 0.15400710557033304,0.45752300715920946,0.30137879136651113,
 0.38551488492515273,0.6528737047388987,0.049862135951577735,
 0.6093199951078387,0.4438779615417361,0.21390531814360136,0.8658558399024789,
 0.6072706948902319,0.2355516223789087,0.55260101801466,0.6742079124495522,
 0.6711567466439097,0.9701897300206033,0.7873683556241569,0.1708579038004291,
 0.6034605803211691,0.8739467191864916,0.2608768004409915,0.6593058630192579,
 0.008116018445652462,0.684681417208382,0.5095396358646022,0.7061516086227745,
 0.35540545804499657,0.568021355609889,0.6609040402009739,0.8602098112383281,
 0.3190781459862252,0.6178305387112886,0.5783797059153759,0.48193661848423774,
 0.941615875102865,0.7684167202113439,0.31932658805948266,0.8866076721984729,
 0.9752795539871175,0.9604602899503383,0.9334796716343537,0.25847059296491526,
 0.7999715017940168,0.00990404675320955,0.4961261341759917,0.5501150712961361,
 0.4332132905779533,0.982808893214698,0.6274868962080842,0.8712170721436852,
 0.4753323357013781,0.5450139323635373,0.12382249083839025,0.2242939959213145,
 0.11279640410707115,0.4123779251463854,0.9518937493362097,
 0.026655611147323843,0.8016439616398898,0.6511476167319644,
 0.28171989294689126,0.9305068914906356,0.7167214697751387,0.9799615563020521,
 0.8283596415437805,0.03445887713459328,0.35536483188435897,
 0.10385601594961835,0.11770236155148117,0.6361884987117978,
 0.7757987088906815,0.884939476454309,0.24642530682307073,
 0.028366145608358817,0.9130733764254284,0.5484753344347884,
 0.9032386142503352,0.6555447042673428,0.705189879521562,0.2431800695505767,
 0.1514624683784035,0.9872663226730386,0.6922497171577673,0.7557182350636291,
 0.16683886777164325,0.04721636610829627,0.06839127427702874,
 0.7698726123086193,0.29169099771016016,0.5128281414998758,
 0.18260566412647272,0.26008296303370293,0.03928727178739988,
 0.18106899854792158,0.4755535315501639,0.5344963875953137,
 0.0036991901277181327,0.26130872040300945,0.9168302599464604,
 0.0625838002891489,0.12481316430132672,0.08915048513547763,
 0.8003372787029212,0.4278110235979492,0.034479578083537854,
 0.7277051965127131,0.5566388886800604,0.42742333511447983,0.7266357663758355,
 0.361331134887961,0.031498022155940886,0.4230886680218492,0.865401640381189,
 0.5889016515657001,0.37196419416564774,0.01461188085595655,
 0.8365618089596072,0.7510254328391841,0.7795732898340368,0.5377721641072531,
 0.3305052304428727,0.15468370759893035,0.6624601353496604,0.6393624708742165,
 0.07068281819459588,0.6716295035928981,0.7939314534896277,0.6298691397580214,
 0.670712712692598,0.5058140857974271,0.2967604133024109,0.8063363286915695,
 0.9417593425253938,0.8267869284392194,0.07893128775062119,0.7770798644417224,
 0.13688280676827946,0.8912606929919255,0.6504585612789313,0.9546626049564408,
 0.48663951294165053,0.46691583069009657,0.9052666215231997,
 0.3503138398590526,0.39073867678158547,0.12914936758780693,
 0.4386232023331058,0.25927513487765486,0.5069455794699814,
 0.44639837779936264,0.6975264213812844,0.464138587448067,0.280383991144153,
 0.3240897913308619,0.9349480417702117,0.7535549370759015,0.2230491305495027,
 0.6980281659492836,0.005710089837737131,0.7274708059325136,
 0.8708841934728729,0.6490502377570233,0.047382610892337085,
 0.38040310728338933,0.1130117305451177,0.8328259678724526,
 0.12284698100675517,0.141973373639949,0.13746686583519185,0.1389109923688432,
 0.07407360521587159,0.7379654205526003,0.9434882784405005,
 0.14078194631330532,0.5097298621247461,0.37040452777414523,
 0.6606144377944403,0.9315744235037031,0.9707809513677081,0.8338976130017373,
 0.7909525824380871,0.6828489657122089,0.8994795638863282,0.15267859418355778,
 0.31106207624222193,0.314819588650866,0.5400983750324491,0.07049920071306459,
 0.910965482698474,0.8691947553632621,0.5700750938678817,0.8686430946714072,
 0.9531107800209577,0.2941231951012193,0.007436082856778992,
 0.3496418125337747,0.4818719982669739,0.6781631946671192,0.1179682249245867,
 0.4498737020234075,0.4839575597997502,0.7050643578963888,0.4987079224721991,
 0.2259638612957977,0.0955975722850636,0.48865805888439484,0.689396014646237,
 0.10116286381689055,0.025075630967759333,0.2850770040829407,
 0.5603213387524342,0.4621891253572429,0.28205363015391327,0.8952477263322033,
 0.7243577224769563,0.11293728367696343,0.9104248188441498,0.5132956513487147,
 0.3220480417160718,0.8656450944752995,0.4230433063722846,0.5357852271603831,
 0.8363850951883225,0.614439631042618,0.39248973267005416,0.6223532384295688,
 0.4970082940326035,0.9412973946831307,0.6745804436704798,0.04530152196087445,
 0.0657562411985424,0.025672927975340598,0.18752424892988506,
 0.9671009281991442,0.0581208613740124,0.5262842484129249,0.18712272524057838,
 0.6125790279178526,0.306035876989163,0.6599393967969265,0.7967012980839605,
 0.5304008388830168,0.1654715331074661,0.2060707151684733,0.540309196282829,
 0.16409822580415223,0.583960931773262,0.4055879134015432,0.6740294860293572,
 0.17471211245637874,0.4907312881055601,0.5082820029982171,0.8127793114008732,
 0.20977643758558573,0.5419977056916906,0.7855757865277716,
 0.45196034076458513,0.5955010258751902,0.06632041585080894,
 0.8303545789385546,0.4526229051959748,0.6150261376356134,0.3987313176467321,
 0.10846125489873704,0.724636626249205,0.6461807043982759,0.6428252237929681,
 0.24310513979117532,0.8973123461126911,0.6686108727910627,0.5930971537343764,
 0.06769020871598919,0.46716078999842703,0.012673081158923072,
 0.9662995576286908,0.13295400699471027,0.5686896647508863,
 0.18008368370067118,0.2912607529014368,0.6500028316195169,0.3866474735613412,
 0.7033876168345691,0.3797846806624823,0.5336176405659881,0.9600102774513981,
 0.506271833680476,0.7688801919878969,0.8811936340450651,0.1939474005794788,
 0.4796311845070257,0.4590906321544713,0.07241865550094795,
 0.48221782469937113,0.6282370869073028,0.28105002021821435,
 0.009467180078926773,0.5290492041349079,0.5361599497413946,
 0.8083612382166587,0.3212717347847115,0.8614917652721404,0.5281561134088679,
 0.3270628956833671,0.3485635081415148,0.7918708159825316,0.2511096230350779,
 0.2506193317728562,0.8499560348165617,0.9395608285784981,0.7165152682595106,
 0.8300215372028823,0.8389074960062176,0.8070345979042741,0.47406840989277255,
 0.30336416933773214,0.3601721361884316,0.9074147127685135,0.857833650572434,
 0.43655026869191227,0.5746135295571071,0.4938346816605751,0.1697055203206147,
 0.9696326158720077,0.7511487408175154,0.5988822984986365,0.16533345573144287,
 0.9360551789998348,0.8042347632654899,0.35799296547974224,0.3589952793090193,
 0.6508549876365597,0.07370685831558976,0.6209199065651007,
 0.38241645393976775,0.7832560270053094,0.9375894238678165,
 0.10779687704077101,0.6528225371078761,0.7282852500871186,
 0.005531749018105714,0.16852724976095868,0.6337372232281251,
 0.13932494147760754,0.3682435344616607,0.15571284728356483,
 0.7032511067823753,0.4266942445654447,0.5069803572670482,0.9522535901683294,
 0.1941208821573246,0.36777865658810327,0.004263839292929283,
 0.978680916467374,0.9914892387105602,0.6106062705837444,0.5385749972115584,
 0.8335199435470219,0.9374273501595255,0.26570721167422495,0.9314530530269329,
 0.5130225098859966,0.2596083669021665,0.16881919965446412,0.7208139047411911,
 0.5705852463400889,0.5784646899021176,0.9012002223816765,0.53528417718464,
 0.042024108509387315,0.8531496513847487,0.8938929437604218,
 0.03477596375978709,0.11493961915374407,0.05437381310906597,
 0.7574659063797357,0.6534878454223308,0.2854132168625363,0.44512136946263403,
 0.01346079539139955,0.6120017937537918,0.3261940453480576,
 0.27306545678747385,0.4301584179089807,0.38227804930748543,
 0.04858702468433984,0.8967933981634401,0.31007648474692373,
 0.9222234883434521,0.12888233415852157,0.9188408183981441,0.6552070391281974,
 0.537798624083471,0.5515855242985442,0.2445043085036227,0.29849350950622233,
 0.8394015181472557,0.9981947293591982,0.3997174320077357,0.7435004177136353,
 0.1691109191942496,0.14975177980724153,0.26664713865043943,
 0.1877678632373747,0.0423882382197579,0.001121525629961151,0.756025618925563,
 0.03550776396002253,0.8327608960760642,0.046733093097196976,
 0.6474452813906867,0.2996367440118842,0.8424258186237633,
 0.041626870360242485,0.17488495795385772,0.2189314836353321,
 0.5684759761111888,0.09026612922801336,0.21940430790856247,
 0.10733335749321771,0.6904800179328101,0.5716753307133469,
 0.06144773008921256,0.41416483136101345,0.8939230852968033,
 0.8095338624117725,0.1490114132172623,0.6293613224716392,0.6712298714008479,
 0.46797240753768476,0.0780493754151691,0.4885599527259975,0.6046653286573573,
 0.5413875081125903,0.7780303845454242,0.4226744809312911,0.08553827928604174,
 0.9163587729856795,0.8122605918707118,0.13201320602352207,
 0.15535084295960566,0.6446372325173807,0.8473364486924371,
 0.10111504040046482,0.1950975540272002,0.3164496530896004,0.4824140522183362,
 0.02983278440394277,0.992227921694932,0.40295050436091595,
 0.05423192908467245,0.9416007610889167,0.21387835979829328,
 0.45503215115701545,0.8759494588989981,0.15333790736874375,
 0.6311341091515956,0.6819669493779341,0.4515398956725587,0.6732721744290984,
 0.1615593176888095,0.08885163662177564,0.20871899763020818,0.356128365119172,
 0.6834663446506988,0.2588014807664476,0.3179426861738879,0.7651790860746922,
 0.6540993156202037,0.2670036812703347,0.5492738587447413,
 0.057808228586300636,0.5943173917960577,0.20875527813830475,
 0.6966631475183255,0.8593197513681361,0.7741729619284649,0.2506044110854554,
 0.34012118599776353,0.6221593112768383,0.29299559572032774,
 0.7168546705894945,0.1522719254665672,0.2195324656023675,0.18209173037083137,
 0.09448715184963596,0.4453197197781791,0.05355918293354933,
 0.8092337709067359,0.32635327070362496,0.5066667747334794,0.7583647760999246,
 0.3275164940060049,0.18372600240679393,0.05373381158693036,0.923168594848315,
 0.18877358297269897,0.5485369135146829,0.9626688870270167,
 0.23694353271899204,0.17562863962839792,0.7742106087022216,
 0.3883673226716118,0.5621820672902161,0.3851818036470227,0.8219081363314708,
 0.04690647392931313,0.4760537324516978,0.11891754805577093,
 0.5560207701746274,0.6489467352724672,0.09735636221280042,0.8817013756403803,
 0.7425460883940623,0.6717956179093847,0.048446319804025806,
 0.0019246971341705388,0.13120988964290758,0.9980057405532446,
 0.5053052822840024,0.7058972804296367,0.2733770504930888,0.9219057870624893,
 0.4644945901957649,0.25131546187657317,0.2267540997500418,0.6834834659883469,
 0.3493117727681445,0.4086460177589637,0.3380810127625479,0.537797888939628,
 0.5908702072323814,0.24777770736620086,0.616292909030849,0.2614575670802932,
 0.8568528293829372,0.6505206443964052,0.1869636343232538,0.4780723213238224,
 0.23249347759724037,0.513587661245258,0.33430228093856773,0.8817638220704543,
 0.5371602303940219,0.7729800371204301,0.8556752286419075,0.9860898862161779,
 0.8328040311121019,0.042375477753215596,0.6992719555006238,
 0.6640439227021582,0.6724841533052703,0.028980712583228763,
 0.27650008619064825,0.31557921906227504,0.8971027229979744,
 0.0038424822504586897,0.5489090600724367,0.907958064226853,
 0.7820201197552601,0.1744032739831839,0.9361392051005563,0.6103426900381923,
 0.17269846751668805,0.21162638316179838,0.4515335024020922,
 0.10163696788344145,0.7353494495983306,0.23634772767360213,
 0.6445417646450381,0.9508830258687953,0.5849892015148113,0.37976264220046185,
 0.6298829034685657,0.2985242116195763,0.8408232482030853,0.9691586699712357,
 0.8563272169259897,0.5729558938817375,0.433947176521822,0.6624742396989596,
 0.8757996432271998,0.6015555744610644,0.4633828728276156,
 0.003824493590630751,0.15943711785596282,0.9534257715581732,
 0.934779300656685,0.507865410579211,0.4083319680187306,0.7168408532402804,
 0.18170511237877762,0.829404248739,0.27494571077723395,0.45225207685052493,
 0.3345484335733838,0.6659525475432015,0.4490536245750111,0.5830247313130273,
 0.5530566198679079,0.16698964441822006,0.6909519019370919,0.7186140126474863,
 0.7931403278759583,0.2029150526882496,0.25642343064780915,0.8730110881999259,
 0.5014581896814532,0.3248892900800977,0.26418433478468817,0.0983034566828519,
 0.24575420836374673,0.26449330617391753,0.6442784082574594,
 0.6233071590695052,0.738328225661427,0.8322141575492816,0.8257633784989423,
 0.07903845019156508,0.7945290198811825,0.991808645572904,0.37627451154905567,
 0.677938530914464,0.3267302134605917,0.18550579843992776,0.18863307574879162,
 0.2149074681124341,0.524528085581181,0.4000525281814542,0.8115770935007038,
 0.6315142414427944,0.1731143385597388,0.1498355291782162,0.5484036320438113,
 0.7976462163926801,0.7659811065116323,0.49871762627635874,0.7717667152599184,
 0.339140513831795,0.15716337695862415,0.6404453790869921,0.0825556074527879,
 0.7825563564639268,0.46617739068267205,0.7715767960714763,0.9503544073953202,
 0.9850739761010915,0.7648401397858262,0.39881328751050815,
 0.03227164234100066,0.7855437172220393,0.3604196454693991,0.5113292459766706,
 0.03097135078610269,0.7333129411165454,0.5253031482885332,0.6836016120503708,
 0.7348716238020641,0.8562613328385233,0.6705472434989054,0.5897669384128944,
 0.10288869114216947,0.4837248473337039,0.29895521762895727,
 0.6309340809304689,0.23650495386476722,0.5359110910095686,
 0.19555063775668147,0.5953616645316846,0.39664141268597464,
 0.24985095036687566,0.7318039967524292,0.7506505839368485,0.7426960578369162,
 0.8793796714302233,0.9813283847130018,0.7393982011516159,0.40823270623482755,
 0.20833232877021546,0.5040364846487653,0.5519434520721676,0.188032240410543,
 0.7674523189799244,0.9240583601451956,0.16298336288952453,
 0.48178846995231617,0.2585890344565711,0.5353029364004098,
 0.11181309242531001,0.06596430631042283,0.7072222719448565,
 0.20348288355878952,0.17476215242315984,0.7756983407976733,
 0.6187806788351653,0.8293422695354721,0.9712963878396638,0.3394457597672401,
 0.4679674964759847,0.9723319566778547,0.41339847578963174,
 0.42263068142566096,0.8724473961809098,0.2072314093404638,0.3230417187958071,
 0.42057128052916304,0.9264384630450724,0.8293580164101739,0.9683133530646933,
 0.026797051290806673,0.0011956251941973584,0.4445897402407002,
 0.6344645856782986,0.5131257352539417,0.08926977201820374,0.6740026352333891,
 0.17026080201472693,0.002843153759410244,0.5018307401177658,
 0.8469246971369737,0.9233469396210796,0.6055101526353179,0.1955195281890758,
 0.5779422782979918,0.12188513118638189,0.17055688111752987,
 0.6294694363339037,0.18336940903772958,0.3963052818498939,0.7451065511933053,
 0.58791108613041,0.8088246221396247,0.22240428399590206,0.06420105949312405,
 0.5329198442837441,0.3228523778931146,0.14709939258904026,0.8628990332217377,
 0.3063754604698279,0.21919323137458202,0.8638569299517047,0.5935040391825875,
 0.15524521801598523,0.7600826265273628,0.2145275443196254,0.6045786935226674,
 0.6799259437797404,0.17788404999123797,0.6547770918533166,0.5066953466949888,
 0.7247077824982451,0.5319918228346322,0.4983631714509906,0.1534956670934734,
 0.02417261939785198,0.2455660539697746,0.7740527413375916,0.37739106239245,
 0.05759438019204599,0.3957529838967735,0.357179107993285,0.7109755200610804,
 0.015560202457611583,0.675932471118366,0.3808470275807584,0.9949948461630284,
 0.01659060584868155,0.06289594091179096,0.38140887948034474,
 0.6019702575870952,0.4196977893909436,0.16069806163296452,0.3545030282401106,
 0.6900498623360598,0.5537504615603195,0.37577598902855336,0.2838855795194217,
 0.15356584661350303,0.3747474701570195,0.0437614481098616,0.9036813534732692,
 0.5028708207421118,0.9984823384080334,0.6991458352718328,0.2215401702799784,
 0.3845626577416741,0.890597034681772,0.6710060870659423,0.7864502110791621,
 0.8510768995460471,0.7722090926014906,0.3641726982325042,0.23073337762114932,
 0.1348198761883319,0.6648818162290185,0.5711575418504119,0.6463026596031771,
 0.3819285061189942,0.7087497743348923,0.9486499690235347,0.4962971397572107,
 0.9685570253311757,0.21104591489493707,0.42393491122087434,
 0.9436321698019334,0.3697184631955135,0.5644707815836933,0.9412720749102689,
 0.9254680040991535,0.8468292384314673,0.8049829297923212,
 0.051609784780473844,0.4536858152215333,0.21147526170996123,
 0.7925012231886758,0.32378909551100354,0.42620701131510164,
 0.07862189020733623,0.2734288784529607,0.7953258186883317,
 0.10078662233078306,0.7449224073607789,0.419965028738818,0.07778303934303987,
 0.947864527862534,0.7509473315609207,0.5297951016084179,0.5153865091595534,
 0.7625326141552239,0.02846110380903899,0.718432044154171,0.996631978053257,
 0.5407027767245048,0.12149514117315308,0.030176388333909676,
 0.5044640258034574,0.6247834232393829,0.26390733592205073,0.1269224306713287,
 0.7206704753947663,0.7917702802262074,0.16930301875399867,
 0.11365160463074053,0.8104427236127252,0.4886698880227037,0.2952837677241632,
 0.16527822398628744,0.45111870855541325,0.13092655908659767,
 0.3509412815527688,0.42383565186993266,0.7413273930868375,
 0.24003707201995472,0.30959239810354366,0.9367172143957745,
 0.45014810735738797,0.12522492061450086,0.3715608462515543,
 0.7113258423803867,0.5967487867455687,0.6931680326249943,0.8411268829303281,
 0.9392085856286989,0.9797145486108082,0.39143120545773247,0.9702150027741049,
 0.18006552236841789,0.22576992737194912,0.6076639260139003,
 0.8280474115348333,0.3244679213452155,0.7349127120452332,0.6276179490358128,
 0.5649471860094987,0.8585732191223466,0.02064019711431664,
 0.13293102303779436,0.9762871930495751,0.4804866199397463,0.5247434729020632,
 0.1035886190869012,0.5595563433576007,0.25586789822601075,0.2029175938358312,
 0.7248061004782563,0.9595674389457851,0.4729971410323469,0.9681136424266064,
 0.7109435072300909,0.6975178730884387,0.41175375871466535,0.5612677276423904,
 0.6276572872455476,0.07702166690349821,0.9826708047545761,0.9969956071106598,
 0.5640706104467206,0.4027489323155673,0.335449126501413,0.619296835775016,
 0.18325050608545468,0.12135464457272538,0.7224886501749915,
 0.6293146014139136,0.03605733384129528,0.8191779134274437,0.1324929085937796,
 0.9962670032939759,0.852532559402289,0.46163028714943977,0.5322065713182789,
 0.787357003150251,0.026507434634432725,0.37936278022540615,
 0.07121068280841714,0.300635880744511,0.8686767020458966,0.9589034093712833,
 0.3251745357345651,0.532171341182918,0.47014512426018584,0.48292310675035754,
 0.4395070364101289,0.9587994725034297,0.8208387004121256,0.9482939204870872,
 0.5240856766637981,0.01657198064431764,0.5282900119273924,0.6461046940970864,
 0.8604665720253292,0.5274692162769454,0.9853608494689373,0.8732057599987859,
 0.40923462786676923,0.5332630628787458,0.3632918488301111,0.5188250408747883,
 0.5775037204474467,0.06462047628216783,0.8672079995439665,0.9687180064031352,
 0.3313598904553403,0.5310344497884292,0.36226411714824414,0.9167005915141367,
 0.8970264807166822,0.9260627415822287,0.27570151342423965,0.7657944437349605,
 0.08812633967402483,0.42733746768384084,0.615966916229941,0.8244194499170369,
 0.5354546775740364,0.10091712985862578,0.22033748023449984,
 0.6585763999574252,0.541248806377086,0.9884186764024027,0.10882901090026031,
 0.8710273814836201,0.4052169771761881,0.4532697581612233,0.32319668704410165,
 0.45881168295483066,0.26357222276851344,0.38708596236643045,
 0.026629876513222794,0.28432694453052787,0.9913904713522851,
 0.6295770502278294,0.9803088646997564,0.23147841990221973,
 0.18256135503876914,0.8431067199226628,0.6266098134796037,0.290186806182991,
 0.9852331500484302,0.9689184134169151,0.9112193829410163,0.281336627166899,
 0.5893523011542038,0.6927962301684917,0.14935759442790109,0.3788813722862324,
 0.34569130906684786,0.840270399241152,0.5950613018940349,0.5892058774827134,
 0.9737157920424321,0.44271723765811233,0.6521188932831772,0.20610685751391,
 0.07134768788639878,0.6915555757191263,0.5806740077355627,0.7298116047380514,
 0.34676934134453996,0.201118666905454,0.6504935121324334,0.4064160649927562,
 0.040737273714859246,0.41385446758691924,0.21897156428024467,
 0.5547975344185475,0.5621111730133622,0.189330066036015,0.3394784995396626,
 0.8282767551192968,0.6503959733891522,0.5863620620223684,0.665463735173438,
 0.6376068208048329,0.6531228455203859,0.9216520374749542,0.8381855079766781,
 0.0265570345890076,0.5453890004284181,0.7273705415979157,0.4500542300632542,
 0.5754127434717687,0.3802364864286967,0.7332864590721535,0.4193858194318324,
 0.10951116842184982,0.5580591746847895,0.3256795959024814,0.5463754020832056,
 0.4188870924425979,0.7132851181000031,0.05249715774575614,
 0.24373840885610198,0.6929155957985659,0.9440225549798604,0.7802289441949091,
 0.30824432118234335,0.4797453261783392,0.505828291513284,0.18648588847268266,
 0.7988377476350758,0.18329301467991743,0.5411869790089254,
 0.11445498997471537,0.7333561786289221,0.11605340825808153,
 0.6101221427951342,0.49668631108403083,0.7862162071354171,0.3129586325194976,
 0.4005711534913432,0.4890470581932844,0.2904371664834372,0.11186661047678825,
 0.0562385726642719,0.286624466848748,0.9435745932882209,0.697989264895146,
 0.522437439125157,0.6098910777957421,0.31350018922789724,0.16516887925306922,
 0.6304615875870572,0.9737437297196032,0.51335066861718,0.43243795916194705,
 0.978147390034877,0.7545716122222044,0.5186099924667934,0.09431798206150277,
 0.9941972930268074,0.09597309562669654,0.3913676509752665,
 0.44879806214438533,0.16175970590979316,0.06841954806242523,
 0.15729620276909362,0.8103837873808206,0.20698663273941476,
 0.2496585351113496,0.15489141473270485,0.8009383323953502,0.8761732903619943,
 0.06223064926997779,0.3247959891030977,0.6969641892088402,0.538402106218705,
 0.601685850817932,0.9471888322428222,0.6488467612828955,0.9001628754541707,
 0.200115317467484,0.051388048751783266,0.8291615012133502,
 0.12737209804120875,0.06883757915045519,0.1582827004629095,
 0.39325389001841105,0.6055892975926419,0.053929668725579916,
 0.2819961443691834,0.047321578808925624,0.6576143362090433,
 0.9379161688033093,0.2779931508623428,0.5037013079416706,0.6105990680258707,
 0.045601675823445986,0.17768375669825864,0.11702559925601674,
 0.6755021784423129,0.8179821718256972,0.3564884497148719,0.13109780809006932,
 0.7952636406952783,0.13817744406334476,0.9153927719530867,0.5749801637712804,
 0.4302105932848067,0.2838321404318497,0.18446604378843845,0.4074904885919902,
 0.560312082850068,0.12613126613062897,0.2572317160617229,0.5389326565235015,
 0.7787034297789148,0.16071706075417935,0.45404520334090337,
 0.6820274263615156,0.22633091634740288,0.016488754830911168,
 0.24366100724790085,0.10924302717318302,0.039509326474153283,
 0.326294814196047,0.13372006485184051,0.33817277645636945,0.6904361193059856,
 0.8957284952571894,0.7924267293977982,0.5020608299885394,0.10030921932367987,
 0.1632867436780594,0.050855671926875456,0.2509567555410712,
 0.44428143064781367,0.9567778360698261,0.4075995456361944,
 0.07538217997382424,0.9577387612907958,0.04414462191554591,
 0.5857735658232007,0.983830219291797,0.7370326620746492,0.06094498960401451,
 0.2534747688038458,0.9826454741004941,0.5889806588612839,0.9411995193917483,
 0.9338074427802959,0.7233719663780401,0.7710747846434686,0.08005546119585427,
 0.4632583914588396,0.7117369097346713,0.2217176416718748,0.58392984288224,
 0.9056322032862039,0.6734064219478381,0.5720460862189629,0.430470839294247,
 0.9542310596216025,0.1571699043720265,0.5156372377867771,0.2279048481903051,
 0.2555768563648967,0.7025256187597101,0.37568514203431824,0.8234421875779606,
 0.010484318201562481,0.27571324482993054,0.8618896614768267,
 0.5702542480360309,0.7642373438574321,0.8479183586043443,0.6888756467544412,
 0.8671581596474307,0.3358817911024262,0.9763498099031578,0.29669692328338293,
 0.08685577520752474,0.6621331559658876,0.6476467706593222,0.6300182857361333,
 0.054670227019840034,0.830165692681945,0.7642547483627219,
 0.22190250040071535,0.12842514013528106,0.23476455391200457,
 0.49847509250796274,0.8367350162655729,0.19204068102545602,
 0.10368286047026987,0.4405012527392843,0.3898948672376743,
 0.22578122911591125,0.9679191152931059,0.09796091636899529,
 0.6593766037276724,0.10787038835955076,0.08768933866979411,
 0.5394575775598005,0.9505004277385812,0.9751050551119111,0.19777232784950527,
 0.8401755239057008,0.07245348990053191,0.8086969415196472,0.3380996541094714,
 0.5671361066727385,0.3815190482394567,0.3592624280486931,0.41997408900147815,
 0.5132842671162507,0.7230002172536265,0.8596346983709977,0.4231885323382243,
 0.24728602720282078,0.8025125246437762,0.3478142763302995,
 0.026606093693851585,0.8313549327128006,0.7192120138203876,
 0.8306093503954113,0.15417051772202617,0.2831913074412864,
 0.36267456023873024,0.06896227312785241,0.7768272262266942,
 0.3808340439106739,0.7484314703380639,0.04478351250108059,0.5238978997101651,
 0.5391617809758634,0.5175168035984888,0.3462125421339173,0.8752905685866068,
 0.02912173801083906,0.8807829285214058,0.9944258886817102,0.5531001790530586,
 0.09862521811330349,0.9692976888634697,0.8805516064191472,0.691118660931274,
 0.5456926980356098,0.4258261205888223,0.4843439791125337,0.8811971521679127,
 0.4665111909846119,0.5058443803665119,0.15329823807622445,
 0.06569521027553793,0.7810838224518863,0.51672562723575,0.39177883904761757,
 0.901241141155553,0.2880588784099567,0.8215205542740646,0.8332450254667021,
 0.3803498700647232,0.1974756258596534,0.35439440086318696,0.8933166523139034,
 0.4485690772120373,0.010934486506050223,0.18616266100086543,
 0.09186563352355892,0.4833422690842134,0.6390753620548072,0.7372060119707543,
 0.6007373262482233,0.05071313350721052,0.4780398168089839,0.7348160278053287,
 0.507010616558712,0.17727593214113602,0.9374319295542488,0.8000719064339076,
 0.3423003841922889,0.09632570518583772,0.8977507319886114,0.6926008897536392,
 0.2698675177241976,0.8037264986058037,0.07285292952180478,0.6807495829367693,
 0.6178632219533652,0.12411759841549053,0.9881449382662231,
 0.47095858979787963,0.6522585136017068,0.4506581963614439,0.5146965389089867,
 0.7123183784901905,0.6205531452452197,0.4029855602149892,0.20793277603774862,
 0.7475254970334861,0.21784350486033355,0.0257892032781748,0.9077195452941734,
 0.13354650407763913,0.539276668061312,0.5589845216470362,0.572012801933273,
 0.15953989149577963,0.22291418404334384,0.45014978930277016,
 0.5012039360284756,0.7234624865448929,0.4269603630410821,0.7467506750836406,
 0.6021473570996785,0.6307061406882893,0.6904695291271055,0.04767452884829593,
 0.2902898826792568,0.7178410908818134,0.2918797007176226,0.19046823860867945,
 0.13520252574490388,0.9180158864361238,0.3686704082458626,
 0.17959670011310858,0.4395548206039007,0.20428396194158716,
 0.3211130468585217,0.6566851026672574,0.667032378108414,0.8782780387991908,
 0.725401458522885,0.7402388471857887,0.14622545537098464,0.3911847686443902,
 0.12528838422105582,0.6595549908156232,0.6815499141327984,
 0.026110996763089145,0.5824671342955385,0.2664651515408979,
 0.3131614255206563,0.872030444438786,0.9888302333803634,0.5843348584520958,
 0.8587861510941418,0.2949132887972957,0.8969910048006852,0.49270643180868656,
 0.09989952220111531,0.7419392206348168,0.029429053457195486,
 0.6850354715688273,0.8906073261038411,0.6402469756610137,0.6750818889787674,
 0.7829144774245786,0.02159104025469305,0.5368421273014781,0.8858127757058537,
 0.2575507650462412,0.014076036499037659,0.4738051533244815,
 0.5860053809951302,0.41325767760144627,0.45248821457312827,
 0.7716078607949429,0.11888996467310675,0.27919982499932305,
 0.9944481551400665,0.01728398246086016,0.9068042620651994,
 0.18077361692566962,0.16184004596698354,0.47698380169387533,
 0.6778660951193818,0.47213049941778595,0.42415794646215876,
 0.0234744461941635,0.3226021644011132,0.1284666938960487,0.7877056123246544,
 0.29746542611250004,0.9422215940900032,0.381934587083565,0.8210794625014951,
 0.6691273750157747,0.6896341320594118,0.1251402983763239,0.32293903044006345,
 0.0761994227295163,0.11168870762339556,0.10297652665003298,
 0.3753081159225251,0.04479823954116413,0.307823672436771,0.6054146294024945,
 0.5572736622980787,0.7657430464401112,0.9510440844034488,0.2783342975111751,
 0.6038875425727133,0.25871851164525683,0.31764389553327677,
 0.5890433662083229,0.1829213444989065,0.7046191519504732,0.6114473302143736,
 0.1698482380894033,0.7149932263382408,0.04569143638676676,
 0.23798195115545573,0.22320969322787598,0.9338681823653445,
 0.6684165004709525,0.27789569454935703,0.5989328182113357,0.696669267607136,
 0.43046512798634273,0.33410083655418177,0.3938056553604742,
 0.37358927960392363,0.9370246733400512,0.9805728091406025,0.7918625019080541,
 0.2669976395386937,0.37609616266236745,0.056716220386815897,
 0.26186321491377995,0.39350494914815604,0.2121019790149744,0.69219479253299,
 0.2962333614462106,0.18185623858372724,0.2724487647541395,0.2603482539687345,
 0.6463103241811019,0.4792218947471081,0.24717166838129512,0.5218674692528462,
 0.9249330574887991,0.10559971909365573,0.8526788375064406,0.943888506145388,
 0.6013809650892803,0.7886315780913596,0.5792432077686358,0.9976698162012891,
 0.20177863972926258,0.5631017083469391,0.5048439421117666,0.3176820051511754,
 0.03942871178849616,0.4526474435401523,0.5257059629205664,
 0.09492775783703178,0.5077354070301707,0.7562152500224673,0.9501740742913862,
 0.049133818107903626,0.8121579116682809,0.2701470864550628,
 0.19710950244296777,0.34732666930047307,0.930941368813902,0.3583774945049828,
 0.6825866377626802,0.1064727789103539,0.016753759845926197,
 0.6514452109042556,0.9674989593379483,0.2352963156903337,0.9739376587404789,
 0.44613809205220856,0.7040863681436552,0.4864164465518883,0.6356083709216767,
 0.8714636597596085,0.805969372189454,0.029400983334560427,0.6454181765575678,
 0.14402033740311926,0.7910761988474037,0.2529428344682749,0.843300698693103,
 0.5090084346521826,0.23467774651735318,0.33631453664411115,
 0.9009469324647191,0.03847184653182323,0.4998653780490854,
 0.03338652258183905,0.39704770143658064,0.03170172176998831,
 0.8210845712117667,0.5813402258464268,0.7872118031162278,0.7542660733608499,
 0.09032014472510552,0.6897009271389789,0.3435867652844242,0.6399866644422294,
 0.9626213848527586,0.13064115121970243,0.4687361334013316,0.4419946501820248,
 0.5538137256988062,0.8990102223919971,0.24505565838110954,
 0.48930204391356724,0.5533542379679817,0.8205655446559132,0.4192338832032001,
 0.3188229496946693,0.5403718823463004,0.5986708481235083,0.2671968194582608,
 0.3934761273350761,0.4242160096619896,0.4720179244356557,0.5177293393876911,
 0.23895926739322526,0.17522560180279856,0.6153583417564678,
 0.28566535899547396,0.9153163808936049,0.3759441101387694,0.3018473785263509,
 0.6481698551109636,0.3072923415506621,0.9591712002403239,0.32605261806960706,
 0.5042352958713611,0.8585790217564456,0.14625127885094824,0.8493995020450056,
 0.9672542011029752,0.7209961946535322,0.9312217872113661,0.529706207510045,
 0.2132920494275421,0.24498240583699438,0.3030707906540039,0.9997278597873699,
 0.13849884214582653,0.8241621204536321,0.4568246718389489,0.5699399454183915,
 0.823299175242677,0.8199106993389698,0.5546082259873701,0.8762939175850237,
 0.2840557055405537,0.21784252541719396,0.35160939033726646,
 0.09928917721600572,0.5032305605981197,0.053453136811285384,
 0.48786363398063515,0.27353669665816394,0.8457048595277273,
 0.8022002566786552,0.359796501185799,0.9561276343391061,0.6219523261044947,
 0.4690535930464881,0.3357311255137795,0.7094793091516228,0.7438964804601841,
 0.12920651105838676,0.618649002998744,0.7182844192121364,0.10993914823505557,
 0.19752209257809228,0.34538285844010364,0.4061434620956099,
 0.09921828598053817,0.27903387292672477,0.06683245680286376,
 0.9864639232527055,0.27499862317837576,0.11156117233542728,
 0.19575298687358655,0.5644229441964873,0.7941950154617263,0.991071562519418,
 0.7645810819623259,0.5580010648088559,0.2490093956550138,0.5495399900522046,
 0.971202307539262,0.4343466600070922,0.6251822842913668,0.1745983718564138,
 0.7942831616632784,0.10918953363322748,0.45940453704740236,
 0.9181343539297734,0.6332437484466916,0.8992322650699789,0.9805641865403728,
 0.35952059926040214,0.18679332706475038,0.3483362105436807,
 0.5314995072655281,0.8124236820097369,0.05150757612415424,
 0.17195057093868227,0.6951260249426421,0.041876043803975804,
 0.7324781349643473,0.2299624362978916,0.27688817424017353,0.6676464077372053,
 0.8642797260725711,0.07191949665112185,0.6345640463773805,0.5480841116956916,
 0.3246128701535267,0.013856408936337217,0.5540965237635331,
 0.22820496958104386,0.17621540362293475,0.37390150341172834,
 0.3323102488466644,0.7716596006740732,0.21536861487899273,0.6837388156618067,
 0.3961884327735943,0.23933714434440767,0.4963307494122937,0.6956490544771218,
 0.9802911602966264,0.772127304554875,0.9872282865262494,0.6970303506864621,
 0.08580620129413752,0.7787875209676478,0.7284805564660393,0.4862256399844851,
 0.5081627914259906,0.6938815606721396,0.8705300662937979,0.2443857786566025,
 0.06550263575922122,0.23082644877264602,0.36336419212142135,
 0.7085361323436766,0.6921228344421321,0.3937305773999472,0.40979275687078687,
 0.7214317682915508,0.7115891613967713,0.7028189952998185,0.9922416793830253,
 0.7728654369693597,0.3592974535851048,0.25125471128615595,
 0.19269459633151942,0.9320224485666241,0.05604695050652686,
 0.8678228857710093,0.3121258463712888,0.6952207350444906,0.3631452186983056,
 0.8785573105663593,0.5938754588313784,0.43319517878742597,0.6402464296572251,
 0.5482750346256852,0.9143593546600283,0.12055485432912172,0.9903423930468027,
 0.4788516048412189,0.5038204990692831,0.2024994140118993,0.5053234439803261,
 0.47037689451958486,0.786653128859513,0.7322022187280368,0.3357718094008786,
 0.5088462453894714,0.8674816817785835,0.36616030809251954,
 0.32772973823972906,0.7687244604846698,0.75014519288343,0.4697223988133008,
 0.4170211279519068,0.8054406636678011,0.8446951807443009,0.37410720243207496,
 0.18703723705356845,0.9291801036503575,0.5377630397258205,0.5229730635406941,
 0.7758375029769079,0.1954832023221602,0.6211437048584695,0.6794789933997848,
 0.5310708508592126,0.02526012395209576,0.1513816408962777,
 0.05814636044791344,0.08939847175578541,0.8168741473657591,
 0.4563677356136854,0.909544745373858,0.20957568172222185,0.696181842354704,
 0.4092451875051817,0.18015510260840073,0.7958813397477337,0.5706511298643326,
 0.4888262029080367,0.4696155999549816,0.7539409081330113,0.7665415809841241,
 0.49752544123809683,0.5511841919160966,0.43840896764789294,
 0.5637273941052203,0.41643231206596654,0.459467913639622,0.8582564778435797,
 0.29353294305846467,0.41479926864614514,0.0479014262004448,
 0.6709772502056182,0.7308496445346876,0.33658529952616845,
 0.09740530235771883,0.40575935422661424,0.6898432541803046,
 0.6173062924515704,0.7364373405034297,0.27619250362300296,0.706732322959094,
 0.06991669418390345,0.7279299734133804,0.4418632171407503,0.7108056715180916,
 0.4192910677797973,0.5301646431676983,0.9139850934870631,0.45931679589899543,
 0.08728001578938271,0.8829096547898505,0.23718225973629026,
 0.32799305332566475,0.09805958990584807,0.5930856594753272,
 0.9985980500932231,0.9297218223400063,0.0988639495554795,0.6287510076940135,
 0.5971938914740935,0.7256804565385709,0.06779250698466066,0.9980585451707565,
 0.43556487095632546,0.05604486816741461,0.3549069717136565,
 0.040487491973501655,0.35168566306672344,0.6155677629404013,
 0.2790955990566484,0.4821430159868536,0.17884770490147484,0.9526114728845528,
 0.6197850514584612,0.5147708782698415,0.909903953470761,0.28674044725842696,
 0.910058266325326,0.16331430975730044,0.8582812866337834,0.8212974026440841,
 0.2949285737611792,0.14667340852933264,0.49806863422339376,
 0.9583189638471388,0.5880148930095306,0.605368571975916,0.08012052617830312,
 0.6085011331213316,0.5740116329732231,0.06873067363076724,0.5351426473337701,
 0.06217420588365119,0.670501433203926,0.16947996828491152,
 0.17828309747153737,0.3457060365664417,0.33710585819669026,
 0.40399678854758514,0.40773169970261813,0.5308714116373839,
 0.05546948202344182,0.01079349398162388,0.5090364747167551,
 0.06515882636339865,0.26084579313441414,0.10907236657590225,
 0.014317594906430364,0.5088680914371514,0.18301925592187251,
 0.38869506864816095,0.4885240104469182,0.03215284703553911,
 0.5716783176964241,0.9522556233130031,0.1581053685106074,0.11818583592364573,
 0.6886335081438077,0.6656993610986742,0.4257688784106026,0.9946260070659674,
 0.7572893879090842,0.6113405336435731,0.8617866031553588,0.8310917908746196,
 0.9653736243433074,0.8696121531559514,0.03163081097336551,
 0.46007940658730384,0.5520889112588434,0.46391029578835374,
 0.08314800653351884,0.23199342188253835,0.5307472876050076,
 0.28926930581170385,0.47749718912934835,0.618570619860101,0.7086649447392925,
 0.9602191448790856,0.026004959794112215,0.07214061980747022,
 0.3640102212113502,0.12487324858037196,0.43468710891920703,
 0.9036153908847431,0.5218163094567425,0.7972304945074816,0.12795145122284324,
 0.744596284432872,0.7107786858977536,0.1526729665146892,0.7895173063961494,
 0.5273999792757158,0.014800394055326738,0.14299677130212052,
 0.4090771427156923,0.16644690118997385,0.21109602711734266,
 0.29962297404636695,0.10018596130966662,0.5407030126589903,
 0.05705048948823399,0.8137652661176189,0.11896274930442385,
 0.00971228145021441,0.5712291902657461,0.9161010438572358,
 0.35320311315556374,0.006302598468409637,0.7201911201575911,
 0.8461756144683599,0.6400990459198864,0.49568902768249656,0.5257354052406247,
 0.09931562737999589,0.2758214463099553,0.35192210191289675,
 0.46827896510873146,0.0014020934472418478,0.7257725000146734,
 0.29509340631672687,0.5365102694552254,0.7657958842409133,0.7956286135651446,
 0.7742691923251454,0.5823883265412442,0.327169469247232,0.15210637688124196,
 0.29328005139400937,0.7749949978618124,0.14254309956115618,
 0.3781605330085678,0.4560113512791315,0.6326178914072506,0.891098248896663,
 0.01922597340379051,0.6180067893287637,0.31876231923195175,
 0.9474466947664905,0.19847138283887855,0.6587442212427569,0.569570570040555,
 0.40820780804414436,0.5243367099071278,0.7282396002899599,0.6426503254666339,
 0.24055143869077955,0.8925737390492362,0.07499280217897508,
 0.6689760181024936,0.4640943814715123,0.08418306555610577,0.7122056300162762,
 0.07839283904996508,0.399153634759,0.011159983061905132,0.6139077004904009,
 0.8892010439404938,0.30935040080485476,0.8708991429987609,0.3572635884987997,
 0.8966401257075496,0.8646106418263199,0.4713866886569672,0.12143417048607252,
 0.9648327394699648,0.41436844330571687,0.09732679656261956,
 0.5350965160044083,0.9632695609212399,0.9278352625556551,0.06067681894235699,
 0.018199445466581476,0.9866906662952998,0.9905776138369876,
 0.8003321576232688,0.6916461832175662,0.25894629949355874,0.5598803812803889,
 0.2389725226549686,0.2640211056740014,0.904801018806496,0.5971009238214222,
 0.4673660171343612,0.49978649408609066,0.5832055005017643,0.6337033813862682,
 0.2588915921659054,0.15986836974953744,0.4211988477236174,
 0.21000196616074518,0.2666453763691773,0.15669011522328713,
 0.8894199070835922,0.4409074501504028,0.824133531202869,0.007785803743459496,
 0.24052662240098932,0.4383729124099487,0.6099835470827223,0.6138541131717191,
 0.6249202225883993,0.8038992298113651,0.6226241132080542,0.15748117830791342,
 0.8434687736569201,0.5050738730310667,0.9106201222542814,0.6034566528971619,
 0.3050797851725684,0.05318067776419877,0.9696403714364228,0.5635377845874087,
 0.7992530047915779,0.118783747335516,0.8103486075556571,0.6324072793607988,
 0.7943299162610895,0.50856773652175,0.3496005984171262,0.7267450345193744,
 0.41389750576171414,0.40783089068053946,0.9723101070603662,
 0.5300135183389271,0.6414267711446167,0.29409514554689076,
 0.13513299178948535,0.9665126726855564,0.08717716259520081,
 0.40235865273947113,0.058907775051699196,0.4687401981028205,
 0.5224315051729368,0.9273227071065597,0.3758248941167962,0.341380767289308,
 0.07463006492011737,0.11922623539375488,0.3771169848465302,
 0.3229680549100711,0.16288520618099156,0.8880837139340428,
 0.32649833810667384,0.19070037770618564,0.5134191902141738,
 0.7466712622869527,0.946594295443754,0.301695886754875,0.38616747962509856,
 0.48531692788801173,0.19824959914498042,0.21269745460493628,
 0.6088545873077905,0.8029818010506414,0.20888539440500709,0.1699352720184848,
 0.5590323186928134,0.9228582689574912,0.7649115939828396,0.4748156342429579,
 0.8262216726889223,0.7149953981885848,0.3925576956363669,0.951980048485649,
 0.8009681810100828,0.030894787493782494,0.28941891675477216,
 0.7489150313714881,0.6714442482120021,0.5862192009691691,0.8235287474371379,
 0.4916575443793474,0.7402964293147647,0.5180335711929352,0.3420689824005847,
 0.6092810260451071,0.3291039447397911,0.5569526975772184,0.7749334022546869,
 0.17138856033135097,0.47996351863040543,0.9384967424724945,
 0.27504977296669264,0.538784393634249,0.46779091296167663,
 0.014824364818548319,0.16589477349973736,0.07348358836307367,
 0.8985395137882513,0.7660036635223982,0.9886134629282893,0.4554547407107269,
 0.09584739831386557,0.28917499370511135,0.8868584470494326,
 0.2457582547279813,0.5292982932942373,0.5777694705282205,0.1806568816117533,
 0.16085478354736638,0.5475517068410641,0.4260620725141816,0.5414408237026166,
 0.7413188485402977,0.18468800747417669,0.2851039396703431,0.8177755590286896,
 0.7964322030273033,0.07048401279814553,0.29698236870233474,
 0.02405226023627327,0.01688352738516885,0.7415750636799033,
 0.5993486694624494,0.9937526225120153,0.7545461356004906,0.9667523885235045,
 0.3146037785828395,0.22200232367070105,0.919026778755347,0.3399345226954099,
 0.8605644355349334,0.4124965198959215,0.1632481844621556,0.5217813257346227,
 0.2843943658148015,0.4755090783897349,0.8650856623725327,
 0.025156759536850593,0.5553539493805981,0.8626537505761122,
 0.24177424548258764,0.1663949569057448,0.5084351580451837,
 0.38880389682958505,0.8898268845162671,0.5143753367936353,0.7262443765105033,
 0.10955438044435328,0.710415659081182,0.43012964473436255,0.8958304551517051,
 0.8733979446448279,0.8777504540087644,0.6246313383552743,0.4351750358934703,
 0.5036561092044751,0.7392718742405554,0.28822399038223956,0.7689962835851945,
 0.21437340995583076,0.7679435459525157,0.5861976951755645,0.7791363746182145,
 0.9473500777929675,0.6820285417880563,0.32080362354045755,0.3649641991777932,
 0.040157904268330546,0.16970872182505625,0.9970539131877041,
 0.25254598565595976,0.3453658866614262,0.26377689968690743,
 0.3815874281376146,0.6284604523822146,0.1608993369931948,0.8931733993516087,
 0.18882916454472654,0.4819637596933758,0.2397733950005012,0.9603139347538308,
 0.741222768531813,0.4868743813471432,0.20838694087397092,0.14254960015323437,
 0.02799527976547944,0.27481627769905315,0.12971296247281927,
 0.17587166265521859,0.12360088589258322,0.7034787377519375,
 0.26515984983327845,0.05622316394185156,0.9503418577189636,
 0.8039862906341523,0.9592383378747336,0.7929300749295071,0.7486457444646437,
 0.3245182601895037,0.08757171878677039,0.4723366256604653,0.2667954005699349,
 0.026574460222372864,0.7376752959012758,0.004073688549318444,
 0.021795429871649397,0.368690607250035,0.6425042305034527,0.9595564391529566,
 0.7562773656650049,0.3338689055921452,0.6953852891611221,0.585528378600432,
 0.7814054479882151,0.43899469873476116,0.28320382307387837,
 0.1878882811537158,0.44985634970232935,0.18992776179488757,
 0.7209974076453427,0.7566122666591292,0.3777367136055332,0.5563255612273597,
 0.5268072834606794,0.7622659831288076,0.4594144749373278,0.8344375198811502,
 0.9787408503175425,0.7798643998217326,0.37478665630656893,
 0.06946689893680447,0.849471688117037,0.680928840668885,0.9749233051079719,
 0.6755728680234842,0.6045031990904273,0.41086143371688144,
 0.40134501601092953,0.2526838807248495,0.5166705479270576,0.5870476530854404,
 0.279665795141932,0.2210685659780508,0.2735104001824701,0.4775338239880115,
 0.06320659673327533,0.07521088813980592,0.8982287322740827,
 0.05159885142481557,0.5702709823057248,0.5234197431306509,
 0.08209530671700649,0.762058052834847,0.7793103760285045,0.42590699581104785,
 0.3435330457429666,0.5868872893222171,0.22407375483519165,0.7419240960319116,
 0.20118233167619692,0.009977186347467248,0.1802432324461748,
 0.1936096742264497,0.9291743880595409,0.22392694190474494,0.5682930000848918,
 0.3263668894955345,0.5122478004031116,0.16166913747370026,0.6367826018321684,
 0.7488760578608457,0.8487827432625343,0.30086321413837125,0.8816928717179937,
 0.11606745064632007,0.09624333704448962,0.9998207281887326,
 0.6762524312403609,0.07529058165813962,0.7845697301966518,0.2370114598894466,
 0.0051758624153518396,0.5323620861238241,0.9553321718837673,
 0.9304312838827052,0.24691439339119348,0.6651265413346628,0.6565132323967624,
 0.9061960439614474,0.0763359467853284,0.366951493219024,0.12361521147617505,
 0.27763175789672645,0.6206873035270072,0.19282807447667305,
 0.6512927515549618,0.11159705327335945,0.6912668329149271,0.6376117211463199,
 0.41402016677869935,0.5543608725636948,0.470692274913703,0.20473253453344908,
 0.2612331424960018,0.4883490106646282,0.17857819614365733,
 0.10710230760104289,0.858558455731249,0.18874479007518397,0.5897729285108397,
 0.630960547892743,0.5149998971551537,0.5399810715163843,0.09016980709709765,
 0.8022994129819905,0.8030432532406269,0.30106893777570376,
 0.31146588614879556,0.06551275333912165,0.694646856758035,
 0.40460624342938756,0.30943309065961155,0.6004283205198362,
 0.8620470896080642,0.9749794586740069,0.48898911323713357,0.6697417328503058,
 0.4995377511188174,0.3602576480833881,0.3130166237271472,0.2261104989832431,
 0.5875977627317971,0.5867438066159049,0.47393323672419807,0.924169766584709,
 0.7460619002240174,0.8975031577365263,0.40378784465407047,0.614930913693631,
 0.1746436295109557,0.004231480753173367,0.6157475935782414,
 0.5388784660504707,0.8218796518034183,0.03143859229707546,0.6766998919503973,
 0.8386850211917827,0.5341365791672397,0.05798941883328468,0.9163834144728685,
 0.3262854517424113,0.9801326042302559,0.16839613757257665,
 0.15973735193613892,0.9294148747794823,0.7885247923660673,0.8234865119207478,
 0.6556969360131707,0.6843286878240549,0.10814933712770514,0.9511013624564432,
 0.6101293149922746,0.9508248576657704,0.21264863508865306,0.5459972458474391,
 0.32036195892785,0.27708265377514385,0.9839956660554405,0.4756583059677245,
 0.6293464267750479,0.2976224623133724,0.775933336555075,0.6641122999420912,
 0.3679430569037969,0.058041138202775455,0.9902963862610752,
 0.7485056236781349,0.8521066036299464,0.9767527560703874,0.7649788473701974,
 0.5780962697693898,0.5497813223927872,0.9615965081108425,0.8810644400801082,
 0.4028706279032246,0.39972931312588444,0.8448879035312633,
 0.26220545598520606,0.2945511816831503,0.42025813358641084,
 0.3499798309683453,0.09776064336133061,0.7698122528210285,0.5646060467042447,
 0.6455382531045079,0.383788043720378,0.9943627287649784,0.6368436004954143,
 0.7276449196460343,0.4126732914409468,0.06282550004625953,0.4721071390818423,
 0.41708794857307063,0.15355843280827042,0.18127313228541098,
 0.2538824540270854,0.06851730496423625,0.9734863351520208,0.9405177571030479,
 0.883715120446854,0.027899529628699904,0.48397467422846163,
 0.6167344864102211,0.5175041388776704,0.37170992491469157,0.5436871226526331,
 0.28154661054869956,0.7911409647207278,0.5570033000062811,0.842506455687492,
 0.9829561074534631,0.37192963087020137,0.9213072807560951,0.6060537608296231,
 0.004583468212709607,0.9884072312354695,0.43632454576949087,
 0.4550404434801386,0.9924107815455366,0.5663916697086616,0.3449657142750857,
 0.1971498622311646,0.7582137033899116,0.2628367938332655,0.7627779066735612,
 0.8701587683135857,0.18539670422583776,0.7645291621244419,
 0.31782604527034364,0.011272916065781069,0.3811343026669689,
 0.8224701917445952,0.33075434463836795,0.04647967925854246,
 0.8845332134245404,0.898055129491931,0.033340031629880196,0.5804235892006021,
 0.4305916252951345,0.7471754180403617,0.15753619924739326,
 0.18790732044071923,0.42134530824249683,0.08493538885745544,
 0.06003723565505259,0.9655177867965233,0.1962876458149625,0.6846314308764383,
 0.4524523676246096,0.11800151148004323,0.6321818756643929,
 0.28300536885364724,0.532577151654171,0.3647821352043028,0.28122352559769337,
 0.6586892311449466,0.8684989875091104,0.6375259680500431,0.09445919829623939,
 0.8307568463724917,0.9865570455950904,0.17352999686767778,0.5731361621721621,
 0.43474055328785643,0.74030580100797,0.5789630519063091,0.5689077307391125,
 0.7536697516072206,0.771256697945018,0.28505713902779695,0.6904207042022431,
 0.5966057160433824,0.3241048753723549,0.47434399098952484,0.8888454017345088,
 0.7597526568143939,0.5331514815961587,0.4730560442614107,0.3924038191297623,
 0.5887914414794962,0.014909156385918387,0.7244335214973621,
 0.6264527698989878,0.372078295650471,0.3620982091046101,0.016739181270046588,
 0.24893808284373442,0.3720635779010494,0.7841429689154713,0.4525960832155481,
 0.8735730272692188,0.17464839097378027,0.767506524876123,0.9156191497469226,
 0.27389457146070206,0.6931024591193728,0.7268024307656868,0.2089534281968659,
 0.4042881946132565,0.6482062002222386,0.8547890043108302,0.8223050959070761,
 0.9884085834279336,0.05328190875121286,0.3614822024993827,0.6001760534479847,
 0.5684193495474529,0.10826232985715445,0.6737044944212077,
 0.17611163086972392,0.8062392079884899,0.7727428052526759,0.9743029068883345,
 0.5443888458176314,0.27700468951493773,0.1509418887304047,0.3394635210296906,
 0.7708432733777524,0.356217969354924,0.6452756106006021,0.8675649713030518,
 0.29322002388013424,0.7233678932595846,0.3905116567018274,
 0.14943862092331894,0.8908492249812812,0.5960582141836248,
 0.15571595981246678,0.7626464402241552,0.613483788150018,0.4312100113940006,
 0.43601961660101307,0.8553110266747328,0.676485235618145,0.10357959867575739,
 0.730389173033305,0.22637668191822757,0.13663945253665966,
 0.18900515660217243,0.1566515024270272,0.6354050217215639,0.7314974381556719,
 0.7304992745171214,0.47334654889800176,0.2897077133242274,
 0.39371756057007967,0.24342473000981824,0.3149517899275436,
 0.7837286191471218,0.7707652976184125,0.056889578127935936,
 0.6649583627266258,0.6391420481436413,0.6082887990273704,0.05475714654478858,
 0.6019892584339432,0.023084332991174183,0.25442871326602023,
 0.20894034755135715,0.16044557097744816,0.7843606501277001,
 0.4068811692428026,0.9189087008047223,0.3172294174502056,0.5822088576451172,
 0.4486165534875495,0.6658431018435557,0.5279591514926198,0.9187446570792929,
 0.7663052303569869,0.959313091936033,0.08568825123130697,0.07500079199011367,
 0.9124304870770069,0.7514712541020752,0.47122295317895135,
 0.08325724611965235,0.00411625826098716,0.6352065215085648,
 0.8651331383883377,0.7100127713341803,0.3355665783291678,0.6943418736306984,
 0.9209063785008826,0.8350737499573407,0.9132722953409026,0.19667697510903848,
 0.9155201397153505,0.8625377636037126,0.5092736959528144,0.8106484348832208,
 0.1293364767566625,0.34152054983642854,0.3389810932894267,0.4234728733408709,
 0.7286823617197623,0.46905056364217346,0.42619817743211197,
 0.7800364489305458,0.03364615427842932,0.4869534281156902,0.4515384579356404,
 0.02610592090861974,0.6341832569845343,0.7097069414222877,0.6703722446081928,
 0.8253901695059652,0.2973021482022622,0.22975557006297032,0.543504805886192,
 0.009435179828445661,0.6871047896489563,0.13444767073757236,
 0.8106015657295313,0.7255877033044947,0.7832350466717148,0.6392470905882555,
 0.6511454927239474,0.6834250248364669,0.8720147394584493,0.1868421957008104,
 0.5291284838858801,0.7748638589692834,0.8201208959414681,0.1527989942003074,
 0.8326717274034072,0.8438410368734033,0.9877790884879419,0.8286702422134613,
 0.3223390113105804,0.9869576925559483,0.45005658452843833,0.6836254524264496,
 0.5349664417248162,0.6065928717130967,0.5462231445443798,0.8626308598437925,
 0.31601130154261947,0.8631964173030124,0.25853663022311246,
 0.39762543693000507,0.7969150245120605,0.055048008918691216,
 0.03542320715773184,0.5465837879908039,0.8480185952536919,
 0.03635971996613274,0.7097903903751765,0.15516105934534574,
 0.23172022933847736,0.9968133717121164,0.07609142865953,0.2569931382142836,
 0.4164818236689938,0.009777507038978506,0.1998972416146747,
 0.10188378810695697,0.015292955171358869,0.42072735798699423,
 0.510254821996885,0.2403443884299017,0.4082725954958816,0.04614187480378873,
 0.4385440337563252,0.5119508848539688,0.6862420096607624,0.6083409760945259,
 0.1630716984258087,0.6792877402723576,0.3710069160449181,0.13837945272684227,
 0.4972459746931539,0.15861655396272722,0.08698145847380967,
 0.4442596982090997,0.5396681730415938,0.24815307035371614,0.0630511179316301,
 0.4413618478194179,0.1628620388396843,0.5295140922703574,0.751100922962707,
 0.8083981307364411,0.4552030181831749,0.5112531837882919,0.9394701624124422,
 0.9722245725072193,0.5438092261560002,0.14041298352981724,
 0.23965449817238949,0.17384696068200323,0.45900100082611317,
 0.117987609210779,0.9000681049768366,0.9648784486994596,0.46631196936526753,
 0.2995058929003045,0.5362888388681509,0.543689621665681,0.6238750437879685,
 0.17244017600865635,0.09978345444715209,0.0921552880329275,
 0.6429306148412957,0.1408313396965044,0.1376495112060515,0.5847095774269981,
 0.5101790565376091,0.5763922021041601,0.9111062182647336,0.256307878159326,
 0.8019371903863136,0.25669765289059243,0.12914178716697833,
 0.5793450322726906,0.12949534506220395,0.32394251251669715,
 0.27130689061860624,0.669635409036069,0.3997875399604015,0.6461919133230224,
 0.7800112185922287,0.5613896476303945,0.1747575375778343,0.8358189241938491,
 0.9545551085924552,0.7100437411759999,0.25915536422804253,
 0.16904497486505998,0.09121018354614507,0.41535426459248015,
 0.09765798052183106,0.12490089324106435,0.36133573285506926,
 0.905036736301234,0.07267255902623082,0.29695297172173885,0.8309855640172736,
 0.2417915459500155,0.382967179498214,0.8534067913390988,0.9963982538305888,
 0.9393313412645782,0.8057755748340487,0.7646622970296498,0.9588140894548018,
 0.8148637247069729,0.8894680877725372,0.6605314417270454,0.5292689396746699,
 0.8267081875195322,0.9210056339798673,0.4346809763253915,0.13298802389402975,
 0.439317436615524,0.6969959119868291,0.8997801042705048,0.9584922695289795,
 0.8903157298927425,0.5481855240614691,0.5004257161912612,0.6238980964458786,
 0.4698210947748003,0.6340987077473702,0.0229487751392361,
 0.007355200614251878,0.7199971888989669,0.9977163876439634,
 0.5987258290824569,0.40037544072557885,0.672559826071503,0.15045778454962422,
 0.6077065138923248,0.31165595290820536,0.31709974809427777,
 0.04162075148367772,0.05024250294722066,0.3226411816567816,
 0.8125533273536492,0.1445785332600774,0.03603577488541887,
 0.10156864424789491,0.3099420036028544,0.192781651903996,0.6930753008309961,
 0.5697029934310868,0.1356853796026808,0.781207259057119,0.6584499706518286,
 0.05938699932866198,0.49561785343595466,0.4389940499257091,
 0.06323097661006116,0.5179352295427112,0.02818668458172724,
 0.37187647834803916,0.2396478097079613,0.33509717253477767,
 0.22255373593426875,0.3251336547976952,0.892188613806369,0.13403359728921505,
 0.22509836999731803,0.5761347126580711,0.9332284852455083,0.5238747192716982,
 0.0151170299812593,0.468862491481131,0.5108136364961868,0.18438291511451643,
 0.5845327668066188,0.48723215488238936,0.4004199384469327,
 0.28806469497469256,0.8592317845500774,0.1728297459902477,0.7927312631680151,
 0.04649685734359388,0.4051652446198427,0.17592465523294976,
 0.2335334948541259,0.33041929271543746,0.5892737535818671,0.8012542331063957,
 0.9180152871570653,0.1035650043585783,0.4163758558320906,0.2548507424725014,
 0.5293555901958633,0.08437993927075749,0.9335919601649153,0.8910282215570876,
 0.5045233671894029,0.6964157471529672,0.9262417137341483,0.7374913367414048,
 0.21159154298137584,0.15393464813771007,0.6902128519845352,
 0.5640996031961009,0.5354559921142912,0.14444430512496031,0.9705062510769131,
 0.3239811279593985,0.35427533134815226,0.7795265411237439,0.8559841127895463,
 0.5457678797881483,0.034738951585058575,0.13395303334679298,
 0.028430132642905948,0.02955990027606792,0.9870000801346712,
 0.5811666280832299,0.7291075458380369,0.6309120284217071,0.45170499244293816,
 0.4120060324698879,0.10437524291052458,0.8484883941007069,0.5543999312675388,
 0.09172800390740088,0.8264683833452983,0.2259074626913481,0.6161294655737511,
 0.026370875639321723,0.8344797054152904,0.2983485282433441,
 0.25721856321598013,0.5169552278138072,0.23429181820557576,
 0.43959051703872143,0.44831612938296117,0.3447465003804866,
 0.9705011453789836,0.6337875612913993,0.6832997164457139,0.9843295957647681,
 0.42860567632760926,0.08881132422889448,0.13918529900637688,
 0.7846756655709612,0.4976886501180938,0.4115800576287185,0.26540197126438536,
 0.5434752032171465,0.5701563617857354,0.9952792469512227,0.5366294447600107,
 0.42676150257964984,0.6878226544009673,0.9292596716059378,
 0.025034073633398046,0.8125197770675819,0.6801858428158856,
 0.7537445192048485,0.9243066361815558,0.06821120138367665,0.8641358735651112,
 0.5778893982039365,0.1340474776269528,0.14367934448520525,0.5082515472921757,
 0.9554668671629385,0.02445483568211837,0.8157607724133703,0.7397448406425772,
 0.9528221861208251,0.010414828097944584,0.4776507117617921,
 0.033700113305011326,0.035387017014662625,0.4230683969605845,
 0.9725716548891891,0.6219645781347249,0.41151676618066313,0.613061709131429,
 0.6661136131757591,0.03590606119117101,0.6336952209599493,0.7811515516983502,
 0.158575037161099,0.3467649132341216,0.8419999287816852,
 0.0024479682233892186,0.5390178991110504,0.494254010393403,
 0.39251758300095385,0.4586536362374798,0.2807853240150875,
 0.20051687361143233,0.27426212941321837,0.34327740921470673,
 0.7647812997093121,0.020115334031918675,0.04992104690784904,
 0.6366506612584278,0.10404059008821687,0.7682711840378776,0.3176730067164548,
 0.282402763944398,0.6500635845406176,0.5039633169969199,0.6263901500988541,
 0.5168419950618796,0.9182937005672598,0.7236034929403831,0.6601959607276496,
 0.00163169416505482,0.11680021130202856,0.07527127788123511,
 0.9031134540118235,0.7230310449635493,0.9698978736393807,0.4741506889681363,
 0.8715531440533683,0.2170222193753838,0.24109102485144995,
 0.17223633150021778,0.8856082343677475,0.38840157358111393,
 0.5478155034988436,0.86133598513077,0.9998345849758863,0.5890091550335029,
 0.8564808786129076,0.6100598974702638,0.44691600414732113,
 0.07631223712234791,0.3068761204976853,0.16420737233860883,
 0.9206839397961412,0.8820833248261916,0.5473117760636084,0.6003640995226925,
 0.685784252375697,0.284632444854511,0.6803707519128166,0.5841708329770189,
 0.6234292067459888,0.09256760745301118,0.3950667777328781,0.5337996046562198,
 0.5326148260966089,0.49134823257591853,0.3975941150011886,0.4766102980039215,
 0.1400670834170734,0.22876820762701178,0.646018192234089,0.08786015871791975,
 0.729650963351723,0.1997106535380292,0.08016072972978583,
 0.037213792944081114,0.18863997387532294,0.13301466874198087,
 0.24652987856977315,0.3098008361075324,0.4937623228786642,0.9408780271550587,
 0.8951867025291423,0.48093501853792253,0.9498526902927424,0.6645231144946381,
 0.8530676530426162,0.7420781226205045,0.8613508200912041,0.7770099243843014,
 0.2538376506502864,0.7316908200670297,0.6078051166769262,0.7075844708074136,
 0.16135753247860396,0.9265410816148671,0.2133883604606106,
 0.04812739937624588,0.041995971898044715,0.7149579672641079,
 0.6452147408817993,0.5722433728163457,0.520239250870045,0.19557049675622484,
 0.8011213369930382,0.22515372097577468,0.46370690827142935,
 0.023270117559568426,0.6194378471341444,0.8137885474628369,
 0.13253964426274367,0.7029548165784087,0.7110619728561858,
 0.44624041051757857,0.9646679924617787,0.09667347233682488,
 0.7276531557812046,0.522922060003602,0.31516881905423066,0.23824410990438016,
 0.24626949008024124,0.3373379959281235,0.7363467354065364,0.2798384624500668,
 0.3489443536305692,0.395820943245507,0.4889840816566192,0.8217667006608981,
 0.2463258114013429,0.020206618443788016,0.5660458586031147,
 0.9505763630179517,0.4842422666967344,0.4009448577625262,0.17814483268426384,
 0.30423666578690645,0.38324238573069,0.3759986074295548,0.29204741889732655,
 0.20149811649258131,0.5841633662871906,0.9236331745787671,
 0.43792239960131496,0.5772644970470424,0.3027162983175691,0.6470071143115353,
 0.917759039949992,0.9967405388867018,0.5422154662811605,0.9155716564337771,
 0.45920094067184924,0.16866968775292968,0.763601293017793,0.4554987296303977,
 0.5582479861529566,0.3884568580758675,0.03785410555922475,0.8842921678939472,
 0.25274281033365564,0.09361048035137431,0.8716444913299245,
 0.8472792925846091,0.394148430164605,0.787018370914028,0.13116735196656815,
 0.3749294500878324,0.9569375371991963,0.45156323176219837,0.7212943020935179,
 0.474684696747284,0.6449393550097136,0.22462433593032893,0.6747104066818749,
 0.30562664343227164,0.27076412436955943,0.6243096026542353,
 0.38583101593908586,0.06475604277962466,0.31365257435884875,
 0.4578961717868828,0.35105032764038735,0.34045171185540524,
 0.9675873260510304,0.30263567592978213,0.3964338220141419,0.1472854805975986,
 0.3910342591820638,0.269962604306766,0.3537937750195469,0.11860529515461615,
 0.6000264191561498,0.7659188656249817,0.11821375740142315,
 0.12501272167953725,0.1164261945798083,0.5754236317850712,
 0.31052485701146726,0.610818051819135,0.5678006926167953,
 0.028426355322056507,0.6248453545332799,0.44951411690306453,
 0.4587686785912244,0.6906511345528381,0.6222244100483817,0.5801015409066826,
 0.9306144439202892,0.09225847785425756,0.1464327972075159,0.8304061620531707,
 0.4754541104264268,0.7250798115416214,0.3264622964984626,0.09477565408419353,
 0.7975343623788195,0.6050311892118706,0.9216745769888396,0.4891067855094263,
 0.21594738572757688,0.0013343787781990857,0.9777776422703075,
 0.4072418436212737,0.3175899837644478,0.8024087812803057,0.7562979959898224,
 0.197090352221147,0.010546841767342219,0.05385837989384823,
 0.0703854833041605,0.13162131580568892,0.5560578955089799,0.8156434479279959,
 0.10016136609827131,0.26015858959904026,0.0505719240382293,
 0.3507868776497146,0.17065716632014416,0.8900661349990642,
 0.42588751194384566,0.8889023637614407,0.14643639413743126,
 0.7613521026195815,0.8762320517155068,0.3775746966391117,0.5172688572503686,
 0.07063672123311182,0.8152227235689365,0.9752060222725779,
 0.16244219064613874,0.3174533959770943,0.7064021331025693,0.7053564488833378,
 0.06937998228833786,0.15087386076928078,0.8454337004733238,
 0.3107363721076185,0.08078732599897043,0.21018538960798927,
 0.6420083284641022,0.0952981596695911,0.5728470607241833,0.16785555537165286,
 0.6811241823409431,0.6052988987811879,0.8482850504310216,0.30297911187042037,
 0.43051446841664665,0.42286542298150254,0.8931987512663433,
 0.5825150212014237,0.31136887568466487,0.7608342067804603,0.0654847583330207,
 0.8027321564201552,0.15949320109552967,0.28305310997932476,
 0.49182355130096245,0.10715548137109565,0.12560957288715358,
 0.3409080494790777,0.2637999839435623,0.8233940852639134,0.6887262716591832,
 0.9017432629392288,0.4199987178935175,0.11529542840966611,0.7485784277003356,
 0.4188081143901736,0.09322806318724008,0.4523689313470757,
 0.0015983143351869789,0.6919152197309637,0.9547285885590635,
 0.5439456727195725,0.6186097056763662,0.2543200678255676,0.5239427782249116,
 0.1221650907497237,0.13247172127425433,0.9765916367047617,0.3865959804134207,
 0.7197396510092559,0.20879498029435706,0.2825450472575394,0.2864971283801325,
 0.5340431036217504,0.6613602751594483,0.28939926815937467,0.5252551859074119,
 0.5367082528614135,0.8710076238193654,0.9961724624217254,0.6659945917774059,
 0.914882867672094,0.16412290197606016,0.7154235915024815,0.4397086901610451,
 0.5426000381728354,0.7400918862657634,0.7350387843172129,0.5518165063966731,
 0.5144641719499644,0.49252774128512555,0.8654682312548021,0.9590326547305896,
 0.5387214840210642,0.9895406271830092,0.5117197154325833,0.6241821790195611,
 0.0023263989361279602,0.5115137248150387,0.29674601700778325,
 0.7828395254007819,0.940449743010626,0.024966843310084363,0.7520329945718937,
 0.46410408924851576,0.7262744298078406,0.12991087160354697,0.940460301408017,
 0.27483151498359015,0.18345134842016053,0.007911512615468186,
 0.8804110572861534,0.21780932490804228,0.4864048273682119,0.2725941719591166,
 0.6877830767801935,0.596723472501053,0.3256169188240581,0.20102010730703157,
 0.8402617410505901,0.5760353510264254,0.8329922615043273,0.1248749455024738,
 0.8181758020283845,0.3915195870629924,0.5952911451927974,0.9025612096967864,
 0.6903231402721729,0.008881201108523196,0.9343377444100461,0.916692540173489,
 0.9303636435201865,0.5083998168070825,0.7724704617676406,0.5088452276094152,
 0.7569844661078396,0.9407269553755385,0.7258834573526602,
 0.044151398658302865,0.6165688356863628,0.13445819834957984,
 0.5562936674837541,0.36160471225567514,0.2032909357038699,0.6071431672238737,
 0.7284369258114318,0.8556086583327503,0.1445401523785107,0.12394832992338867,
 0.27318962630891874,0.9362914862527353,0.806312164149696,0.49713227999710585,
 0.8526650738261303,0.8282062205949178,0.2797423725146958,0.28316006541902305,
 0.8365346520811767,0.9037498196287925,0.3305613820680744,0.48178465598271547,
 0.511925667265522,0.8283456629474529,0.18009325750324723,0.2790442205186351,
 0.6212443694539974,0.3063182158851153,0.9040598019350142,0.34643535803705383,
 0.7404450243541856,0.17372822459372506,0.7504846492990372,0.6499821321581262,
 0.17762242682265816,0.7847433853896115,0.3799839452851107,0.6934481692133775,
 0.7716208421726036,0.9859117101623911,0.7152943670272256,0.16377726716399787,
 0.9135153896930058,0.05658978168565909,0.41985177225254755,0.374183922057021,
 0.7243345504976971,0.9218591469538111,0.14669665723355985,0.652922821693372,
 0.8802746850275052,0.7104469447069668,0.9220806498941531,0.43385224620704343,
 0.2959633104321642,0.935001800445068,0.7496960011424653,0.17067496241713764,
 0.1759904725421626,0.15647514106533045,0.8134170509525456,0.9935111650343297,
 0.025819267488769837,0.6822155151419892,0.7299496225513817,
 0.06627982980125258,0.1968997257470595,0.2034068955664402,0.0440654941021934,
 0.007041769148039889,0.5818853420572606,0.8204129935050214,
 0.20220055547577864,0.923718594122144,0.48244025196536866,
 0.04017348655326414,0.9989677885816988,0.7051055512662695,0.644149350216677,
 0.4060691380040037,0.5026716374599948,0.6021310356561788,0.4982788917996339,
 0.6760273692028718,0.5539695540336571,0.23860892501588815,0.9292332443701417,
 0.2215243279615755,0.3406016398278303,0.9938780048337152,0.724413399019863,
 0.9277313389775531,0.5907079004685334,0.48948918799623387,0.9821274552259525,
 0.2396938615731089,0.435572683460159,0.540528135384502,0.6155539076829202,
 0.034723668668285024,0.7282968515937787,0.13502951391977946,
 0.7392713842570734,0.4831498756611079,0.05638014010206738,
 0.22002438110393696,0.8408982204508364,0.5465433269764454,
 0.33291883205763584,0.581195807180107,0.6922893940249981,0.35232483165538564,
 0.6818645765619411,0.2751207462644589,0.36101849955554943,
 0.44843255055213027,0.43557424600674,0.7702821132187991,0.01591117465399483,
 0.3766065860323081,0.04029791515572634,0.5111232048569764,0.8158437093389204,
 0.9233314232376975,0.010835277708558855,0.6645324478851377,
 0.9599469503815462,0.9641003617489103,0.8228243624567338,0.2978060751410786,
 0.7425683738979717,0.5094643260673986,0.9639988320661715,0.4579943195459121,
 0.13489998009448456,0.0894050930481084,0.26676308223327183,0.961949045018232,
 0.5213379588205868,0.7296470669077707,0.44915887791467046,0.8220057386559785,
 0.30051590900837666,0.3423989231984328,0.5518887771440856,0.5950674764711078,
 0.6490888128839061,0.2628701264682183,0.45935501438472515,0.5109142743853646,
 0.5810254994543609,0.38128109781403474,0.5977282505376658,0.1874369534644056,
 0.22723689186745932,0.8192745413491437,0.301051778429699,0.9795855195459081,
 0.2004625411671236,0.3130635037742565,0.5453735557771449,0.7087413688586092,
 0.6557601942891849,0.3751407430137219,0.8853402795090477,0.21411379995248514,
 0.40365553276453214,0.10091943240315193,0.47581374176345603,
 0.10644806313963573,0.44556813870150813,0.34178505987276064,
 0.044992474496324486,0.9734902270163709,0.0010000277720785267,
 0.08908025873009162,0.21196554046392824,0.9134391564201304,
 0.8859282134488202,0.5149435251108123,0.8976794069236754,0.1609365464925886,
 0.7348453332437668,0.34073097464412927,0.34327653452816886,
 0.14530360165193634,0.42106733131734075,0.6277304671129973,
 0.9109761527228825,0.7893913562392386,0.5229110893932308,0.14981321665728875,
 0.5160173615718416,0.6342988301350261,0.3911392962930039,0.903292086587485,
 0.5841188642279682,0.39400415445520953,0.4813096629505791,
 0.04768909213146555,0.6684995745748016,0.7208896511652285,
 0.27043556520231693,0.5061510057375291,0.7276927812278455,
 0.27893683333952335,0.35890951777796687,0.5472031729324303,0.073038315835521,
 0.6608892073612241,0.8113989052893337,0.25691616577535337,0.8317461066941628,
 0.5321554175049643,0.5712668024675592,0.1284211860283173,0.4164959617389461,
 0.5948793396353842,0.02914775999695529,0.4676808996662607,0.9339555198651447,
 0.45414705083413964,0.7708303883804346,0.5905079322009898,
 0.19661648534496257,0.030190567106514687,0.9947406102413313,
 0.8991120154309544,0.2696879816153128,0.7137704942881666,0.7970542193589507,
 0.6171638102391853,0.41782352402874223,0.4209776301631888,0.8682356498443299,
 0.30078032375797215,0.056882336791842736,0.42515072762794415,
 0.163012799445159,0.9366505142794687,0.441101704423546,0.5066482119334303,
 0.692226303358285,0.37531798638309044,0.9374945768481004,0.6353636650334131,
 0.15220036631891576,0.8802107849093436,0.3966800681514374,0.5075971404252011,
 0.24739257820553628,0.13633093379273808,0.06058789264309994,
 0.6626964507706531,0.8698317157001952,0.243556930640086,0.7570513463826347,
 0.05355777582281862,0.907438917261769,0.7278341970451329,0.35428231422459544,
 0.31118899720686477,0.022511632852126984,0.46099566301651074,
 0.9871481571237051,0.12199579860278298,0.6554230849009003,0.8258391743909038,
 0.613591005893287,0.2684386485538156,0.33385170443140166,0.9778214495656946,
 0.9058132387552265,0.27958865066123784,0.5531034781027029,
 0.49119212934469314,0.9412408465190317,0.6372765407867125,0.6288402077957361,
 0.13430404233098758,0.12497522104712044,0.8581706003460778,
 0.6538950480225144,0.6968073699819571,0.7447551222885306,0.7554453877664193,
 0.5870611541428485,0.6945381798375369,0.4857796105985669,0.7535689330869859,
 0.8539359229090782,0.617975242930183,0.39348482510260585,0.32657283231122336,
 0.03883876075722292,0.04638686905846923,0.6323820875298557,
 0.6024000392838235,0.34796557710066645,0.506496603234308,0.44377965371491745,
 0.055291025646901915,0.21761290761614127,0.07266579185090116,
 0.08193488472463895,0.6851829302115668,0.054769459087122274,
 0.9553959970257299,0.2907480432362237,0.22138788497269346,0.7627735448983538,
 0.4575808197078154,0.3572847911574245,0.32031121917366245,
 0.29158371983483933,0.5146062069223841,0.637301074406841,0.14767546040001134,
 0.8570858069758737,0.467854169761288,0.8198994549061277,0.5109517721569086,
 0.2715354317360247,0.7197217284091373,0.2955306588528498,0.5023886706759546,
 0.25886603379942064,0.8326036093895461,0.4008178345802106,
 0.012854652229673302,0.4639105510010536,0.2678910139766544,
 0.3686265440103855,0.42129202866921256,0.9208431814228614,0.8176360608325263,
 0.5425940712690757,0.2136255025697159,0.9062618532803672,0.05381468858926741,
 0.28354185842035307,0.5336366834772812,0.5634763309374223,0.3816972490425684,
 0.383677994449542,0.08686952454675001,0.191481321438691,0.8117748160898718,
 0.5838045700781118,0.7458080689956356,0.5816405946572561,0.1445150202685559,
 0.41910794302270715,0.7450225076609042,0.3022129819769499,
 0.17752322902907147,0.5836807883401259,0.7327341710321157,0.7077979119480572,
 0.8488796457495889,0.31448121392218586,0.6667427345546586,
 0.10725916206929609,0.9040397923059876,0.8564373146349084,0.8041797093615405,
 0.9765266002862645,0.2699039452319486,0.2881126322269876,0.26286989748176626,
 0.8189221005341666,0.9389091041330189,0.10839298406771125,
 0.11302758151703962,0.9313063845210515,0.8412955318099802,0.2351596923112942,
 0.46364508771400326,0.14734773125584866,0.917686919154226,
 0.20536254186373204,0.3376669423839296,0.9357182107058375,0.6687171593798248,
 0.9013897133942472,0.9779087959032466,0.9040030533822646,0.5124827045393647,
 0.22971392418216818,0.8870131344615607,0.29397931244116593,
 0.8270460289427933,0.2440247535836072,0.22258764806804698,
 0.31348118869255837,0.6509665511434541,0.2070866531111939,
 0.28209212141002216,0.1870021923117412,0.43393644752508354,0.681717639325129,
 0.007426591644677738,0.3878067730540413,0.07020178659159682,
 0.9025809672852303,0.22059787417865784,0.2864296827322592,0.7326169672025304,
 0.646795347791104,0.3460916560650814,0.88068901306946,0.9834898135473606,
 0.9420657720934627,0.5762420622539801,0.8460345477682173,0.6367962576887378,
 0.6157736077864152,0.7984362373774438,0.3420703204633565,0.418630654426033,
 0.5947930070858521,0.40028720284187747,0.8998506806007316,
 0.20903008904231135,0.4598388264820157,0.10436096108659632,
 0.19731656284419963,0.6331440446623012,0.7532588400431731,0.4657928570703447,
 0.15402691973393645,0.8175992451528807,0.21017299034445147,
 0.6385707643198311,0.5314160572064291,0.8988308646364711,0.26270455552556404,
 0.4986415364731065,0.9843974504151278,0.6214180008320912,0.4165806924554428,
 0.32488188745280877,0.6841061795021275,0.7844215409543911,
 0.17322074562904632,0.3864540177355592,0.029598029887050514,
 0.6618199231207667,0.7086974679961786,0.9944162580666939,0.18727363475334668,
 0.5270667607321663,0.7320345371655314,0.6832373611772429,0.6708726961504251,
 0.2776906376830338,0.7707192246599215,0.29813249581596546,0.1263791027229425,
 0.07049674161991504,0.6321898093942876,0.44426960609779775,
 0.8968762897866474,0.9368349827151762,0.04843541562682874,0.7627770512747761,
 0.6341840958896201,0.2262380358697561,0.8466176925198267,0.07781837691947224,
 0.041491847442990615,0.10726755498956819,0.30437015072492113,
 0.452442088912722,0.48124905694236886,0.08552147605367866,
 0.059109904762013055,0.044991284958683875,0.1242818663656512,
 0.7104651254579255,0.7792522260723451,0.8082167877040453,0.793818977936421,
 0.6931249651914468,0.7464639710793929,0.48418695713542337,0.9622860657899357,
 0.5924910375027219,0.022567764726105333,0.8473840984765713,
 0.5317014829089644,0.21065824032296532,0.04896705646719557,
 0.07426219710224047,0.7915615710011448,0.14194591888879982,
 0.6423417834964723,0.07844900978581526,0.4328859118320425,0.7421632289762775,
 0.780438934041011,0.7042994138936223,0.11048122364358315,0.12215305396351928,
 0.6371772007188503,0.9486284569435568,0.6360981305798988,0.02714117811201644,
 0.383169268766989,0.8791165607162036,0.2783080297448032,0.9455443859840669,
 0.25908530457651024,0.13069931480441477,0.2353581787153557,
 0.19990382741668877,0.6415248465757938,0.8964131345977485,0.4504964768100179,
 0.09742143000794623,0.5658469368216025,0.1995797600208653,0.8264421497874869,
 0.2969171887000227,0.3488392390962364,0.31123904349771925,0.4132197527055377,
 0.39305527500949244,0.4497439237054822,0.05805940424601941,
 0.12926987315709593,0.17869276788750255,0.5453431173802203,
 0.9654341688091718,0.303584682111142,0.7360357684460792,0.5572567316058414,
 0.21664688308742097,0.5939048237390484,0.8424676785998699,0.4011340713814131,
 0.7132875464450115,0.5144483619312789,0.12954080064123674,0.575139672262652,
 0.6568978664176919,0.36785207553632215,0.9542587830670887,0.5431527341560406,
 0.6243641307853321,0.7542518253059979,0.3542855279731626,0.20756655038178695,
 0.8459299979487265,0.3984384582245101,0.5177751947015194,0.29036148400630923,
 0.028905054095735627,0.8125987008076476,0.805019620466787,0.5473808283221392,
 0.49315725400758004,0.6194864409114091,0.32577169773271375,
 0.7145873534542938,0.6291956978083658,0.8693386726220148,0.6323129655328223,
 0.458361185296681,0.8172722954957683,0.8245801890652467,0.8590419002935133,
 0.32521326012756924,0.8803093883282831,0.3382890545238233,0.9054726168157387,
 0.6793519075760505,0.188674045803494,0.030524636107632452,
 0.021584112326436977,0.65673206499759,0.15161018327834763,0.2792454652331474,
 0.025833748954839808,0.49250434990760894,0.8436712088934188,
 0.33139913206051563,0.43389030850901256,0.6131379023326133,
 0.7069786201148988,0.791943236196739,0.5348423593163463,0.30325546602542763,
 0.66560375314341,0.7006038695006727,0.43037708622879056,0.4902350790420167,
 0.791623579078441,0.8392173118588344,0.9115032719515134,0.789733293776375,
 0.5284196869266568,0.7908475005900593,0.2837289359830323,0.4596604305560975,
 0.21648116010536156,0.06639731293040563,0.6229056664031689,
 0.18576956697923486,0.9620506366516354,0.22316050125083153,
 0.11088111009807289,0.7666944352532223,0.12948198193164284,
 0.5970499119251799,0.46626811483848196,0.9350054640969815,0.8543151074363946,
 0.12490776367133816,0.8963231263923455,0.19968598138209803,
 0.6579758513726872,0.5245896219882196,0.05550906677440959,0.2854497570760298,
 0.19562051113871426,0.08471565671872783,0.7486886669549044,0.54185701630381,
 0.8946007709570143,0.9794455762785174,0.3098041365059979,0.687651155132479,
 0.10283029532482113,0.08118048966902791,0.9839423600982832,
 0.5959133379205891,0.9093236716984152,0.015232423017141006,
 0.8991028342487238,0.03530573233866696,0.728670839489852,0.8838462679902686,
 0.710927718374767,0.4323083130408617,0.6872104315241212,0.4362934348291926,
 0.6449767536462467,0.6732576102946608,0.8090718296716457,0.6453130502428674,
 0.4545940548786451,0.6907655460773883,0.6728558428101455,0.9760026368139213,
 0.4197799617895993,0.34810122669146504,0.7674777541334953,
 0.00504362623316823,0.8923443009813589,0.28813312627683396,
 0.016625827833462448,0.8730390610931251,0.5825713767337526,
 0.4482558738961011,0.12442715327987541,0.968055341419946,0.9873898135945057,
 0.7024991853014927,0.5000866893842927,0.14396688153774195,
 0.03996272492298014,0.6303323025545298,0.9725153546885831,
 0.48547944392567943,0.2916506605403787,0.28935177320766614,
 0.22682070471224014,0.2903872378336423,0.4758313478264944,0.5157517034261488,
 0.862572631542488,0.5252248040553362,0.9035167374780891,0.1914492262438232,
 0.1535481244523459,0.5224257195702611,0.06568810352525989,0.6273165941596366,
 0.8402307712720569,0.9726347258087948,0.6296532953233908,0.1521887376182457,
 0.6991758869060978,0.6655495536046463,0.9656336900583786,0.9583670096491792,
 0.9358802187874922,0.4884875448736403,0.03282646864388861,0.7828039961106095,
 0.27986427891824994,0.34976403231017095,0.6158955227640603,
 0.09803343329925873,0.22200933235512443,0.9009533381247479,
 0.6617420229117306,0.439751964791955,0.01786185147202657,0.12030554011500261,
 0.7859726582160173,0.9372210091998395,0.37027435857520274,0.3152930802549494,
 0.4846026525122845,0.9466394923474566,0.8509536700210081,0.0916353355838071,
 0.919708075038677,0.8630562584148534,0.3834938853561314,0.4794795899814054,
 0.31384418099767053,0.8898113794636952,0.48399496501186756,
 0.8905725428719286,0.7526777319537626,0.8222050333113483,0.8704353221840324,
 0.07203206618313784,0.12511583274719362,0.7791945704215031,
 0.5602801661719868,0.5251718469281705,0.02265727829712627,0.8443930108500238,
 0.46397133156947246,0.11077169251713859,0.6197562126368872,
 0.9202073577546659,0.6807804040005481,0.5586519015293202,0.6296891934868811,
 0.0726190778884469,0.6966509832700475,0.6252539398993421,0.4686388956635871,
 0.4407069389420333,0.6211524025167308,0.7320667369016781,0.7040477398331373,
 0.9181704785587416,0.7057165103540685,0.651997989425846,0.8144368033410612,
 0.3447491552596519,0.40699976843864594,0.9790269174138637,
 0.23054079451191634,0.17776860069923395,0.5348842094412979,
 0.9749168307539808,0.7966601851229005,8.140781377263817e-4,0.702063334618989,
 0.431855229389578,0.8949170255743248,0.37505922293777894,0.5770171973092822,
 0.4004106596913649,0.3484133102445949,0.05976036834076459,0.692289793057017,
 0.04176832086107152,0.21412592258978702,0.01574387600811622,0.2885216000923,
 0.10895524847721572,0.9163733160541427,0.4431680966805379,0.2050243246431379,
 0.6258772825911775,0.30710071299488295,0.6307630992017165,0.9520491976350381,
 0.11199604384073891,0.7232894471961536,0.16232818721265108,
 0.7912596781494996,0.27639626770100423,0.15790740960780414,
 0.1773717342890596,0.40726896160188375,0.47468443977283536,
 0.6534925372388293,0.5683184349593734,0.6814844338281258,0.837904666951411,
 0.9693452939900198,0.26571751194418214,0.523740885211073,0.9461311301386348,
 0.92830923091406,0.4112013134779928,0.6088959585299237,0.5817857089608987,
 0.4732840174024654,0.2529368397912297,0.9455486448114786,0.49345292864030454,
 0.48689194938065494,0.4399660374953711,0.9687262807688553,
 0.14023046204337586,0.9199758126041417,0.27809932401529003,
 0.6917139823383011,0.935870711979525,0.563906321251936,0.09486960995379146,
 0.8657179096092016,0.4806150552177265,0.1056319836133427,0.9157406632816505,
 0.4480949925086797,0.984765533113574,0.4308177118614658,0.6526896916675087,
 0.4969580664387103,0.3830670330245364,0.03656564688954955,0.9334414948080364,
 0.10499527688776455,0.7398798586499176,0.7496394035576225,0.7512229385984013,
 0.17418230098359522,0.24854574224444337,0.3347990264577667,
 0.7722401685945799,0.37869063090084376,0.5026178364853527,
 0.22826506288913773,0.5027784237837243,0.535460985106007,0.39560614486205825,
 0.7542898801806688,0.9801202518873638,0.8793788027205502,0.0667406496362799,
 0.3859795658193542,0.9369783846986588,0.8987394052433082,0.9726257328228525,
 0.552404991799745,0.03776191433610132,0.21348869098780288,0.7437804893431045,
 0.4009880504498603,0.20549357831587955,0.5569952194795708,0.722929932909064,
 0.4338346265309365,0.21173993110659795,0.869474721384725,0.9232307147209456,
 0.47382889493839364,0.9207693034480802,0.32478378114702733,
 0.5063148284020436,0.06679211192979861,0.845092910923093,0.3391656601509305,
 0.24281373648416205,0.8230872042927704,0.7776205902165622,0.5613669231671956,
 0.7526597298625672,0.7714009875234924,0.8023907661261315,0.11693823129281822,
 0.8047576889501069,0.40698767570585104,0.8478981423058601,
 0.33549728866967854,0.2724545066349815,0.9308786957984017,0.3537298790228096,
 0.8688319103294875,0.41355139439914435,0.8951919414241021,0.6896769258348527,
 0.14580863102330355,0.7049810989016656,0.45829493344832495,
 0.8702045690513228,0.6780265220856793,0.27540091149901724,0.6356551716408616,
 0.46581093821794484,0.04270704657583113,0.5645291923071162,
 0.25618579299648914,0.30787211978349205,0.22369869494421335,
 0.9897419830983629,0.8381642181164544,0.925021479982223,0.17681432780292572,
 0.9211927179797087,0.03220131623552047,0.3544021170058864,
 0.39170659234789995,0.9061741709409202,0.4095225699559616,0.4243685610369792,
 0.7029922563702425,0.5264993106419558,0.10262433096013235,0.4745912793217565,
 0.16886345619050647,0.4485675952763841,0.21460940996201516,
 0.21306975659585237,0.08511936507041529,0.44908211898305994,
 0.9003923892180671,0.6392771028904003,0.5919820965114792,0.1318259919278657,
 0.24369320680932138,0.61296846994891,0.6545693014335748,0.8004108092938125,
 0.29194947488814227,0.7166493746761021,0.32181309755611476,
 0.6874080147181516,0.9903018476889287,0.9697127945841286,0.6429831679685729,
 0.557928334542565,0.038564869742424235,0.49245331196819087,0.414796886810879,
 0.8340464175270417,0.7753981647026464,0.04518537517504084,0.1167679782348846,
 0.406789193760684,0.18158685039548572,0.34634850017540686,0.1329367979131134,
 0.5490565994154532,0.7432878661468751,0.051198822229175045,
 0.5209239215673097,0.8223241892432275,0.056044651034223736,
 0.9182207929337834,0.7654498403101836,0.7259543145564836,0.5728866824202082,
 0.5074432203683041,0.9315851418173574,0.1021868585758624,0.1779648224943785,
 0.10743120005550855,0.9937573420937603,0.08570217039481731,
 0.9591384034861294,0.5840272004911111,0.9502874394674198,0.8451655247261565,
 0.6614646426747953,0.8833994780932878,0.8159335074538461,
 0.044908434847568235,0.6980566049361492,0.8403034724238163,
 0.2114715377294656,0.35083279804658785,0.2264176866673795,0.7383453098394912,
 0.2534715521836318,0.3630914322131338,0.35281682190536967,0.5502859694314387,
 0.05832951966152333,0.911683645719858,0.7525165397214063,0.22223339670125408,
 0.5885923135777663,0.4757151235258751,0.7106387021364616,0.5525861523447285,
 0.6672793723576321,0.9109181735130384,0.6357578309190864,0.191710198672721,
 0.3364976406361493,0.3525240211516547,0.8382135670340093,0.30107677914523134,
 0.5966341853039081,0.5683518956678709,0.5993363163066544,
 0.027194716709700084,0.48368642313220467,0.5599318506821074,
 0.040061849153534834,0.5915846519019755,0.9134428642373105,
 0.06543285490929207,0.5114707005725817,0.8671278513938496,
 0.10222451588206749,0.9695788462132905,0.3713624447909081,0.9822207049928903,
 0.04467162520856094,0.19896289765560993,0.8935300566712464,
 0.5806809410434373,0.1257835855737176,0.21371350052364646,0.3411802152872774,
 0.8976275708179027,0.2916368126512625,0.698927266599574,0.6756770010969636,
 0.8755530957903133,0.36411794432140254,0.7694706633396194,0.1606714315460701,
 0.38323047808655697,0.9422713236277841,0.5164076302202135,
 0.49299985590574713,0.696628553572393,0.64767900449558,0.7979470088022756,
 0.6797791978388161,0.765183480562317,0.7108358798973191,0.6004545205741505,
 0.9291147155535443,0.7321410125828149,0.2565314887823872,0.5775135662883161,
 0.6897420980738513,0.833593285321603,0.39061171537317074,
 0.052583161161090075,0.46320723055151536,0.9014133334690946,
 0.25872664922957134,0.4020962050991146,0.3231117343316171,0.4404317673633652,
 0.785059725171235,0.9728412994243184,0.12205267918722129,0.12986057634369697,
 0.5886408646368155,0.8596266552526579,0.8072305637189924,0.6920256722818934,
 0.03336755580495865,0.6853724821084956,0.6916317535690465,
 0.028793341829427943,0.394253844459716,0.8935022820656541,0.51626123668007,
 0.05928938213210677,0.7549539815203675,0.49959363807332147,
 2.4362321838833356e-4,0.9155998604832589,0.2372644085374701,
 0.3081772744118616,0.08983599083052485,0.8637468191805321,0.2694440428771413,
 0.027443935101758465,0.9619600928701656,0.8255249760495654,
 0.46353598486321934,0.6798292582248943,0.8672491207369707,
 0.28335559969402824,0.0024265967738043015,0.6153068208212993,
 0.013026301376006688,0.45468445169938776,0.13194455711102382,
 0.5772579446319259,0.13018659559714807,0.09155732415062745,
 0.4510622719376839,0.768653421611145,0.005840344748779702,0.526380731851221,
 0.8066553504051549,0.9462443305955103,0.179488379388977,0.3352278118572486,
 0.6145010407512943,0.6835393131091895,0.3286517097948929,0.45760649998606007,
 0.24700314021843306,0.3941977317526585,0.6990243974401366,0.3442144725647517,
 0.22532662943738446,0.4259144734693644,0.2953325278157519,
 0.37030342351323997,0.5219046514050962,0.27266729596412265,
 0.3442831922614432,0.7201588070121616,0.7036705304576021,0.9654060890263146,
 0.3427841454769721,0.13540345085074246,0.17386418785887336,
 0.9290810035830865,0.4076083214154804,0.14426986068704217,
 0.23915835005021224,0.8283734851864777,0.3845240381745365,0.2801888157883434,
 0.2996075450916911,0.018275041283534743,0.5595784418841387,
 0.31954441172400694,0.8741380193526445,0.028254170711034643,
 0.1471116169843516,0.32306085000022255,0.5468920438967749,0.3510038351025113,
 0.3900946625142653,0.3233819816130421,0.6741135561017805,0.05778517865008137,
 0.7082939889101255,0.3298302419093192,0.8480897743966535,0.8374385800408058,
 0.06738765954785764,0.681465715653393,0.03904518929831102,
 0.39484802821395437,0.9693775211495779,0.525835870228358,0.5035562721829279,
 0.24285835875619888,0.21952420166844477,0.4665644644505208,
 0.8908459436320344,0.44998060891583724,0.12590804752183238,
 0.8539760025146991,0.01840417389902993,0.918110208850141,0.6317720298986479,
 0.30035313543258013,0.5865166076548021,0.5608425782983489,0.6028260487928175,
 0.6980685889592421,0.9009962113219813,0.8458586061935521,0.3316199334282578,
 0.1377062210847122,0.455447297671972,0.027317421455388735,0.3769782851566471,
 0.26522530415791323,0.260441540515582,0.6322588726939763,0.7158967423621856,
 0.5368374171125587,0.7788874894298004,0.7451379854841926,0.02447142711321304,
 0.10560973800004536,0.29917056542935483,0.04761357668086408,
 0.9357223060078317,0.7837391592944283,0.661309374296234,0.6222991489484676,
 0.9840814056744973,0.7100259662866371,0.7331052687824788,0.7345440054078769,
 0.27541578136353506,0.8619023996854652,0.5826185998635168,0.6944601229933981,
 5.703016101961111e-4,0.7865576337911241,0.37179405367347784,
 0.8105335370562343,0.7764898840043252,0.40013724905803605,0.8335113665046354,
 0.4547864069511185,0.5160791940798943,0.79808859626567,0.4273739652218259,
 0.5098219000276762,0.8770524085497576,0.8221537612167118,0.7059382433102273,
 0.17828173380787782,0.5562015571773284,0.33205778011089604,
 0.3916152460892792,0.7655171581408752,0.08173899190029787,
 0.43134988763136217,0.8925211586692637,0.12489714282248243,
 0.019598260108507737,0.618725967337582,0.8068594276806099,0.8941246513427147,
 0.3920947430262317,0.9401195934258383,0.9737083846795392,0.9827729633646702,
 0.01227826594428083,0.7002114828848574,0.8727758704272557,0.5273942944820368,
 0.519667394703268,0.40714890698708694,0.9371502103804217,0.8392012953080293,
 0.06673461268160186,0.017141658460499642,0.48981755106252356,
 0.9646358447135716,0.7197712471804256,0.855910199399489,0.5803569537744788,
 0.7252174049053339,0.7470159671039361,0.25599662970729475,0.9366342103195476,
 0.7538061292935799,0.6376658686262767,0.22820330608133843,
 0.15557353365036697,0.2616484125815648,0.8278505294262448,
 0.29293705735883724,0.9916539543441794,0.21853643692289781,
 0.9530020065799625,0.09126208430733429,0.2160189358081367,
 0.34747855749766676,0.6896609627130547,0.47975519983953185,
 0.9731231873584792,0.8175279182785847,0.36018689761062983,
 0.28563475079948897,0.7126246194513841,0.7428359239809185,0.6917683660112419,
 0.07222205324067943,0.797739290509428,0.7979304379699711,0.5917332143876013,
 0.8580826183325182,0.6502571097874745,0.1650347258066741,0.9257218400987574,
 0.6183396266993623,0.9609733112239174,0.47780084033173675,
 0.37749954985105383,0.980076526025994,0.8679540810688428,0.5940043390413228,
 0.4135720660511364,0.8227504645667105,0.2803388509018889,0.7652149835463027,
 0.6141319294339938,0.46169726533771505,0.5291704046985269,0.8536872875399617,
 0.9100194528672843,0.5767807734049268,0.9732146892472967,0.7077332177771232,
 0.005118614964676804,0.04348370633929344,0.6703926871249641,
 0.1689822418651017,0.8728396389276224,0.6232612405401259,0.1748997906546239,
 0.6416585045985084,0.6082917772725438,0.8185121069758592,0.6497701323592996,
 0.7692169986992122,0.9466770349298013,0.6394204172478758,0.9322737177610706,
 0.39768344056800475,0.5849032356662744,0.5594912719787548,0.6136462987033905,
 0.300733908941621,0.40771608435002016,0.6721579221282706,0.1239404917486936,
 0.1117743508365393,0.9518636075618897,0.3951708242737261,0.4238158734359677,
 0.1612868081413532,0.7885740529576365,0.360821614916903,0.6162293875621716,
 0.6319210074219348,0.8646639047889288,0.23737651295801,0.22509878132559968,
 0.6920002404997403,0.2193576117052255,0.049892224944698804,
 0.49567023416639244,0.23296752962063472,0.18684521826886258,
 0.1926875209045441,0.8537659707680255,0.8867631221731088,0.3416762991031348,
 0.18784378922682776,0.6140842736817905,0.7325426358560384,
 0.38058481367159025,0.6617405179848125,0.7077809306216842,0.9479576860011104,
 0.44740627102925057,0.9808548227261951,0.8573944231949303,
 0.10658556956363008,0.5853827282371705,0.4768327716683185,0.9691772060762859,
 0.863566671214192,0.9526113926718195,0.5959843604783733,0.028887989786153145,
 0.419255659153542,0.6551051074846659,0.8721505129606802,0.5705612659149774,
 0.16685800221526903,0.4269519725550077,0.27240070985442666,
 0.14570500953938403,0.7023452833759455,0.2768159299489741,
 0.22881203136432304,0.595654162130836,0.5805156640475886,0.9056557274225993,
 0.4328309229674213,0.9979715383153149,0.46827168566140753,0.4291093991392608,
 0.878762910211596,0.4008806303666368,0.392701048723356,0.8890174600234101,
 0.9630677389125739,0.4256338530660314,0.7448432179948055,0.06463120556921842,
 0.984888114514683,0.384148758491741,0.9287345564245522,0.8863174932285466,
 0.5739909444740465,0.03232300305558766,0.722399432735808,0.14451791705582273,
 0.4519979248884991,0.7157052991756898,0.6197249824904689,0.8034248870515172,
 0.3098798992123276,0.05904325554765022,0.747359608629297,0.6198801013128843,
 0.697128123015635,0.3092269531563888,0.5812304750545593,0.006180653087085819,
 0.29713153166606354,0.6330143741509453,0.0160795020322958,0.8469378618764483,
 0.7131081374708408,0.19128805067435994,0.6060301308683874,
 0.27380777060086625,0.09741175998198737,0.062381912278615226,
 0.6227146057379161,0.38898264738632127,0.44100091342078906,
 0.4916987345278536,0.3571096440113578,0.34234876386547664,
 0.23468689662976439,0.12346875483425834,0.6369594961978693,
 0.5231305691173305,0.39679348028171335,0.4088888546481838,
 0.015397946624678527,0.6778372861811658,0.5047031072333635,
 0.7560523461230665,0.5258420711140898,0.33172100992676024,
 0.38702948930086734,0.7879705482289168,0.4553468382484305,0.6693025188789269,
 0.40087320757612077,0.5573220600298144,0.3412107660842415,0.6998748698550052,
 0.2969056890728725,0.7481262101220084,0.8812663471444285,
 0.024584338372359515,0.8546790902420592,0.5160862282680317,
 0.6851992389695035,0.34467770405079723,0.6102707122306412,
 0.040188137451302786,0.30270496938145164,0.41959950994646933,
 0.9802295001757817,0.13421024067301934,0.0876725766104125,
 0.47802501396569075,0.8260873804541738,0.12121282488938401,
 0.0065683780723939655,0.3100301821365691,0.39542861548755703,
 0.8037451120561306,0.11291251256336121,0.3066509667799564,
 0.10698752136951029,0.9832202627524951,0.9371705878619027,0.623195178643756,
 0.5154091399890156,0.2664338209719983,0.2295529625341053,0.2244784798907471,
 0.2860821970802645,0.6058191899884914,0.565522249859563,0.6916331940518265,
 0.9333780390044473,0.7611643006938245,0.8089413935911516,0.8316755116486872,
 0.6815124136058739,0.16845649912958416,0.5510799294695246,0.7223927062452146,
 0.21892705831461512,0.3204161584328695,0.7011362603882325,0.8103320720630838,
 0.39905026813704925,0.7011141517831648,0.02842139623515949,
 0.6208194076671423,0.8506876498303882,0.6896440494453675,0.4613423486027939,
 0.7206839433934031,0.9782862482295313,0.9948179178011233,0.2695771956455838,
 0.9672264671905829,0.4815811741703887,0.5082622188759572,0.7293841522845828,
 0.14792748502321762,0.8712683862830557,0.22829990856163418,
 0.3761015042016378,0.032154467820636023,0.040604025509884734,
 0.3053325967442837,0.10365211334278435,0.852987670881622,0.1644775361200317,
 0.45469013221613563,0.9683030407969093,0.6169204214241191,0.9491623395903647,
 0.9698113522692026,0.16179934212508573,0.7214729716750291,0.4031119032896132,
 0.37248871879389067,0.019278855218900715,0.7349491905234201,
 0.7042402311196387,0.2323440053003908,0.22542592323963007,
 0.18845637615613575,0.5318279528211031,0.8660548320090015,0.8500391055155796,
 0.7758422446270196,0.7315683263328364,0.4098459228070538,0.44679296218324016,
 0.9514264466012842,0.12551214097118513,0.5229989393802519,
 0.21304290350188593,0.6807263032645154,0.4599860611254817,0.4072770395852213,
 0.6634348490875499,0.6368248843844561,0.4419150072166267,0.6189061441392198,
 0.4161011329721809,0.9355038072490927,0.2775034657074842,0.9727882580749945,
 0.061791749953973296,0.6611641302673281,0.06947018323513221,
 0.03620139051505167,0.1809935872505961,0.3984588728562971,
 0.009338802984677486,0.3379114164818602,0.7837940530472991,
 0.1778119158894399,0.4406710686282649,0.7460455958548593,0.8560120684767876,
 0.6541236405108158,0.4741776021980273,0.8878263299594437,0.7060949829317926,
 0.8680512429173413,0.93725118688159,0.35057414853022273,0.3025438102371527,
 0.44005988646829663,0.9386584972789145,0.5070818660982237,0.5374972362763882,
 0.7590076285565406,0.6888139178303692,0.2280955470277387,0.9837332455102119,
 0.7988446275430441,0.05886896753411475,0.7788884093615069,
 0.46499946048247554,0.6343005683368342,0.3685175203189488,0.3057938387607254,
 0.8578264962533579,0.23569651415794657,0.12190199308188476,
 0.7830266314292003,0.5637120036387046,0.12512247577063762,0.7655417719187969,
 0.3988305792012151,0.8489231425963553,0.47748761838098674,0.6116606342361124,
 0.5536448128099034,0.34319921526650243,0.821896939053085,0.5323002515719348,
 0.42163643985630284,0.879504282295961,0.7799440314593548,0.6560054169572441,
 0.7274455580135941,0.14608085528054882,0.03078731087472142,
 0.08648888673317234,0.7840517507865944,0.4053200411634251,0.546340234967703,
 0.6820176158498721,0.049203804407564,0.9748485498032697,0.8133389726756133,
 0.10413889680508048,0.6311564162795974,0.6677042389547311,0.29873171405245,
 0.6881586153566089,0.689761736331423,0.1494874465404843,0.5785968592946973,
 0.16952634570298253,0.5640946415151585,0.9037579623149589,0.827842993149401,
 0.890203074023455,0.7449847121237029,0.8792118638985036,0.8498887856246453,
 0.5750330150819098,0.42906360458666115,0.7079519344717883,0.7767522516690203,
 0.7096646001383415,0.3620381092213458,0.63500402521992,0.9315929851994696,
 0.167741215423869,0.3890970309517402,0.5288140493106313,0.5994357458313052,
 0.760916709448512,0.03085656698430428,0.561264416945004,0.6854710141804472,
 0.46545165440468983,0.1346820536556882,0.7246959576850412,0.4023772939700119,
 0.7045420099899034,0.9808754716321111,0.0019600947333106022,
 0.4471903740710311,0.8754228913227913,0.7109024559704598,0.14400899073763385,
 0.8767296453267593,0.4072661777224388,0.21217272963277078,
 0.21954977111116492,0.9364585995999579,0.21294890281441214,
 0.5843106549659633,0.28609533732950376,0.3086069116226756,
 0.33369416162644183,0.2758490900474877,0.8855487419914819,0.3151186210398307,
 0.4186080453738953,0.6574882089013245,0.7082890059336837,0.8334451084590536,
 0.22383913285003765,0.37364717419627747,0.40588147984755374,
 0.19994515438163507,0.8449438217389769,0.5336957003610063,0.7695577526284756,
 0.3350059962748524,0.6596371390254157,0.7020049330615621,0.3546185534759225,
 0.7080155754902367,0.953036999231434,0.8178290375933033,0.06512094956561865,
 0.278834714332005,0.9161232854014874,0.8782818194770033,0.219793778471584,
 0.36167585353838594,0.013650722489657152,0.1718284540943924,
 0.5047406884272638,0.9436431845855011,0.482203817510939,0.49091695353887055,
 0.7531293256812236,0.09744734689871926,0.28173656758720644,
 0.3882873740437295,0.29099728725376706,0.48318289829660455,
 0.39014848746287467,0.4965914808581108,0.9506555442391199,
 0.39605088397982136,0.30933391222062767,0.23173717298908447,
 0.15329566148455143,0.7867462150245096,0.6243610925499616,
 0.35958746824716414,0.036918184313689206,0.06550834236341718,
 0.759613601563218,0.8405657902984541,0.1445822025372342,0.9569493050724587,
 0.7208226410840308,0.7355619624839274,0.7722122890584666,0.3048627006682961,
 0.9049152273274264,0.2014174960626668,0.5370365729192338,0.3017676525031726,
 0.0053392724927529,0.9559698639057774,0.38586645738836,0.526040944619766,
 0.40340058338551144,0.16019922574538692,0.2391719794640953,
 0.4353566401562746,0.6407289976349463,0.012780949586688961,
 0.6067141746753231,0.39341429776590475,0.06740741494584734,
 0.4815277496315198,0.4849668225948378,0.019925213543700693,
 0.34565349569357195,0.5191197973763428,0.5845012393046202,0.9666645229894109,
 0.02135104073156069,0.46828256563341,0.2781330880584244,0.34763250313022576,
 0.024550686367633467,0.59658577107548,0.6362110928357262,0.5207235647563355,
 0.9275067452781897,0.7933029972688909,0.00789486948601481,0.3694218249514353,
 0.3602370298516877,0.7791340873318278,0.28187432446638094,
 0.13352100269115286,0.27297639431644505,0.36328359029224655,
 0.5105692522308936,0.07194750289025686,0.09111506121120405,
 0.0374746161837366,0.1802272142488548,0.02013593216211651,0.3713702222339488,
 0.8015854644706388,0.8367069923803121,0.21459294279955854,0.8112768671530085,
 0.5854430013166795,0.6022638766343948,0.7438036576466595,0.7140885874795647,
 0.6534228103425802,0.7838466731467563,0.3814479982587423,0.2135963655316201,
 0.5031530190061861,0.735056556635116,0.7193513928295984,0.5598868107486116,
 0.6082833172333326,0.9766999998128312,0.24511424188634245,
 0.0042032678314080085,0.8507096389294284,0.21231205321746194,
 0.5014428641323722,0.2612712797470471,0.7128083981426275,0.48398297998663964,
 0.8889723000684299,0.15187846260531623,0.5147648014082918,0.6922792840840356,
 0.5225942873084,0.9143067095075504,0.2398714365969491,0.16716750306316264,
 0.547034640398718,0.7641446013825864,0.521150703345331,0.15670065717717252,
 0.5867176599391304,0.7831417469842259,0.9393848616062663,0.06452613977367172,
 0.4182127294933322,0.3865182669333227,0.21150145927282593,0.890873461349918,
 0.2514172572868778,0.12868750452080924,0.6292207124220053,0.4290995279997326,
 0.4714970120662402,0.6611985619524462,0.033060439231903604,
 0.23029893702019666,0.3956287962797005,0.5617965309331789,
 0.013323267339054978,0.11605717764544854,0.41331359859319017,
 0.20366501026097794,0.4335466087579577,0.7101117356269775,0.7604168548563937,
 0.17965856767044408,0.7247534827267619,0.9930906614759147,0.9153871862687037,
 0.7774281801324965,0.8069049790762316,0.060367660055721806,
 0.20694728893558167,0.9842380424415205,0.6267449010690797,0.5910600110931119,
 0.5442895804211323,0.22307808436068288,0.8775632493284589,
 0.19832650777954308,0.7007763673064838,0.425864882822641,0.3231181274227305,
 0.2329268213050053,0.08169512805065948,0.35664125514366873,
 0.8439300475181949,0.7673666538816858,0.32935255999971036,0.7953561527987182,
 0.5299087601257424,0.8568010127462129,0.25832476621532274,0.984099317275329,
 0.8351139200154682,0.4774563282693818,0.8925653309013606,0.5791900458393513,
 0.4169787057199964,0.3113723307051498,0.14771418382962653,0.9534392403919919,
 0.5382154057114679,0.11447296886505876,0.18754369200114995,
 0.47826369500832167,0.6206470469515473,0.4655044594249179,0.3396889099144772,
 0.6364602244041522,0.9366416017716095,0.5803763275158164,0.09241470232689686,
 0.6992173293901306,0.8257027765678652,0.9723237008479734,0.08343676585714244,
 0.11040180994354565,0.45436904010022605,0.7469260227559439,0.882527557630876,
 0.02535826170697597,0.5125408084095567,0.9167854623956687,0.6073278713798593,
 0.882204355080297,0.24415721697362325,0.29676651255326625,0.5537021854172116,
 0.5190811802011525,0.9808495915438767,0.53397514511465,0.16384806530784735,
 0.07632592625070567,0.1404487413161486,0.4937854256718365,
 0.11155676622996258,0.0060776028995984666,0.12303824841890132,
 0.8853681422427522,0.735407415296455,0.9297098514977647,0.9940147483394125,
 0.8493684195339541,0.8702206834205037,0.5106620319371761,0.6080429037379197,
 0.5248800400041136,0.031727359034726454,0.8734410731493102,
 0.21196345495635427,0.8203342775393382,0.5555909002035956,0.3447979035726061,
 0.12320073167168688,0.7448758006223358,0.01183499920984965,
 0.3310143583911178,0.8576063606852597,0.6908718674463983,
 0.011125042044619393,0.052536185286421455,0.09134269355271485,
 0.6528788932706862,0.08995149733983188,0.5021021880353143,
 0.31073560397576516,0.36777100742243973,0.9816427093746861,
 0.20113721716990296,0.41017020322400455,0.09806188516369096,
 0.9466096302433932,0.9529241208393907,0.5091945130511897,0.9652019299845747,
 0.34148940567426644,0.8703820745772287,0.02963932930566726,
 0.2623660328181351,0.7373505100254081,0.3275918604502026,0.685505542205119,
 0.20486473269714622,0.8894076918468155,0.1397130215317932,0.6880204515633338,
 0.4577672620330955,0.9105408437390473,0.5153727794528692,0.5291439799807037,
 0.3820641781721603,0.21847384553836224,0.4362917567296458,0.7496037531131817,
 0.6337239159251599,0.8865636072728508,0.10329802956976053,0.7456440114367948,
 0.2881327777193404,0.2369947036080492,0.17272197541070033,0.5743075156121145,
 0.4563645414011356,0.9994679772365298,0.15622597420173645,
 0.24384436898726647,0.5881780845037818,0.8020888548630208,
 0.23317719618826982,0.7824023250756483,0.33382924729536445,
 0.6394144553619572,0.6205571082043878,0.902674885654972,0.4815288375226242,
 0.7598879301104757,0.4846979485228381,0.13256790890916226,
 0.39481832058388433,0.572823923255878,0.6172901194982829,0.3018545897796736,
 0.8864460073488035,0.24295359975078512,0.02936608720158551,
 0.2353474834957472,0.192612372958056,0.23972076029075873,0.16148480888464445,
 0.7170804274171639,0.7698866161209293,0.07505509650055242,0.9748149758555592,
 0.5227031320434474,0.8774409793541867,0.4570949002183433,0.836932433564062,
 0.3335030877227304,0.6858504710841514,0.8554764937865434,0.3556585631946235,
 0.055617118444849956,0.725239868496583,0.22449773969915698,
 0.6059661614419312,0.7355272803288442,0.5522830526639562,0.5640703335557535,
 0.4221686825367026,0.166056307774752,0.9972762552978146,0.1971388981441301,
 0.09443655021491415,0.6484521666034808,0.03936548049477373,
 0.47291389050499977,0.8562329534118042,0.3390374423370578,0.4662024996183399,
 0.5105601624210343,0.6066825548737941,0.2422973206437683,0.925860578936406,
 0.5806040231908164,0.8096994826891883,0.2189820495024255,0.317516045972538,
 0.15075264885748352,0.6974477053121622,0.4243065403995562,0.9652247840427922,
 0.8960246409142709,0.7988237745900808,0.789982071715773,0.19189301626609234,
 0.8384589312636025,0.08193579682342489,0.4063568537231951,0.8130367254545592,
 0.9030979320653995,0.592715693795632,0.6779320385305785,0.6036330887038538,
 0.3783455970526214,0.33128886804626323,0.34539293948791416,
 0.5410966439051148,0.8440984779264692,0.042503118026524844,
 0.16564523052074132,0.0021194493072961285,0.48183380875622983,
 0.9893935258138795,0.46614820214705066,0.7824475013454756,
 0.13364136563596396,0.8942632409517816,0.47421274934138125,
 0.43186370631925886,0.9421415384089143,0.6799432208423731,0.774889329338214,
 0.41130033992551107,0.7957919407427608,0.16830217596660046,
 0.9882424316273418,0.6766183057701193,0.3145276051505529,0.5106058512905628,
 0.5462404212398831,0.5709784474070587,0.11326962860345713,
 0.011019553251471081,0.7607380753113573,0.2211594010920117,
 0.19910317346018347,0.13282465379366215,0.13471784300778722,
 0.7823604382080047,0.7066467777969957,0.10796561190630383,
 0.08914676286255052,0.31963441320029695,0.9615930495514863,0.501341561199641,
 0.39846705375431457,0.8046992120374815,0.6234316544020126,0.9206067485133216,
 0.10641967430511179,0.9887823696969437,0.7031997678101807,0.848187946903715,
 0.2912244937819789,0.9381328712487856,0.9898812374347635,0.2314574241900662,
 0.12908515934123943,0.11654863009990091,0.044829957376982366,
 0.8915940227891876,0.04489553709056504,0.3146281683193345,
 0.029485485027605063,0.44776207764885767,0.4392106912830478,
 0.08975948719524451,0.9991524304253041,0.3616879444000325,
 0.45511868175994163,0.948614993321609,0.77491020508963,0.6046462137673195,
 0.11375539372885879,0.05115954112795895,0.3788348108689027,
 0.32183450003163927,0.34706597890619806,0.001976364928610863,
 0.37193113820440393,0.9002195679786641,0.7101418564724171,0.290980834367649,
 0.0864167953836078,0.9513967928462049,0.019980129313496242,
 0.6840964745436792,0.33177827635316093,0.7929434728328589,0.8857427714708414,
 0.3812717347879758,0.06953457875333657,0.4879618419553503,0.851515215283242,
 0.6832326078566229,0.3671764427740132,0.6447658452213957,0.16483808153873247,
 0.4283817333610743,0.7601468749434487,0.23521661878484001,0.5083642282631551,
 0.14858200314117886,0.5016707452466825,0.4193361036865255,0.5528333365451852,
 0.36669328816762226,0.14858640188098293,0.26342763943989456,
 0.2843984539580958,0.7716817573019447,0.9742655079318407,0.06733421810131035,
 0.5358782305335241,0.009080931391219238,0.2306731899618255,
 0.8714357029731747,0.8400640461905134,0.44667119653115195,0.6503072393314346,
 0.037251740406180955,0.18534315239634624,0.40559280981171997,
 0.9387651383896489,0.5040168369231978,0.7531901690740883,0.5559127703812986,
 0.8369416519735661,0.9445205634037168,0.6896177466479564,0.7412227036669575,
 0.7972344359748491,0.7347792054422556,0.35323153452052647,0.6920342318841216,
 0.2733398327772987,0.7676454350452896,0.04087418225546613,0.5171116662034092,
 0.7441003048547843,0.22932913737009397,0.9979928782042667,0.5106757747842229,
 0.932816444351052,0.25794967846818084,0.539751339138307,0.1185519026512949,
 0.8318381142902002,0.2755598975080725,0.49322429291485503,0.9559512251963821,
 0.08266013572852715,0.2711256029211403,0.49465027382619897,
 0.7507333262195396,0.9004126784994257,0.8695316964372679,0.7893261727571707,
 0.8620607105896607,0.9636232194770916,0.0697488631784331,0.8526213345216156,
 0.6891641840297982,0.12203576543231653,0.9931569978791991,0.8704822717311907,
 0.7168157632277977,0.7497344041985642,0.309070300858455,0.62855416889608,
 0.9180300029049498,0.12704162046887368,0.8036431024208741,0.6899127651921234,
 0.40884966185717075,0.4015321854208682,0.5967284556858139,0.5787397076191515,
 0.14427691249888253,0.4344610200356901,0.40667437164498854,
 0.7665790860231606,0.18860615886522603,0.005228273930159788,
 0.49210592317140645,0.904388522899222,0.363125571970603,0.77559215666097,
 0.5589870177056061,0.6994659070471992,0.5133310328810348,0.04988537703882545,
 0.014134904616362576,0.5217078903903986,0.11096754602664738,
 0.8015920120917727,0.37995121875327387,0.36268553585195695,
 0.5982295973698615,0.8681965906971687,0.5580219655664375,0.38362371654382377,
 0.5646444585684451,0.13002540565515452,0.7478124998957366,0.5718192211921865,
 0.417942211770739,0.18183031185344312,0.0927249718137485,0.7847535138668257,
 0.2901174714995247,0.054886623781138955,0.9835405608601749,
 0.04489639018465963,0.17816888268503894,0.0648342019036916,
 0.7354325082103502,0.6336771507396078,0.9865914510477019,0.5667100504167852,
 0.2201432711199276,0.08202004040607891,0.1793682507415244,
 0.17010936162890544,0.5195123839255421,0.7639824583052094,0.6140639594359617,
 0.21884681188222,0.6631464752759377,0.576803189001661,0.6187482940980048,
 0.4286829156557104,0.1965769388932228,0.32498666076238325,0.5487970233858571,
 0.7477985346960578,0.4672707767824672,0.9357082548347726,0.3515204753135137,
 0.30850950705088687,0.5384231921010625,0.10701037925224244,
 0.13025828045037824,0.5285874892288102,0.5803915033578959,
 0.23771085173930917,0.4702377962375688,0.35203604999542804,
 0.29187797324935827,0.6940205905317416,0.7852830166784001,
 0.47048765866740927,0.8651356872595872,0.5634166135104817,0.6729608223527284,
 0.18685260638030243,0.4153866366681245,0.5765451041978611,
 0.10456960205864241,0.8657603787248991,0.6706280711333896,0.6258159737072964,
 0.7970023097442347,0.6662009743302365,0.23732563053979727,0.6737913341358439,
 0.8152235741255988,0.7591874610085703,0.0577456367188256,0.7408306237386825,
 0.5992371453591927,0.029632766421492263,0.6685051236009394,
 0.20778839626378454,0.574416640716849,0.5970684970767054,0.3255296508312967,
 5.6611465696798646e-5,0.5185320235533272,0.9587771660753189,
 0.45286546287304974,0.5077059473664441,0.478585132820481,0.7305187364585716,
 0.09339860613793183,0.5683737767668705,0.3977374876038571,
 0.22382736363630729,0.808335090667083,0.23003514442978834,
 0.44330175164223173,0.3233201004571471,0.4373959097966107,0.5994692243356968,
 0.7315250398451353,0.5119038193473809,0.4774324897803034,0.15406223365361138,
 0.6684893455190992,0.32081317371763907,0.8134771642763996,0.7899059155450889,
 0.9768744462014983,0.3387246664183339,0.6562749433299361,0.20217846080366553,
 0.5108874637124612,0.9401973068464531,0.9383321253205035,0.5408924275628779,
 0.6962616262795742,0.430846650473345,0.2288703953606157,0.25750157748892843,
 0.2817247762775783,0.27173563558521063,0.5509209858914332,0.3307625510290957,
 0.9195962204059345,0.09358214938107312,0.5484655943816796,
 0.19041565500680602,0.45045920924756677,0.5668705922307694,
 0.6871460491783719,0.10049988214579775,0.40666823245927686,
 0.6714475929091359,0.9006655714138115,0.23840671900003452,0.4999906090658959,
 0.5801325779679953,0.7688750002367626,0.7428939010251459,0.25852435696565734,
 0.12466568258527155,0.9948860562542076,0.08023452119279428,0.831540564068611,
 0.13307501144063427,0.5909327422599814,0.5570195562882528,
 0.44748175773571275,0.44204023609283294,0.14872011197628066,
 0.9539129215879195,0.8116789757413582,0.3408529147712872,0.2018065726197651,
 0.4347381350373595,0.9229653977455764,0.9853550896071315,0.7974553194502213,
 0.0334386855149561,0.028639087419138787,0.8968948013673366,
 0.2876628511369542,0.8423587350445599,0.04154497966879345,
 0.07705702301376827,0.47621015696003954,0.19691296784196188,
 0.2648180323765905,0.9358859867667743,0.7160417485716474,
 0.024271372930607327,0.7666341303594835,0.20853841339189827,
 0.7761893429451097,0.9278566742883997,0.7666870351648782,0.17448788253841407,
 0.9624871283980762,0.8202034255244507,0.8143270951001588,0.4538887195787328,
 0.7844723885850577,0.6646272127695835,0.8297874573555367,0.9025022876657517,
 0.5623634418842182,0.4192152795299544,0.39095621306650596,
 0.49881838296936587,0.8487600914397333,0.46147127173499203,
 0.5513757115842628,0.38655672690010956,0.6711228680713976,0.8063793988322024,
 0.8402793458095389,0.18091506591137785,0.3091249794565396,
 0.21953166958058457,0.5691152126458331,0.455817485665851,0.435111071107383,
 0.794492754430625,0.6133473672441,0.7433785193471852,0.7817833319706876,
 0.5625703975135836,0.500393710629642,0.11550599032997444,0.599581577760122,
 0.14644336428883487,0.2060919310012641,0.30881961715515993,
 0.3691482830184193,0.8126868955405314,0.15332304996628876,0.5675478296984068,
 0.4239622612382832,0.1068822330155933,0.014178671182976998,
 0.6396724434670179,0.7235170509607188,0.9484192442491376,0.3603929508914846,
 0.661965202431622,0.8227898534250778,0.2900725055827871,0.8903534940946937,
 0.6051322299487634,0.033222359645382116,0.6100686375859925,
 0.9719176335192955,0.1741720058429692,0.38834798493227174,0.9628908515342037,
 0.079450809225446,0.15927348366652327,0.9238317423503808,0.31100439602674845,
 0.10030499589198694,0.6958340331260632,0.7755041926688278,0.5724385458367691,
 0.044136066455653045,0.5152432129429338,0.22091878212073723,
 0.35901255106826446,0.44512757772463896,0.12238444904000423,
 0.4981795649854861,0.14648314529581163,0.6641269862970256,0.8203040562790903,
 0.18806926441101757,0.8415422165807422,0.005025290108525482,
 0.6096299065391124,0.4050694965163658,0.08569053774405067,0.5307910638335813,
 0.29222933239825155,0.40451203180638784,0.30344544944149143,
 0.7448631488707737,0.7295813822417768,0.25204237383985784,0.6278815650761574,
 0.9758005570208763,0.9836213972329748,0.7907594802532414,0.3732001174344671,
 0.5345956683053982,0.6055952054064079,0.16411508320020007,
 0.44773831818987764,0.9591408252775312,0.20858444745133453,
 0.3701885706321403,0.73347718429927,0.8299303920719341,0.5594587257433684,
 0.2624538429987914,0.21696572869392483,0.9013185047069021,0.8766190819084219,
 0.7207209550005387,0.39260095828493646,0.3411202587954649,0.5908490607584369,
 0.285165723543259,0.9314344348768819,0.07312618364883072,0.20668570391159757,
 0.1644702956024754,0.40214575354944937,0.2037418664228028,0.9562664651790445,
 0.1742550824410254,0.1553249559155805,0.01818647556465669,
 0.19981831369224767,0.45878962333897766,0.13442109947391945,0.67829600918169,
 0.4948111384502927,0.1449290341305749,0.5380837862833766,0.5114716372854322,
 0.3354629229808921,0.8597680185273436,0.6536146122861165,0.5948534608904336,
 0.5496430513515069,0.5999253571612952,0.2160962137694189,0.33740719523815477,
 0.48558413705727677,0.8921044200572741,0.7267661648899698,0.5834480737764227,
 0.9986976177685313,0.5068747237628635,0.2141869287311008,0.8419417406999788,
 0.3731120481595279,0.6175237934990473,0.29070546026519595,
 0.43299752535520875,0.7223646516100752,0.8041721801710021,
 0.14079281644523167,0.9655201847761603,0.80902432559481,0.7175358105130867,
 0.7849845720635796,0.5715799725936083,0.4673225480486749,0.2191785295785269,
 0.5549299923787185,0.05520279591074739,0.3547056633144924,0.5487518596793782,
 0.0028230586029421856,0.010102575774019518,0.5952897522563023,
 0.4866582702972413,0.9798598721782222,0.3775110441779065,0.9401736425849863,
 0.738899555708194,0.9859039957395116,0.619546180912726,0.4612602458364763,
 0.7968897322223397,0.5458185259770153,0.4828592871039423,0.16858743766760276,
 0.05333800931237531,0.07978996152394457,0.02053712276778452,
 0.5523817776561094,0.9108992653501995,0.40011413590208267,
 0.047234909263996094,0.9237095298051765,0.15051588531683324,
 0.1956578274478462,0.759685828230173,0.8582642143915126,0.18602351859459798,
 0.16893055927587997,0.30434271323815254,0.9399467047235067,
 0.3893432323920496,0.252079606193896,0.3670903123030522,0.34381107869970406,
 0.9180147557172658,0.736959476208368,0.16171043642674832,0.04592858670908884,
 0.1882785634462414,0.22167827550025565,0.43054269731970796,
 0.8857013292866399,0.348110488123771,0.5245135195081181,0.003664521031452095,
 0.3574482559731339,0.5861685917358085,0.09358943666531783,0.5804741430015793,
 0.07673182226462072,0.3375409416582136,0.4816756734065709,0.2922960441699053,
 0.026275649208991947,0.4641419159993988,0.20448515200848372,
 0.3170115028239544,0.046183702477229804,0.4187232835797505,
 0.16133617398862488,0.23810447636262078,0.5386430657513626,
 0.3673383845024819,0.07283255923122933,0.8889727391244426,0.9277829035387862,
 0.6396955440273449,0.6660838360194692,0.643960058639411,0.7290717875392492,
 0.4694544280515529,0.685247236651992,0.9936968677292183,0.03619860882913095,
 0.07781774799065455,0.7512559760090473,0.6949448665679384,0.5313745659411906,
 0.5789336258473079,0.24153679792689853,0.896170558810767,0.9214624326360942,
 0.5274149051781197,0.33134838407756506,0.06454688077149551,
 0.7194194398626739,0.0988295432725469,0.6281409316164647,0.3207908672735008,
 0.3254868649544129,0.756658921179077,0.8116211345849991,0.381232432072226,
 0.16373954169296967,0.5952891591267802,0.28613783520229674,0.682034570388891,
 0.8019491347019917,0.037732250050226956,0.1871824472129563,
 0.026151374373941993,0.0878728568741215,0.8715086891538351,
 0.10831193504025238,0.4281736114029696,0.1900733370149118,0.3415449375325005,
 0.44459912424847914,0.9777478703120714,0.8871034132928171,0.646925952169195,
 0.6497290032361028,0.10211660570982828,0.07868559170300027,
 0.17850585962839338,0.1007756202818515,0.057535588278449845,
 0.05353836326018513,0.17579801137178308,0.7015333328747109,
 0.6322645875088827,0.39582740959881424,0.7352614497054868,0.1463298460884761,
 0.8775095073410039,0.37393957792624766,0.8984295030483954,
 0.11085427519999302,0.4045125862087696,0.05494355926590899,
 0.39502774606474744,0.41252030730396017,0.4618535780263202,
 0.2506298923284229,0.6619488043892583,0.9377075441668197,0.42111626306143013,
 0.48990322464831704,0.764889241821995,0.03764793746109263,0.9357723662196096,
 0.2341455571591734,0.9874798547364567,0.7305362091219475,0.835319419061588,
 0.2867139235843029,0.09102004325482183,0.6183830462859231,
 0.09093696888632485,0.11643067641860205,0.8872141850231257,0.358983990679984,
 0.936636221705009,0.5340541942893989,0.030916926509376497,0.6380347547222011,
 0.7024778739105276,0.20049376375511407,0.5337580689543461,0.4704801816148244,
 0.06898830277510093,0.29348896309788364,0.4339622229303467,
 0.8725592290367542,0.3665478095327921,0.16461791255674663,0.5504718716956168,
 0.9692601875487389,0.1849647850014734,0.6780448306552748,0.2766825618520875,
 0.8070491828441102,0.9179712349997208,0.3475287523763181,0.4654942273970871,
 0.6108121508581634,0.052384844459326785,0.3133639302763459,
 0.37540811098700577,0.6257270990446988,0.194041406880881,0.9024199181306334,
 0.1009868668213143,0.7885642257893654,0.8265314888782492,0.2912658002830878,
 0.8549719466777915,0.5379827486621301,0.4937047222338118,0.4866598388249703,
 0.9076997017344617,0.8940356550836569,0.7109657863693445,0.9839553508354179,
 0.48603278033781017,0.36508400181321443,0.278851752235874,0.843870517981306,
 0.30181900160876707,0.6654263040180682,0.24119576581982338,
 0.17976156203697935,0.3924687488712182,0.7191994891902282,0.9902624528927542,
 0.2487383744826558,0.6371273827783162,0.7902458657101946,0.7134156134256405,
 0.6864919382769918,0.34932588721846414,0.3753098846857186,
 0.33875166752874386,0.8395253227487902,0.5221585066921235,
 0.19865946112783994,0.7862134741482214,0.23451795171789813,
 0.9568123709515759,0.4153152084545737,0.9280112217384229,0.5980656131843681,
 0.6121536231575273,0.7463293048384029,0.17798168347597965,
 0.31474527428345944,0.658302038365168,0.8627573532880563,0.5485801664634333,
 0.8721851267112714,0.8067378411502208,0.8931912632375535,0.25919502932566274,
 0.8162838366765287,0.07667407105571622,0.8353162991910562,0.9373747508572141,
 0.8041001709082518,0.5839349014618767,0.9348840052120639,0.9528553898421818,
 0.4232120867337027,0.45732025699881274,0.8898720079323237,
 0.11647865930489443,0.45582040227313336,0.10870766496164996,
 0.9620115328418372,0.15415849538595572,0.519094002748681,0.6326186750442941,
 0.15277822182091616,0.9825941412501878,0.5078098997459615,0.7034679693068471,
 0.07653589626576973,0.22449039446826846,0.13130603657022144,
 0.5052221845624257,0.18807578140720982,0.47450039527903887,
 0.8279139939451561,0.9746046762541011,0.7407694051789626,0.4903175449624555,
 0.6304505031429056,0.9596748528069879,0.10946625135471932,0.6384774635937203,
 0.7292488001407207,0.32517421299567506,0.5242079569751211,0.1333859421381478,
 0.21248417348176651,0.3134072613179657,0.31477172108813245,
 0.2087325427822151,0.4203972849087342,0.1999410020920177,0.6559296366280852,
 0.25356783331963895,0.8742039752969862,0.7180671282663442,0.4182203188382503,
 0.38342504066012373,0.7071385559750006,0.3796446117309792,0.7532853079664752,
 0.9506922272497931,0.7129184606094727,0.12462425242344488,
 0.49980252655157054,0.7613977326098162,0.37012971451231946,
 0.8305060051967846,0.9385338555454467,0.03623384783317618,
 0.05777633252817438,0.9853522133130665,0.3731052466430178,
 0.031344727384654436,0.4255266977674752,0.1937020982132872,
 0.5095138885571424,0.265205615675446,0.0704532858147846,0.0998852537187711,
 0.25962072276376313,0.7491128064478508,0.3352624563775658,0.5639948803524596,
 0.30987236117741823,0.45073384720279797,0.9751394887980247,
 0.01756783208550694,0.8970798363359332,0.18849852679469303,
 0.6954983195200878,0.9486970455319391,0.6098961457644501,0.30706309523894626,
 0.395579296641586,0.8493545397465222,0.958971338289756,0.5241917492287761,
 0.7417088685243572,0.48637962967032977,0.2687333492696261,0.5185669550357278,
 0.3290592632358771,0.8387034509798751,0.07110748741738915,0.9428455525822296,
 0.3521971164767649,0.02879166254186738,0.3083936419038832,0.8450163027635578,
 0.7813484013980075,0.2001632180354208,0.976372604660549,0.19608151338748325,
 0.3728223011153947,0.674128287906646,0.4844264713065476,0.8923401554775188,
 0.3849801899815234,0.10782253387578689,0.8840745576958374,0.7370182875674814,
 0.42950831346964813,0.9396701425363478,0.9760364756144364,
 0.16161716327029496,0.9283877847705513,0.3196441171940678,0.8740798323904262,
 0.8074579786448355,0.8462895973945134,0.12365706686245104,0.2933485636107912,
 0.8271296218683307,0.6737649509298603,0.9094044749878588,0.2541077097581427,
 0.11495421680080975,0.5262394779367643,0.7931626943449283,0.6527085117589481,
 0.46634052488955147,0.9822180558100565,0.7512528515880279,0.8337339444968521,
 0.5220024085093468,0.9111797856165829,0.16931246864630278,0.8936671039484917,
 0.2452976799667398,0.27297904194805633,0.612708468563691,0.8459171864929975,
 0.04281003587658894,0.15784854099515788,0.9033537030561393,
 0.26474201919647344,0.359721601553156,0.6335395859635851,
 0.004391592907332154,0.37118194377841573,0.5890361470220264,
 0.9774800283180124,0.5689910672742109,0.33942680768111044,0.9512338311877283,
 0.8655824109602811,0.6748961979359399,0.8790296784688675,0.5524276346205629,
 0.6411182329379548,0.910521811180383,0.5634664467748944,0.06833798965800342,
 0.9260976262440984,0.05809315436248852,0.4246731369000465,0.7763375351178626,
 0.7066958528182812,0.9552560663729585,0.7030127239586441,0.5036691253460528,
 0.5281311620491169,0.9169319371355181,0.5127055720505909,0.04346881876104458,
 0.2513092262394684,0.8666948610909199,0.7285236832544554,0.2149047233946737,
 0.7193947705747098,0.4452629673188405,0.5717760623532083,0.2127511250218893,
 0.7987628304534955,0.9424341763509622,0.20965281336816688,0.6584462660488046,
 0.6021559860783198,0.4986586235115005,0.9001398707933094,0.9144346300030746,
 0.5302347490970988,0.5865813182334821,0.5674511597468908,0.48330469379886964,
 0.08698276085008938,0.9183405013543781,0.35990796530636693,
 0.45420433925915715,0.1351560317688536,0.026319332949463203,
 0.09721079308783709,0.5070817170878938,0.5460798846232974,0.8386804899505184,
 0.3921554102694432,0.8574927597981369,0.26001977331818593,
 0.06918013053788341,0.1019839428366186,0.7831716426049583,0.8326850963698268,
 0.03325140240020019,0.6093178215617845,0.15183844902318056,
 0.8764733931275359,0.424547866121054,0.23790747057835182,0.5719235455476539,
 0.8322702797102918,0.8679436081524624,0.19072429445464112,0.9865934248815825,
 0.8271539889407433,0.5057109900611277,0.001956239963573747,
 0.28918928493874674,0.662767134103101,0.33724713495490055,0.5050753444239721,
 0.3639815308755113,0.16237077227005337,0.2891357532182195,0.3954619609701322,
 0.29496852815568486,0.7223282485290216,0.8624530338194178,0.3926159224395086,
 0.6133717961584204,0.07403925283808266,0.09901553811442387,
 0.8196329259276637,0.34675262379890703,0.8583442254337996,0.7075282205373092,
 0.01274257111448851,0.3696430351430451,0.052307155047451204,
 0.8738466840833644,0.827159548482463,0.32167501311076374,0.4419148344969511,
 0.10325350164285674,0.9389719938624359,0.32888968449381406,
 0.6076751020322279,0.7588327309684533,0.22315121595397014,0.47217845833815,
 0.9598237943995084,0.8930257740354596,0.43163789917767836,0.9679012894561924,
 0.8161019974228275,0.2231026177426665,0.16633835312756606,0.9732514855005572,
 0.17741282592676944,0.4553449556306871,0.054663350881059114,
 0.3193352948335899,0.34928542271243923,0.822454509388321,0.23225595457394044,
 0.38146451681089344,0.4377149614298641,0.2591997032115938,0.7568070311754266,
 0.597223412793138,0.5839358698410818,0.1347815894984985,0.22751531158124727,
 0.9563054412996126,0.26582064182975873,0.8689500549513887,0.6708008937982425,
 0.13309805417320653,0.1403563858394814,0.9454104333536234,0.3598622859547008,
 0.6072463105992941,0.10772245664752633,0.4836443158832736,0.0487515940852048,
 0.8468092189452601,0.4511563763722347,0.33994346686539345,0.7370544659832281,
 0.6624189003408825,0.5372982624781045,0.9373540708888859,0.30715108608560104,
 0.6006282185369922,0.11811509753619853,0.42395955951607234,
 0.20544836876665828,0.969285936434131,0.41287684899778465,
 0.49456295331090927,0.8566880625262614,0.23079336175313125,
 0.7817062185571606,0.4902090161468341,0.31785187069255827,0.5482333042546197,
 0.21146127160195272,0.3506243614673943,0.052973453917887325,
 0.8834473689474194,0.8299551407495971,0.8352726069541437,0.17098580405823105,
 0.798838817012417,0.053070476335474015,0.1197823554730506,
 0.42273374859743174,0.9134204596625114,0.2724273423288003,0.7787535355559717,
 0.3186773896391144,0.07731364873820412,0.7481745706790844,0.9448642632190225,
 0.2316888259745502,0.39481399853458465,0.8786109622144214,
 0.053003818722799356,0.3494553399339422,0.4055761198991942,
 0.8868976426955537,0.9750390572765756,0.26776117056450655,
 0.08331725583811911,0.9557460753687117,0.8663575934503398,
 0.13627020609169804,0.6498734782072785,0.5170414708775006,0.7778353117542083,
 0.6078201555168254,0.3063066838961056,0.7694390757752935,0.3742136209116015,
 0.6468586698238485,0.6125602296265972,0.6588795676501208,0.7502638204571164,
 0.7275843413786397,0.2741508645631602,0.18890886461547163,0.7615567797265751,
 0.02251234754082676,0.919085491796123,0.2972635667405126,0.9228408418023345,
 0.8277164177923491,0.19226631281106854,0.5873251982137511,0.6306711565270358,
 0.8979624912958766,0.6954871406533448,0.4544608537028908,0.01685450861154225,
 0.4592008695618889,0.2465285286888257,0.057924287646704764,
 0.9653040055372629,0.05748970993130709,0.765473980152445,0.6988681732273112,
 0.7698315771289792,0.029457316338690753,0.40891404659217734,
 0.3063263600686277,0.006921517502146979,0.45001860164883145,
 0.811713923493637,0.22867111066231582,0.40266169717866085,
 0.13584090623869116,0.17529962024767887,0.04432687195033458,
 0.3299453149741278,0.2928475394555028,0.39660459707374673,0.5329090834907382,
 0.7941734826395893,0.9266750644325265,0.30776787509364845,0.8015869910859696,
 0.35840689681904003,0.29151607096869325,0.8094849338933112,
 0.27918099281071884,0.1195588892916033,0.8238032284080176,0.6529094530437511,
 0.2198469175466452,0.8745297188688709,0.15477770387079215,
 0.08568117436703238,0.7152952973184551,0.8187346171883209,0.8543633118731323,
 0.5440948723524295,0.2842459908228706,0.15034011954846838,
 0.46392442630774433,0.6436152114318799,0.5273464168739922,
 0.47658708622352164,0.8015352980788657,0.43510188695437424,
 0.18412791743101375,0.39666993291946895,0.7266820040477211,
 0.4153194102023927,0.30140577796160517,0.6337003261621712,0.6548085513262487,
 0.8170770033490785,0.7642491511347067,0.8482286149244692,0.9814279592831888,
 0.3523460433343393,0.5299389929948507,0.9225783360426753,0.6807591968562763,
 0.7417973167754258,0.6708375138631361,0.9703974437660021,0.6827913532298264,
 0.08144996405357663,0.235377696179358,0.16193779536400754,0.5743153670819175,
 0.4401022133521205,0.5284163168600784,0.9884234872697842,0.03080947944714707,
 0.8715313655977951,0.5148845406022198,0.03680110831439176,0.8061413380563413,
 0.41615324989074864,0.7599626493527096,0.5431615061332147,0.7313632442278917,
 0.0877997258451525,0.7423119146500978,0.32161050842688343,
 0.39716819213301213,0.43105109150214105,0.8855242344864221,
 0.46189047955481854,0.9176277162924138,0.4710389185747246,0.5249413383168854,
 0.8916550938871537,0.13400435346493422,0.17624641120959306,
 0.9659810664329062,0.6955261214211399,0.6583134309888602,0.2999276059502448,
 0.7924764861351569,0.45878232758857684,0.7308603922322527,0.4077513086501682,
 0.12426276554124072,0.44081986500365833,0.19794507092814617,
 0.31026866488103166,0.9520773079296119,0.07257354403292826,
 0.04847084414234215,0.6279136038796784,0.39894616334783073,
 0.7643583169008866,0.11931725496500933,0.6626245822655222,0.6564039663627553,
 0.01011167397889845,0.13528066694633334,0.7265590293576888,
 0.35326586081667066,0.8268818340665325,0.7141461835927099,0.7114977614308028,
 0.0921149852554084,0.9420727894087632,0.020602530440893108,
 0.6795978728637119,0.5063007241831742,0.6353445193639115,0.5321073588338177,
 0.4795627753762376,0.04860107580164308,0.24228959293387398,
 0.6967490350072096,0.15569775122151164,0.7763692971627686,
 0.38954169135212546,0.35117650853539806,0.0073810169081566235,
 0.8361056561269311,0.7584855332750958,0.8127409398692832,0.3631665246114335,
 0.8995135951312552,0.9475491706068935,0.7160207515762469,0.6765945888313134,
 0.6045656228609346,0.5467111627674799,0.4693511731248047,0.3948541817704885,
 0.19849180808305023,0.5595979289505244,0.7937820994016443,0.8677255964034013,
 0.014153396780016703,0.9748589582703268,0.4595864530038165,
 0.5327279373334374,8.250909810192386e-4,0.5269038659870949,
 0.6441445289214015,0.9134899851583232,0.3902724508721883,0.05736894209212906,
 0.24870588942766847,0.6049423439957735,0.08518529124684959,
 0.9628070715129247,0.9650822479804129,0.3850749464396501,0.8484215753071669,
 0.9140493894141213,0.5136117922706348,0.464826295603867,0.9882908015206293,
 0.9046318992537373,0.696545753662331,0.536549672247722,0.9098151801554138,
 0.49393000097063067,0.19744595457605885,0.47488036671576683,
 0.31603985664823375,0.2856728038427728,0.919481128133169,0.24156889559754002,
 0.6072191444061646,0.2179542066473924,0.7754825675292316,0.18001805464280485,
 0.33587907156313046,0.672298280158225,0.8682208829572375,0.8646441643990542,
 0.04458523562201755,0.8791945544897203,0.7596748695879894,0.074176528612101,
 0.908315217946927,0.1979498729997291,0.6894072099212522,0.2708866478295655,
 0.9393301759205438,0.5436964752353298,0.6312958632611538,
 0.039290855506423195,0.9122092367003409,0.37598546120489906,
 0.7790690816465997,0.5929656715249758,0.5259215309748515,0.9764664939709079,
 0.37444732636526734,0.08829337191928288,0.9533045741323627,
 0.5282170027976465,0.860975837782515,0.012070919923671597,0.5617260945787873,
 0.6165165544127177,0.2063327716604071,0.9611141025780947,0.43020258312373627,
 0.21273165318145604,0.6725554293911907,0.578294019335029,0.7417553909336255,
 0.2659842524918139,0.48116874286926237,0.8426506767768212,0.8071486434225141,
 0.35177952152184266,0.5128136543488404,0.9623297022014407,0.6969228346285038,
 0.4267393149308829,0.14045409271362097,0.31193255427277466,
 0.9800279012004216,0.9492155573743722,0.1115751963199274,0.47871018041824787,
 0.6351318743826301,0.2544938739141671,0.24705868138992904,0.8481084019952928,
 0.358419696802031,0.2183771434569851,0.4425117635492799,0.8881037977511911,
 0.5246858969975143,0.1432143184249579,0.888668895640712,0.9806817688156592,
 0.4598771248609791,0.3753344486664051,0.5115954441835697,0.2717045195489506,
 0.9149830433936307,0.7392459809970708,0.9203076807782383,0.22993418258118525,
 0.47593359625279585,0.47703556203669417,0.9764170653245783,
 0.7806647699836273,0.6780690751165417,0.08532257427325574,0.6053934542236904,
 0.43523204398930393,0.8507565775627653,0.8073608922058685,0.9521094613425644,
 0.025485044036360538,0.38553677171850276,0.5078887182734789,
 0.2473327972561804,0.9498735384092583,0.8486057865009429,0.313087164200864,
 0.1811824552142861,0.10201772028425693,0.23469832732064666,
 0.9825041645579581,0.6443533038418148,0.4511372859201779,0.2622852813457234,
 0.15154663224522913,0.6923387200684079,0.3062074907916389,
 0.16754450899090712,0.2065688792423661,0.2085107113081821,0.6760768317162916,
 0.655179032342942,0.49178663483647667,0.42889615930193614,0.8551103080766063,
 0.88772777083317,0.6947475399057692,0.5634399021935169,0.4725927075934996,
 0.12080031961841309,0.9150629120354243,0.18831717322449382,
 0.05820410872423509,0.1427070306164202,0.6580547583902365,0.1390470511602615,
 0.578935387043406,0.2348321515258298,0.0800704816016965,0.9175740131393338,
 0.730002336599248,0.728454870371219,0.16009335234905586,0.8482549984563289,
 0.11795325850436311,0.4983221747535621,0.6591856183810251,0.961425209600071,
 0.5523427491287024,0.5578130994456663,0.7794318606807363,0.7381169778156779,
 0.40470770942251766,0.1729967254232394,0.6163963274988469,
 0.23280685546426083,0.1407830723277529,0.6914544098936256,0.4478714173956706,
 0.4742742658425273,0.35755810688295864,0.5827747490064147,0.3552579413375603,
 0.8176325367631616,0.06658560770500388,0.4932343861452957,
 0.45397703402160094,0.7081692855868928,0.7597151107659302,
 0.25997643629221445,0.6919050097476134,0.1373422959139151,
 0.19552555161836538,0.8844170689900565,0.7579571163935177,
 0.16405362866875184,0.03765084006761388,0.3539592768264538,
 0.8227826231960288,0.22848174417891265,0.44824880906522035,
 0.9809597002145827,0.3970078620121704,0.7898762153307721,0.26084527615708986,
 0.5647546736484456,0.998392906470442,0.47438760478089514,0.4614506230977048,
 0.9785895782840727,0.38178246278897277,0.383859320927854,0.9108434571767412,
 0.6804101109822432,0.6552522153298704,0.21037751140950411,0.6952975879369218,
 0.8189047195109878,0.6297491081213893,0.424743661412595,0.5930418586880927,
 0.5206368597984616,0.18712257821013467,0.023168472747917557,
 0.8838652217595007,0.8955536142038216,0.16632897625084853,0.677955338250536,
 0.4758536333366806,0.49483460112560707,0.2221906083850097,0.9200952474480306,
 0.950895549940337,0.8819326145951519,0.2537104733409641,0.4389332705051492,
 0.06616509184206465,0.6639611675842838,0.7842770376027817,0.2465651263029517,
 0.8019462129037928,0.6748632479227578,0.4780937372904043,0.36548443142504206,
 0.16467113964039148,0.03315593931253946,0.7168005237712735,0.366110386790779,
 0.6670759250301903,0.8293472017269189,0.6123494045706857,0.5484234075830804,
 0.6899875524132462,0.14710508340807438,0.3943109277243362,0.9664023610843349,
 0.25595483618273585,0.22344149613625786,0.7838431204911537,
 0.3832711176461455,0.9826044273000215,0.2870302181313489,0.7924521000442503,
 0.1543616225998945,0.7847770965838693,0.21456438536113642,
 0.19439063912308285,0.1428361885275834,0.3731665563873805,
 0.056870084229640905,0.7745022217862291,0.6343932891901422,
 0.3823959761506609,0.4708173184876785,0.8587235322580973,0.8286936948761928,
 0.8571481589720948,0.49823673715540173,0.06497761878861152,
 0.18876628554115404,0.4835884671909114,0.9707426149513199,0.7488799016693197,
 0.9611033608205651,0.6463756804356392,0.13638962387575737,0.9355454499378605,
 0.9394807701805693,0.7850593116851486,0.23957043035539838,
 0.42010538405907316,0.2600005926979577,0.3437935568117336,0.3786712457126805,
 0.9747363780635296,0.9054264300860494,0.24982263434662544,0.7294112047902165,
 0.40835024004300347,0.13064247815634755,0.1843757757674167,0.92035212578059,
 0.6336126275463376,0.47948764691606394,0.3643840481800751,0.949944312131575,
 0.2966536643247484,0.6810784445781748,0.8527346451211022,0.9271000604525441,
 0.22076291347848875,0.717144974888786,0.20828088219903995,0.6629771394263491,
 0.5431380123714793,0.38740802913463,0.9870622123657337,0.09754178913680978,
 0.9752558663843693,0.794211742038794,0.5627390487713724,0.7895920194940695,
 0.8481206204321557,0.8411228784819061,0.34767872757441687,0.8956840262413605,
 0.647236673660136,0.793062948075976,0.27212523965257696,0.4093696969416819,
 0.48884152005076087,0.22184380870623122,0.8953658788392356,
 0.48000195863544304,0.7202735524888905,0.7695367209471211,0.6362879515443297,
 0.7709571826008348,0.8390615147203889,0.9573953300894246,0.6142383197459366,
 0.7246449414592285,0.6722039891614722,0.7655549043147414,0.8784237399895208,
 0.48573141394560226,0.18447825741876522,0.5755369630411121,
 0.5148740488325705,0.7775146321900943,0.05865553206327445,0.8151670351930383,
 0.6145385205832755,0.8908453400446537,0.7058026618195821,0.1962773029948146,
 0.915153821947657,0.1793314444918871,0.7513752908755971,0.5194734278117378,
 0.9378809548077993,0.8709112160123726,0.6686097780400606,0.27741987267558077,
 0.9842050584159386,0.07788495090371594,0.9942665589239282,0.6974212995096515,
 0.16220445111083015,0.5978059715302052,0.6530624198129779,0.8358782070333538,
 0.07222237733911818,0.20915555093719063,0.12008162536128797,
 0.3065894955782703,0.6503089555661801,0.6940174757316974,0.639552399364267,
 0.4504057867164065,0.07430415513487232,0.4569744932819485,0.8134322899325306,
 0.35145800030052576,0.7158125609938524,0.2375147026335973,
 0.020362344916020514,0.31649438297222776,0.9740729285333437,
 0.9050635197587149,0.3552121377790648,0.9710550999690374,0.19674989298927392,
 0.6783837412119512,0.4534497493786115,0.9246504767866387,0.6547710468343186,
 0.28369877739369187,0.5277764453500532,0.7619370077683545,0.167456787968287,
 0.2949790435356632,0.3732726334263672,0.7236459806947967,0.4343003922846268,
 0.4246760461118496,0.8703384663838497,0.9465674214063807,0.7396191474159574,
 0.5306188428837607,0.46648961505110254,0.502733372548415,0.6382305084243476,
 0.109902379284611,0.9256715770524524,0.3784843498468994,0.6571311654593003,
 0.6211794147085524,0.7084327787721677,0.5086266286328753,0.8936712957107854,
 0.8499259468196511,0.4440401348765818,0.33933278508881937,0.8116719154873999,
 0.673775626005183,0.49111629230560006,0.30499700919801553,
 0.18367732860362795,0.46945199938701243,0.07127358041972964,
 0.30769726493523397,0.17031859698181384,0.7943858103836455,
 0.3261946562514234,0.3515716207209796,0.10216793506344346,0.2780657991167984,
 0.7298507562696429,0.22139388689224915,0.4699304163362894,
 0.21766027034455837,0.7263898786347849,0.2078228318409826,
 0.13872725691202337,0.08302974518838557,0.7161633545438036,
 0.6277003848730608,0.8982265018012046,0.385280498710732,0.7992729210923205,
 0.8101613807084111,0.46106562324650646,0.6647948824145028,0.5161704152932483,
 0.03944888937980506,0.3977593720330289,0.907800632772998,0.966508804999779,
 0.9230971643360536,0.04788839081504381,0.2732111088619148,0.2671341273691348,
 0.5293656895461905,0.8354733418001601,0.07378255578070292,
 0.14017056368322922,0.14699114868235474,0.2032668366740472,
 0.12457781298669812,0.8159396637775034,0.4344268555784756,
 0.32370155732099204,0.028946540579767532,0.40115561288614177,
 0.8981311713343949,0.5551018438587914,0.3261153527125179,0.05630268880295919,
 0.5691595507325413,0.8241054285321968,0.19138349830575774,
 0.14710025385182324,0.32136192425985743,0.8492482242983354,
 0.8614917026089804,0.10053599614887121,0.24337926733706805,
 0.41687783100353837,0.6000062003079982,0.982893228048196,
 0.011475350971239529,0.32218740530248946,0.1553031850598814,
 0.1732764089046126,0.3042062739373723,0.6320703590044146,0.4208755174221618,
 0.10366283513601515,0.2819025402178347,0.15034681039436493,0.924069593026024,
 0.8876104334454133,0.5301745657493525,0.414977041597029,0.6609127890933224,
 0.3433132956295579,0.8767098993708831,0.5887586847402151,0.3888317469554482,
 0.8332978715767126,0.6470873427172871,0.5967109489646985,0.8255872956153664,
 0.29683125485322415,0.8517046558984573,0.781240508111343,0.10898920464114537,
 0.18967915178599315,0.741074379914899,0.2581007844025881,0.3773281011268431,
 0.015671201493596576,0.6262423769592985,0.2773086360173347,
 0.6483839076289974,0.8883483585747058,0.7328032635989845,0.34746835185829195,
 0.20880370348575228,0.4203244748301823,0.1304991916471363,0.8286900413106256,
 0.008048990812885015,0.9217220593076121,0.6271696688164171,
 0.1682053808425037,0.9820542188950241,0.1959833628651816,0.8354124629721696,
 0.832945705465042,0.7261296876789884,0.1713138081701433,0.4644533661863465,
 0.9102189019135559,0.813125339517405,0.006308255825695275,
 0.07417005161862344,0.5182413579996428,0.6706615058689721,
 0.27074184248113664,0.3996420458891239,0.32888898307152614,
 0.9298015267875874,0.7993414875818561,0.05455936623683999,
 0.24633004514372414,0.5324177692502013,0.4874316069841367,0.8897565554000347,
 0.0788564150288581,0.8775640513482259,0.23776246590378658,
 0.39466427514893154,0.9388465920496799,0.8263581120946151,0.4061012487348208,
 0.7850482518258303,0.03506042342992899,0.3774248469524455,
 0.03835971112704206,0.002041010345530503,0.63501995428937,
 0.25220067480105535,0.1693112695783019,0.9161574735339977,0.6934781045132947,
 0.12938240771406095,0.2564601187191369,0.5181869674962126,0.2908186691184643,
 0.42244469093012316,0.9549266362617748,0.2055792758056475,0.3770750648185499,
 0.9047126887613729,0.3365952363473149,0.6339378711365745,0.9931619211686636,
 0.7507523403166061,0.3678358004170724,0.11343761458485768,0.5050463281286999,
 0.024924530573653625,0.6334377208505085,0.6916336471371997,
 0.5259044086874105,0.24114637356337743,0.061842098822875835,
 0.6130296302931107,0.33595412117330825,0.24053582040205723,
 0.07344698979869546,0.6349672869519232,0.5180038852306206,0.7782659152566167,
 0.17528722405709174,0.4747703777403127,0.6698622172002837,
 0.013893740009054745,0.322353374714184,0.11200743084665832,
 0.05818361289888996,0.5932614330445453,0.6857615413509934,0.5319568504807973,
 0.8681450209463893,0.14445525283369764,0.7659892732363035,
 0.38715841914223914,0.30903061293533374,0.8236209563455155,
 0.08687940016469509,0.9351081712472581,0.3962400591940609,0.9343220819999668,
 0.41101468113810957,0.31276665965133243,0.6827247834003389,
 0.009425594402905468,0.7851428659040852,0.986358734272819,0.5306953539670967,
 0.5306814791741616,0.4404074560924014,0.7377800072538291,0.9719548018821883,
 0.11801437106313495,0.5821891879264671,0.05056034411051913,
 0.4546187231062385,0.982374546007183,0.7390846977486079,0.8348884183709534,
 0.9677559161752676,0.020696871888332247,0.5354806388460578,
 0.8035734409325388,0.4880532705520966,0.23984451239902094,0.8462176145154935,
 0.3694000413177597,0.3542298772033874,0.7411917140715468,0.5029300315249069,
 0.1112617823847245,0.8508657756161794,0.8558290365656477,0.6513115859088002,
 0.41069137368546815,0.5197053136114053,0.031490919112953986,
 0.8842407350598985,0.9235960572104911,0.6493620988885727,0.05195154863780049,
 0.9839555923978872,0.6384193616162679,0.18625802418215542,0.5133589608705149,
 0.23542731851666798,0.02182193540191235,0.18144445592491198,
 0.48465663093001266,0.3692736834362037,0.6351203172651971,0.4154560982027681,
 0.9416103314375285,0.9046594243103299,0.6937216850858046,0.6426421954380814,
 0.03361708797265839,0.5174759189350076,0.6705117855218538,0.9130300651860812,
 0.6102811406447538,0.9418116649772116,0.7596893472219685,0.6836112081045324,
 0.7175933467047998,0.5040690872882044,0.7078144289201104,0.9827695423601858,
 0.3923797076427732,0.498622897579875,0.2892643033147628,0.4000876436684412,
 0.4234516347100399,0.6481786789148105,0.7571159283500963,0.5174171921989844,
 0.5691897086482525,0.45592008629486336,0.2246856862164991,
 0.21226519756372486,0.8894427538853654,0.752370470078006,0.25915345944775714,
 0.6089751600893535,0.261977574115863,0.45078546113965534,0.2797918307177345,
 0.833475394506217,0.2303909527218987,0.8868154828744453,0.36792611170315914,
 0.6590117645267494,0.27577213373959275,0.5657789482259792,
 0.19774886110954992,0.3586403247379232,0.49973446644071584,
 0.43034775113435686,0.689806636260865,0.9765947268760344,0.7645416578979718,
 0.8573156914435687,0.568226541199997,0.00613153801581201,0.9191459346773391,
 0.11771556258182148,0.5394152247759761,0.7787466457205479,0.2888538125353345,
 0.6276972234421667,0.5209216946117343,0.2898632566412739,0.47589818666366623,
 0.4020700490300625,0.2651081439150863,0.2280034875416681,0.16958664262908218,
 0.5832719574960704,0.994003018378734,0.15063692497394943,0.43408811841907513,
 0.9743558692653836,0.11770130776770871,0.23005256858421397,
 0.5563881428599117,0.8681757892182632,0.2397807276837033,0.4731977370589141,
 0.5972856345763286,0.03642687966666813,0.8804217272188267,0.33950443833164,
 0.49468300689377576,0.7382945013941135,0.6575095483558574,0.7729538739544727,
 0.1157239337401682,0.24933241600857392,0.6622833011678212,
 0.39273066696384507,0.19817310668442523,0.22448683130128144,
 0.5188383855198716,0.5736833598702761,0.820141917003812,0.21294920144334728,
 0.30745295398915307,0.12824896179408385,0.19020123875688966,
 0.33477509253619064,0.08106003073278645,0.40005975859915677,
 0.6934518459505026,0.9746821881321281,0.12937343016310732,0.5542797883014119,
 0.543538173580348,0.8347148195774456,0.1768201428129691,0.8783429935659302,
 0.8320311152321014,0.9119942728562402,0.6082885173795486,0.8276867241803639,
 0.44712777041433704,0.700320394403274,0.3895901704414818,
 0.017234390328612736,0.6022490324015018,6.418861980704849e-4,
 0.5895713133801995,0.9332844270358941,0.32419816769648635,0.3106329035746679,
 0.07613386527450994,0.42166864491162415,0.5750090308081383,
 0.30499789832413327,0.284353340017505,0.22763895503192066,
 0.08984171690197229,0.12634006759181315,0.8984682817462681,
 0.3188228262825621,0.28998059073101745,0.8645732464231899,0.708363292840759,
 0.525336202589487,0.03477200502777078,0.31049029944159845,0.5014239909738781,
 0.17781971963830168,0.49940960649246147,0.8354206011022841,
 0.14504798389492413,0.09723680384380329,0.6870144907575014,
 0.05500262431154046,0.31563538983004624,0.18525199369578982,
 0.12209953115815664,0.24150159110569158,0.22817109800296864,
 0.19586077917929468,0.9311317914696863,0.7499892922187766,0.4559530469299071,
 0.1853273069616851,0.8455965288273648,0.42463326829685766,
 0.11313071924926243,0.07617048761314749,0.553410074246731,
 0.20979898228859772,0.6407530213760139,0.6879718946189792,0.7769451213631866,
 0.21036113418919844,0.9447902329146048,0.9631420510849354,0.534330332429018,
 0.5949030282582004,0.7794217873651808,0.08575751417866373,
 0.35660582644265415,0.3895383192824937,0.18978158768690345,
 0.9136775101358053,0.5426780375621866,0.8428871114483307,0.5651919536310932,
 0.24286733512087066,0.15489807196921734,0.8933168792881241,
 0.47901688151541444,0.6332957301159525,0.5845703326123903,0.6899652616867764,
 0.5681739307934628,0.3439182542086169,0.39012295136597464,0.5514743989497326,
 0.0412581259641146,0.9255080910922905,0.7174700042417439,0.46236775239830585,
 0.6473109849542179,0.22129952935600228,0.36215182803902823,
 0.2617220222987825,0.32852120974594423,0.13091589308739804,
 0.9261378649959197,0.8918143787402986,0.5401885108813557,0.9476490478454748,
 0.5832467758126602,0.09514193170774798,0.0645152937188842,0.7993154106042789,
 0.9923973386085029,0.5510004604477167,0.739337171842069,0.008818169301873291,
 0.6956267119986904,0.7283458952577786,0.34888059733861365,0.3563928356836459,
 0.9863056263579092,0.13696386741357036,0.6646038011199592,0.518427861116301,
 0.1297403223284308,0.6640419351446611,0.6333618834095022,0.9418686505329049,
 0.6933001089197923,0.3139624587309202,0.768279739199229,0.6560500872858828,
 0.17247358533398616,0.6401004478377618,0.3932798798511693,0.5265367029832866,
 0.5042686722819177,0.2703802497929251,0.5830776974854188,0.4627355707575854,
 0.5752544670510475,0.46018103202429483,0.3800667115417704,
 0.37913013938380224,0.42282054562272575,0.8297531663648419,
 0.7221381111328378,0.7715748117034783,0.36382411060721564,
 0.04750725020299296,0.8450136681340833,0.6674538707644115,0.8507053153858438,
 0.9080724936212378,0.4772435851800989,0.7615499429344169,0.2479589978782254,
 0.5826481635368743,0.21476339581777493,0.36972401150046574,
 0.13331733734721984,0.9611957212416763,0.41837288800491723,
 0.03834598647811094,0.38382263332407407,0.5098780039579986,
 0.3600419186017333,0.3129127618953442,0.9476333372839726,0.4174123538399028,
 0.1729951840411822,0.8890358128212563,0.3553875628049705,0.9355252298781657,
 0.7022113478967268,0.7274104484747186,0.04011217281706159,0.8127989348076856,
 0.5987671991346968,0.25909631160608615,0.2661782887870706,0.2790372663160017,
 0.9154369999704176,0.5610405147966195,0.5034426191830748,0.3702619986011756,
 0.4440184627676267,0.1908014220653812,0.2594735111947898,
 0.023408518890934094,0.049825906562180666,0.2551639532762301,
 0.7035095069624089,0.8278500393020793,0.26073188083881016,
 0.059423656386934365,0.7773427669567395,0.01146662378010721,
 0.5276039667943767,0.629779034789289,0.5660353155140724,0.62566756066033,
 0.673334146386919,0.8174615911499608,0.7152798719105177,0.8728076691219053,
 0.2590192184984128,0.9213616392307847,0.1910202857966511,0.3487065814762399,
 0.5747800804867358,0.5400257445997438,0.8516333905292223,
 0.042553146668173114,0.44461819172764316,0.27815478344940503,
 0.40346112450125693,0.6062512770552951,0.15922255521306528,0.642201113069557,
 0.7142005832019452,0.005054897012208626,0.6138630882327387,
 0.47447348856435057,0.24059141292800756,0.9352122399828251,
 0.9186680366362536,0.35644370044295715,0.041817804295023464,
 0.9945985238921708,0.8385374243796121,0.6886802689607743,0.1384725094508934,
 0.4269562369735933,0.019545624910675907,0.05377518386487057,
 0.49923034332540717,0.6842122556083975,0.9359223155771272,0.3040208242949639,
 0.5684815402921908,0.5709504306922181,0.22964425976411496,
 0.054789806416513764,0.05757909151445795,0.11911126633756441,
 0.14544414512142256,0.39012728653318696,0.1501484083394803,
 0.17648993226638243,0.48198536345063037,0.7112156742235948,0.932828360137833,
 0.43685154611221355,0.5151163045645804,0.0032672314220302257,
 0.9509075700941307,0.3029567157629047,0.7320009443947127,0.7722944846392263,
 0.7629146207891648,0.14306788625041345,0.1100003008808298,0.3194018794752532,
 0.5209906576751091,0.16777639340724515,0.7476942177346293,0.7212627254186452,
 0.7392829533269512,0.3712999188526338,0.0402420485870113,0.3734161432991985,
 0.030843431134409682,0.958386392793729,0.7540307638839708,0.6570651887015676,
 0.04074281549989789,0.07775603627038918,0.19568015855467846,
 0.14808331565510935,0.960852064056608,0.601894658439658,0.9328540074120694,
 0.3162256093226039,0.8389491617718557,0.49156961674125377,0.7707029128231075,
 0.8722953080570982,0.38079161596412303,0.5495170041037415,0.8125461096581755,
 0.652771660140224,0.9879029901037264,0.330427963034579,0.985121597153956,
 0.08007839178904763,0.9767538968384598,0.5293352601598418,
 0.48449718911637873,0.5191261386662468,0.16944662567641666,
 0.1968162696178467,0.5002244100626015,0.4153515467482053,0.2755733166469717,
 0.5747511740935254,0.5572189140197493,0.6282420220639818,0.6029080307739325,
 0.43070970532749386,0.8276111915946451,0.44071685734837185,
 0.18976954105797883,0.3805613383952635,0.2971982927272343,0.7926515465089733,
 0.005742913792003312,0.04519392236881126,0.19786251542963984,
 0.5611435039624657,0.5796011678223723,0.18940925749235116,0.3892546913148702,
 0.28691159402867883,0.04524273854998462,0.8219109695223525,
 0.5608300716251254,0.058522228159454404,0.24024893799215408,
 0.9293018248748366,0.9413029122634933,0.5119142518102,0.18682341483910014,
 0.9637391564329173,0.10800371725303526,0.318870809429455,0.4210931265840566,
 0.7077671687113222,0.44303658922550126,0.36698270205673134,
 0.4023879939375308,0.6681949334133553,0.27993301395515435,
 0.10672753047568895,0.7735389233618717,0.555365163368031,0.7075133274929576,
 0.19754781434121926,0.11481906128299435,0.887055655756822,
 0.22057691713315464,0.10828970467024324,0.33490452109099234,
 0.08353835661525488,0.41595445823094945,0.10417545100606529,
 0.9770016757059024,0.2678985087893654,0.2906392588606609,0.31900674972093535,
 0.8997422165617898,0.6911693761826,0.7339097573375323,0.24934425003756733,
 0.34308373733337993,0.7100547210640225,0.5313000157374531,0.9686505095261895,
 0.2868850337045925,0.12281113396990084,0.418684440841364,0.9448155354836935,
 0.4124510098877895,0.29914375807136195,0.6235086435562637,0.2529558949751878,
 0.6631391432514512,0.9443396531019161,0.9625245873956148,0.3859424552896502,
 0.6552596093024635,0.2334319860386962,0.9427063248250124,0.34939319371021393,
 0.471975054061199,0.1408019591842472,0.6608395954992898,0.5351000679036201,
 0.9664218992819811,0.5932409869745421,0.615004855937803,0.057799407473367204,
 0.2978612320075851,0.25340457667584304,0.9901270723985455,
 0.06995799913624401,0.655589898680639,0.34822554922734594,0.9873432982163549,
 0.07630490030809811,0.08219750775741508,0.5773161921770611,0.523296588383946,
 0.9363577257348887,0.32472224916413484,0.4494462623087647,0.4680206862192541,
 0.13884963641001413,0.793224442141262,0.593556494989698,0.13884346875298226,
 0.018132965604960916,0.13295586475304988,0.31147869190831834,
 0.8633113072581909,0.06678161747778555,0.49858424560473535,
 0.5776359376772561,0.10108213526968057,0.6528478138676065,0.5921287336192222,
 0.271418162590893,0.7867910724299041,0.3718177860078644,0.5036321972398095,
 0.7283952461631155,0.2257135704615486,0.7764364168083737,0.5839477547474351,
 0.692869621776366,0.7022624942790899,0.09584131161050236,0.3249031804927688,
 0.6302589320387182,0.6403775315942161,0.18187365668217925,0.1680315624469385,
 0.9039698549192581,0.2498849365267778,0.32788564599532166,0.9720795471446004,
 0.12679170691248998,0.8648668461220741,0.5481327977560864,
 0.026348071259672756,0.6143727260000752,0.7669480104433384,0.839310141367663,
 0.5436941187861524,0.01828175045082403,0.5001278392101913,
 0.23775558973253408,0.24571270381596944,0.49826242291752787,
 0.8911548351092031,0.3873084027876421,0.12965230391871918,0.9456403147847429,
 0.35848714459380426,0.7859546738882385,0.6282843660225316,0.8688380524717004,
 0.6499457070736361,0.9880775134639086,0.6605997790127075,0.18341396717838954,
 0.7806390417893645,0.8294502211336874,0.8616726557933689,0.8422491504351772,
 0.5643987421739936,0.14277052762861497,0.5522575954902317,0.8466597811158794,
 0.07908881467900908,0.1256730125430512,0.43162015450156144,
 0.5187945481840734,0.917682233197542,0.737040190923714,0.27123857054832257,
 0.1527590976332064,0.9951047004824857,0.20266367232599192,
 0.34635899206951337,0.6331270258206749,0.20330008667125876,
 0.2290822001721472,0.6019472843952871,0.4080443067923323,0.5185302708713384,
 0.9622060506072753,0.837086916946492,0.633088542092345,0.2369111925945313,
 0.36325662384693624,0.026918111530148336,0.7498363883105131,
 0.3911096246794553,0.7686119000724491,0.3752913427286906,0.4359069669021881,
 0.9183811418837186,0.28552844110557496,0.31236769231805006,
 0.2566744835167071,0.8701934488784613,0.5103976844044614,0.6985772302702753,
 0.9497087599055978,0.9822695590998264,0.9901504405636931,0.5918918671365015,
 0.6327827430718512,0.3101757301455137,0.7570584838270031,0.16261995027134168,
 0.682045628567565,0.9001543113805102,0.19547844066923203,0.00179915608067871,
 0.8079117158255806,0.7346751960473255,0.2172277080981374,0.850608163373067,
 0.6515195801816956,0.2945832975040804,0.7274687866843321,0.45622627039678654,
 0.2980640479316434,0.040048654860084354,0.19420638910088672,
 0.07957964258520833,0.996173167452729,0.6644810549718946,0.13038844538681305,
 0.043220714118763826,0.4498004655395096,0.7427566475797441,
 0.36544944887013686,0.7849563021903141,0.7548648985897929,0.7586944676106793,
 0.017039083742137362,0.49586179155972787,0.40863947745847384,
 0.23474430250539113,0.9942060180125051,0.2205010826224122,
 0.01979908812149922,0.8291161810809463,0.13588564222523303,
 0.045103959516926806,0.4092619125509659,0.5304751935580793,
 0.9634740146199631,0.6713117062062347,0.049864800842387536,
 0.9780418017185537,0.6450060902743131,0.17719494397645696,0.856310640234293,
 0.716893231487201,0.26947779136986183,0.5338328160306055,0.21007582870739072,
 0.46077980324067946,0.012990757136811415,0.04865076980857652,
 0.39725186585523764,0.8088910667765621,0.376541856965803,0.5195189743728124,
 0.5778182698619388,0.3130439415689913,0.09867546053378962,
 0.10655445719750456,0.8812807251672612,0.42748345626871387,
 0.17594444451496427,0.5247457628082579,0.035267916510891895,
 0.8750509863925011,0.6547018727838244,0.27185600832357404,
 0.18927799528984424,0.26656004848308523,0.01591468743698199,
 0.22274318027380102,0.7074943670152573,0.2407945087112452,0.6219539302152114,
 0.8897763355405036,0.11939498894612022,0.49226517270405457,
 0.11237069197982907,0.7514660716095775,0.07927114365481125,0.275895903044616,
 0.1518060260414522,0.7740690414071476,0.7239937299609189,0.09384344590269544,
 0.9682843792266841,0.5816495085651624,0.7389043348645375,0.15039366772374374,
 0.6079754998395203,0.663334163616919,0.5114426100428571,0.8363105427727874,
 0.1536120242074024,0.18517917984190446,0.9944708718857788,0.667138069627422,
 0.9532326079066358,0.22578290081989527,0.11429912793231534,
 0.7543230001775587,0.4979685726855535,0.3841373196111162,0.8293489724276519,
 0.03942814957964624,0.24106952901184187,0.1327533793306378,
 0.17693039709823144,0.38663571866149826,0.9243755310697548,
 0.37874571705697324,0.6422656603881153,0.8433774650821249,0.5632995165900319,
 0.9193888523611553,0.4566424940671011,0.27481187374581106,0.6396646694330516,
 0.7346498379783878,0.9237269140379529,0.6795851347895365,0.16021737854513884,
 0.5600996760914689,0.3579222207780419,0.31960314291153913,
 0.41082734058730175,0.9398835615202754,0.1434754788716186,
 0.05641194499428037,0.6889269518816644,0.8273168259782541,
 0.008903070444582717,0.902772766420801,0.2143641883077766,
 0.20232140394066545,0.43153235266064216,0.5699580292455867,0.265775530318876,
 0.23711662228138597,0.9534237597467534,0.9229959556912234,0.6448601026710342,
 0.7629122913756656,0.1321796441426919,0.4218582417418877,0.3275975411402223,
 0.972308396682966,0.44497427092258557,0.694359805572938,0.607185815802169,
 0.3038151310426409,0.06088910224695743,0.7846425664715673,
 0.19259516598128223,0.33613269155560244,0.7088643284020247,
 0.43596867294909825,0.6214391060462817,0.7639349930850138,0.5081641322275371,
 0.6453482190223587,0.9162152153129641,0.3550730067561354,0.5560610360266913,
 0.27916557656068197,0.14609679409420417,0.4431459802935164,
 0.7106794183407432,0.48652336189433043,0.7669755411194409,0.6198933815593552,
 0.6799738129303425,0.8548105888766497,0.5872917872976359,0.25814822562971707,
 0.8513690204553868,0.9180407522312208,0.8151646362869521,
 0.060755536089742046,0.17936605638183079,0.5929540061878081,
 0.10807390961333718,0.8228153789530801,0.8120574207227456,0.2304448527784142,
 0.7962750080931201,0.18450901269267472,0.2866936402907905,
 0.10738550487965182,0.3734645304588824,0.4178397377877313,0.7367150550008583,
 0.747015484141062,0.2847890634670005,0.5934052469171658,0.06022568715513987,
 0.9975450103226288,0.1722164967755263,0.4926882152522941,0.8072785012694167,
 0.2743422693873404,0.9150824911143681,0.2310737533930055,0.5675893207705655,
 0.4294086628521194,0.7357867213314152,0.9469002441253976,0.5061436547136,
 0.14898983701197022,0.9774242204125342,0.7605996341092668,
 0.24902905383358487,0.5579625301786596,0.17980298921133986,
 0.9610138561467927,0.4110139085153074,0.10629471148966696,
 0.13124864134480052,0.07152093369403933,0.9128181046094226,
 0.4644443762490392,0.8425254836044042,0.6594045489748837,0.45691707135175685,
 0.4570140317927499,0.14198536082388769,0.48394567196425964,
 0.27620994332918425,0.5960254128769703,0.9768996996103401,
 0.21135239486089452,0.3584391998735591,0.130364188870827,0.8515555749213729,
 0.9482529884968975,0.9664132603072847,0.8987201568811105,0.0968669055773963,
 0.8911145912936811,0.48502178009295815,0.29950412378372326,
 0.7325827428068155,0.8007440039402272,0.5480039940128617,0.8390102939760956,
 0.8977077612776749,0.6797342763635158,0.6496534525655181,0.07980974924169626,
 0.14844953692444618,0.8035082177522394,0.09065484979915228,
 0.8563697625154929,0.46966308637271736,0.2008707524647021,0.8312199069975337,
 0.16425459159193256,0.2951067935579015,0.18524545067290465,
 0.10286260052137908,0.30344794610260395,0.3788728478972796,
 0.36617868140369914,0.836317423479669,0.20802679882481623,
 0.01775237668524776,0.491515882077992,0.7288921177559714,0.685856288474189,
 0.7541075892988958,0.4516257783004225,0.9523923373187132,0.841868267815534,
 0.7076961976547761,0.2511903070365874,0.7804801720213562,0.7527992381371202,
 0.18227639342379898,0.4873005737695929,0.6852672062302305,0.9111953168036735,
 0.4413552328286122,0.14378686625699522,0.17159223355307052,
 0.3447202629300765,0.8865544043168961,0.37615035216159853,0.19051692260843,
 0.2466620008042434,0.15936607751916798,0.8540945186481709,
 0.38423141849922327,0.7409875703343622,0.09914290674293758,
 0.9031468906545003,0.018029659946631194,0.40445835895638105,
 0.679061471456464,0.17088111220655444,0.40800835266997026,0.5909101932594558,
 0.28042768459089396,0.7003346882803618,0.9280549843773964,0.7466381389703381,
 0.8876216042344759,0.5784762892781258,0.5090098523962153,0.9441584168344871,
 0.43600198108296384,0.7722747562777305,0.12939719252246273,
 0.4317048646518792,0.7430105238682145,0.30573432768106146,
 0.30127621820936845,0.41460615655481536,0.6750330465036006,
 0.6457467281016951,0.8046853396396602,0.5524957910126256,0.7937622952279806,
 0.6246598103337476,0.5323679855660727,0.43400141103494005,0.8413695634995263,
 0.178797542893112,0.924144697837441,0.8390245604892366,0.6718525757145224,
 0.03693830924259767,0.7041690771861483,0.9783398569792447,0.9369627601569348,
 0.3523848944700553,0.5139461117600093,0.465827351647048,0.25741683284948946,
 0.6082206045868199,0.7838905261169853,0.6221178480660347,0.3202665118281771,
 0.09351786384746053,0.3120431716468577,0.7292385347270296,
 0.12285019151187349,0.23726672499279544,0.40677972823098574,
 0.07228337470210466,0.9528326919600723,0.48415611482537435,
 0.6026288235230692,0.6203043655030267,0.782907627032513,0.13788823473578649,
 0.09995761729200081,0.07037480069708812,0.4408301251803307,
 0.9032905060704509,0.3975631197442496,0.6983394610145847,0.8611236998782138,
 0.35605366159119944,0.20055697380830395,0.7925225706898318,
 0.9930945967382527,0.320407134469896,0.1469946140921139,0.6570626894278241,
 0.5120665163669978,0.38666037604995573,0.5546391096542287,0.7654438142198999,
 0.39824480225138703,0.840743903406364,0.32348437390836937,
 0.46588528536808826,0.13843609968896542,0.6944625678537083,
 0.2612937258036211,0.008698607086399357,0.009828300789965105,
 0.9402441729227763,0.9080482259280914,0.7771100358426087,0.33846711015555275,
 0.9862162666418238,0.5324951688959269,0.2172517845596995,0.8925512189819615,
 0.9128304688699429,0.10163832344674117,0.854757128387725,0.7355237797618402,
 0.7634918562083232,0.7095287800346178,0.6114902666993526,0.8443489868038094,
 0.9494700144251091,0.2611893983763087,0.23140139589337527,
 0.038420096337464904,0.6773752501162605,0.6591846769250352,
 0.7182109569418864,0.017434738301826402,0.9091983160079872,
 0.22040358031695972,0.5914609444936612,0.9446807643833652,0.9424627717852433,
 0.30493029370326186,0.3643421726127617,0.3500638865449397,
 0.09883356321967096,0.7157389881656238,0.3876445772404219,0.1137964374231566,
 0.7337387434510595,0.055369230077874976,0.6006169680598313,
 0.020605327549979013,0.30953364125105554,0.0028376103936865738,
 0.6214036591743654,0.9834974420517223,0.6706439312435128,0.3337333160024015,
 0.14729683277915107,0.028439495861196873,0.3231104607749101,
 0.19514788832276553,0.6146469226492285,0.26035034915377275,
 0.07253058689166103,0.8425172282054941,0.7221703493744188,0.3987707024551772,
 0.22100449804833144,0.6234787440109656,0.21594318019782688,
 0.04813747285125758,0.8136140950041353,0.6010528456894484,0.4101692271518973,
 0.2731554618440133,0.21453930356997086,0.01975772649931251,
 0.24022047581577566,0.11807112322719826,0.7101988723984098,
 0.3225718184503942,0.1284610339923804,0.14021227607783526,0.4641827208835635,
 0.35852286372517495,0.18828322502680728,0.2183836115323481,
 0.5881851755091674,0.19826913406196278,0.8264842922424485,0.6681225006343733,
 0.5933552702578081,0.5016401079221149,0.7663807697878025,0.7157850698211123,
 0.6670547287302484,0.12625541334180046,0.3917606828441761,0.9827213423187212,
 0.8829802463522233,0.6819207644958113,0.9986936977045838,0.38670081728881656,
 0.9488834822397868,0.06801360814836821,0.812404162617671,0.7853194574880692,
 0.5374449758230233,0.777342520182751,0.2307227828501981,0.956540066993154,
 0.9884973701406637,0.12185104752899267,0.5531877916775518,0.9354063656837553,
 0.19273319408252254,0.40880168783389825,0.9183552188115002,
 0.9096236563803419,0.9006688183115003,0.7831769643855299,0.18489998831956034,
 0.7157747561273837,0.03934435762163324,0.6195365683786611,0.9872246459693004,
 0.20979987836559855,0.5253459879676219,0.5768967183984728,0.3982399012717628,
 0.04379659180789153,0.7850433359791786,0.8272188691780524,0.3792539855507301,
 0.2813383796356276,0.8004729158203191,0.32433273440686916,0.4157989855102935,
 0.5970137772888654,0.42309551712051563,0.5511695795311787,0.4932507460952732,
 0.9359223481259065,0.9595853279559083,0.8400513728381273,0.7721832205847381,
 0.47736643495543074,0.12918093353672966,0.1805776214366469,
 0.34606725214129275,0.8455675769360282,0.5620970690193663,
 0.29960861766811664,0.22576079259347626,0.6906828876225467,
 0.11954967731359245,0.41105016009970974,0.8424617505679498,
 0.26100501674424414,0.7829855588808948,0.2134826978384169,
 0.02529401058683134,0.10803250615053095,0.277553420353582,0.9656845962217747,
 0.11343980532802067,0.30054753345997387,0.8715922220024187,
 0.9395805911277563,0.9078533316643375,0.8710784746173851,0.2647607981371092,
 0.8876010615539387,0.3239224273444079,0.8602601026564575,0.41871814861140666,
 0.6193062653423107,0.6852992778636279,0.019182001877863564,0.695029254413358,
 0.3540093776090497,0.6528165958649174,0.5156992777482756,0.649927037682575,
 0.49619554368172514,0.9406982005028048,0.7136481683478644,0.669617318532444,
 0.4219057561418613,0.13042327901920525,0.040684120672006,0.5447060732633487,
 0.33678652896334726,0.3608453709629552,0.6369923110443346,
 0.09849054881406483,0.6561336686471121,0.07857569232642536,
 0.3022155406190842,0.7386364052653145,0.1531304379833891,0.3314469610048547,
 0.012534109292707152,0.9042636088561018,0.08208301841943433,
 0.6932453269650608,0.8290836829152506,0.34473433712812707,
 0.26185913404283634,0.7454491762769481,0.5231902207906487,0.913718106195961,
 0.8150792103982016,0.5081115139293664,0.04653198676271375,
 0.25195832857503264,0.6024014253871092,0.1325398091911294,0.590426608517429,
 0.6733147309859215,0.19597290224494934,0.41535861365881943,
 0.4209500984663409,0.49248801266888564,0.28668756233212145,
 0.6146747359794317,0.3090467084828652,0.9133218931188041,0.08685472116028703,
 0.44637434533906917,0.3578852958630434,0.7376503443837266,
 0.060891927309113636,0.09893964495583119,0.7512914299663469,
 0.36593227447188925,0.7254970975636675,0.07609857182606583,
 0.08166876553338209,0.88460158062452,0.01225377168191033,
 8.577171031820541e-4,0.40949782413197544,0.762825289526765,
 0.6258720391656755,0.6656789763878274,0.39544095776100563,
 0.08188796767186302,0.1651923584159518,0.18530927952860599,
 0.8097725122151769,0.13408938967535589,0.29397550774620673,
 0.31768399461753716,0.1862696952087579,0.25731925031888514,
 0.7369218922181279,0.6169737527110887,0.5317021686412182,0.7987219471060475,
 0.4922044288586621,0.23695841740069823,0.6173922817500996,
 0.24838999938988715,0.25006680531134773,0.969897623318466,0.3406160374804732,
 0.16672345498194185,0.31080712412778044,0.0796778527076013,
 0.11250328731851633,0.6943292421937566,0.7572388460955197,
 0.12737857948685505,0.3988021499404263,0.20100534891714517,0.982156164413757,
 0.9748917299542365,0.5253476748111645,0.4411777761267335,0.6517420374223679,
 0.6827160637297643,0.5845499955286876,0.6481888405696397,0.44354322799132984,
 0.43665681442489435,0.7574983673156997,0.6542931863102583,0.8539224995068326,
 0.3768340121037784,0.7347804565107444,0.4629259470802166,0.4646893358939804,
 0.4118948215974778,0.130298071390056,0.7510658685917327,0.9829284494514057,
 0.5506994695347576,0.11951163982993052,0.6607367492565341,
 0.13696568251243701,0.4236334159290118,0.9103894976009161,0.7613022280404402,
 0.5438059933320598,0.4465990105243163,0.13543103495699083,
 0.34700859621264424,0.9146146756926437,0.0681608060311133,
 0.36468338720100846,0.05168361936715116,0.9249688477913337,
 0.9231909629077908,0.7717507069256428,0.11649699550163561,0.5269617011669916,
 0.798348052605009,0.10607439780368899,0.9629078130917635,0.7845823519862984,
 0.5980941603510768,0.8166636122666394,0.2520570350283309,0.3050589040998607,
 0.9243057102941217,0.9183848430546779,0.5989161393733196,0.809573905554956,
 0.94489408437097,0.7869859742319032,0.6320348133592011,0.7534506541086115,
 0.37655528873268496,0.8742201605312485,0.2848184643834485,0.7396275256159723,
 0.887114197077206,0.3562993711110535,0.31236325592922276,0.6194073982401332,
 0.7177484318146586,0.2833886332511898,0.5121229242097911,0.2490669291961236,
 0.7442449196165928,0.8577007995219621,0.959493204745692,0.8011288016101443,
 0.3592795009591223,0.2809427383391587,0.6429084855356968,0.40058065036580415,
 0.7741941968294381,0.4036556093349416,0.15193545461033797,0.8549020126208221,
 0.7406711514134682,0.667169082306359,0.44843188866644823,0.3998684304068564,
 0.7127245536077643,0.18866362757783417,0.5362168716835388,
 0.08210963698121743,0.10330673197436957,0.6763839794618712,
 0.1903844355446791,0.5363419674053712,0.150980593885795,0.5503279598220504,
 0.5195588142152688,0.7131056542617895,0.9824290687050695,0.19406082256420287,
 0.5902939250183347,0.06774993033002774,0.7218173383299061,0.820089130644062,
 0.6599065078666841,0.5140229151022115,0.3302043769495866,
 0.021414302422978815,0.37373109691149864,0.753742693980298,
 0.3524904035216009,0.9977247513618805,0.2849686096882955,0.19203610116513747,
 0.7665255437519436,0.7684072830366538,0.5612571741565143,0.15665863112126854,
 0.4276041196874757,0.8221195797983811,0.5782478529121894,0.961113637200716,
 0.7674611172954325,0.31819234383786377,0.18005403501033013,
 0.22889723728046463,0.488893498473876,0.08973599379856934,0.604328207843801,
 0.5199954001736681,0.13654909314793795,0.9908858029026409,0.183185030981054,
 0.4214566225527502,0.23838031103590607,0.44896303475695054,
 0.33848095234810216,0.43866403678520227,0.22195214465386792,
 0.5329280327583137,0.942216843343976,0.9559081938708329,0.7414215824968899,
 0.07151578130563363,9.694497584616935e-4,0.8464516294495088,
 0.922530578622108,0.9025794090877327,0.7628101754664491,0.05425577091399647,
 0.14109512327507412,0.8610200270651219,0.8856835777130415,0.7681878449762868,
 0.24375718176803018,0.10463435592149539,0.051743297248746045,
 0.2020490050989916,0.7478137402965314,0.8613989620292087,0.03136302774721289,
 0.5182844646703549,0.4032357859559248,0.4906673662753085,0.2878853035914586,
 0.849421979214017,0.7943174435758071,0.8382015275280941,0.1842535471303911,
 0.9309018996751969,0.5048927479556375,0.30702187843566864,0.71593330337551,
 0.1795269473485448,0.3922139501919941,0.10737540220003006,
 0.41529671135658086,0.9262728128320151,0.4764897540833276,0.2406224882628446,
 0.8964663130621556,0.5517634923467708,0.07300033050279398,
 0.25978142053539377,0.07938922873216558,0.9223999005861101,
 0.5658568924125312,0.4838509846680351,0.20972476813322471,
 0.08143866079603479,0.9430622118032128,0.7164546822768307,
 0.12671879174500555,0.9409003636739994,0.5628387908805557,0.6022877513795323,
 0.3989601218009061,0.02971682865610681,0.4339820360847185,0.9573580048854445,
 0.05895298975030561,0.6541842876427153,0.3469670436359339,0.4464386425802722,
 0.23533791398731863,0.879814617898927,0.7881333084471728,0.8554469588977777,
 0.6470768402961533,0.5324394321767045,0.8974030783814276,0.9944365498945744,
 0.9142001171560651,0.7624733914167079,0.27558952697810035,0.7352793386206764,
 0.38045200585440364,0.17232299646725358,0.4249884714269292,
 0.6142966627741966,0.42224545353495946,0.8802242177354498,0.268530915361327,
 0.36803902864230387,0.23774891808318688,0.760418582467139,0.8033736040675354,
 0.8899771558216119,0.9217388813969821,0.6378463074447016,0.8660754485961714,
 0.7563274798984656,0.807170773953213,0.24786140745964336,0.16842851551897642,
 0.0023825054327049067,0.9893635868553959,0.9043199128443966,
 0.8002366328925494,0.8172119726132272,0.24992045594507173,
 0.34838397007877997,0.8862200055367375,0.4842046679974399,0.8097453616700265,
 0.7961267121369767,0.14416822326767986,0.7329878605029373,
 0.10452997031178013,0.651516732245387,0.19658600111471536,0.7270066776146337,
 0.49386547910245704,0.0865831771609249,0.6191029403238145,
 0.06993569861371274,0.8429034162079259,0.10889555874242163,
 0.6205149864918034,0.936843275285495,0.034214407011471604,0.7538579612833991,
 0.8337155778459532,0.6082837509522686,0.5226192094779131,0.2508913404327827,
 0.1185649578605037,0.9162110012100029,0.3544574554283919,0.9011457956286582,
 0.21684213607059877,0.35078678428489707,0.7454614769276771,
 0.1124753585309004,0.8876202317540489,0.46211011273960834,0.9957620574203386,
 0.625672561817672,0.051348660917982514,0.6461778460098497,0.5362116733905009,
 0.28027517133016333,0.9197350908246534,0.13466592417248358,
 0.8166804158729456,0.044307137004690134,0.24875725029862705,
 0.23264751173027198,0.2628832749427107,0.5268299287662556,0.6416115242922059,
 0.7063462709894909,0.4482103955536316,0.3898260590474575,0.1388827828260918,
 0.5195724690918528,0.989443284692944,0.08830401135003507,0.8195890466839928,
 0.9187014997525491,0.4691368413539767,0.0021296685812779437,
 0.8840381674797874,0.8548671670079042,0.12165665572763273,0.3247954613392615,
 0.6979206203476644,0.47153124003001634,0.9909433450735677,0.7871990352117291,
 0.28578584759852255,0.968888101410095,0.35079527932326726,0.8246047891591685,
 0.4729761809607611,0.4640402373583773,0.2699337715628395,0.9666503025632072,
 0.4102439957405293,0.4542855741667031,0.07007867317592731,
 0.49120040046697144,0.31098765952836294,0.7250678435242948,
 0.24857060851722002,0.5624668346715781,0.29852238503419826,
 0.9532704905527664,0.26930030330536114,0.7487405466159154,0.5374181199359386,
 0.42853991480763065,0.3511001222840622,0.8361078462891527,0.5692368212635212,
 0.5227346633526175,0.9542688338614846,0.4774947617445662,0.12794066755060052,
 0.33464072825161073,0.4462004194235756,0.3841606686829042,0.6314606458688184,
 0.9495755598679395,0.9798520781751219,0.13941472725272974,0.2979058132985172,
 0.6854098452887671,0.10986377883919407,0.6376307468923567,0.9757114674771014,
 0.6066459909073022,0.5098882235146169,0.3812050842810064,0.1796459066114624,
 0.661115768848084,0.6438213566367268,0.44919348305150153,0.22248070246591434,
 0.035705141795168704,0.7946433441202049,0.40403108982988534,
 0.8211573895822804,0.34429479205982405,0.8243698487321874,0.8771598744663206,
 0.18878565620272336,0.2705961927337148,0.3967743576337429,0.5339731539549225,
 0.9478177255473222,0.8108179471961946,0.4142841885102637,0.6130464708308669,
 0.26107492026824763,0.41425380793141453,0.24385964546810224,
 0.6070258709893604,0.187465182614293,0.9525807876403712,0.6492701681510729,
 0.803011652354742,0.3634324194366154,0.8507140165117695,0.4207417400250437,
 0.4837723652644935,0.15021299866630233,0.5661255191752723,0.96038439387669,
 0.6344082770100237,0.8977906688784998,0.024244572238148465,0.995427987101756,
 0.19568572953605867,0.6963860152106875,0.8363256904070813,0.9662083904943011,
 0.2793422443300715,0.49416655446680147,0.620153109964393,0.8754470528549851,
 0.7404667681676269,0.5866355607135738,0.31058770392963375,
 0.35139833767448647,0.6156339452110844,0.12891701900775154,
 0.9370265740302676,0.4738592920007334,0.07367057750094841,0.7647922814579515,
 0.6992635241153304,0.7254423070820085,0.943543175431407,0.3475053392879952,
 0.15707740507269252,0.6109806005086782,0.31991262301076073,
 0.6831539475926659,0.6185547302417744,0.593343996712818,0.3168476333405501,
 0.24366699195782715,0.37341415563487723,0.8575725572920279,
 0.02975066348902622,0.10041485020045227,0.5644295157178606,
 0.05822621204627698,0.6348592279609927,0.8620639099687428,0.3017089158575207,
 0.8233841632448893,0.5598861684988987,0.6217778679611115,0.5966768326073197,
 0.09349036564388646,0.9523624973630846,0.9787106336459552,0.5745138134072388,
 0.45312323685397393,0.5815924391656154,0.2566537466094637,
 0.39962847126424006,0.34824832496892877,0.5760058434436685,
 0.21010649564683526,2.2662883861901406e-4,0.690017027700273,
 0.08289812925635598,0.7864615211205154,0.4760234453619723,0.546218187840612,
 0.6713731534126595,0.7878756526325986,0.9411985589742345,0.6900250309952659,
 0.8495108922667466,0.5802900392686083,0.6125096435028027,0.3162254826339116,
 0.7085041072770706,0.13153638917899713,0.4066829597008791,0.5526552215078914,
 0.07142303834064534,0.0386035378110583,0.9224762647963409,
 0.025911922552580013,0.9606157596517079,0.5785650843285529,
 0.28621006378037706,0.24970958711897628,0.39650305505795114,
 0.23325956047933072,0.8582083783530454,0.9871668323529965,0.7793275342910349,
 0.02358930504808754,0.8540290691305847,0.6203904765973791,0.5890750255188189,
 0.5199872350053085,0.1686735998998331,0.7200270893920577,0.33901006214141194,
 0.6109832964187597,0.10295272952644097,0.5907179435251555,0.9273645537781053,
 0.5275575929732349,0.5792003072132422,0.4176750066226991,0.583169454463332,
 0.0758412395182071,0.023744157385406583,0.08037533884702286,
 0.602815496627901,0.5444385414902433,0.6302146954746621,0.24034971299356067,
 0.8611600064178387,0.932876211263574,0.8848624819205767,0.04733083642750602,
 0.40052664612434563,0.2559189865657976,0.9399290862339645,0.5634138946102131,
 0.8183818556249348,0.562426879310712,0.09003237834838174,0.37676003991361706,
 0.5776104984490106,0.732145139424425,0.3436802164934941,0.5329613753635911,
 0.4780024836652217,0.5811122471911323,0.33300919272541374,
 0.35689552979300665,0.966418806282135,0.8207182106728752,0.17168441108823462,
 0.5898489699451562,0.11325133739780013,0.18503728090624394,
 0.8319879133673117,0.24967846953939432,0.7982202332925724,0.1441880497949769,
 0.06736824311349887,0.7984058900212403,0.10538216753374796,
 0.1719612456752153,0.9930169150875453,0.7498169758447881,0.5645069171863627,
 0.051323217142878685,0.5766142866277495,0.06438202586867336,
 0.8819263815071304,0.43296174256597075,0.6665265505849326,
 0.053767798268735345,0.1240302049363815,0.8922223206832421,
 0.37251980399264606,0.41717272484257684,0.5418101922497867,
 0.3195452448418026,0.8945741050134096,0.1335081083272155,0.9686992365339756,
 0.9479030107486149,0.09562116139879306,0.4298833917773843,
 0.38728249008835336,0.6672737339351738,0.2859929936756995,0.6680876051033574,
 0.39721388383339706,0.44213149201604063,0.18557465746714674,
 0.5211612389578857,0.11053787826762984,0.8896049363491497,0.9255657032381265,
 0.9302363577639934,0.6994326731613472,0.7726193768423331,0.2786648996316481,
 0.5623335676937075,0.012661384322470992,0.06678256910214342,
 0.9754501655642348,0.3230415523125093,0.5594536155743136,0.596219421237743,
 0.9792904381525841,0.9761651299229903,0.31730769335537345,
 0.02643425020071799,0.5878294269343276,0.599790426542654,0.09562390656442821,
 0.33348190041544423,0.5468952220332874,0.6757944829253798,0.8365671654698436,
 0.34383675499066935,0.177358845027779,0.15903568767814946,0.1428806252653455,
 0.0305775744980884,0.5319046869564275,0.23843362167545612,0.290250669601974,
 0.3470109347190582,0.7341738164016112,0.6854642316771106,0.6178366876527778,
 0.13369396885428686,0.3474090730086168,0.972117062300114,0.856032834407629,
 0.4661878335262325,0.15554250073035203,0.04427264812966403,0.639652344448915,
 0.928472637715795,0.6900421045415028,0.09184895678769689,0.9585110256748948,
 0.4870153528439417,0.512397973538363,0.0661050900091864,0.4613444289599874,
 0.8038082974699726,0.2603458929302107,0.7800370238282022,0.38658273302855195,
 0.91943406696829,0.4654930802504583,0.4752411413467065,0.5932094129909871,
 0.5997105900406852,0.8737518188180021,0.8399558511227405,0.8057931433383619,
 0.9827342658620448,0.730325846105389,0.5474635674167285,0.006116808583551792,
 0.9021916572537328,0.8335248342360683,0.3633048900823216,0.5344853180719543,
 0.2391232498605821,0.034656917058351056,0.08815913821144161,
 0.3460332668165058,0.17765618808236083,0.26627787973267714,
 0.10037305090204507,0.9567327894490292,0.22971480297811941,
 0.9624807810687261,0.4514865354337346,0.41988149204594005,0.6777253878117467,
 0.21231460035725302,0.2838453140848638,0.030931052092080025,
 0.023443213102716576,0.2829429754406402,0.26443530460560694,
 0.5787293591109167,0.5728163603750104,0.22266281753540973,0.926664449213539,
 0.5486644616672524,0.9097616351071626,0.008086302301892112,
 0.6181771125722156,0.8531876284362935,0.39072090474133847,0.6514821360639742,
 0.6224627366332801,0.4249618309731015,0.28398504540807534,
 0.35500218388388016,0.8302065656333781,0.8526405845253322,0.5902512905604093,
 0.4666474020279935,0.4474973581653834,0.6153348262519407,0.31048359291755123,
 0.9769624652587297,0.5833347373513346,0.4281055346298298,0.2364245458806441,
 0.7900658839399494,0.5470247851165373,0.5124757187641485,0.49709786075849716,
 0.7666301386196295,0.5669146843566256,0.8672028417555828,0.12255363313836698,
 0.4257834864678127,0.8356260421152794,0.9492114077759923,0.39805192475107065,
 0.5007464534512996,0.009016223731942108,0.21651971128984882,
 0.8912149076788636,0.12466522371606437,0.16646951976835878,
 0.01112481731816195,0.5898905457407314,0.9776458498286796,0.4061274492905961,
 0.8479878682284068,0.2967457712831084,0.9891150786879066,0.9575638814879355,
 0.5393403093025461,0.20074146249065783,0.25710993317328246,
 0.48088778596030324,0.7253695809110039,0.5488196839843216,
 0.34693694074186776,0.48067223994809904,0.5345582236475792,
 0.39182354120056373,0.5913978680506616,0.5235737009508648,0.7871831120445258,
 0.37057959329603485,0.4722310342783844,0.7652048775774539,0.4630361975903369,
 0.5464944581063508,0.1967334557467766,0.5016224509197029,0.21005481609111176,
 0.745272720339915,0.13749876479395284,0.4097638821877364,0.15292833077809165,
 0.938474843087633,0.782236290643959,0.38978066714118886,0.5231682483338279,
 0.831980089504517,0.3330419874070123,0.12629599908407863,0.8329195678029797,
 0.6818361365455513,0.27836190786446346,0.6951734761827113,0.8102401718172431,
 0.33129581843069533,0.15194738283349485,0.7465460641540962,
 0.6070116998995585,0.39906991827438265,0.42396441060576884,
 0.10981891854857306,0.009849627984374942,0.38658640449538895,
 0.2945620718434008,0.3164648689729984,0.13281941341863623,0.8642698541659103,
 0.5127277740469383,0.2461703845089206,0.4371655171242744,0.6279424717934108,
 0.4215769279492234,0.8584396762078559,0.1932851948650447,0.4660728145889267,
 0.04665352854316629,0.6586490613173025,0.5014038949576682,
 0.12120209774382118,0.7709159174824146,0.020361162021087953,
 0.22716831878189958,0.1472354228270002,0.6751583729703567,0.6107500972778036,
 0.8978682793032748,0.931777591753012,0.9235619190512496,0.16520916823900744,
 0.1539109999367757,0.9141599506456136,0.12904434070143722,0.1446718600563749,
 0.8308866726841261,0.4019420070276527,0.3088319290264092,0.7181376766840731,
 0.5638623717882242,0.6489528424577543,0.2760481509251831,0.08275703284502711,
 0.41643819218743006,0.5619933679122616,0.05388376023968178,
 0.19253483388233683,0.8209248067250237,0.57174517250971,0.33100093791140106,
 0.5023654413879827,0.9338602660454802,0.36616257296504584,0.8624044684349285,
 0.6313609915733635,0.7593915650459976,0.04803282455499236,0.5735467246018808,
 0.8446542571495425,0.33148736115395416,0.5090001072133336,
 0.06060592902616235,0.7592198140100482,0.4933556353482471,
 0.49658486369899046,0.6304632975345063,0.38264752703709537,
 0.02545989174481822,0.871698701168835,0.3474423250426719,0.6660865705791981,
 0.4252449707199667,0.07105068466339948,0.29304638841076414,
 0.9008549437440342,0.53404951065401,0.9870859329258077,0.37112435646526687,
 0.47679809958828057,0.6744730353980183,0.09739644741488018,
 0.6207564964352398,0.15108027884619601,0.023348812814574238,
 0.8101375178337537,0.8264171663892613,0.3714429008996025,0.160484066675602,
 0.7421037895291173,0.03674273716664356,0.4502824281601874,0.2610257023721786,
 0.6041980048410527,0.5516724760047119,0.7476488163196044,0.6920592243683136,
 0.3600950874696548,0.3502216522248889,0.05678884178157351,0.9700439960036137,
 0.04496996525723729,0.23242978947494597,0.7582373997821311,
 0.45468930910205885,0.6723525964194608,0.6294564882963543,0.4570925964347412,
 0.47884683462003563,0.7041996111703576,0.641177466425057,0.19413693987186242,
 0.15005936552497534,0.7009743627271914,0.353606402071427,0.6620633537819275,
 0.9079567037643601,0.8226528281176585,0.6306246172634798,0.5711849046015547,
 0.7475704277164127,0.803829801045158,0.4356161192748049,0.10062234724180896,
 0.5961437309973072,0.87797420059526,0.39152904997472393,0.7972320224994311,
 0.9065096138224211,0.15554404065181482,0.46433196982446223,
 0.30965282613464473,0.4170155262741527,0.2441735919162462,
 0.17878363620785132,0.4290341845390868,0.5588365754529523,
 0.01997658742231545,0.33538526844184713,0.062434385706105155,
 0.7891192582229173,0.34266125104679657,0.8636520762672252,0.3487210955059684,
 0.21474501196317786,0.10422940026246985,0.7373024322585224,
 0.7870721381315979,0.6187316565732969,0.8246989203967243,0.8025215339218568,
 0.026052769045928348,0.6442538693452644,0.04762557238023479,
 0.566524021654612,0.5154935765754691,0.24718350192134686,0.9316901243597353,
 0.6362773936910437,0.43520901896042186,0.9412176039210707,0.5059078871857465,
 0.1443992977833265,0.5290670213733657,0.33566876060649153,0.4312808148116909,
 2.390862409288319e-4,0.5548416959420931,0.02938640270031523,
 0.9855866032329411,0.5046323276897967,0.48665679737098233,0.9745459641948201,
 0.03794040243971608,0.7297907936821232,0.41945028020500197,
 0.9124416770539809,0.9492187938477368,0.2508158395098318,0.4593610747354494,
 0.6823815960830587,0.22702986985752094,0.727723782374334,0.3655112971120119,
 0.5539338473647855,0.003026840651677798,0.98597809284539,0.2764302500619815,
 0.773553283259425,0.15703598132157248,0.9380257588763483,0.6817708263824076,
 0.6783475329460913,0.12478901380202678,0.05273627188344765,
 0.9275961095451777,0.5116176847047897,0.011486475575839372,
 0.34842603707569286,0.8387854247752818,0.39712033618557185,
 0.34631477578713765,0.21298909828939128,0.5478072763825459,
 0.9053901240420844,0.7494540675545487,0.23372711802149548,0.8416736807055037,
 0.0817155698511296,0.8555629264758553,0.888564977100986,0.8998069733351486,
 0.00270686842959289,0.10226875421945092,0.13404460324755763,
 0.9693311242460965,0.3493250451752603,0.6281281877640338,0.6653441950828296,
 0.8450211101299032,0.8401599391135756,0.485200655151903,0.5873709657440676,
 0.044049059604083673,0.27879581313064505,0.8872319889945535,
 0.8320405641120853,0.21031729881012484,0.8907634076699078,0.5723585683930379,
 0.4278994398120577,0.2246637152891514,0.8467923616153755,0.04740333052224277,
 0.27151915867722354,0.03378266101971095,0.9380755064233972,
 0.30965884349179795,0.5692073540957384,0.8768390666857593,0.4500075420016707,
 0.7016880547813747,0.6517693998779097,0.20975605743339565,0.902350322713068,
 0.49168161177761427,0.40742379791938144,0.5257142508531427,
 0.26938639728365243,0.7164696806620443,0.3630161947367797,0.5679716600780585,
 0.38935629574065944,0.3669570174719343,0.020983889706614023,
 0.9281261400878704,0.32819385089944175,0.5617242904703985,
 0.32715578063391926,0.9173782680143454,0.5011329379458758,
 0.47592320396338705,0.5462796590108996,0.16695878303623723,
 0.7867682156533891,0.6798732439891519,0.33143614042421854,0.6626723859667594,
 0.6902130716560255,0.5643517811990917,0.3457616926186283,0.44920091718199984,
 0.06425024512228628,0.6103393470217932,0.6878610906163771,
 0.28617170568947015,0.11239729386142305,0.4799144873826269,
 0.1263373315242584,0.32373006840854346,0.36310173749616736,
 0.8293657868428667,0.2914574624842695,0.1833571547922629,0.7258353974370347,
 0.838826994622484,0.510026876377842,0.12917572040305414,0.8253799810749443,
 0.3738870041964273,0.00905771340278605,0.49748232392430103,
 0.11975681136276928,0.222634463152384,0.9130441159066269,0.6404732402398333,
 0.20512613693153592,0.7489215842749078,0.5368547496761044,0.5101385162060494,
 0.3681725429834727,0.19894490794808584,0.39301184480606977,
 0.21193499115818892,0.5825649586942518,0.11212825505725932,0.40302378143896,
 0.9848320329754793,0.9380038058040533,0.5911299193678214,0.9964191797798614,
 0.8155140637075917,0.07496322973740766,0.6163150972156369,0.9415664065252172,
 0.26127704435750687,0.842547051153726,0.47504728458774914,0.9081259800392966,
 0.004293191487410586,0.7958744610493119,0.17125797506169316,
 0.7519244082965261,0.6022765010600346,0.6729162596440299,0.16221885555454063,
 0.2952057338005444,0.3775581697653906,0.8898814222417102,0.2508546318472902,
 0.6122823679820056,0.2830847784256908,0.444653027218163,0.7365765761370588,
 0.7874504423395818,0.11979492494336141,0.8123968979508276,0.9323767535547778,
 0.15375133200119262,0.09719969929192795,0.011390578206006308,
 0.9913937928290281,0.4272326365132656,0.6718857890415291,0.7422105637639329,
 0.8745023750591352,0.6772766451184481,0.0611376715274744,0.5637428736336242,
 0.45369270067977663,0.3521797919186296,0.9649006710484852,0.3863429175910901,
 0.3399185133095267,0.2902739061730478,0.7022816217106314,0.14752183356718396,
 0.01823532134488537,0.8256257262379232,0.9025122549182243,
 0.24217379098989056,0.7302654878050457,0.5797309101142705,0.2604587292021646,
 0.7063114292820764,0.5854640529783268,0.6995878245909777,0.7943219953644111,
 0.16100232007637372,0.22700446429077858,0.6022148615182349,
 0.9954579074056374,0.8616426189318069,0.3137941726575936,0.22928347162966134,
 0.2014761835099096,0.6562872058596132,0.6284588076595871,0.6400759392839765,
 0.9110904074423065,0.975630085046277,0.5521514168527042,0.30717878904235973,
 0.9121227751004676,0.4835671516615673,0.6163785177303244,0.6015788800568804,
 0.9162055012887217,0.8139954317515834,0.28833512514144743,0.7418872036595082,
 0.06684298887067397,0.32825541675971137,0.5693848104683691,
 0.31020173179931865,0.9849523754655949,0.3051559495418177,0.8949857225584734,
 0.4435595267304979,0.9004802693280973,0.8383518541353104,0.7729811427525146,
 0.7567456157889951,0.012077420431715558,0.8534058400255937,
 0.31761610127324125,0.38130425002622637,0.8101865958939004,
 0.3137072793683865,0.4208360683109955,0.10963024986252068,0.5764620167236453,
 0.8360009378387572,0.42605300702855464,0.6949646175558194,0.6762333850858612,
 0.2762799841001966,0.9418496982076088,0.4387616409133205,0.4264889828433207,
 0.18258847070866546,0.036117807053367645,0.05272829838507764,
 0.2368749170517268,0.9093808436807348,0.2155341339777106,0.5514306151038982,
 0.31280097506215,0.7502304332214811,0.4900922278419948,0.4177471569775322,
 0.8566339624066182,0.3345311579100271,0.1212309649181953,0.3061726506881628,
 0.8568828988244159,0.7425073650134916,0.6839872559303883,0.7994651213209192,
 0.6765292258441753,0.5215040380555832,0.14414654992538345,0.7942337579555588,
 0.8806496535440421,0.16476302234030915,0.4329636913431095,
 0.17418852810672625,0.9116264355539925,0.6004796786908451,0.2832230837539853,
 0.4431231507981791,0.802846086723658,0.41108902388372637,0.42315394917877414,
 0.27837223179755577,0.583649715627114,0.3350090918176113,0.9374506053977717,
 0.19054756662649686,0.9641594952520146,0.8751699537983973,0.8994996261636901,
 0.27450050046083163,0.38077031143647744,0.843226082410407,
 0.43756921942498106,0.7802426717136447,0.17146899722057807,
 0.5623013043755047,0.047468507365752366,0.7445889373456843,
 0.6430559368419092,0.18246701050410374,0.206782895687192,0.9621652742670388,
 0.5149070945629141,0.7505759996503314,0.3052300192092169,0.7472324003557655,
 0.8092352368146993,0.9096011380500654,0.960148611423857,0.41390612857410725,
 0.1349932019840493,0.595239332129565,0.06359680883836705,0.8310136998213202,
 0.4308690077952053,0.5111175631996676,0.7833226059571945,0.9594348873187873,
 0.9427240882776241,0.0633911471505646,0.6606973385528168,0.5934909057547143,
 0.048768559135515144,0.43333888284192734,0.41899407377168274,
 0.5934515340395893,0.747178996071876,0.9094139437907908,0.924589061318437,
 0.6883004857760558,0.9888372197680961,0.9208000839019567,0.9600593408926991,
 0.6676459101305818,0.07709648087577581,0.10745572838019213,
 0.4476291728798705,0.46930750216367734,0.2171916490904683,0.8166129010228165,
 0.9378782715518488,0.8914897527093772,0.6686808115486581,0.14918055023141485,
 0.11740199228704462,0.4899318059101717,0.7678702701033635,
 0.44601514024480204,0.16183909610343694,0.3268770759488481,
 0.34723820866377086,0.15915891177433783,0.12697008461914028,
 0.11523136385024124,0.45399330790607095,0.8637367745929555,
 0.21853527772320236,0.8827599751324093,0.7409058395436354,0.7794383016568479,
 0.3077717671135478,0.7681144708611873,0.5552117810494974,0.23128984294387644,
 0.9308749045513508,0.24919879562175917,0.027985414459452573,
 0.605759553319207,0.4178365702671618,0.46919356896558806,0.3939039332483768,
 0.10158792198850075,0.40427946158347416,0.7366837474704486,0.787656214500562,
 0.5472992067033587,0.9544329932834554,0.868173065752411,0.9335666070073237,
 0.8383612513144801,0.7555978364137784,0.43409880732505535,0.5910787580626405,
 0.7817092302903911,0.766828744648202,0.941780785232312,0.4064910178507821,
 0.5387338539202635,0.18461973079546645,0.8912104589696801,0.8470313132918005,
 0.720303561558442,0.6408835687887439,0.42316524281783763,0.7234155707804768,
 0.2746167800004078,0.7374833363832793,0.8590174507673074,0.6717897074062401,
 0.25005882896777254,0.07419600116798586,0.4639047904152809,
 0.6707317642623415,0.31834708486247143,0.7553952580270346,0.4795563745421917,
 0.8874292941057947,0.574888899219692,0.07623612275657288,0.44490056089948826,
 0.6181320561517832,0.456531799217618,0.8422294987437412,0.20780024504571093,
 0.21565113443075257,0.8018634783348837,0.16453668558516665,
 0.8478140023858853,0.8306211272834009,0.7807973726182852,0.9728026645564434,
 0.8688879314111873,0.6059209904602143,0.7882508051913939,0.30220068871453165,
 0.7577281343545897,0.7469868479073899,0.7313290502055885,0.6222342295967564,
 0.8808511542751049,0.3025556053495707,0.9301230898887648,0.5880195891883717,
 0.5817148573364174,0.25924946803276794,0.8679677299933373,0.9039403222537675,
 0.9776566826134094,0.3074621195325319,0.06354387017423246,0.9516907566818145,
 0.6711285495100111,0.6362692973143669,0.07062685230240917,0.9100048238152935,
 0.5675329327125002,0.041589115529497955,0.882541176285567,0.6915894167661191,
 0.34566117278274455,0.6790881567373415,0.39384072033296325,
 0.12240053563063558,0.10213162894125505,0.8123141316191281,
 0.0321361134761321,0.314934851104682,0.41621539799878426,0.5970235700705506,
 0.33583789845423606,0.3636610277603123,0.91686680757159,0.27956728510970374,
 0.25838560070270034,0.9103604346633882,0.44275765420273716,0.692060261158253,
 0.9328143694531175,0.3786180584675404,0.9971582540856074,0.6376500591821057,
 0.958316998479476,0.4572646336877164,0.038085026368803954,0.9576547454118288,
 0.9926359749061893,0.5887014217811852,0.34348690123092884,0.9001068473440454,
 0.3385287149498877,0.7658736882143264,0.405855655308885,0.0719153792775542,
 0.4043169020715657,0.240853098551709,0.6444094312592502,0.8376323850160385,
 0.9200468853926302,0.6392360318181495,0.2962260962765467,0.4643454437079789,
 0.6226663248204621,0.8267567230751488,0.8243750590908369,0.14740808326222476,
 0.8977348055489275,0.7601885425978427,0.7911484697629649,0.7267527092698262,
 0.45533370929371464,0.5951485683957137,0.11838365880427815,
 0.7523466467948579,0.9380568233989921,0.17792065257000766,0.6527474441989666,
 0.3203313160602268,0.9742842317919209,0.8164326322928481,0.5488495988835511,
 0.6922193321308552,0.49061792890195577,0.33976453545384866,
 0.26454417835769806,0.13900860236070023,0.22348761345830426,
 0.8562242067830848,0.3542882417023081,0.7101782697801727,0.7661460100446226,
 0.09512789101245955,0.6975576291214267,0.9156958617157628,0.6364724479714962,
 0.18417028051031525,0.0639623719247373,0.6741994129784251,
 0.18821498470891662,0.687276594273832,0.05725377415699395,
 0.26952733924654493,0.061967636748614274,0.665911289498478,
 0.3859944708477824,0.6736081279746362,0.19963012501802258,0.312018327973521,
 0.003986484389356004,0.4368208174590872,0.8511114540660825,
 0.8135722279975175,0.11224959698473158,0.7032977306099429,
 0.48244609094819446,0.08865948211659858,0.11923860552666221,
 0.1194645982559005,0.389263782190195,0.740413795856487,0.021444639419101463,
 0.979225975840893,0.47219121697671007,0.26863667378421996,
 0.39348915161679965,0.4654900954250738,0.5324275728394546,0.7263980477188587,
 0.341589722018418,0.44734278490171997,0.3851291045637355,0.8318405362648796,
 0.025204884147304374,0.029620649701849633,0.04664221860354245,
 0.7968757815705644,0.29642324677021903,0.6895541905200275,0.5724364256791267,
 0.9956857877461938,0.09270433310186366,0.17292516255095103,
 0.9714123115743052,0.3094652227691699,0.933529740376812,0.9703094648275556,
 0.45970849939659464,0.39859139880948813,0.1286190746695045,
 0.5227865363658933,0.3014008224001744,0.3177430702312709,0.1845653689282658,
 0.8056830345079385,0.47124496078493183,0.5120590608264679,0.841965020660174,
 0.8625459126582856,0.03895717654721187,0.13227745172409455,
 0.13808175245826615,0.4070030149027981,0.5389827854346667,0.9598360592829401,
 0.8448428837897887,0.09899663815134807,0.7440130803065963,
 0.014171625042909919,0.6569301010152375,0.5227200689194457,
 0.11953254096902954,0.9175988530275694,0.8576244904926771,0.7538645732613144,
 0.7795880115105931,0.3605677523546468,0.9833676623333486,0.8329591884813685,
 0.8576962072500419,0.1614126277081881,0.492535802858178,0.5550282121686926,
 0.6771448250622704,0.3118675862120621,0.9916412262161995,0.4429643262643803,
 0.16013710276291437,0.08224792790503233,0.2274868511583802,
 0.9209305565371018,0.2712378187702942,0.9932579419664246,0.8336956794437876,
 0.02692489700152656,0.25154748723427345,0.3552718841340402,
 0.6230483392449511,0.5085496313333155,0.5928385302414626,0.9115543013924802,
 0.09631770635109171,0.15664221329180783,0.45892403825612127,
 0.9228852611501197,0.005764179195762109,0.26712812797841456,
 0.04568697743884864,0.1482349560072006,0.23528227926283618,
 0.8323376010367882,0.9324346351119108,0.20262111173149344,0.6035763451700009,
 0.8882118992851951,0.7798656187525378,0.7192109005096932,0.7133556100869535,
 0.6833674784507678,0.11081504829484168,0.6630852137779404,0.6884635148208321,
 0.1859380222374084,0.8744481249191481,0.7005289995564232,0.8964237341867718,
 0.8604002661816389,0.969696839743399,0.28085637020902154,0.8299624911662804,
 0.5185265090787939,0.61804816821245,0.019199987119876027,0.0668893786438981,
 0.06614041531109083,0.8944169942519827,0.586625307545784,0.34274514839091463,
 0.3541329256672954,0.8982885676153934,0.3354858155828152,0.15216822852637502,
 0.688722661288634,0.2926165496301868,0.9304375967788583,0.08167681882360212,
 0.9132852049669961,0.5350571664751681,0.7200513993233274,0.3532573218978069,
 0.5644855354760107,0.3220760970119223,0.15880732178145873,
 0.16066303749038358,0.719941885498002,0.692734983802735,0.03366208636688128,
 0.11592894329379999,0.9725691669404666,0.0577400712756867,
 0.17888580194592674,0.8339242748940674,0.08865073060911222,
 0.5528229831864393,0.9684947600491188,0.46453242266243944,0.6465235958752147,
 0.7814795882605104,0.4156507372051037,0.4860367111868007,0.20001897012034642,
 0.20424648860138894,0.7379809655766879,0.6755542161451255,0.9694464112127762,
 0.9877856481558257,0.843938681409162,0.8150920208823755,0.7584853774044094,
 0.273466130710306,0.030901365592091334,0.7515199821032238,0.9788892881150922,
 0.18124215597262328,0.5668046083298544,0.09108396267298124,
 0.3117763751690956,0.29720067376069026,0.848911089728923,0.26208375011558416,
 0.34659997143552645,0.936630389349652,0.4558015130331835,0.7573182164813225,
 0.7157972647118562,0.03356413182179907,0.8100531901436316,
 0.12153040963075812,0.6125479502812453,0.18502289025328256,
 0.4397400139062885,0.05184026357485061,0.9411792061018845,
 0.34043129951126616,0.22657617832930377,0.06048706322717767,
 0.40175959628168045,0.48461249648034554,0.9427443078066837,
 0.03204930277215423,0.11318530508292535,0.2670675183790241,
 0.08754987414229609,0.5159121071687067,0.6466750063216765,0.5675151529866862,
 0.9339725090714133,0.6802651442254536,0.2901076549916798,0.8727665672856092,
 0.5796149307533819,0.9953388823612193,0.013860188748801017,
 0.8673311167453631,0.5091392439975332,0.8388890726478588,0.2546536262075415,
 0.42320683860677255,0.6505407637004448,0.12078687270941202,
 0.2736572167697606,0.3287561080672987,0.2881094318444799,0.8326189753064828,
 0.04134963757643062,0.8892932816656847,0.3061266667252467,
 0.23790395270937614,0.02888245438975523,0.8457265052848298,
 0.48318012827647017,0.8062850028343898,0.5080281199696073,
 0.27408493474728035,0.4576098247105398,0.32976546806953144,
 0.17657745842478345,0.3631243016593698,0.9606543098117237,0.1773610685280892,
 0.6091519272799686,0.38273434238298987,0.900896673030784,0.4573354300205986,
 0.8567643585513045,0.18261347360732139,0.9151874972794076,0.7522660875165144,
 0.1286307477451254,0.17574540222416757,0.3424909720731957,0.8312357479792395,
 0.8332283533184095,0.8000225408864721,0.8434837161850196,0.08268623473072345,
 0.7028436666867732,0.6696509860530984,0.5379133746222114,0.20336386534684459,
 0.016544370873169845,0.9793094306694156,0.1170359674853726,0.954727236764452,
 0.08030523405657408,0.6863181919576068,0.7913467931321654,0.7322040222508123,
 0.7495536668831299,0.5611004891748778,0.6265924532641328,0.604404344735855,
 0.6126537848821499,0.09696272213930757,0.24952018468474668,
 0.9961191615276609,0.20514125611542078,0.33859619606912617,
 0.31054199302203567,0.6639070482426942,0.9105694898415858,0.6462996538842681,
 0.6307720625203151,0.8042667820397517,0.2961733008566786,0.07281081069415152,
 0.8372784355279139,0.6621368341449225,0.8523470394077125,0.8720000023720157,
 0.33676873104321603,0.12355177708056975,0.8885734456358646,
 0.6220202404483265,0.7861270660530495,0.2944303647767902,0.3990968836628359,
 0.49689532417520466,0.026656080891520162,0.38951201922269607,
 0.8748865810914134,0.4764867766075175,0.9410896843269216,
 0.005271701304498944,0.16354594732854155,0.8635241807544071,
 0.5692046639816136,0.7399965627767048,0.04708022080936869,0.6253012044018927,
 0.4597307887504287,0.47717221155076084,0.09690138784347713,
 0.7473517191051378,0.7494968868916773,0.4254179956662075,0.6487568363363749,
 0.4852376300015291,0.1543269439327284,0.01233777925505175,
 0.36712993233536273,0.9766372462135114,0.14383507647103355,
 0.4019927243053598,0.89211526403778,0.591508778777959,0.06676069529066431,
 0.7215862132940533,0.7585612846043485,0.30832057690230696,
 0.21684819241855913,0.5926738079929039,0.8912208573725168,0.7223123503032133,
 0.968598026648583,0.33853941350375016,0.28331036860935843,
 0.13900651830200672,0.2397337037607291,0.60735565511384,0.36595832837361675,
 0.5498608509915522,0.7092356265350525,0.8635355426500437,0.24686003009342472,
 0.492010467988628,0.4415714996583888,0.2450016690558574,0.06599835145323862,
 0.630816327078055,0.2464512578068354,0.8795973343694917,0.49459606595219563,
 0.720047482710596,0.009829876811026206,0.9456176764937021,0.6117477065810577,
 0.6606806180846908,0.7822393514688372,0.7915566738032814,0.9075752061466886,
 0.8990698634693826,0.405812880530128,0.12219225222010033,0.4632634355506472,
 0.8049527616176866,0.8902160717056744,0.8956795584411354,0.16466505170554047,
 0.10981639815200805,0.07357479103681364,0.4276097965827983,
 0.9013169025358512,0.11607920120489923,0.1520464926360663,0.2145362188717309,
 0.6590130911410613,0.49455412090466755,0.6179298426222266,0.4275855989640872,
 0.009217897476041292,0.0027784279321018124,0.43923552414557043,
 0.4532536914784818,0.08473364303767728,0.6282352333917745,0.7274804386265503,
 0.5280198117762395,0.635110632563228,0.007678533183504643,0.6332974209003185,
 0.20506815854234595,0.48263175024015814,0.2353600127012334,
 0.44968965701478825,0.6346428688298786,0.0927277117799088,0.3683166786759441,
 0.6151581124232319,0.5115045163791748,0.7723456256761533,0.21415726368319365,
 0.7136149929112614,0.5846652273613299,0.3040788939957,0.5268857343148831,
 0.5398369658660409,0.4149497285802116,0.6562155797725374,0.4013301762454916,
 0.710012377886989,0.9427842395316631,0.10346272114918764,0.4937044733839627,
 0.6794141165685578,0.7208162949967664,0.4204246115634891,0.7853921220547542,
 0.2744336803306475,0.9846182366605205,0.8338182573787433,0.8219085410456708,
 0.045148765956488535,0.1425624813890326,0.994870206967108,
 0.03241054863594928,0.8471113039166978,0.4495562738288096,
 0.41958724137190573,0.859403863034452,0.3200286607810032,0.3149950564949222,
 0.29135304165495546,0.06516830146519392,0.16073761093113836,
 0.3857802902972868,0.2982457930100839,0.6169038132790394,0.44030778069565635,
 0.4736864250910551,0.9036995317078584,0.9068496691748622,0.04280847561562051,
 0.7254723527838527,0.11897131133697281,0.9468739399148993,0.8112119909679079,
 0.8315325657587809,0.4294303145529468,0.7952462620427011,0.9205260533121035,
 0.7042650022881503,0.3584300274843839,0.9253745032289957,0.30521859577491905,
 0.8900136877796249,0.25384181481191737,0.6963144303257489,0.9550023373196084,
 0.44703586145319385,0.26321518256277665,0.4149866900283299,
 0.9679252503152763,0.056388758840623554,0.10710321139310441,
 0.913818655229159,0.9802603659689187,0.3042581386743647,0.7744310286248742,
 0.6052550713913737,0.3193347919614653,0.8755032097191854,0.40216905554090265,
 0.6053202844414733,0.48593153142402434,0.017964744524002763,
 0.34844782746086334,0.18566361480851468,0.8665547340717324,
 0.12951782756769037,0.19385747794831043,0.8438756273278618,
 0.2844532855916173,0.06367114111741223,0.30745858392747616,
 0.7844023679688594,0.6134547903416779,0.6481004406183571,0.5557149563010756,
 0.3391182159752646,0.8178794334364422,0.595898867242894,0.7561873598480685,
 0.06353139356608595,0.8451318357295363,0.643003498204971,0.4814135443541062,
 0.07817688374955156,0.2763759409858346,0.6501592624730381,0.756193319332608,
 0.9586514364615022,0.7117360797473121,0.2643589523207962,0.6017148772397688,
 0.17656914597742113,0.768072372633152,0.7393233994153172,0.15696609323473099,
 0.7261604667757274,0.011830941581476906,0.24657371205353684,
 0.8808245802769887,0.5672736744704395,0.5343565272407409,0.7234818702795613,
 0.00938678105166768,0.19906552437771063,0.029236755909976875,
 0.8762680004835917,0.6254073871145918,0.5158502155301952,0.6868669922877625,
 0.8461127034956973,0.3356613927355182,0.6234915771193851,0.3671708211201017,
 0.6364088280923738,0.7434520656079413,0.9045687407654074,0.4450517938889531,
 0.3052206062509507,0.44149085054658155,0.9958135205575777,0.149004401800093,
 0.7085598278556531,0.58487593398197,0.8948279980591776,0.9634773459950301,
 0.6784845265398831,0.2613095551786113,0.3063083911854402,0.3769671915778653,
 0.6070877255893289,0.7041422761843515,0.6136661709827731,0.7695414231490012,
 0.6639653914797774,0.08916032482095115,0.22451016164902915,
 0.8657564269446254,0.8947852825470095,0.7546405848181612,0.4843967604979802,
 0.002636430108711618,0.8987381152881064,0.6663713553211478,
 0.6673765885201449,0.723271599872739,0.9617493604115006,0.4089942733059454,
 0.793475176175372,0.9508650453969862,0.612683114115396,0.32383430464240737,
 0.486197785840075,0.9153626505067614,0.2277185482282721,0.9798913193497839,
 0.25358779916995766,0.8230735535427418,0.6093518671468015,
 0.45330604722084566,0.3143362198724968,0.8416140292125636,0.3642255583997345,
 0.9652880695376544,0.3561485232102788,0.5358870247688534,0.9343025578534381,
 0.8707089974803843,0.05576186389955373,0.18374654509266586,
 0.42992822167752365,0.22310574494616386,0.24012303025259718,
 0.7193920575686871,0.8204288047078541,0.3303385719909294,0.692733955701769,
 0.6030703237696239,0.024358562837199282,0.8911318069003438,
 0.7809364281035436,0.7122227041625665,0.8298374868924122,0.949413619332243,
 0.8896410215614884,0.9691607028402426,0.6763530827155617,0.09113072208805184,
 0.3682057713979532,0.08611974150635027,0.3314939865936346,
 0.35624285742882433,0.7561255599079783,0.1518317911060838,0.1768088644622885,
 0.8037261596583313,0.20651140912722088,0.7701823291349711,0.5823930327911837,
 0.45072026147990685,0.7186783303310356,1.3618458561825975e-4,
 0.8064387551027238,0.06751839423516526,0.6377114610287028,0.8565345910600326,
 0.8983343410316376,0.44734551782084386,0.9688579566263937,0.8390979881325717,
 0.9551429802447873,0.4022056498558608,0.9201634010668778,0.5399710481730366,
 0.8844604464641269,0.6314612867430647,0.8891245255065093,0.8687032895503521,
 0.8793057442681224,0.9684405959423553,0.6065342031501305,0.04469570592322292,
 0.518030165858415,0.13558175751196422,0.0827318448512595,0.7816252365406018,
 0.1762759983963842,0.7154718514206939,0.9833330493820446,
 0.012348552199872853,0.1525320140070573,0.11311568869151833,
 0.067463332601999,0.3956450507260866,0.5445439106847136,0.43884136238860705,
 0.91271066696472,0.5718832957033754,0.8036371201715324,0.2993785999351821,
 0.7225399191932036,0.08891051110087778,0.17887448485229474,
 0.5068678434638614,0.8611368296920461,0.6536652026584909,0.46677123199411363,
 0.7341303712732321,0.6961517778006823,0.9592181121269907,0.3700798474330913,
 0.32455150502221164,0.89345982573742,0.7678989609191443,0.5727838732297784,
 0.7051200884450253,0.642360629200947,0.8976217414602291,0.5958056307598778,
 0.5884768872606466,0.30769276178257465,0.19573829655359243,
 0.7107570924168205,0.22704861251323616,0.224099021970938,0.5727753042747004,
 0.2843197154618242,0.7714611795323053,0.2276477042469861,0.11038739958108157,
 0.13870515547881512,0.5438379945183136,0.27331870563229743,
 0.9329247147059183,0.24068751502741237,0.09582288403547556,
 0.4695223952790446,0.2943070202946996,0.5286981610405733,0.260827077084932,
 0.04932303281472061,0.8266686414767883,0.0814863379567572,0.7212967348039224,
 0.17306782242744112,0.6358907961303727,0.41963109611043614,
 0.27527451682671444,0.6383350956230995,0.012061269496238758,
 0.6504487379681643,0.9635297771156404,0.4617360407829415,0.9802822244145817,
 0.14110697707522268,0.12070824667394042,0.12226394273182783,
 0.3746508457554508,0.6159503536467009,0.2821936365472143,0.411753595462671,
 0.33125488327040165,0.032320325963028584,0.26782652647968563,
 0.0747225037265874,0.6564910238862081,0.06780722788232196,0.2841415696076095,
 0.44245102249604007,0.7374109040248671,0.2767921504381111,0.7068108708892544,
 0.667582007459802,0.15693322772321572,0.6971370812717046,0.19597865730602382,
 0.5122464300637942,0.64482244329531,0.7792748624316383,0.7949311063643686,
 0.23957251358672527,0.36709696677469517,0.2556331968726644,
 0.20497989215930135,0.6335079647510816,0.688349268753005,0.8354305615564409,
 0.00579970763962423,0.343786679014466,0.9924351268356201,0.8949235808439542,
 0.35611178157546786,0.8176575720175563,0.14285980895952743,
 0.14614963987073015,0.3970410436095583,0.4621390377682041,0.7327637680908009,
 0.7949023380822773,0.19146520960270952,0.7710836044139256,0.927001202083021,
 0.8138500262195818,0.12363406627419238,0.013810423070414801,
 0.00594644911520914,0.0078596669516513,0.40276962801224236,
 0.7656406993460237,0.7965152999198674,0.981457955944041,0.5866257112029658,
 0.72153801676182,0.757845419624504,0.6830643221672532,0.797664305778819,
 0.31332326414471734,0.18767867740215383,0.970243611920445,0.6555953720434227,
 0.4486195564179747,0.8247737390596501,0.4853474231797813,0.6686526999386991,
 0.7357271512686572,0.9702693597697016,0.9934055472848268,0.19526185348693004,
 0.7090260094742245,0.7957083878478071,0.5375818942143457,0.6026943082752126,
 0.992889168351542,0.1639523246686715,0.20576402828983342,0.7357388133684432,
 0.5367917169616151,0.9104869640165243,0.27851649008696877,
 0.24602239748124588,0.9524366050112455,0.07965310712869811,
 0.3750135422821772,0.7876261386270988,0.7653792675884289,0.48287464963888094,
 0.16366323370013847,0.037600725407442304,0.8792185485583712,
 0.7254252089904822,0.5961963648830112,0.17259837981278958,0.3268919492359472,
 0.6329539365097514,0.39430616066313295,0.4914926537109022,
 0.20845815844993676,0.029375288172138747,0.292981350196588,0.353670555654537,
 0.15958238900745125,0.7024662202622105,0.11837463735774556,
 0.19444284106207677,0.5579027078180264,0.7021016231642851,0.9335303089104179,
 0.18284143525458596,0.21439801511834555,0.7648649448049787,
 0.5098678420884931,0.7596160826878953,0.49491293682538817,0.9560159272616301,
 0.7046928158644014,0.8272144785666463,0.9953446783750675,0.635365873766695,
 0.967866195517501,0.044282682995549605,0.36497091779423374,
 0.9651208839544536,0.11138519109498335,0.7993000051535931,0.8216653675885178,
 0.420563008020597,0.11891108856108334,0.4213366226109406,0.0665892727365186,
 0.9717215711286753,0.8731354480667839,0.36492485830846055,0.7292823641542532,
 0.7834254128380335,0.469981997341093,0.13468756748428523,0.8751832208329587,
 0.6873224579284863,0.9667267882228747,0.4888852165021652,0.8425750420272493,
 0.7352501472690025,0.6360366022858571,0.5780981961696315,0.929510232067366,
 0.5124809771866026,0.020473988297261214,0.5677557734228691,
 0.5546555601557022,0.8841481395799087,0.020712879693369435,
 0.6968149338758395,0.5772002853437477,0.47447621640520876,
 0.21125123159002213,0.6963636109557256,0.6293510329724714,0.8436761081355355,
 0.14662271808663352,0.42986885656133933,0.9815504244152524,
 0.3839906059584953,0.7963901191999448,0.8763952328601349,0.1967087851621553,
 0.4202725476892436,0.2735213438930191,0.751361622896593,0.9785290976792588,
 0.7661879894951743,0.515963265778489,0.645461338066023,0.08535923902997211,
 0.24129579808191726,0.3204214765590736,0.3114163525391619,0.4478191341321843,
 0.21018468171598426,0.0881167460797796,0.5358039801837434,
 0.31854475772744173,0.8270449741816773,0.151636117236577,0.8527137778222061,
 0.10478709833844935,0.9708316910159466,0.4632907437854039,
 0.38303432286732664,0.2962481026750463,0.5722854755252812,0.2676787957493101,
 0.20767992213514108,0.38195330864455423,0.8435586524527883,
 0.5606558885292092,0.2523376011751972,0.6818497067974147,0.560692311518439,
 0.9628526535859069,0.8474101070741135,0.5004168915793042,0.934673963738314,
 0.9503502067986289,0.3760761691044452,0.23446605076916938,
 0.47871156347439714,0.31535719425113673,0.2656805290431161,
 0.9936391564683466,0.2566174980886089,0.8930454956784284,0.5877155974392438,
 0.9554172495351948,0.6629369059730094,0.9116390925819835,0.8698063077543366,
 0.2224207626511101,0.21995200459057962,0.501357450270133,0.8069791016634716,
 0.541836142165929,0.3851713052206882,0.347080758635987,0.9894962725958799,
 0.7046999440864893,0.5535674162576785,0.9611992628265822,0.2560563293520972,
 0.19238866170889535,0.4304321891510271,0.27706751138747787,
 0.41731014330179317,0.07203245865975827,0.09031347710818427,
 0.42737491487449486,0.3077091474911341,0.12584219450066514,0.675187118598418,
 0.04840092696115139,0.8867071032462075,0.9716441721170739,
 0.08574985525318835,0.08381870503040334,0.6411110975907716,
 0.23242603436242837,0.0439865123551737,0.8283435324430211,
 0.29528050356301216,0.7185795476568326,0.4941580326684052,
 0.23891338981348076,0.15860610476917836,0.42334764096657607,
 0.6127873473533638,0.5611111459070612,0.9964744680576905,0.5445549641792582,
 0.14905216916524378,0.4778410762926414,0.2260014112879677,
 0.36114060782647284,0.6791440451049987,0.40653270396425645,
 0.10101093055355848,0.1790823634968588,0.9504712750681582,0.8066088638229242,
 0.0962447570949081,0.3512924253879828,0.35029291540592866,
 0.09197069517881573,0.9914573660074704,0.9161937820051435,0.5404864637909874,
 0.18586757004248844,0.7153760492761104,0.6985432240060568,0.3413751472527745,
 0.16146442374593017,0.987602647245399,0.7957925520162873,0.766600181331415,
 0.041885624100472474,0.9041436927280357,0.9488930714906575,
 0.9034953388970923,0.826844269856311,0.5211936783823415,0.26158744168661463,
 0.05310584387697481,0.5757157511598225,0.7926286916027876,0.8018535879567951,
 0.9616319758257516,0.2529260883347926,0.9883181615231567,0.6095824491308078,
 0.5894922695910143,0.3204629221116533,0.41059404018791046,
 0.45928188005796866,0.6210855433240916,0.4501917594749396,0.3342950716523547,
 0.29131460158524924,0.21914115001809498,0.7416877853538214,
 0.5025029293247772,0.8295973953333604,0.012052044020473218,
 0.06973182349009222,0.30442969792724184,0.021511243847133765,
 0.926569037083932,0.44783426809297433,0.27218178435948137,
 0.32847042235081125,0.44117960413134316,0.6747555541013833,
 0.9289888749290323,0.0068716555694959824,0.33272018503015166,
 0.06319078437377235,0.03335532519610884,0.22859132900646695,
 0.8056533746962995,0.9632212076899639,0.18065059186732468,0.5303893992601496,
 0.24199256476262787,0.18704173808336577,0.1546283172505344,
 0.05800511351785265,0.5120917557572425,0.9108222840468273,
 0.38261981070585893,0.5760596014848797,0.5429404892409795,0.665243962575983,
 0.8394659719767812,0.37417075551181955,0.8579012279519677,0.4761762300789252,
 0.43441907329474516,0.2565069965947271,0.794561210167376,0.7661154085270779,
 0.8322158167562449,0.6995301308701236,0.8794842016660219,0.4548302323120844,
 0.5156142572257286,0.9580702249880272,0.17623650374236866,0.8644157904640649,
 0.6450770539869546,0.2465220158374929,0.9664528164696631,0.7162363874262,
 0.2516215517809034,0.5097306436471781,0.5719835492646733,0.7928707381812982,
 0.2563187660345967,0.5161565070774428,0.6729125374243802,0.4718240339261919,
 0.7473034196664596,0.053963557808346274,0.746431011806334,
 0.30494665673188337,0.5139396915173973,0.8055724679000852,
 0.07199749761037166,0.45134924140623145,0.5864374704839287,
 0.7652874980703851,0.576258807815951,0.42206403797542524,0.7254684906122959,
 0.3551279610051302,0.31344422107212533,0.8439348702014576,0.8607221273468526,
 0.5936336398364774,0.6028942122831379,0.7930827384283726,0.1797432500640601,
 0.8143556276756619,0.7872326012173135,0.7103819888718755,0.20987861846514677,
 0.9211253096924564,0.998592379269521,0.1185450892855171,0.05085190163402942,
 0.18689402993402893,0.14808809887178553,0.5460150971464361,
 0.10229891644574118,0.1792104839355002,0.5831673737826728,0.8011743472850837,
 0.42453331353924917,0.8884412653314478,0.2994342711248803,
 0.17998435158241588,0.9821117944220286,0.8814359752934711,0.5313437585432833,
 0.8883831579396978,0.5033212208420466,0.6345640529501596,0.973042878371674,
 0.10787316115399503,0.14777682498214406,0.08155548002861612,
 0.7024061401766225,0.5739689905899013,0.28249050375151885,0.8262772438478163,
 0.7344319398084463,0.13896835559131182,0.035089801242237484,
 0.08231629125943796,0.16210878040372356,0.9517467859365563,
 0.8955691136159623,0.5732541853928155,0.5686769773342972,0.4783779399766728,
 0.6423849815104985,0.8147707974830499,0.3311548093080857,0.1874826735930636,
 0.3007431744115546,0.21177253926827033,0.06043143071551049,0.91961045583843,
 0.49764183638256,0.09260814681873675,0.6885085629857185,0.45285812257735425,
 0.16327910877426788,0.613906360452943,0.6989603546037615,0.1712407729584653,
 0.16093414566754605,0.6325034884396773,0.9235266216093309,0.3977500686950143,
 0.2371623269293669,0.540766266599224,0.7104273120183228,0.6644027655011209,
 0.0532522607922713,0.7851759075577898,0.2412938945588884,0.8260338146811699,
 0.20122225649114278,0.9236074360998812,0.5190971177276257,0.2565486547483312,
 0.6048505312054328,0.47782237005616446,0.8451135888293119,0.5840343128802219,
 0.2103249507975382,0.4290411132179005,0.9770596231200913,0.5588176419736754,
 0.020776232754268875,0.64547684747892,0.7210961878226737,0.23971112675989803,
 0.9594521073389223,0.8770125441307862,0.6265014771771565,0.8195891906049045,
 0.4788527285473806,0.4912892959865738,0.4807870668778724,0.8453632663419425,
 0.596757451229622,0.11743832715282876,0.9426159694519654,0.2015536488916102,
 0.764311740075282,0.5105282606228309,0.7815647806133735,0.7935335569595582,
 0.07685327118399965,0.8490609998834752,0.31115031771470414,
 0.3600386189446628,0.04529176156540482,0.22830823276643641,
 0.029450504260784216,0.027430138934278925,0.9876034015842458,
 0.7630890499677457,0.8115259994907325,0.8886543336798431,0.566686872213455,
 0.11699104506024405,0.3260562077179543,0.38774882032591096,
 0.16223493074222473,0.5925334946564208,0.19950186982409135,
 0.10449820784039865,0.12343417541565449,0.026395318299193393,
 0.8657811336039752,0.14914718226145296,0.5098286926893828,
 0.48875373464930005,0.9113949239766228,0.8806014715893137,0.7975521191402323,
 0.7233701625027511,0.8969730892938276,0.18652520284201823,0.8664337723434135,
 0.5127473766383392,0.271826382270713,0.38108596996865385,0.9493985024257198,
 0.11905354059368722,0.3873732456454073,0.019158265662463037,
 0.5442396428358924,0.557065687088648,0.40262794974936567,0.5177849749591515,
 0.3646729823991056,0.15984739710614504,0.3500067705792522,0.4248804094860056,
 0.8226035365500615,0.28317485049017,0.11178535486573804,0.24669057821709872,
 0.09737765770667439,0.44429450144201954,0.0922000421819511,
 0.5857196172430735,0.15247044179929725,0.5640023741465415,
 0.018579489984383746,0.6683923226992973,0.00669921119969219,
 0.15223714576877645,0.020611348497203585,0.44441481774360947,
 0.8497821278423959,0.34489778941694726,0.1789293080994332,0.7340665458377191,
 0.4056988088232045,0.8942598116334329,0.07415711417298965,0.4185127162026778,
 0.4246812970604009,0.8158612903513585,0.18930748333437053,
 0.22351094747897204,0.7381439689703027,0.5875044675093014,0.7638123996713302,
 0.21049960950012925,0.5734592483415788,0.8806739341350883,0.4716989542309541,
 0.3739253950136534,0.579923178851862,0.7496050970599476,0.37325373208663615,
 0.5805642585416773,0.8790426523607868,0.7916864885557989,0.36579209002831414,
 0.6532094493337743,0.3448364779341486,0.382686083739633,0.3766093884435362,
 0.735223875615534,0.35090248011095404,0.33151640077501576,
 0.21384058419909835,0.28779236255021234,0.38786944004653634,
 0.6760184436313558,0.7304669477743452,0.06970420902483909,
 0.07232025311605583,0.00444297498383206,0.8659434287719403,
 0.13502045856435718,0.7311849811344939,0.09509093019700587,
 0.47384840987550303,0.5563231005420625,0.878016295388135,0.05813282804265418,
 0.4326136530564636,0.15565849237669882,0.26609822435707076,
 0.5226866754217953,0.40013519244388895,0.3606158013527597,0.3358148574060784,
 0.3853743969184448,0.18481589331275106,0.9813977908029279,
 0.032902312991174965,0.9696984425700684,0.6685506278795017,
 0.9520149489605297,0.17139305372810654,0.46801746264513566,0.956224044518831,
 0.4328869812006021,0.12087204365907334,0.4109118294912558,0.5178016343144046,
 0.586340037727871,0.4515855106447271,0.9290818738816374,0.9375905143787501,
 0.11605330678604442,0.9643305230430242,0.09159940687132573,
 0.09089600182472124,0.8764248982103012,0.7274970355967212,
 0.48455652798879223,0.6179663529613871,0.8373161904014146,0.315465012552979,
 0.11795093559761394,0.3839958713572287,0.012567945720136087,
 0.45040906401772673,0.7142267367393919,0.8801482636767884,
 0.44432853999106225,0.9119848074439346,0.17817424743455113,
 0.3867492781548527,0.7726689159861149,0.4286080491773995,0.6164288534100086,
 0.8094398856354181,0.7442078001208564,0.0871763822677748,0.3904175254796407,
 0.9632406537983537,0.11443142509392112,0.6832158323937736,0.7577573235604244,
 0.14651856901150329,0.5801742469739749,0.4758661020572934,0.2815303930735381,
 0.08674717711416591,0.6092028240645517,0.7432939231794298,0.936328072938118,
 0.39393477210677696,0.8420211106492362,0.6423031242941356,
 0.07985338984384305,0.4530494999630327,0.9877912621232893,0.4233310644987174,
 0.17959072570241497,0.21849924447864577,0.9439705311108644,
 0.0624875156838095,0.7221371312277174,0.09293917881334268,0.5895098285538045,
 0.31547755348402806,0.8881475612973958,0.10946523713953404,
 0.09695126336832349,0.1373872528640886,0.5146567559033103,0.5845881645712362,
 0.9350172651330269,0.3266146765013658,0.42988075153336314,0.486546798917707,
 0.8317469034381666,0.6009467907475191,0.015771940330690737,
 0.08623195610627044,0.059475708414443273,0.7641765464343526,
 0.10320628698876588,0.8732713768140513,0.024269697788067424,
 0.5384954719009352,0.46445507537863506,0.5164799793971022,0.4541472037618791,
 0.33784716573985973,0.2098608503442444,0.8961388399654074,0.922839395037394,
 0.7340651583192397,0.5051647076264535,0.5240220674917802,0.3778478422804966,
 0.48969864171174715,0.4252556295015315,0.40386824034838187,
 0.9766021295625709,0.04165877146540686,0.7125030351654169,0.3809632706344557,
 0.42077650895413443,0.41510736146757043,0.7756167284197698,
 0.5207145840853481,0.552812753553782,0.5669639855771891,0.25867281506433315,
 0.29830557775897004,0.13858958011147748,0.29033104661532894,
 0.2356319331122837,0.2527960469498144,0.5022009588767217,0.46034629926873516,
 0.541231425575249,0.8208380342876018,0.2241507305584699,0.9866647062733225,
 0.0016837202355903802,0.9460745300382462,0.47806324326669525,
 0.33204818367912325,0.5439291512335299,0.17334216912348188,
 0.2420780114946708,0.8525575965045811,0.3138391816105168,0.8205097854691645,
 0.08714542571634487,0.2063850430234354,0.8762314746449541,0.3020919687939314,
 0.28344711196741124,0.2986543489307987,0.5408752986425732,0.512560396504629,
 0.9857460670458563,0.7564971236843693,0.055141408377628154,
 0.35250055511297007,0.44020408665159005,0.10089783075438175,
 0.7309710942597367,0.4759033381327199,0.5879712773679657,0.64760021926177,
 0.9768760585858377,0.09554971732804807,0.07070858687804937,
 0.09951540887626553,0.7235759835213774,0.1511728957799895,
 0.12491729915402494,0.8800102650515621,0.44216023166508867,
 0.8649413306487295,0.9305973368929518,0.30111261306414483,0.5737322549747564,
 0.37667137216415036,0.024372106326756904,0.8687017705407449,
 0.5632426750458258,0.9972681474332846,0.559671179504824,0.9760248004357021,
 0.9721846637469042,0.19213883906901286,0.5417752884152344,0.5969311352874915,
 0.7059345226577398,0.7237527669748777,0.12810933856976803,0.723596406614778,
 0.4348872737179742,0.3612093330873629,0.33750182630503633,0.5815202576002055,
 0.8554170992649297,0.8389976968901456,0.9697869496662211,0.9022185683510747,
 0.910528149160275,0.3271655301259532,0.7652256662683203,0.8480707269849248,
 0.6186544732540687,0.6353362048762251,0.8484100540897671,0.2667148752622672,
 0.4114164512374151,0.69165467094415,0.9459023964927188,0.6824683851674722,
 0.6961145087737527,0.704280364305081,0.8039969695097561,0.8063916226627658,
 0.7873269296278744,0.25621427545764086,0.8734888201197619,
 0.38707566242308333,0.47659468252218395,0.2728551074170219,
 0.5019724620837355,0.9827180023957836,0.19036186935600474,0.7029786594777494,
 0.3283840003865709,0.6895704618182432,0.3568414647633519,0.767078381304408,
 0.3230854850124434,0.9395592773999889,0.8095611879204102,0.02552148111004271,
 0.09920592431768338,0.23269086619395907,0.7652086548345628,
 0.4439066027490062,0.3968577960140235,0.7456806512005891,0.4447942734993986,
 0.6757546882597293,0.6305607098424725,0.3921754933392734,0.32777387641746114,
 0.3607000183244805,0.05347773856129978,0.9778450937651739,0.8773624009133778,
 0.08080463099777857,0.9767303097329638,0.5996642016664241,0.2671452079530411,
 0.06953520644913502,0.07026068257602891,0.12619785187200028,
 0.0024325035742887025,0.7624390158402593,0.19888407859398094,
 0.5536365315624767,0.4241331785021041,0.7416834058922426,0.06284493454359064,
 0.33164520481046533,0.6093703725839015,0.7309234686048727,0.957920165958565,
 0.6780009181078208,0.09068550491165173,0.5304047024786755,0.863505895078248,
 0.9376540407478875,0.9624820569410506,0.40788770610256775,0.154406573515523,
 0.8721212353516115,0.17179072391599104,0.1164834613640503,0.7510412976170293,
 0.6765844332629068,0.3102460733586525,0.9989105982042539,0.690415185253166,
 0.9055568392124846,0.9936750645751434,0.2960964953491403,0.326113687678105,
 0.83309195346788,0.8318969976040039,0.59043285883696,0.7725436533107225,
 0.5369725127143578,0.37811150704670293,0.16423297009602877,
 0.05835079828834999,0.5706959089525581,0.9289661173147327,0.9311387963487077,
 0.49152374851331326,0.4139086404528056,0.31372547995136246,
 0.5538106874769952,0.6047871340941535,0.09449383144480861,0.543641701331449,
 0.16925990725646056,0.9539983825248639,0.22375601492505826,
 0.5832130834233721,0.5369285361075494,0.5207662068705738,0.5198195059241189,
 0.9523510655884563,0.9136267421362115,0.6809557254274425,0.14118264813170622,
 0.10861095145349131,0.37457605561198304,0.733679729124237,0.7550340416874893,
 0.7359807794426213,0.049438070269018586,0.9661227388912356,
 0.5846564848678394,0.911028469013465,0.2325260430610664,0.45301090535915645,
 0.0014010911373377732,0.8850067376327505,0.3650284373861392,
 0.3225734851547478,0.5774581423484261,0.8103022587956146,0.6128715725679945,
 0.9954044221313265,0.1625680918984973,0.41662713253431904,0.4254088205345896,
 0.7633872123630183,0.7829742297868828,0.43178679107610995,0.653529284110092,
 0.16762044016147382,0.14887800009804497,0.13636699179222678,
 0.8209948892525334,0.9235200261012498,0.05911607793938389,0.6758385755545269,
 0.26563833099750844,0.10532628117135068,0.9087251987764201,
 0.6477504730658626,0.5178029400532945,0.2332845762346012,0.6422459015125017,
 0.34169854722632564,0.761213828509848,0.4156320048676577,0.6535769184373269,
 0.053669523311324685,0.5333920984347961,0.03987850565160944,
 0.7706151193758894,0.4253390118107374,0.9299852522733614,0.38684513592157965,
 0.0805972067948505,0.18228888594289627,0.5208552951319313,0.8822333510242442,
 0.03898197843462814,0.34625926624400505,0.6105365052622597,0.667480758595389,
 0.1348844355912434,0.8271947946781579,0.4935310729341861,0.8057526353260769,
 0.6858574932962338,0.915240633381351,0.7487178274583044,0.30434519763282253,
 0.8194826806722102,0.16351784856691265,0.5983448543831416,0.4779005428259425,
 0.6874131150923738,0.8107324610320618,0.022973976536212026,
 0.18128050556993047,0.4351079302641493,0.8098722995244576,
 0.09905793055580436,0.547173470503192,0.5597150755315619,0.4982085810160508,
 0.4408845883401701,0.43616539355944883,0.49335822774367555,
 0.08301923646902332,0.27828232911463047,0.5029699888874427,
 0.33478821365583367,0.01585274340334175,0.03735736577776583,
 0.7472937076085122,0.7519842753759954,0.5197877635944619,0.18452465500729487,
 0.5252074483669493,0.9176618637130205,0.98192420359298,0.5564450275547002,
 0.6518939064570408,0.5246892964711667,0.36352689722147113,0.6620206357197577,
 0.9197867245194556,0.41036290036956913,0.8101823224440566,
 0.17999623343004167,0.1472251333729383,0.6292403697181485,
 0.14727757071285574,0.5552120057477281,0.8487784914849583,0.7013998243009479,
 0.20741734379917298,0.8887082727158282,0.7115202737067815,0.8437728506516079,
 0.47370906018373315,0.7977061792079745,0.3246209119363568,
 0.18282851806069644,0.4551585961944261,0.6409166641379939,0.9342826458652412,
 0.15177190495057002,0.27477222358786113,0.8894278590850323,0.963514861595811,
 0.457291482514698,0.6280156627493889,0.4145536183499148,0.7442459873184135,
 0.29042869544699534,0.15601073861529502,0.16729094627879193,
 0.44117688565423874,0.09215471579177181,0.6769244133204021,
 0.3117337996623297,0.06006597735761998,0.2644620915753044,0.797675368770324,
 0.9837691286017769,0.07205439332555819,0.836948773173516,0.794986343929169,
 0.3716779313795264,0.2906832436150265,0.036596816301508794,
 0.29042145929107077,0.6243215044517105,0.23488794773856714,
 0.6150536176241443,0.2935755246521734,0.5183241805457451,0.4038010645620048,
 0.9843963142107968,0.6566785613587124,0.6984437325750803,0.42665608372997077,
 0.6947217547097204,0.41423535072429685,0.1195269292509985,0.4650130318894675,
 0.40205070903123974,0.5102980846627247,0.4205512952946179,
 0.24515541682680175,0.7347137867420885,0.2299820484408901,0.2727502545856242,
 0.6203133954155823,0.8320016942320766,0.9886903527950096,0.23864325903243322,
 0.14864378141073387,0.38527689090969,0.6811765868653908,0.9757596832771411,
 0.8320497415322764,0.0916502510665137,0.6373923817316799,0.04129338481322664,
 0.4392093672650036,0.12977162396252095,0.2848962374705267,0.3797333701723117,
 0.23335432125655098,0.19234179975203558,0.7104905773425847,
 0.3607387756146354,0.8152245561115503,0.3861645437231729,0.47051812334588883,
 0.43248850456730126,0.3861511712836764,0.9224770172386364,
 0.09627726060246367,0.7241821168038822,0.14454880426861505,
 0.02951902794868022,0.20762724452658143,0.8276299508837671,
 0.3434591407998182,0.7640037530518827,0.6075783163446092,0.12394350679162347,
 0.2877143573175305,0.5264181882549646,0.43549216406381563,0.7071935238595903,
 0.4409580380198166,0.5999561020873821,0.8039724171354536,0.6713061979202902,
 0.41776680549004797,0.07188091660546758,0.021869166462780587,
 0.11013071710347289,0.41112044867018094,0.13647955839087078,
 0.7463614933774184,0.46539096556315407,0.5488863169227726,0.5959938593516269,
 0.29043023123069267,0.18741686071155583,0.23323789156666141,
 0.09743482359532307,0.5290616929059115,0.3741228669408816,
 0.20837078329439607,0.4207093610965015,0.18296338324387307,
 0.15515950501490028,0.3441677007410908,0.16056915514301284,
 0.036332377754115464,0.7824537749391882,0.8971753167558657,
 0.2523845922913689,0.37021828699503256,0.1475550509670439,
 0.008929914104135461,0.6358685113692595,0.39189801499048715,
 0.3076242297992047,0.552508668157606,0.43087319008418556,0.16825564741974253,
 0.2879002813260806,0.3524577687462649,0.4851393471322225,0.4550546916087219,
 0.4227892471456175,0.7312899575767245,0.08555894189221136,
 0.18520849565169206,0.20576197521840056,0.5599965620704377,
 0.8448675289156609,0.1669169076378212,0.7849261609938488,
 0.025386861127447702,0.6737826499474665,0.8747976098283051,
 0.04281310997483778,0.8342864419154697,0.8202995695098007,0.3087979196184605,
 0.7385276984399963,0.44209977481453067,0.268477526924078,0.10104899330361161,
 0.13749984449342356,0.9212081267383412,0.15547065513577563,
 0.1258147397158328,0.9861284049556184,0.38842510590855583,0.3286579916429122,
 0.20772020273318115,0.6260810862950817,0.011932892454928234,
 0.12790775110091324,0.2732132551586236,0.9499712959782257,
 0.10455198607769867,0.6209059488791489,0.8127464095245951,0.5167049625887385,
 0.45206869683787676,0.1777796108118671,0.662884593233251,0.19540289605606453,
 0.1348816070833827,0.5488987052032704,0.680130410659773,0.0963933139294395,
 0.547655128677904,0.39812518268488706,0.620877284676219,0.9349151797015461,
 0.13420566602723905,0.8142301083381801,0.6434881163565468,0.7456619544958335,
 0.21490385482926666,0.6316422572609716,0.7476265355568726,0.3880664452664636,
 0.14446519545364267,0.27371294246009625,0.8637548438816748,
 0.9000970585939709,0.8759542299121591,0.9145631838122349,0.23960401662011965,
 0.5392030335432206,0.4043402949576492,0.07431387416645008,0.3834854344024965,
 0.15296395951235375,0.2183084928821778,0.7184997984617163,0.8274545613733498,
 0.162114995295921,0.5182833831551595,0.23279894709783466,0.5070520535684228,
 0.35258314768472165,0.05532564071648427,0.9518015061632722,
 0.9325638696530074,0.0769054762356195,0.42883244635986106,
 0.27045321906966424,0.35597841740795066,0.9916198545051329,
 0.9430316503306485,0.5628296141348826,0.07613691126912847,0.6300590670384745,
 0.11777936310774406,0.9969438346293369,0.6524031983509172,0.794957209507703,
 0.8315940142750657,0.2884200667961304,0.05586754784463899,0.4597318936468284,
 0.26351494389480123,0.18763028461842524,0.536194334110641,0.4649062831495192,
 0.4747503470029465,0.537748951722313,0.7069319060078841,0.5388959598858021,
 0.7483951935741554,0.8656321667475352,0.8282150850610428,0.5661874296450299,
 0.5195675353921992,0.4925888384869077,0.03913245385005071,0.3234451889949943,
 0.008834853441256119,0.7922784092562799,0.05068853658295347,
 0.897124361899193,0.6683388654372125,0.7243251250758431,0.9871998037329855,
 0.38872990456063417,0.8682924544487679,0.3408400079131185,0.7325906728546128,
 0.09790668497841681,0.9663865148461078,0.02100063537751029,
 0.17706784103527262,0.12411708317460823,0.19024537140010245,
 0.00908074290605354,0.35437927658790014,0.7450849380831301,
 0.2943747169189894,0.2809472134665939,0.8459791329533227,0.6270194099986953,
 0.0019671142515199636,0.4244110802933304,0.2186476817723204,
 0.3550284805423666,0.21546929766476064,0.14494526097517801,0.612089741851719,
 0.38389490535646265,0.42598355934418464,0.9189271470460671,
 0.9205884982655655,0.5596006292753047,0.188498975901923,0.9325818959865476,
 0.5679590406697028,0.9576149123973662,0.534696828341077,0.09957639936060447,
 0.8547388794861256,0.4374944538763126,0.1321248862073493,0.7014631877987552,
 0.7393278121416536,0.08605712071203753,0.6481836636601219,0.7232269947135368,
 0.17521847475013785,0.39219329748618303,0.7433059749679671,
 0.4838995311125205,0.0028582466166747755,0.34985739228071366,
 0.044772769700440085,0.6365458700479999,0.581848214986073,0.9708616733706038,
 0.5642984476440602,0.8999887384290306,0.6473919028223867,0.442818959494071,
 0.6346745713072023,0.719107338051336,0.19466115615261748,0.2948330265104351,
 0.03204521575364516,0.9910686288170563,0.8035256864424634,0.5948225967265629,
 0.3955272429926776,0.9629437977390358,0.26781626703882466,0.5486249407108426,
 0.26489380753185066,0.5467953845407095,0.18615491028072484,
 0.49430500873444017,0.833482917423267,0.18399067369786404,
 0.18275109864097383,0.2505656614715339,0.6650557205988507,0.4116862536977731,
 0.08478590859982638,0.05993637273456054,0.20390982332038965,
 0.8003336602745093,0.9888627428030006,0.6281383436004705,0.9263068191460173,
 0.09034924526362598,0.6043586012175106,0.9849905963348599,
 0.06577889594370201,0.9681579227935913,0.6474024893768182,0.7175021956805546,
 0.6681800848966116,0.9624639025876671,0.241248544696119,0.68343976955205,
 0.1168685528957254,0.960123052966978,0.7007679483325592,0.35380459866661895,
 0.3098887420583132,0.3389991838750812,0.7886976564457835,0.3306498852009706,
 0.7544722902740648,0.10583356803918675,0.8549164650703088,0.3547820143317908,
 0.4521599549427242,0.21320848928228475,0.06780885470099007,
 0.21260394847893738,0.3860085033930092,0.752922741081636,0.767888939430921,
 0.8410372196406276,0.011869738370313798,0.42258932320461284,
 0.7132832551579475,0.3465156261747079,0.3618414354105127,0.14500750886151936,
 0.42525749256279666,0.18438653963008234,0.4775503765921676,
 0.14635893568734437,0.7365835168167351,0.8583750318776457,0.8583545623282203,
 0.8879532256550211,0.2743967006361513,0.1751782484518446,0.2999550895546328,
 0.7754501582137463,0.9755483044784986,0.6053430894776191,0.9824082153366822,
 0.9526966311173944,0.16748314068882542,0.8112169294880783,
 0.11074422524712035,0.45438134925628837,0.5204662648113865,
 0.5135662062382866,0.7003543697435498,0.8400978600283071,0.16218226782561873,
 0.5153011506090099,0.8630828083960631,0.8820988685500672,0.9922901845593723,
 0.7320352471419582,0.1431775720398608,0.06288549150191047,
 0.21703057305880846,0.28293111220143374,0.04824934760894828,
 0.06699270414970848,0.5640888485788713,0.1573411628981216,0.2161511976553807,
 0.29691493176236083,0.697355933202678,0.8873730314025836,0.920632769595021,
 0.8265457189030081,0.9444748402243138,0.7800688877929357,0.7317537601708922,
 0.5793335729171616,0.8848723012748628,0.3909549472939283,0.9309792616586833,
 0.28990678359218847,0.2692229746521424,0.7156903361818865,
 0.021272013611414287,0.39743232120020744,0.17280720472080868,
 0.9904189651338491,0.026664963495512728,0.2336103974921515,
 0.4253962979430056,0.3343164424418018,0.3046558789732392,0.7508829907201977,
 0.455363469881089,0.9381767735789355,0.23316054440564582,0.785974169260519,
 0.6331174902655811,0.08645348927777308,0.9994934449084543,0.8672733717211422,
 0.4211973739817332,0.36269991169047067,0.7964806161082821,
 0.29287558448453044,0.3790276539060843,0.9310059425506572,0.649219147163115,
 0.6319040120644228,0.1711971589110246,0.2595690483237372,0.7989143330368847,
 0.59335939066328,0.9385718502722,0.23830676106226,0.9979417716886148,
 0.839110788388948,0.7181407764479796,0.6682461226911591,0.572351504372605,
 0.5881230776491722,0.39764424663835773,0.023267987071209806,
 0.961596990960504,0.19245936419814558,0.7125948841604238,0.6493087404825186,
 0.4414206795434854,0.4903875062689065,0.4930270079531489,0.5730758695824968,
 0.13229444595993256,0.7358110466000619,0.18270875299433453,0.754573488806463,
 0.18995837996636344,0.16915127215661463,0.7591456488219122,
 0.6655741387868221,0.40671018939887604,0.983408398024235,0.3754676227146738,
 0.15610883895290162,0.36149034622223375,0.3542271242146685,
 0.2788582443110904,0.3181668944322765,0.3382399182442474,0.2022806019356871,
 0.36573206137232406,0.28607806776386324,0.7031012636221926,
 0.8982892613276232,0.5971869170793044,0.9331807723228686,0.47246412320202014,
 0.9464415340873349,0.5447230808450382,0.6240864641723073,0.963580784699813,
 0.7589969361380581,0.922248465890011,0.3980909799823166,0.631632133617498,
 0.005557339123038241,0.5513294261622829,0.5449213053531372,0.956812454546691,
 0.34387309157970125,0.1363740076107386,0.9128421498739474,0.4041375055670883,
 0.08447202500232975,0.9397433773930612,0.9089587339553858,
 0.33608098353096794,0.17495603696069595,0.03241823760082285,
 0.3098114774419414,0.36940115883249747,0.8945037853982395,
 0.11240511787937679,0.5519586446140905,0.9779783644571343,0.8960071259736694,
 0.5165111341488391,0.29550205663071916,0.8782539496063878,0.8425246582016911,
 0.20932111525925123,0.5142597660938868,0.5270497998448304,0.2905390131631783,
 0.2602008342160612,0.3695562113833497,0.2987559949557328,0.7155996027889999,
 0.5264648660985639,0.15549252785081846,0.4163114679042489,
 0.04331181794848049,0.686573409764348,0.7605865753804663,0.14703438632605426,
 0.849476079201555,0.4002052590726646,0.47356236112206396,0.787933607607471,
 0.3819775856813842,0.7817168609402911,0.053634692192600975,
 0.8176505921382071,0.9212799287675437,0.8328777014952568,0.5857282488060402,
 0.6403040391133349,0.13177043044270742,0.5076649309032019,0.6480681506399304,
 0.9723069941949704,0.7932477515335147,0.7150954992431012,0.34565558535864627,
 0.25752072091341627,0.4153025581922417,0.32799837470700677,
 0.5862666020854141,0.6632024365436202,0.7617208918994456,0.40067310817139823,
 0.5970595633218879,0.580644026128585,0.29673463454371785,0.7264705547204982,
 0.7254225758801844,0.5654802900603735,0.6226537018425599,0.9876461801776739,
 0.9496371439437042,0.9260247033771121,0.3314627342544276,0.7973656031161527,
 0.7753191970873383,0.8392784314584517,0.23555783893490634,
 0.04268463160068481,0.6428169910761263,0.8682995132144811,
 0.10032899648635696,0.3946434348903901,0.27598716464810547,
 0.9388675266686108,0.24422578968330277,0.6993638552832734,
 0.0011618474161214642,0.4632032495672732,0.45425468246855283,
 0.7556503092382398,0.06481347619572142,0.9700536117623076,0.7345406271288302,
 0.5459596585517194,0.4783203106977305,0.8234379081042457,0.3547731032195013,
 0.8460214043176121,0.11818353374578106,0.7944461161275437,0.7663635159725831,
 0.8648952945094864,0.6902177151888994,0.7290113847346572,0.7695763029533084,
 0.22719266648757575,0.8416038645379624,0.9999866945218413,0.538596447701788,
 0.07749487313134729,0.8847935833783291,0.24816713923756084,
 0.5219170154472385,0.1137278134389077,0.5851912474072343,
 0.030458098455773075,0.7751398108345198,0.922903540947721,0.9594976099305955,
 0.1313556674346532,0.7441614708710067,0.31236511721081717,0.9227443926258379,
 0.6783031225927955,0.04959033301544591,0.9880530438407766,0.5878988993695113,
 0.926272899636276,0.9967373561957357,0.2144375385880115,0.6517030902356076,
 0.8775695731465533,0.4073323435090572,0.5737846131349977,0.44923705043509177,
 0.6564063623006542,0.6129118443802082,0.6084534944961489,0.3125235314298863,
 0.3833020620657728,0.7004784857829903,0.9372625445947321,0.34325512324496465,
 0.9356660992518381,0.714160768162348,0.4844870225866875,0.49803789520770014,
 0.6207701208111954,0.05848070809246875,0.5241259219057116,0.7836800876592679,
 0.5423774929801288,0.2179027311588917,0.4419316986091405,0.6746267166681975,
 0.01912858930916983,0.1694465565497334,0.14939839022599577,
 0.6058704164789411,0.42938728796538117,0.12937209433053343,
 0.3458455504374337,0.7335954052689705,0.5098812934468889,0.2815038911871567,
 0.05862106896300645,0.5318287437997951,0.0937522101456636,
 0.16887833806055075,0.9513875660281519,0.10671078438260784,
 0.44594754196895947,0.8239277249735317,0.8190762926807504,0.953380602136414,
 0.12292138197750335,0.04065696067167268,0.2836560840385196,
 0.20740896742256976,0.15218733973631948,0.38311244915595777,
 0.8920787600871989,0.3083815165216962,0.7517344739286348,0.8206769682664323,
 0.2910803630205635,0.714812568177002,0.44389260964820365,
 0.031766899230500134,0.28348735985000695,0.04445462363521524,
 0.13044111530835523,0.2582345778237862,0.9077853447707455,
 0.07354923426811633,0.3973262210033687,0.018917656669553806,
 0.2671213733375246,0.11940689421995765,0.1291777293398374,
 0.49515227201083567,0.6808352807733966,0.6917109918100025,0.5083541512098918,
 0.8953665373040806,0.5757913781575782,0.33705295389430123,
 0.46270497732186744,0.14047232194597392,0.4313522998557019,
 0.22022549602747155,0.16729088832070793,0.3954413869242456,
 0.7707676572554565,0.47851242620520296,0.0055025650657032354,
 0.5876958363604379,0.09730820933368,0.4236199628169435,0.14677824819122787,
 0.5870623563894708,0.07777602799612093,0.7283400431510827,0.5255267128369372,
 0.16721744446473197,0.02457110461098333,0.6104568847071355,
 0.7264231364645773,0.23614155312834473,0.6992254465432856,0.6244685805073715,
 0.7579378118336345,0.1282908759901632,0.6914134191984241,0.6664242652084795,
 0.3302061093698959,0.4257722714857133,0.2515426955494182,0.784178810836446,
 0.22459465751916452,0.6609410926858666,0.40081935566441007,
 0.09852345266674434,0.059672109010657004,0.21033168163982574,
 0.5376932371533936,0.40831902240605245,0.9517658360328032,0.6446592652146446,
 0.12015448406584239,0.5231315995044876,0.32444358138753127,
 0.4196678352004646,0.9545048730172079,0.09224129483644028,0.7056061764788026,
 0.35087192960755587,0.5471369531532477,0.10717849209085384,
 0.6531892325333435,0.5487816541386173,0.4872217700790389,0.11524897882023333,
 0.9314121074648274,0.25784928872767265,0.6700422577539235,0.7141005576255883,
 0.23397371844699777,0.9724721337238835,0.9320108540956125,0.6800573377848825,
 0.9437388193222501,0.3216765134626348,0.33073412958039705,
 0.18584055481120965,0.7675284128105128,0.7665719719267045,
 0.23399898310567013,0.485227804101501,0.021295686229479416,0.262453656587045,
 0.3664466407018512,0.9105509907748063,0.14168434185556045,0.8339857651379592,
 0.22057030271927036,0.1275384911462354,0.952247304521326,0.3258993080803788,
 0.44211002633702723,0.1687361344205529,0.4344820933572112,0.6924354671067564,
 0.9805155015028185,0.283038295580861,0.6908574389113238,0.4209951686413379,
 0.3986563043638558,0.2819727275003955,0.14605041665017138,0.4414192769914388,
 0.4295973007716364,0.2559608075028885,0.7065181816412108,0.6888207291717791,
 0.875734940746389,0.4716576172620811,0.24774413455848565,0.3737964415338302,
 0.29039016287551733,0.22148022304529635,0.9629731169318765,
 0.6713485528949983,0.6971040047239768,0.25013014472518835,0.5356721370571772,
 0.12271401369513935,0.7610350351909446,0.6669543321679259,0.5248527831859798,
 0.5368981698139784,0.37167422005518236,0.1995354462140515,0.6808180829912144,
 0.38587942943927545,0.22341844659780719,0.8305482801613622,
 0.7835504351744663,0.9133432954564152,0.891902024397141,0.737975383539844,
 0.26727201599560346,0.4301676975787847,0.8875386561782787,0.5049594376857646,
 0.21970962829388874,0.5204711411922356,0.6363032832870709,
 0.020636769978563452,0.15450666681253378,0.9508962468227117,
 0.4693853631746886,0.6980518290569584,0.8940136618747221,0.4591008465875751,
 0.9156807716289322,0.9051102905792938,0.6501439279842212,0.5643599298786172,
 0.2686943334470029,0.3913305503764888,0.06809744383683747,0.4276217067507263,
 0.2697907678593868,0.6043881636227909,0.3269103503678099,0.9935892511698768,
 0.40335338345660265,0.460389848906551,0.011954228821887014,
 0.5284657081639793,0.22126515128747193,0.15966885499667505,
 0.8269749612670108,0.09332784999587973,0.997057142637239,0.8726404151738983,
 0.10327759157255434,0.3620379515182046,0.21733476145229302,
 0.6232180787499818,0.7026999555228328,0.663839680341197,0.5181345313550386,
 0.03088511266477001,0.3045156391389744,0.35850368624893825,0.753006483505268,
 0.07140290149120865,0.31989227501908835,0.6652121273694871,
 0.009406064736135145,0.10804084929439739,0.9320119658131034,
 0.9303168627508402,0.8805395311565893,0.5218468614082485,0.9101868549219938,
 0.3889628689400517,0.9563186871296668,0.43127841148648116,0.308220195672954,
 0.5434372877824585,0.8808739273318765,0.979084686275264,0.49237929980056283,
 0.04814162516552145,0.4379614479549473,0.6420914679003804,0.9509798552814147,
 0.15225200443089215,0.2306859404296583,0.8125388515851104,0.8827141137949405,
 0.5816848853756935,0.30284158421281515,0.01984224088868969,
 0.02732489135370131,0.7601909589910436,0.6281150430973395,0.9742081064437103,
 0.5805734111693942,0.4509595575094586,0.3067563404650072,0.0944708572628139,
 0.48276605676012085,0.8392692472057696,0.6593828386330037,
 0.39823395381821347,0.8411495842390719,0.033179825704316546,
 0.1116170629870199,0.13726037408483271,0.6172641337475077,0.4734007053363233,
 0.7894704350723472,0.9588988546788649,0.9996550040143743,0.6521497933780233,
 0.16356823156876543,0.09752176866478524,0.8099249529620834,
 0.25238668199745085,0.4599998971682754,0.2886454382136564,0.6499266479073004,
 0.5988154393295939,0.29574827537336734,0.44850290146597405,
 0.37341819661041575,0.5723328046744599,0.2284092124531517,
 0.16178771760741184,0.3107797217736703,0.4730749232127843,
 0.39256114121842467,0.9516578399180401,0.35996551978637503,
 0.8473185836857102,0.38051265684672686,0.9964933044899918,0.9577594983007914,
 0.3194901786633344,0.12939343851206686,0.11026013443622074,
 0.17552408015850518,0.3244575743763073,0.7557231322278913,0.9995317769267349,
 0.9611289582409087,0.7879121891161313,0.025260206718271894,0.725212530953901,
 0.5852108665411938,0.7864068438824989,0.22452382508854007,0.8695956659553005,
 0.09970237772937385,0.4288060884401572,0.414867902373641,0.531149219525616,
 0.15473251165943758,0.7884272150684025,0.48715819666692783,0.440192275832885,
 0.028094544206292427,0.9053596266963143,0.7526539513690225,
 0.7953667236890954,0.7449303069331991,0.6196600429295235,0.42601344549268194,
 0.3559800331967733,0.7455573470355507,0.5812885369974495,0.7614092940400226,
 0.09537061581036288,0.9364376633867434,0.3559832009171623,0.880768382506531,
 0.01984926030721912,0.11073824829963586,0.7970778048497245,
 0.7792867560590695,0.34193082474920855,0.690206034947848,0.9606616055150071,
 0.7139774572364238,0.6148807631350788,0.44407330267470424,
 0.31006427594515296,0.31911816998325326,0.4873758900548334,
 0.7276304101688831,0.16216759948347648,0.39601090207465184,
 0.9118286351465531,0.665840181223221,0.9793292596264618,0.9039659948170782,
 0.9873513732038915,0.6097730789110021,0.7810075292227541,0.2722042328785712,
 0.5916919323014607,0.28888612687458937,0.5771688912563435,0.6012133102891033,
 0.3287986262667225,0.35993695793148783,0.0862051205428106,
 0.04499023920141787,0.8202799068869864,0.5004883447742083,0.7870210482542456,
 0.38878380952918545,0.7505530907867074,0.009379125810292654,
 0.5191133102246152,0.3039288969900218,0.9106602898363328,0.07249274432236652,
 0.9760783633966584,0.2082626956040864,0.014749592887840612,0.717869425397343,
 0.9363229805574482,0.21712318184866997,0.9277172560340347,0.9603894609456589,
 0.5931327758713383,0.6095999184546662,0.32038119410424315,0.6766494611969927,
 0.27712775150928826,0.6357131005500352,0.8977047470278382,0.6136443756420917,
 0.8768878898509572,0.43689393102042573,0.9093077688413511,0.6398410717798982,
 0.3784122349315059,0.6706654662520295,0.42229544934345464,0.9171851625872908,
 0.6838640139070066,0.018253822199678416,0.21869188106046267,
 0.4737520370600874,0.2580637407328087,0.38201178054755436,0.3360158370619575,
 0.8121504775199311,0.45324664266590775,0.5471631201622161,
 0.017574846942745248,0.4740712393263815,0.9583733545789574,
 0.7885650031078849,0.622235379524102,0.5769727075130985,0.5453197270586351,
 0.41507248258905827,0.8310579599682505,0.8309608921682534,0.2892734359134417,
 0.6235322842932177,0.35658144106309875,0.07457207019450363,
 0.5944033671599601,0.5925057825816307,0.09861777912752978,0.7235227395535746,
 0.7336569736843304,0.16936261424390386,0.4086691730626405,0.5657894982299844,
 0.8382567124632864,0.005874161135674849,0.8235220922044176,
 0.5012397295270542,0.41386561733200256,0.24436215107517056,0.37650904492835,
 0.5149218823594174,0.88320530191174,0.38895597507092006,0.6123379995812502,
 0.8627355677451731,0.37198922873612994,0.884874416509966,0.6861983053424295,
 0.5624939680905732,0.6343605888057537,0.44122253374968956,0.3993638986908231,
 0.7617626146195066,0.07632034555133715,0.6537040193633656,0.6291051526311238,
 0.6779738770074272,0.4813851306571486,0.2271703174731462,0.39344423397276396,
 0.7020449184764039,0.7395351765913243,0.5144171291883949,0.8998367926732476,
 0.8422436509120572,0.3782222904249726,0.21691293578581416,0.438210757170734,
 0.3403037641900245,0.7316553732785466,0.8007711054062416,0.08777940379625693,
 0.32473662912615564,0.31349161255065194,0.3216496088172479,
 0.08184116215963388,0.5391250763273047,0.9436370446819107,0.7915469431970595,
 0.5305837033077505,0.6835125698239618,0.2581897024984948,0.3437575970658997,
 0.8834586454737592,0.7826076588716239,0.9846229782308292,0.9856626668118544,
 0.10811394916244843,0.5482543531977974,0.34065893239320144,
 0.2912613268352593,0.5252280349054378,0.8222840687617303,0.6814019861085577,
 0.5610215556456262,0.9196709976540666,0.08027356082340809,0.8713391296577115,
 0.36235621989701516,0.8917920919178225,0.8829447965901325,0.6484417103091603,
 0.7294943220048201,0.15932650543071603,0.8105054220337728,0.6642289821976175,
 0.5599559833342036,0.20400621336916558,0.9951144825685223,0.5901951893441004,
 0.5810865545587023,0.9001127396861784,0.23771674534844123,0.394292125513794,
 0.43041083278997616,0.0720768757106971,0.9440300414535227,0.3984984539366404,
 0.45558979125586885,0.2728985570557859,0.8063032761778477,0.6147376154796188,
 0.4134605338674018,0.6475829915636921,0.29382200324308116,0.9606157430159247,
 0.06006366683263076,0.6056653297878506,0.6389914444533732,0.6174065000998908,
 0.06682852467333245,0.6899851155674191,0.015995368112069586,
 0.042730602380477656,0.3491480184663974,0.42796195250545854,
 0.5810938997164293,0.07278833200701962,0.45364369948932637,
 0.43184874616226654,0.5516183904025418,0.013677515367586324,
 0.9973537641301566,0.7728677826505189,0.25390855476520713,0.8814126455986908,
 0.4356407175663335,0.8927192988830445,0.08128391140742997,0.6787910108679106,
 0.620318798365984,0.3538707735899548,0.33025754697766874,0.09072279067005506,
 0.9150843477023157,0.23823227919340162,0.47352527007027534,
 0.7141073493715374,0.5926302092727559,0.7737163343853211,0.4563090457912935,
 0.4370710829463439,0.6982878366339753,0.4790027169761637,0.5082316045662093,
 0.6553259924444023,0.6179109670159955,0.1884397616180844,0.31666777234611954,
 0.8570192928754975,0.4159166441881701,0.6381022865130787,0.8145493705065715,
 0.3225819321706167,0.47940886867239907,0.8393621551750947,0.4750985270332839,
 0.3266567074473119,0.8465370531111316,0.20365469075272669,0.3027067139812898,
 0.3864121604456715,0.9660963154270754,0.8385430969999764,0.6772536949896865,
 0.40436856958205913,0.7428531663415701,0.0053556965728973616,
 0.23956863684071683,0.42791887757014446,0.8337440939147527,
 0.8892275692617209,0.8137719616761735,0.5861908520280257,0.18665258426212172,
 0.8820243470054945,0.11971607186993505,0.2515933691022316,0.7498663276420521,
 0.013602371562244087,0.15011030699935501,0.032293835917869496,
 0.8990790384107238,0.3098417670087228,0.2383183790916681,0.22489391393259495,
 0.259717698669917,0.7267884728951358,0.5622097576462658,0.876246156430783,
 0.7478107152793972,0.6348022954551534,0.9182703902549548,0.8118200791136764,
 0.3389389063917372,0.8584929550302685,0.22913551873392402,
 0.35642163097639223,0.509777784780564,0.9622468403162419,0.41584995220464493,
 0.9506850098873904,0.3262046414553731,0.18479657300123686,
 0.11313982183328608,0.8280904532927086,0.6477055530560099,0.5657202204744447,
 0.8240063799777425,0.7705635253315944,0.35352584532459475,0.8659552195181917,
 0.5425376375093227,0.8076719300015982,0.8335048878320621,0.266591436235573,
 0.26297758609168875,0.05541489422514689,0.7979383514919943,
 0.08021779012166586,0.6981832612735106,0.08586900406775189,0.506281186806692,
 0.052999920980325244,0.3793097194202055,0.782248397414468,0.7808365497603346,
 0.4975723922304335,0.9931864908628721,0.7762678007672605,0.03239375994179361,
 0.06179311706610757,0.8480380710518551,0.6878117884050169,0.1704520098732203,
 0.08751095195336811,0.2151397892682002,0.9180429230621115,
 0.04935233246959847,0.3572181154095928,0.9084632926219944,0.9209178272006515,
 0.6684882169056404,0.5979850272622341,0.7870281753569364,0.5734978127110577,
 0.67142448114091,0.331754661453455,0.05978256408169136,0.3013021673055698,
 0.5335481409287172,0.8708740973606051,0.9010498657369752,0.9255648790783739,
 0.960059974614911,0.3865499272194347,0.5399388678754209,0.3934897463280469,
 0.22864757490372756,0.7595132416877999,0.04300355056266514,0.850352509895883,
 0.8439904470190208,0.06854544529186701,0.2961271571299373,0.4578395507689894,
 0.4300833214484221,0.846761785561778,0.9398038249365074,0.18848808814107854,
 0.9814583068573232,0.28536483657895406,0.8008817132874971,0.9959653681731858,
 0.15186642511333615,0.07776607193136997,0.7394007258026466,
 0.8805516525526038,0.5757905501153329,0.39405553830905915,0.2288627209759584,
 0.10123475989118358,0.0925876593673407,0.05262519245280506,
 0.6502454126805346,0.5582236498757567,0.9589920959014353,0.6101824859319883,
 0.1009987001099728,0.8617123273022731,0.1687539822012104,
 0.011600978303972997,0.34445518091956906,0.2612797717377283,
 0.049176806384792604,0.29453934095842316,0.5025261386270556,
 0.03024493612587098,0.2256284112052429,0.24203757790842018,
 0.6595989236997521,0.06358399724451635,0.5039179779209809,
 0.49542110351981994,0.06136885983240703,0.026516079176918117,
 0.16815947945765197,0.16094576220926837,0.5891305647352045,
 0.03673264599071013,0.1401238347054592,0.583064089031962,0.9835620576875073,
 0.6036357141994207,0.30279907643927917,0.9228953075206148,
 0.07330972091343346,0.6480331793463586,0.3025687773546937,
 0.46089433161334414,0.17786225049615623,0.6101910046905017,
 0.3662968613540222,0.7047479936059822,0.07743783914218849,0.5808430980230138,
 0.9905079470288982,0.6090289391573204,0.7164680694246213,0.3480588823381865,
 0.053900232443982565,0.925948446166326,0.016855835700683053,
 0.7766823188926307,0.12385249890093164,0.5306764095088361,0.6249570530277903,
 0.906956989558579,0.9431435367601475,0.16803729571229942,0.4856071745738765,
 0.2643114344372084,0.10784282447949167,0.2988886800755437,
 0.11264564473332739,0.6906294194625995,0.92660006094843,0.27994600049762486,
 0.7337879448482173,0.5140092924526748,0.7447501044818172,0.14443919771201008,
 0.7328470560525548,0.20253464872357219,0.3320737080565872,0.4955080024342029,
 0.8377196938163367,0.5428775250267485,0.5853005968434759,
 0.015169661528540268,0.6092688966133855,0.38457748324463914,
 0.3626977789342567,0.942743401112482,0.6805132286364357,0.3400883189894993,
 0.9743451844860368,0.6899454703759508,0.8079494249109119,0.3746785445871824,
 0.231349585990538,0.48588472574844344,0.9820818634715855,0.7554977017518427,
 0.5594463822989564,0.1458449243822142,0.8387791666813817,0.1697919026292808,
 0.37292558280820587,0.6148515208575758,0.0017569209697786414,
 0.8764946184965281,0.11690654347960128,0.03822434285557774,
 0.9572066229733868,0.7419315243643748,0.5569105764279696,0.5161396835215392,
 0.2856995566112899,0.5086903119397028,0.6920866625288833,0.9749096315426002,
 0.6625966341790367,0.3000009567490658,0.4301600546249269,0.33290911671807355,
 0.22714497355905372,0.7626548451733465,0.358563185280002,0.31485389122519625,
 0.33024503757559065,0.7174351498827273,0.5262073457289179,0.3198288824865143,
 0.6155372252404403,0.09886509705306712,0.7285117959185218,0.7578230502802439,
 0.9208675879505549,0.2662904099373138,0.5847784717981316,0.8171588343346877,
 0.017436818890331107,0.08128484943805747,0.28836083660716627,
 0.028727771550324244,0.9118786351744493,0.722785830502998,0.7159820228825903,
 0.4745880064306046,0.6290435332470014,0.070732028777174,0.9045195141687283,
 0.5647988185945614,0.14877346563034266,0.11704647515822086,
 0.48603611018981385,0.7994171079071166,0.033153122254973355,
 0.9793637938485464,0.21314307562025847,0.6647502796589473,
 0.49590802015699276,0.5978630601051359,0.4141100692946391,
 0.18192490319122134,0.9357203410747124,0.412976851988043,0.3430087261165268,
 0.9480835912336782,0.7240689120955934,0.711200502882517,0.6846399295333423,
 0.0011242497873220625,0.7704218505799166,0.47890131055142715,
 0.37876087113688117,0.5231694605907361,0.5992931146123109,
 0.09263755835983423,0.7853575806314609,0.7693905847775151,0.8483703966086775,
 0.10100068215386115,0.36547562268977096,0.5634910597004048,
 0.21220608657068307,0.36858144239629664,0.7779885029874994,
 0.7068661858993619,0.5562312858708052,0.7366723719241499,0.23290445334107113,
 0.9443394706882402,0.24253625796514378,0.121579651301713,0.8556285662667439,
 0.8664304071696084,0.11660559972814244,0.5550310959447403,
 0.38631159063665954,0.7413057899576989,0.914380555317026,
 0.008556377630261625,0.3872143708248227,0.8180266668222645,
 0.4756441981788144,0.2839534752878199,0.6666618577321797,0.44583173366792705,
 0.9768107172363347,0.7185604752068017,0.4620179295987804,0.895208935881729,
 0.957935956068126,0.6886067066901556,0.5568821920158754,0.08690373809104912,
 0.25593977776920485,0.6077733253916173,0.6860336672210847,0.9620428960158209,
 0.8083966460372102,0.2676358495951602,0.35908937127920926,
 0.014709503328379836,0.30807742347320255,0.44817164094681494,
 0.6681097419871206,0.622958908350207,0.5112427063747307,0.709825036315512,
 0.21669992359874612,0.7169956776419723,0.6787758519364495,
 0.18019502413586508,0.10806161727548957,0.32214791448057645,
 0.19554271347625962,0.007094775488357796,0.4514947591235484,
 0.15227231454257784,0.10342500730798088,0.012181370256154733,
 0.5014024544626952,0.7920371801280265,0.8847232997227914,0.730812738949009,
 0.5464274083613946,0.9166889882504918,0.3393667741442955,0.9285882729019548,
 0.2314216133382907,0.4958804918623786,0.5602992173774974,0.2599646496726201,
 0.43156260336875696,0.36434733598300395,0.8141011909095218,
 0.8716247062921312,0.197240158660119,0.8358793452706115,0.6674269569934502,
 0.5504078413273686,0.10051667470620307,0.4693667174637969,0.1557529451312073,
 0.2542874547124663,0.9661729443288466,0.4475395277321593,0.42766125636719776,
 0.7971060836342112,0.17663078546133038,0.0582232339657196,0.6776834281121138,
 0.9553336777335258,0.6090840891550131,0.4526619888465384,0.23910033404956188,
 0.3864085392354316,0.809307015124407,0.8034744536147428,0.8845379111406604,
 0.794258055138726,0.7301690052869269,0.9175719818036236,0.6478590478211425,
 0.524665046177975,0.5347340053196706,0.6596664085474682,0.6991033620818979,
 0.0993930894226045,0.8221001840680444,0.42335350369157587,0.2997102335384063,
 0.617459842027655,0.2990458511039852,0.9059425724617339,0.013440504974762013,
 0.608326057242568,0.007601903316752701,0.11327740987655721,
 0.1335051434943706,0.8207956572718775,0.5264205245622763,0.7957267948405151,
 0.38769215852440775,0.9197122302787867,0.5883205851405285,
 0.11753633611560621,0.5842152531419836,0.5438487601259107,0.2481608587991755,
 0.6152729866128149,0.3175803684119388,0.7855016414851147,0.4880751540511361,
 0.7131642242671175,0.27777037171691854,0.44583426889692235,
 0.5361130173913423,0.6047799432796269,0.8602294896387841,0.5896824236993533,
 0.9765724303977398,0.6200163462090983,0.6017433945510686,0.3919430534708104,
 0.8333035129814588,0.44542623431667505,0.9775251162999752,
 0.22793326777797784,0.2385506774741799,0.12264017327626298,0.925727631633697,
 0.8321603983378787,0.8885383220585789,0.2636051709949415,0.1886169907384354,
 0.0892864726116126,0.8589532739257864,0.04789015386529949,0.8120176752518562,
 0.9566625165266315,0.1710423245800109,0.0062261331190022196,
 0.10993297336420227,0.3438165045883589,0.5050321238675459,0.8860070617788381,
 0.9418296070432813,0.21552827420881038,0.035273214370790895,
 0.406325924214639,0.574223468581464,0.9946923934532435,0.4061439826052483,
 0.31163912291826745,0.405085965367427,0.6388096556869096,0.7231301469906538,
 0.21801899453421203,0.9884799315644914,0.052065147072172246,
 0.5161687938013517,0.8463105018786297,0.06884494957636322,0.1664168447919374,
 0.1768086784423536,0.5962954423072409,0.05443323398983424,0.4356643869230029,
 0.9670447464179452,0.5093762052476842,0.9420574105870972,0.08637874252698952,
 0.4492206783928133,0.52472677775607,0.9916382391168571,0.8803629891012547,
 0.016634607793539313,0.5352621290860732,0.5344394622878481,
 0.8410506664943482,0.025336063331765635,0.20329095541160147,
 0.6564121883192764,0.7074154305895732,0.48761744902129256,
 0.04656087750144666,0.9751799833299064,0.03992029212166748,
 0.9270628624864832,0.8784664836580749,0.0838239064047781,0.534576359946693,
 0.5248841342122791,0.3596558679313966,0.01954473703385151,
 0.36672608953769337,0.8815648323626571,0.7341654692479652,0.6251230108131892,
 0.31447519327313,0.7214799622509753,0.7155628344010871,0.8660398426725393,
 0.4490489671352743,0.12142703671928545,0.09967977472957557,
 0.29098653123912166,0.5711123502019048,0.4915582266817904,0.3612105918476237,
 0.08544402497015802,0.359528770176755,0.2679758321707656,0.2785478668042807,
 0.990336216981919,0.19198677703258094,0.662899125610895,0.511681690811725,
 0.11526813471147035,0.17679642927118078,0.8886558679090689,
 0.6827161574333832,0.9560527661504173,0.6596612714222684,0.24081471450938186,
 0.44009906559390255,0.9290907769232465,0.5050735110626398,
 0.16321471049498248,0.7191931543683825,0.7723203603179564,0.6374388652466296,
 0.8346782570145991,0.8472030516947436,0.4816624761219256,0.8653592945911247,
 0.7715473018432246,0.07726835164513923,0.032393795790397584,
 0.5194863521914108,0.7789189348951885,0.16641322647644208,0.9837469030921329,
 0.2715620774167792,0.43458082325261715,0.9841033885030571,
 0.48588909212938225,0.029714100423592904,0.2601726555159144,
 0.06299953766970157,0.12996916128067282,0.5385357643459122,
 0.7733916476420897,0.18430158313318534,0.9236214345131911,
 0.14645730738515894,0.4989129303574489,0.5000485444203393,0.754919191860539,
 0.3254083100516496,0.1786187399429653,0.7891561835000229,0.03955426695516606,
 0.28812156933126243,0.14743853718665245,0.1913154944959815,
 0.22141570925869192,0.9984378942495508,0.5507800189037064,
 0.31373280355703925,0.38334234224689756,0.4852909900071315,
 0.8491494090058014,0.396302108292633,0.23080510592552872,0.6709746226581882,
 0.1197921022111621,0.7103545033998544,0.999029700624452,0.2336568226492841,
 0.7200582402013269,0.9216748700498684,0.34345069158474284,
 0.16094038628292429,0.500049924942089,0.5103627881763759,0.7685294604908606,
 0.4964564274647192,0.3511785861111638,0.9560633397894579,0.10567882155823216,
 0.22138202882095515,0.3641790304924948,0.40629456900781513,
 0.08050689246451126,0.29151209391077026,0.5076932105576982,
 0.1187583359022315,0.47266679773414666,0.9187286361191833,
 0.19600623122200123,0.14772429848244706,0.3953219522379703,0.970932787446618,
 0.7909248454253948,0.924102722162498,0.29695124113696914,0.21036405411100445,
 0.4612117608998504,0.19619228533225364,0.9008966516220897,0.751289578023687,
 0.1963267090514237,0.9033871158075235,0.8471352493300802,0.17301048252372309,
 0.4110076909145254,0.8270417803195111,0.7642387160210212,0.2774973242703336,
 0.5574925996507192,0.931857354241455,0.17717934236576782,0.8091778930771121,
 0.6332504099500392,0.2828249717749163,0.26877512146796967,0.8884391369706287,
 0.3967793163879664,0.2303241731569552,0.26908184224760967,
 0.041013231439869346,0.7601639437740516,0.5600247588859336,
 0.18583900211484594,0.7351753900894193,0.7676700778247444,0.7815164213076776,
 0.7124960700967172,0.8656167270513234,0.4681233137295634,0.23282766556968149,
 0.35420012908523135,0.24267678693438555,0.08911081445448576,
 0.4036858657737231,0.18047303444334029,0.8243417647193069,
 0.18297934016928974,0.9399754947816661,0.827434830003827,0.48525966631630624,
 0.41257937222745766,0.2060332028677827,0.4308049155818443,
 0.39977750640683873,0.3189641767062489,0.16531392378466048,
 0.5863947246175297,0.5366550546800077,0.12055049531855955,
 0.19318961202230256,0.025771967532701057,0.25365168886970524,
 0.8193435711641293,0.5536957818618029,0.6600457948621602,0.14566703311975948,
 0.552515973061621,0.8745733300262551,0.39793351798493193,0.5329954121007409,
 0.13768093272028148,0.22354585945473537,0.7221579726050029,
 0.13796204346117824,0.21336774290092997,0.2682279444252289,
 0.6183744328591012,0.19803794734678515,0.15047048369017768,
 0.7662098506816342,0.3464599161148121,0.12308715172385132,0.7539315154655076,
 0.4772783096095953,0.48902100895660705,0.018788659815033304,
 0.5538341565969429,0.0077940296881325866,0.5672975268372661,
 0.7408813257403531,0.9788001107540232,0.908820040143682,0.2863799417581371,
 0.20856412461877694,0.6331373105147754,0.4024586220583646,
 0.21879613810432408,0.9892459757727774,0.3363240233873487,
 0.43997406536137185,0.8887024289080078,0.7246198412099957,
 0.21815994094411362,0.5930206803615106,0.11503941072417267,0.976480144760035,
 0.3843890741005084,0.13672068228137013,0.39632905315265365,
 0.05061554736528384,0.8469200206930131,0.30326465212074993,
 0.4176997132492859,0.33160591667002626,0.4278767115769685,
 0.019005459467088315,0.23580802815994661,0.1441586156759086,
 0.08073648053024807,0.17048248439846914,0.7595001070359468,
 0.01704445585310055,0.9711108281025949,0.4348617778590059,
 0.18371791507188884,0.21114574598210945,0.8823000678425263,
 0.49640072835694304,0.7624904637347742,0.7902436251590269,
 0.48637114510216883,0.7520785301568034,0.3182014008610109,0.8146930917850955,
 0.34589789770633006,0.20019121312169452,0.5043179958290012,
 0.25974578670616544,0.40476508857564863,0.677947886010918,0.5468684274253255,
 0.2062622525713772,0.3033038502840524,0.0753281766793249,0.07844452162258875,
 0.4555217164326363,0.9136366310699436,0.10658383313671826,0.5481969826913351,
 0.7835241122634466,0.9935401994320308,0.430612755760567,0.5529497121325826,
 0.04950824484446348,0.42681186142855987,0.7823943822999351,0.463004115704193,
 0.0742463869620199,0.19379173872382638,0.4624197037855554,0.4566025160746947,
 0.8406656208653532,0.8992092236199721,0.6311335130759831,0.47820632443739064,
 0.5621705460259447,0.8385825247240171,0.12908167907272616,0.2942685595180634,
 0.8843241070227207,0.9031290841268967,0.7298663309555281,0.09475095586572713,
 0.04957216085682581,0.26444090719874624,0.26024696004309267,
 0.0456494395766287,0.7325348833582189,0.8648590388021846,0.5096615113556474,
 0.4573268733019674,0.6904488458009763,0.27765987995576624,
 0.11986331516485582,0.6798761917913659,0.2199030107498522,0.9551935523040948,
 0.8675752551259233,0.2551261341578257,0.6900340221540302,
 0.029711012653462765,0.5763562963354083,0.4372465805261887,
 0.43466788589038197,0.22543563087000407,0.0569701975831024,
 0.2814333673664072,0.30416103354851154,0.5464784569325803,0.7466746633023167,
 0.30433064863429893,0.1060795176491679,0.5092614591782273,0.7028049609781938,
 0.1567458536759576,0.5102510386122225,0.22796669311737272,0.7571883357888485,
 0.8470025429860364,0.9077448606178841,0.9111693417512916,0.23414767579674245,
 0.35578544152856106,0.8176968823000779,0.14209109812015874,
 0.27458426233279254,0.7015460483677527,0.5622991247826352,
 0.38129729167705584,0.22639924453681437,0.8812143408446794,
 0.9583934308402504,0.4750550038368162,0.9856779741778845,0.6096320462371889,
 0.8598168659373583,0.29879797268752517,0.06249144053156519,
 0.43191434473390855,0.9131594284776615,0.9298766216370424,
 0.14835271201300526,0.7169131360739822,0.507684586314623,0.530932186489775,
 0.8277937964104212,0.5828562868706608,0.9212345739109153,0.294559643591331,
 0.7472391622900125,0.050349566820990455,0.3451316714077688,
 0.6262356339028042,0.5924192526627712,0.49555224745455473,
 2.2882245327116024e-4,0.5399272934096815,0.6195287404634995,
 0.7673589776493605,0.9957207303662129,0.3643294603786653,0.5472555854183421,
 0.9930395883658035,0.5542303338396227,0.3612954362398859,
 0.023241091145885906,0.4849472364362104,0.7432013799427473,
 0.36175011235040844,0.07158560413541848,0.685886733340533,0.996458536340507,
 0.1365850790187323,0.7839917387963857,0.7150853455874466,0.7421122025981691,
 0.6679656662844897,0.1742479060004175,0.1221591552075133,0.8081675002531785,
 0.5590921241165498,0.2819932039630191,0.3392508731875927,0.279665944722733,
 0.755063793877944,0.337645606344429,0.8829075072512592,0.5385314512606448,
 0.6415710511817562,0.39168345511126446,0.6853602555571743,0.7615563824519944,
 0.18603970840726014,0.387082401579269,0.005000396041228594,
 0.6479938030650688,0.016709845087004904,0.42751136648846555,
 0.08385878728830476,0.09162232054589037,0.8940480414614049,
 0.6406093273756867,0.7645657406892223,0.2501129128721611,0.6367784519057524,
 0.8343525980452043,0.0997163318722688,0.7554495125568245,0.7827316514794608,
 0.7382057407915155,0.12440654105679239,0.21805036519190812,
 0.4335366557927778,0.2261892850241849,0.6805059910551088,0.7929209206368291,
 0.6568281580775854,0.10569954520012548,0.6193880561587356,0.7637318108381519,
 0.4578325214376706,0.5311255155519726,0.9828368779300585,
 0.028247668608726406,0.9321308738800705,0.028302238907528254,
 0.17859209091557848,0.2849796109715155,0.522740770367692,0.6932541308497832,
 0.4652357962812639,0.7889157218684815,0.383675673382458,0.8692170240920323,
 0.3936654731783015,0.205007404497625,0.7507670854968438,0.8125931845459577,
 0.550033128623686,0.6946102169119681,0.9441858656425146,0.6914279819342288,
 0.757175156551195,0.4657601184535032,0.03711864599887604,0.7339860886804694,
 0.12123283902575599,0.5485293693812505,0.07625651146116519,
 0.6008436833171932,0.8639897086004003,0.2982228059523596,0.24858478982033672,
 0.7185764251014161,0.24264300545102224,0.4099939600422674,0.9053597154125709,
 0.08314190928545628,0.2328984823672705,0.29224198315452754,
 0.8380347186395198,0.4565833266316033,0.06699773987667379,0.3366823531274168,
 0.7647025443804947,0.2659119317386963,0.5581227341047229,0.6621691165866278,
 0.8594125753937876,0.34095812937621905,0.14779684863955955,0.92237418807147,
 0.8827395678686073,0.2664688438712147,0.23671614820155196,0.3441735411884289,
 0.4024242467076967,0.5246067883256769,0.4455834925956016,0.30161332932350937,
 0.0651126622161542,0.7950553992605233,0.07898482194638345,0.2554476378533409,
 0.3681611075447315,0.23489855467152698,0.5640477120779638,0.5901973410917745,
 0.9027238077486077,0.22875901953472555,0.7903905147333434,0.9575094593863901,
 0.39450225159046204,0.5192036047849696,0.8378623470768914,0.9603573270228085,
 0.9553721741709782,0.9215033691341523,0.7439288423665349,0.44834207059332876,
 0.7276049283325142,0.06983966569248701,0.6131282257358613,0.9774044540018286,
 0.31997500191504447,0.12359158914271673,0.9907213272380145,
 0.6873448842000869,0.023427604301531346,0.48389499325750096,
 0.8161191319578383,0.631195156637959,0.7468090394265473,0.8914794011343415,
 0.699857111388163,0.15031085735538818,0.8822063489402348,0.2664974185799911,
 0.8461586271391666,0.535811283514745,0.1509524148330248,0.4928456984298757,
 0.6686086390103209,0.6439887911312779,0.7725093685432661,0.7630825746585171,
 0.7197281753195023,0.4859339435296677,0.7165019527411859,0.16523268718721984,
 0.8295914454094327,0.3774643489283295,0.02889057528130745,0.5409627908957851,
 0.3511430667906621,0.17939601863023047,0.8842566576886524,0.7559478682002687,
 0.019777531129048453,0.7047392653214866,0.5618141613839659,
 0.31620402892079513,0.11023140975538182,0.6599759423953222,
 0.27269647756013027,0.5703361693802321,0.5799598832915482,0.3253689717243976,
 0.6365603094383077,0.11278453066660943,0.9490362643279833,0.9459679147959851,
 0.6993600431913585,0.7106088980342098,0.9187019086327483,0.49725104908783857,
 0.7134310070635466,0.3212935289336306,0.8771192664815672,0.10050496772829298,
 0.8754242256975039,0.934841137323375,0.4892814541873619,0.27393751265808597,
 0.8885146836218888,0.9975117016653948,0.2961370659729722,0.8666336270104958,
 0.9406516957792391,0.564523881872832,0.01792254965467044,0.5803945981241915,
 0.7211875562605705,0.1506986877852442,0.5515367075121816,0.6399307240500892,
 0.0620670613942893,0.5314492944078636,0.8501917994012386,0.5048961030643081,
 0.7673892620388933,0.36636802372922717,0.02463609164602454,
 0.18541667980654342,0.6471277612895697,0.383709397368858,0.08761758289530897,
 0.37671050123604743,0.8079373334937981,0.3047421195030424,0.6667108422175596,
 0.897361563219907,0.9385647016432831,0.751951464904482,0.023551310209287357,
 0.956123981673594,0.11786515035119383,0.024387421641980045,
 0.3277384784941394,0.24870519632956245,0.23840498822739942,0.589651335221716,
 0.36521186204321754,0.9933824884291663,0.15018599339408423,
 0.046753875365458386,0.4682175209020212,0.18235160120554816,
 0.051804864134426754,0.011353388250979646,0.31970411010526567,
 0.15444164452993925,0.8190956674942266,0.4662839183545686,0.5535491176495935,
 0.28519025355568883,0.12403589051188979,0.7766059142479004,
 0.6707661610250911,0.13531134165644043,0.6608324169370742,0.5069152957606531,
 0.70116112362532,0.7976679497343564,0.017385386902761812,0.45755538758736725,
 0.08355867776363102,0.055625519130344525,0.795406652128722,
 0.4464366826228642,0.9962381849830697,0.8715971011866483,0.5369254593386297,
 0.14014981904643753,0.5082188751373278,0.8514906464168938,0.5215444516558276,
 0.0019385642988597462,0.40934726176861835,0.5394353788076901,
 0.3811373497275464,0.4787102625712548,0.9050087948576917,0.2976951424271814,
 0.18587649095137648,0.0035441275970147235,0.6753403938476499,
 0.24629930271566547,0.13103008769869573,0.05669179851155548,
 0.874971316223155,0.7922758198079549,0.5243954314396587,0.6928782397956043,
 0.49204825190158374,0.10347147126041267,0.2894756040858343,
 0.4689643111295658,0.656571040873787,0.21438595112265169,0.17164624315275934,
 0.21121946731628205,0.566914228067441,0.4537672181653435,
 0.006621753269933528,0.5854595649900549,0.9671661393914464,
 0.0687365564322967,0.9746707599355757,0.26517343282993733,0.6930185176460558,
 0.7734417821082937,0.8134444163861411,0.14774264352059618,
 0.28851604331575853,0.5806195193431775,0.34594279557009533,
 0.1787384266109051,0.6497622590049326,0.26616474654028976,0.9845628026884243,
 0.6990989730252852,0.809998360885782,0.4784803829198817,0.8881145899668978,
 0.46541379703637176,0.4840828601985959,0.34230485334059424,
 0.35469327132530326,0.692284044505019,0.583088503327594,0.3621528164758303,
 0.8917389372716229,0.35171982248389333,0.30974322593777315,
 0.9213851090353535,0.3840681883978465,0.10239764773643678,
 0.07685716011888755,0.5068096844382032,0.5077896837021766,0.9046343563000239,
 0.18119832422187454,0.5209286064341446,0.8334703962297872,0.7818623168315232,
 0.9058382161614831,0.6504468718274358,0.48647627184339404,0.7994969105925074,
 0.47467668690789633,0.2542202112737387,0.4588233560815599,0.8207325858416321,
 0.8885269475055829,0.3676336505468697,0.15421588191175317,0.1534563403928867,
 0.0035504405337833944,0.3345033291502071,0.07434198185884233,
 0.701954212743998,0.9683904336961565,0.21071162213118133,
 0.032276815225294686,0.6478184957082981,0.2942853062856732,
 0.6418923478154692,0.8864783828339239,0.07446387721400238,
 0.15743469828949974,0.09162491109741555,0.3191127192982379,
 0.01536310737701363,0.3685974980219858,0.7147138362161698,0.8412294933416931,
 0.7752155157300584,0.9322749089641401,0.08119538004731319,0.9691496697874947,
 0.3594288390057748,0.2302629757125073,0.6966067937338203,0.5036448426735556,
 0.7156376569194818,0.7283510134520286,0.8768187078564984,0.9312685545202057,
 0.9973801049875295,0.1538806323377192,0.993855212647548,0.30811560748740807,
 0.9838900475419288,0.18314939403552621,0.050152898981454475,
 0.4403987184891067,0.8191907059992309,0.44006793240078057,0.6076317209477047,
 0.39191203755514403,0.35401857116718893,0.36558429746853416,
 0.23729880092487277,0.8619794871439198,0.9592408992383619,0.8793518204323453,
 0.7057514747980136,0.6142829260777724,0.9965243247186106,0.6475293659955149,
 0.5904092295248359,0.1778889321184448,0.1945574948278478,0.9673165650731307,
 0.548558826130925,0.3362037957506413,0.06473183240991265,0.5361980690610996,
 0.5596687657942981,0.3048702159922365,0.7952840291843716,
 0.010793143494307711,0.6957644782374558,0.1075585348622018,
 0.14672047379878672,0.5481527046458857,0.5052270379985614,0.5965275477383278,
 0.6048553581857502,0.7734438779453386,0.3976413204214533,0.5845493999110978,
 0.39339792983313604,0.2273101712387967,0.7934993077771646,0.8096063675993468,
 0.9268709544396508,0.9234013117394799,0.1381595899714304,0.5310979707403191,
 0.5365480116242152,0.05884233304625486,0.8292406266799494,0.8335654239707102,
 0.5921393346893469,0.42594062025243973,0.2428896316999083,0.9153011619629371,
 0.28602113166629173,0.8110411711138454,0.416115278600956,0.8765107688150605,
 0.4359566830646495,0.08519693190818123,0.9431408419154523,
 0.21000556948901627,0.07695477434312625,0.1109436267632059,
 0.12705707041534753,0.30051014122676933,0.12575882938620686,
 0.8541822201470235,0.5424466991422006,0.8842012210441332,0.648652636605638,
 0.9298621538050333,0.5576062419495853,0.31640415605492833,
 0.37216653613326756,0.276373965450726,0.41806649241680627,0.5952653714727394,
 0.5018028602045421,0.516285375904253,0.08086544613221347,0.03118446906137773,
 0.6360514602844027,0.9713654880919264,0.31795477027196595,0.7483041651583616,
 0.4106013038084657,0.402563913436794,0.006726507894490252,0.4692707547990149,
 0.24901841069523878,0.6625575113945559,0.3516134579376752,0.6038571559679955,
 0.7554827798163208,0.14639380097717947,0.8010153482474471,0.4493793458204207,
 0.11088514226106772,0.051140424384887084,0.11241950304891468,
 0.08426647417483402,0.3221611340518691,0.5689168724391098,0.4842959195997184,
 0.3170248785765284,0.14967679687493485,0.3994528573334579,0.6317132392353608,
 0.7690506685947052,0.5880508318354074,0.5330034932158098,0.7976484608739812,
 0.7306444796331393,0.6243134729603718,0.06004467156765059,0.56834632026514,
 0.728679319314567,0.8950567467001614,0.47049751415877494,0.7728606959038886,
 0.28483324453548253,0.2805593530800611,0.32050808995227476,
 0.0059050109397783546,0.007227392964532564,0.06077863601094735,
 0.5456988329171235,0.19414822669090992,0.261664443945374,0.8563750430175497,
 0.6106084519315437,0.18509830178992304,0.10081442990724288,
 0.3041706120419808,0.8138139013037111,0.5474040532746945,0.8597720659421914,
 0.7660527781702191,0.7887675712160936,0.24950756036262955,0.2003624831315709,
 0.7695897842107383,0.13762999479066085,0.645810990926832,0.7414447908365398,
 0.26282958040902205,0.5537604064515818,0.640212110063377,0.550706431135746,
 0.270613947120393,0.38650939150842023,0.7292195450827719,0.3424154747450403,
 0.23636138559414965,0.4438504360596863,0.7236186216416625,0.9113887473161599,
 0.3567435162611201,0.11253325738268294,0.43574718254922584,
 0.6064796015570125,0.9392153354072137,0.8545091130714425,0.4480095409878282,
 0.05896180433669451,0.756190860563431,0.7678676244265915,0.989943182893564,
 0.20436495571157742,0.45858154768040704,0.0357491077253127,
 0.8009279019655218,0.4287055244630703,0.434851239578131,0.33578526273065146,
 0.2031869325329333,0.7316375903517873,0.46228595986611964,0.9930871141490533,
 0.023742724624007838,0.8448884019749272,0.9854753901656188,
 0.6835658888308237,0.026885794191003565,0.24634250254882162,
 0.0021362970423112193,0.1254712785144274,0.7960148505568404,
 0.7576788387149276,0.5186943464608235,0.9279604411994185,0.7847570366937014,
 0.11595574751325777,0.7711952338692099,0.571182661182056,0.25150136787225563,
 0.2078259961162776,0.5374384381205702,0.8336298292908667,0.2687173914833174,
 0.5312887766361051,0.3650338563255445,0.9449062484157843,0.08164798045946708,
 0.2809013984317299,0.6403824138729777,0.6018444886417944,0.7329472190201045,
 0.21161788696403083,0.14377833231282877,0.7407917278186024,
 0.46058130805180675,0.5009005656117939,0.48407441772227766,
 0.5514757538784157,0.33609909706994456,0.9716053809047546,0.6154078384264486,
 0.9267310765893618,0.007942074148967304,0.8464847687590193,
 0.7029606328254537,0.09207741255514623,0.5631317456162835,
 0.49196553611656013,0.6118037590093092,0.3915891572061676,0.6819213330445381,
 0.2295955104994759,0.6959669899842573,0.15045618718437304,0.8577811199647064,
 0.13141064785143508,0.494928391326504,0.2607380798994007,0.8355126675085048,
 0.307047987152251,0.5303218398648283,0.15791280888773662,0.4642855280687186,
 0.9240258848115858,0.5569857735509101,0.2977674570375003,0.4653316746088183,
 0.9327931205685487,0.318337420913414,0.417427800742492,0.7198331984340582,
 0.1292278779957604,0.2856558575376511,0.09022210035682754,0.8905759640162948,
 0.4464219166182557,0.5873923094594641,0.6704618434812927,0.48201922114648155,
 0.05586858807314754,0.24627371488076033,0.30058104832989485,
 0.7720942940371744,0.7402106754542269,0.8639833661827262,0.5860315392183607,
 0.7183500760561635,0.6231489677452442,0.4624896938958275,0.8641506037705619,
 0.48304360121671386,0.2985496042677267,0.9692224905072414,0.2364027372366051,
 0.7879538395029186,0.9179529145838535,0.4595439320552547,0.6435050651621246,
 0.9000327856435006,0.8483400067074983,0.4974880750881656,0.44288908866901244,
 0.40521365324732106,0.35330842923453387,0.4210274475685638,
 0.4211308257511157,0.2853341155169249,0.3787017545258937,
 0.008710059984715968,0.2519387267462908,0.11655177097033087,
 0.05537955508959924,0.5103189238851167,0.029660896009890414,
 0.1679948274741676,0.6020713205263054,0.5614641968402433,0.16228898428659777,
 0.1094175628013978,0.36822587267098084,0.01051128168212434,
 0.9463461995345767,0.6009251059116818,0.7474674631787936,0.1566781366939083,
 0.6018313624829486,0.7975555491141252,0.8139776201624023,0.22332805165684355,
 0.0291795085101747,0.8494658120661946,0.9530335450915306,0.04016135739047055,
 0.0499981207745499,0.24238656387925694,0.7660932145552469,
 0.14788235077888479,0.8189386679401188,0.4143657962268229,0.6915998918111751,
 0.22210153015291145,0.5890717414969742,0.8336029837121047,0.9990750785952918,
 0.938796649234721,0.5832372928971867,0.39286753258333795,0.8732984483139198,
 0.8572148422261664,0.9042794330102674,0.7673696932037203,0.7504886188665384,
 0.7533352514067655,0.015827273107256956,0.25231774240913085,
 0.9848641363568501,0.8360424270513497,0.9944261737085863,
 0.007965010278368156,0.6859242431286159,0.8517749232091634,0.765515203159956,
 0.6330638799567712,0.8056669487187667,0.20833070542871335,0.6397410873944813,
 0.1557512094544674,0.7258391393124488,0.009619997407938397,
 0.39431075069444244,0.7632899353767717,0.3087498191320873,0.3608379726325386,
 0.6301613165098003,0.5577670420436448,0.4033296186494024,0.484894743401566,
 0.15646431379074843,0.42028044863373126,0.14596464327663328,
 0.0487186290687438,0.9317534693411398,0.3149272589536607,0.08285329509464967,
 0.1541747877199694,0.031729016808047406,0.48135228064256186,
 0.13899086310296827,0.9972441010380144,0.7022782774634255,0.2236221178317026,
 0.6419045215869797,0.6689473135550734,0.5959926395445578,0.741984564622171,
 0.3366560082096681,0.5959297130933914,0.21125312208845626,
 0.020050719602859535,0.21542767977724298,0.16987767222591144,
 0.7008286804767545,0.22374585746729714,0.713234992649297,0.8532373549263184,
 0.659455164399139,0.795053212027068,0.4340960554519582,0.2632915631080721,
 0.8784570726443994,0.5280143587306159,0.8064559851725426,0.4461014016512993,
 0.24984600810154167,0.80290394870394,0.11835970268479734,0.07553328538998527,
 0.32605788286699044,0.932454256064743,0.920936659371716,0.9588683615380753,
 0.573487628205781,0.2221550184206308,0.6205621762495477,0.13614001225753536,
 0.16037915117928492,0.9772333920529588,0.04515881787860909,0.861931087978991,
 0.5799966791986089,0.06285257948069456,0.513902205529736,0.28604734532339837,
 0.847514270752105,0.36811418326498346,0.8922457222591061,0.1878814933053985,
 0.7000042680391465,0.6721617895259531,0.9592557373542119,0.8631706967563678,
 0.5178409626799169,0.7363468100484106,0.9614005348425133,0.32195702947861327,
 0.9404367535724335,0.020989102034385754,0.4419253592609964,
 0.31956916987697315,0.5223472552054311,0.7573587090844038,0.5294757323748258,
 0.059514243556681445,0.9005975041526693,0.19744764259280612,
 0.49374885668630863,0.00980229883553474,0.9109361522243661,
 0.44554160903901907,0.1576404188692696,0.3097634851173463,0.6452745745117116,
 0.1577840229764914,0.3730269494394378,0.08540919452477336,0.9385624127702022,
 0.2269671033329893,0.8367229049796435,0.6493674398607993,0.916115295215354,
 0.0993632225499741,0.7430205860805364,0.9623578352272426,0.40704056015898593,
 0.9270915748282795,0.9731535771828056,0.6285967901636843,0.823365725123067,
 0.25969974062961376,0.3656996449578094,0.6791321567714348,0.9519067886403181,
 0.3577426171748066,0.07093327516870374,0.383342698378596,0.6184873595933935,
 0.13415295133445948,0.6550290248806719,0.9001574650480677,0.6521045589569071,
 0.06432055111858925,0.5940282065826396,0.6640787211379884,0.6109140252436441,
 0.6753978883578732,0.47217801879715804,0.4067736639740438,0.9109446549782222,
 0.33630662407604883,0.6253773133852577,0.5378657898357344,0.9573777082381296,
 0.0823622175963975,0.676941718873158,0.9073648465475895,0.34071072883976705,
 0.483738480136943,0.9330524232191839,0.6552358896981416,0.7280571072417916,
 0.48023491115330696,0.317858110614484,0.44989162503323243,
 0.31335113081246413,0.37087314502731683,0.6177280941687722,
 0.5665276752755588,0.46986276041756636,0.43232016277029106,
 0.5063026877988146,0.5344564248930682,0.1723508670471161,0.7304161538976411,
 0.12803491510691412,0.058928419857916636,0.6069014743508494,
 0.3771519146960708,0.8593381418056656,0.7644232100017001,0.8207034652775345,
 0.5782663767103204,0.6780040606457614,0.8028508996944564,0.6426182441890065,
 0.338990465746122,0.6719407092188743,0.7759065250611269,0.25402859142436696,
 0.34434868552987175,0.8674335174253519,0.574998559766756,0.15023807360251396,
 0.23794371454616792,0.7170188218529993,0.7720150218701473,0.7271582329670013,
 0.8386316700718429,0.13906566141129917,0.6576012414043364,0.9552925278975644,
 0.7791220824037919,0.31379682513967655,0.0574127044759104,
 0.41445871217358565,0.15138640739936826,0.25646953646826787,
 0.9228810487226642,0.6248380712312893,0.7323346732934884,0.3573081635059048,
 0.5840063975835369,0.05497253857323159,0.7926837805785198,0.821048996207539,
 0.6911037474533889,0.7065710513763237,0.02787246750372452,
 0.11164082154724198,0.6033576625527689,0.6579210750258001,
 0.31410924920582395,0.27897717823322365,0.3540074218878525,
 0.6979845900272699,0.998627392344216,0.09575970776552456,0.38476704822331464,
 0.551752245058654,0.3896958119006161,0.6796438670659455,0.3629887965006634,
 0.16946683893719583,0.6139610303551296,0.6001980858896154,0.617579634386095,
 0.7541665178095623,0.26468551012823893,0.023590480224169985,
 0.2185611146850699,0.9862942380202646,0.6160554647334819,0.3969705925230841,
 0.6173911797746651,0.8407491210749587,0.3140981744413349,0.4491185356518812,
 0.4844813155376473,0.3342455877946702,0.2936678797820579,0.11282882811326944,
 0.609560671423289,0.6277987969101249,0.8552254434941434,0.23530808287263394,
 0.2815185124085524,0.21060559260360345,0.2408287256185514,0.4118834289129485,
 0.03139402704921712,0.30973383365282403,0.8830501673599632,
 0.7779796893006697,0.03543529346292296,0.7652394628960707,
 0.43415124663854854,0.02780360224589329,0.7775865370801271,
 0.9746228181043843,0.12307923448190616,0.3344046172747719,
 0.23846202961324137,0.7122639887005143,0.42446334081208004,
 0.0777580640093497,0.06685534473297916,0.5136286289439509,
 0.028247260939210372,0.12341615042393972,0.8888390364161929,0.09130126209806,
 0.1215401498076013,0.25768487475656177,0.8930187609906678,
 0.006807620300672279,0.8869585248708919,0.4029396924773675,0.53137983472827,
 0.05790335902525168,0.16399064097721106,0.07838650593230645,
 0.29997737048653583,0.04976035988756544,0.6875005453991313,
 0.23782711117047528,0.8111535682540905,0.04827170308761519,
 0.7139175442045166,0.2737850207234731,0.9486120866044263,0.8758405627010544,
 0.2940638992094946,0.28737438298395346,0.13042158857032005,
 0.9000753335850409,0.3764144326305736,0.4956687612036841,0.10069275841289527,
 0.9796186313196608,0.9301454867550402,0.38958438099974135,
 0.09019399887854307,0.5637009846755875,0.34195216778773907,
 0.011631803909237903,0.4456197125376298,0.5150390258265057,
 0.9023326827280271,0.4878269846570741,0.557451862253657,0.3226878898927118,
 0.7164105502345035,0.9736669253969539,0.6442036593889674,0.6751081696044903,
 0.18718988875425713,0.030252562114388537,0.10607453470190908,
 0.7112859898251409,0.18270661906179841,0.545013820369502,0.33658093983914483,
 0.288431172184232,0.09235302741394813,0.10052941501987311,0.3391649667082761,
 0.6485010404007543,0.12956481817633425,0.9323317621941004,
 0.11667596879903719,0.27410874223936277,0.22411953940367424,
 0.6831813302842591,0.03513026888150217,0.8038847809938243,
 0.42142309162306046,0.45445895932696645,0.28613383519478564,
 0.6304197366210729,0.5068665020747187,0.6327945605082749,0.653855806960671,
 0.9376866197062474,0.46066222889948294,0.6989411854110447,
 0.22426027750339295,0.484439267499169,0.6846511647548826,
 0.041971697489156634,0.6350625454017784,0.07400709319692067,
 0.8774742875083062,0.09679886231096224,0.49112223928000676,
 0.8847637154885897,0.4294634798931509,0.7827117927088312,0.02228512812910899,
 0.07565180245390513,0.8923449368337537,0.9932078799039559,0.9463787834629441,
 0.5234156713415762,0.8999864080253671,0.3539127499019329,0.5337331964280531,
 0.11477476674092524,0.956711816968188,0.47639037114119986,0.6767410196000863,
 0.41522272581637076,0.11596936354141718,0.2953308142555454,
 0.7066940293346513,0.43418292715313467,0.9949858568849741,0.6148191327430751,
 0.45978328896589804,0.17068400917398252,0.8486077955111944,
 0.21990439627215574,0.5432785044838574,0.3303932222801953,0.7730081725104465,
 0.9986742384128581,0.277337711843911,0.542014332449802,0.7048342225282616,
 0.9410958980709476,0.5494498208863576,0.8473558299372314,0.3459836771003758,
 0.5543907869856571,0.6262930514937175,0.9184893662490013,0.31500367530737017,
 0.9876703456345408,0.28329620874589845,0.8691399840599686,0.5417604549815513,
 0.6590431503622742,0.5972674463448704,0.5650044655706696,0.5877961545763107,
 0.9141416415557628,0.2878862748236427,0.1123562808705858,0.3584648981237696,
 0.05912061368795163,0.06778266149262346,0.6653408221471444,
 0.6340960461248937,0.4548692453913825,0.1795202276464406,0.9906796507766672,
 0.12189536453018801,0.7350217450466835,0.49839070348047887,0.83647870944415,
 0.8229311209756796,0.5758016664377823,0.7856923588033747,0.14826597678811415,
 0.1956722941190061,0.07733075183892091,0.6071738338010005,
 0.21316060401113668,0.8123078490612717,0.7426198464266024,0.578699708830941,
 0.9962629887693072,0.05377480701403348,0.318812688840296,0.5961544764273049,
 0.578526559539517,0.6221729973134396,0.02063269620728947,0.9270999868055756,
 0.5521362910259238,0.24627803818159133,0.7411194726937556,0.4016553639854298,
 0.7211011866145028,0.10795153473707542,0.9675376587685492,0.8567668627892219,
 0.04146037819021209,0.9487578123419254,0.48919756271529613,
 0.9245886725461235,0.12593202925608593,0.169041784976652,0.0993339837639795,
 0.6418583236255735,0.40384403748953046,0.3252886350447949,
 0.38297901330121187,0.7390826314450429,0.052889769855221314,
 0.6885847210621897,0.8040808784882866,0.9410680653522602,0.5532564735688779,
 0.8236955672918611,0.3122880665087139,0.9681543700188122,0.7923863050039469,
 0.04407794069135662,0.6111308422716001,0.5970880724191742,0.8295035129597039,
 0.08495457937652517,0.18460248548342106,0.428862728782495,0.7449063298488173,
 0.9935395813953936,0.9832113647618825,0.27866415702314973,0.6766438642793544,
 0.9479719275522615,0.435614604799885,0.44564311560055103,0.3303560074864963,
 0.2107014957247879,0.7412103132479124,0.2682220331628148,0.4997345869443842,
 0.3561343046695695,0.1039394018316172,0.9849771677167634,0.2799384889040284,
 0.7621964887670594,0.9360563187070734,0.02593367045649364,0.604533757191147,
 0.41359119627928476,0.2723408667918681,0.058167797348685824,
 0.703918975102029,0.9887936878492377,0.7204449090519948,0.3957179160003884,
 0.9267984195847142,0.9978693721428885,0.8604103779992515,0.33616903338280224,
 0.23901688567167367,0.9369036216704629,0.15591084146332024,
 0.7384393084512295,0.662570936138054,0.533634200318259,0.28001271309442544,
 0.8083863650672566,0.9058215582131027,0.6804583856456299,0.25933501405809867,
 0.5314918703792922,0.8363098269994858,0.6483706657438094,0.14914465299927848,
 0.9319952164625981,0.3193083286447793,0.35632134349631295,0.2696188726107772,
 0.3961618332193141,0.05361764058603047,0.22405357689329197,
 0.2430715898117467,0.31970267695505583,0.8103079855881541,
 0.009826195276805394,0.16905971245557216,0.018411444866296645,
 0.2667749371734659,0.03686624099834257,0.21359631585170624,
 0.2112402420109557,0.07711255271248585,0.17200301333503631,
 0.6193394869151496,0.49886492404884897,0.4922850353751034,
 0.04448694385809082,0.28710112619131856,0.35378162336685015,
 0.8024558656871464,0.8974586289035995,0.975104648281272,0.49668535375076794,
 0.6416167647640305,0.5222046112164997,0.57979598750543,0.2449341869719498,
 0.242378639861348,0.6890372621174606,0.5741546976157689,0.4939850367913925,
 0.4156665979861852,0.34563158362947344,0.9295022480437551,0.5708680712356926,
 0.13386978832979213,0.3162446139593905,0.15548789420101705,
 0.6002678699436643,0.35826923835270585,0.8393415613586949,0.6682565559753876,
 0.1007503778314518,0.9676871038085461,0.9169452938811723,0.9884900192780327,
 0.9296574813948041,0.6487325464798642,0.44020441436792224,0.4613820802375225,
 0.35089553508174764,0.920089852376917,0.020182669629325578,
 0.20755672521585022,0.41867355285207497,0.15122273105510686,
 0.45519524500631403,0.9558989491573618,0.553892279271855,0.26975552756496946,
 0.8611494804574847,0.10662093583802423,0.4761573183502974,0.9345299708374425,
 0.03604531260406085,0.09084302978424641,0.238446584099959,
 0.06291572729572636,0.5627083590906565,0.05790520144682165,
 0.26275715176193326,0.9432777916789747,0.9978795109246525,
 0.24332339646949797,0.3681242139930394,0.2061349571908263,0.3651010569904536,
 0.5659333325559205,0.7890145075537328,0.21390807230478093,0.133137478950776,
 0.9779372873331144,0.6393526568884806,0.6356850505691882,0.8452952285503794,
 0.13981841662437056,0.0027730755958292885,0.5194410486105786,
 0.29368426090658284,0.7833577479474784,0.4675797817253383,
 0.38707802778116673,0.4699660270613998,0.16390042827611906,
 0.5338591020875199,0.2785547931610667,0.060297425195760646,0.713117397372287,
 0.38028962006151623,0.9612296529381261,0.49417117545485734,
 0.41971320890904595,0.8861546435038863,0.8652405133460754,0.5864034293430017,
 0.744789088810791,0.36628618773347665,0.6767698364548891,0.24241436444775755,
 0.5720890516332309,0.6169684741082432,0.6647048873587942,0.19246115933247543,
 0.14343075305171282,0.6668918118618186,0.39324700620881103,
 0.6416817587746709,0.06708033936338897,0.12708975494613328,
 0.6283063399553538,0.3754145610448554,0.5414504791972081,0.6644834147306797,
 0.20512421394265012,0.2601325116351293,0.3604527921436862,0.439020238540881,
 0.15350657394049871,0.2949440729930577,0.5712446691996181,
 0.22387137299429405,0.5187638623439847,0.4402401788122614,0.6689588429258577,
 0.5394187352880215,0.6383995326094131,0.30340219367278165,0.2387365139587876,
 0.7372573829565708,0.8154984434698012,0.492750254207031,0.11413797951091698,
 0.37742283666505516,0.8143043857121813,0.8467184661941063,0.5247978250754591,
 0.34454912038511987,0.25220119991218415,0.6106397573389312,
 0.6572994607667897,0.44475354465720607,0.22127397468754284,
 0.6221522667407884,0.37613320506287184,0.47825879021161,0.07628112413501231,
 0.2921641953104702,0.7699540455137518,0.5404081166120225,0.8083366172330995,
 0.20162666835420706,0.9774269834530745,0.2426312213235361,0.3895154042078529,
 0.5543189861336296,0.4506174016931469,0.489611494124619,0.4075577065171273,
 0.9675104281892266,0.05169822990616435,0.8991066422957849,0.5188794137689892,
 0.45378529602527706,0.622606370888362,0.40492127740359873,0.6407468116272712,
 0.30451431023189324,0.3222691039422966,0.18437385076280044,
 0.9588060074041359,0.2932720129681139,0.6297327784793425,0.6735997372492025,
 0.7484460776578584,0.6804029927125551,0.2390568856115115,0.9408229627514838,
 0.6732021858051647,0.44504631585010823,0.37788235692586725,
 0.6701072189845008,0.03717427699350606,0.5443165722058547,
 0.06963322476231693,0.9383112437107285,0.1887615203715829,
 0.08793821971047833,0.18624207049445618,0.19872705027126625,
 0.8747538148133862,0.12611442742777712,0.7654703469243399,0.9898163617337942,
 0.7168485298461154,0.5507542463374487,0.40948513343188175,0.7835583968658881,
 0.859074259505137,0.3134389636866044,0.23144372144389214,0.5652625749050965,
 0.7043436218379566,0.6997692873251223,0.959673847210125,0.4012772786477997,
 0.26928086115805505,0.9164972133847448,0.7090278081799503,
 0.33830954526811097,0.4051256017883472,0.9960898353942138,0.8614689531930725,
 0.2627540542063028,0.8326908598045308,0.41299685780656226,0.693851114135229,
 0.6167243737738268,0.7016902355887349,0.12872150731014265,
 0.32108014824242725,0.44366147294274594,0.9395570989961489,
 0.8451521791842369,0.5200840992364514,0.44778268571837676,0.0974843417191944,
 0.8396682050885054,0.8888758150441045,0.3836422300121721,0.6635900038150045,
 0.8917565960443866,0.6809115574502627,0.496742549309804,0.9046698975758599,
 0.8616322369492655,0.3434303969522797,0.9290733780889413,0.6478832749032741,
 0.13589956148528615,0.0941859705789736,0.47497892333903713,
 0.2961869531704172,0.433139791819687,0.2554509780936427,0.14354189041692156,
 0.18199539491211325,0.17014519049967736,0.1614863557718218,
 0.6760550345936925,0.6499111875407276,0.7982775262319036,0.5986326810693418,
 0.02341383542201636,0.8003306903084394,0.6968425302371167,0.8746592258503714,
 0.6712498531839605,0.8218006827253982,0.7818106022459077,0.9784143884513955,
 0.677092548942232,0.8925776684878779,0.7632698213396526,0.34529433156938283,
 0.04281532852635761,0.993765659381658,0.29195296621053524,
 0.07114305846806968,0.794192160793081,0.7731562739002388,0.7169104996055813,
 0.5422400583886545,0.960983508429789,0.46903498579432334,0.02213208851173698,
 0.6971190850964791,0.34513299850842594,0.8418050168860869,0.6425771267268825,
 0.031035742544427092,0.845368401523554,0.25482213185844704,0.558009357776534,
 0.4496287023762666,0.3916760746909319,0.6355458681464035,0.6242392976719089,
 0.9750605179563332,0.6469162765121702,0.5080168519652201,0.8629638675639097,
 0.04497177003655006,0.10868501679343612,0.6429135727862478,
 0.18001429963579507,0.6523260205422086,0.6951299938472744,
 0.25865823693601897,0.3211967558483717,0.8252184337550587,0.8708235929423193,
 0.31991175122789706,0.10761848687692499,0.12342087358885834,
 0.3021264425753293,0.18813224185165667,0.6501909936162776,0.2012853883199166,
 0.7768488995732349,0.2392229061939668,0.4015585463693091,0.38067170672563566,
 0.707659736012082,0.290030918917042,0.179249943902681,0.15127536038491307,
 0.14795397814657685,0.8796859714143217,0.33772281537414006,
 0.7130956632831612,0.434765421450811,0.5575444864512344,0.7771877697313212,
 0.10853184130657278,0.0166799044514494,0.49091182483257045,
 0.8021109763668697,0.3129357985011936,0.32916291252179075,0.5099568908942028,
 0.981371661807078,0.36825711015582174,0.9750460854568326,
 0.019839197982768786,0.621448248981794,0.488440459516142,0.5291958269788226,
 0.5353873580841249,0.9246201949245315,0.6490399517143348,0.7113474737347153,
 0.005921988921629873,0.2275371136386949,0.11619155117957947,
 0.31256852842430016,0.15806856420879067,0.1830031650826086,
 0.8981662979282193,0.9731879954625937,0.1783730399739264,0.4261442967827891,
 0.9906539451103497,0.15570851415738796,0.20420588036052834,
 0.2345071866807985,0.1179900657221098,0.4751274453944724,0.5942045228350046,
 0.3701147787239999,0.9254075100618536,0.3044499255577877,0.04694794260434587,
 0.3152831826662781,0.009128878902375348,0.01284491187537773,
 0.2923264329876779,0.8740469750221105,0.9623843952017056,0.557033538779771,
 0.6310483985566124,0.5707927197188811,0.8979407435347541,0.10040073460763721,
 0.6020411514314675,0.5186010796304943,0.6062023287095518,0.6764817348774382,
 0.975976098965222,0.5250967337067172,0.9895773196112041,0.8646570528063051,
 0.1769988716673706,0.9707880989867599,0.5084918460194153,0.5499067230323263,
 0.2347286734746865,0.7826721493915857,0.17315242480406678,0.895110882049774,
 0.3539275021867523,0.5546022746108958,0.468989784520634,0.5804655204511072,
 0.1798409577106319,0.5573935071730499,0.31544500616766624,
 0.32498309681433335,0.9168877188704095,0.7253922177141655,0.6863214967537539,
 0.8006244962459848,0.4910660819531282,0.4347305888897077,0.5707316547457205,
 0.6632025212204239,0.2366312721133641,0.19982738946437062,0.7642569527266136,
 0.47480353365207273,0.7767957092182762,0.06661803945261291,
 0.8603863462333039,0.3972188447768792,0.8504509608254128,0.2229687862873947,
 0.5958829214320596,0.4612427069203563,0.33427007172308687,0.7580259572660687,
 0.7870660306850759,0.824290034422671,0.34168247410097097,0.7992228063062914,
 0.8103277993434351,0.7688786009263693,0.6983399011097992,0.8980609653609706,
 0.1955197138409953,0.5653210488743787,0.32820754759232584,0.8451874621363195,
 0.8386166249845335,0.3609138171666033,0.3007684072530624,0.11707017030146238,
 0.10776985989779442,0.9936051388726339,0.2753528922473274,0.9424366403480969,
 0.2917112230826897,0.34524967629059555,0.37105652683713597,
 0.45503588349813073,0.1352861117317239,0.1175040075319913,0.5448534799138051,
 0.9104172405805792,0.7908279814551282,0.3128859132584648,0.15925596428849487,
 0.1757798884919496,0.8126381631386286,0.061307333213594495,
 0.6439832319085046,0.3334591607804598,0.28752625920954344,0.1412165969825897,
 0.3725669660312114,0.16910723146209183,0.5357535608688939,0.6839521043860585,
 0.9677498042290523,0.37994880422556443,0.9511602888402846,
 0.15135295174184338,0.07325740980655604,0.9664899195409408,
 0.0202454478771793,0.26974231172771446,0.7615095318520332,0.6144292550278497,
 0.8019685874449307,0.946047202309471,0.9099450122158097,0.1315980379747289,
 0.9243540306294866,0.7660499249921903,0.8620384325657424,0.15272316446907275,
 0.46524361624707133,0.12169207273219484,0.07535828371818665,
 0.9390204471677333,0.56315761077419,0.8540232212702565,0.6176962541871956,
 0.34161429332781035,0.47028866928949153,0.5808726708919567,
 0.7238482270486196,0.688407112981907,0.5943654825812316,0.443472242361169,
 0.1449614542766976,0.4272736307982905,0.38199794082013117,0.6906388067260809,
 0.6480113347655497,0.7635457249579551,0.681557073240548,0.6885688929302608,
 0.823123998647525,0.6006245333885328,0.37210061691038687,0.5030777089033323,
 0.9240793355796154,0.6614060362596943,0.7027738233247443,0.9336557201526811,
 0.043706750994059185,0.011157393016454464,0.647934648756212,
 0.24436102238992763,0.9906534451725757,0.47380167615675695,
 0.49739539065599636,0.539295948991078,0.4867222410145522,0.9556708556215535,
 0.02970496661004307,0.13142365449409354,0.5983175181685523,
 0.5494221687547182,0.5377787110555368,0.33519870352647196,
 0.39026518620855766,0.8440364775349868,0.792536467943665,0.373955540058418,
 0.0891894803697364,0.3205581075120515,0.8093349705252214,0.0969392482788961,
 0.35636036380408587,0.5725611518270507,0.13103229491935053,
 0.14026713787527223,0.25968630397048575,0.20699168423525405,
 0.7694245217609093,0.6066916516419603,0.5057057729925415,0.37000923571972744,
 0.39286981587745506,0.15698563467596394,0.20195344897565493,
 0.48087884474347586,0.5138439124021703,0.27894147200534325,0.763591062846661,
 0.013434187585027235,0.4738262251084878,0.9023068261823487,
 0.4581165955750618,0.322796557042323,0.7864368037264283,0.12784942686415757,
 0.6189492663870193,0.1075591675900045,0.2082796461965788,0.25778969395808293,
 0.8851022912816051,0.31296531692602303,0.6923754980398379,
 0.18210674341898592,0.6176661821570306,0.6700077339437063,0.7283879057919808,
 0.3442935688712495,0.23437561402590656,0.42208395077719163,
 0.5626059249542569,0.8802289846350844,0.4910995044157138,0.849675797469476,
 0.6219648880888882,0.7680764347718075,0.8188583855542864,0.034918621481417,
 0.7419096155159707,0.8145664909065558,0.7694099760658006,0.277385999482455,
 0.2534334482329079,0.027746276398665004,0.35886914868716147,
 0.7534060650797494,0.9787403309957358,0.633699171948652,0.4194474588492241,
 0.04401951397531767,0.17134034705382428,0.4818388733802075,
 0.12809657380800943,0.11225181149808172,0.8311634368642272,
 0.9545896002764114,0.26402529963647736,0.5828800437300742,
 0.48137619962006784,0.06602930261514972,0.6920713779399854,
 0.40022776802183857,0.5072414167823652,0.022133181532859814,
 0.5199366415694016,0.18589988773865462,0.27015672347802244,
 0.4187125697399713,0.8891828299689221,0.8054026103644454,0.6499759395677744,
 0.8310238269937642,0.4063547257556932,0.07943354914717793,0.5399006799243588,
 0.0832012893473385,0.05515406821556934,0.8130438718067455,
 0.23998083725205266,0.26687786232905053,0.7293841267470262,
 0.6109025016044773,0.6483770150420154,0.41072010461475283,
 0.02581006423405452,0.15836658182299168,0.701925673208994,0.5188795543426434,
 0.19397009660844777,0.5610011692490523,0.7075401903093578,0.3230918208204988,
 0.5722756282074448,0.9352572495591994,0.12924962194795597,
 0.16160806501862246,0.642051900853523,0.860585744254759,0.5951188219163566,
 0.0674410967083785,0.7939257494192088,0.0931221156768155,0.11831186448397946,
 0.7299937843918876,0.5462379553864597,0.19720287841308137,
 0.44982122908339695,0.6115477378635573,0.3939154707845588,0.8450360681867232,
 0.6131263381698222,0.628931704164082,0.2023266524418822,0.14135445261508695,
 0.9403990608464164,0.7181028750448117,0.18447614071976703,0.9255101387015864,
 0.19277523240194017,0.7808895153883915,0.27165607353964116,
 0.28826466411873375,0.9153810416758833,0.23597489040826858,
 0.8723504000365406,0.911206264323953,0.5364186131944708,0.31378496693270963,
 0.36537901560123653,0.5088515835178379,0.9417294655498198,0.9423215565139702,
 0.3647267955008484,0.9348823477954801,0.05111728545222327,0.8286104089903228,
 0.8146019686143245,0.33452785976067345,0.1129187051840721,
 0.31027251002113243,0.7659173551451183,0.930161168324426,0.9729889233292368,
 0.9115461821392907,0.3991037822385227,0.715075447644849,0.6869121393558206,
 0.47047607376011313,0.6940409598661338,0.4996796617326642,0.5441215711943365,
 0.7130046491084868,0.31145635571779107,0.6094876184440453,0.5440889015312748,
 0.5171144633073461,0.5780347932383315,0.20091787561550922,0.7731404191177005,
 0.9175497334506513,0.807520213241636,0.8791893962063141,0.163553935515246,
 0.47747163931783143,0.976236937781209,0.10263659712696338,0.4418165420746817,
 0.2676640275662684,0.6378538708905358,0.746726269922799,0.011383828800254459,
 0.09012275554563076,0.3610491025005277,0.2305961651253856,0.6557052937083214,
 0.603308484942702,0.22612648178998784,0.23383363635981702,
 0.20770637956250515,0.26548076105610285,0.21400663034054201,
 0.2729207806434947,0.3364612699191879,0.7294693220037508,0.47370907164005016,
 0.9937410487827673,0.8010541438235215,0.2592547073389505,0.8034490629043844,
 0.4237507430383862,0.9312274367353168,0.7606279254446987,0.21330192459023156,
 0.01942538738374855,0.09683500033998416,0.8065176568950838,
 0.8343661882815536,0.9063872912056038,0.5139173001963905,0.73688693821868,
 0.028999147429959682,0.1423703444139316,0.9352110382445553,
 0.23384416727980162,0.42783296912897817,0.24408404302034648,
 0.401696764887124,0.42874424898375285,0.8474216588415279,0.9271547788055219,
 0.03211678949406105,0.6476034636165557,0.3530919280094419,0.2557144117717126,
 0.5957797284194584,0.8876830748708859,0.5592475407535653,0.01590174659571919,
 0.5048374440479653,0.5300919375859948,0.10468013466020842,
 0.36176549798828195,0.03704828981464059,0.35347294710200583,
 0.4749008692652319,0.21547680272789838,0.9084896671545024,0.8828900726383204,
 0.8299869790751928,0.6779940363561704,0.271842326420934,0.24987539226266342,
 0.722654795779992,0.17342097765796116,0.6697015213787582,0.19740608783225955,
 0.5514363251906726,0.5723697159247016,0.752633889847638,0.7865533494305378,
 0.29455935226804364,0.669347586677905,0.31296077311770243,0.8158868148146019,
 0.008601184223862579,0.2804941319125077,0.8090481644771154,
 0.006174777077329274,0.08984842859908204,0.6016861833427691,
 0.02608463166188779,0.5309279859845961,0.3867542658632954,0.2735939377919411,
 0.34032301528543396,0.2621689682832786,0.5491907171673532,0.1918838528735647,
 0.9134823539493718,0.21340567596998306,0.41355698891231585,
 0.6113714197379272,0.9071549108838542,0.8060488691238018,0.8370354870012914,
 0.8989435658814162,0.39926447973044177,0.8578459155669579,0.6762084090766676,
 0.7394022265749611,0.6060961896833688,0.8255412328432546,0.14847862983686133,
 0.7340341796729273,0.6379769260728873,0.001663295020208122,
 0.5120342852327167,0.6073484678225003,0.35206264865936365,0.553338557449888,
 0.8443674779832778,0.6318183386743587,0.6610550409050493,0.46608092663792533,
 0.5937777918349683,0.29393269912161246,0.8451075897602529,0.9703343670520329,
 0.8261296238651947,0.1434174456355688,0.012234885174753019,
 0.03621232904869487,0.4119486166811296,0.6943736429366638,
 0.20412963355701352,0.8110178759178854,0.15088178157078747,
 0.03506851471386818,0.755856911153606,0.6146057337059876,
 1.713484482586125e-4,0.11826789385046832,0.45926856116533843,
 0.35280810793532325,0.36380766949510124,0.3578770662299888,
 0.8633328423110479,0.14110919530145505,0.5802135707952121,0.2969418271212618,
 0.44484242603092194,0.86801985876125,0.7730284757460567,0.26384603386347116,
 0.21671567530016378,0.7686647328204539,0.41295913169383136,
 0.8808919767841649,0.677098514397846,0.9096477037232519,0.8113637699451343,
 0.051565535026485754,0.5588155595158399,0.039187560293281076,
 0.33419346789368254,0.3693095996387358,0.8220939004247951,0.997132211893195,
 0.9140582880169023,0.5833972227412395,0.5005574591672051,0.8941256162235405,
 0.1451678189547746,0.3685932425011835,0.6824289324008835,0.1645513071097966,
 0.6305717530483327,0.5539059905797395,0.5438314461114467,0.9975553950339274,
 0.6617271907906676,0.12429360898731057,0.021971150564601305,
 0.5717853232483706,0.49779640872304787,0.3714492949688606,0.2267532730338514,
 0.9381459749464501,0.40525342225281,0.6683190212622964,0.45029472440041407,
 0.8221970055855955,0.19962257483004364,0.05554503120169396,
 0.2093320211091112,0.4962231962826167,0.7842243503641494,0.17094268104546784,
 0.6452172739431539,0.7427047742274817,0.2605065233506684,0.8422073874173915,
 0.37226211030729595,0.23899852057483273,0.2052587950703414,
 0.7852851056365017,0.3909306387500512,0.8935192218059467,0.29906206569274174,
 0.20377080544918003,0.40532044998034666,0.16506397052150934,
 0.4437531965585291,0.9722999935410263,0.8536479760330807,0.16573269899780296,
 0.056325304674783694,0.7707464863490388,0.2686115384497474,
 0.8937320334425863,0.9720223085314668,0.5740271571765341,0.1451387613872538,
 0.8499282942567244,0.6250881951207969,0.007666396283249388,
 0.6263009739699127,0.6273041377268208,0.6231805298769365,0.9959789991552943,
 0.08568107925333268,0.7316776965552524,0.49778036971924766,
 0.46991514323824235,0.5204087001641992,0.8128966497093689,
 0.03684482913450038,0.021017896247915724,0.6090654126370059,
 0.0049900310720177865,0.9362144954132002,0.584772369043293,
 0.9777191935828676,0.3569732667933522,0.23256592953911215,
 0.18401943113039554,0.9246811510641252,0.917705051299389,0.5103794474175353,
 0.22102398312696336,0.1533414707859021,0.9680377633141493,0.8654616199421257,
 0.3232567379489445,0.18947562294312925,0.6103152816899009,0.7741929835400985,
 0.5248953674555363,0.7820063444899198,0.8803302542469135,
 0.013532837366821315,0.3774580144387175,0.5498957869720174,
 0.7274689685676551,0.9508087203835544,0.29971993508652794,0.7775306835143048,
 0.48083276603163705,0.7008255926973316,0.7339651122320492,0.6334335054406218,
 0.5566128473133048,0.8858019037556993,0.9583782829379437,0.06324824363395243,
 0.4210891567639514,0.7934000495222426,0.3490637110226954,0.6101137920492715,
 0.011638197624261326,0.8872832925108672,0.27526136554311353,
 0.10146303843711779,0.5972628856213598,0.3387574002651408,
 0.17594981187188252,0.6277367362795305,0.6132265250613385,
 0.48415616961277186,0.08759185284232984,0.9251364729167827,
 0.8961780454813066,0.16861480981138766,0.06756773176742603,
 0.7010467830069785,0.14371135785563172,0.4762724228116606,0.7713376346092116,
 0.8681300797367647,0.45515562363620343,0.2509635695266341,0.2089208194633586,
 0.16485286881326833,0.21604779111472405,0.17984445698359763,
 0.5834359073945035,0.23128037088958808,0.5350369917497145,0.25521663371239,
 0.5672147399347794,0.6323336243497787,0.2977830148460773,0.5503233920055635,
 0.26445250086443817,0.6939996405073359,0.4307233354157768,0.5951670283773249,
 0.946197915550435,0.4022189353001303,0.5543366056050091,0.09499137871073327,
 0.8814069646066429,0.09961551425618276,0.6325566189662758,0.6917811176412306,
 0.8207356554904158,0.458402590669054,0.23274762407201266,0.5003045035440099,
 0.4066227473068176,0.6646960156851005,0.8111039330586495,0.4424751471805288,
 0.47040155300128994,0.48471703498776453,0.28543862060583514,
 0.45151508559762377,0.7341902551383972,0.46820810819375425,
 0.4963938547169695,0.7731640294073785,0.5463712814003379,0.2960421430612892,
 0.9413366563048272,0.2390546678436024,0.28139990102343226,0.1492868745688759,
 0.15325983534770127,0.5895670733465348,0.6687110194031973,0.8621193360911511,
 0.016959486215361474,0.29943613616397746,0.17419839682641247,
 0.35122570288043575,0.04298409748897147,0.2782526908165167,0.822582950760638,
 0.10541770002945605,0.697666187264442,0.21775746028698906,
 0.06947690555624064,0.21576323908155393,0.7389733304818272,
 0.7966283078195868,0.25754911715488404,0.3170355637039889,0.9158964212545463,
 0.2021650232280895,0.10325816119687103,0.6379465287651163,
 0.10663198357931314,0.10838699561214704,0.5673618234372321,
 0.6216609204168473,0.37667503897409116,0.07984851872428655,
 0.07143497381487707,0.4445125537023259,0.5627458680140108,0.887865313673928,
 0.6880152712640002,0.8942092032267697,0.0767225001295525,0.9674300186628905,
 0.08626130589263659,0.6315085956563299,0.30457488360471563,
 0.46140110046667215,0.4476758099304238,0.5519279522763482,0.8701630818130868,
 0.9847750916993738,0.22024807514468014,0.11451817139201559,
 0.6106231474220499,0.5273435921074249,0.41167562853479445,0.8660249709070867,
 0.6407723835137716,0.7615172682222953,0.07610048887664334,
 0.46926647342188044,0.5412022554291445,0.16109115176548305,
 0.6273753261282147,0.5988312656227114,0.9321778882626082,0.38834512936171794,
 0.6468398829292479,0.21540796637183535,0.5422881810836313,
 0.13510932835909073,0.9731994716655864,0.8911438281054667,
 0.020740679463756462,0.48068976695816923,0.7852822151968486,
 0.08908743931093754,0.7317119731779451,0.9285077466790027,0.5894962832169404,
 0.9613106501001812,0.6855968131471152,0.6523507670401784,0.8800646345389083,
 0.7650756877475695,0.18246060437068634,0.6041215131647041,0.8572650805532178,
 0.7920360453296449,0.3088463867442298,0.5296982382832351,0.09526260694925459,
 0.47551533152793846,0.46196056652243245,0.8345073495987935,
 0.18578179056174005,0.4098935644215238,0.18529582709608272,
 0.6360179627492732,0.21817404607934587,0.9033908074821545,0.6839659745975909,
 0.7989790111636376,0.36633708520653196,0.6173103814779715,
 0.27337693760221704,0.9983843566971575,0.3510866646648493,0.3026879052442001,
 0.17425369962675252,0.9137194193663779,0.6659007476234523,
 0.38969493940644284,0.7558942890968714,0.29052138585201115,
 0.9169543128323865,0.3741415421525257,0.7173191477951438,0.7942321401222535,
 0.9359107540061092,0.7352384834895646,0.4226355712370512,0.8941674713433378,
 0.3742528370580752,0.9500984731232667,0.3729294809535929,0.1374101974523969,
 0.29707806780873003,0.10816189305314639,0.38736340407337433,
 0.7545991892795264,0.20686846092114586,0.6117745294300928,0.6839293347695139,
 0.7055886246081898,0.8373383164959842,0.8509272854755737,0.7817894631058465,
 0.9827770778487489,0.9428607504833934,0.05863971399094969,0.8231348121740988,
 0.06733621816888313,0.01491131676684776,0.4613991824410201,
 0.7796391251406773,0.5982130012380245,0.2505618363149873,0.48747292581462953,
 0.23191628871789238,0.8545910317614688,0.07615259010438902,
 0.3944015664154148,0.22889030984463055,0.8518365860447809,0.9322472056546499,
 0.978113864177032,0.21062829565956087,0.9495544673998149,
 0.007676336946440543,0.16707477878610855,0.10167368468490767,
 0.3739829779255194,0.30257968334354457,0.7028095334729642,
 0.22901627112283252,0.756043190518015,0.9944029924871262,0.1663417604465841,
 0.20588518556893842,0.9172912657472277,0.03252421758746904,
 0.8907216845623778,0.7443012847981716,0.4609175179175725,
 0.027241404570084615,0.9448870698820002,0.8822702769295571,
 0.7655661999539243,0.6119701935483994,0.27288224921064663,0.8147335173624072,
 0.2760539531929155,0.5004645778771448,0.4583656364941533,0.2807890162838075,
 0.19335459902525476,0.4058017619031101,0.37113738129776397,
 0.9550985195736271,0.07042516818336952,0.6199378067768729,0.626230335635479,
 0.5670683091602675,0.9044848859669659,0.9447316982582715,0.675851353107982,
 0.9936998866569848,0.9628513137267481,0.1079586013661028,0.28719772830073276,
 0.28013237182930006,0.170485407717895,0.613998094217331,0.24962021524108113,
 0.6138218440796591,0.6960049477306487,0.5865231641324898,0.7363065409778854,
 0.636299258241302,0.326058908129349,0.7852243540160503,0.060093963395316696,
 0.268499516116236,0.8663408217157303,0.6654742887392131,0.44165885250534165,
 0.7739490857690883,0.2922305439148354,0.5473028053971222,0.5487896301485657,
 0.6921281344404617,0.6780225889230503,0.78863469163405,0.10177214306780136,
 0.19578974945068517,0.9620799582584305,0.275796110710929,0.11641890291976698,
 0.6481915143534647,0.6676300663544351,0.18601508561911329,
 0.08189357230587091,0.033786450095926845,0.7774499060616645,
 0.8731752646929094,0.5778012263943338,0.4697208294844093,0.6027190543691971,
 0.3071398630126654,0.4964754566359162,0.010403937977570221,
 0.07756072378713808,0.3326438374425269,0.8727446703628859,0.7645722316967535,
 0.3726785727976639,0.553264664041109,0.027078794341968138,0.352081864772281,
 0.3708840437943316,0.9086546629510748,0.8775533766887441,0.07176720912388423,
 0.2026620797363885,0.005845881025076882,0.9873861842824136,
 0.01584495069835523,0.564524506970774,0.6631705183605638,0.02141048738612783,
 0.7042230662867759,0.35984083477656403,0.360441010655271,0.6257527239724588,
 0.9777892808782972,0.9591913120578581,0.9587231385655738,0.29770253456247076,
 0.01871183560567069,0.9554374737477815,0.3605713083020198,0.1247991075134639,
 0.8088434403584586,0.4149264104143019,0.03341989856579064,
 0.40849612284826975,0.6828884074044385,0.3698966930670142,0.7815945695906986,
 0.721895477587358,0.6207132014966518,0.16486285757960095,0.8581598047158671,
 0.9287787664369488,0.8193840362766908,0.05835183675352151,0.5203773348286379,
 0.977009925700485,0.6961821999789761,0.8334895157257283,0.6273196727556034,
 0.26629810744505156,0.3752858682382656,0.06042474675691811,
 0.6032307405481916,0.2457542308925419,0.0452141300650476,0.19269424983362393,
 0.0069176115112996595,0.6729562099974038,0.53837880643955,
 0.37959520241811373,0.47880963938842547,0.08583166696593347,
 0.7744146924193163,0.7363361770941917,0.7710926513840362,0.0652204278217351,
 0.8802833969347877,0.4405640548549782,0.7909406870511224,0.4549754002757319,
 0.294532954431162,0.9468643936963966,0.984607812820041,0.06395298760357537,
 0.5396353555815242,0.3559413352330911,0.011138933802988471,
 0.7955672078175563,0.38049741469648946,0.7171200759186506,0.1573236859791196,
 0.9702457981197933,0.9247886100204331,0.7173742797174731,0.9522968881779088,
 0.44230756076384237,0.45863038490929453,0.27295387332632437,
 0.1551036690220573,0.1639246352560627,0.6258731990287927,0.9237848316667578,
 0.9412807068948987,0.6923663449254612,0.2843184021230264,0.955409187903979,
 0.5173822335557136,0.6614389480048984,0.29902268987710623,0.6763388814965312,
 0.8820661942239969,0.8142518769419569,0.43608196702378776,0.6013544575954174,
 0.3918902754444039,0.9583773719572143,0.10207238489810844,0.8754171506858202,
 0.3556240966268469,0.42528975350359677,0.5635904513736233,0.6214112835805446,
 0.6104956458137655,0.49914290997305855,0.18648385064336936,
 0.1725429628975359,0.8849381983808479,0.22464062099891235,
 0.10725098049275583,0.6724008215105778,0.8169808994265714,0.1764479956182995,
 0.05131986594331306,0.9174607059719522,0.9332078024544606,0.2390816860791075,
 0.6269599120062724,0.13461095768436382,0.14468343317158217,
 0.1574239636894461,0.6061702880350717,0.01965441668286705,
 0.14617570032321936,0.27483710654263904,0.6130071001460387,
 0.9836233531418598,0.588346626894388,0.5334511269435909,0.5389939069771263,
 0.8694539282060033,0.8872089449249948,0.4613712990894353,0.5370468377640957,
 0.8616204968028164,0.8835280235595487,0.39313882065303063,0.9118430087964403,
 0.31410577449065014,0.4406913588774458,0.6037933992718815,0.6468472598717091,
 0.7050850714167631,0.6695221743399585,0.7282389346085031,0.16206278663653984,
 0.6074417783757082,0.2559770581484613,0.5426908338672316,0.8850466085558574,
 0.7123376540000829,0.6648294275785538,0.08759711764166078,0.6149460436440474,
 0.03782938443593631,0.47146785244883127,0.3931090383505027,
 0.7600351330174605,0.09045862229422008,0.2571936785306612,0.700375249798715,
 0.10949113085902162,0.4151554427840256,0.04517168515771308,
 0.18926682601263833,0.6743634490448034,0.3115075739841171,0.8320872941995259,
 0.6186543120254369,0.10648909389177841,0.979170838687641,0.708882778783665,
 0.9212865229398861,0.7205559028783846,0.2681540776209864,0.2133717483198092,
 0.051507093882534605,0.3533421477869595,0.0361817334679716,
 0.5387323861355896,0.7933652925381779,0.4435024023294434,0.12008656326201994,
 0.21901809439008701,0.7369581669157721,0.16901047109811063,
 0.9239121001408626,0.14826397041681183,0.2505121423159624,
 0.034571629918886626,0.6738461090412197,0.377825554935334,0.3742853277410991,
 0.07902635510069833,0.7980704165326942,0.6039448607729875,0.7687049597962053,
 0.8702090022838639,0.3204791978960382,0.05821617428887571,0.6209461350716574,
 0.05822146551591545,0.08363494487472212,0.7875439585930855,0.859777622963234,
 0.4002042967716386,0.4342217868592406,0.48164302817229987,0.7936389338024264,
 0.5263754096804151,0.09886098622351924,0.8556519473341327,0.4414038854190432,
 0.8000171889687723,0.8517183758430615,0.4533576051286301,0.2870407507532503,
 0.028693430093551853,0.7207172924911136,0.41934005022590637,
 0.21076756693247134,0.5047536644826001,0.12350940302531721,
 0.5932839053330305,0.7684828706298784,0.7122327659016736,0.7603865888406217,
 0.25896492387894576,0.15265486441852416,0.15115100276769822,
 0.9638792506737167,0.3145485374486723,0.6178421256318305,0.09788926442515289,
 0.7391553420720829,0.35303549883128893,0.5024303555155213,
 0.035797391440518966,0.09604028394973951,0.924298791884081,
 0.9251428412934591,0.2693184702547333,0.4653938782253907,0.7621070455747154,
 0.3933079235226552,0.5004680407828127,0.18559855974610462,0.8845940397840379,
 0.21383014628343466,0.6786925396998139,0.23482419743260974,
 0.9780308254190082,0.9994655496878372,0.8221456875132177,0.2800669401380028,
 0.2812574575200859,0.8948109606630386,0.013414677567185729,
 0.5338993231026572,0.8405557165157156,0.05571702044373239,
 0.014451733169432823,0.14648539875713162,0.6138055321465234,
 0.39213892948717544,0.9780244689950254,0.6711982320730152,0.6845425707552426,
 0.27576885163823206,0.7500509189207571,0.12435990346087078,0.788075025993865,
 0.7441166554687247,0.9542991388669958,0.3360792246895027,0.9339346844183389,
 0.8972213814219948,0.8744435515047214,0.19484167413733378,
 0.13720514606116652,0.18450976839388933,0.8732993176509692,
 0.16711557907688046,0.8482497842153836,0.8452004014025329,0.3540511410498989,
 0.7049570287265542,0.49605507558448847,0.18108575192633403,
 0.19652052648210172,0.9451352592328127,0.4879922384753772,0.9338846901857362,
 0.18647205585908755,0.559025230438701,0.3883629546699159,0.9015277509719323,
 0.04595865042051883,0.53217290191818,0.19399168682843781,0.653490801784478,
 0.9663252859290551,0.8002273721404555,0.25226586123117234,0.3768163347950094,
 0.9708850734588421,0.303008218992611,0.4943520566563476,0.1402621810409529,
 0.9754107637114007,0.6355840328260423,0.43005880311212175,0.1430255670770436,
 0.01642084210153505,0.9425983420293553,0.10907184083740995,
 0.11625721597395178,0.4734721319150167,0.027882995164762336,
 0.17578048219342746,0.566127961646344,0.6005079502830343,0.8004560644850643,
 0.3584121842840242,0.4763153281802992,0.5001626345653485,0.29183640502337316,
 0.06155547530452443,0.993071935152934,0.6311980917218165,0.7448256352673107,
 0.1748232027207527,0.7496830327147772,0.0992987976232721,0.5625876471466424,
 0.7697942277882263,0.33873664347406485,0.733780515426786,0.19725465217686933,
 0.4257503218678991,0.042080657796807586,0.7503689541315822,
 0.9001058287403431,0.4239830872858078,0.7021771575368195,0.8780771908782332,
 0.3539721938799366,0.9719802115296335,0.21585712240920774,0.8957982326582871,
 0.2992501679382562,0.30200804742742104,0.07612415790447458,
 0.19906415414331013,0.08770747008198798,0.44577373054969394,
 0.6845402268321686,0.19687491066473672,0.652645183492388,0.3381650032919209,
 0.06049026269602786,0.5108661777006791,0.9388435304309428,
 0.17763846351834667,0.08376965537427083,0.18804807079607344,
 0.20685256415507003,0.6052199753642735,0.7315578290493229,0.6943517306187257,
 0.11691256194941901,0.03148022761925229,0.39010319055987586,
 0.07426339567054707,0.426967661688447,0.30762612957131463,0.7207266300954596,
 0.48356953462337016,0.06796669927430543,0.4587664740698214,
 0.5553735602510049,0.12459066723110135,0.4544613731010767,0.561152839002434,
 0.47025813281760187,0.9524145020801839,0.1660174995117485,
 0.46412432138603343,0.8573864838377234,0.8848668348344372,0.3039944510296353,
 0.8195644373930487,0.7250245502273908,0.3414088103908268,0.14240875949275444,
 0.10592281161908979,0.22428787701709152,0.6069785358246378,
 0.6639212006783709,0.7668797697708144,0.8494612686106586,0.7562748740241215,
 0.7478490863590951,0.2452420356780316,0.9292016323158132,0.48514064267929036,
 0.6277405552887594,0.5495708275299764,0.6010387360309528,0.3475174873335084,
 0.07648166357547781,0.6938489559338938,0.04239440934418259,
 0.8539277440905672,0.5452478316820064,0.17062160430150541,0.704812367186738,
 0.3684092949481148,0.15955501368140834,0.42947291310971725,
 0.4293324048134023,0.3177928618633489,0.07326846313278457,
 0.09669585465676245,0.6804485125239249,0.48883510711956035,
 0.6513178788913739,0.2191048327251368,0.053197442192394306,
 0.13546117455938456,0.7571364712155475,0.04975957613679416,
 0.9693259123009588,0.8886866515659066,0.7185588933223437,0.19956998010714422,
 0.3521470045439751,0.3134665096706962,0.05572106134279242,0.6455692412973889,
 0.9716603537428213,0.24080184635432078,0.9657187416519807,0.5276957039616315,
 0.807498847130916,0.45160078272266924,0.8546270897027933,0.46770788259990503,
 0.22681320581825282,0.2067958694642268,0.6139622182779355,
 0.16145599533921506,0.3775913644189969,0.7969609601788389,0.1088985088411919,
 0.5372197496716217,0.1547695735460024,0.7808375862289453,0.5734954729160163,
 0.781223495447428,0.42655824369938744,0.09524434018433725,0.559332702530751,
 0.7899010510258258,0.4787270079558158,0.11156328972991503,0.6170744247364599,
 0.5106811160426012,0.6657679059629046,0.7581882284557349,0.18433104863802352,
 0.2705549674066484,0.032759512334595065,0.5454520022656629,0.866419946599728,
 0.4323312829734164,0.663827884219636,0.8907031858322364,0.17042312705856288,
 0.6358556372578165,0.6449470670862102,0.5621483854979566,0.6217402760455475,
 0.37620532483193936,0.0937625093732799,0.4558751667666865,0.2100601113911762,
 0.14961394664104388,0.03145983673001396,0.06350294369626142,
 0.3872347151297739,0.941879500992827,0.2029445940331399,0.9844908139367683,
 0.7784663811668386,0.32762596084391615,0.06223967640818695,
 0.9172280168570534,0.8634061557051483,0.42244050793088517,0.2967791280584856,
 0.9812720418694426,0.6251893281681185,0.5146545950449303,0.33520097148782557,
 0.9461409068642432,0.1928929430076728,0.8837603476605458,0.3577075583310313,
 0.41025572580708536,0.7610186951463482,0.6410093607058798,0.2745440500434485,
 0.8715709110236431,0.5716088772610403,0.8271671716779873,0.3829828010482562,
 0.3042018137281737,0.8916950138938279,0.5458130360320996,0.09092905226131665,
 0.7408184613681248,0.9962607587202289,0.7719044126836291,0.47862905556147894,
 0.3917490078714887,0.23216901968866643,0.7972829704679143,0.5530392488298632,
 0.7731601067350554,0.06702340949922703,0.5872135240395177,
 0.47809744070024407,0.31064568041353513,0.5007279210461493,
 0.8036524480402143,0.773154991522777,0.9381630402127695,0.5834513511164492,
 0.7867642255603138,0.6912254902687425,0.702831590965206,0.4565595635149938,
 0.21233960346991176,0.9087453388644765,0.5605446163285244,0.099756952015708,
 0.9255542854965481,0.04292191498584785,0.42489202800735004,
 0.29411416949586144,0.6379003695625861,0.7223490015853623,0.8075774963707905,
 0.25591512055297383,0.18926280973238374,0.3242836890194468,
 0.07691949006115384,0.9193150310329427,0.2994258976343396,
 0.07514424010190268,0.4594800924345579,0.5732237400268285,0.7128512342302414,
 0.634905484097607,0.9505096683176317,0.46973041262340054,0.14743200018887792,
 0.322920723126231,0.28912929527770337,0.3552107869514237,
 0.043116898173359064,0.7912561948410587,0.5400304138713806,
 0.20542102391023787,0.7163086272190414,0.830885318509615,0.9312684684539478,
 0.46818773445085793,0.08070521128108599,0.2613091310205016,0.624826951723876,
 0.05956403885078854,0.12641860229751978,0.9659375256658886,
 0.9486439579269692,0.5546660300149829,0.6601005630815081,0.6732112405645042,
 0.3885410181654263,0.1819429699542402,0.35118046463061015,0.6511546405580356,
 0.11953615950520446,0.8157139792875397,0.13020310328191964,
 0.25694779231899134,0.6723536549887918,0.8627547465726756,0.4214752351310419,
 0.9310175020875235,0.3825118672379103,0.7996124996147138,0.54758743583757,
 0.875167564910417,0.7903095724826561,0.29115488620547025,0.9647779289904743,
 0.6211942899603918,0.513910925555793,0.7259670507634544,0.5371175684681536,
 0.9290652191262817,0.5311369771078689,0.6888828494726669,0.44251783795808675,
 0.534211994755321,0.4498038516364934,0.5465984786968702,0.25662915610279824,
 0.9444341102948092,0.13903064637015983,0.8104749075822409,0.7636196372315125,
 0.35765659933985783,0.7082960782674688,0.2813722316539645,0.9171450842582273,
 0.9764232471449242,0.9489160304087149,0.3613422574708349,0.32897129902485345,
 0.8047164108028618,0.10007390071749411,0.9934416390447621,
 0.37675094737400006,0.031656070189091445,0.5861719720790514,
 0.2623148404822966,0.923220893185265,0.28511356267581023,0.8221607799867396,
 0.8410495699777855,0.8319429371685098,0.24193741009923486,0.8446037815292101,
 0.39813146525397625,0.4586225003600719,0.6225011822264841,0.5485748894688083,
 0.6255072098038595,0.2268927765867016,0.871802170862529,0.6949773543380086,
 0.7507587233390974,0.5493365578649352,0.9057401362124681,0.7051725463004211,
 0.9522860554105501,0.5186602294406417,0.22961733900959636,
 0.13537070042623778,0.9636798177851329,0.1122750912849314,
 0.12289731202100462,0.6072559899148304,0.9315509392083331,0.9298580362985842,
 0.9795692534856592,0.12112060607067943,0.0281258755838929,
 0.22719760151979762,0.32057330863295286,0.9981029269319479,
 0.06332384242340527,0.8122687363772395,0.2002988147633602,0.7760354704962039,
 0.9321406117418456,0.15457937048341497,0.004720519255892008,
 0.27506504645220464,0.7355126571518809,0.7893692322787953,0.5682687801720832,
 0.5309769518854435,0.023754462943826327,0.09091421188805637,
 0.4469313648364662,0.3305549388166352,0.2302140736983156,0.5700406566716174,
 0.23304087001751816,0.9580029906156566,0.1950326405141597,0.6944197608178215,
 0.09311341210625212,0.011307972923005893,0.44426818583453753,
 0.6435502174446974,0.6657972049656298,0.12819396830506935,0.964871889710238,
 0.860942070896981,0.46267112072566297,0.8944757168754257,0.7062801276927847,
 0.2178849128025122,0.8927618842918477,0.002651999413748163,
 0.18695804438116526,0.266669779627826,0.2306981558756349,0.8291570826900787,
 0.4505823202292786,0.005292972107988181,0.04738302281814566,
 0.9592470908412493,0.831307848348927,0.818099729622407,0.6334381757041468,
 0.31587730489683374,0.4116743279910825,0.11065628260173821,
 0.5816143127556495,0.18203824755712117,0.5791399406303086,
 0.03359447909636981,0.14437177829997727,0.4067267199500383,
 0.20735364361749986,0.2969418935797671,0.587388860585582,0.500195765390558,
 0.18444590121649274,0.24726809624587986,0.7381230919081748,
 0.38612341095491187,0.9512837692487408,0.4307616048896822,
 0.18376738140729554,0.8979979650632542,0.7681804286577196,0.8448725812305661,
 0.014878918398528285,0.301278669677584,0.3263580890573272,0.7629299718319579,
 0.686529877960992,0.399772440337271,0.7167314736012054,0.287096045176641,
 0.7591817038594638,0.10719434876947176,0.7088218605543389,0.7138169154319711,
 0.3289674576533175,0.0274570648305289,0.7680242446996891,0.5320439559441438,
 0.09783458725924077,0.23957272440126076,0.6454024501981132,
 0.5485848053098131,0.39718444341641224,0.681554788026243,0.16848108400809203,
 0.1815347553505805,0.7948816530067068,0.32372530110081377,0.9551979846085688,
 0.5578943429789388,0.4814651598307973,0.550083931966401,0.5861736397806873,
 0.6953140699096774,0.25821846332469134,0.7306042690041252,0.3692484669714663,
 0.10910892483826351,0.5466726348294623,0.4327800307837608,0.546249379108267,
 0.1825551458117729,0.5909778902026736,0.15230343726289586,0.954550970606862,
 0.5653588782488135,0.7482299733958024,0.5339954265825043,0.6463327973210936,
 0.22813967120559342,0.3183328609430156,0.5423335277377803,0.6688720125335161,
 0.8251372127700861,0.9982355494460009,0.685489343407747,0.03385962521197561,
 0.9172250649415492,0.9432503870225384,0.7319296299167617,0.2788352563582158,
 0.7218774575412694,0.5737873289698185,0.9650194643693667,
 0.057901394028837094,0.4173738092065119,0.4457535297870614,
 0.05321443210186416,0.5258144708913026,0.32894509431294927,
 0.4936285587772208,0.14562773404679585,0.3530562774881868,
 0.06151857786247161,0.47169935720429335,0.3149649732278006,
 0.9665642554209535,0.7864557226480504,0.122772992774283,0.4243576249517166,
 0.2159729426388527,0.8635140331022408,0.4040218645247201,0.3478558165305524,
 0.5531402949426779,0.299718175844623,0.7281390826065816,0.4312678250951213,
 0.4786016857237835,0.5071693311596138,0.9144576454965286,0.04284205613581016,
 0.9718523261971237,0.40771896647052275,0.8882037976692247,0.6141455899827971,
 0.6535969512420136,0.20379976319081872,0.7917649084029261,0.7861151377357333,
 0.36592677122188466,0.3903257042259445,0.7791263794934186,0.2544135309357193,
 0.5712247474246976,0.606899027767683,0.07680408071201372,0.9238666663292989,
 0.1032156085973659,0.13307599479571308,0.4467390068108986,0.9359588300000069,
 0.4504640090906674,0.39581986940034497,0.9001014915824176,0.7046271305326108,
 0.21628705720214392,0.20191766498136143,0.1820763993371426,
 0.05016354439294024,0.3402731835506969,0.8300777137991555,0.7041481124004683,
 0.8867702791256926,0.14916741539417055,0.7473405240127933,
 0.16263334897174242,0.050953632894287226,0.853797884847723,
 0.10676639869286653,0.2861931610058437,0.3138035526505166,0.6491909424991616,
 0.06338376553815706,0.4487618204796995,0.043358222377993805,
 0.721339680963989,0.3364381474162582,0.5106935059704413,0.5865496160636169,
 0.9249970798002268,0.4624511936812862,0.839309743836518,0.9985757168046194,
 0.0039946203455183404,0.6869568115252936,0.7336333762879341,
 0.01580528569249262,0.46883807039290626,0.7086355891887576,
 0.21828978412504185,0.6475345063794581,0.29787064939242835,
 0.7390525382334906,0.22214772154420193,0.8693459202870727,0.8784312922727164,
 0.8268734707667408,0.8212189687014096,0.013843843319991112,
 0.45002804400839247,0.9423329095293171,0.05042810956978583,0.263788718777882,
 0.724360080693126,0.8468415270403454,0.5598160041327293,0.1382435622716851,
 0.5017577642791853,0.1028917969854598,0.3046450560759061,0.533741781348627,
 0.5777964640857035,0.3357921380478529,0.913214773907291,0.20865692395174606,
 0.9839861253183974,0.05653997523321308,0.6933509306750906,
 0.48841813444957005,0.6338507077082249,0.0510556987086388,0.2707716816787604,
 0.8576260697927046,0.4105638576650328,0.9021820384047963,
 0.006791993894023962,0.007254775962045379,0.5461909594581829,
 0.17497776402857734,0.055643904929726706,0.8379174321175343,
 0.747713322579565,0.1662713450528881,0.21743615267928906,0.6596525755802604,
 0.8571801554091022,0.0542234355912693,0.062403241292896094,
 0.4294023732071497,0.014300375735501247,0.4699657553025629,
 0.08705352049491222,0.46738451661456504,0.4335259279657754,
 0.10536699007314321,0.5540024423320723,0.34137874321813766,
 0.3601030046672278,0.6387945596151999,0.936973801097809,0.7914686472390348,
 0.888645216534506,0.052137088666784126,0.4088373534298063,0.99571454934509,
 0.09550282445967517,0.48296486559735874,0.8681349940203277,
 0.7646540644825266,0.8086420888180127,0.4492245525817531,0.8335152903256285,
 0.6465606399938182,0.7751875304047144,0.16591189325283995,0.6214906950975428,
 0.9873757488881602,0.5504075070376004,0.8938294045716645,0.7674733414744407,
 0.5731002061705173,0.347209572606603,0.6775912033555542,0.030615167948718325,
 0.5057328749212204,0.9625556259374441,0.16656062962420848,
 0.14040698778305027,0.2124031117681996,0.17372796555037906,
 0.6815854593507691,0.7593006724928097,0.6428771615663852,0.9499961990263377,
 0.5507435882902568,0.7312173192959481,0.03313708527218928,
 0.014493821570797483,0.15617489221721215,0.844767779799658,
 0.9195993390126336,0.6145858241137017,0.7762162405479217,0.8325844354059186,
 0.06441540242203214,0.313052667865942,0.16565338070463986,0.5174301995730529,
 0.7771488680918572,0.40096227023681796,0.13779332696217583,
 0.0418055223982533,0.3448955510521682,0.7691634761817773,0.5414316184601158,
 0.9544130175316893,0.9455882161430048,0.9391019259524136,0.8037311963594183,
 0.9683046808860427,0.35667410126892063,0.29938836235586674,
 0.25657251031456674,0.8232299851830832,0.9601788761302364,
 0.46716128946074753,0.2638981289166408,0.6047604620221072,
 0.18489923899151028,0.29698204751590807,0.6064712654603559,
 0.48325660759359856,0.5794284599692187,0.5636523651947383,
 0.24596139309895249,0.8379517041340315,0.4314845945414737,0.2561974929447285,
 0.42461040818532325,0.3097195631478976,0.3770459141481187,0.9885098448533854,
 0.6910467772684243,0.44864608163937936,0.9360005140653265,0.8193934107085346,
 0.42529778111201555,0.44425395146474633,0.41169746010149144,
 0.2658844475490675,0.9821650868375096,0.2607230002600379,0.882343287472776,
 0.666681750720786,0.7331305984445879,0.4502114732726833,0.7287869170000344,
 0.012252307778628069,0.5619401119827739,0.11332040982345648,
 0.26712414204010193,0.022331145399377972,0.7706978217597213,
 0.5793096099063617,0.6000869989411702,0.17233666347272358,0.9147359074737033,
 0.7424178469400488,0.3360247538856337,0.18889265738321592,0.5744321972938116,
 0.7483423673986358,0.38592582148148613,0.6602831243995747,0.0916018594261625,
 0.12040928725775124,0.4190221891341376,0.8151339749779578,0.7727210881663705,
 0.3328835451189148,0.9809469272729167,0.6505636233808518,0.037818794977307,
 0.3870874479126277,0.2737408107348607,0.7564542748289803,0.23851287036744784,
 0.9945359782721224,0.8491106787186142,0.602556279333645,0.2346367464778928,
 0.062495141625548034,0.6508416977362596,0.46569284408372114,
 0.11736678702828895,0.5433648711062464,0.07911564126303294,
 0.9081755715948261,0.1623065402107977,0.2222039673276076,0.20868501456040234,
 0.40804758637119065,0.4710736651986722,0.8627385910666492,0.9257019598493094,
 0.7650693772058599,0.8160204942111383,0.7269595243430802,0.8699426115489758,
 0.5962162137783583,0.34777963045921756,0.11909533852375831,
 0.4155052387747624,0.5067027917304038,0.6617867684173444,0.1458942444378556,
 0.6430768218204356,0.8441454976153252,0.9852722701646045,0.30689073640758013,
 0.7908976535139369,0.7375419061442777,0.7493723820809955,
 0.057055891166875705,0.48488092253258686,0.8991738048658876,
 0.420783973529653,0.14434319923320138,0.12862101734418507,0.5448526358948481,
 0.4797390961182364,0.5432445629036236,0.051877835919190374,0.517998971100945,
 0.75403198036427,0.31853108413318565,0.899330134474795,0.4976972964646835,
 0.19447282517758824,0.23595166931488054,0.1800823638920479,
 0.6611775135252077,0.6634693312588545,0.9604650462980799,0.8340999061765874,
 0.8229632347912867,0.9530095440215113,0.897941056043156,0.7697298196349105,
 0.5589398954900301,0.23810988398477506,0.16431819372266954,
 0.09545784868864071,0.6314720322355326,0.3257398067127014,0.7538692274657555,
 0.6960329929436664,0.2687667734920991,0.3561154487453093,0.1491769676884216,
 0.005882583330587665,0.1974571680441365,0.8602717684860294,
 0.06781160675411146,0.516645713526545,0.2503385291371638,0.8319719793157878,
 0.9280938381892816,0.6778545046904798,0.8486059317885588,0.6497259696884315,
 0.8507199015031081,0.6633040742127965,0.3260706975578601,0.9866532097843713,
 0.14581291776522542,0.09968432209605194,0.35678880876040153,
 0.011700093062501526,0.9782884398491954,0.849030167635551,0.282162461829518,
 0.8545435167669497,0.19186636005345814,0.5801723154454889,
 0.05987125107118285,0.8299385126061601,0.5813477306776882,0.9637161743414315,
 0.7253394455776566,0.1651315859737349,0.15811666982163464,
 0.25634084832698734,0.9922727388528974,0.7637847600842169,0.635814818541781,
 0.06878568927794981,0.022124457321022084,0.4951913402301691,
 0.9357686447390604,0.5076894387013041,0.3342708667002595,0.07149401619025653,
 0.9692759815841483,0.6946865590251214,0.4475943164487659,0.9680380041628646,
 0.2264668959505607,0.10156868452106094,0.9460240677860805,0.6359868118129203,
 0.22525691974234796,0.6100033097946018,0.8638744269573356,
 0.12426495889172096,0.5310084575440716,0.735437543737255,0.7798798792076437,
 0.24333552126481683,0.35479193969170886,0.8122318132442242,
 0.09819820422532799,0.43265835491022964,0.27207937240292734,
 0.5180881384302616,0.10071733465794586,0.10513177670331009,
 0.29340099728608204,0.8904259817977764,0.5913206208497818,0.5630743634034676,
 0.4591892277623779,0.9795546202813906,0.15241062268341765,0.5331275088460907,
 0.4722956391622951,0.9962703930985889,0.2678819763558242,0.9529116969940289,
 0.05814549788323464,0.5464841167432256,0.9861683970708239,
 0.35748210175183937,0.7156976635448901,0.4428887783449419,0.9568492079582961,
 0.714439928775945,0.14638318266608197,0.09230829846128408,
 0.17754465219786986,0.8788484381623286,0.8992835245864689,0.2071580687923119,
 0.5044888314737374,0.37604317493196737,0.7391784259368457,0.9583970807803927,
 0.8578998866935099,0.5805302963127557,0.7841569790855523,0.31648476191464303,
 0.6476540499180855,0.5612647516257387,0.28383165008375844,
 0.28796948361654984,0.4839445507324983,0.40144531290366736,
 0.6271231141687468,0.624452184833631,0.5246632233539492,0.2332371431311847,
 0.13245375199702902,0.558832282185638,0.6822234657560089,0.7461722709716402,
 0.20191674260689185,0.6502100340593617,0.5696712497629429,0.7103513853012376,
 0.4293900194762922,0.7656261503289206,0.0814645664906204,0.6599148999156595,
 0.03867571913704104,0.22373916937749494,0.5323081008080742,
 0.6227071511826193,0.4657061402904308,0.2868460274354365,0.727738810122894,
 0.7748506404017004,0.31203808165048097,0.5488124206826421,0.3168888873449479,
 0.6120313499180771,0.7172039930691368,0.5873440826641749,0.14214920980794266,
 0.66287902560592,0.7789147382278414,0.9195714548413905,0.008268532888640823,
 0.08401703122178239,0.7330254183944374,0.5023545948815084,
 0.11013154107192868,0.3015649029939932,0.5588592876533219,0.8365122792858822,
 0.9632665324243096,0.5431268998427735,0.7904875688998311,0.735073414888236,
 0.7694551911431549,0.20284990843289352,0.6602611648826117,0.8559065135481119,
 0.33927617499283946,0.07167592610018625,0.6287127722897405,
 0.4431648346637884,0.253922583994064,0.761682893452324,0.675630417558753,
 0.05725649329097471,0.5973593201894044,0.17434185427334392,
 0.2404837049719688,0.45288621063571766,0.6250238270797372,0.934044933658424,
 0.03154094179272726,0.5380026570755769,0.419183892938525,0.10855936638016173,
 0.35286030566795723,0.865286990558499,0.11041314834807037,0.3825212367304728,
 0.44348321277667335,0.22115204873188954,0.5328232633346038,
 0.1233523992008212,0.46255382995501537,0.24349086676398723,
 0.5166207326657493,0.7976578814627252,0.4684465068282463,0.44255008425632814,
 0.962765323529295,0.467120056832353,0.30824937170516153,0.6073644267294434,
 0.8003940373978662,0.29680338343789003,0.7965404283116685,0.2196670971034247,
 0.30976280836906844,0.7927756193479942,0.42197056494227614,
 0.9608311096956883,0.6390679544334313,0.608267929680595,0.38524495138869486,
 0.2369161867585844,0.7130894952912301,0.3608391895054015,0.12467137693042818,
 0.36659442878473625,0.29430659416529226,0.4984242035333075,
 0.7263611718601408,0.12153107479680925,0.20283683297580402,
 0.9618159147200136,0.6351740799435653,0.657007871345041,0.4241648991502277,
 0.9732801848391802,0.4316360946998562,0.7028061074999423,0.7934117850322673,
 0.6678775697774445,0.7585790483004899,0.8092662227676599,0.5425818095916776,
 0.8480909403598176,0.4826914712367807,0.91229623151426,7.67837389764825e-4,
 0.25618968004503273,0.9367305537459671,0.14243736839411714,
 0.9769275635854872,0.3019165632026327,0.14861709811887225,0.5464226203777937,
 0.23739837229804361,0.5076104788976926,0.8443637460364187,
 0.19379560137437424,0.5249114929203937,0.7858715317240336,0.9245973865218972,
 0.31708142041328147,0.9462348110627061,0.7710613133215588,
 0.25101322319843733,0.622313825284804,0.9510215151034243,0.48074063809672407,
 0.5007551198327195,0.45664943418822246,0.384870781366785,0.6799223160986628,
 0.8504840631920747,0.17761578118388988,0.2316025992636307,0.9178404129581539,
 0.15068901991884498,0.33449897800449646,0.3245920744673868,
 0.23051738456907145,0.7622034568085901,0.2119610984465663,0.7506949737991934,
 0.5952301610229318,0.39133969623196685,0.46099427835870843,0.45571174654323,
 0.5015250557994052,0.972491200920707,0.8852195662586659,0.1033303618707423,
 0.48538795548571234,0.8034742749900935,0.5981748647182193,
 0.10162925493189334,0.6177534161457081,0.5371080887151285,0.6527761375761055,
 0.6595586090132504,0.15493840489354072,8.804772937454075e-5,
 0.40272609024127415,0.4325923059271608,0.3270850556114062,0.7921241076424284,
 0.9755804202998889,0.5285859172149492,0.5306127597720265,0.4945036253740078,
 0.5669300015447611,0.20795609985753327,0.8822321857902603,0.9410859456828256,
 0.7386478849962538,0.9182364757602712,0.3104160480255964,0.625827310486344,
 0.6831044847729505,0.19557407734183552,0.9132862183350043,0.9405758317661409,
 0.6929508862015219,0.9288090969970928,0.8979075924944964,0.6861440077115403,
 0.17997689149217577,0.8124622915173337,0.7488386048595519,
 0.16357181355552686,0.06792672053891657,0.0720856490230336,
 0.9355827306725514,0.8778058301974341,0.5738755675341145,0.721501975922346,
 0.8165875592725204,0.4844147686174184,0.20464957865841737,0.6038260561542912,
 0.802115527526095,0.8925175374950829,0.9170035769386736,0.7864606045719459,
 0.22527982409681757,0.10327042210487491,0.39986585491648563,
 0.04510173169268883,0.5434542847530325,0.6267337468277703,0.7478346384992146,
 0.5943298343086999,0.38518394532712186,0.2512981773837897,0.9423847669092382,
 0.5686404084832447,0.22481478053429615,0.3902813444775999,
 0.26391471830452107,0.40583545395026177,0.9393320473668182,
 0.08649653818526704,0.7808559771406526,0.06503316754752564,
 0.22551608749822483,0.2526790778661645,0.7847474669659346,0.2862203119694927,
 0.5380973490097838,0.9566513532812784,0.14290151863331613,0.7499999134241712,
 0.23935371471407862,0.7857566218513373,0.9998357255303211,0.9743303512619951,
 0.5298982862611435,0.2695080928320043,0.8194671286777209,0.21308857895049238,
 0.9804930986713709,0.7256582537864584,0.6264996025574731,0.7147144433066186,
 0.7701424281960394,0.6212499823974773,0.2603507540837825,0.06803748428545153,
 0.1137817115486015,0.2239454747307993,0.47435145453116023,0.6604975381467496,
 0.8133714656326134,0.44553335536746996,0.9944881634987828,
 0.015750990261291897,0.7212463961163479,0.782448033569555,0.8863144020817648,
 0.7603268398752525,0.045252045591836376,0.1345117403291326,
 0.2510177515232431,0.1383418311491571,0.6865661170010846,0.3456825471092917,
 0.6622647122864168,0.34792358655787303,0.8698462733307047,
 0.40947797251323803,0.24797525951466748,0.3086049784668373,
 0.2981068023662965,0.07221329545577937,0.7685531880690091,0.7799741314664066,
 0.43342330690376274,0.6464727349434629,0.09798187482826459,
 0.9929186124171371,0.046260302663500785,0.048109310245674664,
 0.5182788390404386,0.22639588672728783,0.05536640223082989,
 0.35405067903044973,0.12334650483400411,0.9246693670643366,
 0.3511060568348987,0.09428518221875215,0.6623787618908034,0.7205728675263955,
 0.7574317280429359,0.6855084820082822,0.2084434270616604,0.9808452511316902,
 0.02427911584133513,0.7200256957950759,0.907122206852108,0.6753471385198191,
 0.8827400479685438,0.31562750341168033,0.3677743319479835,0.7441339622658987,
 0.3733277019854323,0.32163935951696243,0.54103974481013,0.6265564762215399,
 0.2153742127823992,0.6072760068167439,0.3021069236151749,0.8779769478641782,
 0.7095947348624447,0.09473093587132375,0.0759435800562469,0.7363317595614177,
 0.6701819664657688,0.6433777972292557,0.5142814982490762,0.8470333211810274,
 0.6236640681282246,0.4018382094071955,0.3726241925674707,0.6849592823853172,
 0.7203967979451333,0.42189921820227283,0.6895757512566723,0.3908411808183939,
 0.2051204649097622,0.3949862167325989,0.22317511962524095,0.7300097572425482,
 0.9232952542162243,0.12524340292519343,0.05629460284883425,
 0.9391791790776874,0.41861540010730625,0.1288451674415957,
 0.33886127358722284,0.43963244197839946,0.6446213421049097,
 0.8595477397827573,0.037023727198108425,0.3705614433320217,0.893442462430444,
 0.38715825284458627,0.47162755270282397,0.30699751333999115,
 0.34053275074570144,0.8086636891706186,0.1486977821897162,0.8621801850316557,
 0.3269751529654753,0.12316552886746124,0.18525554359237095,
 0.2058943096442316,0.4724011150403108,0.48375383080583734,0.8928821627111427,
 0.5247028551530731,0.2836129207400676,0.23981356240919782,
 0.05256122096315885,0.7771409908643655,0.6311350472180551,0.7033336202992213,
 0.6247661147278818,0.355809946534436,0.2084602397630938,0.5204535997568033,
 0.7117785966836041,0.08682423812761131,0.007062679186989218,
 0.25201202280375534,0.8810473467283245,0.9811624047145849,0.2877856147209482,
 0.5768624414877918,0.3636659392332646,0.511232986477423,0.6809728401500312,
 0.27589716934457187,0.13701018571410462,0.9056145238149145,
 0.9054262523935979,0.668030402618454,0.449946151745382,0.7281209287050737,
 0.7204332775991994,0.004946249596193164,0.22330340841003182,
 0.4328829177252518,0.709193902422485,0.30431793892981307,0.3702711560085823,
 0.9180514375920151,0.6708992097535971,0.3215035559891245,0.2678882202494597,
 0.05351508652808512,0.0355246175245163,0.9171180569818722,0.4425856030789679,
 0.10582937357343647,0.6603884452528358,0.9071175134969357,0.1456438424993336,
 0.7504469165964149,0.09666284567194272,0.7882044174131946,0.6981483462697446,
 0.3723574269952351,0.2569918435940953,0.8719279490690095,0.1869527876410334,
 0.9370324659330975,0.05995706714782745,0.39859816551971405,
 0.11873510102275209,0.9785262770287384,0.02027772999735533,
 0.23646718409284984,0.5439740009113624,0.18685784372360736,
 0.7843008896090627,0.4485562682274128,0.04696702305760492,0.8666959599381682,
 0.8710040197502962,0.08142959111052073,0.23784177901956094,0.591763103315749,
 0.8535917720907418,0.5753751810988205,0.5594521574372556,0.3166552973924477,
 0.4649153410228205,0.9774081199170302,0.1598376523860201,0.6454672298988869,
 0.4092551818404401,0.1052611217439906,0.6022119948203623,0.3921438595645361,
 0.5806375855866563,0.6992523289708636,0.5325269194928978,0.15284717187880303,
 0.8361324738160045,0.06033167613555923,0.5535506036850717,0.3568634036439179,
 0.02471767267767544,0.8539857646385888,0.6180098628277086,0.687213485296819,
 0.09727061929796854,0.2932103712964601,0.18309757015056227,
 0.11343689083637365,0.9072058625872602,0.930942314475148,0.07960958986684163,
 0.007665889904150736,0.08186314719116539,0.6931089430298771,
 0.8648910844924471,0.9325357533930146,0.44059771791910096,0.7307175458286214,
 0.23696095142265017,0.2364059400940144,0.904219785167207,0.3578970952541676,
 0.9167260668542335,0.9531525380774103,0.21261772432644144,0.5396236352700382,
 0.0625356027209395,0.34457479629251553,0.15261705311928386,
 0.018435213223799263,0.42303156896086513,0.018709473413574207,
 0.4021457592122265,0.8468708098466042,0.9065904267780206,0.995563266463822,
 0.10993222454723606,0.5869086663914689,0.5270098525422375,0.3561255855248793,
 0.6971854036413595,0.44983811359438497,0.529555824612292,0.9886601315541074,
 0.0037922838014329274,0.9018890054881901,0.256478327265808,
 0.5497346939136188,0.09471722479167677,0.7386042825003484,
 0.061738808549555824,0.24956646425407303,0.376649571259019,
 0.34331225176767277,0.15738103743942466,0.032553404407524966,
 0.5676641037357258,0.98722302243441,0.14252278989388367,0.22530235646943741,
 0.70709158907987,0.5086312562809223,0.1808398468593152,0.1218653083391682,
 0.15725987827797128,0.2774217701423166,0.5777217268623749,
 0.45378654600113777,0.30505072597791294,0.33186613625816175,
 0.38547384588542377,0.9600955387801126,0.23629648819490046,
 0.26825651971050724,0.28941409393443784,0.6377568017724964,
 0.6265918426264733,0.5765383477435784,0.18569093099452338,
 0.28996449844458194,0.7553631868548344,0.6521513960483725,
 0.14599246431701074,0.84567859127405,0.002222491925048087,0.7709519837736558,
 0.48150106660725,0.9613268507237431,0.23937439146700978,0.41266320589893746,
 0.3843935956175194,0.7118597826035756,0.2846132143690263,0.9122371872268501,
 0.3696177976992332,0.8460224786230088,0.8273502445715311,0.7933135140877572,
 0.3099869158570595,0.001532564663745184,0.14673667343503427,
 0.27349334013488336,0.08137387454327272,0.2473382344888203,
 0.43666270500125537,0.7691332243735863,0.9195747838061239,
 0.006790837262572791,0.5862432013413392,0.644343271742714,0.6666779699180307,
 0.26704356404270313,0.460667468670942,0.779820843181635,0.061796688341612005,
 0.0688090774648209,0.7231922110228748,0.03804356178695323,0.7974141107306475,
 0.5261372430015713,0.21751836110880807,0.8106241595334489,0.8528122747014164,
 0.7023569965145794,0.5288730573009894,0.8888469288183755,0.35760402698781213,
 0.5098859020167583,0.13253529652518004,0.8073082308852985,0.8114351368020803,
 0.50119317519858,0.7967791642561848,0.17785289136944116,0.24878105253815685,
 0.58903295420587,0.061066403523180046,0.5860591210814531,0.04140694267801759,
 0.5245382445046995,0.738087773496245,0.8943404598305671,0.7312100718387838,
 0.650152290374805,0.6498472914166366,0.5258255356860831,0.3800715947652644,
 0.9653656507654503,0.23913655699627123,0.8196813140506473,
 0.12537379326508213,0.16515668170623043,0.5055620101025263,
 0.5029457102004699,0.3606742241046228,0.4210420470120012,0.3546740676017359,
 0.14549381944435613,0.4900757206257087,0.3706127756316808,0.6121269421736739,
 0.7539352763213437,0.3924571518128537,0.7458322924159825,0.44656717390423917,
 0.2299360838428346,0.818159945894799,0.1436753930789343,0.24702768886380932,
 0.006445476884289425,0.7226512425040265,0.6898236780397323,
 0.42169669724673386,0.47231430500881744,0.33581712957013565,
 0.6691359087528365,0.3973076322151764,0.2489266032470001,0.6174782608093605,
 0.08296947040477076,0.7930164709654713,0.22269252462192657,
 0.8928785462705712,0.8376869444594973,0.7667962330246353,0.2503366868136596,
 0.800310654900547,0.5901673687001212,0.3854875334540272,0.762923231915379,
 0.5232058452662814,0.41146307198832743,0.10710052153780802,
 0.5157176901480427,0.18741755883718414,0.8778375083648615,0.5329900425752149,
 0.09333499631840625,0.4554385445790782,0.9633381306362416,0.993098001123281,
 0.13789868201153577,0.6621210551751039,0.3904937113854178,
 0.36737733163627007,0.9385224327564733,0.5702372023861135,0.1240501989208449,
 0.14041631643169428,0.019800516067512097,0.3199538037282432,
 0.8283214911409832,0.039846483525465715,0.16634552046247242,
 0.18979507931060502,0.028344121342537165,0.8364362697654251,
 0.47352783253567765,0.681290804418078,0.673811693403808,0.2436019391101074,
 0.36961690537366887,0.703427136518011,0.21565632624255093,0.8988025671248898,
 0.9942986558440797,0.2809052547706463,0.5424710223633709,0.6037249731618656,
 0.7343887982442797,0.1134657551329894,0.5388979560000525,0.3430950405053851,
 0.2861113056365677,0.4856513951034539,0.016332712952781603,
 0.4028378216759626,0.7149685971905745,0.018981627681328272,
 0.16117587320350513,0.6670648610899299,0.2572689552856364,0.7981047136664182,
 0.8959842954566967,0.5399750177894815,0.06341249123281534,0.270370849864942,
 0.15400133108664393,0.41209672539364806,0.14239001023886144,
 0.5792883651354408,0.1776701421475798,0.9246205675562884,0.6021603969210464,
 0.32719762899619276,0.5812500879075702,0.6005340236991235,0.8776256683348499,
 0.13299891638239725,0.22549395253941085,0.6363214969826076,
 0.3480091162386795,0.757069608835031,0.26644249789780716,0.7641862431625536,
 0.38765899096983547,0.024253586879890876,0.6884861306685756,
 0.7729002256739089,0.9515561661304831,0.41934489066540304,
 0.15533945275993977,0.2367193015800404,0.847843912263499,0.1898313128771526,
 0.9702169458583078,0.2404560861263777,0.7385737724519981,0.34254252873784397,
 0.6600871404070525,0.7525745813461233,0.4874268205606823,0.23334787922139988,
 0.7221434959851821,0.5422672512599964,0.8958102133758609,0.4127184763130576,
 0.15073814420967913,0.7864957991268449,0.39368243105160405,
 0.6670706957471416,0.26809257198876324,0.07281743562297216,
 0.7962668596741652,0.06585257457602745,0.2086521301484182,0.6864531738213039,
 0.9097396914429787,0.6009214906035122,0.9444295795867973,0.2788474863807957,
 0.7684366189493552,0.261663261743055,0.1935112448870746,0.7069464135958039,
 0.5996399532468321,0.5515668340477439,0.6284490168522205,0.18190961514687676,
 0.4777687226275016,0.564060207367793,0.2504244841869012,0.6160621671915303,
 0.35593978424898964,0.8664336214820251,0.43851513963595234,
 0.3659681092772792,0.7921054392228749,0.9566057900931472,0.48425920305469705,
 0.03074218881786528,0.8535169051980113,0.05013024255093368,
 0.7606729645395307,0.9962932033287846,0.7378940407863315,0.12470527838999379,
 0.9289456631698483,0.6797252414629427,0.3525752887941641,0.1292484565878853,
 0.4467032301924764,0.4790418434130338,0.12149486096200557,
 0.15267203967783904,0.4610905668037072,0.8792896552387116,0.7564435697065035,
 0.2305817274473838,0.42453462654291374,0.28358036655829744,
 0.29793207313809167,0.9605459890449078,0.43023880251674473,
 0.3024065664618206,0.13809242401145905,0.4747091209697668,0.8438674618241137,
 0.7119442137766134,0.013991025620212483,0.6718490222225021,
 0.9347088991144976,0.22522556656412662,0.5690195180479976,0.6566248728019962,
 0.8662831749816212,0.9880360837752952,0.16799106816826048,
 0.022774637371752116,0.5353598667427572,0.9517395615551888,
 0.25970184297186893,0.26205017751266246,0.5039468502500283,
 0.9129072680537722,0.9182570865722527,0.7400683381814808,0.3272678810445493,
 0.2497131126805725,0.21825058393477792,0.4517851748722347,0.7820250692017859,
 0.030722604070295922,0.36368946477912756,0.852581544227537,
 0.25204557829175167,0.3879335438173802,0.19280976271903483,
 0.5074536143143125,0.49477202835191325,0.10003871880053916,
 0.6639184166783657,0.6769310019884959,0.17907999951771703,0.3205901971571735,
 0.0944680905738654,0.5296907280584656,0.6342394421100318,0.700045394102095,
 0.4968075056262432,0.27068725529298,0.7271221072677174,0.7525402534004979,
 0.46784995861905077,0.7956098279900342,0.5503579503928637,
 0.35540203821709737,0.46270527360701375,0.2697768009132714,
 0.32844004055587706,0.9994664200249919,0.14018459983346832,0.15651384204149,
 0.13909361624494887,0.537019559598144,0.4002743817695045,0.2692956741372944,
 0.4702284371597305,0.9229927412306966,0.29355007745626005,
 0.20297514049655418,0.8784848606649549,0.23147633152998182,0.618740471748603,
 0.5675013585273717,0.4763473245958628,0.7946970247889399,0.1541965909878782,
 0.9703081516970289,0.19168611506513233,0.07305764062496112,
 0.3624189009046329,0.13768547142808818,0.06307025612864325,
 0.5817643389040005,0.7545756773198269,0.028290987450600547,
 0.26149049736575647,0.8120451109322958,0.9327960278683721,0.7051776161130503,
 0.4245290056717317,0.7381587768264811,0.1041047718470236,0.8551423524078312,
 0.708684023061109,0.4224670216929177,0.08275547916648118,0.8848027100255346,
 0.11453091256493886,0.45905638855813435,0.9774372374990232,
 0.46663088085576654,0.011271620814385819,0.5843277989546303,
 0.9471378681385156,0.8639431030439262,0.7239900578039101,0.24811630419915043,
 0.16601275530043647,0.07579096226866489,0.39097696851984065,
 0.5433576530795313,0.45495491463580784,0.780306788429583,0.42224379427022596,
 0.5767942287925725,0.1209685048944028,0.4180925648500534,0.4000837858555153,
 0.7940464157065151,0.03424411006778161,0.9475423341234256,0.6527690540298072,
 0.8387977706500138,0.5859974415064321,0.8098457591770536,0.5673704752304823,
 0.67640080906051,0.948119281312887,0.6733885283084903,0.16746673227941677,
 0.5055432035846995,0.31160654010495337,0.21826909854834775,
 0.8150477056177784,0.18409387305349967,0.09231033963730206,
 0.5322750590083963,0.5731470158894564,0.16795142974681032,
 0.12410912490867276,0.3691047089078895,0.8230528492645688,0.6827972775012928,
 0.8605989163432445,0.5968887222835031,0.02390070774266162,0.9723064700650644,
 0.050078035858741154,0.8763044305940317,0.16456986649793504,
 0.5771901780161215,0.7847898823541255,0.5302350464701512,0.425639983509033,
 0.18063910769033276,0.534896289002663,0.6343817622356498,0.49867258398185554,
 0.9255130712177019,0.7689557606282249,0.6610462013403667,0.15464572687329503,
 0.0383941846363729,0.9897467101589712,0.503466126868212,0.04294682468937583,
 0.48332455111519357,0.4247261789412371,0.6409812383975343,0.3204167025833824,
 0.47620212021247854,0.02116120132580679,0.8219033185451917,
 0.9707884139142713,0.7716817658086313,0.08083020180764999,
 0.23149951495189003,0.6919894373634468,0.78249398016399,0.75099639262598,
 0.5494922840140883,0.5686845471426932,0.3119228294114458,0.8799559289004795,
 0.7738630499917448,0.1418426123097346,0.8700278962028289,0.36675220204798176,
 0.8370589397507098,0.1942046295512001,0.7595690179183501,0.7895823016469291,
 0.8371150684341064,0.4248914847149384,0.47596793208997745,0.5767990230929002,
 0.30001946781330113,0.4155694944796403,0.03841891779576767,0.917796866724375,
 0.8122151826578394,0.18251349789742488,0.9218657768420967,0.523650043948235,
 0.08541066311249201,0.3713309656747782,0.761967017365772,0.40639072326287173,
 0.6045877256499752,0.26882972334735233,0.061460293230953145,
 0.3477900744278659,0.5310200474004589,0.2352908204606855,0.9131383150857676,
 0.3781838780202187,0.2514856994231265,0.268983085234733,0.787721807189513,
 0.3054599945534676,0.6467978398372872,0.979741601449916,0.9077460926302848,
 0.5944747873862168,0.40952573258297686,0.1638386439470786,0.8661516069708491,
 0.8335926256324098,0.4221101014813994,0.44491580079610515,0.5351788779494255,
 0.672956502687478,0.043071957865558286,0.039918157144058664,
 0.2871606147845035,0.8300140086784016,4.0187374284195165e-4,
 0.5088948663359141,0.4905070958566883,0.28205693708067603,0.2761240244230292,
 0.8167101638532696,0.196782209019566,0.9607784422314444,0.03264126038563153,
 0.86426179194604,0.6924470146318387,0.5321326255129573,0.3361492682725642,
 0.7083677356167463,0.42109847842407944,0.704629811283259,0.5867467169375242,
 0.5315673013803823,0.2753216974254249,0.7920357549538832,0.9512397298781727,
 0.8456888998693013,0.0987897695229163,0.45177405636953416,0.5574473270901836,
 0.8250502655614472,0.14790932834077564,0.020211931707241337,
 0.8653455660624572,0.15119592353962596,0.8767353484504113,0.5647061434794752,
 0.4802860645764233,0.19830467019430276,0.16492518585869376,
 0.3363464025702816,0.2467677199070385,0.9699351623037389,0.7212205858610167,
 0.7462322808486842,0.7647975661204268,0.37963433677516056,0.6646679976117458,
 0.811043442072395,0.3537328241129939,0.6731239345723438,0.06503474393698916,
 0.024559164040716897,0.6430154214343451,0.013795349052092787,
 0.8795942425865795,0.015740516247937597,0.8315571576194284,
 0.3230489157117008,0.7605725238476894,0.8810381656115309,0.8728545768668572,
 0.5493021720073449,0.8095706796590264,0.47441208182879735,
 0.21502431663392596,0.1336785595528508,0.43815753830627724,
 0.3514256657157797,0.7848172011517284,0.7452920685199539,0.3882627630402299,
 0.8332780201171022,0.13163380158123117,0.8131670434281699,
 0.29206602181972374,0.8477906467306382,0.895432504880455,0.8523192321425499,
 0.05713595538296656,0.5079837407482708,0.8302787871575907,
 0.18908262062002235,0.0021722152066441236,0.5382155082650601,
 0.9871084018101763,0.6387813562331134,0.5658262133854894,
 0.015654796327666753,0.4008004720415592,0.3200932119502713,
 0.5844623058431301,0.6026167832218389,0.10912693345462854,0.4408527633653806,
 0.34451538876240717,0.5762877177987434,0.2313893510490519,0.3471712745492257,
 0.5498290861609793,0.5421416929405486,0.4512038816206403,0.6475326157881914,
 0.4971500387951969,0.5302162166753416,0.10497117899280406,0.6979129322437921,
 0.5877079040147141,0.19080995738213513,0.1483452291287457,0.5803450522871829,
 0.08627455832519526,0.15295307731168584,0.7981530010863264,
 0.10154275336542118,0.19800801488375241,0.6780989000870189,
 0.8368574327087674,0.4285913711844682,0.4478448896505358,0.30230577206763265,
 0.1752001163188478,0.923298120958326,0.4878396127291029,0.6707035872340077,
 0.08930257868326297,0.5936646400910741,0.5209344043290268,0.4422663250337304,
 0.15867722524444616,0.6096139377898571,0.7291254132440967,
 0.02082187328659013,0.5193215913944108,0.5384297729612475,0.8358763521170556,
 0.7453205799656076,0.48597581077206464,0.6491382460563696,0.0514909796078179,
 0.863333483033639,0.7310004788985357,0.39552444852612667,0.8321855107953156,
 0.14355657218481488,0.9173255580319253,0.6031181888024173,
 0.17650153459477946,0.9230880049181949,0.4737995185559346,0.7054284080595691,
 0.8824656458801067,0.011184149243278152,0.04782486522306584,
 0.5816515725724442,0.4370490839789952,0.6589491557262264,0.7901652781625041,
 0.9932789755855982,0.663270263946136,0.6314628077489298,0.6105482592120797,
 0.11341226866922283,0.18611017808885022,0.9022729402914915,
 0.059614157960197756,0.8278134717641508,0.933518169991393,0.8994124194850431,
 0.43743357543868044,0.37431081486769247,0.966137525890301,0.8829426085135811,
 0.7254419105586392,0.585267749556127,0.10457900702661871,0.7784057223000033,
 0.6654364975140997,0.5358769434155619,0.40558079714958184,0.6423407658800652,
 0.6740183702086687,0.061939147979932496,0.6325519356260898,0.667478818242554,
 0.525835034427681,0.027498321179243335,0.8425474730216566,
 0.007517872315148666,0.09534570523511587,0.7737953119966766,
 0.10584448296155247,0.8926470751215084,0.23771096198407005,0.806894740734572,
 0.2498497915443748,0.02197182816542642,0.8052598986783851,
 0.38212379025441723,0.32369493374007463,0.18593115027808593,
 0.18672028164849352,0.1138823872915633,0.8410131738994425,
 0.12648216187662875,0.4490638216777115,0.5284952870858785,
 0.43329768992203477,0.05677657020148763,0.41228505136952676,
 0.37583262764913894,0.8637322783123748,0.9647063814455092,0.5061339180189947,
 0.21897225801501763,0.16796601676532108,0.8577223820734077,
 0.5172636426010205,0.3721890497055943,0.394679430906994,0.8456886695765496,
 0.5542774965067817,0.5169065089760096,0.6089710819783934,0.03243332808750865,
 0.42583755701749204,0.9220081160682432,0.4999466412281188,0.2952285333902046,
 0.595425084666865,0.06730242469650938,0.06288612507775593,0.482593935471341,
 0.633049291795367,0.5236530774038739,0.5405925328731896,0.6330463047375836,
 0.9413426367589409,0.6074766795979444,0.31225018765783796,
 0.38178037316840796,0.23350555859887456,0.3392684514894647,
 0.6437007181016987,0.2735220664757314,0.6715682588472667,0.35990384120672103,
 0.34895821656646453,0.23130929902747432,0.748524937892354,0.9442578454077918,
 0.5723221899773461,0.2810007576172795,0.952257495877346,0.225059179005352,
 0.6447494999278559,0.11009382897331255,0.825513106492973,0.7521552004003945,
 0.5391864047928375,0.9823341628704119,0.37786794955345604,0.4276718893352498,
 0.2932186157706749,0.1805289948963158,0.207517864912929,0.4762800247575171,
 0.3921423833336656,0.10905233215612598,0.9689813973788921,0.1637000336469986,
 0.4348487483513609,0.43321919934970554,0.9433967909149379,
 0.00429248543911287,0.7451444730785467,0.9076126805133948,0.8826187462536715,
 0.40130526256257315,0.3552321659721044,0.5263927678656174,0.0120009583977565,
 0.7060010133744242,0.49100227854052536,0.18367640610591085,
 0.8427054849575791,0.7923619640309228,0.003999438783678477,0.727854844942934,
 0.8313985609512323,0.4741834773065732,0.652818472235504,0.3994688071770994,
 0.22593586481349237,0.4373692683993382,0.7259619490110828,0.7478450114993598,
 0.1015996091192033,0.1668568566458859,0.5480685182487879,0.5480538381123691,
 0.608871189932418,0.4680554830679554,0.455881841576364,0.09031884673410939,
 0.0496818354392331,0.9810481608075617,0.4716498176853414,0.5749975274209761,
 0.855965814080961,0.5995687514569755,0.8641988049404986,0.7670274222717521,
 0.7857726269615546,0.6205411633305586,0.4609118389665722,0.5685116622797345,
 0.21076245914769187,0.09352225845637707,0.9265045764642142,
 0.4555492007077685,0.8026700255337688,0.4282888596506197,0.4909107550073497,
 0.42171388421612044,0.9295378453074651,0.5727872988424055,0.6968235306393591,
 0.6488435859843933,0.6949893479366648,0.23483534961203778,0.287466118984945,
 0.716116513508398,0.8496937365196198,0.06220838050495614,0.4291292878363905,
 0.5012138603321065,0.9036549614239802,0.12806423043315285,
 0.12743430399081124,0.43268431349603853,0.578733153100838,0.4565464902854217,
 0.9829560203283483,0.7117559664153554,0.9072862321238714,0.8367598351641509,
 0.9431818026750702,0.39369365853357075,0.8668549832924048,0.9151670820482007,
 0.2712308176191003,0.8121134216131631,0.026509702106449096,
 0.3886168337606035,0.5940142273121307,0.1454682109807348,
 0.016992196699425866,0.6650642700604013,0.21577338450127637,
 0.14758651536476808,0.6073756466692061,0.5002188502048674,0.7228503108555665,
 0.2416655321382335,0.41373257678268205,0.5678595371286549,0.6151447953790825,
 0.8660107498449724,0.196559812158289,0.598174408750981,0.5369185745969273,
 0.45330448716484073,0.39727137010844105,0.9005012660818479,
 0.28056167589411896,0.8457113284126407,0.1878968764519352,0.5500516824989924,
 0.04680472207929165,0.1041678742584754,0.8901884142951875,
 0.22967096650126229,0.8812340959425966,0.8865670151949396,0.5445892173992383,
 0.8770601617801391,0.22973634672351106,0.2669149252526347,
 0.04006018788463617,0.11415665384657803,0.922481664131788,
 0.31793552061747765,0.004643372791008682,0.255769147237512,
 0.7827434660525109,0.6722925003952103,0.7916953870214075,0.5565088993766117,
 0.8541543335074123,0.446117979605108,0.2304200749362304,0.19545815601501038,
 0.9226598502096719,0.6160496042598067,0.8721718835100924,0.28551383678160447,
 0.28943091886232875,0.7001353316426171,0.10596065966120816,
 0.8861294751870755,0.08837064437938436,0.053749004402188105,
 0.9082726559951549,0.5314480598918523,0.7347873539562605,0.8323217423321672,
 0.6099174984735592,0.43891741787160443,0.017064465623082392,
 0.33301774648833915,0.6347521226670669,0.13390952223156116,
 0.17464292391118452,0.3273254043004239,0.6296170625136681,
 0.18916005998831587,0.023168508143580624,0.5484540286539759,
 0.2761437461151427,0.6636438972771771,0.15682055480278134,0.601989235357094,
 0.28094861519044423,0.8514089465665533,0.7412356197891802,
 0.17768330387535802,0.9043199566303142,0.6128079491148212,0.547216853768481,
 0.6629695524390978,0.9975183024676106,0.06578415047015618,
 0.08257421516061114,0.16881571903600623,0.9613459215876974,
 0.07025357729279447,0.14113260985669995,0.7163326354241913,
 0.08156856402912704,0.6164991577858594,0.5294248691917988,
 0.10534578812851925,0.953544442466739,0.91752940005298,0.23362650988342348,
 0.5151181146158796,0.16374848079830906,0.34920583158653706,
 0.8551879998761148,0.8472719337147766,0.8569087021993133,0.49056353545301523,
 0.039116980745279584,0.7822435292612909,0.20168827380594823,
 0.24743221711984342,0.3885780414625656,0.9593889161586358,
 0.031980236581340415,0.2296041132473582,0.8790636927849467,
 0.2477383822438125,0.42506048825241427,0.6370622620725263,0.020919913457217,
 0.8495335440036473,0.09047624491359962,0.4328123834470763,0.8686408543707689,
 0.3577795741990898,0.03773544935605,0.04588300383987165,0.1307811107790835,
 0.05693363659236106,0.7002760680099,0.8689842766525235,0.7966557103658879,
 0.698972695706074,0.7946657014130435,0.42330441256213003,0.32151858302282155,
 0.4890567078124307,0.8071292969942558,0.8039202185240826,0.6654872231618452,
 0.6543318397974122,0.8651477128886781,0.31571163322130924,
 0.37777655303916524,0.32503842305688213,0.5062507502187105,
 0.8427309512991524,0.0393363020619093,0.6268817731286928,0.37762347655087636,
 0.31887336415078815,0.2831804134215008,0.9973505724145938,0.225050848638215,
 0.14668681221919944,0.3233022231526784,0.3870152742053352,0.6048598535221832,
 0.634084196685811,0.6325231453463075,0.9486062305867935,0.2739675525030706,
 0.9212895846876632,0.45791979126382865,0.4247130921317934,0.3278107001826699,
 0.5338040385120888,0.36424376817086057,0.6823876240883591,
 0.49145581368488644,0.4743572888726155,0.28443330621240337,
 0.6657617738984701,0.6104480194366618,0.9886781909791234,0.8329372332526906,
 0.6159741619944387,0.10922261105110298,0.17626686182439477,
 0.5009115418041157,0.9785143820177246,0.297257788021551,0.39905079840055147,
 0.4198166798441598,0.6953492824285529,0.41677857582777633,
 0.08409828834417443,0.5693895697455171,0.3796336578442545,0.7744569879553266,
 0.7721748178273472,0.9070073345999075,0.8554691220185516,0.7887943956131845,
 0.7709846073417438,0.625495836312151,0.21352909655414043,0.893700122023117,
 0.11878686909935265,0.52647599990852,0.4818560466279478,0.9909336926230861,
 0.4620736189300334,0.5333522507969166,0.7384535732264339,0.20124329980968958,
 0.21864194400668324,0.0528124548179153,0.8940357689788019,
 0.22686286888060692,0.25508436671702084,0.3658271009200137,
 0.8749703523082244,0.9024907784951602,0.22401911403562558,
 0.14243143789091828,0.30697429097445217,0.20234351223100266,
 0.7354519350586752,0.9216390603765512,0.9807124131187784,0.3403546730840792,
 0.33426807939864756,0.48324253508590265,0.6163747359012923,0.795170991017389,
 0.10555665547327608,0.49898359135208603,0.08014670638738486,
 0.029782113548657896,0.8972204159743686,0.7410745733713344,
 0.3416788413930886,0.5332603249383795,0.9575828320857517,0.8058528974315612,
 0.27240587067734523,0.9426417461493355,0.4740389849244363,0.7842961149715918,
 0.08632566567684397,0.14658855935507598,0.4122932011134579,
 0.40012324249590514,0.12156349955828982,0.2513423590491246,
 0.18554329372183254,0.05095530521365432,0.6103508479099564,
 0.7492367166957606,0.2266962823812313,0.9310861089440398,0.21552813613328692,
 0.9426642167818955,0.334565351611641,0.531085335087435,0.932716350469092,
 0.01635005193360195,0.23743996163707182,0.35828333405074975,
 0.8988411510547939,0.8741636898186783,0.8139036749628942,0.9838182769166753,
 0.8519066721373636,0.8211766147667835,0.11211410468623217,
 0.19217367152355402,0.08473694473655524,0.6136033119185186,
 0.11647994856572841,0.9197504860597372,0.20528727694217697,
 0.7654330739837192,0.8487284330914409,0.08921650726987274,0.2014246096963621,
 0.8685457589480168,0.10344345675313849,0.1469082197572168,
 0.20249127475339723,0.2510468331692106,0.7936087999446427,0.5736534416060671,
 0.22937516205246566,0.6539547365565381,0.14081553666499613,
 0.6538442082806054,0.6279768366503162,0.0032208208196694565,0.30081491284647,
 0.1478911073906668,0.5999871395273165,0.42363491507687845,0.4319679074740894,
 0.441381750586318,0.845040605538959,0.8580623951915001,0.0789384599339672,
 0.044332554904490395,0.4830801979354489,0.12028294086798064,
 0.5160464896571019,0.33908650724487943,0.3673736776770653,0.852833487225292,
 0.7088837082997161,0.6595106659403386,0.5624910509176831,0.7603529008860811,
 0.37257376540072495,0.0364852252889073,0.9303385947047729,0.3586431396100871,
 0.8709653641924561,0.11464263641878525,0.804632710603952,0.7452232501863599,
 0.21661069711084946,0.9357621814311499,0.3403083559914466,
 0.09494274153210314,0.2525481916247735,0.4742770951023021,0.7450573143565937,
 0.5989277882512585,0.026506575825509637,0.8523175296878672,0.99474495472216,
 0.4794182795958112,0.1723853219291146,0.04856972471631604,0.7450785021015899,
 0.72785877835504,0.23938402454295493,0.5082071323481923,0.9820596369015994,
 0.14740954129118444,0.40414651299121107,0.726470085181123,0.7404690846286986,
 0.5306127783044476,0.7070884397828997,0.5080339679884323,0.6447214020894505,
 0.7700069052336349,0.7222732507618476,0.7076349423744354,0.2349545984335476,
 0.1110572380609599,0.6105671602800432,0.29170617934736265,0.7942971614386641,
 0.7703207519478936,0.40641899329470754,0.6714938076027492,
 0.26078721428458007,0.2031657068769548,0.07247519293189852,
 0.6050581628638357,0.4763809545801617,0.6539437751571839,0.899049020164144,
 0.1492495644560743,0.06385457747930212,0.3283477306571435,
 0.43189667983943514,0.9577436568000747,0.5385788174563395,
 0.38084366206483045,0.07517811134299213,0.35847662706873673,
 0.31904806436052047,0.1589527850608935,0.7166007709174548,0.6506447745888511,
 0.5135387264742072,0.01742215771629385,0.9113564392262986,0.5096239821321662,
 0.8785078937437336,0.33291714346231704,0.5005319752075005,0.682265686036888,
 0.004345365659595091,0.45385585426263997,0.25656954914890573,
 0.14565780626961589,0.7776024101289649,0.2806781436746738,0.6837065842205544,
 0.6192967557052615,0.32262755057015247,0.2083120883063181,
 0.26358388653172227,0.8560050071275553,0.9024726395608397,0.5041728775218665,
 0.9622060841298319,0.12551787020027938,0.7694848416502775,0.672449115745574,
 0.6968910561208713,0.96809779362522,0.4838981096185143,0.4051323188457773,
 0.9156252102926667,0.41483064369888567,0.1655019806258683,
 0.058441025050620654,0.6600208540233291,0.8310582624815557,0.737740009840302,
 0.9282397187850066,0.7502392280658176,0.6460915671767705,0.9883469743041096,
 0.6782841955959616,0.7358301867176565,0.18850329828255685,0.7974066325935755,
 0.9689248980199014,0.4587694984219919,0.599243632760004,0.20885713955048102,
 0.3382077225868031,0.986860046013495,0.27585297672256714,0.26427728021602404,
 0.7829456582335044,0.9489851363972436,0.9580093001842629,0.8236054491686281,
 0.31946617197477745,0.42299743595141326,0.18796285927755496,
 0.21572036764001168,0.1802162891413268,0.6324443492973134,0.9938085210430525,
 0.35243536064614356,0.5767474730637256,0.12607324327316105,
 0.20501996413179502,0.47266766555632955,0.4655411702786507,
 0.47257164013858877,0.0739131276857935,0.3252362165233954,0.181763861635016,
 0.039616648384254494,0.47100726716244334,0.8853533192225403,
 0.9186680043395853,0.32021453595713745,0.834734043709923,0.8237644128020057,
 0.7085214540901067,0.38391137363291505,0.40511228054234505,
 0.8013474076295756,0.3695520370579911,0.8522055368581903,0.5353822892802267,
 0.30037773663150547,0.2576519985640304,0.5425339463135327,0.6728230989358611,
 0.45029703122928133,0.6549059604698735,0.34135823734697224,
 0.8256829604954723,0.8992678043549511,0.29659835882377905,0.4634209211772933,
 0.17830367959532734,0.6815773911965375,0.9122349063785556,0.5004450999308826,
 0.32182756601204976,0.7266063751871454,0.9174513468945947,
 0.34430995414561427,0.1104572100131831,0.3698143206356501,0.5872411848460284,
 0.30433726180904863,0.8015510064857372,0.30865407872599226,
 0.28042516557672126,0.9300101692241525,0.7701978083679792,0.6902536144929123,
 0.8066531667730578,0.27860726891043386,0.36473963297099266,
 0.41970829534380427,0.10581692492664019,0.9861762251637847,
 0.6113804718874376,0.4439819725506504,0.8209270915832523,0.12326450872273365,
 0.6595608425644169,0.7839371049220636,0.14332005262616265,0.7281729155862843,
 0.9767474801478869,0.7973934130612635,0.0981769376978402,0.7666555540656109,
 0.4409195662665776,0.6855219378486584,0.3864757436969617,0.4057804449589978,
 0.43447907951716,0.7586561462154303,0.8821581449319527,0.9746085580536246,
 0.8622294366321835,0.9934705088516069,0.4347418306762708,0.8238420860149617,
 0.47377428455952286,0.43485643164568666,0.42514889264006506,
 0.4276544911303517,0.6011601923294436,0.3295688752880741,0.7814813715904934,
 0.2929389807324234,0.4252615912049922,0.660175833285779,0.6038787297608136,
 0.9329492640452736,0.5617863224558012,0.7830586121443373,0.709873395386821,
 0.41956634645354385,0.4791988174711729,0.8930340453156279,0.2554787982987128,
 0.8049457626453453,0.4599413595476084,0.5980720909370362,0.12244405785142343,
 0.9791879885022766,0.991237661814226,0.025192899381205502,
 0.02790991071408966,0.8193066087955908,0.5305424409083318,0.8172813954344328,
 0.5286673484383051,0.7559540050996267,0.9487157848927935,0.7468884487585173,
 0.767395152861582,0.8148791403126353,0.8152251744590135,0.7496334961481275,
 0.35830099547870564,0.3527458225848811,0.1134691493004728,
 0.11909077675560664,0.29923910440133106,0.0337540683173887,
 0.9284452209906098,0.2644436929143975,0.8894640682717658,0.374263756218413,
 0.94580380410028,0.6011692597501642,0.05060084359080397,0.7617304934201093,
 0.7654967734491614,0.15019904272577556,0.6409373629174634,
 0.01637092523343664,0.40652922966522653,0.20315711698746752,
 0.7791877477610496,0.10609068247953846,0.8197184092846606,0.511834679790295,
 0.439952341687217,0.024188167706417074,0.19470241818469558,
 0.15581112080162807,0.06991000784777313,0.32573448535397753,
 0.24039879346373083,0.10192191124131833,0.22602277023299844,
 0.6804293425772993,8.899756161508954e-4,0.6392377993304009,
 0.42086473658601586,0.02014325635975034,0.27858723439809396,
 0.19633692787062007,0.35181752940716304,0.23277953638055526,
 0.7934752336210026,0.883430316644174,0.11511641147177132,0.44155825451189257,
 0.7957290992907411,0.4758540799029346,0.9814063255716637,0.9405454678511372,
 0.9568873226962805,0.6352555044969361,0.042911526870420325,0.819001419022137,
 0.889738246072385,0.4402674214859146,0.6052279691912816,0.6567051335020084,
 0.9308161389424563,0.9766262611183957,0.15332107422631203,0.7997700881434519,
 0.44240645407193235,0.16134808118267885,0.5263081351509431,
 0.19537842219168888,0.9287324694778384,0.9019307031613774,0.1175557234791561,
 0.11797406582179293,0.8499518655725748,0.7703389042604384,0.4053390017826527,
 0.45369813748349686,0.8154584303981007,0.6024913007034857,0.6692677687080212,
 0.31197865354566245,0.8173576772149443,0.21065790503803494,
 0.40519238043987305,0.31729943231339075,0.8656298772755537,
 0.9791127159481103,0.5535722542971931,0.6288621340396526,0.08248210073963269,
 0.1683092844549935,0.9478180450777174,0.48092811859395346,
 0.23899536097678153,0.38089057813227956,0.4763863930605161,
 0.9777489230306269,0.1505912991898637,0.02921324415426796,0.9052946955350362,
 0.37372477705731555,0.4498293856177109,0.022619000646598608,
 0.7865759719171519,0.6107779290209093,0.9007658898484142,0.11209076242898641,
 0.4166160479581593,0.48857553055433367,0.04568666745307881,
 0.8615337719793872,0.5928832127875663,0.9588411837799344,0.16021887503936638,
 0.4209833047693071,0.9694151554226208,0.7433840677570167,0.03423741767202948,
 0.5740365860598364,0.6144066580698526,0.6955929110825082,0.36686781061399976,
 0.7372149124011617,0.4663328108363394,0.8511156142496711,
 0.031075417330850064,0.955251780691195,0.865052296037615,0.28537870360623463,
 0.12349268656386592,0.7865274639014375,0.2833026761502776,0.5153693755791356,
 0.5216967267952521,0.1723270720602903,0.11326641305114038,0.6574551439027368,
 0.040255408341900556,0.5114021458397832,0.21840810653061316,
 0.7701421602241806,0.4419217500456024,0.7583001949521964,0.5746628521632734,
 0.10751280674702879,0.25721411390136795,0.40729833562261997,0.52820873440658,
 0.5357082518387212,0.12373530008429434,0.0813611868315911,0.08263713126507,
 0.9440485653824888,0.9640438523194983,0.3884015313022773,0.3371045807755204,
 0.638956623565081,0.5396823900658411,0.015853019860668893,0.9466150868550579,
 0.8177043859897607,0.34073740044180323,0.07819880008197055,
 0.9159005770078843,0.15514331606405196,0.8946278865637057,0.0723735207979358,
 0.3244399121935071,0.9863913269711171,0.35596199705574705,0.825494573655314,
 0.4225429629769806,0.6664126768305924,0.9872660290944554,0.44578896450615657,
 0.8998962931610716,0.578218769078836,0.5390354483432458,0.035880205193452586,
 0.7069274302898245,0.8945962424806687,0.2890331823854658,0.4799207628137836,
 0.9472676988964767,0.7448459522414485,0.6891095734332162,0.765638037139044,
 0.45365332681665493,0.31562351124925825,0.10222803873133623,
 0.8554420939761721,0.5877607493001347,0.008206056183423183,
 0.3162023003307497,0.1529462829710767,0.9606200835555322,0.8082414701427231,
 0.2937614219262257,0.5847118569727304,0.4059557370307658,0.8644591982227559,
 0.8326624222612451,0.31667435589759196,0.06013524225659761,
 0.8190132977658593,0.30779493963822957,0.28803843488097625,
 0.5663344145474298,0.31662487375451887,0.21470700699075318,
 0.04137500277510908,0.053277533627904106,0.17587309469030188,
 0.3530184457256026,0.1170025774790433,0.63087939625621,0.1917207054614618,
 0.6069278959381234,0.8654473505461081,0.9669082196591786,0.12660407836820797,
 0.916077610044513,0.7587114731326537,0.6199182830821792,0.5981442077740473,
 0.4461456033514857,0.5010433473557827,0.8522404797034213,0.9324243728940711,
 0.7522449201009626,0.48402918545300466,0.46915835713931997,
 0.6857665567429806,0.1729171448288035,0.46986749622205626,0.5112596238366182,
 0.9581293470032671,0.2625953095535306,0.24413881063661014,
 0.006068429288903721,0.5209932694732269,0.0020381438715019184,
 0.35905251837194097,0.3292524342134481,0.2641988053345524,0.8036109100011246,
 0.09611604250018602,0.8155147122602178,0.5212026168818336,0.8930318924959038,
 0.8801370034607349,0.8252040195508241,0.695659438227095,0.9730788838447981,
 0.27017631026857925,0.4310043121585183,0.9145577321998997,
 0.44254097145497684,0.6322499131115031,0.12499806098088317,
 0.8374840727227001,0.07525317081187766,0.7365947050514401,0.5250428761463862,
 0.00897661638517444,0.2424272479870071,0.06728025436174712,
 0.18518170820183222,0.9818972404410352,0.7684911098783971,0.8033751070546036,
 0.6014293157456014,0.20880212399765685,0.6651328766888632,0.6343347178745277,
 0.8270654695292824,0.7593753330133255,0.6080700425517557,0.5749542653595995,
 0.3705347718585901,0.3041315623781148,0.6765638110677312,0.3238792842683118,
 0.646974843751674,0.5486200478657515,0.4720201958644349,0.31733514205623625,
 0.4753837544673044,0.5538879149090408,0.08439167550439419,
 0.059297575902352406,0.2708213294596873,0.8802280198577059,
 0.6798524495865279,0.46098368125322164,0.6764940518929122,0.925040311304028,
 0.01311717949484803,0.019196833590242157,0.521963471458308,
 0.5773673924758149,0.16409428724288255,0.9597981859134268,
 0.38141692782396586,0.8490448274751093,0.0853071211080404,0.4682779355525145,
 0.20601842417152905,0.7788216651428631,0.5967313985296411,0.8826687046721056,
 0.011694905619781926,0.37557635334981765,0.10739202097289446,
 0.20495688731853445,0.17749519153840354,0.7479145467142871,
 0.7164647162466473,0.5162667343873335,0.306541233946924,0.1646630227279171,
 0.11884720775499003,0.9515791144150938,0.3811610438897831,0.6238158027894084,
 0.7606129351229388,0.9819089392010198,0.7156734301642328,0.6619820717214839,
 0.9118994951493447,0.7049839192844468,0.6060503609636102,0.8052131928968961,
 0.47886437680550575,0.29516102930342225,0.7730322323046658,
 0.9507115636004505,0.544097713366373,0.3238281015924769,0.46793740773593373,
 0.39861234295976666,0.06485991823368475,0.46727725380825413,
 0.4338392343908031,0.18139633918328946,0.4952181520398736,
 0.33982289720256187,0.46505120540388367,0.7043167581986877,
 0.6231538415871684,0.5285885594201613,0.13992253964147583,
 0.08547605701532512,0.9202373491221535,0.3778444122473489,0.8646005318876566,
 0.3002728264462421,0.9121045114255399,0.37856593433176433,0.4796707847569689,
 0.06359521139527269,0.08850046194485994,0.8283307496558443,
 0.8242312942471461,0.19181938999224313,0.08199605733503046,
 0.7240368419194994,0.2553151863499902,0.7921090726424412,0.3927088573339229,
 0.5383598640401506,0.3335621473934418,0.40633112145517547,0.9763278006617394,
 0.23999364589568328,0.40308465935578974,0.6794433398217129,
 0.5302547289048791,0.28003801525766125,0.2812736548424294,0.4673006463691468,
 0.19035681190269738,0.9766040209481476,0.6331854808215553,0.8935437898136241,
 0.7033932334631366,0.17221055915701589,0.12529492140714893,0.267065650681236,
 0.19026334322325622,0.3500555700186293,0.8785076647224218,
 0.11838694975176489,0.1720575017470013,0.7819856455292306,0.6770618194668133,
 0.8576930548877473,0.6863297645114966,0.1988813387253895,0.8645139812669708,
 0.8773097075489903,0.3455238251449495,0.7854582193565931,0.9426011140734488,
 0.2681420772481313,0.42189496716049846,0.028859592662470623,
 0.906170160813936,0.7955627489454982,0.9601350324665057,0.27823773757376613,
 0.583681604865859,0.01803564427430726,0.9105242054093017,0.6071821284498555,
 0.468171526101431,0.29863494012383673,0.9553555613644766,0.03439881973607595,
 0.38001362947968487,0.4350191191983446,0.8682150515187945,
 0.19598669038403427,0.7927183885441115,0.6332004861407569,0.6990816207117346,
 0.12972848182766583,0.5562174453965593,0.04368762265741932,
 0.4523155041716004,0.9552108317096446,0.5164824470793069,0.49077093418166495,
 0.42884763874356546,0.21230952195660147,0.2700701952098462,
 0.9828124742659946,0.8850676177346652,0.4295305209785506,0.13120909628612143,
 0.15177677286171942,0.31416697114560943,0.15819220511368426,
 0.0186902780147733,0.7913052090020702,0.852367972980001,0.05714757026416861,
 0.36799549608933413,0.43049891366244997,0.8423803846219838,
 0.18442509003291396,0.7626353389416118,0.443962596523476,0.5545988387085017,
 0.46318153889750113,0.2059504345544131,0.20325634480944232,
 0.7563649582596803,0.8675175522876026,0.7838799616858176,0.4692293714042186,
 0.9172194930369367,0.3773612153009628,0.7299882067968406,0.14384967272438054,
 0.6410310938760575,0.1626763256652174,0.2829614928314961,0.04522127687555222,
 0.07534457225397395,0.44218236683988277,0.8574826807076426,
 0.40035920251313906,0.10143878290801878,0.19326302078295454,
 0.40448202979696735,0.8453272290777342,0.6055340888535286,0.3380550747672969,
 0.19263697691191894,0.3449996779011736,0.14481923889374881,
 0.010814192997647742,0.15112988111147652,0.2361155805988313,
 0.49573553244768276,0.3398680906262521,0.9350719715455472,
 0.36178156812883266,0.2199546692641996,0.7428819329440504,
 0.038738173431991485,0.19089546646185718,0.27623218482828227,
 0.16156854513251218,0.4569466524433904,0.7850980685883083,0.8723572460163553,
 0.783290440923825,0.8637289142092928,0.24309645617199704,0.692664866828027,
 0.8262802759808292,0.2842556122785195,0.44831808581721155,0.1400947891213813,
 0.6512916653258466,0.3684530765512385,0.6457465387831985,0.9920581928060481,
 0.7718469530058318,0.05365580829920835,0.1392690290364681,0.7486130589484792,
 0.21329224264278024,0.3051788438494447,0.4482561023023731,
 0.16511578941457106,0.8230613711563096,0.6962432529999056,
 0.44907170226883053,0.6859818080576819,0.48790645369752783,
 0.30455399236574643,0.6072692436119554,0.4540241887407931,0.9158348029279826,
 0.7207944737179295,0.13105754522715085,0.8036402928895241,0.9836397961347497,
 0.004763465061236927,0.6220303992674061,0.24158751252712873,
 0.5994695080700757,0.22750576386291765,0.43864754599276135,
 0.7009340043015201,0.41857784093270134,0.6759417269645651,0.9065010532074594,
 0.5430250431707123,0.030648455997265733,0.5197094519307719,
 0.08565193067578769,0.39629763891263325,0.3174503441758292,
 0.7178436539434264,0.5244576153700147,0.0547511191187251,0.179912792208896,
 0.19627479420561478,0.5771128995598274,0.27048663438329745,
 0.26136872863219995,0.5790543149653818,0.8340955743570828,0.5797974503986377,
 0.003848901343674216,0.6378256882854643,0.24428517346845724,
 0.5244285048669596,0.8736293359444717,0.45049310246616414,0.6112704965440636,
 0.23194103176981062,0.513389100232297,0.23263438842489048,0.6920875234047776,
 0.4336254483212879,0.4337069048645872,0.8188832696331659,0.6046747403263323,
 0.9643263736910129,0.9385230327308182,0.9778146204855305,0.6119509887776605,
 0.494247231765086,0.10091030842158022,0.44731720782994244,
 0.48605088431669596,0.4698515797174716,0.4427756406769151,0.5487411111527133,
 0.8847183154678245,0.9953417351834771,0.38322981043558335,0.4479631136554004,
 0.8182395547558448,0.30137361238556193,0.677401104037781,0.6674933418608557,
 0.3729629915082223,0.6232441448659914,0.5639901102948374,0.8430528209791764,
 0.5510059206719322,0.3481865827927758,0.26505533389971636,
 0.35490601765816776,0.18496052124144058,0.859240528882245,0.9111776396829839,
 0.665125136008499,0.10744417089743152,0.5298648587147665,0.6527173689038698,
 0.4587246219342065,0.051666298580725134,0.44534627271874616,
 0.7530481358646162,0.2791672167854916,0.1051106929537029,0.43027363155321163,
 0.42162597530559986,0.8215441715408071,0.6152533033558072,0.5035612490842601,
 0.7903708366019878,0.021453270917160427,0.9022513566789578,
 0.2514044357238182,0.5730940168831818,0.20403882314828048,0.7130931072896637,
 0.7099543378603108,0.43428874582696364,0.5447222473279918,0.5359975323508507,
 0.18707651903732558,0.9308364661961085,0.2090412980421522,0.8818439498825947,
 0.28709748841495286,0.8605199723033076,0.28113426880943027,
 0.16660182903703213,0.26710789196519946,0.3032291338606905,
 0.05874799900980898,0.6108806050620574,0.6051228565861706,0.8572438564890669,
 0.05779462592792295,0.5434631087593602,0.5249965910523207,
 0.07092887674665971,0.9392711465276912,0.9610152212623441,0.7355696219090108,
 0.7878156716083072,0.5681691261018006,0.0882124731690399,0.9861870771395367,
 0.5315431508504714,0.8986341459159525,0.5833026825740953,0.8609585344142843,
 0.6321953382823086,0.8897746785862567,0.5545431613046554,0.0932324319143355,
 0.8077178964447698,0.5600666667539543,0.7969191564867285,0.6117920828743213,
 0.41236263970749965,0.9524695761745555,0.12978384139578025,
 0.9936193595173459,0.2709311537865414,0.14029510001633483,0.7130916929299455,
 0.9856576846587358,0.7932577110897713,0.8028302739142956,0.14445857247700045,
 0.9550110432136757,0.15223563523229466,0.805542155177116,0.17625677060818246,
 0.5028155454467823,0.47729238997926215,0.067380307872004,0.9659399460692099,
 0.9662430004209455,0.2633793725570587,0.7622627893418824,0.5862127710208045,
 0.2857488310963292,0.2103084960206708,0.8628120947883344,0.5713518946952186,
 0.32046038031640856,0.30546788219219345,0.014024960553510102,
 0.8800269366758219,0.5345945820793699,0.16450457968655163,0.8720360221664304,
 0.9190092292208485,0.15738390688708737,0.06062424331037275,
 0.9306160521487719,0.30364977433778706,0.13037475268116694,
 0.7251634917959995,0.2277139015666091,0.4832596004808689,0.1209067517861836,
 0.7006706340057195,0.7269723354620498,0.6460293271680482,0.2005947890852644,
 0.07262161682667512,0.34025321404851894,0.6329375212550157,
 0.22829030370864534,0.2246510705362893,0.09001434467755143,
 0.8435663020155246,0.5960863274308166,0.17062079509015127,0.3021320538768334,
 0.5524497719613599,0.6944858717394563,0.7586063234737508,0.882121270995329,
 0.9559644989912158,0.8386487297182011,0.2771412537641559,0.5115463628834753,
 0.41313106801145305,0.9098419423542475,0.3376583929288203,0.4247068882008801,
 0.2955014944562139,0.06532859676787084,0.7866894721386137,
 0.32714929166717344,0.23617395261421614,0.1972688380644645,
 0.4012347685960225,0.4051445994022751,0.7396540658813906,0.8293740582808833,
 0.8038525467028003,0.37872760341092704,0.15001463834074769,
 0.8832368982472906,0.039494833821299036,0.8003606008431994,
 0.7172722091227887,0.8300665796781928,0.9731450999267318,0.12858501836315495,
 0.33221857208543704,0.4006433942546864,0.3420158498638226,0.7570991791679885,
 0.7257022260173392,0.1736414961069218,0.2728512732537062,0.9170255990963427,
 0.9679482876030705,0.4103690989063533,0.3613344512033052,0.9945403406948508,
 0.7341103442056155,0.8579839465686266,0.08088044453794652,0.7001861342055513,
 0.8845520750368603,0.8084486253136005,0.26312117135034674,
 0.46593428728660125,0.9947564651555326,0.5589015913903975,
 0.39174663084031025,0.7735607692315614,0.6694301772822406,0.5109379989043701,
 0.6207758429974242,0.8019567076741367,0.4326778285796732,0.382999387967079,
 0.7985901250543713,0.32029817842059516,0.894530857388191,0.06204512120857908,
 0.29655368216133415,0.4628340568729251,0.0847853148874288,0.8298272673081818,
 0.701639014974419,0.7092788606684576,0.5455971661200137,0.15974410545889317,
 0.4537977117461207,0.9651575592411391,0.06711110773698059,
 0.22425043126546673,0.4041494929869386,0.6382486433127124,0.3249438349363659,
 0.09932644576734062,0.2280519522801504,0.42193920564045273,
 0.5392114423758646,0.6355348243348798,0.5772232747226463,0.22820559587761158,
 0.08411582554489994,0.6092355081086235,0.8329701672726634,0.8606556335694848,
 0.35352789656209527,0.29958933269391075,0.6759470700468156,
 0.6136893876133169,0.806733742489037,0.710393201358983,0.5369540370937671,
 0.8309801530470028,0.8364569871793026,0.6229118824353518,0.20085335905784352,
 0.16391869780104262,0.32832718422452545,0.134977884547255,
 0.30053966172619695,0.39348556987870975,0.7295276708852299,
 0.4299055262667486,0.7347193017210533,0.22477466209250174,0.7576658456862168,
 0.13316044779347713,0.908203277181269,0.7246255064635663,0.7514288270062661,
 0.31310896544530054,0.47155039608184524,0.6232322618594721,
 0.7810076732917103,0.7523803712433992,0.4752614749744588,0.0924869448197958,
 0.9888924962773038,0.6475176942358688,0.2555896599495383,0.9539021848410781,
 0.6546517632640257,0.9316024785846218,0.4131124230437674,0.9205995264520832,
 0.14410873489430986,0.8335181237786708,0.9824096408657409,0.7834342799499263,
 0.9624219417770269,0.5380448631881883,0.46634444014516274,
 0.32244944046851043,0.6778784609923133,0.49989718538581673,
 0.4824203807865579,0.8845364049072852,0.22478760887674865,
 0.19368533924606446,0.3943876973322167,0.0617460200755654,0.9554589251327962,
 0.7909361734935465,0.3583803563738839,0.7333885373713498,0.30825506983535866,
 0.6913577726572808,0.5117415780263553,0.5647987657949212,0.5951404755968625,
 0.27985893701918896,0.5777718271976354,0.2546105043842304,0.5517102884864922,
 0.035911408481979934,0.04107532315368023,0.9372715862184064,
 0.3804893739821945,0.05024299704502433,0.9923536587721092,0.4145061708489146,
 0.538251200842092,0.6440669654660391,0.3647548198989071,0.15790939255076442,
 0.2504637561112013,0.9178623919391127,0.5860628448701346,0.14110909718613596,
 0.8445170257882948,0.8860535726120735,0.609241656428166,0.3010164894009255,
 0.04983466659582314,0.8800374039340945,0.1829798136571451,
 0.18508725635142365,0.5915746059286917,0.37212382751071515,
 0.6701511067788015,0.5384052683147501,0.16057642444865972,
 0.028418842993064963,0.6924368709576293,0.6739089624662822,
 0.3702747903059479,0.5837182304322267,0.13611061808395508,0.7420719220553684,
 0.30553408699917417,0.25709731319747764,0.5117629096999967,
 0.05526884842934321,0.638794814076489,0.9778324765411577,0.8460535634183842,
 0.9400551424346653,0.25360769627871527,0.8645019880190399,0.5392669316948852,
 0.0043343426605495505,0.08042645296536044,0.638972561508627,
 0.4518208695516873,0.571219672209337,0.6454410559248946,0.19614840090394148,
 0.8724732510382981,0.6839747327695932,0.5282085623174658,0.2144421384747509,
 0.04916134128621219,0.2874614725823351,0.25210903134836693,
 0.7199612923264055,0.6911933200638383,0.8202553966881808,0.15128959895666194,
 0.09396772596200487,0.2223611696117428,0.02478748135735831,
 0.6726764000047538,0.8495745320538325,0.8055191662593756,0.6295582013292766,
 0.6294904506460526,0.10643382847569138,0.7798742099580978,0.4898092717841164,
 0.9740372246255652,0.12742425314858996,0.3508198717537412,0.6081530634728731,
 0.3224483777449373,0.42763595183826164,0.19357406014274314,
 0.8437019648058364,0.9840435417663274,0.10016506396030334,0.9367754997590283,
 0.9581756408360805,0.7118658721331586,0.7580848778393852,0.5522546376908825,
 0.5184606420104932,0.40458340460118447,0.92513169879063,0.35020792522474575,
 0.13594291395430835,0.4574618226781971,0.03311839032655439,
 0.5122117398630099,0.8708843816766051,0.222276523765636,0.8247883721735026,
 0.7448570238380522,0.11640721739025406,0.4012352136497961,
 0.28853759900732123,0.7110197790002202,0.7368392227861958,0.8254567947201158,
 0.22748304789987495,0.8783376314108843,0.3769602905232279,0.5492598122449516,
 0.5421624138024983,0.17252109236940394,0.7334323354906436,0.2830681034638829,
 0.3529617324274069,0.21308704781396026,0.9124709928472284,0.8599116538168852,
 0.07424963369921189,0.9475908742998687,0.31470623103368733,
 0.5511314313461292,0.4874743981686446,0.6103731059759471,0.8671665588049599,
 0.5405102687437141,0.4671541348870265,0.6135083147605432,0.17670036953958435,
 0.9326792856048447,0.859161985752268,0.49793005456879713,0.7267161190790606,
 0.08791605552733506,0.9804701424181306,0.34647095468103306,
 0.5860682304062164,0.684975594784647,0.007087287742341131,
 0.17460486095285943,0.5599960798436685,0.05586914754009287,
 0.7871143137460034,0.16266752098734738,0.09474470712431238,
 0.6990204597508962,0.2227788914215325,0.4403770247125699,0.9223520115808193,
 0.3016367309597925,0.6893003908697404,0.23473787828053072,0.7933159067743263,
 0.6647720376061109,0.5605871489964989,0.12144661255123768,0.9814547951734667,
 0.5358684549637522,0.2184349487956354,0.23534100787729828,0.4849969382285537,
 0.5537346208651541,0.30470621685050725,0.2585919649637711,0.5918029285433896,
 0.8194587878956819,0.5379654528536476,0.7494631835362908,0.6882115476967848,
 0.5498508117625595,0.1964898254634072,0.7862044173710859,0.5048291415761168,
 0.939208910983582,0.46068076633601174,0.9383417431928142,0.23353030245187,
 0.6047776871641956,0.9161140768983148,0.19973544171763935,0.6097325299993231,
 0.24810002572193135,0.5420379355997379,0.4495923155395307,0.7012458456882725,
 0.5443081708198835,0.05393590355589162,0.5494615203249936,
 0.30905993924920105,0.4428105811077315,0.7275499673446932,
 0.22097959565807546,0.288744114709812,0.07581314178774445,0.2521078084072017,
 0.026700935899453326,0.6748995599204537,0.10253647211591577,
 0.5678713691789473,0.9036001029229636,0.38031280505149756,0.2047992797083804,
 0.18630452884347748,0.680148025458047,0.18529319936454525,0.6253730098596654,
 0.06622644124021293,0.5394908810903226,0.9439329050293649,
 0.34016264698385346,0.5807948672829629,0.3991007561830272,0.813229523857113,
 0.8047913344833142,0.49747128601600976,0.9032792829431235,
 0.22893487364132714,0.5001548772914509,0.5402323711358017,0.992021361575855,
 0.955839058346079,0.6041845530574272,0.6475068774631102,0.1681288964095209,
 0.22192124458145257,0.05869645715211158,0.3359987933006817,
 0.4955599195632351,0.6722518908494826,0.9745500992800454,
 0.017990161107586555,0.7484076219612104,0.9854669831489492,
 0.14282714816985054,0.4518404093955586,0.24403123170378682,
 0.4734090867007408,0.9341170532385797,0.9742965107160572,0.21242768764263442,
 0.5108696164251554,0.4901885918005622,0.1879397505870246,0.24397037816259004,
 0.3359878331997136,0.9423650332132403,0.5066945990716916,0.3048845649256172,
 0.7773629974102596,0.4155231024140663,0.7942797111651088,0.880416697349202,
 0.1387020019972489,0.5744194457730283,0.7187653839376844,0.8924780730067169,
 0.7092112226324001,0.045762577296489915,0.41435178474600476,
 0.6382377352695519,0.6503226176706624,0.559452635796859,0.552015226938178,
 0.7580586275557423,0.06142151016243591,0.9377935323823375,0.3402359743994209,
 0.9410222748024866,0.11401316388857419,0.948486633032202,0.02684733896505509,
 0.9632847825349572,0.7444124295007489,0.8183718568196178,0.04882699297525783,
 0.0608535193202161,0.18035564332088838,0.6111165328296554,0.5766463599326006,
 0.17460970919407326,0.7318759947786367,0.3363151369981957,0.8117190670149984,
 0.9413179327046413,0.1716653554665989,0.7311892569748808,0.7659560457172989,
 0.36871516777678626,0.5018659614173892,0.6761586262078492,0.0930644886250559,
 0.04676711590202909,0.8134921808956562,0.8245736174295325,0.2314089120102525,
 0.7389620846340977,0.887164232431584,0.6585373320391283,0.4145851439101087,
 0.43494559982046077,0.45073279488072227,0.5688077633877147,
 0.3296247270602224,0.16505314618234446,0.01933086137066775,
 0.39259321294896643,0.30996679451613673,0.03632873700664063,
 0.9413500603570611,0.6937824763320738,0.9587355007547137,0.09454817316103248,
 0.20159889232023032,0.8321489672847526,0.010600799330320876,
 0.4579790900513966,0.9826536620082467,0.3240436403069853,0.4690362208378508,
 0.8745558226126325,0.5807340605536948,0.11401660976884376,0.3859807621880399,
 0.4493870472574921,0.45270846499653183,0.16313022568689595,
 0.7908628595067055,0.4432776192857977,0.40560063002121793,
 0.43939248051498836,0.29879365041577755,0.3145836263717048,
 0.8496587208780543,0.23959327314550438,0.7886618846287042,0.3611220696443619,
 0.9696181065384509,0.5920387749835818,0.8311008685086899,0.8308029218806362,
 0.551030424244723,0.43624418776153595,0.6708953041573702,0.798804306437123,
 0.3934382814008013,0.3287501759949112,0.1914647282784847,0.5008046045175236,
 0.4954341293876584,0.4354881894428241,0.4511266214998736,0.19116320506912077,
 0.6768971945011057,0.008128759801040442,0.21951286291674565,
 0.40019856977568125,0.5985260089916398,0.8380483434358597,0.571271804889838,
 0.22523570752261224,0.9395659845626811,0.6699288283409629,
 0.27548113239853755,0.44879273596032376,0.3173430821456912,0.713522837797776,
 0.7705877689341394,0.3395200118712206,0.20553951125694836,
 0.42486875036630645,0.8268592421822718,0.3090489354597924,0.5651049395130531,
 0.7118350887259431,0.04211832577507968,0.26245940118786903,
 0.13565914080030028,0.6634128771025614,0.8582328108062788,
 0.12026704753411144,0.5135114355135608,0.5017782648291795,0.6519800028918789,
 0.7454879482714616,0.808981545631223,0.045837709708440366,
 0.08519350039003004,0.590703751382427,0.8480108372666286,0.3610128820576012,
 0.49444438094948495,0.7093555870910585,0.3250798290853051,
 0.19048827360295106,0.9246492825082462,0.1874690951701654,0.9850981910109379,
 0.21680435142878107,0.6330825285368736,0.4560387541313634,0.8388873107774477,
 0.5973278202009431,0.3605828627593013,0.15293950755037855,0.8455521620130938,
 0.6958436908616179,0.0932504944689736,0.5584113175167553,0.37822844498880215,
 0.007114472357449886,0.8079850509558608,0.3777031856249199,
 0.7198371832758446,0.7660412005390529,0.8212585028128714,0.3500570504304349,
 0.8415396239552408,0.7824377415125312,0.8261322889450211,0.06967224999238575,
 0.08282650075918829,0.06468498519151333,0.839281635897422,0.3328820337006143,
 0.5615277386627325,0.9594998231633899,0.6371967141894145,0.4293765355751419,
 0.4532631445819697,0.06982837268355713,0.9539654610044939,
 0.39904192520460346,0.9564783063494385,0.5108186032875505,
 0.003077348746285402,0.5266707726153387,0.4823345909336718,
 0.44549063906292485,0.49413734005132426,0.07561484613491198,
 0.9935824174740382,0.4615653618551848,0.771385277744041,0.925249471594616,
 0.16524175908045224,0.6494601758047492,0.40016134340019094,
 0.20743039965194865,0.48321133292264606,0.34562733641675736,
 0.16500191635315742,0.45257839456457094,0.79181829703974,0.0516139577683945,
 0.3202398655044948,0.9025563453023134,0.2958743973698572,0.08623394146316865,
 0.9520489495318336,0.8186907137295293,0.8598115803655119,0.6194591644275196,
 0.24354216645734061,0.8922830023246083,0.2280550126203037,0.7425528556862184,
 0.060798284157553284,0.9776909204394205,0.8995714205616867,
 0.45575962708160145,0.1705506602387732,0.4277295692948271,
 0.07829625499534965,0.7334142419106631,0.13458708912104123,
 0.3666365610663307,0.17781786318467763,0.31815590849291475,
 0.6165920051852932,0.2687556940304009,0.3466926029557492,0.2590624964650564,
 0.06119358419016152,0.5775154326082707,0.658848904660188,0.314291049550959,
 0.7604297913964342,0.8596627574430294,0.7389858627801954,0.3053628356700322,
 0.4448588243062337,0.8812216446850031,0.9449027326511125,0.5351465133996576,
 0.7848908470555527,0.21959705026151033,0.34440563974641314,
 0.4689390851322265,0.973800150817093,0.6929687965187263,0.19222389946898388,
 0.9886454638187105,0.9095545126966506,0.584962182124202,0.6068293639397061,
 0.23887148784209333,0.7764654246065509,0.23509423030926568,
 0.5534238833484137,0.6747403558635399,0.7866212568077255,0.6209200444492757,
 0.24308004672129213,0.36390106023093993,0.4005369855364589,
 0.5994305736666636,0.8422213269914773,0.6697344460663514,0.03952689611753213,
 0.6336788524137982,0.6841217126797693,0.48391789623141057,0.8673254609913503,
 0.8863833558656202,0.13032040603709272,0.04553752826157731,
 0.9570796482003596,0.5898978626332326,0.0071203726287204505,
 0.6195170257249636,0.648001965457562,0.36404223356976273,0.18751911172541813,
 0.3071642292194109,0.6312742641915728,0.7955442716189484,0.7655324397795223,
 0.9655133418500674,0.39072043638673415,0.7736550923079247,0.7312525248058583,
 0.15100546399620507,0.3327893438794828,0.04159454667420859,
 0.17293236726965666,0.025137963051163115,0.7962169746994822,
 0.5678847468400356,0.5098505150211389,0.8675453390361043,
 0.047767839097997755,0.47968571636084756,0.14190212143746694,
 0.9458177990719078,0.8666343032360664,0.15830856616465516,
 0.18868047910877583,0.01031074890760686,0.8062304035218986,
 0.26870676164250307,0.10508004896861722,0.8781118809459515,
 0.8032213998812124,0.5751220872972179,0.5461338904757024,0.8985137237234143,
 0.9323494771839971,0.38854789620986074,0.10292323485656374,
 0.4691110249542637,0.3143537015167043,0.47566649984769893,
 0.07654614915640201,0.22186856803229926,0.06879692352520439,
 0.20171340356230694,0.734008986111946,0.038874031623186034,
 0.11845562442928326,0.7984317422867606,0.44322678546636574,
 0.6294922826567915,0.5634974453588746,0.2968619665231642,0.44889680721806346,
 0.519734279538686,0.1934325974760137,0.42139578783856724,0.22892058921172576,
 0.10337568432072508,0.263463090410341,0.7019746216594877,0.9427039676073945,
 0.04292096547114044,0.568141987747777,0.021551420102479346,
 0.013601411111810013,0.8129065078165214,0.137077430928931,
 0.09754461156008054,0.8765678667708531,0.8940179389678324,0.9130591451174802,
 0.9658211420913049,0.37154693849858067,1.6519244871227023e-4,
 0.3670865581548678,0.5259224080683979,0.09393061479901188,0.6220601054178544,
 0.6842597295470076,0.37712689838951396,0.5400833560755658,
 0.03764913556678395,0.29262378943843015,0.9580766015388685,
 0.04609181692521591,0.25840016000844324,0.9966194159202393,
 0.46931016738831755,0.88303306182333,0.43457925808714304,0.7503939145421636,
 0.2602456440545544,0.712659282213329,0.17228676637415286,0.2934516043180758,
 0.867796564508455,0.15081729527418586,0.596920347565936,0.9609850261265719,
 0.1408666956844622,0.8780565386604235,0.296687305270675,0.8146017234115719,
 0.21200468475481316,0.9027322488810245,0.8896101126175845,
 0.48661380901773144,0.8247294849166852,0.22653222035342324,
 0.3252975506865343,0.9601256932944254,0.02061349963394432,0.7903878072770935,
 0.655908899429313,0.757465738662504,0.24312584975233975,0.6168836743317727,
 0.7087759632391452,0.9572251985513224,0.48541720256600895,0.6213935312106481,
 0.5776786041772415,0.7944892090518894,0.04264826392551735,0.8141315306715758,
 0.3999103262513136,0.30528538490606194,0.030191921571372404,
 0.12938857444871632,0.7711862984725996,0.3843467694191761,
 0.01499446506286739,0.29766826292620874,0.1269993600058834,
 0.8960966763356276,0.06048578686506545,0.3012410953751198,0.9052652383513805,
 0.21975510534465248,0.6468029306495777,0.2665779364566666,
 0.35073416120707435,0.5194436765132201,0.5259133751413072,
 0.17157193855316233,0.7327618111082154,0.7436187359915227,
 0.05111198821932106,0.6285634502150557,0.6104805558776061,
 0.41176163934830723,0.5811822616938003,0.23749633318575858,
 0.07831782373505203,0.7524291143342564,0.686758489044884,0.4001702203396963,
 0.508161127801553,0.10947286411985169,0.00624307164345228,0.575534724400917,
 0.49250890405017844,0.48394808390716926,0.4673901616397077,0.766667048402929,
 0.5449668975323028,0.6935367789549198,0.7683537783496877,0.6620286824290259,
 0.4182447421520651,0.4268352851890693,0.8374620609936217,0.33161401089176956,
 0.8728813994866871,0.08710923487786593,0.6685008320190226,
 0.44842100069683066,0.1467257163435559,0.8438124041438005,0.7564903270713702,
 0.8710147818065932,0.801664952871833,0.19348840380185595,0.31211767719703365,
 0.3672568228994775,0.3709701934738471,0.7685293794100384,0.5120528544183107,
 0.961662161727549,0.3171077127111338,0.3023487456855727,0.14213924323277893,
 0.9040509138398009,0.5449216627437078,0.8731624261920465,0.17750699941726777,
 0.838929631483882,0.3764322893967499,0.47342129909555586,0.6091728700864018,
 0.8598851689763878,0.7197935874019041,0.6243384488873972,0.06671104800229877,
 0.3339365875622726,0.28063340912224133,0.31789472000471275,
 0.30038524842190906,0.15940735479207258,0.07056276576248655,
 0.10332638810066208,0.5121610916210353,0.6878991688611726,0.39334216613977,
 0.2554766019407725,0.9837672700604392,0.08149980117763678,0.7808913936192317,
 0.026381642264836747,0.9788429827402532,0.6132224087935133,
 0.8955829291514268,0.9389692373716086,0.1066379036690599,0.7547208878908545,
 0.031207060514363727,0.8104777695259563,0.24055467120145935,
 0.3800930946423884,0.9711992498305658,0.9789425644461345,0.4671355205294515,
 0.12779433077728475,0.4401413223769909,0.6203325768250016,0.601072786351589,
 0.2714412458727836,0.9806935271056951,0.6510938416356546,0.1487633118866467,
 0.602768089383507,0.6231613050258051,0.6417244970889522,0.8407819646961547,
 0.40055988660824093,0.015400887811954145,0.16355913764066787,
 0.3806680020184049,0.894859966584169,0.09079071989090881,0.08419735810038365,
 0.8529881374317193,0.4006767023344031,0.9986507732447846,0.16909420713201684,
 0.2199094442946049,0.5930485411665806,0.026454042719503135,
 0.5780971844711622,0.9918302451449075,0.11057944937431341,
 0.004190372609151671,0.5079253137462056,0.21076727981882215,
 0.39762143117314963,0.1495840080900136,0.4958795076551029,0.3181746957528575,
 0.7080303833040471,0.5237351549309852,0.35428996189904804,0.9722211161387808,
 0.13450408660559776,0.30568421643722965,0.5454677735935626,
 0.0025701979205251524,0.028712135455526977,0.12192258532501743,
 0.5199759418284341,0.3368645100680764,0.15175324403670276,
 0.22117550912940054,0.6652989300691206,0.7878692389669625,
 0.45628548823121284,0.24355594948780246,0.8830003125577232,
 0.005072871842003113,0.6951196464166244,0.8543167222344804,
 0.37318357249598744,0.7676169004207697,0.5264794591638295,0.268954418272092,
 0.5831144012887054,0.19976517837597962,0.567093378961599,0.15934604855893264,
 0.12558418809285787,0.12785808490529949,0.8835028681447231,
 0.4061297676073482,0.9790798128382328,0.9476940254353056,
 0.044883249049543994,0.7698071735390113,0.8488726442728609,0.527400607299439,
 0.44514580793165914,0.13349220861529343,0.9040758612325976,
 0.9788362288463396,0.7603262334492447,0.32532566582575806,0.9823154388350825,
 0.8637236810452007,0.34403710441388824,0.646281789495142,0.8534375695947632,
 0.6633209701743822,0.2669103465426874,0.4622090727713388,0.6367087435385701,
 0.21018718529078928,0.30935628017054173,0.36702936028883226,
 0.26580022506042356,0.5826994418808042,0.5191027232110713,0.863439732946504,
 0.25120463277110505,0.25616900059581826,0.8404323777588919,
 0.9385696532077457,0.644515111322731,0.2483434024098189,0.35729846391711706,
 0.42999264213956834,0.9808370131164588,0.36737123155917395,
 0.9589697246548708,0.8041789981403393,0.6012392143919684,0.04494936640787961,
 0.3631865962466251,0.5457013952905098,0.5481363600176726,0.674226188663188,
 0.29388747043091046,0.7223538987623898,0.05650078281344961,
 0.7651440537904098,0.8761947744416265,0.7287562937221619,0.5103876236476752,
 0.5064224398956758,0.6553971436332714,0.87894508085086,0.10674314632002224,
 0.32945265470669804,0.6002623953890573,0.8851652235529818,0.1989268874600707,
 0.40917239053542787,0.38110418272715085,0.837194246590939,0.3358040003906726,
 0.3351067937574769,0.16218051640701114,0.2470619546793359,0.2157080844989745,
 0.9710867814114633,0.4778776423334171,0.671850232618129,0.4578254122137574,
 0.42101315259302086,0.23423391606951094,0.3434932722728634,
 0.7821776251697355,0.16025131571215911,0.5227727976271528,
 0.41892144107735163,0.47095239217603746,0.4070852606800849,
 0.5392908534935154,0.7043216209649964,0.3534725595657232,0.3646528062998716,
 0.801012028727836,0.6227101591535489,0.4283126581796126,0.20456444530264262,
 0.20282362792756492,0.8427081382506754,0.25623067042249537,
 0.8358073744645544,0.7802180303123537,0.34160987445410573,
 0.15452043109807567,0.23016590354732425,0.6997866542507061,0.669314955502018,
 0.5200499029157442,0.25686443842984275,0.848520927611073,0.50314574711042,
 0.5874457271793392,0.12275017784706588,0.5109470256747474,0.7508969210436847,
 0.2743448158869166,0.20927096117460364,0.3289485572003552,
 0.12838343457858714,0.8817774602004219,0.8083600931198669,
 0.10864915811428433,0.022108428857621876,0.24241992398595902,
 0.27119646945680276,0.657033375963824,0.27180866097151857,0.1667453437209716,
 0.8981823282866364,0.9796087110890676,0.43672812037154785,0.3842168125472465,
 0.07502402627994986,0.8539945190264241,0.4620957279220457,
 0.11368087348345757,0.9710585439667777,0.24796849141926736,
 0.2738074115115644,0.38895244598379053,0.6804811751377193,
 0.49160227451149463,0.9245052028629255,0.7824816891693471,0.9306217193162216,
 0.9036963302509761,0.32112399805042924,0.8449949499284786,0.8069087770448516,
 0.18043664281815586,0.375371208937363,0.008097669287638554,
 0.7734930806934697,0.6893391175425241,0.09054941679910022,0.4748954312844196,
 0.8262114301239643,0.19286567324494697,0.6461276559999832,0.0639556230433227,
 0.6908702598172347,0.04917805024321431,0.37736960351734716,
 0.8077922848988329,0.18448314457297743,0.18684422028219527,
 0.5732377773351088,0.4233544480909165,0.09777029432753892,0.5243965099247568,
 0.10383492300167585,0.6985236066400886,0.11612851140568303,
 0.6032693752803369,0.40792158804517087,0.7992244063277105,0.6193850525816381,
 0.22409350609661016,0.3568141532725636,0.4421411742575294,0.1620622230823202,
 0.28099660154582107,0.993767572568272,0.4062286239801153,
 0.009121947595597368,0.034390337662843384,0.6488725250514474,
 0.5013237169579422,0.9637523458724094,0.04056856628410366,0.5223095021426607,
 0.13739067163011853,0.5910358223450985,0.25056765557158567,
 0.3191657645681568,0.1528068741507751,0.2604685926171627,0.5984038124741243,
 0.5484654983182907,0.1277800198641421,0.051832493990958683,0.935542707838803,
 0.4633947583541673,0.7313386767705203,0.40159087406493965,
 0.24695377107655014,0.2088647879756691,0.7128185213938462,0.5377800652022127,
 0.296635727755179,0.4342883215500415,0.23429711153367938,0.7353200772967566,
 0.954451756094127,0.2180415925304512,0.22644179010110754,0.6946066623026903,
 0.9348757184974072,0.13858600717118252,0.31597197422931067,
 0.19510258625751375,0.6118599816258239,0.28673246454988166,
 0.5353879495057534,0.8919531561100348,0.16845727306496638,0.7269907283350631,
 0.6124987498368376,0.5361948432451642,0.8647297701699888,0.16822174802575218,
 0.6366731301071438,0.16099654665254037,0.4430517707948147,0.5699205530350632,
 0.359324152924165,0.5299717944377172,0.5606960717437471,0.24320519059491252,
 0.2698311247669305,0.6948780509226067,0.2781826239831644,0.7245730510415833,
 0.8559877654856075,0.0011628694905351233,0.35377469545721507,
 0.6654758709154989,0.797384113939859,0.2897296401647267,0.3822482619620251,
 0.35147961879378875,0.6845016685866914,0.383243271356705,0.6340395722454897,
 0.547898597152562,0.45612962444600436,0.19107498180241533,
 0.29492789370621264,0.35103365719219637,0.22914895141927838,
 0.7517359337737347,0.2028494063336238,0.013122961268419164,
 0.41798032270648966,0.5726272437446721,0.5045472062896715,
 0.11365302682158607,0.44426334744694906,0.5861885070396373,0.755719250213665,
 0.3312637500931097,0.17029663695825425,0.13800435710553327,
 0.38383799418642195,0.5943273142464458,0.2555593948808045,0.7743800264168098,
 0.1313175008914267,0.702730703931706,0.6095762490334088,0.45464272724308485,
 0.1311249208658789,0.5341551743899395,0.7381101721314388,0.06707552088867763,
 0.16817345886740465,0.1379702971756981,0.610374408240119,0.6859859795995097,
 0.8493083591135866,0.8663642468615937,0.186263922356121,0.3706987731655088,
 0.9716194757558291,0.7748614628037784,0.42582352639017973,0.553162316803831,
 0.03527866878602348,0.9033053285098837,0.08658169519300707,0.558040428619038,
 0.705494443312233,0.8147004877355711,0.6576158521058719,0.2421428750976029,
 0.6060982497258772,0.18386420745077492,0.1641164694519045,0.8694531039697307,
 0.49652231839702465,0.28829263657766013,0.3898285488564107,
 0.15622411099823674,0.0884518805500607,0.4410822469277009,
 0.36418920579309755,0.4478174290615555,0.8715087418316465,0.555733547019537,
 0.9307046807356627,0.2580545138842366,0.4735511724629232,0.06348629829667252,
 0.8772212344699637,0.29796246716320285,0.5031888739114875,0.1329276060074721,
 0.34508059356301424,0.3526542869784265,0.995213550998612,0.7263597233416272,
 0.09989431807886429,0.22305143119328597,0.5299040198517931,
 0.7643819553733708,0.8302053665750687,0.3333595368718413,0.4265394630647181,
 0.4499581530684943,0.8237322089633504,0.20102893640469677,
 0.13887791544462003,0.8407116063261452,0.17620460202198918,
 0.7939929968686724,0.15787903834198735,0.9913010583547812,0.4491317511804874,
 0.8383048472192051,0.029726509636539666,0.05152932610465211,
 0.8063920380659709,0.7879358466667474,0.8369413447844182,0.37463323774519597,
 0.7591058015908526,0.20227523133873138,0.7830388086093443,0.2597769944960133,
 0.9638934509674281,0.6183348115207654,0.26440237045293713,0.7519125202078175,
 0.3117502558208608,0.8804035595108675,0.7192392523829549,0.41453949655193467,
 0.68236227843027,0.8938012482831246,0.058399561057089655,0.9353512160217213,
 0.9712755102626325,0.7330010619370166,0.19117826900753165,0.2623656897722143,
 0.16752326422058583,0.6263333726072127,0.519287516386684,0.17025256040157466,
 0.05943824942067222,0.7086549246293592,0.8722679664289472,0.6298213663093092,
 0.08735822711468788,0.04926302149947903,0.09155652371979572,
 0.2358470742084542,0.4671313174774495,0.6543131527053695,0.3508927841867886,
 0.532784287443659,0.25319111132668426,0.14586370698661766,0.7586528041834893,
 0.6353080387901285,0.6317797237007423,0.13080608035552954,0.3019628998497904,
 0.699262432263346,0.2678872337027195,0.8684903387600303,0.014097866526458724,
 0.6280933669576213,0.17337962199899448,0.9355770754024314,
 0.45932910827853846,0.40164160041655894,0.6921631404478988,
 0.6922008941285578,0.2417310377164137,0.7549860755584036,0.3589649259010548,
 0.212358320369468,0.3519219033686214,0.09511943343211904,0.8195234779487822,
 0.08162170176460537,0.8822554067810293,0.5159879162915892,0.5548905807096878,
 0.9268683476204901,0.36747120631448504,0.026083456139658345,
 0.6035041317664531,0.36511694979471976,0.8530224099761365,0.6718451401868791,
 0.9240084901817822,0.5452973826857912,0.14473048405340216,0.2593841780876778,
 0.6388508866178166,0.5421782268992809,0.9304802727869217,0.20052737825179145,
 0.21674674841925323,0.5484142508361616,0.12937152482386383,
 0.25942940635338296,0.6651410922567554,0.14399390527237943,
 0.0015742387626018761,0.9679836346880818,0.06315705383583203,
 0.3987392003845458,0.09249428660654946,0.26660865781248066,
 0.8307560911874668,0.05175130939970052,0.4088940513307663,0.0766874133531189,
 0.7376023728775678,0.8984496308069593,0.9657180286254652,0.772819492883079,
 0.2421794712849804,0.2355456832735755,0.7350963603530902,0.8665323443209192,
 0.9520670940007863,0.883708088459771,0.3564923009424412,0.08662685528631964,
 0.13529127196222857,0.9469814178618605,0.4991109641400052,
 0.23392325292867477,0.5546693475651744,0.24822020590702676,0.535459793315864,
 0.36731527726213464,0.824766965078637,0.8969550306303788,0.20096697242205674,
 0.40043906281434816,0.7826540460589123,0.6694946544448791,0.6516955957586809,
 0.47950178517915476,0.5773228723865359,0.6538274169519727,0.7074873246326113,
 0.08405428456747166,0.45106237751779255,0.01771867017265716,
 0.3671305515513691,0.03314222409744727,0.651471445283463,0.5228781381864918,
 0.5250081191464063,0.138683546558518,0.7157734770948204,
 0.0019598651892209773,0.17269833010697522,0.27306416881539264,
 0.9737228548058875,0.6372508401224376,0.12461265162822177,0.4650054302756117,
 0.7387830943357283,0.2230930369638945,0.6698094349141481,0.07724230407911092,
 0.07510739280516532,0.751673099333626,0.00822892632126937,0.4926556876253013,
 0.7706130073345694,0.2929969779904241,0.5359571534688524,0.5689349431539573,
 0.1482930993287136,0.1153600095381464,0.24631061743258964,0.4205350552692335,
 0.4565382910575576,0.12028291979864514,0.5358581605042696,0.7594865472087216,
 0.05926228469936057,0.21005905498064714,0.7620032091241177,
 0.4842470877308116,0.10683487716815854,0.8942188092669012,0.3333218725929691,
 0.2750975054401411,0.980175530503284,0.37586093844314694,0.38388993846423314,
 0.6003609164705235,0.6491904917049289,0.5335519380225232,0.6853904479774756,
 0.9904289662950632,0.24484385920470664,0.2980349357643778,0.9867995738105965,
 0.039968616920799116,0.44914482821209145,0.2685986435527029,
 0.3459718596072532,0.27245389598286074,0.1658528669761633,0.8175310897744126,
 0.4870042645050314,0.5888782228752902,0.8741440697418273,0.5329864296022523,
 0.5635207715812425,0.10376614327718547,0.8575581927905143,0.8287551014077192,
 0.9818551908051484,0.18735911880230693,0.05259458388211802,
 0.7173214530884058,0.7674048825413775,0.33665296441863557,
 0.34951154716190835,0.7303879791428902,0.5595948750371683,0.9438962341389235,
 0.195276677549293,0.6337976702180139,0.40953653757157793,0.16643366586394404,
 0.23729330859865638,0.6647124720315039,0.8975082236595593,0.2741320269322608,
 0.9244311166617369,0.13646964874596057,0.06857084028250049,0.588292912444995,
 0.9586528211817923,0.008780940176916663,0.376711276446995,
 0.34204926539633806,0.8458886701919196,0.18953479189421796,
 0.059465512069940774,0.30776080242047443,0.8986503780488009,
 0.37075261799542825,0.537447994017803,0.6823057844683903,0.136442588242921,
 0.4327831555882339,0.6318025503797706,0.37683853215286633,0.6834103588344498,
 0.2987275661333738,0.08548501263919484,0.20959608070205626,
 0.4497317674013652,0.13310916854046861,0.9003652198495604,0.7699534003464169,
 0.10932891743996853,0.849495394907383,0.8968719645403688,0.06057747774477917,
 0.47399919870921703,0.1258818972039345,0.7795864174162068,0.8788714705667159,
 0.17285249170831174,0.6073641610271459,0.626606626130529,0.24803840679148847,
 0.727503068225587,0.6025936390645701,0.6654153393368835,0.4351588806556371,
 0.5348317064703151,0.5176064593446194,0.8567173474595774,0.8823642198430968,
 0.715832596812381,0.9065165263671002,0.3686789695795998,0.21849716546935805,
 0.016257686617172196,0.8721488535929143,0.45898618703636795,
 0.7854950834188096,0.6883550825728862,0.8407170954997427,0.5946837930543082,
 0.8822449131125802,0.9134516129906747,0.29759954067809424,0.5635362399285221,
 0.5620379337357911,0.934478496737731,0.8101343783620357,0.08860241147408099,
 0.6171103303127213,0.4787480531204489,0.5197955104417795,0.3266973648253261,
 0.7632683263059774,0.4705119667125217,0.5843044568576505,0.4615882426985801,
 0.7853938936620665,0.4037103199429992,0.29548048898086543,0.9112748773080893,
 0.6474053166923169,0.2861150096963815,0.7262601276759577,0.16247197102530797,
 0.5762775727316227,0.6615077675308362,0.7241622016369557,0.2177048888782691,
 0.7012812487856738,0.9040197844756438,0.756541622495186,0.6230359803866946,
 0.21570124086503706,0.9268531488103792,0.22663122894367782,
 0.41125628678801873,0.3594245848555002,0.14909026249428659,
 0.4013918678457651,0.5674635623133025,0.6584569529008345,0.9648053699434794,
 0.3522083114759791,0.09856697989800323,0.04526244018086856,
 0.13072721440288648,0.8917773459231001,0.04769550012821744,
 0.24241419050606794,0.5933768717450226,0.35835339922051435,
 0.6458452457928553,0.7268723851859157,0.7980287220025637,0.7645900335846717,
 0.241955940063448,0.1661071156063002,0.003117743639621584,0.5463936024893999,
 0.9612760802735467,0.801100678367365,0.8424568578936072,0.6904999455909311,
 0.5767272139668778,0.6412805578372489,0.24038458706392152,0.9111515957829465,
 0.8752442193462848,0.6876178256126069,0.23209686718816924,0.8682032969431417,
 0.6674144781396891,0.5752713151375106,0.21973385032958548,0.7871814646817792,
 0.5402072544422389,0.01154750024957718,0.3790663343804007,
 0.11464109252036625,0.6388284498234222,0.38489692116792473,
 0.0986704539214488,0.09713449929470397,0.37441576576932867,
 0.9200567325703775,0.7678255263163264,0.6021702501729347,0.4234538511725524,
 0.40526156498634625,0.20243440906195254,0.3721117631153412,
 0.12145116413154644,0.1523285172249199,0.3294070066963867,0.8544606155658766,
 0.4650401110645185,0.4575815064493445,0.38459500743315567,
 0.40061225501853026,0.3819598286239453,0.13796412547684034,
 0.32985554067139833,0.1657061239286699,0.05218893271882785,0.24357378431512,
 0.911918865773867,0.3096564512567421,0.02795635168635724,0.6668091588937234,
 0.09982696012215131,0.24986490979964415,0.8521016085236597,
 0.07779886426643101,0.9499519586285254,0.7363111629961605,0.8663666932885457,
 0.8691117970835277,0.3957896529537268,0.17900530477290832,
 0.36625421555139825,0.2419893969361384,0.8475214074072792,0.5115017706958337,
 0.13067762000334182,0.29760093105733887,0.23018712343780423,
 0.5138511222476251,0.49146483948568465,0.6842648009274783,0.5542914205885154,
 0.5037850647608652,0.8030366598759128,0.297788152870483,0.8207858483375846,
 0.05652606879751487,0.2264213816991144,0.12201911878380522,0.576056099195298,
 0.24044267704618583,0.8558687436193262,0.13769081406348993,
 0.5854474773180608,0.965599009168521,0.5925062304845331,0.4685381545509846,
 0.6319355870391292,0.49150773628216093,0.030487147217217903,
 0.36360478422676823,0.010458422055058647,0.5895046001293777,
 0.7090703582409964,0.31539402040235476,0.9847183799683735,0.5300016697642397,
 0.6498786630124047,0.06677097358872386,0.8551632701999058,
 0.46854466701596575,0.22967810409977218,0.1971019066087013,
 0.20666449384101937,0.8387368866056707,0.7158539265570885,0.6006885195690944,
 0.2874801940645443,0.9402802838768046,0.33720914749179054,0.7676110046439691,
 0.4191916720753266,0.5226965734052196,0.8607186347910029,0.7945404939384278,
 0.2802434737639795,0.1758412964535414,0.26300174052031544,0.7638691206307151,
 0.7549150111705552,0.593445230063564,0.45083541280529627,0.795554686730892,
 0.3833608516333741,0.23996679902820572,0.35640459890594645,
 0.24028882381413708,0.7984459039657785,0.6766607911698508,0.6021609862982211,
 0.36601920904075436,0.03073837181666972,0.38058472420247935,
 0.8526819679669103,0.32879545284470624,0.4555728764363345,0.9796388442336431,
 0.6268250025415014,0.13223844096620152,0.8713788731756871,0.5727380756834569,
 0.8024107675374412,0.13471126494520103,0.3896643247905066,0.3232686772433728,
 0.5724183719413198,0.10113565124907364,0.3950258695489828,
 0.22371356312653246,0.2582537614019669,0.9967070516359375,0.1971002145391736,
 0.5393609514804263,0.7872495335759373,0.17473698131028048,0.9301106321884794,
 0.24117137573358083,0.5016531844289134,0.5463011431843303,0.2133266792768267,
 0.5690645555402764,0.8328161347008509,0.9392416264534311,0.3504095249849991,
 0.08671184858968128,0.8282433201658503,0.6844484403504562,0.8843678286619088,
 0.7043421792374871,0.9668963841023203,0.651518159831974,0.955187676312478,
 0.09770983166835401,0.2622261513757943,0.0869923339888824,0.4914291677953837,
 0.935713531161911,0.5927913289579377,0.2815211114233095,0.507735349552884,
 0.8049316509114153,0.45568513425420043,0.3950271007379107,
 0.24982727034385233,0.42392479851011233,0.15628011038902467,
 0.06003727630641709,0.7441688298019591,0.0059475246527073145,
 0.1942880199455992,0.005453254947986319,0.6397287002489493,
 0.7946121313255129,0.43506242557481,0.06085401311465832,0.8271275638474612,
 0.030076058620705393,0.8854173155844106,0.4946764384244208,
 0.5432455248577037,0.5526090615438248,0.5591124058159439,0.09549148430345233,
 0.4093487460247366,0.7594697900434164,0.19780773462545131,0.7005953123952379,
 0.8755329135914579,0.48461205869650226,0.35016863378848895,
 0.6648345060891462,0.49045229107348565,0.05534201254287152,
 0.4384245389954773,0.5069859552899224,0.8764318525855612,0.7496154851723422,
 0.5693275304826015,0.6359162595432849,0.0879033325356714,0.266536802429029,
 0.21241576478081514,0.3589705800970323,0.2372017505239914,0.7200318525280535,
 0.36004807976271813,0.3208080844614637,0.5854073466828494,0.694028427988418,
 0.8300402634220133,0.703542282154026,0.39547258014242104,0.11594557696403907,
 0.14181387255280187,0.3994749427897837,0.47020308911026054,0.666381837139948,
 0.227955505606256,0.48472783917427287,0.5625882437154044,0.43849502717206346,
 0.21057194042001393,0.33956122300438696,0.927650357442758,0.834656898229947,
 0.998862111472142,0.8388650604358395,0.48031875044129313,0.35303458988369796,
 0.3441382367813983,0.2473725946186338,0.21772694868778797,
 0.08556126204947923,0.6617311354933774,0.3964351661752745,
 0.23708171004032674,0.09912640993038824,0.8477546360707621,
 0.0034053056566983875,0.9782227070473885,0.7099134809430421,
 0.5281139874735675,0.6718738903338506,0.1666431187353994,0.29280843051058647,
 0.09270307861871963,0.9992759383166655,0.9909922414049961,0.3903885629275725,
 0.23783532899765203,0.027207431601856125,0.8174764854272107,
 0.8240886267297398,0.7916698052840826,0.42809793481147806,
 0.39112129118387906,0.6380101103707353,0.4117413606788981,
 0.26214248974374244,0.07284391657014777,0.9511011362420606,0.890966192786095,
 0.9483330137034355,0.9776632015768147,0.5635138411967874,0.9650888997541468,
 0.06994056752306232,0.734588231314071,0.9082226252888308,0.4365650734485006,
 0.790991964674006,0.6369190218326191,0.28320223833292,0.2499275754177478,
 0.049085120796542925,0.10309026213925798,0.1259673179924875,
 0.0743534896928223,0.0014016483278280756,0.6978380091490862,
 0.2948276524855191,0.3571990744561818,0.41243613586622707,0.2967834434922171,
 0.5529588682727649,0.2544597583061715,0.07859821407755163,
 0.29696988253248136,0.05165755888983958,0.20083330377635256,
 0.7025945877819259,0.22631307020519809,0.4489412776782694,0.4409860658230632,
 0.9334776962778514,0.03042203008600841,0.7710796325021859,0.5662486088496027,
 0.024635488243994308,0.31930430056360204,0.5094199548552488,
 0.25667524358946003,0.7451516887362968,0.6527589206632168,
 0.12180473859861751,0.5078738844576931,0.7382352147250596,0.9592691666704435,
 0.22238870542022837,0.7786427296955105,0.11526235859297518,0.795627720105714,
 0.7998827626396985,0.4894404620072671,0.9196082644900722,0.5761167478791537,
 0.275243248906493,0.3798287305595509,0.038420506557904965,0.9191013880060723,
 0.7560882273042719,0.8550810224349339,0.8326454803226895,0.7673388393221388,
 0.44306746100770145,0.20148853074511752,0.6768598492828674,
 0.03618790873114075,0.15502815874553844,0.7640937838079154,
 0.12507028926051778,0.8510007859260459,0.12416277860589198,
 0.9819766406895098,0.6584777018360558,0.5889862932711278,0.8526504857369434,
 0.7993578598819866,0.19211259523766566,0.12646433120167777,
 0.032549764067766884,0.869120143881298,0.0997750044796405,0.7766701599240838,
 0.921850049963834,0.8414573598212614,0.13839146374942723,0.1967556749466961,
 0.29518230919300503,0.9130604263918278,0.6868114792215857,0.9427616469461862,
 0.29541901968941875,0.6682479258880177,0.32879065762712767,
 0.7067764808171275,0.6480256240759763,0.6175231328989382,0.16915926557830951,
 0.534437949064772,0.5441474567528273,0.9845145981209347,0.55678873564485,
 0.10399968196330667,0.09236802649813303,0.8962204841378666,
 0.6530234585888167,0.29113782824263734,0.11665648196530909,
 0.6176599767749384,0.3362481802306314,0.17888068238027865,
 0.06640398039548756,0.08328349129852775,0.047911959072776344,
 0.21265489785959069,0.8678409240378692,0.20373016600163674,
 0.07523113101507528,0.819110491564912,0.16305647276571789,0.6311558561043571,
 0.8930710283769855,0.659795205399367,0.9844962869636372,0.9680818273610771,
 0.777079966604653,0.1718708457611896,0.8069562636129723,0.10928113676604934,
 0.8360563535485338,0.11100997141580204,0.6515089819196477,
 0.13666924556468218,0.5005967833568521,0.9880609237875674,
 0.42856681005382935,0.5606342191384259,0.7360600283781663,0.6912988962799228,
 0.1366953863720176,0.6373124109184016,0.1045907474992519,0.9726450631263445,
 0.10170188476919939,0.728848458023051,0.7024962660737604,0.7638879099949978,
 0.043794187323774825,0.009139852013688277,0.8760800998982968,
 0.016700816581623923,0.29244552200850893,0.9796048188979902,
 0.47209940397318906,0.4179764255225731,0.2979832766588273,
 0.11066764483618052,0.5763143922388323,0.6565669947805074,
 0.04668552011467264,0.744596464063242,0.3998822488681234,0.5517785785584935,
 0.38133097006749195,0.065510226346159,0.2173840456046212,0.13822979571691096,
 0.09264120058588332,0.14102977909952508,0.3664952787280349,
 0.7947705001157891,0.08087533310013739,0.4167934787323023,0.8249484008830146,
 0.534861480926996,0.1035884508485303,0.7871116674578617,0.7503541521636208,
 0.7717182273067638,0.6241902622178573,0.4362707727044496,0.965478816486819,
 0.3487674361188451,0.6906229875670274,0.3092466590010654,0.01421195857719737,
 0.2202094286494929,0.7003081631591106,0.42970163679800655,0.3991597786803669,
 0.6574534921486237,0.8820634714540874,0.8737181901827177,0.8293058720074279,
 0.5176506401225794,0.4027367046161837,0.11612078700964457,0.8369000507637279,
 0.9872860718030667,0.9449119518766664,0.07534767764381833,
 0.33361918648671574,0.42703291258394593,0.049017045929842795,
 0.38557274732603,0.7963878956930748,0.7394729847363213,0.4589574000140587,
 0.8550665144284901,0.338597652995386,0.7721458241946492,0.6305414802804752,
 0.05170380779003503,0.9511262317895235,0.5266805604206166,
 0.055084011035509084,0.4616330530027072,0.8830447135476605,0.537155920303661,
 0.11752639199776882,0.16332196241037455,0.9943679326516373,
 0.3505922010742131,0.6775849807213163,0.6911800850543123,0.14721289560318518,
 0.07152658462728145,0.41377439763471957,0.129802106010524,
 0.06852533237056257,0.812858047606954,0.6038451089058713,0.876307513542399,
 0.057753354222377595,0.9084411115843465,0.648914811252291,0.5828474767163235,
 0.7828489080148195,0.4422459605433181,0.7724948345407345,0.44067585057846115,
 0.203253284973409,0.2848773195107881,0.40549495187257556,0.9203354930776698,
 0.8185238169807529,0.7834635737236172,0.6902081936734998,0.9467184242033202,
 0.913821989546013,0.4890885538317128,0.8921122088903781,0.272614919340447,
 0.9053730716627543,0.013426911968515642,0.1935260760547045,
 0.11545831440240184,0.022875876559067088,0.10695129479996857,
 0.14589763901685004,0.6228129203767769,0.28824504327847444,
 0.8796864336339615,0.6894767427622577,0.8288117424262366,0.31226237368940324,
 0.5104450962356664,0.9674553006595092,0.7702179334912029,0.5451335522618114,
 0.601244369369533,0.9069758627772294,0.9085397875653756,0.6878898302216001,
 0.9347021772145532,0.6929644400883735,0.49547989697544326,0.3537560057356882,
 0.5290748021756305,0.1136251091786068,0.5038081532540735,0.7227358322117315,
 0.7512532759741619,0.747891218439425,0.7735479561250195,0.5243186824019611,
 0.5225619138126351,0.4402081128942038,0.6776864072682501,0.565849993264714,
 0.593286927972313,0.6527081798872323,0.8448566786526662,0.5705262731809295,
 0.3869612182988553,0.9748917620843138,0.8487226671974829,0.13693512947936515,
 0.6501898989503613,0.6901982473480732,0.612771049109779,0.7580128728523737,
 0.34660054778555527,0.17601114881227575,0.807116911918916,0.8835729248530547,
 0.8758173940425539,0.8376925587092201,0.016395053279896543,
 0.4689322168950968,0.7766740971776356,0.5827863410362235,0.272992517195779,
 0.8135304311238888,0.06596106992674378,0.34253595484460164,
 0.35559184020557755,0.08748278094915196,0.5892272610077898,
 0.3680842659503658,0.7247656796446403,0.11428818823322728,0.1503475320376002,
 0.8921413811921322,0.32046030559488425,0.6819126417258724,
 0.06156834015564494,0.839713461719461,0.10710464854955037,0.3845651427734016,
 0.4092277967460849,0.3238237758650828,0.8883394281031732,0.6496262702597577,
 0.7190967091519782,0.6537062282992366,0.582543824500918,0.20374007753475099,
 0.23685124699511895,0.7922455393826281,0.7585940876800524,0.2182519502888759,
 0.16530677736226762,0.9360555877897493,0.9863740008534959,0.4822584611968699,
 0.5402135479180662,0.1475310180371835,0.3465770762313112,0.8093863956385802,
 0.6159959007941802,0.1951566427493452,0.03893033831997417,0.2817086722030606,
 0.6895827671136158,0.8456827193098442,0.9320582540682147,0.7933885552071667,
 0.06835102139124338,0.06515251174212022,0.7261129912238733,
 0.4568975227239134,0.8567495143113465,0.5624784291515718,0.0965675305624929,
 0.35750250901039615,0.9517544357421338,0.4405960658886432,0.4585993242440768,
 0.33376723229621175,0.9363111742842136,0.7337565340200989,
 0.21629844473837678,0.2690185047154797,0.7652871598805434,0.8053322145251679,
 0.9239113798479512,0.6981864854979224,0.7236688010754254,0.5020767557774815,
 0.008599977532363301,0.6108261626575501,0.8975573915757107,
 0.7476059931224573,0.23428043234410822,0.4943535566279722,
 0.10514216887619798,0.7577828316569769,0.08272577995935482,
 0.7594350395883512,0.169849778662578,0.5816189041756797,0.9410626426346108,
 0.24394769906595648,0.10783875565059242,0.8445432816293507,
 0.21623781677427756,0.5042247994546913,0.8438934088954133,0.2044664910202748,
 0.22013525572497272,0.8702361217029283,0.9993267077813389,
 0.44685119031729204,0.5621239082572953,0.5821616328063234,
 0.21673354705095293,0.9291824736866021,0.11199545136895928,
 0.6666615558980871,0.5701297526273846,0.3297861129326771,0.551028592739609,
 0.613843986530061,0.9671503092997041,0.13011285479990609,0.47635999395632467,
 0.5654789486346004,0.11241905273781028,0.11638042287574368,
 0.45013409340233834,0.7868726892629775,0.5325265544588309,0.7803049726064104,
 0.9404215502504362,0.22955992461175512,0.584762117827223,0.7343462168337352,
 0.38996509737589446,0.16744847930339624,0.7979969011011432,
 0.5846370302744204,0.9349380469385757,0.36724510186525616,0.3022981861985172,
 0.24616639611273383,0.8001295451740873,0.5588172705354068,0.2801489788196324,
 0.8072859625418599,0.49922731593200165,0.39264586275859215,
 0.23734288124408587,0.6656150912317877,0.9947960471480676,0.1286579350703294,
 0.7598433288686922,0.03818151683867621,0.9481943698851758,0.3938034224732845,
 0.8291024520061041,0.958228861500459,0.6554229030248111,0.6528961500900474,
 0.3237392284949977,0.04378497585800678,0.6833534908199815,0.5531042287615258,
 0.47360831855504104,0.4866765637749668,0.37158365094917845,
 0.009294431134217795,0.35564058887325856,0.9182808782947948,
 0.6095227103062605,0.7893704027100362,0.11922456257245218,0.8667283552094884,
 0.4786638843805563,0.16291932550648092,0.24237960388049473,
 0.10538829486681911,0.07092955083608188,0.8834648396311748,
 0.7861593147003176,0.11492505827458199,0.9456265009224092,0.5509505090545577,
 0.6869735962969455,0.4017013853442337,0.0836828461873006,0.49196402688512686,
 0.026045530965496888,0.21211407279917827,0.4144302493575561,
 0.900296106029352,0.8390666481854254,0.12943251573941295,0.5560618180543773,
 0.4425153995477792,0.27762202571720307,0.800482059883344,0.36332203895975945,
 0.021034955926026777,0.4006489163644774,0.4080319661484293,0.755878933763122,
 0.20015598240906074,0.497011960086406,0.7495444264191637,0.5277634916815539,
 0.45269472438507485,0.3533206570448726,0.7550144957235277,0.774828317559532,
 0.24600850532710306,0.22309826405992583,0.11514009546547521,
 0.0861699126562242,0.46736642890118696,0.9480019616454082,0.293785326071076,
 0.0753825614945377,0.28974340446675984,0.7024569228141568,
 0.17064664360708548,0.5830702709789456,0.2677478240866844,
 0.024448264366306205,0.20829665614688508,0.475141524502004,
 0.04705268377508309,0.901260720065677,0.558393000083594,0.07918073604245834,
 0.19396733929569854,0.9895401793808848,0.6878516556722004,0.9695771422769172,
 0.2844767305349353,0.22402216922155627,0.1754514049476259,0.648917093408056,
 0.37957458922760157,0.6519839347064403,0.16416619779716835,
 0.7005349523686905,0.814225676588688,0.582767992385201,0.18634136506699162,
 0.3936906459616686,0.23126787237703506,0.015441534700293413,
 0.6763344359515825,0.7742888953377972,0.9211249012244866,0.8071941242351988,
 0.7796090192657064,0.6258386378664652,0.09515572746934087,0.7384208965241597,
 0.7246030949984015,0.3378224988495594,0.7566159712806225,0.5711256851797946,
 0.2156601387846956,0.8084552649993787,0.5407490478291813,0.331121064583762,
 0.06987285021654877,0.8258755033021303,0.8293662206741353,
 0.35592320057189997,0.5608540198608446,0.07287774301450456,
 0.7974094217373023,0.8495842191948714,0.1359736271332047,0.30126796713325554,
 0.43686135246971813,0.22352181259349035,0.07225922092800374,
 0.058216970122417,0.4820039303819593,0.23765697397159413,0.5422221790752514,
 0.757569766922989,0.1497204838300551,0.45730432292515655,0.5447881337169538,
 0.7296602300185601,0.17034818781642302,0.23921827042034882,
 0.2848008223091053,0.5940753929229795,0.954783310039971,0.7434205572468874,
 0.14193964450051033,0.6939879550267587,0.4429333713484882,0.7487132580476161,
 0.0010246328325864695,0.241788066186147,0.40505031871568997,
 0.062130002190290745,0.28850651128794613,0.16158347240853288,
 0.051415356576198556,0.6182450868563635,0.959689671203655,0.6166130283489729,
 0.1438408097815903,0.24842850120290416,0.0937662918570239,0.9863542553638889,
 0.6068983276352813,0.9599798412239958,0.19974619883497213,0.3414860634384034,
 0.5614663627975395,0.5178852219534491,0.3276297447144304,0.1686366932694665,
 0.41071747611680165,0.37699033884323907,0.7836847833761602,
 0.3002410573668819,0.07584324597327474,0.16030004978827317,
 0.6379065812370086,0.4884894780170391,0.9385978444573438,0.09413139281450089,
 0.6383783369506828,0.15443799109871814,0.9302080573987671,0.8263998514885789,
 0.26793038386192025,0.2531207965688528,0.3288640905411918,
 0.29510617342544543,0.3140951755269964,0.8256276272282117,0.8612823200349241,
 0.4617352972729337,0.6929540609430385,0.7725036604697438,0.6623917710124809,
 0.3079515274597132,0.2218689749199686,0.6638733770216689,0.36399306726912084,
 0.343859391489753,0.4658829074770655,0.8723275861672939,0.11095681640913779,
 0.9377776630394659,0.00597908213163556,0.23988105734567577,
 0.39666870373551655,0.8848904781909169,0.7658976344242441,
 0.24928643889883062,0.33702866493551464,0.03955459107424275,
 0.6166083727208568,0.22467326236284413,0.5612294555969025,0.8821542953166558,
 0.7201333923783647,0.8193398088601234,0.6303338353600954,0.2520133117483697,
 0.7832396645499995,0.5697537392512266,0.7201235517168586,0.45953748155265295,
 0.24905442156080027,0.8438513499681498,0.2803561871678233,0.2846030762567281,
 0.07047766380230736,0.3158622969822449,0.5927785069888165,0.6088128179628709,
 0.09803595585076752,0.8910052193367777,0.21773414917470257,
 0.3290158608131156,0.5496536398506018,0.21704530876511519,0.0790423698376933,
 0.9721996279323025,0.21426645303942526,0.4609943687635374,0.4219145618588981,
 0.2405762764282522,0.7231711557565188,0.06737818200055357,0.8366399141284084,
 0.661903852175979,0.7935881580948396,0.9596734233130316,0.6889472053030443,
 0.8096510537012943,0.9952935795297152,0.5760032962556052,0.15741529827743073,
 0.2445678744872385,0.10102773808677534,0.9727959310015674,0.5833545592621295,
 0.37138530020300686,0.43103103137736765,0.6682365257885537,
 0.8444806289020693,0.8594996685224923,0.16530318133435484,0.6079784584710471,
 0.8738877392863746,0.5483193366233832,0.13579950121596152,0.2048977613928935,
 0.5579992123921398,0.7606774387721364,0.9691193780417698,0.6976684443860462,
 0.9903441686045414,0.2367042548775703,0.27528375942654215,0.5278063075366313,
 0.1321924763488842,0.4947710350380824,0.8954284182387291,0.8258382443057039,
 0.8006838797641305,0.2598207885729712,0.9703542633840458,0.1768778487682584,
 0.7365359338797619,0.607115590026739,0.9357169850680866,0.35772847406809305,
 0.5627282284443401,0.30249105050046987,0.7000783649886144,0.6448704028383473,
 0.7255018920635041,0.25694518780034126,0.3016610482503109,0.3679337873375055,
 0.20697219235992959,0.05769982742506918,0.3228219145757547,
 0.1347797870828804,0.8798322278949796,0.6422251635477957,0.7431927877544143,
 0.39213393788402695,0.27809897776980896,0.9059256613632943,
 0.6836692285053401,0.8148255224766415,0.45692723588077944,0.9191931990458352,
 0.6227962669432577,0.9770807470939804,0.7440669912899494,0.09831329938016853,
 0.5491624761306739,0.1839065148755551,0.6279908909634598,0.2858889945466421,
 0.4990870528521112,0.38095727286621806,0.27112781690595655,0.813123833252156,
 0.5383083635792514,0.8148345263467327,0.39910610699423765,0.975518984367828,
 0.05624960379046262,0.4102806984085764,0.8769703518376777,0.9682905790577682,
 0.9153296445241803,0.9849779858983272,0.034375388303880205,
 0.16800164567962117,0.2918905114489274,0.23658024569588587,
 0.04199734876842498,0.7981918848469554,0.14386358016460044,
 0.2862042111991777,0.4467706148323889,0.7501202327085549,0.609929756881491,
 0.09377766332660342,0.13649612499721508,0.13541412019963972,
 0.367702981311529,0.7925332194798778,0.4754251655105385,0.9255181508658774,
 0.7051650806665741,0.236793546248112,0.6711406702926584,0.8241864398159786,
 0.3151152251482855,0.7161399398156557,0.02291844615009575,0.5505240327415264,
 0.4659906284482038,0.3955293856806307,0.9645149961850189,0.03983171912148875,
 0.5839924695218771,0.9534482204907679,0.5636545317330319,0.6617934533359723,
 0.3251159347588012,0.12430267777274184,0.7672614048419301,0.877760874392488,
 0.600545570956991,0.3858753004935689,0.2667222113152502,0.7891675274942053,
 0.1184299622859607,0.28380982963120827,0.5976613826256,0.44692369193640524,
 0.3342089849910338,0.18512308329842508,0.803842667082329,0.5606277650483252,
 0.06784203392674837,0.8096996485454734,0.13923640538105153,
 0.44064852996970416,0.7923084987773711,0.9223743825984521,0.6851239015395958,
 0.012860478132451059,0.05426968048671832,0.5479824406240213,
 0.3549946824862933,0.9212838382219066,0.8252773630791618,0.7999090069015471,
 0.3445004227588162,0.3020150769837723,0.9508529148163143,0.1963292204117668,
 0.6631934714650194,0.5092713717470129,0.8787925056407825,0.8125304501569558,
 0.11191777011004189,0.8083493163335214,0.2504217315953672,0.4371426388689711,
 0.13324929576479472,0.9288736385989383,0.8367420926576751,0.9599320905522346,
 0.14247277155772542,0.8024264423286673,0.9329692340587932,0.4215588084210472,
 0.7883170447116017,0.24025185459831377,0.3535396569212015,0.8214386302966615,
 0.26723774125914457,0.09795717755176969,0.0810189929102354,
 0.07896380546684822,0.6361066496531315,0.3442963470149272,
 0.49636461886399796,0.26973586946128314,0.8858018967632907,
 0.14158959955372896,0.17481863562545752,0.06534719592424199,
 0.44604192042233126,0.7237414475166548,0.09521355755737138,
 0.18673461818745762,0.40761864093075306,0.18620987878376805,
 0.18774391392974643,0.16946559950868323,0.41440665105656094,
 0.09416052162768296,0.17988960366002993,0.03125154490939619,
 0.4609458050693437,0.13899565432516825,0.2808701723899637,0.6421959197233513,
 0.8388070539538514,0.644692277134767,0.1915745560028539,0.7720548602632823,
 0.6187540205626401,0.6167781218281144,0.38951569501670147,
 0.10648423167908905,0.709136781852981,0.5326600580428553,0.34830991288186763,
 0.1335360653727382,0.06422729296676111,0.6487022395093718,
 0.22574267368371048,0.5123279635460668,0.0689233915373022,0.6981058325903782,
 0.6507581006941038,0.3353384143371386,0.2967869374642098,0.39774810793002746,
 0.8154612564238795,0.7357952208908725,0.8133434407661791,0.07255338466864858,
 0.5511645048313251,0.2626404912329333,0.8603736123387237,0.521582501307555,
 0.5150107335819802,0.1561263975697612,0.6528874900809273,0.2720581126963093,
 0.5391233259755775,0.05818242154658837,0.35826216262195276,
 0.29752985812685684,0.006391924408520544,0.47775582946797646,
 0.8872802367259393,0.6912582284402864,0.6377230559524524,0.5089746452732051,
 0.18455618265979168,0.18271095683175942,0.44424468574865106,
 0.0891228206411413,0.5848386589041926,0.5737888531367945,0.7320613869810118,
 0.6021596686879539,0.14300121435272928,0.8327626949611927,0.2616582302587429,
 0.9293886110250132,0.015384099070829826,0.11088165638308645,
 0.7426203981760564,0.46157025525715,0.2493130533388892,0.9139421589130384,
 0.7762389552375166,0.3837528828697645,0.13546311499370978,0.8785377939980152,
 0.6240141536075009,0.876056502578996,0.6769110481579885,0.5860641400238626,
 0.8856291166285123,0.6608487467349078,0.2363710010835982,0.21294861343153515,
 0.7759034835098262,0.06451641969262056,0.6549363141166117,0.7781328726310154,
 0.8750982852189862,0.1187171939813354,0.7094626487961568,0.651353871846287,
 0.5375964367574307,0.2145169188497651,0.03914063534167067,0.8938875220164629,
 0.30027048093631004,0.4314411376915481,0.161203075386944,0.7678409741314326,
 0.02799918212263308,0.9936192863276345,0.24055850479237173,
 0.22338029023244987,0.9488456481915845,0.9064061043867464,
 0.34082899175193526,0.350015149230006,0.6015080031345834,0.525778370741905,
 0.5462641164999607,0.5242914590828482,0.8172569922780742,0.31045953808824955,
 0.06755040886869335,0.6141018046783849,0.27807651922478416,
 0.3919390738335421,0.9149396640153282,0.7775272556089283,0.737359426091927,
 0.5841864981433321,0.9489969690917012,0.890903543775069,0.6708361168557381,
 0.6887798941329772,0.7252181978840396,0.9067752666619142,0.08740995515257599,
 0.2811147934053554,0.07494905093595794,0.41878875537343896,0.630897861728375,
 0.28785909910720453,0.3764285608345037,0.7712606505985509,
 0.28331529523595955,0.5201721194909741,0.782736764336316,0.03882363862392779,
 0.4605042123709602,0.4486010653364292,0.5793113037782686,0.7826366009893899,
 0.6978311611329977,0.8860357817696977,0.22070785117561886,0.5449175943191167,
 0.3876271354459525,0.9346495883251166,0.7506079896345004,0.09662392837725686,
 0.6651638188525343,0.33436952011613064,0.02857130475434355,
 0.8261362583474875,0.6746938296603909,0.25409053180534746,0.7425894214561835,
 0.49200616060407165,0.3306944103979461,0.7910952886320758,
 0.10364209189547469,0.6827961154498549,0.71626699499248,0.6216702317618792,
 0.9554458260064469,0.47535131132011055,0.07794767941667413,
 0.20506667001752565,0.7653551167343388,0.11162463743594653,
 0.48745253726092963,0.5010550713553262,0.43829050233591993,
 0.2351760816570213,0.5345387733681831,0.18356929545021716,0.8973645466397913,
 0.4558055912530855,0.7527138449163626,0.01617946624645683,
 0.47396872418859914,0.7595439853438757,0.18688546035677078,
 0.14987061277717006,0.43152107009639695,0.15714929113062004,
 0.06922297709471914,0.704465598533304,0.449621669424106,0.9603522049667808,
 0.49762330609928906,0.6448234955617552,0.22345585064326146,0.817415918265811,
 0.902009152720527,0.1492318615799445,0.14597714043515053,0.9541477983813706,
 0.2727965346454587,0.7233776916500074,0.49416504304002495,0.9912139633693104,
 0.6357451677472801,0.4791468502535826,0.5973997276393037,0.8614306327345826,
 0.029651032230622043,0.0802513356184944,0.23677634627713962,
 0.9714090664655851,0.4018700556663113,0.5988676682892558,0.40802984092177286,
 0.5646924863345449,0.02555629265511017,0.09520158863732198,
 0.6498349234923265,0.2820659613619423,0.7032012831942887,0.5495925029043782,
 0.44300052415620517,0.022518906544643946,0.9159658587177076,
 0.12840342747424405,0.6871012218497977,0.8380068764139812,0.8687555043341968,
 0.966245197190653,0.4319546153082928,0.563464084912031,0.8800234990560761,
 0.05957801470430668,0.6222830943344178,0.36776042954588717,
 0.21801332337515889,0.9834239188499668,0.9039751675821859,0.4948438717130693,
 0.13940650874275284,0.7829420249347755,0.5606854291650428,0.8479325075234772,
 0.4294758445233863,0.9392456200930716,0.7591203705694807,0.6254094421833262,
 0.7497954751141578,0.8776376755549876,0.8499517108830698,0.8772638214692315,
 0.7458721995120801,0.19355865215451695,0.1843967729083018,0.487001600143271,
 0.2636480185330945,0.24336087546357843,0.515331401647301,0.32128336736704033,
 0.8357134577077525,0.17931981065747504,0.7319225310019288,0.5127422280244825,
 0.8759026404323788,0.26950969204521535,0.17078624007438592,
 0.23982852436066648,0.5767405453007903,0.67939248803894,0.7714683676433309,
 0.09490663171925817,0.2268800824348225,0.7636138191460017,
 0.12135688093326646,0.30510255924412444,0.012585450437714085,
 0.6963020161229928,0.09597245243462538,0.20182659574290318,
 0.5664013764681065,0.13911560807662138,0.674430685432625,0.01474055121398854,
 0.6905345433298178,0.1795302732402586,0.45877240830165533,0.6957244310980464,
 0.9055832609108864,0.5856027726316734,0.7557282708800959,0.8689315761790442,
 0.06709466629198624,0.10253280831028821,0.3629748798852238,0.77590858690305,
 0.6424845669162199,0.328859106038934,0.18044204234044492,0.9814209453073433,
 0.9778849181739941,0.5088667035380283,0.49222968022041724,
 0.03220402515631049,0.6222432065527694,0.37642847464168583,
 0.39198290066007435,0.7333017667251275,0.2643045492831423,
 0.15160810921016554,0.7220348558723297,0.8466974340524532,
 0.07697027062855344,0.5412766445579921,0.427330889377306,0.4245553576696742,
 0.04337383031330433,0.7891497464858374,0.515969824129578,0.9755044778372272,
 0.9507151460924871,0.07007789162524469,0.8370246585010643,0.2724011143176356,
 0.6929718406358537,0.5630359462541248,0.2865187132190231,0.5092412601446611,
 0.485850881098449,0.2347067945894349,0.008909223152094858,0.6491030720838881,
 0.6866972586354341,0.35049931476692053,0.46686452350071306,
 0.5930023914053523,0.09602042370954755,0.17269358910207921,
 0.36983925499029247,0.8832962997567048,0.687820968519857,0.4585393693350509,
 0.2707116841461299,0.8941602106379195,0.652818116813585,0.004411030399339655,
 0.37911834321101,0.005496677164462116,0.6317909854443775,0.6448483880835103,
 0.1741853687197812,0.9188406078497662,0.1203549298591311,0.9641740856084154,
 0.9574729739447888,0.4865401378947458,0.4450818596061845,0.1661415489012017,
 0.153439593923975,0.8391509992138823,0.9910188039533651,0.40520384232856177,
 0.5448771241200131,0.9181114294276733,0.0758151094485302,0.32643966599208807,
 0.6429094067022385,0.7856006663752646,0.7304383836270462,0.5172988113535998,
 0.8349737287819738,0.19535174032494929,0.8524198541406817,0.9345901410420668,
 0.13729682894546757,0.7327428509731835,0.29072763934261625,
 0.5637649825247636,0.04589589325113341,0.10112645898951966,
 0.4850955235640185,0.6728714907385518,0.10967129432490008,0.4075508850749453,
 0.48144981184728686,0.4123343403956512,0.6411332962424666,0.3176602997111967,
 0.471254150588983,0.31056420887821223,0.8673867258721116,0.4864180128586829,
 0.02257672718535797,0.9930720475552147,0.8412303152331433,
 0.10372642460852488,0.6341332988170718,0.10405389312943569,
 0.8642851291398201,0.8100604270848755,0.8029865241768763,0.08148713885364545,
 0.9636558419459768,0.5796657812750815,0.450784338875556,0.48674222300346537,
 0.45673637025916114,0.6182571922665326,0.5451786367474251,0.71161566779264,
 0.2359661526903286,0.13643515239572224,0.26383482687252346,
 0.18108093398960345,0.6755268918861886,0.7163369987901294,0.5018922607028568,
 0.7971082078295803,0.3750033398810829,0.08030813082202959,
 0.024541760119020428,0.14360564764309192,0.7889666078189335,
 0.15435912799258955,0.9123849858839207,0.9323314302741542,0.9473515212367867,
 0.18890112067995224,0.6381997052079987,0.6948028890576742,
 0.32642748878142946,0.7128978338616443,0.9890339983151061,0.6656958431107513,
 0.27583799625969707,0.08982676821498481,0.8276983863940091,
 0.9075974237260086,0.596222426950442,0.8883072398291935,0.8508117407875951,
 0.4488840773037285,0.7535248070663733,0.5251477964019945,0.8980638280191512,
 0.32313121278143975,0.025205125885620627,0.5207137074931046,
 0.11898336965670808,0.5469515184993545,0.5976274158628024,0.8672393056371808,
 0.7576770150852036,0.9650907926982656,0.054225997667148196,
 0.9606713619502202,0.2898852512740224,0.18872459029869437,0.5115038168046275,
 0.8461670487512926,0.1708337732560311,0.4967002681660042,0.10932744518934767,
 0.9067715925944796,0.7964721768233574,0.3788704122529667,0.38424405710130727,
 0.2563971559920727,0.7785327999954499,0.9218225796960061,0.9614515078557132,
 0.9724040402520449,0.49183489276351255,0.2015345290770213,0.3048811310852343,
 0.09529316780085684,0.027250980631279176,0.7604739120886492,
 0.22379021659676357,0.19951991185154738,0.1976950309041834,
 0.5485221530332409,0.1018176029491048,0.24164949265521796,0.8778831885803947,
 0.9716209860198747,0.06591771224898424,0.5174730812032137,0.9557082319488621,
 0.8533019356795659,0.49776747209357053,0.9174846078808829,0.9436440413730328,
 0.40944911034777,0.8689342810611973,0.34069220910170495,0.4972226470462382,
 0.24233572556976568,0.0369031095838509,0.74730121386668,0.714053868642157,
 0.0012275005738709988,0.8240443127278908,0.7640228932467708,
 0.4231822564191746,0.33586227854706463,0.20558246990478068,
 0.5584591095633374,0.7041153008099724,0.5869830432301276,0.22523347211962252,
 0.1312822328350498,0.4528772394446875,0.3855108593029539,0.494442936956088,
 0.45467162205356226,0.8757344399563594,0.04404370952276826,
 0.17524156696621052,0.8051666147614669,0.7216085782220056,0.5238768888190775,
 0.966382893041278,0.38858306800720444,0.7435815976892656,0.8070417005054232,
 0.00812187138652265,0.902986839392755,0.679402816703396,0.3537379259338572,
 0.5617185974611355,0.7921796498151673,0.8738435456014058,0.09040499553092929,
 0.867159191089266,0.8845709718780841,0.002639484356902977,0.4438751536975357,
 0.7007725770864371,0.6708033063472248,0.9799561993482935,0.3757799977933892,
 0.2346069765810247,0.5642100367530504,0.3123331579306834,0.5821053548411146,
 0.6369321008331955,0.419864537558883,0.6493770042958574,0.5468995362197203,
 0.1287329267891978,0.7268321214573897,0.5125766758508384,0.5761976525042369,
 0.9961207896487988,0.39991270858638517,0.8469211217400182,0.8484580772248397,
 0.5249900776239885,0.07418363125176441,0.48297902237316026,
 0.9093876093485767,0.4256836688146627,0.1513920748405172,0.34833176927073706,
 0.28755297778101707,0.02635857541233444,0.22306073638214874,
 0.706849885077368,0.855304454336204,0.13706635008612422,0.09663973784578495,
 0.6394398507586367,0.621311994027926,0.9405772127345263,0.3284319409674942,
 0.18840124793800028,0.9774441984545783,0.6349986134504437,0.5533881573723042,
 0.4957216508381167,0.6430400766136586,0.20536728227907008,0.5483768879690778,
 0.7789053599521579,0.958327026675887,0.11679166480731629,0.609381608636517,
 0.7702655766726554,0.6577925447751599,0.04404269881397782,
 0.013322663491268094,0.21829133912388887,0.7291208640307683,
 0.7051971574122979,0.861546818121106,0.47105516171042205,0.10088033680183417,
 0.6189610115289753,0.7227008489879619,0.32374503906551233,0.700850297851811,
 0.04868309310209418,0.5783960627861036,0.4995572356964787,0.8335892373418646,
 0.7865779553713679,0.23938937663197135,0.14184217020266376,
 0.3194992793078817,0.24186690100619268,0.6935125297930849,0.8357553860457076,
 0.9128983993997608,0.0913933565119267,0.9277431493055893,0.8862644723520964,
 0.8888523918067586,0.945539683521969,0.47081200986342386,0.9935305523556686,
 0.44656556954169724,0.5000206731104955,0.06053493439184243,
 0.30068574889458544,0.49223071118965267,0.9033927310137347,
 0.9815415663885286,0.9087367484998722,0.27746680285842895,0.2771677597207066,
 0.2405993286265239,0.004609069997461912,0.40203165263484275,
 0.11146621667483991,0.2817026811311636,0.03368084957994999,
 0.12417134604448987,0.7687818341402471,0.5028262793604664,
 0.11373488869899429,0.8347413053766696,0.44554120321978874,
 0.5184621734719662,0.9159922124768444,0.35672255123751906,0.1514065326314472,
 0.4697800891741525,0.6336577832215846,0.9391097286433737,0.5213090580943152,
 0.6925685265919842,0.5784127684421727,0.583059893701761,0.6527814276150831,
 0.811479965447308,0.32249350453606973,0.9363197928876137,0.10609732182548115,
 0.6634802321827606,0.32100806366173484,0.3014524098393726,0.1263429662377542,
 0.26931039860945494,0.20459802076828293,0.44422713708737493,
 0.4875399751614713,0.6214563161304392,0.37687435214691245,
 0.029368614069900656,0.7197009661320334,0.3681214102844599,
 0.23229868563182754,0.4770765295646713,0.6319211252416908,0.2594166439723088,
 0.03503225478718064,0.9753313763661609,0.4440132453851232,
 0.48051184674312597,0.22334524688577018,0.29161504800613636,
 0.5711087078930664,0.9388118520310792,0.22082311671539012,0.6623144661813418,
 0.3221884928790564,0.15850923415044538,0.8254121119761869,0.4942072806690224,
 0.8068726147794807,0.5014637976770473,0.06584561227525121,
 0.10902323845009965,0.8738178666066244,0.4284517918297962,
 0.08086198625259966,0.8971546362457958,0.30296502882147036,
 0.8827056821461925,0.27201632456333513,0.6481678946956889,0.3586274516946608,
 0.8897272255437513,0.03773148032686113,0.05117281065336554,
 0.5364099153404212,0.32441491620006324,0.2631877674514783,
 0.02416510490572288,0.5467067899528218,0.17425113835091177,
 0.8780510368082199,0.16997886399001305,0.2702572334455363,0.3797494924600252,
 0.7727933334168975,0.22625327223714686,0.8379963123254697,
 0.25967815096867153,0.37236647420508673,0.4940901736995217,
 0.7631583821546146,0.05508967450273594,0.4230569014870448,0.3255699908873423,
 0.6141771736225705,0.46664018800810103,0.7035914023945236,
 0.03515946570885298,0.4568699199048285,0.4133515864200126,0.7664889724067081,
 0.8594058409242987,0.8895843695941099,0.7118939916440185,0.17590995656895325,
 0.39082117213259293,0.8119554112665248,0.421474935854358,0.682816895175438,
 0.4609191742291838,0.2701698049944711,0.8328476787303997,0.25036159719028594,
 0.3111539885741732,0.037965207831517245,0.018847412876740632,
 0.8772923732767285,0.9804448007781305,0.6499117215307797,0.14727508544173695,
 0.043332488478080466,0.20558241019261547,0.24479343741874504,
 0.20741087697768168,0.016469935363242816,0.08999946278721016,
 0.09396790866979943,0.28160148198052504,0.24239590822407486,
 0.17860893843999348,0.2618599293643138,0.7528572924346885,0.4320110223383664,
 0.577951435315485,0.01026659942219188,0.6363454267587356,0.9699422808971065,
 0.7604369349567036,0.7947076946229201,0.20472069240944546,
 0.45784679149309493,0.42770736961103184,0.7541825877129783,
 0.7128344975581208,0.9052539775303106,0.3322622892502539,0.5091171511581529,
 0.44439839751670585,0.1856843237082607,0.2586698309452464,0.1509867655553796,
 0.10318588923200966,0.5007202152169209,0.9942570662042047,0.2144069427132047,
 0.3626473561370974,0.2958895972015093,0.21083969800204083,0.4853866590282925,
 0.141194454138727,0.492740928570921,0.07451646661732392,0.030978307075420375,
 0.5598969792121973,0.9974342478324703,0.5197867127458075,0.9774006739506128,
 0.3687923946856324,0.6903925019759793,0.6087632526964729,0.7708091847733523,
 0.04403028449610058,0.08886973969681333,0.1437129907624174,
 0.5910522679081186,0.8745454406584369,0.7838319949928465,0.9204411341560251,
 0.8975506722387251,0.6943148453476938,0.6848087452564858,0.16520813129264444,
 0.016085190508159042,0.4119965397593617,0.9328371017165129,
 0.14703342841167166,0.18058494497779465,0.1485750895823127,
 0.41886497964667924,0.395154543624038,0.39642126892634644,
 0.47140975665173346,0.7151772208333586,0.874745572792233,0.9932752347373358,
 0.36283041100133984,0.12203195835619829,0.6757265884160701,
 0.6014312861192119,0.17977717059058762,0.5473933781926967,0.8953789244399133,
 0.24732149492610744,0.7857610643872126,0.21035439709771397,
 0.4089425436332348,0.16134913088170044,0.055700831505259174,
 0.7242861709349165,0.039241197808337014,0.6693857345949201,
 0.9219465437349617,0.9596479267687891,0.6056601659012092,0.23785904908955235,
 0.9461102869817868,0.9874834435782253,0.9116119369558019,0.7830399510699998,
 0.3414525642300973,0.28306538884652566,0.9651063181035957,
 0.20035375249461496,0.2790829377224129,0.7042240192990962,0.6630509914921183,
 0.04150794459344209,0.21537021850422544,0.18549053685842565,
 0.570662269425802,0.5940354847522101,0.00465692954799013,0.4788948378155311,
 0.3873242076832446,0.8905911189916917,0.6897397716755143,0.619269715836768,
 0.10747342417260242,0.3664342946918656,0.07950690328642362,
 0.7424865985624213,0.8685907261964716,0.003605745488053036,
 0.1634015287430881,0.3305816249929364,0.17356630251416694,0.4673607874640999,
 0.8827323262193236,0.22058654726468496,0.3515789772554364,0.8873154061170784,
 0.050081043645228696,0.9669359316428915,0.3175554881284165,
 0.6319633121596845,0.5974088552611199,0.6952664293785215,0.44947904913281767,
 0.31762084418534653,0.6784049661817493,0.9762632578128639,0.340571689042089,
 0.6038436427140343,0.9160619353348768,0.6887356918637001,0.5884666728815711,
 0.5590540307440554,0.7103863039703042,0.10025717008059742,0.9780890996744308,
 0.43472594294099387,0.5180323637537984,0.7993246662720431,0.8156729499746725,
 0.5327135827764955,0.720981696790007,0.043801834409507134,0.2722965870187576,
 0.10158221347506624,0.708767350156452,0.4599892617055379,0.3699120678210174,
 0.460161081704082,0.6377462075871411,0.13076339792569736,0.6165869286495707,
 0.0629305428035557,0.19452922515805826,0.8404770296296533,
 0.04300735170973269,0.9190357559793303,0.4643788685009511,0.4220917780577902,
 0.3247548123675337,0.48489394505291683,0.9197602547741872,0.4813323849462745,
 0.17104961291491871,0.7485171632058871,0.3957930952616405,0.573765887828149,
 0.426613237281655,0.4692108951090518,0.33857496225004047,0.5563918027165276,
 0.43876722703545135,0.53465842026121,0.7429543787951003,0.794274214802818,
 0.6148632922385652,0.7934060165788026,0.2494922659765132,0.4863045453083408,
 0.20409949002895256,0.5937869697468336,0.5595101606998765,0.3601873506687461,
 0.8695056996401144,0.9339868080475824,0.22461857693547171,
 0.33635803834499634,0.8074293675751033,0.2269797018163605,0.6822265507013839,
 0.7731798698839194,0.1560422893999004,0.3246358457519396,0.8026852574501238,
 0.44083921234720014,0.8812862998883523,0.03452127557151252,0.926507913645,
 0.2720011609988582,0.6575849552909163,0.9588794845919499,0.5230070161349388,
 0.39360096145255785,0.8407083688624997,0.6306626649284038,0.6839856162212744,
 0.3084751771648925,0.1261561474055085,0.020240161997983508,
 0.3218509383524518,0.2755754207725212,0.09063536647799364,0.8841149755462923,
 0.4487892630643019,0.8940243706429581,0.6352007088044103,
 0.023003968006993447,0.5898029769169463,0.5470028874012438,
 0.5181919558473127,0.24813940771968102,0.45535989754853357,0.50495114792434,
 0.6231076489988909,0.44377945826295906,0.19751104957270416,
 0.07267992868502238,0.7668474729880759,0.5216750408270464,
 0.21308296055673748,0.5440869073483531,0.8424676017295958,0.2532189137650427,
 0.9543555838893817,0.012156721741227616,0.08936999465873996,
 0.6243517351304755,0.44060813662651266,0.019759390672662525,
 0.5413608867848693,0.1765217094868814,0.8733653093454598,0.2606221214566411,
 0.20580705322897974,0.11833244517712238,0.5060762430650456,
 0.9812894005827592,0.8424330410054216,0.5711606035439549,0.6506392680139385,
 0.7311769357899487,0.8581908816624061,0.5834009951103653,0.7195587657654359,
 0.7260227524493134,0.45529782917339146,0.9777205077156871,0.7916637382332414,
 0.20435093535323046,0.895249657490059,0.26230094749902,0.7057563161312534,
 0.8042369066658818,0.19023342752029992,0.2998354233878,0.24546019505504657,
 0.9942872352415667,0.6093997521438788,0.3679102340785684,0.813931820290039,
 0.9641009045889565,0.16671784558593883,0.8187601446967782,0.2717110869360304,
 0.6426080187607711,0.15610271581091295,0.16653241125360707,
 0.5400573529923853,0.947197286380156,0.5974331540071932,0.19475024769991645,
 0.49488250088383356,0.27253814123232567,0.7732850428489949,
 0.18881093686726158,0.8915767731229971,0.22846401009265416,
 0.9990903096496988,0.9294518776423037,0.303903798378955,0.7788895989121474,
 0.08560000675021329,0.8789152144186961,0.31354540693939825,0.595847440246512,
 0.08238108853317883,0.27748933385435526,0.14319329483184395,0.27031512225092,
 0.13679216794338134,0.2865641454714861,0.20171985778783663,
 0.7479928890985021,0.22508327244007564,0.8425496373956421,
 0.31557433696237613,0.6246520288720919,0.9078888059019228,0.9675159135702857,
 0.11835807555146394,0.7167751370672046,0.45482416960170735,
 0.2106306291349611,0.19702528037020317,0.5339407142247565,0.6080032641438297,
 0.9134598972898587,0.5274702719173111,0.5969437279682026,0.380559070575185,
 0.45177900868679055,0.2627773569410675,0.6293550925837036,
 0.06780401365086319,0.36961160921280867,0.4185600856499081,
 0.7091209824421254,0.4089447223852283,0.34937152854361453,0.2618878878640175,
 0.754404428606096,0.9064869389195893,0.730512811953508,0.9515502318410375,
 0.06900378619506131,0.9873019121918513,0.4446289706800576,0.3317113632814823,
 0.3996319704215763,0.26435401295412253,0.981818215861755,0.9974076261811308,
 0.21140234530012725,0.8287227432374432,0.09562160019773558,
 0.6956428723426887,0.420678446944055,0.010106159527804914,0.4939657400462213,
 0.8103216764837717,0.902810666333304,0.7422136395924355,0.17193750707619815,
 0.2952531536567796,0.962151058285404,0.8685579787971883,0.4306418568465511,
 0.32200074096393405,0.958223694545284,0.9834649562358153,0.3187291144818418,
 0.6723992222954296,0.08237440423626419,0.41811326380646585,
 0.06561707420630447,0.23105098014944114,0.8739980383304228,
 0.5768207233345575,0.30809248250362353,0.5090500731980191,
 0.08049547668233148,0.2482304241193527,0.9979086220760633,0.7160113020228811,
 0.37612915320671125,0.57571972787954,0.42781920643102367,0.39913884443866954,
 0.41993915350244404,0.9601661955783364,0.18197328686677428,
 0.8228436623811552,0.3672425161276861,0.8712904617774849,0.7072485273901038,
 0.16488438648016934,0.26184989773377776,0.4377932845324827,0.84764206821333,
 0.6314367881271713,0.6842423132600768,0.15407571435194023,0.5815454496460953,
 0.058416947918282336,0.5883301466096421,0.5595529168879261,
 0.4580896991654154,0.2825213889731091,0.15207277296998523,0.8835842118036946,
 0.7916647603843049,0.4086421323043852,0.3896305532940917,0.9952563087119801,
 0.5166571012542596,0.5617703887370602,0.736443353069755,0.3007485493389189,
 0.5192932561816649,0.8398876889725042,0.4414598592187218,0.7100351565084888,
 0.08602752878398001,0.8961415705560319,0.7135308963416324,0.7744522896372799,
 0.6289885589890734,0.7754109689129569,0.8257539692491924,0.8410811019401009,
 0.6255135252937352,0.5327921897721907,0.10865361389726713,0.5699413518702823,
 0.6564999623956583,0.7519310954038939,0.1779525816817602,0.21412890600028966,
 0.39354408971993016,0.12460117969627549,0.8896802805559328,0.598162342507561,
 0.938551361737701,0.9242520554523947,0.650219180745243,0.48887699731961765,
 0.6392327725867527,0.6076790902536595,0.6686360549612591,0.7194898877316108,
 0.9344880245225784,0.14892534219123166,0.7299841419002704,
 0.42260604850346395,0.5271120029146684,0.7110630737925007,0.7167960089223705,
 0.9926959522272066,0.25513530078874236,0.8589391794080198,
 0.013317879817642875,0.5885528073464924,0.8963977647118977,
 0.9466185367267616,0.1651903237420076,0.1271856090404846,0.0727927403026919,
 0.4465854455466597,0.31709612867363657,0.6698585247953415,
 0.16266554341411021,0.384890798906411,0.36144455593056535,0.9033961660620417,
 0.0933220695503334,0.3845496556737157,0.6277000404416814,
 0.009574621586840082,0.599318608982991,0.732701916017258,0.30054888591963314,
 0.6004639682828374,0.4984549983847554,0.3518167127043289,0.48941326655055717,
 0.011422441908768066,0.8473670738605544,0.26477119518249326,
 0.05892296004347741,0.0299360938803499,0.45942374062592795,
 0.9013691979156612,0.4305095618276319,0.08285495188207481,
 0.07861887288651992,0.8921268291146299,0.5995539655323556,0.7138453836978348,
 0.03347264199928823,0.05910110902790855,0.3716758387021004,
 0.1414844096445701,0.6192405870500756,0.759014304351687,0.07011567132153729,
 0.31428678034991075,0.2890338677523625,0.5046814279697772,
 0.12069558622939947,0.04769210978219007,0.31381917934845893,
 0.4551517436078609,0.23325299031948166,0.2391297692977029,0.9454404537398914,
 0.46171486249117777,0.40599630334706116,0.538301841265074,0.9980116405649505,
 0.17318348651012938,0.8495125064585796,0.31066237707090816,0.698691366376959,
 0.7325659242466538,0.023380126403086066,0.8419910221067818,
 0.6222985122063719,0.5348991927736793,0.014189499895124524,
 0.35051008757200774,0.654020614847758,0.007636248607554075,
 0.038829503477838645,0.6409634949564728,0.16075845731256222,
 0.24791274637842642,0.7796505522382126,0.6083613038422253,
 0.16406335747944856,0.29821696224110716,0.12720589849350938,
 0.8996087627082343,0.5928530632491534,0.5860448510128545,0.9573414059693288,
 0.5371937234214668,0.4286492817744044,0.23813778358921,0.7565178352603569,
 0.7752526871907959,0.2704715468958423,0.13385992988092466,0.8848778692451009,
 0.9916278771662935,0.906792586232794,0.9489957835170347,0.9027695075527171,
 0.3282008250326105,0.22942112189129005,0.5071970196269419,
 0.22707829526923295,0.6458150028183413,0.8676921024971148,0.2970245977970284,
 0.6225932798715405,0.65047182081728,0.49535015390684567,0.6503053668365809,
 0.8289696812473938,0.6623904854613658,0.1472535572274648,0.7321480304299711,
 0.18809406340338808,0.9540443812069128,0.5817376359567095,
 0.11519595184720222,0.6048841496589895,0.9109719195497293,0.1352061854883463,
 0.9475546319769149,0.6413443930086364,0.24746471524985414,
 0.06716696862705229,0.7350681735086901,0.291164353789527,0.7193691665880935,
 0.18010881843173787,0.005096441396188367,0.02102093708541508,
 0.8478191524610088,0.8659440348557796,0.5250521685208381,0.25522232754833996,
 0.4463691007833315,0.6221981759027944,0.7228901679962613,0.22933500269320728,
 0.19286268208716217,0.5771373959887764,0.3913203687700645,0.2179413622552403,
 0.3754672169028165,0.9395720844731693,0.14666414503620406,0.4568986747591668,
 0.09840927584451375,0.8394155353606942,0.9235564672022958,
 0.43362443283046326,0.04627287920968515,0.5127660135304055,
 0.7958403171780777,0.6842669794149702,0.34534572230253713,
 0.43430299731932154,0.6968474209270399,0.45321299368517176,
 0.7160567077666835,0.6521411106518457,0.08907010201663212,0.9987067614009416,
 0.6193312377161141,0.42822814371536244,0.5941230036394738,0.3954073716452011,
 0.20272887306002318,0.6137032031938625,0.7564776368359837,0.9616931957280355,
 0.0984776855365328,0.7027444763775438,0.7319657722875648,0.4563205588599608,
 0.0012645388026575333,0.13315370719167618,0.21455094075915027,
 0.7756545050952925,0.9606741381417356,0.8315954660280003,0.9048950540163061,
 0.9902912914779857,0.3827863580828659,0.8194696556036404,0.9945277841750495,
 0.8846605811705698,0.78623441647619,0.5427320729527693,0.7196330137592442,
 0.932881695596539,0.15873757060027227,0.1430414500144913,0.11022596079906632,
 0.1816395304921462,0.19867048964774148,0.2396726112573514,0.6749817210744437,
 0.26730966106626486,0.6517317488279968,0.019708326992698977,
 0.09661007348374173,0.6324994139797775,0.40035414510203704,
 0.003601697002375226,0.7253756435121504,0.4667508094170245,
 0.3404727188231318,0.14063995061054013,0.5193036952078026,0.3923142126491572,
 0.8309472636582241,0.00877526611657764,0.3352887354429088,0.5603459869253038,
 0.07856229493949862,0.6107405133651818,0.2905383182065182,0.5338591034069664,
 0.6539854314817434,0.36201669143944093,0.17443761042481354,
 0.3013951655441356,0.6651560747571379,0.361834620293647,0.6634598668393896,
 0.7947436960809787,0.7554956986805619,0.36844948351365925,0.5806039578402039,
 0.7840525041370664,0.06939220453781847,0.5163721849942375,0.2466534193809029,
 0.8721135872484226,0.3520023198844747,0.6365762824825028,0.8450224293950134,
 0.9197168682128954,0.508206422532012,0.7033072036619866,0.957133870755507,
 0.9375879503367717,0.4565320639391105,0.9978061733460803,0.6205925101390751,
 0.526694058902204,0.47515499420627616,0.762447205373577,0.9665197519798522,
 0.9232015673824703,0.025361906994432193,0.09243541738029959,
 0.6133585361261445,0.12142932609850698,0.258209664299759,0.19358740617764225,
 0.8524502403912256,0.1080168771829515,0.061894249616310626,
 0.05464729068185903,0.006599142123942969,0.9200457360749968,
 0.6581095029008652,0.027685498240323625,0.47851341037530626,
 0.7779137078436236,0.5183463462381372,0.9792630289083915,0.09099905679757458,
 0.12366543286575071,0.8615942101966735,0.38996363492094244,
 0.47282848059052895,0.034666556636640755,0.2938681335258332,
 0.519226071298984,0.26700901633662233,0.511021908536807,0.23542784451049936,
 0.7420388871775504,0.25291431232356776,0.5647680846274171,0.3187038724977871,
 0.22426830821951904,0.44941951803777536,0.6405882674451734,
 0.7494266714805511,0.18516839549787933,0.3882311981789546,0.5405848926986466,
 0.6995940081558076,0.7332572230897446,0.8787513946878515,0.725707805007069,
 0.17113160503201075,0.9030811853495251,0.6777179217331616,
 0.026619657610249536,0.22141449939439206,0.9930072268915277,
 0.16267767656667909,0.6759690121006043,0.04137487351059099,0.70817769477655,
 0.1452667031690138,0.0327056983654358,0.46996971255795783,0.6938288760428961,
 0.6788138246022497,0.826071534420445,0.6704293392139526,0.7491551322720651,
 0.63379286753391,0.5805848579121139,0.15544532058273874,0.6929279034874021,
 0.11915125179608244,0.6550533149632001,0.5390670468043632,0.2808688332156676,
 0.687919336991375,0.8983211848471654,0.820402145236891,0.44201703325033737,
 0.8949321463112034,0.6862186646489017,0.39084262559834215,0.2092427074152179,
 0.37799923767601906,0.7561677739792146,0.9823810532927156,
 0.10069553295409828,0.24930502058349024,0.34426206739646115,
 0.8960071551084091,0.2630631154902028,0.08348466338853289,0.4732687888399765,
 0.7729620511481579,0.6682057526803493,0.29327243154846505,0.5561979345450436,
 0.9631780954790652,0.33510579213702674,0.39540482830833257,
 0.8944280223496562,0.6009282134172647,0.6726317342349829,0.8508720040066918,
 0.15197636518240198,0.2146285226897009,0.9756323318065823,
 0.17127971926911256,0.5566775540080076,0.9233149418145279,0.2928367491652978,
 0.3901784023978142,0.9625566962113458,0.7799687360843656,0.2758026466680352,
 0.9296703961008403,0.7215138354018233,0.7588923532636794,0.7233596036738533,
 0.5567518887181457,0.6049096457991385,0.7236816336235679,0.7417598821292318,
 0.5640058429302912,0.4791059071601814,0.5083249548561521,0.4478575916135358,
 0.07333542682042238,0.8062637700902058,0.08574677075088966,
 0.36639519844789015,0.40776531328238697,0.9206147855640955,0.702925771134014,
 0.1504710900718429,0.31348528988436264,0.6146701115674628,0.6605110522933254,
 0.45235596463761285,0.6171639886622735,0.2619902606125465,0.5840044356462462,
 0.4362762582553241,0.14779026224640224,0.8694863076082118,0.3866849231399129,
 0.657623281789499,0.580365455186967,0.815709916057308,0.38523708395879597,
 0.3696605117740732,0.47619650570207817,0.7455315075576092,
 0.07744719113557452,0.715957711647318,0.7875077208743618,0.6100403222114364,
 0.28888652595106934,0.49792273127102005,0.8529587044239935,
 0.9866391250313638,0.40497717812320777,0.10070889444630371,
 0.9198374753716766,0.6783582707647401,0.6471578966499996,0.5516029922954296,
 0.5448800243236853,0.10093005089190077,0.5709649110935082,
 0.17131305598949786,0.7111416293554302,0.29604138961705617,
 0.2138363599517633,0.10601372347745708,0.49863805325940047,
 0.8619559795747488,0.338562854060074,0.4096461218528229,0.08675439881452407,
 0.62452924031087,0.7782045943308981,0.7656769827654171,0.725517336985418,
 0.07326072745674095,0.9629584621990518,0.5996853419425806,0.5883529519785823,
 0.23403948654369533,0.497665383372521,0.47798462144277476,
 0.17340187490091485,0.9565657410984858,0.9542531751333114,0.9045971119865347,
 0.036205707299790335,0.45811870734797877,0.5145513631527097,
 0.8924903523076085,0.24218445856807325,0.7869954883677407,0.3738544055984915,
 0.8541005522469368,0.6374370065466239,0.2683874543490474,0.6582422307739623,
 0.415336204743631,0.946165081737969,0.8205164799692816,0.6136301913975576,
 0.3485529262921475,0.8110740116165758,0.6315982931482192,0.36479928042465437,
 6.809016685861424e-4,0.9912732469517858,0.8802453155001412,
 0.9075276224510745,0.7453355372822866,0.68315100013448,0.8452707383439839,
 0.18139452900198472,0.4832806676043736,0.9573176854370177,
 0.053648190703927345,0.024961933796977975,0.2154145198488966,
 0.5296644444994872,0.0395881173279099,0.9812290049487339,0.22768336290297364,
 0.9583292761964881,0.5173719831365229,0.2803405586764389,0.37123652229409976,
 0.3642404914874696,0.8600376771173934,0.8255658564709323,0.15404731666728988,
 0.34779249972549264,0.9214659026944533,0.8791418867217665,
 0.18262244444342512,0.3342418515110053,0.1380967607156217,0.5855379112891114,
 0.18159141295711378,0.7542487745092088,0.11053131221536872,
 0.2926688907624272,0.6158291515670716,0.38193338247641184,0.2991923034374938,
 0.16695576598398065,0.3132003350402546,0.950646836181092,0.8560364558785354,
 0.4134271425141932,0.6380856509194697,0.4132894046847473,0.8475560473734619,
 0.8676693523033647,0.6597973036052478,0.1365910723344591,0.5612199157891395,
 0.27195219939585213,0.6327515657231475,0.8674559169767332,0.7314552639881381,
 0.07828176113778107,0.9412530310057602,0.9647716117463504,0.5715737583023933,
 0.04497056664638188,0.54402851327437,0.6003270247030227,0.6817970356175617,
 0.5087985611173638,0.14156359355283665,0.7837964375893698,0.6506336532145252,
 0.5707922844120666,0.7925793950301738,0.44424687095465565,0.3002006477612381,
 0.1693854231415045,0.8991159203613742,0.17222492048789428,0.3065394015534859,
 0.9903480391713265,0.9955583477466662,0.2607756855182314,0.734089880965121,
 0.9649738862521717,0.5340286262030707,0.08896859560788284,0.6792713042822255,
 0.21414089572442585,0.8156607962386782,0.9509870866680378,
 0.48784097608871224,0.554971993398984,0.47977721638339754,0.6006952815579352,
 0.7680099273315941,0.8079455143942083,0.03198053210265828,0.9066056536760194,
 0.24661218317998845,0.8640156027735185,0.46871566110368357,
 0.6147628639763556,0.8806524908308679,0.024968594792705412,
 0.8065990974771025,0.8806685765933436,0.3530490078284916,0.40755072740083054,
 0.7436165101702858,0.49630100932432475,0.08886163773414002,
 0.3225421443190344,0.6444658004109034,0.959319380307879,0.2597423355770503,
 0.34963201946121847,0.09514672525509482,0.45601149968042587,
 0.14333978028794203,0.4027551635988844,0.2961298777631962,0.6159269401616025,
 0.869907074567428,0.6425090082857208,0.686976583237634,0.634245662886923,
 0.10103790693347614,0.35078655201668274,0.4321995292437648,
 0.6540930052200274,0.441463345114655,0.8293096071611987,0.5203524907154615,
 0.3247572244350613,0.724905464368223,0.3742076515352084,0.721689062055601,
 0.906175663365339,0.1986327476356975,0.37035134142088155,0.5430819540488538,
 0.502845568516359,0.19219092800596393,0.9743320005465618,0.26625677413903925,
 0.28765048906117097,0.7558139951442959,0.31915034327737457,0.60701264841714,
 0.8217832738499964,0.22729791787919607,0.15738735427262884,
 0.6805342116155328,0.56571564551375,0.006622533786059703,0.4908538232209807,
 0.4413660011174896,0.8496621792353055,0.25003577437964053,
 0.02191454018402439,0.18692448034450426,0.5878075913683062,
 0.41152949202909406,0.4762503729664609,0.7161186670329407,0.9535464978506293,
 0.6595450296835283,0.6941115093129946,0.7298024276583367,0.15210577725808994,
 0.43241834398436585,0.5371638505055417,0.6157122944734679,0.9656766744371196,
 0.111345852453692,0.15381898404279382,0.8644745417995988,0.2277106483560235,
 0.281318308979561,0.1973714191277094,0.043374661184109264,0.9826423509970592,
 0.3943291040247665,0.07208998314600112,0.3324056990943117,0.9679251474351276,
 0.2801260310168301,0.28583888136030755,0.13323877283651941,
 0.32032719985193137,0.6891010106981237,0.3404910718457489,
 0.36470613030720145,0.3673075636645453,0.8968984966459752,
 0.07522280568349249,0.3729172275408905,0.4967042911582905,0.8765884277330206,
 0.17906159909597685,0.9825902471203067,0.19705887319596127,
 0.5449151273833562,0.6401502008865515,0.13895411837126836,
 0.004904009360653161,0.8588467375057162,0.9932456979267776,
 0.2851880916727798,0.5817265017310381,0.38893764894493155,0.9630338478118993,
 0.3482660309041311,0.6190674428018308,0.759189098996218,0.8649799110559692,
 0.5174807201749816,0.09620982251069288,0.19794765263882097,0.306797943800617,
 0.641309469892903,0.9694105971400234,0.84067415127042,0.09282938021095988,
 0.9406234845062567,0.5672548301427703,0.2858109824932398,0.0821989604880552,
 0.544392128348596,0.3680035084392006,0.5938291525916993,0.30740808008151776,
 0.9368617675867571,0.3583268791999604,0.37603965344238277,0.9360090480334675,
 0.6522417912361194,0.20610495759430314,0.21259085415506718,
 0.6987235553095755,0.04765504669504306,0.611636715004481,0.4838302764064002,
 0.8289925914666568,0.9840972553949252,0.09459445829527968,0.4339651094131949,
 0.9591891802192708,0.05514642224174304,0.13591404394899398,0.735769453347884,
 0.35597141862532655,0.4289095112558794,0.26563845834584954,
 0.5224744569133705,0.8502018000751264,0.2976329701981678,0.6241230684840243,
 0.09981825161069169,0.9346290420040043,0.6064880514853461,0.9315168862298169,
 0.9136412254355871,0.5402318229299732,0.4421941834852441,0.8941300501702676,
 0.8725704839292384,0.7924900366804422,0.7211823527889112,0.5846980453439499,
 0.34431852953644715,0.08292171541852789,0.8455139043441073,
 0.22143856735523026,0.883790897713598,0.30952784092496266,0.9703131955961461,
 0.2584687935264448,0.6981657814262635,0.4938599682885204,0.8142838524815902,
 0.12455638139971925,0.762031334601862,0.7982015967829892,0.31069640950796473,
 0.6060828505086794,0.6694446521745232,0.3896910848977819,0.2621696581195273,
 0.8907554695310858,0.9348500779726772,0.26195133572775353,0.7121024834003659,
 0.8295141207018606,0.18725574991261384,0.6824999948724656,
 0.04099046935802675,0.31827814441782365,0.24930807575818892,
 0.24432776486504904,0.11397231589843759,0.8036480468782043,
 0.5476309026055685,0.11674156893518817,0.6354479968346765,
 0.20336407807109114,0.9107058127138238,0.5903787598350881,0.879450512929494,
 0.1304406949790632,0.8508496714387448,0.40749069844120434,
 0.12665854490810646,0.3692051810349123,0.31791245875328966,
 0.1942578244950941,0.34926833227574783,0.11363456691985352,
 0.4300818208123376,0.26322095041843985,0.8283848102440711,
 0.26764282468535894,0.8689357299940295,0.40816878011562907,
 0.25038475405396987,0.04044329957568049,0.45971657436335034,
 0.45732439578375683,0.04142837971306079,0.1704914944113255,
 0.8161618596614044,0.7270795479324359,0.7569616075667291,0.3753990812926937,
 0.6188762048814779,0.03612147628012008,0.9003634249136673,
 0.26494719925292354,0.9929341864130784,0.20436770568988505,
 0.0779729417203685,0.5224578668367434,0.754056540490533,0.49301630850265765,
 0.3577329934613688,0.0630383228432645,0.9755170690397037,0.3443314064023839,
 0.4853195558425999,0.10725905778992684,0.5016741681538214,0.9016377082559613,
 0.857869447860531,0.9134436855808814,0.3938124721775341,0.6645183922187539,
 0.3383670728463577,0.1446890832637141,0.31537579036141095,0.6879977385032486,
 0.5248188242136655,0.5248443774942765,0.31561965073896747,0.5008497795759643,
 0.9915736444715948,0.6563539283925797,0.3266136495199643,0.51626623075519,
 0.6068848401699576,0.9683466771492544,0.07543789575668902,
 0.08024318152408783,0.1653623693602948,0.7124823695007385,0.7881761204667741,
 0.9749295627172764,0.603198179939419,0.9561236955617369,0.7843585240381065,
 0.9875703780113091,0.34920332507967466,0.05483690228737659,
 0.9018197600093906,0.3466389350239497,0.3689202732261756,0.8873993845297526,
 0.014049351781414643,0.5558223116498437,0.6923038191275368,
 0.5026224323264749,0.43622148368566394,0.4624379234011533,0.9333484701983451,
 0.9741235968719663,0.4791292479781577,0.8235747605709499,0.9663651384017602,
 0.6411642301695711,0.41932512897423113,0.49824438518115566,
 0.0237370399244462,0.5061896600547523,0.30242957767581036,0.7828868482926851,
 0.6247438636745074,0.5347931249098176,0.5447340337039885,0.666584465618565,
 0.6863649461711938,0.7667289202363275,0.935763538556184,0.6638442345142683,
 0.9100143544646707,0.40713612371239094,0.8753102895647916,0.9347669313399614,
 0.5205583631800648,0.18872840118095968,0.6071746332668269,
 0.23514516005371033,0.17789834458321452,0.8904907047859062,
 0.7152268289418684,0.7158768140226428,0.9013630845458284,0.43820354939972783,
 0.01969821128645277,0.3067147347165883,0.9079832338723431,
 0.29153547068791485,0.7474999562596626,0.42160540701001925,
 0.08233893588274244,0.7829817533169009,0.03306380553791699,
 0.2336826464722772,0.10495298406888887,0.06668369135622254,
 0.5808714475626638,0.37821367381901294,0.6298664921261872,
 0.17435426653719954,0.4216651965146062,0.05787701077693708,
 0.19072985115282992,0.6458755782820921,0.8389178411829017,
 0.27958072881743645,0.10601989557851965,0.7621429818904958,
 0.8630104595712687,0.8084488004302404,0.7817399534656474,0.5954233047975688,
 0.6691005252286177,0.6264505716006977,0.7470411085037856,0.5256812174256553,
 0.4173993638291097,0.34088748401008717,0.05509170449411738,
 0.38194574717122975,0.04059901386948894,0.3271003839628067,
 0.11620973404911994,0.030551298308994723,0.9015743535201656,
 0.33029532607253653,0.22004872921950247,0.9402994424159633,
 0.2738956882145205,0.02944502616289313,0.5902801342618412,0.7299175815557359,
 0.8743008255933584,0.2914406786462491,0.10417117454119706,
 0.44077405714232765,0.5075619164472444,0.14597686230776963,
 0.40350916941873927,0.9421839527182558,0.5623639185024711,0.5351151058321658,
 0.42051181419392847,0.41622466998255525,0.658711981451067,0.4091815326354342,
 0.8375656197573257,0.03728645716935142,0.4136076212233615,0.5048392641396452,
 0.3970784077165912,0.5635251876853875,0.29867536718237364,
 0.19170325104133468,0.9135275237539925,0.4438516459757227,0.6582579284766441,
 0.6199577653946362,0.8353889769910456,0.8097499425783078,0.9530675250450192,
 0.8045058513942197,0.051440949891047305,0.8845038542222656,
 0.4337531961688711,0.8418401246135232,0.7709263978330647,0.8597512446870637,
 0.9594534205539584,0.7549497458968614,0.7146389935573344,0.6403551347086446,
 0.3617007555777939,0.3500663075919407,0.6603193914364314,0.20399894129297458,
 0.7975720964910936,0.6454742805597774,0.22529468932913255,
 0.06803809901301405,0.9628088585865734,0.4922308310831951,0.5414282585500557,
 0.43204677858444396,0.12068198520691875,0.09639706247938418,
 0.548994980792237,0.9123652519576078,0.6027549059377941,0.4508693389631865,
 0.3559695252602644,0.35760161100961274,0.738873639751675,0.149784367659055,
 0.8841397142960662,0.5011971078273741,0.6076155971417377,0.6486954100347014,
 0.843398810615203,0.3996491660164794,0.7667780741530712,0.1193590102088441,
 0.5544960078823771,0.11835969879730501,0.2904600898121048,
 0.17117474823575973,0.5601836574239297,0.17899580907055324,
 0.8022162387024295,0.8127690163745546,0.4095555744895123,0.05624843665174417,
 0.8260165882500452,0.12811146529263973,0.8707200349255244,0.3225528975191929,
 0.18703515898522172,0.9929800295432623,0.3891719192117664,
 0.35565521696248137,0.23918922594015446,0.06352904679314231,
 0.9259094324548212,0.752043889905417,0.8182963822320516,0.831934696279881,
 0.6254071948469727,0.008206677999853529,0.30493891533482553,
 0.21849555432905077,0.2095929081990362,0.8834870467559475,0.6903405745737218,
 0.5148937779688602,0.695269000786784,0.6701173253528058,0.9906700548630298,
 0.7398534346147723,0.06959403301134415,0.059535006102769295,
 0.636340359503049,0.5710257364539383,0.4878975452583131,0.46514748579446374,
 0.14285139362810728,0.7487864543620075,0.6483260018680248,0.7749298435907331,
 0.9331521739393092,0.8038757772077524,0.21211673761324312,0.6654611473195324,
 0.48011746913619213,0.70731528421215,0.873078967922566,0.09265350947567519,
 0.39544889714248554,0.8389288309329158,0.06846961304309263,
 0.13868868873637596,0.08377942232946689,0.09785706929311744,
 0.5982336611788351,0.8078454241202008,0.27569471620164865,
 0.12417011843549053,0.26389440863337876,0.7582734434391958,
 0.9082447602767624,0.7388667695450719,0.24647837986677945,0.7683730622588031,
 0.6320424387644928,0.17915110954859015,0.6086333704685892,0.7518353906404891,
 0.5396248196114664,0.12028003086433658,0.6265423316498296,0.7516630529672965,
 0.17116001598006914,0.28189377813937644,0.41915145321151914,
 0.4918667180584403,0.36684339832626534,0.9884838435777868,
 0.17498427438809738,0.5406524243821855,0.34523727921476965,
 0.8381533565205704,0.6767448665131489,0.9516754039516014,0.3038634287098533,
 0.34202897590365944,0.6576074733518037,0.6616516048819026,
 0.04150536429126328,0.6805810725563787,0.34490729728576197,
 0.5698974793726663,0.20139531370624464,0.9129358883068626,0.7302131136457288,
 0.09474876551231082,0.2786652494718984,0.33243283186850936,
 0.6412823536365779,0.6015930813835481,0.04665150328584344,0.7649931488187866,
 0.5082560946070914,0.6305450708437826,0.2387823896536666,0.551652049038166,
 0.4575080182565594,0.20862276208541697,0.1285337775399511,0.8919283204768651,
 0.33171091954864906,0.5440237507504049,0.6647736591508586,0.75701820085304,
 0.9013051488530891,0.8975262408002673,0.5034783823245526,0.45697883122537153,
 0.472404378036853,0.4822854739990886,0.7050463064870167,0.31987669041254485,
 0.6369213052555617,0.489597498005508,0.5138349825761691,0.09756999186750692,
 0.7612202514121794,0.37042569241427936,0.04988913551699059,0.934671580981706,
 0.6107904169406985,0.7466614623887636,0.06086496822603249,
 0.35332014995380046,0.8083546972591314,0.8290893939592419,0.5626303699952829,
 0.05035698646961739,0.5671839970568832,0.4880703719599878,
 0.45884618658448806,0.7736699864860654,0.6022340371099375,0.3231404807875107,
 0.6374998060847733,0.11157449668066033,0.460235417775237,0.6502989134577264,
 0.28310203575903947,0.13973686613221847,0.2809615583363284,
 0.41214228827350974,0.8174205687062293,0.5870077623121788,
 0.059124316570795976,0.8133521571896066,0.7304341411940003,
 0.08512865938346603,0.6149435783930983,0.39057444748463255,
 0.6336309693162794,0.24544795860194135,0.5587826357084692,0.8631715146141947,
 0.9328262416617482,0.09140479669911628,0.8321065262137879,
 0.23338666388723972,0.35953229971574896,0.7240556755714949,0.19859302144169,
 0.21913770253240794,0.3264909664612855,0.36708754220436757,
 0.7537894773109584,0.2751885200783296,0.28404236943752925,
 0.05981578237043761,0.4220584768079897,0.795815795659661,0.25620670814495305,
 0.04755730678736736,0.3281052995247121,0.6742535169120722,0.838069003521052,
 0.9190854369656587,0.5305378667172302,0.6641150825758908,0.7931710677651136,
 0.09088744388948622,0.6635463283876395,0.0055067416458610685,
 0.43612896166318993,0.3340023042285174,0.8726757789425936,0.9888155384537317,
 0.17378007586345412,0.4551172862858516,0.4237463368772314,0.4415839120090169,
 0.3979929230911413,0.6560281987840206,0.5824491913650314,
 0.053912387020075725,0.8283938146349555,0.8448072938640656,
 0.14313013837082877,0.7966713676473555,0.8146310724962886,
 0.03279575126064749,0.7534124489298618,0.3045901361644824,
 0.22765851371507084,0.40024351659477064,0.29913180272619444,
 0.9739571992922901,0.9276513465666218,0.828289147930262,0.4221730698212576,
 0.5261537816458203,0.8041980012223845,0.8293374859745632,0.148931592380524,
 0.2566897291417912,0.2629394284771046,0.6126959436868076,0.4164187424463506,
 0.5634219319352387,0.7840719428457268,0.6335445188974596,0.15066667217457508,
 0.16789763812014435,0.6428233162915358,0.4988717931433766,0.0319679018079696,
 0.40203863434259035,0.1448914144723059,0.9118329096076485,
 0.11473127937172434,0.29983660349280417,0.7409497190478734,
 0.11903297715024752,0.9548396289123966,0.8224542629370127,
 0.15899193147009105,0.06512630159085597,0.332684396576998,0.743900751611204,
 0.2664721529797084,0.33756767070937144,0.4650748647217404,
 0.45132387585451683,0.07938503999430613,0.9922923924711337,
 0.6172746716184565,0.14099938579695503,0.6317999743153773,0.864356121858246,
 0.9447648384388445,0.5245453002866806,0.05515989332629179,0.2009555757144641,
 0.7192535253657107,0.20216185853278945,0.05081569447979928,
 0.21145986112690374,0.7193282136799378,0.93017778797603,0.129612981546548,
 0.05973862567772348,0.1159081466090972,0.10303316739535961,
 0.6637554530448965,0.22210660846671626,0.8121642771616837,0.520922389392607,
 0.26601673038483575,0.2289972676461487,0.10082242045019085,
 0.9197213360536989,0.9930632501744047,0.7305706377160837,0.28174088492677407,
 0.498951187767499,0.7883050589735134,0.037899225297394534,0.7042662482463218,
 0.5068598699932808,0.09253673768506898,0.6991745517402741,
 0.19566902774035966,0.8552686616308649,0.9994957129094361,0.4571627368790422,
 0.10224827268327163,0.9510271203723007,0.36293804707826993,
 0.7823643426669715,0.8022216430893612,0.9138034356260287,0.2506929691359463,
 0.6878473913582255,0.0066485246456899905,0.5094117679154133,
 0.09857676452818243,0.44322306711996995,0.5641242563099826,
 0.9622782349995559,0.5372091701654688,0.8605132528020459,0.7249761613810783,
 0.6383162904409538,0.48121235469077384,0.8798854421608671,
 0.01788406048636193,0.11878489706531592,0.016267723885653673,
 0.9758925806230027,0.5522363433420444,0.17616151315794848,0.6054536138903286,
 0.5017109800663317,0.23950472641282228,0.19655177799758028,
 0.8661347822869261,0.5688816909096527,0.7010049938517428,0.1032720033732879,
 0.7535415179257982,0.5292267180182367,0.18144427505740823,0.8709010143170645,
 0.6314752698646867,0.8193018398016101,0.6857184279280977,0.2522214928468768,
 0.6274545885972888,0.45871737698283166,0.25274849779172937,
 0.18903728992990843,0.18416822486534323,0.4285576915456518,
 0.33789365369033975,0.1304351604114652,0.30481803856036827,
 0.9449191579982366,0.2949964615515165,0.5030695105995128,0.05608142130745575,
 0.8771444086236684,0.07373212590798928,0.5667088456281776,
 0.08670947360996939,0.3272438905790669,0.5805637049895044,
 0.21324085463353482,0.3835402769318258,0.029162965705863075,
 0.4016127669506395,0.582154498115532,0.38945895950553855,0.9751668451034764,
 0.4798883828703221,0.892214139335388,0.05678190288359608,0.975388381206928,
 0.7256184276048264,0.24452547746821607,0.6140533183544856,
 0.33899825597026256,0.9083022787872345,0.7528663604390003,0.5711306702870617,
 0.38758585870337003,0.27018433974649614,0.2635188254671469,
 0.8306521971367934,0.27402943283267367,0.0977202075711936,0.8500127470786487,
 0.12955641038229004,0.3447741917403405,0.20775349533570642,
 0.15328137601262515,0.905950418719933,0.6530759992948016,0.9968868127344162,
 0.8737816296094929,0.6090574265141327,0.9137602307301492,0.7435783083030951,
 0.16723105943040606,0.835736956044412,0.40508808879942526,0.3073228377767736,
 0.019060578268662143,0.9467341414390769,0.9380057725248855,
 0.41363588624420355,0.03085325324068766,0.30942683679739336,
 0.8783034776540132,0.5117597647638957,0.5968677853664781,0.04652806197913295,
 0.9465203282204704,0.6099608026211636,0.4173413951146472,0.10910533857286941,
 0.540577145493413,0.5456505599179416,0.19173262359981869,0.24946800989019047,
 0.8947989694005156,0.0997480255366785,0.4290694293878088,0.6001221939429452,
 0.7974625044774019,0.7609911890373191,0.8722330667200348,0.3161433788486425,
 0.9432308472101066,0.8764107424271681,0.14946966018960717,
 0.010983990833987267,0.2508252272596455,0.5618228814997168,
 0.8636447695626223,0.5763236983665879,0.4810213791915827,0.6589471605556099,
 0.2433546142515678,0.7772022502547354,0.9929309143899151,0.9771151764964265,
 0.7075979491581869,0.6456122057310699,0.9253007679863191,0.7216927271361206,
 0.7796750500535367,0.22419717059704114,0.06187730662303015,0.411016520151267,
 0.7780147869566655,0.18524955085682304,0.8158833828961807,0.1366287783821447,
 0.4670545419304777,0.8953020973414075,0.38642476117221525,0.9807564490829848,
 0.5553075517815688,0.7441651838534086,0.7130450820995069,0.8791261560735725,
 0.9560146477805145,0.9853570440789721,0.383580631881823,0.08168300455843269,
 0.5756065793248448,0.9626524420972381,0.7463895192990851,0.40888952805641554,
 0.9175641578619035,0.8170495197053481,0.2823017718273272,0.8806119007379595,
 0.42774502644391765,0.03155114577113671,0.3079147198089729,
 0.5985302503038257,0.6589047035677964,0.14623024403661977,
 0.05505081387075561,0.08820961145990269,0.2941834073667712,
 0.8344019168009595,0.9659190403063052,0.9901043229367684,0.576473758089925,
 0.3888621363877719,0.41896638040134004,0.42531551145523605,
 0.9447651755846822,0.17779697395994143,0.6700280436442014,
 0.31711818351604315,0.7721884369449276,0.21830579674591,0.6369748725716189,
 0.7840940327272504,0.9162571283543218,0.7214143786501676,0.11571716948391009,
 0.18663947195088382,0.2584905673423481,0.4206268281356348,
 0.42465610807044607,0.7359193978181148,0.6427584113727491,0.5235979058163593,
 0.06625182589423706,0.9805170065462384,0.07501248828962293,
 0.8444724354265656,0.9404608052436285,0.877401696186485,0.8892193533880086,
 0.17986151424257957,0.1586950796727571,0.993010023277801,0.46485367156488,
 0.7110447645008987,0.6068638529874,0.42893329027382654,0.16579670756958975,
 0.781515828436401,0.06996001763082549,0.5393084289483959,0.743650034672867,
 0.16256569362989615,0.7084406173466593,0.6231119954981074,0.7076630037281926,
 0.5467205593785232,0.4488177232896421,0.611466367799231,0.19813913388659876,
 0.7534258670497374,0.5481878943932378,0.7307063752205752,0.8559255841876117,
 0.5468058047729778,0.4842630130021992,0.8236049755380095,0.2397990837086421,
 0.7585414036843973,0.2377163961854113,0.08551143177469533,0.2178678470018125,
 0.6448090245089582,0.5158062069667323,0.010499756468378862,0.992873870326094,
 0.5155692851932987,0.44819293976790464,0.26378527559292997,
 0.4137610705916963,0.43448860595317995,0.8983914122136661,
 0.35280952517979414,0.8393088210939385,0.9912738152715095,0.1482698881118334,
 0.40810473005407233,0.898304389477895,0.47636924561764293,0.6587865823727673,
 0.8603913508956,0.6677582008635371,0.24261316501442365,0.2910076902487463,
 0.22589722415258828,0.3080964896285453,0.10714343494654033,
 0.21827919165851262,0.4980940342897242,0.31148956135128014,
 0.31262204867936916,0.8433586660238326,0.5934898714881289,
 0.16170779591996187,0.4724413253113564,0.33427691738761656,
 0.6570336945476872,0.7177025576534785,0.43687002237516626,0.411498791685333,
 0.10902098811681238,0.6488547761594781,0.4930065272414179,0.6333642581727745,
 0.13480067614842872,0.5746424252483845,0.40811825131366253,
 0.6808897870560151,0.21588089298880075,0.03985482075277247,
 0.5846790818270438,0.5219226261858237,0.18805955279640985,
 0.007515185525288071,0.46538432362505877,0.3836267403367808,
 0.009191157218012114,0.19810753974706108,0.6354037407482438,
 0.058719038639588716,0.15447408048580746,0.29353202876952866,
 0.6283286604618876,0.7181110611207039,0.7035370237990218,0.5039353185508357,
 0.6157063065004559,0.7010364321749478,0.6348887734477416,0.12109696873068088,
 0.3445953312085632,0.4807770744019403,0.7440233976385255,0.6746689532048248,
 0.8875692016100104,0.5773432852058255,0.32009074933350345,
 0.47690930639684825,0.9256090946917405,0.8007554417297825,
 0.21208083575184722,0.9053916855194339,0.5848058970095537,
 0.09013985119064305,0.3820067579988392,0.3766080777198998,0.9789898949920516,
 0.17616146701219537,0.7183085457609306,0.22432573495390284,
 0.5154177563527194,0.7742258065436095,0.0648099643150335,0.06569998559049373,
 0.7306762810894373,0.16693983834749038,0.39688466409478296,
 0.7848120148049459,0.3001957029837332,0.35177349712465555,0.9877578229365457,
 0.18272268151334836,0.2546797117874894,0.15959320191927728,
 0.5383134991108354,0.5513630234147706,0.26182935670377816,0.5070158930704767,
 0.4893950525391866,0.9334908753372999,0.47825399687081627,
 0.21225223536662718,0.5736518440926596,0.04744380824649086,
 0.44378429689336985,0.2156614750830812,0.9361169154997081,
 0.04151802293869089,0.9895996462332001,0.7024850664593331,0.2510563832596706,
 0.5619931160661763,0.20409455410374955,0.5829823147881698,
 0.41499218870487686,0.16777639302721137,0.8223056858484686,
 0.7620806065419665,0.2679952419609206,0.4651144127855851,0.9881504982468017,
 0.3320895952797832,0.9641524573567283,0.015841111613083547,
 0.8693724197568634,0.49567143905376165,0.8128651425179506,0.560725201448474,
 0.31577982409306204,0.16014590562720976,0.9782360526213679,
 0.8627815244502303,0.14847036899090682,0.22453107165205766,
 0.37708892880871014,0.8915584575178721,0.42827479564788506,
 0.8059074803829275,0.37510878019994787,0.5812137565276265,0.7818777186527012,
 0.4750599706536318,0.1284759497235617,0.6578430625722115,0.3594388905522832,
 2.171113318689688e-4,0.820446212115087,0.6284990672594808,
 0.05718991004364726,0.15696160530620318,0.4199745784784904,
 0.8070286724564449,0.6833778672420588,0.22189103815139488,0.7896928588687311,
 0.6410195844196187,0.05087805138093804,0.4528915229145576,0.9307216676815353,
 0.047885379294706376,0.9001605935308298,0.48712893435362115,
 0.14926177277162944,0.3402618370552314,0.11830331791048732,
 0.8727934451680686,0.7350832787979135,0.4110601375371131,0.9183029036252515,
 0.38213131646401755,0.6867704791231242,0.6586746622647639,0.742938102771472,
 0.942901366708161,0.03885075341929867,0.859759975701319,0.5593975952046563,
 0.7249194454803449,0.16233897376667872,0.28272650079202455,
 0.40681697074880285,0.7374281651784108,0.3805743365791573,0.8216526870381116,
 0.8345913722975635,0.01223196212849098,0.24616720202804465,
 0.7501984553168959,0.3571406237117297,0.09310097317953081,0.7471578468999629,
 0.7382412549899731,0.46528952925507006,0.18898666830758937,
 0.3254269780232908,0.465181031501683,0.7137755153816849,0.2759324241502945,
 0.4363813479144313,0.4084492162199995,0.7062585866284362,0.355190837202328,
 0.19540641333314657,0.36122268537867186,0.5546115542972594,
 0.3879228422459735,0.4174760223389197,0.6183880021681066,0.6610987140889755,
 0.4923128004716031,0.15249037668560894,0.01241838233736714,0.693867157432386,
 0.3008269596464631,0.010759468192786614,0.9768097882427353,
 0.7898660942240486,0.512113027974821,0.995452366741121,0.768499668187794,
 0.7125824113601006,0.9391741733735302,0.822456949105981,0.4196598417437092,
 0.2840144232576196,0.9210336591398622,0.10096019652299115,
 0.05318767142919778,0.47656204549313785,0.7383514017164163,
 0.7299918978450277,0.55673678051539,0.042046754946413745,0.3342923071411744,
 0.6804162122683476,0.008642857305323104,0.9220534079594541,
 0.5196150662820628,0.8760201372187886,0.6098328041439409,0.09547772150005207,
 0.15785774059036817,0.39074404312390054,0.6450744030521695,
 0.49941812311393985,0.09432540553669166,0.609502797781287,
 0.018223231645719595,0.6643668964982558,0.17223079637768668,
 0.2077656879455122,0.5406742095875778,0.09806684989965575,0.7841686505759462,
 0.6884477568707409,0.3454121529996417,0.5282143935822298,0.5917041968396198,
 0.44173224770316244,0.9292251675377927,0.2043079232965883,
 0.027151059376691267,0.9870996358854743,0.1431508821786367,
 0.9638051502772117,0.8937411626077377,0.5488885525725783,0.2525405995372304,
 0.8298860076819076,0.4076027328020303,0.23005239367591823,
 0.31575783142410896,0.7415628808553395,0.8940377774054821,0.6145554455583602,
 0.4343807200368297,0.846291477182873,0.8720158585645973,0.40125671707223154,
 0.35914891603102506,0.15414411446170662,0.8874591137760255,
 0.13392783750346493,0.21730970884221568,0.015680418137009933,
 0.43349798925213834,0.8485735274708199,0.11692962488927872,
 0.8107649765844833,0.6351745154901606,0.844259630977314,0.3378094334837366,
 0.47255742003027557,0.7057639412098239,0.7381667682951516,
 0.19010675760240536,0.6852408661775021,0.8792598642457559,0.5041502782960194,
 0.42458415840922026,0.14858198139381007,0.7926278320990572,
 0.17909904590004155,0.1524019672499446,0.7546481033742163,
 0.33308122855272715,0.7751569871560071,0.7836289238681774,0.7843719765378114,
 0.1910069310043523,0.6342789039241455,0.8035031140525606,0.3688863999079637,
 0.16806749722397984,0.4531951241273522,0.8490549902636966,
 0.12872527153053637,0.44060976193408075,0.45869367259175375,
 0.6517912771502644,0.8387099794896119,0.49920032046692153,0.7424614999379173,
 0.42200271872052764,0.2365450786350758,0.49666919369891416,
 0.023109995324351318,0.5315630469307353,0.599118319568485,
 0.12051902134064219,0.005368942853471559,0.6740563055471716,
 0.2968804488281721,0.2821889505731461,0.347595254057501,0.1789256913432251,
 0.2668215700252454,0.4419653931021579,0.19016378322950422,
 0.36511867637724904,0.6948457454580349,0.18589699186259778,0.672841910963008,
 0.6475165292827887,0.8514586867322163,0.8780052145421017,0.18707337848091288,
 0.6407606714036767,0.2443851945893608,0.5977215915272519,0.13957216463146593,
 0.9183999150867821,0.09673725562572988,0.3530131061254407,0.9881561319703072,
 0.4509403474396886,0.28534035121357637,0.7434162717638763,0.7726080120679216,
 0.7504477915431214,0.2156373431794808,0.9634551484854227,0.4482030813888125,
 0.3678429759526679,0.0840213174455261,0.11343918636461647,0.6118933329917355,
 0.273191421567191,0.5990923330827944,0.44145527408429186,0.6924618699654608,
 0.6824780445138088,0.6435744867610951,0.8643260460445856,0.956563898091503,
 0.6922586314259531,0.17464240409285747,0.03331577362743798,
 0.6556644246945058,0.7027291010384756,0.5931840274319538,0.672314876960222,
 0.3036377972504989,0.4379582467281319,0.6359817992276612,0.37563978352378913,
 0.41586907737255574,0.1706743184795695,0.5423297584699261,0.962995388412875,
 0.8948456128755439,0.8215650083395717,0.4628404666626944,0.18718712569578522,
 0.7438006348917233,0.943757354283766,0.6386385688472923,0.3222248273051155,
 0.3613275091325986,0.7424969402330487,0.6831481427263227,0.5461007484123991,
 0.18930080663799054,0.3394350965991275,0.2901720680068546,0.9376679417075949,
 0.5683391703895546,0.03241316837026509,0.4033166809235693,0.9722726873007611,
 0.46071652025996135,0.18307127070622564,0.8440890029344502,
 0.5687403622069792,0.21602660429265153,0.8225909152014149,0.8299506800166915,
 0.7492467492993033,0.33752970726563536,0.5073264092698113,0.8990879635674456,
 0.2847800102165948,0.8402966819587853,0.8678722067556257,0.36847253051958206,
 0.2988629349357237,0.6384075044990198,0.860455977941483,0.7617786651308314,
 0.49991642804038905,0.6841437232266503,0.48999492511804055,
 0.6238983844644699,0.5953344860839893,0.4911718103278442,0.4122108080230549,
 0.20431081542094076,0.5549133482674425,0.14352879259367524,
 0.7659380923062408,0.41041648518203233,0.6510898392942782,0.6610138798652369,
 0.5694697722911891,0.889124172160344,0.5668454231743038,0.769958133660761,
 0.7105413846847051,0.5100195614967018,0.9608069437628012,0.28873501113792865,
 0.2248448612542543,0.12083390884455503,0.04733116531698789,
 0.7150552110992549,0.7860216746660561,0.8310411790533325,0.3543765065096236,
 0.8415743258491968,0.3109096402214624,0.7250887151066169,0.21092505125390015,
 0.4091332579573621,0.936360878327137,0.7160782025096217,0.3699116441667383,
 0.8533605399562508,0.9116743946702097,0.45688835266578287,0.9788991938434906,
 0.6213428073064435,0.3814907049527889,0.913829085586036,0.5938868847740643,
 0.7249350258819671,0.5628263612204265,0.8838562609407016,0.7019093459803947,
 0.5504349473169111,0.659352049758864,0.7155111479407226,0.3909585823239895,
 0.6741358436719662,0.5701462471561838,0.11486793652633887,0.8005543441005777,
 0.2790796672797531,0.24764876724145946,0.819122762159173,0.4237201419644644,
 0.09209333588081003,0.05655180654739922,0.36827954580092004,
 0.5333336675546231,0.047109554127839126,0.013271988817861957,
 0.8745072175004047,0.8172946159585037,0.5092824617753511,0.8710210999408965,
 0.6149817860442308,0.6357853900836979,0.1270265940393127,0.6418165543352202,
 0.8683945395499589,0.029775282226340627,0.7052068306200181,
 0.21537798901131922,0.343669055720755,0.3955519268827491,0.5941778515786903,
 0.45552746891799345,0.632066939026064,0.5757938347162387,0.5661377772648629,
 0.5495989068444614,0.7276090032486662,0.13826504383822447,0.799862446592291,
 0.38239462237191835,0.9538465256927605,0.5360889100791664,0.9281786157305921,
 0.33644792571484794,0.9845278298263996,0.028208461396700635,
 0.5393737533683612,0.7290863241302964,0.9259575871057194,0.9643990893306489,
 0.35684032750935835,0.43405375959209014,0.4744960043387003,
 0.3222665258391779,0.3237301606064569,0.9887771125093804,0.687197183838158,
 0.8825971099690499,0.7083354600184444,0.5992887291019235,0.6444623960028911,
 0.648657298843062,0.01902660741273987,0.8653068921186702,0.8035752515966582,
 0.0019227321079640092,0.6469608427147828,0.34074400183599574,
 0.4512907161405597,0.7284594154769258,0.6515824516361645,0.530617633642318,
 0.5764552085015574,0.6810480345016947,0.5758822736400079,0.6402389760544104,
 0.4534811659179149,0.2854616655070914,0.12256635323802978,
 0.36639239215811314,0.8780437533156087,0.03761274931901848,
 0.7230991009138391,0.483614952733602,0.2812009997571814,0.5897821262188324,
 0.07914381175645202,0.5492885917199133,0.8378737000904007,
 0.07094589203879509,0.32254687124623915,0.9047607108785888,
 0.3049567800927362,0.6871781175504991,0.056797754425587654,
 0.9626462501908605,0.18918046617162043,0.8261554082634021,
 0.30327410563098145,0.42487843985235274,0.2378743125431667,
 0.2922753942714442,0.21835254669619908,0.3820360554459785,
 0.05937481147722634,0.02908840761504794,0.5522698736332785,
 0.8314469752435117,0.12418817406221105,0.22957445932348564,
 0.3283986839717221,0.7727859566909946,0.5047774603085264,0.7868175140826558,
 0.41690631630993413,0.4039109772896248,0.9302963669346578,0.7123042281322769,
 0.19827718266811245,0.9341712308095633,0.8317665124863467,0.9135555140314044,
 0.6020270173210881,0.33886185985969564,0.8482177687033083,0.4123731389750319,
 0.5539219811292019,0.4275135991637693,0.6621286598674623,0.38335659621398266,
 0.7226600817706453,0.11177567994568827,0.38551189068551617,
 0.46971629212989485,0.10236683783854128,0.13552873242201913,
 0.4202896864722049,0.4195999524115551,0.0020331439446228394,
 0.06981945221780606,0.34145029198733035,0.06443392672033899,
 0.37899282988853245,0.7394734003448425,0.8135752066735782,0.5467803487700976,
 0.672777776666922,0.2743414889989909,0.7050882340992523,0.9756550806085149,
 0.41964750242959314,0.33826827774071766,0.4866130099550141,
 0.6585280065387189,0.6299902418417455,0.7658544806283254,0.1288331472317502,
 0.572868279115479,0.702200582883484,0.29298375190695614,0.19336442770041729,
 0.4676499871850094,0.08844811726613122,0.7748349024342385,0.7079064551601731,
 0.0049154378143335675,0.04451635568919299,0.5975018575187506,
 0.7502845344453865,0.38273222538898943,0.9217982513232924,0.5599422159449539,
 0.5902792755450312,0.08147463553772372,0.22509570541390223,
 0.5270171586774648,0.74877959526094,0.16155480085363683,0.7377156624106609,
 0.7051312281042909,0.4084311212938999,0.16934495019042772,0.3922557526124857,
 0.43184146368778065,0.7204787737520859,0.028867331672018937,
 0.4644302656257938,0.3643788440030815,0.16023601903320572,0.9622418928774518,
 0.24336893659246284,0.9725634178577849,0.09489833772000966,
 0.30412409000301466,0.09700235742797847,0.7403931421449093,0.840037485651797,
 0.06360193650003909,0.5311491276389823,0.7891534127413535,
 0.28525284708384313,0.9367911519804887,0.7207318650927629,0.7358136278805498,
 0.06248660703321984,0.4529249018430208,0.1605413539478966,0.5813546202510047,
 0.5978857306713556,0.9388065360285717,0.031536973141631286,
 0.4913858779214598,0.12029435870033234,0.41013349218437,0.6493740217697481,
 0.5750061728903385,0.9443434958682667,0.536948852743859,0.6366379747200352,
 0.8809038367079628,0.9962743985657789,0.9610618936520711,0.998546719308828,
 0.5254996569930562,0.11128375911226074,0.8666200635204488,0.8556458183738379,
 0.858861076635413,0.0503888659872197,0.9492616955371052,0.9109162678514365,
 0.17149064079700072,0.4082474503895259,0.1195585449986456,0.8761718401200653,
 0.9432095254165481,0.4095873181673261,0.7064141312670579,0.18781144580813247,
 0.22960850490962426,0.44917573105092246,0.3701455825230423,
 0.014058986179276367,0.26454363917541146,0.09230050038658666,
 0.703461840285302,0.7255868664334728,0.260396172130812,0.23709265396399115,
 0.021557459288368275,0.9411275587446191,0.9007677305624324,
 0.4411575905982924,0.6467374116011204,0.638414700382211,0.7025596077145418,
 0.13783931663155657,0.5243514919229685,0.06616203810985788,
 0.6983338440817599,0.4316232612658276,0.40346829239239934,0.6940158789356093,
 0.09151053970518241,0.27040836072466035,0.19143167128804428,
 0.2529313668042379,0.5594481373907072,0.7201964588204225,0.602983586326302,
 0.8659660827509169,0.42012233869167437,0.4100533681338483,
 0.10794438692523989,0.5153233948067798,0.7555666562682735,
 0.19313337578035938,0.3039051332951651,0.3707264644786984,
 0.23899252224798728,0.5246575866148407,0.014414824580668006,
 0.9676649592528953,0.6935495535432686,0.06754607752692088,0.7292416024365111,
 0.18340821527215068,0.8354161960293006,0.9827707932638713,
 0.15476672690009197,0.20958444645211394,0.479101930348079,0.4977381050403401,
 0.0800351551093813,0.671179075047946,0.17800736555667152,
 0.005602191538667589,0.8413473143768605,0.8089532241958446,
 0.8274628750900902,0.3536546939628987,0.5271666826229232,0.8067551928965546,
 0.6656360845905881,0.21117843286768467,0.14327438730677855,
 0.11827892708754628,0.7490152044288079,0.28264610652448097,
 0.5842387326207233,0.3107845911527696,0.2445118070225053,0.7066335790430668,
 0.6326432238120342,0.916996050771082,0.7721474941476842,0.5225177889773194,
 0.9873689291022567,0.9098203108060771,0.8966197413152658,0.28439556904405705,
 0.5389654958439928,0.29101132874826985,0.9446215253276787,0.3150217664556786,
 0.4955782921401858,0.3862163035128706,0.696458027803104,0.7080096803795493,
 0.23406246954208076,0.9818493487165294,0.1990424888073602,
 0.14759164646033618,0.1728871362522877,0.3369000282052439,
 0.17656173028753752,0.026750091679906163,0.340155282528703,
 0.34862383889474013,0.25003687556827914,0.2690251633338764,
 0.9510985031132841,0.5494712970289551,0.2729114165153558,0.09589746258909715,
 0.5841406705063104,0.9274996317486409,0.7554370808335671,0.0499049655647652,
 0.11747034054461214,0.7389923771281941,0.2066161892802475,0.8249532890020963,
 0.3944738834923387,0.30947713385348774,0.5661282288167366,
 0.45557636108696997,0.5823175829445335,0.5909313772512085,0.6979079461335697,
 0.8163443689452061,0.31587458541839175,0.730879861319598,0.9510787843656806,
 0.9791342894857614,0.4945453392543565,0.21217411746051562,
 0.39053838997548274,0.21657835602285436,0.20297230269313082,
 0.027198904565375104,0.7193213266725278,0.4495990684725324,
 0.8860755012939855,0.18356459403293346,0.9433328394037372,0.7650253782199337,
 0.7854878688939801,0.061771599702962976,0.9288506076339362,
 0.08036547411810024,0.07265053238337293,0.9838191311238151,
 0.2135934360928896,0.8527159176037091,0.6292163015662432,0.7964924636268156,
 0.20320865088687767,0.7366923583392545,0.50169423921479,0.9659494194483863,
 0.28125571050466536,0.07213419160119161,0.5404141020297604,
 0.03189389865389125,0.09582310707466912,0.15699453722357326,
 0.18311081208371194,0.8240585751827185,0.19437311568413018,
 0.06505438850681844,0.5967586361408957,0.01407737987629687,
 0.5232909905884131,0.10383892273344408,0.27336375377909927,
 0.5036510934496907,0.9910134549910612,0.34910325657745955,0.570821442189943,
 0.10820048664116066,0.008136837751060333,0.30260537169315815,
 0.4146439271453307,0.49168571371652925,0.006688530872863363,
 0.5914555294545094,0.6289091978438505,0.388327350577204,0.5506277473377844,
 0.35629534968640564,0.028423374787324818,0.6999426509199251,
 0.6704685123720994,0.03684760927774422,0.1888507565374078,
 0.09585878059496256,0.5828110167978504,0.5252859203160758,0.7174324248993182,
 0.2854152714007282,0.5352725803330967,0.1435020502909906,0.3668199941851593,
 0.11691131899174834,0.7030133227356283,0.8021500406898479,0.8823282268762895,
 0.817235483734544,0.9518061447778315,0.7993556077813344,0.8868437204778414,
 0.5405125170707251,0.0764986587183838,0.9384289074558771,0.5323414373461244,
 0.6052005784924571,0.9290390507395686,0.9166864648861314,0.05441702285629335,
 0.3291379476950561,0.41285174317724416,0.7423014531341854,0.6224288049011487,
 0.20369506252372516,0.7790677375615419,0.9081441525179401,0.3988676620647469,
 0.5804709173683075,0.5688403339298111,0.16292520534232335,
 0.46350494115285124,0.3739576727462951,0.5526365589071359,0.7892335209015693,
 0.45638699024398155,0.638234616514219,0.927234332755788,0.7818423386335838,
 0.7972502580249488,0.789819021704525,0.918591377636907,0.09960966156103446,
 0.778361583310749,0.595639236253495,0.2146060163156267,0.0687386985411167,
 0.4462951038405013,0.5806416901418994,0.3751167711219512,0.4706019514000158,
 0.7964916959621258,0.9360307641004899,0.9249692871184979,0.5754798104579503,
 0.9394001494640641,0.37679264679479263,0.4307597863793359,0.5442441893902636,
 0.1664375528825428,0.6393135751961522,0.9517525765730704,0.986721899375357,
 0.4201973171609642,0.45026193608568355,0.1928435914662765,0.553841693809892,
 0.7589733894761681,0.5710794304652231,0.513988162087953,0.5574858587926242,
 0.34833100175196297,0.6304724954228407,0.28954858284240037,0.712611620597543,
 0.2348389414941352,0.32993529391651666,0.3499409055812531,0.8639751260069283,
 0.945185379415947,0.2095503385124866,0.7675764022074758,0.9081986638397028,
 0.9265773151366424,0.1540328008793549,0.8158123171413734,0.928805556241687,
 0.8580557416492951,0.9937361322143202,0.4879559752167718,0.36573828055501234,
 0.9444458953678239,0.3360982478682235,0.3944797607320023,0.39935967175421494,
 0.5612640217133934,0.7656710463785855,0.5975169712036789,0.6458589859486028,
 0.2487526608910785,0.28947168727421546,0.5452115032875444,0.9509633451430773,
 0.03740647414343479,0.3485817571956117,0.09483902112308051,
 0.2666940230732161,0.9804700026939569,0.7591866117575701,0.7411580276218446,
 0.8931493362185592,0.9759222070775818,0.966828324784344,0.3812989140100542,
 0.6070511695407765,0.7528764266970329,0.19017522918839347,
 0.27195293042982127,0.676994189440919,0.7215731417380025,0.8308972573258638,
 0.31962040250769963,0.6974669333345771,0.6509216524329191,0.7484014274059784,
 0.4545121765878426,0.7719025465998157,0.4171975118397895,0.11207865551045304,
 0.8057620522504353,0.39163286620698357,0.5470173734300173,0.7743534908288792,
 0.5625732741906413,0.29624546615366465,0.40138296076228275,
 0.9025544839276831,0.8036682457067851,0.8736032027929314,0.03298425024552443,
 0.058109913268362234,0.366381263713654,0.6623522274523305,0.2584943106057932,
 0.37914002606488695,0.30633521286736054,0.7659180899578786,
 0.22824151650554358,0.20793543397567926,0.4065334215855092,
 0.5369561348537434,0.04259312776170665,0.3174190822598071,0.9492243431025402,
 0.7653458603056923,0.5027821782660986,0.9732397074273921,0.6596487499935757,
 0.8329068527451725,0.5699313086657305,0.1327141687335638,0.4782172079694811,
 0.33432680939628723,0.5582496571552308,0.11845222537754352,
 0.8204170677820587,0.641860411915876,0.122605661515515,0.34170426396401876,
 0.6751378937079366,0.016097005009331422,0.3538461827184908,
 0.04323412670329019,0.13615205830786103,0.27921460123559716,
 0.6352660424827558,0.8414722884996005,0.9094918029402899,0.6444491715601115,
 0.4447302853485322,0.867255720997847,0.7737142105624526,0.9302962530845742,
 0.9503686025466546,0.06918295642793271,0.910758863909944,0.1510283039686684,
 0.09890351877284798,0.582799574449059,0.2152465185456458,0.48554632827116273,
 0.524962966635055,0.6003208877724117,0.3107982776345065,0.020689973285232588,
 0.8935588213438672,0.5217143265042062,0.9690788721906136,0.6906531378714043,
 0.04855788405446704,0.8166843385382552,0.313086190823214,0.4895012348449337,
 0.7575104889290624,0.8531554190054471,0.5675832712674257,0.29069730300407026,
 0.0780170118821446,0.12784790210547903,0.1602398571402205,0.8519006202087795,
 0.007515329952009209,0.009667015955179825,0.8921377406183685,
 0.9059814764769148,0.8231401109069484,0.7766101213090493,0.3578034863679209,
 0.43695255866792226,0.5082090484070744,0.98277406514233,0.7487616470061709,
 0.8829691243836224,0.497688125807207,0.43049828146585833,0.3985920519126013,
 0.0253775335122034,0.7233342506636808,0.7576150137506775,0.4361164606678105,
 0.42708758849191975,0.031780220700354356,0.4305910722800911,
 0.9155874845502114,0.013414124415341955,0.44752460574132535,
 0.9732523769310026,0.9753034710307132,0.7563560004176804,0.7079948312547035,
 0.9896210049175068,0.300888655598742,0.12457505077820352,0.31797961430637345,
 0.9302012817915588,0.9455841503276452,0.023416732222296788,
 0.9308552723191676,0.7669926082661437,0.8759878208354812,0.8734718361642501,
 0.6961629956404639,0.07239363697058476,0.3387610603252508,
 0.07943205595982328,0.7012813682133491,0.007155951233627222,
 0.11018327570061182,0.5593838574371709,0.5612590579396123,0.2783992313949857,
 0.2012330651453922,0.21602986221298448,0.14540797910328696,
 0.10260716176934581,0.7600459757789559,0.30732883394483157,
 0.7223524742881792,0.5607256487983479,0.8227821810999641,0.30592604776036014,
 0.2237710131356565,0.6089824172626296,0.5784362435889571,0.41685640284089365,
 0.02984203441440958,0.05735253495141068,0.16727509617541747,
 0.46677482125782965,0.86494809503996,0.4417605957776203,0.21059726639912768,
 0.2881893266936011,0.5277598465283079,0.22458008766267168,
 0.17622707681833805,0.7624062305355797,0.5081069131110285,
 0.47332044165666454,0.4964694977974303,0.69762633468713,0.32139028170252326,
 0.8755364109946947,0.8708516076250266,0.7394858215380091,0.770523444270099,
 0.42347768312088885,0.40955245516878236,0.6034297734970331,
 0.03185108820242455,0.29832921378985344,0.7883421768633815,
 0.28672479668085915,0.5243820539814232,0.5779941381892679,0.829151823636126,
 0.7418576628939587,0.4873095362868496,0.9064229870754589,0.7211198930012862,
 0.47527421157546645,0.004997533450773695,0.013304497789884184,
 0.6989533073411798,0.20821120237813573,0.28841316444619947,
 0.7477691813579064,0.8858133387903853,0.7314633217032216,0.5285291912289326,
 0.12246059361528405,0.55202251283218,0.9547099720224671,0.27094525495532906,
 0.2608269287055349,0.3679392132868161,0.6546449582872522,0.386457061034855,
 0.7695851803395349,0.7296371240391303,0.1468307175628426,0.8720528933574258,
 0.3175377112406059,0.41549361469673496,0.5661578955619238,0.4451003623201659,
 0.19538327523942967,0.7234366197580011,0.8894942172081031,0.543380547437994,
 0.989562611462776,0.015108327552559753,0.3114727783718283,0.651527137381061,
 0.2074983123297982,0.1220465955403861,0.9503668066337907,0.7481766725535982,
 0.8582530003944472,0.016677933176604154,0.39285843358003103,
 0.6595073459168753,0.9365487428491504,0.705937659982396,0.02072233574540916,
 0.4118096597847135,0.8219570042480119,0.9035603866591454,0.6703792946764082,
 0.07256337648650613,0.3569710103034782,0.4388582893326334,0.3497202692998862,
 0.030052926945784986,0.4896496995275178,0.7195437020671841,
 0.21628567469788607,0.5944709216710508,0.5043386392616946,0.7273678210078077,
 0.07600204923422393,0.5605070016722801,0.5171313008198347,0.7477174910636213,
 0.5226646624084131,0.7470733685719783,0.525373854546971,0.04549846697332094,
 0.26846381132541397,0.21848759269226214,0.5204656826299533,
 0.1975572165554209,0.0014563109413525677,0.6971541850848944,
 0.6600091893202789,0.3954513107185671,0.4610345500348183,0.29387003096663356,
 0.002729510163226978,0.4648175296906395,0.9144889064736526,
 0.7407709799546239,0.1791324127528232,0.8365713348618048,0.22869583448158692,
 0.8854663739553859,0.5781542692392158,0.08646791976164192,0.8711641331615917,
 0.30096905041462485,0.5867494167099366,0.5710308527636732,0.8865817177229716,
 0.23433338966664918,0.022912010551241146,0.5814661440304317,
 0.0046888294757354565,0.5813932520119977,0.07446220262054892,
 0.46151744278312,0.3717482598874394,0.8972750431667106,0.7150927888565896,
 0.21724895677589606,0.5977760367751911,0.40637779181558975,
 0.3975467108516362,0.428752769260513,0.39203473095398755,0.3459987709074388,
 0.7549633689036179,0.06304457171008848,0.5032380359156438,0.7467820208302219,
 0.029274493463378626,0.06838562014802185,0.5310128805197395,
 0.8705383012195164,0.17413709363040675,0.2368906539090011,0.254025318191928,
 0.2890989733320608,0.1351114917611287,0.8926908522071427,0.4938566019917281,
 0.743144495113602,0.9997813172664721,0.05548151018759828,0.24100047636570754,
 0.6222268004381204,0.03431900947973299,0.15154313555779464,0.730941884931858,
 0.7494646230333128,0.023734672779609056,0.8529515618518163,
 0.28177001162970594,0.5291259392461984,0.18048459245734216,
 0.8431459743424139,0.07752843837918166,0.02074239716972892,
 0.5453820202950483,0.1520654929270171,0.6169581534412842,0.22827728046566453,
 0.9282861827493009,0.6958541209209229,0.7275006010041597,0.7458779642818169,
 0.25668912585887227,0.5934387242568377,0.36606837982477647,
 0.10388593902771237,0.46128440947896043,0.2831334346840194,
 0.06155383126606029,0.9263860874860042,0.237340768446616,0.8960140155249497,
 0.7434556925481197,0.07695989621743504,0.07591854162263789,
 0.9364835557560828,0.23050636683560455,0.8184060716531412,0.9851345332526297,
 0.30741124532751485,0.2701678743237991,0.24626835800130897,
 0.7884052222153575,0.4291383800208338,0.024842319625632436,
 0.8034696067840764,0.6949082793201693,0.40877704922899594,0.5628416821808034,
 0.5885282731373739,0.3810938073668021,0.13089485601619977,0.5092871533516539,
 0.06481589899185458,0.3841460364678435,0.15538267340333956,
 0.5593356453295099,0.7191977764429114,0.8426918390591381,0.4397193663922043,
 0.6579494261656386,0.5534796866873521,0.4124915030566252,0.8256801789119503,
 0.07842807530077445,0.33783217304382207,0.805524058108449,0.0416413877490287,
 0.006582252104199382,0.7977444150151531,0.18136509361450892,
 0.1742252099752425,0.5120564249259433,0.20000189689606807,
 0.39457059318508847,0.9423816980216145,0.5049173102137439,0.7130345704789404,
 0.036315132414000306,0.1292765962395368,0.9278371595239024,
 0.13941211281730959,0.43957197166095785,0.832611679409867,0.7359367492671619,
 0.40237020002867385,0.581940465229062,0.32547891702817777,0.5918378519587283,
 0.813820484230066,0.7904480970391448,0.6679157486100182,0.11222680481250924,
 0.22100917338992,0.4498700116944221,0.44559216005886226,0.6866829976354063,
 0.909036728779696,0.2187725902485984,0.5446600904507508,0.5116206685921113,
 0.64522249717269,0.5559050749462751,0.1101357599189059,0.8953547092678326,
 0.6333060625618581,0.23591637937783416,0.4415885918415122,
 0.44244406635727973,0.5244259615171355,0.32091445403792895,
 0.26535441384376246,0.8926516632431359,0.7500037309337938,0.6601544782465226,
 0.16073055969901373,0.4626013200569308,0.7596477116505034,0.4443423148505542,
 0.5622334434148499,0.5154474081656067,0.6738449650761685,0.5881093717627602,
 0.008592498742019616,0.10100085154863936,0.6059918156301123,
 0.3062126911465122,0.2742626503791996,0.395728268268829,0.15620658686124345,
 0.639556454915407,0.6494452089659967,0.9098594688093056,0.34808801574575965,
 0.6566096324280245,0.2604963752241637,0.45263715426514295,
 0.07175471515850451,0.439611752523963,0.6282201621542421,0.9420457155746131,
 0.25407338573048643,0.3152349714432384,0.37847233883176157,
 0.9087774854287937,0.4262360096268276,0.5040131473677275,0.32800139671023887,
 0.26574089567684245,0.4515280471236175,0.37506097223431556,
 0.9789879000780338,0.5213851662664108,0.6075400269233967,0.9526357199029494,
 0.046387931496456036,0.08833434211605506,0.06600201385511673,
 0.5291320830062407,0.5201674478575715,0.9497560372823399,0.6329929097349392,
 0.9211887497500134,0.8628037811443248,0.2786295840029446,0.29455191267842007,
 0.33681104774849546,0.606786333323646,0.9267769736407053,0.15289511056715666,
 0.4968734867495085,0.5000698734854775,0.7469315828864365,0.3391707026358123,
 0.8718280073594235,0.10839689272881947,0.9540998409880448,0.6590645169806981,
 0.03210832579675138,0.07268906759942517,0.8924363641457731,
 0.7988972768990174,0.9767525907190779,0.6614095183575636,0.04502566121455931,
 0.21975137154155988,0.4402406987499917,0.3372812990025664,
 0.056057909417900675,0.6427080415284585,0.08298936416608094,
 0.9726572437724557,0.06556547685379588,0.7939204361120491,0.6356565461158356,
 0.30070493805782417,0.9301491274645401,0.5478542528616985,0.8415827734005991,
 0.7017961900931159,0.9213833729765111,0.28424164366562876,0.5170185533722281,
 0.7851834006797873,0.2804466733586154,0.9769578802870995,0.8150458407612265,
 0.16156457092154886,0.5533561703847855,0.4034073351049672,0.6577229486434233,
 0.7126312366808746,0.2601881369612812,0.7494680378676667,0.8813439264710646,
 0.20593987474858066,0.03933784001559526,0.8517141679371907,
 0.20166140993113757,0.6721590131329351,0.981730114371231,0.947026854946575,
 0.8541995561529951,0.7582082006940737,0.06362721436359686,
 0.34442379269629686,0.5104280078449848,0.08449289208298194,
 0.4079546048620626,0.1926404984872948,0.3987124108904663,0.1968013507859041,
 0.29290191121382303,0.7597304042835408,0.9654794183762088,0.8841295018227471,
 0.8201469168510626,0.6971920739741554,0.3289365138978254,0.20082882757732434,
 0.2846686751643964,0.4140778068704195,0.8506184369530946,0.17819627282544193,
 0.6250510373345295,0.954244091511086,0.06075734915209907,0.11100574883776781,
 0.9325359018179398,0.08169095304902707,0.6877060527664438,
 0.23015895598903935,0.9799882315667658,0.16297293996364692,
 0.5879580876819435,0.051706687342796154,0.2824060412092976,
 0.5084288270555164,0.24289740724871267,0.5483356982830044,0.9411639033425251,
 0.7431202635837902,0.2877528556117437,0.5242232531315735,
 0.030766137646854652,0.879555765596529,0.9132127298939281,
 0.34330344109118227,0.27910817094613716,0.7249369317198844,
 0.7921905764161461,0.80139863868799,0.24453730598530588,0.8844773987986313,
 0.31194672474189145,0.03890885694512347,0.18446481946308235,
 0.5156965692289044,0.8928951410440007,0.25249411547170153,0.7845349622220411,
 0.09817158890450184,0.055474708236918246,0.6678202882105301,
 0.23990428750203951,0.6905557128207978,0.5742666133554414,0.4136085152519253,
 0.010216563437161486,0.9010981236163014,0.4047509567124199,
 0.06530619940278026,0.7425126581048959,0.582614427890992,0.9553650109899134,
 0.09227668459134686,0.6277220081003778,0.46952092552040103,
 0.5277316657901033,0.9386750787631293,0.4252676812249119,0.1393155170526844,
 0.34054334929649666,0.07760773474140437,0.8709275642412758,
 0.26653476871476245,0.7418566531147857,0.6615163141022622,0.2802621770944197,
 0.6043863452062829,0.3886950997865538,0.07146399587229435,0.7135980405900075,
 0.10555531931927753,0.7569952565235097,0.039411161052421706,
 0.22208671293491955,0.1263051385342423,0.8932893094691348,
 0.48367449203131463,0.1969270115540619,0.7371409430565742,0.8577328539698573,
 0.0621648569137897,0.976711459903986,0.76918343286034,0.5172207028459679,
 0.6521986294997333,0.662341868391821,0.026946499119642087,
 0.13698232200735894,0.584744025536652,0.13469749585212254,0.5385669125721037,
 0.1377056282820711,0.9117598914698566,0.1296535894969485,0.4659670611789585,
 0.07831527653833825,0.3448001353274134,0.34667047377383486,
 0.06458791509055117,0.18935188886391008,0.025625360419683996,
 0.7357628604508235,0.8942310057837994,0.426703036991246,0.5179479522438925,
 0.17350245250122331,0.2275034338774229,0.20757136226647388,0.220438308574881,
 0.9878139394927224,0.8498689010476539,0.36398071892266837,0.4354775035193791,
 0.3552597585499523,0.08290828412321638,0.11735951049738946,
 0.9294125283467917,0.8858775013854666,0.34073326836460505,
 0.25324028887128947,0.5893152933914706,0.559368457233986,0.7046233028423955,
 0.5552160270744058,0.4147183207013204,0.7617650270155645,0.13235107209999353,
 0.6302164820977155,0.29493877999881124,0.14191770764243006,
 0.25790813487760045,0.506172277854547,0.09247082382840677,
 0.15841965197376462,0.1736865627344637,0.8235567086029489,0.8356037026966752,
 0.4462941890467096,0.828689877168031,0.458294537211947,0.2374585621815669,
 0.6797363956777954,0.8108014264292749,0.7683692154645856,0.34712025039086813,
 0.4056241541028467,0.7034213340512016,0.8051873373513984,0.19085421085991616,
 0.9600645612325748,0.3847390167930075,0.2329977408927909,0.6254881452035045,
 0.8937089179928889,0.6468522404741245,0.6134275161141021,0.40026282336463914,
 0.2959621242780669,0.2882762275633025,0.9245090997096,0.4343301791664038,
 0.5260530813036162,0.10279241959593666,0.36211850048652483,
 0.01798568553892288,0.8978986399968438,0.33270063956967266,
 0.4694995092209402,0.7020865533705234,0.13866968483389852,0.7360053998917048,
 0.6445499912580567,0.7114591063656042,0.22444448297199682,
 0.16225614837453806,0.5583661450995134,0.3751992153598349,0.5761861020795424,
 0.24099998020386804,0.3510908198016408,0.49662019095602894,
 0.7225642044325088,0.5755175523184773,0.7170359716739922,0.861582460476509,
 0.5014766553840314,0.11701448964484318,0.6600230490571187,0.6400086544215622,
 0.11237313063802579,0.49763144419328864,0.12724223790414002,
 0.3003971849583149,0.16051051729045152,0.2772982195194005,0.48365203608119,
 0.3109136898063012,0.9277979767878779,0.6261094984253972,0.21682501352564598,
 0.6564604067594206,0.2971159576800134,0.3438582715946833,0.7160380941311983,
 0.5939479115696239,0.9291009901330737,0.24128256718309993,
 0.26144433712303783,0.1773974436821334,0.42386118155707597,
 0.4129331276029151,0.6180855031445286,0.7052111619146916,0.6310672769643013,
 0.4920752167044101,0.7998519518444327,0.0876959234167054,0.3131959466764639,
 0.276937362345295,0.5183483342147507,0.7233454695415075,0.573871214282641,
 0.4104318673647406,0.9868828190180667,0.04758395101770074,
 0.18760677728168507,0.5664853176908098,0.02595757325576109,
 0.4992164151914994,0.7080966446223925,0.7534494412220432,0.4566857573826869,
 0.6508890372272248,0.2417926950473871,0.41104272868117353,
 0.23296333989347617,0.315624790316014,0.6584454436688317,0.161971016533335,
 0.48340702126241963,0.9119438095940818,0.2410350649531665,0.9113614839705069,
 0.7531010335149898,0.8017752831496099,0.3104972547370145,0.6739896672821345,
 0.5739914344120932,0.5348090113893023,0.8667053090968349,0.08500131404935463,
 0.22767507941527632,0.5103732842167957,0.8506560164720763,0.544589987515505,
 0.9482303823727234,0.4497028920631184,0.7180698437520031,0.7476959929248332,
 0.3068832095259685,0.8456520467650235,0.08118798352036516,0.8648272404669931,
 0.17169831964021554,0.6773356817921568,0.5985047510221594,0.6973840730545043,
 0.0314013261990157,0.6991465500574154,0.21895609245932823,0.5852796367947734,
 0.7723450057951531,0.40835183015352816,0.910592268793754,0.47652361151547173,
 0.7762680983303981,0.2806115488970281,0.87837626523328,0.1650914864070513,
 0.8335731109453455,0.9528920471758263,0.46528877181999406,0.5787917051222691,
 0.789473678174053,0.11978798498054777,0.326334677809733,0.10039648210517194,
 0.4717920031817513,0.949371110515027,0.4960638009923961,0.8927500095884507,
 0.28869143770508676,0.5917326028192901,0.08909267152643174,
 0.07410629710561512,0.05139433406080074,0.2236312212434598,
 0.8271992900749516,0.24597199270046333,0.6570394768365178,0.675743461205395,
 0.29862208683521385,0.060167459303719584,0.049446999703200656,
 0.48782358210474064,0.7725509932741227,0.8619211166498906,0.5093728515082864,
 0.2514668908377685,0.15757175054817452,0.8997659784691172,0.629204940400751,
 0.23994837423216242,0.40878348738551185,0.5655578397788177,
 0.46923137931496817,0.24899224878784343,0.23894520317386714,
 0.21118699159483845,0.18068217896029115,0.8943184455494283,0.170033000153319,
 0.11216358137176186,0.7873316692586049,0.7698657457247381,0.2342498319458668,
 0.8192641376430453,0.6876661854414584,0.9702000031582287,0.46575134939137186,
 0.11091309043167663,0.20978586487176287,0.3143518213314571,
 0.9028484121909163,0.2991159791515712,0.66338176968177,0.42917818689325504,
 0.2930952272912206,0.003068585612440722,0.715767069622178,0.5741731364415148,
 0.6331563059316401,0.5133397071928347,0.12206723303996614,
 0.45913723276397633,0.18613486334013885,0.5209341783979782,
 0.6724327798246437,0.7678573138562497,0.26410302709537914,0.6219046973566654,
 0.4097740984048195,0.8748665449089202,0.7090400492136171,0.6212786716350515,
 0.4870894443664262,0.6260643757808618,0.2563774142777371,0.23817090928877271,
 0.7788961899629383,0.9095929764875256,0.14591139629735295,
 0.27321893243439743,0.2966286901708943,0.974916644056883,0.9061737232725475,
 0.4715812836750696,0.4650995672154672,0.8381717616282152,0.6750961027843667,
 0.748710368628326,0.2094733563951482,0.8669158947144626,0.7903407334614948,
 0.044160700787785,0.91010466701323,0.0024370251459112646,0.03995546208700629,
 0.8394962467822403,0.10493278526173289,0.8027165055610013,0.4372012355879953,
 0.9334632127290542,0.09676961851770371,0.6189674804622066,0.7102903518391378,
 0.5942884155971965,0.8392765006289635,0.7740712480221139,0.09059507046592741,
 0.041852216348773896,0.6884235056187884,0.08358627285913134,
 0.9567448434122469,0.26661779031971655,0.07266970948289297,
 0.5341432797730222,0.717197158856477,0.34973243775132157,0.07689928096749166,
 0.04005978201431648,0.4338748441039545,0.43854008547196444,
 0.8078887648977855,0.6113836307991816,0.20120883693572922,
 0.20074507728952984,0.7359524957044689,0.0920588796585653,
 0.10607137935672772,0.43765592316906643,0.7598776581740632,
 0.6296921365688061,0.08150076589294841,0.7749585397816575,
 0.06775267973566923,0.6798268047114716,0.9233201254835841,0.9900311063344325,
 0.8441441886963503,0.6748007272050133,0.360606936537351,0.5460622721059056,
 0.6954110623684951,0.838343805776033,0.6954874235473537,0.32157557365302625,
 0.7708092640253295,0.18333651233077564,0.7238965937060126,0.3076533229184568,
 0.19998638532206214,0.7723105970237887,0.8212071886351182,0.8378109183369139,
 0.9046796457892455,0.12253163702087788,0.11368906738544415,
 0.41640541955767607,0.31843043676405136,0.8915744829713882,
 0.23708317102894616,0.6665007409113667,0.053446442089502,0.8086749176405905,
 0.950015134941372,0.7124542711422678,0.989055021655293,0.5811924718306523,
 0.3191055065260595,0.7320695537474102,0.7796456012734286,0.4041929182969888,
 0.4334532121091126,0.04255782064258806,0.7650620375915268,0.8973281887424974,
 0.5917199985517351,0.7024112744438793,0.2154910863108317,0.8587993875406804,
 0.6511092753252192,0.5899845044707397,0.7828147022854766,
 9.561582353345699e-4,0.9553217779057399,0.7337867880021652,
 0.7114974847084763,0.504081123579695,0.29415679354149027,0.6347200241647422,
 0.42704962356942855,0.5778766994715825,0.7038529866963459,0.5368154004395844,
 0.15507962293688582,0.9633607393333867,0.9278559023860108,0.5605483835493269,
 0.3702585007812763,0.16135760001290045,0.8757631941000387,0.7722058269493105,
 0.44940324778840446,0.11286425824350865,0.5058462812086231,
 0.3773899198185683,0.07050109162165064,0.08221511807876736,
 0.043943031509358854,0.5075272598237717,0.07394488978624059,
 0.8044277821703464,0.6889151869297799,0.5143293441033747,0.3333390705981554,
 0.31682797973291854,0.9522170290854377,0.7502112751316837,
 0.14937736513756528,0.07333361977817787,0.5905690018979851,
 0.6810630538687656,0.595183511022772,0.3999495034978262,0.64188516144234,
 0.15582795023618878,0.3231230461775403,0.11566342209084113,
 0.6016433179393209,0.22088536123269098,0.34853844950046253,
 0.5063320187863506,0.16690493685244479,0.8413887679901083,0.1033682086605422,
 0.5036081461190989,0.05000801105530572,0.14991489948370196,
 0.08402882983236337,0.17911352496127964,0.5517632891865907,
 0.035715296178371325,0.5691843971037908,0.3700566682510268,
 0.7497811659115079,0.7461888555284163,0.3786138312112476,0.47656956022170016,
 0.9382645739628084,0.7144800246336708,0.4243230515485106,0.2303044089998968,
 0.3339606978788099,0.9991847101352395,0.6196852523727253,0.5910467050264278,
 0.27707383711630396,0.5225331431013701,0.6562522133005096,0.9473799730309365,
 0.6012699249715665,0.9018950487516635,0.6495176600523055,0.8369338368810352,
 0.8089072074997495,0.7106021725236319,0.8964777639884954,0.4958391098424457,
 0.9678335422517181,0.9513403007717557,0.4950559592983823,0.06526198422049678,
 0.061264274228164495,0.015144177830913819,0.42408790561385634,
 0.29369711199054915,0.021125575298083454,0.8719365081141726,
 0.5891544956717295,0.6328933960258434,0.5850616424553263,0.5197756797698289,
 0.9746825355444221,0.14677743787562925,0.0777729856678896,
 0.24802714675819304,0.3562295772418167,0.92359812519065,0.5977152816280948,
 0.8178998982366327,0.2011722547628918,0.3414185598566032,0.8420836971571486,
 0.8725761660859958,0.9007126198895319,0.42696958993815315,0.2127523785772909,
 0.5367094610523685,0.6964050483874068,0.38857097084241843,0.6212627518555421,
 0.45143243799249644,0.8449083591464114,0.574269706471778,0.3747558868146751,
 0.5842371999553131,0.5318789994409432,0.5931086189247433,0.7613763996773825,
 0.7376901001166367,0.2418703689054188,0.8472964946797369,0.1879831968040443,
 0.6899598999792834,0.07709901011041165,0.7140578062105006,0.130326662597524,
 0.7209312652095557,0.4482095687965879,0.6452369377624608,0.6522359418592925,
 0.09764429661324492,0.8716136721374772,0.6044994865970494,0.1882909484133486,
 0.9923814764378424,0.11955622518642572,0.6500445408698788,0.7812670959722858,
 0.1929873186040162,0.11417934163488364,0.6922851246011189,
 0.10275631383926198,0.9463551594730183,0.856899402067724,0.1560135605508588,
 0.2820833244316262,0.6657047429362449,0.6699650356292086,0.39895962107367,
 0.9043466733934731,0.9362559393232726,0.24508392773389176,0.44833885255379,
 0.7063188999372211,0.3059219236703291,0.9204045998221501,
 0.018578577504054117,0.09860656082831798,0.6398097823647941,
 0.6472321382764968,0.48828924482508973,0.6215866291796758,0.6413280055516637,
 0.3197240262046327,0.8419642429489773,0.6929471442604755,0.582756094452638,
 0.6339801083821048,0.7092004570197307,0.05706000884943063,0.8345843994290623,
 0.15203291298438426,0.40146304299388835,0.6009921892999227,
 0.8102160318294317,0.14812064617334653,0.7679826111540586,0.8051073140677597,
 0.10166699544423086,0.8472561056510012,0.9680811960254806,0.2655633431331076,
 0.29638969973698037,0.9961120474182849,0.6946207305597566,
 0.24165708190742186,0.5085702666247679,0.42307230653435945,
 0.9449044950481387,0.5029940177787886,0.8510170510839675,0.6299746116064255,
 0.10935606279816412,0.8963357780563891,0.07592284041922892,0.782697959343844,
 0.5931595444294738,0.6905154153278008,0.5476301579173143,0.6806554962946426,
 0.6895747510857956,0.5975480987345729,2.7951903086975705e-4,
 0.21273511670509015,0.3140453646086494,0.04143673655492852,
 0.8539581244861073,0.8279268147741676,0.4015690073511722,0.06107719306861248,
 0.6929410482569484,0.9634209270092481,0.9896806902876905,0.06503886109757118,
 0.7047279640659462,0.4600181776860538,0.526445938374222,0.7075006301925011,
 0.3271385127744033,0.7503023069588288,0.35356069536221946,0.2835176646468607,
 0.9380290907647829,0.5679513636242371,0.9091179935182396,0.29785945789457724,
 0.9158402702708276,0.8882056961381329,0.19452010620486726,0.7448593256872422,
 0.3090589952132179,0.44555883961525333,0.9124833698231323,
 0.13013495893321314,0.4945637580914606,0.903314874584234,0.3818313297997068,
 0.39076275259705784,0.4334601343186235,0.18270230091621398,
 0.21368790716742536,0.9755563528016873,0.6961245192131372,0.5698696744182328,
 0.451113115751008,0.8796816397542379,0.7897295304548532,0.5331143056006628,
 0.5130934826438739,0.5529052129684584,0.5365283983813502,0.12232665954055122,
 0.42570099343243994,0.3763997447137173,0.2469722936177875,
 0.15925021137027884,0.12442232056746705,0.7526297727149989,
 0.2051783575813143,0.2797098409112969,0.42867333267164154,0.526384318895503,
 0.14808976807916008,0.17921749198403447,0.09631146687372916,
 0.2562037857486965,0.46798112006394055,0.58706485791235,0.5794644007822846,
 0.2658207526077827,0.23823709909829915,0.6614550200532081,0.7259287101138829,
 0.7395889258758807,0.3776574997096014,0.04157407401968538,0.834135880419314,
 0.020636040782760046,0.7079705191436412,0.7432634528318879,
 0.2546577944265793,0.8886504122568788,0.31259399356562356,
 0.18762420047230288,0.4306528569988183,0.25197807675000794,
 0.20327133235820438,0.6035537372568738,0.22707645883319738,0.788312500996873,
 0.2860844649131884,0.4389736447868283,0.6295949510987685,0.27498595258538705,
 0.9093414362567587,0.046067701883018675,0.44326169408032423,
 0.3437083846586436,0.01021618425490256,0.5962155572765824,0.2562873557631026,
 0.2601777118904893,0.6005671633184664,0.7358492183002756,0.6776965406439073,
 0.2680600796728354,0.3456531374372688,0.9763540124848314,0.2326210262372933,
 0.8461356773612922,0.4105091549574238,0.09786824320675014,
 0.49357326658010137,0.4099168070918102,0.5017968461585349,0.5239361374541438,
 0.3171449222600836,0.8673280198326899,0.41608364255227603,0.2798013924109617,
 0.4167585345877972,0.18867189391108874,0.7493121004229429,0.8524532415162429,
 0.5188204536744682,0.043385799760169985,0.6035593988101988,
 0.3220348003882094,0.7432151835618585,0.6714421447105122,0.9710984108990944,
 0.3080658985663296,0.7978551476388699,0.4044316996522467,0.542112863145682,
 0.9329399300074924,0.8694563849447176,0.16492445366418917,0.5061080372423258,
 0.9915635082769726,0.12049967237215709,0.1031533629256155,
 0.12590854970661858,0.03564503504414973,0.17069624870702815,
 0.8493173689069653,0.6762163328303062,0.8596887363423159,0.40795939072266396,
 0.13768511524990557,0.7815642223547458,0.6271476525601495,0.1489027789707692,
 0.5788314611671729,0.26406210069916625,0.733917643851703,0.13865490272063608,
 0.2295305472892788,0.5314910243640968,0.0609202818630209,0.6531532035046267,
 0.9492633721450101,0.06624222771578481,0.16191614838777235,
 0.5029675753267235,0.6126620781918871,0.012825015586238164,
 0.5401362495180467,0.10240232589700515,0.5585722775144176,0.8132253350618924,
 0.7139708273492402,0.23652067931143073,0.35620589007155,0.3764825045434974,
 0.7681002772472852,0.5034071495747207,0.7345913103411517,0.25129060292543537,
 0.3704352245306287,0.6564453148601747,0.291428396518037,0.5064911512917931,
 0.9998443593860789,0.9659979765148778,0.6500274791540073,0.2326021584629836,
 0.9395230715257141,0.9134499101496741,0.7026112025999576,0.9648988857890877,
 0.6488790341623587,0.36686832902198474,0.8107869051163037,
 0.32264043339176707,0.19097445941757663,0.5355880907174193,
 0.07611314073789177,0.411574777403346,0.7664377142900721,
 0.057415020472668554,0.5677066746640568,0.5938972585901029,0.959478561133657,
 0.3109067903371395,0.7091854954367287,0.7582764086896284,0.4080831114359682,
 0.7612952318198096,0.12341180012417119,0.717004548169366,0.3686301249585219,
 0.7288251105954212,0.44321111797979174,0.592067912434362,0.7439815623068543,
 0.718278057527461,0.7001124197723381,0.27752058681612735,0.4726784014786998,
 0.09400387268804722,0.9782098665235621,0.5207148816207852,
 0.08974851609932555,0.7981550810011828,0.22550075119274915,
 0.6578890072687973,0.31537014799791707,0.3663199254995475,
 0.26928443712939854,0.9780045736162762,0.03476166619650267,
 0.7323293593150834,0.41550917806175747,0.5002738919458631,0.2504015047963347,
 0.722224308752125,0.8807395441843056,0.22430526539631812,0.8882660582111896,
 0.19744281287649024,0.148088576828577,0.154486120756806,0.12448764003349888,
 0.8564477006241382,0.22171037618984224,0.993055843145001,0.7340830206994483,
 0.9813178009575461,0.28520303822738746,0.1918740431413669,0.6458097415235071,
 0.035756047506381394,0.33876292245025286,0.9294542420551721,
 0.08785179352484929,0.4291650182317226,0.9554798894485411,0.4659853491612863,
 0.22447527348568397,0.44980464462416736,0.45726652474315943,
 0.435342273856984,0.16983962782443018,0.6445380541663205,0.05972377845439758,
 0.9648264099749233,0.27699646969910385,0.5859561754897842,
 0.07415532919668733,0.6656723947056342,0.9328251762475293,0.8564522399166056,
 0.24835224913374487,0.6857014177390541,0.27792716454358524,0.965214649164743,
 0.8219604803584304,0.634333828729465,0.8116116277190186,0.1957020902002603,
 0.9232226116723643,0.16212066222929422,0.676905412714414,0.39758937651134674,
 0.744345171515378,0.07688293186876838,0.46101323193906396,
 0.17725880142604633,0.5121760829015474,0.6735772406129505,0.9875263933659416,
 0.34818241252823057,0.341558139577953,0.7444087319899968,0.9280091304929973,
 0.974800066909401,0.9271475294070217,0.936967196364497,0.1204665985368647,
 0.5186854979111345,0.5341608297814081,0.7250869115756772,0.8052263397621158,
 0.4137507555226635,0.9836493450988276,0.14489413568674014,0.8940041639191756,
 0.8160494162064065,0.15422764352779228,0.029602190749358437,
 0.30472262782751647,0.08130912305692384,0.6244408211984981,
 0.4901519935553048,0.2731212543534548,0.5794565329579446,0.19124725980499246,
 0.3110013645026841,0.8102807241462407,0.2778179370204644,0.02296418455216742,
 0.6733743266683333,0.5271481235078589,0.07509648388069723,0.960930600918122,
 0.6869760426759687,0.942346102735276,0.09783078163964232,0.7408142633853813,
 0.6377472685160099,0.005709654898670036,0.8620901309198379,
 0.24767789391776107,0.8018500911037559,0.29990285266308125,
 0.13832194157579414,0.47179964044205625,0.8395846325265526,0.998567159631949,
 0.6372519406611924,0.29933877201751524,0.15917267536867807,
 0.27523583261983076,0.10168954324044377,0.5375555396657253,
 0.5112390512693172,0.5769046988345782,0.5088265735010686,0.8201440250595019,
 0.16448639552064603,0.7885795612957008,0.4404878500353302,0.2540175548680256,
 0.4770603511731437,0.363869349262838,0.6843371775677368,0.024366573416792825,
 0.43200593272663546,0.3547202636021377,0.9930379583689674,0.8393110347234659,
 0.29668458337747416,0.0712835337647898,0.07660004432318834,
 0.12267515720753053,0.645057735319799,0.4612358490509949,0.37468463919621664,
 0.20727384100116464,0.48845153647992656,0.049389987686951686,
 0.6444562774849427,0.945349399399313,0.8037097447522669,0.9606626415352407,
 0.9613989648820707,0.49727438336069796,0.2650099484240268,0.314932480079705,
 0.12898018852605753,0.14953905539966594,0.8034049481793288,
 0.6804492468495116,0.602211079687678,0.7808111294531115,0.4167247293654379,
 0.7000152193733868,0.5806643497852871,0.8640269538451572,0.15907282618247953,
 0.4215568548912023,0.5108907471038617,0.4808096041701754,0.30769820218165167,
 0.7395663703786958,0.2609203159096599,0.6026998221688654,0.2249389609781246,
 0.9646765950820971,0.789609779072964,0.3750567376895646,0.90260854744738,
 0.30283434899791895,0.9154393760587762,0.18082136282002137,0.567319174444002,
 0.6343623831609099,0.9261514435146444,0.5377867465597533,0.5757939242923699,
 0.23898169158981053,0.03653937198291457,0.7808784619987347,
 0.47133064937502334,0.7087214453209261,0.6789586922030295,0.0797968146869984,
 0.7167890488385815,0.7782557501466887,0.7679059381911517,0.542965235494617,
 0.4370506310199679,0.9280058392846701,0.5216748392876681,0.6476167365568961,
 0.051480183663319634,0.4330686614740137,0.8449092812943666,
 0.3309080033484676,0.0026280052108914598,0.24113828511218127,
 0.5773947993537448,0.4434186702904941,0.9765810156952708,0.30837759638393203,
 0.8465001789592066,0.13992054912350582,0.8776303640007402,0.8155758409802262,
 0.04058306379562393,0.33272143200398807,0.14816854080517783,
 0.7938268157206354,0.32363394301033754,0.8101193919483042,0.7405619415358995,
 0.07378565789994374,0.31858535989103387,0.9038851507910276,
 0.6170617894865155,0.558021777487437,0.6335281644052869,6.276896600743154e-4,
 0.24896989256013824,0.38840602799639556,0.4932902587576803,
 0.20910889144116196,0.4950437139358298,0.07086582451247359,
 0.23263952848488312,0.033695892281494166,0.7130835223942286,
 0.8008126755884292,0.2933037439446873,0.6380287216999961,0.4161337994522831,
 0.014342106201080007,0.17424474683704394,0.8936448132646171,
 0.04705356571308639,0.6195255190612233,0.23090105849664144,
 0.20625740504958912,0.568226935520376,0.755506019000578,0.13831906008426198,
 0.5726282890624299,0.05900614966250306,0.1346608558049558,0.5793328472633794,
 0.04103580949405983,0.3937828168643518,0.7595291771083547,0.9075378592240153,
 0.48762492641715494,0.6278703666960705,0.7719847974837327,
 0.35664061719011153,0.6228721885663978,0.6030129536033754,0.4849034926656176,
 0.5144355910650826,0.8744765952144893,0.3021757882430838,0.17223085921247983,
 0.9445673391794408,0.894564014150329,0.43047156682931076,0.45056120443057956,
 0.1466865180142939,0.027178509207373303,0.13362518078477215,
 0.7261636965321318,0.3624909941743626,0.7447193279625572,0.6050849838882589,
 0.2852071100798421,0.6621545410408842,0.27633813059914514,0.8549496927419605,
 0.9803754972015123,0.9464322190221296,0.38184241987489087,0.4628559666897988,
 0.37455990007446394,0.6407312805082447,0.5148559574468378,0.7502703273229391,
 0.9387210291103223,0.14748708028847302,0.8592123915454023,
 0.45845131558082397,0.593377110089868,0.7804380427683086,0.5551610634873473,
 0.008029648377987586,0.21522863645768708,0.7165640609241799,
 0.19808899427320048,0.5150463592298343,0.24447612820136955,
 0.7750723811383181,0.6358103802080576,0.04494079259693451,0.9598407297546535,
 0.7772587563593322,0.9514513113999448,0.44064196870445227,0.2087989345707777,
 0.5923788344626508,0.8281002330443581,0.09694676032129834,0.1350029467802184,
 0.16824681962249444,0.06795401526715006,0.7070492478560524,
 0.5248839311956514,0.8178046151401032,0.1680934508866775,0.1538275492591472,
 0.4661233338978078,0.4901739788319712,0.961116045429634,0.8227326788306969,
 0.44533356062968565,0.15753180880017137,0.1337768328465414,
 0.7771615138733458,0.020985408512685044,0.6714057132259738,
 0.9714808783602527,0.42047719798460825,0.5670127090465344,
 0.29735982079292467,0.43801359784722005,0.0031769372884098424,
 0.19130141365735776,0.26767273351987186,0.23449723489251717,
 0.8938214788313571,0.29924517500472003,0.25127623823580747,
 0.4075364296983631,0.022530751076433653,0.73061670613552,0.5295509997152211,
 0.7061160547586613,0.5337897769351932,0.7486425252449374,0.32414450917346427,
 0.30711156286711716,0.4174016424520639,0.9745807899901076,
 0.034477654735467156,0.07171973355366779,0.930227728206999,
 0.07496612132293323,0.7971880033272227,0.5229790796024701,0.6579117265666405,
 0.011888125340664524,0.30837912330367057,0.8509356058995765,
 0.1491129423988271,0.7788148539888003,0.8780877595842854,0.8409556378232739,
 0.2839494357892891,0.8618776992908603,0.9932572106823236,0.6700398808661877,
 0.11401165722762374,0.7554666122675723,0.6471538987425056,0.5043401069854305,
 0.5626908403175161,0.6877577753820857,0.9924715898355084,0.4229231677576041,
 0.7807583832704541,0.1565920435738153,0.2742547714176343,0.4118230530859949,
 0.5995683111505615,0.2096487623613411,0.7928053161898785,0.7774520686286651,
 0.3132632768364866,0.9361741709185739,0.9294768509929892,0.12437936280096173,
 0.37603621061978487,0.21283233174337968,0.5951198285066706,
 0.06479993321733324,0.06440966305799167,0.18307124861881252,
 0.546798438443891,0.9716623951987917,0.7747043730219932,0.8720102858905475,
 0.30869066785707944,0.30537563542400703,0.024288257666817747,
 0.7868862200046863,0.6907747085303657,0.18709412360877997,
 0.41862480807163227,0.5916557314261783,0.9167263059635673,
 0.12310029917512211,0.7088270994386845,0.6529735745588028,0.3642106203646538,
 0.32344940221577967,0.99871629928183,0.9091716083177418,0.5888715485675091,
 0.1748941364680061,0.5636942934631438,0.07577206191823205,0.8781964095568302,
 0.22059433546594043,0.1869079211384045,0.8539329165171274,0.8584835217431052,
 0.4885367235631629,0.4456201079403386,0.07124408192525422,0.5700006501062161,
 0.9508119804580945,0.03278833460610919,0.9154777420490204,0.6080022477855855,
 0.9112013725581537,0.9514258708589158,0.2782317010960671,0.7818763193536064,
 0.3101596745686448,0.06561210818184893,0.5394595635308175,0.9873668542156724,
 0.9871391360736236,0.35452766268604485,0.9354405543131983,0.0483338276500892,
 0.6745365941819421,0.052286670885705044,0.07890739092055032,
 0.3732151979912509,0.41638267961859154,0.056167079288842126,
 0.8807714241559114,0.23753151577689402,0.09279772602158254,
 0.8079872028384384,0.2564888735950467,0.29251069602833124,0.5538906016092909,
 0.5563170137468989,0.20785418280504542,0.6545292045763144,0.3494016481112793,
 0.04524384995624675,0.06503535978303454,0.3719737778018717,
 0.4157434987087357,0.982834171373561,0.8784083494048824,0.5786999310630117,
 0.6636656787173396,0.2633493550503988,0.37153078556129293,
 0.09197947557783515,0.789183550382541,0.5281884431399889,0.6688031399747499,
 0.09223009947564842,0.16333121341192247,0.47018435992983443,
 0.4158085455720867,0.6218565243560642,0.5446027535106157,0.6290939297700241,
 0.6501434121844374,0.5558558658752714,0.17452984508609792,
 0.25761948641398913,0.690420372745047,0.8143067810676076,0.16941698063977872,
 0.5962922113995784,0.4776557490886868,0.18882652401262057,0.6299276748206554,
 0.5682037970739895,0.19912314371895679,0.6991311546231641,0.0862576661816532,
 0.8328771382474655,0.7566693682669746,0.1578013162397114,0.5815369909073971,
 0.7158792342522416,0.13889255325963457,0.15069181209154192,
 0.6643428350239075,0.005290929487291374,0.5256776240952302,
 0.1415669402640839,0.15990406607518348,0.37676351881273185,
 0.8930128414370493,0.22581836566557878,0.14108241630933172,
 0.6858885722409163,0.111855026259438,0.7377314523161861,0.45515037837557026,
 0.9034691025383208,0.6077518235906174,0.004535648005085546,
 0.039531822651230675,0.08169019046585879,0.1711922675894837,
 0.5990390536373281,0.4628914618104787,0.7030094138279118,0.33407925638370894,
 0.7193757655666871,0.02181621751726237,0.37597954315402793,
 0.8228578293953572,0.8681222097502729,0.34956806934478635,0.3633423178850165,
 0.2857336982535519,0.9282869436593726,0.5908681727231865,0.05953797089780699,
 0.03916171271921742,0.052774486963312106,0.6399149000993407,
 0.5563136563966656,0.3182289617300825,0.40244188595859853,0.5997593639604993,
 0.7961870825486439,0.44469095492308774,0.9177348210492516,0.4843041245730172,
 0.7010488595362352,0.9283977708343873,0.7982235308939641,0.7522391595547008,
 0.8682609393410015,0.3030660346422617,0.3087722315790399,0.23773201553465084,
 0.2147988551140413,0.35475921831566914,0.495197005254293,0.08807677758970434,
 0.5586036229957533,0.28633779416390737,0.7665596428378401,0.6784765394158261,
 0.662390119790676,0.7692062085060972,0.6836413695225635,0.6830706393432264,
 0.5570440338300251,0.012761814845405728,0.034437830692845894,
 0.9382047896051613,0.2961455770286959,0.5592666578261448,0.8109619389228105,
 0.020260681706893435,0.6630898321359072,0.8669992277154323,
 0.8032924651170763,0.7083779649931768,0.013340895871619018,
 0.48630799999972985,0.8366344124339945,0.6271967558504556,
 0.09239096901211363,0.8189081078395183,0.8546406648966648,0.9408920734317876,
 0.661281330493966,0.8287051171833535,0.12386626504291787,0.04460851187027848,
 0.08914754735953545,0.7590872034736514,0.32304419064744183,
 0.7315496308923148,0.5322265653875753,0.4819408211422609,0.285363997665794,
 0.03424101130166046,0.3126568951004922,0.9802210871140464,0.5090888485238468,
 0.47956646650838297,0.3302737144999126,0.4673791978131373,0.7408981135822522,
 0.19490643995742307,0.4781324513653218,0.6331213591306626,0.7059338775670108,
 0.12018740484706414,0.7457213212813367,0.1632049770954047,
 0.23959863032801998,0.6030637528252144,0.01374854787990909,
 0.4719832930689166,0.8519859151356586,0.2147538611541806,0.47614251704017685,
 0.7586336616123918,0.05982008099762526,0.21209212938854105,
 0.5786446330859323,0.09460968680416171,0.5522593424451739,0.3851717569439489,
 0.30475260465958753,0.07002436056966732,0.8193519167887677,
 0.4720125467817087,0.013803241114243359,0.6799962532377846,
 0.1072930007826951,0.6511276540966893,0.7961753533931882,0.5021436228917658,
 0.787676124098801,0.7214067273042936,0.8742767665394703,0.04741318520488491,
 0.9573344761875411,0.5651648618268009,0.20443896349620083,
 0.12844482371926247,0.1027775403725556,0.5758395725721097,0.9815226966472402,
 0.5492706129548018,0.2728895212648855,0.35795994021522826,
 0.05216687806662634,0.5760698756123234,0.8962024104086324,
 0.42884144909606015,0.07757133828932672,0.2147671906084111,
 0.4195353709636399,0.057571021350980356,0.3946668037991058,
 0.09693745923450359,0.9181259181185142,0.4778180487581768,0.9530427268280579,
 0.8101281810273048,0.022871638023875485,0.90900403380361,0.288305402851206,
 0.9151993397957487,0.12095339290922213,0.9137674857376212,0.3870485797926817,
 0.7302951965349301,0.4688303436281721,0.6793248059838132,0.24133237885519432,
 0.5466906513635466,0.8209265315632002,0.8455190465119985,0.6868793034312148,
 0.7866842048976327,0.6164647360202363,0.5518916244006971,0.3848714144911164,
 0.16822542246030592,0.6808720753674582,0.28379678457682456,
 0.46036138992439923,0.09469408515027766,0.5745923415653157,
 0.9529244482885829,0.4715091497193262,0.15922286038677425,0.5257038799132312,
 0.6673086145816077,0.5485796609199243,0.6280915274139818,0.3265112867355947,
 0.24052197835974154,0.8523140210674889,0.5293219800557263,
 0.17039826423115167,0.9178454435780223,0.3982061818893956,0.9679540803869,
 0.8773191427428236,0.8842885610278799,0.982875612724115,0.3264502257413111,
 0.5427330105003172,0.36199909795225405,0.09481178924609124,
 0.4351674424263561,0.35659649796022586,0.9116679263059405,0.3689022771465478,
 0.8046790556865161,0.6453285377423565,0.4455330359498426,0.2860016112617466,
 0.6907283663696663,0.25376128467420067,0.7257778744153242,
 0.31439405032044565,0.7778536417908869,0.4050000525986729,0.3306743855108032,
 0.7261582980899411,0.9459571677805427,0.5695343679047132,0.5457324074796167,
 0.7172589544490187,0.12370723158281716,0.6525996805769477,
 0.15957168501632968,0.6349790405324526,0.014039597761510092,
 0.3651519432075716,0.49498755266063565,0.7456395096362877,0.3471765172377216,
 0.9426133858203604,0.5233801879568022,0.9412567989283229,0.37380868904202624,
 0.55538791767504,0.9832878029504539,0.9210250631307368,0.7879791706170766,
 0.2358110275746792,0.2506191612756693,0.10616497129529467,
 0.43831183383104455,0.7886220693924361,0.7289880481333855,0.2962843116256463,
 0.6515612254737735,0.9720925020521821,0.47745987556661906,
 0.25731587023431124,0.13186088243251937,0.42119891755443073,
 0.042639314547324614,0.9782504193517986,0.5270588110480027,
 0.9816848649333316,0.13191297172835015,0.2493969924739835,
 0.08789931735482592,0.988900149770175,0.9419191548456194,0.7696613007518459,
 0.3847559341439605,0.4075288058907851,0.5481069028833963,0.4678474093278535,
 0.9200111042149822,0.4602926721311351,0.18285236866599874,0.6491120563674264,
 0.96187799455846,0.44828962214150203,0.8707815652348183,0.5527939263517455,
 0.8388536385725367,0.040186499232381934,0.7807734442977647,
 0.7342003429882132,0.48474946843011557,0.382015508166227,0.07468151159427516,
 0.7310779022405659,0.721449876080236,0.8619907011279804,0.9181271007165169,
 0.21727758237849348,0.6833671929419509,0.17385815523688442,
 0.3303772872826527,0.738219002631544,0.3422498409430641,0.3684129319904421,
 0.7139455536409369,0.563352602876831,0.6147345424582923,0.7160280611974955,
 0.4478674668416781,0.8108217537841962,0.05199576274013451,
 0.20492824909835672,0.1830447485595874,0.937988148612269,0.30749150506719525,
 0.22340363748785252,0.3771573405136591,0.3414145257335376,0.393492374854263,
 0.057335999368522295,0.6056324201427774,0.20669425254024776,
 0.10394966929844662,0.40302270254924266,0.0875649408055923,
 0.15007906486084976,0.3247579934566963,0.06721515666154343,
 0.0735661972193391,0.760656084694143,0.013816777998908902,0.5319897319365531,
 0.6714776664749235,0.9699303613420817,0.33086444509183266,0.488372539254192,
 0.09848200486437342,0.7977311731034077,0.2001387018331784,0.3355806189364292,
 0.8059449799315824,0.6723687713046883,0.5164262187702304,0.38065725725972344,
 0.39859723916449785,0.24409059928119803,0.599600907773596,
 0.10650309212435449,0.27365015780261237,0.5826437081807893,
 0.16194148166435218,0.2184466074815038,0.9672383336529531,
 0.14499366819647674,0.8453715115936448,0.6069919393164965,
 0.43217963670306647,0.4213244075732623,0.8310891048329475,0.8705128267548238,
 0.38315697092840995,0.7171438389625492,0.518721476559062,0.5897917238637372,
 0.03350247897371694,0.2441487083828755,0.6957030023614943,
 0.49484651812494407,0.006876491037018262,0.3583271416146826,
 0.8938166521763511,0.270980877652103,0.8363530443671549,0.41419226468942005,
 0.7642855592462091,0.9434219355828065,0.15512949360016925,0.8696279493463961,
 0.4220357845361141,0.8479659276623633,0.41398124749562404,
 0.41539524220309654,0.6220066945238983,0.23310468293927022,
 0.21973570091409722,0.23937635821284386,0.15000755481974648,
 0.1341157174483627,0.354244091032337,0.13188243760976037,0.46808471158507015,
 0.18273272125984463,0.9510564242401354,0.2665109869848076,
 0.47916892684996504,0.6853261143913385,0.49661743334424846,
 0.07639667629928337,0.7328806801589164,0.011831711286741431,
 0.21822054410021907,0.9621190622798974,0.8284796083070166,
 0.12300587221801196,0.830584618194077,0.9860938794522874,0.9381042208356846,
 0.48953571294729326,0.7724466717535251,0.7803702906338963,0.3079335075321067,
 0.1682580664785578,0.20823656666060064,0.5397320767116063,
 0.26792320284460136,0.10990101826010268,0.6332314736478938,
 0.7238535977703714,0.22788442479191895,0.04206112756775515,
 0.13528226839797797,0.7282036545520842,0.856115125979333,0.5680181577910152,
 0.6667875770578979,0.22133833125868063,0.7082729837200312,0.7294705497894283,
 0.11424758367417076,0.4152492680056743,0.30899576805835915,
 0.8998350974980777,0.19569550389251,0.5161315508532106,0.8012667375586338,
 0.8667918302592978,0.10432987490740075,0.41946746933976353,
 0.9935568714107407,0.9737904438131472,0.3826482153399584,0.06603509164553734,
 0.6664282405168706,0.3722119297723331,0.48725402351836833,
 0.23865085270678166,0.7740971347642616,0.5451159205170559,0.6656934023742486,
 0.6617810411993109,0.19301261645789247,0.6672636316916578,
 0.19897585934990236,0.2632491551605577,0.20946450140960793,
 0.29921984545935976,0.7146204036344161,0.822660040787885,0.9237046462798483,
 0.39633342389448467,0.8529219645274699,0.7680238651299122,
 0.11317243563669566,0.2743495244053964,0.7012721600430033,
 0.10612965752770132,0.6071104303732027,0.5341095900423608,0.5704685501082598,
 0.3624648593028017,0.26935469925550104,0.503232911948211,0.09494087077465285,
 0.9548681067649957,0.9549361193370187,0.3514981085124078,0.02615957055999696,
 0.9878533402192378,0.946745290106986,0.16618254944699706,0.9174163936404099,
 0.060218525453453964,0.44395371572774456,0.5871304622064231,
 0.4025923759985205,0.786272037653849,0.7955107208196875,0.9148606087684062,
 0.10908191270839396,0.6539709888812691,0.7723089097123628,
 0.30000145091132624,0.6095653756405182,0.6803929842061381,0.3686638685332002,
 0.8177248921836837,0.8501550799845958,0.4328866239287652,0.5941953078917133,
 0.8966149372112096,0.9342111872165131,0.6065896598525566,0.7744974457185546,
 0.2864327817096085,0.5758884392797308,0.9649098862948366,0.3541807089624046,
 0.297371750734316,0.42057244180294684,0.8428614168016002,0.36265285587363083,
 0.3381147522686496,0.44264917550833227,0.2491840459655449,
 0.36426904102090674,0.3867852389857548,0.7388584450007234,0.3527777225797768,
 0.7403432707051915,0.1535602201668046,0.320711296855298,0.7133150998895416,
 0.2337965379068453,0.18434615667381737,0.9039566864249317,0.627554748840398,
 0.18645919680280365,0.13060584621245663,0.8542371440789083,
 0.25789604190456894,0.24330382072699042,0.779667045230555,0.5238263636789028,
 0.47194998733226445,0.1710384892782456,0.9568401819274379,0.5868278881316557,
 0.951416471349054,0.2987305494626453,0.2848685600315328,0.383708636864482,
 0.6089925815642182,0.8277497706746981,0.70144337317375,0.8209449853453843,
 0.07592211660602743,0.6147889228728793,0.5279391234689529,
 0.08737282419903414,0.2887926217935607,0.04780731589031584,
 0.17495790722736293,0.12777425132399,0.3799133698180377,0.09715667397045435,
 0.6693943864091992,0.8814367247457854,0.5843316872171702,0.11275756906359224,
 0.256649637594327,0.9607019242081358,0.7488131703591403,0.8781642507987336,
 0.09481377221355147,0.6036031954591011,0.703045089294067,0.4178934728307757,
 0.6197569508321144,0.7824556480227469,0.13372797523330604,0.2628595185423944,
 0.5906638233221142,0.3288794150067246,0.3956685815733675,0.46815429558949284,
 0.7130589118122653,0.09407314607348494,0.052902985004786074,
 0.34410452193866703,0.3521627358171727,0.021977924131009963,
 0.2647617907226597,0.8684285202459601,0.5590425197888245,0.38082233738716553,
 0.7072989249094966,0.9167319515319281,0.3753726538164768,0.2620550540552753,
 0.6308613237921108,0.5849423022152495,0.6470906124435436,0.9594755093238827,
 0.20880821150561846,0.768833773042151,0.6653987210715796,0.8381742642388116,
 0.36368347269727375,0.9904805092382054,0.9220989949050002,0.8948220263395086,
 0.9344755779434226,0.29646654334281497,0.2435591878338077,
 0.11323017448537775,0.6512369938710387,0.045211731613364226,
 0.021674324223111108,0.0376552465058525,0.825924678021477,0.5849562866369653,
 0.9778306804892073,0.09699977204074084,0.6253121591989104,0.2759623075599428,
 0.47820485235352195,0.8778941160112027,0.8656109311218306,0.6252391720933193,
 0.2695609451440545,0.4149667004824533,0.7052154656409821,0.9461892192216418,
 0.4477654509849076,0.0332160320351047,0.5613655557211836,0.983555131405916,
 0.318390671114548,0.4568400077112561,0.2127987177575068,0.5128056080192234,
 0.550502371062958,0.1646850123634952,0.22591004961693617,0.5889353753288789,
 0.7062246947454826,0.7973478887647829,0.15201903216823875,0.2116373873310795,
 0.791386049311531,0.07817894437740758,0.5851915425225793,0.796203912245826,
 0.026903879517171325,0.5382003971391307,0.9930976656008409,
 0.9206775450321265,0.2525276599849604,0.2164606345078235,0.878039984889583,
 0.9310946567199979,0.04617305539707928,0.8972744561004986,0.3111608402053534,
 0.16066027977407027,0.6100226568742875,0.8445125037960701,
 0.060054675327567475,0.7629733590569273,0.7033694436035134,
 0.5230077810590753,0.1933311489440599,0.05105292690972485,0.7994061079636899,
 0.21775684313307953,0.7689602780782383,0.3563379040072582,0.9792815712974554,
 0.8999886772897705,0.4040360101019066,0.06529642097283161,0.5478224775061583,
 0.8642791321876706,0.6673048141158849,0.19582634189863557,0.9364920198179172,
 0.5250585170337497,0.6666223004950836,0.8889109467458808,0.7292874635674884,
 0.594908110142132,0.5066136415394121,0.23236410579036004,0.5602118472737203,
 0.6015282528154634,0.2052727129321663,0.7323454341705831,0.24897802772643796,
 0.34526760693330494,0.12083453868762384,0.7635703233952117,
 0.6856580542940846,0.42266899699697835,0.6660643417860949,0.4717194867868517,
 0.6897981132536426,0.026018395439475817,0.6106575101770332,
 0.04124069784106177,0.8606911634474611,0.9467178563260537,0.7825893978219256,
 0.7938509578074804,0.8942977828545617,0.1305687231895315,0.1274201723897016,
 0.5460738392681624,0.33358810070724576,0.4605559538002312,0.4027144637568907,
 0.9584298026952132,0.6410667894557772,0.8154933797207109,0.7053848797970823,
 0.4367440251627548,0.04906459603834912,0.5162689610406809,
 0.19946435964642173,0.842337511839492,0.6444459020146011,
 0.025082598749493923,0.3829803470664368,0.4606456440858546,
 0.11367172800380287,0.9002061802776306,0.6331605293576958,0.54014088904466,
 0.6313200514143974,0.22773106582366687,0.06563490365428704,
 0.41733466281701637,0.32342205572332805,0.048792134076252935,
 0.4933986585875566,0.10409186955702188,0.20173349426225917,
 0.48504259578620124,0.9899549346835887,0.8260544686820275,
 0.09745074351409722,0.4154432491205696,0.008862411282902327,
 0.43537859879213114,0.05257588301165661,0.2740874161069453,0.393828032052573,
 0.44699214577693314,0.3842777748233762,0.3749456005363052,0.9375705850485131,
 0.24896739239882515,0.1827624761235046,0.7179125816564305,
 0.36123208009909646,0.40888755253231934,0.36459037020686313,
 0.04820360483155062,0.1261040353340367,0.24049785973742788,
 0.6411924132237234,0.16543911402397526,0.12493873392534716,
 0.8960350004428028,0.9983239572572155,0.1631000999868658,0.6285319293142402,
 0.7505392508603158,0.47586330435888946,0.2668054867157127,0.1732102276948918,
 0.7076378478287715,0.3552500786437971,0.8773781397124703,0.7794037173927337,
 0.9080021405980672,0.8812822282360221,0.7046075180565374,0.0544173027483037,
 0.22216819927856868,0.14709444745816336,0.2416146064151884,
 0.6232649840880907,0.7542295588759094,0.2522886949905371,0.2134494187333198,
 0.9198955727764228,0.5453916314047687,0.4293295404346664,0.924463144637593,
 0.7378415543359623,0.1586524036124637,0.9252823156230255,0.3964605990435728,
 0.29590277132854803,0.09877511495896585,0.3871691153987241,
 0.06082710680149328,0.6973010094674936,0.15176980877879576,
 0.09135294958939255,0.15042323481515907,0.48586288583788206,
 0.6360437854786825,0.06630830229927498,0.10632786733677779,
 0.6435847465537008,0.026632190427926195,0.32376771087481493,
 0.8286369624732166,0.9082549426609795,0.5388104049534939,0.9897808038097766,
 0.41699951046447536,0.7704724900851552,0.43718876610677726,0.7486781685928,
 0.754654386918686,0.2586488547140744,0.658614800017033,0.11508412714950655,
 0.3344263721896601,0.6665751592104889,0.7093717403373727,0.48664474789888734,
 0.004124521163238137,0.9697277375435372,0.4733705287276261,
 0.6310802417327976,0.5635842748897435,0.39723954482952395,0.6984171419507818,
 0.09920681663653352,0.07282695270555539,0.188448719461346,0.7834959164180106,
 0.12311953186348246,0.7082424528345505,0.22689795060821094,
 0.29440484568277536,0.5469340324960783,0.5611535214504719,0.6260210011745345,
 0.7908125496050286,0.2096404183233942,0.5497309887753038,0.19356802611056412,
 0.3999108086778369,0.6345929111376004,0.4376401160934851,0.084188936568806,
 0.4473371789228676,0.7857275572390282,0.6418004143144695,0.9876182552847798,
 0.7180742093701529,0.2801304530448482,0.15418859103844818,0.7998516808816499,
 0.6097857330746077,0.4864305889730245,0.5558813826353051,0.9373941138112445,
 0.27828591654581813,0.4826577745312737,0.9005151947690461,0.5987968488552361,
 0.4013724846771478,0.7433713359852052,0.08679792828936539,0.6913564012923356,
 0.08346137207121063,0.9736943039440632,0.6333494773433952,0.8339409909873371,
 0.11659144296358193,0.7406954226944603,0.5241558844156836,0.2378246150721779,
 0.3529270604774062,0.5197477607203344,0.37791625819650665,
 0.007236244576634632,0.7666584525180956,0.6283604757906329,
 0.5785987709373857,0.767230940311838,0.6880133254699154,0.1377626342062811,
 0.7109598020464447,0.831032164135413,0.5235471477344223,0.7263469002578214,
 0.27581340753512795,0.20850528780626698,0.40665711033368657,
 0.4127741662445654,0.027264924317303763,0.0383487605387981,
 0.01812049368212798,0.1080799720625708,0.8193335067649344,
 0.44874571537829455,0.8911783496646329,0.9044464404718875,
 0.17159653025014876,0.5848059758236869,0.08876420959578979,
 0.3269420616445332,0.9643190028418328,0.25450763116572606,
 0.05255461919505067,0.1529644638960571,0.38618836059682593,0.814738106783243,
 0.828449862032516,0.6955760529154187,0.8606428009779646,0.93192771355542,
 0.44229621399642427,0.7016252513013224,0.4631000538173401,
 0.08300158799064905,0.7180169847927389,0.9627431247867542,0.8154977131237525,
 0.2077955433893619,0.025708390185535057,0.7139587142923824,
 0.17452551741156985,0.5866000506267937,0.7552465245327206,0.6359325814290894,
 0.5912652233195599,0.6333926722202102,0.7354569013295551,0.5381018904402186,
 0.33371501971492135,0.17281444800452406,0.6054725394449733,
 0.41485222922771714,0.6201911326198358,0.9486578755712983,0.9788754625046141,
 0.393796157834568,0.5494217878289377,0.3062513605323056,0.24782716940077387,
 0.10436373379261843,0.47540451351211643,0.8558324728801757,0.134460819847961,
 0.5261262285280858,0.20741242951272498,0.5792313063674898,
 0.009169295800846355,0.28622979462170994,0.1981941315821184,
 0.9205611070700972,0.20352976842060144,0.03312260090720287,0.759189909839406,
 0.3850292839111432,0.23709141325642302,0.7676800372838182,
 0.20040191094719118,0.10203823917735944,0.4524906041393887,
 0.46438255254098937,0.3699151165804673,0.6192391639911031,
 0.02646044915169332,0.4293721760655562,0.7714872198886737,0.6771411303903018,
 0.5195274641719401,0.13947111683584645,0.10254407658985087,0.770281832179182,
 0.951141265520797,0.7027577335688266,0.6445095026953522,0.08220232287701701,
 0.16337610545364523,0.1861232307958789,0.9757868040447414,0.3241067565939494,
 0.17829749911730153,0.9254724434534666,0.20466861564386973,
 0.22184674478693778,0.9081776345672632,0.9522804829864778,0.2351518087017035,
 0.186091686312159,0.3079243050685503,0.16459273227444748,0.9653425006968677,
 0.4153936394563639,0.918835020511898,0.9689005281446704,0.35053430050724554,
 0.017235954444842494,0.2545086390248029,0.5173713413948072,
 0.5637655916319074,0.7645066529422722,0.4945229595971713,0.47350050402023613,
 0.9659506469011913,0.1168528101108669,0.9637224039831453,0.15435130873121083,
 0.05967784272491106,0.8783752981803974,0.32312230873346404,
 0.05800290344789394,0.2543561992822516,0.17818712058695452,0.532071703039204,
 0.14513950342986937,0.9225210864140613,0.34327406301468455,
 0.5447499813116367,0.5278092232625896,0.2077970944729095,0.5072448314091773,
 0.10364809485768889,0.8249068193566539,0.5615003130855132,0.768540759972435,
 0.12575144130087124,0.18975659976117654,0.6345729021347681,
 0.2750148752225293,0.819994205148761,0.18861537558265928,0.7887506686735198,
 0.67839982441456,0.26173000830672133,0.12296446281771356,0.443381363903221,
 0.3376821796307836,0.3291999342565247,0.4722526551302846,0.2464993146097878,
 0.22530409095542347,0.17750305560689839,0.724652030095504,0.7883984053129016,
 0.9049521020327772,0.8947080299681922,0.6334701388338118,0.45513978637033914,
 0.4385070934032913,0.41446560354732354,0.9849925961084682,0.5748055697963586,
 0.8693265951271167,0.8429851867458023,0.48586216854806097,
 0.44219931396228906,0.7565039077775401,0.5709203035436908,0.8288010478464417,
 0.5910747765856766,0.7423393931111937,0.4060658169536858,0.39294016519108854,
 0.5744784430796188,0.20250347496302412,0.39478191351281955,0.986948508428399,
 0.3489107248274157,0.026748846054107767,0.3480991244313456,
 0.36925933970745783,0.43567733064666314,0.9673180557512254,
 0.7478477681522054,0.6162355620209434,0.2013403518329746,0.713448998072883,
 0.48155895927369474,0.9892805871951065,0.3119766905810021,0.8460580107458139,
 0.5859349575270271,0.02089120545252865,0.016183698454071083,
 0.32996251142047595,0.17997930849163213,0.764702770487046,
 0.02970922663741926,0.2951117252111365,0.4521938955192919,0.7673632294941868,
 0.2499594279086339,0.38969872130682526,0.9673977933376099,
 0.38726947718721827,0.6517522701015648,0.3301514815760026,0.4364587815836708,
 0.1779921283853122,0.09388400821934517,0.13692549003717602,
 0.3754971586491197,0.9311313566000624,0.8056200908689024,0.25846424906470644,
 0.7413182850488489,0.15407716970460594,0.7904877631671259,
 0.035473422009205224,0.23989013056347286,0.3662589844449612,
 0.650684731856935,0.2738657677654147,0.30786831420024807,0.934759404129496,
 0.983809858273781,0.18324220330448704,0.8548492487143844,0.4624170113507694,
 0.48377641545529015,0.7515198958367469,0.7071732783254236,
 0.29429713755121045,0.688933636733536,0.8859662681080165,0.610836169926457,
 0.7890593205834748,0.29540560035865027,0.41144409501416135,0.632988370086568,
 0.34721925436526946,0.870829479788189,0.006197034703000592,0.285580148330256,
 0.1735057816719957,0.08868445797103253,0.5894613686944447,0.6195061752724585,
 0.28571567263728703,0.49911442213744217,0.43907526104463357,
 0.48695080956564574,0.5886388549933319,0.25172064291368934,
 0.1468366474252385,0.06630792296265442,0.2232239642043925,0.718008045877169,
 0.09848278954942791,0.25899247841036765,0.9548588616533068,
 0.37117983110304253,0.7039076567842862,0.35382329411115454,
 0.9952062260762401,0.7683625196757087,0.505778832029677,0.24154773277380148,
 0.24619597258941472,0.05269072484898596,0.060937541182805144,
 0.1584131567858359,0.3048897710617017,0.05528294224177044,0.571488154425334,
 0.17265954442953368,0.0411946745399856,0.9964723877405118,0.5340381204116664,
 0.8390707371212761,0.8044369926288778,0.4175461633024615,0.4480223219239241,
 0.5932843216285915,0.34675619267087554,0.06071758547777828,
 0.20069341485508496,0.4846815940480046,0.4253678190861977,
 0.08556637765850361,0.1463394079037097,0.9488782678897961,0.6494063170908224,
 0.42666997091229775,0.08852846942688852,0.32080612413642695,
 0.04749806431254855,0.9751683220683628,0.9394931531423214,0.4097576984860063,
 0.013185900546153428,0.22077554887122852,0.283853915795384,
 0.17369705657984824,0.11513381138674994,0.8663543549980492,
 0.37246865183633127,0.010739591681595662,0.8912941710619835,
 0.6211848709901931,0.9392912298555288,0.6119336027935584,0.1292467162455181,
 0.8934027570210787,0.34849122457363646,0.31561080433632016,
 0.016935761498071522,0.3498417679687794,0.7132783508200704,
 0.4749930267208847,0.348938635113032,0.9973635799855844,0.39169482891105845,
 0.11023389243068471,0.9521007187588392,0.7329028188096504,
 0.48317790682521233,0.19251207062239495,0.849294260486748,0.4838555904352293,
 0.5236620676254697,0.8890437951744881,0.8099149016478892,0.7780575873557929,
 0.2976280997491595,0.9542911970166789,0.952275952581633,0.5284073896599322,
 0.5434191395910303,0.43995051345790825,0.7400266413989157,
 0.32105907534741673,0.9647925467352816,0.48270518015804253,
 0.9235468005422993,0.7795140518631228,0.2079701825913013,0.5566321579976449,
 0.5844311779508058,0.869979464985205,0.9131229866180467,0.40484187177134523,
 0.4362517449706509,0.10793347144005316,0.742716422745503,0.33691034003220977,
 0.10577250988349585,0.22804927334222036,0.9946846674907657,
 0.8885660936065716,0.033967379899976446,0.9681591484675146,
 0.8326925834110342,0.06922952948956995,0.8548033825589866,0.9711468320204459,
 0.2717833312095852,0.015968379100820607,0.10966213926490598,
 0.4466139067074212,0.4775359076782587,0.421396773563568,0.3629070860333189,
 0.26323211101446,0.2933271463215965,0.1297032305180953,0.041000721417745134,
 0.4793613821072197,0.486688802243349,0.7892536574872331,0.6995683014185688,
 0.053300464805897896,0.2711625639557448,0.16369109626044065,
 0.19998658909670142,0.45197898171587303,0.4819944959663698,
 0.7934275339805313,0.5400759975264924,0.5447229426951261,0.8974438016699519,
 0.834089323239013,0.20712624552089742,0.08447991440708136,0.840489570343836,
 0.41740471352069386,0.6276853290598021,0.8170532615588676,0.5899221720175853,
 0.857131584092158,0.2556663483117917,0.23097783346954148,0.794570873367841,
 0.5815605353577437,0.07294874255977901,0.4395209592211091,0.830547829331844,
 0.04064557187001294,0.13710504173674243,0.5088843923130305,
 0.8030378091563719,0.7019942784449055,0.1512618734096064,
 0.028115896092275938,0.4022365922067316,0.967141905676842,
 0.42887507759163035,0.08301093025937367,0.4073191192048131,
 0.14081493016388535,0.8038273974945555,0.6647036351252805,0.4780063980639103,
 0.2509397170763822,0.755282761208854,0.0708911037634109,0.05048709244764349,
 0.8101621390972062,0.0772907241777282,0.32184039827419364,0.3886402159988902,
 0.3720084331453053,0.8225713034094285,0.7815087077859831,0.3114731954137807,
 0.19135349149381264,0.7921123516563535,0.08642691623734933,
 0.8678280421140862,0.7929565746173297,0.77233640096526,0.59091347959373,
 0.4524978771775201,0.5466477102179993,0.8009438212517792,0.6888133249151167,
 0.8467535242731409,0.46147135882998513,0.8515599918781035,0.9790914119917197,
 0.8586934732713559,0.9795471184302456,0.3425744603893557,0.4555200287738932,
 0.5872648896827799,0.15032453486531727,0.08296168890972622,
 0.8166778967890507,0.44257131071822287,0.701717575073102,0.5603137310748505,
 0.41252020598591876,0.39207087901649595,0.7364100241781455,
 0.1923006746758832,0.012939432393747952,0.9265838133958074,
 0.8436119903205361,0.8178215211209211,0.5217216045171844,
 0.040062868586859035,0.13837218283972263,0.3251773078897511,
 0.7237449196739679,0.7051481269230316,0.052118458358600406,
 0.5923489938606504,0.9650691134735071,0.7310261099182176,0.05343699074319541,
 0.7035562305679748,0.13978642402138453,0.19675072046707187,
 0.8993365423076914,0.4699680810610448,0.025462088017112694,
 0.5906004380745079,0.5675822515988156,0.35173815352446425,0.4493455085063758,
 0.44400263685405184,0.17649941280701298,0.7548008439298171,
 0.3152866858818195,0.08761096231951992,0.4224696595175721,0.6085415124506588,
 0.6939728553038583,0.37006382390865766,0.3511276516634908,0.385388024294381,
 0.9092940802867727,0.29955422889425687,0.6297039826382229,0.6778272754084607,
 0.5823350988520848,0.9533043359529398,0.41932123106360863,0.911808322165116,
 0.0662776295011911,0.1704073825019492,0.4280086534141979,0.7270400607156837,
 0.8993738922412481,0.15326386907904477,0.9024603581299123,0.6898281887124021,
 0.302067464478068,0.528306368709043,0.5567168096334142,0.013018750709442584,
 0.9201561151305082,0.5396981615223974,0.8796743619677473,0.1944879495051135,
 0.9590610256539904,0.9498575630376709,0.7792011782899888,0.31364961606876673,
 0.8920463009937262,0.3161997882559847,0.9063478694268069,0.3496611598861926,
 0.06433930444054992,0.4687769984706891,0.9591286122298491,
 0.45579524721538756,0.964173539820522,0.10604517767037225,0.9605090181898709,
 0.03365902202247095,0.5226211433100405,0.529452248621802,0.7614532332147161,
 0.613971645745473,0.6781101534879624,0.810328689058175,0.38561595332390874,
 0.8928440424920226,0.5046228196342858,0.4857296990193445,0.813424094789499,
 0.7640215741436269,0.6346966338840226,0.8176637859127281,0.2504885483843883,
 0.0021205493534569575,0.8685070904208257,0.40663965772121224,
 0.40883420476169796,0.24253514376568908,0.015763580066801675,
 0.2884243584924421,0.4989914626514489,0.1565242206814511,0.33985257731358187,
 0.39021427080675075,0.3318293587811314,0.9553624246606641,
 0.20851610774782625,0.6200087353078616,0.14934306479793935,
 0.5073094775657568,0.4028565838263901,0.7951917900872715,0.3778642982932635,
 0.9380349487149768,0.09325627299733696,0.7949620362737863,0.2487343886187613,
 0.3826725001097684,0.8501185591124765,0.4190370569568128,0.6300862483409023,
 0.8388490178594784,0.7021756240707444,0.02305524344201504,
 0.13031450571668013,0.32102427528056887,0.5021667371502896,0.528542226787812,
 0.16003124684923353,0.3539162823023271,0.7295144781874955,0.6136631085448699,
 0.9371694304239848,0.3029709330027287,0.3402803616401726,0.6121087809215037,
 0.9425697453136861,0.6974850325093999,0.07330727493121703,
 0.22413433337575817,0.4226934600593921,0.5739801547172693,0.9527932296285758,
 0.580804797088104,0.9420600866597226,0.3008005194831702,0.9305564669642782,
 0.9843225238403991,0.759053131964738,0.11226602384066553,
 0.022349888470050483,0.46438881717208824,0.8517914965252835,
 0.9743949809707205,0.5773436625584742,0.07338893494989573,0.3440799686878041,
 0.4123523671946767,0.9712963831358188,0.5734802481646721,0.4262868430552187,
 0.4111075901167911,0.6860232740775103,0.43216016166316806,
 0.08903697205398164,0.11388370145393145,0.6262807894877086,
 0.6074758673969736,0.38100579900190423,0.030563594996927845,
 0.12459085559867522,0.9684459504744408,0.6525081466216689,
 0.41014756133349706,0.5000235053511236,0.9419726480493906,0.7064352755221203,
 0.1280626735921342,0.07086571971743194,0.9080316903101084,0.887249608417292,
 0.6460951201173917,0.4039854532751339,0.5539604219656623,0.8849699487467796,
 0.906538818516417,0.026454439104853233,0.3008550718309444,0.5036220948648925,
 0.5693990911259523,0.29169782676072,0.23257076634609208,0.9150666704360513,
 0.6814714940542802,0.18001494411894048,0.7805458395063312,0.7983045122903477,
 0.41404443801037183,0.40537027699942074,0.12707134130752695,
 0.10161640067473199,0.6353717962446295,0.3080284177555078,0.6646349250009141,
 0.11597043036310062,0.9232877673091301,0.19199281369259424,
 0.9356372825876108,0.1716954340153204,0.35093530421340613,
 0.10617206082698416,0.5442413512541797,0.8162660489462633,0.6292226315902235,
 0.562181957128894,0.7487915617985375,0.1728086343069325,0.09471312435598223,
 0.45896579770198376,0.7937702179698659,0.47402320940564346,
 0.9689488786297255,0.22489437645473176,0.33539020965340294,
 0.7741158616363583,0.9380840516073299,0.9313567329034608,0.03068674199997601,
 0.9358577776468744,0.7179953734867048,0.46931438307244777,0.7236483948318486,
 0.7112345997820595,0.9053069686178605,0.2363907397889211,0.2470985275522657,
 0.6656133998038137,0.5211138319140864,0.816783074788577,0.7751415768327707,
 0.7480124591325714,0.3499157840966165,0.11943580253875286,
 0.46715506635732706,0.5017525797253083,0.4572908760780168,0.5221745295383424,
 0.10920619273550103,0.6898576218678899,0.6706445539224146,
 0.37765272708288555,0.9118042930537381,0.7576886032196368,0.607112726623542,
 0.6083153250217926,0.6069140488048173,0.65678047951293,0.4041193618426461,
 0.530921556324383,0.6545395889807322,0.25955616214186006,0.6585157870708371,
 0.9129908950197967,0.6742404308138897,0.5288141559473892,0.8086177120810574,
 0.4033184172799642,0.18602203325715672,0.6232930060107029,0.5026580776366104,
 0.2119840738851897,0.9290800408659006,0.9880646673900212,0.5215925255793032,
 0.7460563910297721,0.38494828566725636,0.5950386578680958,0.8863287167775903,
 0.38896816472772244,0.6229440199843899,0.9461016415716745,0.7819438419414916,
 0.39455435689436724,0.14394664307626315,0.457352088546101,
 0.16944571366727013,0.6078594189814611,0.8817493373073029,0.279791787096767,
 0.5168290155716887,0.681645774651769,0.015081009149299218,0.3979948928666921,
 0.1976614304611758,0.9347423352743612,0.3136583134269687,0.8788020499383089,
 0.8555128737699849,0.5014177806762272,0.5356312868062355,0.5275986573977356,
 0.30153146236892403,0.5647314540376012,0.7479741212277483,0.3405178058543793,
 0.6342316051918249,0.7231768571018988,0.17407576092218247,0.3397084310222772,
 0.7651151592500847,0.9337285557588116,0.26172579692057474,0.5112334549482572,
 0.36200550852809954,0.7228604690566037,0.7711875722894069,0.8334373495953143,
 0.5042459580387306,0.21295262967075146,0.26699917295016795,
 0.1455886373636458,0.34692310032186746,0.5930485968160776,0.3511984453960474,
 0.9575768470621078,0.12848608056742716,0.7490206893423561,
 0.049816521659410085,0.9910285909641676,0.3574186662059837,
 0.5766217784774816,0.6350023038530667,0.17778072396268452,0.446034542713278,
 0.9495033656068508,0.7508324651375714,0.003046397171858728,
 0.9980214357535893,0.8146189861728208,0.01361093911768596,0.4185587196860514,
 0.9190337801497118,0.2924204680727067,0.5588382749787882,0.23084674397037352,
 0.6183625627742855,0.7424591058178094,0.9669497584228338,0.413024201777213,
 0.0010895552630099115,0.2198049630941552,0.5010156785115738,
 0.49758149974395205,0.5095685511175749,0.295475919984274,0.8283315367719999,
 0.14755896428968762,0.5175741632896864,0.1627991791036303,0.9234705057394511,
 0.09391505190156557,0.11736229237007556,0.453608521475351,0.7242632235635615,
 0.6833545886910648,0.692145120241906,0.697399293813193,0.41012821505045405,
 0.5861808622237866,0.9368358883436149,0.825459224812927,0.8861006395591706,
 0.6808408378175961,0.31991854132653397,0.5114933911452421,0.9563759940561479,
 0.3664470739128054,0.16137834204595103,0.4549722971511142,0.4200158613730709,
 0.0735871119687399,0.25093527458290876,0.3304989637287651,
 0.37192805978699894,0.04352073954038782,0.045939053520922424,
 0.8265767080170093,0.1046152694603596,0.5499170285714954,0.49985003792844374,
 0.6267172512117015,0.023906250252286654,0.8054153653303905,
 0.27856598151506096,0.14734677500551863,0.6398018015196251,
 0.1500349113520819,0.010826399490301841,0.05296680625864181,0.52682112768199,
 0.7760582843822824,0.3923151566180436,0.24452221575967426,
 0.34367136007441135,0.6418426913008024,0.5421349641082392,0.5560832286023186,
 0.3411138369196933,0.7120007865773921,0.3979747075849739,0.5341587141341491,
 0.3723745237720273,0.5741423762902944,0.44367144763807853,0.8284592674942552,
 0.2577319190821906,0.5216421277569425,0.13999790117027522,0.6277782381155201,
 0.15500790617622873,0.493340710513117,0.9795297793842986,0.8769740834525785,
 0.7655248175062002,0.5556045526888874,0.3770603316649732,0.9619750688290453,
 0.6879255116883416,0.7582134906165968,0.5389776300574411,0.7788761967563692,
 0.457127726979762,0.6539950984069384,0.4007580876205701,0.4008438786465337,
 0.5582334988681512,0.710673636326865,0.461274007743129,0.8371518796951568,
 0.7656709397458612,0.21512759327719588,0.6097860112388943,0.4672208817381044,
 0.8693479853400438,0.9027795577555323,0.5666041712454354,0.4920126953658418,
 0.7899689524685283,0.2990601397841348,0.2552146890588254,0.5108581001827609,
 0.22852185413324988,0.09003824454075304,0.7954852446469712,
 0.8742871531769616,0.5675337159119568,0.471693099438797,0.6390876412006987,
 0.1706663422331135,0.4912102345238244,0.698295439206253,0.7120605479269907,
 0.6593325511370287,0.16372494690679673,0.21599229532860897,
 0.3766142327187926,0.24153235210286939,0.0333977596977183,
 0.24716354049170985,0.5673734695866675,0.4976401598828636,0.5921656352273066,
 0.11286388374888745,0.24350203573129559,0.6197944105925618,
 0.7880876234124963,0.27554530597985505,0.18450325245797705,
 0.7455643281578921,0.41068056197898906,0.6173848741820264,0.625740305870883,
 0.12130828422962148,0.48566376578283066,0.7371275474079942,
 0.14397347852424836,0.46526653200931367,0.7677276079588834,
 0.42621263540589105,0.44014135982948266,0.34486290614379966,
 0.6253368065897902,0.1856437935880444,0.5076485176581211,0.9588310690116498,
 0.8144457307789468,0.35176020000594754,0.05608677394298056,
 0.5773758334928769,0.33783591957985715,0.17294809691703705,0.815510515294583,
 0.6407285588217166,0.3225051021143288,0.24284044995166076,0.5968903562550038,
 0.18561627875462894,0.540122903583072,0.7231064896936694,0.9529594175502044,
 0.06078695775405352,0.751482572764743,0.24969183707358655,
 0.45519088356396464,0.45631734302687177,0.846726205446244,0.9141302848276794,
 0.9189141734943556,0.08123589580434987,0.5504221624576441,0.2862377617196902,
 0.792969647960335,0.2088819571602103,0.8074258270544813,0.368143270359929,
 0.17685125844528682,0.4610672781456193,0.8050507251268121,0.5807991897172158,
 0.9637049077898857,0.20693559244578774,0.9313797680252622,0.6905803685817172,
 0.609931186486477,0.5525498052034682,0.8861190280954502,0.026893922743436938,
 0.05318604611548128,0.6718111273505174,0.02535514681551554,
 0.9001429070963877,0.23687597713712027,0.37758857538555923,
 0.7172441738816334,0.19614752520199008,0.09849760679931174,
 0.3100741629388316,0.8861395995076224,0.043375197563287005,
 0.6425701069159389,0.3601690374602028,0.41866051931580217,0.2661381850443393,
 0.6437481544441254,0.0963704207089573,0.08002073443600022,
 0.22489696158190464,0.24007879186175318,0.7443636323616414,
 0.14375744272784718,0.359449862494736,0.1410662277706819,0.40924085038224667,
 0.21334826590498046,0.9662072624771246,0.7789476548373578,0.8121238651224992,
 0.5400474467309954,0.8256211065225733,0.32094347881838026,
 0.18158575207133376,0.501765628792658,0.2705650414788867,0.32320750109171503,
 0.48104916595484526,0.8812134768287182,0.3022389176077067,0.5220860407296023,
 0.666141268353588,0.8463769436779981,0.47703901852524955,0.5426346897308085,
 0.32002785261638333,0.9517564458616504,0.6060599047158213,0.8122619363107076,
 0.045332130942569826,0.4763847872372772,0.40255913879819394,
 0.8012528619872498,0.2152020255991064,0.6035710236912643,0.5677414725093753,
 0.42632169638250994,0.17127205883712504,0.007642619213632185,
 0.14933491985073222,0.05088164564955333,0.665188424292968,
 0.08916256959021829,0.6376664955998463,0.17574537114292477,
 0.9984061131883634,0.5140027572647446,0.2399216491773586,0.7483674675035347,
 0.497602768442894,0.6906709760154809,0.03390396224151215,0.26786521985003786,
 0.4408124476044354,0.03225805865831899,0.14627500730043574,
 0.2590011267343407,0.5464643760140898,0.4039817353662277,0.900087236317998,
 0.36173540606276644,0.4463676375688571,0.6263559761392868,
 0.044545711467492755,0.19778556028348526,0.5759130909355508,
 0.012987568105799596,0.0741097613806142,0.9358845452684845,
 0.8273857758554144,0.794732283238579,0.27727672719221885,0.9746145536665101,
 0.2142201023296929,0.4602625134371545,0.44397229792371706,
 0.24954077131619223,0.9692645996826179,0.6410442666517315,0.8669608198250724,
 0.2835489533892106,0.4053003773870142,0.8185649419473497,0.18138933541113655,
 0.3575504148927775,0.3510482799819137,0.6207730595386359,0.8740212589946209,
 0.33593308132045385,0.13824474760945027,0.4715452508255151,
 0.05657233864809286,0.8254399550800533,0.795361464082331,0.5481283536522779,
 0.6612447112079171,0.9399927203381476,0.6609730402462147,0.5746772414977908,
 0.4322659354523939,0.007228226719058317,0.9329740871279847,
 0.04258061689027859,0.8887271955177437,0.9478272324857725,
 0.16999408557689977,0.20648207499321214,0.15431095315916687,
 0.44273711026621476,0.036521990587930375,0.6193526206322421,
 0.046370802560567914,0.8263487614622652,0.18500586156731447,
 0.3824578235281162,0.279612481472727,0.6554341132210132,0.9714741836190406,
 0.56242355451911,0.9631791845054046,0.7708913060191089,0.9626084844575612,
 0.4252860399922086,0.035352810736233975,0.1182402697388969,
 0.9259711919147396,0.2875190631951463,0.019838884736693152,
 0.42388355260472055,0.7882851167208232,0.182515155034227,0.35893618702797814,
 0.2742180706903673,0.4826522968146447,0.5785170853141277,0.16783129020659715,
 0.5883910280744614,0.7673224855712397,0.8658058664996814,0.4540744772810238,
 0.2516584419193856,0.8493764355583758,0.9207287733055056,0.00963901314077753,
 0.932154490997885,0.1505584730604742,0.3013361554958688,0.666232206001971,
 0.7079402543620118,0.4119016287256202,0.6014322846080169,0.49867242884708496,
 0.9972594715398861,0.42522715878648454,0.9611307921610039,0.5600262280139601,
 0.520809248787313,0.08095305511894058,0.2872943335583992,0.3013541950266515,
 0.613559333252078,0.06096409954349169,0.9189173209698974,0.47256076364534316,
 0.7564405669517418,0.3145023580746773,0.9038617719930369,0.37612592087290975,
 0.16962569823602025,0.5913417016989202,0.49972801447118753,
 0.8852831208261714,0.8243560761022407,0.8386768199906935,0.9185733821280275,
 0.727637459049487,0.0011022368307702646,0.32005456707873714,
 0.033828793385017875,0.9317995537577308,0.37574078937017574,
 0.9975459799717646,0.8048479603641007,0.12118759996868345,0.94941429343937,
 0.888368660813421,0.5424416626882488,0.25346798602613374,0.35862923884682074,
 0.9686054601491735,0.05713524588306984,0.7325655859082658,0.3490069711494491,
 0.4979096155737629,0.5435891077150519,0.16775598118222124,
 0.44115780650554925,0.5904181900861314,0.1154240833197887,
 0.43294671444749755,0.4149123652325105,0.5898269224657097,0.5194139821598353,
 0.7832124721360119,0.6696340920287199,0.768010935862312,0.3918443127183163,
 0.30723006075846704,0.975889465395017,0.6223548046441304,0.17098068855930604,
 0.34004933351653877,0.5988313723126919,0.6302879976019333,0.5490690407245639,
 0.022089129937548635,0.6528176947801021,0.4016006579798038,
 0.6025395299886145,0.3149492063329804,0.8261580206230558,0.6732344874806121,
 0.6470089916927461,0.28868481667085644,0.9479249394058445,0.4446833469593182,
 0.5308613312982537,0.7935548400535353,0.4214988756801399,0.1665720503496948,
 0.117627187267902,0.3832345665604604,0.4050261431915825,0.33207708513111156,
 0.7681112331799903,0.07680846364091809,0.8903088010438381,
 0.49719751607536466,0.16737308461727485,0.8507636522110794,
 0.6754555940311547,0.9328350730552017,0.17142915223101574,0.588909079745077,
 0.5025728691713807,0.8972881256571332,0.9407553736640696,0.2816955666350588,
 0.29631393772166126,0.8441985168252782,0.14507427469190892,
 0.6023904156544597,0.32536856868195974,0.8636601077907977,0.3494989616874067,
 0.14548347052787858,0.09857836150328558,0.9999566060583911,
 0.10913956632421629,0.26372467712099745,0.2309413419290871,
 0.7929010635715488,0.4309049232768747,0.46584555938181493,0.2663295011353739,
 0.8344923334426101,0.5338214768357565,0.6757032044223713,0.23057464264313288,
 0.8313400037413918,0.6918833587208503,0.10343489299105824,
 0.16055585401796324,0.6280400931836267,0.8708380585735563,0.3675678477946016,
 0.9254104935333138,0.32566868243519664,0.03882130789678662,
 0.06334441644517108,0.979470908644665,0.4686214541785294,0.4711338052774099,
 0.7521357246986186,0.42912250724939993,0.44490712765476736,
 0.4252024219131708,0.026515457548687338,0.4217125813034944,
 0.9645811480686192,0.2734198820904331,0.9970015384169395,0.6425012781178501,
 0.7150757051905863,0.8093244076232823,0.5544651518318813,0.19218609679641507,
 0.454008989255389,0.759032810346013,0.6999714783449841,0.31202013316293087,
 0.8465916365707957,0.7381151399433246,0.3590663682272339,0.521037903474814,
 0.942821849554436,0.075757717244889,0.9584648334787922,0.7906301171021314,
 0.23832454530339653,0.2681863564366913,0.2907534223109327,0.6537615829576195,
 0.1852412105408332,0.41084088108422234,0.6960698939305455,0.9571782544137561,
 0.5939950088765611,0.529872106939846,0.537673125851539,0.6217443155658768,
 0.6856915398244254,0.990865996199456,0.46161964737405126,0.36901655135183675,
 0.8590319771374829,0.6418157229756505,0.998406371612918,0.7510657161583265,
 0.19057165901451956,0.26172695021153747,0.4908325063499793,
 0.5478479382822093,0.69049275209068,0.4475786546823194,0.07189236262961007,
 0.527222692664537,0.32937506574319986,0.7686396646425838,0.32000661792023455,
 0.6426429984491348,0.757243807990018,0.8161375971705532,0.6769947566924424,
 0.187034574067787,0.8091125171321537,0.08319309447192358,
 0.024111713997830075,0.7159608802558924,0.042458870763030454,
 0.9841530656621001,0.8132050390261111,0.12951001879556234,0.3457000710705933,
 0.28126021033841975,0.4939857438890273,0.8480462825329392,0.9100183743728856,
 0.9291783140694487,0.7047686620280484,0.9539403672332587,0.7789288975102984,
 0.3745895968140256,0.006624202976151405,0.49431293084322436,
 0.5855900808221478,0.8318147290661958,0.502478278681194,0.743631142251459,
 0.2319108364142014,0.9747895939344479,0.23325723932192344,0.4370064968283135,
 0.9194011117051853,0.9115183705438883,0.6476288313316247,0.05037513641088143,
 0.751302957256191,0.019109570527707898,0.5715823091553257,
 0.09407698199364611,0.5904817177057646,0.5164566293115926,
 0.25393886001190413,0.45252628663243677,0.04621023165710014,
 0.6504635186555888,0.923485204710244,0.03202082720812083,0.9566729011770388,
 0.4878158290000538,0.31784579336564867,0.21423496470313452,
 0.22426262684253595,0.27384072913132185,0.9603647682044427,
 0.050803819234434666,0.15134521982717364,0.45970338555529877,
 0.6848179268548207,0.5680105234113814,0.33328910897676045,0.9824505078245562,
 0.17150566967554137,0.752516001292849,0.09414633997293642,0.9906571461925837,
 0.18329402975170828,0.33021913876177844,0.9442548891633737,
 0.3838634348862622,0.9686419358256629,0.3968853245039652,0.39620815103724083,
 0.49788021780615954,0.4262416830829986,0.2797951688143925,
 0.38253022018769123,0.2165745832708572,0.9428196117749155,0.7427155405222536,
 0.5591945706740964,0.1276126488568432,0.372953343837402,0.28197719234103236,
 0.16973929508838126,0.7048849410373184,0.26164229305206443,
 0.02720270135163938,0.13142516302378449,0.6033580574090203,
 0.3306573160366617,0.659940161478706,0.6664474927858476,0.7106367362113485,
 0.7841518727635101,0.23463579905721255,0.8697354881795248,0.8800145395755905,
 0.8871538869788095,0.32298065586191405,0.23434136380958426,
 0.3045240758763522,0.4960070732965296,0.9755331031493298,0.4710926303513556,
 0.218924556964173,0.7982566371416632,0.7769111483123547,0.3697746612359547,
 0.6179561839570049,0.202038984716576,0.6058818384644448,0.16696062173378257,
 0.5910425706439721,0.9778342003461364,0.6620927699930387,0.8048428926596511,
 0.6752990996394623,0.677578685376493,0.41359477056062244,0.5142513391722372,
 0.9717073985563639,0.25699245058584974,0.9023548687459041,0.7699928080923869,
 0.6740501271859184,0.11437063448036211,0.8852060877850976,0.6623506172545276,
 0.3089713379869302,0.7549454648305295,0.5206626726291396,0.21736569476253753,
 0.7095990473688211,0.7076143393894526,0.26652969896210044,
 0.20138495375819776,0.7027900946649654,0.20726549279564388,
 0.8718101895046039,0.3696312947013758,0.2850683829663323,0.10221470527972154,
 0.48370134505186113,0.6831978879018944,0.832594697962199,0.3798181774162551,
 0.14796108135424313,0.07272798093611965,0.34361145668584425,
 0.1858021003362227,0.3206183827151088,0.948440621604878,0.19661855133843165,
 0.7104505359383076,0.8725171381209397,0.2011918941377605,0.5251046547244762,
 0.3761386949717783,0.3173191708085831,0.7275975094124645,0.8757524859412429,
 0.40004718286469154,0.7002533629650943,0.15304022655535532,0.097205685559277,
 0.6884340590909246,0.8043196185765216,0.749424976835641,0.18594696519239395,
 0.6213329619993583,0.24970687091715982,0.5374099146980307,
 0.17846560862405259,0.7517068034013992,0.4393975537069672,0.8325807060398642,
 0.0012032256612609649,0.04924756980170242,0.8512646619806782,
 0.2814198250948148,0.8308516005823192,0.3487133040875682,0.7511635268488173,
 0.4084303685524777,0.27709444548110407,0.344873036815906,0.08636327780038833,
 0.6326586571046489,0.49571931539839986,0.6760860273715413,
 0.12860057822408621,0.18751726992093098,0.9368601211427221,
 0.18297210598911073,0.422293165227398,0.5043260590520191,0.32532896843631354,
 0.7835401118827889,0.692519322533744,0.03165871132868614,0.44358332150228086,
 0.7758428482666526,0.19754344811389302,0.2966102373063556,0.7078201209907459,
 0.09011122625977985,0.25683409890072095,0.20567159303272853,
 0.6988641720063427,0.4449530219877219,0.7779625393179642,0.48131140455258403,
 0.15480918953718248,0.27747256865919845,0.7548079427420241,
 0.24462194678695137,0.5127760415139828,0.01984567871337395,
 0.6646262021425385,0.40153439954015124,0.17958180349437258,
 0.02538449365507711,0.37372620913678534,0.3086653784099298,
 0.22234517450049496,0.17704703504953312,0.060867843216779516,
 0.2173761866263607,0.9637443632872685,0.9522896457571804,0.5409651508702792,
 0.8749820429089823,0.1829815115972886,0.7786717264743711,0.6078407977879012,
 0.6969936102797896,0.8657109548608563,0.5499789028636924,0.492011458618976,
 0.6532973002690055,0.18203747265753178,0.6823621343858532,0.5163152120296457,
 0.5059583260140822,0.16032503907799922,0.4301509247890609,
 0.44573943398451543,0.08104792094661006,0.8628815401686725,
 0.5524327382461699,0.6682936207392438,0.1189865727130136,0.2766891615462348,
 0.35745994717287055,0.6128658882060025,0.40615175368015255,
 0.7116271338051522,0.06725189803807863,0.10604879606152107,
 0.8738812629040354,0.4484533575944887,0.199727169159315,0.3126079888541613,
 0.10724380327232763,0.21533366658760433,0.20299724973861888,
 0.6196548260242016,0.16347157370922394,0.2152255557734618,0.5479844212631442,
 0.16389810122992654,0.8144932061948994,0.7285267291564362,
 0.15313376841752602,0.6179113256697036,0.42763845255647714,
 0.4413064221433789,0.8455681502640988,0.2964637038845177,0.46131769319009797,
 0.16146782639575108,0.69077986373033,0.1491430386313386,0.8894146168734614,
 0.7338327729570893,0.6811823106279039,0.1597859014323727,0.606734049620504,
 0.5964908668017216,0.9877990599358895,0.8607452083055802,0.09642992693241204,
 0.9616874432429281,0.4624631574995183,0.7345877824314355,0.21006558063719183,
 0.025527851362334264,0.3352397924557786,0.16830347430976744,
 0.17756665950661066,0.9916661065647059,0.3180066350254771,0.7550526527481056,
 0.2778084771964021,0.6057830088427882,6.351596027942552e-4,
 0.5409049008521799,0.6193224520757807,0.9812222885049262,0.8322670853309151,
 0.9630144092408377,0.957047847058135,0.6996568839889102,0.38544897100013,
 0.7623885414310702,0.31788261298837295,0.16949826895076148,
 0.006250726561585984,0.5320921632194465,0.15155540043049498,
 0.5096288714333977,0.24691597591530778,0.06944057835798267,
 0.3740389252687544,0.35926663849468177,0.17102106056949595,
 0.5046870459723792,0.822241254490052,0.2097388015238273,0.49264340471021106,
 0.3568121141560867,0.11061071165709035,0.6743378029951848,0.9413693167077841,
 0.08007356131663301,0.5842719395780333,0.9061360083336304,0.9085254672117952,
 0.7548904704815693,0.3682434371125106,0.006766250463330015,
 0.1712166874906227,0.9797994180751397,0.1393696541978945,0.15827381093217063,
 0.4984959138017666,0.4792531651549804,0.16021890896701219,0.7221480202444766,
 0.5602087781042454,0.3105311940865183,0.331183450910971,0.2078487908553459,
 0.8472210668711945,0.9304330955037825,0.04820557060623665,
 0.47228264562172395,0.6939296561682933,0.5590516971018794,
 0.40289927696462935,0.9114452319763142,0.4816083121814463,
 0.17163712270916687,0.9082804613517885,0.021465049488918697,
 0.28019690432250177,0.34623963792063817,0.7311402435134385,
 0.9375614050231453,0.8451562140044215,0.24522247357660842,
 0.06313122787288916,0.00937740634123696,0.6787840694826826,
 0.5309222643803508,0.9947836741667346,0.7442066208029523,0.7100973330306759,
 0.6051157156225907,0.2064138630162945,0.3648327569060833,0.8043476813094556,
 0.9761104605362262,0.35952539360554325,0.4379752247634622,0.2728771162331203,
 0.1315662936203046,0.5504632810051657,0.4079443377709574,0.8633387128768875,
 0.2844050784683576,0.2794862615571826,0.2750713526189934,0.5116608795589471,
 0.0777002994195164,0.24246295490726355,0.07920122587302836,0.889076275059509,
 0.37574010615714837,0.7639595909990898,0.06081174676190726,
 0.4016871327699404,0.22202655672913119,0.4939802700724708,0.2157216848944804,
 0.3462501962622082,0.09984672790958282,0.524402869008254,0.1491813319178732,
 0.33247914964563297,0.4563492428621929,0.5815161694075458,0.8785899922671202,
 0.6460591322205836,0.5750751768526917,0.8973099410100897,0.5026404550035685,
 0.7112041556321406,0.7350157151198625,0.05720613680570363,
 0.03049946434852202,0.22130778394612527,0.3566945948605693,
 0.5529016513344182,0.8297701446447757,0.17575675167833937,0.6915996680189965,
 0.4402295934053745,0.5158475270852754,0.8152927485362325,0.4122962779088689,
 0.033243635120505544,0.43044008131467204,0.20312803449157513,
 0.41176275367632476,0.02783343762995949,0.4312849369905626,
 0.21621775198560922,0.4085428517232025,0.09460861518728336,
 0.3128979633080864,0.020003346276413203,0.6182656713390111,
 0.4604854387057018,0.6037992098788395,0.4107005195001652,0.3411918311766028,
 0.9991636723927236,0.7154120773183766,0.03594379947629944,0.1282945455353205,
 0.2918842063353425,0.7997027953999173,0.7669486027404728,0.7235303980191873,
 0.7935210025785998,0.49115605148580366,0.8863860216468796,0.4772059943364617,
 0.14987331542625104,0.5407297845894268,0.3349062673427361,0.5086995289936039,
 0.1929483890797843,0.3305336363669752,0.8630761719444869,0.8753210462986378,
 0.770814333010074,0.4720296876948745,0.24499324862938276,0.5795333826160531,
 0.6518689857413125,0.1884734701844356,0.6889803363124488,0.8269096875978518,
 0.5293764259394804,0.05510574381873945,0.8518261758972624,0.849075152301953,
 0.4973525731857834,0.2511180980737816,0.9725976844394614,0.8298918510654855,
 0.11018390207735751,0.12374606112757891,0.8418891109345541,
 0.4815599440161724,0.8325430010008733,0.10107533195450469,0.6577776791971109,
 0.5480569045950078,0.4926369266303664,0.4591867864100916,
 0.036363196606744586,0.48993937912967467,0.48715697624758414,
 0.40928173818995917,0.9707350662638962,0.8145946585417659,0.0861529363959681,
 0.5372452804588713,0.7551005350171015,0.03329637828157439,
 0.18558975991792426,0.45985295107589896,0.6598627166570666,
 0.8890139096184286,0.4258411060047258,0.9077124919978592,0.6506994698775803,
 0.3854872569036485,0.6303096002250438,0.5388568286778945,0.747373171507295,
 0.964578740085223,0.7382552741715559,0.49290351019490874,0.2947712598606216,
 0.33105221277928765,0.8595394785930464,0.8991524490224192,
 0.12380704266948483,0.18212935448835832,0.06615096051290559,
 0.10091213421246958,0.02607157700029905,0.7226062144339302,
 0.6382797449419502,0.611808876818814,0.34338495146016834,0.32986528077441524,
 0.6882490493662561,0.2839920100106792,0.4961043324705354,0.22360464741009067,
 0.11418483058824991,0.5253709516649356,0.24673426552122568,
 0.11395161444837132,0.9852895192429081,0.6176443487323577,0.84277790722105,
 0.8472381828200984,0.8358362507844955,0.64773738149398,0.39946511899608483,
 0.7952105728936515,0.22608109628429285,0.8733855538051748,
 0.40900947460924386,0.20414027473340302,0.17282483403347437,
 0.09057919988305596,0.5781157239668409,0.5298794634474502,
 0.030102826947423833,0.4939253793474857,0.7562679714108451,
 0.6719682800879411,0.4361323322768377,0.5442191103465988,0.40091168342978156,
 0.8531649624419156,0.9639854886826966,0.06803139391121438,
 0.25021001723326564,0.5299428065155026,0.40877545235415,0.22888907864585306,
 0.009654393093461167,0.8433043115758523,0.7437337213824013,
 0.5333007571343659,0.4160621828610108,0.27613874000979644,
 0.14566283826320203,0.4702949545265569,0.6756928795558345,0.8019137726621091,
 0.257065193053192,0.47948339716366195,0.6358609358354591,0.22663319832564877,
 0.7593328217110842,0.5574197203605332,0.5078763512339709,0.5434037315419749,
 0.9172722980869136,0.9585334509220051,0.8836376926675735,0.32613801692648636,
 0.3636997676622541,0.17157075924846887,0.6401875939726391,
 0.19727363488548677,0.050983746442262445,0.8126898451114519,
 0.3793609414209901,0.1973519473007701,0.2088760764048092,0.9036748991864799,
 0.5400251758561495,0.6272887644095276,0.5812103281065052,0.09191415154725435,
 0.46159733993581764,0.740475854542354,0.9298944017033296,0.9530994961674373,
 0.043451145563492166,0.9427930225555112,0.011834196749631154,
 0.027120848363499928,0.8421440603310801,0.48175536477005776,
 0.4213065385463898,0.3183230219575015,0.055753659438079084,
 0.8907740835327628,0.11249464794451824,0.23252657628544804,
 0.3145491313902119,0.6400842594289506,0.3077653961444977,0.09963672938151291,
 0.1383915739595054,0.48029453921752807,0.24025698817406094,
 0.27319649716161243,0.8364447954363056,0.834006991506935,0.12554469646486444,
 0.42449455683183857,0.20572849640952473,0.8777169559883015,
 0.12576217739034556,0.7595894351274625,0.6477923510513486,
 0.30114253052364925,0.1880850679937227,0.22433202214703196,
 0.4622442549790877,0.9656778762231948,0.4450440269822057,0.24018225851351915,
 0.409778088769273,0.5284935926252251,0.344723851784384,0.37592365138828476,
 0.5759738757715974,0.484765369545644,0.32495642352649745,0.17348305877067094,
 0.7455192975070025,0.3659134236748681,0.2610073776595758,0.7919993500964241,
 0.8328513082476036,0.30728678253130215,0.7790179615083973,0.7774701918895484,
 0.34758425629895706,0.23713074440565396,0.18226973239374322,
 0.6838998508385071,0.1649979187537588,0.4881792525275648,0.544556269804406,
 0.6165058001903533,0.6574409086702466,0.19572108839569624,0.8295213518449372,
 0.7724556584662876,0.8524331571707647,0.5758469704676815,0.8439523322924195,
 0.9222521514025145,0.507034225630083,0.23881514631907264,0.08580403849086626,
 0.4644567373617786,0.20724540442523787,0.9912410234201245,0.5308219086592829,
 0.6263683477804972,0.39900925310536595,0.7240969614763771,0.4123588946353812,
 0.9100382608073693,0.4796307853501557,0.08304492019072973,
 0.21911021782943751,0.12738027901882387,0.4778965550515979,
 0.26871104283359637,0.5558993797075488,0.1519103529314415,0.9722291332816796,
 0.9851740416157089,0.27713957833478,0.19604834820211192,0.6005658151655252,
 0.21309135949966684,0.9535477417296355,0.9965596900608872,0.7968819643099119,
 0.4591125087378666,0.46976237507701,0.21402251200173683,0.9323972287624064,
 0.6667885903635505,0.5265645547485248,0.7106975574224296,0.9381831060328492,
 0.021639367589827985,0.8428616561446141,0.2794544475602251,
 0.1942134590000676,0.7162424105826483,0.6536046100329359,0.9600763221422258,
 0.6836992056110641,0.6161204879389565,0.7050231057352974,0.2967920589394617,
 0.20291833483259947,0.6355814416704754,0.014537088484840699,
 0.45153235264685865,0.6664523864022268,0.5070338776309792,0.6639654289407314,
 0.681326925066977,0.4836965022831612,0.8476968662293336,0.5857014308256996,
 0.12119121167897973,0.9170179515447936,0.08788454588905004,
 0.0978288274950877,0.20739718086227354,0.6934317365038323,0.6683028423864431,
 0.4552764888866859,0.19147750795357843,0.11636847786538185,
 0.9582379135894105,0.8439685679812023,0.9523343621662779,0.8392661754323559,
 0.501144862872139,0.524614874291451,0.5415658470192007,0.9861610174805541,
 0.04350160845159412,0.36009256386627086,0.28560287058929834,
 0.970053648353016,0.5103615563486539,0.971499788879789,0.5429497063823039,
 0.2898208205671041,0.07745014730098254,0.09875762415760103,
 0.6737199705178577,0.52894036074175,0.6663040576963494,0.17467243686661105,
 0.8777827452069871,0.683446160153493,0.7154688458725931,0.4387357217392215,
 0.49904005186919353,0.25561218845686584,0.1334950094820856,
 0.05027070759053742,0.32880849098828036,0.29239863342215466,
 0.3718512775058571,0.0065602049841688714,0.955793584964269,
 0.025984993928995337,0.42869586203405186,0.0609860669245782,
 0.6178071873579729,0.7074311227870088,0.6095675574618129,0.9914754842312906,
 0.33508433838363916,0.1204342619134563,0.8411733372303403,0.7675151418181617,
 0.143753422001208,0.7390866490900797,0.6913621775672139,0.5494097610333616,
 0.4527703798652313,0.4807965790180061,0.04303833590120121,
 0.07916578410133057,0.24365506374559198,0.5984834336324171,
 0.6312588352888625,0.6827734533900189,0.8520549300887469,0.6609239237891297,
 0.12310594759011162,0.9074023765714733,0.9977383714646214,0.6268390116999205,
 0.6910668591332969,0.0025570782899722033,0.0026749600323535683,
 0.8159297221897849,0.8855494605860039,0.9921131937115453,0.5059171564651256,
 0.5232813045155227,0.18345806720413593,0.9928005535837681,0.3934587564245269,
 0.6948433409711003,0.8849522993660813,0.18500726406110546,0.3066876785431345,
 0.6772713344575352,0.5017505594434113,0.436118291179602,0.03257755952210928,
 0.06637361314555179,0.9873863962114794,0.5346497822426709,0.6207283528721164,
 0.05706063855331878,0.6141447530030161,0.19257846131312806,
 0.12043892172955228,0.41403030055326984,0.43568912924443937,
 0.26744020959258075,0.2830209795035469,0.49076776739115047,
 0.20161176879082066,0.33150285144841973,0.2632766525294131,
 0.11387207992589121,0.5784799703938871,0.1861401953928915,0.3340254719137823,
 0.21218825769554406,0.9741696738515269,0.8848411982539813,
 0.17398635021283237,0.15077307592600087,0.010735158789736743,
 0.43301745025713023,0.835621051238759,0.1480748004630179,0.2504514276786005,
 0.24418004423754036,0.45126303364326115,0.42253333098189305,
 0.753190118564125,0.48948353879050865,0.6229066912595392,0.8841647855926028,
 0.4517959902039892,0.19769559816871085,0.42093368594692016,
 0.9280662570927294,0.4581043954455799,0.2576999093222849,0.7629092377115982,
 0.8790180688926462,0.8097849757379952,0.22034495569719204,0.7330665943835317,
 0.7394987976940611,0.8153215791449314,0.6103916926303257,0.6823725850384099,
 0.9141379624347516,0.15236798205347324,0.021335576369061582,
 0.9465942580930649,0.9374276553233286,0.3686602982760201,0.9830792469676094,
 0.615055552654335,0.23143910301570136,0.6087984564782619,0.9003900684666545,
 0.45912056470638785,0.8402162953309854,0.03869504373396215,
 0.26581520917933343,0.8568884078285973,0.13999306040302506,
 0.8973482579734209,0.3105055751594652,0.8910138272257587,0.6310407932935518,
 0.7343416211305205,0.679851384608771,0.3862897776102141,0.7832140775808437,
 0.7678553466541558,0.31594124831985315,0.6599281860723499,0.7580920039881286,
 0.4860891800736944,0.712367406920929,0.09347715037944837,0.8492081490103798,
 0.5568044434414731,0.8187787603055777,0.11024491337049036,0.403639385047205,
 0.2331757578062108,0.40456648279734,0.051418000122986585,0.45012178584487916,
 0.5639900213296465,0.3884430356590154,0.01126337110461062,0.5709667844363226,
 0.32150335969893207,0.5086189679031161,0.3283331967532058,0.796340648505592,
 0.7568121993764988,0.7802600438347689,0.6252658870005854,0.20060123360184057,
 0.3491682183940097,0.7644844092536588,0.2706836063622893,0.3665372220458052,
 0.9644964252873786,0.31678177746562364,0.7324392683399834,0.7516264234493119,
 0.21786578870146156,0.6006786010694983,0.06807793914698934,
 0.13674047588337634,0.8115918538682824,0.4279455999980333,0.9231081610460364,
 0.45918604182384226,0.5961433031576507,0.7609081117080375,0.7775412637989764,
 0.08490861524983562,0.8809847904037722,0.7383426571721186,
 0.47445656028798977,0.25170446037050576,0.020182005755637666,
 0.7315495552811285,0.22087134419482646,0.5245386405317659,0.9501128246872833,
 0.6836964108723685,0.9729440715065228,0.6107011224252986,0.714609860984815,
 0.13894481688173266,0.3350140453673588,0.616024030671572,0.3673713824235266,
 0.9244450620420772,0.975804987745928,0.022167935296579677,0.691804654941444,
 0.3873376783206186,0.6872583864604775,0.10795685872009042,0.615430395373252,
 0.16003817687175215,0.46900716609005866,0.1961303196980445,
 0.4080266901955206,0.8295902956810274,0.35400027013047697,0.5261443140682116,
 0.9651580153031698,0.27126950191303634,0.2084323866925416,
 0.15201086810544417,0.4107257367396716,0.5564086338924019,0.5058657345489754,
 0.044996972439050986,0.10910806391347794,0.3917963047479227,
 0.9658696455163489,0.4104387871426126,0.2877364135351441,0.729378666747076,
 0.18598202505175465,0.35893988249331077,0.23197326819135133,
 0.5430367714263982,0.0321724083251943,0.16701655215524724,0.7698689683245457,
 0.2671981277314419,0.37947748758883226,0.3925385573393503,
 0.28175239345483993,0.6427569920076446,0.462477883261438,0.23058373085300765,
 0.8648643098075268,0.7017285657778415,0.9002021328171729,0.8661929026301803,
 0.4804081660613295,0.9225510452558511,0.9098233591337297,0.5527863915624446,
 0.4719211711711633,0.01580994530268831,0.880956530004408,0.4181463123008986,
 0.21719726415702145,0.08245525397844722,0.42734013304679797,
 0.2737798808313694,0.2505122484462665,0.4680185701654811,0.1514868398394411,
 0.9365275640260948,0.8915580263140214,0.04500856468616621,0.3676309523062232,
 0.5811786323379848,0.7026177060542286,0.12950848882268207,0.7932691038731355,
 0.32647319307063116,0.16769767293132798,0.9026736872190505,
 0.17634122371715766,0.5389716877601183,0.15850132998722344,0.607268255092538,
 0.11859774315056293,0.24067774352867866,0.802997841532215,
 0.18097825893854758,0.2644386742212377,0.3875232715318534,0.081450298963468,
 0.16922496057187564,0.20039932744600542,0.3652923728065942,
 0.19883239297264677,0.34115848610776334,0.07565843318363341,
 0.937564885636758,0.10505726805346483,0.9125988716082124,0.7642452486432044,
 0.5673990489615257,0.0681329066037707,0.9445566856072037,0.47771976109171677,
 0.09621898902550963,0.3479835917965354,0.13208723395073352,
 0.37520322880571366,0.05140076910296165,0.4387356850275199,
 0.20027981330983025,0.8051925205326061,0.19056401395396394,
 0.23267488520191904,0.15683980864565839,0.46212557308771607,
 0.8183232982135374,0.3438711694262325,0.24237139363235216,
 0.48753041930314245,0.49032875001980036,0.7959358373325083,
 0.46118225471675867,0.34290839329081735,0.6441370713034503,
 0.11210484214175687,0.3312634629066441,0.5613865377196251,0.6956472211968636,
 0.4022396990654815,0.5578695238682552,0.5268047078442024,
 0.030156987039068772,0.9847616751621558,0.08878955157527024,
 0.26840676872449487,0.8447915749707064,0.22943469130884886,
 0.8940255935803991,0.5613802032450689,0.06202400456612056,
 0.36536371959532477,0.9156508826604431,0.40412207873685435,
 0.36103514756235244,0.08721007216588295,0.9440475141561783,
 0.8759286462054603,0.614135162182456,0.9620925861430503,0.7181567467692078,
 0.007011991023690989,0.6309190632812797,0.10107555056661133,
 0.3280805806440821,0.07484214993315885,0.832283105927098,0.7796355279836841,
 0.18116044136652132,0.31916575654902146,0.3648304305012794,
 0.7057182623512328,0.2851537194143041,0.40926135398301056,0.256961663110545,
 0.2812574128555426,0.36091044800475147,0.42181738148648995,
 0.6349363754646615,0.9509408393785407,0.12417029018278236,0.7065057379275366,
 0.7000292988713309,0.33833339224394,0.6251556385251916,0.4917081462026389,
 0.8168109077606027,0.8463302885358295,0.6829357810497608,0.20452702819210977,
 0.44539161626047896,0.0021000616429861507,0.5044339927573915,
 0.9829667821504618,0.7142196021322607,0.7545111018949777,0.14176864637006825,
 0.47369133676456343,0.991615101120883,0.2423864825769949,0.6208667497327163,
 0.07652624226811755,0.11003891408513444,0.8303783186333717,
 0.17191106236778053,0.07084240134037534,0.18334211057470218,
 0.22267365096682668,0.23093879024027852,0.5351527169801565,
 0.7558558447089873,0.6739253459094026,0.3019158034520718,0.8030066936280272,
 0.7041003958852597,0.17808009697677396,0.6308755140465453,0.8069398514268158,
 0.8523187387986481,0.5799311258302574,0.3223716097745434,0.7259875179719932,
 0.10180793200470839,0.1308829072039892,0.759690754465294,
 0.029556201989733855,0.24832270938918866,0.5538468233615839,
 0.8358111612557078,0.2619118298282397,0.14376995988486774,0.8681652741226811,
 0.597428052193014,0.49113435434744934,0.37672235047884706,0.9008392043183884,
 0.2777505097711066,0.40760145780087964,0.47525087361765506,
 0.18175420299928502,0.7123021095661208,0.8318194917872183,0.8024094207290944,
 0.13979459865003108,0.20261338493823122,0.707562623261528,
 0.43473297172281355,0.8377706793641866,0.07385178425599115,
 0.6400701328729722,0.34711603949920145,0.14007133277766415,
 0.3241578846026374,0.514559232032343,0.9522061905302588,0.10524422003669609,
 0.31483027745635295,0.4974935995786284,0.15962408340421308,
 0.8403875497333734,0.30105733310934646,0.32956127660116696,
 0.8086342416047586,0.4350478154527916,0.012595091768826006,
 0.020798160862625203,0.8872276226715737,0.12252431862934143,
 0.5797668137683947,0.1283292039146342,0.4831920982058563,0.08453459792735607,
 0.4285071338512282,0.6625644660043954,0.426157837325145,0.5896012645156187,
 0.6084262105811153,0.7948938992233066,0.3067505195157534,0.9870173213724548,
 0.5181149321595132,0.9207680232983046,0.7999861985014889,0.4915388490053616,
 0.07023549334533485,0.748492269483327,0.050460609943644985,
 0.27721743913966346,0.9286836373598869,0.5921211285389364,
 0.046930390429036795,0.07330526247181601,0.23667109679705656,
 0.13362060149079535,0.8041985547892399,0.8590317594684477,0.7136792419279502,
 0.45551674936476305,0.8652658021989273,0.9310983569755987,
 0.28506540729890506,0.9877083635941071,0.544448545530514,0.2638428996382802,
 0.4576903090342662,0.3871184879541878,0.10410733988010623,0.6272577865054677,
 0.4564674130925096,0.18834297419289514,0.07064785459487921,
 0.9229992914907574,0.10733478360855064,0.30043695467064724,
 0.6530203016364214,0.0035558066274523625,0.5837671476466595,
 0.43790774343801186,0.12660489289712906,0.9030999778477558,
 0.5239390107157745,0.873876758663374,0.937150243596235,0.3461607198132145,
 0.29467654462584225,0.9466662556610308,0.32553768020260265,
 0.29667503996770894,0.46358660561390397,0.86505036431149,0.939012789226178,
 0.4418003696193302,0.9635155574860307,0.8100634927922156,0.14094394433974522,
 0.05325373019580115,0.6014707945455746,0.19852452401375298,
 0.9486126543419039,0.4008808677223208,0.5472227324922785,0.1491958418827034,
 0.8332489828958694,0.8868364349847558,0.3399224737979688,0.7828121987643231,
 0.13926537355083413,0.6871743536156206,0.39637703882379316,
 0.5769197430786531,0.919721257715964,0.419646918244842,0.5707808117766277,
 0.5231483333152835,0.5977138384111279,0.06464157710652552,
 0.23949006423169594,0.1515139663134123,0.508115600698998,0.25111074749190554,
 0.44575734607428796,0.5989315829187875,0.8244456352440385,
 0.02275565198687579,0.7487868654760607,0.44434820005652265,0.927632591133659,
 0.5704371354548011,0.7433975113115626,0.9709491501296329,0.12797344175794234,
 0.07750004202421734,0.499261721630896,0.9793575659236865,0.4043295857298914,
 0.5420328041515907,0.06295746440547001,0.5083536095818401,0.5354050408985169,
 0.7936498466506491,0.473390026875921,0.5289566857071601,0.3069491768184708,
 0.10874467178815417,0.029304387026449663,0.10735631540762336,
 0.24407510474278205,0.20092706408548655,0.27034796075970136,
 0.7441795788785043,0.3288538102495049,0.864842947182828,0.46800968856679426,
 0.19458473088612982,0.1429542143590644,0.38918971627987886,
 0.01653855536949722,0.32825757003370426,0.498235881518723,0.8204942184923399,
 0.6640694837410384,0.6651931752060342,0.8127009968499117,0.3686322811821905,
 0.415602496065672,0.4439741503322534,0.860204223901096,0.544268850010209,
 0.6840204717398033,0.48159412027023507,0.915980542410071,0.33968464485185046,
 0.016922541400605118,0.8215463042906761,0.4033492251546209,
 0.9702934377046094,0.6244108427571617,0.14257272729928439,
 0.11992072140546917,0.027440172539822005,0.17720625872084483,
 0.9146842707331937,0.8504824482058349,0.2963259770580924,0.6173771274959183,
 0.4410711063530527,0.7117982274420707,0.11667663323305144,0.6395210266404183,
 0.25199983176928065,0.8614904060316075,0.2280911721434825,0.6510020759093312,
 0.16087677862206096,0.5787967550109544,0.636611695356913,0.7394330417429108,
 0.039853748409759526,0.13009607596324424,0.6859500490177887,
 0.7054103642530467,0.34506855790058044,0.6741164604632245,
 0.22089792506240158,0.2135696145474315,0.97776363376576,0.4394093486679316,
 0.2572777487057337,0.1117795773101119,0.5410997364368664,0.1835036746910822,
 0.043928814845335507,0.3973892787115372,0.18149903850104132,
 0.7134675399014528,0.14498057215705773,0.9682068891073471,0.6976297895384208,
 0.5225668087895586,0.2709961715539191,0.9314577063433461,0.07700249422105288,
 0.44643390731305554,0.4766606761632938,0.15051583934123836,
 0.7016748840015915,0.1237141023613445,0.18186034728529687,
 0.42019002779645265,0.11773343587818164,0.2538404629316793,
 0.6471139439588159,0.658762185547291,0.2276534536715905,0.7164329286945619,
 0.1789770512674086,0.7114477154790783,0.342915409515271,0.856929386601731,
 0.16668182027716738,0.3675438091429427,0.7047775561776124,0.7011975669136437,
 0.8998119266726174,0.4331246839334745,0.9458999279463718,0.6151745810984588,
 0.6176121870476798,0.12417721114047564,0.1994616832600714,0.9201932899564147,
 0.19259239679674012,0.4469531218201771,0.4731875144554575,
 0.36758892580552227,0.6501306568173341,0.6396250083945054,
 0.05348241195059966,0.39976538027910047,0.712168059187417,
 0.058704315404117025,0.7486766001439153,0.26391236002694984,
 0.8827078990694085,0.9975827370156145,0.9719453542348324,0.325672628120515,
 0.7498376663628403,0.6993328503821088,0.8299985933877887,0.6191015023979906,
 0.3296562294452412,0.7544433612202907,0.9027803772133087,0.7454891528278285,
 0.8741477656931687,0.33869029030236186,0.8818880651761432,0.4634329316901078,
 0.792836846053911,0.6624489738777266,0.33933664774843986,0.30679065103728786,
 0.7747890428126526,0.674652734341419,0.45882913790482094,0.9236964061089034,
 0.39478940115442807,0.6320216311761684,0.7369615089191686,0.5986301538351978,
 0.1716249340255872,0.30410879689328674,0.4528660297935414,0.6973070587664163,
 0.04800979335635547,0.003055794970135195,0.7419107789526391,
 0.29480628858026026,0.3540640475330379,0.7389988187339764,0.4176871108503205,
 0.8282177559524596,0.5449923119225509,0.6893009425540543,0.9879834260066758,
 0.19590057362811675,0.05373276496434398,0.09421625241813747,
 0.8036854116716317,0.6152254068388916,0.0548508861618876,0.5087183610510404,
 0.01979687924454887,0.020237802074170252,0.6415306748384432,
 0.09199653580547518,0.32628010556594433,0.7945516104888996,
 0.19939659052684022,0.6057450553631896,0.5901926222257003,
 0.18182861418440188,0.13557373730709854,0.8131970003354347,
 0.3050754631705215,0.3452741307593489,0.8093930806807645,0.379801889639344,
 0.9450039971547316,0.9880698267237391,0.7410951275929173,0.1521713415575262,
 0.774864569418312,0.2849191968027939,0.6069149163297581,0.8766275348858372,
 0.10532611352472365,0.6836161781103051,0.41683673005948574,
 0.7894998321025142,0.5022234983607803,0.3142797436116178,0.3903969162779516,
 0.23394817166718251,0.22299786628674756,0.5508037333012923,
 0.8004471688183159,0.10077215525965932,0.6056911594604893,
 0.09455151840709508,0.29967707501529384,0.596113992933452,
 0.19514827716956995,0.27576060044243433,0.3271245638576378,
 0.42381113052193586,0.7880508365776507,0.11906307390562154,0.90295512085823,
 0.06213947566648248,0.926929142029075,0.969638461396945,0.39485411330722975,
 0.8460493527276481,0.8757517335892233,0.23137295537636304,
 0.17690646596449322,0.34294614072271257,0.6148156326080604,0.640645203730565,
 0.6241716456761496,0.8789958704051677,0.6987486563141788,
 0.017248085597666996,0.43262572801233035,0.21992215764301215,
 0.3243306624277901,0.29637578548794563,0.8560196223738732,
 0.006348422628973727,0.7701059067437299,0.243383256803975,
 0.06745341639020075,0.7235949665404653,0.82545598746856,0.011267939217441847,
 0.5211892828953889,0.9231442100570453,0.32037955453272526,0.7417090806019517,
 0.15975907262870304,0.34699738681720305,0.8124828915792299,
 0.2252169992798556,0.11820657598466877,0.9851382594723787,0.7125948381622452,
 0.510366765416818,0.20081116451235992,0.7480705366513473,0.9838016555067987,
 0.35252145359279385,0.0810029601944171,0.8311338154653934,0.5427054157433013,
 0.4572417318838988,0.5951582218200205,0.6011766053295048,0.3006232293015928,
 0.6544126576683349,0.7538804609483052,0.43976884869518096,0.9091786348384842,
 0.2708086891115853,0.15343866713278909,0.6893015582191571,0.9619159608500888,
 0.9278308763282592,0.8302136546194907,0.2546360263667391,0.29146332862950286,
 0.5248736392674445,0.5762997458957632,0.3607598383758278,0.732344060174622,
 0.5970774531119983,0.9829684491955418,0.28072421466647557,
 0.07823174009269807,0.00185342335441685,0.3007435925438485,
 0.8058617913150127,0.27304235191870707,0.8418096977750513,0.3667324863037552,
 0.6585590295508196,0.135185103486694,0.46580858210556997,0.46211299398974726,
 0.2694287195524414,0.6836810317951555,0.8644707295025685,0.11310396887521568,
 0.16174014890131794,0.9713703577013943,0.3889427630255806,
 0.30115825361901705,0.9544963980247715,0.5662257688884436,0.2191619637844815,
 0.6358048232033049,0.7598594312770106,0.8229889235882715,0.8248254799257534,
 0.905743224015415,0.8991071267771242,0.25882495328775346,0.8972946365860371,
 0.7275397568946563,0.6852415342899382,0.04238386789285831,0.4297670627212138,
 0.8356664350261228,0.19136499130302764,0.4567878398443711,0.5122581849233093,
 0.8169745375115123,0.9674950615225418,0.7830780324163762,0.6651002709964714,
 0.4932684722967918,0.26766768375945915,0.2523270818687591,
 0.18963803939892065,0.6245995030160982,0.3046381284883478,0.7529554451842032,
 0.5533107847509819,0.04546868927173631,0.8088709210321203,0.5716313588953026,
 0.5281953112880471,0.8874533149117165,0.36119702923515484,0.7294895821096284,
 0.74792309733389,0.18119084471407287,0.6885677903060872,0.5532158746860207,
 0.6253334512040571,0.083368694535494,0.09570694680218139,0.40584186251538656,
 0.3735157141244698,0.6590217977515249,0.5096245267680688,0.8956571492036789,
 0.6916086946814046,0.23803606793574206,0.12080391837474891,
 0.7641721349397179,0.7702150189780753,0.3384886537872578,
 0.028623835132072784,0.9563181709805301,0.5621304904050867,
 0.6451632208800627,0.16734154376823396,0.9624981654959398,0.9786046749259911,
 0.0037988165490214776,0.2635385117209984,0.25539877237934117,
 0.6757505591355063,0.9664051354996547,0.10155952191470496,0.8482973239757903,
 0.1046640467433897,0.9930697259392592,0.2926791164475675,0.10165083109519168,
 0.7486087513721975,0.5153405934658928,0.19254187956052415,0.9230249725020981,
 0.5160903511583537,0.7952899430019001,0.9873164953050252,0.9201737608185938,
 0.6833157348933032,0.3666787195461658,0.12119113594423947,
 0.21662031963933837,0.6976711083637122,0.263872224898526,0.8997672191083821,
 0.033758256865317526,0.6465880853778717,0.6208255006980182,
 0.9204189147976993,0.2757454130111201,0.09865597566477402,0.729264640647726,
 0.6785148163552462,0.48663611749377556,0.4002013270375421,0.8664066048816935,
 0.5609854501734716,0.604593004885851,0.9574057433807053,0.5472714929317244,
 0.4313628229483427,0.6775204715963552,0.42171117544279824,
 0.04524601852384558,0.3125790217184694,0.8729933648344083,0.4283762687049085,
 0.5367324938527922,0.10092082278498149,0.04933404445163847,
 0.5342894248265354,0.19064876895621374,0.9330700585584335,
 0.16069025111776858,0.7043152262773472,0.24634543833885802,
 0.031505215921487584,0.7226864272427028,0.7001570130675471,
 0.2470545820498723,0.9810293323800512,0.9004002242532086,0.7892054490993721,
 0.2750652179325196,0.8733948621745663,0.9369692625400515,0.3005771599778124,
 0.6041078244911963,0.8613330404036682,0.8910880800555422,0.2348770354565941,
 0.0725213026864423,0.6394696193845574,0.1686266442690072,0.338509329829665,
 0.5056351266189258,0.10106156604716843,0.24871118290222416,
 0.7676986918946673,0.0628788727019326,0.4055379313717621,0.39341223147350024,
 0.4849380285464493,0.5417679425652039,0.8978306592999891,0.27601057975168397,
 0.7939564810076242,0.7440539402355146,0.0828767957189428,0.6612181878485409,
 0.8294193213004726,0.4141194363755578,0.24858783296623443,
 0.09168882274620294,0.9099454537589586,0.20074713893636154,
 0.24085728610239987,0.28253054870716043,0.35715273057113417,
 0.3230556679838914,0.8190914359991261,0.4910317077424826,0.44435386098757756,
 0.11900830457938127,0.6757388631559819,0.8793634491703081,0.5709074853781992,
 0.5211679568019122,0.38544959673894574,0.43287913525073907,
 0.40499135343505843,0.6845848279594149,0.34941203703669255,
 0.6025028468271502,0.09774380678450045,0.5121003654621633,0.6659977406902334,
 0.17409785375752873,0.4767615688237229,0.07633002371025432,
 0.6989799990804578,0.5271061656357124,0.4325274531908999,0.21639958573305895,
 0.1707265138421099,0.013834083862203861,0.6169659325405924,
 0.8470422914428595,0.48643699633882587,0.3886857884885172,
 0.14783569367345795,0.4878822042272041,0.3444911772693714,0.979653377759699,
 0.6386747506581628,0.9655571541554777,0.4195693324607458,
 0.045647834679844124,0.2688408609623175,0.5781412808904185,
 0.8512187755741973,0.10571840944158473,0.8388398038548767,0.5859441992737888,
 0.8433429284352342,0.2908863786272974,0.6153970897539787,0.18749081733423012,
 0.5910501198924845,0.0753182267731024,0.30103010424073573,0.657545726844495,
 0.35433018972233776,0.924124696563839,0.27260990122802875,
 0.39220865582311504,0.9921315282556654,0.48163366835338384,
 0.13724385816239892,0.37231136278247,0.6965999303924088,0.08785089451153383,
 0.07587858361676236,0.7568904238697972,0.41177038328257787,
 0.21522612004665354,0.3380250921190955,0.6692126031127187,0.2200722839725303,
 0.42990112192661256,0.06148293510363867,0.7536331373414352,0.988770913246637,
 0.400725288754629,0.18438524184300464,0.46704727099556176,0.8947384185057888,
 0.4438555931149474,0.3762267709627468,0.998543296350467,0.28815540325225664,
 0.24178974242006124,0.22599693239373952,0.04201621930385335,
 0.6741309715784144,0.4807232320284266,0.3675552921026942,0.5049562149862532,
 0.22920897964687958,0.4554644183032509,0.24410106398726583,
 0.2526469152426789,0.5107751373049277,0.03582750061435758,
 0.14658508229551548,0.11833489185177704,0.4183052134485886,
 0.38924804494498144,0.6776579944472242,0.36204563327952377,
 0.4440930323889767,0.5071373795100297,0.7870760561713894,0.3639943836894023,
 0.7628761592170654,0.8737820626459432,0.9112214073775109,0.40289253205457387,
 0.7545775292255851,0.7162075173791568,0.2833698477981196,0.9035255479574039,
 0.8907525177492654,0.38069812690814,0.9696583981807515,0.8759911126441694,
 0.22745627462604145,0.44463739390347035,0.46065567810256614,
 0.19329190345246172,0.8384889499469601,0.02115131729586639,
 0.5296438916826133,0.019223420930259527,0.8395272776573043,
 0.9085122541097347,0.4489310890353485,0.64475735654679,0.21247144807876261,
 0.509441530952735,0.25297478990714894,0.23593228520249776,0.8473238547152551,
 0.44694546034153426,0.6606248400073192,0.6298792328243324,0.7786608517203807,
 0.6466622464961118,0.5940197337560249,0.6440315612830799,0.11619520033115871,
 0.6677483886221224,0.4757004460098838,0.2915532950223795,0.1031115150619707,
 0.8895252646929156,0.6055186233602512,0.5405792555058676,0.4077449589909712,
 0.4976333029932164,0.04094709428544541,0.3068823994828379,0.9669516888102372,
 0.19238514366528814,0.7487635059982373,0.1945202484451709,0.7706906823515087,
 0.17580466856465737,0.18825150369008647,0.06966815991061304,
 0.8943944379193842,0.6024196769436018,0.8099827835652159,0.5208366810772894,
 0.3187868599847191,0.6442729549005454,0.8774375993062248,0.524969268447118,
 0.34327358509913364,0.3417405266489725,0.13632850291367382,
 0.09814277937065574,0.8363737369604034,0.7770554587148644,0.845105762168411,
 0.3574744387954245,0.48370334182790276,0.8949159346672283,
 0.35099387564740236,0.7597825302674708,0.09561649752397994,
 0.7944172907097544,0.6885790247277133,0.06549623390151227,0.4357572347564933,
 0.5341115057507322,0.27496323949280366,0.5631858773548378,0.4925257857646548,
 0.36555446985574536,0.6704589470821801,0.46180614428175837,
 0.27032543535738285,0.34987154799921927,0.19533437277928023,
 0.8902769606143582,0.6455166746312924,0.8862488730002601,0.44196996395307275,
 0.646438155419116,0.18609670956025803,0.7430372904804273,0.13172626900284123,
 0.41258612759137425,0.05541910475378953,0.5266278840551479,
 0.13124009638170464,0.13070833097778956,0.19981573395615737,
 0.1797065200332134,0.25755294869421086,0.33873371542636144,
 0.5948596742537263,0.9744385292509659,0.9608945254811438,0.6668366301137146,
 0.9750989801749436,0.33020905905487363,0.4348007232734987,
 0.35310231076209186,0.3473998403238091,0.5581996223733314,
 0.23845796164022337,0.17138624861186302,0.7830356945228095,
 0.40212725743668876,0.8061710806346633,0.3808105234085216,0.6887794511646235,
 0.2222495359239116,0.9612406151062554,0.3402170746377515,0.02514766630856724,
 0.3068694168146455,0.7534225135917081,0.74347635164537,0.16870283703089162,
 0.9747705884822176,0.9495806643756538,0.7902156064177931,0.48596930007059136,
 0.9805897164884032,0.04369709377540254,0.5513058966857297,
 0.26821129690734613,0.8689007208456863,0.12588646700439576,
 0.07003129430127208,0.14379830377081915,0.1614743056080511,
 0.8815156488223134,0.8899905278532165,0.8418738432038618,0.5311926174080076,
 0.7672217815621412,0.6685876125406667,0.3852351694643018,0.49749555674991397,
 0.40266901130391497,0.03986263752554087,0.29623841017810626,
 0.6770059877831908,0.7264716445727876,0.3077377052139949,0.08639247362000457,
 0.9343207287096831,0.24890621558119985,0.0611296890240417,
 0.11931738990411389,0.14407148932432967,0.6041969490787557,
 0.6448560845667232,0.7010807038392863,0.9811552541107881,0.7120867966808132,
 0.3266629601763137,0.15776741634567848,0.00216258016957116,
 0.7851981738747676,0.35683825145886805,0.2720699810004489,0.9191682219321372,
 0.7964140610096617,0.5294666169295142,0.6555814142646466,0.6361549976872396,
 0.026679250843340774,0.7779902767375655,0.6034631476623263,
 0.10803490558500595,0.6030927548532162,0.6150718679020494,0.5166251410580759,
 0.13367849652369834,0.8384470053095632,0.7116299353248071,0.8312695624470823,
 0.1326591275944855,0.20272156784698414,0.42879323189275864,
 0.5418464145257021,0.28755930137911545,0.5355674012506377,0.3025115896930143,
 0.7811088135616399,0.28998634271542,0.7767852662922213,0.2767014192274335,
 0.7921452437161149,0.005435326767185611,0.996428758294634,0.6829068390260823,
 0.41996306445284537,0.12773834547195884,0.656549938973339,0.9798725509909827,
 0.09866040967903011,0.13809756920485183,0.6878915130118767,
 0.8920590532576927,0.5836170525511677,0.43191577535322623,0.6047229125822454,
 0.47453851217959464,0.7914039702023618,0.9672662042565066,0.34175034588622,
 0.5003784573162708,0.2461865994103445,0.3097526630722285,0.7919176450950819,
 0.20307907274400094,0.7186777300405081,0.10897494304305011,
 0.6388992981674813,0.49158592572245563,0.13555533942209352,
 0.5600559200148136,0.61739385002302,0.19616830589326661,0.13400399654453,
 0.4597272818968068,0.2196241887230398,0.39817669825874935,
 0.15635474069849142,0.19162474638129479,0.9573949115364878,
 0.1962432500239385,0.6467817028858326,0.7348886464577763,0.7161015915904664,
 0.3189785875408072,0.6268804987614162,0.9600500017965428,0.04175284341657015,
 0.9115586352894373,0.7894403791286532,0.16165138264170176,0.4480457280708088,
 0.6362509936856496,0.30007001877115824,0.8505686937059808,0.9561484848925668,
 0.36580581161285197,0.9823122125654008,0.6197389309306187,
 0.30663178331483365,0.9638936547606617,0.9771484636316271,0.4546083141924414,
 0.5512843064907276,0.10163996626794158,0.23529942103802715,
 0.12860282546194057,0.5487763184250799,0.6925954980804482,0.9178542757389654,
 0.4751888714350143,0.04404225407132334,0.26734476521117934,
 0.29594066174159395,0.6435262287565084,0.9655567819694519,0.9247015445525819,
 0.6115156505061702,0.634992740710113,0.3039002762737012,0.18624337192206997,
 0.05142887012867814,0.15785700462203167,0.565260463454242,0.7665771205555233,
 0.0874862651859254,0.6467554453218123,0.8546030295729823,0.4451109745692369,
 0.9973162272836557,0.5112772101972016,0.8899255930269945,0.6742055830379684,
 0.5334448104795765,0.47643896877360825,0.4430063971604731,0.890860819160374,
 0.9897034156887474,0.677465675573617,0.8790368894637294,0.9435072278598112,
 0.08507051663620024,0.4131968417663877,0.17732994603058927,
 0.36678484893566476,0.442495412002735,0.8997831356236744,
 0.006604229041073606,0.5123154076210673,0.23930414011654988,
 0.7580189479422375,0.5123767745443569,0.5914751202630463,
 0.004028576478423362,0.49250933607702596,0.082372210488598,
 0.9494465028976459,0.6461284152282437,0.7042584561802752,0.6076058472121714,
 0.4527575985782094,0.5090159777322831,0.7539179669825026,0.5431309227905449,
 0.6746151814314552,0.028404002528969796,0.1257733108137632,
 0.056978466361110036,0.4599373386892568,0.981820547034237,
 0.07196415274547285,0.22464645715426745,0.7377519419802858,
 0.8493655902078452,0.29503823001408835,0.8957467250661358,0.615532650313162,
 0.9913961129084949,0.8155743925805956,0.20057347426985028,0.4741335037314559,
 0.16329086724626352,0.6362764618100263,0.2202032561706293,0.9905057928875787,
 0.4925837104226777,0.11706197593002043,0.20641866517564844,
 0.5606732675977302,0.6150123411822757,0.21430872909019127,0.6550759246827956,
 0.6888831568917121,0.5972276775169408,0.9422482917142402,0.2535145910123484,
 0.667876879779911,0.3272594409398787,0.3425641198572472,0.8261267064897331,
 0.030029260320902695,0.43184645231543095,0.9789203552775976,
 0.3551165350948753,0.5292779305323404,0.23224014006957505,
 0.14878329139206736,0.7279482768161092,0.3478129305664046,0.382548079373638,
 0.6079980141980368,0.6437103306652391,0.7342787946330209,0.6527404041814238,
 0.7156417999346194,0.6634270845586676,0.517063546018248,0.21958764888148397,
 0.23166611283789063,0.943829673762013,0.703740742902642,0.012089576384790468,
 0.3865312734277805,0.4344743953441882,0.3783593250850479,0.7866364507468413,
 0.45911896780890005,0.06322679346215199,0.6438635905529188,
 0.4002716088025199,0.2620983299173739,0.1907665883687425,
 0.006384599278201897,0.8487106625776246,0.15916468355106927,
 0.1815263117899063,0.08174509859126933,0.8985976838903275,
 0.17699160359635446,0.5065087007680886,0.6036416373351354,0.9606453451979913,
 0.200131404702654,0.9143360279968578,0.11745681821678489,0.645317787621705,
 0.01792440144570484,0.6723574411614952,0.6542641492662953,0.3196674155245711,
 0.32196416325900346,0.7032251887251911,0.3969548080662224,0.5031783970725465,
 0.2061755242496568,0.6157677865993273,0.7479543129232572,0.5237314159190424,
 0.13562568988987056,0.7855476636496564,0.709531689295194,
 0.012553029430627527,0.35539489816067504,0.6316136434200915,
 0.14593450651706596,0.9596714986958562,0.3909375069654556,0.5692132675945748,
 0.2347823288587224,0.10661904288038615,0.6575236926107652,
 0.16662391574437074,0.33248112110338746,0.6142031069400612,
 0.5448341591030554,0.010518628653186646,0.6974921896566029,
 0.1523872353758824,0.9532263015832484,0.9513283247447585,0.5024026881000804,
 0.22718702706910943,0.02675393626056488,0.7531843658842892,
 0.38729456120764394,0.4880895769016037,0.16048107805747747,
 0.5651270620770751,0.995128627631996,0.2168055670394231,0.5892055494435329,
 0.050291684178987994,0.9004085275623035,0.3118883174340833,
 0.7323798092290005,0.6878405497726929,0.2592010183648472,0.6562101946828977,
 0.8482946755871003,0.24207941648443643,0.26988653624307046,0.726044695149025,
 0.17680461016652782,0.5492894660826382,0.5249032259949995,0.5928572347682682,
 0.563607379381405,0.24273777333100788,0.10421286279179032,0.6298215677235968,
 0.018922662499785248,0.5100971033691282,0.9158570797610505,
 0.39807833271406967,0.01066953944417881,0.4123009092473262,
 0.34272484672414394,0.4605738180311014,0.673490095559925,0.9536801012396585,
 0.9279229542830041,0.8817290233794793,0.5368785733561989,0.7322592417547145,
 0.05157629937048813,0.03712137332108412,0.08512450316209497,
 0.7392914338730558,0.2929411847167185,0.13700826602514504,0.7612039694308543,
 0.6411383841492545,0.5125118449794785,0.483166585124302,0.04432000551962112,
 0.4707145529657417,0.2950473593401589,0.6269585818133429,0.48411212007438564,
 0.6575416165592106,0.3240073890807639,0.9076627184403685,0.16100595413728713,
 0.1967135539020437,0.746091950566875,0.6397521297129976,0.38491569257596814,
 0.4830447295364726,0.3394975917388351,0.2743460844989597,0.8300731654766798,
 0.5183843360752924,0.5795806683341611,0.4208268998641821,0.6550787339780916,
 0.44557758596772523,0.6034508963544426,0.9816008729133546,0.3819785683094762,
 0.04828169149959305,0.8116395088070111,0.9036634748461077,0.832392944991009,
 0.8436461402265271,0.2609490460852113,0.9859959832449242,
 0.042825796737876676,0.7850334201311552,0.6526318980194213,
 0.05687341036243809,0.2429604482770542,0.6508736123516929,0.1824676286676249,
 0.8298811501245157,0.5061583997886646,0.7777122374102591,0.11042113627182415,
 0.5070087345783064,0.9238703886254305,0.47751029020161473,
 0.05071879196114404,0.3170422177307821,0.194884272298241,0.8851221170394135,
 0.9430234748754168,0.005326495194347469,0.6180131450304017,
 0.7591943246449622,0.5495365548038886,0.3181081301149101,0.02746757192370375,
 0.5448292543889338,0.9829031311906928,0.8337094452341702,0.562501618456696,
 0.41650143539447426,0.5711543512803517,0.5767334746937618,
 0.41811327663299924,0.949835705410325,0.9232983683843949,0.18448626166904902,
 0.6987765340866547,0.7981791064988257,0.6329194066378814,0.5755851184365495,
 0.6687568056132662,0.32893873674546514,0.9538564902858935,
 0.13244430705604893,0.5622043584042693,0.36336254218536534,
 0.025192086710988826,0.3530227923790845,0.6926181242206937,
 0.5290726147853381,0.612384874093075,0.7076633868334362,0.3956958541897102,
 0.5736005489043027,0.3423935908284085,0.9293932439697031,0.13233277339076233,
 0.12925290538065848,0.21521843334883028,0.4087731637890828,
 0.9556025850340907,0.114334836802093,0.8309459432264616,0.6224717413184283,
 0.26966513240950674,0.09664464883000279,0.7808044415347652,
 0.9353425053698223,0.3361394245223138,0.4814284667546792,0.8336017863704916,
 0.2006291415442456,0.34585320783190754,0.7767585300342685,
 0.11536317455958789,0.215228650458396,0.14952745252655786,0.1103674162194701,
 0.9956116137582309,0.5106431039323999,0.9616622307289429,0.5421900969148089,
 0.700950431902407,0.4747521708906379,0.4302282133701002,0.7825297579096202,
 0.03444446509619925,0.9092805050210888,0.2855972902709294,0.5070310260181051,
 0.7451178079488562,0.6554454394144201,0.8485554816587281,0.5207622297420804,
 0.5501265839876354,0.14362216825800633,0.9689705864126803,
 0.42986782652510813,0.9779042849967111,0.8776634245517196,
 0.21598833594245115,0.6906422247762192,0.2789324109663801,
 0.036674442774766614,0.5878303661881492,0.2555174132217328,
 0.6547192001686128,0.29204885316527807,0.9712467964826964,0.8557307747540561,
 0.6020164270177997,0.3218234180293079,0.3861667351756979,0.8575926817921028,
 0.9865130614021929,0.988789335172493,0.33310510662026294,0.25069112325672094,
 0.19083751743689403,0.6770473604621449,0.15136747314846288,
 0.3610412217207001,0.438523470740819,0.8755555204506088,0.04671165931178478,
 0.6594253479346028,0.8283786705419315,0.9597653530067063,0.7305767637188393,
 0.7442386358637528,0.3277686753642115,0.24399787994204125,0.4148708794485302,
 0.5296206708965384,0.4115284172556801,0.12166228293940917,0.7171739285694705,
 0.03786144176872164,0.5398441318031788,0.10655251797124066,
 0.16830088771276042,0.7498189082071984,0.5455854093541826,
 0.38177828839444394,0.9052953334588933,0.7402868299955336,0.78390588837251,
 0.00658993954755871,0.4793050263817238,0.4983249479149223,0.6741178715534323,
 0.8061566467084766,0.6284701459018446,0.908741635062668,0.25687085971002677,
 0.17709418539287458,0.35107653419214424,0.2206560558871029,
 0.9549594372712175,0.03895899585890372,0.5276169655214361,0.4299604746295589,
 0.45011167860118717,0.7581318615647792,0.5537916913387994,
 0.15054367117821166,0.07398342836407923,0.4325617974788951,
 0.8768134438116553,0.8972349115858546,0.16973431001507988,
 0.28845882267134626,0.836351112130655,0.47697568335026086,0.5706826132749019,
 0.08848370490305557,0.39926170586116505,0.5652057834004438,
 0.5033116364900456,0.303608264762242,0.4911569539913233,0.8906359533175744,
 0.18640931079826195,0.8920561409339587,0.5791899738382821,0.8213034488236963,
 0.34386404610718435,0.8719354828189548,0.03694178526569791,
 0.21843627769626306,0.8411026123991935,0.847319538965147,0.9504817712424742,
 0.052063704275364775,0.13938820311357936,0.3570452802360349,
 0.026400091100839407,0.050647710029426585,0.9159358138648565,
 0.7021903268557753,0.6724531074347041,0.42602553510575847,0.6542119297100382,
 0.5469155317500723,0.36462598308238503,0.1721575979722515,0.4429756364601366,
 0.9430523401632043,0.8578202948795615,0.3257557287560292,0.31429278120764803,
 0.6230909523036487,0.7609863675645107,0.3157961128792883,0.7170230244236984,
 0.8345831656120016,0.780587454979052,0.6404131652696541,0.6849172924397987,
 0.8124237859366991,0.861063700992216,0.31876708715822266,0.7259625120842438,
 0.5227124831409415,0.718316284030748,0.6502306139348563,0.7455998119184645,
 0.48430135591848,0.7893590281952345,0.06733406886816906,0.08183244424926839,
 0.7935322267673113,0.9601004524155173,0.4527688141670966,0.8988276965164125,
 0.7211610641279441,0.14180182828055066,0.2772751940601097,
 0.12782602584353797,0.04863283815075348,0.41147699476347455,
 0.3206085783527203,0.5996273410253138,0.26893698392273446,0.8141377856539835,
 0.9003903609840642,0.49653678780839017,0.7669484836645346,0.6461599469111633,
 0.7142640561659502,0.8093674589380185,0.3092177963070143,0.04431305423731435,
 0.6097388373562125,0.8168641097321714,0.06611880791867164,0.6089452199716375,
 0.5505938431033285,0.007076916825607404,0.004869352008775363,
 0.6259896917676446,0.7453669801339662,0.6529043029303516,0.6808273347416592,
 0.7153746711482069,0.6366441603427907,0.29859270956393424,0.6509332132909266,
 0.72980450996343,0.6895176423937397,0.1364880958427751,0.8359727273891925,
 0.13044103651440953,0.8976125739524367,0.17649972973913097,
 0.7793282042991811,0.11287397469206939,0.22202610891222652,
 0.12737669895560477,0.5627767663593664,0.11816413362847844,
 0.6319180908315438,0.8078632869560683,0.134941980973578,0.22842226149316036,
 0.20060549431798191,0.8251335035907504,0.7917236934629055,
 0.33916926447760476,0.9357079038860268,0.3303298877697082,0.8407969724878032,
 0.0333999623926593,0.3588775541380935,0.9908215200894722,0.407691867205383,
 0.5652907350633487,0.002677231172180883,0.8380006015073667,
 0.4266917745092119,0.6368919142168564,0.8723177949756039,0.14279248817060464,
 0.8521307445424243,0.31799230431493297,0.49845625395712356,
 0.11171830753092926,0.7795064610555582,0.45010190807213024,
 0.9884127453148744,0.05736843775415512,0.9896477309533226,
 0.44905410447515526,0.9497571148022087,0.6701098556313214,0.5152373889697797,
 0.3989659000549328,0.9023455323562445,0.8783439037386439,0.11343935991184528,
 0.3403343203085707,0.1992186834482248,0.2571495031722184,0.40372781149705195,
 0.3286643045526143,0.9946104232759949,0.7895544108163302,0.23827859548922836,
 0.1254424252843156,0.014331000027670071,0.705936609494603,0.5293510251399804,
 0.38366114427702125,0.4097721444081267,0.6798374734695415,0.5347250835274431,
 0.9881452909726722,0.06992282273742467,0.49730941282904806,
 0.09548140729303334,0.09554102075067672,0.14089862531084663,
 0.046980953354790134,0.8088302434552508,0.3390073201484246,
 0.6776133983672976,0.010576667800569872,0.13069553990803895,
 0.8838666040509688,0.7188621917322862,0.8313518011462195,0.3748489484957357,
 0.9257708586028697,0.6068639696577762,0.9252719253780449,0.08015689949420757,
 0.9591939783615477,0.9574517515319226,0.36249363370587784,
 0.44464409680007677,0.3448352323153532,0.19376299067797953,
 0.7810997472006471,0.14810009124012324,0.21130130283900428,
 0.8410254919669438,0.24752920914457643,0.6850004430436358,
 0.20982766849716594,0.3043424908735972,0.09662270371721093,
 0.4091615696238966,0.06109326523084058,0.7465063625687074,0.4887066781677347,
 0.3266853674003076,0.4244312187582263,0.7605196921269146,0.4359566953487448,
 0.6132684369123413,0.2816276375025323,0.13571336268776735,
 0.34001228194816824,0.8627949515847679,0.570983751378672,0.39270004736331376,
 0.04689189686521111,0.0984133726217904,0.42252341875092025,
 0.8657823683185644,0.5930891404019525,0.22067018727149146,0.1902554942032364,
 0.587668125399825,5.118171416976791e-4,0.6059896973741216,
 0.41471183279384605,0.5364079125127486,0.07096664916552142,
 0.2917734832008072,0.22629275863401443,0.8098017845589215,
 0.005301345447010886,0.9806328724862554,0.8612561312373708,
 0.5680252234344113,0.6957982762895115,0.9854643219309932,0.39286420830075885,
 0.5609827646156512,0.3340194595313706,0.6587369363539705,0.5223596566835792,
 0.4558762512177975,0.5438592668963279,0.8776629809977163,0.9829859427952903,
 0.986556064950418,0.35076135992946333,0.25174412759463893,0.7242670113168571,
 0.7375119213361673,0.8225014972832622,0.7087222378347997,0.797083399427795,
 0.6420315068838859,0.46587352951903294,0.06559433413370375,
 0.5487017387346353,0.119855004408708,0.8068521772870887,0.7102723619784989,
 0.751891698708597,0.48223440127266815,0.8292673134798203,0.6240290452782574,
 0.6540878115100227,0.9449419877007348,0.726551355689958,0.7064458419072065,
 0.011496766890208732,0.987002810846376,0.09518628999092726,
 0.02703281822236503,0.6133691594883249,0.41393598851349755,
 0.1353130768964913,0.3411664587027764,0.023763888051509197,
 0.5028488517071619,0.9057880846243606,0.6887479042392077,0.06992978520335391,
 0.9351156180879763,0.8685699309040691,0.2890788349392106,0.738740043228983,
 0.37475716747024446,0.355709242159967,0.3404629095180267,0.5648292592470716,
 0.38345820780440165,0.6069892215128596,0.9709935659767844,0.7640314114916236,
 0.5770707815423401,0.8266462937275167,0.9284744238145375,0.4460967575569611,
 0.7248575162703247,0.5283244544918131,0.5478137326258526,0.25987539584481634,
 0.20862995020521713,0.853664576079129,0.08322025945525113,
 0.14503532978066858,0.552566893026211,0.01893074243990822,0.5302466985959807,
 0.655413200046053,0.05642048161604518,0.7722533142625412,0.883734842954338,
 0.9934754260853733,0.6628140978826365,0.347918251915462,0.605246714216381,
 0.46513264788037234,0.5849504179557703,0.15722398445644492,
 0.8405427757064242,0.9152305790062714,0.35187531750527423,
 0.49502760116309186,0.10359205277375205,0.4650420546864982,
 0.6451741149713031,0.6387419535727331,0.2948185380291528,0.3193417912684432,
 0.47544953316660066,0.5104112973687616,0.028420871492204625,
 0.22143913039600105,0.8294629409569877,0.9805197843427799,
 0.23110864175898937,0.46558432290711294,0.42214213590098715,
 0.042821793703388566,0.35538040909405,0.6669500175965034,0.579434694507926,
 0.729869194526243,0.06336134139483818,0.13230502235439368,0.5004376987265772,
 0.11364430415237603,0.3093248603829849,0.045515153023040766,
 0.6571491161940508,0.7623608908035431,0.7914061219183215,0.932724698938253,
 0.3703041730724004,0.667371214158011,0.6429896147606979,0.5630749348787365,
 0.8892451526868249,0.45347271406098666,0.6927630419185561,
 0.20243995803277337,0.9733043674917564,0.7824643863008638,0.8330354869112689,
 0.6495184376031303,0.44001581927970146,0.8494770496768379,
 0.36784013282607464,0.09310759663051726,0.2704902066454523,
 0.6451811924451475,0.5168087454894155,0.09071158908205912,
 0.10812283682704826,0.6235307865657248,0.24208491389805964,
 0.19771229102061594,0.3159462913097084,0.8392033938053635,0.9445112252399047,
 0.42414024756256574,0.5225697466879703,0.7684841860528075,
 0.24552870880781663,0.4072171159149275,0.9943077595893823,0.915540713171943,
 0.18025914650425567,0.269789939354228,0.40814414682719136,0.6881901531936997,
 0.3889927334353719,0.4668064917573339,0.08623298690389436,0.7332332689075265,
 0.892051695326943,0.9793976933776513,0.058657027050906474,0.2439085785906151,
 0.26486310282748393,0.5288463183218813,0.9745159395130453,0.1240140059276843,
 0.049957221296174414,0.7126545701730872,0.056218773588700754,
 0.3353755015166996,0.6952864119483375,0.4705911224343239,0.30389388794439975,
 0.3919027970040678,0.4406557613941643,0.5808058599804397,0.47362039755603913,
 0.8606717289253326,0.5505997872978281,0.9566013272876064,0.1402687299956944,
 0.38204453129203064,0.995887959888346,0.7991878444568787,0.5484599930690279,
 0.03875676444795306,0.08164981184946829,0.8560296988048413,
 0.3317388947127372,0.32989306739042923,0.4804243023689523,
 0.48696107943154077,0.8712317535617277,0.25363459002484245,
 0.07702645250100093,0.3294464973723207,0.27831866805652894,
 0.14042984166401906,0.36679989937865476,0.8684876047733041,
 0.6653056083662475,0.04409309146646656,0.09630347303995457,
 0.5998924216923813,0.13606918425142112,0.27101299526692446,
 0.6732401869854534,0.5090859858553184,0.6966232435053105,0.5921883332394383,
 0.4306779887146919,0.8367457050233446,0.5187692078861366,0.26331180640458174,
 0.48930325773764893,0.5015915233946173,0.7586998227639558,0.6573342843373394,
 0.21648082937372115,0.4958411468890016,0.5700706294497205,0.3060658655918582,
 0.34678742685114716,0.3056650424796158,0.08804397618121784,
 0.10741725781915423,0.7036961104544585,0.9662390537963302,0.5051178979094321,
 0.9351555540907813,0.621485913673231,0.5657933715042437,0.5503516811238497,
 0.1892422703752773,0.753809115546348,0.8185499406204914,0.622555107961977,
 0.035371633288239,0.4222800694586666,0.18050489268511993,0.5604243159845939,
 0.6468974179766177,0.4184652182423527,0.8212587429821354,0.2873286357850505,
 0.8455607841514051,0.5250803534448029,0.18159607285094914,0.7313773937830587,
 0.8770984102470791,0.07236015956316388,0.4638749758053031,
 0.04504672498053186,0.9340902169587942,0.819137792921003,0.22423082852314546,
 0.9583855354217458,0.09239286138878122,0.6345706680027173,0.1384542837798286,
 0.3610385494709627,0.07852986924698235,0.631286942115064,0.03649454438140631,
 0.5430557167374667,0.4048423733850801,0.8278955557109792,0.7781893502526867,
 0.7062847228139889,0.6444225866390499,0.09175561487847617,
 0.14399274053027467,0.05543650937079725,0.1629029298959831,0.669920726301461,
 0.04614196644596502,0.4823873317921593,0.32738132583971746,
 0.5988866469893238,0.4539971674360227,0.29932350685343745,0.6644933709863566,
 0.8269703621209962,0.7802494054821318,0.9619360605368517,0.44409673546163886,
 0.7441949926702519,0.9114739289037309,0.1563309287230803,0.6080267745816121,
 0.045320432079299255,0.19966412950045576,0.9025156171422548,
 0.02591038912082755,0.1633449884235657,0.4320533058277408,
 0.08531656815085631,0.9993336803260497,0.1152844672198805,0.8949636846577709,
 0.950429593189693,0.5053847874822552,0.6414634626279139,0.6442806643242037,
 0.12814809515497771,0.5229116573269973,0.05671916571883873,
 0.34878599328702586,0.5452250990270391,0.031139156443333427,
 0.4223942912930929,0.4144495603205083,0.5871114678117829,0.1747923080521533,
 0.2274916464747192,0.740246733269586,0.04956984927742003,0.6134498810657255,
 0.13545357481148923,0.8141244032059622,0.29719534987325624,
 0.15582806436124985,0.9745809198927805,0.9065539523064862,
 0.38226397203574614,0.7353347672522011,0.8845959850196977,0.2970012151787903,
 0.13696117359429283,0.21715560229472142,0.94071056593052,0.5969039266808579,
 0.15596250726520056,0.7600692952213508,0.4452826081328758,0.8682387489082445,
 0.4344750330222096,0.21830214806579862,0.39816993677033585,
 0.9187034066460683,0.701093288958147,0.6496449280384138,0.3361971129607725,
 0.07234293976328443,0.9015017514297241,0.36877716609660394,
 0.7238787331633568,0.9661536976792245,0.49591361401501466,
 0.052347787941121715,0.4786707112438432,0.9815764546203355,
 0.021086409266512507,0.6617634915043589,0.2935381228829206,
 0.7705804980751141,0.2989859620314207,0.13080781520215223,0.4905604404737448,
 0.7639824500008847,0.9389592259408169,0.5883818585798426,0.33159563970251593,
 0.6103725868163851,0.7833780729872208,0.10660953592238176,0.9770513168899234,
 0.5716703748552824,0.09854929563953119,0.8200310179110147,0.5597818726614918,
 0.969021074173491,0.6594772691634532,0.4029425445417917,0.20096740244459954,
 0.3842879811894856,0.648905779344164,0.9166913462386861,0.6770075774500054,
 0.6637346968527996,0.23871207934033067,0.25933124552965325,
 0.48746704973352994,0.3770721406809261,0.4573250140693038,
 0.19534107500781572,0.2082920091089261,0.24456224520911096,
 0.8746055012433553,0.42358020844224176,0.5811252895404573,0.5227632630480376,
 0.5678792372185983,0.21525520242052998,0.38142840066799044,
 0.33275166157284963,0.5949299571318065,0.5150880605480123,
 0.12626074364741502,0.40834379429601997,0.5229519336575591,
 0.43105414500210315,0.3935525994301672,0.7189156388197291,0.4328227048594441,
 0.9175269221880707,0.781355158641058,0.10604259238473246,0.6274348202971345,
 0.087975659886955,0.9078317490096001,0.8964227731351664,0.11790620099620575,
 0.6468889755213655,0.5271019192740208,0.18529805660128318,0.3390250178218416,
 0.9153624147683214,0.4000428457056532,0.11500501500932259,
 0.09914811727697914,0.9719323600629939,0.24646433836526893,
 0.15184717988854002,0.28830738476878615,0.7715173475003005,0.228659231182748,
 0.42520391356010345,0.8996218220537566,0.07644894851996897,
 0.12367368841274762,0.5025642918039074,0.9586548575217743,0.9705984916902518,
 0.0641941947984972,0.2173761658352995,0.5755480434481095,0.37491972666131057,
 0.18022080570243149,0.5882413303086815,0.04691980808208496,
 0.1833381343943491,0.5294811147639373,0.16560538667088742,0.9401716685209097,
 0.718714115943532,0.7160533631274907,0.7915888104856196,0.25145340208395484,
 0.08350743766059232,0.24980900234703107,0.15528557810624954,
 0.9059504573214824,0.6087487977230103,0.8110282390989129,0.9870243097082113,
 0.05890580662310141,0.604010305748868,0.8091675412714997,0.5473060627388586,
 0.6331041154926563,0.2497237488208891,0.12853109024122478,0.8692185422319371,
 0.03122605778343146,0.5856922793846178,0.1383969545449366,0.0690210795084627,
 0.7869749530190493,0.23907790856319022,0.022143361300028364,
 0.07771297494789386,0.29317671492571185,0.5023046467065917,
 0.6219092338613703,0.26268828035017333,0.21599296741349372,
 0.22873150714692914,0.33272658115349596,0.5543172424839643,
 0.8321553949806504,0.9750824316888556,0.5849684727008895,0.6849800901867926,
 0.9851971610562472,0.7979300948219468,0.1893534719133524,0.9426914664936678,
 0.9585344184040518,0.7838544301850092,0.48558874128245755,0.919107415683468,
 0.7117770269746688,0.20743844436073677,0.47618048354188536,
 0.20350463165849664,0.3529398908781679,0.2935927862475026,0.4092286579714387,
 0.11747107762524844,0.5183970817853041,0.05957259045590346,
 0.43730635930118755,0.6279454015027555,0.08328991158564603,
 0.7289396810029365,0.06627477321457431,0.012331164330266264,
 0.8436971259872236,0.19762279726659315,0.34588844409955555,0.659059522055337,
 0.6078666108888509,0.2594990461705926,0.9395825604325612,0.43204290638713194,
 9.418076721068314e-4,0.5272643265672607,0.28480376637805205,
 0.009392426107357577,0.9091080996548284,0.34852091877669966,
 0.39423008748367805,0.19856510251785042,0.5763635820773737,
 0.6231587841761234,0.960170977174932,0.006732652858428723,
 0.10605268841005788,0.23161005527734946,0.12080342649261344,
 0.07419595607016916,0.22043618494008488,0.13520345662483502,
 0.8753407462194048,0.6991029129767974,0.5996586471567954,0.04692930538276019,
 0.49116577055290656,0.22122890695625919,0.6152001906205362,
 0.010146250916429178,0.10053556845396039,0.19619735332954769,
 0.17220425557760488,0.11347762528302141,0.5030231839198037,
 0.3714001359996737,0.777406140897557,0.026094670460429126,
 0.006964826176710792,0.582669959820914,0.5386768353744287,
 0.09482757763273442,0.8750155368246547,0.21745479351263197,
 0.7657209431875748,0.7032412984931619,0.5283488066889088,0.45483303345305337,
 0.798967987019349,0.9607045000560797,0.24526593793138518,0.990343163003897,
 0.5614158449460075,0.22942204912149067,0.3674737270750662,
 5.982569163872631e-4,0.5113264112947914,0.854369071766483,0.2004497143706021,
 0.3149800445335005,0.5974650798714227,0.24334160155799944,0.7638079615170711,
 0.16207840787056838,0.7216835099009622,0.3722155973691794,0.5151700845948899,
 0.5437591076578092,0.3989277314486037,0.30065127179606166,0.4088352278299022,
 0.7371502102034594,0.24105162747260567,0.460963233367954,0.16269390142646423,
 0.3282065395568776,0.9373621576561504,0.031748684587446285,
 0.9495922931713198,0.31103181838244365,0.45400874294893645,
 0.5286372236998722,0.31157936001379716,0.36894381283941513,
 0.16383909160487975,0.06448351793831986,0.31023329058160354,
 0.47161294897783024,0.8311467330692123,0.825152062412914,0.23042815032728003,
 0.26445124160068034,0.8968596282195509,0.46682517291565184,
 0.3667095228733952,0.28112558157869916,0.0785301286160841,
 0.11114338376677213,0.956001789751676,0.08537346424128722,0.4417996270472945,
 0.5738272067756395,0.21864179575131504,0.5906942486998177,0.530060347142806,
 0.9172162831886206,0.4567639331122737,0.04343036065530048,0.7822764186356961,
 0.5112456551962051,0.42153658528362004,0.5453425703768776,0.0974262661014208,
 0.23927450956567997,0.6913287363173118,0.1711393722492781,0.2957902281621554,
 0.2400222833800444,0.4762133338498482,0.8973794928179083,0.08431710091918809,
 0.02837735200054159,0.6395478232035995,0.056464344811633205,
 0.7731429095187807,0.7771161075195514,0.4197335973538008,0.525377806521546,
 0.7144339813742452,0.5515928365003476,0.5401143599726771,0.47617826065944746,
 0.922879195454031,0.1692855661183793,0.2970362242673209,0.3354567856780897,
 0.009409011529459721,0.4975737385539728,0.5022940312507023,
 0.8919491880604022,0.09650264300741118,0.17619349003058837,
 0.4508073683964131,0.5575292772315257,0.025883762243651853,
 0.2592680829630455,0.6997518008599449,0.13094826277802962,0.4401512770986934,
 0.5131581565794804,0.3737119567639544,0.9726544962350656,0.8070121632204035,
 0.5821194113003173,0.6650373860216727,0.5105513085111995,0.01348044159131756,
 0.4957212735383617,0.8870392797354182,0.18270891448380722,0.7516192823980313,
 0.8464931923378174,0.5376390962982105,0.9511292995851864,0.45922676310965715,
 0.2326376626756692,0.3389588332373934,0.8469789000625358,0.5368428496068521,
 0.5312067113800195,0.946223373891968,0.68500611445639,0.22530233470200123,
 0.24658053006223457,0.2852090757984217,0.40119092458773675,
 0.9849117529844836,0.23268433354115026,0.9163818360893173,0.6985641153568439,
 0.37961700178476965,0.471583037173707,0.031246450492494637,
 0.03518848198212654,0.6183463861778445,0.001431145992001448,
 0.6322096714391037,0.14499433365693182,0.4317569408617177,
 0.17776143814076595,0.942130671213364,0.7964028841602088,0.2089143911468222,
 0.6855869185835446,0.7565645274406273,0.18080386245772784,
 0.40308209898839764,0.8495166688023577,0.7148016690666366,0.8664574353659933,
 0.8803404817718836,0.3420339151174697,0.747166109623856,0.4307513886410822,
 0.5611294704034346,0.34924925394510853,0.45656965821847617,
 0.6827496317557219,0.9995063766688768,0.44472204835020457,
 0.14521901507088952,0.20352194714146432,0.7600155475991648,
 0.6793907746929141,0.24432006550613639,0.5222464388735663,
 0.09042615419924982,0.8927185158411026,0.9386501406482576,0.6704785767578724,
 0.1567563451261743,0.20537134863434037,0.9837736544209005,0.8722801138968556,
 0.7126439694820513,0.7062079954647298,0.1839566752612728,0.9370845553418222,
 0.35363893908343724,0.08992186157531568,0.8064453869623677,
 0.6958340955370159,0.009746959004381828,0.12276162339558572,
 0.29903080510541014,0.16220057916669084,0.36966424783178764,
 0.42361389373460767,0.7424505446472991,0.2236795232760722,0.4716160391373039,
 0.3520225067900371,0.3555120093775912,0.6366674516030263,0.6498469178235265,
 0.651853382772074,0.8410251476585984,0.5352306663767631,0.8447254056503803,
 0.7812340295288179,0.22644964058592443,0.691231920820382,0.23311748560204482,
 0.7427000914807009,0.42684154345399805,0.7742056617888924,0.9584562475455694,
 0.5092811236240865,0.04893608980860187,0.065671007367055,0.4199500444904738,
 0.13441349772996347,0.0302821049333738,0.592093376257871,0.6219850300033127,
 0.9568508699185354,0.09113235863931513,0.5370191688460444,
 0.07970127091061552,0.6267079068809109,0.20159594202675335,
 0.3146040062516062,0.9091160115710725,0.14379364041793197,0.7629846911794309,
 0.05135017016120158,0.3910724897045219,0.21715630948263687,
 0.4926377742102501,0.6442583734001065,0.0772814308320856,0.03628542934876999,
 0.7426153805912197,0.9663915988700337,0.5890529585577575,0.2954350880082065,
 0.501944488877495,0.916481149156284,0.07535250140604066,0.7780655988027205,
 0.5500054691734722,0.040095823747784975,0.4594856606389528,
 0.5251821117776264,0.21189921902816788,0.7670671165552772,0.9709667057974269,
 0.3444284951344114,0.3413233164844598,0.582232193301351,0.49001856425154655,
 0.3127787581372531,0.0042615011528136515,0.14360228275673637,
 0.19723591675906338,0.3777084942062009,0.8637721548732206,0.5957261502237988,
 0.8029219611504826,0.7367128944390772,0.7484409577613134,0.9183119670389613,
 0.8804395750210333,0.0968377191142974,0.7514545982967443,0.7392063984162384,
 0.48127348906397127,0.9207531885478567,0.6608642190567835,0.8564597526414452,
 0.6876380966674622,0.8960179834633216,0.13878410389333729,0.0808378206982504,
 0.7496111209938512,0.964328673770313,0.015620408929540197,
 0.11442854988343343,0.41158203921227465,0.1250927958414817,
 0.12440824883568202,0.637582770293506,0.5822279066059641,0.7222216588764072,
 0.24179960337310302,0.16071461606742743,0.763891998585783,
 0.47871964321696703,0.9292699597770667,0.5269141883556915,0.5726026669042693,
 0.7632527922690705,0.03363321857024404,0.91232130631121,0.646407383697371,
 0.7252043188449275,0.8919248369997876,0.7055653679381592,0.41773733005759717,
 0.04103649556329236,0.3305334717884275,0.5481372532693749,0.591979663873325,
 0.027153623544841854,0.9649258451321293,0.3836574189423465,
 0.6055844905268564,0.18926895482189487,0.8870349347690092,
 0.10439662240380065,0.2660203218316932,0.6184858594941839,0.8459116677199663,
 0.742717104500205,0.8073942384317222,0.61212234137293,0.11590571227154156,
 0.28218739053058306,0.19237661789559624,0.6544322169583607,
 0.4361788871649953,0.6330476514016313,0.9592419582362632,0.5201346100887457,
 0.02317782373248911,0.17927516484789785,0.4232831245484572,
 0.7047040336865461,0.2957446972232868,0.827203091271934,0.04699122616551277,
 0.3998084735700742,0.5456631780046887,0.15890104024857155,0.5416968159678548,
 0.6555403307601918,0.9720039314916398,0.22015744574759755,0.7394886995696357,
 0.09797338329504257,0.958883534612859,0.5421982601870637,0.21316733876743188,
 0.5324098772590463,0.2063514975344618,0.9602134885989296,0.5921032756211196,
 0.5820929815969108,0.2707074961455851,0.9848119929421397,0.03134062395615056,
 0.6486857910460017,0.9799969131074897,0.5425162267551105,0.5531811646537405,
 0.6630389530040196,0.65865319562604,0.8681834086926493,0.40266012567330545,
 0.20075797854798472,0.5128236830780841,0.7124065054250769,0.8508309809427391,
 0.5731123223649504,0.8680930617061527,0.6343910129967835,0.5369636748349417,
 0.9150906528677397,0.5049555760434201,0.5143218198294291,0.4078512420207333,
 0.5012904255827005,0.28375926177615096,0.7379989985651804,
 0.10355945125489607,0.7970964925633292,0.07262943764688157,
 0.8392248776760898,0.47546744567549026,0.3369181585367973,0.2774047396922542,
 0.20721275292087293,0.5201255925055859,0.6279586158193996,
 0.10710396046541937,0.6420605787881268,0.11004836035465893,
 0.7464703291376332,0.6345503642596534,0.6252283829934119,0.5022938721431192,
 0.7442014999239067,0.15444923900813368,0.6006620768371371,0.72595962060772,
 0.23068043093154933,0.061959933648759624,0.5140826723065697,
 0.11228312266871399,0.47306570002382753,0.5006512073471435,
 0.29940927309296217,0.7803754247858767,0.4993990745352931,
 0.14468689382167998,0.6813929212081893,0.48767028592382644,
 0.0680745830243279,0.30049089144702346,0.6154070710023782,
 0.05323332099772771,0.9400956498468106,0.09621887505384341,
 0.4994781658604286,0.04337656551637936,0.9535422828238419,0.5936545021492134,
 0.40577057935786276,0.045822115646579675,0.4584832045497538,
 0.6364342511957473,0.7202388445738943,0.9813221153191481,0.11362844718066478,
 0.09475891749549636,0.991031491585544,0.6806002730829687,0.947933418467403,
 0.547246578206847,0.18988338386214532,0.8048533375281195,0.6634811689129759,
 0.5111939167578048,0.2862643128533431,0.05746165045470353,0.9950436058669618,
 0.25340659134446275,0.10528688110384021,0.6789649735386305,
 0.8393219058078553,0.8451903261382734,0.06600824766200464,0.635087618432469,
 0.5081712020333568,0.3686642579042603,0.7836444916838999,0.31818553824745854,
 0.7617259847334741,0.029399683451482583,0.4299748656432868,
 0.9754071817800263,0.2219323104044385,0.8154125227142143,0.8215942340219761,
 0.6376711583383182,0.594657354197536,0.8163827929835409,0.9369449752952621,
 0.7060912270297375,0.5983362770507192,0.04950267234277439,0.6962848941004767,
 0.5455543683571495,0.11591273831417215,0.6227718594429597,0.5016942955510749,
 0.660843475297543,0.8087296665924556,0.30682274714010804,0.11572877068924603,
 0.41770040343347004,0.966470502276195,0.5359620488926872,0.7269488913536271,
 0.05596679986054598,0.5305632838513225,0.32370477133161213,
 0.45633011629837905,0.0191883139117216,0.22402588913592414,
 0.25172789615629887,0.19936787399259615,0.5927142040484397,0.892035974439282,
 0.9601315407274984,0.46793005387543396,0.7754507787409852,0.5650315394065238,
 0.4650622187721395,0.14281945723955503,0.1441088999148883,
 0.26834300835199654,0.36210700298178,0.500408802127523,0.8529459180452594,
 0.574003038637588,0.6770408751721607,0.8129503893907413,0.36762493277996366,
 0.8634295753035625,0.5845048396706278,0.06448125504864954,0.1899089509994124,
 0.6534162001140587,0.7267561699564005,0.8100114500150488,0.4086812250245021,
 0.27896444720765545,0.21065108313825387,0.18875788798965132,
 0.5463416650422062,0.8532996475489572,0.7625868012774236,0.30162522770786926,
 0.35008861393534274,0.3886651475707733,0.7614760239147789,0.5420734735556882,
 0.6373798102113001,0.7765693621247901,0.9434222912561793,0.24863153182797382,
 0.6961829246195249,0.4936497390660194,0.9174375941368214,0.9472170294439168,
 0.3927859896611552,0.4870624496571392,0.34076347528658135,0.1971755926108516,
 0.5993570258773153,0.4760874517301038,0.763293641744011,0.2705284816222867,
 0.5104257203010394,0.4673613140747088,0.8157689615197062,0.1653120737638929,
 0.4906208959898062,0.473642499205405,0.4465686766403463,0.46411430020602573,
 0.7296752839158624,0.016094327396158903,0.7849663566977052,
 0.24666696993559611,0.9180764123691851,0.8321323427509051,
 0.19970735032213027,0.13399303998054235,0.7432079944306389,
 0.21491576933555234,0.2921828247267453,0.448953822112672,0.928795592661442,
 0.750691378413119,0.7137334465846308,0.4571215214790697,0.845847332404989,
 0.8047880433365129,0.7109623653253849,0.08662072342707028,
 0.04911383574921002,0.6159816826548534,0.651638084128805,0.6173031072583498,
 0.4457139028952368,0.5676071741573085,0.6990805659353795,0.3427837564021664,
 0.7467663871176421,0.6780470143288435,0.8721212789700349,0.04800846271083148,
 0.5757952319234861,0.8131516344062417,0.11502793557426527,0.448191849993432,
 0.5146521516602499,0.27660027492791284,0.5943802749046784,
 0.23244965815283447,0.4676816578315487,0.8838105201315285,0.5978596692368667,
 0.33581230802815343,0.7246086944173342,0.5751574346768323,0.5885379705889899,
 0.7179432642762195,0.7862269703732909,0.289492419644077,0.9972357043865276,
 0.5814414854509452,0.776939671437888,0.9309946654923673,0.2100586724131881,
 0.2749710809967705,0.6798513705912277,0.4089227600072808,0.4242856221592012,
 0.7543847927750218,0.8676221774620897,0.5824292597017063,0.8691868075680439,
 0.9312995619668104,0.4266613904803377,0.2896211003893989,0.6321882152595915,
 0.8172879765394063,0.3028147735240144,0.7844241104071799,0.943836369308125,
 0.3951819762161036,0.7918066513667356,0.8719261449466089,0.5419521458142454,
 0.6953019194152869,0.3714759113867785,0.18916266455151087,0.5135699157694482,
 0.9647869478992795,0.5492123332918992,0.5930622875111036,0.6802831600858318,
 0.5487697888021654,0.7471770207350458,0.025054309266865693,
 0.12985130099562314,0.6795370780820471,0.05957446948621614,
 0.12654768823025597,0.5750668744391563,0.5187421286626412,0.968571146548527,
 0.5023403536107397,0.6922200869366372,0.25247401548865245,
 0.16334179852554576,0.12994582080070516,0.5050679004626706,
 0.3041451601015799,0.2152553003155132,0.4563246112316164,0.6876003870556061,
 0.3235425169808971,0.08893299760987994,0.32414879082418957,
 0.4451593933690543,0.44226089570053306,0.003188170106711019,
 0.4209951279784896,0.8959011701812591,0.9222971256758042,0.1958409857150354,
 0.03968924712626798,0.030958858299821368,0.8973567097908773,
 0.24382419202411598,0.3493580448689422,0.4957069089142486,0.683392627196915,
 0.4005332781819062,0.4809396773228307,0.4300345050203095,0.23065338354790277,
 0.06441285517184403,0.8480324270432444,0.47582969396429964,
 0.5368331249730662,0.9403360873365811,0.6217757933579825,0.21480130840141687,
 0.8697229798399605,0.2514829773017879,0.7582291348602316,0.5222593427361764,
 0.13523307310336707,0.21463474926285164,0.2380420677399807,
 0.8084065951890372,0.02170557083678526,0.8856308697351982,0.9070462458778794,
 0.9110025671313113,0.3883545973873417,0.05414888500342807,
 0.36619609410737386,0.23654908808063602,0.3399020097424894,
 0.5558133081950322,0.03854865365049165,0.10476527071481179,
 0.7753957308502742,0.2664652121114979,0.5833627258605676,0.5042891813387914,
 0.019820226827683696,0.32052943981088533,0.8232781013256076,
 0.5308964322566759,0.05721642481907019,0.1717757959119981,0.2854383506701398,
 0.8876268722003412,0.34677549875078817,0.643610272747083,0.4463204571046542,
 0.8409177812170394,0.2331893507233025,0.8451898902925337,0.21446300164993892,
 0.4147394829242962,0.5813875345147592,0.31339684596925255,0.1563979573585491,
 0.2207190882917751,0.3472325863202653,0.12822739712409648,0.7414292906505479,
 0.4376797366394135,0.5892138987343456,0.9617440516281035,0.39054501756607074,
 0.48525671720627506,0.9206585526528379,0.06552377444102508,
 0.07510295264311273,0.594680968504704,0.5254007373711607,0.41932907006523823,
 0.2852382348855067,0.6051268730604985,0.2272403546697258,0.7604211775247828,
 0.03670219352955861,0.3254322801814238,0.7583002479582991,
 0.060294473985255004,0.5123567780069496,0.39846856359521,0.09697198641407234,
 0.8953821689177714,0.38584715490955956,0.8021772458542232,0.1819969425771849,
 0.9717654007782146,0.12297312582911446,0.30456313860363204,
 0.3661532643606519,0.6629361988374151,0.023382718856355122,
 0.9529022524722729,0.9020843179931752,0.5089549163031043,0.9801323381157343,
 0.06462332918362779,0.9617657655217373,0.9407667553748301,0.8686592248659442,
 0.433464899959328,0.09288239301015333,0.23086509556601564,0.3501371540508428,
 0.766592117324421,0.26045715138843106,0.3145062139711148,0.37185070155334377,
 0.11485351285639112,0.9873262803979667,0.9834902555897234,
 0.37401982165946634,0.5624776689039297,0.5120107550193831,0.9406608962106308,
 0.4757964226470818,0.39610035289192347,0.7469298407898242,
 0.16697365658323104,0.6520704670478048,0.005389921216345295,
 0.304090391987451,0.6312335922919485,0.1553971915088267,0.08334337984873263,
 0.4645453505349286,0.2273575623835834,0.1773326554387873,0.566426925227992,
 0.4081381335244496,0.029962842513594268,0.6923857885121787,
 0.17835444683294377,0.6202387650827258,0.8073456626032315,0.4356287508074681,
 0.7908835493701011,0.7424107541677485,0.6570484498141281,0.7904932051740668,
 0.416425537377608,0.15662491075835305,0.9860651977218804,0.650122647434403,
 0.0692445395177066,0.019070645995447166,0.1307833443765104,
 0.032313458608691725,0.9794858847721714,0.38953666592561187,
 0.17883440137646656,0.006848627496694792,0.02900447627482383,
 0.25226132796393086,0.07525751710718953,0.7140451877158486,
 0.2475759789687808,0.9391338584189982,0.6101649175747282,0.1299255094897248,
 0.09201319364380167,0.17122891141174346,0.9128027943608422,
 0.05775532304882469,0.3238445386752409,0.3319833668718485,
 0.38062583832250363,0.1335727222342843,0.17432275583084655,
 0.27373765829310315,0.4829317499835567,0.7901305504015955,
 0.25407072367651784,0.07962427311905529,0.25312648957384454,
 0.31237286915557505,0.35125949575458437,0.7791825288268898,
 0.4357862120193192,0.9108460319321161,0.5158696068623653,0.9996601259150001,
 0.32091022119069434,0.23757196060546493,0.8775868715284277,
 0.13505351770884355,0.23763615879982392,0.9580522871725463,
 0.11163530717405867,0.795092611435891,0.635114340256134,0.24286673135900327,
 0.30578971126326393,0.7314478714520688,0.8227771151485977,0.9838186389441339,
 0.2808642489754013,0.016468133167513943,0.7793861166095329,
 0.5046008070822218,0.20848672256341894,0.09881912108725621,
 0.12550145599331786,0.7024871012976865,0.3438629237579346,0.7945775945931608,
 0.5930674823137337,0.39943282296567995,0.7566446486352693,0.4593355081153919,
 0.7456170405749594,0.775960073484706,0.5670462768570113,0.7838297136636385,
 0.3389372822857998,0.861752045564846,0.9925579167463028,0.7805578620701884,
 0.424746794331199,0.6391370651522121,0.0742768264009841,0.5100615992926203,
 0.7791797905159835,0.08104817575703971,0.6447523659030632,0.7042909835872777,
 0.6066253212856662,0.9542957286560646,0.269427708889895,0.5174174939083215,
 0.3943605002820938,0.5680189141580549,0.7863903335643477,0.8332604082365083,
 0.6058380812939371,0.6470601619400969,0.11316897593013564,0.3998189774643174,
 0.054952742963235046,0.6483887526169352,0.21299368458946444,
 0.17293882481165612,0.1438168122650788,0.47174803158403356,
 0.10610032058924579,0.696722668423125,0.03144390892777338,
 0.13116444888794343,0.03318981710818658,0.73328551581793,0.1261719376320659,
 0.9031692379804841,0.03458152534393366,0.4845070114563169,0.9951429436005904,
 0.7536090327791832,0.9168901283943544,0.32616494172645827,0.3640352922071197,
 0.3210416164722726,0.024445425257127118,0.8661569490509416,
 0.23260585452130667,0.16924039678606673,0.17568489093145656,
 0.8221623744579797,0.9392456265578425,0.7269443881269426,0.37504565080121277,
 0.24945687034836728,0.43897488429165243,0.5232920344086699,
 0.9017222293741832,0.5496818261892327,0.8495780971556576,0.9427573644174969,
 0.38732524243114086,0.3784836033799278,0.5267898615020864,0.251917155058079,
 0.16815651607248794,0.7812835226362702,0.6968040429119873,0.7190987449604654,
 0.4904495920507391,0.34166487728390726,0.3704889068569972,0.8905224022234728,
 0.3646344078105367,0.3703285904176936,0.9578980671678363,0.31466411225064816,
 0.06553774619232666,0.9015171597034763,0.9569852620435167,0.7511913986478951,
 0.8244837712398104,0.13318387638501783,0.7865479266935756,
 0.43233300792620943,0.27067584237326336,0.8834057990112232,
 0.8541250351760737,0.5389526375658951,0.3428399977893377,0.0071464766490138,
 0.6768756078018083,0.4034351568591791,0.17038593648384592,0.9655675036316007,
 0.922090927446552,0.2793815010977275,0.608375707935712,0.15026991633003295,
 0.31681127079107363,0.42828660009076935,0.36771087137156155,
 0.2374004211211288,0.8443751059040584,0.04046631919505861,0.3038406368568368,
 0.979762806623602,0.2618461389955622,0.055447646527677974,0.7250206831921646,
 0.059221639890589484,0.5310023568181993,0.08333212091735209,
 0.8669931927203894,0.3215937665831955,0.20709449027341265,
 0.24824748065007407,0.5034871892397692,0.8375393265443787,0.6712435554075831,
 0.22616216420753632,0.6091332721550364,0.5745800609992728,0.8403143063370367,
 0.38274287110728733,0.03797738062202094,0.02468663370017954,
 0.4827658767583296,0.8558515078424004,0.8044217179163413,0.9493933591865824,
 0.9456322092904499,4.721349703440403e-5,0.32635475421411364,
 0.9274000478298061,0.7967319245757034,0.8728816756792528,0.22966305628857986,
 0.9477956971647374,0.40915521521379117,0.5920402775953517,0.5593462194473361,
 0.06417245268448235,0.47256695291574213,0.9037568261153132,
 0.3775226407414417,0.14564546889574548,0.3860444989545385,
 0.36414197109514834,0.1564584165991454,0.7858694175445267,
 0.18732429585999932,0.331753110752961,0.44018071103999423,0.364622114947844,
 0.5379275367677823,0.39872649232407487,0.8315548513330943,
 0.11079767982624467,0.052576115563714,0.7823544032241468,0.9918996740843208,
 0.12123406736892783,0.7642618343618128,0.5133510936901208,0.3315889651636089,
 0.32580799286267426,0.8805121953605103,0.376783088295247,0.8607175818366264,
 0.5056743074604694,0.9088682079808672,0.298634988687144,0.02514561952054095,
 0.019495542180097192,0.15947178927712968,0.25161847025418616,
 0.7279989820230924,0.733986811648169,0.3156689613557693,0.912627541474307,
 0.3923722896859829,0.47030351419117067,0.008274534084658836,
 0.4238233813727914,0.3944713628626748,0.5394014177548654,0.1878564129098419,
 0.6366998926826922,0.4283599671208014,0.8387736429763831,0.7450800816175195,
 0.5246818149579947,0.7574948840989697,0.7694143064140322,0.6963171984427866,
 0.42704511044455185,0.9273108825160616,0.5952131960761657,0.8046294931483363,
 0.7645923762229855,0.6372213340970556,0.6816546393767339,0.8350982295884223,
 0.15654870739683924,0.20804750605724742,0.936618638275494,0.4563938832598604,
 0.023962884087610825,0.14342649638047955,0.5809966860622603,
 0.526778272201717,0.09351050601556121,0.1285060093737055,0.8231893723780063,
 0.8416648208070414,0.13167346036829886,0.5666455503712478,0.6292432833937807,
 0.4158794167051164,0.07808993471990389,0.2523990712132932,0.4325084269171885,
 0.9218997197905056,0.4835005430557273,0.13119207526120336,0.0882391512365358,
 0.8224553195946358,0.11418921504186641,0.37010022820350597,
 0.42091592795374344,0.7659016577032063,0.8482500237274692,0.902806822522986,
 0.09511724329623894,0.756516723796948,0.33313838450988065,0.5924129859491098,
 0.23340650887871928,0.9575543475496787,0.27919182245089347,
 0.4462475866384097,0.7820656464184218,0.15909396464371317,0.785386561626464,
 0.08551211084733801,0.39469708134251835,0.1098155489869791,
 0.7754190801790354,0.29557548832236336,0.7998046216342163,0.5210225286931384,
 0.12269959076969073,0.5731010328819599,0.5597590713906208,0.6330547338549628,
 0.2037124248338804,0.2585416338278501,0.2157703014539727,0.4973131663128276,
 0.54176910155028,0.04334469404044339,0.34191921158748995,0.6387419507135669,
 0.8219256174413517,0.17137766818674516,0.2848702169466921,
 0.47592327242757404,0.1076798876212639,0.07619786821499897,
 0.6908395268264871,0.20295782317931244,0.18642237655681793,
 0.06345196446970569,0.4396525183830762,0.23781455519185002,
 0.08135949595639114,0.0772644087933897,0.8892585538591623,
 0.04279646667304371,0.7421412476927857,0.5764818783115742,0.8887606036251887,
 0.9654589470091725,0.2559592789955608,0.3604366009621609,0.22837002900026782,
 0.33050447064386823,0.6574379419021705,0.3802783382202284,0.8809117739451743,
 0.48459262179065965,0.22746331744287285,0.43543705464262994,
 0.024911161621969602,0.3917894298087221,0.801107825067411,0.7091559247269468,
 0.20054788690042236,0.5039920671724186,0.3386416198354355,0.0974555094933075,
 0.14819911377260597,0.48628350462722036,0.5086234400075109,
 0.49146233708219056,0.9589213527947973,0.48283988253089094,
 0.042666713019824054,0.7340201524103923,0.8376546146727496,
 0.6785335270579582,0.10567710914627992,0.26078905535921104,
 0.8105733135321828,0.01650154409807636,0.42144078965128995,0.538415963674461,
 0.8027037270896588,0.17847400143312453,0.039111331371386227,
 0.26672678119066284,0.6992488901300987,0.5745126578826724,
 0.37642883805353733,0.7090246646871785,0.283297856628583,0.7691281919483073,
 0.9725314173004429,0.09070714772510446,0.8988482488285283,0.2865758752494302,
 0.43160980943171035,0.6734146895642915,0.3274119235580065,
 0.13890431818593718,0.3512782382714221,0.521378038341415,0.7838932503674213,
 0.8812846564167985,0.443387416490178,0.2431840522123322,0.33861684927303937,
 0.5230140965643233,0.1571138846226765,0.441991095656755,0.7359771510876688,
 0.3227498520852419,0.7752519051042692,0.9720596504129394,0.7814361568104787,
 0.9667454850443195,0.9514054444557606,0.9269933792308805,0.6868868614944881,
 0.6927346702741284,0.5745356574819583,0.8560356974440384,0.24413246899880114,
 0.14642240888309588,0.9641681405879052,0.7129445673824526,0.9224000984169921,
 0.14250151687932844,0.45161815930409466,0.25773965490106776,
 0.5207363554492738,0.6516287397262359,0.12598896155312223,0.600343699969785,
 0.07930397644165876,0.6219256948123474,0.12664507394854885,0.854055451623722,
 0.46098933456751257,0.9703924631861811,0.8341056480694317,0.2175185859788623,
 0.06825095512941515,0.9613311683772974,0.2770827475040719,0.7613541774125684,
 0.7775840868812668,0.6962787419488625,0.8402711563955851,0.11713724057176922,
 0.9325254378724223,0.5599318238119255,0.9910034124062514,0.604543264439258,
 0.020542018482739732,0.8643099138495545,0.14965064868233602,
 0.5674215524820219,0.9737798021025504,0.5339024312774439,0.45547830025391056,
 0.5674897604422906,0.2785583752493195,0.6869635669726301,0.28859713503797546,
 0.564773177937466,0.3555016410659344,0.36153822348157705,0.829567536042283,
 0.6054832755502685,0.940168919125731,0.23146364749100168,0.08249775751944266,
 0.7063704373374073,0.15471215597763022,0.46483249533144455,
 0.19353930423738785,0.3478490558012486,0.09812372846808703,
 0.8255476383746284,0.0753427774703046,0.7014306371549082,0.7832443149255646,
 0.21731757336643165,0.4654647443463298,0.773254373044779,0.8447687978731977,
 0.5295484333623038,0.8820788759913907,0.07797739824152194,0.949589803457899,
 0.8640694421616537,0.28787293404266623,0.9978892933932331,0.955062147015046,
 0.9956172646988177,0.3914486110303673,0.3888095267477385,0.45619655196161624,
 0.8626031983522309,0.9163644612847328,0.15670738335875412,
 0.24608785129397714,0.07539724941151471,0.07204946851048966,
 0.362814137374733,0.628366218442681,0.2172799478054812,0.5178397564348923,
 0.2954759390164742,0.7249199476888407,0.08031697378178126,
 0.07070316731188431,0.42224428051011864,0.7681666453676333,
 0.6069666627007035,0.33107592635491645,0.7643345814113935,0.3002394044032157,
 0.45600016604966553,0.12594731360673084,0.42515048783367504,
 0.8224857874559005,0.7633179642472228,0.1180415131031024,0.6407682810058362,
 0.896558243536658,0.6824422377763093,0.5282030700311235,0.6193856580051504,
 0.8226021573008748,0.26105314031710625,0.7284041259484223,
 0.39334512536803024,0.483543288604267,0.9745114929213327,0.3701753723607747,
 0.6476515952015295,0.7044084104846717,0.0981930890498266,0.9049591170951727,
 0.9795461149561142,0.3763000787041154,0.22895783982806195,
 0.14079852632769496,0.7610179945122733,0.1392174994748948,0.9615257465741941,
 0.18133543305687727,0.6845905965208858,0.6377663631191183,
 0.033359127731699356,0.5480471931886639,0.38617982106740967,
 0.8656867167225939,0.7200285131935623,0.4509904417574985,0.8595646042747599,
 0.0034000172983646992,0.8627223353005563,0.7751402426218433,
 0.789282416867031,0.8122791994787307,0.6715527284016887,0.40298192804510347,
 0.2899083151331727,0.9738403351445019,0.5808635378538043,0.7401349498298414,
 0.09017620104634738,0.8139095559404402,0.9505113629871205,0.7189743408098794,
 0.17061775909219423,0.4599598252219129,0.27424111384361305,
 0.15317675906112138,0.5364598971516342,0.5434923121177964,0.2241156916608843,
 0.7863489979955728,0.5942799685783561,0.7907553040658011,0.8461461366887457,
 0.4817412343455014,0.5666685593479345,0.7105771439197884,0.2882287696369089,
 0.844389261044587,0.3980637007505964,0.6815836722689624,0.4436349010614439,
 0.42309227905535063,0.7117328324377961,0.1779440026958764,
 0.04499081164592211,0.8973896638761865,0.937062459048937,0.3517551152060483,
 0.7362614668553992,0.1418008316762559,0.7989602342540367,0.29044658620250974,
 0.9095833273043785,0.7273935405673937,0.9741327229634871,0.45964344460124607,
 0.09383530143299346,0.501500264870128,0.8106927414526287,0.7406483586193607,
 0.8917004353606095,0.47534646345644704,0.7922725823142195,0.3806280328646692,
 0.339124279589504,0.20776134210201413,0.5012307636798706,
 0.028419243973786923,0.7383572385402783,0.556556816516048,
 0.41834512088844844,0.8632064093028233,0.6170784068328995,0.90889725640561,
 0.5898218407500905,0.7451246025466385,0.1233310434520406,0.6442279196826702,
 0.0528781287690383,0.38770229992355265,0.23562897547260098,
 0.5643647863941222,0.03759794196318067,0.6976855711240592,0.7085591916195068,
 0.9516290636183631,0.7803580634599941,0.986914625120259,0.22119996908353934,
 0.030018063363497216,0.6305701998861816,0.8080646084217319,
 0.8931882649421815,0.49074613711946435,0.9178096736624111,0.9420968455254214,
 0.8796270627400297,0.9029181134240876,0.6062559643961912,0.6227219230233018,
 0.35056266312059814,0.110938891643649,0.5297330681645507,0.2818640544927057,
 0.2477345020828995,0.7263909893189656,0.8800002672123935,0.5652349211021919,
 0.693397271178941,0.7603498011789511,0.5323924985043255,0.2734012985570681,
 0.23437011800330387,0.5446551203244501,0.4650871687649052,0.5198394521093579,
 0.510917200316331,0.1826456035501699,0.5285132247816938,0.12570247992397543,
 0.654205426397097,0.031151114563042648,0.021002382200647762,
 0.6095355172640007,0.0397024637832708,0.4268029207342663,0.19437236673195535,
 0.7730890654005949,0.8725919395285145,0.5679786300361255,0.1344077241910292,
 0.2534553415908626,0.6195897417340405,0.05387902484517204,0.4040949980520495,
 0.41701049821832326,0.7942438553599762,0.3384093470643085,0.8195070882595946,
 0.6052373466126825,0.20343188968366266,0.29720416882548983,
 0.09638521358160557,0.9191786636095018,0.8358222120090332,0.450120491970495,
 0.12876377470902967,0.994018139993388,0.5903362680344558,0.8404171678600557,
 0.017550589144202378,0.8696621163947639,0.3800718214899911,
 0.6433561199023216,0.9939560404773264,0.5289716454904659,0.3187816126551103,
 0.38705512657763574,0.21113542241776861,0.7433804577950546,
 0.9546544004193088,0.827292199223499,0.3138892928734308,0.42311506064089455,
 0.4241327193780977,0.09988299006971779,0.8609640664355213,
 0.17385903123404833,0.9509874506377455,0.2520503968822462,0.4176167137884079,
 0.7634845822564659,0.9978876038235429,0.3013434892453326,0.8125101412787241,
 0.8202667190979017,0.442980140195567,0.46136431649945486,0.6645333718708114,
 0.11070837755444374,0.053429187006197365,0.21563448432029841,
 0.7635184336164821,0.7069282515982718,0.965648367287789,0.6729289799044661,
 0.14239438177548414,0.7323169251357183,0.9492183430633047,
 0.32426616547349374,0.7419643869701195,0.040333786716735154,
 0.9338670286277211,0.08275868432323263,0.8064192476235936,0.946559897896703,
 0.8487383778394649,0.24439902642977773,0.1731258805633904,
 0.15513936175361231,0.6397819587713557,0.28450771229237737,
 0.6183354285833549,0.2063872289323634,0.7992122427466303,0.42025930115102006,
 0.7035615097805424,0.4971139129482057,0.24543425993584345,0.8533337200420876,
 0.14774825110721357,0.55865930194837,0.26618826122860884,0.5740008018902159,
 0.3407482145539519,0.044210120434261646,0.8672111825840705,
 0.06965486821627509,0.014415773918202701,0.42407934419881677,
 0.19180175247165687,0.3734037737047958,0.7951858138638684,
 0.13413393018781772,0.8783485166054685,0.5604813104294482,0.8678218007497954,
 0.18204806198119672,0.6218364427043276,0.5831126690822171,0.627320706711225,
 0.6759734467687708,0.734200975453241,0.402627663719502,0.846282993690514,
 0.47413299134706177,0.08800261849980917,0.4686048980529387,
 0.8795642294132706,0.9081684820676743,0.25007634299157444,0.7700157763392447,
 0.9017074093914829,0.003350318317974832,0.2513278777407373,
 0.07996706891156613,0.24203049816236155,0.666813561946193,0.840280233755121,
 0.7789989207028097,0.7896601196816859,0.5152252223948852,0.865018275469663,
 0.3049822468707666,0.4942761511346693,0.1162931867085999,0.5582580436639313,
 0.6912114887766247,0.28240211345544197,0.1449029925874762,0.4548961917034826,
 0.43859336039444163,0.9024193546990232,0.0761836892652249,0.4346486401575531,
 0.7800082883451461,0.7913788468726618,0.31451973650051324,0.5895106417591442,
 0.9798835235594544,0.326192064039576,0.6321230953169088,0.25224201570978944,
 0.23376419704646878,0.15055507060333917,0.4137449872945941,
 0.7703675594346525,0.8413649113601805,0.3570411927570267,0.7448565657512951,
 0.08747437518877688,0.6383618216209922,0.009171658028675589,
 0.4770739376302472,0.7406705752062117,0.3976034906974064,0.76861423029966,
 0.027487539685631668,0.03228670277961998,0.5859645078830368,
 0.14351318102834998,0.4309959654440203,0.41301430424255203,
 0.6186644758386448,0.6332144451091545,0.9606781066994486,0.8369500718750412,
 0.9135489221776346,0.8838062627535235,0.06891657584794741,0.8500989128906651,
 0.3349762983980077,0.6304286484476086,0.4107661048525788,0.21994046052316374,
 0.12434050213293713,0.5235931716613517,0.34959166457142377,
 0.8736352437099756,0.8274595499540073,0.2611545069638004,0.9240958071893923,
 0.02074845734525499,0.7896555649038267,0.2261826441941741,0.623876119704985,
 0.5356794854392367,0.9793502103310933,0.2449322473148423,0.2951035164077276,
 0.8184402469310447,0.9770143266042828,0.5942784135211195,0.4981643190175453,
 0.5104916025471333,0.8724745693731811,0.21961293876620502,0.5772865108828137,
 0.3186430840697392,0.26077066186236963,0.20810521184062525,0.820345044636234,
 0.3000621409093771,0.5581313196036107,0.4018680310156215,0.9775505007636737,
 0.41839123371310427,0.8218568628544026,0.1408953518265581,0.6869706461358014,
 0.8156148057227551,0.3176443116779091,0.6658956395005267,0.43138471330169736,
 0.016988819356173024,0.3507427684708997,0.5247973453856182,
 0.15659069369659462,0.21272086680817182,0.3406683016993559,
 0.5996907718405665,0.014325058671767699,0.94484554393593,0.26800829815042904,
 0.9175062980531037,0.3758973904076721,0.9272513559238418,0.39644102267789805,
 0.2955084398700999,0.12419604944102947,0.7055348805373839,0.816919768229966,
 0.41013591555002205,0.9534123261015472,0.8493160557252332,
 0.09166192378781857,0.6430189442079127,0.9386609189248089,
 0.03671817635780683,0.39779808848180065,0.39166137398253464,
 0.05681304959876832,0.20439199021991994,0.7518034175096777,
 0.7957655616467878,0.9991816464150018,0.2173541325820193,0.7735447031237992,
 0.5977099143553783,0.19131786998243272,0.905692514124081,0.2717265316748667,
 0.6788375597394938,0.1409077178499708,0.06377312598330565,0.4913637303652204,
 0.5608888386623941,0.01939138305594379,0.30863842261876406,
 0.7698385676506377,0.4147946251629211,0.3362303314407592,0.7866692316765849,
 0.9719241323507289,0.8693016117445025,0.08594372250403248,
 0.28037868925414466,0.7629663827510632,0.6703244922827957,0.6377825214343101,
 0.5078630338283868,0.4292692940654871,0.3899688132704251,0.996597674738803,
 0.05314146148871912,0.9494319320955517,0.7593311444041758,0.1748673386513001,
 0.1949951396207834,0.05231305706365763,0.20206077263449895,
 0.8176990984722569,0.9875877440893691,0.08233407984244412,0.8371169675444717,
 0.7468252078879387,0.165060523058301,0.8637721796825722,0.6458313542732549,
 0.5029396890015689,0.8370861908288891,0.9007158767701823,0.6000029249244048,
 0.2539311520991592,0.7430764730891154,0.7564289564250128,0.9469373444841657,
 0.8908389499540065,0.7183656731843958,0.4552290000428626,0.07169331401574985,
 0.16042959266712808,0.5477864856519734,0.7729491368728826,0.6907880692444425,
 0.11699108628341504,0.05449060415800988,0.9347964497978852,
 0.6011645871314193,0.5137394247048681,0.8347546623605735,0.4756430332853535,
 0.776822011653611,0.21461684334336395,0.3069595137708737,0.11254644854299467,
 0.3027922289046491,0.1608472169894828,0.41830019709576094,0.1271038273788342,
 0.834523355535513,0.24272081354619346,0.13956047098242264,0.5157469336360487,
 0.13563132544583523,0.22281693263026936,0.38548600395097965,
 0.8527277916115562,0.2188133106876453,0.5125484948071302,0.5442600914436181,
 0.6120013417870326,0.8666998252770778,0.224568030049333,0.01179332359226426,
 0.6842733825215325,0.5453446391778152,0.355127056834991,0.6733722229304059,
 0.6129924605519121,0.1755671626255253,0.06427393369483159,0.3330609293379021,
 0.6853123654284748,0.9951690341830618,0.8890227318607153,0.8127662133994452,
 0.8473496908412785,0.31134456990134707,0.5367803397527149,0.2041558895747032,
 0.7464714431869337,0.6732874227660656,0.5720000625809379,0.9926628021835182,
 0.15372764812145334,0.27091506540415267,0.530512971710278,
 0.47908148105327464,0.046114116351219314,0.04387755516272063,
 0.4875832698500956,0.5711735981951589,0.7329677161349645,0.9861598033504617,
 0.8336557127416949,0.02498809664977031,0.2038122648526609,
 0.48952482355969873,0.29114965653400726,0.7897274117914896,
 0.04532046710302795,0.2890456670625563,0.5940306545932729,
 0.050569638752795965,0.37672253652445153,0.9254879698466248,
 0.11835934108597279,0.999484194087797,0.7515454150959976,0.38037884469199457,
 0.3824574407898509,0.31479141624510076,0.2881524746854379,0.8804842647333725,
 0.7151224999959936,0.5414058248709006,0.15921743773927755,0.6308556332189934,
 0.6474018458342291,0.6558408307920053,0.6496985276221815,0.21095608590861192,
 0.23376411436412758,0.7210451576308337,0.515005239969474,0.3711588245869558,
 0.15133794109225196,0.36607609906040595,0.5232346230350666,
 0.5226765461834478,0.06505060731108159,0.9604165879594186,0.6043287324065509,
 0.7821832124022234,0.4966726555675183,0.07208557251120118,0.4755479048107366,
 0.7744981565139697,0.4784682419708385,0.31885031124295626,0.8827665015107254,
 0.12004141670940327,0.12012007222929921,0.25794058989427004,
 0.33157383485670755,0.5174202209865336,0.49202355802376563,
 0.7350405333153474,0.550245548207806,0.6802529306856445,0.972593833268604,
 0.23274820887532788,0.5002850426218834,0.9308396483569024,0.8425764570002715,
 0.004578196615659835,0.7268591483138575,0.855191381759221,0.9042017636377989,
 0.5266238092208302,0.5680053138863744,0.2065907418583911,0.830289969015773,
 0.4845439111438339,0.1908127132189944,0.6042005851785479,0.8183465332908204,
 0.1514090601453798,0.3088508965785328,0.5353974535918788,0.05368465618248641,
 0.5918351662842829,0.4108695156173563,0.9758076297608644,0.2723202021207177,
 0.20009046367059757,0.11364490044846565,0.6756362328180762,
 0.7018820870505666,0.7585954957887819,0.7061293130851957,0.9250297994064681,
 0.7983304349054606,0.6704911532318089,0.29215286516034533,0.4319799677525563,
 0.22500538782514112,0.49579054895638386,0.31467917010022284,
 0.3205838668239829,0.8469658912900733,0.8769945225054163,0.1863299502194793,
 0.5229197401220347,0.9525328831888409,0.08134071447758273,
 0.07612223920937478,0.046278574932490546,0.8263122315328751,
 0.40222631318304636,0.49568772521335214,0.9751409442646152,
 0.5106284127479712,0.6164058564319697,0.09238197119845626,0.2314742448005056,
 0.9056408065658407,0.19302478027025616,0.4620680152175449,0.6741729583946406,
 0.7151364048123079,0.11860095310654384,0.27649164160428863,
 0.5751252806468782,0.6888333057448781,0.11166798982452042,0.5264647769397354,
 0.3636536000849552,0.24122308786553093,0.6116701595735546,0.7465523478634334,
 0.504860130752274,0.25649120014877924,0.5184058136855556,0.4961123150637904,
 0.9509063171846616,0.3092365814492871,0.032211613758892255,
 0.7997610125483743,0.688035153887205,0.9023140004072763,0.9235654649258039,
 0.8960541514016489,0.5810378495741941,0.08350418786376712,
 0.17311142517619782,0.9522117687583868,0.2160902514007529,0.8395749058724506,
 0.19095116803547152,0.3277760221358339,0.42274872500032257,
 0.4747262618849101,0.1730341625433679,0.19903000848114394,
 0.01444296898159092,0.7495575580352024,0.016873724255917377,
 0.07325192115479195,0.9699771623791811,0.39987270036661504,
 0.5138847040606221,0.4026482039387427,0.7458565805058163,0.1846976648667107,
 0.7039173227404163,0.4595355996678282,0.2784993986180484,0.20525238994848571,
 0.09823570667648962,0.10797288681939743,0.2535496313970692,
 0.17161920743192705,0.9365929854342132,0.8394129254222984,0.9686283395720284,
 0.9154239784541212,0.6870044967424489,0.6765990658571187,0.7379096799923704,
 0.44097005050049576,0.545404804393772,0.1979325969573097,0.8611752159956837,
 0.30293427692045516,0.48160138678910014,0.2777134963443645,
 0.8167115282633216,0.03335971273564298,0.283948281423546,0.08084486812903025,
 0.7862895279670779,0.09818566733424083,0.3264748084323088,
 0.003110412248735872,0.8613398225293285,0.8612515573631604,
 0.13982516222232666,0.5650705441227013,0.920773856564472,0.9411802724213382,
 0.6072637297092892,0.21141568658031873,0.9024479151800495,
 0.29509676253771144,0.28729726874502115,0.2524871339376912,
 0.1823778682872993,0.1052870784909612,0.6367031734977024,0.964215956807713,
 0.43765415557438303,0.7061994872182779,0.6548548688694501,0.3772976353934606,
 0.5817420585433241,0.15756719912858297,0.7075759961061913,0.6188098922373745,
 0.02402323803924078,0.8588908855094821,0.05093843852749247,
 0.3033915619984233,0.16834593497232875,0.9407292284392338,
 0.10325524060457703,0.10570279200574928,0.0464233517546625,
 0.14844945656631003,0.41873804315078167,0.004441727115832683,
 0.37562460114155827,0.34747352084473815,0.5174883687920063,
 0.6332144066843889,0.12511395129503966,0.7501151970651657,0.7590879114382351,
 0.1746913933714549,0.77439603329992,0.8914744293156807,0.2591045151984066,
 0.3407628185290107,0.028232020489797227,0.08590728415898585,
 0.7877065605316491,0.7995078026721204,0.02417388880001281,0.9615656072187379,
 0.839243168992635,0.9680817255136156,0.6347487928473515,0.5174635630490786,
 0.7321984066163537,0.6005393390416925,0.8767024159824818,0.16300069239187498,
 0.07203372946433917,0.04233531252456335,0.3213697567362921,
 0.42305214695000615,0.020603496805196464,0.3109261316909443,
 0.44141219990120095,0.65378351039354,0.9984512533348181,0.9355481072459967,
 0.1694502933751988,0.7722163109968287,0.5541637149977727,0.6514942496540309,
 0.16110927039446488,0.6535176502641845,0.6705257496958239,
 0.42423226943014136,0.10744916496282297,0.9128709026757037,
 0.6861423085582923,0.9733722225683007,0.6377115661662505,0.8554167281669942,
 0.9006249925967649,0.6896557334815234,0.6881750758314452,0.2244147208665238,
 0.5471111655315144,0.8589550299825491,0.7501065803696308,0.6010354969447715,
 0.727507267340758,0.03198447989424569,0.9529511793453035,0.5161397394093696,
 0.576211689889022,0.5819532963945606,0.7452917880423157,0.04654712886398149,
 0.8209754693902629,0.031191486110062283,0.08645779091637817,
 0.09147552229673428,0.4708228774446527,0.6696819934599655,
 0.22058636196492187,0.04677366443561326,0.8546596899994601,
 0.9038619111100566,0.6304669021299482,0.0451944144007681,0.39048121046632644,
 0.004277213585871076,0.8592964846626193,0.4294146316591293,
 0.3495480764977359,0.737696572310219,0.0769075873191799,0.6784160616549172,
 0.6246133467329066,0.1819443366598662,0.3655856575024814,0.5580893084364235,
 0.24861223738542093,0.30945861508127004,0.7431285858780211,
 0.2386969290871399,0.4701254157229845,0.4779267660273324,0.10981215342569839,
 0.49121508914883083,0.7834809555087644,0.8449044447968794,0.4195885503399669,
 0.0022720531476951855,0.6655396853299236,0.8955128764647309,
 0.7288475032765614,0.9486328242177041,0.6569163934853182,0.15914047878956306,
 0.6845647758538178,0.05311420243774656,0.8227390708495705,0.7696272483596124,
 0.22875424971285974,0.8280578357236512,0.9559171133679276,0.8575750341546876,
 0.10193146037233414,0.29086328067828804,0.7359481052437167,
 0.8285296172880119,0.6937216414819423,0.6834078448680753,0.15978884683076255,
 0.3173100110820857,0.6594065615876825,0.9224630467522206,0.30211024501799866,
 0.8908537481701396,0.5957778389773499,0.9390926721093674,0.10056789062955973,
 0.8657609424489336,0.8469193445275882,0.6324831833997833,0.35991533205757786,
 0.7842991712351203,0.42323631038598264,0.2663791304362637,
 0.24084253720327198,0.6664979046787559,0.6307817384170772,
 0.17318509709196628,0.6271026812922724,0.2867777805762769,0.8457678168012176,
 0.3913432488722932,0.28059110099791407,0.9442235529572104,0.8249278672198682,
 0.05457376174138662,0.8762128600710557,0.15811809077782035,
 0.7324919442854536,0.3206257396340657,0.60665476902752,0.2162330137563081,
 0.9684805560674716,0.09490786360167891,0.14535157031354962,
 0.7798845978775968,0.6998695794644626,0.5793795539482919,0.3604672453182216,
 0.7906880741475771,0.595709046902713,0.7980265546670665,0.004404265667557139,
 0.6838939810246024,0.06674098454424593,0.15015053464164307,
 0.4534401543894925,0.23316214379272104,0.2752428551476812,0.5657180423814436,
 0.2796565076825597,0.2579628952850912,0.49411082682206797,0.3720553714922591,
 0.044659759941588106,0.7351732484464499,0.799099763786475,
 0.31186768272798304,0.719168545436231,0.1953740262772694,0.5973584451878085,
 0.6438613067614188,0.31925541325320905,0.7327592502394888,0.6252332160344831,
 0.8760645032577024,0.5770910850721267,0.46903881326365127,0.9081547890396732,
 0.5392291779240521,0.16069744276219244,0.9593522517570755,
 0.36625990562723265,0.6216559308436,0.006836256598650947,
 0.017520480815421124,0.20705310199764215,0.9121730773303747,
 0.08420428397296564,0.224106086364963,0.7571129904052445,0.7636807374657149,
 0.15909770399887246,0.2422331998850018,0.4794684605293923,0.9186298562898599,
 0.705381525408526,0.07567701949813366,0.4491236365289676,0.5852379031599031,
 0.04273271449472649,0.03321382292779762,0.11352107026760905,
 0.6286147720227545,0.3651920917602033,0.4057356422346605,0.04196785419793769,
 0.531821610024055,0.9814570783552424,0.46217600454765695,0.5923880303202597,
 0.23201066452676444,0.29386296843338666,0.1344408811874903,
 0.22632118899339437,0.7778200187055752,0.01878326966114341,0.210322023418257,
 0.9856416308879726,0.7621939005418426,0.3153302847533572,0.8335948413266235,
 0.37841491117242587,0.9643254232256175,0.9885812765942493,0.9640475211596542,
 0.7104031812316154,0.6115911837596817,0.07093711313828266,0.8842903374934895,
 0.43425961880317865,0.24843889726033974,0.08857865316707592,
 0.03877738187376567,0.3995851907000523,0.7976358025550291,0.5337396545012477,
 0.7411173974300077,0.5310567060701257,0.1361603233449361,0.7651591708934741,
 0.7718044108805722,0.09696014475599912,0.6583756184647787,
 0.11053158470071356,0.5810153818148636,0.49434834607209177,
 0.4733296150288675,0.09901234382552127,0.7296547733532974,0.5755905786420016,
 0.7964727357877398,0.9279845495813075,0.7079529387487549,0.9736380909858874,
 0.7149326948786041,0.6128374565108794,0.6265999057904736,0.6194464982432589,
 0.7557299880025566,0.38702775270148637,0.416271868509445,0.34475634196129756,
 0.23557602169782854,0.7122716057730364,0.7807173177801253,0.5426105632285104,
 0.496492245571985,0.8971061007307835,0.6091179741021174,0.5339433216173826,
 0.20033677692790375,0.5344712443284269,0.7194833989266225,0.3780816465291297,
 0.47536383075750677,0.41430301828194116,0.06970253746836785,
 0.9863339636811442,0.37707594243252307,0.950159115871678,0.8963244058886625,
 0.43661318068839783,0.41970358970301447,0.58842758175417,0.17792829363314377,
 0.1133164138937135,0.3190779058200065,0.8862397723180084,0.14984953979019944,
 0.7869713680268376,0.10609931625149094,0.12427681934405965,0.519613387484849,
 0.3660335876547035,0.6124722779003227,0.0058251643007958265,
 0.3226423016373082,0.8411390994528812,0.5716086555087267,0.1342426874499668,
 0.8427657298929901,0.8827615177636502,0.48700986697807847,0.7325568971147508,
 0.9574012840961885,0.8962917467001219,0.4111408007534251,0.4646893332392492,
 0.758432819952661,0.7512550748329905,0.22613057899196498,0.7026482125560163,
 0.17277846795174856,0.5652578098360198,0.7710816690331994,0.8136577093734939,
 0.9772318574705777,0.9959301697633558,0.27178907533944463,0.9256569278838376,
 0.2412605895193728,0.5272464073320092,0.3917534637215888,0.20672797838549006,
 0.5850074313095355,0.5360893144904509,0.6353560710081649,0.5071941235499746,
 0.4383203932394635,0.9261643225019808,0.5400188718268416,0.6445978148405704,
 0.35299211368061056,0.4106545603581496,0.7898618802378828,0.9094777408195646,
 0.4313960521047774,0.24352724663705372,0.2732090366535972,
 0.25210641531446343,0.9139115200346238,0.5922975666161759,0.9948102522374734,
 0.8469907608684902,0.36727022761010497,0.3502408543341239,
 0.03380178862063099,0.5336001099297694,0.6295412546895314,
 0.15448296134133233,0.5256197717557133,0.11186922380684305,
 0.4528823273442142,0.7384568303411772,0.7164298922084955,0.23024069727898522,
 0.14248154727035256,0.2271391943724862,0.3795620448515631,0.7524138096451392,
 0.9008969578093303,0.13093121860241863,0.33810274388778927,
 0.5017873190769422,0.7998754252910115,0.5994951333184485,0.8445903518331557,
 0.09282276082280028,0.37487488215844256,0.6739426654819742,
 0.4404878625369011,0.554704452366056,0.8232967723658161,0.6568235011928474,
 0.6460725152032244,0.9699870766391265,0.9858374050669598,0.7939789996118174,
 0.7106886031821663,0.42186772588093746,0.03409161479873157,
 0.5856451443764867,0.7822826781865119,0.12987411619934996,
 0.07855752889769685,0.3173850995086859,0.575703209114829,0.9861364580881898,
 0.4234349640217313,0.29816242846212293,0.7156760928222117,0.3501447441036444,
 0.8866659526693126,0.8666531996052554,0.5212388000582422,0.8883327439459231,
 0.8503692742395905,0.34483436054662464,0.6565094546609219,0.5533688829178349,
 0.9544309715165212,0.012336047905992631,0.9962761492308789,
 0.3511433163052553,0.0477754297609152,0.9876488600728744,0.8177984854988101,
 0.5025844257691997,0.9544240427710939,0.29341578757396725,
 0.42158089873378946,0.0778866953825551,0.34685928281214673,
 0.11028610146544138,0.59592496979671,0.151443239902338,0.3072402196618751,
 0.8446086442071183,0.12931874994379366,0.5334204352736577,0.5906178067390659,
 0.2056057378152868,0.0022721998944048583,0.9248153923100542,
 0.9310540336909514,0.3044160692693073,0.9320451723665135,0.07440027180186948,
 0.9883594568119017,0.46798925523454316,0.26934395848470516,
 0.6430114407676567,0.8282160022314687,0.3339533789944603,0.8509597231649473,
 0.34244040090929784,0.46327147406301883,0.8595545180208954,
 0.29471960226787297,0.10532123301291274,0.4852995300764896,
 0.35839345388776955,0.3075357090082329,0.520827286627442,0.6053110289163983,
 0.9013851434066796,0.7834917150315241,0.9396982777589072,0.600770953504964,
 0.8775822313391631,0.3919384033937827,0.71129859165287,0.1421790963648688,
 0.43912178239421673,0.2232998333967653,0.04754976783366216,
 0.3613801878774845,0.8611957184264559,0.07728115030067706,
 0.11036260084090654,0.9256716436296974,0.8693878171768059,
 0.037435862305130185,0.6708653961891788,0.7343766482372979,
 0.3252683456517951,0.8137028155120322,0.5464645639945368,0.6232169818452344,
 0.07809977609931673,0.9903748164653282,0.6026363296854959,
 0.14057000494332028,0.806753079104483,0.15905324339097104,0.684181739477771,
 0.19927722705095385,0.815744182738324,0.7638229587296947,0.16926544486585338,
 0.9103938913574637,0.5346194251870362,0.2571819608140329,0.34533769365658396,
 0.1838477938445251,0.6241041539918075,0.2740978253866153,0.6661281954446049,
 0.6198647358620107,0.26058004331676454,0.9291814810289167,0.6302610294057431,
 0.6544699133295946,0.30466557126104266,0.29412560270643784,
 0.18942741651889916,0.1746097127913947,0.7267631736238567,0.726765936606737,
 0.4884663988930571,0.42472787192502914,0.4892650564631328,
 0.027002810978097536,0.7121266980527607,0.25285594726683447,
 0.7435473190112348,0.41397462869524726,0.7803393484257063,0.2905893560559537,
 0.2663282203978462,0.7575314176626133,0.42946345717194057,0.4018175831562535,
 0.07682523863451252,0.5707832384422662,0.5256676558497159,
 0.22910523492883095,0.18325154058998927,0.07337774619313675,
 0.44374101515597075,0.7876199666696435,0.833435602535463,0.21853904110447808,
 0.7942990108711535,0.5216564858794854,0.7520059104798482,0.875022148954422,
 0.3464969814505252,0.27717901549109325,0.4967580451649898,0.2573878570830246,
 0.67535282411073,0.9724586627199457,0.5150872982377811,0.3754558644792383,
 0.048989498780057206,0.38151430041256906,0.8707799395695841,
 0.16723622228762924,0.7528508993138558,0.13077534893895826,
 0.18347533209619138,0.019166063871897232,0.29469222688789676,
 0.9722714470283154,0.3644384322092663,0.04896534475110648,0.729373302615018,
 0.5136458552306388,0.41592314733573976,0.809607350471308,0.3241754848139523,
 0.43935486689103387,0.7636595572042054,0.3972447736700082,0.6394521109899842,
 0.15818572513583695,0.12199740748395627,0.21908125473993745,
 0.22943918630410554,0.22827147291824823,0.47481357649782174,
 0.13742705452587622,0.5957937020795006,0.5669138165561199,0.5969115333097325,
 0.002926484393207307,0.8832402068074033,0.3493599412829702,0.951569674635061,
 0.47726672542932225,0.03990528338304389,0.8776378985449419,
 0.24856370670644556,0.9094677059901177,0.7289526063084962,
 0.21958991240582415,0.6555640518954712,0.3934937028659087,0.5645239462595685,
 0.6786135513000424,0.38312870721347925,0.31872985274380095,
 0.09846927207196088,0.400570890307278,0.3727594720554833,0.22583414438500082,
 0.3258471923908324,0.42769743029262264,0.38460111585719214,0.870474193477557,
 0.7156376091187818,0.9574084148375929,0.9172229773437338,0.6053394074668937,
 0.552655158083123,0.30776011327589525,0.24955899348813326,0.3958188208207616,
 0.08394715717938372,0.26294202361494223,0.698986829191266,
 0.06432592163650541,0.24998502900594566,0.8748470774811499,
 0.3334771387292701,0.9502241082587548,0.22578286894885202,0.2735983285757466,
 0.8684397221259837,0.17284212443386693,0.8111397802436742,0.9175794674248967,
 0.7732396647141597,0.4339118000427602,0.3829137086505108,0.6406093270053539,
 0.3949252802547263,0.27815200277930807,0.03125089501933043,
 0.20961449943598254,0.11841142408802086,0.8719578340777229,
 0.8387788746298883,0.6581902212780996,0.5036369589422716,0.18268464441082444,
 0.7389999892244011,0.02874142990780637,0.3706400484950967,
 0.07883750787729416,0.25375683369656443,0.9738716260686215,
 0.15730724027806708,0.7331731734346081,0.16674255099234914,
 0.8707917630468724,0.4532145642969404,0.3637548915616625,0.12009406356914465,
 0.3289742632813405,0.09224173134069125,0.9909851479406088,0.8429418172914858,
 0.40269194898490224,0.8694974206405705,0.6808590092497264,0.2354474963291504,
 0.5158094315387932,0.9711046454480721,0.12211228839408295,
 0.09631789110310196,0.049218570537813644,0.647170770032419,
 0.5715695512787249,0.5723023477374577,0.573021642439154,0.8484318126059723,
 0.4827257936843883,0.7975185896504967,0.017642593733247836,
 0.6492844966171614,0.6152958752312921,0.08568613531138236,
 0.11217807161439508,0.38987624233084706,0.9027888269932214,
 0.09420811720731737,0.8632069467515027,0.20078628708208923,
 0.9383458116237023,0.9680852773613828,0.3294471747205202,0.58352051521032,
 0.49416553157516874,0.016569848746808558,0.39309893566839493,
 0.49866925091511616,0.30163584564768864,0.8350778364075658,
 0.7397865359360418,0.780188837146289,0.29043455944670526,0.7580314891602223,
 0.5987490969043346,0.8471679065340517,0.738148176832998,0.07240510777980103,
 0.34675526790039046,0.5523657912628648,0.4890918059341649,0.3410725318091361,
 0.15240531213908437,0.8663733135061656,0.16622650144221174,
 0.8492693861507618,0.21897324704910448,0.13299699094202544,
 0.49439179156702595,0.21971003962695346,0.6413359580077825,
 0.31183740632442003,0.9206318684058732,0.22500190994204883,
 0.9991938554252006,0.1645478906460751,0.6552678557183025,0.6755326672699653,
 0.7630653856767141,0.45207073686023347,0.2334369664964029,0.6880032378854972,
 0.524996562554161,0.4883513741472464,0.43943089147493664,0.45442490109858646,
 0.055904784365782056,0.20735440496334578,0.9943197422416666,
 0.9454253847812457,0.8211523658885875,0.6773677955725508,0.11373999715606276,
 0.960403350062901,0.8886524430532118,0.214169837421059,0.6376043636759804,
 0.11449977012602697,0.5392417048228332,0.5261040338045788,0.8294489760245537,
 0.6724059972015234,0.829337753005536,0.834035572792441,0.196384202698482,
 0.8565659771227504,0.8797455957496018,0.46350862375119684,0.7381111462675635,
 0.9617551118230161,0.2633039318502459,0.9018015284012888,0.9566704911548195,
 0.7640577777106237,0.8431730899899962,0.4686237642783272,0.01704122293273591,
 0.9445042668097663,0.11724243832932124,0.7835907258469966,0.3229691334182475,
 0.5464662456827788,0.43458619513466246,0.5264735546146008,0.5034899351854152,
 0.7739965197871967,0.5584603393270772,0.14666005525655312,0.5147420328476087,
 0.9051688978822909,0.043802549450366324,0.11840852273905944,
 0.7073747987906257,0.3226166507268251,0.5855594263751984,0.2190739083317631,
 0.09304430074537229,0.6002506778852721,0.5880150964739621,0.5668346522632544,
 0.7901283236866831,0.03240345131282285,0.4620474773769765,0.6543190451297064,
 0.7895330860817833,0.5048548320461143,0.2976899163253576,0.7625255488191791,
 0.5457317446055076,0.5431574442549347,0.2208102224515851,0.12280154194367887,
 0.14539348093636428,0.49829617369252543,0.5098526370347419,
 0.8004141806875057,0.9323985659776102,0.7238279518248034,0.13841219798687032,
 0.0713852281562587,0.8682774104187547,0.02187767609312241,
 0.031909837456086554,0.7756357772320911,0.6929852850200214,
 0.4443478035252587,0.8087259405219588,0.12357645404815143,0.5804087878002364,
 0.27546195670485674,0.10675475636941767,0.9870340538917426,
 0.48767203809128923,0.8600262319085983,0.9726345423836991,0.7846682203680049,
 0.540658214754653,0.23724440809661462,0.8700656240093633,0.3056408353021408,
 0.3193505722626834,0.103572911608793,0.6190660546166428,0.8376342017833761,
 0.8077811565155848,0.6207770789199205,0.5129392787157284,0.45794730413408324,
 0.038794640048589124,0.4453961264121674,0.30584625014655387,
 0.284839463264408,0.4945955773865598,0.8243003473405389,0.06959522815972019,
 0.3974063079195056,0.808147094938208,0.9799517115619011,0.419335380979732,
 0.5312768106292871,0.4984966957406868,0.7752482320793622,0.880162386815947,
 0.6724901579105245,0.5235036148560614,0.31744303544824737,0.6677803401578996,
 0.1843171356971255,0.7148648923468255,0.792547110379982,0.07596515599621556,
 0.022547016257593144,0.7886617538478778,0.007356723960868394,
 0.49247580347549125,0.326461670752779,0.6842444993148563,0.7194422441590956,
 0.6062445384453307,0.11034339808419069,0.9903443098794287,0.9827824398372157,
 0.40424578614090134,0.8118108671811178,0.11523614209781796,
 0.23988353649718697,0.39472027023437306,0.5245124576472461,
 0.6389284647701181,0.25528054549346535,0.3024501719418815,
 0.33820378599813894,0.839846502450555,0.6339951462437146,0.03945016163942161,
 0.8036679100414549,0.5190164228951782,0.9340479369383693,0.9158071483877007,
 0.9064911925489095,0.25950217179316026,0.2046921544037743,0.8970448580321074,
 0.2411926157610459,0.773237477952781,0.5995701005915237,0.32388103901879317,
 0.5120037460487691,0.23978474899951452,0.45406709302816217,
 0.4831274649502696,0.16271728684507636,0.6565646260300122,
 0.39928413212832825,0.7542178021987636,0.6579455027139071,0.2637984487464433,
 0.2925213375262671,0.8113010497077963,0.058383478407280975,
 0.9805128292484233,0.0997995279664583,0.4248563272511656,0.8410913624016007,
 0.5134983328028857,0.22904044848480487,0.42975542212290707,0.820226225825049,
 0.062407308806307515,0.4965005560008525,0.7983874012515648,
 0.6933205228551069,0.8515070566435508,0.45367653963898213,0.5499665630419255,
 0.4796670682383234,0.2780419066127904,0.35382824122403045,
 0.42868024251374826,0.7315357218828332,0.582157376036458,0.16933488751050496,
 0.3683927409506007,0.028104180278158974,0.445996434122398,0.991187930035716,
 0.2214409707670849,0.21112545126089743,0.08267627827758994,
 0.6780619380000724,0.22568683786408472,0.818093380965299,0.9187255757614629,
 0.8984115667523593,0.5260165168665398,0.2192167877090906,0.7169970273008299,
 0.23528672282771135,0.09401815576933281,0.44954017009834224,
 0.28035914910916315,0.15566960213097658,0.6209325963794283,
 0.3729709140347036,0.1605340628005869,0.7128822493307037,0.9930944314432864,
 0.3355483352074574,0.2727223346836827,0.6585913903880408,0.24657775117015845,
 0.34071920843275205,0.1337848433437081,0.008136774218663545,
 0.8937338528980265,0.39709890533983494,0.9618249205792944,
 0.26376150532075726,0.6731436541161482,0.2742144980804383,0.5339794469577405,
 0.48999151991390655,0.45303446766925426,0.01144247868406112,
 0.49580985675243383,0.5874694246642748,0.29532029039316365,
 0.7827870937629194,0.09427363662757271,0.5154764824769525,
 0.06595891660825648,0.31315573510102257,0.986745631112256,0.990679641121219,
 0.47478029055212834,0.759694883244636,0.804102614427865,0.24076854086247734,
 0.46044457527894134,0.8171804456136822,0.3928063855264303,0.281237105579733,
 0.6493600383907672,0.2537843288994246,0.8454102906541153,0.6817233471116642,
 0.9721811687878181,0.8008810853305398,0.35933507804862386,
 0.03580773697380546,0.24687839633659947,0.7167629539741041,
 0.7364449620319095,0.23684355812229252,0.753022643744761,0.5303474792766036,
 0.6250438293045912,0.07360010721427268,0.8986699416163896,0.5951643660039657,
 0.6215298827933754,0.10668590734311079,0.11380591466975765,
 0.8302791134797908,0.7343635523439309,0.7581670896266997,0.6095491743408329,
 0.7114141791867905,0.32428664882911606,0.21929885528809945,
 0.5145490320384358,0.34371441378809775,0.36423431683194474,
 0.8606354736746212,0.043537667927695534,0.7245092102848354,
 0.22438828547922096,0.2703143409502081,0.5924176887854407,0.7979547397006994,
 0.036033948229264534,0.09245480039583454,0.4547187205840556,
 0.9449172146329592,0.6654446691559578,0.030794545408206475,
 0.7001767339405293,0.9064116519043274,0.19409541731369095,0.679301141893048,
 0.058130200323250114,0.27411105740263786,0.23916640128661704,
 0.6650932601432006,0.8173234475023274,0.6525820448515058,0.5470352012734361,
 0.5964399538419398,0.9725289358708216,0.8393250015642508,0.3411312638269224,
 0.5827993706989898,0.6219998486657035,0.7184527293980452,0.7754329908916073,
 0.10183876625950272,0.07320214100924582,0.348665030344088,0.7270102853633579,
 0.11396673053331297,0.22493204705330894,0.2814208620145706,
 0.052224614348431286,0.45129219646538066,0.3531255037331731,
 0.8055015945795986,0.9622665620238203,0.4877590284021097,0.6776669356147088,
 0.5360923133080555,0.51131058657794,0.8512681674883323,0.8544669420340029,
 0.3799806744142299,0.5565786251505318,0.39395426619135754,0.7962210975412649,
 0.9058191475949326,0.24657723199453108,0.9168057900537124,0.6658707197495912,
 0.23946965634312867,0.034923966245991034,0.3223399366808173,
 0.7524079798726732,0.15444464533589197,0.4609275971381196,0.5689255643362214,
 0.21564327460454402,0.3513904347273208,0.9248730251415903,0.8413612407648482,
 0.9751108208606799,0.5181085241449246,0.9308381135311955,0.8531776973510758,
 0.1653812058733517,0.040786047548645565,0.4765796355188221,
 0.5044540596556257,0.18644136373919518,0.6928105707995075,0.9161245492555338,
 0.7493897788053313,0.5135421034141643,0.24625105113488388,0.6930730140462266,
 0.5074944012027266,0.9725954114237491,0.500428585632232,0.14391924022781222,
 0.6379998285293842,0.8208483010537094,0.915061000531406,0.0834810853528184,
 0.4277188816764863,0.6882269627619205,0.6036358620786768,0.16312384098713062,
 0.5958008426489858,0.9822385540065236,0.36163456777900516,0.3525396059186322,
 0.38674983569439647,0.33711354461022236,0.6506471635230907,
 0.3798008297457838,0.9358365535093528,0.6926927916249805,0.12674335105229229,
 0.02541833056150755,0.25036191929039164,0.47239961883769355,
 0.8548070216470353,0.11741502190682107,0.2093325184691499,0.8670834507073915,
 0.5651719310008309,0.8936601792406736,0.4472360956094178,0.9675435806411736,
 0.9687015856915767,0.013290054688368191,0.07944056022708657,
 0.5684526013118283,0.42121764833478414,0.08211204324877563,
 0.8024205602031687,0.9037296940008946,0.6028208791505032,0.8770413177378732,
 0.2801267255309612,0.7656464639839927,0.05698152255552458,0.9936793173720634,
 0.6049682459160728,0.23077717139916432,0.9550503155035285,0.6749995261204341,
 0.8329923700005852,0.14772929451933525,0.17235921357992057,
 0.12668131069242583,0.19234171900644081,0.20286480257544415,
 0.21522999997764947,0.05267686734977339,0.879628382099738,
 0.13261924450742868,0.4352425197838955,0.31582014674464254,0.330319036201581,
 0.5384968304158567,0.7061915174404311,0.438886401523192,0.6196676832253484,
 0.3377484150643001,0.8036330505626454,0.9338494339409049,0.3679170387919485,
 0.9070807084268009,0.05570606785797916,0.04807560202430461,
 0.9401396484698239,0.452848696665614,0.02801008872034183,0.5858260415281419,
 0.3195436769535438,0.7668024730851254,0.2578118158124565,0.17868098262576337,
 0.4965579898367193,0.47921736895022493,0.35142266151509505,
 0.06083417797834201,0.47923760731982157,0.8521607968174744,
 0.8536347748154371,0.9508435410589848,0.570316585514632,0.8182933795133914,
 0.11588888451217694,0.6846349636260847,0.1279969508740958,
 0.029575207347503296,0.29815586302826946,0.6051248577582098,
 0.2828798137095364,0.6188688956708115,0.1232605736979353,0.8324660251248428,
 0.9209468189276522,0.9749370341243622,0.04070408558904404,0.1111605808049736,
 0.4344390474852702,0.6713784442473818,0.8926653806934884,0.7653044221282342,
 0.9162879911404592,0.6794588983315335,0.27298160198384036,0.6713173782098231,
 0.11919677077677093,0.39065626899077044,0.2296873192810771,
 0.4365313771709056,0.13393544586342576,0.354207419515162,0.8342117562647733,
 0.6660224742253438,0.7077645421014661,0.41345711297030296,0.8978144148087728,
 0.4315675539854189,0.5231150078191186,0.03809470630721057,0.6869812283006541,
 0.11636258165184354,0.9554891829119652,0.06847860743226974,
 0.3652054427050594,0.22253735469483782,0.6239446521231213,0.3730572870048905,
 0.9963208948255898,0.5770599149633366,0.48638127744471793,0.5843028730970443,
 0.3399799924703564,0.3171334968086671,0.09363887271595406,0.4558799886299084,
 0.45527841891422915,0.24037700647726956,0.48443250199470644,
 0.3126807029077763,0.6149490038389337,0.5995945014055009,0.7330269826638602,
 0.35878298092232763,0.8810657109323399,0.29168648348224,0.6725258985539868,
 0.0811870767779912,0.18062558893042113,0.4864068019076587,0.6763506502370512,
 0.046380744945639596,0.6547005075943069,0.7727971616753517,
 0.42475673010611337,0.23415905354267774,0.2797433491791562,
 0.4748423700100607,0.7876234866319272,0.5606458300553941,0.7843845879685751,
 0.042243967362802004,0.13007960572766963,0.05815303136568917,
 0.30575366510503066,0.6079434082240631,0.10246550324950754,
 0.042475800858005286,0.6244574156382844,0.9481707503633137,
 0.4697025967566467,0.3490690882355414,0.22990392568124174,
 0.09980392351645262,0.8397541424085323,0.2767880898932876,0.1806290033571194,
 0.34036764556254107,0.7078063813525158,0.5112822120793157,
 0.21373458906181808,0.3875169724674733,0.7460294763234313,
 0.34474443373080044,0.5413520151702452,0.012232208397581323,
 0.018510068982508843,0.14080174423513625,0.16759472586352953,
 0.01721694964922893,0.1981782724496739,0.9903589706544276,
 0.39138033716751697,0.4518543336789792,0.7235399392425615,
 0.13285375626636664,0.3147333411713641,0.2641804458068364,0.6089038100885978,
 0.15925089618512356,0.7227991558867065,0.42402637484735217,0.688302517212075,
 0.3496595721494492,0.24288875788165998,0.6475020082138361,
 0.35736269072112115,0.3175726258255238,0.6723491960410201,0.0700087209880772,
 0.37387305225555323,0.08588361839430114,0.9035092118343843,
 0.6719445989550847,0.13271138793344028,0.054550006467812495,
 0.8562731730636683,0.29654370865898616,0.6174169042923777,
 0.46857701654135253,0.499171950738418,0.47192102310672734,0.3807448452937757,
 0.37621923805405477,0.4642866051612079,0.44140253748250013,
 0.49126783335234825,0.6947968016318951,0.8990965922407965,
 0.39740848873275025,0.140775125747593,0.12054133274458068,0.9082352638219644,
 0.3334397401020657,0.661677101238128,0.7678847189047384,0.5232165325132944,
 0.8819571307247771,0.19652440386031145,0.020088530098025137,
 0.04629990272029916,0.8206844070963256,0.909605296881165,0.6248060901053203,
 0.8965358547294016,0.13045381579116677,0.37553581005002923,
 0.3418705524028106,0.009257779818375056,0.7503201063572383,
 0.06580281733056792,0.48719309299939806,0.2110636384395761,
 0.9491103595682011,0.7895320062522755,0.09565677219770585,
 0.18877399619558555,0.9857006770603262,0.7422307054414972,0.9438519760578331,
 0.9160790649280152,0.9110072627141396,0.15590577772956393,0.6416719250299708,
 0.1596669527196113,0.11530159615615143,0.30151496494032515,
 0.9065595741784864,0.5061679406879117,0.3740223221022143,0.4789238323708762,
 0.46432072105466526,0.7735567738942871,0.6769036571614707,0.9500106466088186,
 0.9411065993935678,0.09464060009148856,0.9274666237307687,
 0.21949761469541673,0.3196321221655277,0.2363014165498455,0.4285638646669273,
 0.5106425558759984,0.7915628061994378,0.620564912494634,0.16566717671699915,
 0.5824401662116616,0.05493086330804031,0.25846247598852123,
 0.6350122110402863,0.17345803972868157,0.3264654840293294,
 0.21497334871428941,0.17850265862879355,0.01681370918962477,
 0.9980849671501278,0.8318289086961478,0.4542096149292212,0.0876180032061209,
 0.4426048416347341,0.6972225240141136,0.5187793267511477,0.8133103283102286,
 0.02992600700248793,0.5333661294132513,0.06287773963679466,
 0.3944649958743154,0.4864304567891643,0.1964294192870053,0.08398768819841262,
 0.7607846326777883,0.8862726428194384,0.2516586217628728,0.5087389435355574,
 0.310815508398647,0.6057120512605306,0.9557713746001618,0.13133267512065439,
 0.7117533790883102,0.1313124046430374,0.2552802744728615,0.11053339797812922,
 0.09084554621141461,0.7943494491836949,0.6857658328529614,0.2215435036158515,
 0.09639193106855093,0.05871321692382825,0.36847931287129376,
 0.7853183933520167,0.6382230824725023,0.9684605443131503,0.4242443277119834,
 0.8826245744876744,0.42194691598308165,0.8858983676824903,
 0.08405952418140694,0.6083713633299659,0.9007897858234157,
 0.47402523359793625,0.47037556444319106,0.6847478745108502,
 0.9289315375322482,0.40781624749851886,0.5568171755572416,0.1863151011385784,
 0.04666838769334225,0.29844724336793116,0.7274356119563735,
 0.18270487907373023,0.25030263410760945,0.6311213747459868,
 0.17009789536670172,0.7069156792566974,0.46734843937699444,
 0.9939202160951754,0.6038621246973089,0.723585876032511,0.42008763288271267,
 0.19860836146275573,0.3210513528722082,0.3377755275846085,
 0.09356541288528286,0.15612588288100593,0.19548074854643638,
 0.34429923081009295,0.14401262511902302,0.015760053968750354,
 0.3440759500599555,0.3613813977727294,0.28463601398057714,0.8310805404040322,
 0.5776040698973007,0.3578215395768417,0.9336266500700486,0.6437887424728926,
 0.6984757166047848,0.0481663954217163,0.5599878130390429,0.6879707595067217,
 0.23928907283195378,0.04976756244421132,0.6188298371510278,
 0.9611617172568901,0.5386182386259615,0.3089337320958997,0.472818076510108,
 0.7247938850517557,0.48979648864905,0.835448978850849,0.07872374294385032,
 0.34866582465582974,0.2950210116568627,0.553211373346432,0.27918933052602346,
 0.4146227928506401,0.4706544957953871,0.3068487839993468,0.28734826504128463,
 0.6621778355451382,0.6227602154098336,0.8109920050596652,0.9205409100535532,
 0.07177695632378389,0.9491961828735502,0.9107696811344437,0.4089446728440512,
 0.37233162034747336,0.5774545899377871,0.4116045361185684,0.4425575670128117,
 0.6062803153431303,0.2552172420015755,0.5153308126779215,0.5894204476426801,
 0.7986581273929031,0.8405615471086123,0.9772890098136573,0.49434199526031,
 0.3946382134162465,0.5639671625246078,0.4176490824460116,0.914695269949483,
 0.2328839875385864,0.2465701092301119,0.1897134729006008,0.5717851787259628,
 0.5786920265397644,0.8007535609430959,0.9720231736196825,0.9209646462787957,
 0.48460056240836136,0.7837938059635102,0.8274889907057368,0.8235582914189876,
 0.7545830167924908,0.35441316672652845,0.4276665578536252,
 0.004083756484408285,0.39486249690348585,0.5435189475452127,
 0.29483779689452805,0.1989078111498872,0.7456997176044369,
 0.19710320094018907,0.48232996810692685,0.24269690030390267,
 0.7941693098413298,0.9782074295146069,0.9645609187844432,0.6009028193752577,
 0.8715549169948076,0.3877372684876986,0.6641768547610752,0.12045434336437999,
 0.16667126862313575,0.49835946179462254,0.6443613266642554,0.526856396371902,
 0.7329091044802482,0.8111318208545376,0.41678405558751674,0.1588996726081866,
 0.049037449805684474,0.48185173884713417,0.3167305902095,0.9162763756776086,
 0.677736756492749,0.24847428025582408,0.88874924525214,0.48259792294884674,
 0.8514747020019912,0.28397608209739367,0.34446179676951316,
 0.4703203572787381,0.7586294922020247,0.27121754694524125,0.5705951838649426,
 0.7606570040194989,0.9669054601132148,0.9388895831573073,0.250641557999005,
 0.5018965165295702,0.5182182386381456,0.43869620000794907,
 0.39964195518094137,0.5091908474222985,0.145612885311038,0.9859301294858384,
 0.14747230628476993,0.1260727688394745,0.12372851923757766,0.135277283674481,
 0.845762706441409,0.8006486553577932,0.3463915697131914,0.6965033972370502,
 0.4376329932436275,0.5731436196099549,0.33652115828078066,0.1414877732947325,
 0.7306325341404178,0.4399427946096012,0.4008768523731321,0.439235341968785,
 0.09519015522680874,6.038088633087968e-4,0.6478897032865669,
 0.46029053246161034,0.47242911503776563,0.04500905437218283,
 0.9651649848671214,0.08167431597643837,0.8956037852376748,
 0.011209979115811897,0.5002922791974829,0.24105732565107774,
 0.3562026121806843,0.6405019885055507,0.07597986969559611,0.6439530459180188,
 0.5812860639082067,0.03706145920149739,0.7523688429535558,0.7772451732001175,
 0.07340902386333936,0.8580683645160898,0.28785410608606454,
 0.06031972279205444,0.9303525333845167,0.3231665022526068,0.3069022956279256,
 0.5902894260974414,0.3105383075558166,0.27652473799108956,0.8105709446812182,
 0.19278085180456483,0.1875051878007321,0.8968091245957639,
 0.17909844979285516,0.8795192496263599,0.9189306542596591,0.8040659258735655,
 0.13592001579066781,0.7304072640420457,0.6590427243598719,0.3432969957686517,
 0.19197046820475216,0.5559991474760453,0.7506063385444646,0.2508075516624372,
 0.7524299698712871,0.6267211129045145,0.4140415604276486,0.8723212007240913,
 0.9512039926887343,0.1871901020329778,0.1995164443389359,0.5967843130348889,
 0.4000566930172862,0.60385372857645,0.6963846015580868,0.5916000448501662,
 0.6728498828392518,0.9239129194893457,0.0497376063862377,0.8576776893791253,
 0.5085095467502024,0.22851255640823354,0.40033890624445156,
 0.12371138947013549,0.957187145889494,0.47264270369723993,0.8131602821097058,
 0.867348731083748,0.9844121688319576,0.9193402973338639,0.7098263778270354,
 0.833081236633467,0.5008755233457032,0.7058844350829059,0.21619363027388527,
 0.8305689508318562,0.42827527663698395,0.13063822769128142,
 0.3773513937959484,0.5624318567984906,0.30658018468375436,0.9775228798861797,
 0.27279218679753126,0.10316740926537205,0.34825987706167183,
 0.048312121482868675,0.3086637455912986,0.9991549595373284,
 0.06644493606852153,0.21570403777536662,0.7333024164733067,
 0.9752733383225065,0.8297419647182888,0.5290688469174873,0.6215566074102522,
 0.5252424860910023,0.31986173033465126,0.4869826925439401,
 0.039478470812439936,0.22044265328944745,0.223744919469399,
 0.3245449940170735,0.2152112080517139,0.8699629182594414,0.17075416376872565,
 0.3946730020920981,0.056008877413645575,0.0909938426817094,
 0.6237146234536028,0.2697742152104261,0.62029641202235,0.8488198607320241,
 0.037912332885943534,0.7387975035090584,0.20107734108937947,
 0.7188619305366241,0.631525539472222,0.7961000108917345,0.23023139797595893,
 0.136379458343438,0.602683686832982,0.13104678758135757,0.47017561621635784,
 0.38295039087364735,0.31386387602384946,0.17813431936939916,
 0.1590928139170099,0.3193444661285342,0.13434027272430915,0.7988204424538328,
 0.7186887556107928,0.2471298386729598,0.5005564966623979,0.86402768543702,
 0.9892750978907282,0.6498407523245407,0.14030696623124883,
 0.22433618504237052,0.004662724760791548,0.2930408348682413,
 0.3132688703560872,0.7251846028532788,0.5168065317896112,0.13686323008898493,
 0.6877474938769382,0.645450770718419,0.33842396473541114,0.7857689403692906,
 0.505718170935476,0.43451320916210145,0.6579986816913008,0.672660129708395,
 0.4264521810490157,0.3161787138994717,0.5214517234497923,0.7172920248024754,
 0.4342003086372902,0.48757650192336566,0.2955949977043494,0.6793538392796189,
 0.5241116640388699,0.9996925912831284,0.9589801966875087,0.16552719608457178,
 0.23811191218896754,0.38670714837021514,0.9614287639836689,0.380262256275421,
 0.793394294663625,0.2874953502381983,0.1715276275790827,0.8571715532541279,
 0.263405659780795,0.3289788374567507,0.6795714163770441,0.8914275150595544,
 0.44621319825567995,0.6301216508744509,0.8052826109422976,0.8762566694951354,
 0.5949749787081746,0.18549681150849284,0.9723756575155524,
 0.13329130745792694,0.8917414679233675,0.8278933167187972,
 0.48373623452777226,0.6796817209580985,0.43136472174153795,0.688279980405048,
 0.7287269723994582,0.11834223512715625,0.18508224555319452,
 0.7666302752787777,0.47264616250421254,0.6638857818887667,
 0.23401576717544303,0.6951793959132571,0.8274494655900538,0.8833262886734391,
 0.2540222047101499,0.9016122666188412,0.3797479589189958,0.4640924446041411,
 0.8797768943263222,0.3283102951941457,0.7599589358293395,0.3776394243412947,
 0.3300769763542769,0.44844161237882907,0.947680668536036,
 0.011228976441714944,0.784621091228827,0.035012635098576994,
 0.9417618975255556,0.2946507368533592,0.22885306565293917,
 0.13897929634233852,0.35335608422254483,0.5272846359447962,
 0.7980388470754501,0.5241951841110946,0.2072857575015703,0.8934453394232362,
 0.6921941114696901,0.05190484038948773,0.48195601131932786,
 0.011567358829251084,0.30098329564873993,0.7170419097837812,
 0.9387265913757701,0.9554342612046455,0.8390558270532608,0.5586829672837013,
 0.4334203741492936,0.506573506561402,0.6678720764399749,0.9976486741848276,
 0.6096979080507987,0.6904257968607064,0.4776892826334299,0.4840377176110412,
 0.38676480398100266,0.23397559537963364,0.1401741511986877,
 0.3057771695119811,0.42712911865978864,0.1345311594553772,0.5677256885426695,
 0.8157802726871304,0.8853789177295602,0.3544998318399094,0.05158627944539507,
 0.571916232831192,0.42177768711011965,0.1162342847671971,0.8504610587026274,
 0.6970425391580316,0.5797530449214828,0.7515862648348176,0.7748843834500709,
 0.7164787189439943,0.3458774062359994,0.4689238915010632,0.2502433983954203,
 0.8658792756460857,0.20326171921811953,0.3960566375617479,
 0.43201021690523933,0.333638454228534,0.7791785806224213,0.9163064408285818,
 0.19456484037456767,0.7357463816396677,0.7988049000897853,
 0.10360260832250268,0.2522891695161459,0.030367093769197995,
 0.712583112003737,0.050843155352591096,0.47313403297511203,
 0.1860700706695837,0.4031102975444838,0.03720026876528726,0.5247267946466507,
 0.09666603086357595,0.9602547563263282,0.19351088395677252,
 0.4271626652718242,0.6276496956774671,0.8813464329940516,0.2845437025813935,
 0.21410252273531016,0.5049779666024484,0.3178769256901378,0.2873994154658145,
 0.2240437568107626,0.31274238705060087,0.6499579835523458,0.5906475989738527,
 0.6633460470420686,0.024090369240837517,0.1888030832027432,0.557594481703995,
 0.5199421255718681,0.1802923673576985,0.9441442793197082,0.2810565424094691,
 0.012747204749652408,0.6795951153080895,0.712155535770837,0.5235967822435157,
 0.8401605196606701,0.9142562046600675,0.20088042209074897,0.6781352136034022,
 0.047337632606379576,0.7652252504148199,0.8951430221871983,
 0.11871209215962442,0.41648109740917616,0.4282808082490185,
 0.6599822819801947,0.652252219789266,0.9500290872760794,0.8955230142825805,
 0.12225706758695509,0.616669166831342,0.39193902187027463,0.8732200200543483,
 0.9935897603300307,0.16038149230636123,0.3078367442298914,0.8890413968848185,
 0.20587782011326072,0.9892793191404361,0.9911265908392196,0.4592779470766991,
 0.16644216343172957,0.6475289212159003,0.34062036058584955,
 0.3390445343754033,0.003141807760301818,0.8171563040307174,
 0.4686849012221974,0.9328404468764725,0.5892645686252783,0.20692810253678928,
 0.2605431539317563,0.1542111221909156,0.1934666334866314,0.3918107002352288,
 0.7431219631814677,0.535247207312524,0.1701468904196759,0.1489996625208313,
 0.7995095289751188,0.3939384672140973,0.31969467608033264,0.2505330123932754,
 0.3483085919572467,0.31911701788748126,0.020721377120688556,
 0.2906046091214173,0.24656995641022628,0.5155434148293377,0.6951557150628813,
 0.9488721803563699,0.12026362745784425,0.33164057913812817,
 0.6404938022980283,0.9735927200382444,0.5757451688257937,0.9157047285659701,
 0.12823879231849955,0.8270808961602224,0.055294972182443725,
 0.7037759905902736,0.6816716593481236,0.04621705175462487,0.5194867693954911,
 0.8861715328369626,0.7699003564255846,0.026637737231612335,0.343518577716559,
 0.025542336572062396,0.07744997347706173,0.9782674435200848,
 0.6899441227569192,0.7947123232950777,0.6238428489895022,0.39007818827331253,
 0.9744943426052172,0.17012997750744108,0.08114364354745984,
 0.32988935610142844,0.20304472475565083,0.8504659704833472,
 0.34301236968263316,0.02625957970911752,0.2385136181872649,
 0.7202404282534249,0.992192637246502,0.31985221664550245,0.8700820402620714,
 0.013938717554622282,0.652179580542794,0.36344340358039096,
 0.7731917053027261,0.7227676196640059,0.15973727798833692,0.8755777115875876,
 0.5161183643009378,0.5969495466215617,0.4892200724274758,0.5238905689675715,
 0.27711239718213587,0.902598471414773,0.9007355375321052,0.632864002501642,
 0.5216877396898688,0.43912029287799115,0.29976671008911593,
 0.12065009143118932,0.7600355362072827,0.06180877482328073,
 0.9682612496685242,0.13484002840297538,0.2997567617819301,
 0.38949950692214175,0.4132936446924499,0.1274742866744738,0.9167557201027575,
 0.299000666703057,0.18576058918552618,0.9798504418191627,0.16854443222277016,
 0.6466939804792997,0.948221590315516,0.3062860965184808,0.4511601348040529,
 0.5050990655889162,0.8476887173395228,0.810497366087651,0.5902220289628687,
 0.19210213278821242,0.9318412188718412,0.43654844451843455,
 0.6744708423403195,0.15091738596077786,0.39603843899249247,
 0.20067509260569072,0.8601727517207636,0.41427670441195386,
 0.40629664330383675,0.5205661806580204,0.3700327605869552,
 0.33096933093670633,0.14228339352666053,0.2855255246838355,
 0.7654155725188736,0.5966283757044152,0.6430236151975809,0.8749814411540556,
 0.5219696565035369,0.320487752129068,0.2788271404838336,0.8697194177005231,
 0.573602600925686,0.03999313326400855,0.34547895895896685,
 0.35519942895274026,0.5274123576571369,0.1078517795027868,
 0.17908392451870092,0.8295432418034319,0.7116061437320986,
 0.03276828695293954,0.3414292549366892,0.775779268312724,0.03124612944856886,
 0.1918273169017547,0.8831832464152797,0.909641432242987,0.36637076740118457,
 0.11423131300393663,0.8362855234418918,0.22746832775394044,
 0.8534373763871026,0.5997255222432287,0.7684281192746886,0.8303913469537461,
 0.9587974578313817,0.06848114346854106,0.07086054999375291,
 0.5188005103879259,0.7022726781779388,0.7954354227569045,0.902295993430914,
 0.08990229340453249,0.9545941063891414,0.8656858398902932,0.8288053766959982,
 0.10665215192242938,0.4967036607142896,0.5143576929331208,0.9948485051546226,
 0.7292746257599007,0.5237874010070747,0.5545929292538057,0.42648794508693455,
 0.18463393617559476,0.4037925723331083,0.3288902735997026,0.6429799742073349,
 0.2941626901297939,0.6179249389228859,0.9460585007112081,0.6130183646320282,
 0.33272601567994897,0.1418267112983188,0.29939528874156074,
 0.19938629187605172,0.8049035130760933,0.46804025205891153,
 0.8020904131373376,0.6849615075034514,0.8763847438866059,0.5803120168787139,
 0.8700437604797929,0.5477064254194858,0.1575217885774094,0.22143618344256044,
 0.3167762572855817,0.8482434272620931,0.7155159894904174,0.8954012952859406,
 0.8921321197522746,0.3423086143375901,0.6444159458606541,0.8204482850567358,
 0.22116743134437866,0.3115473413829253,0.020176662928861,0.6588192180273069,
 0.02645911875732765,0.601429317082413,0.10362464644232439,0.7345249174184375,
 0.9858741712731224,0.16141793479498534,0.5596787870790069,0.6011113372109028,
 0.958561652987802,0.19949985651013358,0.34947364964656136,
 0.12705133132493174,0.9246291851661429,0.4771507309940777,0.8044778892016287,
 0.5126885256640447,0.38488287560998813,0.47800934266066786,
 0.31275613261559787,0.7061110551935029,0.034794069358227486,
 0.23718025060524184,0.9550501632119551,0.3091302299654425,
 0.20429020881145976,0.715053162561351,0.787384950782698,0.33430915072787604,
 0.4147874795069857,0.5144933772330772,0.6790077553926775,0.5722240696947777,
 0.5265904452409417,0.9750655911859138,0.30032705701955986,
 0.23446373104500706,0.5836241399999191,0.5153075829860656,
 0.36487496685467613,0.9517955087291273,0.5238261922736684,0.8275479508030481,
 0.6011372504510175,0.2870775091207578,0.5459446118064797,0.7518314009441663,
 0.21867580612128545,0.6907861382389928,0.8780925840299955,
 0.11489760010149563,0.17014669762590107,0.919621360288084,
 0.08860936885611403,0.8554498200758222,0.8647129853904421,
 0.12823059099172152,0.5073597716428996,0.846954143866497,0.3138416769814756,
 0.8907333537407902,0.9621879070908367,0.3219219622378665,0.3031720905300689,
 0.26637317518075654,0.10224826611960813,0.8402142722752225,
 0.19141967655169312,0.09096779950857714,0.2821332563659634,
 0.8857387047903852,0.6066298452444372,0.5942850284962722,0.2200961531278086,
 0.01551454092955995,0.7509103129251858,0.6761327442176608,0.604004114244298,
 0.7684887650006643,0.18994974063339132,0.16316521617478297,
 0.5998940960028042,0.3644661954903132,0.5547121466592149,0.7476847231178568,
 0.8771385317640688,0.16169434375816938,0.8934196095748553,0.4738809518349276,
 0.8822013553758783,0.7126196877534409,0.8326340059335258,0.5651259953571119,
 0.1038689700478147,0.6723839991258903,0.6064077783984542,0.12361770558220109,
 0.641212296607049,0.6978034043048517,0.9777661347362313,0.31955406446531887,
 0.6799129967722449,0.14827103753917759,0.5269198930700872,0.5868314223675606,
 0.6308109848512778,0.4659147779052242,0.16476540277662166,0.2747367664603496,
 0.7898215826569246,0.9000967303259506,0.6918498699689534,
 0.005636457934365291,0.5725186726334335,0.2349437616046226,
 0.07535187936326337,0.5014733953094587,0.5327147795119034,0.3466241300905064,
 0.4333951588869831,0.7075253230403378,0.9097582083687074,0.2810437305576794,
 0.9333661173784614,0.9436205872040369,0.45167960478996083,0.6117308258817835,
 0.1857014145179422,0.39285912457947814,0.9255315268624665,
 0.44706002068779516,0.9553995706398695,0.7831353329585661,
 0.013854832033452946,0.8331135232855877,0.4280743486942711,
 0.5975487744605557,0.8663631708614423,0.391865626297472,0.35489225444273076,
 0.7844637744577136,0.7573498119532387,0.4403482942766198,0.2223801222358297,
 0.1779193089687876,0.5213918669118098,0.11033276871592501,
 0.39466945061225567,0.28208230730228623,0.8261086082098511,
 0.6236262460776341,0.6918428286335765,0.7461284792387615,0.651670085079034,
 0.6023786338298879,0.22825296911888093,0.5588777521474307,0.0898146052139337,
 0.9768549642200273,0.22817193327219776,0.91907369301651,0.8025395853406343,
 0.524710449003961,0.8527751872506983,0.7007585611594223,0.2516271903198478,
 0.14028107833385528,0.730180821573551,0.6847796196904017,0.07361442237317606,
 0.04955838253577627,0.18971166453325894,0.5745260398722746,
 0.40095326081935756,0.558040612762296,0.14080748898905826,0.291119471116405,
 0.6691205245201935,0.6954793452805499,0.33988036381618725,
 0.41372746028934526,0.13645815090288194,0.10130280201447128,
 0.3875633117799263,0.01447436013811354,0.7282054552403578,0.1879354789630724,
 0.14648841672136703,0.6439487698660036,0.8510417058027865,0.8241572206910664,
 0.7287871763552446,0.4453555908395199,0.9476262790922776,0.41510557820716254,
 0.3616080974045638,0.6145414626356915,0.36134494733937716,0.8621707664869528,
 0.001392035983215023,0.5938923087561031,0.8330321857542179,
 0.6337075794319693,0.3281281898111015,0.767600267197414,0.0901661428857774,
 0.25122567780698635,0.09006946821080786,0.7798380306724441,0.36710253622169,
 0.4647949509465319,0.7569493602151525,0.40478058057244737,
 0.17822955602507773,0.27550328954413206,0.7362321244176788,
 0.05197866563600928,0.20604483840845544,0.223437556047129,0.8188912587685043,
 0.5938955781797683,0.22096038773611593,0.3737770085114336,0.7568294552696266,
 0.7900113239142669,0.47230712383752693,0.1538343300157018,0.7238387927731256,
 0.5804338909745139,0.12646381845509935,0.7618253665706801,
 0.004022553275882412,0.2067433933261743,0.5535493439404837,
 0.4482894289430426,0.7374867967812874,0.4851071276006669,0.7754059111833197,
 0.16258829340680592,0.45450955632622514,0.5928606053353012,
 0.1849490485028661,0.32685984581624006,0.029194096080564225,
 0.4223014408143829,0.5155842089872286,0.04770201855530831,
 0.32830357012260236,0.9204722793831107,0.20938904055895835,
 0.05290252790363126,0.3014760192114806,0.6523448280659636,0.8356266570514252,
 0.6990075921169357,0.44152185826200174,0.7239030087314348,
 0.03291650366879317,0.8524802497202295,0.04527372387001816,
 0.4137121404355856,0.8070912264927584,0.03424801207526751,0.9180278581516452,
 0.9007640196530092,0.4122639759423121,0.2139070872639557,
 3.5633013844282857e-4,0.20789012317054223,0.6503255725781131,
 0.7566785347806788,0.10463574775024687,0.5989361210445086,0.9218761710174657,
 0.40615961087772867,0.1562009470084148,0.737426832935559,0.231044269677658,
 0.324905613237361,0.5377068406225118,0.6898309281649109,0.10791864065095758,
 0.27483891303048535,0.16200053984815244,0.3437099272016062,
 0.21340367615723976,0.02886252922483945,0.24789462514110194,
 0.0026017951291473773,0.21960762253103638,0.7055096281808264,
 0.6368496279459199,0.45053370358070244,0.6942242452998245,0.4669367819933681,
 0.6572077095710376,0.6699761754074095,0.8578128667089422,0.6217309501283489,
 0.7961147521596956,0.23058939161048253,0.46045076143703667,0.658106082162653,
 0.7726182428114775,0.37390680725031844,0.8816991965220469,0.8272204996191934,
 0.2840602647864633,0.7702450257660827,0.18710655204526627,0.3928355975383483,
 0.1455308855497266,0.8914947281375539,0.6942388159783984,0.3570949710211826,
 0.9273005985002172,0.6017757342123052,0.850226364847919,0.1369910953520258,
 0.7855644261311395,0.8827529211287617,0.7074523840836371,0.7822107308088443,
 0.1319307557565661,0.7301857108008917,0.789321325326636,0.4099356837525633,
 0.36061745549116764,0.7132223396489483,0.41859694451425045,
 0.053972363677098345,0.7976629700441856,0.13976166801315093,
 0.6120296605796441,0.028554606484584077,0.085629935358988,
 0.41660554391059257,0.49691381023375636,0.50740699211445,0.6147139722574725,
 0.010538913559764551,0.08403632054070476,0.63960340336086,0.7377906533155038,
 0.8892038933292947,0.035439479978662036,0.11649417051091138,
 0.43281794974390986,0.884330641011805,0.8185261238325651,0.9194501880675394,
 0.2996111926165943,0.8370823714196066,0.6021347183398807,0.835822351514296,
 0.4696100544878361,0.6817953487593147,0.526198118020074,0.18451181742572476,
 0.29875550162691655,0.11922187340273904,0.48496469072762194,
 0.011995602646199699,0.36264853642635786,0.664676884194341,0.731414206709165,
 0.6281174081916878,0.6375280546731928,0.33088625503116975,
 0.0026982275590228255,0.30637737885210026,0.5772787679870484,
 0.2897657874141317,0.08475107100722323,0.7148966925632687,
 0.06201287127686672,0.4115744148227627,0.4740790047563118,0.4258679366372037,
 0.05273092803832247,0.015178439438879865,0.9551355710507098,
 0.8999373268314933,0.02127823653676708,0.5744119456742185,
 0.10391381234768415,0.7764135027543768,0.7921679652751767,0.0773326513815098,
 0.4915010087267232,0.7999114126600225,0.43253783848023786,0.7225812525345444,
 0.6147179247945961,0.6720652548969831,0.1698467189385482,0.0687109683461562,
 0.11391063551810932,0.6910529005664456,0.6470295031707003,
 0.16450899998046287,0.8745189790967712,0.6416275270393164,0.1195608125193004,
 0.2826638739319203,0.7162163166847596,0.7607514671680132,0.19840737445220102,
 0.47935767584734656,0.1599859936013992,0.3069504293944577,0.7675607432209642,
 0.8358771455038446,0.4273920257854208,0.19595931015823942,
 0.07262949344575442,0.15192796012634013,0.08199677869601962,
 0.15973777546493206,0.08333545392729358,0.36287397714067393,
 0.718550861777741,0.6391337437997399,0.8917109058913297,0.13458140707289445,
 0.7784460984846173,0.6011986956052662,0.05782325899378782,
 0.01625948005014477,0.18122313432925252,0.7119953005644664,
 0.6574151420319032,0.9692800604757135,0.6742033342662802,0.27254627178919477,
 0.8886445800466283,0.8151906644675923,0.5877544618876577,0.36437350651571454,
 0.7874468631851044,0.49741325163829286,0.08170231332514233,
 0.9783560716495833,0.8058549196810757,0.6337431489388656,0.4640209420624142,
 0.2864408382035959,0.7802355991032849,0.2951834154996855,
 0.036950848232355615,0.8889146832123175,0.3997801585866311,
 0.5563105367537784,0.26823565968491203,0.4535073335281149,0.6308128113039829,
 0.1830803653777775,0.931065249776817,0.2244841255484924,0.3249507398710455,
 0.44026021223735334,0.8836830990293132,0.4227078941537984,0.4245574741474797,
 0.8775000045960039,0.8916517456453473,0.24219572683843726,0.7159757409382463,
 0.6206821810017782,0.20967940441830013,0.3492323633899088,0.6154549042818973,
 0.10677836599056634,0.5298703978974335,0.14695137994827145,
 0.8834491654108324,0.8028125044651437,0.9710790973663851,0.6552085498795054,
 0.6029007603268688,0.9057458058886144,0.8572232820046903,0.7085925708019536,
 0.056378026717480156,0.42777849950289315,0.8543544320271864,
 0.8316962903842874,0.43886793472384933,0.7462507761791564,
 0.23433401926858488,0.9184162361187599,0.885762981128638,0.15137379448609056,
 0.09179225186259787,0.6619226061460306,0.9002603703989952,0.8187539493822458,
 0.027068701430267517,0.25626005063932666,0.32157804374044663,
 0.6029334819749966,0.8789457306057299,0.8114640611481011,0.5581129471135238,
 0.9019049222079992,0.3783617747423519,0.7482615021145811,0.5484153692804276,
 0.8395540893337,0.8827788502697826,0.6693612366634646,0.6721305146010789,
 0.30128011342948524,0.8525311825315792,0.7053018346600302,0.817310514696155,
 0.8303670367045464,0.0603907444340539,0.38292016574426546,0.7188782298037499,
 0.9716473300397356,0.30574081176595536,0.05548716573887624,
 0.7285398190202731,0.5910337131264836,0.48296182432647417,0.1340582222049269,
 0.2610862830263716,0.0011215526285983657,0.5948817328438983,
 0.8741491846082174,0.01828418888612038,0.9489369868019605,0.2597112353347085,
 0.1244965835941344,0.019447471331812682,0.7891623433524424,
 0.9154043080256468,0.36793575879617024,0.8469855471832891,0.7759642739198334,
 0.43567942228143275,0.11806405249786145,0.35635766418018233,
 0.6599392988581803,0.06349456767414074,0.7405197435785297,0.1929042549381429,
 0.11725055770795456,0.3829862910276409,0.9157937095616414,0.8719089174102336,
 0.07945138151695075,0.24627059997718903,0.2633915067193351,
 0.6194755895759417,0.37970724145673607,0.1654653230883386,0.5632926533985,
 0.024939960640674053,0.9884205888973902,0.669783377520519,0.4378676347510936,
 0.5344368411716136,0.4455249517762736,0.186610895889466,0.4050611758132804,
 0.4230659850115288,0.3640857032091678,0.07021633385200876,0.9900374830450451,
 0.9028572507021397,0.8483341561615536,0.7306314418952332,0.19147054933281815,
 0.040989891148126234,0.5248817606993257,0.5411240171147217,
 0.4605053466717348,0.7235887332127512,0.022388758334981662,
 0.5287101041709702,0.948706910651611,0.08791021385017639,0.25578677535447925,
 0.5453145159194316,0.8057041456305584,0.8096545017927952,0.6315111336906215,
 0.1942182202212066,0.42048775384041504,0.39152770564260597,
 0.49248678811525504,0.0343191452112942,0.448120257972496,0.1541224978803779,
 0.7656499057097449,0.32802716553906053,0.6168918610910659,0.9320909797690843,
 0.30198457989137495,0.3052789811809119,0.41694879192957923,
 0.2589460477119607,0.9439628558580387,0.17212945487075837,0.1285607574600579,
 0.9638187696777601,0.49025982840117066,0.4885303125810213,
 0.49884362689257455,0.454013306714271,0.38759774408615955,
 0.29028639079255913,0.4305359784445275,0.8358807213032443,0.9483491206963466,
 0.7401731560334562,0.9435780657787162,0.04906155021165659,0.5382943671079549,
 0.8237870955936534,0.33145922581912135,0.5989800582965531,0.8996789361564524,
 0.987616496575964,0.2724626623366494,0.7126279663576232,0.1683076577102246,
 0.322470536300583,0.7020376930158709,0.9258630855623267,0.9316146827517933,
 0.6539218193242619,0.03863693328732576,0.6237926960603176,0.5076395426548358,
 0.19480774698501202,0.7092214578841065,0.23606143224733156,
 0.46937880639339313,0.66406380407937,0.8409327364679007,0.4495473591909537,
 0.16903197700871808,0.8938041051465158,0.1148471987240024,0.8228992615844524,
 0.07782612372097852,0.023612694595862682,0.4971388185263904,
 0.9796711963912653,0.009930298115906733,0.37227203859833224,
 0.7328946479347987,0.23113940895332175,0.3148857564727515,0.9517155706349777,
 0.06640957590227059,0.8319602866498538,0.08772939930190005,
 0.12988435712303037,0.2685077027325693,0.5064483620659528,0.3203052325980925,
 0.7146697012612291,0.41395741156069965,0.10837245549206942,
 0.1039626168825255,0.6698559438519578,0.8108987822079353,0.9187691121730943,
 0.7985089432407133,0.8587192651232428,0.7674297094236747,0.63289397882367,
 0.7498994513865003,0.0058234390120479596,0.7095508434257503,
 0.1846552960926533,0.7851137711245371,0.39652950418324984,0.8637549873243714,
 0.8770952854124995,0.1346963618521213,0.09807802406042965,0.7589674678501969,
 0.8504195579471421,0.14838587536995596,0.6491805049809609,0.7686554247144328,
 0.056872480699762584,0.24784960691985314,0.0654606965217398,
 0.10486451979011802,0.801888139955487,0.20897148790137732,0.946770347664462,
 0.4839460497464,0.22826229386156938,0.10607086781417252,0.48373371638766505,
 0.4874145936856379,0.36107394544408333,0.148936884400651,0.5375955287256259,
 0.43690722819660976,0.38666559003937373,0.8187132999215024,
 0.9859378430545743,0.22390933401772128,0.23071787062794813,
 0.6147525433647627,0.8533229981673942,0.9639666522332021,0.6775823089976081,
 0.6179561127743379,0.5544885862422947,0.49999877463982,0.25389632112902305,
 0.6924906543814194,0.157668653289044,0.573873882045216,0.5086803725717266,
 0.25988548274315904,0.9308482028139535,0.3010221436106899,0.7835663415389569,
 0.3361190062089551,0.5093819994372689,0.00431944825348507,0.1356698282727532,
 0.21701305085847755,0.46089407164156904,0.02958254240660052,
 0.35172616115078703,0.42890636465429255,0.6062955313614019,0.781727988373294,
 0.812605917498582,0.8940057379072669,0.15549664770161953,0.26414197363055136,
 0.9209091833415323,0.1202730280168468,0.4089813983381875,0.12026284360671535,
 0.07538180995345134,0.27047665714059876,0.2974040049525595,
 0.5693079061335847,0.45386168290688267,0.43889425750832023,0.637424703044164,
 0.6633087517327863,0.6097878272336031,0.07989231527847174,0.9393235268547773,
 0.8335926915127063,0.5454389929202387,0.2329727671435623,0.21988360966543574,
 0.6756273514748672,0.29236302422325333,0.9214794213598384,
 0.008540917626221445,0.8179938438606261,0.29156885344523265,
 0.694268220627512,0.37675242281920185,0.6225279828368235,0.4562925648479945,
 0.9179379962612538,0.9624183619627774,0.24749225106172468,0.3183467604523701,
 0.13079609446795282,0.6545744647712528,0.25890742407699896,
 0.5597372466622306,0.45957393100207755,0.12939303828538062,
 0.2851501212648446,0.5210958498554021,0.3401671305615436,0.6328546340298195,
 0.9604496428186718,0.695912532671985,0.48232291114289905,0.3027483076826448,
 0.9366578469536828,0.24322315677823414,0.0036178664292230156,
 0.5286457343043967,0.4384567335176759,0.48111839535211454,0.80010756644173,
 0.010288687949608799,0.9405064132918097,0.2876179979648451,
 0.12192443381808749,0.8140053507564482,0.6235526034201335,0.347683522827375,
 0.30649569459313586,0.10023229409842427,0.01807655514343043,
 0.9313434747148983,0.340186361366107,0.9276434850421855,0.15346108212646714,
 0.3177709244404876,0.8149364771744627,0.8109304367348945,0.8605598757857699,
 0.3051870547209552,0.11098326666272529,0.25079785786145614,
 0.8985673638372849,0.7042980285451218,0.04420147828113341,0.5021431058836341,
 0.7305875033300261,0.9953055534943546,0.5381242633659921,0.926371706193662,
 0.07475256465568703,0.4766217421063832,0.37496605000119654,
 0.7305204726920091,0.04322259335948275,0.09846917103699915,
 0.7508893888982574,0.7239059215018091,0.2344743395660348,0.5051185502061412,
 0.047836813312432835,0.3975719527762587,0.49354425762987586,
 0.19760935885319542,0.6243034958698307,0.5919437703280208,
 0.21946304116683035,0.8056633010012911,0.9176270087827656,0.534804408978899,
 0.14047099686337183,0.7426963920380911,0.28359326871250246,
 0.5455273736752576,0.3656272211786097,0.8567948831532834,0.8011935620848594,
 0.910117074153257,0.11649044460878932,0.38221175316740175,
 0.27307104638987023,0.5985013917022232,0.9423319386404675,
 0.03968143059962925,0.7957735287182721,0.8806953927093569,0.8646329815768615,
 0.9846971965349649,0.49170560143867537,0.7766259415156718,
 0.31600453964678166,0.047697157674789326,0.7666143986234164,
 0.2990197998242018,0.5955169112917854,0.02877855338646207,
 0.20590524549993627,0.30889102647249167,0.18196386670040698,
 0.4462567480082805,0.8002161970404064,0.6370751266896317,0.1268722123636562,
 0.20803218087501807,0.29117765726359846,0.7984170339596788,
 0.7319349419411292,0.336453329855378,0.6331078983326516,0.3086436843831504,
 0.2670329895786939,0.596024638067956,0.7497934900927252,0.17506741734928388,
 0.2045558267499178,0.9906012534686618,0.6127164504321383,0.8748467497734765,
 0.9743289084157447,0.8650771502422452,0.4841740301144438,0.6341541307724574,
 0.9554081057422685,0.3743725757195272,0.88957023721575,0.2948315337953793,
 0.06101949104344784,0.5668444784622022,0.6204140612924537,
 0.49028031491952273,0.3549573299354847,0.43458689642739423,
 0.4377171111905128,0.07326463057846477,0.16448575963807288,
 0.5775278095454721,0.14176445384093572,0.8338693116725694,
 0.17074900179636865,0.9162856579037436,0.4116183834774617,0.6743372927579542,
 0.21946462785142296,0.6865848475853736,0.9750079713372788,0.5212550477684312,
 0.5004937179670845,0.49542035784859895,0.41007869450502776,
 0.7664287563627442,0.1929408600705329,0.4611699257817683,0.3921988272867847,
 0.9528807680764751,0.5129556104986166,0.6718313775153688,0.44902421901104883,
 0.8278558537201957,0.4978978343876561,0.8077481038832901,0.27654240761296667,
 0.9453329526586283,0.707950895287242,0.5560878509027594,0.039434119525385,
 0.9663504744541576,0.19928061393547036,0.4244924312161561,0.1705465889544301,
 0.4331578297678148,0.2164259082676533,0.9968586890555802,0.5541893750314574,
 0.8008795510731681,0.059046269494331005,0.10406381297710876,
 0.6921008158741505,0.6639742913122948,0.7305974069779909,0.9941569858647619,
 0.8425052566048006,0.1575170116738619,0.6489663217694817,0.5049288180653791,
 0.20716043762323388,0.13538017934154567,0.35316211708384726,
 0.37933575389394714,0.6202490715939009,0.2542998069769091,0.7531958013429073,
 0.21838450255308672,0.5437331021881268,0.695480025261022,0.1376485120022718,
 0.4118746970340438,0.08917582655877265,0.6505555373202286,0.7477691042460374,
 0.9380828510147791,0.07010238445526251,0.17967492826748643,
 0.4423549893898161,0.943584305925665,0.8907667393572859,0.8110045943332717,
 0.3906474528840119,0.586374410939722,0.7749290997568031,0.12474771780442417,
 0.5506648025802655,0.9059569495691109,0.7930133775328906,0.5397597926125058,
 0.7327794397233351,0.9475267982392417,0.25103416689158253,0.8443381500059576,
 0.8853319652395957,0.7052363467471112,0.24020021708800343,0.1880046909737152,
 0.6875040691786746,0.26018705726191793,0.2155518333434958,0.7706221461297955,
 0.7197606322509806,0.32404193271120585,0.1233556984009726,
 0.013109437405989799,0.39861506612813047,0.32002623179856005,
 0.4312458106823993,0.2885153898182202,0.9436447163505605,0.49067523710861005,
 0.1541402922021412,0.6662592085039526,0.6619718623331827,0.7131493405262523,
 0.8015449639778092,0.5466302855331993,0.5100290277653616,0.6915842515753886,
 0.40613058886799225,0.9226124600732137,0.9130304764237717,0.4524454274763562,
 0.4075337211782821,0.22080521299813283,0.805060232661205,0.3656720915778313,
 0.5640048611161449,0.05430743268965055,0.10225827835472578,
 0.6005407195139455,0.517911692913983,0.8593575565828477,0.7928920717358396,
 0.48825960405460334,0.7009002353248293,0.345908338481801,0.7381914722768539,
 0.928048822325581,0.5748511688871071,0.8210963654065843,0.4291102585611848,
 0.11455965027214932,0.7837767575791827,0.4842992105926712,0.6470865416195928,
 0.2733630160856497,0.1278052309939196,0.10781609883540244,0.8444289949468117,
 0.3052250402136062,0.6590785750218966,0.17399310376427413,0.7203631778583353,
 0.6698437289193185,0.9344007137833269,0.568849599940668,0.11271400056595038,
 0.8120135237910802,0.9319444962096155,0.29136844063226863,0.6190382398918792,
 0.9471030012991107,0.6484634529396329,0.22445054080022653,0.8457286282251696,
 0.42643021629011024,0.6304616374885462,0.47608431173005483,
 0.10027095969852007,0.3345899774622989,0.1173944798614639,0.3512240151297068,
 0.32498319736649894,0.8129060076060245,0.7407285702988191,
 0.23343154628859109,0.6632979716002417,0.34527064708634625,
 0.8570532058468888,0.492998789004347,0.6789239576541739,0.3762206919824642,
 0.17652356396071256,0.39201590973907785,0.3442499627459027,
 0.7199757749097403,0.2141119421214135,0.6752660539418642,0.758774881412152,
 0.6958505633103952,0.22566923111841297,0.9930407247703269,0.6879158368729651,
 0.10908426251314818,0.12071556030226971,0.5304137410991766,
 0.4070920867473773,0.5007858566615554,0.7317076108608185,0.05473300946020143,
 0.07505353064453213,0.9140186380882919,0.2743300575415478,0.3147881117328427,
 0.625911074566293,0.24163859262706056,0.4481932229818206,0.5644568008472286,
 0.9795001379172336,0.5532934656571503,0.12288987268277352,0.9274962181122536,
 0.9009896409283605,0.06619036647645693,0.12108087808338952,
 0.42873396905892114,0.47002663140161405,0.0312903375876834,
 0.6035709873150872,0.03199765084686701,0.19346818423184053,0.696845403909693,
 0.749481993004063,0.28122272623340094,0.2549253517885809,0.48776215741176354,
 0.9475594531273452,0.36178415080220494,0.8648201727577777,0.7492650746782943,
 0.5609209539298347,0.7138770365436948,0.5827838412825372,0.8738220768287898,
 0.13299217195114144,0.4864646726247177,0.07917704002917325,
 0.5083897221643228,0.8984358255868206,0.3388443664277222,0.07987385224928734,
 0.10654577029728884,0.23560265468818264,0.6400162534592093,
 0.49057107013916224,0.5739473466953591,0.7305109839592944,0.920210475720791,
 0.5456274564411591,0.3039963634660414,0.7657841880145987,0.8028507063376759,
 0.25606843370745525,0.896233692867101,0.13277834725644422,0.937540443054163,
 0.3785351377311237,0.7990430974979996,0.9001165221070526,0.39425997654512224,
 0.11365305248497559,0.6397847135376786,0.9108738414781982,0.9916704471032363,
 0.9290354751484221,0.3247757668599285,0.9688848612721004,0.7499731644515929,
 0.5040324713855792,0.1842317558553418,0.13074703038514424,0.4130681059455228,
 0.27620784480943406,0.7832828587849784,0.9299007241887848,0.8053617757561867,
 0.8712187449556864,0.9046642837197377,0.1199576606633298,0.3808396872307319,
 0.056478277535670474,0.04970291863970944,0.5527582788116014,
 0.22920585203130706,0.728528675904151,0.9306484696601112,0.1308877118287659,
 0.83574106836928,0.7390753228695768,0.24243687127326174,0.5932881608464539,
 0.6647394267360559,0.46087607532608876,0.9581252948684275,0.5392555791188115,
 0.22751018234272236,0.8435101124893949,0.28127126403869773,
 0.9170730784980493,0.4505696510860443,0.6278239156716995,0.25222239818918013,
 0.7009829426267407,0.2141464969183997,0.7820899769495008,0.40864104952522307,
 0.8189579307456869,0.5029121704901982,0.06604396038145977,
 0.04691109042325037,0.23160003665063522,0.5231226352897911,
 0.7224120833817875,0.8409300605761806,0.545365801370408,0.026222394970417895,
 0.9115721591719029,0.5630323005941875,0.8305315400268589,0.086702319874146,
 0.06493497348936972,0.17948811093648143,0.584249831682398,
 0.46368135782263686,0.4507473049189338,0.9368062200417079,0.6637070679992807,
 0.5552990720468489,0.7924415494713613,0.8913873193745718,0.21276662973588678,
 0.4837204591305898,0.8561243021231439,0.12536192961715031,
 0.20222469973815382,0.8243573766821698,0.6303781277723737,0.5742841489425812,
 0.1947958709114357,0.022693447050692206,0.42491799300226385,
 0.41113006595890456,0.7293407762103774,0.6737973518046059,0.4151777525507354,
 0.21637355003786252,0.02557887867772668,0.4252634321276846,
 0.059537178904299104,0.6997323753841259,0.05395017449606687,
 0.5617231890491888,0.13053171553085074,0.590367329044249,0.06087129338900654,
 0.7137275580618779,0.046709199089055575,0.4227769726971109,
 0.09610813713034838,0.9348345729324717,0.24169024047133003,
 0.4353857470884348,0.9240516832161781,0.06634023159127089,0.6070794152782187,
 0.9393975695339449,0.9432591522055165,0.6080653808616853,0.2157640254957317,
 0.22564462421643883,0.29426677119588707,0.7787044158754439,
 0.46445119606379626,0.8879867075446122,0.053588867241605276,
 0.7203101913923504,0.1672194419565921,0.8710425280768642,0.08570522337260889,
 0.8426915817153154,0.042635506344025575,0.9675446260616161,
 0.5978983939022054,0.2709455517306253,0.8943558255019552,0.8465366013662046,
 0.8355088878073686,0.5191489255968901,0.5286504125382385,0.579395061522426,
 0.5090376825911288,0.56291011492686,0.07293055976143914,0.1326913457010026,
 0.7151355432200541,0.6373081181976684,0.6655780912335804,0.1702890075207647,
 0.25242376030077174,0.7423720379293283,0.9121444664396898,0.419481876982851,
 0.5988385273530655,0.7415424570678727,0.8177947519071054,0.10212242723284437,
 0.45528405555783535,0.5000475867597405,0.5561332833751921,0.8077476799646011,
 0.8658167786468396,0.5724223289914414,0.43099511503319166,
 0.38529876742124647,0.7742156933548477,0.3643900318506844,0.9270070042838232,
 0.24351645777702313,0.9407882854928568,0.20758532885487235,
 0.19213547541948228,0.7703035856396626,0.6369905578873434,0.8063835950526175,
 0.9459122774888754,0.2069621150445199,0.2165476911083768,0.7235425661070843,
 0.7433257042513798,0.33702421010619643,0.6553560211119056,0.6642155501990689,
 0.8034184109933765,0.9579400243676622,0.6729484381506508,0.3306599713864069,
 0.18535802857197736,0.348302222581105,0.4713670373725638,0.7683473328505593,
 0.7572336594732578,0.20098710748640136,0.45537017617932707,
 0.9534856166839468,0.20777125478663439,0.6504407327027639,0.2162676370435489,
 0.3769946338964453,0.6674821753605475,0.8321226556898341,0.5731505393320528,
 0.4264681701039058,0.9781202734648283,0.3647355131135581,0.26722095968457027,
 0.22750908717108675,0.6346632713927262,0.7670425797752994,0.6135749359850262,
 0.5866638242708175,0.9332449097376052,0.9980615125319142,0.6907874084666898,
 0.5350108973230907,0.4945253730634189,0.7335345017676855,0.9742201793692326,
 0.8340363158333599,0.8655348054970149,0.12560798731896172,0.6777823788941568,
 0.9079254303637009,0.20819725398194266,0.5161037172326979,0.5111302478139582,
 0.19388041579551285,0.18699196462133738,0.7688369888166309,
 0.09220315684216607,0.5446043159455074,0.5778012098123589,0.8360893690103619,
 0.2982341989540527,0.3824230905774101,0.34716288660515626,0.3201837180385543,
 0.8490223396322977,0.3459860965614875,0.4702366493427851,0.6467789949448411,
 0.8176164029701504,0.7893346928402205,0.05319969526143564,0.9651817355385128,
 0.6564525463017746,0.04061436023510412,0.7763601000804572,
 0.14875187629769604,0.09165248209184862,0.6998579072822644,
 0.12077027625455583,0.15515970416117242,0.7441991639352308,
 0.6199712674290453,0.48118553462885394,0.7344756795431175,0.7352424269635819,
 0.40959064649342825,0.8559242266337657,0.3746359270720241,0.9701298750773828,
 0.5098325208991676,0.7634771064092712,0.9380563126756452,0.3469771479765168,
 0.38605420729797923,0.8080248666607927,0.3987343970716576,0.888771482214898,
 0.12298107206198028,0.6486396304748341,0.7797077726905167,0.511609691853178,
 0.7059971947420598,0.738463865702457,0.8792340698579502,0.21858229566602982,
 0.5341210086441418,0.4603707684561311,0.059624919991924985,
 0.8964284008199033,0.09927439772706559,0.39944326656073537,
 0.8946506032666296,0.7154132673735271,0.9811838995434792,0.2758124567217304,
 0.6533873618768489,0.20148897218157302,0.6266105280486455,0.5471035562474857,
 0.4539568721925783,0.5082264690168691,0.8375809214478234,0.48422090891400216,
 0.5832008976863798,0.08784764555254221,0.7028820673228372,0.9094206911966988,
 0.39930673590218235,0.12448406369160825,0.9462654528257434,
 0.05720001444825806,0.6746063757915787,0.3915592701702668,
 0.056643561984606294,0.6376462116417977,0.2818993677970957,
 0.3209833617566098,0.29755096884351195,0.025115498944179304,
 0.243959951514358,0.25498547321505627,0.5257909022677937,0.9871068929807065,
 0.4031580221688119,0.7433848014150697,0.5673118250548813,0.3233721035052701,
 0.25490053139756386,0.8856425463531064,0.3736129749824344,0.6685142586226648,
 0.7912839534953031,0.8279602011912532,0.7818644004043469,0.5260071865243598,
 0.5355215473807851,0.32909916364987235,0.007265564857486306,
 0.5499745864068553,0.009851085101583834,0.08719355286334429,
 0.6841584560568094,0.2445068937127347,0.31388291940206203,
 0.15967143684304408,0.9705357254117934,0.563116280225633,0.5175337936566016,
 0.5995664505698218,0.8215545397764874,0.7602082922412058,0.7899818812062379,
 0.9160368666098709,0.20441041530484427,0.11289235654700402,
 0.7660520102734664,0.9225853683437321,0.5912196016643854,0.1269063661223997,
 0.5171178435345432,0.05948446745676039,0.9683915489930585,0.5155710789990791,
 0.836683493382471,0.22982397186359005,0.8617425127176944,0.2395858770476954,
 0.9873098869186643,0.045684569604689806,0.591240426577102,0.7648080123249306,
 0.7634357127175939,0.6602549394143655,0.8170690505913047,0.46721011407582014,
 0.4947481652930552,0.35775535004790804,0.22200741678184155,
 0.9279665889135943,0.5705692274587754,0.5503781440852935,0.4874455169293608,
 0.6429321380353387,0.03993055621947006,0.3374194712898648,0.5307729915978849,
 0.0628704631507796,0.5023722304340876,0.7136982554820405,0.46680854848714903,
 0.8262490095912897,0.4927846670391127,0.33023846244132404,
 0.19503184821545416,0.9392709157932511,0.35293323465386794,
 0.18037986447237198,0.9444945543978231,0.028464628178305862,
 0.759735868866393,0.3589206770716483,0.6190193171144505,0.866048022542681,
 0.8148915111433822,0.961001896662884,0.48108147268265444,0.2901835126528154,
 0.18347482286437278,0.47698507188970707,0.35751839965723997,
 0.24501985114014613,0.30192716464865554,0.9062712435184652,
 0.1028904273030743,0.6808250027624174,0.6436670892169152,0.7571613562632428,
 0.38674441409980564,0.7886589209916582,0.7952028811825895,0.5341491121100831,
 0.21624458211884467,0.5505492199305583,0.020690928195870706,
 0.15610053497369092,0.6450213025815164,0.895017924442882,0.8279107173575043,
 0.12241649437494462,0.6571422663880485,0.8563364746109087,0.5692688052954671,
 0.098040987405004,0.7733836701809876,0.3779473366414903,0.8836099399336028,
 0.13521140829079648,0.719735064860048,0.031139462676338936,
 0.5568810065829779,0.708802402458357,0.03474444956017453,0.4172690051600264,
 0.01199153974535716,0.19237617175434685,0.02448372667600196,
 0.6091981664220563,0.27185837340904573,0.4082871765160886,0.3071280833642772,
 0.5694598310088094,0.09716149277443054,0.4798595067112025,
 0.47733893443986064,0.8212687367704182,0.5488351655886775,
 0.05898166332153287,0.6862958980187187,0.43786187442628743,
 0.8771711655171375,0.984506832260641,0.3828134050345907,0.6184874395109858,
 0.7542144629685483,0.4159222612749489,0.8203504395888577,0.8759903286745107,
 0.010858284916681171,0.8167108853703753,0.37344710794703195,
 0.7048271230915649,0.13783832306549337,0.9539698475285011,0.4832136020782003,
 0.7712579616054386,0.24650225290233796,0.3689687917618376,
 0.033753977365934884,0.5158445393282588,0.6379075071203493,
 0.7287071086311077,0.6686610593702806,0.6639361265064284,0.14686092863244837,
 0.39466480624291733,0.39569887643652335,0.6026021595129877,
 0.9981380162759974,0.6059667002418225,0.4923660509423097,0.05641050639691725,
 0.5935299200707655,0.6951815973613497,0.13474096764757182,
 0.26158258457186156,0.02597663335992717,0.0689761498537611,
 0.34112913608876083,0.5800122186478386,0.3839168609008696,0.5513395569249301,
 0.5836958297280602,0.5182581466649229,0.3831289807510574,0.5453654355846401,
 0.8404423821867741,0.9692362365962961,0.5410710883007585,0.48111060889698654,
 0.03784970670424448,0.7779807488200505,0.46204734812777537,
 0.06721591180783715,0.21752812160309032,0.41118724024117814,
 0.0950180206095047,0.5193187790161184,0.20991290843028043,0.6699285789040506,
 0.8607964779699827,0.10295776080406893,0.16259747901644295,
 0.9535854222180766,0.1704834533205133,0.7488280920610995,0.917539230491939,
 0.13942126624947604,0.6191255469684767,0.6188351065797565,0.718904095229172,
 0.09397287002733978,0.7601927311215122,0.3876554840019777,0.795838388179087,
 0.022136933189620733,0.25899655789768583,0.21842335753315445,
 0.8216281957377318,0.3471733673493119,0.6722856038928733,0.4277909350599369,
 0.9978131870970883,0.06518913891573319,0.7927502843810225,
 0.23977001493197136,0.43032134530117505,0.1537463008706721,
 0.8969854303416298,0.9448110390939171,0.5442428563184939,0.07527369162051478,
 0.7077200697630305,0.6978852022736172,0.23977329892638877,0.2694946351186315,
 0.11236423456179045,0.6092241741904935,0.05171858752305314,
 0.3030129539571821,0.30958194937968075,0.5927600693196101,
 0.44034233800732225,0.34171519670771233,0.10352391724792664,
 0.0168530234053903,0.255745441132358,0.6943983182565302,0.3008748112072993,
 0.2278889142166699,0.41978363026885,0.5184312379866242,0.3471757452451465,
 0.8537710098519558,0.0029169444007353107,0.45310126989274613,
 0.8118183214746624,0.19767652035575634,0.957732318750774,0.9920676422945648,
 0.08976625317252085,0.3992970863488756,0.6774668311773039,
 0.23646459950411802,0.18206188347775476,0.33762627673151147,
 0.9801374320431252,0.27182168795675365,0.19106381529822336,
 0.33727718335697354,0.6561706904589805,0.9030186112782578,
 0.07066126871282696,0.5671652593410914,0.7586622805914301,0.4555693663543535,
 0.8715397526988984,0.1557145702515027,0.7842495523607286,0.7212139332498249,
 0.3227676616393259,0.2803710600571535,0.38031911630352866,0.5641530858209888,
 0.6922903305108513,0.6155877099257251,0.9362611079175822,0.6394453846011869,
 0.8074624508437904,0.13195344910587126,0.27214695484301543,
 0.7045028740999703,0.85873866561996,0.34022465483525555,0.47660824458361617,
 0.6705155813863305,0.4914194550466341,0.5634415441091433,0.47124637418174986,
 0.5055688612494582,0.30034003970968226,0.6431769722883592,0.7876378243764514,
 0.47416696478908793,0.9469943990552866,0.39785743461590994,
 0.5718177995541032,0.7194714439783187,0.5235326218494465,0.44608468039103055,
 0.8823133670348169,0.861804130873866,0.3823425102683585,0.7864478148310038,
 0.10817423084608346,0.6309197610237606,0.569745754724527,0.1535554637844876,
 0.6273752413328179,0.03143642523119072,0.22492629954723442,
 0.6307230896542833,0.741590285402309,0.5102715530222534,0.6745226814996858,
 0.9224092042982992,0.1966600751977513,0.5345064040964066,0.16340901092451987,
 0.13283453549829294,0.4292918448344675,0.02926709516715864,0.746473950374434,
 0.9669286086449692,0.3230031423998395,0.7721107472931381,0.21923307278726623,
 0.7096997582098932,0.0014498815880721327,0.07680039682956374,
 0.4549673069396257,0.30147357738219027,0.45563288334900176,
 0.5822536974507629,0.9386777647326416,0.7342089525350958,0.6732793218305733,
 0.7407010706331123,0.35889942247331674,0.37396732598786253,
 0.7884162616725269,0.32010311024134275,0.43437955877140666,
 0.9643211890945993,0.0888764197297287,0.7450456054585959,0.8050023599727201,
 0.5414029623461225,0.8975013634147079,0.41735272512155164,
 0.18391251413093845,0.21770891866527908,0.08226742347682725,
 0.9135557403296455,0.30587750106711353,0.13006226320589942,
 0.7490769201229055,0.03639248831490838,0.0797538185187816,0.7088488528839876,
 0.24918825076250006,0.6518598477709404,0.21648601467143047,
 0.5320587915401987,0.5085684655345604,0.008159177189951516,
 0.9984435987489576,0.3825448100632408,0.9458991035013068,0.9897210622617311,
 0.0365362335145587,0.08057896456957414,0.5904658954720302,0.3397318758279644,
 0.8964955210702567,0.3756149117874217,0.9408290764512101,0.4073193622115442,
 0.6617994499903296,0.8196580837944917,0.23550354148527064,0.1969966321123665,
 0.44607542266623046,0.47175953061642606,0.8220674230239163,
 0.10641973565249063,0.39177028100428946,0.5497576052008343,
 0.7014429089144071,0.008340077537446144,0.05436105186509188,
 0.19241310758294217,0.10295033943818022,0.9953603189342731,
 0.5361466247069956,0.9206175042042467,0.8227945501048498,0.00432052508200198,
 0.9748566290500336,0.1555878294565307,0.9583419269389593,0.6242847975023282,
 0.36580221916820244,0.7796025181940712,0.10596761765688678,
 0.3508774888514652,0.09102246066320085,0.3962972923331771,0.6465053237366502,
 0.8369452927058141,0.2445783228925359,0.6560153228259736,0.37677881536415625,
 0.8448037148957015,0.5649548902806304,0.6295425714580584,0.46507584434509763,
 0.6127277194966355,0.27262358160938505,0.555484470863675,0.6247304656732228,
 0.981332033851662,0.9747068989275064,0.7431489331119212,0.19706329804085032,
 0.15954577307157536,0.20982463428209042,0.7438900756696222,
 0.4438450924485783,0.6610423955148025,0.5092335155108302,0.6102142794709557,
 0.3742795537299217,0.5866068353837813,0.16886711535260757,0.3502963218745083,
 0.6795125135205085,0.6975706221414848,0.0019214942984712158,
 0.10976495212106785,0.6540499586123483,0.16396581329861215,
 0.6765375164893377,0.7635944208961798,0.7746471094161332,0.05969464083041198,
 0.3485960709050627,0.5565470855297779,0.28763676959105167,0.942803603148748,
 0.008840803732244273,0.9560276991191206,0.22281066454800513,
 0.16250103882379374,0.5482739273133294,0.01820371303445556,
 0.5784368037933766,0.6611942166108856,0.39368021816368837,0.2972512853461472,
 0.4795017527712657,0.06575374511617882,0.23007997801966273,
 0.8408187127141816,0.2842990659056436,0.28607183368413736,0.5769362388849228,
 0.9085489773744534,0.7048248048045479,0.3696024524216367,0.6944619471728554,
 0.6228393532154973,0.740387175309731,0.573643227348022,0.8180604936478364,
 0.6697421377778772,0.2811234922484319,0.11649549078455013,
 0.22124357419323726,0.38496760155913934,0.7039987234427674,
 0.3437653386755257,0.7256939541316436,0.39832429738920494,
 0.45843612620401997,0.5152176057177706,0.11041918284245045,
 0.12472721026069078,0.11180143229139106,4.7307649912275984e-4,
 0.17338832949852145,0.44190358494711945,0.2316025744431528,
 0.043955683456547545,0.156926782402248,0.7866688660194745,0.8016287006027197,
 0.7454557352150604,0.5935861077474974,0.17902387302544942,0.9638332587962549,
 0.7317049383380955,0.15812860543880303,0.8594380440682315,0.5201133954956256,
 0.9542716688565617,0.3828712511036976,0.7732566260586371,0.39815665261562505,
 0.8498570752756028,0.8676295258422796,0.360608265857002,7.206790108329475e-4,
 0.7277463264222553,0.9692165577474738,0.883644344867241,0.5912150630077766,
 0.9028860211718922,0.9545257365695925,0.7953089680851879,0.18199599841371594,
 0.28030750966757734,0.8450408300998474,0.2278947278443184,
 0.15536305658764904,0.7171051276095414,0.32851908838634025,
 0.8048488584228063,0.40298317919942417,0.5781003257284886,
 0.22820513059991288,0.3535770993220002,0.4334802010489549,0.888224655479811,
 0.07847198736208172,0.04469680300191359,0.6945405933010963,
 0.7653570179809633,0.9435154238090446,0.5029928634629953,0.15302766747011276,
 0.9761647944282208,0.40151573543121044,0.6979955024396068,0.7587175065249626,
 0.8790114958777882,0.3645901510319055,0.3179367037320192,0.7359866348910398,
 0.25726390724849924,0.3577339520560934,0.49303689559757624,
 0.5170206032810616,0.7689928527493899,0.25743520146938337,0.2940945904126039,
 0.08977600994104573,0.002868059156417635,0.10636428786669305,
 0.3243164276903687,0.13068199093474975,0.0070799365313583,0.5295337374895218,
 0.08063739254969204,0.9255402599886301,0.9510079393764224,0.514234308654144,
 0.37011669289951643,0.4105884576815523,0.3025732542734707,0.8295292933216747,
 0.9468789044364351,0.32878783758908603,0.025672787160608612,
 0.5127988500599738,0.496545979495178,0.9196222916999915,0.19995233558885173,
 0.4526739727257698,0.42899074537913395,0.9259474222428379,0.7620229927267687,
 0.6949448395979776,0.3526878654114093,0.792272729607874,0.9393065876051752,
 0.9653458578232216,0.1746395285566088,0.5160452232658199,
 0.022519386085412485,0.05806948215463037,0.15326051594921086,
 0.3154999762882378,0.3120698815580454,0.785862578569743,0.5381610161626166,
 0.20132805087093242,0.7584676993469238,0.5670100990712079,
 0.33263456024581206,0.9479121021554697,0.8014012807337564,0.7728628320307255,
 0.19320723224404024,0.11076716829702449,0.688594024786034,
 0.03806161902044891,0.6520502329723815,0.5454039607098273,0.8057591299646634,
 0.5072763975716242,0.9919217605895042,0.9518558677976979,0.77241386579459,
 0.5353457045652337,0.9491734038715213,0.9265604217510014,0.9050604536946255,
 0.198531814657412,0.13484449203663518,0.39146039916755115,
 0.43889284579422827,0.6012622948249304,0.9493183509743528,0.9646669375186139,
 0.27818155128201894,0.33884443489622784,0.18278383108705698,
 0.41926035044021503,0.6616980098195409,0.3137286560336541,0.6883310856669168,
 0.4592055546496615,0.5793489496073148,0.7133275437364812,0.4521688948410505,
 0.8342447457051937,0.8378816882620753,0.9414417212162126,0.3194937319251083,
 0.5529319431508003,0.5642862340566189,0.7568455604971123,0.7562943970652303,
 0.16492811640126204,0.49759602106862166,0.9043257049373642,
 0.4114761386937015,0.891129773591343,0.24833165522376188,0.7326507827829486,
 0.28056860354404134,0.646140987614229,0.31850441365851834,0.6193752818099645,
 0.17059417390033116,0.09473484580444269,0.7549010270039054,
 0.9818036852188046,0.736149391178897,0.28870164251635844,0.10087326774513128,
 0.26651591065750857,0.48403928720027933,0.7173328283782785,
 0.8957477702495599,0.06601684828365939,0.1199709910498622,0.9041765926135134,
 0.3487120677149932,0.8758306037440671,0.34955535896001955,0.7290272114774552,
 0.18327775195235607,0.9472936375789688,0.7998969149674497,0.6829455444367264,
 0.1673614186568788,0.01699002552205786,0.25957774696927804,
 0.1202765803912178,0.4640754520388294,0.43287407935523325,0.7039420226513172,
 0.2649652540574544,0.010699465811938591,0.31779360646568344,
 0.24698550653787466,0.11269123417272409,0.18628531991277342,
 0.5073400428907002,0.23202243469545414,0.6097964663871348,
 0.034673439316291854,0.7936789837558502,0.34657449703436305,
 0.5973208007191254,0.3274737020354197,0.1836247002531458,0.752667526133767,
 0.29303211018312036,0.7154567710913852,0.7390974450659554,0.6882371369703576,
 0.9762338429367339,0.7962723188473608,0.9801319003867281,0.18654174576901184,
 0.8509420508653381,0.6210080569699633,0.3259114189682073,0.7859282552006548,
 0.4742009426598619,0.005094400834433355,0.9389913122225364,
 0.4186284576199747,0.07740913480067368,0.21065483717463174,
 0.17917903465818152,0.27475038410343045,0.6436120659601521,
 0.7458247911853617,0.45164670419821373,0.07351840968366297,
 0.9517036953893648,0.13813395538672324,0.39708922397660573,0.712278275085419,
 0.5094476904480673,0.8195431989463026,0.48003661280708365,0.3985978235349278,
 0.9892720819888632,0.7998263679494808,0.3581499166327513,0.7520480659978799,
 0.3802299211971568,0.9320036143970807,8.500626531138256e-5,
 0.8071667166373979,0.31495670056503133,0.30342931038523313,
 0.6683057540113129,0.44871017411181513,0.42845842596253736,
 0.6046207728997174,0.4170596352090542,0.8673274745202266,0.9020816112482349,
 0.6841636618793708,0.2523666147670085,0.17986438249670622,0.5145896825316957,
 0.9550846716839956,0.13854595945793624,0.6887497124549149,0.9756173982819947,
 0.8441336573975033,0.9994880216078095,0.21177473429822768,0.5744241846340997,
 0.9475757235798108,0.25854542516160095,0.2596890117402776,0.8301590188277916,
 0.00785126243159806,0.9690312410623647,0.7531969818500478,0.3203379147120826,
 0.1814874501527708,0.056397173442652715,0.9826190522417663,
 0.35175248620069943,0.27599922937074695,0.11596021945643653,
 0.04133628745024154,0.8241011643750389,0.732861772766999,0.19331429292767555,
 0.8063023588035726,0.5325177923564157,0.6845249724530056,0.6259049126758347,
 0.7038972876375741,0.41942420092731014,0.710199137228078,0.49219183177509873,
 0.2836176274260712,0.5993122221911484,0.2624743470892018,
 0.054613338260492084,0.5428902561223476,0.579471206137089,0.393772879999423,
 0.7525813628384668,0.2573706271172468,0.28558023731296556,0.7550989421591383,
 0.8543324956726466,0.9690465912835947,0.19515719545534127,
 0.42720666800947815,0.7923551898015516,0.5314117210317097,0.5517612630118406,
 0.5967428496925881,0.8491468884648393,0.6300155897271998,
 0.019820282861145233,0.4657139204902421,0.899650572205016,
 0.42079928880717643,0.11309974177049631,0.03182428860923425,
 0.9285719168258737,0.35944209073869904,0.7881906403775456,0.4254404686438077,
 0.6026463214408196,0.7369459405162845,0.23600567952279405,0.5093300829731893,
 0.6097680453998839,0.012689853714942023,0.39337552348164817,
 0.6687040558158347,0.13531925886458185,0.7403466492234051,0.7729149273060958,
 0.09688495338381697,0.8659626381433981,0.7496145012835321,0.7680557559149113,
 0.8610917792142163,0.4674537290568703,0.35439469089889764,
 0.05999238424297482,0.5694810288819431,0.038261743204159315,
 0.831408824711082,0.6092528279004679,0.541590008164485,0.9692134542390808,
 0.5112593401008257,0.4543576735021104,0.05618735626951932,0.7837942342397446,
 0.47268631041361864,0.7945773408578216,0.924542207836373,0.8658241481772008,
 0.027506902086020246,0.18670712059011407,0.6335312649764807,
 0.4082337214186602,0.23327879793824757,0.8025871253734239,0.3568827932857941,
 0.457862158800014,0.8608640930675648,0.6043690931355175,0.15513680478883696,
 0.3173148334738436,0.7870782720389438,0.6912903263321949,0.3279868015938747,
 0.9194706296614874,0.47223911935172147,0.5049732925188887,0.4666549203309237,
 0.5872488121120923,0.14832444685713753,0.5801246631054074,0.5284570815034015,
 0.19496086978158478,0.6772671880690081,0.14902503636482345,
 0.6660660059215242,0.45176517138052696,0.7185970061313237,0.275333396276527,
 0.974787791901451,0.8598815495292509,0.2459319684665423,0.10615556540162241,
 0.2978266504907122,0.7439562698558327,0.29556374780158723,0.4275315442309068,
 0.5590679695518239,0.42384628061244256,0.623960144513148,0.14774742892453563,
 0.5980802192594258,0.0010207882982565808,0.09872534108372066,
 0.20047258883864205,0.9234036252224737,0.48734909087546063,0.856637662413154,
 0.41116914328060006,0.9383819202285105,0.9674345739270617,
 0.20633073917552558,0.5915286777774259,0.3617233929153749,0.677655749810479,
 0.4265898610613208,0.18091111933153692,0.255008754729418,0.3482202151040301,
 0.3817714532630663,0.12968311041031477,0.9316227454145753,0.5364530895181749,
 0.8013747751787494,0.6527025925332332,0.45039637594244963,0.3617364335801938,
 0.909825385412485,0.03336232364160008,0.6472386809886108,0.4693519385269326,
 0.09806692194123123,0.4291523313156068,0.6798449591512479,
 0.30162580083731805,0.2589012242948411,0.6984556772451617,0.7702916226073209,
 0.7020035201267701,0.11884932614267973,0.4382095332927668,
 0.14121269495902655,0.18999650012843305,0.31868515427357513,
 0.1879731830011806,0.19879312994018605,0.4202754908567621,0.8613802701794218,
 0.7093468625675217,0.26812912362599384,0.1795377458817038,
 0.23106906543766692,0.5585302208359271,0.8919753383420556,
 0.07954839611227371,0.30502763925625953,0.5422624767242059,
 0.5036949798867725,0.462882033175126,0.37597603265746593,0.34565263869463003,
 0.702270414673844,0.6989334967886536,0.559799690340889,0.2471638962325342,
 0.971802489082,0.6331564754309835,0.39052306083255106,0.9372990785618412,
 0.47370273381318606,0.17437811015850635,0.05520910704558557,
 0.37514252136704396,0.9148181109403357,0.08366081958414517,
 0.30210726559536627,0.8369439130534334,0.5226623728325162,0.3208880758581777,
 0.8537262214657241,0.9231796492687184,0.1534875359501312,0.5591682262048077,
 0.261506356552507,0.2731068453505986,0.03510942818238327,0.6526585901485629,
 0.4108774346918638,0.8686685333474642,0.39448071628057557,0.539621992783589,
 0.615688843181954,0.14156044993866346,0.46277250146902116,0.4054351517925917,
 0.9489367101339566,0.49277477589830276,0.5514555948637034,0.1761256341006977,
 0.24113000694453302,0.5739095101490738,0.9191668546469681,0.6341649184989243,
 0.12013884548628906,0.8359785452730117,0.32653284569268937,
 0.8099224344602126,0.131018764767867,0.8795468739891643,0.5897695563774906,
 0.45430378290154405,0.7242849471074777,0.04263013713078001,
 0.08047379474879235,0.42030145104837513,0.7240435021729673,
 0.5520662602908011,0.09298321392301279,0.28632602549590347,
 0.9369289383247827,0.3410213669581419,0.9126353415117744,0.8823184842374614,
 0.4116661395387827,0.16917349761267675,0.21596232838003104,
 0.4477883588224276,0.9636337956806275,0.810083812262278,0.3795305769107784,
 0.2066788234397603,0.01558142672371765,0.761272362670779,0.7448420514852896,
 0.903341127752836,0.8035392027253278,0.07959363336718295,0.16997661127194186,
 0.8531978334099898,0.08030079171480264,0.8719031054790136,0.7709502019503174,
 0.6957547481589862,0.7237070083519034,0.1863032657208934,0.7229743437982908,
 0.8315675342119211,0.2957277712839106,0.29204744118792914,
 0.35323805210569803,0.2814749968756607,0.4610941688913055,0.9691711685426939,
 0.8087710884083779,0.8491817204442271,0.6379125384906118,0.13053624253152563,
 0.1803158322610975,0.22305285704397904,0.4258293190174113,0.9780584862043216,
 0.5880294624327841,0.4670735724140812,0.857984497531135,0.9753287491387532,
 0.2953427515176027,0.6690719331488643,0.5858836417839608,0.19271330462641723,
 0.6266982799947167,0.725843209196279,0.3259062249149497,0.7742529337706461,
 0.7641048598521558,0.8740398597129417,0.011860970225726408,
 0.9810107654319301,0.48039290095676235,0.9778677777291466,
 0.06261695632433895,0.8782454955536694,0.38736768812560685,0.431712561472095,
 0.7226691797167538,0.15117687526164458,0.8467275448477286,0.9985403602170937,
 0.43048738877924,0.8887245698549614,0.1641453345655598,0.2865552415730689,
 0.056507803297411785,0.3563107014453011,0.8214813355419768,
 0.26852436372608024,0.9970691854448755,0.3306092492269569,
 0.37713865185266915,0.3693846199998998,0.19235864095155508,
 0.6729829305727902,0.6094408661353415,0.5895646209873433,0.6733934285959123,
 0.4785365369199285,0.12775848492746944,0.5397859983669003,0.5530228564026216,
 0.6171286738817499,0.6062796666676339,0.21430715757141106,0.3875984851333425,
 0.3169875114200609,0.7404534948773547,0.8115792849708388,0.4559936740591737,
 0.6405752258623927,0.8737854735918469,0.7217606221778718,0.2594432625054661,
 0.7291829723706447,0.16529959772903302,0.8643079202894999,
 0.22646390843844455,0.76847070399328,0.6199140937440178,0.40659104061841356,
 0.05107607331191366,0.7296065236148855,0.5808601624415181,0.7354633173805507,
 0.06820364565408721,0.35885247544826604,0.530094337637949,0.4001031122498031,
 0.40525682189450174,0.22248239079010756,0.4249554113640367,
 0.15463839848728267,0.7204523802715268,0.24133106921062952,0.51936763708889,
 0.9004486261338821,0.7880317640829182,0.8775352559159592,0.8742773743981553,
 0.39037421801758054,0.8024192719828634,0.10111619121358317,
 0.36321687968597993,0.21155038133502702,0.9767110060795439,
 0.17706908917838893,0.6833533113053329,0.025036923016427348,
 0.5821492569583837,0.6636982583361067,0.5270039821302605,0.22378670343831852,
 0.737142307310375,0.6182152315673487,0.5370012054660405,0.5813288300833922,
 0.7905564164319795,0.306238318534765,0.42939512198792773,0.5230809415363615,
 0.9002786127232311,0.2231335299104592,0.42748239362716944,0.9768741651094272,
 0.6884746060684561,0.8228014451863377,0.9989049775370368,0.9932039025072217,
 0.08899637759562473,0.18662469574832052,0.0630409818863249,
 0.7261468593232411,0.5372431347138904,0.35796927292469083,0.8067544564857508,
 0.41155747629052686,0.12599702420751502,0.6259050051676969,
 0.8346106888424735,0.6452081804329489,0.9276730532665067,0.41809349700884413,
 0.48500896413538164,0.7232292745001411,0.3930337891745548,
 0.18128478700193962,0.814595445280836,0.019843079996175184,
 0.06906646687779816,0.45803578242994203,0.373769574226386,0.9049997498230063,
 0.8541959751160191,0.9649632835745032,0.9121270603408769,0.760587961096801,
 0.23895539941653487,0.4003555496374619,0.4676168879652345,0.4957074410512743,
 0.10477637962119657,0.9280683986270818,0.47857825950294375,
 0.5538689016086015,0.5329522554951209,0.25341131487804314,0.3109838186149463,
 0.1607040749952212,0.3666509489121005,0.08555990438477667,0.688603284844495,
 0.1869702977711627,0.6792153925404848,0.5213998540588195,0.7520353521177584,
 0.912639275097588,0.25380628324115384,0.4515871589428716,0.38910148164764025,
 0.4624760656907513,0.6016658250643571,0.7853125832177571,
 0.005299400769221307,0.6994095614769595,0.5143749851075992,
 0.7171170528221593,0.5975837315629631,0.6080029175493332,0.8247649583958147,
 0.9754934246408848,0.06897266024073068,0.5723240213216603,
 0.02201037170447906,0.9203145126966031,0.14459105957747476,0.873112426862168,
 0.6389849942542875,0.19669354625727054,0.9629979864094562,0.4372854320993177,
 0.16480800027793907,0.09642626984476332,0.502096340216946,0.8020271938825536,
 0.8819410003126767,0.6745184102285846,0.7493310636080691,0.18214150755717917,
 0.07605503647051426,0.9185254597488256,0.6325296189377693,0.6714614659514262,
 0.9292517339834779,0.31376879976103667,0.34065783927542304,
 0.2991709435656368,0.364324091991979,0.26499760062775446,0.28522316972764195,
 0.45166361638601116,0.3661223687154558,0.04941485887260566,
 0.9407501165719948,0.09322185141372308,0.7491352480662609,
 0.07803519118403157,0.39177085889139385,0.002718203458439983,
 0.6534132930439677,0.8679316060892488,0.17781493961734773,0.8180738519009887,
 0.8355467527731756,0.43245042977345205,0.7715622240708904,0.7263360932608763,
 0.5865487086682215,0.2830729216259795,0.39428215909311093,
 0.14748316731338829,0.57437979551356,0.7252860289795018,0.1861544680989874,
 0.2559267000678892,0.6601331519908747,0.1659220840610659,0.654258430161541,
 0.9974327535758839,0.5003853787301364,0.7257736688517983,0.7462433951206825,
 0.02874421279355177,0.8429467171483578,0.18323601800930978,
 0.16089824738942493,0.7609110953228893,0.19105882405133368,
 0.28375624220076034,0.11256490354997617,0.8352817833404556,0.951394653622982,
 0.9167906664325547,0.2581232794783811,0.846504897615689,0.45559943528763314,
 0.8116347472571992,0.38856255411279683,0.06321672427253322,
 0.7673492950123294,0.6714093263924129,0.05122945155777159,0.2588168553893039,
 0.5948001000213485,0.7470017156598426,0.006604319762482991,
 0.6265880473977515,0.9865018611792425,0.2463196143705071,0.4589739071098151,
 0.7846223746152351,0.28706332176593563,0.7389109131852662,0.6845021870484236,
 0.6772402993469552,0.3445671482233805,0.9966337481008624,0.13024643003548675,
 0.2630621803127161,0.4829909327307692,0.024851417946310184,
 0.7622973367749867,0.47115426752958856,0.1527136946273776,0.6982560617417191,
 0.6337986710757513,0.9531903150748614,0.9490825138506038,0.9547816326218689,
 0.7649193910511867,0.6035149961393969,0.5517717651169264,0.28774274533568456,
 0.9317529407501413,0.47503472512523603,0.07245080942305471,
 0.8136543967704689,0.32339958456787254,0.12634368417172326,
 0.010182823113970008,0.2940671230616071,0.9320990371877172,
 0.4402515967137096,0.18183761790953468,0.049415692914269504,
 0.7062193534984956,0.4994727797193129,0.998765879517379,0.6265113585561152,
 0.06642118169821387,0.9691962650856595,0.880299873930093,0.26701627198640854,
 0.38439040543401937,0.9425608804455872,0.5917382418819559,0.8335797005804211,
 0.4694933463309836,0.4641253615034686,0.6836426715123645,0.8568135058841486,
 0.32446643796039143,0.4820105436081903,0.5488968992893415,0.7292842231315839,
 0.8462134872582223,0.33062357204497916,0.594501838334778,0.38065033373306967,
 0.20834215198040773,0.6686487232016678,0.19910491931251206,
 0.3501970107192601,0.6539961897442559,0.38272834025256064,0.6398089108611258,
 0.7625902345774376,0.4251153135026573,0.01782478269681198,0.9001779089017278,
 0.6820738290653587,0.6733542027858852,0.03396642416834006,0.831945031236877,
 0.9435417890876138,0.5555792509411537,0.6120508630380174,0.9793927007950831,
 0.9425978737158924,0.8465043624210363,0.7243361079918648,0.8355877980960968,
 0.31190277552986156,0.9904868885226599,0.7772526311522072,0.8150301068587764,
 0.1594120222341192,0.13957150467235724,0.11846712656026126,0.69930383254297,
 0.9254672244024644,0.12816659499337257,0.16518456386168534,
 0.6289069446057507,0.7783514675443219,0.5729414047830146,0.4321381418650163,
 0.406217924845953,0.029832423120115803,0.21489677694624432,
 0.8539689156237643,0.44809205987181244,0.40344537761469723,
 0.5122188670702503,0.5830720358892576,0.8463800940442533,0.21284820583090958,
 0.5938408532771681,0.68283405527434,0.27628505422944716,0.4693557886053943,
 0.7136231271053131,0.4401689196795503,0.2684649676889904,0.1098383036922429,
 0.17606128140894972,0.9002033822584944,0.0751823617360119,0.2533195602348479,
 0.28315600452576395,0.734887932059526,0.3983237852662259,0.0419662886667822,
 0.3124194190783509,0.8903966577519071,0.8501638944477976,0.8576021178576194,
 0.7254933536579422,0.9595420205798,0.10193240600237596,0.7839472461819474,
 0.8013186606036795,0.9599445364041433,0.27356457984330373,
 0.06525240449480729,0.939751680459581,0.1481833254231839,0.7180512475129737,
 0.7109914600525864,0.07118924027950102,0.6916758759447138,
 0.25837826162326405,0.16844981934881198,0.46278786573485775,
 0.777698875163206,0.6705425748729321,0.9729897648019141,0.7829095492117153,
 0.255717335536749,0.24864911647859633,0.49263181512876864,0.6530835841594571,
 0.057369403257450546,0.029006096625525002,0.8181726429160037,
 0.8479740308472151,0.8784161160131287,0.5435181932063309,0.8229440161975299,
 0.5584820341954538,0.14607642376156438,0.6633785921474493,0.3270816516476114,
 0.3808608945687859,0.2863866681472106,0.46433595071286116,0.4401939197332765,
 0.35629866617958217,0.7593212980074033,0.30981082747312094,
 0.5881054289041919,0.9554662275598108,0.23852905007666114,0.8350144859176234,
 0.674991461518336,0.7095255093309638,0.7454429850812948,0.63181979586049,
 0.8975527860355451,0.43978412928393984,0.7705717994910551,0.47869622647858,
 0.2030890627191374,0.7075866711791762,0.5847086870076359,0.4820427093620797,
 0.9167307096542391,0.6250469658334749,0.5961964844638934,0.2846775352798617,
 0.31141324008144955,0.8290504085888036,0.7469385118450679,0.9476750960167524,
 0.9921321655691512,0.988401537101069,0.3295681816202046,0.7423632457400234,
 0.025600011924519883,0.12349023334455467,0.28684107672128656,
 0.4428601342074864,0.7774814354907138,0.4158425083764684,0.9597830285611683,
 0.8554082398867323,0.8188250748784235,0.45373518496098064,0.7398721944548764,
 0.9805914453406981,0.4879667857730684,0.3399023643829644,0.0548285647517861,
 0.7466726876973205,0.8017653114475853,0.6369984102964256,0.9514415446951965,
 0.06382603980008383,0.8370972398752056,0.785898745005507,0.5776063109650158,
 0.722325691922574,0.6942117925851541,0.10748800681902293,0.9627513578804506,
 0.6407312969343673,0.3777375169163204,0.5826509779531056,0.9489394195242219,
 0.8029997337891144,0.6357736206573212,0.4609154073189845,0.4674238346440537,
 0.5522533681390844,0.30638889051077434,0.8266121608547993,0.6009276989991056,
 0.22650781289954025,0.2972411258334289,0.7517798834065331,
 0.09831937770684152,0.16594267378302752,0.9932126918737545,
 0.15499617095887352,0.7979079220714412,0.03855571007265002,0.591742555261856,
 0.12236115770254641,0.017577907798917436,0.906805687638674,
 0.6838896910439247,0.5710144400725001,0.9536214210439271,0.55921823809191,
 0.23632676565117017,0.8111894336217255,0.807179123136087,0.2377137748539554,
 0.9881958357742425,0.3723829751216293,0.7253436828216323,0.3872773630768248,
 0.939861959207912,0.862638503442514,0.9305551812998597,0.8269681922804666,
 0.003855861744310074,0.6556560780707233,0.31178884123006145,
 0.9074433166361245,0.9030445936726073,0.3453419587426605,0.17942036110005533,
 0.945244692829946,0.10175899754654139,0.7407234238248083,0.31042740438993677,
 0.7090786571300343,0.3028275460357821,0.49664317348185827,0.7233058128509924,
 0.3034528782524041,0.5435687009849315,0.07417612929902828,0.3073405657484003,
 0.16080141494848554,0.5822310938892278,0.4188229810261822,
 0.22963340165161217,0.6280729854789896,0.1887831714777506,0.6008841231574862,
 0.10549235212362096,0.5150626470890944,0.8893013795034411,0.6711946976231881,
 0.8327031029772161,0.7696288418202648,0.4827008248246729,0.8555046211118975,
 0.24490230519439493,0.10773102061671813,0.0015485653350354456,
 0.3466329151571901,0.6801941178905775,0.7471912469649944,0.60093948490462,
 0.8260823823086181,0.6487315151517205,0.6702626326989828,0.2746520559089427,
 0.9312190226982597,0.36060974462233364,0.4649311922460604,0.3729489622600698,
 0.06888302505199728,0.12785869244095405,0.7554149270578784,
 0.23746848763098516,0.4041740507498389,0.6680768442179749,0.2546787535319768,
 0.20091138704541356,0.29461868505254984,0.6069034988720481,
 0.15178165462823023,0.24344686877045496,0.7208342028560994,
 0.022161648584175575,0.14793751947551126,0.022969537701901688,
 0.3875707478858872,0.8781826889827823,0.5494650153651205,0.49788248928251,
 0.9203003638010384,0.4396785569847975,0.7841343227464175,0.9757176617625525,
 0.21785023402633918,0.31146486368978665,0.4365311606158706,0.692395054567164,
 0.08738063506943028,0.5169525162625765,0.19677128951598444,
 0.8443526771131101,0.25370002874973874,0.1640391039195166,0.7347086270831638,
 0.8063978034618103,0.3239579217611962,0.2756965452611517,0.15013695589903975,
 0.31258610346059734,0.20183694863073998,0.18894200424973517,
 0.39558973192538505,0.6139713797738221,0.9528425082088132,0.6023962937259717,
 0.625350167448189,0.12513442710340794,0.06900516146456614,0.5616031593721186,
 0.64597220072103,0.22407270069906648,0.8267764508177511,0.8250038479233269,
 0.45515174565887806,0.3640470139928328,0.5307447683584537,0.514337555771078,
 0.5448675404722628,0.620036828132411,0.4793499176232773,0.127644761815638,
 0.6709590794098121,0.44285348899881716,0.009935216458582419,
 0.6406737880607294,0.5790238591580538,0.20377679059110598,0.2403656643483355,
 0.8479854517341995,0.16143858093778385,0.7838442850850791,
 0.056192864700862044,0.2658801815243319,0.519818804342413,0.4010181714833335,
 0.8878271818421635,0.794553950271203,0.5147002970806296,0.9330741476463749,
 0.17937679490379055,0.9680914352204344,0.6720257401928719,
 0.47075729387264076,0.7974654762736518,0.013080140068752799,
 0.6660677440992152,0.8641737317924653,0.4361266118717715,
 0.060415906064147706,0.40372952241417137,0.5805703813382654,
 0.8660003346115964,0.0034368893426151015,0.8395537689661654,
 0.14871592390729615,0.6850186557386743,0.15715459119164177,
 0.30348453313202517,0.1900784863738081,0.3034533114444452,0.8299992691283343,
 0.015255686681985514,0.6649790081681209,0.7939777445717153,
 0.19613930194795037,0.6561625669233746,0.030731007377683972,
 0.5819481942790563,0.2734904203484686,0.6802120470377077,0.1393933459582124,
 0.7320668945616553,0.8427282521783936,0.3948416142182549,0.7259835474794294,
 0.70189564502664,0.5741505301751832,0.03236485390083843,0.38426358468404354,
 0.3112417993703256,0.5427315064205352,0.942955857207098,0.353647314573464,
 0.8913771338849003,0.7043951764978925,0.8272844269845372,0.38467235789360676,
 0.07084581688774283,0.49184159822362916,0.7782272379947497,
 0.7482583263226992,0.11587544017002693,0.3729098673359549,0.932691458405393,
 0.7911227300806184,0.5169883618198479,0.23438177813893657,0.3878829527519805,
 0.8722966929704418,0.7601832422561318,0.07991719964398059,
 0.18112263121584393,0.9831042837583989,0.47889120265353113,
 0.27792629114578193,0.341182427558141,0.6486751438397707,0.3301693986516845,
 0.7294688130521039,0.8304131022262647,0.34019055943287757,
 0.17871344299065206,0.6906923846888313,0.7304854987268931,0.5054789632247975,
 0.9903221052932887,0.562795646884418,0.11854839424108421,0.14328278734341793,
 0.3884681240736729,0.5897529166524689,0.9033453726945829,0.7537002268514706,
 0.6193114233824223,0.9866343979086629,0.898375727321791,0.7457169356002133,
 0.1718877106483213,0.0013990649806310262,0.5607134957128986,
 0.3196371784775127,0.5946888002113292,0.3709828896070454,0.34749849611024064,
 0.030619357306207462,0.9793778741167207,0.7708376061221189,
 0.6204308190637065,0.6670934793131583,0.9318933223910316,0.7970105919676982,
 0.7844289938758906,0.3182399554326091,0.8837693682361243,0.6945251266840934,
 0.7300338282606258,0.17431741045722327,0.3342588740824578,
 0.35177101051441373,0.87456402804413,0.21351059559069185,0.5385583333001662,
 0.3380757321696434,0.5461671702059723,0.7094313166903432,0.7586743876809321,
 0.6696976313919438,0.30678503871086793,0.7887709650546276,
 0.06691660441794922,0.7562854706594089,0.7578995702594683,0.8091081243072992,
 0.6200506177490281,0.5422591603082064,0.5477126426991296,0.7153367971942982,
 0.43648205824899955,0.4625441800321437,0.4498470998763252,0.6047859059831558,
 0.47589830701277935,0.5670582632350446,0.697477038446251,0.6954601680526667,
 0.06331332652729393,0.4149427025196656,0.9431728580258699,
 0.19251145761510247,0.1285903536667038,0.426051345495322,0.06886923396588951,
 0.3167479910716472,0.820112112533688,0.1206094240638621,0.2582103627965332,
 0.5267126537865189,0.46397546189699457,0.27845127622865984,
 0.21922534604041888,0.5452210341451456,0.4832291247812569,0.0738940780390207,
 0.4128799380210836,0.864844351939941,0.7167843194788474,0.7242806943870073,
 0.22453494239804594,0.6625192068175236,0.43783418292665366,
 0.3291517677319059,0.48364941267551553,0.10642813344972535,
 0.17016670016427227,0.37284901862541153,0.412000383711876,
 0.36596473132451235,0.05046602753331442,0.5350423804144229,
 0.41957593979504193,0.8637282292579144,0.9542025604095382,0.9487682611295486,
 0.783923781607007,0.931639951120533,0.0726475999447378,0.7764977164278889,
 0.37039408004116947,0.22651447680630266,0.5635640140863467,
 0.13492490095534004,0.5898537425322836,0.9547076724877881,0.3619458193468147,
 0.5336364229044326,0.3322707140166288,0.35082827903645675,0.5950726973498318,
 0.5019162730748936,0.7231791162287735,0.031460373193509805,
 0.7193321697759393,0.9576521225508304,0.2571107439830572,0.526498910783959,
 0.8186690613194562,0.9090608162750363,0.9708583179548473,0.7701788922069727,
 0.8919551503402963,0.846285642447759,0.0038080087905518756,
 0.6511143949425423,0.49348458565336184,0.19590254021697462,
 0.8412234489663255,0.378003968063341,0.8708748241722044,0.4149645688187469,
 0.1351559219960452,0.12556427664642555,0.7395108759838697,0.524326864683007,
 0.3440004309422088,0.06936425874863139,0.8882427015005074,
 0.17807260984804263,0.008458886021700751,0.06485152613085932,
 0.853623445812082,0.4836531432532597,0.022218049137511575,
 0.0019160319016932048,0.8406046028099334,0.11423531124596809,
 0.7772727001991027,0.6686617834620003,0.011218058309970025,0.824991463051567,
 0.9967212186800011,0.9096071290730421,0.7968298612456561,0.5858478953097043,
 0.9891241698556065,0.8564956455512602,0.6997937761720164,0.34671587927196157,
 0.8861923339755927,0.21440973666050533,0.011748254504367561,
 0.28226031626484716,0.5943269118955716,0.8713830851476632,
 0.10757203008553684,0.689029198312999,0.04126162608278572,0.1383873904252132,
 0.5304710255460854,0.15136909765162143,0.9529097392162883,
 0.06920482643033499,0.18130430836286993,0.4570727036883674,
 0.3695164303838582,0.2980656118591857,0.2250538001058755,0.7135125352027174,
 0.12634390568427767,0.7113156883041258,0.31829070334217957,
 0.7000789411581159,0.3202185748644333,0.16162949764847423,0.8745220723124452,
 0.026675294522273574,0.3886834630243893,0.46372934679126365,
 0.5908059575889203,0.2000756587200685,0.38761378395120405,
 0.06346749480439451,0.8966036226718052,0.8282166613149711,
 0.43530195997325616,0.6825621646276901,0.3945254688852602,0.7439500322876281,
 0.5685920090002509,0.6657049254187766,0.7317665751746996,0.4993465370388306,
 0.5178867802136199,0.38668047483901213,0.7419610563899331,
 0.28303268733325515,0.680835116425412,0.6506626747203169,0.9794658719103653,
 0.2738615389850175,0.2550827863859103,0.33773551506283983,0.505193841336426,
 0.09688778155964584,0.016052202146972894,0.38182334515536764,
 0.10821068976610648,0.5274849967502229,0.7461237897931143,0.7610565059365519,
 0.7376639346705923,0.10574050384896794,0.4573237502058918,
 0.21398777801218105,0.5154112880251911,0.153948134982982,0.6431216395360404,
 0.46706214124574075,0.6895180535252501,0.3692322001289946,0.7816411145938154,
 0.6428279217167031,0.016696302765168758,0.4239545012223874,
 0.3388038428245729,0.8458503085403515,0.5937608593664904,0.1047774945334311,
 0.6510534985572195,0.2936150812624789,0.8386248406230662,0.7265257398587482,
 0.5127188809839217,0.18031988791996367,0.335850206505536,
 0.056356188718931666,0.6153996188140095,0.6821980855844477,0.868367966185807,
 0.4645425362698473,0.9014993501985628,0.609721848623977,0.870370375194087,
 0.8066869099377365,0.7545297365659308,0.9165905996145784,0.6558396386552003,
 0.8570164480432837,0.3798342360839748,0.41377343190578775,
 0.29939288727632474,0.012311268649771168,0.8799185291638433,
 0.049032295109618484,0.19295002444787035,0.5787840850557149,
 0.9527293365656669,0.054057790891607116,0.38105525343502544,
 0.5469266317548567,0.0714402804694022,0.636767142503105,0.3556094278592774,
 0.8378674181361774,0.16834402906634272,0.36895403537846205,
 0.16440365640700627,0.1967161056675396,0.4117320770817754,0.9070588558362829,
 0.4939594078288727,0.3517711786747756,0.6086289754840796,0.9636345534295233,
 0.470086979812398,0.46047558841796954,0.6432870481343915,0.03320131201166909,
 0.7599441313336469,0.22340279101760296,0.5934362194270585,
 0.42807315318851946,0.9313847646718922,0.13849171984598874,
 0.8475345335557004,0.01226248837325672,0.0064155749867523815,
 0.0053999028030008445,0.6804327617996865,0.7279174160204981,
 0.18187354155388813,0.9835912631330274,0.6850073648972711,0.6088600367564371,
 0.7022313630065266,0.7954876087165607,0.3291084301354783,
 0.014507280535741685,0.8996214102806914,0.53278056713367,0.4399501676957662,
 0.12793889705663153,0.682791609760998,0.6918477474972877,0.6089376291546278,
 0.702393196417646,0.02684215659870448,0.5128663725916794,0.9171181836090857,
 0.11335067291643397,0.42218634266270005,0.054896865894943914,
 0.693145302988289,0.7676618522931631,0.22226393638622255,0.6194182583989317,
 0.7002869400467949,0.355307519523097,0.6137550980116023,0.9096899646459692,
 0.074579091525387,0.8266661601129237,0.7801932549870274,0.9457343238459429,
 0.19441774223341368,0.45638268934947024,0.06173545137880043,
 0.25878706603423896,0.17206006957512288,0.6483694932472948,
 0.6842794340767999,0.37181282315773256,0.9043894136431442,
 0.08826951766447311,0.8842994041798979,0.19995042632033433,
 0.17328822493667606,0.23975785925373638,0.7096045064426637,
 0.6494622054567959,0.15702532383722778,0.07415858284536037,
 0.8554171462846003,0.8178438637407222,0.04482123548106298,0.7362501421971444,
 0.8633185250195669,0.9695051815500761,0.8363024839499672,0.7151580170744991,
 0.08824877550552324,0.022750787468718947,0.06424224754674546,
 0.6512530320022698,0.9977222595247619,0.2725294017686859,0.7659138149941656,
 0.7022239966021756,0.4285589610818762,0.002422958092403893,0.320865935778026,
 0.7823842484402068,0.039733857161541875,0.35030443310018666,
 0.5867723120599615,0.6765154421399731,0.35469919888286183,0.9216145205374913,
 0.32911896666000506,0.022154430558014915,0.9249549412816118,
 0.4006697974778328,0.25489915983949185,0.9756322493262821,0.4439812087774784,
 0.17844580314928227,0.7228582003461201,0.3214000996726679,0.4342203922000264,
 0.5029176548379717,0.7540772633609664,0.9787943177219681,0.6873794380288547,
 0.3691978712232473,0.27797438109518313,0.8478306593087239,0.934173540969126,
 0.9232000515844554,0.8174540352021202,0.501733534687153,0.18597163497820546,
 0.45693130456342446,0.9373639813621226,0.8689767771266307,0.3998110876471772,
 0.4444539353736563,0.7444959178129751,0.4830213371678631,0.9400650116341237,
 0.5182901739892498,0.5136243535107021,0.023700270743206353,
 0.35041239632452936,0.3463011046983966,0.9388424836114334,0.5225059897803386,
 0.1291083623804964,0.7055153528445779,0.5868183528257958,0.8920534955234022,
 0.2529511044510704,0.0031685392418860925,0.8913268738432556,
 0.36478121505996397,0.5498824180274219,0.8468496255774004,
 0.20323891021022833,0.11871300033120147,0.8848581034067741,
 0.6413925648452835,0.6474774611483314,0.9126349571194383,0.29910039319958615,
 0.5286741590068426,0.26146428415587386,0.19020141636429144,
 0.06923077415509626,0.14000572840338765,0.32586970650334723,
 0.019360087086011224,0.18237703097452707,0.7047846366274034,
 0.06110539784947444,0.3034244676601735,0.49160283892518963,
 0.7681839345641912,0.8051620038979989,0.9812093623355052,0.8004552190481548,
 0.5616444614172085,0.38712179400142455,0.5089557787309343,0.8363811296268827,
 0.30700401158841084,0.6249181075453341,0.5106467467282793,0.7830259548905791,
 0.0365070588897477,0.3973279297766461,0.0398291840995213,0.8686804378044148,
 0.28954918203826774,0.4423584730293536,0.15291941879321747,
 0.8702611778782012,0.7350511411683633,0.45236029118406706,0.3585855778153255,
 0.5925320301343213,0.1611681139443183,0.6419618622171023,
 0.018654639283212204,0.3618423456040012,0.5037723480206404,
 0.2097839223814839,0.04438023697430338,0.02665448439789242,
 0.5433772898268475,0.34884438384064653,0.9729775156074149,0.8222414068152301,
 0.533540851107625,0.7897470614611012,0.1811006088835121,0.8950991354265555,
 0.9288676719745756,0.07742287815265037,0.9061051889730528,0.8884187122717409,
 0.705089844710117,0.7600473488817174,0.12860897906791802,0.41239439020487756,
 0.3741602178700849,0.38817954985028347,0.00376371800770936,
 0.6960845770469812,0.6717465607503168,0.8778222777692138,0.1184833933709738,
 0.9372487187548673,0.083020553157375,0.24867267948249006,0.04219343831929556,
 0.5294498009210102,0.7941311113758973,0.04682353506487047,0.9713676760877245,
 0.8574197299083843,0.6514185990953294,0.5094691988158148,0.8406973639939496,
 0.9766179094303,0.02128831777897222,0.40627430689456034,0.6614488975504524,
 0.22542607248726465,0.3856111313388568,0.48705802780519336,
 0.6915044494766023,0.6302222888033187,0.8420844406358593,0.1941484313703723,
 0.685945596789478,0.31121619251448307,0.3372590077704224,0.10988994182115464,
 0.9379077728703112,0.7285660231577364,0.8491565578071664,0.3967513739087597,
 0.6808053674265718,0.8436233713089685,0.9447190129665456,0.9706706462025978,
 0.6383705878477782,0.7146086847782418,0.0187808474099328,0.35324006903914473,
 0.6356638410767583,0.42522284785847386,0.5458434057321989,0.5975964954612026,
 0.3765726730487199,0.8447696504402991,0.6968781469094503,0.9908439820106497,
 0.8673130442595518,0.27877850899045764,0.07289671574502776,0.663703720505708,
 0.09808863953011171,0.506884795492819,0.08256783295239456,0.7870978902074811,
 0.7472781656973413,0.7229093506200317,0.15975605744243726,0.836269462985963,
 0.5774828294321345,0.41527682328116167,0.5994643074987867,0.3672875425094444,
 0.19662387071174248,0.27362700475826796,0.6142018869735567,
 0.15695555240425563,0.5278828449990026,0.9626789276332902,0.9560731075996697,
 0.9990915931094712,0.3160711908195919,0.04994810240945036,
 0.006468813892098413,0.12427756469355644,0.9141841455108979,
 0.004663190924764193,0.29963016893634364,0.4229438877291727,
 0.22148610797739843,0.3310505883969437,0.41865711229967906,
 0.5089361354516314,0.9528237029350147,0.7814448828224477,0.03877689371652937,
 0.9970397901619767,0.8277880023682137,0.017417536677354906,
 0.3344908205711754,0.2944048920368393,0.4898784621113177,0.976430650660357,
 0.8332356184409663,0.33704912426569233,0.9169979979740416,0.8531018814240651,
 0.1332673289355809,0.8289495025971745,0.6940334739207515,0.2571647988733782,
 0.11181687336289536,0.4157860017921957,0.5885119882654779,0.5426607172621397,
 0.46196057836521076,0.1988381425740826,0.8596231849149152,0.4867403652234329,
 0.3650861879803512,0.7443028873810741,0.19163378660197183,0.6939526611011618,
 0.8237335852484224,0.23251566171357552,0.6184942333704199,0.3862866171551852,
 0.9121214857839648,0.7665205475828056,0.24146913716368712,0.4805350705321617,
 0.9862980581408389,0.9352453509975788,0.5370530233748819,0.18476205696874282,
 0.6421821505045725,0.5744451566012726,0.29820443247279405,0.625857202337246,
 0.30059786665623356,0.7509873121056914,0.5095098403682514,
 0.06642139312521678,0.4598651555411496,0.3173134225321478,0.4669443785691936,
 0.025655222245592246,0.08964659933891961,0.24516404677669446,
 0.22034915823942025,0.8617695991640509,0.24247702592457065,
 0.34146311710849175,0.7791183047988628,0.5201142033884057,0.5177691033313093,
 0.6719909900584281,0.6368940151167277,0.8013001906213973,0.43199863984222064,
 0.18508796332855926,0.30169394013804873,0.4834892410130627,
 0.5522116211741035,0.8115496174036692,0.41302366631772436,0.6908519297346895,
 0.7403947383756772,0.2640560453538583,0.8131755167972715,0.968008436247138,
 0.34679609400724376,0.34997353939919373,0.506440361769565,0.7171536064225514,
 0.16142540183566834,0.7876143716464357,0.08330445021101185,
 0.3268049537490204,0.5708301808205191,0.3506469139466837,0.8783360693589538,
 0.6776332994689782,0.21169552501077604,0.3181908015715609,
 0.022122969897451994,0.011915681404587541,0.8466845144655673,
 0.5647921014245303,0.17252130410536737,0.7931534499270864,0.7522803371235632,
 0.25205243564550583,0.4223648142699432,0.11984282885217201,
 0.47653220564817667,0.0355775716804585,0.5785704962013083,0.9638025940023509,
 0.16169157974174686,0.5055948739828353,0.5479978166758731,0.2488933386802582,
 0.7164994506867396,0.1947803013853403,0.753723598286375,0.7481123239590338,
 0.21712572179012202,0.545341592517226,0.3583513251770011,0.02478565129010013,
 0.9130620736976847,0.7698524329898606,0.06338720416342858,
 0.22256052395889425,0.41184283105987274,0.7084366940536587,
 0.5527729127562239,0.3292193068164896,0.8169755410022428,0.15631924178573497,
 0.9719122939573815,0.9223650445845748,0.8995029834896398,0.5875257012273623,
 0.6085304696549068,0.5085433112821773,0.9806045805228604,0.4523164467802755,
 0.8410873430117527,0.5188012737300048,0.9962892083278143,0.5582889152650864,
 0.4145708222906974,0.6996145588940903,0.4507619357712992,0.7059551990531893,
 0.34442001231878994,0.0564154753534678,0.4869083678830154,
 0.36538029968635366,0.270886161528997,0.9630518159180724,0.8943501206017302,
 0.7944781946805636,0.12106782471665689,0.5997959039588272,0.4743139486203959,
 0.8749219791617884,0.2770196500311154,0.6066881087733436,0.41298217337651877,
 0.09866628350667728,0.5231846224449749,0.029634436386685503,
 0.8838279480998676,0.14939403386885675,0.021895686716366658,
 0.7841903567955387,0.32282319968901463,0.5853368130307959,0.8264853076785876,
 0.8538783026747814,0.07697529006366666,0.6778120801925079,
 0.42683017123419864,0.049848202108352724,0.4265884473003825,
 0.3446691631334029,0.5772247390209818,0.8040771194640254,0.04302728145896306,
 0.16490507789526598,0.5814049657389765,0.21374254565180162,
 0.06641278071604861,0.055450563538555375,0.3663475159339943,
 0.6059129582642488,0.9425925389509452,0.5773246369989798,0.7174873532181084,
 0.33538756211851806,0.4229490949580873,0.20495563382754867,
 0.03759380864786366,0.1906846617851553,0.12446529272177842,
 0.42681331493213903,0.8049409101802618,0.2260969646643749,
 0.12374464803727958,0.19643169036679442,0.34784982847186263,
 0.6695030217269057,0.30045838196129615,0.06077311908205696,
 0.3342240855562779,0.2758431785639158,0.5546654603552386,0.8062276593122182,
 0.38399613968701096,0.2984098340700232,0.661271308046755,0.8413016010604669,
 0.6331685329310749,0.19418258085344342,0.16111868983380662,
 0.7648744304309605,0.04667504907840514,0.9210834827125329,0.6258743489846783,
 0.4246745832240336,0.5223175885613816,0.0029047917689636327,
 0.08443141443300661,0.3099700408653999,0.1287078817973244,0.2987705225965688,
 0.7427935689358989,0.2523975685978417,0.12679493032483946,0.6794959338696893,
 0.7132590304981201,0.030867156194762546,0.30389294610669093,
 0.13097080611450185,0.23140908827649942,0.3711576014618264,0.125039264372644,
 0.4359507050268753,0.1372594470731494,0.9916546103554212,0.9983963199067469,
 0.6036897178376546,0.1333146474791853,0.9854706438394328,0.4630208687433739,
 0.39415479354213134,0.12017343920559431,0.20202791574888246,
 0.20354947840882287,0.3631361027757819,0.9250329511448434,0.9953074694464619,
 0.6593360821932547,0.5963763158062038,0.9169847239711724,0.9524500162718655,
 0.8466962787439889,0.963098612624122,0.5297931469975848,0.14928454656956425,
 0.9830864395548436,0.6710305194857856,0.7309489493399847,0.10420855216061509,
 0.3066721037893543,0.6388828038835427,0.7868708176993022,0.8524596387557128,
 0.9642311269322252,0.3005447370631068,0.5434681831937829,
 0.056936306470519615,0.7669297217147977,0.8403777321768023,
 0.6471326687694914,0.06154470784965993,0.2461545383061131,0.3497949484604569,
 0.3777542866535235,0.21099395220045192,0.6257176894033739,0.2273714457388396,
 0.34839786361933167,0.5915554340534074,0.45037230447790977,
 0.36532273373450885,0.28582760637329585,0.2332530115069431,
 0.08786237028770028,0.3692988904451565,0.8598568170111249,0.9711354694365247,
 0.5704004494529018,0.15654287559740476,0.30261052772598485,
 0.3286903346767651,0.24177171073060844,0.38625165523472693,
 0.18753617636051134,0.903973346302628,0.9529440913704879,0.2935190892378673,
 0.9031038865734752,0.6676227424157293,0.32684717789066753,0.7944268149542901,
 0.713812975452591,0.6765838414057839,0.32666200104551,0.9764042299396097,
 0.4513873140548146,0.26657993055018336,0.8124336355481105,0.6238667254584841,
 0.2136825335068423,0.8263827726498377,0.8675115570767379,0.2510384098027838,
 0.7318948107264648,0.051705618449211954,0.5567759522036042,
 0.040711324027392415,0.3568579547663119,0.2864318147034538,
 0.8685453884784179,0.3849972571485136,0.22047529772041696,0.7243302176618094,
 0.5896299898654443,0.23918371209421396,0.5019498164817724,0.4385151886084001,
 0.6362540121183828,0.4619813662025116,0.7578836609606092,0.8476037767459517,
 0.36330018421417043,0.3892230253211809,0.18179703539590897,
 0.07295203933092798,0.4040318212239059,0.13765360452959918,
 0.19323543335196713,0.036724246653639936,0.3458671327295936,
 0.8685937635817816,0.2698553553765636,0.5613961060862543,0.18605378752824986,
 0.946393844738318,0.8456705384884605,0.6526952477440543,0.005078635375954432,
 0.48251384321051083,0.8019441239016296,0.622840997386899,0.8427068837801468,
 0.6989452329079311,0.922536298590652,0.8952426322182134,0.06740847657371374,
 0.06991569217033766,0.11505201907418916,0.4834521046550734,
 0.4058269950171929,0.7305039589025604,0.4614424040782439,0.49893445826484617,
 0.29955451952384615,0.24242247694638208,0.015144109475632939,
 0.027635469576613225,0.31336623750900594,0.6171367409388626,
 0.16383417082026996,0.35319069169234896,0.2705271522851601,
 0.3599785176798549,0.597487299145434,0.8116815781918848,0.7415166810552655,
 0.4248835763840916,0.6710557666186482,0.5600273662249823,0.1973819512036925,
 0.8196157443968272,0.7290438173265223,0.9658075161817365,0.8543427609643661,
 0.36864277281349356,0.4543820101787137,0.9429651712718291,0.2428856848581522,
 0.15243825577594505,0.20977451886189147,0.2919986751493473,
 0.4167533656437572,0.6967475978835126,0.3745173415448789,0.3263851517765757,
 0.7706711640578296,0.9857353974135863,0.4505726250518176,
 0.025745682461706276,0.6269632067535752,0.09944843245492663,
 0.7888825368778339,0.2985227167612362,0.5340432518734218,0.686506499957615,
 0.28746344346130415,0.5981838272355972,0.7884418372014692,0.5598256930775679,
 0.13583433992860794,0.7656505491829028,0.9385908106946501,0.7530851589447795,
 0.5386012236960261,0.13868467354812974,0.04189435398048602,0.741657319847004,
 0.4771192258192628,0.7231510813452491,0.5384024650096165,0.587776780328038,
 0.8677237138686649,0.22906088949040093,0.8932004741259388,0.548339542948184,
 0.34381523055006213,0.1903139488165364,0.4824252978348683,0.3992575490762752,
 0.3172034444075046,0.6914143234849445,0.078671309110951,0.3127854091124028,
 0.9346428084276649,0.027996368982357356,0.21083226127268573,
 0.2822019199970074,0.38679415689953267,0.22389015161080916,
 0.3187201493638955,0.518811868043006,0.45439531925534216,0.4731991400043921,
 0.49704964613871416,0.6691760712546462,0.5208096605009145,0.8029573839954651,
 0.44121706608620537,0.07449393518965797,0.3935230776350289,
 0.6269447760833046,0.24150608437660614,0.8923385350593277,0.2792449980367264,
 0.12103510383588201,0.7631042924948191,0.5018511972349888,
 0.14046017240368625,0.25961732904700996,0.055329586918093976,
 0.6894421240960101,0.7347634636319946,0.4280925678238877,0.34437198471625974,
 0.24627016680206648,0.6693431777548136,0.9335243957458216,0.770886325254657,
 0.5624519376620292,0.17740110966811617,0.8168515672972742,0.1850919079435771,
 0.16070448858982678,0.6735896360021996,0.8280286492983127,0.779329950535623,
 0.9620686878244075,0.6255686991980351,0.5650762858885398,0.47304009778825973,
 0.5524174959951793,0.9755899443190396,0.9572605703354103,0.6649953053531714,
 0.7574102688534793,0.8001806421192913,0.4519098930894323,0.07377062014255598,
 0.30738902191671513,0.2308733953065133,0.44468973528993905,
 0.4421783559324284,0.7747710423265866,0.7841601882967262,0.9582759644061798,
 0.09255947558952482,0.8964155753578302,0.24180436013882545,
 0.15880560465603022,0.908010817006592,0.09917898925506297,0.70784997063952,
 0.20945372266641693,0.11105635279561599,0.0835181838255431,
 0.9985607238335027,0.1424810736409886,0.8967018408810502,0.3701926764145138,
 0.8373871702838578,0.7746324806745857,0.3506307609229691,0.04603266412956397,
 0.44259393222778354,0.9135002145857636,0.7622375337874033,0.6512115759776379,
 0.6450699204486487,0.24531497410221337,0.14107198445256197,
 0.41213071834804726,0.18222272480643575,0.2680279934123163,
 0.5452883191348235,0.27581476234159563,0.724609808769282,0.2451583990057291,
 0.7439403133567554,0.3540537762942362,0.10478660310565746,
 0.29895987189622775,0.8929422153023867,0.5277499576572713,0.6713732474852138,
 0.3255032028654734,0.9485993379952314,0.11881515367119677,0.2220582081706035,
 0.5322365851542838,0.11707169576838217,0.4970754959349669,0.7689389727508762,
 0.17603846285918223,0.6962243235538095,0.7518797787460801,0.4209301885558101,
 0.7169664971627041,0.7297076452765997,0.24913857274657636,0.9850395652689622,
 0.4138755287959265,0.1538387043222419,0.49623916350537967,0.5938138367060738,
 0.520892539138051,0.028392338023737862,0.055338958516985226,
 0.7130521743608509,0.20042863093013263,0.7900197574265795,0.637689224019527,
 0.6366223889317166,0.8139335253608027,0.17337230267514436,
 0.21199301703268036,0.7970551119310426,0.5348591771156269,0.4501364758678873,
 0.2587091555742034,0.529494778215051,0.01691619036018266,0.2248426967489472,
 0.4822096975052038,0.920254909720567,0.10201785197793323,0.24730745534362153,
 0.8704704138689858,0.7855202673541959,0.8880078434428105,0.5799260149248474,
 0.9671640560170165,0.24611104477294865,0.9618676823434328,
 0.020352089009362473,0.041150576235646996,0.023208484116870376,
 0.43568604671439326,0.42757372975452657,0.6036686246471816,
 0.7145671531553776,0.1615167983415613,0.33924968375087805,0.8828941087385408,
 0.5878915572185495,0.6607580518501786,0.8702976482654385,0.8143963388835596,
 0.30682098817061676,0.0034494405998810596,0.7513225509524253,
 0.9707860812446091,0.01950935501847839,0.36462428990601103,
 0.09572186539583483,0.19268775554581574,0.7382741609988629,
 0.6535740277838566,0.0980604898675227,0.13865501857997753,
 0.42343932347483515,0.27213154902474457,0.8744103706103062,
 0.7738295196598081,0.05707503215511989,0.575361709363669,0.3008077337470434,
 0.19688556209363173,0.6221820521549938,0.8064966252974893,
 0.07714339129267289,0.0777836339971002,0.08189016779494984,0.75205319379794,
 0.6742092799703695,0.6943217096082742,0.7538563727086379,0.8906656092746228,
 0.4024463774840391,0.16131398658921825,0.1492827764699982,0.4221365999230342,
 0.9147148347003295,0.6602350136321916,0.47970238378884245,0.7511009817945664,
 0.8138131396952808,0.6234007359753949,0.6623314682141432,0.751754972551064,
 0.37976977966918735,0.0020154301372696803,0.5582485966108474,
 0.2894849419319966,0.6923987383310664,0.4602521495687819,0.33640173132662987,
 0.05180345633772998,0.5424996303877727,0.7183863499741862,0.5253978466525737,
 0.20038160145445616,0.08215236216120236,0.6382743338211252,
 0.6000852022534704,0.5042318702725795,0.32174816908983006,0.7594210621718652,
 0.29067204265632385,0.7291400076064676,0.4205210953892109,0.5617521854697626,
 0.42840466226231355,8.240908872746422e-4,0.9130053450428296,
 0.4770063502417671,0.6915657970862455,0.8290425073055671,0.4165400022530402,
 0.5468295011050491,0.1486782061654115,0.12547661774314867,0.7642081890519326,
 0.19603666327812197,0.03226687334341172,0.019340397181625835,
 0.33182273244872695,0.01326756815350938,0.3575672136687449,
 0.6295794284255243,0.5545161831689888,0.3274546938006452,0.40600292389025194,
 0.5009424832863432,0.00750792413323198,0.2785109591415952,
 0.35905577040608205,0.5589184914290591,0.9676300804545571,0.4552470207647984,
 0.42119821210767217,0.3865805189722762,0.5967468430326801,0.7055254098738944,
 0.28473351339326636,0.12086867711249549,0.22901745024407716,
 0.40436343187930324,0.25007874650773654,0.17972998442330068,
 0.8294812000747958,0.9350038402971016,0.3042157527718412,0.9850383700423708,
 0.8220379905070822,0.8981321999694559,0.4993260783520661,0.8010660445400473,
 0.23222906313446023,0.36736282756131533,0.5397553859674277,
 0.5579950300220071,0.032293294276512285,0.12550223199994215,
 0.19261233467532712,0.9281821400713718,0.7242919353041091,
 0.12300018851976491,0.07007364657257231,0.920834681988936,0.5338574286875538,
 0.5247428575766038,0.8888288021801016,0.042474153923659275,
 0.13437955036668503,0.26663931160735144,0.4104110551050333,
 0.24026081042642744,0.34689906222114486,0.2045118371546314,
 0.7899863532856394,0.6700492480183845,0.9275658234323212,0.6919568946409235,
 0.04817884147415308,0.8578993061153443,0.6311244787600078,0.2973071508193783,
 0.8911465379943797,0.05041873868029101,0.9482448986145203,0.101071485898929,
 0.856249656060182,0.11595216773758799,0.9870844724443383,0.11396625656312664,
 0.12886253521916802,0.36242531206527906,0.48941615782076464,
 0.23661695114562709,0.026339039715624568,0.7399732964685513,
 0.8772906409082517,0.842714668689233,0.9812238196195158,0.1309997281094446,
 0.677311518817315,0.6253843471581184,0.36539708614628996,0.7321051676330254,
 0.029827386216245366,0.06760648670307512,0.09957826388919244,
 0.5766286880439246,0.22200533338317008,0.13233853203464607,
 0.5929729385302401,0.9656439719779281,0.5745643276785882,0.08757253115520325,
 0.6897548668441676,0.4385379621084152,0.052543606803635456,
 0.11196571424292656,0.14325299716565032,0.16498728369675253,
 0.6715208983656561,0.3050451019577439,0.5937276395658986,0.9286966417480608,
 0.7266235972731618,0.5709296906905296,0.031209548273769316,
 0.45112495072630443,0.8219872552405265,0.26854794609962307,
 0.0758940638973008,0.2223139089434586,0.7893407055311774,0.323371883609084,
 0.7934105934146457,0.7729487185003916,0.6003004225901101,0.7544567591917943,
 0.2427343104742442,0.8000926022522621,0.28992559711199795,0.6049477538525868,
 0.025637951289889105,0.03155210138832598,0.33301967625111184,
 0.7610144942715462,0.1798773466618957,0.18418736477618392,0.4936510676493584,
 0.1632896252362258,0.21920899778561942,0.2995087260116188,
 0.31646643675134956,0.9388285955403173,0.7043011778886186,0.7301301246724549,
 0.14710030663480766,0.6847093381287875,0.9269756523462652,
 0.18654389564517881,0.82678019551651,0.1400608802244312,0.6796235103506307,
 0.9335646879824887,0.9720445280897118,0.5804692605535454,0.5137668427748993,
 0.0018521995251493673,0.5292619445370821,0.9256330550332355,
 0.6014116456591919,0.6238043648104603,0.27844349733758345,0.6131317818452555,
 0.9991248000227807,0.732266323535734,0.22429027461273743,0.8874463696693796,
 0.6804516080811184,0.38426478481840265,0.3939524309711967,0.8098536428148279,
 0.9544139394891729,0.4549762911723786,0.8058564120374999,0.7527529476378845,
 0.35319386503185934,0.5793103910822031,0.27610622513154204,
 0.06925748308112001,0.480145085376928,0.3837506743667931,0.657857883602736,
 0.7304490887729569,0.039840610889334105,0.18794499396200337,
 0.03503869511020263,0.7528804447630807,0.10116942324808997,
 0.7450521892979597,0.615792683183974,0.6332000657663936,0.38066693720669986,
 0.4802963979654926,0.4589604648200757,0.3531072952332943,0.5220300075336741,
 0.8680043911846874,0.373939778511423,0.39968589627058493,0.20297968757129725,
 0.5454460514960515,0.06191478003912565,0.692264095957708,0.9902835252051828,
 0.12134536774238591,0.7241235032509952,0.10777781701478661,
 0.06021596143632901,0.43312249421894566,0.31327918538896904,
 0.3727158678628191,0.4946271829181986,0.032850309406151856,
 0.10522193091923115,0.8910144475411987,0.9699919866115971,0.4105642130903614,
 0.12976592968245337,0.4344608042010981,0.24964986632097408,
 0.9552955100330364,0.1778372430174806,0.14054930510898567,0.6168228388522502,
 0.6770445537822809,0.6959110603385172,0.26895111358707957,
 0.37878360670817923,0.48043650745113087,0.2863505656519566,
 0.20866040319322499,0.8468396243739937,0.44839541543085526,
 0.5498238044383305,0.2720111843485735,0.4832688278858154,0.8992511782605556,
 0.8529969527850717,0.15877842947064524,0.1838254923860172,0.5659298072499993,
 0.06166927539110856,0.4245649637463986,0.16211705483815508,
 0.6717534358631598,0.517170208683575,0.528771968906898,0.6745842743968434,
 0.9851829483466956,0.5039586218091758,0.3446237568900892,0.599576587187902,
 0.8362036520280418,0.8163198034383878,0.5717290521761668,
 0.028023606207395413,0.8322327940968076,0.8310417789891782,
 0.5572083861433816,0.2712130335834261,0.5040033832545477,0.4380894665795615,
 0.9882120993643786,0.37413343219132744,0.6448961802356097,
 0.40873735483181384,0.05687770963663896,0.5923110329027378,
 0.22731280197698434,0.6036411533995332,0.1553102944837642,0.7366948541620336,
 0.6474129850557994,0.2573687714549725,0.8050684596065196,0.8646428189759564,
 0.20757404159889603,0.9467176513193996,0.749116325265379,0.667111479597499,
 0.5418228023416167,0.4990082267445676,0.3729624968321016,
 0.026921515681341823,0.3786284397169333,0.9734227031005913,
 0.5900651731482278,0.6056517907187464,0.6076427559367821,0.905560408356503,
 0.5020990900215787,0.06596486644831223,0.24709380230932132,0.540429658866567,
 0.7878196716003014,0.8014584890360789,0.5203442937364015,0.8626140027330812,
 0.2912679173893986,0.473676119111617,0.03239703487143031,0.9357202070843386,
 0.24888811620438844,0.7019540519248018,0.9211955487565958,0.3988763712817778,
 0.6650577472752961,0.26655481088095756,0.9575017618638046,0.5236391252938066,
 0.10177978022031797,0.20803316944263217,0.8179858243101757,
 0.4922785831474027,0.04853748480983011,0.3170157334562046,
 0.008223017821228362,0.807609643554674,0.7057861074521351,
 0.37521976106626675,0.3818536556154467,0.23140740138500404,
 0.22557678488387456,0.5525604054842208,0.33560157752601016,
 0.5466551390263128,0.0595840051814055,0.8950724517090933,0.3058326440279764,
 0.47469304707957916,0.0716958660434528,0.7647326335051939,
 0.17296850765276617,0.5905849728996274,0.3428838391131389,0.8195076959730355,
 0.5288200937331063,0.7237688785763834,0.6255006138259994,0.6861478940230388,
 0.43981178151982225,0.2543028302465066,0.8332256870779118,0.8739950801607802,
 0.13542945066533796,0.4542735975227037,0.14094584677333744,
 0.32363539803118546,0.998584605856613,0.9511922299324037,0.03418174506195337,
 0.5286511236576902,0.8380344807640394,0.5968029491242335,0.7204437480053206,
 0.7822397111114178,0.5326148715379161,0.5123155251129025,0.7208579314489456,
 0.7434383236428188,0.83650067069354,0.20745098088083014,0.6379466982953742,
 0.5193206006688258,0.8588867140567159,0.20054814987851244,
 0.11492049882099842,0.5031276077530143,0.6018773572168394,0.8786803646857024,
 0.19975094086603262,0.13833654960771857,0.1159217758205594,
 0.3947588203523449,0.7405157760410677,0.6960280862467599,0.5410596097794678,
 0.8857976605865463,0.9410503212000405,0.00878234852241222,
 0.28305013794731926,0.5678686983651664,0.985893746400295,
 0.016243592974175436,0.5253131444649778,0.286370194545303,0.9974941745564629,
 0.3690062812355521,0.3762697972857414,0.9747034271250101,0.26299926738780244,
 0.0835582415704661,0.6629264336210023,0.9862428222712748,0.853362620204186,
 0.7215937572958844,0.6954352525056289,0.768840121249169,0.5210655942907763,
 0.20612400134608688,0.35816510322472905,0.5481290430813455,
 0.15391723811164815,0.24458193114591953,0.5478832936527678,0.646587117342881,
 0.15306136151289584,0.056229104354583725,0.7045925099328687,
 0.5546582418082201,0.8151845049187327,0.985829843820105,0.11602268565210638,
 0.3422638282985373,0.19562692447035157,0.36176785409413315,
 0.02099025947568567,0.1277981361721302,0.9812723143454615,
 0.06064712687678464,0.9906059636894406,0.967797291775932,0.35017881301148523,
 0.5672500188942275,0.35276804724722366,0.7596052659364749,0.8408462103235456,
 0.4907840750023479,0.07779903747916328,0.02823903033927877,
 0.7797853915934522,0.1257926557239486,0.8531701089960041,0.163200217190146,
 0.07880979146183087,0.5788365701727768,0.17557694717566674,
 0.7367479699310469,0.9298596857180794,0.26800767685217386,
 0.16820957421147875,0.17350240401702288,0.5499339771017459,
 0.7855400148048842,0.9225508563547719,0.6302705393996924,0.31857726874882963,
 0.21357551159394006,0.6113948214368007,0.8747224382055896,0.7501232318348839,
 0.7709958616647528,0.6624799792820188,0.020019269507139836,
 0.8512670589481884,0.3898595517024287,0.03289298248965278,0.8133180749823984,
 0.8427111348305895,0.668435751631215,0.4190767833071598,0.6974983100480152,
 0.9672097039409608,0.23004333091614815,0.7331817618842965,0.7176817319014488,
 0.8586199518092591,0.40721555172939483,0.5020972424968252,
 0.15605624314626765,0.3654853589530743,0.09794585872161221,
 0.8665446191070992,0.06428191824263496,0.9573176506923413,
 0.12737294794237153,0.33312293389469905,0.8902600861967129,
 0.6612757060793191,0.8669392215174564,0.4211753402895325,0.2779284810383925,
 0.6985931506941392,0.7352542437305921,0.20193774227876915,0.6856402539119082,
 0.23787827352787694,0.8304293315064176,0.7250019372648746,0.5573065291939029,
 0.020884876726792845,0.34886558087244957,0.601822273067141,
 0.5499851840089879,0.3213652198282686,0.4371578835327936,0.4562614867331346,
 0.7530399883095376,0.8901468667186058,0.3788619378660051,0.7436600081247731,
 0.7235256647814168,0.11773469723478569,0.24229178979405752,
 0.22006830836378577,0.5790905857092838,0.6156134101560508,
 0.15099790628374576,0.11799526847538133,0.2939824328823164,
 0.07322884495320825,0.9248002460558032,0.16238546545852905,
 0.28171314919627655,0.835701969385462,0.4294982590344225,0.03594926389968378,
 0.16460549187334728,0.4756575303184446,0.7460100350180505,0.7614974532494307,
 0.24248642962814138,0.9413881567371563,0.09163025358309385,0.473435548917822,
 0.01977203967458996,0.8960765335250109,0.8292181141659286,0.2795062196810666,
 0.2126317980259641,0.005003397558739575,0.3033838089581723,
 0.09372607567504165,0.6760950623957356,0.6510679409200718,
 0.48137805656764676,0.2946756462401342,0.3036373589242751,0.7332342474355281,
 0.28468816828231924,0.12359211340399678,0.22054421042936556,
 0.014214481985834415,0.5201973541149737,0.09110673118315982,
 0.5701097946929093,0.406125027901424,0.4178415009421529,0.27721963800856997,
 0.28256814274042785,0.07440592252222689,0.2500376839663859,
 0.4331874667334258,0.5621507569957044,0.5665310671256797,0.06907142336458993,
 0.4658798989741655,0.9633223849161762,0.33743581381275,0.2595051103152838,
 0.5103276817079913,0.2662397989153579,0.40407170913626833,0.8949877460343336,
 0.8171580709215847,0.11184401941946187,0.3059552740372349,
 0.33296682138280465,0.4947561379687009,0.8353700048331628,0.3858000643214239,
 0.3570714113520864,0.6271293373552681,0.3744026542151835,0.3081751660476775,
 0.7993305121120042,0.9344311810776603,0.040315652121053924,
 0.7501791681013839,0.6104963120819036,0.6556628256201174,0.3846664726848784,
 0.3178475480277152,0.5710747083531797,0.9506875574344987,0.2597064001218138,
 0.5656851481372067,0.10416650513931691,0.7436310253190456,0.6283132085852852,
 0.9074939628755975,0.3584749629297136,0.7960595528107866,0.28290436700552146,
 0.7036040902161115,0.3795415273781161,0.2286807955706477,0.2740450673698376,
 0.6857030316007448,0.08319872456754662,0.7791091609818838,0.436914597246698,
 0.6108827339229631,0.5386579321267414,0.12611973518791075,
 0.05901795771288432,0.10153618458464941,0.5413436996277147,0.678366458366053,
 0.1578873074212518,0.5512832557515114,0.9749063065709016,0.5938433727228214,
 0.4303643187544587,0.9931740190517796,0.5586550730209671,0.12371823715979169,
 0.26869564836473325,0.8509735470146742,0.885094972543808,0.29894745278927815,
 0.030404750889869936,0.7482694572172732,0.9947050570020931,
 0.8798675102881689,0.6711929102055352,0.2967881702449625,0.6500740179007428,
 0.7489303045457595,0.9852776844245877,0.07040413289199365,
 0.060636492088619875,0.9477928291161211,0.3822452833308968,
 0.19512344015012228,0.7280788054883638,0.5791796126001811,0.6850665638670385,
 0.7532038444451691,0.5263555568406033,0.44364001812414955,0.6997446142307051,
 0.18959753422410353,0.02509440478856484,0.026844554012141475,
 0.943114537496529,0.4307549476550282,0.1232010382695593,0.3702741326620749,
 0.2228518705260818,0.5661760396373064,0.40343853808711416,0.8689701008760922,
 0.15284108691713005,0.3810263515416096,0.806246485823129,0.7169690575618316,
 0.5681347054356186,0.6464495609430158,0.7769358702494581,0.2563189413627195,
 0.7543890692794828,0.9755920597572765,0.787190279320092,0.2386736117509718,
 0.8455548532039538,0.04664863845162859,0.44948241913215126,
 0.05145217806276958,0.2441491427802024,0.27919004973085304,
 0.48261645653571383,0.23657682261376234,0.02576560546719453,
 0.375841267248628,0.19349311687069704,0.2920092571559795,0.9481463175585865,
 0.5823069005506469,0.05460253966664219,0.143494842471505,0.521730788202608,
 0.6542083251893723,0.07798545943557345,0.13675257167984645,
 0.7632433762551681,0.519478363785885,0.6642891165698053,0.03805849903407288,
 0.9995081523608059,0.5454256167907335,0.4681493925987693,0.9349707449892143,
 0.8409395107936904,0.22292801525508588,0.13826322390007484,
 0.3130167029944142,0.8103561117967021,0.5080884656881942,0.5259054612313149,
 0.24208250882286042,0.846614379265861,0.21218888662167013,0.8919932005396193,
 0.6136105588672307,0.3571034519142171,0.6274198723853097,0.5464906375867036,
 0.6449942409904742,0.805057223347461,0.5933922672204467,0.5531399190396545,
 0.06087692276930812,0.09412223843065515,0.8553353813603259,
 0.8695569565215104,0.3880081147978367,0.09719094587686206,0.3307954665666104,
 0.16708874794376305,0.3574456355092439,0.4629739459280916,0.8345785068809844,
 0.7697524588800003,0.8726976247031252,0.8444302639129693,0.6524247268800847,
 0.365804481977547,0.2228094370479864,0.2650221248875331,0.1282132718343203,
 0.8145430682495511,0.9552580053108192,0.5205661261517038,0.5999127058282283,
 0.9692158114423646,0.7767714567709489,0.6231182806071645,0.4760125980747568,
 0.483134592727404,0.7651743756407471,0.2740412666076707,0.4603236840552599,
 0.8096574841204466,0.4846105057693365,0.969663591058994,0.7784609765927359,
 0.10658172694980994,0.39387743626406335,0.7222693380025517,
 0.6936988488819305,0.8673299486838658,0.6572012922462114,0.1390171531773603,
 0.8734271732746405,0.8449867437387166,0.8354318198593177,0.8500042256292022,
 0.3355059095508195,0.24805909834108353,0.2532162407090669,
 0.44332648946777053,0.17172471187377858,0.23057872703121463,
 0.3368735318182502,0.8406915956732832,0.21867688176922173,
 0.35062633021282663,0.6765629973287631,0.9891330021332521,
 0.23321599774927104,0.29309400172108657,0.48501475736835253,
 0.5127669568915006,0.9138970785404178,0.06148062555908773,0.5798951830916251,
 0.7895581813337855,0.41808021619687064,0.6441713906243788,
 0.29194458154699854,0.08191285163338058,0.6566317564024823,
 0.9166862109928721,0.8742761100791557,0.03359516136279006,0.6511079631571087,
 0.5007725513248288,0.8067661670686038,0.16559989887953575,0.5850391601832583,
 0.47332983743249024,0.5430318434866209,0.35917305397107024,
 0.46017771363992865,0.49627458077551534,0.6553286480651269,
 0.44420496788087316,0.7158321506252892,0.5162763823052613,0.6057716326882322,
 0.7231878865704061,0.23670686664400808,0.3754611561339304,0.6958106717615536,
 0.03841285428386654,0.6986064673117439,0.983037808627373,0.44778472369327316,
 0.6886986338380776,0.5107040857827965,0.9901713749128273,0.06816417114530082,
 0.39368049380783277,0.13510137998122695,0.30746748060634244,
 0.2536416560126503,0.77316514490574,0.9083090564203399,0.0780770954947636,
 0.2964198242046693,0.8753081192322739,0.7944773386806756,
 0.031100152049495122,0.36600571130123893,0.24063565720171487,
 0.17673143712552086,0.162586627132844,0.7072185251069717,0.5267917770617611,
 0.8257333156008707,0.0024412399250863803,0.03960685390513374,
 0.5098799130686751,0.6631695997213195,0.6640385329099799,0.4137948677286327,
 0.24234799208519597,0.06660542847584883,0.04474889900959034,
 0.5783984962683231,0.15873250772480552,0.31811438709953566,
 0.7590824421364689,0.26708359417020733,0.3631527679961879,0.3056264031559284,
 0.2426900954296306,0.15801193696854088,0.127292181412902,0.21516777271692378,
 0.11507816813908311,0.5647382974449894,0.6354121694820456,0.8435650512293252,
 0.9077660939657267,0.2320183975055149,0.21982986863033283,
 0.33689789517451474,0.9483633793907542,0.03292160349194584,
 0.5601303020527437,0.29246739787246734,0.4545172869575582,
 0.16667961737786796,0.0658323830831733,0.31958643005121723,
 0.10486700369345325,0.8837770073929405,0.7097902720397021,0.7314110911065691,
 0.03671382846206772,0.06824571676938229,0.12780084460652652,
 0.5321378188171774,0.4514765463406255,0.6273292486058974,0.8372200980342767,
 0.5752326871710837,0.3214385583169842,0.5684345508084955,0.4528459522137458,
 0.2989769992442872,0.8987914118652938,0.8615613983125736,0.6736652454157315,
 0.30933756714501226,0.2604666210483467,0.22817292050695193,
 0.8980925123477981,0.805895452180262,0.49390536223879855,0.560592993606861,
 0.19361708401326205,0.06077322780354921,0.1262422187907447,
 0.6227485558644689,0.3205652698502519,0.7845511289284836,0.659507600124646,
 0.5713307833956451,0.875259545752255,0.008267942717170462,
 0.12894790283257906,0.10517790420337936,0.8950780919511305,
 0.5315429239488564,0.1720352793087515,0.7805717610305354,0.44463037225418034,
 0.6977610016332747,0.9477897696844062,0.15040702867767874,
 0.005177664999498677,0.3151334795610148,0.41417966513727766,
 0.7846990497453591,0.3199382293526615,0.1791469818592123,0.40943715694862415,
 0.963542088094509,0.6951173676470097,0.7144225857694728,0.674843942408965,
 0.19608609439837743,0.7481235317483943,0.08945192852976547,
 0.4915946261966351,0.6049156331321311,0.9010480687125415,0.6448475113883732,
 0.19911434127037309,0.9517382550571524,0.8047554013812072,0.6516151355599431,
 0.8804583426815136,0.8394479267543274,0.3448263262942829,0.3928908131792195,
 0.0597543328286706,0.8772448582469538,0.2870020691012215,0.2595004085730238,
 0.40247002865413517,0.7500026149216527,0.2346199700484145,0.709428639715855,
 0.12439720344107807,0.12847480557765967,0.9142569034420601,
 0.7241555569422888,0.5556808414103441,0.4115594895465652,0.9932969180919413,
 0.2433481664939805,0.08362885094736838,0.5238833760401437,
 0.07331800872934968,0.5528587869005049,0.054257666546848715,
 0.9923844029547554,0.05137578148491184,0.8910289630951128,0.6045869081623344,
 0.18901136337719437,0.7156869807528496,0.6314666700887032,0.5644356294633017,
 0.7721634943297854,0.5138943818438356,0.5635156244666804,0.7592663791358589,
 0.12458726405951648,0.8293608058501938,0.849004335835606,0.05914080433570279,
 0.5825503489221043,0.757030685560422,0.30282726679262983,0.1418331231645923,
 0.8610617603670705,0.3668555810476001,0.1957438440223761,0.33842362475592247,
 0.1293535872399847,0.6596782891008872,0.2146363672414634,0.7088885206660936,
 0.7347080754901496,0.02556431475553589,0.9053823363432866,
 0.12218287309876807,0.24256382488078365,0.1610548483076556,
 0.8577822703054776,0.5983120070074237,0.9822658114223031,0.3122995189605642,
 0.645076905261185,0.393201410353637,0.3428492160612798,0.0013473206846372765,
 0.8728651236900282,0.9579350954535829,0.04314865883585994,
 0.024078759047692078,0.579036232672542,0.736502630036738,0.8555586250967877,
 0.7390972469087893,0.3381966013502967,0.940233031226719,0.12215437349220293,
 0.18978887558946989,0.3913799018772818,0.9641425636265167,
 0.23162386667543666,0.5881844337550393,0.1113090083987962,0.5609582827849299,
 0.3844098336601437,0.11266928731447867,0.2716812160712878,0.5653515763917903,
 0.8259960575864841,0.17314788410647308,0.44300242891077146,
 0.17824521142702843,0.913793896904234,0.8948538369277675,0.42117871844858534,
 0.19462588527810076,0.9645296221163068,0.1975731022479137,0.7536219818838279,
 0.40956129914392414,0.4600787181754814,0.6072427935370037,0.6371345946621856,
 0.2554804708730246,0.5977111585064856,0.13350683381088957,
 0.062395983934857346,0.30686179889211695,0.43139439898657694,
 0.43641606616615647,0.2526586236481232,0.14923845096183241,
 0.34759989937165026,0.8420865596906784,0.5724801904362358,0.8423020636958646,
 0.9815333904035021,0.7481747376979296,0.9356240040478121,0.297159838668569,
 0.2787853451491574,0.04751564358296756,0.15327196018445532,
 0.3581966396391334,0.5587405869674482,0.38185055694884884,0.6752792967255513,
 0.9114129930872303,0.45474030169415625,0.17542006646998376,
 0.7931442850555696,0.3751748559545822,0.5055352943321413,0.8560105259448931,
 0.38268620803030784,0.21358860098822463,0.964152295092155,0.5481671280733998,
 0.4161912682356279,0.4731391142523441,0.8196121125547118,0.7110259568542642,
 0.636271019926749,0.8915073198115694,0.7614413883707358,0.2549052626157028,
 0.2381698084604178,0.6752214022035511,0.7422701752106657,0.6155201683182076,
 0.5667542892842818,0.5262792838375286,0.42219673138966485,
 0.18227010327481352,0.8198445371688953,0.4952875583282186,
 0.07232796886412074,0.7990534898844619,0.05652186941428017,
 0.33323467359691616,0.9885830337959224,0.14749033810849943,
 0.4575554416482308,0.8874408671062284,0.14418623450218537,0.6844635542432143,
 0.723734503651381,0.5567120257354329,0.12053123138618971,0.9054977984474866,
 0.4974948514122377,0.8380218197407261,0.32688959948859775,0.1758439955507095,
 0.6662343523052463,0.17311594884729686,0.9970536089262327,0.5528798745416823,
 0.6587896274265632,0.1539871874395765,0.28527301318545983,
 0.08168812245881651,0.7238369617867948,0.7790417985990274,0.514442574172054,
 0.16365346704579742,0.42838513839823233,0.6598992390123006,
 0.16754884043456397,0.34602799949848706,0.7683570906402702,
 0.8873392966004138,0.013081322530201778,0.7764466334393965,
 0.7981109317244229,0.3737990038572895,0.425809984052419,0.9500737984699466,
 0.5543186266416646,0.9201735862320557,0.7388232935486205,0.5542361473016519,
 0.5314947681471447,0.6697655020010861,0.05898032626540628,0.8874156340802275,
 0.36036191250445704,0.8546279853132276,0.26817526278972137,
 0.3605784604551945,0.6204357191233437,0.4857489335011842,0.15633105903538747,
 0.061876923475165824,0.7105568090402672,0.29946111472116166,
 0.7969126226373465,0.9723158714230625,0.17337792009209096,0.5924558960578347,
 0.730896399020315,0.5109540469674232,0.20602441458543064,0.8756419706948312,
 0.39988607549645583,0.5021787718860352,0.6057798045028477,0.4383698434378569,
 0.9956269727236966,0.13126971186533054,0.5726117562357765,
 0.28395693159004776,0.0867679159631205,0.8267133820403829,0.8774414424153711,
 0.19833671918664353,0.4328004770256504,0.5621051053170468,
 0.07305655400839883,0.6484403257966742,0.6018011578576287,0.474784592121675,
 0.8142156494635517,0.5713799089526549,0.8315627135438071,0.5891662803648693,
 0.07304233999115706,0.5652625326497622,0.09254975902754925,
 0.22211754972107167,0.9085183238366534,0.18509019952223027,
 0.6193493632290695,0.7840942207700676,0.023240733523958945,
 0.38724080781404757,0.1193693844142033,0.20046226420937885,
 0.8884846506786728,0.953319804500087,0.6239714476458893,0.9282460089729901,
 0.9050416043339141,0.26432449241781664,0.8340073444423993,
 0.10480154103075645,0.9176823829204276,0.30457382722840154,
 0.2420291915178977,0.5722623525823455,0.5274163957859189,0.8350076219676453,
 0.23868760188321192,0.575332861981749,0.9901524806547434,0.8418213900407059,
 0.04001303665351996,0.6984469473529149,0.01621640422761439,
 0.10610267156169029,0.7093700728859351,0.44887774477852205,
 0.9768794289394064,0.10736362129552679,0.7941919134571656,
 0.005424980740262986,0.008133222297165288,0.2206011715914843,
 0.6709764693125018,0.7116148659534557,0.04877119403891139,
 0.045247690991340184,0.7172854952589655,0.8841845559002732,
 0.8260423455613298,0.32838123403620834,0.02218490948160068,
 0.7736402874718185,0.6941026656229046,0.06979227683746214,0.5599575901850017,
 0.19699463234202086,0.34269489386107665,0.194576346141222,0.8731621796554047,
 0.7619293195786474,0.583361295695779,0.4944380483042595,0.8515770187690628,
 0.25308115604255677,0.12715156559135954,0.3758743968415993,
 0.7617232215305526,0.7362787917658941,0.540499989253854,0.5393945729156114,
 0.08632640519926826,0.018150813365496177,0.24851912949179877,
 0.33140949112683404,0.8276766066496809,0.33599038250144075,
 0.6449197695174222,0.18463394100110675,0.2017527399554988,
 0.33987337229624803,0.9896964426354167,0.8778490206045174,
 0.08777327452064254,0.08149202070122141,0.18840432810645158,
 0.9245247426249951,0.5166435179268665,0.5222445456161813,0.1974700577033337,
 0.03584641910444353,0.8724628005579522,0.1629712237282529,0.4420599448842444,
 0.8658403097003805,0.41167992838569634,0.8265806301656712,0.6317668139174941,
 0.24040901734401,0.09550529869171775,0.7721618073677026,0.3372899971885772,
 0.46284362347565655,0.2698761169737368,0.11427772151131865,
 0.4460380121869182,0.5867451780462279,0.09950039571117952,0.7065876385954502,
 0.6707284826214245,0.7091552736410822,0.9213177992269626,0.8277171058137222,
 0.8243678439024874,0.3406709657196073,0.5139444556383426,0.6335158518853046,
 0.9663846001878135,0.4162923552844403,0.24078341585470542,0.1997035862540697,
 0.2844495038516638,0.08596973649079387,0.3675708275892633,
 0.24623709001212002,0.1419077241369069,0.19352493426051898,
 0.9797528877756725,0.5675073779763302,0.015278443713260592,
 0.12647706588479268,0.5669626469416222,0.660540455955365,0.3148737980405697,
 0.006793816916269346,0.9944148727888166,0.192498259002425,0.8732527085814832,
 0.34150898166834287,0.3690220194957796,0.4030515402765886,0.0435122812665647,
 0.2613481759207006,0.12449907076313615,0.45912910008961294,
 0.8033767053098468,0.8424415630742561,0.6887261770303442,0.7284276721945155,
 0.17647516555596876,0.15127401390252193,0.728877042052041,0.5674859860480812,
 0.00881016170485771,0.3992634661681236,0.011992067265462647,
 0.42219679648061426,0.2713740431829056,0.07436648429972248,
 0.27664461497334647,0.3576724652256612,0.4529689250425629,0.7972212274678915,
 0.8737412853852481,0.8851636416456722,0.634639801086713,0.5185035507623991,
 0.9462897079859328,0.9211881425732723,0.6091806844248533,0.26198257501004296,
 0.4932987871351032,0.7119149376475977,0.16492864844228539,0.4942417383987817,
 0.06144414417257493,0.39466471922850976,0.7576186627048973,0.142438008974898,
 0.24526761548119025,0.2638347788934986,0.3374258301007763,0.6247998464267419,
 0.5104594792983131,0.7498456452306383,0.9471918365574796,0.8303007986115606,
 0.9008922937986007,0.6788849673313604,0.9867319445609621,0.357769916737934,
 0.715894255751487,0.6656147963842237,0.26862916310833374,0.6034808843061357,
 0.9504681169420405,0.39503890157231836,0.7716064125532639,0.2396166831123533,
 0.9412512570482421,0.31192262776164137,0.8616579747897616,0.1226112233424963,
 0.6519472590140444,0.4451835318843632,0.9935901530163637,0.3112833902988864,
 0.8974784442858025,0.47015459710962526,0.07551965925347026,
 0.9785516059190287,0.5327810814457726,0.3323530534313681,0.15205850560746736,
 0.08108034057785762,0.13549158892765878,0.205390140464633,
 0.23774396356962246,0.3847965622155979,0.768559706615587,0.2516826770291585,
 0.3968598441240535,0.22189338851769436,0.5225680064472121,0.9921753313213515,
 0.38304989418035396,0.688460295252188,0.9215219313212845,0.9229298921180518,
 0.5715374815226534,0.3778463683720428,0.17461535066270728,0.4307033822463163,
 0.9033974867344959,0.43545486897013386,0.3601611344931077,
 0.22784040703339858,0.6730001593261337,0.6701817395637256,
 0.43626647159665666,0.2093820603672345,0.020160486777380027,
 0.12968949363912197,0.6483284552068044,0.5870814924314408,
 8.906862147707306e-4,0.23395673049739407,0.8202002744955557,
 0.875741967629775,0.9670698240518447,0.9091138217978507,0.30489644092190116,
 0.044327976815391135,0.8129819195167634,0.5342222939125764,
 0.37483895210280904,0.7988104617744572,0.16022041986162172,
 0.7030411747273558,0.9178029008225408,0.9990580803098937,0.7837272965953497,
 0.8773723718843057,0.13539597630412947,0.0602189314608329,
 0.20617205012371065,0.2299050575531454,0.5530016633825007,0.6425719219589997,
 0.25980994375122457,0.7287575373555366,0.20335914684006318,
 0.8167919549929213,0.9773258479671227,0.35317222638180823,0.6393813753533992,
 0.6940109454121112,0.22790328742761234,0.045831885588045296,
 0.3313119450135551,0.5765319563346982,0.32652168016681016,0.6036960686806288,
 0.5038171373228795,0.893310286612457,0.2472143253190554,0.7453103307746355,
 0.13769745018051627,0.9667379706596781,0.6840925782963082,0.6584459003260468,
 0.4550386624436963,0.8917085600877175,0.9463797729867522,0.8545666811248743,
 0.23463133128543845,0.24470962746387437,0.7620139685729138,
 0.7561293435420782,0.09957458424307175,0.17122609097440222,
 0.8008497521990737,0.6087795915873123,0.4316263180640034,0.03682132712315489,
 0.8269050393301589,0.19083968099162218,0.7259034704825065,0.8953150765389303,
 0.46636172657867736,0.24738400729528653,0.8153081720399004,0.18372366132105,
 0.9855583464341757,0.8269202236093949,0.24350914737764306,0.866829814633277,
 0.8222174893795645,0.2545299445566629,0.40382705951968145,0.8314432950180631,
 0.8785490515833612,0.943065513764908,0.6208151251867418,0.6280255563413845,
 0.1599438977852402,0.8507667187967962,0.009712337967630491,
 0.5318793179670338,0.19078090707594997,0.38346502314887265,
 0.2989006331005133,0.2942262546731236,0.7916049923292063,0.7382193833140657,
 0.24629801544799856,0.37792006266239153,0.9403667616653145,
 0.3919319174115047,0.6580444172856778,0.1347408989854909,0.9297989686103365,
 0.3282904701773889,0.05466319312106127,0.28530595664764746,
 0.40048033831869945,0.5777827850438764,0.9411582594754359,0.8740455294986096,
 0.038888298139804656,0.877520564036689,0.22810028919061787,0.392096210713601,
 0.33348781324714394,0.4920942557071182,0.16267997451058802,
 0.15639107722395995,0.15655938516188694,0.24245617982317158,
 0.5146227502229881,0.09648168148810543,0.08047076590743829,
 0.6043858049147721,0.904049077848217,0.1645934765818463,0.4377617537740407,
 0.2690826069700174,0.04447903387708363,0.8859775418845466,0.3605649777565515,
 0.6347915608086192,0.8075521469230629,0.663379631080742,0.5776262250498682,
 0.8641180493741644,0.4971355769600053,0.41877627108156934,0.9174980817709752,
 0.9421443822649143,0.31520753516613476,0.12758641585106678,
 0.6296759403146315,0.6591439509618031,0.6960204806331455,0.9775100388801052,
 0.47806600268148025,0.4215157873830745,0.8137398265105933,
 0.0034869002439833485,0.609696189034503,0.9740124699352802,
 0.4687626272714974,0.2581268002224282,0.0090547342759707,0.9193023210994582,
 0.414398656524281,0.11391324363395672,0.9235618544819433,0.7927410846603746,
 0.6332827441791234,0.8125696889803982,0.5123640338739712,0.93368639962537,
 0.44516265234310626,0.817585355689113,0.4101092662825152,0.762897457368507,
 0.2379033224876369,0.86625761767907,0.5843075995780997,0.10139042631351736,
 0.2865177968411352,0.2836841828509352,0.8869510899872107,0.08362035824774872,
 0.7246287577591223,0.18952531497435288,0.8736197759816237,
 0.10074117705634211,0.36909894104177876,0.17466728816828292,
 0.4510594758134745,0.3687359609523072,0.8216846263191202,0.9756184592946917,
 0.19739144728548363,0.9300327711040873,0.45114690123856005,0.916507025243663,
 0.06313201089086251,0.24416274008228656,0.6482536566754291,
 0.7674036986992829,0.059833599279170646,0.5790084452234209,
 0.5556535621189879,0.43514781498777455,0.2511931070255887,0.4244861060240467,
 0.3169318203118794,0.8204828427640153,0.7132790318857155,0.606765375450054,
 0.32072331898253004,0.533858643597385,0.5681431266651257,0.32520967735646344,
 0.6131484853991013,0.6182431493800162,0.2661048394536012,0.22819834046065424,
 0.9400225285610287,0.9218921655797238,0.9048163802261063,0.12992852294535184,
 0.31609561927869967,0.24050501463405427,0.3881067756531793,
 0.5059250630699522,0.8158453525025047,0.08141015845858435,
 0.40804568519765194,0.850081216678876,0.2577249564523171,
 0.056933001532685656,0.3598226497531335,0.8875887982379319,
 0.6485101689772335,0.23676259185847182,0.18909167882723332,
 0.31060127244256563,0.6367991778370976,0.21581386841852768,
 0.7248047772590027,0.8554814371691118,0.7336518956601626,0.7101437313445059,
 0.8296918234524063,0.09293332144097732,0.9515434604160609,
 0.46131486119260245,0.6590301386128549,0.03486088120363706,
 0.16883729403995584,0.0448781390092482,0.689297196907335,0.1300430606703259,
 0.7487208881120262,0.9684611016551772,0.802333110903782,0.4552634707594516,
 0.2142634045528622,0.6748242089873707,0.7026978611139885,0.08292437036079736,
 0.3791299092526491,0.27257853181617697,0.9643007426280796,0.3105655384615563,
 0.24421076418050314,0.9641572190549306,0.013341237994588306,
 0.2873188950271267,0.5695877131633896,0.7091729269683977,0.0855946780232344,
 0.48975768857656476,0.9026572013753288,0.7422683441378637,0.4576826410032726,
 0.4226680286822496,0.023022839587162114,0.48343851200285426,
 0.8378339903501196,0.6748082847067687,0.34135744352723685,
 0.18322673366926667,0.27535093090368723,0.8658300665725753,
 0.5412646925403719,0.11947565010522454,0.20370898692555062,
 0.08259322251731338,0.16456230479172862,0.6346186961197582,
 0.31098337694916256,0.17391330864024745,0.4754616978421423,
 0.9239897227189846,0.5456229259608181,0.8850639715289266,0.31598888564514405,
 0.3669994122089232,0.2799765027632377,0.6111208046316047,0.3195994578220491,
 0.7445712019176643,0.19323257741775146,0.8259659997890141,0.6230572757092372,
 0.7210887783871186,0.9399301521115959,0.43636593828538495,0.571364671188649,
 0.5523639842509499,0.9349823855291818,0.3307450549607017,0.4137345515862778,
 0.31587388160011787,0.8692122496244121,0.17783037758625564,
 0.7985104642536877,0.010245876310529534,0.910209805656206,
 0.40140211457504604,0.3818955061703748,0.5785809919838331,
 0.32610389788284255,0.9388145411495269,0.6218560674084364,0.5518659162767667,
 0.4580472167379749,0.45833321611019173,0.6464711473388045,0.0931206132560537,
 0.5008627516649624,0.42245881349293846,0.5589454753946977,
 0.39908513702781845,0.5421628910162056,0.8926389687226453,0.1940592753435404,
 0.5946692122468222,0.6093437492314899,0.31686664961750655,
 0.18336783160468884,0.915012234625057,0.27432435138193223,0.9882438800140766,
 0.8427757486862111,0.3437125577998599,0.2806454172760069,0.46337017790414936,
 0.29972794186054164,0.10896274369366843,0.9505557522232051,
 0.5279701026656811,0.8406366617146594,0.3605514945913302,0.36534149158711104,
 0.40649372763470293,0.43487243318157,0.1374446427698527,0.07717504567548006,
 0.7276688410935932,0.49256263331300887,9.220047902148698e-4,
 0.42344120729822055,0.17067644906655655,0.36040114620976493,
 0.1113642967173405,0.6368481594848163,0.8235853345767555,0.5761385259163114,
 0.47130064352672996,0.285628086803709,0.9423776010430974,0.641547961669705,
 0.7730065632496925,0.8886130272947493,0.7278234361944765,0.8115167530506946,
 0.25856316949346037,0.005467118960115647,0.5163673649296125,
 0.36910336057717374,0.9990377794446457,0.617042538402727,0.3710451520842488,
 0.9574321655818159,0.53492850872092,0.15655589516495383,0.4306585039141032,
 0.17311399678087191,0.13770722567220695,0.0483786398571443,0.44620170472041,
 0.014097287769823552,0.10761738786141728,0.019982862352893482,
 0.26644046759732576,0.16055271732591525,0.9904854777062663,
 0.2813752653797249,0.20393562575863056,0.038944716115717926,
 0.42753293585828755,0.5196292910333581,0.7933994699921317,
 0.45505565374611523,0.47236057814845966,0.9425797889629717,
 0.9801310447881337,0.5273007916561643,0.5848091309099818,0.3733744274988996,
 0.9550978837967405,0.7148795034579304,0.4617539041199312,0.329665940326467,
 0.40530060352364494,0.09510276236827597,0.3473154023625675,
 0.3715574805253544,0.2847835131634039,0.21195515299092216,0.3787851738193112,
 0.21771679876004635,0.6485253026191615,0.16130661142501146,
 0.18089788244084803,0.1374883042769055,0.7179631870622599,0.6865272944755845,
 0.3834065612473556,0.057455025562650586,0.8543267715347571,
 0.2513771760526621,0.9613645330397906,0.09545208330815913,0.8827044309867315,
 0.9473942250816725,0.6807458104187527,0.5711991877123086,0.35244540204804164,
 0.796191986177738,0.9243179652166896,0.024799275017518885,0.8250018643143542,
 0.173140790702345,0.41050312246899257,0.30434974661404346,
 0.10603846425437391,0.7234924778844184,0.537220656568037,
 0.053799392747372776,0.5598767923772053,0.4736168561977355,
 0.8977542627190319,0.47179827391797313,0.43016731021289,0.5444607486611712,
 0.5098188787449589,0.7914946491744335,0.24370843642762174,
 0.45303226360789495,0.7746634215823622,0.05358489550201373,
 0.2475675151657467,0.46004058745970666,0.12566599880857032,
 0.7272930065168605,0.0028448711244526415,0.9908517089397997,
 0.6709193863129874,0.9811365177811245,0.2813417204447446,0.6179539373136104,
 0.6289079726079747,0.7266521578946797,0.9415019805522388,0.8036194778148805,
 0.42591227037077894,0.2676667876778185,0.6468081132279775,0.5725782836104502,
 0.8079828430183602,0.6510579731408724,0.059004054652047166,
 0.4256182793040577,0.37014147933019825,0.7921865723122805,0.7291572961385818,
 0.39822185257563536,0.6533430841284001,0.014560525886311182,
 0.7122173110222888,0.07983496069972429,0.34983164014164003,0.357070886913602,
 0.06943087921674818,0.07893963059573483,0.45451930542213026,
 0.2613359556134871,0.8011077320709349,0.05025193661467653,
 0.17355828622130776,0.8524324950718746,0.02327587962167077,
 0.3506783339614534,0.6202317724325304,0.8925721147592858,0.3225755116945054,
 0.025267880842469337,0.6475846992341989,0.6014091800699002,
 0.20342366888710528,0.06730644131867947,0.7799483630290401,
 0.22447604147537004,0.8181029156886644,0.29431982744776364,
 0.09717549832707562,0.9659945090843167,0.497800565075603,0.6337672201124428,
 0.7275075814035415,0.7203012275949964,0.14422270545077143,0.4610213918168431,
 0.17415767226720158,0.2100311577351206,0.3545053021211936,
 0.10123105677718491,0.5980056307937862,0.23280674094113518,0.964774185891845,
 0.7347583587920499,0.7508134493840399,0.05488706808744781,0.7513932378034683,
 0.016042423602906974,0.20170148149769462,0.15102867184864033,
 0.218742996000024,0.8934798930377322,0.7997415456382503,0.03093608779228285,
 0.7009057217761396,0.9118950851224078,0.886094588069419,0.9798445739078115,
 0.6793826717024573,0.5118125181422926,0.9336765389238362,0.8115398327329936,
 0.371853324277032,0.5343376490547483,0.982206459911621,0.765593982402712,
 0.8977292083180395,0.8483937508949667,0.2959484599098432,0.6483691901163505,
 0.42815023823047826,0.2789715789997961,0.11914612792453227,
 0.9265353567251133,0.6146894364682983,0.3704681445496648,0.7668170731352838,
 0.9349383411710265,0.27838952644614856,0.8133853411216662,0.6801488286871097,
 0.8317991892400514,0.527829503551416,0.579410241922149,0.9476093308884233,
 0.6424627332663545,0.4518949476812668,0.14985151697099697,0.8333767410367852,
 0.3783403348091249,0.7288963391504627,0.35953203329446226,0.9420905515325888,
 0.6797615964159257,0.5895296568448897,0.15317585563965563,0.5348518720543203,
 0.5810696110308751,0.11667353869953323,0.6837196553602075,
 0.42896921830116597,0.23438514062692806,0.4815276403736768,
 0.19145586290328032,0.8737779652450791,0.45513319636581895,
 0.007470550299916834,0.7272064534124247,0.2672788331811461,
 0.5079498238282317,0.19145244729351563,0.18467407671848235,
 0.7085204957894791,0.975833215805598,0.9064561636699964,0.75733466718762,
 0.8538058462897778,0.43578146547959007,0.0011246008813289299,
 0.03258968058060907,0.4729713209782791,0.8349921605744045,0.8143321609415244,
 0.4449377347462281,0.8018104923226621,0.49086260009266236,0.9556524333765621,
 0.9390645020758658,0.7505788379500579,0.7744930917602835,0.6480187812228745,
 0.10555792440314649,0.08107056187765083,0.2060989523895208,
 0.29405772057136026,0.4203059338118218,0.4158799663377847,0.7413087161772095,
 0.23924151280049366,0.31942017719544535,0.7377957663197391,0.436162751823319,
 0.3993437121348128,0.22537972549796048,0.3727409639535375,0.4870351834227442,
 0.9500756778106675,0.8501507338077502,0.8693503947372965,0.7529613774783455,
 0.12334887244408888,0.24192098754823965,0.7273737536401884,
 0.21943810024346477,0.7485951279116733,0.3458430292279795,
 0.46423028656137766,0.01323090943933658,0.4020256310734248,
 0.8074507667786758,0.04314721240768449,0.6308729923264232,0.9447091206935894,
 0.003731348519324662,0.23827111581005578,0.6868043138353955,
 0.021692978837899113,0.1802853234810038,0.4861463336672214,
 0.47054687445043497,0.6721763910319539,0.1329899831847816,0.6465006525684094,
 0.3893338216131975,0.2859895208827208,0.19199331441333578,0.4292546954650963,
 0.14033241920067074,0.4115818445347441,0.21169563372366595,
 0.6532452407980786,0.32979346227239514,0.3254689398935162,0.4233184789318506,
 0.4185263455575624,0.14937557832108894,0.9415261539842471,0.719093097574875,
 0.6693800150318227,0.9874153942842708,0.5928612428157037,0.9687488228191597,
 0.5328289999565878,0.897143250558074,0.38653225836539007,0.09305791360771254,
 0.3691036439779508,0.048660744034360826,0.17781289342158424,
 0.3725095946909044,0.7480112966772097,0.36075867321382693,0.7152040311286831,
 0.18050033156960643,0.13258874925659725,0.15648620681470138,
 0.6892562231147368,0.9957289077323571,0.20401202390612583,
 0.20674555437255382,0.8837055593214833,0.8162884466675729,0.2709350073036264,
 0.48187854415461884,0.630443555593395,0.3182395991793002,0.5712126268533291,
 0.4713842723872024,0.48832454052002483,0.745954358621443,0.9884595774367585,
 0.01658987483383767,0.3335476367604482,0.313740477003424,0.6542690604621775,
 0.3699902971711668,0.1472110395852767,0.8059298254218585,0.5718706875923341,
 0.6592399535667587,0.7504481256091262,0.6186884613601524,0.911862482019284,
 0.8892799100367104,0.6907415047256164,0.5344653604393421,0.3000685380631021,
 0.9500923802852952,0.1882158644183589,0.9222148609736613,0.25461411393353806,
 0.3809886337038173,0.9997569819959001,0.5973077574347624,0.7992128785502204,
 0.3615071533405142,0.08254420852960953,0.20751703175996528,
 0.4073569403507439,0.28921915641545404,0.5659820726244913,
 0.13931195596790313,0.39222787027996986,0.3649095906620119,
 0.18477964604203512,0.6398860989751141,0.5672279713966748,0.642146619336327,
 0.5173240949612644,0.01045343571766566,0.7096073578991804,0.5348440656245862,
 0.9773595076468173,0.40444685080385856,0.03172792048645079,
 0.6228082796081373,0.3059930610551064,0.6329742411983064,0.9324471484363533,
 0.3555036797864791,0.06748484537140964,0.9740819092171918,
 0.13260256290741668,0.5139793742485371,0.596503987855353,0.31496240975469936,
 0.06627035541815285,0.475497931299276,0.13105663623107322,0.7193238137532543,
 0.7204468302713443,0.09236628520572354,0.7997302075983007,
 0.07267146276202074,0.7080839799023373,0.1990795770908611,
 0.16769237182125873,0.6977837737186628,0.7569595728360257,
 0.43723882023619964,0.6661333597597233,0.711203334096293,0.5057086372346414,
 0.017358073571161814,0.07267043729019618,0.5472681848514319,
 0.36844976105400995,0.7330109483046086,0.5124151409100713,0.2645061980069918,
 0.5130005684668165,0.40981505270267515,0.7096621482681591,
 0.31651453257124984,0.4268336780788181,0.4417887701928773,
 0.11498750933776247,0.5694552071500317,0.3264328173747181,0.6786782911837917,
 0.8171590515699466,0.599757064742506,0.16493031825607585,0.43223666479893086,
 0.7603528328075376,0.4076108464243128,0.7366205921702259,0.08464043541786714,
 0.7969383452027641,0.05527831450416665,0.813578247824231,0.7951352549848617,
 0.5532020997222854,0.24530992953817843,0.6048774448999693,
 0.16583179628649725,0.1439408488459455,0.6649351070813727,0.5452849270986272,
 0.29086927266523643,0.05623897567138636,0.09698845412634005,
 0.3736493634546898,0.647798973650391,0.15319839773917532,0.6855284241884239,
 0.45467269938951216,0.4453125966012076,0.1781435859980881,0.5026915891739525,
 0.8271993127815565,0.9774778858890932,0.4179027301625038,0.14730901277015485,
 0.4088629546269269,0.8422922595187805,0.1360319951396789,0.8228443918192065,
 0.5487546513979522,0.8389895308124742,0.7882840862391854,0.0518060656449012,
 0.23137950363806337,0.8296532174170204,0.5543938609795744,
 0.13152562794994282,0.6724397417504528,0.5017115222422939,0.6969329530645534,
 0.3422785598451462,0.07760160466495503,0.10934817406505337,
 0.1673179139010461,0.34339000861345936,0.6854925507086855,0.2776357313175244,
 0.9282745414243566,0.02577013860170463,0.3446971202139114,
 0.07549659311478929,0.7950027032550461,0.9147087244935281,0.807870282329449,
 0.28497019700586546,0.6741853974249045,0.9078839173928814,0.2959271328960902,
 0.7184305380735271,0.1534086897971797,0.526887699937326,0.5534285048800647,
 0.6617866307745632,0.8271226222161578,0.7359557910529453,0.8315048550698261,
 0.8791994181905061,0.8896871873502139,0.25122206516373147,0.2948198971706726,
 0.04539163326049056,0.030244775846595684,0.5681792317132305,
 0.327840856926362,0.17306452855421817,0.32508222916013807,
 0.13515057403917108,0.8170825917692875,0.5099519055955262,
 0.37855780121145877,0.5491328299753665,0.09553207974011657,
 0.8957326370278402,0.7638882522990276,0.4824804149482017,0.39829440604278854,
 0.7335419093430601,0.0435456687157082,0.3784393559418099,0.4229247308997812,
 0.8932095326208619,0.582327052882084,0.3156494808516168,0.2909198713279859,
 0.18234965281976767,0.882949403763745,0.4403867392015355,0.15951590573050956,
 0.5892603214857954,0.8369347585241586,0.9310096162441849,0.6018326229042295,
 0.4346119877513197,0.5746000307442504,0.5214204729132147,0.24393998552026142,
 0.9817947013850084,0.7105074156884759,0.5848312442030782,0.5580430747133821,
 0.14969651260171923,0.8522248015230621,0.8674651898412733,0.601849504353106,
 0.43006344630587834,0.44946705693266453,0.6820976689021887,
 0.8380923654058865,0.08011263100544297,0.2996837017751447,0.9001047178674801,
 0.8752118238736072,0.19128673577770294,0.9699779035885833,
 0.026232758874698336,0.4452266328280634,0.5605016116745071,
 0.14307444071086028,0.4959206220365491,0.24135588078736514,0.997932201153839,
 0.3556280566902891,0.868809166193701,0.7846731426004173,0.018272754267456892,
 0.7279900179401216,0.2799499991492045,0.6081992638567291,0.7540135806704921,
 0.45660242279944174,0.6909824903514635,0.255173373129404,0.6048643178066717,
 0.3649549601996289,0.5386231960601988,0.7850307201125586,0.8919819596559088,
 0.08135971326009417,0.011748359143693654,0.908575907782759,
 0.7276360458661694,0.08108965328984041,0.46539346750730093,
 0.4145848288689029,0.620416862768339,0.27936140554902955,0.801570184064314,
 0.5486992377218607,0.6380606929375637,0.36524858432590335,0.3668680705633711,
 0.5240565850965645,0.8603129328047391,0.18472003768330558,0.7512882624345983,
 0.09787860705444518,0.6359870258891248,0.03564013692770818,
 0.36737989821153283,0.15367493461923243,0.9288642362961473,
 0.6388555985238876,0.8844690430587097,0.882305616286397,0.10472014625857418,
 0.8720040317754647,0.8796089424335348,0.11068824925086751,
 0.47890450813604934,0.11962769766268988,0.5811252242226287,
 0.3272577629714354,0.9632462240058658,0.29189554482801605,0.9590627343940724,
 0.21694950765508159,0.6429328163932204,0.6654748568132988,
 0.43193728859497416,0.8238169216727869,0.7203244833442013,0.3674836587486956,
 0.9146958406258257,0.4547734905191221,0.3129671164324834,0.18953659288293512,
 0.3026953517366724,0.6906098295017257,0.5952508440691853,0.9758538025868679,
 0.713441910670824,0.6332830261879882,0.1855158370708031,0.3624385691710341,
 0.37412386748018656,0.8280942165531856,0.6633824615843866,
 0.05011490785523842,0.3933983422155326,0.8048454583719044,0.6676389307328678,
 0.4114553152833409,0.9745057614311385,0.7940900299109845,0.7155036471187678,
 0.5228459575330391,0.21670390158482034,0.4679538038000598,0.6788256626835176,
 0.908280934238175,0.11685546625073118,0.6363959391157178,0.7613472436522135,
 0.6138117995270292,0.6319141137626412,0.9568708146595643,0.9270309645095715,
 0.34481416122113384,0.13303154584486176,0.6593849448041181,
 0.7470785967427761,0.8296227164181869,0.21519378710846526,
 0.049419120223138324,0.1442222259710737,0.6389071739970409,
 0.10702411490307162,0.9223761341006188,0.42618756501558197,
 0.37184841864412055,0.7184523752100025,0.09009363060080455,
 0.4667498768521363,0.49974386917928193,0.2172440269823458,0.9996556053744449,
 0.45790772451246986,0.8568833034192282,0.2474215570905518,
 0.19945806707829472,0.9767198572767898,0.211358586467013,0.5166489381448409,
 0.81505391075235,0.6806574840604056,0.39639584186641275,0.7593846393240488,
 0.9425344785194762,0.6901868201701596,0.6228252137782774,0.8540535657278644,
 0.7145352702503293,0.5514228330017199,0.5610788853008637,0.295901624234051,
 0.2669925127730236,0.36144397858513344,0.9007450510561761,0.7127006763575356,
 0.5421602067560873,0.16141941825239603,0.118402090511733,0.7190226039411516,
 0.6645769231613567,0.37393752271389813,0.6216363868368712,0.878081622539747,
 0.6226332008556887,0.44881561753383614,0.3957784425241442,
 0.36878656687236033,0.4713704337724448,0.40994587686564843,
 0.1189974653625201,0.7212056613969742,0.9445698251469228,0.3582851361878766,
 0.9466995585163036,0.9000862223167096,0.776132800638712,0.771592122695602,
 0.9308082337443154,0.5821630166957936,0.5372131352400965,0.6006976630073878,
 0.4013871912403375,0.7477776668005159,0.48546496351313395,
 0.16194065936665913,0.17947371131926004,0.7685997471773577,
 0.3242703079003868,0.6651190649047197,0.043706815789927855,
 0.0016209847167035774,0.8968045719519373,0.2966301212122969,
 0.4210154497394586,0.8344427789879378,0.11605972630293959,
 0.12678509504873203,0.2884449285781918,0.5615391986718169,
 0.21327984252786525,0.32610999587110234,0.22815393328617162,
 0.26915141550826915,0.40280413820196315,0.6069211362155806,
 0.25672568446180677,0.019347423451183676,0.762045165070818,
 0.8256328241557493,0.9765863116113357,0.9036649478493404,0.1761517314985681,
 0.9237019071562624,0.5021074762006867,0.12758514520002984,
 0.25490520838686137,0.5989545225765583,0.2265539776521056,0.2264857528966211,
 0.7586506121903792,0.9852754093977816,0.18015247582520133,0.6986365765726904,
 0.880438250331235,0.8734375156401061,0.7552740535508476,0.09193861719535301,
 0.012818453103297878,0.6852677264921483,0.9421253504735486,
 0.9079859027361927,0.7634399375523484,0.2795789014371244,0.2634593361958948,
 0.4699542191048014,0.6425909592933606,0.20530362005219083,
 0.07485502285781154,0.02919112706489313,0.9829428557369632,
 0.5042670944717838,0.020952554139649027,0.42595291355919374,
 0.5853103355007541,0.7526308540694042,0.45905055834022535,
 0.40838151077463725,0.9219147985401394,0.990762696419064,0.2512734834570546,
 0.8898473314741756,0.7843674988385919,0.1030619854796968,0.8154573658098486,
 0.878379257885023,0.4516158835583781,0.3084168824326702,0.5039643077886986,
 0.6813306553139228,0.45668184063804007,0.05593165067831807,
 0.7248655170494875,0.18894913154099735,0.9508746302487734,0.112883967989579,
 0.8191787347754307,0.6603044813325741,0.5124854824162326,0.04818504980293625,
 0.8740710694370786,0.32891420680129346,0.7285411467202171,0.763152600936714,
 0.07417626379005182,0.4102780466569278,0.5414324641521578,
 0.15397973394922948,0.9348704490206202,0.4141329306557463,0.7578054986655133,
 0.9232066818939765,0.9811215227866029,0.7980746264322728,0.1599665178228249,
 0.03522852632799145,0.3638059962770467,0.38910826549934563,
 0.15511317955306203,0.8174567502492743,0.17062324080850078,
 0.3675721804454284,0.48641106533304046,0.9271784624461482,0.9356891479665632,
 0.61733128318034,0.2055452162039435,0.3702914316631054,0.0013929960066939584,
 0.580041679665183,0.636678137570213,0.9036577031524504,0.8663930315312665,
 0.7629190906075836,0.21474439206707707,0.40060941230386904,
 0.8139818797710566,0.014800013171162307,0.5674390440050968,
 0.8015526267306866,0.6292916768767459,0.45716603932896893,0.8512472755498556,
 0.1703341438414535,0.7540369601775205,0.2787185155081312,
 0.012578596068526404,0.6409915109767209,0.5658631099979742,
 0.7626425463442883,0.31622684285193436,0.22621061376369123,
 0.21718758894951373,0.31871051256179395,0.4548104741700103,
 0.3161882157284168,0.08098560359309614,0.859058029012962,0.326569336509857,
 0.35300735679618245,0.9206584789868358,0.7404967482974297,0.2328736891957368,
 0.3850669768602031,0.10376783953233992,0.8594107515264975,0.7429309653063287,
 0.35171864293474464,0.2931777431105792,0.42468333960784,0.778225534763293,
 0.9250534433600099,0.9661323035616339,0.03890601563989171,0.8097102006027043,
 0.3032554760494446,0.018645218645577333,0.6888923745485402,
 0.25441277782731886,0.08267264154819498,0.16858043893182673,
 0.4153366405633082,0.7427831775210311,0.08548332780806855,0.7340764720249295,
 0.9662506460746599,0.263425199555565,0.2678641310051644,0.8685976354025537,
 0.24572245566634932,0.3491485555801507,0.34414826796715037,
 0.9078289125834075,0.4473051813527884,0.6167934007508917,0.6797658975106108,
 0.7099422703039346,0.05079699316195829,0.1951078151463821,0.9718557491291003,
 0.0656271920918301,0.7568891196807678,0.9226796491099516,0.06460819875725288,
 0.8338323606460658,0.07539145665182867,0.4390383473670274,0.8396990561169895,
 0.3762764264951618,0.9674010085822902,0.5584512360260655,0.568157891747826,
 0.6359460748386717,0.4624423302106446,0.6095757681344565,0.31070993544830006,
 0.9667368499969905,0.6821411456927771,0.5682721921978612,0.44930285377481516,
 0.035888893945021705,0.0514506695462823,0.45089483383002327,
 0.3824824194465192,0.25337336656005915,0.6304171703256966,0.7456404628458974,
 0.9031037467015811,0.6883196362766695,0.2844082003643599,0.05015200712522283,
 0.9495974301638225,0.8315954758969298,0.6535867201094994,0.0677248749750915,
 0.9911883641357063,0.6742999598486707,0.7530894484286128,0.7652407323398789,
 0.25762268739204464,0.4937959253254972,0.9626596273089497,0.9077933244552734,
 0.9021081861634314,0.6101273439364348,0.48534156988436017,
 0.016393403467436785,0.842465567377811,0.6803905104324265,0.9825642884488506,
 0.1664389126113881,0.606191646544878,0.117798642432837,0.4308885991139929,
 0.8241988137437821,0.7845265158496115,0.24902892337068883,0.709595033073689,
 0.6305147289190897,0.9801233678315056,0.6259226166234654,0.42493211491502103,
 0.010893460764352936,0.4526317352731657,0.3298277863140978,
 0.42672999610376927,0.027849935697142447,0.25036126849303164,
 0.9646361934390478,0.8209948891399994,0.5386621300675121,0.48483491143333235,
 0.15818847431572336,0.0857667737106752,0.2276632542533965,0.8554369108129295,
 0.15277400258175833,0.8914184344844764,0.24323453777256043,
 0.6883692153111247,0.9749587003432392,0.842633411197586,0.22042093081955816,
 0.6625226056686337,0.03747283674035251,0.47150621229114176,
 0.030377063344751987,0.4466063733897371,0.8495685989842522,
 0.7907127300590122,0.03565841472305653,0.8898467022619345,0.1609681410762771,
 0.14368304228392337,0.15963142825084553,0.6230053070514014,
 0.08973783894888243,0.5157289358426402,0.7960650042332871,
 0.027843305006417074,0.7581413137280664,0.2741061882794088,
 0.5478457114084395,0.8428183888764931,0.9968642392776992,0.8900401892015388,
 0.4682260777610807,0.022511096593507696,0.9127930483732978,0.948744359114135,
 0.1770476472491347,0.3863563185005344,0.5821609971783707,0.2678079005450096,
 0.41612348042376146,0.6259916413108857,0.8788294679404532,
 0.09264110580173535,0.26534862837945916,0.5258452934397104,
 0.3934591619845016,0.06514031200680515,0.3515077344174782,
 0.13995022139818847,0.5662059710112701,0.5875673339948151,0.5834144811217892,
 0.9498831381030759,0.5345608051693672,0.6052484714657187,0.6448808477361102,
 0.44511110229954587,0.33434313077852496,0.7547095735201403,0.737096779123435,
 0.8874386815537192,0.7984431742660746,0.8069606646724161,0.4228765404800304,
 0.310125135689577,0.21111709736718542,0.5706172251794593,
 0.028411253367207068,0.531282333877693,0.6533155132210118,
 0.030150600444581288,0.5578647336670813,0.4413055812541842,
 0.3844974271013789,0.6499434864902357,0.5637231657191912,0.9538055219779452,
 0.6320130628673418,0.4453584491652055,0.8317973818780872,0.9104318279395474,
 0.7232414716211913,0.9906934231564912,0.7202361857739596,0.48916014465941415,
 0.3716732986498502,0.14582627728213482,0.409957522440182,0.587226646929247,
 0.798515126359792,0.4120543871793534,0.42269979107203703,0.13877533553956634,
 0.4265823066140727,0.2643429752991586,0.9149132863738462,0.5118452414730438,
 0.9393005962592217,0.02436711764588284,0.8399675573300727,0.846541124715353,
 0.24169389483807746,0.9302100629727399,0.7588952357146457,0.6279539082307694,
 0.20699438379108237,0.0387338315011827,0.7582312246222999,0.6120859927018761,
 0.9543422200139605,0.5247807792600816,0.27896999068440154,0.3001173404950279,
 0.7164150729863142,0.6530534138248006,0.09651843011536498,0.7798953788139551,
 0.5044237550614219,0.645994740201065,0.9186744855665051,0.7687461243666484,
 0.5641708128498282,0.6970479388059554,0.09056403420016101,
 0.009637689303783725,0.7627451248794999,0.879814308652806,0.4374161823574617,
 0.6077170506173719,0.14121845073884376,0.9968033683167911,0.0870067753620477,
 0.03918412832315288,0.2914293273041155,0.8774235540657065,0.9014131352809693,
 0.8230042462219693,0.2793779733215569,0.8766813003768075,0.659788345277712,
 0.04835985055300651,0.08804435783821207,0.3417557619325948,
 0.3307312214775966,0.6903937875059126,0.42772143430908494,
 0.009099300744221828,0.8202241082423137,0.7207926785092429,
 0.9437095815123111,0.20378782738446422,0.35523141778383716,
 0.8479577762652362,0.8655368128337807,0.2967300434176243,0.07708775292652936,
 0.10164739320372085,0.23171246833205839,0.35487034594323763,
 0.003034164205493295,0.9940535761906382,0.45536595003244196,
 0.9499570458082689,0.5861150818664245,0.6108463409624401,0.6206870792053248,
 0.1508350969788429,0.9595226298329983,0.045475258279461306,
 0.1476716261100378,0.14356075437821036,0.2562190509041994,0.7830127222594556,
 0.8577647274003514,0.8446594430410017,0.14571063983325538,0.7035826984988777,
 0.20669212810846682,0.2314227504339933,0.2748926922365794,0.6716158467436968,
 0.2965822361005589,0.014549527025113695,0.2755317108045998,
 0.7130958213638854,0.8535216515883128,0.48125859029302376,0.6403987145828729,
 0.29917054453367753,0.10954837431617093,0.6045304168533314,
 0.026886409993050364,0.2958999706065111,0.6959443429191116,
 0.2646718599612655,0.8863843703122286,0.049108207797474934,
 0.9175392732310582,0.5279007084033992,0.23552627794579428,0.8109302721578393,
 0.8916474064308835,0.6355785488688248,0.009463502254876688,
 0.9848587428415461,0.43806212837140235,0.9475429526057471,0.2930238147227485,
 0.5996238739792388,0.6026806686790811,0.6347032492004144,0.6270171038253499,
 0.07875555783397425,0.12568718339635898,0.33097220377093883,
 0.5126026510908577,0.7775282167641986,0.7237784042571191,0.364779330194845,
 0.6973338424382755,0.051978795229475794,0.007495166326128788,
 0.597863576945174,0.8447601694157435,0.09504833253579681,0.7974376049449348,
 0.6760598770958485,0.6878560086343929,0.3749736213365523,0.5153238370500673,
 0.003947184667344006,0.08857120275098662,0.14602776270083095,
 0.1535180367504847,0.3475156567149885,0.10157080388288353,0.95640222804613,
 0.7208624920907809,0.7358476793209876,0.8978256407381124,0.33444226943411626,
 0.12238145486642205,0.3073150863579456,0.28425432223179714,
 0.5154224080635809,0.7593018239187312,0.8182107350124935,0.4451318594314042,
 0.5726581985977018,0.9981790799873684,0.7775256244561239,0.7977259386250392,
 0.45840832856970815,0.9107844397409558,0.9155169908523614,0.7915041924331583,
 0.39414501034379446,0.5187693212194027,0.38441458973258313,
 0.31872726099839277,0.2511959190303207,0.08679660295286329,
 0.9728404984039053,0.09952263697652608,0.39453076879838156,
 0.5010824369005723,0.2312622660012038,0.2880915477902579,0.5492380946041117,
 0.31296726851945567,0.9561112370997762,0.8917531247120246,
 0.19951422156277387,0.8548897526521883,0.8071491875437197,0.8849642939491038,
 0.6794231809133207,0.5035103861536797,0.19683500791964348,
 0.09812084495544304,0.6851203901437176,0.20060598053527967,
 0.5572115269589474,0.3919944891451699,0.4040127921362344,0.23930558106373012,
 0.021270357684686103,0.5262261459135746,0.15224665025699657,
 0.1039776253847775,0.757584168427621,0.6951177998507383,0.22180310726679453,
 0.8057429900797026,0.4026243795794958,0.3440441420309637,0.5825877843376268,
 0.49318680436326723,0.26315956455245215,0.7686529670596789,
 0.5137073175912396,0.48324086993868676,0.4021875869709597,0.6787359742111483,
 0.7372649956911936,0.8115705034629161,0.7372042034691211,0.73385931357787,
 0.3102746107594544,0.10770207917231034,0.9146088316477538,0.4080612827606457,
 0.8080556810651872,0.7679154233403718,0.2749342998975782,0.6093017001789552,
 0.7736230478932142,0.6818326068900805,0.11721275205880377,0.8086614671212246,
 0.08661094798255298,0.7625213938292046,0.26846495684264904,
 0.6187658900236521,0.9577497845938421,0.6142959027994551,0.7262445469764702,
 0.21382887255848104,0.550088832483283,0.18562646398168525,0.1117066525591297,
 0.3643860431239542,0.22405431023775946,0.47260951582997135,
 0.9198101643246902,0.3375257436597918,0.8966582211651274,0.0321301409201028,
 0.9164165011492722,0.4262889414700709,0.7694595108280105,0.8301036122981816,
 0.2484053921491457,0.820795333352812,0.1534370147816071,0.4512246955951801,
 0.5470441280954587,0.30783733258116985,0.9961048533181497,
 0.15513358679221723,0.38576535842212767,0.9449005553609027,
 0.9301821753547701,0.44238018178982186,0.9098547690262632,
 0.40734580574005597,0.7656947848931892,0.22739205996940492,
 0.9806439460887373,0.0907066481659502,0.6222138608537339,0.5272855884134013,
 0.6924806996543833,0.3084021711840421,0.9865440841459439,0.9482851132413752,
 0.2102440376544803,0.8169975515672115,0.4423156051875202,0.5381996869245576,
 0.33809474601962264,0.2515207858997903,0.31994418092291377,
 0.20674726203007587,0.4989138827738555,0.949372982948006,0.2923977162742205,
 0.310456224688488,0.9325904382796675,0.013397116784240715,
 0.18322855445204855,0.21551725479002248,0.2301428851752264,
 0.7585976597498518,0.6288382048849569,0.7515208843000895,0.51743372193869,
 0.012865465617311278,0.5762938241814601,0.4188740617679906,
 0.6641959753254026,0.17680886317483857,0.6154474580048878,0.890884076960422,
 0.5984547447347697,0.78952145604975,0.9262702789689063,0.5032377374164544,
 0.08946882352540364,0.34508273828859803,0.4114479254925786,
 0.19305610173166077,0.23279264988151782,0.32711602784550664,
 0.553289034301679,0.0373606692027042,0.6924065702031068,0.6086722170650749,
 0.5865617409986537,0.6355169209735098,0.7390800067687395,0.7014042504424703,
 0.42666817189797945,0.6473523073761115,0.42987630537393995,
 0.5046132992743501,0.5026158749746361,0.6729065619001218,0.2803912942434934,
 0.9437702037066911,0.4089095654624244,0.6415448994334059,0.9012947648223294,
 0.7657323286719886,0.080078044949917,3.897709466782473e-4,
 0.38601141381846293,0.12240769015039032,0.7110995452850092,
 0.19557407626203327,0.16381905276642983,0.7749072958704917,
 0.2187929727177229,0.06899211252841264,0.09499761589303646,
 0.5226584830570784,0.3026648814404911,0.1784801555136878,0.605859861577755,
 0.7649948592732309,0.5487311672215492,0.8717236561901154,0.697078960392684,
 0.8150365463001195,0.24617308624199052,0.27512716870807397,
 0.8910874505330788,0.9289767341153332,0.019559299431223365,
 0.9550410752074855,0.3790936267510705,0.255040842215257,0.7888222559447837,
 0.30238132835513953,0.12672482136744878,0.11478856594101638,
 0.27712343836729536,0.26161522734934817,0.1650655744339633,
 0.3115012093599395,0.3551214090737044,0.13974545049621045,0.640871658239237,
 0.839309751854241,0.5027421129781724,0.9475758528730187,0.44065036780386413,
 0.8563990939064717,0.3898904477183808,0.8774450173721502,0.6469246654789451,
 0.6297027512892656,0.6226038197921053,0.4624127058269112,0.06642749184478003,
 0.5221446809957179,0.3623870145221899,0.5018846344088699,0.04766114041370617,
 0.8489492159833842,0.2726112481258526,0.6733605966090638,0.44910279447876167,
 0.06146311210184552,0.3118048699774537,0.9350737542142012,0.6806665712003117,
 0.900183296295025,0.2990552622495506,0.5599765677456423,0.5945304409555523,
 0.447400690918936,0.09270592709062231,0.6616078475680891,0.5204828131157858,
 0.15060916525953705,0.15607672037293063,0.07771941360720813,
 0.7066502734086009,0.06010149062362635,0.8960741102692796,0.4202764300198124,
 0.449056784104244,0.6158310442048821,0.9333742506435602,0.21892490221908656,
 0.041133998562776775,0.5524379152452502,0.2455406281496938,
 0.6496562351438853,0.7018049306484251,0.6402814897217766,0.49377552870942376,
 0.2980159146985387,0.20965114255192674,0.9813537573785984,
 0.14114317227947693,0.7857211710209031,0.46339650315153835,
 0.4306350592867638,0.24550075206974398,0.2739397053902699,
 0.06007368289449888,0.152305359080113,0.8201662827283269,0.7214336519120018,
 0.4489228613724404,0.4858233141317553,0.7505868782677928,0.4693581037571666,
 0.10002812643911829,0.9221794350862768,0.7592529855807693,
 0.31511826515701924,0.8946623383802219,0.45934509319469885,
 0.44499943703803424,0.28933120060641615,0.5577426944319454,0.674415137145145,
 0.06246516501824351,0.4060922563634767,0.17785315672767055,
 0.5641483033547857,0.37676690246553135,0.1312387778153652,0.9998052896611628,
 0.7234506821682171,0.0919815897749039,0.2982639891816645,0.6845728671067945,
 0.9861036166455481,0.5885304183211935,0.921478512237437,0.21039858983605841,
 0.8469816623951879,0.4930080156116996,0.10220895853223211,0.655047229974554,
 0.6250887471384142,0.3542554011353082,0.05486627940056055,0.6925236665179842,
 0.1762130236381354,0.8286325894958315,0.11931446462698081,
 0.21548087020072348,0.5282943130296647,0.35882416322522936,
 0.47516783266265084,0.5608703402643063,0.19845021442937982,
 0.47973952244712836,0.29024969670144285,0.4574462326474614,0.670907066638156,
 0.12409573312544286,0.9407965142155951,0.9332187552406136,
 0.48794912920922084,0.7793604396596464,0.5215947856301965,
 0.43573013655293846,0.06785356857424907,0.9539358830458576,
 0.1923513794307582,0.3024810621633913,0.8383699002444653,0.541935994382062,
 0.02651674655130809,0.9988089210736911,0.2943335763189122,0.249004886766245,
 0.7750498528238686,0.12790751092876151,0.4655141329306799,
 0.13507438541561634,0.18236496150919268,0.06179701322697417,
 0.32176147093079954,0.50147055030133,0.7325893470432275,0.2733379626419594,
 0.7447735802702251,0.3604682959086418,0.31864947992105264,
 0.057543545138666885,0.6355252631979542,0.45124211669863723,
 0.5829506003116267,0.3457950217026342,0.8460899987847259,0.3196896501476142,
 0.5327732402777868,0.5993317858839489,0.12215085193383113,
 0.41359150857661053,0.15737926535765023,0.13296268832076408,
 0.5768014129507064,0.7838194624868989,0.7646268316405584,0.7014666770064099,
 0.31485453533214003,0.8055835431398398,0.28483918471554626,
 0.1244650913657166,0.012487131619492442,0.6380574167179325,
 0.2601984999793091,0.7610366215783466,0.4643407430977984,0.547557269225435,
 0.9463364016665975,0.8822126769187457,0.08710489935494092,0.908367267664845,
 0.06756584598448778,0.7913075723575275,0.7754012131720756,0.529207981633348,
 0.4844465556143186,0.6489000155243915,0.5884813292588432,0.2394840292764231,
 0.19868443838376393,0.6019500282128228,0.2720164546059677,0.8977512972075778,
 0.8738782726586027,0.2071490533174204,0.6648939214264639,0.7664687032420082,
 0.6993937617530737,0.4183927586792865,0.7237146527587066,0.891583635893014,
 0.5162697515132803,0.2798682911529262,0.28984137440279834,0.6868085435686434,
 0.5291127403333991,0.44520890503142163,0.8519305504808886,
 0.07068763398301514,0.5414820455974041,0.6616418808018444,
 0.12233856113703145,0.9742207419943314,0.043613186759544176,
 0.8937104175493321,0.06251074011696511,0.37624641109636925,
 0.6748456624600521,0.19154368251305243,0.852405314721528,0.13468574087100815,
 0.007607856186412909,0.17221708721758933,0.689857770383387,
 0.04725082294751415,0.1277275947702068,0.33651689411033237,
 0.7891221569425371,0.7987853140387078,0.6080175632524941,0.11289145309683268,
 0.4092814549207562,0.8261205964944451,0.18135673096736127,0.7844997364798768,
 0.3927489289715409,0.34986886402738127,0.04485294104488635,
 0.7659187963274584,0.17478206354287595,0.3806740833516713,
 0.43291262879373116,0.6322632928357889,0.4332117341122613,0.4124299337145332,
 0.5197215614526693,0.20734050477208055,0.03464316654338495,0.350506409172916,
 0.46570532029533807,0.31416898479611144,0.06298518648864715,
 0.10219253872023581,0.331459896194364,0.5410905548043448,0.11181106249612405,
 0.9263667997288804,0.5306303131139108,0.16535416545743375,
 0.14360564683104104,0.6827456182523354,0.9218626101609294,0.7187691452240901,
 0.46830157966786246,0.1402562840855085,0.5009513179038174,0.3765861596868868,
 0.6384402691880039,0.31256566355340065,0.04625614458505278,
 0.7167120424337801,0.9196199707382372,0.23553153869224164,0.7819915521269287,
 0.9535739456627778,0.15525980027273956,0.18891783715745958,
 0.9925511683018262,0.9514244008145119,0.8892714583572523,0.16747904786816248,
 0.7641443860053342,0.8935337642691981,0.9750050623034292,0.3410261486739208,
 0.26315383631723055,0.5365527057265652,0.6533501753965494,0.7198257777232261,
 0.7644484893889864,0.32246699314567384,0.9640917478018303,0.8264767099358252,
 0.5965931821121486,0.5526840888825105,0.06409112905881031,
 0.27052714923144816,0.8579470810690896,0.9917374494657518,
 0.13837687985804592,0.22163582443112084,0.3703980265083411,0.287331867545485,
 0.3306515327330496,0.421384528883913,0.13204915520367844,0.7276287133258088,
 0.21063936006551232,0.4972191274561204,0.03968654183628051,
 0.30225208112031454,0.7576118631354247,0.3392383146349649,0.5287507436484479,
 0.6902605619636673,0.36009796173094677,0.4390496600098075,0.473212208026355,
 0.9935809312397553,0.7614723167124443,0.23972856488497962,0.3441320766902819,
 0.21430943930549806,0.8519783024959119,0.1355604740796187,0.5319146119744946,
 0.1352469735653843,0.508737047505007,0.5499040658224825,0.484067765527617,
 0.4667304097493864,0.5132080959712014,0.43538519744740267,
 0.05567315065556533,0.4807492060198204,0.21535341218394266,
 0.2043058068958442,0.6735261466660898,0.20301056585365584,0.1928080483873087,
 0.347518776388857,0.4694886644726932,0.0465727519996918,0.5557873059050529,
 0.7327108731256204,0.7558657980034391,0.28814771768452707,0.9934689292411585,
 0.4417728018162522,0.721376304970847,0.4601104310829852,0.9836211036710919,
 0.13010328896205792,0.354017990850483,0.559164745842937,0.05425706526399643,
 0.6846117690723744,0.6136644345458557,0.39597575013634934,0.8521830668200796,
 0.1355682501895139,0.24927354924866607,0.7514041103032215,0.7698918569671931,
 0.760219760504413,0.38899179306735243,0.02185284502517504,0.2610610442308231,
 0.9224709768431005,0.11826038145305229,0.09743455807887735,
 0.21395415549520147,0.9433203394373407,0.5101736389225653,0.9136596128986438,
 0.17222873825703033,0.6096489356520144,0.9835773641159826,0.4487953050031107,
 0.600301340446855,0.6876513446010469,0.5495885565431903,0.3561642142733492,
 0.8581951669842366,0.6071997232111822,0.26463109320664624,0.8436196009443329,
 0.826049168233805,0.024167692257909712,0.9911867403646285,0.5821751350263388,
 0.8609477730720277,0.293824240885994,0.3282003326558003,0.46606701556952346,
 0.9824528477603989,0.8028875140160968,0.8485613754657167,0.18162457304930069,
 0.1502775053112967,0.7724689202240315,0.34800890345646895,0.5003808777319182,
 0.25566256203398674,0.533742917919428,0.9023124795281612,0.07755148496956632,
 0.8911770942202053,0.02714785429500799,0.6959794515766072,0.3601105422299209,
 0.2900707787148158,0.48496718651568194,0.6652632519144919,0.5566088201859503,
 0.7811132403028461,0.04037099429310931,0.39883427567558505,
 0.9048329395511503,0.30280952615986256,0.062327937719972026,
 0.19372945113884532,0.1307340037483944,0.4445605984767166,0.0660138863247145,
 0.18685422188707268,0.17361765805573792,0.2185691905554038,
 0.7823718740309394,0.08139824979787502,0.17052089201858012,
 0.6749929540928946,0.8212204361666868,0.8300294293392602,0.5902676533180908,
 0.6606864907173837,0.6420756657650598,0.3621858497405427,0.23922742360889715,
 0.6187257778473508,0.6074661233749793,0.6695893973064305,0.02279416221394248,
 0.288518360152098,0.8591125864531848,0.8610649482734339,0.9247131236598197,
 0.8945519890977348,0.0739533855522495,0.3703473665675573,0.7276455966440257,
 0.47354462817519194,0.9183068071592669,0.5614412737088,0.9243819949376565,
 0.2661024550775992,0.6978598420866458,0.10727963570960397,0.4812464922034906,
 0.3642478128159117,0.6248459016155743,0.9159041456310364,0.05763014506489794,
 0.10635762269741789,0.9617135728401558,0.7348096241142517,0.2576656334701265,
 0.5661623649777949,0.1290722323641561,0.8048242709357273,0.2191931296292402,
 0.6555935052856217,0.9232738386179613,0.8707143701161619,0.9754706429170209,
 0.5417747453711841,0.805423455184659,0.4885583982360546,0.023752811202397695,
 0.2895781074189442,0.7541073587808949,0.3125166850357619,0.6199022254727538,
 0.9361521660601019,0.3418998756005591,0.7965789770353753,0.9173733290159894,
 0.6339771544816768,0.6998093338259526,0.17928684938503414,
 0.060867725066987455,0.8686566638092622,0.25217840692144733,
 0.49913446142396223,0.7455677147537109,0.8596499996144439,0.9329710474728451,
 0.9757644252934229,0.8075349636772204,0.8987569863829526,0.9415510395826588,
 0.6755232776110507,0.8657761878597936,0.1268791403333298,0.9779729752671491,
 0.3940469178403776,0.6345212063967666,0.9398276709434956,0.8468689693636268,
 0.6394677266027862,0.7887921358725298,0.15860717266138025,0.2526382976807182,
 0.9593424506338863,0.30370684044411944,0.797669173560589,0.9128664912207363,
 0.5587508141891835,0.330092142496888,0.6898711777372107,0.040775674708283804,
 0.8018210240928698,0.33689040154143746,0.12645011645236648,
 0.018637164646914783,0.7926270288893467,0.887278066296725,0.6738462045907831,
 0.5132731152242092,0.11657919136122685,0.9453204019928334,
 0.24130759138476154,0.4811941558260211,0.3381579396614306,0.9100683994608265,
 0.6015494312119674,0.34375663341453044,0.8106801281435009,0.5182193600782771,
 0.13002428806436106,0.8841556349684155,0.1024034567010339,0.7105277747901201,
 0.9635240960553809,0.6257038703998137,0.3430646941539721,0.715444351226715,
 0.5478706354194375,0.6499062199661161,0.1064157173331709,0.7160252692696977,
 0.6686386108660503,0.5671090810794894,0.20681271572691795,0.542037824337226,
 0.8087902999662642,0.5992047522536128,0.6652971765835861,0.344073010336845,
 0.6193876284699081,0.39700662956954946,0.5493533432115865,0.2733558999566462,
 0.5178799640979262,0.8072935492997356,0.057768326422271365,
 0.5741051703379675,0.10630794890500761,0.20876885560826064,
 0.5209565846265009,0.8689703269016954,0.8083088059746873,0.7558445914443835,
 0.5200122871852475,0.3833949453831006,0.957359306600031,0.1573741157456806,
 0.1989928558945444,0.14996897949484567,0.9267478996558004,0.6204572164684288,
 0.010643855861670493,0.15242317157729146,0.4686272222643808,
 0.9975332922272047,0.7290893137042769,0.5141293217009635,0.9384882249026989,
 0.6443167447961777,0.5898046041083564,0.7790030561295395,0.2702557410309493,
 0.8423436082588158,0.40188166314963647,0.6629722460414984,0.5664744539421169,
 0.9017295518988012,0.28940649211089553,0.3073972657581515,0.7529532924627025,
 0.06750567310619093,0.23725345868736158,0.09220981898976988,
 0.8253828015576288,0.3705400460085798,0.5302630268396405,0.7711570093344519,
 0.0656245614380857,0.17402148963861164,0.527010717553213,0.7987615880673441,
 0.03326040166082267,0.8729435581036733,0.7093472041189409,0.7511693178722945,
 0.10240794026158528,0.06119357329788522,0.12802813134116997,
 0.1370406179177448,0.24976409734851857,0.9579331640528109,0.8711666480189938,
 0.5790393046405706,0.22392171430572538,0.35611335306982606,0.451208254128754,
 0.053021458128836074,0.4333093165977754,0.8660119133249363,
 0.8675445096404579,0.6616446698677909,0.9029089875407589,0.6034485098194907,
 0.4315285689837218,0.28569183813803045,0.9515117571183893,0.606902643928946,
 0.9968736747879922,0.2277876461756715,0.3553806996207113,0.9071525776065446,
 0.3988416835628772,0.9015096050546888,0.038164886437313195,0.959322184618651,
 0.42843022513739326,0.3713078579078507,0.01893257766224954,
 0.49406972295705154,0.9364607549453906,0.08393412125816291,
 0.2655126570598374,0.876072273138608,0.9377876947333678,0.1520135370821195,
 0.7166105730208132,0.6549695117649657,0.6697796573042343,0.9614588855683606,
 0.3498443506710255,0.516019349369159,0.23815110650201687,0.9241202873785899,
 0.8676060312421922,0.9854239971324598,0.3646933229028818,0.25244572622700656,
 0.4658940099058668,0.7396870382225657,0.955444286483881,0.9335569063950041,
 0.5441181455075771,0.8766317789681758,0.7120713285936475,0.9177853515277962,
 0.21965431829503623,0.8844900039497006,0.6441621187862986,0.4527277103823326,
 0.7267657502244065,0.17709906292236532,0.32426443978916053,
 0.4114769020444554,0.5004332069602773,0.3607558430436356,0.5390802589487965,
 0.45600156767136635,0.2052500256389379,0.9656149168020246,0.7242391758156268,
 0.9505716675626457,0.5593958618020656,0.8402427954181187,0.05038040402620614,
 0.17192587854561092,0.48787925847869584,0.13772697628589503,
 0.1675279037244951,0.2503036520655182,0.2673825353401317,0.5515090649924466,
 0.6450249113697217,0.8399940831997998,0.5728199958763093,0.4168751555041207,
 0.7685473710007016,0.15329059405965428,0.27503795868501524,
 0.4334937776710909,0.9839875018205393,0.6407056869454872,0.12376646749541731,
 0.2716285892830633,0.03582969840330741,0.031589114510700456,
 0.03393487589499622,0.9005731295546932,0.5306081008021206,0.6192977623437409,
 0.12348689023688053,0.3199942797681552,0.8813463679023998,0.8599076587347773,
 0.20252309272058155,0.11682598145282908,0.9460137531503954,
 0.09033897898516674,0.3167652430375615,0.649761263720845,0.6550921116054519,
 0.02340613588647522,0.4170690367392548,0.6024363404730856,0.5923597978049853,
 0.04826132567106878,0.45641195715111726,0.06783043619120255,
 0.4927428447355644,0.4277160661627851,0.28530483876444224,
 0.050509917059665455,0.5303128505429162,0.6524739506762315,
 0.8831192026879562,0.8320547143402284,0.27110593519698867,0.7799387275801044,
 0.7918817268790094,0.3102861608470433,0.6175024735677543,0.5667398683583903,
 0.906839931720365,0.3235880965793222,0.861456166483994,0.02998499807034649,
 0.8019608876909234,0.722037014538363,0.02726678247401504,0.7896574312326052,
 0.41732267802384737,0.24699692495182912,0.2300356705646236,
 0.8905694567183208,0.316888478300171,0.07141646789532263,0.2906919156603577,
 0.18436271202607946,0.9444461325931421,0.7843154283831285,0.9793174844272851,
 0.5304560259162007,0.4298659041623726,0.7662648401265177,0.9934487447515458,
 0.13253865078392357,0.3060365082349097,0.47340907768839746,
 0.9273374141878216,0.6036592857452279,0.49390732092816036,0.5942140996282361,
 0.20293873666903062,0.5053056971960012,0.39554038562695715,
 0.4789727198109591,0.08460016075723076,0.8368977463218716,0.6390759422360683,
 0.8007434953760314,0.32887069098074195,0.4486544398614295,0.5399894442407062,
 0.7632769723282117,0.9643498761576675,0.4967558745276278,0.7293437089727972,
 0.15982700208670864,0.397845173484523,0.4067785872678532,0.8948037837497465,
 0.7599710613005837,0.7646945937435841,0.0394072324813417,0.8007118668479398,
 0.5480076550433925,0.6879332334260879,0.06517869198680071,
 0.43903733421790714,0.9424934176855894,0.887969487923312,0.7343569692254858,
 0.837809172760099,0.7353039354056308,0.3092067436654271,0.5130849624184819,
 0.8558768667470835,0.19292404528173834,0.5720048058778323,
 0.36446653177624144,0.2332873786852575,0.3122297641218692,
 0.05811825923107039,0.4769362814324559,0.7338463344095973,0.7549186628558704,
 0.3061196190978036,0.9067968588802037,0.21351959815980126,0.854595641544662,
 0.5484791240232648,0.06628583302127689,0.23284844966012064,
 0.7665785239362563,0.12382017371889775,0.9883730665420971,0.1271306313918934,
 0.30598371429368587,0.15834678014386738,0.7364659633608857,
 0.02716259421254552,0.930042076135144,0.7530711857730811,0.926180368771517,
 0.9211749175213807,0.3599558604365123,0.6993053063341891,0.7670079873824538,
 0.941969868019486,0.46959864881712887,0.43689078554847405,
 0.14200554042543034,0.8695185396560501,0.6871194552592809,
 0.04927927778533403,0.6250260460979529,0.6022719680220144,0.1750639777057741,
 0.9315440615169862,0.49985780338211994,0.5066001087284209,0.5721705078393315,
 0.9204033882315501,0.5212422433357486,0.5723186103191752,0.6526032419416344,
 0.4136130156315373,0.8664228321610865,0.2699113137680581,0.6900435675269744,
 0.9326395679983244,0.1443291201880822,0.8468285442974524,0.8930794511221228,
 0.9355707945864853,0.6313181529483016,0.4433503858972738,0.25878692194662245,
 0.7486139961923935,0.4477940905772507,0.6952977273498256,0.32467198313320433,
 0.365152265190934,0.5900069295992196,0.5204046205972382,0.2693288028805496,
 0.7667652523864342,0.3447383709428221,0.9561779726475053,0.1539271569870062,
 0.04762193667344128,0.5681837105647561,0.9985116178429764,0.8254614891436163,
 0.2629268463211304,0.9074763363552536,0.16917983416114313,
 0.00521132165153082,0.36628233160125023,0.087817423759597,0.5047410459810059,
 0.40261314641137613,0.6388902005731341,0.9242268443655691,0.7170468647274946,
 0.11745769618851232,0.6912124180407729,0.7015858405809796,0.7390973640762741,
 0.36905860338483243,0.9268305280342023,0.7639505353445923,
 0.20183939348540214,0.9038476590782831,0.9823520508217726,
 0.34325356445410726,0.6767777482831567,0.17137825170421894,
 0.8412157637198585,0.15734495170141172,0.6884587083121572,0.6773904826938326,
 0.5578484160407629,0.5655663968563097,0.030344114783294907,
 0.8098275840259505,0.9248395454743732,0.330241653792936,0.888010615207734,
 0.1972012400401013,0.5016392335433197,0.7088665823153028,0.6317962288038099,
 0.7835296490651168,0.8440922826259977,0.4528498543826027,0.8088418171165639,
 0.4460604449161202,0.16643112584581765,0.9234774350907397,0.1390004088185265,
 0.7578463213524607,0.553663172925088,0.9634467285237983,0.7826579889752351,
 0.8872159292476651,0.00727320802671616,0.16511996050708921,
 0.37800962722752995,0.11896207313532436,0.7569470981246007,
 0.8562984224010827,0.6917836283379147,0.25161474837379805,
 0.25553928161297845,0.613708903117783,0.9861060411901141,0.9564417181395026,
 0.43094389018088086,0.6021882990362999,0.01451991370986816,
 0.17527534909836762,0.1329638393853988,0.751451197007646,0.8913329542900414,
 0.1951266477087975,0.08623534094186625,0.6557405903466982,0.7874537662460194,
 0.52073298939577,0.8976264937597658,0.5788427344098028,0.8677022385342602,
 0.8974377041836467,0.7477653976694457,0.4121651362371661,0.3981018307791221,
 0.8821898047335859,0.775812607239841,0.5846285066831064,0.8813705250706784,
 0.2814800433369695,0.5635404861257963,0.6202331892293772,0.291035997183138,
 0.13182055905750212,0.8573755675460109,0.7302672196571941,0.2323142092310142,
 0.7801730137472063,0.5521426237009943,0.3368635174813861,0.9235499335044773,
 0.14064844779851327,0.19989098857490606,0.5206635487843617,
 0.7051791306139306,0.8644171197475504,0.5599791404561911,0.8930662164094774,
 0.1351745454679869,0.46844380102149585,0.2853226964615191,0.8982954960891815,
 0.5067336766738547,0.30772821869361433,0.5316232946632385,0.8351791204380202,
 0.8625084839782371,0.614376393455052,0.8832188196017082,0.21966220682012194,
 0.02259057052415958,0.9761913806443667,0.32388372959699097,
 0.1323561360161445,0.7877023101075042,0.5284100224180082,0.6194617625851047,
 0.7816996094549712,0.6414107928766092,0.8023608747516532,0.323542843009897,
 0.632497734536035,0.01552807786406707,0.24314142100075975,0.1627721449079188,
 0.5939628453553971,0.06573262936587909,0.9089504220210206,
 0.19756566242453877,0.8180013529913436,0.14962879601419643,
 0.019462364795788334,0.7712071865812223,0.3209901082199047,
 0.7119307449243202,0.5312928299780897,0.32830697386372576,
 0.22951514896559744,0.6822971261331423,0.9024741137410467,0.1288806584348624,
 0.8967252886576113,0.46492300489620575,0.09905801469613795,
 0.6904768684875164,0.589552866253093,0.5658219501263851,0.8618536411372624,
 0.1446650959372946,0.20181110163489424,0.6202982916704057,0.7151906795026568,
 0.15872004364886072,0.788298096995564,0.5102244074591061,0.9956925639253388,
 0.26722980664291474,0.14551542324752587,0.0637255928783953,
 0.4480583831292547,0.5736495040732028,0.9972191685241741,0.49784690867599646,
 0.3406440280976799,0.3212938239866162,0.26769213310206996,0.9707486487733564,
 0.8454572088148256,0.9785922060473725,0.7885741896974453,0.6332846589901776,
 0.6567485148200887,0.5868841111115338,0.27118941916407846,
 0.32737240651951327,0.2511246957942087,0.38603700492652226,
 0.10690591347620193,0.48713043896302666,0.5501169474067176,
 0.007759339058627024,0.7086743775816345,0.8040128584674558,
 0.8148609401379692,0.1071457813370078,0.19412271210737708,0.2118730196713302,
 0.932433047735069,0.33672751948313684,0.768824489057233,0.9172216003907573,
 0.8703153362236183,0.7517978623699575,0.567778998317254,0.8300143164928921,
 0.8190141689253982,0.587176525444613,0.6386513498051513,0.16316125328382824,
 0.6966356131858408,0.21845945962007918,0.47434868831073107,
 0.37244727066690286,0.4730703724929076,0.16098849415867622,0.82549328313297,
 0.10713257679810162,0.2882140870746702,0.6125615859522896,
 0.18463431503931083,0.11445691492992738,0.0440367080681785,
 0.6540547650108915,0.8375715552349157,0.4786316965428312,0.21103855706879948,
 0.6453677633663065,0.022186532816968696,0.02939708556484577,
 0.9639665826179835,0.9896442185496699,0.8870708467778359,0.4472568996035271,
 0.05375438754007156,0.5513663718317048,0.9973286451973082,0.596788691074033,
 0.8310703337121145,0.963283611662368,0.7911839238664334,0.1265135767144755,
 0.7283508544129155,0.33066260044945506,0.07341908153249599,
 0.12771379792185245,0.5470556925561457,0.21848954181086755,
 0.33289056624430624,0.5574816432636659,0.5807623777152605,0.7845671067982277,
 0.08796218561672198,0.41391089930246183,0.4801893816011521,
 0.5620350959144829,0.5832600153370455,0.08480261957863666,0.5729002261023951,
 0.774521587244155,0.369077121382239,0.9232375036739611,0.2661800782615713,
 0.2073461444056266,0.25197822100115086,0.7813475430128527,0.6008912036381079,
 0.9513186022947817,0.15094868380523274,0.42479656730796334,
 0.3964317400572476,0.4606238222897767,0.8884553367624917,0.9480263091052801,
 0.3111693642449871,0.6023848338475517,0.13235152783370652,0.9578630689957326,
 0.8017725378853819,0.3040010714706559,0.46369675200431026,
 0.017856293396507072,0.558932797933551,0.5893945600874104,0.3336445681302612,
 0.2419196579320566,0.7905918138769623,0.9844540089057858,0.6102128608916041,
 0.5582074796718758,0.38771549568421904,0.7168808950383014,0.5514392068666402,
 0.05061259520129968,0.6565932947633482,0.3149797593165685,0.0921089965012839,
 0.668708302379383,0.3483372957624411,0.18564978745601157,0.2762882069587065,
 0.27158491078767644,0.6677603176422342,0.1349148798349623,0.5411643539597152,
 0.4338548815933986,0.1761713373017323,0.30384321259712743,0.9731101464296255,
 0.37769372393657286,0.35704538828895305,0.08236513769636855,
 0.0075150575915038376,0.05953467215301467,0.5515127760579213,
 0.21113985154571013,0.0807098366534269,0.45289761110836846,0.706800279327175,
 0.6121555270409771,0.9544762343549806,0.7860948332418551,0.08047864143954664,
 0.26905425768806923,0.09724982316811726,0.6487230023225367,
 0.06462916215716275,0.7057695629892445,0.4130123629764555,
 0.11361032736281873,0.48957591510917664,0.37267598447505323,
 0.47591672707643917,0.434428684694801,0.34571965290533235,0.5670270961460272,
 0.20865277683706762,0.22267815747933817,0.8911778852897553,
 0.6223493215437137,0.8701519850882331,0.743257676694071,0.3046810932769968,
 0.4101747105397521,0.4264801913841566,0.3042909323284628,0.7722083448794295,
 0.8345339486167892,0.655161967284899,0.6884101676267879,0.3508734884053084,
 0.6737120692843299,0.13160192478885202,0.4147503889907448,
 0.06254395731283802,0.6701321741450204,0.9231422898162027,0.2992526475096594,
 0.40493675869203405,0.6009357160420978,0.6787954179779161,
 0.10883267042323252,0.294090580765786,0.4568592894239698,0.16673779368240393,
 0.3302183897563409,0.43351952333872545,0.32958208574433656,
 0.7792352836684112,0.3615806735656948,0.4240159545032297,0.5730561194892936,
 0.6201198948805058,0.6156240428753338,0.17498800572127982,0.6583624250730167,
 0.7706357131066719,0.6472862593403255,0.10776841396488157,0.7434936900677483,
 0.48492135018666727,0.1255428913454516,0.9757016560971133,0.5860058686497218,
 0.1770297853773175,0.11197150930496802,0.6410653492523767,0.972194957697873,
 0.44623622872511204,0.058179961526127144,0.3857751384114996,
 0.6370628209569766,0.9907284608061468,0.17672076283001537,0.2434911928558956,
 0.9069525874957336,0.48925356610438897,0.5066793188791467,0.5435468773481766,
 0.9101054639835003,0.7135845628133307,0.29328465890953515,0.4689366422201974,
 0.1500745843342539,0.7544298967304621,0.9658792039821321,0.6938361194870659,
 0.3812478013155284,0.23947939307807042,0.422769579339467,0.11645886217704382,
 0.591927330459464,0.5414656484499644,0.03054212877822593,0.539766311877194,
 0.16223673318744214,0.6098950266613704,0.3316402803028563,0.6735070868485047,
 0.35371085445399786,0.008474144571579245,0.6101792504635102,
 0.36673394474518395,0.8624818028022239,0.927455051324507,0.7590610252443251,
 0.3115463326816341,0.27925321733753605,0.7748267166866323,0.5754396181165468,
 0.3414675273409462,0.2725197684945918,0.697097668396488,0.226012303373902,
 0.08332703386103979,0.24378287263433462,0.33967687955392434,
 0.47166486604547453,0.9422317090778491,0.7955986503553977,
 0.37411144532685203,0.7059374653983961,0.8173848812652479,
 0.12525314164492163,0.2665952722493262,0.11328489309997913,0.031613969259215,
 0.47065345882241916,0.3936219804226766,0.27969374909748423,
 0.8726990036299812,0.37269267661171235,0.1994777994011736,
 0.009278154968664798,0.5369638910694295,0.3993336989631252,
 0.7141343008468937,0.5717412359627848,0.17107872128653268,0.8159484131978146,
 0.2641628906982918,0.6748364456582927,0.9644680172018802,0.06553039205571642,
 0.9593309899289884,0.5182274769303056,0.5906299848082573,0.85947608529988,
 0.6274033842743577,0.7685289645377007,0.5590537246199387,0.9668379020574999,
 0.032170498260386204,0.09852162525607644,0.6800328613730382,
 0.9438173438814284,0.19316331030323752,0.547277967671818,0.267448136800409,
 0.911074111046691,0.668861817289145,0.9948589971530974,0.13139472281569753,
 0.05167114776817061,0.23163406319282476,0.26733430786419965,
 0.13888780327667094,0.3016325294932308,0.47183729990931844,
 0.5624669050185861,0.360512314638772,0.4054060726707671,0.1447596188239475,
 0.37382035912027456,0.485124677040768,0.6031165161444936,0.29230729205946726,
 0.19471308654232944,0.5896219100557978,0.21566740587155564,
 0.44139360650723714,0.06208566528933934,0.7449456889493925,
 0.40739051637308676,0.19592521697246923,0.5664563846497053,
 0.8558777830072642,0.39059669430708954,0.6194306651272355,0.3341335419901206,
 0.3465399586569533,0.8751565851978332,0.5190884088255902,0.02627799727925928,
 0.00817023358458624,0.9752247714470998,0.4094520419379549,
 0.28673906125037496,0.5473900883815297,0.2862010921026321,0.8799308225319562,
 0.5985645425956723,0.28605630642867574,0.788372274602535,
 0.030887501232638725,0.30026142832137626,0.9271607809403115,
 0.2321880360260289,0.6883902029898522,0.7374792055290635,0.7166976155973128,
 0.19253401769985412,0.08239737855002705,0.9010808365933404,
 0.011677944711210753,0.4698373196458565,0.924317054605165,0.6800119758289979,
 0.3187081751959856,0.053347120219138455,0.4645942033570908,
 0.35153676519881294,0.9253265020133705,0.9929702786275932,
 0.23167374848682998,0.06683233209256267,0.5904431005015246,
 0.8309641797126195,0.5892131473446793,0.9070854468835071,0.8284913931026082,
 0.2717678564778431,0.2148182707773617,0.37513215475137307,
 0.03462558140901495,0.35784962840137313,0.7893213844194268,
 0.19427285089372126,0.753416053949785,0.6079706842118611,0.5196565978793903,
 0.5971584151116567,0.6446357348587871,0.7842238374296732,0.23493386672802263,
 0.07192958083547851,0.33016898683492735,0.2196744765292289,
 0.3556714040433717,0.2111285431043688,0.9518201114778042,0.2271002931367636,
 0.5129773879704839,0.8591009909192748,0.1471973885808895,0.9428568656527364,
 0.4854622194877152,0.18441650472631022,0.8649595095610914,
 0.37497232842405526,0.9620492447949353,0.8657103291428778,
 0.46915517693447406,0.11376282704878737,0.48519668245539493,
 0.47036193135721016,0.6893467336317426,0.31271895577099773,
 0.04234193308322631,0.40019520892784755,0.07643117321059545,
 0.06352085518472439,0.7021340300215712,0.6463379679652881,0.5781461106742168,
 0.07434618531431203,0.8318978963434186,0.4651062052415542,0.9191815034873138,
 0.7646731302160068,0.3738869590473912,0.034446060596176276,
 0.07017020207833535,0.6700239977647936,0.7354797957404159,
 0.02927628684798378,0.6645692736562414,0.5914996763694487,
 0.034308181143609895,0.9547353107990011,0.11321914654840304,
 0.5752686517923511,0.6010893000442428,0.39637614044638936,
 0.37393209593983456,0.10926730017690489,0.7394975702275132,
 0.17606770169973773,0.02937853433117421,0.21688464609863878,
 0.5969956351657452,0.5995730179624812,0.787666649329813,0.9457148781471405,
 0.3891254650122019,0.6598952967969552,0.4507291493450526,0.48103177095492333,
 0.2245592223736479,0.7952899766138067,0.3036050437641439,0.09024045978750117,
 0.5711095475647356,0.4339695367454862,0.4874376541586689,0.04577713590750143,
 0.9428462837492619,0.9660419771191161,0.512135508486773,0.7476122074459772,
 0.3055334169402154,0.48216969749980676,0.35827055596659085,
 0.5020795824344928,0.3634361492641405,0.9143513458923578,0.1797098575402506,
 0.056590226336492466,0.2974024190682013,0.1355847206995634,0.171870154773187,
 0.8198679048193614,0.45840321822447316,0.7123274730401091,
 0.004178007558298336,0.7639233151167567,0.1359676970719571,
 0.0467814021491062,0.31459905338942595,0.8191277981697654,0.94424741209237,
 0.34798040553921505,0.2903678410397935,0.7723934281769544,0.4998019110556877,
 0.6069946469372719,0.9393301224975281,0.49001991391579847,0.786042343639625,
 0.8104222202010961,0.14509384038589268,0.017821299615912167,
 0.32917559177292566,0.5153336195944348,0.7180757741051799,0.3712068623991841,
 0.05227984155009746,0.6207207996610347,0.59367900832151,0.5746547582766983,
 0.5665255676042713,0.6760533782366147,0.24818734516775454,0.5016279486847133,
 0.7540991024571371,0.11402609920365092,0.34160136717095646,
 0.22855869242153792,0.9491262945679347,0.40346870583189287,
 0.46842494940998625,0.3098188153011344,0.40064080990183515,
 0.5804380745448074,0.9718461988883862,0.15266313708835577,0.5207792134159843,
 0.8019797385023357,0.4619562954701385,0.8488178993987353,0.32629785649449583,
 0.1701599518599597,0.8502369065710865,0.9744686439751307,0.7203851658518946,
 0.8140778199237355,0.9856622763782867,0.313403146193183,0.6635732448754667,
 0.12687807389477634,0.2567870178507098,0.06338719898597511,0.616683938824753,
 0.840651500185663,0.9544365358383073,0.47285989904636083,0.28513820370440524,
 0.4020301248257049,0.2521587612848363,0.9121777276412408,
 0.030010841137026123,0.062074304629397004,0.177699448652354,
 0.5871322882983506,0.6288722949619916,0.9089551921377816,0.9998168785752265,
 0.6730547462877435,0.4571460751798717,0.43938409044411186,
 0.38892971586883274,0.038988344581922574,0.17248906889828763,
 0.3870123000355917,0.08211915842739692,0.6790221774246011,0.8422553218411073,
 0.48737790003225623,0.28343969220381615,0.8637986554184011,
 0.8290814547658742,0.10268636575491252,0.005686386405648003,
 0.9588219552705372,0.844457117283783,0.43122148130265203,0.7787433553479784,
 0.7603113236927509,0.9019719255353645,0.590063338125537,0.22828366532734445,
 0.8434562805388323,0.7158568794594875,0.836947346932722,0.6087870771284046,
 0.6178116068480801,0.5376729632385135,0.27599336506838934,0.7367010512436016,
 0.369138720464364,0.5310231015273532,0.4681616934541477,0.9999509640698528,
 0.23607629738174984,0.19464715594634918,0.9688364535870603,
 0.0738466337796061,0.05106079705289779,0.6048423075901433,0.7526126396840981,
 0.44162989360333527,0.9664041753760457,0.8072443404064652,
 0.20661569075608766,0.26214957799976324,0.44415294487976087,
 0.6033964575437332,0.48514616391757626,0.7187420914394711,0.8987622406160347,
 0.535015554417273,0.8903220720714846,0.01810446696209489,0.49486240673926873,
 0.04374927189703348,0.11391201289200259,0.2739630225424541,
 0.5060174329380107,0.16394555897218732,0.003832579847087292,
 0.14515748350053292,0.09808084158195973,0.11672393117903168,
 0.5418058859549731,0.8414069994840052,0.7291499669468925,0.8518216297071306,
 0.721220075200543,0.21586869993144697,0.12303041078250898,0.2989512133167592,
 0.3997037310916436,0.8735896310379361,0.8853297531296838,
 0.012794853944401607,0.40890930692734984,0.8532512801928612,
 0.5743984304672164,0.3262225758119145,0.5233550513462303,0.6085012152890261,
 0.06380586219021667,0.48731948451127294,0.6375796160329106,
 0.8813536273933806,0.12128512502418065,0.1576505718953657,0.7923211675500005,
 0.6544837470355943,0.5773564683000632,0.8158187261714264,0.403591001764787,
 0.02731307255523674,0.9669379609087576,0.3224697550226491,0.915254862443558,
 0.8982305592443811,0.9956895478595149,0.5241462015061185,0.5145027678044288,
 0.3097227302462563,0.4170103109928328,0.49648186174749076,0.7637295466910128,
 0.0702138301966917,0.8973054806908138,0.45876356197863166,0.2560215439023662,
 0.38702795975701376,0.7102867883186923,0.9698418305726688,0.5336291740931627,
 0.908948698779174,0.6588794028166085,0.8493291674612379,0.32979277524322304,
 0.5534607155413503,0.8989318917509994,0.4243505026515547,0.24199834208229698,
 0.2541161676138952,0.3810132552564547,0.662542234140201,0.7702339167309004,
 0.5587496100718321,0.2481592110196329,0.78149273639048,0.10433255605493863,
 0.5415727578824937,0.042499344604972955,0.49359734028423996,
 0.14948270646746875,0.7865165494085523,0.8645163722111252,0.6781386413381421,
 0.5391255611661646,0.7660439971303474,0.26977612129378736,0.5859867072137495,
 0.8827136826950899,0.7087296399987982,0.6710800695650121,0.2791197759367813,
 0.32469927137634547,0.5690445108580581,0.8103599481285548,0.3369571196757829,
 0.17667801706762032,0.11473857502955132,0.6404791892442333,
 0.5159257275485646,0.4677952216498291,0.009555829679508077,
 0.3532961512487174,0.4395768724764335,0.8084431806134433,0.27241612518539904,
 0.8918578165926782,0.8438829140362193,0.7782987512899903,0.31660393573066714,
 0.803278098029987,0.604571659449448,0.2944309148311237,0.5450736659035607,
 0.6065652149916028,0.8623470308786527,0.6342888845446912,0.9737023582504428,
 0.654765513497022,0.12196607017286798,0.6303360860868609,0.35569825380523634,
 0.43364597369466984,0.5409381807589514,0.9394304066738273,
 0.36376196143773476,0.15374816938129088,0.5356303369766655,
 0.35511027489077274,0.31345643840592485,0.043857938587759215,
 0.3402789525242894,0.48810541783385386,0.26920844762001905,
 0.5773097106004965,0.7753380901700576,0.9296515947566129,0.511947076314262,
 0.20033198691909937,0.9839199105686467,0.10708408899814703,
 0.5081912934108045,0.33364273796712673,0.0266046089501355,0.6811290116334915,
 0.6316887006761756,0.5306728076353352,0.5376743702680221,0.617173006253281,
 0.7711676867778503,0.5237197774557117,0.4124671883222498,0.46528713133211663,
 0.781333173444818,0.33470558968491926,0.87221734203147,0.13586787151881718,
 0.2761711402501976,0.8476041175577538,0.4968159392439183,0.8307891531007422,
 0.7538433873462526,0.6063419388180742,0.9058035615934437,
 0.028548428346018007,0.9237148868549939,0.3128395867390701,
 0.7524655002709028,0.10553698810989032,0.29223381990618447,
 0.0389022890190327,0.24966589457130917,0.8410304862436455,0.7386258555117848,
 0.3620736493932457,0.4134445017082857,0.05591300479675332,
 0.026796329982796463,0.4160670579789947,0.727695398395204,0.3800645087204937,
 0.35307977982111205,0.8963114521274403,0.8367970890729054,
 0.10078794174836625,0.47359713848374874,0.2855677476946,0.22544091648178932,
 0.44071114909089615,0.8193324634202761,0.4478124807345625,0.2957975328542678,
 0.5375519835936314,0.7479542005425143,0.24259351963956688,0.6874447291377606,
 0.3256305504134708,0.16746629263806456,0.7591997388881749,0.8954319485006872,
 0.8375065872825312,0.9172105038993925,0.013487410015520007,
 0.9964579144705555,0.2015403493468677,0.5199411083792208,0.29833867884146503,
 0.4537271014621117,0.3138779872692167,0.6230238191996755,0.16633226400977574,
 0.3718065732011968,0.9261904542851529,0.4248871850612601,0.6588225127661954,
 0.7233380003966656,0.03704293178780227,0.5822531558167859,0.7531999125598807,
 0.8040644970864194,0.775276836624491,0.2992298466564156,0.6005361162262648,
 0.5615363673239545,0.5378418621400227,0.2836198335407305,0.23731555019014194,
 0.01782737605124418,0.3039337139875964,0.9950305036348893,0.2679353189649235,
 0.5786460363817804,0.7315590001118779,0.0761537190711864,0.14213328509308232,
 0.5675843925386669,0.5752187794006691,0.5190916211290469,0.478086511293478,
 0.8749759763394871,0.14069450889149693,0.008744585219262468,
 0.8218089387087391,0.23737510077701407,0.6065006854889108,0.9837308878135076,
 0.7761244670684524,0.8185340459112033,0.10069280621813848,0.7898548208962873,
 0.4380901542450486,0.43056090006415015,0.7222954336768329,
 0.11030521531192417,0.6287949041310178,0.6249119288742143,0.7722839392963758,
 0.14830667150559118,0.4140555067437528,0.0341732263785296,0.5611264008689405,
 0.9139852425917082,0.6770476760078461,0.26601249770456226,
 0.13121356185907818,0.2718986533310769,0.0706091423377535,0.1262375588039646,
 0.2513075833793875,0.6312402110313459,0.04293727960856575,0.8155754414224441,
 0.9275582321397254,0.49279559670504613,0.2771231584449597,
 0.13087782430358552,0.7276803221727693,0.8919405483179439,
 0.19956293575753836,0.09842568434155574,0.4007600664573542,
 0.12177605508886835,0.9111030197650145,0.32797440329834326,
 0.3184017861179562,0.38897730284060417,0.7185704749558423,0.8523134458945305,
 0.7700963428120077,0.06764239507235759,0.241172689827594,0.5250359794209206,
 0.8264486241146838,0.5104562614137449,0.12267793823233886,0.9403898086357381,
 0.7592433858554501,0.6054139625237164,0.7882613000461149,0.2238961435698844,
 0.336799541343076,0.7244495181972698,0.012277688023252464,0.9108231818948689,
 0.8171072472840778,0.2870381446998804,0.8322273201735485,
 0.0023266392933897206,0.029327382893988063,0.857766062167439,
 0.9878016645745542,0.02526791880654089,0.7384157702993434,0.7984326715185128,
 0.8964851808123145,0.47868789742211704,0.3008041812693608,
 0.41825203613720496,0.7183142500067738,0.880385928445206,0.7831770788724572,
 0.6343190360139057,0.0328827532399244,0.765652825577828,0.5300215824254063,
 0.09301484410590521,0.47581036969801604,0.14350718700259024,
 0.4516960791642587,0.34744965631164626,0.09998207767748724,
 0.7696955201576475,0.09856638549566987,0.9699687030278095,0.638887772055376,
 0.1368777231299817,0.901417072858456,0.6251645888414255,0.9143751933221983,
 0.7363107296872567,0.370842216275893,0.5463559974442056,0.3937157518809348,
 0.542543866957395,0.2625683296252407,0.20747980917967035,0.3630501780339812,
 0.62931153434831,0.22367109860472267,0.06449889167882239,0.0719927165664378,
 0.39189799694794414,0.8016645399930701,0.3245319280172655,0.2648083018951687,
 0.24152931154521873,0.9570801730405385,0.5563847737466158,
 0.12577603968826023,0.6912999688087269,0.5705419479792497,
 0.20831603364541573,0.015213661096864284,0.7068383102242183,
 0.2014905158097693,0.5809748123363659,0.9863126851929261,0.7993014781752531,
 0.26760680580296126,0.6131525366944667,0.5860882957977079,
 0.05378937509605708,0.47586425357063056,0.7328198578215959,
 0.1940010208862305,0.44593467940428444,0.1086815965479847,0.5406685402297351,
 0.730518690936503,0.7027956233104525,0.3972484282216485,0.3720251963467486,
 0.39370546933896444,0.2114962645910866,0.8236453707875208,0.7099353884155034,
 0.673497587614266,0.7297192200759586,0.5186433974891663,0.04184848512903616,
 0.6587411511387791,0.6482225317698154,0.22180841368673387,
 0.39652101047155963,0.8910289917811701,0.7795600920165682,0.8478649811820365,
 0.384022560246286,0.32275658183038114,0.6558789096376023,0.6951473818145708,
 0.10796871344783465,0.02109930874227961,0.27704383472127836,
 0.4506532269265494,0.4229924216159342,0.6632008038836343,0.9343198191356932,
 0.7879985445275558,0.35413046353490396,0.9845922121513492,
 0.14645287107535687,0.843662767747384,0.4436873190234796,0.24510242792807335,
 0.8326256732517172,0.5948784962168712,0.5536127494874041,0.2771580739103502,
 0.2707226945820652,0.017130412184954147,0.023456413426279354,
 0.14334135772983314,0.21429814646538825,0.431056164002252,0.6228822198437041,
 0.7961514367056284,0.08930357756190832,0.5873726545287229,
 0.09252614102510748,0.19821409411730762,0.14107964627000535,
 0.872502708648609,0.816813658214255,0.6381869776104833,0.48914864576510286,
 0.23519080901445055,0.3312577004919387,0.8180853300130098,0.7714976780626216,
 0.2340348649373759,0.9129419709079307,0.30245097370860874,
 0.0014971068929315878,0.8727688841728625,0.616254378490317,
 0.7852898245841385,0.4898340999258862,0.03975108816688766,0.1985646261915236,
 0.9903913323809199,0.3407839726879218,0.9546080391289343,0.9101889149000756,
 0.046877754989598897,0.17028266127468428,0.9945280985820601,
 0.28280019666466394,0.32663452457664965,0.8265614712677882,
 0.3933181411081331,0.07758704645765413,0.15666608039546948,
 0.19930772340735692,0.5799787863802689,0.39373388057465375,
 0.6913288482078862,0.8684191645462642,0.9555411055555286,0.7044803082028727,
 0.1461324199464804,0.7649773190105486,0.5624519936033789,0.2802989814796043,
 0.1849668093067316,0.043001863841986454,0.530987497877131,0.9999833445433518,
 0.5334663516268718,0.8234966857550589,0.3031701826529972,0.20766740630167047,
 0.4804127158501429,0.1170951882781206,0.10410968962667422,
 0.09143666887576851,0.3307125652725851,0.8263578335229442,0.502105942965989,
 0.12658931963571818,0.35212011767095563,0.9842541690095732,
 0.002112486028780114,0.4316216302190733,0.17248639208796257,
 0.2766214822602173,0.8158245283411988,0.43379380674857904,0.818274852412803,
 0.737055978620428,0.28186060895525866,0.0395339167471912,0.5539311969959159,
 0.550248723860588,0.2643652351614174,0.8572567998219973,0.33584586300273134,
 0.46325185888573095,0.6181207226092582,0.09332856896758679,
 0.9918646519601811,0.25050228494358695,0.9560114161032973,
 0.20662466842236882,0.9531486692759653,0.7741947688935481,
 0.48490278962492317,0.10638713481931128,0.45051504141912413,
 0.9722529232159802,0.17552080992771235,0.7863125435616942,
 0.37513171109207244,0.09886503534515645,0.9085684937304163,
 0.8677260178437145,0.4399384120162466,0.7913285013239144,0.7370432172522343,
 0.2868034513778316,0.7787384811682729,0.4136563030849195,0.9220346730617746,
 0.9992171822501736,0.7950325279210727,0.3147773252516505,0.2434577310062339,
 0.18071987994316374,0.527963580840381,0.18323260552799803,0.5545957142203884,
 0.7947835129462845,0.7857437514496919,0.43901685523566836,0.9002736098649944,
 0.5674380008119383,0.4348048536029566,0.720980530965274,0.37921378880656986,
 0.7662648255826785,0.136014496567902,0.8782031997923834,0.647111163491921,
 0.9085168104111075,0.5756975309345592,0.1738994472447556,0.5619808816428277,
 0.3535561028822799,0.15893648320915754,0.22563041218467839,
 0.8067447268142313,0.16459839357889772,0.2884412276861057,0.9581117646029886,
 0.2057718553255674,0.038596779970293316,0.37392467557815845,
 0.7110311836540105,0.011481893713308322,0.7574512870492514,
 0.7183635784304943,0.3536190519792609,0.7692529364716508,0.3726914911791236,
 0.5125599067709508,0.9154352162325642,0.556511134699209,0.1662557672908943,
 0.27351844992609486,0.01676583214522631,0.03269470355447268,
 0.6196898015735102,0.5634035513776776,0.1422168156200847,0.7638615968961376,
 0.6958978612983823,0.7627380397540359,0.23328693368729214,
 0.46893444189491884,0.9355527084963502,0.4353503582294753,
 0.31817287068202593,0.3385756408326641,0.09715201007257357,
 0.2585175578344904,0.7081211212270515,0.21235827782280925,0.7099046304817285,
 0.2655211051014891,0.49798159293880595,0.5260173761100287,0.1682180769216739,
 0.37530918307117944,0.5353928734014695,0.5871677588749057,
 0.019409385711282034,0.6810813429480121,0.9613815711394501,
 0.7424666174114997,0.6506227713821445,0.7796133231243779,0.6507217542501404,
 0.7963913466763958,0.2699739565497947,0.776858385967745,0.3218356619073892,
 0.7218474144736068,0.4721788972117068,0.40261007921826397,0.783407805765608,
 0.023408833770123594,0.3523346418485933,0.6610358947967991,
 0.6928607674003573,0.2607294372614595,0.4262850912711985,0.0223315871819999,
 0.04450112467831391,0.9884272670669149,0.6770442829288505,0.2665638743160611,
 0.6268376994674798,0.6238378516166686,0.5226290017320424,0.596108495003129,
 0.8158699818346703,0.7338711253862381,0.3919919237943553,0.8187010344754972,
 0.18203075078032183,0.7386436555709742,0.8061596184594976,0.7839204106651144,
 0.4434870169274545,0.26813712248682875,0.7137625364091191,
 0.41322583608403474,0.20159862027323638,0.5172281818520791,
 0.46540760881861976,0.10136598150048859,0.20908277612551684,
 0.1940597149287504,0.09435513428231612,0.5900447720787197,
 0.14857437470209978,0.6592560738446982,0.7725902361369901,
 0.49291116625563647,0.9373311946827334,0.34531087628401436,0.475494880239959,
 0.5484607759226975,0.8322186023484633,0.5129173782757642,0.6687579643194733,
 0.30431253084484766,0.4271840308832864,0.5164427653244285,0.8611294461986809,
 0.03914449219291527,0.029236445489978835,0.5972912677617017,
 0.880818900241113,0.6846046665076186,0.3147992391652147,0.5447232790545636,
 0.7209113198523505,0.23549692285145296,0.35264613029999226,
 0.18132727231002976,0.24801912682420424,0.5700278238237031,0.661271003561529,
 0.5137187890436777,0.31080871174510216,0.9990440027868818,0.4090662956508939,
 0.30649506693143425,0.08404786809992293,0.7733242446395393,0.965456504624047,
 0.43909594676024066,0.40509181194112664,0.22017741788660594,
 0.5830710504406645,0.07522047268076737,0.7147472523605725,
 0.37271111437275795,0.2470652560163613,0.011516140981998269,
 0.6195528844905812,0.8209387707040534,0.547967907471048,0.7911864490145505,
 0.9572649468777142,0.10656864374673514,0.18682796935580626,
 0.8145606532293368,0.3458725393405724,0.3617102896829013,0.20786656764958444,
 0.7775468426014657,0.9322009843078521,0.3044380324691385,0.3142716832858321,
 0.07312894318835395,0.4237291695434332,0.8562324572698169,0.9098677339300743,
 0.9870036770793741,0.6462619601297456,0.647497824231458,0.28712613536171794,
 0.25212899700417724,0.1523458165432644,0.5090877517132167,0.9620805229779494,
 0.9092891001970802,0.10224143461553648,0.3849248031133269,
 0.029135860141308978,0.5827993798158619,0.10777482463187327,
 0.9951599170686878,0.6401314488892497,0.9878293284943711,0.9434990609620723,
 0.8821490792074265,0.6469119377978005,0.13453121846196736,0.2643720272288699,
 0.2926929426181264,0.4756417296762212,0.7730998338332267,0.30984331566412016,
 0.1001511290349697,0.40044253519767725,0.9383946994713825,0.6908051603624646,
 0.04597541542739192,0.7990332952603459,0.7008604521735962,0.9336477467545439,
 0.20238334556766024,0.537876483121487,0.22310721568859437,0.8506196876794234,
 0.46159358657294514,0.4351138318471949,0.05019929697765091,
 0.22274522191296398,0.3424135377777342,0.2865993099704378,0.8207627361468277,
 0.15590126849500174,0.8388113867468429,0.22586002313724773,
 0.40508865474945943,0.17365907095882505,0.43154543085237407,
 0.4205587160394759,0.7487711727208572,0.07736928487811301,
 0.37518827898185336,0.7917748336766429,0.04609293060116049,
 0.15903440684136227,0.26835222264132585,0.05265469655679689,
 0.39187018882997804,0.4793444525441565,0.016730708982318654,
 0.5016868775791992,0.6050517702430485,0.9960736153316717,0.4680811237118724,
 0.6440116854142075,0.7563815664375744,0.20828603050079686,
 0.32730075379655044,0.29163221384940996,0.034674947488001484,
 0.5436333113647724,0.946209355411015,0.253593983248952,0.6035001789420847,
 0.10222770839116624,0.5079435263537039,0.5957604375262949,0.9933689986735903,
 0.44030629132934296,0.6371118979551931,0.32791046925540224,
 0.3128078464580555,0.8411050540654349,0.19841692900068286,0.4105752020081308,
 0.015920686865010025,0.4069779519450253,0.5673153948439804,
 0.04494301502731113,0.5726276992616183,0.8682420235929255,0.9726611139194731,
 0.2034413480361814,0.3662624109347884,0.06987872316013177,0.9139499103067622,
 0.48580048401012477,0.8085237014251107,0.6705604246566275,0.9372236524123684,
 0.12202559149678094,0.8125936210061053,0.6749149769939273,0.8254315863897177,
 0.28107501370843024,0.5736324608460027,0.1924172708148194,0.5738611835456824,
 0.2007324440665047,0.9060641672481647,0.5999438828932456,0.9759775719090171,
 0.9687799963555059,0.725766958583219,0.10169076404792976,0.36451824976299363,
 0.6863127097208275,0.4349054731554307,0.5791694526313689,0.8516415150242136,
 0.8499350715075699,0.7213095246304329,0.199534440476964,0.8342505465465228,
 0.9266931028690722,0.17172768227385948,0.32543250001178137,
 0.9304206058759819,0.1518154124385659,0.765792281556524,0.28673610280797635,
 0.95962989364413,0.3126246193854302,0.8313283257833287,0.3970508720031263,
 0.6685861994640104,0.2449346589135626,0.1322159602073678,0.3702806044811263,
 0.7223040952295281,0.6160488989925602,0.00386183617769853,
 0.023348152423151025,0.2732738568083666,0.9792986063169131,
 0.47038505658774576,0.6013874348747177,0.45631134926502503,
 0.8189272210519625,0.4036115869433794,0.32909015601820435,0.3036199331968088,
 0.8858269555668458,0.22071956903529422,0.22315795900638746,
 0.9665319294007092,0.6698909461121222,0.46599590368330146,0.6091863573674988,
 0.07775800433795899,0.38462606281465606,0.5550481063277379,
 0.8429428227859042,0.16106578626548684,0.3735677818189068,0.6675828904270372,
 0.09223922723585753,0.7792528069954151,0.41563730683086986,
 0.22465361383213645,0.026185792391388185,0.25638112647049804,
 0.45697479246599493,0.6007844669985686,0.3852964771514682,0.479937486055268,
 0.5767651189514141,0.05201778831103243,0.043824419273298276,
 0.4278429042575176,0.0031813608128711257,0.3660615522888194,
 0.7750171094378755,0.35915937328448133,0.03287454007362778,
 0.8634341918962113,0.05661156417346913,0.2174785728299078,0.6939907121975109,
 0.9066562553526981,0.940236538550056,0.5797709184682422,0.7037811532599609,
 0.5036944934898785,0.6290520589299377,0.4220371439785964,0.5745396460001384,
 0.4324730699442505,0.6803205697475883,0.29812581399411464,0.7657271498239532,
 0.24508122317446157,0.7024036105681762,0.9098685000314651,
 0.037193102552992796,0.7902637279879517,0.3999252997016156,
 0.030678133224219084,0.4168966017238399,0.9565087265674033,
 0.6825610013446126,0.35172293751398176,0.8030452828332886,0.8659430438394361,
 0.07747902637117743,0.2856262117842465,0.8104542798906532,
 0.24500001849218844,0.6227262809310287,0.4655879076346774,0.1135120973255127,
 0.17470350176092175,0.32110080197672963,0.7277478064297521,
 0.8364356686679288,0.7301148360212781,0.6357200848761054,0.2507444412663391,
 0.2514364639676082,0.0862631550556191,0.817385524967108,0.255460098939588,
 0.24313277681931678,0.44242336658035175,0.6177165743604528,
 0.9886770443486487,0.23425996002440475,0.03363508754431144,
 0.4479132958110995,0.3643428399146056,0.41432433127293367,0.3110711122252233,
 0.7397684543112658,0.9827013914863623,0.31934420783594075,
 0.47771649909819447,0.11274432702537118,0.9464126901788043,
 0.05297356445295398,0.27358252876732536,0.8531865346048284,
 0.5542671957761596,0.13890155445467134,0.9648731393086005,
 0.24771307746135374,0.4053008725126943,0.9339678675059152,
 0.004622255549976084,0.7079433938000613,0.46205652208544,0.2580299851537834,
 0.4244425482730968,0.053973883565954894,0.499847514983661,0.687732457940239,
 0.3406123707225014,0.1801032992840802,0.2574455087959707,0.8532492525396521,
 0.6405057268772956,0.8842372212449583,0.8237945300023224,0.20218166431340734,
 0.8751420040097715,0.22216110933223732,0.4389159746411331,0.7003404324103049,
 0.5810257074904962,0.36412651708138144,0.5512675826415205,0.2892045981630791,
 0.07338794742395138,0.007637903336782825,0.9391873374746188,
 0.43068091561207844,0.02185823246153329,0.20859605229268996,
 0.25106781565273995,0.08383576285432304,0.9760151297889796,
 0.2776485042859016,0.8354663958518063,0.03022923522613552,
 0.0053441151074902216,0.27986842737055795,0.15592977247036066,
 0.11574785067999716,0.029981038106909552,0.6998331722223271,
 0.9801576150244611,0.27569906109978726,0.8260340441884003,
 0.31623378367864763,0.05569817468820981,0.8182709356264419,
 0.41459959486150455,0.3786668176755599,0.79459552435898,0.4864977719993008,
 0.796342609647027,0.8485083371400693,0.5706482313868049,0.5813133632542491,
 0.6971213886699476,0.8460714078424382,0.8315334586475263,0.4532095261746566,
 0.264687446350116,0.4825081511825887,0.7386648772876291,0.47656234458903646,
 0.7534173665080067,0.03658383298456269,0.3931780574667828,0.6020566045772079,
 0.6195134160770566,0.03159355494665195,0.2652553001235537,0.1570485099631762,
 0.5797625056101494,0.26489504791642415,0.5137228071982364,
 0.08120314766044423,0.022951971160454776,0.8863676393154771,
 0.8109072968176035,0.9095134666684108,0.8560120166035127,0.7763425713657054,
 0.26247962134830805,0.7442588544227013,0.8305977859184911,
 0.34801100433880683,0.9100247232281479,0.25426759248795394,
 0.5318595901332933,0.3098740433886581,0.07311901427583478,0.0096908845305691,
 0.7082817935188183,0.4053400873449967,0.9973080534918148,0.2607107890181076,
 0.6723118800373703,0.6426342657908415,0.5854139202989197,0.09525948019115626,
 0.14210228453314444,0.8811506797143347,0.0941807150579711,0.7757632191552385,
 0.2724060126122382,0.5499370153301517,0.516506430231961,0.6260683052126559,
 0.5297613554093722,0.9074277197401627,0.28853010325592454,0.8092003947265926,
 0.5226502829991819,0.028992179780137683,0.21654855783093963,
 0.5200252521672861,0.933548426551983,0.857027974894679,0.8956357913776755,
 0.7182492940597663,0.6530659688919478,0.18970023971088246,
 0.11443118301977329,0.673206122471657,0.30348218031073715,0.3619735906076673,
 0.5945159741202029,0.15427381824691055,0.5275851212319319,0.545108484176129,
 0.3532111408505648,0.015136492282169622,0.20662246284828834,
 0.32999667047961256,0.5433812100389295,0.9707900565622346,0.1437279690262212,
 0.8953754387601686,0.46265223279154877,0.02747559085063611,
 0.047972248767887415,0.8342443915951238,0.1194586492980918,
 0.05202032659978095,0.567968829209005,0.5844395681545501,0.21319368206132894,
 0.9247442645301629,0.6342511177455011,0.31831620132792626,0.4945106705919057,
 0.5113444191715066,0.8833162396277391,0.4326461179819856,0.5695401888862022,
 0.5283258917811847,0.1543865141943982,0.5028890258351266,0.9965447178096316,
 0.38188214213205973,0.038502858619032576,0.9160007206992146,
 0.14556044273341884,0.3218357217405424,0.331453987369162,0.748755862777187,
 0.45380660915477733,0.5966038051241783,0.7104655594696615,0.85774198739197,
 0.1130796094179013,0.5240101645242095,0.49930362837514297,
 0.03942359546049812,0.6037472137508725,0.4687424830293728,0.8539213191443527,
 0.10100288210192032,0.9732601445188105,0.20947186716636967,
 0.8443514260345331,0.843150201315192,0.9421224025392994,0.992575510944178,
 0.08119527231795443,0.4358694219718543,0.9105721621478053,0.6205716774679436,
 0.5945487975505199,0.38295143590228653,0.2769732328788732,
 0.08820054924718956,0.2785459863045686,0.7527316419242077,
 0.19773262405089975,0.05832543703187165,0.06947230210314048,
 0.32666557784854233,0.7856834264180114,0.6493573436979891,0.5475043634938923,
 0.40959967265640707,0.525760233089302,0.8033167902979995,0.598749330960542,
 0.3856141307466616,0.2790402939968748,0.9540918325193053,0.18574091357427036,
 0.2875331813572738,0.5122728594135255,0.8037009056720792,0.8884399659392516,
 0.8141594480515669,0.8388486855571171,0.5954351261905897,0.5665433753386817,
 0.47927740298118215,0.32602226365212394,0.15303346441945476,
 0.7113604095175432,0.4973429217217721,0.08719499493395688,0.2694674832133874,
 0.5807552326351967,0.2396812029761337,0.8202088733291785,0.5179025386620209,
 0.29151744925188094,0.08323455165420846,0.5679821376115671,
 0.7563822191328627,0.5462340736194021,0.6851727163605588,0.6378727663410134,
 0.4615462752592374,0.8347251394693713,0.231076730502062,0.987062887684276,
 0.2713121392006781,0.8881457028861304,0.618775325675053,0.8316500314523949,
 0.9554102595668273,0.5317180162622859,0.2804228610935532,0.706105263018076,
 0.43100826172603257,0.07239033403437312,0.9691454393095783,
 0.19423824654338162,0.7179491263966813,0.04480896619731145,
 0.20941021508009883,0.32839582083592045,0.2430303504688609,
 0.5572342181992957,0.34435423609682403,0.05597933201208882,
 0.7765755667016195,0.7931542673651762,0.4261255870189381,0.22345769674504934,
 0.13318473788014884,0.23524257794582804,0.12309108423608817,
 0.5220025165826384,0.8359827233601059,0.6554103949949136,0.5410437044686549,
 0.07831041390851512,0.0365240169395078,0.6797157457822096,
 0.018764232000923364,0.820381370883088,0.1250179811793739,0.3395651142270326,
 0.7973013186549294,0.07720622694633739,0.4611028960006889,0.8595030385326163,
 0.41313021186715115,0.6560322687474954,0.326896825962947,0.324605071860272,
 0.20348367669599254,0.08962486568297579,0.43266394303415856,
 0.9756425427594367,0.8573181732636931,0.18328841101535898,0.6501329585716354,
 0.1504049534560128,0.25979440529953624,0.7477733060090336,0.7026226241660177,
 0.9674801626555902,0.8790503840795072,0.7709528400246315,0.8703877598432931,
 0.2553101595994234,0.04141672704443411,0.763612131578531,0.25871298524543096,
 0.23830760663981632,0.3186480960565937,0.6623643180369181,0.6308224663345878,
 0.5513292157778829,0.9854946669672171,0.4588595632276339,0.9230985114500749,
 0.28473298977476225,0.6711090480200599,0.9139239881205159,
 0.30750299775631706,0.5930647220730259,0.3119150949623274,0.7479551628712914,
 0.12135302894404576,0.8427941477612796,0.1610274737898213,0.6418965546293228,
 0.8234533607045265,0.36867308847026425,0.589580757155592,0.7884899342109667,
 0.4194145926080318,0.0870157192557387,0.6563837772923176,0.237828685545582,
 0.2859842321056054,0.8828547385368621,0.7536355599943382,0.2357113211118821,
 0.8351151448763023,0.19521197017154068,0.6365778055684789,0.5642169189069912,
 0.6099087006463475,0.19485902320546644,0.5716449190243609,0.7516814466150248,
 0.8673043066149938,0.9776663337777699,0.0700978968749898,0.03360032200430341,
 0.9153366014847657,0.26622375208952564,0.012073533852756224,
 0.9156238020904719,0.18537898849056067,0.5378889559685656,0.3821191710214785,
 0.4068823223383351,0.49316665922758063,0.6110519393454701,
 0.013993764716708545,0.41371617990878407,0.21625440854333244,
 0.9352131098463238,0.6166702338466036,0.047503030183081885,
 0.17041816158241807,0.20117067091368512,0.5097958831566654,
 0.35924116537101825,0.07443316501443076,0.07584418461239228,
 0.7543110580933126,0.5629383773871012,0.8991548736221531,0.42212112213498976,
 0.46308915846807497,0.48058707870549067,0.2934505964624381,
 0.5211288153840213,0.24742699041466765,0.5740147309945571,0.7197916697144148,
 0.8899719657794938,0.07758051414918854,0.8058258830157302,0.2863247232329109,
 0.34887865036664545,0.9358243305396259,0.18439876820604506,
 0.8846814911889529,0.3590574081822837,0.6930028978619566,0.9492155123433551,
 0.014720257597440223,0.9114317501806155,0.906617129061247,0.7014659132574406,
 0.3207699037118712,0.44369477517040967,0.7535728016732685,0.2886991086341787,
 0.8932071465047221,0.7203806561767724,0.3003435871924849,0.646949923217221,
 0.37115618300044484,0.6449484936909573,0.029242685806616264,
 0.376411028344977,0.9223575464836469,0.6057892099160291,0.557488080287775,
 0.5776771176443767,0.7161829430067017,0.868144254713026,0.7237769478258259,
 0.69148346222833,0.26713449446025594,0.5330421638585792,0.44551773694882946,
 0.39550193547602464,0.08750161860628836,0.7935822256631525,
 0.7015315709795698,0.5192411495090496,0.021750251619241268,
 0.5187558539249812,0.2611382284649021,0.4676763710647486,0.9023877397965849,
 0.6493882375334282,0.4031068698355198,0.9309544629386606,0.23525498584884064,
 0.8786441901858922,0.21009153103246814,0.5723528248823905,
 0.031689510305081936,0.42487629627039436,0.6232166248946295,
 0.28057709988036494,0.612662715110601,0.4395390944150066,0.47879272250486093,
 0.7806494664547853,0.6441954982968953,0.29327416879785906,
 0.08888829703501833,0.5489946499388733,0.9877208625302655,0.3350339528429138,
 0.708797704971322,0.2649610943893438,0.9896830300984585,0.6513177333067675,
 0.4418190333707125,0.5505353989953073,0.9317786204399122,
 0.028333215971618975,0.275960882809974,0.23148774041974407,
 0.3121236930989022,0.38518640001808124,0.2550802515883721,0.8006108834047126,
 0.48405261498549534,0.4348497169482727,0.48326474221529125,
 0.15119188366314862,0.4071574810478923,0.07707403620781272,
 0.2876461986464306,0.7631620786119933,0.878745191929322,0.5366477602502886,
 0.7384984434506151,0.18772786181059298,0.77982111194944,0.8411932521700708,
 0.02282925819940973,0.5637130199642152,0.6612657119443478,0.8567731020040464,
 0.596453105183019,0.9377330945418236,0.7009767490024512,0.2486310911832803,
 0.3750762263273608,0.7833520007226298,0.5016405582298736,0.9406949848349868,
 0.9892030513944831,0.18059842834946938,0.8094053413581158,0.4997500078731746,
 0.5531362837192988,0.023241678245210617,0.6685595555912479,
 0.6682100680456217,0.698596597370933,0.024929251685047227,0.5138918217785758,
 0.06595174533423831,0.48071111119947185,0.6150794767492357,
 0.5781390877189873,0.6202986451129172,0.9123786337606549,0.11831162077348245,
 0.6606098648041481,0.38583115938464885,0.6521863439758903,0.9177914888218498,
 0.07944414901703956,0.4857760824516494,0.10449677786926992,
 0.11808957094683004,0.10961512034272425,0.04353255295837122,
 0.5398849030052296,0.6432160946639076,0.26827696749678087,0.6002919344807186,
 0.2257044658965266,0.7850182318558392,0.2749191112721051,0.6306061131575137,
 0.2069898940469509,0.3499817430948997,0.8009302302089202,0.24351470048006874,
 0.1671970643552595,0.1561479564430035,0.5154850593003931,0.4095027048391596,
 0.3357775712837885,0.2953465710377884,0.8767023324692949,0.6272834748310769,
 0.7943290672414197,0.04747291796118769,0.43383352644179474,
 0.24871101564562637,0.7372061789712973,0.2520823902428412,
 0.36478662409873264,0.575360618601052,0.10128547542305633,0.683484078619387,
 0.035735022020882856,0.46310809040915724,0.237519275649705,
 0.7531526403605394,0.026349724857018475,0.2722511390483451,
 0.5632962917725577,0.21439992048539813,0.517385691696019,0.3604843074570182,
 0.22364691401944647,0.053306342554160224,0.5784321878563354,
 0.18498689532562407,0.13626928029804142,0.24502817312620273,
 0.49994639191464463,0.34764404558441464,0.5113669921135566,
 0.8038358176642356,0.4041277762471466,0.8852578488282213,0.9804553787955872,
 0.03698741240070402,0.3916323285576634,0.7218569129297607,0.9186350947706562,
 0.1248874572164429,0.5870013389218918,0.3635586780224588,0.02566204837278907,
 0.059622743745718854,0.6999922216949557,0.7655143127855195,
 0.6752655746666867,0.8207622344696737,0.6101259645098296,0.2609314053273085,
 0.03573912358531861,0.5544836220325902,0.6914141719490932,0.8498331582038391,
 0.6632916739313487,0.5297837541985574,0.41733522614127583,0.9491822146729881,
 0.9048535408955554,0.06647231831323641,0.13705512411896237,
 0.29890935240381433,0.1699720264026775,0.7240173384062768,
 0.44986057627345466,0.17531537209279024,0.98268172791585,0.8582852235326974,
 0.10712421086349899,0.7333152334130182,0.38971635074773214,0.997348396099795,
 0.7785938640212222,0.7868364214125528,0.8269313868407561,0.5458013655633467,
 0.3950403867772633,0.15384351025045562,0.918686760346571,0.7200256848999622,
 0.37699921695488525,0.2940580221767607,0.8875800427987258,
 0.25787240651282506,0.4818743153581252,0.5687113981446703,0.9909569605199088,
 0.9471991673339698,0.7063165460439302,0.21045239048537334,0.2295153855143922,
 0.4772177023854409,0.34179236834392324,0.6009641258902211,0.5334385375367783,
 0.5188863508209864,0.5313292405057024,0.3962384222794413,0.22683677483881737,
 0.9654792552182687,0.18321963108537853,0.2423113025437279,
 0.48318878645530916,0.12029694429188864,0.749727884353277,0.2906435990022841,
 0.9062286017141812,0.11526437222372499,0.5899780163906148,
 0.037896470872598176,0.18844287233373924,0.5982932947922306,
 0.03220320767521068,0.09544082431405831,0.3920676761836849,
 0.9341823653149708,0.8043670049435057,0.7543173868366712,0.8825341737676466,
 0.007093702609038188,0.18086959667841063,0.9021917347371033,
 0.5094324815804784,0.4589187376256081,0.9783741697346902,0.5199051565445737,
 0.99370725385677,0.18016265933187725,0.3017365087154481,0.8145183222479333,
 0.4047292247422284,0.0792568485525682,0.47399619636109813,0.8106605748103508,
 0.8737844439039512,0.9572500791562849,0.07179278239082709,0.9159969761621919,
 0.4928494243631252,0.019094235497494227,0.39597875755157086,
 0.036134143777234784,0.2465223926129726,0.283007418144675,0.5602720864060262,
 0.12180418258442116,0.1402440756825981,0.6662138099947459,0.4930662350028676,
 0.7491101443195874,0.15302256564149264,0.3276827313381355,0.3570794917962069,
 0.3400397713244776,0.17540477541299648,0.02235673073940081,
 0.8913976258034547,0.9503769974937912,0.09830883208541841,0.0424668878701413,
 0.04757913262184965,0.9709330866379515,0.24814777533336918,0.584643857375114,
 0.003786575826143612,0.836991162965651,0.2484948855514073,
 0.13504434385316966,0.903219863335843,0.6626713182389998,0.45353693310267496,
 0.3045409662144958,0.10232436218915619,0.31526283760624496,
 0.9784860408627463,0.7761919497819366,0.7969961997911044,0.8158252011208312,
 0.21204609896260562,0.009560675256681916,0.7860732628924917,
 0.02589287018138675,0.06751029741232173,0.9639563011466956,
 0.8053655503142938,0.960260903915887,0.23016931853441513,0.24883988782577537,
 0.3744330187623095,0.8689242022595571,0.8071868734814706,0.9179342777311217,
 0.45797056042678974,0.29839801041318037,0.789627037781162,0.7784777959016708,
 0.09378544980758508,0.19213166851373142,0.4154716988077922,
 0.9577501182515707,0.4785552869864792,0.8647421440649471,0.6532144525889926,
 0.3047152928513378,0.21228631101246076,0.07192577900260955,
 0.020285694569461832,0.22221314398417513,0.5730774809756981,
 0.7705619556597645,0.5511429318749527,0.2837671061726911,0.3049272606776292,
 0.9605037771275462,0.26268942808910145,0.4809913186386624,
 0.20826857610453253,0.6122711542133978,0.2834871982795836,0.7549820789309601,
 0.6112957595866874,0.2994479746754417,0.06665040456546345,0.6470744167482521,
 0.7651293638504942,0.47965826142421575,0.11606216532605251,
 0.14395163594035654,0.6377368685154238,0.15444422719086026,
 0.9512749437228493,0.19921335588195754,0.47434863615907363,
 0.4185053752712793,0.5531386466358069,0.1290472329016592,0.22913307692870089,
 0.6544594562237412,0.6914759466249778,0.6522244395403206,0.3309477470279384,
 0.9048990353720863,0.04617566957967367,0.7706875869599076,
 0.49235586392344366,0.9981071672657097,0.5602643823862401,0.6107217782931955,
 0.38900624580808596,0.07682834962604779,0.9268983149735837,
 0.3361840451583096,0.1567791800321212,0.9408153880252614,0.7193416686433189,
 0.959754893285477,0.2703966797823155,0.28259054836045727,0.7931359172518144,
 0.9706078522068184,0.7715830556468466,0.36753590466684516,
 0.41054572424767477,0.7310968540543934,0.31996093699804007,
 0.7999473465038737,0.3925640067007561,0.8243506255375148,0.4978827745389388,
 0.7125454015677207,0.5296292878022755,0.4945100233323515,0.9540230257609893,
 0.9666474745419785,0.9236205060362188,0.7049205385288663,0.5131582587956702,
 0.15398862529698087,0.6342821626690605,0.509209861724815,0.7662484968663096,
 0.2067012015276305,0.7305935014015036,0.5760178828573361,
 0.037503893920764475,0.7684905201149501,0.6279027635057653,
 0.5865076610565987,0.011968241782479128,0.9129763255191234,
 0.1368698160847287,0.08914214476390137,0.54955917282947,0.8051747428744508,
 0.9173009542445546,0.0432839769511304,0.6204042463958186,0.8877353968817754,
 0.22604348574514288,0.11112237001874314,0.16436632214306024,
 0.39493802120492283,0.6004371504418418,0.07435353786082044,
 0.13753320206229647,0.3630304382567412,0.07063889620521302,
 0.11414408840997159,0.46864858143911714,0.0014779072054142972,
 0.3113174283674802,0.5303697209017465,0.8483223806837991,0.1754325259521906,
 0.418656641825896,0.9028178144296533,0.06404724837579279,0.7259995004398762,
 0.1338412743853521,0.3916943941897927,0.8587105586538661,
 0.035290904324668526,0.016730103719249545,0.44831388984970555,
 0.4777618281754449,0.45521979468428186,0.8097187249331596,0.2267244634762976,
 0.3228855850804234,0.3690941765087712,0.5558281865782716,0.8462204311216301,
 0.49769282163511674,0.4071080565924924,0.5432130354283656,0.566205764685893,
 0.7397045586197984,0.3083926652078207,0.5641437740784256,0.20768441425578854,
 0.5897550133466428,0.7057354884136647,0.20203912381888844,0.9064490197879624,
 0.1367223742622412,0.32497725038102065,0.09429784844846689,
 1.0853707368041476e-4,0.9326369923454794,0.7718141260889788,
 0.9960102416728611,0.709008248339109,0.1980191885776318,0.5181757335456454,
 0.6282625430578229,0.8683451885780475,0.6931198790829918,0.40435129515402624,
 0.15724343910576877,0.7633127510893768,0.18114320390604643,
 0.9550357852690365,0.8936627666584476,0.650152754522061,0.8262950699223072,
 0.40011447347126494,0.4020998852078368,0.5523388509774203,0.6134423207984401,
 0.21324806295056398,0.01812357942491316,0.37457190630705395,
 0.1545916108495765,0.45310289714081364,0.09499007313935015,
 0.4314498156807024,0.6770880462872241,0.05873157294887965,0.5634599880379992,
 0.9748293942680952,0.8318194802709122,0.7785233033081548,0.8204197989605553,
 0.4408418460797887,0.8376025820866644,0.5591069039462427,0.06433235927329195,
 0.650831060609247,0.15959585491648043,0.18244051914850035,
 0.48363473141471824,0.10061773200114643,0.8104955012043823,
 0.5830571071706969,0.7393275722716264,0.3765640606554287,0.9197834500270566,
 0.15663066451253138,0.6245544286933732,0.8624238997136879,
 0.36563400850326944,0.3840414629748695,0.5284813044412429,0.5461563261307888,
 0.9984671136413584,0.4657122232343115,0.618786382329942,0.652722959972147,
 0.9736510196638032,0.06094640941448337,0.987901588369376,0.31151825742043227,
 0.4685133524563192,0.9907726465874792,0.8320078545397533,0.7758811566600294,
 0.15265715739981345,0.10760775489967855,0.5144706892188352,0.998099692551704,
 0.4563360746178433,0.9878923796816221,0.639687871821595,0.26468296023355786,
 0.05366864250393155,0.10386005348046567,0.0275597246709004,
 0.9110212165399109,0.8186388389535465,0.1492969435412188,0.41463929567997015,
 0.17034389381423343,0.40670485893586594,0.4650025941393774,
 0.0038026377230904185,0.2542477714825683,0.6129932229371322,
 0.9416652814118084,0.4941222294382791,0.1736189456477497,0.2747777669265732,
 0.5877937863693903,0.8993988025143062,0.24984608036275502,0.8554065611018227,
 0.9739060382078765,0.8359715180004298,0.966687368109107,0.3334306821175469,
 0.19801811565279803,0.35265283141499537,0.26750545726506636,
 0.12276949389809744,0.9525889615764292,0.41424647263046843,
 0.023752583586581166,0.11574520911927477,0.8829665904819339,
 0.733045380299814,0.8667787262723102,0.9584287780141195,0.7144212119111656,
 0.8084180182899611,0.9600113113521774,0.9547048301795205,0.9964350636378948,
 0.9178584055220937,0.7737478004448216,0.1692078991739514,0.4554563559978526,
 0.6900272330672477,0.7665496413471874,0.646803538270372,0.47208976525292456,
 0.4253996651503613,0.8688580815129673,0.5510018602746873,0.7007546994032723,
 0.07632834667684252,0.1572822214141123,0.6537166299301211,0.7357827647225388,
 0.14865225856869124,0.6975745118240528,0.4189455986490034,0.6735808583387418,
 0.5731753552020806,0.7634926481395622,0.9154307848417002,0.4934069515807722,
 0.3511920631886909,0.6294582841361575,0.5804731298619786,0.0625197204161082,
 0.4517302875253816,0.9349735415336792,0.5881356969796508,
 0.039030726493280654,0.45965287176752223,0.30020359220192716,
 0.8406324612522534,0.595632945015883,0.029855356983691794,0.9351150194977773,
 0.5548012277697671,0.9542792128421081,0.8112167378725692,0.16727950155092708,
 0.49587700837877735,0.4971307081262739,0.14435783312271488,
 0.5726287132967707,0.18137493338178523,0.6492420022126647,0.5695907817578671,
 0.5128523977128353,0.9534262142236538,0.25889205698996964,0.6238932308145414,
 0.8066767772952298,0.8010352427535371,0.3202648789637057,0.14759361275862226,
 0.48341529623541324,0.3049163414096402,0.5599623319771301,
 0.27524445046333224,0.8087173155192807,0.15208063847267828,
 0.02467455608330804,0.37865382390592384,0.8317130716141714,0.730052672486081,
 0.3277932682685746,0.4967881285051088,0.8914973002531398,0.6797477431726613,
 0.017692620623193545,0.6190926912745771,0.1442844852963724,
 0.8584010467115126,0.017892975140745615,0.4051908842415023,
 0.1616338502093968,0.2031556826717661,0.5901747467506248,0.658796527095423,
 0.6954560482597982,0.8777891357400625,0.05365507623046417,
 0.18918343812790028,0.7532815071433744,0.7633963715976485,0.9017186705746284,
 0.720611598644767,0.8281290550893325,0.980431995075451,0.7905528866021692,
 0.684729188991283,0.4610117198859236,0.7965790833329218,0.03774945335343505,
 0.002356377634715745,0.2650371206871861,0.030512218633306,0.4568169806973551,
 0.05330674601887153,0.3225515096591812,0.4727817723713613,
 0.00391872580921282,0.8058639586731328,0.24607442697770887,
 0.37923071640281336,0.20482821600397627,0.023130360260429472,
 0.7406535086581911,0.950001708447653,0.7378004948854296,0.510939541235134,
 0.9907932211495661,0.8263216713955873,0.477956491772178,0.08313424309119388,
 0.056198667034702554,0.14940563448080124,0.8305238387007418,
 0.4802791419597696,0.685224574728348,0.733154692370581,0.25309941109198564,
 0.5352013666699477,0.010261583844358824,0.8001393648113941,
 0.8768681376966977,0.7068671188999015,0.027185436345403713,
 0.9577256699219346,0.8592983113806761,0.8564221215361109,0.3315942337329132,
 0.7891963203359293,0.6752832845128931,0.8897638141228857,0.01017902512608293,
 0.9665411140235236,0.027077527448376992,0.20796735295753788,
 0.26894904179373746,0.3482962379966459,0.6861527773862479,0.4208929184935609,
 0.3951871635180222,0.6288165155547734,0.06161749532355121,0.5607835618394978,
 0.9890471991815396,0.8474939550723156,0.26366680217371385,
 0.13522062412076674,0.6009659711097536,0.1680828495335247,0.7742874852010323,
 0.7553779921688646,0.08103600331105154,0.8172341983115367,
 0.08371833705618592,0.12412460255637303,0.02499239782592788,
 0.7346855284245646,0.5783467318240691,0.711974900592755,0.3762222281872911,
 0.24084571206675043,0.634630721347065,0.4947523157410676,0.6380164921643989,
 0.9638121852130058,0.6244181640085955,0.033250358125446855,
 0.7424858269893445,0.07144671518140822,0.45966026590615916,
 0.02796414345220366,0.9501892742609788,0.9157982925311997,0.6871344212070241,
 0.5086320246802315,0.48988532722920897,0.019267356975168903,
 0.565405910803589,0.8463995772086773,0.2741692560800737,0.9775007893088857,
 0.16225526740325746,0.49269221479633385,0.8567012764909865,0.985868865481399,
 0.4463292516491816,0.14931088442630824,0.7027829039408633,0.3528216362053096,
 0.6516817481634299,0.04310124848573549,0.853470301767743,0.8008302374461236,
 0.8762454531883372,0.5289662978093199,0.6013438966688497,0.11621680451073202,
 0.10622609122084048,0.7990351759441633,0.05178616721321527,
 0.9147683098046919,0.40236783917041685,0.9970321469987786,0.6472568199206026,
 0.4777777778819695,0.5568777411541189,0.5827311958271149,0.9046655388783407,
 0.2043411296668899,0.5923192275618402,0.7969535801458255,0.7014297651888824,
 0.19799907211056478,0.8576171027952316,0.02121785158713485,
 0.9721364985452462,0.4753442323139556,0.6337730438126211,0.8219347741866563,
 0.97144975537032,0.40896280861988665,0.6927293336163561,0.9360208008423001,
 0.84602451174633,0.7083061975659306,0.3515197282314575,0.4405829226546252,
 0.9658114537101915,0.14311308647098409,0.8146726865435756,0.3282504173292138,
 0.6862394697933135,0.5124930902011353,0.10342050572895767,
 0.22582890191104932,0.1320207329083677,0.3930164555943487,0.4130711843638323,
 0.7388453095771289,0.9323988507356118,0.15745148463938774,
 0.45348137108652153,0.9903057178762942,0.3493806697943489,0.6361512023590306,
 0.6766547397714038,0.1877226722212405,0.18591353065784855,0.8902487644727842,
 0.9428285027706729,0.8643607576734755,0.37483699033245543,0.7797796237486725,
 0.10029095862284454,0.3765753338707405,0.41300650505735537,
 0.3008794544625861,0.27882719375958986,0.13039071661785395,0.299078957209596,
 0.9402981697536061,0.8209668320736616,0.6051103992336461,0.5483137573732368,
 0.5915625923296933,0.2254279951261462,0.6761802556168868,0.1890297683202541,
 0.22796802923479664,0.0406850059945828,0.9082334104122596,
 0.11890455171888958,0.5284470300396741,0.6983599088326283,
 0.09657373678440095,0.061896986949878774,0.23446685621916707,
 0.8584266007950208,0.7209026637903053,0.6477359051196669,0.8762761488935139,
 0.04340917418242973,0.030354068423496083,0.31207366376231305,
 0.5211994812350795,0.7685754119182866,0.13374716309869017,0.0952850515871475,
 0.22928269520705624,0.691085376181331,0.8783141728691432,0.7695379542749999,
 0.3454319267793414,0.8239372945222648,0.9438619658706245,0.9249786152636775,
 0.27811861520010184,0.7067721961190948,0.2715526777418924,0.740680689420866,
 0.6649481449309114,0.3018703339276887,0.44622107683333634,0.8318508452555791,
 0.13185262503356643,0.6397018466044941,0.22308845336030303,
 0.25574382858144495,0.1233656167668481,0.36444547003290295,
 0.8090240807600975,0.13587346197159766,0.34973275036072127,
 0.7703133574251837,0.742500902940873,0.9855266230420338,0.7113521295651641,
 0.33685272252043863,0.6145968565343801,0.026366468834444823,
 0.8210908540824094,0.5976349306969122,0.007385896038707096,
 0.21461921981365562,0.3899510563788283,0.05201564176835838,
 0.015050353010251372,0.7997812624865807,0.8337012675807403,
 0.9595515028280099,0.27636603564234696,0.3044332182973465,
 0.27774044716297375,0.3631618157911345,0.08936939895195584,
 0.22187128272769385,0.5227762032999386,0.6283830069204704,0.889449440899477,
 0.48107997495754873,0.39037613796214976,0.7967644732279805,
 0.9497116847189846,0.6855549986745267,0.1460559027961863,0.823176529977488,
 0.23187038760195577,0.5053331580717457,0.4264876095951592,
 0.05766619016379737,0.5268954315028028,0.12372556962308578,
 0.34664732027230405,0.5107529724879876,0.1939043580239771,
 0.09695284651377267,0.34360883043807267,0.6016993916595336,
 0.4153866468067844,0.023729016652994206,0.8922153192359507,
 0.1460592233412491,0.703911100781149,0.08621114442772415,0.8575859653563312,
 0.9200556170896868,0.839736202885443,0.22803502194036562,0.29210364590093807,
 0.13388185225283888,0.6482714489291994,0.5671326130385042,0.8952353425521709,
 0.5342760478022209,0.3358105912676119,0.30465030227781154,0.7779352565978727,
 0.7377249829520997,0.37199189920802234,0.11961802859708937,
 0.04836126835505694,0.8468716098246205,0.11083791962885958,
 0.3954067726876669,0.2656375232343503,0.326049943131709,0.5878085596619393,
 0.04409953458250593,0.9049953027901486,0.4290071125462219,
 0.17367031839712233,0.39040075970022126,0.937233297313875,
 0.05758006718774655,0.4702238617205329,0.6227594546767681,
 0.048785583306362956,0.12018132276932414,0.8071579547511208,
 0.8776749541359538,0.53813968703879,0.6426218613945531,0.5620994773094375,
 0.4018378907988126,0.13424975381998883,0.9445874058866908,0.4534939571621569,
 0.9637327245986398,0.7195818092230704,0.7278192186471295,0.9420759957792486,
 0.5719645631351219,0.7500910930769561,0.4863031945124491,0.07713272568997842,
 0.8462304271810962,0.20915705537846674,0.30033657419699356,
 0.7501918535130032,0.8333567234908016,0.1474666909809409,0.8097323724619676,
 0.18398475877395648,0.4215497465118181,0.37355553519919027,
 0.3625561421347865,0.8046287824721322,0.7059892889655275,0.934739475295286,
 0.10450043201934744,0.7250910922825178,0.45585825550470704,
 0.3259897555302439,0.8662585444360673,0.3005161850236723,0.3936955873112129,
 0.9105650062303502,0.7882297602957271,0.7226999320960258,0.22420432664149803,
 0.030829424370898972,0.9295700684038495,0.8931605648783522,
 0.5272438473615658,0.9894538116528806,0.01752100750465202,0.9084042515930841,
 0.7669548305714236,0.6629809630269354,0.24173923242622597,0.264320848937566,
 0.09632102343733584,0.8966519243184919,0.06875511206213014,
 0.9448546739895866,0.30115309853302674,0.18197137501860716,
 0.09810860531422994,0.4467100947195024,0.9000289294963214,0.6351552036066385,
 0.7837677799065386,0.7435933016745014,0.8656136477924921,0.8579384537161562,
 0.9969589636327244,0.1820124662275293,0.03730321042107576,0.3563277047582958,
 0.8828494225214998,0.1147809293090084,0.3747841726098954,0.5593360113346832,
 0.1494980044784363,0.4324836719693339,0.3007563895141806,0.43522456695731093,
 0.18209021725691832,0.01782321765275907,0.8383854647138853,
 0.16040467547677173,0.5156249197472706,0.9389281686977685,0.6815973957987536,
 0.5691811131180915,0.32741651980857733,0.7960252562006557,0.2054748468563562,
 0.6815347407775959,0.14274728471115317,0.2167407670614172,0.5346229960523478,
 0.34245732529781003,0.325503794239113,0.4572174716993689,0.0778459097832237,
 0.12317890174704949,0.6380006762291057,0.161597282719248,0.7050152848779614,
 0.9793409205105089,0.45220694425447316,0.7276172320627948,
 0.27574647223395143,0.9646950329793346,0.2618933909730681,
 0.22617891182491212,0.44756499024524,0.3535908992213035,0.6466315719024013,
 0.6198711753812228,0.45287804464974746,0.3109243894172743,
 0.15044434976335597,0.31531469849946636,0.5165434104225426,
 0.10158044148139589,0.553049239688561,0.8071014723593104,0.9713211960986539,
 0.4785566377255993,0.21412847502829846,0.6941806328240241,0.7978523988003395,
 0.563448619399438,0.14832635395708837,0.4973481981384804,0.19472317118329951,
 0.1442033890559694,0.277059934098855,0.4077818234188558,0.5598092629300185,
 0.32897924738188244,0.49584159503064495,0.12328068922172153,
 0.9297166573224241,0.6998257878228227,0.7287441886999038,0.6674596690905948,
 0.9218939902227098,0.9156401485537141,4.110048083592055e-4,
 0.5732554031087235,0.7174659487176316,0.6901690008354567,0.9573916968454887,
 0.9533976314068362,0.6700377114253008,0.6855772306418018,0.9931937822292334,
 0.9541615845627037,0.15687074929230937,0.04509591143489966,
 0.3874888990162987,0.657904805854838,0.7463883632328318,0.982883708801928,
 0.15006488646921656,0.5734702351434138,0.6630942090072249,0.9022393860539618,
 0.5994126162792229,0.18835856974318332,0.12213748474365138,
 0.5446383004837545,0.7170786866710679,0.0780977440083035,0.4708013269368754,
 0.33215211580660564,0.17005596520861443,0.5192031836957884,
 0.8976645795274699,0.5088612119996093,0.37225629674954874,0.5939415744121715,
 0.21178697254610068,0.23291949233125298,0.10444586507845377,
 0.17057942334967025,0.32397159621861793,0.9990616192178864,
 0.2987805734489436,0.9381178031660949,0.7712607177694557,0.9429780895171957,
 0.11943180885401583,0.03091432231600133,0.123486744030737,0.3977651475743965,
 0.5622607184908955,0.9117581533232505,0.5014529175864996,0.9181457088807643,
 0.5811340300804225,0.8441999323680204,0.6179509715415613,0.9697968717513503,
 0.7022833207426272,0.18964425805974872,0.24618751717971432,
 0.9660965180789475,0.46164535266060536,0.9243427786512377,0.7560915089153868,
 0.670134402359751,0.7523381695037152,0.9992204427312394,0.36599073746340505,
 0.2951836914632381,0.8149908611076508,0.09130647670178138,0.249676425874085,
 0.5085069816659008,0.38585086977071403,0.13280373185245475,
 0.10070886578226979,0.7764992324122422,0.5014029798465208,0.5604350758351428,
 0.6213197919353003,0.1105042651252478,0.6397553866370719,0.3206366870429861,
 0.42315843514284457,0.35555013880398334,0.9415943041740064,
 0.3012229749910843,0.40247246580131946,0.023226712332668042,
 0.257652073692785,0.48571945422448937,0.3402422952416644,0.5048791004471117,
 0.5862095691751176,0.21501267910128097,0.7419879021086445,0.5431857051446529,
 0.7706653631316449,0.8398593450285146,0.663106315495396,0.015111164604305394,
 0.7631497861543171,0.9353108800669528,0.9859095324695102,0.1322179830814293,
 0.109919814110079,0.7726834804362694,0.2650291431078956,0.5549934817929861,
 0.03393468296448343,0.1225382734397078,0.9509193020711196,0.1582337439021435,
 0.28052531220498655,0.6191624488630443,0.6210161617732171,0.6408366714752848,
 0.03309383840224356,0.252150224700571,0.09841072572149445,0.7101384442373343,
 0.006500674935298711,0.7646218160886504,0.7976431602390831,
 0.03533185442038356,0.8284430868378043,0.36428991174330494,0.572533648814975,
 0.04557926862025741,0.663186961776687,0.7039052526640344,0.16933041312135655,
 0.17859475219323717,0.3580889905621505,0.559714266904665,0.724450627442698,
 0.32960789906060384,0.18895593859072024,0.7281805114587675,
 0.23069471375895167,0.7735402211363306,0.28170745685278364,
 0.42029058169920586,0.5116680810979066,0.466194764328578,0.1951495820128647,
 0.9947070522277595,0.7551152787663942,0.576191656893283,0.7967995138174259,
 0.3091830252977076,0.07634234109722315,0.06259265956409787,
 0.6448671444364421,0.5263335347919176,0.1777341523999113,0.895832702753324,
 0.24364906459017077,0.3909347164480388,0.24658104770437284,
 0.3736374979741701,0.3630185782040807,0.11396793097481661,0.5002561680429543,
 0.39618618607012623,0.7876630966666368,0.7797497873377841,0.2309492890564394,
 0.24507563114744335,0.5223290448042041,0.3898254917452849,0.8346948852650432,
 0.6357096687731608,0.8534726197046921,0.6671345244219086,0.8359805553809923,
 0.34194522157555574,0.19845840757286837,0.889257402871392,0.7826239910078104,
 0.2598869291590812,0.8856149021055135,0.025676271017881613,0.882777212722244,
 0.1837506997485585,0.3679064363129201,0.06330774481889911,
 0.052177297189727856,0.8337888853402806,0.4758246764301025,
 0.9969035099172958,0.02740935161748359,0.23532217489790097,
 0.0997549258551006,0.9487762766733134,0.14015307870666593,0.6954873302163497,
 0.37171719498235256,0.26665597386227413,0.025664859006844787,
 0.5097384079943749,0.7474342235133632,0.22543494636348205,0.7988340122117026,
 0.519685647743008,0.6429864446023065,0.9929696670390211,0.7611342742063335,
 0.8021432862485415,0.9360902010696237,0.6554566639434595,0.5606656675529592,
 0.22329751813029497,0.8206215519982474,0.8036522622723401,0.583526652664978,
 0.9097913879000609,0.8180324393644968,0.469100334969771,0.07042598581619797,
 0.6430078731385025,0.12445060936920227,0.02297533370082161,
 0.2615371853801336,0.5056565981840268,0.5462173377157183,0.8781774605187325,
 0.8319329494298973,0.8091224191879977,0.8921049418960869,0.3076515931820263,
 0.801087599603483,0.2811125957632995,0.16382608691830947,0.713562918346908,
 0.03495917254343062,0.5650724678852143,0.3930769776781301,
 0.10533480729687317,0.36846542861377685,0.6808768423253864,
 0.4556007180156726,0.03347933389410018,0.08396479767799792,
 0.6735824838213353,0.5151167261287615,0.01930947706192554,
 0.11724999558770932,0.018039234019785466,0.4036209276137127,
 0.49801156917102185,0.022809554082648176,0.9788123936409072,
 0.729247111317571,0.8102765565637065,0.1609654358244943,0.7684672181808156,
 0.5327244227636087,0.03355243870675284,0.10451806254509366,
 0.8070335411804432,0.08045982759033898,0.9741476473526784,
 0.08227933817948696,0.8289157703638833,0.5120381354940049,0.3561531532547183,
 0.10260101764105767,0.41012012415042787,0.06391082391826719,
 0.5417754034119111,0.11183619869234729,0.03017060685886619,
 0.7844258617426035,0.34397629508858296,0.18309308825280268,
 0.48875798484953636,0.28102152886010057,0.09887382881678108,
 0.08088349507893866,0.29415836205635526,0.8973501620379776,
 0.7998641783710543,0.6776768296475667,0.1498826522177259,0.863278225991755,
 0.45617329430113385,0.8735247690875938,0.3655218212014779,0.9401879611211789,
 0.22931384545380618,0.9119165052232581,0.6765001994582865,
 0.03508852474733626,0.3399367641945371,0.7051631366359614,
 0.13171936988415722,0.4694381371252363,0.9812519037318654,0.6737839071654628,
 0.06440042801510382,0.10568694324268146,0.21754515831081567,
 0.49438622568479984,0.8899528846930393,0.11084473082935897,
 0.26346923034325664,0.40439413318354134,0.2743428707966371,
 0.35802429220933085,0.19749922694390365,0.40209634748452916,
 0.09267695474394622,0.10348847261746996,0.569975579409556,0.8538270575753865,
 0.4789885181874055,0.8271510744838921,0.3415352844212701,0.15128841076699495,
 0.6459356186391396,0.6457658770917053,0.388999681025755,0.693349337493399,
 0.6902487401368327,0.4745144132883099,0.06127458062794722,
 0.26340100096721597,0.22803255512245924,0.4909245293865372,
 0.9358448328094165,0.2529388667464819,0.8858704371184143,0.46193970882897717,
 0.847812596221803,0.9793361107141508,0.45216317928474115,0.827806883625783,
 0.38074228871516813,0.5551199061367942,0.39034032265057694,
 0.8094911006682457,0.29142107240123816,0.9053497207384352,0.4399421138155508,
 0.022155503224016937,0.6654105092406617,0.19034834282551016,
 0.07326757283283913,0.5588513772500024,0.18623165049414558,
 0.29060172217966274,0.6602443182215271,0.6219618156928292,0.9614573853323631,
 0.9574514287920329,0.9684274827712837,0.8759028927834037,0.6410745977708175,
 0.48830036633175145,0.801676449491454,0.4159099007328413,0.06865327532007948,
 0.1771444427545198,0.5856214043478856,0.4216579574270559,0.7891441798248313,
 0.7498470188673237,0.6882273270951348,0.058499438720665875,
 0.3818093255016247,0.7569776487757405,0.3533968569832979,0.5280727227401623,
 0.3435056818659419,0.01582222929532562,0.13633034716109493,
 0.8660330355037473,0.07678110214577694,0.7162739023145958,0.9326267897886348,
 0.38482651308379745,0.4946411041090284,0.20753199350730456,
 0.18138985762266369,0.6807575476616531,0.8620675251226434,0.8701034734571323,
 0.9050916972241634,0.8163432197050078,0.9411542449667619,
 5.797266978160298e-4,0.4311667402432432,0.3854934922994405,
 0.33307962905634336,0.7793549459460243,0.2437430709843884,
 0.19820682876454365,0.3570153570046978,0.8964431054891576,0.1739395098513239,
 0.19792910677937792,0.19459687324868058,0.25371288510855283,
 0.43227469523604845,0.1509078368066159,0.17926910116979577,
 0.5878522328465227,0.4809112491104416,0.6524587100077672,0.9069426249528275,
 0.22755298594034234,0.1819753481645977,0.5593780426105375,0.3876795852839139,
 0.46707382411046205,0.5579653808631033,0.05139724832861925,0.616048873073733,
 0.2547291455743135,0.9812498716613496,0.08975529869024823,
 0.04404493615918115,0.934140457844146,0.29110106717732476,0.5057203910771825,
 0.4595264963150598,0.5187535294680036,0.670835377534028,0.018377330014513693,
 0.3977331151942768,0.852844357689937,0.9456678635638371,0.12374652181265244,
 0.1518053653778274,0.7745965196610503,0.12824670382476533,0.5096827238143891,
 0.05112670180181633,0.1287803351688872,0.6615470980653984,
 0.07161692323912916,0.7866733701853303,0.3253473863798696,0.3609872755555703,
 0.48337351295505415,0.011719004393418442,0.35032001204674534,
 0.2894606125698187,0.794913760813479,0.8942382997561175,0.6767743612462546,
 0.33375557499154795,0.8852025855663227,0.5452300900908846,
 0.09341016919057077,0.7587631687841113,0.8671211786111402,0.6801842283783678,
 0.5090990350051325,0.6332605151632994,0.862132155672829,0.6242082438799493,
 0.8026294541115565,0.6831398180089905,0.9868164986933248,0.5683903206438394,
 0.18603877259742196,0.42711593263098346,0.053864128965488756,
 0.2550858895331435,0.8937927895561958,0.8015986239767756,0.802793428091474,
 0.4984548341089565,0.9513994292665557,0.2844022096899832,0.21367486469201835,
 0.4877530364224907,0.6223307632874985,0.9575722047820613,0.22487884282052506,
 0.37452552917134196,0.22576614839986342,0.7025989886428002,
 0.8492314911299669,0.8333999455760506,0.905598665909048,0.6498574559766304,
 0.3364402354146222,0.707461092365197,0.08006005712018749,0.3604056194581461,
 0.45598586923098805,0.8242089932220023,0.8689691722064562,0.6915675180810575,
 0.8070526412028062,0.3474854699510228,0.7285111335847401,0.9419368736430218,
 0.9713544615524455,0.0905389428839668,0.40096568226433427,0.9891299365918484,
 0.15265177694293786,0.7471029039939685,0.14545993497002951,0.295660821749414,
 0.3957275838684474,0.21032541814700612,0.3903102344848143,0.4430345269207032,
 0.5617413645767786,0.27516341825338764,0.8061628405595127,0.8995961299525205,
 0.4606813633275725,0.9484360357269677,0.99820126504058,0.3422708136028454,
 0.04488353513767984,0.5916867747978409,0.5285285847801202,0.5232324706895893,
 0.637713720485427,0.12612138916309845,0.9319812776293891,0.8938371470765284,
 0.7537720797602208,0.1839423773195369,0.20600672424969324,0.4934482146705049,
 0.5111536166674201,0.10917526353698892,0.18872925528069318,
 0.35560507952507603,0.44139217344601434,0.524560298513944,0.4915780211671168,
 0.18884974203349636,0.7296558999765045,0.5656463268852819,0.4144333414627299,
 0.8520130246968483,0.7297491566684094,0.7616390302406731,0.3465067667311881,
 0.18793417202705442,0.11945522696467048,0.4456318208898413,
 0.5012646776481775,0.31946800185154023,0.875837487744152,0.41916411320557545,
 0.38396290020803103,0.6379794727738034,0.9051462769896054,0.5679811420770091,
 0.7541476101843989,0.32030026726767846,0.23542556808361603,
 0.8847792062757943,0.5500195238869946,0.3566701823199959,0.5839974376755694,
 0.13555537180611565,0.8380119493787881,0.545988475746323,0.9483647271417254,
 0.12908333265614236,0.6085535318434596,0.9130809795400139,
 0.13280308782868455,0.7338349482536759,0.4824300846576227,
 0.46962563748912733,0.44614575253468247,0.46312688914414013,
 0.22601462037904185,0.5840166938728943,0.4918401644291922,
 0.43246171018682344,0.2723202068125251,0.7519719292113399,
 0.20769207565054626,0.990623616910077,0.8574744570669536,0.9547739672564852,
 0.6918063983845915,0.8206800862270051,0.28659605082684103,0.4571904604054132,
 0.10480344548578746,0.9581232313526662,0.8814375247378525,0.7848658467815421,
 0.15850293266821147,0.2545488687514398,0.45162833296168226,
 0.09958412018477392,0.3265371472288854,0.435839522074393,0.6051561252579382,
 0.673271444595463,0.9620620393060659,0.8354459308246811,0.6010542132118202,
 0.38202423789430506,0.39501519035844157,0.7690188316811204,
 0.36452325131852104,0.8665154129532762,0.12038388386647425,0.744426066044187,
 0.5871781536249444,0.860246933445987,0.9740181786459395,0.6968392707011362,
 0.8067055885849174,0.3657912612051776,0.7139761392288606,0.9131425611592453,
 0.4596438658836294,0.29274847003144977,0.15828595358313902,
 0.6386305288091216,0.4957397167297196,0.018710058194298607,
 0.8674865631375483,0.5507133165071093,0.15180805985640633,
 0.42662331679656706,0.7768260148567744,0.3156555787030475,
 0.01663285716230778,0.33829961315423795,0.31602429431738177,
 0.2914550489637715,0.3275340162214653,0.5085402661518336,0.30089924088113384,
 0.009990943806381436,0.33158572128835406,0.687477131683853,
 0.5066684787690304,0.5922148212786896,0.08077966144849003,0.1703935151199074,
 0.7661219839022152,0.11579908639317882,0.4412584519788151,
 0.42205245817845194,0.6029780343903589,0.8647312657412702,0.8705155111071496,
 0.28336828529395197,0.9628976013014807,0.1693994243006507,
 0.17106525043259724,0.1475102479785244,0.26418162121073285,
 0.2718989268806674,0.04989034373407941,0.21610452550282078,
 0.6942634677162066,0.6756059299773732,0.25879230164067124,
 0.24701501551637495,0.586247206591558,0.8491456967803521,0.4375991604500744,
 0.4385935006074545,0.6466119336666145,0.36697703405008575,0.8689568129940722,
 0.4009795552479144,0.057577438874168996,0.5488602417186067,
 0.5561954168933205,0.4958047267135749,0.23490087634096424,0.6299193026160275,
 0.9470524119429125,0.039683649076417526,0.4561543559051886,
 0.8393958865144787,0.06026831354186568,0.12173671017839771,
 0.23914807098360624,0.6421526355273393,0.48882557676965577,
 0.052499332107721064,0.18707957264581,0.39390277880977276,0.9626508058730694,
 0.5357085704318374,0.6639120996630317,0.7128708178870902,0.0715368921951981,
 0.030785228736745274,0.11994106417058048,0.4894109289403097,
 0.6499681868022988,0.8615118822183176,0.10505845693037785,0.429669722942972,
 0.5700743713102799,0.5912626267648823,0.6622313988593761,0.04932230491007439,
 0.773189866087364,0.10376070506371149,0.002866286526684192,
 0.9815226100875014,0.7181904993084817,0.02199733171337359,0.8614525948450499,
 0.07076239244338978,0.3622826916402847,0.7611065234318335,
 0.17803005942955974,0.19818821751316018,0.34706672356668855,
 0.32079769084315224,0.06802812070405884,0.7039309860192867,
 0.7128959710481617,0.289731952348396,0.21656274310522639,0.664400250440923,
 0.45406220485729787,0.7363784518477821,0.1470585490750428,0.0158074392278158,
 0.9047197709368553,0.5753836676027453,0.09145599774771751,
 0.36416152670635793,0.5819978233699244,0.31181712376633786,
 0.08796686493853001,0.524645657810241,0.5066366292719472,
 0.007660417046021273,0.6139937147507175,0.41866473493951795,
 0.5493342679782379,0.5817131383686021,0.84195119217823,0.4629687160652921,
 0.6251028263693279,0.24632381010370463,0.7857422460109751,0.806240537662743,
 0.5358554798431862,0.6324676395077469,0.6085725039768881,0.992354247265697,
 0.6692249580075855,0.8333594736963004,0.19381172629800625,0.7377624915065037,
 0.19979241651299107,0.9448470873363156,0.3790138035499563,0.9850228389501209,
 0.4492685389708705,0.08446042328980352,0.08836480145324332,
 0.22911603339738773,0.7306859905521076,0.2889264678974305,0.4558771998738951,
 0.9956618595387473,0.1593810882286354,0.8043561716139715,0.6959256469347753,
 0.6902388072679201,0.30654340056131724,0.10004653081272619,
 0.4199116537632729,0.79137218310799,0.6126093476753048,0.11791003707168668,
 0.5791852304930767,0.8501674452782944,0.39695864193690045,0.8919144697231919,
 0.6952531843870152,0.7108446363703862,0.8805941609476097,0.7589925700037891,
 0.3872332595559782,0.810385464251837,0.8476257244270347,0.7429250371983324,
 0.37417103212284175,0.5247035262567646,0.3028000253179888,0.3575436088970827,
 0.3867780429826947,0.5074428780924481,0.45095110313223863,0.758085357609221,
 0.4345277815895545,0.3681694758230526,0.4873163899505265,0.4636751573362605,
 0.8063149069577026,0.8204893747223366,0.3904556174465651,0.7557060408421541,
 0.06969171416552877,0.14489666783014632,0.12379613642662324,
 0.45732648406364806,0.3136975119219061,0.21345971882881964,
 0.9589854817220888,0.14799384235521806,0.26167085570679205,
 0.4661983933672802,0.7585331287228828,0.5658900315859363,
 0.018896147517290074,0.11831202743611102,0.0708945399907015,
 0.05851744829732508,0.21298956069708375,0.5395020941281539,
 0.8346506015658409,0.8950585209965404,0.34708903048456585,
 0.21918282281768597,0.4304481727603793,0.6976986286229235,0.2950227376985165,
 0.11672831567702235,0.06882663693871427,0.13803895052279092,
 0.6774957607142327,0.08788217972385781,0.27595104459960895,
 0.7891362982863671,0.07858346837166474,0.3450188156022098,0.2114423048282038,
 0.11977204850522039,0.713893193760533,0.20708586054762101,0.7508084852161148,
 0.31477039205437474,0.8576151714965198,0.32009331747317415,
 0.24769420602022063,0.5273215251354261,0.29888197451833676,0.472755878231535,
 0.8310611267154571,0.7772034883178616,0.5058481052007218,0.3232816858158595,
 0.0240404843402654,0.1413052351825672,0.532817344160463,0.663407652251922,
 0.18642234774571154,0.5220784600445019,0.45064108158227867,
 0.6690293509058586,0.8001541618669817,0.5673482915191105,0.29910440980228636,
 0.8309007194722113,0.5799261126407376,0.501134512467829,0.6063994093539745,
 0.7579945351257711,0.2665443079273654,0.6411850223214007,0.7066434932026596,
 0.13983564895416067,0.6057116392414809,0.7633860154413529,0.7534028343943786,
 0.49737121583241417,0.4024889137577492,0.09476565190522646,
 0.3865147014913154,0.28674500042277784,0.7959547864590066,
 0.44115143582610905,0.1977725781727463,0.6180937164789836,0.7108327856826214,
 0.890115298302843,0.3211396631748843,0.12831805392437157,0.7548773295440494,
 0.5715100660122143,0.8265340330155078,0.1129788182485475,0.01798646105126822,
 0.42001526434323333,0.440106260221233,0.9202185772239075,0.5993717406460418,
 0.007549757714602423,0.8062553686118741,0.16558263216789504,
 0.3375296870853135,0.5404856003481721,0.38782204008525856,0.5993950594969528,
 0.2941142319118344,0.9695352715469426,0.5272662126902385,0.26225831555777945,
 0.803028347053504,0.3844397258168619,0.34310898337788553,0.16676238344221594,
 0.25060443229480833,0.9975707116581289,0.01557812173492057,
 0.8907323568364518,0.6292213033044499,0.8079540245550908,0.6860460804791488,
 0.49615020101405427,0.6280361021351301,0.718308334971591,0.38241075627396315,
 0.8189811915517786,0.4630718478873339,0.8828902417524316,0.6082128277037012,
 0.6582256919842565,0.8157615935918728,0.9079483316237452,0.10732984011676638,
 0.24741256020475588,0.14228942471310302,0.3783611466550494,
 0.2619690914776558,0.7870720967637075,0.019358371510119432,
 0.22456520491535847,0.5808601597280348,0.21877756806131188,
 0.8641515852234141,0.20345261178718932,0.6466030568491241,0.6487642156647373,
 0.4039551895904624,0.34383056752409935,0.5030296527164498,0.5364418063037693,
 0.4674281077819551,0.029075699337081984,0.5589350495060981,
 0.9896903510580367,0.3338516058888882,0.4554231587524775,0.3802881251732322,
 0.07649912213314236,0.06645678347984141,0.08099926493076892,
 0.8380331368652452,0.7865244039886081,0.0448578464521493,
 0.040357382501400196,0.39737691660772734,0.7078641658318552,
 0.09586781670999456,0.1614654319067872,0.12569783450202943,
 0.22467772594170243,0.42654275628567495,0.2775516835487275,
 0.1445091495490407,0.4459808688411385,0.829753170769516,0.6895755964174767,
 0.5826287542151903,0.2952857674398446,0.05669330740148082,
 0.43539914455683515,0.759046420733876,0.6825882174446706,0.4190807197229576,
 0.2181861113352549,0.19861123865491292,0.5537441628209423,0.2030651776926451,
 0.9134631026840431,0.7718281449879822,0.8229200576925397,0.3945537473727737,
 0.7189060067996367,0.24376531766256593,0.9745365267901425,0.5555967513732977,
 0.04387469344366124,0.927595149133007,0.1035380897565581,0.929827433051399,
 0.5312328746664557,0.33921875370935783,0.25066833055395543,
 0.10707602150586315,0.37796261528521335,0.37238283840730835,
 0.10717170606383142,0.2430932793877949,0.20362271525747788,
 0.5188798262512184,0.9094836205567418,0.9243717561924,0.2198714208162652,
 0.7945056538717763,0.9162508202748567,0.6271116876161862,0.1498255538687665,
 0.9508349002588755,0.19907576880295053,0.585617720546446,0.5414474619532911,
 0.2689656262591519,0.9445150271825801,0.38554309311513046,
 0.27640485188659636,0.12237072725585729,0.5657785924545626,
 0.048264651537437975,0.11193094838343265,0.6553802280629573,
 0.6343992841752578,0.019925866895354583,0.14161325337267583,
 0.15790988435912046,0.4266211172765837,0.1879400998669546,0.3092105769374256,
 0.2419004046648009,0.1215943574011189,0.7233857511618262,0.36269643531591966,
 0.5610864953399398,0.17076437760812335,0.8459239178273976,0.7677904944570118,
 0.2962277905339068,0.7505571170036465,0.8844951469997978,0.1852258115925074,
 0.18407305406461538,0.22560728891990733,0.1505743887835288,
 0.30884843463343936,0.9435421283620369,0.005229955232027628,
 0.12919774031612008,0.16477118823707626,0.7428013141163172,0.984088121058016,
 0.23062908136626448,0.6212109217536337,0.05217636084670274,
 0.018361758686516683,0.21967269887557717,0.8925995144786825,
 0.6142509889278398,0.9171409925493244,0.5365723271572429,0.8879359152813722,
 0.12565786759532604,0.25678550784847975,0.9674868840406537,
 0.045070311550998166,0.01855164241560492,0.37732218484077495,
 0.005000220353789109,0.24191112414989668,0.47869544602742686,
 0.7829787599546864,0.4720989063921094,0.9436862080261994,0.7484349622253532,
 0.04078938861021053,0.1988034316731505,0.45206509316290955,
 0.5439267205158218,0.9624481445203747,0.8514669935081787,0.5576235306083952,
 0.9496937902810023,0.18561483912488086,0.4027407932247866,0.5586298796112485,
 0.6448961659348077,0.6003296514855884,0.53079493584217,0.07541345905480235,
 0.35990064986469283,0.3683460486209913,0.69975863745479,0.0860685796784103,
 0.4158678165808145,0.09673813928127117,0.2662133004299031,
 0.12641333353908402,0.6775732754362489,0.7899083517267959,
 0.019434268639246355,0.07006992687083713,0.9678249845995645,
 0.20932884955937958,0.9383509863602353,0.8250577162639745,
 0.31007960666931234,0.703123661475354,0.2668314853442689,0.10200438736970652,
 0.9510988043699309,0.11597595924185944,0.3810927800600189,
 0.0043772122075677045,0.9191472338623253,0.05584140025605955,
 0.23028269574208304,0.0802846682786712,0.5900148560585424,0.9499398933871089,
 0.9322859794671163,0.15238870614346756,0.7314057926130045,0.9627520977437722,
 0.17024893193776647,0.567535700470483,0.3524396669183668,0.49449956525302496,
 0.7940452709732925,0.693556047555786,0.2648148609530542,0.9125128551698435,
 0.13376332217672504,0.15792111173541112,0.495688237778115,
 0.22183258387446936,0.9196351446868566,0.9275214954924009,0.8715988272467907,
 0.8094284709878978,0.8548899834709843,0.8639538974388025,0.8743536198639948,
 0.8025383587565659,0.6531959807059983,0.9086660656008431,0.1400000509708056,
 0.8285310641645085,0.5130214053229989,0.47699504636500656,0.8847908207090798,
 0.39379166640887175,0.9966206846903378,0.9334593109135549,0.5923452919537481,
 0.5544259088616518,0.14583979259165736,0.7670798844519789,
 0.04055833227944028,0.12441482005012072,0.3275554867876238,
 0.15825769292386305,0.5873082004698962,0.06664423364306571,
 0.4667253086416887,0.576482857630236,0.7797352305452135,0.21751239655106192,
 0.7816662652698553,0.9074424866883297,0.35751373536332254,
 0.024528057482733567,0.862117854260522,0.3657233158125982,
 0.13698081570396914,0.7201367155129629,0.4842375786285674,0.5977669694376025,
 0.251089105962631,0.7585787112637535,0.03921765258688947,0.6075369813508459,
 0.5943511835627193,0.23708429628337135,0.03847793367977581,
 0.1976496473216648,0.8940620479131778,0.5463236791982505,0.5422992762180575,
 0.7399363486902313,0.6594242196821585,0.5905862584536694,0.25232473970381086,
 0.7079978065359104,0.5581548690001737,0.31007479245744074,0.901477889843436,
 0.41256825094747507,0.82868242071127,0.16112895206844624,0.0233773325533686,
 0.595055382427562,0.24401505783721944,0.3822613699166463,0.4107278521183625,
 0.7170036950888392,0.9123224628044878,0.8668656837203829,0.622568795324799,
 0.2950878847021541,0.07569374929240691,0.7633760168105225,
 0.47695715895868673,0.8646264724859712,0.1933763690565642,
 0.18260233827219274,0.4173877267328052,0.22089083277977206,
 0.35947459534146675,0.5085570477070506,0.448901493179629,0.41083546147917227,
 0.7816814683062322,0.032115954658957335,0.7552817225608143,0.664225233105457,
 0.33409746139944363,0.44517193769945973,0.5771303612139514,
 0.6107315290519691,0.5522162283486012,0.4883274577694312,0.2781237640533465,
 0.8753420190516699,0.7046029366899094,0.6329908381742885,0.02007282110917108,
 0.9132589522262683,0.4546449833708276,0.347258685004321,0.17445181191633063,
 0.07101260232045059,0.7815143983201929,0.514338594855352,0.7128509348275063,
 0.04957356256598022,0.4440088007239703,0.08687827166487927,
 0.8884874684276006,0.9432177133724293,0.8878367094942844,0.7552889538371063,
 0.18172216458090573,0.9927532661286352,0.47012784821465814,0.898437057266239,
 0.798023563358969,0.8902013457518997,0.4743692403904586,0.1963651087796079,
 0.0784520145630917,0.7360548414980903,0.08648974678977617,0.1514297667173492,
 0.5365289159767942,0.855431249259206,0.024131727073474107,0.6117756121877235,
 0.2005073493489602,0.5391795826695637,0.5588232467998284,0.9020959480371952,
 0.9087299899587772,0.28812539492990075,0.7923981588232234,
 0.35567717763451134,0.4475128840478182,0.06446179484108416,
 0.5429886705530427,0.3280114531855223,0.6366826437579814,0.9517495560719161,
 0.520223303483073,0.6989296362295171,0.742921595086763,0.8408892606231645,
 0.19144720230195889,0.9387860518464797,0.3926963540415451,0.5016264444456603,
 0.22469012181238512,0.12354620642245318,0.04515633894788218,
 0.32583608642288253,0.9848532657738904,0.0021674840167058562,
 0.6344223117536829,0.7303581919946753,0.07975889316467466,
 0.009966182843296756,0.5742859621872044,0.23646277525796278,
 0.5457165981741079,0.5259020528205745,0.21769754536053443,
 0.35327152174376875,0.3852331586253618,0.5601862016633986,0.6269359950354838,
 0.7828731275781312,0.909509164425853,0.9592830129067723,0.049362078374924234,
 0.5030182752329746,0.5562374731636018,0.19051954009471195,
 0.015121025326173987,0.3062249744790637,0.19164646071313118,
 0.35608208127166086,0.6071011581742566,0.990844095529268,
 0.008358969074062972,0.7183671438328124,0.5170111346810169,
 0.36654983673410413,0.5075864371010602,0.732689189041537,0.9305015459726956,
 0.7442886455720493,0.24350153967245525,0.15769738478961481,
 0.24062754286005816,0.21132828637639112,0.010745836064225078,
 0.663785771967029,0.062181006956387286,0.756924443343978,0.439664743997611,
 0.5210206287696146,0.9178571999914268,0.9734578050598568,0.5325226647496547,
 0.28553273256210165,0.6771308159850404,0.7203194439038852,
 0.016738730279706182,0.34979871977431,0.02036745582273558,
 0.31186006730264393,0.0015634220058142745,0.7168027757380647,
 0.5751806055353889,0.25546690376258696,0.22882092969784307,0.350113086201414,
 0.958275818477029,0.053935263818785995,0.9291496604376872,0.8980675266238739,
 0.9686263958991732,0.4928342168700045,0.9372471806485704,0.39681516545458595,
 0.8375007183219144,0.03566888832418835,0.8731244326730635,0.6441622051030291,
 0.7296577460467758,0.9285118682967076,0.24008989974564532,0.7577099864436474,
 0.4091449942511277,0.8493908040429914,0.4218187373515069,0.8401927850939784,
 0.023814028261011644,0.1549901159000754,0.7947636941418426,
 0.41347796086028765,0.8103005590365665,0.6049639522574293,
 0.04257974785719343,0.2200879049027692,0.1813845738933867,0.3297313724888855,
 0.5579890111580036,0.23442186736028825,0.11373468736497416,
 0.13840246087814112,0.8495463291797135,0.42469508355155927,
 0.7961171629139637,0.23061687214786386,0.24926297713567125,
 0.9039886950142011,0.6085425473836962,0.6585222553751621,0.9178287989162766,
 0.6050718018997598,0.8987806880865383,0.25262098582487336,0.8272703715605654,
 0.32547921285355685,0.3101425721414468,0.029347398349690135,
 0.3429079967703387,0.7958840730031299,0.021035563609669294,
 0.4063509062242273,0.7686671625405168,0.5246230746033266,
 0.0024878283696389003,0.5421827577059052,0.43580326106946543,
 0.2897300598666698,0.06675511361991326,0.4124591624722166,0.9753843821713284,
 0.19814674107100017,0.13757858619798058,0.3381796535809989,
 0.6455540586601386,0.0641070572748309,0.4675228411934723,0.7605576673339306,
 0.6015251173005004,0.11364583215575708,0.4260646617863,0.308129187736244,
 0.1541974955087846,0.7763322766449893,0.7240847006423273,0.9106640286030958,
 0.5537156804994244,0.8445686319823296,0.6987247393343519,0.961169279295815,
 0.9540877816691548,0.13301454222359255,0.1045478312686714,
 0.46109509165926077,0.6747812387848212,0.451094156399512,0.7263742209033445,
 0.14007537832585015,0.9358668516217006,0.5354266693611457,0.7954728542195714,
 0.1489033088071875,0.4585973327376043,0.6985516856834851,0.6452519660670033,
 0.18413272844100192,0.5074942919721763,0.3964414448927637,0.8872247534209209,
 0.2616885733559212,0.8051287030645528,0.7951171603677938,0.9502513122159639,
 0.4921713106109362,0.5910550290833958,0.7693849105431606,0.5620600936814015,
 0.8386053038338485,0.7510368227287931,0.2912511575320047,0.792182519683176,
 0.5654850190178559,0.5499986130625664,0.11534266734480003,0.794759409047575,
 0.4017900349898892,0.13637225159580257,0.06060823204749832,
 0.9770829608037126,0.05189507962462536,0.10817121506074057,
 0.6545588175043878,0.6751270556265643,0.3541537320912721,0.9398181050737163,
 0.7787513334271372,0.7042432739271357,0.371953641379583,0.30710046524127843,
 0.21875744330908709,0.6955928455194025,0.1897339685410847,0.4128231422683645,
 0.12061608117189937,0.3057341273170311,0.8435221585774085,0.8271281851656022,
 0.5966310174896365,0.5226844182664854,0.42482579450805724,0.6857060988165598,
 0.23262975246088513,0.23904591034000233,0.7144610246554287,
 0.0189481445477353,0.017591266541966966,0.22407370802875204,
 0.579784988168722,0.0396139588032447,0.4902223482276965,0.35079215328851954,
 0.9493708702959521,0.2907286825755254,0.6861072134152798,0.6059794373072795,
 0.2968060069165632,0.6852402046203089,0.35997560572290044,0.4635736555702952,
 0.090635350974685,0.632486150487924,0.9072330522036881,0.2134179960829593,
 0.4774272459403014,0.7474626849407073,0.7654340195642998,0.41447185074302717,
 0.20619042684946298,0.33257442934884407,0.8487291298238268,
 0.8831302124050341,0.02813180564740181,0.9347370008584545,0.1789607634268786,
 0.8018126213036475,0.1971589694123247,0.3870219872081361,0.4475865211530292,
 0.32473512192287646,0.16579298505796425,0.8333869086078738,
 0.3291715300361715,0.900731164975933,0.9772156012409494,0.18967725019839654,
 0.7177723222417665,0.49722688826717265,0.03202380378117442,
 0.8351412312692351,0.2137639275296237,0.9486321122641153,0.5532203212061918,
 0.47630861958484394,0.09539009438922652,0.23822901303403,0.3650044992091803,
 0.8062335159349052,0.5460030291207785,0.11861548216576834,
 0.12580174330378258,0.4802567920179671,0.9509731302916582,
 0.25914096425295297,0.2222024519251209,0.624645108299286,0.7970851706039045,
 0.5235383064736284,0.7870043052721306,0.7645803872056857,0.8274830313212167,
 0.46536268790946134,0.16382860391125453,0.5024847038742983,
 0.2661644022847872,0.9543190386465958,0.13550443667026635,0.1680172766599778,
 0.3109834600001311,0.8020147324334307,0.33047570017517125,0.6107110234519173,
 0.0045471667113682335,0.39550029586770663,0.16457617721930573,
 0.524884677919557,0.38674467252185785,0.7406378563203744,
 0.020141311606956536,0.8325010879806432,0.7822279686609216,0.730764900657648,
 0.06579338615198493,0.6375957175921296,0.20349305916896165,
 0.30275543402583727,0.26546397561749235,0.6976853399760764,
 0.28479478572690364,0.2803918393372724,0.6450942607123744,0.8373855201317328,
 0.569403374504918,0.8896310634847668,0.008589578961050837,0.9553376261314013,
 0.14115054162883445,0.500988549934582,0.14373495164579075,0.7852571612287975,
 0.6069630154533254,0.036329203114668585,0.3942152664800742,
 0.7003607194042761,0.4046956287963491,0.28910996167737646,0.4584507483001876,
 0.9427660686735686,0.8305448076920363,0.5873340573913822,0.20970943097875472,
 0.30958498055981565,0.391000243204658,0.23424273328474,0.7792936407306219,
 0.1058182806752731,0.3731573444890852,0.6867571754724934,0.10295080229108877,
 0.37782645113531554,0.5515860415277943,0.34946044490496864,
 0.2098863207701398,0.5658437819741385,0.47986097904150693,
 0.017970953058814132,0.6103480327011077,0.5388466816241567,0.21898575857614,
 0.840228805958408,0.7990942469453461,0.6255305237941042,0.048976162130356715,
 0.3319650126931837,0.4836918901761176,0.5414827528170469,0.5159875243845382,
 0.8552883907884884,0.34924817389614504,0.6311424226812089,0.2595798596166101,
 0.5028578844645851,0.37791281450493286,0.893047488038467,0.6442844789940051,
 0.7006094087190441,0.10302534287606713,0.8272509306443057,
 0.34591284754961193,0.852292855346177,5.01949348812758e-4,
 0.30882173216995956,0.9009351584110129,0.36806258680939374,
 0.8727223691195339,0.8480485408692373,0.36987072793877585,0.9873305913554805,
 0.3165656736233017,0.5756828937947945,0.32117373581160713,0.9107413945101428,
 6.546650649308372e-4,0.5243320207873926,0.7248264472951773,
 0.20953419438125798,0.027692218884649122,0.6409172882574317,
 0.8264704577399249,0.48177655885686244,0.37344383907516987,
 0.021310276215869672,0.026185506165007477,0.6135799023672758,
 0.5138944624773745,0.5568381225570858,0.6123802767526794,0.31059390205731663,
 0.13733247771099144,0.5879463623838863,0.719678098498434,0.5960840703976779,
 0.5963886631542521,0.9325896895498956,0.8447859362885649,0.630413673843291,
 0.1303784493182365,0.626409431466898,0.5007345560355065,0.24797376533977,
 0.21427987739678478,0.90406334469346,0.9436141311255085,0.4585768100384714,
 0.7354991795714197,0.45851776678755185,0.3697511886855015,0.6205931721027989,
 0.5753618383974214,0.7650803083141626,0.8111226483872402,0.34228883503732577,
 0.8305754417518952,0.6628519308028628,0.720242838751196,0.13491473856196423,
 0.39983912410892297,0.3929009525531093,0.02635638437028276,
 0.2388834807396767,0.15012204121792916,0.9418921904711002,0.3251888199772619,
 0.23555263413544125,0.4539581202775964,0.30272272775563447,0.643534860698233,
 0.8637389851388222,0.732537023815811,0.958336090777533,0.37869805678458135,
 0.40530456096150735,0.7445433948443141,0.7714868943164742,0.5416120212344023,
 0.9386979591586606,0.9114759065507577,0.37574740547488217,0.323170177455516,
 0.5607296282038725,0.08665069392019076,0.6105488866582718,
 0.08445588495629708,0.8960552201346621,0.90393045148438,0.41359255336040257,
 0.4519268542324726,0.7092101580450718,0.05418485165861797,0.7116450586533669,
 0.9293496174746493,0.4512237050683945,0.7355277725377716,0.5325841806480749,
 0.01026278980638995,0.026343510172259,0.7178310429573173,0.3568647745773159,
 0.6266381298599549,0.7015853928239567,0.8692728864120014,0.17590436192597725,
 0.8359091814186278,0.34393159933385886,0.03639031862795328,
 0.8406141650784532,0.0583804992873006,0.9841988434078821,0.6058727969077515,
 0.3060685381068575,0.9982790801510898,0.07611833126902212,
 0.26621844350944035,0.8043271943885024,0.4458812598749229,0.1301306040128941,
 0.7070851892476994,0.7415668825023951,0.9684724314065329,0.4166989192081474,
 0.7715555914779345,0.1504578686669069,0.822082935222063,0.6961938952355056,
 0.363651441281698,0.5315398133436623,0.13471758179211568,0.629943473685766,
 0.6157234979763896,0.5295713172788639,0.21266730606284656,
 0.23209799047088187,0.3418911328254417,0.8101011851872577,0.7419331420045251,
 0.49542553642979736,0.6942276492987638,0.9941126090433043,0.8779352314004033,
 0.6489785889646847,0.6375229983803926,0.6148787518187528,0.601741673447604,
 0.43841965449370246,0.8506155545693184,0.25062871382661045,
 0.6019799487421789,0.7868306601243497,0.02579434157236371,
 0.06236897946546649,0.5968104454917875,0.2994936316551752,0.7264209348509749,
 0.012038848065148988,0.01578928294966564,0.2346695018271271,
 0.2348543124635304,0.1419700698028672,0.0027164984278227333,
 0.8935895746866022,0.018465498618795873,0.02728175391663623,
 0.15761220844235213,0.06389480225497157,0.571801700032896,0.5055978157166905,
 0.9107922297587412,0.17631033868422996,0.06701263361638554,
 0.9159092334610175,0.8977345801700399,0.4125468916243159,0.7345434708814149,
 0.9021265017710058,0.10024045229422829,0.12922456459925513,0.84628600499668,
 0.6193585840129807,0.5049703652812969,2.1810009253719898e-4,
 0.8509600088470222,0.8568606624073406,0.9229832823564337,0.43509707224572725,
 0.8830285195476355,0.38318507099296395,0.37376231618350886,
 0.22467124746253986,0.4115496216070055,0.4473577549566863,0.5369149201247454,
 0.14082729283054785,0.9621359756794106,0.7916308352988728,
 0.060228417978429816,0.36538520302283317,0.017064202511194337,
 0.13267428148848792,0.8260998839441929,0.523145809989313,0.13436124277389616,
 0.5703950659421058,0.7987114130854847,0.008472093074097464,
 0.13640894187207409,0.2989200109965039,0.45524522443554416,
 0.5981229882009146,0.9165321681893531,0.36822321955518733,0.9257520770766401,
 0.0615648676953372,0.8132078165621541,0.021564900002806087,
 0.9124790468086523,0.5865633020001828,0.9626170481135993,0.3859968019717983,
 0.8727734889401061,0.8717443484216814,0.392183076386438,0.5281884470629179,
 0.8159129903082956,0.04401285579827796,0.052021901481739685,
 0.5499407194908675,0.32613832511841767,0.8788901159355896,0.8952417282177636,
 0.12618398528900965,0.20764248211637126,0.3624044768081667,
 0.5410515006334485,0.25314674273538407,0.8805845629271359,0.7900103811053123,
 0.16314945351790533,0.07692168683682821,0.2270046968700994,
 0.7710203169237256,0.40034660305153047,0.8184601669072139,0.5934591940385873,
 0.8914980182147574,0.47723386296893267,0.6637144840739547,0.5489632184960953,
 0.7442170590907271,0.33437912394022407,0.8810438785989375,0.4626762087182533,
 0.4874104911412763,0.01872475012620234,0.428932376473366,0.7686267627337868,
 0.35428645688812344,0.1469904377585426,0.19988150073709465,0.779670959613404,
 0.06322002202560983,0.5311641899291326,0.835393181281539,0.6126197775402316,
 0.001090026031819069,0.41745576938703677,0.42710246156821496,
 0.34272389085799415,0.7652587777535188,0.6938203172654925,0.990060158137742,
 0.07659628291836917,0.0710539319305229,0.8722152903840353,0.9414948554067317,
 0.7331606655755463,0.27669235545597015,0.8446485514779447,
 0.46804247522096065,0.28313763629726296,0.4930397775208213,
 0.6474552498650753,0.4169525535996067,0.7837274171614373,0.39121808905255717,
 0.5339596689933964,0.5149788756319371,0.46898324260070934,0.4846729289560403,
 0.02492511335444314,0.8632247000198389,0.25435931990627325,
 0.16324922811391074,0.47660012409864405,0.5298872904399314,
 0.9266246362079642,0.8478490571184282,0.8412263606650867,0.26014117050609087,
 0.45623311826166735,0.36917364451665313,0.6168666821695525,
 0.1746906145256769,0.8461977018591668,0.9068209629836947,0.16007214027796657,
 0.35439921473356106,0.054510457621671726,0.8004112539131092,
 0.20862965748200957,0.9798449222022079,0.5656587164134412,0.4283519030096785,
 0.8695848751193442,0.2353130566818613,0.3479362217627261,0.6801612083593517,
 0.7234426727162472,0.8835181260612921,0.9241281848246816,0.45620094305027536,
 0.17548666019821257,0.03853062017809883,0.2656956567813943,
 0.7816338838888729,0.4569099702795789,0.8960571205170097,0.928824521304559,
 0.9868377737913876,0.6549803112502796,0.32042598446791903,
 0.040599276366977244,0.06454760622264721,0.6309298123838842,
 0.5973273635327603,0.3591427133252536,0.42360018816437506,0.9052167956386858,
 0.21768113967883773,0.6351379696301255,0.9718121321515395,0.6337511790245731,
 0.41984467201975184,0.9283605246580786,0.571945617531132,0.339130598973581,
 0.8004456164516403,0.25466815496915507,0.24816028761898445,
 0.5976894579709847,0.23935886142159768,0.3572063081282302,0.2215455384505588,
 0.6206416490958961,0.7255899259064209,0.35702859216096006,0.6950019051924986,
 0.3627623585909008,0.2939581107073187,0.4466172705292625,0.17180037747388321,
 0.32608737262205423,0.13800260789267005,0.3935073832925765,
 0.1455311897606416,0.5892194156201536,0.2110811330776009,0.45212298630391023,
 0.5549916007399289,0.7974690315737334,0.6808581270022999,0.34741489519405044,
 0.47101191665763353,0.6611770194053118,0.9878482330733998,0.4317617460484221,
 0.5741886022043979,0.9303431872533342,0.0239088722714611,0.06536816558503478,
 0.5719414465650988,0.13885742962914072,0.7771508129515786,0.617251666917974,
 0.860744082124956,0.36117594990375956,0.16556822277685002,0.7946306207166073,
 0.2684127899233576,0.4006023298942121,0.4867062887718071,0.02456946952993022,
 0.06726558123226012,0.055997528345091196,0.2412039929113926,
 0.8177860616723347,0.35772285207492416,0.19285646203935947,
 0.5496769707914007,0.9663081824390672,0.19585296554488696,
 0.008621474674931684,0.5302143114530595,0.9362210005482042,
 0.9000343121068716,0.3191310091947708,0.2043083325303804,0.05009316622712445,
 0.2873940164044322,0.5096143546782989,0.24181156103856183,0.5824301737881923,
 0.615404259950933,0.9701551982191516,0.8246703891924952,0.28437800153340476,
 0.014459142307538086,0.8927954340385781,0.801546894794271,0.8769026616774457,
 0.38562353622214773,0.791559791020626,0.8729974767557425,0.5034468803376767,
 0.5698715205968266,0.07566715545790004,0.6637155186755941,0.709577767856078,
 0.35419942249646286,0.12724503709504587,0.9295950876427219,
 0.4296796337924116,0.887174639394975,0.1264939245699227,0.373354268311196,
 0.08710281757302918,0.0833690760474185,0.8082190981481958,0.4437139524672199,
 0.7747116855787497,0.9875114143217903,0.15287299400777687,0.8807763186705592,
 0.4145670073058696,0.1737732133932436,0.6234095472866695,0.25865234916681445,
 0.06611387531829649,0.23293347620797433,0.08141387220443752,
 0.3034055508315382,0.939964925560714,0.17872352629000832,0.03605386009764433,
 0.1729309407014975,0.5199417717730349,0.6268446445800978,0.09204492216716353,
 0.8379144848218151,0.6526984654320331,0.02235712600513151,0.1814129576343565,
 0.03162993652845669,0.19783934321940344,0.48421098431569964,
 0.5096097199264895,0.06726057738753344,0.3064808172239384,0.8631335693798249,
 0.11955817194619389,0.17915634328145447,0.29614002655086646,
 0.7558760092542363,0.12127833077295613,0.19117579010062147,
 0.3958979910296433,0.8451090817396913,0.08455509347918,0.23960131270554985,
 0.5065750780368542,0.44935353924937904,0.36033226927733697,
 0.20152347482976407,0.15633583723286404,0.9773813127367025,
 0.04497423685188573,0.7330590431184322,0.08832427884417071,
 0.7263663582996271,0.5436402194085587,0.07478143609687571,0.8360855674385548,
 0.844806029130012,0.2612764292312659,0.5676807312453416,0.7589537016459436,
 0.782993474211738,0.7630610304603326,0.8196209630428306,0.3187465996940818,
 0.536353733644565,0.3230508069401796,0.004555558864173803,0.7628265656611655,
 0.6715780167875647,0.1178868062429207,0.6100817944368379,0.46015359794783206,
 0.19690110146072537,0.7070876949829612,0.4371593953193016,
 0.28950599398885546,0.30363846414034723,0.19301530852366555,
 0.19809385615456687,0.57139531421248,0.49344147823530937,0.27179652680608857,
 0.016116043522328094,0.6898582841303341,0.806699608253063,0.8186685913983215,
 0.9273114045911333,0.025160027661671802,0.6112680925891094,
 0.29046767495644554,0.18043764881242563,0.009886805511263641,
 0.46919534058089196,0.8472380560176134,0.7513252192963484,
 0.09586037606607989,0.7590079049454906,0.9124564471127931,0.7064106443738295,
 0.7035242365797797,0.14001881312022124,0.5483257685703323,0.6118611512094838,
 0.4824018271223457,0.3734993933943881,0.10969939285829766,0.5215898091050309,
 0.172389091180723,0.8217219783404842,0.02198264712578868,0.9376488143925702,
 0.12643157540152883,0.9161434948571243,0.6779227701482725,
 0.08346016567706804,0.7206353522818981,0.49117608084430786,
 0.03305799699934131,0.9778533048408005,0.222815657261911,0.7226962994204016,
 0.9635114017344939,0.4326779730784438,0.6991432319854622,0.38118450174375174,
 0.772578358142343,0.7889698327331827,0.12480829437862873,0.8460793075964426,
 0.9432519267211802,0.4672209556439648,0.9798226554567686,0.8808825780065392,
 0.6075081165635818,0.03802681153142862,0.9201643410789989,0.9178406336034862,
 0.5292989559536033,0.23364679626319595,0.729112699743574,0.7736002301632148,
 0.8262074131272831,0.7798608023076072,0.20317050543160142,0.6769824527765385,
 0.6145759227460479,0.07865066184927683,0.3877707258196279,0.4947954620216525,
 0.6209852912166771,0.24515245384631923,0.9257671060435682,
 0.18184743261000824,0.40810814201556345,0.21498150077158806,
 0.1442278655383613,0.5492644901745086,0.749691714369243,0.9849775050300539,
 0.017401113886171715,0.3052854820855474,0.5078723042907074,
 0.6266728105276876,0.47538841547247,0.06685610293046396,0.04039451378060821,
 0.33704534672568964,0.3925145311258521,0.6080700144234685,0.8427643073722684,
 0.49769796515406817,0.04738824670554864,0.9556598826745132,
 0.04910663900146828,0.4013423595161798,0.6754294290272589,
 0.058251424166771315,0.8133491765040884,0.3793091272479592,
 0.9023336251836727,0.6044212039281454,0.3288152834958781,0.232480429876196,
 0.6997129728498293,0.8696821962033447,0.2515788032355716,0.2256726836706482,
 0.4419331423756053,0.0025482318535058335,0.19288505502649833,
 0.610299127339711,0.4766802100716132,0.36033934665135736,0.4750564759974628,
 0.9678045742943238,0.6233906570961115,0.8984066387374124,0.03750750026992511,
 0.10496424703782503,0.23332667185401423,0.05304637734427309,
 0.8910784347163133,0.8012996009201024,0.5741358097076659,0.6270712455249432,
 0.183632506336057,0.09240217440689924,0.35127972357976966,0.9147524547913044,
 0.1392462282451128,0.4396817839036874,0.7265487958051053,0.7693090638611604,
 0.3434769656302643,0.8858648012601037,0.8195938432892038,0.6883347946564102,
 0.921694509331785,0.028778974887851416,0.7779492383706104,0.6860493706759677,
 0.8291258985320925,0.8367121512460524,0.5288884228393261,0.7811890813059632,
 0.5533224228910003,0.47897779204754465,0.2241345730612485,
 0.09899209028830891,0.9996783068740712,0.606416892485147,0.8280129947322463,
 0.0958614505232035,0.9888265555302183,0.8514045834571069,0.20702063832837447,
 0.6622015191309552,0.4593856499113156,0.10097154228732752,0.4967047537293584,
 0.8837169306408169,0.09678692659112409,0.8861937909584219,
 0.07971450103622946,0.7692782598627532,0.31942412483813354,0.595041780606849,
 0.9596134119274239,0.4007621092181691,0.6902861056447459,0.6920168774425,
 0.6348726121271964,0.3351886079998776,0.9164929869349135,0.15674697507690238,
 0.2786276007939661,0.45978812310591755,0.7603908051480528,
 0.14338695947699598,0.741682470046625,0.25889931004841926,0.7436099354976922,
 0.474043486063905,0.8472788920966532,0.7551258749087927,0.19434667963674335,
 0.020489811919158885,0.9442749635466231,0.4339253773115973,
 0.9173356662402342,0.7665798829971228,0.7708635938287216,0.12649417916968342,
 0.5383073185442153,0.5436083392732876,0.6547852620533834,0.21442313291916426,
 0.8769883544934464,0.4103472668347439,0.33274925210265993,
 0.07582894758196224,0.1559392984099196,0.16556786856315753,
 0.4691601975351467,0.12279342276565275,0.1784623016401713,0.8609800134705475,
 0.9879572741218284,0.8263281813807328,0.13816943169149998,0.8714286587643251,
 0.005004340483897973,0.9699594924053245,0.13847621246624398,
 0.31522158330318373,0.7142483605181232,0.13878658449923764,
 0.7099988757882508,0.46798740388157634,0.1445529772459142,0.3166814559022797,
 0.009257532452346173,0.1975258432543895,0.03459983817801171,
 0.7392539157358097,0.42843278064474744,0.3916836915428562,0.7954388319481953,
 0.7645622606774445,0.48478413367949913,0.9393533288160727,
 0.14968189511350694,0.9406314595426692,0.7932502590196346,
 0.18045737562385067,0.24156268899381006,0.6024198006736423,
 0.8521504655913431,0.3994146554569342,0.6788379479813662,0.4649170361558682,
 0.21626362426259405,0.029563151132137144,0.34181820290903075,
 0.3104357947062315,0.18366396552041597,0.6913071365158983,0.5543549588750188,
 0.9320304196066331,0.2030415007099785,0.5935993876367244,0.6134460441287576,
 0.37129682903850203,0.7078736594290157,0.3726206750825849,
 0.11907813154266833,0.915481660805112,0.14839679905100878,
 0.28855386716268094,0.32977128042953463,0.5746603751741808,
 0.3671978957584894,0.975954985718646,0.4418932264388059,0.7788631044151837,
 0.06615907995589321,0.16010042847543526,0.7248846343337009,
 0.8424921890553193,0.7481276255671738,0.37320856061021745,0.7577573238403408,
 0.3883877434080837,0.2869976867066035,0.10645412455252401,0.3668813659918364,
 0.37639603933094934,0.6845746463887479,0.7374609809044356,
 0.004622249770100173,0.8086703279871758,0.3203957122015747,
 0.22737049624678052,0.7164278605545689,0.607400238060348,0.01231851730620337,
 0.08929706455117714,0.8535443543022456,0.12094521384846701,
 0.8496445603612502,0.6985325672732436,0.5509919835195626,0.06523799736749059,
 0.32637567448480376,0.08347701205942859,0.41780338008189544,
 0.44833671644550055,0.7486049823112984,0.984536473322105,0.1928183695353245,
 0.5256738332725525,0.4415000228546616,0.8353216656402662,0.6237942845855086,
 0.3848970518368394,0.3047323466382681,0.6707551986670294,0.04852729636492781,
 0.6967470595231398,0.6683316994213032,0.14131518343210636,0.7768313128651829,
 0.026491669933483175,0.20117080386269515,0.8987235551373831,
 0.6013265831878303,0.797432821171416,0.1570385120533555,0.1065806985554717,
 0.2489048745387915,0.9081111356591617,0.30167422656043164,
 0.20810634920846627,0.416493317563396,0.729192416549706,0.8403657635056273,
 0.7573543014585294,0.23084362194556485,0.06363221557893473,0.889181730054232,
 0.7828255245796188,0.8824606137785229,0.22712801676304473,0.1981803308470953,
 0.21696665393345516,0.1876358721886595,0.46671740546426443,
 0.3933816538145507,0.9178507436326392,0.6490478883374694,0.16199072450198415,
 0.8303427797196311,0.050637457949749054,0.6599406728564715,
 0.11324610974732707,0.4170359414027168,0.09780587445347944,
 0.1223669943217065,0.42673213845356184,0.854882213916229,0.24114050080704652,
 0.1468586220383954,0.7187310774498753,0.24278688934815584,0.846328134063497,
 0.8173621745588976,0.6002148450730415,0.2939850738335532,0.27909344220508614,
 0.04188506054357677,0.2450756418930059,0.9337683937989166,0.8751111388152346,
 0.7700071533900674,0.2763883766989871,0.667465182469624,0.7958299053032243,
 0.8740185161077989,0.5816261204809524,0.43047839122103704,
 0.05472849543669933,0.6811452277478752,0.6124315772596896,0.8280420969172514,
 0.53397615035198,0.9868253978681994,0.9617853110580946,0.5802597359977649,
 0.2703435479243669,0.11910614499608008,0.8735911134976214,0.9591244408725959,
 0.22630461850301953,0.46011148894193277,0.38888989966063325,
 0.5749873253152654,0.430036951846236,0.5814025227749952,0.7424499793054592,
 0.05033138940999371,0.06901388958311683,0.35125229530768953,
 0.9794338316663145,0.5129516924219606,0.5271153065091333,0.14215041827881125,
 0.10395120895977406,0.1134387269062822,0.6790771816650355,0.9421812748668035,
 0.6286695537249206,0.17640236045469426,0.515946046165516,0.791565916762173,
 0.5599808922404965,0.313717141144347,0.411234043589905,0.331396299799374,
 0.7880487951264529,0.35168462392222966,0.7647774815521071,0.8593219360983905,
 0.6454357813987934,0.22262975159772602,0.6748937013475936,0.7397465730761958,
 0.055636389541627196,0.6813654559538355,0.7495708242777043,
 0.8227344858574719,0.5325298388086633,0.48999034650580275,0.7779502138147468,
 0.5197905768348337,0.2168766356566144,0.9474213103039478,0.7923719836292563,
 0.15080227860437712,0.6179055239726746,0.9651564856962422,0.5631578726186471,
 0.4227561378355522,0.8303566927394236,0.8622039140776414,0.1435144523874068,
 0.4875686699122541,0.20293717043219328,0.41238607874531163,
 0.3007214538517432,0.3585129071745654,0.0038097575071537326,
 0.557167774873862,0.9383897414304548,0.6495235432316032,0.899670052438392,
 0.3727949460324467,0.6617753070303155,0.047361936889650624,
 0.5865378509825181,0.22725855816652207,0.29111144259114796,
 0.7646952672151468,0.45880840598803774,0.6187973856055393,0.8297000988194415,
 0.625979738153072,0.8188170842229884,0.4907898567491552,0.6232871850934091,
 0.42863875368422033,0.037154642990136044,0.9258300537346136,
 0.9366047930134186,0.6301685929024318,0.0493364072508804,0.8235896019716977,
 0.8786966911211028,0.5287747848852843,0.5343265047924458,0.37824235155647046,
 0.2499537915125798,0.4871107061871307,0.31798643718684827,0.9961598696869313,
 0.026121458585168345,0.6919331288476895,0.6478632692351842,
 0.7237751806107607,0.7707536336507432,0.745308993161849,0.8150629822805168,
 0.2831770073362505,0.3139588915134065,0.8584106084872526,0.38000835553141343,
 0.8690252882019839,0.9203306572298936,0.05414168401058694,0.7639332657021819,
 0.10897891776564717,0.6433615088726735,0.7825231269309411,
 0.23654420422631772,0.6160120592034191,0.8237942802188555,
 0.14932251641856598,0.9407803842776393,0.02791824361886619,
 0.7287360004845894,0.8072559475736248,0.6743778716308494,0.20963265740352954,
 0.7374572342857161,0.21339193998365946,0.14512631631133566,
 0.2854165208800117,0.45087839819582654,0.3343195452064125,
 0.03473507888441196,0.9962241079032363,0.14491065106076084,
 0.08183784568123198,0.7706539831425557,0.637149702605529,0.5224858758347141,
 0.4118620954418807,0.6408788828986954,0.8686458763760277,0.2550160234126352,
 0.5332274226204278,0.8093343328860919,0.9525999016515168,0.6768167062425657,
 0.38456255453422283,0.7980069198610257,0.3147116885638609,0.9857584706388209,
 0.05160102406786149,0.9945347954260043,0.49463299198338406,
 0.7576796005637723,0.48877451477755396,0.3860343884970078,0.72489162659755,
 0.13988901993124359,0.7857404933882456,0.07979235523190842,
 0.34344591432167937,0.14554030114375216,0.5891335027088869,
 0.022433998134404498,0.8831264844640867,0.7865724167401973,
 0.7335351595124082,0.6128015395773645,0.8269635385984926,0.5904982392513229,
 0.30378505300037684,0.3475360150476483,0.0340072711349082,0.4238468028429506,
 0.14408184999767726,0.996755484950005,0.5154102085338224,0.9916336080146669,
 0.169502078466781,0.29766538295757083,0.5985875349545023,0.9853752066060303,
 0.6369709998443968,0.7018104838220656,0.4556134502944753,0.48898968121701114,
 0.19896711660844413,0.5907675402900471,0.6406615502401667,0.4050209837682933,
 0.43500599994910893,0.8797215044068291,0.7952149405547858,0.2513277701520171,
 0.24479548388382488,0.3023879179899782,0.8712186133642356,
 0.31341132147484463,0.29028333807906725,0.5550623957045828,
 0.7531121122589898,0.6902541966862048,0.5508261440346088,0.9790886112320971,
 0.888483871910742,0.7539628190682124,0.170692570359692,0.7362645962369954,
 0.43015237732605827,0.11732503566739483,0.7234273158324374,
 0.19297967335549293,0.2124099950677406,0.2665569409722294,0.618476941106936,
 0.8019022046289213,0.019028012688987128,0.3505365629817262,
 0.37685995508205217,0.9914396933691658,0.8954444073275019,0.2898510261803183,
 0.5650095173748334,0.7491702047374272,0.20093892811620984,
 0.11855137079041389,0.6559900189307697,0.6567814043669908,
 0.03251807997704537,0.9537163637638294,0.8836786333800177,
 0.39604820283305153,0.7692334406715713,0.7627986987603286,0.9947833450995772,
 0.06198131134697582,0.7481836130494866,0.8902567977679502,0.9472657375155652,
 0.8594644695827955,0.8084648993658266,0.9454651078353655,0.8922419091881658,
 0.5835764420189993,0.60505859503302,0.6542059498718406,0.9672583440273905,
 0.043289659036892414,0.0564505033479914,0.40366563704509273,
 0.621871869673833,0.27828299279424895,0.048342134615345644,
 0.7620443652235731,0.5373321711112062,0.41322012870137725,0.7820271592178152,
 0.5383719491385921,0.18042821700964984,0.3937802169649737,0.364186272322616,
 0.03994799541345828,0.8374361488206243,0.4395409824900026,0.465544639019694,
 0.6999374384556183,0.35840341741058035,0.1548657882852209,0.2253550618479533,
 0.16420312641014267,0.2785298179667993,0.364698555187165,0.23605252544174293,
 0.6900049320294733,0.8203731307233407,0.8958585171828413,0.8158744246632399,
 0.3872097726452999,0.677373370822208,0.12374021852667638,0.24727119994148095,
 0.15706794885423958,0.6308647546467079,0.1795105672454348,0.3474623539725772,
 0.055523327310446624,0.09097254493126705,0.654332619180267,
 0.035079272158727526,0.23708735358225808,0.253588998125421,
 0.7212709375663519,0.26309633319585657,0.5773514469870097,0.9468663974886258,
 0.7998359491854146,0.5760303631652416,0.24502368846795786,0.9287000049507026,
 0.867673172666205,0.8929752551364478,0.8799687605771315,0.02169757959471097,
 0.6759798661209024,0.9999982069053269,0.7726455624488776,0.5795711014908118,
 0.8917617470912597,0.9445845651434026,0.44928898493844827,0.800942442667195,
 0.07374735361607232,0.776326768779152,0.5146729798965648,0.02510657545849282,
 0.20508407462913525,0.816134285131638,0.7112567292677252,
 0.0073548210601278585,0.635489985618441,0.02321504765671023,
 0.44080035945563445,0.20524167485023903,0.2596833273469328,
 0.14035779109532753,0.5479983470441019,0.23974190158922126,
 0.011308318240293214,0.8908437230938784,0.7539574103029411,
 0.5349848038207602,0.8269740641189063,0.16647090465833647,0.9843561492035463,
 0.07198365347313695,0.8632150380730754,0.4695934950205043,
 0.35092094771615256,0.7960226390671448,0.8683735985221471,0.9022124661847579,
 0.0019301785286883755,0.898258000866729,0.4892757515833839,
 0.0936344924584328,0.7450583474735633,0.589225288316851,0.8804126958748065,
 0.9291208165798652,0.7553406007946116,0.4699116743298405,0.5748565440683213,
 0.7639537950584139,0.49150956956610725,0.5013075325239039,0.6085446664863516,
 0.4834766392846188,0.33768264095606804,0.7521944631062049,
 0.38480938736026227,0.349681656550905,0.534969797198823,0.6720498855618015,
 0.8942235236953473,0.7373579674336564,0.027372337429009796,0.609790009945077,
 0.17645739533902416,0.37784348974283966,0.777460001933999,
 0.37602202774302873,0.7775006151466827,0.95973737484166,0.431697227112399,
 0.3456099259600366,0.4087741073828024,0.7442108241984476,0.1244385379218752,
 0.5497365167538687,0.9341158432184378,0.4545626173272843,0.22435675508502684,
 0.3596975470687619,0.7050963021797784,0.0905938646723291,0.3767595802137357,
 0.5911468727639064,0.4600166465194845,0.9275362495864379,0.5470801666335116,
 0.3184574323661047,0.9637213172622627,0.5190817022720057,0.8602782792597288,
 0.8011512730842825,0.05097561155206565,0.8965417866760015,
 0.06899579566306635,0.09040275957663291,0.15329930980860174,
 0.1705825607869237,0.26919283159764706,0.2246962365316436,0.6409905066403069,
 0.4104462410276195,0.7949774318796223,0.3540292955672195,0.6151898702936283,
 0.21239348621819298,0.31652373792349353,0.5092899317856591,
 0.3972529709248288,0.2668100531233196,0.9395200643291787,0.48240320391464964,
 0.6293723895254507,0.7876730758361012,0.3207504932203611,0.5565620588725565,
 0.06491492672704124,0.9691763257427323,0.26154573209245124,
 0.41024177758550395,0.5735363508812178,0.5042520921342276,
 0.24674643692429354,0.019479648323135312,0.24097191419617436,
 0.30498583732523477,0.27683818520664305,0.720889861421916,0.4906395262268731,
 0.10000064634704109,0.27629292208867984,0.95308589331621,0.5522826360741969,
 0.2553896216427012,0.1718696812174909,0.6202217526596179,0.8806680935147415,
 0.546841235395461,0.11302187169977129,0.2278833653461635,0.19354560448735503,
 0.508141287171817,0.5384600288947203,0.5842949973509824,0.07701152037075842,
 0.8075189505493586,0.8352116059310509,0.6426626420625088,0.20528627069062844,
 0.06233418560471826,0.5530800668073691,0.8472233474254973,0.543068275223264,
 0.5348294592649251,0.8058163874340734,0.995870212295243,0.18196184622829648,
 0.8731017536874157,0.8592276223736057,0.22144899522766637,0.3079313408738853,
 0.47730278867942877,0.6337574295728804,0.3709117185045516,
 0.21942902370520712,0.7143666237006646,0.12513826461157795,0.824384275503561,
 0.24309153391131433,0.7824613108133277,0.7164083104116543,0.9968115931907138,
 0.6684420226653078,0.8888061991780989,0.5249701327570437,0.205192820272583,
 0.40516946299530954,0.49316619768461323,0.7799281545531788,
 0.6373921949802451,0.5947731483635381,0.26753156105840115,
 0.043530058403498995,0.9271163242954796,0.23324609034760913,
 0.6739105073078753,0.12236219832209805,0.4187140387488264,0.6534823530397785,
 0.10169742859684872,0.3140592113472168,0.02583598455811842,0.384502049394561,
 0.9448893625327498,0.4883764076841892,0.6000792736342021,0.5658651944119333,
 0.7574722698873554,0.41303602252614036,0.09177954422165624,
 0.6195930948345547,0.17588403477119152,0.3174197251030172,0.6495545050100193,
 0.6494770871759443,0.4007338957769151,0.6766033460044014,0.7342463875389035,
 0.9657588303365424,0.6964494267276956,0.8428357923123493,0.9582187834057714,
 0.1758894183550812,0.07037447074773961,0.30557807300572193,
 0.31038912028535204,0.5373687133762082,0.9192373158510374,0.1703627772965286,
 0.2616506640258357,0.005446639202643322,0.1186268210452639,
 0.8600559530041585,0.46599638120237463,0.6629725946726261,0.6569175563318288,
 0.8318627999711037,0.6924089454377741,0.02355168191554624,0.6701788183880457,
 0.2663043556763146,0.49176645845305433,0.7631276741397501,0.5529101758847602,
 0.23282150266996016,0.8631119987598577,0.36837472598301335,
 0.08123877386702649,0.6090373784942263,0.7518315223945107,0.1594365602968847,
 2.4345876441855818e-4,0.9026185680091553,0.08021129848776598,
 0.3971450685040151,0.3046208142039879,0.6235629451077953,0.5167297185541633,
 0.703785346829803,0.06052823559903886,0.8967641837016266,0.6581093323618112,
 0.7619921458699699,0.04892472712533047,0.7034028481658845,
 0.22959006840312457,0.6854157498602582,0.9705272653853786,
 0.36113802969913056,0.9769994003331495,0.9843471022499091,0.4707256077463411,
 0.7407355636541968,0.07167283017586634,0.6071462713616909,0.9037875426840772,
 0.09653531753797018,0.01593468473638704,0.738001036878537,0.5535769940939705,
 0.4296321650083238,0.16981860541015303,0.10721445436531418,
 0.13261738522234823,0.22244482483742067,0.5506030141747416,
 0.6968365060894116,0.8973976362754437,0.5839149317275547,0.6250774407221852,
 0.2993686842033222,0.7243098322598501,0.6781332965907079,0.217950980473687,
 0.9381881870675953,0.013540531864908889,0.07140781085096393,
 0.9806462057805454,0.23989901123673674,0.09986779305435334,
 0.2632372202916383,0.16930727011489943,0.5471372087552313,
 0.37198656899401783,0.3986181566007414,0.07464949306046309,
 0.9076689943718639,0.8340732899657287,0.5191721345283038,0.8853662792193495,
 0.6684536708403983,0.19973446427007702,0.595109520045497,0.47829660003078445,
 0.46604701466699927,0.7532393378964929,0.6590981887173323,
 0.03802659398644215,0.6108187193816499,0.8809627944482757,0.7629192128612562,
 0.5998912272768907,0.2226985036471082,0.6250944867599596,0.4117560537655782,
 0.021335378141940975,0.6873553380952131,0.054772018058070415,
 0.7024885483008219,0.7113101703215644,0.02185309672787361,
 0.34056571284286175,0.13037177884528695,0.8567102236455942,
 0.9516961210524779,0.7152602383127621,0.34315689962944074,0.8030636148680043,
 0.9252899305463363,0.49310624902132805,0.541260940270281,0.2302820416340996,
 0.8560026627064652,0.6245130100242848,0.22184360620645982,
 0.05436870992301057,0.22156296179762236,0.33349076130021826,
 0.41333815945693253,0.8801501568601788,0.747043752933802,0.3193482510203962,
 0.21029831443614067,0.07048987163294307,0.9879632937341497,
 0.25748417526259726,0.25288077841133694,0.8205693303459196,
 0.5593562296049216,0.6419782722610531,0.3694266758289375,0.24775550881595,
 0.8542901586602007,0.40107272549369577,0.19638421499500958,
 0.5177294511059893,0.8767824618400637,0.047666090854261545,
 0.9976500308299067,0.24602619607897536,0.08505655449671734,
 0.3810398349721852,0.296671407958635,0.09354110614500422,0.9746059728869056,
 0.38225819536571193,0.8838148498117625,0.9141045211578658,0.364458033890919,
 0.5714745931323625,0.2631857810042061,0.41060875967091737,
 0.48632230405051535,0.41629007345011637,0.35503656890365165,
 0.9741600894749388,0.1278585372665133,0.07301294888720755,0.8177635404902277,
 0.34608907251893883,0.40952737037120013,0.7026962311633295,
 0.11765381086652749,0.8571416546074009,0.5033121647245791,0.3913327462942333,
 0.007166004447443308,0.9827622881047997,0.0947351907739662,
 0.8975020452119506,0.9161380123045004,0.1546853813368938,0.8315876132892085,
 0.6410212073414586,0.06727804189385134,0.19502960275927972,
 0.5815496704240881,0.500634624711442,0.3596189101165498,0.6241258439952035,
 0.6802584567627037,0.5342546147565929,0.3066376436064664,0.5542685496333543,
 0.728727036043077,0.29367973792536306,0.7845747684906835,0.15631772999945504,
 0.1527267429868724,0.7180283682349314,0.3532734709326074,0.5169860424390266,
 0.4443001186333634,0.02450201872479918,0.7654063713252228,0.8735623903467372,
 0.3496362354613449,0.29888882952147244,0.13987262147963708,
 0.7247375063585128,0.3763087857769213,0.9165728253186891,0.578560048170076,
 0.4297297684186284,0.9196215387476693,0.4090742961056848,0.36836009295747507,
 0.2841271790205685,0.991586126277745,0.0717286610387804,0.24610736956543577,
 0.6048631961149021,0.6717422343512656,0.0562315213172625,0.03538713917865843,
 0.37532667003382225,0.7735367902745536,0.21288319352479357,
 0.20189898664212014,0.9576501464080681,0.030817214043385777,
 0.24236505237385386,0.5198689007956414,0.06099584753872256,
 0.6948919141322475,0.8757024340089735,0.6981870617850815,0.22124055929845798,
 0.8402481965234376,0.007643542696149641,0.055529041283450375,
 0.775798879200139,0.39456938972696476,0.39359263980412296,0.8047036037628199,
 0.161126545602862,0.5408709770825447,0.2956172076445144,0.7879869494740053,
 0.15109570919852644,0.9230247846349012,0.866712410572909,0.7800034506776565,
 0.6712537825610861,0.3885280587308315,0.6638611677011812,0.1774493366229234,
 0.13701127001390823,0.8975991435812023,0.720983880459731,0.4683976321694059,
 0.01803605657410745,0.07239651448584794,0.4119289016212676,
 0.020619735353132995,0.8713685300609275,0.8723211501811555,
 0.1617795174834038,0.6305955515904939,0.38778285188696393,0.370779314566106,
 0.4473348602117888,0.6804769522138654,0.6530710079672173,0.5839951986939589,
 0.5011978577842856,0.61048027953056,0.27050755803093607,0.929390338496632,
 0.1233077915125369,0.06562855475086926,0.2913506799198058,
 0.041517755502663345,0.6310248844975256,0.6382597827728009,
 0.4275342834220698,0.19305765107000517,0.12382192628818534,
 0.6599432624487401,0.7218115283167722,0.1670153005837164,0.22024271349430136,
 0.09447720841072804,0.2631599462478017,0.2768498194708632,0.7604258313319985,
 0.514797811484235,0.11837919492098337,0.2266556870624794,0.5562091942114769,
 0.06480800966493094,0.12126980692362865,0.41287593535802525,
 0.026654785272381742,0.4587309133256121,0.4852590336796264,
 0.1434315733216357,0.8042188779151318,0.015954133207913723,
 0.0027833517659358042,0.6010213070762861,0.34304733549024824,
 0.408974067230374,0.5924614137094469,0.16672646459878893,0.824260374891396,
 0.40147521212009096,0.8044757281574514,0.8593538808699028,0.5979065731481719,
 0.32507375834914176,0.8107749276853851,0.15174334042297077,
 0.8393870279517809,0.790428961707823,0.9080057282390597,0.565212290941219,
 0.7356483019732792,0.3163581405377226,0.04091275184102139,0.8083796456865661,
 0.7698009383283922,0.898084629136079,0.29829220010665225,0.5689781917902051,
 0.9828114310438363,0.8309627903336076,0.8242294276127987,0.662532068439842,
 0.7303798362935695,0.5624230863288373,0.3073037059719177,0.518147200067328,
 0.47512267331018465,0.00289577426994736,0.09041059947151353,
 0.9461054341027989,0.8730352547825877,0.6076235743786266,0.9842525506605909,
 0.9838013009659379,0.17380016824831446,0.48803538952708647,
 0.5512898494021471,0.5428656051707147,0.14615121025070055,0.9493751580211998,
 0.10960066439807326,0.23285593760048806,0.2141400056320728,
 0.49659485000449577,0.0672783158553677,0.05996550353645724,0.938814630630606,
 0.7621191059969132,0.9355919807885988,0.8598997142461029,0.9102580293660522,
 0.2377729117223938,0.7605427939190094,0.9507691545069301,0.5851890690282072,
 0.7847599191003027,0.10345717341821825,0.5732739607849373,0.9332826711158962,
 0.7310978702916375,0.9125742563011265,0.4058377396514198,0.4861280164822458,
 0.537197665442109,0.05781777900833274,0.17130991987792332,0.9663520271050426,
 0.4993656665376561,0.7658817039852317,0.4485948522042552,0.9462509019354711,
 0.382262254833179,0.6792838544411691,0.3995618512173037,0.4241987481371102,
 0.9499643945183311,0.15005720093987907,0.30145374527441415,
 0.6105890368338057,0.1933853753410275,0.6691979563774797,0.2892016992433796,
 0.8593711916352598,0.3574356199427515,0.4135193783321509,0.9220916954905,
 0.5940815198389439,0.8060887692657386,0.09261168370561823,
 0.40791303514275234,0.06548950723091496,0.0878894819104431,
 0.28178582615760395,0.7528852887824099,0.333993659384364,0.6836696557283176,
 0.2045022983772936,0.018563123920957825,0.9279947593424651,
 0.7169311431083324,0.9037977007035177,0.7844617455061859,0.07414012492320199,
 0.08788537619177084,0.47660079153639656,0.16180090155510407,
 0.8136706516481289,0.43544643083308576,0.3330857196803637,
 0.43047387122988645,0.504535979062967,0.7759080561740895,
 0.002351855183327145,0.7913587803246529,0.34133213050931177,
 0.6002517490674816,0.17096697866703603,0.12001933546962551,
 0.14140359137004377,0.5785481724527166,0.8542906177643016,
 0.40873887907307604,0.7974345892803218,0.14886606273350378,
 0.004077784735173284,0.25239765191122554,0.13488466399622812,
 0.8476065735342581,0.6480960026016814,0.317386741622816,0.6578460812068221,
 0.27280787394501516,0.7747597474410556,0.5059289306308312,0.8882519309864617,
 0.4001988915076893,0.019728260771032846,0.3469097274745542,
 0.9767470473670633,0.7159952232027093,0.9961041886076613,0.26786200903187474,
 0.49371027550479407,0.048531549261655704,0.7521261935931972,
 0.7165280971110677,0.9615572214481178,0.1242546918314269,0.7383315889914682,
 0.4441993790515312,0.15460350015021107,0.9858291867662232,0.5893143657420632,
 0.6208480011318677,0.5813078959934403,0.17051013456693886,0.9670424532840141,
 0.6368709816680435,0.2943426195478659,0.9241573750026849,0.44278240523291423,
 0.4911677679569204,0.4149709767290909,0.3341417650238755,0.3946118306444272,
 0.12044357021659902,0.8094994051916924,0.902240585625919,0.7007640911373837,
 0.5081405479431818,0.33974419647633514,0.353719608789518,0.3146804452793035,
 0.16233718403132158,0.5766776116330256,0.6192122037572634,0.7667591906060821,
 0.16593719825400277,0.2442365281505836,0.02256311973474101,
 0.01042739402896864,0.6064658699882913,0.29559814138574037,
 0.8017632934977404,0.9637785066159856,0.8501786306481445,0.1808551420814556,
 0.8370556353265198,0.05854534923069621,0.5676012515545094,0.7244563217408205,
 0.3989342914747397,0.23119713320816493,0.04969339936236583,
 0.07858489054784001,0.6240725738459748,0.13249672586983063,
 0.26435245228381965,0.20158937786938125,0.18604102997219651,
 0.47404766414964083,0.24912859431383538,0.6979403705300242,
 0.40199110859499876,0.6247396219188625,0.25975683199364397,
 0.9011773654848821,0.3791804817081732,0.6138269097787974,0.411212458305222,
 0.17445802194209736,0.671770621666398,0.9094094103630446,0.2678282790368638,
 0.44502703779673647,0.4703457147388186,0.6310342877595947,0.5448985093085557,
 0.763604549923602,0.3053496709910718,0.6026582229407814,0.9138815313986128,
 0.7002505839074473,0.20622758868135782,0.9250796099134477,0.5619465060192314,
 0.07412774412199341,0.9653519939636183,0.822587377517505,0.2985232730544888,
 0.9052599258619229,0.5758361113988202,0.23047602009886947,0.8840621211082849,
 0.9427514157268094,0.28986389186224926,0.8163936716296866,
 0.43056797205037534,0.4779804202744766,0.050418658067027944,
 0.06176486265880521,0.45684826561285086,0.4564569974233863,
 0.5951498474452787,0.8658712387597544,5.700336899167047e-4,
 0.050277945767117194,0.6902144307669352,0.48122472839961616,
 0.6527873741765989,0.9498522322537756,0.23417726798725214,0.9893981569109239,
 0.8705432165946387,0.8361869488132523,0.31381680646392784,0.9265246452067216,
 0.09377195601149424,0.6190494366001499,0.34290240648394754,
 0.3400626967127768,0.01922281962719219,0.9285332703777786,0.3712850056527439,
 0.25805401213225254,0.9040718833547609,0.2219267622096438,0.5502482893954647,
 0.673654012171238,0.8296581812280337,0.9044786643329837,0.42623174415621456,
 0.6464800058348016,0.3990586786523629,0.5971660284862549,0.38884236797350935,
 0.3912986532158109,0.5636000009290945,0.061709233910275385,
 0.6988396250928738,0.8199395144432426,0.3915783712770683,0.3178697468408367,
 0.8916660364231815,0.044978762737805544,0.029845763126654656,
 0.6436877881713016,0.20070791258534415,0.8832105177713959,0.6783214339961382,
 0.7619931313230012,0.04188358422142946,0.010426639233866886,
 0.11964731330243505,0.9453430816659659,0.8053053380303982,
 0.27634615466537427,0.043833911065398135,0.26607406206442463,
 0.9785417907085099,0.8959680329406219,0.03327537292344818,
 0.33075726547453776,0.6362843711240056,0.10762664566384228,
 0.0397793188983232,0.9713495802859958,0.08961040185052715,0.6748525037284233,
 0.13386881775470827,0.9524279707258119,0.14362220053072272,
 0.020967593021800845,0.10007395659751561,0.7893520048904704,
 0.8874972013841134,0.6292432928176233,0.66560306430563,0.8286820025383341,
 0.8070492162555373,0.711494583243381,0.44796276002918356,0.6429556511159275,
 0.5244771977260595,0.24736829559025608,0.45179965611436756,
 0.2732818968747003,0.7791637781883664,0.09666914156045425,0.802112599463083,
 0.4786149860925787,0.33255113758500776,0.03624382204139187,
 0.4909084464195026,0.8074122834039398,0.2690254320502199,0.9291711077982431,
 0.5421272367375567,0.5048084937333643,0.13859304515134607,0.2608229126887883,
 0.04242764727776005,0.08928656915052657,0.6268264931817233,
 0.7782517014206705,0.8978538546256926,0.1148162477344028,0.7319786961634327,
 0.7907838500006047,0.11808706669021429,0.32776551389238917,
 0.09839661496779517,0.4781055645317438,0.39457743136900403,
 0.6636074120068483,0.14081802698540768,0.41228988029309777,
 0.26971231384611627,0.6508206812963201,0.29890623223052304,
 0.7762645524346314,0.1311289758016383,0.847534411777469,0.1803503070030006,
 0.11937288522417533,0.9229364459477127,0.3791926457939272,0.6705947407804203,
 0.3462814226199723,0.9162368110880924,0.0860737108290568,0.9213785028694542,
 0.07590392055621242,0.35888099015044794,0.23609220200463887,
 0.9978910480173386,0.8609155803378631,0.3834014479607879,0.0334445715952113,
 0.5234111578262814,0.5487805868153623,0.13894647284912864,0.6072343374785145,
 0.35926537336668574,0.5052619836156552,0.09272718952201786,
 0.4170354279455458,0.510291510138989,0.2755334723494247,0.6559292168912525,
 0.22815835786569616,0.9796205489805239,0.07670719430852158,
 0.5955371410282352,0.34995801903813617,0.29361649399172995,0.533985252799416,
 0.5488374179652694,0.8508662867673916,0.7632923309935309,0.5520349131192855,
 0.20494252191585138,0.41285668110685314,0.6546116867472209,
 0.9922309169947477,0.547886163550492,0.4764298594152222,0.7189734636943542,
 0.5760863107162294,0.11472529362998574,0.7482423404087658,0.3310007942577581,
 0.7807714337420749,0.5183023811291355,0.4894350424410485,0.9169833858361671,
 0.8661735038224405,0.5602057372399664,0.01571319396543902,
 0.0013779277495435949,0.2817376157965755,0.3357472666336583,
 0.26763188580816544,0.19569145293194745,0.6572783086046613,
 0.4393400427826546,0.5759219260232675,0.11216753718530226,0.4025300001464598,
 0.8923674315970577,0.2016108963910137,0.20812431949289323,0.1622882584586285,
 0.4257287985106213,0.3383844519357726,0.5433054270191943,0.10138421626200511,
 0.9141045599803512,0.09561367776356577,0.6143090100314484,0.7068231440710997,
 0.9106925753068895,0.683431185812463,0.8962818042847199,0.29090120052900437,
 0.22383156301571194,0.22483806189965017,0.4375824457980909,
 0.4629554736707382,0.8336467989131333,0.05386376128396775,
 0.04141024640469271,0.022372669949304202,0.012319625056072425,
 0.16630068120566033,0.7382354043285477,0.9819554322256546,
 0.13181136710546426,0.6902900868517211,0.8704252336867708,0.4715525728292893,
 0.00581923191762912,0.483723235403239,0.5566098917403124,0.30124694092559234,
 0.3616848826553586,0.014357380126492458,0.9662481910589759,
 0.31579958472811376,0.3725681484262082,0.24180018414933924,
 0.1022684879229463,0.9977601465870104,0.8163035005201877,0.8595257940144991,
 0.9012771152327587,0.5163312259214816,0.4691276711084147,0.7670874184802119,
 0.7197075111104887,0.841547403884995,0.3216228886342928,0.005864245457923745,
 0.6430456986136672,0.4176847339464056,0.2564109824352222,0.5675378321961339,
 0.9536623152208786,0.001135611280147053,0.3647925455596944,
 0.4805821720864535,0.7091307985185988,0.8971587579162325,0.591934633688151,
 0.16567615604598296,0.5260384791150695,0.19872570226993347,
 0.5394490454912642,0.6299251656016067,0.7050912197484042,0.8977408288709309,
 0.8686925506802259,0.6191289919093196,0.27768929930518804,0.5720919168356258,
 0.7630405791035768,0.33886526972345965,0.646294598851153,0.11206945881874586,
 0.7018034888642923,0.21581115838951848,0.4561693703730172,0.9686673408866531,
 0.7491880543988296,0.16566794090039894,0.8122424190566453,0.7386324860864962,
 0.06457833111263112,0.9672184269111048,0.08829268574697158,
 0.9252187300181072,0.09529950738801019,0.4821434604210464,
 0.11824909902041159,0.3669731692853606,0.015858336099851322,
 0.0060302884979934746,0.10181168256901729,0.9975539958060867,
 0.4251650355472587,0.8755133092177371,0.1494096115827821,0.9410307775936516,
 0.1258963000674137,0.6257635937049469,0.683539666806394,0.9698230918164847,
 0.08872426821512402,0.16804595874985528,0.5185117219630074,
 0.47320794782897224,0.30889376502964194,0.7815927748480814,0.241341688025001,
 0.7857141396418914,0.46455264000371255,0.10060163322524285,
 0.30171156829759593,0.4154676068593367,0.18833987428714416,
 0.9507550062832337,0.8219480752875781,0.5751183889265833,0.8365220547135539,
 0.7316700970628021,0.9987027267609209,0.1427232204890796,0.46406472137712695,
 0.5300994467081837,0.19468942596863426,0.8958325133277389,0.9762911521228529,
 0.6271095171434333,0.04932292448994957,0.6113425396106276,0.9071675888442815,
 0.9194246759066981,0.004390046483061072,0.00467617952558419,
 0.7551174696053992,0.8947039131645173,0.16833287214994053,0.9938292699094428,
 0.6352510068103631,0.10946814972478314,0.24325565071466326,
 0.06308685064539177,0.9897166192041303,0.21038628687358574,
 0.06732354220958714,0.42723220691826513,0.035344358666699605,
 0.19253139036587796,0.5967479797443168,0.2393736504815106,0.7411275796809413,
 0.4729337846470377,0.30320920662263917,0.7706158793274345,0.9788951486260594,
 0.7255996580628008,0.9058567416129495,0.36554991667847037,
 0.031030108428909187,0.2730889487369258,0.3987666613649137,
 0.49403985973567444,0.4684744833727019,0.5995486312907004,0.5710920343820142,
 0.852131754353099,0.26578023992405386,0.7733284415379543,0.3336178673179815,
 0.025277289597263497,0.24763171792614758,0.6791414988530193,
 0.21897347032426095,0.7270911756493926,0.5950794153592436,
 0.18230852773619777,0.13473651084866634,0.035387400527568835,
 0.2911318815520785,0.8891702820980822,0.9532712499278198,0.6003236295023202,
 0.7932410808910694,0.4148775006262987,0.4055357445659148,0.9168077890191932,
 0.9599305567965007,0.10108436097096152,0.30974269823788425,
 0.14376268003057857,0.882052578076074,0.08559247010425541,0.3806531204182001,
 0.3091073979052068,0.8302599592176279,0.620998644754308,0.4876225285649264,
 0.11046014712449992,0.2946345123905192,0.611433845023439,0.8669953878914474,
 0.2853169120601683,0.8737206746694428,0.6255808303450964,0.6499165719760132,
 0.9083097970580264,0.7913282303621035,0.8763541532255588,0.9277920866035564,
 0.13030798985389458,0.03051545158821667,0.20478370641721266,
 0.7204424276195879,0.291051018141567,0.07906933201645128,0.9358905375914066,
 0.5332771558417523,0.36228880522025364,0.8266257794499838,0.2861316592510047,
 0.10381449166304779,0.9088673640875234,0.36795860036896466,
 0.8211851186457201,0.4546371228100827,0.001750458042971359,0.385201262022004,
 0.6125524396842454,0.8843900190280994,0.16705013929904422,0.6153521159044268,
 0.533204453268195,0.09561469628388441,0.308384755711569,0.0918710820615033,
 0.7183831512734528,0.12871420737976014,0.19659972475828857,
 0.23412110399510722,0.89725658382558,0.8042184473320542,0.05451669952586935,
 0.6231502958531042,0.10154400754079962,0.3098421016179379,
 0.14657206099533526,0.5202782288221722,0.6638412429839893,0.7045397798919864,
 0.3846615469970861,0.49129952629594364,0.4378213309219541,0.9124479412945078,
 0.6002711500025864,0.5921808868442987,0.42239618953638103,0.39689365051127,
 0.9104863514360768,0.6217357986222489,0.030999101550054253,
 0.9620601781016507,0.9273101607296292,0.6656874785153233,0.3395009105090585,
 0.04584337357117996,0.8244597984096105,0.21973197185957616,
 0.9469678602701984,0.6264939231248591,0.5639122473668472,0.07550348663131534,
 0.6179429345038071,0.6116613589746766,0.3688092922028203,0.32024308979021854,
 0.8398819928843426,0.6696750669037299,0.9553893650250639,0.5641309831594803,
 0.07183704937562974,0.4165337816487602,0.863891528741473,0.22711826789872014,
 0.4440874554116149,0.4460034543109024,0.19965577017635272,0.9369730138954344,
 0.7198287261617461,0.0876399603193142,0.3955955553213033,0.2239652645282021,
 0.15472406717855247,0.91167816511541,0.1070302137239807,0.9961144159431862,
 0.48088479328546196,0.050458901760624286,0.8999106008906558,
 0.5500582516149504,0.8383411482432063,0.011362082697439657,
 0.41556913161814246,0.7752032755342989,0.41868792502152796,
 0.006214565974095643,0.9453639623186436,0.44607275280158865,
 0.02477250509037432,0.12989300904646306,0.567409720185183,0.3304931916500071,
 0.0769824491576202,0.436387630627034,0.10874518112368159,0.46546692348714747,
 0.7240413236715574,0.30832011195171327,0.2720632740649399,0.4543792416099188,
 0.9531786954067523,0.48332404897138237,0.00516963451475505,
 0.6985719934770425,0.7415458713262194,0.2668362400259907,0.19255415046193125,
 0.039081045120851865,0.31205316031159513,0.2345746322168728,
 0.6288323703073442,0.0943611229878798,0.1218095154408051,0.24639708213943834,
 0.34441534141461305,0.20707124191418436,0.6173881616781672,
 0.8662073425642878,0.6573070450307227,0.53477361940349,0.4541250426605493,
 0.6922899162321586,0.18135017684890187,0.4619151527806846,0.6091235732632719,
 0.4104730476174262,0.18395024897482193,0.22004724755539196,
 0.06920351327952823,0.1069637328278027,0.9900899727365587,0.6030359713602115,
 0.43507034545229895,0.7115768072256086,0.598231439762881,0.3362479023284555,
 0.359355882932768,0.9646216300660884,0.5203237294590096,0.8668321833789833,
 0.3139840793462779,0.12981489628880605,0.5760632617223828,0.9136800649076228,
 0.5920227887134276,0.38622191712609144,0.6735891727433315,0.810589377395133,
 0.4213236024560354,0.8748591402926256,0.16108806113277307,0.3053162610543769,
 0.5262805588236814,0.42205953800458207,0.802691415903721,0.43449232305882135,
 0.22561650346702133,0.3012106410255223,0.42911802390620646,
 0.3830584700437537,0.42098785168318564,0.6627680855740803,
 0.39697991260745635,0.8679987035316259,0.4367567013230672,0.5975291805203745,
 0.8858375815863593,0.8979288329380158,0.05180590683203823,
 0.21404414219312085,0.1887607089130794,0.173220210735463,0.3336364025499865,
 0.926651898454265,0.17139482360158587,0.7619182699125725,0.4685557121622419,
 0.931333198654969,0.9208318415286185,0.5937454439448967,0.7754715421930194,
 0.778491393902595,0.23859309156127884,0.3243507118985537,0.9756954229099781,
 0.4597976496196672,0.213147795740252,0.7241875574495469,0.6723079351552411,
 0.8523165744663799,0.822757366194527,0.26199859426244143,0.6960049771897937,
 0.906629257193604,0.6703492245078607,0.15636569794936117,0.06258277552488178,
 0.24215466514686357,0.04701540263338411,0.8204640467051595,
 0.8500228240624416,0.6196613293662041,0.04234772403892739,0.7409116457906907,
 0.6270536683248744,0.8068663533674091,0.013545160554410574,
 0.7697114165685179,0.6344740623950348,0.6208260537063046,0.49030641911700357,
 0.18045484189755945,0.9467485490157157,0.4791877625312694,
 0.44256953426370493,0.24221082562301655,0.43160920065016084,
 0.30886901327416094,0.8458063159875698,0.8556450480381091,0.1695983325067404,
 0.1807737376678381,0.08658766357010128,0.8721018786221449,0.5138823460915816,
 0.8956392995685413,0.9941920649586532,0.9349549304316639,0.4285699165697525,
 0.3140246447845839,0.440490285957433,0.6739242745043064,0.7372811822103734,
 0.2263571852351387,0.2960669556990879,0.8887570835281906,0.2445608801784791,
 0.9776334571811356,0.7509220996735055,0.3855728701780867,0.26751209238816176,
 0.12837977777822962,0.3774114239672537,0.6643846009687352,
 0.19050144168176353,0.4073603634778825,0.2545670160427934,
 0.22036054599747823,0.39153422804216076,0.7068984947734421,
 0.5640867935912433,0.22825648193302095,0.4814699939335081,
 0.13988041451864985,0.6075869086615433,0.24222333451300382,
 0.35236352477830546,0.9856426934549557,0.8205639390150742,
 0.06746349066748192,0.8402139233049752,0.5261140383144279,0.543297100775926,
 0.6914031665821959,0.06273264223069386,0.7309683859713436,
 0.16649046310922966,0.10453424616063212,0.18741022636629479,
 0.8463861219992282,0.3789915040571942,0.01581665662842391,0.685406080780921,
 0.553730056730877,0.8312477739736169,0.6205891905424279,0.04696684207460078,
 0.047220610145924935,0.4732308369919831,0.8067720122183458,
 0.4999257403458006,0.18134768009607716,0.7612761766786792,0.9396517230157079,
 0.6199196821718329,0.6235349179090962,0.3481326175414028,0.268509594024129,
 0.13969078118725908,0.21609605990014147,0.36011941830456595,
 0.42780546849404166,0.7097862925308063,0.8431548232655752,0.3966420114807431,
 0.8100073698558271,0.6308150699596242,0.2862576388481296,0.2943306587284893,
 0.17749935034426434,0.4546772233243993,0.410787091724647,0.20095602862940154,
 0.49656710346071986,0.23013885080133933,0.8182172248965471,
 0.34785970431009794,0.08031880490882326,0.9122034539279256,
 0.2978348301370781,0.3392278274566546,0.35498249476132826,
 0.038775840247505244,0.5828926938744386,0.058521503676785436,
 0.7072674123724746,0.276596123040837,0.5778325939131681,0.9147874467035896,
 0.8231484643012298,0.08754457806540095,0.9719783112343654,0.6547664133529618,
 0.07888119687947293,0.5010737424512033,0.5746695811017489,
 0.046124690738481355,0.5849346259072157,0.6211312703126454,
 0.10223782124424974,0.885826272791977,0.2605418939882722,0.22352780803673922,
 0.5932667361114927,0.2476077285153927,0.6268007129561843,0.6327123882110097,
 0.6685332718437266,0.37904000381415537,0.8460488436599061,
 0.32471790544165924,0.83327625331278,0.19544731981080277,0.20218091749239642,
 0.13893591641176906,0.21898803664457378,0.29999789263237897,
 0.6308273454811794,0.3265219713447941,0.4844998792560713,0.20564606746488334,
 0.9078753263501271,0.36059804760102354,0.25324040527762504,
 0.9351005107259301,0.5506033708572406,0.37068468810768274,0.6558039388705453,
 0.2948506304260706,0.06600810000524726,0.7165915508059499,
 0.28997804931586657,0.8886190957170235,0.4866187108592991,
 0.22397866508114928,0.4898481276875112,0.17597479405078686,
 0.3296051872810065,0.3879042736725029,0.06923262940815456,
 0.03255870227471114,0.9565189016908365,0.7393510409555212,0.4772407409559142,
 0.17333461931802374,0.9916942679594175,0.6103290596574446,
 0.09894185009681022,0.47881975989711845,0.47121644654705286,
 0.34003834501817143,0.32080885127188685,0.005453904755777894,
 0.15512572718325834,0.20982509608123,0.17793410340815896,0.914009549677634,
 0.04400427217559966,0.5352341321236763,0.719962236122167,0.5411337453319529,
 0.619655757780215,0.9616051484505961,0.9855759810138569,0.9779764140782045,
 0.9008792395316618,0.46502986129966817,0.9690072304238533,
 0.27752917328179827,0.07950189808413688,0.2508011394811148,
 0.7720462400876708,0.06197097504047411,0.12130760883314373,
 0.9320583942238375,0.30734369388483374,0.24244302265877127,
 0.4337535081223258,0.6573736718802048,0.7494936481279235,0.6876438048492977,
 0.15476038458886454,0.2976717793613163,0.14771911856955566,
 0.5074925836007547,0.927536143642488,0.4455561952145318,0.1738568996731129,
 0.16253303576147426,0.6495643479123612,0.7935356682471607,0.5243746308411261,
 0.13345109811448208,0.06809883722329113,0.6669364105307705,
 0.7935130806448641,0.3960478762760714,0.06892055086269888,0.9023092612256834,
 0.2539266346554563,0.5229959532255822,0.5855389357639491,0.6653645530200045,
 0.8831186568707721,0.38401281991622105,0.3555297367612119,0.9220689902208163,
 0.5344991326624476,0.22306317116297025,0.7988313921126801,0.7825849667509708,
 0.45037050914916865,0.4240651720721913,0.6752132460396871,0.6678734952740493,
 0.3553793978276144,0.9623372968269994,0.5055571451085238,0.2461334196426388,
 0.03427020609911813,0.49649193767115285,0.12698959060570736,
 0.5218824003375978,0.7247890480890317,0.6898491363308757,0.3709854005717491,
 0.7420896660138443,0.9060424378768153,0.016627010564038924,
 0.4580550160983258,0.3678989226812671,0.0371416604004845,0.23736048926780462,
 0.724909617774703,0.19302106531727548,0.858080525294566,0.3512276422886842,
 0.780064244040352,0.13462107096946896,0.21893457534593175,0.3786406552264685,
 0.06392836957103998,0.7995869054200063,0.11617986586923634,
 0.9820511604269253,0.7504619938924015,0.7507596352090868,0.41093975015832784,
 0.060069256987494546,0.21896999622396496,0.6837909906471318,
 0.43694420213516005,0.19383445941769406,0.6262075396574551,
 0.12223826960915485,0.6249737358567247,0.5507527092452738,0.8135903329598836,
 0.41113647380876395,0.648346267193075,0.47742374340025795,
 0.48304920494528725,0.33540826278054436,0.10150638697416203,
 0.17083222189253355,0.8525367115278691,0.9744338405915106,0.2869333281751074,
 0.3171106015907702,0.24301831597388635,0.061889178994678606,
 0.33637967958413806,0.28965057722975973,0.976717678878531,0.3552719177065784,
 0.7461457515678088,0.5436107770545848,0.9296957519451462,0.4270759757684359,
 0.009895013491313942,0.624766649516744,0.2040454769400184,
 0.22798821757502452,0.11017838320995255,0.806498958763433,
 0.45137606428458754,0.5122955223121086,0.680958475414777,0.5938769655760996,
 0.5561452764845631,0.9152770541171013,0.4244097326484273,0.806639117493198,
 0.7721427216763814,0.8882104055763835,0.08611597507648028,
 0.24518787639063988,0.5089137993297933,0.7359423543001102,0.6327445664240834,
 0.7020132640029397,0.4550538200773594,0.07856652845806011,
 0.00720845811243187,0.7309571942513258,0.6911551204061955,0.8119149092350229,
 0.3641178751248493,0.13973415661557353,0.4751428821995377,0.2622024738327693,
 0.025375463484390348,0.7900152443391261,0.22154625648911042,
 0.3456235058607968,0.40292925823294246,0.37986425495553666,
 0.9044576636297387,0.3507894438183251,0.8465290030139532,0.19032519446054863,
 0.0021813236325680663,0.5405421347639892,0.33496160390990126,
 0.15056438178719522,0.24842354616332607,0.12074395821529549,
 0.7507815623203664,0.2691110221743035,0.8799104998090761,0.3648854650618426,
 0.9820743997777347,0.8000833253413596,0.30096032975106646,0.7240259843918645,
 0.7598906387282358,0.9149618061660655,0.39450406861508913,0.4400593409456677,
 0.46195464315698354,0.08189381661975104,0.8597192827927306,0.632715241875073,
 0.3097705540808713,0.3568488250236116,0.7071674573164172,0.2763216412253856,
 0.3931924795665862,0.34945825119042906,0.7072224916127516,
 0.11484578938552137,0.38940889102953846,0.8075959793101519,
 0.7776325483859982,0.4242721450683069,0.8628169125280563,0.9073317387432254,
 0.952823718793888,0.9347059129590167,0.07222360097626124,0.383624974052326,
 0.9390608043291182,0.8040210429051426,0.7009217041457035,0.7140467212917623,
 0.06426942785580936,0.3479622467188064,0.6785674462840406,
 0.28597279650904417,0.3302390025462476,0.8030125390931474,0.8149886709341496,
 0.29252728364592073,0.27978477943505076,0.7647641262735243,0.76434472172589,
 0.7970890572845446,0.05504722648024618,0.6288850394155775,
 0.33175957279105595,0.41969867122953475,0.28994558750034904,
 0.22057127392052334,0.7652313482856792,0.6708258539146473,0.60723602641091,
 0.030622040516988136,0.2345642923469511,0.5588121527417971,
 0.6452534936184101,0.1469239761808202,0.7454102985592368,0.9315731844532026,
 0.0451786109062412,0.858486136530861,0.9633535904896445,0.28165401771048926,
 0.6113297529435219,0.5311786246326495,0.9681218071275806,0.7906529203549747,
 0.07923916799599096,0.23487779847824353,0.4828677522385355,
 0.6152964583707163,0.47746830057435363,0.12271905519095316,
 0.22307961154446687,0.6369078227666598,0.43911366872189583,
 0.8261481905555907,0.5129589840299696,0.3659813783807856,0.6445340671531531,
 0.5417587999887019,0.47922551198134644,0.08714915555580038,0.510158665492367,
 0.4789159350285802,0.05036112334510523,0.2793010246425789,0.7755480190702838,
 0.5396744393369374,0.8672571187926408,0.43872080974375594,0.7802804652781674,
 0.1955065331192587,0.033960460591716424,0.9947117492671551,
 0.9098450017441646,0.9200116464705497,0.16632478141108642,0.1535331216968212,
 0.6195901185979023,0.7109450645359113,0.03794952165780685,0.7729595835359969,
 0.4818622988750455,0.9411446152754585,0.36881324676004346,
 0.05606636789292052,0.47105780324677093,0.7069102326071065,
 0.7825193883255892,0.9993746840156429,0.7767836673417152,0.4393630414324181,
 0.43131443895091093,0.7763340753799919,0.5643724178814259,0.6230814728526136,
 0.4192026077780703,0.3476741291435378,0.03104757532604918,
 0.30834049078074277,0.8852931368473232,0.8838892759975954,0.1081292788198448,
 0.7156507462706971,0.08827166769875505,0.6993413980956288,0.8516772552316537,
 0.2707317517682617,0.3756335889892479,0.12023288024803025,0.6110906736641074,
 0.6141593631309288,0.9505459868530376,0.8181689069509273,0.17645576216632253,
 0.8688702815459537,0.13395522629618084,0.02429394937388596,
 0.16729758726406918,0.8935313147418122,0.3288429409424247,0.5439293255275313,
 0.8407960008753189,0.6654472468550556,0.755669148722947,0.41354780050916407,
 0.8075698215186521,0.678684799232898,0.9012742156531729,0.7136348043709184,
 0.5458751816387241,0.6032763320086673,0.37215754527547007,0.1540767380158532,
 0.5888886175188552,0.5645691170712385,0.5872208468768331,0.2344709172206857,
 0.32757764051860594,0.9120481486855512,0.5921345139681443,0.9929173862618598,
 0.7437419602930275,0.8500441414086934,0.6360034628912554,0.26420481151945185,
 0.29988842600943455,0.7380799077483217,0.879925247617408,0.32433021841917054,
 0.7679631589585946,0.8963975600944238,0.5623195396894942,
 0.032248740635076656,0.006744396322268531,0.41972246063355956,
 0.5849912709317088,0.5163656449526864,0.6522206498907428,0.7271323940102683,
 0.12111077600535802,0.4204933589240816,0.4917177471759162,
 0.49690027492533395,0.06374620957038979,0.09715133832013478,
 0.6893127907219151,0.18637562860913903,0.7538779604027566,
 0.48636144242811297,0.34533242472604586,0.1278277064527027,
 0.1057018814797952,0.043093988502849956,0.450471850916602,
 0.09836623240174924,0.6446954043997828,0.42490832819842805,
 0.43927705388380545,0.524951567872794,0.5582188242891347,0.5901674703604972,
 0.5839955362736367,0.9543143868720012,0.0465685408754265,0.12545090203404408,
 0.16239999242466074,0.1459515580344104,0.5751280768393635,0.3576013809827878,
 0.9522406019771269,0.29808274817457603,0.8695391413663369,0.9566846479654432,
 0.5496367814533338,0.9265152843451728,0.9405163573184079,0.5963964727735878,
 0.8357945169841683,0.5357933098500853,0.012801445808285239,
 0.30615575376178805,0.4179889062259827,0.837304741996856,0.9423755749980464,
 0.32319657780669175,0.08340995848643862,0.32158858795501555,
 0.4866029220899851,0.7473222001434566,0.5369402441029463,0.9304871537423124,
 0.5060612326550659,0.2391067108055217,0.4131409853073582,0.3790664541294072,
 0.15616599030277412,0.031605988610939995,0.8011878385489408,
 0.5182772072751021,0.16428155274265333,0.949433165640937,0.4470700447221876,
 0.22029745313214366,0.7009242366151637,0.9994627331523371,0.5560767780484016,
 0.9851156411684718,0.2692103550683542,0.7765695895476914,0.8590663292249318,
 0.3774248795346492,0.6123411317565567,0.7232255364327415,0.14010099715991264,
 0.15289790169117734,0.780359753067145,0.5488828908275406,0.8380444455345009,
 0.8703835467391547,0.3428152507455392,0.6394993027342328,0.14089114531407831,
 0.2161297179655155,0.05218976950015053,0.43780407400658494,
 0.4502537443112664,0.5012749778124789,0.039868522636983705,
 0.3713661723040741,0.06962290501589807,0.8810325390212345,0.8571277063685632,
 0.44993360371329116,0.1357502759484165,0.2134713716252279,
 0.10511920879535774,0.9990702625772965,0.47434307699855094,
 0.7584181852560736,0.28757212567607937,0.21216427184870912,
 0.13588047083125687,0.30719553536594657,0.1131835073641293,
 0.8365142382829632,0.058487880799613756,0.3529341697014239,
 0.4074344633652788,0.7286984874712005,0.5243739387012615,0.7107138459215172,
 0.20409200948011463,0.6169629148917467,0.7576115938144373,0.7235410958437452,
 0.076049216653019,0.07177081393085405,0.021243695522327855,
 0.060416953180691624,0.3256935044581626,0.7165815437735582,
 0.29211211179038354,0.9050110072458172,0.5718559619715162,0.5056375085889367,
 0.7620036972124021,0.6095185229942723,0.15765533086422956,
 0.25654539622928185,0.08304107578036013,0.00600234167475544,
 0.5438196110080384,0.41211833127054576,0.15159927611689017,
 0.6329247395652595,0.08244393871496425,0.10900453454337489,
 0.9156561282422186,0.3523996628012209,0.4333506307717382,0.8790277580764789,
 0.9507155200473681,0.43691359101292504,0.7741172573311529,
 0.48077866059317453,0.5511631066267717,0.7729473970503628,0.9547487187420539,
 0.08143292600586616,0.16505680819453494,0.8014928269814717,
 0.5556546493983029,0.8523439168204704,0.8871157250762941,0.04475730248097032,
 0.39798896171891907,0.706173481790394,0.8202932406196327,0.32586622565462053,
 0.6508455167061208,0.08363628264462664,0.6325709107883725,
 0.35116886671590686,0.21305721951532675,0.333852036125852,0.0891233181498875,
 0.08402973454418539,0.9899995027388626,0.0067656433350372636,
 0.9520158924719696,0.42227836081789427,0.7266551682417443,
 0.48790237701644157,0.3601276584625728,0.5189958484508073,0.6507781299677649,
 0.8912512485865332,0.24100788488893277,0.288828233188406,0.0973953504403694,
 0.6560166809824333,0.5401195362441348,0.32263404892614767,0.6491689876527367,
 0.019335111343874112,0.2800133310509072,0.2642420114872359,
 0.6581323703542268,0.8168496462954986,0.5140506712344782,0.9484569792882013,
 0.06802689713987631,0.24523877629159707,0.009221288054748977,
 0.7091744434876788,0.7157031234580296,0.11227193719248652,0.2640990490547688,
 0.8182742691099202,0.9582729985179413,0.017814824887013248,
 0.26827582753312273,0.8717337882499855,0.8205733395356565,
 0.32520437724270357,0.4184134448694419,0.9892976480923782,0.7292310553681574,
 0.13200880956278793,0.7996655398666043,0.640248399363326,0.5271033036482546,
 0.2958274402613833,0.9531476252790828,0.4090212001300193,0.7101026709741614,
 0.4921430301484624,0.9546361137993367,0.1031593489306295,0.24431754793293137,
 0.007982667049156289,0.8253240206121464,0.34465356968432126,
 0.820155067034214,0.27254585442432977,0.15359348055952315,0.0088895202313104,
 0.7810306230440045,0.8109736919474111,0.19215557126813398,
 0.05772983097245599,0.9363780864057385,0.6681245960926824,
 0.06591663521079338,0.7031437472853672,0.7608859696930357,0.647039453185281,
 0.2611816640441874,0.2938761658753747,0.4836687263199685,0.3804646321022763,
 0.9794330296132344,0.10685249209744985,0.04272432745121402,
 0.8021770134355243,0.8792411102426284,0.5814060501595475,0.9828707776748885,
 0.9253695814800572,0.19815858394732822,0.42788382932054847,
 0.058622473583585855,0.12906990952960118,0.4329368268201418,
 0.5704901732569604,0.125608216998233,0.9837258533057672,0.9967730436804226,
 0.32835141458072525,0.38358597193317456,0.6039626025971921,0.53242828482508,
 0.39412439898419926,0.8595128187577779,0.11168674694420766,
 0.45365293520332417,0.9562818725360976,0.8684272798768088,0.6960637869380984,
 0.06346095360195225,0.8270217052059199,0.9850646468346713,0.2987002949323062,
 0.8952062022070496,0.6002703431530783,0.4888875447813036,0.3735620316882504,
 0.7373678847662255,0.8218947039895443,0.32897250348196605,0.6024002511829022,
 0.34031787020362525,0.5130513990949831,0.4965584579612953,0.3632635602570803,
 0.07778538057968332,0.057124492349015066,0.03546430587991889,
 0.9060546614278379,0.6857956360545574,0.4493224089000989,0.4079814947783965,
 0.807292361789644,0.6796897062592108,0.7486429690036394,0.05558604946833057,
 0.7872283093008915,0.7944841790972926,0.49890847702221275,
 0.12058401057336032,0.3294682141405728,0.9154333755766652,0.4299071143721297,
 0.7160210625314373,0.1354108504169962,0.5913247504200346,0.3442901052034908,
 0.5448119984741923,0.4652146311849703,0.25122643042796466,0.3785708565639009,
 0.8949657184168919,0.26577094011683755,0.4790793552666476,
 0.11480188435745253,0.8764414987927129,0.07034496090509201,
 0.6311861700342705,0.26582780072107115,0.3861627044177286,
 0.009132399763316812,0.17468808638210365,0.7986084118873205,
 0.9127316962787941,0.42154293684779154,0.5917042009552954,0.5709221884894029,
 0.8492638309086197,0.49523213402147814,0.21622469971629188,
 0.19746304829551609,0.4389746258029824,0.7693892502769161,0.5571081534624188,
 0.17140432735348854,0.7376078306905817,0.3266365428558706,0.2590769075488689,
 0.15548131197021342,0.062394515662994476,0.5090730345192527,
 0.8325630469727455,0.13660439297994909,0.8233811009236884,0.8179762041236338,
 0.11535821425645831,0.5785815645114383,0.794482838237454,0.934525387702327,
 0.2950927705162867,0.5022481912272984,0.03213045716447649,0.4163817422732451,
 0.2937292713231252,0.0887447622476526,0.4896715219878117,0.35394429559922813,
 0.9443090812151084,0.42665610954516786,0.4417315461947905,
 0.04036977317496726,0.4966013333064432,0.9388759516772992,0.9239850988560392,
 0.6762909265641868,0.3434700783684278,0.6603988627680523,0.8871383469945251,
 0.6530742729511811,0.9408433975851356,0.804083723786456,0.24651130083460193,
 0.43057156221924875,0.11343920356904591,0.5672127686831602,
 0.7219072625021656,0.471201526222899,0.24819442488920407,0.49453694821375827,
 0.27061868826861646,0.673900997094169,0.2068072712194592,0.40292539989890297,
 0.5928524249285657,0.8944882680063537,0.47153930558569035,0.7782031136603499,
 0.7764265700394183,0.962988229955524,0.6640805832673374,0.5539541278437853,
 0.5843054313101521,0.6967736631786134,0.3001232765437769,0.372095775898019,
 0.7232937298412468,0.9873083532321125,0.112054358006624,0.8458986713567955,
 0.7123857647356416,0.005187236250463245,0.5317340965610273,
 0.8847774286237949,0.19241422458603663,0.4850636112605765,0.7136165211143454,
 0.3439625062676942,0.9471171409855802,0.8763848271073937,0.04020307717855398,
 0.19059093033884966,0.7627485764662748,0.8248560330111033,
 0.18780888134002993,0.799378151368876,0.0902869106815376,0.15669724833028598,
 0.28966871416564666,0.1846337966021422,0.5836003090798786,
 0.20420518419906608,0.938355394200191,0.19914498103591194,0.8769313262401296,
 0.05930736521121005,0.7668248659380493,0.5277916372791163,0.8791290848989162,
 0.5450583779835768,0.7402952012623054,0.5235873020416233,0.4961952470428924,
 0.9049821306260185,0.20702390221765032,0.9649471195786756,0.8703778632329203,
 0.06438623407747857,0.4652097989599403,0.965928418609309,0.8192244009719521,
 0.8975846941155013,0.9019432914405767,0.8765123227605972,0.592412610890536,
 0.838637239297935,0.7169754569855139,0.27137425812219684,0.06195309735238286,
 0.26427959348703667,0.2630181845682843,0.4124154639877585,0.8700394258305275,
 0.9886456608201714,0.004099976859153531,0.1859650233862814,
 0.03156232185821173,0.20524631954414507,0.05260120165895332,
 0.13945144630369766,0.024284532662036784,0.012562487114280319,
 0.8032907558011856,0.9531960094873627,0.7958915792903685,0.1037577070781216,
 0.13412906642030753,0.14779590128710196,0.5637761494292242,
 0.3169547386217971,0.38793693111926586,0.19582727874482386,
 0.16905094617764704,0.9162900034516114,0.8569671419608287,
 0.14488087975789077,0.4691308896777442,0.6184071343962438,0.8846584161002193,
 0.28551265223167077,0.9731739886199137,0.41397780645589155,
 0.9188171776252412,0.6259186247172218,0.13159872164776143,0.0547982327436235,
 0.9967640382868982,0.6742506583705773,0.10156122633004694,
 0.09719126872598283,0.6595130678359866,0.6420812083812892,0.9829071617505104,
 0.18468074926236866,0.3543656858841451,0.5591951281754464,
 0.24573671998615787,0.5270992277932803,0.13033263837862308,
 0.16777959791807318,0.2649674493265639,0.5444922370676887,
 0.15268013298791372,0.2790238333318714,0.5399802772566001,0.4056070803960714,
 0.5863781064521123,0.6062397717752103,0.2897803457330381,0.3364846882143949,
 0.1002329866764412,0.4438162099593813,0.7093973202434813,0.5930622931861104,
 0.8603609623224155,0.08877879513802467,0.39277995669012133,
 0.6793887773469023,0.5900469052652448,0.8462176578271853,0.42077810829659723,
 0.6607925442397167,0.9147218675551965,0.9204149513601086,0.5383594622335077,
 0.0559131509422669,0.8605088241723611,0.6486898989451769,0.11995920598783294,
 0.37171955906508636,0.5604870444715555,0.1801033829671127,0.5284618221061407,
 0.7259310483906695,0.8971929454651376,0.31045711014849264,0.4048935819256282,
 0.30634691970660977,0.8558504877868381,0.8899914535960591,
 0.014710881693137168,0.46945028682899603,0.19090174875157495,
 0.7661489360601714,0.4350994347159698,0.7861396997059218,0.5785259582297797,
 0.8014614693126824,0.064824990665304,0.7133501071518978,0.6760533135732093,
 0.5162768610887299,0.7874861475703526,0.18256526807479545,
 0.039603313179121025,0.8091602280637449,0.8640754466340426,
 0.8347617695764078,0.9045644167498216,0.8086020826839092,0.5078830736529473,
 0.6666545987717399,0.0018615538472672721,0.3440522702380513,
 0.4787565057230283,0.20043012961582252,0.08328366876879456,
 0.5177993356028437,0.7152023554191991,0.1191481995761301,0.9885992358932424,
 0.19051386988420238,0.6127160015378948,0.44098251158297175,
 0.15018758343254945,0.9849396137518414,0.9766922278182509,0.7690434855688856,
 0.11041546093977173,0.9478425719732291,0.42229063007302836,
 0.028794251912650948,0.214961356446139,0.4308359898948123,0.4718488580626121,
 0.2530176680378198,0.8730479713016013,0.8432481673266129,0.02013792345276033,
 0.21746214129112973,0.894446220773473,0.8581084499683498,0.43424354835861556,
 0.4820330718125573,0.14718077868337986,0.6715502280973289,
 0.007171641514528093,0.5027549636239413,0.28638186935622434,
 0.5654381812784228,0.038033913405302044,0.0035723727317380494,
 0.7869400079353165,0.1257261507529338,0.18113470583404734,0.6822913357180748,
 0.39182480201909775,0.011800194617847648,0.4397739478216174,
 0.649113065636535,0.609347910216733,0.8845670636243,0.05143855110823192,
 0.25458655090630133,0.1748643170390446,0.040188077638473096,
 0.4739790587425575,0.5987651830166774,0.5998515402820555,0.4967301370933852,
 0.1458999412270674,0.14715539254114018,0.16169239389556678,
 0.26854179316236615,0.38603739661272674,0.5793321668819785,
 0.37663767518441527,0.6528419101559552,0.897766771549427,0.983191069238026,
 0.22240617188565592,0.4187560774530188,0.7657348521669374,0.8847869694892647,
 0.9378662934922566,0.4314405258232471,0.921074706969788,0.9034591483161534,
 0.14767099847121368,0.6586431120846931,0.769812697145955,0.9745231322233794,
 0.5477033169918294,0.7962524155924192,0.11528019944035506,0.8685308667731334,
 0.13576205923423734,0.0999642804455313,0.7209213506039807,0.936678723991523,
 0.8251270003950248,0.5709087142913767,0.48594522774116766,0.7466763388948208,
 0.35294573552951425,0.5697552323575598,0.33745038714394027,
 0.5656352811969876,0.09494064320059281,0.9918295579477625,0.6696605466483158,
 0.7623213629069446,0.7005046539568723,0.20732225225550738,0.2142813317179687,
 0.5502730257547402,0.3361963990212561,0.49820385237635534,0.9690065437009797,
 0.8660647057172608,0.5645935064332277,0.002165336915132765,0.72246727226803,
 0.043723131003911675,0.6875330643968015,0.6116962847902073,
 0.45263233796224467,0.6262072766643166,0.22629965483935777,
 0.12506457934312087,0.46027617052286485,0.6289977686034012,
 0.06371853002693928,0.8136078088406343,0.21483899045066224,
 0.9412086811930512,0.25565968198238254,0.7221579519777819,0.2118251211300849,
 0.8827277897892692,0.461700407948479,0.7232174919128462,0.32586325530264215,
 0.05240757734765733,0.8466223314395578,0.31405679556580435,
 0.6344729034543212,0.7628413105015889,0.4296019895590345,0.09834456285573256,
 0.09328584262846595,0.06332051173619269,0.9383798630740805,
 0.7121924988705642,0.8676694838243239,0.9419422097486007,0.5638828847901165,
 0.23223251715602788,0.38931207391800604,0.008507674993399217,
 0.4851092960026129,0.13052867925736866,0.2113679921964441,
 0.07704305972589653,0.9777362684086778,0.7563500351996321,
 0.49872480130103836,0.2689146484950725,0.6455583066428853,0.2857167463865904,
 0.1546844260804615,0.24158975749006528,0.3462116093284502,0.8331959506205037,
 0.7587947527505818,0.34361798140766286,0.34169996979269046,
 0.11056992829059265,0.14571769414686087,0.05111007795461753,
 0.13801030335207742,0.0581998212778958,0.8703564184740535,0.0685084323554187,
 0.5470651615468802,0.11143324997713078,0.13335790011338922,
 0.3395881505383166,0.8371725511570326,0.47347553499330575,
 0.43220592217848774,0.7942845590104218,0.838706718421565,0.12286876608607777,
 0.20262136994572122,0.13236268177745425,0.0764244944332001,0.256836422557177,
 0.24356921929540887,0.6003677819972731,0.7104490672932022,0.5966557627522048,
 0.060894219548086936,0.7437473541013793,0.4805108677531552,
 0.7965812327763113,0.37322525201405776,0.9985682942621796,0.2989161628400241,
 0.9877716944588423,0.09100026675175954,0.028950389845215696,
 0.5134142087559513,0.6692528311206889,0.15953608891378646,
 0.22192123495779048,0.5145827728874736,0.3359391319239502,0.8754871268237479,
 0.7878574272708735,0.3762569352209586,0.1274698131803622,0.625272944253318,
 0.2130054722221949,0.22149915921909147,0.9383831791627119,0.2769153878378233,
 0.47544452987662456,0.3210191340635855,0.16420489086201662,
 0.7707969639177971,0.014949137998099604,0.25122995069321075,
 0.3149286301191513,0.3182706579507093,0.03608270576049377,
 0.30916938248079207,0.5412181861668532,0.350639343048611,0.8135598562089923,
 0.3907020584759159,0.026512862509512014,0.10763804774564767,
 0.21365271678213626,0.6419184039711999,0.6482415551197441,0.9595574117654306,
 0.18591429685121508,0.5044113908587745,0.08635752345915915,
 0.19820264151552702,0.6965847125133194,0.16439419137685285,
 0.5534602873135281,0.6088838639426714,0.2976505703422927,0.43363691035030894,
 0.0293845006501412,0.36941666241237026,0.007922189553609349,
 0.7406155576309913,0.8571422141297066,0.39584360761072923,0.2939785666081711,
 0.7610724505867131,0.9450764575294325,0.8228746379973256,0.5123249546297807,
 0.34726103088975935,0.8763732638299865,0.4352040772778436,0.7765787256182753,
 0.4526689280396723,0.0067015301308218644,0.3346403515135339,
 0.16131729367854453,0.8447151009048224,0.4931296188189044,0.8220868552532365,
 0.3035008982322648,0.9707746275708702,0.3164376617187137,0.1165748995446535,
 0.9107263387516729,0.6936749571562795,0.7061471564692657,0.17257268210607601,
 0.20218017014887324,0.21169041859978754,0.35955350202847824,
 0.03014336250771077,0.8059128337602274,0.6825686101372348,0.86775334256947,
 0.5483230078915569,0.5127499394107604,0.07765715255435568,0.4036376675865413,
 0.18526366845457898,0.22675529112201387,0.48225635324500604,
 0.8578900456552621,0.3298898554921498,0.32055928220875274,
 0.29209334631508954,0.694406583048812,0.8493788655723438,0.6996573527304715,
 0.9243724835477586,0.1614846422914853,0.6832450008624344,0.7457476500684062,
 0.507419143977514,0.18129705963352027,0.11998780325166725,0.7231372047810369,
 0.35894164446196763,0.002296443507198709,0.7088896030913948,
 0.6963472389461035,0.6107139859074706,0.5793302574077084,0.2575049887858354,
 0.7063240236094865,0.5075284582673786,0.8651453693862797,0.5542974821940916,
 0.49381262154938355,0.834898938360772,0.9334307914410257,0.9685909978057566,
 0.15416497453387745,0.03911506405192511,0.42277673120122516,
 0.8736057393714972,0.9611462165122799,0.3374618055851645,0.5831178280241557,
 0.6912224243438618,0.8952799167382348,0.6843827633078836,0.6209808590974437,
 0.43236343666198096,0.40926592062065814,0.21236066908843965,
 0.2977207956459472,0.5741345667031855,0.20990401406388304,0.2535164436665983,
 0.5167350566334278,0.36648833716710005,0.5223746966969613,0.8898264609883921,
 0.8129455259006779,0.2685046365258894,0.873703434317138,0.520195307750934,
 0.3072581877544708,0.9673702745088375,0.8624540904632345,0.4613933940326578,
 0.3571944593021007,0.4716867633478983,0.5171071500561149,0.9114565620061285,
 0.5188266099047608,0.9994296676495688,0.6094341778776012,0.364025554147934,
 0.9227069140615827,0.27041809844363107,0.7374607896888624,0.8070921740529204,
 0.7837402803632812,0.09862168135054339,0.984000816829936,0.5906738830022407,
 0.6778721272795025,0.4025797626844676,0.006198194593821094,0.555725576011052,
 0.48198323491500306,0.53553793898208,0.14206288744555395,0.7567307925604739,
 0.6039640859785302,0.0812604444183318,0.9789081753174922,0.32777311557498334,
 0.4351896874307396,0.6627828619172217,0.7899644703352757,0.29480049969569344,
 0.976997213347785,0.5528933064768434,0.05416393834007127,0.49948330671668806,
 0.060095999876967454,0.11288461168259545,0.9387855168289089,
 0.6274717477566633,0.9630304102195945,0.11621229309477499,0.8024217364471019,
 0.4752458160162659,0.7481439874089316,0.8545816371579562,0.46992515102838306,
 0.7732625025807814,0.8098389368172184,0.7336833240643008,0.981978512722016,
 0.9422114415459341,0.2741556939969336,0.3020223746636357,0.0684638057042064,
 0.3622289971488277,0.19268625656846095,0.32280133865149263,
 0.47659565012709737,0.953156666453913,0.7873000267457456,0.3940312199002467,
 0.9998315202387011,0.327203594620612,0.8938040264761579,0.36979749879434365,
 0.5545209796343034,0.9216726464673999,0.5498509766485377,
 0.011013240173868066,0.5394940054703631,0.9596471227456513,
 0.06379487629835934,0.7886099782779457,0.4459074734185222,
 0.39458379305325886,0.9513333548221773,0.33544110840767405,
 0.35714530273790324,0.6108560632069947,0.19049626678987053,
 0.5524497294010258,0.25616380586278686,0.7376396818740578,0.594536159921873,
 0.29887113139502464,0.40085499844013317,0.2227821176858632,
 0.6736726498014634,0.7306229462724741,0.9730141500714362,0.06103636041374183,
 0.21111893330049702,0.32753961382053365,0.4535602011357289,
 0.18881409550394412,0.340712495322394,0.9840592764774713,0.8620798419626547,
 0.3284515367841053,0.5107273933432801,0.1386651329078873,0.577442252743517,
 0.8301223484225188,0.7413781704006196,0.7454586182735882,0.398034317718129,
 0.8251442606968367,0.7841509348556456,0.7955242789829071,0.49725162713995474,
 0.06559227452177618,0.2541448272886797,0.5714133313182466,0.418143325387449,
 0.6286014417522231,0.8019099638982631,0.8298273286711861,0.16675781946617652,
 0.36477845401294096,0.022878601442852364,0.19145916609186509,
 0.09294271011253363,0.9441179818800011,0.676389462851395,0.10947200501842735,
 0.9270810607307701,0.8465885023904149,0.538936914834964,0.7613767907009359,
 0.9725998204512731,0.9889652296666539,0.14736315548841472,0.7429281713085845,
 0.7894174975212271,0.1939940316990686,0.7132644147307575,0.6516473070784927,
 0.514907291224913,0.2707840022890283,0.7291193105420014,0.4044256861380918,
 0.9664050835038883,0.7576691940822647,0.6959071937563409,0.09551244925857305,
 0.7643724369058764,0.009108500624327776,0.9033828457762485,0.792678194066352,
 0.21213348545067867,0.015504844596428313,0.18311528441789293,
 0.3091330722364136,0.8946820022410384,0.3509781051502817,0.8406004175737112,
 0.1224371723602371,0.9764449405164217,0.9578939439652268,0.34503695207894525,
 0.8546428671463158,0.6647593578330735,0.9873241147281708,0.24420678308841548,
 0.7551817718891806,0.5287008691498187,0.7729901201671854,0.25763725239380175,
 0.8860372852102134,0.17586643012866765,0.41334362618495035,
 0.35080013200970117,0.46748804535509847,0.5022231313414873,
 0.9104116047231079,0.39230437792084916,0.2871979985892038,
 0.46645922344441104,0.13889552049428477,0.9950977495301788,
 0.9988468501886035,0.43324285481260594,0.3498269617239388,0.9316231366215391,
 0.4722346306824783,0.0547861020561633,0.1368659265631289,0.57738172698785,
 0.38113286392145196,0.7420153340257414,0.6902583738603343,0.6436365499526965,
 0.8507867159645434,0.9859226852149255,0.9705816931328172,0.4734516073645858,
 0.9073643625852615,0.2478463084069923,0.9573090927649006,0.7569801358199676,
 0.9301929396167812,0.6715300759029409,0.9994550938904015,0.5399557789240491,
 0.3409183771930133,0.730732704852851,0.990959748499242,0.08500270998783832,
 0.6362874379406139,0.6044550745532298,0.5555117097392006,0.9510636774610434,
 0.32387906122207344,0.32565504358145614,0.21253624052568565,
 0.9335415715432247,0.5953159822892918,0.8741671680328646,0.8872109488312141,
 0.22740609811525592,0.9252923117947542,0.05993785253615369,
 0.9623025934314032,0.8979522503356594,0.817949728449394,0.237968388144421,
 0.12188711850376799,0.6446381690330968,0.17967667369481,0.9453997814596939,
 0.25410498902867873,0.29104629608333155,0.8582505459027624,
 0.7520241938958727,0.25907146747732557,0.25903272974875313,
 0.49959802405365283,0.011448034857083789,0.6221813924383344,
 0.6234194265468669,0.7276913831189528,0.31288008679295687,0.9277732248061737,
 0.016678753041249106,0.056138330901662425,0.3629704074952125,
 0.23018645179556663,0.01674507001832648,0.08239431896112659,
 0.9926845002909034,0.3980613776456817,0.5350542100376042,0.4044981419257214,
 0.06731878894114851,0.06543730678180959,0.8056389647480434,
 0.9285604569426743,0.6212972753703803,0.9169576211894066,0.745031034729985,
 0.47354417500133106,0.8044686631494724,0.08920687118637216,
 0.11984477941983751,0.7035773873507711,0.7164249525303662,0.6031636273213028,
 0.5941197311597224,0.7537942772326625,0.6300331048480031,0.6735821680160547,
 0.21921780425867587,0.8140571275699583,0.9209447634837488,
 0.17362128739261407,0.4052314720560496,0.6458238814494446,
 0.49377459836784476,0.6770990985719952,0.9787563785986193,
 0.16384848902045246,0.40553943117661284,0.27453453485783363,
 0.38471560217757417,0.11820832034198459,0.6520621790071337,
 0.7938083769483432,0.22642452567631555,0.05724455893512648,
 0.4104632164997202,0.19868622881699127,0.7927282311074286,
 0.09297152145755305,0.4913370681121523,0.7934562390470079,0.8094934877916663,
 0.8829883179865504,0.22809359012217412,0.02590690365836923,
 0.030652933589020548,0.8127792205151739,0.08389240592266478,
 0.40889725080998596,0.31101736552416215,0.28337661444788587,
 0.14860170464696654,0.007793914617346354,0.054698206868318056,
 0.94563755176102,0.6686466544978713,0.050540064692870756,0.796748566242732,
 0.42560901415451957,0.049739486062717275,0.6985479156557978,
 0.6609050381863935,0.5189818484591181,0.18375426062374411,
 0.42112609996229755,0.5599638756963651,0.14742401844078756,
 0.12517934609903647,0.6462826336857929,0.8503340928492419,0.6858730726369616,
 0.7951766851391358,0.433202764029349,0.7086587186910207,0.5944911194411483,
 0.8741221048432373,0.7406885548019592,0.19916575025319083,0.6813669812202722,
 0.3705239684628293,0.3936274255126643,0.9432719035327014,0.42054780218246834,
 0.39400295873007773,0.1209137618823598,0.4049146435903206,
 0.31392803377602363,0.6565559188323296,0.1477041215800029,
 0.37647977661221055,0.6126073414623475,0.3128838616457281,
 0.13354321761279153,0.2608759968628632,0.8434065586059981,0.9268949330254508,
 0.7523293717109242,0.5075376768194798,0.17669696116009992,0.8070747194318911,
 0.4822406514263162,0.3956882268432561,0.30297906797933505,
 0.28537473150027703,0.12824437165462577,0.4474481238078175,
 0.5846303063391022,0.34040878459822954,0.1743515296661906,0.5419877984450934,
 0.23921903896911223,0.4482261226217772,0.033105742583717124,
 0.5527526188794143,0.10161320987973554,0.057031091088490915,
 0.6577907061219276,0.9237269572616572,0.2679782100308362,0.45153723652797073,
 0.13647850517469418,0.37993987859121303,0.2867549798246275,
 0.9617870745927765,0.3302120732528592,0.5947971144502484,0.6532580678987057,
 0.390832271638508,0.03634962153259713,0.35191516955600555,0.8742742933058029,
 0.9420049742127163,0.5700233703682673,0.07396543158832503,0.4991747772611953,
 0.5507064853395713,0.5050103504575565,0.051241282956831946,
 0.44123018661375646,0.479472427268248,0.1412173247345444,0.7612370443238753,
 0.4124226150335988,0.7208605641670663,0.3308390906928882,0.3082186231190579,
 0.9742592415728197,0.6847052106912164,0.09850559121202973,
 0.05344157141710126,0.22596956316694117,0.08994888350868724,
 0.19265061736104871,0.8636546766632561,0.5073029055084667,
 0.01377012350194362,0.8612016461134323,0.7437247580803977,0.9449797699672662,
 0.3392763640879788,0.9665816546793174,0.09441185472300184,0.9616300129019026,
 0.4515351350777531,0.9232535923187489,0.7867079559802154,0.28647548443255566,
 0.3727863970898103,0.6930079285276983,0.691195328841851,0.864785919843165,
 0.7972601665688723,0.39791587396216344,0.8286596644787739,0.3894178947563981,
 0.34634086792910057,0.7908839482015361,0.8584068577874969,0.7242843749001634,
 0.9666468464326974,0.5130784754911111,0.06664892143144707,0.4836595377612358,
 0.3203866073343329,0.10557517154571061,0.05123506457712823,
 0.42230650817127957,0.12530149064554785,0.4729910417665346,
 0.30455005155037573,0.21239669502897218,0.7104999241774931,
 0.2777268904227105,0.8038843872750595,0.7589400957572414,0.7204859211319043,
 0.7531936031549761,0.6344686112514908,0.5095823687744883,
 0.024554756521906418,0.9808693059265454,0.8387235301354931,
 0.8016218938313943,0.9040900987526307,0.12884117733776845,
 0.21798304293769477,0.9113359947687815,0.4200499341778753,
 0.40524233004971755,0.21839223455470858,0.6099909835681857,
 0.39732116788758276,0.358151210853203,0.646773032387754,0.24623616517421265,
 0.6364711499056764,0.02705712474968469,0.2574084394248689,0.9534875646056005,
 0.5070876591726172,0.9543858425022145,0.745917279396996,0.7693776427857021,
 0.688401035358095,0.9063208791580853,0.6418692563781699,0.5719966496525845,
 0.630423917852748,0.8500071620236895,0.3389307173428886,0.5307646525683019,
 0.7675623047282943,0.5383190815762349,0.6423985675259996,0.8639043528166871,
 0.2688137454117192,0.6311178215036519,0.8351338831234393,
 0.017292424812461404,0.14112445347429536,0.30749996008564207,
 0.7011582308818896,0.6643995406837109,0.8837980535509214,0.6839194872889618,
 0.8396806268822292,0.16427438488376145,0.11280447015362749,
 0.16352676038077274,0.781632153495668,0.7969226622768996,0.3744409476545413,
 0.6627777978166595,0.5644139945529496,0.6505224454236317,0.1317285369061696,
 0.5627570042157573,0.13161769363949005,0.38166193901963585,
 0.9176730423133908,0.21214974492708916,0.5618169397070524,0.9286452398970583,
 0.40514043367296193,0.9095460639793786,0.725469777016041,0.27992840664148544,
 0.6669128130468783,0.93969364847915,0.95473868255181,0.6137833909817278,
 0.5557923775754355,0.44190082840540434,0.9322380439467401,0.2701020708862538,
 0.7488760783951369,0.5753281951416976,0.7738043047016973,0.3031570219066452,
 0.5960158088906168,0.2629195245753718,0.02677194717064424,
 0.006418512456901726,0.010321011006578895,0.8261269520260309,
 0.5380367688312092,0.22066673504329293,0.34568667968923406,
 0.4029816324071396,0.9172510304273525,0.9217013014909707,0.4274221274054103,
 0.3680772519000195,0.6342447272059499,0.5431258509568183,0.906233378168672,
 0.1588534271251556,0.6970603315667121,0.16017198041661773,
 0.18865429915969312,0.9901716235876752,0.675043937683433,0.561753561458416,
 0.7864021659374826,0.4716260308607161,0.8758207159863638,0.45693047955474253,
 0.49790032634327674,0.41852616785029784,0.7811494352157473,
 0.8495807256230133,0.22232967299933237,0.516617326018974,0.8269251752983916,
 0.05661472406932433,0.7019752736059696,0.28268941900389155,
 0.5244334621253454,0.7292668429321295,0.6343344843010834,0.6078231484680487,
 0.6371081402424841,0.10132067223830954,0.14521309315410003,
 0.7656096186794995,0.7444995455383526,0.3858047085831938,0.37823646407651823,
 0.0845504080137448,0.8326368759362617,0.5059067223545326,0.7182468303735201,
 0.41591619187120954,0.8381124224127999,0.6758984067326379,0.8071088004172668,
 0.7427764942156793,0.6242234595221327,0.5981128417425875,0.465415867677483,
 0.7855342629779846,0.7503504262457558,0.1419571146977514,0.6863406517883205,
 0.4131167142176242,0.1545881386021547,0.5915839697824294,0.9462408780317044,
 0.45039585621265454,0.97428504643526,0.6479595982835205,0.6861072128962039,
 0.45717686957150483,0.9223153474568869,0.6048384303905562,0.8371232912081392,
 0.05660121739058288,0.9180265547615435,0.575320337994164,0.7930588172815614,
 0.45623997542300065,0.8956172902938524,0.8362795897327057,
 0.10858269564710765,0.07356414777734277,0.37656478423665407,
 0.36774726318867623,0.9968627109063128,0.6312459343449472,
 0.12529376196249475,0.13552394264204592,0.020747702276345037,
 0.768505587315257,0.5370108234940256,0.5424460714503669,0.9507405374581577,
 0.49601654021175545,0.428626337989772,0.8664148817696423,0.40125440929020306,
 0.6897875312628711,0.12215078190287876,0.31017071231990734,
 0.9165780691953527,0.5055102652547759,0.8247692006213061,0.8764875234711111,
 0.22540738797045212,0.024936046891553598,0.337768388632665,
 0.1534641790868826,0.40175338395800286,0.9600169303747296,0.983762205346444,
 0.6844081326244054,0.8740617080272248,0.3766416073899823,0.5883374324852647,
 0.592284296147808,0.3851180930718263,0.8927691084074119,0.30266669811990465,
 0.3649428766245111,0.35825397401124914,0.026913130960312448,
 0.01346115568912365,0.48461461588847476,0.5288147022764539,
 0.05684395416293331,0.906201616417485,0.7992921032517817,0.279214655637518,
 0.5415466566409103,0.4926906073154479,0.0014256020242058298,
 0.4673988207336889,0.19493184175965295,0.30733017822705844,
 0.5028800578981221,0.6809977214973137,0.7367955042467285,0.6349950302110935,
 0.21972376970977447,0.5362790945844887,0.013416286151481827,
 0.46190106640579043,0.25195350875172196,0.5721213100021916,
 0.8107725575065312,0.789808862104533,0.6672402239259084,0.8424908289621923,
 0.812492228910052,0.6302770030192191,0.08749255562646108,0.49888776749613417,
 0.5999518311590085,0.16530758670510692,0.45111639018403893,
 0.8543845138283304,0.21812495313856506,0.9317917372819587,
 0.011178646568977157,0.7757960453710188,0.5786385114624429,
 0.8355886466178211,0.9073490189582482,0.4353274253025189,0.25991650675378075,
 0.9861775520008846,0.5909035904555073,0.6433670123170797,0.9895048707182038,
 0.210131208863241,0.29570365134228516,0.6890326673159927,0.5514022422159018,
 0.3828106303358281,0.6695333624846557,0.291523797949336,0.07089552425800216,
 0.6424606543185019,0.9209280787071231,0.7706081406517076,0.632358774328802,
 0.4626592418048401,0.475043956554963,0.3908542275504211,0.09996755997153084,
 0.24276573534629964,0.3282767948720161,0.41637122387789904,0.67570515926807,
 0.6190576895464481,0.744664077104509,0.19389651831924792,0.6817921821808458,
 0.273689756763441,0.35440950321048614,0.4981000717692241,0.5648399950433184,
 0.494110044506638,0.46144620808109804,0.706446145401181,0.2389256814433991,
 0.043814847631005516,0.3469418991503761,0.9951052921052466,
 0.5585342408172871,0.732616431996742,0.5862978333066309,0.30411395736265945,
 0.7035321069478678,0.19575425535566637,0.3905842891888277,0.5341412623445988,
 0.2784477815199554,0.6144076934072552,0.2464095911022699,0.6555531416341894,
 0.2708870552273155,0.039034109366125724,0.4747266027505227,
 0.6144416013700333,0.8536696768784884,0.027969788224355474,
 0.009608919871345822,0.2177267819492681,0.6070155278944869,
 0.2624617232068748,0.5360407744629379,0.7902181036356707,0.1390615690650987,
 0.42999557093410956,0.7962469430104301,0.42787650603229876,
 0.2504655072710469,0.21905850167971908,0.6345904392754621,0.3546944873747987,
 0.3268776572775832,0.6204640316726013,0.8926779144436618,0.6719704015491487,
 0.7820428992684519,0.45388740246444437,0.6145647433145525,
 0.16155607272014494,0.19687363065169639,0.507842671742704,
 0.016500421703063872,0.5322874872422121,0.1946754944127651,0.143655505999832,
 0.44665947485015267,0.9048568909670318,0.11301947705300996,
 0.3705065976103533,0.8383345241311541,0.9551373786607686,0.6942470839111072,
 0.4365678457268056,0.47504830834818623,0.38539995267909966,
 0.11006406437783134,0.7278681423125761,0.6262532666033142,
 0.30497410418371673,0.6697806969229257,0.5380265804848374,0.4411283211601603,
 0.9645674630323701,0.8275141022250255,0.8284229387275202,0.7201215872243072,
 0.6784001356650546,0.7216895076108913,0.11998926087584627,0.2983234159178324,
 0.7772399180817726,0.04750226998771301,0.21002486291392075,
 0.8230977106822757,0.4827184702382028,0.4167429552230679,0.08305031206485214,
 0.49122561899521633,0.266784586050948,0.8359798006071945,0.7176117984903335,
 0.8472151155889179,0.3133423128001379,0.7565365975261343,0.10038787071587063,
 0.42666071234095604,0.7433673270218757,0.9839299465036425,
 0.16104246815288947,0.8050349220850166,0.8375982838442059,0.976982085630268,
 0.6698613370813004,0.169676839039134,0.7062922925376756,0.536186217780051,
 0.8935621543973375,0.7410198421564168,0.5146511819860436,0.24300252736659067,
 0.6893742223637745,0.0848344328338716,0.7947621623630041,0.16608974091188267,
 0.30841133569990076,0.11405404837960009,0.5989789893198427,
 0.13763777840643787,0.700899654243378,0.157201335113338,0.5810876106159427,
 0.036509123641865315,0.34648591657927585,0.661706988592676,
 0.7082333418500458,0.35012802106404817,0.45545778449245666,0.8903522368244,
 0.9324964418240617,0.5797209783139826,0.9940371290744686,0.8378452392509748,
 0.015400088077398677,0.29187234037295473,0.3648226290696913,
 0.9296303461760321,0.9089646245125349,0.9326465539798058,0.33194372480779477,
 0.1341757508454634,0.3270416726822699,0.37754154406238793,0.9173201852744636,
 0.3650269484894264,0.18579118657518245,0.3961104601165606,0.6000053799696897,
 0.21988179052522538,0.044971748857893346,0.05790601793036343,
 0.36773587064759816,0.26007996808836986,0.36393664660754255,
 0.9849071697487146,0.15176875706582083,0.7978049064135898,0.9288474607930984,
 0.004300196277431612,0.02679298898035154,0.08233082731475694,
 0.015327771313472516,0.26124896710424816,0.7434005380413533,
 0.037284414006212296,0.06367800602118656,0.5951950120874162,
 0.8232291042506605,0.1974789592717302,0.23204419304691648,
 0.30771784413880965,0.7255624425998911,0.9794883160075624,0.624830283701491,
 0.7263885020618598,0.6627154167942416,0.7835491702060234,0.08239338544507979,
 0.5740550845664274,0.15581097887624118,0.08132787579893941,
 0.41865957019286426,0.3527393385608759,0.9007713266047852,0.3684837942974655,
 0.782368404197506,0.7514847398973248,0.0886937501213142,0.2625743569118417,
 0.4567025209795874,0.025293791626304518,0.3830593757141052,
 0.23756211701600105,0.6540468648974542,0.0641240060116558,
 0.42578630325325584,0.7136842401377931,0.6092838776765102,
 0.19629980488236043,0.7804998997490764,0.03277446105247972,
 0.7782542588151136,0.6734191817021596,0.5283323234058785,0.3184893689548911,
 0.6474000261134757,0.8916054839632056,0.6172086755507388,0.12473025687015571,
 0.2911408746569303,0.8680739369876106,0.2562789862598256,0.7322956124275953,
 0.17058058936741383,0.9644634977181213,0.5864801609495478,0.6467179027871062,
 0.08708433528214632,0.02339907045416678,0.27979661622468266,
 0.5906583964333506,0.5501067293187054,0.21578614598085277,0.5385440851912249,
 0.6055252857557902,0.07316898469717781,0.7892872977090954,0.8141778946234444,
 0.525524597931837,0.6364928300394972,0.9398858395808156,0.6654924058882127,
 0.43177356846538395,0.16269730993103482,0.6584651507822646,
 0.060681385807463295,0.8772506314308408,0.15870506443671695,
 0.971419888695826,0.6293832670910676,0.35112177437256986,0.8686416730782295,
 0.612846316662044,0.6817257711162765,0.09765270419077032,0.8877541981384031,
 0.801947685778215,0.15295826814044822,0.5142776556350218,0.7177729256971928,
 0.681589803382186,0.6342663407826283,0.16552672233475063,0.1900828013818734,
 0.2796791624141325,0.11492967841286306,0.25559763165027394,0.570802355058991,
 0.2333021560121632,0.5038370067416303,0.6377354017604966,0.03542872233883232,
 0.8484323750301697,0.07944479865674581,0.46037557129630335,
 0.0366169122358565,0.652866265201821,0.06861803677628531,0.4581763320720974,
 0.15149658968234236,0.9142190406698557,0.32866231723623063,
 0.8558491731009709,0.24108062461724367,0.412430015367083,0.7958219373171315,
 0.08550908693263604,0.9333504081080913,0.05494391722605352,
 0.1324163846116393,0.35576773190479793,0.6694809103360219,0.7051267037547724,
 0.261509594705296,0.9584394765721902,0.3974664712543753,0.8862623070557047,
 0.7904223404687646,0.7805612957832927,0.10468569789668858,0.1291395308154084,
 0.5824281084307235,0.31063804838600784,0.8470204644935116,0.8225683726497799,
 0.2103800374177911,0.008978832568018125,0.32132867384807806,
 0.21085552695162413,0.8108014698371403,0.2174973495926188,
 0.25046478323765464,0.834506952909709,0.9258055710199461,0.8589148563597613,
 0.5355228642185048,0.5365890538229982,0.6675095418764615,0.0945589507381629,
 0.35525226888481176,0.5975409746017826,0.10413069413933596,
 0.6795962960540314,0.00435194592889232,0.19013506093066335,
 0.1493114785409997,0.2109303247211899,0.7641270686916319,0.5078061920666731,
 0.8035618091120911,0.09995497907562345,0.9686736409865153,0.5173535519628926,
 0.31691488045375593,0.8027857284694889,0.3949855414719342,0.669612582481875,
 0.7568474575233646,0.7470586428049747,0.2618669208368892,0.8713597637294332,
 0.1774535800418875,0.9417965954979937,0.31859506018408135,
 0.08793104380182482,0.5578152109275539,0.9469960241204916,
 0.22161974599940704,0.31015725986222353,0.5178960402442272,
 0.6935435565085454,0.34343269106559327,0.728183455731142,0.6132517031415643,
 0.7480244229262517,0.6172676254132515,0.08487632226288355,0.3011572987292688,
 0.29727599963726137,0.7726487060553355,0.7655643605100874,0.397312074257824,
 0.8671125117750009,0.6012296235958061,0.8268329093545597,0.7646120694546364,
 0.4143287581936701,0.27443258834415574,0.36779114746100017,0.799768092986913,
 0.45069637883125857,0.5826777562434787,0.844793737428353,
 0.039816893055856006,0.45902668466375274,0.2472085878137541,
 0.9327305956542326,0.6922417039479916,0.4577119855094325,0.07376716040290077,
 0.0719627545319459,0.15835395462002433,0.38448170532027326,
 0.6842443980957363,0.8016269637534088,0.6375333235684383,0.43462708331621724,
 0.12179580854594318,0.07130548615110088,0.3041129972723261,
 0.3259704959982702,0.6269369042399481,0.04438731187639422,0.345096742913528,
 0.31752079832855995,9.958419318495393e-4,0.14587872615121977,
 0.8175348576800404,0.7010279970048242,0.8602770969378623,0.5164035002265693,
 0.46978130530024276,0.43239128419058803,0.07881181289367645,
 0.32066640776493915,0.7808989807778481,0.11830734538023835,
 0.19515658270257885,0.4806161535389245,0.9685532149616893,
 0.12454015470312907,0.9644265507874276,0.20814185230232296,
 0.15288204080862466,0.519166859715207,0.9652362757261538,0.13474497210887915,
 0.2534378865589776,0.5771305005104408,0.936594516199357,0.7887912623561723,
 0.9724514285542589,0.5182764758906863,0.3725756659846323,0.12450203771280821,
 0.3316572864865289,0.10823549896067519,0.7526335416602241,
 0.36325163201488264,0.35832640977390906,0.0010251968545726964,
 0.6631444305941532,0.5284896837586419,0.5642423972914785,0.48002079265935216,
 0.9588809416553122,0.41669023675736305,0.9147048683396882,0.5728817342919746,
 0.5733763437745754,0.7929323909947295,0.007347547819041678,
 0.41228147157803563,0.9949091209451782,0.6411508239643176,0.4416889964042986,
 0.02753187928638845,0.5714995287027315,0.5257327339778666,0.9108727576867934,
 0.355851846944056,0.18085430241051514,0.619764631526807,0.8223446354407986,
 0.21084796192863609,0.5389733904752784,0.09484015273049196,
 0.5374999785852665,0.7136731481414268,0.7295068137956664,0.29728004691884957,
 0.9511089047957901,0.009501464352975275,0.1890569834175393,
 0.31352160479253344,0.4451395767592159,0.13741056417911068,
 0.0010606704214002338,0.23206649117585265,0.17484952991040015,
 0.6329120314275011,0.6265602485352726,0.8308204610336504,0.22215395245786695,
 0.7771027210703028,0.6349590282818806,0.6350072674522582,0.8494775749754836,
 0.44668064334061686,0.5439956359788043,0.6292461204371937,0.9810888323915885,
 0.2821228120882644,0.1914172003815482,0.13030469387319799,0.8257495585136037,
 0.8046078198172422,0.9497435661257742,0.1239821408531463,0.9472230968390931,
 0.06425423689913112,0.29388116604945125,0.7144959277215989,
 0.2926727336703453,0.3753278440547033,0.0886844864173928,0.2739750329275463,
 0.6097644222927996,0.47733712409715756,0.5171753134739608,0.6960052703690576,
 0.03603129251436543,0.5250191488237015,0.6307027987366622,0.6006828386251655,
 0.412884196614243,0.24547386071153854,0.07324268387105615,0.7588286881086665,
 0.908200929235357,0.6235463875304854,0.18478406067603814,0.8510071054795323,
 0.5208540229826691,0.9079965602843535,0.9637781570548356,0.778497097174912,
 0.13581647117552098,0.023546827485504962,0.20220720857628105,
 0.9458379248399581,0.28991512938736497,0.052847155022422765,
 0.2215607155799426,0.04140344447080824,0.649923615446028,0.17534540376792496,
 0.8250832707550406,0.31558092813539296,0.35246143625848103,
 0.1976791364061672,0.21109256507064544,0.6653045203313154,0.3975526994214037,
 0.33771260376441337,0.3273219896654449,0.358978568769286,0.8505389201356492,
 0.9563009074839212,0.11190879009110821,0.9713983309557652,0.5065188577804238,
 0.7266385471144199,0.3336872059326722,0.7601865052355968,
 0.013542470736458201,0.8660807727135788,0.6200780849054148,
 0.39618507891292465,0.6783615668303214,0.7997547276388112,0.6176886925024372,
 0.8311804430187977,0.4124149833582794,0.9211982753810274,0.8997068809582385,
 0.5265405955330384,0.6659089266312022,0.37456358732550754,0.4612922399594058,
 0.050199683761506964,0.8954213591163979,0.374786627004736,0.8019918096628441,
 0.5496593630953066,0.22879079111279987,0.7576414019411613,0.7053263846382773,
 0.3539994592841449,0.1988191755511698,0.8609071391010124,0.8865089123704303,
 0.15950712423073232,0.8798474631655806,0.2517333868193523,
 0.26749612535444933,0.15294425569984327,0.23350545105301135,
 0.5194455526052035,0.2365951758399445,0.7500900356968143,0.7747613447980565,
 0.1283976343630544,0.49892041857022207,0.3127518236756579,
 0.20922657354346963,0.8191174539748315,0.07801527272030473,
 0.7495518468643892,0.8139316003598028,0.9507620353736421,0.8114805815757122,
 0.9256476107579881,0.021390726174815966,0.9141537468610585,
 0.6240048001056494,0.8698799634592951,0.5881554571073366,0.38460953108061413,
 0.8689804463543367,0.5618002587944174,0.8478076508201456,0.6380755541518299,
 0.49959117692652955,0.381817966011083,0.23363685457168148,0.8461558236561734,
 0.4771778180731636,0.84862220120015,0.6753378795529048,0.8394228126364514,
 0.08332465510129661,0.42923397378854433,0.4839596274581105,
 0.9960417002835751,0.2876800146708106,0.5260967504889223,0.04408535100960442,
 0.7522933125960101,0.7110757076174938,0.7092974633288005,0.05702644242860044,
 0.9093010859480833,0.8425923358306439,0.3547175534953926,0.36449674674271004,
 0.016971974249721966,0.232156726395231,0.7826578156730333,0.4789287008413303,
 0.5093493485957609,0.22292736395588697,0.9774480566703287,0.8041011057542952,
 0.5768699585543096,0.7626826708056595,0.055614795413402196,
 0.8293305361192629,0.4222375248266491,0.07416690737374843,0.9404676910403644,
 0.9990735654205514,0.5011412461754559,0.5989895577321652,0.8853700119739601,
 0.8674374068296572,0.26137189048748166,0.5941634360794406,
 0.12140690029535273,0.2518889824446938,0.3601304984773507,0.1570288103767702,
 0.8043745618637507,0.06986351482795583,0.5886651491671459,0.3893576383728544,
 0.11260565493784169,0.38215403433311923,0.3188223537407382,
 0.3875205979154447,0.9507259252404752,0.9280860189765977,0.6911651340697655,
 0.24871971557726935,0.1007443305756418,0.05476933071438317,
 0.5458503754914945,0.22704665396654078,0.6142517381351729,0.3219785239559402,
 0.6283283148682653,0.5004772776465662,0.20078913460023862,0.8572830611939007,
 0.008864681046618195,0.9535038042989225,0.2613761330859943,
 0.051066892580574175,0.9983574060026432,0.8848588031679149,
 0.9970194250540596,0.5831253824011235,0.5485443371001815,0.20838519209672335,
 0.7437593869707207,0.5860976148895067,0.5901921487114141,0.3793800739170663,
 0.570837769557514,0.6400085826634805,0.9493207794780867,0.17466409262341787,
 0.3785162617566801,0.26391337447494045,0.5090581212567379,
 0.35386298022873264,0.44941248801445655,0.5360026046291965,
 0.9331116543841587,0.944557973230463,0.32560500181472085,0.931615100225082,
 0.5929861810157432,0.31239390240866216,0.1305996120186974,0.3120857567756685,
 0.5795606252870642,0.5666937700801669,0.7517881911356764,0.8395975220153177,
 0.6037258297877515,0.2830391985155182,0.18581689575694427,0.6394028759189492,
 0.6137457790582698,0.4454700096737051,0.6136207396111114,0.7031531979965207,
 0.4466736082545082,0.0464338738071991,0.20739329546594,0.05240706122321981,
 0.8255010275327519,0.5761362097877201,0.14024141991191685,0.7994685642586903,
 0.9737462432921186,0.8760344964000583,0.08057993420712384,0.7882636549382713,
 0.20403615232183547,0.21512328931368052,0.5658294842088102,
 0.5823661767032458,0.3357626463165799,0.32288712073047066,0.903116540469064,
 0.8152670031974025,0.06639014756015216,0.7842618625584111,0.5168649240723064,
 0.8374152019070725,0.7011836470017501,0.16933298295144783,0.5195484359348664,
 0.8328165210394126,0.7085750591637046,0.8937932562087096,0.618050681347028,
 0.16602983913962988,0.5479774844363406,0.014996513290026381,
 0.4896251821847446,0.8296810183849361,0.17901132725832059,0.6949374036628682,
 0.9734630836903001,0.6351751068156962,0.8035952342610369,0.9082556190738508,
 0.2736978375149841,0.857632668527635,0.6692921747155312,0.7257747325513461,
 0.021051632513610974,0.540432358480207,0.6707771902531057,
 0.17259535040241147,0.752391682346314,0.4013363010410047,0.5394474252682246,
 0.958936392082335,0.2639550392832479,0.6500103374129694,0.8241037192759857,
 0.5822248647591803,0.3973234083460977,0.7895181056916722,0.9112436848832145,
 0.1939948540905453,0.8007931398327277,0.6060787623353645,0.6427162234871202,
 0.11647840287283528,0.7070779845549419,0.1084586529401631,
 0.20132834057038818,0.11860618059209926,0.3326722429225619,
 0.26160618147648473,0.29312790503012787,0.6467745375687733,
 0.6379636767821183,0.2616150274911433,0.3161791452013405,0.9708024875462267,
 0.6320224338613353,0.7532163784509334,0.7070700892391384,0.756526453669554,
 0.8043775366667948,0.5000488487917197,0.004967989295854847,
 0.9204206254026687,0.05517156204773732,0.09012551216094078,
 0.5476850821076088,0.09741149820020278,0.8940802325327458,0.6189170260479266,
 0.5185480042654858,0.09069671109126598,0.6775869356861505,0.3940537277047599,
 0.2853331293358473,0.3833349939322661,0.7136498914498498,0.7986701459155484,
 0.7726861868777684,0.7249772398566716,0.803528673562808,0.5300187872870377,
 0.3660154575142953,0.3264440229609715,0.975646643559299,0.05543276851887502,
 0.9987426541066524,0.832229607122696,0.9590628592327755,0.041650936841211905,
 0.5420215354770477,0.6435549277668978,0.9789131059773578,0.05015380197869501,
 0.5991497167151314,0.3037393523231421,0.4844398862703988,0.21925104457529332,
 0.7093988350213779,0.8262206273499998,0.08366218230474087,0.3039802598404905,
 0.9088629590386939,0.781700495665952,0.6153259457715557,0.5520132171625391,
 0.5615825296202888,0.06878117983560472,0.3290121217044172,
 0.03150464463369729,0.47651233785029845,0.5784540168665887,
 0.8315168226932222,0.2087942476565552,0.7037973285425337,0.10574868161418527,
 0.5978856173594088,0.8333191287799478,0.06991515641366997,0.5624483245143409,
 0.5059480486881338,0.8548024195074282,0.7757441074756676,0.19347863955568545,
 0.9610761815069848,0.5276327002526006,0.03464902294600447,0.5381449320330984,
 0.19522401789716304,0.8850847100782478,0.27387652195615453,
 0.4501545120714252,0.6778941146546751,0.20172871093181344,0.8718630857061322,
 0.39882493116537465,0.2130475120099613,0.8542943825175046,0.4049500705713662,
 0.6640718625275465,0.19944089136677523,0.4446422988153891,0.7142080227274624,
 0.9335526709078855,0.6707720074786431,0.675634457985368,0.6968998359790313,
 0.6435377514726235,0.8175416594919152,0.2159197575430446,0.8108864789535831,
 0.7322757462185483,0.6982072929858648,0.021370053397009148,
 0.9052332075314822,0.22181989903120347,0.25111637253538144,
 0.22828562792958096,0.8803315495568598,0.4210260069071433,
 0.48209870367485996,0.9392882768998552,0.028332300001272293,
 0.11096905104328347,0.287183566033546,0.5326952446234554,
 0.026467190060919465,0.4341579930054842,0.7203954707183037,
 0.7159172184308931,0.9319810079317802,0.11471612745536486,0.3291745341523984,
 0.6593789421655685,0.24582611097861284,0.713466536298503,0.7560021171603353,
 0.9119558383414267,0.877299513922327,0.7640192701669918,0.11253663780468304,
 0.4598209235604298,0.957967389308222,0.4565262787914097,0.9423514339921837,
 0.47137574364016943,0.30588012862825265,0.6577280334734004,
 0.16388829366823954,0.4698611103821413,0.8148354939092172,0.2820604157073393,
 0.7733894516686626,0.6847251683396667,0.8072633025771055,0.1838206625036377,
 0.347486162134546,0.5739533369274614,0.20381805697851596,0.9004582383677028,
 0.2789382983561457,0.23340470317789086,0.08061443094007847,
 0.9770845543586243,0.8388806534022811,0.30962250690053683,0.8455917712586097,
 0.8485906873833308,0.036773294599134476,0.013180404533637002,
 0.7834058131949289,0.5669312185280843,0.8111718651338997,0.6939075614725478,
 0.14757327342780202,0.3118130795346119,0.8318558332227258,0.7749994339774272,
 0.4657239766003709,0.9411771712568396,0.25508435888829084,0.4100709693851001,
 0.25998816483471965,0.23480063447356325,0.46531553771631606,
 0.9240952315281769,0.2820623312727045,0.7532098102301837,0.5268414959601797,
 0.8833852255549298,0.6720818496519577,0.19373880118191267,0.9413654901982785,
 0.7524808422924091,0.06167527064203526,0.14031914311523797,
 0.17511129310631313,0.7616147746794684,0.6099514798988193,0.5843552932664517,
 0.9170991397634377,0.48826755239727604,0.1751160922457471,0.6945798703541722,
 0.7008162622152039,0.10953761582298405,0.8618112627998247,0.6377297082381835,
 0.3015510542518316,0.21911056559629405,0.2390342285287541,
 0.26910630977144234,0.8805566852015219,0.42029557581801635,
 0.7077745475743774,0.5042927673547404,0.8079593574335299,0.4330210528584866,
 0.49459774186164596,0.5249667241999973,0.7694867585838769,0.9177394610156977,
 0.2505709450767497,0.4467159697050931,0.001876279934298286,
 0.1537331333192854,0.6024335405670055,0.8047121142866682,0.43240548667748924,
 0.7166641314074655,0.7863445264689731,0.2281741150828236,0.36514793534683543,
 0.5582008592895717,0.7077945806918999,0.2421392102413924,0.8760884772236989,
 0.4940975335275575,0.6215697968607468,0.8248417761266755,0.19536580188777186,
 0.09757524707499177,0.06372539098144092,0.2802378775687546,
 0.23657489394004194,0.36572863450545956,0.5501493966788626,
 0.6303034887370732,0.9362900131397076,0.08144130826564722,0.8777379587881554,
 0.3243846085381932,0.9165908091894126,0.5560688253426204,0.5331641614182908,
 0.3671774020462223,0.8402550919822986,0.18347688772193804,0.9135660750859302,
 0.12167043794322518,0.9570084164105197,0.36092662080953497,
 0.20517372191048122,0.04924823639722242,0.835270824758607,0.1591624649255312,
 0.5972420661192565,0.3149828373172806,0.2212871601703802,0.9179735094581013,
 0.32798470956412407,0.49553707659015533,0.07248998386369421,
 0.6964163132296464,0.6674625336188169,0.2733471645197777,0.5948990296293258,
 0.3034245272469328,0.7485729239221646,0.7521121569452912,0.5102466474525359,
 0.7806227934326472,0.20602028514062476,0.5949438531803642,0.1871173611976149,
 0.5803770250992623,0.6289005670822643,0.16124645992268938,0.7000874798447295,
 0.39595335571926205,0.4823525967871729,0.9079295953191284,0.897892081141328,
 0.40795419810765454,0.30972894939288886,0.7392531110314914,
 0.8729762173498856,0.8335143833494145,0.7969156350495333,
 0.009870715487561799,0.871834212741192,0.7561572873702331,
 0.09603568775764382,0.4294944391485248,0.9395411560411004,0.7315838280500899,
 0.5330473704822966,0.3968298593890516,0.029474471141211822,
 0.19524722861480837,0.6001480484317876,0.23240878413847743,0.636423275975127,
 0.6501134633807674,0.4289505414493938,0.1965992564694764,0.8642198023819985,
 0.8282062137391497,0.12196063087259679,0.623278373693452,0.2805947828241294,
 0.9150713140796206,0.4485387268674764,0.6111667922020713,0.387524332694998,
 0.3517233222806514,0.5181847219890039,0.2347240952119053,0.6258252143115616,
 0.9735480645134307,0.58794779000325,0.280785141459552,0.11635143394041436,
 0.8910055719479651,0.2848872460476547,0.8645091046981765,0.8963455056645282,
 0.4617060501152239,0.8500651002928441,0.2673003803909477,0.18443440473710127,
 0.45380090225186254,0.640538284322278,0.7748093292307209,0.8107676584291633,
 0.48182816480198043,0.47616090874933914,0.346329729650116,0.7866839131309175,
 0.722598812544565,0.5503026255970688,0.2770949672268056,0.8828664802500397,
 0.02510509286806939,0.09330189808807621,0.462169156185514,0.5466945669282006,
 0.9939823715358903,0.837713442603608,0.011166183900885751,0.9797323416461742,
 0.20378173343694295,0.4259538863208163,0.9446632457330384,0.5808873791786581,
 0.44359456604309966,0.5487913056715134,0.08331458760665877,
 0.7023153905070987,0.09486443503075082,0.015725144297709437,
 0.06502214101081782,0.5917473942925773,0.4253092368482141,0.6334850731769522,
 0.8525802874604622,0.8063075350681144,0.31784418954920224,
 0.027303629987467715,0.7917765568411959,0.9558827287797699,
 0.5749871047026534,0.2382127123888782,0.7234515033505108,0.3993289679779649,
 0.3828704605016032,0.08717399018885752,0.5274733518918718,0.9169290079501023,
 0.929847404436647,0.6884865776296669,0.7330676545833864,0.6198956222701437,
 0.544975118477952,0.1960954234816894,0.28844575032034103,0.0545267311484704,
 0.7356487766968652,0.6858670350228484,0.9896048576879743,
 0.026179212945507957,0.8017302498486183,0.5172724007794858,
 0.8123256520464502,0.1857619046315684,0.1044620737279287,0.4257131379587792,
 0.02240507260310476,0.8925938359743851,0.20588842603132762,
 0.38986511162943904,0.49406452130940137,0.640030060064916,0.2260997610911425,
 0.5050829515067242,0.31175320515693405,0.9825167936469059,0.5699198612897594,
 0.36855518422108136,0.39633370616356234,0.12471829576085058,
 0.22409958693130339,0.9421836794987417,0.8920261211296749,0.2971218421908758,
 0.570125874323391,0.8164851997153809,0.9258782151864487,0.9624556430156473,
 0.6840890810364213,0.07135811554339111,0.5393864210818684,0.819871998384741,
 0.13717053633297405,0.48967709615356947,0.6119420086063065,
 0.26467379458524554,0.07129952787209082,0.11910957184676396,
 0.4106576997764032,0.6281645472604138,0.1368313888838294,0.7474164126576069,
 0.06579712953693806,0.02388235643077319,0.251117713384458,0.4262119217352407,
 0.9550800407633224,0.8089755290957751,0.4878658177598769,0.9611931667646606,
 0.5045810333648845,0.6989053156101112,0.5796394825896141,0.7200526465547341,
 0.40628139654771944,0.9554006099640187,0.9300441581482737,
 0.10005457861230527,0.5100965021128673,0.7558775709198189,0.5068932707801806,
 0.9479471125455177,0.5873292301873484,0.5239711410665592,
 0.004008136677633334,0.18955684726013033,0.35681872367954015,
 0.6180447683073587,0.7146543302609776,0.06996198624914363,0.9917098374002635,
 0.10610686017318616,0.0623329998313521,0.45770000947997347,
 0.2157295076606467,0.9477471377818288,0.7434527444474818,0.6159676778141877,
 0.05898991877913817,0.9106711443958367,0.22655764864059413,
 0.4860895544586421,0.19197521453264998,0.5714251560668371,
 0.40692225650146163,0.8978004746649567,0.016180605259948133,
 0.04014563216474021,0.13320069113578847,0.31858690108131227,
 0.4338283553160459,0.0862927463543377,0.13944075102729758,0.9286688994336558,
 0.617695455335571,0.5454907404059458,0.9253075024924936,0.7418107278908032,
 0.1624536932615851,0.07250226223743506,0.07819569233768642,
 0.4268424421365965,0.2614217917275268,0.9775713547992461,0.1490044587884084,
 0.8119522587334176,0.8758705158466449,0.4221470971407456,0.8860184269858712,
 0.7017382811290953,0.2560086334810925,0.8979103665653955,0.08139488346460266,
 0.18623515580881156,0.0086877696174541,0.8056003251216353,0.6744766652914269,
 0.6963959341685632,0.8776989639232979,0.7058803642753477,0.7294550079123694,
 0.35748491013271866,0.6468027250575732,0.45933444313026284,
 0.3678337306244224,0.03890570142306837,0.027143414265370924,
 0.9739652073899477,0.9496302208296777,0.286161040165401,0.3426192739004579,
 0.5461717603399779,0.03187012097946684,0.8814090731777116,0.6014862533310762,
 0.3879113677627597,0.23386956079912835,0.10497862893997945,
 0.4190064515158972,0.13714562030055233,0.9465438145991599,
 0.25637996434415344,0.7245069735005754,0.35358669572821544,
 0.9127713358797399,0.6493861522569719,0.6714263112455868,0.8042276829705316,
 0.29099020440225054,0.29654345703707197,0.31116296962740086,
 0.26392862972386033,0.4029919939332872,0.10392348310926636,
 0.6956450143326031,0.5426189464044329,0.9378514523624105,0.32180533515535226,
 0.9137574896610885,0.9836811406518855,0.7808198018816432,0.35806375932155843,
 0.692681144795741,0.33001577098874413,0.28279993788822433,0.6092049915031589,
 0.5790590369431399,0.23619247285551137,0.5091632514379705,
 0.20265015747241322,0.35347350979320924,0.7582815929871157,
 0.7720702486919012,0.008303835818670113,0.6858242547091148,
 0.5882141263193058,0.9819757228236842,0.3993493113391158,0.757025124776313,
 0.4269605761258104,0.1065866107652299,0.7505645756264886,0.5905825035635169,
 0.9029686244390838,0.7658493912604709,0.5695589298916852,0.4638264765718033,
 0.5893078671286733,0.05505199258980742,0.45001119240787835,
 0.15350669962461538,0.7452596374265156,0.27477687772165027,
 0.02606439024332674,0.15775883225076637,0.15124985928524548,
 0.6740136669662977,0.4739477464450762,0.4274491283520243,0.1191789657203508,
 0.9037858621728847,0.42320757661960495,0.7134270903702415,
 0.25307055661655176,0.8357093032079281,0.6044817955699424,0.8041618054740222,
 0.06404587245690607,0.44650552991710646,0.14937276028430457,
 0.10086952999869081,0.44730023164303767,0.4238234040270452,
 0.9394425630579947,0.0028427444031207116,0.7623211527888201,
 0.30156894238318377,0.6380688610551277,0.3217957167491813,
 0.37060038470185275,0.4786722216821272,0.19006026761935024,
 0.3145627549740203,0.38446636580576177,0.5569644581001884,
 0.28455787733994886,0.12634941554827028,0.7541936150337836,0.785096158611484,
 0.14925408158603815,0.7635102874582058,0.18225842809515924,0.688615019951264,
 0.778357053654071,0.6161467020409055,0.013284356896373417,0.6618382866950981,
 0.6217050246356981,0.24828857458576103,0.21337085515370535,
 0.6542919436419298,0.2528734171369551,0.6544444619865919,0.9535900971339817,
 0.6064245763630847,0.7047236169723987,0.8854275408498277,0.17940463577495613,
 0.10180040213819086,0.0069944753891472145,0.5458622554892141,
 0.8634232784514566,0.8424449644268525,0.8406906790305259,0.2009070137516923,
 0.8075785257584942,0.8029953999848387,0.1339407539098867,0.9276781869752375,
 0.28817964663175655,0.07904746594312373,0.6317407045685812,0.134853811690693,
 0.9391238688278964,0.7337466285448647,0.8689671660328512,0.16511483201025023,
 0.44426714958144864,0.20574748089375094,0.7100318954759275,
 0.6368797838974257,0.2003851213151493,0.31189893751915987,
 0.40130144502199583,0.923161669789527,0.26066690083754196,0.6436300522231941,
 0.13427936126489903,0.6465447966562261,0.47470082416643034,
 0.9983337323812105,0.7362097738705113,0.690211997422873,0.8090713784805171,
 0.08117675799371094,0.9775741749844027,0.12161542596463326,0.771414458685543,
 0.6764670669923667,0.7113205184312331,0.6706719720537062,0.3730506609573717,
 0.6957188434514417,0.11534859235806305,0.5927748392665715,0.9709482582395711,
 0.6112381115243782,0.9789687880385904,0.7690409892943313,0.532633825276012,
 0.9861066198647463,0.6900638197377108,0.05228944450468098,
 0.40929470252937983,0.5113039429556656,0.7066990726804991,
 0.10898620671397907,0.1984776077834075,0.6360435879880854,0.3646308597142077,
 0.5857834023766731,0.9659306672538861,0.65847275930455,0.6907193622783208,
 0.7077030131708074,0.5735875799252255,0.25575429901186364,0.1755818327120423,
 0.8107301904851985,0.832379329760472,0.4455741382955952,0.5826857141650935,
 0.9330520310971998,0.3466313635835696,0.5116970613888006,0.3055990817131611,
 0.7186815869611467,0.33792069603426,0.011618356396291607,0.6292930957689334,
 0.8434087066170703,0.7636297131769417,0.020542932112063705,
 0.7615028316194852,0.5602749556291996,0.977304067135736,0.8559841366113125,
 0.6512132864803515,0.08427393685963191,0.7658920791002888,0.822573846732997,
 0.9018471268517834,0.984798818527714,0.6336494278873187,0.3182048067387352,
 0.596318186351836,0.04678938415894485,0.404305820721503,0.01140054729350759,
 0.6248394984048311,0.3716793968953672,0.8122894988681076,0.08534085459263363,
 0.8716372107635098,0.3880373188086931,0.2895525919574973,0.9083622380557068,
 0.1325921039211191,0.07638604915169978,0.8168390167867622,0.6836636953544466,
 0.32020012909229956,0.827872058174957,0.6561312682823238,0.4613590706930141,
 0.5888186685269309,0.183119219226751,0.09294475409733916,0.285007203262004,
 0.11313783613122541,0.45290464915422246,0.6103008883581206,
 0.48912882992279694,0.6531081612370239,0.16938283552392785,
 0.9525749480280925,0.5019668023080921,0.12074013899587621,0.9242975331714546,
 0.6327673393193998,0.3552756055899208,0.030699472970411423,
 0.42528010930989324,0.9913878459069658,0.3968263687276097,0.7143833347894156,
 0.27042354294367443,0.5750023171261733,0.24484111666779118,
 0.9691348439655129,0.49639264406533545,0.689759457224582,
 0.050899239643903416,0.03663747121564376,0.7273831241884262,
 0.29692801179159034,0.7092624524506885,0.9124739025297264,0.7625331360671915,
 0.6544436098260801,0.5779809749562212,0.13426321950300957,0.5345352858333936,
 0.4132378927541127,0.7896444662240159,0.12617837173907787,
 0.06685139575548193,0.44703026746301855,0.453663549711527,0.4194418031846343,
 0.7368176008137072,0.10576837938620898,0.6728065648161567,0.3211791410398267,
 0.933475885437031,0.8066639796122027,0.7414190178524869,0.637222047327456,
 0.6484123542369752,0.5178228766579321,0.691896655738465,0.6622630771881252,
 0.20708053830197404,0.9061869089334663,0.1989401975174402,0.5362557494298937,
 0.2178199183434132,0.41815633398005736,0.8285165697968315,0.1978581670896209,
 0.7072024410296829,0.5583794469034982,0.43405213742025484,0.75820044490845,
 0.12482252573358843,0.10800944905608867,0.325869440785438,0.335810884533874,
 0.18889683912253752,0.5792897004563657,0.8209168532208642,0.1722924695418243,
 0.8196416589320366,0.16601501948870678,0.9428411007814116,0.6004710985046037,
 0.4791978467032525,0.4311586559153484,0.8402888624958913,0.5224265344686513,
 0.026838756941021225,0.4745299703265473,0.7927744546662613,
 0.9226181439914564,0.8215798298254868,0.05778305662902272,
 0.06598665242676571,0.6608751422797906,0.08975706666694916,
 0.5406745763807683,0.6494529888692391,0.5872093331892692,0.36535896905148324,
 0.1989702807928022,0.7963667548802993,0.4230640719545672,0.7887075086939911,
 0.20430256075141617,0.10163807945442338,0.9907472100195188,
 0.5553539583597834,0.8976669200685761,0.3303063929487482,0.38309813677859483,
 0.5225910282317454,0.00794452769217957,0.31697012756345455,
 0.7490833584851458,0.7792948910528114,0.7038197069034217,0.12615209719813558,
 0.5037547592679268,0.7458324780997256,0.6002567802977165,0.6368706084077908,
 0.8841785719399655,0.7370083334313869,0.1306758703051114,0.9213709169633802,
 0.23910443069708132,0.6170642208823365,0.6516757869451789,0.7174621816301385,
 0.1742080620236932,0.949088873039004,0.24508390352311027,0.03382947889521715,
 0.6978651160359066,0.05660093137374789,0.26919310598974266,
 0.7421720223854177,0.26637877263150767,0.12300877221122186,
 0.7443854840777655,0.6672424192329429,0.10361542806889579,
 0.46300689118259786,0.8136906583433876,0.10000395791806238,0.255671400906075,
 0.11853259425736773,0.47978706253965586,0.9343012284987963,
 0.7333270497071652,0.0562189958075886,0.6091663883976057,0.430157543427769,
 0.6830041864979451,0.7085524807783258,0.9534433987660812,0.10806271753826724,
 0.3018195159617587,0.7404575529478885,0.6908652796022625,0.35572480363647385,
 0.6509953520813361,0.8189465520321308,0.966367459308802,0.4362178675612457,
 0.15172138056500994,0.4567153384141722,0.9855896199586491,
 0.014507159849817874,0.15851959076361255,0.9206174759668596,
 0.9935029505847967,0.827421572352421,0.9389970683530803,0.03261405857121913,
 0.20589903136757493,0.374345233431265,0.7775751042363211,0.8484182711384024,
 0.18065752555375936,0.030261202545419308,0.8456194385212665,
 0.3550372395562764,0.4155504804968757,0.002858609540772994,
 0.9864491915675766,0.04117823248231356,0.3664533261365692,0.6001736739243884,
 0.9065941153344789,0.010785618695749899,0.15466280230734264,
 0.49464734905692953,0.2701738652807171,0.473520216752485,0.1469194492560566,
 0.49343164102325143,0.42335766665260455,0.5541236399320808,
 0.15326910037384356,0.5984619484108874,0.6638224954079126,
 0.22692902598603082,0.5956165916400814,0.21279937447310326,
 0.25959391865623904,0.7642301434911394,0.8470303897015697,
 0.004479484703711956,0.8463633207525676,0.14657326788422065,
 0.22978814418690363,0.11298222547767489,0.2962952545032218,
 0.5021709224414208,0.32617786741001553,0.8749114700010799,0.7061505299113212,
 0.855390435365093,0.23894100733216495,0.7228894037230642,0.4127868576890047,
 0.5956019162012242,0.16815835507030386,0.989477158414712,0.6839421085638648,
 0.07950578398492858,0.7978979719674997,0.5141246986913636,0.5744200156834246,
 0.8573116872731772,0.23499147065892978,0.35222575072294615,
 0.9394319691916123,0.4204229108332441,0.29811677900165434,0.9189027090400851,
 0.9860995370300404,0.8632391595982375,0.6201908517114718,0.6932705477095814,
 0.7191231829021401,0.3636842036537937,0.23442861549501826,0.3347007205437631,
 0.9206884228758595,0.38003200977051455,0.8348447781636861,0.5488549611678486,
 0.8159555145702116,0.9932945492468861,0.21843055074081508,0.9495158700262872,
 0.8818571409477034,0.7583372465729237,0.8895812524146568,
 0.023036794181581666,0.10416533306105125,0.6938138569055863,
 0.4457349223795778,0.20756315438969586,0.6706666792525614,0.9724997959885419,
 0.8784493243592689,0.8609636607101248,0.25196942001169675,0.5136290423206351,
 0.2701656876914377,0.8729716241526403,0.23695599966067915,0.5967880113008035,
 0.5786706549649095,0.9336398224403503,0.1080533404161248,0.6307075766405078,
 0.08563020445821135,0.6199360918633058,0.235156348162033,0.589418120284479,
 0.8700406940251172,0.9155333028044432,0.06506828013767141,0.7575105539345143,
 0.3937018181576166,0.5689503785970955,0.4046593727231397,0.5844248172060438,
 0.9745340851871289,0.5122072462796365,0.9971575091530488,0.5175632114814733,
 0.009394281479182642,0.44723621498017796,0.2503314123790712,
 0.9383254035609236,6.605692765767612e-4,0.8988473065041506,
 0.9919685247240151,0.02147762740801229,0.06170721890880404,
 0.6809457010978576,0.6916489547718465,0.023547575830989875,
 0.014843438047567936,0.06114465227260635,0.9685666870606059,
 0.08856915028235357,0.6840805989671652,0.47344847613984087,
 0.14981367962782233,0.7845536402951733,0.7077415231103309,0.8260297927425342,
 0.3752655507150551,0.6266694822267443,0.9510019479021647,0.8986261598873542,
 0.6781191248845191,0.8472240463598912,0.20462380587039053,0.7928339563203716,
 0.3565837035196243,0.392926715005944,0.04317468984939721,0.6756890351931453,
 0.9136985643424731,0.14022226775032076,0.7234616782226676,0.1904711681820236,
 0.27711745761021145,0.6449433524329146,0.3063158407230452,0.8925081358246021,
 0.5530519909328878,0.44361193047568204,0.6608218259571339,0.6336509593757734,
 0.27946990370117764,0.06758338111071027,0.43112435751357936,
 0.18200341839469658,0.15962207135621187,0.8948221021069546,
 0.41243291837701934,0.6977211557745827,0.11323164772485794,
 0.36594223773659207,0.27747925890432423,0.3402831398267513,
 0.7771695136918593,0.5884827647553691,0.24130098627941066,
 0.49832307464525505,0.7217270509395192,0.5193499373423904,
 0.41495507695661016,0.8784602367376743,0.37274733211786626,
 0.8097896921887789,0.9879878253786427,0.48043318965007964,0.8636370053872539,
 0.14863666985749957,0.8064742489682384,0.4790881518367216,0.5468466832839761,
 0.535833653769673,0.34193415874301203,0.21205556030450146,0.6668420785541249,
 0.7496674844645499,0.7452766370329865,0.11382592267819736,0.8415537051562758,
 0.7404375118796617,0.8091062496007121,0.21743113955577464,
 0.38556482756805144,0.860753616845408,0.9230377093940049,0.26792666595014336,
 0.894014519774593,0.48249359875460684,0.7134349351663171,0.1060252986837652,
 0.8548379018682887,0.16066584315371524,0.7271971613850694,0.8572271289708382,
 0.6256937078277135,0.35211235125850904,0.6518735022092668,
 0.19711859366161966,0.9222475613570703,0.03552877627091289,
 0.9722336026041852,0.8941868066808178,0.5991137923409723,0.46337883333411656,
 0.4997400519652633,0.033778550632657556,0.6829895415678672,
 0.4020580810630051,0.3959826964620755,0.13080284946844045,0.3713134442486149,
 0.6658187533452609,0.32280913228757746,0.6377544967289823,0.4145538778625021,
 0.3150630050752904,0.04432213368491433,0.5117386746281956,0.3525810416754953,
 0.975946211593079,0.3296807614498578,0.49754725242308595,0.6865982206918163,
 0.7868339283839654,0.2588190778420981,0.5225202154801566,
 0.027253309032128303,0.027012891213926693,0.7566392518709847,
 0.2660978941061244,0.13287093496072588,0.6038689534058889,0.4617306592372016,
 0.402977498823214,0.3553071380729179,0.16647591923780092,0.3165087220670205,
 0.44648105830684803,0.6881328831157077,0.09268406175445465,
 0.015492179288305152,0.506507808540898,0.7938081187265362,0.8373770217581674,
 0.9052680706862168,0.5019119909854368,0.16148747879043834,
 0.22837843001119107,0.3671901334759359,0.9162880135900654,0.4802081646161174,
 0.18961087847025115,0.3612215053053687,0.3044262950109141,0.8079789896962588,
 0.6691567619136586,0.0667059611264097,0.2980747601452556,
 0.013136902308182097,0.603469342673675,0.6371177117887065,0.7812343100722006,
 0.43841690950933865,0.12395260138986253,0.20131139386105534,
 0.9720911395960584,0.6983280607211423,0.4239595365560962,0.857988175980607,
 0.5350609860872245,0.1484207273947047,0.5118729468099209,0.5766973511530802,
 0.49291573929810417,0.6638245307068968,0.8428373725650788,
 0.13261972767467767,0.42013840812360204,0.7956017835084382,
 0.38230608556712564,0.4572426694206999,0.12802821165335176,
 0.11307147098272341,0.3062074919499602,0.403288279228211,0.9461152237418249,
 0.9516832482949954,0.5644527068715725,0.08699911875232014,
 0.36395375611048253,0.3832536692878268,0.9691722860142198,0.6227336265620315,
 0.9662485258342777,0.944677595251196,0.26841604411077535,0.9581931199575157,
 0.9615941978613844,0.19655025156965278,0.110236614161762,0.2529830183185523,
 0.9134069633901376,0.6715098348293689,0.1503909976343899,0.8695195921493697,
 0.9125808314104624,0.35596057875637177,0.8108484511246393,0.7460911887303615,
 0.12713388111033885,0.34742631852068795,0.6943436690979536,
 0.12381878760583964,0.827294399809546,0.08735612195539455,
 0.056870365963011316,0.5044977113987092,0.6429380519337159,
 0.41394221775308637,0.8863030152093957,0.9784541096606526,0.4683775107626952,
 0.8766914004409793,0.5711840712266568,0.48389969436819724,0.6868109421497092,
 0.22553155677015813,0.9581020323090437,0.9908401875029244,
 0.08002559644148644,0.3112896648274114,0.5275045205959608,0.9695211110872568,
 0.5470238168641084,0.28474490213954107,0.162099537126271,0.7947039007387678,
 0.6633503983715443,0.11330356730977265,0.744365615192437,0.05185345876491154,
 0.503595987102355,0.6238929431637984,0.22156709531592222,0.9810572881179178,
 0.04446953336261694,0.8728133387610404,0.6628403744927522,0.7091322458182674,
 0.2839484339122579,0.6777458568652479,0.8788448904954975,0.9883946121486111,
 0.17265959602374992,0.043074952757502616,0.8235636461704258,
 0.18933596557159027,0.13594843879503837,0.4767626336270365,
 0.031237369426242778,0.5062621261829046,0.58267372357128,0.11845339422208068,
 0.0012265856800687258,0.9571130061449358,0.03333025149867086,
 0.10596623564697438,0.8223259524252903,0.23209523565719503,
 0.9081341942220194,0.973846465981454,0.1180781615538733,0.8292574321498252,
 0.8871520871975994,0.43563154362716483,0.9914363505899827,0.6210458086623587,
 0.6634190164505572,0.49264257906525655,0.23769640350260868,
 0.18793000406693106,0.5121838790371598,0.4380717723630584,0.8513933760503983,
 0.7708894366222964,0.35832178833320316,0.8698413054640843,
 0.37045698827329965,0.25367517857771216,0.09724689361563554,
 0.5448239866659825,0.31489204746542276,0.6590751026528211,
 0.24676210631027617,0.15790921118686485,0.1375585171908692,
 0.6749116654433971,0.964949736178815,0.6112260276999256,0.5264159124743985,
 0.059057092660409705,0.41962936175896426,0.5762704949840007,
 0.1472183055273799,0.8014981507868376,0.3073536252244372,0.4043245004080176,
 0.5379377596325007,0.18608574573135317,0.17023801391859894,
 0.7709901005858335,0.6268536883160721,0.4018641887642924,0.6270574240208908,
 0.3122849992921338,0.7326076334189878,0.7441391460284733,0.24234176246653671,
 0.11767252750672763,0.10279434333514836,0.7200946192036777,
 0.40073425160471055,0.8703916930814267,0.8440179045956416,
 0.39497749990614406,0.9064279728557945,0.04120080997852782,
 0.1758512271404391,0.39458470291011827,0.9376446826003615,0.599936045837836,
 0.8618262719959828,0.6607793367367628,0.4533723929178626,0.03018935807276646,
 0.7226815675947215,0.24657552198558252,0.10145411946583693,
 0.9830763203047022,0.7586758173023381,0.5213753312543468,0.7112076995942159,
 0.36483347901980356,0.39803825431256123,0.6478961143071754,
 0.18230825290188435,0.1597220850212907,0.22279288137823694,
 0.4088477608672846,0.2915019481912552,0.8049980473678926,0.7276901361756015,
 0.7028357732576361,0.8998053712969319,0.8205135613160174,0.1468377383445859,
 0.436901057542478,0.8666308429078572,0.716874037556935,0.6932377497691851,
 0.8597712073228024,0.4128060599280843,0.8118199548886482,0.22486567439418048,
 0.5524587063006732,0.40530022286840917,0.11310828184389488,
 0.3522454281144257,0.34609912988004843,0.19889891613291522,
 0.05437362717239058,0.3491078533678651,0.09126367167220417,
 0.6156759279570215,0.029349680848527182,0.44403352574467525,
 0.5186202383441852,0.7781989750901208,0.22562234343182686,0.890365893178708,
 0.9104470560606484,0.2452738823061995,0.6830830255485386,0.2913627800873003,
 0.3373926291721182,0.5566021390706585,0.26850251214921983,0.5734452089472326,
 0.399634417641628,0.6605058981576378,0.6797145870989526,0.400125863218195,
 0.006837943125923118,0.6494263582017973,0.21015387740089608,
 0.417299389150092,0.4866221182889777,0.8667782948469998,0.49321672203999656,
 0.1873111912570986,0.4505764194737183,0.7941517727204142,0.2922114946707417,
 0.3089284775336151,0.911267087132962,0.6663823244250674,0.5507022625773363,
 0.2441413862132107,0.7073256929954517,0.23560440205532784,0.4150916467876632,
 0.3873412350494365,0.26605031955203007,0.9331850953109719,0.4555754408438596,
 0.6377107517863042,0.15906780592286307,0.7040845677907352,
 0.24043824790039925,0.7824198056546767,0.9133821994440449,
 0.030286980838546307,0.22744399318477682,0.3059446936188144,
 0.20257578106616714,0.6551723852654638,0.7150175315253787,
 0.29517757523374866,0.7995197529021518,0.2897692814456988,0.8235064434553592,
 0.1576043980526054,0.8788835033743829,0.5242618371796134,
 0.010247352730504367,0.9637618849555998,0.37990964929491,0.8303625919372984,
 0.11237899324017009,0.5039882923072966,0.6839794021374828,
 0.06407001215423336,0.35263727088881525,0.8236303635252373,
 0.6165563472792154,0.6242702450215217,0.5107929524271259,0.7528947435510058,
 0.09215321609063354,0.07409855356935913,0.526901101523818,0.9725511910201117,
 0.0633666579752824,0.4326836111210415,0.6121346168841858,0.2731566785920615,
 0.71656553961606,0.3989809477077094,0.8632699035360649,0.9893106330728759,
 0.2946194282508908,0.19203840850211162,0.06872699931636816,
 0.23811418646420235,0.3422431017570783,0.0985323144212632,0.5243982333128293,
 0.873745377885757,0.394005440770801,0.5419774009826182,0.07740420214011412,
 0.945718890020327,0.8434921138705282,0.15508257451492202,0.3085374131506108,
 0.47597945033361366,0.9279969391199401,0.3360972798078641,0.8108631726332343,
 0.15121848754907297,0.4319403651180942,0.5515201490830657,0.206959155222172,
 0.6291932587865485,0.19888677224315143,0.4852262660909731,
 0.28140950770636264,0.6749833910230648,0.6262716572210753,
 0.41818171282541716,0.025133834854807025,0.3927735933718096,
 0.7346215400482746,0.5499562070880158,0.6232816489191356,0.12747520805683266,
 0.103981289168388,0.10522461275036277,0.36336078521092363,0.3128711522252855,
 0.7428007608617726,0.9972479048974116,0.5232290938557274,
 0.026307036846073495,0.12121284013546996,0.32968473879114346,
 0.9786990125787082,0.38782394785089813,0.5609312552072565,
 0.14746620268258726,8.541371519661389e-4,0.43109070738571087,
 0.37021269444795213,0.877483806728523,0.004039486088681099,
 0.5944917699789358,0.44691838478448054,0.8018297699767828,0.998115114793147,
 0.5877847782514706,0.03891060341742858,0.6507223353526208,0.5043613583794612,
 0.9615744486608289,0.6763464802642581,0.9989899061439679,0.992127171091469,
 0.13333531407953114,0.41237383740298694,0.5913570837531957,
 0.9753991343676051,0.7300155377949304,0.30443765004232903,0.8725805101608595,
 0.07499621378339683,0.005335213830754748,0.4087940739690551,
 0.11192554754330197,0.8430986543720884,0.05166903071246254,
 0.7954098284381828,0.12185158445147959,0.16167457362539017,
 0.9934896081924856,0.5728593172491256,0.34976400808407904,0.5848075663912079,
 0.14703122234488986,0.6593556321179372,0.7098725179384826,0.1975427026917177,
 0.3830410909580826,0.32506553132617466,0.3059250014198468,
 0.40553487204697236,0.808896641948317,0.32789928727403794,0.6222365384002699,
 0.8795646536226542,0.5930566016560388,0.02831629360162935,0.3520504742712982,
 0.9439674255635531,0.1118613803060825,0.5201493929646014,0.7923408792490506,
 0.5498382993556475,0.916667591749355,0.630141439838019,0.5636270840297557,
 0.26813580100677226,0.08820805555617428,0.08619662639006576,
 0.4813236264423436,0.07222279040920265,0.5939256829999294,0.0758933529800474,
 0.21449861645800772,0.9762533724598349,0.34717708266394287,
 0.04876534503116159,0.5883901025630759,0.32107054685918723,
 0.2104334292920721,0.049528966243895134,0.4016924127275041,0.977340568509264,
 0.6090856784354176,0.24073135040717508,0.44697235149692016,
 0.8036760298209609,0.696173960573817,0.9182494700700855,0.31965081086099656,
 0.7893158286377863,0.8980128505332174,0.17663030184320705,0.2549868561594726,
 0.5508439336980981,0.5442230261754533,0.054906657815352644,
 0.4026623472760965,0.2287683648545733,0.8494299602028114,0.9945388221808165,
 0.7982485105725166,0.9604259911399958,0.061781604402335155,
 0.8728557389115803,0.21974860264532647,0.36797039218106864,
 0.4604242661214266,0.9811763267835838,0.32534594076114953,0.8072820119341333,
 0.3555923641319573,0.42114943696761176,0.6715538135864689,0.9825388674623752,
 0.029793700986239458,0.15305278270422362,0.32887691707781774,
 0.2996378789460803,0.6422342107525933,0.5961622391889148,0.9160863812236457,
 0.9925598940996541,0.41872429190669,0.42204457778262705,0.07562905887744309,
 0.06260244655417135,0.9604871126115861,0.9340780731794569,0.6008447494264979,
 0.04839550102019785,0.7749608110007624,0.0964192074119753,
 0.04711023939541015,0.5116002435480496,0.8426612951519147,0.7217881185414408,
 0.8752727502852814,0.18150836037719098,0.6918896523070214,0.7323888101013458,
 0.8913345143367991,0.8818780691741879,0.44033296140107425,
 0.22677950169825734,0.38451413572746085,0.19457789182141982,
 0.8563661581478277,0.6045693693366782,0.15950649179880494,0.5001239847057505,
 0.6722763463550421,0.5670443901170318,0.5835695477630827,0.7764726274530748,
 0.400195523535153,0.6206527253578864,0.8212960796712631,0.3668050182213656,
 0.43762092521731,0.3926387596853754,0.6441741376697304,0.6669551456976622,
 0.6005959810645307,0.8105960555757704,0.9071592286827423,0.480065261944399,
 0.4372653999448647,0.7789683828472328,0.4278722998938075,0.8661822970911693,
 0.902769306722159,0.7564155419950164,0.420162826922549,0.47302451518240474,
 0.7216764799031621,0.06263036254295862,0.16721598871921883,
 0.6328256849930716,0.5604405119654977,0.7028143272702992,0.6429990002204747,
 0.77743410934219,0.4775915283218637,0.2381642851871918,0.2613358312401317,
 0.38206627437403284,0.33558916959952534,0.6644692379974306,
 0.9417865750545446,0.6788654143708093,0.532460228594347,0.9509510503943961,
 0.772353338914488,0.1949872909584892,0.23615145856017472,0.6036324061540853,
 0.7223716424055144,0.16470694510188044,0.9087196559906114,
 0.030451070445056017,0.6935637710134708,0.9844539304419001,
 0.4289129659047073,0.9059995915792952,0.22396766220779785,0.6607371268462091,
 0.18016026911578764,0.4747543173402101,0.08163374668160961,
 0.22360193971082054,0.6029503624610257,0.08282617647137969,
 0.9771012974030397,0.38338625074278754,0.8462701083001912,
 0.11529684461253864,0.7791606189446109,0.3674187920366534,0.4971782226720476,
 0.9847606914273206,0.47862139490876765,0.06836430346535805,
 0.4155045156025374,0.16083649548271206,0.9619455072851162,0.6315241708721264,
 0.35013179409745043,0.008901592720436002,0.2727089294056414,
 0.8966119518377478,0.9896024807336541,0.7873138592482755,0.21103178070991635,
 0.9932403795007292,0.09541827884555398,0.15242598819978348,
 0.07918425700512155,0.41726228192143455,0.8880773574217214,0.50293680776406,
 0.23517585423528964,0.8954559602054439,0.950411944510904,0.666121389328374,
 0.27854857793504806,0.6249289140996588,0.19129400362044346,
 0.9399284834417374,0.5485905492648974,0.4308265704347567,0.3962586897797967,
 0.770298023939231,0.09504931590022114,0.4925692663392368,0.34274122699353227,
 0.1842502669896634,0.7144184913004848,0.5918234678511456,0.10984723610242686,
 0.17150945708013543,0.2264740059268362,0.05811513639101318,
 0.14758829064599288,0.6398281650730958,0.10684838315841116,
 0.2843493250901167,0.9594866822421823,0.5970201351591479,0.7721193243823546,
 0.638931327952688,0.8066882925706353,0.9059989869216087,0.5757321158023014,
 0.4810381469094692,0.8656353223119847,0.6239857908455276,0.11096699905126173,
 0.38875493693576957,0.41528689712737277,0.26865781896214214,
 0.5894274965780971,0.21681066984023523,0.18512727724351175,
 0.06423886084381913,0.8858140017370872,0.6005281111274352,
 0.03550366222882728,0.6960019812436488,0.3743611940185314,0.4174729278275525,
 0.5377943399261377,0.9680418115966517,0.25665913167028354,0.9458000486622296,
 0.4632019441002101,0.21305944678217614,0.354075544384049,0.2886861182930347,
 0.21410288447259052,0.2578217512255153,0.12224868531889621,
 0.41145479165639365,0.6420838734231387,0.49680815745232065,
 0.36227018503943853,0.1036810192642863,0.6240917514916884,0.991376134005408,
 0.8799133501357914,0.3946217871215596,0.7718880432055137,0.8741584176284019,
 0.40971226153993756,0.03189842338562465,0.8473963197169301,
 0.8903361490942511,0.0874409432597113,0.546424274553722,0.7928766401829597,
 0.2990930401855446,0.48279854328297056,0.4704981345760373,0.2713611658336482,
 0.8745793840836903,0.6482288964587736,0.022140810662338084,
 0.7226399341231967,0.3874736159768829,0.7940430849831296,0.4429682493153211,
 0.6773615367247939,0.012334048521285279,0.9127113423523292,
 0.46287164336444886,0.826039778538866,0.9005902068044996,0.4496148300727829,
 0.09022584727804883,0.9141317060112442,0.32607433744387804,
 0.26640260808189575,0.8648917698275685,0.604866186581891,0.28193630621872345,
 0.09140273009818156,0.15786744655271967,0.6666485526151154,
 0.3466515014068712,0.043228998370698046,0.5273524939080071,
 0.7217510122841212,0.18172494586216426,0.6253876025725436,
 0.39226940079367634,0.02483941372881593,0.8941902150782782,
 0.6106163304258274,0.9753655806625632,0.44865306850274245,0.4868843352094079,
 0.6237067173125604,0.7404676496946333,0.6868412343619185,
 0.013757092733265641,0.8525159180828114,0.9004805096928573,
 0.8370238078620038,0.7179958670215765,0.6221079929820073,0.4660726024942764,
 0.02636995829554123,0.20166243576670084,0.9751234792989344,
 0.19526326672308092,0.8527412265428789,0.5898106116712489,0.2813683315464439,
 0.23907568490702513,0.7844544875789751,0.12952348012019266,
 0.5923485461712019,0.37385954982501424,0.4334956639168299,0.7144148170554975,
 0.779266212572804,0.08500049815294575,0.40620239286359805,0.4267899703912267,
 0.15438594927492533,0.4929765600864302,0.17832576670075095,
 0.2924466567123407,0.8802672986762877,0.0487128324409114,0.2417320573395133,
 0.04673915785653715,0.9591273716420601,0.5873200614522428,
 0.21101838324208755,0.5114483578273828,0.9799319653781225,0.8393769179875357,
 0.9597412202188701,0.19755807591034447,0.2075491362162587,0.5673898090570045,
 0.16065251546752657,0.6741310103401642,0.7482940835135196,
 0.20216660234760486,0.06850501687356125,0.27414520610858273,
 0.8455576005875098,0.7364041401170495,0.21747982120158615,0.6441178933924371,
 0.4957172989231281,0.7653920283556268,0.4761954946976278,0.2946525906881178,
 0.13197698437723637,0.9803605797852338,0.4404161624019354,0.5932047512354574,
 0.9961885376119088,0.8133162347672505,0.47015821690873383,0.5154391966209544,
 0.7802081681548249,0.5854330464477119,0.8306249014210105,0.9640240014610147,
 0.6808697155382543,0.34371203486462565,0.13444604392120274,
 0.09498076319536297,0.13029679983884324,0.6137964464377946,
 0.12583712018413895,0.5359564795403564,0.33698448036278816,
 0.41787394904088304,0.739506001850976,0.4408172624719162,0.9204241569499654,
 0.8582929521120801,0.2816474743508931,0.3079158819955743,0.3895371148794359,
 0.6283373723783787,0.11547569598477458,0.7134303243836265,0.6080450473621495,
 0.06473281068268677,0.6329551573758695,0.4008537424191594,0.9613232138668393,
 0.5734154912963078,0.36658465871304524,0.6244310334635972,0.5868784085744699,
 0.026388097038743563,0.34364773070684207,0.2150308678209205,
 0.2967369650907872,0.2849070852275233,0.5748672697731055,0.5976617960338712,
 0.6744505645108319,0.5849258004793157,0.44703698979270956,
 0.15881430283196207,0.23707338185369486,0.08173822701798716,
 0.8073866406161154,0.5988813836276459,0.03612732676089214,0.8488280437152561,
 0.37582668407240594,0.5041796580985956,0.6178527969867881,
 0.20801849142645756,0.005530957942538883,0.06672310527386671,
 0.4695885655153802,0.67986698466632,0.597405933703579,0.20286076394110764,
 0.5403374388876065,0.710224657263719,0.05091803194271716,0.9889460036649742,
 0.09763664542943917,0.7409532150403964,0.5803887436778571,
 0.24230185356283485,0.3183528765505741,0.47589705616552336,
 0.10532687044117306,0.11644981603623983,0.09448777684208232,
 0.8580638368705973,0.5556090600507367,0.2698119782542079,0.35635076131082566,
 0.697323726496496,0.7626113568531843,0.07575433086805139,0.7794689202279023,
 0.48275228263236036,0.7773307523136919,0.8752616096629858,0.599622086510406,
 0.05608531575067721,0.8612367825852197,0.09796591116447473,
 0.5606780879826087,0.44658753175053145,0.4153960486330349,0.1959108740895853,
 0.20267473052863894,0.2795840448479404,0.7104092455638479,0.1180678411114714,
 0.03386575881270437,0.06021520739197683,0.9898563958839084,
 0.08213434812681264,0.6095139233332081,0.648825673430905,0.45319178843926955,
 0.16534117792230296,0.11948031704394868,0.7951253613715042,0.852426330947395,
 0.17596253861097388,0.7311018866582542,0.07245195011595085,0.175293319780656,
 0.9523212109381644,0.7888171810312249,0.04349136474244697,0.8473945349794337,
 0.7094659785565834,0.34948087588388155,0.8383654715217893,0.6020010114500124,
 0.043788278510874745,0.045616453787082456,0.3798207116468262,
 0.48925332952454226,0.28495925943988165,0.6025643145846661,
 0.47380418338023844,0.11259453844777634,0.831088219354988,0.9573914853813451,
 0.5629812054357474,0.7173862241809598,0.564742036489986,0.12295692954784898,
 0.7348238710808103,0.23395304451067966,0.7313249572496565,0.8761278784023117,
 0.1833973056416648,0.4226341021204236,0.5329712114430445,0.6626070470388956,
 0.6176604659873438,0.40267342007634044,0.7095355269381408,0.6091378205758042,
 0.07477737078872415,0.8628220702226654,0.4377167989268206,
 6.53078842062671e-4,0.057371385838874134,0.23627799104813585,
 0.2947651458190619,0.21327528480364433,0.009467433786790047,
 0.022302923902331018,0.033006184678680106,0.2653736609484387,
 0.40907326323374604,0.529402262627205,0.07290133284910505,
 0.040508982780386615,0.8551190825648795,0.046256179004659725,
 0.6345802699399983,0.4867153063359534,0.17227932285231795,
 0.29041222788198995,0.5494919032356222,0.4984766755583615,0.4567075709504944,
 0.03603351996733828,0.3169172619694829,0.8340593709037373,0.9996989547421161,
 0.36109361268184237,0.6376338123564174,0.3206838065199078,
 0.49854360868116454,0.9539227345294886,0.7440954041277912,0.9631232383795474,
 0.3978393601645548,0.9350007931304968,0.5627887816262569,0.38287280444938,
 0.3620324376925974,0.46921513027941186,0.13390892206553662,
 0.8083929910631106,0.6892600566821745,0.08661405454590188,0.7165248883235386,
 0.34811186125685867,0.8705123620243604,0.9255537950887386,0.7796572551899419,
 0.24364887319779538,0.38899808840390804,0.6697706424641205,
 0.8589911824737616,0.20575283406249467,0.21073867138435776,
 0.8323822837336314,0.4728568045351841,0.12775548966929373,0.5343603344836697,
 0.9160611060675717,0.8868978357324608,0.19658157333675108,0.6028336961356455,
 0.5917960385026408,0.10908115256513129,0.4100266791903564,
 0.47960436061026623,0.19268287293180897,0.2781446863480944,
 0.6784461500127839,0.24008903446386376,0.7296148249324604,0.6148541366123879,
 0.6045487026460953,0.6118683659997765,0.26482986966168454,0.7213721661027792,
 0.5405616941725645,0.28365289840050356,0.0981974309018383,0.548562940507199,
 0.10626834163839272,0.03664497999191074,0.8012856975953399,
 0.3252507769040417,0.9149780387433001,0.270582800593981,0.6402414432102455,
 0.8285310600264122,0.6847266622358261,0.36167163727929086,
 0.40609230863275625,0.8239985376389034,0.2124513734005511,0.1321701587502574,
 0.8909521589355214,0.9843134918473001,0.9673742908777239,0.43069783310748555,
 0.3563436114790125,0.0896244723793489,0.7396651336256211,0.37753980524992026,
 0.6914315535992861,0.18131326886037302,0.12857744634893375,
 0.7275702042147337,0.9932500786084988,0.7468664446480532,0.6429215721820694,
 0.7397523900962955,0.6870048792129527,0.4123445720132395,0.1259863781463545,
 0.6707395588984572,0.03734808592681693,0.8182063760442304,0.6749964556130315,
 0.5446390088973238,0.3900788611991697,0.10287450473378934,0.9243083218430419,
 0.6366839093433637,0.4059389029744973,0.6068623798903856,0.13451530169815573,
 0.7866393815850358,0.44584160045144483,0.45036361205019715,
 0.7546485338419651,0.8834046668969154,0.7760808751815582,0.8148669168127717,
 0.6072114749733808,0.1865355502971372,0.1367217448581428,0.3444960438626149,
 0.30669656197084905,0.36141998799875497,0.6933040040328723,
 0.17442033731689865,0.13759904529820477,0.45478906483329606,
 0.2531144455934261,0.8857903121856744,0.31739928593459,0.1248399365524856,
 0.30458675664898693,0.39201108919499883,0.6743272095299632,
 0.0038210281101735966,0.029355707578324286,0.8955064985755427,
 0.050870683615743095,0.7310580276779708,0.27614172236468193,
 0.6934776880745002,0.7243234448121322,0.7602796912359726,0.11164546450989088,
 0.8617531484869373,0.028252262285059837,0.6796590260052383,
 0.4615623872953529,0.01730279086305031,0.3800843570740675,
 0.19144464330649358,0.2865367167023748,0.6306799075933098,0.8473479439622403,
 0.13951075420815184,0.45220537742564526,0.6969882898948074,
 0.2715565742722792,0.21697647180717117,0.680543073451937,0.30772967067062895,
 0.5331227057985654,0.6857699803244972,0.9466261497932793,0.2526216964386706,
 0.78089369857513,0.9673833083376613,0.3388758464413528,0.575659717100385,
 0.23507551281123107,0.8042144317979601,0.6849299438224905,0.7476931795303158,
 0.33934174317511967,0.24126257626463588,0.27999751245243165,
 0.06807624739414952,0.9630928743473265,0.6866731390838661,0.9971783954324772,
 0.9675573528347451,0.41447153806088566,0.04789016731821549,
 0.28792673227771826,0.9713098463679813,0.014564718301415591,
 0.805714986282136,0.5607523027779726,0.7169027483755701,0.9837296819918523,
 0.3199640220981227,0.022164175027388655,0.426057054602725,0.8053434438142962,
 0.4272204157081203,0.4722313201665891,0.01478834815964336,0.8076047880673293,
 0.20730044464440822,0.49748222047033963,0.5867176078115209,
 0.3951709712475536,0.786182271588801,0.13246553217744406,0.899542574709221,
 0.006479545043222346,0.16542407295965267,0.08017299628837848,
 0.4949015191641941,0.9473960709172601,0.6480974216614215,0.6400009782361535,
 0.057657801280755705,0.3227816579145719,0.1453613934084209,
 0.04695713717016592,0.04926475542267439,0.5934907326968639,
 0.3196583710388814,0.672149717407295,0.1215081058465105,0.32147778263937055,
 0.15302472416212176,0.23267266740946535,0.6307224637265991,
 0.22146702563671317,0.6403608895794043,0.8869194421820369,0.3011463289605224,
 0.9085097839190149,0.1269646820225434,0.9883386208105176,0.9415828287269575,
 0.010244486735880076,0.06152206805924132,0.7599792514448022,
 0.426809074719817,0.6353701513828542,0.5262062891862657,0.8884351434876088,
 0.9391218710628225,0.2653624551711993,0.619536333786742,0.30417964468901193,
 0.43050328669787585,0.0799359719278101,0.06540070061765446,
 0.33840420257907067,0.1940743384624949,0.575252275331575,0.03020748437114751,
 0.9790320480050532,0.0517516534339042,0.8589827623966678,0.26296501839051956,
 0.6216860315765922,0.3497764557539781,0.49218098405266975,0.9989480651337086,
 0.8272509196169426,0.9771238806665735,0.34073264020184246,0.7132891894565654,
 0.9251382294260977,0.5901478710338492,0.9262733011193549,0.6688023877998835,
 0.9819445381447844,0.3808876872090675,0.3852959620309101,0.7458916429781435,
 0.8591854522167597,0.9906968515862635,0.279318013614521,0.5175903671012801,
 0.7023689625761838,0.11489930566942941,0.9815845823856628,0.8761425021016805,
 0.2653337149940752,0.4436213037600836,0.6064496510600829,0.5168252818825335,
 0.347951534558149,0.33438681735778675,0.23489032117153008,0.7987586997720266,
 0.17922840880674917,0.6877485073921784,0.18765791412688937,
 0.2385351132285346,0.47432599959773125,0.5359930419699687,
 0.22298303219722193,0.8841893763040952,0.7650391163624152,0.5335245944085479,
 0.8587613529048865,0.8751222650418473,0.13457522347758044,
 0.20313069169942244,0.2672841372943,0.4280998519877013,0.17631622821488668,
 0.24518362562439422,0.12688101489060288,0.5176745310676961,
 0.09752296565383511,0.27812311635856857,0.7125841410862471,
 0.9463688749484415,0.8656194089318807,0.2556555550693964,0.38320003967378546,
 0.2906595125738346,0.24636513685324524,0.05462657477010191,
 0.8132253190934464,0.6626836357594611,0.3942793170843172,0.1701609373601427,
 0.6444880174508189,0.035372982509814244,0.48740799290663595,
 0.9044219817365796,0.5777543123279727,0.6710828437861721,0.5776838064146127,
 0.24702910302949144,0.9640299187111961,0.37845099818463124,
 0.8746661951012487,0.9045581798323192,0.8762758526611055,0.399127844668982,
 0.3439574676803214,0.13114343944866236,0.7589364873840891,0.9443075099226385,
 0.6942453423948219,0.39537638247359375,0.6534725264647241,0.6094392660473233,
 0.02725618367479332,0.0700537160786745,0.6527949701845214,0.9866559741150585,
 0.489394738741481,0.3586063783580904,0.7258864639811609,0.44840357190697255,
 0.7241012690292273,0.9546018747897133,0.6658607737617109,0.6050086682399894,
 0.8818167170489246,0.1117197322691953,0.5379991481851429,0.5522272468114451,
 0.2766324282141126,0.6738176361853484,0.5728900451086261,0.21916597181830577,
 0.47915238951469097,0.9440866336303311,0.4931940898639968,
 0.16761595450681066,0.8891581630018782,0.04521613413067693,0.323567340301414,
 0.7572944709643146,0.7621564454680446,0.4562694866383805,0.10867344049138616,
 0.6145935787291859,0.49088620918930803,0.5842556421700359,0.6272493727964124,
 0.13759637646275968,0.4093498041501147,0.08129435654777273,
 0.9762766690210392,0.4714071163241377,0.3779300556833717,0.13956678362481445,
 0.40528133420115287,0.38130206309324144,0.8930297435147511,
 0.7357865375250274,0.6607331175175878,0.11785121839208801,0.2831660391286106,
 0.4727867092295156,0.48704069536953454,0.4659500857709509,0.571249667007135,
 0.0048384956130503864,0.6997642236001723,0.8146897071233735,
 0.5575279240967751,0.21228125372728857,0.9695294110470823,0.7196024830653486,
 0.7351993134493482,0.5431348041978925,0.13360444196665244,0.9034140216384154,
 0.02569414520719704,0.14455207647121837,0.027500216397112864,
 0.2552265142073824,0.2612081003850927,0.01160655514766773,0.5200292596943508,
 0.557537914697209,0.013421056078136595,0.06973694024860999,0.518832583912478,
 0.6656025333723641,0.3096330249731061,0.9352015037857242,0.37433774516423646,
 0.5795966601471433,0.8532953348747804,0.5953277515860053,0.12160356696565977,
 0.4772139302401013,0.13379394820955837,0.07855836349585565,
 0.36543255545978237,0.6123141971901066,0.5555788248110171,
 0.21365636890643291,0.5257063149318979,0.8450655765907146,0.8995406645450111,
 0.6224226277609219,0.7175859758228502,0.5549294010948433,0.611572059655251,
 0.7749995162724284,0.8678262995550573,0.24851291179154344,0.3626346884753048,
 0.9436360445072902,0.1944424419986771,0.5009835922359882,0.6353365422070204,
 0.9154913451539479,0.27711825732371553,0.005294868949960008,
 0.7996535311348538,0.10861315616565825,0.37008846828643815,
 0.31073814943143085,0.669364317032321,0.20535857687681403,0.3072539523635728,
 0.08831000902486552,0.5038602059570465,0.787890320051812,0.2179174909197541,
 0.5234182642440046,0.599878847605299,0.6171078961208145,0.49379503942798575,
 0.3058224287373006,0.943537507984686,0.8049141833764939,0.1902130845210639,
 0.987037781635912,0.9348704050944593,0.40756876862662805,0.9072960478220333,
 0.7527486412362188,0.497002578103269,0.19500243768753345,0.09540634621126665,
 0.9197311634153659,0.23453987099314233,0.12440387161389854,
 0.029762112483872838,0.9303860966382491,0.8710126931749256,
 0.14894903977645801,0.8870692682496489,0.49857722013053685,
 0.13754804414315114,0.714168018607403,0.9687847391159872,0.5147536769262883,
 0.7173333859528798,0.8126956324896928,0.12312887431029962,0.3004372901674526,
 0.026350849576716406,0.28848887365607556,0.8926461259553665,
 0.43316244386737646,0.8277820087394694,0.451829185397319,0.8326005680324073,
 0.7552616564893058,0.7311748644922552,0.5227741669587863,0.17796117772497,
 0.2711182600092813,0.6771666440017734,0.8160388495821259,0.2314171118965772,
 0.6038173111755223,0.7453799705176773,0.9948681881657743,0.6125945653629346,
 0.7031493724077363,0.10730062827751607,0.6093892575139803,0.9377249606953936,
 0.3940001306084263,0.02301840389908205,0.16584910240996198,
 0.05057728318213228,0.3575173175461701,0.11166822368144402,
 0.6185829219322662,0.513638434860496,0.2722516899554588,0.32038005945957426,
 0.08756832086491673,0.03870223474871948,0.5328367566523835,
 0.17074559532134692,0.5254274506393803,0.9883613699962356,0.9784762104219029,
 0.22464416714934865,0.8528810664418236,0.1764715269588204,0.7217337430224952,
 0.7646316494635252,0.5253547121174074,0.3974981898700498,0.25098981815913735,
 0.031529928018827036,0.294218635407243,0.9054955608430093,0.4463182660299061,
 0.23507153641820266,0.8772375542438895,0.3705039515673717,0.8462132818001108,
 0.3285409108322651,0.3113918636433004,0.006464091934750282,
 0.7763764124629612,0.13249265756417572,0.9230872257315578,
 0.07380394489535247,0.4425721841804896,0.765259942372624,0.6135380099082127,
 0.9661913761323679,0.23992867948040197,0.04096598746557434,
 0.6614367916073141,0.7847186906929398,0.2691869151470072,0.8527303571610076,
 0.6610943796973899,0.733276719593373,0.8039076905419891,0.2770588187107621,
 0.5893819855905293,0.4129774148217783,0.03748380032115317,0.5969346834761265,
 0.6586258222954141,0.5371846870908228,0.7138887662455206,0.8695354444200827,
 0.8162944185195156,0.01634990179764939,0.03286498283526618,
 0.6822812502299203,0.6802921291508386,0.016652043745336687,0.867875757232557,
 0.875870993347773,0.08403738173536235,0.5861147705623576,0.5011903934243768,
 0.6510186923927946,0.9742913902073586,0.6779752081329395,0.3987326544499282,
 0.2631765753557659,0.5095680858196113,0.3578114083492818,0.28702714782935057,
 0.2471069830721475,0.5278318760497541,0.09231101693947552,0.0492924492435769,
 0.42960826894855086,0.3330936614479698,0.8801508934796948,0.3323459203391639,
 0.556389031871965,0.03827072705342849,0.9963140005513645,0.8305610129896285,
 0.7847802995306817,0.841601771810066,0.6405360902821328,0.12937350952305304,
 0.7946276738912539,0.180751723020538,0.15155374267142385,0.30711963479975124,
 0.40808672840304283,0.499143598737941,0.9025881792259892,0.7828951586512488,
 0.09726124495534993,0.7607437460199595,0.9502247339709239,
 0.43717971886897844,0.6071631474992117,0.8467307993460735,0.4394418922403228,
 0.5035503341900918,0.1342585626020656,0.9655311155465294,0.9889679914983097,
 0.5866735035950524,0.1889141024868367,0.7987866122428147,0.3905800956771308,
 0.10659089314476267,0.8324847483925802,0.25839225847622327,
 0.8903252585905319,0.21763949611131994,0.2709293339871415,0.7417540526738607,
 0.8464020188977188,0.9823071216867394,0.14482628476796822,
 0.45624424115127304,0.4108790612689236,0.5356900469153796,0.6315424105105555,
 0.6512690950908234,0.244384172950594,0.986662882536797,0.9069230561195,
 0.07692008039846399,0.24022608431636938,0.88729763970862,0.38309782161736305,
 0.18256882793817486,0.5504429952820347,0.3774705858677503,0.8356558746769963,
 0.10489125950187905,0.13660131883154902,0.07105864195846467,
 0.0969704292688448,0.7180046020751907,0.20344722807065052,0.6748883401946099,
 0.4957132095626079,0.10827512887280577,0.23090825868867004,
 0.16223112883068025,0.3664288543831795,0.5896092001232989,
 0.31873708743916995,0.3844706913236845,0.5442868567813067,0.1321629316003804,
 0.10567312409797003,0.45969550562216965,0.7433152059347631,
 0.6707030865661777,0.15849783836786313,0.6177670525633339,0.7258217812494469,
 0.7817720979871133,0.3677872337124437,0.6590870851272752,0.151047447117866,
 0.13467583216362344,0.731257469590001,0.6154349820259222,0.8728080159104246,
 0.06763676470753577,0.9413210622983127,0.7518048965211075,0.7155183656532138,
 0.22550393386235756,0.14257610728180425,0.7874501375055999,
 0.9551147776693212,0.09848946805191428,0.9697886301817109,
 0.36706254686705564,0.2721900568691712,0.7127244103507725,0.68068920105177,
 0.3599760325256548,0.9663913875208663,0.33645103025443457,
 0.45254014200374093,0.9422671176019025,0.191598116320017,0.8005340336708342,
 0.35911390527832765,0.9301359171105982,0.09770279203160193,
 0.13744899496905594,0.20607646833702475,0.8435652614583311,
 0.5638855696746446,0.6636777097974341,0.15180801053123982,
 0.16545310591531648,0.950544755514596,0.9190344566004534,0.9346985327888662,
 0.42273976218145504,0.46940370331004266,0.10504920861675948,
 0.9849444778139955,0.6484864606896665,0.41181481405781595,0.8297331666242739,
 0.9609976527521781,0.6543078230188393,0.02340307273609976,
 0.35759437649235426,0.7335426372313123,0.11692868414548818,
 0.1902734566332407,0.23301260842171456,0.31034323640323835,
 0.7595440633623278,0.8851322462491611,0.559775315530987,0.2471570290025168,
 0.7812789895697677,0.5041972720487209,0.07366627486140309,
 0.33119523838736376,0.271800974464832,0.08336689374189021,0.2582614519370151,
 0.27959632845444815,0.3856116640734,0.35711375660149036,0.0770979657440003,
 0.04684943566646793,0.7645560751762696,0.052755492265956994,
 0.12996793018170738,0.6173073217021566,0.6042823949248064,
 0.11501088612633614,0.9557444308822443,0.9047833008086559,0.6547628817428552,
 0.46195249299624463,0.7314801294201925,0.8790004942420759,
 0.24260140504364258,5.729236961451445e-4,0.7393900940566841,
 0.4914192451507772,0.6790316402331312,0.9018207775923335,0.1870921262669174,
 0.7896322627855352,0.6810243458727562,0.48489833148125383,0.5116063075675217,
 0.6310419179839557,0.05054219560220452,0.3546450516438904,0.5659460884928028,
 0.2820209418777998,0.7657333366673785,0.9057247298826803,
 0.035709102546221394,0.9017098763176635,0.41194537246882945,
 0.5639638048237307,0.864281099771164,0.17929453642700632,0.6863699548394714,
 0.6872082956931826,0.7077670308472356,0.30291835374767895,0.562276776481406,
 0.931373413988869,0.37107351073769124,0.9582245773652811,0.6027584900790799,
 0.03512626098629856,0.6298005190195917,0.04451587628121301,
 0.7983824037041145,0.15736969926884936,0.9516448919424692,0.88920579165959,
 0.5277016854108821,0.26813782401558683,0.6195312099309337,0.4267183870249641,
 0.9328717630936001,0.9173768410007951,0.4857853843357687,0.07572619830516436,
 0.8193028856460763,0.27116058631148765,0.12241692174714902,
 0.5680034693526409,0.8398713638649367,0.031282898293746864,
 0.6206493041337289,0.08206076827212594,0.8007657694643049,0.559224755905596,
 0.849405049864319,0.34608569296920055,0.38880999929053406,
 0.04137172939932454,0.014657541821619269,0.8510893157496233,
 0.9425894503101256,0.6878946145572126,0.05471809907901504,0.5419110407367972,
 0.837712331674327,0.270226808053044,0.1171712481730165,0.537644679258074,
 0.38461506679400825,0.991185259115381,0.020745599647268076,
 0.017562613571567298,0.1825018638742577,0.7753562264987163,
 0.48342888524767924,0.44519400930714914,0.40345711131580986,
 0.3822685769412989,0.4333312694434601,0.44311068490091143,0.7578340979874667,
 0.43823175964344063,0.2318354647212304,0.8462154633913432,0.5483861105493502,
 0.12566037064114355,0.9614063181857748,0.9683571369246532,
 0.09748688495797375,0.7443314889485986,0.13230850130948346,
 0.3835892378336997,0.49959634817633825,0.2685238439286929,0.5338291087029194,
 0.8982701624179079,0.24345871740696823,0.768576883885899,0.8632830361647876,
 0.18987529150394222,0.8034930469614062,0.5459550364514083,0.7957795616517676,
 0.036004727393749714,0.38140023517612864,0.632587006097581,
 0.3070134178653001,0.17590998406763625,0.8564912245981287,0.7270593740254083,
 0.7538269028022037,0.5406341359745934,0.7697633265333563,0.49839600267897577,
 0.8667106145635227,0.6214158508071004,0.8542205055133366,0.7708230359379674,
 0.3176323152986731,0.9283447740605699,0.23022875634876516,0.957836750493427,
 0.13602279643482973,0.8388030159801548,0.07540604698293896,
 0.8356599260109097,0.641418192435899,0.07680891568607928,0.18759188287535078,
 0.2906449457383312,0.5068152141833382,0.5538277310644876,0.438360649689184,
 0.29412377553145796,0.3109995136133036,0.9743293932991589,0.2832340728340952,
 0.7548029450878769,0.19089542576094898,0.4581586167443291,
 0.05408354115970393,0.2846559322717579,0.6969208597538492,0.782911378080875,
 0.7591994013637257,0.16930806612887206,0.06094638504253658,
 0.8638068755856921,0.32934340447315325,0.32070269796483275,
 0.039086223665098174,0.7420560810975148,0.7531846212749885,
 0.07784580585596823,0.42487659459852545,0.7057313394503401,
 0.22713673553991431,0.9105546257349806,0.44009003088732235,0.967827135048293,
 0.6459484707418645,0.3741845562551893,0.09286852637747112,
 0.36146523473126146,0.35900758261010246,0.027034294219992705,
 0.3815135471774028,0.7756958371224535,0.3802018378894173,0.8564825821135326,
 0.1095655037980352,0.5237741200971318,0.33318549451311263,0.5260289260378659,
 0.02096610668544696,0.2355912814822212,0.6245446089943159,0.8079772496146499,
 0.7694489352477057,0.09852901902499434,0.03510311227433016,
 0.7310533710939973,0.21714072961360043,0.051592756709580745,
 0.5382124281803625,0.7473140785335296,0.15469794789705116,0.8099692353457578,
 0.8585428893923174,0.526314253951953,0.07214156647764025,0.9936048462117786,
 0.6108320849586094,0.62776266429011,0.8942121040136556,0.2169478575853021,
 0.8070931894711824,0.7703457627636894,0.08348574099656259,
 0.07003433664387515,0.43276977091513935,0.4080581535010497,
 0.9107643816739643,0.29735595951022364,0.22717696717111457,0.914762647663115,
 0.7404085081278338,0.5535897083967354,0.6966741821499057,0.8923711382140268,
 0.6712710951593379,0.15515754004807603,0.58023737236274,0.48138873776523183,
 0.9338719258821536,0.7329944667456518,0.7824942413344697,0.7777313040073423,
 0.8324404425163271,0.268079491732956,0.9129489096184736,0.8317038817622351,
 0.4256035383310397,0.44789585612490057,0.7936148821894176,
 0.30539289519308954,0.9780500509920368,0.42580999373576545,
 0.29301286169894636,0.4626508032241722,0.7839025451670432,0.9174660307551602,
 0.341404772172999,0.3761356039711381,0.14384307461811763,0.3518344594995626,
 0.6517654438757657,0.2093598192455972,0.8874389355283167,0.4812163200340932,
 0.18917772728490867,0.8289921204569988,0.02346574494785114,
 0.11696195626015826,0.9585551960955873,0.19655905041197963,
 0.6614251092756994,0.8055077244205229,0.7776741223464533,0.6138517517484517,
 0.3961148030831293,0.850008308782615,0.7632024926444803,0.4173462030181261,
 0.8999653154555205,0.06095670225338823,0.33307893149581647,
 0.8005316629959232,0.40722790334460557,0.40539281330018007,
 0.44260190732158455,0.32371711776839684,0.7013723383536776,
 0.10922893183250626,0.10601106241470848,0.5785493720187916,
 0.9423990500607918,0.7646081325346368,0.8435563777385378,0.267240219918667,
 0.878153739760037,0.28383784854112826,0.5259103535377283,0.7006954588241392,
 0.3889329018267834,0.7056597485975249,0.5987789722726052,0.5406780324877503,
 0.7803367216456369,0.7952634586205329,0.9234652878587185,0.05303171026856346,
 0.4654708543124316,0.6744992796977197,0.5288432723605598,0.1642423319664329,
 0.19175469255324618,0.529556058570391,0.06803862674729644,0.823293536057159,
 0.9074179283588846,0.43760852892836566,0.33602773250115514,
 0.8180888193177517,0.9141112602805832,0.053376891294401596,
 0.29551713748042796,0.4716145929920996,0.9075653430368322,
 0.23053818583083208,0.5461548566215216,0.8722479443655548,0.3944712278840067,
 0.7087981918249544,0.9735700662535952,0.665832344336895,0.994143484089101,
 0.5150193542311661,0.34982141738081507,0.1921479154339436,0.53123776073027,
 0.7828275001272762,0.5877706407394568,0.4214914679595415,0.18260834035281137,
 0.4161285655303678,0.739596616231166,0.21956014313317473,0.40476856838057385,
 0.017699058974287363,0.2476723394047946,0.32628006813032306,
 0.3807324558540165,0.267275311261294,0.20989985827979862,0.7049288696625193,
 0.5825902448256821,0.34318778596496025,0.06314839808265926,0.861027279490553,
 0.2348309292056372,0.45071684304492743,0.9477312929816945,0.4883321692934781,
 0.8584932881845166,0.6965045784436896,0.39227406019235145,0.3917618428091322,
 0.8227736501377402,0.6967235558270216,0.3473124584457089,0.36757309067981647,
 0.14470068637106803,0.6710078965654102,0.541693269292492,0.9989635651984431,
 0.9634332723205905,0.011921921951109082,0.36133355254754396,
 0.2995918531222238,0.5176980803700104,0.2041185932881302,0.462514690669767,
 0.49590776490106103,0.9377119528256327,0.04214732781661923,
 0.8491935442927335,0.3587001508693616,0.5523675604786772,0.42004690918956655,
 0.7686803542449663,0.7317193879905147,0.3187659044845573,0.47715162772998876,
 0.9842772201834493,0.36696975695786926,0.3739768737281779,0.8539896656809111,
 0.3299781599970205,0.9416032110421366,0.46007734899536257,0.9842851946730563,
 0.23413809805362207,0.36909760081335596,0.6652132533956455,
 0.6724832566602973,0.3125805614166378,0.4781905248541598,0.8429800124560879,
 0.17141373207419053,0.4824998391753903,0.37288801167417684,
 0.4535877376395887,0.2532968324510221,0.7779998405758743,0.8445828502146885,
 0.07252685359563538,0.7841527690548689,0.3578459944904232,0.9901083144274788,
 0.5713471418195332,0.7405295430237724,0.9312605685679677,0.39146031708100504,
 0.6110233933926424,0.9225530333572882,0.9049225702539063,0.07218918543538089,
 0.5461795454196363,0.7769018451336459,0.4179405791388252,0.3230595397712529,
 0.05999066960651778,0.6186327431976429,0.41068672683585616,
 0.22588077049971955,0.8352393418608417,0.6451950735758285,0.17966392567848,
 0.052100697532510565,0.11049165399257577,0.3707041594211673,
 0.7094734211945654,0.7653323300776641,0.017707926365507665,
 0.5732108090379958,0.507494569440558,0.5422774533317735,0.24818399072331376,
 0.2998743162075137,0.7013269179909233,0.08122549603012374,0.3588099937173024,
 0.9035410993629158,0.6395294629289704,0.9219195530658868,0.22978834511054003,
 0.1830874071037708,0.4085152751240856,0.22308305015797025,0.4336936779966841,
 0.23351262966849207,0.9622673924758043,0.7933337068237183,
 0.25761539068567574,0.04491730567758534,0.4671818100822118,
 0.6488472110133741,0.3957081632166508,0.9753505607622464,0.8092251884532938,
 0.667436249818094,0.6904761850459297,0.4445578523499021,0.950108742021136,
 0.07849618842206862,0.8459634636058132,0.09289673268729282,
 0.6601877327954058,0.4780279575319273,0.19100666692854396,0.8444622313421432,
 0.7613498166894644,0.8763838139023596,0.5219127927875384,0.7392507001280592,
 0.5863793093823579,0.06972780287346936,0.06970921926676465,
 0.8001039081283625,0.387137426158379,0.8058265818299617,0.7398874101018028,
 0.7361007053409726,0.3518238774610589,0.21439378354758165,0.6676009228687474,
 0.10914604661960292,0.7055747902546902,0.6300905353860001,0.8944035052008124,
 0.48604819453843606,0.576628010439203,0.1275612682788657,0.9047797970920526,
 0.6136255613572361,0.787267114847608,0.8808905332388655,0.7679072091011009,
 0.536673584945288,0.28895774326091905,0.08357257257516815,0.695981074355715,
 0.27389037100247327,0.16650884998624627,0.19308574291251057,
 0.9286025698568332,0.15139444441676286,0.6322254885388231,0.6647300146110835,
 0.8457074808602408,0.9276166282939715,0.5206013163250295,0.06569055664555845,
 0.33743647091984774,0.023837188609186377,0.3579952383849446,
 0.9006173456125977,0.1850446750259288,0.7670381462786391,0.196146820016192,
 0.16951325733816003,0.9900794358189768,0.4083847634266514,0.737367704463637,
 0.2929553329120659,0.762131509125615,0.12554906485367667,0.8945741963498478,
 0.5472485928109019,0.9965138662570935,0.03940322470548163,0.8980583241780337,
 0.032001961888176944,0.5194029049520692,0.15132286382065152,
 0.3941412582559016,0.238095444960588,0.6142185296905197,0.6259341393622607,
 0.9831634635411933,0.7745943791953305,0.217502315026462,0.5169079574769551,
 0.04281688245594306,0.06700642173668503,0.3946979813660504,
 0.8237179565553497,0.7072767118409999,0.16035989879856904,0.4420656437430671,
 0.13129575737113564,0.4981783025631421,0.7811748642155608,
 0.06837903895891106,0.4325165524326645,0.02709769531863493,
 0.6584441635745713,0.8423025208363383,0.26921296597540223,0.5688322914443618,
 0.23528499683516657,0.5135516212923678,0.43269066247873655,
 0.6475750332195787,0.7628580136341601,0.038315990913420084,
 0.7432554965276215,0.761795251380921,0.1803289749289897,0.5291434008805821,
 0.8936461707064844,0.8093964287682125,0.0833095049865531,0.5310353216246267,
 0.6510393960607673,0.8107019953212161,0.3048206385529253,0.35520389841579036,
 0.13377056357370165,0.8117386498868709,0.9458201476392712,
 0.012766106942452171,0.7226403521154826,0.22456267802692553,
 0.3422213673855268,0.3870208784410152,0.4431483079688584,0.4311419805620468,
 0.5463137196734453,0.0774894438506607,0.7441762556296462,0.9612611362222678,
 0.12743953994580615,0.1952259657969414,0.7029105105638243,0.9782646292473979,
 0.831754268022008,0.62034544482683,0.008327120361268148,0.8919340422399955,
 0.29828859332110635,0.06991262097877082,0.8189656099813059,
 0.2737424859043258,0.390606922882457,0.554511140632142,0.8935470519490939,
 0.6055470139249974,0.7998527910761211,0.13297844313507445,
 0.12967719907360364,0.7530534076690718,0.07521830685402209,
 0.23442506380655626,0.1530342079501199,0.9479058235706567,0.6395877563840514,
 0.2076808534297273,0.04578467319596058,0.305484863087367,0.17968010753134922,
 0.15045255235947852,0.12737941834306665,0.9252011064878788,
 0.4843226598955056,0.5767799042898002,0.3771214246774981,0.15172447269442702,
 0.2459507676623499,0.15114164429694876,0.809999622985419,0.997462028722102,
 0.20992664320488164,0.550167020986145,0.43216405358814636,0.8961236921619378,
 0.246503251852228,0.1246489214811457,0.2549428621595554,0.962931483700109,
 0.8296542018530294,0.695745649551044,0.2641437681814366,0.05541877868364631,
 0.26420519778272034,0.7190811424276708,0.900850436980715,0.38419373399660917,
 0.5575081773848849,0.30511516289139096,0.9111647615623527,0.6463030371018122,
 0.1575820470681062,0.331965394001408,0.4351961203540171,0.6626014218250784,
 0.811475678063469,0.1084155556746762,0.9344634869604261,0.14278186054956887,
 0.2987147985478187,0.9745910254039115,0.5140006776164806,0.11290678432774637,
 0.29232675592612123,0.4162483183428819,0.24384473826644137,
 0.7090325586130074,0.8185068167270052,0.5803098697499942,
 0.028954160594878875,0.10810096463058527,0.5813345392586527,
 0.29438889211034147,0.2041837476903825,0.4670517089006907,0.6747944642181343,
 0.15701523429247255,0.13600444283031965,0.20564382795911762,
 0.06080785986381032,0.2176047023137534,0.9542203239024456,0.9194314554678249,
 0.10905076567378158,0.40629252217984624,0.199835555478716,
 0.03332494918167228,0.7233076909672498,0.1760749178807094,0.4480480000944407,
 0.5829967179029703,0.11840783600244065,0.6346652289244877,0.7738034933618735,
 0.8530859774562248,0.11887884931922876,0.13680388351430794,
 0.6956746297807439,0.6331954824464568,0.4102142938806108,0.06254290325064793,
 0.5510401032525284,0.23828816414095821,0.48365795298494785,
 0.24961780327894934,0.5462534210180133,0.7027314432236776,
 0.08304150505707097,0.8381225325766006,0.1006452513680165,0.256917879665422,
 0.7547234328921499,0.5931535722451539,0.05562859672938403,
 0.04670309698478459,0.7977596937478263,0.5174151808254972,0.819631748381587,
 0.7921926410300593,0.4058272110170148,0.4862273016942965,0.9589785740014534,
 0.07702551816077485,0.17554158858558055,0.7430656288303212,
 0.6981460811158114,0.3727669634616365,0.5733253268706571,0.4526256785641777,
 0.041302211111150844,0.5715768482388854,0.2829725118680033,
 0.5007024106970328,0.9738653186987409,0.8716977498644342,0.32154948754633683,
 0.8313933109715144,0.2729183378363932,0.6409269963196954,0.3566765073602294,
 0.3718877103538277,0.9141615336850956,3.406882920524712e-4,
 0.4488339785076806,0.7157374300067545,0.9860535789041976,0.9119612905682702,
 0.18468729597176203,0.048742283993241986,0.3584396214650498,
 0.6538567337069583,0.6934101306376262,0.017152179701207304,
 0.007648781354507861,0.07869560796738717,0.021219802364740747,
 0.25915760466197435,0.09351462518286491,0.03271186029417772,
 0.191041452994399,0.19956361208554707,0.14401418604338678,
 0.05654605153591141,0.0401489520233681,0.6661327017843645,0.6607184149946927,
 0.3980467789106933,0.35298175817504296,0.15395531625094894,
 0.46589581308047956,0.24778813818476342,0.2661213130484348,
 0.8702210908210173,0.6489937341946064,0.8002096404634129,0.4972788653619892,
 0.39381143859715906,0.015252142405458913,0.7102253055273562,
 0.38771573460086484,0.18046673850284978,0.6586354013390188,
 0.44813435932939316,0.34744200645319834,0.3217053539911424,0.758333467932569,
 0.7838590591880537,0.07979457213495111,0.3245471775488269,
 0.11338646978411837,0.03425352526054437,0.773360771950313,0.6122322337036239,
 0.12222416699902205,0.35076212104415305,0.8861446898549014,
 0.07897288137433078,0.41335996128744745,0.3876315251926373,
 0.7895656371240087,0.47538626225188985,0.6978686136455131,
 0.32368880441527625,0.910348070836615,0.2768633944517098,
 0.024128587426498926,0.9774680001541484,0.07856511937531618,
 0.5541457672026677,0.3411579808006986,0.7799379233282653,0.7375708967138455,
 0.8305976945457092,0.6646431260326532,0.9735424323678916,0.6803490191176776,
 0.9337523876862621,0.7606776758381261,0.44598977636368164,0.5917576494684718,
 0.42793421347667016,0.3510579500730501,0.6165581914831847,0.8805887257097348,
 0.026236765772950976,0.9532319629220515,0.47273622412135974,
 0.7555865366185746,0.9895220370703646,0.6198685832531438,0.7551805373178284,
 0.8981881698770553,0.09908055745728428,0.5254877954859549,0.8175232822357973,
 0.7657860134611668,0.1515861938055335,0.906958461843891,0.5972546615546119,
 0.6779229670679541,0.1187313622818662,0.3059397944390567,0.20909580858756294,
 0.9829614232244541,0.840283775807297,0.6958668073924049,0.899856157491274,
 0.22855959575953544,0.31553248701571246,0.9471908713479502,0.647676105654213,
 0.004128017698498798,0.09456313219550161,0.34288078368093444,
 0.30249648387962114,0.015564094997345568,0.6000126151695562,
 0.4657633599220158,0.1781200967153217,0.9811425205392533,0.2971324493459593,
 0.20513977452590415,0.98545247667134,0.3973456496424357,0.02977569170155303,
 0.580181299918019,0.7673357583431223,0.37753622876425375,0.06461797960365234,
 0.5354958547809837,0.4389185540616928,0.4000950749428903,0.23095329894738115,
 0.45887027968535754,0.038498073332650096,0.8248256424669097,
 0.28555472425986517,0.2382029819913437,0.6683571658125016,0.6971371617463922,
 0.6826756120494358,0.32253632428458456,0.9214116890141164,0.8649195847868965,
 0.6646904421951528,0.16350254133010245,0.6085880328589095,0.5178007072887957,
 0.14403821673080464,0.830386684934938,0.17474745687241178,0.6166755764429738,
 0.6766575390312602,0.13598328598202425,0.6929773428835,0.877719923169088,
 0.6255708912174603,0.881772417015341,0.7579346581042938,0.4961895276784607,
 0.24999685354005685,0.11453451046828644,0.36756223581886627,
 0.1814661841390185,0.16048867327180494,0.9739581216807291,
 0.057694812817986274,0.6822215998488431,0.020466932865792042,
 0.813699538682177,0.5029479181864301,0.6238160035227059,0.41956558370261066,
 0.17487345611457827,0.11781909317607875,0.6801197433163851,
 0.2445679933607291,0.6076823068266054,0.6884966528656731,0.8774647268271853,
 0.4255948393224648,0.46508339598684756,0.4633102517518821,0.575138449469897,
 0.5151071208326855,0.8231167488177155,0.11656449158778015,
 0.16027682834238166,0.05675593735690032,0.871893178700667,0.6888156954763854,
 0.08809180845413866,0.3503109117739155,0.9561455019940649,0.7379847552763374,
 0.8482162434465369,0.49179957498193905,0.3704458286005512,
 0.34666462340410864,0.5510524391141731,0.1788864848805085,0.2567951130482895,
 0.5883596179844901,0.2722783438085282,0.4704453397464726,0.8620161166963513,
 0.14553378352203628,0.7290208690917475,0.15995946550645346,
 0.21080981783564046,0.32475579496354035,0.5309729177097997,
 0.9022509021717419,0.17091108781883213,0.6557423714178762,
 0.20356819124669545,0.35691883830271953,0.9882065592105564,
 0.8587976307498232,0.5270562157542248,0.5333597000055466,0.42982501189517486,
 0.47582295265022356,0.87424580618616,0.6814307617371909,0.8159492684290169,
 0.339001253867499,0.4158754131723922,0.972169642532427,0.38958078493098125,
 0.20827249509084544,0.5637942441398525,0.8089215021864263,0.1285897449311214,
 0.6642008012095356,0.819362376044376,0.4719955629235375,0.3948197888502336,
 0.7578422513009252,0.29392094285562265,0.9967870394387544,0.6056406503613604,
 0.9134253381749845,0.17064483221586446,0.13189163679996052,0.893468958074473,
 0.31312205869115717,0.24249923598634526,0.9257013105843178,
 0.3027712409306611,0.4589728483434681,0.24171275429066963,0.5131582771223299,
 0.5553659468719957,0.554902638679817,0.3554165746777611,0.03130944422129933,
 0.4808282781658757,0.06510269968271887,0.07997234519189,0.9443332258251536,
 0.8952367657202469,0.23093802282943288,0.9148903737213856,0.301438346600002,
 0.624802368376802,0.22042958272904212,0.38836840166874054,
 0.29278286007414756,0.4456928064740746,0.09144180269067514,
 0.47557348092786245,0.9727993088178823,0.47202236684624965,
 0.8413789512678421,0.2070253269927852,0.6233826832922448,0.6786071239586728,
 0.834489512550303,0.001813566185963178,0.826014311781304,0.977389687162766,
 0.3620589285348501,0.17571988164930952,0.14228487833009096,
 0.5544537767448764,0.779311589648038,0.275666255067644,0.056885954221325474,
 0.16949217931957383,0.10407669386384488,0.9089792903892607,0.859961141773561,
 0.09455919077454222,0.7096497707462313,0.9805027800144428,0.7067422680054819,
 0.8953105608008106,0.46704311757154127,0.8988394806684092,0.1004920982574864,
 0.20497453256302334,0.5094455533459054,0.14620052291054453,
 0.6078163630716628,0.6055516012600444,0.8786750315317456,0.18522650510331806,
 0.9597766002700479,0.6382728816484018,0.7561113053283472,0.03235968698799452,
 0.017647491532939763,0.8010438697362134,0.5835809308957904,
 0.035936608340204845,0.20402249125688998,0.09553469342288323,
 0.8309011920777428,0.41021415932534633,0.3126580360511175,
 0.45927085099773224,0.21218187628770935,0.07104530924156705,
 0.17882946448878512,0.6608947795791199,0.8558604661620604,0.6763334335140079,
 0.23638680694977288,0.28436510723649366,0.0482838067022584,0.701888203580211,
 0.8109295320009171,0.16449391373420752,0.46733421344507753,
 0.4324687674293405,0.6674453886407314,0.47238001309506206,0.5380984974515008,
 0.9172269375313303,0.7780469007709541,0.15283111678583483,0.7955076877129387,
 0.262843250138055,0.8841230615480876,0.6894231021090637,0.2509900657845616,
 0.6183131869584962,0.011954099179973476,0.5432514452557344,
 0.9239026117613799,0.4053439328968421,0.4401716605484203,0.1322858380004942,
 0.8252670207499885,0.30202473880314273,0.8105008470148201,0.9774167979346652,
 0.412440481569596,0.6267354484476332,0.5251773787018542,0.5678141467090775,
 0.8478815711365397,0.007843467333637166,0.3611722694486388,
 0.8176271271497916,0.20896391925288227,0.6300754827150528,
 0.23201535584057698,0.6193974718529514,0.21157686475127668,0.381660621525699,
 0.4618054943356794,0.974358016507507,0.21693692258865382,0.8289728897422711,
 0.9922973767877052,0.601680995012319,0.28141203728454767,0.9010958741703539,
 0.044511701005606974,0.406832650129499,0.4831753352119825,0.4940938847653784,
 0.363609935678433,0.1776709896202585,0.023805677397287894,0.377848769937721,
 0.029435509534200377,0.6916504754900993,0.33970432923893557,
 0.535445197625582,0.6820180384566251,0.9513916884977865,0.11412333170545019,
 0.9448212333396631,0.7194838539831733,0.3941523810307155,0.3082715930152071,
 0.9379278985599994,0.9326649669789335,0.7762861924372434,0.10188212546486997,
 0.027549108577786363,0.018351912620455657,0.2812996048559452,
 0.1238673447910621,0.569762932987995,0.6726169484147094,0.32849725644915806,
 0.3008024217180125,0.5245118854591565,0.6091199281425256,0.9430252436885104,
 0.6791763391081699,0.9791177672392266,0.9527587599747029,0.8292539325745467,
 0.09155054134886653,0.4378969346916348,0.7128294627678848,
 0.026691786391691563,0.26404789160939446,0.3431718732107254,
 0.19981320526608715,0.7513588496314001,0.3743991992622233,0.571892767727968,
 0.6062852893745421,0.8565005131882485,0.16473215753355208,0.5176458412848376,
 0.4836871208584519,0.22893055375939553,0.7588379072504908,0.6456745043065539,
 0.21148744517375273,0.3685353547436522,0.4403220278226295,0.7984229328342961,
 0.9363671035514981,0.9213482720161722,0.2795471896672572,0.27786608473554764,
 0.8034770731290538,0.7372344767982202,0.5460265849036422,0.5526887000847258,
 0.12342022190807067,0.5436324983260258,0.16457278563429745,
 0.7559190167101073,0.27908451049378813,0.22608377376214306,
 0.37315752506249333,0.06503938592348957,0.5662861794004699,
 0.5919391587588204,0.6108135146657165,0.4439215935950269,0.8778329901855514,
 0.7584773367170948,0.5378672829818526,0.17918737982348776,0.8095739054996512,
 0.9757494730863927,0.9187386313561988,0.04550307497831296,0.6226380974159177,
 0.8323519909191771,0.8498238661741274,0.19688804522047754,0.298458550730899,
 0.5269539964281114,0.1784460097020335,0.8055028161421887,0.4214282442335895,
 0.6804005381592045,0.3191749366664376,0.8011023353412827,0.323104187184976,
 0.6034446956503744,0.5976620982707717,0.9041854514032888,0.8993209963535029,
 0.30891750800750784,0.22556066934774177,0.7290884584646382,
 0.7009391965341436,0.986516860866124,0.7540631508464387,0.8147747282649318,
 0.9790331153805376,0.8152854810842594,0.5346219633262008,0.17350714207144224,
 0.8112375664995939,0.9762424824786954,0.8645595261408492,0.7696543918037071,
 0.35459862789752783,0.9883188422734666,0.8547554139505417,0.9488923003743107,
 0.9971687508639854,0.3573417999319092,0.7149144333628883,0.9335794319918247,
 0.25169362673991824,0.9696123505004066,0.6640801043879755,0.4266722240390698,
 0.08533592559612402,0.27090392849187084,0.5389664004601248,
 0.8951474754156602,0.9203419876739735,0.07794644042777388,
 0.26166793482492234,0.803834530711856,0.812813006060654,0.8421670030789258,
 0.9996411041174892,0.9438371813468096,0.20850642951563836,0.483903804785765,
 0.018290278878477673,0.6320159641113375,0.06124659927354459,
 0.6368545542158796,0.32015417857676987,0.6957433416384557,
 0.23426222337135494,0.1258254377098026,0.4116394228304334,0.3769380867244495,
 0.040315777388313934,0.5310756631065812,0.6995198500676815,0.516533947211042,
 0.9887404472010536,0.07090504276193166,0.5237214153977007,
 0.10738509066385082,0.19822985533985737,0.5117407472164289,
 0.2922310229143912,0.687059471683698,0.6529342864378056,0.9483799440237347,
 0.5517084120279266,0.9913802743121589,0.6077922812923223,0.7723366093970149,
 0.2964974488970771,0.8536777327376994,0.8606632502941536,0.18313182069183298,
 0.12247487875011032,0.5753630607947875,0.6257097513630149,
 0.047021764292380364,0.9287363808161071,0.6078056136754484,
 0.6930423876764943,0.25694769296679765,0.8471551568653515,
 0.26493113342123675,0.49474107239418974,0.4087263428744008,
 0.5470025371632999,0.1719662759650079,0.6984467487475317,0.4792151816156699,
 0.1937271040844346,0.47955113731209675,0.6760055282379194,0.7596384987233278,
 0.1961989466898305,0.011789242119678245,0.17284530236649587,
 0.8083408000925699,0.5631871603969486,0.619311269382975,0.08947570638458457,
 0.437879773482335,0.9754033447879634,0.7195915760710787,0.3627087016400955,
 0.9483315910564346,0.8564872959266379,0.3121569983357513,0.28168449627413295,
 0.17130946221201815,0.6877409206766965,0.30406105005463646,
 0.02884281408165046,0.8242991914186868,0.38309370265369425,
 0.4546646114428813,0.6624611264974546,0.7984502961713621,0.636582513927664,
 0.14818862661119359,0.9680385539353685,0.3297993020044062,0.3937122464425421,
 0.45864683220419233,0.2405690130663154,0.1836149968223143,0.9901819826077728,
 0.9954953245527813,0.4236956857394435,0.09330434307967195,
 0.28498297083909807,0.13571744489110138,0.5868611591568561,0.640186634975372,
 0.7314191972633151,0.5136229655132706,0.7019695273379756,0.42639622698375457,
 0.6804905607745175,0.6364380983265239,0.14411866752237135,
 0.22537412851818472,0.48607850481840464,0.9401660361950033,
 0.13858303268531202,0.03206744974239906,0.6965171271310049,
 0.4978436486326354,0.6629917615386721,0.25680677885713454,0.6786915090693373,
 0.20609300151278942,0.055287010316974916,0.2374970993254556,
 0.25696094946613623,0.7742317005845348,0.2294944812425601,0.7967941997141873,
 0.45358036662278,0.5804389537077369,0.3598572619433644,0.5944138517862065,
 0.8962292989255114,0.12070034995933665,0.2613691663458154,
 0.05017270011389319,0.6098787286883709,0.4548968213540352,0.682729155785082,
 0.5741363003268605,0.5906188104834001,0.6906220267683696,0.41309781057212835,
 0.7008370447410448,0.6606015135768261,0.11102740382465281,0.8292767593614304,
 0.44937852201619033,0.946937698757073,0.8377039357783118,0.9593248249625549,
 0.6870852012662736,0.6397044855864249,0.686386593719435,0.9871340994538852,
 0.1062067037610741,0.36486177386672547,0.8849241623200056,0.6722753874166394,
 0.9522592553560318,0.6923196936439467,0.11065940212330194,0.917486139307301,
 0.9422887520798805,0.09502737595322852,0.02899027905638074,0.227992358931415,
 0.7442707796969922,0.951715844548333,0.32895260344948896,0.6269278933895756,
 0.4297156835124454,0.9925570565619094,0.16227343796374205,0.4891842340231142,
 0.7427782406760762,0.5918393171451071,0.9132595849818189,0.5140263752927144,
 0.7220291801433601,0.7627015021858704,0.5508540813053052,0.7417493027757502,
 0.030291734437349338,0.32820161393758107,0.7802021303223109,
 0.8077676745278078,0.039798196369626004,0.8152500106682495,
 0.3858547892203875,0.8659826636477098,0.8281885604426,0.2359471265165587,
 0.6763449426902461,0.3199649138661327,0.19455732841375162,
 0.40318289623865344,0.6157637618359938,0.639554402547923,0.12567448772369838,
 0.11040735619168152,0.7699560346801175,0.08380427273241442,
 0.7054116818822367,0.2044727687752278,0.6161452806017506,0.990057700484729,
 0.024588663892971874,0.9827998399929898,0.07279099551936963,
 0.4388529470090211,0.945347510126014,0.7481759391611222,0.7684326897281006,
 0.6193262614977016,0.6175061983307394,0.02703539579013925,
 0.03587441124333279,0.7757537486675075,0.3366884711112873,0.5914960979788547,
 0.2585600514119796,0.2907372627525795,0.9544254516640139,0.9347355899144812,
 0.7751791394360428,0.18860149899357415,0.7726412293536588,
 0.41098700319154347,0.05987945822552765,0.4646859124533811,
 0.24334604578546892,0.21578698394596285,0.231654738271992,
 0.10569900643968089,0.6934428366949552,0.25531014255543283,
 0.49336896055916635,0.5201440216685456,0.6252246730490436,0.5977674394030152,
 0.5957603341474446,0.6421864601537638,0.09874215168026002,0.815654531812089,
 0.5653270564580194,0.22126621820966508,0.6369172051298452,0.5904832196517824,
 0.1660722650702856,0.9212210816605216,0.4008080306594921,0.32065966884181696,
 0.04625481444675794,0.49673045785823977,0.8397202418507119,
 0.28039038224333657,0.46676653506241095,0.13243872220325292,
 0.03385698269588289,0.06251100995703274,0.7461885487745867,
 0.5614237779569343,0.41349863339880244,0.7299991183568532,0.5100229315600676,
 0.5028900007200835,0.7291103096880461,0.3489379165909685,0.9160254254636299,
 0.06073639238853512,0.4762814194096173,0.2470696700863827,
 0.05541248470649951,0.3557921461473149,0.3146545188253206,
 0.19131379263151904,0.6250614757054623,0.4035130073428762,0.7809530399151624,
 0.10647683828516519,0.126238658152249,0.11752895276170583,0.9321175991224938,
 0.7260834106428129,0.8678978534449319,0.5770850111667014,0.5807512816780893,
 0.5354301404780911,0.8055849704125884,0.8982602214934724,0.9076217957688879,
 0.48932775752533453,0.274242910152907,0.6914693756710462,0.38385943060966043,
 0.3134910131848093,0.35633843000067844,0.8526768966159566,
 0.19744431493300263,0.3769412588827652,0.16093800294855332,
 0.9219048803730037,0.9485355468421501,0.2503100423240463,0.6734094252271511,
 0.028314177325450385,0.5762269333711307,0.9707934728380814,
 0.46869342611616815,0.4037218310990196,0.9337025803473913,
 0.16891243263645994,0.1778142621763672,0.6191376135228833,
 0.0018047469393329685,0.9252016259583398,0.6231321781434709,
 0.14755324918758206,0.2319413257502746,0.9838722554502284,
 0.15416221988007794,0.8866104586538606,0.15951250460378807,
 0.6615314029286774,0.03838550014703468,0.7749723550154166,
 0.05912888100148228,0.6535662005046637,0.0762434204037804,0.6946265010667508,
 0.5222266816926109,0.6010333597304607,0.20306204147787088,0.801193105317244,
 0.9808723524893642,0.2706131266176185,0.6004721664724995,0.41074502452844563,
 0.498647408660032,0.02446889928254692,0.602983080455662,0.7952627661891042,
 0.8220919505983596,0.33607686459484376,0.4990968825736828,
 0.17007570297801022,0.3554243578589208,0.5097494418709111,0.4182330458850264,
 0.13462357126293056,0.9463502376285207,0.7246764936067348,0.4241007422967703,
 0.7376263373929791,0.0507432818831266,0.6387822883056389,0.09720632249278038,
 0.06786992252701529,0.6561912579495268,0.7748486589107986,0.7297801915849251,
 0.4774114434087544,0.39285324963129686,0.6589713007788842,
 0.08751870526212024,0.9369184379406621,0.3114517555190527,0.7438225199022517,
 0.7520099248579574,0.010403793493839464,0.7368655179406937,
 0.43393998017550783,0.1657110646667337,0.2876979850813465,0.9859153865796582,
 0.516366171306512,0.46920642937845713,0.3844308399952069,0.38648557135793593,
 0.567060915923684,0.7323023459892071,0.8036118818679748,0.04961014141941422,
 0.7109345646787945,0.0226780901089636,0.9134642620077018,0.5154234067918086,
 0.5499037435572199,0.591831140722197,0.6852468427044324,0.6890413606898946,
 0.3945330463354342,0.7797763983641751,0.22017979912749286,0.9923185346954218,
 0.6763412820912584,0.005705717882186656,0.066392487378792,0.7865924060748037,
 0.8225779520390665,0.9321340763191244,0.6841132152505665,
 0.023761277070917464,0.946028360005895,0.6845592390927712,
 0.03935459660253471,0.46453134027103726,0.04810806497841014,
 0.6519321325954446,0.8837189184093102,0.6433418980610274,0.4798159929978477,
 0.883769402339013,0.46617315007576465,0.5230588465911219,0.6300189693404528,
 0.8316030858602375,0.4519425616572845,0.17401742219863947,0.8582124080675071,
 0.4352669808120544,0.6537374113299806,0.48974644390455513,0.8855407282979986,
 0.254124084729913,0.9761842046088595,0.15242782768759522,0.5131673066867982,
 0.3817389887701602,0.9904522126522193,0.9613541710497182,0.19995917257967688,
 0.2896501479916107,0.45754884592300693,0.5688363737386394,
 0.49844240797562755,0.48074125679745183,0.8827531378573525,
 0.19332457276188175,0.3389319949734786,0.016722373105054866,
 0.43304598007872563,0.32110012431398904,0.9993589621334168,
 0.9740855077030399,0.6798422575726758,0.6180443671709925,0.9766737783051211,
 0.8537296206445435,0.12330803692683767,0.5370945824956613,0.9165612343847049,
 0.6555004232151417,0.3052571736793309,0.9915686585037399,0.14011506126188444,
 0.24743340523185142,0.8005156075156475,0.9795579037256583,0.8388531919666298,
 0.5006030215723433,0.3122527506305769,0.9739848821319882,0.360423268090758,
 0.7333514729881973,0.5269805091654636,0.15220346462244516,0.6919581042795621,
 0.053472443369798284,0.9975504058144714,0.9738988172868815,
 0.8693980794932148,0.42984004038111023,0.9741855060311984,
 0.17755932991115608,0.4071804827521629,0.5163317114542285,0.7636326583281284,
 0.10605446637082472,0.6971523433806714,0.4981483196080504,
 0.003306017278787876,0.9016235829054386,0.5198729221295528,
 0.28774911533075376,0.6053544240083327,0.8715822340747756,0.1411465713187271,
 0.7765016911877253,0.8383982973365435,0.13823566000165988,0.9185600090014198,
 0.37742178492787715,0.7535440225699145,0.06150372793986225,
 0.2756374083034182,0.8079478868367217,0.6788202419133762,0.5692322128149193,
 0.14096444922769225,0.012495484784271538,0.9235043224120936,
 0.07118077226207586,0.13211888242199976,0.24902796166435126,
 0.7427812575255813,0.506525026469228,0.9339675412686201,0.7184290974172827,
 0.8684090733533782,0.4206708726000482,0.08612319125323331,0.6720924299725319,
 0.8950686893370843,0.6068840256901645,0.18458331041098508,0.8455934134226473,
 0.5781480509731833,0.7536963266884666,0.13677177073634827,0.4805176753409508,
 0.6838375404784852,0.84336253395945,0.29876228079107436,0.35435766244580325,
 0.7156922016233984,0.4794982664353591,0.9948066273670739,0.9366226755551237,
 0.26219159832311223,0.23993618564005847,0.5158557715554096,
 0.6484924973584724,0.646076554828554,0.5711978134341611,0.7827940947613509,
 0.15628187733926568,0.1475479189975587,0.8769475216686088,0.6602480633896346,
 0.8767708869955082,0.8325627311935548,0.7209079895427244,0.3236166258578135,
 0.4868789907600072,0.8571542814254756,0.07456484202224334,0.7243915835397152,
 0.49853761445326905,0.8558447430049162,0.834915681661238,0.10684095681934502,
 0.9175843486122897,0.6480643753126303,0.06400089074822868,0.1900353138408596,
 0.6898175320893447,0.851286010040221,0.6856496710458703,0.3463949746513386,
 0.40187802498915304,0.01574955817783885,0.34313192638119494,
 0.48217844316155745,0.48474747820401687,0.0760759903128666,
 0.41361348700249223,0.3030164816441303,0.020732867473537753,
 0.4692078257338598,0.6601842791910181,0.6962128911274084,0.1550901568707278,
 0.07019909212018105,0.8731407860384276,0.03360155099242845,
 0.7736477478307368,0.7600322334682402,0.6270458661608252,0.7384121826991461,
 0.3743887115687936,0.6110022629552917,0.382939970126993,0.3947666201754578,
 0.4117230605764448,0.7996261586892658,0.80053850076749,0.3541556318112348,
 0.613963905874834,0.3198540284475848,0.743632966568266,0.7681747861533348,
 0.2925495139987666,0.34389664722576363,0.9422166659316462,0.2744588008512143,
 0.06713514717829461,0.3082374287985803,0.4254619386454147,0.236172033099028,
 0.7329285447797351,0.6132899843701451,0.09749292339348525,
 0.12412490855298497,0.6345046933368733,0.08368039907185942,0.975775678697128,
 0.6664288459915352,0.23135123774567068,0.6461364402780644,0.0793329622035569,
 0.8005594690399291,0.16589959473868088,0.2001849938542628,0.6543331376111227,
 0.7613528947582938,0.43930089066640954,0.18298879471322116,
 0.26709961490597234,0.06881660196136585,0.4517819359589883,
 0.6354570077651304,0.40771141410255074,0.26242762621670535,
 0.6624216038504498,0.45173879408305806,0.702536231671895,0.8190077530501181,
 0.23791576733584285,0.6335537035350206,0.8739263111742813,
 0.23993677180956163,0.6767738818885339,0.8326146100072888,0.2007829920299804,
 0.9952455167224785,0.16429022877840804,0.626147271718752,0.410312462093399,
 0.9656852652483477,0.6348905167546669,0.30997839246382564,
 0.21516064925228695,0.7792474181436869,0.29095755546828217,
 0.3220230686769675,0.23794552137370606,0.8789605277823851,
 0.34962311776683064,0.3030606476295844,0.4823051636441974,0.8336157709889208,
 0.2942045879205264,0.7007311867769088,0.27749903542126386,0.741142915741671,
 0.8598207418351775,0.6647341937066717,0.32474636098119525,
 0.19650333873109993,0.9476554144201874,0.809095691511492,0.1208347779964023,
 0.35953531349746903,0.22011650867678612,0.5276299176919292,
 0.9952583265041759,0.590682596675684,0.4614596517260605,0.4691103705504115,
 0.0044340044784362265,0.1849264966130626,0.7461217528332275,
 0.10272921782863076,0.8014785436997358,0.20717247036306152,
 0.9341264144986156,0.3618790551940806,0.6971637480909562,0.740115076614091,
 0.08417053404418728,0.6047760408086567,0.7766924615964355,
 0.40017672629283574,0.6714427209978426,0.5377551534308528,0.4959189860565888,
 0.46127139616959834,0.9686215188153279,0.26710788830629606,
 0.058680803215315325,0.8371660013377882,0.992226131053984,0.2819169513688745,
 0.44005420675393203,0.9181262389051877,0.4625879727574832,0.515089919389109,
 0.74867216775353,0.5648001604990118,0.2077557166122047,0.617268918003722,
 0.34662892317268756,0.6366549657756699,0.13002856702987742,
 0.7412887838409912,0.37607886128007484,0.4294793000796645,
 0.025083117323970505,0.11400208979363946,0.6524635250499741,
 0.04897778321581925,0.9153661992376506,0.8064450640523823,
 0.013723582865504902,0.9818358799715056,0.6781125820103525,
 0.17627374324517675,0.9142279556208703,0.6170984096921911,
 0.24380983910878073,0.7332113086340655,0.8890231752696304,0.2421787442994433,
 0.8611360953667004,0.4051682791516702,0.22917477212056803,0.5719346237970109,
 0.2650134367552872,0.1705474306999868,0.7601874493306778,0.8497692163502848,
 0.16595530990528173,0.6335502386856876,0.9642133858823138,0.6265085252577848,
 0.6617569963512326,0.7060161716109513,0.7387260417911392,0.21091028340974305,
 0.17993880538770912,0.1737518950421315,0.28724376753741543,
 0.9081292255513261,0.8859898166595135,0.13963367607788701,0.5753485135457764,
 0.8659196061309826,0.0492192584201514,0.11784990271084606,0.5331347671095678,
 0.23236635466855704,0.2985494512891528,0.14169778698179203,
 0.5358317386314791,0.9381869475252312,0.1153001329830754,0.8701871489153242,
 0.6796809064789249,0.7065997131119296,0.6759601846264283,0.6264647418151253,
 0.787340373462589,0.9452608129328721,0.4562227694202793,0.7807852305636278,
 0.6785282890653519,0.17467903239385496,0.6320252541156228,0.2931789774381255,
 0.16046008580072846,0.9594078415287668,0.740606512677381,0.1507335056542829,
 0.1564059933466495,0.7187194763080491,0.10734103944676776,0.9084062084920422,
 0.41672437739551516,0.7350553882017806,0.8062438637623782,0.8894095234131786,
 0.8947237195422113,0.5477932335147515,0.21362638361865982,0.6066591018929971,
 0.46486485111725684,0.46708937533835904,0.3637938724507417,
 0.11956164323166507,0.31972082491867193,0.6299759297733015,
 0.10292263703337379,0.5835918824214952,0.16266674538065617,
 0.06179551271709027,0.5433589663027683,0.10316414818426145,
 0.8744726103754935,0.5574914471734066,0.6531678133104762,0.610946296137641,
 0.2349906672681813,0.328302378528071,0.8145924908878902,0.05279746061473367,
 0.9026398267533065,0.5195872601151238,0.8050677484628203,0.4684669775687571,
 0.30630614691275226,0.32173628986113934,0.30196721848188546,
 0.46528654079818554,0.014765585519943447,0.0021311874459467584,
 0.37339200635118663,0.3906605061963754,0.17450641349166796,
 0.3993616581813927,0.6199254042489936,0.4001320339406874,0.24651697298544173,
 0.13911608819293353,0.19623871943122229,0.5204273515797883,
 0.8203701789504689,0.3213148533942354,0.0661133710282904,0.6690212560616497,
 0.10004184774887726,0.9096247028089697,0.5617134910263754,0.1471087296978114,
 0.018829545256785374,0.5023147152398741,0.6364342291246016,0.853296378762753,
 0.017299621744377713,0.47495466627592,0.6979585220170401,0.9088048870691109,
 0.005904641922542697,0.0033988841811295467,0.4464975037480827,
 0.38712370179747013,0.20889222841775468,0.11377060868323996,
 0.4405861183876141,0.861234863595671,0.8030129067710722,0.05779101258963715,
 0.5636459803655701,0.09623355953188106,0.601458457098771,0.5817881745028807,
 0.7762129529047779,0.2295492305182194,0.37475142315023857,0.8962856001010235,
 0.15094275781393118,0.8683857258422476,0.5473968399322445,
 0.01595854760389548,0.7425852554922387,0.8234277528839687,0.9047495891434074,
 0.12655839514955858,0.7753750683200762,0.35583726090141254,
 0.40361606849063003,0.9037795048986903,0.45786275936235377,
 0.7942682416946939,0.18697052334663444,0.21422601449870615,
 0.9554035531284082,0.7370797436134964,0.6173064236346487,0.4680308122357435,
 0.036351225292876266,0.7724407010827662,0.7987725246394162,
 0.6274850644895327,0.23812690753432797,0.2995786992774352,
 0.18769888005220936,0.11254532397928096,0.551047029453115,0.5057317925069356,
 0.6616826661511802,0.5197826152064391,0.41991192547701617,
 0.09252562645916074,0.8169179885175235,0.9726155166246939,0.7566285008409606,
 0.5088452563404565,0.5497078130997275,0.36179750382545206,
 0.14629029505360325,0.9475048427218947,0.532113915237971,0.8973935582721229,
 0.6620562562764498,0.8718455308275972,0.7914492972171401,0.5095149004726952,
 0.13589971234231124,0.6697192068155665,0.2717561666594047,0.2816970066093524,
 0.6062750322780217,0.02619343341221425,0.6419326313631573,
 0.18153196181860487,0.7896123740127206,0.9674731506855554,0.7715414239820942,
 0.3231338106896563,0.2929577678481263,0.5796505719587975,0.5315664919393723,
 0.11072503233121866,0.10806121289135717,0.4585578484628148,
 0.8226409974164742,0.16608796901244083,0.597296978319985,0.6767242543354258,
 0.9722182668239155,0.9551690643259256,0.21905870874073918,0.7902475173131233,
 0.06507017920872926,0.31752691941625744,0.14130064466919146,
 0.037534491429024874,0.7143292960540375,0.07591410587421055,
 0.23314834371535909,0.39329477451413086,0.36000103305944486,
 0.33126725712432115,0.3709751369841383,0.47456264616968125,
 0.5221805706146904,0.05875546904493467,0.9027121486024858,0.291358577896604,
 0.06602756377082442,0.7239995045722926,0.8935972092686936,0.7099949762287899,
 0.3113574611747103,0.888540895625637,0.9580289076591976,0.5320742958317184,
 0.7941238932485905,0.39696717408421367,0.6512976298229528,0.7078665315833379,
 0.8629105922319714,0.8471560394323686,0.17710479492270415,0.9851640084129304,
 0.15376738321580885,0.19462921124514398,0.016147925184554324,
 0.7085325147480609,0.6436738620161429,0.011760986439142251,
 0.5121746010466894,0.3352607414931471,0.5833349357765193,0.3255197105317854,
 0.6477452700220209,0.35687590417438453,0.11429934813788867,
 0.9739035881360799,0.8732923342069238,0.8416244604891705,0.9210883184378662,
 0.35530111768704864,0.4808109094744982,0.8205914702699457,
 0.05764618568389401,0.9622268563921175,0.5030414342547985,0.908404665155364,
 0.014442980301856978,0.5190345316469004,0.10768027559482607,
 0.9662509707934476,0.4268534252535695,0.49397492653496977,
 0.025954045899433176,0.4197386423035244,0.18029921401622506,
 0.9162935030260876,0.623724320512207,0.268700468940829,0.4275635400247836,
 0.9965653985963303,0.6218811652910601,0.11100356584560667,0.6381371345932788,
 0.4366598446572002,0.39015280001764086,0.13424997199964595,
 0.17536506123662332,0.2670721189738432,0.18731665107897522,
 0.38782008692535697,0.035097723748059195,0.97098108989056,0.9164549244078825,
 0.0911108778886276,0.8525495485542363,0.8961743589620471,0.3172539319607055,
 0.8532125670134694,0.49661970879440587,0.35645669225059007,
 0.5660738599040962,0.7604697558452278,0.30714320901448844,0.1841566793893057,
 0.49305931152157556,0.178872702309525,0.8693343776127901,0.19246968827861144,
 0.5751891063543533,0.8111033871643774,0.994008686634353,0.5473866346933799,
 0.22098949918837651,0.8798288101207414,0.9769750591579203,0.9914516742266516,
 0.7730531240677275,0.9897896210281525,0.19795928286405629,0.9776161037418198,
 0.3442190761006878,0.36642415247694626,0.6987372642931027,0.6790728095064257,
 0.8491332352933277,0.8668100481478279,0.8820793950673095,0.25567659100306817,
 0.524375106334027,0.585978434030838,0.9823170230331474,0.8123242693177155,
 0.4694926157010779,0.5644711694863491,0.9419573029739768,0.7971891307183895,
 0.6469992605038885,0.662958125151433,0.37949140249946267,0.10391344498527166,
 0.5960288853183564,0.5932764856740524,0.5095508511936262,0.09826277340780032,
 0.7865956036136909,0.3768045216172149,0.2993339305086228,0.9203050904164943,
 0.648640461998049,0.07877694792526979,0.5371145254860554,0.4850056610146656,
 0.0343576955653222,0.08559658723189723,0.87574483589265,0.8951954999178228,
 0.20297540899021937,0.02928928380496898,0.07205684823298641,
 0.17840990397671397,0.7937275003581756,0.2828715890371416,0.7734027635299995,
 0.01681796462741969,0.07211832618248915,0.8548276597385068,
 0.7447179983834674,0.5225975978891622,0.9419961110072005,0.45182550156712475,
 0.5376154212266315,0.04555197459191751,0.437815698345151,0.3511129707098879,
 0.4261944853573564,0.19055072579254262,0.29929975335571257,
 0.43808010462374214,0.539228770521974,0.6295098128835573,0.5264318504211515,
 0.25340933864033144,0.5044125052775085,0.8888814658618002,
 0.10212783133222991,0.6956007691640128,0.2200536914652894,
 0.17554999145821426,0.568532481519,0.7101987706366755,0.3503283422137875,
 0.7587497499235595,0.013155860790270646,0.1895454869852261,
 0.11435827643519025,0.0290281429303777,0.014924782181873097,
 0.6566063277746729,0.8720189080188337,0.47095612827453315,
 0.018335331923368225,0.8807655679849546,0.7032044769436174,
 0.9930930845864384,0.7945169281942582,0.527670647763926,0.44833822293094705,
 0.7154693728037074,0.9258679929589028,0.763364910826499,0.5262435751191255,
 0.4398836559441248,0.6261461670539734,0.6615355029849167,0.540932005328075,
 0.6591113016334231,0.04744297368935313,0.49243482080487344,
 0.14208464682617294,0.28370652899162163,0.4471681987156998,
 0.4555262403565611,0.9868184603619854,0.7876579024066991,0.7968155317353722,
 0.3421699014388391,0.2542139696362894,0.5408605698089677,0.3417283099448867,
 0.5874720425633624,0.5518643197125683,0.482555522894897,0.0327845085104318,
 0.5140818393452996,0.9010365722756963,0.9590825821699945,0.986958651859964,
 0.6973048758326624,0.3326867983759174,0.5879764647852249,0.6730335956111908,
 0.3087575324419133,0.7550676638238527,0.5837817295538137,0.2783722302785049,
 0.9770029172375562,0.98682690375262,0.6520922625975758,0.14795476939965735,
 0.4348365603348817,0.07861247253165216,0.6119191680596741,0.8633954129962788,
 0.8122580996675841,0.7164012944328301,0.7865476405728,0.12992340082940268,
 0.9523309646980782,0.1680138902655065,0.8251752244202909,0.8820718424425699,
 0.04638039708332253,0.7110232062517482,0.08549963455277865,
 0.5855358216126832,0.8059267376609748,0.24428244549219946,
 0.29657732365289635,0.9509726617474479,0.9665478207683718,
 0.27187421845522075,0.5188115754727708,0.7461839073566042,
 0.21845526332206944,0.0062933001563568736,0.8628694243467655,
 0.5828719025101344,0.5530454357570406,0.8663576259052761,0.8284180471281526,
 0.49430477304723996,0.9749508964627427,0.9229326365633586,
 0.02092162030415412,0.9064087981161573,0.21571713001022408,
 0.3663201635015847,0.4703917055290797,0.9485904464271274,0.03900825698839239,
 0.7580392393812179,0.9223692181444471,0.563302792797095,0.45099297888504175,
 0.8821305245938112,0.604991407800662,0.4360970366627165,0.6663606635513836,
 0.04587916947401116,0.2539832232768332,0.5834299301288967,0.6405761884294606,
 0.9885381962234756,0.9517606417187726,0.3144182983611089,0.5595686276719827,
 0.12338161369027834,0.2380348938536535,0.9473093284868191,0.9451803697905262,
 0.6827663643243717,0.3081806807692682,0.47479106154701733,0.8969913898225601,
 0.035623564410631525,0.2299689491860425,0.5107835179078624,0.802836797811993,
 0.1991492221456218,0.3681911826997428,0.33142958654488475,
 0.033339147858533424,0.670435260284036,0.19635077406878598,
 0.9940286221486867,0.242795694145306,0.5006545816641863,0.5386704768776328,
 0.470818196401494,0.06947434109254025,0.36906737256277644,
 0.09836573187816255,0.49069960889095643,0.688812251465811,
 0.09808000328891775,0.666734039657195,0.6760009801493112,0.4300483342324112,
 0.3071630305854106,0.6870585835564532,0.8210304964883581,0.8908985154151836,
 0.33084393136252843,0.24090915689701775,0.28446699059319336,
 0.45794773200411154,0.11305753431564203,0.29224385854152,0.833574586371571,
 0.10902967165516986,0.38679372906713794,0.5797533863606223,
 0.6961828714455054,0.6157641894639498,0.05587357649031177,0.5625938221871469,
 0.8539138119134341,0.24761228730405727,0.19524722166807817,
 0.2619975998580164,0.6738585978825018,0.2824626465592298,0.5209961103006011,
 0.0676151300532788,0.566853951321742,0.1909028297759623,0.6458693412295313,
 0.934804790381456,0.26397330478238135,0.16017595369228566,
 0.48729300943512266,0.0515532136585688,0.22081994001334004,
 0.23944912379955063,0.3866302757591612,0.7072737451695409,0.6428674421783664,
 0.23912447731253128,0.6376516831447426,0.608918303181754,0.5071157348941124,
 0.2042467807151288,0.22657237504496508,0.42087965814546835,
 0.1584557033134174,0.543605419477669,0.409155801820171,0.9529035047389144,
 0.5627658761295239,0.09424937728534954,0.07908525154415624,0.867222366965544,
 0.8219827914303094,0.6587279044219266,0.9300441414603534,0.5949451467678666,
 0.5697828163017573,0.620393421542898,0.06837547976587044,0.8430897452424724,
 0.6798350304008192,0.4493597810570832,0.6844963707186987,
 0.010062142349000713,0.49845106957423246,0.8068999036809297,
 0.029440379091424385,0.574797616437108,0.1859926197087881,
 0.42685837524437864,0.8055724064896613,0.8257197389960017,
 0.14305902763822909,0.503738031974918,0.35009815351737816,0.2520641815960415,
 0.2023002005945468,0.879994192200849,0.3955327207673499,0.39935601153277567,
 0.5905033604658142,0.9594293127078916,0.8596301234272496,0.3743314669976253,
 0.5294544851263217,0.16876309592921301,0.6868522015746145,0.8463978285556444,
 0.25386810923069747,0.7123337199330252,0.1301829389109268,0.5704201675114897,
 0.3253751201399171,0.5323218933725544,0.11575817785855191,0.5970730655759612,
 0.9538815526799095,0.09801160036324841,0.6214402321200124,
 0.019581370753574467,0.6278654118147089,0.4319918149717531,0.64531262940143,
 0.3240828778096204,0.956796419352842,0.2352913029950956,0.04635670900258759,
 0.682092116345882,0.4621564703597315,0.9170648995921715,0.44684370028273745,
 0.5950999904015564,0.13462427744990746,0.9480174024038862,0.9078897046803293,
 0.6041934732493742,0.3861342454024853,0.485380679575413,0.7316948027281246,
 0.3301720792690568,0.797529709939869,0.4118024001925331,0.6908949366092968,
 0.2639418477164064,0.8756772205299285,0.6843323630445672,0.303504814303566,
 0.43453849007997747,0.9131361210104662,0.1764233951877432,
 0.31502537631951655,0.5779334974791988,0.23302701848464702,
 0.25731570738718246,0.43578227544739534,0.6921313693320473,
 0.7029843777444209,0.3434460553968892,0.920939350078392,0.07318933507154968,
 0.7568143175618013,0.11086461488224841,0.6452431057186421,0.6743619815213557,
 0.4201360437126703,0.6707305042750759,0.14617420017748362,0.1977640414054217,
 0.49594579914526804,0.09192907261424432,0.7361622273706174,
 0.3213888914973807,0.5723551035994723,0.4029212504615838,0.6181524449938574,
 0.7929302538700663,0.14642306587538068,0.8540563056954924,0.6863998581737618,
 0.6832848608259555,0.4178754607272632,0.6198936949915241,0.24726099933894685,
 0.06712571937041378,0.12288142914816835,0.7947522257972524,
 0.34172720269081003,0.8301594242773795,0.611597264182641,0.4465035495418529,
 0.2247938271661183,0.19821863366336112,0.4911760039434474,0.692589962679845,
 0.661061424846966,0.3929692577532564,0.8279863571889493,0.34650593901498683,
 0.2849530790050908,0.0975399303893143,0.44105778554662267,0.1950934619605681,
 0.5325423523589883,0.10956732429529037,0.45739475627183845,
 0.12156623014530266,0.26416406159303774,0.8166224620865828,
 0.5404980690658563,0.21256451642146523,0.5804903600108186,0.8311212653211396,
 0.999808287734197,0.17297450277877746,0.598800539776694,0.2843550885187405,
 0.7108478937383585,0.7781707353485116,0.4434034864094082,0.10241005025850314,
 0.22318603452328767,0.5729507970536915,0.20274221154267735,
 0.6490683956210741,0.825035526165577,0.6035681761510252,0.24114072410568488,
 0.27488740101736664,0.4974369198817048,0.7503774720275658,0.1498959403914728,
 0.9179810717412011,0.5109188222266052,0.6598702458461214,0.03673246638162231,
 0.12080063168512312,0.7195334964049535,0.47184475681496396,
 0.20884157838344863,0.1602241012895389,0.48878554560967746,0.836046235666878,
 0.2833306555328844,0.22045066513531353,0.2172982455187531,0.7590590754460953,
 0.2957709637122061,0.29402717339994244,0.2871375979843399,
 0.35252239817182807,0.8454355162053044,0.7771826187767938,0.3282338226867866,
 0.39514637633548255,0.06019923951075912,0.7945026459674437,
 0.9449135676200464,0.7656099798275501,0.42193035229430376,
 0.09892799573430655,0.7270353438008799,0.8125174508504751,0.6951064256836698,
 0.9497961986194083,0.12118424141747808,0.29735178593092537,
 0.13072998849856676,0.9920223522307408,0.9941070515015357,0.4470934844843941,
 0.21995471303448222,0.2868860732951741,0.4932086532850314,
 0.37397678505026766,0.41589930682276977,0.3469782470512677,
 0.8311114865111956,0.6451937204086285,0.32490651204203147,0.3636923014307306,
 0.07995408186437047,0.16335821791102234,0.09362096718651336,
 0.3847944183732426,0.5681762466755309,0.2045346273688794,
 0.019199052618045664,0.758356275960344,0.22044504696624956,
 0.06688112308160032,0.3289983093642387,0.1285143438418498,0.7725120929828039,
 0.6420164385943731,0.5336226791571318,0.5792956356347245,0.21393342989607045,
 0.009622939617164095,0.6437356261002366,0.4570793587864499,
 0.8890717151832727,0.1720750612543509,0.2724405127689471,0.8112614819257197,
 0.05208325381666268,0.48812536619744695,0.9503062444557435,
 0.18215998986380666,0.3930710152864426,0.1953501909819284,
 0.37680582703271215,0.13095875914754163,0.7782804508392386,
 0.46651991817644056,0.46677736124050573,0.3288548773170056,
 0.06382212279183097,0.23479137051631926,0.7760513946366437,
 0.6895827597600841,0.38563310189419164,0.65892265157191,0.3263104800903659,
 0.9077562896106216,0.6993936122864611,0.6981856936916452,0.8593587155427911,
 0.3477450314749453,0.14614298951707838,0.2442821640731737,0.3514511603260517,
 0.10817227607817803,0.8671901620775989,0.5306179229017085,0.8387588130582525,
 0.3439441765658793,0.03873774873715552,0.33519009212414597,
 0.5411677962200527,0.027128680204502875,0.16217776730298694,
 0.9242500308255198,0.41235884635550113,0.22366696534871655,
 0.02521993687812485,0.9728586914978224,0.17220056670503625,
 0.34167492179960934,0.3764755832256984,0.4947506545314998,0.3359007309241199,
 0.9962914723627723,0.5719517010877717,0.8961681370178086,0.8236157274019498,
 0.9303725668557075,0.28612653306219626,0.35326458479139544,
 0.07956638034755525,0.7521957990459829,0.15840365771133968,
 0.6487939445520885,0.2834665338400588,0.023149554460864974,
 0.7530780616848639,0.5637467469017337,0.653693227719704,0.26337821615387935,
 0.1632280760217686,0.6766445543165146,0.4222084639171543,0.7046045626948318,
 0.37309907099621675,0.6714454283762905,0.11722768537383976,
 0.6155723243480774,0.7516407254867818,0.24938854205997796,0.9787153243222576,
 0.9086692856757289,0.21703256569065943,0.318428019469426,0.8366608752614164,
 0.37232436752206133,0.30137373028074255,0.29023175036812165,
 0.5407442773820823,0.640460350755022,0.32238985692367916,0.18407633165575155,
 0.047814174658902586,0.33091179603525644,0.4176121286341462,
 0.23397711590078885,0.7497141630897223,0.8240833685436039,0.7750852161164015,
 0.9546385241887321,0.431897022823608,0.9572031476021566,0.8676280802141257,
 0.16507317674846156,0.47325612829936103,0.30123074809629535,
 0.013716616289160766,0.0724723149882387,0.4428160390557898,
 0.1815967188328017,0.28403209624323655,0.8377157675289675,0.6548227264721279,
 0.3424720891860883,0.2588168581696054,0.6942314351125699,0.06894011333944217,
 0.6306102683622182,0.9134983414122619,0.38433690568965395,0.9965114800066779,
 0.9741546518396702,0.34394387387287284,0.8440552046801877,
 0.09777060065454668,0.5178723211075137,0.665770670467722,0.18325395253020627,
 0.7128546332688999,0.8096922480474752,0.7572018800734762,0.08448951633540058,
 0.5184111460501604,0.27128813955907627,0.2008576041543595,0.5934493180948336,
 0.6680538851364527,0.9415296632735389,0.6676309745666478,0.08997719690222183,
 0.021357291501062914,0.9461272400381424,0.8697479009895516,
 0.8581968483713774,0.3185388456307304,0.5697545813630146,0.8018995023868319,
 0.1270911958872465,0.8619450822916424,0.10560024265953405,0.8285343159734986,
 0.800587191531235,0.8964841265299139,0.8372562097895847,0.2799305113532957,
 0.757243587799339,0.24269659030182844,0.9118613131436281,0.9213503655744432,
 0.7211807409955234,0.9568427955965325,0.3857450536968887,0.444828709600244,
 0.31133676738034743,0.18507329886216783,0.9670467807579071,
 0.5865036925773696,0.3119603715711531,0.6224395178915827,0.5753938117787389,
 0.07874778388310277,0.5905570728938424,0.871530454576106,0.533162748304909,
 0.601466147774865,0.16200503408762856,0.05916682896286174,0.6034222237949977,
 0.11122859429150078,0.2625121510322299,0.4946482573870572,0.5162916687709445,
 0.9095642645000953,0.8606659696743029,0.7019385310549585,0.7284054909405568,
 0.3947176827971539,0.9772532140869138,0.38242766944440554,0.5929087213870796,
 0.28786050816379105,0.059732490158617924,0.4744675486225449,
 0.7675136048990086,0.49319984163870145,0.6665963955065668,
 0.010000166763218798,0.17964045335825762,0.905299972502199,
 0.12644579836753778,0.16514708281120383,0.49576878362020727,
 0.5594100910136424,0.159882820921311,0.09845784817091141,0.359000995120507,
 0.7550520158565062,0.3160091281893167,0.7876359545116698,0.9288078339996663,
 0.24799179203617583,0.40517399577048097,0.03887722360572421,
 0.11203035388055571,0.8649316712685721,0.3871273460024949,0.8923687095998086,
 0.6097235417628415,0.4409324884517225,0.734129482722491,0.4891705205198017,
 0.8191340740671026,0.18604275892720323,0.8181510841327266,0.9531992734069828,
 0.1395598372611554,0.4993116147559866,0.5511719439984493,0.1703483955224201,
 0.0021776894367844246,0.33080846991625734,0.19082777002632223,
 0.6426480192186869,0.10258914680344222,0.4779607125903065,0.3645922223118151,
 0.3217678092777386,0.7779968251578682,0.6119351687574077,0.8471230463597469,
 0.9144471594981606,0.8540066804865364,0.2127179496671724,0.7502508552099251,
 0.3296503726471577,0.030593650055701582,0.13525884122368081,
 0.2175827573037733,0.5632971156088181,0.3927352858433818,0.17760750206434506,
 0.7671929509102594,0.05644121070235486,0.5867879305935157,
 0.12694384341384524,0.7431989752157153,0.5339297261510918,0.24551973401603,
 0.20960193105999247,0.5659480126668515,0.9206776519412732,0.7543408837351536,
 0.9735519697812162,0.45047866060579156,0.833641669085792,0.02737109674406868,
 0.43495783521166065,0.5569177276543414,0.162863774759042,0.7967326924867102,
 0.373111578845002,0.34618865740080873,0.8368388886083317,0.4785133938549193,
 0.7338163600561073,0.4533967717236608,0.7749771395704699,0.5899224145178424,
 0.8490988501094149,0.058156637493391505,0.23159646112127352,
 0.5032706780139964,0.024799825704054035,0.37028287732475296,
 0.7962656597914093,0.08656555596299986,0.7556202753290012,
 0.22626104324438034,0.054711375028027476,0.5821868069187931,
 0.6428635731489086,0.9775200057412761,0.36212132247802675,0.9371846370905305,
 0.9065706206296064,0.1570758754987196,0.48149114759819245,0.6976592566155642,
 0.6405332738477115,0.46879718161175155,0.8831281210045326,0.3669230132013409,
 0.24884778516249428,0.9367871977518992,0.25228020824856134,
 0.5513000013410078,0.0851915755951207,0.49133597097889403,0.7921415361615717,
 0.7892719526019426,0.8767603091374333,0.8001328688971381,0.561529049576088,
 0.35428394004556296,0.06586323201665212,0.45348767824729497,
 0.29642022171462046,0.6540894415689712,0.4949012166941984,0.7254294837421091,
 0.6004057481388732,0.40304535589503043,0.8207450429267712,0.9112006077149486,
 0.40858445081895234,0.5950281999158503,0.0642567820304264,
 0.06819487976684824,0.1354117216788664,0.4908298534768666,0.3450013740222527,
 0.5263193834493622,0.8064108003458345,0.5772958371694464,0.6347357299852486,
 0.4975248621745334,0.45741265242113305,0.5336542636165122,
 0.04869413237779707,0.5492483139513361,0.03191101153892528,
 0.5570293757098925,0.6313220246503077,0.12548888008494175,0.7082790597229063,
 0.6035980141342926,0.18986180418950727,0.09636666204935118,
 0.021139336198077174,0.6526972871470657,0.6983245248926111,
 0.9988106577295442,0.17322933397193174,0.5867639470656599,0.5737374286481693,
 0.7838765345724168,0.16362505215182543,0.1108635353457107,0.9780377498995626,
 0.3102402882353025,0.7467730037036824,0.24433038250255845,0.7183826892432784,
 0.7461805425460223,0.5853233503186501,0.6581067188371191,0.8936547810736937,
 0.9598011733562969,0.7225022179993189,0.09899858536798312,0.1155360275837154,
 0.74491358717359,0.8189115654760045,0.09610296138294072,0.22268936393856797,
 0.8148279556099277,0.4843669832741624,0.8008154903601579,0.5202068893034976,
 0.35105875190321445,0.0290960870456074,0.4277634090631153,0.7679892825835254,
 0.042676165842972624,0.7704700261748582,0.3424217368310767,
 0.5836761668884903,0.7372568397368038,0.5258008673766977,0.23977759403413013,
 0.5690813021366677,0.5616030313287159,0.4941541164449843,0.5939086217275267,
 0.43898285638412904,0.5876925633824762,0.6724976159388585,0.9383407654045166,
 0.46940116030241597,0.2244569302624475,0.6983041666710746,
 0.024457934339562448,0.8663459081038071,0.6012663531889615,
 0.01194742970691709,0.23632991116203894,0.03201685394784182,
 0.1985238249597674,0.01799704276674241,0.21377964901082014,
 0.24228713856894635,0.0037154147112161695,0.48541199569317295,
 0.28930265817699663,0.8424206416435115,0.12286829256599685,
 0.10865494745434556,0.7948462494643378,0.04074854546447115,
 0.8318566240953896,0.8492871453068906,0.983874178702956,0.5816379017230544,
 0.5264868685628354,0.395692746032251,0.16982598204907084,0.7818578451029882,
 0.23892018090966982,0.8638147038110129,0.33357575351210056,
 0.4745950789627238,0.7904187689479973,0.3446070146472765,0.271570843062663,
 0.5684298834519944,0.9091427803418807,0.0081949099945815,0.1103193983935703,
 0.48725793242489723,0.9127933050489132,0.9255546453914515,0.7767646718156501,
 0.25684253196567985,0.04646955717369006,0.2614711513285505,
 0.8802991495528901,0.40365351530658167,0.08084181043942573,
 0.5471832650217467,0.28998733119919784,0.32466643299855813,
 0.8701935522560478,0.27527577751754007,0.5453692068581821,0.7811445388398321,
 0.16174216417057174,0.4695925377329385,0.1335863501749336,0.6890210717059829,
 0.39656267708194115,0.5905671049523042,0.49687759147846067,
 0.11852472206540376,0.7653263456519965,0.530795663743065,0.8813171157754947,
 0.5873023274044521,0.7322536921739021,0.2512354797731394,0.8486789582377974,
 0.01594901357337042,0.27843701922969577,0.24301838345577031,
 0.9100592606506608,0.21758367718359195,0.9005606439400256,
 0.43668844902042014,0.011334603268082333,0.057771435933979154,
 0.6883668374329976,0.04637258625145224,0.5292910934194593,0.9600807648373406,
 0.26999616090071477,0.6948460312618431,0.17228736864955896,
 0.5930296447800681,0.09704186047539776,0.6764594659565546,0.6019641474432874,
 0.9045953679282536,0.817523708193019,0.29588149693182786,0.17585988665491337,
 0.6070159725476398,0.7086659302325592,0.6799184455793043,0.2929857565681543,
 0.9079672195535735,0.2374805028295356,0.6092156004323672,0.7545514121506672,
 0.5952566673646058,0.6865583441730673,0.7034531568386962,0.8228764267129436,
 0.5265430844255338,0.7855967403112305,0.6533729344595998,0.23893542407872304,
 0.9432547103221983,0.4163285601054012,0.7873301530580187,0.06828125642676464,
 0.37699104314771514,0.9626826914941201,0.9575794952243322,0.7880691888132406,
 0.0713223390426646,0.8722255796931971,0.7890162800562726,0.8893812375220888,
 0.23078697602001585,0.5884587674700145,0.004543166295619683,
 0.3284767658854275,0.6594178315319354,0.45683129519317933,0.9878317474443716,
 0.6941382687393287,0.6794582081167505,0.8279539155262852,0.14753050889978436,
 0.5607903451087389,0.7906275935883305,0.6111493651331583,0.41050641991187664,
 0.08507689473689461,0.7750619608854443,0.15704122027465295,
 0.3474852668829016,0.26734895331553155,0.13545194270670124,
 0.21125292470840173,0.8870068674745385,0.15766886423417026,
 0.7022805077177187,0.1567064163325057,0.21924242698608876,0.977739519433398,
 0.425804860849756,0.8021540757991968,0.17218192819317946,0.7527924499266551,
 0.9587915727455197,0.0247507253166267,0.4737044551528067,0.9175878343001452,
 0.3666382203597205,0.027905657916620474,0.02469234680165866,
 0.1359912763134944,0.18696689452682091,0.00731663713662023,
 0.5541301897942912,0.03594672132692367,0.39249747636954047,
 0.3337726508977923,0.45664456243378837,0.5901897217042279,0.9327986424546736,
 0.7427404895150039,0.4671002390912281,0.45989145804042475,
 0.20133223785231968,0.6919516385599094,0.38152477726928324,
 0.1102711486267327,0.02043896203478368,0.19618876550745057,
 0.45597810502343006,0.879013568549498,0.9645341445918052,0.7765751856577874,
 0.5633190766083693,0.6876509243193785,0.13121241917785886,
 0.42415103888739547,0.45228617889562406,0.6427545842763941,
 0.40326893275036224,0.7143995409288616,0.6393269506540048,0.3168193329223076,
 0.22688177387936992,9.616933107977621e-4,0.7059546950690465,
 0.2900199033876545,0.11110096387253066,0.8991987262391995,
 0.10947114591102414,0.7390373165351987,0.026285767258006798,
 0.31472478460098563,0.49542400736885306,0.2339357208969406,
 0.4178607296158877,0.9980287867087627,0.32484257129194716,
 0.045089370769729076,0.5229313514289091,0.9183290997527253,
 0.6493616617517513,0.9529266773502851,0.7944569336188033,0.5384555408482317,
 0.19441699326741013,0.4469099125047251,0.028856887406418785,
 0.14929343364552206,0.8105317543846337,0.27528193685333935,
 0.5128660475022333,0.06025078530492911,0.8631764705249587,0.6622533535761179,
 0.1555425691544391,0.1883797271144475,0.9710394703502756,0.7612378824592431,
 0.1407270082873694,0.757377405127118,0.2830255948119955,0.6837550857106858,
 0.8216792896447025,0.5249235115450354,0.49493759760373457,0.2669164629856541,
 0.16973755321044093,0.8996554979795742,0.2905938736326342,0.9647838133768584,
 0.5983578619559718,0.46822929486858333,0.8873162356979728,
 0.17277575718966565,0.5973757580755036,0.41170714281094534,
 0.9183620197962966,0.9576826964118427,0.4944943928582637,0.6380801159434726,
 0.04913990044911265,0.028808497539372868,0.7209851154217373,
 0.9002123083273614,0.7149800211532459,0.0954032492483341,0.5126079554460964,
 0.29762192387314845,0.49597088353535634,0.6050784802987993,
 0.4749775171641256,0.8862121408189532,0.27282447877720006,0.6970272168154691,
 0.42050947128133387,0.6132271814177442,0.8550916987154775,
 0.35056031639413077,0.304363778627296,0.18285122109161056,0.4150976028193193,
 0.6058872012804186,0.41474805078360866,0.6531872719773393,0.5360009056915002,
 0.5497725188715424,0.6178056836701769,0.9233820079005325,0.17873973735197013,
 0.642782016333323,0.030580256452378052,0.48421709026794535,
 0.17495817698553595,0.9983396423448783,0.5388076490762204,
 0.008167856379517557,0.6064949661388307,0.4388742913805368,
 0.8297689735101911,0.9485223147555698,0.8308398850619904,0.8158008291652938,
 0.042447433120183065,0.2593057542390649,0.9207793045253563,
 0.7975701896072132,0.8123855056047538,0.11312996642995066,0.5793796513689323,
 0.5878199515994245,0.17674508187556903,0.7291601452139914,0.8522400013259971,
 0.6202073418185932,0.45648453149145385,0.681943138863888,0.24545180635809405,
 0.22631824723086047,0.17769537907786526,0.7213129017607485,
 0.05026733866597022,0.03135562488412347,0.6641340463898648,
 0.11698107843516037,0.19859450322341676,0.7530689492675456,0.447884095748337,
 0.05582566946823686,0.4232603786114345,0.9372994191483772,0.7606801479068948,
 0.9607950909871397,0.5070212778955209,0.4330754014651149,0.4797859640114197,
 0.8701952734647035,0.7954131935042427,0.5318877354331045,0.525914643880374,
 0.1408921891415561,0.6053005983667569,0.49862963056191056,
 0.14004091399536078,0.6527335464752517,0.5389714267463965,0.5075266410568736,
 0.009166722548943462,0.3733146441737797,0.6229341752366053,
 0.6347427238297858,0.7089709708907546,0.2322582568868995,0.649091296933811,
 0.8049232866150156,0.1068616582649401,0.791124410792573,0.12129280477603599,
 0.604399705246237,0.8227589084079654,0.6049997826473851,0.1991630523613067,
 0.9485394747096024,0.2471596129107433,0.13624455333656216,0.8092266429145952,
 0.13390400059871643,0.10966745736505645,0.859641200830953,0.9592489886082105,
 0.04097459838122108,0.4846723058123298,0.6029168333767383,
 0.020084433968490023,0.5604366641935318,0.014625480355182296,
 0.5482650800989749,0.09970272533761981,0.25219413107686905,
 0.29285197527455387,0.30388711970896987,0.4119093997593156,
 0.8871681161022307,0.18724604431081138,0.4777433206904347,0.7401790178000951,
 0.9381051189040015,0.5226804783510934,0.16966581243830103,
 0.21375847571354312,0.5972409108228517,0.12906127300064152,
 0.29653482108988616,0.6430205053084993,0.06845901298963653,
 0.40399286505126086,0.11717799512706906,0.9296231199100232,
 0.49427354404835655,0.8589425887820414,0.6883379178735642,0.5567761851670516,
 0.7471255356432535,0.5406417513866446,0.31912683339743986,
 0.07915885686763302,0.03750239567223013,0.21607158372534796,
 0.9903158835697945,0.9696214194582852,0.03284321023031511,0.5433980072368055,
 0.24072410123701915,0.7586665742418779,0.46557016691247877,0.349639851994745,
 0.619866072524287,0.20239764106169877,0.47821688683370245,0.7321425254174146,
 0.6788301713353007,0.6791077563631774,0.8618645524696972,0.45153726477794054,
 0.5407646704231506,0.5827938945298432,0.8165757300724199,0.7245247900470851,
 0.708184132878853,0.8898893599853632,0.9106811526415669,0.4345454491488231,
 0.9558344965556185,0.5955751574699599,0.8169338837485096,0.7101285033094957,
 0.5717522473666986,0.028910082636381507,0.4626616464281472,0.38793128358134,
 0.2751311694980022,0.31432175529269946,0.17008417805072407,
 0.9628383743852589,0.6054799975385899,0.8726626126664002,0.2849336197522523,
 0.2643400450232516,0.3333547053334279,0.5212683010161809,0.3515719364715244,
 0.44889229249915585,0.5300147274053452,0.9914939041984244,0.8666551522736808,
 0.3190489794768947,0.113000849979354,0.8759807824698604,0.3629342110479896,
 0.8346996409474584,0.030568595216011296,0.8580547031307982,
 0.9530226261652096,0.798637589717291,0.6182793570926057,0.8399026808549648,
 0.01547012915923296,0.2840031610888747,0.2777479375190194,
 0.24867398151503373,0.8844219221198091,0.5117005053129469,0.5934100358631508,
 0.1437120886819545,0.8897123498349674,0.11640298292900053,
 0.00632861976517296,0.5098914782973651,0.46993064564096243,
 0.9112561956665813,0.3619857353579661,0.628602390031417,0.27110711005380983,
 0.7165849402792787,0.3037596716125478,0.024150634127575787,
 0.048185862019282766,0.8079779441391173,0.12669859068254963,
 0.04344674538538429,0.4864689332816331,0.23133026716411076,
 0.22291397506528599,0.3276558291954754,0.8381789345391157,0.5000031760741875,
 0.9035528850543024,0.12195011290693547,0.8599102069303688,
 0.03387892462222253,0.2588140614654022,0.8606206751453672,0.2304014422700067,
 0.8374532597461097,0.628224407167157,0.14715199000212775,0.307481202571686,
 0.5467217088795681,0.44790961422719955,0.3799045158490284,
 0.20734780020050625,0.14975499635487088,0.12037800443987279,
 0.32524336877964566,0.7601413272326454,0.4616096487962983,0.4203426399062903,
 0.30938122572413973,0.5553164087289671,0.3791600547940587,0.6718794104691121,
 0.11303671745996646,0.576532885699832,0.372848306413748,0.4974239472245596,
 0.8137535135424818,0.58159373533333,0.6879084742373917,0.8328022737883936,
 0.8286019136795081,0.013595206290212936,0.38363848230169983,
 0.7827131755354237,0.7798194370497387,0.3124516585248466,0.14432706435752019,
 0.9846822527040999,0.8691470574346305,0.20803445613266836,0.9777869674022652,
 0.27279469744475715,0.817727818230479,0.41193723411332417,
 0.04584695850502252,0.5516398448458505,0.8767430617608781,0.3526601264019491,
 0.6690087362145283,0.796714427774893,0.34794549995384083,0.49259680967151176,
 0.4455568747199953,0.010837212020521458,0.7586089794497467,
 0.015793135834185956,0.6299686793863785,0.0993561873931702,
 0.9802181983422056,0.8294974007099903,0.36359942948252644,0.9154465716587658,
 0.5298652599275467,0.03465060955214627,0.8438563883347409,
 0.14714019154792168,0.21257975097803627,0.5575491067192595,
 0.5587705725286528,0.5633812795779357,0.2844878096150052,0.1775465155225937,
 0.5427275048795809,0.3660308526338648,0.8543508709776791,0.10249923113602755,
 0.2880803098306366,0.550992424679413,0.3466076711395889,0.5234364952213402,
 0.6849385907009707,0.8296403102459227,0.5482870098612018,0.6657208292984653,
 0.6310317899849426,0.3271239434353882,0.04759519208876528,0.9404808091800659,
 0.873396043234628,0.1054666478822287,0.9878325438800178,0.629980865232656,
 0.4793433677134462,0.46368931710086614,0.45389881635147933,
 0.2791353502739572,0.607033764055403,0.5724309974016975,0.6627552885170267,
 0.8086829738697555,0.323452236259297,0.8682178299376597,0.2001113142002171,
 0.9803931634692844,0.3757799601896652,0.3759995404682507,0.2056743429462624,
 0.5840691915376679,0.21737870691101546,0.9999167013038908,0.3681201249815731,
 0.1476228483122347,0.2084649356163727,0.08893570911196613,0.6168266562145477,
 0.9462807765046455,0.39579771798381014,0.5033148951679932,
 0.03224633593686843,0.7148043237610224,0.2968198222149512,0.2588973196627382,
 0.7377206669130066,0.6074082658593754,0.3923180482746643,0.08352358202254784,
 0.7854649417021362,0.5143408568674169,0.8565858374457419,0.2636591476755582,
 0.03502369837676178,0.5705906620368355,0.20102198433235896,
 0.22392502752080146,0.5042533322041393,0.33188663393877926,
 0.20274729084970278,0.15506996173288734,0.7336443428508168,
 0.3889768963795529,0.5536431609830071,0.203565308601922,0.8910794904193329,
 0.9571004377144832,0.29110051655366764,0.4938950177715543,0.7662253994195396,
 0.07519951342589537,0.7979120682878038,0.5264978129801036,0.6069377277635167,
 0.5198967031040788,0.3824506790573885,0.9959043091658808,0.7569484932613152,
 0.8946829227055315,0.7700419499076417,0.6386661464376707,0.7119438178042274,
 0.19142465626088911,0.003711657194191531,0.6285270212626848,
 0.5030316989592326,0.6365494765891951,0.04440117800539234,0.6314813009428478,
 0.025775927751924055,0.8295659003761781,0.6608236968906924,0.505096145230223,
 0.4849778252458308,0.730741736471329,0.6536109924770301,0.7650262229536238,
 0.8678855264364902,0.27999900141204814,0.24902836846568321,
 0.3088808913438633,0.26752592622407734,0.2439759675048605,0.9877554543982594,
 0.8864206984577835,0.6375546945917304,0.4710968345777473,0.6487936345093478,
 0.16285138346146866,0.3142288673396877,0.4619264632543063,0.4425079533000865,
 0.5707523556801429,0.172744810420162,0.7475656554127502,0.2795216771518273,
 0.9318642909510872,0.7835058920663017,0.18956759763605324,0.5275468991972845,
 0.6849108671440597,0.4296441481654789,0.576538573928958,0.8657918111328341,
 0.832906371100773,0.001313821390413139,0.8177191311211732,
 0.11565072965675205,0.6452457608679867,0.19851957449116608,0.82915293917661,
 0.6545177088117632,0.6082683442517746,0.348597059572324,0.2146798856691764,
 0.11610353728044587,0.641628150488392,0.47534476287819916,0.5918830138039359,
 0.47630618860643426,0.4148148549998817,0.6793290772524403,0.8802304617421622,
 0.3546125545269635,0.14654522904885858,0.36166019985125986,
 0.5610634019235834,0.2985158324115996,0.2875912940457278,0.19256485689543845,
 0.8329790881721904,0.6267440819982028,0.219042419232229,0.2669752960714791,
 0.13144416482362709,0.6612107182881348,0.9386470066612044,0.6940893568237099,
 0.5141403489613424,0.12763217577584818,0.809247159668384,0.20818649088138752,
 0.5435032804794008,0.42770277491443354,0.4535217738186632,
 0.31774572302261095,0.08358342886934123,0.14270619856670663,
 0.8583252439863668,0.8445205368589134,0.9483177788604971,0.4184914700319544,
 0.7791550911595065,0.6712363991358486,0.28381627623418304,0.9422922377590713,
 0.8720211527840567,0.0764514684750387,0.8083858558201543,0.805552399938518,
 0.9588986624027519,0.8778915120769006,0.659252303962073,0.807096233720093,
 0.5424079195086838,0.21519903908430427,0.6540123998623125,0.1909283958121637,
 0.29904117805045427,0.4520269927193509,0.3140822535458003,
 0.35161665505104933,0.4484456536800998,0.006275253415401805,
 0.394302251973051,0.13542985640137106,0.39813689732031154,0.5220793802382571,
 0.9626966780064824,0.4454335627945327,0.8121615132643105,0.30206775961868504,
 0.3666093195659377,0.7447632153776019,0.005519463115695289,
 0.1998438021392892,0.007972178285388543,0.6261664242212184,
 0.11394827445525879,0.07753973820531512,0.32870125288148944,
 0.758789611772983,0.7294985487659739,0.23403148930742024,
 5.246971686881929e-4,0.9251298583921747,0.9947144163353736,
 0.41854976107497077,0.5920004359174726,0.8009144866698966,0.7028977715841825,
 0.1228627158912704,0.6069626034923925,0.25468615416374973,0.6122547820938596,
 0.3559931544366137,0.6996173541890782,0.9747483089395897,0.02342566262689416,
 0.08825675470408756,0.3763590836948122,0.6231118710360584,0.7141934114133612,
 0.6295782204178377,0.8474197830543898,0.6098886062831047,0.6858864440805004,
 0.9691058553764306,0.7686157583032799,0.5038618896566633,0.6015366182567367,
 0.18594324209031954,0.9427401635565533,0.8412735071478108,0.4073074195508941,
 0.17368533857610524,0.8056142712411165,0.48281067204250516,
 0.6922120962828181,0.040016670605610094,0.24618235882649453,
 0.05753838686071999,0.8753795464339829,0.4432304496197803,0.8470543381499959,
 0.9193341650783643,0.1024725798533328,0.7530730882437102,0.9092670546757065,
 0.9621510831221565,0.7477370196645826,0.433893108012779,0.8031547848340725,
 0.8088233665302343,0.5449299668679526,0.6604072950896969,0.4332257858246321,
 0.9050950407051094,0.9207247659909568,0.06917837456454046,0.7352724020727303,
 0.006384931393026871,0.38879797244490355,0.035527076080672404,
 0.6103569815342796,0.03222871683557016,0.4146610716919321,
 0.11338719818905596,0.13655997512837792,0.2797761254817416,
 0.23599420992663622,0.9215200051141884,0.4771348528886342,
 0.22355323408851546,0.8736143891473933,0.06368233574534099,
 0.4126043514676232,0.5435820291965475,0.6147033541980009,0.6760407184898716,
 0.06648996725725387,0.37868543301790336,0.9619075565225199,
 0.2103766989489082,0.38419940046447465,0.9483071031445143,0.6913736828012317,
 0.4468967133711288,0.9742007742879351,0.9038630953667034,0.5087177146775766,
 0.7858004083014473,0.6056490208694371,0.2350279210716989,0.31911631237927673,
 0.46754333817298044,0.8609786360290516,0.40409159667730044,
 0.17275056567426983,0.08934455683408948,0.6446993770983744,
 0.09894702003778277,0.5757827438744041,0.08197069896787057,
 0.3593125953170202,0.4654329926673838,0.0029375141216447798,
 0.41467998678283735,0.059390152383542394,0.9227139168522731,
 0.24764194129700245,0.7146007408998644,0.1393352017697007,
 0.031181855800672897,0.2451953124654359,0.6877497483313214,
 0.14447631102208291,0.689204786524872,0.09396706174569558,0.5026985232674674,
 0.7441020466384817,0.23416632945743965,0.5153072827067684,
 0.11930413887142532,0.7724416425977588,0.9645824490720482,
 0.17920799971986545,0.5056355608434,0.20823936882887772,0.10741044802286903,
 0.5400614747089124,0.7612769938832911,0.1368628791646741,0.116061719065355,
 0.2001183158190475,0.04894975614763708,0.08265691952719101,
 0.9366410779070033,0.8930368878273178,0.006674741894604153,
 0.3346688353170638,0.29094611229507217,0.4650164808927869,0.703790846743559,
 0.049736837323783534,0.5289872499553461,0.7871425502636135,
 0.7916270313045681,0.14795202241258598,0.33545144999234666,0.611661424602791,
 0.24338705270363736,0.9353219443163262,0.930029492515155,0.7988180517571098,
 0.9087778794816217,0.5178249093513574,0.8423869124971828,0.6826533809793669,
 0.9637770620294858,0.5750906944739235,0.7010801963530331,0.8198392594875219,
 0.2815599992350043,0.5209953922644244,0.16402721947668952,0.539751241502076,
 0.3473611456683272,0.2871299206044974,0.7834907728576272,0.5038615762442743,
 0.8708488153691272,0.9074395648899178,0.07215971747450822,
 0.43184870248195684,0.617756047300935,0.6396333312104228,0.9284012161104283,
 0.5978184666352182,0.797724823453874,0.02595166364427992,
 0.013175501750396834,0.13043274729839371,0.9095943205243393,
 0.7184267356927818,0.6140041224772169,0.8641923549557737,0.9572014739052488,
 0.9120381536170913,0.003580624683819922,0.5327550880621926,
 0.36469665434178344,0.23556588296748937,0.3226818712194415,
 0.7521889247918682,0.1691977627384892,0.9125523162573091,0.10435944005474729,
 0.7292204180234327,0.30157939504970865,0.36989247653611956,0.42105474107248,
 0.5390277624774683,0.22178596197988898,0.7557492909365128,0.8463944646802386,
 0.3892730750977307,0.02743846371741321,0.8017022288790359,
 0.09128697087542492,0.6610069899081974,0.314466252588402,0.9123331158580007,
 0.6533850538925473,0.8138869258446397,0.7299944210445206,0.5439567107141577,
 0.1747642545586967,0.11743035664479551,0.5658501865496293,
 0.03558939370567782,0.1926257638387796,0.46657234056611174,
 0.8771122654515427,0.7031185439660265,0.2524535704647509,0.13630931648017908,
 0.29906211766170254,0.1571896991021121,0.13322600985966382,0.968119961700291,
 0.02639899249269939,0.0718687837003974,0.979150439166302,0.8546177655618492,
 0.7361037935113162,0.5198720666272129,0.2600460551533009,0.7075105199496273,
 0.5222507206128097,0.6397257550511997,0.3653834449322073,0.6723356398392525,
 0.29843323853184167,0.542569475552123,0.03846143268099944,0.5262761815450588,
 0.9997695368606129,0.6423662256854867,0.10066286781354661,
 0.11922387863919026,0.03238140829081859,0.3006906723522722,
 0.6078806829157335,0.7876404318546752,0.02766151003050532,0.6805833212529961,
 0.6537975270556087,0.7047344155354166,0.19676648592151258,0.9445387551176089,
 0.08663916933592308,0.26046653110876794,0.4861518695450824,
 0.1850689883135569,0.16001913824287273,0.9261483464360433,0.9945220425875774,
 0.34409063233730663,0.7731397553958705,0.4395116195161045,0.6840866159862143,
 0.9197577651479082,0.8716069267018706,0.5873581896279265,0.36775988104428914,
 0.3745100898097269,0.6765309464713154,0.02573920732225965,
 0.34401039201418715,0.034218796370450555,0.6037782314777681,
 0.3812549621495226,0.16065566204437753,0.7059872844970407,
 0.35137574768810553,0.46061372187806005,0.052460001393066724,
 0.5949444148293409,0.13811547844553684,0.9190289395896015,0.9007310383775271,
 0.9060394286131439,0.2967109444382927,0.9700340507840842,0.2672155880159609,
 0.8264566512623023,0.2919232472267881,0.5969168500803372,0.35745945644854116,
 0.9148867366526975,0.40315872496071326,0.16391740822254797,
 0.23494904007387707,0.5806661831295177,0.7536990159480284,
 0.22617428926965333,0.033746520321980666,0.4097463998591969,
 0.12320880254298738,0.15827771317081096,0.896474150616309,
 0.34382770897374315,0.6845818065299808,0.8368284577217832,0.947707425920201,
 0.18768486305073773,0.46998052436098503,0.47487844006154667,
 0.4365699175828839,0.4345434067575137,0.23095063106174352,0.2126394648942842,
 0.475722201830029,0.5573198414266516,0.8833210160648592,0.16862178474055445,
 0.5150430236825012,0.366411091281166,0.3663003705645954,0.4443219465996342,
 0.8628127550654027,0.8258363505963836,0.5610933586391987,0.2310288523993388,
 0.3320641842053387,0.7180793699174695,0.5761100438958501,0.7080179808710196,
 0.256037600109961,0.29288403535763463,0.6756076547126532,0.6187473070710652,
 0.33217494939657954,0.25645542698360446,0.88274869294144,0.3588080599730995,
 0.11822340575399659,0.18950577562179638,0.4972368689847618,
 0.4643985750701416,0.9655356776047865,0.45487415828547406,
 0.14535457383405814,0.9068487126062986,0.3783505066772609,0.636978194976572,
 0.9967018157176368,0.5206659264171656,0.08206170087694808,0.9157152154307242,
 0.2434296207227553,0.7717402613851133,0.20999162676736816,
 0.15114496134394173,0.03535326477062095,0.2668022944218954,
 0.9524210163870911,0.9699983340762125,0.368276668028212,0.23907445615334333,
 0.4955313592095544,0.7389871056324422,0.43798760573759354,0.5995009265601607,
 0.16704907550829717,0.1691672206184186,0.7354256973818836,
 0.26909905681087287,0.9465814732227099,0.5890018275376355,0.5552461968363889,
 0.3391646485951778,0.1911916616680831,0.7365777013277236,0.7648629141923288,
 0.7027726351485445,0.30902371996878863,0.600811079704034,0.7858057265528511,
 0.2724685271688362,0.14139266581703103,0.756583827959018,0.8015804915709228,
 0.027742052317389798,0.044292480592574845,0.673457787928427,
 0.9157480901283438,0.7748242261434855,0.753759065705635,0.7408690800712886,
 0.23729411336881678,0.8166996647485705,0.5347916415275891,0.7025067705338963,
 0.013485790222402372,0.9217203053766571,0.8597956190696656,
 0.24195069675115077,0.8285855863432795,0.9753026318108554,0.6248301512800017,
 0.2533583632111387,0.43206420750031116,0.8862572988942966,0.7841453769300019,
 0.6976607140761979,0.4114636323621155,0.8659191117772954,0.21835367054812083,
 0.4300953107244526,0.42201215485121946,0.44251461576707807,
 0.8983292539246264,0.3966267947230341,0.5814137754674858,0.33485959921046515,
 0.16488378667269932,0.535077646375402,0.8462381414509133,0.37057253944763047,
 0.7378498545420993,0.5431341362966392,0.43792702197758127,
 0.15136094389972288,0.04234021585853742,0.9009168229101129,
 0.4824930736036721,0.6305305823905443,0.14190590023660288,0.6710532860241588,
 0.4553453703000232,0.2639229162509773,0.9368263642017182,0.9240043476824797,
 0.444016566952511,0.18800879316087615,0.36830943287436313,
 0.11214826259638144,0.2765295888314734,0.7219356910321397,0.7226399825368077,
 0.849756038719008,0.46952008015720803,0.9118941894034275,0.7340236783004198,
 0.5028299173046884,0.5965764809364614,0.11083648095560827,
 0.44091536430435874,0.6264412376237022,0.9527038613533045,
 0.16098390506491722,0.4637586786342993,0.16787976158694884,
 0.3097908512580745,0.7702244166620228,0.9359149110777147,0.878923322523942,
 0.38111665742416034,0.40841063499037034,0.37451638275158483,
 0.40243067842205704,0.8245270983136044,0.8357191748256663,0.5246418153007119,
 0.6000993934134641,0.9761532772737319,0.42748195362162555,0.3997658352712553,
 0.4147204272613465,0.23780478474286082,0.41891906455479677,
 0.2982738363105656,0.13486577107605346,0.4868870255221125,0.8497793056338251,
 0.381952583168779,0.05312645586325715,0.8958982163909491,
 0.013286488077910219,0.30398466307916916,0.23228905230333718,
 0.7432338896505906,0.9398231026464519,0.48241749588817584,
 0.011504523066629968,0.7547191187314939,0.4065073935763497,
 0.5743361657126038,0.7408718515866343,0.7954577961650781,0.26049168144598345,
 0.7914144881970521,0.592002256962059,0.013276131993198437,0.5960029381799878,
 0.3405854284516542,0.7615006045620442,0.8432662195606506,0.5436647952622978,
 0.5572774071447784,0.6578657992037674,0.46308889957369415,
 0.09041262580871479,0.7625177495589968,0.21547872878059215,
 0.13489707135328577,0.22985302368076854,0.2895406481136841,
 0.27490060479817113,0.5810390682974296,0.20478852587432705,
 0.25822252236793164,0.5679204395346101,0.9477424697675044,0.8746763487302222,
 0.16358588406177021,0.928616469634246,0.31343343298370496,0.6776158971637889,
 0.6789283115345826,0.3186963086689316,0.835947348063409,0.4461084682865367,
 0.6158944222723765,0.5196231839791214,0.25627099499242223,0.5692524789314554,
 0.47749783111110344,0.7123915281872932,0.2493217826355767,
 0.09459118350066809,0.18354259635508974,0.3731158688979306,
 0.29798292876228494,0.0076755141482515166,0.21287780338283047,
 0.037532745870080486,0.4372564179952446,0.2228228415678054,
 0.9001368518212778,0.19042492393485855,0.20583244361297892,
 0.06582154745661595,0.9823183608827234,0.0651491548665094,
 0.24150738484244083,0.4527438133392889,0.5890750176149804,0.3896878150055493,
 0.26184508795725814,0.8349364014774703,0.011001286448228398,
 0.032768059247409886,0.5822349272830294,0.1051106062653635,
 0.10990601187086824,0.351667844613885,0.7272104695167967,0.3472545503467541,
 0.8316322121944737,0.18271901577804894,0.38051785633278823,
 0.5437196228491424,0.21810003687803792,0.26911812682591585,
 0.8706706293348933,0.2509670651788648,0.36418698486768697,0.2777084729357244,
 0.9129375608217897,0.02390805197246859,0.9718248238784549,0.2399235458401462,
 0.616315635378726,0.8761805256701933,0.8571512424627061,0.9796821792763457,
 0.21298530612096056,0.28619214511264934,0.07423224456257915,
 0.27324976251996624,0.25605457216949745,0.4264248028444244,
 0.48385876849820875,0.3827966456319113,0.6415983470403548,0.9779884340301808,
 0.0841281406519514,0.34979075163232565,0.6396850362208987,
 0.02412255348043657,0.24894355703664572,0.11716306573599145,
 0.17179502292082915,0.8902393398628008,0.29883083164831814,0.853291398795203,
 0.004188021970080502,0.47703892011056603,0.22295408609151834,
 0.9249269283011111,0.9302979542131891,0.14546707687229188,0.6708868522321109,
 0.5880305040285525,0.5167220055431028,0.12144725471903306,0.5584358803765459,
 0.4903763223241271,0.8806136372092441,0.584668960180599,0.5070999944632849,
 0.007044261145737751,0.6994988130770113,0.375262214081981,0.8216713878054107,
 0.4654844367754998,0.49691391739225976,0.008718749109278612,
 0.3287592200438858,0.3231235147040339,0.2251310598871924,0.8746499685075859,
 0.30552631721424184,0.8339038817293619,0.46812121652410144,
 0.004041919536098892,0.7609438305746139,0.507717499980207,0.4773461676024814,
 0.4967325368763488,0.5648728549820099,0.1723830139732183,0.01142176140569695,
 0.5887258779749464,0.4494858803042753,0.6939824991246833,0.9334852728458888,
 0.17702467639062514,0.6413005979335141,0.23128571254299102,
 0.6833999547695048,0.7763518207735599,0.04246973923103936,0.5329038279325634,
 0.6235836476356645,0.4534923679634427,0.024359274597011815,
 0.05467916260721406,0.23151356073476204,0.017339846563550587,
 0.4086875358657569,0.6898174569173835,0.27982717848777217,0.7844875730925901,
 0.6590178076393837,0.4390386610972423,0.19121453060908455,0.6182820236837933,
 0.5630996526505003,0.9481304541409669,0.8542516844600767,0.8320605654434243,
 0.0608626599196429,0.7943268769043339,0.8272191220572511,0.795016896600359,
 0.0646107909968292,0.8152491499016282,0.8127431424225467,0.13813754127304023,
 0.9029688280269872,0.13324458576110387,0.8289909168011458,
 0.20072305560553294,0.07703187893477237,0.5409130175106203,
 0.02440942613814201,0.48777656795986435,0.8847269676128895,
 0.48091102768302896,0.652570680473548,0.06131967240720959,0.8389637401676298,
 0.9277117252723118,0.30408414052745447,0.4241389067068623,0.3401345225421306,
 0.5913564055390365,0.7562744941926071,0.09787108694191482,0.7964517113106901,
 0.410170073503891,0.04031169244993227,0.24361329041005264,0.8911540829201984,
 0.6383915797971481,0.9499545128784308,0.8362214054177285,0.5543791644071966,
 0.1776753328160695,0.9789333440405684,0.32138281877273545,0.8333993924045556,
 0.16752691900618621,0.20406205148179968,0.929626183195379,0.6633670936683355,
 0.6501190009438956,0.7407801160415377,0.5838899626240892,0.5867086054716529,
 0.7771319132677208,0.3253851670551636,0.7322008184407856,0.767788726246685,
 0.65554923700351,0.22653803662093963,0.019861616152764716,0.5836111089804684,
 0.9374355172200356,0.17800204518250684,0.7153130974426736,0.3075107336720966,
 0.25854957131620815,0.10817233488722655,0.48044707703511014,
 0.8014027718818224,0.8884419572404707,0.09969479940729642,0.6905645846197236,
 0.6179435340330873,0.34016807999805987,0.8917790634133651,0.6824905801005209,
 0.7393661600071694,0.53928510047141,0.5809376996751094,0.8113330731737103,
 0.5545494293772565,0.578609107525391,0.9227586154814961,0.26226495106807945,
 0.8941864075732114,0.15604000817743235,0.3323454164787606,0.6429086450586805,
 0.28903972777834386,0.14100089916242842,0.4926223572171219,
 0.8518124036939135,0.28956762178623174,0.9320722240689807,0.1816853095642762,
 0.2758268750259347,0.2282598179966947,0.43275475122406426,0.2380200928572801,
 0.6129742209476382,0.21088812004726165,0.5501702281687284,0.1879817690793406,
 0.6908044127398821,0.7581344557331394,0.8979671111404499,0.4385103840575332,
 0.37189737799718037,0.7238570076975118,0.33249158133224477,
 0.24608040080135363,0.4153010237856607,0.711791990030451,0.3115701239884122,
 0.31734602281257085,0.5640379014963258,0.44626189283741735,
 0.18201477726459359,0.5167627846428933,0.4132076292451401,0.5973681694735398,
 0.9839412837025764,0.9148968814281402,0.9627686016843363,0.5215220185500821,
 0.5943420899630393,0.5237180331116236,0.9354212271579087,0.1598630899536041,
 0.25674271761828094,0.02310046096028584,0.058847637765304106,
 0.42726435879808033,1.0433960089872851e-4,0.14821201926001748,
 0.24037997389661847,0.5520341135119597,0.6465317349114772,
 0.03894671229941338,0.8932479107659022,0.23928769083272805,
 0.8382146195767233,0.1336941633489569,0.14134602257746076,0.3460064312810951,
 0.7275842508185046,0.5660105703688131,0.3015710621133174,0.2143479248948139,
 0.3731364890813824,0.8235428872734478,0.27547434580296815,0.6011274291365098,
 0.43373921593901,0.5637145188403196,0.5319206170167744,0.23538292140516215,
 0.8235310394324189,0.6337615820027775,0.20060993312708353,
 0.25908375159317054,0.6072088384692036,0.013613900540003332,
 0.39743822580059407,0.534066493142579,0.821878153253653,0.022911706652346986,
 0.5388894328382595,0.44143535636022035,0.5852301443532127,0.7141101504711913,
 0.7020607525423502,0.8967105501702255,0.6238988528691914,0.46536325385254684,
 0.898565140703984,0.3009121536684837,0.6545387192004767,0.010125428500477929,
 0.7336522398116679,0.5412363053175142,0.8426768925326088,0.1753404705340158,
 0.058204678047965475,0.5322757466149917,0.5108015662097641,0.680772186763909,
 0.9790739840642851,0.3191366283505811,0.2277321423465105,0.17253640174376383,
 0.5060485799119979,0.5776953948113237,0.4277412738712518,0.38895222977768407,
 0.840162313255387,0.3843043735407232,0.6537708949799951,0.09379460573176468,
 0.8414842018016131,0.89766104516933,0.760883722064681,0.7053637356526504,
 0.6049887456819213,0.7392725805493348,0.9576518892567079,0.918232335138222,
 0.8846701674026458,0.0878648318291303,0.3312055990008107,0.4848894981237569,
 0.23573939288136603,0.9694334564081802,0.39682881474808906,
 0.6106244148147713,0.2853343045860315,0.2703858343961153,0.3172750245488031,
 0.5665572174312277,0.3594861336317482,0.720609264479504,0.508644528518496,
 0.6793516022032845,0.3963651919386444,0.5922017870177688,0.9559120795773776,
 0.5941730726831759,0.2861451076327405,0.9872061891830007,0.10855560408767262,
 0.21185047126598255,0.3783494335675256,0.35042668257491916,
 0.6346945778263511,0.6623845858110113,0.11152976491918765,0.3685139498660902,
 0.45184487991458944,0.8336623991418739,0.33321250687438975,
 0.9369543651170593,0.9521565758557469,0.7079983635200154,0.24342444352834336,
 0.7220213456370604,0.6635095161536373,0.647111195377455,0.6654590611985631,
 0.8492003674981266,0.8234090439976605,0.9417452581416341,0.8075529226083041,
 0.6112502986582407,0.05660299967348437,0.3440153693737329,0.316579151220274,
 0.47441461618053127,0.1540400024440971,0.9162665995031407,0.1164775394877976,
 0.8932209610573234,0.7015685096986386,0.07426517912712405,0.7920936322536364,
 0.4873509581978184,0.8066984877664051,0.5899478005331694,0.19115513484217672,
 0.7508415205006967,0.9732224613000863,0.9605385056061096,0.21533205493942886,
 0.3011166507976102,0.4177882189184513,0.5694388320835732,0.40670739083471563,
 0.4064195667850087,0.37109388222291406,0.4767874580274982,
 0.15058846714441843,0.4500175997295308,0.019604017985165645,
 0.9749365278728078,0.7099872873991757,0.7773665525423256,0.4945905605592118,
 0.5259146326203045,0.6196140481216051,0.14255271602116848,
 0.05841966180774316,0.7832551973983224,0.9379881834593365,0.930575543851328,
 0.48044434284273496,0.859343954109102,0.4237128316143348,0.6047282882594098,
 0.2677261926649569,0.044111814027937846,0.23800531604869646,
 0.48417658656545115,0.1874620906913419,0.42855340897679894,
 0.2581068916732904,0.6906544732339661,0.13948467936303321,0.3538904662147583,
 0.15967876515349033,0.8313705309950387,0.9220923700271552,0.5729809269126563,
 0.8907798691493494,0.560569282425276,0.7523265131483305,0.8146410303085742,
 0.8950389997083104,0.9604713519730946,0.08299361864461563,0.9666110592030563,
 0.266010538043176,0.16539071524357907,0.34337924960021216,0.3402238153662043,
 0.25128788518359435,0.7117583344103389,0.1497293707563334,0.7713270483613395,
 0.36697478512001014,0.6237976218074857,0.7411717811197961,
 0.22070992192989458,0.2779466891487318,0.22171122306942692,
 0.16359617771564883,0.6557123129020536,0.33028230315909857,0.502674604310625,
 0.048378270196608564,0.6434005315828963,0.2709959210672479,0.881197788884648,
 0.8267291882891444,0.7456096593077046,0.28949598177209024,0.7194881919247688,
 0.762222075441473,0.09095866061166236,0.16820706009837028,
 0.058623013458129036,0.968783633601208,0.8854861217950024,0.5148310790011699,
 0.46098118450589,0.09597073948516555,0.05760529422020755,0.24344172498153271,
 0.9095595456889314,0.2592047052208495,0.20986030354783103,0.3294544667366398,
 0.06392491724138638,0.16581109466356092,0.12622694345698782,
 0.7324059694095139,0.411092923365282,0.2752701728278555,0.2402704864431895,
 0.647679188645358,0.3585935368669898,0.9246528656613361,0.7347251907913858,
 0.09276352695894374,0.3974964796513303,0.24818740712105203,
 0.2487307219826257,0.698407693849271,0.057387040902883935,
 0.25743949831226787,0.10588861381977166,0.975075887559651,0.5297696597777879,
 0.9257624967275507,0.3303170091084112,0.2889375493063737,0.23957630960162746,
 0.5040571825083876,0.7647695205224487,0.3953500000097674,0.4779716274825946,
 0.9872657906437609,0.35117254610308,0.8938513691473087,0.4492503878791585,
 0.3087808845393747,0.18030337142808062,0.223776009133973,0.7618406082067738,
 0.1880472557066688,0.8993425889885814,0.9089526746789721,0.1444433194397946,
 0.6384167522820288,0.7403901881271128,0.39553077594808106,
 0.23370989904124917,0.855096590465688,0.8744550546861906,0.09403372656183151,
 0.32576521505034994,0.12206087712587865,0.42617430404673384,
 0.014677177229152627,0.917993657694011,0.6172080247320151,
 0.48070969214525405,0.28676700793875975,0.8411525666776043,
 0.14359821324037036,0.4644219119661086,0.040136349886819245,
 0.7107367162964707,0.0407922055513017,0.10958942325012577,0.7180009106584734,
 0.6025281042784985,0.3345634619528387,0.1705687798843597,
 0.010511261679864603,0.9701680571031345,0.7071977399342693,
 0.8612315985632654,0.6512628125779416,0.45863367188407467,0.9807895958395703,
 0.044223564006864446,0.694827704704817,0.035666392564013605,
 0.3862583856014066,0.7816855866711653,0.47224520002512616,
 0.16123956946279683,0.21052481450224692,0.6902931713588906,
 0.40087650449842394,0.5491580036111197,0.35219605980360247,
 0.4546448944180683,0.6229450730036593,0.7620028670723189,0.33345429030537543,
 0.5357196209801056,0.3199340131972238,0.5177401124863609,0.6451202613190417,
 0.22180193341217658,0.2839329482362959,0.5306511484764769,0.3278025583193417,
 0.1288276425395738,0.3782742968917142,0.5796151604479213,0.7178765885221269,
 0.6236357622733816,0.20574967013890322,0.6435403020640859,0.4400154853500382,
 0.8878598856080766,0.6450337440937861,0.8761154200419804,
 0.062098706869085474,0.23453518806895524,0.41703084414455993,
 0.38291122288895163,0.31996055625987885,0.2518314216489703,
 0.9932871206147073,0.17251771835389462,0.5969508957000955,0.6219401801795076,
 0.40996980578908004,0.20730456969969935,0.7575933564133875,
 0.3091228320884447,0.13129895352795495,0.7248874478260164,0.4454082454915461,
 0.07464243683540117,0.5697342640712442,0.3860456273270102,
 0.46476004840600826,0.9902090304748278,0.9063761882767929,
 0.16274703675976532,0.213471031039016,0.052390027504051195,
 0.49648340428824267,0.03132918044611921,0.5611080992209614,
 0.5771320556067691,0.15793966173803686,0.5950190242875313,0.9398576042164618,
 0.6077181376918607,0.10200054059687047,0.9347046658753889,0.4251571010787004,
 0.9408824994693084,0.2531993405445109,0.12494186808562435,0.9874629139370821,
 0.21262704308597957,0.3768699513491284,0.06134477667027327,
 0.9979615651663035,0.7363312413214961,0.6321520942401927,0.11683723506123755,
 0.40783606378854065,0.07205009739104073,0.37688265959954326,
 0.9253569454492235,0.7223312868794374,0.16140974807078579,
 0.36317565564956933,0.6980226168337378,0.4013136277390621,0.4565995808921509,
 0.9486791140834705,0.16462476283149363,0.7759020514468662,
 0.16406770442905794,0.5943717744524717,0.5472256976917853,
 0.36099439254107657,0.19610176619404296,0.26317125878758907,
 0.2783535552539007,0.7935167562215305,0.24978097707257518,0.3380824187656184,
 0.9487819277479426,0.257833374084794,0.16389798292085822,0.8680950475683206,
 0.10927032997990738,0.6008900113104061,0.5663356320305908,0.5355334710842306,
 0.29490845970621804,0.2598472317470816,0.12925896273447912,
 0.40104702420445326,0.7579883889895952,0.627712715906691,0.45333484694804405,
 0.08883309965943043,0.4574363341095926,0.6292715566323924,0.4721809593996944,
 0.24883221492914342,0.05773551610385719,0.1828700182706109,
 0.12689502770627914,0.30864880410759565,0.34325856466803184,
 0.19916065619491452,0.7029672501086791,0.14470876031317637,
 0.9593002485661036,0.30541806532884275,0.5558461867466968,0.9953176171525366,
 0.19816347822123714,0.45902838078176855,0.010015393429235409,
 0.06421869146224712,0.046684342312208305,0.3590018867934429,0.69071605467194,
 0.573479424380569,0.5783725198584988,0.47003521764727085,0.624415283006222,
 0.3502445930551955,0.7293404241752689,0.42986282730580205,0.8835733282706212,
 0.8983923942979877,0.7447162825991569,0.2877630677557248,0.7294541974091574,
 0.2339484933565863,0.9452399797755834,0.9425708702391882,0.9502373503143683,
 0.8639184099432616,0.009767719608237191,0.0984350378912694,
 0.21686563945240966,0.9902696616263583,0.4571740382955979,0.9517013059925655,
 0.3503381472943994,0.9143089194236104,0.7124907899647226,0.993240398576057,
 0.7085258979373604,0.16643510904599135,0.19837937124925842,
 0.7358228342378617,0.5480010730707039,0.7333308712382097,0.19567748888415692,
 0.4367145602068855,0.5948276948177328,0.8473496140257566,0.8911746513877326,
 0.5623874068995115,0.22048512189945169,0.22426153152834927,0.871039825453298,
 0.1624643313500853,0.22966878633529597,0.39888303943903836,
 0.4607391122257416,0.3511450452240117,0.45284908535654267,0.5367645513788036,
 0.5927681579080338,0.5042798753339712,0.015156861670780564,
 0.4515847728136252,0.7355046789818918,0.9969803563093632,0.14046476049550805,
 0.2989548356294712,0.9786816737361386,0.8246241791732306,0.5619866827677856,
 0.5898768642235136,0.011138634827587368,0.6684927714122034,
 0.4916999022829769,0.24571039717737753,0.14436587905753195,
 0.6713419342330382,0.9109072520210503,0.725693859147557,0.37357919161260467,
 0.6942759580158122,0.755430396331852,0.9925768936471038,0.27862839333147804,
 0.8125917798014008,0.3972786771811829,0.7344173592098124,0.40314001773847064,
 0.23631782959138004,0.9790250874373224,0.5062872665932561,0.8996697531804942,
 0.049094931408405595,0.8366992966006846,0.14364828350499692,
 0.10013623931297477,0.024306158424250435,0.8597151372737765,
 0.7249185582766193,0.9066417751895546,0.9844000567627033,0.8837547905235225,
 0.12229372591647547,0.2590021434882841,0.39323997669064437,
 0.32996454699181066,0.7406468120713212,0.8827369999850138,0.9193890166749514,
 0.1774983345005252,0.34453493693195913,0.5289577311153572,
 0.04253406447602992,0.7849254662552255,0.6988339228794104,
 0.18467073446484716,0.6487654026463646,0.07209836338679099,
 0.7437507749772505,0.7454450341132417,0.37808837649817606,0.5627733069041351,
 0.8135080127232015,0.8848644799076825,0.4968848918448727,0.34564508285017737,
 0.19455542587292052,0.8729928490146481,0.3593480214482735,0.7744887086114655,
 0.6747629232593083,0.8576911332532778,0.10146870876185232,0.9040945888675109,
 0.6967542147179085,0.6589546371775927,0.5954127704161418,0.1696051661318787,
 0.5928435396012914,0.0033347189957871892,0.1990635735422127,
 0.6122689072039822,0.33697723735103524,0.34770370308823106,
 0.30388680644767896,0.189620792160495,0.886697947314005,0.24549522729628526,
 0.7517756722871108,0.05623400969105119,0.738121339307847,0.8161577304545707,
 0.6154842809906785,0.8034713145764311,0.4368276282056647,0.5249353834277166,
 0.7390425946845564,0.0782845819643565,0.1827233414116587,0.06573889801906785,
 0.5989192308424615,0.15756729003954528,0.8785670845259813,0.7141001403123918,
 0.3643446302760929,0.19562548578610073,0.5072305223458762,
 0.40420118134375826,0.47763026356375027,0.6741875887231208,
 0.9586014506568176,0.04734096327913484,0.023335006042215167,
 0.4300475703166692,0.05169633139481522,0.6833256602124951,0.732103224000185,
 0.6248263478520629,0.3994572689255599,0.7677127719278762,0.1562621427287184,
 0.904913461413392,0.34031520315218744,0.7441408132161161,0.900801185958413,
 0.6479269201943858,0.296164052019002,0.2963000329480939,0.7773033209384027,
 0.9756248489782986,0.44225306279391974,0.03271205832346147,
 0.05395403106572405,0.07527476908665787,0.4958098460035121,
 0.6754333618272168,0.5362380193851146,0.08731330910259139,0.9325146510935685,
 0.9920146780621044,0.2698068378721161,0.5993002171941811,0.38922591596937006,
 0.9407169236324394,0.7247409622635556,0.1909241642326006,0.2513985429611224,
 0.46010919829558494,0.47793310227754826,0.45805886040937493,
 0.46508827731927616,0.8247659517004414,0.7692802551099502,0.6760413488282648,
 0.1897450111866802,0.889627842249591,0.3492681096479059,0.9486492421831795,
 0.7481116020386901,0.7153658763706541,0.2925076170357447,0.4275670196867313,
 0.14535653745704868,0.7617059563331422,0.24439640221906433,
 0.3160242318419235,0.703995446739281,0.01976021430141328,0.01518313060983656,
 0.0475097663408347,0.6871171105661094,0.024686394123317634,
 0.9515814417450693,0.5840280966279003,0.0279784362233505,0.8909038543177565,
 0.46259191763475094,0.5621086927891458,0.34193467958426393,
 0.48284617096878446,0.7753605319557555,0.08679833626070765,
 0.9753645358572604,0.9586022989831897,0.262225803393086,0.018136845170316995,
 0.004476728138514852,0.8308289522954588,0.40697537291325814,
 0.23846521000807708,0.8015984238303808,0.3484533838528363,0.6427406264968744,
 0.8176891665571705,0.3120594460060778,0.01581610516757126,
 0.16844882650109794,0.07562616105308317,0.6001442023260829,
 0.6219492997311544,0.3263620575608287,0.36218630284657727,0.3495891993868808,
 0.5285295621092472,0.26321608847349665,0.8186471436018316,
 0.08688220035207794,0.3374550582894861,0.5905979737035213,0.2968607143700551,
 0.07286338595606712,0.5681466312860255,0.24431785900130165,
 0.7513024852180852,0.6304394820522665,0.6272197399122857,0.4696599675329003,
 0.7176470404951106,0.5039258839132176,0.6730602980249064,0.8456462164613496,
 0.7665210416164548,0.22702360761258278,0.7571596025846311,0.5828578787736363,
 0.8884630077344655,0.6769464714423409,0.34747017844982875,0.3823758308706853,
 0.5581330351121105,0.2061239226867395,0.28130757395841743,0.9198930631282706,
 0.1681662082289722,0.13207075247437394,0.19792158114350844,
 0.2696178429983157,0.38568733477342043,0.5690755608728,0.1683126698448285,
 0.20399081775504868,0.18539286854023562,0.5321667781313575,
 0.7028350894832691,0.957666108303716,0.5303100195269622,0.0956600629435066,
 0.8885815428783326,0.5773576457010408,0.45049855996942956,0.5128282925967285,
 0.8364901715058466,0.4487408832237486,0.3649518118736482,0.8302514017774388,
 0.3565023543448892,0.7495228831242433,0.6364059224108817,0.7472272987288126,
 0.5865205396695986,0.7604653566067507,0.9706604922704296,0.16243386646781188,
 0.36486335914052903,0.34178822350081994,0.6751109126471048,
 0.2758274822757105,0.45463797157704033,0.8976928371449688,
 0.46617136949015736,0.05539003210956983,0.05934849433691214,
 0.5182109822432129,0.7075545938984871,0.6543047671079351,0.26021795966564554,
 0.9258049138930692,0.10102572119796382,0.6489709076844338,0.2154255257523272,
 0.6530283224416308,0.8876140694113509,0.35273931901654587,0.3284510282508789,
 0.969985286033573,0.9395744533487339,0.12163301115518776,0.6063803729818287,
 0.47396775183264206,0.5824398710904205,0.7559968882824913,
 0.49772932570225104,0.6264382586428879,0.21466149570852666,
 0.3073027158546989,0.4875537496010578,0.443907312165316,0.5162837507841965,
 0.39404470302836714,0.4844281289953778,0.5923044789070387,0.26453076652179,
 0.19490887603022744,0.005391154819974653,0.45342453846635766,
 0.47596991373700903,0.08985141591725043,0.7912696980918625,
 0.5367839118071717,0.5900099961408048,0.7444281152394724,0.4274508366318065,
 0.3578884155745621,0.39876036581936836,0.9602863743237133,0.4575528364133723,
 0.3921058228067521,0.36364236360216484,0.2280712322842403,
 0.23263390977053477,0.3710297057596875,0.34159323171788036,
 0.19190159693723863,0.33436068361910476,0.18493857287282212,
 0.07497068129460471,0.22506039895981667,0.22803327938358942,
 0.29054315812054343,0.4280057975040221,0.8046930599630046,0.8519543080769724,
 0.3566741998304417,0.7435204351013694,0.4209273252050044,
 0.020559685340925893,0.21786961039887698,0.1131705504555437,
 0.6096195391009424,0.9999008389065072,0.8154740241401841,0.15031795365378742,
 0.06267118417061412,0.8141531350137454,0.659178729692372,0.4757953716559119,
 0.09773424793705776,0.7062998767706736,0.1732368651369084,
 0.04952404967296031,0.29988121291406067,0.954671632106031,
 0.18008454370230953,0.05961436345743576,0.8679090460427478,
 0.38207603269123946,0.8435944772888048,0.21318697702556877,
 0.6796961191868938,0.46853967376894345,0.27179183173630106,
 0.46705860206307837,0.3325840616983232,0.8323972166813248,0.2682026823644954,
 0.0750810401185632,0.7050535521722281,0.02798047801090109,0.1870292302250791,
 0.7572101906325863,0.8852328256780099,0.5402769383246856,0.30470847000363854,
 0.9423540866623327,0.28370051055006584,0.5005914675638501,0.3899552551698843,
 0.27644466180635374,0.9379297319194604,0.822298825224733,0.7357783065471813,
 0.40030314306590487,0.671079707627884,0.18407851874262082,0.9060407620882733,
 0.03687045750613316,0.9933019566981163,0.9006508316072759,0.5290671734096851,
 0.19017746614203568,0.22321927966414612,0.34109220982541355,
 0.7743892632194233,0.4359816653827222,0.8599101673489289,0.9213210245346879,
 0.4674884797726905,0.03803318406530953,0.014156674014982773,
 0.7177069422617501,0.7766434391677037,0.7698107492527013,0.27177441101027033,
 0.4568224973095647,0.21775049575015581,0.01948182146478228,
 0.6846867165130553,0.054707956884529785,0.584933245107383,0.7308176773886725,
 0.09877690425650765,0.007673519667019901,0.0923780164961675,
 0.5149017900344316,0.3455688370348018,0.7820240392668337,0.2290859070420589,
 0.5098084268548404,0.6558547160639999,0.6179320102870003,0.9692988629900524,
 0.18802998456987796,0.688936780901057,0.7739342485168009,0.46142659481571335,
 0.6151095930715708,0.8184561518136456,0.019507828823760343,
 0.7944916022167005,0.665556666544253,0.3339623709246209,0.4357892089186086,
 0.3476095876693689,0.8988750212414125,0.665770854693726,0.6998344589476563,
 0.14459221421162038,0.4331218806312217,0.122927080539739,0.7379535878211737,
 0.5660263977646938,0.42968173333759174,0.7151097649150407,
 0.15557253582486752,0.49800567710700927,0.8407496249767767,
 0.2547692219836112,0.052128561617000635,0.9337717037183679,
 0.6600663258457112,0.7168184557037477,0.9851746510828364,0.9663865976672034,
 0.21766269060979182,0.6222664360135866,0.669219712811664,0.03173463964785683,
 0.8886414719114936,0.9355506666574557,0.33169962163136824,0.8913616810776506,
 0.3934096249350292,0.5080089359294259,0.36521393348895126,0.707062295752971,
 0.2112237967349605,0.3621239396860949,0.8723860855778464,0.5943462014459954,
 0.9714993695612956,0.809356848478348,0.9583604336175816,0.4143066754992939,
 0.1971928293528724,0.4802248008484442,0.6007334272169075,0.6213980205290293,
 0.8958567253229608,0.13075807703206355,0.7174414526174119,
 0.04315284489578408,0.19817651743646003,0.457693092218708,0.9985107759302004,
 0.0028537385482458877,0.7853914492509725,0.8079124413038596,
 0.5701402792863987,0.769183765428426,0.622820011076463,0.24107819048319812,
 0.512400220145126,0.47107151345591436,0.04983976166455728,0.5360588728767982,
 0.9815550095577894,0.6853046319690475,0.25818092819517147,0.1451879616262246,
 0.6804601972077341,0.2589024960018004,0.1642876434634506,0.9297831643100952,
 0.7758437475630094,0.3015011306102119,0.4762709204589026,0.15630599936032796,
 0.4675956383301305,0.09869002329365106,0.4030498811009958,0.6297615778468192,
 0.9415607695334476,0.09524770572519792,0.8628575124503204,0.7992729250278834,
 0.4450404463939591,0.4476501828670597,0.3927044245019671,0.3994803650604781,
 0.780092676373717,0.06310922954933518,0.3423848929431851,0.8071748279469431,
 0.45719023650285495,0.22673198761638114,0.6870484427909628,
 0.5296050427362853,0.7058592487110203,0.2541878722403852,0.7188131609401659,
 0.9260741112876335,0.07564840143197815,0.6304941879201301,0.927457767979271,
 0.727411534252195,0.7084218436882195,0.19897766388094928,0.47904987075196614,
 0.029134003077659787,0.15200604226214154,0.3624857237462663,
 0.7010281221440425,0.6285924166035265,0.5815074623315366,0.7804767001628338,
 0.006411913797927049,0.7002209269028956,0.5229223996329969,
 0.10541846849805303,0.26870247915243783,0.36643695863771597,
 0.30077586969240266,0.3621124989447738,0.43412557562353715,
 0.4140977911354252,0.3797098868988551,0.4570459977035025,0.5320112654196284,
 0.9064597820188368,0.9077337651458277,0.3632705628072034,0.8134810536923172,
 0.5963601286414298,0.8647160429864239,0.5901937381355452,0.7479413660778653,
 0.06307153115756148,0.37641121315283155,0.29901576116536766,
 0.16416568818823807,0.35225823913101195,0.6324404052659219,
 0.7527748684259633,0.027410501847250357,0.787244021912193,0.3937093080953633,
 0.9486452836069625,0.890491128690472,0.40994050097811596,0.935455911767208,
 0.5222230336331529,0.7481274841099959,0.7848305653177181,0.4037661486132411,
 0.7787271496123751,0.33516164953303007,0.4816941080791941,
 0.29086853044615757,0.45051270075317107,0.7057393151832568,0.807767003774757,
 0.22171550484779923,0.15458883856121053,0.5278239705261084,
 0.24544077959950628,0.9320358022598455,0.2635281701025556,0.9926669278865915,
 0.8972328605793731,0.8074814287985975,0.17142702161237855,0.0895621935565114,
 0.809188756408455,0.9189163114286726,0.7908448884466219,0.1497793213166334,
 0.35713827818902133,0.493091418030135,0.014893610794044498,
 0.9463030235333467,0.33925617737106895,0.4425868800938916,0.5343806083881331,
 0.3918512024753783,0.0029071296764563437,0.8712919241372861,
 0.7512719911903698,0.08430753872149055,0.06350466462146287,
 0.28611813468992353,0.7099850860916529,0.15803889534913684,
 0.8067821604688274,0.03265149253117394,0.4912387769265606,0.836887004891893,
 0.3054132309084159,0.9058466190855338,0.2684113513335,0.4634563483748746,
 0.362073371063109,0.7591505960786655,0.9015380269804882,0.9889494326954917,
 0.6950399438441556,0.31884419101857464,0.3590034876585304,0.9760509279155097,
 0.0902727359796276,0.9408452771050703,0.9623264018332979,0.18479183140744493,
 0.538332488816026,0.5329033840307373,0.6036498048306705,0.4271445271571096,
 0.17491810389294737,0.29032279627196955,0.40977091727942616,
 0.6726772870750225,0.30838959197906657,0.16743087664365564,
 0.22050284783382068,0.05681281392945747,0.8818699909830041,
 0.35716795318700345,0.630979127132508,0.8297606378351439,0.9142750385481315,
 0.5507160866657284,0.6270771512383835,0.838423797537065,0.4461777901410964,
 0.7714589733925858,0.5008133613249854,0.6327611215140245,0.46629810275680594,
 0.4803544923731875,0.34171522102859764,0.22570388576286704,0.379391703496074,
 0.6472264390983982,0.3061538932857374,0.07243328421582285,0.4598682276236967,
 0.5056424211160386,0.01290875766623778,0.9286012205026655,0.8599709531920188,
 0.6708596793866407,0.7597577135363112,0.6737375213365318,0.4223426492520802,
 0.0692233831163912,0.8532922291461724,0.6058522093179827,0.31067168724556593,
 0.5689955573845921,0.9372995313390974,0.16857806947046372,
 0.22445628405724793,0.7757057822179328,0.8510019717795193,0.3601886529181648,
 0.7888395701968012,0.8111940852381658,0.33284968226413814,
 0.28769200802898753,0.6888842429260473,0.14286949417115602,
 0.7454333461445877,0.4213065417727472,0.4419768377419764,0.7452907076898311,
 0.2895430866557881,0.8028670775271133,0.9604849664808651,0.7747333181745111,
 0.6103234593487417,0.31538369725711013,0.9067523903368158,0.4256613607172194,
 0.7878783397114935,0.8662755472258828,0.3251263442326825,0.5083222392776469,
 0.9680913775732144,0.5053995621869543,0.9164468264393311,0.6767625392280913,
 0.7761781637369576,0.6717631963110815,0.6816495206254288,0.4570004613859415,
 0.7088005937821362,0.7206358702235955,0.2889725128350633,0.35632807967360147,
 0.7190633553773047,0.03244189207501391,0.7166403106377641,
 0.48771378718803626,0.9513871744064623,0.5048919563755576,0.8030306477690851,
 0.2140141838599725,0.9197964743097605,0.8749258742614093,0.43702622774479294,
 0.2295885083940239,0.9704319998307358,0.9061336820344938,0.26965456589309,
 0.8809393604893371,0.3394506363222728,0.24107002334329042,0.5259788706335294,
 0.4868045815051726,0.06060087201119857,0.07488280439147144,
 0.7092985499950146,0.33228588397463654,0.5314073307956723,0.7533707265615254,
 0.9796164508011319,0.8454886789746006,0.6731620048072764,0.39509098344608684,
 0.9129509795498029,0.5494074499279927,0.9559235951001367,0.2954497737709607,
 0.8306924950523883,0.6524306695849552,0.09447179836214004,0.9848580175178657,
 0.3851292505579509,0.12402711702602054,0.6170380799381834,0.9609625294869519,
 0.7216072923948509,0.388098612211923,0.7536371360599547,0.42236454160183634,
 0.18494311167849464,0.26503214388676777,0.5768920774590699,
 0.5963071743688642,0.11465804179582739,0.967575766593789,0.8596963015634965,
 0.6112281289659683,0.43116745458017314,0.18452442367002275,
 0.6925516207879924,0.7230265410580698,0.27138856250729615,
 0.40839753678927027,0.9903799880697998,0.08433724510456331,0.63275867455464,
 0.5614584427120297,0.8331801514297558,0.9252316683391504,0.3496518383078133,
 0.2135114133579935,0.9501931389602877,0.7132350404074661,
 0.059280121400105745,0.7581566017012891,0.15052888813877785,
 0.7933158993189031,0.5392997787012016,0.43072243078445405,0.4298716432924402,
 0.6972181618171479,0.47641733459193225,0.5767622249482711,0.3379891126603096,
 0.30972189285800134,0.45507521190457934,0.6563217592046598,
 0.7575640111561132,0.6985823354247276,0.6328046704371306,0.161973718468712,
 0.2775729759578871,0.8572708594951022,0.7961293881872527,0.5571700900881369,
 0.9383354390003721,0.5974264503445381,0.6692396576545384,0.5609166953264704,
 0.30491041684881004,0.7134676941490782,0.4318575173023105,0.180344835316254,
 0.6660557318166801,0.2458652787670833,0.7768289765985137,0.12468661450022578,
 0.6067505804338722,0.8400469994779751,0.6524315190595237,0.1807753298256679,
 0.3724235854909954,0.08890528024032374,0.5374137395237928,0.9456494388615171,
 0.45353532847524547,0.7786603869954596,0.5626733489810727,0.5282995576736726,
 0.07602785417067981,0.8402695838954816,0.7695481985021886,0.7044008987233686,
 0.2113489930096537,0.34694203793576217,0.560684689635969,0.5562359780053595,
 0.6087675851460944,0.26451458625852986,0.6888035405889215,0.9772585603238116,
 0.8743204808155711,0.7742313395449739,0.7237840289488117,0.39522079919298114,
 0.4170232692125352,0.5562298777614832,0.9820592427868087,0.7908453068964147,
 0.8815645102204518,0.18955859304885536,0.37391050426693384,
 0.9492536044853433,0.6952609865579977,0.36755606729812595,
 0.11142005130821697,0.08171520924862363,0.6301492340298536,
 0.33694133919008085,0.5156495378710547,0.33483735664976333,0.874091237315815,
 0.14384800327261826,0.286232506700559,0.5053830929117481,0.1705218505367282,
 0.3969212113588796,0.6130747337788254,0.18652943199588268,
 0.18264972038350413,0.8103823882219191,0.42759732752075164,
 0.37338052762209517,0.15584269595720823,0.025423979230149496,
 0.2446448398955935,0.21382147767856707,0.29278985614057873,
 0.7723796011249907,0.9098732420845153,0.18774842862906382,0.3203712323380994,
 0.7093566068196349,0.407385835593582,0.1480105853609438,0.7286320622998996,
 0.6561381309230685,0.429993462462983,0.4064047227605818,0.12473025224321743,
 0.3304835067105628,0.402297203969062,0.35731004727122195,0.8967522045268937,
 0.7496475339869022,0.29622865462848114,0.5941473911930151,0.6488679492580948,
 0.7588726717049433,0.6383866092802672,0.3550857946793874,0.9301039654490117,
 0.9717882411742331,0.21717416853851057,0.1661082009183774,
 0.24244046530956131,0.756495956584665,0.5733821250021326,0.1262145118602227,
 0.07912300713311504,0.3001915728799115,0.6028563162396942,0.8093706071623968,
 0.5721775567582558,0.31586815494346354,0.6608530350428892,
 0.08935687661679481,0.5424248525159499,0.1377060230388395,0.0689953652400569,
 0.3132794000768513,0.25989101720112584,0.9995327402564061,0.7575045842680683,
 0.6733152939716739,0.4953532362136138,0.4794987203155394,0.07313906659055525,
 0.8996396860989234,0.9044707708458949,0.34593456345927265,0.8207405344793941,
 0.09013728335725146,0.4672410474329416,0.9963300102035613,0.50766575859418,
 0.8920267743195891,0.5109010034901172,0.7866115786953494,
 0.049187457934026235,0.5664791487591748,0.9747487561808512,
 0.25021073877021727,0.6521835087750169,0.3075680601155196,0.7960261212667548,
 0.8338481307597763,0.040499808181232,0.7260465434592923,0.22537027768549311,
 0.6928747876234645,0.8790116446476473,0.3903016531189075,0.9549629098653161,
 0.9062978935322047,0.3398280695545063,0.7255893372357711,0.7511744885107339,
 0.6995051795825924,0.5301818695101839,0.11673505685281182,0.6311444574408066,
 0.22156664162744455,0.2366365101795943,0.03492167315160899,
 0.7642673332734595,0.8173141271129931,0.23790414903308799,0.715245847767913,
 0.3676504262249036,0.941463582361727,0.20631280000863061,0.03995125608382011,
 0.18673596120472025,0.8405720703282492,0.46403600288280034,
 0.2903981996521978,0.46501326504778606,0.9705436865858689,0.2958546776423059,
 0.7758581165644971,0.8061574370465125,0.6321079038162107,0.02852227224487891,
 0.8937066543753961,0.6402792831527438,0.8266721792830837,0.17286908304359394,
 0.5598586188141327,0.2300159993253461,0.7106334495380215,0.799917292474046,
 0.21774187893271968,0.35268370087876133,0.7601876798145556,
 0.8189007856956273,0.9486479443169685,0.6854016437318258,0.7852989497912738,
 0.3057870525439147,0.3008235335702093,0.00889560572061443,0.5157795593021287,
 0.591277408390349,0.30058237481204464,0.9820437014597077,0.34551958149164086,
 0.22904166485572797,0.3475575026525548,0.6510209165062343,0.0982127928630131,
 0.9313775511078017,0.627121269761999,0.20418458116227023,0.4138199052336814,
 0.5635709224999694,0.19918769445191897,0.2869996527945613,0.0409347188834267,
 0.0691779700720141,0.10206373289116011,0.8495374987306101,0.3168789392472251,
 0.936522219665392,0.36945661713597555,0.3541816273969194,0.2932448755053668,
 0.6225274737009455,0.38764269122719874,0.3315053656932654,0.9164540734420495,
 0.3039629936744067,0.06077717552108086,0.4933054971507249,
 0.21312070552887674,0.24656600623342106,0.25258287348710673,
 0.9519026247521996,0.5702919949290393,0.4084702976734599,0.10156048696150133,
 0.5605264335670102,0.7410111419874199,0.9126987163604253,0.8865664973541696,
 0.3199340785757675,0.2895670172744558,0.8496206677843906,0.46691094487303486,
 0.34291877977511676,0.4653011157979242,0.5306002172720123,
 0.007931816533430558,0.5264378718235313,0.8358735519331515,
 0.15020459488754367,0.8684131601576168,0.8148931098255714,0.7474970973968472,
 0.48703487787476274,0.7416107970968614,0.21213222165396617,
 0.7630347095492284,0.851660452482358,0.7682108231261099,0.6609278757409709,
 0.16427495125415859,0.7431055357770738,0.19958717461928643,
 0.3359247749385903,0.7813504557135126,0.24052443820490343,0.7187630182946887,
 0.5475357199571276,0.4296846923962562,0.6255259980296769,0.5603550917293804,
 0.42170450105497637,0.17904942595076112,0.03528090399180206,
 0.2908978116198728,0.47440062564140606,0.48145311097236065,
 0.2532592202928432,0.5957236117156358,0.8343149080073281,0.2473656413016997,
 0.867753299842105,0.2552167352453717,0.2738004428795271,0.09990763399536695,
 0.748082230027178,0.5135488143270146,0.0368642841598954,0.7764605474642882,
 0.6086314487776702,0.7131488364561106,0.5837425849908322,0.06257448633903517,
 0.8100145921918501,0.9183159727142807,0.8494068891426039,0.4199255413882288,
 0.7774043644204895,0.4489574896200149,0.05624965554454886,0.6009085011072531,
 0.01911557848967549,0.35534455408731835,0.9906014945877923,
 0.2929868228790855,0.5326136676646578,0.9812032433874652,0.303866348384194,
 0.866676990956284,0.9444542247905253,0.6030977556602252,0.8994600075291768,
 0.1822394618020191,0.90401716822525,0.8642142346567162,0.8492442400521587,
 0.7952381957952039,0.08883130969293251,0.09990088172895528,
 0.3976630013797767,0.6896980983439531,0.7699926255307037,0.4121461044585638,
 0.5407058715625601,0.1911672163969106,0.5143693152001062,0.8122383152838989,
 0.16742792540371088,0.837779876924193,0.7629378046684979,0.0304399485851794,
 0.6716682394152924,0.06970533750529373,0.13220194968719168,
 0.25579811327942714,0.15072837215062018,0.6550824222216346,
 0.17905573245099016,0.3828232642701864,0.699648496200693,0.01621877630125157,
 0.1742987200416921,0.037774594048749055,0.09857965864716356,
 0.4478484692083783,0.8087104902624728,0.21702503819317898,
 0.36736275563396914,0.5130668494613053,0.6432655457010493,
 0.013392605036036564,0.22414209719144873,0.7849835315449173,
 0.7871267406988764,0.5979938523945596,0.2904584302053579,0.42895636059003406,
 0.936846714424929,0.9567459634711738,0.6942303725487294,0.9552454069365617,
 0.4253823697304675,0.07518054500444249,0.7507344925425932,0.9073091906894876,
 0.6234059493272386,0.8928470279311926,0.8514282719354139,0.28996279597382113,
 0.7096647731130079,0.5102476327908035,0.3174192108799061,0.5838383987221045,
 0.4321365778145906,0.36258202681032004,0.20738969714734024,
 0.11251471202427188,0.8025719181429801,0.04073224229675376,
 0.43233023094208833,0.694855526355836,0.19549393251670788,0.83337487330323,
 0.4231692631868492,0.8715594226582426,0.9633024156252794,0.5988403457159861,
 0.01340714556324396,0.3986040408272711,0.3090706523215244,0.6395690165272612,
 0.9129466858494023,0.44713044869212326,0.5250286216983142,
 0.21271282496578592,0.9083912369048268,0.4857669293468371,0.7715617142352451,
 0.050993974944381404,0.08749126410027441,0.3714054726255952,
 0.4760899796526734,0.30247446864661676,0.3992835586554716,
 0.09959805612092332,0.5037582471615178,0.053391846203929116,
 0.14675344837147586,0.9131386580939171,0.9910040414148599,0.3170190415698305,
 0.4474088085859106,0.44240302234044104,0.5239889933681301,0.6484805730802852,
 0.30937339252859375,0.2958269003851469,0.5503385982082756,
 0.15938774582556758,0.3055241544110088,0.44977887957522245,
 0.46600018496282736,0.14960149186220517,0.2336864276329177,
 0.6609447748871968,0.18402365150523503,0.23511614966799332,
 0.8002283481749646,0.35436067402972404,0.8689338594889293,0.5621704432106264,
 0.5709650049222217,0.9977234708831194,0.4074402159477627,0.10233360566016259,
 0.6554667058440009,0.14481482629002096,0.8290453953496266,0.9599481792990424,
 0.9879634515302016,0.22332974339214673,0.9675468629687533,
 0.018268984769517105,0.6610883342057229,0.25965256258868286,
 0.552197421136492,0.9168629686418204,0.3695478550841812,0.5932611094317758,
 0.344443616813188,0.6231365380904366,0.8380297016058532,0.0457915728207654,
 0.6351060824496231,0.936954824900552,0.20477432772574133,0.12470274203114395,
 0.9450050269513491,0.7547257453024788,0.9801958739719172,0.3004145069421743,
 0.25941843226632944,0.40348217860093505,0.678241449575954,
 0.49608332230692664,0.542207353606878,0.18871926984578802,
 0.22333226363543324,0.9974102455031039,0.5128390535336395,0.8929912047659827,
 0.7097394532963042,0.3922230959724152,0.23171572139508811,
 0.04682420155061162,0.7786511377786568,0.10933069590277145,
 0.3739254696790254,0.17269416637711266,0.7496381752327672,
 0.29019319227779916,0.8035307439790619,0.6025353392026906,0.3336588842439543,
 0.6696429337950769,0.8772254594334152,0.4476003050532972,0.9691020117806232,
 0.07370743861573481,0.9084794626370101,0.37590112611935256,
 0.4218434800117383,0.4458988411015853,0.3897272661110215,0.5573492648671937,
 0.06159370629173411,0.295409351171248,0.2942316911755817,0.9605085401456277,
 0.5503794000056836,0.6987443174460503,0.40429822286465844,0.7487527228952501,
 0.0037717965729238134,0.30815868904712085,0.9031823782832452,
 0.23200157980039948,0.7812919811795026,0.2773255228913546,0.9857942192859864,
 0.3417547503987449,0.12762600194150642,0.8485086540172811,
 0.32765990971120684,0.6263850858410835,0.7504317631142515,0.8667508084763641,
 0.7617479840599797,0.4338286640131035,0.7811070840206763,0.7549847489356117,
 0.6649475911656406,0.4504327066298106,0.8619969394526464,0.6546481298552516,
 0.9621241790049017,0.5428337030891561,0.8189391275433934,0.2530763837958919,
 0.839649777951428,0.5914326380171572,0.5468017848393476,0.9062059409314454,
 0.45175444968581946,0.6821857953950586,0.5874179002886474,
 0.004945300678617981,0.3396099677595581,0.43019265497355563,
 0.08681200597251626,0.1424858247066636,0.8524492678288977,0.6107348737564896,
 0.6868147014016442,0.9631770689911057,0.9139244844611221,
 0.004660708187189333,0.2294648930219223,0.6071996456873933,
 0.13381804297489497,0.08119231201825461,0.14565124229613924,
 0.6756175996943187,0.25774130437454756,0.8425792889028152,0.3411119516623634,
 0.8797974949694255,0.4552124817914044,0.9270345038426696,0.10430681091365968,
 0.3439218264172821,0.40744057699830594,0.8849288315986011,0.8985046808448812,
 0.7677304143051611,0.7710414801832332,0.3086162603712358,0.15082277690124635,
 0.016851641749259794,0.6429241284235999,0.7595352086678954,
 0.4124969540908727,0.41928171625021127,0.4730654527102065,0.9539509448077853,
 0.8481637062218784,0.2890397294838155,0.5402423622233972,0.7106541987417883,
 0.027206996082588253,0.1497091616031757,0.41209346840500694,
 0.06392141915863792,0.3629166476832043,0.42006861462122724,
 0.0189603988841901,0.7759459757359704,0.847283255981621,0.7013416692144573,
 0.675765446112465,0.9673825239498651,0.7107092947134954,0.7674614867652202,
 0.08128666421511421,0.14099819494555632,0.567832035995798,0.8336115234498106,
 0.9665429094796446,0.7991203353705812,0.47628549074555904,0.4871926787387393,
 0.15737115628200282,0.011577067586703516,0.35371809519185926,
 0.3879114788019864,0.9985175392580258,0.6767359252368573,0.3314226926636834,
 0.9056508405390099,0.9624489590994425,0.7626459332185866,0.5470467766154251,
 0.6424938118370415,0.06679825049516108,0.4333127422832518,0.5650837624578144,
 0.34018009392585735,0.6903037296737142,0.38330021960405736,
 0.7174002629923536,0.8137921856602676,0.992894111290023,0.6976945894653229,
 0.8261754589102712,0.5957617139607772,0.6828084485062154,0.7840153232040206,
 0.47117518402444303,0.5089118593812004,0.5756002264338308,
 0.43247621304895734,0.044953599144070044,0.9296156825733725,
 0.48282579818454974,0.5839318976974769,0.1958612372022046,0.8433237476595176,
 0.19567112927866603,0.7858572749120689,0.21144842060183167,
 0.9432036769416708,0.8065512534552718,0.8590242849057549,0.05624235859834803,
 0.19089793423016455,0.3679939496033706,0.5310639726817835,
 0.22150093774348611,0.6900101882860966,0.7452058175500262,
 0.09245964208829827,0.7930111028411463,0.2774148289980529,
 0.07961743965115409,0.5464682462285427,0.2706363892083673,
 0.10940414473780558,0.157876730866225,0.9761701569675197,
 0.025036882842994324,0.05592753068949119,0.6572531125668183,
 0.6105355141930761,0.4508040495719299,0.48307447910829626,0.703540720535291,
 0.8818866175875235,0.6663021151764127,0.6800123226125765,0.8764362248427959,
 0.11045808573033045,0.6326177915916513,0.14762012892147114,
 0.5758071524986319,0.7789313462946088,0.9510820707663742,0.3607455185717674,
 0.48858693124879116,0.4889576152768935,0.10889064508732282,
 0.032586213381354256,0.035608988029348754,0.47787896455784384,
 0.17350938876174204,0.10282486992284967,0.059253223166204516,
 0.6725088801532728,0.45386072095315255,0.7792037059153992,0.809549634054929,
 0.0258408853940989,0.9789606936399126,0.13897205124939316,
 0.14969290276891167,0.44985498711192584,0.5127891679080555,
 0.09819176215592629,0.502264012668304,0.4682897103954854,0.582226701555505,
 0.8656614810096739,0.513450182224186,0.20647610278442696,0.1818692472601946,
 0.15618453808079225,0.6907823433778164,0.3863752068025538,0.1960398382387124,
 0.431451656166859,0.572373379110014,0.4081768479978135,0.7097479427738712,
 0.5358694447889853,0.5135875247259867,0.8053206667275978,0.1092161713528651,
 0.7219386263879971,0.5190732836389234,0.6562387558691919,0.7623019948412533,
 0.7686250092436021,0.33652492817503865,0.6878772048932483,0.3768903982771846,
 0.933112857889456,0.689801013864502,0.8213942765563892,0.43595777558904303,
 0.5732844640663566,0.23380710434008367,0.7924353921629553,
 0.11027888788667739,0.496171055868172,0.6756549314434598,
 6.922817771713996e-4,0.9360873137446359,0.1893019674062052,
 0.5711167043349618,0.8247319475114547,0.7314495200340339,0.7664194956208059,
 0.26082538272405364,0.7997063260346519,0.4674506695375138,0.4601660598999925,
 0.3481668764208755,0.022454941434644216,0.03853129056449256,
 0.8201833257570369,0.7999595978679389,0.05422533488873116,
 0.13087516603707328,0.37943431440159603,0.8868138614922731,
 0.6324428203996889,0.9060697956382636,0.3675251840104684,
 0.021570919902355157,0.1299603283937878,0.191839518028752,0.8738668363359384,
 0.27414036018923216,0.06597672996904891,0.8107174099387603,
 0.26850567353629273,0.5437877191550002,0.9568701069574544,
 0.28561548691454064,0.2998736895545774,0.6053230475396811,0.3214634751389249,
 0.7921213247152218,0.5701935341726704,0.24733131595779223,
 0.45666355419564797,0.49982580989742287,0.9820624439722685,
 0.2662104269984271,0.3536606179984092,0.20628479046148207,0.5317451905976522,
 0.4849173079878908,0.476824234594313,0.3810183991205395,0.9925762851124675,
 0.32091914704559876,0.9370042779171145,0.7130490194481294,
 0.48369731349065304,0.792997269810441,0.4172861805865815,0.3925855414511209,
 0.17643264549084026,0.49082322166292036,0.8000617344367412,
 0.9490501014961026,0.6000607825363814,0.6697652005440711,0.9913039126695518,
 0.23856523756629677,0.9536679786597763,0.5347234625130703,0.8166002224406359,
 0.6460265666913918,0.09225920092604678,0.9197302537459422,
 0.24721081741213835,0.19625275871665038,0.8091250457527118,
 0.018176437795682898,0.8328820003857427,0.7303383778406031,
 0.4900444386031011,0.7880297361884565,0.9763404091393888,0.819306847628722,
 0.9362750394693381,0.24528816311188917,0.7948975794324802,
 0.26113005266961964,0.002069812324094422,0.3539506460111269,
 0.09459316464758372,0.11558162660848437,0.009173128277777609,
 0.8441899861932126,0.5468053773804894,0.24104395573117976,0.5676848456126922,
 0.8550813658961983,0.45625025323408663,0.5804882261622974,
 0.19833249478184778,0.6439778357469134,0.9316165911440967,0.8307054848874443,
 0.7222013946889521,0.6741878738808068,0.36239683400027556,0.704891643530423,
 0.5042841841396507,0.9720457795638867,0.32671013111027836,0.7502953290768577,
 0.988833222634203,0.9125265771229524,0.4027072301680473,0.40199369679210595,
 0.23419043543808105,0.6545427759746836,0.5749819626271752,0.8846432111135949,
 0.04188839080047513,0.368924657269858,0.641221975473023,0.3729024381398608,
 0.30186658534402633,0.21692050625935544,0.20909471830757842,
 0.2598723892569176,0.4866262861057843,0.6931098904517161,0.7095416374348992,
 0.43295976100936706,0.16801943390991703,0.7223359207116928,
 0.7772708254971894,0.29389267443697553,0.3277735062912851,0.6870432085751601,
 0.6845545731521558,0.1327755047196124,0.4401078450448527,0.7290088751646611,
 0.19038675894450363,0.8961008019701375,0.06537681946033258,
 0.5015885633184674,0.24833249999658946,0.6055902526127925,
 0.29167208166601033,0.7844320704015167,0.4718570375134099,
 0.14592536747136353,0.7919202017276461,0.44949136149038327,
 0.8307279333757531,0.897998133073727,0.312837390161935,0.9842706548238498,
 0.7734244104171777,0.5088450322402608,0.18696386764992123,0.8998353964087982,
 0.5908554821623596,0.14613258285900743,0.37582086945707216,
 0.7614838903969916,0.9510684582671491,0.5375197925234922,0.8491684669498019,
 0.7502871010414371,0.9942528228098475,0.28566958410544074,0.9452397473954077,
 0.5395677453606654,0.5184037570075203,0.7290659638005025,0.43925037207531104,
 0.10782422316166351,0.4473321690693355,0.6928239004029703,0.0997902627783469,
 0.8149780205468581,0.12575224315951994,0.47305099759080926,
 0.6435742126000156,0.24951170471426254,0.9946700330674291,0.7433938070527681,
 0.8734267784011503,0.7006870576360845,0.979563231171908,0.4144129218405842,
 0.5302857199244507,0.6599440949466417,0.10627390461561392,0.6266575868100589,
 0.8782144904840985,0.843794422169353,0.5686874687801732,0.6322215244086598,
 0.22591458351039462,0.49534831016055847,0.198892343635804,0.4261984545966023,
 0.031661136879590046,0.4677534434510997,0.9599738250655842,0.676528757377801,
 0.135760516373451,0.2960945428380042,0.018591647727238314,0.5028388471239507,
 0.7305735101916718,0.10066815451879318,0.9340700412079779,0.7223286369780212,
 0.5502749790753777,0.4966906006487587,0.4277957671346291,0.6079015156286991,
 0.15121928023676,0.11516967217042495,0.18285211380967725,0.6738750879120913,
 0.924249562949325,0.9614015256933097,0.4539693150592028,0.4881798233039424,
 0.537404980756678,0.46387657354825673,0.23072482567182118,0.8448519198131866,
 0.8457426392105336,0.3020084724284904,0.1605406579795573,0.33304464288097346,
 0.02760532110233649,0.571659096857191,0.3328472469684679,0.7524255815471291,
 0.5205065752382787,0.3424008643938169,0.627183182736477,0.06302039819787786,
 0.9180675295874098,0.45417076524830524,0.3841295189656848,0.844273282620243,
 0.4375447909512854,0.8019918003369029,0.4702767840300175,0.05433863987265386,
 0.5707412485478969,0.3484254131886678,0.7174176192633313,0.06473191844164128,
 0.33658046387402485,0.33836014328060793,0.5339202342468707,
 0.9395792170171878,0.8057951641978186,0.024839000979121617,
 0.21983645886995717,0.40857673528707183,0.7448573831824437,
 0.6505997365652694,0.9016364918578629,0.3329614633282407,0.5383315586268369,
 0.9342354142495397,0.8443587025333583,0.7093663181065321,0.7435610793895733,
 0.192839083026913,0.689419117958532,0.9070176113073932,0.5008924401851147,
 0.9371752404841847,0.9802678451024195,0.6429730266021206,0.7507663022043768,
 0.3825130194304016,0.41460970350359305,0.32698624035496837,
 0.8218335385913464,0.9894050005401381,0.7559596039734423,0.609846028792077,
 0.4341711577496741,0.7074953692175341,0.9453383115806238,0.8497655643896519,
 0.5597420361403952,0.0759349351694848,0.11791405755689155,0.9129694152768925,
 0.4276572622354191,0.7029934161191979,0.6011699853721737,
 0.015571754847466424,0.43305415838449324,0.4713268980119988,
 0.7105690178575923,0.8572541111696015,0.5318473369206628,0.47889026443495075,
 0.9754444160164285,0.698440803417851,0.11421477996727503,0.7405510433419334,
 0.9554292274555048,0.7623374412977904,0.16886586659054148,0.7022804623060458,
 0.10534653587673415,0.7906419945396118,0.5935129653314326,0.5586884507722699,
 0.5841916400276391,0.6798787689721574,0.07995449832930768,0.3248850013899037,
 0.910933112575151,0.22386020773935622,0.5182866036704711,0.06407580372053279,
 0.06124059421283701,0.6446528529281925,0.3029387354681632,
 0.28034033781418777,0.35910186168630276,0.49126194656581923,
 0.3910197262156757,0.843641687219815,0.2930322861375523,0.29958656577968856,
 0.08830933607235014,0.7351391142404717,0.7920431453325847,
 0.02807345689723717,0.4641667707374276,0.12154223830346811,
 0.2255933061000105,0.7276962754240561,0.5242878891408079,0.519226297030172,
 0.9886968443235836,0.7712078650421944,0.0966104951658886,0.06937046765301147,
 0.3783901859251819,0.6705661726252227,0.18420446585735883,0.893123821061596,
 0.5805763696363553,0.21068763020458836,0.39819563352979337,
 0.3225401567264604,0.35646193982238494,0.6914440895711241,0.7139115669036287,
 0.5964804243074923,0.6656411269982127,0.6822190540251345,0.8277078530076835,
 0.673193482161277,0.7843410066256506,0.21974643196250598,0.827316930577282,
 0.06426405348324615,0.2337785325596854,0.8801955545602633,0.2488373336697568,
 0.39781375955051224,0.3079994571298874,0.2573800459579825,
 0.06744317496691032,0.24612719984190712,0.5639388430096058,
 0.6878603974008901,0.6056835298487994,0.257165786723333,0.44662757398067865,
 0.4461326627869058,0.664031469518613,0.28460550877575996,0.8715355587127481,
 0.5295687575875714,0.3369519350622885,0.13282197595495004,
 0.26620046781409745,0.8498306115412535,0.5249664047730569,0.4213803457908383,
 0.6253650799040807,0.6051898921451548,0.7436865135263603,0.2425365463919742,
 0.035123006746644925,0.6881078467946496,0.8312800198063777,
 0.7619095471513986,0.9632490903449703,0.2723303919515574,0.9547554298468012,
 0.27013989099876945,0.661497856692344,0.9883014066404818,0.7154164889518095,
 0.2521066493872067,0.4895240927306531,0.27993789309670103,0.5167853915139141,
 0.5202470415902033,0.6797610742100474,0.6059009065935651,0.8084557093333259,
 0.4164579532428918,0.6622009040155598,0.35795906776017905,0.4913567773038202,
 0.36073087590764064,0.21423731703106874,0.671595747284515,
 0.022711209293822865,0.1273011498012282,0.036401610071482904,
 0.8158281928030476,0.20726344856515888,0.3108908235460158,0.2718012759759111,
 0.17661239934106987,0.8898945709961225,0.15775195910138617,
 0.2535766159198499,0.47377046563868486,0.9977272493768953,0.3897126156985158,
 0.076666659168381,0.7772922201566552,0.6960398510566266,0.15556574061498407,
 0.7741340715648786,0.03293081843501744,0.14442106736948446,
 0.11943405410567376,0.4099277355790828,0.3988676588647059,0.7847080464512134,
 0.8739093259091302,0.676239324402653,0.2700241565609196,0.69908596478944,
 0.6712917309007689,0.07163608396673682,0.8526827087028295,
 0.42516178411099614,0.3016717361188015,0.8724830579363343,0.5591510870482604,
 0.6830357237583978,0.05643672132011224,0.14241156503240981,
 0.3492781002862302,0.12599975700118926,0.37757739240056865,
 0.2440249887457553,0.99204057610636,0.3354378191815691,0.03675601518709348,
 0.1233168939177508,0.4081619971474755,0.3157372535048486,0.09632362632667824,
 0.5475945104915985,0.3647672962971583,0.26918225178165556,0.7151391107491668,
 0.7147221488216425,0.618892054908509,0.822018619116974,0.14370173174273737,
 0.512211062721396,0.8152465927911695,0.7510835648497297,0.7209253162019765,
 0.5026125012314865,0.4661678077319882,0.6822827615473748,0.4310173397568008,
 0.18707056990140658,0.6856398188878096,0.3467840860548561,0.3398511938810064,
 0.3320422597574858,0.4397811812700645,0.5712378054885993,
 0.0037042093488373062,0.8095849468405506,0.7462264409186277,
 0.8315058314120867,0.8244105696353388,0.4364566469494119,0.9851938209012522,
 0.7918354243553974,0.2689864929050977,0.9389206571532472,0.04221349835798016,
 0.44067398983194295,0.7993497869932911,0.8237325855260125,0.7621321338087613,
 0.39340227320995325,0.6588336908103534,0.637271071263356,0.0753470593076625,
 0.21406615854350752,0.21536758414417523,0.056663206657455834,
 0.3269169528993987,0.25464706183152463,0.8569969229678019,0.1444873469076815,
 0.03200922868461398,0.21587862075193642,0.19495102619540172,
 0.7653893997201509,0.455387156489911,0.7577979953734797,0.24363930810590828,
 0.19935527901761807,0.5358785773146816,0.21042374070180614,
 0.5121176809185158,0.8182716988027351,0.8309434908423456,0.21751838342522312,
 0.28685974974850736,0.6732208578927824,0.8570792619251467,0.5733970363343013,
 0.9581766687102283,0.11379093529504458,0.8367153714513004,0.7494809132733998,
 0.4557257856887287,0.5890125532685349,0.5782006427050213,0.6821979637439349,
 0.9797586290954338,0.4474488877089138,0.5542596878832332,0.03066283357352395,
 0.9566981114085403,0.2440888485366366,0.8970803454670452,0.04970364585112408,
 0.9953951877766997,0.15812177666831362,0.6091592022413825,0.8137048469897518,
 0.7677326245120287,0.0896551111149343,0.28634967252916643,0.7217453915588034,
 0.1485311384675222,0.8196189190996956,0.2737861493054412,0.7449230533137661,
 0.8479344283709125,0.8822366998116443,0.18244356931742256,0.6793506313285651,
 0.008561512915922664,0.13892178694269885,0.8477069595162399,
 0.3082898745792396,0.8591025673746515,0.778308162832947,0.6968307992482534,
 0.4588087147616928,0.5002949475488073,0.6904038160179372,0.7420557368940264,
 0.012531913448111776,0.24932480967207338,0.43223694093430665,
 0.6035676823368438,0.6861712752794613,0.07221018065266815,0.9180194391396925,
 0.27349033222807706,0.9235945643027955,0.3709789531919302,0.6225177918402051,
 0.14885278008279057,0.04680550244877357,0.32513486981230966,
 0.505655366531703,0.18826316502125662,0.8020859261226254,0.23834099162414812,
 0.9348658018368395,0.0885369826651673,0.8303520929733035,0.9830479038674558,
 0.234010348381965,0.16512420108373493,0.8019049208707438,0.5626018655323232,
 0.21152702322259398,0.0022953729863532946,0.040426770056668504,
 0.6336734073498798,0.7520419324490271,0.3570143479688339,0.32800546438125,
 0.8449849511668084,0.31354199166495944,0.6774347131028112,
 0.07352081007699407,0.18301565183997337,0.6747304629667795,
 0.04089804042689005,0.825649316895777,0.2132936017060303,0.8004348969469902,
 0.7791233391760385,0.11325234483986624,0.9745395714545987,
 0.18234848007198512,0.733485685065131,0.05677216578065769,0.6840094645092962,
 0.498772990033039,0.3875546688742926,0.500930182399252,0.3467799277757746,
 0.4396112371687797,0.08045792719259293,0.12351452445280853,
 0.7030189787460253,0.03277131838605285,0.7151215256842626,0.128949550746989,
 0.3158478160111381,0.2827709238377021,0.45926501672322306,0.7871866868378569,
 0.07648752418171956,0.4146821924737085,0.31438617815125625,
 0.3471981227229849,0.2558995243338935,0.8672502412375788,0.033560096968132,
 0.0020316536704951815,0.09015451666847529,0.6604062826246071,
 0.26722564749285826,0.5581986571543698,0.1265373065942269,0.4930327442368806,
 0.12094253351456419,0.959628607128764,0.8316565706286125,0.7065407725522751,
 0.6222913249731362,0.9185520062088738,0.1839788741635866,0.5756313932414967,
 0.8142048214290806,0.5600380592926858,0.8976312304111665,0.7039692747942109,
 0.5181222403583172,0.24945295492502784,0.5063206866289237,0.9418643587961406,
 0.06105318815977223,0.2607624068205018,0.9839861206193155,
 0.16514779712366168,0.9475284303967467,0.21758045957782512,
 0.9092817791443355,0.06153804661824813,0.6522454946761148,
 0.28302579702835795,0.7147594898731255,0.014400842220623433,
 0.2073991502573701,0.8655685288376194,0.8400479474089735,0.31978130855197473,
 0.7459504249139052,0.508814815277355,0.8001366490972284,0.16617341656331297,
 0.722267340837931,0.02775141260134939,0.10691656396172566,
 0.48990045782124936,0.20070704193042643,0.22461217695825286,
 0.7164299548500496,0.08989293261078235,0.017744188895290502,
 0.1713588920374497,0.8926034605495544,0.3313742716131084,0.12707574498418628,
 0.6238649266163507,0.05778015438364381,0.23729641061760487,
 0.4560778619792397,0.10788883738879629,0.5883962156648898,0.5552363176232881,
 0.2909855767226782,0.20976689216902922,0.45254584358931194,
 0.9223058090727174,0.17658157354625303,0.9946575468378926,0.7604159262339084,
 0.001642347914227793,0.1949977120033366,0.22416527119558305,
 0.2797362595980142,0.4952103606042074,0.9901912854081035,0.442834046763788,
 0.010647804261748162,0.4767502232408074,0.3043317459151704,
 0.8549687314028459,0.4287201171887731,0.9076564442260115,0.4717696996036925,
 0.7448805703024836,0.06951370512895239,0.14285004355619557,
 0.8406725709188382,0.02002480642893445,0.05654261726578791,
 0.9936646828459096,0.21116701919379632,0.6572767659099183,
 0.42427196012015334,0.20366151101163166,0.05759175182269216,
 0.12093380224774819,0.5451202152169976,0.6651345731829957,0.723249466747146,
 0.3381727074068076,0.05301124392551693,0.09615953122647614,
 0.36310233175427553,0.35436670522254865,0.38405814052454,0.3684107233882641,
 0.8827337289610452,0.16834754769494475,0.05188108610857356,
 0.0963230009562468,0.7779764757320887,0.45474216756758035,0.7007054786702103,
 0.5242374139241384,0.9651716235608586,0.7676696785856153,0.2205616503097546,
 0.9192047409077132,0.5096471457038624,0.4527385270827352,0.11731523295964064,
 0.4293543590675296,0.826783865181074,0.5394056358236363,0.40626462244360173,
 0.7093397924398063,0.567117323153426,0.025054886030230206,0.9803606211890564,
 0.6246130150281992,0.20396316559085503,0.4295443420934816,0.8202979028343368,
 0.46297124785459653,0.6142949993948563,0.07842994733396658,
 0.7454897077416038,0.3898481887707137,0.05605298977504869,
 0.012190230842174299,0.07488056882762106,0.46662055081502574,
 0.08765659672117532,0.40660211812114744,0.31722116775053955,
 0.5349278742114291,0.47718859039093076,0.5050255204178449,
 0.03276791840013227,0.7106479187268891,0.10317025047437545,
 0.8189735007303365,0.5004376252280617,0.6252925324352405,0.8358142713034225,
 0.7222254254925501,0.6975537403829495,0.1595979450285928,0.4851129023832065,
 0.761273525606266,0.08552294495837542,0.8030743417615647,0.6867897902688289,
 0.7973192202154139,0.9315225764659989,0.10608021956734104,0.8001364768809389,
 0.5973681468340133,0.8630968313935481,0.34577104669164116,
 0.16728579408584643,0.17384609392088346,0.45489749415776026,
 0.6915780114643483,0.15762284277969185,0.4928926016628332,0.8939045266764587,
 0.4188530649033583,0.6894794630584138,0.5956702069607458,
 0.040149273635618155,0.3676346196119945,0.8858630944359153,
 0.9704893511296048,0.30711608309454275,0.07664276991417007,0.989381492591329,
 0.8671939144241273,0.2748683537994816,0.07648485059685484,0.8072573854667037,
 0.787236310591212,0.24673578753620296,0.5712532849991684,0.16487496722605355,
 0.7349339724439923,0.7771075696351193,0.8413079471779956,0.501751049520059,
 0.8680503377123155,0.4909815746038858,0.13873719732966494,
 0.19416579608975526,0.5752547946054883,0.9348492727838517,
 0.44452038684833217,0.7754079934433835,0.7696936711290764,
 0.19935455338423735,0.9783372921469615,0.03036759758483254,
 0.08190750686017068,0.3689563481830649,0.4526413157476019,0.3187222171227695,
 0.07383679954194333,0.2649292308581557,0.23968529119816062,
 0.07541456510702582,0.017404896702211792,0.42194967706706554,
 0.9577418602443264,0.9604149964674971,0.6280597783619624,0.11987931536528351,
 0.9183809101521003,0.2978294300092492,0.32573117538802354,0.8899026615397405,
 0.27595790603916914,0.654036762833699,0.39272165769286826,0.6998278970533703,
 0.3329331494319797,0.11948263726659991,0.6516243231288639,0.7173145058826149,
 0.050406307753626645,0.370358551997115,0.5020661415140885,
 0.18976900084325832,0.3077349205522505,0.547153756891446,0.0834239316817259,
 0.7061358048284025,0.5459181988397682,0.6068085955950099,0.8100002260163591,
 0.7124380997934847,0.5711932264622399,0.6209595513228869,0.7660542702373477,
 0.6229577909986195,0.6625503060981184,0.37363611041520983,0.5386437271635255,
 0.7295293298814669,0.11947062121491503,0.4542584732275151,0.8341939439847031,
 0.7980740524138494,0.1016094439928863,0.4668968661264943,0.9027353503949712,
 0.09028594053999517,0.03653852078691122,0.1253107803867285,
 0.8195976675237091,0.8034440381824525,0.03122187480014349,
 0.39436667920800517,0.6392897809451679,0.7677498656978697,0.6906926846554314,
 0.28255092451328423,0.24799497254098002,0.3100092851481715,
 0.8743615900232549,0.6521874319733946,0.33663785162055626,0.9910867136733468,
 0.0578870650891633,0.2656122671427643,0.0582349781577175,0.4832226248163951,
 0.6516235863309057,0.8479674854612704,0.4794582197080769,0.44751370047747085,
 0.9627919437289472,0.17841831300489874,0.8817821949675794,
 0.41670549197144513,0.11862492532444602,0.6374416211648397,
 0.24805073706530778,0.4507058207970921,0.7916968075715178,
 0.21445001132366348,0.4393300883069753,0.6868481564527054,0.4629964394202494,
 0.5928024467780982,0.09791704904644638,0.19692456311772544,
 0.48299648725114375,0.9256471650800759,0.8827389225370823,0.5750752375897867,
 0.17542046712154502,0.7359657878751407,0.3831697884187173,
 0.13045106230335413,0.08616388023843591,0.23617263312164116,
 0.6095636302306933,0.04185551138029997,0.3927312068216896,
 0.12401960909758514,0.03326399535074087,0.4057473456738506,
 0.3365141089089301,0.817010423032621,0.6690412371197297,0.7398940245128941,
 0.534811457156491,0.8154405783195243,0.5224817308274434,0.7959178405285239,
 0.5101904687580926,0.2160640774616276,0.5764818028090327,0.5583794140754296,
 0.1651127592320002,0.5958517972078589,0.42392019202743336,
 0.11799540425816457,0.11413897219822844,0.025567843350216557,
 0.5201047931554793,0.24220063415017312,0.6527425688166797,0.3613692745962723,
 0.16556836833111777,0.729171244593784,0.47946451567855686,
 0.10146647169812328,0.8921835735598432,0.9504853038834427,0.8826268755043004,
 0.603270775167492,0.5086299951600426,0.23875539620084762,0.43885130401031536,
 0.9728151151192136,0.8930082131242847,0.591587424568826,0.02179066357901127,
 0.029104675083506337,0.6387511119268585,0.3073239617899096,
 0.9632335592743307,0.4366132903140989,0.38678071728664243,
 0.39622725102680134,0.7721235032405487,0.3274769191417852,0.5463736521219961,
 0.2974052021387116,0.8857989206795871,0.7129116178042616,0.46947218848468875,
 0.013764050765679658,0.3950193173106664,0.24377544140985263,
 0.13507367712865748,0.889748615864652,0.8286654049665371,0.5106360640461121,
 0.8982520283682895,0.2737185865141626,0.9206163909625102,0.7537285698997789,
 0.390488863788101,0.44011100583471696,0.8347251582480292,0.0232370783211161,
 0.42582995369143084,0.35739839436188936,0.5930991429244393,
 0.7410065836083753,0.5369658313821863,0.9528594494519237,0.6314852271003414,
 0.5274250287056051,0.8173806326408712,0.8452962408054258,
 0.031859171843054424,0.5327775514216841,0.48018026874389363,
 0.8347994264866758,0.6589119076495062,0.03526987637018486,0.526551009238467,
 0.3665855203905567,0.13359374623414666,0.07772258384647868,0.442191353368781,
 0.7526864065276472,0.6880287430596219,0.22318775766182053,
 0.03553610965376053,0.7177092350358396,0.9853056350034463,
 0.012224555684901883,0.667708910039463,0.049388136830438634,
 0.9963949677244854,0.37489225146332195,0.08369652402182914,
 0.4314020595924226,0.054150236846254085,0.0714802223098614,
 0.09423866574888651,0.628230606125337,0.8281036560417344,0.5771365735847538,
 0.7524144608910889,0.15435657347209286,0.015984452807491945,
 0.8798830708327661,0.4489131980314949,0.7588318811649548,0.8731327083940731,
 0.47394588158451145,0.6247806586003105,0.6006390873066245,0.6220808786721316,
 0.7858524657756086,0.5864649485972937,0.5776751808948276,0.8673632258196874,
 0.05152501869669979,0.4676085766284105,0.9488603748621752,0.8896091403083367,
 0.7931400384080136,0.7549460385451766,0.728213347556851,0.06037245622784004,
 0.05866604362032879,0.19582968236028608,0.8901459552513008,
 0.6396668363061734,0.360471888342064,0.6576971010458247,0.10854912219382684,
 0.7751936487347473,0.5538123037014536,0.6983985492190808,0.6924639878919298,
 0.6379727348970448,0.029028919892135296,0.6676320799082827,
 0.5621350728488755,0.20297403808193837,0.29104870359520774,
 0.38861195448540853,0.17031019585383267,0.4780318416130427,
 0.15344923654768028,0.4613655586567278,0.7889277175273056,0.3110889601319444,
 0.7435199279165394,0.9233480357318706,0.7143048687983392,0.7424967751100366,
 0.5939913028076375,0.9695503114078207,0.3045229188519156,
 0.016740127651077907,0.5602734600948648,0.7337937806902979,
 0.31534947336445907,0.8130417913074209,0.8400235339932475,0.4535442048073406,
 0.830364866830879,0.6942204081720649,0.2308264480420481,0.5639973312746576,
 0.7223998130829657,0.04471740589398654,0.6719434548043626,0.8192138830687105,
 0.42547284399234786,0.8421498406473507,0.017757662436370558,
 0.5550757789338291,0.1378566281264414,0.17257401910908254,
 0.20792270919352496,0.850251091875563,0.16319186903076943,0.9312153680821453,
 0.2446693016460666,0.13465929777228658,0.5448133997749488,
 0.18861611770300812,0.9447681178440398,0.15224031666002125,
 0.49828778476885927,0.1404038124234922,0.7267919756375633,0.684127226379471,
 0.7651787544376047,0.9165846764011135,0.4344101417985513,0.6716207557696378,
 0.0875840530823987,0.5069097354716842,0.5180995202620746,0.7322634253053408,
 0.5705130827277622,0.8565384347231368,0.8717267904312972,0.1579990470015924,
 0.7775946407619073,0.41321827578675574,0.8840244326208088,
 0.19838643763873665,0.742484052973349,0.8283013127554565,0.1795593670285175,
 0.8813529286531743,0.30937533245225735,0.5222050063100836,0.2311495422882902,
 0.09953968103927802,0.6991200804208476,0.9052811168465443,
 0.21129430206105626,0.36860485315604286,0.30415049268625116,
 0.2421102736153039,0.43063954779190194,0.5841213493687722,
 0.036243136559648836,0.32085533770379215,0.561984776496084,
 0.7717577046637285,0.7105701422314435,0.31049061218025664,0.87790155450846,
 0.6107170316495832,0.03268841897009611,0.9778609302567439,0.9561857152304317,
 0.8551957749717404,0.1204981370378686,0.4706500981347296,0.8782393054348165,
 0.5482593620853106,0.42447703465865727,0.6416948414516513,
 0.03276900004628658,0.5265498375128168,0.5108817069918878,0.638652368031709,
 0.4255285934443237,0.974305792808163,0.7895779331331506,0.5849260595103577,
 0.5434788028557593,0.8871826940885521,0.8187329262635112,0.19033630684066527,
 0.5806535535459872,0.6312812915285693,0.2690696514252283,0.8471064974465397,
 0.2554161556800696,0.6671071236308178,0.26514089245994954,0.7716502740052057,
 0.4191362427862605,0.2600443422514046,0.48980488679867706,
 0.12996448020636897,0.5348642469625229,0.35467971619453753,
 0.04333801865479858,0.06876593382736818,0.8010379053346203,
 0.6006325070882994,0.9395946176816445,0.9989504605217735,0.14142965188734535,
 0.9305667461258043,0.8009586281675077,0.5640234278056921,0.8318345644537527,
 0.9778836503658453,0.4870982246231238,0.8224744563114297,0.9839464838946266,
 0.2808356220871926,0.4776840287729831,0.453906745496512,0.39276958536080475,
 0.3267664068805839,0.7508907806272194,0.5941009484134554,0.16424831583969146,
 0.4788253054029893,0.6256176019252716,0.26418830702751817,0.7127927258024324,
 0.1568573647006688,0.19671862561707754,0.7784151001979371,
 0.10994210318384723,0.3260683597412224,0.15234427407712248,
 0.28535736702927494,0.8005273792321235,0.8314937487248735,
 0.13126166002977113,0.03541536715378879,0.07670425119800017,
 0.6546396595506571,0.6870622737095337,0.4903966473475597,0.09009023982098541,
 0.5485107749164291,0.19650821041701227,0.5747356141041876,
 0.19541446433256227,0.6285667045281238,0.10962497709063623,
 0.12867021130122225,0.990935720308743,0.7123457813415768,0.3839170500469319,
 0.6470573473721589,0.7506642893240694,0.7264444538101534,0.6536659363578132,
 0.6946083844551048,0.2524576276291802,0.32809581565852775,
 0.18365110848126243,0.9396234330098948,0.6756387602165745,
 0.05026858451812288,0.6233348009018248,0.05992467303950888,
 0.6817595048445715,0.3787112954729529,0.1569453364773179,0.5025196626551316,
 0.8734688860237577,0.5235004016896397,0.2825562064421431,0.8283783903444933,
 0.9801309269051983,0.7913968367465947,0.7788668644391509,0.3615226564325714,
 0.42531350184429373,0.11265240798347376,0.3631606621177643,
 0.42408951166852493,0.4369733510136362,0.056920226154176246,
 0.14466150639626585,0.08954076755748841,0.09815033911047011,
 0.18672177321993333,0.7198656548174399,0.6614336803127646,
 0.022451923190200618,0.604642576291306,0.08914781286704954,
 0.2026859686424518,0.7011184646189368,0.8917501430154063,0.18355498020165184,
 0.3549818958886469,0.6874174868804802,0.757589225504582,0.822663096355155,
 0.4017214688676516,0.22884764454343198,0.2193363242783335,
 0.25377670705465183,0.23414721975839947,0.1130822554253178,
 0.08468615378426148,0.3853233602360062,0.7382219450219868,0.669187935041901,
 0.30321669108801397,0.11985679090825241,0.20378955444438385,
 0.6873678803048071,0.010175146484387665,0.663731686194749,0.8757193682386859,
 0.14142358249246578,0.9064903768453735,0.2523816258972641,
 0.49107720948324807,0.0065029977301629405,0.5815840609835649,
 0.8750266738747405,0.23972431625434187,0.13889140819811308,0.977590204945108,
 0.10083616560559872,0.40497926620516,0.5759678672871545,0.9358703865355511,
 0.12755172067742038,0.04501950015754952,0.5535986397254283,
 0.16892842719044343,0.5667797361880975,0.9493157860123072,0.9672097416396044,
 0.7150890725948623,0.3490336339565293,0.26720946740343987,0.8156251930180762,
 0.60605887831755,0.5783967815583626,0.42239769016377016,0.6657569624536854,
 0.7678110819022519,0.8921031400533412,0.9132200894337744,0.21618211968964351,
 0.004601727139787459,0.4174518357890198,0.8686907173760603,
 0.022220545546476345,0.2512702232305213,0.3937347345421629,
 0.9634357439228155,0.06938720119896147,0.71138049454613,0.9889285163502562,
 0.5412794617567,0.8375878020766906,0.9695747998329901,0.5204357011753431,
 0.5984191123814441,0.9897581312263157,0.24470507219235615,0.6289845842726068,
 0.5274421873028564,0.9390459970096195,0.4749864496485787,0.6192911583483919,
 0.9282706487412684,0.4262457610499506,0.08326920085777534,0.5757758537460665,
 0.2727871264522679,0.35143590580196715,0.9922998209433219,0.017395179473637,
 0.3408980695555841,0.8931778774975142,0.31371052403061206,0.3632675547248597,
 0.9979284957871007,0.8429161797799458,0.963137404201498,0.19601683341692955,
 0.4812543739529772,0.026063279339497702,0.6890572184502386,
 0.7696222175954959,0.05029689549274918,0.4093104800392423,0.9627543169102477,
 0.33492418658350653,0.03565458696355306,0.0294151233880674,
 0.42015035444806625,0.6904087730109424,0.7982103319861721,0.750431774171313,
 0.011250727249031756,0.501812973438747,0.8955555721201911,0.5665846351198836,
 0.09554578784260381,0.6122729591790521,0.4931792288329755,0.5178778967806842,
 0.9959592396287661,0.38611538349489816,0.7286764775091485,0.4331578428879891,
 0.6445115152426535,0.9731888673462924,0.36782951430784017,0.1263902750162409,
 0.8738723183924747,0.3772574682372325,0.4648033441647592,0.7320130241468462,
 0.4975700701774608,0.434828604589405,0.7763060105592329,0.3563101362600676,
 0.9984865619403276,0.2354951227016212,0.020356454674194557,
 0.4524747932205859,0.8216166266932646,0.1022392017749163,0.3740583208266157,
 0.9726089782731626,0.6439023955492945,0.5882454256345402,0.7171647449902396,
 0.5855949088340152,0.790727026496513,0.5535021423254944,0.7253458599455609,
 0.04576297987337441,0.3091628018427174,0.6565063263355561,
 0.23099853210966048,0.08322502903547213,0.8003532057672955,0.67721332986714,
 0.4529252083533437,0.7694781142243053,0.531263942292118,0.026103713646919213,
 0.5569536347265911,0.25591566470721805,0.990855615237823,0.9110073714842669,
 0.06168408545642545,0.869120611038699,0.5643705532120381,
 0.006613325209738563,0.502667885105934,0.24788046866862468,
 0.1638743491879453,0.3175937782219257,0.053589587680955786,
 0.07048649809337315,0.35000897449302104,0.9058837506646121,
 0.03922105507737905,0.7024256698718087,0.8314102970064073,0.2906297056956163,
 0.3018386627686088,0.2427539598624311,0.01658418813755791,0.5515946395454266,
 0.38444375863246794,0.31482683251620214,0.6224239148990007,
 0.9366699063524355,0.9014276873975081,0.09475864196447736,0.7389685742159974,
 0.439869208078401,0.018275049508721497,0.5456995217785775,0.5028767437611282,
 0.5134569269483141,0.0871887634794497,0.07401685658857493,0.694751885223291,
 0.5815650286420553,0.07315869754037074,0.6974639443266804,0.2609714478618895,
 0.8817316342517303,0.08262078503606674,0.7510279211203832,0.871149060023444,
 0.9579897686456055,0.15745788920540016,0.6710978470074006,0.8124090989271855,
 0.9300204530382851,0.840625075474891,0.36490362843366353,0.1389070888353483,
 0.8574250228665186,0.5701199615161863,0.906219193028732,0.48865700466886386,
 0.6337621028303878,0.19729065102240795,0.4320206996953708,
 0.27133653879831954,0.03816590294407152,0.6298546071602804,
 0.47013988971922627,0.7323947912442278,0.10739149407422488,
 0.48417266451960284,0.03174104676359413,0.7232442205867584,
 0.7380889587451336,0.7343383942760249,0.7774780473721363,0.30285858623267303,
 0.9916845975513561,0.6193758514241736,0.3557023733940793,0.7217074566495787,
 0.8006812179763141,0.40801048325300304,0.6230417795798696,0.9550964825961056,
 0.8977801872424265,0.8565920867952737,0.3799813580447772,0.9298954077225303,
 0.450460923257771,0.35910336733994164,0.9659429985295771,0.7958225171212768,
 0.9058875360051728,0.5043482079283024,0.38722502929810465,0.8911635314837235,
 0.28568936442511217,0.2873056487259531,0.909011137590202,0.8639454808061535,
 0.15747037024888355,0.77554704160091,0.9516167769195316,0.5297581962580903,
 0.7829421311532119,0.578195309889891,0.8339845989378908,0.288157721256067,
 0.14690020778382262,0.4735330607270962,0.3703840150375065,
 0.14958916673836375,0.7141719569288701,0.1609038773057455,0.3964683002517738,
 0.7047772246403823,0.4661205709770523,0.2818575197350497,0.1573066398095101,
 0.8797288268296274,0.04976056333094003,0.12522558938880757,0.841397222206019,
 0.4921330418794324,0.497892208797846,0.4817978917347323,0.2943586455453776,
 0.8901870229555578,0.5855510367819919,0.6284822773332698,0.8791879976012822,
 0.6641972782315746,0.12317892502823613,0.0389393471655578,0.5061035790496636,
 0.4814828624542118,0.6923244764981513,0.6019049023780376,0.37200128543029676,
 0.9484027078708023,0.7007609720113719,0.6496320047455508,0.9596756647269474,
 0.6807968643629344,0.6325687600656706,0.3369866781087225,
 0.006155737413535478,0.820928809655921,0.6326124266647212,0.6441298096789871,
 0.1540872954810416,0.6584521045400942,0.27516142522367004,0.4462253480707544,
 0.5960735844830993,0.7018474243602542,0.11009954586206061,0.6094188065526358,
 0.4831044480520944,0.7673966939535304,0.976498680591763,0.8678152263117384,
 0.7183606967691866,0.8206436958843124,0.5993289040512407,0.6964720172026508,
 0.7841391165114809,0.9805202242914819,0.21627041446429018,
 0.20354977750124792,0.9400305914198193,0.31776570108578084,
 0.41427565741788996,0.49393119781250383,0.6135897102948488,
 0.07429902390824972,0.3598576993681464,0.9740458493377735,0.9837314454762602,
 0.9765343835489634,0.2964626176326798,0.15897109405161047,
 0.09090292145115852,0.7044877661145137,0.7453629170759367,0.9824013193759648,
 0.23097316205506868,0.48577651463096416,0.9668494543524878,
 0.1375027521633887,0.18544543904500455,0.03182842745900882,
 0.40537513080337484,0.597805719257397,0.8242047331238254,0.765933288535585,
 0.2918050033032865,0.32153064340159565,0.8694112195380994,0.4197905370225917,
 0.1695998390731165,0.8157572910256707,0.2831000708245176,0.25799908978751984,
 0.5306720003405354,0.6535955872899771,0.8850695301960857,0.49341500761347257,
 0.32924920159940985,0.2696338361716487,0.5838027021296402,
 0.08607453920202457,0.7711363066024506,0.677604182786719,0.5232068043727052,
 0.5544771001128316,0.6541997306715567,0.4447230279411538,0.5522197254828072,
 0.06572257835556128,0.8226303049943393,0.09930314712359778,
 0.8828037832514324,0.03662182069534736,0.6845730741683972,0.98263167303505,
 0.7118936223404699,0.7177540194710126,0.6597848851608528,0.2353766864057356,
 0.621230488863576,0.5130949059675813,0.5776768196636718,0.7507530602660868,
 0.1374898636103945,0.9668703898624873,0.8661186710663208,0.2409352580641958,
 0.09393598909677803,0.973008466196047,0.7065821278682374,0.7590332207162692,
 0.34672090135891676,0.17531557043297474,0.7013620187624129,0.673005342889105,
 0.12900692015880366,0.28517086772457867,0.6914812477791585,
 0.09653615742735155,0.08157203869627627,0.39334090631590146,
 0.2794173797821916,0.35276521976150277,0.640882408110398,0.2645182839315453,
 0.5635224283753777,0.9948296708677538,0.641641771670616,0.8316025282439327,
 0.02845994173102362,0.7498321020033439,0.46673990825202993,
 0.08777662226800609,0.04210944171467268,0.395427194353682,0.6875484077748315,
 0.08675436525272562,0.1255201871764493,0.6671079206520805,
 0.055281239316403274,0.5737855862327026,0.3758801259696567,
 0.8523976103467962,0.7836114434632221,0.3304905003514561,0.5526642308662811,
 0.3192784438911699,0.9582853917037184,0.7041630076123211,0.4763214935928899,
 0.6053162587869154,0.9810453376568625,0.5430959918094325,0.1374626403922108,
 0.9165001505650668,0.32646927558604566,0.6107722673336288,
 0.04031335632381006,0.680274744731666,0.843303353872065,0.14559541424352274,
 0.8221799981769605,0.4596812535703555,0.4002297862186375,0.8226718761209981,
 0.7214663040312155,0.33907838869660456,0.974931712910502,0.4530211693739119,
 0.16222798493922874,0.8105232744421741,0.1533022499343566,0.9549687282009451,
 0.21463327817951283,0.6806308238655101,0.2675372971308102,
 0.46987778782270684,0.6459772600416289,0.06288770805163835,
 0.12034406780889118,0.12134650053969254,0.4099017964510768,
 0.5215032792408092,0.5480166737438503,0.4557330099038579,0.2502899471745035,
 0.7053654559464074,0.7680718715193177,0.4630101570464177,0.6722066698676032,
 0.7225105483058663,0.7451488838779533,0.9143888104691263,0.5662066341975623,
 0.7234173575566768,0.11613340252312354,0.8849906792577494,
 0.06371620368180597,0.9404486886647053,0.4604355908656278,0.7183738281118687,
 0.2694292303015442,0.8672351740802468,0.7859414464848458,0.811777133433152,
 0.3853202741399474,0.2303413322507435,0.3081192291283118,0.5880076945614123,
 0.6543375084245765,0.9735855762373677,0.22258600057939049,0.5565380344346247,
 0.43483228940748,0.40745249110062676,0.18997845506917754,0.8150234982026142,
 0.8489804357061329,0.9762811644449244,0.8167926430708458,0.43345613987523857,
 0.5760011949314615,0.364939962199751,0.09808030899387765,0.04673049349429692,
 0.7335316086817678,0.5637895935984236,0.9103707046768531,0.2899194771447885,
 0.3768678103963732,0.7763187831647292,0.007463033828065124,
 0.34821995630861713,0.008858019683474327,0.7174068928129254,
 0.424387023795053,0.008908657107987383,0.33729568752059835,
 0.5080384040627737,0.7369122939340245,0.7790628087600979,0.5999128840882358,
 0.9326478226766581,0.22383583337739366,0.1229937047130053,0.6099353333863621,
 0.3164362151898885,0.8379650518557352,0.767417648527059,0.5727648610312379,
 0.14044646725396592,0.7541575892242209,0.7489894562759489,0.4064880621846533,
 0.41556320238691913,0.06765377529990646,0.028841797724775575,
 0.8010122606189831,0.5725600910493223,0.7015739203848964,0.3398669232357264,
 0.7566305810351981,7.705325905844607e-4,0.9613509238037163,0.42983913144471,
 0.2508350806596378,0.6524995090758924,0.6420100003408363,0.7127166344839215,
 0.6936808297000993,0.7054433463914607,0.607889233697845,0.6260110010555778,
 0.12228220284187008,0.4165559778104253,0.624761845820384,0.14458670502690008,
 0.6575013800271194,0.40286241275944035,0.731470444768709,0.8031998058165739,
 0.9267951747648578,0.02521367660317375,0.6065326967837403,0.5674709895533476,
 0.9001016738307981,0.47124902622141684,0.07206657495496527,
 0.22820279513016095,0.7353541869663217,0.27190232672639936,
 0.19709743759293108,0.6537961933897802,0.5965842182999337,0.7498930064438922,
 0.3870318281505891,0.5782428348210099,0.5821645467128951,0.8262467963077823,
 0.4937973242561754,0.2526006451866203,0.9146082342212235,0.5598728780201103,
 0.5339803811391529,0.026995010424530563,0.3814180092802568,
 0.8614044977405443,0.8863773235644303,0.3830813774814902,0.3350991905181324,
 0.11147868323466514,0.8664044445732468,0.10513266983990166,
 0.2949143213056198,0.5246117190707331,0.05269728955990172,0.6261268011300636,
 0.7355611550609813,0.6804998907458849,0.2173209996118406,0.09811132720767246,
 0.31578959518005767,0.9804222461018155,0.025551685353676024,
 0.09656264229536893,0.13454076891184819,0.5255286941528832,
 0.4813634454876561,0.6423247008027269,0.8299030433175959,0.7907872887258102,
 0.8851198891242851,0.205285371272256,0.07330542879599977,0.7851205629962543,
 0.3040019342130095,0.7250217353301174,0.14921590713424215,0.5335495499303208,
 0.6884839290134829,0.977829156249844,0.4938415647866734,0.6855127517150685,
 0.36979793024992735,0.28008140511447766,0.8163760710494956,
 0.09578656292006793,0.18625987964072233,0.7265107085670653,
 0.5772157960814626,0.7246609282486822,0.2939235810014882,0.2023698483432056,
 0.09148790611315227,0.32716499810270094,0.9257586019383519,
 0.4823066148065186,0.7932253565153274,0.44273445420880764,0.8018194351860057,
 0.3588315618776715,0.7485577288376766,0.7963936359403313,0.59650378648508,
 0.5702377356007565,0.4175994651894631,0.19777832899998982,0.6165473176565202,
 0.8073288436911841,0.22860802244610567,0.2123810462366874,0.4008613904780294,
 0.7346539073045353,0.5850551431775423,0.6765093139560825,0.7390885207881128,
 0.0549760743493104,0.2780729853074445,0.059326986649749225,
 0.05912444559785246,0.33180186526615696,0.726113934130734,
 0.0058791994710302165,0.6235017611877012,0.16724973824408707,
 0.03844537774406076,0.14076020481671425,0.9481513833211856,
 0.4788372838077444,0.8630474492362871,0.0067236405261964505,
 0.5172525548453238,0.21286716610857392,0.14495118691556785,
 0.5202579436383805,0.7050711979077977,0.8294324714017147,0.22818581203901667,
 0.8038713809879531,0.06524918161400284,0.5287600284782805,0.3366619029750981,
 0.42735741924708803,0.24625686382793477,0.7182012554199517,
 0.39036841097817554,0.27372682975852736,0.41927878093827464,
 0.42681651741651694,0.6047595211778114,0.8019898154613593,0.1363572120115928,
 0.4710143216942019,0.14106187976328433,0.3958731835446252,0.5447226057615289,
 0.23310117564683264,0.5953944704332168,0.8114811437427294,0.5637827347933608,
 0.9377286796990884,0.3480274314097038,0.4170606032069544,0.8719665980532081,
 0.8773332067111943,0.7240344699366306,0.7849540068304447,0.8907893570782502,
 0.5902868117600514,0.5190863823687113,0.6489235239986921,0.5980089506674288,
 8.974998432869086e-4,0.6245085768847515,0.7863218866869268,
 0.6041787158466527,0.5700924751927984,0.04029937027215669,
 0.30502615913537856,0.23165977769743384,0.43091287811348944,
 0.6410213464947442,0.2007217147642174,0.41968847609264337,0.9491963130680919,
 0.8844399298348707,0.7153577481984931,0.9101210825198024,0.9217365804456662,
 0.32126663678874057,0.30882598325961674,0.528691758748941,0.5730250132399948,
 0.03699102054168635,0.528105848822458,0.1977275083690243,0.1616965567104932,
 0.0063985278617830765,0.8707189256931047,0.44510559819409534,
 0.7664988986269164,0.09471484215165937,0.35985178628571723,0.877239794002887,
 0.7912838828652653,0.3819842738443078,0.6350613294957193,0.3724806065961227,
 0.9529529901836763,0.12732515366098873,0.1416465375185314,0.334001342455856,
 0.5209077798919188,0.6400296297241517,0.5315013503034277,0.8071645268739791,
 0.2025311765165263,0.011271664750881705,0.4046098051425835,
 0.21983707180209044,0.04651778230973114,0.39711763448614046,
 0.46349935903470574,0.3513250802044505,0.6120188734197024,0.5130343100701514,
 0.23421132772545605,0.9963390051064529,0.37588353927020146,
 0.06228665814586243,0.5277165061233852,0.6544369867539079,0.7316722533891303,
 0.29170979167292044,0.6065911118873393,0.9584570398830491,
 0.06599822625923502,0.760051832600537,0.7540703394334168,
 0.054260964227784036,0.26276161123698294,0.563374933575707,
 0.6985571502905773,0.12695904243873346,0.8742702256725523,0.2754083030893525,
 0.669733424999191,0.28901570140464317,0.45707987797003335,0.2013997040604395,
 0.4335089542414632,0.10699669381987265,0.6519174748002324,
 0.04631429357414274,0.19770977927443179,0.5558972964839208,
 0.5610014280042491,0.29837310813507234,0.3258265603598436,0.4493915082240644,
 0.5327794228700174,0.29967691656936823,0.36840911459559544,
 0.5425835497305813,0.5238463711566355,0.9158043926760203,0.17279465234108793,
 0.28268969990313875,0.909958321153091,0.6409991759863356,0.32768082724501135,
 0.21828730855248124,0.2528292507786889,0.9854069784456633,0.8583421000636791,
 0.6010909174103121,0.4806453595596456,0.2839436467274178,0.9286049704429931,
 0.04275008608149378,0.1354406402101893,0.883358648443476,0.05408019823044996,
 0.9883966503716617,0.293511762351942,0.5637878997417686,0.6110152405673761,
 0.26075735093131036,0.49172394133538055,0.5334642110540853,
 0.9255232612931851,0.6164618565324904,0.9223249568017172,0.1874005627740033,
 0.3770456738153535,0.47431237605520404,0.8232519551209745,
 0.32803172355323473,0.4272942155946714,0.1329414302554235,0.2311341910535143,
 0.654658776943676,0.8375768986008607,0.2843914825106073,0.6483879569936803,
 0.7479748425629579,0.03957422424842649,0.7326054892358453,
 0.007200261591034485,0.43486916611855453,0.6197796321648582,
 0.9967474926584635,0.7963488236101783,0.43149966523092576,0.8499561150212336,
 0.19680518717560402,0.8047447220479171,0.8928429297284101,
 0.35174079515905665,0.3990427194115783,0.22541697902843405,
 0.9973382188578414,0.9328469171340051,0.9350098939527003,0.49308561413221463,
 0.9461270552449195,0.9627307807022898,0.3401622940976874,0.04803905585786983,
 0.9255224101823879,0.08309177713907456,0.564389630988046,0.44997824373493267,
 0.4470225898708222,0.4718349873082195,0.46837831430436594,0.7572114108051179,
 0.1079974498849624,0.5332312702353756,0.32902258020717967,0.7695664674555046,
 0.3734905179503468,0.6611224921927058,0.18701278151657807,0.9078937966420659,
 0.38328934177552043,0.6947138005383596,0.5292911984616805,0.2902184054700573,
 0.6901847472977449,0.9042386436750597,0.7567299103577776,0.2533583695670436,
 0.8808675804465914,0.24947077058332467,0.49425023909257626,
 0.8797087358271588,0.052038985674481175,0.2244267838798193,
 0.6166914162927243,0.7741741217943247,0.5166526110780517,0.24729696138861945,
 0.2491523083838737,0.9520352955337273,0.911365468776712,0.07198498246444995,
 0.77137410994851,0.30534026955842597,0.09719863686305874,0.6842641536780893,
 0.3122842709714475,0.938690959844809,0.7996175038603678,0.5113312111221385,
 0.4667620534272052,0.010989898566241774,0.3246217493262702,
 0.5800816724470169,0.7878073826187089,0.28429091549034435,
 0.17286230801715552,0.739327671709724,0.9888903661026051,0.10193554110143044,
 0.250090578029845,0.300867355525968,0.7715225211608618,0.4952584853337678,
 0.9790175862998335,0.6389283105993951,0.7077702411300079,0.7681809401070716,
 0.9391801162636928,0.8493173592596216,0.26896585182926025,
 0.01651501078856976,0.13915153076865494,0.9355316379511491,
 0.4724061268971884,0.4999147627459055,0.45214536018963436,0.637259935144504,
 0.062355493250322525,0.901198215832445,0.5877406787999404,0.9274727406981492,
 0.16996883508097238,0.37054354871136197,0.4203748379943828,
 0.5228211094086687,0.8140847966157893,0.5087301261010759,0.3929657202166099,
 0.1768968685005019,0.18396952614886997,0.2749845010001223,0.5408812118909857,
 0.7883841927366689,0.09974379886353524,0.5942083112804072,
 0.21503789420240027,0.25496612839606114,0.13105129151344097,
 0.040540130322850754,0.31834654978534993,0.6613746379005655,0.82328720213462,
 0.5847327126387231,0.9764782354470675,0.26774484309072755,0.5422969898154981,
 0.6563389133183001,0.9364217776575682,0.9611779100434445,0.04648899178610366,
 0.8519283634198896,0.8894440609393406,0.41697812988040517,
 0.19472814967356933,0.4905360556214341,0.051237355880751734,
 0.37205668904678124,0.7083996112139104,0.20612736112310515,
 0.8199163607932634,0.4760933063642875,0.6250143250315505,0.8520850750598961,
 0.7210431455803293,0.4218703325092552,0.7710308525118992,0.7878348665179881,
 0.3750465723502915,0.41452054467509125,0.6867019267518445,0.1000518759366994,
 0.11245808640963229,0.837527508445515,0.895557039850729,0.6781907499030677,
 0.6047364600039767,0.9663244445366781,0.7297042428414726,0.7456886426246685,
 0.12044482073271634,0.16384377676419037,0.4027957489622839,
 0.8934647401030887,0.9757436371030935,0.907887326379102,0.24564111925269305,
 0.616038440184828,0.34065432425143605,0.5903262882896192,0.41598413859628613,
 0.24241081846554102,0.6426317654349266,0.7502951077409598,0.6973618902282208,
 0.4409085559255138,0.1488307851520242,0.002041810173635805,
 0.048959658715108545,0.08493383385058861,0.4830493291123339,
 0.33885763789698975,0.5898457208925604,0.8819531313828736,
 0.45678853474865067,0.20380106711429513,0.858827384078769,0.3045334352659813,
 0.714035713117069,0.6086130883171574,0.6623602346365715,0.3848751819338032,
 0.6603339930369961,0.24791474967551963,0.5944852569686149,0.4872149984368691,
 0.47171709372875337,0.8409705500189305,0.32640465760798654,
 0.31102843880967446,0.9597238037459969,0.8450698980585921,
 0.32277851041551675,0.622900498554874,0.14732936380196182,0.8626589484482694,
 0.47184572598888275,0.7960396976473297,0.95182248109999,0.15822415479148977,
 0.7827169327089057,0.7219421862420138,0.14700795970181257,0.4010627821036379,
 0.28873357996683335,0.030532774914089456,0.7322382552285551,
 0.2352741529382103,0.34679255248816543,0.5520060195199554,0.3847818278024354,
 0.09692683539369862,0.13850074210439745,0.7948570986164274,
 0.5265285820773705,0.9840040863724661,0.48599645977069805,0.5977880535561617,
 0.5989630997339633,0.969853192554847,0.9659737373460875,0.4710553981939367,
 0.4584110537348095,0.7406908572123297,0.4492821120720385,0.07256209497705701,
 0.34336700069431636,0.8022488373691807,0.9318664098575323,0.9615101408145039,
 0.18985214912813597,0.0783315887026046,0.3954938739687124,0.5995796950855232,
 0.35585114326541867,0.6603669709284572,0.08719620779368664,
 0.8757586343708308,0.1638961170681098,0.09289024505975418,0.9709391335543134,
 0.43877717912078873,0.4319629165567337,0.7432543482267974,0.6848218575261701,
 0.6825052608685422,0.44314969848413044,0.039264856402213444,
 0.04316238226037794,0.42255696337980786,0.9528521929580798,
 0.5495229732229979,0.7698472836083594,0.10563526957801406,0.2860266887566809,
 0.5892685772295748,0.572381890940947,0.584003186629698,0.7091959147289654,
 0.5000473571580526,0.08792972747784511,0.9832821223502517,0.4656898822369382,
 0.09409587904373184,0.0027157970154043642,0.7770164949425528,
 0.3597763771087248,0.9017262043478622,0.3223977869085203,0.14433039241766754,
 0.8645142223110688,0.0459138296840883,0.6583066516105761,0.15862060373775744,
 0.555818085596087,0.9904136878610812,0.60478324765382,0.26726987250575607,
 0.7530590582070633,0.480683255165397,0.41683705021567274,0.872642874994294,
 0.9655451139620592,0.4663142121842012,0.425476948267286,0.3737687558874977,
 0.6763565082176851,0.28755389193642256,0.24344678685146393,
 0.5254558670618622,0.9884899328116783,0.25559164548599156,0.5713191942010893,
 0.34034316001299025,0.4700567572070127,0.7970508808386679,0.9909493771140974,
 0.262849528718322,0.2806332610865454,0.1744124413978695,0.44091978519697705,
 0.279069259972913,0.6507914543200246,0.9393506409537439,0.6743260546200543,
 0.5899494980592672,0.915276443941747,0.1028417887922406,0.8186240011085811,
 0.3790640996698259,0.175458314353239,0.3143220048588068,0.4240229979702759,
 0.943462612155893,0.4534612727049635,0.727937073705318,0.7510280589639021,
 0.8010961633195425,0.6951252880339592,0.3649903139912304,0.31306578741476665,
 0.5292295999025028,0.8030746756097223,0.2357186229497582,0.18064129860047862,
 0.5262573241645838,0.7726535689233817,0.4230888124363562,0.2833661648647191,
 0.37360833391344783,0.925296193246973,0.49280675292807885,
 0.20152983060011997,0.9132554975092094,0.9977655198036063,0.7868068034959546,
 0.5433007064196468,0.9373890565324652,0.06291937445270757,
 0.46106147343993187,0.11973080893013421,0.594166844992998,
 0.05035824943695655,0.7187081642843987,0.04161605171225835,
 0.03786283146611269,0.07865104218174745,0.44322519182321507,
 0.17333618391188754,0.2482104726956056,0.6246241924008986,
 0.36245972825566086,0.1033578411087619,0.4773423160218089,0.970281108812493,
 0.7627962066767736,0.5729914494070484,0.9055688425633749,0.645520375523993,
 0.9964031498471939,0.5511455108581491,0.3885156387586455,0.8525893204641992,
 0.031974268641998926,0.36851378240351895,0.7444080321415957,
 0.9922236604481579,0.7854765472068994,0.9649715497327171,0.02171974548706246,
 0.5548317830236259,0.07168433733581936,0.633899375592127,
 0.020506076912551907,0.6311823633665294,0.1691739379405336,0.85412696777328,
 0.940916487351991,0.6124418922305882,0.374444103010237,0.30694183695798816,
 0.5247747298044461,0.3762143079357996,0.43919021780969114,0.9618472980758108,
 0.1455936280247766,0.33174358240817314,0.7062999049234767,0.7812125705287291,
 0.1951184613182667,0.11420170287627884,0.6871482206289143,0.3705075513590661,
 0.2623406474407929,0.9492614081512037,0.796523699194502,0.5998694417144474,
 0.7888745253890924,0.22006055200451125,0.6450542481311439,0.5001719389788072,
 0.5743089181297635,0.07519866657849406,0.44159703291985086,
 0.7515046976020674,0.48624998045883006,0.023469807991112535,
 0.6565307492035912,0.8377218913730315,0.6197911374595322,0.8902499748654359,
 0.2803844368715993,0.9121936747944854,0.3251221697894703,0.8694739682017566,
 0.31287922375303145,0.985480194271285,0.9515448731490036,0.5673882068954557,
 0.7115147659241234,0.13560277862120318,0.8361562318645499,0.2530087826048407,
 0.5041506424189628,0.9659134779934399,0.908506460725208,0.79574603261285,
 0.5270791711523102,0.8819734421889778,0.009368145399810679,
 0.6819942030874266,0.31254323947602947,0.2864548049274004,0.6170229932244257,
 0.6012111831871957,0.7816424816897101,0.03365852105982159,0.9179936457710978,
 0.34926544196083453,0.1995212814863152,0.9119450193322428,0.4826655854403561,
 0.3696356555369229,0.7999824151109294,0.8485401179681026,0.9099551453136886,
 0.0544775417580482,0.7269218101493642,0.19128263735000506,
 0.27707567165792013,0.6644606279706609,0.3239653638964488,0.9626135635546077,
 0.23303164991250713,0.7230914744938086,0.9564210820326069,0.9231042238326493,
 0.6431283636737104,0.21953567177987354,0.9855416065750816,0.9230272672006985,
 0.3681041224573405,0.8603841764694996,0.5174332192224211,0.6101227342024593,
 0.5977265703623367,0.2177473581788958,0.0738841354691756,0.7344580599918211,
 0.3781717722492268,0.6230686939552751,0.31885904656029196,0.1166711097803873,
 0.33173203915733784,0.7758941493272619,0.21071482646079087,
 0.25280032262642516,0.45489632674438596,0.11830864584286316,
 0.2470700400721757,0.7997785670036905,0.8989493724533006,0.9525143715481359,
 0.8852968975058672,0.058949789618809034,0.6790854309123007,
 0.0031003224329485546,0.4100687524591001,0.7751820839678376,
 0.4143386928312317,0.6537771295435904,0.5100626215846625,0.8435926768623382,
 0.5636803544892811,0.2920749684699242,0.9198522889038961,0.2570870686117781,
 0.2684928217573006,0.13683717551058472,0.29616165061922417,
 0.3172711419583235,0.6260858184316426,0.6289656192365516,0.8526826220986231,
 0.1379111369090672,0.1753317786357672,0.8310940007643509,0.6678994682202732,
 0.8825640709702053,0.9764367878757336,0.3284989064167231,0.46648635899099944,
 0.01119253349919691,0.5855205138476194,0.7933014236342054,0.3005130403694831,
 0.13746635069061452,0.9971912704173569,0.027389505913467227,
 0.42094502581035365,0.03286163919087903,0.13644516854070998,
 0.5957241416146327,0.052592087692208356,0.1791305133795008,
 0.6625128394536843,0.6042924346580802,0.4725723156376148,0.9571225096665964,
 0.8654137150299689,0.7988256750409414,0.13250977292007837,0.5847599360996911,
 0.3564025114679461,0.1749193282786261,0.19646309867088962,0.6440932029055011,
 0.8409176755940361,0.5347531395090432,0.0965772943628922,0.22076908004453222,
 0.6141263218508345,0.11321613881447634,0.21312490821278873,
 0.4240049984908707,0.9921508106767536,0.1290192471386078,0.8227829439734001,
 0.4085714052038859,0.0935842123207411,0.030796442260386914,
 0.8995701182079271,0.3059361036956435,0.9100079927089018,0.47683571761321886,
 0.007782504692737224,0.8894701149827677,0.3315297277661322,
 0.8108537404712892,0.2813549137486191,0.8042311583759656,0.5677183150007522,
 0.644139652312937,0.6706084463525724,0.9661567008463386,0.6147125956410018,
 0.776175707550654,0.5911377444803328,0.8463358882265095,0.5312767788198849,
 0.7880357288591523,0.3957354319318096,0.5606264346189651,0.6618828952569784,
 0.7653499773612212,0.7430179807784141,0.33749410289279624,0.9779922343027495,
 0.8837609355383708,0.4617360618786992,0.7298079731772182,0.7830284380679433,
 0.8811853472037474,0.008162955869879918,0.8909298702515016,
 0.014517754348151501,0.1759097191905954,0.008573722023240138,
 0.591125027512927,0.1161294921183379,0.648304606999904,0.08799290421996353,
 0.7735742176142228,0.08778008052024566,0.8153425163901605,0.7172161689385167,
 0.04098645407628676,0.05227683152343332,0.8213364367241118,
 0.8951985196018994,0.1782684091608474,0.039347462272647626,
 0.2860916774715412,0.004521393515484307,0.14078955820360495,
 0.4905875994836204,0.4757518687948844,0.2772082945555351,0.0315709451532038,
 0.0375365291679326,0.26767678362218383,0.06555282533727647,
 0.7540529863486393,0.017659604504925097,0.9975556487375916,
 0.7276224159963477,0.26402593665182594,0.041548929935831036,
 0.3335775269810829,0.7219904459918002,0.3249373934147831,0.43206734802489744,
 0.29399889791322675,0.16316998029944507,0.5208345318654964,
 0.5800510561578548,0.7886411191619913,0.3118635763206382,0.681372651707871,
 0.6479480628977309,0.18798787616969237,0.38163388327272063,0.957051726051862,
 0.2859267355017938,0.056151187141403325,0.8912284573118128,
 0.03193239608143683,0.4213455819871599,0.6056186174497241,0.3267997588491919,
 0.7908341787158432,0.17879350008497685,0.25003317970426375,
 0.8862862669557008,0.4142246480736832,0.16328655221204091,0.5834614308806556,
 0.5599170687527826,0.7177744429684063,0.8575703525659629,0.8101611074882229,
 0.5249142441670217,0.9104748930611061,0.140655884964637,0.4935091477132725,
 0.5208422914441401,0.36462045568556567,0.5241009133641888,
 0.005337200202196524,0.25192097101159183,0.5170859468452027,
 0.7053141436457853,0.1740769868176213,0.009180567837858256,
 0.7152374895166227,0.983752210674995,0.9800510274305957,0.891190731637066,
 0.15598882647675028,0.13898289983957335,0.9190697474090592,
 0.48727417760524205,0.8297903184227512,0.36922498536123693,
 0.4937693630186679,0.5492477753011716,0.061124836297326124,
 0.8266836520973124,0.8068484260074165,0.3531508175463931,0.2720730654676923,
 0.698391221116525,0.4133922383860462,0.11693677547479486,0.5407142215390902,
 0.8231973988403649,0.6652631608455704,0.7815720197012054,0.6009632958950475,
 0.014563318502217149,0.4867767267677878,0.6693075537721451,
 0.8645708799682627,0.7284490676515879,0.35756282602174017,0.6293938667770047,
 0.8268476924152606,0.4508852411216685,0.005954795516030287,
 0.5832829061298354,0.05979457455951098,0.22852372563042278,
 0.11180648386433067,0.8708707317936583,0.7892445186389134,
 0.02177589084999143,0.7145402248798036,0.9598094233842179,
 0.11345728250261222,0.936322939177759,0.003960422756568938,0.477720539741499,
 0.8406027353568509,0.7188173295481316,0.7952857530953202,0.2712592848726042,
 0.4746715914866976,0.23589481044753946,0.28553861513411916,
 0.34482275914836946,0.722417709103363,0.4675929611958186,0.30783991228018304,
 0.26521847132297616,0.21710936078982124,0.8053637541997195,
 0.2203469653751784,0.04664224182349541,0.8342813444128557,0.860283949159396,
 0.36687735736051563,0.9603639357651785,0.4839536710466491,0.3989152482464634,
 0.8660437140275494,0.33762504150391515,0.4452254240388197,
 0.009421644867937506,0.46896171100216455,0.8122167937313497,
 0.5128629343828254,0.5559733154826331,0.8044368265118057,0.6554416153543093,
 0.05755832764235691,0.9929963171019297,0.9513776232102923,0.4925169567980663,
 0.09709855995284977,0.9924034063834776,0.06929852421383498,
 0.9679688002920974,0.09261749824722942,0.3939712807399173,0.5233116994216759,
 0.6775632757289648,0.36207481089765425,0.16260120505386189,
 0.8821497808322901,0.8468277343623765,0.0017430899050098203,
 0.9342635922078536,0.8664664333536,0.9525802259347287,0.4540710707386987,
 0.6627448951194682,0.3907212488431614,0.43359027591081034,0.9044351331526206,
 0.265858690057478,0.6823317214379613,0.36611519297901896,0.9836474222433437,
 0.638093629654245,0.4627499415061068,0.3111356479486621,0.4507134006107031,
 0.5061441318362606,0.05132022943789105,0.8952862412852506,0.2935829913853306,
 0.30492432773364087,0.8918121776887855,0.2128793633423346,0.3269164514484938,
 0.8805741966337212,0.9971310620193669,0.0662982329540528,0.16966764723099204,
 0.4501086726291156,0.9322095304507945,0.5245102242051987,0.669547009634633,
 0.10478371007432463,0.3081263460915121,0.3997486733091671,0.1309011369397186,
 0.14035413878360736,0.035001486342308485,0.20826763495242828,
 0.19489808112321783,0.2738899222647847,0.5109818107649621,
 0.23870950900725407,0.2841186432719882,0.43222773697768224,
 0.09031002701164548,0.31082768301546193,0.4447432557956883,
 0.07022191206784523,0.6080364471101433,0.33281648476780223,
 0.04704009507500784,0.9740172681254854,0.45524872383932324,
 0.0873361637819916,0.07206735432204092,0.9333039567194321,0.8913188267090772,
 0.3752439718195937,0.49707027498951717,0.7209227493300955,
 0.30189792232843726,0.5587701848124902,0.33829261074074846,
 0.012972148256600313,0.8452194704564191,0.8806389998971589,
 0.2802000204519244,0.49762798570005273,0.1618472063378953,0.2823978811329946,
 0.3896257279908959,0.1114723742221404,0.9015083467807131,0.0386799936557326,
 0.7838229160542824,0.7378189778184314,0.4927561815616385,0.8368756167986365,
 0.24354132308260634,0.5128576545315054,0.17119374093361917,
 0.8407581008186197,0.5758421599867763,0.8700904970249884,0.5673694551441315,
 0.3870297379016562,0.15774846514282026,0.6290075400681038,0.30013722463878,
 0.22568961069193993,0.032592374256637635,0.8181141516383138,
 0.5640715583595279,0.8146603022946172,0.6089806541476097,0.5313474401919591,
 0.459254789441388,0.6641926114946849,0.1615769565662526,0.882768000960943,
 0.7915173976428673,0.8046998918229614,0.3672578686700759,0.2277211402454431,
 0.6549528723344827,0.24673820649014677,0.3851347977582361,0.9060133515612125,
 0.8980830478243773,0.8551379066616664,0.5996246246360669,0.04447118787999271,
 0.2083377704890339,0.009147679032875722,0.2725247354298417,
 0.9692310333610692,0.4240171061851612,0.24745625513534675,
 0.07324152768854941,0.7647145018665273,0.4232985126328932,0.3074005977979771,
 0.4508355249221342,0.784147393727717,0.12942183216871272,0.5831317401507339,
 0.21897020183974736,0.022699001132975627,0.31234560740844686,
 0.24245682169408322,0.9972829919898091,0.2772911091428586,0.7706449386910813,
 0.004990971332897498,0.5198434052532821,0.9787196120977342,
 0.44309473802240074,0.9138399201711973,0.44106097998645144,
 0.7920174666458106,0.7393892097420561,0.05800636014588845,
 0.11659665487320459,0.6389797178043203,0.3719313570209455,0.1406706667301877,
 0.975756777001024,0.6903633182412954,0.3677055310818621,0.2313000183271826,
 0.7909933199379648,0.7048455599071362,0.16724230678474117,0.821277101625877,
 0.8216292923290786,0.18326251816492056,0.6994419642191939,0.908693899334781,
 0.3533483122805656,0.2347983284648706,0.41500267487405307,0.8812688040358262,
 0.550804799465013,0.5272617660824926,0.9724018097922543,0.5757485743816734,
 0.38185178409037523,0.895828026099878,0.7403289393029804,0.636171467200605,
 0.008299357069664337,0.2856011871228268,0.025770273852710468,
 0.735554486078298,0.41487597098537,0.9562882926255366,0.3139239386564848,
 0.25663257807771933,0.4701096843847926,0.23823126167844377,
 0.16686750897457303,0.9268372616300632,0.7542656118793206,
 0.15009192001408023,0.83165079852977,0.5277829760474039,0.7415451085974256,
 0.9314690843862532,0.3380874261965675,0.9193819698221759,0.8790254628092197,
 0.5021389187572748,0.3095782196378951,0.5040750879212015,0.437456392436393,
 0.4954248165177728,0.7530387867798533,0.7702667425001585,0.6409304632380098,
 0.9602323739973367,0.8381021406235333,0.2074380607367199,0.9066545832086786,
 0.909135780772677,0.0038865250664695816,0.35068947730814737,
 0.4882836118783447,0.39585933045083554,0.19838586920346535,
 0.7027456209160218,0.3373526857728605,0.5438134110464151,0.6617019762076596,
 0.7409282840591469,0.746146128476946,0.07185148045755929,0.9475259762451003,
 0.860439343390702,0.42796904805387803,0.5495213688423508,0.8501188729566895,
 0.7964898931072811,0.2937281580036165,0.3336167608615227,0.6672419473902239,
 0.749553885704378,0.9222530893714564,0.21063169919699432,0.803498160889051,
 0.973245395063688,0.5104905729416358,0.24285309249629206,0.9673975659373301,
 0.9441704733866769,0.5541842419974958,0.5939682157500148,0.36192678084540775,
 0.8172522675264113,0.9480265098227294,0.27930040058730776,
 0.43555560246768366,0.3476053157984721,0.3554009674775499,0.5330643960437628,
 0.8412748544956158,0.0014793528304093417,0.3958077791013186,
 0.45009362478027715,0.15889016752697005,0.4656561769757205,
 0.4342764638937381,0.6234654313713237,0.4868794330887466,0.16307117158234907,
 0.5204971197636153,0.7807758878868792,0.580002329715512,0.9875829359229671,
 0.2529410571937105,0.008822351496790892,0.2678978038703941,
 0.7513339153281804,0.712514849247194,0.9728245258740774,0.9859594072435861,
 0.9634227637980085,0.19359120691977283,0.5110451127599382,
 0.08313208809983352,0.14267612538539332,0.7895317516785607,
 0.7057570668705142,0.37421525714991,0.659847381559662,0.9519013479330467,
 0.9669125577843211,0.1755984615702899,0.17935178053328826,0.7842597615390107,
 0.9732167054083005,0.9999417473421085,0.32722222527940414,0.4138138736397624,
 0.5136184928135771,0.5662146996594281,0.8974053271384308,0.7380384204005609,
 0.9316882313927675,0.8859107042148979,0.1363686936114049,0.348074332990412,
 0.050350489781255336,0.05133361160130012,0.7026024580790989,
 0.19052613066252788,0.4991886304452484,0.6994532516627521,0.8855119250097905,
 0.7491410887533402,0.40694922315380955,0.4725053698281174,
 0.026830866560146327,0.378853620577311,0.5649935925147103,
 0.14183439006271292,0.09936040308462879,0.13344258057908642,
 0.8449832443005619,0.9656404134479297,0.2840284903863336,0.2100124073102988,
 0.44992704984296095,0.6971456457756642,0.18014818116533715,
 0.02531560859302573,0.8031252148314633,0.12331973618995806,
 0.7056272771708636,0.7200575639854222,0.6166816898330908,0.9096647707919026,
 0.4634926622726391,0.3326846106241761,0.9483691170827372,0.5886948446878052,
 0.20084435891535923,0.13365109500473826,0.018474251928719876,
 0.21221464737555296,0.508218981537435,0.2682969717141612,0.9336996942734972,
 0.18969124281189043,0.4296243133430159,0.6906879908932066,
 0.10078596673674933,0.9524849443188568,0.4472895978578513,0.2002631804153806,
 0.09285762473730419,0.56854940711711,0.03321859452317932,0.6437462483929544,
 0.4528608618884,0.5428660348405243,0.710997635501134,0.15054184229820455,
 0.20830824070980736,0.5510901702366285,0.22076346053094675,
 0.8520733102360628,0.5455036897447012,0.6186003128953569,0.4847401456524064,
 0.7926669760475636,0.8457267999943905,0.7984610245125976,0.3734009372223256,
 0.019330316451814378,0.3052320850306697,0.7131042387417966,
 0.9586752716383022,0.655405800721229,0.7769751383454656,0.22408628031394517,
 0.5304610525562993,0.2768323472352656,0.6540695268627927,0.7218828795081532,
 0.0059318031725471565,0.7551592822910174,0.5977289859572152,
 0.428329950297219,0.5758239581799046,0.24822236888864224,0.4629998682572416,
 0.3274586123865183,0.22519234394258913,0.26484140054791994,
 0.14217557825449167,0.8539423645438196,0.34354765546572696,0.784328438255153,
 0.6454224404166519,0.49127398390384724,0.12596307794645567,0.722591971674817,
 0.6790729470237573,0.6561798529948102,0.16835588904965437,0.7894371652913261,
 0.2103322973529146,0.22325574517438607,0.33423325431865725,
 0.48600230070582273,0.7821856695054885,0.7018092448391776,0.1918423190399351,
 0.454581376311168,0.8226368300053627,0.16798657586207,0.6963575804738626,
 0.307536936231803,0.2422052846040067,0.19548860010669045,0.9554717951441283,
 0.8609938572380478,0.12349692872996099,0.1597727898601824,
 0.16215170544499202,0.43994049595647333,0.785096350400712,
 0.13320796887355524,0.9149052592815561,0.5106892539683139,0.8621180068607668,
 0.556192773774536,0.9409464756939085,0.6591203051378446,0.10168244882525879,
 0.9299105930703346,0.6647201293138633,0.10456001513977176,
 0.07540912381684617,0.8128069095374787,0.6650973242216178,0.9861431240220335,
 0.745029763244977,0.7815914128063804,0.6903675449303871,0.010345050708482129,
 0.8793182507960137,0.6638635038230155,0.0957712632583646,0.49122500890194787,
 0.4364067702348604,0.14371639518828971,0.8614244663810354,0.6339156895250588,
 0.8868353189515807,0.7391128388089188,0.02293681450110252,0.6996431388418884,
 0.13068815093830977,0.664603251115562,0.5420569339271668,0.1311343293855184,
 0.2914761034909038,0.5906960537654212,0.5178716782062115,0.8300583531976455,
 0.15101154327805477,0.9469351375204872,0.9407745909430768,
 0.04724550498697866,0.6546581511868391,0.2660086737673083,0.5959602668470326,
 0.6718829289354186,0.23117740193701464,0.5139815453262958,0.7068228565958747,
 0.4777109607264576,0.11684410639414489,0.678526604794536,
 0.010487669132869026,0.8026781040580788,0.2904925641314403,
 0.6105776509209554,0.8620643169160572,0.8517752719261148,0.18012436969857348,
 0.9502917473261804,0.5446347336442124,0.013466966337317587,
 0.32824786806191875,0.23647319021772772,0.10530485828981873,
 0.934105978982221,0.7223358903564551,0.49884883850106787,0.20615485055225247,
 0.7407148789123087,0.5524276368921023,0.18323234656529364,0.517931164490039,
 0.3945250553414784,0.875505787790622,0.5181662819329315,0.9441455664677461,
 0.7089195391953714,0.04676180147150688,0.9293441981952402,0.2569201022299771,
 0.35373524849362115,0.32807179956636645,0.4523643131493156,
 0.4584015991052871,0.7162770808601209,0.30491555276004945,0.77201592019084,
 0.759507149318753,0.8338368084890101,0.15838372003358814,0.8249614390177452,
 0.9896304704632124,0.8098661219226265,0.5450982706359172,0.7328110560890386,
 0.32433189990320377,0.6841482796062033,0.41955765751902163,
 0.5908474743322367,0.49189838847168565,0.14935499775581185,
 0.5286416811851598,0.32650767935045255,0.49532310278796965,
 0.45553691706302946,0.13501613541907465,0.8955834917973096,
 0.2632230628332337,0.777113490606608,0.8548448515424827,0.08316130142699496,
 0.21212339002838498,0.09287235941758798,0.6331588304852928,
 0.4722252019643862,0.7706209406267048,0.2913306056668421,0.3187238735566972,
 0.8672844863916933,0.6754790693318848,0.6349507448543736,0.7635012371513403,
 0.3536839646592267,0.36286416457674875,0.34617951099440836,
 0.1211570861081801,0.22783831127032017,0.39935309390977825,
 0.6689654883527171,0.6266794351967401,0.4079089941240911,0.47425761840955283,
 0.9687513623400462,0.12003273326514319,0.222054748874277,
 0.051875289849499184,0.8277821453233838,0.17685683994740242,
 0.5513520189378267,0.14924206715341415,0.8479164295450586,0.2741214467525105,
 0.9830839238915372,0.070861947975416,0.800681014843427,0.6294853153074046,
 0.8241829577701619,0.3170087775608902,0.7905007653655665,0.17689168533106425,
 0.3091614657153632,0.38305986444126483,0.9551856729322793,0.4944198632610901,
 0.05511094948624973,0.6789884879009799,0.5047266577088348,0.5469549189982654,
 0.6101040540011375,0.7570799763974707,0.5515756685912381,0.05745898523894133,
 0.9714972801485653,0.9883899007401113,0.7067127951887238,0.19268582477074436,
 0.4457258767728548,0.013340142000011102,0.286362628630501,0.0805246358373728,
 0.4301255022290922,0.21410589729866236,0.6909795025083569,
 0.44483860919922924,0.4187966439080617,0.9266997550577938,0.4299188486615507,
 0.6315718837980712,0.5078387816144412,0.2055594633914256,0.30238067039846417,
 0.4157131369106324,0.8367441253960268,0.8694889862734516,0.8422659134124474,
 0.7099880878988878,0.16392043124798328,0.6367241025420532,0.4488963919082565,
 0.1419489370557866,0.48964295581604367,0.6378391168479899,
 0.19178523521234458,0.12283403889212419,0.19451311236802404,
 0.8671816719818429,0.497304747178659,0.1415652048173056,0.6520466367768953,
 0.8452621588290332,0.03666391887876652,0.21671741379764942,
 0.3491880621388219,0.8111243114471627,0.7967727169590282,0.7775851204764215,
 0.36257744179900786,0.3529547970426389,0.5264592408800904,0.5569327737027532,
 0.9370043725793121,0.9550472576893696,0.2802732706540121,0.18991311302793468,
 0.02850514723206332,0.637927314139071,0.6995877389326293,0.06565539263941433,
 0.39967068638493675,0.9121498527099873,0.9892173141322744,
 0.021988217757198614,0.15924660485358888,0.4961295090042197,
 0.7585113997087773,0.6867545844105692,0.21307172295079357,0.9733242701862338,
 0.7592748082580231,0.8519140112611733,0.1973040040290096,0.5721884456130327,
 0.8028078132512487,0.6654976402847924,0.9698991841467105,0.715255908282732,
 0.5242326170772457,0.5314837165479562,0.9552835843565943,0.4710705556389393,
 0.8108644856469687,0.04298104825243343,0.46035599534451144,
 0.8396190620091525,0.947485332792942,0.14474651062169075,0.9721150917021465,
 0.8329514998755779,0.5240133763677053,0.3311726671096663,0.10215740593541478,
 0.4588370315504686,0.7770961061204332,0.7489693952019507,0.15587940103405928,
 0.17205193431155008,0.5924882119566421,0.8728088029360547,
 0.09543616995215609,0.4936898157445362,0.13334108985727033,
 0.9736374990086566,0.09256021870511544,0.3482578053094668,0.9610005800290443,
 0.9418005674982691,0.6845570775751973,0.4477143268608603,0.7754800749297561,
 0.10197302274078934,0.12790706064241764,0.09369488194142794,
 0.05718494409380814,0.7328844107325744,0.13562985448530818,
 0.7364156332446997,0.5786974708712029,0.04768869370050588,0.9697116695757155,
 0.8525054957109666,0.6385608493446217,0.6674871023716172,0.0242828983969402,
 0.5901090081228737,0.7641861144518779,0.9309092253104888,0.4712793093120088,
 0.7581472571958489,0.3493346119820121,0.6722518031556911,0.1968114329169205,
 0.6898760532736372,0.4909645529100203,0.044163717010099246,
 0.9281118616491343,0.5449472287019534,0.8074308355672527,0.10354777843889496,
 0.8276560333465178,0.8119644379044886,0.9379542618228447,0.28830059310906264,
 0.31568993214710694,0.12098511248099042,0.5643187233746391,
 0.34185603959890787,0.1513461503232736,0.16920256992646832,0.803622465593002,
 0.6362589354440751,0.9055700217120174,0.008781565106085298,
 0.06544617847063028,0.2089010914963627,0.6971666610892545,0.1760945540604707,
 0.15874192061931713,0.36610440111649545,0.6736617191527182,
 0.010465328907419469,0.716534657379702,0.2491629830707075,0.7518608155482266,
 0.8677569657406429,0.609572747598226,0.159295574031444,0.35179060202417245,
 0.7625511492519884,0.6789074269819896,0.8426348647629269,0.6441834453246031,
 0.7884591497318438,0.3330808037076417,0.1476817046550848,0.15480301017593534,
 0.05495114322318573,0.09662476805327458,0.3695470715311401,
 0.4240371286950029,0.14193201715952763,0.46204938599315026,
 0.23049483653019376,0.9816655191876366,0.5044141310395165,0.4799390856591421,
 0.9448750684136296,0.5372980668696958,0.8276503651354652,0.971000236701004,
 0.4451377335566893,0.21153212051705905,0.5581434929848843,
 0.15534523161190528,0.6671527387488934,0.4854058968341287,0.4513014643124005,
 0.2167032426791109,0.6298244035787834,0.48516823218237715,0.9863902794580579,
 0.945721067991801,0.9263780595796562,0.11678171367001766,0.8358344127941366,
 0.23145938134249788,0.8049277467378833,0.7158898535673048,0.362592706971537,
 0.6120186223933697,0.5516442938965188,0.0139871501729526,0.1396518865809846,
 0.7796590254843911,0.13402091187344667,0.41325315798483253,
 0.2966806566276621,0.960822254714787,0.861859741598315,0.20764687554321204,
 0.20641461785577264,0.9306846066568383,0.12284536138111313,
 0.46456489863924144,0.03868056932636166,0.1612219276784288,0.478644081237104,
 0.12316568641163583,0.045330309846498995,0.47943165435271773,
 0.16107552335357433,0.3581444943964551,0.8417658153770832,0.9589244722494243,
 0.7683308113307131,0.884299088953909,0.550876147707179,0.08591819581869142,
 0.24555047049810808,0.13020714506242914,0.5433852478525701,
 0.6163195981895844,0.8914963340266143,0.22060838629975266,0.8809450509265709,
 0.7351994002719326,0.3485674702729794,0.76173563458918,0.7407652474509837,
 0.31389414584706143,0.8722566848168438,0.5361528545354367,
 0.22512295629493473,0.04633017462172506,0.6675014706631871,
 0.48357885641981624,0.07935795845493199,0.6418997384600491,0.952833624340018,
 0.2426504750351003,0.4705151713611472,0.20221748457456667,0.7750831545268791,
 0.40489386333358435,0.2620429534176757,0.5915946042879358,0.8448942049567425,
 0.7113663203283614,0.4069823983678754,0.3868263382596161,0.43546184869609494,
 0.47805911548248803,0.564671835933797,0.5697839372002914,0.2948755495072507,
 0.1685900322221564,0.2939260769464219,0.024055445994298652,
 0.6782698362663298,0.5968507206991093,0.9414681419976183,0.7630606588330425,
 0.7753220562149575,0.8346717991204946,0.2632295582637254,0.6282902035232333,
 0.6249417794923113,0.549486734498349,0.21942502326294233,0.828686616291554,
 0.6458052915252686,0.7376356404137746,0.47580237057784125,0.4649104462107889,
 0.43674257759485147,0.3242227182009709,0.804409787923074,0.8052121211122305,
 0.8954103926340301,0.48726222181690193,0.2348323034085964,0.9956455308551624,
 0.785993131664735,0.14130650768193154,0.1869788472919629,0.47468765006886093,
 0.34879973391178976,0.9333029519067437,0.7243504271905954,0.3345980891755088,
 0.253756066424481,0.27853580835990255,0.005255738755691297,
 0.16490492239839383,0.3547616815295107,0.4752187137558864,
 0.46003489330963954,0.3365951370639807,0.3240803983278515,0.1427260256002001,
 0.04596610874722862,0.23670431073949194,0.5980354242690877,
 0.5029233457802998,0.018356498322841253,0.9012756207829118,
 0.046041991549620964,0.577235443818592,0.414566759424445,0.18413223500730236,
 0.15199591128027656,0.19946867396009926,0.02366376333652076,
 0.12910089440577477,0.6254719387338781,0.5717614979683479,0.7774495336502291,
 0.6260892751272532,0.8823956822129128,0.7048356370833155,0.16662689865654468,
 0.7497988409475734,0.49848257183534683,0.3511947704065793,0.5600752847964635,
 0.0011343410075239646,0.9250405057937243,0.7140089562529534,
 0.13445344651528734,0.9644900438327073,0.6009247936491685,
 0.32902581293867217,0.3642380395307945,0.07313899136469892,
 0.6392501201189454,0.3071953385710996,0.08506006649670639,0.5456107951869523,
 0.30867717887855517,0.9436467262875731,0.09163116992483877,
 0.31000763844545476,0.987914227046701,0.6211106037362563,0.4240328951857405,
 0.9414176412080544,0.4421455682119029,0.33413611143321675,0.9686922175688102,
 0.9240176293456,0.21100887789780798,0.4147531205770173,0.6131024201903013,
 0.02496006699423481,0.5427369195508035,0.41049030578149304,
 0.32469369212744537,0.2948508102241032,0.30523875823385116,
 0.6511150518807673,0.8795001668535747,0.6416996276123834,0.1553277007306153,
 0.05992331304910814,0.4946811955220508,0.8540996611775874,
 0.18907142384987152,0.6155559931595058,0.9731578020992692,
 0.07821249593063673,0.9221620095675661,0.7034547361796456,0.5713820456285488,
 0.5476188528246567,0.2835451394189573,0.035002111298832794,
 0.42056699869261926,0.6083547789207444,0.8477733110422313,
 0.24143742730760043,0.6383452278468518,0.4263561063816397,
 0.21567603992657147,0.33554493998368007,0.5535989463339787,
 0.7761808262218015,0.5626910033297007,0.5564442628680834,0.3838262075152733,
 0.0612482585708396,0.6427487135754391,0.46028609825610634,
 0.17709689059148892,0.3704608487480594,0.17997905960556904,
 0.41538173293292735,0.7855239056436365,0.7443495115352996,0.4325391658153911,
 0.0035845525884297125,0.9787611686953916,0.013564874179583963,
 0.7740497448834743,0.33134601686449083,0.4897634648409165,
 6.246275697736614e-4,0.5144659565449166,0.08727377639696654,
 0.47019288866047115,0.8799225948042392,0.1489944985751641,0.0603701890453886,
 0.6604608023722398,0.9338667260205452,0.5527452613290831,0.21681515342648572,
 0.46448639916030565,0.7409077930512037,0.7142064258387175,0.9104890915719542,
 0.3671465987654413,0.3406369521523417,0.7807331918633126,0.04845084948343692,
 0.3668235780152076,0.777772579159147,0.46512880784243704,0.9532505697603851,
 0.15054929278037454,0.9210960750509288,0.016081978674231112,
 0.6129604740498654,0.3735133069033587,0.007897340370663142,0.983756600296335,
 0.5485140086439532,0.6196254869469696,0.8359268764031703,0.15351386010449564,
 0.17839373678375137,0.2138547727948753,0.11391223030955977,
 0.4904205797179447,0.996785321512603,0.20226480819131898,0.2706480041241355,
 0.7764534261622607,0.6849532116502713,0.4472115646510215,0.09764748539559909,
 0.6996444559827742,0.6415012942185695,0.15363302495616504,0.3142633038588718,
 0.06702076290841452,0.3901686931145709,0.7380865339583894,0.956227913358175,
 0.6205153563435772,0.6560252112612032,0.8919140256029888,0.4576555204801993,
 0.863381084522623,0.6952309595918045,0.8132668356416727,0.8297179517020509,
 0.8194264846418265,0.9533097010368413,0.9562279821452143,0.2220027405760685,
 0.15286920107930002,0.8542637972116869,0.5237872676438724,
 0.15801085645972002,0.6218074203689228,0.05590717315897287,
 0.4796259967342412,0.7492402296563585,0.3843639770159266,0.9517283547574947,
 0.5083845084300737,0.041403125594909884,0.19532793772613266,
 0.512837875391434,0.8689010734495509,0.8986007822237857,0.48902338804291,
 0.7918758536415377,0.0898796087024204,0.36702557767987487,0.1712353787321047,
 0.05583765294397436,0.03193996456847248,0.8653187168252554,0.851314934089565,
 0.7597522504862272,0.25884516338594465,0.43428702441758127,
 0.32999154313222345,0.5503155427243119,0.7143760606787234,0.2740702312658325,
 0.33037925080794384,0.20359395117728796,0.6649486278106954,
 0.7178737945183791,0.7301042500500771,0.7797531255274767,0.9075664856001922,
 0.674544335517827,0.7729546254880111,0.19357921494966035,0.5630312341158277,
 0.4688640066926024,0.572840351152654,0.45543495362485176,0.14447988362099018,
 0.7454391075040123,0.3840180610851326,0.4225819533678744,0.4541341742142556,
 0.8924784957820637,0.9455905179723192,0.3784993111316325,0.6601184395760145,
 0.4807222580789626,0.05582087012420478,0.9166591267001336,0.5060605634700862,
 0.6135060957224033,0.9340335131990164,0.8829882751471335,0.01585010444096857,
 0.1811169114505342,0.4341628194743038,0.15747630526977252,0.9833143715532283,
 0.6923978802313567,0.11366259308793969,0.3772282259367422,
 0.23543709052683226,0.8157207365560926,0.9035087458551649,
 0.12949942128106073,0.5452783705641482,0.28808296091584773,
 0.6250034625449793,0.4481082149779164,0.7471402351115662,0.9642438365072161,
 0.9800568327112641,0.8908386772314584,0.2707189888809799,0.5629041151604417,
 0.4880964208716416,0.4099108313454155,0.5486451336645176,0.5463763096637569,
 0.39208630162384883,0.23706638002478853,0.054059236231599916,
 0.6948074414512103,0.9640494942440512,0.6003044857764602,0.36681378931402153,
 0.04220655358478298,0.7034784465710675,0.25185188847120643,
 0.9050402052776966,0.6917307002267327,0.7536910346376671,0.9067818863609457,
 0.4198324720156995,0.7490124853174256,0.48650831869225375,0.7470940434376941,
 0.25010901081456083,0.15761449398502114,0.9118102727331496,
 0.03263503613219465,0.31973549766302656,0.8102527353752059,
 0.11868056955156059,0.25800738861041494,0.10204332436564556,
 0.7292317899855043,0.7453029353778959,0.3647928554711555,0.461567149859492,
 0.411379257286006,0.48627040590398696,0.38501443217918463,0.9278083552066899,
 0.5177412746490825,0.9956237211964647,0.061068980370669546,
 0.20775991805987104,0.2387916976269795,0.6675500338286141,0.624911196647741,
 0.2829852573460596,0.40609100285949706,0.6319294175549751,0.6327363461268769,
 0.08855422799007961,0.7208650201210798,0.26650117470244583,0.564221371231058,
 0.31075124866475756,0.3994350533740866,0.1659453090766182,0.5085855154449668,
 0.18711609962463815,0.4283657035655398,0.34670509319041487,
 0.35615629912811564,0.3204148663210118,0.8061838828612549,0.3192786210744689,
 0.7032367271214648,0.4063091217637266,0.8113521406768254,0.9667094580775419,
 0.2582076975315213,0.09220935493719173,0.15926296697984332,
 0.6025908895180296,0.9663897423583216,0.23683401681649618,0.8258451512194279,
 0.8057201361488507,0.808733240633569,0.20505290894803097,0.9898869318581287,
 0.15516957682813726,0.07691061978143243,0.6735728013022588,
 0.8022193564168498,0.3642847447063329,0.7634025209736351,0.4195844450102747,
 0.7567010591806711,0.002643620756138998,0.5497706155186525,
 0.1293389150633455,0.7842957503058081,0.23471469498715658,0.827457121048893,
 0.27703066360622086,0.04905548878380539,0.21946296480846705,
 0.5977868203691948,0.33795887613787523,0.2787065729216409,0.9773924413394677,
 0.9414600480619991,0.46420967713506356,0.484120665682796,0.893766214896726,
 0.8660490502126366,0.9739306409860681,0.17651011640499192,0.2544437107749731,
 0.6871178097252898,0.993883020149142,0.08640021788450714,0.1188454211830291,
 0.9751785631583603,0.38035573629812514,0.3428259726446202,
 0.13585801945471987,0.5685948122115887,0.491550687177442,0.4572992465174974,
 0.7353207125313602,0.22848455362277864,0.14968726170788083,
 0.6130802922933503,0.5205476456401861,0.11357689995101783,0.2058971341771907,
 0.4545527658117743,0.5149487885472681,0.5727236111035964,0.5732775191698897,
 0.24859940121624335,0.5911392737625805,0.2447114378997669,0.7783608531671078,
 0.6175153908974109,0.2990075561704816,0.3077467701370705,0.7854992023793985,
 0.9260735755710896,0.5855775847982616,0.7480765298182758,0.9294767255200208,
 0.7994155877769962,0.02108970783342723,0.7129002121181836,0.7442875053019364,
 0.06854771477199739,0.8206037102311096,0.7075991257906421,
 0.21806956388394627,0.7516038130162475,0.048480114790859896,
 0.8729611551472903,0.41234677005729703,0.03560946678933008,
 0.4080576520866305,0.9810502244400257,0.603846470013034,0.9980621932474112,
 0.282727069411827,0.4137262486716192,0.13406556459810226,0.36057081274986635,
 0.6583545547129512,0.15534581923166113,0.872387818842675,0.1725097253002933,
 0.33503275650773057,0.6018154283306356,0.271957866148421,0.2438764970966979,
 0.922394758893397,0.6493290132180248,0.7595511262780547,0.14422661215849408,
 0.30011945306411614,0.9508595904964281,0.9000911098970176,0.7830335389347645,
 0.22636371476200412,0.7881163007035341,0.7025385107022147,0.4114840283772516,
 0.6943196979899958,0.22674931551201538,0.5500582848466536,
 0.26125173784376143,0.7530790309147741,0.838238067603331,0.2579971736426233,
 0.6769193799600446,0.2093095744385145,0.17567336198954953,0.493754642910859,
 0.2780657961317856,0.5353193378552294,0.9079048470191036,0.761280286742305,
 0.7704763212027652,0.13364099154153286,0.4527791478183252,
 0.42311862948873113,0.8516467592198256,0.21812587878408818,
 0.33332876171318704,0.47791204964153533,0.0027416599500221794,
 0.3091746939819604,0.692922737483476,0.9755310901687517,0.9979534640978575,
 0.350815120225332,0.8222111524861454,0.9080169096850793,0.7297104771483666,
 0.701347156859871,0.4617964789470952,0.4873701003432641,0.49524047364113555,
 0.5196915743461036,0.007735361059762469,0.6294549521590914,
 0.7727762796858086,0.2737832203998636,0.698780081441313,0.19006218532648844,
 0.9769583292149427,0.3199273330128739,0.9889710624455619,0.555854078412985,
 0.6851767795273112,0.18469944584452225,0.5526708621440367,0.8427151888589972,
 0.20193922854144541,0.3813342634998502,0.7532608443579923,0.7913576304524775,
 0.5150607549334475,0.996955774665053,0.09251554314091026,0.0705244437428082,
 0.6352668657182942,0.09420013038442465,0.46077958905815963,
 0.01966862839783512,0.9359527658257716,0.8732379720484822,0.5983561849224102,
 0.14302172997875084,0.09847186368189464,0.9764173832685755,
 0.2385278766687341,0.9024935723515268,0.41918219842121207,
 0.03439760355977284,0.8496841220877592,0.5969029005331631,0.3789604544093119,
 0.421088054727932,0.08620056268238918,0.5071351223891786,0.8082108712347923,
 0.17435803461622013,0.07241523814789408,0.9236433249097494,0.586899429043942,
 0.3991089445305923,0.709568985804705,0.2851854266363276,0.40933495303611434,
 0.824209616847618,0.4089949477219752,0.36614789868597797,0.9112792275395313,
 0.5040938053249022,0.6189828192941108,0.930859661645544,0.26641783634506333,
 0.7895749595772104,0.9373036277850699,0.04368974009282178,
 0.04657001527027749,0.4289336271195656,0.6515689341261163,0.2710732586982393,
 0.005920502694863994,0.301168770540689,0.7439111430184243,0.808318450790777,
 0.16011159132499964,0.9496543788246098,0.4039416253605863,
 0.11545463223989505,0.6840953212195773,0.952027918969636,0.03645748837285989,
 0.6360452034346946,0.6307261877156641,0.9696442036591487,0.18038433667227727,
 0.6041766982232643,0.6348070959412314,0.20612143072398115,0.8010950960357681,
 0.6753689911247822,0.9036704061433054,0.9305620510163859,0.8392727302618489,
 0.7842698248649214,0.918007967079328,0.9589705760962531,0.23917717668979943,
 0.30654776874613276,0.35904527178012213,0.7350286235055048,
 0.2581013854045273,0.5410810843806103,0.0925437309238556,0.7063544977211877,
 0.9553942749443983,0.64097034707669,0.24813651195545794,0.8590132850186079,
 0.9992434778779714,0.23967089161930233,0.9936170135113034,0.6231431381129557,
 0.057685324383025205,0.545050706767313,0.749611710645255,0.5143025225200353,
 0.05902021669522295,0.4741485253188803,0.07186866243102408,
 0.5334970842813715,0.09328261153959694,0.3884891032683617,
 0.13668145222046946,0.26034876015634434,0.9160065440993782,
 0.4683882920406385,0.6861921910004529,0.8715362818852949,0.1064469204716616,
 0.6107017708028267,0.6468378365373529,0.8311441228000291,0.6127372979764301,
 0.01927060882995968,0.5488698467711515,0.2229397476374031,0.3503160896205928,
 0.7249486133726102,0.9485217410702947,0.03576295618148917,0.9238167681984335,
 0.6201080635138949,0.24108445439060033,0.005974564640359836,
 0.5776424969677129,0.4916164598082966,0.4698632428141378,0.4605497427111377,
 0.6493732258938936,0.7389051029464454,0.4602427018291111,0.1852398184317342,
 0.21708697121717346,0.6124084538536587,0.20155552127963805,
 0.004085770688143309,0.7613711238105265,0.9455646668351587,
 0.25619343168516906,0.20148609324139943,0.07495235975254866,
 0.9513616258936536,0.27410407855264707,0.5541880988984224,0.9506608219193639,
 0.7882837780252331,0.5866979077783075,0.732722260107344,0.16775657064787464,
 0.6267433336020829,0.9453266050856474,0.10398265623051628,0.7848816521150473,
 0.2832954560230311,0.8884379120408542,0.24932498615795806,
 0.24137743766665554,0.8213502395762804,0.9938864332724839,0.8539068675100123,
 0.7293944362533379,0.04784176417730812,0.002067186240166441,
 0.3441331306320645,0.26817064375511557,0.972189899879347,0.3703522729410982,
 0.36916416308967825,0.4435738967255014,0.1633242407497868,0.7831638045344097,
 0.5790401100637642,0.7312844691171825,0.1771318256487292,0.8611039081890074,
 0.5663728288642942,0.5669081658704733,0.46154445689690804,0.2905079240127352,
 0.12405027406223901,0.4451760872138002,0.7774902321521467,0.1182916706904894,
 0.6764704237268997,0.4042646069950584,0.0454334397267937,0.47216894791152797,
 0.9225110399464108,0.38483515941529856,0.21842179652659566,
 0.3919582489139849,0.3704068519569368,0.6859872029594395,0.3925275959453234,
 0.7664611609477032,0.9899206486374555,0.3203034739770748,0.5347836897329091,
 0.22895989522705007,0.2901018579841952,0.4431292146388446,
 0.04066835921897638,0.7480896240816001,0.9273654089978668,0.6177567698896462,
 0.6423773594594135,0.7083385604867247,0.41886996242062224,0.7437925389877215,
 0.5391681516344499,0.6510177248371684,0.1854913678634238,0.17158599700886035,
 0.9918490643224988,0.9066649741807847,0.3136255187181216,
 0.020977019437784516,0.6200800588924442,0.6979002571978907,
 0.08867625209346752,0.2910164227673855,0.8795244720139085,0.5035794693694495,
 0.988716031982339,0.4315672756472325,0.8525363562399246,0.4742812390407164,
 0.08020056791886565,0.03388573372659387,0.6291085764011057,
 0.5237419713434561,0.14663635071204317,0.951207601729085,0.32552143292642466,
 0.0666144060787367,0.1268346429794387,0.45783469528319376,
 0.20359843955983548,0.4045624488450351,0.5538882214893208,
 0.15236878317723412,0.9344542852242961,0.3818709506611815,
 0.18153952712329013,0.2668955761234274,0.5411358650151662,0.4941238389586613,
 0.8329494562827133,0.5085110384236732,0.21184547792378672,0.5177554018486984,
 0.04973640367468257,0.5997335006043196,0.6369596748310664,0.6144347451959252,
 0.4074568336975829,0.39305712594166464,0.30577539527457775,
 0.4225225588109557,0.5280144466637648,0.3924177934710201,0.3003290395798327,
 0.4995199827432213,0.5052612753938251,0.24772862391086592,0.2143209063738809,
 0.31706043147196294,0.6777153580181693,0.18858895000106823,
 0.7915608191238497,0.614291168624735,0.6451343399624173,0.6503492299978715,
 0.34269900970628286,0.9908806608237872,0.6125245214960084,0.5439558354535192,
 0.32839387367966355,0.4236457802601514,0.23003114195049312,
 0.11036573000743544,0.01506515996112745,0.5185772566239917,0.133373387970962,
 0.8260383909259069,0.9191855202591867,0.31875251233418056,
 0.49586133818390443,0.7898001948528318,0.0959549250387145,0.3287619528338511,
 0.9153672499870441,0.461154925623512,0.3114491205139559,0.08217941928803962,
 0.4398280650044851,0.8054421333989759,0.1908440901173386,0.9531837601387968,
 0.246070179896256,0.5631578344992336,0.24197933499962787,0.9583479490623722,
 0.21710570042279764,0.9812464076910834,0.2941696902216583,0.9729622086354264,
 0.2383088593466871,0.10674170562019025,0.21153159859075488,0.608800312737558,
 0.6136778038205071,0.8850759553303504,0.8460727320179425,0.26449298655212294,
 0.9917456504571385,0.9851242148547225,0.1617899489973782,0.10094502912352676,
 0.40483709329337647,0.025470870332128248,0.6771411735071008,
 0.21167965647119635,0.6401274020239822,0.5820538866117534,0.7056031301865427,
 0.38979910391235084,0.03316203363868708,0.09362042990193142,
 0.5273136366793866,0.24174701181917868,0.5265719408381658,0.4476063686377335,
 0.27589556455580455,0.005764773008033686,0.20699302349614324,
 0.18560675554888606,0.9481884828421854,0.5925194510580712,
 0.27816265897840653,0.7550824088327582,0.47574565485272813,0.688846561755865,
 0.34537520959038104,0.05053036431269409,0.5791860179864957,
 0.2564950096800661,0.8766760491914284,0.8162013089575922,0.7677148169599365,
 0.23614795705048097,0.5601335351785598,0.32983734445595225,
 0.8963500705864962,0.3557723302987349,0.30631924624853435,
 0.49998102056411253,0.9542387058362818,0.17548246821382152,
 0.3740771683740973,0.5978661880171746,0.9466222650700212,0.7491140602786974,
 0.3071069561252051,0.5295409593593187,0.950141878185871,0.556340546507646,
 0.6819683275603203,0.4651422769077108,0.6786528221307303,0.4241549378161136,
 0.03777873136559129,0.5212347891013562,0.09751400648119224,
 0.8547819307075791,0.2664874347950754,0.21767319942663288,0.9919368419711982,
 0.5062569041456159,0.0991212250338257,0.5739506141959795,0.22653214226378737,
 0.4943417697069815,0.7803815111339041,0.8344910705212191,0.08015602397654864,
 0.4930218475407765,0.47389369379165414,0.8314928294411994,0.3954198361317248,
 0.220180273548798,0.25765959316384346,0.8665155748915326,0.657068144864075,
 0.7113875739581291,0.2882535969629927,0.2590165538148319,0.5826516383167063,
 0.5189548110929131,0.5290084952909528,0.7373605032394734,0.03153930549179407,
 0.32304778972367665,0.34156796390974375,0.8766534737800853,
 0.2620533932181349,0.5080846801943598,0.5153258581580974,0.3898612683906266,
 0.6072191876161639,0.4956897253236441,0.007318444895781706,
 0.2816038007945275,0.5365023891656109,0.7643420228162482,0.06344577299599441,
 0.22787922519770865,0.28505496436573086,0.14164601627888718,
 0.8670851144501762,0.1639018209989791,0.056362425425331875,
 0.5837462289598455,0.8835091568488583,0.44645962979098774,
 0.19869365626578506,0.03727171695516662,0.8736433909970259,
 0.6739009355691445,0.9593832796707024,0.5611518404089958,0.9570590459111973,
 0.6640391806398176,0.46033695756546533,0.4383840758417805,0.705812641896738,
 0.024675626418074303,0.9356497220096396,0.933870242888792,0.4306790674725931,
 0.9358720472770368,0.6404748144069732,0.053428759502798906,
 0.11118490252239499,0.7675372794260162,0.5247547081080378,0.5771720718434444,
 0.7052326122688956,0.6954216027337143,0.04713746627675075,0.7582717747582303,
 0.4894884876029164,0.8260267433872767,0.3548201301087631,0.8731209284396124,
 0.9033999780830294,0.25457861762469014,0.07320253055595816,
 0.6449594021499196,0.9185963911916111,0.6214831098290943,0.7390698769551455,
 0.5792850553521646,0.7938687063888363,0.027692350898844253,0.868582870224075,
 0.1630965470668897,0.1834463361387243,0.06858821024098705,0.8646315568448032,
 0.9970431036648189,0.36355575981283494,0.40615410178643596,0.470455455340544,
 0.30360162861893325,0.9211021634880544,0.5289297698113936,
 0.04501116889931944,0.5582713848708477,0.8157174238598126,0.5386621900764226,
 0.016366697203320513,0.1889358435576758,0.9355634571079106,
 0.25120587961356566,0.32715755654878564,0.1289336711469864,
 0.35816899904576716,0.2085728960118336,0.35730195835032463,
 0.8890284855638435,0.6211794923128907,0.08783995570843761,
 0.037926947866474436,0.6790842251788574,0.7103070798242501,
 0.09494831565467376,0.07516683024468784,0.037396729338217805,
 0.10026994140692091,0.32024779974391504,0.4103275905019257,
 0.27031559966222185,0.08958421141509643,0.911882623925773,0.8203497846505908,
 0.8332335976428543,0.6968982236928057,0.6319287803622993,0.07112673916814283,
 0.12385034633972869,0.9364401584105202,0.6199607755152103,
 0.03819614022348372,0.7839240864483876,0.47534581617881244,
 0.8412339196633922,0.288499862107108,0.9689364361129987,0.02604667493919821,
 0.9034319355133985,9.8833179490454e-4,0.36308952358337643,0.7929602757275717,
 0.33029388896156653,0.3600781478154985,0.3798971075938702,0.386298762583136,
 0.22563180002581196,0.1665743843394446,0.7774937659682448,0.4291354979694906,
 0.01640864684735066,0.5318610332563926,0.29187292948081955,
 0.8397546178940031,0.1314040605801775,0.847926591394464,0.07665408898104531,
 0.5386669517304024,0.22460050621846406,0.7415143572457527,0.6306468722057001,
 0.5821797501796272,0.3501960273593635,0.4625363113144032,0.9143552821713192,
 0.06933095891865326,0.38597151039102917,0.9157401322470916,0.941356686420693,
 0.5287618048539457,0.41597313381487355,0.2080429122527373,0.4390291640832511,
 0.9579573476962675,0.9476577320550443,0.6128746674364245,0.3139518639400627,
 0.4123346448411209,0.9497785260341859,0.6249472872419349,0.35639788747852874,
 0.7211811142412986,0.6916033089509467,0.7223014941841593,0.6571280650601355,
 0.4018076062294409,0.7937658157967955,0.04615617411515993,
 0.026744064270920576,0.24989072945211577,0.3869039075191407,
 0.34265685482122354,0.6780537674462567,0.5624208946302905,0.8308750459361652,
 0.38754620013436647,0.30698685181512153,0.8141937068708072,
 0.2545015345385141,0.32915867819835654,0.3791810045580587,0.2369048049173974,
 0.22380184627853705,0.5333398410304838,0.16175217544469134,
 0.16903999902190014,0.06113138905261495,0.7200871458964193,
 0.8092975539996389,0.1860986637589952,0.06469991277354148,
 0.25214595923693783,0.47169842240707316,0.026422659145457383,
 0.018455219051747518,0.7697550260703394,0.562065804055794,
 0.14402954894380415,0.8312322135733009,0.8886692302252818,0.7387112744300041,
 0.6516444664309025,0.40760412654222944,0.02213532157650111,
 0.8199072440407479,0.4603801093680947,0.8916022468646616,0.07463659883249163,
 0.8164182554946225,0.596256511805368,0.5932767268547714,0.047840912479821585,
 0.21606891675329676,0.6060515256435265,0.8875725408269393,0.2648871555080573,
 0.79070915780789,0.5462537824165162,0.8761268282596475,0.6131377154638834,
 0.43838762616301186,0.5975860860045259,0.583925577465066,0.798075816036702,
 0.0775388601047563,0.13660684451929617,0.18321022827021105,
 0.18856786237320833,0.9532362082621406,0.8118422701341815,
 0.16606609141743156,0.0404379921532374,0.10342441234944211,
 0.2843729456489158,0.07815881885859954,0.403689236098438,0.8294273606454342,
 0.40369360265517373,0.3412739135531957,0.028594501401895234,
 0.6466179644531095,0.9845070695095406,0.32543837723023517,0.9369112885666596,
 0.8393770630115281,0.2911154665375628,0.4910660425248209,0.5663522738992228,
 0.9388317608168448,0.8511543720852601,0.5594685672031832,0.18871226673113362,
 0.30826628743677165,0.658776238304596,0.34602088867560044,0.5227078484464577,
 0.9221883942841396,0.5840275651904823,0.15823280628106295,0.3126293249245652,
 0.1675932461700156,0.7508879247224998,0.43072457728278746,0.7941517411025614,
 0.6089277097074177,0.04208851217093801,0.6628401868143444,0.5416753356501729,
 0.15551574066845797,0.49064917463190794,0.3616478160983292,
 0.9828845940046557,0.5122478116058308,0.11677328892353689,
 0.013862560298513071,0.4814595411155863,0.05022681695070408,
 0.8033270436288324,0.599318761975325,0.22057978493232433,0.337104091496333,
 0.2369381774046042,0.4246228042209246,0.9700657107472774,0.5992713436483716,
 0.8488827321363664,0.9555613706823708,0.9744001161043085,0.9686280243906102,
 0.013663072916679564,0.8947304165145936,0.33022764806174243,
 0.5723432733052274,0.9897210948766055,0.6321019993659371,0.7446855641653912,
 0.5448703004001462,0.26104366625788566,0.7136319576741987,0.6442095307533612,
 0.4097242313237943,0.4018654353693376,0.8199687134872992,0.33152348264049936,
 0.5197498318285193,0.5807470100552444,0.7704117132778353,0.38803426415040376,
 0.2529005478668237,0.39391466424690325,0.8631428345510106,
 0.46872430370340723,0.5115512306690262,0.3908160540657144,0.5334923894166629,
 0.9850913415693844,0.962448886332852,0.7724334367216282,0.09410883349202892,
 0.7968021840064149,0.18459579007660298,0.6153111752042082,0.5952757898553677,
 0.160340891494398,0.33463070326396394,0.8726214952449878,0.9570823887142799,
 0.1053566995436882,0.04980072457380469,0.8690417325372071,
 0.01025111893713937,0.8647583758032129,0.49212509649759273,
 0.8667159050675733,0.95600324291405,0.8858804528974709,0.0686577334897357,
 0.3918843579892487,0.8450616764097476,0.5073605184927781,0.7762105549179867,
 0.7689021457322538,0.4150236164377782,0.9087471822517303,0.05052026682592203,
 0.47661420079653216,0.24007375020887134,0.13618585705809672,
 0.15815992187553685,0.9694503849479272,0.49938316462336296,
 0.5881324693824657,0.4863582889440178,0.4003613902993033,0.7294537052255015,
 0.5861289857426174,0.6396485829192952,0.8953175815013983,0.12457352539783328,
 0.12840748914176625,0.8433632981106226,0.5083140284931893,0.9881113804475053,
 0.4806892412323771,0.21625131290280242,0.9213268879831589,
 0.13516341062861437,0.589446951765437,0.7763086154274228,0.17212687411893146,
 0.14186830758437818,0.4309634259530788,0.5887398367137555,0.7021619134917781,
 0.8251473818263022,0.9278126994809993,0.9418801179379859,0.18981279873646906,
 0.5647681415621817,0.6557926659982567,0.5606500864937278,0.5292332127258899,
 0.8545037356448857,0.21619598673207774,0.27055887327225425,
 0.8121888048416454,0.7947138517476928,0.39507346030642165,0.3369957938649336,
 0.9063590536565151,0.10702503696383237,0.7739997320025869,0.3379199873532198,
 0.7895387732335679,0.18338509295066519,0.5199437567194428,0.5232222009589763,
 0.6077123914223911,0.17937720732604123,0.9134757568815102,0.9107815168490301,
 0.8318795050469584,0.5210884427363878,0.1044505726800049,0.06660827245915757,
 0.0423330255404194,0.7125139885610658,0.6537855494991248,0.5257383249815835,
 0.020360151940226667,0.6725967313030259,0.750630330249904,0.7650925139407815,
 0.6519864118668394,0.26704572152602823,0.6130566951483127,0.3605720566149223,
 0.11790579519229338,0.446798673759474,0.12860530940494008,0.0399870068820819,
 0.6455950264517212,0.8592139671373242,0.021414294327725125,
 0.4948249546820356,0.5994554802788646,0.18216508917882746,0.7405089184711395,
 0.40308239592893824,0.6443582814156351,0.6894019521643293,0.9638916359494057,
 0.8774790585434582,0.15468127521737884,0.30226874415654614,
 0.6745800200722372,0.8501464319453743,0.7122810331425561,0.7426029990238946,
 0.013283710664549031,0.31430799364178563,0.09548318983271087,
 0.030965515165048974,0.9475564912534054,0.15726731930832716,
 0.022081560019802438,0.19322390738596906,0.251875544914275,
 0.27342248554969817,0.3317180387327827,0.785959438747476,0.2823914206598592,
 0.32109540494531463,0.7580275016290186,0.5536726767706706,0.7055533025392355,
 0.1250765477655651,0.18361554694469873,0.8937091878824583,0.7625631469611837,
 0.5874691983533764,0.2500702649291293,0.9088190558316142,0.8719933343399976,
 0.5529704271002314,0.1924241816054194,0.8398041972319087,0.8766830144884878,
 0.6276939799757524,0.46665436230123236,0.0019254059360823694,
 0.8050033542927556,0.9398220607311789,0.30330590886877296,
 0.05774915562393068,0.30229383166295287,0.7559861066976277,
 0.07655862169809546,0.07568862429966061,0.894415752124253,0.6503899966857709,
 0.37592687929395474,0.29483285912247903,0.05021038134333944,
 0.6546532254929189,0.8538685576870297,0.6525632968113365,0.5609051323938228,
 0.49453457963882363,0.2522908636307276,0.2604906265722302,0.8064878648603827,
 0.30347247594109916,0.3519918016720084,0.237869441325521,0.4968258926999273,
 0.28813861619413694,0.237157673844694,0.8476212819755341,0.2626489201408234,
 0.44714906486582295,0.5767826591236672,0.36296516067708195,
 0.20872948803435165,0.6558215650897277,0.4336755760133004,0.4662826222140073,
 0.9121486028072058,0.24436817486435047,0.15088188456872687,0.093362376169143,
 0.8381495350850812,0.3425988125239663,0.2990533782519291,0.7117473203742024,
 0.872889047383129,0.6622238892910919,0.3036921085405444,0.920043465170003,
 0.4155159640549586,0.34602978874101264,0.7033759150871113,0.7887412542935449,
 0.31746672384399155,0.567564703890444,0.28689216458212374,0.5474950708306058,
 0.6375279811055928,0.3894467144331275,0.6574311630965861,0.2439015835782764,
 0.26998749863612315,0.15576285328090345,0.516636447831898,0.8725484239810036,
 0.5493198037371878,0.26204912115276935,0.1837335825722134,0.739994820809767,
 0.9305782413811947,0.5024360838440403,0.8113943137513748,0.3754699338519908,
 0.4541811772821265,0.14929203394643042,0.6655771374025968,0.973042302530217,
 0.08943349874564444,0.9833644055775798,0.998465664001778,0.3262430084624336,
 0.3548554802289644,0.6354947368515272,0.1598224670705296,0.3058374772344512,
 0.595018982687767,0.6421423522306181,0.46162491891891433,0.7087233829505922,
 0.9752117750410547,0.03460966251872222,0.32312711793379356,
 0.6107980694991602,0.36222090284023734,0.6616052997063577,0.650618402303421,
 0.9760306490215647,0.885834529533834,0.5363736885083611,0.6399169388076205,
 0.7902944552688886,0.06619324339241184,0.7036176894230619,0.5641411150727357,
 0.33951964010006974,0.520524195644063,0.3674849363107441,0.3570326542129659,
 0.8766249830108687,0.17206156800163597,0.23975028228165562,
 0.22663969429143516,0.586422007354451,0.8864346963266483,0.2351121869943904,
 0.8176648060496734,0.8731050727734578,0.008854228624228688,
 0.27271283729267504,0.5607339931885764,0.6705219087400103,0.4245095130303148,
 0.9403605808389018,0.8243390693087076,0.7992560944578679,0.6197349264887713,
 0.4467418256982848,0.6748083286249775,0.4134763781143549,0.6718015035576076,
 0.8975881000520313,0.9816199708079011,0.6188307257741835,0.04543140018599989,
 0.12416435240052925,0.7584440268812309,0.7020145864386885,0.6040758751683174,
 0.19629812426532645,0.5650195875340219,0.8502714996898953,0.165893448423574,
 0.5401302146417143,0.9361336296616691,0.2860575182240179,0.2904023359005481,
 0.38079324183381147,0.24075192176662696,0.9909805285699873,
 0.8497397249364504,0.24545529581530334,0.6459515230340808,0.5813948015786901,
 0.4585198146020122,0.22800636591815848,0.7919968232759337,
 0.11531318599856544,0.6712570540016732,0.22142021069230677,0.884213727968798,
 0.45011644405871254,0.8930093478663693,0.030175107909722665,
 0.07711290945829441,0.32466968538446417,0.12582684251352483,
 2.157414037817773e-4,0.790678558680978,0.7295760133947122,0.6453875924728454,
 0.5315014115103985,0.9978011186810424,0.08633945872188253,0.6324856176634966,
 0.014046784754079167,0.08424594761914261,0.733571046804961,
 0.37844044352248596,0.06554170217758504,0.9609404546288426,
 0.6806928324426189,0.7475324197697046,0.4113414055884079,0.6279254800138281,
 0.16750026416882124,0.47214677875237687,0.9270994656496729,
 0.23401141515901014,0.700437314737613,0.3505207689624219,0.6102056768387825,
 0.06806594498313234,0.966128076359785,0.04874729779042564,0.8733468878106935,
 0.0119250667424462,0.7961548303473677,0.6668371210964349,0.2694322262024208,
 0.7063724472102123,0.23210294862343162,0.32341293306312446,
 0.8244887605413447,0.6471777256830837,0.10113185406683245,0.7421325568280579,
 0.4874219460533499,0.26662692342582195,0.34569179856272836,
 0.09251053480799287,0.8274258115960147,0.8044810743813435,
 0.38152152131730876,0.05849666062122438,0.3179074348234858,
 0.9261352904287392,0.8455553881038711,0.6358816230484012,
 0.028453492922293933,0.4280386127493735,0.24577397120581723,
 0.5583311708663083,0.759500976598726,0.09171932036872554,0.22550474644911633,
 0.9619139348714955,0.9193226240683255,0.6448587084553611,0.7925634478430488,
 0.5995482769430126,0.8931722894941309,0.8713972878706957,0.0592160109262746,
 0.596323010469109,0.7591421648846018,0.0846885156435635,0.5817004429706585,
 0.9811652356844232,0.9442066859986785,0.5546071058291536,0.6429566428141895,
 0.05599789553019563,0.35653994011664747,0.845286805857062,0.5450903872949153,
 0.8561849816073883,0.5060616958014813,0.7622120691035859,0.7040848884615254,
 0.44267884821132586,0.45620578679130086,0.12532782071281412,
 0.815149507125969,0.9693357029250859,0.7491631938635631,0.8866987611147754,
 0.9131976159373396,0.9229264605679782,0.8899960068641415,0.19677970013396207,
 0.18102549926166245,0.8545574966744062,0.9449628857814147,0.8199276917441487,
 0.07927085656053678,0.47172011679063774,0.21769845799678667,
 0.2880368290286981,0.2833189563172285,0.5266430023056629,0.05347231471626346,
 0.6589519344351844,0.061590516530790174,0.6992652894785396,
 0.9480899074030089,0.712263080102149,0.6745832352946171,0.9919648366937758,
 0.7780491544377286,0.14581106862278748,0.725701833231355,0.6004630495745716,
 0.4164134143329328,0.9646470936588584,0.04322246997141943,
 0.05157883870555546,0.2216923553641289,0.20503315835721492,
 0.0043192898673150015,0.018377439956443364,0.47566786193751687,
 0.9011966160877806,0.04836493668088737,0.5679649783390331,0.8396618750917911,
 0.194693532143132,0.771256526916096,0.5306581835516663,0.3721995004713283,
 0.9579183379900594,0.12317639611070796,0.024174636389543314,
 0.39904883407983394,0.39543077854819486,0.4563061625865742,
 0.5879434834286279,0.058594798113907665,0.030996924350500388,
 0.011735669520082648,0.005495012785430742,0.2579599946678979,
 0.3964143698297934,0.33253694108563114,0.6748554830544438,0.3523077721237239,
 0.5504108698899883,0.6817250489049359,0.5765453272120685,0.7430720612392316,
 0.22423121253231038,0.39281994600845127,0.9080705502446819,
 0.054852532284594524,0.40448864223026204,0.6430543881164968,
 0.15903812562190245,0.8158366537825242,0.665490127636547,0.5129195949657676,
 0.4167313514036438,0.528015022052635,0.7983018023546232,0.6983522553037225,
 0.3336359196528056,0.39020788918926097,0.4350259516150804,
 0.19978189263383617,0.5284977435817051,0.095104784344053,
 0.010548491901060775,0.36749587813345663,0.49779574019372885,
 0.3804681444042315,0.4461885495039566,0.22854811216798598,0.7174022844156692,
 0.4706680274845172,0.35124040030163406,0.3845189830825928,0.7079504938311633,
 0.5080937766126412,0.4050423658077664,0.008403849235143923,
 0.6759650929899177,0.9269936975027516,0.02165552778537283,0.741889178104801,
 0.04897133201278747,0.11889000829104202,0.25705579493641073,
 0.45633663092390764,0.3030117060438904,0.5660884301627608,0.8889645991984997,
 0.8887051306459518,0.9545903328245743,0.7583011990588593,0.9683015906812155,
 0.11206969258749033,0.019445847960056883,0.7556008002087535,
 0.9275357036165863,0.22634591848216223,0.4018396545268379,0.4280514892561953,
 0.5750802248509366,0.8724405241299376,0.06888962765666151,0.3315276458263331,
 0.9674773613436385,0.8497928716682968,0.42574366734230984,0.5289856336010917,
 0.4475740471335673,0.659985278667615,0.27712273779875063,0.17638273104152902,
 0.8262181724502747,0.3548639908143363,0.22968231469184774,
 0.21585586123878775,0.6193685955503114,0.2275226760736847,0.7711610618577888,
 0.3067494131936317,0.7709688473829658,0.425812553952546,0.7122912698490478,
 0.20385898233755118,0.5040696557993465,0.8171971672167435,0.390456445452256,
 0.4836745702455363,0.6761904029312075,0.5845997161355224,0.4569516387595238,
 0.378274863510884,0.9855944391175826,0.23641801870683077,0.8164046061268141,
 0.21031865015782159,0.9164699132362031,0.8888340406825453,0.7057214154307908,
 0.10470548575030747,0.9672304845725135,0.10769572494144564,
 0.2970705034754486,0.6036443746558523,0.0226330197247544,0.7461143113799432,
 0.7717399518826991,0.338535793254653,0.11198898327207951,0.45291473360896894,
 0.21443726382424388,0.9506199483240892,0.33194274829960124,
 0.6962641763203599,0.024781424356268733,0.346599630190647,0.7123788422219935,
 0.7325168303685589,0.430723607403803,0.9470011098382918,0.9925237688524843,
 0.7250491816341177,0.3234178135895295,0.730429775210545,0.6665158572323157,
 0.5715851083334679,0.5292364937623258,0.19050722466606707,
 0.15192196172271577,0.6490461122693115,0.8045426581002183,0.7270320610487593,
 0.4351613879530035,0.027979295467304155,0.8163998976945608,
 0.14513144985237847,0.991788919928379,0.39552754084509434,
 0.021862763160489074,0.6641908759196301,0.9928913684209661,
 0.9789284976637415,0.132845683955079,0.5169471241275517,0.2142464084545046,
 0.9387106115327791,0.7592569530963205,0.42847727942163505,0.650276432702326,
 0.7873382590960569,0.3379421078507986,0.9881480233410349,0.4058451736382136,
 0.06050686984049347,0.04419052213059893,0.5936081922639548,
 0.8874091767313731,0.46398548454862565,0.5108094474493149,0.9308555304915581,
 0.19224389259624663,0.6885454234719182,0.7421637507048173,
 0.33073072113742885,0.6326620917390983,0.8406895654697409,0.3557947244339039,
 0.7957380103588694,0.3267860446362586,0.3567206329455125,0.9291494655384885,
 0.42872267779156303,0.3474848123232763,0.6938080814814104,0.7164027137361225,
 0.9913180515898435,0.6347767502341806,0.36639952500119133,
 0.49592191106165096,0.5804577811791802,0.6617254561517291,
 0.037948763832924226,0.08773307389913887,0.2674910503298292,
 0.2597553593936157,0.6054239296207413,0.2509191466651727,
 0.0066916833977530565,0.9488858125988302,0.27731237436293754,
 0.7773214711822013,0.15937812851376998,0.496746601872989,0.1333111987292096,
 0.3150399822749125,0.38135784583046517,0.20533039176896783,
 0.5012395527144594,0.2570075390922314,0.3282507462032799,0.606245618950206,
 0.8669154221799052,0.76716855532481,0.1474448047049407,0.37139765700997296,
 0.5101288277565991,0.5264276863570263,0.2227077250026217,0.6910493029266156,
 0.8537640443778585,0.34839668755589626,0.6317560180058238,0.7229822995006938,
 0.5040940537418576,0.26395150044876214,0.6184367967827114,0.8755855446006977,
 0.012485457201626304,0.9463254831462232,0.6257173740436721,
 0.7709322939516897,0.341483932264927,0.01968268446409205,
 0.005058194251664316,0.6714131985913835,0.2280705202177249,
 0.4929346959342279,0.3103710675990137,0.9696007215915677,0.7126792888709964,
 0.5358120955968885,0.8626872730822991,0.7401114416937067,0.3226938539477411,
 0.20001292419777794,0.27335285074922344,0.5060856708666631,
 0.3270444952766749,0.12725533868110206,0.3542997946544163,0.77431014165234,
 0.1979276328750571,0.7503043596227112,0.8414701575029295,0.3119447283642911,
 0.5881153614359549,0.17320106622852505,0.21049769880733782,
 0.08136419560820085,0.46403085067125227,0.508954920172495,
 0.04289270095268927,0.942312197878915,0.6481634277870449,0.07390236985277909,
 0.7543265161528395,0.4090857667252842,0.16689509231665156,
 0.34851912312795985,0.2517299113460797,0.9411912497156973,0.3787035272842487,
 0.6763887466551445,0.46851007166328695,0.6395075495151639,0.459844538342703,
 0.7783336459101189,0.8319563638074574,0.020379780596018304,
 0.03517107537829545,0.3602116282636266,0.48517654891274975,
 0.11906693872877838,0.6912031086967189,0.7094452930742909,0.5084361483455716,
 0.9054674802995676,0.46644993106924204,0.07760741747965538,
 0.14338259833509104,0.5520312279472452,0.769351419834668,0.1883537288917998,
 0.9748917250498655,0.030088973552332976,0.6318358511985134,
 0.44259157651224357,0.32008247760561925,0.5027598441218956,
 0.3138252097842781,0.6267090281701747,0.4438529928031554,0.06802409921426289,
 0.34684493781891346,0.8132363185939321,0.8848022621168647,
 0.16729223531192106,0.4241542473917751,0.7994133135360952,
 0.024498300460303324,0.10588899903206606,0.18594193461548125,0.377258531489,
 0.41177629352319267,0.412995145854931,0.6238597390988273,0.13134603810990475,
 0.204785313712184,0.527980844540212,0.4835285260474115,0.6987340671819183,
 0.7103755663041312,0.041811454342069565,0.19632664582845238,
 0.34493789589850965,0.7445763476234883,0.7984625691566445,0.6540054724869573,
 0.3697240110752045,0.7418777431101121,0.23370430472788972,
 0.10598502044974412,0.9678769198874799,0.07222854438025961,
 0.8106588303006703,0.9658734852570177,0.896177171029074,0.5699473377805635,
 0.8604296810350665,0.9336731634127602,0.04449217946856976,0.1400245074312887,
 0.5863397807694041,0.3066679855453245,0.0645281189855178,0.9891913581535646,
 0.43892649575779197,0.6839326159980013,0.827017183253909,0.5824822729500119,
 0.0837194507987109,0.6905090236381989,0.14444301811221472,0.3505233446013324,
 0.029695491970877508,0.9715031592152527,0.2309855817733384,
 0.7863103881949823,0.904718619673643,0.2314955538173351,0.7569652291417567,
 0.30932686126085973,0.9526930153753433,0.05426420715307956,
 0.5763750133149792,0.4104145935733089,0.8267196211745701,0.9729420784230702,
 0.7786206937003607,0.7106119472708826,0.7202010645867201,0.5040421268262165,
 0.8490532158854535,0.8830352410495949,0.7714578181880956,0.07909157805254208,
 0.45554834462747096,0.3411851336407289,0.48869918953484115,
 0.9773771069489811,0.9873804385580135,0.2138786281495203,0.8998635996382357,
 0.24257355278628157,0.20208667732652674,0.3072319626458242,
 0.48019348044775834,0.2449660134830215,0.9284258555734936,0.4650885381524974,
 0.4567849376559475,0.4256943187429244,0.9326644490885737,0.5975419720084284,
 0.7905488119646452,0.23624038959504712,0.07479640840101653,0.827867783446069,
 0.5329919787608507,0.9605335771458212,0.5060695051327188,0.18616664711070363,
 0.39896488852632483,0.003628331117594863,0.2699708846369455,
 0.6421617598461089,0.2811513718878943,0.18450291965236199,0.6034001944107802,
 0.011272820193747868,0.7021279757697708,0.5488183144609353,
 0.8995737511704768,0.8304674505291771,0.6537688859480608,0.7998317510226147,
 0.5652106246136861,0.5273443562425769,0.8767414199724566,0.09719536489374447,
 0.22365506857258177,0.19185118323765515,0.0035498131761172447,
 0.6000994999529496,0.10325920159359225,0.8390469987619088,
 0.06583356802979945,0.23950106088433354,0.7654484995248898,
 0.6816405215551742,0.4296378641527199,0.6011254898338733,0.3009730097234583,
 0.7105199864651784,0.4379978027354198,0.8808806958049089,0.1420140692023708,
 0.7553642551435258,0.4759667985039957,0.3216088658455486,
 0.0038714660572630732,0.382935483075701,0.7046042155900408,
 0.3277357533921861,0.9646304427774112,0.7007330236459508,0.5109871969527088,
 0.440937607451354,0.29785280970475314,0.03674534722901346,0.5042846012797535,
 0.3437392578893851,0.1392641553562255,0.07805704821950088,
 0.31971536756740315,0.322929906740946,0.900440323122535,0.731310913191711,
 0.5444703543179349,0.7980553945050524,0.18352588303800155,
 0.11234596763197635,0.035074607527943,0.09329759684623828,0.3429202255631787,
 0.09206194634780651,0.5924694606480354,0.2755642958437974,0.2893333748248905,
 0.5175671564018247,0.4958645716851885,0.07695651384098767,
 0.13234454003993834,0.49258856466681544,0.6481562825051254,
 0.8890275341407421,0.6011039869828471,0.3766804246955896,0.5423270408405396,
 0.9928186245577192,0.30383420483707435,0.5332041360057793,0.6788545229310743,
 0.10580645955173296,0.9976066040978523,0.8995626740930924,0.43301358979992,
 0.42255658108065397,0.9897015827906005,0.7817307826410441,
 0.14240469046549775,0.9023474128879303,0.8319663931684397,0.6582465738735015,
 0.10807832794413008,0.7893649005234948,0.39142794871161857,
 0.7651562939742471,0.9658651004946507,0.33485809429466074,0.9722505715888105,
 0.5152661083049294,0.902325104388689,0.5743562993332564,0.5136829193818755,
 0.5569245520720657,0.2481399542276006,0.03721389867880631,0.7101993056488439,
 0.4279008218408371,0.2079837805613467,0.16718245545606003,
 0.06511743054487162,0.14630767948602608,0.6852917738627753,
 0.6989223986337358,0.12520994255298945,0.3122313043258491,0.5771994735156045,
 0.6884259086780143,0.9809584770230706,0.43438660108097216,0.6548633387015944,
 0.28711924550956947,0.9354844446273541,0.9071458797982574,
 0.21891578969388947,0.21578467153087488,0.6642478840728026,
 0.17292870201778976,0.920428353970569,0.549006974974637,0.9265936753400081,
 0.7353829059786426,0.21786910166051332,0.3820433785705545,0.746024885511029,
 0.01455372470064098,0.07046395795917548,0.3883212690754023,
 0.5253399314787782,0.5126464995323309,0.5937296097626816,0.27683269901041796,
 0.988921938286736,0.17849133132626038,0.5938135668242388,0.38178963127424725,
 0.912093568560982,0.3763181890637731,0.12691245121456052,0.38499555759524695,
 0.6999622387807888,0.5292996305640784,0.06435113522777758,
 0.14218453698291167,0.24371026832148102,0.2804559296962056,
 0.3970505220119004,0.5433089512546342,0.025646517760224974,
 0.07856616403697114,0.8757710084875991,0.5193353005224615,
 0.20431552167107736,0.5317439740478349,0.7252252372221732,
 0.33589312455610765,0.5130859660741777,0.8312260306011241,0.2650377350968258,
 0.4183484717348589,0.9465991892108602,0.7485069732758544,
 0.037116008162815795,0.49467453190369337,0.40803138237426373,
 0.14323462241706386,0.9874930926636021,0.008730506137666239,
 0.5765988574449373,0.34378482797226906,0.8021514352202934,0.6788369341859778,
 0.5954381700017972,0.20002010805327797,0.4195929434405521,0.1491918290729366,
 0.3712124463243405,0.324834981988261,0.08286955911370897,0.9935834058887258,
 0.7888383536790698,0.9827402150598679,0.10148724801195819,
 0.02339339247750727,0.8344247053397473,0.24774479892710077,
 0.8415361110629177,0.22752635541622046,0.19376608548113172,
 0.6780344442412235,0.9792562428267645,0.13174728707531735,0.7382739240258984,
 0.06364697145443954,0.20014756518266952,0.8848871213261503,
 0.4341023343009609,0.023572574177373906,0.7350728027632427,
 0.19194500790945201,0.21704064935468437,0.8248157867342771,
 0.9787776415227936,0.1535138652431216,0.8193931288978907,0.30292829084135975,
 0.6173025634936258,0.40776874656142215,0.5548837445761178,
 0.22945087879691128,0.9160717053824723,0.7913137349752419,
 0.20261860996950243,0.10946345318685013,0.0965813638983739,0.831576972353881,
 0.6757856272040611,0.8413651517981753,0.25930104082495675,0.5160452805150416,
 0.4746866051217703,0.4621370917439934,0.054389012468863296,0.905991693232775,
 0.5373357039321649,0.8842964222639106,0.46860924937738024,0.2766302089063124,
 0.3568988896020936,0.12823994928212068,0.08071361337659955,
 0.021176993299718117,0.02906585277632423,0.08352240429806768,
 0.33824130783829176,0.15363499958022997,0.5436606133008826,
 0.7515622054578485,0.45490992402684816,0.3373453726173221,0.9397567786010984,
 0.08820017284319004,0.8287708503042708,0.43215209500738316,
 0.5893202947366214,0.46641348215820366,0.09563521407100528,
 0.4701348709448604,0.1393310541831858,0.21020088853537355,
 0.10269732119002462,0.7753733619202602,0.9477994955849603,0.9637556690670345,
 0.0784712138582645,0.8013036362751004,0.9084717663752019,0.1935095940492243,
 0.8839272946431539,0.660431284926228,0.6451549798489334,0.6918189805266988,
 0.8511009567345409,0.23372553992239298,0.5997480158400323,
 0.08117052474682362,0.14361148303717142,0.03811301171778958,
 0.281337929451247,0.23678559542645394,0.13093297912316848,0.6026154978700079,
 0.125070233792413,0.9714150114202865,0.21635089211513592,0.44020941781682965,
 0.05510863115442344,0.646874362263124,0.425353129103091,0.9332480630274325,
 0.8495119757147648,0.9933068310498606,0.4451154801402625,0.5008406382176027,
 0.8649774672122325,0.42859776106307,0.422120975168773,0.8857609949696326,
 0.608237070080917,0.5018540091623384,0.6056115695362425,0.1439707315759018,
 0.4069484450632709,0.8612847207760979,0.2809828240698977,0.03271486255870637,
 0.9524820726548096,0.7726320777191029,0.6858787754144804,0.5587462691882967,
 0.4328706614278257,0.8690657443171058,0.6192742921565042,0.751637932520317,
 0.18390359410864288,0.7876090938654187,0.16879726722853583,
 0.5737063678287292,0.9963826830739615,0.83076419808701,0.458052842186758,
 0.3259783584765903,0.405135403724192,0.24828884513069804,0.6482585844508842,
 0.8312072013131744,0.8989843014238229,0.721214614351348,0.15700750621955406,
 0.3601148937650879,0.7998563376734091,0.8416446364093939,0.39647214655754537,
 0.8170658662076147,0.004919724060755359,0.9976282543830809,
 0.3603932253272025,0.6846385275755592,0.5871496713223188,0.7749535940349731,
 0.8080792967654887,0.7687323090395898,0.19484004402286037,0.5259088252926644,
 0.20888743135224508,0.934708660060758,0.8837506195851499,0.27285042563648465,
 0.31738276996864057,0.6594377730724024,0.28591611107351944,
 0.7799713744944684,0.9058674384344141,0.8925261448087808,0.5377763730230508,
 0.6829906478207669,0.23618007607499747,0.7294827374080074,0.8040025888217096,
 0.8335946744093871,0.4843556868153718,0.7346089897227029,0.6316663831960516,
 0.02332212003340861,0.1261114231702467,0.30070653499668953,
 0.18792737171520302,0.34990166213525187,0.5453389196858214,
 0.49100804920063656,0.764345924969331,0.539410954817204,0.18680977493872764,
 0.47263365583301997,0.254182469090924,0.05705271848001714,0.4218556083644138,
 0.33847442560150265,0.4513814327050186,0.8464871039455962,0.5059232580251616,
 0.41141383387193,0.1427515627712982,0.5730412852593859,0.0029015277707513665,
 0.5682537429468069,0.06264013009312586,0.8990844346308506,0.831081204111221,
 0.42990404372045743,0.5717516886101714,0.8226663650004615,0.4873007142240807,
 0.05489351003825094,0.7853792806529091,0.7019092761168083,0.6864799766557141,
 0.1309477361526824,0.07531240422237473,0.6088457318081185,0.548615779069954,
 0.6067967079934753,0.5963923196592142,0.31664375001751344,0.7496084055503988,
 0.4313504128375132,0.6244214502709852,0.36143459920360177,0.7856770324385572,
 0.8546942077621424,0.6692755921101877,0.034060490677595756,
 0.24695049246123302,0.9579661495614129,0.4653074875142581,0.3118481840149141,
 0.6253524848098557,0.10325037060445619,0.10429899453473324,
 0.6486320881583312,0.16498002852550597,0.23148598379571028,
 0.04494535145112122,0.1357264591177224,0.5584340836251713,0.186979884955929,
 0.4054447306643415,0.058579504396032656,0.13096614122484174,
 0.7723726614918027,0.9490307908339644,0.1960533515127646,0.4795826169741817,
 0.2765528354475686,0.9127223782606061,0.9526639520769947,0.7085367662192735,
 0.20950881436971214,0.14985191178126256,0.00737264995353204,
 0.467730548239205,0.9416144711595491,0.4059058906216235,
 0.0035889720909346723,0.06117096216270401,0.07386192968733174,
 0.8157285073272392,0.8153836817069418,0.66423781501996,0.2299189414795495,
 0.3113920098312001,0.07304441431594688,0.6363340881546172,0.7384882774497592,
 0.536115305090842,0.7345283004897186,0.470193767799129,0.5669718525850922,
 0.8288098842025425,0.864070869592717,0.17732673924790043,0.8689627828944944,
 0.6834372007706555,0.6509674131089017,0.16141596376466383,0.958252232282242,
 0.5313944005258293,0.565115148735843,0.07934621888009596,0.4068382320385384,
 0.0668723078533855,0.35859027509318997,0.5422982210912162,0.7102130850226934,
 8.025004862729501e-4,0.7123060255523929,0.594974569097173,0.74839622655138,
 0.995750856130574,0.06737313994962801,0.365673453831204,0.4700193488373309,
 0.4670615490227842,0.9908733307931082,0.46962139041562123,0.5422390130822898,
 0.9329513512397727,0.18712283399097607,0.19288444962637608,
 0.39389452842377015,0.863914700762816,0.566761144187373,0.39291811931099907,
 0.3653037543265072,0.7386695089998562,0.25032606024856696,0.6872807382349645,
 0.06908459969256187,0.5366107866555722,0.38123869329439675,
 0.09356013204846658,0.9998612818702456,0.05863409123593899,
 0.8324936410736904,0.42138746909624325,0.5425316074927915,
 0.03511395143497098,0.8290566302103171,0.6859360442702856,0.7026720336226179,
 0.6076014261930771,0.2298431650707955,0.8111920880730492,0.3808450316941222,
 0.4624313583921077,0.067977401565532,0.5249922723741897,0.7389781947263919,
 0.3368941709279456,0.1742814943694515,0.3932984372616596,0.8485046129700039,
 0.6122893791540162,0.5588664603264009,0.4522844274546364,0.7810361816223808,
 0.34570157355639786,0.316176578460999,0.41841679934272136,0.6477925867827087,
 0.4707793632573831,0.9923260395837746,0.3538090258771557,0.07825704615596896,
 0.16463994261323434,0.57700539134703,0.8862731627566827,0.6607189875918293,
 0.6978212286756402,0.2014185773532542,0.310648792524328,0.6611336648482738,
 0.6671267471263335,0.5329371326273514,0.6433024753694334,0.29536203020546214,
 0.547214132213021,0.6630252087927726,0.8692440498287981,0.5437516950986733,
 0.605681342607339,0.7258323235114538,0.3398025026235246,0.6049152391737069,
 0.6018832437904762,0.7052444600557869,0.116597079687904,0.42995603990714915,
 0.7721851115037838,0.04252356759518716,0.5954635765016484,0.5738046404200119,
 0.7990854583737431,0.6324588726673827,0.7522925617658811,
 0.051380755590412996,0.5022429563377414,0.6420113538340001,
 0.9724184810839627,0.04164287304542291,0.8564456683172906,0.5191236135259922,
 0.7132944493664269,0.6025975807594133,0.8654423711924534,0.21052927964946777,
 0.34761157337935034,0.6536559217728531,0.46000470443591723,
 0.5105450600719228,0.9877416055805757,0.40052102635282916,
 0.11088377752362788,0.9602599067721619,0.2653509061201722,0.8907099658432102,
 0.6175156504998656,0.6492248233457861,0.38110450395687057,
 0.42346617398744235,0.8254444851584847,0.9514143232326042,
 0.40905721698245545,0.16704450018485617,0.9612259280746983,0.061903079804144,
 0.03023171316965556,0.6800089012923214,0.15993359619082892,
 0.6752416644990953,0.0021528405817551333,0.1459691504929912,
 0.11069927270919844,0.6265800367876326,0.397142980151317,0.18565567699731977,
 0.031177344787277406,0.9438444524811141,0.26551518780425976,
 0.012948039942676015,0.37758842490543143,0.9437672104314702,
 0.3608901205282843,0.7134902636640604,0.5690061953990662,0.7856066113121858,
 0.535192460448194,0.9742652291549024,0.6109495950510815,0.8005786256429359,
 0.05631279533808897,0.9194861715296719,0.6485442869212512,0.9280599407275733,
 0.141617844491575,0.8840594934866152,0.19531989132202376,0.9623116189143569,
 0.6630194228416522,0.16666715647274144,0.852875625239251,0.5257582861779818,
 0.08671074555397196,0.6446902528635873,0.5991432549776237,0.9259895813566706,
 0.032519035120665674,0.8769359115276695,0.4839750575057957,
 0.3837585685581394,0.9151391508243663,0.339732461368738,0.8885407438705945,
 0.2501373454510698,0.30383231008574074,0.9257778258421699,0.1676140988831929,
 0.8992136306196866,0.4428291263079529,0.6252805936691281,0.660370620502118,
 0.7532406896883713,0.7856800572475168,0.15136205097822764,0.6408576823502886,
 0.6086402487156798,0.5112136696382669,0.3315532356036792,0.8263639092025924,
 0.026413900674083335,0.3806083865673089,0.9653975090133649,0.194531292869089,
 0.1638539310725331,0.7200456567948613,0.8835135514401428,0.21064688838319856,
 0.8202267981964094,0.23227485887821064,0.8892573120872782,0.9291719223274468,
 0.7092964086040016,0.4542857441813578,0.13668781694754417,0.5587707484048807,
 0.7415419315036644,0.20968498800561475,0.32329514435046414,
 0.5854281720504017,0.7669243149611487,0.6233201431977362,0.5694188079167115,
 0.46455470963642687,0.8726117774283874,0.06766936647930866,
 0.47049543831061913,0.8376903055623074,0.2631626822970259,0.6971618209013319,
 0.36037449565790614,0.3935727183120352,0.3790723343259649,0.9876958614822164,
 0.953594144261993,0.6535550632015523,0.8107928693042495,0.7406051482550572,
 0.6570154378790969,0.1129551488874595,0.7823267563754204,0.8169910876542882,
 0.41329659253763373,0.7275651402696136,0.1342406619690928,0.8971132287579006,
 0.1267090382251247,0.16662577660115796,0.7630296200849579,0.9367512435255406,
 0.9443659323922452,0.13537296830218393,0.5501679893166285,0.3259173571175985,
 0.1192655679888951,0.18457078212983435,0.9812548631476692,0.5983134828734534,
 0.47413789981576626,0.34429812358870837,0.7460843495661519,
 0.3118939551511798,0.4171003389899214,0.0771921512407222,0.27157370314159057,
 0.3236061516920389,0.6555781992827104,0.4887855900981163,0.4967195359500114,
 0.398685348289223,0.5358751845181662,0.36276518227720855,0.42142659159576634,
 0.15820658568719836,0.8055124706751629,0.5728852400418207,0.7543311003924966,
 0.7302926578978216,0.33150283276841175,0.2100946079215784,0.6523251696449726,
 0.30130409507075107,0.3575525101197675,0.8587800542767519,0.302196606313204,
 0.34885288309834817,0.8765547685018048,0.023291732511069196,
 0.7324169260134493,0.11441048637660522,0.8886025144804459,0.4612812469362646,
 0.9285143959245611,0.5561833215900314,0.09674812030900737,
 0.06094193675432902,0.23733337797200305,0.8698488468243081,
 0.6434981760679077,0.37726905395262134,0.6076841018950017,0.6822786037305822,
 0.15699642457017016,0.8315686194741099,0.18389908685221668,
 0.5630273036961458,0.16884409262779743,0.7557398444982109,
 0.31623507651835425,0.8548634593201947,0.8725215454417924,0.6355089465505924,
 0.9856173008875503,0.8187585892000989,0.8489361066574965,0.05114671571505969,
 0.014020619406254764,0.22711871893347912,0.37101029487773274,
 0.7664726066774201,0.4613622622522934,0.9631337487169468,0.2956196058921221,
 0.6367092849369391,0.11666988648015741,0.36328148861506926,
 0.6676816207531318,0.6345504235305107,0.08471533538954645,
 0.03878152437256377,0.191201752210857,0.6970767994237617,0.23080620402471652,
 0.5663961716610734,0.9725211167587784,0.16794282626231816,0.474260294078079,
 0.21889745826659357,0.4390225366725581,0.46832355294136185,0.645677462343249,
 0.2292852308389587,0.8392531685579795,0.09814302022247268,0.8648331101896066,
 0.18954221586772468,0.512466997461905,0.3496625163640944,0.7313395241577444,
 0.4810704105354101,0.5751952796567813,0.6071560439924354,0.2842345857440933,
 0.9203036801566197,0.16842792320887945,0.8178358007680313,0.3665119419527856,
 0.03199722497081103,0.37427982988766273,0.34381607537532566,
 0.32234457608174627,0.016339442800363146,0.6518916855505312,
 0.5531225870193167,0.11995308975337504,0.15891198890460922,0.72298400653335,
 0.12119532863805915,0.7795995949658258,0.11969022845448807,0.875124702557595,
 0.5666756570239385,0.6982425255296982,0.17216658505164528,0.9937009228028576,
 0.2002698079045716,0.5805986141931815,0.4083726524940494,0.47409616721068604,
 0.5815954124959114,0.5257562481626312,0.21333874749366744,0.2789651952565405,
 0.5176890442804423,0.78819356201278,0.7054396952146128,0.8191426007443394,
 0.4201438445309963,0.7155412190667152,0.19707983119619144,0.741774282412895,
 0.801523278997854,0.6337177018631542,0.17344323926557248,0.7540468078056735,
 0.41628433784380003,0.48633502614150137,0.27710469031382057,
 0.6512532126601196,0.041897920373063724,0.2448270430463737,
 0.09262282396000732,0.8301638730839629,0.650284540699512,0.7850814719805393,
 0.8498153609809528,0.5511078362208387,0.45553281438145443,0.4102509972921238,
 0.5509160346174093,0.273625086693454,0.6257496313679506,0.7672661468883947,
 0.3305978698656594,0.7013063551586934,0.27539533668658067,0.6046554604693555,
 0.17475151862905947,0.028169072076207335,0.7250258522389522,
 0.814418554031468,0.14365360405634753,0.8531857887082265,0.9499035705920432,
 0.6697431650479788,0.8707285984270866,0.9960309096254372,0.21611852207367288,
 0.3592938337049282,0.28137929905262604,0.5066750817115162,
 0.43084619173520455,0.6208207030513226,0.7351501757298567,0.8924529485059881,
 0.1956496148608241,0.021927986358661222,0.5013529801162974,
 0.15122849897244028,0.18377857852065427,0.6403961058440857,
 0.7246695438842535,0.1406467540055104,0.14032424835609025,
 0.22748324819949606,0.3562710093044079,0.7921015092903021,0.2453891291954735,
 0.43503891549627305,0.82700783437086,0.6505023828200689,0.2380140225237266,
 0.8411360922537189,0.45400685638544946,0.7891002627086169,
 0.48501816052648916,0.3567175699203635,0.16005576646350628,
 0.07536619081041374,0.9371011972693092,0.030687324201559063,
 0.608331270443565,0.6263801581684503,0.26565314582737287,0.9183605084022863,
 0.5516213042592907,0.21607044038522016,0.004351052015947321,
 0.4942530296517219,0.6291986280191181,0.8747903026498051,0.06437083555795098,
 0.4224251844805752,0.9421339319644937,0.0739238814778489,0.5362601148565197,
 0.49649343154566195,0.7331423487838442,0.40420933208360876,
 0.9143042232449178,0.2420253456785283,0.830471330369361,0.7863949639184599,
 0.938315902277941,0.591705716613453,0.10567860596011347,0.6574364872534408,
 0.6060326167232009,0.6272009827745748,0.8398352831016259,0.8975949399599443,
 0.07199853285751301,0.907521157748974,0.04978915308765142,0.7029882522959138,
 0.7010380270205944,0.07326955968866211,0.7452653147742223,0.7938317562195795,
 0.6159078209273469,0.7063300912272955,0.5575108795469297,0.6125539452622728,
 0.10069126709151377,0.5792322514793903,0.0010269508829672258,
 0.684964705163053,0.285858640119272,0.779264914387792,0.5433449487100471,
 0.46517072789787894,0.14319733663492618,0.17329235991610925,
 0.8695990278957947,0.506121892629107,0.043445095742526196,0.5525422058823375,
 0.934422154283633,0.6864352886743669,0.782841886945221,0.47356772968117056,
 0.41289272615345507,0.7168186288388971,0.7907073760433928,0.9107033253984337,
 0.2854889471405667,0.710116605640233,0.31235357650208073,0.5048279381513165,
 0.12207533497701939,0.6688654308472997,0.7510337254142524,0.5966486600270562,
 0.610578190930547,0.7062342190702006,0.36704871862412825,0.9856174237669851,
 0.8685783131662906,0.7451672478127933,0.14555052854902906,
 0.06941867496947207,0.5641686834955645,0.47831274900370013,
 0.3605226324603724,0.9463359872143481,0.523231201851424,0.36724861713295864,
 0.7117620264228883,0.11193361506591404,0.9493392477952128,
 0.007984976413439981,0.28270500112722274,0.12192083520109465,
 0.8138149059193975,0.784877840538625,0.8117530690403708,0.18692127949524784,
 0.5307296396605508,0.7223067357603464,0.2238749079607677,0.23498717908895417,
 0.2920980299546705,0.03720990548340175,0.4667551746257447,0.2690362397672831,
 0.8843100722992655,0.9371159722795528,0.28678771055071584,
 0.004702128496402058,0.7136609623536693,0.3294660694562034,
 0.5617944859057605,0.3216965012834704,0.11063369330195727,0.6948011938415315,
 0.24620696634885741,0.6717724001080316,0.45956202255555845,
 0.9195019957919912,0.29949429337042044,0.645695035597217,0.3578712023860683,
 0.07964289522132739,0.8800724596781567,0.9051421646249792,0.7146675022925577,
 0.4450332721998356,0.6378785151143118,0.524463340256904,0.06096693538812736,
 0.9804793744966422,0.1597926940567579,0.5545478633727454,0.6226194376663384,
 0.7427347563463358,0.6287337038276236,0.14618505076674726,0.8487778036064842,
 0.7393189305531168,0.03384343608814422,0.888056537057706,0.5703901566796784,
 0.3362906674633337,0.2593042203093069,0.5044066736263861,0.8350449189093984,
 0.061727549052664865,0.21101149865510482,0.10170573066007738,
 0.8836381893614662,0.26662981988917256,0.6081964862257159,
 0.059138035555632684,0.6680978116580759,0.42114572935754124,
 0.3686091664879384,0.6016074865504191,0.9221404144761567,0.6575440311922187,
 0.03963151998105385,0.5753451392396954,0.5852318676826616,
 0.39998303185334383]).

uncache(Result) ->
    element(1, Result).

uncache_skewness(State) ->
    uncache(skewness(State)).

uncache_kurtosis(State) ->
    uncache(kurtosis(State)).

uncache_stddev(State) ->
    uncache(standard_deviation(State)).

normal_distribution_value([], _, _) ->
    [];
normal_distribution_value([Uniform1, Uniform2 | L], Mean, StandardDeviation) ->
    % use Box-Muller transformation to generate Gaussian noise
    %
    % George Edward Pelham Box, Mervin Edgar Muller.
    % A Note on the Generation of Random Normal Deviates.
    % The Annals of Mathematical Statistics,
    % vol. 29, no. 2, pp. 610–611, 1958.
    K = StandardDeviation * math:sqrt(-2.0 * math:log(Uniform1)),
    Radians = 2 * math:pi() * Uniform2,
    Normal1 = Mean + K * math:cos(Radians),
    Normal2 = Mean + K * math:sin(Radians),
    [Normal1, Normal2 | normal_distribution_value(L, Mean, StandardDeviation)].

normal_distribution() ->
    Mean = 0,
    StandardDeviation = 1.0,
    normal_distribution_value(?RANDOM_VALUE_LIST, Mean, StandardDeviation).

t_random_value() ->
    L = ?RANDOM_VALUE_LIST,
    % standard deviation of a uniform distribution [a .. b]
    %   with a = 0.0 and b = 1.0 : ((b - a)^2 / 12)^0.5 = 0.288675...
    % skewness of a uniform distribution is 0.0
    % kurtosis of a discrete uniform distribution is -1.2
    StateX6 = add_from_list(L, new()),
    StateM6_1 = add_from_list(lists:sublist(L, 500000), new()),
    StateM6_2 = add_from_list(lists:sublist(L, 500001, 500000), new()),
    StateM6A = merge(StateM6_1, StateM6_2),
    StateM6B = merge(StateM6_2, StateM6_1),
    1000000 = count(StateX6),
    "9.99999e-01" = erlang:float_to_list(maximum(StateX6), [{scientific, 5}]),
    "1.53123e-06" = erlang:float_to_list(minimum(StateX6), [{scientific, 5}]),
    true = (count(StateX6) =:= count(StateM6A)) andalso
           (count(StateM6A) =:= count(StateM6B)),
    true = (minimum(StateX6) =:= minimum(StateM6A)) andalso
           (minimum(StateM6A) =:= minimum(StateM6B)),
    true = (maximum(StateX6) =:= maximum(StateM6A)) andalso
           (maximum(StateM6A) =:= maximum(StateM6B)),
    "5.000021791129e-01" = erlang:float_to_list(mean(StateX6),
                                                [{scientific, 12}]),
    "5.000021791129e-01" = erlang:float_to_list(mean(StateM6A),
                                                [{scientific, 12}]),
    "5.000021791129e-01" = erlang:float_to_list(mean(StateM6B),
                                                [{scientific, 12}]),
    "2.886061744484e-01" = erlang:float_to_list(uncache_stddev(StateX6),
                                                [{scientific, 12}]),
    "2.886061744484e-01" = erlang:float_to_list(uncache_stddev(StateM6A),
                                                [{scientific, 12}]),
    "2.886061744484e-01" = erlang:float_to_list(uncache_stddev(StateM6B),
                                                [{scientific, 12}]),
    "-6.7071408756e-04" = erlang:float_to_list(uncache_skewness(StateX6),
                                               [{scientific, 10}]),
    "-6.7071408756e-04" = erlang:float_to_list(uncache_skewness(StateM6A),
                                               [{scientific, 10}]),
    "-6.7071408756e-04" = erlang:float_to_list(uncache_skewness(StateM6B),
                                               [{scientific, 10}]),
    "-1.198695443626e+00" = erlang:float_to_list(uncache_kurtosis(StateX6),
                                                 [{scientific, 12}]),
    "-1.198695443626e+00" = erlang:float_to_list(uncache_kurtosis(StateM6A),
                                                 [{scientific, 12}]),
    "-1.198695443626e+00" = erlang:float_to_list(uncache_kurtosis(StateM6B),
                                                 [{scientific, 12}]),
    StateX5 = add_from_list(lists:sublist(L, 100000), new()),
    StateM5_1 = add_from_list(lists:sublist(L, 50000), new()),
    StateM5_2 = add_from_list(lists:sublist(L, 50001, 50000), new()),
    StateM5A = merge(StateM5_1, StateM5_2),
    StateM5B = merge(StateM5_2, StateM5_1),
    "4.992619967244e-01" = erlang:float_to_list(mean(StateX5),
                                                [{scientific, 12}]),
    "4.992619967244e-01" = erlang:float_to_list(mean(StateM5A),
                                                [{scientific, 12}]),
    "4.992619967244e-01" = erlang:float_to_list(mean(StateM5B),
                                                [{scientific, 12}]),
    "2.8844011070567e-01" = erlang:float_to_list(uncache_stddev(StateX5),
                                                 [{scientific, 13}]),
    "2.8844011070567e-01" = erlang:float_to_list(uncache_stddev(StateM5A),
                                                 [{scientific, 13}]),
    "2.8844011070567e-01" = erlang:float_to_list(uncache_stddev(StateM5B),
                                                 [{scientific, 13}]),
    "-1.0857863674e-04" = erlang:float_to_list(uncache_skewness(StateX5),
                                               [{scientific, 10}]),
    "-1.0857863674e-04" = erlang:float_to_list(uncache_skewness(StateM5A),
                                               [{scientific, 10}]),
    "-1.0857863674e-04" = erlang:float_to_list(uncache_skewness(StateM5B),
                                               [{scientific, 10}]),
    "-1.199931486373e+00" = erlang:float_to_list(uncache_kurtosis(StateX5),
                                                 [{scientific, 12}]),
    "-1.199931486373e+00" = erlang:float_to_list(uncache_kurtosis(StateM5A),
                                                 [{scientific, 12}]),
    "-1.199931486373e+00" = erlang:float_to_list(uncache_kurtosis(StateM5B),
                                                 [{scientific, 12}]),
    StateX4 = add_from_list(lists:sublist(L, 10000), new()),
    StateM4_1 = add_from_list(lists:sublist(L, 5000), new()),
    StateM4_2 = add_from_list(lists:sublist(L, 5001, 5000), new()),
    StateM4A = merge(StateM4_1, StateM4_2),
    StateM4B = merge(StateM4_2, StateM4_1),
    "5.0003127683634e-01" = erlang:float_to_list(mean(StateX4),
                                                 [{scientific, 13}]),
    "5.0003127683634e-01" = erlang:float_to_list(mean(StateM4A),
                                                 [{scientific, 13}]),
    "5.0003127683634e-01" = erlang:float_to_list(mean(StateM4B),
                                                 [{scientific, 13}]),
    "2.89731623863123e-01" = erlang:float_to_list(uncache_stddev(StateX4),
                                                  [{scientific, 14}]),
    "2.89731623863123e-01" = erlang:float_to_list(uncache_stddev(StateM4A),
                                                  [{scientific, 14}]),
    "2.89731623863123e-01" = erlang:float_to_list(uncache_stddev(StateM4B),
                                                  [{scientific, 14}]),
    "-1.210219089670e-02" = erlang:float_to_list(uncache_skewness(StateX4),
                                                 [{scientific, 12}]),
    "-1.210219089670e-02" = erlang:float_to_list(uncache_skewness(StateM4A),
                                                 [{scientific, 12}]),
    "-1.210219089670e-02" = erlang:float_to_list(uncache_skewness(StateM4B),
                                                 [{scientific, 12}]),
    "-1.2014726987524e+00" = erlang:float_to_list(uncache_kurtosis(StateX4),
                                                  [{scientific, 13}]),
    "-1.2014726987524e+00" = erlang:float_to_list(uncache_kurtosis(StateM4A),
                                                  [{scientific, 13}]),
    "-1.2014726987524e+00" = erlang:float_to_list(uncache_kurtosis(StateM4B),
                                                  [{scientific, 13}]),
    StateX3 = add_from_list(lists:sublist(L, 1000), new()),
    StateM3_1 = add_from_list(lists:sublist(L, 500), new()),
    StateM3_2 = add_from_list(lists:sublist(L, 501, 500), new()),
    StateM3A = merge(StateM3_1, StateM3_2),
    StateM3B = merge(StateM3_2, StateM3_1),
    "4.92339360873928e-01" = erlang:float_to_list(mean(StateX3),
                                                  [{scientific, 14}]),
    "4.92339360873928e-01" = erlang:float_to_list(mean(StateM3A),
                                                  [{scientific, 14}]),
    "4.92339360873928e-01" = erlang:float_to_list(mean(StateM3B),
                                                  [{scientific, 14}]),
    "2.91773681207196e-01" = erlang:float_to_list(uncache_stddev(StateX3),
                                                  [{scientific, 14}]),
    "2.91773681207196e-01" = erlang:float_to_list(uncache_stddev(StateM3A),
                                                  [{scientific, 14}]),
    "2.91773681207196e-01" = erlang:float_to_list(uncache_stddev(StateM3B),
                                                  [{scientific, 14}]),
    "2.8651679541386e-02" = erlang:float_to_list(uncache_skewness(StateX3),
                                                 [{scientific, 13}]),
    "2.8651679541386e-02" = erlang:float_to_list(uncache_skewness(StateM3A),
                                                 [{scientific, 13}]),
    "2.8651679541386e-02" = erlang:float_to_list(uncache_skewness(StateM3B),
                                                 [{scientific, 13}]),
    "-1.2135372152268e+00" = erlang:float_to_list(uncache_kurtosis(StateX3),
                                                  [{scientific, 13}]),
    "-1.2135372152268e+00" = erlang:float_to_list(uncache_kurtosis(StateM3A),
                                                  [{scientific, 13}]),
    "-1.2135372152268e+00" = erlang:float_to_list(uncache_kurtosis(StateM3B),
                                                  [{scientific, 13}]),
    uniform = uncache(describe_distribution(StateX3)),
    approximately_symmetric = uncache(describe_skewness(StateX3)),
    platykurtic = uncache(describe_kurtosis(StateX3)),
    StateX2 = add_from_list(lists:sublist(L, 100), new()),
    StateM2_1 = add_from_list(lists:sublist(L, 50), new()),
    StateM2_2 = add_from_list(lists:sublist(L, 51, 50), new()),
    StateM2A = merge(StateM2_1, StateM2_2),
    StateM2B = merge(StateM2_2, StateM2_1),
    "4.706669625907499e-01" = erlang:float_to_list(mean(StateX2),
                                                   [{scientific, 15}]),
    "4.706669625907499e-01" = erlang:float_to_list(mean(StateM2A),
                                                   [{scientific, 15}]),
    "4.706669625907499e-01" = erlang:float_to_list(mean(StateM2B),
                                                   [{scientific, 15}]),
    "2.87131585305909e-01" = erlang:float_to_list(uncache_stddev(StateX2),
                                                  [{scientific, 14}]),
    "2.87131585305909e-01" = erlang:float_to_list(uncache_stddev(StateM2A),
                                                  [{scientific, 14}]),
    "2.87131585305909e-01" = erlang:float_to_list(uncache_stddev(StateM2B),
                                                  [{scientific, 14}]),
    "3.863417141553e-02" = erlang:float_to_list(uncache_skewness(StateX2),
                                                [{scientific, 12}]),
    "3.863417141553e-02" = erlang:float_to_list(uncache_skewness(StateM2A),
                                                [{scientific, 12}]),
    "3.863417141553e-02" = erlang:float_to_list(uncache_skewness(StateM2B),
                                                [{scientific, 12}]),
    "-1.04629011133886e+00" = erlang:float_to_list(uncache_kurtosis(StateX2),
                                                   [{scientific, 14}]),
    "-1.04629011133886e+00" = erlang:float_to_list(uncache_kurtosis(StateM2A),
                                                   [{scientific, 14}]),
    "-1.04629011133886e+00" = erlang:float_to_list(uncache_kurtosis(StateM2B),
                                                   [{scientific, 14}]),
    StateX1 = add_from_list(lists:sublist(L, 10), new()),
    State5 = add_from_list(lists:sublist(L, 5), new()),
    StateM1_1 = State5,
    StateM1_2 = add_from_list(lists:sublist(L, 6, 5), new()),
    StateM1A = merge(StateM1_1, StateM1_2),
    StateM1B = merge(StateM1_2, StateM1_1),
    StateM1C = merge(merge(add(lists:nth(9, L), new()),
                           add(lists:nth(5, L), new())),
                     merge(merge(add(lists:nth(2, L), new()),
                                 merge(merge(add(lists:nth(4, L), new()),
                                             add(lists:nth(6, L), new())),
                                       merge(add(lists:nth(10, L), new()),
                                             new()))),
                           merge(add(lists:nth(8, L), new()),
                                 merge(merge(add(lists:nth(7, L), new()),
                                             add(lists:nth(1, L), new())),
                                       merge(merge(new(),
                                                   new()),
                                             add(lists:nth(3, L), new())))))),
    10 = count(StateX1),
    "9.45664e-01" = erlang:float_to_list(maximum(StateX1), [{scientific, 5}]),
    "1.22548e-02" = erlang:float_to_list(minimum(StateX1), [{scientific, 5}]),
    true = (count(StateX1) =:= count(StateM1A)) andalso
           (count(StateM1A) =:= count(StateM1B)) andalso
           (count(StateM1B) =:= count(StateM1C)),
    true = (minimum(StateX1) =:= minimum(StateM1A)) andalso
           (minimum(StateM1A) =:= minimum(StateM1B)) andalso
           (minimum(StateM1B) =:= minimum(StateM1C)),
    true = (maximum(StateX1) =:= maximum(StateM1A)) andalso
           (maximum(StateM1A) =:= maximum(StateM1B)) andalso
           (maximum(StateM1B) =:= maximum(StateM1C)),
    "5.11755648024292e-01" = erlang:float_to_list(mean(StateX1),
                                                  [{scientific, 14}]),
    "5.11755648024292e-01" = erlang:float_to_list(mean(StateM1A),
                                                  [{scientific, 14}]),
    "5.11755648024292e-01" = erlang:float_to_list(mean(StateM1B),
                                                  [{scientific, 14}]),
    "5.11755648024292e-01" = erlang:float_to_list(mean(StateM1C),
                                                  [{scientific, 14}]),
    "3.4614190350208296e-01" = erlang:float_to_list(uncache_stddev(StateX1),
                                                    [{scientific, 16}]),
    "3.4614190350208296e-01" = erlang:float_to_list(uncache_stddev(StateM1A),
                                                    [{scientific, 16}]),
    "3.4614190350208296e-01" = erlang:float_to_list(uncache_stddev(StateM1B),
                                                    [{scientific, 16}]),
    "3.4614190350208296e-01" = erlang:float_to_list(uncache_stddev(StateM1C),
                                                    [{scientific, 16}]),
    "-5.5822101200515e-01" = erlang:float_to_list(uncache_skewness(StateX1),
                                                  [{scientific, 13}]),
    "-5.5822101200515e-01" = erlang:float_to_list(uncache_skewness(StateM1A),
                                                  [{scientific, 13}]),
    "-5.5822101200515e-01" = erlang:float_to_list(uncache_skewness(StateM1B),
                                                  [{scientific, 13}]),
    "-5.5822101200515e-01" = erlang:float_to_list(uncache_skewness(StateM1C),
                                                  [{scientific, 13}]),
    "-1.24910058319438e+00" = erlang:float_to_list(uncache_kurtosis(StateX1),
                                                   [{scientific, 14}]),
    "-1.24910058319438e+00" = erlang:float_to_list(uncache_kurtosis(StateM1A),
                                                   [{scientific, 14}]),
    "-1.24910058319438e+00" = erlang:float_to_list(uncache_kurtosis(StateM1B),
                                                   [{scientific, 14}]),
    "-1.24910058319438e+00" = erlang:float_to_list(uncache_kurtosis(StateM1C),
                                                   [{scientific, 14}]),
    "7.319162775309527e-01" = erlang:float_to_list(mean(State5),
                                                   [{scientific, 15}]),
    "1.5566327007068292e-01" = erlang:float_to_list(uncache_stddev(State5),
                                                    [{scientific, 16}]),
    "1.889408700207494e-01" = erlang:float_to_list(uncache_skewness(State5),
                                                   [{scientific, 15}]),
    "-1.25653596075256e+00" = erlang:float_to_list(uncache_kurtosis(State5),
                                                   [{scientific, 14}]),
    State4 = add_from_list(lists:sublist(L, 4), new()),
    "7.651283348666424e-01" = erlang:float_to_list(mean(State4),
                                                   [{scientific, 15}]),
    "1.5796982786491284e-01" = erlang:float_to_list(uncache_stddev(State4),
                                                    [{scientific, 16}]),
    "-2.64734534845686e-01" = erlang:float_to_list(uncache_skewness(State4),
                                                   [{scientific, 14}]),
    "-9.8440316642523e-01" = erlang:float_to_list(uncache_kurtosis(State4),
                                                  [{scientific, 13}]),
    State3 = add_from_list(lists:sublist(L, 3), new()),
    3 = count(State3),
    0.7779093128442551 = maximum(State3),
    0.5602545451350247 = minimum(State3),
    "7.04949698542946e-01" = erlang:float_to_list(mean(State3),
                                                  [{scientific, 14}]),
    "1.25311173304007e-01" = erlang:float_to_list(uncache_stddev(State3),
                                                  [{scientific, 14}]),
    "-7.0703087628703e-01" = erlang:float_to_list(uncache_skewness(State3),
                                                  [{scientific, 13}]),
    "-1.50000000000000e+00" = erlang:float_to_list(uncache_kurtosis(State3),
                                                   [{scientific, 14}]),
    moderately_skewed = uncache(describe_skewness(State3)),
    platykurtic = uncache(describe_kurtosis(State3)),
    % Gamma family distributions tests
    GammaRate = 0.25,
    % Exponential distribution from random values
    StateG1 = add_from_list(fun(X0) ->
        -math:log(X0) / GammaRate
    end, L, new()),
    true = round(mean(StateG1)) =:= round(1 / GammaRate),
    true = round(uncache_skewness(StateG1)) =:= 2,
    true = round(uncache_kurtosis(StateG1)) =:= 6,
    exponential = uncache(describe_distribution(StateG1)),
    % Erlang-2 distribution from random values
    ErlangK = 2,
    StateG2 = add_from_list(fun(X0, X1) ->
        -(math:log(X0) + math:log(X1)) / GammaRate
    end, L, new()),
    true = round(mean(StateG2)) =:= round(ErlangK / GammaRate),
    true = round(uncache_skewness(StateG2) * 100) =:=
           round(200 / math:sqrt(ErlangK)),
    true = round(uncache_kurtosis(StateG2)) =:= (6 div ErlangK),
    gamma_family = uncache(describe_distribution(StateG2)),
    % Log-Normal family distributions tests
    NormalL = normal_distribution(),
    StateLN1 = add_from_list(NormalL, new()),
    normal = uncache(describe_distribution(StateLN1)),
    StateLN2 = add_from_list(fun(X0) ->
        % base-e exponential function
        math:exp(X0)
    end, NormalL, new()),
    log_normal_family = uncache(describe_distribution(StateLN2)),
    ok.

-endif.
